/*
 *M*****************************************************************************
 *M*  Product:  XEROX OSG/OSCG/EIDC S2X Board                                  *
 *M*****************************************************************************
 *M*                                                                           *
 *M*                    COPYRIGHT (C) 2001 XEROX CORPORATION                   *
 *M*                    ALL RIGHTS RESERVED                                    *
 *M*                                                                           *
 *M*****************************************************************************
 *M*
 *M*    MODULE NAME  :  jbig2.c
 *M*
 *M*    PURPOSE      :  JBIG2 compression
 *M*
 *M*    DESCRIPTION  :  processes a line of input image and performs JBIG
 *M*							compression
 *M*
 *******************************************************************************
 **     AUTHOR IDENTITY                                                        *
 *******************************************************************************
 **
 **  Initials  Name
 **  --------  ------------------------------
 **  mdd       Marlin Diehl
 **   nl       Nathan Lund
 **   rn       Ramesh Nagarajan
 **   pt       Phil Turkovitz
 **
 *******************************************************************************
 **     SVCS MODULE REVISION HISTORY:                                          *
 *******************************************************************************
 **  Revision    Date    Author  Description
 **  --------  --------  ------  ----------------------
 **  0.1			02.01.21  ??      
 *******************************************************************************
 */

/*****************/
/* Include files */
/*****************/
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <time.h>
#include "jbig2.h"

// SILX headers
#include "dgp_err_handle.h"
#include "silx_encode.h"
/*********************/
/* Constant #defines */
/*********************/


#define	PRINTF(x)		//printf x

/*******************************/
/* Declaration specific to PC  */
/*******************************/
#define PCVERSION
#ifdef PCVERSION
	#define  inline
	#define  restrict
	#define  if_prob(a)
	typedef  unsigned int	n32;
#endif


/*********************/
/* Global Variables  */
/*********************/
#define JB2HEADERSIZE 8
static Dgp_UInt8 jb2header[] = { 0x97, 0x4a, 0x42, 0x32, 0x0d, 0x0a, 0x1a, 0x0a };

/*F*****************************************************************************
 *F*
 *F*  Function Name:  compressOneRow
 *F*
 *F*  Operation:      Compresses rows of input image using jbig2 compression.
 *F*
 *F*  Function
 *F*  Returns:        Nothing
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  02.01.21		??		
 *F****************************************************************************/
static void compressOneRow (uint8 *restrict imgptr, JBIG2State *restrict state, int rows)
{
	uint8	*restrict curline;
	int   iii, jj;
 	int	bytesperline = state->bytesperline;


	for (iii=0; iii<rows; iii++) {
		PRINTF (("<compressOneRow> LINE %d processing...\n", iii));

		// Process each line of input image and put the encoded bitstream to state->outCurAddress
		// Currently output is the same as the input image

		curline = &imgptr[iii*bytesperline];

		for (jj=0; jj<bytesperline; jj++)
			(*state->outCurAddress++) = curline[jj];

		state->compressedSize += bytesperline;

	}
}




/*F*****************************************************************************
 *F*
 *F*  Function Name:  jbigEncode
 *F*
 *F*  Operation:      entry point to JBIG2 encoding algorithm
 *F*
 *F*  Function
 *F*  Returns:        Nothing
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  02.01.21		??
 *F****************************************************************************/
int jbigEncode (uint8 *restrict imgptr, JBIG2State *restrict state, int rows)
{ 

	Dgp_Int32 ret = cSuccessful;

	switch (state->processingState) {
	case initialize:    // initilaizing and processing the firstBand
        {
            Dgp_State* dgp_state = NULL;
            SilxEncCookie sCookie = NULL;
            int i;
            /* Init dgp state */    
            ret = Dgp_initDgpState(&dgp_state, 
                    NULL,
                    NULL,
                    stderr, 
                    Dgp_fileErrPrinter, 
                    DGP_TRUE, 
                    DGP_TRUE);
            ret += silx_initEncState(dgp_state, &sCookie);

            // write jbig2header to buffer
            for(i = 0; i< JB2HEADERSIZE; i++){
                *(state->outCurAddress)= jb2header[i];
                state->outCurAddress++;
            }
            /* Write 0x00: random-access organisation, page count known */
            /* Write 0x01: sequential organisation, page count known */
            /* Write 0x02: random-access organisation, page count not known */
            /* Write 0x03: sequential organisation, page count not known */
            *(state->outCurAddress) = 0x03;
            state->outCurAddress++;
            if ( ret  != cSuccessful){
		        if(dgp_state)	
                    Dgp_cleanupDgpState((Dgp_State **)(&dgp_state));
                if(sCookie) 
                    silx_cleanEncState(dgp_state, &sCookie);
                return 0;
            }
            state->dgp_state  = (void* )dgp_state;
            state->enc_cookie = (void* )sCookie;
		    PRINTF (("<jbigEncode> Initialize\n"));
        }
	    break;

	case process:
        {
            int lastBufInPage;
            //encode one buffer at a time
            if(state->bandNumber >= state->totalBandCnt)
                lastBufInPage = 1;
            else 
                lastBufInPage = 0;
            // encode one buffer
            ret = silx_encMonoBufInPage(
                     state->dgp_state, 
                     state->enc_cookie, 
                     imgptr,            // input buffer
                     state->pixels,     // buffer width (same for  page width)
                     rows,              // buffer Height
                     state->bytesperline,
                     state->scanlines,  // height of the page
                     state->bandNumber -1,
                     lastBufInPage);
            state->compressedSize += state->bytesperline * rows;
            // finish and get the data
            if(lastBufInPage){
                uint8* dictData = NULL, *pgData = NULL;
                uint32 dictDataSize = 0, pgDataSize = 0;
                ret += silx_encMonoBufGetPGData(
                     state->dgp_state, 
                     state->enc_cookie,
                     &dictData,   
                     &dictDataSize,
                     &pgData,
                     &pgDataSize);
	            //copy dict
                if(dictDataSize!=0 && dictData){
                    memcpy(state->outCurAddress, dictData, dictDataSize);
                    state->outCurAddress +=dictDataSize;
                    free(dictData);
                }
                if(dictDataSize!=0 && dictData){
                    memcpy(state->outCurAddress, pgData, pgDataSize);
                    state->outCurAddress += pgDataSize;
                    free(pgData);
                }
            }
		    PRINTF (("<jbigEncode> Process\n"));
		    //compressOneRow (imgptr, state, rows); skip this function
        }
		break;
	
	case finish:
        //clean up 
        silx_cleanEncState(state->dgp_state, (SilxEncCookie*)(&(state->enc_cookie)));
        Dgp_cleanupDgpState((Dgp_State**)(&(state->dgp_state)));

		//Any completion work
		PRINTF (("<jbigEncode> Finish\n"));
		break;
	}

	return 1;

}
