#include "stdafx.h"

#include <math.h>

#include "Buffer.h"
#include "misc.h"
#include "debug.h"


static Stopwatch DownBufFlushTmr( "DownBuf::flushData" );
static Stopwatch BufferWriteTmr( "Buffer::write" );
static Stopwatch IsSolidTmr( "isSolid" );

Buffer::~Buffer( void )
{
    DbgStream() << BufferWriteTmr << std::endl;
    delete [] m_buf; m_buf = NULL;
}

void Buffer::setSize( size_t newSize )
{
    if( size() != newSize )
    {
        delete [] m_buf;
        m_buf = new unsigned char [newSize];
        m_size = newSize;
    }

    m_idx = 0;
}

void Buffer::write( unsigned char * buf, size_t count )
{
    BufferWriteTmr.start();

    DBG_TRACE_OBJ( "Buffer::write" );
    DBG_WATCH( "count", count );

    size_t bufByte = 0;

    size_t total = m_idx + count;

    if( (m_idx == 0) && (count >= size()) )
    {
        m_idx = size();
        BufferWriteTmr.stop();
        flushData( buf );
        BufferWriteTmr.start();
        count -= size();

        if( count )
        {
            write( buf + size(), count );
        }
    }
    else 
    {
        if( total < size() )
        {
            // We will not fill the buffer, so just copy the data and
            // advance the index.
            memcpy( &(m_buf[m_idx]), buf, count );
            m_idx += count;
        }
        else //thus( total >= size() )
        {
            // We have enough or too much data.

            // Figure out how much we need to fill the buffer.
            size_t toEnd = size() - m_idx;

            // Copy just enough data and adjust index accordingly.
            // call flushData().
            memcpy( &(m_buf[m_idx]), buf, toEnd );
            m_idx += toEnd;
            BufferWriteTmr.stop();
            flushData( m_buf );
    
            BufferWriteTmr.start();

            count -= toEnd;

            if( count )
            {
                // Now, call ourself recursively for the remaining data.
                write( buf + toEnd, count );
            }
        }
    }

    BufferWriteTmr.stop();
}

DownBuf::DownBuf( size_t fact, size_t xRes, Buffer * output )
: m_factor( fact ),
  m_xRes( xRes ),
  m_output( output )
{
    setSize( m_xRes * m_factor * 3 ); //3 bytes per pixel!
}

DownBuf::~DownBuf( void )
{
    DbgStream() << DownBufFlushTmr  << std::endl;
}

void DownBuf::flushData( unsigned char * data )
{
    DownBufFlushTmr.start();

    DBG_TRACE_OBJ( "DownBuf::flushData" );

    // Is there anything to do?
    if( m_idx == 0 )
    {
        DownBufFlushTmr.stop();
        return;
    }

    // If the output is NULL, we just swallow what we were given.
    if( !m_output )
    {
        DBG_MSG( "NULL m_output." );
        m_idx = 0;
        DownBufFlushTmr.stop();
        return;
    }

    // The number of bytes in a row of the full-resolution image.
    const size_t bytesRow = (m_xRes * 3);

    // downRows will usually be m_factor, but it may be something
    // else if we are at the bottom of the page.
    const size_t downRows = m_idx / bytesRow;

    // However, it is a mistake to call this method with a non-integeer
    // number of rows!
    assert( !(m_idx % bytesRow) );

    // lines[] is an array of pointers to each row that we are downsampling.
    // These will move through the image in parallel as we process the image.
    Ptr24Bit * lines = new Ptr24Bit[downRows];

    {
        Ptr24Bit ptr( data );

        for( size_t ii=0; ii < downRows; ++ii, ptr.offset(bytesRow) )
        {
            lines[ii] = ptr;
        }
    }

    // Note the trucation.  downXRes is the number of full down sampled
    // columns.  We have an extra loop at the bottom to deal with any
    // extra columns at the end.
    const size_t downXRes = (m_xRes/m_factor);

    // The number of columns not counted for by the truncated integer division
    // used to calculate downXRes.
    const size_t extraCols = m_xRes % m_factor;

    // The "real" downsample resolution, including an extra column if needed.
    const size_t realDownXRes = downXRes + !!extraCols;    
   
    // Get a pointer to the first scan-line in the buffer.  This
    // is where we will write the down-sampled data.  (We are
    // downsampling in place.)
    Ptr24Bit line0( data );

    const size_t downCols = m_factor;

    // For each pixel in the scan-line.  The += 3 is 3 bytes for
    // 24 bit color.
    {for( size_t ii = 0; ii < downXRes; ++ii, ++line0 )
    {        
        // The number of non-white pixels we see in our downsampling
        // area.
        size_t ct = 0;

        // The sum of the red, blue, and gree components from the care
        // pixels.
        unsigned int sumR = 0;
        unsigned int sumG = 0;
        unsigned int sumB = 0;

        // Loop over all of the pixels we are down sampling.  Count and 
        // sum the care pixels.
        for( size_t row = 0; row != downRows; ++row )
        {
            for( size_t col = 0; col != downCols; ++col, ++lines[row] )
            {
                if( !lines[row].isDontCare() )
                {
                    ++ct;
                    sumR += lines[row].r();
                    sumG += lines[row].g();
                    sumB += lines[row].b();
                }
            }
        }

        // Place the average care count into the don't care pixels.
        if( ct )
        {
            switch( ct )
            {
                case 1:  break;

                case 2:  sumR >>=  1; sumG >>=  1;  sumB >>=  1; break;
                case 4:  sumR >>=  2; sumG >>=  2;  sumB >>=  2; break;
                case 8:  sumR >>=  3; sumG >>=  3;  sumB >>=  3; break;
                case 16: sumR >>=  4; sumG >>=  4;  sumB >>=  4; break;

                default: sumR  /= ct; sumG  /= ct;  sumB  /= ct; break;
            };

            line0.setRGB( sumR, sumG, sumB );

            // We must be very careful that we do not create a don't care
            // pixel because of truncation.  For example, if four pixels
            // all of saturates blue but 2 have 1 count of red and no green
            // while 2 have 1 count of green and no read, we end up with
            // saturated blue.
            line0.forceCare();        
        }
        else
        {
            line0.markDontCare();
        }
        
    }}
   
    if( extraCols )
    {
        // The number of non-white pixels we see in our downsampling
        // area.
        size_t ct = 0;

        // The sum of the red, blue, and gree components from the care
        // pixels.
        unsigned int sumR = 0;
        unsigned int sumG = 0;
        unsigned int sumB = 0;

        // Loop over all of the pixels we are down sampling.  Count and 
        // sum the non-white ones.
        for( size_t row = 0; row != downRows; ++row )
        {
            for( size_t col = 0; col != extraCols; ++col, ++lines[row] )
            {
                if( !lines[row].isDontCare() )
                {
                    ++ct;
                    sumR += lines[row].r();
                    sumG += lines[row].g();
                    sumB += lines[row].b();
                }
            }
        }

        // Place the average non-white value in the output.

        if( ct )
        {
            switch( ct )
            {
                case 1:  break;
                case 2:  sumR >>=  1; sumG >>=  1;  sumB >>=  1; break;
                case 4:  sumR >>=  2; sumG >>=  2;  sumB >>=  2; break;
                case 8:  sumR >>=  3; sumG >>=  3;  sumB >>=  3; break;
                case 16: sumR >>=  4; sumG >>=  4;  sumB >>=  4; break;
                default: sumR  /= ct; sumG  /= ct;  sumB  /= ct; break;
            };

            line0.setRGB( sumR, sumG, sumB );

            // We must be very careful that we do not create a don't care
            // pixel because of truncation.  For example, if four pixels
            // all of saturates blue but 2 have 1 count of red and no green
            // while 2 have 1 count of green and no read, we end up with
            // saturated blue.
            line0.forceCare();
        }
        else
        {
            line0.markDontCare();
        }
    }

    DownBufFlushTmr.stop();

    m_output->write( data, 3 * realDownXRes );

    delete [] lines;

    m_idx = 0;
}

void StorageBuf::write( unsigned char * buf, size_t count )
{
    DBG_TRACE_OBJ( "StorageBuf::write" );
    DBG_WATCH( "count", count );

    size_t bufByte = 0;

    size_t total = m_idx + count;

    if( total <= size() )
    {
        // We will not fill the buffer, so just copy the data and
        // advance the index.
        memcpy( &(m_buf[m_idx]), buf, count );
        m_idx += count;
        DBG_WATCH( "Storage buffer at ", m_idx );
    }
    else // thus( total > size() )
    {
        // We have too much data.

        // Figure out how much we need to fill the buffer.
        size_t toEnd = size() - m_idx;

        // Copy just enough data and adjust index accordingly.
        memcpy( &(m_buf[m_idx]), buf, toEnd );
        m_idx = size();
        DBG_WATCH( "Storage buffer overfow by ", count - toEnd );
		m_overflow = TRUE;
    }
}

void StorageBuf::fillWhite ()
{
	memset (m_buf, 0, size());
	m_idx = size();
}

ColorStorageBuf::~ColorStorageBuf( void )
{
    DbgStream() << IsSolidTmr << std::endl;
}

bool ColorStorageBuf::isSolid( const unsigned char val ) const
{
    const size_t dwords = index() / sizeof( unsigned int );
    
    // This is an optimization.  We get over 2x speed improvement for
    // large images if we use integers instead of bytes here for as much
    // of the image as possible.  Of course for really tiny images
    // we take a speed hit here because of the overhead involved in
    // cacluating the number of whole words.
    if( dwords )
    {
        unsigned int tmp;

        memset( &tmp, val, sizeof( tmp ) );
            
        const unsigned int intVal = tmp;

        unsigned int * dwPtr = reinterpret_cast<unsigned int*>(m_buf);

        unsigned int * const dwEnd = dwPtr + dwords;

        for( ; (dwPtr != dwEnd) && (*dwPtr == intVal); ++dwPtr ) {}

        if( dwPtr != dwEnd )
        {
            return false;
        }
    }
    
    const size_t bytes = index() % sizeof( unsigned int );

    unsigned char * bytePtr = m_buf + (dwords * sizeof(unsigned int));

    unsigned char * const byteEnd = bytePtr + bytes;

    for(; (bytePtr != byteEnd) && (*bytePtr == val); ++bytePtr ) {}
 
    return bytePtr == byteEnd;
}
