#ifndef DEBUG_H_INCLUDED
#define DEBUG_H_INCLUDED

#include <time.h>

void DbgTrace( const std::string & func,
               const std::string & file,
               const int line,
               const void * thisPtr );

std::ostream & DbgStream( void );

std::string indent( void );

void IncrIndent( void );
void DecrIndent( void );


#if 0

#define DBG_TRACE( f ) const std::string FUNC(f); DbgTrace( FUNC, __FILE__, __LINE__, NULL ); IncrIndent();
#define DBG_TRACE_OBJ( f ) const std::string FUNC(f); DbgTrace( FUNC, __FILE__, __LINE__, (void *)(this) ); IncrIndent();

#define DBG_ERROR( msg ) DbgStream() << indent() << "ERROR --> " << (msg) << std::endl;

#define DBG_MSG( msg ) DbgStream() << indent() << (msg) << std::endl;

#define DBG_WATCH( expr, val ) DbgStream() << indent() << (expr) << " == " << (val) << std::endl;

#define DBG_RET DecrIndent(); DbgStream() << indent() << "<-- Leaving " << FUNC << std::endl;

#define DBG_RETVAL( val ) DecrIndent(); DbgStream() << indent() << "<-- Returning " << (val) << " from " << FUNC << std::endl;

#else

#define DBG_TRACE( f )
#define DBG_TRACE_OBJ( f )

#define DBG_MSG( msg )

#define DBG_ERROR( msg )

#define DBG_WATCH( expr, val )

#define DBG_RET 

#define DBG_RETVAL( val )

#endif


class Stopwatch
{
public:

    Stopwatch( const std::string & n = "anon" ) : m_name(n) { reset(); }

    double read( void )
    {
        if( m_running )
        {
            const clock_t check = clock();
            m_elapsed          += check - m_lastCheck;
            m_lastCheck         = check;
        }

        return double(m_elapsed)/double(CLOCKS_PER_SEC);
    }

    void start( void )
    { 
        if( !m_running )
        {
            m_running   = true;
            m_lastCheck = clock();
        }
    }
    
    void stop( void )
    {        
        if( m_running )
        {
            m_elapsed += clock() - m_lastCheck;
            m_running = false;
        }
    }

    void reset( void )
    {
        m_running   = false;
        m_elapsed   = 0;
        m_lastCheck = 0;
    }

    double lap( void )
    {
        double lapTime = 0.0;

        if( m_running )
        {
            const clock_t delta = clock() - m_lastCheck;
            lapTime = double(delta)/double(CLOCKS_PER_SEC);
            m_lastCheck += delta;
            m_elapsed   += delta;
        }

        return lapTime;
    }

    bool isRunning( void ) const { return m_running; }
        
    const std::string & name( void ) const { return m_name; }

private:

    bool m_running;

    clock_t m_elapsed;
    clock_t m_lastCheck;

    std::string m_name;
};

inline std::ostream & operator<<( std::ostream & lhs, Stopwatch & rhs )
{      
    if( rhs.isRunning() )
    {
        lhs << '(' << rhs.name() << ") lap: " << rhs.lap();
        lhs << "   total: " << rhs.read();
    }
    else
    {
        lhs << '(' << rhs.name() << ") total: " << rhs.read();
    }

    return lhs;
}

#endif