#ifndef OemBase_h_INCLUDED
#define OemBase_h_INCLUDED

#include <set>

#include <PRCOMOEM.H>

#define OEM_SIGNATURE   (DWORD('MSFT'))
#define OEM_VERSION     (DWORD(0x00000001L))

/*

  CustomGdiSurface serves primarily as a wrapper that ties together a
  SURFOBJ with its accompanying bitmap handle.

*/
class CustomGdiSurface
{
public:

    CustomGdiSurface( void )
    : m_pSurf(NULL), m_hBmp(NULL) {}

    virtual ~CustomGdiSurface( void ) { Destroy(); }
    
    void Create( SURFOBJ * pSrc, const size_t hRes, const size_t vRes );
    void Destroy( void );

    void take( CustomGdiSurface & src )
    {
        this->m_pSurf = src.m_pSurf;
        this->m_hBmp  = src.m_hBmp;

        src.m_pSurf = NULL;
        src.m_hBmp  = NULL;    
    }

    size_t rows( void ) const { return m_pSurf? m_pSurf->sizlBitmap.cy : 0; }
    size_t cols( void ) const { return m_pSurf? m_pSurf->sizlBitmap.cx : 0; }

    size_t delta( void ) const { return m_pSurf? m_pSurf->lDelta : 0; }

    unsigned char * data( void )
    {
        return m_pSurf? 
            reinterpret_cast<unsigned char *>(m_pSurf->pvScan0) : NULL;
    }

    SURFOBJ * surf( void ) { return m_pSurf; }

    void fill( const unsigned char val )
    {
        if( data() )
        {
            memset( data(), val, rows() * delta() );
        }
    }

private:

    CustomGdiSurface( const CustomGdiSurface & );//disabled.
    CustomGdiSurface & operator=( const CustomGdiSurface & );//disabled

    SURFOBJ * m_pSurf;
    HBITMAP m_hBmp;
};

/*
  First off.  You really need to have access to the Microsoft Driver
  Development Kit to make any sense out of IOemBase (or its 
  derived classes).  If you don't this will probably seem like
  Sanskrit.  I have assumed that you have access to the documentation
  that comes with said DDK, and have not reexmplained the meanings
  of classes or methods defined there.
  
  IOemBase is our base implementation of the IPrintOemUni interface.
  It contains a whole lot of simple implementation that should be
  standard for all implementations of IPrintOemUni.  For example,
  we have implemented the IUnknown portion of the interface.  We
  have also implemented a bunch of stuff that should be similar for
  most printer drivers.  For example, we keep track of the page
  number, stash the state given to us by GDI in EnableDriver() and
  StartDoc(), etc.

  Perhaps the most important thing we have done, however, is that we
  have completely encapsulated the entire hooking and punting
  mechanism as virtual methods on our new base class.  The file DdiHook.cpp
  contains all of the possible global hook methods.  These have been
  implemented to turn around and call virtual methods with the same name
  on IOemBase.  The default implementations of these methods typically
  just turn around and punt back to GDI.

  Although this scheme may seem elaborate, and it does add extra overhead,
  it also makes hooking GDI calls much easier.  All you have to do is
  override the virtual method.

  Finally, I have chosen an architecture in which IPrintOemUni is the PDEV
  object passed back to GDI in EnablePDEV.  This is necessary (given the
  current architecture) to make the hooking mechanism work.  You can add
  data members to your derived class, of course, but you must use your
  IOemBase derivative as the PDEV object.

*/

class IOemBase : public IPrintOemUni
{
public:

    IOemBase( void );

    virtual ~IOemBase( void );

    //////////////////////////////////////////////////////////////////////
    // IUnknown methods (See Windows VC++ SDK documentation.)
    //
    
    STDMETHOD(QueryInterface) (THIS_ REFIID riid, LPVOID FAR* ppvObj);
    STDMETHOD_(ULONG,AddRef)  (THIS);
    STDMETHOD_(ULONG,Release) (THIS);

    //////////////////////////////////////////////////////////////////////
    // IPrintOemUni methods (See Win 2000 Driver DDK documentation.)
    // 
 
    // Implemented

    virtual HRESULT __stdcall GetImplementedMethod( PSTR );
    virtual HRESULT __stdcall GetInfo( DWORD, PVOID, DWORD, PDWORD );
    virtual HRESULT __stdcall PublishDriverInterface( IUnknown* );  

    virtual HRESULT __stdcall  EnableDriver( DWORD, DWORD, PDRVENABLEDATA );
    virtual HRESULT __stdcall DisableDriver( void );

    virtual HRESULT __stdcall  EnablePDEV( PDEVOBJ, PWSTR, ULONG, HSURF*, ULONG, GDIINFO*, ULONG, DEVINFO*, DRVENABLEDATA*, OUT PDEVOEM* );
    virtual HRESULT __stdcall   ResetPDEV( PDEVOBJ, PDEVOBJ );
    virtual HRESULT __stdcall DisablePDEV( PDEVOBJ );   

    // Not implemented.

    // None of these "need" to be implemented (beyond returning E_NOTIMPL.)
    // I haven't needed to implement any of these for any of my ventures into
    // derived classes.  I fyou need to, they are described in the DDK
    // documentation.  All you you have to remember for your contract with
    // IOemBase is to add the appropriate implementMethod() call in your
    // constructor.

    virtual HRESULT __stdcall DevMode( DWORD, POEMDMPARAM )               { return E_NOTIMPL; }
    virtual HRESULT __stdcall CommandCallback( PDEVOBJ, DWORD, DWORD,
                              PDWORD, OUT INT* )                          { return E_NOTIMPL; }
    
    virtual HRESULT __stdcall Compression( PDEVOBJ, PBYTE, PBYTE, DWORD,
                              DWORD, OUT INT* )                           { return E_NOTIMPL; }
    
    virtual HRESULT __stdcall DownloadCharGlyph( PDEVOBJ, PUNIFONTOBJ,
                              HGLYPH, PDWORD, OUT DWORD* )                { return E_NOTIMPL; }
    
    virtual HRESULT __stdcall DownloadFontHeader( PDEVOBJ, PUNIFONTOBJ,
                              OUT DWORD* )                                { return E_NOTIMPL; }
    
    virtual HRESULT __stdcall DriverDMS( PVOID, PVOID, DWORD, PDWORD )    { return E_NOTIMPL; }
    virtual HRESULT __stdcall FilterGraphics( PDEVOBJ, PBYTE, DWORD )     { return E_NOTIMPL; }
    virtual HRESULT __stdcall HalftonePattern( PDEVOBJ, PBYTE, DWORD,
                              DWORD, DWORD, DWORD, PBYTE, DWORD )         { return E_NOTIMPL; }
    
    virtual HRESULT __stdcall ImageProcessing( PDEVOBJ, PBYTE,
                              PBITMAPINFOHEADER, PBYTE, DWORD, PIPPARAMS,
                              OUT PBYTE *)                                { return E_NOTIMPL; }
    
    virtual HRESULT __stdcall MemoryUsage( PDEVOBJ, POEMMEMORYUSAGE )     { return E_NOTIMPL; }
    virtual HRESULT __stdcall OutputCharStr( PDEVOBJ, PUNIFONTOBJ, DWORD,
                              DWORD, PVOID )                              { return E_NOTIMPL; }

    virtual HRESULT __stdcall SendFontCmd( PDEVOBJ, PUNIFONTOBJ,
                              PFINVOCATION )                              { return E_NOTIMPL; }

    virtual HRESULT __stdcall TextOutAsBitmap( SURFOBJ*, STROBJ*,
                              FONTOBJ*, CLIPOBJ*, RECTL*, RECTL*,
                              BRUSHOBJ*, BRUSHOBJ*, POINTL*, MIX )        { return E_NOTIMPL; }

    virtual HRESULT __stdcall TTDownloadMethod( PDEVOBJ, PUNIFONTOBJ,
                              OUT DWORD* )                                { return E_NOTIMPL; }
    
    virtual HRESULT __stdcall TTYGetInfo( PDEVOBJ, DWORD, PVOID, DWORD,
                               DWORD* )                                   { return E_NOTIMPL; }

    //////////////////////////////////////////////////////////////////////
    // DDI Hook Rendering methods (Implemented)
    //

    // These 6 mehtods are the basic control methods for printing.  The
    // rest of the calls are all to do with actual drawing, fonts, and
    // printer escapes.  These 6 methods, however, will form the core of
    // any driver.  These 6 methods are also special because the default
    // implementations here manage some generic internal state.  Thus,
    // when implementing any of these methods, you should usually call 
    // the base class version.  This will ensure that m_curPage, m_docName,
    // m_jobId, and m_isBanded are all set correctly.  Typically you will
    // want to call them as your very first operation, although there 
    // may be some exceptions.  You can then use these protected variables
    // to determine the state of the print job.

    virtual BOOL     DrvStartDoc( SURFOBJ *, PWSTR, DWORD ); 
    virtual BOOL    DrvStartPage( SURFOBJ * ); 
    virtual BOOL     DrvSendPage( SURFOBJ * );
    virtual BOOL DrvStartBanding( SURFOBJ *, POINTL * ); 
    virtual BOOL     DrvNextBand( SURFOBJ *, POINTL * );
    virtual BOOL       DrvEndDoc( SURFOBJ *, FLONG );
    
    // Instead of punting, I have stubbed this method out to return 0 always.
    // That indicates that no escape sequences are supported, which seems
    // to be the correct default behaviour.

    virtual ULONG DrvEscape( SURFOBJ *, ULONG, ULONG, PVOID, ULONG, PVOID ); 

    //////////////////////////////////////////////////////////////////////
    // DDI Hook Rendering methods (punted)
    //

    virtual BOOL         DrvAlphaBlend( SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, RECTL *, RECTL *, BLENDOBJ * );
    virtual BOOL         DrvBitBlt( SURFOBJ *, SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, RECTL *, POINTL *, POINTL *, BRUSHOBJ *, POINTL *, ROP4 );
    virtual BOOL         DrvCopyBits( SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, RECTL *, POINTL * );
    virtual ULONG        DrvDitherColor( DHPDEV, ULONG, ULONG, ULONG * );
    virtual BOOL         DrvFillPath( SURFOBJ *, PATHOBJ *, CLIPOBJ *, BRUSHOBJ *, POINTL *, MIX, FLONG );
    virtual ULONG        DrvFontManagement( SURFOBJ *, FONTOBJ *, ULONG, ULONG, PVOID, ULONG, PVOID ); 
    virtual ULONG        DrvGetGlyphMode( DHPDEV, FONTOBJ * );
    virtual BOOL         DrvGradientFill( SURFOBJ *, CLIPOBJ *, XLATEOBJ *, TRIVERTEX *, ULONG, PVOID, ULONG, RECTL *, POINTL *, ULONG );
    virtual BOOL         DrvLineTo( SURFOBJ *, CLIPOBJ *, BRUSHOBJ *, LONG, LONG, LONG, LONG, RECTL *, MIX ); 
    virtual BOOL         DrvPlgBlt( SURFOBJ *, SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, COLORADJUSTMENT *, POINTL *, POINTFIX *, RECTL *, POINTL *, ULONG ); 
    virtual BOOL         DrvQueryAdvanceWidths( DHPDEV, FONTOBJ *, ULONG, HGLYPH *, PVOID, ULONG ); 
    virtual PIFIMETRICS  DrvQueryFont( DHPDEV, ULONG, ULONG, ULONG * );
    virtual LONG         DrvQueryFontData( DHPDEV, FONTOBJ *, ULONG, HGLYPH, GLYPHDATA *, PVOID, ULONG ); 
    virtual PVOID        DrvQueryFontTree( DHPDEV, ULONG, ULONG, ULONG, ULONG * );
    virtual BOOL         DrvRealizeBrush( BRUSHOBJ *, SURFOBJ *, SURFOBJ *, SURFOBJ *, XLATEOBJ *, ULONG ); 
    virtual BOOL         DrvStretchBlt( SURFOBJ *, SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, COLORADJUSTMENT *, POINTL *, RECTL *, RECTL *, POINTL *, ULONG );
    virtual BOOL         DrvStretchBltROP( SURFOBJ *, SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, COLORADJUSTMENT *, POINTL *, RECTL *, RECTL *, POINTL *, ULONG, BRUSHOBJ *, ROP4 ); 
    virtual BOOL         DrvStrokeAndFillPath( SURFOBJ *, PATHOBJ *, CLIPOBJ *, XFORMOBJ *, BRUSHOBJ *, LINEATTRS *, BRUSHOBJ *, POINTL *, MIX, FLONG ); 
    virtual BOOL         DrvStrokePath( SURFOBJ *, PATHOBJ *, CLIPOBJ *, XFORMOBJ *, BRUSHOBJ *, POINTL *, LINEATTRS *, MIX );
    virtual BOOL         DrvTextOut( SURFOBJ *, STROBJ *, FONTOBJ *, CLIPOBJ *, RECTL *, RECTL *, BRUSHOBJ *, BRUSHOBJ *, POINTL *, MIX );
    virtual BOOL         DrvTransparentBlt( SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, RECTL *, RECTL *, ULONG , ULONG ); 


    // Returns true if the brush is neither black nor white.
    static bool brushHasColor( const BRUSHOBJ * br )
    {
        // Note that the test of br first will cause evaluation to stop
        // if br is NULL.
        return br && (br->iSolidColor != 0) && (br->iSolidColor != 0xffffff);
    }

protected:

    // This is needed so that we can implement PublishDriverInterface()
    // and call DrvWriteSpoolBuf().
    //
    IPrintOemDriverUni * m_pOEMHelp;

    //
    // Values stashed from calls to EnablePDEV() and EnableDriver().
    // IOemBase does not actually use these, but we dutifully stash
    // these values in our structure so that derived classes can get at
    // them.  I felt that this was preferable than having everybody who
    // creates a derived class implement EnableDriver and make their own
    // versions of these variables.
    //
    
    DWORD m_driverVersion;
    std::basic_string<WCHAR> m_printerName;

    GDIINFO m_gdiInfo;
    DEVINFO m_devInfo;

    //
    // These are the values passed to DrvStartDoc().  Once again, IOemBase
    // does not actually use these, but I felt that storing them in protected
    // variables was better than making everybody implement DrvStartDoc().
    //

    std::basic_string<WCHAR> m_docName;
    DWORD m_jobId;

    // Used by DrvStartBanding() and DrvNextBand() to keep track of where
    // we are on the page.
    POINTL m_bandOrigin;
    
    // This gets initialized to false in DrvStartDoc() and then set to
    // true if DrvStartBanding() is called.  It is used by DrvSendPage()
    // to tell whether or not the page needs to be sent to the printer.
    bool m_isBanded;

    // This is just a convenient way of keeping track of what page we are
    // on.  Pages start at 1, so 0 indicates we are not priting.
    size_t m_curPage;

    
    // This should never be called anywhere other than in the constructor
    // of either IOemBase or a class decended from IOemBase.
    // This is how you tell IOemBase that you have implemented a method
    // so that GetImplementedMethod() will do the right thing.  You 
    // should call implementMethod() in your constructor once for each
    // method that you have implemented.  (Don't worry about whether you
    // are overriding a base class method or not.  The values are stored
    // in a std::set, so inserting an element twice is a noop.)
    void implementMethod( const std::basic_string<CHAR> & );

    // Used for implementation of ResetPDEV().  Alows initialization of
    // one IOemBase from another of the same kind.  Note that the object
    // passed to this must be able to dynamic_cast itself to the same
    // derived class as this.  By default we fail.  However, the
    // base class version must be called by any derived classes because
    // it handles copy of IOemBase specific information.
    virtual HRESULT initFrom( IOemBase * );

    // Returns the color depth for the current printing mode.
    const size_t bitsPixel( void ) const;

    std::string getSetting( PDEVOBJ pdevobj, const std::string & setting );

    // This is a utility that turns a null-terminated array of WCHARs
    // into std::basic_string<WCHAR>.
    static std::basic_string<WCHAR> makeStr( const WCHAR * );

    // This function calculates how far a surface stretches past the bottom
    // of a page.  I only ever imagine that it will be called in
    // DrvNextBand().  The calculation is simple, but it occurs exactly
    // the same way in every implementation of DrvNextBand() that I have
    // ever written.  The value returned will be negative if there is space
    // remaining in on the page after the current band, 0 if this band goes
    // to the bottom of the page exactly, and positive, if this band
    // overruns the bottom of the page.
    long bandOverrun( SURFOBJ * ) const;

    // bandRows() will probably always be called right after bandOverrun().
    // It exists for a similar reason to bandOverrun().  It's a simple
    // caculation done almost all the time.
    size_t bandRows( SURFOBJ * ) const;

    // We often need black or white brush objects around.
    static BRUSHOBJ m_whiteBr;
    static BRUSHOBJ m_blackBr;

private:

    IOemBase( const IOemBase & ); // disabled
    IOemBase & operator=( const IOemBase & ); // disabled

    // The number of active references to this object.  (This is used by
    // the IUnknown methods.)
    LONG m_refCount;

    // This is here so that we can avoid const_casting in EnableDriver().
    // It represents an object-specific copy of the hook function table.
    DRVFN * m_hookPtr;

#pragma warning( disable : 4251 )

    // Contains the names of the pure virutal methods of the IPrintOemUni
    // that we have implemented.  This collection is used by
    // GetImplementedMethod() to determine if a method has been implemented.
    // It is the job of each derived classes to call implementMethod()
    // once for each method that it implements.
    std::set<std::basic_string<CHAR> > m_methodNameTable;

    // m_puntFunctions contains the functions we can punt to.  getPuncFunc()
    // is basically just a convenient way of extracting functions.  We
    // dont' have other methods for manipulating m_puntFunctions because
    // the methods on std::set are good enough.  Searching is a bit 
    // cumbersome, however.  The only use of m_puntFunctions is for
    // implementing default versions of DrvXxxx functions that can
    // punt.

    std::set<DRVFN> m_puntFunctions;
    PFN getPuntFunc( ULONG iFunc );
};

#endif


