#include "stdafx.h"

#include "OemTurboSeg.h"
#include "Debug.h"
#include "misc.h"
#include "Segmentor.h"

static Stopwatch IOemTurboSegRenderTmr( "IOemTurboSeg::Drv*Render" );
static Stopwatch downTmr ( "Down Sampling" );
static Stopwatch jedTmr ( "Jed" );

// Here was are just initializing our brushes to be solid black and
// solid white.

IOemTurboSeg::IOemTurboSeg( void )
: 
 m_downSample( 4 ),
 m_fgJpegQuality( 90 ),
 m_bgJpegQuality( 90 ),
 m_hasColor( false ),
 m_numGdiCallsBand( 0 ),
 m_selectorOnly( false ),
 m_smearHalftones( true ),
 m_bgFgOnly( 0 ),
 m_pTfxOut    ( NULL ),
 m_pG4        ( NULL ),
 m_pJpeg      ( NULL ),
 m_pFgDownBuf ( NULL ),
 m_pBgDownBuf ( NULL ),
 m_pBgBuf     ( NULL ),
 m_pFgBuf     ( NULL ),
 m_pSelectBuf ( NULL ),
 m_passThrough( false )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::IOemTurboSeg" );

    m_pTfxOut = new ITfxOutput;
    m_pJpeg   = new IJpeg;
    m_pG4     = new IG4; 
}

IOemTurboSeg::~IOemTurboSeg( void )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::~IOemTurboSeg" );

    delete m_pTfxOut;     m_pTfxOut    = NULL;   
    delete m_pG4;         m_pG4        = NULL;
    delete m_pJpeg;       m_pJpeg      = NULL;
    delete m_pBgDownBuf;  m_pBgDownBuf = NULL;
    delete m_pFgDownBuf;  m_pFgDownBuf = NULL;
    delete m_pBgBuf;      m_pBgBuf     = NULL;
    delete m_pFgBuf;      m_pFgBuf     = NULL;
    delete m_pSelectBuf;  m_pSelectBuf = NULL;

    DbgStream() << IOemTurboSegRenderTmr << std::endl;
    DbgStream() << downTmr << std::endl;
    DbgStream() << jedTmr << std::endl;
}

/*

  This function is very simple.  We just loop over all of the
  pixels in the image and force them to be care pixels.

  rowPtr -- points to the first pixel in the image.
  rows   -- the number of rows in the image
  cols   -- the number of columns in the image
  delta  -- the number of bytes between the beginnings of successive rows

*/

void IOemTurboSeg::
removeDontCares( Ptr24Bit rowPtr, const size_t rows,
                 const size_t cols, const size_t delta )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::removeDontCares" );

    const Ptr24Bit endRow( rowPtr.bytePtr() + (rows*delta) );

    for( ; rowPtr != endRow; rowPtr.offset(delta) )
    {
        Ptr24Bit pixPtr( rowPtr );
    
        const Ptr24Bit endPix( rowPtr + cols );

        for( ; pixPtr != endPix; ++pixPtr )
        {
            pixPtr.forceCare();
        }
    }
}

/*
  This method works on exactly one row at a time.  

  This function takes a row of the selector and the full-color
  image as inputs.  Using the selector, it moves appropriate pixels
  to the background and marks masked-off pixels as don't cares.

  What this function does is sort of counterintuitive.  It actually
  doubles the total amount of color data.  However, half of that
  data is marked as don't care.  This is the prelude to the
  downampling step that is to follow.

  selectRow -- pointer to the beginning of the selector row.  The selector
               must already have been created.
  fgRow     -- pointer to the beginning of the foreground row.  The
               foreground must already have been rendered.
  bgRow     -- pointer to an unitialized block of memory that will have the
               background put into it.
  cols      -- the number of pixels in a row.
               
*/

void IOemTurboSeg::
applySelector( BitPtr selectRow, 
               Ptr24Bit fgRow, Ptr24Bit bgRow, size_t cols )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::applySelector" );

    const size_t numColorBytes = cols * 3;

    // For each pixel in the row, either mark the background
    // as don't care, or copy the foreground to the background and
    // mark the foreground as don't care, depending on the selector.
    
    const Ptr24Bit fgEnd( fgRow + cols );

    for( ; fgRow != fgEnd; ++selectRow, ++fgRow, ++bgRow )
    {
        if( *selectRow )
        {
            bgRow.markDontCare();
        }
        else
        {            
            bgRow.copyDataFrom( fgRow );
            fgRow.markDontCare();
        }
    }
}

/*

  doBand() is the meat of both DrvNextBand() and DrvSendPage().
  It should not be called except for by these two methods.
  
  pso  -- pointer to the SURFOBJ passed to either DrvNextBand()
          or DrvSendPage().

*/

BOOL IOemTurboSeg::doBand( SURFOBJ * pso )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::doBand" );

    const size_t rows = bandRows( pso );
    const size_t cols = pso->sizlBitmap.cx;  // For convenience.

    assert( rows <= pso->sizlBitmap.cy );

    // Our driver works at two color depths mono-chome (black and white) and
    // 24-bit true color.
    if( bitsPixel() == 1 )
    {
        // For the monochrome case, we just write the data from our surface
        // to the output one scanline at a time.
        
        // Get a properly typed pointer to our bitmaps.
        Ptr24Bit base( static_cast<unsigned char *>(pso->pvScan0) );

        // The surface we normally think of as 24-bit is actually monochrome.
        unsigned char * rowPtr = base.bytePtr();

        for(size_t row=0; row != rows; ++row, rowPtr += pso->lDelta)
        {
            
            unsigned char * const endRow = (rowPtr + pso->lDelta);
            
            // JBTODO:  This is a kluge.  We should change the tfx file.
            for( unsigned char * ptr = rowPtr; ptr != endRow; ++ptr )
            {
                *ptr = ~*ptr;
            }

            m_pSelectBuf->write( rowPtr, m_bwBytesRow );
        }
    }
    else if( bitsPixel() == 24 )
    {
        // Get a properly typed pointer to our bitmaps.
        Ptr24Bit clrBase( static_cast<unsigned char *>(pso->pvScan0) );

        // Does our image actually have any color?  (Or, more correctly, is
        // there a change it has color?)
        if( m_hasColor )
        {
            // If yes we must set any pixels which are accidental don't cares
            // to something that is not a don't care.
            removeDontCares( clrBase, rows, cols, pso->lDelta );

            // Now, we must build the selector.
            BinarySegmentor seg;

            seg.setSmearing( m_smearHalftones );

            seg.addImage( clrBase, pso->lDelta );
            seg.addSelector( m_selSurf.data(), m_downSample, rows, cols, 
                             m_selSurf.delta() );
            seg.segment();
        }
        

        {
            // Write the selector 1 row at a time (to take care of alignment
            // issues.
            unsigned char * rowPtr = m_selSurf.data();

            const size_t selDelta = m_selSurf.delta();

            for( size_t row=0; row != rows; ++row, rowPtr += selDelta )
            {
                m_pSelectBuf->write( rowPtr, m_bwBytesRow );
            }
        }
        
        // If there is color then we must deal with the foreground and
        // background.
        if( m_hasColor )
        {
            // You might ask why we do not have a special case
            // for m_downSample == 1 in here.  The short answer is
            // that I tried it and it ran more slowly.  What we
            // have in here if there is no downsampling is effectively
            // a random selector because we never made one.  We have
            // two options.  The first is memsetting the selector to
            // 0xff and setting each of the background pixels to don't
            // cares.  The other (which we have chosen here) is that
            // we can just pretend that the state of the selector is
            // deliberate and continue as before.  applySelector() is
            // basically just producing the appropriate don't care
            // pixels then.  In our second pass, the JpegSegmentor
            // will end up fixing the selector anyway.  Intuitively,
            // I would think that option 1 is faster, but it does not
            // seem to be.  Option 2 has less code.

            // The number of bytes in a row in the undownsampled bitmap.
            const size_t clrBytesBmp = cols * 3;

            unsigned char * bgRowBuf = new unsigned char[clrBytesBmp];

            Ptr24Bit fgRow( clrBase );
            BitPtr selectRowPtr( m_selSurf.data() );

            Ptr24Bit bgRow( bgRowBuf );

            const size_t selDelta = m_selSurf.delta();

            for( size_t row = 0; row < rows; ++row, fgRow.offset(pso->lDelta),
                 selectRowPtr.offset(selDelta) )
            {
                applySelector( selectRowPtr, fgRow, bgRow, pso->sizlBitmap.cx );

            	downTmr.start ();
                m_pBgDownBuf->write( bgRow.bytePtr(), clrBytesBmp );
                m_pFgDownBuf->write( fgRow.bytePtr(), clrBytesBmp );
	            downTmr.stop ();
            }

            delete [] bgRowBuf; bgRowBuf = NULL;
        }
    }
    else
    {
        DBG_ERROR( "Bits per pixel set to an illegal value." );
        DBG_WATCH( "Bits per pixel", bitsPixel() );
        return FALSE;
    }

    // flush all buffers so that we are sure that all data is there
    m_pBgDownBuf->flush();
    m_pFgDownBuf->flush();

    if( m_pBgBuf->index() && m_pFgBuf->index() )
    {
        JpegSegmentor seg;

        seg.addColor( Ptr24Bit(m_pFgBuf->getDataPointer()), 
                      Ptr24Bit(m_pBgBuf->getDataPointer()) );
        seg.addSelector( m_pSelectBuf->getDataPointer(), m_downSample,
                         rows, m_gdiInfo.ulHorzRes );

        seg.resegment();

        assert( !(m_pBgBuf->index() % 3) );

        if( m_pBgBuf->isSolidWhite() )
        {
            m_pBgBuf->flush();
        }

        // Now we find the edges of the foreground.  We assume that the
        // background extends all the way to either side because it normally
        // does.  In almost all cases, there is a roughly 1/2-inch margin.

        // Start by assuming that the left edge is all the way at the right
        // and the right edge is all the way at the left.  This is the most
        // generous guess we can make.  Then we look at each row to make our
        // guess more conservative if needed.

        m_fgLeft  = cols;
        m_fgRight = 0;

        BitPtr rowPtr( m_pSelectBuf->getDataPointer() );

        for( size_t row = 0; row < rows; ++row, rowPtr.offset(m_bwBytesRow) )
        {                
            // Find new minimum for left edge.
            {
                BitPtr lPtr( rowPtr );

                size_t col;

                for( col = 0; (col != m_fgLeft) && !*lPtr; ++col )
                {
                    ++lPtr;
                }

                m_fgLeft = col;
            }

            // Find new maximum for right edge.
            {
                BitPtr rPtr( rowPtr );

                rPtr += (cols - 1);

                size_t col;

                for( col = (cols-1); (col != m_fgRight) && !*rPtr; --col )
                {
                    --rPtr;
                }

                m_fgRight = col;
            }            
        }

        DbgStream() << "left, right (" << m_fgLeft;
        DbgStream() << ", " << m_fgRight << ")" << std::endl;
    }

    if( m_selectorOnly )
    {
        m_pBgBuf->flush();
        m_pFgBuf->flush();
    }
    else if( m_bgFgOnly )
    {
        m_pSelectBuf->flush();

        unsigned char byte = (m_bgFgOnly == 1)? 0xff : 0x00;

        unsigned char * buf = new unsigned char[m_bwBytesRow];

        memset( buf, byte, m_bwBytesRow );

        for( size_t row=0; row != rows; ++row )
        {
            m_pSelectBuf->write( buf, m_bwBytesRow );
        }
        
        delete [] buf; buf = NULL;

        if( m_bgFgOnly == -1 )
        {
            m_pFgBuf->flush();
        }
        else
        {
            m_pBgBuf->flush();
        }
    }

    return TRUE;
}


BOOL IOemTurboSeg::writeCEF( SURFOBJ * pso )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::writeCEF" );

    const ULONG y_pos = m_bandOrigin.y;
    const size_t selRowsToWrite = bandRows( pso );
    
    // write out the layers, one after another
	// the compressors have built in buffer handling. All we need is a pointer to the data
    BOOL status = FALSE;
    unsigned char * comp_buf = NULL;
    ULONG comp_buf_size;
	ULONG layer, bitsPpixel, channels;
    ULONG colorRowsToWrite = (selRowsToWrite + (m_downSample - 1)) / m_downSample;

/**********************************************************
 * LAYER 2     selector layer, compressed with G4
 */

	layer      = 2;
	bitsPpixel = 1;
	channels   = 1;

	// tell the compressor the size of the buffer
    status = m_pG4->SetSize( m_gdiInfo.ulHorzRes, selRowsToWrite, 
                             m_bwBytesRow, bitsPpixel, channels);

	// get the compressed data
    if (status)
    {
        status = m_pG4->Compress (m_pSelectBuf->getDataPointer(), m_pSelectBuf->index (), 
            &comp_buf, &comp_buf_size);
    }

	// With the first selector strip we need to store page information, like size and resolution
    if (status)
    {
        if (y_pos == 0)
        {
            status = m_pTfxOut->TfxWriteFirstSelectorStrip(m_gdiInfo.ulHorzRes, m_gdiInfo.ulVertRes, 
                    pso->sizlBitmap.cy, m_gdiInfo.ulLogPixelsX, m_gdiInfo.ulLogPixelsY, 
                    comp_buf, comp_buf_size, IO_COMP_MMR);
        }
        else
        {
            status = m_pTfxOut->TfxWriteNextSelectorStrip (comp_buf, comp_buf_size);
        }
    }

    comp_buf = NULL;

/*************************************************
 * LAYER 1  background layer, compressed with JPEG
 */

	layer = 1;
	bitsPpixel = 24;
	channels = 3;

	// check if there is any data that is not black. If there isn't we don't need to write this
	// layer
    if( m_pBgBuf->index() )
    {
		// tell the compressor the size of the buffer
        if (status)
        {
            status = m_pJpeg->SetSize (m_clrPixRow, colorRowsToWrite, 
                    m_clrPixRow * 3, bitsPpixel, channels);
        }

        // get a pointer to the compressed data
		if (status)
        {
            m_pJpeg->setQuality( m_bgJpegQuality );
            status = m_pJpeg->Compress (m_pBgBuf->getDataPointer(), m_pBgBuf->index (), 
                &comp_buf, &comp_buf_size);
        }

		// save the compressed data
        if (status)
        {
            status = m_pTfxOut->TfxAddImage (layer, 0, y_pos,
                     m_clrPixRow, colorRowsToWrite, 
                     m_gdiInfo.ulLogPixelsX / m_downSample, m_gdiInfo.ulLogPixelsY / m_downSample, 
                     comp_buf, comp_buf_size, IO_COMP_JPEG);  
            comp_buf = NULL;
        }
    }

/*********************************
 * LAYER 3, foreground layer, JPEG
 */

	layer      = 3;
	bitsPpixel = 24;
	channels   = 3;

	// check if there is any data that is not black. If there isn't we don't need to write this
	// layer
    if( m_pFgBuf->index() )
    {
		// tell the compressor the size of the buffer
        if (status)
        {
            status = m_pJpeg->SetSize (m_clrPixRow, colorRowsToWrite, 
                    m_clrPixRow * 3, bitsPpixel, channels);
        }

		// get a pointer to the compressed data
        if (status)
        {
            m_pJpeg->setQuality( m_fgJpegQuality );
            status = m_pJpeg->Compress (m_pFgBuf->getDataPointer(), m_pFgBuf->index (), 
                &comp_buf, &comp_buf_size);
        }

		// save the compressed data
        if (status)
        {
            status = m_pTfxOut->TfxAddImage (layer, 0, y_pos,
                     m_clrPixRow, colorRowsToWrite, 
                     m_gdiInfo.ulLogPixelsX / m_downSample, m_gdiInfo.ulLogPixelsY / m_downSample, 
                     comp_buf, comp_buf_size, IO_COMP_JPEG);  
            comp_buf = NULL;
        }
    }
   
/**********************
 *  all layers written 
 */

    m_pBgBuf->    flush();
	m_pFgBuf->    flush();
    m_pSelectBuf->flush();
	
    return status;
}

/////////////////////////////////////////////////////////////////////////
//
// IOemBase methods
//
//////////////////////////////////////////////////////////////////////////

HRESULT IOemTurboSeg::initFrom( IOemBase * old )
{
    DBG_TRACE( "IOemTurboSeg::initFrom" );

    IOemTurboSeg * oldThis = dynamic_cast<IOemTurboSeg *>(old);

    if( !oldThis )
    {
        return E_FAIL;
    }

    this->IOemBase::initFrom( old );

    this->m_passThrough = oldThis->m_passThrough;
    this->m_passFile    = oldThis->m_passFile;

    this->m_downSample = oldThis->m_downSample;

    this->m_fgJpegQuality = oldThis->m_fgJpegQuality;
    this->m_bgJpegQuality = oldThis->m_bgJpegQuality;

    this->m_smearHalftones = oldThis->m_smearHalftones;

    this->m_selectorOnly = oldThis->m_selectorOnly;
    this->m_bgFgOnly     = oldThis->m_bgFgOnly;

    this->m_selSurf.take( oldThis->m_selSurf );

    this->m_pTfxOut    = oldThis->m_pTfxOut;     oldThis->m_pTfxOut    = NULL;
    this->m_pJpeg      = oldThis->m_pJpeg;       oldThis->m_pJpeg      = NULL;
    this->m_pG4        = oldThis->m_pG4;         oldThis->m_pG4        = NULL;
    this->m_pBgDownBuf = oldThis->m_pBgDownBuf;  oldThis->m_pBgDownBuf = NULL;
    this->m_pFgDownBuf = oldThis->m_pFgDownBuf;  oldThis->m_pFgDownBuf = NULL;
    this->m_pBgBuf     = oldThis->m_pBgBuf;      oldThis->m_pBgBuf     = NULL;
    this->m_pFgBuf     = oldThis->m_pFgBuf;      oldThis->m_pFgBuf     = NULL;
    this->m_pSelectBuf = oldThis->m_pSelectBuf;  oldThis->m_pSelectBuf = NULL;
    
    return S_OK;
}

BOOL IOemTurboSeg::DrvStartDoc( SURFOBJ *pso, PWSTR pwszDocName, DWORD dwJobId )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::DrvStartDoc" );

    if( m_passThrough )
    {
        DBG_MSG( "StartDoc aborted early because we are doing a passthrough." );
        return TRUE;
    }

    // Fulfill our requirment to the base class   

    if( !IOemBase::DrvStartDoc( pso, pwszDocName, dwJobId ) )
    {
        return FALSE;
    }

    // Some default values in case we have trouble reading them.
    m_downSample     = 4;
    m_fgJpegQuality  = 90;
    m_bgJpegQuality  = 90;
    m_smearHalftones = true;
    m_selectorOnly   = false;
    m_bgFgOnly       = 0;

    // For convenience.
    PDEVOBJ pdevobj = reinterpret_cast<PDEVOBJ>(pso->dhpdev);

    // Read the downsample factor.

    std::string setting = getSetting( pdevobj, "FgDownSampleFactor" );

    if( setting == std::string( "NONE" ) )
    {
        m_downSample = 1;
    }
    else if( setting == std::string( "COLOR_DOWN_2" ) )
    {
        m_downSample = 2;
    }
    else if( setting == std::string( "COLOR_DOWN_4" ) )
    {
        m_downSample = 4;
    }

    // Read the background JPEG quality.

    setting = getSetting( pdevobj, "BgJpegQuality" );    

    sscanf( setting.c_str() + strlen( "JPEG_" ), "%d", &m_bgJpegQuality );


    // Read the foreground JPEG quality.

    setting = getSetting( pdevobj, "FgJpegQuality" );    

    sscanf( setting.c_str() + strlen("JPEG_"), "%d", &m_fgJpegQuality );

    // Read the whether we are in one of the partial-rendering debug modes.

    setting = getSetting( pdevobj, "OutputImage" );
    
    if( setting == std::string("SelectorOnly") )
    {
        m_selectorOnly = true;
    }
    else if( setting == std::string("BgOnly") )
    {
        m_bgFgOnly = -1;
    }
    else if( setting == std::string("FgOnly") )
    {
        m_bgFgOnly = 1;
    }

    setting = getSetting( pdevobj, "HalftoneSmearing" );
    
    if( setting == std::string("ON") )
    {
        m_smearHalftones = true;
    }
    else if( setting == std::string("OFF") )
    {
        m_smearHalftones = false;
    }

    return m_pTfxOut->TfxStartDoc();
}

BOOL IOemTurboSeg::DrvStartPage( SURFOBJ *pso )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::DrvStartPage" );
    DBG_WATCH( "Page number", m_curPage );

    if( m_passThrough )
    {
        DBG_MSG( "StartPage aborted early because we are doing a passthrough." );
        return TRUE;
    }
    
    IOemBase::DrvStartPage( pso );
    
    // Set the main surfaces to flat white for the next page.
    memset( pso->pvScan0, 0xff, pso->sizlBitmap.cy * pso->lDelta );

    // Create (or recreate) the selector surface.
    m_selSurf.Create( pso, m_gdiInfo.ulHorzRes, m_gdiInfo.ulVertRes );

    // Set this to flat black by default.
    m_selSurf.fill( 0x00 );

    // We start each page by assuming there is no color and there 
    // have been no calls.
    m_hasColor        = false;
    m_numGdiCallsBand = 0;

    // set horizontal sizes, assuming that bw is 1 bpp, color is 24 bpp.
    // This is the byte-aligned number used for output, not the GDI-surface
    // lDelta number.
    m_bwBytesRow = (m_gdiInfo.ulHorzRes + 7) / 8;

    m_clrPixRow = (m_gdiInfo.ulHorzRes/m_downSample) + !!(m_gdiInfo.ulHorzRes%m_downSample);
           
    // colorSize is the amount of data (in bytes) that will be in the
    // uncompressed output image.  This value is byte aligned.
    // Here we assume that there are only two options: (bitsPixel() == 1) or 
    // (bitsPixel() == 24).  If we are monochrome, there is not color data.
    // we are in color, we need enough rows the downsampled inmage.  This is
    // the number of rows in the full-resolution image divided by the 
    // downsample factor rounded up.
    const size_t colorSize = (bitsPixel() == 1)?  0 :
        m_clrPixRow * ((pso->sizlBitmap.cy + (m_downSample - 1)) / m_downSample) * 3; 
        
    // Get rid of any existing buffers and make new ones.
        
    delete m_pBgBuf;
    delete m_pFgBuf;
    delete m_pSelectBuf;
    delete m_pBgDownBuf;
    delete m_pFgDownBuf;

    m_pBgBuf      = new ColorStorageBuf( colorSize ); 
    m_pFgBuf      = new ColorStorageBuf( colorSize );
    m_pSelectBuf  = new StorageBuf( m_bwBytesRow * pso->sizlBitmap.cy );
    m_pBgDownBuf  = new DownBuf( m_downSample, m_gdiInfo.ulHorzRes, m_pBgBuf );
    m_pFgDownBuf  = new DownBuf( m_downSample, m_gdiInfo.ulHorzRes, m_pFgBuf );
 
    PDEVOBJ pdevobj = reinterpret_cast<PDEVOBJ>(pso->dhpdev);

    // let the TIFF-FX generator know that we are starting a new page
	// pass along the spool buffer information

    return m_pTfxOut->TfxStartPage( pdevobj, m_pOEMHelp );
}

BOOL IOemTurboSeg::DrvNextBand( SURFOBJ * pso, POINTL * pptl )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::DrvNextBand" );

    if( m_passThrough )
    {
        DBG_MSG( "Aborting DrvNextBand because we are in pass through mode." );
        m_bandOrigin.x = -1;
        m_bandOrigin.y = -1;
        *pptl = m_bandOrigin;

        return TRUE;
    }

    IOemBase::DrvNextBand( pso, pptl );

    // I believe that pso and pptl should never be NULL.
    if( !pso  ) { assert( pso  ); return FALSE; }
    if( !pptl ) { assert( pptl ); return FALSE; }

    // Do segmentation and get everything set up in buffers ready for
    // compression.  m_p*DownBuf are flushed at the end of this.
    if( !doBand(pso) ) { return FALSE; }

    // Compress the buffers and write them out.  The rest of the buffers
    // are flushed at the end of this.
    if( !writeCEF(pso) ) { return FALSE; }

    BOOL status = TRUE;

    //  If there is work left to be done, overrun will be negative.
    if( bandOverrun(pso) < 0 )
    {
        // Incriment m_bandOrigin to point to the top of the undraw
        // region.
        m_bandOrigin.y += pso->sizlBitmap.cy;

        // Set the surface to flat white and flat black for the next band.
        memset( pso->pvScan0, 0xff, pso->lDelta * pso->sizlBitmap.cy );

        m_selSurf.fill( 0x00 );

        m_hasColor        = false;
        m_numGdiCallsBand = 0;
    }
    else
    {        
        // Setting both x and y to -1 is how we tell our caller there is
        // nothing more to be done.
        m_bandOrigin.x = -1;
        m_bandOrigin.y = -1;

        // Let the TIFF-FX generator know we are done with the page.
        status = m_pTfxOut->TfxEndPage();
    }

    *pptl = m_bandOrigin;

    return status;
}

BOOL IOemTurboSeg::DrvSendPage( SURFOBJ *pso )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::DrvSendPage" );

    if( m_passThrough )
    {
        DBG_MSG( "SendPage aborted early because we are doing a passthrough." );
        return TRUE;
    }

    IOemBase::DrvSendPage( pso );

    // The documentation seems to say that DrvSendPage() gets
    // called even when banding is turned on, however it does not
    // seem to get called in realitiy.  In any case, we check for
    // it just to make sure.
    if( m_isBanded )
    {
        return TRUE;
    }

    // If we are successful doing segmentation.
    if( doBand(pso) )
    {
        // If we are successful compression.
        if( writeCEF(pso) )
        {
        	// Let the TIFF-FX generator know we are done with the page.
            return  m_pTfxOut->TfxEndPage();
        }
    }

    return FALSE;
}

BOOL IOemTurboSeg::DrvEndDoc( SURFOBJ *pso, FLONG fl )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::DrvEndDoc" );

    BOOL status = TRUE;

    if( !m_passThrough )
    {
        m_passThrough = false;

        // tell the TIFF-FX generator that we are done with this 
        // document and close the file
        status = m_pTfxOut->TfxEndDoc();

        // In principle, we do not have to destroy this stuff here.
        // Most of the time the destructor will run almost immediately
        // after DrvEndDoc().  However, we are finished with this stuff,
        // so why not dump it?
        m_selSurf.Destroy();

        delete m_pBgDownBuf; m_pBgDownBuf = NULL;
        delete m_pFgDownBuf; m_pFgDownBuf = NULL;
        delete m_pBgBuf;     m_pBgBuf     = NULL;
        delete m_pFgBuf;     m_pFgBuf     = NULL;
        delete m_pSelectBuf; m_pSelectBuf = NULL;

        // Note the order here.  By putting DrvEndDoc() first, it always gets
        // evaluated.
        return IOemBase::DrvEndDoc( pso, fl ) && status;
    
    }
    else
    {
        const size_t bufSize = (1024*1024);

        FILE * inFile = fopen( m_passFile.c_str(), "rb" );

        char * buf = new char[bufSize];

        size_t numRead;

        PDEVOBJ pdevobj = reinterpret_cast<PDEVOBJ>(pso->dhpdev);

        do
        {
            DWORD dwResult;

            numRead = fread( buf, 1, bufSize, inFile );
            m_pOEMHelp->DrvWriteSpoolBuf( pdevobj, buf, numRead, &dwResult );
        }
        while( numRead == bufSize );

        delete [] buf;
        
        fclose( inFile );
    }

    return TRUE;
}

// We clearly cannot handle any escapes by default.
ULONG IOemTurboSeg::
DrvEscape( SURFOBJ * pso, ULONG iEsc, ULONG cjIn, PVOID pvIn, 
           ULONG cjOut, PVOID pvOut )
{
    DBG_TRACE_OBJ( "IOemTurboSeg::DrvEscape" );

    DBG_WATCH( "iEsc", iEsc );

    // Indicates escape is "not supported."
    ULONG supported = 0;

    if( iEsc == QUERYESCSUPPORT )
    {
        std::string data( reinterpret_cast<const char *>(pvIn) );

        DBG_WATCH( "pvIn", data );

        if( data == "Xerox CEF Driver" )
        {
            DBG_MSG( "PASSTRHOUGH MODE!" );
        
            m_passThrough = true;
            
            supported = 1;
        }
    }
    else if( m_passThrough && (iEsc == 0xdeadbeaf) )
    {
        DBG_MSG( "File Name for pass through is?" );

        const char * data = reinterpret_cast<const char *>(pvIn);
        
        m_passFile = std::string( data );

        DBG_WATCH( "file name", m_passFile );

        supported = 1;
    }
    
    return supported; 
}

/////////////////////////////////////////////////////////////////////////
// IOemBase methods -- Drawing methods
//
//    Methods that draw to the text plane.
//////////////////////////////////////////////////////////////////////////

BOOL IOemTurboSeg::DrvTextOut(
    SURFOBJ    *pso,
    STROBJ     *pstro,
    FONTOBJ    *pfo,
    CLIPOBJ    *pco,
    RECTL      *prclExtra,
    RECTL      *prclOpaque,
    BRUSHOBJ   *pboFore,
    BRUSHOBJ   *pboOpaque,
    POINTL     *pptlOrg,
    MIX         mix )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvTextOut" );
    
    ++m_numGdiCallsBand;

    m_hasColor |= brushHasColor( pboFore   );
    m_hasColor |= brushHasColor( pboOpaque );

    BOOL ok = EngTextOut( pso, pstro, pfo, pco, prclExtra, prclOpaque, pboFore, pboOpaque, pptlOrg, mix ); 

    if( !m_hasColor )
    {
        BRUSHOBJ * pOpa = NULL;
        BRUSHOBJ * pFor = NULL;

        if( pboFore )
        {
            pFor = pboFore->iSolidColor? &m_blackBr : &m_whiteBr ;
        }

        if( pboOpaque )
        {
            pOpa = pboOpaque->iSolidColor? &m_blackBr : &m_whiteBr;
        }

        ok = EngTextOut( m_selSurf.surf(), pstro, pfo, pco, prclExtra, prclOpaque, pFor, pOpa, pptlOrg, mix );
    }

    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvLineTo(
    SURFOBJ    *pso,
    CLIPOBJ    *pco,
    BRUSHOBJ   *pbo,
    LONG        x1,
    LONG        y1,
    LONG        x2,
    LONG        y2,
    RECTL      *prclBounds,
    MIX         mix )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvLineTo" );

    m_hasColor |= brushHasColor( pbo );
    ++m_numGdiCallsBand;

    BOOL ok = EngLineTo( pso, pco, pbo, x1, y1, x2, y2, prclBounds, mix );

    if( !m_hasColor )
    {
        BRUSHOBJ * pBr = NULL;

        if( pbo )
        {
            pBr = pbo->iSolidColor? &m_blackBr : &m_whiteBr;
        }

        ok = EngLineTo( m_selSurf.surf(), pco, pBr, x1, y1, x2, y2, prclBounds, mix );
    }

    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvStrokePath(
    SURFOBJ    *pso,
    PATHOBJ    *ppo,
    CLIPOBJ    *pco,
    XFORMOBJ   *pxo,
    BRUSHOBJ   *pbo,
    POINTL     *pptlBrushOrg,
    LINEATTRS  *plineattrs,
    MIX         mix )
{
    IOemTurboSegRenderTmr.start();
    
    DBG_TRACE_OBJ( "IOemTurboSeg::DrvStrokePath" );
    
    ++m_numGdiCallsBand;
    m_hasColor |= brushHasColor( pbo );

    BOOL ok = EngStrokePath( pso, ppo, pco, pxo, pbo, pptlBrushOrg, plineattrs, mix );

    if( !m_hasColor )
    {
        BRUSHOBJ * pBr = NULL;

        if( pbo )
        {
            pBr = pbo->iSolidColor? &m_blackBr : &m_whiteBr;
        }

        ok = EngStrokePath( m_selSurf.surf(), ppo, pco, pxo, pBr, pptlBrushOrg, plineattrs, mix );
    }

    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvFillPath(
    SURFOBJ    *pso,
    PATHOBJ    *ppo,
    CLIPOBJ    *pco,
    BRUSHOBJ   *pbo,
    POINTL     *pptlBrushOrg,
    MIX         mix,
    FLONG       flOptions )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvFillPath" );

    ++m_numGdiCallsBand;
    m_hasColor |= brushHasColor( pbo );

    BOOL ok = EngFillPath( pso, ppo, pco, pbo, pptlBrushOrg, mix, flOptions );

    if( !m_hasColor )
    {
        BRUSHOBJ * pBr = NULL;

        if( pbo )
        {
            pBr = pbo->iSolidColor? &m_blackBr : &m_whiteBr;
        }

        ok = EngFillPath( m_selSurf.surf(), ppo, pco, pBr, pptlBrushOrg, mix, flOptions );
    }

    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvStrokeAndFillPath(
    SURFOBJ    *pso,
    PATHOBJ    *ppo,
    CLIPOBJ    *pco,
    XFORMOBJ   *pxo,
    BRUSHOBJ   *pboStroke,
    LINEATTRS  *plineattrs,
    BRUSHOBJ   *pboFill,
    POINTL     *pptlBrushOrg,
    MIX         mixFill,
    FLONG       flOptions )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvStrokeAndFillPath" );

    ++m_numGdiCallsBand;
    m_hasColor |= brushHasColor( pboStroke );
    m_hasColor |= brushHasColor( pboFill   );


    BOOL ok = EngStrokeAndFillPath( pso, ppo, pco, pxo, pboStroke, plineattrs, pboFill, pptlBrushOrg, mixFill, flOptions );

    if( !m_hasColor )
    {
        BRUSHOBJ * pStr = NULL;
        BRUSHOBJ * pFil = NULL;

        if( pboStroke )
        {
            pStr = pboStroke->iSolidColor? &m_blackBr : &m_whiteBr;
        }

        if( pboFill )
        {
            pFil = pboFill->iSolidColor? &m_blackBr : &m_whiteBr;
        }

        BOOL ok = EngStrokeAndFillPath( m_selSurf.surf(), ppo, pco, pxo, pStr, plineattrs, pFil, pptlBrushOrg, mixFill, flOptions );
    }

    IOemTurboSegRenderTmr.stop();

    return ok;
}

/////////////////////////////////////////////////////////////////////////
// IOemBase methods -- Drawing methods
//
//    Methods that do not draw to the text plane.
//////////////////////////////////////////////////////////////////////////

BOOL IOemTurboSeg::DrvBitBlt(
    SURFOBJ   *psoTrg,
    SURFOBJ   *psoSrc,
    SURFOBJ   *psoMask,
    CLIPOBJ   *pco,
    XLATEOBJ  *pxlo,
    RECTL     *prclTrg,
    POINTL    *pptlSrc,
    POINTL    *pptlMask,
    BRUSHOBJ  *pbo,
    POINTL    *pptlBrush,
    ROP4      rop4 )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvBitBlt" );

    DBG_WATCH( "psoSrc",  (void *)(psoSrc ) );
    DBG_WATCH( "psoMask", (void *)(psoMask) );
    DBG_WATCH( "pco",     (void *)(pco) );
    DBG_WATCH( "pxlo",    (void *)(pxlo) );
    DBG_WATCH( "pbo",     (void *)(pbo) );
    
    if( pbo )
    {
        DBG_WATCH( "pbo->iSolidColor", (void*)(pbo->iSolidColor) );
        DBG_WATCH( "pbo->pvRbrush",     (void*)(pbo->pvRbrush) );
        DBG_WATCH( "pbo->flColorType", (void*)(pbo->flColorType) );
    }

    DBG_WATCH( "rop4", (void *)(rop4) );

    ++m_numGdiCallsBand;
    m_hasColor |= brushHasColor( pbo );
    m_hasColor |= !!psoSrc;

    BOOL ok = EngBitBlt( psoTrg, psoSrc, psoMask, pco, pxlo, prclTrg, pptlSrc, pptlMask, pbo, pptlBrush, rop4 );

    if( !m_hasColor )
    {
        BRUSHOBJ * pBr = NULL;

        if( pbo )
        {
            pBr = pbo->iSolidColor? &m_blackBr : &m_whiteBr;
        }

        ok = EngBitBlt( m_selSurf.surf(), psoSrc, psoMask, pco, pxlo, prclTrg, pptlSrc, pptlMask, pBr, pptlBrush, rop4 );
    }

    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvStretchBlt(
    SURFOBJ         *psoDest,
    SURFOBJ         *psoSrc,
    SURFOBJ         *psoMask,
    CLIPOBJ         *pco,
    XLATEOBJ        *pxlo,
    COLORADJUSTMENT *pca,
    POINTL          *pptlHTOrg,
    RECTL           *prclDest,
    RECTL           *prclSrc,
    POINTL          *pptlMask,
    ULONG            iMode )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvStretchBlt" );

    ++m_numGdiCallsBand;
    m_hasColor = true;

    BOOL ok = EngStretchBlt( psoDest, psoSrc, psoMask, pco, pxlo, pca, pptlHTOrg, prclDest, prclSrc, pptlMask, iMode );
    
    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvStretchBltROP(
    SURFOBJ         *psoDest,
    SURFOBJ         *psoSrc,
    SURFOBJ         *psoMask,
    CLIPOBJ         *pco,
    XLATEOBJ        *pxlo,
    COLORADJUSTMENT *pca,
    POINTL          *pptlHTOrg,
    RECTL           *prclDest,
    RECTL           *prclSrc,
    POINTL          *pptlMask,
    ULONG            iMode,
    BRUSHOBJ        *pbo,
    ROP4             rop4 )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvStretchBltROP" );

    ++m_numGdiCallsBand;
    m_hasColor = true;

    BOOL ok = EngStretchBltROP( psoDest, psoSrc, psoMask, pco, pxlo, pca, pptlHTOrg, prclDest, prclSrc, pptlMask, iMode, pbo, rop4 );
    
    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvCopyBits(
    SURFOBJ        *psoDest,
    SURFOBJ        *psoSrc,
    CLIPOBJ        *pco,
    XLATEOBJ       *pxlo,
    RECTL          *prclDest,
    POINTL         *pptlSrc )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvCopyBits" );

    ++m_numGdiCallsBand;
    m_hasColor = true;

    const BOOL ok = EngCopyBits( psoDest, psoSrc, pco, pxlo, prclDest, pptlSrc );

    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvTransparentBlt(
    SURFOBJ    *psoDst,
    SURFOBJ    *psoSrc,
    CLIPOBJ    *pco,
    XLATEOBJ   *pxlo,
    RECTL      *prclDst,
    RECTL      *prclSrc,
    ULONG      iTransColor,
    ULONG      ulReserved )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvTransparentBlt" );

    ++m_numGdiCallsBand;
    m_hasColor = true;

    const BOOL ok = EngTransparentBlt( psoDst, psoSrc, pco, pxlo, prclDst, prclSrc, iTransColor, ulReserved );

    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvGradientFill(
    SURFOBJ    *psoDest,
    CLIPOBJ    *pco,
    XLATEOBJ   *pxlo,
    TRIVERTEX  *pVertex,
    ULONG       nVertex,
    PVOID       pMesh,
    ULONG       nMesh,
    RECTL      *prclExtents,
    POINTL     *pptlDitherOrg,
    ULONG       ulMode )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvGradientFill" );

    ++m_numGdiCallsBand;
    m_hasColor = true;

    const BOOL ok = EngGradientFill( psoDest, pco, pxlo, pVertex, nVertex, pMesh, nMesh, prclExtents, pptlDitherOrg, ulMode );

    IOemTurboSegRenderTmr.stop();

    return ok;
}

BOOL IOemTurboSeg::DrvAlphaBlend(
    SURFOBJ    *psoDest,
    SURFOBJ    *psoSrc,
    CLIPOBJ    *pco,
    XLATEOBJ   *pxlo,
    RECTL      *prclDest,
    RECTL      *prclSrc,
    BLENDOBJ   *pBlendObj )
{
    IOemTurboSegRenderTmr.start();

    DBG_TRACE_OBJ( "IOemTurboSeg::DrvAlphaBlend" );

    ++m_numGdiCallsBand;
    m_hasColor = true;

    const BOOL ok = EngAlphaBlend( psoDest, psoSrc, pco, pxlo, prclDest, prclSrc, pBlendObj );   

    IOemTurboSegRenderTmr.stop();

    return ok;
}