#ifndef OemTurboSeg_h_INCLUDED
#define OemTurboSeg_h_INCLUDED

#include "OemBase.h"
#include "TfxOutput.h"
#include "jpeg.h"
#include "g4.h"
#include "buffer.h"
#include "misc.h"

/*

  IOemTurboSet is a TFX printer driver that uses a rather cool and
  aggressive segmentation algorithm on the image.  For the most
  part, it uses no hints.  It's output is a three-layer tiffx file
  that contains two JPEG layers, foreground and background, and a
  monochrome selector (compressed with G4) that chooses between the two.

  In most cases, the image is rendered in full 24-bit color as usual
  (but see below.)  If the image is to be downsampled, then a crude
  cluster-based segmentation is applied separately to each group of
  pixels being down-sampled.  The idea here is to choose a palette
  of at most two colors for each region of the image.  Thus, if you are
  downsampling a region that is n pixels by n pixels, you immediately
  get a (lossy) compression ratio less than or equal to 
  (24n^2)/(n^2 + 48).  For n=2, this give you a ratio 1.85:1.  For n=4,
  this gives you a ratio of 6:1 or better.  However, you should note
  that on simple documents, this compression will be lossless.  Here,
  "simple" ends up meaning, "No nxn down-sampling region contains more
  than 2 colors."  This case actually happens much of the time for
  documents created from an electronic source.
  
  Now every nxn 24-bit area in the image has been replaced by an nxn 1-bit
  selector and two 1x1 layers: the foreground and the background.


  Example of 4x4 downsampling (showing two 4x4 areas):


     +-+     +-+ 
     | |     | |     <--24-bit foreground
     +-+     +-+ 
    /   \   /   \
   /     \ /     \
  +-+-+-+-+-+-+-+-+
  | | | | | | | | |
  +-+-+-+-+-+-+-+-+
  | | | | | | | | |
  +-+-+-+-+-+-+-+-+ <-- 1-bit selector
  | | | | | | | | |
  +-+-+-+-+-+-+-+-+
  | | | | | | | | |
  +-+-+-+-+-+-+-+-+
   \     / \     /
    \   /   \   /
     +-+     +-+    
     | |     | |    <--24-bit background
     +-+     +-+ 


  In cases where the 4x4 area is uniform (or considered to be a half-tone,
  see below) it is possible that the selector will be trivial and only one
  color (either background or foreground) is needed.  In this case, the one
  that is not needed is labeled as a don't-care pixel.

  Once this first-pass segmentation and compression is done, we do a second
  pass.  On this pass we use the fact that we can freely exchange background
  and foreground pixels if we also invert the selector between them.  Up
  until now, the choice of which color ended up in which plane has been
  arbitrary.  Now we try to do two things:

  1.  Arange 8x8 squares in the color layers (which correspond to 8n x 8n
      squares in the original) such that there is as little high-frequency
      data as possible.  JPEG has poor image quality and file size properties
      on things with sharp edges an other high-frequency components.
  2.  When possible (and consistant with #1), push white to the background.
      The background defaults to white.  In some cases we can get rid
      of the background entirely by pushing white pixels to the background.

  (Of course, don't care pixels provide extra freedom here.)

  Once the optimal arrangement of the selector has been chosen, the don't care
  pixels are set through a bleeding algorithm.  Each 8x8 square is divided
  into 4x4 quadrants.  The average value of the care pixels in each quadrant is
  found, as is the average value of the care pixels for the entire 8x8 square.
  A don't care pixel is set to the average care pixel in its quadrent or the
  average for the whole 8x8 square if there are no care pixels in its quadrant.
  If the entire 8x8 square is don't care, then it is set to either black or white
  depending on whether it is in the foreground or background.

  In reality, there are a couple more steps.

  1.  We count the GDI calls.  If there are none, we know that band is blank
      and we skip the above process entirely, sending only an empty selector.
  2.  We keep track of whether or not we have seen color (or something we
      didn't understand) in the GDI calls.  As long as we are sure that the
      band is black and white, we render it in both 1-bit and 24-bit color.
      If when the band is complete it is still known to be black and white,
      we can skip the segmentation step and just use the the monochrome
      bitmap as the selector directly.
  3.  In the 4x4 down-sampling case, we attempt to detect and smear half-
      tones.  We do this by looking for the presense of high-frequency 
      components in the selector.
*/

class IOemTurboSeg : public IOemBase
{
public:

    IOemTurboSeg( void );

    virtual ~IOemTurboSeg( void );
       
    //
    // DDI hook methods.
    // 

    virtual BOOL DrvStartDoc ( SURFOBJ *, PWSTR, DWORD ); 
    virtual BOOL DrvStartPage( SURFOBJ * ); 
    virtual BOOL DrvNextBand ( SURFOBJ *, POINTL * );
    virtual BOOL DrvSendPage ( SURFOBJ * );
    virtual BOOL DrvEndDoc   ( SURFOBJ *, FLONG );

    virtual BOOL DrvTextOut( SURFOBJ *, STROBJ *, FONTOBJ *, CLIPOBJ *, RECTL *, RECTL *, BRUSHOBJ *, BRUSHOBJ *, POINTL *, MIX );
    virtual BOOL DrvLineTo( SURFOBJ *, CLIPOBJ *, BRUSHOBJ *, LONG, LONG, LONG, LONG, RECTL *, MIX ); 
    virtual BOOL DrvStrokePath( SURFOBJ *, PATHOBJ *, CLIPOBJ *, XFORMOBJ *, BRUSHOBJ *, POINTL *, LINEATTRS *, MIX );
    virtual BOOL DrvFillPath( SURFOBJ *, PATHOBJ *, CLIPOBJ *, BRUSHOBJ *, POINTL *, MIX, FLONG );
    virtual BOOL DrvStrokeAndFillPath( SURFOBJ *, PATHOBJ *, CLIPOBJ *, XFORMOBJ *, BRUSHOBJ *, LINEATTRS *, BRUSHOBJ *, POINTL *, MIX, FLONG ); 

    virtual BOOL DrvBitBlt( SURFOBJ *, SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, RECTL *, POINTL *, POINTL *, BRUSHOBJ *, POINTL *, ROP4 );
    virtual BOOL DrvStretchBlt( SURFOBJ *, SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, COLORADJUSTMENT *, POINTL *, RECTL *, RECTL *, POINTL *, ULONG );
    virtual BOOL DrvStretchBltROP( SURFOBJ *, SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, COLORADJUSTMENT *, POINTL *, RECTL *, RECTL *, POINTL *, ULONG, BRUSHOBJ *, ROP4 );   
    virtual BOOL DrvCopyBits( SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, RECTL *, POINTL * );
    virtual BOOL DrvTransparentBlt( SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, RECTL *, RECTL *, ULONG , ULONG );
    virtual BOOL DrvGradientFill( SURFOBJ *, CLIPOBJ *, XLATEOBJ *, TRIVERTEX *, ULONG, PVOID, ULONG, RECTL *, POINTL *, ULONG );
    virtual BOOL DrvAlphaBlend( SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, RECTL *, RECTL *, BLENDOBJ * );

    // I haven't figured out how to do this one yet. JBTODO!
    //virtual BOOL DrvPlgBlt( SURFOBJ *, SURFOBJ *, SURFOBJ *, CLIPOBJ *, XLATEOBJ *, COLORADJUSTMENT *, POINTL *, POINTFIX *, RECTL *, POINTL *, ULONG );

    virtual ULONG DrvEscape( SURFOBJ *, ULONG, ULONG, PVOID, ULONG, PVOID ); 

protected:

    // IOemBase methods.

    virtual HRESULT initFrom( IOemBase * );

private:

    IOemTurboSeg( const IOemTurboSeg & );    // Disabled
    IOemTurboSeg & operator=( const IOemTurboSeg & ); // Disabled
  
    bool m_passThrough;
    std::string m_passFile;

    // How much should we down sample the color layers?
    size_t m_downSample;
    
    // Set at the beginning of each page, this is the number of bytes per
    // scan-line including padding up to the nearest full byte.  It is not
    // the delta in a black and white GDI surface.  That value is stored
    // in the SURFOBJ and is dword aligned.
    ULONG m_bwBytesRow;
    
    // Tells whether we should try to smear halftones.
    bool m_smearHalftones;

    // Allows us to generate a tfx file containing only the selector.
    // This is very useful for debugging.
    bool m_selectorOnly;

    // 0 indicates write both.  -1 indicates bg.  +1 indicates fg.
    // Again, this is here for debugging.
    int m_bgFgOnly;
                
    // Starts out false at the begining of a band.  Each drawing call is
    // trapped,  If any of them have color brushes or put an image onto the
    // document, this is set to true.  When m_hasColor is false, we only
    // send the selector to the output.
    bool m_hasColor;

    // This is the number of calls written to the text surface this band.
    // At present this is effectively a boolean used for detection of trivial
    // bands.  If it is 0, we don't do any processing.  We just write a blank
    // selector.  Actually, because of the way things are written, this value
    // ends up not being used.  (m_hasColor turns out to be all we need.) 
    // For the present, I have left this here because it is a pain to put it
    // in.
    size_t m_numGdiCallsBand;

    // These values are the left-most and right-most boundaries of the
    // foreground.  They are set during doBand().  Thus code in DrvNextBand()
    // and DrvSendPage() that happens below doBand() can use them.  Their
    // purpose is to allow as small an image as possible in the final output.
    size_t m_fgLeft, m_fgRight;

    // You send this function one row of foreground data and one row of
    // selector.  This method lifts any background pixels off of the 
    // foreground and replaces them with a don't care value for later
    // bleeding.  It then writes the background row by calling
    // writeBackground().  (You must still write the foreground when
    // of this is done.)
    void applySelector( BitPtr selectRow, Ptr24Bit fgRow, Ptr24Bit bgRow,
                        size_t cols );

    // Chop off the last bit of any pixels that just happen to be don't
    // cares by accident.
    void removeDontCares( Ptr24Bit, const size_t rows, const size_t cols,
                          const size_t delta );

    // Much of DrvNextBand() and DrvSendPage() are identical.  As such,
    // We have written a functions to handle the common work of both
    // methods.  doBand() contains the segmentation code.  writeCEF()
    // does copression and outputs the tfx file.  In both methods, pso
    // is the same pso that was passed as an argument to either
    // DrvSendPage() or DrvNextBand().  doBand() and writeCEF() should
    // only be called from DrvNextBand() or DrvSendPage().
    BOOL doBand( SURFOBJ * pso );
	BOOL writeCEF( SURFOBJ * pso );

    // A 1-bit per pixel version of the document.  We try rendering this
    // way first.  Regardless of whether we get away with 1-bit rendering,
    // this surface also ends up serving as the selector.  This is why
    // we refer these objects as the "Sel"ector surface.
    CustomGdiSurface m_selSurf;

    // I/O lib stuff
    ITfxOutput * m_pTfxOut;

    // compression objects
    IJpeg * m_pJpeg;
    IG4   * m_pG4;

    // The desired JPEG quality for each of the color layers.
    int m_fgJpegQuality;
    int m_bgJpegQuality;

    // m_clrPixRow represents number of color pixels per row which will be
    // identical in the foreground and the background.  It gets set in 
    // DrvStartPage().
    ULONG m_clrPixRow;

    // There are used for downsampling the image.  They are written to
    // in doBand().  They are flushed at the end of doBand().  They
    // dump the resulting data into the corresponding m_pBgBuf and 
    // m_pFgBuf objects.
    DownBuf  * m_pBgDownBuf;
    DownBuf  * m_pFgDownBuf;

    // Storage buffers where the three layers are stashed for processing
    // before compression.
    ColorStorageBuf * m_pBgBuf;
    ColorStorageBuf * m_pFgBuf;
    StorageBuf      * m_pSelectBuf;
};

#endif
//#ifndef OemTurboSeg_h_INCLUDED