#include "StdAfx.h"

#include <algorithm>

#include "Segmentor.h"
#include "misc.h"
#include "debug.h"

// This method takes an array of four bytes, which are supposed
// to represent a 4x4 square in a mono-chrome bitmap.  It 
// returns the vertical "transition count" for all of those 
// bytes.
inline static size_t vertTransCount( const unsigned char n[] )
{
    unsigned char x[3];

    x[0] = n[0] ^ n[1];
    x[1] = n[1] ^ n[2];
    x[2] = n[2] ^ n[3];

    
    size_t sum = 0;

    unsigned char mask = 0x01;

    for( size_t ii=0; ii != 4; ++ii, mask <<= 1 )
    {
        size_t val  = ((x[0] & mask) >> ii);
               val += ((x[1] & mask) >> ii);
               val += ((x[2] & mask) >> ii);

        sum += (val==3)? 4 : val;
    }

    return sum;
}




/* Put a 24 bit rgb pixel into one of 10 buckets:

    0 -- white (all bits set)
    1 -- black (all bits clear)

        (high order bit only)

       | red | green | blue  
-------+-----+-------+----------          
    2  |  0  |   0   |   0
    3  |  1  |   0   |   0
-------+-----+-------+----------
    4  |  0  |   1   |   0
    5  |  1  |   1   |   0
-------+-----+-------+----------
    6  |  0  |   0   |   1
    7  |  1  |   0   |   1
-------+-----+-------+----------
    8  |  0  |   1   |   1
    9  |  1  |   1   |   1

 Note that white and black black would fall into buckets 2 and
 9 if there were not checked for explicitly.

*/

inline size_t BinarySegmentor::bucket( const Ptr24Bit & pixel )
{
    // Note that the order of the tests here is very important for
    // speed.  We assume that a white is most common, then black.
    // Only if we have neither of these cases to we attempt to bucket.
    
    if( pixel.isWhite() )
    {
        return WHITE;
    }
    else if( pixel.isBlack() )
    {
        return BLACK;
    }

    const unsigned char r =  (pixel.r() >> 7);
    const unsigned char g = ((pixel.g() >> 7) << 1);
    const unsigned char b = ((pixel.b() >> 7) << 2);

    return (r | g | b) + 2;
}


// Operates on rows rows and cols columns.  Thus you can use this anywere
// in the image.  Note, however, that it automatically clears the 
// existing histogram before starting.
inline void BinarySegmentor::
buildHisto( Ptr24Bit image, const size_t delta, 
            size_t rows, const size_t cols )
{
    init();

    for( ; rows; --rows, image.offset(delta) )
    {
        Ptr24Bit pixPtr( image );

        const Ptr24Bit endPtr( pixPtr + cols );

        for( ; pixPtr != endPtr; ++pixPtr )
        {
            ++(m_histo[bucket(pixPtr)]);
        }
    }
}

// Operates on rows rows and cols columns.  Thus you can use this anywere
// in the image.  Note, however, that it automatically clears the 
// existing histogram before starting.
inline void BinarySegmentor::
buildHisto4x4( Ptr24Bit image, const size_t delta )
{
    init();

    const Ptr24Bit imageEnd( image.bytePtr() + (delta*4) );

    for( ; image != imageEnd; image.offset(delta) )
    {
        Ptr24Bit pixPtr( image );

        // Because we know the amount of the offset, getting the pointer
        // and adding 4 columns * 3 bytes ourselves is faster.
        const Ptr24Bit endPtr( pixPtr.bytePtr() + (4*3) );

        for( ; pixPtr != endPtr; ++pixPtr )
        {
            ++(m_histo[bucket(pixPtr)]);
        }
    }
}


/* 
 This returns a distance between 0 and 6 specifying the distance
 between two buckets.Identical buckets are obviuosly not
 separated.
 
 Buckets that are neither black nor white are separated by their
 hamming distance.  This makes sense because each bit represents
 1 color.

 Black and white buckets get two terms in their distance from other
 buckets.  The first term is the distance that would have been
 calculated if white and black did not have special buckets.  The
 second term is an offset (3 for white, 1 for black) that makes
 white and black more distant than other pixels.
*/

inline size_t BinarySegmentor::bucketDistance( size_t b1, size_t b2 )
{
    // Just get rid of the common case of identical pixels.
    if( b1 == b2 )
    {
        return 0;
    }

    // Note that from now on, b1 != b2 is implied.  Thus I can add to 
    // distance for white and black based only on one bucket without
    // checking the other.

    size_t dist = 0;

    switch( b1 )
    {
        case WHITE: dist += 3; b1  = 7; break;
        case BLACK: ++dist;    b1  = 0; break;
        default   :            b1 -= 2; break;
    }

    switch( b2 )
    {
        case WHITE: dist += 3; b2  = 7; break;
        case BLACK: ++dist;    b2  = 0; break;
        default   :            b2 -= 2; break;
    }

    // Compute the hamming distance.

    const size_t b1XorB2 = (b1 ^ b2);

    dist +=  (b1XorB2 & 0x01);
    dist += ((b1XorB2 & 0x02) >> 1);
    dist += ((b1XorB2 & 0x04) >> 2);

    return dist;
}

/*
 Returns the Euclidean distance squared between two pixels,
 plus some extra for black or white.
*/
inline size_t BinarySegmentor::
distance2( const Ptr24Bit & lhs, const Ptr24Bit & rhs )
{
    // If pixels are equal, we are done.
    if( lhs.hasSameData( rhs ) )
    {
        return 0;
    }

    // I made up the numbers for the white and black offsets
    // by messing with them and seeing what worked...
    // Sorry, no theory here.

    int offset = (lhs.isWhite() || rhs.isWhite())? 0x10000 : 0;

    if( lhs.isBlack() || rhs.isBlack() )
    {
        offset += 0x1000;
    }

    const int dr = int(lhs.r()) - int(rhs.r());
    const int dg = int(lhs.g()) - int(rhs.g());
    const int db = int(lhs.b()) - int(rhs.b());

    return (dr*dr) + (dg*dg) + (db*db) + offset;
}


/*
  Turns a bucket into the pixel at the center of said bucket.  The
  24 bit value is returned in the first three bytes of the return
  value.

  Operates on rows rows and m_8 columns.  You cannot use this on the last
  odd columns. Returns true if background bucket is white.

*/

inline unsigned int BinarySegmentor::bucketToPixel( size_t bucket )
{
    if( bucket == WHITE )
    {
        return 0x00ffffff;
    }
    else if( bucket == BLACK )
    {
        return 0x00;
    }

    bucket -= 2;

    unsigned int val = 0;

    Ptr24Bit ptr( reinterpret_cast<unsigned char *>(&val) );

    ptr.r() = (bucket & 0x01)? 0xc0 : 0x40;
    ptr.g() = (bucket & 0x02)? 0xc0 : 0x40;
    ptr.b() = (bucket & 0x04)? 0xc0 : 0x40;

    return val;
}

/*

  image -- points to the upper-left-hand corner of the square we are
           segmenting.
  iDelta -- the number of bytes needed to advance a full row in the
            image.  (This may not be a multiple of 3!)
  selector -- points the upper-left-hand corner of the square we are
              segmenting in the selector.
  sDelta -- the nubmer of bytes needed to advance one full row in
            the image.
  rows -- the number of rows in the square we are segmenting.
  blocksize -- the size of the block we are segmenting.  This ends
               up being the number of columns.  The reason that we
               need rows is to handle the odd rows at the end of each
               band.
*/

void BinarySegmentor::
makeSelector( Ptr24Bit image,  const size_t iDelta, 
              BitPtr selector, const size_t sDelta,
              const size_t rows, const size_t blockSize )
{
    buildHisto( image, iDelta, rows, blockSize );

    // In general, maxBuck is the bucket of greatest mass.  However
    // in the case of white, we require less mass.  Either way,
    // maxBuck is the first "cluster" in our crude clustering algorithm.
    // Start by guessing white is going to get a bucket.
    unsigned int maxSize = m_histo[0];
    size_t maxBuck = 0;

    // Do we have a significant number of white pixels?
    // I use rows here because I am thinking of a single
    // white row or column being significant.  That gives us
    // blockSize for the general case, but in the case that
    // we are at the bottom of the band, it gives us rows.
    if( maxSize < rows )
    {
        // If no, then find our best bucket.
                
        for( size_t idx = 1; idx != NUM_BUCKETS; ++idx )
        {
            if( m_histo[idx] > maxSize )
            {
                maxSize = m_histo[idx];
                maxBuck = idx;
            }
        }
    }

    // The ares in pixels of our current block.
    const size_t area = blockSize * rows;

    // Was there actually any segmentation?
    if( maxSize != area )
    {
        // farBuck is our other cluster.  It is the bucket furthest from
        // maxBuck using the bucketDistance function.  If there is a tie
        // for distance, the bucket with the greatest mass wins.

        //Start by guessing that if maxBuck was white, farBuck is
        //black, and vice-versa.

        size_t farDist   = 0;
        size_t farBuck   = !maxBuck;
        size_t farWeight = 0;
        
        // If either black or white is maxBuck, and the other color is
        // present, then there is no need to search for farBuck.
        if( (maxBuck > 1) || !m_histo[farBuck] )
        {
            // Otherwise, zip through the buckets looking for the bucket
            // furthest from maxBuck.
            for( size_t idx = 0; idx != NUM_BUCKETS; ++idx )
            {
                if( m_histo[idx] && (idx != maxBuck) )
                {
                    const size_t newDist = bucketDistance( idx, maxBuck );

                    if( (newDist > farDist)                                  ||
                        ((newDist == farDist) && (m_histo[idx] > farWeight))  )
                    {
                        farDist   = newDist;
                        farBuck   = idx;
                        farWeight = m_histo[idx];
                    }
                }
            }
        }

        // If there are only two buckets present, we can segment very quickly,
        // by testing only the buckets.
        if( (m_histo[maxBuck] + m_histo[farBuck]) == area )
        {
            for( size_t rIdx=0; rIdx != rows; ++rIdx,
                 selector.offset(sDelta), image.offset(iDelta) )
            {
                Ptr24Bit pixPtr( image );

                BitPtr selBitPtr( selector );

                const Ptr24Bit endPix( pixPtr + blockSize );

                for( ; pixPtr != endPix; ++pixPtr, ++selBitPtr )
                {
                    // We put the winning bucket in the foreground if it is
                    // non-white and the background if it is white.
                    const bool val = (bucket(pixPtr) == maxBuck);                             
                    selBitPtr.set( (maxBuck != WHITE)? val : !val );
                }
            }        
        }
        else // thus((m_histo[maxBuck] + m_histo[farBuck]) != area )
        {
            // We must segment based on the distance2() function.

            unsigned int maxPix = bucketToPixel( maxBuck );
            unsigned int farPix = bucketToPixel( farBuck );

            const Ptr24Bit maxPixPtr( reinterpret_cast<unsigned char *>(&maxPix) );
            const Ptr24Bit farPixPtr( reinterpret_cast<unsigned char *>(&farPix) );

            for( size_t rIdx=0; rIdx != rows; ++rIdx,
                 selector.offset(sDelta), image.offset(iDelta) )
            {
                Ptr24Bit pixPtr( image );

                BitPtr selBitPtr( selector );

                const Ptr24Bit endPix( pixPtr + blockSize );

                for( ; pixPtr != endPix; ++pixPtr, ++selBitPtr )
                {
                    const size_t dMax = (distance2(maxPixPtr, pixPtr) );
                    const size_t dFar = (distance2(farPixPtr, pixPtr) );

                    const bool val = (dMax < dFar);
                              
                    // We put the winning bucket in the foreground if it is
                    // non-white and the background if it is white.
                    selBitPtr.set( (maxBuck != WHITE)? val : !val );
                }
            }
        }
    }
    else //thus( maxSize == area )
    {
        // All we have to do is decide if we want to place stuff in
        // the foreground or the background.  We put the winning
        // bucket in the foreground if it is non-white and the background
        // if it is white.

        const bool val = maxBuck? 1 : 0;

        for( size_t rIdx=0; rIdx != rows; ++rIdx, selector.offset(sDelta) )
        {
            BitPtr selBitPtr( selector );

            for( size_t col=0; col != blockSize; ++col, ++selBitPtr )
            {                                  
                selBitPtr.set( val );
            }
        }
    }
}

/*

  makeSel4Fast() is a highly-optimized version of makeSelector()
  designed for our most common case, 4x4 squares.

  image -- points to the upper-left-hand corner of the square we are
           segmenting.
  iDelta -- the number of bytes needed to advance a full row in the
            image.  (This may not be a multiple of 3!)
  selector -- points the upper-left-hand corner of the square we are
              segmenting in the selector.
  nDelta -- the nubmer of nibbles needed to advance one full row in
            the image.
  smear -- are we smearing halftones?

  Other than nibble-specific coding, makeSel4Fast() is identical to
  makeSelector().
*/

void BinarySegmentor::
makeSel4Fast( Ptr24Bit image,     const size_t iDelta, 
              NibblePtr selector, const size_t nDelta )
{
#define BLOCK_SIZE 4

    //buildHisto( image, iDelta, BLOCK_SIZE, BLOCK_SIZE );
    buildHisto4x4( image, iDelta );

    // Start by guessing white is going to get a bucket.
    unsigned int maxSize = m_histo[0];
    size_t maxBuck = 0;

    // Do wehave a significant number of white pixels?
    if( maxSize < BLOCK_SIZE )
    {
        // If no, then find our best bucket.        
        for( size_t idx = 1; idx != NUM_BUCKETS; ++idx )
        {
            if( m_histo[idx] > maxSize )
            {
                maxSize = m_histo[idx];
                maxBuck = idx;
            }
        }
    }

    // Was there actually any segmentation?
    if( maxSize != 16 )
    {
        // If yes, then we must find the far bucket.  Here, what is going on
        // is that we are guessing/hoping that maxBuck is either black or 
        // white and the farBuck is either white or black.  If this assumption
        // is wrong, then we search the histogram.
        size_t farDist   = 0;
        size_t farBuck   = !maxBuck;
        size_t farWeight = 0;

        if( (maxBuck > 1) || !m_histo[farBuck] )
        {
            for( size_t idx = 0; idx != NUM_BUCKETS; ++idx )
            {
                if( m_histo[idx] && (idx != maxBuck) )
                {
                    const size_t newDist = bucketDistance( idx, maxBuck );

                    if( (newDist > farDist)                                  ||
                        ((newDist == farDist) && (m_histo[idx] > farWeight))  )
                    {
                        farDist   = newDist;
                        farBuck   = idx;
                        farWeight = m_histo[idx];
                    }
                }
            }
        }

        // Here we are cashing the result of the test for
        // whiteness.  The reason that 0x11 is used is for
        // nibble optimization.  This, combined with set4Fast()
        // allows us to set the correct nibble (in the byte)
        // without testing which nibble we are on.  (Basically
        // it saves us a comparison and a shift.)
        const unsigned char  eqMax = (maxBuck != WHITE)? 0x11 : 0x00;
        const unsigned char neqMax = (eqMax ^ 0x11);

        if( (m_histo[maxBuck] + m_histo[farBuck]) == 16 )
        {
            NibblePtr selPtr( selector );

            for( size_t rIdx=0; rIdx != BLOCK_SIZE; ++rIdx, 
                 image.offset(iDelta), selPtr += nDelta )
            {
                Ptr24Bit pixPtr( image );

                unsigned char val = 0;

                const Ptr24Bit endPix( pixPtr + BLOCK_SIZE );

                // Note that our loop constructs our nibble.
                for( ; pixPtr != endPix; ++pixPtr )
                {
                    val <<= 1;

                    val |= (bucket(pixPtr) == maxBuck)? eqMax : neqMax;
                }
            
                // Only at the end to we actually make the changes.
                selPtr.set4Fast( val );
            } 

            // Now we look for half-tones to smooth out if we are smearing.
            if( m_smearHalftones )
            {
                NibblePtr selOscPtr( selector );

                // The horizontal transition number.
                size_t hTrans = 0;

                // The bytes we are examining.  If hTrans exceeds our threshold
                // then we will use this for finding the vertical transition
                // number.
                unsigned char selSq[4];
            
                for( size_t row=0; row != BLOCK_SIZE; ++row, selOscPtr += nDelta )
                {
                    selSq[row] = *selOscPtr;
                    hTrans += selOscPtr.transCount();
                }

                if( (hTrans > 15) && (vertTransCount(selSq) > 15) )
                {
                    for( size_t r=0; r != BLOCK_SIZE; ++r, selector += nDelta )
                    {
                        selector.setAll();
                    }
                }
            }
        }
        else //thus( (m_histo[maxBuck] + m_histo[farBuck]) != 16 )
        {
            // This section is like the section above except that we
            // have at least three buckets so we must use our distance
            // test.
            unsigned int maxPix = bucketToPixel( maxBuck );
            unsigned int farPix = bucketToPixel( farBuck );

            const Ptr24Bit maxPixPtr( reinterpret_cast<unsigned char *>(&maxPix) );
            const Ptr24Bit farPixPtr( reinterpret_cast<unsigned char *>(&farPix) );

            const unsigned char  eqMax = (maxBuck != WHITE)? 0x11 : 0x00;
            const unsigned char neqMax = (eqMax ^ 0x11);

            NibblePtr selPtr( selector );

            for( size_t rIdx=0; rIdx != BLOCK_SIZE; ++rIdx, 
                 image.offset(iDelta), selPtr += nDelta )
            {
                Ptr24Bit pixPtr( image );

                unsigned char val = 0;

                const Ptr24Bit endPix( pixPtr + BLOCK_SIZE );

                for( ; pixPtr != endPix; ++pixPtr )
                {
                    val <<= 1;

                    const size_t dMax = (distance2(maxPixPtr, pixPtr) );
                    const size_t dFar = (distance2(farPixPtr, pixPtr) );

                    val |= (dMax < dFar)? eqMax : neqMax;
                }

                selPtr.set4Fast( val );
            } 

            NibblePtr selOscPtr( selector );

            size_t hTrans = 0;

            unsigned char selSq[4];

            for( size_t row=0; row != BLOCK_SIZE; ++row, selOscPtr += nDelta )
            {
                selSq[row] = *selOscPtr;
                hTrans += selOscPtr.transCount();
            }

            if( (hTrans > 9) && (vertTransCount(selSq) > 9) )
            {
                for( size_t r=0; r != BLOCK_SIZE; ++r, selector += nDelta )
                {
                    selector.setAll();
                }
            }
        }
    }
    else // thus( maxSize == 16 )
    {
        // We have only one bucket.  All we need to do is set the
        // selector appropriately for white or not white.
        if( maxBuck != WHITE )
        {
            for( size_t r=0; r != BLOCK_SIZE; ++r, selector += nDelta )
            {
                selector.setAll();
            }
        }
        else // thus( maxBuck == WHITE )
        {
            for( size_t r=0; r != BLOCK_SIZE; ++r, selector += nDelta )
            {
                selector.clearAll();
            }
        }
    }

#undef BLOCK_SIZE
}

void BinarySegmentor::segment( void )
{
    const size_t rowBlocks = m_rows / m_ratio;
    const size_t colBlocks = m_cols / m_ratio;
    
    // For hi-resolution images, 4x4 is our most common case, thus
    // we treat it separately.
    if( m_ratio == 4 )
    {
        const size_t blockSize = 4;

        // The distance in nibbles between the beginning of two
        // scanlines.
        const size_t nDelta = m_selectDelta * 2;

        // The distance in bytes between the beginning of two
        // row blocks in the image.
        const size_t iRBDelta = m_colorDelta * blockSize;

        // The distance in bytes between the beginning of two
        // row blocks in the selector.
        const size_t sRBDelta = m_selectDelta * blockSize;

        Ptr24Bit rowPtr( m_image );

        unsigned char * selRowPtr = m_selector;

        // The number of pixels in full blocks horizontally.
        // This is an optimization.  Inside the loop below, we
        // need this value for each rowBlock.
        const size_t blockDistance = (blockSize * colBlocks);

        for( size_t rowBlock = 0; rowBlock < rowBlocks; ++rowBlock,
             rowPtr.offset(iRBDelta), selRowPtr += sRBDelta )
        {
            Ptr24Bit  pixPtr( rowPtr );
            NibblePtr selPtr( selRowPtr );
            
            const Ptr24Bit endPix( pixPtr + blockDistance );

            for( ;pixPtr != endPix; pixPtr += blockSize, ++selPtr )
            {
                makeSel4Fast( pixPtr, m_colorDelta, selPtr, nDelta );
            }
        }
    }
    else  if( m_ratio != 1 )
    {
        // We check to see if m_ratio is 1 here because there
        // is no first-pass segmentation if there is no downsampling.

        Ptr24Bit rowPtr( m_image );
        BitPtr selRowPtr( m_selector );

        // These are not really important to understanding the code.
        // They are just expressions whose values do not change for
        // the life of the enclosed loops.  Therefore, I pulled them
        // out for speed and clarity.
        const size_t downDelta     = m_ratio * m_colorDelta;
        const size_t selDownDelta  = m_ratio * m_selectDelta;
        const size_t downColBlocks = m_ratio * colBlocks;

        const Ptr24Bit endRow( m_image.bytePtr() + (rowBlocks * downDelta) );

        for( ; rowPtr != endRow; rowPtr.offset(downDelta),
             selRowPtr.offset(selDownDelta) )
        {
            Ptr24Bit pixPtr( rowPtr );
            BitPtr selPtr( selRowPtr );

            const Ptr24Bit endPix( rowPtr + downColBlocks );

            for( ;pixPtr != endPix; pixPtr += m_ratio, selPtr += m_ratio )
            {
                makeSelector( pixPtr, m_colorDelta, selPtr, 
                              m_selectDelta, m_ratio, m_ratio );
            }
        }
    }

    // Now we handle any rows that are left-over at the bottom of
    // the band.  We do not bother with worrying about 4x4
    // optimizations here since the overall algorithm is basically
    // the same regardless.  This code handles such a small fraction
    // of the image that there is no need to optimize.

    const size_t remainderRows = m_rows % m_ratio;

    // Note that if m_ratio is 1, then this block gets skipped.
    if( remainderRows )
    {
        Ptr24Bit imagePtr( m_image + (rowBlocks * m_ratio * m_colorDelta) );
        BitPtr selPtr( m_selector + (rowBlocks * m_ratio * m_selectDelta) );

        for( size_t colBlock = 0; colBlock < colBlocks;
             ++colBlock, imagePtr += m_ratio, selPtr += m_ratio )
        {
            makeSelector( imagePtr, m_colorDelta, selPtr,
                          m_selectDelta, remainderRows, m_ratio );
        }
    }
}

inline size_t JpegSegmentor::bucket( const Ptr24Bit & pixel )
{
    // Note that the order of the tests here is very important for
    // speed.  We assume that a white is most common, then black.
    // Only if we have neither of these cases to we attempt to bucket.
    
    if( pixel.isWhite() )
    {
        return WHITE;
    }
    else if( pixel.isBlack() )
    {
        return BLACK;
    }
    else if( pixel.isDontCare() )
    {
        return DONT_CARE;
    }

    const unsigned char r = (pixel.r() >> 7);
    const unsigned char g = (pixel.g() >> 7);
    const unsigned char b = (pixel.b() >> 7);

    return ((b << 2) | (g << 1) | r) + 2;
}

inline size_t JpegSegmentor::distance( size_t b1, size_t b2 )
{
    // Just get rid of the common case of identical pixels.
    if( b1 == b2 )
    {
        return 0;
    }

    // Note that from now on, b1 != b2 is implied.  Thus I can add to 
    // distance for white and black based only on one bucket without
    // checking the other.

    size_t dist = 0;

    switch( b1 )
    {
        case WHITE: dist += 3; b1  = 7; break;
        case BLACK: ++dist;    b1  = 0; break;
        default   :            b1 -= 2; break;
    }

    switch( b2 )
    {
        case WHITE: dist += 3; b2  = 7; break;
        case BLACK: ++dist;    b2  = 0; break;
        default   :            b2 -= 2; break;
    }

    // Compute the hamming distance.

    const size_t b1XorB2 = (b1 ^ b2);

    dist +=  (b1XorB2 & 0x01);
    dist += ((b1XorB2 & 0x02) >> 1);
    dist += ((b1XorB2 & 0x04) >> 2);

    return dist;
}

inline void JpegSegmentor::
buildHisto( const size_t rowBlock, const size_t colBlock, 
                  size_t numRows,  const size_t numCols )
{
    initHisto();
    
    const size_t colBase = colBlock * 8;
    const size_t rowBase = rowBlock * 8;

    Ptr24Bit fgRowPtr( getFgPix( rowBase, colBase ) );
    Ptr24Bit bgRowPtr( getBgPix( rowBase, colBase ) );

    const size_t delta = getColorDelta();

    for(; numRows; --numRows, fgRowPtr.offset(delta), bgRowPtr.offset(delta) )
    {
        Ptr24Bit fgPtr( fgRowPtr );
        Ptr24Bit bgPtr( bgRowPtr );

        for( const Ptr24Bit endFg(fgPtr + numCols); fgPtr != endFg; ++fgPtr, ++bgPtr )
        {
            ++(m_histo[bucket(fgPtr)]);
            ++(m_histo[bucket(bgPtr)]);
        }    
    }
}

inline void JpegSegmentor::
flipSelectSquare( BitPtr selBitPtr, size_t numRows )
{
    if( m_ratio == 4 )
    {
        const size_t nibDelta = (getSelectDelta() * 2);

        for( NibblePtr ptr( selBitPtr ); numRows; --numRows, ptr += nibDelta )
        {
            ptr.flipAll();
        }
    }
    else
    {
        const size_t delta = getSelectDelta() * 8;

        for( ; numRows; --numRows, selBitPtr += delta )
        {
            BitPtr ptr( selBitPtr );

            for( size_t ii=0; ii != m_ratio; ++ii, ++ptr )
            {
                ptr.flip();
            }
        }
    }
}

// JBTODO:: I am not sure how this function interacts with the possibility that
// there is no color in the background, etc.
void JpegSegmentor::
fixSelector( const size_t rowBlock, const size_t colBlock,
             size_t numRows,        const size_t numCols,
             const size_t lastSize )
{
    buildHisto( rowBlock, colBlock, numRows, numCols );
    
    // Start by guessing white is going to get a bucket.
    unsigned int maxSize = m_histo[0];
    size_t maxBuck = 0;

    // Do we have a significant number of white pixels?
    // I use blockSize here because I am thinking of a single
    // white row or column being significant.
    if( maxSize < numRows )
    {
        // If no, then find our best bucket.

        for( size_t idx = 1; idx != NUM_BUCKETS; ++idx )
        {
            if( m_histo[idx] > maxSize )
            {
                maxSize = m_histo[idx];
                maxBuck = idx;
            }
        }
    }
    
    size_t farDist = 0;
    size_t farBuck = maxBuck;

    for( size_t idx = 0; idx != NUM_BUCKETS; ++idx )
    {
        if( m_histo[idx] )
        {
            const size_t newDist = distance( idx, maxBuck );

            if( newDist > farDist )
            {
                farDist = newDist;
                farBuck = idx;
            }
        }
    }

    const size_t colBase = (colBlock * 8);
    const size_t rowBase = (rowBlock * 8);

    Ptr24Bit fgRowPtr( getFgPix(rowBase, colBase) );
    Ptr24Bit bgRowPtr( getBgPix(rowBase, colBase) );

    const size_t cDelta = getColorDelta();
    const size_t sDelta = (getSelectDelta() * m_ratio * 8);

    BitPtr selectRowPtr = getSelectPix( rowBase * m_ratio, colBase * m_ratio );

    for( ; numRows; --numRows, fgRowPtr.offset(cDelta),
         bgRowPtr.offset(cDelta), selectRowPtr += sDelta )
    {
        Ptr24Bit fgPtr( fgRowPtr );
        Ptr24Bit bgPtr( bgRowPtr );

        const size_t selRows = (numRows == 1)? lastSize : m_ratio;

        BitPtr selectPtr( selectRowPtr );

        const Ptr24Bit endFg( fgPtr + numCols );

        for( ; fgPtr != endFg; selectPtr += m_ratio, ++fgPtr, ++bgPtr )
        {

            size_t fgBuck = bucket( fgPtr );
            size_t bgBuck = bucket( bgPtr );
            
            bool flip = false;

            if( fgBuck == maxBuck )
            {
                flip = !maxBuck;
            }
            else if( bgBuck == maxBuck )
            {
                flip = !!maxBuck;
            }
            else if( fgBuck == DONT_CARE )
            {
                flip = !maxBuck;
            }
            else if( bgBuck == DONT_CARE )
            {
                flip = !!maxBuck;
            }
            else
            {
                // We needt to reverse the roles of the foreground and background
                // if the maxBuck is white.
                if( maxBuck == WHITE ) { std::swap( fgBuck, bgBuck ); }

                const size_t dFgMax = distance(fgBuck, maxBuck);
                const size_t dBgMax = distance(bgBuck, maxBuck);

                if( dFgMax > dBgMax )
                {
                    flip = true;
                }
                else if( dFgMax == dBgMax )
                {
                    const size_t dFgFar = distance(fgBuck, farBuck);
                    const size_t dBgFar = distance(bgBuck, farBuck);

                    flip = (dBgFar > dFgFar);
                }
            }

            if( flip )
            {
                flipSelectSquare( selectPtr, selRows );
                fgPtr.swapData( bgPtr );
            }

        }
    }
}

void JpegSegmentor::
bleedSquare( const Ptr24Bit & image, size_t numRows,
             const size_t numCols, const unsigned char fill )
{
    size_t count = 0;

    int sum[3] = {0, 0, 0};
    
    const size_t delta = getColorDelta();

    Ptr24Bit rowPtr( image );

    {for( size_t rowIdx = 0; rowIdx != numRows; ++rowIdx, rowPtr.offset(delta) )
    {
        Ptr24Bit pixPtr( rowPtr );

        const Ptr24Bit endPix( rowPtr + numCols );

        for( ; pixPtr != endPix; ++pixPtr )
        {
            if( !pixPtr.isDontCare() )
            {
                ++count;
                sum[0] += pixPtr.r();
                sum[1] += pixPtr.g();
                sum[2] += pixPtr.b();
            }

        }    
    }}


    if( !count )
    {        
        const size_t bytes = numCols * 3;

        for( rowPtr = image; numRows; --numRows, rowPtr.offset(delta) )
        {        
            memset( rowPtr.bytePtr(), fill, bytes );
        }        
    }
    else if( count != (numRows * numCols) )
    {
        sum[0] /= count;
        sum[1] /= count;
        sum[2] /= count;

        for( rowPtr=image; numRows; --numRows, rowPtr.offset(delta) )
        {        
            Ptr24Bit pixPtr( rowPtr );

            const Ptr24Bit endPix( rowPtr + numCols );

            for( ; pixPtr != endPix; ++pixPtr )
            {
                if( pixPtr.isDontCare() )
                {
                    pixPtr.setRGB( sum[0], sum[1], sum[2] );
                }                   
            }    
        }
    }
}


void JpegSegmentor::
bleedQuads( const Ptr24Bit & image, size_t numRows,
             const size_t numCols, const unsigned char fill )
{
    size_t qCount[4] = { 0, 0, 0, 0 };

    int qSum[4][3] = { {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0} };

    const size_t delta = getColorDelta();

    Ptr24Bit rowPtr( image );

    {for( size_t rowIdx = 0; rowIdx != numRows; ++rowIdx, rowPtr.offset(delta) )
    {
        Ptr24Bit pixPtr( rowPtr );

        for( size_t colIdx = 0; colIdx != numCols; ++colIdx, ++pixPtr )
        {
            if( !pixPtr.isDontCare() )
            {
                const size_t quad = ((rowIdx/4) * 2) + (colIdx/4);

                ++qCount[quad];
                qSum[quad][0] += pixPtr.r();
                qSum[quad][1] += pixPtr.g();
                qSum[quad][2] += pixPtr.b();
            }

        }    
    }}

    size_t count = 0;
    int sum[3] = {0, 0, 0};

    {for( size_t quad = 0; quad != 4; ++quad )
    {
        count += qCount[quad];
        sum[0] += qSum[quad][0];
        sum[1] += qSum[quad][1];
        sum[2] += qSum[quad][2];
    }}
   
    if( !count )
    {        
        const size_t bytes = numCols * 3;

        for( rowPtr = image; numRows; --numRows, rowPtr.offset(delta) )
        {        
            memset( rowPtr.bytePtr(), fill, bytes );
        }        
    }
    else if( count != (numRows * numCols) )
    {
        sum[0] /= count;
        sum[1] /= count;
        sum[2] /= count;

        {for( size_t quad = 0; quad != 4; ++quad )
        {
            if( qCount[quad] )
            {
                qSum[quad][0] /= qCount[quad];
                qSum[quad][1] /= qCount[quad];
                qSum[quad][2] /= qCount[quad];
            }
            else
            {
                qSum[quad][0] = sum[0];
                qSum[quad][1] = sum[1];
                qSum[quad][2] = sum[2];
            }
        }}

        rowPtr = image;
        
        for( size_t rowIdx = 0; rowIdx != numRows; ++rowIdx, rowPtr.offset(delta) )
        {        
            Ptr24Bit pixPtr( rowPtr );

            for( size_t colIdx = 0; colIdx != numCols; ++colIdx, ++pixPtr )
            {
                if( pixPtr.isDontCare() )
                {
                    const size_t quad = ((rowIdx/4) * 2) + (colIdx/4);

                    pixPtr.setRGB( qSum[quad][0], qSum[quad][1], qSum[quad][2] );
                }                   
            }    
        }
    }
}

void JpegSegmentor::resegment( void )
{

    const size_t numFullRowBlocks = m_rows / (8 * m_ratio);
    const size_t numFullColBlocks = m_cols / (8 * m_ratio);

    {for( size_t rowBlock = 0; rowBlock < numFullRowBlocks; ++rowBlock )
    {
        Ptr24Bit fgPixPtr( getFgRow(rowBlock * 8) );
        Ptr24Bit bgPixPtr( getBgRow(rowBlock * 8) );

        for( size_t colBlock = 0; colBlock < numFullColBlocks; ++colBlock,
             fgPixPtr += 8, bgPixPtr += 8 )
        {
            fixSelector( rowBlock, colBlock, 8, 8, m_ratio );
            bleedQuads( fgPixPtr, 8, 8, 0x00 );
            bleedQuads( bgPixPtr, 8, 8, 0xff );
        }
    }}

    size_t remainderRows = m_rows % (8 * m_ratio);

    remainderRows = (remainderRows / m_ratio) + !!(remainderRows % m_ratio);

    if( remainderRows )
    {
        size_t lastSize = m_rows % m_ratio;

        if( !lastSize )
        {
            lastSize = m_ratio;
        }

        for( size_t colBlock = 0; colBlock < numFullColBlocks; ++colBlock )
        {
            fixSelector( numFullRowBlocks, colBlock, remainderRows, 8, lastSize );

            const size_t row = numFullRowBlocks * 8;
            const size_t col = colBlock * 8;

            Ptr24Bit fgPtr( getFgPix( row, col ) );
            Ptr24Bit bgPtr( getBgPix( row, col ) );

            bleedSquare( fgPtr, remainderRows, 8, 0x00 );
            bleedSquare( bgPtr, remainderRows, 8, 0xff );                
        }
    }
}
