#ifndef Segmentor_H_INCLUDED
#define Segmentor_H_INCLUDED

#include "misc.h"

enum Buckets
{
    WHITE            = 0,
    BLACK            = 1,
    DONT_CARE        = 10,
    NUM_BUCKETS      = 10,
    NUM_JPEG_BUCKETS = 11

};

/*

 In our two-pass segmentation scheme, BinarySegmentor operates on squares
 the size of the downsample area (usually a 4x4.)  Visit each 4x4 area of
 the image.  On each area you call makeSelector().  The image will not be
 modified, but the selector will be generated.

 The working assumption here is that any given 4x4 area of the image can
 reasonably by a pallette of one or two colors.  If the pallette contains
 only one color, then the selector is trivial.  If it contains two then
 it selects the color for each pixel.

*/

class BinarySegmentor
{
public:
    
    BinarySegmentor( void )
    :
     m_rows(0),
     m_cols(0),
     m_selector(NULL),
     m_ratio(1),
     m_smearHalftones( false ),
     m_selectDelta(0),
     m_colorDelta(0)
    {
    }

    void setSmearing( bool b ) { m_smearHalftones = b; }

    void addImage( const Ptr24Bit & image, const size_t delta )
    {
        m_image      = image;
        m_colorDelta = delta;
    }

    void addSelector( unsigned char * sel, size_t ratio, size_t rows, 
                      size_t cols, size_t delta )
    {
       m_rows        = rows;
       m_cols        = cols;
       m_selector    = sel;
       m_ratio       = ratio;
       m_selectDelta = delta;
    }

    void segment( void );

    virtual ~BinarySegmentor( void ) {}

protected:
private:

    // Disable copy semantics.
    BinarySegmentor( const BinarySegmentor & );
    BinarySegmentor & operator=( const BinarySegmentor & );

    // The histogram.
    size_t m_histo[NUM_BUCKETS];

    Ptr24Bit m_image;

    unsigned char * m_selector;
    
    size_t m_ratio;

    size_t m_selectDelta, m_colorDelta;

    size_t m_rows, m_cols;

    bool m_smearHalftones;

    // Clear out the histogram.
    void init( void ) { memset( m_histo, 0x00, sizeof(m_histo) ); }

    // Build the histogram for a given area.
    void buildHisto( Ptr24Bit image, const size_t delta,
                     size_t rows, const size_t cols );

    // A highly optimized version of the histogram builder for the 4x4
    // case.  It is mathemtatically equivalent to calling buildHisto()
    // for four rows and four columns.
    void buildHisto4x4( Ptr24Bit image, const size_t delta );

    // Return the bucket that this pixel lies in.
    static size_t bucket( const Ptr24Bit & );
    
    // Turns a bucket into the pixel at the center of the bucket.
    unsigned int bucketToPixel( size_t bucket );

    // Move the pointer "image" to the begining of each down-sampling
    // square in the image.
    void makeSelector( Ptr24Bit image,  const size_t iDelta, 
                       BitPtr selector, const size_t sDelta,
                       const size_t rows, const size_t blockSize );

    // Like makeSelector() above, but highly optimized for 4x4 case.
    // Also, this version does half-tone smearing.
    void makeSel4Fast( Ptr24Bit image,     const size_t iDelta, 
                       NibblePtr selector, const size_t nDelta );


    //
    // Note that bucketDistance() and distance2() are not directly
    // related!  You should not compare outputs.
    //

    // The distance between two buckets.
    static size_t bucketDistance( size_t, size_t );

    // The distance between two pixels.
    static size_t distance2( const Ptr24Bit &, const Ptr24Bit & );

};

/*

 The Jpeg Segmentor is the second pass in our segmentation.  After
 we have finished segmenting using the binary segmentor and downsampling,
 we still have some freedom left to help our compressor.  Each downsample
 block in the original image now corresponds to a pixel in the foreground
 and another in the background (with the selector choosing which to use.)
 We still have the freedom to exchange these two pixels if we invert the
 associated selector square.  JpegSegmentor's job is to decide if this
 exchange and inversion is needed and then do it.

 After this step there may still be don't-care pixels left from the previous
 steps.  JpegSegmentor chooses a suitable color for the don't-care pixels
 to help Jpeg to its job better.

 To use JpegSegmentor, you must call addColor() with the foreground and
 background.  Then you call addSelector() with the selector, the
 downsampling ratio, and the x resolution.

*/

class JpegSegmentor
{
public:

    JpegSegmentor( void )
    : m_rows(0),
      m_cols(0),
      m_selector(NULL),
      m_ratio(1),
      m_selectDelta(0),
      m_colorDelta(0)
    {
    }

    virtual ~JpegSegmentor( void ) {}

    void addColor( const Ptr24Bit & fg, const Ptr24Bit & bg )
    {
        m_fg = fg;
        m_bg = bg;
    }

    void addSelector( unsigned char * sel, size_t ratio, size_t rows, size_t cols )
    {
       m_rows = rows;
       m_cols = cols;

       m_selector   = sel;
       m_ratio      = ratio;

       m_selectDelta = ((cols + 7) & ~ULONG(7)) / 8;
       m_colorDelta  = ((cols / m_ratio) + !!(cols % m_ratio)) * 3;
    }

    void resegment( void );

protected:
private:

    // Disable copy semantics.
    JpegSegmentor( const JpegSegmentor & );

    JpegSegmentor & operator=( const JpegSegmentor & );

    Ptr24Bit m_fg, m_bg;

    unsigned char * m_selector;
    
    size_t m_ratio;

    size_t m_selectDelta, m_colorDelta;

    size_t m_rows, m_cols;

    unsigned int m_histo[NUM_JPEG_BUCKETS];

    // Clear out the histogram.
    void initHisto( void ) { memset( m_histo, 0x00, sizeof(m_histo) ); }

    void buildHisto( const size_t, const size_t, size_t, const size_t );

    void flipSelectSquare( BitPtr, size_t );

    size_t getSelectDelta( void ) const { return m_selectDelta; }
    size_t getColorDelta ( void ) const { return m_colorDelta;  }
    
    Ptr24Bit getFg( void ) { return m_fg; }
    Ptr24Bit getBg( void ) { return m_bg; }

    Ptr24Bit getFgRow(size_t r) {return getFg().offset(r*getColorDelta());}
    Ptr24Bit getBgRow(size_t r) {return getBg().offset(r*getColorDelta());}

    Ptr24Bit getFgPix(size_t r, size_t c) {return getFgRow(r) + c;}
    Ptr24Bit getBgPix(size_t r, size_t c) {return getBgRow(r) + c;}
    
    unsigned char * getSelect( void ) { return m_selector; }

    unsigned char * getSelectRow(size_t r)
    {
        return getSelect() + (r*getSelectDelta());
    }

    BitPtr getSelectPix(size_t r, size_t c)
    {
        BitPtr ptr( getSelectRow(r) );

        ptr += c;

        return ptr;
    }

    void fixSelector( const size_t, const size_t, size_t, const size_t, const size_t );

    void bleedSquare( const Ptr24Bit &, size_t, const size_t, const unsigned char );

    void bleedQuads( const Ptr24Bit &, size_t,  const size_t, const unsigned char );

    static size_t bucket( const Ptr24Bit & );

    static size_t distance( size_t, size_t );
};



#endif //#ifndef Segmentor_H_INCLUDE