#include "stdafx.h"

#include "TiffOutSurf.h"
#include "debug.h"

#define NOCOMPRESS 1

static Stopwatch TiffOutSurfWriteTmr( "TiffOutSurf::writeSpoolBuf" );

TiffOutSurf::~TiffOutSurf( void )
{
    DbgStream() << TiffOutSurfWriteTmr << std::endl;
}


void TiffOutSurf::
init( size_t xResolution, size_t yResolution, size_t colorDepth )
{
    DBG_TRACE_OBJ( "TiffOutSurf::init" );
    DBG_WATCH( "TiffOutSurf instance", m_fileName );

    // First we must set up our image resolution variables.  We
    // will reference these later!
    m_xRes[0] = xResolution;
    m_yRes[0] = yResolution;
    m_xRes[1] = 1;
    m_yRes[1] = 1;

    m_bitsPixel = colorDepth;

    // The next IFD will start after the header and the complete image.
    // For the tiff header, we just set the value directly.  For m_ifd
    // we set nextIfdOffset to point just past the tiff header.  This
    // is because of how we do our booking keeping.  send*PageIfd() will
    // add the size of the image and then the size of the IFD to it.
    // (I think this is simpler than the alternative, althought it seems
    // somewhat counter-intuitive to me.  The issue that this bookeeping
    // scheme neatly takes care of is the issue of m_fields[5], strip
    // offset.)
    m_ifd.m_nextIfdOffset = m_th.bufSize();
    m_th.setIfdOffset( m_ifd.m_nextIfdOffset + bytesImage() );

    // Initialize the IFD struture.
    m_ifd.m_fields[0].setLongValue( xRes() ); //Image Width --  # bits in scanline
    m_ifd.m_fields[1].setLongValue( yRes() ); //Image Length
    m_ifd.m_fields[2].setShortValue( (bitsPixel() == 1)? 1 : 8 ); // Bits per sample.
    m_ifd.m_fields[3].setShortValue( NOCOMPRESS ); //Compression   
    m_ifd.m_fields[4].setShortValue( (bitsPixel() == 1)? 1 : 2 ); //Photometric Interpretation   

    m_ifd.m_fields[6].setShortValue( (bitsPixel() == 1)? 1 : 3 ); // Samples per pixel.
    m_ifd.m_fields[7].setLongValue( yRes() );  //Rows per strip
    m_ifd.m_fields[8].setLongValue( bytesImage() ); //Strip byte count -- # bytes per strip    

    // Next, delete any file that used to be there.
    std::ios_base::openmode mode = ( std::ios_base::out    |
                                     std::ios_base::trunc  |
                                     std::ios_base::binary );
   
    std::ofstream spoolStream( m_fileName.c_str(), mode );
    spoolStream.close();

    writeSpoolBuf( m_th.getBuf(), m_th.bufSize() );    
}


ULONG TiffOutSurf::bytesRow( void ) const
{
    // Round up to next byte.
    return (((bitsPixel() * xRes()) + 7) & ~ULONG(7)) / 8;
}

ULONG TiffOutSurf::bytesImage( void ) const
{
    // Round up to next integer word.
    return ((bytesRow() * yRes()) + 1) & ~ULONG(1);
}

HRESULT TiffOutSurf::
writeSpoolBuf( unsigned char * buf, size_t cbSize )
{
    TiffOutSurfWriteTmr.start();

    DBG_TRACE_OBJ( "TiffOutSurf::writeSpoolBuf" );
    DBG_WATCH( "TiffOutSurf instance", m_fileName );

    DBG_WATCH( "cbSize", cbSize );

    std::ios_base::openmode mode = ( std::ios_base::out    |
                                     std::ios_base::app    |
                                     std::ios_base::binary );
    
    std::ofstream spoolStream( m_fileName.c_str(), mode );
    spoolStream.write( reinterpret_cast<const char*>(buf), cbSize );
    spoolStream.close();

    TiffOutSurfWriteTmr.stop();

    return S_OK;
}

void TiffOutSurf::sendVeryLastPageIfd( void )
{    
    DBG_TRACE_OBJ( "TiffOutSurf::sendVeryLastPageIfd" );
    DBG_WATCH( "TiffOutSurf instance", m_fileName );
    
    m_ifd.m_fields[5].setLongValue( m_ifd.m_nextIfdOffset ); //Strip Offsets -- offset to first scan line
    m_ifd.m_nextIfdOffset += bytesImage() + m_ifd.bufSize();

    m_ifd.m_fields[ 9].setOffset( m_ifd.m_nextIfdOffset );// offset to X Resolution
    m_ifd.m_nextIfdOffset += sizeof( m_xRes );

    m_ifd.m_fields[10].setOffset( m_ifd.m_nextIfdOffset );// offset to Y Resolution
    m_ifd.m_nextIfdOffset += sizeof( m_yRes );

    m_ifd.m_nextIfdOffset = 0;

    // We need to word align the image if it isn't already.
    if( bytesImage() > (bytesRow() * yRes()) )
    {
        unsigned char byte = 0x00;
        writeSpoolBuf( &byte, sizeof(byte) );
    }

    writeSpoolBuf( m_ifd.getBuf(), m_ifd.bufSize() );
    writeSpoolBuf( reinterpret_cast<unsigned char*>(m_xRes), sizeof(m_xRes) );
    writeSpoolBuf( reinterpret_cast<unsigned char*>(m_yRes), sizeof(m_yRes) );
}

void TiffOutSurf::
sendLastPageIfd( size_t newXRes, size_t newYRes, size_t newColorDepth )
{
    DBG_TRACE_OBJ( "TiffOutSurf::sendLastPageIfd" );
    DBG_WATCH( "TiffOutSurf instance", m_fileName );

    m_ifd.m_fields[5].setLongValue( m_ifd.m_nextIfdOffset ); //Strip Offsets -- offset to first scan line
    m_ifd.m_nextIfdOffset += bytesImage() + m_ifd.bufSize();

    m_ifd.m_fields[ 9].setOffset( m_ifd.m_nextIfdOffset );// offset to X Resolution
    m_ifd.m_nextIfdOffset += sizeof( m_xRes );

    m_ifd.m_fields[10].setOffset( m_ifd.m_nextIfdOffset );// offset to Y Resolution
    m_ifd.m_nextIfdOffset += sizeof( m_yRes );

     // We need to word align the image if it isn't already.
    if( bytesImage() > (bytesRow() * yRes()) )
    {
        unsigned char byte = 0x00;
        writeSpoolBuf( &byte, sizeof(byte) );
    }

    const size_t newBitsPixel = newColorDepth;
    const size_t newBytesRow = (((newBitsPixel * newXRes) + 7) & ~ULONG(7)) / 8;
    const size_t newBytesImage = ((newBytesRow * newYRes) + 1) & ~ULONG(1);

    m_ifd.m_nextIfdOffset += newBytesImage;

    writeSpoolBuf( m_ifd.getBuf(), m_ifd.bufSize() );
    writeSpoolBuf( reinterpret_cast<unsigned char*>(m_xRes), sizeof(m_xRes) );
    writeSpoolBuf( reinterpret_cast<unsigned char*>(m_yRes), sizeof(m_yRes) );

    m_ifd.m_nextIfdOffset -= newBytesImage;

    m_xRes[0] = newXRes;
    m_yRes[0] = newYRes;
    m_xRes[1] = 1;
    m_yRes[1] = 1;

    m_bitsPixel = newBitsPixel;

    //Reinitialize the structure.
    m_ifd.m_fields[0].setLongValue( xRes() ); //Image Width --  # bits in scanline
    m_ifd.m_fields[1].setLongValue( yRes() ); //Image Length
    m_ifd.m_fields[2].setShortValue( bitsPixel() ); // Bits per sample.
    m_ifd.m_fields[3].setShortValue( NOCOMPRESS ); //Compression   
    m_ifd.m_fields[4].setShortValue( (bitsPixel() == 1)? 1 : 2 ); //Photometric Interpretation   

    m_ifd.m_fields[6].setShortValue( (bitsPixel() == 1)? 1 : 3 ); // Samples per pixel.
    m_ifd.m_fields[7].setLongValue( yRes() );  //Rows per strip
    m_ifd.m_fields[8].setLongValue( bytesImage() ); //Strip byte count -- # bytes per strip   
}