#ifndef TiffOutSurf_h_included
#define TiffOutSurf_h_included

#include <fstream>
#include <string>

#include "tiff.h"
#include "Buffer.h"
#include "debug.h"

/*

 A TiffOutSurf is an encapsulation of a tiff file as a printer
 spool.  Basically, you name the file.  Upon initialization, the
 file is deleted if it already exists.  The standard 8-byte tiff
 header is placed at the beginning of the file.

 Then, you just send your image data to it using writeSpoolBuf().
 When you have finished writing an image, you call
 sendLastPageIfd() to write the IFD for that page and the offset
 to the next IFD.  (This is why you need to pass arguments to it.)
 Note that while you must send byte-aligned image rows to
 TiffOutSurf, the send*PageIfd() methods take care of word aligning
 the image as a whole.

 After the last page, you call sendVeryLastPageIfd().  (This
 has no arguments because the offset to the next IFD is null.)

 The rest of the methods are for internal use and/or convenience.

 Because these objects are based on my current tiff.h file, only
 uncompressed monochrome and 24-bit color will work.
*/


class TiffOutSurf
{
public:

    TiffOutSurf( const std::string & fileName )
    : m_fileName( fileName ),
      m_bitsPixel( 0xafbeadde )
    {}

    // Copy semantics.
    //
    // Although there are clear uses for copying TiffOutSurf objects,
    // make sure that you understand what you are doing.  The two
    // objects will share a file, etc.  The only reason that we
    // have copy semantics on TiffOutSurf is to simplify
    // IOemBase::initFrom() implementations.
    //
    // We can use default implementations for both.
    //
    // TiffOutSurf( const TiffOutSurf & );
    // TiffOutSurf & operator=( const TiffOutSurf & );

    virtual ~TiffOutSurf( void );

    HRESULT writeSpoolBuf( unsigned char *, size_t );

    void init( size_t xResolution, size_t yResolution, size_t colorDepth );
    
    void sendLastPageIfd( size_t newXRes, size_t newYRes, size_t newColorDepth );
    void sendVeryLastPageIfd( void );

    ULONG bitsPixel ( void ) const { return m_bitsPixel; } 
    ULONG bytesRow  ( void ) const;
    ULONG bytesImage( void ) const;

    ULONG xRes( void ) const { return m_xRes[0]; }
    ULONG yRes( void ) const { return m_yRes[0]; }

protected:
private:

    TiffHeader m_th;
    Ifd m_ifd;

    std::string m_fileName;
    
    size_t m_bitsPixel;

    // X and Y resolution expressed as rationals (in pixels).
    DWORD m_xRes[2];
    DWORD m_yRes[2];
};

/*
 
 TiffOutBuf is an adapter for a TiffOutSurf that allows it to be used
 in a Buffer chain.  (See Buffer.h and Buffer.cpp.)  It's sole purpose
 is to wrap writeSpoolBuf() with write.  Note that you still have to
 call Send*PageIfd() yourself.

*/

class TiffOutBuf : public Buffer
{
public:

    TiffOutBuf( void ) : m_pTiff( NULL ) {}

    TiffOutBuf( const TiffOutBuf & ); // Disabled!
    TiffOutBuf & operator=( const TiffOutBuf & ); // Disabled!

    virtual ~TiffOutBuf( void ) {} // Implements Buffer.

    // As TiffOutBuf is just an adapter with no state, we can
    // reset our adapter to wrap anything that we want at any
    // time.
    void setOutSurf( TiffOutSurf * tiff ) { m_pTiff = tiff; }

    // Pass Buffer::write() calls to TiffOutSurf::writeSpoolBuf() calls.
    virtual void write( unsigned char * buf, size_t count )
    {
        DBG_TRACE_OBJ( "TiffOutBuf::write" );

        DBG_WATCH( "count", count );
        DBG_WATCH( "m_pTiff", (void *)m_pTiff );

        if( m_pTiff )
        {
            m_pTiff->writeSpoolBuf( buf, count );
        }
    }

    // Since we do not buffer anything, we never have anything to flush, so 
    // flush() is a noop.
    virtual void flush( void ) {}


private:

    // We do not own the object pointed to by m_pTiff and never delete it!
    TiffOutSurf * m_pTiff;
};

#endif