#ifndef Buffer_h_Included
#define Buffer_h_Included

#include "debug.h"

/*

  A buffer is something with a definite size that you can write to.
  When the size is exceeded, the buffer gets flushed.  (You can
  also force a flush.)

  Because the buffers we are interested in store image data, the base
  class keeps around an array of unsigned chars.

*/

class Buffer
{
public:

    Buffer( void ) : m_buf(NULL), m_idx(0), m_size(0) {}

    Buffer( const Buffer & ); // Disabled!    
    Buffer & operator=( const Buffer & ); //Disabled.
    
    virtual ~Buffer( void );

    virtual void write( unsigned char *, size_t );

    void flush( void ) { flushData( m_buf ); }

    void setSize( size_t );

    size_t size( void ) const { return m_size; }

protected:

    // The buffer.
    unsigned char * m_buf;

    // The current position within the buffer.
    size_t m_idx;

    // The size (in bytes) of the buffer.
    size_t m_size;

    // As an optimization, we allow write() to flush data directly from
    // the data passed to it when possible.  write() normally copies data
    // into its m_buf storage area.  However, when the index is 0 and
    // m_size or more bytes are sent in a single write, each full chunk
    // of size m_size in that write can be flushed directly without being
    // copied.
    virtual void flushData( unsigned char * data ) = 0;

private:

};

/*

  A DownBuf downsamples the data sent into it.  You initialize it
  with a resolution and a factor.  When it gets factor full rows
  of data, it creates a single output row that has xRes/factor
  pixels.  Each output pixel represents a factor x factor square
  in the original data.  The average care-value of the input
  square is used as the output square.  If all of the pixels are
  don't care pixels, then a don't care pixel is used.  (It must
  be bled over later.

  Warnings:

  --DownBuf has only ever been tested with factors or 1, 2, and 4.
  --A factor of 1 actually traces a different code path than
    a factor of anything other than 1 because we skip all of the
    downsampling code in this case.
  --DownBuf assumes 24-bit color.
  --DownBuf assumes the don't care pixels are r=g=0x00, b=0xff.
  --You can flush a DownBuf safely only if it has an integer amount
    of row data in it.

  That last one is a big one.  Note than an assert() will fail if
  you violate this rule.  The reason for this behaviour is that for
  most of our purposes (in a GDI printer driver) we do not need to
  flush non-integer numbers of rows, and it would probably be a bug
  if we did.  Also, I like the idea that flush forces a buffer to
  empty.  What would happen to the left over data if there were a
  non-integer number of rows were flushed?

  Note, however, that you can flush an integer number of rows that
  is smaller than factor.  This is required so that the downsampled
  image has a representation of all of the data somewhere.  Basically,
  just remember that using DownBuf causes rounding up, not truncation.

*/


class DownBuf : public Buffer
{
public:

    // DownBuf does not take ownership of output.
    DownBuf( size_t fact, size_t xRes, Buffer * output );

    DownBuf( const DownBuf & ); // Disabled!    
    DownBuf & operator=( const DownBuf & ); //Disabled.

    virtual ~DownBuf( void );

    virtual void write( unsigned char * b, size_t s )
    {
        DBG_TRACE_OBJ( "DownBuf::write" );

        // If m_factor is 1, then we have a trivial case and there
        // is no need to buffer anything.
        if( m_factor == 1 )
        {
            m_output->write( b, s );
            m_idx = 0;
        }
        else
        {
            Buffer::write( b, s );
        }
    }

    virtual void flushData( unsigned char * );

    // setOutput() allows DownBuf, BleedBuf, and TiffOutBuf (or another
    // as of yet unwritten class) to be used in a buffer chain.  This
    // simplifies the rest of the printer driver by allowing it to be
    // written in terms of one buffer as a write() target.  The flushed
    // output is sent to the write() method of the next buffer in the
    // chain.
    //
    // Note that you can set output to NULL (although that seems sort of
    // dumb.)  write() will still queue up data, but flush() will just
    // swallow anything it gets.
    //
    // setOutput() does not take ownership of output.  You are still
    // responsible for deleting it when you are done.
    void setOutput( Buffer * output ) { m_output = output; }

private:

    // This is the 1-d downsampling factor.  It gets applied both
    // horizontally and vertically. Thus, if m_factor is 2, we
    // end up with 1/4th the data.
    const size_t m_factor;

    // The x resolution or number of pixels in a row.
    const size_t m_xRes;

    // When we have flushed, we send the downsampled data to m_output.
    // We do not own the object refered to by m_output and should never
    // delete it!
    Buffer * m_output;
};

class StorageBuf : public Buffer
{
public:

    explicit StorageBuf( size_t size ) : m_overflow(FALSE) { setSize (size); }

    virtual ~StorageBuf( void ) {}

    virtual void write( unsigned char * b, size_t s );

    virtual void flushData( unsigned char * ) { m_idx = 0; }

	BOOL overflow ( void ) { return m_overflow;}

	unsigned char* getDataPointer ( void ) { return m_buf; }

    size_t index ( void ) const { return m_idx; }

	void fillWhite ();

protected:

	BOOL m_overflow;
};

class ColorStorageBuf : public StorageBuf
{
public:

    explicit ColorStorageBuf( size_t size ) : StorageBuf(size) {}

    virtual ~ColorStorageBuf( void );

    // Looks for a buffer filled on ly with the given value.  This
    // is basically just factoring so we don't have to implement
    // the two methods below twice.
    bool isSolid( const unsigned char ) const;

    // These check for solid white and solid black buffers.
    bool isSolidWhite( void ) const { return isSolid(0xff); }
    bool isSolidBlack( void ) const { return isSolid(0x00); }

private:

};

#endif