
#include "stdafx.h"
#include "feeder.h"
#include "Debug.h"

LONG FileBlock::read (LONG index, signed char *buffer, LONG byteCount)
{
	DBG_TRACE_OBJ( "FileBlock::read" );

	DBG_WATCH( "index ", index );
	DBG_WATCH( "byteCount ", byteCount );
	DBG_WATCH( "m_fp ", m_fp );

    LONG end = m_startingAddress + BLOCKSIZE, numBytes, total_read = 0;

	if (index >= m_startingAddress && index < end)
	{
		numBytes = end - index;
		if (numBytes > byteCount) numBytes = byteCount;
		LONG bytes_read;

		while (numBytes > 0 && (bytes_read = fread((void*) buffer, 1, numBytes, m_fp)) > 0) 
		{
			numBytes -= bytes_read;
			buffer += bytes_read;
			total_read += bytes_read;
			DBG_WATCH( "read bytes ", bytes_read );
		}
	DBG_WATCH( "read bytes ", bytes_read );
    }
	
	return total_read;
}

LONG FileBlock::write(LONG index, const signed char *buffer, LONG byteCount)
{
	DBG_TRACE_OBJ( "FileBlock::write" );

	LONG end = m_startingAddress + BLOCKSIZE, numBytes, total_write = 0;

	if (index >= m_startingAddress && index < end)
	{
		numBytes = end - index;
		if (numBytes > byteCount) numBytes = byteCount;
		LONG bytes_write;

		while (numBytes > 0 &&
				(bytes_write = fwrite((const void*) buffer, 1, numBytes, m_fp)) > 0)
		{
			numBytes -= bytes_write;
			buffer += bytes_write;
			total_write += bytes_write;
			DBG_WATCH( "wrote bytes ", bytes_write );
		}
	}
	else
	{
		DBG_WATCH( "index ", index );
		DBG_WATCH( "starting address ", m_startingAddress );
	}
	
	return total_write;
}

LONG MemoryBlock::read (LONG index, signed char *buffer, LONG byteCount)
{
	DBG_TRACE_OBJ( "MemoryBlock::read" );

	LONG end = m_startingAddress + BLOCKSIZE, numBytes = 0;
	LONG offset = index - m_startingAddress;
	DBG_WATCH( "offset ", offset );

	if (index >= m_startingAddress && index < end)
	{
		signed char* start_pointer = m_buffer + offset;

		numBytes = end - index;
		if (numBytes > byteCount) numBytes = byteCount;
		DBG_WATCH( "numBytes ", numBytes );

		memcpy ((void*) buffer, (const void*) start_pointer, numBytes);
	}
	
	return numBytes;
}

LONG MemoryBlock::write(LONG index, const signed char *buffer, LONG byteCount)
{
	DBG_TRACE_OBJ( "MemoryBlock::write" );

	LONG end = m_startingAddress + BLOCKSIZE, numBytes = 0;
	LONG offset = index - m_startingAddress;
	DBG_WATCH( "offset ", offset );

	if (index >= m_startingAddress && index < end)
	{
		signed char* start_pointer = m_buffer + offset;

		numBytes = end - index;
		if (numBytes > byteCount) numBytes = byteCount;
		DBG_WATCH( "numBytes ", numBytes );

		if (numBytes > 0)
		{
			memcpy ((void*) start_pointer, (const void*) buffer, numBytes);
		}
	}
	else
	{
		DBG_WATCH( "index ", index );
		DBG_WATCH( "starting address ", m_startingAddress );
	}
	
	return numBytes;
}

TiffSpoolBufferFeeder::TiffSpoolBufferFeeder () :
m_dataArray (10),
m_currentBlock (0),
m_currentIndex (0),
m_highestBlock (0),
m_highestIndex (0),
m_markedBlock (0),
m_markedIndex (0),
m_fp (NULL)
{
	DBG_TRACE_OBJ( "TiffSpoolBufferFeeder::TiffSpoolBufferFeeder" );

	// allocate the first block
	Block *pB = new MemoryBlock ();
	//FileBlock *fb = new FileBlock (m_fp);
	m_dataArray[0] = pB;

	// set the other pointers to NULL
	for (int i = 1; i < m_dataArray.size (); i++) m_dataArray[i] = (Block*) NULL;
}

TiffSpoolBufferFeeder::~TiffSpoolBufferFeeder ()
{
	DBG_TRACE_OBJ( "TiffSpoolBufferFeeder::~TiffSpoolBufferFeeder" );

	// close the file
	if (m_fp) fclose (m_fp);

	// delete all blocks
	for (int i = 0; i < m_dataArray.size (); i++) 
	{
		delete m_dataArray[i];
	}

}

BOOL TiffSpoolBufferFeeder::addBlock ()
{
	DBG_TRACE_OBJ( "TiffSpoolBufferFeeder::addBlock" );

	// check if there is room for a new block
	int last = m_dataArray.size() - 1;
	Block *pB = m_dataArray[last];
	while (!pB && last > 0)
	{
		last--;
		pB = m_dataArray[last];
	}
	DBG_WATCH( "last Block ", last );

	// if necessary, resize the vector
	if (m_dataArray.size() - 2 <= last)
	{
		m_dataArray.resize (m_dataArray.size() * 2);
		if (m_dataArray.size () - 2 <= last) return FALSE;
		for (int i = last + 1; i < m_dataArray.size (); i++) 
			m_dataArray[i] = (FileBlock*) NULL;
		DBG_WATCH( "resized to ", m_dataArray.size () );
	}

	// allocate new block
	if (m_highestBlock < MAXMEMORYBLOCKS)
	    {
		pB = new MemoryBlock ();
		DBG_WATCH( "created Memory Block ", last + 1 );
	    }
	else
	    {
		if (!m_fp) m_fp = tmpfile ();
		if (!m_fp) m_fp = fopen (TEMPFILENAME, "w+b");
		if (!m_fp) return FALSE;
		pB = new FileBlock (m_fp);
		DBG_WATCH( "m_fp ", m_fp );
		DBG_WATCH( "created File Block ", last + 1 );
		
		// copy the first memory block to the temp file
		LONG first = m_highestBlock - MAXMEMORYBLOCKS;
		DBG_WATCH( "working on Memory Block ", first );
		MemoryBlock* pMB = dynamic_cast<MemoryBlock*> (m_dataArray[first]);
		if (!pMB) 
		{
			DBG_WATCH( "no Memory Block found at ", first );
			return FALSE;
		}
		// copy the starting address
		LONG address = pMB->get_startingAddress ();
		pB->set_startingAddress (address);
		// seek to the end of the file
		fseek (m_fp, 0, SEEK_END);
		pB->write (address, pMB->get_bufferPointer(), BLOCKSIZE);

		// now that we have saved the data to the temp file,
		// we can recycle the Memory Buffer
		m_dataArray[first] = pB;
		pB = pMB;
	    }

	// set the starting address by offsetting from the previous one
	pB->set_startingAddress ( m_dataArray[last]->get_startingAddress () +
				   BLOCKSIZE );
	DBG_WATCH ("starting address of new block ", pB->get_startingAddress ());

	// add the block to the data array
	m_dataArray.at(last + 1) = pB;

	return TRUE;
}

void TiffSpoolBufferFeeder::increasePositionBy (LONG offset)
{
	DBG_TRACE_OBJ( "TiffSpoolBufferFeeder::increasePositionBy" );

	LONG n_block = offset / BLOCKSIZE;
	LONG n_index = offset - n_block * BLOCKSIZE;

	// take care of all full blocks
	m_currentBlock += n_block;

	// advance the index
	if (m_currentIndex + n_index >= BLOCKSIZE)
	{
		m_currentBlock++;
		m_currentIndex = m_currentIndex + n_index - BLOCKSIZE;
	}
	else
	{
		m_currentIndex += n_index;
	}

	DBG_WATCH( "current Block ", m_currentBlock );
	DBG_WATCH( "current Index ", m_currentIndex );

}

BOOL TiffSpoolBufferFeeder::sendToSpoolBuf (PDEVOBJ pdevobj, IPrintOemDriverUni* pOEMHelp)
{
	DBG_TRACE_OBJ( "TiffSpoolBufferFeeder::sendToSpoolBuf" );

	BOOL status = writeSpoolBuf (pdevobj, pOEMHelp, m_highestBlock, m_highestIndex);

	// close file
	if (m_fp) fclose (m_fp);
	m_fp = NULL;

	return status;

}

BOOL TiffSpoolBufferFeeder::sendMarkedToSpoolBuf (PDEVOBJ pdevobj, IPrintOemDriverUni* pOEMHelp)
{
	DBG_TRACE_OBJ( "TiffSpoolBufferFeeder::sendMarkedToSpoolBuf" );

	// don't write if there is less than 1 kb
	// this prevents writing of the data before the first page
	if (m_markedBlock == 0 && m_markedIndex < 1024) return TRUE;

	LONG markedSize = m_markedBlock * BLOCKSIZE + m_markedIndex;
	BOOL status = writeSpoolBuf (pdevobj, pOEMHelp, m_markedBlock, m_markedIndex);
	if (!status) return FALSE;

	// now clean up the fileStructure
	LONG spoolSize = writtenToSpool();

	// delete all complete blocks that were written
    for (int count = 0; count < m_markedBlock; count++)
    {
		delete m_dataArray[count];
		m_dataArray[count] = (Block*) NULL;
		DBG_WATCH( "Deleted block ", count );
	}

	// clean up the marked block
	Block* pB = m_dataArray[m_markedBlock];
	LONG data_read, data_write, data_size = BLOCKSIZE - m_markedIndex;
	if (m_markedBlock == m_highestBlock)
		data_size -= BLOCKSIZE - m_highestIndex;
	DBG_WATCH( "Data to copy in marked block ", data_size );
	signed char *data_buf = new signed char [data_size];

	// get the bytes left in the marked block
	LONG position = pB->get_startingAddress () + m_markedIndex;
	data_read = pB->read (position, data_buf, data_size);
	DBG_WATCH( "Data read in marked block ", data_read );

	// and write them to the beginning of this block
	position = pB->get_startingAddress ();
	data_write = pB->write (position, data_buf, data_read);
	DBG_WATCH( "Data written in marked block ", data_write );
	delete [] data_buf;

	// move the block to beginning of the dataArray
	m_dataArray[m_markedBlock] = NULL;
	m_dataArray[0] = pB;

	// change the starting address of the first block
	pB->set_startingAddress (spoolSize + markedSize);
	DBG_WATCH( "Starting address of block 0 ", writtenToSpool());

	// clean up the current markers
	m_currentBlock = 0;
	m_currentIndex = data_write;

	// check for more blocks
	if (m_markedBlock < m_highestBlock)
	{

		// need a temporary array to hold the additional blocks,
		// because I need to free the entries in the dataArray
		LONG n_block = m_highestBlock - m_markedBlock;
		Block** temp_array = new Block* [n_block];
		DBG_WATCH( "have additional blocks ", n_block);
		for (LONG i = m_markedBlock + 1; i <= m_highestBlock; i++)
		{
			temp_array[i - m_markedBlock - 1] = m_dataArray[i];
			m_dataArray[i] = (Block*) NULL;
		}

		// need to save the highest block and index, because I am going to 
		// write to the dataArray now
		LONG highestB = m_highestBlock, highestI = m_highestIndex;

		m_highestBlock = 0;
		m_highestIndex = data_write;
		data_buf = new signed char [BLOCKSIZE];

		for (i = 0; i < n_block - 1; i++)
		    {
			pB = temp_array[i];
			pB->read (pB->get_startingAddress(), data_buf, BLOCKSIZE);
			delete pB;
			data_write = mrcioWrite (reinterpret_cast<LONG> (this), data_buf, BLOCKSIZE);
			DBG_WATCH( "wrote extra block ", data_write);
		    }

		pB = temp_array[n_block-1];
		pB->read (pB->get_startingAddress(), data_buf, highestI);
		delete pB;
		data_write = mrcioWrite (reinterpret_cast<LONG> (this), data_buf, highestI);
		DBG_WATCH( "wrote extra block ", data_write);

		delete [] temp_array;
		delete [] data_buf;
	}
	else
	{
		// clean up the high markers
		m_highestBlock = 0;
		m_highestIndex = data_write;
	}
		
	// just for tidyness
	m_markedBlock = m_markedIndex = 0;

	// close file
	if (m_fp) fclose (m_fp);
	m_fp = NULL;

	return TRUE;
}


BOOL TiffSpoolBufferFeeder::writeSpoolBuf (PDEVOBJ pdevobj, IPrintOemDriverUni* pOEMHelp,
										  LONG lastBlock, LONG lastIndex)
{
	DBG_TRACE_OBJ( "TiffSpoolBufferFeeder::writeSpoolBuf" );

    signed char *buf = new signed char [BLOCKSIZE];
    DWORD dwResult;
    size_t numRead   = 0;
	LONG position = -1;
 
	// reset file and position
	if (m_fp) position = fseek (m_fp, 0, SEEK_SET);
    DBG_WATCH ("seek returned ", position);
	m_currentBlock = 0;
	m_currentIndex = 0;

    DBG_MSG( "Copying..." );

    for (int count = 0; count <= lastBlock; count++)
    {
		if (!m_dataArray[count]) continue;

		position = writtenToSpool() + count * BLOCKSIZE;
		if (count == lastBlock)
		{
	        numRead = m_dataArray[count]->read( position, buf, lastIndex );
		}
		else
		{
	        numRead = m_dataArray[count]->read( position, buf, BLOCKSIZE );
		}

        HRESULT ok = pOEMHelp->DrvWriteSpoolBuf( pdevobj, buf, numRead, &dwResult );

        if( ok != S_OK )
        {
            DBG_MSG( "DrvWriteSpoolbuf() != S_OK" );
			return FALSE;
        }

        if( dwResult != numRead )
        {
            DBG_MSG( "DrvWriteSpoolbuf() ... dwResult != numRead" );        
			return FALSE;
        }

    }
    
    delete [] buf;
	return TRUE;
}

LONG mrcioSeek(ULONG mrcioID, LONG offset)
{
	DBG_WATCH("mrcioSeek", offset);
	
	TiffSpoolBufferFeeder *p_sbh = reinterpret_cast<TiffSpoolBufferFeeder*> (mrcioID);
	LONG block, index, retval;

	offset -= p_sbh->writtenToSpool ();
	block = offset / BLOCKSIZE;
	index = offset - block * BLOCKSIZE;
	DBG_WATCH( "Block ", block );
	DBG_WATCH( "Index ", index );

	if (block <= p_sbh->m_highestBlock ||
		(block == p_sbh->m_highestBlock && index <= p_sbh->m_highestIndex))
	{
		p_sbh->m_currentBlock = block;
		p_sbh->m_currentIndex = index;
		DBG_WATCH( "current Block ", p_sbh->m_currentBlock );
		DBG_WATCH( "current Index ", p_sbh->m_currentIndex );

		// check if we need to take care of the temp file
		if (p_sbh->m_highestBlock >= MAXMEMORYBLOCKS)
		    {
			LONG n_fileBlocks = p_sbh->m_highestBlock - MAXMEMORYBLOCKS;
			if (block <= n_fileBlocks)
			    {
				retval = fseek(p_sbh->m_fp, offset, SEEK_SET);
				if (retval != 0)
				    {
					return (LONG) 0;
				    }
			    }
		    }


		return offset + p_sbh->writtenToSpool ();
	}

	DBG_WATCH( "highest Block ", p_sbh->m_highestBlock );
	DBG_WATCH( "highest Index ", p_sbh->m_highestIndex );
	return (LONG) 0;
}
	
LONG mrcioFileSize(ULONG mrcioID)
{
	DBG_WATCH("mrcioFileSize", 0);

	TiffSpoolBufferFeeder *p_sbh = reinterpret_cast<TiffSpoolBufferFeeder*> (mrcioID);
	
	p_sbh->m_currentBlock = p_sbh->m_highestBlock;
	p_sbh->m_currentIndex = p_sbh->m_highestIndex;
	DBG_WATCH( "current Block ", p_sbh->m_currentBlock );
	DBG_WATCH( "current Index ", p_sbh->m_currentIndex );

	return p_sbh->getLinearPosition () + p_sbh->writtenToSpool ();
}

LONG mrcioTell(ULONG mrcioID)
{
	DBG_WATCH("mrcioTell", 0);

	TiffSpoolBufferFeeder *p_sbh = reinterpret_cast<TiffSpoolBufferFeeder*> (mrcioID);

	return p_sbh->getLinearPosition () + p_sbh->writtenToSpool ();
}

LONG mrcioRead(ULONG mrcioID, signed char *buffer, LONG byteCount)
{

	DBG_WATCH("mrcioRead", byteCount);

	TiffSpoolBufferFeeder *p_sbh = reinterpret_cast<TiffSpoolBufferFeeder*> (mrcioID);
	LONG bytes, bytes_total = byteCount, position;
	LONG index = p_sbh->m_currentBlock;

	while (p_sbh->m_dataArray.at(index) && bytes_total > 0) 
	{
		position = p_sbh->getLinearPosition() + p_sbh->writtenToSpool ();
		bytes = p_sbh->m_dataArray[index]->read (position, buffer, bytes_total);
		DBG_WATCH( "bytes read ", bytes );
		buffer += bytes;
		bytes_total -= bytes;
		p_sbh->increasePositionBy(bytes);
		index++;
		DBG_WATCH( "current Block ", p_sbh->m_currentBlock );
		DBG_WATCH( "current Index ", p_sbh->m_currentIndex );
	}

	return (byteCount - bytes_total);
}

LONG mrcioWrite(ULONG mrcioID, const signed char *buffer, LONG byteCount)
{
	DBG_WATCH("mrcioWrite", byteCount);

	TiffSpoolBufferFeeder *p_sbh = reinterpret_cast<TiffSpoolBufferFeeder*> (mrcioID);
	LONG bytes = 1, bytes_total = byteCount, position;
	LONG index = p_sbh->m_currentBlock;

	while (bytes > 0 && bytes_total > 0) 
	{
		position = p_sbh->getLinearPosition() + p_sbh->writtenToSpool ();
		bytes = p_sbh->m_dataArray[index]->write (position, buffer, bytes_total);
		DBG_WATCH( "bytes write ", bytes );
		buffer += bytes;
		bytes_total -= bytes;
		p_sbh->increasePositionBy(bytes);
		index++;
		DBG_WATCH( "current Block ", p_sbh->m_currentBlock );
		DBG_WATCH( "current Index ", p_sbh->m_currentIndex );
		if (p_sbh->m_currentBlock > p_sbh->m_highestBlock)
		{
			p_sbh->m_highestBlock = p_sbh->m_currentBlock;
			p_sbh->m_highestIndex = p_sbh->m_currentIndex;
			DBG_WATCH( "highest Block ", p_sbh->m_highestBlock );
			DBG_WATCH( "highest Index ", p_sbh->m_highestIndex );
		}
		else if (p_sbh->m_currentBlock == p_sbh->m_highestBlock &&
				 p_sbh->m_currentIndex > p_sbh->m_highestIndex)
		{
			p_sbh->m_highestIndex = p_sbh->m_currentIndex;
			DBG_WATCH( "highest Block ", p_sbh->m_highestBlock );
			DBG_WATCH( "highest Index ", p_sbh->m_highestIndex );
		}
		if (!p_sbh->m_dataArray[p_sbh->m_highestBlock]) 
		    p_sbh->addBlock ();
	}

	return (byteCount - bytes_total);

}
