#ifndef Feeder_h_Included
#define Feeder_h_Included

#include <vector>
#include <stdio.h>

#include <PRCOMOEM.H>

/*static const LONG BLOCKSIZE = 64;         // stress test
static const LONG MAXMEMORYBLOCKS = 1000; // parameters    */
static const LONG BLOCKSIZE = 524288; // 0.5 Mb
static const LONG MAXMEMORYBLOCKS = 4;  // total of 2 Mb in memory
static const char* TEMPFILENAME = "c:\\printemp.tfx"; // if tempfile() does not open a tempfile
							                          // we will try to open a file with this name

class Block
{
public:

	// initialize with BlockSize
    Block( void ) : m_startingAddress (0) {}
    
    virtual ~Block( void ) {}

	virtual LONG read(LONG index, signed char *buffer, LONG byteCount) = 0;
	virtual LONG write(LONG index, const signed char *buffer, LONG byteCount) = 0;

	// access to startingAddress
	LONG get_startingAddress ( void ) const { return m_startingAddress; }
	void set_startingAddress ( LONG address ) { m_startingAddress = address; }

protected:

	LONG m_startingAddress;
};


class FileBlock : public Block
{
public:

	FileBlock (FILE *fp) : m_fp(fp) {}
	virtual ~FileBlock ( void ) {};

	virtual LONG read(LONG index, signed char *buffer, LONG byteCount);
	virtual LONG write(LONG index, const signed char *buffer, LONG byteCount);

protected:

	FILE *m_fp;

};

class MemoryBlock : public Block
{
public:

	MemoryBlock () { m_buffer = new signed char [BLOCKSIZE]; }
	virtual ~MemoryBlock ( void ) { delete [] m_buffer; };

	virtual LONG read(LONG index, signed char *buffer, LONG byteCount);
	virtual LONG write(LONG index, const signed char *buffer, LONG byteCount);
	signed char* get_bufferPointer ( void ) { return m_buffer; }

protected:

	signed char *m_buffer;

};

class TiffSpoolBufferFeeder
{

public:

	TiffSpoolBufferFeeder ();
	~TiffSpoolBufferFeeder ();

	// send what we have to the spool buffer
	BOOL sendToSpoolBuf (PDEVOBJ pdevobj, IPrintOemDriverUni* pOEMHelp);

	// send everything before the mark to the spool buffer
	BOOL sendMarkedToSpoolBuf (PDEVOBJ pdevobj, IPrintOemDriverUni* pOEMHelp);

	// mark the current highest position
	void markCurrentPosition ( void )
	{
		m_markedBlock = m_highestBlock;
		m_markedIndex = m_highestIndex;
	}

	// set of primitive I/O functions that MRCIO needs
	// these need to be frients to have access to the private members
	friend LONG mrcioRead(ULONG mrcioID, signed char *buffer, LONG byteCount);
	friend LONG mrcioWrite(ULONG mrcioID, const signed char *buffer, LONG byteCount);
	friend LONG mrcioSeek(ULONG mrcioID, LONG offset);
	friend LONG mrcioFileSize(ULONG mrcioID);
	friend LONG mrcioTell(ULONG mrcioID);

private:

	// convert block and index to linear length
	LONG getLinearPosition ( void ) 
	    { 
		return m_currentBlock * BLOCKSIZE + m_currentIndex; 
	    }

	// increase block and index by some offset
	void increasePositionBy (LONG offset);

	// get the number of bytes which we already sent to the spool buffer
	LONG writtenToSpool ( void )
	    {
		if (m_dataArray[0]) return m_dataArray[0]->get_startingAddress ();
		return 0;
	    }

	// write to spool Buffer
	BOOL writeSpoolBuf (PDEVOBJ pdevobj, IPrintOemDriverUni* pOEMHelp,
						LONG lastBlock, LONG lastIndex);

	// add another block to the fileStructure
	BOOL addBlock ( void );

	// the temp file in case we run out of memory
	FILE *m_fp;

	// the array that hold all blocks
	std::vector<Block*> m_dataArray;

	// variables to keep track of positions in the file
    LONG  m_currentBlock;
    LONG  m_currentIndex;
    LONG  m_highestBlock;
    LONG  m_highestIndex;
    LONG  m_markedBlock;
    LONG  m_markedIndex;

};

#endif



