/*
 * -- CCITTFaxCompressImpl.mesa
 *   -- Copyright (c) 1998 by Xerox Corporation.  All rights reserved.
 *   -- Russ Atkinson (RRA) October 17, 1998 10:19 pm PDT
 *
 * Translated to C by James Ahlborn in January 1999.
 */

#include "StdAfx.h"
#include "g4.h"
#include "debug.h"
#include <stdio.h>

IG4::IG4 (void)
{
  G4Create(&m_cookie, stderr, stdout);
}

IG4::~IG4 (void)
{
  G4CleanUp(&m_cookie);
}

BOOL
IG4::Compress (unsigned char* imageData, ULONG dataSize,
			 unsigned char** compressedData, 
			 ULONG* compressedDataSize)
{
  LONG compressionFactor = 10;
  ULONG size;
  BOOL status;

  do {
	  AllocateOutputBuffer (compressionFactor);

	  size = m_compressedDataSize;
	  m_memErr = FALSE;

	  DBG_WATCH ("G4 compression with buffer at ", compressionFactor);
	  DBG_WATCH ("G4 compression buffer size ", size);

	  status = G4EncodeC(imageData, &size, TRUE);

	  if (compressionFactor > 1) compressionFactor /= 2;
	  else if (compressionFactor == 1) compressionFactor = -2;
	  else compressionFactor *= 2;

  } while (m_memErr);

  if (!status) return status;

	*compressedData = m_pCompressedData;
	*compressedDataSize = size;

	return TRUE;
}


/* c implementation of compression standard */

#define UCHAR  unsigned char
#define USHORT unsigned short
#define _LITTLE_ENDIAN
#define set_err_info(funct_name, pdgp_err_state)  \
  LONG *__PERR_STATE__ =  NULL
#define throw_err(string) do { \
    DBG_WATCH( string , 0 ); \
    goto bailout; } while (0)
#define rethrow_err(string) do { \
    DBG_WATCH( string , 0 ); \
    goto bailout; } while (0)
#define throw_err0() do { \
    goto bailout; } while (0)
#define rethrow_errf(string, var) do { \
    DBG_WATCH( string , var ); \
    goto bailout; } while (0)

/* constants */

const ULONG WORD_LAST = ((ULONG)0xFFFFFFFF);

const ULONG eolBits = 12;
const ULONG eolCode = 1;
  
const ULONG codeSplit     =   64;
const ULONG codeLongSplit = 2560;

const ULONG maxCodeBits =  13;
const ULONG runSplit    = 512;

const ULONG TAB_COL_WHITE = 0;
const ULONG TAB_COL_BLACK = 1;
const ULONG TAB_LEN_SHORT = 0;
const ULONG TAB_LEN_LONG  = 1;

const LONG G4_K_VAL    = -1;
const LONG G3_1D_K_VAL =  0;
const LONG G3_2D_K_VAL =  1;

const ULONG BITS_ULONG = 32;
const ULONG BITS_UCHAR = 8;

/* data structures */

typedef struct {
  USHORT val;
  USHORT len;
} TabEntry;

typedef struct {
  char *data;
  ULONG length;
  ULONG maxLength;
} RefText;

typedef struct {
  ULONG len;
  ULONG max;
  ULONG *data;
} Trans;

typedef struct {
  ULONG old_bpl;

  ULONG bpl;
  RefText outBuffer;
  UCHAR outMod;
  ULONG outWord;
  Trans *transOld;
  Trans *transNew;
  FILE *debug;
  FILE *err_out;
} CompressData;

/* compression functions */
static BOOL
EncodeInit(CompressData *cd, ULONG bpl);

static BOOL
LineToTrans(CompressData *cd, ULONG *ptr);

static BOOL
TransToG4Code(CompressData *cd, BOOL uncompressed);

static BOOL
Terminate(CompressData *cd, LONG k);

/* tables */

const TabEntry whiteShortCodeTab[codeSplit] = {
  {53, 8}, {7, 6}, {7, 4}, {8, 4}, {11, 4}, {12, 4}, {14, 4}, {15, 4}, 
  {19, 5}, {20, 5}, {7, 5}, {8, 5}, {8, 6}, {3, 6}, {52, 6}, {53, 6}, 
  {42, 6}, {43, 6}, {39, 7}, {12, 7}, {8, 7}, {23, 7}, {3, 7}, {4, 7}, 
  {40, 7}, {43, 7}, {19, 7}, {36, 7}, {24, 7}, {2, 8}, {3, 8}, {26, 8}, 
  {27, 8}, {18, 8}, {19, 8}, {20, 8}, {21, 8}, {22, 8}, {23, 8}, {40, 8}, 
  {41, 8}, {42, 8}, {43, 8}, {44, 8}, {45, 8}, {4, 8}, {5, 8}, {10, 8}, 
  {11, 8}, {82, 8}, {83, 8}, {84, 8}, {85, 8}, {36, 8}, {37, 8}, {88, 8}, 
  {89, 8}, {90, 8}, {91, 8}, {74, 8}, {75, 8}, {50, 8}, {51, 8}, {52, 8}};

const TabEntry blackShortCodeTab[codeSplit] = {
  {55, 10}, {2, 3}, {3, 2}, {2, 2}, {3, 3}, {3, 4}, {2, 4}, {3, 5}, 
  {5, 6}, {4, 6}, {4, 7}, {5, 7}, {7, 7}, {4, 8}, {7, 8}, {24, 9}, 
  {23, 10}, {24, 10}, {8, 10}, {103, 11}, {104, 11}, {108, 11}, {55, 11}, 
  {40, 11}, {23, 11}, {24, 11}, {202, 12}, {203, 12}, {204, 12}, {205, 12},
  {104, 12}, {105, 12}, {106, 12}, {107, 12}, {210, 12}, {211, 12}, {212, 12},
  {213, 12}, {214, 12}, {215, 12}, {108, 12}, {109, 12}, {218, 12}, {219, 12},
  {84, 12}, {85, 12}, {86, 12}, {87, 12}, {100, 12}, {101, 12}, {82, 12},
  {83, 12}, {36, 12}, {55, 12}, {56, 12}, {39, 12}, {40, 12}, {88, 12},
  {89, 12}, {43, 12}, {44, 12}, {90, 12}, {102, 12}, {103, 12}};

const TabEntry whiteLongCodeTab[codeSplit] = {
  {0, 0}, {27, 5}, {18, 5}, {23, 6}, {55, 7}, {54, 8}, {55, 8}, {100, 8}, 
  {101, 8}, {104, 8}, {103, 8}, {204, 9}, {205, 9}, {210, 9}, {211, 9},
  {212, 9}, {213, 9}, {214, 9}, {215, 9}, {216, 9}, {217, 9}, {218, 9},
  {219, 9}, {152, 9}, {153, 9}, {154, 9}, {24, 6}, {155, 9}, {8, 11},
  {12, 11}, {13, 11}, {18, 12}, {19, 12}, {20, 12}, {21, 12}, {22, 12},
  {23, 12}, {28, 12}, {29, 12}, {30, 12}, {31, 12}, {0, 0}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}};

const TabEntry blackLongCodeTab[codeSplit] = {
  {0, 0}, {15, 10}, {200, 12}, {201, 12}, {91, 12}, {51, 12}, {52, 12},
  {53, 12}, {108, 13}, {109, 13}, {74, 13}, {75, 13}, {76, 13}, {77, 13},
  {114, 13}, {115, 13}, {116, 13}, {117, 13}, {118, 13}, {119, 13}, {82, 13},
  {83, 13}, {84, 13}, {85, 13}, {90, 13}, {91, 13}, {100, 13}, {101, 13},
  {8, 11}, {12, 11}, {13, 11}, {18, 12}, {19, 12}, {20, 12}, {21, 12},
  {22, 12}, {23, 12}, {28, 12}, {29, 12}, {30, 12}, {31, 12}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}, {0, 0}};

const TabEntry *shortTabs[2] = {whiteShortCodeTab, blackShortCodeTab};
const TabEntry *longTabs[2] = {whiteLongCodeTab, blackLongCodeTab};


/* 
   Resets the cd object to the initial state.  This is useful when desiring
   to use the same compression handle over several pages, since the storage
   can be reused.
   -- Internal buffers will be allocated according to the bpl parameter for
      outBuffer, transOld, and transNew.
   */
BOOL
EncodeInit(CompressData *cd, ULONG bpl)
{
  set_err_info("EncodeInit", cd->err_out);

  ULONG lim = bpl + 4;
  Trans *tnew = cd->transNew, *told = cd->transOld;


  if(bpl > cd->old_bpl) {
    
    if(tnew->data) (free)(tnew->data);

    if(!(tnew->data = 
	 (ULONG *)(malloc)(lim * sizeof(*(tnew->data)))))
      throw_err("malloc failed");
    tnew->max = lim;

    if(told->data) (free)(told->data);

    if(!(told->data = 
	 (ULONG *)(malloc)(lim * sizeof(*(told->data)))))
      throw_err("malloc failed");
    told->max = lim;

    if(cd->outBuffer.data) (free)(cd->outBuffer.data);

    if(!(cd->outBuffer.data = 
	 (char *)(malloc)(sizeof(*(cd->outBuffer.data)) * (bpl+10))))
      throw_err("malloc failed");
    cd->outBuffer.maxLength = bpl+10;
    cd->old_bpl = bpl;
  }

  cd->bpl = bpl;
  tnew->len = 0;
  told->len = 0;
  tnew->data[0] = tnew->data[1] = bpl;
  told->data[0] = told->data[1] = bpl;

  cd->outBuffer.length = 0;

  cd->outMod = 0;
  cd->outWord = 0;

  return TRUE;

bailout:

  return FALSE;
}


/* cleanup memory stored in comress data struct */
void
IG4::G4CleanUp(void **_cd)
{
  CompressData *cd = static_cast<CompressData*> (*_cd);
  Trans *trans;

  *_cd = NULL;

  if(cd) {

    trans = cd->transNew;
    if(trans) {
      if(trans->data) (free)(trans->data);
      trans->data = NULL;
      (free)(trans);
    }
    cd->transNew = NULL;

    trans = cd->transOld;
    if(trans) {
      if(trans->data) (free)(trans->data);
      trans->data = NULL;
      (free)(trans);
    }
    cd->transOld = NULL;

    if(cd->outBuffer.data) (free)(cd->outBuffer.data);
    cd->outBuffer.data = NULL;

    (free)(cd);
  }
}


/*
  -- Creates a new handle for compression.
  */
BOOL
IG4::G4Create(void **_cd, FILE *err_out, FILE *db_out) 
{
  set_err_info("EncodeCreate", err_out);

  CompressData *cd = NULL;

  
  *_cd = NULL;

  if(!(cd = (CompressData *)(calloc)(1, sizeof(*cd))))
    throw_err("malloc failed");

  if(!(cd->transNew = (Trans *)(calloc)(1, sizeof(*(cd->transNew)))))
    throw_err("malloc failed");

  if(!(cd->transOld = (Trans *)(calloc)(1, sizeof(*(cd->transOld)))))
    throw_err("malloc failed");

  cd->err_out = err_out;
  cd->debug = db_out;

  *_cd = cd;

  return TRUE;

bailout:

  G4CleanUp((void **)&cd);

  return FALSE;
}


/*
  -- Converts a line of raw data to transitions.  All lines begin with the
  first (leftmost) bit of the word addressed by ptr.
  -- The transitions in transNew are moved to those in transOld before the
  conversion.
  -- The new transitions are then put into transNew.
  */
BOOL
LineToTrans(CompressData *cd, ULONG *ptr)
{
  set_err_info("LineToTrans", cd->err_out);
  ULONG sense = 0, pos = 0, w = 0, len = (ULONG)cd->bpl, rem, xor, tmp;
  Trans *trans = cd->transOld;
  ULONG *dst = trans->data;
  ULONG *dp = dst;
  UCHAR *bptr_in, *bptr_out;

#define AddRun(k)				\
  do {sense = ~sense;				\
  dp[0] = pos;					\
  dp++;						\
  pos += (k);} while(0)

#define AddRun2(k1, k2)			       		\
  do {dp[0] = pos;						\
  pos = pos + (k1);					\
  dp[1] = pos;						\
  dp += 2;				 		\
  pos += (k2);} while(0)

  cd->transOld = cd->transNew;
  cd->transNew = trans;

  /* Do whole words in this loop */
  while(len >= BITS_ULONG) {
    rem = BITS_ULONG;
#ifdef _LITTLE_ENDIAN
	bptr_in = (UCHAR*) ptr;
	bptr_out = (UCHAR*) &w;
	*bptr_out = *(bptr_in + 3);
	*(bptr_out + 1) = *(bptr_in + 2);
	*(bptr_out + 2) = *(bptr_in + 1);
	*(bptr_out + 3) = *bptr_in;
#else
    w = ptr[0];
#endif
    ptr++;
    len -= rem;
    
    /* First of all, take care of long runs */
    if(sense != w) {
      /* Quick test for the entire word being OK */
      while(rem >= 4) {
	xor = w ^ sense;
	tmp = xor >> (BITS_ULONG-4);
	if(tmp == 0) {
	  /* Quick test for the remainder being a single run */
	  if((xor >> (UCHAR)(BITS_ULONG-rem)) == 0) {
	    break;
	  }
	  pos += 4;
	} else if(tmp == 1) {
	  pos += 3;
	  AddRun(1);
	} else if(tmp == 2) {
	  pos += 2;
	  AddRun2(1, 1);
	} else if(tmp == 3) {
	  pos += 2;
	  AddRun(2);
	} else if(tmp == 4) {
	  pos += 1;
	  AddRun2(1, 2);
	} else if(tmp == 5) {
	  pos += 1;
	  AddRun2(1, 1);
	  AddRun(1);
	} else if(tmp == 6) {
	  pos += 1;
	  AddRun2(2, 1);
	} else if(tmp == 7) {
	  pos += 1;
	  AddRun(3);
	} else {
	  /* In wrong sense, so cut current run short and try again */
	  AddRun(0);
	  continue;
	}
	w *= 16;
	rem -= 4;
      }
    }
    pos += rem;
  }

  if(len != 0) {
    /* Finish up remainder of line, bit by bit */
#ifdef _LITTLE_ENDIAN
	bptr_in = (UCHAR*) ptr;
	bptr_out = (UCHAR*) &w;
	*bptr_out = *(bptr_in + 3);
	*(bptr_out + 1) = *(bptr_in + 2);
	*(bptr_out + 2) = *(bptr_in + 1);
	*(bptr_out + 3) = *bptr_in;
#else
    w = ptr[0];
#endif
    while(len > 0) {
      if(((LONG)(w ^ sense)) < 0) {
	AddRun(1);
      } else {
	pos++;
      }
      len -= 1;
      w += w;
    }
  }
  
  if(pos != cd->bpl)
    throw_err("invalid data");

  dp[0] = pos;
  dp++;

  /* Compute the length and terminate with 2 entries */
  trans->len = ((ULONG)(dp-dst)) /*/ sizeof(ULONG)*/;
  dp[0] = dp[1] = pos;

  return TRUE;

bailout:

  return FALSE;
}


/* Takes nb bits, right-justified in w, and puts them into the output bit
   stream. */
#define PBtmp(w, n) do {       						\
  ULONG _w = (w);							\
  UCHAR _n = (n);	       						\
  UCHAR _oc = outMod;							\
  outMod = (_n + outMod) % BITS_ULONG;	       					\
  if(outMod > _oc) {							\
    /* Just storing a partial field */					\
    _w = _w << ((UCHAR)(BITS_ULONG - outMod));					\
    outWord |=  _w;							\
  } else if(_n == 0) {							\
    /* No bits to write */						\
  } else {								\
    /* Crossing a word boundary */					\
    ULONG _newlen = rt->length + sizeof(ULONG);	       			\
    if(outMod == 0) {							\
      if(_newlen > rt->maxLength) throw_err("internal buffer error");	\
      /* Filling a word exactly */					\
      wp[0] = outWord | _w;						\
      outWord = 0;							\
    } else {								\
      if(_newlen >= rt->maxLength) throw_err("internal buffer error");	\
      /* There are bits left over for the next word */			\
      wp[0] = outWord | (_w >> outMod);					\
      outWord = _w << ((UCHAR)(BITS_ULONG - outMod));				\
    }									\
    wp++;								\
    rt->length = _newlen;				       		\
  }									\
} while(0)

#ifdef canDebug

#define PB(w, n) do {							\
  if(db) fprintf(db, "PB[%lx, %lu]\n", (ULONG)(w), (ULONG)(n));		\
  PBtmp((w), (n));     							\
} while(0)

#else

#define PB(w, n) PBtmp((w), (n))

#endif


#define PutEnt(ent) PB((ent)->val, (ent)->len)

#define PutHLenTmp(len, state) do {			       		    \
  ULONG _len = (len);							    \
  UCHAR _state = (state);						    \
  const TabEntry *_ent;							    \
  _ent = &(longTabs[(LONG)_state][(codeLongSplit / codeSplit) % codeSplit]); \
  while(_len >= codeLongSplit) {					    \
    PutEnt(_ent);							    \
    _len -= codeLongSplit;						    \
  }									    \
  if(_len >= codeSplit) {						    \
    _ent = &(longTabs[(LONG)_state][(_len / codeSplit) % codeSplit]);	    \
    PutEnt(_ent);							    \
  }									    \
  _ent = &(shortTabs[(LONG)_state][_len % codeSplit]);			    \
  PutEnt(_ent);								    \
} while(0)


#ifdef canDebug								    

#define PutHLen(len, state) do {       					     \
  if(db) fprintf(db, "PH [%lu, %d]\n", (ULONG)(len), (LONG)(state));          \
  PutHLenTmp((len), (state));						     \
} while(0)

#else

#define PutHLen(len, state) PutHLenTmp((len), (state))

#endif


/* 
   Takes n bits, right-justified in w, and puts them into the output bit
   stream. 
   */
static BOOL
PutBits(CompressData *cd, ULONG w, UCHAR n)
{
  set_err_info("PutBits", cd->err_out);

  UCHAR oc = cd->outMod;
  UCHAR outMod = (n+oc) % BITS_ULONG;
  ULONG outWord = cd->outWord;
  RefText *rt;
  LONG oldlen;
  ULONG *wp;
  ULONG newlen;

#if canDebug
  if(cd->debug) fprintf(cd->debug, "PB[%lx, %lu]\n", (ULONG)w, (ULONG)n);
#endif

  if(outMod > oc) {
    /* Just storing a partial field */
    w = w << ((UCHAR)(BITS_ULONG-outMod));
    outWord |= w;
  } else if(n == 0) {
    /* No bits to write */
  } else {
    /* Crossing a word boundary */
    rt = &(cd->outBuffer);
    oldlen = rt->length;
    wp = (ULONG *)(rt->data + oldlen);
    newlen = oldlen + sizeof(ULONG);
    
    if(outMod == 0) {
      if(newlen > rt->maxLength) throw_err("internal buffer error");
      /* Filling a word exactly */
      wp[0] = outWord | w;
      outWord = 0;
    } else {
      if(newlen >= rt->maxLength) throw_err("internal buffer error");
      /* There are bits left over for the next word */
      wp[0] = outWord | (w >> outMod);
      outWord = w << ((UCHAR)(BITS_ULONG - outMod));
    }
    rt->length = newlen;
  }
  
  cd->outMod = outMod;
  cd->outWord = outWord;

  return TRUE;

bailout:

  return FALSE;
}



/*
  Converts the transitions in transNew to G4 coding in the outBuffer.
  */
BOOL
TransToG4Code(CompressData *cd)
{
  set_err_info("TransToG4Code", cd->err_out);
  UCHAR outMod = cd->outMod;
  ULONG outWord = cd->outWord;
  RefText *rt = &(cd->outBuffer);
  UCHAR initMod = outMod;
  LONG initPos = rt->length;
  ULONG *wp = (ULONG *)(rt->data + initPos);
  ULONG bpl = cd->bpl;
  ULONG rx = 0;
  ULONG *runs = cd->transNew->data;
  ULONG *lag = cd->transOld->data;
  ULONG lagLen = cd->transOld->len;
  ULONG lagx = 0, a0 = 0, b1 = lag[lagx], b2 = 0, d = 0, a1, a2, w;
  UCHAR color,  nb;
#if canDebug
  FILE *db = (cd->debug);
#endif

  while(a0 < bpl) {
    a1 = runs[rx];
    b2 = lag[lagx+1];
    if(a1 > b2) {
      /* Pass mode, skip a transition on the lagging line
         The clause "AND (lagx+8 > lagLen OR a1 < lag[lagx+8])" can be
	 appended to the above test "a1 > b2" to improve the compression
	 ratio (and decompression speed), but some G4 decompressors can't
	 handle the generality of skipping over several lagging transitions
	 and demand the pass codes instead. Both DocuPrint and PostScript
	 can handle the missing pass codes.
	 */
#if canDebug
      if(db) fprintf(db, "P, a1: %lu, b2: %lu, ", (ULONG)a1, (ULONG)b2);
#endif

      /* pass code */
      PB(1, 4);

      a0 = b2;
      lagx += 2;
      if(lagx > lagLen) {
	lagx = lagLen;
      }
      /* We do NOT have to scan to find the right lagx, since the color is
	 already OK and we know that the next b1 > a0.  This can save a bit
	 of time. */
      b1 = lag[lagx];
      continue;
    }
    d = b1 - a1 + 3;

    if(d < 7) {
      /* Vertical mode */
      w = 1;
      nb = 1;
      
      switch(d) {
      case 0:
	/* -- a1 = b1 + 3 */
	w = 3;
	nb = 7;
	break;
      case 1:
	/* -- a1 = b1 + 2 */
	w = 3;
	nb = 6;
	break;
      case 2:
	/* -- a1 = b1 + 1 */
	w = 3;
	nb = 3;
	break;
      case 4:
	/* -- a1 = b1 + 2 */
	w = 2;
	nb = 3;
	break;
      case 5:
	/* -- a1 = b1 + 3 */
	w = 2;
	nb = 6;
	break;
      case 6:
	/* -- a1 = b1 + 3 */
	w = 2;
	nb = 7;
	break;
      }
#if canDebug
      if(db) fprintf(db, "V, a1: %lu, b1: %lu, ", (ULONG)a1, (ULONG)b1);
#endif
      PB(w, nb);
      rx++;
      a0 = a1;
    } else {
      /* Horizontal Mode */
      a2 = runs[rx + 1];
      color = rx % 2;
#if canDebug
      if(db) fprintf(db, "H, a1: %lu, a2: %lu, ", (ULONG)a1, (ULONG)a2);
#endif
      /* horizontal code */
      PB(1, 3);

      PutHLen((a1-a0), color);
      PutHLen((a2-a1), 1-color);
      rx = rx+2;
      a0 = a2;
    }

    /* At this point the value of b1 (based on lag[lagx]) is probably
       wrong, so we have to update it to correspond to the new a0. */
    if(lagx > 0) {
      lagx -= 1;
    }

    while(TRUE) {
      b1 = lag[lagx];
      if((b1 > a0) && ((rx % 2) == (lagx %2))) break;
      if(lagx >= lagLen) break;
      lagx++;
    }
  }

  cd->outMod = outMod;
  cd->outWord = outWord;

  /* Flush partial bits to the buffer, but don't update the state */
  if(outMod != 0) {
    wp[0] = outWord;
  }

  if(a0 != bpl)
    throw_err("a0 not equal to bpl");
  
  return TRUE;
  
bailout:

  return FALSE;
}


/*
  Puts out a G3/G4 terminating code in the outBuffer, determined by the K
  parameter (see DoPage description).
  */
BOOL
Terminate(CompressData *cd, LONG k)
{
  set_err_info("Terminate", cd->err_out);

#ifndef _LITTLE_ENDIAN

  LONG i;
  LONG times = ((k < 0) ? 2 : 6);
  LONG tCode = ((k > 0) ? ((eolCode * 2) + 1) : eolCode);
  LONG tLen = ((k > 0) ? (eolBits + 1) : eolBits);
  
  for(i = 0; i < times; i++) {
    if(!(PutBits(cd, tCode, tLen)))
      rethrow_err("failed writing bits to output buffer");      
  }
#else    

  ULONG temp = 0;
  UCHAR *ptemp = ((UCHAR*) &temp);

	*ptemp = 0x00;
DBG_WATCH( "G4 terminator" , temp );
    if(!(PutBits(cd, temp, 8)))
      rethrow_err("failed writing bits to output buffer");      

	*ptemp = 0x10;
DBG_WATCH( "G4 terminator" , temp );
    if(!(PutBits(cd, temp, 8)))
      rethrow_err("failed writing bits to output buffer");      

	*ptemp = 0x01;
DBG_WATCH( "G4 terminator" , temp );
    if(!(PutBits(cd, temp, 8)))
      rethrow_err("failed writing bits to output buffer");      

#endif

  return TRUE;

bailout:

  return FALSE;
}




/*
 * Take the input image, which is packed with MSB=left, width
 * by height bits, and each row having longsPerRow longs, and encode it
 * using G4. 0 bits are white, and 1 bits are black.
 *
 * output points to the location that the result must be put into.
 * Initially, *outlen gives the size of the output buffer. At return, *outlen
 * should be set to the number of bytes actually written into the buffer.
 * Return TRUE on success, FALSE on failure. Running out of buffer space
 * is a failure.
 *
 * No EOL codes are used, either between lines or at the end. Encoded rows
 * are not padded; the only padding done is on the final byte.
 *
 * If eofb is TRUE, an EOFB (two EOL codes) is written at the end.
 */


/*
 * This function is the same as the above function, except without the
 * creation and deletion of the CompressData struct each time.
 */

BOOL 
IG4::G4EncodeC(unsigned char *image, ULONG *outlen, BOOL eofb)
{
  CompressData *cd = (CompressData *) m_cookie;
  set_err_info("g4EncodeC", cd->err_out);

  ULONG *long_ptr;
  unsigned char* outbuf = m_pCompressedData;
  LONG i, j, rem;
  ULONG outBufMax = *outlen, outBufLen = 0, tmp_ulong, tmpBufLen;
#ifdef _LITTLE_ENDIAN
  ULONG *lbuf;
#endif

  /* initialize the compress data struct */
  if(!(EncodeInit(cd, m_width)))
    rethrow_err("failed initializing compress data struct");

  *outlen = 0;

  /* process each line of input image */
  for(i = 0; i < m_height; i++) {

	long_ptr = (unsigned long*) image;
    /* convert lines to transitions */
    if(!(LineToTrans(cd, long_ptr)))
	{
      DBG_WATCH("could not compress line ", i);
	  goto bailout;
	}

    /* G4 encoding */
    if(!(TransToG4Code(cd)))
	{
      DBG_WATCH("could not compress line ", i);
	  goto bailout;
	}

    /* output new data */
    tmpBufLen = cd->outBuffer.length;
    if((tmpBufLen + outBufLen) > outBufMax)
	{
	  m_memErr = TRUE;
      throw_err0();
	}

#ifdef _LITTLE_ENDIAN
    
    lbuf = (ULONG *)(cd->outBuffer.data);
    for(j = 0; j < tmpBufLen; j+=sizeof(ULONG)) {
      tmp_ulong = *(lbuf++);
      if(BITS_ULONG == 64) {
	*(outbuf++) = (UCHAR)((tmp_ulong >> 56) & 0xFF);
	*(outbuf++) = (UCHAR)((tmp_ulong >> 48) & 0xFF);
	*(outbuf++) = (UCHAR)((tmp_ulong >> 40) & 0xFF);
	*(outbuf++) = (UCHAR)((tmp_ulong >> 32) & 0xFF);
      }
      *(outbuf++) = (UCHAR)((tmp_ulong >> 24) & 0xFF);
      *(outbuf++) = (UCHAR)((tmp_ulong >> 16) & 0xFF);
      *(outbuf++) = (UCHAR)((tmp_ulong >> 8) & 0xFF);
      *(outbuf++) = (UCHAR)(tmp_ulong & 0xFF);
    }

#else

    /* copy output buffer to 'output' */
    memcpy(outbuf, cd->outBuffer.data, tmpBufLen);
    outbuf += tmpBufLen;

#endif

    outBufLen += tmpBufLen;
    cd->outBuffer.length = 0;
    image += m_bpl;
  }


  /* Write optional EOFB */
  if(eofb) {
    if(!(Terminate(cd, G4_K_VAL)))
      rethrow_err("failed writing terminate code");

    /* output new data */
    tmpBufLen = cd->outBuffer.length;
	if (tmpBufLen)
	{

DBG_WATCH("G4 flushing final data ", tmpBufLen);

    if((tmpBufLen + outBufLen) > outBufMax)
	{
	  m_memErr = TRUE;
      throw_err0();
	}

#ifdef _LITTLE_ENDIAN
    
    lbuf = (ULONG *)(cd->outBuffer.data);
    for(j = 0; j < tmpBufLen; j+=sizeof(ULONG)) {
      tmp_ulong = *(lbuf++);
      if(BITS_ULONG == 64) {
	*(outbuf++) = (UCHAR)((tmp_ulong >> 56) & 0xFF);
	*(outbuf++) = (UCHAR)((tmp_ulong >> 48) & 0xFF);
	*(outbuf++) = (UCHAR)((tmp_ulong >> 40) & 0xFF);
	*(outbuf++) = (UCHAR)((tmp_ulong >> 32) & 0xFF);
      }
      *(outbuf++) = (UCHAR)((tmp_ulong >> 24) & 0xFF);
      *(outbuf++) = (UCHAR)((tmp_ulong >> 16) & 0xFF);
      *(outbuf++) = (UCHAR)((tmp_ulong >> 8) & 0xFF);
      *(outbuf++) = (UCHAR)(tmp_ulong & 0xFF);
    }

#else

    /* copy output buffer to 'output' */
    memcpy(outbuf, cd->outBuffer.data, tmpBufLen);
    outbuf += tmpBufLen;

#endif

    outBufLen += tmpBufLen;
    cd->outBuffer.length = 0;
	}
  }

	/* flush remaining bytes if necessary */
  if(cd->outMod > 0) {
    rem = (cd->outMod + BITS_UCHAR - 1) / BITS_UCHAR;
    tmp_ulong = cd->outWord;

DBG_WATCH("G4 flushing bytes ", rem);

    if((outBufLen + rem) > outBufMax)
	{
	  m_memErr = TRUE;
      throw_err0();
	}
    
    for(j = 0; j < rem; j++) {
      *(outbuf++) = (UCHAR)((tmp_ulong >> ((BITS_ULONG - BITS_UCHAR) -
					   (j * BITS_UCHAR))) & 0xFF);
    }

    outBufLen += rem;

    cd->outWord = 0;
    cd->outMod = 0;
  }

  /* return output data size */
  *outlen = outBufLen;

  return TRUE;

bailout:

  return FALSE;
}

