#ifndef Misc_H_INCLUDED
#define Misc_H_INCLUDED

/*

  The functions in this file are just alternate versions of the << 
  streaming operator for some of the vairous objects we find in out
  printer driver.  These are used only for debugging.

*/

// Note that you are going to need winddi.h for STROBJ.

inline std::ostream & operator<<( std::ostream & lhs, const STROBJ & rhs )
{
    if( rhs.pwszOrg )
    {
        for( size_t ii=0; ii < rhs.cGlyphs; ++ii )
        {
            WCHAR val = rhs.pwszOrg[ii];

            if( val < 0x0100 )
            {
                lhs << char(val);
            }
            else
            {
                lhs << '\\' << (void*)(val);
            }
        }
    }
    else
    {
        lhs << "<NULL STROBJ>";
    }

    return lhs;
}

inline std::ostream & operator<<( std::ostream & lhs, const WCHAR * rhs )
{
    if( rhs )
    {
        lhs << "\"";

        for( size_t ii=0; rhs[ii]; ++ii )
        {
            WCHAR val = rhs[ii];

            if( val < 0x0100 )
            {
                lhs << char(val);
            }
            else
            {
                lhs << '\\' << (void*)(val);
            }
        }

        lhs << "\"";

    }
    else
    {
        lhs << "<NULL WCHAR STRING>";
    }

    return lhs;
}


struct C8Dot4
{
    C8Dot4( const FIX & fix )
    {
        assert( sizeof(long) == 4 );

        fract_ = fix & 0x0f;

        value_ = fix >> 4;

        if( fix < 0 )
        {
            value_ |= 0xF8000000;
        }
    }

    long value_;
    long fract_;
};

inline std::ostream & operator<<( std::ostream & lhs, const C8Dot4 & fix )
{
    return lhs << long(fix.value_) << '.' << long(fix.fract_);
}

inline std::ostream & operator<<( std::ostream & lhs, const POINTFIX & pfx )
{
    return lhs << "( " << C8Dot4(pfx.x) << ", " << C8Dot4(pfx.y) << " )";
}

inline std::ostream
operator<<( std::ostream & lhs, const std::basic_string<WCHAR> & rhs )
{
    for( size_t ii=0; ii < rhs.length(); ++ii )
    {
        lhs << char(rhs[ii]);
    }

    return lhs;
}

inline static bool operator<( const DRVFN & lhs, const DRVFN & rhs )
{
    return lhs.iFunc < rhs.iFunc;
}


/*

  BitPtr is a pointer-like object that allows you to conveniently access
  single bits in an array of bytes.  This is particularly useful for
  dealing with monochrome displays.  Be warned however, it does not
  quite have true pointer semantics.

  For the most part, it behaves exactly as a pointer.  operator++() moves
  the pointer 1 bit.  operator+=() moves it an arbitrary number of bits.
  operator* is a const, non-reference operator, because there is no way
  to return a reference to a single bit.  Most bit manipultions 
  wiil probably happen with set(), flip(), and clear().

  Because BitPtr was built for images, the bits go 0x80, 0x40 ... 0x02, 
  0x01.  That is to say, the high-order bit is the first one accessed
  in each byte.
*/

class BitPtr
{
public:

    explicit BitPtr( unsigned char * ptr = NULL ) { reset( ptr ); }

    // Default copy semantics.
    // BitPtr( const BitPtr & rhs );
    // BitPtr & operator=( const BitPtr & rhs );

    ~BitPtr( void ) {} // Not virtual

    // Advance pointer 1 bit.
    BitPtr & operator++( void )
    {       
        m_byte += (m_bit & 0x01);
        m_bit = ( (m_bit << 7) | (m_bit >> 1) );

        return *this;
    }

    // Advance pointer 1 bit.
    BitPtr & operator--( void )
    {       
        m_byte -= ((m_bit & 0x80) >> 7);
        m_bit = ( (m_bit >> 7) | (m_bit << 1) );

        return *this;
    }

    // Advance pointer n bits.
    BitPtr & operator +=( size_t n )
    {
        // Move byte the number of whole bytes.
        m_byte += (n/8);

        // Find out the left-over number of bits.
        n %= 8;

        if( n )
        {
            // Put our bit in a larger integer.
            unsigned int val = m_bit;

            // Now copy the bit one byte over.  When we shift off remain
            // the bit that is in the low-order byte will be the one that
            // we want.
            val |= (val << 8);

            // Right shift the bit reamin spaces.  If this is enough to take
            // us to the next byte, the orginal bit is shifted off.  And the
            // one we just created ends up in the right place.  If it is not
            // enough, then the original bit ends up in the right position
            // and the new bit ends up staying in the higher byte.
            val >>= n;

            // Set bit to the new value.
            m_bit = (val & 0xff);

            // Increase byte if needed.  If both bits are still present, then
            // one of the bits will be in the 9th place or higher.  Both bits
            // still present indicates that we have not rolled over.
            m_byte += (val <= 0x80);
        }
    
        return *this;
    }

    // Read the bit currently pointed to.
    bool operator*( void ) const { return !!((*m_byte) & m_bit); }
    
    //  Each of set(), clear(), and flip() operate on the current bit.
    void set( bool val ) { val? set() : clear(); }  
    void set  ( void ) { *m_byte |=  m_bit; }
    void clear( void ) { *m_byte &= ~m_bit; }
    void flip ( void ) { *m_byte ^=  m_bit; }

    // Reassign this pointer to reference some other data.
    void reset( unsigned char * p ) { m_byte = p; m_bit  = 0x80; }

    // Move this pointer n bytes.
    void offset( size_t n ) { m_byte += n; }

    // Get the pointer to the current byte.  Beware that the current
    // bit may  not by 0x80.
    unsigned char * bytePtr( void ) const { return m_byte; }

    unsigned char bit( void ) const { return m_bit; }

    // Bet the orderinal number of the current bit.
    size_t bitNum( void ) const
    {
        switch( m_bit )
        {
            case 0x01 : return 7;
            case 0x02 : return 6;
            case 0x04 : return 5;
            case 0x08 : return 4;

            case 0x10 : return 3;
            case 0x20 : return 2;
            case 0x40 : return 1;

            case 0x80 : 
            default   : break;
        };

        return 0;
    };

private:

    // Points to the data we are currently accessing.
    unsigned char * m_byte;
    
    // A mask telling us which bit we are on.  m_bit should never be
    // anything other than a power of 2.  That is, exactly one bit
    // should always be set.
    unsigned char m_bit;
};


/*

 NibblePtr is a relative of BitPtr.  It's job is to simplify operations
 on groups of 4 bits.
   
*/

class NibblePtr
{
public:

    explicit NibblePtr( unsigned char * ptr = NULL ) { reset( ptr ); }

    explicit NibblePtr( const BitPtr & bp )
    : m_byte( bp.bytePtr() ),
      m_nibble( (bp.bit() == 0x80)? 0xf0 : 0x0f )

    {
        // This construction really only makes sense if the BitPtr is
        // lined up on a nibble.
        assert( (bp.bit() == 0x08) ||
                (bp.bit() == 0x80)   );
    }

    // Default copy semantics.
    // BitPtr( const BitPtr & rhs );
    // BitPtr & operator=( const BitPtr & rhs );

    ~NibblePtr( void ) {} // Not virtual

    // Advance our pointer 1 nibble.
    NibblePtr & operator++( void )
    {       
        m_byte   += (m_nibble & 0x01);
        m_nibble ^= 0xff;

        return *this;
    }

    NibblePtr & operator +=( size_t n )
    {
        m_byte += (n/2);

        if( n % 2 ) { ++(*this); }
    
        return *this;
    }
    
    unsigned char operator*( void ) const
    {
        unsigned char val = (*m_byte) & m_nibble;

        val |= (val >> 4);

        return val & 0x0f;
    }
    
    // All of these operate on all four bits together.
    void setAll  ( void ) { *m_byte |=  m_nibble; }
    void clearAll( void ) { *m_byte &= ~m_nibble; }
    void flipAll ( void ) { *m_byte ^=  m_nibble; }

    // set4() sets each of the bits in the nibble.  only the low-order
    // four bits of val are used regardless of whether we are on an even
    // or odd nibble.  It is illegal to have any of the high-order bits
    // set.  (Note the assertion.)
    void set4( unsigned char val )
    {
        assert( !(val & 0xf0) );

        set4Fast( val | (val << 4) );
    }

    // This is like set4(), but it assumes that you have put the data
    // into the correct nibble.  Because  it turns out that allowing
    // garbage to be in the other nibble is sometimes useful for this
    // optimization case, this method will safely ignore the nibble
    // that we are not interested in.  You can see how this is used
    // by set4() above.
    void set4Fast( unsigned char val )
    {
        *m_byte &= ~m_nibble;
        *m_byte |= (val & m_nibble);
    }

    // Make our pointer reference new data.
    void reset( unsigned char * ptr )
    {
        m_byte   = ptr;
        m_nibble = 0xf0;
    }

    unsigned char * bytePtr( void ) const { return m_byte; }

    // Count the transitions in our nibble.  I am not sure whether
    // this method really belongs here or deep in our segmentation
    // code.
    //
    // A "transition" is a switch from a 0 to a 1.  Thus, 0x00 and
    // 0x0f (or 0xf0) have no transitions.  0x05 and 0x0A have 
    // 3 transitions.
    //
    // In the case of three transitions, we actually return 4 because
    // we consider this an important case.
    size_t transCount( void ) const
    {
        unsigned char val = (*m_byte) & m_nibble;

        if( val & 0x88 )
        {
            val = (~val) & m_nibble;
        }

        if( val & 0x11 )
        {
            return (val == (0x55 & m_nibble))?  4 : 1;
        }

        return val? 2 : 0;
    }

private:

    // Pointer to our data.
    unsigned char * m_byte;
    
    // A mask for the current nibble.  This must always be one of
    // 0xf0 or 0x0f.
    unsigned char m_nibble;
};

/*

 Ptr24Bit allows convenient access to 24-bit image data packed into
 an array of bytes.  operator++() moves the pointer 3 bytes, etc.

*/

class Ptr24Bit
{
public:
    
    explicit Ptr24Bit( unsigned char * p = NULL ) : m_ptr(p) {}

    // Default copy semantics.
    // Ptr24Bit( const Ptr24Bit & );
    // Ptr24Bit & operator=( const Ptr24Bit & );

    ~Ptr24Bit( void ) {}
    
    // All of the following comparisons have the standard pointer
    // arithmatic meanings.
    bool operator==(const Ptr24Bit &rhs) const {return m_ptr == rhs.m_ptr;}
    bool operator!=(const Ptr24Bit &rhs) const {return m_ptr != rhs.m_ptr;}
    bool operator< (const Ptr24Bit &rhs) const {return m_ptr <  rhs.m_ptr;}
    bool operator> (const Ptr24Bit &rhs) const {return m_ptr >  rhs.m_ptr;}
    bool operator<=(const Ptr24Bit &rhs) const {return m_ptr <= rhs.m_ptr;}
    bool operator>=(const Ptr24Bit &rhs) const {return m_ptr >= rhs.m_ptr;}
    
    // Move our pointer 1 24-bit pixel.
    Ptr24Bit & operator++( void ) { m_ptr += 3; return *this; }
    Ptr24Bit & operator--( void ) { m_ptr -= 3; return *this; }

    // Move our pointer n 24-bit pixels.
    Ptr24Bit & operator+=( size_t n ) { m_ptr += n*3; return *this; }
    Ptr24Bit & operator-=( size_t n ) { m_ptr -= n*3; return *this; }

    Ptr24Bit operator+( size_t n ) { return Ptr24Bit( m_ptr + (n*3) ); }
    Ptr24Bit operator-( size_t n ) { return Ptr24Bit( m_ptr - (n*3) ); }

    void reset( unsigned char * p ) { m_ptr  = p; }
    
    // Move our pointer n bytes.
    Ptr24Bit & offset( size_t n ) { m_ptr += n; return *this; }

    unsigned char * bytePtr( void ) const { return m_ptr; }

    // The rgb functions assume that r is the low-order byte and g
    // is the high-order byte.

    unsigned char & r( void ) const { return m_ptr[0]; }
    unsigned char & g( void ) const { return m_ptr[1]; }
    unsigned char & b( void ) const { return m_ptr[2]; }

    void setRGB( unsigned char r_, unsigned char g_, unsigned char b_ ) const
    {
        r() = r_;
        g() = g_;
        b() = b_;
    }

    // Copy the data pointed to by another Ptr24Bit to our own data
    // area.
    void copyDataFrom( const Ptr24Bit & src ) const
    {
        this->m_ptr[0] = src.m_ptr[0];
        this->m_ptr[1] = src.m_ptr[1];
        this->m_ptr[2] = src.m_ptr[2];
    }

    void swapData( Ptr24Bit & that ) const
    {
        unsigned char tmp[3];

        tmp[0] = that.m_ptr[0];
        tmp[1] = that.m_ptr[1];
        tmp[2] = that.m_ptr[2];

        that.m_ptr[0] = this->m_ptr[0];
        that.m_ptr[1] = this->m_ptr[1];
        that.m_ptr[2] = this->m_ptr[2];

        this->m_ptr[0] = tmp[0];
        this->m_ptr[1] = tmp[1];
        this->m_ptr[2] = tmp[2];       
    }

    bool hasSameData( const Ptr24Bit & that ) const
    {
        return (this->r() == that.r()) &&
               (this->g() == that.g()) &&
               (this->b() == that.b());
    }

    bool isWhite( void ) const { return (r() & g() & b()) == 0xff; }
    bool isBlack( void ) const { return (r() | g() | b()) == 0x00; }

    bool isDontCare( void ) const { return !(r() | g() | (unsigned char)(~b())); }
    void markDontCare( void ) const { setRGB( 0, 0, 0xff ); }

    void forceCare( void ) const { if( isDontCare() ) { m_ptr[2] = 0xfe; } }

private:

    // Pointer to the first byte of the current 24-bit pixel.
    unsigned char * m_ptr;
};


#endif
//#ifndef Misc_H_INCLUDED
