#ifndef tiff_h_included
#define tiff_h_included

/*

  TiffHeader is a simple class that ecapsulates the 8 byte header
  at the beginning of a tiff file.  For information on what all of
  the values really mean, you will have to look at a tiff file
  specification, however, the description here should be sufficient
  for most purposes.

  To use this object to write tiff header, create it.  Then, set the
  endianness you want and the value of the IFD offset.

  I have not bothered making a virtual destructor as I cannot imagine
  anybody every inheritting from TiffHeader.

*/

class TiffHeader
{
public:

    TiffHeader( void ) : m_byteOrder('II'), m_ifdOffset(0) {}

    // Default copy semantics.
    //
    // TiffHeader( const TiffHeader & );
    // TiffHeader & operator=( const TiffHeader & );

    void setLittleEndian( void ) { m_byteOrder = 'II'; }
    void setBigEndian   ( void ) { m_byteOrder = 'MM'; }
    
    bool isLittleEndian( void ) const { return m_byteOrder == 'II'; }
    bool isBitEndian   ( void ) const { return m_byteOrder == 'MM'; }

    void  setIfdOffset( DWORD o ) { m_ifdOffset = o; }
    DWORD getIfdOffset( void ) const { return m_ifdOffset; }

    // bufSize() and getBuf() are used to write this data to a file.
    // bufSize() will always return 8, of course.  We do this, as
    // opposed to reinterpret_casting this-> as an unsigned char *
    // to make sure everything is platform, compiler, and optimization
    // independent.
    const size_t bufSize( void ) const { return sizeof(m_buf); }

    unsigned char * getBuf( void )
    {
        // 42 is a magic number used in tiff headers.
        const WORD tiffId = 42;

        assert( sizeof(m_byteOrder) == 2 );
        assert( sizeof(tiffId)      == 2 );
        assert( sizeof(m_ifdOffset) == 4 );

        unsigned char * ptr = m_buf;

        memcpy( ptr, &m_byteOrder, sizeof(m_byteOrder) );
        ptr += sizeof(m_byteOrder);

        memcpy( ptr, &tiffId, sizeof(tiffId) );
        ptr += sizeof(tiffId);

        memcpy( ptr, &m_ifdOffset, sizeof(m_ifdOffset) );

        return m_buf;
    }

private:

    WORD  m_byteOrder;
    DWORD m_ifdOffset;

    // A tiff header is 8 bytes long.
    unsigned char m_buf[8];
};

/*

  And IfdEntry is an entry in an Ifd (see below.)  An IfdEntry is always
  12 bytes, although the last 8 bytes may be an offset to more data.

  I have not bothered with the virtual destructor here.

*/

class IfdEntry
{
public:
    
    IfdEntry( void ): m_tag(0), m_type(0), m_count( 1 ), m_valueOffset(0) {}

    enum { IFD_ENTRY_SIZE = 12 };

    enum Type
    {
        BYTE      =  1,
        ASCII     =  2,
        SHORT     =  3,
        LONG      =  4,
        RATIONAL  =  5,
        SBYPE     =  6,
        UNDEFINED =  7,
        SSHORT    =  8,
        SLONG     =  9,
        SRATIONAL = 10,
        FLOAT     = 11,
        DOUBLE    = 12
    };

    // Methods for setting the values in an entry.
    void setTag   ( WORD  t ) { m_tag  = t; }
    void setType  ( Type  t ) { m_type = static_cast<WORD>(t); }
    void setCount ( DWORD c ) { m_count = c; }
    void setOffset( DWORD o ) { m_valueOffset = o; }

    // If the offset is actually data, we use the following
    // 4 methods to set the data instead of calling setOffset(). 
    // Data is stored in the offset if it is less than or equal
    // to 4 bytes.
    void setByteValue( unsigned char b, size_t loc = 0)
    {
        setType( BYTE );
        assert( loc < 4 );
        (reinterpret_cast<unsigned char *>(&m_valueOffset))[loc] = b;
    }

    void setShortValue( WORD s, size_t loc = 0 )
    {
        setType( SHORT );
        assert( loc < 2 );
        (reinterpret_cast<WORD *>(&m_valueOffset))[loc] = s;
    }

    void setLongValue( DWORD l )
    {
        setType( LONG );
        m_valueOffset = l;
    }

    // bufSize() and getBuf() are used to write the IFD entry to a file.
    // bufSize() will always return 12, of course.  We do this, as
    // opposed to reinterpret_casting this-> as an unsigned char *
    // to make sure everything is platform, compiler, and optimization
    // independent.
    const size_t bufSize( void ) const { return sizeof(m_buf); }

    unsigned char * getBuf( void )
    {
        assert( (sizeof(m_tag  )      +
                 sizeof(m_type )      +
                 sizeof(m_count)      +
                 sizeof(m_valueOffset)) == bufSize() );

        unsigned char * ptr = m_buf;
        
        memcpy( ptr, &m_tag, sizeof(m_tag) );
        ptr += sizeof(m_tag);
        
        memcpy( ptr, &m_type, sizeof(m_type) );
        ptr += sizeof(m_type);
        
        memcpy( ptr, &m_count, sizeof(m_count) );
        ptr += sizeof(m_count);

        memcpy( ptr, &m_valueOffset, sizeof(m_valueOffset) );

        return m_buf;
    }

private:

    WORD  m_tag;
    WORD  m_type;
    DWORD m_count;
    DWORD m_valueOffset;

    unsigned char m_buf[IFD_ENTRY_SIZE];
};

/*
 
 IFD stands for image file directory.  This is where all of the header
 information about an image and how it is encoded goes.  The version 
 we have implemented here is a quick-and-dirty implementation that will
 only work for uncompressed monochrome and 24 bit color images.

 I have not bothered making the destructor virtual as I cannot imagine
 any derived classes (without a thorough rewrite).

*/

class Ifd
{
    // This is the number of entries in our IFD.
    enum { ENTRY_COUNT = 11 };

public:
    
    Ifd( void ) : m_nextIfdOffset(0)
    {
        m_fields[0]. setTag( 256 );
        m_fields[1]. setTag( 257 );
        m_fields[2]. setTag( 258 );
        m_fields[3] .setTag( 259 );
        m_fields[4]. setTag( 262 );
        m_fields[5]. setTag( 273 );
        m_fields[6]. setTag( 277 );
        m_fields[7]. setTag( 278 );
        m_fields[8]. setTag( 279 );
        m_fields[9]. setTag( 282 );
        m_fields[10].setTag( 283 );

        m_fields[0] .setType( IfdEntry::LONG     );
        m_fields[1] .setType( IfdEntry::LONG     );
        m_fields[2] .setType( IfdEntry::SHORT    );
        m_fields[3] .setType( IfdEntry::SHORT    );
        m_fields[4] .setType( IfdEntry::SHORT    );
        m_fields[5] .setType( IfdEntry::LONG     );
        m_fields[6] .setType( IfdEntry::SHORT    );        
        m_fields[7] .setType( IfdEntry::LONG     );
        m_fields[8] .setType( IfdEntry::LONG     );
        m_fields[9] .setType( IfdEntry::RATIONAL );
        m_fields[10].setType( IfdEntry::RATIONAL );    
    }
    
    // Default copy semantics.
    // Ifd( const Ifd & );
    // Ifd & operator=( const Ifd & );

    // bufSize() and getBuf() are used to write this data to a file.
    virtual const size_t bufSize( void ) const { return sizeof(m_buf); }

    virtual unsigned char * getBuf( void )
    {       
        unsigned char * ptr = m_buf;
        
        const WORD entryCount = ENTRY_COUNT;

        memcpy( ptr, &entryCount, sizeof(entryCount) );
        ptr += sizeof(entryCount);
        
        for( size_t field = 0; field < ENTRY_COUNT; ++field )
        {
            memcpy( ptr, m_fields[field].getBuf(), m_fields[field].bufSize() );
            ptr += m_fields[field].bufSize();
        }
        
        memcpy( ptr, &m_nextIfdOffset, sizeof(m_nextIfdOffset) );
        ptr += sizeof(m_nextIfdOffset);

        return m_buf;
    }

    // For now I have just left these public, as they need to be
    // manipulated by people who use this class.
    IfdEntry   m_fields[ENTRY_COUNT];
    DWORD      m_nextIfdOffset;

private:

    // Ok.  There are some annoying constant numbers here.  An ifd consists
    // of:
    // 1.  A 2-byte WORD indicating the number of entries in the directory.
    // 2.  ENTRY count 12 bytes entries.
    // 3.  A 4-byte DWORD indicating the location of the next IFD or 0 if
    //     this is the final IFD.
    unsigned char m_buf[2 + (size_t(ENTRY_COUNT)*size_t(IfdEntry::IFD_ENTRY_SIZE)) + 4];
};

#endif
//#ifndef tiff_h_included
