/******************************************************
 * dgpver
 *
 * 5.10.2000 / dhtesq.
 * Following Jeff Breidenbach's demonstration of using background dataglyphs
 * to hold a document's content, we are close to building a document
 * "verifier" or "tamper-proofer", which can compare the glyph contents
 * with a scan of a received page.
 *
 * This component does the actual comparison.  Given two raw black & white
 * images, it runs the lossy tokenizer and requires an exact dictionary match
 * for the two images.  In general it is not feasible to require a position
 * block match, because one of the images comes from a scan path.  This
 * way we can avoid the registration problem.
 *
 * Most of this comes from the encoder.
 *
 ******************************************************/

#include "verify.h"
#include "dgp_err_control.h"

static CLOpts cl_opts = {   DEF_VERBOSE,
                            DEF_FRACHAUSD, 
                            DEF_WIDTHDIFF, 
                            DEF_HEIGHTDIFF, 
                            DEF_MAXDENSITY,
                            DEF_COMMITTHRESH,
                            DEF_FREEZEHISTTHRESH,
                            DEF_MAKEHISTTHRESH,
                            DEF_RETURNHISTTHRESH,
                            DEF_DILATIONRADIUS,
                            DEF_SIZETHRESH,
                            DEF_WRITETHRESH,
                            DEF_NEARESTNEIGHBORS,
                            DEF_XSHIFT,
                            DEF_XSLOPECHANGE,
                            DEF_XNEWSLOPE,
                            DEF_NONTP,
							//DEF_LDL,
                            //DEF_INCRDICT,
                            //DEF_CLEARDICT,
                            DEF_BASELINEORIENT,
                            DEF_DESKEWACCURACY,
                            //DEF_HALFTONEDETECT,
                            //DEF_OUTPUT_RES,
                            DEF_DRIFT,
                            DEF_NODESKEW,
                            DEF_NOALIGN,
                            //DEF_WRITECLASSES,
                            //DEF_INSTRUMENT,
                            //DEF_G4TOKENS,
                            DEF_ERRFILENAME,
                            DEF_NOERROROUT,
                            DEF_POSTMORTEM,
                            DEF_BOUNDINGBOXES };

// Displayed when showing usage information.
static const Dgp_Int8 *cl_header = "Usage: dgpver [options] trusted.tif suspicious.tif\n \
    \tVerifates first pages of input files. Results sent to stdout.";

/*
 * The definition of the command line options.
 */
static DgpCl_Option cl_cmdln_options[] = {
  DGPCL_HELP_CMD,
  DGPCL_VERBOSE_CMD(cl_opts.verbose),
  DGPCL_NEW_FLOAT_CMD("ph", cl_opts.fracHausd, "Pct. of Hausdorff distance to use for model match", DEF_FRACHAUSD_STR),
  DGPCL_NEW_INT_CMD("wd", cl_opts.widthDiff, "Highest allowed width difference for matching boxes", DEF_WIDTHDIFF_STR),
  DGPCL_NEW_INT_CMD("hd", cl_opts.heightDiff, "Highest allowed height difference for matching boxes", DEF_HEIGHTDIFF_STR),
  DGPCL_NEW_INT_CMD("md", cl_opts.maxDensity, "Allowed number of mismatches per 3 x 3 cell", DEF_MAXDENSITY_STR),
  DGPCL_NEW_INT_CMD("ct", cl_opts.commitThresh, "Population threshold for committing a class", DEF_COMMITTHRESH_STR),
  DGPCL_NEW_INT_CMD("ft", cl_opts.freezeHistThresh, "Population threshold for freezing class histogram", DEF_FREEZEHISTTHRESH_STR),
  DGPCL_NEW_INT_CMD("mt", cl_opts.makeHistThresh, "Population threshold for class histogram creation", DEF_MAKEHISTTHRESH_STR),
  DGPCL_NEW_INT_CMD("rt", cl_opts.returnHistThresh, "Population threshold for class histogram return", DEF_RETURNHISTTHRESH_STR),
//10
  DGPCL_NEW_INT_CMD("dr", cl_opts.dilationRadius, "Dilation radius (0..9). Select 10, 14, or 20 for non-tp preserving 1, sqrt(2), and 20", 
    DEF_DILATIONRADIUS_STR),
  DGPCL_NEW_INT_CMD("st", cl_opts.sizeThresh, "Lower size threshold for CCs", DEF_SIZETHRESH_STR),
  DGPCL_NEW_INT_CMD("wt", cl_opts.writeThresh, "Population threshold for writing a class", DEF_WRITETHRESH_STR),
  DGPCL_NEW_INT_CMD("nn", cl_opts.nearestNeighbors, "Neighborhood size for matching shifts", DEF_NEARESTNEIGHBORS_STR),
  DGPCL_NEW_INT_CMD("xshift", cl_opts.xShift, "Small token error curve, offset parameter", DEF_XSHIFT_STR),
  DGPCL_NEW_INT_CMD("xsc", cl_opts.xSlopeChange, "Small token error curve, slope change parameter", DEF_XSLOPECHANGE_STR),
  DGPCL_NEW_FLOAT_CMD("xns", cl_opts.xNewSlope, "Small token error curve, new slope parameter", DEF_XNEWSLOPE_STR),
  DGPCL_NEW_FLAG_CMD("nontp", cl_opts.nontp, "Force non-topology-preserving dilation"),
  //DGPCL_NEW_INT_CMD("ldl", cl_opts.lowerDilationLimit, "Lower dilation limit; only dilate cc's this size and larger", DEF_LDL_STR),
  //DGPCL_INC_DICT_CMD(cl_opts.incrementalDict, DEF_INCRDICT_STR),
  //DGPCL_CLR_DICT_CMD(cl_opts.clearDict, DEF_CLEARDICT_STR),
  DGPCL_NEW_STRING_CMD("bo", cl_opts.baselineOrient, "Baseline orientation for deskewing [\"down\", \"left\", \"right\", \"up\"]", 
    DEF_BASELINEORIENT),
//20
  DGPCL_NEW_INT_CMD("da", cl_opts.deskewAccuracy, "Deskew accuracy: 0 (fast) to 5 (accurate)", DEF_DESKEWACCURACY_STR),
  //DGPCL_HALFTONE_CMD(cl_opts.halftoneDetection),
  //DGPCL_RES_CMD(cl_opts.outputRes, DEF_OUTPUT_RES_STR),
  DGPCL_NEW_INT_CMD("rd", cl_opts.relativeDrift, "Allowed relative drift between cc's", DEF_DRIFT_STR),
  DGPCL_NEW_FLAG_CMD("nodeskew", cl_opts.noDeskew, "Do not deskew"),
  DGPCL_NEW_FLAG_CMD("noalign", cl_opts.noAlign, "Do not perform shifts when determining bounding box positions"),
  //DGPCL_NEW_FLAG_CMD("wc", cl_opts.writeClasses, "Write classes to tmp file indexed by class and dictionary number"),
  //DGPCL_NEW_FLAG_CMD("inst", cl_opts.instrument, "Instrument run time"),
  //DGPCL_NEW_FLAG_CMD("g4t", cl_opts.g4Tokens, "Encode singleton and outlier tokens as group 4"),
  DGPCL_NEW_STRING_CMD("e", cl_opts.err_filename, "Set error file name", DEF_ERRFILENAME),
  DGPCL_NEW_FLAG_CMD("noe", cl_opts.noe, "Suppress error output; overrides -e"),
  DGPCL_NEW_STRING_CMD("pm", cl_opts.postMortem, "Write postmortem tif to this location", DEF_POSTMORTEM),
//30
  DGPCL_NEW_STRING_CMD("bb", cl_opts.boundingBoxes, "Write bounding box info to this location", DEF_BOUNDINGBOXES),
  {NULL}
};

/*************************************************************************/
// An argument validator...
static ErrCode
validateArgs(Dgp_State *dgp_state,
             Dgp_Int32 unparsed_argc, 
             Dgp_Int8 **unparsed_argv,
             FILE *err_out,
             void *validate_cookie)
{
  DGP_PROLOGUE("validateArgs()", DGP_PERRSTATE);
  CLOpts *my_opts = (CLOpts *)validate_cookie;

  unparsed_argc = unparsed_argc;
  unparsed_argv = unparsed_argv;

  /*
  if((my_opts->outputRes > 2400) || (my_opts->outputRes < 0)) {
    DGPCL_PRINT_ERRF0(err_out, "Valid resolutions are 0 < res < 2400...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  */
  if((my_opts->fracHausd < (Dgp_Float32)0.0) || (my_opts->fracHausd > (Dgp_Float32)1.0)){
    my_opts->fracHausd = my_opts->fracHausd / (Dgp_Float32)100;
    DGPCL_PRINT_ERRF0(err_out, "Interpreting Hausdorff distance as percentage...");
    //DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->fracHausd < (Dgp_Float32)0.0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid Hausdorff distance...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->widthDiff < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid width difference...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->heightDiff < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid height difference...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->maxDensity < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid max density...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->commitThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid commit threshold...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->freezeHistThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid historgram freeze threshold...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->makeHistThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid histogram make threshold...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->returnHistThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid histogram return threshold...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if((my_opts->dilationRadius < 0) || ((my_opts->dilationRadius > 9) &&
      (my_opts->dilationRadius != 10) && (my_opts->dilationRadius != 14) && (my_opts->dilationRadius != 20))){
    DGPCL_PRINT_ERRF0(err_out, "Invalid dilation radius (use 0...9, 10, 14, or 20).");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->sizeThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid size threshold...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->writeThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid write threshold...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  /*
  if(my_opts->incrementalDict < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid incremental dictionary parameter...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->clearDict < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid clear dictionary parameter...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  */
  if(my_opts->relativeDrift < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid relative drift parameter...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if((my_opts->deskewAccuracy < 0) || (my_opts->deskewAccuracy > 5)){
    DGPCL_PRINT_ERRF0(err_out, "Invalid deskew accuracy...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if((my_opts->nearestNeighbors != 0) && (my_opts->nearestNeighbors != 4) && (my_opts->nearestNeighbors != 8)
      && (my_opts->nearestNeighbors != 24)){
    DGPCL_PRINT_ERRF0(err_out, "Invalid nearest neighbor parameter; specify 0, 4, 8, or 24.");
    DGP_ERR(cErrInvalidCmdLine);
  }
  switch(my_opts->baselineOrient[0]){
  case 'd': case 'D':   case 'u':   case 'U':
  case 'l': case 'L':   case 'r':   case 'R':
      break;
  default:
    DGPCL_PRINT_ERRF0(err_out, "Invalid baseline orientation; specify \"down\", \"up\", \"left\", \"right\", \
        or at least the first letter.");
    DGP_ERR(cErrInvalidCmdLine);
  }

  NORMAL_RETURN;
}
/*************************************************************************/

/*********************************************************************
 *  Make an oversized guess about how many dictionaries there will be.
 *
 *  It assumes one dictionary per page.  And then adds extras for 
 *   the incremental and cleared dictionaries.
 *
 *  clearDict is how many pages before a completely fresh dictionary;
 *     0 means don't ever write a clear dictionary (until the end).
 *  incrDict is how many pages before a incremental dictionary;
 *     0 means don't ever write an incremental dictionary.
 *
 *********************************************************************/     
Dgp_Int32 
estimateNumDicts(Dgp_Int32 num_pages, Dgp_Int32 clearDict, Dgp_Int32 incrDict)
{
  Dgp_Int32 numIncr, numClear;

  if (incrDict == 0) {
    incrDict = num_pages + 1;
  }

  if (clearDict == 0) {
    clearDict = num_pages + 1;
  }

  numIncr = num_pages / incrDict;
  numClear = num_pages / clearDict; 
  
  return numIncr + numClear + num_pages + 1;
} 

/*
void setXiffCallbacks(OutputCallbacks *cb)
{
	cb->createOutputState = (CreateOutputState) createXiffCB;
	cb->closeOutputState = (CloseOutputState) closeXiffCB;
	cb->writeOutputPageDictionary = (WriteOutputPageDictionary) writeXiffPageDictionaryCB;
	cb->finishOutputPage = (FinishOutputPage) finishXiffPageCB;
	cb->writeOutputAnnotations = (WriteOutputAnnotations) writeXiffAnnotationsCB;
	cb->writeOutputPosBlock = (WriteOutputPosBlock) writeXiffPosBlockCB;
	cb->writeOutputDictionary = (WriteOutputDictionary) writeXiffDictionaryCB;
	cb->writeOutputBinaryPixr = (WriteOutputBinaryPixr) writeXiffBinaryPixrCB;
	cb->freeOutputDictblock = (FreeOutputDictblock) freeXiffDictblockCB;
	cb->userData = NULL;
}
*/

void setJbig2Callbacks(OutputCallbacks *cb)
{
	cb->createOutputState = (CreateOutputState) createJb2StateCB;
	cb->closeOutputState = (CloseOutputState) closeJb2StateCB;
	cb->writeOutputPageDictionary = (WriteOutputPageDictionary) writeJb2PageDictionaryCB;
	cb->finishOutputPage = (FinishOutputPage) finishJb2PageCB;
	cb->writeOutputPosBlock = (WriteOutputPosBlock) writeJb2PosBlockCB;
	cb->writeOutputDictionary = (WriteOutputDictionary) writeJb2DictionaryCB;
	cb->writeOutputBinaryPixr = (WriteOutputBinaryPixr) writeJb2BinaryPixrCB;
	cb->freeOutputDictblock = (FreeOutputDictblock) freeJb2DictblockCB;
	cb->userData = NULL;
}


/*****************************************************************
 * Main program...
 *****************************************************************/
int
main(int argc, char **argv)
{
  DGP_MAIN_PROLOGUE("main()", stderr, Dgp_fileErrPrinter);

  Dgp_Int8      **infile_names = (Dgp_Int8 **)NULL;
  Dgp_Int8      *errfname = NULL, *postMortem = NULL, *boundingBoxes = NULL;
  Dgp_Bool      no_err_out = DGP_FALSE;
  Dgp_State     *dgp_state = NULL;

  FILE          *errfile = (FILE *)NULL;
  FILE          *bb = (FILE *)NULL;
  Dgp_Int32     num_dicts, tempArgc;
  TCLYDocCookie tdc = NULL;
  DgpTdc_Cookie llTdc = NULL;
  TCLYSectionBlockPtr sbPtr = NULL;
  //DgpTdc_SectionBlock *llSb = NULL;
  Dgp_Bool      report = DGP_FALSE;
  tokOptions    tokopt;
  LlTokOptions  *llTokOpt;
  TCL_options   opts;
  OutputCallbacks callbacks;


  if(argc == 1){
      DgpCl_printHelp(stderr, cl_header, cl_cmdln_options);
      MAIN_NORMAL_RETURN;
  }

  /* Init dgp state. */
  DGP_ERR_PASS(Dgp_initDgpState(&dgp_state, NULL, NULL, stderr, Dgp_fileErrPrinter,
      DGP_TRUE, DGP_FALSE));
  DGP_SET_ERR_STATE(DGP_PERRSTATE);
  DGP_ERR_PASS_CLEANUP(w_strel1Inits( dgp_state->ipwstate));

  // Defaults set in this function will be overridden by defaults for command-line interface.
  tokSetDefaults(&tokopt);
  tokopt.flags |= DEF_OPTIONS_FLAGS;

  /* Parse command line. */
  tempArgc = (Dgp_Int32)argc;
  DGP_ERR_CHECK_NO_MSG(DgpCl_parse(dgp_state, &tempArgc, argv, cl_cmdln_options, stderr, cl_header, validateArgs, &cl_opts)){
      if(DGP_ERROR_IS(cErrInvalidCmdLine)){
          MAIN_NORMAL_RETURN;
      }
  }
  argc = tempArgc;
  if(argc != 2){
      DGPCL_PRINT_ERRF0(stderr, "Specify two input files for comparison.");
      DGP_ERR(cErrInvalidCmdLine);
  }
  infile_names = argv;
  
  //opts.xiff = DGP_TRUE;
  tokopt.nneigh = cl_opts.nearestNeighbors;
  tokopt.maxDense = cl_opts.maxDensity;
  tokopt.widthDiff = cl_opts.widthDiff;
  tokopt.heightDiff = cl_opts.heightDiff;
  tokopt.fracHausd = cl_opts.fracHausd;
  tokopt.commitThresh = cl_opts.commitThresh;
  tokopt.freezeHistThresh = cl_opts.freezeHistThresh;
  tokopt.makeHistThresh = cl_opts.makeHistThresh;
  tokopt.xshift = cl_opts.xShift;
  tokopt.xslopechange = cl_opts.xSlopeChange;
  tokopt.newslope = cl_opts.xNewSlope;
  tokopt.dilationRadius = cl_opts.dilationRadius;
  tokopt.flags |= ((cl_opts.nontp == DGP_TRUE) ? TOK_OPT_NONTPDIL:0);
  tokopt.lowerDilationLimit = 0;
  if((cl_opts.dilationRadius == 10) || (cl_opts.dilationRadius == 14) || (cl_opts.dilationRadius == 20)){
      tokopt.flags |= TOK_OPT_NONTPDIL;
  }
  opts.sizeThresh = cl_opts.sizeThresh;
  opts.classThresh = cl_opts.writeThresh;
  //opts.incrDict = cl_opts.incrementalDict;
  //opts.clearDict = cl_opts.clearDict;
  opts.halftone = DGP_FALSE;//cl_opts.halftoneDetection;
  //opts.forceRes = cl_cmdln_options[22].found == DGP_TRUE ? cl_opts.outputRes:-1;
  opts.skewAccuracy = cl_opts.deskewAccuracy;
  switch(cl_opts.baselineOrient[0]){
  case 'd': case 'D':
      opts.baselineOrient = BO_DOWN;
      break;
  case 'u': case 'U':
      opts.baselineOrient = BO_UP;
      break;
  case 'l': case 'L':
      opts.baselineOrient = BO_LEFT;
      break;
  case 'r': case 'R':
      opts.baselineOrient = BO_RIGHT;
      break;
  }
  opts.deskew = !cl_opts.noDeskew;
  if(cl_opts.noAlign){
      tokopt.flags &= ~TOK_OPT_ALIGN;
  }
  opts.writeclasses = DGP_FALSE;//cl_opts.writeClasses;
  if(opts.writeclasses){
      tokopt.flags |= TOK_OPT_COPYBITMAPS;
  }
  opts.instrument = DGP_FALSE;//cl_opts.instrument;
  //opts.g4_toks = cl_opts.g4Tokens;
  errfname = cl_opts.err_filename;
  no_err_out = cl_opts.noe;
  if(cl_cmdln_options[25].found == DGP_TRUE){
      postMortem = cl_opts.postMortem;
  }
  if(cl_cmdln_options[26].found == DGP_TRUE){
      boundingBoxes = cl_opts.boundingBoxes;
      if(strcmp(boundingBoxes, "stdout")) {
      DGP_ERR_GEN_CLEANUPF1((!(bb = fopen(boundingBoxes, "w"))), cErrFileNotFound,
          "I can't open bounding boxes file \"%s\" for writing.", boundingBoxes);
    } else {
      bb = stdout;
    }
  }
  
  if(!no_err_out){
    if(strcmp(errfname, "stderr")) {
      DGP_ERR_GEN_CLEANUPF1((!(errfile = fopen(errfname, "w"))), cErrFileNotFound,
	      "I can't open error log file \"%s\" for writing.", errfname);
      dgp_state->err_state->userCookie = errfile;
    } else {
      errfile = stderr;
    }
  } else {
    dgp_state->err_state->userCookie = NULL;
  }
  report = cl_opts.verbose;
  tokopt.dgp_state = dgp_state;

  //////////////////////////

  // Setup callback functions. 
  setJbig2Callbacks(&callbacks);

  // This call assumes one page per document.
  num_dicts = estimateNumDicts(argc, DEF_CLEARDICT, DEF_INCRDICT);

  DGP_ERR_PASS_CLEANUP(tcly_doc_init(	dgp_state, 
                                        &tdc, 
                                        DGP_TRUE, 
                                        //color, 
                                        opts.halftone, 
                                        opts.sizeThresh, 
                                        DGP_FALSE,
										DGP_FALSE,
                                        &tokopt,
                                        opts.deskew,
                                        opts.skewAccuracy,
                                        opts.baselineOrient,
                                        -1,
                                        report));

  // These lossless tokenizer options are all unused now, but if at some point
  // you want to verify with a mixture of lossless and lossy matching (like the
  // normal mono encoder), they'll start to mean something.

  DGP_CHECK_ALLOC_CLEANUP(llTokOpt = (LlTokOptions *)DGP_MALLOC(sizeof(LlTokOptions)));

  llTokOpt->ascii = DGP_FALSE;
  llTokOpt->color = DGP_FALSE;
  llTokOpt->colorTokenizerParams = NULL;
  llTokOpt->dgp_state = dgp_state;
  llTokOpt->flags = 0;
  llTokOpt->halftone = DGP_FALSE;
  llTokOpt->lossy = DGP_TRUE;
  llTokOpt->maxSamptokH = -1;
  llTokOpt->maxSamptokH = -1;
  llTokOpt->noascii = DGP_TRUE;
  llTokOpt->numStrips = 1;
  llTokOpt->sizeThresh = opts.sizeThresh;
  llTokOpt->tokenizerParams = NULL;
  llTokOpt->verbose = cl_opts.verbose;
  llTokOpt->washThresh = 301;
  DGP_ERR_PASS_CLEANUP(DgpTdc_initDoc(	dgp_state, 
										&llTdc, 
										llTokOpt->lossy,
										llTokOpt->color,
										llTokOpt->halftone,
										llTokOpt->sizeThresh,
										(Dgp_UInt16)llTokOpt->washThresh,
										DGPTDC_SAMP_COMPTYPE_JPEG, 
										NULL,
										llTokOpt->tokenizerParams,
										llTokOpt->colorTokenizerParams,
										llTokOpt->verbose));
  ((tdoc_state *)llTdc)->llTokenizerOptions = llTokOpt;
  

  DGP_ERR_PASS_CLEANUP(DgpVer_processInfiles(	dgp_state, 
                                                tdc, 
												llTdc,
                                                infile_names,
                                                &tokopt,
                                                cl_opts.relativeDrift,
                                                postMortem,
                                                bb,
                                                report));

  DGP_ERR_PASS_CLEANUP(DgpTdc_endDoc(&llTdc, ((tdoc_state *)llTdc)->curBlock));
  sbPtr = (TCLYSectionBlockPtr)((TCLYDocState *)tdc)->curBlock;
  DGP_ERR_CHECK_NO_MSG(tcly_doc_end(tdc, sbPtr)) {
      DGP_WARNF0("Problem cleaning up; continuing anyway.");
      DGP_CLEAR_CUR_ERROR;
  }
  tdc = NULL;
  sbPtr = NULL;
  
  DGP_ERR_PASS_CLEANUP(w_strel1Term(dgp_state->ipwstate));
  
  if(errfile && (errfile != stderr)){
      fclose(errfile);
  }
  if(bb && (bb != stdout)){
      fclose(bb);
  }

  Dgp_cleanupDgpState(&dgp_state);

  MAIN_NORMAL_RETURN;
  
ERR_CLEANUP:
  if(errfile && (errfile != stderr)){
      fclose(errfile);
  }
  if(bb && (bb != stdout)){
      fclose(bb);
  }

  if(tdc != NULL) tcly_doc_end(tdc, sbPtr);

  if(dgp_state != NULL)  Dgp_cleanupDgpState(&dgp_state);
  
  MAIN_ERR_RETURN;
}
