/*
 *
 * verify-utils.c
 *
 * 20.10.2000 / dhtesq.
 * Support functions for dgpver.
 *
 *
 */

#include "verify.h"


// Comparision function for qsort.
// Sorts in raster order, not distance from the origin.
int
compareULPosition(  const void *elem0,
                    const void *elem1)
{
    dgplssy_cc *cc0 = *(dgplssy_cc **)elem0;
    dgplssy_cc *cc1 = *(dgplssy_cc **)elem1;

    if(cc0->y > cc1->y){
        return 1;
    } 
    
    if(cc0->y == cc1->y) {
        if(cc0->x > cc1->x) {
            return 1;
        } else if(cc0->x == cc1->x) {
            return 0;
        }
    }
    return -1;
}    

// Comparision function for qsort.
// Sorts in ascending x.
int
compareXPosition(  const void *elem0,
                    const void *elem1)
{
    dgplssy_cc *cc0 = *(dgplssy_cc **)elem0;
    dgplssy_cc *cc1 = *(dgplssy_cc **)elem1;
    
    if(cc0->x > cc1->x) {
        return 1;
    } else if(cc0->x < cc1->x) {
        return 0;
    }

    return -1;
}    


/*
  Returns an array of cc pointers sorted according to the qsort compare function.

  The cc's are taken from the passed list.

  No. of cc's is returned too.

  Passed parameters are unchanged in the event of an error.
*/
ErrCode
DgpVer_sortListByPosition(  Dgp_State *dgp_state,
                            DgpAdt_List list,
                            dgplssy_cc ***ccArray,
                            Dgp_Int32 *ccArrayMemberCount)
{
  DGP_PROLOGUE("DgpVer_sortListByPosition()", DGP_PERRSTATE);

  Dgp_Int32 ccCount, i;
  dgplssy_cc **ccs = NULL;
  DgpAdt_ListNode ln0;

  ccCount = DGPADT_LILEN(list);

  // Allocate for cc pointers.
  DGP_CHECK_ALLOC_CLEANUP(ccs = (dgplssy_cc **)DGP_MALLOC(ccCount * sizeof(*ccs)));

  // Put cc's into unsorted array.
  i = 0;
  DGPADT_FOREACH(ln0, list){
      ccs[i++] = (dgplssy_cc *)DGPADT_LIGET(ln0);
  }

  // Now sort them.
  qsort((void *)ccs, (size_t)ccCount, sizeof(*ccs), compareXPosition);

  if(ccArrayMemberCount){
      *ccArrayMemberCount = ccCount;
  }
  if(ccArray){
      *ccArray = ccs;
  }

  NORMAL_RETURN;

ERR_CLEANUP:
  if(ccs){
      DGP_FREE(ccs);
  }
  ERR_RETURN;
}

/*
Retrieves local registration information by searching the passed xref list for the passed
cc, and averaging as best it can in a local neighborhood whose size is set by the relativeDrift parameter.

The xref{} contains previously recorded position difference information (reference - scanned).
*/
ErrCode
DgpVer_retrieveLocalRegistration(   Dgp_State   *dgp_state,
                                    DgpAdt_List  xrefList,
                                    dgplssy_cc          *box,
                                    Dgp_Int32   relativeDrift,
                                    Dgp_Int32   xRegGlobal,
                                    Dgp_Int32   yRegGlobal,
                                    Dgp_Int32   *xRegLocal,
                                    Dgp_Int32   *yRegLocal)
{
  DGP_PROLOGUE("DgpVer_retrieveLocalRegistration()", DGP_PERRSTATE);

  DgpAdt_ListNode ln0, aln;
  ccXref *xref, *box_xref = NULL;
  tokCBoxType *cbox = NULL;
  dgplssy_cc *scan_box, *ref_box, *closest_box;
  Dgp_Float64 d, d_min;
  Dgp_Int32 xd, yd, x_reg_offset, y_reg_offset, reg_count;

  DGP_ASSERT(xRegLocal);
  DGP_ASSERT(yRegLocal);

  x_reg_offset = y_reg_offset = reg_count = 0;
  box_xref = NULL;

  DGPADT_FOREACH(ln0, xrefList){
      xref = (ccXref *)DGPADT_LIGET(ln0);
      DGP_ASSERT(xref != NULL);
      scan_box = xref->box;
      DGP_ASSERT(scan_box->pgnum >= 1);

      if(scan_box == box){
          if((xref->xReg != -1) && (xref->yReg != -1)){
              *xRegLocal = xref->xReg;
              *yRegLocal = xref->yReg;
              NORMAL_RETURN;
          } else {
              box_xref = xref;
          }
      }

      //xref->xReg = xref->yReg = -1;

      if((Dgp_Float64)(DGPVER_INTERBOX(scan_box, box)) > (Dgp_Float64)(DGPVER_LOCAL_REG_MEMBER_RADIUS * relativeDrift)){
          continue;
      }

      // Might have a loner...
      if(!xref->classSiblings) continue;
      closest_box = NULL;
      d_min = (Dgp_Float64)INT_MAX;

      // When looking for local tokens to register against, only look at matches...
      DGPADT_FOREACH(aln, xref->classSiblings){
          cbox = tokCMemberBox(aln);
          DGP_ASSERT(cbox != NULL);
          ref_box = (dgplssy_cc *)cbox->userdata;
          DGP_ASSERT(ref_box != NULL);
          if(ref_box->pgnum >= 1) continue;
          
          xd = abs(ref_box->x - xref->box->x - xRegGlobal);
          yd = abs(ref_box->y - xref->box->y - yRegGlobal);
          d = sqrt(xd*xd + yd*yd);
          if(d < d_min){
              d_min = d;
              if(d < (DGPVER_LOCAL_REG_RADIUS * (Dgp_Float64)(relativeDrift))){
                  closest_box = ref_box;
              }
          }
      }

      if(!closest_box){
          continue;
      }
      x_reg_offset += closest_box->x - xref->box->x;      // (reference - scanned)...
      y_reg_offset += closest_box->y - xref->box->y;
      reg_count++;

      closest_box = NULL;
  }
////////////////////////////

  //DGP_ASSERT(box_xref);

  if(!reg_count){
      DGP_WARNF0("Couldn't find local registration tokens.  Local neighborhood too small?");
      *xRegLocal = *yRegLocal = 0;
      NORMAL_RETURN;
  }

  *xRegLocal = (Dgp_Int32)((Dgp_Float32)x_reg_offset/(Dgp_Float32)reg_count);  
  *yRegLocal = (Dgp_Int32)((Dgp_Float32)y_reg_offset/(Dgp_Float32)reg_count);

  if(box_xref){
      box_xref->xReg = *xRegLocal;
      box_xref->yReg = *yRegLocal;
  }
  
  NORMAL_RETURN;
}


/*
Generates global registration data, using those scan tokens for which the tokenizer was able to find
a nearby match in the reference.  Averaged over a whole page, global registration is typically not quite 
good enough for verification.

The registration numbers follow the convention (reference - scanned).

For those tokens for which we have confident registration, we stick them on the passed list (confidents).
*/
ErrCode
DgpVer_retrieveGlobalRegistration(  Dgp_State   *dgp_state,
                                    DgpAdt_List xrefList,
                                    DgpAdt_List confidents,
                                    Dgp_Int32   relativeDrift,
                                    Dgp_Int32   *xRegGlobal,
                                    Dgp_Int32   *yRegGlobal,
                                    Dgp_Int32   *scanTokensFound,
                                    Dgp_Bool    report)
{
  DGP_PROLOGUE("DgpVer_retrieveGlobalRegistration()", DGP_PERRSTATE);

  Dgp_Int32 x_reg_offset, y_reg_offset, reg_count, xd, yd;
  Dgp_Float64 d, d_min;
  DgpAdt_ListNode ln, aln, retnode;
  ccXref *xref;
  dgplssy_cc *scan_box, *closest_box, *ref_box;
  tokCBoxType *cbox;

  DGP_ASSERT(xRegGlobal);
  DGP_ASSERT(yRegGlobal);
  DGP_ASSERT(scanTokensFound);

  // The idea here is to roughly register the text on the scan with respect to the 
  // text on the fiducial.
  // We'll traverse the xref list, looking for the closest corresponding reference token.
  // For more exact registration, take position differences on a token-by-token basis.
  x_reg_offset = y_reg_offset = reg_count = 0;

  DGPADT_FOREACH(ln, xrefList){
      xref = (ccXref *)DGPADT_LIGET(ln);
      DGP_ASSERT(xref != NULL);
      scan_box = xref->box;

      xref->xReg = xref->yReg = -1;

      closest_box = NULL;
      d_min = (Dgp_Float64)INT_MAX;
      // Might have a loner...
      if(!xref->classSiblings) continue;

      DGPADT_FOREACH(aln, xref->classSiblings){
          cbox = tokCMemberBox(aln);
          DGP_ASSERT(cbox != NULL);
          ref_box = (dgplssy_cc *)cbox->userdata;
          DGP_ASSERT(ref_box != NULL);
          if(ref_box->pgnum >= 1) continue;
          
          xd = abs(ref_box->x - scan_box->x);
          yd = abs(ref_box->y - scan_box->y);
          d = sqrt(xd*xd + yd*yd);
          if(d < d_min){
              d_min = d;
              if(d < (DGPVER_GLOBAL_REG_RADIUS * (Dgp_Float64)relativeDrift)){                  
                  closest_box = ref_box;
              }
          }
      }

      if(!closest_box){
          // It may have mistakenly matched something elsewhere on the page (not the corresponding reference token).

          //DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(lonerList, (void *)scan_box, &retnode));
          (*scanTokensFound)--;          

          
          //DGP_REPORTF0("Problems registering a matched token within specified tolerance; continuing...");
          DGP_REPORTF4(report, "Unregisterable (but matched) %d x %d scan token @%d, %d", scan_box->bitmap.width, scan_box->bitmap.height,
              scan_box->x, scan_box->y);
          continue;
      }
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(confidents, (void *)scan_box, &retnode));
      // Mark reference box.
      closest_box->tok = DGPVER_GLOBAL_REGISTRATOR;
      x_reg_offset += closest_box->x - scan_box->x;      // (reference - scanned)...
      y_reg_offset += closest_box->y - scan_box->y;
      reg_count++;
      closest_box = NULL;
  }

  if(!reg_count){
      //DGP_REPORTF0(report, "DgpVer_retrieveGlobalRegistration() / Nothing to register.  Is it possible you gave me an empty cross-reference list?");
      NORMAL_RETURN;
  }

  *xRegGlobal = (Dgp_Int32)((Dgp_Float32)x_reg_offset/(Dgp_Float32)reg_count);
  *yRegGlobal = (Dgp_Int32)((Dgp_Float32)y_reg_offset/(Dgp_Float32)reg_count);

  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}


/*
Writes bounding box information of problem areas to disk.

Internal but trivial format.
*/
void
DgpVer_writeStinkers(   DgpAdt_List stinkers, 
                        FILE *bb, 
                        Dgp_Int8 **infiles, 
                        Dgp_Int32 x_reg,
                        Dgp_Int32 y_reg,
                        Dgp_Bool useReg)
{

    DgpAdt_ListNode ln;
    dgplssy_cc *box;

    fprintf(bb, "# Reference:  %s\n", infiles[0]);
    fprintf(bb, "# Scanned:  %s\n", infiles[1]);

    fprintf(bb, "# Rough registration (x, y) (ref - scan):  %d, %d\n", (int)x_reg, (int)y_reg);
    fprintf(bb, "# Bottom right of bounding boxes (w x h + r + b):\n");

    if(stinkers){
        DGPADT_FOREACH(ln, stinkers){
            box = (dgplssy_cc *)DGPADT_LIGET(ln);
            
            if(useReg == DGP_TRUE){
                // box is a reference token...
                fprintf(bb, "%dx%d+%d+%d\n", 
                    box->bitmap.width, box->bitmap.height, (int)(box->x + x_reg + box->bitmap.width - 1), 
			        (int)(box->y + y_reg + box->bitmap.height - 1));
            } else {
                // box is already a scan token...
                fprintf(bb, "%dx%d+%d+%d\n", box->bitmap.width, box->bitmap.height, 
    			(int)(box->x + box->bitmap.width - 1), (int)(box->y + box->bitmap.height - 1));
            }
        }
    }
    return;
}


/*
Blat the passed list of cc's off (XOR) the passed image.

The 'offset' parameter is added to the x and y positions of all cc's in the list.
This is useful when blatting to a pixr image (e.g.) with a fixed-size boundary.
*/
ErrCode
DgpVer_removeCCsFromImage(  Dgp_State       *dgp_state,
                            DgpAdt_List     confidents,
                            UInt8           *image,
                            Dgp_Int32       width,
                            Dgp_Int32       height,
                            Dgp_Int32       bpl,
                            Dgp_Int32       offset)
{
  DGP_PROLOGUE("DgpVer_removeCCsFromImage()", DGP_PERRSTATE);

  DgpAdt_ListNode ln;
  dgplssy_cc *box;

  //offset = 0;
  DGPADT_FOREACH(ln, confidents){
      box = (dgplssy_cc *)DGPADT_LIGET(ln);
      DGP_ASSERT(box);

      DgpBlat_blitBitsOp((Dgp_UInt32 *)image,
                         (bpl) / sizeof(Dgp_Int32),
                         0, 0,
                         width + offset, height + offset,
                         box->bitmap.bits,
                         box->bitmap.longsperline,
                         box->bitmap.width,
                         box->bitmap.height,
                         box->x + offset, box->y + offset,
                         DGPBLAT_OP_COVER,
                         dgp_state);
  }
  NORMAL_RETURN;
}
