/******************************************************
 * verify.h
 *
 * 5.10.2000 / dhtesq.
 * Following Jeff Breidenbach's demonstration of using background dataglyphs
 * to hold a document's content, we are close to building a document
 * "verifier" or "tamper-proofer", which can compare the glyph contents
 * with a scan of a received page.
 *
 * This component does the actual comparison.  Given two raw black & white
 * images, it runs the lossy tokenizer and requires an exact dictionary match
 * for the two images.  In general it is not feasible to require a position
 * block match, because one of the images comes from a scan path.  This
 * way we can avoid the registration problem.
 *
 ******************************************************/

#ifdef _WIN32
/* Include winsock2.h under NT. */
#include <winsock2.h>
#endif

#include <limits.h>
#include <math.h>
#include "dgp_state.h"
#include "tc-tools-exec.h"
#include "dgp_sys_err_control.h"
#include "tc-tiffio.h"
#include "tclydoc.h"
#include "tclyout.h"
#include "writejbigCB.h"
//#include "writeXiff3CB.h"
#include "dgp_cl.h"         // Standardized command-line interface definitions.
#include "tiffio.h"

#include <libiow.pub>
#include <libpixrw.pub>
#include <strel.pub>
#include "compress.h"
#include "blit.h"
#include "compressUtil.h"
#include "token-compress-priv.h"
#include "tiff_cc2box.h"

// These are the definitions and tables for the fast cross-dilation used in the tokenizer.
#include "cross.h"


// Baseline orientations for deskewing.
#define BO_DOWN   0
#define BO_UP     1
#define BO_LEFT   2
#define BO_RIGHT  3

/*
 * Default values for command line arguments.
 */
#define DEF_VERBOSE     DGP_FALSE

#define DEF_FRACHAUSD   (Dgp_Float32)(0.96)
#define DEF_FRACHAUSD_STR   "0.96"

#define DEF_WIDTHDIFF   2
#define DEF_WIDTHDIFF_STR   "2"

#define DEF_HEIGHTDIFF  2
#define DEF_HEIGHTDIFF_STR  "2"

#define DEF_MAXDENSITY  3
#define DEF_MAXDENSITY_STR  "3"

#define DEF_COMMITTHRESH 10000
#define DEF_COMMITTHRESH_STR "10000"

#define DEF_FREEZEHISTTHRESH 10000
#define DEF_FREEZEHISTTHRESH_STR "10000"

#define DEF_MAKEHISTTHRESH 10
#define DEF_MAKEHISTTHRESH_STR "10"

#define DEF_RETURNHISTTHRESH 0
#define DEF_RETURNHISTTHRESH_STR "0"

#define DEF_DILATIONRADIUS 14
#define DEF_DILATIONRADIUS_STR "14"

#define DEF_SIZETHRESH  5
#define DEF_SIZETHRESH_STR  "5"

#define DEF_WRITETHRESH 10
#define DEF_WRITETHRESH_STR "10"

#define DEF_NEARESTNEIGHBORS 24
#define DEF_NEARESTNEIGHBORS_STR "24"

#define DEF_XSHIFT      3
#define DEF_XSHIFT_STR      "3"

#define DEF_XSLOPECHANGE 6
#define DEF_XSLOPECHANGE_STR "6"

#define DEF_XNEWSLOPE   2.0
#define DEF_XNEWSLOPE_STR   "2.0"

#define DEF_NONTP       DGP_FALSE

#define DEF_LDL			0
#define DEF_LDL_STR		"0"

#define DEF_INCRDICT    0
/*
#define DEF_INCRDICT_STR    "0"
*/

#define DEF_CLEARDICT   2
/*
#define DEF_CLEARDICT_STR   "2"
*/

#define DEF_BASELINEORIENT "down"

#define DEF_DESKEWACCURACY 3        // [0...5]
#define DEF_DESKEWACCURACY_STR "3"

#define DEF_HALFTONEDETECT DGP_FALSE


#define DEF_OUTPUT_RES      300
/*
#define DEF_OUTPUT_RES_STR  "same as input"
*/

// Allowed relative drift of cc's...
#define DEF_DRIFT       14
#define DEF_DRIFT_STR   "14"

#define DEF_NODESKEW    DGP_FALSE
#define DEF_NOALIGN     DGP_FALSE
#define DEF_WRITECLASSES DGP_FALSE
#define DEF_INSTRUMENT  DGP_FALSE
//#define DEF_G4TOKENS    DGP_FALSE

#define DEF_ERRFILENAME "stderr"
#define DEF_NOERROROUT  DGP_FALSE

#define DEF_POSTMORTEM  "<none>"
#define DEF_BOUNDINGBOXES "<none>"

#define DEF_OPTIONS_FLAGS (TOK_OPT_ALIGN | TOK_OPT_RETURNOFFSET | TOK_OPT_COPYBITMAPS)

typedef struct {
  Dgp_Bool verbose;
  Dgp_Float32 fracHausd;
  Dgp_Int32 widthDiff;
  Dgp_Int32 heightDiff;
  Dgp_Int32 maxDensity;
  Dgp_Int32 commitThresh;
  Dgp_Int32 freezeHistThresh;
  Dgp_Int32 makeHistThresh;
  Dgp_Int32 returnHistThresh;
  Dgp_Int32 dilationRadius;
  Dgp_Int32 sizeThresh;
  Dgp_Int32 writeThresh;
  Dgp_Int32 nearestNeighbors;
  Dgp_Int32 xShift;
  Dgp_Int32 xSlopeChange;
  Dgp_Float32 xNewSlope;
  Dgp_Bool nontp;
  //Dgp_Int32 lowerDilationLimit;
  //Dgp_Int32 incrementalDict;
  //Dgp_Int32 clearDict;
  Dgp_Int8 *baselineOrient;
  Dgp_Int32 deskewAccuracy;
  //Dgp_Bool halftoneDetection;
  //Dgp_Int32 outputRes;
  Dgp_Int32 relativeDrift;
  Dgp_Bool noDeskew;
  Dgp_Bool noAlign;
  //Dgp_Bool writeClasses;
  //Dgp_Bool instrument;
  //Dgp_Bool g4Tokens;
  Dgp_Int8 *err_filename;
  Dgp_Bool noe;
  Dgp_Int8 *postMortem;
  Dgp_Int8 *boundingBoxes;
} CLOpts;

typedef struct {
    Dgp_Int32 sizeThresh;
    Dgp_Int32 classThresh;
    Dgp_Int32 incrDict;
    Dgp_Int32 clearDict;
    Dgp_Int32 skewAccuracy;
    Dgp_Int32 baselineOrient;
    Dgp_Int32 rotation;
    Dgp_Bool halftone;
    Dgp_Int32 forceRes;
    Dgp_Bool writeclasses;
    Dgp_Bool instrument;
    Dgp_Bool deskew;
    Dgp_Bool g4_toks;
    Dgp_Bool xiff;              /* If True, output xiff (default), else output jbig2. */
} TCL_options;

// For finding distance between two cc's...
#define DGPVER_INTERBOX(cc0, cc1)       sqrt(pow(cc0->x - cc1->x, 2) + pow(cc0->y - cc1->y, 2))

// Factor by which relative drift parameter is multiplied to arrive at a radius within which
// matched scan tokens are to be considered confidently registered.
#define DGPVER_GLOBAL_REG_RADIUS        (3.0)

// Factor by which relative drift parameter is multiplied to arrive at a radius within which
// fellow scan tokens are to be considered in local registration computations.
#define DGPVER_LOCAL_REG_MEMBER_RADIUS        (10.0)

// Factor by which relative drift parameter is multiplied to arrive at a radius within which
// scan and reference tokens are to be considered confidently registered for purposes of calculating
// local registration.
#define DGPVER_LOCAL_REG_RADIUS        (3.0)

// Factor by which relative drift parameter is multiplied to arrive at a radius within which
// a reference token is a candidate for overblatting an unmatched scan token.
#define DGPVER_OVERBLAT_RADIUS          (0.75)

// Factor by which relative drift parameter is multiplied to arrive at a radius within which
// we require reference tokens not have a match if they are to be used for overblatting.
#define DGPVER_NEIGHBORHOOD_MATCH_RADIUS          (0.5)

// Factor by which relative drift parameter is multiplied to arrive at a radius within which
// scan tokens are allowed to wander before we scream erasure.
#define DGPVER_POSITIONAL_MATCH_RADIUS          (1.3)

// My code for fiducial cc's that have been used to break joined scanned cc's.
#define DGPVER_SIAMESETWIN      (void *)0x01

// ...for fiducial cc's that have been used to register scanned cc's (globally).
#define DGPVER_GLOBAL_REGISTRATOR    (void *)0x02

void
DgpVer_reportFindings( Dgp_State *dgp_state,
                Dgp_Int8 **infiles,
                Dgp_Bool documentVerified,
                Dgp_Int32 refBoxCount,
                Dgp_Int32 scanBoxCount,
                Dgp_Int32 extraScanClassCount,
                DgpAdt_List stinkers,
                Dgp_Int32 x_reg,
                Dgp_Int32 y_reg,
                FILE *bb);

void
DgpVer_writeStinkers(   DgpAdt_List stinkers, 
                        FILE *bb, 
                        Dgp_Int8 **infiles, 
                        Dgp_Int32 x_reg,
                        Dgp_Int32 y_reg,
                        Dgp_Bool useReg);

ErrCode
DgpVer_processInfiles(
	Dgp_State *dgp_state, 
	TCLYDocCookie tdc, 
	DgpTdc_Cookie _llTdc,
	Dgp_Int8 **infiles,
    tokOptions *tokopt,
    Dgp_Int32 driftTolerance,
    Dgp_Int8 *postMortem,
    FILE *bb,
	Dgp_Bool report) ;

#if 0
ErrCode
DgpVer_dilateCC(    Dgp_State *dgp_state,
                    dgplssy_cc *box,
                    PIXR **dest);
#endif

ErrCode
DgpVer_sortListByPosition(  Dgp_State *dgp_state,
                            DgpAdt_List list,
                            dgplssy_cc ***ccArray,
                            Dgp_Int32 *ccArrayMemberCount);

ErrCode writeMonoPage(  Dgp_State                   *dgp_state,
                        TCImagePtr                  tcImagePtr,
                        Dgp_Int8                    *outName,
                        TIFF                        **tiff,
                        Dgp_Bool                        firstPage,
                        Dgp_Bool                        lastPage);

ErrCode
DgpVer_retrieveGlobalRegistration(  Dgp_State   *dgp_state,
                                    DgpAdt_List xrefList,
                                    DgpAdt_List confidents,
                                    Dgp_Int32   relativeDrift,
                                    Dgp_Int32   *xRegGlobal,
                                    Dgp_Int32   *yRegGlobal,
                                    Dgp_Int32   *scanTokensFound,
                                    Dgp_Bool    report);

ErrCode
DgpVer_retrieveLocalRegistration(   Dgp_State   *dgp_state,
                                    DgpAdt_List  xrefList,
                                    dgplssy_cc *box,
                                    Dgp_Int32   relativeDrift,
                                    Dgp_Int32   xRegGlobal,
                                    Dgp_Int32   yRegGlobal,
                                    Dgp_Int32   *xRegLocal,
                                    Dgp_Int32   *yRegLocal);

ErrCode
DgpVer_removeCCsFromImage(  Dgp_State       *dgp_state,
                            DgpAdt_List     confidents,
                            UInt8           *image,
                            Dgp_Int32       width,
                            Dgp_Int32       height,
                            Dgp_Int32       bpl,
                            Dgp_Int32       offset);
