  
#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif
#include <stdio.h>

#include "silx_encode.h"
#include "silx_encode_priv.h"
#include "tclydoc.h"
#include "jtokapi.h"
#include "tc-tiffio.h"
#include "dgp_sys_err_control.h"
#include "dgp_cl.h"
#include "jb2mmuncompress.h"
#include "iowrap.h"
#include "pixr.h"
#include "dgp_err_printers.h"
#include "simple_utils.h"
#include "tok-lossless.h"
#include "libpixrw.pub"

static double utime ()
{
        return (double)clock()/CLOCKS_PER_SEC;
}

/********************************************************************
 * Set general encoding options. Put NULL if you don't want to set that value
 * You have to call silx_initEncState first.
 *
 * @param    cookie SilxEncCookie
 * @param    *halftoneDetection Flag to attempt to do halftone detection when set to DGP_TRUE. Default is DGP_FALSE.
 * @param    *deskew            Flag to attempt to do deskew when set to DGP_TRUE. (Default= DGP_FALSE).
 * @param    *deskewAccuracy    Deskew accuracy: 0 (fast) to 5 (accurate)(Default = 3)
 * @param    *outputTokRes      Default input file resolution. Only used if no resolution specified in file
 *                              (Default = 300)
 * @param    **baselineOrient   Baseline deskewing orientation ("d[own]", "l[eft]", "r[ight]", or "u[p]")
 * @param    *rotation          Rotate clockwise by this multiple of pi/2 before encoding(Default = 0)
 * @param    *jbig2Profile      Jbig2 profile(Default = 2)
 * @param    *g4Tokens          Flag to encode singleton and outlier tokens as group 4 when set to DGP_TRUE. 
 *                              (Default = DGP_FALSE)
 * @param    *tokenSize         Flag disables standard limit of 10% of the page size for max size of tokens
 *                              when set to DGP_TRUE. (Defalut = DGP_FALSE)
 * @param    *verbose           Flag to verbose output when set to DGP_TRUE. (Default = DGP_FALSE)
 * @param    *membond           Token dictionary memory upper limit [-1 / 1 / 2 / 3] to stand for
 *                              [disable / 1 Meg / 2 Meg / 3 Meg]. Default is -1
 * @param    *no_err_out        Flag to disable error output, when set to DGP_TRUE. (Default = DGP_FALSE)
 * @param    *lossless,         Flag to encode in full lossless mode (electronic source) when set to DGP_TRUE. (Default = DGP_FALSE)
 * @param    *incrementalDict   Page interval for incremental dictionaries. Default is 50 pages.
 * @param    *clearDict         Page interval for clearing dictionaries. Default is 100 pages.
 * @param    *redTokNoise       Flag to reduce token noise for each page when set to DGP_TRUE. Color images only. (Default = DGP_FALSE)
 * @param    *jb2Only           Flag to output jb2 stream instead of TFX file when set to DGP_TRUE. (Default = DGP_FALSE)
 */
DGP_LIBEXPORT void 
silx_setGeneralEncOptions(
             SilxEncCookie    cookie,
             Dgp_Bool  *halftoneDetection,
             Dgp_Bool  *deskew,
             Dgp_Int32 *deskewAccuracy,
             Dgp_Int32 *outputTokRes,
             Dgp_Int8  **baselineOrient,
             Dgp_Int32 *rotation,
             Dgp_Int32 *jbig2Profile,
             Dgp_Bool  *g4Tokens,
             Dgp_Bool  *tokenSize,
             Dgp_Bool  *verbose,
             Dgp_Int32 *membond,
             Dgp_Bool  *no_err_out,
             Dgp_Bool  *lossless,
             Dgp_Int32 *incrementalDict,
             Dgp_Int32 *clearDict,
             Dgp_Bool  *redTokNoise,
             Dgp_Bool  *jb2Only)
{
    /* API calls to initiate tokenizer */
    SilxEncState *silxState = (SilxEncState *) cookie;
    DgpApi_Options			*opts=NULL;
    LlTokOptions *llTokOpt = NULL;
    tokOptions *tokOpts = NULL;

    if(!cookie) return;
    opts = &silxState->opts;
    llTokOpt = &silxState->llTokOpt;
    opts = &silxState->opts;
    tokOpts = &silxState->tokOpt;

    if(no_err_out) silxState->no_err_out = *no_err_out;
    if(lossless){
        silxState->lossless = *lossless;
        tokOpts->lowerDilationLimit = silxState->lossless ? INT_MAX:tokOpts->lowerDilationLimit;
    }
    if(halftoneDetection)   {
        opts->halftone  	= *halftoneDetection;
        llTokOpt->halftone = *halftoneDetection;
    }

    if(deskew)              opts->deskew        = *deskew;
    if(outputTokRes) {
        opts->forceRes = *outputTokRes;
        silxState->outputRes = *outputTokRes;
    }
    if(baselineOrient){
        switch((*baselineOrient)[0]){
        case 'd': case 'D':
            opts->baselineOrient = BO_DOWN;
            break;
        case 'u': case 'U':
            opts->baselineOrient = BO_UP;
            break;
        case 'l': case 'L':
            opts->baselineOrient = BO_LEFT;
            break;
        case 'r': case 'R':
            opts->baselineOrient = BO_RIGHT;
            break;
        }
    }
    if(deskewAccuracy) 
        opts->skewAccuracy   = *deskewAccuracy;
    if(rotation){ 
        opts->rotation  	    = *rotation % 4;
        opts->rotation = opts->rotation < 0 ? opts->rotation + 4:opts->rotation;
    }
    if(jbig2Profile)     opts->jbig2Profile  = *jbig2Profile;
    if(g4Tokens)         opts->g4_toks		= *g4Tokens;
    if(tokenSize)         opts->bitmapSizeLimit= !(*tokenSize);
    if(verbose){
        opts->verbose		= *verbose;
        llTokOpt->verbose = *verbose;
    }

    if(membond) {
        switch(*membond){
            case -1:
                break;
            case 1:
                opts->membond = 1024000;    // 1Meg
                break;
            case 2:
                opts->membond = 2048000;    // 2 Meg
                break;
            case 3:
                opts->membond = 32768000;   // 32 Meg
                break;
            default:
                opts->membond = 32768000;   // 32 Meg
                break;
        }
        if(opts->membond > 0){
            // deduct dictionary overhead
            if(opts->jbig2Profile ==2){
                opts->membond -= TOKGENOVERHEAD;
            }else{
                opts->membond -= (TOKGENOVERHEAD + TOKARITHOVERHEAD);
            }
            // deduct halftone dictionary overhead
            if(opts->halftone){
                opts->membond -= TOKHTOVERHEAD;
            }
        }
    }
    if(incrementalDict)
        silxState->incrementalDict = *incrementalDict;
    if(clearDict)
        silxState->clearDict = *clearDict;
    if(redTokNoise)
        silxState->redTokNoise = *redTokNoise;
    if(jb2Only)
        silxState->jb2Only = *jb2Only;
}
/*************************************************************
 * Set Jpeg releated options. Put NULL if you don't want to set that value
 * You have to call silx_initEncState first.
 *
 * @param cookie SilxEncCookie
 * @param *jpegQuality Set the quality of jpeg compression (range 0-100)(Default = 90)
 * @param *jpegRes  Set resolution of output jpegs (Default = 100)
 * @param *jpegScaleFactor Set the scale factor of output jpeg (Default = 1.0f)
 * 
 */
DGP_LIBEXPORT void
silx_setJpegOptions(
             SilxEncCookie          cookie,
             Dgp_Int32              *jpegQuality,
             Dgp_Int32              *jpegRes,
             Dgp_Float32            *jpegScaleFactor)
{
    DgpTdc_JpegCompInfo         *jpegComp;
    DgpApi_Options              *opts;
    SilxEncState *silxState = (SilxEncState *) cookie;
    if(!cookie) return;
    jpegComp = &silxState->jpegComp;
    opts = &silxState->opts;

    if(jpegQuality) {
        jpegComp->jpQuality	= *jpegQuality;
        silxState->jpegQual = *jpegQuality;
    }
    if(jpegRes) silxState->jpegRes = *jpegRes;
    if(jpegScaleFactor) jpegComp->scaleFactor = (Dgp_Float32)*jpegScaleFactor;
}

/**************************************************************************
 * Set tokenization related parameters. Put NULL if you don't want to set that value.
 * You have to call silx_initEncState first.
 *
 * @param classWriteThresh  Population threshold for writing a class (Default = 10)
 * @param fracHausd Pct. of Hausdorff distance to use for model match (Default = 0.97)
 * @param maxDensity Allowed number of mismatches per 3 x 3 cell(Default = 3)
 * @param widthDiff  Highest allowed width difference for matching boxes (Default = 1)
 * @param heightDiff  Highest allowed height difference for matching boxes (Default = 1)
 * @param *commitThresh Population threshold for committing a class(Default = 10000)
 * @param *freezeHistThresh Allowed number of mismatches per 3 x 3 cell(Default = 3)
 * @param *makeHistThresh  Population threshold for class histogram creation (Default = 10)
 * @param *returnHistThresh Population threshold for class histogram return(Default = 0)
 * @param *nearestNeighbors  Neighborhood size for matching shifts (Default = 8)
 * @param *xShift  Small token error curve, offset parameter (Default = 3)
 * @param *xSlopeChange  Small token error curve, slope change parameter (Default = 6)
 * @param *xNewSlope  Small token error curve, new slope parameter (Default = 2.0)
 * @param *dilationRadius   Dilation radius (0..9). 10, 14, or 20 for non-tp, r = 1, sqrt(2), and 2 (Default = 1)
 * @param *lowerDilationLimit When matching, only dilate tokens this size and larger(Default = 6)
 * @param *ccSizeThresh  Minimum cc size, inclusive (Default = 1 (2 for color))
 * 
 */
DGP_LIBEXPORT void 
silx_setTokenOptions(
    SilxEncCookie cookie,
    Dgp_Int32 *classWriteThresh,
    Dgp_Float32 *fracHausd,
    Dgp_Int32 *maxDensity,
    Dgp_Int32 *widthDiff,
    Dgp_Int32 *heightDiff,
    Dgp_Int32 *commitThresh,
    Dgp_Int32 *freezeHistThresh,
    Dgp_Int32 *makeHistThresh,
    Dgp_Int32 *returnHistThresh,
    Dgp_Int32 *nearestNeighbors,
    Dgp_Int32 *xShift,
    Dgp_Int32 *xSlopeChange,
    Dgp_Float32 *xNewSlope,
    Dgp_Int32 *dilationRadius,
    Dgp_Int32 *lowerDilationLimit,
    Dgp_Int32 *ccSizeThresh)
{
    SilxEncState *silxState = (SilxEncState *)cookie;
    DgpApi_Options			*opts=NULL;
    tokOptions *tokOpts = NULL;
    LlTokOptions *llTokOpts = NULL;

    fprintf ("in setTokenOptions %x\n", cookie);
    if(!cookie) return;
    opts = &silxState->opts;
    tokOpts = &silxState->tokOpt;
    llTokOpts = &silxState->llTokOpt;

    // Fill in tokenizer options too...
    if(ccSizeThresh)   {
        opts->sizeThresh    = *ccSizeThresh;
        llTokOpts->sizeThresh = *ccSizeThresh;
    }
    if(classWriteThresh)         opts->classThresh   = *classWriteThresh;
    if(fracHausd) tokOpts->fracHausd	= (Dgp_Float32)*fracHausd;
    if(maxDensity) tokOpts->maxDense	= *maxDensity;
    if(widthDiff) tokOpts->widthDiff	= *widthDiff;
    if(heightDiff) tokOpts->heightDiff	= *heightDiff;
    if(commitThresh) tokOpts->commitThresh	    = *commitThresh;
    if(freezeHistThresh) tokOpts->freezeHistThresh   = *freezeHistThresh;
    if(makeHistThresh) tokOpts->makeHistThresh     = *makeHistThresh;
    if(returnHistThresh) tokOpts->returnHistThresh   = *returnHistThresh;
    if(nearestNeighbors) tokOpts->nneigh		    = *nearestNeighbors;
    if(xShift) tokOpts->xshift		    = *xShift;
    if(xSlopeChange) tokOpts->xslopechange   = *xSlopeChange;
    if(xNewSlope) tokOpts->newslope		= *xNewSlope;
    if(dilationRadius) tokOpts->dilationRadius = *dilationRadius;
    if(lowerDilationLimit) tokOpts->lowerDilationLimit = silxState->lossless ? INT_MAX : *lowerDilationLimit;

}
/**********************************************************************
 * Free SilxEncCookie.
 *
 * @param *dgp_state Pointer to Dgp_State
 * @param *cookie Pointer to SilxEncCookie
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */

DGP_LIBEXPORT ErrCode
silx_cleanEncState (
    Dgp_State *dgp_state,
    SilxEncCookie *sCookie)
{
    DGP_PROLOGUE("silx_cleanEncState()", DGP_PERRSTATE);
    SilxEncState *silxState;
    DgpAdt_ListNode ln;
    DgpApi_PageData *pd_read;

    DGP_ERR_GEN(!sCookie, cErrInvalidParm);
    silxState = (SilxEncState *) *sCookie;
    if(silxState){
        /* close tokenizer */
        if(silxState->pd_list){
                ln = silxState->pd_list->first;
                while (ln != DGP_NULLLISTNODE){
                    // get the page data 
                    pd_read = (DgpApi_PageData*) ln->userdata;
                    if(pd_read) DGP_FREE ((void*) pd_read);
                    DgpAdt_liRemAndNext (ln, &ln);
                } 
            DgpAdt_liFree (silxState->pd_list);
        }
        if(silxState->api_handle) DgpApi_close (silxState->api_handle);
        /* close file */
        if(silxState->pCDoc){

                Dgp_closeJbig2Stream (silxState->pCDoc);

        }
        silxState->api_handle = NULL;
        silxState->pd_list= DGP_NULLLIST;
        silxState->pCDoc = NULL;

        DGP_FREE(silxState);
        *sCookie = NULL;
    }
    NORMAL_RETURN;

}
/**********************************************************************
 * Initialize SilxEncCookie. Has to be called.
 *
 * @param *dgp_state Pointer to Dgp_State
 * @param *cookie Pointer to SilxEncCookie
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */
DGP_LIBEXPORT ErrCode
silx_initEncState (
             Dgp_State              *dgp_state,
             SilxEncCookie          *cookie)
{

    DGP_PROLOGUE("silx_initEncState()", DGP_PERRSTATE);
    SilxEncState *silxState = NULL;
    DgpTdc_JpegCompInfo *jpegComp;		// Includes quality & scale factor.
    DgpApi_Options *opts=NULL;
    tokOptions *tokopt = NULL;
    LlTokOptions *llTokOpt = NULL;
    uds						*dump;

    /* allocate State struct */

    DGP_CHECK_ALLOC_CLEANUP (silxState = DGP_MALLOC (sizeof (*silxState)));

    tokopt = &silxState->tokOpt;
    llTokOpt = &silxState->llTokOpt;
    opts = &silxState->opts;
    jpegComp = &silxState->jpegComp;

    silxState->lossless = DGP_FALSE;
    silxState->jpegRes = DEF_JPEG_OUTPUT_RES;
    silxState->jpegQual = DEF_JPEG_QUALITY;
    silxState->incrementalDict = DEF_INCRDICT;
    silxState->clearDict = DEF_CLEARDICT; 
    silxState->redTokNoise = DEF_REDTOKNOISE;
    silxState->jb2Only = DEF_JB2ONLY;
    silxState->outputRes = DEF_OUTPUT_RES;
    silxState->pCDoc = NULL;
    silxState->api_handle = NULL;
    silxState->pd_list = DGP_NULLLIST;
    silxState->page_num = 0;
    /*set up DgpApi_Options*/
    opts->sizeThresh    =   DEF_SIZETHRESH;
    opts->classThresh   =   DEF_WRITETHRESH;
    opts->halftone      =   DEF_HALFTONEDETECT;
    opts->deskew        =   DEF_DESKEW;
    opts->forceRes		=   DEF_FORCE_RES;
    opts->baselineOrient=   DEF_BASELINEORIENT;
    opts->skewAccuracy  =   DEF_DESKEWACCURACY;
    opts->rotation      =   DEF_ROTATION;
    opts->jbig2Profile  =   DEF_JBIG2PROFILE;
    opts->g4_toks		=   DEF_G4TOKENS;
    opts->bitmapSizeLimit=  DEF_TOKENSIZELIMIT;  //limit token size be 10% of page width
    opts->verbose		=   DEF_VERBOSE;
    opts->membond       =   -1;             // default turn off mem upper limit
    opts->writeclasses  =   DGP_FALSE;
    opts->instrument    =   DGP_FALSE;
    
    
    /* Initialize parameters to default settings if caller hasn't supplied an alternative. */
    tokopt->dgp_state			 = dgp_state;
    tokopt->fracHausd			 = DEF_FRACHAUSD;
    tokopt->maxDense             = DEF_MAXDENSITY;
    tokopt->widthDiff			 = DEF_WIDTHDIFF;
    tokopt->heightDiff		     = DEF_HEIGHTDIFF;
    tokopt->commitThresh       = DEF_COMMITTHRESH;
    tokopt->freezeHistThresh   = DEF_FREEZEHISTTHRESH;
    tokopt->makeHistThresh     = DEF_MAKEHISTTHRESH;
    tokopt->returnHistThresh   = DEF_RETURNHISTTHRESH;
    tokopt->nneigh             = DEF_NEARESTNEIGHBORS;
    tokopt->xshift             = DEF_XSHIFT;
    tokopt->xslopechange       = DEF_XSLOPECHANGE;
    tokopt->newslope           = DEF_XNEWSLOPE;
    tokopt->dilationRadius     = DEF_DILATIONRADIUS;
    tokopt->lowerDilationLimit = DEF_LOWERDILATIONLIMIT;
    tokopt->flags              = TOK_OPT_ALIGN | TOK_OPT_RETURNOFFSET;
    tokopt->mergeCallback = tokopt->callbackCookie = NULL;
    
    llTokOpt->dgp_state		 = dgp_state;
    llTokOpt->flags			 = 0;
    llTokOpt->verbose			 = opts->verbose;
    llTokOpt->ascii			 = DGP_FALSE;
    llTokOpt->color			 = DGP_FALSE;
    llTokOpt->halftone		 = DGP_FALSE;
    llTokOpt->lossy			 = DGP_FALSE;
    llTokOpt->maxSamptokH		 = -1;
    llTokOpt->maxSamptokW		 = -1;
    llTokOpt->noascii			 = DGP_TRUE;
    llTokOpt->numStrips		 = 1;
    llTokOpt->sizeThresh		 = DEF_SIZETHRESH;
    llTokOpt->washThresh		 = 301;
    llTokOpt->tokenizerParams	 = NULL;
    llTokOpt->colorTokenizerParams = NULL;
    
    jpegComp->jpQuality		 = DEF_JPEG_QUALITY;
    jpegComp->scaleFactor		 = 1.0f;
    
    dump = &silxState->outUds;
    dump->fwrite = (fwrite_func)fwrite;
    dump->fread = (fread_func)fread;
    dump->fseek = (fseek_func)fseek;
    dump->ftell = (ftell_func)ftell;
    dump->ffeof = (feof_func)feof;

    // default supress error output
    *cookie= (SilxEncCookie)silxState;   
    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode
read_data (Dgp_State *dgp_state, FILE *fp, DgpApi_PageData *pd);

/***********************************************************************
 *   Encode one page at a time from buffer.
 *
 * @param *dgp_state    pointer to Dgp_State
 * @param  cookie       pointer to SilxEncState
 * @param  *image       input image in DgpImage format. 
 * @param  lastPage     flag to indicate if this is the last page
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */
DGP_LIBEXPORT ErrCode
silx_encPage(
             Dgp_State *dgp_state,
             SilxEncCookie cookie,
             void     *image,
             Dgp_Bool lastPage)
{
    DgpApi_Handle			handle = NULL;
    DgpApi_PageData			*pd = NULL, *pd_read;
    DgpAdt_List				pd_list = NULL;
    DgpAdt_ListNode			ln;
    DgpApi_Jbig2Data		dict, posBlock;
    PIXR                    *pixr;
    Dgp_Int32               bpp;
    
    Dgp_UInt32				sec_page_num;
    Dgp_Bool                clear = DGP_FALSE;
    
    /* color stuff */
    DgpImage_Gray           grayImage = NULL;
    
    /* TIFF FX writer */
    Dgp_CompressedDocument	*pCDoc = NULL;
    Dgp_CompressedPage		*pCPage = NULL;
    Dgp_CompressedImage		*pCImage = NULL;
    
    //memory limit
    Dgp_Bool overflow = DGP_FALSE;
    
    SilxEncState                *silxState = (SilxEncState *) cookie;
    DgpApi_Options	            *opts =NULL;
    
    DGP_PROLOGUE("silx_encPage()", DGP_PERRSTATE);
    DGP_ERR_GEN(!silxState, cErrInvalidParm);
    opts = &silxState->opts;
    
    handle = silxState->api_handle;
    pd_list = silxState->pd_list;
    pCDoc = silxState->pCDoc;
    DGP_ERR_GEN(!pCDoc, cErrInvalidParm);
    DGP_ERR_GEN(!handle, cErrInvalidParm);
    DGP_ERR_GEN(!pd_list, cErrInvalidParm);

fprintf(stderr, "encode page ..\n");

    /* loop over input files */
    DGP_SEC_MALLOC(pd, (DgpApi_PageData *), sizeof(*pd));

    //convert input image into PIXR
    DGP_ERR_PASS_CLEANUP(tc_convertToPixr(dgp_state, image, &pixr));

    // Get information from PIXR
    DGP_ERR_PASS_CLEANUP( w_pixrGetWidth(DGP_PIPWSTATE, pixr, &pd->width));
    DGP_ERR_PASS_CLEANUP( w_pixrGetHeight(DGP_PIPWSTATE, pixr,&pd->height));
    DGP_ERR_PASS_CLEANUP( w_pixrGetBpl(DGP_PIPWSTATE, pixr, &pd->bpl));
    DGP_ERR_PASS_CLEANUP( w_pixrGetResolutionDPI(DGP_PIPWSTATE, pixr,&pd->xRes, &pd->yRes));
    DGP_ERR_PASS_CLEANUP( w_pixrGetDepth(DGP_PIPWSTATE, pixr, &bpp));
    /* Check if we are dealing with binary image.  If not, abduct bpl. */
    if (bpp > 1){
	    pd->bpl = 0;
	    pd->image = NULL;
    }
    else{
        DGP_ERR_PASS_CLEANUP(w_pixrGetImageAll(DGP_PIPWSTATE, pixr, &(pd->image)));    
    }

    DGP_ERR_PASS_CLEANUP(Dgp_createNewPage (pCDoc, pd->width, pd->height, &pCPage));
    if (pd->bpl != 0) 
    {
        // monochrome image
        DGP_ERR_PASS_CLEANUP(DgpApi_addMonoPage (handle, pd, &overflow));
        
    } 
    
    // if dictionary memory within bound 
    if(overflow != DGP_TRUE){
        DGP_ERR_PASS_CLEANUP(Dgp_addNewPage (pCDoc, pCPage));
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pd_list, (void*) pd, NULL));
        silxState->page_num++;
        
    }else{
        DGP_ERR_PASS_CLEANUP(Dgp_freePage (pCDoc, &pCPage));
        DGP_FREE ((void*) pd);              
    }
    pd = NULL;
    if((silxState->page_num % silxState->incrementalDict) == 0 || 
        (silxState->page_num % silxState->clearDict) == 0  ||  
        overflow == DGP_TRUE ||
        lastPage == DGP_TRUE){
        // check if we can write out some data, only if we are not over dictionary limit 
        if ((silxState->page_num % silxState->clearDict) == 0 || overflow == DGP_TRUE || lastPage == DGP_TRUE) { 
            // clear the dictionary
            DGP_ERR_PASS_CLEANUP(DgpApi_endSection(handle, DGP_CLEAR_DICT));
            if(silxState->redTokNoise){
                DGP_ERR_PASS_CLEANUP(DgpApi_redTokNoise(handle));
            }
            overflow = DGP_FALSE;
            clear = DGP_TRUE;
        } else if ((silxState->page_num % silxState->incrementalDict) == 0) {
            if(silxState->redTokNoise){
                // clear the dictionary, don't do merge and reduce token noise
                DGP_ERR_PASS_CLEANUP(DgpApi_endSection(handle, DGP_CLEAR_DICT));
                DGP_ERR_PASS_CLEANUP(DgpApi_redTokNoise(handle));
            }else{
                // commit the classes, keep the instances 
                DGP_ERR_PASS_CLEANUP(
                    DgpApi_endSection(handle, DGP_INCR_DICT));
            }
            clear = DGP_FALSE;
        }
        DGP_ERR_PASS_CLEANUP(DgpApi_getDict(handle, clear, &dict));
        Dgp_storeCompTokDict (pCDoc, dict.data, dict.data_size);
        ln = pd_list->first;
        sec_page_num = 0;
        // loop over all elements of the list
        //each element represents one page that needs to be written out 
        while (ln != DGP_NULLLISTNODE){
            
            // get the page data 
            pd_read = (DgpApi_PageData*) ln->userdata;
            DGP_ERR_PASS_CLEANUP(DgpApi_getNextPosBlock (handle, &posBlock));
            Dgp_getPageOfSection (pCDoc, ++sec_page_num, &pCPage);
            DGP_CHECK_ALLOC (pCImage = DGP_MALLOC (sizeof(*pCImage)));
            
            pCImage->width = pd_read->width;
            pCImage->height = pd_read->height;
            pCImage->xres = pd_read->xRes;
            pCImage->yres = pd_read->yRes;
            pCImage->x = pCImage->y = 0;
            pCImage->layer = 2;
            pCImage->data = posBlock.data;
            pCImage->data_size = posBlock.data_size;
            pCImage->extData_size = posBlock.extData_size;
            Dgp_storeCompPosBlock (pCPage, pCImage);
            DGP_FREE ((void*) pd_read);
            DGP_ERR_PASS_CLEANUP(DgpAdt_liRemAndNext (ln, &ln));
        }

            Dgp_writeSectionJbig2Only (pCDoc, clear); // clear dict 

        // Start a section
        DGP_ERR_PASS_CLEANUP(DgpApi_startSection(handle));

    }    
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if(grayImage) DGPIMAGE_FREE(grayImage);
    if(pd) DGP_FREE(pd);
    ERR_RETURN;
    
}


/***********************************************************************
 *   Start Encoding...
 *
 * @param *dgp_state        pointer to Dgp_State
 * @param  cookie       pointer to SilxEncState
 * @param  *outfp       Output file stream pointer. Default is FILE. 
 *                      If you use other stream. You have to set uds in SilxEncState to handle that stream.
 * @param  *errfname    File name to error messages. Default is stdout. 
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */
DGP_LIBEXPORT ErrCode
silx_encStart(
    Dgp_State *dgp_state,
    SilxEncCookie cookie,
    void     *outfp,
    Dgp_Int8 *errfname)
{
    DgpApi_Handle			handle = NULL;
    DgpAdt_List				pd_list = NULL;

    /* TIFF FX writer */
    Dgp_CompressedDocument	*pCDoc = NULL;

    SilxEncState                *silxState = (SilxEncState *) cookie;
    DgpApi_Options	            *opts =NULL;
    tokOptions                  *tokopt = NULL;
    LlTokOptions                *llTokOpt = NULL;
    FILE                        *errfile = NULL;

    DGP_PROLOGUE("silx_encStart()", DGP_PERRSTATE);
    DGP_ERR_GEN(!silxState, cErrInvalidParm);

    fprintf (stderr, "entered routine \n");
    fprintf (stderr, "cookie : %x\n", cookie);
    fprintf (stderr, "silxState : %x\n", silxState);

    llTokOpt = &silxState->llTokOpt;

    fprintf (stderr, "1 \n");
    tokopt = &silxState->tokOpt;
    fprintf (stderr, "2 \n");
    opts = &silxState->opts;
    fprintf (stderr, "3 \n");
    fprintf (stderr, "4 \n");
    silxState->page_num =0; 
    fprintf (stderr, "5 \n");
    silxState->outUds.fp = outfp;
    fprintf (stderr, "6 \n");
    dgp_state->err_state->userCookie = stderr;

    fprintf (stderr, "open stream \n");
    if(silxState->pCDoc == NULL){
           DGP_ERR_PASS_CLEANUP(Dgp_createJbig2Stream (dgp_state, &(silxState->outUds), &pCDoc)); 
        silxState->pCDoc = pCDoc;
    }

    fprintf (stderr, "start list of images \n");
    if(silxState->pd_list == NULL){
        /* Init List of pages */
        DgpAdt_liNew (&pd_list, dgp_state);
        silxState->pd_list = pd_list;
    }

    fprintf (stderr, "open api handle \n");
    if(silxState->api_handle == NULL){
        DGP_ERR_PASS_CLEANUP(DgpApi_init (&handle, NULL, NULL, opts, tokopt,llTokOpt, NULL));
        silxState->api_handle = handle;
        // Start a section
        /* loop over multiple pages in mutiple files */
        DGP_ERR_PASS_CLEANUP(DgpApi_startSection(handle));
    }

    if(errfile && (errfile != stderr)){
        fclose(errfile);
    }
    NORMAL_RETURN;

ERR_CLEANUP:
    if(errfile && (errfile != stderr)){
        fclose(errfile);
    }

    ERR_RETURN;

}
/***********************************************************************
 * Encode document one file at a time
 *
 * @param *dgp_state    pointer to Dgp_State
 * @param  cookie       pointer to SilxEncState
 * @param  *inFile      Input file names.
 * @param  lastFile     Flag to indicate if this is the last file
 *
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */
DGP_LIBEXPORT ErrCode
silx_encDocument(
                 Dgp_State *dgp_state,
                 SilxEncCookie cookie,
                 Dgp_Int8  *inFile,
                 Dgp_Bool lastFile)
{
    DgpApi_Handle			handle = NULL;
    DgpApi_PageData			*pd = NULL, *pd_read;
    DgpAdt_List				pd_list = NULL;
    DgpAdt_ListNode			ln;
    DgpApi_Jbig2Data		dict, posBlock;
    
    Dgp_Int32				file_page = 0;
    Dgp_UInt32				sec_page_num;
    Dgp_Bool                clear = DGP_FALSE;
    
    /* color stuff */
    DgpImage_Gray           grayImage = NULL;
    
    /* TIFF FX writer */
    Dgp_CompressedDocument	*pCDoc = NULL;
    Dgp_CompressedPage		*pCPage = NULL;
    Dgp_CompressedImage		*pCImage = NULL;
    
    //memory limit
    Dgp_Bool overflow = DGP_FALSE;
    
    SilxEncState                *silxState = (SilxEncState *) cookie;
    DgpApi_Options	            *opts =NULL;
    Dgp_Int32 pageWritten;

    FILE *fp;
    ErrCode retcode;
    
    DGP_PROLOGUE("silx_encDocument()", DGP_PERRSTATE);
    DGP_ERR_GEN(!silxState, cErrInvalidParm);
    opts = &silxState->opts;
    pCDoc = silxState->pCDoc;
    handle = silxState->api_handle;
    pd_list = silxState->pd_list;
    DGP_ERR_GEN(!pCDoc, cErrInvalidParm);
    DGP_ERR_GEN(!handle, cErrInvalidParm);
    DGP_ERR_GEN(!pd_list, cErrInvalidParm);
    
    pageWritten = silxState->page_num;
    /* loop over input files */

    fp = fopen (inFile, "rb");
    DGP_ERR_GEN_CLEANUPF1((fp == (FILE *)NULL), cErrNullPtr,
        "can't open %s\n", inFile);

    //    do{

      DGP_SEC_MALLOC(pd, (DgpApi_PageData *), sizeof(*pd));

      retcode = read_data (dgp_state, fp, pd);
      DGP_ERR_GEN_CLEANUP( retcode != cSuccessful, cErrNullPtr);

      fprintf( stderr, "time: %f\n", utime());
                
      DGP_ERR_PASS_CLEANUP(Dgp_createNewPage (pCDoc, pd->width, pd->height, &pCPage));
      if (pd->bpl != 0) 
	{
	  // monochrome image
	    DGP_ERR_PASS_CLEANUP(DgpApi_addMonoPage (handle, pd, &overflow));
	  
	  DGP_FREE (pd->image);
	  pd->image = NULL;
	  
	} 

      // if dictionary memory within bound 
	if(overflow != DGP_TRUE){
	  DGP_ERR_PASS_CLEANUP(Dgp_addNewPage (pCDoc, pCPage));
	  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pd_list, (void*) pd, NULL));
	  file_page++;
	  silxState->page_num++;
	  
	}else{
	  DGP_ERR_PASS_CLEANUP(Dgp_freePage (pCDoc, &pCPage));
	  DGP_FREE ((void*) pd);              
	}

      pd = NULL;
    /*    } while((silxState->page_num % silxState->incrementalDict) != 0 && 
            (silxState->page_num % silxState->clearDict) != 0  &&
            overflow != DGP_TRUE &&
            file_page !=0); */
        
    // check if we can write out some data, only if we are not over dictionary limit 
      if (((silxState->page_num % silxState->clearDict) == 0 ||
	   overflow == DGP_TRUE || 
	   (lastFile == DGP_TRUE && file_page == 0) == DGP_TRUE || // lastPage of lastfile, we write out
	   (silxState->page_num % silxState->incrementalDict) ==0 ) &&
	  (silxState->page_num != pageWritten )  // this last condition make sure that there is something new to be written
	  ) {
	if ((silxState->page_num % silxState->clearDict) == 0 || 
	    overflow == DGP_TRUE || 
	    (lastFile == DGP_TRUE && file_page == 0)) { 
	  // clear the dictionary
	    DGP_ERR_PASS_CLEANUP(DgpApi_endSection(handle, DGP_CLEAR_DICT));
	  if(silxState->redTokNoise){
	    DGP_ERR_PASS_CLEANUP(DgpApi_redTokNoise(handle));
	  }
	  overflow = DGP_FALSE;
	  clear = DGP_TRUE;
	} else if ((silxState->page_num % silxState->incrementalDict) == 0) {
	  if(silxState->redTokNoise){
	    // clear the dictionary, don't do merge and reduce token noise
	      DGP_ERR_PASS_CLEANUP(DgpApi_endSection(handle, DGP_CLEAR_DICT));
	    DGP_ERR_PASS_CLEANUP(DgpApi_redTokNoise(handle));
	  }else{
	    // commit the classes, keep the instances 
	      DGP_ERR_PASS_CLEANUP(
				   DgpApi_endSection(handle, DGP_INCR_DICT));
	  }
	  clear = DGP_FALSE;
	}
	DGP_ERR_PASS_CLEANUP(DgpApi_getDict(handle, clear, &dict));
	Dgp_storeCompTokDict (pCDoc, dict.data, dict.data_size);
	ln = pd_list->first;
	sec_page_num = 0;
	// loop over all elements of the list
	//each element represents one page that needs to be written out 
        while (ln != DGP_NULLLISTNODE){
	  
	  // get the page data 
	    pd_read = (DgpApi_PageData*) ln->userdata;
	  DGP_ERR_PASS_CLEANUP(DgpApi_getNextPosBlock (handle, &posBlock));
	  Dgp_getPageOfSection (pCDoc, ++sec_page_num, &pCPage);
	  DGP_CHECK_ALLOC (pCImage = DGP_MALLOC (sizeof(*pCImage)));
	  
	  pCImage->width = pd_read->width;
	  pCImage->height = pd_read->height;
	  pCImage->xres = pd_read->xRes;
	  pCImage->yres = pd_read->yRes;
	  pCImage->x = pCImage->y = 0;
	  pCImage->layer = 2;
	  pCImage->data = posBlock.data;
	  pCImage->data_size = posBlock.data_size;
	  pCImage->extData_size = posBlock.extData_size;
	  Dgp_storeCompPosBlock (pCPage, pCImage);
	  DGP_FREE ((void*) pd_read);
	  DGP_ERR_PASS_CLEANUP(DgpAdt_liRemAndNext (ln, &ln));
	}

      fprintf( stderr, "time: %f\n", utime());
	  DGP_ERR_PASS_CLEANUP(Dgp_writeSectionJbig2Only (pCDoc, clear)); // clear dict 

          // Start a section
	  DGP_ERR_PASS_CLEANUP(DgpApi_startSection(handle));

	pageWritten = silxState->page_num;
      }// end if 
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if(grayImage) DGPIMAGE_FREE(grayImage);
    if(pd) DGP_FREE(pd);
    ERR_RETURN;
    
}

ErrCode
read_data (Dgp_State *dgp_state, FILE *fp, DgpApi_PageData *pd)
{
      
  unsigned char lbuf[4];
  size_t size, read, missing, new_data;
  DGP_PROLOGUE ("read_data", DGP_PERRSTATE);
  

  fprintf (stderr, "Entering read_data\n");
  if (fp == NULL || pd == NULL)
    {
      DGP_ERRF0 (200, "invalid parameters to read_data()\n");
    }
  
  /* read parameters width, hight, bytes_per_line, xRes, yRes */
  new_data = fread (lbuf, 1, 4, fp);
  if (new_data != 4)
    {
      DGP_ERRF0 (200, "No more data in file\n");
    }
  pd->width = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3];
  
  new_data = fread (lbuf, 1, 4, fp);
  if (new_data != 4)
    {
      DGP_ERRF0 (200, "Not sufficient data in file\n");
    }
  pd->height = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3];
  
  new_data = fread (lbuf, 1, 4, fp);
  if (new_data != 4)
    {
      DGP_ERRF0 (200, "Not sufficient data in file\n");
    }
  pd->bpl = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3];
  
  new_data = fread (lbuf, 1, 4, fp);
  if (new_data != 4)
    {
      DGP_ERRF0 (200, "Not sufficient data in file\n");
    }
  pd->xRes = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3];
  
  new_data = fread (lbuf, 1, 4, fp);
  if (new_data != 4)
    {
      DGP_ERRF0 (200, "Not sufficient data in file\n");
    }
  pd->yRes = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3];
  
  /* read large image data block */
  size = pd->bpl * (pd->height + 64);
  read = 0;
  DGP_CHECK_ALLOC (pd->image = (unsigned char*) DGP_MALLOC (size));
  while (read < size)
    {
      missing = size - read;
      new_data = fread (pd->image + read, 1, missing, fp);
      if (new_data == 0)
	{
	  DGP_FREE (pd->image);
	  pd->image = NULL;
	  DGP_ERRF0 (200, "Not sufficient data in file\n");
	}
      read += new_data;
    }

  pd->xRes = 300;
  pd->yRes = 300;
  fprintf (stderr, "%dx%d %d  %d %d\n", pd->width, pd->height, pd->bpl, 
	   pd->xRes, pd->yRes);

  NORMAL_RETURN;
}
