/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	asfloyd.c:	Code to do setup and make 1-bit dithering tables
 *			required by all (?) assembly implementations of the
 *			Floyd-Steinberg algorithm.
 *
 *	    General setup:
 *		Int32	ip_floydSetup()
 *		Int32	ip_floyd2Setup()
 *		Int32	ip_floyd4To1Setup()
 *
 *          Lookup tables:
 *              Int32	ip_makeErrorDiffusionTables() 
 *
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "imageref.pub"
#include "defines.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrip.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile $; $Revision $; $Date $");

/* If this is the PC assembly implementation, we do 1 line per
 * pass across the image.  If this is a SPARC, do 7 lines in each
 * sweep across the page.
 */
#if defined(_386_ARCH_)
#define cDBlockHeight 1
#elif defined(_SPARC_ARCH_)
#define cDBlockHeight 7
#endif

/* The error diffusion tables used by assembly code differs from that used
 * by C code in that the assembly code can take advantage of the carrr
 * flag to do double shifts.  That means the upper portions of the
 * horizontal error tables are filled with 1's in the assembly tables.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_getSizeOfDitherGrayBuf()
 *		Returns size of gray buffer used for Floyd-Steinberg dithering.
 *
 *	Input: 
 *		gBpl:			gray image's bytes/line
 *		w:			width of the image in pixels without
 *					frame.
 *		pGrayBufSize:		pointer to returned size.
 *
 *	Results:
 *		The size of a gray buffer is calculated.
 *
 *	Returns:
 *		ia_successful:		it worked
 **************************************************************************/
Int32  CDECL 
ip_getSizeOfDitherGrayBuf (
    IPState    *pIPState,
    UInt32	gBpl,
    Int32	w,
    Int32      *pGrayBufSize)
{
PROLOGUE;
 UInt32      	 cyclesOnRight;		/* number of extra cycles executed
					 * on right side of image to flush
					 * pipeline. */

	/* Calculate size of the grayBuffer.
	 * We ask for a buffer of size gBpl + 4 so that if we're doing
	 * a bunch of dithers from rectangles of the same gray image, we'll
	 * probably keep getting the same buffer (which ought to avoid
	 * memory fragmentation, right?).
	 * The weird cyclesOnRight term is mostly for documentation.
	 * As explained in other dither documenation, this is the
	 * number of extra bytes we want to stuff into GrayBuf.
	 * The extra cDBlockHeight bytes are needed because GrayBuf
	 * is written with trash from beyond the *left* edge of the
	 * rectangle by ip_ditherFloydLine while the pipeline is
	 * filling.  Since the leftmost byte of the
	 * bottom line is cDBlockHeight bytes to the left of the
	 * rectangle, there are an extra cDBlockHeight bytes in GrayBuf.
	 * Since we're using gBpl as the basic width of pGrayBuf instead of
	 * w (and gBpl is likely to be big), these funny little extra terms
	 * shouldn't make any difference.
	 */
	/* Why is cyclesOnRight calculated this way? Wellll,  
	 * we want to make sure that the bottom line has its last error
	 * value calculated correctly.  To do this we need cDBlockHeight 
	 * extra pixels plus some adjustment for 32 bit alignment
	 */
    cyclesOnRight = cDBlockHeight + ((32 - (w % 32)) & 0x1F);

    *pGrayBufSize = gBpl + cyclesOnRight + cDBlockHeight + 4;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_floydSetup()
 *		Does the implementation-dependent setup required before
 *		starting to dither an image.
 *
 *	Input: 
 *		pVDitherErrorTable:	pointer to vertical error table
 *		pHDitherErrorTable:	pointer to horizontal error table
 *		pDitherTruncateTab:	pointer to pixel truncation table
 *		pDBlockHeight:		pointer to word that will contain
 *					the number of image lines that will
 *					be dithered in one pass of
 *					ip_ditherFloydLine.
 *		pG:			pointer to the gray image.
 *		pGrayBuf:		ptr to buffer used to hold
 *					error data for a single gray scan line.
 *		pDRCTable:		Ptr to DRC table used to convert
 *					raw gray image data before dithering.
 *		w:			width of the image in pixels without
 *					frame.
 *
 *	Results:
 *		pVDitherErrorTable, pHDitherErrorTable and pDitherTruncateTab
 *		are copied to known global positions if needed (PC assembly
 *		only).
 *		The block height for the current implementation is stored
 *		in pDBlockHeight.
 *		A pre-allocated gray buffer is initialized with data from
 *		the gray image as modified by the DRC table.  The size of
 *		the gray buffer must be at least as large as the number
 *		returned by ip_getSizeOfDitherGrayBuf.
 *
 *	Returns:
 *		ia_successful:		it worked
 **************************************************************************/
Int32  CDECL 
ip_floydSetup (
    IPState    *pIPState,
    UInt32     *pHDitherErrorTable,
    UInt32     *pVDitherErrorTable,
    UInt8      *pDitherTruncateTab,
    Int32      *pDBlockHeight,
    UInt8      *pG,
    UInt8      *pGrayBuf,
    UInt8      *pDRCTable,
    Int32	w)
{
    PROLOGUE;
    Int32     	 cyclesOnRight;		/* number of extra cycles executed
					 * on right side of image to flush
					 * pipeline. */
    UInt8        *pAGrayBuf;		/* pointer inside buffer that
					   holds modified pixels from
					   previous passes. pAGrayBuf is
					   guranteed to be aligned on a word
					   boundary. */
    Int32     	 byte;			/* interation var */

/* if this is the setup for the PC assembly language, we need these tables */
#ifdef _386_ARCH_
extern Int32	 vdErrorTable;
extern Int32	 hdErrorTable;
extern Int8	 dTruncTable;

	/* copy the error diffusion and truncate tables to a place
	 * that the pc knows about so that it can use a constant
	 * offset to access the tables instead of needing to keep
	 * the address of the tables in registers, which the pc
	 * has very few of.
	 * Since these tables are always the same, we don't worry
	 * about overwriting them with new stuff.
	 */
    memcpy(&vdErrorTable, pVDitherErrorTable, cDitherVErrorTabSize);
    memcpy(&hdErrorTable, pHDitherErrorTable, cDitherHErrorTabSize);
    memcpy(&dTruncTable, pDitherTruncateTab, cDitherTruncateTabSize);
#else
	/* keep the compiler from complaining */
    pVDitherErrorTable = pVDitherErrorTable;
    pHDitherErrorTable = pHDitherErrorTable;
    pDitherTruncateTab = pDitherTruncateTab;
#endif

	/* Set the number of lines dithered per horizontal pass */
    *pDBlockHeight = cDBlockHeight;

	/* Fill the grayBuffer.  Make pAGrayBuf point to first
	 * word boundary in GrayBuf.  We'll use pAGrayBuf for efficiency.
	 */
	/* Why is cyclesOnRight calculated this way? Wellll,  
	 * we want to make sure that the bottom line has its last error
	 * value calculated correctly.  To do this we need cDBlockHeight 
	 * extra pixels plus some adjustment for 32 bit alignment
	 */
    cyclesOnRight = cDBlockHeight + ((32 - (w % 32)) & 0x1F);

    pAGrayBuf = (UInt8 *) ((UInt32)(pGrayBuf + 3) & ~3);

	/* copy the topmost line of pixels into GrayBuf to get things started.
	 * We'll mimic ip_ditherFloydLine, which puts trash into the first
	 * cDBlockHeight bytes of GrayBuf.
	 * The real code also reads up to index w+cyclesOnRight+cDBlockHeight
	 * + 3, so we should initialize that much stuff.  The bytes read
	 * on the far right end are never used, but this keep Purify happy.
	 */
    for (byte = 0; byte < w+cyclesOnRight+cDBlockHeight+3; byte++)
    {
	/* usually the access to pAGrayBuf would use the macro
	 * UCHAR_ACCESS. However, the pc assembly code only reads
	 * and writes a byte at a time so it is ok not to use 
	 * UCHAR_ACCESS.
	 */
	pAGrayBuf[byte+cDBlockHeight] = 
	  pDRCTable[UCHAR_ACCESS(&pG[byte])];
    }

        /* Make the first cDBlockHeight pixels zeros */
    for (byte = 0; byte < cDBlockHeight; byte++) 
   	pAGrayBuf[byte] = 0;

    NORMAL_RETURN;
}


/**************************************************************************
 *  ip_getSizeOfDither2GrayBuf()
 *		Returns size of gray buffer used for Floyd-Steinberg dithering.
 *
 *	Input: 
 *		gBpl:			gray image's bytes/line
 *		w:			width of the image in pixels without
 *					frame.
 *		pGrayBufSize:		pointer to returned size.
 *
 *	Results:
 *		The size of a gray buffer is calculated.
 *
 *	Returns:
 *		ia_successful:		it worked
 **************************************************************************/
Int32  CDECL 
ip_getSizeOfDither2GrayBuf (
    IPState    *pIPState,
    Int32	gBpl,
    Int32	w,
    Int32      *pGrayBufSize)
{
    PROLOGUE;
    Int32      	cyclesOnRight;		/* number of extra cycles executed
					 * on right side of image to flush
					 * pipeline. */

	/* Calculate size of the grayBuffer.
	 * We ask for a buffer of size gBpl + 4 so that if we're doing
	 * a bunch of dithers from rectangles of the same gray image, we'll
	 * probably keep getting the same buffer (which ought to avoid
	 * memory fragmentation, right?).
	 * The weird cyclesOnRight term is mostly for documentation.
	 * As explained above, this is the number of extra bytes we want
	 * to stuff into GrayBuf.
	 * The extra cDBlockHeight bytes are needed because GrayBuf
	 * is written with trash from beyond the *left* edge of the
	 * rectangle by ip_ditherFloydLine.  Since the leftmost byte of the
	 * bottom line is cDBlockHeight bytes to the left of the
	 * rectangle, there are an extra cDBlockHeight bytes in GrayBuf.
	 * Since we're using gBpl as the basic width of pGrayBuf instead of
	 * w (and gBpl is likely to be big), these funny little extra terms
	 * shouldn't make any difference.
	 */
	/* Why is cyclesOnRight calculated this way? Wellll,  
	 * we want to make sure that the bottom line has its last error
	 * value calculated correctly.  To do this we need cDBlockHeight 
	 * extra pixels plus some adjustment for 32 bit alignment
	 */
    cyclesOnRight = cDBlockHeight + ((16 - (w % 16)) & 0xF);

    *pGrayBufSize = gBpl + cyclesOnRight + cDBlockHeight + 4;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_floyd2Setup()
 *		Does the implementation-dependent setup required before
 *		starting to dither an image.
 *
 *	Input: 
 *		pVDitherErrorTable2:	pointer to vertical error table
 *		pHDitherErrorTable2:	pointer to horizontal error table
 *		pDitherTruncateTab2:	pointer to pixel truncation table
 *		pDBlockHeight:		pointer to word that will contain
 *					the number of image lines that will
 *					be dithered in a single pass of
 *					ip_ditherFloydLine.
 *		pG:			pointer to the gray image.
 *		pGrayBuf:		ptr to buffer used to hold
 *					error data for a single gray scan line.
 *		pDRCTable:		Ptr to DRC table used to convert
 *					raw gray image data before dithering.
 *		w:			width of the image in pixels without
 *					frame.
 *
 *	Results:
 *		pVDitherErrorTable, pHDitherErrorTable and pDitherTruncateTab
 *		are copied to known global positions if needed (PC assembly
 *		only).
 *		The block height for the current implementation is stored
 *		in pDBlockHeight.
 *		A gray buffer is initialized with data from
 *		the gray image as modified by the DRC table.  The gray
 *		buffer must be at least the size indicated by
 *		ip_getSizeOfDither2GrayBuf.
 *
 *	Returns:
 *		ia_successful:		it worked
 **************************************************************************/
Int32  CDECL 
ip_floyd2Setup (
    IPState    *pIPState,
    UInt32     *pHDitherErrorTable2,
    UInt32     *pVDitherErrorTable2,
    UInt8      *pDitherTruncateTab,
    Int32      *pDBlockHeight,
    UInt8      *pG,
    UInt8      *pGrayBuf,
    UInt8      *pDRCTable,
    Int32	w)
{
    PROLOGUE;
    Int32      	 cyclesOnRight;		/* number of extra cycles executed
					 * on right side of image to flush
					 * pipeline. */
    UInt8        *pAGrayBuf;		/* pointer to buffer that
					   holds modified pixels from
					   previous passes. pAGrayBuf is
					   guranteed to be aligned on a word
					   boundary. */
    Int32      	 byte;			/* interation var */

/* if this is the setup for the PC assembly language, we need these tables */
#ifdef _386_ARCH_
    extern Int32 vdErrorTable2;
    extern Int32 hdErrorTable2;
    extern Int8	 dTruncTable;

	/* copy the error diffusion and truncate tables to a place
	 * that the pc knows about so that it can use a constant
	 * offset to access the tables instead of needing to keep
	 * the address of the tables in registers, which the pc
	 * has very few of.
	 * Since these tables are always the same, we don't worry
	 * about overwriting them with new stuff.
	 */
    memcpy(&vdErrorTable2, pVDitherErrorTable2, cDitherVErrorTabSize);
    memcpy(&hdErrorTable2, pHDitherErrorTable2, cDitherHErrorTabSize);
    memcpy(&dTruncTable, pDitherTruncateTab, cDitherTruncateTabSize);
#else
	/* keep the compiler from complaining */
    pVDitherErrorTable2 = pVDitherErrorTable2;
    pHDitherErrorTable2 = pHDitherErrorTable2;
    pDitherTruncateTab = pDitherTruncateTab;
#endif

	/* Set the number of lines dithered per horizontal pass */
    *pDBlockHeight = cDBlockHeight;

	/* allocate the grayBuffer.  Make pAGrayBuf point to first
	 * word boundary in GrayBuf.  We'll use pAGrayBuf for efficiency.
	 * We ask for a buffer of size gBpl + 4 so that if we're doing
	 * a bunch of dithers from rectangles of the same gray image, we'll
	 * probably keep getting the same buffer (which ought to avoid
	 * memory fragmentation, right?).
	 * The weird cyclesOnRight term is mostly for documentation.
	 * As explained above, this is the number of extra bytes we want
	 * to stuff into GrayBuf.
	 * The extra cDBlockHeight bytes are needed because GrayBuf
	 * is written with trash from beyond the *left* edge of the
	 * rectangle by ip_ditherFloydLine.  Since the leftmost byte of the
	 * bottom line is cDBlockHeight bytes to the left of the
	 * rectangle, there are an extra cDBlockHeight bytes in GrayBuf.
	 * Since we're using gBpl as the basic width of pGrayBuf instead of
	 * w (and gBpl is likely to be big), these funny little extra terms
	 * shouldn't make any difference.
	 */
	/* Why is cyclesOnRight calculated this way? Wellll,  
	 * we want to make sure that the bottom line has its last error
	 * value calculated correctly.  To do this we need cDBlockHeight 
	 * extra pixels plus some adjustment for 32 bit alignment
	 */
    cyclesOnRight = cDBlockHeight + ((16 - (w % 16)) & 0xF);

    pAGrayBuf = (UInt8 *) ((UInt32)(pGrayBuf + 3) & ~3);

	/* copy the topmost line of pixels into GrayBuf to get things started.
	 * We'll mimic ip_ditherFloydLine2, which puts trash into the first
	 * cDBlockHeight bytes of GrayBuf.
	 * The real code also reads up to index w+cyclesOnRight+
	 * cDBlockHeight + 3, so we should initialize that
	 * much stuff.  The bytes read on the far right end are never used,
	 * but this keep Purify happy.
	 */
    for (byte = 0; byte < w+cyclesOnRight+cDBlockHeight+3; byte++)
    {
	/* usually the access to pAGrayBuf would use the macro
	 * UCHAR_ACCESS. However, the pc assembly code only reads
	 * and writes a byte at a time so it is ok not to use 
	 * UCHAR_ACCESS.
	 */
	pAGrayBuf[byte+cDBlockHeight] = 
	  pDRCTable[UCHAR_ACCESS(&pG[byte])];
    }

        /* Make the first cDBlockHeight pixels zeros */
    for (byte = 0; byte < cDBlockHeight; byte++) 
        pAGrayBuf[byte] = 0;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_getSizeOfDither4to1GrayBuf()
 *		Returns size of gray buffer used for Floyd-Steinberg dithering.
 *
 *	Input: 
 *		gBpl:			gray image's bytes/line
 *		w:			width of the image in pixels without
 *					frame.
 *		pGrayBufSize:		pointer to returned size.
 *
 *	Results:
 *		The size of a gray buffer is calculated.
 *
 *	Returns:
 *		ia_successful:		it worked
 **************************************************************************/
Int32  CDECL 
ip_getSizeOfDither4to1GrayBuf (
    IPState    *pIPState,
    Int32	gBpl,
    Int32	w,
    Int32      *pGrayBufSize)
{
    PROLOGUE;
    Int32      	 cyclesOnRight;		/* number of extra cycles executed
					 * on right side of image to flush
					 * pipeline. */

	/* Calculate size of the grayBuffer.
	 * What is the size of the gray buffer?  We need enough for every
	 * pixel in the image; hence the gBpl*2 term.  There are 
	 * approximately gBpl*2 pixels (2 pixels per byte) in the image.
	 * Each 2 pixels require 2 bytes so we need gBpl*2 bytes.
	 * The extra cDBlockHeight bytes are needed because GrayBuf is
	 * written with trash from beyond the *left* edge of the rectangle
	 * by ip_ditherFloydLine.
	 * Since the leftmost byte of the bottom line is cDBlockHeight
	 * bytes to the left of the rectangle, there are an extra
	 * cDBlockHeight bytes in GrayBuf.
	 * The extra cyclesOnRight term is needed to for the right
	 * side finish.
	 * The extra 10 bytes are pad bytes.  The are used to write
	 * gray data for the part of the image just over the right side
	 * of the image and to allow for alignment of the buffer on a 4
	 * byte boundary.
	 */
	/* Why is cyclesOnRight calculated this way? Wellll,  
	 * we want to make sure that the bottom line has its last error
	 * value calculated correctly.  To do this we need cDBlockHeight
	 * extra pixels plus some adjustment for 32 bit alignment
	 */
    cyclesOnRight = cDBlockHeight + ((32 - (w % 32)) & 0x1F);

    *pGrayBufSize = gBpl*2 + cyclesOnRight + cDBlockHeight + 10;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_floyd4To1Setup()
 *		Does the implementation-dependent setup required before
 *		starting to dither an image.
 *
 *	Input: 
 *		pVDitherErrorTable:	pointer to vertical error table
 *		pHDitherErrorTable:	pointer to horizontal error table
 *		pDitherTruncateTab:	pointer to pixel truncation table
 *		pDBlockHeight:		pointer to word that will contain
 *					the number of image lines that will
 *					be dithered in a single pass of
 *					ip_ditherFloydLine.
 *		pG:			pointer to the gray image.
 *		pGrayBuf:		ptr to buffer used to hold
 *					error data for a single gray scan line.
 *		pHDRCTable:		Ptr to DRC table used to convert
 *					high nibble of raw gray image data
 *					before dithering.
 *		pLDRCTable:		Ptr to DRC table used to convert
 *					low nibble of raw gray image data
 *					before dithering.
 *		w:			width of the image in pixels without
 *					frame.
 *
 *	Results:
 *		pVDitherErrorTable, pHDitherErrorTable and pDitherTruncateTab
 *		are copied to known global positions if needed (PC assembly
 *		only).
 *		The block height for the current implementation is stored
 *		in pDBlockHeight.
 *		A gray buffer is initialized with data from
 *		the gray image as modified by the DRC table.  The gray
 *		buffer must be at least the size indicated by
 *		ip_getSizeOfDither2GrayBuf.
 *
 *	Returns:
 *		ia_successful:		it worked
 **************************************************************************/
Int32  CDECL 
ip_floyd4To1Setup (
    IPState    *pIPState,
    UInt32     *pHDitherErrorTable,
    UInt32     *pVDitherErrorTable,
    UInt8      *pDitherTruncateTab,
    Int32      *pDBlockHeight,
    UInt8      *pG,
    UInt8      *pGrayBuf,
    UInt8      *pHDRCTable,
    UInt8      *pLDRCTable,
    Int32	w)
{
    PROLOGUE;
    Int32	cyclesOnRight;		/* number of extra cycles executed
					 * on right side of image to flush
					 * pipeline. */
    UInt8      *pAGrayBuf;		/* pointer to buffer that
					   holds modified pixels from
					   previous passes. pAGrayBuf is
					   guranteed to be aligned on a word
					   boundary. */
    Int32	ibyte, obyte;		/* interation vars */


/* if this is the setup for the PC assembly language, we need these tables */
#ifdef _386_ARCH_
    extern Int32 vdErrorTable;
    extern Int32 hdErrorTable;
    extern Int8	 dTruncTable;

	/* copy the error diffusion and truncate tables to a place
	 * that the pc knows about so that it can use a constant
	 * offset to access the tables instead of needing to keep
	 * the address of the tables in registers, which the pc
	 * has very few of.
	 * Since these tables are always the same, we don't worry
	 * about overwriting them with new stuff.
	 */
    memcpy(&vdErrorTable, pVDitherErrorTable, cDitherVErrorTabSize);
    memcpy(&hdErrorTable, pHDitherErrorTable, cDitherHErrorTabSize);
    memcpy(&dTruncTable, pDitherTruncateTab, cDitherTruncateTabSize);
#else
	/* keep the compiler from complaining */
    pVDitherErrorTable = pVDitherErrorTable;
    pHDitherErrorTable = pHDitherErrorTable;
    pDitherTruncateTab = pDitherTruncateTab;
#endif

	/* Set the number of lines dithered per horizontal pass */
    *pDBlockHeight = cDBlockHeight;

	/* Fill in the grayBuffer.  Make pAGrayBuf point to first
	 * word boundary in GrayBuf.  We'll use pAGrayBuf for efficiency.
	 */
	/* Why is cyclesOnRight calculated this way? Wellll,  
	 * we want to make sure that the bottom line has its last error
	 * value calculated correctly.  To do this we need cDBlockHeight 
	 * extra pixels plus some adjustment for 32 bit alignment
	 */
    cyclesOnRight = cDBlockHeight + ((32 - (w % 32)) & 0x1F);

    pAGrayBuf = (UInt8 *) ((UInt32)(pGrayBuf + 3) & ~3);

	/* copy the topmost line of pixels into GrayBuf to get things started.
	 * We'll mimic ip_ditherFloydLine, which puts trash into the first
	 * cDBlockHeight bytes of GrayBuf.
	 * The real code also reads up to index w+cyclesOnRight+
	 * cDBlockHeight + 3, so we should initialize that
	 * much stuff.  The bytes read on the far right end are never used,
	 * but this keep Purify happy.
	 */
    for (ibyte = 0, obyte = cDBlockHeight;
	 obyte <= w+cyclesOnRight + cDBlockHeight+3;
	 ibyte++, obyte += 2)
    {
	    /* Usually we would use UCHAR_ACCESS to write into the
	     * gray buffer but since the pc reads in byte order,
	     * we don't use it.
	     */
	pAGrayBuf[obyte]   = pHDRCTable[UCHAR_ACCESS(&pG[ibyte])];
	pAGrayBuf[obyte+1] = pLDRCTable[UCHAR_ACCESS(&pG[ibyte])];
    }

        /* Make the first cDBlockHeight pixels zeros */
    for (obyte = 0; obyte < cDBlockHeight; obyte++) 
    	pAGrayBuf[obyte] = 0;

    NORMAL_RETURN;
}



/* PRIVATE */
/**************************************************************************
 *  ip_makeErrorDiffusionTables()
 *		Fills in hDitherErrorTable (7/16 of error) and
 *		vDitherErrorTable ({3/16 | 5/16 | 1/16} of error).
 *		Fills in pDitherTruncateTab which truncates pixels to
 *		8 bits.
 *		All tables are allocated and freed by caller.
 *		Used when dithering to 1 bit/pixel.
 *
 *	Input:
 *		pVDitherErrorTable:	pointer to vertical error table
 *		pSingHDitherErrorTable:	pointer to single line dithering
 *					horizontal error table
 *					(Pass in NULL if SingleLine dithering
 *					 will not be used)
 *		pHDitherErrorTable:	pointer to horizontal error table
 *		pDitherTruncateTab:	pointer to pixel truncation table
 **************************************************************************/
Int32  CDECL 
ip_makeErrorDiffusionTables (
    IPState    *pIPState,
    UInt32     *pHDitherErrorTable,
    UInt32     *pSingHDitherErrorTable,
    UInt32     *pVDitherErrorTable,
    UInt8      *pDitherTruncateTab)
{
    PROLOGUE;
    Int32	 pix, error;		/* pixel and error val */

	/* check the tables */
    if ((pVDitherErrorTable == NULL) || (pHDitherErrorTable == NULL) ||
	(pDitherTruncateTab == NULL))
	return ia_invalidParm;

	/* clear all three tables */
    memset (pVDitherErrorTable, 0, cDitherVErrorTabSize);
    memset (pHDitherErrorTable, 0, cDitherHErrorTabSize);
    memset (pDitherTruncateTab, 0, cDitherTruncateTabSize);

        /* If the SingleLine functions are to get called, we notice the
	 * horizontal dithering error tables are different depending on whether
	 * we run C or assembly.  We are using assembly for the regular block
	 * dithering calls, so we need the C version of the horizontal error
	 * for the C coded SingleLine dithering prmitives!
	 */
    if (pSingHDitherErrorTable != NULL)
    {
	memset (pSingHDitherErrorTable, 0, cDitherHErrorTabSize);

	    /* Fill the SingleLine dithering horizontal error table just as
	     * done in (commented) mifloyd.c:ip_makeErrorDiffusiontables.
	     */
	for (pix = 0; pix < 128; pix++)
	    pSingHDitherErrorTable[pix] =  ((7*pix)/16) << 22;

       	for (pix = 128; pix < 256; pix++) 
	{
	    error = pix - 255;
	    pSingHDitherErrorTable[pix] = ((((7*error)/16) & 0x3FF) << 22);
	}

	    /* See the long commented note below concerning completing the
	     * horizontal and vertical error tables in this manner.
	     */
    }
    
	/* Fill the tables.  For each pixel value, we compute the error,
	 * then the fraction of the error that goes into each pixel.
	 */

	/* first, fill in all the terms for which errors are positive. */

    for (pix = 0; pix < 128; pix++)
    {
	error = pix;
	pHDitherErrorTable[pix] =  ((7*error)/16) << 11;
	pVDitherErrorTable[pix] = (((3*error)/16) << 22) |
				  (((5*error)/16) << 11) |
				   ((1*error)/16);
    }

	/* now the area in which the terms for which errors are negative.
	 * We allocate 10 bits for each pixel, leaving a buffer bit
	 * between them.  The buffer bit may occasionally allow a carry
	 * from one pixel to the next, but we'll assume that's ok. */
    for (pix = 128; pix < 256; pix++)
    {
	error = pix - 255;
	pHDitherErrorTable[pix] = ((((7*error)/16) & 0x3FF) << 11) |
				  (0x3FFUL << 22);
	pVDitherErrorTable[pix] = ((((3*error)/16) & 0x3FF) << 22) |
				  ((((5*error)/16) & 0x3FF) << 11) |
				   (((1*error)/16) & 0x3FF);
    }

    	/* Note that there is a slight problem with the completed horizontal
	 * and vertical error tables as such.
	 * When the error is small (or when, precisely ABS(error*7/16) < 0!),
	 * there will be no propogated horizontal or vertical error
	 * (pHDItherErrorTable[] = pVDitherErrorTable[] = 0).
	 * This means that essentially we have "banding" at these values,
	 * that when dithering, they can all look like the same value as
	 * they propogate the same error.
	 *
	 * In the above case, this happens when the error is 0, 1, or 2, for
	 * index values 0,1,2 and when the error is 0, -1, or -2, for index
	 * values 253,254,255.  Thus, white(255) and nearly white(254,253)
	 * gray values will propogate exactly the same error values.
	 *
	 * Is this a problem?  Well, that depends.  If you feed us an image
	 * like this:
	 *  255 255 255 .... (repeat for whole row)
	 *  255 255 255 ....
	 *  255 255 255 ....
	 *  254 254 254 ....
	 *  254 254 254 ....
	 *  254 254 254 ....
	 *  253 253 253 ....
	 *  253 253 253 ....
	 *  253 253 253 ....
	 *  252 252 252 ....
	 * then you will get back a completely white image for all rows above
	 * the first 252 row.
	 *
	 * Can we fix it?  Well, the only way would be to push the "banding"
	 * area in terms of values [0..255] further one way, but there would
	 * need to be one somewhere due to the nature of us not counting
	 * error all the way down to 1/16.
	 */

	/* Values of pix between 256 and 512 represent big
	 * positive numbers.  These will be digitized as 0's (white in
	 * binary).  Since they are bigger than 255, their error values
	 * should be 0.  We don't want to make surrounding pixels even
	 * more positive (whiter) than they already are just because the
	 * the printer can't print whiter than white.  Therefore, we just
	 * leave the array alone; calloc already filled it with 0's.
	 * Similarly, all values of pix between 512 and 1024 represent
	 * negative numbers.  These will be printed as the blackest
	 * 1's the printer can produce.  Surrounding pixels should not be
	 * made more negative (blacker), so their error values are left at
	 * 0 also.
	 * HOWEVER, the carry flag should be set when any horizontal (7/16)
 	 * error is applied corresponding to a pixel between 256 and 512.
	 * Even so, we get away with filling in only a quarter of
	 * the vertical diffusion table and half of the horizontal table.
	 */
    for (pix = 256; pix < 512; pix++)
    {
	pHDitherErrorTable[pix] = (0x3FFUL << 22);
    }


	/* Make the dither truncate table.  This is used to turn 10 bit
	 * pixel values into 8 bit pixel values so we can pack them into
	 * the GrayBuf.  The 10 bit values can be positive or negative.
	 * We truncate them back to the range [0..255].
	 * Start with the pixels that are in the right range.
	 */
    for (pix = 0; pix < 256; pix++)
    {
	pDitherTruncateTab[pix] = (UInt8)pix;
    }

	/* Truncate excessively positive pixels */
    for (pix = 256; pix < 512; pix++)
	pDitherTruncateTab[pix] = 255;

	/* The rest of the pixels represent negative values, so leave
	 * them set to 0.
	 */
    NORMAL_RETURN;

}



/*

 $Log: asfloyd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:09  jlin
 Imported sources

 Revision 1.13  1998/01/13 23:46:39  ddavies
 ddavies on Tue Jan 13 15:46:39 PST 1998

 asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.12  1997/04/23 21:29:40  ddavies
 asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v

 Lots of changes to do with making DitherState contain all the tables and
 buffers directly instead of having pointers to them.  Also generally
 pre-allocate tables and buffers, letting lower level routines fill them
 in instead of having the lower level routines allocate them.
 ip_getSizeOfDitherGrayBuf and ip_getSizeOfDither4to1GrayBuf are new.
 These are used by the client to get the size of the gray buffer used to
 hold error values between passes.
 Ditto for ip_getSizeOfDither2GrayBuf.
 ip_makeErrorDiffusionTables now takes pointers to pre-allocated tables
 instead of allocating the tables itself.  This makes the allocation and
 freeing of tables more consistent since it all happens at the same level.

 Revision 1.11  1995/12/13 02:41:09  ejaquith
 ejaquith on Tue Dec 12 18:41:09 PST 1995

 asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.10  1995/10/13  23:48:43  dferrell
 * dferrell on Fri Oct 13 16:48:06 PDT 1995
 *
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.9  1994/10/27  23:07:12  ejaquith
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * Fixed Uninited Memory Read that Purify sniffed out when creating the
 * dithering gray buffer.
 *
 * Revision 1.8  1994/10/21  18:09:02  ejaquith
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * Added comments concerning construction of horizontal and vertical error
 * tables in reference to uniformly very white or very black regions.
 * We won't be distribuing small errors manually, as was toyed with below.
 *
 * Revision 1.7  1994/10/18  01:42:28  ejaquith
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * took out previous table small error fix (temporarily)
 *
 * Revision 1.6  1994/10/18  00:36:29  ejaquith
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * Fixed a "bug" that occurs with uniformly very white or very black images.
 * The horizontal error tables at certain values were allocating a small
 * error of magnitude less than 2 to none of its neighbors as integer math
 * was truncating [0..1]*7/16 to be 0.
 * Manually distribute small errors now.
 *
 * Revision 1.5  1994/08/25  17:15:24  ejaquith
 * ejaquith on Thu Aug 25 10:11:45 PDT 1994
 *
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * Made the FloydSetup routines clear the first cDBlockHeight bytes of GrayBuf.
 * Made ip_makeErrorDiffusionTables take another parameter,
 * pSingHDitherErrorTable, because the FloydSingleLine routines need both
 * the C and assembly versions of the horizontal error table.
 *
 * Revision 1.4  1994/08/10  21:46:42  ejaquith
 * ejaquith on Wed Aug 10 14:46:23 PDT 1994
 *
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * Removed unused parameter to some private functions.
 *
 * Revision 1.3  1994/07/27  15:39:02  ejaquith
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Changed table creation function to modify passed in table pointers.
 *
 * Revision 1.2  1994/02/04  20:12:10  dferrell
 * dferrell on Fri Feb 4 12:11:25 PST 1994
 *
 * asfloyd.c -> /product/ipcore/ipshared/src/RCS/asfloyd.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/12  01:13:17  ddavies
 * Initial revision
 *

 */


