/*************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  asglyph.c:  contains C stubs of convolution routines and a wrapper for
 *		code that generates machine code.  This is code common
 *		to all assembly implementations.
 *
 *          Contains stub implementations of subroutines
 *
 *                Int32         ip_convolve1x1()
 *                Int32         ip_convolve3x3()
 *
 *	    Contains SPARC implementations of
 *		  UInt32	ip_genConvAsm()
 *
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"

    /*  prototypes */
#include "shrip.pub"
#include "shrglyph.prv"		/* If we were using any of the public
				 * pointers to private structs, this would
				 * have to go first since it's where the
				 * private structures are defined.  Today,
				 * we're not using any, but you never know. */
#include "shrglyph.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: asglyph.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:09 $");
/* SET THIS cTrue TO SEE A CODE LISTING */
#ifdef DEBUG
#undef DEBUG
#endif /* DEBUG */
/* #define DEBUG	0 */

/* Most of the comments for the glyph functions are in cglyph.c.  This file
 * contains stubs for ip_convolve1x1 and ip_convolve3x3.  It contains a real
 * version of ip_genConvAsm, which calls ip_gen1x1ConvAsm and ip_gen3x3ConvAsm to
 * generate machine code to perform glyph convolutions.
 *
 * To keep me sane, we'll actually have different procedures to generate
 * the two types of convolution code.  One type of convolution code produces
 * a single convolution sum (like sglyph.c's version of ip_convolve1x1).
 * The other type generates a 3x3 array of output sums, picks the biggest
 * one that's less than pKernel->upperThreshold and returns it while also 
 * modifying the image pointers to point to the position corresponding 
 * to the biggest sum (like sglyph.c's version of ip_convolve3x3).
 * ip_genConvAsm simply figures out which type of code to generate and 
 * calls the appropriate procedure.
 */

/* This is a private routine */
/* PRIVATE */
/*
 *  ip_genConvAsm():
 *        This routine generates machine code to process a kernel of a
 *	  particular shape.  The code is specialized for each platform.
 *	  If we don't know how to generate code for this platform, return
 *	  a pointer to the general C implementation.
 *
 *	pKernel:	Pointer to IAKernel data structure containing
 *			information about the kernel's shape.
 *	allocator:	Ptr to function that is used to allocate memory
 *			for the code.  Function's single argument is
 *			the number of BYTES requested.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of malloc
 *			provided by WATCOM on the PC!!
 *	freer:		Ptr to function used to free memory in case of
 *			error
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of free
 *			provided by WATCOM on the PC!!
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		couldn't allocate internal buffers
 *					
 */
Int32 CDECL 
ip_genConvAsm (
    IPState    *pIPState,
    IAKernel   *pKernel,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	),
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *))
{
    PROLOGUE;
    UInt32	codeSize;

	/* Keep the compiler happy */
    if (pKernel->is3x3)
    {
	__error__ = ip_gen3x3ConvAsm (pIPState, pKernel,
				      allocator, freer, &codeSize);
#ifdef DEBUG
	if (__error__ == ia_successful)
	    ERR_PASS( i_disassembleSPARCOps(pIPState,
					    (UInt32 *)pKernel->conv3Code,
					    codeSize));
#endif
    }
    else
    {
	__error__ = ip_gen1x1ConvAsm (pIPState, pKernel,
				     allocator, freer, &codeSize);
#ifdef DEBUG
	if (__error__ == ia_successful)
	    ERR_PASS( i_disassembleSPARCOps(pIPState,
					    (UInt32 *)pKernel->conv1Code,
					    codeSize));
#endif
    }

    NORMAL_RETURN;
}

/* This is a private routine */
/* PRIVATE */
/*
 *  ip_convolve1x1():
 *        This routine does one convolution operation producing one sum.
 *
 *	pImage:		Pointer to word containing upper left corner of
 *			image block to be convolved.
 *	centerX:	ScaledInt1616 holding position of leftmost bit
 *			within word pointed to by pImage
 *	wpl:		Number of words/line in image.
 *	pKernel:	Pointer to IAKernel data structure used to
 *			define the convolution kernel.
 *
 *   Returns:
 *	convResult:	The convolution sum.
 *					
 */
 /* This is a stub since all the kernels should have machine code
  * implementations. */
Int32 CDECL 
ip_convolve1x1 (
    UInt32     *pImage,
    ScaledInt1616 centerX,
    UInt32	wpl,
    IAKernel   *pKernel)
{
    pImage = pImage;		/* leave no variable unused! */
    centerX = centerX;
    wpl = wpl;
    pKernel = pKernel;
    return 0;
}


/* This is a private routine */
/* PRIVATE */
/*
 *  ip_convolve3x3():
 *        This routine produces 9 convolution sums in a 3x3 array.
 * 	  It returns the largest of the sums that's less than the upper
 *	  threshold.  It also adjusts the input address pointer and
 *	  input value of centerX to point to the position of the largest
 *	  sum relative to the original estimated position.  The original
 *	  position is in the center of the 3x3 array.  If the center's
 *	  sum is as large as the largest sum, it is chosen.
 *
 *	ppImage:	Pointer to pointer to word containing upper
 *			left corner of estimated position of the glyph
 *			to be found.  The image pointer will be adjusted
 *			to point to the glyph center actually found.
 *	pCenterX:	Pointer to ScaledInt1616 holding estimated 
 *			position of leftmost bit of the glyph 
 *			within word pointed to by *ppImage.
 *	wpl:		Number of words/line in image.
 *	pKernel:	Pointer to IAKernel data structure used to
 *			define the convolution kernel.
 *	numOutBit:	The LS bit of this number is used to determine
 *			the order in which we search the 3x3 output grid
 *			for the biggest score.  If 0, we search left
 *			to right, top to bottom.  If 1, we search right
 *			to left, bottom to top.  This tends to keep
 *			us from wandering off the line of glyphs.
 *
 *   Returns:
 *	convResult:	The largest convolution sum less than
 *			pKernel->upperThreshold.
 *	in *ppImage, *pCenter:	Location of the image word containing
 *			the upper left corner of the image
 *			block corresponding to the best convolution sum.
 *			For example, if the sum in the upper left corner
 *			of the 3x3 array was the biggest legal one,
 *			*ppImage and *pCenter would be adjusted to the
 *			left by 1 and up by 1 from their original values.
 *					
 */
 /* This is a stub since all the kernels should have machine code
  * implementations. */
Int32 CDECL 
ip_convolve3x3 (
    UInt32    **ppImage,
    ScaledInt1616 *pCenterX,
    UInt32	wpl,
    IAKernel   *pKernel,
    UInt32	numOutBit)
{
    ppImage = ppImage;		/* leave no variable unused! */
    pCenterX = pCenterX;
    wpl = wpl;
    pKernel = pKernel;
    numOutBit = numOutBit;
    return 0;
}

/*

$Log: asglyph.c,v $
Revision 1.1.1.1  1999/09/20 18:18:09  jlin
Imported sources

Revision 1.5  1998/01/13 23:46:41  ddavies
asglyph.c -> /product/ipcore/ipshared/src/RCS/asglyph.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1997/07/11 01:48:23  ddavies
ddavies on Thu Jul 10 18:48:22 PDT 1997

asglyph.c -> /product/ipcore/ipshared/src/RCS/asglyph.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.3  1995/12/13 02:41:15  ejaquith
asglyph.c -> /product/ipcore/ipshared/src/RCS/asglyph.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/10/18  03:51:20  ddavies
 * ddavies on Tue Oct 17 20:50:30 PDT 1995
 *
 * asglyph.c -> /product/ipcore/ipshared/src/RCS/asglyph.c,v
 *
 * Revision 1.1  1994/01/22  03:47:46  ddavies
 * Initial revision
 *
 * Revision 1.8  1993/12/02  19:21:12  ddavies
 * fglyph.c -> /product/ipcore/ipshared/src/RCS/fglyph.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.7  1993/11/05  01:24:46  ddavies
 * fglyph.c -> /product/ipcore/ipshared/src/RCS/fglyph.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


