/*************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * bintable.c--procedures that make binarization tables
 *
 *           Int32     i_makeBinarizationTables()		
 *           Int32     i_make4To2bppBinarizationTable()
 *           Int32     i_make2bppBinarizationTables()
 *           Int32     i_makeInvBinarizationTables()
 *	     Int32     i_make2To1BinarizationTables()
 *	     Int32     i_make2To1InvBinarizationTables()
 *
 *          Lookup tables for reduction:
 *           Int32     ip_makePermuteLUT()
 *	     Int32     ip_makeInvPermuteLUT()
 *           Int32     ip_make4PermuteLUT()
 *	     Int32     ip_makeRedLUT2()
 *	     Int32     ip_makeRedLUT4()
 *
 *	    Sequence tables for non-interpolated rotation:
 *	     Int32     ip_makeNoInterpRotTable()
 */

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "lmask.prv"
#include "memory.pub"
#include "defines.pub"

    /* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: bintable.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:09 $");

/*-----------------------------------------------------------------------*
 *                Binarization Bit-Unscrambling LUTs                     *
 *-----------------------------------------------------------------------*/

/* PUBLIC */
/*************************************************************************
 *  i_getBinarizationTableSizes()
 *	This routine returns the size in bytes of the tables needed for
 *	the i_thresholdXXX, i_highpassXXX, and i_passfiltXXX routines.
 *	Most of these routines need four tables.  They are all the same
 *	size.  That size is returned here.
 *
 *      Input:
 *              srcDepth:	depth of source image
 *		destDepth:	depth of dest image
 *		invert:		set to cTrue when tables will be used for
 *				i_thresholdIXXX, i_highpassIXXX or
 *				i_passfiltIXXX operations will be called.
 *		pTableSize:	ptr to word containing table size.
 *
 *************************************************************************/
Int32  CDECL 
i_getBinarizationTableSizes (
    IPState    *pIPState,
    UInt32	srcDepth,
    UInt32	destDepth,
    UInt32	invert,
    UInt32     *pTableSize)
{
	/* Currently, the invert tables are the same size as the non-invert
	 * tables.  Reference the invert flag to keep the compiler happy.
	 */
    PROLOGUE;
    invert = invert;

	/* And, well golly.  the source and dest depths don't currently
	 * matter either.  However, let's check them just to make sure
	 * that the client isn't spoofing..
	 */
    if ((srcDepth != 8) && (srcDepth != 4) && (srcDepth != 2))
	return ia_invalidParm;

    if ((destDepth != 4) && (destDepth != 2) && (destDepth != 1))
	return ia_invalidParm;

	/* Well, ok, the 8 to 4 case doesn't use a table */
    if ((srcDepth == 8) && (destDepth == 4))
	*pTableSize = 0;
    else
	    /* return the right number for all the other combos */
	*pTableSize = 256 * sizeof(UInt32);

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************
 *  i_makeBinarizationTables():  Allocates and fills Permute tables
 *		used to unscramble binarization bytes.
 *		These tables are used when converting 8 bit/pixel
 *		images to 1 bit/pixel images.  They are used by
 *		the straight threshold (threshold = 128), highpassfilt
 *		(threshold is an arg) and bandpass filter routines.
 *
 *	NOTE: The following tables may be allocated by the caller or
 *	      may be allocated here.  If allocated by the caller, then
 *	      ALL FOUR of the *pTabBinarizeByteX pointers must already
 *	      be pointing to tables.  If the tables are to be allocated
 *	      here, then *pTabBinarizeByte0 must be NULL.
 *	      The size of the tables is found by calling
 *	      i_getBinarizationTableSizes.  All four tables have the
 *	      same size.
 *
 *      Input:
 *              pTabBinarizeByte0:    ptr to ptr to permute table for byte 0
 *              pTabBinarizeByte1:    ptr to ptr to permute table for byte 1
 *              pTabBinarizeByte2:    ptr to ptr to permute table for byte 2
 *              pTabBinarizeByte3:    ptr to ptr to permute table for byte 3
 *
 *     On return:
 *             permute tables are generated
 *************************************************************************/
Int32  CDECL 
i_makeBinarizationTables (
    IPState    *pIPState,
    UInt32    **pTabBinarizeByte0,
    UInt32    **pTabBinarizeByte1,
    UInt32    **pTabBinarizeByte2,
    UInt32    **pTabBinarizeByte3)
{
    PROLOGUE;
    Int32	addr;
    UInt32	tableSize;
    UInt32     *pTabB0, *pTabB1, *pTabB2, *pTabB3;

    
        /* allocate if necessary, clear the tables if they're already
	 * allocated.
	 */
    i_getBinarizationTableSizes (pIPState, 8, 1, cFalse, &tableSize);
    if (*pTabBinarizeByte0 == NULL)
    {
	*pTabBinarizeByte0 = (UInt32 *) CALLOC (tableSize, 1);
	*pTabBinarizeByte1 = (UInt32 *) CALLOC (tableSize, 1);
	*pTabBinarizeByte2 = (UInt32 *) CALLOC (tableSize, 1);
	*pTabBinarizeByte3 = (UInt32 *) CALLOC (tableSize, 1);
	if (!*pTabBinarizeByte0 || !*pTabBinarizeByte1 ||
	    !*pTabBinarizeByte2 || !*pTabBinarizeByte3)
	    return ia_nomem;
    }
    else
    {
	memset((void *)*pTabBinarizeByte0, 0, tableSize);
	memset((void *)*pTabBinarizeByte1, 0, tableSize);
	memset((void *)*pTabBinarizeByte2, 0, tableSize);
	memset((void *)*pTabBinarizeByte3, 0, tableSize);
    }

        /* set the local table pointers */
    pTabB0 = (UInt32 *) *pTabBinarizeByte0;
    pTabB1 = (UInt32 *) *pTabBinarizeByte1;
    pTabB2 = (UInt32 *) *pTabBinarizeByte2;
    pTabB3 = (UInt32 *) *pTabBinarizeByte3;


	/* The *pTabBinarizeByteX tables un-interdigitate the inputs.  The
	 *   *pTabBinarizeByteX tables are indexed with addresses of the form
	 *   *pTabBinarizeByte0: 31, 27, 23, 19, 15, 11, 7, 3
	 *   *pTabBinarizeByte1: 30, 26, 22, 18, 14, 10, 6, 2
	 *   *pTabBinarizeByte2: 29, 25, 21, 17, 13,  9, 5, 1
	 *   *pTabBinarizeByte3: 28, 24, 20, 16, 12,  8, 4, 0
	 * The output bits are unscrambed to be (x is a bit position containing
	 * zero):
	 *   *pTabBinarizeByte0: 31xxx27xxx23xxx19xxx15xxx11xxx7xxx3xxx.
	 *   *pTabBinarizeByte1: x30xxx26xxx22xxx18xxx14xxx10xxx6xxx2xx.
	 *   *pTabBinarizeByte2: xx29xxx25xxx21xxx17xxx13xxx09xxx5xxx1x.
	 *   *pTabBinarizeByte3: xxx28xxx24xxx20xxx16xxx12xxx08xxx4xxx0.
	 */
   for (addr = 0; addr < 256; addr++)
   {
	pTabB0[addr] =  ((addr & 0x01) << ( 3-0) ) | /* bit 0 */
	                ((addr & 0x02) << ( 7-1) ) | /* bit 1 */
			((addr & 0x04) << (11-2) ) | /* bit 2 */
			((addr & 0x08) << (15-3) ) | /* bit 3 */
			((addr & 0x10) << (19-4) ) | /* bit 4 */
			((addr & 0x20) << (23-5) ) | /* bit 5 */
			((addr & 0x40) << (27-6) ) | /* bit 6 */
			((addr & 0x80) << (31-7) );  /* bit 7 */

	pTabB1[addr] =  ((addr & 0x01) << ( 2-0) ) | /* bit 0 */
	                ((addr & 0x02) << ( 6-1) ) | /* bit 1 */
			((addr & 0x04) << (10-2) ) | /* bit 2 */
			((addr & 0x08) << (14-3) ) | /* bit 3 */
			((addr & 0x10) << (18-4) ) | /* bit 4 */
			((addr & 0x20) << (22-5) ) | /* bit 5 */
			((addr & 0x40) << (26-6) ) | /* bit 6 */
			((addr & 0x80) << (30-7) );  /* bit 7 */

	pTabB2[addr] =  ((addr & 0x01) << ( 1-0) ) | /* bit 0 */
	                ((addr & 0x02) << ( 5-1) ) | /* bit 1 */
			((addr & 0x04) << ( 9-2) ) | /* bit 2 */
			((addr & 0x08) << (13-3) ) | /* bit 3 */
			((addr & 0x10) << (17-4) ) | /* bit 4 */
			((addr & 0x20) << (21-5) ) | /* bit 5 */
			((addr & 0x40) << (25-6) ) | /* bit 6 */
			((addr & 0x80) << (29-7) );  /* bit 7 */

	pTabB3[addr] =  ((addr & 0x01) << ( 0-0) ) | /* bit 0 */
	                ((addr & 0x02) << ( 4-1) ) | /* bit 1 */
			((addr & 0x04) << ( 8-2) ) | /* bit 2 */
			((addr & 0x08) << (12-3) ) | /* bit 3 */
			((addr & 0x10) << (16-4) ) | /* bit 4 */
			((addr & 0x20) << (20-5) ) | /* bit 5 */
			((addr & 0x40) << (24-6) ) | /* bit 6 */
			((addr & 0x80) << (28-7) );  /* bit 7 */
    }

    NORMAL_RETURN;
}


/*-----------------------------------------------------------------------*
 *                Binarization Bit-Unscrambling LUTs                     *
 *-----------------------------------------------------------------------*/

/* PUBLIC */
/*************************************************************************
 *  i_make4To2bppBinarizationTable():  Allocates and fills Permute table
 *			for unscrambling binarization bytes.
 *			This table is used when converting 4 bit/pixel
 *			images to 2 bit/pixel images by selecting the 2 MS
 *			bits of each nybble.
 *
 *	NOTE: The following table may be allocated by the caller or
 *	      may be allocated here.  If allocated by the caller, then
 *	      the *pTab4To2bpp pointer must already
 *	      be pointing to space for the table.  If the table is to
 *	      be allocated here, then *pTab4To2bpp must be NULL.
 *	      The size of the table is found by calling
 *	      i_getBinarizationTableSizes
 *
 *      Input:
 *             pTab4To2bpp:    ptr to ptr to byte 0 of Permute table
 *************************************************************************/
Int32  CDECL 
i_make4To2bppBinarizationTable (
    IPState    *pIPState,
    UInt32    **pTab4To2bpp)
{
PROLOGUE;
 register Int32   addr;
UInt32		 tableSize;
UInt32          *pTab;

	/* Allocate space for the 2 bit/pixel tables if necessary.
	 * Otherwise, clear out the already-allocated tables.
	 */
    i_getBinarizationTableSizes (pIPState, 4, 2, cFalse, &tableSize);
    if (*pTab4To2bpp == NULL)
    {
	*pTab4To2bpp = (UInt32 *) CALLOC (tableSize, 1);
	if (*pTab4To2bpp == NULL)
	    return ia_nomem;
    }
    else
    {
	memset((void *)*pTab4To2bpp, 0, tableSize);
    }


    /* Fill in the tables.  These are similar to the ones
     * used above, except we rearrange the data in units of 2 bit pixels
     * instead of 1 bit pixels.  First, allocate the new tables.
     * First, set the local table pointers */
    pTab = (UInt32 *) *pTab4To2bpp;

    /* The *pTab4To2bpp table un-interdigitate the inputs.
     * The output bits are unscrambed to be (x is a bit position containing
     * zero):
     *   *pTab4To2bpp: 0706xxxxxx0504xxxxxx0302xxxxxx0100xxxxxx
     */

   for (addr = 0; addr < 256; addr++)
   {
	pTab[addr] =  ((addr & 0x03) << (12-0) ) | /* bits 1,0 */
		      ((addr & 0x0C) << (28-2) ) | /* bits 3,2 */
		      ((addr & 0x30) << (14-4) ) | /* bits 5,4 */
		      ((addr & 0xC0) << (30-6) );  /* bits 7,6 */
    }

    NORMAL_RETURN;

}  /* i_make4To2bppBinarizationTables () */


/* PUBLIC */
/*************************************************************************
 *  i_make2bppBinarizationTables():  Allocates and fills Permute tables
 *			for unscrambling binarization bytes.
 *			These tables are used when converting 8 bit/pixel
 *			images to 2 bit/pixel images by selecting the 2 MS
 *			bits of each byte.
 *
 *	NOTE: The following tables may be allocated by the caller or
 *	      may be allocated here.  If allocated by the caller, then
 *	      ALL FOUR of the *pTabTo2bppByteX pointers must already
 *	      be pointing to tables.  If the tables are to be allocated
 *	      here, then *pTabTo2bppByte0 must be NULL.
 *	      The size of the tables is found by calling
 *	      i_getBinarizationTableSizes.  All four tables have the
 *	      same size.
 *
 *      Input:
 *             pTabTo2bppByte0:    ptr to ptr to permute table for byte 0
 *             pTabTo2bppByte1:    ptr to ptr to permute table for byte 1
 *             pTabTo2bppByte2:    ptr to ptr to permute table for byte 2
 *             pTabTo2bppByte3:    ptr to ptr to permute table for byte 3
 *
 *      On return:
 *             permute tables are generated
 *************************************************************************/
Int32  CDECL 
i_make2bppBinarizationTables (
    IPState    *pIPState,
    UInt32    **pTabTo2bppByte0,
    UInt32    **pTabTo2bppByte1,
    UInt32    **pTabTo2bppByte2,
    UInt32    **pTabTo2bppByte3)
{
    PROLOGUE;
    Int32	addr;
    UInt32	tableSize;
    UInt32     *pTabB0, *pTabB1, *pTabB2, *pTabB3;

	/* Allocate space for the 2 bit/pixel tables if necessary.
	 */
    i_getBinarizationTableSizes (pIPState, 8, 2, cFalse, &tableSize);
    if (*pTabTo2bppByte0 == NULL)
    {
	*pTabTo2bppByte0 = (UInt32 *) CALLOC (tableSize, 1);
	*pTabTo2bppByte1 = (UInt32 *) CALLOC (tableSize, 1);
	*pTabTo2bppByte2 = (UInt32 *) CALLOC (tableSize, 1);
	*pTabTo2bppByte3 = (UInt32 *) CALLOC (tableSize, 1);
	if (!*pTabTo2bppByte0 || !*pTabTo2bppByte1 ||
	    !*pTabTo2bppByte2 || !*pTabTo2bppByte3)
	    return ia_nomem;
    }
    else
    {
	memset((void *)*pTabTo2bppByte0, 0, tableSize);
	memset((void *)*pTabTo2bppByte1, 0, tableSize);
	memset((void *)*pTabTo2bppByte2, 0, tableSize);
	memset((void *)*pTabTo2bppByte3, 0, tableSize);
    }

        /* set the local table pointers */
    pTabB0 = (UInt32 *) *pTabTo2bppByte0;
    pTabB1 = (UInt32 *) *pTabTo2bppByte1;
    pTabB2 = (UInt32 *) *pTabTo2bppByte2;
    pTabB3 = (UInt32 *) *pTabTo2bppByte3;

	/* The *pTabTo2bppByteX tables un-interdigitate the inputs.
	 * the *pTabTo2bppByteX tables are indexed with addresses of the form
	 *   *pTabTo2bppByte0: 31,30, 23,22, 15,14, 7,6
	 *   *pTabTo2bppByte1: 29,28, 21,20, 13,12, 5,4
	 *   *pTabTo2bppByte2: 27,26, 19,18, 11,10, 3,2
	 *   *pTabTo2bppByte3: 25,24, 17,16,  9, 8, 1,0
	 * The output bits are unscrambed to be (x is a bit position containing
	 * zero; the apparent overlap is an artifact of putting 2 digits into
	 * a single bit position):
	 *   *pTabTo2bppByte0: 3130xxxxxx2322xxxxxx1514xxxxxx0706xxxxxx
	 *   *pTabTo2bppByte1: xx2928xxxxxx2120xxxxxx1312xxxxxx0504xxxx
	 *   *pTabTo2bppByte2: xxxx2726xxxxxx1918xxxxxx1110xxxxxx0302xx
	 *   *pTabTo2bppByte3: xxxxxx2524xxxxxx1716xxxxxx0908xxxxxx0100
	 */
   for (addr = 0; addr < 256; addr++)
   {
	pTabB0[addr] =  ((addr & 0x03) << ( 6-0) ) | /* bits 1,0 */
	                ((addr & 0x0C) << (14-2) ) | /* bits 3,2 */
			((addr & 0x30) << (22-4) ) | /* bits 5,4 */
			((addr & 0xC0) << (30-6) );  /* bits 7,6 */

	pTabB1[addr] =  ((addr & 0x03) << ( 4-0) ) | /* bits 1,0 */
	                ((addr & 0x0C) << (12-2) ) | /* bits 3,2 */
			((addr & 0x30) << (20-4) ) | /* bits 5,4 */
			((addr & 0xC0) << (28-6) );  /* bits 7,6 */

	pTabB2[addr] =  ((addr & 0x03) << ( 2-0) ) | /* bits 1,0 */
			((addr & 0x0C) << (10-2) ) | /* bits 3,2 */
			((addr & 0x30) << (18-4) ) | /* bits 5,4 */
			((addr & 0xC0) << (26-6) );  /* bits 7,6 */

	pTabB3[addr] =  ((addr & 0x03) << ( 0-0) ) | /* bits 1,0 */
			((addr & 0x0C) << ( 8-2) ) | /* bits 3,2 */
			((addr & 0x30) << (16-4) ) | /* bits 5,4 */
			((addr & 0xC0) << (24-6) );  /* bits 7,6 */
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************
 *  i_makeInvBinarizationTables():  Allocates and fills Permute tables
 *                     for unscrambling binarization bytes.  Identical to
 *		       i_makeBinarizationTables, except contents of tables
 *		       are inverted.
 *
 *	NOTE: The following tables may be allocated by the caller or
 *	      may be allocated here.  If allocated by the caller, then
 *	      ALL FOUR of the *pTabBinarizeByteXInv pointers must already
 *	      be pointing to tables.  If the tables are to be allocated
 *	      here, then *pTabBinarizeByte0Inv must be NULL.
 *	      The size of the tables is found by calling
 *	      i_getBinarizationTableSizes.  All four tables have the
 *	      same size.
 *
 *       Input:
 *              pTabBinarizeByte0Inv:    ptr to ptr to permute table for byte 0
 *              pTabBinarizeByte1Inv:    ptr to ptr to permute table for byte 1
 *              pTabBinarizeByte2Inv:    ptr to ptr to permute table for byte 2
 *              pTabBinarizeByte3Inv:    ptr to ptr to permute table for byte 3
 *
 *       On return:
 *             permute tables are generated
 *************************************************************************/
Int32  CDECL 
i_makeInvBinarizationTables (
    IPState    *pIPState,
    UInt32    **pTabBinarizeByte0Inv,
    UInt32    **pTabBinarizeByte1Inv,
    UInt32    **pTabBinarizeByte2Inv,
    UInt32    **pTabBinarizeByte3Inv)
{
    PROLOGUE;
    UInt32	addr, iAddr;
    UInt32	tableSize;
    UInt32     *pTabB0, *pTabB1, *pTabB2, *pTabB3;

        /* allocate if necessary, clear the tables if they're already
	 * allocated.
	 */
    i_getBinarizationTableSizes (pIPState, 8, 1, cTrue, &tableSize);
    if (*pTabBinarizeByte0Inv == NULL)
    {
	*pTabBinarizeByte0Inv = (UInt32 *) CALLOC (tableSize, 1);
	*pTabBinarizeByte1Inv = (UInt32 *) CALLOC (tableSize, 1);
	*pTabBinarizeByte2Inv = (UInt32 *) CALLOC (tableSize, 1);
	*pTabBinarizeByte3Inv = (UInt32 *) CALLOC (tableSize, 1);
	if (!*pTabBinarizeByte0Inv || !*pTabBinarizeByte1Inv ||
	    !*pTabBinarizeByte2Inv || !*pTabBinarizeByte3Inv)
	    return ia_nomem;
    }
    else
    {
	memset((void *)*pTabBinarizeByte0Inv, 0, tableSize);
	memset((void *)*pTabBinarizeByte1Inv, 0, tableSize);
	memset((void *)*pTabBinarizeByte2Inv, 0, tableSize);
	memset((void *)*pTabBinarizeByte3Inv, 0, tableSize);
    }

        /* set the local table pointers */
    pTabB0 = (UInt32 *) *pTabBinarizeByte0Inv;
    pTabB1 = (UInt32 *) *pTabBinarizeByte1Inv;
    pTabB2 = (UInt32 *) *pTabBinarizeByte2Inv;
    pTabB3 = (UInt32 *) *pTabBinarizeByte3Inv;


	/* See comments in i_makeBinarizationTables for description of action.
	 * Here, we really want to invert the data bits, but NOT invert
	 * the filler bits.  We do this by inverting addr before relocating
	 * each bit.
	 */
   for (addr = 0; addr < 256; addr++)
   {
	iAddr = ~addr;
	pTabB0[addr] =		  ((iAddr & 0x01) << ( 3-0) ) |	/* bit 0 */
				  ((iAddr & 0x02) << ( 7-1) ) |	/* bit 1 */
				  ((iAddr & 0x04) << (11-2) ) |	/* bit 2 */
				  ((iAddr & 0x08) << (15-3) ) |	/* bit 3 */
				  ((iAddr & 0x10) << (19-4) ) |	/* bit 4 */
				  ((iAddr & 0x20) << (23-5) ) |	/* bit 5 */
				  ((iAddr & 0x40) << (27-6) ) |	/* bit 6 */
				  ((iAddr & 0x80) << (31-7) );	/* bit 7 */

	pTabB1[addr] =		  ((iAddr & 0x01) << ( 2-0) ) |	/* bit 0 */
				  ((iAddr & 0x02) << ( 6-1) ) |	/* bit 1 */
				  ((iAddr & 0x04) << (10-2) ) |	/* bit 2 */
				  ((iAddr & 0x08) << (14-3) ) |	/* bit 3 */
				  ((iAddr & 0x10) << (18-4) ) |	/* bit 4 */
				  ((iAddr & 0x20) << (22-5) ) |	/* bit 5 */
				  ((iAddr & 0x40) << (26-6) ) |	/* bit 6 */
				  ((iAddr & 0x80) << (30-7) );	/* bit 7 */

	pTabB2[addr] =		  ((iAddr & 0x01) << ( 1-0) ) |	/* bit 0 */
				  ((iAddr & 0x02) << ( 5-1) ) |	/* bit 1 */
				  ((iAddr & 0x04) << ( 9-2) ) |	/* bit 2 */
				  ((iAddr & 0x08) << (13-3) ) |	/* bit 3 */
				  ((iAddr & 0x10) << (17-4) ) |	/* bit 4 */
				  ((iAddr & 0x20) << (21-5) ) |	/* bit 5 */
				  ((iAddr & 0x40) << (25-6) ) |	/* bit 6 */
				  ((iAddr & 0x80) << (29-7) );	/* bit 7 */

	pTabB3[addr] =		  ((iAddr & 0x01) << ( 0-0) ) |	/* bit 0 */
				  ((iAddr & 0x02) << ( 4-1) ) |	/* bit 1 */
				  ((iAddr & 0x04) << ( 8-2) ) |	/* bit 2 */
				  ((iAddr & 0x08) << (12-3) ) |	/* bit 3 */
				  ((iAddr & 0x10) << (16-4) ) |	/* bit 4 */
				  ((iAddr & 0x20) << (20-5) ) |	/* bit 5 */
				  ((iAddr & 0x40) << (24-6) ) |	/* bit 6 */
				  ((iAddr & 0x80) << (28-7) );	/* bit 7 */
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************
 *  i_make2To1BinarizationTables():  Allocates and fills Permute tables
 *			for unscrambling binarization bytes during 2 bit/pixel
 *			to 1 bit/pixel binarization.
 *
 *	NOTE: The following tables may be allocated by the caller or
 *	      may be allocated here.  If allocated by the caller, then
 *	      ALL FOUR of the *pTab2To1bppByteX pointers must already
 *	      be pointing to tables.  If the tables are to be allocated
 *	      here, then *pTab2To1bppByte0 must be NULL.
 *	      The size of the tables is found by calling
 *	      i_getBinarizationTableSizes.  All four tables have the
 *	      same size.
 *
 *      Input:
 *             pTab2To1bppByte0:   ptr to ptr to permute table for byte 0
 *             pTab2To1bppByte1:   ptr to ptr to permute table for byte 1
 *             pTab2To1bppByte2:   ptr to ptr to permute table for byte 2
 *             pTab2To1bppByte3:   ptr to ptr to permute table for byte 3
 *
 *      On return:
 *             permute tables are generated
 *************************************************************************/
Int32  CDECL 
i_make2To1BinarizationTables (
    IPState    *pIPState,
    UInt32    **pTab2To1bppByte0,
    UInt32    **pTab2To1bppByte1,
    UInt32    **pTab2To1bppByte2,
    UInt32    **pTab2To1bppByte3)
{
    PROLOGUE;
    Int32	addr;
    UInt32	tableSize;
    UInt32     *pTabB0, *pTabB1, *pTabB2, *pTabB3;

        /* allocate if necessary, clear the tables if they're already
	 * allocated.
	 */
    i_getBinarizationTableSizes (pIPState, 2, 1, cFalse, &tableSize);
    if (*pTab2To1bppByte0 == NULL)
    {
	*pTab2To1bppByte0 = (UInt32 *) CALLOC (tableSize, 1);
	*pTab2To1bppByte1 = (UInt32 *) CALLOC (tableSize, 1);
	*pTab2To1bppByte2 = (UInt32 *) CALLOC (tableSize, 1);
	*pTab2To1bppByte3 = (UInt32 *) CALLOC (tableSize, 1);
	if (!*pTab2To1bppByte0 || !*pTab2To1bppByte1 ||
	    !*pTab2To1bppByte2 || !*pTab2To1bppByte3)
	    return ia_nomem;
    }
    else
    {
	memset((void *)*pTab2To1bppByte0, 0, tableSize);
	memset((void *)*pTab2To1bppByte1, 0, tableSize);
	memset((void *)*pTab2To1bppByte2, 0, tableSize);
	memset((void *)*pTab2To1bppByte3, 0, tableSize);
    }

        /* set the local table pointers */
    pTabB0 = (UInt32 *) *pTab2To1bppByte0;
    pTabB1 = (UInt32 *) *pTab2To1bppByte1;
    pTabB2 = (UInt32 *) *pTab2To1bppByte2;
    pTabB3 = (UInt32 *) *pTab2To1bppByte3;


	/* The *pTab2To1bppByteX tables un-interdigitate the inputs.
	 * the *pTab2To1bppByteX tables are indexed with addresses of the form
	 *   *pTab2To1bppByte0: 31, 15, 30, 14, 29, 13, 28, 12
	 *   *pTab2To1bppByte1: 27, 11, 26, 10, 25,  9, 24,  8
	 *   *pTab2To1bppByte2: 23,  7, 22,  6, 21,  5, 20,  4
	 *   *pTab2To1bppByte3: 19,  3, 18,  2, 17,  1, 16,  0
	 * The output bits are unscrambed to be (x is a bit position containing
	 * zero):
	 *   *pTab2To1bppByte0: 31302928xxxx xxxxxxxx 15141312xxxx xxxxxxxx.
	 *   *pTab2To1bppByte1: xxxx27262524 xxxxxxxx xxxx11100908 xxxxxxxx.
	 *   *pTab2To1bppByte2: xxxxxxxx 23222120xxxx xxxxxxxx 07060504xxxx.
	 *   *pTab2To1bppByte3: xxxxxxxx xxxx19181716 xxxxxxxx xxxx03020100.
	 */
   for (addr = 0; addr < 256; addr++)
   {
	pTabB0[addr] =            ((addr & 0x01) << (12-0) ) |  /* bit 0 */
				  ((addr & 0x02) << (28-1) ) |	/* bit 1 */
				  ((addr & 0x04) << (13-2) ) |	/* bit 2 */
				  ((addr & 0x08) << (29-3) ) |	/* bit 3 */
				  ((addr & 0x10) << (14-4) ) |	/* bit 4 */
				  ((addr & 0x20) << (30-5) ) |	/* bit 5 */
				  ((addr & 0x40) << (15-6) ) |	/* bit 6 */
				  ((addr & 0x80) << (31-7) );	/* bit 7 */
	
	pTabB1[addr] =            ((addr & 0x01) << ( 8-0) ) |	/* bit 0 */
				  ((addr & 0x02) << (24-1) ) |	/* bit 1 */
				  ((addr & 0x04) << ( 9-2) ) |	/* bit 2 */
				  ((addr & 0x08) << (25-3) ) |	/* bit 3 */
				  ((addr & 0x10) << (10-4) ) |	/* bit 4 */
				  ((addr & 0x20) << (26-5) ) |	/* bit 5 */
				  ((addr & 0x40) << (11-6) ) |	/* bit 6 */
				  ((addr & 0x80) << (27-7) );	/* bit 7 */

	pTabB2[addr] =            ((addr & 0x01) << ( 4-0) ) |	/* bit 0 */
				  ((addr & 0x02) << (20-1) ) |	/* bit 1 */
				  ((addr & 0x04) << ( 5-2) ) |	/* bit 2 */
				  ((addr & 0x08) << (21-3) ) |	/* bit 3 */
				  ((addr & 0x10) << ( 6-4) ) |	/* bit 4 */
				  ((addr & 0x20) << (22-5) ) |	/* bit 5 */
				  ((addr & 0x40) << ( 7-6) ) |	/* bit 6 */
				  ((addr & 0x80) << (23-7) );	/* bit 7 */
	
	pTabB3[addr] =            ((addr & 0x01) <<  ( 0-0) ) |	/* bit 0 */
				  ((addr & 0x02) <<  (16-1) ) |	/* bit 1 */
				  ((addr & 0x04) >> -( 1-2) ) |	/* bit 2 */
				  ((addr & 0x08) <<  (17-3) ) |	/* bit 3 */
				  ((addr & 0x10) >> -( 2-4) ) |	/* bit 4 */
				  ((addr & 0x20) <<  (18-5) ) |	/* bit 5 */
				  ((addr & 0x40) >> -( 3-6) ) |	/* bit 6 */
				  ((addr & 0x80) <<  (19-7) );	/* bit 7 */
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************
 *  i_make2To1InvBinarizationTables():  Allocates and fills Permute tables
 *			for unscrambling binarization bytes during 2 bit/pixel
 *			to 1 bit/pixel binarization.  These complemented tables
 *			are used by the accelerated version of
 *			thresholdI2to1bpp.
 *
 *	NOTE: The following tables may be allocated by the caller or
 *	      may be allocated here.  If allocated by the caller, then
 *	      ALL FOUR of the *pTab2To1bppByteXInv pointers must already
 *	      be pointing to tables.  If the tables are to be allocated
 *	      here, then *pTab2To1bppByte0Inv must be NULL.
 *	      The size of the tables is found by calling
 *	      i_getBinarizationTableSizes.  All four tables have the
 *	      same size.
 *
 *      Input:
 *             pTab2To1bppByte0Inv:   ptr to ptr to permute table for byte 0
 *             pTab2To1bppByte1Inv:   ptr to ptr to permute table for byte 1
 *             pTab2To1bppByte2Inv:   ptr to ptr to permute table for byte 2
 *             pTab2To1bppByte3Inv:   ptr to ptr to permute table for byte 3
 *
 *      On return:
 *              permute tables are generated
 *************************************************************************/
Int32  CDECL 
i_make2To1InvBinarizationTables (
    IPState    *pIPState,
    UInt32    **pTab2To1bppByte0Inv,
    UInt32    **pTab2To1bppByte1Inv,
    UInt32    **pTab2To1bppByte2Inv,
    UInt32    **pTab2To1bppByte3Inv)
{
    PROLOGUE;
    Int32	addr, iAddr;
    UInt32	tableSize;
    UInt32     *pTabB0, *pTabB1, *pTabB2, *pTabB3;

        /* allocate if necessary, clear the tables if they're already
	 * allocated.
	 */
    i_getBinarizationTableSizes (pIPState, 2, 1, cTrue, &tableSize);
    if (*pTab2To1bppByte0Inv == NULL)
    {
	*pTab2To1bppByte0Inv = (UInt32 *) CALLOC (tableSize, 1);
	*pTab2To1bppByte1Inv = (UInt32 *) CALLOC (tableSize, 1);
	*pTab2To1bppByte2Inv = (UInt32 *) CALLOC (tableSize, 1);
	*pTab2To1bppByte3Inv = (UInt32 *) CALLOC (tableSize, 1);
	if (!*pTab2To1bppByte0Inv || !*pTab2To1bppByte1Inv ||
	    !*pTab2To1bppByte2Inv || !*pTab2To1bppByte3Inv)
	    return ia_nomem;
    }
    else
    {
	memset((void *)*pTab2To1bppByte0Inv, 0, tableSize);
	memset((void *)*pTab2To1bppByte1Inv, 0, tableSize);
	memset((void *)*pTab2To1bppByte2Inv, 0, tableSize);
	memset((void *)*pTab2To1bppByte3Inv, 0, tableSize);
    }

        /* set the local table pointers */
    pTabB0 = (UInt32 *) *pTab2To1bppByte0Inv;
    pTabB1 = (UInt32 *) *pTab2To1bppByte1Inv;
    pTabB2 = (UInt32 *) *pTab2To1bppByte2Inv;
    pTabB3 = (UInt32 *) *pTab2To1bppByte3Inv;


	/* The *pTab2To1bppByteXInv tables un-interdigitate the inputs.  The
	 * *pTab2To1bppByteXInv tables are indexed with addresses of the form
	 *   *pTab2To1bppByte0Inv: 31, 15, 30, 14, 29, 13, 28, 12
	 *   *pTab2To1bppByte1Inv: 27, 11, 26, 10, 25,  9, 24,  8
	 *   *pTab2To1bppByte2Inv: 23,  7, 22,  6, 21,  5, 20,  4
	 *   *pTab2To1bppByte3Inv: 19,  3, 18,  2, 17,  1, 16,  0
	 * The output bits are unscrambed to be (x is a bit position containing
	 * zero, other bit positions are to be logically complemented):
	 *   *pTab2To1bppByte0Inv: 31302928xxxx xxxxxxxx 15141312xxxx xxxxxxxx.
	 *   *pTab2To1bppByte1Inv: xxxx27262524 xxxxxxxx xxxx11100908 xxxxxxxx.
	 *   *pTab2To1bppByte2Inv: xxxxxxxx 23222120xxxx xxxxxxxx 07060504xxxx.
	 *   *pTab2To1bppByte3Inv: xxxxxxxx xxxx19181716 xxxxxxxx xxxx03020100.
	 */
   for (addr = 0; addr < 256; addr++)
   {
	iAddr = ~addr;
	pTabB0[addr] =               ((iAddr & 0x01) << (12-0) ) | /* bit 0 */
				     ((iAddr & 0x02) << (28-1) ) | /* bit 1 */
				     ((iAddr & 0x04) << (13-2) ) | /* bit 2 */
				     ((iAddr & 0x08) << (29-3) ) | /* bit 3 */
				     ((iAddr & 0x10) << (14-4) ) | /* bit 4 */
				     ((iAddr & 0x20) << (30-5) ) | /* bit 5 */
				     ((iAddr & 0x40) << (15-6) ) | /* bit 6 */
				     ((iAddr & 0x80) << (31-7) );  /* bit 7 */

	pTabB1[addr] =               ((iAddr & 0x01) << ( 8-0) ) | /* bit 0 */
				     ((iAddr & 0x02) << (24-1) ) | /* bit 1 */
				     ((iAddr & 0x04) << ( 9-2) ) | /* bit 2 */
				     ((iAddr & 0x08) << (25-3) ) | /* bit 3 */
				     ((iAddr & 0x10) << (10-4) ) | /* bit 4 */
				     ((iAddr & 0x20) << (26-5) ) | /* bit 5 */
				     ((iAddr & 0x40) << (11-6) ) | /* bit 6 */
				     ((iAddr & 0x80) << (27-7) );  /* bit 7 */
	
	pTabB2[addr] =               ((iAddr & 0x01) << ( 4-0) ) | /* bit 0 */
				     ((iAddr & 0x02) << (20-1) ) | /* bit 1 */
				     ((iAddr & 0x04) << ( 5-2) ) | /* bit 2 */
				     ((iAddr & 0x08) << (21-3) ) | /* bit 3 */
				     ((iAddr & 0x10) << ( 6-4) ) | /* bit 4 */
				     ((iAddr & 0x20) << (22-5) ) | /* bit 5 */
				     ((iAddr & 0x40) << ( 7-6) ) | /* bit 6 */
				     ((iAddr & 0x80) << (23-7) );  /* bit 7 */

	pTabB3[addr] =               ((iAddr & 0x01) <<  ( 0-0) ) | /* bit 0 */
				     ((iAddr & 0x02) <<  (16-1) ) | /* bit 1 */
				     ((iAddr & 0x04) >> -( 1-2) ) | /* bit 2 */
				     ((iAddr & 0x08) <<  (17-3) ) | /* bit 3 */
				     ((iAddr & 0x10) >> -( 2-4) ) | /* bit 4 */
				     ((iAddr & 0x20) <<  (18-5) ) | /* bit 5 */
				     ((iAddr & 0x40) >> -( 3-6) ) | /* bit 6 */
				     ((iAddr & 0x80) <<  (19-7) );  /* bit 7 */
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/*************************************************************************
 *  i_make4To1BinarizationTables():  Allocates and fills Permute tables
 *			for unscrambling binarization bytes during 4 bit/pixel
 *			to 1 bit/pixel binarization.
 *
 *	NOTE: The following tables may be allocated by the caller or
 *	      may be allocated here.  If allocated by the caller, then
 *	      ALL FOUR of the *pTab4To1bppByteX pointers must already
 *	      be pointing to tables.  If the tables are to be allocated
 *	      here, then *pTab4To1bppByte0 must be NULL.
 *	      The size of the tables is found by calling
 *	      i_getBinarizationTableSizes.  All four tables have the
 *	      same size.
 *
 *     Input:
 *             pTab4To1bppByte0:   ptr to ptr to permute table for byte 0
 *             pTab4To1bppByte1:   ptr to ptr to permute table for byte 1
 *             pTab4To1bppByte2:   ptr to ptr to permute table for byte 2
 *             pTab4To1bppByte3:   ptr to ptr to permute table for byte 3
 *
 *     On return:
 *             permute tables are generated
 *************************************************************************/
Int32  CDECL 
i_make4To1BinarizationTables (
    IPState    *pIPState,
    UInt32    **pTab4To1bppByte0,
    UInt32    **pTab4To1bppByte1,
    UInt32    **pTab4To1bppByte2,
    UInt32    **pTab4To1bppByte3)
{
    PROLOGUE;
    Int32	addr;
    UInt32	tableSize;
    UInt32     *pTabB0, *pTabB1, *pTabB2, *pTabB3;

        /* allocate if necessary, clear the tables if they're already
	 * allocated.
	 */
    i_getBinarizationTableSizes (pIPState, 4, 1, cFalse, &tableSize);
    if (*pTab4To1bppByte0 == NULL)
    {
	*pTab4To1bppByte0 = (UInt32 *) CALLOC (tableSize, 1);
	*pTab4To1bppByte1 = (UInt32 *) CALLOC (tableSize, 1);
	*pTab4To1bppByte2 = (UInt32 *) CALLOC (tableSize, 1);
	*pTab4To1bppByte3 = (UInt32 *) CALLOC (tableSize, 1);
	if (!*pTab4To1bppByte0 || !*pTab4To1bppByte1 ||
	    !*pTab4To1bppByte2 || !*pTab4To1bppByte3)
	    return ia_nomem;
    }
    else
    {
	memset((void *)*pTab4To1bppByte0, 0, tableSize);
	memset((void *)*pTab4To1bppByte1, 0, tableSize);
	memset((void *)*pTab4To1bppByte2, 0, tableSize);
	memset((void *)*pTab4To1bppByte3, 0, tableSize);
    }

        /* set the local table pointers */
    pTabB0 = (UInt32 *) *pTab4To1bppByte0;
    pTabB1 = (UInt32 *) *pTab4To1bppByte1;
    pTabB2 = (UInt32 *) *pTab4To1bppByte2;
    pTabB3 = (UInt32 *) *pTab4To1bppByte3;

	/* The *pTab4To1bppByteX tables un-interdigitate the inputs.
	 * the *pTab4To1bppByteX tables are indexed with addresses of the form
	 *   *pTab4To1bppByte0: 31, 23, 15,  7, 30, 22, 14,  6
	 *   *pTab4To1bppByte1: 29, 21, 13,  5, 28, 20, 12,  4
	 *   *pTab4To1bppByte2: 27, 19, 11,  3, 26, 18, 10,  2
	 *   *pTab4To1bppByte3: 25, 17,  9,  1, 24, 16,  8,  0
	 * The output bits are unscrambed to be (x is a bit position containing
	 * zero):
	 *   *pTab4To1bppByte0: 3130xxxxxx 2322xxxxxx 1514xxxxxx 76xxxxxx.
	 *   *pTab4To1bppByte1: xx2928xxxx xx2120xxxx xx1312xxxx xx54xxxx.
	 *   *pTab4To1bppByte2: xxxx2726xx xxxx1918xx xxxx1110xx xxxx32xx.
	 *   *pTab4To1bppByte3: xxxxxx2524 xxxxxx1716 xxxxxx98   xxxxxx10.
	 */
   for (addr = 0; addr < 256; addr++)
   {
       pTabB0[addr] =            ((addr & 0x01) << ( 6-0) ) |	/* bit 0 */
				 ((addr & 0x02) << (14-1) ) |	/* bit 1 */
				 ((addr & 0x04) << (22-2) ) |	/* bit 2 */
				 ((addr & 0x08) << (30-3) ) |	/* bit 3 */
				 ((addr & 0x10) << ( 7-4) ) |	/* bit 4 */
				 ((addr & 0x20) << (15-5) ) |	/* bit 5 */
				 ((addr & 0x40) << (23-6) ) |	/* bit 6 */
				 ((addr & 0x80) << (31-7) );	/* bit 7 */

       pTabB1[addr] =             ((addr & 0x01) << ( 4-0) ) |	/* bit 0 */
				  ((addr & 0x02) << (12-1) ) |	/* bit 1 */
				  ((addr & 0x04) << (20-2) ) |	/* bit 2 */
				  ((addr & 0x08) << (28-3) ) |	/* bit 3 */
				  ((addr & 0x10) << ( 5-4) ) |	/* bit 4 */
				  ((addr & 0x20) << (13-5) ) |	/* bit 5 */
				  ((addr & 0x40) << (21-6) ) |	/* bit 6 */
				  ((addr & 0x80) << (29-7) );	/* bit 7 */

       pTabB2[addr] =             ((addr & 0x01) <<  ( 2-0) ) |	/* bit 0 */
				  ((addr & 0x02) <<  (10-1) ) |	/* bit 1 */
				  ((addr & 0x04) <<  (18-2) ) |	/* bit 2 */
				  ((addr & 0x08) <<  (26-3) ) |	/* bit 3 */
				  ((addr & 0x10) >> -( 3-4) ) |	/* bit 4 */
				  ((addr & 0x20) <<  (11-5) ) |	/* bit 5 */
				  ((addr & 0x40) <<  (19-6) ) |	/* bit 6 */
				  ((addr & 0x80) <<  (27-7) );	/* bit 7 */

       pTabB3[addr] =             ((addr & 0x01) <<  ( 0-0) ) |	/* bit 0 */
				  ((addr & 0x02) <<  ( 8-1) ) |	/* bit 1 */
				  ((addr & 0x04) <<  (16-2) ) |	/* bit 2 */
				  ((addr & 0x08) <<  (24-3) ) |	/* bit 3 */
				  ((addr & 0x10) >> -( 1-4) ) |	/* bit 4 */
				  ((addr & 0x20) <<  ( 9-5) ) |	/* bit 5 */
				  ((addr & 0x40) <<  (17-6) ) |	/* bit 6 */
				  ((addr & 0x80) <<  (25-7) );	/* bit 7 */
    }

    NORMAL_RETURN;
}



/* PUBLIC */
/*************************************************************************
 *  i_make4To1InvBinarizationTables():  Allocates and fills Permute tables
 *			for unscrambling binarization bytes during 4 bit/pixel
 *			to 1 bit/pixel binarization.
 *
 *	NOTE: The following tables may be allocated by the caller or
 *	      may be allocated here.  If allocated by the caller, then
 *	      ALL FOUR of the *pTab4To1bppByteXInv pointers must already
 *	      be pointing to tables.  If the tables are to be allocated
 *	      here, then *pTab4To1bppByte0Inv must be NULL.
 *	      The size of the tables is found by calling
 *	      i_getBinarizationTableSizes.  All four tables have the
 *	      same size.
 *
 *     Input:
 *             pTab4To1bppByte0Inv:   ptr to ptr to permute table for byte 0
 *             pTab4To1bppByte1Inv:   ptr to ptr to permute table for byte 1
 *             pTab4To1bppByte2Inv:   ptr to ptr to permute table for byte 2
 *             pTab4To1bppByte3Inv:   ptr to ptr to permute table for byte 3
 *
 *     On return:
 *            permute tables are generated
 *************************************************************************/
Int32  CDECL 
i_make4To1InvBinarizationTables (
    IPState    *pIPState,
    UInt32    **pTab4To1bppByte0Inv,
    UInt32    **pTab4To1bppByte1Inv,
    UInt32    **pTab4To1bppByte2Inv,
    UInt32    **pTab4To1bppByte3Inv)
{
    PROLOGUE;
    Int32	addr, iaddr;
    UInt32	tableSize;
    UInt32     *pTabB0, *pTabB1, *pTabB2, *pTabB3;

        /* allocate if necessary, clear the tables if they're already
	 * allocated.
	 */
    i_getBinarizationTableSizes (pIPState, 4, 1, cTrue, &tableSize);
    if (*pTab4To1bppByte0Inv == NULL)
    {
	*pTab4To1bppByte0Inv = (UInt32 *) CALLOC (256, sizeof(UInt32));
	*pTab4To1bppByte1Inv = (UInt32 *) CALLOC (256, sizeof(UInt32));
	*pTab4To1bppByte2Inv = (UInt32 *) CALLOC (256, sizeof(UInt32));
	*pTab4To1bppByte3Inv = (UInt32 *) CALLOC (256, sizeof(UInt32));
	if (!*pTab4To1bppByte0Inv || !*pTab4To1bppByte1Inv ||
	    !*pTab4To1bppByte2Inv || !*pTab4To1bppByte3Inv)
	    return ia_nomem;
    }
    else
    {
	memset((void *)*pTab4To1bppByte0Inv, 0, tableSize);
	memset((void *)*pTab4To1bppByte1Inv, 0, tableSize);
	memset((void *)*pTab4To1bppByte2Inv, 0, tableSize);
	memset((void *)*pTab4To1bppByte3Inv, 0, tableSize);
    }

        /* set the local table pointers */
    pTabB0 = (UInt32 *) *pTab4To1bppByte0Inv;
    pTabB1 = (UInt32 *) *pTab4To1bppByte1Inv;
    pTabB2 = (UInt32 *) *pTab4To1bppByte2Inv;
    pTabB3 = (UInt32 *) *pTab4To1bppByte3Inv;

	/* The *pTab4To1bppByteX tables un-interdigitate the inputs.
	 * the *pTab4To1bppByteX tables are indexed with addresses of the form
	 *   *pTab4To1bppByte0Inv: 31, 23, 15,  7, 30, 22, 14,  6
	 *   *pTab4To1bppByte1Inv: 29, 21, 13,  5, 28, 20, 12,  4
	 *   *pTab4To1bppByte2Inv: 27, 19, 11,  3, 26, 18, 10,  2
	 *   *pTab4To1bppByte3Inv: 25, 17,  9,  1, 24, 16,  8,  0
	 * The output bits are unscrambed to be (x is a bit position containing
	 * zero):
	 *   *pTab4To1bppByte0: 3130xxxxxx 2322xxxxxx 1514xxxxxx 76xxxxxx.
	 *   *pTab4To1bppByte1: xx2928xxxx xx2120xxxx xx1312xxxx xx54xxxx.
	 *   *pTab4To1bppByte2: xxxx2726xx xxxx1918xx xxxx1110xx xxxx32xx.
	 *   *pTab4To1bppByte3: xxxxxx2524 xxxxxx1716 xxxxxx98   xxxxxx10.
	 */
   for (addr = 0; addr < 256; addr++)
   {
	iaddr = ~addr;
	pTabB0[addr] =           ((iaddr & 0x01) << ( 6-0) ) |	/* bit 0 */
				 ((iaddr & 0x02) << (14-1) ) |	/* bit 1 */
				 ((iaddr & 0x04) << (22-2) ) |	/* bit 2 */
				 ((iaddr & 0x08) << (30-3) ) |	/* bit 3 */
				 ((iaddr & 0x10) << ( 7-4) ) |	/* bit 4 */
				 ((iaddr & 0x20) << (15-5) ) |	/* bit 5 */
				 ((iaddr & 0x40) << (23-6) ) |	/* bit 6 */
				 ((iaddr & 0x80) << (31-7) );	/* bit 7 */

	pTabB1[addr] =  	 ((iaddr & 0x01) << ( 4-0) ) |	/* bit 0 */
				 ((iaddr & 0x02) << (12-1) ) |	/* bit 1 */
				 ((iaddr & 0x04) << (20-2) ) |	/* bit 2 */
				 ((iaddr & 0x08) << (28-3) ) |	/* bit 3 */
				 ((iaddr & 0x10) << ( 5-4) ) |	/* bit 4 */
				 ((iaddr & 0x20) << (13-5) ) |	/* bit 5 */
				 ((iaddr & 0x40) << (21-6) ) |	/* bit 6 */
				 ((iaddr & 0x80) << (29-7) );	/* bit 7 */

	pTabB2[addr] =           ((iaddr & 0x01) <<  ( 2-0) ) |	/* bit 0 */
				 ((iaddr & 0x02) <<  (10-1) ) |	/* bit 1 */
				 ((iaddr & 0x04) <<  (18-2) ) |	/* bit 2 */
				 ((iaddr & 0x08) <<  (26-3) ) |	/* bit 3 */
				 ((iaddr & 0x10) >> -( 3-4) ) |	/* bit 4 */
				 ((iaddr & 0x20) <<  (11-5) ) |	/* bit 5 */
				 ((iaddr & 0x40) <<  (19-6) ) |	/* bit 6 */
				 ((iaddr & 0x80) <<  (27-7) );	/* bit 7 */

	pTabB3[addr] =           ((iaddr & 0x01) <<  ( 0-0) ) |	/* bit 0 */
				 ((iaddr & 0x02) <<  ( 8-1) ) |	/* bit 1 */
				 ((iaddr & 0x04) <<  (16-2) ) |	/* bit 2 */
				 ((iaddr & 0x08) <<  (24-3) ) |	/* bit 3 */
				 ((iaddr & 0x10) >> -( 1-4) ) |	/* bit 4 */
				 ((iaddr & 0x20) <<  ( 9-5) ) |	/* bit 5 */
				 ((iaddr & 0x40) <<  (17-6) ) |	/* bit 6 */
				 ((iaddr & 0x80) <<  (25-7) );	/* bit 7 */
    }

    NORMAL_RETURN;
}



/*********************************************************************
 ************        2x Threshold reduction LUTs         *************
 *********************************************************************/

/* This is a private procedure */
/* PRIVATE */
/*
 *  ip_makePermuteLUT():  Allocates and fills Permute table
 *                     used in 2:1 threshold reduction routines.  These tables
 *		       are also used to convert 2 bit/pix images to 1 bit/pix
 *		       images.
 *                     Originally, we used one 64K x 2 byte table.  That
 *                     was pretty big.  We now use a 256 x 2 byte table.
 *                     There are more opcodes/translation, but the cache
 *                     statistics should be better.  It's also a lot easier
 *                     to make the table as needed instead of having it hang
 *                     around forever.
 *     Input:
 *             pTabPermute:  ptr to ptr to Permute byte table
 */
Int32  CDECL 
ip_makePermuteLUT (
    IPState    *pIPState,
    UInt16    **pTabPermute)
{
    PROLOGUE;
    Int32	addr, data;
    UInt16     *pTab;

        /* allocate */
    *pTabPermute = (UInt16 *) CALLOC (256, sizeof(UInt16));
    if (!*pTabPermute)
        return ia_nomem;

    pTab = (UInt16 *) *pTabPermute;
 
        /* The *pTabPermute tables un-interdigitate (a new word?) the inputs.
         * Code will do two accesses to this table to convert 16 bits into
         * 16 permuted bits.  In each access, we convert bytes of the form
         * {11, 3, 10, 2, 9, 1, 8, 0} into 16-bit halfwords of the form
         * {z, z, z, z, 11, 10, 9, 8, z, z, z, z, 3, 2, 1, 0} where the "z"
         * are zeros.  The results of the access using the high order input
         * byte are shifted left by 4 and ORed with the results of the access         * using the low order byte.
         */
    for (addr = 0; addr < 256; addr++)
    {
        data =  ((addr & 0x0001)     ) |
                ((addr & 0x0004) >> 1) |
                ((addr & 0x0010) >> 2) |
                ((addr & 0x0040) >> 3) |
                ((addr & 0x0002) << 7) |
                ((addr & 0x0008) << 6) |
                ((addr & 0x0020) << 5) |
                ((addr & 0x0080) << 4);
        pTab[addr] = (UInt16) data;
    }

    NORMAL_RETURN;
}


/* This is a private procedure */
/* PRIVATE */
/*
 *  ip_makeInvPermuteLUT():  Allocates and fills Permute table
 *			  used in 2:1 threshold reduction routines.
 *                        Originally, we used one 64K x 2 byte table.  That
 *                        was pretty big.  We now use a 256 x 2 byte table.
 *                        There are more opcodes/translation, but the cache
 *                        statistics should be better.  It's also a lot easier
 *                        to make the table as needed instead of having it hang
 *                        around forever.
 *			  This routine is just like ip_makeInvPermuteLUT except
 *			  the result tables entries are complemented.  It is
 *			  used by thresholdI2To1bppPr to complement the image
 *			  while it's being thresholded to 1 bit/pixel.
 *     Input:
 *            pTabPermuteInv:   ptr to ptr to Permute byte table
 */
Int32  CDECL 
ip_makeInvPermuteLUT (
    IPState    *pIPState,
    UInt16    **pTabPermuteInv)
{
    PROLOGUE;
    Int32	addr, iAddr, data;
    UInt16     *pTab;

        /* allocate */
    *pTabPermuteInv = (UInt16 *) CALLOC (256, sizeof(UInt16));
    if (!*pTabPermuteInv)
        return ia_nomem;

    pTab = (UInt16 *) *pTabPermuteInv;
 
        /* The *pTabPermuteInv tables un-interdigitate (new word?) the inputs.
         * Code will do two accesses to this table to convert 16 bits into
         * 16 permuted bits.  In each access, we convert bytes of the form
         * {11, 3, 10, 2, 9, 1, 8, 0} into 16-bit halfwords of the form
         * {z, z, z, z, 11, 10, 9, 8, z, z, z, z, 3, 2, 1, 0} where the "z"
         * are zeros.  The results of the access using the high order input
         * byte are shifted left by 4 and ORed with the results of the access
	 * using the low order byte.
	 * This is the same table as is made in ip_makePermuteLUT except each
	 * entry is complemented.  i_thresholdI2To1bpp uses these tables to
	 * created complemented threshold images.
	 * Here, we really want to invert the data bits, but NOT invert
	 * the filler bits.  We do this by inverting addr before relocating
	 * each bit.
	 */
    for (addr = 0; addr < 256; addr++)
    {
	iAddr = ~addr;
        data =  ((iAddr & 0x0001)     ) |
                ((iAddr & 0x0004) >> 1) |
                ((iAddr & 0x0010) >> 2) |
                ((iAddr & 0x0040) >> 3) |
                ((iAddr & 0x0002) << 7) |
                ((iAddr & 0x0008) << 6) |
                ((iAddr & 0x0020) << 5) |
                ((iAddr & 0x0080) << 4);
        pTab[addr] = (UInt16) data;
    }

    NORMAL_RETURN;
}


/*********************************************************************
 ************        4x Threshold reduction LUTs         *************
 *********************************************************************/

/* This is a private procedure */
/* PRIVATE */
/*
 *  ip_make4PermuteLUT():  Allocates and fills Permute table
 *                         used in 4:1 threshold reduction routines.
 *     Input:
 *            pTab4Permute:   ptr to ptr to Permute byte table
 */
Int32  CDECL 
ip_make4PermuteLUT (
    IPState    *pIPState,
    UInt8     **pTab4Permute)
{
    PROLOGUE;
    Int32	addr, data;
    UInt8      *pTab;

        /* allocate */
    *pTab4Permute = (UInt8 *) CALLOC (256, sizeof(Int8));
    if (!*pTab4Permute)
        return ia_nomem;

    pTab = (UInt8 *) *pTab4Permute;

        /* The *pTab4Permute table un-interdigitates the inputs.
         * It is called with an address whose bits are in the form
         *          7,3,5,1,6,2,4,0.
         * The output bits are unscrambed to be
                    7,6,5,4,3,2,1,0.
        */
   for (addr = 0; addr < 256; addr++)
   {
        data =  ((addr & 0x01)     ) |          /* bit 0 */
                ((addr & 0x10) >> 3) |          /* bit 1 */
                ((addr & 0x04)     ) |          /* bit 2 */
                ((addr & 0x40) >> 3) |          /* bit 3 */
                ((addr & 0x02) << 3) |          /* bit 4 */
                ((addr & 0x20)     ) |          /* bit 5 */
                ((addr & 0x08) << 3) |          /* bit 6 */
                ((addr & 0x80)     );           /* bit 7 */
        pTab[addr] = (UInt8) data;
    }

    NORMAL_RETURN;
}




/* ip_makeNoInterpRotTable simulates generating one image line.  The image line 
 * is formed from runs of input bits.  The length of the runs is 
 * computed from the rotation angle and the current y coordinate
 * within a pixel.  When a diagonal line drawn from the current 
 * fractional y coord crosses the top edge of the current row of 
 * pixels, the run ends.
 *
 * Runs may or may not end at output word boundaries.  Masks always 
 * end at output word boundaries.  To allow executing code to tell 
 * the difference between the end of a run and a continuing run, 
 * this routine inserts a source address modifier into the sequence of
 * masks each time a mask touches the LS bit of the output word.  The 
 * code that uses the masks will always subtract sWpl from the 
 * source address at the end of a run, so will subtract sWpl at the end
 * of every dest word too.  If the run continues into the next word,
 * this routine will store either (sWpl<<2) or sWpl into that word.
 * Otherwise, it stores 0 in the mask word.
 *
 * We store (sWpl*4) if assembly code will be interpreting the table
 * instead of sWpl because the assembly routine uses this value directly
 * instead of shifting it left by 2 first (as required by the C semantics
 * of modifying word ptrs)
 *
 */

/* PRIVATE */
/*
 *  ip_makeNoInterpRotTable():  Makes the sequence of masks and offsets used
 *			for small angle rotation.
 *
 *			This is a sequence of masks used to insert input
 *			bits into output words.  The same sequence is
 *			used in every line.
 *
 *	sinRotAngle:	rotation angle in radians.  This is assumed to be
 *			positive (makeNoInterRotTable is called after
 *			the angle and starting location have been corrected
 *			if counterclockwise rotation was specified).
 *
 *	maskTable:	pointer to returned rotation table.  This proc
 *			allocates the table and returns the address in
 *			this pointer.  It's up to the caller to free
 *			the table.
 *
 *	wS:		Width of source image in pixels.  Used to compute
 *			length of mask sequence.
 *
 *	sWpl:		Words per line in source image.
 *
 *	depth:		number of bits/pixel in image, used to decide
 *			how wide to make the masks.
 *
 *	forAssembler:	cTrue if table will be interpreted by assembly routine.
 *
 *	Return code:
 *	 ia_nomem		no memory available for table
 *	 ia_depthNotSupported	cannot make tables for rotation of
 *				images with this depth
 *	 ia_successful		good completion
 *
 *
 */
Int32  CDECL 
ip_makeNoInterpRotTable (
    IPState    *pIPState,
    Int32	sinRotAngle,
    UInt32    **maskTable,
    Int32	wS,
    Int32	sWpl,
    Int32	depth,
    Int8	forAssembler)
{
    PROLOGUE;
    UInt32	*localTab;		/* pointer to table */
    Int32	 word;			/* interation vars */
    UInt32	 mask;			/* this is the mask we calculate */
    UInt32	 runLength;		/* length of current run */
    Int32	 shortRunLen, longRunLen; /* the two possible run lengths */
    Int32	 shortRunDiff, longRunDiff; /* differences between short
					 * and long run lengths (integers)
					 * and the ideal run length */
    Int32	 bitPosition;		/* position of current run in
					 * output word */
    Int32	 numMasks;		/* number of runs in the table */
    Int32	 numWords;		/* number of output words in a
					 * scan line */
    Int32	 diffSum;		/* running sum of differences
					 * in run length from the idea value */

	/* can we make tables for images of this depth? */
    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8))
        return ia_depthNotSupported;

	/* compute length of the table.  There's one entry per run
	 * plus one entry per word boundary.  How many runs are there?
	 * Well, we go up sinRotAngle for every pixel we move
	 * horizontally, so each run is of length 1/sinRotAngle.
	 * Therefore, there are wS / (1/sinRotAngle) runs.  BUT,
	 * at each word boundary, there's a chance that we'll need
	 * to generate an extra mask, so that adds a word to the
	 * table.  Also, each word boundary adds another word to
	 * hold sWpl<<2 if the runs does not end at the word or
	 * 0 if it does.
	 * Remember that sinRotAngle is a split integer so we must
	 * shift the product right by 16 to get the integer part of the
	 * sum.  Then, we need to add 1 because there's probably one
	 * left over.
	 * AND, round sinRotAngle, don't truncate it.
	 */
    numWords = ((wS * depth) >> 5) + 1;
    numMasks = ((wS * (sinRotAngle + 0x8000)) >> 16) + 1 + numWords;

	/* allocate the table */
    *maskTable = (UInt32 *) CALLOC (numMasks + numWords, sizeof(UInt32));

    if (*maskTable == NULL)
	return ia_nomem;

    localTab = (UInt32 *) *maskTable;

	/* Find the two possible run lengths.  We'll use split integers,
	 * so the binary point is between bits 15 and 16.  For purposes of
	 * illustration, say the correct run length for this rotation
	 * angle is 19.3 bits.  The run length calculation in pixels is
	 * 1/sinRotAngle.  In bit's it's depth * (1/sinRotAngle).  Since
	 * sinRotAngle has already been multiplied by (1<<16), we must
	 * put that constant into the numerator also, producing
	 * depth * (1<<16) / sinRotAngle, or just
	 * (depth << 16) / sinRotAngle.
	 * Also, note that the two run lengths differ by 1 pixel, not by
	 * 1 bit.  Therefore, the longer run length equals the shorter one
	 * plus depth. */
	/* Oringinally, we calculated shortRunDiff as 
	 * -((depth << 16) % sinRotAngle) but that doesn't give us a 
	 * fractional value.  We need to keep the fraction bits around
	 * somehow.  (Yeah, we could use floating point but we're 
	 * trying not to use floating point unless necessary.)
	 * So, when we do the initial calculation to find the shortRunLen,
	 * we'll shift left by 28 instead of by 16.  This should yield:
	 * | 31 30 29 28 | 27 26 25 ... 14 13 12 | 11 10 9  ... 2 1 0 |
	 *    0 0  0  0    integer run length	   fraction
	 *
	 * BUG FIX: we only want to move over in units of "depth" bits.
	 * Just dividing depth by sinRotAngle moves over by the right
	 * total distance, but doesn't necessarily take all steps in units
	 * of complete pixels.  To fix this, we compute the short and
	 * long run lengths as though the depth were 1, then multiply
	 * the run lengths by depth at the end.
	 */
    shortRunLen = (1 << 28) / sinRotAngle;		/* say this is 19.3 */
    shortRunDiff = -(shortRunLen & 0xFFF);		/* This becomes -0.3 */
	/* now get rid of the fractional bits in shortRunLen */
    shortRunLen = (shortRunLen >> 12) * depth;		/* This becomes 19
							 * in units of pixels */
    longRunLen = shortRunLen + depth;			/* this becomes 20 */
    longRunDiff = 0x1000 + shortRunDiff;		/* this becomes +0.7 */
    diffSum = 0;					/* the diff from 0 */

	/* Loop through the runs until we've filled enough to process
	 * a scan line. */
    word = 0;
    bitPosition = 0;
    while (word < numWords)
    {
	    /* compute length of current run */
	if (diffSum < 0)
	{
	    runLength = longRunLen;
	    diffSum += longRunDiff;
	}
	else
	{
	    runLength = shortRunLen;
	    diffSum += shortRunDiff;
	}

	while (runLength > 0)	/* compute masks */
	{
	    if (runLength <= 32)
	        mask = leftMask[runLength-1];	/* mask fits in a word */
	    else
		mask = 0xFFFFFFFF;	/* mask might not fit in a word */
	    mask >>= bitPosition;	/* shift it into position */
	    *localTab++ = mask;		/* store it into the table and */

	    if ((runLength + bitPosition) < 32)	/* more to do in this word */
	    {
		bitPosition += runLength;
		runLength = 0;		/* done with that run */
	    }
	    else
	    if ((runLength + bitPosition) == 32) /* run ended at word boundary */
	    {
		*localTab++ = 0;	/* don't tell the user of the table
					 * complete the run later */
		word++;
		if (word >= numWords)
		    break;
		runLength = 0;
		bitPosition = 0;
	    }
	    else
		/* run crossed word boundary */
	    {
		if (forAssembler)
		    *localTab++ = (sWpl<<2); 	/* tell assembly code
						 * to complete the run later */
		else
		    *localTab++ = sWpl; 	/* tell C code to complete the
						 * run later */

		word++;
		if (word >= numWords)
		    break;
		runLength -= (32 - bitPosition);
		bitPosition = 0;
	    }
	}	/* end of processing one run */
    }
    NORMAL_RETURN;
}


/* PRIVATE */
/*
 *  ip_makeRedLUT2()	Allocates and fills lookup table for 2x subsampled
 *			reduction.
 *
 * Input:		redTable - pointer to pointer to the table
 *
 * Returns:
 *	ia_nomem if unable to allocate memory for the table
 *	ia_successful on successful completion
 */
Int32 CDECL 
ip_makeRedLUT2 (
    IPState    *pIPState,
    UInt8     **redTable)
{
    PROLOGUE;
    Int32       i;
    UInt8      *tabR2;

    if ((*redTable = (UInt8 *) MALLOC(256*sizeof(UInt8))) == NULL)
	return ia_nomem;
    tabR2 = (UInt8 *) *redTable;


    for (i = 0; i < 256; i++)
        switch(i & 0xAA)
        {
        case    0x00:                   /* 0000 0000 */
            tabR2[i] = 0;
            break;
        case    0x02:                   /* 0000 0010 */
            tabR2[i] = 1;
            break;      
        case    0x08:                   /* 0000 1000 */
            tabR2[i] = 2;
            break;
        case    0x0A:                   /* 0000 1010 */
            tabR2[i] = 3;
            break;      
        case    0x20:                   /* 0010 0000 */
            tabR2[i] = 4;
            break;      
        case    0x22:                   /* 0010 0010 */
            tabR2[i] = 5;
            break;      
        case    0x28:                   /* 0010 1000 */
            tabR2[i] = 6;
            break;      
        case    0x2A:                   /* 0010 1010 */
            tabR2[i] = 7;
            break;      
        case    0x80:                   /* 1000 0000 */
            tabR2[i] = 8;
            break;      
        case    0x82:                   /* 1000 0010 */
            tabR2[i] = 9;
            break;
        case    0x88:                   /* 1000 1000 */
            tabR2[i] = 10;
            break;
        case    0x8A:                   /* 1000 1010 */
            tabR2[i] = 11;
            break;
        case    0xA0:                   /* 1010 0000 */
            tabR2[i] = 12;
            break;
        case    0xA2:                   /* 1010 0010 */
            tabR2[i] = 13;
            break;      
        case    0xA8:                   /* 1010 1000 */
            tabR2[i] = 14;
            break;
        case    0xAA:                   /* 1010 1010 */
            tabR2[i] =15;
            break;
        }

    NORMAL_RETURN;
}


/* PRIVATE */
/*
 *  ip_makeRedLUT4()	Allocates and fills lookup table for 4x subsampled
 *			reduction.
 *
 * Input:		redTable - pointer to pointer to the table
 *
 * Returns:
 *	ia_nomem if unable to allocate memory for the table
 *	ia_successful on successful completion
 */
Int32 CDECL 
ip_makeRedLUT4 (
    IPState    *pIPState,
    UInt8     **redTable)
{
    PROLOGUE;
    Int32	i;
    UInt8      *tabR4;

    if ((*redTable = (UInt8 *) MALLOC(256*sizeof(UInt8))) == NULL)
	return ia_nomem;
    tabR4 = (UInt8 *) *redTable;

    for (i = 0; i < 256; i++)
        switch(i & 0x88)
        {
        case    0x00:                   /* 0000 0000 */
            tabR4[i] = 0;
            break;
        case    0x08:                   /* 0000 1000 */
            tabR4[i] = 1;
            break;
        case    0x80:                   /* 1000 0000 */
            tabR4[i] = 2;
            break;              
        case    0x88:                   /* 1000 1000 */
            tabR4[i] = 3;
            break;
        }

    NORMAL_RETURN;
}

/*

$Log: bintable.c,v $
Revision 1.1.1.1  1999/09/20 18:18:09  jlin
Imported sources

Revision 1.29  1998/01/13 23:46:43  ddavies
bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.28  1995/12/13 02:41:19  ejaquith
bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.27  1995/10/18  03:51:36  ddavies
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.26  1995/02/18  05:49:02  ddavies
 * ddavies on Fri Feb 17 21:48:39 PST 1995
 *
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Fix bug in ip_makeNoInterpRotTable that caused the pixel masks generated
 * for gray image to end in the middles of pixels instead of at the ends.
 * The code was dividing depth by sinRotAngle instead of calculating
 * 1/sinRotAngle, getting rid of the fraction bits, then multiplying by depth.
 * This fixes a problem reported by Macduff Hughes in which rotated images
 * contained vertical strings of dots at the jaggie boundaries.
 *
 * Revision 1.25  1994/09/20  02:45:01  ddavies
 * ddavies on Mon Sep 19 19:43:07 PDT 1994
 *
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Cast first arg to memset to be a (void *) in all calls to memset.
 *
 * Revision 1.24  1994/09/17  17:42:52  ddavies
 * ddavies on Sat Sep 17 10:42:13 PDT 1994
 *
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Change the binarization table making procedures so that they only allocate
 * tables if the incoming table pointers are NULL.  Otherwise, use the tables
 * that are passed in.  Also, make these procedures public (i_) instead of
 * private (ip_) so clients can call them.  The new procedure names are:
 * i_getBinarizationTableSizes, i_makeBinarizationTables,
 * i_make4To2bppBinarizationTable, i_make2bppBinarizationTables,
 * i_makeInvBinarizationTables, i_make2To1BinarizationTables,
 * i_make2To1InvBinarizationTables, i_make4To1BinarizationTables,
 * i_make4To1InvBinarizationTables.
 *
 * Revision 1.23  1994/09/09  21:14:32  ddavies
 * ddavies on Fri Sep 9 14:13:19 PDT 1994
 *
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Try rounding sinRotAngle instead of truncating it when calculating the
 * * size of the table needed in ip_makeNoInterpRotTable.  This may cure the bug
 * Art found in which the code was writing (and later reading) off the end of
 * the table.
 *
 * Revision 1.22  1994/07/30  00:04:32  dferrell
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Wrote new ip_make4To2bppBinarizationTable () which creates the
 * un-interdigitating table for use by 4->2 bit conversions.
 *
 * Revision 1.21  1994/07/27  15:23:52  ejaquith
 * ejaquith on Wed Jul 27 08:19:56 PDT 1994
 *
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Changed all table creation calls to pass in pointers to pointers to
 * table bytes that get created.  Made reduction table functions take
 * passed in pointers to pointers as well.
 *
 * Revision 1.20  1994/02/04  20:12:22  dferrell
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.19  1993/12/02  19:20:03  ddavies
 * ddavies on Thu Dec 2 11:19:58 PST 1993
 *
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.18  1993/11/05  01:23:03  ddavies
 * ddavies on Thu Nov 4 17:22:54 PST 1993
 *
 * bintable.c -> /product/ipcore/ipshared/src/RCS/bintable.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


