/*****************************************************************
 *  Copyright (c) 1993 Xerox Corporation.  All rights reserved.  *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"

/* prototypes */
#include "fxpmath.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: fxpmath.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");


/* These routines were written by Ron Danisewicz at XIS.  There are a
 * few changes in format and type names to adapt them to ipshared.
 * Ron supplied a great many comments with the original file.
 * Dan Davies is to blame for any errors in these comments.
 */

/*
 *************************************************************
 *                  SCALING/FRACTIONAL ARITHMETIC
 *   NOTE, ALL ROUTINES ASSUME Int32 IS AT LEAST 32 bits long.
 *
 * INPUTS and OUTPUTS are 16 bit integers embedded in the lower halves
 * of Int32s.
 *
 * Suppose theta is a floating point number expressed in radians;
 *
 * fract_cos and fract_sin accept args of the form 
 *	ScaledAngle = (Int32) ( 32768.0 * (theta/(2.0*pi)) )
 *
 * where -32768 <= ScaledAngle <= +32768; i.e. -2*pi <= theta <= +2*pi.
 *
 * fract_cos and fract_sin produce results that have 17 integer bits (which
 * are almost always just used to hold the sign) and 15 fraction bits.
 * These are of type ScaledInt1715.
 *
 * fract_sin(ScaledAngle) =
 *	(UInt32) ((32768.0) * sin(2.0*pi * (ScaledAngle/32768.0)) )
 *
 * fract_cos(ScaledAngle) =
 *	(UInt32) ((32768.0) * cos(2.0*pi * (ScaledAngle/32768.0)) )
 *
 * i.e. the scaled outputs range from -32768 to +32768 as the ideal outputs
 * range from -1.0 to +1.0.
 *
 * iatan2 outputs ScaledAngles as well.  It takes as arguments the coordinates
 * of a point and computes the angle between a line drawn from (0, 0) to
 * (x, y) and a line drawn from (0, 0) to (1, 0).
 *   Assuming
 *	  -32768 < x < 32768
 *	  -32768 < y < 32768
 *
 *	       pi/2    . (x, y)
 *		|     /
 *		|    /
 *		|   /
 *		|  / output
 *		| /  angle
 *  -pi ________|/)________ 0
 *		|
 *		|
 *		|
 *	      -pi/2
 *
 * Then iatan2(x, y) = (Int32) ((32768.0 / (2.0 * pi)) * atan(y/x))
 * Furthermore, the real output is restricted to the range [-pi, +pi],
 * so the scaled output is restricted to the range [-32768/2, +32768/2].
 *    
 **********************
 * isqrt accepts a UInt32, x, and returns a
 * long y, such that y*y is as close to x as possible.
 * x and y are interpreted as unsigned integers.
 *
 * fract_sqrt accepts a long mantissa and long exponent where the
 * mantissa is interpreted as fractional, and the exponent
 * is interpreted as an integer.
 * 0 <= mantIn < 32768
 * 
 * x = ((float)mantIn/32768.0)*2.0^expIn
 *
 * and computes y = sqrt(x) by returning mantOut
 * and expOut where 
 *
 * y = (float)(mantOut/32768.0)*2.0^expOut
 * 
 *
 * 
 *****************************************************************
 *
 *    NOTES ON SCALING IMPLEMENTATION
 * 
 * Suppose x and y are floating point numbers that are less than 1:
 * Let (x)' denote the 16 bit fractional representation of x
 * 
 * x'=(Int32)(2.0^15)*x
 * y'=(Int32)(2.0^15)*y
 *
 * (x*y)'=(x*y)(2.0^15)= (x'*y')/(2.0^15)= (x'*y') >> 15 
 *
 *  
 *
 * Essentially, we could simulate a machine that has a fractional multiplier
 * and adder:
 *  (most DSP chips do it this way)
 *
 * 
 *       sign bit
 *       |
 *       |
 *       | radix point
 *       | |
 *       | |                                 
 *       | |                           accumulator
 * ext   v v     high accumulator                   low accumulator
 * _ _ | _ . _ _ _ _ _ _ _ _ _ _ _ _ _ _ _  | _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
 *      
 * ext                               operand register
 * _ _ | _ . _ _ _ _ _ _ _ _ _ _ _ _ _ _ _  |
 *
 *
 * A 16-bit fractional multiply can be simulated by a fixed point multiply
 * into the whole destination accumulator followed by a 1 bit arithmetic shift
 * left.  Just after the multiply, the low accumulator contains the LS 16 bits
 * of the product and the high accumulator contains the MS 14 bits.  Our
 * fractional representation has 15 bits of fraction, so the left shift is
 * needed to put the complete answer into the high accumulator.
 * An add is just a fixed-point add into the high accumulator. The radix
 * point in the accumulator is shifted 16 places left from that of the register,
 * so that it is these 16 bits (high accumulator) that we move back into
 * the register when we want to do an operation on the result.  The accumulator
 * extension bits are needed when a temporary result is greater than 1.
 *
 *   
 *                   .____  .____  ----
 *                      |     |        |
 *                      v     v        |
 *                                     |
 *                      *     *        |
 *                       *   *         |
 *                        ***          |
 *                         |           |
 *             ___ _ ______v________   ^ 
 *             ext| hi acc | low acc | |
 *             sss|s.hhhhhh|lllllllll| |
 *                      |              |
 *                      |              |
 *                      |              |
 *                      |              |
 *                      v              |
 *                       -------------- 
 *
 * Of course, we don't have this hardware.  However, if we could read/write
 * the 16 bit halves of the accumulator (and other registers), the normalization
 * required after a multiply could be performed efficiently by moving
 * the high half of the accumulator into the high half of a register.
 *
 * Since we can't read/write halves of registers and there are a max of
 * 32 bits in each register (no extension bits), and I'm  not guaranteed
 * that my intermediate computations will be < than 1 (but they will be
 * less than 2, I could rescale to guarantee < 1, but it's not worth it),
 * I'll replace the 1-bit shift left after the fixed-point multiply with
 * a 15 bit shift right for scaling after a multiply.  This convention
 * leaves integers in the lower 15 bits (plus sign) of each register.
 *
 * If I had more than 32 bits to play with (say if long int was 64 bits,
 * which my C manual tells me is compiler-implementation dependent), I
 * could split my word into a 32-bit extension and a 16-bit high acc and
 * a 16-bit low acc as in my diagram.  Then, whenever I needed to scale
 * after a multiply,  I'd just test to see if the magnitude of the long
 * int was less than 2^31.  If it was, I could just read the 16 bit short
 * that corresponded to the hi acc to perform the scaling necessary to
 * operate on the result again. But then again, I think I'd have to worry
 * about endianess for this read to work on all machines.  Let's just shift
 * right by 15 after each multiply.  If we ever put this junk in a tight
 * loop, I can redo it the fast way.
 * 
 * [DD's note from ignorance - We could normally keep the sign + 15 fraction
 * bits in either the LS 16 bits or the MS 16 bits of a register.  If we
 * keep them in the LS 16 bits, we need a right shift after each multiply.
 * If we keep them in the MS 16 bits, we need two right shifts before each
 * multiply.  Seems like the first scheme (the current scheme) is better,
 * right?]
 *   
 * 
 * So we're going to implement polynomial expansions using this
 * kind of scaling model.
 *
 *  y = cC0*x + cC1*x^2 + cC2*x^3 + .......
 *
 *
 * Where we assume x, cCk, and y are in fractional form.
 *
 */

/* Define generally useful constants */
#define cOne		32768		/* scaled version of 1.0 */
#define cOneHalf	16384		/* scaled version of 0.5 */

#define cPi		16384
#define c2Pi		32768
#define cPiOver2	 8192 
#define c3PiOver2	24576

/* define constants used for the fract_sin and fract_cos polynomials */
#define cC1 (-18704)
#define cC3   21165
#define cC5  (-2603)
#define cC7     142

/****************************************************************
 *
 * fract_cos()
 *	Input:	Angle in scaled radians where -32768 corresponds to -2*pi
 *		and +32768 corresponds to +2*pi.
 *		
 *	Output: (UInt32) ((32768.0) * cos((2.0 * pi * InputAngle)/32768.0))
 *
 *	That is, the output ranges between -32768 (representing -1.0) and
 *	32768 (representing +1.0).
 *
 * Cosine is computed using polynomial approximation (over -pi/2 to pi/2).
 * The polynomial coefficients were determined using a least squares fit.
 *
 *	Returns:
 *		ia_invalidParm	ABS(angle) is too big.
 *		ia_successful	result was computed
 ****************************************************************/
Int32 CDECL 
fract_cos (
    IPState    *pIPState,
    ScaledAngle	angle,
    ScaledInt1715Ptr result)
{
	/* declare terms of the polynomial */
    PROLOGUE;
    ScaledInt1715	arg2, arg3, arg5, arg7, accumOver2;

    if ((angle < -c2Pi) || (angle > c2Pi))
	return ia_invalidParm;

	/* The cosine function is symmetrical around 0 */
    angle = ABS(angle);
    
	/* We compute the cosine in two steps.  In the first one, we
	 * map the range of angles into the range of outputs using
	 * two straight lines.  In the second step, we bend the lines
	 * so they look more like cosine.  The initial mapping is:
	 * [0, pi/2, pi, 3pi/2, 2pi] -> [-cOne, 0, cOne, 0, -cOne].
	 * One of the lines goes from (0, -cOne) to (pi, cOne).  The
	 * other one goes from (pi, cOne) to (2pi, -cOne).  This
	 * is a linear approximation to cosine.
	 */
    angle = cPiOver2 - ABS(cPi - angle);
    angle = angle * 4;		/* converts pi/2 to cOne */

	/* If cos were a triangular function, we'd be done now
	 * Except for inverting the sign of the output.
	 * It's not, so we need to bend the curves a little
	 * with the polonomial given below.
	 * Set arg3 = angle^3, arg5 = angle^5, arg7 = angle^7.
	 * A stunning trick of this type of multiply is that
	 * SCALED_MULTIPLY(cOne, cOne) = cOne.  This is saying
	 * that 1 * 1 = 1.
	 * Lets do a quick sanity check.  fract_cos(0) => at this point
	 * angle = cOne, so all powers of angle also equal
	 * cOne.  Therefore, (accumOver2/cOne) is set to -cOne + C1
	 * + C3 + C5 + C7 = 32768 = cOne.  This is
	 * correct.  For fract_cos(pi/2), angle = 0 at this point,
	 * so all powers of angle also equal 0, so the sums are all 0.
	 * Zowee!
	 */
    arg2 = SCALED_MULTIPLY(angle, angle);
    arg3 = SCALED_MULTIPLY(arg2, angle);
    arg5 = SCALED_MULTIPLY(arg2, arg3);
    arg7 = SCALED_MULTIPLY(arg5, arg2);

    accumOver2 = -cOne*angle + cC1*angle + cC3*arg3 + cC5*arg5 + cC7*arg7;
    *result = ((accumOver2 + 0x4000) / cOne);	/* round the result and
						 * shift fraction bits 
						 * into position */
    NORMAL_RETURN;
}


/****************************************************************
 *
 * fract_sin()
 *	Input:	Angle in scaled radians where -32768 corresponds to -2*pi
 *		and +32768 corresponds to +2*pi.
 *		
 *	Output: *result =
		(UInt32) ((32768.0) * sin((2.0 * pi * InputAngle)/32768.0))
 *
 *	That is, the output ranges between -32768 (representing -1.0) and
 *	32768 (representing +1.0).
 *
 * Sine is computed using polynomial approximation (over -pi/2 to pi/2).
 * The polynomial coefficients were determined using a least squares fit.
 *
 *	Returns:
 *		ia_invalidParm	ABS(angle) is too big.
 *		ia_successful	result was computed
 ****************************************************************/
Int32 CDECL 
fract_sin (
    IPState    *pIPState,
    ScaledAngle	angle,
    ScaledInt1715Ptr result)
{
	/* declare terms of the polynomial */
    PROLOGUE;
    ScaledInt1715	arg2, arg3, arg5, arg7, accum;
    ScaledInt1715	sign;

    if ((angle < -c2Pi) || (angle > c2Pi))
	return ia_invalidParm;

	/* sign(sin(angle)) = sign(angle), so remember the original sign
	 * and use the absolute value of the angle below.  This reduces
	 * the range of angles in which we must construct a linear
	 * approximation.
	 */
    if (angle < 0) 
      sign = -1;
    else 
      sign = 1;
    angle = ABS(angle);

	/* Conversion takes place in two steps.  First, we map the angle
	 * into two straight lines that approximate a sine.  Then we bend
	 * the lines appropriately.  The two lines should map
	 * [-2pi, -3pi/2, -pi, -pi/2, 0, pi/2, pi, 3pi/2, 2pi] ->
	 * [  0,    cOne,   0,  -cOne,  0, cOne,  0,   -cOne,  0 ].
	 * By taking the absolute value of the angle above, we've reduced
	 * the mapping to:
	 * [0, pi/2, pi, 3pi/2, 2pi] -> [0, cOne, 0, -cOne, 0].
	 * Note there are 2 inflection points on the curve; at pi/2 and
	 * 3pi/2.  We only want to use one ABS value in the calculation
	 * below, so can have only one inflection point.  To get around
	 * this, we map all angles in the range [3pi/2, 2pi] to the
	 * range [-pi/2, 0].  These two parts of the sine curve have the
	 * same values, so it's ok.
	 */
    if (angle > c3PiOver2)
	angle = angle - c2Pi;

	/* Do the mapping described above. */
    angle = ABS(angle - cPiOver2) - cPiOver2;
    angle = angle * 4;			/* convert angles to outputs */

	/* If sin were a triangular function, we'd be done now
	 * except for inverting the sign of the output.
	 * It's not, so we need to bend the curves a little
	 * with the polonomial given below.
	 * Set arg3 = angle^3, arg5 = angle^5, arg7 = angle^7.
	 * A stunning trick of this type of multiply is that
	 * SCALED_MULTIPLY(cOne, cOne) = cOne.  This is saying
	 * that 1 * 1 = 1.
	 * Let's do a quick sanity check.  fract_sin(pi/2) => at this point
	 * angle = cOne, so all powers of angle also equal
	 * cOne.  Therefore, (accum/cOne) is set to -cOne + C1
	 * + C3 + C5 + C7 = 32768 = scaled value of 1.  This is
	 * correct.  For fract_sin(0), angle = 0 at this point,
	 * so all powers of angle also equal 0, so the sums are all 0.
	 * Zowee!
	 */
    arg2 = SCALED_MULTIPLY(angle, angle);
    arg3 = SCALED_MULTIPLY(arg2, angle);
    arg5 = SCALED_MULTIPLY(arg2, arg3);
    arg7 = SCALED_MULTIPLY(arg5, arg2);
    accum = -cOne*angle + cC1*angle + cC3*arg3 + cC5*arg5 + cC7*arg7;
    accum = ((accum + 0x4000) / cOne);	/* round the result and shift
					 * fraction bits into position */

	/* flip the polarity if the original sign was negative. */
    if (sign < 0)
	accum = -accum;

    *result = accum;
    NORMAL_RETURN;
}

/****************************************************************
 *
 * isqrt()
 *	Integer square root routine
 *	Input:	an unsigned number.
 *		
 *	Output: square root of input
 *
 *	Returns:
 *		ia_successful	result was computed
 *
 *	This routine uses binary search, setting or resetting 1 bit
 *	at a time.
 ****************************************************************/
Int32 CDECL 
isqrt (
    IPState    *pIPState,
    UInt32	square,
    UInt32Ptr	squareRoot)
{
    PROLOGUE;
    UInt32	root, currentBit, test;
    currentBit = 0x8000;		/* the MS bit of the square root */
    root = 0;
    
	/* do the binary search.  At each stage, we add in a bit if it
	 * doesn't make the square root too big.
	 */
    while (currentBit != 0)
    {
	test = root + currentBit;
	currentBit >>= 1;
	if (test*test <= square)
	    root = test;
    }

	/* At this point, root is either exactly correct or too small.
	 * We round the result by saying if (root+1)^2 is closer
	 * to square than root^2, we set the result to root+1.
	 * Note that (root+1) * (root+1) = root^2 + 2*root + 1, so
	 * we can calculate (root+1)^2 by adding 2*root + 1 to root^2.
	 */
    test = root * root;
    if ( ((test + 2*root + 1) - square) < (square - test) )
	root++;

	/* post the results */
    *squareRoot = root;
    NORMAL_RETURN;
}

/****************************************************************
 *
 * fract_sqrt()
 *	Input:	number in the form of a 32 bit exponent and 32 bit
 *		mantissa where
 *		x = ((float)mantIn/32768.0) * 2.0^expIn
 *		0 <= mantIn < 32768
 *
 *		This square-root routines uses a 2-word floating point
 *		format.  The Mantissa is interpreted as a 16 bit
 *		fractional representation, the exponent is simply
 *		interpreted as a long integer.
 *
 *	Output: *mantOut and *expOut are updated so that
 *		sqrt(x) = (float)(mantOut/32768.0)*2.0^expOut
 *
 *	Returns:
 *		ia_invalidParm	mantIn is negative or too big.
 *		ia_successful	result was computed
 ****************************************************************/

Int32 CDECL 
fract_sqrt (
    IPState    *pIPState,
    ScaledInt1715 mantIn,
    UInt32	expIn,
    ScaledInt1715Ptr mantOut,
    UInt32Ptr	expOut) 
{
    PROLOGUE;
    UInt32	root, currentBit, test;

	/* if the incoming mantissa is negative or too big, we can't
	 * handle it. */
    if ((mantIn < 0) || (mantIn > cOne))
	return ia_invalidParm;

	/* If the incoming mantissa is 0, the output is always 0 */
    if (mantIn == 0) 
    {
	*mantOut = 0;
	*expOut = 0; 
	NORMAL_RETURN;
    }

	/* Normalize the incoming fraction if it wasn't already */
    while (mantIn < cOneHalf) 
    {
	mantIn = mantIn * 2;
	expIn--;
    }

	/* When taking the square root, we'd like to divide the exponent
	 * by 2.  However, we can only do this if it was even.  If
	 * the exponent was odd, increase it by 1 and shift the mantissa
	 * to accomadate the increase.  This may discard the LS bit of
	 * the mantissa, but big deal.
	 */
    if (0x1 & expIn) 
    {
	expIn++;
	mantIn = mantIn / 2; 
    }

	/* Divide the corrected exponent by 2.  This will be the final
	 * exponent.
	 */
    *expOut = expIn / 2;

	/* Now we need to find root such that SCALED_MULTIPLY(root, root) =
	 * mantIn.  We'll use a search, testing each bit of root in turn.
	 * Since each SCALED_MULTIPLY does a right shift by 15, we can
	 * save some time if we just shift mantIn left by 15 first and do
	 * regular multiplies.  The compares work the same either way.
	 * We work our way down from the MS bit of root, at each step seeing
	 * if root^2 is too big.  If so, we discard the current new bit.
	 * If not, the new bit is retained.
	 */
    mantIn <<= 15;
    currentBit = 0x8000;		/* set to MS bit of root */
    root = 0;				/* this probably isn't necessary,
					 * but it doesn't take much time.
					 * We know mantissa is > 0 and is
					 * normalized, so we'll always
					 * put something into root.
					 */
    while (currentBit != 0)
    {
	test = root + currentBit;
	currentBit >>= 1;  
	if (test*test <= (UInt32)mantIn)
	    root = test;		/* if the new value of root isn't
					 * too big, keep it */
    }

	/* At this point, root is either exactly correct or too small.
	 * We round the result by saying if (root+1)^2 is closer
	 * to square than root^2, we set the result to root+1.
	 * Note that (root+1) * (root+1) = root^2 + 2*root + 1, so
	 * we can calculate (root+1)^2 by adding 2*root + 1 to root^2.
	 */
    test = root * root;
    if ( ((test + 2*root + 1) - mantIn) < (mantIn - test) )
	root++;

	/* post the results and return */
    *mantOut = root;
    NORMAL_RETURN;
}


/****************************************************************************
 * The next routine computes arctan(y/x)
 * where x and y are 16-bit signed integers embedded in 32 bit integers.
 * They represent the coordinates of a point.  The angle returned is the
 * angle between a line from (0,0) to (x,y) and a line from (0,0) to (1,0).. 
 *
 * iatan(x, y) =
 *	(UInt32) ((32768.0 / (2.0 * pi)) * atan(y/x))
 *
 * This output ranges from -pi to pi, so the scaled angle ranges from
 * -32768/2 to 32768/2, where 32768 corresponds to 2*pi
 * 
 * 
 * This routine uses a 2 piece polynomial approximation.  The first segment
 * uses a Taylor polynomial expanded at 0 degrees for good accuracy
 * around 0, and good accuracy up to about 27 degrees.
 * 
 * The second segment uses a Taylor polynomial expanded around 38 degrees
 * for good accuracy between 27 degrees and 45 degrees.                    
 *                                                      
 * The input angle is mapped so that the output will be in first octant
 * (0, 45 degrees) for the computation.  The angle is mapped to the proper
 * octant before returning.                                              
 *                                                          
 * The output of the calculation is naturally in (fractional) radians.
 * The nice thing about radians is that they are essentially a distance,
 * just like y/x is a distance.  We will divide by 2PI after
 * computing the polynomial to be consistent with the format of
 * ScaledAngles required by fract_sin and fract_cos.     
 * 
 * The polynomial used when arctan(y/x) < 27 degrees:
 * quot = y/x
 * angle = (quot + cSA3*quot^3 + cSA5*quot^5 + cSA7*quot^7) / 2*pi             
 * 
 * The polynomial used when arctan(num/denom) > 27 degrees:      
 * quot = (y/x - cNominalQuot);
 * angle = (cNominalAngle + cBA1*quot   + cBA2*quot^2 +
 *			    cBA3*quot^3 + cBA4*quot^4)) / 2*pi
 *
 ****************************************************************************/

#define cSA3 (-10923)
#define cSA5  6554
#define cSA7 (-4681)

#define cNominalQuot 25601		/* equals ScaledInt1715 version of
					 * arctan(38 degrees) */
#define cNominalAngle 712114176		/* equals 2pi * ScaledAngle of 38
					 * degrees (32768 * 32768 * 0.66
					 * radians) */
#define cBA1 (20348)
#define cBA2 (-9872)
#define cBA3 (2174)
#define cBA4 (1483)

#define cTanOf27Degrees 16696		/* equals 32768 * tan(27 degrees) */

/****************************************************************
 *
 * iatan2()
 *	Input:	x and y coords of a point.
 *	  -32768 < x < 32768
 *	  -32768 < y < 32768
 *
 *	This function computes the tangent of the line drawn from
 *	(0,0) to (x,y) with respect to a line drawn from (0,0) to (1,0).
 *
 *	Output: *angle = (Int32) ((32768.0 / (2.0 * pi)) * atan(y/x))
 *
 *	That is, the output ranges between -32768/2 (representing -pi) and
 *	32768/2 (representing +pi).
 *
 *	Returns:
 *		ia_invalidParm	x or y were too big or both are 0.
 *		ia_successful	result was computed
 ****************************************************************/
Int32 CDECL 
iatan2 (
    IPState    *pIPState,
    Int32	x,
    Int32	y,
    ScaledAnglePtr angle)
{
	/* declare vars to hold powers of input quotient.  quotX =
	 * quotient to the X power. */
   PROLOGUE;
   Int32	quot, quot2, quot3, quot4, quot5, quot7, accum;
   UInt32	quad, octant;	 	/* these hold octant info */

	/* check to see if x and y are legal */
    if ((ABS(x) >= cOne) || (ABS(y) >= cOne))
    {
	*angle = c2Pi + 1;
	return ia_invalidParm;
    }
    if ((x == 0) && (y == 0))
    {
	*angle = 0;
	return ia_invalidParm;
    }

	/* remember the quadrant quadrant containing (x,y) and map
	 * the coord to the 1st quadrant.
	 */
    if (y >= 0)
    {
	if (x >= 0)
	    quad = 1;			/* both x and y are pos */
	else
	    quad = 2;			/* y is pos, x is neg */
    }
    else	/* y < 0 */
    {
	if (x >= 0)
	    quad = 4;			/* y is neg, x is pos */
	else
	    quad = 3;			/* y is neg, x is neg */
    }

    y = ABS(y);				/* map to first quadrant */
    x = ABS(x);
    
	/* If (x,y) is in the upper octant of the first quadrant, map it
	 * to the lower octant by switching x and y when computing
	 * the quotient.  This has the effect of reflecting the
	 * angle around the 45 degree line.  We do the equivalent of scaled
	 * multiplication, i.e. shift the numerator left by 15 before
	 * doing the divide (instead of shifting the product right
	 * by 15 after doing the multiply).
	 */
    if (y > x)	/* in upper octant */
    {
	quot = (x << 15) / y;
	octant = 1; 
    }
    else	/* in lower octant */
    {
	quot = (y << 15) / x;
	octant = 0;
    }

    if (quot < cTanOf27Degrees) /* small angle polynomial approx. */
    {
	quot2 = SCALED_MULTIPLY(quot, quot); 
	quot3 = SCALED_MULTIPLY(quot2, quot);
	quot5 = SCALED_MULTIPLY(quot2, quot3);
	quot7 = SCALED_MULTIPLY(quot5, quot2);
	accum = cOne*quot + cSA3*quot3 + cSA5*quot5 + cSA7*quot7;
    }
    else /* large angle polynomial approx. */
    {
	quot = quot - cNominalQuot;	/* compute difference from 38 degrees */
	quot2 = SCALED_MULTIPLY(quot, quot); 
	quot3 = SCALED_MULTIPLY(quot2, quot);
	quot4 = SCALED_MULTIPLY(quot2, quot2);
	accum = cNominalAngle + cBA1*quot + cBA2*quot2 + cBA3*quot3 +
				cBA4*quot4;
    }
    /* accum now contains 32768 * 32768 * theta, where theta is really
     * expressed in radians.  The first 32768 came from the multiplies
     * in the last step.  The second one came from the left shift of
     * the numerator used to compute quot by 15 before we started.
     * We want to return 32768 * (theta / 2pi).  Note this is the REAL
     * 2.0 * pi = 6.283186...  Therefore, we want to divide accum by
     * 32768, then multiply it by 1/6.283185....  We'd also
     * like to retain as many relevant bits as possible while making sure
     * nothing interesting overflows off the top.  We'll divide by (32768/2),
     * multiply by 4 * 32768/6.283186 (= 20861), then divide by 8 * 32768.
     */ 
    accum = (accum + 32768/4) / (32768/2); /* round accum, then
					 * accum = 2 * 32768 * theta */
    accum = accum * 20861;		/* multiply by 4 * 32768/6.28, so
					 * accum = 8 * 32768^2 * theta/2pi */
    accum = (accum + 4*32768) / (8 * 32768);	/*  round accum, then
					 * accum = 32768 * theta/2pi, which
					 * is the proper format for
					 * ScaledAngles
					 */

    /* map back to proper octant.  We moved to the lower octant of each
     * quadrant by switching x and y.  The is the same as reflecting the
     * angle around the 45 degree line.  Therefore, if we were in the
     * upper octant, reinvert the angle and subtract it from pi/2. */
    switch (quad)
    {
    case 1:
	if (octant != 0)		/* if octant == 0, we're done already */
	{
	 accum= cPiOver2 - accum;
	}
	break;
    case 2:
	if (octant == 0)
	    accum = cPi - accum;
	else
	    accum= cPiOver2 + accum;
	break;
    case 3:
	if (octant == 0)
	    accum = -cPi + accum;
	else
	    accum = -cPiOver2 - accum;
	break;
    default:	/* case 4 */
      if (octant == 0)
	 accum = -accum;
      else
	 accum = -cPiOver2 + accum;
    }

    *angle = accum;			/* store the result */
    NORMAL_RETURN;
}

/*

$Log: fxpmath.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.8  1998/01/13 23:46:47  ddavies
fxpmath.c -> /product/ipcore/ipshared/src/RCS/fxpmath.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1997/07/25 21:02:22  ddavies
ddavies on Fri Jul 25 14:00:45 PDT 1997

fxpmath.c -> /product/ipcore/ipshared/src/RCS/fxpmath.c,v

Fixed illegal arg check in iatan2.  Had been checking for y == 0, which
ruled out the prefectly valid angles of 0 degrees and 180 degrees.
It now checks for (x==0) && (y==0), for which there is no angle.
Also fixed the comment that talks about this check.

Revision 1.6  1997/07/22 22:37:29  aber
aber on Tue Jul 22 15:35:44 PDT 1997

fxpmath.c -> /product/ipcore/ipshared/src/RCS/fxpmath.c,v

If taking arctan with dy = 0, return 0 and error (like comments claim).

 * Revision 1.5  1995/12/13  02:41:26  ejaquith
 * fxpmath.c -> /product/ipcore/ipshared/src/RCS/fxpmath.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.4  1994/04/03  02:04:18  ddavies
 * ddavies on Sat Apr 2 18:04:13 PST 1994
 *
 * fxpmath.c -> /product/ipcore/ipshared/src/RCS/fxpmath.c,v
 *
 * Incorporate pointer types for use in SantaFe project.
 *
 * Revision 1.3  1993/12/02  19:21:49  ddavies
 * fxpmath.c -> /product/ipcore/ipshared/src/RCS/fxpmath.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:25:45  ddavies
 * fxpmath.c -> /product/ipcore/ipshared/src/RCS/fxpmath.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


