/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_absdif.c: produce the absolute value of the difference of two images.
 *
 *                  Int32     	i_absValDiff()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /*  prototyes */
#include "shrrast.pub"
#include "shrip.pub"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_absdif.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");
/* The procedure is used to compare two images.  At every pixel
 * of the destination image, it produces the absolute value of the difference
 * between the two source pixels.  If the images are binary, this operation
 * is a simple XOR.
 * 
 * This routine deals with depths of 1, 2, 4 and 8 bits/pixel..
 *
 * Since this result of this routine will be used to detect small differences,
 * we can't make any approximations that might speed things up.  Also, we
 * expect that this algorihm will only be used to do quality control, so
 * it isn't necessary that it be especially fast.  If performance becomes
 * an issue, we'll work on it.
 */


/*********************************************************************
 *************************      Procedures     ***********************
 *********************************************************************/
/* PUBLIC */
/*
 *  i_absValDiff()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (includes frame)
 *	sBpl1 =	 source image 1 bytes/line
 *	pS2 =	 ptr to source image 2 (includes frame)
 *	sBpl2 =	 source gray image 2 bytes/line
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	depth =	 bits/pixel of all gray images
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *	dBpl =	 dest gray image bytes/line
 *
 *	On return: 
 *		destination gray image that equals absolute value of
 *		difference between source gray images.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	image depth was not 1, 2, 4 or 8
 */
Int32   CDECL 
i_absValDiff (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt32	sBpl1,
    UInt8      *pS2,
    UInt32	sBpl2,
    UInt32	w,
    UInt32	h,
    UInt32	depth,
    UInt8      *pD,
    UInt32	dBpl)
{

	/* point to the images inside the frames */
    PROLOGUE;
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pS1 += sBpl1 << cLogFrameBits;	/* do left frame adjustment below */
    pS2 += sBpl2 << cLogFrameBits;
    pD += dBpl << cLogFrameBits;


	/* perform the correct operation given the image depth */
    switch (depth)
    {
    case 1:
	    /* Copy one of the sources to the destination, then xor
	     * it with the other one.  If one of the source pointers
	     * already points to the destination, just do the xor.
	     */
	pS1 += cFrameBytes;
	pS2 += cFrameBytes;
	pD += cFrameBytes;
	if (pS1 == pD)
	{
	    i_rasterOp (pIPState, pD, 0, 0, w, h, 1, dBpl,
		        PIX_SRC^PIX_DST,
		        pS2, 0, 0, 1, sBpl2);
	}
	else
	if (pS2 == pD)
	{
	    i_rasterOp (pIPState, pD, 0, 0, w, h, 1, dBpl,
		        PIX_SRC^PIX_DST,
		        pS1, 0, 0, 1, sBpl1);
	}
	else
	{
	    i_rasterOp (pIPState, pD, 0, 0, w, h, 1, dBpl,
		        PIX_SRC,
		        pS1, 0, 0, 1, sBpl1);
	    i_rasterOp (pIPState, pD, 0, 0, w, h, 1, dBpl,
		        PIX_SRC^PIX_DST,
		        pS2, 0, 0, 1, sBpl2);
	}
	break;

    case 2:
	pS1 += 2 * cFrameBytes;
	pS2 += 2 * cFrameBytes;
	pD += 2 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
		/* compute difference in one line of 2 bit pixels */
	    ip_absValDiffLine2 (pIPState, pS1, pS2, w, pD);
	}
	break;

    case 4:
	pS1 += 4 * cFrameBytes;
	pS2 += 4 * cFrameBytes;
	pD += 4 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
		/* compute difference in one line of 4 bit pixels */
	    ip_absValDiffLine4 (pIPState, pS1, pS2, w, pD);
	}
	break;

    case 8:
	pS1 += 8 * cFrameBytes;
	pS2 += 8 * cFrameBytes;
	pD += 8 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
		/* compute difference in one line of 8 bit pixels */
	    ip_absValDiffLine8 (pIPState, pS1, pS2, w, pD);
	}
	break;

    default:
	return ia_depthNotSupported;
    }

    NORMAL_RETURN;
}

/* This is where the real work gets done.  These procedures compute the
 * absolute values of the differences in one image line.
 *
 * The first procedure, absValDiffLine2, computes this for 2 bit pixels.
 * It would be very tedious to look at every 2-bit pixel.  Instead, we use
 * a trick.
 *
 * What happens if we just XOR the pixels together as we did in the 1-bit
 * case?
 *	Pix1	Pix2	XOR	Desired Result
 *	00	00	00	00
 *	00	01	01	01
 *	00	10	10	10
 *	00	11	11	11
 *
 *	01	00	01	01
 *	01	01	00	00
 *	01	10	11	01	WRONG
 *	01	11	10	10
 *
 *	10	00	10	10
 *	10	01	11	01	WRONG
 *	10	10	00	00
 *	10	11	01	01
 *
 *	11	00	11	11
 *	11	01	10	10
 *	11	10	01	01
 *	11	11	00	00
 *
 * What's amazing about this is that it gets the right answer 14 out of 16
 * times!  What can we do to detect the two bad cases?  Note that whenever
 * one of the two arguments is 01 or 10 and the other one is 10 or 01, we
 * can safely XOR the raw XOR with 10 and get the right answer.  This will
 * also fix the problem we're seeing here.  So we compute the result as
 * (Pix1 ^ Pix2) ^ ((Pix1 ^ (Pix1 << 1)) & (Pix2 ^ (Pix2 << 1)).
 */

/* PRIVATE */
/*
 *  ip_absValDiffLine2()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 2 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals absolute value of
 *		difference between source gray images.
 *
 */
void CDECL 
ip_absValDiffLine2 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	sWord1, sWord2, dWord;		/* holds source data */
    UInt8      *pDEnd;				/* pts to last dest word + 1 */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 15)/16) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);
	dWord = sWord1 ^ sWord2;		/* works for all cases except
						 * 01-10 and 10-01. */
	dWord &= ~((sWord1 ^ (sWord1 << 1)) &	/* catch these two cases */
		   (sWord2 ^ (sWord2 << 1)) &
		   0xAAAAAAAA);			/* and make the result 01
						 * instead of 11 */
	ULONG_ACCESS(pD) = dWord;
    }
}


/* The 4 and 8 bit/pixel routines could be more straightforward, but what
 * fun would that be?  Especially in the 4 bpp case, doing 8 shifts,
 * subtracts and so on is slow.  Instead, we'll use a sort of complicated
 * trick.
 *
 * Say we always operate on 32 bit words.  Call them Source1 and Source2.
 * If each pixel in Source1 were greater than or equal to the corresponding
 * pixel in Source2, we could get the proper result by simply computing
 * Source1 - Source2.  We will go to some trouble to rearrange the pixels
 * in Source1 and Source2 to guarantee that this is true.  Note that if
 * we compute the difference between Source1 and Source2 without guaranteeing
 * that all the big pixels are in Source1, we would get borrows that would
 * corrupt the results.
 *
 * To do rearrange the pixels, we need a mask that has 1's covering all the
 * pixel positions where Source1 is greater than or equal Source2 and 0's
 * elsewhere.  To compute this mask, we'll read Source1 and Source2 from
 * memory, then set the MS bit in each pixel in Source1 and clear them
 * in Source2.  We then subtract Source1 - Source2.  If the LS bits of
 * the pixel in Source1 were greater than or equal to the LS bits of the 
 * pixel in Source2, the MS bit in each pixel is now 1, otherwise, it's 0.
 * No borrows have gone between pixels.  The MS bit of each pixel now has
 * the correct mask value unless the MS bit of Source1's pixel was 0 and
 * the MS bit of Source2's pixel was 1.  We invert the original Source1,
 * OR it with the original Source2 and AND the result with the mask word
 * we've been computing.  The mask word now has the correct mask bits in
 * the MS bit of each pixel position.
 *
 * How do we fill in the rest of the mask bits?  We clear out the other
 * (trash) bits and use a sequence of shifts and OR's to fill out the
 * mask bits.
 *
 * What do we do with the mask?  We set a variable called diffWord to
 * the XOR of the original values of Source1 and Source2.  We then AND
 * diffWord with the mask.  It now has Source1 ^ Source2 in pixels where
 * Source1 >= Source2 and 0's elsewhere.  We set Source1 equal to
 * Source1 XOR diffWord.  This puts Source2 pixels into all the positions
 * where the original Source1 pixel was greater or equal and Source1
 * pixels elsewhere.  Finally, we XOR Source2 and diffWord.  This produces
 * a result that's opposite of Source1.  The final result word is obtained
 * by subtracting Source2 - Source1 (because of the way the pixels were
 * rearranged).
 *
 * This sounds complicated, but I think it's significantly faster than
 * dealing with each pixel individually.
 */

/* PRIVATE */
/*
 *  ip_absValDiffLine4()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 4 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals absolute value of
 *		difference between source gray images.
 *
 */
void CDECL 
ip_absValDiffLine4 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	sWord1, sWord2,		/* holds source data */
		dWord, xWord;			/* holds dest and intermediate*/
    UInt8      *pDEnd;				/* pts to last dest word + 1 */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 7)/8) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);

        xWord = sWord1 ^ sWord2;                /* we use this twice, so
	                                         * keep it around */

	dWord = (sWord1 | 0x88888888) -
		(sWord2 & 0x77777777);		/* MS bit of each pix has
						 * mask contribution from
						 * LS bits */
	dWord = ((~xWord) & dWord) +            /* determine whether MS bits */
	          (sWord1 & (~sWord2));		/* cause the pixel swap */

	dWord &= 0x88888888;			/* clear out unwanted bits */
	dWord = 0x88888888 - (dWord >> 3);      /* fill the pixel mask with */
	dWord ^= 0x88888888;                    /* 1's for swap, 0's for not */
	dWord &= xWord;				/* in each pix where
						 * sWord1 >= sWord2, we get
						 * pix1 ^ pix2.  In the other
						 * pixels, we have 0.
						 */
	sWord1 ^= dWord;			/* In each pix where sWord1
						 * is greater, we get sWord2's
						 * pixel.  In the remaining
						 * pixels, we get sWord1's pix
						 */
	sWord2 ^= dWord;			/* in each pix where sWord1
						 * is greater, we get sWord1's
						 * pix.  In the remaining
						 * pixels, we get sWord2's pix.
						 */

	ULONG_ACCESS(pD) = sWord2 - sWord1;
    }

}


/* PRIVATE */
/*
 *  ip_absValDiffLine8()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 8 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals absolute value of
 *		difference between source gray images.
 *
 */
void CDECL 
ip_absValDiffLine8 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	sWord1, sWord2,		/* holds source data */
		dWord, xWord;			/* holds dest and intermediate*/
    UInt8      *pDEnd;				/* pts to last dest word + 1 */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 3)/4) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);

	xWord = sWord1 ^ sWord2;                /* we use this twice, so
	                                         * keep it around */
	dWord = (sWord1 | 0x80808080) -
		(sWord2 & 0x7F7F7F7F);		/* MS bit of each pix has
						 * mask contribution from
						 * LS bits */
	dWord = ((~xWord) & dWord) +            /* determine whether MS bits */
	          (sWord1 & (~sWord2));		/* cause the pixel swap */

	dWord &= 0x80808080;			/* clear out unwanted bits */
	dWord = 0x80808080 - (dWord >> 7);      /* fill the pixel mask with */
	dWord ^= 0x80808080;                    /* 1's for swap, 0's for not */
	dWord &= xWord;				/* in each pix where
						 * sWord1 >= sWord2, we get
						 * pix1 ^ pix2.  In the other
						 * pixels, we have 0.
						 */
	sWord1 ^= dWord;			/* In each pix where sWord1
						 * is greater, we get sWord2's
						 * pixel.  In the remaining
						 * pixels, we get sWord1's pix
						 */
	sWord2 ^= dWord;			/* in each pix where sWord1
						 * is greater, we get sWord1's
						 * pix.  In the remaining
						 * pixels, we get sWord2's pix.
						 */

	ULONG_ACCESS(pD) = sWord2 - sWord1;
    }

}

/*

$Log: i_absdif.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.10  1998/01/13 23:46:51  ddavies
i_absdif.c -> /product/ipcore/ipshared/src/RCS/i_absdif.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.9  1996/06/19 00:19:52  dferrell
dferrell on Tue Jun 18 17:17:27 PDT 1996

i_absdif.c -> /product/ipcore/ipshared/src/RCS/i_absdif.c,v

Fixed bug in 4 and 8 bit functions which for certain pixel values would
cause a borrow to propagate between pixels and mess up the result.

 * Revision 1.8  1995/12/13  02:41:34  ejaquith
 * i_absdif.c -> /product/ipcore/ipshared/src/RCS/i_absdif.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.7  1994/07/12  22:21:15  ejaquith
 * ejaquith on Tue Jul 12 15:20:49 PDT 1994
 *
 * i_absdif.c -> /product/ipcore/ipshared/src/RCS/i_absdif.c,v
 *
 * Removed inclusion of file "iadef.prv"
 *
 * Revision 1.6  1994/02/04  20:12:47  dferrell
 * i_absdif.c -> /product/ipcore/ipshared/src/RCS/i_absdif.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.5  1993/12/02  19:21:59  ddavies
 * i_absdif.c -> /product/ipcore/ipshared/src/RCS/i_absdif.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.4  1993/11/05  01:26:17  ddavies
 * i_absdif.c -> /product/ipcore/ipshared/src/RCS/i_absdif.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


