/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_aieana.c	
 *
 *	AIE "default analayzer" module
 */


#include <stdio.h>
#include <string.h>
#include <math.h>

#ifdef _DEBUG
#ifdef PC_ENV
#define _POSIX_
#include <io.h>
#else
#include <unistd.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include <limits.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"
#include "aiedefs.pub"
#include "aiedefs.prv"
#include "defines.pub"

/* Prototypes */
#include "shrip.pub"
#include "shros.pub"
#include "shrip.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_aieana.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

#ifdef AIE_DEBUG
void
dump_trc(aTable *table, const char *filename);

void
dump_trc(aTable *table, const char *filename)
{
	     
    long point, points = table->length;
    long *trc = ip_tableData (pIPState, table);
    FILE *fp;

    if ((fp = fopen(filename, "wb")) == NULL) {
	fprintf(stderr, "Can't open %s\n", filename);
    }
    for (point = 0; point < points; point++) {
	fwrite(trc++, sizeof(long), 1, fp);
    }
    fclose(fp);
}
#endif


/*******************************************************************************
 *
 * i_aieAnalyzeRGB
 *
 *	Do default analysis of an RGB image. A fourth image is passed in, 
 *	which should contain the Y (luminance) channel.
 *
 * Arguments:
 *	pDefaults =	    The set of AIE default values
 *	pRed
 *	pGreen
 *	pBlue =		    Pointers to the color planes
 *	pLuminance =	    Pointer to the luminance plane
 *	Bpl =		    Bytes per line for color/luminance planes
 *	w, h =		    Size of images
 *	ppRedTRC =	    TRC computed which corrects exposure, contrast, color
 *	ppGreenTRC =	    TRC computed which corrects exposure, contrast, color
 *	ppBlueTRC =	    TRC computed which corrects exposure, contrast, color
 *	pSharpness =	    Sharpness value computed by sharpness analyzer
 *
 * Returns:
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_aieAnalyzeRGB (
    IPState    *pIPState,
    aAieDefaults *pDefaults,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    UInt8      *pLuminance,
    Int32	Bpl,
    Int32	w,
    Int32	h,
    aTable    **ppRedTRC,
    aTable    **ppGreenTRC,
    aTable    **ppBlueTRC,
    Float64    *pSharpness)
{
    PROLOGUE;
    Int32	in[2], out[2], arr100[4], edgeFactorIndex, error = ia_aborted;
    aTable	*pGammaTRC, *pContrastTRC;
    Float64	satVal, whiteOff, blackOff, variance;
    UInt8	blackPoint, whitePoint;
    UInt8	*pScaledLuminance, *pScaledRed, *pScaledGreen, *pScaledBlue;
    Int32	 scaledW, scaledH;
    UInt32	uniqueColors = 0, uniqueGrays = 0;
    Int32	scaledX, scaledY;
    Int32	selectX, selectY;

    /* Linearly-interpolate some TRCs to initialize them */
    in[0] = 0;
    in[1] = 255;
    out[0] = 0;
    out[1] = 255;

    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], ppRedTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], ppGreenTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], ppBlueTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], &pGammaTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], &pContrastTRC);

    /* Compute interior image region to be used in each analysis,
     * note that this removes the PIXR frame, plus another frame
     * 5% in on each side. */
    ExceptGoto (i_aieScale (pIPState, pDefaults, pRed, w, h, Bpl, 
			    &pScaledRed, &scaledX, &scaledY,
			    &scaledW, &scaledH, &selectX, &selectY), fail);
    ExceptGoto (i_aieScale (pIPState, pDefaults, pGreen, w, h, Bpl, 
			    &pScaledGreen, &scaledX, &scaledY,
			    &scaledW, &scaledH, &selectX, &selectY), fail);
    ExceptGoto (i_aieScale (pIPState, pDefaults, pBlue, w, h, Bpl, 
			    &pScaledBlue, &scaledX, &scaledY,
			    &scaledW, &scaledH, &selectX, &selectY), fail);
    ExceptGoto (i_aieScale (pIPState, pDefaults, pLuminance, w, h, Bpl, 
			    &pScaledLuminance, &scaledX, &scaledY,
			    &scaledW, &scaledH, &selectX, &selectY), fail);

    /* Count the number of unique colors, and the number of unique grays */
    ExceptGoto (i_countColors (pIPState, pRed, pBlue, pGreen, w, h, Bpl, 257UL,
			       &uniqueColors), fail);

    ExceptGoto (i_countGrays (pIPState, pLuminance, w, h, Bpl,
			      &uniqueGrays), fail);

    /* Analyze sharpness */
    ExceptGoto (i_aieSharpness (pIPState, &(pDefaults->sharpness),
				pScaledLuminance, 
				scaledW, scaledH, Bpl, 
				&edgeFactorIndex, pSharpness), fail);

    IP_YIELD (abort);

    /* Analyze color balance */
    ExceptGoto (i_aieColorBalance (pIPState, &(pDefaults->color),
				   pScaledRed, pScaledGreen, pScaledBlue, 
				   scaledW, scaledH, Bpl, 
				   *ppRedTRC, *ppGreenTRC, *ppBlueTRC, 
				   &satVal, &whiteOff, &blackOff,
				   uniqueColors), fail);

#ifdef AIE_DEBUG
    dump_trc(*ppRedTRC, "ipc.colR.lut");
    dump_trc(*ppGreenTRC, "ipc.colG.lut");
    dump_trc(*ppBlueTRC, "ipc.colB.lut");
#endif

    IP_YIELD (abort);
    /* Analyze exposure */
    ExceptGoto (i_aieExposure (pIPState, &(pDefaults->exposure),
			       pScaledLuminance, scaledW, scaledH, Bpl,
			       pGammaTRC, &arr100[0],
			       &variance, NULL, uniqueColors), fail);

    IP_YIELD (abort);
    /* Apply gamma curve to the TRCs */
    ip_tableFold (pIPState, pGammaTRC, *ppRedTRC, *ppRedTRC);
    ip_tableFold (pIPState, pGammaTRC, *ppGreenTRC, *ppGreenTRC);
    ip_tableFold (pIPState, pGammaTRC, *ppBlueTRC, *ppBlueTRC);

    
    /* Analyze contrast */
    blackPoint = 255, whitePoint = 0;
    ExceptGoto (i_aieContrast (pIPState, &(pDefaults->contrast), 
			       pScaledLuminance, scaledW, scaledH, Bpl,
			       pGammaTRC, &arr100[0], variance,
			       whiteOff, blackOff, 
			       edgeFactorIndex, blackPoint,
			       whitePoint, pContrastTRC,
			       selectX, selectY,
			       uniqueColors, uniqueGrays),
		fail);

#ifdef AIE_DEBUG
    dump_trc(pGammaTRC, "ipc.gamma.lut");

    dump_trc(pContrastTRC, "ipc.contrast.lut");
#endif

    IP_YIELD (abort);
    /* Fold that contrast curve in */
    ip_tableFold (pIPState, pContrastTRC, *ppRedTRC, *ppRedTRC);
    ip_tableFold (pIPState, pContrastTRC, *ppGreenTRC, *ppGreenTRC);
    ip_tableFold (pIPState, pContrastTRC, *ppBlueTRC, *ppBlueTRC);


#ifdef AIE_DEBUG
    dump_trc(*ppRedTRC, "ipc.finalR.lut");
    dump_trc(*ppGreenTRC, "ipc.finalG.lut");
    dump_trc(*ppBlueTRC, "ipc.finalB.lut");
#endif

    /* Free up garbage */
    ip_tableFree (pIPState, pGammaTRC);
    ip_tableFree (pIPState, pContrastTRC);
		 
    NORMAL_RETURN;

abort:
    error = ia_aborted;
fail:
    /* Free up everything */
    ip_tableFree (pIPState, pGammaTRC);
    ip_tableFree (pIPState, pContrastTRC);
    ip_tableFree (pIPState, *ppRedTRC);
    ip_tableFree (pIPState, *ppGreenTRC);
    ip_tableFree (pIPState, *ppBlueTRC);
    *ppRedTRC = *ppGreenTRC = *ppBlueTRC = 0;

    return error;
}

/*******************************************************************************
 *
 * i_aieAnalyzeRGB_init
 *
 *	Banded AIE initialization.
 *
 * Arguments:
 *	pDefaults	-   The set of AIE default values
 *	w, h		-   Total size of the image to process
 *	bpl		-   Bytes/line in image
 *	ppData		-   The intermediate data structure.
 *
 * Returns:
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_aieAnalyzeRGB_init (
    IPState    *pIPState,
    aAieDefaults *pDefaults,
    Int32	w,
    Int32	h,
    Int32	bpl,
    aIntermediates **ppData)
{
    PROLOGUE;
    UInt8 *lum;

    /* Create the intermediate data structure */
    if (!(*ppData = (aIntermediates *) MALLOC (sizeof (aIntermediates))))
	return ia_nomem;

    (*ppData)->defaults = pDefaults;
    (*ppData)->w = w;
    (*ppData)->h = h;
    (*ppData)->bpl = bpl;
    (*ppData)->currentLine = 0;

    if( w == 0 || h == 0 || bpl == 0)
	(*ppData)->isSegs = cTrue;
    else (*ppData)->isSegs = cFalse;

    /* Compute the scaled (deframed) portion to get statistics from */
    i_aieScale (pIPState, pDefaults, NULL,
		w + 2 * cFrameBits, h + 2 * cFrameBits, bpl,
		&lum, &(*ppData)->scaledX, &(*ppData)->scaledY,
		&(*ppData)->scaledW, &(*ppData)->scaledH,
		&(*ppData)->selectX, &(*ppData)->selectY);

    /* Initialize the analyzers */
    i_aieSharpness_init (pIPState, *ppData);
    i_aieColorBalance_init (pIPState, *ppData);
    i_aieExposure_init (pIPState, *ppData, cFalse);
    i_aieContrast_init (pIPState, *ppData);
		 
    NORMAL_RETURN;
}

/***********************************************************************
 * 
 * i_aieAnalyzeRGB_data 
 *
 *	Shove some data past the AIE analyzers 
 *
 * Arguments:
 *	pData	    -	    Data structure returned by i_aieAnalyzeRGB_init
 *	pRed	    -	    Red plane band - points to frame
 *	pGreen	    -	    Green plane band - points to frame
 *	pBlue	    -	    Blue plane band - points to frame
 *	pLuminance  -	    Luminance channel - points to frame
 *	lines	    -	    Number of lines to process, WITHOUT frame
 *
 * Returns:
 *	status	    -	    Error status, 0=ok.
 ***********************************************************************/

Int32 CDECL 
i_aieAnalyzeRGB_data (
    IPState    *pIPState,
    aIntermediates *pData,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    UInt8      *pLuminance,
    Int32	lines)
{
    PROLOGUE;
    if( pData->isSegs == cTrue )
    {
	UInt8 *lum;
	pData->currentLine = 0;
    /* Compute the scaled (deframed) portion to get statistics from */
	i_aieScale (pIPState, pData->defaults, NULL, pData->w + 2 * cFrameBits, 
		    pData->h + 2 * cFrameBits, pData->bpl,
		    &lum, &pData->scaledX, &pData->scaledY,
		    &pData->scaledW, &pData->scaledH,
		    &pData->selectX, &pData->selectY);
	pData->selectX = pData->selectY = 1;
	pData->localWidth = pData->scaledW;
	pData->localHeight = pData->scaledH;
    }

    /* Give that data to each analyzer */
    ExceptPass (i_aieSharpness_data (pIPState, pData, pLuminance, lines));
    ExceptPass (i_aieColorBalance_data (pIPState, pData,
					pRed, pGreen, pBlue, lines));
    ExceptPass (i_aieExposure_data (pIPState, pData, pLuminance, lines));
    ExceptPass (i_aieContrast_data (pIPState, pData, pLuminance, lines));

    /* Update the line counter */
    pData->currentLine += lines;

    NORMAL_RETURN;
}

/***********************************************************************
 * 
 * i_aieAnalyzeRGB_finish 
 *
 *	We've passed all of the data through; finish up the analysis
 *	and compute what we need, then free up data structures
 *
 * Arguments:
 *	pData	    -	    Data structure returned by i_aieAnalyzeRGB_init
 *	ppRedTRC    -	    TRC computed which corrects exposure, contrast, color
 *	ppGreenTRC  -	    TRC computed which corrects exposure, contrast, color
 *	ppBlueTRC   -	    TRC computed which corrects exposure, contrast, color
 *	pSharpness -	    Sharpness value computed by sharpness analyzer
 *
 * Returns:
 *	ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
i_aieAnalyzeRGB_finish (
    IPState    *pIPState,
    aIntermediates **ppData,
    aTable    **ppRedTRC,
    aTable    **ppGreenTRC,
    aTable    **ppBlueTRC,
    Float64    *pSharpness)
{
    PROLOGUE;
    Int32	in[2], out[2], arr100[4], edgeFactorIndex,
	        error = ia_aborted;
    aTable	*pGammaTRC = NULL, *pContrastTRC = NULL;
    aIntermediates *pData;
    Float64	satVal, whiteOff, blackOff, variance;

    pData = *ppData;
    /* Linearly-interpolate some TRCs to initialize them */
    in[0] = 0;
    in[1] = 255;
    out[0] = 0;
    out[1] = 255;

    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], ppRedTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], ppGreenTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], ppBlueTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], &pGammaTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], &pContrastTRC);

    /* Analyze sharpness */
    ExceptGoto (i_aieSharpness_finish (pIPState, pData,
				       &edgeFactorIndex, pSharpness), 
		fail);

    /* Analyze color balance */
    ExceptGoto (i_aieColorBalance_finish (pIPState, pData, 
					  *ppRedTRC, *ppGreenTRC, *ppBlueTRC, 
					  &satVal, &whiteOff, &blackOff),
		fail);

#ifdef AIE_DEBUG
    dump_trc(*ppRedTRC, "ipcb.colR.lut");
    dump_trc(*ppGreenTRC, "ipcb.colG.lut");
    dump_trc(*ppBlueTRC, "ipcb.colB.lut");
#endif

    /* Analyze exposure */
    ExceptGoto (i_aieExposure_finish (pIPState, pData, pGammaTRC, &arr100[0],
				      &variance, NULL), fail);

    /* Apply gamma curve to the TRCs */
    ip_tableFold (pIPState, pGammaTRC, *ppRedTRC, *ppRedTRC);
    ip_tableFold (pIPState, pGammaTRC, *ppGreenTRC, *ppGreenTRC);
    ip_tableFold (pIPState, pGammaTRC, *ppBlueTRC, *ppBlueTRC);

    
    /* Analyze contrast */
    ExceptGoto (i_aieContrast_finish (pIPState, pData,pGammaTRC, 
				      &arr100[0], variance, 
				      whiteOff, blackOff, 
				      edgeFactorIndex, pContrastTRC), fail);

#ifdef AIE_DEBUG
    dump_trc(pGammaTRC, "ipcb.gamma.lut");

    dump_trc(pContrastTRC, "ipcb.contrast.lut");
#endif

    IP_YIELD (abort);
    /* Fold that contrast curve in */
    ip_tableFold (pIPState, pContrastTRC, *ppRedTRC, *ppRedTRC);
    ip_tableFold (pIPState, pContrastTRC, *ppGreenTRC, *ppGreenTRC);
    ip_tableFold (pIPState, pContrastTRC, *ppBlueTRC, *ppBlueTRC);

#ifdef AIE_DEBUG
    dump_trc(*ppRedTRC, "ipcb.finalR.lut");
    dump_trc(*ppGreenTRC, "ipcb.finalG.lut");
    dump_trc(*ppBlueTRC, "ipcb.finalB.lut");
#endif

    /* Free up garbage */
    ip_tableFree (pIPState, pGammaTRC);
    ip_tableFree (pIPState, pContrastTRC);

    /* Free up the intermediates thing itself */
    FREE(pData);
    *ppData = NULL;

    NORMAL_RETURN;

abort:
    error = ia_aborted;
fail:
    /* Free up garbage */
    if (pGammaTRC)  ip_tableFree (pIPState, pGammaTRC);
    if (pContrastTRC) ip_tableFree (pIPState, pContrastTRC);

    return error;
}


/******************************************************************************
 * i_aieAnalyzeGray
 *
 *	Do default analysis of grayscale image. 
 *
 * Arguments:
 *	pDefaults =	    The set of AIE default values
 *	pLuminance =	    Pointer to the luminance plane
 *	Bpl =		    Bytes per line for luminance plane
 *	w, h =		    Size of images
 *	ppTRC =		    TRC computed which corrects exposure, contrast
 *	pSharpness =	    Sharpness value computed by sharpness analyzer
 *
 * Returns:
 *	ia_successful, ...
 *****************************************************************************/

Int32 CDECL 
i_aieAnalyzeGray (
    IPState    *pIPState,
    aAieDefaults *pDefaults,
    UInt8      *pLuminance,
    Int32	Bpl,
    Int32	w,
    Int32	h,
    aTable    **ppTRC,
    Float64    *pSharpness)
{
    PROLOGUE;
    Int32	 in[2], out[2], arr100[4], edgeFactorIndex, error = ia_aborted;
    aTable	*pGammaTRC, *pContrastTRC;
    Float64	 variance;
    UInt8	*pScaledLuminance;
    Int32	 scaledW, scaledH;
    UInt8	black, white;
    UInt32	uniqueGrays = 0;
    Int32	scaledX, scaledY;
    Int32	selectX, selectY;

    /* Linearly-Interpolate some TRCs to initialize them */
    in[0] = 0;
    in[1] = 255;
    out[0] = 0;
    out[1] = 255;

    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], ppTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], &pGammaTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], &pContrastTRC);

    /* Compute interior image region to be used in each analysis,
     * note that this removes the PIXR frame, plus another frame
     * 5% in on each side. */
    ExceptGoto (i_aieScale (pIPState, pDefaults, pLuminance, w, h, Bpl, 
			    &pScaledLuminance, &scaledX, &scaledY,
			    &scaledW, &scaledH, &selectX, &selectY), fail);

    /* Count the number of unique grays */
    ExceptGoto (i_countGrays (pIPState, pLuminance, w, h, Bpl, &uniqueGrays),
		fail);

    /* Analyze sharpness */
    ExceptGoto (i_aieSharpness (pIPState, &(pDefaults->sharpness),
				pScaledLuminance, 
				scaledW, scaledH, Bpl, 
				&edgeFactorIndex, pSharpness), fail);

    IP_YIELD (abort);

    /* Analyze exposure */
    ExceptGoto (i_aieExposure (pIPState, &(pDefaults->exposure),
			       pScaledLuminance, scaledW, scaledH,
			       Bpl, pGammaTRC, &arr100[0],
			       &variance, NULL, uniqueGrays), fail);

    IP_YIELD (abort);

    /* Apply gamma curve to the TRC */
    ip_tableFold (pIPState, pGammaTRC, *ppTRC, *ppTRC);
    
    IP_YIELD (abort);

    /* Analyze black-point and white-point */
    ExceptGoto (i_aieWhitepoint (pIPState, pScaledLuminance, scaledW, scaledH,
				 Bpl, &black, &white), fail);

    IP_YIELD (abort);

    /* Analyze contrast */
    ExceptGoto (i_aieContrast (pIPState, &(pDefaults->contrast), 
			       pScaledLuminance, scaledW, scaledH,
			       Bpl, pGammaTRC, &arr100[0], variance,
			       0.0, 0.0, edgeFactorIndex, black,
			       white, pContrastTRC,
			       selectX, selectY,
			       0, uniqueGrays),
		fail);

    IP_YIELD (abort);

    /* Fold that contrast curve in */
    ip_tableFold (pIPState, pContrastTRC, *ppTRC, *ppTRC);


#ifdef AIE_DEBUG
    dump_trc(pGammaTRC, "ipc.gamma.lut");

    dump_trc(pContrastTRC, "ipc.contrast.lut");

    dump_trc(*ppTRC, "ipc.final.lut");
#endif

    /* Free up garbage */
    ip_tableFree (pIPState, pGammaTRC);
    ip_tableFree (pIPState, pContrastTRC);

    NORMAL_RETURN;

abort:
    error = ia_aborted;
fail:
    /* Free up everything */
    ip_tableFree (pIPState, pGammaTRC);
    ip_tableFree (pIPState, pContrastTRC);
    ip_tableFree (pIPState, *ppTRC);
    *ppTRC = 0;

    return error;
}

/******************************************************************************
 * function:
 *	i_aieScale ()
 *
 * parameters:
 *	pDef		Scale defaults
 *	pLuminance	Pointer to luminance image, including frame
 *	w, h		Image dimensions, including frame
 *	Bpl		Bytes per line of source image
 *	ppScaledLuminance	Resulting (non-framed) relevant image
 *	pScaledX	Resulting scaled X position
 *	pScaledY	Resulting scaled Y position
 *	pScaledW	Resulting scaled width
 *	pScaledH	Resulting scaled height
 *	pSelectX	Resulting number of tiles in X direction
 *	pSelectY	Resulting number of tiles in Y direction
 *
 * returns:
 *	ia_success, ia_nomem, ...
 *
 * purpose:
 *	to compute the image region to be analyzed by the other AIE
 *	routines (sharpness, color balance, and contrast). Current to
 *	Sandpiper 4.0, an interior region is used for these routines,
 *	and the exterior 5% of image is ignored.
 *****************************************************************************/

Int32 CDECL 
i_aieScale (
    IPState    *pIPState,
    aAieDefaults *pDef,
    UInt8      *pLuminance,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    UInt8     **ppScaledLuminance,
    Int32      *pScaledX,
    Int32      *pScaledY,
    Int32      *pScaledW,
    Int32      *pScaledH,
    Int32      *pSelectX,
    Int32      *pSelectY)
{
    PROLOGUE;
    Int32	x, y;
    aScaleDefaults	def;
    Int32	selectX, selectY;
    
    def = pDef->scale;

    /* Remove the PIXR frame */
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

    /* Compute the autocrop region */
    x = (w * def.autocrop_left_pct + 50) / 100;
    y = (h * def.autocrop_top_pct + 50) / 100;
    w = (w * def.autocrop_width_pct + 50) / 100;
    h = (h * def.autocrop_height_pct + 50) / 100;
    if (w < 512 && h < 512) {
	selectX = selectY = 1;
    } else {
	selectX = pDef->contrast.selectX;
	selectY = pDef->contrast.selectY;
    }

    /* Return the results */
    *ppScaledLuminance = pLuminance + (cFrameBits + y) * Bpl + cFrameBits + x;
    *pScaledX = x;
    *pScaledY = y;
    *pScaledW = w;
    *pScaledH = h;
    *pSelectX = selectX;
    *pSelectY = selectY;

    NORMAL_RETURN;
}

/*******************************************************************************
 *
 * i_aieAnalyzeGray_init
 *
 *	Banded AIE initialization.
 *
 * Arguments:
 *	pDefaults	-   The set of AIE default values
 *	w, h		-   Total size of the image to process
 *	bpl		-   Bytes/line in image
 *	ppData		-   The intermediate data structure.
 *
 * Returns:
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_aieAnalyzeGray_init (
    IPState    *pIPState,
    aAieDefaults *pDefaults,
    Int32	w,
    Int32	h,
    Int32	bpl,
    aIntermediates **ppData)
{
    PROLOGUE;
    UInt8 *lum;

    /* Create the intermediate data structure */
    if (!(*ppData = (aIntermediates *) MALLOC (sizeof (aIntermediates))))
	return ia_nomem;

    (*ppData)->defaults = pDefaults;
    (*ppData)->w = w;
    (*ppData)->h = h;
    (*ppData)->bpl = bpl;
    (*ppData)->currentLine = 0;

    if( w == 0 || h == 0 || bpl == 0)
	(*ppData)->isSegs = cTrue;
    else (*ppData)->isSegs = cFalse;

    /* Compute the scaled (deframed) portion to get statistics from */
    i_aieScale (pIPState, pDefaults, NULL,
		w + 2 * cFrameBits, h + 2 * cFrameBits, bpl,
		&lum, &(*ppData)->scaledX, &(*ppData)->scaledY,
		&(*ppData)->scaledW, &(*ppData)->scaledH,
		&(*ppData)->selectX, &(*ppData)->selectY);

    /* Initialize the analyzers */
    i_aieSharpness_init (pIPState, *ppData);
    i_aieExposure_init (pIPState, *ppData, cFalse);
    i_aieContrast_init (pIPState, *ppData);
		 
    NORMAL_RETURN;
}

/***********************************************************************
 * 
 * i_aieAnalyzeGray_data 
 *
 *	Shove some data past the AIE analyzers 
 *
 * Arguments:
 *	pData	    -	    Data structure returned by i_aieAnalyzeGray_init
 *	pLuminance  -	    Luminance channel (ptr to frame)
 *	lines	    -	    Number of lines to process
 *
 * Returns:
 *	status	    -	    Error status, 0=ok.
 ***********************************************************************/

Int32 CDECL 
i_aieAnalyzeGray_data (
    IPState    *pIPState,
    aIntermediates *pData,
    UInt8      *pLuminance,
    Int32	lines)
{
    PROLOGUE;
    if( pData->isSegs == cTrue )
    {
	UInt8 *lum;
	pData->currentLine = 0;
    /* Compute the scaled (deframed) portion to get statistics from */
	i_aieScale (pIPState, pData->defaults, NULL, pData->w + 2 * cFrameBits, 
		    pData->h + 2 * cFrameBits, pData->bpl,
		    &lum, &pData->scaledX, &pData->scaledY,
		    &pData->scaledW, &pData->scaledH,
		    &pData->selectX, &pData->selectY);
	pData->selectX = pData->selectY = 1;
	pData->localWidth = pData->scaledW;
	pData->localHeight = pData->scaledH;
    }

    /* Give that data to each analyzer */
    ExceptPass (i_aieSharpness_data (pIPState, pData, pLuminance, lines));
    ExceptPass (i_aieExposure_data (pIPState, pData, pLuminance, lines));
    ExceptPass (i_aieContrast_data (pIPState, pData, pLuminance, lines));

    /* Update the line counter */
    pData->currentLine += lines;

    NORMAL_RETURN;
}

/***********************************************************************
 * 
 * i_aieAnalyzeGray_finish 
 *
 *	We've passed all of the data through; finish up the analysis
 *	and compute what we need, then free up data structures
 *
 * Arguments:
 *	pData	    -	    Data structure returned by i_aieAnalyzeGray_init
 *	ppTRC       -	    TRC computed which corrects exposure, contrast
 *	pSharpness  -	    Sharpness value computed by sharpness analyzer
 *
 * Returns:
 *	ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
i_aieAnalyzeGray_finish (
    IPState    *pIPState,
    aIntermediates **ppData,
    aTable    **ppTRC,
    Float64    *pSharpness)
{
    PROLOGUE;
    Int32	 in[2], out[2], arr100[4], edgeFactorIndex = 300, error = ia_aborted;
    aTable	*pGammaTRC = NULL, *pContrastTRC = NULL;
    aIntermediates *pData;
    Float64	 variance = 0.0;

    pData = *ppData;

    /* Linearly-interpolate some TRCs to initialize them */
    in[0] = 0;
    in[1] = 255;
    out[0] = 0;
    out[1] = 255;

    *pSharpness = 1.0;

    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], ppTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], &pGammaTRC);
    ip_tableInterpolate (pIPState, 2, &in[0], &out[0], &pContrastTRC);

    /* Analyze sharpness */
    ExceptGoto (i_aieSharpness_finish (pIPState, pData,
				       &edgeFactorIndex, pSharpness), 
		fail);


    /* Analyze exposure */
    ExceptGoto (i_aieExposure_finish (pIPState, pData, pGammaTRC, &arr100[0],
				      &variance, NULL), fail);

    /* Apply gamma curve to the TRC */
    ip_tableFold (pIPState, pGammaTRC, *ppTRC, *ppTRC);

    
    /* Analyze contrast */
    ExceptGoto (i_aieContrast_finish (pIPState, pData, pGammaTRC, 
				      &arr100[0], variance, 
				      0.0, 0.0, 
				      edgeFactorIndex, pContrastTRC), fail);

    /* Fold that contrast curve in */
    ip_tableFold (pIPState, pContrastTRC, *ppTRC, *ppTRC);

#ifdef AIE_DEBUG
    dump_trc(pGammaTRC, "ipcb.gamma.lut");

    dump_trc(pContrastTRC, "ipcb.contrast.lut");

    dump_trc(*ppTRC, "ipcb.final.lut");
#endif

    /* Free up garbage */
    ip_tableFree (pIPState, pGammaTRC);
    ip_tableFree (pIPState, pContrastTRC);

    FREE (pData);
    *ppData = NULL;
    NORMAL_RETURN;

fail:
    /* Free up garbage, but not ppData, because it could have been
       an error in allocating the resulting tables, which doesn't
       invalidate the ppData data.
     */
    if (pGammaTRC)  ip_tableFree (pIPState, pGammaTRC);
    if (pContrastTRC) ip_tableFree (pIPState, pContrastTRC);

    return error;
}

/***********************************************************************
 * 
 * i_aieDestroyTable()
 *
 *	Free a TRC table which was allocated by the AIE code
 *
 * Arguments:
 *	
 *	ppTable     -	    Table, pointer is cleared on exit
 *
 * Returns:
 *	status	    -	    Error status, 0=ok.
 ***********************************************************************/

Int32 CDECL 
i_aieDestroyTable (
    IPState    *pIPState,
    aTable    **ppTable)
{
    PROLOGUE;

    ERR_PASS( ip_tableFree (pIPState, *ppTable));
    *ppTable = NULL;

    NORMAL_RETURN;
}


/*

$Log: i_aieana.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.18  1998/01/13 23:46:52  ddavies
i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.17  1997/12/04 00:33:24  xzhu
xzhu on Wed Dec 3 16:31:02 PST 1997

i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v

Made analyze data also support segmented image. If input is segmented image, do scale again.

Revision 1.16  1997/11/14 18:22:17  aber
aber on Fri Nov 14 10:21:23 PST 1997

i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v

Occlude debug includes.

Revision 1.15  1997/11/13 23:21:11  aber
aber on Thu Nov 13 15:21:11 PST 1997

i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v

CommonCheck: Updated the AIE algorithms to match (or closely match) the latest DITC
AIE used in Sandpiper. The updates include the following:
1. Exclude a frame (5%) around the exterior in almost all statistics.
2. Make sure the black-point and white-point are included in the
   computations.
3. ... this was improved upon to not use just the first region with a
   black-or-white-point, but to use the ones with the largest variance.
4. The statistics are accumulated in cells 4 across and 4 down. This
   is not done if the width and height are less than 512.
5. Black-point and white-point do not include 0 and 255 respectively.
   They are the darkest and whitest not quite at these saturation points.
6. If the number of colors or gray levels is too low, less aggressive
   correction is applied.
7. Lots of thresholds changed.
8. Luminance channel is computed with a different formula.

Revision 1.14  1996/09/26 21:54:53  jcraig
jcraig on Thu Sep 26 14:53:38 PDT 1996

i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v

IP_YIELD now goes to another target which sets the return code
to ia_aborted and then goes through the cleanup code and returns.

Revision 1.13  1996/03/19 00:09:59  macleod
macleod on Mon Mar 18 16:09:41 PST 1996

i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v

Added i_aieDestroyTable, which allows one to free a table that AIE
analyzers made. This is a stop-gap, really--the memory allocation
in AIE either needs to be reworked, or perhaps aie should be taken
out of ipshared altogether.

Revision 1.12  1995/12/13 02:41:38  ejaquith
i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.11  1995/11/22  01:52:45  macleod
 * macleod on Tue Nov 21 17:52:31 PST 1995
 *
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * The termination routines for banded AIE analysis, i_aieAnalyzeRGB_finish
 * and i_aieAnalyzeGrayFinish weren't freeing the intermediate-results
 * structure that the initialization routines had created, so they
 * were modified to do so. Also, the pointer passed for the structure
 * was changed to a double-indirect, so that the pointer could be
 * zeroed on exit.
 *
 * Revision 1.10  1994/10/11  16:36:52  lperry
 * lperry on Tue Oct 11 09:35:39 PDT 1994
 *
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Added include of shros.pub to get IP_YIELD def
 *
 * Revision 1.9  1994/10/06  00:38:39  macleod
 * macleod on Wed Oct 5 17:38:16 PDT 1994
 *
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Eliminated an unreferenced parameter warning
 *
 * Revision 1.8  1994/10/05  22:20:08  macleod
 * macleod on Wed Oct 5 15:16:57 PDT 1994
 *
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Added support for banded AIE (analysis phase). Added three routines to each
 * analyzer, and six new routines to this file (three for RGB, three for gray).
 * For each analyzer, there's an "init", a "data", and a "finish" routine.
 * The "init" routine allocates stuff and sets up a data structure for holding
 * intermediate values needed for analysis. The "data" routine takes a band of
 * data, and does whatever data gathering is needed. The "finish" routine takes
 * the intermediate values and actually computes the analysis. The reason that
 * the analyzers had to be split like this is that the analysis must be run
 * in order, but one doesn't want to have to read through all of the bands more
 * than once, so one would prefer to pass each band to all of the analyzers in order.
 *
 * Revision 1.7  1994/07/12  22:22:23  ejaquith
 * ejaquith on Tue Jul 12 15:22:09 PDT 1994
 *
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Removed #include "iadef.prv"
 *
 * Revision 1.6  1994/06/23  05:59:36  macleod
 * macleod on Wed Jun 22 22:59:30 PDT 1994
 *
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Added some IP_YIELD calls
 *
 * Revision 1.5  1994/02/04  20:12:53  dferrell
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/21  23:24:41  macleod
 * macleod on Tue Dec 21 15:23:49 PST 1993
 *
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Added a parameter to i_aieExposure
 *
 * Revision 1.3  1993/12/02  19:22:04  ddavies
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:26:25  ddavies
 * i_aieana.c -> /product/ipcore/ipshared/src/RCS/i_aieana.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

