/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_aiecnt.c	
 *
 *	AIE Contrast Analysis module
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#ifndef __WATCOMC__
#ifndef __GNUC__
/* #include <memory.h> */
#endif
#endif
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"
#include "aiedefs.pub"
#include "aiedefs.prv"
#include "defines.pub"

/* Prototypes */
#include "shrip.pub"
#include "shrip.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_aiecnt.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

#define cHistogramMax         (32767)	/* max for scale factor */

/******************************************************************************
 * function:
 *	i_aieContrast ()
 *
 * parameters:
 *	pDef		Contrast enhancement defaults
 *	pLuminance	Pointer to luminance image, without a frame
 *	w, h		Image dimensions, without a frame
 *	Bpl		Bytes per line of source image
 *	pTRCGamma	The TRC that the exposure analyzer produced
 *	arr100[]	100% range that the exposure analyzer produced
 *	globalVariance	The variance computed by exposure analyzer
 *	whiteOff	% white was off, as computed by color analyzer
 *	blackOf		% black was off, as computed by color analyzer
 *	edgeFactorIndex	Index into sharpening table, as computed by
 *			sharpness analyzer
 *	blackPoint	black point
 *	whitePoint	white point
 *	pTRCContrast	Resulting contrast-enhancing TRC, which should already
 *			be allocated
 *	xBins		number of bins partitioned in X direction
 *	yBins		number of bins partitioned in Y direction
 *	uniqueColors	number of unique colors (if multichannel, 0 otherwise)
 *	uniqueGrays	number of unique grays (in luminance channel)
 *
 * returns:
 *	ia_success, ia_nomem, ...
 *
 * purpose:
 *	to look at both the global histogram and local region histograms
 *	to determine if there are enough valid regions to be used to
 *	compute a power value and variance factor.
 *	In English: 
 *****************************************************************************/

Int32 CDECL 
i_aieContrast (
    IPState    *pIPState,
    aContrastDefaults *pDef,
    UInt8      *pLuminance,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    aTable     *pTRCGamma,
    Int32	arr100[],
    Float64	globalVariance,
    Float64	whiteOff,
    Float64	blackOff,
    Int32	edgeFactorIndex,
    UInt8	blackPoint,
    UInt8	whitePoint,
    aTable     *pTRCContrast,
    Int32	xBins,
    Int32	yBins,
    UInt32	uniqueColors,
    UInt32	uniqueGrays)
{
    PROLOGUE;
    aContrastDefaults	def;

    Float64          factor, variance;
    Int32            i, j;
    Int32            localWidth, localHeight;
    Int32            m, n, mn;
    Int32            numValid;
    Int32            computeHistoTRC;
    aTable          *histoGamma = NULL;
    Int32            confidence;
    aTable	    *globalHistogram = NULL;
    aTable	    *histoUse = NULL;
    aTable	   **localHistograms = NULL;
    aHistStat	    *localHistogramStats = NULL;
    
    def = *pDef;
    globalVariance = 0;	/* Shut up WATCOM */
#ifdef PRINT_MESSAGE
    fprintf(stderr, "/********************/\n");
    fprintf(stderr, "/* contrast anlysis */\n");
    fprintf(stderr, "/********************/\n");
    fprintf(stderr, "variance = %f\n", globalVariance);
#endif

    /*
     *
     * some of this could be computed in exposure...
     * [though some of it gets trashed]
     */
    ERR_PASS( ip_tableAllocate (pIPState, 256, &globalHistogram));
    memset(ip_tableData (pIPState, globalHistogram), 0, 256 * sizeof(Int32));
    m = xBins;
    n = yBins;
    mn = m * n;

    /*
     * What's going on here is that a global histogram and mXn local
     * histograms (the picutre divided up into mXn rectangles) are
     * computed.
     */
    localWidth = (w + n - 1 ) / n;
    localHeight = (h + m - 1 ) / m;
    ExceptRaise (!(localHistograms = (aTable **) CALLOC(mn, sizeof(aTable *))), ia_nomem);
    ExceptRaise (!(localHistogramStats = (aHistStat *) CALLOC(mn, sizeof(aHistStat))), ia_nomem);

    for (i = 0; i < mn; i++)
    {
	ERR_PASS( ip_tableAllocate (pIPState, 256, &(localHistograms[i])));
	localHistogramStats[i].valid = 1;
    }

    /* Compute the global histogram */
    i_histoRect (pIPState, pLuminance, Bpl, 0, 0, w, h,
		 ip_tableData (pIPState, globalHistogram));

    /* Compute the local histograms. */
    for (i = 0; i < m; i++)
    {
	for (j = 0; j < n; j++)
	{
	    i_histoRect (pIPState, pLuminance, Bpl,
			 j * localWidth, i * localHeight,
			 MIN_1(localWidth, w - j * localWidth),
			 MIN_1(localHeight, h - i * localHeight),
			 ip_tableData (pIPState, localHistograms[i * m + j]));
	}
    }

    ip_doHistogramCompute (pIPState, (Float64) -1.0, 
			  (Float64) def.histogramVariance,
			  m, n, 
			  globalHistogram,
			  localHistograms, 
			  localHistogramStats, 
			  def.histogramMinValid,
			  blackPoint, whitePoint,
			  &variance
			  );

    if (variance > def.maxGlobalVariance)
    {
	numValid = 0;
	for (i = 0; i < m; i++)
	{
	    for (j = 0; j < n; j++)
	    {
		if (localHistogramStats[i * m + j].variance <=
		    def.maxLocal1Variance)
		    numValid++;
	    }
	}
	if (numValid >= def.histogramMinValid)
	{
	    factor = def.maxLocal1Variance / variance;
	    def.autoPower = def.power1;
	}
	else
	{
	    numValid = 0;
	    for (i = 0; i < m; i++)
	    {
		for (j = 0; j < n; j++)
		{
		    if (localHistogramStats[i * m + j].variance <=
			def.maxLocal2Variance)
			numValid++;
		}
	    }
	    if (numValid >= def.histogramMinValid)
	    {
		def.autoPower = def.power2;
		factor = def.maxLocal2Variance / variance;
	    }
	    else
	    {
		def.autoPower = def.power3;
		factor = def.factor3;
	    }
	}
	/*
	 * this recomputes the global histogram, hence extra output.
	 * we need to turn off debug, recompute, then reset debug.
	 */
	ip_doHistogramCompute (pIPState, variance, factor, m, n,
			      globalHistogram, localHistograms,
			      localHistogramStats, def.histogramMinValid,
			      blackPoint, whitePoint, &variance);
    }

    confidence = 0;

    /*
     * * build up some confidence that we have been here before...
     */
    if (arr100[3] > 230)
	confidence++;
    if (arr100[3] > 250)
	confidence++;
    if (edgeFactorIndex > 210)
	confidence++;
    if (whiteOff <= 0.01)
	confidence++;
    if (blackOff <= 0.01)
	confidence++;

    if (confidence >= 3)
    {
	def.autoPower = 0.0;
#ifdef PRINT_MESSAGE
	fprintf(stderr, "confidence level of %ld stretching histogram\n ", confidence);
#endif
    }
    else
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "confidence level of %ld too low\n", confidence);
#endif
    }

    computeHistoTRC = 1;
    if (uniqueColors == 0)	/* Was: if (ins->image_parts == 1) */
    {
	if (uniqueGrays <= 32)
	{
	    if (def.autoPower >= 0.0)
	    {
		def.autoPower = 0.0;
		computeHistoTRC = 0;
#ifdef DEBUG
		aie_message(globals, ins, "not enough colors, setting power to %.4lf\n", def.autoPower);
#endif
            }
	}
	else if (uniqueGrays <= 64)
	{
	    if (def.autoPower >= 0.1)
	    {
		def.autoPower = 0.1;
		computeHistoTRC = 0;
#ifdef DEBUG
		aie_message(globals, ins, "not enough colors, limit exceeded, setting power to %.4lf\n", def.autoPower);
#endif
	    }
	}
    }
    else
    {
	if (uniqueColors <= 128)
	{
	    /*
	    ** force not to compute histo_trc
	    */
	    computeHistoTRC = 0;
	    if (def.autoPower >= 0.0)
	    {
		def.autoPower = 0.0;
#ifdef DEBUG
		aie_message(globals, ins, "not enough colors, setting power to %.4lf\n", def.autoPower);
#endif
            }
	}
	else if (uniqueColors <= 256)
	{
	    /*
	    ** force not to compute histo_trc
	    */
	    computeHistoTRC = 0;
	    if (def.autoPower >= 0.1)
	    {
		def.autoPower = 0.1;
#ifdef DEBUG
		aie_message(globals, ins, "not enough colors, limit exceeded, setting power to %.4lf\n", def.autoPower);
#endif
            }
	}
    }

    if (computeHistoTRC)
    {
	Int32           *histp, prod;
	Int32            sums[256], *sumsp;
	Float64          scale;
	Int32            total;

	ip_tableApply (pIPState, globalHistogram, pTRCGamma, &histoGamma);

	/* first must scale it to 256 */
	ERR_PASS( ip_tableAllocate (pIPState, 256, &histoUse));
	ip_tableScale (pIPState, 2, histoGamma, histoUse);

	total = 0;
	histp = ip_tableData (pIPState, histoUse);

	for (i = 0; i < m; i++)
	{
	    for (j = 0; j < n; j++)
	    {
		if (localHistogramStats[i * n + j].valid)
		    total++;
	    }
	}

	if (total == 0)		/* if no select, use all */
	    total = (m * n);

	for (i = 0; i < 256; histp++, i++)
	    *histp = (Int32) ((*histp * (m * n)) / total);
	
	ip_tableSmooth (pIPState, def.enhanceAverage, histoUse);

	histp = ip_tableData (pIPState, histoUse);

	for (i = 0; i < 256; histp++, i++)
	{
	    if (*histp)
		*histp = (Int32) pow((Float64) *histp, def.autoPower);
	}

#ifdef PRINT_MESSAGE
	fprintf(stderr, "x ** %f \n", def.autoPower);
	fprintf(stderr, "histogram\n");

	for (i = 0; i < 256; ++i)
	{
	    if ((i != 0) && ((i % 16) == 0))
		fprintf(stderr, "\n");
	    fprintf(stderr, "%ld ", histoUse->table[i]);
	}

	fprintf(stderr, "\n");
#endif

	histp = ip_tableData (pIPState, histoUse);
	sums[0] = *histp;
	histp++;

	/* add each hist value to the last sum, place in sums array */
	for (i = 1; i < 256; i++)
	    sums[i] = sums[i - 1] + *histp++;

	/* compute scale factor for trc */
	scale = (Float64) (histoUse->length - 1) /
	    (sums[histoUse->length - 1] - sums[0]);

	for (i=255; i>=0; i--)
	{
	    sums[i] -= sums[0];
	}

	/* compute TRC */
	sumsp = sums;
#ifdef PRINT_MESSAGE
	fprintf(stderr, "trc\n");
#endif
	for (i = 0; i < histoUse->length; i++)
	{
	    prod = (Int32) (scale * (Float64) *sumsp + .5);
	    *sumsp++ = prod;
#ifdef PRINT_MESSAGE
	    if ((i != 0) && ((i % 16) == 0))
		fprintf(stderr, "\n");
	    fprintf(stderr, "%ld ", prod);
#endif
	}
#ifdef PRINT_MESSAGE
	fprintf(stderr, "\n");
#endif

	for (i = 0; i < 256; i++)
	{
	    pTRCContrast->table[i] = sums[i];
	}

	ip_tableFree (pIPState, histoUse);
	ip_tableFree (pIPState, histoGamma);

    }

    /* Free that which we've allocated */
    for (i = 0; i < mn; i++)
    {
	ip_tableFree (pIPState, localHistograms[i]);
    }

    ip_tableFree (pIPState, globalHistogram);
    FREE (localHistogramStats);
    FREE (localHistograms);

    NORMAL_RETURN;
}


/******************************************************************************
 * function:
 *	i_countColors ()
 *
 * parameters:
 *	pRed		Pointer to red channel, including frame
 *	pGreen		Pointer to green channel, including frame
 *	pBlue		Pointer to blue channel, including frame
 *	w, h		Image dimensions, including frame
 *	Bpl		Bytes per line of source image
 *	colorLimit	Maximum number of colors to search for
 *	pUniqueColors	Pointer to the number of unique colors found
 *
 * returns:
 *	ia_successful, ia_nomem, ...
 *
 * purpose:
 *	This function counts the number of unique colors found in the image.
 *	This function can be costly as the number may be large, since there
 *	are 16M colors possible in 3-channel 8-bit data. However, this is
 *	what the DITC contrast routine requires, so here it is. Since the AIE
 *	heuristics depend on whether 257 or more colors exist, they stop
 *	looking once 257 are found. It would be wise to limit the number,
 *	since the search can be very time costly.
 *	This function uses i_accumulateColors().
 *****************************************************************************/

Int32 CDECL 
i_countColors (
    IPState    *pIPState,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    UInt32	colorLimit,
    UInt32     *pUniqueColors)
{
    PROLOGUE;
    UInt32 rows, cols;
    UInt8 *redLine, *greenLine, *blueLine;
    UInt8 *colorList = NULL;
    Int32 error = 0;

    /* Zero out the return value */
    *pUniqueColors = 0;

    if (colorLimit == 0) {
	NORMAL_RETURN;
    }

    /* Remove the PIXR frame */
    cols = w - 2 * cFrameBits;
    rows = h - 2 * cFrameBits;
    redLine = pRed + cFrameBits * Bpl + cFrameBits;
    greenLine = pGreen + cFrameBits * Bpl + cFrameBits;
    blueLine = pBlue + cFrameBits * Bpl + cFrameBits;

    /* Allocate tables for colors previously seen. */
    ExceptRaise (!(colorList = 
		   (UInt8 *) CALLOC(colorLimit * 3, sizeof(UInt8))), ia_nomem);

    error = i_accumulateColors (pIPState, redLine, greenLine, blueLine,
				cols, rows,
				Bpl, colorLimit, colorList, pUniqueColors);

    /* Throw away the colors found */
    FREE(colorList);

    return error;
}

/******************************************************************************
 * function:
 *	i_accumulateColors ()
 *
 * parameters:
 *	pRed		Pointer to red channel raw data (NOT including frame)
 *	pGreen		Pointer to green channel raw data
 *	pBlue		Pointer to blue channel raw data
 *	w, h		Image dimensions of relevant data (NOT including frame)
 *	Bpl		Bytes per line of source image
 *	colorLimit	Maximum number of colors to search for
 *	colorList	Pointer to array of colors, will have space for
 *			'colorLimit' RGB triplets, and already be filled with
 *			'*pUniqueColors' colors
 *	pUniqueColors	Pointer to the number of unique colors found -- will
 *			also contain the number already populated in
 *			colorList[]
 *
 * returns:
 *	ia_successful, ...
 *
 * purpose:
 *	This function adds to the count of unique colors found in the image.
 *	This function can be costly as the number may be large, since there
 *	are 16M colors possible in 3-channel 8-bit data. However, a reasonable
 *	limit on the number of colors may tell a heuristic algorithm whether
 *	to do on thing or another. For example, if only 256 colors exist, then
 *	a conversion from 24-bit to 8-bit paletted will be lossless.
 *	This routine expects the parent routine to manage the memory in
 *	colorList[].
 *****************************************************************************/

Int32 CDECL 
i_accumulateColors (
    IPState    *pIPState,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    UInt32	colorLimit,
    UInt8      *colorList,
    UInt32     *pUniqueColors)
{
    PROLOGUE;
    UInt32 color, colorsFound = *pUniqueColors;
    UInt32 row, col, rows, cols;
    UInt8 *redLine, *greenLine, *blueLine;
    UInt8 r, g, b;
    Bool colorSeen;
    UInt8 *prevColor;

    /* Determine if we've previously reached the max */
    if (*pUniqueColors >= colorLimit) {
	NORMAL_RETURN;
    }

    /* Access the raw data */
    redLine = pRed;
    greenLine = pGreen;
    blueLine = pBlue;
    cols = w, rows = h;

    /* Loop through the image data, checking for a color match, and
     * adding the new colors to the list. Break out of the loop when
     * the max is hit. */
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    r = UCHAR_ACCESS(redLine + col);
	    g = UCHAR_ACCESS(greenLine + col);
	    b = UCHAR_ACCESS(blueLine + col);

	    /* Check if color previously seen */
	    colorSeen = 0;
	    prevColor = colorList;
	    for (color = 0; color < colorsFound; color++) {
		if (r == prevColor[0] && g == prevColor[1] &&
		    b == prevColor[2]) {
		    colorSeen = 1;
		    break;
		}
		prevColor += 3;
	    }

	    /* If not, add it */
	    if (colorSeen == 0) {
		colorList[colorsFound * 3] = r;
		colorList[colorsFound * 3 + 1] = g;
		colorList[colorsFound * 3 + 2] = b;
		if (++colorsFound >= colorLimit) {
		    goto doneSearching;
		}
	    }
	}
	redLine += Bpl;
	greenLine += Bpl;
	blueLine += Bpl;
    }

 doneSearching:

    /* Return the result */
    *pUniqueColors = colorsFound;

    NORMAL_RETURN;
}

/******************************************************************************
 * function:
 *	i_countGrays ()
 *
 * parameters:
 *	pLuminance	Pointer to the luminance image, including frame
 *	w, h		Image dimensions, including frame
 *	Bpl		Bytes per line of source image
 *	pUniqueGrays	Pointer to the number of unique gray values found
 *
 * returns:
 *	ia_successful, ia_nomem, ...
 *
 * purpose:
 *	This function counts the number of unique gray values found in the
 *	image. This function is much more reasonable than i_countColors()
 *	since an 8-bit LUT can be used.
 *****************************************************************************/

Int32 CDECL 
i_countGrays (
    IPState    *pIPState,
    UInt8      *pLuminance,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    UInt32     *pUniqueGrays)
{
    PROLOGUE;
    UInt32 col, rows, cols;
    UInt8 *line;
    UInt8 grayList[256];

    /* Zero out the return value */
    *pUniqueGrays = 0;

    /* Remove the PIXR frame */
    cols = w - 2 * cFrameBits;
    rows = h - 2 * cFrameBits;
    line = pLuminance + cFrameBits * Bpl + cFrameBits;

    /* Clear the Boolean table */
    for (col = 0; col < 256; col++) {
	grayList[col] = 0;
    }

    ERR_PASS( i_accumulateGrays (pIPState, line, cols, rows, Bpl,
			         grayList, pUniqueGrays));

    NORMAL_RETURN;
}


/******************************************************************************
 * function:
 *	i_accumulateGrays ()
 *
 * parameters:
 *	pLuminance	Pointer to the raw luminance data (NOT including frame)
 *	w, h		Image dimensions of raw data (NOt including frame)
 *	Bpl		Bytes per line of source image
 *	grayList	List of which gray values have occured (will be
 *			updated)
 *	pUniqueGrays	Pointer to the number of unique gray values found
 *
 * returns:
 *	ia_successful ...
 *
 * purpose:
 *	This function counts the number of unique gray values found in the
 *	image. This function is much more reasonable than i_accumulateColors()
 *	since an 8-bit LUT can be used.
 *****************************************************************************/

Int32 CDECL 
i_accumulateGrays (
    IPState    *pIPState,
    UInt8      *pLuminance,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    UInt8      *grayList,
    UInt32     *pUniqueGrays)
{
    PROLOGUE;
    UInt32 graysFound = *pUniqueGrays;
    UInt32 row, col, rows = h, cols = w;
    UInt8 *line;
    UInt8 intensity;

    /* Check to see if we've already found them all */
    if (*pUniqueGrays >= 256) {
	NORMAL_RETURN;
    }

    /* Loop through the data, and add to the list of grays seen */
    line = pLuminance;
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    intensity = UCHAR_ACCESS(line + col);

	    /* Check if gray value previously seen */
	    if (grayList[intensity] == 0) {

		/* If not, add it */
		grayList[intensity] = 1;
		if (++graysFound >= 256) {
		    goto finish;
		}
	    }
	}
	line += Bpl;
    }

 finish:

    /* Return the result */
    *pUniqueGrays = graysFound;

    NORMAL_RETURN;
}


/******************************************************************************
 * function:
 *	ip_doHistogramCompute()
 * parameters:
 *	g			normalized global variance
 *	vv			variance factor
 *	m, n			number of region in X,Y directions
 *	globalHistogram		the global histogram for the image
 *	localHistograms		the local histograms for each of mXn regions
 *	localHistogramStats	the histogram statistics for each local histo
 *	black			black point
 *	white			white point
 *	pVariance		the variance we compute
 *
 * purpose:
 *	use the normalized global variance and variance factor
 *	to determine if a region is valid.  the global histogram
 *	is then modified to contain the contents of the valid
 *	local histogram data.
 *******************************************************************************/

Int32 CDECL 
ip_doHistogramCompute (
    IPState    *pIPState,
    Float64	gn,
    Float64	vvv,
    Int32	m,
    Int32	n,
    aTable     *globalHistogram,
    aTable    **localHistograms,
    aHistStat  *localHistogramStats,
    Int32	minNumValid,
    UInt8	blackPoint,
    UInt8	whitePoint,
    Float64    *pVariance)
{
    PROLOGUE;
    Int32	i, j;
    Int32	mn = m * n;
    Float64	scale;
    Float64	temp;
    Float64	sum, sqsum;
    Float64	ttemp;
    Float64	ave, sigma2;
    Float64	variancefactor, variance;
    Int32	numValid = 0;
    Int32	minBlack, maxWhite;
    Float64	bestVariance;
    Int32	bestVarIndex;

    variancefactor = vvv;
    sum = sqsum = 0.0;
#ifdef PRINT_MESSAGE
    fprintf(stderr, "variance factor = %f\n", variancefactor);
#endif
    for (i = 0; i < 256; i++)
    {
	ttemp = (Float64) globalHistogram->table[i];
	sum += ttemp;
	sqsum += ttemp * ttemp;
    }
    ave = sum / (Float64) 256;
    sigma2 = sqsum / (Float64) 256 - ave * ave;
    sigma2 = sigma2 * (1024 * 1024) / (sum * sum);

    if (gn != -1.0)
	sigma2 = gn;

    variance = sigma2;
#ifdef PRINT_MESSAGE
    fprintf(stderr, "normalized global variance:%5ld\n\n", (Int32) sigma2);
#endif

    for (j = 0; j < mn; j++)
    {
	sum = sqsum = 0.0;

	for (i = 0; i < 256; i++)
	{
	    ttemp = (Float64) ((localHistograms[j])->table[i]);
	    sum += ttemp;
	    sqsum += ttemp * ttemp;
	}

	ave = sum / 256.0;
	sigma2 = sqsum / 256.0 - ave * ave;

	if (sum > 0.0)
	    sigma2 = sigma2 * (1024.0 * 1024.0) / (sum * sum);
	else
	    sigma2 = -1.0;

	localHistogramStats[j].variance = (Int32)sigma2;

	if ((sigma2 > (variance * variancefactor + 0.5)) || (sigma2 == -1.0))
	    localHistogramStats[j].valid = 0;
	else
	    localHistogramStats[j].valid = 1;

#ifdef PRINT_MESSAGE
	fprintf(stderr, " %5ld    ", (Int32) sigma2);
	if (j % n == (n - 1))
	    fprintf(stderr, "\n");
#endif

    }

    /*
     * If we haven't included a good dark and light region, add them in
     */

    /* Search for the blackest black */
    minBlack = 255;
    for (i = 0; i < mn; i++)
    {
	if (localHistogramStats[i].valid)
	{
	    for (j = 0; j < minBlack; j++)
	    {
		if (localHistograms[i]->table[j] > 0)
		{
		    minBlack = j;	/* this will cause j loop to break */
		}
	    }
	}
    }

    /* This deviates from the DITC routines. Their routines find the first
     * 16th region with the black and white points, and then oddly miscompute
     * position to alway be in the top quarter of the analyzed image.
     *
     * Here instead I search for the the region with a black point and keep
     * the one with the largest variance. Same for the white point */

    if (minBlack > blackPoint)
    {
	/* Search the invalid regions for one with a black-point. Keep the
	 * one with the largest variance */
	bestVarIndex = -1;
	bestVariance = 0.0;
	for (i = 0; i < mn; i++)
	{
	    if (localHistogramStats[i].valid == 0)
	    {
		for (j = 0; j <= blackPoint; j++)
		{
		    if (localHistograms[i]->table[j] > 0)
		    {
			if (localHistogramStats[i].variance > bestVariance)
			{
			    bestVariance = localHistogramStats[i].variance;
			    bestVarIndex = i;
			    break;
			}
		    }
		}
	    }
	}

	/* Mark this one valid too */
	if (bestVarIndex >= 0)
	{
	    localHistogramStats[bestVarIndex].valid = 1;
	}
    }

    /* With possible new region included, search for the brightest white */
    maxWhite = 0;
    for (i = 0; i < mn; i++)
    {
	if (localHistogramStats[i].valid)
	{
	    for (j = 255; j > maxWhite; j--)
	    {
		if (localHistograms[i]->table[j] > 0)
		{
		    maxWhite = j;	/* this will cause j loop to break */
		}
	    }
	}
    }

    if (maxWhite < whitePoint)
    {
	/* Search the invalid regions for one with a white-point. Keep the
	 * one with the largest variance */
	bestVarIndex = -1;
	bestVariance = 0.0;
	for (i = 0; i < mn; i++)
	{
	    if (localHistogramStats[i].valid == 0)
	    {
		for (j = 255; j >= whitePoint; j--)
		{
		    if (localHistograms[i]->table[j] > 0)
		    {
			if (localHistogramStats[i].variance > bestVariance)
			{
			    bestVariance = localHistogramStats[i].variance;
			    bestVarIndex = i;
			    break;
			}
		    }
		}
	    }
	}

	/* Mark this one valid too */
	if (bestVarIndex >= 0)
	{
	    localHistogramStats[bestVarIndex].valid = 1;
	}
    }
		
#ifdef PRINT_MESSAGE
    fprintf(stderr, "\n");

    for (i = 0; i < n; i++)
    {
	for (j = 0; j < m; j++)
	    fprintf(stderr, "%s", localHistogramStats[j + i * m].valid ? "[]" : "..");
	fprintf(stderr, "\n");
    }

    fprintf(stderr, "\n");
#endif

    for (i = 0; i < mn; i++)
    {
	if (localHistogramStats[i].valid)
	    numValid++;
    }

    if (numValid >= minNumValid)
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "numValid = %ld, using local histogram\n", numValid);
#endif
	memset(ip_tableData (pIPState, globalHistogram),
	       0, 256 * sizeof(Int32));
	for (i = 0; i < mn; i++)
	    if (localHistogramStats[i].valid)
		for (j = 0; j < 256; j++)
		    globalHistogram->table[j] += (localHistograms[i])->table[j];
    }
    else
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "numValid = %ld, expected %ld, using global histogram\n", numValid, minNumValid);
#endif
    }

    /* normalize globalHistogram array to fit 16 bit integers */
    temp = 0L;
    for (i = 0; i < 256; i++)
    {
	if (globalHistogram->table[i] > temp)
	    temp = globalHistogram->table[i];
    }

    scale = (Float64) cHistogramMax / (Float64) temp;

    if (scale < 1.0)
    {
	for (i = 0; i < 256; i++)
	{
	    /* Yep, you really have to do it this way to get the typecasting
	     * right for certain compilers.  Yick.
	     */
	    globalHistogram->table[i] = (Int32)(globalHistogram->table[i] * scale);
	}
    }

    *pVariance = variance;
    NORMAL_RETURN;
}


/*******************************************************************************
 * function:
 *	ip_tableSmooth()
 * parameters:
 *	avg_num		average amount
 *	hist		histogram
 * purpose:
 *	to average the histogram to make it appear continuous.
 *	this gives us better results after applying a histogram
 *	to a trc.
 *******************************************************************************/

Int32 CDECL 
ip_tableSmooth (
    IPState    *pIPState,
    Int32	avg_num,
    aTable     *hist)
{
    PROLOGUE;
    Int32           val;
    Int32           i, j;
    Int32           histo_tmp[256];
    Int32           half;

    half = avg_num / 2;
    for (i = 0; i < 256; i++)
    {
	val = 0;

	if (i >= half)
	{
	    if (i < (255 - half))
	    {
		for (j = 0; j < half; j++)
		    val += hist->table[i - half + j];

		/* half on each side of current index */
		if (avg_num & 0x01)
		{
		    for (j = half; j < avg_num; j++)
			val += hist->table[i - half + j];

		    histo_tmp[i] = val / (avg_num);
		}
		else
		{
		    for (j = half; j <= avg_num; j++)
			val += hist->table[i - half + j];

		    histo_tmp[i] = val / (avg_num + 1);
		}
	    }
	    else
	    {
		for (j = i - half; j < 256; j++)
		    val += hist->table[j];

		histo_tmp[i] = val / (j - (i - half));
	    }
	}
	else
	{
	    for (j = 0; j <= half + i; j++)
		val += hist->table[j];

	    histo_tmp[i] = val / (half + (j - half));
	}
    }

    for (i = 0; i < 256; i++)
	hist->table[i] = histo_tmp[i];

    NORMAL_RETURN;
}

/***************************************************************************
 * i_aieContrast_init (data)
 *
 *	Initialize banded contrast-analysis
 *
 * parameters:
 *	pData	    -	The "intermediate data values" structure, which should
 *			have the pointer to the "defaults" structure
 *			filled in
 *
 * returns:
 *	ia_success, ia_nomem, ...
 *
 * purpose:
 *  
 ***************************************************************************/

Int32 CDECL 
i_aieContrast_init (
    IPState    *pIPState,
    aIntermediates *pData)
{
    PROLOGUE;
    aContrastDefaults	def;

    Int32            i, j, w, h;
    Int32            m, n, mn;

    /* Don't really know why we make a copy of this.. */
    def = pData->defaults->contrast;
    w = pData->scaledW;	/* Note: scaled width */
    h = pData->scaledH;	/* ... and height */
    m = pData->selectX;
    n = pData->selectY;
    mn = m * n;

    /*
     * What's going on here is that a global histogram and mXn local
     * histograms (the picutre divided up into mXn rectangles) are
     * computed.
     */

    ERR_PASS( ip_tableAllocate (pIPState, 256, &(pData->globalHistogram)));
    pData->localWidth = (w + n - 1) / n;
    pData->localHeight = (h + m - 1) / m;
    ExceptRaise (!(pData->localHistograms = 
		   (aTable **) CALLOC(mn, sizeof(aTable *))), ia_nomem);

    for (i = 0; i < mn; i++)
    {
	ERR_PASS_CLEANUP(
		ip_tableAllocate (pIPState, 256, &(pData->localHistograms[i])));
    }

    pData->numGrays = pData->numColors = 0;
    memset(pData->grayList, 0, sizeof(pData->grayList));

    NORMAL_RETURN;

ERR_CLEANUP:
    for (j = 0; j < i; j++)
	ip_tableFree (pIPState, pData->localHistograms[j]);

    FREE (pData->localHistograms);

    ERR_RETURN;
}

/***********************************************************************
 *
 * i_aieContrast_data
 *
 *	Accumulate a band of data for the purpose of contrast
 *	analysis. 
 *
 * Arguments:
 *	pData	    -	    The "intermediate data" structure
 *	pLuminance  -	    Luminance channel data, pointer to frame
 *	lines	    -	    Number of lines to analyze
 * 
 * Returns: 
 *	ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
i_aieContrast_data (
    IPState    *pIPState,
    aIntermediates *pData,
    UInt8      *pLuminance,
    Int32	lines)
{
    PROLOGUE;
    Int32   i, j, m, n, hStart, hEnd, hLines, bStart, bEnd, px, py, pw, ph;
    Int32   w, bpl, currentLine;

    currentLine = pData->currentLine;
    w = pData->w;
    bpl = pData->bpl;
    m = pData->selectX;
    n = pData->selectY;
    bStart = currentLine;
    bEnd = currentLine + lines - 1;

    /* Add to the list of gray levels found in entire luminance channel */
    ExceptPass(i_accumulateGrays (pIPState, pLuminance, w, lines,
				  bpl, pData->grayList,
				  &pData->numGrays));

    /* Compute the relevant local histograms */
    for (i = 0; i < n; i++)
    {
	/* Do the lines passed overlap this row of histograms? */
	hStart = pData->scaledY + pData->localHeight * i;
	hEnd = hStart + MIN_1(pData->localHeight, \
			 pData->scaledH - i * pData->localHeight);

	if (bStart >= hStart && bStart <= hEnd)
	{
	    hStart = bStart;
	    hLines = MIN(bEnd, hEnd) - hStart + 1;
	}
	else if (bEnd >= hStart && bEnd <= hEnd)
	{
	    hLines = bEnd - hStart + 1;
	}
	else if (bStart <= hStart && bEnd >= hEnd)
	{
	    hLines = hEnd - hStart + 1;
	}
	else
	    hLines = 0;

	if (hLines)
	{
	    for (j = 0; j < m; j++)
	    {
		px = pData->scaledX + j * pData->localWidth;
		py = hStart - bStart;
		pw = MIN_1(pData->localWidth, \
			   pData->scaledW - j * pData->localWidth);
		ph = hLines;

		/* i_histoRect doesn't clear the histogram, so it
		   can be used for accumulating to the histogram */
		i_histoRect (pIPState, pLuminance, bpl, px, py, pw, ph,
			     ip_tableData (pIPState,
					   pData->localHistograms[i * m + j]));
	    }
	}
    }

    if (currentLine < pData->scaledY) {
	py = pData->scaledY - currentLine;
	lines -= pData->scaledY - currentLine;
	currentLine = pData->scaledY;
    } else {
	py = 0;
    }

    if (currentLine + lines > pData->scaledY + pData->scaledH) {
	lines = pData->scaledY + pData->scaledH - currentLine;
    }

    if (lines > 0) {

	/* i_histo doesn't clear the histogram first, so we can use it
	 * to accumulate histograms */
	i_histoRect (pIPState, pLuminance, pData->bpl,
		     pData->scaledX, py, pData->scaledW,
		     lines, ip_tableData (pIPState, pData->globalHistogram));
    }

    NORMAL_RETURN;
}

/*******************************************************************************
 * i_aieContrast_finish ()
 *
 *	We've gathered the statistics, now finish processing and produce
 *	a TRC to correct for contrast. Then, free data structures in
 *	the "intermediates" data structure that this analyzer allocated.
 *	Looks at both the global histogram and local region histograms
 *	to determine if there are enough valid regions to be used to
 *	compute a power value and variance factor.
 *
 * Arguments:
 *	pDdata	    -	Intermediates structure returned by .._init
 *	pTRCGamma   -	The TRC that the exposure analyzer produced
 *	arr100[]    -	100% range that the exposure analyzer produced
 *	globalVariance -The variance computed by exposure analyzer
 *	whiteOff    -	% white was off, as computed by color analyzer
 *	blackOff    -	% black was off, as computed by color analyzer
 *	edgeFactorIndex-Index into sharpening table, as computed by sharpness
 *			analyzer
 *	pTRCContrast -	Resulting contrast-enhancing TRC, which should already
 *			be allocated
 *
 * returns:
 *	ia_success, ia_nomem, ...
 *
 *******************************************************************************/

Int32 CDECL 
i_aieContrast_finish (
    IPState    *pIPState,
    aIntermediates *pData,
    aTable     *pTRCGamma,
    Int32	arr100[],
    Float64	globalVariance,
    Float64	whiteOff,
    Float64	blackOff,
    Int32	edgeFactorIndex,
    aTable     *pTRCContrast)
{
    PROLOGUE;
    aContrastDefaults	def;

/*    Float64        minStatVariance = 0.0;*/
    Float64          factor, variance;
    Int32            i, j;
    Int32            m, n, mn;
    Int32            numValid;
    Int32            computeHistoTRC;
    aTable          *histoGamma;
    Int32            confidence;
    aTable	    *histoUse;
    aHistStat	    *localHistogramStats;
    Int32	     w, h;
    UInt8	     blackPoint = 0, whitePoint = 255;
    
    def = pData->defaults->contrast;
    globalVariance = 0;	/* Shut up WATCOM */
#ifdef PRINT_MESSAGE
    fprintf(stderr, "/********************/\n");
    fprintf(stderr, "/* contrast anlysis */\n");
    fprintf(stderr, "/********************/\n");
    fprintf(stderr, "variance = %f\n", globalVariance);
#endif

    m = pData->selectX;
    n = pData->selectY;
    mn = m * n;
    w = pData->w;
    h = pData->h;

    ExceptRaise (!(localHistogramStats = (aHistStat *) CALLOC(mn, sizeof(aHistStat))), ia_nomem);

    for (i = 0; i < mn; i++)
    {
	localHistogramStats[i].valid = 1;
    }

    ip_doHistogramCompute (pIPState, (Float64) -1.0,
			  (Float64) def.histogramVariance,
			  m, n,
			  pData->globalHistogram,
			  pData->localHistograms,
			  localHistogramStats,
			  def.histogramMinValid,
			  blackPoint, whitePoint,
			  &variance
			  );

    if (variance > def.maxGlobalVariance)
    {
	numValid = 0;
	for (i = 0; i < m; i++)
	{
	    for (j = 0; j < n; j++)
	    {
		if (localHistogramStats[i * m + j].variance <= def.maxLocal1Variance)
		    numValid++;
	    }
	}
	if (numValid >= def.histogramMinValid)
	{
	    factor = def.maxLocal1Variance / variance;
	    def.autoPower = def.power1;
	}
	else
	{
	    numValid = 0;
	    for (i = 0; i < m; i++)
	    {
		for (j = 0; j < n; j++)
		{
		    if (localHistogramStats[i * m + j].variance <=
			def.maxLocal2Variance)
			numValid++;
		}
	    }
	    if (numValid >= def.histogramMinValid)
	    {
		def.autoPower = def.power2;
		factor = def.maxLocal2Variance / variance;
	    }
	    else
	    {
		def.autoPower = def.power3;
		factor = def.factor3;
	    }
	}
	/*
	 * this recomputes the global histogram, hence extra output.
	 * we need to turn off debug, recompute, then reset debug.
	 */
	ip_doHistogramCompute (pIPState, variance, factor, m, n,
			       pData->globalHistogram, pData->localHistograms, 
			       localHistogramStats, def.histogramMinValid,
			       blackPoint, whitePoint,
			       &variance);
    }

    confidence = 0;

    /*
     * * build up some confidence that we have been here before...
     */
    if (arr100[3] > 230)
	confidence++;
    if (arr100[3] > 250)
	confidence++;
    if (edgeFactorIndex > 210)
	confidence++;
    if (whiteOff <= 0.01)
	confidence++;
    if (blackOff <= 0.01)
	confidence++;
    if (confidence >= 3)
    {
	def.autoPower = 0.0;
#ifdef PRINT_MESSAGE
	fprintf(stderr, "confidence level of %ld stretching histogram\n ", confidence);
#endif
    }
    else
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "confidence level of %ld too low\n", confidence);
#endif
    }

    computeHistoTRC = 1;
    if (pData->numColors == 0)	/* Was: if (ins->image_parts == 1) */
    {
	if (pData->numGrays <= 32)
	{
	    if (def.autoPower >= 0.0)
	    {
		def.autoPower = 0.0;
		computeHistoTRC = 0;
#ifdef DEBUG
		aie_message(globals, ins, "not enough colors, setting power to %.4lf\n", def.autoPower);
#endif
            }
	}
	else if (pData->numGrays <= 64)
	{
	    if (def.autoPower >= 0.1)
	    {
		def.autoPower = 0.1;
		computeHistoTRC = 0;
#ifdef DEBUG
		aie_message(globals, ins, "not enough colors, limit exceeded, setting power to %.4lf\n", def.autoPower);
#endif
	    }
	}
    }
    else
    {
	if (pData->numColors <= 128)
	{
	    /*
	    ** force not to compute histo_trc
	    */
	    computeHistoTRC = 0;
	    if (def.autoPower >= 0.0)
	    {
		def.autoPower = 0.0;
#ifdef DEBUG
		aie_message(globals, ins, "not enough colors, setting power to %.4lf\n", def.autoPower);
#endif
            }
	}
	else if (pData->numColors <= 256)
	{
	    /*
	    ** force not to compute histo_trc
	    */
	    computeHistoTRC = 0;
	    if (def.autoPower >= 0.1)
	    {
		def.autoPower = 0.1;
#ifdef DEBUG
		aie_message(globals, ins, "not enough colors, limit exceeded, setting power to %.4lf\n", def.autoPower);
#endif
            }
	}
    }

    if (computeHistoTRC)
    {
	Int32           *histp, prod;
	Int32            sums[256], *sumsp;
	Float64          scale;
	Int32            total;

	ip_tableApply (pIPState, pData->globalHistogram,
		       pTRCGamma, &histoGamma);

	/* first must scale it to 256 */
	ERR_PASS( ip_tableAllocate (pIPState, 256, &histoUse));
	ip_tableScale (pIPState, 2, histoGamma, histoUse);

	total = 0;
	histp = ip_tableData (pIPState, histoUse);

	for (i = 0; i < m; i++)
	{
	    for (j = 0; j < n; j++)
	    {
		if (localHistogramStats[i * n + j].valid)
		    total++;
	    }
	}

	if (total == 0)		/* if no select, use all */
	    total = (m * n);

	for (i = 0; i < 256; histp++, i++)
	    *histp = (Int32) ((*histp * (m * n)) / total);

	ip_tableSmooth (pIPState, def.enhanceAverage, histoUse);
	histp = ip_tableData (pIPState, histoUse);

	for (i = 0; i < 256; histp++, i++)
	{
	    if (*histp)
		*histp = (Int32) pow((Float64) *histp, def.autoPower);
	}

#ifdef PRINT_MESSAGE
	fprintf(stderr, "x ** %f \n", def.autoPower);
	fprintf(stderr, "histogram\n");

	for (i = 0; i < 256; ++i)
	{
	    if ((i != 0) && ((i % 16) == 0))
		fprintf(stderr, "\n");
	    fprintf(stderr, "%ld ", histoUse->table[i]);
	}

	fprintf(stderr, "\n");
#endif

	histp = ip_tableData (pIPState, histoUse);
	sums[0] = *histp;
	histp++;

	/* add each hist value to the last sum, place in sums array */
	for (i = 1; i < 256; i++)
	    sums[i] = sums[i - 1] + *histp++;

	/* compute scale factor for trc */
	scale = (Float64) (histoUse->length - 1) /
	    (sums[histoUse->length - 1] - sums[0]);

	for (i=255; i>=0; i--)
	{
	    sums[i] -= sums[0];
	}

	/* compute TRC */
	sumsp = sums;
#ifdef PRINT_MESSAGE
	fprintf(stderr, "trc\n");
#endif
	for (i = 0; i < histoUse->length; i++)
	{
	    prod = (Int32) (scale * (Float64) *sumsp + .5);
	    *sumsp++ = prod;
#ifdef PRINT_MESSAGE
	    if ((i != 0) && ((i % 16) == 0))
		fprintf(stderr, "\n");
	    fprintf(stderr, "%ld ", prod);
#endif
	}
#ifdef PRINT_MESSAGE
	fprintf(stderr, "\n");
#endif

	for (i = 0; i < 256; i++)
	{
	    pTRCContrast->table[i] = sums[i];
	}

	ip_tableFree (pIPState, histoGamma);
	ip_tableFree (pIPState, histoUse);
    }

    /* Free that which we've allocated */
    for (i = 0; i < mn; i++)
    {
	ip_tableFree (pIPState, pData->localHistograms[i]);
    }

    /* NOTE that the global histogram is shared by the exposure module--
       we compute it in this one, so we free it here */
    ip_tableFree (pIPState, pData->globalHistogram);
    FREE (localHistogramStats);
    FREE (pData->localHistograms);

    NORMAL_RETURN;
}


/*

$Log: i_aiecnt.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.16  1998/04/28 19:47:34  aber
aber on Tue Apr 28 12:46:51 PDT 1998

i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v

Bug fix to histogram computation on multiple segments beyond the first one.

Revision 1.15  1998/01/13 23:46:54  ddavies
i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.14  1997/11/15 19:11:47  aber
aber on Sat Nov 15 11:11:34 PST 1997

i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v

Initialize new variable.

Revision 1.13  1997/11/13 23:21:13  aber
i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v

CommonCheck: Updated the AIE algorithms to match (or closely match) the latest DITC
AIE used in Sandpiper. The updates include the following:
1. Exclude a frame (5%) around the exterior in almost all statistics.
2. Make sure the black-point and white-point are included in the
   computations.
3. ... this was improved upon to not use just the first region with a
   black-or-white-point, but to use the ones with the largest variance.
4. The statistics are accumulated in cells 4 across and 4 down. This
   is not done if the width and height are less than 512.
5. Black-point and white-point do not include 0 and 255 respectively.
   They are the darkest and whitest not quite at these saturation points.
6. If the number of colors or gray levels is too low, less aggressive
   correction is applied.
7. Lots of thresholds changed.
8. Luminance channel is computed with a different formula.

Revision 1.12  1997/07/03 21:52:09  jcraig
jcraig on Thu Jul 3 14:51:44 PDT 1997

i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v

use i_histoRect instead of i_histo.

Revision 1.11  1995/12/13 02:41:42  ejaquith
i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.10  1995/10/13  23:49:03  dferrell
 * i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.9  1994/10/05  22:20:16  macleod
 * i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v
 *
 * Added support for banded aie
 *
 * Revision 1.8  1994/07/12  22:22:58  ejaquith
 * ejaquith on Tue Jul 12 15:22:47 PDT 1994
 *
 * i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v
 *
 * Removed #include "iadef.prv"
 *
 * Revision 1.7  1994/07/12  00:48:59  macleod
 * macleod on Mon Jul 11 17:48:52 PDT 1994
 *
 * i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v
 *
 * Fixed a couple of memory leaks
 *
 * Revision 1.6  1994/02/04  20:12:59  dferrell
 * i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.5  1994/01/22  03:48:36  ddavies
 * i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v
 *
 * Commented out "include <memory.h>" since no compiler seemed to actually
 * use it and mkmf on the PC issued a warning since it couldn't find it.
 *
 * Revision 1.4  1993/12/02  19:22:08  ddavies
 * i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:26:33  ddavies
 * i_aiecnt.c -> /product/ipcore/ipshared/src/RCS/i_aiecnt.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

