/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_aiecol.c	
 *
 *	AIE Color Balance module
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"
#include "aiedefs.pub"
#include "aiedefs.prv"

/* Prototypes */
#include "shrip.pub"
#include "shrip.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_aiecol.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

#define DO_SATURATION	(0)
#define cChannels		3

/*******************************************************************************
 * function:
 *	i_aieColorBalance
 *
 * parameters:
 *	pDef		Color balance default processing parameters
 *	pR, pB, pB	Image planes, without a frame
 *	w, h		Image dimensions, without a frame
 *	Bpl		Bytes per line of source image
 *	pTRCRed,
 *	pTRCGreen,
 *	pTRCBlue	TRCs to which color balance should be applied
 *	pSatVal		Computed saturation value
 *	pWhiteOff	Amount white is off
 *	pBlackOff	Amount black is off
 *	uniqueColors	Number of unique colors (if multichannel, 0 otherwise)
 *
 * purpose:
 *	this function handles both the color balance and saturation
 *	analysis of the image.  currently saturation is NOT applied.
 *	to determine if color balance or gray balance should be
 *	applied to the image requires determining both white and
 *	black points for each of the r, g, and b separations. in both
 *	cases, we look at non-saturated values.  we then choose the
 *	correction based on the LOWER of the 2 values:  white_point_off
 *	and gray_balance_off.
 *
 *	to determine saturation ratio, we look at the max(r, g, b) -
 *	min(r, g, b) divided by max(r, g, b) for a given range of values.
 *	to get the saturation value, we take the defined saturation point
 *	add 1.0 and subract the saturation ratio. if both of the hints
 *	hint90 and hint100 are both either "Image is too dark" or
 *	"Image is too light",  we will decrease the saturation by 10 percent.
 *
 *	color_balance() only makes sense on color images.
 *
 * returns:
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_aieColorBalance (
    IPState    *pIPState,
    aColorDefaults *pDef,
    UInt8Ptr	pRed,
    UInt8Ptr	pGreen,
    UInt8Ptr	pBlue,
    Int32	w,
    Int32	h,
    UInt32	Bpl,
    aTable     *pTRCRed,
    aTable     *pTRCGreen,
    aTable     *pTRCBlue,
    Float64Ptr	pSatVal,
    Float64Ptr	pWhiteOff,
    Float64Ptr	pBlackOff,
    UInt32	uniqueColors)
		  
{
    PROLOGUE;
    aColorDefaults   def;
    int              i, j;
#if DO_SATURATION
    int              A, B;
#endif
    Int32            temp;
    Int32            numPoints;
    UInt8	    *rptr, *gptr, *bptr;
    register Int32   r, g, b;
    int              ibuf[3];
    Int32	     bmr, bmg, bmb;
    Int32	     wmr, wmg, wmb;
    Float64          mult;
    Int32            in[4], out[4];
    aTable	    *pTRCRedBalance, *pTRCGreenBalance, *pTRCBlueBalance;
    Int32            channelSum[3];
    Int32	     whitePoint[3], blackPoint[3];
    Float64	     Ya[4], Yw[4];
    Int32	     aveMin, aveMid, aveMax, wMin, wMax, bMin, bMax;
    Float64	     d_aveMin, d_aveMid, d_aveMax;
    Float64	     grayBalanceOff, whiteOff, blackOff;
    Int32	     whiteMin, blackMax;
    Int32	     tableSize;
    Float64	     tmp;

    /* Make a copy because it gets trashed */
    def = *pDef;
#ifdef PRINT_MESSAGE
    fprintf(stderr, "/***********************/\n");
    fprintf(stderr, "/*    color balance    */\n");
#if DO_SATURATION
    fprintf(stderr, "/*         and         */\n");
    fprintf(stderr, "/* saturation analysis */\n");
#endif
    fprintf(stderr, "/***********************/\n");
#endif
    whitePoint[0] = whitePoint[1] = whitePoint[2] = 255;
    blackPoint[0] = blackPoint[1] = blackPoint[2] = 0;
    channelSum[0] = channelSum[1] = channelSum[2] = 0;
    numPoints = 0;
    whiteMin = 0;
    blackMax = 255;
    tableSize = 256;	/* This should be passed in */

#if DO_SATURATION
    /*
     * we can do saturation analysis at the same time...
     */
    *pSatVal = 0.0;
    minRange = 100;
    maxRange = 200;
    ratio = 0.0;
#endif
    *pSatVal = 0.0;

    /*
     * Grind through the image
     */
    rptr = pRed;
    gptr = pGreen;
    bptr = pBlue;
    for (j = 0; j < h; j++)
    {
	for (i = 0; i < w; i++)
	{
	    r = UCHAR_ACCESS(rptr + i);
	    g = UCHAR_ACCESS(gptr + i);
	    b = UCHAR_ACCESS(bptr + i);
	    channelSum[0] += (long) r;
	    channelSum[1] += (long) g;
	    channelSum[2] += (long) b;
	    numPoints++;
#if DO_SATURATION
	    A = MAX_2(r, g, b);
	    B = MIN_2(r, g, b);
	    if ((A > minRange) && (A < maxRange))
	    {
		ratio = (Float64) ((Float64) (A - B) / (Float64) A);
		if (ratio > *pSatVal)
		    *pSatVal = ratio;
	    }
#endif

	    /*
	     * Search for white-point and black-point:
	     */

	    if ((r >= 255) || (g >= 255) || (b >= 255))
		; /* skip saturated white pixels */
	    else if ((r <= 0) || (g <= 0) || (b <= 0))
		;/* skip saturated black pixels */
	    else
	    {
		if (r >= whiteMin)
		{
		    temp = r;
		    if (g >= whiteMin)
		    {
			temp = MIN_1(temp, g);
			if (b >= whiteMin)
			{
			    whiteMin = MIN_1(temp, b);
			    whitePoint[0] = r;
			    whitePoint[1] = g;
			    whitePoint[2] = b;
			}
		    }
		}
		
		if (r <= blackMax)
		{
		    temp = r;
		    if (g <= blackMax)
		    {
			temp = MAX_1(temp, g);
			if (b <= blackMax)
			{
			    blackMax = MAX_1(temp, b);
			    blackPoint[0] = r;
			    blackPoint[1] = g;
			    blackPoint[2] = b;
			}
		    }
		}
	    }
	}

	/* Move to the next line */
	rptr += Bpl;
	gptr += Bpl;
	bptr += Bpl;
    }

    aveMin = wMin = bMin = INT_MAX;
    aveMax = wMax = bMax = 0;
    aveMid = 0;

    for (i = 0; i < cChannels; i++)
    {
	ibuf[i] = channelSum[i] / numPoints;
	if (channelSum[i] < aveMin)
	    aveMin = channelSum[i];
	if (channelSum[i] > aveMax)
	    aveMax = channelSum[i];

	if (whitePoint[i] < wMin)
	    wMin = whitePoint[i];
	if (whitePoint[i] > wMax)
	    wMax = whitePoint[i];
	if (blackPoint[i] < bMin)
	    bMin = blackPoint[i];
	if (blackPoint[i] > bMax)
	    bMax = blackPoint[i];
    }

    for (i = 0; i < cChannels; i++)
    {
	if ((channelSum[i] < aveMax) && (channelSum[i] > aveMin))
	    aveMid = channelSum[i];
    }
#ifdef PRINT_MESSAGE
    fprintf(stderr, "averages:  ");
    for (i = 0; i < cChannels; i++)
	fprintf(stderr, "%d ", ibuf[i]);
    fprintf(stderr, "\n");
#endif

    d_aveMin = (Float64) aveMin / (Float64) numPoints;
    d_aveMax = (Float64) aveMax / (Float64) numPoints;
    d_aveMid = (Float64) aveMid / (Float64) numPoints;
    grayBalanceOff = ((Float64) aveMax / (Float64) aveMin) - 1.0;
    whiteOff = ((Float64) wMax / (Float64) wMin) - 1.0;
    blackOff = ((Float64) (255 - bMin) / (Float64) (255 - bMax)) - 1.0;

#ifdef PRINT_MESSAGE
    fprintf(stderr, "avemin:  %.4f\n", d_aveMin);
    fprintf(stderr, "avemid:  %.4f\n", d_aveMid);
    fprintf(stderr, "avemax:  %.4f\n", d_aveMax);
#endif

    for (i = 0; i < cChannels; i++)
    {
	Ya[i] = (Float64) aveMid / (Float64) channelSum[i];
	Yw[i] = (Float64) wMax / (Float64) whitePoint[i];
    }

#ifdef PRINT_MESSAGE
    fprintf(stderr, "Ya:  %.4f %.4f %.4f\n", Ya[0], Ya[1], Ya[2]);
    fprintf(stderr, "Yw:  %.4f %.4f %.4f\n", Yw[0], Yw[1], Yw[2]);
    fprintf(stderr, "whitePoint:  ");
    for (i = 0; i < cChannels; i++)
	fprintf(stderr, "%ld ", whitePoint[i]);
    fprintf(stderr, "\n");
    fprintf(stderr, "whitePoint_off:  %.4f\n", whiteOff);
    fprintf(stderr, "blackPoint:  ");
    for (i = 0; i < cChannels; i++)
	fprintf(stderr, "%ld ", blackPoint[i]);
    fprintf(stderr, "\n");
    fprintf(stderr, "blackPoint_off:  %.4f\n", blackOff);
    fprintf(stderr, "gray_balance_off:  %.4f\n", grayBalanceOff);
#if DO_SATURATION
    fprintf(stderr, "saturation ratio:  %.4f\n", *pSatVal);
#endif
#endif

    tmp = R_CONST * blackPoint[0] + G_CONST * blackPoint[1] +
	B_CONST * blackPoint[2];
    if ((blackOff >= 0.00) && (blackOff <= def.maxBlackOff))
    {
	bmr = (Int32)(2.0 * blackOff * (Float64)blackPoint[0] +
		      (1.0 - 2.0 * blackOff) * tmp + 0.5);
	bmg = (Int32)(2.0 * blackOff * (Float64)blackPoint[1] +
		      (1.0 - 2.0 * blackOff) * tmp + 0.5);
	bmb = (Int32)(2.0 * blackOff * (Float64)blackPoint[2] +
		      (1.0 - 2.0 * blackOff) * tmp + 0.5);
    }
    else
    {
	bmr = blackPoint[0];
	bmg = blackPoint[1];
	bmb = blackPoint[2];
    }
    bmr = (bmr < 0) ? 0 : (bmr > 255) ? 255 : bmr;
    bmg = (bmg < 0) ? 0 : (bmg > 255) ? 255 : bmg;
    bmb = (bmb < 0) ? 0 : (bmb > 255) ? 255 : bmb;

    tmp = R_CONST * whitePoint[0] + G_CONST * whitePoint[1] +
	B_CONST * whitePoint[2];
    if ((whiteOff >= 0.00) && (whiteOff <= def.maxWhiteOff))
    {
	wmr = (Int32)(2.0 * whiteOff * (Float64)whitePoint[0] +
		      (1.0 - 2.0 * whiteOff) * tmp + 0.5);
	wmg = (Int32)(2.0 * whiteOff * (Float64)whitePoint[1] +
		      (1.0 - 2.0 * whiteOff) * tmp + 0.5);
	wmb = (Int32)(2.0 * whiteOff * (Float64)whitePoint[2] +
		      (1.0 - 2.0 * whiteOff) * tmp + 0.5);
    }
    else
    {
	wmr = whitePoint[0];
	wmg = whitePoint[1];
	wmb = whitePoint[2];
    }
    wmr = (wmr < 0) ? 0 : (wmr > 255) ? 255 : wmr;
    wmg = (wmg < 0) ? 0 : (wmg > 255) ? 255 : wmg;
    wmb = (wmb < 0) ? 0 : (wmb > 255) ? 255 : wmb;

    mult = (Float64) tableSize / 256.0;

#ifdef DEBUG
    if (blackOff <= def.maxBlackOff)
	aie_message(globals, ins, "changing black point to %d %d %d\n", bmr, bmg, bmb);
#endif

    if ((whiteOff <= grayBalanceOff) || (def.maxWhiteOff == -1.0))
    {
	if (whiteOff <= def.maxWhiteOff)
	{
	    in[0] = 0;
	    in[1] = (int) (mult * blackPoint[0]);
	    in[2] = (int) (mult * whitePoint[0]);
	    in[3] = tableSize - 1;
	    out[0] = 0;
	    out[1] = (int) (mult * bmr);
	    out[2] = (int) (mult * wmr);
	    out[3] = tableSize - 1;
	    ip_tableInterpolate (pIPState, 4, in, out, &pTRCRedBalance);
	    in[0] = 0;
	    in[1] = (int) (mult * blackPoint[1]);
	    in[2] = (int) (mult * whitePoint[1]);
	    in[3] = tableSize - 1;
	    out[0] = 0;
	    out[1] = (int) (mult * bmg);
	    out[2] = (int) (mult * wmg);
	    out[3] = tableSize - 1;
	    ip_tableInterpolate (pIPState, 4, in, out, &pTRCGreenBalance);
	    in[0] = 0;
	    in[1] = (int) (mult * blackPoint[2]);
	    in[2] = (int) (mult * whitePoint[2]);
	    in[3] = tableSize - 1;
	    out[0] = 0;
	    out[1] = (int) (mult * bmb);
	    out[2] = (int) (mult * wmb);
	    out[3] = tableSize - 1;
	    ip_tableInterpolate (pIPState, 4, in, out, &pTRCBlueBalance);
#ifdef PRINT_MESSAGE
	    fprintf(stderr, "applying colorbalance...\n");
#endif
	}
	else
	{
	    in[0] = 0;
	    in[1] = tableSize - 1;
	    out[0] = 0;
	    out[1] = tableSize - 1;
	    ip_tableInterpolate (pIPState, 2, in, out, &pTRCRedBalance);
	    ip_tableInterpolate (pIPState, 2, in, out, &pTRCGreenBalance);
	    ip_tableInterpolate (pIPState, 2, in, out, &pTRCBlueBalance);

	}
    }
    else
    {
	in[0] = 0;
	in[1] = tableSize - 1;
	out[0] = 0;
	out[1] = tableSize - 1;
	ip_tableInterpolate (pIPState, 2, in, out, &pTRCRedBalance);
	ip_tableInterpolate (pIPState, 2, in, out, &pTRCGreenBalance);
	ip_tableInterpolate (pIPState, 2, in, out, &pTRCBlueBalance);
    }
    if ((grayBalanceOff < whiteOff) || (def.maxGrayOff == -1.0))
    {
	if ((grayBalanceOff <= def.maxGrayOff) || (def.maxGrayOff == -1.0))
	{
	    if (!pTRCRedBalance)
	    {
		in[0] = 0;
		in[1] = tableSize - 1;
		out[0] = 0;
		out[1] = tableSize - 1;
		ip_tableInterpolate (pIPState, 2, in, out, &pTRCRedBalance);
		ip_tableInterpolate (pIPState, 2, in, out, &pTRCGreenBalance);
		ip_tableInterpolate (pIPState, 2, in, out, &pTRCBlueBalance);
	    }
#ifdef PRINT_MESSAGE
	    fprintf(stderr, "applying graybalance...\n");
#endif
	    ip_tableApplyGain (pIPState, Ya[0],
			       pTRCRedBalance, pTRCRedBalance);
	    ip_tableApplyGain (pIPState, Ya[1],
			       pTRCGreenBalance, pTRCGreenBalance);
	    ip_tableApplyGain (pIPState, Ya[2],
			       pTRCBlueBalance, pTRCBlueBalance);
	}
	else
	{
	    /*
	     * * we may apply modified graybalance, similar to modified color
	     * balance.
	     */
#ifdef PRINT_MESSAGE
	    fprintf(stderr, "gray balance off %f is too high, max is %f.  ", grayBalanceOff, def.maxGrayOff);
	    fprintf(stderr, "skipping graybalance...\n");
#endif
	}
    }

    if (pTRCRedBalance)
    {
	if (uniqueColors > 32)
	{
	    ip_tableFold (pIPState, pTRCRedBalance, pTRCRed, pTRCRed);
	    ip_tableFold (pIPState, pTRCGreenBalance, pTRCGreen, pTRCGreen);
	    ip_tableFold (pIPState, pTRCBlueBalance, pTRCBlue, pTRCBlue);
	}

	ip_tableFree (pIPState, pTRCRedBalance);
	ip_tableFree (pIPState, pTRCGreenBalance);
	ip_tableFree (pIPState, pTRCBlueBalance);
    }

    *pWhiteOff = whiteOff;
    *pBlackOff = blackOff;

    NORMAL_RETURN;
}

/***********************************************************************
 * 
 * i_aieColorBalance_init 
 *
 *	Initialize for banded color-balance analysis
 *
 * Arguments:
 *	pData	-	"Intermediate Data" structure
 *	
 * Returns:
 *	ia_successful or error status
 *
 ***********************************************************************/

Int32 CDECL 
i_aieColorBalance_init (
    IPState    *pIPState,
    aIntermediates *pData)
{
    PROLOGUE;
    pData->whitePoint[0] = pData->whitePoint[1] = pData->whitePoint[2] = 255;
    pData->blackPoint[0] = pData->blackPoint[1] = pData->blackPoint[2] = 0;
    pData->channelSum[0] = pData->channelSum[1] = pData->channelSum[2] = 0;
    pData->numPoints = 0;
    pData->whiteMin = 0;
    pData->blackMax = 255;
    pData->numColors = 0;
    NORMAL_RETURN;
}

/***********************************************************************
 *
 * i_aieColorBalance_data
 *
 *	Gather info from a band of image for color balance analysis.
 *
 * Arguments:
 *	pData	    -	    "Intermediate Values" structure
 *	pRed	    -	    Pointer to red band, without a frame
 *	pGreen	    -	    Pointer to green band, without a frame
 *	pBlue	    -	    Pointer to blue band, without a frame
 *	lines	    -	    Number of lines to proces, WITHOUT frame
 * 
 * Returns:
 *	ia_successful or error code
 *
 ***********************************************************************/

Int32 CDECL 
i_aieColorBalance_data (
    IPState    *pIPState,
    aIntermediates *pData,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    Int32	lines)
{
    PROLOGUE;
    Int32    w, bpl, i, j, temp;
    Int32    r, g, b;
    UInt8   *rptr, *gptr, *bptr;
    Int32    currentLine;

    w = pData->w; 
    bpl = pData->bpl;
    currentLine = pData->currentLine;

    /*
     * Add in any new colors found, up to max of 257
     */
    ExceptPass(i_accumulateColors (pIPState, pRed, pGreen, pBlue, w, lines, bpl,
				  257, pData->colorList, &pData->numColors));

    /*
     * Grind through the analysis portion of the image, accumulate color sums,
     * black and white points.
     */
    if (currentLine < pData->scaledY) {
	pRed += (pData->scaledY - currentLine) * bpl;
	pGreen += (pData->scaledY - currentLine) * bpl;
	pBlue += (pData->scaledY - currentLine) * bpl;
	lines -= pData->scaledY - currentLine;
	currentLine = pData->scaledY;
    }

    if (currentLine + lines > pData->scaledY + pData->scaledH) {
	lines = pData->scaledY + pData->scaledH - currentLine;
    }

    if (lines <= 0) {
	NORMAL_RETURN;
    }

    rptr = pRed + pData->scaledX;
    gptr = pGreen + pData->scaledX;
    bptr = pBlue + pData->scaledX;
    for (j = 0; j < lines; j++)
    {
	for (i = 0; i < pData->scaledW; i++)
	{
	    r = UCHAR_ACCESS(rptr + i);
	    g = UCHAR_ACCESS(gptr + i);
	    b = UCHAR_ACCESS(bptr + i);
	    pData->channelSum[0] += (long) r;
	    pData->channelSum[1] += (long) g;
	    pData->channelSum[2] += (long) b;
	    pData->numPoints++;

	    if ((r >= 255) || (g >= 255) || (b >= 255))
		; /* skip saturated white pixels */
	    else if ((r <= 0) || (g <= 0) || (b <= 0))
		;/* skip saturated black pixels */
	    else
	    {
		if (r >= pData->whiteMin)
		{
		    temp = r;
		    if (g >= pData->whiteMin)
		    {
			temp = MIN_1(temp, g);
			if (b >= pData->whiteMin)
			{
			    pData->whiteMin = MIN_1(temp, b);
			    pData->whitePoint[0] = r;
			    pData->whitePoint[1] = g;
			    pData->whitePoint[2] = b;
			}
		    }
		}
		
		if (r <= pData->blackMax)
		{
		    temp = r;
		    if (g <= pData->blackMax)
		    {
			temp = MAX_1(temp, g);
			if (b <=  pData->blackMax)
			{
			    pData->blackMax = MAX_1(temp, b);
			    pData->blackPoint[0] = r;
			    pData->blackPoint[1] = g;
			    pData->blackPoint[2] = b;
			}
		    }
		}
	    }
	}

	/* Move to the next line */
	rptr += bpl;
	gptr += bpl;
	bptr += bpl;
    }

    NORMAL_RETURN;
}

/***********************************************************************
 *
 * i_aieColorBalance_finish
 *
 *	Finish color balance analysis and clean up.
 *	To determine if color balance or gray balance should be
 *	applied to the image requires determining both white and
 *	black points for each of the r, g, and b separations. In both
 *	cases, we look at non-saturated values.  we then choose the
 *	correction based on the LOWER of the 2 values:  white_point_off
 *	and gray_balance_off.
 *
 *	To determine saturation ratio, we look at the max(r, g, b) - min(r, g, b)
 *	divided by max(r, g, b) for a given range of values.
 *	To get the saturation value, we take the defined saturation point
 *	add 1.0 and subract the saturation ratio. If both of the hints
 *	hint90 and hint100 are both either "Image is too dark" or
 *	"Image is too light",  we will decrease the saturation by 10 percent.
 *
 *	color_balance() only makes sense on color images.
 *
 * Arguments:
 *	pData =		"Intermediate Values" data structure
 *	pTRCRed,
 *	pTRCGreen,
 *	pTRCBlue =	TRCs to which color balance should be applied
 *	pSatVal	=	Computed saturation value
 *	pWhiteOff =	Amount white is off
 *	pBlackOff =	Amount black is off
 *	
 * Returns:
 *	ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
i_aieColorBalance_finish (
    IPState    *pIPState,
    aIntermediates *pData,
    aTable     *pTRCRed,
    aTable     *pTRCGreen,
    aTable     *pTRCBlue,
    Float64    *pSatVal,
    Float64    *pWhiteOff,
    Float64    *pBlackOff)
{
    PROLOGUE;
    aColorDefaults   def;
    int              i;
    int              ibuf[3];
    Int32	     bmr, bmg, bmb;
    Int32	     wmr, wmg, wmb;
    Float64          mult;
    Int32            in[4], out[4];
    aTable	    *pTRCRedBalance, *pTRCGreenBalance, *pTRCBlueBalance;
    Float64	     Ya[4], Yw[4];
    Int32	     aveMin, aveMid, aveMax, wMin, wMax, bMin, bMax;
    Float64	     d_aveMin, d_aveMid, d_aveMax;
    Float64	     grayBalanceOff, whiteOff, blackOff;
    Int32	     tableSize = 256;
    Float64	     tmp;

    def = pData->defaults->color;

    aveMin = wMin = bMin = INT_MAX;
    aveMax = wMax = bMax = 0;
    aveMid = 0;
    *pSatVal = 0.0;

    for (i = 0; i < cChannels; i++)
    {
	ibuf[i] = pData->channelSum[i] / pData->numPoints;
	if (pData->channelSum[i] < aveMin)
	    aveMin = pData->channelSum[i];
	if (pData->channelSum[i] > aveMax)
	    aveMax = pData->channelSum[i];

	if (pData->whitePoint[i] < wMin)
	    wMin = pData->whitePoint[i];
	if (pData->whitePoint[i] > wMax)
	    wMax = pData->whitePoint[i];
	if (pData->blackPoint[i] < bMin)
	    bMin = pData->blackPoint[i];
	if (pData->blackPoint[i] > bMax)
	    bMax = pData->blackPoint[i];
    }

    for (i = 0; i < cChannels; i++)
    {
	if ((pData->channelSum[i] < aveMax) && (pData->channelSum[i] > aveMin))
	    aveMid = pData->channelSum[i];
    }
#ifdef PRINT_MESSAGE
    fprintf(stderr, "averages:  ");
    for (i = 0; i < cChannels; i++)
	fprintf(stderr, "%d ", ibuf[i]);
    fprintf(stderr, "\n");
#endif

    d_aveMin = (Float64) aveMin / (Float64) pData->numPoints;
    d_aveMax = (Float64) aveMax / (Float64) pData->numPoints;
    d_aveMid = (Float64) aveMid / (Float64) pData->numPoints;
    grayBalanceOff = ((Float64) aveMax / (Float64) aveMin) - 1.0;
    whiteOff = ((Float64) wMax / (Float64) wMin) - 1.0;
    blackOff = ((Float64) (255 - bMin) / (Float64) (255 - bMax)) - 1.0;

#ifdef PRINT_MESSAGE
    fprintf(stderr, "avemin:  %.4f\n", d_aveMin);
    fprintf(stderr, "avemid:  %.4f\n", d_aveMid);
    fprintf(stderr, "avemax:  %.4f\n", d_aveMax);
#endif

    for (i = 0; i < cChannels; i++)
    {
	Ya[i] = (Float64) aveMid / (Float64) pData->channelSum[i];
	Yw[i] = (Float64) wMax / (Float64) pData->whitePoint[i];
    }

#ifdef PRINT_MESSAGE
    fprintf(stderr, "Ya:  %.4f %.4f %.4f\n", Ya[0], Ya[1], Ya[2]);
    fprintf(stderr, "Yw:  %.4f %.4f %.4f\n", Yw[0], Yw[1], Yw[2]);
    fprintf(stderr, "whitePoint:  ");
    for (i = 0; i < cChannels; i++)
	fprintf(stderr, "%ld ", pData->whitePoint[i]);
    fprintf(stderr, "\n");
    fprintf(stderr, "whitePoint_off:  %.4f\n", whiteOff);
    fprintf(stderr, "blackPoint:  ");
    for (i = 0; i < cChannels; i++)
	fprintf(stderr, "%ld ", pData->blackPoint[i]);
    fprintf(stderr, "\n");
    fprintf(stderr, "blackPoint_off:  %.4f\n", blackOff);
    fprintf(stderr, "gray_balance_off:  %.4f\n", grayBalanceOff);
#if DO_SATURATION
    fprintf(stderr, "saturation ratio:  %.4f\n", *pSatVal);
#endif
#endif

    tmp = R_CONST * pData->blackPoint[0] + G_CONST * pData->blackPoint[1] +
	B_CONST * pData->blackPoint[2];
    if ((blackOff >= 0.00) && (blackOff <= def.maxBlackOff))
    {
	bmr = (Int32)(2.0 * blackOff * (Float64)pData->blackPoint[0] +
		      (1.0 - 2.0 * blackOff) * tmp + 0.5);
	bmg = (Int32)(2.0 * blackOff * (Float64)pData->blackPoint[1] +
		      (1.0 - 2.0 * blackOff) * tmp + 0.5);
	bmb = (Int32)(2.0 * blackOff * (Float64)pData->blackPoint[2] +
		      (1.0 - 2.0 * blackOff) * tmp + 0.5);
    }
    else
    {
	bmr = pData->blackPoint[0];
	bmg = pData->blackPoint[1];
	bmb = pData->blackPoint[2];
    }
    bmr = (bmr < 0) ? 0 : (bmr > 255) ? 255 : bmr;
    bmg = (bmg < 0) ? 0 : (bmg > 255) ? 255 : bmg;
    bmb = (bmb < 0) ? 0 : (bmb > 255) ? 255 : bmb;

    tmp = R_CONST * pData->whitePoint[0] + G_CONST * pData->whitePoint[1] +
	B_CONST * pData->whitePoint[2];
    if ((whiteOff >= 0.00) && (whiteOff <= def.maxWhiteOff))
    {
	wmr = (Int32)(2.0 * whiteOff * (Float64)pData->whitePoint[0] +
		      (1.0 - 2.0 * whiteOff) * tmp + 0.5);
	wmg = (Int32)(2.0 * whiteOff * (Float64)pData->whitePoint[1] +
		      (1.0 - 2.0 * whiteOff) * tmp + 0.5);
	wmb = (Int32)(2.0 * whiteOff * (Float64)pData->whitePoint[2] +
		      (1.0 - 2.0 * whiteOff) * tmp + 0.5);
    }
    else
    {
	wmr = pData->whitePoint[0];
	wmg = pData->whitePoint[1];
	wmb = pData->whitePoint[2];
    }
    wmr = (wmr < 0) ? 0 : (wmr > 255) ? 255 : wmr;
    wmg = (wmg < 0) ? 0 : (wmg > 255) ? 255 : wmg;
    wmb = (wmb < 0) ? 0 : (wmb > 255) ? 255 : wmb;

    mult = (Float64) tableSize / 256.0;

    if ((whiteOff <= grayBalanceOff) || (def.maxWhiteOff == -1.0))
    {
	if (whiteOff <= def.maxWhiteOff)
	{
	    in[0] = 0;
	    in[1] = (int) (mult * pData->blackPoint[0]);
	    in[2] = (int) (mult * pData->whitePoint[0]);
	    in[3] = tableSize - 1;
	    out[0] = 0;
	    out[1] = (int) (mult * bmr);
	    out[2] = (int) (mult * wmr);
	    out[3] = tableSize - 1;
	    ip_tableInterpolate (pIPState, 4, in, out, &pTRCRedBalance);
	    in[0] = 0;
	    in[1] = (int) (mult * pData->blackPoint[1]);
	    in[2] = (int) (mult * pData->whitePoint[1]);
	    in[3] = tableSize - 1;
	    out[0] = 0;
	    out[1] = (int) (mult * bmg);
	    out[2] = (int) (mult * wmg);
	    out[3] = tableSize - 1;
	    ip_tableInterpolate (pIPState, 4, in, out, &pTRCGreenBalance);
	    in[0] = 0;
	    in[1] = (int) (mult * pData->blackPoint[2]);
	    in[2] = (int) (mult * pData->whitePoint[2]);
	    in[3] = tableSize - 1;
	    out[0] = 0;
	    out[1] = (int) (mult * bmb);
	    out[2] = (int) (mult * wmb);
	    out[3] = tableSize - 1;
	    ip_tableInterpolate (pIPState, 4, in, out, &pTRCBlueBalance);
#ifdef PRINT_MESSAGE
	    fprintf(stderr, "applying colorbalance...\n");
#endif
	}
	else
	{
	    in[0] = 0;
	    in[1] = tableSize - 1;
	    out[0] = 0;
	    out[1] = tableSize - 1;
	    ip_tableInterpolate (pIPState, 2, in, out, &pTRCRedBalance);
	    ip_tableInterpolate (pIPState, 2, in, out, &pTRCGreenBalance);
	    ip_tableInterpolate (pIPState, 2, in, out, &pTRCBlueBalance);

	}
    }
    else
    {
	in[0] = 0;
	in[1] = tableSize - 1;
	out[0] = 0;
	out[1] = tableSize - 1;
	ip_tableInterpolate (pIPState, 2, in, out, &pTRCRedBalance);
	ip_tableInterpolate (pIPState, 2, in, out, &pTRCGreenBalance);
	ip_tableInterpolate (pIPState, 2, in, out, &pTRCBlueBalance);
    }
    if ((grayBalanceOff < whiteOff) || (def.maxGrayOff == -1.0))
    {
	if ((grayBalanceOff <= def.maxGrayOff) || (def.maxGrayOff == -1.0))
	{
	    if (!pTRCRedBalance)
	    {
		in[0] = 0;
		in[1] = tableSize - 1;
		out[0] = 0;
		out[1] = tableSize - 1;
		ip_tableInterpolate (pIPState, 2, in, out, &pTRCRedBalance);
		ip_tableInterpolate (pIPState, 2, in, out, &pTRCGreenBalance);
		ip_tableInterpolate (pIPState, 2, in, out, &pTRCBlueBalance);
	    }
#ifdef PRINT_MESSAGE
	    fprintf(stderr, "applying graybalance...\n");
#endif
	    ip_tableApplyGain (pIPState, Ya[0],
			       pTRCRedBalance, pTRCRedBalance);
	    ip_tableApplyGain (pIPState, Ya[1],
			       pTRCGreenBalance, pTRCGreenBalance);
	    ip_tableApplyGain (pIPState, Ya[2],
			       pTRCBlueBalance, pTRCBlueBalance);
	}
	else
	{
	    /*
	     * * we may apply modified graybalance, similar to modified color
	     * balance.
	     */
#ifdef PRINT_MESSAGE
	    fprintf(stderr, "gray balance off %f is too high, max is %f.  ", grayBalanceOff, def.maxGrayOff);
	    fprintf(stderr, "skipping graybalance...\n");
#endif
	}
    }

    if (pTRCRedBalance)
    {
	if (pData->numColors > 32)
	{
	    ip_tableFold (pIPState, pTRCRedBalance, pTRCRed, pTRCRed);
	    ip_tableFold (pIPState, pTRCGreenBalance, pTRCGreen, pTRCGreen);
	    ip_tableFold (pIPState, pTRCBlueBalance, pTRCBlue, pTRCBlue);
	}

	ip_tableFree (pIPState, pTRCRedBalance);
	ip_tableFree (pIPState, pTRCGreenBalance);
	ip_tableFree (pIPState, pTRCBlueBalance);
    }

    *pWhiteOff = whiteOff;
    *pBlackOff = blackOff;

    NORMAL_RETURN;
}

/******************************************************************************
 * function:
 *	i_aieWhitepoint ()
 *
 * parameters:
 *	pDef		Color balance defaults
 *	pLuminance	Pointer to luminance image, without a frame
 *	w, h		Image dimensions, without a frame
 *	Bpl		Bytes per line of source image
 *	pBlackPoint	minimum black found
 *	pWhitePoint	maximum white found
 *
 * returns:
 *	ia_successful
 *
 * purpose:
 *	To replace the function of colorbalance when using luminance only.
 *	The min and max "inside" of the saturation points are used as the
 *	black-point and white-point, and are to be considered "valid"
 *	quadrants when used to compute statistics in the contrast routines.
 *****************************************************************************/

Int32 CDECL 
i_aieWhitepoint (
    IPState    *pIPState,
    UInt8      *pLuminance,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    UInt8      *pBlackPoint,
    UInt8      *pWhitePoint)
{
    PROLOGUE;
    register UInt8 pixelVal, maxWhite, minBlack, *pRow;
    register UInt8 saturatedBlack, saturatedWhite;
    Int32 xpos, ypos;

    saturatedBlack = 0, saturatedWhite = 255;
    maxWhite = 0, minBlack = 255;
    pRow = pLuminance;
    for (ypos = 0; ypos < h; ypos++, pRow += Bpl)
    {
	for (xpos = 0; xpos < w; xpos++)
	{
	    pixelVal = UCHAR_ACCESS(pRow + xpos);
	    if (pixelVal < minBlack && pixelVal > saturatedBlack)
	    {
		minBlack = pixelVal;
	    }
	    if (pixelVal > maxWhite && pixelVal < saturatedWhite)
	    {
		maxWhite = pixelVal;
	    }
	}
    }

    /* Return the values found */
    *pBlackPoint = minBlack;
    *pWhitePoint = maxWhite;

    NORMAL_RETURN;
}



/*

$Log: i_aiecol.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.12  1998/01/13 23:46:56  ddavies
i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.11  1997/11/13 23:21:14  aber
i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v

CommonCheck: Updated the AIE algorithms to match (or closely match) the latest DITC
AIE used in Sandpiper. The updates include the following:
1. Exclude a frame (5%) around the exterior in almost all statistics.
2. Make sure the black-point and white-point are included in the
   computations.
3. ... this was improved upon to not use just the first region with a
   black-or-white-point, but to use the ones with the largest variance.
4. The statistics are accumulated in cells 4 across and 4 down. This
   is not done if the width and height are less than 512.
5. Black-point and white-point do not include 0 and 255 respectively.
   They are the darkest and whitest not quite at these saturation points.
6. If the number of colors or gray levels is too low, less aggressive
   correction is applied.
7. Lots of thresholds changed.
8. Luminance channel is computed with a different formula.

Revision 1.10  1995/12/13 02:41:46  ejaquith
i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.9  1994/10/06  00:39:04  macleod
 * i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v
 *
 * Eliminated an unreferenced parameter warning
 *
 * Revision 1.8  1994/10/05  22:20:30  macleod
 * i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v
 *
 * Added support for banded AIE
 *
 * Revision 1.7  1994/07/12  23:54:33  ejaquith
 * ejaquith on Tue Jul 12 16:52:49 PDT 1994
 *
 * i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v
 *
 * removed #include "iadef.prv"
 *
 * Revision 1.6  1994/07/12  00:50:08  macleod
 * i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v
 *
 * Color balance was being computed, but the code that was to apply it
 * had its test wrong, so it would only get applied (and structures freed)
 * if the structures weren't allocated in the first place. This means
 * that color-balance correction never got applied.
 *
 * Revision 1.5  1994/02/04  20:13:07  dferrell
 * i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:22:14  ddavies
 * i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:26:41  ddavies
 * i_aiecol.c -> /product/ipcore/ipshared/src/RCS/i_aiecol.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

