/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_aiepro.c	
 *
 *	AIE processor. The functions in this module process an
 *	image according to the values that AIE analysis produced.
 *	This is typically a TRC or set of TRCs, and a sharpness
 *	value.
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"
#include "aiedefs.pub"
#include "aiedefs.prv"

/* Prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shros.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_aiepro.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/*******************************************************************************
 * i_aieProcessRGB 
 *
 *	Process an RGB image given three TRCs and a sharpness value.
 *	If the sharpness value is 1.0, the image is not sharpened.
 *
 * Arguments:
 *
 *	pRed
 *	pGreen
 *	pBlue =		    Image planes to process
 *	sBpl =		    Bytes per line in a source image plane
 *	pRedCorr
 *	pGreenCorr
 *	pBlueCorr =	    Corrected Image planes, i.e. result
 *	dBpl =		    Bytes per line in a result image plane
 *	w, h =		    Size of an image plane, including frame
 *	pRedTRC
 *	pGreenTRC
 *	pBlueTRC =	    Tone Response Curves for each plane
 *	sharpness =	    Sharpness value
 * 
 * Returns:
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_aieProcessRGB (
    IPState    *pIPState,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    Int32	sBpl,
    UInt8      *pRedCorr,
    UInt8      *pGreenCorr,
    UInt8      *pBlueCorr,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    aTable     *pRedTRC,
    aTable     *pGreenTRC,
    aTable     *pBlueTRC,
    Float64	sharpness)
{
    PROLOGUE;

    /* We can, in fact, do each plane separately as a gray plane */
    /* this would not be the case if we sharpened the Y image */
    ERR_PASS (i_aieProcessGray (pIPState, pRed, sBpl, pRedCorr, dBpl, 
				w, h, pRedTRC, sharpness));
    IP_YIELD (fail);
    ERR_PASS (i_aieProcessGray (pIPState, pGreen, sBpl, pGreenCorr, dBpl,
				w, h, pGreenTRC, sharpness));
    IP_YIELD (fail);
    ERR_PASS (i_aieProcessGray (pIPState, pBlue, sBpl, pBlueCorr, dBpl, 
				w, h, pBlueTRC, sharpness));

fail:
    NORMAL_RETURN;

}

/*******************************************************************************
 * i_aieProcessGray 
 *
 *	Process an Gray image given a TRC and a sharpness value.
 *	If the sharpness value is 1.0, the image is not sharpened.
 *
 * Arguments:
 *
 *	pSrc =		    The image plane to process
 *	sBpl =		    Bytes per line in a source image plane
 *	pCorr =		    Corrected Image plane, i.e. result
 *	dBpl =		    Bytes per line in a result image plane
 *	w, h =		    Size of the image plane, including frame
 *	pGrayTRC =	    Tone Response Curves for the image
 *	sharpness =	    Sharpness value
 * 
 * Returns:
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_aieProcessGray (
    IPState    *pIPState,
    UInt8      *pSrc,
    Int32	sBpl,
    UInt8      *pCorr,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    aTable     *pGrayTRC,
    Float64	sharpness)
{
    PROLOGUE;
    UInt8      *trc;
    Int32      *trc32P;
    UInt8       trc8[256];
    Int32       i;

    sharpness = 0;  /* Shut up WATCOM */

    /* Convert the table to 8-bit */
    trc = &trc8[0];
    for (i = 0, trc32P = ip_tableData (pIPState, pGrayTRC); i < 256; i++)
    {
	trc[i] = (UInt8)(trc32P[i]);
    }

    /*
      Call the 8 bit mapping routine. This works
      for PC or Sun,etc
    */
    ERR_PASS( i_map8To8 (pIPState, pSrc, w, h, sBpl, pCorr, dBpl, trc));

    NORMAL_RETURN;
}

/*

$Log: i_aiepro.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.13  1998/01/13 23:46:57  ddavies
i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.12  1995/12/13 02:41:50  ejaquith
i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.11  1995/10/13  23:49:11  dferrell
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.10  1994/10/28  01:52:38  lperry
 * lperry on Thu Oct 27 18:50:41 PDT 1994
 *
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * Change to use i_map8To8, rather than have special
 * LUT mapping code up at this level. Now makes the
 * same call regardless of architecture. That
 * special casing was moved down to i_map8To8.
 *
 * Revision 1.9  1994/10/17  19:10:05  lperry
 * lperry on Mon Oct 17 12:04:03 PDT 1994
 *
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * Modified function i_aieProcessGray for improved performance
 * Runs different code for PC and Sun.
 * On Sun, now calls i_map8To8 to do the tone reproduction lookup
 * table mapping (DRC).
 * On the PC, the UCHAR_ACCESS functions were eliminated, since
 * byte order doesn't matter for point transformations (ends of
 * lines are handled by special case code).
 *
 * Results in about 50% speedup on Sun and 80% on PC.
 *
 * Revision 1.8  1994/10/11  16:37:03  lperry
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * Added include of shros.pub to get IP_YIELD def
 *
 * Revision 1.7  1994/07/12  23:54:57  ejaquith
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * removed #include "iadef.prv".
 *
 * Revision 1.6  1994/06/23  05:59:44  macleod
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * Added some IP_YIELD calls
 *
 * Revision 1.5  1994/02/04  20:13:15  dferrell
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:22:20  ddavies
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:26:49  ddavies
 * i_aiepro.c -> /product/ipcore/ipshared/src/RCS/i_aiepro.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

