/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_aieshr.c	
 *
 *	AIE Sharpness analyzer
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"
#include "aiedefs.pub"
#include "aiedefs.prv"

/* Prototypes */
#include "shrip.pub"
#include "shrip.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_aieshr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");



/******************************************************************************
 * function:
 *	i_aieSharpness()
 *
 * parameters:
 *	pDef =		    Sharpness default values
 *	pLuminance =	    Pointer to a luminance channel to analyze
 *	w, h =		    Size of the image, including frame
 *	Bpl =		    Bytes/line of the image
 *	pEdgeFactorIndex =  Exported edge factor index, used by contrast
 *	pEdgeFactor =	    Computed sharpening factor
 *
 * returns:
 *	ia_successful, ...
 *
 * purpose:
 *	determine the min and max pixel difference for a given NxM
 *	region (currently 3x2) and use the maximum difference for the
 *	image as the edge factor index into the sharp.dat data file.
 *	the edge multiplier and offset is then applied to the edge
 *	factor.  if the edge factor does not meet the minimum/maximum
 *	edge factor requirements, they are set appropriately.
 *****************************************************************************/

Int32 CDECL 
i_aieSharpness (
    IPState    *pIPState,
    aSharpnessDefaults *pDef,
    UInt8      *pLuminance,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    Int32      *pEdgeFactorIndex,
    Float64    *pEdgeFactor)
{
    PROLOGUE;
    register UInt8	a, b, c, d, e;
    register UInt8	*odd, *even;
    register Int32	diff, maxDiff;
    register Int32	i, j;

    /* Even:  C  D  E
     * Odd:   B  A  F
     *
     * Sharpness computed as maximum difference of: |A - B|, |A - C|,
     * |A - D|, and |A - E|.
     */

#ifdef PRINT_MESSAGE
    fprintf(stderr, "/******************/\n");
    fprintf(stderr, "/* sharp analysis */\n");
    fprintf(stderr, "/******************/\n");
#endif
    maxDiff = 0;
    even = pLuminance;
    odd = even + Bpl;

    for (i = 1; i < h; i++)
    {

	for (j = 1; j < w - 1; j++)
	{
	    a = (int)UCHAR_ACCESS(odd + j);
	    b = (int)UCHAR_ACCESS(odd + j - 1);
	    c = (int)UCHAR_ACCESS(even + j - 1);
	    d = (int)UCHAR_ACCESS(even + j);
	    e = (int)UCHAR_ACCESS(even + j + 1);
	    diff = a - b;
	    if (diff < 0) diff = -diff;
	    if (diff > maxDiff) maxDiff = diff;
	    diff = a - c;
	    if (diff < 0) diff = -diff;
	    if (diff > maxDiff) maxDiff = diff;
	    diff = a - d;
	    if (diff < 0) diff = -diff;
	    if (diff > maxDiff) maxDiff = diff;
	    diff = a - e;
	    if (diff < 0) diff = -diff;
	    if (diff > maxDiff) maxDiff = diff;
	}
	even = odd;
	odd += Bpl;
    }
    *pEdgeFactorIndex = maxDiff;
#ifdef PRINT_MESSAGE
    fprintf(stderr, "edge factor index = %d\n", *pEdgeFactorIndex);
#endif

    if (*pEdgeFactorIndex <= 220)
    {
	*pEdgeFactor = 3.65 - (*pEdgeFactorIndex >> 1) * 0.013 -
			((*pEdgeFactorIndex + 1) >> 1) * 0.012;
    }
    else
    {
	*pEdgeFactor = 1.0;
    }

#ifdef PRINT_MESSAGE
    fprintf(stderr, "edge factor = %f\n", *pEdgeFactor);
#endif
    if (*pEdgeFactor == -1.0)
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "edge factor screwed up, setting to 1.0\n");
#endif
	*pEdgeFactor = 1.0;
    }
    if (pDef->multiplier != 1.0)
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "multiplying edge factor by %f\n", pDef->multiplier);
#endif
	*pEdgeFactor *= pDef->multiplier;
    }
    if ((pDef->offset != 1.0) && (pDef->offset != 0.0))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "adding offset %f to edge factor\n", pDef->offset);
#endif
	*pEdgeFactor += pDef->offset;
    }
    if (*pEdgeFactor < pDef->minEdge)
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "edge_factor too low, setting to %f\n", pDef->minEdge);
#endif
	*pEdgeFactor = pDef->minEdge;
    }
    if (*pEdgeFactor > pDef->maxEdge)
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "edge_factor too high, setting to %f\n", pDef->maxEdge);
#endif
	*pEdgeFactor = pDef->maxEdge;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * i_aieSharpness_init (data)
 *
 *	Initialize banded sharpness-analysis
 *
 * parameters:
 *	pData	    -	The "intermediate data values" structure, which should
 *			have the pointer to the "defaults" structure filled in
 *
 * returns:
 *	ia_successful
 *
 * purpose:
 *  
 *****************************************************************************/

Int32 CDECL 
i_aieSharpness_init (
    IPState    *pIPState,
    aIntermediates *pData)
{
    PROLOGUE;
    pData->maxDiff = 0;
    pData->sharpness = 1.0;

    NORMAL_RETURN;
}

/******************************************************************************
 *
 * i_aieSharpness_data
 *
 *	Accumulate a band of data for the purpose of sharpness analysis. 
 *
 * Arguments:
 *	pData	    -	    The "intermediate data" structure
 *	pLuminance  -	    Luminance channel data, pointer to frame
 *	lines	    -	    Number of lines to analyze
 * 
 * Returns: 
 *	ia_successful or error code
 *****************************************************************************/

Int32 CDECL 
i_aieSharpness_data (
    IPState    *pIPState,
    aIntermediates *pData,
    UInt8      *pLuminance,
    Int32	lines)
{
    PROLOGUE;
    Int32	w, bpl;
    Float64	sharpness;
    Int32	maxDiff = 0;
    UInt8	*lum;
    Int32	currentLine;

    w = pData->w;
    bpl = pData->bpl;
    lum = pLuminance;
    currentLine = pData->currentLine;

    /* Adjust for deframed image */
    if (currentLine < pData->scaledY) {
	lum += (pData->scaledY - currentLine) * bpl;
	lines -= pData->scaledY - currentLine;
	currentLine = pData->scaledY;
    }

    if (currentLine + lines > pData->scaledY + pData->scaledH) {
	lines = pData->scaledY + pData->scaledH - currentLine;
    }

    if (lines <= 0) {
	NORMAL_RETURN;
    }

    /* Run the normal algorithm on the band */
    ExceptPass (i_aieSharpness (pIPState, &(pData->defaults->sharpness),
				lum + pData->scaledX, pData->scaledW, lines,
				bpl, &maxDiff, &sharpness));

    /* Add in the new statistics if we've found a sharper edge */
    if (pData->maxDiff < maxDiff) {
	pData->maxDiff = maxDiff;
	pData->sharpness = sharpness;
    }

    NORMAL_RETURN;
}

/******************************************************************************
 * i_aieSharpness_finish ()
 *
 *	We've gathered the sharpness value, now show the result.
 *
 * Arguments:
 *	pDdata	    -	Intermediates structure returned by .._init
 *	pEdgeFactorIndex-Resulting index into sharpening table, as computed
 *			by sharpness analyzer. This is actually the maximum
 *			pixel difference between two pixels in any of several
 *			(see i_aieSharpness()) directions
 *	pSharpness -	Resulting sharpness value
 *
 * returns:
 *	ia_successful
 *
 *****************************************************************************/

Int32 CDECL 
i_aieSharpness_finish (
    IPState    *pIPState,
    aIntermediates *pData,
    Int32      *pEdgeFactorIndex,
    Float64    *pSharpness)
{
    PROLOGUE;

    *pEdgeFactorIndex = pData->maxDiff;
    *pSharpness = pData->sharpness;

    NORMAL_RETURN;
}


/*

$Log: i_aieshr.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.8  1998/01/13 23:46:59  ddavies
i_aieshr.c -> /product/ipcore/ipshared/src/RCS/i_aieshr.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1997/11/13 23:21:15  aber
i_aieshr.c -> /product/ipcore/ipshared/src/RCS/i_aieshr.c,v

CommonCheck: Updated the AIE algorithms to match (or closely match) the latest DITC
AIE used in Sandpiper. The updates include the following:
1. Exclude a frame (5%) around the exterior in almost all statistics.
2. Make sure the black-point and white-point are included in the
   computations.
3. ... this was improved upon to not use just the first region with a
   black-or-white-point, but to use the ones with the largest variance.
4. The statistics are accumulated in cells 4 across and 4 down. This
   is not done if the width and height are less than 512.
5. Black-point and white-point do not include 0 and 255 respectively.
   They are the darkest and whitest not quite at these saturation points.
6. If the number of colors or gray levels is too low, less aggressive
   correction is applied.
7. Lots of thresholds changed.
8. Luminance channel is computed with a different formula.

Revision 1.6  1995/12/13 02:41:53  ejaquith
i_aieshr.c -> /product/ipcore/ipshared/src/RCS/i_aieshr.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1994/07/12  23:55:19  ejaquith
 * i_aieshr.c -> /product/ipcore/ipshared/src/RCS/i_aieshr.c,v
 *
 * removed #include "iadef.prv".
 *
 * Revision 1.4  1994/02/04  20:13:48  dferrell
 * dferrell on Fri Feb 4 12:13:38 PST 1994
 *
 * i_aieshr.c -> /product/ipcore/ipshared/src/RCS/i_aieshr.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:22:24  ddavies
 * i_aieshr.c -> /product/ipcore/ipshared/src/RCS/i_aieshr.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:26:58  ddavies
 * i_aieshr.c -> /product/ipcore/ipshared/src/RCS/i_aieshr.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

