/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_aiexpo.c	
 *
 *	AIE exposure (gamma) correction
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"
#include "aiedefs.pub"
#include "aiedefs.prv"

/* Prototypes */
#include "shrip.pub"
#include "shrip.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_aiexpo.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/*******************************************************************************
 * function:
 *	i_aieExposure
 *
 * parameters:
 *	pDef		Pointer to exposure defaults
 *	pLuminance	Pointer to a luminance pixmap (without a frame)
 *	w, h		Width of image, without a frame
 *	sBpl		Bytes per line in above
 *	pGammaTRC	Resulting TRC which should correct for exposure. Should
 *			be initialized to linear as passed.
 *	arr100		100% rage, exported so that contrast can use it
 *	pVariance	the variance as computed
 *	pGammaComputed	place to stick the gamma value
 *	uniqueLevels	unique gray/color levels (color if source is color,
 *			gray otherwise)
 * returns:
 *	ia_successful, ia_nomem, etc.
 * purpose:
 *	to look at the 90% and 100% range of the luminance histogram and
 *	generate 2 gamma values.  if these values are somewhat reasonable,
 *	we will simply average the 2 values for a suggested gamma value.
 *	we also set hint90 and hint100 strings for future use with saturation.
 *	in this analyzer, we are looking at the global histogram.
 *
 * notes:
 *	Some documentation implied that the "hint" stuff would be useful;
 *	nobody uses it right now, so it's not exported. If it is to be
 *	exported, the hints should be changed from text to constants.
 *	(English-language text in code is a pain when one wants to
 *	internationalize)
 *
 *******************************************************************************/

Int32 CDECL 
i_aieExposure (
    IPState    *pIPState,
    aExposureDefaults *pDef,
    UInt8      *pLuminance,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    aTable     *pGammaTRC,
    Int32      *arr100,
    Float64    *pVariance,
    Float64    *pGammaComputed,
    UInt32	uniqueLevels)
{
    /* We need to make a copy of the defaults, since we trash them */
    PROLOGUE;
    aExposureDefaults def;

    Int32           i, j;
    Int32           minPixVal = 255;
    Int32           maxPixVal = 0;
    Int32	    minVal, maxVal;
    Int32           topThreshold;
    Int32           bottomThreshold;
    Int32           arr[4];
    Int32	    arr50[4], arr60[4], arr70[4], arr80[4], arr90[4];
    Int32          *hd;
    Int32          *histp;
    aTable	   *globalHistogram;
    Float64         sum;
    Float64         sqsum;
    Float64         sigma2;
    Float64         average;
    Int32           min90, max90;
    Int32           min100, max100;
    Float64         darkLog90, brightLog90;
    Float64         darkLog100, brightLog100;
    Float64	    temp_log;
    Int32           adjust;
    Int32           localWhite = 255;
    Int32           localTableSize = 256;
    Float64         gamma90 = -1.0, gamma100 = -1.0, gammaSuggest;
    Bool	    doInvert = 0;

    def = *pDef;

#ifdef PRINT_MESSAGE
    fprintf(stderr, "/*********************/\n");
    fprintf(stderr, "/* exposure analysis */\n");
    fprintf(stderr, "/*********************/\n");
#endif

    /*
     * Compute a global histogram for the luminance channel
     */
    ERR_PASS( ip_tableAllocate (pIPState, 256, &globalHistogram));

    /* Uh, DITC routine replaced by IPCore routine: */
    i_histoRect (pIPState, pLuminance, sBpl, 0, 0, w, h,
		 ip_tableData (pIPState, globalHistogram));

    /* Go through and find a min/max in the histogram */
    minVal = -1;
    for (i = 0; i < 256; i++)
    {
	if ((globalHistogram->table[i] != 0) && (minVal == -1))
	    minVal = i;
	if (globalHistogram->table[i] != 0)
	    maxVal = i;
    }

    hd = histp = ip_tableData (pIPState, globalHistogram);

    /* Compute the sum and sum-of-squares for the histogram */
    sum = sqsum = 0L;
    for (i = 0; i < 256; i++)
    {
	sum += (Float64) hd[i];
	sqsum += (Float64) ((Float64) hd[i] * (Float64) hd[i]);
    }

    /* 
     * Compute the mean and "normalized" variance for the histogram. 
     * Here, "normalized" variance means that it was multiplied by a large
     * number.
     */
    average = sum / 256.0;
    sigma2 = sqsum / 256.0 - average * average;
    sigma2 = sigma2 * (1024.0 * 1024.0) / (sum * sum);
    *pVariance = sigma2;

#ifdef PRINT_MESSAGE
    fprintf(stderr, "normalized variance: %15d\n", (int) sigma2);
#endif

    /* 
     * Compute min/max/range values for 100, 90, 80, 70, 60, and 50%
     * of the histogram. For example, in the 90% case, we would exclude
     * 5% on either end. This is 5% by "volume", that is, percentage of the
     * total sum of the histogram. 
     */
    for (j = 0; j < 6; j++)
    {
	topThreshold = bottomThreshold = 0;

	/* Search up from the bottom until we've accumulated j/2 percent */
	for (i = 0; i < 256; ++i)
	{
	    bottomThreshold += hd[i];
	    if (bottomThreshold > ((sum * j) / 20))
	    {
		minPixVal = i;
		break;
	    }
	}

	/* Search down from the top until we've accumulated j/2 percent */
	histp = hd + 256 - 1;
	for (i = 256 - 1; i >= 0; i--)
	{
	    topThreshold += hd[i];
	    if (topThreshold > ((sum * j) / 20))
	    {
		maxPixVal = i;
		break;
	    }
	}

	arr[0] = 100 - j * 10;	/* Percent of power */
	arr[1] = minPixVal;
	arr[2] = maxPixVal;
	arr[3] = (maxPixVal - minPixVal) + 1;

	/* Crikey. I'll have to clean this up at some point */
	if (arr[0] == 50)
	{
	    arr50[0] = arr[0];
	    arr50[1] = arr[1];
	    arr50[2] = arr[2];
	    arr50[3] = arr[3];
	}
	else if (arr[0] == 60)
	{
	    arr60[0] = arr[0];
	    arr60[1] = arr[1];
	    arr60[2] = arr[2];
	    arr60[3] = arr[3];
	}
	else if (arr[0] == 70)
	{
	    arr70[0] = arr[0];
	    arr70[1] = arr[1];
	    arr70[2] = arr[2];
	    arr70[3] = arr[3];
	}
	else if (arr[0] == 80)
	{
	    arr80[0] = arr[0];
	    arr80[1] = arr[1];
	    arr80[2] = arr[2];
	    arr80[3] = arr[3];
	}
	else if (arr[0] == 90)
	{
	    arr90[0] = arr[0];
	    arr90[1] = arr[1];
	    arr90[2] = arr[2];
	    arr90[3] = arr[3];
	}
	else if (arr[0] == 100)
	{
	    arr100[0] = arr[0];
	    arr100[1] = arr[1];
	    arr100[2] = arr[2];
	    arr100[3] = arr[3];
	}
#ifdef PRINT_MESSAGE
	fprintf(stderr, "%3ld%% - low %3ld, high %3ld, range %3ld\n",
		100 - j * 10, minPixVal, maxPixVal, (maxPixVal - minPixVal) + 1);
#endif
    }

    def.dark90 = (Int32) (((Float64) def.dark90 /
			   (Float64) (localWhite + 1)) * localTableSize);
    def.bright90 = (Int32) (((Float64) def.bright90 /
			     (Float64) (localWhite + 1)) * localTableSize);
    def.dark100 = (Int32) (((Float64) def.dark100 /
			    (Float64) (localWhite + 1)) * localTableSize);
    def.bright100 = (Int32) (((Float64) def.bright100 /
			      (Float64) (localWhite + 1)) * localTableSize);

    if (def.dark90 != 0)
    {
	darkLog90 = log((Float64) def.dark90 /
			(Float64) (localTableSize - 1));
    }
    else
    {
	darkLog90 = 0;
    }
    if (def.bright90 != 0)
    {
	brightLog90 = log((Float64) def.bright90 /
			  (Float64) (localTableSize - 1));
    }
    else
    {
	brightLog90 = 0;
    }
    if (def.dark100 != 0)
    {
	darkLog100 = log((Float64) def.dark100 /
			 (Float64) (localTableSize - 1));
    }
    else
    {
	darkLog100 = 0;
    }
    if (def.bright100 != 0)
    {
	brightLog100 = log((Float64) def.bright100 /
			   (Float64) (localTableSize - 1));
    }
    else
    {
	brightLog100 = 0;
    }

    min100 = (arr100[1] * (localTableSize - 1)) / localWhite;
    max100 = (arr100[2] * (localTableSize - 1)) / localWhite;

    min90 = (arr90[1] * (localTableSize - 1)) / localWhite;
    max90 = (arr90[2] * (localTableSize - 1)) / localWhite;

    if ((min90 <= def.dark90) && (max90 >= def.bright90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is fine\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	hint90 = strdup("Image is fine");
*/
	gamma90 = 1.0;
    }
    else if ((min90 >= def.dark90) && (max90 >= def.bright90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is too light\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	hint90 = strdup("Image is too light");
*/
	temp_log = log((Float64) min90 / (Float64) (localTableSize - 1));
	if (temp_log != 0.0)
	{
	    gamma90 = darkLog90 / temp_log;
	}
	else
	{
	    gamma90 = darkLog90;
	}
    }
    else if ((min90 <= def.dark90) && (max90 <= def.bright90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is too dark\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	hint90 = strdup("Image is too dark");
*/
	/*
	 * * old calculation: *   gamma90 = brightLog90 / log((Float64)
	 * max90/(Float64) (localTableSize-1)); * invert histogram
	 * for dark images to maintain symmetry
	 */
	doInvert = 1;
	min90 = localWhite - (arr90[2] * (localTableSize - 1)) / localWhite;
	max90 = localWhite - (arr90[1] * (localTableSize - 1)) / localWhite;
	temp_log = log((Float64) min90 / (Float64) (localTableSize - 1));
	if (temp_log != 0.0)
	{
	    gamma90 = darkLog90 / temp_log;
	}
	else
	{
	    gamma90 = darkLog90;
	}
    }
    else if ((min90 >= def.dark90) && (max90 <= def.bright90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is screwed up\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	hint90 = strdup("Image is screwed up");
*/
	gamma90 = -1;
    }

    if ((min100 <= def.dark100) && (max100 >= def.bright100))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is fine\n");
#endif
/*
	if (hint100 != (char *) 0)
	    free(hint100);
	hint100 = strdup("Image is fine");
*/
	gamma100 = 1.0;
    }
    else if ((min100 >= def.dark100) && (max100 >= def.bright100))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is too light\n");
#endif
/*
	if (hint100 != (char *) 0)
	    free(hint100);
	hint100 = strdup("Image is too light");
*/
	temp_log = log((Float64) min100 / (Float64) (localTableSize - 1));
	if (temp_log != 0.0)
	{
	    gamma100 = darkLog100 / temp_log;
	}
	else
	{
	    gamma100 = darkLog100;
	}
    }
    else if ((min100 <= def.dark100) && (max100 <= def.bright100))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is too dark\n");
#endif
/*
	if (hint100 != (char *) 0)
	    free(hint100);
	hint100 = strdup("Image is too dark");
*/
	/*
	 * * old calculation: *   gamma100 = brightLog100 /
	 * log((Float64) max100/(Float64) (localTableSize-1)); *
	 * invert histogram for dark images to maintain symmetry
	 */
	doInvert = 1;
	min100 = localWhite - (arr100[2] * (localTableSize - 1)) / localWhite;
	max100 = localWhite - (arr100[1] * (localTableSize - 1)) / localWhite;
	temp_log = log((Float64) min100 / (Float64) (localTableSize - 1));
	if (temp_log != 0.0)
	{
	    gamma100 = darkLog100 / temp_log;
	}
	else
	{
	    gamma100 = darkLog100;
	}
    }
    else if ((min100 >= def.dark100) && (max100 <= def.bright100))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is screwed up\n");
#endif
/*
	if (hint100 != (char *) 0)
	    free(hint100);
	hint100 = strdup("Image is screwed up");
*/
	gamma100 = 1.0;
    }
    if ((min100 == max100) || (min90 == max90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image, what image?\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	if (hint100 != (char *) 0)
	    free(hint100);
	hint90 = strdup("Image, what image?");
	hint100 = strdup("Image, what image?");
*/
	gamma100 = 1.0;
	gamma90 = 1.0;
    }

    /* Limit gamma if few colors/grays */
    if (uniqueLevels <= 16) {
	if (gamma90 > 1.0) {
	    gamma90 = 1.0;
	}
	if (gamma100 > 1.0) {
	    gamma100 = 1.0;
	}
    } else if (uniqueLevels <= 64) {
	if (gamma90 > 1.75) {
	    gamma90 = 1.75;
	}
	if (gamma100 > 1.75) {
	    gamma100 = 1.75;
	}
    }

#ifdef PRINT_MESSAGE
    fprintf(stderr, "Gamma90 = %f, Gamma100 = %f\n", gamma90, gamma100);
#endif
    adjust = 0;
    if ((gamma100 - gamma90 > def.gammaConfidence) || (gamma90 - gamma100 > def.gammaConfidence))
	adjust = 1;

    if (gamma100 < def.minGamma100)
	gamma100 = def.minGamma100;
    if (gamma100 > def.maxGamma100)
	gamma100 = def.maxGamma100;

    if (gamma90 > def.maxGamma90)
	gamma90 = def.maxGamma90;
    if (gamma90 < 0.0)
	gamma90 = gamma100;	/* order dependent */
    if (gamma90 < def.minGamma90)
	gamma90 = def.minGamma90;


    gammaSuggest = (Float64) ((gamma90 + gamma100) / 2.0);

#ifdef PRINT_MESSAGE
    fprintf(stderr, "Suggest = %f\n", gammaSuggest);
#endif
    if (adjust)
    {
	gammaSuggest = gamma100;
#ifdef PRINT_MESSAGE
	fprintf(stderr, "changing gamma value to %f\n", gammaSuggest);
#endif
    }

    if (gammaSuggest != 1.0 && pGammaTRC)
    {
	ip_tableApplyGamma (pIPState, gammaSuggest,
			    pGammaTRC, pGammaTRC, doInvert);

	/* No, don't do this, the caller will apply the curve. */
	/* apply_gamma(gammaSuggest, trc_r_gamma, table_size, trc_r_gamma); */
	/* apply_gamma(gammaSuggest, trc_g_gamma, table_size, trc_g_gamma); */
	/* apply_gamma(gammaSuggest, trc_b_gamma, table_size, trc_b_gamma); */
    }
   
    if (pGammaComputed)
	*pGammaComputed = gammaSuggest;

    ip_tableFree (pIPState, globalHistogram);
    NORMAL_RETURN;
}

/*******************************************************************************
 * i_aieExposure_init
 *
 *	Initialize banded exposure processing. The only thing this module
 *	needs is a histogram, and if the contrast module is also running,
 *	we should let it take the histogram.
 *
 * Arguments:
 *	pData	    -	The "intermediate data values" structure, which 
 *			should have the pointer to the "defaults" structure.
 *			(This is returned by i_aieAnalyzeXXX_init)
 *	takeHisto   -	If non-zero, compute a global histogram in the _data
 *			function. One only does this if one is not also
 *			passing the data to the contrast analyzer, as
 *			it computes the global histogram.
 * Returns:
 *	ia_successful, ia_nomem, etc.
 *
 *******************************************************************************/

Int32 CDECL 
i_aieExposure_init (
    IPState    *pIPState,
    aIntermediates *pData,
    Bool	takeHisto)

{
    PROLOGUE;
    pData->xpoTakesHisto = takeHisto;

    if (takeHisto)
    {
	ERR_PASS(ip_tableAllocate (pIPState, 256, &(pData->globalHistogram)));
    }

    NORMAL_RETURN;
}

/***********************************************************************
 * 
 * i_aieExposure_data
 *
 *	Pass a band of data through the exposure analyzer.
 *	If xpoTakesHisto true, accumulate the data into the global
 *	histogram.
 *
 * Arguments:
 *	pData	    -	    The "intermediate data" structure
 *	pLuminance  -	    Pointer to some luminance data (frame)
 *	lines	    -	    Number of lines to process
 *
 * Returns:
 *	ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
i_aieExposure_data (
    IPState    *pIPState,
    aIntermediates *pData,
    UInt8      *pLuminance,
    Int32	lines)
{
    /* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
     * NOTE: this hasn't had the scale function factored in yet
     * XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX */

    PROLOGUE;
    if (pData->xpoTakesHisto)
    {
	/* i_histo doesn't clear the histogram first, so we can use it
	   to accumulate histograms */
	ExceptPass (i_histoRect (pIPState, pLuminance, pData->bpl,
				 0, 0, pData->w,
				 lines,
			         ip_tableData (pIPState,
					       pData->globalHistogram)));
    }

    NORMAL_RETURN;
}

/***********************************************************************
 *
 * i_aieExposure_finish
 *
 *	Finish computing the exposure information.
 *	We look at the 90% and 100% range of the luminance histogram and
 *	generate 2 gamma values.  If these values are somewhat reasonable,
 *	we will simply average the 2 values for a suggested gamma value.
 *	We also set hint90 and hint100 strings for future use with saturation.
 *	In this analyzer, we are looking at the global histogram.
 *
 * notes:
 *	Some documentation implied that the "hint" stuff would be useful;
 *	nobody uses it right now, so it's not exported. If it is to be
 *	exported, the hints should be changed from text to constants.
 *	(English-language text in code is a pain when one wants to 
 *	internationalize)
 * 
 * Arguments:
 *	pData	    -	    "Intermediate Data" structure
 *	pGammaTRC   -	    Resulting TRC which should correct for exposure. 
 *			    Should be initialized to linear as passed.
 *	arr100      -	    100% rage, exported so that contrast can use it
 *	pVariance   -	    the variance as computed
 *	pGammaComputed -    place to stick the gamma value
 *	
 * Returns:
 *	ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
i_aieExposure_finish (
    IPState    *pIPState,
    aIntermediates *pData,
    aTable     *pGammaTRC,
    Int32      *arr100,
    Float64    *pVariance,
    Float64    *pGammaComputed)
{
    /* We need to make a copy of the defaults, since we trash them */
    PROLOGUE;
    aExposureDefaults def;
    Int32           i, j;
    Int32           minPixVal = 255;
    Int32           maxPixVal = 0;
    Int32	    minVal, maxVal;
    Int32           topThreshold;
    Int32           bottomThreshold;
    Int32           arr[4];
    Int32	    arr50[4], arr60[4], arr70[4], arr80[4], arr90[4];
    Int32          *hd;
    Int32          *histp;
    Float64         sum;
    Float64         sqsum;
    Float64         sigma2;
    Float64         average;
    Int32           min90, max90;
    Int32           min100, max100;
    Float64         darkLog90, brightLog90;
    Float64         darkLog100, brightLog100;
    Int32           adjust;
    Int32           localWhite = 255;
    Int32           localTableSize = 256;
    Float64         gamma90 = -1.0, gamma100 = -1.0, gammaSuggest;
    Float64	    temp_log;
    Bool	    doInvert = 0;

    def = pData->defaults->exposure;
		      
#ifdef PRINT_MESSAGE
    fprintf(stderr, "/*********************/\n");
    fprintf(stderr, "/* exposure analysis */\n");
    fprintf(stderr, "/*********************/\n");
#endif

    /* Go through and find a min/max in the histogram (well, really
       a first and last non-zero value) */
    minVal = -1;
    hd = histp = ip_tableData (pIPState, pData->globalHistogram);

    for (i = 0; i < 256; i++)
    {
	if ((hd[i] != 0) && (minVal == -1))
	    minVal = i;
	if (hd[i] != 0)
	    maxVal = i;
    }

    /* Compute the sum and sum-of-squares for the histogram */
    sum = sqsum = 0L;
    for (i = 0; i < 256; i++)
    {
	sum += (Float64) hd[i];
	sqsum += (Float64) ((Float64) hd[i] * (Float64) hd[i]);
    }

    /* 
     * Compute the mean and "normalized" variance for the histogram. 
     * Here, "normalized" variance means that it was multiplied by a large number
     */
    average = sum / 256.0;
    sigma2 = sqsum / 256.0 - average * average;
    sigma2 = sigma2 * (1024.0 * 1024.0) / (sum * sum);
    *pVariance = sigma2;

#ifdef PRINT_MESSAGE
    fprintf(stderr, "normalized variance: %15d\n", (int) sigma2);
#endif

    /* 
     * Compute min/max/range values for 100, 90, 80, 70, 60, and 50%
     * of the histogram. For example, in the 90% case, we would exclude
     * 5% on either end. This is 5% by "volume", that is, percentage of the
     * total sum of the histogram. 
     */
    for (j = 0; j < 6; j++)
    {
	topThreshold = bottomThreshold = 0;

	/* Search up from the bottom until we've accumulated j/2 percent */
	for (i = 0; i < 256; ++i)
	{
	    bottomThreshold += hd[i];
	    if (bottomThreshold > ((sum * j) / 20))
	    {
		minPixVal = i;
		break;
	    }
	}

	/* Search down from the top until we've accumulated j/2 percent */
	histp = hd + 256 - 1;
	for (i = 256 - 1; i >= 0; i--)
	{
	    topThreshold += hd[i];
	    if (topThreshold > ((sum * j) / 20))
	    {
		maxPixVal = i;
		break;
	    }
	}

	arr[0] = 100 - j * 10;	/* Percent of power */
	arr[1] = minPixVal;
	arr[2] = maxPixVal;
	arr[3] = (maxPixVal - minPixVal) + 1;

	/* Crikey. I'll have to clean this up at some point */
	if (arr[0] == 50)
	{
	    arr50[0] = arr[0];
	    arr50[1] = arr[1];
	    arr50[2] = arr[2];
	    arr50[3] = arr[3];
	}
	else if (arr[0] == 60)
	{
	    arr60[0] = arr[0];
	    arr60[1] = arr[1];
	    arr60[2] = arr[2];
	    arr60[3] = arr[3];
	}
	else if (arr[0] == 70)
	{
	    arr70[0] = arr[0];
	    arr70[1] = arr[1];
	    arr70[2] = arr[2];
	    arr70[3] = arr[3];
	}
	else if (arr[0] == 80)
	{
	    arr80[0] = arr[0];
	    arr80[1] = arr[1];
	    arr80[2] = arr[2];
	    arr80[3] = arr[3];
	}
	else if (arr[0] == 90)
	{
	    arr90[0] = arr[0];
	    arr90[1] = arr[1];
	    arr90[2] = arr[2];
	    arr90[3] = arr[3];
	}
	else if (arr[0] == 100)
	{
	    arr100[0] = arr[0];
	    arr100[1] = arr[1];
	    arr100[2] = arr[2];
	    arr100[3] = arr[3];
	}
#ifdef PRINT_MESSAGE
	fprintf(stderr, "%3ld%% - low %3ld, high %3ld, range %3ld\n",
		100 - j * 10, minPixVal, maxPixVal, (maxPixVal - minPixVal) + 1);
#endif
    }

    def.dark90 = (Int32) (((Float64) def.dark90 / (Float64) (localWhite + 1)) * localTableSize);
    def.bright90 = (Int32) (((Float64) def.bright90 / (Float64) (localWhite + 1)) * localTableSize);
    def.dark100 = (Int32) (((Float64) def.dark100 / (Float64) (localWhite + 1)) * localTableSize);
    def.bright100 = (Int32) (((Float64) def.bright100 / (Float64) (localWhite + 1)) * localTableSize);

    if (def.dark90 != 0)
    {
	darkLog90 = log((Float64) def.dark90 /
			(Float64) (localTableSize - 1));
    }
    else
    {
	darkLog90 = 0;
    }
    if (def.bright90 != 0)
    {
	brightLog90 = log((Float64) def.bright90 /
			  (Float64) (localTableSize - 1));
    }
    else
    {
	brightLog90 = 0;
    }
    if (def.dark100 != 0)
    {
	darkLog100 = log((Float64) def.dark100 /
			 (Float64) (localTableSize - 1));
    }
    else
    {
	darkLog100 = 0;
    }
    if (def.bright100 != 0)
    {
	brightLog100 = log((Float64) def.bright100 /
			   (Float64) (localTableSize - 1));
    }
    else
    {
	brightLog100 = 0;
    }

    min100 = (arr100[1] * (localTableSize - 1)) / localWhite;
    max100 = (arr100[2] * (localTableSize - 1)) / localWhite;

    min90 = (arr90[1] * (localTableSize - 1)) / localWhite;
    max90 = (arr90[2] * (localTableSize - 1)) / localWhite;

    if ((min90 <= def.dark90) && (max90 >= def.bright90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is fine\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	hint90 = strdup("Image is fine");
*/
	gamma90 = 1.0;
    }
    else if ((min90 >= def.dark90) && (max90 >= def.bright90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is too light\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	hint90 = strdup("Image is too light");
*/
	temp_log = log((Float64) min90 / (Float64) (localTableSize - 1));
	if (temp_log != 0.0)
	{
	    gamma90 = darkLog90 / temp_log;
	}
	else
	{
	    gamma90 = darkLog90;
	}
    }
    else if ((min90 <= def.dark90) && (max90 <= def.bright90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is too dark\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	hint90 = strdup("Image is too dark");
*/
	/*
	 * * old calculation: *   gamma90 = brightLog90 / log((Float64)
	 * max90/(Float64) (localTableSize-1)); * invert histogram
	 * for dark images to maintain symmetry
	 */
	doInvert = 1;
	min90 = localWhite - (arr90[2] * (localTableSize - 1)) / localWhite;
	max90 = localWhite - (arr90[1] * (localTableSize - 1)) / localWhite;
	temp_log = log((Float64) min90 / (Float64) (localTableSize - 1));
	if (temp_log != 0.0)
	{
	    gamma90 = darkLog90 / temp_log;
	}
	else
	{
	    gamma90 = darkLog90;
	}
    }
    else if ((min90 >= def.dark90) && (max90 <= def.bright90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is screwed up\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	hint90 = strdup("Image is screwed up");
*/
	gamma90 = -1;
    }

    if ((min100 <= def.dark100) && (max100 >= def.bright100))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is fine\n");
#endif
/*
	if (hint100 != (char *) 0)
	    free(hint100);
	hint100 = strdup("Image is fine");
*/
	gamma100 = 1.0;
    }
    else if ((min100 >= def.dark100) && (max100 >= def.bright100))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is too light\n");
#endif
/*
	if (hint100 != (char *) 0)
	    free(hint100);
	hint100 = strdup("Image is too light");
*/
	temp_log = log((Float64) min100 / (Float64) (localTableSize - 1));
	if (temp_log != 0.0)
	{
	    gamma100 = darkLog100 / temp_log;
	}
	else
	{
	    gamma100 = darkLog100;
	}
    }
    else if ((min100 <= def.dark100) && (max100 <= def.bright100))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is too dark\n");
#endif
/*
	if (hint100 != (char *) 0)
	    free(hint100);
	hint100 = strdup("Image is too dark");
*/
	/*
	 * * old calculation: *   gamma100 = brightLog100 /
	 * log((Float64) max100/(Float64) (localTableSize-1)); *
	 * invert histogram for dark images to maintain symmetry
	 */
	doInvert = 1;
	min100 = localWhite - (arr100[2] * (localTableSize - 1)) / localWhite;
	max100 = localWhite - (arr100[1] * (localTableSize - 1)) / localWhite;
	temp_log = log((Float64) min100 / (Float64) (localTableSize - 1));
	if (temp_log != 0.0)
	{
	    gamma100 = darkLog100 / temp_log;
	}
	else
	{
	    gamma100 = darkLog100;
	}
    }
    else if ((min100 >= def.dark100) && (max100 <= def.bright100))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image is screwed up\n");
#endif
/*
	if (hint100 != (char *) 0)
	    free(hint100);
	hint100 = strdup("Image is screwed up");
*/
	gamma100 = 1.0;
    }
    if ((min100 == max100) || (min90 == max90))
    {
#ifdef PRINT_MESSAGE
	fprintf(stderr, "Image, what image?\n");
#endif
/*
	if (hint90 != (char *) 0)
	    free(hint90);
	if (hint100 != (char *) 0)
	    free(hint100);
	hint90 = strdup("Image, what image?");
	hint100 = strdup("Image, what image?");
*/
	gamma100 = 1.0;
	gamma90 = 1.0;
    }

    /* Limit gamma if few colors/grays */
    if (pData->numGrays <= 16) {
	if (gamma90 > 1.0) {
	    gamma90 = 1.0;
	}
	if (gamma100 > 1.0) {
	    gamma100 = 1.0;
	}
    } else if (pData->numGrays <= 64) {
	if (gamma90 > 1.75) {
	    gamma90 = 1.75;
	}
	if (gamma100 > 1.75) {
	    gamma100 = 1.75;
	}
    }

#ifdef PRINT_MESSAGE
    fprintf(stderr, "Gamma90 = %f, Gamma100 = %f\n", gamma90, gamma100);
#endif
    adjust = 0;
    if ((gamma100 - gamma90 > def.gammaConfidence) || (gamma90 - gamma100 > def.gammaConfidence))
	adjust = 1;

    if (gamma100 < def.minGamma100)
	gamma100 = def.minGamma100;
    if (gamma100 > def.maxGamma100)
	gamma100 = def.maxGamma100;

    if (gamma90 > def.maxGamma90)
	gamma90 = def.maxGamma90;
    if (gamma90 < 0.0)
	gamma90 = gamma100;	/* order dependent */
    if (gamma90 < def.minGamma90)
	gamma90 = def.minGamma90;


    gammaSuggest = (Float64) ((gamma90 + gamma100) / 2.0);

#ifdef PRINT_MESSAGE
    fprintf(stderr, "Suggest = %f\n", gammaSuggest);
#endif
    if (adjust)
    {
	gammaSuggest = gamma100;
#ifdef PRINT_MESSAGE
	fprintf(stderr, "changing gamma value to %f\n", gammaSuggest);
#endif
    }

    if (gammaSuggest != 1.0 && pGammaTRC)
    {
	ip_tableApplyGamma (pIPState, gammaSuggest,
			    pGammaTRC, pGammaTRC, doInvert);
    }
   
    if (pGammaComputed)
	*pGammaComputed = gammaSuggest;

    if (pData->xpoTakesHisto)
    {
	ip_tableFree (pIPState, pData->globalHistogram);
    }
    NORMAL_RETURN;

}

/*

$Log: i_aiexpo.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.13  1998/01/13 23:47:01  ddavies
i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.12  1997/11/13 23:21:17  aber
i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v

CommonCheck: Updated the AIE algorithms to match (or closely match) the latest DITC
AIE used in Sandpiper. The updates include the following:
1. Exclude a frame (5%) around the exterior in almost all statistics.
2. Make sure the black-point and white-point are included in the
   computations.
3. ... this was improved upon to not use just the first region with a
   black-or-white-point, but to use the ones with the largest variance.
4. The statistics are accumulated in cells 4 across and 4 down. This
   is not done if the width and height are less than 512.
5. Black-point and white-point do not include 0 and 255 respectively.
   They are the darkest and whitest not quite at these saturation points.
6. If the number of colors or gray levels is too low, less aggressive
   correction is applied.
7. Lots of thresholds changed.
8. Luminance channel is computed with a different formula.

Revision 1.11  1997/07/03 21:52:43  jcraig
i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v

use i_histoRect instead of i_histo.

Revision 1.10  1995/12/13 02:42:01  ejaquith
i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.9  1995/10/13  23:49:18  dferrell
 * i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.8  1994/10/05  22:20:41  macleod
 * i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v
 *
 * Added support for banded AIE
 *
 * Revision 1.7  1994/07/12  23:55:33  ejaquith
 * i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v
 *
 * removed #include "iadef.prv".
 *
 * Revision 1.6  1994/02/04  20:13:59  dferrell
 * i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.5  1994/01/05  23:55:58  dferrell
 * dferrell on Wed Jan 5 15:54:36 PST 1994
 *
 * i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v
 *
 * Commented out references to hint90 and hint100 including the use of the
 * non-ANSI function strdup ().  The file now compiles on the Mac.
 *
 * Revision 1.4  1993/12/21  23:24:53  macleod
 * i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v
 *
 * Exposure computation can now export the computed gamma value
 *
 * Revision 1.3  1993/12/02  19:22:33  ddavies
 * i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:27:19  ddavies
 * i_aiexpo.c -> /product/ipcore/ipshared/src/RCS/i_aiexpo.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

