/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_allrn1.c: routines to find all the runs in a scan line.
 *
 *		Int32	i_findAllRunsInLine()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "imageref.pub"
#include "defines.pub"
#define	_USE_FIRST_BIT_TABLE_		/* want both tables from iopbyte.prv */
#include "iopbyte.prv"
#include "lmask.prv"			/* contains leftMask[] */
#include "rmask.prv"			/* contains rightMask[] */


/* prototypes */
#include "shrcc.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_allrn1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* It is more efficient to find all the runs of 1's in a scan line at
 * once rather than find them one at a time, as is done with
 * i_nextRunToTheRight.  i_findAllRunsInLine returns a linked list of
 * all the runs in a single scan line.
 *
 * This routine is a lot like i_nextRunToTheRight, except that it
 * allocates new linked list elements as necessary (it can use an
 * existing list, just adding to the end as needed), and it loops
 * around after finding the right end of a run.
 *
 * One of the obvious efficiencies of an entire run at once is
 * that one ought to be able to take advantage of the fact that if a run
 * begins and ends in a single word, and it begins at bit X,
 * it will end somewhere to the right of bit X.  That is, we don't need to
 * do a completely general search of the word to find the end of the run.
 * One of the efficiencies of finding all the runs at once is that we
 * can do the same if we know that a new run starts in the same word
 * where an old run ended.  But how do we implement a fancy searcher?
 *
 * We could build a nice state machine with state numbers and a switch
 * statement at the top.  This is lovely, but sort of slow for something
 * we'll be doing a lot.  Just to be daring, I'm going to use, gasp,
 * goto's.
 *
 * The main loop will contain four sections:
 *   1.	Skip over words containing all 0's.
 *   2.	Find the left end of a run.
 *   3. Skip over words containing all 1's.
 *   4. Find the right end of a run and allocate a new run structure
 *	on the linked list if necessary.
 *
 * Clearly, if you know at the end of step 2 that the run ends in the
 * same work where it started (as is often the case), you should be able
 * to skip step 3 entirely.  Furthermore, if you know that the run began
 * in byte X and it finishes in the same word, you need only look in
 * byte X and bytes to the right.  Therefore, if this case occurs and we
 * found the beginning of the run in byte 0, we go to the stage of step 4
 * that starts looking for the end in byte 0.  Similarly, if the run
 * starts in byte 1 and ends in the same word, we go to the stage in step
 * 4 that looks in byte 1 and so on.
 *
 * Similarly, if we find the end of a run in step 4 and know that a new
 * run starts in the same word, we go to the portion of step 1 that starts
 * looking for the run in the appropriate byte of the word.
 * Using goto's to do this is unclean, but seems significantly faster than
 * any alternative I've been able to find.
 *
 * The normal path through steps 2 and 4 does a binary search on the bytes
 * of the word in question, then uses a table to find the bit within the
 * selected byte.  The bits to the left of the point found are filled in
 * in step 2 and cleared in step 4.  This allows us to find the end of the
 * run or beginning of the next run without being confused by the previous
 * run.
 */

/* PUBLIC */
/**************************************************************************
 *   i_findAllRunsInLine()
 *
 *	This procedure starts from the given offset in a scan line and
 *	looks to the right for runs of 1's.  The coordinates of each
 *	run found are returned in a linked list of TCScanLineRun
 *	structures.
 *
 *	THIS PROCEDURE ONLY WORKS ON BINARY IMAGES.  Since we expect
 *	to be calling it zillions of times, it doesn't check the image
 *	depth.
 *
 *	If there is a 1 at the initialOffset, this routine will return
 *	the initialOffset as the first 1.
 *
 *	pList must point to a non-null list of TCScanLineRun's.  This
 *	routine reuses structures in the list as needed.  If it runs
 *	out of structures, it allocated more from the heap of
 *	TCScanLineRun's pointed to by pRunHeap.
 *
 *	Input:
 *		pS:		Pointer to leftmost 32-bit word in the
 *				scan line OUTSIDE the frame.
 *		rightLimit:	rightmost pixel at which we should look
 *				for bits.  By definition, no run can
 *				begin or end to the right of this offset.
 *				Runs ending beyond this offset will be
 *				chopped off here.  This is relative to the
 *				OUTSIDE of the frame.
 *		initialOffset:	starting offset in line relative to 
 *				left edge of image OUTSIDE frame.
 *		pList:		pointer to list of TCScanLineRuns that will
 *				hold the output.  Coordinates of these runs
 *				will be relative to the OUTSIDE of the frame.
 *		pRunHeap:	pointer to heap of TCScanLineRuns.  This
 *				heap is created using i_allocateMemoryBlock
 *				and freed using i_freeAllMemoryBlocks.
 *
 *	Returns:
 *		ia_successful		found a run
 *		ia_invalidParm		pList or pRunHeap was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_findAllRunsInLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	rightLimit,
    Int32	initialOffset,
    TCScanLineRun *pList,
    TCBlockHeader *pRunHeap)
{
    PROLOGUE;
    Int32	 byteOffset;	    /* byte offset into the source line */
    Int32	 rightByteLimit;    /* byte offset of the word holding the
				     * rightmost pixel */
    UInt32	 inputWord;	    /* the word we read from the image */
    Int32	 theEnd;	    /* holds fragment of word containing the
				     * end of a run; or it contains the
				     * numerical offset of this run from
				     * the end of the word
				     */


	/* See if we were given a valid list */
    if ((pList == NULL) || (pRunHeap == NULL))
	return ia_invalidParm;

    pList->pClient = NULL;	/* initialize the very first element of the
				 * list.  The other are initialized as we
				 * go */
	/* calculate the byte offset of the word holding the initial offset */
    byteOffset = (initialOffset/8) & ~0x3;
	/* turn rightLimit into the offset of the word holding the last byte */
    rightByteLimit = (rightLimit/8) & ~0x3;
    if (byteOffset > rightByteLimit)	/* we're done already */
	goto finishFindAllRunsInLine;

    inputWord = ULONG_ACCESS(&pS[byteOffset]);	/* get the first word */

	/* mask off the bits on the left that we should ignore */
    inputWord &= rightMask[(~initialOffset) & 0x1F];
    if (inputWord != 0)
	goto findLeftEnd;	/* a run starts in the very first word */

	/* MAIN LOOP */
	/* In this loop, we skip words full of 0's, find the left end
	 * of a run, skip words full of 1's and find the right end of
	 * a run.  We also allocate new TCScanLineRun's as required.
	 */
    while (byteOffset <= rightByteLimit)
    {
	    /* STEP 1: SKIP OVER WORDS OF 0's */

	    /* We only get here if inputWord == 0, so go look at the
	     * next word.
	     * BUG FIX: Well, yes, inputWord might equal 0, but
	     * that doesn't mean that we're not at the right end
	     * now.  So we can't always advance byteOffset blindly.
	     * On the other hand, we can't stop just because
	     * the memory location at byteOffset holds 0, because
	     * it might not.  Remember that the current value of inputWord
	     * has been modified, so may not reflect what's in memory.
	     * So, we only advance if we're not already at the end of the line.
	     */
	if (byteOffset < rightByteLimit)
	{
	    byteOffset += 4;
	    while ( (byteOffset < rightByteLimit) &&
		    (ULONG_ACCESS(&pS[byteOffset]) == 0) )
		byteOffset += 4;

		/* At this point, we're either pointing at a non-zero word or
		 * at the last word of the scan line (or both).  Put this
		 * word into inputWord.  If it's 0, just stop.
		 */
	    inputWord = ULONG_ACCESS(&pS[byteOffset]);
	}
	if (inputWord == 0)
	    goto finishFindAllRunsInLine;


	    /* STEP 2: FIND LEFT END OF RUN */

	    /* Find the leftmost 1.  Do this by locating the 1 down to
	     * the byte, then using tables to find it in a byte.
	     * Note that we also set the 0's to the left of the leftmost
	     * 1 to 1.  This prevents us from finding them when we
	     * look for the first 0.
	     */
findLeftEnd:
	if ((inputWord & 0xFFFF0000) != 0)
	{
	    theEnd = inputWord >> 16;	/* get the upper half.  We use
					 * theEnd so we don't mess up
					 * inputWord.  We'll need it
					 * below to find the right
					 * end of the run
					 */
	    if ((theEnd & 0xFF00) != 0)
	    {
		    /* Left End of Run is in Byte 0 */
    
		theEnd = firstBitTable[theEnd >> 8];	/* get offset within
							 * word */
		inputWord |= leftMask[theEnd];
		theEnd += (byteOffset * 8);
		    /* if that was off the right end of the area we're
		     * allowed to search, just quit.
		     */
		if (theEnd > rightLimit)
		    goto finishFindAllRunsInLine;
		else
		    pList->runLeft = (Int16) theEnd;

		    /* fill in white to the left of the run so we can find the
		     * right end of the run.  If the run ends in this word,
		     * skip the section of code that skips over entire words
		     * filled with 1's.
		     */
		if (inputWord != 0xFFFFFFFF)
		{
		    inputWord = ~inputWord;
		    goto findRightEnd;
		}
	    }
	    else
findLeftEndByte1:	/* theEnd contains inputWord >> 16 */
	    {
		    /* Left End of Run in Byte 1 */
		theEnd = 8 + firstBitTable[theEnd & 0xFF];
		inputWord |= leftMask[theEnd];
		theEnd += (byteOffset * 8);
		    /* if that was off the right end of the area we're
		     * allowed to search, just quit.
		     */
		if (theEnd > rightLimit)
		    goto finishFindAllRunsInLine;
		else
		    pList->runLeft = (Int16) theEnd;

		    /* if inputWord is now not all 1's, find the other end
		     * of the run by starting to look in byte 1.
		     */
		if (inputWord != 0xFFFFFFFF)
		{
		    inputWord = ~inputWord;
		    if ((inputWord & 0xFFFF0000) != 0)	/* end is in byte 1 */
		    {
			theEnd = inputWord >> 16;
			goto findRightEndByte1;
		    }
		    else
			goto findRightEndBytes23;
		}
	    }
	}
	else	/* The left end of the run starts in the lower
		 * half of the word */
findLeftEndBytes23:
	{
	    if ((inputWord & 0xFF00) != 0)
	    {
		    /* Left End of Run in Byte 2 */
    
		theEnd = 16 + firstBitTable[inputWord >> 8];
		inputWord |= leftMask[theEnd];
		theEnd += (byteOffset * 8);
		    /* if that was off the right end of the area we're
		     * allowed to search, just quit.
		     */
		if (theEnd > rightLimit)
		    goto finishFindAllRunsInLine;
		else
		    pList->runLeft = (Int16) theEnd;

		    /* if inputWord is now not all 1's, find the other end
		     * of the run by starting to look in byte 2.
		     */
		if (inputWord != 0xFFFFFFFF)
		{
		    inputWord = ~inputWord;
		    goto findRightEndBytes23;
		}
	    }
	    else
findLeftEndByte3:
	    {
		    /* Left End of Run in Byte 3 */

		theEnd = 24 + firstBitTable[inputWord & 0xFF];
		inputWord |= leftMask[theEnd];
		theEnd += (byteOffset * 8);
		    /* if that was off the right end of the area we're
		     * allowed to search, just quit.
		     */
		if (theEnd > rightLimit)
		    goto finishFindAllRunsInLine;
		else
		    pList->runLeft = (Int16) theEnd;

		    /* if inputWord is now not all 1's, find the other end
		     * of the run by starting to look in byte 3.
		     */
		if (inputWord != 0xFFFFFFFF)
		{
		    inputWord = ~inputWord;
		    goto findRightEndByte3;
		}
	    }
	}

	    /* STEP 3: SKIP OVER WORDS OF 1's */

	    /* If we get here, inputWord contains all 1's.  Keep looking
	     * at input words until we either run out of room or
	     * we find one with some 0's in it.  That marks the right
	     * end of the run.
	     * BUG FIX: The current word might contain all 1's AND
	     * be the last word of the line, so we can't blindly
	     * advance byteOffset.
	     * On the other hand, we can't stop just because
	     * the memory location at byteOffset doesn't hold all 1's, because
	     * it might not.  Remember that the current value of inputWord
	     * has been modified, so may not reflect what's in memory.
	     * On the third hand, we have to do something to detect the
	     * case when the last run really goes to the right boundary,
	     * In my first bug "fix", inputWord contained all 1's
	     * if we started at the right boundary and it contained
	     * all 0's if we didn't.  Hmmm.
	     * Ok, so we always advance byteOffset and just don't
	     * do the read if byteOffset >= rightByteLimit.  Then we
	     * check to see if inputWord == 0 OR byteOffset >= rightByteLimit.
	     * In either case, the run ends at the right boundary.
	     */
	byteOffset += 4;
	while ( (byteOffset < rightByteLimit) &&
		(ULONG_ACCESS(&pS[byteOffset]) == 0xFFFFFFFF) )
	    byteOffset += 4;

	    /* At this point, we're either pointing at a non-1's word or
	     * at the last word of the scan line (or both) OR we're
	     * pointing one word beyond the last word if we started
	     * in the last word.  If there's a word here, put it's
	     * inverse into inputWord.  If that inverse is all
	     * 0's or we're now beyond the rightLimit, return rightLimit
	     * as the right end of the run.
	     */
	if (byteOffset <= rightByteLimit)
	{
	    inputWord = ~(ULONG_ACCESS(&pS[byteOffset]));
	}
	else
	{
	    inputWord = 0;	/* our way of saying that the run ended
				 * at the rightLimit.
				 */
	}

	if (inputWord == 0)
	{
	    pList->runRight = (Int16) rightLimit;
	    if (pList->pNext == NULL)
	    {
		__error__ = i_allocateStruct (pIPState, pRunHeap,
					    (void **)&(pList->pNext));
		if (__error__ != ia_successful)
		    goto finishFindAllRunsInLineWithError;
		pList = pList->pNext;
		pList->pNext = NULL;
		pList->pClient = NULL;
	    }
	    else
		pList = pList->pNext;

	    goto finishFindAllRunsInLine;
	}


	    /* STEP 4: FIND RIGHT END OF RUN */

	    /* At this point, inputWord has been inverted and does not
	     * contain all 0's.  Just start looking for the first 1's.
	     */
	    /* Find the rightmost 1 in the real word, or the (leftmost
	     * 1 in the inverted word) - 1.  Do this by locating the 1 down to
	     * the byte, then using tables to find it in a byte.
	     * When we're done finding the end of the run, we set the
	     * entire run to 0's so we can find the start of the next
	     * run..
	     */
findRightEnd:	/* inputWord has already been complemented */
	if ((inputWord & 0xFFFF0000) != 0)
	{
	    theEnd = inputWord >> 16;	/* get the upper half.  We use
					 * theEnd so we don't mess up
					 * inputWord.  We'll need it
					 * to find the next run.
					 */
	    if ((theEnd & 0xFF00) != 0)
	    {
		    /* Right End of Run is in Byte 0 */

		    /* Before doing anything else, allocate a new
		     * run if necessary.  We'll need it.
		     */
		if (pList->pNext == NULL)
		{
		    __error__ = i_allocateStruct (pIPState, pRunHeap,
						(void **)&(pList->pNext));
		    if (__error__ != ia_successful)
			goto finishFindAllRunsInLineWithError;
		    pList->pNext->pNext = NULL;
		}
		theEnd = firstBitTable[theEnd >> 8];	 /* get offset within
							  * word */
		inputWord = ~(inputWord | leftMask[theEnd]);
		theEnd += (byteOffset * 8) - 1;
		    /* if that was off the right end of the area we're
		     * allowed to search, set the limit to rightLimit
		     * and quit.
		     */
		if (theEnd > rightLimit)
		{
		    pList->runRight = (Int16) rightLimit;
		    pList = pList->pNext;
		    pList->pClient = NULL;
		    goto finishFindAllRunsInLine;
		}
		else
		{
		    pList->runRight = (Int16) theEnd;
		    pList = pList->pNext;
		    pList->pClient = NULL;
		}

		    /* if a new run starts in this word, go find it.
		     */
		if (inputWord != 0)
		    goto findLeftEnd;
	    }
	    else
findRightEndByte1:  /* inputWord has been complemented and theEnd contains
		     * (complemented inputWord) >> 16
		     */
	    {
		    /* Right End of Run is in Byte 1 */
		    /* Before doing anything else, allocate a new
		     * run if necessary.  We'll need it.
		     */
		if (pList->pNext == NULL)
		{
		    __error__ = i_allocateStruct (pIPState, pRunHeap,
						(void **)&(pList->pNext));
		    if (__error__ != ia_successful)
			goto finishFindAllRunsInLineWithError;
		    pList->pNext->pNext = NULL;
		}

		theEnd = 8 + firstBitTable[theEnd & 0xFF];
		inputWord = ~(inputWord | leftMask[theEnd]);
		theEnd += (byteOffset * 8) - 1;
		    /* if that was off the right end of the area we're
		     * allowed to search, set the limit to rightLimit
		     * and quit.
		     */
		if (theEnd > rightLimit)
		{
		    pList->runRight = (Int16) rightLimit;
		    pList = pList->pNext;
		    pList->pClient = NULL;
		    goto finishFindAllRunsInLine;
		}
		else
		{
		    pList->runRight = (Int16) theEnd;
		    pList = pList->pNext;
		    pList->pClient = NULL;
		}

		    /* if inputWord is now not all 0's, find the start
		     * of the next run by starting to look in byte 1.
		     */
		if (inputWord != 0)
		{
		    if ((inputWord & 0xFFFF0000) != 0)	/* start is in byte 1 */
		    {
			theEnd = inputWord >> 16;
			goto findLeftEndByte1;
		    }
		    else
			goto findLeftEndBytes23;
		}
	    }
	}
	else	/* The right end of the run ends in the lower
		 * half of the word.  Remember that inputWord
		 * has already been complemented so we're looking
		 * for the leftmost 1's
		 */
findRightEndBytes23:	/* inputWord has already been complemented */
	{
	    if ((inputWord & 0xFF00) != 0)
	    {
		    /* Right End of Run is in Byte 2 */
    
		    /* Before doing anything else, allocate a new
		     * run if necessary.  We'll need it.
		     */
		if (pList->pNext == NULL)
		{
		    __error__ = i_allocateStruct (pIPState, pRunHeap,
						(void **)&(pList->pNext));
		    if (__error__ != ia_successful)
			goto finishFindAllRunsInLineWithError;
		    pList->pNext->pNext = NULL;
		}

		theEnd = 16 + firstBitTable[inputWord >> 8];
		inputWord = ~(inputWord | leftMask[theEnd]);
		theEnd += (byteOffset * 8) - 1;
		    /* if that was off the right end of the area we're
		     * allowed to search, set the limit to rightLimit
		     * and quit.
		     */
		if (theEnd > rightLimit)
		{
		    pList->runRight = (Int16) rightLimit;
		    pList = pList->pNext;
		    pList->pClient = NULL;
		    goto finishFindAllRunsInLine;
		}
		else
		{
		    pList->runRight = (Int16) theEnd;
		    pList = pList->pNext;
		    pList->pClient = NULL;
		}

		    /* if inputWord is now not all 0's, find the start
		     * of the next run by starting to look in byte 2.
		     */
		if (inputWord != 0)
		    goto findLeftEndBytes23;
	    }
	    else
findRightEndByte3:	/* inputWord has already been complemented */
	    {
		    /* Right End of Run is in Byte 3 */

		    /* Before doing anything else, allocate a new
		     * run if necessary.  We'll need it.
		     */
		if (pList->pNext == NULL)
		{
		    __error__ = i_allocateStruct (pIPState, pRunHeap,
						(void **)&(pList->pNext));
		    if (__error__ != ia_successful)
			goto finishFindAllRunsInLineWithError;
		    pList->pNext->pNext = NULL;
		}

		theEnd = 24 + firstBitTable[inputWord & 0xFF];
		inputWord = ~(inputWord | leftMask[theEnd]);
		theEnd += (byteOffset * 8) - 1;
		    /* if that was off the right end of the area we're
		     * allowed to search, set the limit to rightLimit
		     * and quit.
		     */
		if (theEnd > rightLimit)
		{
		    pList->runRight = (Int16) rightLimit;
		    pList = pList->pNext;
		    pList->pClient = NULL;
		    goto finishFindAllRunsInLine;
		}
		else
		{
		    pList->runRight = (Int16) theEnd;
		    pList = pList->pNext;
		    pList->pClient = NULL;
		}

		    /* if inputWord is now not all 0's, find the start
		     * of the next run by starting to look in byte 3.
		     */
		if (inputWord != 0)
		    goto findLeftEndByte3;
	    }
	}
    }	/* if we get to the bottom of the loop, the next run starts in
	 * some other word.  Go look for it.
	 */



finishFindAllRunsInLine:
    pList->runLeft = pList->runRight = cVeryLargeNum;
    __error__ = ia_successful;

finishFindAllRunsInLineWithError:
    NORMAL_RETURN;
}


/*

 $Log: i_allrn1.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:10  jlin
 Imported sources

 Revision 1.7  1998/01/13 23:47:03  ddavies
 i_allrn1.c -> /product/ipcore/ipshared/src/RCS/i_allrn1.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.6  1996/05/21 18:43:13  ddavies
 ddavies on Tue May 21 11:40:02 PDT 1996

 i_allrn1.c -> /product/ipcore/ipshared/src/RCS/i_allrn1.c,v

 Fix comments.  Fix bug introduced by last change.  We did the right thing
 when looking for the left endpoint of runs about not running off the right
 edge of the image, but introduced a bug when looking for right endpoints.
 We now always increment the byteOffset, only look at the word if it's in
 bounds, load the inverse of the ending word if its in bounds or load 0
 otherwise.  This has the effect of correctly stopping all runs that end
 on word boundaries and at the right edge of the image.

 Revision 1.5  1996/05/18 00:21:42  ddavies
 ddavies on Fri May 17 17:20:11 PDT 1996

 i_allrn1.c -> /product/ipcore/ipshared/src/RCS/i_allrn1.c,v

 Fix bug in i_findAllRunsInLine found by William Rucklidge.  It was possible
 to read off the right edge of frameless images.  We now check position
 before moving right to skip over words containing all 1's or all 0's.

 Revision 1.4  1995/12/13 02:42:05  ejaquith
 i_allrn1.c -> /product/ipcore/ipshared/src/RCS/i_allrn1.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/10/18  03:51:58  ddavies
 * i_allrn1.c -> /product/ipcore/ipshared/src/RCS/i_allrn1.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.2  1995/07/10  23:10:48  ddavies
 * ddavies on Mon Jul 10 16:09:03 PDT 1995
 *
 * i_allrn1.c -> /product/ipcore/ipshared/src/RCS/i_allrn1.c,v
 *
 * Change casts in calls to i_allocateMemoryBlock from (TCBlockHeader **) to
 * (void **).  This is really more correct since we're only pointing to the
 * first entry of a free list.  Ooops, that's i_allocateStruct, not
 * i_allocateMemoryBlock.
 *
 * Revision 1.1  1995/05/09  02:46:51  ddavies
 * Initial revision
 *
 * Revision 1.4  1995/01/11  22:22:41  dferrell
 * i_allrun.c -> /product/ipcore/ipshared/src/RCS/i_allrun.c,v
 *
 * Changed include of shrcmprs.pub to shrcc.pub.
 *
 * Revision 1.3  1994/08/16  00:46:12  ddavies
 * ddavies on Mon Aug 15 17:44:11 PDT 1994
 *
 * i_allrun.c -> /product/ipcore/ipshared/src/RCS/i_allrun.c,v
 *
 * Fixed bug in i_findAllRunsInLine.  When the image was a multiple of 32
 * pixels in width and a large (wider than 32) block abuted the right edge
 * of the image, the routine wrote the EOL value on top of the rightmost
 * run instead of in the next element of the run list.
 *
 * Revision 1.2  1994/06/08  03:53:13  ddavies
 * ddavies on Tue Jun 7 20:51:57 PDT 1994
 *
 * i_allrun.c -> /product/ipcore/ipshared/src/RCS/i_allrun.c,v
 *
 * Some formatting changes; change comments for i_findAllRunsInLine to say
 * that the pointers and dimensions are relative to the outside of the frame
 * like other public routines.  This actually required no code changes.
 *
 * Revision 1.1  1994/06/06  19:34:35  ddavies
 * Initial revision
 *

 */

