/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_allrn2.c: routines to find all the runs in a scan line, putting
 *		the results in a list.
 *
 *		Int32	i_findAllRunsInLineArray()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "imageref.pub"
#include "defines.pub"
#define	_USE_FIRST_BIT_TABLE_		/* want both tables from iopbyte.prv */
#include "iopbyte.prv"
#include "lmask.prv"			/* contains leftMask[] */
#include "rmask.prv"			/* contains rightMask[] */


/* prototypes */
#include "shrcc.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_allrn2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* It is more efficient to find all the runs of 1's in a scan line at
 * once rather than find them one at a time, as is done with
 * i_nextRunToTheRight.  i_findAllRunsInLine returns a linked list of
 * all the runs in a single scan line.  The routine in this file,
 * i_findAllRunsInLineArray returns the run coordinates in an array
 * instead of in a linked list.
 *
 * This routine is a lot like i_nextRunToTheRight, except that it loops
 * around after finding the right end of a run.
 *
 * One of the obvious efficiencies of an entire run at once is
 * that one ought to be able to take advantage of the fact that if a run
 * begins and ends in a single word, and it begins at bit X,
 * it will end somewhere to the right of bit X.  That is, we don't need to
 * do a completely general search of the word to find the end of the run.
 * One of the efficiencies of finding all the runs at once is that we
 * can do the same if we know that a new run starts in the same word
 * where an old run ended.  But how do we implement a fancy searcher?
 *
 * We could build a nice state machine with state numbers and a switch
 * statement at the top.  This is lovely, but sort of slow for something
 * we'll be doing a lot.  Just to be daring, I'm going to use, gasp,
 * goto's.
 *
 * The main loop will contain four sections:
 *   1.	Skip over words containing all 0's, if any.
 *   2.	Find the left end of a run.
 *   3. Skip over words containing all 1's, if any.
 *   4. Find the right end of a run and allocate a new run structure
 *	on the linked list if necessary.
 *
 * Clearly, if you know at the end of step 2 that the run ends in the
 * same work where it started (as is often the case), you should be able
 * to skip step 3 entirely.  Furthermore, if you know that the run began
 * in byte X and it finishes in the same word, you need only look in
 * byte X and bytes to the right.  Therefore, if this case occurs and we
 * found the beginning of the run in byte 0, we go to the stage of step 4
 * that starts looking for the end in byte 0.  Similarly, if the run
 * starts in byte 1 and ends in the same word, we go to the stage in step
 * 4 that looks in byte 1 and so on.
 *
 * Similarly, if we find the end of a run in step 4 and know that a new
 * run starts in the same word, we go to the portion of step 1 that starts
 * looking for the run in the appropriate byte of the word.
 * Using goto's to do this is unclean, but seems significantly faster than
 * any alternative I've been able to find.
 *
 * The normal path through steps 2 and 4 does a binary search on the bytes
 * of the word in question, then uses a table to find the bit within the
 * selected byte.  The bits to the left of the point found are filled in
 * in step 2 and cleared in step 4.  This allows us to find the end of the
 * run or beginning of the next run without being confused by the previous
 * run.
 */

/* PUBLIC */
/**************************************************************************
 *   i_findAllRunsInLineArray()
 *
 *	This procedure starts from the given offset in a scan line and
 *	looks to the right for runs of 1's.  The coordinates of each
 *	run found are returned in an array of Run structures.
 *
 *	THIS PROCEDURE ONLY WORKS ON BINARY IMAGES.  Since we expect
 *	to be calling it zillions of times, it doesn't check the image
 *	depth.
 *
 *	If there is a 1 at the leftLimit, this routine will return
 *	the leftLimit as the first 1.
 *
 *	pArray must point to a non-null array of Run32's.  If this
 *	routine runs off the end of the array, it returns with
 *	ia_bufferExceeded.
 *
 *	The address of the array element just beyond the last valid
 *	array element is returned in *ppDummy.  Clients can mark this
 *	element as they wish.
 *
 *	Input:
 *		pS:		Pointer to leftmost 32-bit word in the
 *				scan line OUTSIDE the frame.
 *		rightLimit:	Rightmost pixel at which we should look
 *				for bits.  By definition, no run can
 *				begin or end to the right of this offset.
 *				Runs ending beyond this offset will be
 *				chopped off here.  This is relative to the
 *				OUTSIDE of the frame.
 *		leftLimit:	Starting offset in line relative to 
 *				left edge of image OUTSIDE frame.
 *		pArray:		Pointer to array of Run structures.
 *		numRuns:	Number of Run32 structures in the array
 *		xOffset:	Add this number of each run coordinate
 *				found.
 *		ppDummy:	Pointer to pointer to the dummy array element
 *				that's just to the right of the last valid
 *				array element.  This routine fills in the
 *				address of this dummy element.  Clients can
 *				mark this dummy as they wish.
 *
 *	Returns:
 *		ia_successful		found a run
 *		ia_invalidParm		pS or pArray was NULL or numRun == 0
 *		ia_bufferExceeded	ran out of Run32s in array
 *
 **************************************************************************/
Int32  CDECL 
i_findAllRunsInLineArray (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	rightLimit,
    Int32	leftLimit,
    Run32      *pArray,
    Int32	numRuns,
    Int32	xOffset,
    Run32     **ppDummy)
{
    PROLOGUE;
    Int32	 byteOffset;		/* byte offset into the source line */
    Int32	 rightByteLimit;	/* byte offset of the word holding the
					 * rightmost pixel */
    UInt32	 inputWord;		/* the word we read from the image */
    Int32	 theEnd;		/* holds fragment of word containing the
					 * end of a run; or it contains
					 * the numerical offset of this
					 * run from the end of the word
					 */
    Run32	*pArrayEnd;		/* ptr just past end of array, used to
					 * check for overflow. */


	/* See if we were given valid parameters */
    if ((pArray == NULL) || (pS == NULL) || (numRuns == 0))
	return ia_invalidParm;

	/* Calculate address of the end of the list.  Used for
	 * bounds checking. */
    pArrayEnd = &(pArray[numRuns]);

	/* calculate the byte offset of the word holding the initial offset */
    byteOffset = (leftLimit/8) & ~0x3;
	/* turn rightLimit into the offset of the word holding the last byte */
    rightByteLimit = (rightLimit/8) & ~0x3;
    if (byteOffset > rightByteLimit)	/* we're done already */
	goto finishFindAllRunsInLineArray;

    inputWord = ULONG_ACCESS(&pS[byteOffset]);	/* get the first word */

	/* mask off the bits on the left that we should ignore */
    inputWord &= rightMask[(~leftLimit) & 0x1F];
    if (inputWord != 0)
	goto findLeftEnd;	/* a run starts in the very first word */

	/* MAIN LOOP */
	/* In this loop, we skip words full of 0's, find the left end
	 * of a run, skip words full of 1's and find the right end of
	 * a run.  We also check for the end of the array.
	 */
    while (byteOffset <= rightByteLimit)
    {
	    /* STEP 1: SKIP OVER WORDS OF 0's */

	    /* We only get here if inputWord == 0, so go look at the
	     * next word.
	     * BUG FIX: Well, yes, inputWord might equal 0, but
	     * that doesn't mean that we're not at the right end
	     * now.  So we can't always advance byteOffset blindly.
	     * On the other hand, we can't stop just because
	     * the memory location at byteOffset holds 0, because
	     * it might not.  Remember that the current value of inputWord
	     * has been modified, so may not reflect what's in memory.
	     * So, we only advance if we're not already at the end of the line.
	     */
	if (byteOffset < rightByteLimit)
	{
	    byteOffset += 4;
	    while ( (byteOffset < rightByteLimit) &&
		    (ULONG_ACCESS(&pS[byteOffset]) == 0) )
		byteOffset += 4;
    
		/* At this point, we're either pointing at a non-zero word or
		 * at the last word of the scan line (or both).  Put this
		 * word into inputWord.  If it's 0, just stop.
		 */
	    inputWord = ULONG_ACCESS(&pS[byteOffset]);
	}
	if (inputWord == 0)
	    goto finishFindAllRunsInLineArray;


	    /* STEP 2: FIND LEFT END OF RUN */

	    /* Find the leftmost 1.  Do this by locating the 1 down to
	     * the byte, then using tables to find it in a byte.
	     * Note that we also set the 0's to the left of the leftmost
	     * 1 to 1.  This prevents us from finding them when we
	     * look for the first 0.
	     */
findLeftEnd:
	if ((inputWord & 0xFFFF0000) != 0)
	{
	    theEnd = inputWord >> 16;	/* get the upper half.  We use
					 * theEnd so we don't mess up
					 * inputWord.  We'll need it
					 * below to find the right
					 * end of the run
					 */
	    if ((theEnd & 0xFF00) != 0)
	    {
		    /* Left End of Run is in Byte 0 */
    
		theEnd = firstBitTable[theEnd >> 8];	/* get offset within
							 * word */
		inputWord |= leftMask[theEnd];
		theEnd += (byteOffset * 8);
		    /* if that was off the right end of the area we're
		     * allowed to search, just quit.
		     */
		if (theEnd > rightLimit)
		    goto finishFindAllRunsInLineArray;
		else
		{
		    if (pArray < pArrayEnd)
			pArray->runLeft = theEnd + xOffset;
		    else
		    {
			__error__ = ia_bufferExceeded;
			goto finishFindAllRunsInLineArrayWithError;
		    }
		}

		    /* fill in white to the left of the run so we can find the
		     * right end of the run.  If the run ends in this word,
		     * skip the section of code that skips over entire words
		     * filled with 1's.
		     */
		if (inputWord != 0xFFFFFFFF)
		{
		    inputWord = ~inputWord;
		    goto findRightEnd;
		}
	    }
	    else
findLeftEndByte1:	/* theEnd contains inputWord >> 16 */
	    {
		    /* Left End of Run in Byte 1 */
		theEnd = 8 + firstBitTable[theEnd & 0xFF];
		inputWord |= leftMask[theEnd];
		theEnd += (byteOffset * 8);
		    /* if that was off the right end of the area we're
		     * allowed to search, just quit.
		     */
		if (theEnd > rightLimit)
		    goto finishFindAllRunsInLineArray;
		else
		{
		    if (pArray < pArrayEnd)
			pArray->runLeft = theEnd + xOffset;
		    else
		    {
			__error__ = ia_bufferExceeded;
			goto finishFindAllRunsInLineArrayWithError;
		    }
		}

		    /* if inputWord is now not all 1's, find the other end
		     * of the run by starting to look in byte 1.
		     */
		if (inputWord != 0xFFFFFFFF)
		{
		    inputWord = ~inputWord;
		    if ((inputWord & 0xFFFF0000) != 0)	/* end is in byte 1 */
		    {
			theEnd = inputWord >> 16;
			goto findRightEndByte1;
		    }
		    else
			goto findRightEndBytes23;
		}
	    }
	}
	else	/* The left end of the run starts in the lower
		 * half of the word */
findLeftEndBytes23:
	{
	    if ((inputWord & 0xFF00) != 0)
	    {
		    /* Left End of Run in Byte 2 */
    
		theEnd = 16 + firstBitTable[inputWord >> 8];
		inputWord |= leftMask[theEnd];
		theEnd += (byteOffset * 8);
		    /* if that was off the right end of the area we're
		     * allowed to search, just quit.
		     */
		if (theEnd > rightLimit)
		    goto finishFindAllRunsInLineArray;
		else
		{
		    if (pArray < pArrayEnd)
			pArray->runLeft = theEnd + xOffset;
		    else
		    {
			__error__ = ia_bufferExceeded;
			goto finishFindAllRunsInLineArrayWithError;
		    }
		}

		    /* if inputWord is now not all 1's, find the other end
		     * of the run by starting to look in byte 2.
		     */
		if (inputWord != 0xFFFFFFFF)
		{
		    inputWord = ~inputWord;
		    goto findRightEndBytes23;
		}
	    }
	    else
findLeftEndByte3:
	    {
		    /* Left End of Run in Byte 3 */

		theEnd = 24 + firstBitTable[inputWord & 0xFF];
		inputWord |= leftMask[theEnd];
		theEnd += (byteOffset * 8);
		    /* if that was off the right end of the area we're
		     * allowed to search, just quit.
		     */
		if (theEnd > rightLimit)
		    goto finishFindAllRunsInLineArray;
		else
		{
		    if (pArray < pArrayEnd)
			pArray->runLeft = theEnd + xOffset;
		    else
		    {
			__error__ = ia_bufferExceeded;
			goto finishFindAllRunsInLineArrayWithError;
		    }
		}

		    /* if inputWord is now not all 1's, find the other end
		     * of the run by starting to look in byte 3.
		     */
		if (inputWord != 0xFFFFFFFF)
		{
		    inputWord = ~inputWord;
		    goto findRightEndByte3;
		}
	    }
	}

	    /* STEP 3: SKIP OVER WORDS OF 1's */

	    /* If we get here, inputWord contains all 1's.  Keep looking
	     * at input words until we either run out of room or
	     * we find one with some 0's in it.  That marks the right
	     * end of the run.
	     * BUG FIX: The current word might contain all 1's AND
	     * be the last word of the line, so we can't blindly
	     * advance byteOffset.
	     * On the other hand, we can't stop just because
	     * the memory location at byteOffset doesn't hold all 1's, because
	     * it might not.  Remember that the current value of inputWord
	     * has been modified, so may not reflect what's in memory.
	     * On the third hand, we have to do something to detect the
	     * case when the last run really goes to the right boundary,
	     * In my first bug "fix", inputWord contained all 1's
	     * if we started at the right boundary and it contained
	     * all 0's if we didn't.  Hmmm.
	     * Ok, so we always advance byteOffset and just don't
	     * do the read if byteOffset >= rightByteLimit.  Then we
	     * check to see if inputWord == 0 OR byteOffset >= rightByteLimit.
	     * In either case, the run ends at the right boundary.
	     */
	byteOffset += 4;
	while ( (byteOffset < rightByteLimit) &&
		(ULONG_ACCESS(&pS[byteOffset]) == 0xFFFFFFFF) )
	    byteOffset += 4;
    
	    /* At this point, we're either pointing at a non-1's word or
	     * at the last word of the scan line (or both) OR we're
	     * pointing one word beyond the last word if we started
	     * in the last word.  If there's a word here, put it's
	     * inverse into inputWord.  If that inverse is all
	     * 0's or we're now beyond the rightLimit, return rightLimit
	     * as the right end of the run.
	     */
	if (byteOffset <= rightByteLimit)
	{
	    inputWord = ~(ULONG_ACCESS(&pS[byteOffset]));
	}
	else
	{
	    inputWord = 0;	/* our way of saying that the run ended
				 * at the rightLimit.
				 */
	}

	if (inputWord == 0)
	{
	    pArray->runRight = rightLimit + xOffset;
	    pArray++;
	    goto finishFindAllRunsInLineArray;
	}


	    /* STEP 4: FIND RIGHT END OF RUN */

	    /* At this point, inputWord has been inverted and does not
	     * contain all 0's.  Just start looking for the first 1's.
	     */
	    /* Find the rightmost 1 in the real word, or the (leftmost
	     * 1 in the inverted word) - 1.  Do this by locating the 1 down to
	     * the byte, then using tables to find it in a byte.
	     * When we're done finding the end of the run, we set the
	     * entire run to 0's so we can find the start of the next
	     * run..
	     */
findRightEnd:	/* inputWord has already been complemented */
	if ((inputWord & 0xFFFF0000) != 0)
	{
	    theEnd = inputWord >> 16;	/* get the upper half.  We use
					 * theEnd so we don't mess up
					 * inputWord.  We'll need it
					 * to find the next run.
					 */
	    if ((theEnd & 0xFF00) != 0)
	    {
		    /* Right End of Run is in Byte 0 */

		theEnd = firstBitTable[theEnd >> 8];	 /* get offset within
							  * word */
		inputWord = ~(inputWord | leftMask[theEnd]);
		theEnd += (byteOffset * 8) - 1;
		    /* if that was off the right end of the area we're
		     * allowed to search, set the limit to rightLimit
		     * and quit.
		     */
		if (theEnd > rightLimit)
		{
		    pArray->runRight = rightLimit + xOffset;
		    pArray++;
		    goto finishFindAllRunsInLineArray;
		}
		else
		{
		    pArray->runRight = theEnd + xOffset;
		    pArray++;
		}

		    /* if a new run starts in this word, go find it.
		     */
		if (inputWord != 0)
		    goto findLeftEnd;
	    }
	    else
findRightEndByte1:  /* inputWord has been complemented and theEnd contains
		     * (complemented inputWord) >> 16
		     */
	    {
		    /* Right End of Run is in Byte 1 */

		theEnd = 8 + firstBitTable[theEnd & 0xFF];
		inputWord = ~(inputWord | leftMask[theEnd]);
		theEnd += (byteOffset * 8) - 1;
		    /* if that was off the right end of the area we're
		     * allowed to search, set the limit to rightLimit
		     * and quit.
		     */
		if (theEnd > rightLimit)
		{
		    pArray->runRight = rightLimit + xOffset;
		    pArray++;
		    goto finishFindAllRunsInLineArray;
		}
		else
		{
		    pArray->runRight = theEnd + xOffset;
		    pArray++;
		}

		    /* if inputWord is now not all 0's, find the start
		     * of the next run by starting to look in byte 1.
		     */
		if (inputWord != 0)
		{
		    if ((inputWord & 0xFFFF0000) != 0)	/* start is in byte 1 */
		    {
			theEnd = inputWord >> 16;
			goto findLeftEndByte1;
		    }
		    else
			goto findLeftEndBytes23;
		}
	    }
	}
	else	/* The right end of the run ends in the lower
		 * half of the word.  Remember that inputWord
		 * has already been complemented so we're looking
		 * for the leftmost 1's
		 */
findRightEndBytes23:	/* inputWord has already been complemented */
	{
	    if ((inputWord & 0xFF00) != 0)
	    {
		    /* Right End of Run is in Byte 2 */
    
		theEnd = 16 + firstBitTable[inputWord >> 8];
		inputWord = ~(inputWord | leftMask[theEnd]);
		theEnd += (byteOffset * 8) - 1;
		    /* if that was off the right end of the area we're
		     * allowed to search, set the limit to rightLimit
		     * and quit.
		     */
		if (theEnd > rightLimit)
		{
		    pArray->runRight = rightLimit + xOffset;
		    pArray++;
		    goto finishFindAllRunsInLineArray;
		}
		else
		{
		    pArray->runRight = theEnd + xOffset;
		    pArray++;
		}

		    /* if inputWord is now not all 0's, find the start
		     * of the next run by starting to look in byte 2.
		     */
		if (inputWord != 0)
		    goto findLeftEndBytes23;
	    }
	    else
findRightEndByte3:	/* inputWord has already been complemented */
	    {
		    /* Right End of Run is in Byte 3 */

		theEnd = 24 + firstBitTable[inputWord & 0xFF];
		inputWord = ~(inputWord | leftMask[theEnd]);
		theEnd += (byteOffset * 8) - 1;
		    /* if that was off the right end of the area we're
		     * allowed to search, set the limit to rightLimit
		     * and quit.
		     */
		if (theEnd > rightLimit)
		{
		    pArray->runRight = rightLimit + xOffset;
		    pArray++;
		    goto finishFindAllRunsInLineArray;
		}
		else
		{
		    pArray->runRight = theEnd + xOffset;
		    pArray++;
		}

		    /* if inputWord is now not all 0's, find the start
		     * of the next run by starting to look in byte 3.
		     */
		if (inputWord != 0)
		    goto findLeftEndByte3;
	    }
	}
    }	/* if we get to the bottom of the loop, the next run starts in
	 * some other word.  Go look for it.
	 */


	/* If we're still in bounds, return happily.  Otherwise complain.
	 * Since we would have incremented the pArray pointer after
	 * inserting the last run, it might legally be equal to
	 * pArrayEnd at this point.
	 */
finishFindAllRunsInLineArray:
    if (pArray <= pArrayEnd)
    {
	*ppDummy = pArray;
	__error__ = ia_successful;
    }
    else
	__error__ = ia_bufferExceeded;

finishFindAllRunsInLineArrayWithError:
    NORMAL_RETURN;
}



