/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_allrn3.c: routines to find all the runs in a scan line.
 *
 *		Int32	i_findAllRunsInLineWithState()
 *					finds all runs in a scan line
 *					from a position in a bitmap
 *					described in an FindRunsState
 *					struct.
 *
 *		Int32	i_getFindRunsStateSize()
 *					returns size of a FindRunsState block
 *
 *		Int32	i_initFindRunsState()
 *					fills in a FindRunsState block
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"

/* prototypes */
#include "shrcc.pub"
#include "shrcc.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_allrn3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* The standard connected component routines find all connected components
 * in a region of a bitmap.  Two new routines, i_findBoundingBoxesForSomeCCs
 * and i_findRunsForSomeCCs have two new features each.  They allow their
 * heaps to be allocated beforehand to decrease overhead on each call.
 * They also can be used on images represented in some form other than
 * bitmaps.  The way this is done is by allowing their clients to pass
 * in procedures that produce lists of TCScanLineRuns from whatever
 * image format they have.
 *
 * The procedures in this file are examples of such procedures.  They
 * happen to operate on bitmaps.  They are useful to clients which want
 * the low overhead features of the new connected component routines, but
 * still want to operate on bitmaps.
 *
 * This state block will contain a pointer to a bitmap, which might move
 * around.  We recommend that clients reinitialize the contents of this
 * structure immediately before calling i_findBoundingBoxesForSomeCCs or
 * i_findRunsForSomeCCs.
 *
 * i_getFindRunsStateSize returns the size of the FindRunsState block.
 * The client is responsible for allocating a block of this size.
 *
 * i_initFindRunsState fills in the entries in a FindRunsState block.  This
 * routine should be called to reinitialize the block immediately before
 * every call to i_findBoundingBoxesForSomeCCs or i_findRunsForSomeCCs..
 *
 * i_findAllRunsInLineWithState extracts the relevant information from the
 * FindRunsState struct, hands it to i_findAllRunsInLine and updates the
 * image pointer.
 */


/* PUBLIC */
/**************************************************************************
 *   i_getFindRunsStateSize()
 *
 *	This procedure returns the size of a FindRunsState block
 *
 *	Input:
 *		pFindRunsStateSize:	Pointer to returned size	
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pFindRunsStateSize is NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getFindRunsStateSize (
    IPState    *pIPState,
    UInt32     *pFindRunsStateSize)
{
    PROLOGUE;
    if (pFindRunsStateSize == NULL)
	return ia_invalidParm;

    *pFindRunsStateSize = sizeof(FindRunsState);
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_initFindRunsState()
 *
 *	This procedure initializes a FindRunsState block.
 *
 *	Input:
 *		pFindRunsState	Pointer to the FindRunsState block to be
 *				initialized.
 *		pImage:		Pointer to leftmost 32-bit word in the
 *				scan line OUTSIDE the frame.
 *		leftLimit:	left edge of region of interest in line
 *				relative to left edge of image's OUTSIDE frame.
 *		rightLimit:	rightmost pixel at which we should look
 *				for bits.  By definition, no run can
 *				begin or end to the right of this offset.
 *				Runs ending beyond this offset will be
 *				chopped off here.  This is relative to the
 *				OUTSIDE of the frame.
 *		bpl:		Image's bytes/line.
 *
 *	Action:
 *		Puts the args into the slots in the structure.
 *
 *	Returns:
 *		ia_successful		did it
 *		ia_invalidParm		pFindRunsState is NULL
 *
 **************************************************************************/
Int32  CDECL 
i_initFindRunsState (
    IPState    *pIPState,
    FindRunsState *pFindRunsState,
    UInt8      *pImage,
    Int32	leftLimit,
    Int32	rightLimit,
    Int32	bpl)
{
	/* is the pointer ok? */
    PROLOGUE;
    if (pFindRunsState == NULL)
	return ia_invalidParm;

	/* we're happy, initialize the entries as the client asked */
    pFindRunsState->pImage = pImage;
    pFindRunsState->leftLimit = leftLimit;
    pFindRunsState->rightLimit = rightLimit;
    pFindRunsState->bpl = bpl;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_findAllRunsInLineWithState()
 *
 *	This procedure starts from the given offset in a scan line and
 *	looks to the right for runs of 1's.  The coordinates of each
 *	run found are returned in a linked list of TCScanLineRun
 *	structures.
 *
 *	THIS PROCEDURE ONLY WORKS ON BINARY IMAGES.  Since we expect
 *	to be calling it zillions of times, it doesn't check the image
 *	depth.
 *
 *	If there is a 1 at the initialOffset, this routine will return
 *	the initialOffset as the first 1.
 *
 *	pList must point to a non-null list of TCScanLineRun's.  This
 *	routine reuses structures in the list as needed.  If it runs
 *	out of structures, it allocated more from the heap of
 *	TCScanLineRun's pointed to by pRunHeap.
 *
 *	Unlike i_findAllRunsInLine, this procedure takes the image pointer
 *	and horizontal region boundaries in a structure.  It is meant to
 *	be used with i_findBoundingBoxesForSomeCCs and i_findRunsForSomeCCs.
 *	The state block is declared to be a pointer to void so that we
 *	match the generic prototype for procedures of this type.
 *
 *	Input:
 *		pStateBlock:	Pointer to state block containing image info.
 *		pList:		pointer to list of TCScanLineRuns that will
 *				hold the output.  Coordinates of these runs
 *				will be relative to the OUTSIDE of the frame.
 *		pRunHeap:	pointer to heap of TCScanLineRuns.  This
 *				heap is created using i_allocateMemoryBlock
 *				and freed using i_freeAllMemoryBlocks.
 *
 *	Returns:
 *		ia_successful		found a run
 *		ia_invalidParm		pStateBlock, pList or pRunHeap was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_findAllRunsInLineWithState (
    IPState    *pIPState,
    void       *pStateBlock,
    TCScanLineRun *pList,
    TCBlockHeader *pRunHeap)
{
PROLOGUE;
 FindRunsState	*pFindRunsState;	/* get the correct type */


	/* if any of the pointers are NULL, quit */
    if ((pStateBlock == NULL) || (pList == NULL) || (pRunHeap == NULL))
	return ia_invalidParm;

	/* fix the type of the state block */
    pFindRunsState = (FindRunsState *) pStateBlock;

	/* Have i_findAllRunsInLine do the real work */
    __error__ = i_findAllRunsInLine (pIPState, pFindRunsState->pImage,
				   pFindRunsState->rightLimit,
				   (UInt32)(pFindRunsState->leftLimit),
				   pList,
				   pRunHeap);

    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* it worked.  Point to the next scan line and return */
    pFindRunsState->pImage += pFindRunsState->bpl;

    NORMAL_RETURN;
}



/*

 $Log: i_allrn3.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:10  jlin
 Imported sources

 Revision 1.3  1998/01/13 23:47:06  ddavies
 i_allrn3.c -> /product/ipcore/ipshared/src/RCS/i_allrn3.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1995/12/13 02:42:12  ejaquith
 i_allrn3.c -> /product/ipcore/ipshared/src/RCS/i_allrn3.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1995/07/14  02:39:11  ddavies
 * Initial revision
 *

 */
