/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_bndbx1.c: Find the bounding boxes of connected components in a
 *		rectangle of an image.
 *
 *		Int32	i_findConnectedComponentBoundingBoxes()
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

/* prototyes */
#include "shrcc.pub"
#include "shrcc.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_bndbx1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* It is often useful to find blobs in an image.  What does "find" mean?
 * For the the purposes of this code, "find" means find the minimum
 * rectangle which completely encloses the blob.
 * i_findConnectedComponentBoundingBoxes works only on binary images.  
 *
 * The previous version of this function did its job by individually erasing
 * each pixel of each blob and remembering where it used to be.  That's
 * pretty slow.  This version of the function will make a single sweep through
 * the image looking for blobs.  The location of each blob will be recorded
 * as we go.  We plan on using this technique for symbol-based compression,
 * so this is a good place to try it out.
 *
 * The interface to this function is interesting.  It needs the obvious
 * things: a pointer to the image data as well as the bytes/line.
 * In addition, it takes x, y, w, h specifying a rectangle within the image
 * where we will be looking.  Blobs that cross the boundary are clipped
 * at the edges.
 *
 * The function also takes a minimum width and height.  Bounding boxes that
 * are less than either the minimum width OR the minimum height are ignored.
 *
 * The interesting feature of the interface is the way we return the
 * results.  Our clients are very particular about memory allocation and
 * data structures.  At Leigh Klotz's suggestion, we will call a client-
 * supplied callback procedure for each bounding box found instead of
 * creating a list of rectangles.  That allows the client to
 * define any sort of data structure.
 *
 * Here's a brief description of the procedure.  After stripping the frame
 * from the arguments, we allocate memory blocks for our two data structures.
 * The structures are allocated from our own heap (see i_allocateStruct and
 * i_freeStruct in i_heap.c), so we need memory for the heap.
 *
 * Our two data structures are a CCSymbol and a TCScanLineRun.  To
 * save typing, I call connected components "symbols."
 *
 * We go through the image one horizontal line at a time.  Each horizontal
 * line of an image is composed of a sequence of black and white runs.
 * There is one TCScanLineRun for each black run on the current horizontal
 * line, or scan line.  These structures are used to connect runs on
 * one scan line with overlapping runs on the next.  Doing this, we can
 * construct entire connected components by connecting vertical sequences of
 * horizontal runs.
 *
 * Since connected components may fork and join as we move vertically
 * through them, a single symbol may have several active runs in a single
 * scan line.  We use a CCSymbol structure as the central symbol
 * identification element shared by all the TCScanLineRun structs in a
 * given symbol.  The central element also keeps the dimensions of the symbol,
 * which is what we really care about anyway.
 *
 * Once our little heaps are allocated, we start looking for horizontal
 * runs of black bits (1's) in the input image.  We keep two sets of runs,
 * one for the line 0 and one for the line 1.  Line 0 is above line 1.
 * Keeping all the runs for both lines makes it easy to compare the two.
 * The two sets of data structures are used as a two line ring buffer -
 * that is, each set of data structures is filled with new run data (becomes
 * line 1), then used as the line 0, then filled with new data again and so on.
 *
 * The first thing that happens on a new line is that we fill one of the
 * sets of data structures with the run data from the new line.  This is
 * done without any further processing of that run data by i_findAllRunsInLine.
 * We hope this allows i_findAllRunsInLine to be very efficient.
 *
 * Next, we make a pass through line 0 looking for new symbols
 * and joins between existing symbols.  When a run with no existing
 * symbol is found, we allocate a new symbol for it.  When adjoining runs
 * on the previous line overlap a single run on the current
 * line, we have found a join.  When this happens, we consolidate the
 * symbol's bounding boxes and make the symbol we're going to discard point
 * to the new symbol.  As a consequence of this process, a run may point
 * to a list of symbols.  All but the last symbol in the list are obsolete
 * symbols that were merged with other symbols.  The obsolete symbols
 * will be discarded as the number of active runs associated with them
 * goes to 0.  This should happen by the time we finish processing line 0.
 *
 * Having straightened out all the symbols, we make another pass through
 * the runs.  For each run encountered on line 0, we take the following
 * actions:
 *   1. Find the leftmost run on line 1 which overlaps the run on line 0,
 *	if any.
 *   2. If there is an overlapping run on line 1:
 *		* Update the symbol's maxY and minX.
 *		* Go through the obsolete symbols associated with the
 *		  run on line 0, stopping before reaching the real symbol.
 *			* Decrement each symbol's numActiveRuns.
 *			* if the count goes to 0, free the obsolete symbol.
 *		* If the run on line 1 has already been given a symbol,
 *		  a join must have occured between runs on line 0, so
 *		  decrement the real symbol's numActiveRuns.
 *		* Look for runs on line 1 to the right of the original
 *		  run on line 1.  For each of these that overlaps the run
 *		  on Line 0, tag the run on line 1 with the real symbol
 *		  belonging to the run on line 0 and increment the 
 *		  symbol's numActiveRuns.
 *		* When we have found the rightmost overlapping run on line
 *		  1, update the symbol's maxX using this run's right endpoint.
 *   3. If there were no overlapping runs on line 1 (the symbol might end):
 *		* Go through the obsolete symbols associated with the
 *		  run on line 0, stopping before reaching the real symbol.
 *			* Decrement each symbol's numActiveRuns.
 *			* if the count goes to 0, free the symbol.
 *		* Decrement the real symbol's numActiveRuns.
 *		* If the real symbol's numActiveRuns goes to 0, call
 *		  the client's callback proc and free the symbol.
 *
 * When the last line of the rectangle has been scanned, we make one more
 * pass through the remaining scan line structs to finish off any existing
 * blobs.  We simply pretend that a line of white was found below them.
 *
 * FEATURE ADDITION: The XIS folks would like to know the number of pixels
 * in the connected components being returned.  This is pretty easy, we
 * just add the length of each run to a new field in the symbol structure
 * and include the sum as one of the args to the callback proc.
 *
 * FEATURE ADDITION: Doug Ferrell and Julie Craig would also like to see
 * the number of runs and a point that is actually on the cc.  The former
 * is used to quickly sort out interesting cc's.  The latter is used to
 * mark interesting ones for later seed filling.
 */


/* PUBLIC */
/**************************************************************************
 *   i_findConnectedComponentBoundingBoxes()
 *
 *	This procedure returns the bounding boxes of all connected
 *	components within a rectangle of a binary image.  The bounding
 *	boxes must be at least minWidth pixels wide AND at least
 *	minHeight pixels tall to be returned.
 *
 *	This procedure finds 8- or 4-connected objects.  In the 8-connected
 *	world, diagonal connections count.  In the 4-connected world,
 *	only horizontal and vertical connections count.
 *
 *	The bounding boxes are returned by calling a client-supplied
 *	callback proc.  This allows the data structure used by the client
 *	to hold the bounding boxes to be hidden from this routine.
 *
 *	Input:
 *		pImage:		Pointer to upper left corner of image
 *				INCLUDING frame.
 *		bpl:		The image's bytes/line.
 *		clipX:		X offset of the upper left corner of the
 *				  clipping rectangle INCLUDING the frame.
 *		clipY:		Y offset of the upper left corner of the
 *				  clipping rectangle INLCUDING the frame.
 *		clipW:		Width of clipping rectangle in pixels
 *		clipH:		Height of the clipping rectangle in pixels
 *		minWidth,	Connected components whose bounding boxes
 *		minHeight:	have widths less than minWidth
 *				OR heights less than minHeight are discarded,
 *				i.e. the client's callback proc is not
 *				called for these cc's.
 *		connectType:	0 => 4-connected, 1 => 8-connected.  Other
 *				values produce undefined results.
 *		pRectData:	A pointer to client supplied data that
 *				  we include as an argument when calling
 *				  the client's callback proc.
 *		rectReturnProc:	A pointer to the client's callback proc.
 *				  This routine is called with rectInfo and
 *				  the x, y, w, h of the bounding box.  "m"
 *				  is the mass, or number of active pixels,
 *				  inside the bounding box. px,py is a point
 *				  on the connected component to be used for
 *				  seed fills.  nRuns is the number of runs
 *				  in the connected component.
 *		pRectReturnCode: A pointer to the code returned by the client's
 *				   rectReturnProc.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		connectType > 1
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *		ia_callbackError	The client's callback proc
 *					returned an error.
 *
 **************************************************************************/
Int32  CDECL 
i_findConnectedComponentBoundingBoxes (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	bpl,
    Int32	clipX,
    Int32	clipY,
    Int32	clipW,
    Int32	clipH,
    Int32	minWidth,
    Int32	minHeight,
    const Int32	connectType,
    void       *pRectData,
    Int32 (CDECL *rectReturnProc) (
		    IPState    *pIPState,
		    void       *,
		    Int32	x,
		    Int32	y,
		    Int32	w,
		    Int32	h,
		    Int32	m,
		    Int32	nRuns,
		    Int32	px,
		    Int32	py),
    Int32      *pRectReturnCode)
{
    PROLOGUE;
    TCBlockHeader	*pSymbolHeap;		/* heap for CCSymbol structs */
    TCBlockHeader	*pRunHeap;		/* heap for TCScanLineRuns */
    TCScanLineRun	*pLine0;		/* ptr to older scan line */
    TCScanLineRun	*pLine1;		/* ptr to newer scan line */
    TCScanLineRun	*pLine0Left, *pLine0Right; /* ptrs to left and
						    * right ends of a joined
						    * section on line 0 */
    TCScanLineRun	*pLine1Run;		/* ptr to a run in line 1 */
    Int32		 currentLine;		/* current line number */
    CCSymbol	*pCurrentSym, *pNewSym;	/* ptrs to symbols */


	/* see if connectType is illegal */
    if (connectType > 1)
	return ia_invalidParm;

	/* initialize heap pointers so we can deallocate them even if
	 * they don't exist yet. */
    pSymbolHeap = pRunHeap = NULL;

	/* allocate memory for data structures */
    __error__ = i_allocateMemoryBlock (pIPState, &pSymbolHeap, sizeof(CCSymbol),
				     cCCSymbolBlockSize,
				     cFalse,
				     NULL, NULL,
				     &i_compressAlloc,
				     &i_compressFree);
    if (__error__ != ia_successful)
	goto findCCBBReturn;

    __error__ = i_allocateMemoryBlock (pIPState, &pRunHeap, sizeof(TCScanLineRun),
				     cTCScanLineRunBlockSize,
					cFalse,
					NULL, NULL,
					&i_compressAlloc,
					&i_compressFree);
    if (__error__ != ia_successful)
	goto findCCBBReturn;

	/* Allocate two ending runs for each line */
    __error__ = i_allocateStruct (pIPState, pRunHeap, (void **) &pLine0);
    if (__error__ != ia_successful)
	goto findCCBBReturn;

    pLine0->runLeft = -cVeryLargeNum;
    pLine0->runRight = -cVeryLargeNum;
    pLine0->pClient = NULL;
    pLine0->pNext = NULL;

    __error__ = i_allocateStruct (pIPState, pRunHeap, (void **) &(pLine0->pNext));
    if (__error__ != ia_successful)
	goto findCCBBReturn;

    pLine0->pNext->runLeft = cVeryLargeNum;
    pLine0->pNext->runRight = cVeryLargeNum;
    pLine0->pNext->pClient = NULL;
    pLine0->pNext->pNext = NULL;


    __error__ = i_allocateStruct (pIPState, pRunHeap, (void **) &pLine1);
    if (__error__ != ia_successful)
	goto findCCBBReturn;

    pLine1->runLeft = -cVeryLargeNum;
    pLine1->runRight = -cVeryLargeNum;
    pLine1->pClient = NULL;
    pLine1->pNext = NULL;

    __error__ = i_allocateStruct (pIPState, pRunHeap, (void **) &(pLine1->pNext));
    if (__error__ != ia_successful)
	goto findCCBBReturn;

    pLine1->pNext->runLeft = cVeryLargeNum;
    pLine1->pNext->runRight = cVeryLargeNum;
    pLine1->pNext->pClient = NULL;
    pLine1->pNext->pNext = NULL;



	/* MAIN LOOP */
	/* In this loop, we get new set of runs; assign symbols to all of
	 * the runs in the upper line; then assign symbols to all the runs
	 * in the lower line.
	 * We do one too many lines do the last line is actually retired.
	 * In the last line, we just say there are no runs.
	 */
    for (currentLine = clipY, pImage += clipY * bpl;
	 currentLine <= clipY + clipH;
	 currentLine++, pImage += bpl)
    {
	    /* exchange the structures for the current and previous lines */
	pLine0Left = pLine1;
	pLine1 = pLine0;
	pLine0 = pLine0Left;

	if (currentLine < clipY + clipH)
	{
		/* go get a new set of runs for line 1 */
	    __error__ =
		i_findAllRunsInLine (pIPState, pImage,
				     clipX + clipW - 1, /* right bdry */
				     clipX,		/* left bdry */
				     pLine1->pNext,	/* ptr to first struct
							 * to be written into */
				     pRunHeap);		/* alloc new structs
							 * from here. */
	    if (__error__ != ia_successful)
		goto findCCBBReturn;
	}
	else
	{
	    pLine1->pNext->runLeft = cVeryLargeNum;	/* make a blank line */
	    pLine1->pNext->runRight = cVeryLargeNum;
	}

	    /* ----- FIRST PASS -------- */

	    /* loop through the runs in line 0 looking for runs without
	     * symbols or runs that are joined on line 1 */
	for (pLine0Left = pLine0, pLine0Right = pLine0->pNext,
		pLine1Run = pLine1, pNewSym = NULL;
	     pLine0Right->runLeft < cVeryLargeNum;
	     pLine0Left = pLine0Right, pLine0Right = pLine0Right->pNext)
	{
	    pCurrentSym = pNewSym;
	    pNewSym = (CCSymbol *)pLine0Right->pClient;
	    while ((pNewSym != NULL) && (pNewSym->pNewSymbol != NULL))
		pNewSym = pNewSym->pNewSymbol;		/* find the real symbol
							 * associated with
							 * this run */

		/* try to find a run that joins pLine0Left and pLine0Right */
	    while (pLine1Run->runRight < pLine0Right->runLeft - connectType)
		pLine1Run = pLine1Run->pNext;

	    if ((pLine1Run->runLeft <= pLine0Left->runRight + connectType) &&
		(pLine1Run->runRight >= pLine0Right->runLeft - connectType))
	    {
		    /* pLine0Left should be joined to pLine0Right.  Are
		     * they already part of the same symbol?  Note that
		     * we can only have a join between two real runs
		     * (not the terminating runs on the ends) since
		     * the leftmost terminating run as a big negative
		     * coordinate and all real runs have positive coords.
		     */
		if (pCurrentSym != pNewSym)	/* need to do a join */
		{
		    if (pNewSym == NULL)
			/* cool, just add this new thing to the existing sym */
		    {
			pNewSym = pCurrentSym;
			pLine0Right->pClient = pCurrentSym;
			pCurrentSym->numActiveRuns++;
			pCurrentSym->numRuns++;
			pCurrentSym->mass += pLine0Right->runRight -
					     pLine0Right->runLeft + 1;
			if (pCurrentSym->maxX < pLine0Right->runRight)
			    pCurrentSym->maxX = pLine0Right->runRight;
		    }
		    else
		    {				/* need to join two real syms */
			pNewSym->pNewSymbol = pCurrentSym;
			if (pCurrentSym->minX > pNewSym->minX)
			    pCurrentSym->minX = pNewSym->minX;
			if (pCurrentSym->maxX < pNewSym->maxX)
			    pCurrentSym->maxX = pNewSym->maxX;
			if (pCurrentSym->minY > pNewSym->minY)
			    pCurrentSym->minY = pNewSym->minY;
			if (pCurrentSym->maxY < pNewSym->maxY)
			    pCurrentSym->maxY = pNewSym->maxY;
			pCurrentSym->numActiveRuns += pNewSym->numActiveRuns;
			pCurrentSym->numRuns += pNewSym->numRuns;
			pCurrentSym->mass += pNewSym->mass;
			pNewSym = pCurrentSym;
		    }
		}
	    }

		/* If this run has no symbol, give it one */
	    if (pNewSym == NULL)
	    {
		__error__ = i_allocateStruct (pIPState, pSymbolHeap,
					    (void **) &pNewSym);
		if (__error__ != ia_successful)
		    goto findCCBBReturn;

		pNewSym->minX = pLine0Right->runLeft;
		pNewSym->maxX = pLine0Right->runRight;
		pNewSym->minY = currentLine - 1;
		pNewSym->maxY = 0;			/* not used here */
		pNewSym->numActiveRuns = 1;
		pNewSym->numRuns = 1;
		pNewSym->px = pLine0Right->runLeft;
		pNewSym->py = currentLine - 1;
		pNewSym->mass = pLine0Right->runRight -
				pLine0Right->runLeft + 1;
		pNewSym->pNewSymbol = NULL;
		pLine0Right->pClient = pNewSym;
	    }
	}	/* End of first pass through pLine0.  When we exit this
		 * loop, each run has a symbol and all joins have happened.
		 */

	    /* ----- SECOND PASS -------- */

	    /* Now we go through the runs again, determining how to change
	     * the number of active runs for each symbol and copying symbol
	     * pointers to runs on line 1 that overlap runs on line 0.
	     */
	for (pLine0Left = pLine0->pNext, pLine1Run = pLine1->pNext;
	     pLine0Left->runLeft < cVeryLargeNum;
	     pLine0Left = pLine0Left->pNext)
	{
		/* First, try to find an overlapping run on line 1 */
	    while (pLine1Run->runRight < pLine0Left->runLeft - connectType)
		pLine1Run = pLine1Run->pNext;

	    if (pLine1Run->runLeft <= pLine0Left->runRight + connectType)
	    {
		    /* In this case, we know there is a run under this
		     * one, so the real symbol to which this run belongs
		     * cannot end here.
		     * However, any superceded symbol may end since we
		     * don't count the run under this one as being part
		     * of a superceded symbol.
		     * So, we go through all the symbols associated with
		     * pLine1Run; decrementing numActiveRuns for each one.
		     * If the count goes to 0 AND the symbol has been
		     * superceded, we free it.
		     */
		for (pCurrentSym = (CCSymbol *)pLine0Left->pClient;
		     pCurrentSym->pNewSymbol != NULL;
		     )
		{
		    pCurrentSym->numActiveRuns--;
		    if (pCurrentSym->numActiveRuns <= 0)
		    {
			pNewSym = pCurrentSym;
			pCurrentSym = pCurrentSym->pNewSymbol;
			i_freeStruct (pIPState, pSymbolHeap,
				      (void **) &pNewSym);
		    }
		    else
			pCurrentSym = pCurrentSym->pNewSymbol;

		}

		    /* At this point, pCurrentSym points to the real
		     * symbol for this run.  If the run below this
		     * one has not been tagged yet, we don't change
		     * this symbol's number of active runs (one above
		     * becomes one below) but do tag the run below
		     * with this symbol and increment the total number
		     * of runs associated with this symbol.  Also,
		     * update the symbol's bounding box using this
		     * run's coords.
		     * If it has already been tagged, decrement the
		     * number of active runs for this symbol to
		     * account for the join.
		     */
		if (pLine1Run->pClient == NULL)
		{
		    pLine1Run->pClient = pCurrentSym;
		    if (pCurrentSym->minX > pLine1Run->runLeft)
			pCurrentSym->minX = pLine1Run->runLeft;
		    pCurrentSym->mass += pLine1Run->runRight -
					 pLine1Run->runLeft + 1;
		    pCurrentSym->numRuns++;
		}
		else
		{
		    pCurrentSym->numActiveRuns--;
		}

		    /* Now tag any more runs on line 1 that overlap
		     * this run in line 0.
		     */
		while (pLine1Run->pNext->runLeft <=
			pLine0Left->runRight + connectType)
		{
		    pLine1Run = pLine1Run->pNext;
		    pLine1Run->pClient = pCurrentSym;
		    pCurrentSym->numActiveRuns++;
		    pCurrentSym->numRuns++;
		    pCurrentSym->mass += pLine1Run->runRight -
					 pLine1Run->runLeft + 1;
		}

		    /* now that we're at the right end of the runs
		     * that overlap this run, fix up the symbol's maxX.
		     */
		if (pCurrentSym->maxX < pLine1Run->runRight)
		    pCurrentSym->maxX = pLine1Run->runRight;
	    }
	    else	/* there is no overlapping run below this one */
	    {
		    /* This case starts out like the one above it in
		     * that we free all the superceded symbols whose
		     * numActiveRun counts go to 0.  However, when we
		     * get to the real symbol, we also decrement it's
		     * numActiveRuns.  If it goes to 0, the symbol is
		     * finished, so we call the client's callback and
		     * free the symbol.
		     */
		for (pCurrentSym = (CCSymbol *) pLine0Left->pClient;
		     pCurrentSym->pNewSymbol != NULL;
		     )
		{
		    pCurrentSym->numActiveRuns--;
		    if (pCurrentSym->numActiveRuns <= 0)
		    {
			pNewSym = pCurrentSym;
			pCurrentSym = pCurrentSym->pNewSymbol;
			i_freeStruct (pIPState, pSymbolHeap,
				      (void **) &pNewSym);
		    }
		    else
			pCurrentSym = pCurrentSym->pNewSymbol;
		}

		    /* see if the real symbol is done */
		pCurrentSym->numActiveRuns--;
		    /* if it went to 0, set the lower y coordinate and
		     * return the bounding box */
		if (pCurrentSym->numActiveRuns <= 0)
		{
			/* only tell the client about boxes that have
			 * adequate dimensions.  Note that since we did not
			 * subtract the frame above, we don't need to
			 * add it back in now.
			 */
		    if ((pCurrentSym->maxX - pCurrentSym->minX + 1 >= minWidth)
		        &&
			(currentLine - pCurrentSym->minY >= minHeight))
		    {
			__error__ =
			    rectReturnProc(
			        pIPState,
				pRectData,
				pCurrentSym->minX,
				pCurrentSym->minY,
				pCurrentSym->maxX - pCurrentSym->minX + 1,
				currentLine - pCurrentSym->minY,
				pCurrentSym->mass,
				pCurrentSym->numRuns,
				pCurrentSym->px, pCurrentSym->py);
			if (__error__ != ia_successful)
			    goto findCCBBBadCallback;
		    }

		    i_freeStruct (pIPState, pSymbolHeap,
				  (void **) &pCurrentSym);
		}
	    }	/* end of case where there is not run on line1 overlapping
		 * current run on line 0
		 */
	}	/* end of second pass through runs on line 0 */
    }	/* end of loop through the entire image */

	/* Things must have worked out ok */
    __error__ = ia_successful;
    goto findCCBBReturn;

	/* Something bad happened.  Free the heaps and return */
findCCBBBadCallback:
    *pRectReturnCode = __error__;
    __error__ = ia_callbackError;

findCCBBReturn:
    if (pSymbolHeap != NULL)
	i_freeAllMemoryBlocks (pIPState, &pSymbolHeap);
    if (pRunHeap != NULL)
	i_freeAllMemoryBlocks (pIPState, &pRunHeap);

    NORMAL_RETURN;
}



/*

 $Log: i_bndbx1.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:10  jlin
 Imported sources

 Revision 1.5  1998/01/13 23:47:08  ddavies
 i_bndbx1.c -> /product/ipcore/ipshared/src/RCS/i_bndbx1.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1997/07/11 01:48:26  ddavies
 i_bndbx1.c -> /product/ipcore/ipshared/src/RCS/i_bndbx1.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.3  1997/04/24 23:26:13  ddavies
 ddavies on Thu Apr 24 16:25:37 PDT 1997

 i_bndbx1.c -> /product/ipcore/ipshared/src/RCS/i_bndbx1.c,v

 Change all TCSymbols to CCSymbols.

 Revision 1.2  1995/12/13 02:42:17  ejaquith
 i_bndbx1.c -> /product/ipcore/ipshared/src/RCS/i_bndbx1.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1995/07/10  23:00:16  ddavies
 * Initial revision
 *
 * Revision 1.12  1995/01/11  22:23:35  dferrell
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Removed include of shrcmprs.pub and added shrcc.pub and shrcc.prv.
 *
 * Revision 1.11  1994/11/07  18:53:52  ddavies
 * ddavies on Mon Nov 7 10:53:13 PST 1994
 *
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Clarify comments concerning minWidth and minHeight in
 * i_findConnectedComponentBoundingBoxes.
 *
 * Revision 1.10  1994/10/21  03:27:07  ddavies
 * ddavies on Thu Oct 20 20:26:04 PDT 1994
 *
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Add code to i_findConnectedComponentBoundingBoxes so that it returns
 * the number of runs in the cc and a point on the cc to the callback
 * procedure.  This is easy to calculate and will be useful for both quickly
 * sorting through cc's and marking the cc's selected.
 *
 * Revision 1.9  1994/06/08  03:55:35  ddavies
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Change code that takes account of the frame since i_findAllRunsInLine
 * wants pointers and such relative to the outside of the frame anyway.  This
 * means clipX and clipY are no longer adjusted; we move pImage down by
 * clipY and we don't add the frame back in to the bounding box coords
 * when calling the client's callback routine.
 *
 * Revision 1.8  1994/06/06  19:37:39  ddavies
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Completely reworked version of i_findConnectedComponentBoundingBoxes.
 * The new version uses i_findAllRunsInLine instead of finding the runs
 * one at a time.  The code for parsing the runs is also quite a bit
 * simpler.  It's somewhat faster - the "somewhat" depending on the
 * environment and the platform.  Running a test program on a simple
 * image, it's maybe 1.1x faster.  Running in the middle of the
 * regression test, it's about 1.2x faster.  Anyway, this new routine
 * will share most of it's code with the new i_findConnectedComponentRuns.
 * That routine will return a list of all of the runs in each connected
 * component in addition to the bounding box.
 *
 * Revision 1.7  1994/04/26  02:13:26  ddavies
 * ddavies on Mon Apr 25 19:12:25 PDT 1994
 *
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Change lastLeft and lastRight members of the TCScanLineStruct to
 * runLeft and runRight for consistency.
 *
 * Revision 1.6  1994/04/02  01:24:05  jcraig
 * jcraig on Fri Apr 1 17:22:30 PST 1994
 *
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Change connectType from a UInt32 to an Int32.  This was necessary because
 * in the glyph demo test, a run started at offset 0.  We test to see if
 * the run below it is connected by seeing if it is to the left of 0 -
 * connectType, or -1.  Unfortunately, this turned into a very large
 * positive number before the comparison.  I hate C.
 *
 * Revision 1.5  1994/03/31  03:01:43  ddavies
 * ddavies on Wed Mar 30 19:00:23 PST 1994
 *
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Add the connectType parameter to i_findConnectedComponentBoundingBoxes.
 * When set to 0, we use 4-connected rules to define connected components.
 * When set to 1, we use 8-connected rules.  I don't think this routine will
 * slow down noticably, but let me know if it does.
 *
 * Revision 1.4  1994/03/24  20:34:07  ddavies
 * ddavies on Thu Mar 24 12:32:45 PST 1994
 *
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Change "Figure" to "Symbol" in a terminology change.  Also update use of
 * functions that allocate and deallocate structures so they use an allocator
 * and a free routine that are passed in instead of using malloc and free.
 *
 * Revision 1.3  1994/01/19  18:51:59  dferrell
 * dferrell on Wed Jan 19 10:49:44 PST 1994
 *
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Changed declaration syntax of i_find ConnectedComponentBoundingBoxes ()
 * so that MPW C could understand the function pointer parameter.
 * Made same change to ip_retireBBRun ().
 *
 * Revision 1.2  1994/01/07  04:15:59  ddavies
 * ddavies on Thu Jan 6 20:15:18 PST 1994
 *
 * i_bndbox.c -> /product/ipcore/ipshared/src/RCS/i_bndbox.c,v
 *
 * Added the mass parameter to the callback proc used in
 * i_findConnectedComponentBoundingBoxes.  This parameter specifies the number
 * of pixels actually set in a connected component.  Also calculated the mass
 * of each connected component.
 *
 * Revision 1.1  1993/12/17  22:13:17  ddavies
 * Initial revision
 *

 */

