/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_bswap.c
 *
 *  Public routines:
 *      i_swapImageBytes():	routine to swap the bytes within every
 *				32-bit word of an image
 *	ip_copyLine():		C implementation of routine that copies
 *				one source image line to one dest image line.
 *	ip_invertLine():	C implementation of routine that inverts
 *				a source line on the way to the dest
 */


#include <stdio.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

    /* prototypes */
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shrrast.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_bswap.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* On little endian machines, images are often stored so the leftmost bytes
 * is in byte 0 of a 32 bit word, instead of being in the leftmost byte.
 * On these occasions, we need a routine that swaps all the bytes within
 * each 32-bit word.
 *
 * We organize the routine in two levels.  The top one deals with parameters
 * as usual.  The bottom one swaps the bytes in a line of image.  The bottom
 * one can be implemented in assembly language, especially on i486 machines
 * that have a byteswap opcode.  We use ip_getByteSwapProc to give us a pointer
 * to the procedure that works best on the current platform.  This must be
 * decided at runtime on a PC.
 */


/* This is a public procedure */
/* PUBLIC */
/*
 *  i_swapImageBytes()
 *			Input: description of an image.
 *			Output: all the bytes of each word in the
 *				image have been swapped - i.e. if the
 *				the original contents of a word were
 *
 *				A | B | C | D
 *
 *				The new contents are
 *
 *				D | C | B | A
 *
 *			Since this routine may be called on framed or
 *			frameless images, we assume that the pointer
 *			we are given points to the upper left corner
 *			of whatever area is to have its bytes swapped.
 *			We will always byteswap bytesPerLine/4 words
 *			in each line - so that will include the frame
 *			words.  Therefore, for frameless images, pass
 *			in a pointer to the upper left corner of the
 *			image.  For framed images, pass in a pointer to
 *			the upper left corner of the image, or for a
 *			slight increase in efficiency, pass in a pointer
 *			to the beginning of the first line containing image
 *			bytes (and decrease imageHeight by 2*cFrameBits).
 *
 *	pImage:			Pointer to upper left corner of frame in
 *				  image.
 *	bytesPerLine:		Number of bytes/line in image including
 *				  border, frame and pad.  This is the number
 *				  one adds to the dest image pointer to
 *				  move down 1 line.
 *	imageHeight:		Image height including the frame.
 *
 * This routine always returns a status code of ia_successful.
 */
Int32  CDECL 
i_swapImageBytes (
    IPState    *pIPState,
    UInt8      *pImage,
    UInt32	bytesPerLine,
    register Int32 imageHeight)
{
    PROLOGUE;
    GraySwapNearProc	swapLineProc;

	/* figure out which swap routine to use.  We will not be inverting
	 * the image bytes.  Both the source and destination image pointers
	 * are near pointers.
	 */
    ip_getByteSwapProc (pIPState, cFalse, cPixrToPixr,
			NULL, NULL, &swapLineProc);

	/* have ip_swapLineBytes do the real work.  Do not invert the bytes.
	 * The source image is also the dest image. */
    for (; imageHeight > 0; imageHeight--)
    {
        (*swapLineProc)(pIPState, (UInt32 *)pImage,
			(UInt32 *)pImage, bytesPerLine/4);
	pImage += bytesPerLine;
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/*
 *  ip_copyLine()
 *		This routine copies the words in an image line
 *		from source to dest.
 *
 *	pS:		Pointer to left end of source image line.
 *	pD:		Pointer to left end of dest image line.
 *	wordsPerLine:	Number of 32-bit words in the image line.
 *
 */
void  CDECL 
ip_copyLine (
    IPState    *pIPState,
    UInt32     *pS,
    UInt32     *pD,
    Int32	wordsPerLine)
{
    PROLOGUE;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy((void *) pD, (void *) pS, wordsPerLine*4);
}


/* PRIVATE */
/*
 *  ip_invertLine()
 *		This routine inverts the words in an image line while
 *		transferring them from source to dest.
 *
 *	pS:		Pointer to left end of source image line.
 *	pD:		Pointer to left end of dest image line.
 *	wordsPerLine:	Number of 32-bit words in the image line.
 *
 */
void  CDECL 
ip_invertLine (
    IPState    *pIPState,
    UInt32     *pS,
    UInt32     *pD,
    Int32	wordsPerLine)
{
	/* pretend the image is 1 bit/pixel and call i_rasterOp */
    PROLOGUE;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    i_rasterOp (pIPState,
		(UInt8 *) pD,		/* dpr */
	        0,			/* dx */
	        0, 			/* dy */
	        wordsPerLine*32,	/* dw */
	        1,			/* dh */
	        1,			/* dd */
	        wordsPerLine*4,		/* dLinebytes */
	        PIX_NOT(PIX_SRC),	/* op */
	        (UInt8 *) pS,		/* spr */
	        0,			/* sx */
	        0,			/* sy */
	        1,			/* sd */
	        wordsPerLine*4);		/* sLinebytes */
}


/*

$Log: i_bswap.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.7  1998/01/13 23:47:11  ddavies
i_bswap.c -> /product/ipcore/ipshared/src/RCS/i_bswap.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1995/12/13 02:42:25  ejaquith
i_bswap.c -> /product/ipcore/ipshared/src/RCS/i_bswap.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/10/18  03:52:07  ddavies
 * i_bswap.c -> /product/ipcore/ipshared/src/RCS/i_bswap.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.4  1995/09/13  23:27:48  ddavies
 * ddavies on Wed Sep 13 16:27:11 PDT 1995
 *
 * i_bswap.c -> /product/ipcore/ipshared/src/RCS/i_bswap.c,v
 *
 * Back out of adding invertImage parameter to i_swapImageBytes.  It swaps the
 * frame, so it also inverts the frame.  Avoiding this is a lot of work.  We'lllet the NT build test ip_swapLineBytesInvertedNear using the bmp/Pixr conversion
 * routines.
 *
 * Revision 1.3  1995/09/13  00:10:36  ddavies
 * ddavies on Tue Sep 12 17:10:22 PDT 1995
 *
 * i_bswap.c -> /product/ipcore/ipshared/src/RCS/i_bswap.c,v
 *
 * Add the invertImage parameter to i_swapImageLine.  This will be used to
 * test the new ip_swapLineBytesInvertedNear routine that inverts an image
 * line while byteswapping using near pointers for both source and dest.
 * The ip_ routine will be used primarily in NT and Win32 systems.
 *
 * Revision 1.2  1994/11/07  22:05:12  ddavies
 * ddavies on Mon Nov 7 14:04:30 PST 1994
 *
 * i_bswap.c -> /product/ipcore/ipshared/src/RCS/i_bswap.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.1  1994/01/22  03:47:21  ddavies
 * Initial revision
 *
 * Revision 1.7  1994/01/22  01:06:29  ddavies
 * ddavies on Fri Jan 21 17:04:55 PST 1994
 *
 * cbswap.c -> /product/ipcore/ipshared/src/RCS/cbswap.c,v
 *
 * change "include shrio." to "include shrpixr." since ipshared's libio is
 * being absorbed by libpixr.
 *
 * Revision 1.6  1993/12/02  19:20:17  ddavies
 * cbswap.c -> /product/ipcore/ipshared/src/RCS/cbswap.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.5  1993/11/05  01:23:13  ddavies
 * cbswap.c -> /product/ipcore/ipshared/src/RCS/cbswap.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/



