/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_ccrun1.c: Find the bounding boxes of connected components in a
 *		rectangle of an image.  Unlike
 *		i_findConnectedComponentBoundingBoxes, this routine
 *		also returns the runs of each connected component.
 *
 *		Int32	i_findConnectedComponentRuns()
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

/* prototyes */
#include "shrcc.pub"
#include "shrcc.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_ccrun1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* It is often useful to find blobs in an image.  What does "find" mean?
 * For the the purposes of this code, "find" means find the minimum
 * rectangle which completely encloses the blob.  The routine
 * i_findConnectedComponentBoundingBoxes does this.  Sometimes, it is
 * also useful to get the runs that compose each connected component.
 * XIS uses such a routine to extract characters from pages.  The routine
 * in this file, i_findConnectedComponentRuns returns a pointer to a
 * linked list of runs as well as the bounding box.
 *
 * i_findConnectedComponentRuns is heavily based on
 * i_findConnectedComponentBoundingBoxes.  In fact, most of the lines
 * of code are the same.  Here is a copy of the block comments for
 * i_findConnectedComponentBoundingBoxes:
 *
 *---------------------------------------------------------------------------  
 *
 * The previous version of this function did its job by individually erasing
 * each pixel of each blob and remembering where it used to be.  That's
 * pretty slow.  This version of the function will make a single sweep through
 * the image looking for blobs.  The location of each blob will be recorded
 * as we go.  We plan on using this technique for symbol-based compression,
 * so this is a good place to try it out.
 *
 * The interface to this function is interesting.  It needs the obvious
 * things: a pointer to the image data as well as the bytes/line.
 * In addition, it takes x, y, w, h specifying a rectangle within the image
 * where we will be looking.  Blobs that cross the boundary are clipped
 * at the edges.
 *
 * The function also takes a minimum width and height.  Bounding boxes that
 * are less than either the minimum width OR the minimum height are ignored.
 *
 * The interesting feature of the interface is the way we return the
 * results.  Our clients are very particular about memory allocation and
 * data structures.  At Leigh Klotz's suggestion, we will call a client-
 * supplied callback procedure for each bounding box found instead of
 * creating a list of rectangles.  That allows the client to
 * define any sort of data structure.
 *
 * Here's a brief description of the procedure.  After stripping the frame
 * from the arguments, we allocate memory blocks for our two data structures.
 * The structures are allocated from our own heap (see i_allocateStruct and
 * i_freeStruct in i_heap.c), so we need memory for the heap.
 *
 * Our two data structures are a CCSymbol and a TCScanLineRun.  To
 * save typing, I call connected components "symbols."
 *
 * We go through the image one horizontal line at a time.  Each horizontal
 * line of an image is composed of a sequence of black and white runs.
 * There is one TCScanLineRun for each black run on the current horizontal
 * line, or scan line.  These structures are used to connect runs on
 * one scan line with overlapping runs on the next.  Doing this, we can
 * construct entire connected components by connecting vertical sequences of
 * horizontal runs.
 *
 * Since connected components may fork and join as we move vertically
 * through them, a single symbol may have several active runs in a single
 * scan line.  We use a CCSymbol structure as the central symbol
 * identification element shared by all the TCScanLineRun structs in a
 * given symbol.  The central element also keeps any other centralized
 * symbol info, like the dimensions.
 *
 * Once our little heaps are allocated, we start looking for horizontal
 * runs of black bits (1's) in the input image.  We keep the runs for
 * consecutive pairs of scan lines.  We arbitrarily call the upper
 * scan line "line 0" and the lower one "line 1".  Keeping all the runs
 * for both lines makes it easy to compare the two.  The two sets of
 * data structures are used as a two line ring buffer - that is, each set
 * of data structures is filled with new run data (becomes line 1), then
 * used as the line 0, then filled with new data again and so on.
 *
 * The first thing that happens on a new line is that we fill one of the
 * sets of data structures (becomes line 1) with the run data from the
 * new line.  This is done without any further processing of that run
 * data by i_findAllRunsInLine.  We hope this allows i_findAllRunsInLine
 * to be very efficient.
 *
 * Next, we make a pass through line 0 looking for new symbols
 * and joins between existing symbols.  When a run with no existing
 * symbol is found, we allocate a new symbol for it.  When adjoining runs
 * on the previous line overlap a single run on the current
 * line, we have found a join.  When this happens, we consolidate the
 * symbol's bounding boxes and make the symbol we're going to discard point
 * to the new symbol.  As a consequence of this process, a run may point
 * to a list of symbols.  All but the last symbol in the list are obsolete
 * symbols that were merged with other symbols.  The obsolete symbols
 * will be discarded as the number of active runs associated with them
 * goes to 0.  This should happen by the time we finish processing line 0.
 *
 * Having straightened out all the symbols, we make another pass through
 * the runs.  For each run encountered on line 0, we take the following
 * actions:
 *   1. Find the leftmost run on line 1 which overlaps the run on line 0,
 *	if any.
 *   2. If there is an overlapping run on line 1:
 *		* Update the symbol's maxY and minX.
 *		* Go through the obsolete symbols associated with the
 *		  run on line 0, stopping before reaching the real symbol.
 *			* Decrement each symbol's numActiveRuns.
 *			* if the count goes to 0, free the obsolete symbol.
 *		* If the run on line 1 has already been given a symbol,
 *		  a join must have occured between runs on line 0, so
 *		  decrement the real symbol's numActiveRuns.
 *		* Look for runs on line 1 to the right of the original
 *		  run on line 1.  For each of these that overlaps the run
 *		  on Line 0, tag the run on line 1 with the real symbol
 *		  belonging to the run on line 0 and increment the 
 *		  symbol's numActiveRuns.
 *		* When we have found the rightmost overlapping run on line
 *		  1, update the symbol's maxX using this run's right endpoint.
 *   3. If there were no overlapping runs on line 1 (the symbol might end):
 *		* Go through the obsolete symbols associated with the
 *		  run on line 0, stopping before reaching the real symbol.
 *			* Decrement each symbol's numActiveRuns.
 *			* if the count goes to 0, free the symbol.
 *		* Decrement the real symbol's numActiveRuns.
 *		* If the real symbol's numActiveRuns goes to 0, call
 *		  the client's callback proc and free the symbol.
 *
 * When the last line of the rectangle has been scanned, we make one more
 * pass through the remaining scan line structs to finish off any existing
 * blobs.  We simply pretend that a line of white was found below them.
 *
 * FEATURE ADDITION: The XIS folks would like to know the number of pixels
 * in the connected components being returned.  This is pretty easy, we
 * just add the length of each run to a new field in the symbol structure
 * and include the sum as one of the args to the callback proc.
 *
 *---------------------------------------------------------------------------.  
 *
 * How does i_findConnectedComponentRuns differ?  First, one trivial
 * difference - we don't calculate the mass of the symbol.  We let the
 * client do that if necessary.
 *
 * The more interesting difference is that we change the way the TCScanLineRun
 * structures are used.  i_findConnectedComponentBoundingBoxes
 * keeps two lists of them; alternately filling them with runs from new
 * lines.  Rather than keep the lists around, i_findConnectedComponentRuns
 * generates and consumes the lists as it goes.  i_findAllRunsInLine still
 * builds a new list for each new line.  However, when
 * i_findConnectedComponentRuns is done with a run in line 0, it
 * moves the run to the chain of runs associated with the run's symbol.
 * At the end of the line, the last (dummy) buffer run is connected directly to
 * the first (dummy) run again, making a list that contains no real runs.
 *
 * There's a little more to it.  After returning from the client's callback
 * proc, we free all the TCScanLineRun's before freeing the CCSymbol.
 * That's easy.  The slightly more complicated issue has to do with
 * joins.
 *
 * When we join two distinct symbols, we have to do something with their
 * lists of runs.  We could just tack one list on the end of the other, but
 * that doesn`t work.  Why not?  Because the runs have no vertical info
 * in them.  As a consequence of being taken directly from line 0, the runs
 * for a symbol appear in top-to-bottom, left-to-right order.  Clients
 * parse them this way.  They know when they get to the end of a line
 * because:
 *	nextLeftXCoord <= prevRightXCoord + 1
 * If just tacked the lists together end-to-end, the left to right order
 * would be violated - it would look like a very tall, thin symbol instead
 * of two symbols side by side.
 *
 * So, we must sort the list.  Before starting, let's simplify things by
 * retaining the symbol whose starting point is closest to the top of
 * the image.  If both symbols have the same height above the join,
 * keep the one on the left.  Since we're moving list elements from
 * the merged symbol to the surviving symbol, we'll call the merged
 * symbol the source and the surviving symbol the destination.
 *   1. Starting at the beginning of the run list of the dest
 *	symbol, skip over all the runs belonging to lines that are
 *	above the top of the merged symbol.  Use the criteria given above
 *	to detect ends of lines.
 *   2. Keep pointers to both lists:
 *	pSrcRun point to the next source run to be merged.
 *	pDestRun points to the next dest run to be merged.
 *	pPrevRun points to the last merged run.
 *	nextSrcX = x coord of the next source run to be merged or
 *		cVeryLargeNum of we're at the end of a source scan line.
 *	nextDestX = x coord of the next dest run to be merged or
 *		cVeryLargeNum of we're at the end of a dest scan line.
 *	Start by setting both nextSrcX and nextDestX to cVeryLargeNum and
 *	pSrcRun and pDestRun to the next source and dest runs respectively.
 *   Loop while pSrcRun != NULL or pDestRun != NULL:
 *   3. If both nextSrcX and nextDestX equal cVeryLargeNum, we're at the
 *	end of a scan line for the symbol.  Set nextSrcX to pSrcRun's
 *	x coordinate and nextDestX to pDestRun's x coordinate.  We will use
 *	these to decide which run to merge first.
 *   4. Compare nextSrcX to nextDestX.  If nextSrcX is smaller, link
 *	pPrevRun to pSrcRun and move through source runs until we reach
 *	a run to the right of nextDestX or the end of the source line
 *	or the last source run.  Similarly, if nextDestX is smaller,
 *	link pPrevRun to pDestRun and move down the list of dest runs.
 *	If we reach the end of a line or the end of the list of runs,
 *	set the nextSrcX or nextDestX to cVeryLargeNum.  Otherwise, set
 *	them to pSrcRun->runLeft and pDestRun->runLeft respectively.  Note
 *	that pPrevRun is left pointing to the end of the list and
 *	pSrcRun or pDestRun points to the run just beyond the end.
 *   5.	Repeat steps 3 and 4 until both lists are exhausted.
 *
 ***************************************************************************
 *
 * That was lovely and eventually worked.  However, on images where there
 * was a large piece of noise generating lots of joins, we spent a lot of
 * time running down through the list of runs for the noisy trash finding
 * the place where sorting should happen.
 *
 * I thought about putting in an extra pointer in each TCScanLineRun, so
 * we would have one in each direction.  We would use the new pointer to
 * find our way from the tail of the list back up to the place where
 * merging should start.  This has the drawback of adding
 * space and time to every operation to fix a (relatively) rare problem.
 *
 * Then, I had the dumb, but effective, idea of simply reversing the
 * direction of all of the pointers in the list of runs attached to a
 * symbol.  We add new runs to the tail of the list as before, but the
 * runs point towards the head instead of towards the tail.  When we
 * want to merge the runs for two symbols, we start from the tail.  This
 * scheme as the tiny drawback that you have to read the runs from the
 * lower right towards the upper left, but who cares?
 *
 * Why keep the head pointer around?  We can only use it to get at the
 * topmost, leftmost, image run.  That run points to NULL.  Well, when
 * merging symbols, we would like to avoid checking for pDestRun equalling
 * NULL.  We can do this if we guarantee that we'll always run out of
 * source runs before running out of dest runs.  This will be true if
 * the dest symbol already has the topmost, leftmost run.  We find
 * that out by looking at both head pointers.
 * So, the new algorithm is:
 *
 * pDestRun points to the next dest run to be merged;
 * pSrcRun points to the next source run to be merged;
 * pPrevRun points to the last merged run;
 * nextSrcX is the x coord of pSrcRun or -cVeryLargeNum of we've reached
 *	the left end of a scan line;
 * nextDestX is the x coord of pDestRun or -cVeryLargeNum of we've reached
 *	the left end of a scan line;
 *
 * Start by setting both nextSrcX and nextDestX to -cVeryLargeNum.
 *
 * Loop while pSrcRun != NULL (note that this automatically eliminates
 * any time taken to mess around if the source symbol is 1 pixel tall).
 * 1. If both nextSrcX and nextDestX equal -cVeryLargeNum, set
 *	nextSrcX to pSrcRun's x coord and nextDestX to pDestRun's x coord.
 *
 * 2. If nextSrcX is greater than (to the right of) nextDestX, link
 *	pPrevRun to pSrcRun and follow the chain of source runs until
 *	we find a run to the left of pDestRun, or find the end of the
 *	source line, or find the end of the list of source lines.
 *	If we found the end of the source list, just link pPrevRun to
 *	pDestRun and quit now.
 *	Similarly, if pDestRun was to the right of pSrcRun, link pPrevRun
 *	to pDestRun and follow destination runs until we find one to the
 *	left of pSrcRun or the end of the dest line.
 *	If the end of the line was found, set nextSrcX or nextDestX to
 *	-cVeryLargeNum.
 * 3. Repeat steps 1 and 2 until pSrcRun is NULL.
 *
 * FEATURE ADDITION: Several folks wanted to know the mass and number of runs
 * in the callback procedure without counting them explicitly.  We now
 * return the number of runs in the callback.
 */

/* PUBLIC */
/**************************************************************************
 *   i_findConnectedComponentRuns()
 *
 *	This procedure returns the bounding boxes and runs of all connected
 *	components within a rectangle of a binary image.  The bounding
 *	boxes must be at least minWidth pixels wide AND at least
 *	minHeight pixels tall to be returned.
 *
 *	This procedure finds 8- or 4-connected objects.  In the 8-connected
 *	world, objects can be connected with only diagonal pixels.
 *	In the 4-connected world, only horizontal and vertical
 *	connections matter.
 *
 *	The bounding boxes and runs are returned by calling a client-supplied
 *	callback proc.  This allows the data structure used by the client
 *	to hold the bounding boxes to be hidden from this routine.  Clients
 *	may copy the runs to some other structure or analyze them to produce
 *	other data or both.
 *
 *	Input:
 *		pImage:		Pointer to upper left corner of image
 *				INCLUDING frame.
 *		bpl:		The image's bytes/line.
 *		clipX:		X offset of the upper left corner of the
 *				  clipping rectangle INCLUDING the frame.
 *		clipY:		Y offset of the upper left corner of the
 *				  clipping rectangle INCLUDING the frame.
 *		clipW:		Width of clipping rectangle in pixels
 *		clipH:		Height of the clipping rectangle in pixels
 *		minWidth:	The minimum width of a bounding box that
 *				  we are allowed to return.
 *		minHeight:	The minimum height of a bounding box that
 *				  we are allowed to return.
 *		connectType:	0 => 4-connected, 1 => 8-connected.  Other
 *				values produce undefined results.
 *		pRectData:	A pointer to client supplied data that
 *				  we include as an argument when calling
 *				  the client's callback proc.
 *		rectReturnProc:	A pointer to the client's callback proc.
 *				  This routine is called with rectInfo and
 *				  the x, y, w, h of the bounding box and
 *				  the mass and number of runs of the CC.
 *				  This callback also has a pRuns arg.
 *				  pRuns is a pointer to a linked list of
 *				  TCScanLineRuns.  NOTE!! The order of the
 *				  runs is from the bottom-to-top, right-to-left
 *				  within a line!  The elements of this
 *				  list are freed after the callback proc
 *				  returns.
 *		pRectReturnCode: A pointer to the code returned by the client's
 *				   rectReturnProc.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		connectType > 1
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *		ia_callbackError	The client's callback proc
 *					returned an error.
 *
 **************************************************************************/
Int32  CDECL 
i_findConnectedComponentRuns (
    IPState    *pIPState,
    UInt8      *pImage,
    UInt32	bpl,
    Int32	clipX,
    Int32	clipY,
    Int32	clipW,
    Int32	clipH,
    Int32	minWidth,
    Int32	minHeight,
    const Int32	connectType,
    void       *pRectData,
    Int32 (CDECL *rectReturnProc) (
		    IPState    *pIPState,
		    void       *,
		    Int32	x,
		    Int32	y,
		    Int32	w,
		    Int32	h,
		    Int32	m,
		    Int32	nRuns,
		    TCScanLineRun *pRuns),
    Int32      *pRectReturnCode)
{
    PROLOGUE;
    TCBlockHeader	*pSymbolHeap;	/* heap for CCSymbol structs */
    TCBlockHeader	*pRunHeap;	/* heap for TCScanLineRun structs */
    TCScanLineRun	*pLine0;	/* ptr to older scan line */
    TCScanLineRun	*pLine1;	/* ptr to newer scan line */
    TCScanLineRun	*pLine0Left, *pLine0Right; /* ptrs to left and right
						    * ends of a joined section
						    * on line 0 */
    TCScanLineRun	*pLine1Run;	/* ptr to a run in line 1 */
    Int32		 currentLine;	/* current line number */
    CCSymbol		*pCurrentSym, *pNewSym;	/* ptrs to symbols */
    Int32		 nextSrcX, nextDestX;	/* used when merging lists
						 * of runs */
    TCScanLineRun	*pSrcRun, *pDestRun;	/* used to point to lists
						 * of runs to be merged */
    TCScanLineRun	*pPrevRun;		/* points to run immediately
						 * before pSrcRun or pDestRun
						 * during searches */


	/* see if connectType is illegal */
    if (connectType > 1)
	return ia_invalidParm;

	/* initialize heap pointers so we can deallocate them even if
	 * they don't exist yet. */
    pSymbolHeap = pRunHeap = NULL;

	/* allocate memory for data structures */
    __error__ = i_allocateMemoryBlock (pIPState, &pSymbolHeap, sizeof(CCSymbol),
				     cCCSymbolBlockSize,
				     cFalse,
				     NULL, NULL,
				     &i_compressAlloc,
				     &i_compressFree);
    if (__error__ != ia_successful)
	goto findCCRReturn;

    __error__ = i_allocateMemoryBlock (pIPState, &pRunHeap, sizeof(TCScanLineRun),
				     cTCScanLineRunBlockSize,
				     cFalse,
				     NULL, NULL,
				     &i_compressAlloc,
				     &i_compressFree);
    if (__error__ != ia_successful)
	goto findCCRReturn;

	/* Allocate two ending runs for each line */
    __error__ = i_allocateStruct (pIPState, pRunHeap, (void **) &pLine0);
    if (__error__ != ia_successful)
	goto findCCRReturn;

    pLine0->runLeft = -cVeryLargeNum;
    pLine0->runRight = -cVeryLargeNum;
    pLine0->pClient = NULL;
    pLine0->pNext = NULL;

    __error__ = i_allocateStruct (pIPState, pRunHeap, (void **) &(pLine0->pNext));
    if (__error__ != ia_successful)
	goto findCCRReturn;

    pLine0->pNext->runLeft = cVeryLargeNum;
    pLine0->pNext->runRight = cVeryLargeNum;
    pLine0->pNext->pClient = NULL;
    pLine0->pNext->pNext = NULL;


    __error__ = i_allocateStruct (pIPState, pRunHeap, (void **) &pLine1);
    if (__error__ != ia_successful)
	goto findCCRReturn;

    pLine1->runLeft = -cVeryLargeNum;
    pLine1->runRight = -cVeryLargeNum;
    pLine1->pClient = NULL;
    pLine1->pNext = NULL;

    __error__ = i_allocateStruct (pIPState, pRunHeap, (void **) &(pLine1->pNext));
    if (__error__ != ia_successful)
	goto findCCRReturn;

    pLine1->pNext->runLeft = cVeryLargeNum;
    pLine1->pNext->runRight = cVeryLargeNum;
    pLine1->pNext->pClient = NULL;
    pLine1->pNext->pNext = NULL;



	/* MAIN LOOP */
	/* In this loop, we get new set of runs; process all the joins
	 * in the upper line; then assign symbols to all the runs
	 * in the lower line.
	 * We do one too many lines do the last line is actually retired.
	 * In the last line, we just say there are no runs.
	 */
    for (currentLine = clipY, pImage += clipY * bpl;
	 currentLine <= clipY + clipH;
	 currentLine++, pImage += bpl)
    {
	    /* exchange the structures for the current and previous lines */
	pLine0Left = pLine1;
	pLine1 = pLine0;
	pLine0 = pLine0Left;

	if (currentLine < clipY + clipH)
	{
		/* go get a new set of runs for line 1 */
	    __error__ =
		i_findAllRunsInLine (pIPState, pImage,
				     clipX + clipW - 1, /* right bdry */
				     clipX,		/* left bdry */
				     pLine1->pNext,	/* ptr to first struct
							 * to be written into */
				     pRunHeap);		/* alloc new structs
							 * from here. */
	    if (__error__ != ia_successful)
		goto findCCRReturn;
	}
	else
	{
	    pLine1->pNext->runLeft = cVeryLargeNum;	/* make a blank line */
	    pLine1->pNext->runRight = cVeryLargeNum;
	}

	    /* ----- FIRST PASS -------- */

	    /* loop through the runs in line 0 looking for runs without
	     * symbols or runs that are joined on line 1 */
	for (pLine0Left = pLine0, pLine0Right = pLine0->pNext,
		pLine1Run = pLine1, pNewSym = NULL;
	     pLine0Right->runLeft < cVeryLargeNum;
	     pLine0Left = pLine0Right, pLine0Right = pLine0Right->pNext)
	{
	    pCurrentSym = pNewSym;
	    pNewSym = (CCSymbol *)pLine0Right->pClient;
	    while ((pNewSym != NULL) && (pNewSym->pNewSymbol != NULL))
		pNewSym = pNewSym->pNewSymbol;		/* find the real symbol
							 * associated with
							 * this run */

		/* try to find a run that joins pLine0Left and pLine0Right */
	    while (pLine1Run->runRight < pLine0Right->runLeft - connectType)
		pLine1Run = pLine1Run->pNext;

	    if ((pLine1Run->runLeft <= pLine0Left->runRight + connectType) &&
		(pLine1Run->runRight >= pLine0Right->runLeft - connectType))
	    {
		    /* pLine0Left should be joined to pLine0Right.  Are
		     * they already part of the same symbol?  Note that
		     * we will only detect a join between two real runs
		     * (not the terminating runs on the ends) since
		     * the leftmost terminating run as a big negative
		     * coordinate and all real runs have positive coords.
		     */
		if (pCurrentSym != pNewSym)	/* need to do a join */
		{
		    if (pNewSym == NULL)
			/* cool, just add this new thing to the existing sym */
		    {
			pNewSym = pCurrentSym;
			pLine0Right->pClient = pCurrentSym;
			pCurrentSym->numActiveRuns++;
			pCurrentSym->numRuns++;
			pCurrentSym->mass += pLine0Right->runRight -
					     pLine0Right->runLeft + 1;
			if (pCurrentSym->maxX < pLine0Right->runRight)
			    pCurrentSym->maxX = pLine0Right->runRight;
		    }
		    else
		    {				/* need to join two real syms */
			    /* Keep the taller symbol, or the one of the
			     * left if they're the same height.  Why does
			     * the test below work?  If they're different
			     * heights, the first clause works.  If they`re
			     * the same height and that height is 1, we
			     * took care of that case above.  Therefore,
			     * the second clause only comes into play when
			     * they are the same height and the height is
			     * greater than 1, therefore, both symbols
			     * already have non-null pRunHead pointers.
			     */
			if ((pNewSym->minY < pCurrentSym->minY) ||
			    ((pNewSym->minY == pCurrentSym->minY) &&
			     (pNewSym->pRunHead != NULL) &&
			     (pNewSym->pRunHead->runLeft <
			      pCurrentSym->pRunHead->runLeft)))
			{	/* pNewSym will survive, so exchange
				   pNewSym with pCurrentSym.
				 */
			    CCSymbol	*pTempSym;
			    pTempSym = pNewSym;
			    pNewSym = pCurrentSym;
			    pCurrentSym = pTempSym;
			}

			pNewSym->pNewSymbol = pCurrentSym;
			if (pCurrentSym->minX > pNewSym->minX)
			    pCurrentSym->minX = pNewSym->minX;
			if (pCurrentSym->maxX < pNewSym->maxX)
			    pCurrentSym->maxX = pNewSym->maxX;
			if (pCurrentSym->minY > pNewSym->minY)
			    pCurrentSym->minY = pNewSym->minY;
			if (pCurrentSym->maxY < pNewSym->maxY)
			    pCurrentSym->maxY = pNewSym->maxY;
			pCurrentSym->numActiveRuns += pNewSym->numActiveRuns;
			pCurrentSym->numRuns += pNewSym->numRuns;
			pCurrentSym->mass += pNewSym->mass;

			    /* Now for the fun part.  We need to merge the
			     * lists of runs for the two symbols.
			     * Since the runs point from the lower
			     * right to the upper left, we start
			     * merging from the bottom and work up.
			     * Since we chose pCurrentSym to have the
			     * topmost, leftmost run, we're sure that
			     * we'll run of source runs before we
			     * run out of dest runs.
			     *
			     * NOTE: we don't put runs onto the list
			     * for a symbol until we're done with line 0.
			     * If a symbol, say pNewSym, started one
			     * line ago, no runs have been put on its
			     * list, so pNewSym->pRunTail == NULL.
			     * Therefore, we shouldn't touch it.
			     */
			pSrcRun = pNewSym->pRunTail;
			pDestRun = pCurrentSym->pRunTail;
			    /* We want pPrevRun to point somewhere harmless
			     * so that we can set pPrevRun->pNext without
			     * checking to see if pPrevRun is ok all the
			     * time.  We know that we'll never get to
			     * the first run in the destination symbol,
			     * so this is a harmless place.  After we're
			     * done with the merge, we set pCurrentSym->pRunTail
			     * to the value that was stored in
			     * pCurrentSym->runHead->pNext since it
			     * will have been set to point to the
			     * rightmost, bottomost run.
			     */
			pPrevRun = pCurrentSym->pRunHead;

			    /* This is the loop that does the main merging.
			     * Within a line, we will alternate some runs
			     * from the source with some runs from the
			     * dest.  Scan lines may start and end with
			     * any combination of source and dest runs.
			     * It is this consideration that leads us to
			     * a fairly general loop.
			     *
			     * pPrevRun points to the last run in the merged
			     * list so far.
			     * pSrcRun points to the next source run.
			     * pDestRun points to the next dest run.
			     * nextSrcX is the value we give to the
			     * left coord of the next source run.  This is
			     * set to -cVeryLargeNum at the left end of the
			     * source line.  Remember that we work from
			     * right to left in each scan line.
			     * Ditto for nextDestX.
			     * At the top of the loop, we see if BOTH
			     * nextSrcX and nextDestX are equal to
			     * -cVeryLargeNum.  If so, we're at the left end of
			     * a scan line, so set them to the left coords
			     * of pSrcRun and pDestRun respectively.
			     * Then decide which is larger, nextSrcX or
			     * nextDestX.  If the source is larger, the
			     * next source run goes into the list next.
			     * make pPrevRun->pNext point to it.  The
			     * similar thing happens if the dest is to the
			     * right instead of the source.  Then we follow
			     * whichever list of runs (source or dest) we've
			     * chosen until we find the end of the scan
			     * line or we've gone past the right endpoint of
			     * the set of runs we didn't follow.  We leave
			     * pPrevRun pointing at the last run before
			     * the one that went too far.  If we ran off
			     * the end of the scan line, we set nextSrcX
			     * (if we followed the source) or nextDestX
			     * (if we followed the dest) to -cVeryLargeNum.
			     * Otherwise, we set it to the left coord of
			     * the run that was too far to the right.
			     * We then return to the top of the loop.
			     * This will cause us to follow the other set
			     * of runs and so on.
			     * Note that we need not test for pDestRun
			     * being NULL because we're guaranteed to run
			     * out of source runs before we run out of
			     * dest runs.
			     */
			nextSrcX = -cVeryLargeNum;
			nextDestX = -cVeryLargeNum;
			while (pSrcRun != NULL)
			{
				/* Are we at the left end of a scan line?
				 * If so, we had better have a run for both
				 * source and dest (no need to check for NULL)
				 */
			    if ((nextSrcX == -cVeryLargeNum) &&
				(nextDestX == -cVeryLargeNum))
			    {
				nextSrcX = pSrcRun->runLeft;
				nextDestX = pDestRun->runLeft;
			    }

				/* Should we follow source runs or dest
				 * runs? */
			    if (nextSrcX > nextDestX)
			    {	/* follow source runs */
				pPrevRun->pNext = pSrcRun;

				if (nextDestX > -cVeryLargeNum)
				{	/* we're looking for the last src run,
					 * or the end of the src line, or
					 * a gap ending the bunch of
					 * contiguous source runs that fit
					 * before the next dest run.
					 */
				    while (cTrue)
				    {
					pPrevRun = pSrcRun;
					pSrcRun = pSrcRun->pNext;
					if ((pSrcRun == NULL) ||
					    (pSrcRun->runRight >=
					      pPrevRun->runLeft - 1))
					{
					    nextSrcX = -cVeryLargeNum;
					    break;
					}
					if (pSrcRun->runLeft < nextDestX)
					{
					    nextSrcX = pSrcRun->runLeft;
					    break;
					}
				    }
				}
				else
				{	/* we're looking for the last src run,
					 * or the end of the src line */
				    do
				    {
					pPrevRun = pSrcRun;
					pSrcRun = pSrcRun->pNext;
				    } while ((pSrcRun != NULL) &&
					     (pSrcRun->runRight <
					      pPrevRun->runLeft - 1));
				    nextSrcX = -cVeryLargeNum;
				}
			    }
			    else	/* follow dest runs */
			    {
				pPrevRun->pNext = pDestRun;

				if (nextSrcX > -cVeryLargeNum)
				{	/* we're looking for the end of the
					 * dest line, or
					 * a gap ending the bunch of
					 * contiguous source runs that fit
					 * before the next source run.
					 */
				    while (cTrue)
				    {
					pPrevRun = pDestRun;
					pDestRun = pDestRun->pNext;
					if (pDestRun->runRight >=
					      pPrevRun->runLeft - 1)
					{
					    nextDestX = -cVeryLargeNum;
					    break;
					}
					if (pDestRun->runLeft < nextSrcX)
					{
					    nextDestX = pDestRun->runLeft;
					    break;
					}
				    }
				}
				else
				{	/* we're looking for the end of
					 * the dest line */
				    do
				    {
					pPrevRun = pDestRun;
					pDestRun = pDestRun->pNext;
				    } while (pDestRun->runRight <
					      pPrevRun->runLeft - 1);
				    nextDestX = -cVeryLargeNum;
				}
			    }
			}	/* end of loop that merges the rest of the
			     * runs in the two symbols
			     */

			    /* We exit this loop with pPrevRun pointing
			     * to the last run from the source.  Make it
			     * point to the next run from the dest.
			     */
			pPrevRun->pNext = pDestRun;

			pNewSym = pCurrentSym;	/* merge these so that the
						 * "pCurrentSym = pNewSym"
						 * statement at the top of
						 * the loop works.
						 */
			    /* set the symbol's end pointer to point to
			     * the new end of the list of runs.  This was
			     * set the to point to the very first run in
			     * the list because we initialized pPrevRun
			     * to be pCurrentRun->pRunHead.
			     * However, if there weren't really any runs
			     * attached to pNewSym, we have done nothing
			     * to any of the pointers, so don't try to
			     * undo it (pCurrentSym->pRunHead still
			     * equals NULL!).
			     */
			if (pCurrentSym->pRunHead->pNext != NULL)
			{
			    pCurrentSym->pRunTail =
						pCurrentSym->pRunHead->pNext;
			    pCurrentSym->pRunHead->pNext = NULL;
			}

		    }	/* End of case where two existing symbols are
			 * joined.
			 */
		}	/* End of case where joined runs do not already
			 * belong to the same symbol (one might below to
			 * to no symbol at all yet).
			 */
	    }	/* end of case where there was a join */

		/* If this run has no symbol, give it one */
	    if (pNewSym == NULL)
	    {
		__error__ = i_allocateStruct (pIPState, pSymbolHeap,
					    (void **) &pNewSym);
		if (__error__ != ia_successful)
		    goto findCCRReturn;

		pNewSym->minX = pLine0Right->runLeft;
		pNewSym->maxX = pLine0Right->runRight;
		pNewSym->minY = currentLine - 1;
		pNewSym->maxY = 0;			/* not used here */
		pNewSym->pRunHead = pNewSym->pRunTail = NULL;
		pNewSym->numActiveRuns = 1;
		pNewSym->numRuns = 1;
		pNewSym->mass = pLine0Right->runRight -
				pLine0Right->runLeft + 1;
		pNewSym->pNewSymbol = NULL;
		pLine0Right->pClient = pNewSym;
	    }
	}	/* End of first pass through pLine0.  When we exit this
		 * loop, each run has a symbol and all joins have happened.
		 */

	    /* ----- SECOND PASS -------- */

	    /* Now we go through the runs again, determining how to change
	     * the number of active runs for each symbol and copying symbol
	     * pointers to runs on line 1 that overlap runs on line 0.
	     */
	for (pLine0Left = pLine0->pNext, pLine1Run = pLine1->pNext;
	     pLine0Left->runLeft < cVeryLargeNum;
	     pLine0Left = pLine0Right)
	{
		/* First, we go through all the obsolete symbols associated
		 * with pLine0Left; decrementing numActiveRuns for each one.
		 * If the count goes to 0, we free it.
		 */
	    for (pCurrentSym = (CCSymbol *)pLine0Left->pClient;
		 pCurrentSym->pNewSymbol != NULL;
		 )
	    {
		pCurrentSym->numActiveRuns--;
		if (pCurrentSym->numActiveRuns <= 0)
		{
		    pNewSym = pCurrentSym;
		    pCurrentSym = pCurrentSym->pNewSymbol;
		    i_freeStruct (pIPState, pSymbolHeap, (void **) &pNewSym);
		}
		else
		    pCurrentSym = pCurrentSym->pNewSymbol;

	    }

		/* At this point, pCurrentSym points to the real symbol for
		 * pLine0Left.  Move this run to the list maintained for its
		 * symbol.  We need to do this before calling the
		 * client's callback proc, so we might as well do it now.
		 */
	    pLine0Right = pLine0Left->pNext;
	    pLine0Left->pNext = pCurrentSym->pRunTail;
	    if (pLine0Left->pNext == NULL)
		pCurrentSym->pRunHead = pLine0Left;
	    pCurrentSym->pRunTail = pLine0Left;

		/* Next, try to find an overlapping run on line 1 */
	    while (pLine1Run->runRight < pLine0Left->runLeft - connectType)
		pLine1Run = pLine1Run->pNext;

	    if (pLine1Run->runLeft <= pLine0Left->runRight + connectType)
	    {
		    /* In this case, we know there is a run under this
		     * one, so the real symbol to which this run belongs
		     * cannot end here.
		     * At this point, pCurrentSym points to the real
		     * symbol for this run. We know it cannot end since
		     * there is a run under it.  If that run has not yet
		     * been tagged, we don't change
		     * this symbol's number of active runs (one above
		     * becomes one below) but do tag the run below
		     * with this symbol.  Also, update the symbol's
		     * bounding box using this run's coords.
		     * If it has already been tagged, decrement the
		     * number of active runs for this symbol to
		     * account for the join.
		     */
		if (pLine1Run->pClient == NULL)
		{
		    pLine1Run->pClient = pCurrentSym;
		    if (pCurrentSym->minX > pLine1Run->runLeft)
			pCurrentSym->minX = pLine1Run->runLeft;
		    pCurrentSym->mass += pLine1Run->runRight -
					 pLine1Run->runLeft + 1;
		    pCurrentSym->numRuns++;
		}
		else
		{
		    pCurrentSym->numActiveRuns--;
		}

		    /* Now tag any more runs on line 1 that overlap
		     * this run in line 0.
		     */
		while (pLine1Run->pNext->runLeft <=
			pLine0Left->runRight + connectType)
		{
		    pLine1Run = pLine1Run->pNext;
		    pLine1Run->pClient = pCurrentSym;
		    pCurrentSym->numActiveRuns++;
		    pCurrentSym->numRuns++;
		    pCurrentSym->mass += pLine1Run->runRight -
					 pLine1Run->runLeft + 1;
		}

		    /* now that we're at the right end of the runs
		     * that overlap this run, fix up the symbol's maxX.
		     */
		if (pCurrentSym->maxX < pLine1Run->runRight)
		    pCurrentSym->maxX = pLine1Run->runRight;

	    }
	    else	/* there is no overlapping run below this one */
	    {
		    /* see if the real symbol is done */
		pCurrentSym->numActiveRuns--;
		    /* if it went to 0, set the lower y coordinate and
		     * return the bounding box */
		if (pCurrentSym->numActiveRuns <= 0)
		{
			/* only tell the client about boxes that have
			 * adequate dimensions.  Note that since we did not
			 * subtract the frame above, we don't need to
			 * add it back in now.
			 */
		    if ((pCurrentSym->maxX - pCurrentSym->minX + 1 >= minWidth)
		        &&
			(currentLine - pCurrentSym->minY >= minHeight))
		    {
			__error__ =
			    rectReturnProc(
			        pIPState,
				pRectData,
				pCurrentSym->minX,
				pCurrentSym->minY,
				pCurrentSym->maxX - pCurrentSym->minX + 1,
				currentLine - pCurrentSym->minY,
				pCurrentSym->mass,
				pCurrentSym->numRuns,
				pCurrentSym->pRunTail);
			if (__error__ != ia_successful)
			    goto findCCRBadCallback;
		    }

			/* free all the runs associated with this symbol */
		    for (pSrcRun = pCurrentSym->pRunTail;
			 pSrcRun != NULL;
			 pSrcRun = pDestRun)
		    {
			pDestRun = pSrcRun->pNext;
			i_freeStruct (pIPState, pRunHeap, (void **) &pSrcRun);
		    }
			/* and free the symbol itself */
		    i_freeStruct (pIPState, pSymbolHeap,
				  (void **) &pCurrentSym);
		}
	    }	/* end of case where there is not run on line1 overlapping
		 * current run on line 0
		 */
	}	/* end of second pass through runs on line 0 */

	    /* At this point, pLine0Left is pointing to the last buffer
	     * run in line 0.  Attach it to the first buffer run
	     * to make line 0 legally empty.
	     */
	pLine0->pNext = pLine0Left;

    }	/* end of loop through the entire image */

	/* Things must have worked out ok */
    __error__ = ia_successful;
    goto findCCRReturn;

	/* Something bad happened.  Free the heaps and return */
findCCRBadCallback:
    *pRectReturnCode = __error__;
    __error__ = ia_callbackError;

findCCRReturn:
    if (pSymbolHeap != NULL)
	i_freeAllMemoryBlocks (pIPState, &pSymbolHeap);
    if (pRunHeap != NULL)
	i_freeAllMemoryBlocks (pIPState, &pRunHeap);

    NORMAL_RETURN;
}



/*

 $Log: i_ccrun1.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:10  jlin
 Imported sources

 Revision 1.5  1998/01/13 23:47:12  ddavies
 i_ccrun1.c -> /product/ipcore/ipshared/src/RCS/i_ccrun1.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1997/07/11 01:48:33  ddavies
 i_ccrun1.c -> /product/ipcore/ipshared/src/RCS/i_ccrun1.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.3  1997/04/24 23:27:02  ddavies
 i_ccrun1.c -> /product/ipcore/ipshared/src/RCS/i_ccrun1.c,v

 Change all TCSymbols to CCSymbols.

 Revision 1.2  1995/12/13 02:42:29  ejaquith
 i_ccrun1.c -> /product/ipcore/ipshared/src/RCS/i_ccrun1.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1995/07/10  23:01:06  ddavies
 * Initial revision
 *
 * Revision 1.4  1995/06/08  20:53:09  ddavies
 * ddavies on Thu Jun 8 13:52:10 PDT 1995
 *
 * i_ccruns.c -> /product/ipcore/ipshared/src/RCS/i_ccruns.c,v
 *
 * Have i_findConnectedComponentRuns compute each CC's mass and number of runs
 * as well as everything else.
 * Add these as args to i_findConnectedComponentRuns' callback proc.
 *
 * Revision 1.3  1995/01/11  22:25:24  dferrell
 * i_ccruns.c -> /product/ipcore/ipshared/src/RCS/i_ccruns.c,v
 *
 * Removed include of shrcmprs.pub and added shrcc.pub and shrcc.prv.
 *
 * Revision 1.2  1994/06/09  03:27:00  ddavies
 * i_ccruns.c -> /product/ipcore/ipshared/src/RCS/i_ccruns.c,v
 *
 * Change the implementation of i_findConnectedComponentRuns so that the runs
 * are returned in bottom-to-top, right-to-left order.  Basically, the pointers
 * in the list are just turned around.  This simplifies putting new runs
 * on the list a little bit and speeds up join processing a lot.  We now merge
 * lists of runs from the bottom (where the join happens) up rather from the
 * top down.  In the old world, finding the group of runs in a list that should
 * be merged required starting at the top of the list and searching down.
 * The change improved performance on ccmail.1 from 8.4 sec to 2.9 sec.
 *
 * Revision 1.1  1994/06/08  03:57:59  ddavies
 * Initial revision
 *

 */

