/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_ccs.c
 *
 *    This file provides functions to deal with the ConnectedComponent
 *    data structure. This is an "abstract" entity which contains
 *    no pointers, so is theoretically DAE-friendly (well, somewhat)
 *    but represents a set of connected runs. We will represent the
 *    concept of a set of connected components at a higher level,
 *    since that will involve ickyness like "pointers".
 *
 * Operations on connected components:
 *    i_CCSet () 
 *    i_CCClear ()
 *
 * Freeing, etc. 
 *    i_CCDestroy ()
 *
 * Conversion from TCScanLineRunDecl to ConnectedComponent:
 *    i_RunListToCC()
 *    i_RunListToCCSize()
 *    i_RunListToCCCopy()
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "lmask.prv"
#include "imageref.pub"
#include "memory.pub"

/* prototyes */
#include "shrcc.pub"
#include "shrcc.prv"

#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_ccs.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

static UInt32   rightMask[32] =
{   0xFFFFFFFF, 0x7FFFFFFF, 0x3FFFFFFF, 0x1FFFFFFF,
    0x0FFFFFFF, 0x07FFFFFF, 0x03FFFFFF, 0x01FFFFFF,
    0x00FFFFFF, 0x007FFFFF, 0x003FFFFF, 0x001FFFFF,
    0x000FFFFF, 0x0007FFFF, 0x0003FFFF, 0x0001FFFF,
    0x0000FFFF, 0x00007FFF, 0x00003FFF, 0x00001FFF,
    0x00000FFF, 0x000007FF, 0x000003FF, 0x000001FF,
    0x000000FF, 0x0000007F, 0x0000003F, 0x0000001F,
    0x0000000F, 0x00000007, 0x00000003, 0x00000001 };

/* PUBLIC */

/***********************************************************************
 * i_CCSet ()
 *
 *   This function paints the runs of a ConnectedComponent into a bitmap.
 *   The runs are painted as black (1). No other pixels in the bitmap should
 *   be affected.
 *
 * Arguments:
 *   pDest:            Pointer to upper left corner of the destination bitmap,
 *                     INCLUDING frame
 *   bpl:              Distance between image lines lines, in bytes
 *   destX:            X-Offset into destination bitmap, including frame
 *   destY:            Y-offset into destination bitmap, including frame
 *   component:        The connected component to paint
 *
 * Returns:
 *   ia_successful:    all was well
 *
 ***********************************************************************/

Int32 CDECL 
i_CCSet (
    IPState    *pIPState,
    UInt8      *pDest,
    Int32	bpl,
    Int32	destX,
    Int32	destY,
    ConnectedComponent *component)
{
    PROLOGUE;
    Int32 i, x0, x1, lastx0, oWord, currOffset;

    /* Move the dest pointer to the bottommost line in the cc */
    pDest += (destY + component->y + component->height - 1) * bpl;

    lastx0 = 0x7fffffff;

    /* This code is mostly stolen from ip_symbolToBitmap */
    for (i = 0; i < component->runCount; i++)
    {
	x0 = component->runs[i].runLeft + destX;
	x1 = component->runs[i].runRight + destX;
	if (x1 >= lastx0 - 1)
	    pDest -= bpl;
	lastx0 = x0;

	oWord = leftMask[x1 & 31];
	currOffset = x1 & (~31);

	if (x0 < currOffset)
	{
	    ULONG_ACCESS(&pDest[currOffset / 8]) |= oWord;

	    oWord = 0xffffffff;
	    currOffset -= 32;

	    /* Fill in the intermediate words, stopping before
	     * we get to the word containing the left endpoint
	     * The test works because currOffset always points to the
	     * beginning of a word. 
	     */
	    while (x0 < currOffset)
	    {
		ULONG_ACCESS(&pDest[currOffset / 8]) = oWord;
		currOffset -= 32;
	    }
	}
	    
	oWord &= rightMask[x0 & 31];
	ULONG_ACCESS(&pDest[currOffset / 8]) |= oWord;
    }

    NORMAL_RETURN;
}

/* PUBLIC */

/***********************************************************************
 * i_CCClear ()
 *
 *   This function paints the runs of a ConnectedComponent into a bitmap.
 *   The runs are painted as white (0). No other pixels in the bitmap should
 *   be affected.
 *
 * Arguments:
 *   pDest:            Pointer to upper left corner of the destination bitmap,
 *                     INCLUDING frame
 *   bpl:              Distance between image lines lines, in bytes
 *   destX:            X-Offset into destination bitmap, including frame
 *   destY:            Y-offset into destination bitmap, including frame
 *   component:        The connected component to paint
 *
 * Returns:
 *   ia_successful:    all was well
 *
 ***********************************************************************/

Int32 CDECL 
i_CCClear (
    IPState    *pIPState,
    UInt8      *pDest,
    Int32	bpl,
    Int32	destX,
    Int32	destY,
    ConnectedComponent *component)
{
    PROLOGUE;
    Int32 i, x0, x1, lastx0, oWord, currOffset;

    /* Move the dest pointer to the bottommost line in the cc */
    pDest += (destY + component->y + component->height - 1) * bpl;

    lastx0 = 0x7fffffff;

    /* This code is mostly stolen from ip_symbolToBitmap */
    for (i = 0; i < component->runCount; i++)
    {
	x0 = component->runs[i].runLeft + destX;
	x1 = component->runs[i].runRight + destX;
	if (x1 >= lastx0 - 1)
	    pDest -= bpl;
	lastx0 = x0;

	oWord = ~leftMask[x1 & 31];
	currOffset = x1 & (~31);

	if (x0 < currOffset)
	{
	    ULONG_ACCESS(&pDest[currOffset / 8]) &= oWord;

	    currOffset -= 32;
	    oWord = 0;
	    /* Fill in the intermediate words, stopping before
	     * we get to the word containing the left endpoint
	     * The test works because currOffset always points to the
	     * beginning of a word. 
	     */
	    while (x0 < currOffset)
	    {
		ULONG_ACCESS(&pDest[currOffset / 8]) = 0;
		currOffset -= 32;
	    }
	}
	    
	oWord = ~(~oWord & rightMask[x0 & 31]);
	ULONG_ACCESS(&pDest[currOffset / 8]) &= oWord;
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/***********************************************************************
 * i_CCInterpolateEdges
 *
 *    The purpose of this function is to paint what's under a connected
 *    component with an interpolated value which runs smoothly from the value
 *    at the left edge to the value at the right edge.
 *
 *    The destination image, in this initial implementation, must be 8
 *    bits/pixel. 
 *
 * Arguments:
 *   pDest:            Pointer to upper left corner of the destination image,
 *                     INCLUDING frame
 *   bpl:              Distance between image lines lines, in bytes
 *   destX:            X-Offset into destination image, including frame
 *   destY:            Y-offset into destination image, including frame
 *   component:        The connected component to paint
 *
 * Returns:
 *   ia_successful:    all was well
 ***********************************************************************/

Int32 CDECL 
i_CCInterpolateEdges (
    IPState    *pIPState,
    UInt8      *pDest,
    Int32	bpl,
    Int32	destX,
    Int32	destY,
    ConnectedComponent *component)
{
    PROLOGUE;
    Int32 i, x0, x1, lastx0, l0, l1, dl, x, l;

    /* Move the dest pointer to the bottommost line in the cc */
    pDest += (destY + component->y + component->height - 1) * bpl;

    lastx0 = 0x7fffffff;

    for (i = 0; i < component->runCount; i++)
    {
	x0 = component->runs[i].runLeft + destX;
	x1 = component->runs[i].runRight + destX;
	if (x1 >= lastx0 - 1)
	    pDest -= bpl;
	lastx0 = x0;

	/* Get the endpoints */
	l0 = UCHAR_ACCESS (pDest + x0);
	l1 = UCHAR_ACCESS (pDest + x1);

	/* Use 1/256ths to interpolate l */
	x = x0;
	l = l0 << 8;
	dl = ((l1 - l0) * 256);
	if (x1 != x0)
	    dl /= x1 - x0;
        
	/* x is the "physical" index, l is the interpolated value */
	while (x <= x1)
	{
	    UCHAR_ACCESS (pDest + x) = (UInt8)(l >> 8);
	    l += dl;
	    x++;
	}
    }

    NORMAL_RETURN;
}

/* PUBLIC */

/***********************************************************************
 * i_CCDestroy
 *
 *    Destroy a ConnectedComponent object
 *
 * Arguments:
 *    component:         the cc to destroy
 *
 * Returns:

 *    a null pointer to which the cc pointer can be assigned
 ***********************************************************************/

ConnectedComponent * CDECL 
i_CCDestroy (
    IPState    *pIPState,
    ConnectedComponent *component)
{
    PROLOGUE;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    IP_FREE (component);

    return (ConnectedComponent *) NULL;
}

/* PUBLIC */

/***********************************************************************
 * i_runListToCC
 *
 *    Convert a TCScanLineRun to a ConnectedComponent. The ConnectedComponent
 *    is allocated and returned.
 *
 * Arguments:
 *    scanLineRuns:              The TCScanLineRun to convert
 *    x, y, w, h:                Bounding box of the CC
 *    nRuns:                     Number of runs in list. If -1, count them.
 *    ppCC:			 Pointer to returned CC pointer
 *
 * Returns:
 *    The ConnectedComponent created, or NULL if the MALLOC failed.
 *
 ***********************************************************************/

ErrCode CDECL 
i_runListToCC (
    IPState    *pIPState,
    TCScanLineRun *scanLineRuns,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	nRuns,
    ConnectedComponent **ppCC)
{
    PROLOGUE;
    ConnectedComponent *component;
    Int32	numBytes;

    ERR_PASS(i_runListToCCSize (pIPState, scanLineRuns, nRuns, &numBytes));
    component = (ConnectedComponent *) MALLOC (numBytes);

    if (component != NULL)
    {
	ERR_PASS(i_runListToCCCopy (pIPState, component,
				    scanLineRuns, x, y, w, h));
	*ppCC = component;
    }
    else
	__error__ = ia_nomem;

    NORMAL_RETURN;
}

/* PUBLIC */

/***********************************************************************
 *
 * i_runListToCCSize
 * 
 *     Computes the number of bytes which will be needed to represent
 *     a run list as a single connected component structure (which has
 *     an array of runs inside of it)
 *
 * Arguments:
 *     scanLineRuns:            The TCScanLineRun for which to compute size
 *     nRuns:                   Number of runs, or -1 if we should count them
 *     pNumBytes:		Pointer to returned number of bytes.
 * 
 * Returns:
 *     number of bytes
 *
 ***********************************************************************/

Int32 CDECL 
i_runListToCCSize (
    IPState    *pIPState,
    TCScanLineRun *scanLineRuns,
    Int32	nRuns,
    Int32      *pNumBytes)
{
    /* Count the runs, if needed */
    PROLOGUE;
    if (nRuns < 0)
    {
	nRuns = 0;
	while (scanLineRuns != NULL)
	{
	    nRuns++;
	    scanLineRuns = scanLineRuns->pNext;
	}
    }

    /* Don't forget that the first run is included in the size of the
       ConnectedComponent structure itself. */
    *pNumBytes = sizeof (Run16) * (nRuns - 1) + sizeof (ConnectedComponent);

    NORMAL_RETURN;
}

/* PUBLIC */
/***********************************************************************
 *
 * i_runListToCCCopy
 *
 *    Convert a TCScanLineRun structure into a ConnectedComponent structure.
 *    
 * Arguments:
 *    component:       The connected component to copy into 
 *    scanLineRuns:    Run list, as a TCScanLineRun list
 *    x, y, w, h:      Bounding box of the CC
 *
 * Returns:
 *    ia_successful
 *
 ***********************************************************************/

Int32 CDECL 
i_runListToCCCopy (
    IPState    *pIPState,
    ConnectedComponent *component,
    TCScanLineRun *scanLineRuns,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h)
{
    PROLOGUE;
    Int32 i;

    component->x = x;
    component->y = y;
    component->width = w;
    component->height = h;
    component->runCount = 0;

    for (i = 0; scanLineRuns != NULL; scanLineRuns = scanLineRuns->pNext, i++)
    {
	component->runs[i].runLeft = scanLineRuns->runLeft;
	component->runs[i].runRight = scanLineRuns->runRight;
	component->runCount++;
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/***********************************************************************
 * i_CCPlugHorizontalGaps
 *
 *    Transforms a connected component into one where there is only
 *    one run per scanline, starting at the left edge of the leftmost
 *    run of the scanline, and ending at the right edge of the rightmost
 *    run of the scanline. This guarantees that we plug all interior holes,
 *    which allows one to use fairly stupid schemes for "peeling off" the
 *    cc and filling the background with some color (like the mean or
 *    median around the edges of the cc, for example)
 *    The new component should be allocated by the caller; call 
 *    i_CCPluggedSize to find out how big to make it.
 *
 * Arguments:
 *    dest:        The resulting connected component with horizontal gaps
 *                 eliminated -- must have been allocated by the caller
 *    src:         The connected component from which to work
 *
 * Returns:
 *    ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
i_CCPlugHorizontalGaps (
    IPState    *pIPState,
    ConnectedComponent *dest,
    ConnectedComponent *src)
{
    PROLOGUE;
    Int32 i, x0, x1, lastx0, lowest, highest, runs, outRun;

    lastx0 = 0x7fffffff;
    runs = 0;
    lowest = 0x7fffffff;
    highest = 0;
    outRun = 0;

    /* runs go right-left, bottom-top. */
    for (i = 0; i < src->runCount; i++)
    {
	x0 = src->runs[i].runLeft;
	if (x0 < lowest)
	    lowest = x0;
	x1 = src->runs[i].runRight;
	if (x1 > highest)
	    highest = x1;
	
	if (x1 >= lastx0)
	{
	    if (runs > 1)
	    {
		/* More than one run; stick in a new run */
		dest->runs[outRun].runLeft = (Int16)lowest;
		dest->runs[outRun++].runRight = (Int16)highest;
	    }
	    else if (runs == 1 && i > 0)
	    {
		/* Just copy that run */
		dest->runs[outRun++] = src->runs[i - 1];
	    }

	    runs = 1;
	    lowest = x0;
	    highest = x1;
	}
	else
	    runs++;
	lastx0 = x0;
    }

    dest->runCount = outRun;

    /* It should be impossible for the bounding box to change */
    dest->x = src->x;
    dest->y = src->y;
    dest->width = src->width;
    dest->height = src->height;
    dest->next = NULL;
 
    NORMAL_RETURN;
}

/* PUBLIC */
/***********************************************************************
 * i_CCPlugSize
 *
 *    If one wants to call i_CCPlugHorizontalGaps, call this first to 
 *    find out how big to make the destination ConnectedComponent, which
 *    the caller is responsible for creating
 *
 * Arguments:
 *    src:         The connected component from which to work
 *    pCCSize:	   Pointer to returned Connected Component size in bytes
 *
 * Returns:
 *    # of bytes in the new connected component, or -1 on error
 ***********************************************************************/

Int32 CDECL 
i_CCPlugSize (
    IPState    *pIPState,
    ConnectedComponent *src,
    Int32      *pCCSize)
{
    PROLOGUE;
    Int32 i, x0, lastx0, runs;

    lastx0 = 0x7fffffff;
    runs = 0;

    /* runs go right-left, bottom-top. */
    for (i = 0; i < src->runCount; i++)
    {
	x0 = src->runs[i].runLeft;
	
	if (src->runs[i].runRight >= lastx0)
	{
	    runs++;
	}
	lastx0 = x0;
    }

	/* find the size of a CC structure that holds those runs */
    ERR_PASS(i_runListToCCSize (pIPState, NULL, runs, pCCSize));

    NORMAL_RETURN;
}

