/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_mhc.c: High level utilities used in CCITT (MH, MR, MMR) compression.
 *
 *   	Int32	i_getSizeOfCCITTCompressionInfo()
 *			Returns size of opaque data structure used in
 *			CCITT compression.
 *
 *   	Int32	i_getSizeOfCCITTDecompressionInfo()
 *			Returns size of opaque data structure used in
 *			CCITT compression.
 *
 *	Int32	i_initializeCCITTCompressionInfo()
 *			This routine initializes the opaque data structure
 *			used in CCITT compression.
 *
 *	Int32	i_initializeCCITTDecompressionInfo()
 *			This routine initializes the opaque data structure
 *			used in CCITT decompression.
 *
 *	Int32	i_flushCCITT()
 *			This routine flushes the pipeline.
 *
 *	Int32	i_writeCCITTEOL()
 *			This routine writes and EndOfLine code.
 *
 *      Int32	i_getMostRecentRuns()
 *			Returns pointer to last array of runs decoded.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#define	cDummyMarker	0xFFFFFFFF	/* use to mark dummy run beyond end of
					 * valid runs */

IP_RCSINFO(RCSInfo, "$RCSfile: i_ccttut.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* This file contains routines used to general CCITT compression.  It
 * doesn't contain any of the compression routines themselves.
 */

/* PUBLIC */
/**************************************************************************
 *   i_getSizeOfCCITTCompressionInfo()
 *
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform Modified
 *	Huffman compression.
 *
 *	Input:
 *		w:	width of image including frame
 *		pSize:	pointer to Int32 that will contain the size.
 *			of the buffer required to hold the symbol
 *			information in bytes.
 *
 *	Note:
 *		This compression should only be used on binary images.
 *
 *	Returns:
 *		ia_successful		 it worked
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfCCITTCompressionInfo (
    IPState    *pIPState,
    Int32	w,
    Int32     *pSize)
{
	/* remove the frame */
    PROLOGUE;
    w -= 2 * cFrameBits;

	/* Return size of CCITT compression structure
	 * plus two arrays big enough to hold the endpoints
	 * of the runs (a Run for every 2 pixels).
	 * There's also one dummy struct at the beginning
	 * of each run array and two dummy structs at the end.
	 */
    *pSize = sizeof(CCITTCompressionInfo) +
	     2 * ((((w+1)/2) + 3) * sizeof(Run32));

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getSizeOfCCITTDecompressionInfo()
 *
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform Modified
 *	Huffman decompression.
 *
 *	Input:
 *		w:	width of image including frame
 *		pSize:	pointer to Int32 that will contain the size.
 *			of the buffer required to hold the symbol
 *			information in bytes.
 *
 *	Note:
 *		This decompression should only be used on binary images.
 *
 *	Returns:
 *		ia_successful		 it worked
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfCCITTDecompressionInfo (
    IPState    *pIPState,
    Int32	w,
    Int32     *pSize)
{
	/* remove the frame */
    PROLOGUE;
    w -= 2 * cFrameBits;

	/* Return size of CCITT decompression structure
	 * plus two arrays big enough to hold the endpoints
	 * of the runs (a Run for every 2 pixels).
	 * There's also one dummy struct at the beginning
	 * of each run array and two dummy structs at the end.
	 */
    *pSize = sizeof(CCITTDecompressionInfo) +
	     2 * ((((w+1)/2) + 3) * sizeof(Run32));

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_initializeCCITTCompressionInfo()
 *
 *	This procedure fills a block of CCITT compression info
 *	with necessary information.
 *
 *	Input:
 *		pCCITTCompInfoIn:	pointer to CCITT compression info
 *		w:			width of uncompressed image including
 *					frame.
 *		useMR2D:		cTrue if we're going to do MR
 *					(CCITT Group 3) compression and
 *					will use the 1D/2D hybrid mode.
 *		maxK:			set to 2 or 4 if useMR2D is true.
 *					Equals the k factor defined in
 *					2D Group 3 mode.
 *		padType:		Type of scan line padding.
 *					 If compression is MH:
 *					    cPadToByte => pad each scan
 *						line to a byte boundary.
 *					    cPadToWord => pad each scan
 *						line to a word boundary.
 *					 if compression is MR:
 *					    cNoPad => insert no pad before
 *						EOL.
 *					    cPadToByte => insert pad before
 *						EOL so EOL ends on byte bdry.
 *					 if compression = MMR:
 *						ignored.
 *		compInfoBufSize:	the number of bytes in the
 *					compression info.
 *
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pCCITTCompInfoIn was NULL or buffer was
 *					wrong size
 *
 **************************************************************************/
Int32  CDECL 
i_initializeCCITTCompressionInfo (
    IPState    *pIPState,
    void       *pCCITTCompInfoIn,
    Int32	w,
    Bool	useMR2D,
    Int32	maxK,
    UInt32	padType,
    Int32	compInfoBufSize)
{
    PROLOGUE;
    CCITTCompressionInfo *pCCITTCompInfo;
    Int32		 realCompInfoBufSize;
    Int32		 run;
    Run32		*pRun, *pRun2;
    

	/* initialize a useful pointer */
    pCCITTCompInfo = (CCITTCompressionInfo *)pCCITTCompInfoIn;

	/* Get the size the block should be */
    __error__ = i_getSizeOfCCITTCompressionInfo (pIPState, w,
					         &realCompInfoBufSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* See if any of the parameters were illegal */
    if ((pCCITTCompInfo == NULL) ||
	(realCompInfoBufSize > compInfoBufSize) ||
	(useMR2D && (maxK != 2) && (maxK != 4)) ||
	((padType != cNoPad) && (padType != cPadToByte) &&
	 (padType != cPadToWord)))
	return ia_invalidParm;

	/* Remove frame from the width */
    w -= 2*cFrameBits;

	/* Initialize all the rest of the default stuff */
    pCCITTCompInfo->imageInfoIsValid = 0;
    pCCITTCompInfo->lineNumber = 0;
    pCCITTCompInfo->runArrayOffset = sizeof(CCITTCompressionInfo);
    pCCITTCompInfo->run2ArrayOffset = pCCITTCompInfo->runArrayOffset +
					((((w+1)/2) + 3) * sizeof(Run32));
    pCCITTCompInfo->run2ArrayIsCoding = cFalse;
    pCCITTCompInfo->currentRun = 0;
    pCCITTCompInfo->refRun = 0;
    pCCITTCompInfo->use2D = useMR2D;
    pCCITTCompInfo->k = maxK;
    pCCITTCompInfo->maxK = maxK;
    pCCITTCompInfo->padType = padType;
    pCCITTCompInfo->compInfoIsValid = 0;
    pCCITTCompInfo->compState = cCCITTCompNewLine;
    pCCITTCompInfo->accumBits = 0;
    pCCITTCompInfo->vacantBit = 0;
    pCCITTCompInfo->remainingLength = 0;

	/* Initialize the array of runs.  The first run has coords of
	 * -1, -1.  Set the rest to 0,0.
	 */
    pRun = (Run32 *)((UInt8 *)pCCITTCompInfo +
				pCCITTCompInfo->runArrayOffset);
    pRun2 = (Run32 *)((UInt8 *)pCCITTCompInfo +
				pCCITTCompInfo->run2ArrayOffset);
    pRun[0].runLeft = -1;
    pRun[0].runRight = -1;
    pRun2[0].runLeft = -1;
    pRun2[0].runRight = -1;

    for (run = ((w + 1)/2) + 2; run > 0; run--)
    {
	pRun[run].runLeft = 0;
	pRun[run].runRight = 0;
	pRun2[run].runLeft = 0;
	pRun2[run].runRight = 0;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_initializeCCITTDecompressionInfo()
 *
 *	This procedure fills a block of CCITT decompression info
 *	with necessary information.
 *
 *	Input:
 *		pCCITTDecompInfoIn:	pointer to CCITT decompression info
 *		w:			width of uncompressed image including
 *					frame.
 *		useMR2D:		cTrue if we're going to do MR
 *					(CCITT Group 3) decompression and
 *					will use the 1D/2D hybrid mode.
 *		padType:		Type of scan line padding.
 *					 If compression is MH:
 *					    cPadToByte => pad each scan
 *						line to a byte boundary.
 *					    cPadToWord => pad each scan
 *						line to a word boundary.
 *					 if compression is MR:
 *					    cNoPad => insert no pad before
 *						EOL.
 *					    cPadToByte => insert pad before
 *						EOL so EOL ends on byte bdry.
 *					 if compression = MMR:
 *						ignored.
 *		decompInfoBufSize:	the number of bytes in the
 *					decompression info.
 *
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pCCITTDecompInfoIn was NULL or
 *					buffer was wrong size
 *
 **************************************************************************/
Int32  CDECL 
i_initializeCCITTDecompressionInfo (
    IPState    *pIPState,
    void       *pCCITTDecompInfoIn,
    Int32	w,
    Bool	useMR2D,
    UInt32	padType,
    Int32	decompInfoBufSize)
{
    PROLOGUE;
    CCITTDecompressionInfo *pCCITTDecompInfo;
    Int32		 realDecompInfoBufSize;
    Int32		 run;
    Run32		*pRun, *pRun2;
    

	/* initialize a useful pointer */
    pCCITTDecompInfo = (CCITTDecompressionInfo *)pCCITTDecompInfoIn;

	/* Get the size the block should be */
    __error__ = i_getSizeOfCCITTDecompressionInfo (pIPState, w,
						   &realDecompInfoBufSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* See if any of the parameters were illegal */
    if ((pCCITTDecompInfo == NULL) ||
	(realDecompInfoBufSize > decompInfoBufSize) ||
	((padType != cNoPad) && (padType != cPadToByte) &&
	 (padType != cPadToWord)))
	return ia_invalidParm;

	/* Remove frame from the width */
    w -= 2*cFrameBits;

	/* Initialize all the rest of the default stuff */
    pCCITTDecompInfo->width = w;
    pCCITTDecompInfo->imageInfoIsValid = 0;
    pCCITTDecompInfo->lineNumber = 0;
    pCCITTDecompInfo->runArrayOffset = sizeof(CCITTDecompressionInfo);
    pCCITTDecompInfo->run2ArrayOffset = pCCITTDecompInfo->runArrayOffset +
					((((w+1)/2) + 3) * sizeof(Run32));
    pCCITTDecompInfo->run2ArrayIsCoding = cFalse;
    pCCITTDecompInfo->run2ArrayMostRecent = cFalse;
    pCCITTDecompInfo->currentRun = 0;
    pCCITTDecompInfo->refRun = 0;
    pCCITTDecompInfo->use2D = useMR2D;
    pCCITTDecompInfo->EOLType = cEOL1D;
    pCCITTDecompInfo->padType = padType;
    pCCITTDecompInfo->decompInfoIsValid = 0;
    pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
    pCCITTDecompInfo->accumBits = 0;
    pCCITTDecompInfo->vacantBit = 0;
    pCCITTDecompInfo->remainingLength = 0;

	/* Initialize the array of runs.  The first run has coords of
	 * -1, -1.  Set the rest to 0,0.
	 */
    pRun = (Run32 *)((UInt8 *)pCCITTDecompInfo +
				pCCITTDecompInfo->runArrayOffset);
    pRun2 = (Run32 *)((UInt8 *)pCCITTDecompInfo +
				pCCITTDecompInfo->run2ArrayOffset);
    pRun[0].runLeft = -1;
    pRun[0].runRight = -1;
    pRun2[0].runLeft = -1;
    pRun2[0].runRight = -1;

    for (run = ((w + 1)/2) + 2; run > 0; run--)
    {
	pRun[run].runLeft = 0;
	pRun[run].runRight = 0;
	pRun2[run].runLeft = 0;
	pRun2[run].runRight = 0;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_flushCCITT()
 *
 *	This procedure flushes the compression pipeline to the output
 *	data stream.
 *
 *	Input:
 *		pCCITTCompInfoIn:	Pointer to CCITT compression info
 *		pCompData:		Pointer to first vacant byte
 *					of buffer into which compressed
 *					data will be placed.
 *		compDataBufSize:	Number of vacant bytes remaining
 *					in the compressed data buffer.
 *		pBytesWritten:		ptr to returned number of compressed
 *					bytes written.
 *
 *	Note:
 *		This routine should only be called at the end of strips,
 *		tiles or images.
 *
 *	Returns:
 *		ia_successful		 it worked
 *		ia_invalidParm		 pCCITTCompInfoIn or pCompData is
 *					 NULL
 *		ia_bufferExceeded	 not enough room in compressed
 *					 data buffer.
 *
 **************************************************************************/
Int32  CDECL 
i_flushCCITT (
    IPState    *pIPState,
    void       *pCCITTCompInfoIn,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Int32     *pBytesWritten)
{
    PROLOGUE;
    CCITTCompressionInfo	*pCCITTCompInfo;
    UInt8			*pCompDataEnd;	/* address 1 beyond the last
						 * byte of compressed data */
    UInt8			*pCurCompData;	/* current compressed data
						 * pointer */
    

	/* check for null pointers */
    if ((pCCITTCompInfoIn == NULL) || (pCompData == NULL) ||
	(pBytesWritten == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pCCITTCompInfo = (CCITTCompressionInfo *) pCCITTCompInfoIn;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Call the routine that does the flushing.
	 */
    __error__ = ip_flushCCITT (pIPState, (CCITTCompressionInfo *)pCCITTCompInfoIn,
			     &pCurCompData, pCompDataEnd);

	/* Record the number of bytes written and the return code.
	 * If __error__ == ia_bufferExceeded, the client should
	 * supply us with more data and try again.
	 */
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_writeCCITTEOL()
 *
 *	This procedure flushes the compression pipeline to the output
 *	data stream.
 *
 *	Input:
 *		pCCITTCompInfoIn:	Pointer to CCITT compression info
 *		padType:		If equal to cPadToByte, insert pad
 *					before EOL so that EOL ends on a
 *					byte boundary.  If equal to cNoPad, 
 *					do not insert padding before EOL.
 *		EOLType:		If equal to cEOL1D, follow the EOL
 *					with a bit indicating that the next 
 *					line will be encoded with 1D encoding.
 *					If equal to cEOL2D, append a bit 
 *					indicating that the next line will 
 *					be encoded using 2D codes.  If 
 *					cPlainEOL, do not append a bit.
 *		pCompData:		Pointer to first vacant byte
 *					of buffer into which compressed
 *					data will be placed.
 *		compDataBufSize:	Number of vacant bytes remaining
 *					in the compressed data buffer.
 *		pBytesWritten:		ptr to returned number of compressed
 *					bytes written.
 *
 *	Note:
 *		This routine should only be called at the end of strips,
 *		tiles or images.
 *
 *	Returns:
 *		ia_successful		 it worked
 *		ia_invalidParm		 pCCITTCompInfoIn or pCompData is
 *					 NULL
 *		ia_bufferExceeded	 not enough room in compressed
 *					 data buffer.
 *
 **************************************************************************/
Int32  CDECL 
i_writeCCITTEOL (
    IPState    *pIPState,
    void       *pCCITTCompInfoIn,
    UInt32	padType,
    UInt32	EOLType,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Int32      *pBytesWritten)
{
    PROLOGUE;
    CCITTCompressionInfo	*pCCITTCompInfo;
    UInt8			*pCompDataEnd;	/* address 1 beyond the last
						 * byte of compressed data */
    UInt8			*pCurCompData;	/* current compressed data
						 * pointer */
    

	/* check for null pointers */
    if ((pCCITTCompInfoIn == NULL) || (pCompData == NULL) ||
	(pBytesWritten == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pCCITTCompInfo = (CCITTCompressionInfo *) pCCITTCompInfoIn;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Call the routine that does the flushing.
	 */
    __error__ = ip_writeCCITTEOL (pIPState,
				(CCITTCompressionInfo *)pCCITTCompInfoIn,
				padType, EOLType,
				&pCurCompData, pCompDataEnd);

	/* Record the number of bytes written and the return code.
	 * If __error__ == ia_bufferExceeded, the client should
	 * supply us with more data and try again.
	 */
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getMostRecentRuns()
 *
 *	This procedure is used when we are converting encoded stuff
 *	directly to runs.  It returns a pointer to the beginning of
 *	the most recently written array of runs..
 *
 *	Input:
 *		pCCITTDecompInfoIn:	pointer to CCITT decompression info
 *
 *	Output:
 *		pMostRecentRunArray:	pointer to ptr to array of most
 *					recently decoded runs.
 *
 *	Returns:
 *		ia_successful		 it worked
 *		ia_invalidParm		 pCCITTDecompInfoIn is NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getMostRecentRuns (
    IPState    *pIPState,
    void       *pCCITTDecompInfoIn,
    Run32     **pMostRecentRunArray)
{
    PROLOGUE;
    CCITTDecompressionInfo *pCCITTDecompInfo;

	/* initialize a useful pointer */
    pCCITTDecompInfo = (CCITTDecompressionInfo *)pCCITTDecompInfoIn;

	/* is it null? */
    if (pCCITTDecompInfo == NULL)
	return ia_invalidParm;

    if (pCCITTDecompInfo->run2ArrayMostRecent)
	*pMostRecentRunArray = (Run32 *)((UInt8 *)pCCITTDecompInfo +
				pCCITTDecompInfo->run2ArrayOffset +
				sizeof(Run32));
    else
	*pMostRecentRunArray = (Run32 *)((UInt8 *)pCCITTDecompInfo +
				pCCITTDecompInfo->runArrayOffset +
				sizeof(Run32));
    NORMAL_RETURN;
}

/*

 $Log: i_ccttut.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:10  jlin
 Imported sources

 Revision 1.9  1998/01/13 23:47:17  ddavies
 i_ccttut.c -> /product/ipcore/ipshared/src/RCS/i_ccttut.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.8  1997/08/20 20:23:21  ddavies
 ddavies on Wed Aug 20 13:22:44 PDT 1997

 i_ccttut.c -> /product/ipcore/ipshared/src/RCS/i_ccttut.c,v

 Change musrMR2D to useMR2D in comments.

 Revision 1.7  1996/07/18 23:35:22  ddavies
 ddavies on Thu Jul 18 16:34:06 PDT 1996

 i_ccttut.c -> /product/ipcore/ipshared/src/RCS/i_ccttut.c,v

 Add initialization of remainingLength entry to
 i_initializeCCITTDecompressionInfo.  It was left out by mistake.

 Revision 1.6  1996/05/21 18:44:44  ddavies
 i_ccttut.c -> /product/ipcore/ipshared/src/RCS/i_ccttut.c,v

 Add i_getMostRecentRuns.  This procedure returns a pointer to the most
 recently decoded array of runs from the CCITTDecompressionInfo structure.

 Revision 1.5  1996/05/11 00:59:58  ddavies
 i_ccttut.c -> /product/ipcore/ipshared/src/RCS/i_ccttut.c,v

 Fix a bug found by William Rucklidge!  The run arrays used by the
 CCITTDecompression routines should have ((((w+1)/2) + 3) entries, not
 (((w+1)/2) + 1.  I added two extra dummy elements during debugging the
 MMR code and forgot to increase the array sizes.
 Add useMR2D, maxK and padType args to i_initializeCCITTCompressionInfo.
 The first two are used in the 2D option for CCITT Group 3 compression.
 The padType was an argument to the compression routines that didn't
 ever change inside an image, so was moved to the structure.
 Add useMR2D and padType to i_initializeCCITTDecompressionInfo.  maxK
 is not needed since we decide the compression type of each line by
 looking at the type of the EndOfLine code preceeding it.

 Revision 1.4  1996/05/06 18:44:08  ddavies
 i_ccttut.c -> /product/ipcore/ipshared/src/RCS/i_ccttut.c,v

 Fix CCITTDecompInfo initialization routine so that it touches the
 structure entries that really exist in the latest structure (refRunRight
 and refRunLeft -> refRun).

 Revision 1.3  1996/05/01 23:11:39  ddavies
 i_ccttut.c -> /product/ipcore/ipshared/src/RCS/i_ccttut.c,v

 Add initialization of second array of runs in both compression and
 decompress structures used by CCITT code.  Also initialize associated
 entries in those structures.  Add prototypes for two new routines:
 i_flushCCITT and i_writeCCITTEOL.  Whoops, this is a C file so we're
 adding the routines themselves, not the prototypes.  i_flushCCITT
 flushes the pipeline used in CCITT compression to the output data stream.
 It's used in MR and MMR compression in which individual scan lines are
 not padded to byte boundaries.  i_writeCCITTEOL is used to write one
 end of line code during MR and MMR compression.  It may be preceeded
 by optional padding and followed by an optional flag bit.  That bit
 is used in MR compression to indicate whether the following scan line
 is encoded using 1D or 2D compression.

 Revision 1.2  1996/04/25 18:40:31  ddavies
 ddavies on Thu Apr 25 11:38:35 PDT 1996

 i_ccttut.c -> /product/ipcore/ipshared/src/RCS/i_ccttut.c,v

 Add routines used to manage CCITT decompression opaque state block.
 These routines are i_getSizeOfCCITTDecompressionInfo and
 i_initializeCCITTDecompressionInfo.  The first routine simply returns
 the size of the state block.  The second initializes the state block
 after the client has allocated it.  The client is responsible for freeing
 the block.

 Revision 1.1  1996/04/10 22:04:16  ddavies
 Initial revision

 */


