/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_ccutil.c: Utilities used by the versions of connected component
 *		finding code that don't allocate and destroy their
 *		heaps every time they're called.
 *
 *		Int32	i_createCCState()	allocate and initialize
 *						state used by
 *						i_findCCBoundingBoxesWithState
 *						and i_findRun32sWithState
 *		Int32	i_destroyCCState()	frees this state.
 *
 *		Int32	ip_fixupCCStatePtrs	fix pointers in CCState
 *						after garbage collection
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

/* prototyes */
#include "shrcc.pub"
#include "shrcc.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_ccutil.c,v $; $Revision: 1.2 $; $Date: 2001/05/08 06:43:35 $");

/* The XIS folks would like to use the connected component finding code
 * on many small regions of an image.  They found that the original versions
 * of this code spent the bulk of its time allocating and freeing the
 * heaps used for runs and symbols.  They requested a version of the
 * code that can used a pre-allocated, pre-initialized heap.  This module
 * contains the routines that allocate and destroy the state structures
 * used for this purpose.
 */

/* PUBLIC */
/**************************************************************************
 *   i_createCCState()
 *
 *	This procedure allocates the opaque data block used by the
 *	lightweight connected component finding code.
 *
 *	Input:
 *		ppNewState:	Pointer to pointer to the new state.
 *				This is how the state pointer is returned.
 *		allocator:	Used to allocate the private structure
 *				and both heaps.  The second arg,
 *				fixupPointers will always be set to true
 *				since these are all persistent data
 *				structures.
 *		freer:		Used to destroy the private structure and
 *				both heap later.  The length arg contains
 *				the length of the block to be freed.
 *
 *	Action:
 *		Allocates the private CCState structure and initializes
 *		the two heaps used by both connected component routines.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_nomem		ran out of memory
 *		other			returned by called proc
 *
 **************************************************************************/
Int32  CDECL 
i_createCCState (
    IPState    *pIPState,
    void      **ppNewState,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	bufSizeInBytes,
		    UInt32	fixupPointers),
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *pBlock,
		    UInt32	length))
{
PROLOGUE;
 CCState		*pNewState;
TCBlockHeader	*pNewHeap;
TCScanLineRun	*pLine0;		/* ptr to older scan line */
TCScanLineRun	*pLine1;		/* ptr to newer scan line */


	/* First, allocate the new structure */
    pNewState = (CCState *)allocator(pIPState, sizeof(CCState), cTrue);

    if (pNewState == NULL)
	return ia_nomem;

	/* set the pointer fixup section */
    __error__ = ip_initPointerFixup (pIPState, &(pNewState->fixup),
				   sizeof(CCState),
				   NULL,	/* this is the first element
						 * of the ring */
				   ip_fixupCCStatePtrs);

    if (__error__ != ia_successful)
    {
	freer(pIPState, pNewState, sizeof(CCState));
	NORMAL_RETURN;
    }

	/* We now have the state structure.  Initialize the symbol heap
	 * and attach it to the state structure.
	 */
    __error__ = i_allocateMemoryBlock (pIPState, &pNewHeap,
				     sizeof(CCSymbol),
				     cCCSymbolBlockSize,
				     cTrue,
				     &(pNewState->fixup),
				     ip_fixupCCSymbolPtrs,
				     allocator,
				     freer);

    if (__error__ != ia_successful)
    {
	freer(pIPState, pNewState, sizeof(CCState));
	NORMAL_RETURN;
    }
    else	 /* it worked */
	pNewState->pSymbolHeap = pNewHeap;

	/* do the same for the run heap */
    __error__ = i_allocateMemoryBlock (pIPState, &pNewHeap,
				       sizeof(TCScanLineRun),
				       cTCScanLineRunBlockSize,
				       cTrue,
				       &(pNewState->fixup),
				       ip_fixupScanLinePtrs,
				       allocator,
				       freer);

    if (__error__ != ia_successful)
    {
	i_freeAllMemoryBlocks (pIPState, &(pNewState->pSymbolHeap));
	freer(pIPState, pNewState, sizeof(CCState));
	NORMAL_RETURN;
    }
    else	 /* it worked */
	pNewState->pRunHeap = pNewHeap;

	/* Allocate two ending runs for each line.  It's more efficient
	 * to do this here and have the state structure keep a pointer
	 * to the lists of allocated runs than to have the CC finding
	 * routine put structures back into the heap and take them
	 * out again on each call.
	 */
    __error__ = i_allocateStruct (pIPState, pNewState->pRunHeap,
				(void **) &pLine0);
    if (__error__ != ia_successful)
	goto createCCReturn;

    pLine0->runLeft = -cVeryLargeNum;
    pLine0->runRight = -cVeryLargeNum;
    pLine0->pClient = NULL;
    pLine0->pNext = NULL;

    __error__ = i_allocateStruct (pIPState, pNewState->pRunHeap,
				(void **) &(pLine0->pNext));
    if (__error__ != ia_successful)
	goto createCCReturn;

    pLine0->pNext->runLeft = cVeryLargeNum;
    pLine0->pNext->runRight = cVeryLargeNum;
    pLine0->pNext->pClient = NULL;
    pLine0->pNext->pNext = NULL;

    pNewState->pLine0 = pLine0;

    __error__ = i_allocateStruct (pIPState, pNewState->pRunHeap,
				(void **) &pLine1);
    if (__error__ != ia_successful)
	goto createCCReturn;

    pLine1->runLeft = -cVeryLargeNum;
    pLine1->runRight = -cVeryLargeNum;
    pLine1->pClient = NULL;
    pLine1->pNext = NULL;

    __error__ = i_allocateStruct (pIPState, pNewState->pRunHeap,
				(void **) &(pLine1->pNext));
    if (__error__ != ia_successful)
	goto createCCReturn;

    pLine1->pNext->runLeft = cVeryLargeNum;
    pLine1->pNext->runRight = cVeryLargeNum;
    pLine1->pNext->pClient = NULL;
    pLine1->pNext->pNext = NULL;

    pNewState->pLine1 = pLine1;

	/* Everything worked.  Set the allocator and freer entries,
	 * set the client's new state pointer and return successfully.
	 * We need the freer to free this block when we're done.  I'm
	 * not sure why we need to keep the allocator, but it's symmetric
	 * this way.  You never know.
	 */
    pNewState->ipcAllocator = allocator;
    pNewState->freer = freer;

    *ppNewState = pNewState;
    __error__ = ia_successful;

	/* we also get here after an error, so just return whatever code
	 * we had */
createCCReturn:
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_destroyCCState()
 *
 *	This procedure frees the opaque data block used by the
 *	lightweight connected component finding code.
 *
 *	Input:
 *		ppFreedState:	Pointer to pointer to the state to be freed.
 *
 *	Action:
 *		Frees the private CCState structure and
 *		the two heaps associated with it.  Sets
 *		*ppFreedState to NULL.
 *
 *	Returns:
 *		ia_successful		it worked
 *		other			returned by called proc
 *
 **************************************************************************/
Int32  CDECL 
i_destroyCCState (
    IPState    *pIPState,
    void      **ppFreedState)
{
PROLOGUE;
 CCState	*pFreedState;		/* more convenient to use */

	/* get a pointer to the freed block */
    pFreedState = *ppFreedState;

	/* free the heaps */
    i_freeAllMemoryBlocks (pIPState, &(pFreedState->pRunHeap));
    i_freeAllMemoryBlocks (pIPState, &(pFreedState->pSymbolHeap));

	/* free the state block itself */
    (pFreedState->freer)(pIPState, pFreedState, sizeof(CCState));

	/* Set the client's pointer to NULL */
    *ppFreedState = NULL;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_fixupCCStatePtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a CCState block.
 *
 *	Input:
 *		pThisBlock:	address of the CCState block.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in a CCState block to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupCCStatePtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
PROLOGUE;
 CCState	*pCCState;
UInt32	oldAddr;
UInt32	addrDiff;

	/* convert the local pointer to the right type */
    pCCState = (CCState *) pThisBlock;

	/* prepare for the comparisions */
    oldAddr = (UInt32) pOldAddr;

	/* Just go through all the pointers in a CCState struct, testing
	 * each one to see if it pointed to something that moved.
	 * We do the comparison by subtracting the old address from
	 * the pointer, treating the result as an unsigned number.  If
	 * the pointer was less than the old address (the old address
	 * always points to the beginning of the moved block), we
	 * don't want to touch the pointer.  In this case, the difference
	 * will be a very large unsigned number.  If the pointer was
	 * past the end of the block, the difference will be larger than
	 * blkLength.  In either case, the difference is larger than
	 * blkLength, so that's the test.
	 */

	/* The allocator and freer don't move, so we don't have to check
	 * them.
	 * So, the only set of things are the pointers to the heaps. */
    addrDiff = ((UInt32)(pCCState->pSymbolHeap) - oldAddr);
    if (addrDiff < blkLength)
	pCCState->pSymbolHeap = (TCBlockHeader *)
					(addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pCCState->pRunHeap) - oldAddr);
    if (addrDiff < blkLength)
	pCCState->pRunHeap = (TCBlockHeader *)
					(addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pCCState->pLine0) - oldAddr);
    if (addrDiff < blkLength)
	pCCState->pLine0 = (TCScanLineRun *) (addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pCCState->pLine1) - oldAddr);
    if (addrDiff < blkLength)
	pCCState->pLine1 = (TCScanLineRun *) (addrDiff + (UInt32)pNewAddr);

	/* That's it! */
    NORMAL_RETURN;
}




/*

 $Log: i_ccutil.c,v $
 Revision 1.2  2001/05/08 06:43:35  jlin
 change ->allocator to ->ipcAllocator

 * Revision 1.1.1.1  1999/09/20  18:18:10  jlin
 * Imported sources
 *
 Revision 1.5  1998/01/13 23:47:19  ddavies
 i_ccutil.c -> /product/ipcore/ipshared/src/RCS/i_ccutil.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1997/07/11 01:48:41  ddavies
 i_ccutil.c -> /product/ipcore/ipshared/src/RCS/i_ccutil.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.3  1997/04/24 23:27:45  ddavies
 i_ccutil.c -> /product/ipcore/ipshared/src/RCS/i_ccutil.c,v

 Change all TCSymbols to CCSymbols.
 Change ip_fixupSymbolPtrs to ip_fixupCCSymbolPtrs.

 Revision 1.2  1995/12/13 02:42:42  ejaquith
 i_ccutil.c -> /product/ipcore/ipshared/src/RCS/i_ccutil.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1995/07/14  02:37:02  ddavies
 * Initial revision
 *

 */


