/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_cdith:	    Color ordered dither to linear palette (
 *
 * Exports:
 *	Table Creation:
 *	Int32	    i_orderedRGBToLinearMakeTable ()
 *	Int32	    i_orderedRGBToLinearTableSize ()
 *
 *	Dither to linear palette (8-bit only):
 *	Int32	    i_orderedRGBToLinear8 ()
 *	Int32	    i_rescaleRGBToLinear8 ()
 *
 *	Dither to 4-bit palette:
 *	[Int32	    i_orderedRGBToPalette4 ()] NYI
 *	Int32	    i_rescaleRGBToPalette4 ()
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "imageref.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#include "odith.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

	/* prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"

	/* Local static data. This is recomputed each time, but is always
           the same, so re-entrancy shouldn't be an issue.*/
static UInt8 scaledDither8[cDitherPeriod][cDitherPeriod];


IP_RCSINFO(RCSInfo, "$RCSfile: i_cdith.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/*
 * The idea behind an ordered dither is quite simple. Suppose we're
 * converting an image from 24 bits to an 8 bit "linear combination"
 * palette, such as 6 red, 7 green, 6 blue for a total of 252 colors.
 * If we were to just scale each R, G, and B component to its
 * color in the 6-7-6 palette, we incur some error when we truncate,
 * resulting in countouring. With ordered dither, the idea is to
 * add an additional "fraction" to each component, such that the rounding
 * is "broken up" so that we don't see contouring. The way we do this
 * is to create a "dither matrix" which is a matrix which will be
 * sampled periodically through the input space. For example, if we had
 * a 16x16 dithering matrix, the corresponding value in the dither matrix
 * is (x mod 16, y mod 16). Theer is a bunch of theory on how to construct
 * a dither matrix such that the values are well distributed for best
 * visual effect (i.e. least objectionable pattern), one should refer to
 * Digital Halftoning by Ulichney for more detail. The matrix we're going
 * to use is just taken from that book and used as a constant.
 *
 * So how do we do this efficiently? Our approach will be fairly simple:
 * take (RGB) component, mask out the upper 6 bits, and shift up by three
 * such that we have 6 bits of "pixel" with 5 empty. The dither matrix
 * will be scaled such that max. is 5 bits, so we then OR in those bits
 * into the lower 5 bits, so we now have a pixel represented as 6 bits of
 * integer part and 5 bits of fractional part. Look this up in a scaling
 * table which converts the 11 bits into 8 output bits corresponding to
 * the correctly rounded channel value in the palette space. There will
 * be one LUT each for Red, Green, and Blue.  We will therefore need
 * 3x2KB = 6KB of look-up table. We then take the sum of these three
 * values (they're scaled up appropriately so that they add up correctly)
 * and this is our output pixel.
 * This could be done a bit more efficiently by having a look-up table
 * for each dither matrix position, thus doing one look-up, but on
 * most machines, that would fill the cache, so it's probably not a real
 * good deal. 6KB is already pushing it quite a bit on a 486, so we
 * might need to make it smaller. We could easily drop a bit out
 * of blue, and get another 1KB back, or we could use only 4 bits of
 * fraction, and get 3KB back. We'll have to experiment, to see where
 * things start breaking down.
 */

/* PUBLIC */
/***********************************************************************
 * i_orderedRGBToLinearMakeTable
 *
 *	Given a count of (r, g, b) values and some allocated memory
 *	(size computed by colorDitherTableSize), fill the tables.
 *
 * Arguments:
 *	reds, greens, blues -	Color counts
 *	rTRC, gTRC, bTRC    -	Tone-correction curves for each channel,
 *				if NULL then no tone-correction is done
 *	ditherTab	    -   pointer to block of memory allocated by
 *				the caller, in which table will be filled.
 *
 * Returns:
 *	error status
 ***********************************************************************/

Int32 CDECL 
i_orderedRGBToLinearMakeTable (
    IPState    *pIPState,
    Int32	reds,
    Int32	greens,
    Int32	blues,
    UInt8      *rTRC,
    UInt8      *gTRC,
    UInt8      *bTRC,
    void       *ditherTab)
{
    PROLOGUE;
    UInt8   *pR, *pG, *pB;
    Int32    i, j, pixVal, ditherFrac, r, g, b, dVal;
    Int32    rs, gs, bs, rd, gd, bd;

    pR = (UInt8 *) ditherTab;
    pG = pR + cChannelTableSize;
    pB = pG + cChannelTableSize;

    for (i = 0; i < cChannelTableSize; i++)
    {
	pixVal = i >> cPixShiftDown;
	ditherFrac = ((i & cPixFracMask) + 1) * cDitherFracScale - 1;
	r = g = b = pixVal;

	/* Look it up in the TRC, if passed */
	if (rTRC)
	{
	    r = rTRC[pixVal << cChanToPixShift] >> cChanToPixShift;
	}

	if (gTRC)
	{
	    g = gTRC[pixVal << cChanToPixShift] >> cChanToPixShift;
	}

	if (bTRC)
	{
	    b = bTRC[pixVal << cChanToPixShift] >> cChanToPixShift;
	}

	rs = ((r + 1) * (reds - 1)) - 1;
	rd = (rs + ditherFrac) / cPixMax;
	pR[i] = (UInt8)(MIN(rd, reds - 1) * greens * blues);

	gs = ((g + 1) * (greens - 1)) - 1;
	gd = (gs + ditherFrac) / cPixMax;
	pG[i] = (UInt8)(MIN(gd, greens - 1) * blues);

	bs = ((b + 1) * (blues - 1)) - 1;
	bd = (bs + ditherFrac) / cPixMax;
	pB[i] = (UInt8)(MIN(bd, blues - 1));

/* printf("i %ld pixVal %ld ditherFrac %ld results r %d g %d b %d\n", 
       i, pixVal, ditherFrac, pR[i], pG[i], pB[i]); */

    }

    /* Scale the dither matrix to 5 bits per entry. */
    for (i = 0; i < cDitherPeriod; i++)
    {
	for (j = 0; j < cDitherPeriod; j++)
	{
/*	    dVal = bigDitherMatrix[i][j] * 256 / (cDitherPeriod * cDitherPeriod); */
	    dVal = (UInt32) dither8[i][j] * 256 / 
		   (UInt32) (cDitherPeriod * cDitherPeriod);

	    scaledDither8[i][j] = (UInt8)(dVal >> 3);
	}
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/***********************************************************************
 * i_orderedRGBToLinearTableSize
 *
 *	Returns the size of a dither table, so that the caller can 
 *	allocate one
 *
 * Arguments:
 *	reds, greens, blues -	Color counts
 *      pTableSize - ptr to returned table size
 *
 * Returns:
 *	ia_successful, ...
 ***********************************************************************/

ErrCode CDECL 
i_orderedRGBToLinearTableSize (
    IPState    *pIPState,
    Int32	reds,
    Int32	greens,
    Int32	blues,
    Int32      *pTableSize)
{
    PROLOGUE;
    reds = greens = blues;
    *pTableSize = cChannelTableSize * 3;
    NORMAL_RETURN;
}


/* PUBLIC */
/***********************************************************************
 *
 * i_orderedRGBToLinear8
 *
 *	Do an ordered dither to a linear palette. Uses an implicit
 *	dither matrix.
 *
 * Arguments:
 *	pDest			-   Result image, .r to frame
 *	dBpl			-   Bytes between lines in dest image
 *	pRed, pGreen, pBlue	-   Pointers to source color plane, .r to frame
 *	sBpl			-   Bytes between lines in color planes
 *	x, y			-   Offset into source image (from frame)
 *	w, h			-   Result image size
 *	xStart			-   Starting x-offset, used to determine
 *				    the starting column in the dither matrix
 *				    Does NOT affect which pixel is processed.
 *	yStart			-   Starting y-offset, used to determine
 *				    the starting line in the dither matrix
 *				    Does NOT affect which line is processed.
 *	ditherTab		-   Look-up table created by i_makeOrderedRGBToLinearTable
 *
 * Returns:
 *	error status
 ***********************************************************************/

 Int32 CDECL 
i_orderedRGBToLinear8 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    Int32	sBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	xStart,
    Int32	yStart,
    void       *ditherTab)
{
    PROLOGUE;
    UInt8   *pRLut, *pGLut, *pBLut, *pDOrig;
    UInt8   *ditherRow;
    Int32    ditherVal;
    Int32    row, col, word, colCtr;
    UInt32   outWord, redWord, greenWord, blueWord, clearPadBits;
    UInt32  *pRWord, *pGWord, *pBWord, *pDWord;

    pRLut = (UInt8 *) ditherTab;
    pGLut = pRLut + cChannelTableSize;
    pBLut = pGLut + cChannelTableSize;

    /* Move the pointers to the right place */
    pD += (dBpl << cLogFrameBits) + cFrameBits;
    pR += (sBpl * y) + x;
    pG += (sBpl * y) + x;
    pB += (sBpl * y) + x;
    pDOrig = pD;
    clearPadBits = ~((1 << (8 * (4 - (w & 3)))) - 1);

    for (row = 0; row < h; row++)
    {
	/* Get a pointer to the row */
	ditherRow = &(scaledDither8[(row + yStart) % cDitherPeriod][0]);

	pRWord = (UInt32 *) pR;
	pGWord = (UInt32 *) pG;
	pBWord = (UInt32 *) pB;
	pDWord = (UInt32 *) pD;
	colCtr = xStart % cDitherPeriod;
	for (col = 0, word = 0; col < w ; word++, col += 4)
	{
	    redWord = pRWord[word];
	    greenWord = pGWord[word];
	    blueWord = pBWord[word];

	    /* First pixel */
	    ditherVal = ditherRow[colCtr++];
	    outWord  = pRLut[((redWord >> (24 - cPixIntShift)) & cPixIntMask) | ditherVal];
	    outWord += pGLut[((greenWord >> (24 - cPixIntShift)) & cPixIntMask) | ditherVal];
	    outWord += pBLut[((blueWord >> (24 - cPixIntShift)) & cPixIntMask) | ditherVal];
	    
	    outWord <<= 8;
	    if (colCtr >= cDitherPeriod)
		colCtr = 0;

	    /* Second  pixel */
	    ditherVal = ditherRow[colCtr++];
	    outWord += pRLut[((redWord >> (16 - cPixIntShift)) & cPixIntMask) | ditherVal];
	    outWord += pGLut[((greenWord >> (16 - cPixIntShift)) & cPixIntMask) | ditherVal];
	    outWord += pBLut[((blueWord >> (16 - cPixIntShift)) & cPixIntMask) | ditherVal];

	    outWord <<= 8;
	    if (colCtr >= cDitherPeriod)
		colCtr = 0;

	    /* Third pixel */
	    ditherVal = ditherRow[colCtr++];
	    outWord += pRLut[((redWord >> (8 - cPixIntShift)) & cPixIntMask) | ditherVal];
	    outWord += pGLut[((greenWord >> (8 - cPixIntShift)) & cPixIntMask) | ditherVal];
	    outWord += pBLut[((blueWord >> (8 - cPixIntShift)) & cPixIntMask) | ditherVal];

	    outWord <<= 8;
	    if (colCtr >= cDitherPeriod)
		colCtr = 0;

	    /* Fourth pixel */
	    ditherVal = ditherRow[colCtr++];
	    outWord += pRLut[((redWord << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pGLut[((greenWord << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pBLut[((blueWord << cPixIntShift) & cPixIntMask) | ditherVal];

	    if (colCtr >= cDitherPeriod)
		colCtr = 0;
	    pDWord[word] = outWord;
	}

	/* Clear the pad, if necessary */
	if (w & 3)
	{
	    pDWord[word - 1] &= clearPadBits;
	}

	pR += sBpl;
	pG += sBpl;
	pB += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/***********************************************************************
 *
 * i_rescaleRGBToLinear8
 *
 *	Do an ordered dither to a linear palette. Uses an implicit
 *	dither matrix. Does a non-interpolating (i.e. nearest neighbor
 *	resampling.
 *
 *	This uses floating point to compute the resampling table.
 *	We expect that machines which can't have floating point won't
 *	be doing any color image processing anyway. We don't use
 *	floating point for speed-critical parts, so it shouldn't matter
 *	if it's emulated
 *  
 * Arguments:
 *	pDest			-   Result image, .r to frame
 *	dBpl			-   Bytes between lines in dest image
 *	wD, hD			-   Width, height of dest. image
 *	pRed, pGreen, pBlue	-   Pointers to source color plane, .r to frame
 *	sBpl			-   Bytes between lines in color planes
 *	xS, yS			-   Offset into source image (from frame)
 *	wS, hS			-   Source image size
 *	xStart			-   Starting x-offset, used to determine
 *				    the starting column in the dither matrix
 *				    Does NOT affect which pixel is processed.
 *	yStart			-   Starting y-offset, used to determine
 *				    the starting line in the dither matrix
 *				    Does NOT affect which line is processed.
 *	ditherTab		-   Look-up table created by i_makeOrderedRGBToLinearTable
 *
 * Returns:
 *	error status
 ***********************************************************************/

 Int32 CDECL 
i_rescaleRGBToLinear8 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    Int32	wD,
    Int32	hD,
    UInt8      *pRBase,
    UInt8      *pGBase,
    UInt8      *pBBase,
    Int32	sBpl,
    Int32	xS,
    Int32	yS,
    Int32	wS,
    Int32	hS,
    Int32	xStart,
    Int32	yStart,
    void       *ditherTab)
{
    PROLOGUE;
    UInt8   *pRLut, *pGLut, *pBLut, *pR, *pG, *pB;
    UInt8   *ditherRow;
    Int32    ditherVal;
    Int32    row, col, word, colCtr, i;
    UInt32   outWord, clearPadBits;
    UInt32  *pDWord, *xOffsetTab, yOffset, xOffset;
    Float64  sourceX, sourceXInc, sourceY, sourceYInc;

    pRLut = (UInt8 *) ditherTab;
    pGLut = pRLut + cChannelTableSize;
    pBLut = pGLut + cChannelTableSize;

    sourceXInc = (Float64) wS / (Float64) wD;
    sourceYInc = (Float64) hS / (Float64) hD;

    /* Move the pointers to the right place in Y, but keep them at the frame
       in X */
    pD += (dBpl << cLogFrameBits) + cFrameBits;
    pRBase += (sBpl * yS);
    pGBase += (sBpl * yS);
    pBBase += (sBpl * yS);
    clearPadBits = ~((1 << (8 * (4 - (wD & 3)))) - 1);

    /* Create an x-offset table */
    xOffsetTab = (UInt32 *) MALLOC (sizeof(UInt32) * ((wD + 3) & ~0x3));
    if (!xOffsetTab)
	return ia_nomem;

    for (i = 0, sourceX = xS; i < ((wD + 3) & ~0x3); i++)
    {
	xOffset = (UInt32) sourceX;

	/* Handle the byte ordering in the table */
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	
	xOffsetTab[i] = xOffset ^ 3;
#else
	xOffsetTab[i] = xOffset;
#endif
	sourceX += sourceXInc;
    }

    sourceY = 0.0;

    for (row = 0; row < hD; row++)
    {
	/* Get a pointer to the dither row */
	ditherRow = &(scaledDither8[(row + yStart) % cDitherPeriod][0]);

	/*
	 * Get pointers to the source line. In rescaleNoInterp, we
	 * copied duplicate lines, but we don't bother here, since
	 * it will dither differently for each output line.
	 */
	yOffset = sBpl * (Int32) sourceY;
	pR = pRBase + yOffset;
	pG = pGBase + yOffset;
	pB = pBBase + yOffset;
	pDWord = (UInt32 *) pD;
	colCtr = xStart % cDitherPeriod;

	for (col = 0, word = 0; col < wD ; word++, col += 4)
	{
	    /* We should be able to mask the column # once, since we 
	       shouldn't cross a dither matrix boundary */

	    /* First pixel */
	    ditherVal = ditherRow[colCtr++];
	    xOffset = xOffsetTab[col];
	    outWord  = pRLut[((pR[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pGLut[((pG[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pBLut[((pB[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];

	    outWord <<= 8;
	    if (colCtr >= cDitherPeriod)
		colCtr = 0;

	    /* Second  pixel */
	    ditherVal = ditherRow[colCtr++];
	    xOffset = xOffsetTab[col + 1];
	    outWord += pRLut[((pR[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pGLut[((pG[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pBLut[((pB[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];

	    outWord <<= 8;
	    if (colCtr >= cDitherPeriod)
		colCtr = 0;

	    /* Third pixel */
	    ditherVal = ditherRow[colCtr++];
	    xOffset = xOffsetTab[col + 2];
	    outWord += pRLut[((pR[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pGLut[((pG[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pBLut[((pB[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord <<= 8;
	    if (colCtr >= cDitherPeriod)
		colCtr = 0; 

	    /* Fourth pixel */
	    ditherVal = ditherRow[colCtr++];
	    xOffset = xOffsetTab[col + 3];
	    outWord += pRLut[((pR[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pGLut[((pG[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    outWord += pBLut[((pB[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    if (colCtr >= cDitherPeriod)
		colCtr = 0;

	    pDWord[word] = outWord;
	}

	/* Clear the pad, if necessary */
	if (wD & 3)
	{
	    pDWord[word - 1] &= clearPadBits;
	}

	pD += dBpl;
	sourceY += sourceYInc;
    }

    FREE(xOffsetTab);

    NORMAL_RETURN;
}

/*
 * 4 bit ordered dithering to arbitrary color palettes
 * 
 * The need here is to go to some arbitrary color map with a 24 bit/pixel
 * image, and have it not look "too bad", and still run at a reasonable
 * speed. The way we'll do this is as follows: take the upper 3 bits of
 * each pixel, along with 3 bits of dither fraction (from the ordered dither
 * table) yielding a 12-bit index into a table of output pixels. The table
 * is computed by finding the nearest color in the table to the RGB generated
 * by the three bits, rounded using the dither fraction. This rounding is
 * a lot trickier than in the linear case--we have to round differently
 * for every interval in the color space. Deciding how to do this is not
 * straightforward: we're finding points in 3-space, and which point we
 * pick as the point to "interpolate towards" is ambiguous. One somewhat
 * unappealing option is to dither into something close, like a 6 bit linear
 * (2 bits each) palette, and then use that as a secondary look-up. What
 * we're going to attempt instead is to find the two nearest colors in
 * the color space to the color in question, and use the dither value
 * to decide which color to pick by computing the delta-E between the
 * target color and each of the two colors (call them dE1 and dE2), and
 * then compute dE1/(dE1+dE2), and compare it, scaled to the range of
 * the dither "fractions", to the current dither fraction, and picking the
 * right color depending on whether the fraction was > dE1/(dE1+dE2).
 * This should give a reasonably "colorimetric" dithering, though when
 * dithering to VGA palettes, it will always look bad to some degree.
 * When we say "delta-E", we probably mean a color distance in some
 * color space like Lab or Luv, though just a sqrt(r^2+g^2+b^2) distance
 * might be good enough...
 *
 * ----------------------------------------------------------------
 *
 * The above scheme did something, but it was a bit of a disaster.
 * Figuring out which colors to dither between is difficult, and I never
 * really got it to look decent, so I've moved to doing a constant,
 * known thing--dithering to three bits, (i.e. a "linear palette" with
 * two colors per channel) and then looking those bits up in a table
 * to get the final output bits. The advantage of this scheme is that
 * the three planes are dithered separately, rather than trying to get
 * them to dither together. This scheme does require that the final 
 * palette has all the combinations of bits needed, i.e. full-on 
 * black, red, green, blue, cyan, magenta, yellow, and white. This code
 * is probably only going to be used when dithering to a VGA palette,
 * which does fulfill that requirement. If one truly needs to dither
 * to an *arbitrary* palette, one must use error diffusion, because
 * that has a reasonable heuristic for actually keeping track of the error,
 * and one can therefore just find the nearest color, accumulate the error,
 * and be done with it. 
 * 
 */

/*
 * This is a palette decode table for the "standard VGA" palette. 
 * it converts a 1-bit red, 1-bit green, 1-bit blue value (0-7) into
 * the VGA palette index for that color. 
 */
UInt8    VGAPaletteDecode[] =
{	
    0, 12, 10, 14, 9, 13, 11, 15, 0, 0, 0, 0, 0, 0, 0, 0 
};
			
/* PUBLIC */
/***********************************************************************
 *
 * i_rescaleRGBToPalette4
 *
 *	Do an ordered dither to an arbitrary palette. Uses an implicit
 *	dither matrix. Does a non-interpolating (i.e. nearest neighbor
 *	resampling.
 *
 *	This uses floating point to compute the resampling table.
 *	We expect that machines which can't have floating point won't
 *	be doing any color image processing anyway. We don't use
 *	floating point for speed-critical parts, so it shouldn't matter
 *	if it's emulated.
 *
 *	In our initial version, we tried finding pairs of 'nearest'
 *	colors in RGB space and dithering between them. It was a 
 *	disaster. What we do instead is to dither each color separately
 *	into a "linear" 1 bit/pixel palette, and then look the result
 *	up in a table at the end. 
 *  
 * Arguments:
 *	pDest			-   Result image, .r to frame
 *	dBpl			-   Bytes between lines in dest image
 *	wD, hD			-   Width, height of dest. image
 *	pRed, pGreen, pBlue	-   Pointers to source color plane, .r to frame
 *	sBpl			-   Bytes between lines in color planes
 *	xS, yS			-   Offset into source image (from frame)
 *	wS, hS			-   Source image size
 *	xStart			-   Starting x-offset, used to determine
 *				    the starting column in the dither matrix
 *				    Does NOT affect which pixel is processed.
 *	yStart			-   Starting y-offset, used to determine
 *				    the starting line in the dither matrix
 *				    Does NOT affect which line is processed.
 *	ditherTab		-   Look-up table created by i_makeOrderedRGBToLinearable
 *	paletteDecode		-   LUT to de-scramble linear palette into
 *				    the destination palette
 *				
 *
 * Returns:
 *	error status
 ***********************************************************************/

 Int32 CDECL 
i_rescaleRGBToPalette4 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    Int32	wD,
    Int32	hD,
    UInt8      *pRBase,
    UInt8      *pGBase,
    UInt8      *pBBase,
    Int32	sBpl,
    Int32	xS,
    Int32	yS,
    Int32	wS,
    Int32	hS,
    Int32	xStart,
    Int32	yStart,
    void       *ditherTab,
    UInt8      *pPaletteDecode)
{
    PROLOGUE;
    UInt8   *pR, *pG, *pB, *ditherRow, *pRLut, *pGLut, *pBLut;
    Int32    row, col, word, i, ditherVal, ditherCol;
    UInt32   outWord, val;
    UInt32  *pDWord, *xOffsetTab, yOffset, xOffset;
    Float64  sourceX, sourceXInc, sourceY, sourceYInc;

    pRLut = (UInt8 *) ditherTab;
    pGLut = pRLut + cChannelTableSize;
    pBLut = pGLut + cChannelTableSize;

    sourceXInc = (Float64) wS / (Float64) wD;
    sourceYInc = (Float64) hS / (Float64) hD;

    /* Move the pointers to the right place in Y, but keep them at the frame
       in X */
    pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;
    pRBase += (sBpl * yS);
    pGBase += (sBpl * yS);
    pBBase += (sBpl * yS);

    /* Create an x-offset table */
    xOffsetTab = (UInt32 *) MALLOC (sizeof(UInt32) * ((wD + 7) & ~0x7));
    if (!xOffsetTab)
	return ia_nomem;

    for (i = 0, sourceX = xS; i < ((wD + 7) & ~0x7); i++)
    {
	xOffset = (UInt32) sourceX;

	/* Handle the byte ordering in the table */
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	
	xOffsetTab[i] = xOffset ^ 3;
#else
	xOffsetTab[i] = xOffset;
#endif
	sourceX += sourceXInc;
    }

    sourceY = 0.0;

    for (row = 0; row < hD; row++)
    {
	/* Get a pointer to the dither row */
	ditherRow = &(scaledDither8[(row + yStart) % cDitherPeriod][0]);

	/*
	 * Get pointers to the source line. In rescaleNoInterp, we
	 * copied duplicate lines, but we don't bother here, since
	 * it will dither differently for each output line.
	 */
	yOffset = sBpl * (Int32) sourceY;
	pR = pRBase + yOffset;
	pG = pGBase + yOffset;
	pB = pBBase + yOffset;
	pDWord = (UInt32 *) pD;
	outWord = col = word = 0;
	ditherCol = (col + xStart) % cDitherPeriod;

	while (col < wD)
	{
	    xOffset = xOffsetTab[col];
	    ditherVal = ditherRow[ditherCol];
	    val  = pRLut[((pR[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    val += pGLut[((pG[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];
	    val += pBLut[((pB[xOffset] << cPixIntShift) & cPixIntMask) | ditherVal];

	    outWord = (outWord << 4) | pPaletteDecode[val];
	    col++;
	    ditherCol++;
	    if (ditherCol >= cDitherPeriod)
		ditherCol = 0;

	    /* If we've filled up a word, output it */
	    if ((col & 7) == 0)
	    {
		pDWord[word++] = outWord;
		outWord = 0;
	    }
	}

	/* Don't forget about that last word */
	if (wD & 7)
	{
	    pDWord[word] = outWord << ((8 - (col & 7)) * 4);
	}

	pD += dBpl;
	sourceY += sourceYInc;
    }

    FREE(xOffsetTab);

    NORMAL_RETURN;
}

/*
$Log: i_cdith.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.13  1998/01/13 23:47:20  ddavies
i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.12  1997/05/02 01:18:19  ddavies
ddavies on Thu May 1 18:17:52 PDT 1997

i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v

Remove file header comment about i_sumChannelToLinear8.

Revision 1.11  1995/12/13 02:42:46  ejaquith
i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.10  1995/10/18  03:52:16  ddavies
 * i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.9  1995/10/13  23:49:46  dferrell
 * i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.8  1995/02/14  02:59:35  macleod
 * i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v
 *
 * Put in a cast to get around an annoying acc warning about the ANSI
 * standard that changed years ago
 *
 * Revision 1.7  1995/02/10  22:47:12  macleod
 * macleod on Fri Feb 10 14:46:17 PST 1995
 *
 * i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v
 *
 * Reverted to 16x16 dither matrix rather than 100x100 matrix, due to
 * complaints about random vs. fixed-pattern noise.
 *
 * Revision 1.6  1995/02/07  21:36:05  macleod
 * macleod on Tue Feb 7 13:35:14 PST 1995
 *
 * i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v
 *
 * Moved the declaration of a static table into here. This table gets
 * re-initialized (with the same data) on every call to make a dithering
 * table, which is stupid.
 *
 * Revision 1.5  1995/02/07  21:26:30  macleod
 * macleod on Tue Feb 7 13:25:49 PST 1995
 *
 * i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v
 *
 * Fixed table creation code so that dithering (255,255,255) always results
 * in white in the output.
 *
 * Revision 1.4  1995/02/01  23:14:15  macleod
 * macleod on Wed Feb 1 15:14:14 PST 1995
 *
 * i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v
 *
 * Added 4-bit ordered rescale/dither. The first attempt tried to find pairs of
 * "close" colors and dither between them, but the right color to use is kind
 * of ambiguous, and it ended up looking bad for a lot of effort.  What I ended
 * up doing is to dither each channel separately into a 1-bit/pixel palette,
 * and then convert that into the output palette with a look-up table at the
 * end. This assumes that one's target 4-bit palette does have all of the
 * combinations of 255 and 0 (i.e. black, r, g, b, c, m, y, white). This is the
 * case with the VGA palette. In order to use this routine, first call the
 * table creation routine with 2 reds, 2 greens, and 2 blues.
 * Also
 * - Added "xStart" parameter, which changes the starting column index into
 *   the dither matrix like the "yStart" parameter does for the row.
 * - Removed i_sumChannelToLinear8, since no one was using it.
 *
 * Revision 1.3  1995/01/24  01:43:51  macleod
 * macleod on Mon Jan 23 17:42:12 PST 1995
 *
 * i_cdith.c -> /product/ipcore/ipshared/src/RCS/i_cdith.c,v
 *
 * Contains two new routines:
 * i_sumChannelToLinear8 allows one to do ordered color dither one
 * channel at a time. It's slower, but was requested for low-memory
 * situations.
 * i_rescaleRGBToLinear8 does a combination non-interpolated rescale
 * and ordered dither of an RGB image.
 *
*/
