/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 *	i_cfloyd.c: Color Floyd-Steinberg dither to a color map
 *
 *	First hack: just edit down cfloyd.c
 *
 *	Functions:
 *	   
 *		    Int32     i_ditherFloydRGBToCMap()
 *		    Int32     i_ditherFloydRGBTo8 ()
 */



#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"

    /*	prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrrast.pub"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_cfloyd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

    /*	Constants */
#define cPlanes	    3	    /* Always do three planes */
#define cRed	    0
#define cGreen	    1
#define cBlue	    2

    /* These constants are used for the color LUT stuff */
#define cColorLutSize		4096
#define cColorLutRedMask	0xf
#define cColorLutGreenMask	0xf
#define cColorLutBlueMask	0xf
#define cColorLutRedShift	8
#define cColorLutGreenShift	4
#define cColorLutBlueShift	0
#define cColorLutRedBits	4
#define cColorLutGreenBits	4
#define cColorLutBlueBits	4

/* static variables */
static Int16 zero[] = { 0 , 0 };
			
/*
 * What we are doing is to take an RGB image (actually, three
 * 8-bit planes, since that's what everything seems to deal with)
 * and dithers it down to a color-mapped 8 bit image. The color
 * map can have up to 256 entries, but no more, since we'd have to
 * have more than one output routine in that case, and it probably
 * wouldn't be very useful anyway.
 *
 * Sandpiper has a function called "steinberg" (as opposed to "floyd")
 * which does approximately this. It doesn't appear to be too efficient,
 * since it finds the nearest (distance in color space) color in the color
 * table for a given output by brute force searching of the color table.
 *
 * There are several methods that we're going to use to speed things 
 * up.
 *
 * First, we can get the color search time down from linear search to 
 * a single look-up by encoding the (R,G,B) into 12 bits, and looking it
 * up in a"nearest color" look-up table. If we use 4 bits each,
 * we'll get 16 levels per color, which isn't great, but is probably better
 * than the accuracy of the color table anyway. In any case, the actual
 * error caclulation happens in 24-bit RGB, since all we get from this
 * table is the closest color. We could speed things up quite a bit more
 * by sticking a computed error into that table, but with only 4 bits
 * per color, it would probably be too inaccurate. More likely, we'll have
 * to compute the error on each component individually. We can either store
 * the color itself in the table, which makes it 3 times larger, or
 * we can store the color table index, which causes an extra lookup per
 * pixel.  
 *
 * Once we've looked up the (RGB) for the mapped color we found, we
 * have to compute and propagate the errors for the three channels.
 * Obviously, the trick is to make this speedy. In the gray-to-binary
 * case, we can get awawy with keeping more stuff in registers, i.e. 
 * it can keep 32 output pixels in one register, whereas we can keep
 * only four output pixels in one register. The modified pixels (e.g.
 * pixels plus error, which we shouldn't just write back) can only be
 * packed one per register, since they're RGB.
 * 
 * In this first version, we're going to do one line at a time, and then
 * later we can see whether we can do more than that.
 *
 * 9/17/93: add a parameter to control the # of output bits. For now,
 * it will support 4 and 8.
 */

/* PUBLIC */
/*
 * i_ditherFloydRGBToCMap ()
 *
 *      dither an RGB image down to an 8-bit color-mapped image.
 *	In this version, the color map can be arbitrary
 *      
 *
 *      Input:
 *              pR = pointer to Red plane of source image
 *              pG = pointer to Green plane of source image
 *              pB = pointer to Blue plane of source image
 *              sBpl = bytes per line for each source plane
 *              x = x coordinate of the top left corner of the
 *		    source rectangle relative to OUTSIDE of frame.
 *              y = y coordinate of the top left corner of the
 *		    source rectangle relative to OUTSIDE of frame.
 *              w = width of the rectangle
 *              h = height of the rectangle
 *		pCMap = pointer to color map, organized as bytes of R,G,B,X
 *		colors = number of colors in color map
 *		pCLut = pointer to color LUT, created from color map
 *			by i_cMapToLut
 *              pRedLUT = Look-up table for red channel
 *              pGreenLUT = Look-up table for green channel
 *              pBlueLUT = Look-up table for blue channel
 *              outputOffset = offset to add to output values
 *              pD = pointer to destination image (including frame).
 *		     Output starts at first non-frame pixel.
 *              dBpl = destination bytes per line
 *		dBits = Bits/destination pixel
 *     
 *      Return Code:
 *              ia_successful   good completion
 *              ia_nomem        unable to allocate memory
 *
 *      Assumes that frame has already been placed in destination area.
 */

Int32 CDECL 
i_ditherFloydRGBToCMap (
    IPState    *pIPState,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    UInt32	sBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8      *pCMap,
    UInt16	colors,
    UInt8      *pCLut,
    UInt8      *pRedLut,
    UInt8      *pGreenLut,
    UInt8      *pBlueLut,
    UInt8	outputOffset,
    UInt8      *pD,
    UInt32	dBpl,
    UInt32	dBits)
{
    PROLOGUE;
    ADitherState	*state;

    ExceptRaise (dBits != 8 && dBits != 4, ia_depthNotSupported);
    /* Initialize tables and error buffer */
    ERR_PASS(i_ditherFloydRGBToCMap_Init (pIPState,
					  pRedLut,
					  pGreenLut,
					  pBlueLut,
					  pCMap,
					  colors,
					  pCLut,
					  outputOffset,
					  w,
					  (UInt8)dBits,
					  &state));

    ERR_PASS( i_ditherFloydRGBToCMap_Process (pIPState, state,
					      pR, pG, pB, sBpl, 
					      pD, dBpl,
					      x, y, w, h));
    ERR_PASS( i_ditherFloydRGBToCMap_Finish (pIPState, state));

    NORMAL_RETURN;
}

 /*
 * i_ditherFloydRGBToCMap_Init ()
 *
 *	When processing bands, this is the public routine for initialization
 *	for i_ditherFloydRGBToCMap()
 *
 *	Input:
 *		sBpl = bytes per line for each source plane
 *		dBpl = destination bytes per line
 *		pRedLUT = Look-up table for red channel
 *		pGreenLUT = Look-up table for green channel
 *		pBlueLUT = Look-up table for blue channel
 *		pCMap = pointer to color map, organized as bytes of R,G,B,X
 *		colors = number of colors in color map
 *		outputOffset = offset to add to output values
 *		lineLength = # of pixels in a line
 *		dBits = Bits per pixel of output
 *		state = Place to stuff state structure
 *     
 *	Return Code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes that frame has already been placed in destination area.
 */

Int32 CDECL 
i_ditherFloydRGBToCMap_Init (
    IPState    *pIPState,
    UInt8      *pRedLut,
    UInt8      *pGreenLut,
    UInt8      *pBlueLut,
    UInt8      *pCMap,
    UInt16	colors,
    UInt8      *pCLut,
    UInt8	outputOffset,
    UInt32	lineLength,
    UInt8	dBits,
    ADitherState **st)
{
    PROLOGUE;
    Int16 *err_p;
    Int16 i, tmp;
    Int32 err_val, b_err, c_err, d_err;
    Float64 b_frac, c_frac, d_frac;
    ADitherState *state;

    ExceptRaise (dBits != 8 && dBits != 4, ia_depthNotSupported);
    ExceptRaise (!(*st = (ADitherState *) CALLOC (1, sizeof (ADitherState))), ia_nomem);
    state = (ADitherState *) *st;

    /* Fill the clip lut */
    ExceptRaise (!(state->clipLut = (UInt8 *) MALLOC (1024)), ia_nomem);
    for (i = 0; i < 1024; i++)
    {
	tmp = i - 512;
	if (tmp < 0) tmp = 0;
	if (tmp > 255) tmp = 255;
	state->clipLut[i] = (UInt8)tmp;
    }

    /* Allocate and fill the error lut */
    ExceptRaise (!(state->errorLut = (Int16 *) MALLOC (512 * 4 * sizeof (Int16))), ia_nomem);

    b_frac = 1.0 / 16.0;
    c_frac = 5.0 / 16.0;
    d_frac = 3.0 / 16.0;

    for (i = 0, err_p = state->errorLut; i < 512; ++i)
    {
	err_val = i - 256;

	d_err = *err_p++ = (Int16)(err_val * d_frac + 0.5);  /* D error */
	c_err = *err_p++ = (Int16)(err_val * c_frac + 0.5);  /* C error */
	b_err = (Int16)(err_val * b_frac + 0.5);	      /* B error */
	*err_p++ = (Int16)(err_val - d_err - c_err - b_err);   /* A error */
	*err_p++ = (Int16)b_err;
    }

    /* Allocate the error buffer used to hold the error for the next line */
    /* Calloc is supposed to always zero the allocated memory */
    ExceptRaise (!(state->errorBufI = (Int16 *) CALLOC (lineLength + 2, sizeof(Int16) * 3)), ia_nomem);

    /* Set up the rest of the state variables */
    state->redLut = pRedLut;
    state->greenLut = pGreenLut;
    state->blueLut = pBlueLut;
    state->colors = colors;
    state->pCLut = pCLut;
    state->pCMap = pCMap;
    state->dBits = dBits;
    state->outputOffset = outputOffset;

    NORMAL_RETURN;
}

/* PUBLIC */
/*
// i_cMapToLut
//
//	Crank a color LUT from a color map. The color distance 
//	will be computed in RGB for this version, which is perhaps
//	not the correct space. One might want to make tradeoffs for
//	better luminance, chrominance, saturation, etc.
//
//	The LUT is computed using compile-time constants. If there
//	is a need, we can always make them parameters...
*/

Int32 CDECL 
i_cMapToLut (
    IPState    *pIPState,
    UInt8      *pCMap,				/* Input color map */
    UInt16	colors,				/* # of colors in map */
    UInt8      *colorLut)			/* n:n:n LUT */
{
    PROLOGUE;
    register UInt8 r, g, b, mapIndex, rInc, gInc, bInc;
    register UInt16 lutIndex, k;
    UInt16 i, j;

    /* Fill in the table */
    r = 0; g = 0; b = 0;
    rInc = 1 << cColorLutRedBits;
    gInc = 1 << cColorLutGreenBits;
    bInc = 1 << cColorLutBlueBits;

    for (i = 0; i < rInc; i++, r += rInc)
    {
	for (j = 0; j < gInc; j++, g += gInc)
	{
	    for (k = 0; k < bInc; k++, b += bInc)
	    {
		/* Find the nearest color to this (R, G, B) */
		lutIndex = (i << cColorLutRedShift) |
			   (j << cColorLutGreenShift) |
			   (k << cColorLutBlueShift);
		mapIndex = (UInt8)ip_findNearestColor (pIPState,
						       r, g, b, pCMap, colors);
		colorLut[lutIndex] = mapIndex;
	    }
	}
    }
    NORMAL_RETURN;
}

/* PUBLIC */
/*
// i_cMapToLut1
//
//	Mapping priciple is same as i_cMapToLut. But the color samples'
//	value used in i_cMapToLut are 0x0, 0x10, 0x20, .... 0xf0. It
//	covers [0, 240]. This will cause problem in white point. Here
//	the value 0x0, 0x11, 0x22, ..... 0xff are used. It covers
//      [0, 255]  and enhance white and black point.
//      
*/

Int32 CDECL 
i_cMapToLut1 (
    IPState    *pIPState,
    UInt8      *pCMap,				/* Input color map */
    UInt16	colors,				/* # of colors in map */
    UInt8      *colorLut)			/* n:n:n LUT */
{
    PROLOGUE;
    register UInt8 r, g, b, mapIndex, rInc, gInc, bInc;
    register UInt8 rInc1, gInc1, bInc1;
    register UInt16 lutIndex, k;
    UInt16 i, j;

    /* Fill in the table */
    rInc = 1 << cColorLutRedBits;
    gInc = 1 << cColorLutGreenBits;
    bInc = 1 << cColorLutBlueBits;
    rInc1 = rInc | 1;
    gInc1 = gInc | 1;
    bInc1 = bInc | 1;

    for (i = 0, r =0; i < rInc; i++, r += rInc1)
    {
	for (j = 0, g =0; j < gInc; j++, g += gInc1)
	{
	    for (k = 0, b =0; k < bInc; k++, b += bInc1)
	    {
		/* Find the nearest color to this (R, G, B) */
		lutIndex = (i << cColorLutRedShift) |
			   (j << cColorLutGreenShift) |
			   (k << cColorLutBlueShift);
		mapIndex = (UInt8)ip_findNearestColor (pIPState,
						       r, g, b, pCMap, colors);
		colorLut[lutIndex] = mapIndex;
	    }
	}
    }
    NORMAL_RETURN;
}		
		
/* PRIVATE */
/* 
// ip_findNearestColor
//
//	Given an (r, g, b) triple, locate the nearest color in the 
//	color LUT. The nearest color calcuation is currently done
//	in RGB space.
//
*/

Int16 CDECL 
ip_findNearestColor (
    IPState    *pIPState,
    UInt8	r,
    UInt8	g,
    UInt8	b,
    UInt8      *pCMap,
    UInt16	colors)
{
    PROLOGUE;
    Int32	minDelta, delta;
    Int16	d1, d2, d3;
    Int16	i, index;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Always do better than this */
    minDelta = 0x7fffffff;

    index = 0;

    for (i = 0; i < (Int16) colors; i++, pCMap += 4)
    {
	d1 = (Int16) r - (Int16) pCMap[0];
	d2 = (Int16) g - (Int16) pCMap[1];
	d3 = (Int16) b - (Int16) pCMap[2];
	/* Well, delta^2 actually, but the minimum is the same */
	delta = d1 * d1 + d2 * d2 + d3 * d3;
	if (delta < minDelta)
	{
	    minDelta = delta;
	    index = i;
	}
    }

   return index;
}

/* PUBLIC */
/*
 * i_ditherFloydRGBToCMap_Process
 *
 *	Given that we've initialized, process some data. The salient
 *	feature is that bands can be processed sequentially, and the
 *	error terms will be propagated between them correctly. 
 *	Note that the caller keeps track of the rectangles being
 *	processed, which is, perhaps, dumb, but easier than having
 *	another call to tell the caller how many lines it's going to 
 *	put out this time, so that the caller can allocate an output buffer
 *	of the right size. (Yes, it could just allocate a maximum, but, yuck.)
 *	
 *	Input:
 *		state = ADitherState returned from ..._Init
 *              pR = pointer to Red plane of source image
 *              pG = pointer to Green plane of source image
 *              pB = pointer to Blue plane of source image
 *		pD = pointer to destination image
 *		     Output starts at first non-frame pixel.
 *              x = x coordinate of the top left corner of the rectangle
 *		    in source relative to OUTSIDE of frame.
 *              y = y coordinate of the top left corner of the rectangle
 *		    in source relative to OUTSIDE of frame.
 *              w = width of the rectangle
 *              h = height of the rectangle
 *
 *	Return Code:
 *		ia_successful	good completion
 */

Int32 CDECL 
i_ditherFloydRGBToCMap_Process (
    IPState    *pIPState,
    ADitherState *state,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    UInt32	sBpl,
    UInt8      *pD,
    UInt32	dBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h)
{
    PROLOGUE;
    register Int16 *pErr, *errorP, *errorLut, i;
    register UInt8 *pClipLut, *mapP, outPixel, *dest;
    register Int16 r, g, b, *rLast, *gLast, *bLast;
    register Int16 mapIndex;
    Int16 zero[2], line;
    zero[0] = 0; zero[1] = 0;

    /* JAC note added 4/17/97 just in case I have to make a bug fix to
     * this code again (and I don't want to figure out the variables
     * again)...
     * there are 6 main structures used in this code. 
     * 1) color map supplied by the calling routine.  the error to be
     * diffused in each plane is the difference between the actual pixel
     * on that plane and the color map's pixel value.
     * 2) a LUT that translates from the pixels in the image (well the
     * 4 MSB of each) to the index of the nearest color in the color map.
     * 3) a TRC for each plane.
     * 4) an error buffer.  this buffer holds the error to be diffused 
     * to the next line from the pixels in the current line.  it is
     * arranged as a single buffer with RGB values for each pixel.  thus
     * you find RGB RGB RGB ....
     * 5) an clip LUT.  this is used to clip each pixel back down to 8 bits.
     * 6) an error LUT.  this table is indexed by amount of error to be
     * diffused to the surrounding pixels.  since the error may be 
     * positive or negative, the pointer to this table is actually in the
     * middle of the table so that we can index it by any number in the
     * range [-255, +255].  we diffuse error as 	  * A
     *							D C B	where
     * * is the current pixel.  each entry in the table then has
     * entries arranged as D error, C error, A error, B error.  this 
     * leads to the last kinda confusing part.  when we are computing
     * the value of the current pixel we add in the A error from the
     * previous.  this is pointed to by rLast, bLast, and gLast.   after
     * using those pointers they are incremented to point to the B error.
     * this is used when calculating the error (to be placed in the error
     * buffer) for the pixel in the line below the current one.
     */

    /* Point in the middle so we can use negative indices */
    errorLut = state->errorLut + 256 * 4;
    pClipLut = state->clipLut + 512;
    /* since the first pixel on the line has no error from the previous
     * pixel, point to a location containing 0.
     */
    rLast = gLast = bLast = &(zero[0]);
    mapP = state->pCMap;
    outPixel = 0;

    /* Move the pointers to the right point */
    pR += (y * sBpl) + x;
    pG += (y * sBpl) + x;
    pB += (y * sBpl) + x;
    pD += (dBpl << cLogFrameBits) + ((state->dBits == 4) ? cFrameBits >> 1 : cFrameBits);

    for (line = 0; line < h; line++)
    {
	dest = pD;
	/* skip over 1 RGB pixel - we diffuse error below and to the left
	 * so we have to allow room for this to happen.
	 */
	pErr = state->errorBufI + 3;

	for (i = 0; i < w; i++, pErr += 3)
	{
	    /* Get the pixels and add in their errors. */
	    r = pClipLut[state->redLut  [UCHAR_ACCESS(pR + i)] + pErr[0] + *rLast++];
	    g = pClipLut[state->greenLut[UCHAR_ACCESS(pG + i)] + pErr[1] + *gLast++];
	    b = pClipLut[state->blueLut [UCHAR_ACCESS(pB + i)] + pErr[2] + *bLast++];
	    
	    mapIndex = state->pCLut[((r & 0x0f0) << 4) |
				    (g & 0x0f0) |
				    ((b & 0x0f0) >> 4)];

	    /* Output that pixel */
	    if (state->dBits == 4)
	    {
		if ((i & 1) == 1)
		{
		    UCHAR_ACCESS(dest++) = outPixel | (mapIndex + state->outputOffset);
		}
		else
		{
		    outPixel = (mapIndex + state->outputOffset) << 4;
		}
	    }
	    else
	    {
		UCHAR_ACCESS(dest + i) = mapIndex + state->outputOffset;
	    }

	    /* Compute the error. [next, try storing the error in the LUT] */
	    mapIndex *= 4;	/* each entry is 4 bytes */
	    /* mapP = &pCMap[mapIndex * 4]; */

	    /* Propagate the errors for the channels */
	    /* Following is not *real* efficient on the SPARC */
	    /* since it uses constant offsets to registers, which is */
	    /* reasonably efficient on 486 and 680x0s. */
	    /* But since we expect to code the sparc version in assembly, */
	    /* that's life. And since the sparc has fast shifts and */
	    /* all those registers, we'll probably keep the errors in */
	    /* registers and write them a word at a time */
	    /* The usual picture: 

		- * A
		D C B
	    
		where * is the pixel we're on
	    */

	    /* 4 => each entry is 4 Int16's */
	    errorP = errorLut + ((r - mapP[mapIndex++]) * 4);

	    /* the error buffer is RGB RGB RGB so if the current pixel is:
	     *                         |
	     * then getting to the previous red is -3.
	     */
	    pErr[-3] += *errorP++;		    /* red D */
	    pErr[ 0]  = *errorP++ + *rLast;	    /* red C + last B */
	    rLast     =  errorP;		    /* red A, B */

	    errorP = errorLut + ((g - mapP[mapIndex++]) * 4);

	    pErr[-2] += *errorP++;		    /* green D */
	    pErr[ 1]  = *errorP++ + *gLast;	    /* green C + last B */
	    gLast     =  errorP;		    /* green A, B */

	    errorP = errorLut + ((b - mapP[mapIndex  ]) * 4);

	    pErr[-1] += *errorP++;		    /* blue D */
	    pErr[ 2]  = *errorP++ + *bLast;	    /* blue C + last B */
	    bLast     =  errorP;		    /* blue A, B */

	}

	if (state->dBits == 4 && ((i & 1) == 1))
	{
	    /* Don't forget about that last pixel. */
	    UCHAR_ACCESS(dest) = outPixel;
	}

	/* Move to the next line */
	pR += sBpl;
	pG += sBpl;
	pB += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;;
}
/*
 * i_ditherFloydRGBToCMap_Finish 
 *
 *	We're done; clean up. Warning: state will be freed, so don't
 *	pass it again.
 *
 *	Input:
 *		state = ADitherState returned from ..._Init
 *	
 *	Return Code:
 *		ia_successful	good completion
 */

Int32 CDECL 
i_ditherFloydRGBToCMap_Finish (
    IPState    *pIPState,
    ADitherState *state)
{
    PROLOGUE;
    ExceptRaise (!state, ia_invalidParm);

    if (state->clipLut)
    {
	FREE (state->clipLut);
	state->clipLut = 0;
    }

    if (state->errorBufI)
    {
	FREE (state->errorBufI);
	state->errorBufI = 0;
    }
    
    if (state->errorLut)
    {
	FREE (state->errorLut);
	state->errorLut = 0;
    }

    FREE (state);

    NORMAL_RETURN;
}

/* PUBLIC */
/*
 * i_ditherFloydRGBTo8 ()
 *
 *	Dither an RGB image down to an 8-bit color-mapped image.
 *	There are (at least) two different routines which do something
 *	like this in the "sandpiper" library. One does what this does,
 *	which dithers to a linear color space made out of the
 *	three color counts passed. Another, called "steinberg", takes
 *	any color table and dithers to it (slowly).
 *	
 *	The number of colors output = redCount x GreenCount x BlueCount.
 *	The product of these three counts should not be > 256.
 *
 *	This is a hack-down of "cfloyd.c" from sandpiper.
 *
 *	Input:
 *		pR = pointer to Red plane of source image
 *		pG = pointer to Green plane of source image
 *		pB = pointer to Blue plane of source image
 *		sBpl = bytes per line for each source plane
 *		x = x coordinate of the top left corner of the
 *		    source rectangle relative to OUTSIDE of frame.
 *		y = y coordinate of the top left corner of the
 *		    source rectangle relative to OUTSIDE of frame.
 *		w = width of the rectangle
 *		h = height of the rectangle
 *		pRedLUT = Look-up table for red channel
 *		pGreenLUT = Look-up table for green channel
 *		pBlueLUT = Look-up table for blue channel
 *		redCount = number of reds in color table
 *		greenCount = number of greens in color table
 *		blueCount = number of blues in color table
 *		outputOffset = offset to add to output values
 *		pD = pointer to destination image (including frame).
 *		     Output image starts at first non-frame pixel.
 *		dBpl = destination bytes per line
 *     
 *	Return Code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes that frame has already been placed in destination area.
 */

Int32 CDECL 
i_ditherFloydRGBTo8 (
    IPState    *pIPState,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    UInt32	sBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8      *pRedLut,
    UInt8      *pGreenLut,
    UInt8      *pBlueLut,
    UInt8	redCount,
    UInt8	greenCount,
    UInt8	blueCount,
    UInt8	outputOffset,
    UInt8      *pD,
    UInt32	dBpl)
{
    PROLOGUE;
    ADitherState	*state;

    /* Initialize tables and error buffer */
    ERR_PASS(i_ditherFloydRGBTo8_Init (pIPState,
				       pRedLut,
				       pGreenLut,
				       pBlueLut,
				       redCount,
				       greenCount,
				       blueCount,
				       outputOffset,
				       w,
				       &state));
    ERR_PASS( i_ditherFloydRGBTo8_Process (pIPState, state, pR, pG, pB, sBpl,
					   pD, dBpl, x, y, w, h));
    ERR_PASS( i_ditherFloydRGBTo8_Finish (pIPState, state));

    NORMAL_RETURN;
}


/*
 * i_ditherFloydRGBTo8_Init ()
 *
 *	When processing bands, this is the public routine for initialization
 *	for i_ditherFloydRGBTo8()
 *
 *	Input:
 *		pRedLUT = Look-up table for red channel
 *		pGreenLUT = Look-up table for green channel
 *		pBlueLUT = Look-up table for blue channel
 *		redCount = number of reds in color table
 *		greenCount = number of greens in color table
 *		blueCount = number of blues in color table
 *		outputOffset = offset to add to output values
 *		lineLength = length of a destination line
 *
 *	Return Code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes that frame has already been placed in destination area.
 */

Int32 CDECL 
i_ditherFloydRGBTo8_Init (
    IPState    *pIPState,
    UInt8      *pRedLut,
    UInt8      *pGreenLut,
    UInt8      *pBlueLut,
    UInt8	redCount,
    UInt8	greenCount,
    UInt8	blueCount,
    UInt8	outputOffset,
    UInt32	lineLength,
    ADitherState **st)
{
    PROLOGUE;
    Int16 levels[3];
    Int16 mult[3];
    Int32 i, k;
    Int16 *err_p;
    Int32 b_err, c_err, d_err, err_val;
    Float64 frac, b_frac, c_frac, d_frac;
    ADitherState *state;

    ExceptRaise (!(*st = (ADitherState *) MALLOC (sizeof (ADitherState))), ia_nomem);
    state = (ADitherState *) *st;
    
    /* initialize the table pointers so that we don't free trash on error */
    for (i = 0; i < cPlanes; i++)
    {
	state->outputLevels[i] = 0;
	state->closestLevels[i] = 0;
    }

    levels[0]  = redCount;
    levels[1]  = greenCount;
    levels[2]  = blueCount;

    mult[0] = greenCount*blueCount;
    mult[1] = blueCount;
    mult[2] = 1;

    /* set up look up tables */
    /* set pixels per gap between levels */
    for (k = 0; k < cPlanes; k++)
    {
	frac = 255.0 / (Float64) ((Int16) levels[k] - 1);

	/* Allocate space for the output and closest tables */
	ExceptRaise (!(state->outputLevels[k] = (UInt8 *) MALLOC (512 * sizeof (UInt8))), ia_nomem);
	ExceptRaise (!(state->closestLevels[k] = (Int16 *) MALLOC (512 * sizeof (Int16))), ia_nomem);

	/* set outputLevels to integers between 0 and levels - 1 */
	for (i = 0; i < 256; i++)
	{
	    state->outputLevels[k][i+128] = (UInt8) (i / frac + 0.5);
	}

	/* set closestLevels to contain the levels spread out over 0 - 255 */
	for (i = 0; i < 256; i++)
	{
	    /* mult the levels by # of pixels between levels to spread out */
	    state->closestLevels[k][i+128] =
		(Int16)(((Float64)(state->outputLevels[k][i+128]))*frac + 0.5);
	}

	/* scale up levels to correspond to color LUT base */
	for (i = 0; i < 256; i++)
	{
	    state->outputLevels[k][i+128] *= mult[k];
	}
    }

    /* Allocate the error table, which is 256 x 4 x 2 */
    ExceptRaise (!(state->errorLut = (Int16 *) MALLOC (1024 * sizeof(Int16))), ia_nomem);

    /* compute error table */
    b_frac = 1.0 / 16.0;
    c_frac = 5.0 / 16.0;
    d_frac = 3.0 / 16.0;

    for (i = 0, err_p = state->errorLut; i < 256; ++i)
    {
	/* error_table[i] will be the errors for i - 127, since 127 is our
	   maximimum error */
	/* Note that the order in memory is D,C,A,B */
	err_val = i - 127;
	d_err = *err_p++ = (Int16)(err_val * d_frac + 0.5);  /* D error */
	c_err = *err_p++ = (Int16)(err_val * c_frac + 0.5);  /* C error */
	b_err =            (Int16)(err_val * b_frac + 0.5);  /* B error */
	*err_p++ = (Int16)(err_val = err_val - d_err - c_err - b_err); /* A error */
	*err_p++ = (Int16)b_err;				      /* Store B error */
    }

    for (k = 0; k < cPlanes; k++)
    {
	ExceptRaise (!(state->errorBuf[k] = (Int16 *) CALLOC ((lineLength + 2), sizeof (Int16))), ia_nomem);

	++state->errorBuf[k];	/* so we can do errorBuf[k][-1] in do_cfloyd */

	/* pad out_levels and closest_levels */
	/* this builds the "(desired_value - closest_levels[thresh_desired] + 127) * 4"
	   calculations that used to be within the main loop of do_cfloyd into the
	   closest_levels[] table
	*/

	for (i=0; i <128; i++)
	{
	    state->outputLevels[k][i] = state->outputLevels[k][128];
	    state->outputLevels[k][i+384] = state->outputLevels[k][383];
	}
	for (i=0; i <128; i++)
	{
	    state->closestLevels[k][i] = (i - 1  - state->closestLevels[k][128]) * 4;
	}
	for (i=384; i <512; i++)
	{
	    state->closestLevels[k][i] = (i - 1  - state->closestLevels[k][383]) * 4;
	}
	for (i=128; i <384; i++)
	{
	    state->closestLevels[k][i] = (i - 1  - state->closestLevels[k][ i ]) * 4;
	}
    }

    /* Remember a few things */
    state->redLut = pRedLut;
    state->greenLut = pGreenLut;
    state->blueLut = pBlueLut;
    state->redCount = redCount;
    state->greenCount = greenCount;
    state->blueCount = blueCount;
    state->outputOffset = outputOffset;

    NORMAL_RETURN;
}


/*
 * i_ditherFloydRGBTo8_Process
 *
 *	Given that we've initialized, process some data. The salient
 *	feature is that bands can be processed sequentially, and the
 *	error terms will be propagated between them correctly. 
 *	Note that the caller keeps track of the rectangles being
 *	processed, which is, perhaps, dumb, but easier than having
 *	another call to tell the caller how many lines it's going to 
 *	put out this time, so that the caller can allocate an output buffer
 *	of the right size. (Yes, it could just allocate a maximum, but, yuck.)
 *	
 *	Input:
 *		state = ADitherState returned from ..._Init
 *              pR = pointer to Red plane of source image
 *              pG = pointer to Green plane of source image
 *              pB = pointer to Blue plane of source image
 *		pD = pointer to destination image.
 *		     Output starts at first non-frame pixel.
 *              x = x coordinate of the top left corner of the rectangle
 *		    relative to OUTSIDE of frame.
 *              y = y coordinate of the top left corner of the rectangle
 *		    relative to OUTSIDE of frame.
 *              w = width of the rectangle
 *              h = height of the rectangle
 *
 *	Return Code:
 *		ia_successful	good completion
 */

Int32 CDECL 
i_ditherFloydRGBTo8_Process (
    IPState    *pIPState,
    ADitherState *state,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    UInt32	sBpl,
    UInt8      *pD,
    UInt32	dBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h)
{
    PROLOGUE;
    register Int16 *lookup;
    register Int16 desired, i;
    register Int16 *err_ptr;
    register Int16 *prev_ab;
    UInt8 *out_levels,*out_ptr;
    Int16 *closest_levels;
    Int16  line;
    Int16 *base;

    base = state->errorLut;

    /* Move the pointers to the right point in the planes */
    pR += x + (y * sBpl);
    pG += x + (y * sBpl);
    pB += x + (y * sBpl);
    pD += (dBpl << cLogFrameBits) + cFrameBits;

    for (line = 0; line < h; line++)
    {
	err_ptr = state->errorBuf[0];
	out_levels = &state->outputLevels[0][128];
	closest_levels = &state->closestLevels[0][128];
	prev_ab = &zero[0]; 
	out_ptr = pD;
	
	for (i = 0; i < w; i++, err_ptr++)
	{
	    desired = state->redLut[UCHAR_ACCESS(pR + i)] + err_ptr[0] + prev_ab[0];
	    UCHAR_ACCESS(out_ptr++) = out_levels[desired] + state->outputOffset;
	    
	    /* lookup is index into error-producing arrays,
	    closest_levels[desired] is the graylevel that is nearest
	    the desired brightness of the incoming pixel, subtracted from the
	    desired value, and multiplied by 4;
	    */
	    lookup = base + closest_levels[desired];
	    err_ptr[-1] += lookup[0];			/* D error */
	    err_ptr[ 0]  = lookup[1] + prev_ab[1];	/* C error */
	    prev_ab = lookup + 2;			/* B error */
	}
	   
	err_ptr = state->errorBuf[1];
	out_levels = &state->outputLevels[1][128];
	closest_levels = &state->closestLevels[1][128];
	prev_ab = &zero[0];
	out_ptr = pD;

	for (i = 0; i < w; i++, err_ptr++)
	{
	    desired = state->greenLut[UCHAR_ACCESS(pG + i)] + err_ptr[0] + prev_ab[0];
	    UCHAR_ACCESS(out_ptr++) += out_levels[desired];
	    lookup = base + closest_levels[desired];
	    err_ptr[-1] += lookup[0];			/* D error */
	    err_ptr[ 0] =  lookup[1] + prev_ab[1];	/* C error */
	    prev_ab = lookup + 2;			/* B error */
	}

	err_ptr = state->errorBuf[2];
	out_levels = &state->outputLevels[2][128];
	closest_levels = &state->closestLevels[2][128];
	prev_ab = &zero[0];
	out_ptr = pD;

	for (i = 0; i < w; i++, err_ptr++)
	{
	    desired = state->blueLut[UCHAR_ACCESS(pB + i)] + err_ptr[0] + prev_ab[0];
	    UCHAR_ACCESS(out_ptr++) += out_levels[desired];
	    lookup = base + closest_levels[desired];
	    err_ptr[-1] += lookup[0];		/* D error */
	    err_ptr[ 0] =  lookup[1] + prev_ab[1];	/* C error */
	    prev_ab = lookup + 2;			/* A error */
	}

	/* Move to the next line */
	pR += sBpl;
	pG += sBpl;
	pB += sBpl;
	pD += dBpl;
    }
    NORMAL_RETURN;
}


/* PUBLIC */
/*
 * i_ditherFloydRGBTo8_Finish 
 *
 *	We're done; clean up. Warning: state will be freed, so don't
 *	pass it again.
 *
 *	Input:
 *		state = ADitherState returned from ..._Init
 *	
 *	Return Code:
 *		ia_successful	good completion
 */

Int32 CDECL 
i_ditherFloydRGBTo8_Finish (
    IPState    *pIPState,
    ADitherState *state)
{
    PROLOGUE;
    int k;

    for (k = 0; k < cPlanes; k++)
    {  
	/* Free the outputLevels thing */
	if (state->outputLevels[k])
	{
	    FREE (state->outputLevels[k]);
	    state->outputLevels[k] = 0;
	}

	if (state->closestLevels[k])
	{
	    FREE (state->closestLevels[k]);
	    state->closestLevels[k] = 0;
	}

	if (state->errorBuf[k])
	{
	    /* remember in init_cfloyd we did ++errorBuf */

	    /* NOTE: The following line causes MPW C to generate an odd address:
	     * FREE (state->errorBuf[k] - 1);
             *
	     * so it has been replaced with the equivalent (RDF) :
	     */
	    --state->errorBuf[k];
	    FREE (state->errorBuf[k]);
	}
    }

    if (state->errorLut)
	FREE (state->errorLut);

    FREE (state);

    NORMAL_RETURN;
}

/*

$Log: i_cfloyd.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.17  1998/01/21 21:22:29  ddavies
i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v

Fix Doug's weird syntax for FREE - don't decrement the thing freed in the
macro invocation.

Revision 1.16  1998/01/13 23:47:22  ddavies
i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.15  1997/11/26 19:02:32  xzhu
xzhu on Wed Nov 26 10:41:46 PST 1997

i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v

Created new function i_cMapToLut1. The difference between i_cMapToLut and i_cMapToLut1 is selecting representing value in each indexing interval. Based on Dan's suggestion, the representing value in every  indexing interval is 0x11, 0x22, ... 0xff instead of 0x10, 0x20, ... 0xf0. The new representing value will enhance white and black points.

Revision 1.14  1997/05/02 01:19:21  ddavies
i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v

Change comments to specify that the rectangles are relative to the outside
of the source image frames and that the output always goes to the first
non-frame pixel of the dest image.

Revision 1.13  1997/04/18 01:20:04  jcraig
jcraig on Thu Apr 17 18:17:54 PDT 1997

i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v

increase size of error lut from 256 to 512 to fix bug #114.  the
difference in the pixel from the image and the color map pixel value
was greater than abs(128) and so we referenced off the end of the
table.
added comments about tables used by the code.

Revision 1.12  1996/02/10 01:26:23  dferrell
i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v

Added extra parens around & operations so that gcc 2.7.2 won't complain.

Revision 1.11  1995/12/13 02:42:49  ejaquith
i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.10  1995/10/18  03:52:22  ddavies
 * i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.9  1995/10/13  23:49:54  dferrell
 * i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.8  1994/07/19  23:17:29  dferrell
 * dferrell on Tue Jul 19 16:17:05 PDT 1994
 *
 * i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v
 *
 * Corrected my last change which was not quite right (as Purify was
 * quick to note).
 *
 * Revision 1.7  1994/07/15  18:54:15  dferrell
 * dferrell on Fri Jul 15 11:53:56 PDT 1994
 *
 * i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v
 *
 * Changed a line in i_ditherFloydRGBTo8_Finish () which made MPW C
 * generate incorrect code which, in turn, caused a crash on the Mac.
 *
 * Revision 1.6  1994/07/13  00:02:50  ejaquith
 * i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v
 *
 * Removed #include "iadef.prv"
 * Made a couple variable parameters get allocated directly (instead
 * of some temporary pointer)
 *
 * Revision 1.5  1994/06/28  01:57:17  macleod
 * i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v
 *
 * A couple of routines were calling calloc and free when they
 * should have been calling CALLOC and FREE
 *
 * Revision 1.4  1994/05/19  00:44:57  macleod
 * macleod on Wed May 18 17:44:34 PDT 1994
 *
 * i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v
 *
 * Roll back changes to get rid of buggy Luv-based stuff
 *
 * Revision 1.2  1994/02/04  20:14:04  dferrell
 * i_cfloyd.c -> /product/ipcore/ipshared/src/RCS/i_cfloyd.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/22  03:48:56  ddavies
 * Initial revision
 *
 * Revision 1.12  1993/12/02  19:26:17  ddavies
 * scfloyd.c -> /product/ipcore/ipshared/src/RCS/scfloyd.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.11  1993/11/24  17:03:17  macleod
 * macleod on Wed Nov 24 09:02:03 PST 1993
 *
 * scfloyd.c -> /product/ipcore/ipshared/src/RCS/scfloyd.c,v
 *
 * Wasn't handling byte ordering correctly on the PC version. This caused two
 * problems. First, the frame was getting stomped, and second, the
 * error was being sent to the wrong place, which didn't really look all that
 * bad, but was different than the SPARC version, anyway.
 *
 * Revision 1.10  1993/11/05  01:53:42  ddavies
 * scfloyd.c -> /product/ipcore/ipshared/src/RCS/scfloyd.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


