/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_cmask.c
 *
 *	Make an image-is-in-color mask. This mask is a 1-bit mask where ON
 *	bits indicate pixels which have color under them. "having color"
 *	is defined as having a difference between the three channels 
 *	greater than some constant.
 *	We may also want to suppress the color mask in regions where
 *	low-contrast areas are gray, but edges are falsely colored
 *	due to registration errors. The question is how to know whether or not
 *	the edge's color is unique, i.e. to mask high-contrast edges in colored
 *	regions, and not mask high-contrast edges in gray regions. This will
 *	require some experimentation, but I suspect that if one looks at the 
 *	last few mask bits, one can get a pretty good idea of whether the region
 *	is gray or not. That won't work on the left edge, of course.
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"

/* Prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
#include "cmask.pub"
#include "cmask.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_cmask.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");
    
/*******************************************************************************
 * i_cmask
 *
 * 	Create a color mask for a region. 
 *
 * Arguments:
 *	pRed, pGreen, pBlue =	Pointers to the image data, including frame
 *	sBpl =			Bytes/line in the R, G, and B planes
 *	pMask =			Place to put resulting mask plane
 *	dBpl =			Bytes/line in resulting mask plane
 *	w =			Width of a plane, including frame, in pixels
 *	h =			Height of a plane, including frame, in pixels
 *
 * Returns:
 *
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_cmask (
    IPState    *pIPState,
    UInt8      *pRed,
    register UInt8 *pGreen,
    UInt8      *pBlue,
    Int32	sBpl,
    register UInt8 *pMask,
    Int32	dBpl,
    Int32	w,
    Int32	h)
{
    PROLOGUE;
    register Int32 j, green, delta, lastGreen, delta1;
    register UInt32 bits, *outP, lastBits;
    int i;
    UInt8 *saveMask;

    saveMask = pMask;

    /* Move past the frame */
    pRed += (sBpl << cLogFrameBits) + cFrameBits;
    pGreen += (sBpl << cLogFrameBits) + cFrameBits;
    pBlue += (sBpl << cLogFrameBits) + cFrameBits;
    pMask += (dBpl << cLogFrameBits) + cFrameBytes;
    h -= 2*cFrameBits;
    w -= 2*cFrameBits;
/* 
#undef UCHAR_ACCESS
#define UCHAR_ACCESS(p) *(p)
*/

    for (i = 0; i < h; i++)
    {
	/* Do a line */
	outP = (UInt32 *) pMask;
	bits = 0;
	lastGreen = 0;

	/*
	 * "lastBits" always contains the rolling bits as they go by, which is
	 * used to determine whether the last few bits were colored or not
	 */
	lastBits = 0;
	for (j = 0; j < w; j++)
	{
	    /*
	     * Compute the magnitude of difference between this pixel and its left 
	     * and upper neighbors. We use the green channel for this test, which
	     * obviously won't work on completely non-green image. In that case,
	     * it doesn't matter, since we only care about not introducing color
	     * edges where there weren't color edges before.
	     */
	    green = UCHAR_ACCESS(pGreen + j);
	    /* I don't know whether to say "green+green+green", which is fairly
	       safe, or "green*3", which can be done in one inst. on some processors
	       (most notably the 486) but the compiler might not find. Try it later.
	    */
	    delta = (green + green + green) - 
		    (UCHAR_ACCESS(pGreen - (sBpl + j)) +
		     UCHAR_ACCESS(pGreen - (sBpl - 1 + j)) +
		     lastGreen);

	    if (delta < 0) delta = -delta;

	    if (delta < cLowContrastCutoff|| (lastBits & 3) == 3 || 1)
	    {
		/* Subtract the red and blue pixels */
		delta = green - UCHAR_ACCESS(pRed + j);
		if (delta < 0) delta = -delta;
		delta1 = green - UCHAR_ACCESS(pBlue + j);
		if (delta1 < 0) delta1 = -delta1;
		if (delta + delta1 > cutoffTable[green >> 4])
		{
		    /* Shift a one into the output bits */
/*		    bits = (bits <<  1) | 1; */
		    bits = (bits  >> 1) | 0x80000000;
		    lastBits = (lastBits << 1) | 1;
		}
		else
		{
		    /* Shift a zero into the output bits */
/*		    bits <<= 1; */
		    bits >>= 1;
		    lastBits <<= 1;
		}
	    }
	    else
	    {
		/* Shift a zero into the output bits */
/*		bits <<= 1; */
		bits >>= 1;
		lastBits <<= 1;
	    }

	    /* If we're on the last pixel of a 32-pixel boundary, write the bits we've accumulated */
	    if ((j & 0x1f) == 0x1f)
	    {
		/* We have some bits to write */
		*outP++ = bits;
		bits = 0;
	    }
		    
	    lastGreen = green;
	}

	/* Output the last one, if needed */
	if (j & 31)
	    *outP++ = bits;

	pRed += sBpl;
	pGreen += sBpl;
	pBlue += sBpl;
	pMask += dBpl;
    }

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    i_swapImageBytes (pIPState, saveMask + (dBpl << cLogFrameBits), dBpl, h);
#endif

    NORMAL_RETURN;
}

/******************************************************************************
 * i_ccluster
 *
 * 	Create color cluster. 
 *
 * Arguments:
 *	pRed, pGreen, pBlue =	Pointers to the image data, not including frame
 *	sBpl =			Bytes/line in the R, G, and B planes
 *	ppCluster =		Point of table
 *	x0 =		        left corner coordinate
 *	y0 =			left corner coordinate
 *	w  =			Width of image, in pixels
 *	h  =			Height of image, in pixels
 *      ccNum =                 Number of cluster
 *
 * Returns:
 *
 *	ia_successful, ...
 *****************************************************************************/
Int32 CDECL 
i_ccluster(     
    IPState    *pIPState,
    UInt8      *pl,
    UInt8      *pa,
    UInt8      *pb,
    Int32	sBpl,
    Int32       x0,
    Int32       y0,
    Int32	w,
    Int32	h,
    CCCLUSTER  *pCluster,
    Int32      *ccNum)
{
    PROLOGUE;
    UInt8        *pll, *pca, *pcb;
    Int32         bpl, x, y, index;
    Int32         classNum, idx, num;
    CCCLUSTER    *pClu;
    UInt8         ll, ca, cb, loop;
    Int32         newIdx, il, ia, ib;

    ERR_GEN( pCluster == NULL, cErrNullPtr);

    pClu = pCluster;
    for(index=0; index<TABLENGTH; ++index)
	(pClu+index)->index = index;

    bpl = sBpl;
    pll = pl + y0*bpl +x0;
    pca = pa + y0*bpl +x0;
    pcb = pb + y0*bpl +x0;
    for(y=0; y<h; ++y, pll += bpl, pca += bpl, pcb += bpl)
    {
	for(x=0; x<w; ++x)
	{
	    ll = UCHAR_ACCESS(pll + x);
	    ca = UCHAR_ACCESS(pca + x);
	    cb = UCHAR_ACCESS(pcb + x);
	    index = COLORTOINDEX1((ll/cLFac),(ca/cABFac),(cb/cABFac));
	    ++ (pClu+index)->num;
	    (pClu+index)->l += (Int32) ll;
	    (pClu+index)->a += (Int32) ca;
	    (pClu+index)->b += (Int32) cb;
	}
    }

    classNum =0;
    loop =  0;
    do
    {
	num = classNum;
	classNum = 0;
	for(index=0; index<TABLENGTH; ++index)
	{
	    if((idx = (pClu+index)->num) < 1) continue;
	    il = (((pClu+index)->l /idx) + cHLFac)/cLFac;
	    ia = (((pClu+index)->a /idx) + cHABFac)/cABFac;
	    ib = (((pClu+index)->b /idx) + cHABFac)/cABFac;
	    (pClu+index)->center = CENTER2INDEX1(il, ia, ib);
	}

	for(index=0; index<TABLENGTH; ++index)
	{
	    if((pClu+index)->num < 1) continue;
	    ++ classNum;
	    newIdx = (pClu+index)->center;
	    for(idx = index+1; idx<TABLENGTH; ++idx)
	    {
		if((pClu+idx)->num < 1) continue;
		if( newIdx != (pClu+idx)->center) continue;
		(pClu+index)->l += (pClu+idx)->l;
		(pClu+index)->a += (pClu+idx)->a;
		(pClu+index)->b += (pClu+idx)->b;
		(pClu+index)->num += (pClu+idx)->num;
		(pClu+idx)->num =0;
	    }
	    
	}
    }while(num != classNum && classNum > cMINClass && ++loop < cMAXLoop);

    classNum =0;
    for(index=0; index<TABLENGTH; ++index)
    {
	if((pClu+index)->num < 1) continue;
	++ classNum;
    	num = (pClu + index)->num;
	(pClu+index)->ll = (UInt8) ((pClu+index)->l /num);
	(pClu+index)->aa = (UInt8) ((pClu+index)->a /num);
	(pClu+index)->bb = (UInt8) ((pClu+index)->b /num);
	ERR_PASS(i_LabToRGB(pIPWState, (pClu+index)->ll, 
			    (pClu+index)->aa, (pClu+index)->bb,
			    &(pClu+index)->red, &(pClu+index)->green,
			    &(pClu+index)->blue));
    }

    *ccNum = classNum;

    NORMAL_RETURN;

}

/*******************************************************************************
 * i_createIndexMapTable
 *
 * 	Create color index map table. 
 *
 * Arguments:
 *	pl, pa, pb  =	        Pointers to the image data, not including frame
 *	sBpl =			Bytes/line in the R, G, and B planes
 *	pMapTable =		Point of table
 *	w =			Width of image, in pixels
 *	h =			Height of image, in pixels
 *      pIndex =                Point to the index image
 *
 * Returns:
 *
 *	ia_successful, ...
 *****************************************************************************/
Int32 CDECL 
i_createIndexMapTable(     
    IPState    *pIPState,
    UInt8      *pl,
    UInt8      *pa,
    UInt8      *pb,
    Int32	sBpl,
    Int32	w,
    Int32	h,
    UInt16     *pIndex,
    CCMAPTABLE *pMapTable)
{
    PROLOGUE;
    UInt8        *pll, *pca, *pcb;
    UInt16       *pidx;
    Int32         bpl, x, y, index;
    Int32         num;
    UInt8         ll, ca, cb;
    CCMAPDATA    *pClu = NULL;

    ERR_GEN( pMapTable == NULL, cErrNullPtr);

    pClu = (CCMAPDATA *) IP_CALLOC(TABLENGTH, sizeof(CCMAPDATA));
    ERR_GEN(pClu == NULL, cErrNomem);
    
    MEMSET((void *)pClu, 0, TABLENGTH*sizeof(CCMAPDATA));

    pll = pl;
    pca = pa;
    pcb = pb;
    pidx = pIndex;
    bpl = sBpl;
    for(y=0; y<h; ++y, pll += bpl, pca += bpl, pcb += bpl, pidx += w)
    {
	for(x=0; x<w; ++x)
	{
	    ll = UCHAR_ACCESS(pll + x);
	    if( ll == 0 ) continue;
	    ca = UCHAR_ACCESS(pca + x);
	    cb = UCHAR_ACCESS(pcb + x);
	    index = COLORTOINDEX(ll,ca,cb);
	    *(pidx+x) = (UInt16) index;
	    ++ (pClu+index)->num;
	    (pClu+index)->l += (Int32) ll;
	    (pClu+index)->a += (Int32) ca;
	    (pClu+index)->b += (Int32) cb;
	}
    }

    for(index=0; index<TABLENGTH; ++index)
    {
	if((num = (pClu+index)->num) < 1) continue;
	(pMapTable+index)->l = ll = (UInt8) ((pClu+index)->l /num);
	(pMapTable+index)->a = ca = (UInt8) ((pClu+index)->a /num);
	(pMapTable+index)->b = cb = (UInt8) ((pClu+index)->b /num);

	ERR_PASS_CLEANUP(i_LabToRGB(pIPWState, ll, ca, cb, 
				    &(pMapTable+index)->red, 
				    &(pMapTable+index)->green, 
				    &(pMapTable+index)->blue));
	(pMapTable+index)->num = num;
    }
    
    IP_FREE( pClu );

    NORMAL_RETURN;

 ERR_CLEANUP:

    if( pClu ) IP_FREE( pClu );

    ERR_RETURN;
}

/*

$Log: i_cmask.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.16  1998/11/13 18:18:21  xzhu
xzhu on Fri Nov 13 10:16:36 PST 1998

i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v

Added new input to handle rectangle area. Changed structure define to reduce memory space.

Revision 1.15  1998/08/11 23:46:41  xzhu
xzhu on Tue Aug 11 16:46:30 PDT 1998

i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v

Added new function.

Revision 1.14  1998/07/13 23:32:56  xzhu
xzhu on Mon Jul 13 16:32:31 PDT 1998

i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v

Added protection for accessing inxList.

Revision 1.13  1998/06/29 23:23:10  xzhu
xzhu on Mon Jun 29 16:22:53 PDT 1998

i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v

Added include cmask.

Revision 1.12  1998/01/13 23:47:23  ddavies
i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.11  1995/12/13 02:42:53  ejaquith
i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.10  1995/10/18  03:52:27  ddavies
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.9  1995/09/13  23:29:17  ddavies
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * Remove invertImage parameter from i_swapImageBytes.  I backed out of that
 * change.
 *
 * Revision 1.8  1995/09/13  00:11:14  ddavies
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * Set the new invertImage parameter to cFalse in the call to i_swapImageBytes.
 *
 * Revision 1.7  1994/07/13  00:03:29  ejaquith
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * Removed #include "iadef.prv"
 * Changed 2 constant globals to macro references
 *
 * Revision 1.6  1994/06/16  22:47:34  macleod
 * macleod on Thu Jun 16 15:47:24 PDT 1994
 *
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * Took out an inappropriate comment
 *
 * Revision 1.5  1994/02/04  20:14:09  dferrell
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1994/01/22  01:07:00  ddavies
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * change "include shrio." to "include shrpixr." since ipshared's libio is
 * being absorbed by libpixr.
 *
 * Revision 1.3  1993/12/02  19:22:38  ddavies
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:27:27  ddavies
 * i_cmask.c -> /product/ipcore/ipshared/src/RCS/i_cmask.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

