/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_cmpgry.c: compare two gray images, putting results into a binary image.
 *
 *                  Int32     	i_compareGrayImages()
 *                  void     	ip_makeCompareGrayScaleTable()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

    /*  prototyes */
#include "shrip.pub"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_cmpgry.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");
/* The procedure is used to compare two gray images, producing a binary image.
 * All images have the same dimensions.  At every position where the pixel
 * in image 1 times multiplier mult1 is strictly greater than the pixel in
 * image 2 times mult2, we put a 1 in the binary image.
 *
 * This algorithm is used for page segmentation.  Pixels are tagged with the
 * length of the horizontal or vertical run they're in by some previous
 * operation.  This routine is used to find pixels that are mainly in
 * horizontal or mainly in vertical runs.  See pageseg.c in the XIS code.
 *
 * Currently, the only legal depth for the gray images is 8 bit/pixel.  We'll
 * leave a depth argument for the future.
 *
 * The multipliers are clearly a problem.  Without them, we could compare
 * four pixels at once.  Too bad.  Instead, we do one of two comparisons:
 * 	Original comparison: mult1 * p1 > mult2 * p2
 *	New comparison:
 *		if mult1 >= mult2  p1 > mult2/mult1 * p2
 *		if mult1 < mult2   mult1/mult2 * p1 > p2
 *
 * We build a table of either mult1/mult2 or mult2/mult1.  The table is
 * indexed with a pixel value to implement the multiply.  The result is
 * another byte, which we use in the comparison.
 *
 * Note that since we are comparing for mult1*p1 strictly greater than mult2*p2,
 * we can change which pixel indexes the table, but we can't change the
 * order of the operands in the subtraction.  If we change the order of the
 * operands and invert the results, everything works for the cases where
 * mult1*p1 != mult2*p2, but fails when they are equal.  The inverse of
 * (mult1*p1 > mult2*p2) is (mult1*p1 <= mult2*p2), not (mult1*p1 < mult2*p2).
 *
 * The XIS code this routine replaces finds the ends of a runs of 1's in the
 * output image, then fills in the middle.  To deal with each pixel then you'd
 * need a couple of reads to get the pixels, another read for the lookup table,
 * a subtract and a branch.  Say we just insert the MS bit of the subtract into
 * the binary result.  This has the same reads and subtract, followed by two
 * shifts and an OR.  Or, if we process the line from right to left, we can
 * get away with 1 shift, a mask operation to get the MS bit of the result
 * of the subtract, and an OR.  Since Mask ops are cheaper than shifts on a
 * PC, this is slightly preferable.
 *
 * At least it takes this much stuff in C.  In assembly, we can use the carry
 * flag to replace the two shifts and OR with a single ADC (add with carry)
 * operation that does the shift and inserts the new bit all in one glorious
 * (1 cycle) operation.  In C, the extra overhead of inserting
 * the new bit each time is offset by the overhead of calling the procedures
 * that find the beginning and end of each run and actually creating the runs.
 * In Assembly, it's no contest.
 *
 * Ok, ok, the scheme that creates the runs can just skip over the white parts
 * of the image, so it gains something there.  If there is lots of white
 * space in the result, this might count for something.  Of course, we
 * do spend time clearing out the binary image beforehand, so this isn't a
 * complete gain.  We'll stick with the new scheme and tell Luc he can
 * skip the step that clears out the image.
 */


/*********************************************************************
 *************************      Procedures     ***********************
 *********************************************************************/
/* PUBLIC */
/***********************************************************************
 *  i_compareGrayImages()
 *	Input: 
 *	pS1 =	 ptr to gray source image 1 (includes frame)
 *	sBpl1 =	 source image 1 bytes/line
 *	mult1 =	 mulitplier for gray source image 1
 *	pS2 =	 ptr to gray source image 2 (includes frame)
 *	sBpl2 =	 source gray image 2 bytes/line
 *	mult2 =	 mulitplier for gray source image 2
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	depth =	 bits/pixel of all gray images
 *	pD =	 ptr to dest binary image
 *	dBpl =	 dest binary image bytes/line
 *
 *	On return: 
 *		Binary image has a 1 in every pixel where
 *		mult1 * PixelFromImage1 > mult2 * PixelFromImage2.
 *		This routine writes into every location of the
 *		destination image, it need not be cleared before
 *		calling this routine.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	image depth was not 8
 ***********************************************************************/
Int32   CDECL 
i_compareGrayImages (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt32	sBpl1,
    UInt32	mult1,
    UInt8      *pS2,
    UInt32	sBpl2,
    UInt32	mult2,
    UInt32	w,
    UInt32	h,
    UInt32	depth,
    UInt8      *pD,
    UInt32	dBpl)
{
PROLOGUE;
 UInt8	scaleTable[256];		/* table of scale factors (mult1/mult2
					 * or mult2/mult1)
					 */

	/* Check the depth of the source images */
    if (depth != 8)
	return ia_depthNotSupported;

	/* point to the images inside the frames */
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pS1 += sBpl1 << cLogFrameBits;	/* do left frame adjustment below */
    pS1 += 8 * cFrameBytes;
    pS2 += sBpl2 << cLogFrameBits;
    pS2 += 8 * cFrameBytes;
    pD += dBpl << cLogFrameBits;
    pD += cFrameBytes;

	/* fill in the table */
    if (mult1 >= mult2)
    {
	    /* make a table of mult2/mult1 */
	ip_makeCompareGrayScaleTable (pIPState, mult2, mult1, scaleTable);

	    /* do the real work 1 line at a time comparing
	     * *pS1 with scaleTable[*pS2]  (or p1 with (mult2/mult1 * p2))
	     */
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
	    ip_compareGrayLineScaleP2 (pIPState, pS1, pS2, pD, scaleTable, w);
	}

    }
    else
    {
	    /* make a table of mult1/mult2 */
	ip_makeCompareGrayScaleTable (pIPState, mult1, mult2, scaleTable);

	    /* do the real work 1 line at a time comparing
	     * scaleTable[*pS1] with *pS2  (or (mult1/mult2 * p1) with p2)
	     */
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
	    ip_compareGrayLineScaleP1 (pIPState, pS1, pS2, pD, scaleTable, w);
	}
    }

    NORMAL_RETURN;
}


/* This is the routine that makes the tables.  It computes mult2/mult1 to
 * 16 fractional places, then repeatedly adds the quotient to a running sum
 * and putting the proper 8 bits from the running sum into the table
 */

/* PRIVATE */
/**************************************************************************
 *  ip_makeCompareGrayScaleTable()  
 *		Computes a fractional table used by i_compareGrayImages.
 *
 *	Input:
 *		numerator =	numerator of fraction
 *		denominator =	denominator of fraction
 *		pTable =	points to table to be filled.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_makeCompareGrayScaleTable (
    IPState    *pIPState,
    UInt32	numerator,
    UInt32	denominator,
    UInt8      *pTable)
{
    PROLOGUE;
    ScaledInt1616	 fraction;	/* numerator/denominator with 16
					 * fraction bits */
    ScaledInt1616	 sum;		/* used to accumulate multiples of
					 * fraction. */
    UInt32		 index;		/* used to address table */

	/* keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* compute the fraction.  This will have 16 fraction bits. */
    fraction = (numerator << 16) / denominator;

    for (index = 0, sum = 0; index < 256; index++)
    {
	pTable[index] = (UInt8)(((UInt32) sum) >> 16);
	sum += fraction;
    }
}

/*

$Log: i_cmpgry.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.7  1998/01/13 23:47:25  ddavies
i_cmpgry.c -> /product/ipcore/ipshared/src/RCS/i_cmpgry.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1995/12/13 02:42:56  ejaquith
i_cmpgry.c -> /product/ipcore/ipshared/src/RCS/i_cmpgry.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/10/13  23:50:02  dferrell
 * i_cmpgry.c -> /product/ipcore/ipshared/src/RCS/i_cmpgry.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.4  1994/07/12  23:55:38  ejaquith
 * i_cmpgry.c -> /product/ipcore/ipshared/src/RCS/i_cmpgry.c,v
 *
 * removed #include "iadef.prv".
 *
 * Revision 1.3  1993/12/02  19:22:42  ddavies
 * i_cmpgry.c -> /product/ipcore/ipshared/src/RCS/i_cmpgry.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:27:35  ddavies
 * i_cmpgry.c -> /product/ipcore/ipshared/src/RCS/i_cmpgry.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

