/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_cntbit.c: counts the number of bits in an image.
 *
 *           Int32     	i_countBits()
 *
 */


#include <stdio.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "bitcount.pub"
#include "lmask.prv"
#include "rmask.prv"

    /*  prototyes */
#include "shrip.pub"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_cntbit.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/****************************************************************************
 *  i_countBits():  count the number of 1's in a rectangle of an image.
 *
 *	Input:
 *		pS:	Pointer to source image data WITHOUT frame
 *		x:	x offset within image
 *		y:	y offset within image
 *		w:	width of source rectangle
 *		h:	height of source rectangle
 *		d:	depth of source image
 *		bpl:	source image bytes/line.
 *		pCount:	pointer to UInt32 that will contain the result
 *
 *	Returns:
 *		ia_successful		 it worked
 *
 ************************************************************************/
Int32 CDECL 
i_countBits (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	d,
    Int32	bpl,
    Int32      *pCount)
{
    PROLOGUE;
    UInt32	total;			/* sum of bits seen so far */
    UInt32	leftSideMask;
    UInt32	rightSideMask;		/* masks used to discard bits on the
					 * left and right sides of the rect */
    Int32	leftBitNum;
    Int32	rightBitNum;		/* indexes of leftmost and rightmost
					 * bits in the rectangle */

    if (w == 0)
    {
	*pCount = 0;
	NORMAL_RETURN;
    }

	/* Calculate the left and right masks.  The static array
	 * leftMask[x] contains x+1 1's at the left side of the
	 * word.  This is exactly what we want to get rid of bits
	 * on the right side of the rightmost word.
	 * The funny calculation says "get the index of the leftmost bit
	 * of the rightmost pixel in the line ((x + w - 1) * d), then
	 * add the number of bits required to point to the rightmost
	 * bit of that pixel (d - 1).
	 */
    rightBitNum = ((x + w - 1) * d) + (d - 1);
    rightSideMask = leftMask[rightBitNum & 0x1F];

	/* The static array rightMask[x] contains x+1 1's on the right
	 * side of the word.  We really need all 1's when the leftmost
	 * bit offset is 0, 30 1's on the right when the index is 1 and
	 * so on.  This can be done by simply complementing the index.
	 */
    leftBitNum = (x * d);
    leftSideMask = rightMask[(~leftBitNum) & 0x1F];

	/* Having made the masks, move leftBitNum and rightBitNum
	 * to word boundaries.  We'll adjust the address pointer
	 * to there and calculate the width from those.
	 */
    leftBitNum &= ~0x1F;
    rightBitNum &= ~0x1F;

	/* Calculate the number of interesting words/line.  We always
	 * start and stop on word boundaries to keep from being
	 * confused by the little-endian / big endian issue.
	 */
    w = ((rightBitNum - leftBitNum) / 32) + 1;
    total = 0;

	/* Adjust the starting address to word containing leftmost bit
	 * bit of the rectangle.
	 */
    pS += y * bpl + (leftBitNum / 8);

	/* Loop through the lines */
    for ( ; h > 0; h--, pS += bpl)
    {
	Int32 curTotal;
	ERR_PASS( ip_countBitsLine (pIPState, pS,
				    w,
				    leftSideMask,
				    rightSideMask,
				    (const UInt8 *) &(bitCount[0]),
				    &curTotal));
	total += curTotal;
    }

	/* return the result */
    *pCount = total;

    NORMAL_RETURN;
}

/*

 $Log: i_cntbit.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:10  jlin
 Imported sources

 Revision 1.7  1998/01/13 23:47:26  ddavies
 i_cntbit.c -> /product/ipcore/ipshared/src/RCS/i_cntbit.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.6  1995/12/13 02:43:00  ejaquith
 i_cntbit.c -> /product/ipcore/ipshared/src/RCS/i_cntbit.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1994/06/20  20:48:52  ddavies
 * i_cntbit.c -> /product/ipcore/ipshared/src/RCS/i_cntbit.c,v
 *
 * Add omitted CDECL to declaration of i_countBits.
 *
 * Revision 1.4  1994/01/11  22:33:15  ddavies
 * ddavies on Tue Jan 11 14:32:58 PST 1994
 *
 * i_cntbit.c -> /product/ipcore/ipshared/src/RCS/i_cntbit.c,v
 *
 * Now includes bitcount.pub instead of bitcount.prv.
 *
 * Revision 1.3  1993/12/02  19:22:46  ddavies
 * i_cntbit.c -> /product/ipcore/ipshared/src/RCS/i_cntbit.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/16  02:54:27  ddavies
 * i_cntbit.c -> /product/ipcore/ipshared/src/RCS/i_cntbit.c,v
 *
 * The procedure i_countBits was changed to count bits within a rectangle of
 * the image instead of in the entire image.  The rectangle is specified
 * relative to the upper left corner of the frame so we can count bits in the
 * frame too.
 * The function of the inner loop was moved to ip_countBitsLine.  This is
 * implemented in both C and SPARC assembler.
 *

 */

