/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_color.c
 *
 *	Color transformations and format conversions
 *
 * Exports:
 *
 *	i_RGBToY	- Convert an RGB image to a luminance (Y) channel
 *	i_RGBToY2	- Convert an RGB image to a DITC luminance (Y) channel
 *	i_RGBToYIP	- Convert an RGB image to a luminance (Y) channel
 *	i_RGBToYFast	- Convert an RGB image to a luminance (Y) channel
 *	i_indexed8ToRGB	- Convert an indexed-color image to seperate (R,G,B)
 *                        planes
 *	i_indexed4ToRGB	- Convert an indexed-color image to seperate (R,G,B)
 *                        planes
 *	i_RGBToLuvInit	- Initialize for RGB->LUV conversion (specify white)
 *	i_RGBToLuv	- Convert RGB to LUV. SLOW!
 *      i_RGBToLuvFinish- Finalize RGB->LUV conversion
 *	i_RGBToLabInit	- Initialize for RGB->Lab conversion (specify white)
 *	i_RGBToLab	- Convert RGB to Lab. SLOW!
 *      i_RGBToLabFinish- Finalize RGB->Lab conversion
 *	i_RGB2Lab	- Convert RGB to Lab.
 *	i_Lab2RGB	- Convert Lab to RGB.
 */



#include <stdio.h>
#include <string.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"
/*#include "aiedefs.pub" */
/* #include "aiedefs.prv" */

/* Prototypes */
#include "shros.pub"
#include "shrip.pub"
#include "shrip.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_color.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");


/*
 * look-up tables
 *
 *	These tables are an expansion of:
 *
 *	Y = (0.299)Red + (0.587)G + (0.114)B
 *
 *	so that we can do three look-ups and two adds, rather than
 *	three multiplies and two adds.
 *	The coefficients are the standards from textbooks.
 */

 /* These are static const tables but the gcc compiler will complain if the
  * "const" appears here.  This is notable because otherwise it might not
  * quite look reentrant, but it is. */

static UInt8 redLut[] = 
{
 0,    0,    0,    0,    1,    1,    1,    2,
 2,    2,    2,    3,    3,    3,    4,    4,
 4,    5,    5,    5,    5,    6,    6,    6,
 7,    7,    7,    8,    8,    8,    8,    9,
 9,    9,   10,   10,   10,   11,   11,   11,
11,   12,   12,   12,   13,   13,   13,   14,
14,   14,   14,   15,   15,   15,   16,   16,
16,   17,   17,   17,   17,   18,   18,   18,
19,   19,   19,   20,   20,   20,   20,   21,
21,   21,   22,   22,   22,   23,   23,   23,
23,   24,   24,   24,   25,   25,   25,   26,
26,   26,   26,   27,   27,   27,   28,   28,
28,   29,   29,   29,   29,   30,   30,   30,
31,   31,   31,   31,   32,   32,   32,   33,
33,   33,   34,   34,   34,   34,   35,   35,
35,   36,   36,   36,   37,   37,   37,   37,
38,   38,   38,   39,   39,   39,   40,   40,
40,   40,   41,   41,   41,   42,   42,   42,
43,   43,   43,   43,   44,   44,   44,   45,
45,   45,   46,   46,   46,   46,   47,   47,
47,   48,   48,   48,   49,   49,   49,   49,
50,   50,   50,   51,   51,   51,   52,   52,
52,   52,   53,   53,   53,   54,   54,   54,
55,   55,   55,   55,   56,   56,   56,   57,
57,   57,   58,   58,   58,   58,   59,   59,
59,   60,   60,   60,   60,   61,   61,   61,
62,   62,   62,   63,   63,   63,   63,   64,
64,   64,   65,   65,   65,   66,   66,   66,
66,   67,   67,   67,   68,   68,   68,   69,
69,   69,   69,   70,   70,   70,   71,   71,
71,   72,   72,   72,   72,   73,   73,   73,
74,   74,   74,   75,   75,   75,   75,   76
};


static UInt8 greenLut[] = 
{
 0,      0,    1,      1,     2,     2,     3,     4,
 4,      5,    5,      6,     7,     7,     8,     8,
 9,      9,    10,    11,    11,    12,    12,    13,
 14,    14,    15,    15,    16,    17,    17,    18,
 18,    19,    19,    20,    21,    21,    22,    22,
 23,    24,    24,    25,    25,    26,    27,    27,
 28,    28,    29,    29,    30,    31,    31,    32,
 32,    33,    34,    34,    35,    35,    36,    36,
 37,    38,    38,    39,    39,    40,    41,    41,
 42,    42,    43,    44,    44,    45,    45,    46,
 46,    47,    48,    48,    49,    49,    50,    51,
 51,    52,    52,    53,    54,    54,    55,    55,
 56,    56,    57,    58,    58,    59,    59,    60,
 61,    61,    62,    62,    63,    63,    64,    65,
 65,    66,    66,    67,    68,    68,    69,    69,
 70,    71,    71,    72,    72,    73,    73,    74,
 75,    75,    76,    76,    77,    78,    78,    79,
 79,    80,    81,    81,    82,    82,    83,    83,
 84,    85,    85,    86,    86,    87,    88,    88,
 89,    89,    90,    90,    91,    92,    92,    93,
 93,    94,    95,    95,    96,    96,    97,    98,
 98,    99,    99,   100,   100,   101,   102,   102,
103,   103,   104,   105,   105,   106,   106,   107,
108,   108,   109,   109,   110,   110,   111,   112,
112,   113,   113,   114,   115,   115,   116,   116,
117,   117,   118,   119,   119,   120,   120,   121,
122,   122,   123,   123,   124,   125,   125,   126,
126,   127,   127,   128,   129,   129,   130,   130,
131,   132,   132,   133,   133,   134,   135,   135,
136,   136,   137,   137,   138,   139,   139,   140,
140,   141,   142,   142,   143,   143,   144,   144,
145,   146,   146,   147,   147,   148,   149,   149
};

static UInt8 blueLut[] = 
{
 0,    0,    0,    0,    0,    0,    0,    0,
 0,    1,    1,    1,    1,    1,    1,    1,
 1,    1,    2,    2,    2,    2,    2,    2,
 2,    2,    2,    3,    3,    3,    3,    3,
 3,    3,    3,    3,    4,    4,    4,    4,
 4,    4,    4,    4,    5,    5,    5,    5,
 5,    5,    5,    5,    5,    6,    6,    6,
 6,    6,    6,    6,    6,    6,    7,    7,
 7,    7,    7,    7,    7,    7,    7,    8,
 8,    8,    8,    8,    8,    8,    8,    9,
 9,    9,    9,    9,    9,    9,    9,    9,
10,   10,   10,   10,   10,   10,   10,   10,
10,   11,   11,   11,   11,   11,   11,   11,
11,   11,   12,   12,   12,   12,   12,   12,
12,   12,   12,   13,   13,   13,   13,   13,
13,   13,   13,   14,   14,   14,   14,   14,
14,   14,   14,   14,   15,   15,   15,   15,
15,   15,   15,   15,   15,   16,   16,   16,
16,   16,   16,   16,   16,   16,   17,   17,
17,   17,   17,   17,   17,   17,   18,   18,
18,   18,   18,   18,   18,   18,   18,   19,
19,   19,   19,   19,   19,   19,   19,   19,
20,   20,   20,   20,   20,   20,   20,   20,
20,   21,   21,   21,   21,   21,   21,   21,
21,   22,   22,   22,   22,   22,   22,   22,
22,   22,   23,   23,   23,   23,   23,   23,
23,   23,   23,   24,   24,   24,   24,   24,
24,   24,   24,   24,   25,   25,   25,   25,
25,   25,   25,   25,   25,   26,   26,   26,
26,   26,   26,   26,   26,   27,   27,   27,
27,   27,   27,   27,   27,   27,   28,   28,
28,   28,   28,   28,   28,   28,   28,   29
};


/*
 * look-up tables (2)
 *
 *	These tables are an expansion of:
 *
 *	Y = (0.252939)Red + (0.684458)G + (0.062603)B
 *
 *	so that we can do three look-ups and two adds, rather than
 *	three multiplies and two adds.
 *	These coefficients are NOT standard, but taken from DITC
 *	and used primarily for AIE (Auto image enhancement).
 */

static UInt8 redLut2[] = {
         0,     0,     1,     1,     1,     1,     2,     2,
         2,     2,     3,     3,     3,     3,     4,     4,
         4,     4,     5,     5,     5,     5,     6,     6,
         6,     6,     7,     7,     7,     7,     8,     8,
         8,     8,     9,     9,     9,     9,    10,    10,
        10,    10,    11,    11,    11,    11,    12,    12,
        12,    12,    13,    13,    13,    13,    14,    14,
        14,    14,    15,    15,    15,    15,    16,    16,
        16,    16,    17,    17,    17,    17,    18,    18,
        18,    18,    19,    19,    19,    19,    20,    20,
        20,    20,    21,    21,    21,    21,    22,    22,
        22,    23,    23,    23,    23,    24,    24,    24,
        24,    25,    25,    25,    25,    26,    26,    26,
        26,    27,    27,    27,    27,    28,    28,    28,
        28,    29,    29,    29,    29,    30,    30,    30,
        30,    31,    31,    31,    31,    32,    32,    32,
        32,    33,    33,    33,    33,    34,    34,    34,
        34,    35,    35,    35,    35,    36,    36,    36,
        36,    37,    37,    37,    37,    38,    38,    38,
        38,    39,    39,    39,    39,    40,    40,    40,
        40,    41,    41,    41,    41,    42,    42,    42,
        42,    43,    43,    43,    44,    44,    44,    44,
        45,    45,    45,    45,    46,    46,    46,    46,
        47,    47,    47,    47,    48,    48,    48,    48,
        49,    49,    49,    49,    50,    50,    50,    50,
        51,    51,    51,    51,    52,    52,    52,    52,
        53,    53,    53,    53,    54,    54,    54,    54,
        55,    55,    55,    55,    56,    56,    56,    56,
        57,    57,    57,    57,    58,    58,    58,    58,
        59,    59,    59,    59,    60,    60,    60,    60,
        61,    61,    61,    61,    62,    62,    62,    62,
        63,    63,    63,    63,    64,    64,    64,    64
};

static UInt8 greenLut2[] = {
         0,     1,     1,     2,     3,     3,     4,     5,
         5,     6,     7,     8,     8,     9,    10,    10,
        11,    12,    12,    13,    14,    14,    15,    16,
        16,    17,    18,    18,    19,    20,    21,    21,
        22,    23,    23,    24,    25,    25,    26,    27,
        27,    28,    29,    29,    30,    31,    31,    32,
        33,    34,    34,    35,    36,    36,    37,    38,
        38,    39,    40,    40,    41,    42,    42,    43,
        44,    44,    45,    46,    47,    47,    48,    49,
        49,    50,    51,    51,    52,    53,    53,    54,
        55,    55,    56,    57,    57,    58,    59,    60,
        60,    61,    62,    62,    63,    64,    64,    65,
        66,    66,    67,    68,    68,    69,    70,    70,
        71,    72,    73,    73,    74,    75,    75,    76,
        77,    77,    78,    79,    79,    80,    81,    81,
        82,    83,    84,    84,    85,    86,    86,    87,
        88,    88,    89,    90,    90,    91,    92,    92,
        93,    94,    94,    95,    96,    97,    97,    98,
        99,    99,   100,   101,   101,   102,   103,   103,
       104,   105,   105,   106,   107,   107,   108,   109,
       110,   110,   111,   112,   112,   113,   114,   114,
       115,   116,   116,   117,   118,   118,   119,   120,
       120,   121,   122,   123,   123,   124,   125,   125,
       126,   127,   127,   128,   129,   129,   130,   131,
       131,   132,   133,   133,   134,   135,   136,   136,
       137,   138,   138,   139,   140,   140,   141,   142,
       142,   143,   144,   144,   145,   146,   146,   147,
       148,   149,   149,   150,   151,   151,   152,   153,
       153,   154,   155,   155,   156,   157,   157,   158,
       159,   159,   160,   161,   162,   162,   163,   164,
       164,   165,   166,   166,   167,   168,   168,   169,
       170,   170,   171,   172,   172,   173,   174,   175
};

static UInt8 blueLut2[] = {
         0,     0,     0,     0,     0,     0,     0,     0,
         1,     1,     1,     1,     1,     1,     1,     1,
         1,     1,     1,     1,     1,     1,     1,     1,
         2,     2,     2,     2,     2,     2,     2,     2,
         2,     2,     2,     2,     2,     2,     2,     2,
         3,     3,     3,     3,     3,     3,     3,     3,
         3,     3,     3,     3,     3,     3,     3,     3,
         4,     4,     4,     4,     4,     4,     4,     4,
         4,     4,     4,     4,     4,     4,     4,     4,
         5,     5,     5,     5,     5,     5,     5,     5,
         5,     5,     5,     5,     5,     5,     5,     5,
         6,     6,     6,     6,     6,     6,     6,     6,
         6,     6,     6,     6,     6,     6,     6,     6,
         7,     7,     7,     7,     7,     7,     7,     7,
         7,     7,     7,     7,     7,     7,     7,     7,
         8,     8,     8,     8,     8,     8,     8,     8,
         8,     8,     8,     8,     8,     8,     8,     8,
         9,     9,     9,     9,     9,     9,     9,     9,
         9,     9,     9,     9,     9,     9,     9,     9,
        10,    10,    10,    10,    10,    10,    10,    10,
        10,    10,    10,    10,    10,    10,    10,    10,
        11,    11,    11,    11,    11,    11,    11,    11,
        11,    11,    11,    11,    11,    11,    11,    11,
        12,    12,    12,    12,    12,    12,    12,    12,
        12,    12,    12,    12,    12,    12,    12,    12,
        13,    13,    13,    13,    13,    13,    13,    13,
        13,    13,    13,    13,    13,    13,    13,    13,
        14,    14,    14,    14,    14,    14,    14,    14,
        14,    14,    14,    14,    14,    14,    14,    14,
        15,    15,    15,    15,    15,    15,    15,    15,
        15,    15,    15,    15,    15,    15,    15,    15,
        16,    16,    16,    16,    16,    16,    16,    16
};


/*
 * The following structure is used internally by the RGB To luminance
 * conversion functions.
 */
typedef struct
{
    Float64	   zWhite;
    /* specific to Lab conversion */
    Float64	   xWhite;
    Float64        yWhite;
    /* specific to Luv conversion */
    Float64        uPrimeWhite;
    Float64        vPrimeWhite;
} RGBConversionState;

/*  D50, reference sandpiper */
#define  D50_X    0.964
#define  D50_Z    0.824
#define  TABLELEN 2048
#define  CTHRESH  0.008856

#define  FUNC_INV( y )   (( y<0.206892706 ) ? ( y-0.137931034)/7.787 \
			  : (y * y *y) )
#define  LIMIT( x, min, max )      (( x < min ) ? min : ( (x > max) ? \
						max : x ) )

typedef struct
{
    Int32    r_X[256], r_Y[256], r_Z[256];
    Int32    g_X[256], g_Y[256], g_Z[256];
    Int32    b_X[256], b_Y[256], b_Z[256];
    Float64  tabPow[TABLELEN + 10];
    Int32    l[TABLELEN + 10];
}RGB2LABTable;

/******************************************************************************
 *
 * i_RGBToY
 *
 *	Convert an image consisting of three RGB color planes to a
 *	luminance channel. 
 *
 * Arguments:
 *      pIPState  =             Pointer to standard state block. 
 *	pRed, pGreen, pBlue =	Pointers to the image data, including frame
 *	sBpl =			Bytes/line in the R, G, and B planes
 *	pLum =			Place to put resulting plane
 *	dBpl =			Bytes/line in resulting plane
 *	w =			Width of a plane, including frame, in pixels
 *	h =			Height of a plane, including frame, in pixels
 *
 * Returns:
 *
 *	ia_successful, ...
 *******************************************************************************/

ErrCode CDECL 
i_RGBToY (
    IPState    *pIPState,
    register UInt8 *pRed,
    register UInt8 *pGreen,
    register UInt8 *pBlue,
    Int32	sBpl,
    register UInt8 *pLum,
    Int32	dBpl,
    Int32	w,
    Int32	h)
{
    PROLOGUE;
    register Int32 j, lum;
    register UInt8 *l1, *l2, *l3;
    int i;


    /* Move past the frame */
    pRed += (sBpl << cLogFrameBits) + cFrameBits;
    pGreen += (sBpl << cLogFrameBits) + cFrameBits;
    pBlue += (sBpl << cLogFrameBits) + cFrameBits;
    pLum += (dBpl << cLogFrameBits) + cFrameBits;

    l1 = &redLut[0];
    l2 = &greenLut[0];
    l3 = &blueLut[0];
    h -= 2*cFrameBits;
    w -= 2*cFrameBits;

    /* If we go past the edge of the picture into the frame,
	we shouldn't need to worry about byte ordering */
    if (w & 3)
	w += 4 - (w & 3);

    for (i = 0; i < h; i++)
    {
	for (j = 0; j < w; j++)
	{
	    lum = l1[pRed[j]] + l2[pGreen[j]] + l3[pBlue[j]];
	    pLum[j] = (UInt8)lum;
	}

	pRed += sBpl;
	pGreen += sBpl;
	pBlue += sBpl;
	pLum += dBpl;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 *
 * i_RGBToY2
 *
 *	Convert an image consisting of three RGB color planes to a
 *	luminance channel, using DITC's formula.
 *
 * Arguments:
 *	pRed, pGreen, pBlue =	Pointers to the image data, including frame
 *	sBpl =			Bytes/line in the R, G, and B planes
 *	pLum =			Place to put resulting plane
 *	dBpl =			Bytes/line in resulting plane
 *	w =			Width of a plane, including frame, in pixels
 *	h =			Height of a plane, including frame, in pixels
 *
 * Returns:
 *
 *	ia_successful, ...
 *******************************************************************************/

ErrCode CDECL 
i_RGBToY2 (
    IPState    *pIPState,
    register UInt8 *pRed,
    register UInt8 *pGreen,
    register UInt8 *pBlue,
    Int32	sBpl,
    register UInt8 *pLum,
    Int32	dBpl,
    Int32	w,
    Int32	h)
{
    PROLOGUE;
    register Int32 j, lum;
    register UInt8 *l1, *l2, *l3;
    int i;


    /* Move past the frame */
    pRed += (sBpl << cLogFrameBits) + cFrameBits;
    pGreen += (sBpl << cLogFrameBits) + cFrameBits;
    pBlue += (sBpl << cLogFrameBits) + cFrameBits;
    pLum += (dBpl << cLogFrameBits) + cFrameBits;

    l1 = &redLut2[0];
    l2 = &greenLut2[0];
    l3 = &blueLut2[0];
    h -= 2*cFrameBits;
    w -= 2*cFrameBits;

    /* If we go past the edge of the picture into the frame,
	we shouldn't need to worry about byte ordering */
    if (w & 3)
	w += 4 - (w & 3);

    for (i = 0; i < h; i++)
    {
	for (j = 0; j < w; j++)
	{
	    /* Compute the near equivalent to:
	     *  lum = (unsigned char)((double)pRed[j] * 0.252939 +
	     *		(double)pGreen[j] * 0.684458 +
	     *		(double)pBlue[j] * 0.062603 + 0.5);
	     */
	    lum = l1[pRed[j]] + l2[pGreen[j]] + l3[pBlue[j]];
	    pLum[j] = (UInt8)lum;
	}

	pRed += sBpl;
	pGreen += sBpl;
	pBlue += sBpl;
	pLum += dBpl;
    }

    NORMAL_RETURN;
}


/************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *	Computes close approximation of RGB->Y transform.
 *
 * LongDescription:::
 *	The correct formula is:  Y = 0.299R + 0.587G + 0.114B and the
 *	approximation is: Y = 0.247R + 0.620G + 0.122B which saturates
 *	at 252 rather than 255.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPState::    Pointer to standard state block. 
 *	pRed::    Pointer to red channel source image including all frames
 *	pGreen::  Pointer to green channel source image including all frames
 *	pBlue::   Pointer to blue channel source image including all frames
 *	sBpl::    Bytes per line of the source image (must be %4)
 *	dBpl::    Bytes per line of the dest image (must be %4)
 *	w::       Width of image including frames
 *	h::       Height of image including frames
 *
 * Output:::
 *	pLum::    Pointer to dest image including all frames
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *	Source and dest images must be the same width and height.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *  See any book on color or image processing which covers the RGB->Y
 *  transform for the correct coefficients.
 *
 * EndDocumentation:::
 ************************************************************************/
ErrCode CDECL 
i_RGBToYFast (
    IPState    *pIPState,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    Int32	sBpl,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    register UInt8 *pLum)
{
    PROLOGUE;
    Int32  row, col, wpls,
           widthInWords;      /* width in words */
    UInt32 redWord, 
           greenWord, 
	   blueWord,
	   yWord;
    UInt32 *ispR, *ispG, *ispB, *idp;


    /* point to the image data inside the frames */
    pRed += (sBpl << cLogFrameBits) + cFrameBits;
    pGreen += (sBpl << cLogFrameBits) + cFrameBits;
    pBlue += (sBpl << cLogFrameBits) + cFrameBits;
    pLum += (dBpl << cLogFrameBits) + cFrameBits;

    /* get the unframed width & height */
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

    /* since we're going to read and write only full words, calculate
     * the number of words in a row which have any real (non-frame)
     * pixels in them as well as the stride in words.
     */
    wpls = sBpl / 4;
    widthInWords = (w + 3) / 4;

    for (row = 0; row < h; row++)
    {
        /* increment the source and dest pointers to the next row */
	ispR = (UInt32 *)(pRed + (row * sBpl));
	ispG = (UInt32 *)(pGreen + (row * sBpl));
	ispB = (UInt32 *)(pBlue + (row * sBpl));
	idp = (UInt32 *)(pLum + (row * dBpl));

        for (col = 0; col < widthInWords; col++)
	{
	    /* read a word of RGB data */
	    redWord = ULONG_ACCESS (ispR);
	    greenWord = ULONG_ACCESS (ispG);
	    blueWord = ULONG_ACCESS (ispB);

            /* compute approximate Y */
	    yWord = (redWord & 0xfcfcfcfc) >> 2;
	    greenWord = (greenWord & 0xfefefefe) >> 1;
	    greenWord += (greenWord & 0xfcfcfcfc) >> 2;
	    yWord += greenWord + ((blueWord & 0xf8f8f8f8) >> 3);

	    /* write out the intensity word */
	    ULONG_ACCESS (idp) = yWord;

	    idp++;
	    ispR++;
	    ispG++;
	    ispB++;

	}

	IP_YIELD (aborted);
    }

    /* either we bomb or it worked */
    NORMAL_RETURN;

aborted:
    return (ia_aborted);

}  /* i_RGBToYFast () */


/* PUBLIC */
/*
 * i_indexed8ToRGB ()
 *
 *	Convert an 8-bit pseudocolor image to three RGB planes.
 *
 * Arguments:
 *	pSrc	-	Pointer to upper left corner of source image's frame
 *	x, y	-	Upper left crop point, relative to frame (0, 0)
 *	w, h	-	Size of image
 *	sBpl	-	Bytes between lines of source image
 *	pRed	-	Pointer to red plane frame (0, 0)
 *	pGreen	-	Pointer to green plane frame (0, 0)
 *	pBlue	-	Pointer to blue plane frame (0, 0)
 *	dBpl	-	Bytes between lines in destination planes
 *	pCTable	-	Color table, organized as (r, g, b, 0), 4 bytes per entry, 256 entries
 * 
 * Returns:
 *	ia_succesful, ...
 */

Int32 CDECL 
i_indexed8ToRGB (
    IPState    *pIPState,
    UInt8      *pSrc,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    Int32	dBpl,
    UInt8      *cTable)
{
	     PROLOGUE;
	     Int32   i;
    register Int32   j;
    register UInt8  *pTable;
    

    /* Move the source pointer to the right place */
    pSrc += (y * sBpl) + x;

    /* Move the destination pointers past the frame */
    pRed   += (dBpl << cLogFrameBits) + cFrameBits;
    pGreen += (dBpl << cLogFrameBits) + cFrameBits;
    pBlue  += (dBpl << cLogFrameBits) + cFrameBits;

    for (i = 0; i < h; i++)
    {
	for (j = 0; j < w; j++)
	{
	    pTable = cTable + (UCHAR_ACCESS(pSrc + j) << 2);
	    UCHAR_ACCESS (pRed + j) = pTable[0];
	    UCHAR_ACCESS (pGreen + j) = pTable[1];
	    UCHAR_ACCESS (pBlue + j) = pTable[2];
	}

	pRed += dBpl;
	pGreen += dBpl;
	pBlue += dBpl;
	pSrc += sBpl;
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/*
 * i_indexed4ToRGB ()
 *
 *	Convert a 4-bit pseudocolor image to three RGB planes.
 *	*warning*: doesn't handle odd image offsets!
 *
 * Arguments:
 *	pSrc	-	Pointer to upper left corner of source image's frame
 *	x, y	-	Upper left crop point, relative to frame (0, 0)
 *	w, h	-	Size of image
 *	sBpl	-	Bytes between lines of source image
 *	pRed	-	Pointer to red plane frame (0, 0)
 *	pGreen	-	Pointer to green plane frame (0, 0)
 *	pBlue	-	Pointer to blue plane frame (0, 0)
 *	dBpl	-	Bytes between lines in destination planes
 *	pCTable	-	Color table, organized as (r, g, b, 0), 4 bytes per entry, 256 entries
 * 
 * Returns:
 *	ia_succesful, ...
 */

Int32 CDECL 
i_indexed4ToRGB (
    IPState    *pIPState,
    UInt8      *pSrc,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    Int32	dBpl,
    UInt8      *cTable)
{
	     PROLOGUE;
	     Int32   i, offset, k, odd;
    register Int32   j;
    register UInt8  *pTable;
             UInt8   compositeTable[256][6];

    /* Make a composite table */
    for (i = 0; i < 16; i++)
    {
	for (j = 0; j < 16; j++)
	{
	    offset = (i << 4) + j;

	    for (k = 0; k < 3; k++)
	    {
	        compositeTable[offset][k] = cTable[i*4 + k];
	        compositeTable[offset][k + 3] = cTable[j*4 + k];
	    }
	}
    }

    /* Move the source pointer to the right place */
    pSrc += (y * sBpl) + (x / 2);

    /* Move the destination pointers past the frame */
    pRed   += (dBpl << cLogFrameBits) + cFrameBits;
    pGreen += (dBpl << cLogFrameBits) + cFrameBits;
    pBlue  += (dBpl << cLogFrameBits) + cFrameBits;

    /* If width is odd, we'll have to handle an extra nibble alone */
    odd = w & 1;
    w /= 2;

    for (i = 0; i < h; i++)
    {
	for (j = 0, k = 0; j < w; j++)
	{
	    pTable = &(compositeTable[UCHAR_ACCESS(pSrc + j)][0]);
	    UCHAR_ACCESS (pRed + k) = pTable[0];
	    UCHAR_ACCESS (pGreen + k) = pTable[1];
	    UCHAR_ACCESS (pBlue + k++) = pTable[2];
	    UCHAR_ACCESS (pRed + k) = pTable[3];
	    UCHAR_ACCESS (pGreen + k) = pTable[4];
	    UCHAR_ACCESS (pBlue + k++) = pTable[5];
	}

	/* Handle the last nibble, if necessary.  We get the
	 * last byte and mask off the top nibble, which gives
	 * us index * 16.  There are only 4 entries per color
	 * in cTable, so we divide by 4 to get index * 4.
	 */
	if (odd)
	{
	    pTable = cTable + (UCHAR_ACCESS(pSrc + j) & 0xf0) / 4;
	    UCHAR_ACCESS (pRed + k) = pTable[0];
	    UCHAR_ACCESS (pGreen + k) = pTable[1];
	    UCHAR_ACCESS (pBlue + k) = pTable[2];
	}

	pRed += dBpl;
	pGreen += dBpl;
	pBlue += dBpl;
	pSrc += sBpl;
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/****************************************************************************
 * i_RGBToLuvInit
 *
 *	Initialize RGB To luminance conversion.
 * 
 * Arguments:
 *	X, Y, Z	    -	   XYZ of white point. If zero, uses D50.
 *      state       -      internal state pointer for conversion
 *
 * Returns:
 *	ia_successful or ia_nomem
 ****************************************************************************/

Int32 CDECL 
i_RGBToLuvInit (
    IPState    *pIPState,
    Float64	x,
    Float64	y,
    Float64	z,
    void      **state)
{
    PROLOGUE;
    Float64		 d;
    RGBConversionState 	*s;

    *state = (void *)MALLOC(sizeof(RGBConversionState));
    if (!*state)
        return ia_nomem;
    s = (RGBConversionState *) *state;

    if (x == 0.0 && y == 0.0 && z == 0.0)
    {
	/* Just use D50 */
	x = 0.9640;
	y = 1.0000;
	z = 0.8246;
    }

    /* Compute u'white, v'white */
    d = x + 15.0 * y + 3.0 * z;
    s->uPrimeWhite = (4.0 * x) / d;
    s->vPrimeWhite = (9.0 * y) / d;

    /* Remember Y */
    s->yWhite = y;

    NORMAL_RETURN;
}

/* PUBLIC */
/****************************************************************************
 * i_RGBToLuv
 *
 *	Convert RGB to L*u*v*. This is Very Slow. It's only
 *	intended use is to initialize a table. If you pass it 
 *	image data, you get what you deserve. 
 * 
 *	Most of the constants came out of Sandpiper's cspace3.c. I
 *	haven't really checked them very thouroughly for accuracy,
 *	so be careful.
 *
 * Arguments:
 *	r, g, b		    -	Input data
 *      state               -   internal state pointer for conversion
 *	L, u, v		    -	[out] Output data
 *
 * Returns:
 *	status: ia_successful or some error code
 ****************************************************************************/

Int32 CDECL 
i_RGBToLuv (
    IPState    *pIPState,
    Int32	r,
    Int32	g,
    Int32	b,
    void       *state,
    Float64    *L,
    Float64    *u,
    Float64    *v)
{
    PROLOGUE;
    Float64		 x, y, z, d, lum, yRatio, uPrime, vPrime;
    RGBConversionState 	*s;

    s = (RGBConversionState *) state;

    /* Convert to XYZ. A look-up would be faster, but.. */
    x = (0.468682 * r + 0.356608 * g + 0.138620 * b) / 255.0;
    y = (0.252939 * r + 0.684458 * g + 0.062603 * b) / 255.0;
    z = (0.022318 * r + 0.109283 * g + 0.693101 * b) / 255.0;

    /* Compute a denominator for the pixel */
    d = x + 15.0 * y + 3.0 * z;

    /* Compute luminance */
    yRatio = y / s->yWhite;
    if (yRatio > .01)
	lum = 116.0 * pow(yRatio, 0.333333333) - 16.0;
    else
	lum = 903.3 * yRatio;

    /* Compute u' and v' */
    if (d)
    {
	uPrime = (4.0 * x) / d;
	vPrime = (9.0 * y) / d;
    }
    else
    {
	uPrime = vPrime = 0.0;
    }

    /* Compute L*, u*, and  */
    *L = lum;
    *u = 13.0 * lum * (uPrime - s->uPrimeWhite); 
    *v = 13.0 * lum * (vPrime - s->vPrimeWhite);

    NORMAL_RETURN;
}

/* PUBLIC */
/****************************************************************************
 * i_RGBToLuvFinish
 *
 *	Finish RGB To luminance conversion by freeing internal structures.
 * 
 * Arguments:
 *      state       -      internal state pointer for conversion
 *
 * Returns:
 *      ia_successful
 ****************************************************************************/

Int32 CDECL 
i_RGBToLuvFinish (
    IPState    *pIPState,
    void       *state)
{
    PROLOGUE;
    RGBConversionState *s;

    s = (RGBConversionState *) state;

    FREE(s);
    NORMAL_RETURN;
}


/************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Initialize RGB To Lab conversion..
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPState::          Pointer to standard state block. 
 *	x::		XYZ of white point. If zero, uses D50.
 *	y::		XYZ of white point. If zero, uses D50.
 *	z::		XYZ of white point. If zero, uses D50.
 *
 * Output:::
 *	state::		internal state pointer for conversion
 *
 * Returns:::
 *	ia_successful, ia_nomem, ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig
 *
 * EndDocumentation:::
 ************************************************************************/
ErrCode CDECL 
i_RGBToLabInit (
    IPState    *pIPState,
    Float64	x,
    Float64	y,
    Float64	z,
    void      **state)
{
    PROLOGUE;
    RGBConversionState 	*s;

    if (x == 0.0 && y == 0.0 && z == 0.0)
    {
	/* Just use D50 */
	x = 96.42;
	y = 100.0;
	z = 82.49;
    }
    if (x == 0.0 || y == 0.0 || z == 0.0)
	return ia_invalidParm;

    *state = (void *)MALLOC(sizeof(RGBConversionState));
    if (!*state)
        return ia_nomem;
    s = (RGBConversionState *) *state;

    /* Remember white */
    s->xWhite = x;
    s->yWhite = y;
    s->zWhite = z;

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a single [R,G,B] triplet to a single [L*a*b*] triplet.
 *
 * LongDescription:::
 *  This routine converts a single [R,G,B] triplet to a single
 *  [L*a*b*] triplet.  This is Very Slow. It's only
 *  intended use is to initialize a table. If you pass it 
 *  image data, you get what you deserve.
 *  30-01-98 
 *  Changed RGB->XYZ transform matrix in order to be same in
 *  all routines. We copy the transform matrix from sandpiper. It is
 *  good for CRT. The old matrix is CIE XYZ primaries.
 *  0.49 , 0.31, 0.2 ; 0.1769, 0.8124, 0.01063; 0, 0.01, 0.99
 *
 * Input:::
 *  pIPState::       Pointer to standard state block. 
 *  r::              Red intensity, between 0 and 255.
 *  g::              Green intensity, between 0 and 255.
 *  b::              Blue intensity, between 0 and 255.
 *  state::          Pointer to private state block allocated by
 *		     w_RGBToLabInit.
 *
 * Output:::
 *  L::              L* portion of output pixel.
 *  aStar::          a* portion of output pixel.
 *  bStar::          b* portion of output pixel.
 *
 * Returns:::
 *	ia_successful
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  JCraig
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
i_RGBToLab (
    IPState    *pIPState,
    Int32	r,
    Int32	g,
    Int32	b,
    void       *state,
    Float64    *L,
    Float64    *aStar,
    Float64    *bStar)
{
    PROLOGUE;
    Float64 		x, y, z;
    Float64		xRatio, yRatio, zRatio, yPower, xPower, zPower;
    RGBConversionState *s;

    s = (RGBConversionState *) state;

    /* Convert to XYZ. A look-up would be faster, but.. */
    x = 0.468682 * r + 0.356608 * g + 0.138620 * b;
    y = 0.252939 * r + 0.684458 * g + 0.062603 * b;
    z = 0.022318 * r + 0.109283 * g + 0.693103 * b;

    /* Compute luminance */
    yRatio = y / s->yWhite;
    yPower =  pow(yRatio, 0.333333333);
    if (yRatio > CTHRESH)
	*L = 116.0 * yPower - 16.0;
    else
	*L = 903.3 * yRatio;

    if (yRatio <= CTHRESH)
    {
	yRatio = (7.787 * yRatio) + (16.0 / 116.0);
	yPower =  pow(yRatio, 0.333333333);
    }
    if ((xRatio = x / s->xWhite) <= 0.01)
	xRatio = (7.787 * xRatio) + (16.0 / 116.0);
    if ((zRatio = z / s->zWhite) <= 0.01)
	zRatio = (7.787 * zRatio) + (16.0 / 116.0);

    /* now compute a and b */
    xPower = pow(xRatio, 0.333333333);
    *aStar = 500 * (xPower - yPower);
    zPower = pow(zRatio, 0.333333333);
    *bStar = 200 * (yPower - zPower);

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Deallocated the internal structures used to convert RGB to L*a*b*
 *
 * LongDescription:::
 *  w_RGBToLab uses a private state block to convert one RGB pixel to
 *  one L*a*b* pixel.  That routine is likely to be called a lot.
 *  This routine is called by clients to free the private state block
 *  when finished.
 *
 * Input:::
 *
 * Output:::
 *  pIPState::      Pointer to standard state block. 
 *  state::          Pointer to pointer to private state block that will
 *		     be used during the calculation.  **state
 *		     (and associated memory blocks) will be freed.
 *
 * Returns:::
 *	ia_successful
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  JCraig
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
i_RGBToLabFinish (
    IPState    *pIPState,
    void       *state)
{
    PROLOGUE;
    RGBConversionState *s;

    s = (RGBConversionState *) state;

    FREE(s);
    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert RGB image to Lab image.
 *
 * LongDescription:::
 *  This routine converts buffer of RGB image to buffer of Lab image. The
 *  dest image buffer must be equal to or greater than source image buffer.
 *  If image is very small ( < 2048), don't use this function, because 
 *  setuping the tables consumes more time than color transform.
 *
 * Input:::
 *  pIPState::      Pointer to standard state block. 
 *  pRed::          Pointer to red image buffer to be converted. 
 *  pGreen::        Pointer to green image buffer to be converted. 
 *  pBlue::         Pointer to blue image buffer to be converted. 
 *  bplS::          Source (RGB) image bytes per line.
 *  bplD::          Dest. (Lab) image bytes per line.
 *  width::         Source (RGB) image width. 
 *  height::        Source (RGB) image height.
 *
 * Output:::
 *  pL::            Pointer to L image buffer of dest image.
 *  pA::            Pointer to a image buffer of dest image.
 *  pB::            Pointer to b image buffer of dest image.
 *
 * Returns:::
 *	ia_successful
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
i_RGB2Lab (
    IPState    *pIPState,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    Int32       bplS,
    Int32       bplD,
    Int32       width,
    Int32       height,
    UInt8      *pL,
    UInt8      *pA,
    UInt8      *pB )
{
    PROLOGUE;
    Int32    lutLen, idx, idxH, idxW;
    Float64  val;
    Int32    *r_X, *r_Y, *r_Z, *g_X, *g_Y, *g_Z, *b_X, *b_Y, *b_Z;
    Float64  c_rx = 0.468682, c_ry = 0.252939, c_rz = 0.022318;
    Float64  c_gx = 0.356608, c_gy = 0.684458, c_gz = 0.109283;
    Float64  c_bx = 0.138620, c_by = 0.062603, c_bz = 0.693103;
    Int32    red, green, blue, x, y, z, thre, l, a, b;
    Float64  yy;
    Float64  *tabPow;

    lutLen = 256 * 9 * sizeof(Int32);
    r_X = (Int32 *)MALLOC(lutLen);
    if (r_X ==  NULL) return ia_nomem;
    lutLen = 256;
    r_Y = r_X + lutLen;
    r_Z = r_Y + lutLen;

    g_X = r_Z + lutLen;
    g_Y = g_X + lutLen;
    g_Z = g_Y + lutLen;

    b_X = g_Z + lutLen;
    b_Y = b_X + lutLen;
    b_Z = b_Y + lutLen;

    for(idx=0; idx<256; ++idx)
    {
	val = (Float64)idx / 255.0;
	r_X[idx] = (Int32)(val * c_rx * (Float64)TABLELEN /(Float64) D50_X);
	r_Y[idx] = (Int32)(val * c_ry * (Float64)TABLELEN) ;
	r_Z[idx] = (Int32)(val * c_rz * (Float64)TABLELEN /(Float64) D50_Z);
	g_X[idx] = (Int32)(val * c_gx * (Float64)TABLELEN /(Float64) D50_X);
	g_Y[idx] = (Int32)(val * c_gy * (Float64)TABLELEN);
  	g_Z[idx] = (Int32)(val * c_gz * (Float64)TABLELEN /(Float64) D50_Z);
	b_X[idx] = (Int32)(val * c_bx * (Float64)TABLELEN /(Float64) D50_X);
	b_Y[idx] = (Int32)(val * c_by * (Float64)TABLELEN);
	b_Z[idx] = (Int32)(val * c_bz * (Float64)TABLELEN /(Float64) D50_Z); 
    }

    lutLen = (TABLELEN+ 10)* sizeof(Float64);
    tabPow = (Float64 *)MALLOC(lutLen);
    if (tabPow ==  NULL) return ia_nomem;

    thre = (Int32) ((Float32)TABLELEN * (Float32)CTHRESH);
    for(idx=0; idx<thre; ++idx)
    {
	val = (Float64)idx / (Float64)TABLELEN;
	tabPow[idx] = 7.787 * val + 0.137931034;
    }
    for(idx=thre; idx<= (TABLELEN+2); ++idx)
    {
	val = (Float64)idx / (Float64)TABLELEN;
	tabPow[idx] = pow(val, 0.3333);
    }

/*  RGB to Lab */    
    for(idxH=0; idxH<height; ++idxH)
    {
	for(idxW=0; idxW<width; ++idxW)
	{
	    red = (Int32) UCHAR_ACCESS(pRed +idxW);
	    green = (Int32) UCHAR_ACCESS(pGreen +idxW);
	    blue = (Int32) UCHAR_ACCESS(pBlue +idxW);
	    x = r_X[red] + g_X[green] + b_X[blue];
	    y = r_Y[red] + g_Y[green] + b_Y[blue];
	    z = r_Z[red] + g_Z[green] + b_Z[blue];
	    yy = tabPow[y];
	    l = (Int32)( 295.8 * yy - 40.3);
	    a = (Int32)( 500.0 * (tabPow[x] - yy) + 128.5);
	    b = (Int32)( 200.0 * (yy - tabPow[z]) + 128.5);
	    UCHAR_ACCESS(pL +idxW) = (UInt8)LIMIT(l, 0, 255);
	    UCHAR_ACCESS(pA +idxW) = (UInt8)LIMIT(a, 0, 255);
	    UCHAR_ACCESS(pB +idxW) = (UInt8)LIMIT(b, 0, 255); 
	}
	pRed += bplS;
	pGreen += bplS;
	pBlue += bplS;
	pL += bplD;
	pA += bplD;
	pB += bplD;
    }

    FREE(r_X);
    FREE(tabPow);

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert Lab image to RGB image.
 *
 * LongDescription:::
 *  This routine converts buffer of Lab image to buffer of RGB image. The
 *  dest image buffer must be equal to or greater than source image buffer.
 *  This is i_RGB2Lab's reverse transform.
 *
 * Input:::
 *  pIPState::      Pointer to standard state block. 
 *  pL::            Pointer to L image buffer to be converted. 
 *  pA::            Pointer to a image buffer to be converted. 
 *  pB::            Pointer to b image buffer to be converted. 
 *  bplS::          Source (Lab) image bytes per line.
 *  bplD::          Dest. (RGB) image bytes per line.
 *  width::         Source (Lab) image width. 
 *  height::        Source (Lab) image height.
 *
 * Output:::
 *  pRed::          Pointer to red image buffer of dest image.
 *  pGreen::        Pointer to green image buffer of dest image.
 *  pBlue::         Pointer to blue image buffer of dest image.
 *
 * Returns:::
 *	ia_successful
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
i_Lab2RGB (
    IPState    *pIPState,
    UInt8      *pL,
    UInt8      *pA,
    UInt8      *pB,
    Int32       bplS,
    Int32       bplD,
    Int32       width,
    Int32       height,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue )
{
    PROLOGUE;
    Int32    lutLen, idx, idxH, idxW;
    Float64  val;
    Int32   *fx, *fy, *fz, *ffy1, *ffy2, *ffy3;
    Int32   *ffx1, *ffx2, *ffx3, *ffz1, *ffz2, *ffz3;
    Int32    l, a, b, x, z, tmp;
    Float64  y, yy;

    lutLen = 256 * 6 * sizeof(Int32);
    fx = (Int32 *)MALLOC(lutLen);
    if (fx ==  NULL) return ia_nomem;

    lutLen = 256;
    fy = fx + lutLen;
    fz = fy + lutLen;
    ffy1 = fz + lutLen;
    ffy2 = ffy1 + lutLen;
    ffy3 = ffy2 + lutLen;

    for(idx=0; idx<256; ++idx)
    {
	val = (Float64)idx;
	y = (val + 40.8)/ 295.8;
	fy[idx] = (Int32)((val + 40.8) * (Float64)TABLELEN / 295.8); 
	fx[idx] = (Int32)((val - 128.0) * (Float64)TABLELEN / 500.0);
	fz[idx] = (Int32)((val - 128.0) * (Float64)TABLELEN / 200.0);
	yy = FUNC_INV(y);
	ffy1[idx] = (Int32)((Float64)372.500685 * LIMIT( yy, 0.0, 1.0) + 0.4); 
	ffy2[idx] = (Int32)((Float64)516.566505 * LIMIT( yy, 0.0, 1.0) + 0.4);
	ffy3[idx] = (Int32)((Float64)69.45384 * LIMIT( yy, 0.0, 1.0) + 0.4);
    }

    lutLen = (TABLELEN +2 ) * 6 * sizeof(Int32);
    ffx1 = (Int32 *)MALLOC(lutLen);
    if (ffx1 ==  NULL) return ia_nomem;

    lutLen = TABLELEN +2;
    ffx2 = ffx1 + lutLen;
    ffx3 = ffx2 + lutLen;
    ffz1 = ffx3 + lutLen;
    ffz2 = ffz1 + lutLen;
    ffz3 = ffz2 + lutLen;

    for(idx=0; idx<=TABLELEN; ++idx)
    {
	val = (Float64)idx/(Float64)TABLELEN;
	y = FUNC_INV(val);
	yy = (Float64) D50_X * LIMIT( y, 0.0, 1.0);
	ffx1[idx] = (Int32)((Float64)750.65829 * yy + 0.4);
	ffx2[idx] = (Int32)((Float64)279.21939 * yy + 0.4);
	ffx3[idx] = (Int32)((Float64)19.85379 * yy + 0.4);
	yy = (Float64) D50_Z * LIMIT( y, 0.0, 1.0);
	ffz1[idx] = (Int32)((Float64)116.48655 * yy + 0.4);
	ffz2[idx] = (Int32)((Float64)9.186375 * yy + 0.4);
	ffz3[idx] = (Int32)((Float64)370.214355 * yy + 0.4);
    }

/* Lab to RGB */
    for(idxH=0; idxH<height; ++idxH)
    {
	for(idxW=0; idxW<width; ++idxW)
	{
	    idx = (Int32)UCHAR_ACCESS(pL+idxW); 
	    l = fy[idx];
	    a = fx[UCHAR_ACCESS(pA +idxW)];
	    b = fz[UCHAR_ACCESS(pB +idxW)];
	    tmp = l+a;
	    x = LIMIT( tmp, 0, TABLELEN);
	    tmp = l -b;
	    z = LIMIT( tmp, 0, TABLELEN);
	    l = ffx1[x] - ffy1[idx] - ffz1[z];
	    a =-ffx2[x] + ffy2[idx] + ffz2[z];
	    b = ffx3[x] - ffy3[idx] + ffz3[z];
	    UCHAR_ACCESS(pRed + idxW) = (UInt8)LIMIT(l, 0, 255);
	    UCHAR_ACCESS(pGreen + idxW) = (UInt8)LIMIT(a, 0, 255);
	    UCHAR_ACCESS(pBlue + idxW) = (UInt8)LIMIT(b, 0, 255);
	}
	pL += bplS;
	pA += bplS;
	pB += bplS;
	pRed += bplD;
	pGreen += bplD;
	pBlue += bplD;
    }

    FREE( fx );
    FREE( ffx1 );
    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a single [L*a*b*] triplet to a single [R,G,B] triplet.
 *
 * LongDescription:::
 *  Convert a single [L*a*b*] triplet to a single [R,G,B] triplet.
 *
 * Input:::
 *  pIPState::      Pointer to standard state block. 
 *  L::             L to be converted. 
 *  A::             a to be converted. 
 *  B::             b to be converted. 
 *
 * Output:::
 *  pRed::          Pointer to red.
 *  pGreen::        Pointer to green.
 *  pBlue::         Pointer to blue.
 *
 * Returns:::
 *	ia_successful
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
i_LabToRGB (
    IPState    *pIPState,
    UInt8       L,
    UInt8       A,
    UInt8       B,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue )
{
    PROLOGUE;
    Float64  val;
    Int32    ffy1, ffy2, ffy3, ffx1, ffx2, ffx3, ffz1, ffz2, ffz3;
    Float64  y, l, a, b, yy;
    Int32    ll, ca, cb;

/* Lab to RGB */
    val = (Float64) L;
    l = (val + 40.8) / 295.8; 
    val = (Float64) A;
    a = (val - 128.0) / 500.0;
    val = (Float64) B;
    b = (val - 128.0) / 200.0;

    yy = FUNC_INV(l);
    ffy1 = (Int32)((Float64)372.500685 * LIMIT( yy, 0.0, 1.0) + 0.4); 
    ffy2 = (Int32)((Float64)516.566505 * LIMIT( yy, 0.0, 1.0) + 0.4);
    ffy3 = (Int32)((Float64)69.45384 * LIMIT( yy, 0.0, 1.0) + 0.4);
    val = l+a;
    y = FUNC_INV(val);
    yy = (Float64) D50_X * LIMIT( y, 0.0, 1.0);
    ffx1 = (Int32)((Float64)750.65829 * yy + 0.4);
    ffx2 = (Int32)((Float64)279.21939 * yy + 0.4);
    ffx3 = (Int32)((Float64)19.85379 * yy + 0.4);
    val = l-b;
    y = FUNC_INV(val);
    yy = (Float64) D50_Z * LIMIT( y, 0.0, 1.0);
    ffz1 = (Int32)((Float64)116.48655 * yy + 0.4);
    ffz2 = (Int32)((Float64)9.186375 * yy + 0.4);
    ffz3 = (Int32)((Float64)370.214355 * yy + 0.4);


    ll = ffx1 - ffy1 - ffz1;
    ca =-ffx2 + ffy2 + ffz2;
    cb = ffx3 - ffy3 + ffz3;
    *pRed = (UInt8)LIMIT(ll, 0, 255);
    *pGreen = (UInt8)LIMIT(ca, 0, 255);
    *pBlue = (UInt8)LIMIT(cb, 0, 255);

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create table for converting RGB image to Lab image.
 *
 * LongDescription:::
 *  This routine create table for converting RGB to Lab. 
 *
 * Input:::
 *  pIPState::      Pointer to standard state block. 
 *
 * Output:::
 *  ppTable::       Point to pointer of table, *ppTable must be NULL. 
 *
 * Returns:::
 *	ia_successful
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
i_createRGB2LabTable (
    IPState    *pIPState,
    void      **ppTable )
{
    PROLOGUE;
    Int32    idx, thre;
    Float64  val, len;
    Float64  c_rx = 0.468682, c_ry = 0.252939, c_rz = 0.022318;
    Float64  c_gx = 0.356608, c_gy = 0.684458, c_gz = 0.109283;
    Float64  c_bx = 0.138620, c_by = 0.062603, c_bz = 0.693103;
    RGB2LABTable   *pTab;

    pTab = (RGB2LABTable *) MALLOC(sizeof(RGB2LABTable));
    if (pTab ==  NULL) return ia_nomem;

    len = (Float64)TABLELEN;
    for(idx=0; idx<256; ++idx)
    {
	val = (Float64)idx / 255.0;
	pTab->r_X[idx] = (Int32)(val * c_rx * len/ (Float64) D50_X);
	pTab->r_Y[idx] = (Int32)(val * c_ry * len);
	pTab->r_Z[idx] = (Int32)(val * c_rz * len / (Float64) D50_Z);
	pTab->g_X[idx] = (Int32)(val * c_gx * len / (Float64) D50_X);
	pTab->g_Y[idx] = (Int32)(val * c_gy * len);
  	pTab->g_Z[idx] = (Int32)(val * c_gz * len / (Float64) D50_Z);
	pTab->b_X[idx] = (Int32)(val * c_bx * len / (Float64) D50_X);
	pTab->b_Y[idx] = (Int32)(val * c_by * len);
	pTab->b_Z[idx] = (Int32)(val * c_bz * len / (Float64) D50_Z); 
    }

    thre = (Int32) ((Float32)TABLELEN * (Float32)CTHRESH);
    for(idx=0; idx<thre; ++idx)
    {
	val = (Float64)idx / len;
	pTab->tabPow[idx] = 7.787 * val + 0.137931034;
    }
    for(idx=thre; idx<= (TABLELEN+4); ++idx)
    {
	val = (Float64)idx / len;
	pTab->tabPow[idx] = pow(val, 0.3333);
	pTab->l[idx] = (Int32)( 295.8 * pTab->tabPow[idx] - 40.3);
    }

    *ppTable = (void *)pTab;

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert RGB image to Lab image by use of RGB2LABTable.
 *
 * LongDescription:::
 *  This routine converts buffer of RGB image to buffer of Lab image by
 *  use of RGB2LABTable created in i_createRGB2LabTable. If there is
 *  mask image, it just convert the pixels whose position are '1' in 
 *  mask image. 
 *
 * Input:::
 *  pIPState::      Pointer to standard state block. 
 *  pTable::        Pointer to RGB2LABTable.
 *  pRed::          Pointer to red image buffer to be converted. 
 *  pGreen::        Pointer to green image buffer to be converted. 
 *  pBlue::         Pointer to blue image buffer to be converted.
 *  pMask::         Pointer to mask image buffer. 
 *  bplS::          Source (RGB) image bytes per line.
 *  bplD::          Dest. (Lab) image bytes per line.
 *  bplM::          Mask image bytes per line.
 *  width::         Source (RGB) image width. 
 *  height::        Source (RGB) image height.
 *
 * Output:::
 *  pL::            Pointer to L image buffer of dest image.
 *  pA::            Pointer to a image buffer of dest image.
 *  pB::            Pointer to b image buffer of dest image.
 *
 * Returns:::
 *	ia_successful
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
i_RGB2LabUseTable (
    IPState    *pIPState,
    void       *pTable,
    UInt8      *pRed,
    UInt8      *pGreen,
    UInt8      *pBlue,
    UInt8      *pMask,
    Int32       bplS,
    Int32       bplD,
    Int32       bplM,
    Int32       width,
    Int32       height,
    UInt8      *pL,
    UInt8      *pA,
    UInt8      *pB )
{
    PROLOGUE;
    Int32    idx, idxH, idxW, idxB, w;
    Int32    red, green, blue, x, y, z, l, a, b;
    UInt8    byte;
    Float64  yy;
    RGB2LABTable  *pTab;

    pTab = (RGB2LABTable *) pTable;
/*  Convert all of pixels */
    if( pMask == NULL )
    {
	for(idxH=0; idxH<height; ++idxH)
	{
	    for(idxW=0; idxW<width; ++idxW)
	    {
		red = (Int32) UCHAR_ACCESS(pRed +idxW);
		green = (Int32) UCHAR_ACCESS(pGreen +idxW);
		blue = (Int32) UCHAR_ACCESS(pBlue +idxW);
		x = pTab->r_X[red] + pTab->g_X[green] + pTab->b_X[blue];
		y = pTab->r_Y[red] + pTab->g_Y[green] + pTab->b_Y[blue];
		z = pTab->r_Z[red] + pTab->g_Z[green] + pTab->b_Z[blue];
		yy = pTab->tabPow[y];
		l = pTab->l[y];
		a = (Int32)( 500.0 * (pTab->tabPow[x] - yy) + 128.5);
		b = (Int32)( 200.0 * (yy - pTab->tabPow[z]) + 128.5);
		UCHAR_ACCESS(pL +idxW) = (UInt8)LIMIT(l, 0, 255);
		UCHAR_ACCESS(pA +idxW) = (UInt8)LIMIT(a, 0, 255);
		UCHAR_ACCESS(pB +idxW) = (UInt8)LIMIT(b, 0, 255); 
	    }
	    pRed += bplS;
	    pGreen += bplS;
	    pBlue += bplS;
	    pL += bplD;
	    pA += bplD;
	    pB += bplD;
	}
    }
    else
    {
	for(idxH=0; idxH<height; ++idxH)
	{
	    for(idxW=0, idxB=0; idxW < width; ++ idxB, idxW +=8)
	    {
		byte = UCHAR_ACCESS(pMask +idxB);
		if( byte == 0 ) continue;
		w = idxW + 8;
		if( w> width ) w=width;
		for( idx = idxW; idx<w; ++idx)
		{
		    if( byte & 0x80 )
		    {
			red = (Int32) UCHAR_ACCESS(pRed +idx);
			green = (Int32) UCHAR_ACCESS(pGreen +idx);
			blue = (Int32) UCHAR_ACCESS(pBlue +idx);
			x = pTab->r_X[red] +pTab->g_X[green] + pTab->b_X[blue];
			y = pTab->r_Y[red] +pTab->g_Y[green] + pTab->b_Y[blue];
			z = pTab->r_Z[red] +pTab->g_Z[green] + pTab->b_Z[blue];
			yy = pTab->tabPow[y];
			l = pTab->l[y];
			a = (Int32)( 500.0 * (pTab->tabPow[x] - yy) + 128.5);
			b = (Int32)( 200.0 * (yy - pTab->tabPow[z]) + 128.5);
			UCHAR_ACCESS(pL +idx) = (UInt8)LIMIT(l, 2, 255);
			UCHAR_ACCESS(pA +idx) = (UInt8)LIMIT(a, 0, 255);
			UCHAR_ACCESS(pB +idx) = (UInt8)LIMIT(b, 0, 255); 
		    }
		    byte <<= 1;
		}
	    }
	    pRed += bplS;
	    pGreen += bplS;
	    pBlue += bplS;
	    pL += bplD;
	    pA += bplD;
	    pB += bplD;
	    pMask += bplM;
	}
    }

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free RGB2LABTable.
 *
 * LongDescription:::
 *  This routine free table created in i_createRGB2LabTable. 
 *
 * Input:::
 *  pIPState::      Pointer to standard state block. 
 *
 * Output:::
 *  ppTable::       Point to pointer of table. 
 *
 * Returns:::
 *	ia_successful
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
i_freeRGB2LabTable (
    IPState    *pIPState,
    void      **ppTable )
{
    PROLOGUE;
    RGB2LABTable   *pTab;

    pTab = (RGB2LABTable *) *ppTable;

    FREE(pTab);

    *ppTable = NULL;

    NORMAL_RETURN;
}
    
/*

$Log: i_color.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.18  1998/08/27 23:22:34  xzhu
xzhu on Thu Aug 27 16:19:29 PDT 1998

i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

Seperate creating table and RGB to Lab convertion. It will save time when we convert image strip by strip. If there is mask image, just do convertion to the masked pixels.

Revision 1.17  1998/02/13 23:41:30  xzhu
xzhu on Fri Feb 13 15:39:33 PST 1998

i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

Use sandpiper XYZ matrix for all RGB->Lab and Lab->RGB transforms. Created new function i_LabToRGB.

Revision 1.16  1998/01/23 22:48:59  xzhu
xzhu on Fri Jan 23 14:47:57 PST 1998

i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

Created new functions to convert RGB to Lab and Lab to RGB.

Revision 1.15  1998/01/13 23:47:28  ddavies
i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.14  1997/11/13 23:23:51  aber
aber on Thu Nov 13 15:21:23 PST 1997

i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

Add new RGB to luminance conversion, using DITC formula used in AIE.

Revision 1.13  1997/05/21 02:04:19  jcraig
jcraig on Tue May 20 19:03:55 PDT 1997

i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

add procedures i_RGBToLabInit, i_RGBToLabFinish and i_RGBToLab which
convert between RGB and L*a*b*

Revision 1.12  1997/04/15 19:28:56  ddavies
ddavies on Tue Apr 15 12:26:05 PDT 1997

i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

Fix two bugs in i_indexed4ToRGB.  It was trying to use w % 1 to decide if
a line length was odd.  The code executed when the corrected odd indicator
was set was also broken in that it addressed cTable with index * 16 instead
of index * 4.

Revision 1.11  1997/04/01 21:49:59  dferrell
dferrell on Tue Apr 1 13:46:23 PST 1997

i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

Added new function i_RGBToYFast () which computes a close approximation
of the RGB to Y transform in about half the time of the LUT based
functions.
Changed the LUTs for the other RGB to Y functions.  For some reason they
were made using some approximate coefficients instead of the standard
ones.  Now we have slow and accurate or closely approximate and fast
versions.

Revision 1.10  1995/12/13 02:43:04  ejaquith
i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.9  1995/10/13  23:50:09  dferrell
 * i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.8  1994/07/27  22:28:07  ejaquith
 * ejaquith on Wed Jul 27 15:25:59 PDT 1994
 *
 * i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Changed i_RGBToLuv and i_RGBToLuvInit functions to have pointers to
 * internal state information (instead of globals).
 * Added function i_RGBToLuvFinish function which must be used to clean
 * up this state information.
 * Made statically defined compositeTable a regular variable.
 *
 * Revision 1.7  1994/07/25  17:34:51  macleod
 * macleod on Mon Jul 25 10:33:18 PDT 1994
 *
 * i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v
 *
 * Fixed a couple of bugs in i_indexed4ToRGB. First, the output addressing was wrong,
 * resulting in the image being squished, second, the frame wasn't being skipped
 * correctly, and third, one of the channels in a look-up table was always zero.
 *
 * Revision 1.6  1994/05/13  02:32:15  macleod
 * i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v
 *
 * Moved (very slow) rgb->luv routines here, added indexed 4-bit to
 * RGB conversion
 *
 * Revision 1.5  1994/02/04  20:14:14  dferrell
 * i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/15  20:23:54  macleod
 * macleod on Wed Dec 15 12:23:43 PST 1993
 *
 * i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v
 *
 * added i_indexed8ToRGB()
 *
 * Revision 1.3  1993/12/02  19:22:50  ddavies
 * i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:27:43  ddavies
 * i_color.c -> /product/ipcore/ipshared/src/RCS/i_color.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

