/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_cyclsd.c	routines for seed placement in cycles of a given length
 *
 *	Int32 i_horizCycleSeed()
 *
 */


#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "memory.pub"
#include "frames.pub"
#include "lmask.prv"
#include "runhist2.prv"

/* Prototypes */
#include "shrcc.pub"
#include "shrcc.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_cyclsd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* i_horizCycleSeed finds all cycles of length
 * [minCycleThresh..maxCycleThresh] and puts a seed pixel in the 
 * leftmost on pixel of the run. 
 * A cycle is defined to be a series of 1's followed by a series of 0's.
 *
 * This code was based on the horizontal run length histogram code.
 * We use 3 tables.  pCyclesOnEndsBytePtr is used for runs spanning bytes.
 * The first byte indicates when a run from the previous byte ends and gives 
 * the distance that the run extends into this byte.  When the run ends in  
 * the current byte, the value is negative.  Note that this number is a
 * ones complement table.  See i_rnhst1.c for details.  The second byte
 * of this table contains the length of the run on the left that
 * extends out of the current byte into the next byte.
 * The other 2 tables, pByteEndSeed and pByteSeed, contain the seed
 * pixels that will be written into the destination.  pByteEndSeed is
 * used whenever the cycle that extends between the previous and current
 * bytes must be counted.  pByteSeed is used whenever the cycle that extends
 * between the 2 bytes is either too large or too small to be counted.
 */

/* PUBLIC */
/**************************************************************************
 * i_horizCycleSeed()
 * Find all cycles of length [minCycleThresh..maxCycleThresh] and put
 * a seed pixel in the leftmost on pixel of the run. 
 *
 * Input:
 *	pSImage		pointer to source image
 *	sBpl		bytes per line of source image
 *	w		width of source image
 *	h		height of source image
 *	minCycleThresh	the minimum cycle length to be seeded
 *	maxCycleThresh	the maximum cycle length to be seeded
 *	pDImage		pointer to destination image
 *	dBpl		bytes per line of souce image
 *
 * Returns:
 *	ia_successful	successful completion
 *	ia_nomem	unable to allocate memory
 *	
 **************************************************************************/
Int32 CDECL 
i_horizCycleSeed (
    IPState    *pIPState,
    UInt8      *pSImage,
    Int32	sBpl,
    Int32	w,
    Int32	h,
    Int32	minCycleThresh,
    Int32	maxCycleThresh,
    UInt8      *pDImage,
    Int32	dBpl)
{
PROLOGUE;
 Int32	 rightWordMask;		/* used to delete areas outside
				 * of rectangle in right word of scan line. */
Int32	 rightOffset;		/* the byte offset of the right word of the
				 * rectangle.
				 */

UInt16	*pByteSeed = NULL;	/* points to tables used to contain seed
				 * pixels */
UInt16	*pByteEndSeed = NULL;

UInt16	 seedBits[9] = {0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80, 0x100};
UInt32	 index;			/* current table index */
UInt16	 shiftedIndex;		/* index is shifted to find cycles */
Int32	 cycleStarts;		/* bit in which cycle starts */
Int32	 curBit;		/* current bit being examined */
UInt16	 output, outputPartial; /* current table entry */

	/* allocate the seed tables.  We want one entry for every
	 * possible byte value.  We really index this thing with
	 * 9 bit things and each entry is 9 bits wide.  It would have
 	 * been nice if this table could have been only 8 bits wide
	 * but we need to account for the case in which the last one
	 * bit of the cycle is the most significant bit of the previous
	 * byte.
	 */
    pByteSeed = (UInt16 *) MALLOC(512 * sizeof(UInt16));
    if (pByteSeed == NULL)
	return ia_nomem;
    pByteEndSeed = (UInt16 *) MALLOC(512 * sizeof(UInt16));
    if (pByteEndSeed == NULL)
    {
	FREE(pByteSeed);
	return ia_nomem;
    }
    for (index = 0; index < 512; index++)
    {
	output = outputPartial = 0;
	cycleStarts = -1;
	/* note that without this bit or'd in, if the high order bit of 
	 * the index is one, the check below will succeed and if the
	 * cycle length is below the threshold, a one will be put in
	 * the output byte.  this may or may not be correct, depending
	 * on how many ones are in the next byte.
	 */
	shiftedIndex = (UInt16)(index | 0x200);
	for (curBit = 0; curBit <= 8; curBit++)
	{
	    /* we've found the end of a cycle */
	    if ((shiftedIndex & 0x3) == 0x1)
	    {
		if (cycleStarts == -1)
		{
		    /* this is the first cycle that we've found */
		    cycleStarts = curBit;
		    outputPartial = seedBits[curBit];
		}
		else if (((curBit - cycleStarts) <= maxCycleThresh) &&
			 ((curBit - cycleStarts) >= minCycleThresh))
		    output |= seedBits[curBit];
	    }
	    shiftedIndex = shiftedIndex >> 1;
	}
	pByteSeed[index] = output;
	pByteEndSeed[index] = output | outputPartial;
    }

	/* skip past the frames on the images */
    pSImage += (sBpl << cLogFrameBits) + cFrameBytes;
    pDImage += (dBpl << cLogFrameBits) + cFrameBytes;
    h = h - 2 * cFrameBits;
    w = w - 2 * cFrameBits;

	/* how far away is the right edge of the image? */
    rightOffset = ((w - 1)/32) * 4;
    rightWordMask = leftMask[(w-1) % 32];

	    /* go get the data from the image a line at a time */
    for ( ; h > 0; pSImage += sBpl, pDImage += dBpl, h--)
    {
	__error__ = ip_horizCycleSeedLine (pIPState,
					 pSImage, rightOffset, rightWordMask,
					 pDImage,  (UInt8 *)pByteSeed, 
					 (UInt8 *)pByteEndSeed, 
					 (Int8 *) &cyclesOnEnds,
					 minCycleThresh, maxCycleThresh);
	if (__error__ != ia_successful)
	    break;
    }

	/* that's it.  We just free the stuff we've allocated and return */
    if (pByteSeed != NULL)
	FREE(pByteSeed);
    if (pByteEndSeed != NULL)
	FREE(pByteEndSeed);
    NORMAL_RETURN;
}


/*

$Log: i_cyclsd.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.6  1998/07/28 19:05:59  aber
aber on Tue Jul 28 12:05:37 PDT 1998

i_cyclsd.c -> /product/ipcore/ipshared/src/RCS/i_cyclsd.c,v

Explicit parenthesis to avoid gcc compiler warning.

Revision 1.5  1998/01/21 21:22:55  ddavies
i_cyclsd.c -> /product/ipcore/ipshared/src/RCS/i_cyclsd.c,v

Remove cast in FREE invocation.

Revision 1.4  1998/01/13 23:47:29  ddavies
i_cyclsd.c -> /product/ipcore/ipshared/src/RCS/i_cyclsd.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1995/12/13 02:43:08  ejaquith
i_cyclsd.c -> /product/ipcore/ipshared/src/RCS/i_cyclsd.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/10/13  23:50:15  dferrell
 * i_cyclsd.c -> /product/ipcore/ipshared/src/RCS/i_cyclsd.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.1  1995/04/14  01:20:41  jcraig
 * Initial revision
 *

*/
