/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_dibpix.c
 *
 *  Public routines:
 *      i_DIBImageFromToPixr():	converts between the image portion of
 *				a Microsoft DIB to the style of image we use.
 *
 * Neither of these routines deal with the BITMAPINFOHEADER or RGBQUADS of
 * a DIB.
 */


#include <stdio.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

    /* prototypes */
#include "shrpixr.pub"
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_dibpix.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* Our image format:
 * * Pixels are stored in units of 32-bit words (pixels are read left to right
 *   within 32 bit words)
 * * The upper left corner of the image is stored at the lowest memory address.
 * * A pixel value of 0 is white for binary images.
 * * A pixel value of 0 is black for gray images.
 * * Our images are surrounded by 32 pixel frames filled with 0's.
 *
 * Microsoft DIB format:
 * * Pixels are stored in units of 8-bit bytes (luckily still left to right
 *   within a byte).
 * * The *lower* left corner of the image is stored at the lowest
 *   memory address.
 * * A pixel value of 0 is white for binary images (Hooray).
 * * A pixel value of 0 is *white* for gray images (Boo).
 * * There are no frames, though images are padded to 32-bit boundaries.
 *
 * In big endian machines, the order of pixels within 32-bit words is the
 * same since bytes are ordered from MS to LS within a 32 bit word.  However,
 * in little endian machines, the bytes within a 32 bit word must be
 * swapped since left-to-right would naturally correspond to LS to MS
 * within a 32 bit word.  Swapping the bytes is most of the work.
 *
 * In all machines, we must reverse the order of the lines within the image.
 * 
 * We must invert the pixels if the pixels are gray.
 *
 * We organize the routine in two levels.  The top one deals with parameters
 * as usual.  The bottom one swaps the bytes in a line of image.  The bottom
 * one can be implemented in assembly language, especially on i486 machines
 * that have a byteswap opcode.  We use ip_getByteSwapProc to give us a pointer
 * to the procedure that works best on the current platform.  This must be
 * decided at runtime on a PC.
 *
 * The EZScan folks would sometimes like to convert images between PIXRs
 * and DIBs one scan line at a time.  This saves them memory on the DIB
 * side when they're scanning and generally minimizing overhead.  We will
 * have two routines that return pointers to procedures that convert a
 * line at a time.  One type of line procedure does gray or binary images.
 * The other type of proc converts RGB DIBs to 3 color PIXRs.  The routine
 * that finds the correct line-at-a-time proc is called i_getDIBPixrLineProc.
 */


/* This is a public procedure */
/* PUBLIC */
/***************************************************************************
 *  i_getDIBPixrLineProc()
 *
 *		This routine returns a pointer to a procedure that can
 *		be used to translate between a single scan lines of a DIB
 *		and a single scan line of 1 binary or gray PIXR or 1 scan
 *		line of each of three gray image for color DIBs.  In the
 *		case of gray images, this means swapping the bytes inside
 *		each word.  In the case of interleaved color images, we
 *		also convert between one interleaved line. In DIB little endian
 *		order:	word 0:	| B | R | G | B |
 *			word 1: | G | B | R | G |
 *			word 2:	| R | G | B | R |
 *
 *		All the far pointer nonsense is used in Windows apps
 *		where they can't guarantee that the buffer holding a
 *		DIB will be in our address space.  In NT and Win95 and
 *		Unix and Mac apps, only near pointers are used.  In
 *		those worlds, __far is turned into nothing at compile time.
 *
 *	invertImage: 	TRUE => bytes should be inverted
 *	direction:	cDIBToPixr => translation from a DIB
 *			to a PIXR (source pointer is __far in gray images);
 *			cPixrToDIB => translation from a PIXR
 *			to a DIB (dest pointer is __far in gray images).
 *			cPixrToPixr => Neither pointer is far in gray
 *			images.  Note that this value does not make sense
 *			for color images if __far addresses exist on the
 *			current platform.
 *	isColor:	TRUE => the DIB is an RGB image.  Otherwise,
 *			the DIB is a gray or binary image.
 *	grayToFarProc:	A procedure that byteswaps a scan line containing
 *			a single channel of binary or gray image data where
 *			the source is a near pointer and the dest is
 *			a far pointer.
 *	grayFromFarProc: A procedure that byteswaps a scan line containing
 *			a single channel of binary or gray image data where
 *			the source is a far pointer and the dest is
 *			a near pointer.
 *	grayNearProc:	A procedure that byteswaps a scan line containing
 *			a single channel of binary or gray image data where
 *			both the source and dest pointers are near.
 *	colorProc:	A procedure that converts between an interleaved
 *			line of three-channel data and three near buffers
 *			each holding one channel.  This form is used both
 *			when converting to an interleaved DIB and when
 *			converting from an interleaved DIB.
 *
 *			The types GraySwapToFarProc, GraySwapFromFarProc,
 *			GraySwapNearProc and ColorSwapProc are defined in
 *			ipshared/include/shrpixr.pub.
 *
 *			pSource is a pointer to the left end of the line of
 *			gray or binary source pixels.
 *			pDest is a pointer to the left end of the line of
 *			gray or binary destination pixels.
 *			wordsPerLine is the number of words to byteswap in
 *			the scan line.
 *			pDIBImage is a pointer to the left end of an
 *			interleaved line in the order shown above.
 *			pRed is a pointer the left end of the red plane of
 *			one image line.
 *			pGreen is a pointer to the left end of the green
 *			plane of one image line.
 *			pBlue is a pointer to the left end of the blue plane
 *			of one image line.
 *			width is the number of 24-bit color pixels (not bytes)
 *			in the image line.
 *
 *
 * REMEMBER that the vertical order of scan lines in DIBs is reversed from
 * that in PIXRs!!!  Line 0 is the top image line in Pixrs and the bottom
 * image line in DIBs.
 *
 * This routine always returns a status code of ia_successful.
 ***************************************************************************/
Int32  CDECL 
i_getDIBPixrLineProc (
    IPState    *pIPState,
    UInt32	invertImage,
    UInt32	direction,
    UInt32	isColor,
    GraySwapToFarProc *pGrayToFarProc,
    GraySwapFromFarProc *pGrayFromFarProc,
    GraySwapNearProc *pGrayNearProc,
    ColorSwapProc *pColorProc)
{

	/* If we're converting color DIBs, return the address of
	 * ip_combineDIBTo3Line or ip_combine3ToDIBLine.
	 */
    PROLOGUE;
    if (isColor)
    {
	if (pColorProc == NULL)
	    return ia_invalidParm;
	if (direction == cDIBToPixr)
	    *pColorProc = ip_combineDIBTo3Line;
	else
	    *pColorProc = ip_combine3ToDIBLine;
    }
    else	/* is gray or binary */
    {
	    /* Check the proc pointers before calling this function
	     * so that if it's implemented in assembly language, it
	     * won't have to.
	     */
	switch (direction)
	{
	case cDIBToPixr:
	    if (pGrayFromFarProc == NULL)
		return ia_invalidParm;
	    break;
	case cPixrToDIB:
	    if (pGrayToFarProc == NULL)
		return ia_invalidParm;
	    break;
	default:	/* cPixrToPixr */
	    if (pGrayNearProc == NULL)
		return ia_invalidParm;
	    break;
	}
	ip_getByteSwapProc (pIPState, invertImage, direction, pGrayToFarProc,
			    pGrayFromFarProc, pGrayNearProc);
    }

    NORMAL_RETURN;

}



/* This is a public procedure */
/* PUBLIC */
/*
 *  i_DIBImageFromToPixr()
 *		This routine converts the image portion of a Microsoft
 *		DIB into an ipshared-style image.  The ipshared image need
 *		not have a frame.
 *
 *	pDIBImage:	Pointer to the DIB's image bits.
 *	pD:		Pointer to upper left corner of destination image
 *			(NOT including frame).
 *	height:		height of image in pixels (NOT including frame).
 *	invertImage:	cTrue => invert the image in the conversion.
 *	DIBBpl:		number of bytes/line in DIB image.
 *	dBpl:		number of bytes/line in dest image.
 *	direction:	When equal to cDIBToPixr, we're converting from
 *			DIB to Pixr.  When equal to cPixrToDIB,
 *			we're converting from Pixr to DIB.
 *
 * This routine always returns a status code of ia_successful.
 */
Int32  CDECL 
i_DIBImageFromToPixr (
    IPState    *pIPState,
    UInt8 __far *pDIBImage,
    UInt8      *pD,
    UInt32	height,
    UInt32	invertImage,
    UInt32	DIBBpl,
    UInt32	dBpl,
    UInt32	direction)
{
    PROLOGUE;
    GraySwapToFarProc	swapGrayToFar;
    GraySwapFromFarProc	swapGrayFromFar;
    GraySwapNearProc	swapGrayNear;

	/* figure out which swap routine to use.  The image will be
	 * inverted if it's gray.
	 */
    ip_getByteSwapProc (pIPState, invertImage, direction, &swapGrayToFar,
			&swapGrayFromFar, &swapGrayNear);

	/* point pD to the bottom of the destination image since pDIBImage
	 * is already pointing to the bottom of the source image.
	 */
    pD += (height-1) * dBpl;

    if (direction == cDIBToPixr)  /* converting to Pixr */
    {
	if (swapGrayFromFar == NULL)
	    return ia_noImpl;

	    /* have the routine we found above do the real work. */
	for (; height > 0; pDIBImage += DIBBpl, pD -= dBpl, height--)
	{
	    (*swapGrayFromFar)(pIPState,
			       (UInt32 __far *)pDIBImage, (UInt32 *) pD,
			       DIBBpl/4);
	}
    }
    else			  /* converting to DIB */
    if (direction == cPixrToDIB)
    {
	if (swapGrayToFar == NULL)
	    return ia_noImpl;

	    /* have the routine we found above do the real work. */
	for (; height > 0; pDIBImage += DIBBpl, pD -= dBpl, height--)
	{
	    (*swapGrayToFar)(pIPState,
			     (UInt32 *)pD, (UInt32 __far *) pDIBImage,
			     DIBBpl/4);
	}
    }
    else	/* direction == cPixrToPixr */
    {
	if (swapGrayNear == NULL)
	    return ia_noImpl;

	    /* have the routine we found above do the real work. */
	for (; height > 0; pDIBImage += DIBBpl, pD -= dBpl, height--)
	{
	    (*swapGrayNear)(pIPState,
			    (UInt32 *)pD, (UInt32 *) pDIBImage,
			    DIBBpl/4);
	}
    }


    NORMAL_RETURN;
}



/* PUBLIC */
/*
 *  i_DIBImageFromTo3Pixrs()
 *		This routine combines 3 ipshared-style images into the 
 *		image portion of a Microsoft DIB or vice versa.  One
 *		image contains the red portion, one the green portion,
 *		and the other the blue portion of the RGB triple that is 
 *		put into or pulled out of the DIB.  The ipshared images
 *		need not have a frame.
 *		This routine assumes that each component of the triple
 *		is 8 bits wide.
 *
 *	pDIBImage:	Pointer to the DIB's image bits.
 *	rS:		Pointer to upper left corner of destination image
 *			of the red component (NOT including frame).
 *	gS:		Pointer to upper left corner of destination image
 *			of the green component (NOT including frame).
 *	bS:		Pointer to upper left corner of destination image
 *			of the blue component (NOT including frame).
 *	height:		height of image in pixels (NOT including frame).
 *	width:		width of image (NOT including frame)
 *	DIBBpl:		number of bytes/line in DIB image.
 *	pixrBpl:	number of bytes/line in dest image.
 *	direction:	When equal to cDIBToPixr, we're converting from
 *			DIB to Pixr.  When equal to cPixrToDIB, we're converting
 *			from Pixr to DIB.
 *
 * Returns
 *		ia_successful	if on a pc
 *		ia_noImpl	otherwise
 * 
 */
Int32  CDECL 
i_DIBImageFromTo3Pixrs (
    IPState    *pIPState,
    UInt8 __far *pDIBImage,
    UInt8      *rS,
    UInt8      *gS,
    UInt8      *bS,
    UInt32	height,
    UInt32	width,
    UInt32	DIBBpl,
    UInt32	pixrBpl,
    UInt32	direction)
{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

	/* point to the bottom of the source image since pDIBImage
	 * is already pointing to the bottom of the destination image.
	 */
    PROLOGUE;
    rS += (height-1) * pixrBpl;
    gS += (height-1) * pixrBpl;
    bS += (height-1) * pixrBpl;

    if (direction == cDIBToPixr)  /* converting to Pixr */
    {
	    /* have the routine we found above do the real work. */
	for ( ;
	     height > 0;
	     height--, rS -= pixrBpl, gS -= pixrBpl,
		bS -= pixrBpl, pDIBImage += DIBBpl)
	{
	    ip_combineDIBTo3Line (pIPState, pDIBImage, rS, gS, bS, width);
	}
    }
    else			  /* converting to DIB */
    {
	    /* have the routine we found above do the real work. */
	for ( ;
	     height > 0;
	     height--, rS -= pixrBpl, gS -= pixrBpl,
		bS -= pixrBpl, pDIBImage += DIBBpl)
	{
	    ip_combine3ToDIBLine (pIPState, pDIBImage, rS, gS, bS, width);
	}
    }

    NORMAL_RETURN;

#else
    pIPState = pIPState;  /* keep compiler quiet */
    pDIBImage = pDIBImage;
    rS = rS;
    gS = gS;
    bS = bS;
    height = height;
    width = width;
    DIBBpl = DIBBpl;
    pixrBpl = pixrBpl;
    direction = direction;

    return ia_noImpl;
#endif

}

/*

$Log: i_dibpix.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.13  1998/01/13 23:47:34  ddavies
i_dibpix.c -> /product/ipcore/ipshared/src/RCS/i_dibpix.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.12  1995/12/13 02:43:18  ejaquith
i_dibpix.c -> /product/ipcore/ipshared/src/RCS/i_dibpix.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.11  1995/10/18  03:52:42  ddavies
 * i_dibpix.c -> /product/ipcore/ipshared/src/RCS/i_dibpix.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.10  1995/05/10  21:54:07  ejaquith
 * ejaquith on Wed May 10 14:52:26 PDT 1995
 *
 * i_dibpix.c -> /product/ipcore/ipshared/src/RCS/i_dibpix.c,v
 *
 * Changed #ifdef _WATCOMC_ to #ifdef _386_ARCH_ because this was the
 * run on the win32 compiled PC code (since _WATCOMC_ is compiler
 * specific).  Brian Hart encountered this then the byte swap procedure
 * was NULL!
 *
 * Revision 1.9  1994/10/05  18:19:29  dferrell
 * dferrell on Wed Oct 5 11:17:02 PDT 1994
 *
 * i_dibpix.c -> /product/ipcore/ipshared/src/RCS/i_dibpix.c,v
 *
 * Made use of direction parameter on big endian machines so compilers
 * won't complain about unused parameters.
 *
 * Revision 1.8  1994/01/22  01:07:30  ddavies
 * i_dibpix.c -> /product/ipcore/ipshared/src/RCS/i_dibpix.c,v
 *
 * Change "include shrio." to "include shrpixr." since ipshared's libio is
 * being absorbed by libpixr.
 *
 * Revision 1.7  1993/12/02  19:22:59  ddavies
 * i_dibpix.c -> /product/ipcore/ipshared/src/RCS/i_dibpix.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.6  1993/11/05  01:27:59  ddavies
 * i_dibpix.c -> /product/ipcore/ipshared/src/RCS/i_dibpix.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

