/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_dilbig.c -- includes
 *
 *  	i_rectDilate():  fast rectangular dilation
 */

#ifndef _GNUC_
#include <string.h>
#endif

#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"

/* prototypes */
#include "shrpatt.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_dilbig.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/*
 * This file was written at Xerox PARC by William Rucklidge, a researcher
 * in the Systems Practices Laboratory in the Image Understanding Area.
 * Eric Jaquith of ipcore did minor cosmetic changes to bring it into
 * ipshared, the major work of which was de-macroizing about 20 macros
 * into regular code expansions in the routines, into actual function calls,
 * or simply taking them out (as the original file allowed for dilations
 * in Y of 1...7 and in X of 1,2,4 as well -- these size dilations are small
 * enough that calling the i_fastDilate() routine will be faster.)
 * This editing reduced the object file size from ~23K to ~5K, with no
 * expense of speed.
 *
 * How much faster is this routine than i_fastDilate(), even using assembly
 * routines when possible and breaking (when faster) the single square SxS
 * dilation into two dilations, an Sx1 horizontal and a 1xS vertical?
 * For a 1MB binary image, on a SPARC5:
 *     size:  time for i_fastDilate():   time for i_rectDilate()
 *      2x2    0.07                       0.18
 *      4x4    0.13                       0.21
 *      8x8    0.23*                      0.21
 *    16x16    0.34*                      0.33
 *    32x32    2.00*                      0.37
 *    64x64    6.98*                      0.44
 *  128x128   13.76*                      0.57
 *
 * i_rectDilate()'s   32x32 is about  5 times faster
 * i_rectDilate()'s   64x64 is about 15 times faster
 * i_rectDilate()'s 128x128 is about 24 times faster!
 *
 * The "*" indicates i_fastDilate() used Sx1 and 1xS dilations.
 */

/* This gives a number with only the least significant set bit of l set. */
#define	LSSB(l)	(l ^ (l & (l - 1)))

/*
 * This table gives a number which has every bit set from the LSB
 * to (and including) the MSSB (most sig set bit) of its arg, or 0 for 0.
 */
static UInt8 toMSSB[256] = {
    0x00, 0x01, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07,
    0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
    0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f,
    0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f,
    0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
    0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
    0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
    0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    };



/*
 * Here are some variants on the "dilate in Y" theme. What these do:
 * you can dilate by, say, 32 in Y by considering Y as 8 images, each
 * consisting of every 8th row (so one image would have rows 0,8,16,..., the
 * next would have 1,9,17,... etc; these are called "row-sets"), dilating each
 * one of those by 4, then dilating the whole Y image by 8. I break it down
 * like that rather than as, say dilating by 16 on 2 row sets, then by 2
 * overall because dilating by 8 is the most efficient; more than that and
 * your temporary variables start using up more registers than you have,
 * in the SUN SPARC world.
 */

/*
 * Dilate in Y, propagating up by 4 steps. The steps are "every" row high,
 * and start at "offset" from the bottom of the image. This is basically
 * "dilate up by 4 in Y" except that only every "every"th row in the image
 * is considered, and "offset" says which of the row sets to consider.
 * The "INNER" flag here means that this assumes that the outer part
 * of DILY above (the X loop) has been done, so we can use basep, baseq,
 * and x (variables from the calling routine).
 */
static void CDECL 
ip_dilYProp4InnerOne (
    IPState    *pIPState,
    UInt32     *basep,
    UInt32     *baseq,
    UInt32     *dest,
    Int32	srcSkipPitch,
    Int32	destSkipPitch)
{
    PROLOGUE;
    UInt32   down = 0, down2 = 0, down3 = 0;
    UInt32   t;

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    while (baseq >= dest)
    {
	t = *basep;
	*baseq = t | down | down2 | down3;
	down3 = down2;
	down2 = down;
	down = t;

	basep -= srcSkipPitch;
	baseq -= destSkipPitch;
    }
}


/*
 * Like dilYProp4InnerOneEvery8() but with "every" hardwired to 8.
 */
static void CDECL 
ip_dilYProp4InnerOneEvery8 (
    IPState    *pIPState,
    UInt32     *basep,
    UInt32     *baseq,
    UInt32     *dest,
    Int32	srcSkipPitch,
    Int32	destSkipPitch)
{
    PROLOGUE;
    UInt32   down = 0, down2 = 0, down3 = 0;
    UInt32   t;

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    while (baseq >= dest)
    {
	t = *basep;
	*baseq = t | down | down2 | down3;
	down3 = down2;
	down2 = down;
	down = t;

	basep -= srcSkipPitch;
	baseq -= destSkipPitch;
    }
}


/*
 * Like dilYProp4InnerOne() but propagating by 8 and having "every"
 * hardwired to 8.
 */
static void CDECL 
ip_dilYEvery8InnerOne (
    IPState    *pIPState,
    UInt32     *basep,
    UInt32     *baseq,
    UInt32     *dest,
    Int32	srcSkipPitch,
    Int32	destSkipPitch)
{
    PROLOGUE;
    UInt32   down = 0, down2 = 0, down3 = 0, down4 = 0;
    UInt32   down5 = 0, down6 = 0, down7 = 0;
    UInt32   t;
    
	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    while (baseq >= dest)
    {
	t = *basep;
	*baseq = t | down | down2 | down3 | down4 | down5 | down6 | down7;
	down7 = down6;
	down6 = down5;
	down5 = down4;
	down4 = down3;
	down3 = down2;
	down2 = down;
	down = t;

	basep -= srcSkipPitch;
	baseq -= destSkipPitch;
    }
}


/*
 * Like dilYProp4InnerOne() except that only the image "dest" is used;
 * "src" isn't considered.
 * Move up every 32 rows.
 */
static void CDECL 
ip_dilYProp4InnerOneOneIm (
    IPState    *pIPState,
    UInt32     *baseq,
    UInt32     *dest,
    Int32	destPitch)
{
    PROLOGUE;
    UInt32   down = 0, down2 = 0, down3 = 0;
    UInt32   t;

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Now make destPitch so it skips every 32 rows */
    destPitch <<= 5;

    while (baseq >= dest)
    {
	t = *baseq;
	*baseq = t | down | down2 | down3;
	down3 = down2;
	down2 = down;
	down = t;

	baseq -= destPitch;
    }
}


/* PUBLIC */
/****************************************************************************
 * Do a large rectangular dilation of an image
 *
 * Description:
 *  Take an image and do a hit-only binary morphology dilation on a binary
 *  image using a structure element, described solely by the nx,ny parameters.
 *  This routine restricts the structure element to being of the form:
 *    1 1 1 1 1 1 1 1
 *    1 1 1 1 1 1 1 1
 *    1 1 1 1 1 1 1 1
 *    1 1 1 1 1 1 1 1
 *    1 1 1 1 1 1 1 1
 *    1 1 1 1 1 1 1 1
 *    1 1 1 1 1 1 1 1
 *    1 1 1 1 1 1 1 1c
 *  nx and ny are the number of STREL columns and rows, respectively.
 *  nx must be 1, 2, 4, 8, 16, 32, 64, or 128
 *  ny must be 1, 2, 4, 8, 16, 32, 64, or 128
 *  These restrictions allow a faster routine than i_fastDilate() can offer
 *  (even with an assembly dilate operation) of the same STREL.<BR>
 *
 *  While sizes with width AND height both at least 8 will be faster than
 *  i_fastDilate(), sizes with both smaller will not.  However, when one
 *  dimension is at least 8 and the other is not, this routine will generally
 *  be faster, but you should check yourself if speed is an issue.
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_invalidParm   	invalid passed in nx and/or ny value(s)
 *
 * Notes:
 *  The coordinates of the structure element's center is the lower righthand
 *  corner in this routine.
 *****************************************************************************/
Int32 CDECL 
i_rectDilate (
    IPState    *pIPState,
    UInt8      *dest,	  /* Ptr to upper left corner of dest image,
			   * including frame */
    UInt8      *src,	  /* Ptr to upper left corner of source image,
			   * including frame */
    Int32	imageWidth, /* Width and height of images in pixels, */
    Int32	imageHeight, /* including frame, excluding pad */
    Int32	destPitch, /* Number of bytes/line in dest and source image, */
    Int32	srcPitch, /* including frame, border, and pad */
    Int32	nx,	  /* Structure element has nx columns and ny rows. */
    Int32	ny)         /* nx and ny must be 1, 2, 4, 8, 16, 32, 64, or 128 */
{    
    PROLOGUE;
    UInt32  *p, *q, *basep, *baseq;
    Int32    x;
    Int32    numWords;

    /* Skip binary frames */
    src  += (srcPitch  << cLogFrameBits) + cFrameBytes;
    dest += (destPitch << cLogFrameBits) + cFrameBytes;
    
    if (ny != 1)
    {
	/* Make the bytes per line pitches be words per line pitches */
	srcPitch  >>= 2;
	destPitch >>= 2;
    }
    /* (If ny == 1 we need pitches in bytes per line for Y dilation!) */

    /* Make dimens exclude frame */
    imageWidth -= 2*cFrameBits;
    imageHeight -= 2*cFrameBits;
    
    /* This is the number of words per row that need to be processed */
    numWords = (imageWidth + 31) >> 5;

    
    /*
     * Dilate in Y. In this section, we read from "src" and write to
     * "dest", so we're copying and dilating. It's free, and means we don't
     * have to do any other copying.
     */
    switch (ny)
    {
    case 1:
    {
	UInt8   *pSrc, *pDst, *pEnd;
	Int32    nbytes;  /* Number bytes of data per row */
	
	pSrc = src;
	pDst = dest;
	pEnd = dest + imageHeight * destPitch;

	/* Get the number of significant bytes per image row */
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	nbytes = (imageWidth + 7) >> 3;
#else /*_ALPACA_IMAGE_FMT_ == cBigEndianFmt*/
	/* This makes the number of bytes a multiple (rounding up) of 4.
	 * On the Sun, it's unnecessary.  But on the PC, since the image
	 * is byteswapped, we need to get all significant bytes in the
	 * last word when it is only partially filled.  Since we know
	 * an image row is ALWAYS word-padded, we know this is safe.
	 */
	nbytes = ((imageWidth + 31) >> 5) << 2;
#endif /*_ALPACA_IMAGE_FMT_ == cBigEndianFmt*/

	/* Just copy the source image data to the dest image data! */
	for ( ; pDst < pEnd; pSrc += srcPitch, pDst += destPitch)
	    memcpy((char *)pDst, (char *)pSrc, nbytes);


	/* Make the bytes per line pitches be words per line pitches */
	srcPitch  >>= 2;
	destPitch >>= 2;

	break;
    }
    case 2:
    {
	UInt32 down;
	UInt32 t;
	for (x = numWords,
		 basep = ((UInt32 *)src) + (imageHeight-1)*srcPitch + x - 1,
		 baseq = ((UInt32 *)dest) + (imageHeight-1)*destPitch + x - 1;
	     x;
	     x--, basep--, baseq--)
	{

	    /*
	     * Dilate by 2 in Y.
	     */
	    down = 0;
	    for (p = basep, q = baseq;
		 q >= (UInt32 *)dest;
		 p -= srcPitch, q -= destPitch)
	    {
		t = *p;
		*q = t | down;
		down = t;
	    }
	}
	break;
    }
    case 4:
    {
	UInt32 down, down2, down3;
	UInt32 t;
	for (x = numWords,
		 basep = ((UInt32 *)src) + (imageHeight-1)*srcPitch + x - 1,
		 baseq = ((UInt32 *)dest) + (imageHeight-1)*destPitch + x - 1;
	     x;
	     x--, basep--, baseq--)
	{

	    /*
	     * Dilate by 4 in Y.
	     */
	    down = down2 = down3 = 0;
	    for (p = basep, q = baseq;
		 q >= (UInt32 *)dest;
		 p -= srcPitch, q -= destPitch)
	    {
		t = *p;
		*q = t | down | down2 | down3;
		down3 = down2;
		down2 = down;
		down = t;
	    }
	}
	break;
    }
    case 8:
    {
	UInt32 down, down2, down3, down4, down5, down6, down7;
	UInt32 t;
	for (x = numWords,
		 basep = ((UInt32 *)src) + (imageHeight-1)*srcPitch + x - 1,
		 baseq = ((UInt32 *)dest) + (imageHeight-1)*destPitch + x - 1;
	     x;
	     x--, basep--, baseq--)
	{

	    /*
	     * Dilate by 8 in Y.
	     */
	    down = down2 = down3 = down4 = down5 = down6 = down7 = 0;
	    for (p = basep, q = baseq;
		 q >= (UInt32 *)dest;
		 p -= srcPitch, q -= destPitch)
	    {
		t = *p;
		*q = t | down | down2 | down3 | down4 | down5 | down6 | down7;
		down7 = down6;
		down6 = down5;
		down5 = down4;
		down4 = down3;
		down3 = down2;
		down2 = down;
		down = t;
	    }
	}
	break;
    }
    case 16:
    {
	UInt32 down, down2, down3;
	UInt32 t;
	Int32  srcSkipPitch, destSkipPitch;

	/*
	 * Make two passes, dilating by 4 every fourth the first pass, and
	 * by 4 the second pass.
	 */
	
	/* Make pitches that skip every 4 rows */
	srcSkipPitch = srcPitch << 2;
	destSkipPitch = destPitch << 2;
	
	for (x = numWords,
		 basep = ((UInt32 *)src) + (imageHeight-1)*srcPitch + x - 1,
		 baseq = ((UInt32 *)dest) + (imageHeight-1)*destPitch + x - 1;
	     x;
	     x--, basep--, baseq--)
	{
	    p = basep;
	    q = baseq;
	    ip_dilYProp4InnerOne (pIPState, p, q, (UInt32 *)dest,
				  srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOne (pIPState, p, q, (UInt32 *)dest,
				  srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOne (pIPState, p, q, (UInt32 *)dest,
				  srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOne (pIPState, p, q, (UInt32 *)dest,
				  srcSkipPitch, destSkipPitch);

	    /*
	     * Dilate in Y by 4. Use only "dest".
	     */
	    down = down2 = down3 = 0;
	    for (q = baseq; q >= (UInt32 *)dest; q -= destPitch)
	    {
		t = *q;
		*q = t | down | down2 | down3;
		down3 = down2;
		down2 = down;
		down = t;
	    }
	}
	break;
    }
    case 32:
    {
	UInt32 down, down2, down3, down4, down5, down6, down7;
	UInt32 t;
	Int32  srcSkipPitch, destSkipPitch;
	
	/* Do by 4, then 8 */

	/* Make pitches that skip every 8 rows */
	srcSkipPitch = srcPitch << 3;
	destSkipPitch = destPitch << 3;

	for (x = numWords,
		 basep = ((UInt32 *)src) + (imageHeight-1)*srcPitch + x - 1,
		 baseq = ((UInt32 *)dest) + (imageHeight-1)*destPitch + x - 1;
	     x;
	     x--, basep--, baseq--)
	{

	    /*
	     * Dilate all the 8 every-8 row sets by 4.
	     */
	    p = basep;
	    q = baseq;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);

	    /*
	     * Dilate by 8 in Y, using only the image "dest".
	     */
	    down = down2 = down3 = down4 = down5 = down6 = down7 = 0;
	    for (q = baseq; q >= (UInt32 *)dest; q -= destPitch)
	    {
		t = *q;
		*q = t | down | down2 | down3 | down4 | down5 | down6 | down7;
		down7 = down6;
		down6 = down5;
		down5 = down4;
		down4 = down3;
		down3 = down2;
		down2 = down;
		down = t;
	    }
	}
	break;
    }
    case 64:
    {
	UInt32 down, down2, down3, down4, down5, down6, down7;
	UInt32 t;
	Int32  srcSkipPitch, destSkipPitch;

	/* Make pitches that skip every 8 rows */
	srcSkipPitch = srcPitch << 3;
	destSkipPitch = destPitch << 3;
	
	/* Do by 8 twice */
	for (x = numWords,
		 basep = ((UInt32 *)src) + (imageHeight-1)*srcPitch + x - 1,
		 baseq = ((UInt32 *)dest) + (imageHeight-1)*destPitch + x - 1;
	     x;
	     x--, basep--, baseq--)
	{

	    p = basep;
	    q = baseq;
	    ip_dilYEvery8InnerOne (pIPState, p, q, (UInt32 *)dest,
				   srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYEvery8InnerOne (pIPState, p, q, (UInt32 *)dest,
				   srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYEvery8InnerOne (pIPState, p, q, (UInt32 *)dest,
				   srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYEvery8InnerOne (pIPState, p, q, (UInt32 *)dest,
				   srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYEvery8InnerOne (pIPState, p, q, (UInt32 *)dest,
				   srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYEvery8InnerOne (pIPState, p, q, (UInt32 *)dest,
				   srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYEvery8InnerOne (pIPState, p, q, (UInt32 *)dest,
				   srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYEvery8InnerOne (pIPState, p, q, (UInt32 *)dest,
				   srcSkipPitch, destSkipPitch);

	    /*
	     * Dilate by 8 in Y, using only the image "dest".
	     */
	    down = down2 = down3 = down4 = down5 = down6 = down7 = 0;
	    for (q = baseq; q >= (UInt32 *)dest; q -= destPitch)
	    {
		t = *q;
		*q = t | down | down2 | down3 | down4 | down5 | down6 | down7;
		down7 = down6;
		down6 = down5;
		down5 = down4;
		down4 = down3;
		down3 = down2;
		down2 = down;
		down = t;
	    }
	}
	break;
    }
    case 128:
    {
	UInt32 down, down2, down3, down4, down5, down6, down7;
	UInt32 t;
	Int32  srcSkipPitch, destSkipPitch;

	/* Do by 4, 4, and 8 */

	/* Make pitches that skip every 8 rows */
	srcSkipPitch = srcPitch << 3;
	destSkipPitch = destPitch << 3;

	for (x = numWords,
		 basep = ((UInt32 *)src) + (imageHeight-1)*srcPitch + x - 1,
		 baseq = ((UInt32 *)dest) + (imageHeight-1)*destPitch + x - 1;
	     x;
	     x--, basep--, baseq--)
	{

	    /*
	     * Dilate all the 8 every-8 row sets by 4.
	     */
	    p = basep;
	    q = baseq;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch,destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);
	    p -= srcPitch;
	    q -= destPitch;
	    ip_dilYProp4InnerOneEvery8 (pIPState, p, q, (UInt32 *)dest,
				        srcSkipPitch, destSkipPitch);

	    q = baseq;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);

	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);

	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);
	    q -= destPitch;
	    ip_dilYProp4InnerOneOneIm (pIPState, q, (UInt32*)dest, destPitch);

	    /*
	     * Dilate by 8 in Y, using only the image "dest".
	     */
	    down = down2 = down3 = down4 = down5 = down6 = down7 = 0;
	    for (q = baseq; q >= (UInt32 *)dest; q -= destPitch)
	    {
		t = *q;
		*q = t | down | down2 | down3 | down4 | down5 | down6 | down7;
		down7 = down6;
		down6 = down5;
		down5 = down4;
		down4 = down3;
		down3 = down2;
		down2 = down;
		down = t;
	    }
	}
	break;
    }
    default:
    {
	return ia_invalidParm;
    }
    }


    
    /* Dilate in X */
    /*
     * Tricks used here:
     * - dilating by a and then by b is equivalent to dilating by (a+b-1)
     * - the carryover from one word to the next depends only on the
     *   MSSB of the first word
     *
     * Each loop proceeds as follows:
     *   Iterate over every row in the image "dest", scanning backwards over
     *   every row. A "clearing" statement is executed once at the beginning of
     *   each row, and an "assignment" statement is executed once per word.
     *   p points at the word to be dilated during "assignment".
     */
    switch (nx)
    {
    case 1:
	/* done! */
	break;
    case 2:
    {
	UInt32 left;
	UInt32 t;
	/* Similarly for 2. */
	
	baseq = (UInt32 *)dest;
	for (q = baseq + imageHeight*destPitch; baseq < q; baseq += destPitch)
	{
	    left = 0;
	    for (x = numWords-1; x >= 0; x--)
	    {
		t = baseq[x];
		baseq[x] = t | (t << 1) | left;
		left = t >> 31; /* It's either 0x0 or 0x1!! */
	    }
	}
	break;
    }
    case 4:
    {
	UInt32 left;
	UInt32 t;
	UInt32 t2;
	/* Similarly for 4. */
	
	baseq = (UInt32 *)dest;
	for (q = baseq + imageHeight*destPitch; baseq < q; baseq += destPitch)
	{
	    left = 0;
	    for (x = numWords-1; x >= 0; x--)
	    {
		t = baseq[x];
		t2 = t | (t << 1);
		baseq[x] = t2 | (t2 << 2) | left;
		left = toMSSB[t >> 29];
	    }
	}
	break;
    }
    case 8:
    {
	UInt32 left;
	UInt32 t;
	UInt32 t2;
	UInt32 t3;
	/* Similarly for 8. */
	
	baseq = (UInt32 *)dest;
	for (q = baseq + imageHeight*destPitch; baseq < q; baseq += destPitch)
	{
	    left = 0;
	    for (x = numWords-1; x >= 0; x--)
	    {
		t = baseq[x];
		t2 = t | (t << 1);
		t3 = t2 | (t2 << 2);
		baseq[x] = t3 | (t3 << 4) | left;
		left = toMSSB[t >> 25];
	    }
	}
	break;
    }
    case 16:
    {
	UInt32 left;
	UInt32 t;
	UInt32 t2;
	UInt32 t3;
	UInt32 t4;
	/*
	 * And for 16. Computing our own result is the same technique, but
	 * we have to be a bit careful with the carry because the toMSSB
	 * table is only 8 bits wide, so check the most-sig byte first,
	 * and only if it's zero does the next-most-sig byte matter.
	 */
	baseq = (UInt32 *)dest;
	for (q = baseq + imageHeight*destPitch; baseq < q; baseq += destPitch)
	{
	    left = 0;
	    for (x = numWords-1; x >= 0; x--)
	    {
		t = baseq[x];
		t2 = t | (t << 1);
		t3 = t2 | (t2 << 2);
		t4 = t3 | (t3 << 4);
		baseq[x] = t4 | (t4 << 8) | left;
		left = (t >> 24) ? (0xff | (toMSSB[t >> 25] << 8)) :
		    toMSSB[t >> 17];
	    }
	}
	break;
    }
    case 32:
    {
	UInt32 left;
	UInt32 t;
	/*
	 * 32 is interesting. At this point, only our LSSB and MSSB matter
	 * at all; everything else inside the word gets clobbered. So computing
	 * our own result is easy, but we have to jump through a few hoops
	 * to find our MSSB.
	 */
	baseq = (UInt32 *)dest;
	for (q = baseq + imageHeight*destPitch; baseq < q; baseq += destPitch)
	{
	    left = 0;
	    for (x = numWords-1; x >= 0; x--)
	    {
		t = baseq[x];
		baseq[x] = (UInt32)-((Int32)LSSB(t)) | left;
		left = (t >> 24) ? (0xffffff | (toMSSB[t >> 25] << 24)) :
		    ((t >> 16) ? (0xffff | (toMSSB[t >> 17] << 16)) :
		     ((t >> 8) ? (0xff | (toMSSB[t >> 9] << 8)) :
		      toMSSB[t >> 1]));
	    }
	}
	break;
    }
    case 64:
    {
	UInt32 left, left2;
	UInt32 t;
	/*
	 * 64 - we can carry 2 words left. Be a bit clever: if the next
	 * word is nonzero, then it doesn't matter at all what we carry
	 * out, as they're going to clobber it with all 1s anyway, so don't
	 * bother computing it. This will pay off if the image has a lot
	 * of 1s, which is why we're running *after* the dily phase...
	 * running later also helps our most-sig byte be nonzero more often,
	 * which means we don't do as many tests.
	 */
	baseq = (UInt32 *)dest;
	for (q = baseq + imageHeight*destPitch; baseq < q; baseq += destPitch)
	{
	    left = left2 = 0;
	    x = numWords-1;
	    while (x >= 0)
	    {
		t = baseq[x];
		baseq[x] = (UInt32)-((Int32)LSSB(t)) | left;
		left = t ? 0xffffffff : left2;
		x--;
		left2 = ((x >= 0) && baseq[x]) ? 0 : 
		    ((t >> 24) ? (0xffffff | (toMSSB[t >> 25] << 24)) :
		     ((t >> 16) ? (0xffff | (toMSSB[t >> 17] << 16)) :
		      ((t >> 8) ? (0xff | (toMSSB[t >> 9] << 8)) :
		       toMSSB[t >> 1])));
	    }
	}
	break;
    }
    case 128:
    {
	UInt32 left, left2, left3, left4;
	UInt32 t;
	/*
	 * 128 - we can carry 4 words left.
	 */
	
	baseq = (UInt32 *)dest;
	for (q = baseq + imageHeight*destPitch; baseq < q; baseq += destPitch)
	{
	    left = left2 = left3 = left4 = 0;
	    x = numWords-1;
	    while (x >= 0)
	    {
		t = baseq[x];
		baseq[x] = (UInt32)-((Int32)LSSB(t)) | left;
		x--;
		if (t)
		{
		    left = left2 = left3 = 0xffffffff;
		}
		else
		{
		    left = left2;
		    left2 = left3;
		    left3 = left4;
		}
		left4 = ((x >= 0) && baseq[x]) ? 0 : 
		    ((t >> 24) ? (0xffffff | (toMSSB[t >> 25] << 24)) :
		     ((t >> 16) ? (0xffff | (toMSSB[t >> 17] << 16)) :
		      ((t >> 8) ? (0xff | (toMSSB[t >> 9] << 8)) :
		       toMSSB[t >> 1])));
	    }
	}
	break;
    }
    default:
    {
	return ia_invalidParm;
    }
    }

    NORMAL_RETURN;
}


/*

$Log: i_dilbig.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.8  1998/01/13 23:47:37  ddavies
i_dilbig.c -> /product/ipcore/ipshared/src/RCS/i_dilbig.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1995/12/13 02:43:25  ejaquith
i_dilbig.c -> /product/ipcore/ipshared/src/RCS/i_dilbig.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/10/30  18:57:05  ejaquith
 * ejaquith on Mon Oct 30 10:55:46 PST 1995
 *
 * i_dilbig.c -> /product/ipcore/ipshared/src/RCS/i_dilbig.c,v
 *
 * Whoops.  Forgot that on the PC, memcpy() for source row to dest row
 * will only work if we pad the number of bytes out to be a multiple of 4.
 * (Which is of course always safe since our image rows are always word padded)
 *
 * Revision 1.5  1995/10/30  18:11:49  ejaquith
 * ejaquith on Mon Oct 30 10:10:26 PST 1995
 *
 * i_dilbig.c -> /product/ipcore/ipshared/src/RCS/i_dilbig.c,v
 *
 * Added functionality to i_rectDilate() to allow dilation of images
 * additionally by STRELs of dimens (width or height) 1,2,4 (supplements
 * current ability for 8,16,32,64,and,128).  Added helpful comments
 * for client.
 *
 * Revision 1.4  1995/09/14  19:00:04  ejaquith
 * i_dilbig.c -> /product/ipcore/ipshared/src/RCS/i_dilbig.c,v
 *
 * CommonCheck: Converted all routines that were private in files to be of form
 * ip_*(), which they weren't.  No functionality/noticeable changes.
 *
 * Revision 1.3  1995/03/29  01:57:29  ejaquith
 * ejaquith on Tue Mar 28 17:57:01 PST 1995
 *
 * i_dilbig.c -> /product/ipcore/ipshared/src/RCS/i_dilbig.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images.
 *
 * Revision 1.2  1995/03/15  01:08:55  ejaquith
 * i_dilbig.c -> /product/ipcore/ipshared/src/RCS/i_dilbig.c,v
 *
 * Updated indentation style
 * Killed many unneeded multiplications and extra variables throughout
 * file so dilation uses almost no multiplies now.
 * (Alas only minimally faster on SPARC).
 *
 * Revision 1.1  1995/03/01  22:11:35  ejaquith
 * Initial revision
 *

*/
