/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  i_dtrans.c -- IPShared distance transform public interface
 */


#include <stdio.h>
#include <stdlib.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

#include "dtrans.prv"
#include "shrip.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: i_dtrans.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Computes the 8 bit distance transform of a binary input image
 *
 * LongDescription:::
 *  Uses 4 or 8 neighbor "search" to compute the distance from either 1) a
 *  white pixel to the nearest black pixel, or 2) a black pixel to the nearest
 *  white pixel. Takes argument for assumed "color" (black or white) of the
 *  universe outside of the input image. Note that the distance saturates at
 *  255 in the output.
 *
 * Input:::
 *  srcImage::          pointer to binary input image data to be transformed
 *                      including frames
 *  w::                 Framed width of input image
 *  h::                 Framed height of input image
 *  sBpl::              Bytes per line in input image
 *  connectType::       Either cDT4Connected or cDT8Connected for 4 or 8
 *                      neighbors, respectively
 *  fromBlack::         If cTrue, computes distances of all white pixels from
 *                      nearest black pixels. Otherwise computes distances of
 *                      all black pixels from nearest white pixels.
 *  universeColor::     Assumed color of universe outside of the input image (1
 *                      for black, 0 for white).
 *  dBpl::              Bytes per line of output image
 *
 * Output:::
 *  destImage::         Pointer to output 8 bit image data including frames
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *  See standard morphology texts/papers for full description
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_distanceTransform (
    IPState    *pIPState,
    UInt8      *srcImage,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	connectType,
    Int32	fromBlack,
    Int32	universeColor,
    UInt8      *destImage,
    Int32	dBpl)
{
    PROLOGUE;
    


    switch (connectType)
    {
        case cDT4Connected:
	    __error__ = i_fourNeighborDistanceTransform (pIPState, srcImage, sBpl,
						       w, h, 
						       fromBlack, 
						       universeColor,
						       dBpl, destImage);
	    break;

	case cDT8Connected:
	    __error__ = i_eightNeighborDistanceTransform (pIPState,
							srcImage, sBpl,
							w, h, 
							fromBlack, 
							universeColor,  
							dBpl, destImage);
	    break;

	case cDT2HConnected:
	case cDT2VConnected:
	    __error__ = ia_noImpl;
	    break;

	default:
	    __error__ = ia_invalidParm;
	    break;
    }

    NORMAL_RETURN;

}  /* i_distanceTransform () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Computes the 8 bit, 8 neighbor distance transform of a binary input image
 *
 * LongDescription:::
 *  Uses 8 neighbor "search" to compute the distance from either 1) a
 *  white pixel to the nearest black pixel, or 2) a black pixel to the nearest
 *  white pixel. Takes argument for assumed "color" (black or white) of the
 *  universe outside of the input image. Note that the distance saturates at
 *  255 in the output.
 *
 * Input:::
 *  srcImage::          pointer to binary input image data to be transformed
 *                      including frames
 *  w::                 Framed width of input image
 *  h::                 Framed height of input image
 *  sBpl::              Bytes per line in input image
 *  fromBlack::         If cTrue, computes distances of all white pixels from
 *                      nearest black pixels. Otherwise computes distances of
 *                      all black pixels from nearest white pixels.
 *  universeColor::     Assumed color of universe outside of the input image (1
 *                      for black, 0 for white).
 *  dBpl::              Bytes per line of output image
 *
 * Output:::
 *  destImage::         Pointer to output 8 bit image data including frames
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *  See standard morphology texts/papers for full description
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_eightNeighborDistanceTransform (
    IPState    *pIPState,
    UInt8      *srcImage,
    Int32	sBpl,
    Int32	w,
    Int32	h,
    Int32	fromBlack,
    Int32	universeColor,
    Int32	dBpl,
    UInt8      *destImage)
{
    PROLOGUE;
#if 0
    UInt8  *srcRowP, *destRowP;
    Int32   row;

    srcRowP = srcImage + (cFrameBits * sBpl) + cFrameBytes;
    destRowP = destImage + (cFrameBits * dBpl) + cFrameBytes;
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;
#endif

    /* keep the compiler quiet */
    srcImage = srcImage;
    sBpl = sBpl;
    w = w;
    h = w;
    fromBlack = fromBlack;
    universeColor = universeColor;
    dBpl = dBpl;
    destImage = destImage;

    __error__ = ia_noImpl;
    NORMAL_RETURN;

}  /* i_eightNeighborDistanceTransform () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Computes the 8 bit, 4 neighbor distance transform of a binary input image
 *
 * LongDescription:::
 *  Uses 4 neighbor "search" to compute the distance from either 1) a
 *  white pixel to the nearest black pixel, or 2) a black pixel to the nearest
 *  white pixel. Takes argument for assumed "color" (black or white) of the
 *  universe outside of the input image. Note that the distance saturates at
 *  255 in the output.
 *
 * Input:::
 *  srcImage::          pointer to binary input image data to be transformed
 *                      including frames
 *  w::                 Framed width of input image
 *  h::                 Framed height of input image
 *  sBpl::              Bytes per line in input image
 *  fromBlack::         If cTrue, computes distances of all white pixels from
 *                      nearest black pixels. Otherwise computes distances of
 *                      all black pixels from nearest white pixels.
 *  universeColor::     Assumed color of universe outside of the input image (1
 *                      for black, 0 for white).
 *  dBpl::              Bytes per line of output image
 *
 * Output:::
 *  destImage::         Pointer to output 8 bit image data including frames
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *  See standard morphology texts/papers for full description
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_fourNeighborDistanceTransform (
    IPState    *pIPState,
    UInt8      *srcImage,
    Int32	sBpl,
    Int32	w,
    Int32	h,
    Int32	fromBlack,
    Int32	universeColor,
    Int32	dBpl,
    UInt8      *destImage)
{
    PROLOGUE;
    UInt8  *srcRowP, *destRowP;
    Int32   row = 0;


    /************************************************
    Philosophy of this implementation of the distance transform;
    or why did you turn such a simple concept into such a huge
    mess of strange code?

    As usual, the answer is: speed.  This implementation is
    about 1.7 times faster than (takes 0.6 the time as) the standard
    pixel at a time two pass version published everywhere.  Reads
    and writes are minimized by operating on 32-bit sized chunks of
    image at a time (that's 32 binary input pixels and 4 8-bit gray
    output pixels per memory op).  

    The standard method of computing the distance transform is a two
    pass algorithm.  The first pass is in raster order, considering
    only previously computed neighbors.  The second pass is anti-raster
    order considering only neighbors to the right and below.  While
    this is simple to describe, it is in fact difficult to implement
    with word-at-a-time memory ops as the unrolled inner loop for the
    ant-raster order pass needs branches within each output word (a
    bad thing to have in the inner loop).  So instead we decompose
    the algorithm and reformulate it in such a way that all operations
    can be done word-at-a-time with no inner loop branching.

    Several basic concepts are used to decompose the distance
    transform into a series of simpler operations.  One of these is
    "ramp building."  A ramp starts at 1 and increments by one for
    every linear pixel step moved.  The ramp resets to zero whenever
    an on pixel appears in the input (the distance from an on pixel
    to itself is zero).  If one were to compute a 1-D distance transform,
    the operation could be decomposed into left to right ramp building,
    followed by right to left ramp building where the final result is the
    minimum of the two ramp results.  It can be easily seen that the
    MIN operation can be built into the second ramp generation function
    so that in two passes of an image row, the 1-D distance transform is
    computed.

    For the binary sequence:

    1100001000000001111100010000011000000001

    The left to right ramp would be:

    0012340123456780000012301234500123456780

    The right to left ramp would be:

    0043210876543210000032105432100876543210

    And the minimum of the two:

    0012210123443210000012101232100123443210

    is the 1-D distance transform of the binary input.

    Note that there are 1's on the end of the binary row.  If there
    were zeros we'd be forced to ask "What color is the universe beyond
    the image" to compute the answer correctly (hence the universeColor
    parameter).  Another real problem is overflow/saturation in the output 
    image which is only 8 bits deep.  Extra computation is required to 
    ensure that values that would be greater than 255 are saturated at 255 
    without affecting other pixels.

    To compute a 2-D transform we must consider the rows above and below
    the current row being computed.  To do this we have to add two basic
    steps for middle rows: increment the previous result from the row
    above and take the minimum of this and the current row result (as
    shown above).  Note that the increment step also has to handle the 
    saturation problem.

    We decompose the problem again into
    raster and anti-raster scans of the input image. First we will scan
    each row in raster order looking only at left neighbors (building
    left to right ramps), then we will scan the row in anti-raster order
    building ramps from right to left, taking the minimum of these ramps
    and those from the raster order result, and finally taking the minimum
    of this and the previous row incremented by one to produce the raster
    order result.  After scanning each row from top to bottom, we scan from
    bottom to top taking the minimum of the current row with the row directly
    below incremented by one.  This produces the correct 4 neighbor distance
    transform and each step can be computed in 32 bit (or larger) chunks
    which reduces the memory bottleneck.
    ************************************************/


    srcRowP = srcImage + (cFrameBits * sBpl) + cFrameBytes;
    destRowP = destImage + (cFrameBits * dBpl) + (8 * cFrameBytes);
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

    /*** Scan the image in raster order ***/

    /* top row has no "real" upper neighbor */
    ip_leftNeighborDTRasRow (pIPState, srcRowP, w, 
                             cTrue, fromBlack, universeColor,
			     destRowP);
    ip_rightNeighborDTAntiRasRow (pIPState, srcRowP, w, 
				  fromBlack, universeColor, 
				  destRowP);

    /* middle rows have upper neighbors and are not edge rows */
    row++;
    for (; row < h - 1; row++)
    {
	srcRowP += sBpl;
	destRowP += dBpl;
	ip_leftNeighborDTRasRow (pIPState, srcRowP, w, 
				 cFalse, fromBlack, universeColor,
				 destRowP);
	ip_topRightNeighborDTAntiRasRow (pIPState, srcRowP, w, 
				         fromBlack, universeColor, 
				         dBpl, destRowP);
    }

    /* bottom row is an edge row */
    srcRowP += sBpl;
    destRowP += dBpl;
    ip_leftNeighborDTRasRow (pIPState, srcRowP, w, 
			     cTrue, fromBlack, universeColor,
			     destRowP);
    ip_topRightNeighborDTAntiRasRow (pIPState, srcRowP, w, 
				     fromBlack, universeColor, 
				     dBpl, destRowP);

    /* zero out the right frame which now contains garbage */
    ip_fixRightFrameRow (pIPState, destRowP, w);

    /*** Now scan the image in anti-raster order ***/

    /* skip the bottom row which should now be correct */
    row--;
    for (; row >= 0; row--)
    {
	srcRowP -= sBpl;
	destRowP -= dBpl;
	ip_bottomNeighborDTRasRow (pIPState, w, dBpl, destRowP);

	/* zero out the right frame which now contains garbage */
        ip_fixRightFrameRow (pIPState, destRowP, w);
    }

    NORMAL_RETURN;

}  /* i_fourNeighborDistanceTransform () */



/*
 * $Log: i_dtrans.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:10  jlin
 * Imported sources
 *
 * Revision 1.3  1998/01/13 23:47:39  ddavies
 * i_dtrans.c -> /product/ipcore/ipshared/src/RCS/i_dtrans.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.2  1997/01/03 17:54:13  dferrell
 * i_dtrans.c -> /product/ipcore/ipshared/src/RCS/i_dtrans.c,v
 *
 * Removed unused parameters from call to ip_bottomNeighborDTRasRow ().
 * Made use of all parameters in i_eightNeighborDistanceTransform ()
 * stub to keep PC compilers happy.
 *
 * Revision 1.1  1996/12/31 22:01:14  dferrell
 * Initial revision
 *
 *
 */


