/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_erode.c--includes
 *
 *           fastErode(): upper level of fast erosion operation
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

    /* prototypes */
#include "shrmorph.pub"
#include "shrmorph.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrrast.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_erode.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/*
 *  i_fastErode():  Performs a hit-miss binary morphology erosion on a
 *			binary image using a structure element.
 *		    Result placed in memory pointed to by destImage.
 *
 *	This is a version of erode that uses a combination of c and
 *	assembler if the assembly is available.  If the assembly code is
 *	not around, it uses rasterOp calls just like the version in serode.c.
 *	The assembler will be missing if the structure element was too big.
 *
 *	Here is a sample structure element.  We'll discuss erode ops with
 *	this example,
 *	 | |H|H|
 *	 |M| |H|
 *	 |M|M| |
 *
 *	If we acheive expected performance gains, we'll put in C code
 *	that dynamically generates assembly code for structure elements.
 *
 *	The problem with existing morphological code is that the entire
 *	image is blitted (with possible shifts) for every hit or miss
 *	in the structure element.  To cut down on all this overhead,
 *	we want to do as much as possible with each word read from memory.
 *
 *	The SPARC reads 32 or 64 bit horizontal words from memory.  It can
 *	hold several of these words in registers at once.  Combining two
 *	words logically is cheap - 1 cycle.  Shifting a word horizontally
 *	is relatively expensive - two shifts, 1 OR.  Therefore, we'd like
 *	to combine an input word with all the output image words it could
 *	affect, then shift it, combine it with another set of output image
 *	words and so on.  This minimizes horizontal shifting.
 *
 *	For example, in the first structure element above (destination is
 *	in the center of the SE), we'd like to read a word and combine it
 *	with the word below (Hit) and the word above (Miss).  This retires
 *	the central column of the SE.  Then, we shift the input word right,
 *	filling the vacant bit with the bit from the input word on the
 *	right.  This shifted word is combined with the original word (Miss)
 *	and the word above (Miss).  This retires the left column of the SE.
 *	Finally, we shift the original input left and combine it with the
 *	output image words on the same row (Hit) and the row below (Hit).
 *
 *	Having done all that, which way do we move in the image?  We need to
 *	store some complete or partial results, possibly load new partial
 *	results, load a new input word and do all the combining.  The caches
 *	are set up for to allow the processor to efficiently stream through
 *	the image horizontally.  However, doing this means we would have to
 *	write out, and reload, partial results.  Each line of output would
 *	be computed with three horizontal passes through the input image, one
 *	using the line above as input, one using the line itself as input
 *	and one using the line below as input.  At each step of each pass,
 *	we'd have to read three words of partial results, read 1 word of
 *	input, do the calculations, then write three words of partial result.
 *
 *	We could also go through the image vertically.  Assume we start the
 *	loop with 2 partial results in registers.  At each step, we read the
 *	input word immediately below the last one read.  This will form the
 *	new partial result.  We also read one word of input and the words
 *	on either side of it.  We do the computations, then write out 1
 *	completed result.
 *
 *	In the horizontal pass scheme, we did 4 reads and 3 writes for each
 *	step (4*2 + 3*3 = 17 cycles).  In the vertical pass scheme, we do
 *	four reads and one write (4*2 + 1*3 = 11 cycles).  The vertical scheme
 *	also generalizes well to larger structure elements since we can usually
 *	do the same four reads (new partial result, new input, words on side
 *	of center) and one write.  This will be limited by the number of
 *	registers available to hold partial results and the number of bits to
 *	be shifted horizontally (more than 32 => need more inputs on the
 *	sides).  All this makes us favor the vertical approach.
 *
 *	A pure vertical approach to large images would generate a lot of
 *	cache misses.  There are enough lines in a large image that an access
 *	to one word in each line would cause the cache to be overfilled.
 *	Each access to a new vertical line would cause a cache miss.  At
 *	24 cycles/miss, this makes the vertical pass scheme a loser.
 *
 *	There is an easy fix for this problem.  We divide the image into
 *	blocks that are one word wide and, say, 128 lines tall.  We process
 *	a block using the vertical pass method, then do the next block to
 *	the right.  Processing the first block will cause lots of cache
 *	misses, but the second block's data will already be in the cache.
 *	This sort of hybrid image traversal preserves the efficient portions
 *	of both horizontal and vertical schemes.
 *
 *	How would we like to divide functionality between C and assembly
 *	code?  Presumably, C code is easier to write and already understands
 *	how arguments should be passed.  Assembly code is faster.  Assembly
 *	code for specific structure elements is be generated by C Code, so
 *	we'd rather not generate a lot of it.
 *
 *	We will have the assembly code process the vertical image blocks
 *	and have the C code do everything else.  Assuming the C Compiler is
 *	reasonably efficient, the difference between C Code and assembler on
 *	the occasional overhead stuff won't be noticeable.
 *
 *	To keep things simple, the assembly code will assume that images are
 *	padded with 0's.  This allows it to shift in the right stuff without
 *	special cases on the edges.  However, we must reproduce the pads in
 *	the output image.  We do this by simply leaving them alone.  The pads
 *	are assumed to be cleared when the pixrect is initialized and never
 *	touched again.
 *
 *	For the 386 world, the major change is in how the blocks are set up.
 *	The 386 code processes the entire image as one large block.  Therefore,
 *	We set curBlockHeight and blockWidthInBytes to be the entire image
 *	instead of one block.
 *
 *	Long Erodes, 12/11/96: We're adding the ability to do erosion with
 *	very long words (>= 64 bits).  The first instance of this will be
 *	using Intel's MMX opcodes.
 *
 *	There are several conditions that must be required for this to work.
 *	1) The current platform must support long operations.  2) The
 *	source and destination images being passed in must meet alignment
 *	constraints.  3) We must have assembly code for long operations.
 *	The first two conditions are checked using a call to
 *	ip_checkImageAccess.  The last one is checked by seeing if
 *	we have a pointer to a long erode procedure.  If all conditions
 *	are true, we call the long erode op.  If not, we call a regular
 *	erode op, if present.  If there's no regular erode op and we can't
 *	do a long erode op, we use bitblts.
 */

/* This is a public procedure */
/* PUBLIC */
/*
 *  i_fastErode():  Performs a hit-miss binary morphology erosion on a
 *			binary image using a structure element.
 *
 *	destImage:		Ptr to upper left corner of dest image's frame
 *	srcImage:		Ptr to upper left corner of src image's frame
 *	erodeOp:		Ptr to optimized assembly routine that does
 *				the real work.  If NULL, operation is done
 *				in C using rasterOp calls.  See i_genErodeAsm.
 *	longErodeOp:		Ptr to optimized assembly routine that does
 *				the real work when long (>= 64 bit) accesses
 *				are enabled.  If NULL, operation is done
 *				using erodeOp or in C using rasterOp calls.
 *				See i_genLongErodeAsm.
 *	imageWidth:		Width of image in pixels including frame
 *				and border BUT NOT pad.  This is the normal
 *				definition of width passed to all the "i_"
 *				routines.
 *	imageHeight:		Height of image in pixels including frame.
 *	srcPitchInBytes:	Number of bytes/line in source image including
 *				image, border, frame and pad.  This equals the
 *				number you add to a source address to move
 *				down one line.
 *	destPitchInBytes:	Ditto for dest image.
 *	blockWidthInBytes:	Parameter used by optimized assembly code.
 *				Returned by i_genErodeAsm when machine
 *				code is generated.
 *	longBlockWidthInBytes:	Parameter used by long optimized assembly code.
 *				Returned by i_genLongErodeAsm when machine
 *				code is generated.
 *	nx, ny:			Structure element has nx columns and ny rows.
 *	cx, cy:			The coordinates of the structure element's
 *				center relative to the upper left corner.
 *				(0,0) is the upper left element of the SE.
 *	strelData:		Pointer to an array of nx * ny bytes specifying
 *				the structure element.  Array is arranged so
 *				that the first ny elements contain the leftmost
 *				column of the SE from top to bottom.  The
 *				next ny elements have the next column and so on.
 *				Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *
 *	Return code:
 *		ia_successful	good completion
 *
 * Note that if you are linking with the "s" (slow) libararies, you must
 * always supply nx, ny, cx, cy and strelData.  erodeOp and longErodeOp
 * will be ignored.
 * If you are linking with the "f" libraries, the operation may be done
 * in several different ways:
 *   1)	longErodeOp is not NULL, the image is correctly aligned in
 *	memory, long operations are present and enabled on the
 *	current platform: the operation is done using the long
 *	assembly code.  All parameters are used.
 *   2) Condition 1 is false and erodeOp is not NULL: the operation is
 *	done using normal assembly code. cx, cy and StrelData not used.
 *   3)	Conditions 1 and 2 are false: operation is done using bitblts
 *	based on cx, cy and strelData.  All parameters are tested or used.
 */	
Int32  CDECL 
i_fastErode (
    IPState    *pIPState,
    UInt8      *destImage,
    UInt8      *srcImage,
    ErodeProc	erodeOp,
    LongErodeProc longErodeOp,
    Int32	imageWidth,
    Int32	imageHeight,
    Int32	srcPitchInBytes,
    Int32	destPitchInBytes,
    Int32	blockWidthInBytes,
    Int32	longBlockWidthInBytes,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    UInt8      *strelData)
{
    PROLOGUE;
    Int32	imageWidthInBytes;	/* number of non-pad bytes/line */
    Int32	addrOffset;		/* address offset to right
					 * of big block */
    Int32	remainingHeight;	/* number of lines left to process */
    Int32	curBlockHeight;		/* height of current blocks */
    UInt8      *srcBigBlkAddr, *dstBigBlkAddr;
					/* addr of upper left
					 * corner of leftmost block executed
					 * in current horizontal run
					 * of blocks */
    Bool	first;			/* boolean to tell if we've
					 * processed the first SE element. */
    Int32	index;			/* into the strelData array */
    Int32	row, col;		/* coords in SE */
    Int32	destXOffset, destYOffset; /* args to i_rasterOp */
    Int32	srcXOffset, srcYOffset; /* args to i_rasterOp */
    Int32	rectWidth, rectHeight;	/* args to i_rasterOp */
    Int32	remainingXOffset, remainingYOffset, remainingWidth;
    Int32	longestAccess;		/* longest legal access for current
					 * platform and image */
    Int32	cacheWidthInBytes, cacheHeight;
    Int32	numActiveCols;		/* the number of active columns
					 * in strel */
    UInt8      *pActiveCols = NULL;	/* an array of bytes holding hit and/or
					 * miss flags for each strel column. */
    
    /* special 386 variables */
    UInt32	destBlockAddrSpan, srcBlockAddrSpan;
    UInt32     *pInputCache = NULL;	/* ptr to cache of shifted inputs. */

	/* Initialize parameters we'll need to process the image. */
    srcBigBlkAddr = (UInt8 *)srcImage + 
		      (srcPitchInBytes << cLogFrameBits) + 
		      cFrameBytes;		/* skip the pad */
    dstBigBlkAddr = (UInt8 *) destImage +
		      (destPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */
    remainingHeight = imageHeight - 2 * cFrameBits;

	/* Before bothering to do anything else, check for the
	 * degenerate 1x1 case.  If we see it, just use i_rasterop.
	 * This case is a pain for the code generators and is faster
	 * this way anyway.
	 */
    if ((nx == 1) && (ny == 1))
    {
	if (strelData[0] == cHit)	/* just copy input to output */
	{
	    __error__ = i_rasterOp (pIPState, dstBigBlkAddr, 0, 0,
				  imageWidth-2*cFrameBits,
				  remainingHeight,
				  1, destPitchInBytes, PIX_SRC,
				  srcBigBlkAddr, 0, 0, 1, srcPitchInBytes);
	}
	else
	if (strelData[0] == cMiss)
	{
	    __error__ = i_rasterOp (pIPState, dstBigBlkAddr, 0, 0,
				  imageWidth-2*cFrameBits,
				  remainingHeight,
				  1, destPitchInBytes, PIX_NOT(PIX_SRC),
				  srcBigBlkAddr, 0, 0, 1, srcPitchInBytes);
	}
	NORMAL_RETURN;
    }

	/* imageWidthInBytes is only used by the accelerated code.
	 * It should be rounded up to the next multiple of 32 before
	 * being divided by 8 (it will later be divided by 4).
	 */
    imageWidthInBytes = ((imageWidth - 2 * cFrameBits + 31) & (Int32) 0xFFFFFFE0) / 8;

	/* If there's no strel, just return with invalid parameters */
    if ((nx <= 0) || (ny <= 0))
	return ia_invalidParm;

	/* Determine the longest memory access we're allowed on this
	 * platform with this image.
	 */
    ip_checkImageAccess (pIPState, srcBigBlkAddr, srcPitchInBytes,
			 dstBigBlkAddr, destPitchInBytes, &longestAccess);
    if ((longestAccess != 4) && (longestAccess != 8))
	return ia_internal;

	/* If ((a) we're not allowed to do a long erode op, or (b) there is no
	 * long erode op) AND (there is no regular erode op), then
	 * do the operation using rasterops */
    if (((longestAccess == 4) || (longErodeOp == NULL)) &&
	(erodeOp == NULL))
    {
	imageWidth -= 2 * cFrameBits;
	imageHeight -= 2 * cFrameBits;
	first = cTrue;			/* haven't seen and SE elements yet */
	for (index = 0, row = 0; row < ny; row++)
	{
	    for (col = 0; col < nx; col++, index++)
	    {
		    /* Where do all these weird offsets come from?  In an
		     * erode, we want to combine all of the pixels around the
		     * center with the center pixel.  For example, for
		     * a 3x3 Structure Element, if the element in the upper
		     * left corner is a hit, we want to shift the entire image
		     * down and to the right before doing a COPY or AND.  As
		     * the X (and Y) coords get bigger, we want to shift
		     * right by smaller distances.  By the time we're in the
		     * center's column, the image is shifted neither left nor
		     * right.  For elements to the right of the center's
		     * column, we shift the image to the left.  Therefore, the
		     * amount by which we shift right is cx - col;
		
		     * However, we don't want to specify rectangles that go
		     * off the source or destination image.  To shift the
		     * source image left, we use a positive offset in the
		     * source image (skipping the columns that would have
		     * fallen off the edge of the dest image) and an offset
		     * of 0 in the dest image.  Similar operations are used
		     * for shifting right or shifting up/down.  The width and
		     * height of the rectangle are modified to keep the right
		     * edges of the source and dest rects both legal.
		     *
		     * BUG FIX: That didn't quite work.  Since we carefully
		     * clipped the destination rectangle, some dest rows
		     * and/or cols wouldn't be touched by most operations.
		     * Those rows or cols may well have interesting stuff
		     * that should have been cleared out by the implied
		     * "sea of 0's" surrounding the image.  Or, if this
		     * is the miss AND it's the first SE element encountered,
		     * The area should be filled with 1's.
		     *
 		     * So, we turn each blit into two blits.  The first one
		     * does the normal clipped operation.  The second one
		     * depends on whether the SE op was a hit or miss:
		     *
 		     * SE op	FIRST?	What to do with rest of dest
		     *  cHit		 yes	Fill with 0's
		     *  cHit		 no	Fill with 0's (result of ANDing
		     *			with 0's = fill with 0's)
		     *  cMiss	 yes	Fill with 1's
		     *  cMiss	 no	Leave alone (result of ANDing
		     *			with 1's = do nothing.
		     */

		destXOffset = MAX(0, -col + cx);
		destYOffset = MAX(0, -row + cy);
		srcXOffset = MAX(0, col - cx);
		srcYOffset = MAX(0, row - cy);
		rectWidth = imageWidth - ABS(-col + cx);
		rectHeight = imageHeight - ABS(-row + cy);
		    /* set remainingXOffset and remainingYOffset to point
		     * to the part of the dest rect that wasn't touched
		     * by this operation */
	   	remainingXOffset = (destXOffset == 0 ? rectWidth : 0);
	   	remainingYOffset = (destYOffset == 0 ? rectHeight : 0);
	   	remainingWidth = imageWidth - rectWidth;
		remainingHeight = imageHeight - rectHeight;


		if (first == cTrue)	/* no blits yet */
		{
		    if (strelData[index] == 1)	/* first active element is a
						 * cHit, so copy the image.
						 * This will give us something
						 * to AND with in later steps */
		    {
			i_rasterOp (pIPState, dstBigBlkAddr,
				    destXOffset, destYOffset,
				    rectWidth, rectHeight,
				    1,			/* dest depth */
				    destPitchInBytes,
				    PIX_SRC,		/* op = copy */
				    srcBigBlkAddr,
				    srcXOffset, srcYOffset,
				    1,			/* src depth */
				    srcPitchInBytes);

		   	if (remainingWidth > 0)	/* clear cols we didn't
						 * touch */
			    i_rasterOp (pIPState, dstBigBlkAddr,
					remainingXOffset, 0,
					remainingWidth, imageHeight,
					1,		/* dest depth */
					destPitchInBytes,
					PIX_CLR,	/* op = clear */
					NULL,
					0, 0,
					1,		/* src depth */
					0);

		   	if (remainingHeight > 0)	/* clear rows we didn't
						 * touch */
			    i_rasterOp (pIPState, dstBigBlkAddr,
					0, remainingYOffset,
					imageWidth, remainingHeight,
					1,		/* dest depth */
					destPitchInBytes,
					PIX_CLR,	/* op = clear */
					NULL,
					0, 0,
					1,		/* src depth */
					0);
			first = cFalse;
		    }
		    else
		    if (strelData[index] == 2)	/* first active element is a
						 * cMiss, so copy inverted
						 * image. */
		    {
			i_rasterOp (pIPState, dstBigBlkAddr,
				    destXOffset, destYOffset,
				    rectWidth, rectHeight,
				    1,			/* dest depth */
				    destPitchInBytes,
				    PIX_NOT(PIX_SRC),	/* op = inverted copy */
				    srcBigBlkAddr,
				    srcXOffset, srcYOffset,
				    1,			/* src depth */
				    srcPitchInBytes);

		   	if (remainingWidth > 0)	/* set cols we didn't
						 * touch */
			    i_rasterOp (pIPState, dstBigBlkAddr,
					remainingXOffset, 0,
					remainingWidth, imageHeight,
					1,		/* dest depth */
					destPitchInBytes,
					PIX_SET,	/* op = set */
					NULL,
					0, 0,
					1,		/* src depth */
					0);

		   	if (remainingHeight > 0)	/* set rows we didn't
						 * touch */
			    i_rasterOp (pIPState, dstBigBlkAddr,
					0, remainingYOffset,
					imageWidth, remainingHeight,
					1,		/* dest depth */
					destPitchInBytes,
					PIX_SET,	/* op = set */
					NULL,
					0, 0,
					1,		/* src depth */
					0);
			first = cFalse;
		    }
		}  /* end of "first active SE element seen" clause */
		else
		if (strelData[index] == 1)
		{		/* should do src AND dst (hit) */
		    i_rasterOp (pIPState, dstBigBlkAddr,
				destXOffset, destYOffset,
				rectWidth, rectHeight,
				1,			/* dest depth */
				destPitchInBytes,
				PIX_SRC & PIX_DST,	/* op = AND */
				srcBigBlkAddr,
				srcXOffset, srcYOffset,
				1,			/* src depth */
				srcPitchInBytes);

		   if (remainingWidth > 0)	/* clear cols we didn't
						 * touch */
			i_rasterOp (pIPState, dstBigBlkAddr,
				    remainingXOffset, 0,
				    remainingWidth, imageHeight,
				    1,			/* dest depth */
				    destPitchInBytes,
				    PIX_CLR,		/* op = clear */
				    NULL,
				    0, 0,
				    1,			/* src depth */
				    0);

		    if (remainingHeight > 0)	/* clear rows we didn't
						 * touch */
			i_rasterOp (pIPState, dstBigBlkAddr,
				    0, remainingYOffset,
				    imageWidth, remainingHeight,
				    1,			/* dest depth */
				    destPitchInBytes,
				    PIX_CLR,		/* op = clear */
				    NULL,
				    0, 0,
				    1,			/* src depth */
				    0);
		}
		else
		if (strelData[index] == 2)
		{		/* should ~src AND dst (miss) */
		    i_rasterOp (pIPState, dstBigBlkAddr,
				destXOffset, destYOffset,
				rectWidth, rectHeight,
				1,			/* dest depth */
				destPitchInBytes,
				PIX_NOT(PIX_SRC)&PIX_DST, /* op = AND INVERSE */
				srcBigBlkAddr,
				srcXOffset, srcYOffset,
				1,			/* src depth */
				srcPitchInBytes);

			/* Leave any pixels we didn't touch alone.  This
			 * is the same as ANDing them with 1's, which is
			 * the formally correct thing to do here */
		}
	    }   /* end of col loop */
	}	/* end of row loop */
	    /* if we didn't do anything, set the output image to all 1's.
	     * A strel containing only Don't Care terms matches anywhere,
	     * so write 1's everywhere.
	     * This is done in the case where the original output image
	     * had not been cleared in anticipation of this routine filling
	     * it.
	     */
	if (first == cTrue)
	{
	    i_rasterOp (pIPState, dstBigBlkAddr,
		        0, 0,
		        imageWidth, imageHeight,
		        1,			/* dest depth */
		        destPitchInBytes,
		        PIX_SET,		/* op = set */
		        NULL,
		        0, 0,
		        0,			/* src depth */
		        0);
	}
    }		/* end of case that does erode using blits */
    else
    if ((longestAccess == 4) || (longErodeOp == NULL))
    {
	    /* Either because the current processor doesn't support
	     * long accesses or because the current image is not
	     * correctly aligned, we're going to use the regular
	     * assembly code.
	     */

	    /* The image is divided into major and minor
	     * blocks.  SPARCStation assembly code processes minor blocks, each
	     * cMaxBlockHeight tall and blockWidthInBytes wide.  PC assembly
	     * code's minor blocks are cMaxBlockHeight tall and 1 word wide.
	     * A horizontal concatenation of minor blocks that
	     * covers the image is a major block.
	     * The outer loop below does major blocks, the inner loop
	     * does minor blocks within major blocks.
	     * The SPARC assembly language routines take as arguments:
	     * (initial source addr, initial dest addr, srcPitchInBytes,
	     * destPitchInBytes, height, width).
	     * The PC assembly language routines use an cache of shifted
	     * inputs.  Memory for that cache is created here.  The PC
	     * routines take as arguments:
	     * (initial source addr, initial dest addr, srcPitchInBytes,
	     * destPitchInBytes, the source address at which the assembly
	     * code should stop processing, pInputCache address).
	     */
#ifdef _386_ARCH_
	pInputCache = (UInt32 *)CALLOC(nx*(ny+cMaxBlockHeight),
				    sizeof(UInt32));
	if (pInputCache == NULL)
	    return ia_nomem;

	blockWidthInBytes = 4;	/* width of all PC blocks */
#endif

	while (remainingHeight > 0)
	{
	    curBlockHeight = MIN(remainingHeight, cMaxBlockHeight);


#ifdef _386_ARCH_
	        /* figure out how far the source address will move
		 * when processing a block.  The final value of source
		 * address is passed to the 386 code as the end
		 * condition.
		 */
	    if (curBlockHeight == cMaxBlockHeight)
	        destBlockAddrSpan = destPitchInBytes << cLogBlockHeight;
	    else
		destBlockAddrSpan = curBlockHeight * destPitchInBytes;
#endif

	    for (addrOffset = 0;
		 addrOffset < imageWidthInBytes;
		 addrOffset += blockWidthInBytes)
	    {
#ifdef _SPARC_ARCH_	/* call routine with SPARC args */
		(erodeOp)(srcBigBlkAddr+addrOffset,
			  dstBigBlkAddr+addrOffset,
			  srcPitchInBytes, destPitchInBytes,
			  curBlockHeight,
			  MIN((imageWidthInBytes-addrOffset),
			      blockWidthInBytes) >> 2);
#endif
#ifdef _386_ARCH_	/* call routine with 386 args */
		(erodeOp)(srcBigBlkAddr+addrOffset,
			  dstBigBlkAddr+addrOffset,
			  srcPitchInBytes, destPitchInBytes,
			  dstBigBlkAddr+addrOffset+destBlockAddrSpan,
			  pInputCache);
#endif
	    }
		/* The assembly code could conceivably have written trash
		 * into the right frame.  This could only have happened if
		 * the first active element's copy operation shifted the
		 * image right.  Clear the bits now while the TLB still
		 * has the real addresses of the dest image. */
	    ip_clearRightFrame (pIPState, dstBigBlkAddr,
				imageWidth - 2 * cFrameBits,
			        curBlockHeight, destPitchInBytes, 1);
	    remainingHeight -= curBlockHeight;
	    srcBigBlkAddr += (srcPitchInBytes << cLogBlockHeight);
	    dstBigBlkAddr += (destPitchInBytes << cLogBlockHeight);
	}  /* end of loop processing image with asm code */
    }	/* end of case where we use regular asm code to process image */
    else
    {
	    /* we will use long assembly code to proces the image.
	     * This is a lot like the case above, but the IFDEFs
	     * get sort of crazy when we try to mix them.
	     */

	    /* The image is divided into major and minor
	     * blocks.  PC MMX assembly code's minor blocks are
	     * cMaxLongBlockHeight tall and 1 dword wide.
	     * A horizontal concatenation of minor blocks that
	     * covers the image is a major block.
	     * The outer loop below does major blocks, the inner loop
	     * does minor blocks within major blocks.
	     * The PC MMX assembly language routines use an cache of shifted
	     * inputs.  Memory for that cache is created here.  The PC
	     * routines take as arguments:
	     * (initial source addr, initial dest addr, srcPitchInBytes,
	     * destPitchInBytes, number of rows in input cache,
	     * height of block, pointer to input cache).
	     */
	pActiveCols = (UInt8 *)MALLOC(nx);
	if (pActiveCols == NULL)
	{
	    __error__ = ia_nomem;
	    goto fastCleanup;
	}

	__error__ = ip_getLongErodeSEParms (pIPState, nx, ny, strelData,
					    &cacheWidthInBytes, &cacheHeight,
					    &numActiveCols, &(pActiveCols[0]));
	if (__error__ != ia_successful)
	    goto fastCleanup;

	    /* Only allocate an input cache if it will be used.  It's
	     * not used in purely horizontal strels.
	     */
	if (ny > 1)
	{
	    pInputCache = (UInt32 *)CALLOC(cacheWidthInBytes*cacheHeight,
					   sizeof(UInt8));
	    if (pInputCache == NULL)
	    {
		__error__ = ia_nomem;
		goto fastCleanup;
	    }
	}

	    /* Move the source address up by cy lines.  In the regular
	     * 386 code, this is done by the generated code each
	     * time, but why bother?
	     */
	srcBigBlkAddr -= cy * srcPitchInBytes;
	srcBlockAddrSpan = srcPitchInBytes * cMaxLongBlockHeight;
	destBlockAddrSpan = destPitchInBytes * cMaxLongBlockHeight;

	    /* do whatever's necessary to enter long operation mode. */
	ip_initLongOps (pIPState);

	curBlockHeight = 0;	/* to keep compiler happy */
	while (remainingHeight > 0)
	{
	    curBlockHeight = MIN(remainingHeight, cMaxLongBlockHeight);
	    if (curBlockHeight < cMaxLongBlockHeight)
	    {
		    /* the input cache should be shorter too.
		     * However, it should stay a multiple of 2.
		     */
		cacheHeight -= (cMaxLongBlockHeight - curBlockHeight);
		cacheHeight = (cacheHeight + 1) & ~0x1;
	    }

	    for (addrOffset = 0;
		 addrOffset < imageWidthInBytes;
		 addrOffset += longBlockWidthInBytes)
	    {
		(longErodeOp)(srcBigBlkAddr+addrOffset,
			      dstBigBlkAddr+addrOffset,
			      srcPitchInBytes, destPitchInBytes,
			      cacheHeight, curBlockHeight,
			      pInputCache);
	    }
		/* The assembly code could conceivably have written trash
		 * into the right frame.  This could only have happened if
		 * the first active element's copy operation shifted the
		 * image right.  Clear the bits now while the TLB still
		 * has the real addresses of the dest image.
		 * Since we're using long words, we might affect two
		 * words of frame.  Therefore, we need to clear all
		 * the frame bits.  ip_clearRightFrame only clears
		 * the bits in the word that's next to the image.
		 * Let's do this all at once. */
	    i_rasterOp (pIPState, dstBigBlkAddr,
		        imageWidth - 2*cFrameBits, 0,	/* x, y */
		        cFrameBits, curBlockHeight,	/* w, h */
		        1,				/* dest depth */
		        destPitchInBytes,
		        PIX_CLR,			/* op = clear */
		        NULL,
		        0, 0,
		        0,				/* src depth */
		        0);

	    remainingHeight -= curBlockHeight;
	    srcBigBlkAddr += srcBlockAddrSpan;
	    dstBigBlkAddr += destBlockAddrSpan;
	}  /* end of loop processing image with asm code */

	    /* issue any opcodes necessary to exit from long access mode */
	ip_finishLongOps (pIPState);

	    /* Purely horizontal strels are always processed two lines at
	     * a time.  If that was a purely horizontal strel AND the
	     * height of the image was odd AND there are only MISSes in
	     * the strel, clear the top line of the bottom frame.
	     */
	if ((ny == 1) && (imageHeight & 0x1))
	{
		/* check the columns of the strel */
	    for (col = 0; col < nx; col++)
	    {
		if (pActiveCols[col] == cHit)
		    break;
	    }
	    if (col == nx)
	    {
		    /* there were no hits so clear the
		     * line just below the bottom of
		     * the image.  Since we added destBlockAddrSpan
		     * to dstBigBlkAddr at the bottom of the
		     * loop, subtract it back off.  Then move
		     * down by curBlockHeight so we're pointing
		     * to the line below the last one.
		     */
		i_rasterOp (pIPState, dstBigBlkAddr - destBlockAddrSpan,
			    0,
			    curBlockHeight,
			    imageWidth, 1,
			    1,			/* dest depth */
			    destPitchInBytes,
			    PIX_CLR,		/* op = clear */
			    NULL,
			    0, 0,
			    0,			/* src depth */
			    0);
	    }
	}
    }	/* end of case where we use long asm code to process image */

    __error__ = ia_successful;

fastCleanup:
    if (pInputCache != NULL)
	FREE(pInputCache);
    if (pActiveCols != NULL)
	FREE(pActiveCols);

    NORMAL_RETURN;
}

/*

$Log: i_erode.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.10  1998/01/13 23:47:41  ddavies
i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.9  1997/07/10 22:44:30  ddavies
ddavies on Thu Jul 10 15:42:34 PDT 1997

i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v

Add support for the degenerate 1x1 strel.  The code generators don't always
handle this case well since it's both purely horizontal and purely
vertical, yet these types of strels are supposed to be mutually exclusive.
A simple RasterOp is faster than the fancy generated code anyway.

Revision 1.8  1996/12/18 22:58:32  ddavies
i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v

Add ability to use long (>= 64 bit) operations to i_fastErode.  This means
adding the longErodeOp and longBlockWidthInBytes parameters, figuring out
whether long operations are allowed on this image and invoking the
long routines if everything's ok.  The code that invokes the long routines
is separated from the code that invokes the normal machine code to
reduce the number of ifdefs.  The code that uses rasterOps and short machine
code is basically unchanged.

 * Revision 1.7  1995/12/13  02:43:29  ejaquith
 * i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.6  1995/10/18  03:52:52  ddavies
 * i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.5  1995/10/03  21:11:16  ddavies
 * i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v
 *
 * Include shrmorph.pub/prv instead of shrip.pub/prv.
 *
 * Revision 1.4  1995/10/03  00:29:34  ddavies
 * i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v
 *
 * Replace the funky old declarations for runtime machine code procedures
 * void CDECL (*erodeOp)() with types.  The types are conditionally defined
 * in shrip.pub so they vary with platform.
 *
 * Revision 1.3  1994/12/08  04:33:27  ddavies
 * i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v
 *
 * Catch the case in which the structure element contained no hits or misses.
 * In this case, fill the dest images with 1's since the structure element
 * matches everywhere.
 *
 * Revision 1.2  1994/02/04  20:14:31  dferrell
 * i_erode.c -> /product/ipcore/ipshared/src/RCS/i_erode.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/05  03:55:30  ddavies
 * Initial revision
 *
 * Revision 1.20  1993/12/02  19:20:53  ddavies
 * ferode.c -> /product/ipcore/ipshared/src/RCS/ferode.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.19  1993/11/05  01:24:09  ddavies
 * ferode.c -> /product/ipcore/ipshared/src/RCS/ferode.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


