/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_expand.c: contains calls to lower-level subroutines for
 *                expansion.
 *
 *         2x, 4x, 8x, 16x, 32x expansion procedures:
 *		Int32		i_expand2()
 *		Int32		i_expand4()
 *		Int32		i_expand8()
 *		Int32		i_expand16()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_expand.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* default: suppress all debugging */
#define  TIMING         0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING         1
#endif  /* DEBUG_ON */

/* The lookup tables used by the expansion code are generated in exptable.c
 */


/*********************************************************************
 ***************       2x Expansion Procedures       *****************
 ********************************************************************/

/* PUBLIC */
/*
 *  i_expand2()  
 *              Each pixel in the source image is mapped to a 2x2 square
 *              of 4 identical pixels in the destination.
 *
 *	Input:
 *		pD = 	destination image (pointer to image including frame)
 *		dBpl = 	bytes per line of destination (including frame)
 *		pS = 	source image (pointer to image including frame)
 *		wS = 	width of source in pixels including frame
 *		hS = 	height of source including frame
 *		depth =	depth of source and dest images
 * 		sBpl = 	bytes per line of source (including frame)
 *
 *	Return code:
 *		ia_successful	      good completion
 *		ia_nomem	      unable to allocate memory
 *              ia_depthNotSupported  bad table depth
 */
Int32  CDECL 
i_expand2 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	depth,
    Int32	sBpl)
{
PROLOGUE;
 
UInt16        *pTabExp2;

	/* make the table */
    pTabExp2 = NULL;
    if ((__error__ = i_makeExp2LUT (pIPState, depth, &pTabExp2)) != ia_successful)
	NORMAL_RETURN;
	    
	/* adjust the source and dest pointers to avoid the pads */
    pS = pS  + (sBpl << cLogFrameBits) + depth * cFrameBytes;
    pD = pD  + (dBpl << cLogFrameBits) + depth * cFrameBytes;

    ip_expand2Asm (pIPState, pS, sBpl, pD, dBpl, (wS - 2 * cFrameBits) * depth,
		   hS - 2 * cFrameBits, pTabExp2);

    FREE(pTabExp2);
    NORMAL_RETURN;
}




/*********************************************************************
 ****************      4x  Expansion Procedures       ****************
 ********************************************************************/

/* PUBLIC */
/*
 *  i_expand4()
 *              Each pixel in the source image is mapped to a 4x4 square
 *              of 16 identical pixels in the dest.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 *		wS = width of source in pixels including frame
 *		hS = height of source including frame
 *		depth =	depth of source and dest images
 * 		sBpl = bytes per line of source (including frame)
 *
 *	Return code:
 *		ia_successful	      good completion
 *		ia_nomem	      unable to allocate memory
 *              ia_depthNotSupported  bad table depth
 */
Int32  CDECL 
i_expand4 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	depth,
    Int32	sBpl)
{
PROLOGUE;
 
UInt32          *pTabExp4;

	/* make the table */
    pTabExp4 = NULL;
    if ((__error__ = i_makeExp4LUT (pIPState, depth, &pTabExp4)) != ia_successful)
	NORMAL_RETURN;
	    
	/* adjust the source and dest pointers to avoid the pads */
    pS = pS  + (sBpl << cLogFrameBits) + depth * cFrameBytes;
    pD = pD  + (dBpl << cLogFrameBits) + depth * cFrameBytes;

    ip_expand4Asm (pIPState, pS, sBpl, pD, dBpl, (wS - 2 * cFrameBits) * depth,
		   hS - 2 * cFrameBits, pTabExp4);

    FREE(pTabExp4);
    NORMAL_RETURN;
}


/*********************************************************************
 ****************      8x  Expansion Procedures       ****************
 ********************************************************************/
/* PUBLIC */
/*
 *  i_expand8()
 *              Each pixel in the source image is mapped to an 8x8 square
 *              of 64 identical pixels in the dest.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 *		wS = width of source in pixels including frame
 *		hS = height of source including frame
 *		depth =	depth of source and dest images
 * 		sBpl = bytes per line of source (including frame)
 *
 *	Return code:
 *		ia_successful	      good completion
 *		ia_nomem	      unable to allocate memory
 *              ia_depthNotSupported  bad table depth
 */
Int32  CDECL 
i_expand8 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	depth,
    Int32	sBpl)
{
PROLOGUE;
 
UInt32          *pTabExp8;

	/* make the table */
    pTabExp8 = NULL;
    if ((__error__ = i_makeExp8LUT (pIPState, depth, &pTabExp8)) != ia_successful)
	NORMAL_RETURN;
	    
	/* adjust the source and dest pointers to avoid the pads */
    pS = pS  + (sBpl << cLogFrameBits) + depth * cFrameBytes;
    pD = pD  + (dBpl << cLogFrameBits) + depth * cFrameBytes;

    ip_expand8Asm (pIPState, pS, sBpl, pD, dBpl, (wS - 2 * cFrameBits) * depth,
		   hS - 2 * cFrameBits, pTabExp8);

    FREE(pTabExp8);
    NORMAL_RETURN;
}


/*********************************************************************
 ****************     16x  Expansion Procedures       ****************
 ********************************************************************/

/* PUBLIC */
/*
 *  i_expand16()
 *              Each source pixel is mapped to a 16x16 square
 *              of 256 identical pixels in the dest.
 *
 *	Input:
 *		pD =	destination image (pointer to image including frame)
 *		dBpl =	bytes per line of destination (including frame)
 *		pS =	source image (pointer to image including frame)
 *		wS =	width of source in pixels including frame
 *		hS =	height of source including frame
 *		depth =	number of bits/pixel
 * 		sBpl =	bytes per line of source (including frame)
 *
 *	Return code:
 *		ia_successful         good completion
 *		ia_nomem              unable to allocate memory
 *              ia_depthNotSupported  bad table depth
 */
Int32  CDECL 
i_expand16 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	depth,
    Int32	sBpl)
{
PROLOGUE;
 UInt32		*pTabExp16;


	/* make the table */
    pTabExp16 = NULL;
    if ((__error__ = i_makeExp16LUT (pIPState, depth, &pTabExp16)) !=
								ia_successful)
	NORMAL_RETURN;
	    
	/* adjust the source and dest pointers to avoid the pads */
    pS = pS  + (sBpl << cLogFrameBits) + depth * cFrameBytes;
    pD = pD  + (dBpl << cLogFrameBits) + depth * cFrameBytes;

    ip_expand16Asm (pIPState, pS, sBpl, pD, dBpl, (wS - 2 * cFrameBits) * depth,
		    hS - 2 * cFrameBits, pTabExp16);

    FREE(pTabExp16);
    NORMAL_RETURN;
}

/*

$Log: i_expand.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.6  1998/01/13 23:47:42  ddavies
i_expand.c -> /product/ipcore/ipshared/src/RCS/i_expand.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1995/12/13 02:43:33  ejaquith
i_expand.c -> /product/ipcore/ipshared/src/RCS/i_expand.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/09/17  17:52:33  ddavies
 * i_expand.c -> /product/ipcore/ipshared/src/RCS/i_expand.c,v
 *
 * Change i_expand2, i_expand4, i_expand8 and i_expand16 so they call
 * i_makeExp2LUT, i_makeExp4LUT, i_makeExp8LUT and i_makeExp16LUT instead of
 * ip_makeExp2LUT, ip_makeExp4LUT, ip_makeExp8LUT and ip_makeExp16LUT
 * respectively.  At some point, we should probably allow these procedures
 * to take tables passed in by clients, but not now.
 *
 * Revision 1.3  1994/07/27  15:47:12  ejaquith
 * i_expand.c -> /product/ipcore/ipshared/src/RCS/i_expand.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Modified all function calls to add pointers to argument list in place
 * of globals.
 *
 * Revision 1.2  1994/02/04  20:15:04  dferrell
 * dferrell on Fri Feb 4 12:14:53 PST 1994
 *
 * i_expand.c -> /product/ipcore/ipshared/src/RCS/i_expand.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/06  03:27:26  ddavies
 * Initial revision
 *
 * Revision 1.16  1993/12/02  19:20:58  ddavies
 * fexpand.c -> /product/ipcore/ipshared/src/RCS/fexpand.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.15  1993/11/05  01:24:17  ddavies
 * fexpand.c -> /product/ipcore/ipshared/src/RCS/fexpand.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


