/*****************************************************************
 *  Copyright (c) 1988, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_expndm.c: contains calls to assembler-level subroutines for
 *                expand-and-mask operations.
 *
 *         2x, 4x, 8x expand and mask procedures:
 *		Int32		i_expandAndMask2()
 *		Int32		i_expandAndMask4()
 *		Int32		i_expandAndMask8()
 *
 *	   No expansions larger than 8x are implemented here because
 *	   no one seems to use them, especially for masking full scale
 *	   images.
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_expndm.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* default: suppress all debugging */
#define  TIMING         0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING         1
#endif  /* DEBUG_ON */

/* The lookup tables used by the expansion code are generated in exptable.c
 *
 * A common operation in our style of image processing is to make a mask
 * at some reduced scale, expand it to full scale and use it to mask the
 * original image.  This requires another full scale image.  In the
 * interest of saving memory, we'll introduce a new set of operations that
 * expand the image piece by piece, mask the original with the expanded
 * pieces, then toss the expanded image.
 *
 * These routines will work a lot like the expansion routines, except the
 * destination pointers point to the image that will be masked.
 *
 * We depend on the fact that reduction always rounds down in image size.
 * This means the number of lines in the expanded image will always be less
 * than or equal to the number of lines in the destination image.  That
 * means we don't need to compute the minimum of these numbers.
*/

/*********************************************************************
 ************       2x Expand and Mask Procedures       **************
 ********************************************************************/

/* PUBLIC */
/*
 *  i_expandAndMask2()  
 *              Each pixel in the source pixrect is mapped to a 2x2 square
 *              of 4 identical pixels in the destination.  The original
 *		destination pixels are ANDed with the expanded source
 *		pixels, which are then discarded.
 *
 *	Input:
 *		pD = 	destination image (pointer to image including frame)
 *		dBpl = 	bytes per line of destination (including frame)
 *		pS = 	source image (pointer to image including frame)
 *		wS = 	width of source in pixels including frame
 *		hS = 	height of source including frame
 * 		sBpl = 	bytes per line of source (including frame)
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 */
Int32  CDECL 
i_expandAndMask2 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl)
{
PROLOGUE;
 
UInt16        *pTabExp2;

	/* make the table */
    pTabExp2 = NULL;
    if ((__error__ = i_makeExp2LUT (pIPState, 1, &pTabExp2)) != ia_successful)
	NORMAL_RETURN;
	    
	/* adjust the source and dest pointers to avoid the pads */
    pS = pS  + (sBpl << cLogFrameBits) + cFrameBytes;
    pD = pD  + (dBpl << cLogFrameBits) + cFrameBytes;

    ip_expandAndMask2Asm (pIPState, pS, sBpl, pD, dBpl, wS - 2 * cFrameBits,
		          hS - 2 * cFrameBits, pTabExp2);

    FREE(pTabExp2);
    NORMAL_RETURN;
}


/*********************************************************************
 *************      4x  Expand and Mask Procedures       *************
 ********************************************************************/

/* PUBLIC */
/*
 *  i_expandAndMask4()
 *              Each pixel in the source pr is mapped to a 4x4 square
 *              of 16 identical pixels in the dest.  The original
 *		destination pixels are ANDed with the expanded source
 *		pixels, which are then discarded.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 *		wS = width of source in pixels including frame
 *		hS = height of source including frame
 * 		sBpl = bytes per line of source (including frame)
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 */
Int32  CDECL 
i_expandAndMask4 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl)
{
PROLOGUE;
 
UInt32          *pTabExp4;

	/* make the table */
    pTabExp4 = NULL;
    if ((__error__ = i_makeExp4LUT (pIPState, 1, &pTabExp4)) != ia_successful)
	NORMAL_RETURN;
	    
	/* adjust the source and dest pointers to avoid the pads */
    pS = pS  + (sBpl << cLogFrameBits) + cFrameBytes;
    pD = pD  + (dBpl << cLogFrameBits) + cFrameBytes;

    ip_expandAndMask4Asm (pIPState, pS, sBpl, pD, dBpl, wS - 2 * cFrameBits,
		          hS - 2 * cFrameBits, pTabExp4);

    FREE(pTabExp4);
    NORMAL_RETURN;
}



/*********************************************************************
 *************      8x  Expand and Mask Procedures       *************
 ********************************************************************/
/* PUBLIC */
/*
 *  i_expandAndMask8()
 *              Each pixel in the source image is mapped to an 8x8 square
 *              of 64 identical pixels in the dest.  The original
 *		destination pixels are ANDed with the expanded source
 *		pixels, which are then discarded.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 *		wS = width of source in pixels including frame
 *		hS = height of source including frame
 * 		sBpl = bytes per line of source (including frame)
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 */
Int32  CDECL 
i_expandAndMask8 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl)
{
PROLOGUE;
 
UInt32          *pTabExp8;

        /* make the table */
    pTabExp8 = NULL;
    if ((__error__ = i_makeExp8LUT (pIPState, 1, &pTabExp8)) != ia_successful)
        NORMAL_RETURN;
	    
	/* adjust the source and dest pointers to avoid the pads */
    pS = pS  + (sBpl << cLogFrameBits) + cFrameBytes;
    pD = pD  + (dBpl << cLogFrameBits) + cFrameBytes;

    ip_expandAndMask8Asm (pIPState, pS, sBpl, pD, dBpl, wS - 2 * cFrameBits,
		          hS - 2 * cFrameBits, pTabExp8);

    FREE(pTabExp8);
    NORMAL_RETURN;
}

/*

$Log: i_expndm.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.6  1998/01/13 23:47:44  ddavies
i_expndm.c -> /product/ipcore/ipshared/src/RCS/i_expndm.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1995/12/13 02:43:36  ejaquith
i_expndm.c -> /product/ipcore/ipshared/src/RCS/i_expndm.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/09/17  17:55:01  ddavies
 * i_expndm.c -> /product/ipcore/ipshared/src/RCS/i_expndm.c,v
 *
 * Change i_expandAndMask2, i_expandAndMask4, i_expandAndMask8 and
 * i_expandAndMask16 to use the newly public procedures i_makeExp2LUT,
 * i_makeExp4LUT and i_expandAndMask8 (Ooops, there is no i_expandAndMask16).
 * They used to use the ip_ versions of the same table making procs.
 * At some point, these procs should probably be converted so they can take
 * tables passed in by clients, but not now.
 *
 * Revision 1.3  1994/07/27  15:48:24  ejaquith
 * i_expndm.c -> /product/ipcore/ipshared/src/RCS/i_expndm.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Modified all function calls to add pointers to argument list in place
 * of globals.
 *
 * Revision 1.2  1994/02/04  20:15:11  dferrell
 * i_expndm.c -> /product/ipcore/ipshared/src/RCS/i_expndm.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/06  03:28:02  ddavies
 * Initial revision
 *
 * Revision 1.8  1993/12/02  19:21:02  ddavies
 * fexpandm.c -> /product/ipcore/ipshared/src/RCS/fexpandm.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.7  1993/11/05  01:24:26  ddavies
 * fexpandm.c -> /product/ipcore/ipshared/src/RCS/fexpandm.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


