/*****************************************************************
 *  Copyright (c) 1995.                              		 *
 *  Xerox Corporation.  All rights reserved.                     *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_exprc.c: fast functions for binary row/column expansion
 *
 *	Public:
 *		Int32		i_expandRows ()
 *		Int32		i_expandCols ()
 *
 *
 */

#include <string.h>

#include "types.pub"
#include "ipscore.prv"

#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"

#include "iopbyte.prv"

#include "memory.pub"
#include "shrscal.pub"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_exprc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/**********************************************************************
 * i_expandRows
 *		replicate each row in a pixr factor times
 *
 * Arguments:
 *		pD = 	destination image (pointer to image including frame)
 *		dBpl = 	bytes per line of destination (including frame)
 *		pS = 	source image (pointer to image including frame)
 *		wS = 	width of source in pixels including frame
 *		hS = 	height of source including frame
 *		depth =	depth of source and dest images
 * 		sBpl = 	bytes per line of source (including frame)
 *
 *	Return code:
 *		ia_successful	      good completion
 *		ia_nomem	      unable to allocate memory
 *
 **********************************************************************/
Int32 CDECL 
i_expandRows (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    Int32 f, wSb;

    /* quick sanity check */
    if (factor != 2 && factor != 4 && factor != 8 && factor != 16)
	return (ia_invalidParm);

    /* adjust base pointers to skip the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    hS -= 2 * cFrameBits;

    wSb = (wS + 7 ) / 8;

    /* loop through the source rows */
    for (;hS > 0; hS--)
    {
        /* duplicate each source row factor times in the destination */
        for (f = factor; f > 0; f--)
	{
	    MEMCPY (pD, pS, wSb);
	    pD += dBpl;
	}
	pS += sBpl;
    }

    NORMAL_RETURN;

}  /* i_expandRows () */


/**********************************************************************
 * i_expandCols
 *		replicate each column in a pixr factor times
 *
 * Arguments:
 *		pD = 	destination image (pointer to image including frame)
 *		dBpl = 	bytes per line of destination (including frame)
 *		pS = 	source image (pointer to image including frame)
 *		wS = 	width of source in pixels including frame
 *		hS = 	height of source including frame
 *		depth =	depth of source and dest images
 * 		sBpl = 	bytes per line of source (including frame)
 *
 * Returns:
 *		ia_successful	      good completion
 *		ia_nomem	      unable to allocate memory
 *
 * Notes:
 *		Uses the same expansion table making functions as the
 *		2D expansion functions.
 **********************************************************************/
Int32 CDECL 
i_expandCols (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    

    /* adjust base pointers to avoid the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;


    switch (factor)
    {
        case 2:
	{
	    UInt16        *pTabExp2;

	    /* make the table */
	    pTabExp2 = NULL;
	    if ((__error__ = i_makeExp2LUT (pIPState, 1, &pTabExp2)) !=
								ia_successful)
		NORMAL_RETURN;

	    __error__ = ip_expand2Cols (pIPState, pS, sBpl, pD, dBpl, 
	                              (wS - 2 * cFrameBits),
			              hS - 2 * cFrameBits, 
				      pTabExp2);
	
	    FREE(pTabExp2);
	}
	break;

	case 4:
	{
	    UInt32          *pTabExp4;

	    /* make the table */
	    pTabExp4 = NULL;
	    if ((__error__ = i_makeExp4LUT (pIPState, 1, &pTabExp4)) !=
								ia_successful)
		NORMAL_RETURN;
		    
	    __error__ = ip_expand4Cols (pIPState, pS, sBpl, pD, dBpl, 
	                              (wS - 2 * cFrameBits),
			              hS - 2 * cFrameBits, 
				      pTabExp4);
	
	    FREE(pTabExp4);
	}
	break;

	case 8:
	{
	    UInt32          *pTabExp8;

	    /* make the table */
	    pTabExp8 = NULL;
	    if ((__error__ = i_makeExp8LUT (pIPState, 1, &pTabExp8)) !=
								ia_successful)
		NORMAL_RETURN;
		    
	    __error__ = ip_expand8Cols (pIPState, pS, sBpl, pD, dBpl, 
				      (wS - 2 * cFrameBits),
				      hS - 2 * cFrameBits, 
				      pTabExp8);
	
	    FREE(pTabExp8);
	}
	break;

	case 16:
	{
	    UInt32		*pTabExp16;

	    /* make the table */
	    pTabExp16 = NULL;
	    if ((__error__ = i_makeExp16LUT (pIPState, 1, &pTabExp16)) !=
								ia_successful)
		NORMAL_RETURN;
	    
	    __error__ = ip_expand16Cols (pIPState, pS, sBpl, pD, dBpl, 
	                               (wS - 2 * cFrameBits),
			               hS - 2 * cFrameBits, 
				       pTabExp16);

	    FREE(pTabExp16);
	}
	break;

	default:
	    __error__ = ia_invalidParm;
	    break;
    }

    NORMAL_RETURN;

}  /* i_expandCols () */



/*
 $Log: i_exprc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:10  jlin
 Imported sources

 Revision 1.4  1998/01/13 23:47:45  ddavies
 i_exprc.c -> /product/ipcore/ipshared/src/RCS/i_exprc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1995/12/13 02:43:40  ejaquith
 i_exprc.c -> /product/ipcore/ipshared/src/RCS/i_exprc.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/03/01  18:01:48  dferrell
 * dferrell on Wed Mar 1 10:01:21 PST 1995
 *
 * i_exprc.c -> /product/ipcore/ipshared/src/RCS/i_exprc.c,v
 *
 * Added include of string.h so that Watcom could find the prototype for
 * memcpy.
 *
 * Revision 1.1  1995/02/27  23:59:57  dferrell
 * Initial revision
 *

 */

