/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_fill.c
 *
 *  Public routines:
 *      i_fillClipSeedTwoPass():  pretty fast two pass, 4-connected
 *				  fill operation based on C, as opposed to
 *				  assembly, routines
 *	i_fillClipSeed8Conn():	  8-connected, two pass fill operation
 *	i_fillClipSeedHorizontal(): fills in horizontal direction only
 *	i_fillClipSeedVertical(): fills in horizontal direction only
 *
 *  Private routines:
 *	ip_fillClipLine():	 	does a 4-connected fill of 1 one line.
 *	ip_fillClip8ConnLine():		does a 8-connected fill of 1 one line.
 *	ip_fillClipRightOnlyLine():	fills 1 line horizontally to the right.
 *	ip_fillClipLeftOnlyLine():	fills 1 line horizontally to the left.
 *	ip_fillClipHorizontalLine():	fills 1 line horizontally left and
 *					right..
 */


#include <stdio.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

    /* prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrrast.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_fill.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* Filling a seed into a mask is an essential part of morphological image
 * processing.  For example, text/image segmentation is normally done by
 * creating a seed image having pixels only in areas that are image and a
 * mask image containing a separate blob completely covering each individual
 * object that might possibly be image (usually including some text).  We
 * attempt to guarantee that every image area has at least one seed pixel and
 * non-image areas have no seed pixels.  Seeds in the seed image are grown
 * until they fill one or more of the distinct masked areas.  The areas
 * filled become the image masks.
 *
 * Being good scientists, we have been using fill algorithms that work for
 * arbitrary mask topologies.  Filling spirals is tricky.  The best algorithms
 * used so far are recursive.  From one seed pixel, you remember the current
 * location and look north to see if that one is vacant and inside the mask.
 * If so, start over there.  If not, do the same stuff starting east, south
 * and west.  Once no further progress is possible from the current location,
 * go back to the previous location and explore its remaining directions.
 * Continue until there's nowhere else to go.
 *
 * The algorithms illustrated is a 4-connected fill because pixels are deemed
 * to be attached only if they share a complete edge.  An 8-connected fill
 * also connects a pixel to its diagonal neighbors.  Dan Bloomberg says
 * he's never seen a significant difference in the results of 4- and
 * 8-connected fills in real images.  We'll cheat, doing bits and pieces of
 * 4- and 8-connected as is convenient.  For example, it's a real pain to do
 * an 8-connected fill that crosses word boundaries since it means reading
 * 3 words instead of 1.  On the other hand, it's a often a pain to restrict
 * oneself to a pure 4-connected fill because it means you must dilate the
 * current source word by 1 before ORing it with the word above/below.
 *
 * This method clearly has performance problems.  It colors the image one
 * *pixel* at a time. It maintains moderately complex state for an unbounded
 * number of individual pixels.  We'd like something that fills in groups of
 * pixels and doesn't spend so much time remembering where it is.
 *
 * Luc Vincent from XIS and Dan Davies talked about filling algorithms that
 * increase speed at the cost of precision.  The basic idea is to make
 * multiple passes through the whole image.  On each pass, seed pixels are
 * grown only in a restricted set of dimensions.  After two passes (say
 * one is top to bottom, the other bottom to top), a very large percentage
 * of all mask pixels are filled.  In fact, for purely convex masks (no curly
 * tendrils), all pixels will be filled.  This case includes all rectangular
 * masks, which make up the vast majority of image masks on real pages.
 *
 * There are several choices for how seeds could be grown on each pass.  One
 * could restrict oneself to two flavors of pass through the image,
 * left->right:top->bottom and right->left:bottom->top.  On the first type
 * of pass, a vacant pixel would be turned on if either the pixel above or
 * to the left (or above and to the left) were on and the candidiate pixel
 * were inside the mask.  During the second pass, a candidate pixel would be
 * turned on if either the pixel below or to the right (or below and to the
 * right) were on and the candidate were covered by the mask.
 *
 * This works pretty well.  However, in the first pass, things only grow
 * towards the lower right.  We would get better coverage if things grew
 * left and right from any seed, then towards the bottom in the first pass;
 * then left and right and towards the top in the second pass.  Since many of
 * our halftone masks have ragged edges, we'll try the second approach.  It's
 * fairly important to try to avoid extra passes.  After the first two passes,
 * the number of new pixels filled in per pass will be pretty small, so we'd
 * like to avoid these passes if possible.
 *
 * This routine only does two passes through the image.  This will suffice for
 * most images.  If a more complete fill is desired, one could execute this
 * routine some fixed number of times or count the number of pixels in
 * the destination after each call and stop when the number of new
 * pixels filled in went below some threshold.
 *
 * This routine takes as input a seed image, mask image and a destination
 * image.  Normally, it returns the seed and mask unchanged with the
 * destination containing the filled image.  Some callers may wish to fill
 * in place, making the seed and destination point to the same image.
 *
 * Normally, we initially set the dest image to SEED & MASK.  This gets rid
 * of any seed pixels that accidently spilled into the spaces between the
 * masks.  There are several places where it would be inefficient to explicitly
 * apply the mask when we're in the middle of the inner loops.  On the other
 * hand, this routine might be called iteratively to fill little tendrils.
 * In that case, we don't need to spend the extra time required to do the
 * SEED & MASK step after the first iteration.  The routine takes a variable
 * called clipFirst that's set when we should do the seed clipping step.
 *
 * One more point about high level routines.  An important routine is edgeFill.
 * This fills an image from an imaginary seed drawn around the outside of
 * the image, and masked by the mask.  Dan Bloomberg's edgeFill routine takes
 * two images, a source and a dest.  Sometimes, they're the same.  To use
 * this routine, higher level code would create new seed and mask images.
 * The mask would simply be the inverse of the input image.  The seed would be
 * an empty image with lines one pixel wide drawn around the outside.  The
 * destination image would be ORed with the source to produce the final
 * result.
 *
 * One might think we could save some time by just telling
 * i_fillClipSeedTwoPass to treat the 0's in the mask as the "valid" indicators
 * instead of the 1's.  That would be fine except it would be painful to
 * distinguish the frame's 0's from the image's 0's.  We'll just
 * require the caller to invert the mask first (which leaves the frame and
 * pad alone).  
 *
 * How is this routine structured?  It first copies the seed into the dest
 * image (unless they already point to the same image) and applies the mask
 * (unless clipFirst == cFalse).  It then calls a low level routine for each
 * line of the image.  The low level routine starts out processesing each
 * line right to left.  Originally, we went left to right, but moving
 * right to left lets us use a single variable for both the address offset
 * and the line counter.  For purposes of illustration, let's start out by
 * describing a state machine with four states:
 *
 *	1. Looking for some seed (right to left)
 *	  a. Candidiate seed word = (dest word above/below) OR 
 *	     (dest word) OR (previous result shifted left 1) OR
 *	     (that result shifted right 1).  That calculates an 8-connected
 *	     seed growth.
 *	  b. If the resulting word ANDed with the mask contains
 *	     any pixels, finish filling it, write it to memory, remember
 *	     the current location.  If the rightmost bit of the word written
 *	     was set, go to state 2.  If not and the leftmost bit of the
 *	     word written was on, go to state 3.  If neither end was active,
 *	     stay in state 1.
 *
 *	2. Back-Filling seed from left to right.
 *	     Extend the seed as far as possible to the right.  Do this by
 *	     first testing the next mask word to the right.  If it's all set,
 *	     just set the dest word to 1's and keep going right.  If the mask
 *	     has some off bits, fill in the leftmost bits.  Do this by reading
 *	     the dest word that's already there (in case something had
 *	     already been written there), ORing a 1 into the MSbit and
 *	     iteratively dilating and masking until the result no longer
 *	     changes.
 *
 *	     Note that there's no need to try to look at any of the dest
 *	     words in the neighborhood of the candidate dest word.  If
 *	     they had any legal seed bits, they would have been already have
 *	     been spread.  If their run had intersected the left->right run,
 *	     then the left->right run would never have started because the
 *	     right->left run would have covered the starting place.
 *
 *	     It's also not necessary to keep a count that detects the right end
 *	     of the line.  We can depend on the mask's frame to stop rightward
 *	     travel.
 *
 *	3. Filling seed from right to left.
 *	     Reload the address where step 1 stopped before we began the
 *	     left->right run.  Inspect the word written there.  If that word's
 *	     MSbit is off, go to step 1.  If the word's MSB is set, we need
 *	     to fill to the left.  Step left one word.  If the next dest word
 *	     is already filled with seed, go to step 4.  If the corresponding
 *	     mask word is all on, fill the dest word with all 1's and continue
 *	     with the step-right loop.  If neither of these is true, we need
 *	     to do a partial fill.  OR the dest word above/below here with
 *	     the current dest word.  OR in the LSB, which corresponds
 *	     to the connection to the word to the left.  Mask the result.
 *
 *	     Now, we do a trick that extends the seeds as far to the left as
 *	     they can go.  Add the mask to the seed.  Bits that are to the
 *	     left of seed bits, are under the mask, and are not themselves
 *	     seed bits turn to 0's.  The only other 0's are places that
 *	     aren't under the mask.  And the Mask with the complement of that
 *	     sum.  This produces 1's in all the new, non-seed bits that
 *	     should be turned on.  OR this with the seed.  In each section
 *	     of mask, there will now be 1's to the left of the LS bit of
 *	     the seed.  We need only dilate to the right to finish filling
 *	     in the seed.  Do this in a loop that shifts right 1, ORs with the
 *	     previous word, applies the mask and tests to see if anything has
 *	     changed.  When the word stops changing, we're done.  Write that
 *	     word.
 *
 *	     If the MSB of the word written was on, continue with step 3.
 *	     If not, go to step 1.
 *
 *	4. Skipping over filled areas.
 *	     Go through the image right to left.  So long as the dest words
 *	     read are all 1's, stay in step 4.  As soon as one is found with
 *	     0's, go to step 3. 
 *
 * This state machine is invoked once for each image line in a pass through
 * the image from top to bottom.  The process is repeated in a pass from
 * bottom to top.  That's it.  Most, if not all, of the bits of nice masks will
 * be filled in with these two passes.
 *
 * After some contemplation, we can find optimizations.  First, it's
 * nice to have a loop that just goes across the line without having to
 * back up.  Such a loop can have all of the pointer modification in one
 * place.  So let's convert the process of backfilling to the right into a
 * simple loop inside the "find Seed" state.  That way, we can localize weird
 * pointer backups.
 *
 * Second, there's a test right at the beginning of the "fill to left" state
 * that goes to the "skipping over already filled areas" state.  That state
 * has a very similar test that keeps it in the same state.  Why not just
 * have the "fill to left" state quickly decide to stay in the same state if
 * the dest word is already filled?  This simplifies things by eliminating
 * the skippng state.  The skipping state was a little messy anyway since
 * it had to figure out what to when it found something it shouldn't skip; 
 * either duplicate state 2's filling code or jump to the middle of state
 * 2.
 *
 * The C implementation of this function puts the line state machine in C.
 * The assembly implementation puts the state machine in assembler.
 */
/* JAC note.
 * Before any shifting is done (growing the seed) the candidate word
 * is masked.  This effectively keeps us from doing an 8-connected fill.
 * For example consider a destination with the following 2 lines:
 *	1 0 0 0
 *	0 0 0 0
 * And the corresponding mask for the 2 lines:
 *	1 0 0 0
 *	0 1 0 0
 * In processing the second line, we or the 2 lines together getting 
 * (1 0 0 0) but before we shift the bits left and right we apply the 
 * mask (0 1 0 0), getting (0 0 0 0).  Thus bits that would be turned 
 * on only because of diagonal bits will not be turned on.
 * If the mask for the second line was (1 1 0 0) the bit would be turned
 * on.  
 * We will call this a 4-connected fill the ultimate result is that
 * only the bits that a 4-connected fill would turn on are turned on
 * (even though it's impossible to say which line caused them to be turned
 * on).
 */


/* This is a public procedure */
/* PUBLIC */
/*
 *  i_fillClipSeedTwoPass()
 *		This is a two pass approximate filling algorithm.
 *		Given a seed and a mask, it makes one pass from
 *		top to bottom (left to right in each line).  Each
 *		time a seed pixel is found, it is extended as far
 *		as it will go left and right before coming to the edge
 *		of the mask or the edge of the image.  This blob is
 *		then extended as far as possible downwards before it
 *		comes to the edge of the mask or the edge of the
 *		image.  The second pass does the same thing from
 *		bottom to top.  Any purely convex shape will be
 *		filled in two passes.  Shapes that are primarily vertical
 *		with some diagonal zig-zags will be filled in two passes
 *		also.  Shapes with lots of vertical zig-zags (like a
 *		line of text filled with M's and W's, will generally
 *		require several calls to this routine.
 *		This procedure returns a writeCount, which equals the
 *		number of writes actually done.  When this goes to 0,
 *		further calls to i_fillClipSeedTwoPass will have no effect.
 *		It is legal to have the seed and destination image pointers
 *		point to the same image.  This feature is used to do
 *		multiple iterations of this routine.
 *
 *	pDestImage:		Pointer to upper left corner of frame in
 *				  destination image.
 *	pSeedImage:		Pointer to upper left corner of frame in
 *				  seed image.
 *	pMaskImage:		Pointer to upper left corner of frame in
 *				  mask image.
 *	destPitchInBytes:	Number of bytes/line in dest image including
 *				  border, frame and pad.  This is the number
 *				  one adds to the dest image pointer to
 *				  move down 1 line.
 *	seedPitchInBytes:	Ditto for seed image.
 *	maskPitchInBytes:	Ditto for mask image.
 *	imageWidth:		Width of image in pixels including image,
 *				  border and frame, but not pad.
 *	imageHeight:		Height of image including border and frame.
 *	clipFirst:		Boolean.  When cTrue (non-zero), the seed is
 *				  logically ANDed with the mask before any
 *				  filling is done.  This aids performance
 *				  because the inner loop can then expand
 *				  every seed it encounters without testing
 *				  to see if it's under a mask.  When cFalse,
 *				  this step is skipped.  One would set
 *				  clipFirst to cTrue on the first of a series
 *				  of iterative calls to i_fillClipSeedTwoPass
 *				  and cFalse thereafter. Since the mask step
 *				  was done in the first call, it is not
 *				  needed thereafter.
 *	pWriteCount:		Pointer to number of words written in the
 *				  image over the course of two passes.  If this
 *				  count is returned 0 (or the same as it
 *				  started) no filling was done and further
 *				  iterations will have no effect.
 *
 * This routine always returns a status code of ia_successful.
 */
Int32  CDECL 
i_fillClipSeedTwoPass (
    IPState    *pIPState,
    UInt8      *pDestImage,
    UInt8      *pSeedImage,
    UInt8      *pMaskImage,
    Int32	destPitchInBytes,
    Int32	seedPitchInBytes,
    Int32	maskPitchInBytes,
    Int32	imageWidth,
    Int32	imageHeight,
    Bool	clipFirst,
    Int32      *pWriteCount)
{
    PROLOGUE;
    UInt8      *destPtr, *maskPtr;	/* addr of left end of
					   lines currently being
					   processed */
    UInt8      *seedPtr;
    Int32	row;		  	/* iteration var in SE */

	/* Initialize parameters we'll need to process the image. */
    destPtr = (UInt8 *)pDestImage + 
		      (destPitchInBytes << cLogFrameBits) + 
		      cFrameBytes;		/* skip the pad */
    seedPtr = (UInt8 *) pSeedImage +
		      (seedPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */
    maskPtr = (UInt8 *) pMaskImage +
		      (maskPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */

    imageWidth -= 2 * cFrameBits;
    imageHeight -= 2 * cFrameBits;

	/* do initial step of seed to mask */
    if (seedPtr != destPtr)			/* copy seed to dest */
	i_rasterOp (pIPState, destPtr,
		    0, 0,			/* dest x, y offset */
		    imageWidth, imageHeight,
		    1,				/* dest depth */
		    destPitchInBytes,
		    PIX_SRC,			/* op = COPY */
		    seedPtr,
		    0, 0,			/* src x, y offset */
		    1,				/* src depth */
		    seedPitchInBytes);

	/* get rid of extraneous seeds.  This lets us avoid doing writes
	 * when we find that SEED & MASK = 0.  If we hadn't gotten rid of
	 * the extra bits here, we'd have to write every dest word in which
	 * it might be necessary to turn off erroneous seed bits.  It also
	 * lets us ignore all big blobs of seed.  Otherwise, we'd have to worry
	 * about whether these might lie completely outside mask areas.
	 */
    if (clipFirst)
	i_rasterOp (pIPState, destPtr,
		    0, 0,			/* dest x, y offset */
		    imageWidth, imageHeight,
		    1,				/* dest depth */
		    destPitchInBytes,
		    PIX_SRC & PIX_DST,		/* op = AND */
		    maskPtr,
		    0, 0,				/* src x, y offset */
		    1,				/* src depth */
		    maskPitchInBytes);

	/* Go through the rows top to bottom */
    for (row = 0, *pWriteCount = 0;
	 row < imageHeight;
	 row++, destPtr += destPitchInBytes, maskPtr += maskPitchInBytes)
    {
	*pWriteCount +=
	    ip_fillClipLine (pIPState, destPtr, destPitchInBytes, maskPtr,
			     (imageWidth+31) >> 5);
    }

	/* Go through the rows bottom to top */
    for (row--, destPtr -= destPitchInBytes, maskPtr -= maskPitchInBytes;
	 row >= 0;
	 row--, destPtr -= destPitchInBytes, maskPtr -= maskPitchInBytes)
    {
	*pWriteCount +=
	    ip_fillClipLine (pIPState, destPtr, -destPitchInBytes, maskPtr,
			     (imageWidth+31) >> 5);
    }
    NORMAL_RETURN;
}



/* PUBLIC */
/*
 *  i_fillClipSeed8Conn()
 *		This is a two pass 8 connected filling algorithm.
 *		Given a seed and a mask, it makes one pass from
 *		top to bottom (right to left in each line).  Each
 *		time a seed pixel is found, it is extended as far
 *		as it will go left and right before coming to the edge
 *		of the mask or the edge of the image.  This blob is
 *		then extended as far as possible downwards before it
 *		comes to the edge of the mask or the edge of the
 *		image.  The second pass does the same thing from
 *		bottom to top.  Any purely convex shape will be
 *		filled in two passes.  Shapes that are primarily vertical
 *		with some diagonal zig-zags will be filled in two passes
 *		also.  Shapes with lots of vertical zig-zags (like a
 *		line of text filled with M's and W's, will generally
 *		require several calls to this routine.
 *		This procedure returns a writeCount, which equals the
 *		number of writes actually done.  When this goes to 0,
 *		further calls to i_fillClipSeed8Conn will have no effect.
 *		It is legal to have the seed and destination image pointers
 *		point to the same image.  This feature is used to do
 *		multiple iterations of this routine.
 *		This is very similiar to i_fillClipSeedTwoPass except that
 *		this version is 8 connected.
 *
 *	pDestImage:		Pointer to upper left corner of frame in
 *				  destination image.
 *	pSeedImage:		Pointer to upper left corner of frame in
 *				  seed image.
 *	pMaskImage:		Pointer to upper left corner of frame in
 *				  mask image.
 *	destPitchInBytes:	Number of bytes/line in dest image including
 *				  border, frame and pad.  This is the number
 *				  one adds to the dest image pointer to
 *				  move down 1 line.
 *	seedPitchInBytes:	Ditto for seed image.
 *	maskPitchInBytes:	Ditto for mask image.
 *	imageWidth:		Width of image in pixels including image,
 *				  border and frame, but not pad.
 *	imageHeight:		Height of image including border and frame.
 *	clipFirst:		Boolean.  When cTrue (non-zero), the seed is
 *				  logically ANDed with the mask before any
 *				  filling is done.  This aids performance
 *				  because the inner loop can then expand
 *				  every seed it encounters without testing
 *				  to see if it's under a mask.  When cFalse,
 *				  this step is skipped.  One would set
 *				  clipFirst to cTrue on the first of a series
 *				  of iterative calls to i_fillClipSeed8Conn
 *				  and cFalse thereafter. Since the mask step
 *				  was done in the first call, it is not
 *				  needed thereafter.
 *	pWriteCount:		Pointer to number of words written in the
 *				  image over the course of two passes.  If this
 *				  count is returned 0 (or the same as it
 *				  started) no filling was done and further
 *				  iterations will have no effect.
 *
 * This routine always returns a status code of ia_successful.
 */
Int32  CDECL 
i_fillClipSeed8Conn (
    IPState    *pIPState,
    UInt8      *pDestImage,
    UInt8      *pSeedImage,
    UInt8      *pMaskImage,
    Int32	destPitchInBytes,
    Int32	seedPitchInBytes,
    Int32	maskPitchInBytes,
    Int32	imageWidth,
    Int32	imageHeight,
    Bool	clipFirst,
    Int32      *pWriteCount)
{
    PROLOGUE;
    UInt8      *destPtr, *maskPtr;	/* addr of left end of
					   lines currently being
					   processed */
    UInt8      *seedPtr;
    Int32	row;		  	/* iteration var in SE */

	/* Initialize parameters we'll need to process the image. */
    destPtr = (UInt8 *)pDestImage + 
		      (destPitchInBytes << cLogFrameBits) + 
		      cFrameBytes;		/* skip the pad */
    seedPtr = (UInt8 *) pSeedImage +
		      (seedPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */
    maskPtr = (UInt8 *) pMaskImage +
		      (maskPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */

    imageWidth -= 2 * cFrameBits;
    imageHeight -= 2 * cFrameBits;

	/* do initial step of seed to mask */
    if (seedPtr != destPtr)			/* copy seed to dest */
	i_rasterOp (pIPState, destPtr,
		    0, 0,			/* dest x, y offset */
		    imageWidth, imageHeight,
		    1,				/* dest depth */
		    destPitchInBytes,
		    PIX_SRC,			/* op = COPY */
		    seedPtr,
		    0, 0,				/* src x, y offset */
		    1,				/* src depth */
		    seedPitchInBytes);

	/* get rid of extraneous seeds.  This lets us avoid doing writes
	 * when we find that SEED & MASK = 0.  If we hadn't gotten rid of
	 * the extra bits here, we'd have to write every dest word in which
	 * it might be necessary to turn off erroneous seed bits.  It also
	 * lets us ignore all big blobs of seed.  Otherwise, we'd have to worry
	 * about whether these might lie completely outside mask areas.
	 */
    if (clipFirst)
	i_rasterOp (pIPState, destPtr,
		    0, 0,			/* dest x, y offset */
		    imageWidth, imageHeight,
		    1,				/* dest depth */
		    destPitchInBytes,
		    PIX_SRC & PIX_DST,		/* op = AND */
		    maskPtr,
		    0, 0,			/* src x, y offset */
		    1,				/* src depth */
		    maskPitchInBytes);

	/* Go through the rows top to bottom */
    for (row = 0, *pWriteCount = 0;
	 row < imageHeight;
	 row++, destPtr += destPitchInBytes, maskPtr += maskPitchInBytes)
    {
	*pWriteCount +=
	    ip_fillClip8ConnLine (pIPState, destPtr, destPitchInBytes, maskPtr,
			          (imageWidth+31) >> 5);
    }

	/* Go through the rows bottom to top */
    for (row--, destPtr -= destPitchInBytes, maskPtr -= maskPitchInBytes;
	 row >= 0;
	 row--, destPtr -= destPitchInBytes, maskPtr -= maskPitchInBytes)
    {
	*pWriteCount +=
	    ip_fillClip8ConnLine (pIPState, destPtr, -destPitchInBytes, maskPtr,
			          (imageWidth+31) >> 5);
    }
    NORMAL_RETURN;
}


/* PUBLIC */
/*
 *  i_fillClipSeedHorizontal()
 *		This is a one pass filling algorithm which grows the
 *		seeds into the destination using a horizontal
 *		fill.
 *		The direction is specified by the fillDirection parameter
 *		since a pixel may be filled if the pixel to its left is
 *		on or if the pixel to its right is on or either the pixel
 *		to its left or its right is on.
 *		Given a seed and a mask, it makes one pass from
 *		top to bottom (right to left in each line).  Each
 *		time a seed pixel is found, it is extended as far
 *		as it will go before coming to the edge
 *		of the mask or the edge of the image.  
 *		It is legal to have the seed and destination image pointers
 *		point to the same image.
 *
 *	pDestImage:		Pointer to upper left corner of frame in
 *				  destination image.
 *	pSeedImage:		Pointer to upper left corner of frame in
 *				  seed image.
 *	pMaskImage:		Pointer to upper left corner of frame in
 *				  mask image.
 *	destPitchInBytes:	Number of bytes/line in dest image including
 *				  border, frame and pad.  This is the number
 *				  one adds to the dest image pointer to
 *				  move down 1 line.
 *	seedPitchInBytes:	Ditto for seed image.
 *	maskPitchInBytes:	Ditto for mask image.
 *	imageWidth:		Width of image in pixels including image,
 *				  border and frame, but not pad.
 *	imageHeight:		Height of image including border and frame.
 *	fillDirection		propagate seed pixels to the right only,
 *				  to the left only, or in both directions
 *				  (cFillRight, cFillLeft, cFillRightLeft)
 *
 * This routine always returns a status code of ia_successful.
 */
Int32  CDECL 
i_fillClipSeedHorizontal (
    IPState    *pIPState,
    UInt8      *pDestImage,
    UInt8      *pSeedImage,
    UInt8      *pMaskImage,
    Int32	destPitchInBytes,
    Int32	seedPitchInBytes,
    Int32	maskPitchInBytes,
    Int32	imageWidth,
    Int32	imageHeight,
    Int32	fillDirection)
{
    PROLOGUE;
    UInt8      *destPtr, *maskPtr;	/* addr of left end of
					   lines currently being
					   processed */ 
    UInt8      *seedPtr;
    Int32	row;		  	/* iteration var in SE */

	/* Initialize parameters we'll need to process the image. */
    destPtr = (UInt8 *)pDestImage + 
		      (destPitchInBytes << cLogFrameBits) + 
		      cFrameBytes;		/* skip the pad */
    seedPtr = (UInt8 *) pSeedImage +
		      (seedPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */
    maskPtr = (UInt8 *) pMaskImage +
		      (maskPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */

    imageWidth -= 2 * cFrameBits;
    imageHeight -= 2 * cFrameBits;

	/* do initial step of seed to mask */
    if (seedPtr != destPtr)			/* copy seed to dest */
	i_rasterOp (pIPState, destPtr,
		    0, 0,			/* dest x, y offset */
		    imageWidth, imageHeight,
		    1,				/* dest depth */
		    destPitchInBytes,
		    PIX_SRC,			/* op = COPY */
		    seedPtr,
		    0, 0,			/* src x, y offset */
		    1,				/* src depth */
		    seedPitchInBytes);

	/* get rid of extraneous seeds.  This lets us avoid doing writes
	 * when we find that SEED & MASK = 0.  If we hadn't gotten rid of
	 * the extra bits here, we'd have to write every dest word in which
	 * it might be necessary to turn off erroneous seed bits.  It also
	 * lets us ignore all big blobs of seed.  Otherwise, we'd have to worry
	 * about whether these might lie completely outside mask areas.
	 */
    i_rasterOp (pIPState, destPtr,
		0, 0,				/* dest x, y offset */
		imageWidth, imageHeight,
		1,				/* dest depth */
		destPitchInBytes,
		PIX_SRC & PIX_DST,		/* op = AND */
		maskPtr,
		0, 0,				/* src x, y offset */
		1,				/* src depth */
		maskPitchInBytes);

	/* Go through the rows top to bottom */
    switch (fillDirection)
    {
    case cFillRight:
        for (row = 0; row < imageHeight;
	     row++, destPtr += destPitchInBytes, maskPtr += maskPitchInBytes)
	    ip_fillClipRightOnlyLine (pIPState, destPtr,
				      maskPtr, (imageWidth+31) >> 5);
	break;

    case cFillLeft:
        for (row = 0; row < imageHeight;
	     row++, destPtr += destPitchInBytes, maskPtr += maskPitchInBytes)
	    ip_fillClipLeftOnlyLine (pIPState, destPtr,
				     maskPtr, (imageWidth+31) >> 5);
	break;

    case cFillRightLeft:
        for (row = 0; row < imageHeight;
	     row++, destPtr += destPitchInBytes, maskPtr += maskPitchInBytes)
	    ip_fillClipHorizontalLine (pIPState, destPtr,
				       maskPtr, (imageWidth+31) >> 5);
	break;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*
 *  i_fillClipSeedVertical()
 *		This is a two pass filling algorithm which grows the
 *		seeds into the destination using a vertical
 *		fill.
 *		The direction is specified by the fillDirection parameter
 *		since a pixel may be filled if the pixel above it is
 *		on or if the pixel below it is on or either the pixel
 *		above it or below it is on.
 *		Given a seed and a mask, it makes one pass from
 *		top to bottom (right to left in each line).  Each
 *		time a seed pixel is found, it is extended as far
 *		as it will go before coming to the edge
 *		of the mask or the edge of the image. The second pass
 *		does the same thing from bottom to top. 
 *		It is legal to have the seed and destination image pointers
 *		point to the same image.
 *
 *	pDestImage:		Pointer to upper left corner of frame in
 *				  destination image.
 *	pSeedImage:		Pointer to upper left corner of frame in
 *				  seed image.
 *	pMaskImage:		Pointer to upper left corner of frame in
 *				  mask image.
 *	destPitchInBytes:	Number of bytes/line in dest image including
 *				  border, frame and pad.  This is the number
 *				  one adds to the dest image pointer to
 *				  move down 1 line.
 *	seedPitchInBytes:	Ditto for seed image.
 *	maskPitchInBytes:	Ditto for mask image.
 *	imageWidth:		Width of image in pixels including image,
 *				  border and frame, but not pad.
 *	imageHeight:		Height of image including border and frame.
 *	fillDirection		propagate seed pixels to the up only,
 *				  to the down only, or in both directions
 *				  (cFillUp, cFillDown, cFillUpDown)
 *
 * This routine always returns a status code of ia_successful.
 */
Int32  CDECL 
i_fillClipSeedVertical (
    IPState    *pIPState,
    UInt8      *pDestImage,
    UInt8      *pSeedImage,
    UInt8      *pMaskImage,
    Int32	destPitchInBytes,
    Int32	seedPitchInBytes,
    Int32	maskPitchInBytes,
    Int32	imageWidth,
    Int32	imageHeight,
    Int32	fillDirection)
{
    PROLOGUE;
    UInt8      *destPtr, *maskPtr;	/* addr of left end of
					   lines currently being
					   processed */
    UInt8      *seedPtr;
    Int32	row;		  	/* iteration var in SE */

	/* Initialize parameters we'll need to process the image. */
    destPtr = (UInt8 *)pDestImage + 
		      (destPitchInBytes << cLogFrameBits) + 
		      cFrameBytes;		/* skip the pad */
    seedPtr = (UInt8 *) pSeedImage +
		      (seedPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */
    maskPtr = (UInt8 *) pMaskImage +
		      (maskPitchInBytes << cLogFrameBits) +  
		      cFrameBytes;		/* skip the pad */

    imageWidth -= 2 * cFrameBits;
    imageHeight -= 2 * cFrameBits;

	/* do initial step of seed to mask */
    if (seedPtr != destPtr)			/* copy seed to dest */
	i_rasterOp (pIPState, destPtr,
		    0, 0,			/* dest x, y offset */
		    imageWidth, imageHeight,
		    1,				/* dest depth */
		    destPitchInBytes,
		    PIX_SRC,			/* op = COPY */
		    seedPtr,
		    0, 0,			/* src x, y offset */
		    1,				/* src depth */
		    seedPitchInBytes);

	/* get rid of extraneous seeds.  This lets us avoid doing writes
	 * when we find that SEED & MASK = 0.  If we hadn't gotten rid of
	 * the extra bits here, we'd have to write every dest word in which
	 * it might be necessary to turn off erroneous seed bits.  It also
	 * lets us ignore all big blobs of seed.  Otherwise, we'd have to worry
	 * about whether these might lie completely outside mask areas.
	 */
    i_rasterOp (pIPState, destPtr,
		0, 0,				/* dest x, y offset */
		imageWidth, imageHeight,
		1,				/* dest depth */
		destPitchInBytes,
		PIX_SRC & PIX_DST,		/* op = AND */
		maskPtr,
		0, 0,				/* src x, y offset */
		1,				/* src depth */
		maskPitchInBytes);

    switch (fillDirection)
    {
    case cFillUp:
            /* Go through the rows bottom to top */
        for (row = imageHeight, destPtr += (imageHeight-1)*destPitchInBytes, maskPtr += (imageHeight-1)*maskPitchInBytes;
             row > 0;
             row--, destPtr -= destPitchInBytes, maskPtr -= maskPitchInBytes)
        {
            ip_fillClipVerticalLine (pIPState, destPtr,
				     -destPitchInBytes, maskPtr,
				     (imageWidth+31) >> 5);
        }	
	break;

    case cFillDown:
            /* Go through the rows top to bottom */
        for (row = 0; row < imageHeight;
             row++, destPtr += destPitchInBytes, maskPtr += maskPitchInBytes)
        {
            ip_fillClipVerticalLine (pIPState, destPtr,
				     destPitchInBytes, maskPtr,
				     (imageWidth+31) >> 5);
        }
	break;

    case cFillUpDown:
            /* Go through the rows top to bottom */
        for (row = 0; row < imageHeight;
             row++, destPtr += destPitchInBytes, maskPtr += maskPitchInBytes)
        {
            ip_fillClipVerticalLine (pIPState, destPtr,
				     destPitchInBytes, maskPtr,
				     (imageWidth+31) >> 5);
        }

            /* Go through the rows bottom to top */
        for (row--, destPtr -= destPitchInBytes, maskPtr -= maskPitchInBytes;
             row >= 0;
             row--, destPtr -= destPitchInBytes, maskPtr -= maskPitchInBytes)
        {
            ip_fillClipVerticalLine (pIPState, destPtr,
				     -destPitchInBytes, maskPtr,
				     (imageWidth+31) >> 5);
        }	
        break;
    }
    NORMAL_RETURN;
}


/* This is a private routine */
/* PRIVATE */
/*
 *  ip_fillClipLine()
 *		This routine does the bulk of the work of i_fillClipSeedTwoPass.
 *		It is written in assembler in the optimized version of code.
 *		It assumes the seed has already been copied into the dest,
 *		so needs only pointers to the current seed line and the
 *		current mask line.
 *		The routine goes along one line of destination image from
 *		left to right.  When it find either a seed or pixels under
 *		the mask vertically adjacent to the seed, it extends the
 *		seed horizontally until it reaches either the edge of the
 *		mask or the edge of the image.  The routine returns after
 *		processing one destination line.
 *
 *	destPtr:		Pointer to left edge of current dest line, NOT
 *				  including the frame.
 *	destPitchInBytes:	The number which when added to the dest
 *				  image pointer causes one to point to the
 *				  pixel immediately below the initial pixel.
 *				  Alternatively, the total number of image,
 *				  border, frame and pad bytes in a line.
 *				  When we're sweeping from top to bottom, this
 *				  number is positive.  When sweeping from
 *				  bottom to top, it's negative.
 *	maskPtr:		Pointer to left edge of current mask line, NOT
 *				  including the frame.
 *	imWidth:		Width of image in units of 32-bit words.
 *
 * ip_fillClipLine returns the number of words written.
 *
 */
Int32  CDECL 
ip_fillClipLine (
    IPState    *pIPState,
    UInt8      *destPtr,
    Int32	destPitchInBytes,
    UInt8      *maskPtr,
    Int32	imWidth)
{
    PROLOGUE;
    UInt32	 state;	  	/* state machine's state */
    UInt32	 word, lastWord, maskWord;
    Int32	 rightOffset;	/* used to back up */
    UInt32	*destLong, *maskLong;
    Int32	 longDestPitch;	/* used with long ptrs */
    UInt32	 writeCount;

/* DO NOT CHANGE THESE STATE ASSIGNMENTS!  We often use the fact that
 * sLookForSeed = 0 and sFillToLeft = 1 to set the next state by setting
 * state to (word >> 31).
 */
#define	sLookForSeed		0		/* state machine states */
#define	sFillToLeft		1

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    destLong = (UInt32 *) destPtr;
    maskLong = (UInt32 *) maskPtr;
    longDestPitch = destPitchInBytes >> 2;
    writeCount = 0;

    state = sLookForSeed;

    for(imWidth--; imWidth >= 0; imWidth--)
    {
	if (state == sLookForSeed)
	{
	    lastWord = destLong[imWidth];
	    word = lastWord | destLong[(imWidth-longDestPitch)];
	    if (word == 0)			/* there's no seed.
						 * This is the normal case,
						 * so we should dispense with
						 * it as soon as possible */
		continue;

		/* There's some seed, can we expand it into the mask?
		 * Don't bother with expansion if the mask ended. */
	    maskWord = maskLong[imWidth];

		/* if we have reached the edge of a region where the
		 * mask has already been filled, it's likely that the
		 * mask word already equals the seed word */
		/* Anding word with mask keeps this from being 
		 * 8 connected. */
		/* BUG: We should not quit if there's seed in bit 0,
		 * the rightmost bit.  If this is true, there's a
		 * chance that we should extend it back to the
		 * word to the right.
		 */
	    word = word & maskWord;
	    if ((word == 0) ||
		((lastWord == maskWord) && ((lastWord & 0x1) == 0)))
	    {
		state = lastWord >> 31;
		continue;
	    }
	    else
	    {
		    /* There's still a reasonable chance that we've come
		     * up to the border between two or more regions of mask
		     * where the ones that are supposed to be filled have
		     * already been filled.  Do the first dilation step
		     * explicitly just to see if anything changes.  If
		     * not, just quit.
		     * We must also keep going if there's a 1 in bit 0
		     * of this seed word and there's a 1 in bit 31 of
		     * the mask word to the right.
		     */
		word = (word | (word >> 1) | (word << 1)) & maskWord;
		if ((word == lastWord) &&
		    (((word & 0x1) == 0) ||
		     ((Int32)(maskLong[imWidth+1]) >= 0)))
		{
		    state = word >> 31;
		    continue;
		}

		    /* Otherwise, we really need to expand something.  Do
		     * the expansion to the left.  Now that I read this
		     * again, it's sort of cute.  The ~(word + maskWord)
		     * step just sets the mask bits to the left of
		     * any portion of seed bits to 1.  We mask this
		     * with maskWord again to get rid of carries and
		     * trash.  Then we OR the original seed back in to
		     * fill in the places where the seed was turned
		     * to 0's by the add/complement.
		     */
		word = (~(word + maskWord) & maskWord) | word;

		do	/* dilate word to the right to fill maskWord */
		{
		    lastWord = word;
		    word = (word | (word >> 1)) & maskWord;
		}
		while (lastWord != word);

		    /* at this point, the dest word is fully expanded */
		destLong[imWidth] = word;
		writeCount++;

		    /* If MSB of original word was set, we will fill to the
		     * left */
		state = word >> 31;

		    /* was the LSB of that word set?  If so, fill to the
		     * right */
		if (word & 0x1)
		{
		    for (rightOffset = imWidth+1; ; rightOffset++)
		    {
			maskWord = maskLong[rightOffset];
			if (maskWord == 0xFFFFFFFF)
			{
				/* mask covers the entire word to the left */
			    destLong[rightOffset] = maskWord;
			    writeCount++;
			}

			else	/* the mask has a break.  Fill 1's into just
				 * the contiguous section on the MS end.
				 * The loop is sort of cute.  Since we're
				 * doing an arithmetic right shift, the
				 * contiguous section of the mask attached
				 * to the MSB remains constant.  Other
				 * sections of the mask get eaten away. */
			{
			    word = destLong[rightOffset];
				/* turn off any sections that have already
				 * been filled in on a right->left sweep. */
			    maskWord = maskWord ^ word;
			    do
			    {
				lastWord = maskWord;
				maskWord = (((Int32) (maskWord)) >> 1) &
					     maskWord;
			    }
			    while (lastWord != maskWord);

			    destLong[rightOffset] = word | maskWord;
			    writeCount++;
			    break;	/* we're done filling to the right */
			}
		    }  /* of right fill loop */
		}  /* of right fill case */
	    } /* end of case where we did some filling */
	}

	else	/* sFillToLeft */
	{
	    lastWord = destLong[imWidth];
		/* if this is already filled with seed, skip it */
	    if (lastWord == 0xFFFFFFFF)
	    {
		continue;
	    }

		/* if this mask word is 1's, fill this word and keep going. */
	    maskWord = maskLong[imWidth];
	    if (maskWord == 0xFFFFFFFF)
	    {
		destLong[imWidth] = maskWord;
		writeCount++;
		continue;
	    }

		/* Now, we might have just exited (or found a hole in) a
		 * previously filled area.  If that were the case, the mask
		 * would probably equal the dest word.
		 * This is a cheap test to make, so we'll do it. */
	    if (lastWord == maskWord)
	    {
		state = lastWord >> 31;
		continue;

	    }

		/* There's work to do.  We're near the edge of a filled area
		 * but we can't blindly fill the whole thing in.  We want to
		 * fill this with the 4-connected word derived from the word
		 * to the left and the word above.  The loop optimizes the
		 * case of actually doing some filling.  That is, the mask
		 * doesn't happen to completely erase the new stuff.
		 */
	    word = lastWord | destLong[imWidth-longDestPitch];
	    word |= 0x1;			/* contrib from wd to right */
		/* This keeps an 8 connected fill from happening. */
	    word &= maskWord;			/* make sure we don't jump
						 * any boundaries on 1st step */

		/* There's still a reasonable chance that we've come
		 * up to the border between two or more regions of mask
		 * where the ones that are supposed to be filled have
		 * already been filled.  Do the first dilation step
		 * explicitly just to see if anything changes.  If
		 * not, just quit.
		 */
	    word = (word | (word >> 1) | (word << 1)) & maskWord;
	    if (word == lastWord)	/* nothing changed after dilation */
	    {
		state = word >> 31;
		continue;
	    }

		/* do the expansion to the left */
	    word = (~(word + maskWord) & maskWord) | word;

	    do	/* dilate word to the right */
	    {
		lastWord = word;
		word = (word | (word >> 1)) & maskWord;
	    }
	    while (lastWord != word);

		/* word has been fully filled  */
	    destLong[imWidth] = word;
	    writeCount++;

		/* if the MSB of the filled word is NOT set, go back to
		 * looking for new seed again.  If it is set, keep filling */
	    state = word >> 31;

	}   /* end of sFillToLeft state */
    }	/* end of loop that goes through the line */

    return writeCount;
}  /* end of ip_fillClipLine */


/* PRIVATE */
/*
 *  ip_fillClip8ConnLine()
 *		This routine does the bulk of the work of i_fillClipSeed8Conn.
 *		It is written in assembler in the optimized version of code.
 *		It assumes the seed has already been copied into the dest,
 *		so needs only pointers to the current seed line and the
 *		current mask line.
 *		The routine goes along one line of destination image from
 *		right to left.  When it find either a seed or pixels under
 *		the mask vertically adjacent to the seed, it extends the
 *		seed horizontally until it reaches either the edge of the
 *		mask or the edge of the image.  The routine returns after
 *		processing one destination line.
 *
 *	destPtr:		Pointer to left edge of current dest line, NOT
 *				  including the frame.
 *	destPitchInBytes:	The number which when added to the dest
 *				  image pointer causes one to point to the
 *				  pixel immediately below the initial pixel.
 *				  Alternatively, the total number of image,
 *				  border, frame and pad bytes in a line.
 *				  When we're sweeping from top to bottom, this
 *				  number is positive.  When sweeping from
 *				  bottom to top, it's negative.
 *	maskPtr:		Pointer to left edge of current mask line, NOT
 *				  including the frame.
 *	imWidth:		Width of image in units of 32-bit words.
 *
 * ip_fillClip8ConnLine returns the number of words written.
 *
 */
Int32  CDECL 
ip_fillClip8ConnLine (
    IPState    *pIPState,
    UInt8      *destPtr,
    Int32	destPitchInBytes,
    UInt8      *maskPtr,
    Int32	imWidth)
{
    PROLOGUE;
    UInt32      state;	  		/* state machine's state */
    UInt32      word, lastWord, maskWord;
    Int32	rightOffset;		/* used to back up */
    UInt32     *destLong, *maskLong;
    Int32	longDestPitch;		/* used with long ptrs */
    UInt32      writeCount;

/* DO NOT CHANGE THESE STATE ASSIGNMENTS!  We often use the fact that
 * sLookForSeed = 0 and sFillToLeft = 1 to set the next state by setting
 * state to (word >> 31).
 */
#define	sLookForSeed		0		/* state machine states */
#define	sFillToLeft		1

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    destLong = (UInt32 *) destPtr;
    maskLong = (UInt32 *) maskPtr;
    longDestPitch = destPitchInBytes >> 2;
    writeCount = 0;

    state = sLookForSeed;

    for(imWidth--; imWidth >= 0; imWidth--)
    {
	if (state == sLookForSeed)
	{
	    lastWord = destLong[imWidth];
	    maskWord = maskLong[imWidth];

		/* if we have reached the edge of a region where the
		 * mask has already been filled, it's likely that the
		 * mask word already equals the seed word.
		 * BUG: even if lastWord == maskWord, there's still a
		 * chance that we should be expanding the seed back
		 * to the right where we`ve just been.*/
	    if ((lastWord == maskWord) && ((lastWord & 0x1) == 0))
	    {
		state = lastWord >> 31;
		continue;
	    }

	    word = lastWord | destLong[imWidth-longDestPitch];
	    word = (word | (word >> 1) | (word << 1) |
	        ((destLong[imWidth-longDestPitch-1] & 0x1) << 31) |
	        ((destLong[imWidth-longDestPitch+1] & 0x80000000) >> 31)) &
		maskWord;

	    if (word == 0)			/* there's no seed.
						 * This is the normal case,
						 * so we should dispense with
						 * it as soon as possible */
		continue;


	    else
	    {
		    /* There's still a reasonable chance that we've come
		     * up to the border between two or more regions of mask
		     * where the ones that are supposed to be filled have
		     * already been filled.  Quit if the seed and mask are
		     * the same AND the LS bit of the seed is 0 or the
		     * MS bit of the mask word to the right is 0.
		     */

		if ((word == lastWord) &&
		    (((word & 0x1) == 0) ||
		     ((Int32)(maskLong[imWidth+1]) >= 0)))
		{
		    state = word >> 31;
		    continue;
		}

		    /* Otherwise, we really need to expand something.  Do
		     * the expansion to the left */
		word = (~(word + maskWord) & maskWord) | word;

		do	/* dilate word to the right to fill maskWord */
		{
		    lastWord = word;
		    word = (word | (word >> 1)) & maskWord;
		}
		while (lastWord != word);

		    /* at this point, the dest word is fully expanded */
		destLong[imWidth] = word;
		writeCount++;

		    /* If MSB of original word was set, we will fill to the
		     * left */
		state = word >> 31;

		    /* was the LSB of that word set?  If so, fill to the
		     * right */
		if (word & 0x1)
		{
		    for (rightOffset = imWidth+1; ; rightOffset++)
		    {
			maskWord = maskLong[rightOffset];
			if (maskWord == 0xFFFFFFFF)
			{
				/* mask covers the entire word to the left */
			    destLong[rightOffset] = maskWord;
			    writeCount++;
			}

			else	/* the mask has a break.  Fill 1's into just
				 * the contiguous section on the MS end.
				 * The loop is sort of cute.  Since we're
				 * doing an arithmetic right shift, the
				 * contiguous section of the mask attached
				 * to the MSB remains constant.  Other
				 * sections of the mask get eaten away. */
			{
			    word = destLong[rightOffset];
				/* turn off any sections that have already
				 * been filled in on a right->left sweep. */
			    maskWord = maskWord ^ word;
			    do
			    {
				lastWord = maskWord;
				maskWord = (((Int32) (maskWord)) >> 1) &
					     maskWord;
			    }
			    while (lastWord != maskWord);

			    destLong[rightOffset] = word | maskWord;
			    writeCount++;
			    break;	/* we're done filling to the right */
			}
		    }  /* of right fill loop */
		}  /* of right fill case */
	    } /* end of case where we did some filling */
	}

	else	/* sFillToLeft */
	{
	    lastWord = destLong[imWidth];
		/* if this is already filled with seed, skip it */
	    if (lastWord == 0xFFFFFFFF)
	    {
		continue;
	    }

		/* if this mask word is 1's, fill this word and keep going. */
	    maskWord = maskLong[imWidth];
	    if (maskWord == 0xFFFFFFFF)
	    {
		destLong[imWidth] = maskWord;
		writeCount++;
		continue;
	    }

		/* Now, we might have just exited (or found a hole in) a
		 * previously filled area.  If that were the case, the mask
		 * would probably equal the dest word.
		 * This is a cheap test to make, so we'll do it. */
	    if (lastWord == maskWord)
	    {
		state = lastWord >> 31;
		continue;

	    }

		/* There's work to do.  We're near the edge of a filled area
		 * but we can't blindly fill the whole thing in.  We want to
		 * fill this with the 8-connected word derived from the word
		 * to the left and the word above.  The loop optimizes the
		 * case of actually doing some filling.  That is, the mask
		 * doesn't happen to completely erase the new stuff.
		 */
	    word = lastWord | destLong[imWidth-longDestPitch];

	        /* contrib from wd to right by putting a 1 in the LSB
		 * if that bit is under the mask (since we don't want
		 * to jump any boundaries on 1st step */
		/* There's still a reasonable chance that we've come
		 * up to the border between two or more regions of mask
		 * where the ones that are supposed to be filled have
		 * already been filled.  Do the first dilation step
		 * explicitly just to see if anything changes.  If
		 * not, just quit.
		 */
	    word = ((maskWord & 0x1) | 
		word | (word >> 1) | (word << 1) |
	        ((destLong[imWidth-longDestPitch-1] & 0x1) << 31) |
	        ((destLong[imWidth-longDestPitch+1] & 0x80000000) >> 31)) &
		maskWord;

	    if (word == lastWord)	/* nothing changed after dilation */
	    {
		state = word >> 31;
		continue;
	    }

		/* do the expansion to the left */
	    word = (~(word + maskWord) & maskWord) | word;

	    do	/* dilate word to the right */
	    {
		lastWord = word;
		word = (word | (word >> 1)) & maskWord;
	    }
	    while (lastWord != word);

		/* word has been fully filled  */
	    destLong[imWidth] = word;
	    writeCount++;

		/* if the MSB of the filled word is NOT set, go back to
		 * looking for new seed again.  If it is set, keep filling */
	    state = word >> 31;

	}   /* end of sFillToLeft state */
    }	/* end of loop that goes through the line */

    return writeCount;
}  /* end of ip_fillClip8ConnLine */


/* PRIVATE */
/*
 *  ip_fillClipHorizontalLine()
 *		This routine does the bulk of the work of i_fillClipSeedHorizontal.
 *		It assumes the seed has already been copied into the dest,
 *		so needs only pointers to the current seed line and the
 *		current mask line.
 *		The routine goes along one line of destination image from
 *		right to left.  When it find either a seed or pixels under
 *		the mask vertically adjacent to the seed, it extends the
 *		seed horizontally until it reaches either the edge of the
 *		mask or the edge of the image.  The routine returns after
 *		processing one destination line.
 *
 *	destPtr:		Pointer to left edge of current dest line, NOT
 *				  including the frame.
 *	maskPtr:		Pointer to left edge of current mask line, NOT
 *				  including the frame.
 *	imWidth:		Width of image in units of 32-bit words.
 *
 */
void  CDECL 
ip_fillClipHorizontalLine (
    IPState    *pIPState,
    UInt8      *destPtr,
    UInt8      *maskPtr,
    Int32	imWidth)
{
    PROLOGUE;
    UInt32      state;	  		/* state machine's state */
    UInt32      word, lastWord, maskWord;
    Int32	rightOffset;		/* used to back up */
    UInt32     *destLong, *maskLong;

/* DO NOT CHANGE THESE STATE ASSIGNMENTS!  We often use the fact that
 * sLookForSeed = 0 and sFillToLeft = 1 to set the next state by setting
 * state to (word >> 31).
 */
#define	sLookForSeed		0		/* state machine states */
#define	sFillToLeft		1

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    destLong = (UInt32 *) destPtr;
    maskLong = (UInt32 *) maskPtr;

    state = sLookForSeed;

    for(imWidth--; imWidth >= 0; imWidth--)
    {
	if (state == sLookForSeed)
	{
	    lastWord = destLong[imWidth];
	    if (lastWord == 0)			/* there's no seed.
						 * This is the normal case,
						 * so we should dispense with
						 * it as soon as possible */
		continue;

		/* There's some seed, can we expand it into the mask?
		 * Don't bother with expansion if the mask ended. */
	    maskWord = maskLong[imWidth];

		/* if we have reached the edge of a region where the
		 * mask has already been filled, it's likely that the
		 * mask word already equals the seed word.
		 * HOWEVER, if word == maskWord, we might still have
		 * to fill back to the right.  This might be true if
		 * the LS bit of word is set.
		 */
	    word = lastWord & maskWord;
	    if ((word == 0) ||
		((lastWord == maskWord) && ((lastWord & 0x1) == 0)))
	    {
		state = lastWord >> 31;
		continue;
	    }
	    else
	    {
		    /* There's still a reasonable chance that we've come
		     * up to the border between two or more regions of mask
		     * where the ones that are supposed to be filled have
		     * already been filled.  Do the first dilation step
		     * explicitly just to see if anything changes.  If
		     * not, just quit.
		     * We must also keep going if there's a 1 in bit 0
		     * of this seed word and there's a 1 in bit 31 of
		     * the mask word to the right.
		     */
		word = (word | (word >> 1) | (word << 1)) & maskWord;
		if ((word == lastWord) &&
		    (((word & 0x1) == 0) ||
		     ((Int32)(maskLong[imWidth+1]) >= 0)))
		{
		    state = word >> 31;
		    continue;
		}

		    /* Otherwise, we really need to expand something.  Do
		     * the expansion to the left */
		word = (~(word + maskWord) & maskWord) | word;

		do	/* dilate word to the right to fill maskWord */
		{
		    lastWord = word;
		    word = (word | (word >> 1)) & maskWord;
		}
		while (lastWord != word);

		    /* at this point, the dest word is fully expanded */
		destLong[imWidth] = word;

		    /* If MSB of original word was set, we will fill to the
		     * left */
		state = word >> 31;

		    /* was the LSB of that word set?  If so, fill to the
		     * right */
		if (word & 0x1)
		{
		    for (rightOffset = imWidth+1; ; rightOffset++)
		    {
			maskWord = maskLong[rightOffset];
			if (maskWord == 0xFFFFFFFF)
			{
				/* mask covers the entire word to the left */
			    destLong[rightOffset] = maskWord;
			}

			else	/* the mask has a break.  Fill 1's into just
				 * the contiguous section on the MS end.
				 * The loop is sort of cute.  Since we're
				 * doing an arithmetic right shift, the
				 * contiguous section of the mask attached
				 * to the MSB remains constant.  Other
				 * sections of the mask get eaten away. */
			{
			    word = destLong[rightOffset];
				/* turn off any sections that have already
				 * been filled in on a right->left sweep. */
			    maskWord = maskWord ^ word;
			    do
			    {
				lastWord = maskWord;
				maskWord = (((Int32) (maskWord)) >> 1) &
					     maskWord;
			    }
			    while (lastWord != maskWord);

			    destLong[rightOffset] = word | maskWord;
			    break;	/* we're done filling to the right */
			}
		    }  /* of right fill loop */
		}  /* of right fill case */
	    } /* end of case where we did some filling */
	}

	else	/* sFillToLeft */
	{
	    lastWord = destLong[imWidth];
		/* if this is already filled with seed, skip it */
	    if (lastWord == 0xFFFFFFFF)
	    {
		continue;
	    }

		/* if this mask word is 1's, fill this word and keep going. */
	    maskWord = maskLong[imWidth];
	    if (maskWord == 0xFFFFFFFF)
	    {
		destLong[imWidth] = maskWord;
		continue;
	    }

		/* Now, we might have just exited (or found a hole in) a
		 * previously filled area.  If that were the case, the mask
		 * would probably equal the dest word.
		 * This is a cheap test to make, so we'll do it. */
	    if (lastWord == maskWord)
	    {
		state = lastWord >> 31;
		continue;

	    }

		/* There's work to do.  We're near the edge of a filled area
		 * but we can't blindly fill the whole thing in.  We want to
		 * fill this with the 4-connected word derived from the word
		 * to the left and the word above.  The loop optimizes the
		 * case of actually doing some filling.  That is, the mask
		 * doesn't happen to completely erase the new stuff.
		 */
	    word = lastWord;
	    word |= 0x1;			/* contrib from wd to right */
	    word &= maskWord;			/* make sure we don't jump
						 * any boundaries on 1st step */

		/* There's still a reasonable chance that we've come
		 * up to the border between two or more regions of mask
		 * where the ones that are supposed to be filled have
		 * already been filled.  Do the first dilation step
		 * explicitly just to see if anything changes.  If
		 * not, just quit.
		 */
	    word = (word | (word >> 1) | (word << 1)) & maskWord;
	    if (word == lastWord)	/* nothing changed after dilation */
	    {
		state = word >> 31;
		continue;
	    }

		/* do the expansion to the left */
	    word = (~(word + maskWord) & maskWord) | word;

	    do	/* dilate word to the right */
	    {
		lastWord = word;
		word = (word | (word >> 1)) & maskWord;
	    }
	    while (lastWord != word);

		/* word has been fully filled  */
	    destLong[imWidth] = word;

		/* if the MSB of the filled word is NOT set, go back to
		 * looking for new seed again.  If it is set, keep filling */
	    state = word >> 31;

	}   /* end of sFillToLeft state */
    }	/* end of loop that goes through the line */

}  /* end of ip_fillClipHorizontalLine */


/* PRIVATE */
/*
 *  ip_fillClipLeftOnlyLine()
 *		This routine does the bulk of the work of
 *		i_fillClipSeedHorizontal if we are fill only to the left.
 *		It assumes the seed has already been copied into the dest,
 *		so needs only pointers to the current seed line and the
 *		current mask line.
 *		The routine goes along one line of destination image from
 *		right to left.  When it find either a seed or pixels under
 *		the mask to the right of the seed, it extends the
 *		seed to the left until it reaches either the edge of the
 *		mask or the edge of the image.  The routine returns after
 *		processing one destination line.
 *
 *	destPtr:		Pointer to left edge of current dest line, NOT
 *				  including the frame.
 *	maskPtr:		Pointer to left edge of current mask line, NOT
 *				  including the frame.
 *	imWidth:		Width of image in units of 32-bit words.
 *
 */
void  CDECL 
ip_fillClipLeftOnlyLine (
    IPState    *pIPState,
    UInt8      *destPtr,
    UInt8      *maskPtr,
    Int32	imWidth)
{
    PROLOGUE;
    UInt32      state;	  		/* state machine's state */
    UInt32      word, lastWord, maskWord;
    UInt32     *destLong, *maskLong;

/* DO NOT CHANGE THESE STATE ASSIGNMENTS!  We often use the fact that
 * sLookForSeed = 0 and sFillToLeft = 1 to set the next state by setting
 * state to (word >> 31).
 */
#define	sLookForSeed		0		/* state machine states */
#define	sFillToLeft		1

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    destLong = (UInt32 *) destPtr;
    maskLong = (UInt32 *) maskPtr;

    state = sLookForSeed;

    for(imWidth--; imWidth >= 0; imWidth--)
    {
	if (state == sLookForSeed)
	{
	    lastWord = destLong[imWidth];
	    if (lastWord == 0)			/* there's no seed.
						 * This is the normal case,
						 * so we should dispense with
						 * it as soon as possible */
		continue;

		/* There's some seed, can we expand it into the mask?
		 * Don't bother with expansion if the mask ended. */
	    maskWord = maskLong[imWidth];

		/* if we have reached the edge of a region where the
		 * mask has already been filled, it's likely that the
		 * mask word already equals the seed word */
	    word = lastWord & maskWord;
	    if ((word == 0) || (lastWord == maskWord))
	    {
		state = lastWord >> 31;
		continue;
	    }
	    else
	    {
		    /* There's still a reasonable chance that we've come
		     * up to the border between two or more regions of mask
		     * where the ones that are supposed to be filled have
		     * already been filled.  Do the first dilation step
		     * explicitly just to see if anything changes.  If
		     * not, just quit.
		     */
		word = (word | (word << 1)) & maskWord;
		if (word == lastWord)
		{
		    state = word >> 31;
		    continue;
		}

		    /* Otherwise, we really need to expand something.  Do
		     * the expansion to the left */
		word = (~(word + maskWord) & maskWord) | word;

		    /* at this point, the dest word is fully expanded */
		destLong[imWidth] = word;

		    /* If MSB of original word was set, we will fill to the
		     * left */
		state = word >> 31;

	    } /* end of case where we did some filling */
	}

	else	/* sFillToLeft */
	{
	    lastWord = destLong[imWidth];
		/* if this is already filled with seed, skip it */
	    if (lastWord == 0xFFFFFFFF)
	    {
		continue;
	    }

		/* if this mask word is 1's, fill this word and keep going. */
	    maskWord = maskLong[imWidth];
	    if (maskWord == 0xFFFFFFFF)
	    {
		destLong[imWidth] = maskWord;
		continue;
	    }

		/* Now, we might have just exited (or found a hole in) a
		 * previously filled area.  If that were the case, the mask
		 * would probably equal the dest word.
		 * This is a cheap test to make, so we'll do it. */
	    if (lastWord == maskWord)
	    {
		state = lastWord >> 31;
		continue;

	    }

		/* There's work to do.  We're near the edge of a filled area
		 * but we can't blindly fill the whole thing in.  We want to
		 * fill this with the word derived from the word
		 * to the left.  The loop optimizes the
		 * case of actually doing some filling.  That is, the mask
		 * doesn't happen to completely erase the new stuff.
		 */
	    word = lastWord;
	    word |= 0x1;			/* contrib from wd to right */
	    word &= maskWord;			/* make sure we don't jump
						 * any boundaries on 1st step */

		/* There's still a reasonable chance that we've come
		 * up to the border between two or more regions of mask
		 * where the ones that are supposed to be filled have
		 * already been filled.  Do the first dilation step
		 * explicitly just to see if anything changes.  If
		 * not, just quit.
		 */
	    word = (word | (word << 1)) & maskWord;
	    if (word == lastWord)	/* nothing changed after dilation */
	    {
		state = word >> 31;
		continue;
	    }

		/* do the expansion to the left */
	    word = (~(word + maskWord) & maskWord) | word;

		/* word has been fully filled  */
	    destLong[imWidth] = word;

		/* if the MSB of the filled word is NOT set, go back to
		 * looking for new seed again.  If it is set, keep filling */
	    state = word >> 31;

	}   /* end of sFillToLeft state */
    }	/* end of loop that goes through the line */

}  /* end of ip_fillClipLeftOnlyLine */


/* PRIVATE */
/*
 *  ip_fillClipRightOnlyLine()
 *		This routine does the bulk of the work of
 *		i_fillClipSeedHorizontal if filling to the right only
 *		It assumes the seed has already been copied into the dest,
 *		so needs only pointers to the current seed line and the
 *		current mask line.
 *		The routine goes along one line of destination image from
 *		right to left.  When it find either a seed or pixels under
 *		the mask left of the seed, it extends the
 *		seed to the right until it reaches either the edge of the
 *		mask or the edge of the image.  The routine returns after
 *		processing one destination line.
 *
 *	destPtr:		Pointer to left edge of current dest line, NOT
 *				  including the frame.
 *	maskPtr:		Pointer to left edge of current mask line, NOT
 *				  including the frame.
 *	imWidth:		Width of image in units of 32-bit words.
 *
 */
void  CDECL 
ip_fillClipRightOnlyLine (
    IPState    *pIPState,
    UInt8      *destPtr,
    UInt8      *maskPtr,
    Int32	imWidth)
{
    PROLOGUE;
    UInt32      word, lastWord, maskWord;
    Int32	rightOffset;	/* used to back up */
    UInt32     *destLong, *maskLong;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    destLong = (UInt32 *) destPtr;
    maskLong = (UInt32 *) maskPtr;

    for(imWidth--; imWidth >= 0; imWidth--)
    {
	lastWord = destLong[imWidth];
	if (lastWord == 0)			/* there's no seed.
					     * This is the normal case,
					     * so we should dispense with
					     * it as soon as possible */
	    continue;

	    /* There's some seed, can we expand it into the mask?
	     * Don't bother with expansion if the mask ended. */
	maskWord = maskLong[imWidth];

	    /* if we have reached the edge of a region where the
	     * mask has already been filled, it's likely that the
	     * mask word already equals the seed word.
	     * HOWEVER, if word == maskWord, we might still have
	     * to fill back to the right.  This might be true if
	     * the LS bit of word is set.
	     */
	word = lastWord & maskWord;
	if ((word == 0) ||
	    ((lastWord == maskWord) && ((lastWord & 0x1) == 0)))
	{
	    continue;
	}
	else
	{
		/* There's still a reasonable chance that we've come
		 * up to the border between two or more regions of mask
		 * where the ones that are supposed to be filled have
		 * already been filled.  Do the first dilation step
		 * explicitly just to see if anything changes.  If
		 * not, just quit.
		 * We must also keep going if there's a 1 in bit 0
		 * of this seed word and there's a 1 in bit 31 of
		 * the mask word to the right.
		 */
	    word = (word | (word >> 1)) & maskWord;
	    if ((word == lastWord) &&
		(((word & 0x1) == 0) ||
		 ((Int32)(maskLong[imWidth+1]) >= 0)))
	    {
		continue;
	    }

	    do	/* dilate word to the right to fill maskWord */
	    {
		lastWord = word;
		word = (word | (word >> 1)) & maskWord;
	    }
	    while (lastWord != word);

		/* at this point, the dest word is fully expanded */
	    destLong[imWidth] = word;

		/* was the LSB of that word set?  If so, fill to the
		 * right */
	    if (word & 0x1)
	    {
		for (rightOffset = imWidth+1; ; rightOffset++)
		{
		    maskWord = maskLong[rightOffset];
		    if (maskWord == 0xFFFFFFFF)
		    {
			    /* mask covers the entire word to the left */
			destLong[rightOffset] = maskWord;
		    }

		    else	/* the mask has a break.  Fill 1's into just
			     * the contiguous section on the MS end.
			     * The loop is sort of cute.  Since we're
			     * doing an arithmetic right shift, the
			     * contiguous section of the mask attached
			     * to the MSB remains constant.  Other
			     * sections of the mask get eaten away. */
		    {
			word = destLong[rightOffset];
			    /* turn off any sections that have already
			     * been filled in on a right->left sweep. */
			maskWord = maskWord ^ word;
			do
			{
			    lastWord = maskWord;
			    maskWord = (((Int32) (maskWord)) >> 1) &
					 maskWord;
			}
			while (lastWord != maskWord);

			destLong[rightOffset] = word | maskWord;
			break;	/* we're done filling to the right */
		    }
		}  /* of right fill loop */
	    }  /* of right fill case */
	} /* end of case where we did some filling */

    }	/* end of loop that goes through the line */

}  /* end of ip_fillClipRightOnlyLine */


/* PRIVATE */
/*
 *  ip_fillClipVerticalLine()
 *		This routine does the bulk of the work of i_fillClipSeedVertical.
 *		It assumes the seed has already been copied into the dest,
 *		so needs only pointers to the current seed line and the
 *		current mask line.
 *		The routine goes along one line of destination image from
 *		right to left.  When it find either a seed or pixels under
 *		the mask vertically adjacent to the seed, it extends the
 *		seed vertically by one pixel.  The routine returns after
 *		processing one destination line.
 *
 *	destPtr:		Pointer to left edge of current dest line, NOT
 *				  including the frame.
 *	destPitchInBytes:	The number which when added to the dest
 *				  image pointer causes one to point to the
 *				  pixel immediately below the initial pixel.
 *				  Alternatively, the total number of image,
 *				  border, frame and pad bytes in a line.
 *				  When we're sweeping from top to bottom, this
 *				  number is positive.  When sweeping from
 *				  bottom to top, it's negative.
 *	maskPtr:		Pointer to left edge of current mask line, NOT
 *				  including the frame.
 *	imWidth:		Width of image in units of 32-bit words.
 *
 *
 */
void  CDECL 
ip_fillClipVerticalLine (
    IPState    *pIPState,
    UInt8      *destPtr,
    Int32	destPitchInBytes,
    UInt8      *maskPtr,
    Int32	imWidth)
{
    PROLOGUE;
    UInt32      word, lastWord, maskWord;
    UInt32     *destLong, *maskLong;
    Int32	longDestPitch;	/* used with long ptrs */

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    destLong = (UInt32 *) destPtr;
    maskLong = (UInt32 *) maskPtr;
    longDestPitch = destPitchInBytes >> 2;


    for(imWidth--; imWidth >= 0; imWidth--)
    {
	lastWord = destLong[imWidth];
	word = lastWord | destLong[(imWidth-longDestPitch)];
	if (word == 0)			/* there's no seed.
					 * This is the normal case,
					 * so we should dispense with
					 * it as soon as possible */
		continue;

	    /* There's some seed, can we expand it into the mask? 
	     * or has it already been done? */
	maskWord = maskLong[imWidth];
	word = word & maskWord;
	if ((word != 0) && (lastWord != maskWord))
	{
	    destLong[imWidth] = word;
	}
    }	/* end of loop that goes through the line */

}  /* end of ip_fillClipVerticalLine */

/*

$Log: i_fill.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.6  1998/01/13 23:47:47  ddavies
i_fill.c -> /product/ipcore/ipshared/src/RCS/i_fill.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1997/04/24 23:28:28  ddavies
i_fill.c -> /product/ipcore/ipshared/src/RCS/i_fill.c,v

Change type of clipFirst in the i_fillClipXXX routines from Int32 to Bool.

Revision 1.4  1996/02/10 01:26:15  dferrell
i_fill.c -> /product/ipcore/ipshared/src/RCS/i_fill.c,v

Added extra parens around & operations so that gcc 2.7.2 won't complain.

Revision 1.3  1995/12/13 02:43:44  ejaquith
i_fill.c -> /product/ipcore/ipshared/src/RCS/i_fill.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1994/02/15  00:28:15  ddavies
 * ddavies on Mon Feb 14 16:20:00 PST 1994
 *
 * i_fill.c -> /product/ipcore/ipshared/src/RCS/i_fill.c,v
 *
 * Fix bug that was common to ip_fillClipLine, ip_fillClip8ConnLine,
 * ip_fillClipRightOnlyLine and ip_fillClipHorizontalLine.  In all cases,
 * a seed that ended at the least significant bit of a 32-bit word would not
 * be extended to the right even though there were mask bits to the right.
 * This effects 4-connected filling, 8-connected filling, horizontal filling
 * and horizontal filling to the right.
 *
 * Revision 1.1  1994/01/22  03:49:11  ddavies
 * Initial revision
 *
 * Revision 1.9  1993/12/02  19:26:40  ddavies
 * sfill.c -> /product/ipcore/ipshared/src/RCS/sfill.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.8  1993/11/05  02:21:43  ddavies
 * sfill.c -> /product/ipcore/ipshared/src/RCS/sfill.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


