/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_filter:	Filters for arbitrary scale-to-gray
 *
 * Exports:
 *	Int32	    i_boxFilterInfo ()	    
 *	Int32	    i_boxFilterGen ()	    
 *
 * There used to be tent and sinc2 filters, but they were taken out
 * because their use was no longer supported in exchange for execution 
 * speed.
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscgr.pub"
#include "shrscgr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_filter.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/***********************************************************************
 *
 * i_boxFilterInfo
 *
 *	"info" function for box filter. This function is passed
 *	to i_makeScaleToGrayArbFilters (and i_scaleToGrayArbFilterBytes)
 *	in order to determine the size of a filter given a reduction 
 *	factor. There is one of these functions for each filter that we
 *	implement. The width and height computed should be the maximum
 *	width and height for any phase [0-1).
 *
 * Arguments:
 *	reductionX
 *      reductionY 	-	Reduction factor in horizontal and vertical
 *	scaling		-	Extra scaling, i.e. 1.0 for a 
 *				1:1 filter, 2.0 for a twice-width filter.
 *	width, height	-	Place to stuff width and height 
 *				that filter should be for the
 *				reduction factor.
 *
 * Returns:
 *	nothing
 ***********************************************************************/

void CDECL 
i_boxFilterInfo (
    IPState    *pIPState,
    Float64	reductionX,
    Float64	reductionY,
    Float64	scaling,
    Int32      *width,
    Int32      *height)
{
    PROLOGUE;
    Int32 size;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    size = (Int32)(ceil (1.0 / reductionX) * scaling);
    *width = size;
    size = (Int32)(ceil (1.0 / reductionY) * scaling);
    *height = size;
}

/***********************************************************************
 * i_boxFilterGen 
 *
 *	Filter generation function for a box filter. We're passed
 *	an "ideal" and "actual" width. The "ideal" width is what
 *	i_boxFilterInfo computed. The "actual" width is the actual
 *	width that the caller wants, which is probably truncated to
 *	some maximum. It is left up to individual filters what to do
 *	with a truncated width; in this filter, we'll just make
 *	a narrower box, i.e. trade off some aliasing for some sharpness.
 *
 * Arguments:
 *	reduction   -	Reduction factor, 0-1, where 0 is infinitely small
 *			and 1 is no reduction
 *	width	    -	Width of filter
 *	height	    -	Height of filter
 *	idealWidth  -	"Ideal" width, ignored by this func.
 *	idealHeight -	"Ideal" height, ignored by this func.
 *	rowPhase    -	Source pixel y-phase, [0-1)
 *	colPhase    -	Source pixel x-phase, [0-1)
 *	filterParm  -	"extra" parameter, which we don't use here 
 *	coefs	    -	Place to stuff coefs, organized as height rows of
 *			width coefficients, scaled so that the sum is 1.0.
 *
 * Returns:
 *	nothing
 ***********************************************************************/

void CDECL 
i_boxFilterGen (
    IPState    *pIPState,
    Float64	reduction,		    /* Reduction factor */
    Int32	width,			    /* Width of filter */
    Int32	height,			    /* Height of filter */
    Int32	idealWidth,		    /* ideal width, pre-truncation */
    Int32	idealHeight,		    /* ideal height, pre-truncation*/
    Float64	rowPhase,		    /* y-phase */
    Float64	colPhase,		    /* x-phase */
    Float64	filterParm,		    /* filter-defined parm */
    Float64    *coefs)	    		    /* Place to stuff coefs */
{
    PROLOGUE;
    Float64	val;
    Int32	i, j;

	/* Make compilers silent */
    pIPState = pIPState;
    __error__ = __error__;
    rowPhase = colPhase = 0; idealWidth = idealHeight = 0;
    filterParm = reduction = 0;

	/* It's just a uniform filter */
    val = 1.0 / (width * height);
    for (i = 0; i < height; i++)
    {
	for (j = 0; j < width; j++)
	{
	    coefs[i * width + j] = val;
	}
    }

    return;
}

/*
$Log: i_filter.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.6  1998/01/13 23:47:48  ddavies
i_filter.c -> /product/ipcore/ipshared/src/RCS/i_filter.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1997/11/12 17:51:06  xzhu
xzhu on Wed Nov 12 09:49:32 PST 1997

i_filter.c -> /product/ipcore/ipshared/src/RCS/i_filter.c,v

Use reductionX to estimate filter width and reductionY for filter height.

Revision 1.4  1995/12/13 02:43:48  ejaquith
i_filter.c -> /product/ipcore/ipshared/src/RCS/i_filter.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/11/21  01:03:20  macleod
 * macleod on Mon Nov 20 17:03:11 PST 1995
 *
 * i_filter.c -> /product/ipcore/ipshared/src/RCS/i_filter.c,v
 *
 * Got rid of tent and sinc2 filters. Only box filter remains.
 *
 * Revision 1.2  1995/10/16  17:51:07  dferrell
 * i_filter.c -> /product/ipcore/ipshared/src/RCS/i_filter.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.1  1995/01/04  01:14:19  macleod
 * Initial revision
 *
*/
