/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_floyd.c:	Assembler assisted version of floyd.c
 *
 *          Upper level dithering code using Floyd-Steinberg algorithm
 *		From Nova.c by Rob Tow and tidied by Dan Bloomberg.
 *		Performance rework by Dan Davies.
 *
 *          Single resolution Floyd-Steinberg dither:
 *                  Int32     i_ditherFloyd()
 *                  Int32     i_ditherFloyd2()
 *                  Int32     i_ditherFloyd4to1()
 *
 *	    Banded Floyd-Steinberg dither:
 *		    Int32     i_getSizeOfDitherState()
 *		    Int32     i_ditherFloydInit()
 *		    Int32     i_ditherFloydProcess()
 *		    Int32     i_ditherFloydFinish()
 *
 *		    Int32     i_getSizeOfDither4to1State()
 *		    Int32     i_ditherFloyd4to1Init()
 *		    Int32     i_ditherFloyd4to1Process()
 *		    Int32     i_ditherFloyd4to1Finish()
 *
 *	    Single line Floyd-Steinberg dither (to be used with banded
 *	     routines above):
 *		    Int32     i_ditherFloydSingleProcess()
 *		    Int32     i_ditherFloyd4to1SingleProcess()
 *
 *	    Lookup tables:
 *		    Int32     ip_make2ErrorDiffusionTables();
 *
 *                [I'm not doing 8->2 banded, since only DAE
 *                uses it, and they don't want banded]
 */


#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrrast.pub"
#include "shros.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_floyd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");



/* This module is the upper level code to implement Floyd-Steinberg
 * error diffusion.  That's a technique for turning gray images into
 * binary images while attempting to maintain the range of intensities
 * present in the gray images.
 *
 * Here are the files associated with this function
 *	i_floyd.c	This file, contains the upper level code and 
 *			code to make tables used for 8->2 bit dithering..
 *	mifloyd.c	C implementation of 8->1 bit table making routines,
 *			machine-dependent setup code and dithering primitives.
 *	asfloyd.c	Code to make tables for 8->1 bit dithering for
 *			all assembly routines and machine-dependent setup
 *			code for all assembly versions of dithering.
 *	ssfloyd.s	SPARC assembly version of dithering primitives
 *	pcfloyd.asm	PC assembly version of dithering primitives.
 *
 * The following comments explain the overall philosophy and implementation
 * of the dithering algorithms.
 *
 * ---------------------------------------------------------------------
 *
 * The purpose of this module is to convert 8-bit gray scale images to
 * binary, dithered images.  This is not a simple binary threshold.  The
 * number of black dots in a region of the binary image should be (roughly)
 * proportional to the darkness of the gray image.  The binary image should
 * be a faithful reproduction of the original.
 *
 * In addition, we will also apply dynamic range correction to the
 * gray pixels.  Each pixel's intensity will be mapped to some other
 * intensity using a table provided by the caller.  Doing the mapping
 * here saves the time required to write another complete gray image.
 *
 * There are many, many algorithms for converting gray images to binary.
 * The module ~bloomber/halftoning/Nova.c contains examples of 14
 * algorithms.  People invent more every day.  "Digital Halftoning"
 * by Robert Ulichney (MIT Press, 1987 call number T385.U45 1987 in the
 * PARC TIC) illustrates dozens more.
 *
 * We have chosen the Floyd-Steinberg method of dithering with error
 * diffusion.  It is simple to implement and gives very nice results.
 *
 * Like many diffusion algorithms, we start processing in the upper left
 * corner of an image.  There are two basic steps - threshold compare and
 * error diffusion.  The threshold compare is simple, we just use the most
 * significant bit of the pixel as the binary value.  Printing a "1" is
 * really only completely correct if the gray value is 255.  Printing a
 * 0 is only completely correct if the gray value is 0.  Using this
 * interpretation, we calculate the "error" as the difference between
 * the gray pixel and the "displayed" value.  For example, if the gray
 * pixel value p is less than 128, the pixel is actually p too large for
 * the value to be displayed.  Similarly, if p >= 128, the error is
 * p - 255 (a number less than or equal to 0).
 *
 * Now the magic comes in.  In Floyd-Steinberg, we apportion this error value
 * among 4 other gray pixels as shown below.  The "*" represents the
 * newly thresholded pixel.  The fractions represent the proportions of the
 * error value that is *added* to the pixel to the right and the three pixels
 * below.  The values of these fractions have been carefully chosen to make
 * the resulting images look nice.  I don't know why they work better than
 * some other set.  Nick Briggs says that the fractions are roughly
 * proportional to the distance from the center of the thresholded pixel to
 * to the other pixels.  Leigh Klotz says that FX says many combinations of
 * fractions work so long as there is no accumulated error.  Note that the
 * fractions do add to 1.
 *
 * |      |  *   | 7/16 |
 * | 3/16 | 5/16 | 1/16 |
 *
 * After the error is added to the original pixel values and previously
 * modified pixel values, we are supposed to truncate each addition back
 * to the range [0..255].  If a pixel value becomes greater than 255, we
 * set it to 255.  If it goes negative, we set it to 0.
 *
 * In the classic implementation, each line is processed from left to right
 * creating a new line of modified pixels below it.  This line of pixels
 * is processed in turn, leaving a new line and so on.  At least 4 write
 * operations are done for every pixel; more if truncation is required.
 *
 * How can we improve the speed of this beauty?  The first thing we notice
 * is the number of writes.  How can this be reduced?
 *
 * One approach is to calculate several lines of dithered image at once.
 * Assuming the error values can be held in processor registers until they
 * are binarized, we can avoid writing the error values to memory.  If we
 * calculate n lines of dithered image at once, we only have to write out
 * 1 line of partially-processed error values, so save writing n-1 lines of
 * pixels.
 *
 * A significant source of writes in some implementations is the binary
 * output.  For example, the code in Nova.c does a read, an AND, an OR, a
 * shift and a write for every bit of output image.  We'll just accumulate
 * 32 bits of output for each image line before writing.  We could mess
 * around with double writes, but it uses extra registers and has a very
 * small effect on overall performance (reducing the number of destination
 * writes by a factor of 32 makes the number of cycles spent on writes
 * insignificant).
 *
 * Another source of writes is the individual error diffusion calculations.
 * It would certainly be nice to avoid doing 4 separate READ-ADD-TRUNCATE-WRITE
 * sequences for each thresholded value.  Say we pack four pixels into
 * a register and allocate one register per line of image being calculated.
 * In addition, say we also add all three error values to the pixels in the
 * line below the thresholded pixel in one step.  This is done by fetching the
 * {3/16 | 5/16 | 1/16} triple from a table indexed by error value.  This
 * reduces the number of READ-ADD-TRUNCATE-WRITE operations from 4 to 2.
 * We explored schemes that packed all the pixels receiving error values
 * into one register so there would be only one READ-ADD-TRUNCATE-WRITE op.
 * The extra shifting and masking required wiped out the speed advantage.
 * The best of these schemes was very complex, and no faster.
 *
 * Truncation is also messy, especially if the pixels are packed.  We need
 * at least one bit between the pixels to catch the carries.  These buffer
 * bits can then be formed into masks used for truncation.  This is pretty
 * slow.  However, if we reserve 2 buffer bits between pixels, we can avoid
 * the truncation task altogether.  Say we shift each pixel right by 2 before
 * use.  We use the two upper bits for overflow as shown below:
 *
 * Value	Interpretation
 * 00xxxxxx	small positive number
 * 01xxxxxx	big positive number
 * 10xxxxxx	big negative number
 * 11xxxxxx	small negative number
 *
 * What would the range of pixel values be?  The biggest value would come from
 * a bunch of 127 pixels surrounding a 255.  Every 127 pixel would have an error
 * of 127.  The sum of all the weighted errors would be 127 * (7+1+3+5)/16
 * = 127.  The highest value a pixel could attain would be 255 + 127 = 382.
 * This can be represented in 9 bits plus a sign bit. Disregarding the bottom
 * 2 bits, it fits into 8 bits.  The most negative pixel value comes from
 * a 0 pixel surrounded by 128's.  It would similarly have a modified value of
 * 0 + (-128) = -128.  This fits in 8 bits, or 6 bits disregarding the lower
 * 2 bits. This means we can't generate a "big negative number," the only
 * attainable values are 00xxxxxx, 01xxxxxx and 11xxxxxx.
 *
 * Is is really ok to use the most significant 6 bits out of each pixel?  We'll
 * try it and see.  My reasoning is that the amount of information we
 * lose going from 256 gray levels to 2 gray levels is virtually
 * indistinguishable from that lost by going from 64 levels to 2 levels.
 * Even if we had the world's best scanner, humans would be very hard
 * pressed to distinguish 256 levels of dithered gray from 64 levels of
 * dithered gray.
 *
 * Where else can we improve things?  The binary values could be calculated by
 * comparing with 127>>2 and setting the appropriate bit in the output image
 * if the pixel was greater.  This would require a write per pixel. Yuck.
 * A faster way (at least in assembler, so we'll try it here and hope the
 * compiler is smart) is to subtract 127, invert the sign bit, shift it into
 * the carry (add reg to itself) and shift the carry into the accumulated
 * binary output.  We can skip inverting the sign bit by simply adding
 * 01100000 to the pixel value instead of subtracting 127>>2.  Since the pixel
 * values are restricted to 00xxxxxx, 01xxxxxx and 11xxxxxx; the sum will
 * have a 1 in the sign bit if and only if the original value was positive
 * and greater than 00011111 ( = 127 >> 2).
 *
 * Well, ooops, a bug already.  In gray scale land, big numbers mean BRIGHT
 * (like white).  In binary land, 1 means black.  The code being replaced
 * by this routine produces images that have to be inverted.  What we really
 * want to do is have the MS bit be a 0 if and only if the orignal value
 * was positive and greater than 00011111.  So add 11100000.  If the
 * pixel was 000xxxxx, we get 1 in the MS bit.  If the pixel was 001xxxxx,
 * 010xxxxx or 011xxxxx, we get a 0 in the MS bit.  If the pixel was
 * 110xxxxx or 111xxxxx (negative), we get a 1 in the MS bit.  Just what
 * we want.
 *
 * Since we're trying to explore how things might work in the assembly code,
 * we'll write C code that handles 3 images lines at a time - the top, middle
 * and bottom lines.  Assembly code will have more "middle" lines.  The table
 * below distinguishes them.  Gray Buffer is a private array of pixels one
 * line wide.
 *
 *	Line	Source		Binary Dest 	Gray Dest
 *	------  -----------	-----------	-----------
 *	Top	Gray Buffer	Dest Image	N.A.
 *	Middle	Gray Image	Dest Image	N.A.
 *	Bottom	Gray Image	N.A.		Gray Buffer
 *
 * The areas of image being processed are not quite vertically aligned.
 * Each line's locus of interest is down and 2 pixels to the right as
 * shown below.  Dan Bloomberg suggests we call this the "Knight Shift".
 *
 * |       |       |       |  T1  |  E11  |
 * |       |  T2   |  E21  |  E22 |  E23  |
 * |  E31  |  E32  |  E33  |      |       |
 *
 * The inner loop of the algorithm is:
 *
 *  1. Compute Binary value for T1.  Shift it into binary output reg for
 *	top line.  If that reg is full, write it to the destination image.
 *  2. Shift Top line left by 1 pixel, incoming pixel coming from Gray Buffer.
 *  3. Add 7/16 error value from T1 to E11.
 *  4. Add {3/16 | 5/16 | 1/16} of T1's error to {E21 | E22 | E23}.
 *  5. Compute Binary value for T2.  Shift into binary output reg for
 *	middle line.  If that reg is full, write it to the destination image.
 *  6. Shift middle line left by 1 pixel, incoming pixel from Source image.
 *  7. Add 7/16 error value from T2 to E21.
 *  8. Add {3/16 | 5/16 | 1/16} of T2's error to {E31 | E32 | E33}.
 *  9. Shift Bottom line left by 1 pixel.  Accumulate leftmost gray pixel
 *	in gray output reg.  If gray output reg is now full (only holds
 *	4 pixels) write it to gray buffer.
 *  10. Repeat 1-9 until top line has just processed the rightmost pixel
 *	in its line.
 *
 * Nice pipeline, but how do we start and end a line?
 *
 *  START OF LINE
 *  1. Top line reg is loaded with first four pixels from Gray Buffer.
 *	Leftmost pixel corresponds to leftmost pixel in the line.
 *  2. Middle line reg is loaded with {p[2,-2]:p[2,-1]:pl[2,0]:p[2,1]}
 *	where p[2,-2]:p[2,-1] are the two pixels immediately to the
 *	left of the second line and p[2,0]:p[2,1] are the leftmost two
 *	pixels in the second line.  Note that we explicitly use the
 *	area surrounding the halftone to generate edge values.  This seems
 *	more reasonable than blindly assuming white or black edges. 
 *  3. The bottom reg is loaded with {p[3,-4]:p[3,-3]:p[3,-2]:p[3,-1]}.
 *
 * The scheme for entering and exiting the pipeline should not add crud to
 * the inner loop.  Define cDBlockHeight as the number of lines of binary
 * data produced on each pass across the image (i.e. doesn't include the
 * line of new error data).  We'll have two copies of the loop; one
 * for the left edge and one for the center and right edge of a line.
 * The left edge loop counts the number of horizontal steps taken by the
 * top line (= number of times we've gone through the loop).  The code for
 * the left loop is the same as the main inner loop shown above except no
 * binary or gray error data is written.  When the bottom line is filled
 * with valid pixel data (2*cDBlockHeight times through the loop), we write
 * that data to GrayBuf (the buffer containing the error data) and fall into
 * the main line code.
 *
 * The next loop is executed long enough for the lowest line producing
 * dithered data (the middle line in our case) to have written out its
 * last word.  We assume cDBlockHeight is less than 16, so the top line won't
 * try to write out an extra word on the right while we're waiting for the
 * last line.  How many cycles are required?  Say we start counting from the
 * left edge.  That's where the top line was aligned with the leftmost pixel.
 * The loop needs w cycles to for the top line to have computed all the valid
 * pixels.  It needs another (32 - (w MOD 32)) cycles to get to the next
 * word boundary.  Ooops, that's the right function unless w MOD 32 = 0.
 * Ok, it needs another ((32 - (w MOD 32)) & 0x1F) cycles to get to the
 * next word boundary.  Each line below the top line needs another 2 cycles
 * to get to the same point on its line, for a total of 2*cDBlockHeight cycles.
 * Let's define
 *	cyclesOnRight = 2*cDBlockHeight + ((32 - (w MOD 32)) & 0x1F)
 * We need to keep executing until loopCt (which started at 0
 * in the left loop and wasn't re-initialized) reaches (w + cyclesOnRight).
 *
 * We're not quite done yet.  The GrayBuf should be filled with enough data
 * to keep the top line happy.  How many bytes is that?  When we stop the loop,
 * the register accumulating GrayBuf's data holds at least 2 bytes off the
 * right side of the image.  The top line has just read an extra cyclesOnRight
 * bytes in the loop above, so we should keep accumulating, DRC'ing and storing
 * GrayBuf data until loopCt gets up to
 * (w + cyclesOnRight + (cyclesOnRight-2)).  After that, we keep going until
 * we hit a word (4 pixel) boundary to flush the last pixels.
 *
 * Much like the assembly version, the C code will have two routines:
 * ditherFloydPr and ip_ditherFloydLine.  ditherFloydPr does all the high level
 * stuff (checing args for errors, allocating pixrects, etc) and calls
 * ip_ditherFloydLine for every group of lines to be processed. In this C
 * version ip_ditherFloydLine will put out 2 lines of dithered data at each
 * pass.  The assembly version will put out more lines per pass.   
 *
 * There's still a question about what to do at the bottom of the image.  If
 * we always put out several lines, we run the risk of writing into the bottom
 * frame of the dithered image.  There are (at least) four choices:
 * 1. write a version of ip_ditherFloydLine that only does one line at a time
 * and call it appropriately at the bottom of the image; 2. Put a test and
 * branch into ip_ditherFloydLine at each line boundary; 3. Allow
 * ip_ditherFloydLine to mess up the bottom frame and clear it later.  By fiat,
 * choice 1 is too horrible to comtemplate (especially in the assembly
 * version).  I've got enough trouble just getting one version to work right.
 * Besides, it's slow.  Let's think about 2 vs 3.
 *
 * Say each frame-fixup word write takes 12 cycles and the test/branch pair
 * takes 2.  Since we produce one bit at a time, the test/branch pair will be
 * executed 32 times as often as the frame-fixup writes ( = 64 cycles/word).
 * Since 64 is bigger than 12, we'll do better to let ip_ditherFloydLine smash
 * the bottom frame, then fix it up at the end.
 *
 * **********************  FIRST REWORK + ASSEMBLER  ***********************
 *
 * Eventually (after debugging) that approach worked ok, but the image quality
 * wasn't quite as good as we would like.  The dynamic range was limited.
 * It was really only noticeable in the very light or very dark sections, but
 * it was noticeable there.
 *
 * We'd like to use more gray levels.  How can this be done without doing all
 * the error diffusion calculations in separate registers?  After some
 * contemplation, I had the brainstorm that we'd have lots of room if we only
 * kept *3* pixels in a register instead of 4.  Say we keep the image in
 * memory and in the GrayBuf packed at 4 pix/word, but spread things out
 * when pixels are processed.  We'll also have to rearrange the order in which
 * pixels are processed - adding the 7/16 error horizontally to a pixel before
 * adding the 1/16 error from the pixel above, but addition is communtative.
 * Here's the new inner loop with a new pixel diagram.  Note the inter-line
 * horizontal shift is 1 pixel, not 2.  At the start of the loop, the top
 * line's register holds E11:E12:E13.  The next line's reg holds
 * E21:E22:E23, etc.  The Ex3 pixel is held in bits 10:0.  Ex2 is held in bits
 * 21:11.  Ex1 is held in bits 31:22.  That leaves at least 2 sign bits between
 * each pixel (plus one extra bit above the lower pixels).  Since we are now
 * writing ip_ditherFloydLine in assembler, we can take advantage of the carry
 * flags to do compute the dither data at the same time we apply the 7/16
 * error diffusion term.
 *
 * |       |       |  E11 |  E12  |  E13  |
 * |       |  E21  |  E22 |  E23  |       |
 * |  E31  |  E32  |  E33 |       |       |
 *
 *  1. Put E11 into a separate register.
 *  2. Use E11 to select a 7/16 error value.  Add this error value to
 *	the gray scale reg containing E11:E12:E13.  This has two effects.
 *	The diffusion term is applied to E12.  If E11 was supposed to result
 *	in a white binary bit (0), something is added to the E11 pixel causing
 *	the carry flags to be set.  Otherwise, the carry flag remains clear.
 *  3. The binary register for the top line is added to itself with carry.
 *	This has the effect of shifting it left and inserting the new dithered
 *	bit.  Initially, the binary registers for all lines are set to 1.
 *	When this initial 1 drops into the carry bit after this "shift", we
 *	know the dithered register is full of valid data.
 *  4. Add {3/16 | 5/16 | 1/16} of E11's error to {E21 | E22 | E23}.
 *  5. If the dithered register was filled in step 3, invert it (we want
 *	bright pixels to become 0's, not 1's) and write it to memory.
 *  6. Shift the top gray register left by 11 bits (1 pixel), inserting the
 *	next pixel from GrayBuf into the bottom.
 *  7. Repeat steps 1 - 6 for the middle line.  Error terms are added to
 *	E22 and {E31 | E32 | E33} instead of E12 and {E21 | E22 | E23}.
 *	The new source value (new E23) is a DRC corrected version of the
 *	next pixel from the middle line of the source image.
 *  8. The bottom line is a bit easier since there's no error diffusion.
 *	Truncate E31 and shift the resulting 8-bit pixel into a register
 *	holding new GrayBuf pixels.  The truncation operation maps values
 *	in [0..255] to themselves, values greater than 255 to 255 and
 *	negative values to 0.  If the output register is now full, write it
 *	to GrayBuf.  Shift the bottom line's gray register left 1 pixel,
 *	inserting a newly corrected value from the source image.
 *  9. Repeat steps 1-8 until the middle line (lowest one producing
 *	dithered data) has just written its last full word.
 *
 * How do we get started?  Picture the left edge of the rectangle to be
 * dithered as the line to the left of E11 above.  E11:E12:E13 get the
 * first three pixels from GrayBuf.  E21 is the pixel just to the left
 * of the rectangle, E22:E23 are the leftmost two gray pixels.  E31:E32 are
 * uninitialized (since they'd be discarded) and E33 is the leftmost pixel
 * in the third line.
 *
 * The GrayBuf is filled with processed pixels.  It is written every 4th pixel
 * starting from the very first pixel processed.  But, we see that the first
 * cDBlockHeight pixels in the bottom line are trash, right?  No problem, we
 * just start reading the GrayBuf cDBlockHeight pixels from the beginning.
 * This skips the trash and keeps the inner loop simple.
 *
 * How do we finish?  As before, we need to process enough lines so that
 * the lowest line producing dithered bits (the middle line in this case)
 * has processed its last pixel, then gets to a word boundary.  w cycles gets
 * the top line across the rectangle.  Set cyclesOnRight to
 * 	cyclesOnRight = cDBlockHeight + ((32 - (w MOD 32)) & 0x1F)
 * So we do w + cyclesOnRight processing cycles.  At this point, the bottom
 * line has read 2 pixels to the right of the image.  So, as in the
 * first scheme, we read cyclesOnRight-2+4 more bytes and put them into
 * GrayBuf (the +4 term gets them into the output register.
 *
 * One nice feature of this scheme is that we no longer need to shift each
 * incoming pixel right before use.  A nasty feature is that we need to
 * truncate each modified pixel before storing it into GrayBuf.  As usual,
 * we'll do that with a table.
 *
 * 10/7/92: The previous version of the code occasionally wrote bits into
 * the right frame.  We now use ip_clearRightFrame to get rid of them.
 *
 */

/* PUBLIC */
/****************************************************************************
 *i_ditherFloyd()
 *		Takes an 8 bit/pixel image and produces a dithered
 *		1 bit/pixel image (Floyd-Steinberg algorithm) in the given
 *		rectangle (which may be the entire image).
 *	
 *	Input:
 *		pG = pointer to source image (including frame)
 *		gBpl = source bytes per line
 *		x = x coordinate of the top left corner of the
 *		    source rectangle relative to OUTSIDE of frame.
 *		y = y coordinate of the top left corner of the
 *		    source rectangle relative to OUTSIDE of frame.
 *		w = width of the rectangle
 *		h = height of the rectangle
 *		pDRCTable = pointer to DRC table
 *		pD = pointer to destination image (including frame)
 *		     Output starts at first non-frame pixel.
 *		dBpl = destination bytes per line
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes frame has already been placed in destination image.
 ****************************************************************************/
Int32  CDECL 
i_ditherFloyd (
    IPState    *pIPState,
    UInt8      *pG,
    Int32	gBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8      *pDRCTable,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    UInt32     	 gBlockStride, dBlockStride;
					/* amount to add to addresses
					 * to move down cDBlockHeight lines */
    Int32     	 line;			/* interation var */
    UInt8      	*pGrayBuf, *pAGrayBuf;	/* pointer to buffer that
					 * holds modified pixels from
					 * previous passes. pAGrayBuf is
					 * guranteed to be aligned on a word
					 * boundary. */
    Int32	 grayBufSize;		/* size of gray buffer */
    UInt8	*pDOrig;		/* pointer to upper left corner of
					 * dithered image.  Used to clear
					 * its frame. */
    Int32	 dBlockHeight;		/* number of lines of dithered
					 * image put out by 
					 * ip_ditherFloydLine in each pass */
    UInt32      *pHDitherErrorTable;	/* pointer to horizontal error table */
    UInt32      *pVDitherErrorTable;	/* pointer to vertical error table */
    UInt8       *pDitherTruncateTab;	/* pointer to pixel truncation table */

        /* On entry, [x, y, w, h] specifies a rectangle with respect to
	 * the upper left corner of the *frame*, not the upper left
	 * corner of the image!  The output binary image is always justified to
	 * the upper left corner of its framed pixrect (i.e. inside the
	 * frame). 
	 */

	/* Set these to null in case we need to bail out before they're
	 * set properly.
	 */
    pGrayBuf = NULL;
    pHDitherErrorTable = NULL;
    pVDitherErrorTable = NULL;
    pDitherTruncateTab = NULL;

	/* point to the upper left corner of the rect in gray image.  Note that
	 * x and y include the frame. */
    pG += (y * gBpl) + x;

    pD += (dBpl << cLogFrameBits) + cFrameBytes; /* skip the binary frame */

	/* remember upper left corner of dest image so we can clear its
	 * frame at the end. */
    pDOrig = pD;

        /* Make the error diffusion tables.  First allocate them,
	 * then call ip_makeErrorDiffusionTables to initialize them.
	 */
    pHDitherErrorTable = (UInt32 *) MALLOC(cDitherHErrorTabSize);
    if (pHDitherErrorTable == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloydFin;
    }
    pVDitherErrorTable = (UInt32 *) MALLOC(cDitherVErrorTabSize);
    if (pVDitherErrorTable == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloydFin;
    }
    pDitherTruncateTab = (UInt8 *) MALLOC(cDitherTruncateTabSize);
    if (pDitherTruncateTab == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloydFin;
    }
    if ((__error__ = ip_makeErrorDiffusionTables (pIPState, pHDitherErrorTable,
					        NULL,
					        pVDitherErrorTable,
					        pDitherTruncateTab))
	!= ia_successful)
        goto ditherFloydFin;

	/* It is unfortunate, but each of the primitives that do the
	 * real dithering work slightly differently.  We use a machine
	 * dependent setup routine to do whatever's necessary to
	 * get them going.  First, allocate the gray buffer after
	 * figuring out how big it is.
	 */
    __error__ = ip_getSizeOfDitherGrayBuf (pIPState,  gBpl, w, &grayBufSize);
    if (__error__ != ia_successful)
	goto ditherFloydFin;
    pGrayBuf = MALLOC(grayBufSize);
    if (pGrayBuf == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloydFin;
    }
    __error__ = ip_floydSetup (pIPState, pHDitherErrorTable, pVDitherErrorTable,
			       pDitherTruncateTab, &dBlockHeight,
			       pG, pGrayBuf, pDRCTable, w);
    if (__error__ != ia_successful)
	goto ditherFloydFin;

    pAGrayBuf = (UInt8 *) ((UInt32)(pGrayBuf + 3) & ~3);

	/* point to the next line of source image.  That's where
	 * ip_ditherFloydLine will get its first input data. */
    pG += gBpl;

	/* Have ip_ditherFloydLine actually do the work in units of
	 * dBlockHeight lines. */
    gBlockStride = dBlockHeight * gBpl;
    dBlockStride = dBlockHeight * dBpl;
    __error__ = ia_aborted;		/* which will be true if we
					 * are forced to stop in the middle */
    for (line = 0; line < h;
	 line += dBlockHeight, pG += gBlockStride, pD += dBlockStride)
    {
	ip_ditherFloydLine (pIPState, pG, gBpl, pD, dBpl,
			    pAGrayBuf, w, dBlockHeight,
			    pHDitherErrorTable, pVDitherErrorTable,
			    pDitherTruncateTab, pDRCTable);
	IP_YIELD(ditherFloydFin)
    }

	/* fix up the bottom frame. */
    if ((h % dBlockHeight) != 0)
	i_rasterOp (pIPState,
		    pD - ((dBlockHeight - (h % dBlockHeight)) * dBpl),
		   				/* destination image */
		    0,				/* x offset */
		    0,				/* y offset */
		    w+31,			/* width of dest rectangle
						 * Make sure we get to the
						 * next word boundary since
						 * ip_clearRightFrame doesn't
						 * reach down here */
		    dBlockHeight - (h % dBlockHeight), /* dest rect height */
		    1,				/* depth of dest */
		    dBpl,			/* line bytes */
		    PIX_CLR,
		    NULL,			/* source image */
		    0,				/* x offset */
		    0,				/* y offset */
		    0,				/* depth */
		    0);				/* line bytes */
		  
	/* Clear the right frame bits that have been written by
	 * ip_ditherFloydLine. */
    ip_clearRightFrame (pIPState, pDOrig, w, h, dBpl, 1);
    __error__ = ia_successful;

	/* That's it, free grayBuf and return the dithered pixrect */
ditherFloydFin:
    if (pGrayBuf != NULL)            FREE(pGrayBuf); 
    if (pHDitherErrorTable != NULL)  FREE(pHDitherErrorTable);
    if (pVDitherErrorTable != NULL)  FREE(pVDitherErrorTable);
    if (pDitherTruncateTab != NULL)  FREE(pDitherTruncateTab);
    NORMAL_RETURN;
}




/* PUBLIC */
/*****************************************************************************
 * i_getSizeOfDitherState()
 *	Set up for banded Floyd-Steinberg dither. 
 *	This routine returns the size of the internal dither state block
 *	so clients can allocate it.
 *	
 * Arguments:
 *		gBpl = bytes/line of gray image.
 *		w = width of source image
 *		pSize = pointer to returned size of private dither state block.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *
 *****************************************************************************/

Int32  CDECL 
i_getSizeOfDitherState (
    IPState    *pIPState,
    Int32	gBpl,
    Int32	w,
    Int32      *pSize)
{
    PROLOGUE;
    Int32	grayBufSize;
    

	/* A dither state block is composed of a structure preceeding a
	 * buffer that holds one line of gray pixels.  The size of the
	 * structure is a constant.  Get the size of the gray buffer.
	 */
    __error__ = ip_getSizeOfDitherGrayBuf (pIPState, gBpl, w, &grayBufSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* return the real size */
    *pSize = sizeof(DitherState) + grayBufSize;

    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * i_ditherFloydInit()
 *	Set up for banded Floyd-Steinberg dither. 
 *	Warning: this copies the first line to the "gray" buffer, and
 *	sort of expects that the second line will be passed. 
 *	The net result is that it'll always dither the first line twice.
 *	Fixing it would require the caller to pass an extra line in the
 *	first block.
 *	
 * Arguments:
 *		pG = pointer to source image's frame
 *		gBpl = bytes/line for gray (input) image
 *		x, y = starting (x,y) offset of source rectangle
 *		       relative to OUTSIDE of frame.
 *		w = width of source images
 *		pDRCTable = pointer to DRC table
 *		blockHeight = the number of lines in a band, set by client.
 *		pState = pointer to state block to be initialized.  This
 *			 should have the correct number of bytes as
 *			 determined by i_getSizeOfDitherState.
 *		stateSize = number of bytes in state block being passed in.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_invalidParm	input state block is too small
 *
 *
 *****************************************************************************/

Int32  CDECL 
i_ditherFloydInit (
    IPState    *pIPState,
    UInt8      *pG,
    Int32	gBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    UInt8      *pDRCTable,
    Int32	blockHeight,
    void       *pState,
    Int32	stateSize)
{
    PROLOGUE;
    DitherState *s;
    Int32	 reqStateSize;
    UInt32	*pVDitherErrorTable;
    UInt32	*pHDitherErrorTable;
    UInt32	*pSingHDitherErrorTable;
    UInt8	*pDitherTruncateTable;
    UInt8	*pGrayBuf;
    Int32	 i;
    

	/* see if passed in state block is big enough */
    __error__ = i_getSizeOfDitherState (pIPState, gBpl, w, &reqStateSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    if (stateSize < reqStateSize)
	return ia_invalidParm;

	/* initialize the internal pointers */
    s = (DitherState *) pState;
    pVDitherErrorTable = &(s->VDitherErrorTable[0]);
    pHDitherErrorTable = &(s->HDitherErrorTable[0]);
    pSingHDitherErrorTable = &(s->SingHDitherErrorTable[0]);
    pDitherTruncateTable = &(s->DitherTruncateTab[0]);
    pGrayBuf = (UInt8 *)s + sizeof(DitherState);

        /* Make the error diffusion tables */
    if ((__error__ = ip_makeErrorDiffusionTables (pIPState, pHDitherErrorTable,
					        pSingHDitherErrorTable,
					        pVDitherErrorTable,
					        pDitherTruncateTable))
	!= ia_successful)
	goto ditherFloydErr;

	/* point to the upper left corner of the rect in gray image.  Note that
	 * x and y include the frame. */
    pG += (y * gBpl) + x;

	/* It is unfortunate, but each of the primitives that do the
	 * real dithering work slightly differently.  We use a machine
	 * dependent setup routine to do whatever's necessary to
	 * get them going.
	 */
    __error__ = ip_floydSetup (pIPState, pHDitherErrorTable, pVDitherErrorTable,
			     pDitherTruncateTable, &(s->dBlockHeight),
			     pG, pGrayBuf, pDRCTable, w);
    if (__error__ != ia_successful)
	goto ditherFloydErr;

        /* Set up how many dithering multiple line calls to do and how many
	 * dithering single line calls to do each Process call */
    s->blocksPer  = blockHeight / s->dBlockHeight;
    s->singlesPer = blockHeight % s->dBlockHeight;

    s->grayBufOffset =  ((UInt32)(pGrayBuf + 3) & ~3) - (UInt32)s;

    s->w = w;
    s->depth = 8;

	/* copy the DRC table to the private state */
    for (i = 0; i < cDitherNumDRCTabEntries; i++)
	s->DRCTable[i] = pDRCTable[i];

    NORMAL_RETURN;

ditherFloydErr:
    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * i_ditherFloydProcess
 *
 *	Dither a block of lines. The number of lines dithered will be
 *	dBlockHeight * blocksPer + singlesPer
 * 
 * Arguments:
 *	src = Source image buffer, pointer to frame
 *	sBpl = source image bytes/line
 *	xOffset = source image X offset relative to outside of frame
 *	yOffset	= source image Y offset relative to outside of frame
 *	pDest = destination buffer, pointer to frame
 *		Output is written to first non-frame pixel.
 *	dBpl = bytes/line in destination buffer
 *	state = state thing returned by the i_ditherFloydInit
 *
 * Returns:
 *	error status. If aborted, call finish to clean up.
 *****************************************************************************/

Int32 CDECL 
i_ditherFloydProcess (
    IPState    *pIPState,
    UInt8      *src,
    Int32	sBpl,
    Int32	xOffset,
    Int32	yOffset,
    UInt8      *dest,
    Int32	dBpl,
    void       *state)
{
    PROLOGUE;
    Int32 i, dBlockStride, sBlockStride;
    DitherState *s;
    UInt8 *dp;

    s = (DitherState *) state;
    if (s->depth == 4)
    {
	ERR_PASS( i_ditherFloyd4to1Process (pIPState, src, sBpl,
					    xOffset, yOffset, dest,
					    dBpl, state));
	NORMAL_RETURN;
    }

    /* Move past the frame in the dest buffer */
    dest += (dBpl << cLogFrameBits) + cFrameBytes;
    dp = dest;

    dBlockStride = dBpl * s->dBlockHeight;
    sBlockStride = sBpl * s->dBlockHeight;

    /* Move to the right place in the source buffer */
    src += sBpl * yOffset + xOffset;

    /* Do the specified number of multiple line output dithering calls */
    for (i = 0; i < s->blocksPer; i++)
    {
	ip_ditherFloydLine (pIPState, src, sBpl, dp, dBpl,
			    (UInt8 *)((UInt32)s + s->grayBufOffset),
			    s->w,
			    s->dBlockHeight,
			    &(s->HDitherErrorTable[0]),
			    &(s->VDitherErrorTable[0]),
			    &(s->DitherTruncateTab[0]),
			    &(s->DRCTable[0]));
	IP_YIELD(ditherFloydProcessFin)
	src += sBlockStride;
	dp += dBlockStride;
    }
    
    /* Do the specified number of single line output dithering calls */
    for (i = 0; i < s->singlesPer; i++)
    {
	ip_ditherFloydSingleLine (pIPState, src, sBpl, dp, dBpl,
			          (UInt8 *)((UInt32)s + s->grayBufOffset),
				  s->w,
				  s->dBlockHeight,
				  &(s->SingHDitherErrorTable[0]),
				  &(s->VDitherErrorTable[0]),
				  &(s->DitherTruncateTab[0]),
				  &(s->DRCTable[0]));
	IP_YIELD(ditherFloydProcessFin)
	src += sBpl;
	dp += dBpl;
    }

    /* Clear the right frame bits that have been written by the previous
     * two functions: ip_ditherFloydLine and ip_ditherFloydSingleLine. */
    ip_clearRightFrame (pIPState, dest, s->w, (dp - dest) / dBpl, dBpl, 1);

ditherFloydProcessFin:
    NORMAL_RETURN;
}

/* PUBLIC */
/*****************************************************************************
 * i_ditherFloydSingleProcess
 *
 *	Dither one line.
 * 
 * Arguments:
 *	src = Source image buffer, pointer to frame
 *	sBpl = source image bytes/line
 *	xOffset = source image X offset relative to outside of frame
 *	yOffset	= source image Y offset relative to outside of frame
 *	pDest = destination buffer, pointer to frame
 *		Output is written to first non-frame pixel.
 *	dBpl = bytes/line in destination buffer
 *	state = state thing returned by the i_ditherFloydInit
 *
 * Returns:
 *	error status. If aborted, call finish to clean up.
 *****************************************************************************/

Int32 CDECL 
i_ditherFloydSingleProcess (
    IPState    *pIPState,
    UInt8      *src,
    Int32	sBpl,
    Int32	xOffset,
    Int32	yOffset,
    UInt8      *dest,
    Int32	dBpl,
    void       *state)
{
    PROLOGUE;
    DitherState *s;

    s = (DitherState *) state;
    if (s->depth == 4)
    {
	ERR_PASS( i_ditherFloyd4to1SingleProcess (pIPState, src, sBpl,
					          xOffset, yOffset,
					          dest, dBpl, state));
	NORMAL_RETURN;
    }

    /* Move past the frame in the dest buffer */
    dest += (dBpl << cLogFrameBits) + cFrameBytes;

    /* Move to the right place in the source buffer */
    src += sBpl * yOffset + xOffset;
    
    ip_ditherFloydSingleLine (pIPState, src, sBpl, dest, dBpl,
			      (UInt8 *)((UInt32)s + s->grayBufOffset),
			      s->w,
			      s->dBlockHeight,
			      &(s->SingHDitherErrorTable[0]),
			      &(s->VDitherErrorTable[0]),
			      &(s->DitherTruncateTab[0]),
			      &(s->DRCTable[0]));

    /* Clear the right frame */
    ip_clearRightFrame (pIPState, dest, s->w, 1, dBpl, 1);

    IP_YIELD(ditherFloydProcessFin)
    src += sBpl;
    dest += dBpl;
    
ditherFloydProcessFin:
    NORMAL_RETURN;
}

/*****************************************************************************
 * i_ditherFloydFinish 
 *
 *	We're done dithering (or aborting), clean up any junk we might
 *	have left in the frame
 *
 * Input:
 *	pD = pointer to destination image (including frame)
 *	dBpl = destination bytes per line
 *	w = width of the rectangle
 *	h = height of the rectangle
 *	state = state thing initialized by the i_ditherFloydInit
 *
 * Returns:
 *	error status. 
 *****************************************************************************/
Int32 CDECL 
i_ditherFloydFinish (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    void       *state)
{
    PROLOGUE;
    DitherState *s;

    s = (DitherState *) state;

    if (s->depth == 4)
	return i_ditherFloyd4to1Finish (pIPState, pD, dBpl, w, h, state);

	/* That's it.  Return the dithered pixrect */
    NORMAL_RETURN;
}


/* The monochrome dithering routines worked just fine.  However, our first
 * real printer takes 2 bit/pixel images, not monochrome.  We need a dither
 * routine that produces 2 bpp images from 8 bpp images.
 *
 * In the long run, it might be nicer to have one dither routine that was
 * passed an argument telling whether to produce a 1 or 2 bpp output.  That
 * would require changes to the current interface.  This is a very touchy
 * subject.  For now, we'll introduce a new routine name.  At some later point,
 * these may be consolidated.
 *
 * The changes aren't too bad.  Generating a bigger output image is trivial.
 * There are two more interesting changes.  First, each shift step producing
 * output bits gets two bits instead of one.  Output registers are written
 * to memory twice as often.
 *
 * The other interesting change is in the diffusion tables.  The three values
 * written are:
 *
 * Value	Intensity
 *  00		completely black
 *  01		1/3 white
 *  10		2/3 white
 *  11		completely white
 *
 * This changes the error values.  We still divide the input range into four
 * equally slices by the two upper bits.  However the error values are
 *
 * 	x in Range		Basic Error value	Digital Output
 * [-infinity, 00111111]	MAX(0, x)		  00
 * [01000000,  01111111]	x - (1*255/3)		  01
 * [10000000,  10111111]	x - (2*255/3)		  10
 * [11000000, +infinity]	MIN(x - 255, 0)		  11
 *
 * The first error value is always >= 0, the last always <= 0.  The
 * middle two might be positive or negative.  These basic values are
 * distributed using the familiar 7/16, 3/16, 5/16 and 1/16 fractions.
 *
 * How do we generate the correct 2 bit data?  As seen from the table above,
 * we can't just grab a pair of bits from the thresholded pixel because it
 * hasn't been truncated yet.  It would be unpleasant to add an explicit
 * truncation operation.
 *
 * Errors in the 7/16 table are normally added only to the middle pixel of
 * the register register containing the thresholded pixel.  For example,
 * we would add 7/16 of E11's basic error value from the table above to
 * E21.  Say we do this before shifting the register containing E11 and E21.
 * This means E11 is still in the most significant bits of the register.
 * In this code, we'll add some number to the upper pixel so that the
 * top two bits [31:30] of the result end up being the desired output value
 * for that pixel.  For example, if the thresholded pixel is negative
 * (bits [31:30] = 11), we add a number having 01 in bits [31:30].  After
 * the 7/16 value is added in, we use bits [31:30] as the truncated,
 * digitized output value of the pixel.
 *
 */

/* PUBLIC */
/**************************************************************************
 *  i_ditherFloyd2():
 *		Takes an 8 bit/pixel image and produces a dithered
 *		2 bit/pixel image (Floyd-Steinberg algorithm) in the given
 *		rectangle (which may be the entire image).
 *	
 *	Input:
 *		pG = pointer to source image (including frame)
 *		gBpl = source bytes per line
 *		x = x coordinate of the top left corner of the
 *		    source rectangle relative to outside of frame.
 *		y = y coordinate of the top left corner of the
 *		    source rectangle relative to outside of frame.
 *		w = width of the rectangle
 *		h = height of the rectangle
 *		pDRCTable = pointer to DRC table
 *		pD = pointer to destination image (including frame)
 *		     Output starts at first non-frame pixel.
 *		dBpl = destination bytes per line
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes frame has already been placed in destination area.
 **************************************************************************/
Int32  CDECL 
i_ditherFloyd2 (
    IPState    *pIPState,
    UInt8      *pG,
    Int32	gBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8      *pDRCTable,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    UInt32     	 gBlockStride, dBlockStride; /* amount to add to addresses
					 * to move down dBlockHeight lines */
    Int32     	 line;			/* interation vars */
    UInt8      	*pGrayBuf, *pAGrayBuf;	/* pointer to buffer that
					   holds modified pixels from
					   previous passes. pAGrayBuf is
					   guranteed to be aligned on a word
					   boundary. */
    Int32	 grayBufSize;		/* size of gray buffer */
    UInt8	*pDOrig;		/* pointer to upper left corner of
					 * dithered image.  Used to clear
					 * its frame. */
    		/* return code */
    Int32	 dBlockHeight;		/* number of lines of dithered
					 * image put out by 
					 * ip_ditherFloydLine in each pass
					 */
    UInt32	*pHDitherErrorTable2;   /* ptr to vertical error table */
    UInt32	*pVDitherErrorTable2;   /* ptr to horizontal error table */
    UInt8	*pDitherTruncateTab;    /* ptr to pixel truncation table */


	/* On entry, [x, y, w, h] specifies a rectangle with respect to
	 * the upper left corner of the *frame*, not the upper left
	 * corner of the image!  The output binary image is always justified to
	 * the upper left corner of its framed pixrect (i.e. inside the
	 * frame). 
	 */

	/* Set these to null in case we need to bail out before they're
	 * set properly.
	 */
    pGrayBuf = NULL;
    pHDitherErrorTable2 = NULL;
    pVDitherErrorTable2 = NULL;
    pDitherTruncateTab  = NULL;

	/* point to the upper left corner of the rect in gray image.  Note that
	 * x and y include the frame. */
    pG += (y * gBpl) + x;

    pD += (dBpl << cLogFrameBits) + 2*cFrameBytes; /* skip the 2 bpp frame */

	/* remember upper left corner of dest image so we can clear its
	 * frame at the end. */
    pDOrig = pD;

        /* Make the error diffusion tables.  First allocate them,
	 * then call ip_make2ErrorDiffusionTables to initialize them. */
    pHDitherErrorTable2 = (UInt32 *) MALLOC(cDitherHErrorTabSize);
    if (pHDitherErrorTable2 == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd2Fin;
    }
    pVDitherErrorTable2 = (UInt32 *) MALLOC(cDitherVErrorTabSize);
    if (pVDitherErrorTable2 == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd2Fin;
    }
    pDitherTruncateTab = (UInt8 *) MALLOC(cDitherTruncateTabSize);
    if (pDitherTruncateTab == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd2Fin;
    }

    if ((__error__ = ip_make2ErrorDiffusionTables (pIPState, pHDitherErrorTable2,
						 pVDitherErrorTable2,
						 pDitherTruncateTab))
	!= ia_successful)
        goto ditherFloyd2Fin;

	/* It is unfortunate, but each of the primitives that do the
	 * real dithering work slightly differently.  We use a machine
	 * dependent setup routine to do whatever's necessary to
	 * get them going
	 */
    __error__ = ip_getSizeOfDither2GrayBuf (pIPState, gBpl, w, &grayBufSize);
    if (__error__ != ia_successful)
	goto ditherFloyd2Fin;
    pGrayBuf = MALLOC(grayBufSize);
    if (pGrayBuf == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd2Fin;
    }
    __error__ = ip_floyd2Setup (pIPState,
			      pHDitherErrorTable2, pVDitherErrorTable2,
			      pDitherTruncateTab, &dBlockHeight,
			      pG, pGrayBuf, pDRCTable, w);
    if (__error__ != ia_successful)
	goto ditherFloyd2Fin;

    pAGrayBuf = (UInt8 *) ((UInt32)(pGrayBuf + 3) & ~3);

	/* point to the next line of source image.  That's where
	 * ip_ditherFloydLine2 will get its first input data. */
    pG += gBpl;

	/* Have ip_ditherFloydLine2 actually do the work in units of
	 * dBlockHeight lines. */
    gBlockStride = dBlockHeight * gBpl;
    dBlockStride = dBlockHeight * dBpl;
    __error__ = ia_aborted;		/* which will be true if we
					 * are forced to stop in the middle */
    for (line = 0; line < h;
	 line += dBlockHeight, pG += gBlockStride, pD += dBlockStride)
    {
	ip_ditherFloydLine2 (pIPState, pG, gBpl, pD, dBpl,
			     pAGrayBuf, w, dBlockHeight,
			     pHDitherErrorTable2, pVDitherErrorTable2,
			     pDitherTruncateTab, pDRCTable);
	IP_YIELD(ditherFloyd2Fin)
    }

	/* fix up the bottom frame. */
    if ((h % dBlockHeight) != 0)
    {
	i_rasterOp (pIPState,
			/* dest image ptr */
		    pD - ((dBlockHeight - (h % dBlockHeight)) * dBpl),
	            0,				/* x offset */
	            0,				/* y offset */
		    w+15,			/* width of dest rect.  Make
						 * sure we get to next word
						 * boundary since
						 * ip_clearRightFrame doesn't
						 * get down here. */
		    dBlockHeight - (h % dBlockHeight), /* dest rect height */
		    2,				/* depth of dest */
		    dBpl,			/* dest image bytes/line */
		    PIX_CLR,
		    NULL,			/* source image */
		    0,				/* x offset */
		    0,				/* y offset */
		    0,				/* depth */
		    0);				/* line bytes */
    }
	
	/* Clear the right frame bits that have been written by
	 * ip_ditherFloydLine. */
    ip_clearRightFrame (pIPState, pDOrig, w, h, dBpl, 2);
    __error__ = ia_successful;

ditherFloyd2Fin:
	/* That's it, free grayBuf and return the dithered pixrect */
    if (pGrayBuf)             FREE(pGrayBuf); 
    if (pHDitherErrorTable2)  FREE(pHDitherErrorTable2);
    if (pVDitherErrorTable2)  FREE(pVDitherErrorTable2);
    if (pDitherTruncateTab)   FREE(pDitherTruncateTab);
    NORMAL_RETURN;
}

/* The purpose of this routine is to convert 4-bit gray scale images to
 * binary, dithered images.  This is not a simple binary threshold.  The
 * number of dots in a region of the binary image should be (roughly)
 * proportional to the gray value of the image.  The binary image should
 * be a faithful reproduction of the original.
 *
 * In addition, we will also apply dynamic range correction to the
 * gray pixels.  Each pixel's intensity will be mapped to some other
 * intensity using a table provided by the caller.  Doing the mapping
 * here saves the time required to write another complete gray image.
 *
 * Like many diffusion algorithms, we start processing in the upper left
 * corner of an image.  There are two basic steps - threshold compare and
 * error diffusion.  The threshold compare is simple, we just use the most
 * significant bit of the pixel as the binary value.  Printing a "1" is
 * really only completely correct if the gray value is 0.  Printing a
 * 0 is only completely correct if the gray value is 255.  Using this
 * interpretation, we calculate the "error" as the difference between
 * the gray pixel and the "displayed" value.  For example, if the gray
 * pixel value p is less than 8, the pixel is actually p too large for
 * the value to be displayed.  Similarly, if p >= 8, the error is
 * p - 15 (a number less than or equal to 0).
 *
 * In Floyd-Steinberg, we apportion this error value
 * among 4 other gray pixels as shown below.  The "*" represents the
 * newly thresholded pixel.  The fractions represent the proportions of the
 * error value that is *added* to the pixel to the right and the three 
 * pixels below.
 * |      |  *   | 7/16 |
 * | 3/16 | 5/16 | 1/16 |
 *
 * We convert the incoming pixels to 8 bit per pixel values by 
 * concatenating 4 bits of 0 to the pixel so that xxxx becomes xxxx0000.
 * This means that we can keep 8 bits of error value and hopefully get
 * a better looking image that if we only kept 4 bits.
 *
 * Since we use 8 bits per pixel this code now looks much like the 
 * 8 to 1 error diffusion code.  See comments at the beginning of this
 * file for specific comments about that algorithm.  I'll mention only
 * differences from that algorithm from now on.
 *
 * The inner loop of the algorithm is:
 *
 * 1. Process the top line:
 *	Add horizontal error to top line and get next dest pixel.
 *	Add vertical error to middle line.
 *	Has marker bit fallen off top of destination output word?
 *		Yes, write destination word.
 *		Reset destination word with marker bit.
 *		Is loop count 32?
 *			Yes, set marker bits in middle lines.
 *			No, do nothing
 *	Fetch new pixel from the gray buffer
 *
 * 2. Process middle lines:
 *	Add horizontal error to middle line and get next dest pixel
 *	Add vertical error to bottom line
 *	Has marker bit fallen off top of destination output word?
 *		Write destination word.
 *		Reset destination word with marker bit.
 *	Fetch new pixel from the image.
 *
 * 3. Process bottom line:
 *	Truncate next pixel and place in grayLineOut
 *	Is grayLineOut full of data?
 *		Yes, write grayLineOut to gray buffer
 * 	Fetch new pixel from the image.
 *
 * 4. Do 1, 2 and 3 again.
 *
 * In order to keep from reading the image twice for each pixel
 * within a 8 bit word in the image, an 8 bit quantity is buffered in
 * a variable and the 2 pixels are used from it.  One read must be done
 * per loop.  For the middle line the 8 bit quantity is read during (2)
 * and the second have used during (4).  For the bottom line, the 8 bit
 * quantity is used during (2) and read during (4).
 *
 * What will the drc table look like?  The caller passes in a table with
 * 16 entries each 1 byte in length.  We expect the caller to use all 8
 * bits as this seems to be very susceptible to changes in the drc table.
 * We will transfer that to 2 tables.  One table will be used to convert
 * the high nibble.  The table is indexed by an 8 bit quantity,
 * xxxxyyyy, xxxx is the pixel in the high nibble and yyyy is any value
 * probably the pixel in the low nibble but it is ignored.  This means
 * we don't have to do any shifting to access the drc table.  The 
 * second table will be used to convert the low nibble.  The table is 
 * indexed by an 8 bit quantity, xxxxyyyy, where yyyy is the pixel in the
 * low nibble and xxxx is any value probably the pixel in the high
 * nibble but it is ignored.  Again, this is so we don't have to do
 * any shifting to access the table.
 *
 */

/* PUBLIC */
/*****************************************************************************
 * i_ditherFloyd4to1()
 *		Takes an 4 bit/pixel image and produces a dithered
 *		1 bit/pixel image (Floyd-Steinberg algorithm) in the given
 *		rectangle (which may be the entire image).
 *	
 *	Input:
 *		pG = pointer to source image (including frame)
 *		gBpl = source bytes per line
 *		x = x coordinate of the top left corner of the
 *		    source rectangle relative to outside of frame.
 *		y = y coordinate of the top left corner of the
 *		    source rectangle relative to outside of frame.
 *		w = width of the rectangle
 *		h = height of the rectangle
 *		pDRCTable = pointer to DRC table
 *		pD = pointer to destination image (including frame)
 *		     Output starts at first non-frame pixel.
 *		dBpl = destination bytes per line
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes frame has already been placed in destination area.
 ****************************************************************************/
Int32  CDECL 
i_ditherFloyd4to1 (
    IPState    *pIPState,
    UInt8      *pG,
    Int32	gBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8      *pDRCTable,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    UInt32	 gBlockStride, dBlockStride;
					/* amount to add to addresses
					 * to move down cDBlockHeight lines */
    Int32	 line, ibyte, obyte;	/* interation vars */
    UInt8	*pGrayBuf, *pAGrayBuf;	/* pointer to buffer that
					 * holds modified pixels from
					 * previous passes. pAGrayBuf is
					 * guranteed to be aligned on a word
					 * boundary. */
    Int32	 grayBufSize;		/* size of gray buffer */
    UInt8	*pDOrig;		/* pointer to upper left corner of
					 * dithered image.  Used to clear
					 * its frame. */
    		/* return code */
    UInt8	*pHDRCTable;		/* drc table adjusted for our
					 * internal usage of pixel in high 
					 * nibble */
    UInt8	*pLDRCTable;		/* drc table adjusted for our
					 * internal usage of pixel in low 
					 * nibble */
    UInt32	 pixel;
    Int32	 dBlockHeight;		/* number of lines of dithered
					 * image put out by 
					 * ip_ditherFloydLine in each pass
					 */
    UInt32	*pHDitherErrorTable;	/* pointer to horizontal error table */
    UInt32	*pVDitherErrorTable;	/* pointer to vertical error table */
    UInt8 	*pDitherTruncateTab;	/* pointer to pixel truncation table */

	/* On entry, [x, y, w, h] specifies a rectangle with respect to
	 * the upper left corner of the *frame*, not the upper left
	 * corner of the image!  The output binary image is always justified to
	 * the upper left corner of its framed pixrect (i.e. inside the
	 * frame). 
	 */

	/* Set these to null in case we need to bail out before they're
	 * set properly.
	 */
    pGrayBuf = NULL;
    pHDRCTable = NULL;
    pHDitherErrorTable = NULL;
    pVDitherErrorTable = NULL;
    pDitherTruncateTab = NULL;

	/* point to the upper left corner of the rect in gray image.  Note that
	 * x and y include the frame. 
	 * x is the number of pixels over.  We need to multiply by
	 * bits per pixel and then divide by bits per byte to find the
	 * number of bytes that need to be added to pG
	 */
    pG += (y * gBpl) + (x*4/8);
    pixel = x & 1;

    pD += (dBpl << cLogFrameBits) + cFrameBytes;  /* skip the binary pad */

	/* remember upper left corner of dest image so we can clear its
	 * frame at the end. */
    pDOrig = pD;

        /* Make the error diffusion tables.  First allocate them,
	 * then call ip_makeErrorDiffusionTables to initialize them.
	 */
    pHDitherErrorTable = (UInt32 *) MALLOC(cDitherHErrorTabSize);
    if (pHDitherErrorTable == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd4to1Fin;
    }
    pVDitherErrorTable = (UInt32 *) MALLOC(cDitherVErrorTabSize);
    if (pVDitherErrorTable == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd4to1Fin;
    }
    pDitherTruncateTab = (UInt8 *) MALLOC(cDitherTruncateTabSize);
    if (pDitherTruncateTab == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd4to1Fin;
    }
    if ((__error__ = ip_makeErrorDiffusionTables (pIPState, pHDitherErrorTable,
					        NULL,
					        pVDitherErrorTable,
					        pDitherTruncateTab))
	!= ia_successful)
	goto ditherFloyd4to1Fin;

	/* translate the drc table into two suitable for our internal
	 * use.
	 */
    pHDRCTable = (UInt8 *)MALLOC(sizeof(UInt8)*256*2);
    pLDRCTable = pHDRCTable + 256;
    if (pHDRCTable == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd4to1Fin;
    }
    for (ibyte = 0; ibyte < 16; ibyte++)
    {
		/* first do the drc table for the pixel in the high
		 * nibble */
	for (obyte = ibyte << 4; obyte < (ibyte << 4) + 16; obyte++)
	    pHDRCTable[obyte] = pDRCTable[ibyte];
		/* now do the drc table for the pixel in the low
		 * nibble */
	for (obyte = ibyte; obyte < 256; obyte += 16)
	    pLDRCTable[obyte] = pDRCTable[ibyte];
    }

	/* It is unfortunate, but each of the primitives that do the
	 * real dithering work slightly differently.  We use a machine
	 * dependent setup routine to do whatever's necessary to
	 * get them going.  First, allocate the gray buffer after
	 * figuring out how big it is.
	 */
    __error__ = ip_getSizeOfDither4to1GrayBuf (pIPState, gBpl, w, &grayBufSize);
    if (__error__ != ia_successful)
	goto ditherFloyd4to1Fin;
    pGrayBuf = MALLOC(grayBufSize);
    if (pGrayBuf == NULL)
    {
	__error__ = ia_nomem;
	goto ditherFloyd4to1Fin;
    }
    __error__ = ip_floyd4To1Setup (pIPState,
				   pHDitherErrorTable, pVDitherErrorTable,
				   pDitherTruncateTab, &dBlockHeight, pG,
				   pGrayBuf, pHDRCTable, pLDRCTable, w);

    if (__error__ != ia_successful)
	goto ditherFloyd4to1Fin;

    pAGrayBuf = (UInt8 *) ((UInt32)(pGrayBuf + 3) & ~3);

	/* point to the next line of source image.  That's where
	 * ip_ditherFloydLine will get its first input data. */
    pG += gBpl;

	/* Have ip_ditherFloydLine actually do the work in units of
	 * dBlockHeight lines. */
    gBlockStride = dBlockHeight * gBpl;
    dBlockStride = dBlockHeight * dBpl;
    __error__ = ia_aborted;		/* which will be true if we
					 * are forced to stop in the middle */
    for (line = 0; line < h;
	 line += dBlockHeight, pG += gBlockStride, pD += dBlockStride)
    {
	ip_ditherFloydLine4to1 (pIPState, pG, gBpl, pD, dBpl,
				pAGrayBuf, w, dBlockHeight,
				pHDitherErrorTable, pVDitherErrorTable,
				pDitherTruncateTab, pHDRCTable, pLDRCTable,
				pixel);
	IP_YIELD(ditherFloyd4to1Fin)
    }

	/* fix up the bottom frame. */
    if ((h % dBlockHeight) != 0)
	i_rasterOp (pIPState,
		    pD - ((dBlockHeight - (h % dBlockHeight)) * dBpl),
		   				/* destination image */
	            0,				/* x offset */
	            0,				/* y offset */
		    w+31,			/* width of dest rectangle.
						 * Make sure we get to the
						 * next word boundary since
						 * ip_clearRightFrame doesn't
						 * get down here. */
		    dBlockHeight - (h % dBlockHeight), /* height of dest rect */
		    1,				/* depth of dest */
		    dBpl,			/* line bytes */
		    PIX_CLR,
		    NULL,			/* source image */
		    0,				/* x offset */
		    0,				/* y offset */
		    0,				/* depth */
		    0);				/* line bytes */
		  

	/* Clear the right frame bits that have been written by
	 * ip_ditherFloydLine. */
    ip_clearRightFrame (pIPState, pDOrig, w, h, dBpl, 1);
    __error__ = ia_successful;

	/* That's it, free grayBuf and return the dithered pixrect */
ditherFloyd4to1Fin:
    if (pGrayBuf)            FREE(pGrayBuf);
    if (pHDRCTable)          FREE(pHDRCTable);
    if (pHDitherErrorTable)  FREE(pHDitherErrorTable);
    if (pVDitherErrorTable)  FREE(pVDitherErrorTable);
    if (pDitherTruncateTab)  FREE(pDitherTruncateTab);
    NORMAL_RETURN; 
}


/* PUBLIC */
/*****************************************************************************
 * i_getSizeOfDither4to1State()
 *	Set up for banded Floyd-Steinberg dither. 
 *	This routine returns the size of the internal dither state block
 *	so clients can allocate it.
 *	
 * Arguments:
 *		gBpl = bytes/line of gray image.
 *		w = width of source image
 *		pSize = pointer to returned size of private dither state block.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *
 *****************************************************************************/

Int32  CDECL 
i_getSizeOfDither4to1State (
    IPState    *pIPState,
    Int32	gBpl,
    Int32	w,
    Int32      *pSize)
{
    PROLOGUE;
    Int32	grayBufSize;
    

	/* A dither state block is composed of a structure preceeding a
	 * buffer that holds one line of gray pixels.  The size of the
	 * structure is a constant.  Get the size of the gray buffer.
	 */
    __error__ = ip_getSizeOfDither4to1GrayBuf (pIPState, gBpl, w, &grayBufSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* return the real size */
    *pSize = sizeof(DitherState) + grayBufSize;

    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * i_ditherFloyd4To1Init()
 *	Set up for banded Floyd-Steinberg dither of a 4-bit source
 *	
 *	Input:
 *		pG = pointer to source image's frame
 *		gBpl = bytes/line for gray (input) image
 *		x, y = starting (x,y) offset of source rectangle
 *		       relative to outside of frame
 *		w = width of image to dither
 *		pDRCTable = pointer to DRC table
 *		blockHeight = the number of lines in a band, set by client.
 *		pState = pointer to state block to be initialized.  This
 *			 should have the correct number of bytes as
 *			 determined by i_getSizeOfDitherState.
 *		stateSize = number of bytes in state block being passed in.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *
 *****************************************************************************/
Int32  CDECL 
i_ditherFloyd4to1Init (
    IPState    *pIPState,
    UInt8      *pG,
    Int32	gBpl,
    Int32	x,
    Int32	y,
    Int32	w,
    UInt8      *pDRCTable,
    Int32	blockHeight,
    void       *pState,
    Int32	stateSize)
{
    PROLOGUE;
    DitherState *s;
    
    Int32	 ibyte, obyte;
    UInt32	*pVDitherErrorTable;
    UInt32	*pHDitherErrorTable;
    UInt32	*pSingHDitherErrorTable;
    UInt8	*pDitherTruncateTab;
    UInt8	*pGrayBuf;
    Int32	 reqStateSize;

	/* see if passed in state block is big enough */
    __error__ = i_getSizeOfDitherState (pIPState, gBpl, w, &reqStateSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    if (stateSize < reqStateSize)
	return ia_invalidParm;

	/* initialize the internal pointers */
    s = (DitherState *) pState;
    pVDitherErrorTable = &(s->VDitherErrorTable[0]);
    pHDitherErrorTable = &(s->HDitherErrorTable[0]);
    pSingHDitherErrorTable = &(s->SingHDitherErrorTable[0]);
    pDitherTruncateTab = &(s->DitherTruncateTab[0]);
    pGrayBuf = (UInt8 *)s + sizeof(DitherState);

        /* Make the error diffusion tables */
    if ((__error__ = ip_makeErrorDiffusionTables (pIPState, pHDitherErrorTable,
					        pSingHDitherErrorTable,
					        pVDitherErrorTable,
					        pDitherTruncateTab))
	!= ia_successful)
        goto ditherFloyd4to1Err;

    for (ibyte = 0; ibyte < 16; ibyte++)
    {
	    /* first do the drc table for the pixel in the high
	     * nibble */
	for (obyte = ibyte << 4; obyte < (ibyte << 4) + 16; obyte++)
	    s->HDRCTable[obyte] = pDRCTable[ibyte];
	    /* now do the drc table for the pixel in the low
	     * nibble */
	for (obyte = ibyte; obyte < 256; obyte += 16)
	    s->LDRCTable[obyte] = pDRCTable[ibyte];
    }

	/* point to the upper left corner of the rect in gray image.  Note that
	 * x and y include the frame. */
    pG += (y * gBpl) + (x*4/8);

	/* It is unfortunate, but each of the primitives that do the
	 * real dithering work slightly differently.  We use a machine
	 * dependent setup routine to do whatever's necessary to
	 * get them going.
	 */
    __error__ = ip_floyd4To1Setup (pIPState,
				   pHDitherErrorTable, pVDitherErrorTable,
				   pDitherTruncateTab, &s->dBlockHeight,
				   pG, pGrayBuf,
				   &(s->HDRCTable[0]),
				   &(s->LDRCTable[0]), w);
    if (__error__ != ia_successful)
	goto ditherFloyd4to1Err;

        /* Set up how many dithering multiple line calls to do and how many
	 * dithering single line calls to do each Process call */
    s->blocksPer  = blockHeight / s->dBlockHeight;
    s->singlesPer = blockHeight % s->dBlockHeight;

    s->grayBufOffset = ((UInt32)(pGrayBuf + 3) & ~3) - (UInt32)s;

    s->w = w;
    s->depth = 4;

    NORMAL_RETURN;

ditherFloyd4to1Err:
    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * i_ditherFloyd4to1Process
 *
 *	Dither a block of lines. The number of lines dithered will be
 *	dBlockHeight * blocksPer + singlesPer.
 * 
 * Arguments:
 *	src = Source image buffer, pointer to frame
 *	sBpl = source image bytes/line
 *	xOffset = source image X offset relative to outside of frame
 *	yOffset	= source image Y offset relative to outside of frame
 *	pDest = destination buffer, pointer to frame
 *		Output starts at first non-frame pixel.
 *	dBpl = bytes/line in destination buffer
 *	state = state thing returned by the i_ditherFloyd4to1Init
 *
 * Returns:
 *	error status. If aborted, call finish to clean up.
 *****************************************************************************/

Int32 CDECL 
i_ditherFloyd4to1Process (
    IPState    *pIPState,
    UInt8      *src,
    Int32	sBpl,
    Int32	xOffset,
    Int32	yOffset,
    UInt8      *dest,
    Int32	dBpl,
    void       *state)
{
    PROLOGUE;
    Int32 i, dBlockStride, sBlockStride;
    DitherState *s;
    UInt8 *dp;		/* Destination image pointer */

    s = (DitherState *) state;

    /* Move past the frame in the dest buffer */
    dest += (dBpl << cLogFrameBits) + cFrameBytes;
    dp = dest;

    /* Move to the right place in the source buffer */
    src += sBpl * yOffset + (xOffset * 4 / 8);

    dBlockStride = dBpl * s->dBlockHeight;
    sBlockStride = sBpl * s->dBlockHeight;

    /* Do the specified number of multiple line output dithering calls */
    for (i = 0; i < s->blocksPer; i++)
    {
	ip_ditherFloydLine4to1 (pIPState, src, sBpl, dp, dBpl,
			       (UInt8 *) ((UInt32)s + s->grayBufOffset),
			       s->w,
			       s->dBlockHeight,
			       &(s->HDitherErrorTable[0]),
			       &(s->VDitherErrorTable[0]),
			       &(s->DitherTruncateTab[0]),
			       &(s->HDRCTable[0]), &(s->LDRCTable[0]),
			       xOffset & 1);
	IP_YIELD(ditherFloydProcessFin);
	src  += sBlockStride;
	dp += dBlockStride;
    }

    /* Do the specified number of single line output dithering calls */
    for (i = 0; i < s->singlesPer; i++)
    {
	ip_ditherFloydSingleLine4to1 (pIPState, src, sBpl, dp, dBpl,
				      (UInt8 *) ((UInt32)s + s->grayBufOffset),
				      s->w,
				      s->dBlockHeight,
				      &(s->SingHDitherErrorTable[0]),
				      &(s->VDitherErrorTable[0]),
				      &(s->DitherTruncateTab[0]),
				      &(s->HDRCTable[0]), &(s->LDRCTable[0]),
				      xOffset & 1);
	IP_YIELD(ditherFloydProcessFin);
	src  += sBpl;
	dp += dBpl;
    }

    /* Go back and clear the right frame in the destination */
    ip_clearRightFrame (pIPState, dest, s->w, (dp - dest) / dBpl, dBpl, 1);

ditherFloydProcessFin:
    NORMAL_RETURN;
}

/* PUBLIC */
/*****************************************************************************
 * i_ditherFloyd4to1SingleProcess
 *
 *	Dither one line.
 * 
 * Arguments:
 *	src = Source image buffer, pointer to frame
 *	sBpl = source image bytes/line
 *	xOffset = source image X offset relative to outside of frame
 *	yOffset	= source image Y offset relative to outside of frame
 *	pDest = destination buffer, pointer to frame
 *		Output starts at first non-frame pixel.
 *	dBpl = bytes/line in destination buffer
 *	state = state thing returned by the i_ditherFloyd4to1Init
 *
 * Returns:
 *	error status. If aborted, call finish to clean up.
 *****************************************************************************/

Int32 CDECL 
i_ditherFloyd4to1SingleProcess (
    IPState    *pIPState,
    UInt8      *src,
    Int32	sBpl,
    Int32	xOffset,
    Int32	yOffset,
    UInt8      *dest,
    Int32	dBpl,
    void       *state)
{
    PROLOGUE;
    DitherState *s;

    s = (DitherState *) state;

    /* Move past the frame in the dest buffer */
    dest += (dBpl << cLogFrameBits) + cFrameBytes;

    /* Move to the right place in the source buffer */
    src += sBpl * yOffset + (xOffset * 4 / 8);

    ip_ditherFloydSingleLine4to1 (pIPState, src, sBpl, dest, dBpl,
				  (UInt8 *) ((UInt32)s + s->grayBufOffset),
				  s->w,
				  s->dBlockHeight,
				  &(s->SingHDitherErrorTable[0]),
				  &(s->VDitherErrorTable[0]),
				  &(s->DitherTruncateTab[0]),
				  &(s->HDRCTable[0]), &(s->LDRCTable[0]),
				  xOffset & 1);

	/* Clear the right frame bits that have been written by
	 * ip_ditherFloydLine. */
    ip_clearRightFrame (pIPState, dest, s->w, 1, dBpl, 1);

    IP_YIELD(ditherFloydProcessFin);
    src  += sBpl;
    dest += dBpl;
    
ditherFloydProcessFin:
    NORMAL_RETURN;
}

/*****************************************************************************
 * i_ditherFloyd4to1Finish 
 *
 *	We're done dithering (or aborting), clean up any junk we might
 *	have left in the frame, and then free whatever we allocated
 *
 * Input:
 *	pD = pointer to destination image (including frame)
 *	dBpl = destination bytes per line
 *	w = width of the rectangle
 *	h = height of the rectangle
 *	state = state thing returned by the i_ditherFloyd4to1Init
 *
 * Returns:
 *	error status. 
 *****************************************************************************/

Int32 CDECL 
i_ditherFloyd4to1Finish (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    void       *state)
{
    PROLOGUE;
    DitherState *s;

    s = (DitherState *) state;

    pD = pD;
    dBpl = dBpl;
    w = w, h = h;

	/* That's it.  Return the dithered pixrect */
    NORMAL_RETURN;
}



/**************************************************************************
 **********      Routines to Make Error Diffusion Tables      *************
 **************************************************************************/

/* The 2-bit error diffusion tables are the same for all the assembly
 * and C implementations, so we put the common code here for everyone's use.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_make2ErrorDiffusionTables():
 *		Generates hDitherErrorTable2 (7/16 of error) and
 *		vDitherErrorTable2 ({3/16 | 5/16 | 1/16} of error).
 *		Generates pDitherTruncateTab which truncates pixels to
 *		8 bits.
 *		Used when dithering to 2 bits/pixel.
 *
 *	Input:
 *		pVDitherErrorTable2:	pointer to vertical error table
 *		pHDitherErrorTable2:	pointer to horizontal error table
 *		pDitherTruncateTab:	pointer to pixel truncation table
 **************************************************************************/
Int32  CDECL 
ip_make2ErrorDiffusionTables (
    IPState    *pIPState,
    UInt32     *pHDitherErrorTable2,
    UInt32     *pVDitherErrorTable2,
    UInt8      *pDitherTruncateTab)
{
    PROLOGUE;
    Int32	 pix, error;		/* pixel and error val */

	/* check the tables */
    if ((pVDitherErrorTable2 == NULL) || (pHDitherErrorTable2 == NULL) ||
	(pDitherTruncateTab == NULL))
	return ia_invalidParm;

	/* clear all three tables */
    memset (pVDitherErrorTable2, 0, cDitherVErrorTabSize);
    memset (pHDitherErrorTable2, 0, cDitherHErrorTabSize);
    memset (pDitherTruncateTab, 0, cDitherTruncateTabSize);

	/* The 2 bit/pixel tables are a little more complex.  There are
	 * four output levels, hence the inputs are divided into four
	 * sections:
	 * 	x in Range		Basic Error value	Digital Output
	 * [-infinity, 00111111]	MAX(0, x)		  00
	 * [01000000,  01111111]	x - (1*255/3)		  01
	 * [10000000,  10111111]	x - (2*255/3)		  10
	 * [11000000, +infinity]	MIN(x - 255, 0)		  11
	 *
	 * The other difference from the 1 b/p tables is that we need
	 * to have the horiz table compute the digital output.  The
	 * pixel receiving the 7/16 error is the middle one, the thresholded
	 * pixel is still in the top.  We need to put the right number
	 * in this top position to guarantee that bits [31:30] of the
	 * result correctly represent the digital output.  For values in
	 * the range of 0 through 255, we just add correct top two bits
	 * since the corresponding pixel bits are 00.  For pixels in the
	 * range 256-511 (top two bits are 01), we add 10 to get the desired
	 * value of 11.  Pixels in (512..1023) represent negative numbers
	 * so we want the final value to be 00.  If the pixel is in
	 * [512..767] (top bits = 10), we add 10.  If the pixel is in
	 * [768..1023] (top bits = 11), we add 01.
	 */

	/* first, fill in all the terms for which errors are positive. */
    for (pix = 0; pix < 64; pix++)
    {
	error = pix;
	pHDitherErrorTable2[pix] = (0 << 30) |  (((7*error)/16) & 0x3FF) << 11;
	pVDitherErrorTable2[pix] = ((((3*error)/16) & 0x3FF) << 22) |
				   ((((5*error)/16) & 0x3FF) << 11) |
				    (((1*error)/16) & 0x3FF);
    }

    for (pix = 64; pix < 128; pix++)
    {
	error = pix - (255/3);
	pHDitherErrorTable2[pix] = (1 << 30) |  (((7*error)/16) & 0x3FF) << 11;
	pVDitherErrorTable2[pix] = ((((3*error)/16) & 0x3FF) << 22) |
				   ((((5*error)/16) & 0x3FF) << 11) |
				    (((1*error)/16) & 0x3FF);
    }

    for (pix = 128; pix < 192; pix++)
    {
	error = pix - (2*255/3);
	pHDitherErrorTable2[pix] = (2UL << 30) | (((7*error)/16) & 0x3FF) << 11;
	pVDitherErrorTable2[pix] = ((((3*error)/16) & 0x3FF) << 22) |
				   ((((5*error)/16) & 0x3FF) << 11) |
				    (((1*error)/16) & 0x3FF);
    }

    for (pix = 192; pix < 256; pix++)
    {
	error = pix - 255;
	pHDitherErrorTable2[pix] = (3UL << 30) | (((7*error)/16) & 0x3FF) << 11;
	pVDitherErrorTable2[pix] = ((((3*error)/16) & 0x3FF) << 22) |
				   ((((5*error)/16) & 0x3FF) << 11) |
				    (((1*error)/16) & 0x3FF);
    }

	/* here are the "too postive" values.  Note that the error term is
	 * always zero here.  We don't want to adjust surrounding values
	 * because the printer can't print whiter than white.  We need
	 * only set the upper 2 bits of HDitherErrorTable2.
	 */
    for (pix = 256; pix < 512; pix++)
    {
	pHDitherErrorTable2[pix] = (2UL << 30);
    }

	/* here are the negative pixel values.  Note that the error term is
	 * still zero.  We don't want to adjust surrounding values
	 * because the printer can't print blacker than black.  We need
	 * only set the upper 2 bits of HDitherErrorTable2.
	 */
    for (pix = 512; pix < 768; pix++)	/* upper bits = 10 */
    {
	pHDitherErrorTable2[pix] = (2UL << 30);
    }

    for (pix = 768; pix < 1024; pix++)	/* upper bits = 11 */
    {
	pHDitherErrorTable2[pix] = (1UL << 30);
    }


	/* Make the dither truncate table.  This is used to turn 10 bit
	 * pixel values into 8 bit pixel values so we can pack them into
	 * the GrayBuf.  The 10 bit values can be positive or negative.
	 * We truncate them back to the range [0..255].
	 * Start with the pixels that are in the right range.
	 */
    for (pix = 0; pix < 256; pix++)
    {
	pDitherTruncateTab[pix] = (UInt8) pix;
    }

	/* Truncate excessively positive pixels */
    for (pix = 256; pix < 512; pix++)
	pDitherTruncateTab[pix] = (UInt8)255;

	/* The rest of the pixels represent negative values, so leave
	 * them set to 0.
	 */
    NORMAL_RETURN;
}


/*

$Log: i_floyd.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.22  1998/02/25 01:29:47  aber
aber on Tue Feb 24 17:25:37 PST 1998

i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v

Since error diffusion forces errors from on pixel to the surrounding
pixels, this causes data to propogate into the right frame. In banded
mode, the frame was being cleared in the finish routine, but the
image band was not guaranteed to be in pieces of a whole image (as
is the case that Tim Platt is encountering). So I've moved the
frame clearing from the finish routine, into the band processing
routine; for both 1-bit and 4-bit.

Revision 1.21  1998/01/13 23:47:50  ddavies
i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.20  1997/05/02 01:19:57  ddavies
i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v

Change comments to specify that the rectangles are relative to the outside
of the source image frames and that the output always goes to the first
non-frame pixel of the dest image.

Revision 1.19  1997/04/23 21:08:41  ddavies
ddavies on Wed Apr 23 13:41:44 PDT 1997

i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v

Lots of changes to do with changing DitherState to be a structure that's
allocated by the client and now contains no pointers.
Include ansiprot.h to get proto for memset.
Fix lots of indentation to be new style.
In i_ditherFloyd, allocate error diffusion and truncation tables instead
of letting ip_makeErrorDiffusionTable do it.  Also allocate grayBuf
instead of letting ip_floydSetup do it.
Continue to free all these at this level.  It's neater.
Add i_getSizeOfDitherState and i_getSizeOfDither4to1State.  These are
used by a client to get the size of the DitherState structure used
in banded dithering.
i_ditherFloydInit now takes a pointer to the pre-allocated state, and its
size, instead of allocating the state internally (was using MALLOC!).
Ditto for i_ditherFloyd4to1Init.
Lots of changes in both Init routines to account for the fact that the
DitherState structure contains all the tables and buffers directly
instead of using pointers to other blocks of memory.
Corresponding changes in the xxxProcess routines when passing pointers
to tables.
To be consistent, do the same stuff for i_ditherFloyd2 - it allocates and
frees all tables instead of letting ip_make2ErrorDiffusionTables and
ip_floyd2Setup allocate them.
Move all the 4to1 Init, Process and Finish routines to before the section
of the file that says it includes Routines to Make Error Diffusion Tables.
Make ip_make2ErrorDiffusionTables take pre-allocated tables.  It clears
and fill them with the right stuff.  It no longer allocates them.

Revision 1.18  1995/12/13 02:43:52  ejaquith
i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.17  1995/10/18  03:53:01  ddavies
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.16  1995/10/13  23:50:30  dferrell
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.15  1994/11/07  22:05:23  ddavies
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.14  1994/08/31  19:02:39  ejaquith
 * ejaquith on Wed Aug 31 12:01:38 PDT 1994
 *
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Fixed the two free-already-freed memory errors in i_ditherFloydFinish
 * and i_ditherFloyd4to1Finish.
 *
 * Revision 1.13  1994/08/29  17:09:17  ejaquith
 * ejaquith on Mon Aug 29 10:07:53 PDT 1994
 *
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Made i_ditherFloydSingleProcess and i_ditherFloyd4to1SingleProcess
 * primitives, which dither one line.  Useful for dithering last lines of an
 * image when there are less than the banded dithering's block height lines
 * left.
 *
 * Revision 1.12  1994/08/25  17:26:16  ejaquith
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Changed calls to ditherFloydLine routines to include the dBlockHeight value
 * as a parameter.  Added calls to ip_clearRightFrame in i_ditherFloydFinish
 * and i_ditherFloyd4to1Finish (and added the needed parameters) to clean
 * the right frame out from potential cases of dithering writing into it.
 *
 * Revision 1.11  1994/08/10  21:47:52  ejaquith
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Removed five unused fields in DitherState struct.  Removed an unused
 * parameter from some private functions.
 *
 * Revision 1.10  1994/08/09  17:20:11  ejaquith
 * ejaquith on Tue Aug 9 10:19:53 PDT 1994
 *
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Added flexibility to i_ditherFloydProcess and i_ditherFloyd4to1Process by
 * allowing the number of lines dithered in Process to be completely arbitrary.
 * Did this by adding new field, singlesPer to ditherState struct and calls
 * to ip_ditherFloydSingleLine and ip_ditherFloydSingleLine4to1 for excess
 * lines over full multiples of dBlockHeight.
 *
 * Revision 1.9  1994/07/27  15:36:18  ejaquith
 * ejaquith on Wed Jul 27 08:31:36 PDT 1994
 *
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Made all pointers to error diffusion tables local pointers.
 * Added neccessary checks for freeing dithering structure elements.
 * Changed ip_make2ErrorDiffusionTables to modify passed in table pointers.
 *
 * Revision 1.8  1994/07/11  23:05:27  ddavies
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * When clearing the bottom frame, clear out to the next word boundary on the
 * right since ip_clearRightFrame doesn't go down to the bottom frame.
 *
 * Revision 1.7  1994/04/29  01:43:28  macleod
 * macleod on Thu Apr 28 18:42:59 PDT 1994
 *
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Made a change to banded dithering to support brain-dead usage.
 *
 * Revision 1.6  1994/04/19  00:21:32  macleod
 * macleod on Mon Apr 18 17:20:20 PDT 1994
 *
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Changed the banded dithering code so that the caller passes the source buffer
 * on each "process" routine, so that one doesn't need to keep the entire source
 * image in memory. The consequence is that the caller needs to keep track of the
 * current offset into the image.
 *
 * Revision 1.5  1994/04/13  23:45:35  macleod
 * macleod on Wed Apr 13 16:41:59 PDT 1994
 *
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Added support for band-at-a-time dithering of 4 or 8 bit images to
 * 1 bit results. For both source types, there is an "init", a "process"
 * and a "finish" call. The "init" call passes back a state structure.
 * The "process" call takes a state structure and a buffer and dithers
 * the "current" band into it. The "finish" call frees resources.
 * Note that when one calls "init", one requests some number of lines
 * to process per band. The number given back may not be the same, due
 * to the fact that some of our dithering code processes more than one
 * line at a time (PC Fast: 1 Sparc Fast: 7 "C": 3) and we decided to
 * avoid strange buffering schemes by always rounding down to the nearest
 * multiple of the lines-at-once constant.
 *
 * Revision 1.4  1994/04/05  02:52:34  ddavies
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Fix bug detected by gcc where a table could have been freed before being allocated
 * as a result of another error. The table pointer is pHDRCTable.
 *
 * Revision 1.3  1994/03/02  03:50:30  ddavies
 * ddavies on Tue Mar 1 19:50:11 PST 1994
 *
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Add IP_YIELD to give up control during dithering.
 *
 * Revision 1.2  1994/02/04  20:15:15  dferrell
 * i_floyd.c -> /product/ipcore/ipshared/src/RCS/i_floyd.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/12  01:03:54  ddavies
 * Initial revision
 *
 * Revision 1.18  1993/12/02  19:21:07  ddavies
 * ffloyd.c -> /product/ipcore/ipshared/src/RCS/ffloyd.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.17  1993/11/05  01:24:36  ddavies
 * ffloyd.c -> /product/ipcore/ipshared/src/RCS/ffloyd.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

