/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* 
 *  i_frame.c:  upper level of frame routines:
 *	   Int32    i_addFrame()
 *	   Int32    i_addFrameInPlace()
 *	   Int32    i_removeFrame()
 *	   Int32    i_removeFrameInPlace()
 *	   Int32    ip_clearRightFrame()
 *	   Int32    i_clearRightBottomFrames()
 *	   Int32    i_clearAllFrames()
 *
 *	This is also a good place to put two routines that are used
 *	to calculate the number of bytes/line needed for an image
 *	and the number of total bytes needed for an image.  The
 *	second routine calculates the offset of the image within the
 *	memory buffer.
 *	   Int32    i_calcImageProperties()
 *	   Int32    i_calcImageOffset()
 */


#include <stddef.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "lmask.prv"

    /* procedure declarations */
    /* prototypes */
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shrrast.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_frame.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/*********************************************************************
 *                         add and remove frame                      *
 ********************************************************************/

/* PUBLIC */
/********************************************************************
 *  i_addFrame()  
 *	Input:  
 *		pS =	pointer to source image (without frame)		
 *		ws =	the width in pixels of the image (before frame)
 *		hs =	the height of the image (before frame)
 *		depth =	depth of source image.  must be 1, 2, 4, or 8
 *		sBpl =	bytes/line (before frame)
 *	        pD =	destination image (Image+frame)
 *
 *	On return: 
 *		dest image has the source image with cFrameBits of  
 *		frame added on all 4 sides.
 *
 *	The width of the source image is not assumed to be a multiple
 *	of 32 bits.
 *
 *	Return code:
 *		ia_successful	successful completion
 *
 ********************************************************************/
Int32   CDECL 
i_addFrame (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	ws,
    Int32	hs,
    Int32	depth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    Int32	line;
    Int32	frameInBytes;
    Int32	imageInBytes;
    UInt32	rightWordMask;

	/* We assume the image is actually big enough to hold the image
	 * plus frame.  The image is crammed into the top portion
	 * of the image.  We copy it down to the bottom, fix the width
	 * and height, then clear out the frame.
	 */

    if (sBpl == 0) NORMAL_RETURN;

	/* sBpl is NOT known to be a multiple of 4.
	 * cFrameBytes is a multiple of 4.
	 * Set the number of bytes we'll copy in each line.
	 * The copying routines know to round this up to a multiple
	 * of four while copying, but they need to know the real
	 * number of bytes so they can clear the right frame of each
	 * line properly.
	 * If we copy too much, we'll clear out the excess when we
	 * clear the right frame.
	 */
    frameInBytes = (cFrameBytes * depth);
    imageInBytes = ((ws * depth) + 7) / 8;

	/* Calculate the mask used to clear out all the bits of the
	 * right frame that are in the rightmost word.  If the width
	 * is a multiple of 32, there are no frame bits here.  If the
	 * width%32 = 1, there are 31 frame bits and so on.
	 */
    rightWordMask = leftMask[((ws*depth)-1) % 32];

	/* copy each line down into position starting with the bottom line.
	 * The starting position for the destination accounts for the number
	 * of upper pad lines (cFrameBits), the number of image lines below
	 * the top of the image (hs-1) and the width of the left frame
	 * (frameInBytes).  We also clear out the left and right frames
	 * for each line copied.
	 * Note that ip_addFrameToLine must account for the fact that
	 * the source may not be word aligned since sBpl may not be
	 * a multiple of 4.  We can't just use rasterOp since it operates
	 * from top to bottom and we need to copy from bottom to top.
	 * We can't just use memcpy because it does the wrong thing on
	 * little endian machines when starting or stopping on non-word
	 * boundaries.  That's why we have our own routine.
	 * ip_addFrameToLine also clears out the frame on either side
	 * of the destination line.
	 */
    pS += (hs-1)*sBpl;
    pD += (cFrameBits + (hs-1)) * dBpl;

    for (line = hs; line > 0;
	 pS -= sBpl, pD -= dBpl, line--)
    {
	ip_addFrameToLine (pIPState, pD, pS,
			   imageInBytes, frameInBytes, rightWordMask);
    }

	/* We've moved all the data into place and cleared the right
	 * and left frames adjacent to the data.  Clear out the
	 * top and bottom frames too.
	 * pD points to the top of the image (minus one line).
	 */
    pD -= (cFrameBits - 1) * dBpl;
    i_rasterOp (pIPState, pD,	/* destination image */
				/* dest is pS + sBpl - cause the loop
                                above subtracts one line extra */
		0,		/* x offset */
		0,		/* y offset */
		ws+2*cFrameBits, /* rectangle width */
		cFrameBits,	/* rectangle height */
		depth,
		dBpl,		/* line bytes */
		PIX_CLR,
		NULL, 0, 0, 0, 0);
    i_rasterOp (pIPState, pD,	/* destination image */
				/* dest is pS + sBpl - cause the loop
                                 * above subtracts one line extra */
		0,		/* x offset */
		hs+cFrameBits,	/* y offset */

		ws+2*cFrameBits, /* rectangle width */
		cFrameBits,	/* rectangle height */
		depth,
		dBpl,		/* line bytes */
		PIX_CLR,
		NULL, 0, 0, 0, 0);


    NORMAL_RETURN;
}


/* PUBLIC */
/********************************************************************
 *  i_addFrameInPlace()
 *	Input:
 *		sPtr =	pointer to source image that has no frame bits
 *			but has enough allocated to hold image+frame	
 *		sW =	the width in pixels of the image (before frame)
 *		sH =	the height of the image (before frame)
 *		depth =	depth of source image.  must be 1, 2, 4, or 8
 *		sBpl =	bytes/line (before frame)
 *
 *      On return:
 *		the image has cFrameBits of frame added on all four sides
 *
 *	Return code:
 *		ia_successful	successful completion
 *
 ********************************************************************/
Int32   CDECL 
i_addFrameInPlace (
    IPState    *pIPState,
    UInt8      *sPtr,
    Int32	sW,
    Int32	sH,
    Int32	depth,
    Int32	sBpl) 
{
    PROLOGUE;
    Int32	dBpl, line;
    Int32	frameInBytes;
    Int32	imageInBytes;
    UInt8      *dPtr;
    UInt32	rightWordMask;


	/* We assume the image is actually big enough to hold the image
	 * plus frame.  The image is crammed into the top portion
	 * of the image.  We copy it down to the bottom, fix the width
	 * and height, then clear out the frame.
	 */
    if (sBpl == 0) NORMAL_RETURN;

	/* Calculate the framed images's bytes/line by using
	 * the legal i_calcImageProperties routine instead of
	 * some ad hoc calculation here.
	 */
    frameInBytes = (cFrameBytes * depth);
    imageInBytes = ((sW * depth) + 7) / 8;
    if (i_calcImageProperties (pIPState, sW, sH,
			       depth, cFrameBits, &dBpl, &line) !=
	ia_successful)
	return ia_internal;

	/* initialize the source and destination pointers.  We assume
	 * that clients always start a frameless image at the same
	 * offset as a framed image, so dPtr == sPtr
	 */
    dPtr = sPtr;

	/* Calculate the mask used to clear out all the bits of the
	 * right frame that are in the rightmost word.  If the width
	 * is a multiple of 32, there are no frame bits here.  If the
	 * width%32 = 1, there are 31 frame bits and so on.
	 */
    rightWordMask = leftMask[((sW*depth)-1) % 32];

	/* copy each line down into position starting with the bottom line.
	 * The starting position for the destination accounts for the number
	 * of upper pad lines (cFrameBits), the number of image lines below
	 * the top of the image (hs-1) and the width of the left frame
	 * (frameInBytes).  We also clear out the left and right frames
	 * for each line copied.
	 * Note that ip_addFrameToLine must account for the fact that
	 * the source may not be word aligned since sBpl may not be
	 * a multiple of 4.  We can't just use rasterOp since it operates
	 * from top to bottom and we need to copy from bottom to top.
	 * We can't just use memcpy because it does the wrong thing on
	 * little endian machines when starting or stopping on non-word
	 * boundaries.  That's why we have our own routine.
	 * ip_addFrameToLine also clears out the frame on either side
	 * of the destination line.
	 */
    sPtr += (sH-1)*sBpl;
    dPtr += (cFrameBits + (sH-1)) * dBpl;

    for (line = sH; line > 0;
	 sPtr -= sBpl, dPtr -= dBpl, line--)
    {
	ip_addFrameToLine (pIPState, dPtr, sPtr,
			   imageInBytes, frameInBytes, rightWordMask);
    }

	/* We've moved all the data into place and cleared the right
	 * and left frames adjacent to the data.  Clear out the
	 * top and bottom frames too.  We use sPtr here because it
	 * point to the top of the top frame (minus one line).
	 * dPtr points to the top of the image (minus one line).
	 */
    i_rasterOp (pIPState, sPtr + sBpl,	/* destination image */
				/* dest is sPtr + sBpl - cause the loop
                                 * above subtracts one line extra */
		0,		/* x offset */
		0,		/* y offset */
		sW+2*cFrameBits, /* rectangle width */
		cFrameBits,	/* rectangle height */
		depth,
		dBpl,		/* line bytes */
		PIX_CLR,
		NULL, 0, 0, 0, 0);
    i_rasterOp (pIPState, sPtr + sBpl,	/* destination image */
				/* dest is sPtr + sBpl - cause the loop
                                 * above subtracts one line extra */
		0,		/* x offset */
		sH+cFrameBits,	/* y offset */

		sW+2*cFrameBits, /* rectangle width */
		cFrameBits,	/* rectangle height */
		depth,
		dBpl,		/* line bytes */
		PIX_CLR,
		NULL, 0, 0, 0, 0);
    NORMAL_RETURN;
}

/* PUBLIC */
/********************************************************************
 *  i_removeFrame()
 *	Input:
 *      	pS = 	source image (that has Frame Bits)
 *		ws =	width in pixels of the image (including frame)
 *		hs =	the height of the image (including frame)
 *		depth =	bits/pixel. must be 1, 2, 4, or 8.
 *		sBpl =	source bytes/line (including frame)
 *		pD =	destination image pointer
 *		dBpl =	dest bytes/line (not including frame)
 *
 *	On return: 
 *		the dest image has the source image with cFrameBits
 *		of pad removed on each side.
 *
 *	Return code:
 *		ia_successful	successful completion
 *
 ********************************************************************/
Int32   CDECL 
i_removeFrame (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	ws,
    Int32	hs,
    Int32	depth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    Int32	 wd, hd;
    Int32	 frameInBytes;
    Int32	 imageInBytes;

    wd = MAX(1, ws - 2 * cFrameBits);  /* don't let it get smaller than 1 */
    hd = MAX(1, hs - 2 * cFrameBits);  /*    ditto  */

	/* initialize the source and destination pointers */
    if (sBpl == 0) NORMAL_RETURN;

	/* sBpl is known to be a multiple of 4 since all images are padded
	 * to 32-bit boundaries.  cFrameBytes is also a multiple of 4.
	 * Therefore, frameInBytes is a multiple of 4.
	 * However, it appears we must deal with destination images whose
	 * pitches are NOT multiples of 4.  We calculate the number of
	 * bytes of image data to transfer anyway..
	 */
    frameInBytes = (cFrameBytes * depth);
    imageInBytes = ((wd * depth) + 7) / 8;

	/* copy each line up into position starting with the topmost image line.
	 * The starting position for the source accounts for the number
	 * of upper pad lines (cFrameBits) and the width of the left frame
	 * (frameInBytes).
	 */

    pS += (sBpl << cLogFrameBits) + frameInBytes;
    for (; hd > 0; pS += sBpl, pD += dBpl, hd--)
    {
	ip_removeFrameFromLine (pIPState, pD, pS, imageInBytes);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/********************************************************************
 *  i_removeFrameInPlace()
 *	Input:
 *      	sPtr = 	pointer to source image (includes frame)
 *		sW =	the width in pixels of the image (includes frame)
 *		sH =	the height of the image (includes frame)
 *		depth =	bits/pixel of the image.  must be 1, 2, 4, or 8.
 *		sBpl =	bytes/line (includes frame)
 *		dBpl =  destination bytes/line (not necessarily a multiple of
 *			2 or 4).
 *
 *	On return:
 *		The image has cFrameBits of frame removed
 *              on each side.  The image bits are crammed into 
 *		the top portion.  Each resulting image line has dBpl
 *		bytes/line.  The frameless image lines are padded to
 *		8, 16 or 32-bit boundaries as necessary.  Longer pads
 *		are not supplied.
 *
 *	Return code:
 *		ia_successful	successful completion
 *
 ********************************************************************/
Int32   CDECL 
i_removeFrameInPlace (
    IPState    *pIPState,
    UInt8      *sPtr,
    Int32	sW,
    Int32	sH,
    Int32	depth,
    Int32	sBpl,
    Int32	dBpl)
{
    PROLOGUE;
    Int32	imageInBytes;
    UInt8      *dPtr;

	/* initialize the source and destination pointers */
    if (sBpl == 0) NORMAL_RETURN;
    dPtr = sPtr;

	/* sBpl is known to be a multiple of 4 since all images are padded
	 * to 32-bit boundaries.  cFrameBytes is also a multiple of 4.
	 * Therefore, frameInBytes is a multiple of 4.
	 * However, it appears we must deal with destination images whose
	 * pitches are NOT multiples of 4.  We calculate the number of
	 * bytes of image data to transfer anyway..
	 */
    imageInBytes = (((sW - 2*cFrameBits) * depth) + 7) / 8;

	/* copy each line up into position starting with the topmost image line.
	 * The starting position for the source accounts for the number
	 * of upper pad lines (cFrameBits) and the width of the left frame
	 * (frameInBytes).
	 */

    sPtr += (sBpl << cLogFrameBits) + (cFrameBytes * depth);
    for (sH = sH - 2*cFrameBits;
	 sH > 0;
	 sPtr += sBpl, dPtr += dBpl, sH--)
    {
	ip_removeFrameFromLine (pIPState, dPtr, sPtr, imageInBytes);
    }
    NORMAL_RETURN;
}


/* The following routine is called after we have mistakenly written trash
 * into the right frame.  Several routines that operate only on 32-bit words,
 * notably the dilation code, do so.  The SPARC version of this routine tests
 * to see if there are any bits in each word of the right frame, erasing them
 * if so.  The test is there because write operations are so slow.  It's also
 * a good idea on the 386/486 to keep the write buffers from filling up.
 * This routine clears the first 32 bit word of the right frame independent
 * of the depth of the image.
 */

static UInt32 rightFrameMask[32] =
    {0x80000000, 0xC0000000, 0xE0000000, 0xF0000000,
     0xF8000000, 0xFC000000, 0xFE000000, 0xFF000000,
     0xFF800000, 0xFFC00000, 0xFFE00000, 0xFFF00000,
     0xFFF80000, 0xFFFC0000, 0xFFFE0000, 0xFFFF0000,
     0xFFFF8000, 0xFFFFC000, 0xFFFFE000, 0xFFFFF000,
     0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00, 0xFFFFFF00,
     0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0, 0xFFFFFFF0,
     0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE, 0xFFFFFFFF};

/* PRIVATE */
/****************************************************************************
 *  ip_clearRightFrame()
 *	Input:
 *      	sPtr = 	pointer to dest image (inside frame)
 *		sW =	the width in pixels of the image (not including frame)
 *		sH =	the height of the image (not including frame)
 *		sBpl =	bytes/line (includes frame)
 *		depth = number of bits/pixel.  Used to find rightmost
 *			image bit.
 *
 *	On return:
 *		Zeros have been put into the right frame of the image.
 *
 *	Return code:
 *		ia_successful	successful completion
 *		ia_invalidParm	depth had illegal value
 *
 ****************************************************************************/
Int32   CDECL 
ip_clearRightFrame (
    IPState    *pIPState,
    register UInt8 *sPtr,
    Int32	sW,
    Int32	sH,
    register Int32 sBpl,
    Int32	depth)
{
    PROLOGUE;
    UInt32	rightWordOffset, rightmostBit;
    UInt32	imageMask, frameMask;

	/* Get offset of rightmost word containing some image.  From this,
	 * get number of rightmost bit containing image bits.  */
    switch (depth) {
    case 1:
	rightWordOffset = (sW-1) >> 5;		/* 32 pixels/word */
	rightmostBit = (sW-1) - (rightWordOffset << 5);
	break;
    case 2:
	rightWordOffset = (sW-1) >> 4;		/* 16 pixels/word */
	rightmostBit = ((sW-1) - (rightWordOffset << 4)) * 2 + 1;
	break;
    case 4:
	rightWordOffset = (sW-1) >> 3;		/* 8 pixels/word */
	rightmostBit = ((sW-1) - (rightWordOffset << 3)) * 4 + 3;
	break;
    case 8:
	rightWordOffset = (sW-1) >> 2;		/* 4 pixels/word */
	rightmostBit = ((sW-1) - (rightWordOffset << 2)) * 8 + 7;
	break;
    case 16:
	rightWordOffset = (sW-1) >> 1;		/* 2 pixels/word */
	rightmostBit = ((sW-1) - (rightWordOffset << 1)) * 16 + 15;
    default:
	return ia_invalidParm;
    }

    if (rightmostBit == 31)	/* rightmost word contained only image */
	NORMAL_RETURN;

	/* else, we need to clear some bits.  Set up the masks for the
	 * image and the frame */
    imageMask = rightFrameMask[rightmostBit];
    frameMask = ~imageMask;

	/* point to the rightmost word */
    sPtr += (rightWordOffset << 2);

	/* we're ready to clear the frame.  Test the frame portion of each
	 * word.  If it has bits in it, clear them.  This assembly routine
	 * does the real work. */
    ip_clearRightFrameBlock (pIPState, sPtr, sBpl, sH, imageMask, frameMask);

    NORMAL_RETURN;
}

/* PUBLIC */
/****************************************************************************
 *  i_clearAllFrames()
 *	Input:
 *      	sPtr = 	pointer to dest image (outside frame)
 *		sW =	the width in pixels of the image (including frame)
 *		sH =	the height of the image (including frame)
 *		sBpl =	bytes/line (includes frame)
 *		depth = number of bits/pixel.  Used to find rightmost
 *			image bit.
 *
 *	On return:
 *		Zeros have been put into the entire frame of the image.
 *
 *	Return code:
 *		ia_successful	successful completion
 *
 ****************************************************************************/
Int32   CDECL 
i_clearAllFrames (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	depth)
{
	/* clear top frame */
    PROLOGUE;
    i_rasterOp (pIPState, pS,	/* dest ptr */
		0,		/* dest x */
		0,		/* dest y */
		wS,		/* width of dest in pixels */
		cFrameBits,	/* height of dest */
		depth,		/* dest depth */
		sBpl,		/* dest bpl */
		PIX_CLR,	/* operation */
		NULL,		/* src pointer */
		0, 0,		/* src x, src y */
		0,		/* src depth */
		0);		/* src bytes/line */

	/* clear left frame */
    i_rasterOp (pIPState, pS,	/* dest ptr */
		0,		/* dest x */
		cFrameBits,	/* dest y */
		cFrameBits,	/* width of dest in pixels */
		hS-2*cFrameBits, /* height of dest */
		depth,		/* dest depth */
		sBpl,		/* dest bpl */
		PIX_CLR,	/* operation */
		NULL,		/* src pointer */
		0, 0,		/* src x, src y */
		0,		/* src depth */
		0);		/* src bytes/line */

	/* clear right frame */
    i_rasterOp (pIPState, pS,	/* dest ptr */
		wS-cFrameBits,	/* dest x */
		cFrameBits,	/* dest y */
		cFrameBits,	/* width of dest in pixels */
		hS-2*cFrameBits, /* height of dest */
		depth,		/* dest depth */
		sBpl,		/* dest bpl */
		PIX_CLR,	/* operation */
		NULL,		/* src pointer */
		0, 0,		/* src x, src y */
		0,		/* src depth */
		0);		/* src bytes/line */

	/* clear bottom frame */
    i_rasterOp (pIPState, pS,	/* dest ptr */
		0,		/* dest x */
		hS-cFrameBits,	/* dest y */
		wS,		/* width of dest in pixels */
		cFrameBits,	/* height of dest */
		depth,		/* dest depth */
		sBpl,		/* dest bpl */
		PIX_CLR,	/* operation */
		NULL,		/* src pointer */
		0, 0,		/* src x, src y */
		0,		/* src depth */
		0);		/* src bytes/line */

    NORMAL_RETURN;
}


/* PUBLIC */
/****************************************************************************
 *  i_clearRightBottomFrames()
 *	Input:
 *      	sPtr = 		pointer to dest image (outside frame)
 *		sW =		the width in pixels of the image
 *				(including frame)
 *		sH =		the height of the image (including frame)
 *		sBpl =		bytes/line (includes frame)
 *		depth =		number of bits/pixel.  Used to find rightmost
 *				image bit.
 *		rightCols =	number of columns to clear in right frame.
 *		bottomRows =	number of rows to clear in bottom frame.
 *
 *	On return:
 *		Zeros have been put into the right and bottom frame of
 *		the image.
 *
 *	Return code:
 *		ia_successful	successful completion
 *
 ****************************************************************************/
Int32   CDECL 
i_clearRightBottomFrames (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	depth,
    Int32	rightCols,
    Int32	bottomRows)
{
    PROLOGUE;
    UInt8	*pInsideFrame;

	/* get a pointer to the inside of the image frame.  It's
	 * used by ip_clearRightFrame.
	 */
    pInsideFrame = pS + (sBpl << cLogFrameBits) + depth * cFrameBytes;

	/* if the right cols can be cleared with ip_clearRightFrame,
	 * use that.
	 */
    switch (depth)
    {
    case 1:
	if (((rightCols + wS)/32) == (wS / 32))
	    ip_clearRightFrame (pIPState, pInsideFrame,
				wS-2*cFrameBits, hS-2*cFrameBits,
			        sBpl, 1);
	else
	{
		/* cleared area crosses word boundaries right frame */
	    i_rasterOp (pIPState, pS,	/* dest ptr */
			wS-cFrameBits,	/* dest x */
			cFrameBits,	/* dest y */
			rightCols,	/* width of dest in pixels */
			hS-2*cFrameBits, /* height of dest */
			1,		/* dest depth */
			sBpl,		/* dest bpl */
			PIX_CLR,	/* operation */
			NULL,		/* src pointer */
			0, 0,		/* src x, src y */
			0,		/* src depth */
			0);		/* src bytes/line */
	}
	break;

    case 2:
	if (((rightCols + wS)/16) == (wS / 16))
	    ip_clearRightFrame (pIPState, pInsideFrame,
				wS-2*cFrameBits, hS-2*cFrameBits,
			        sBpl, 2);
	else
	{
		/* cleared area crosses word boundaries right frame */
	    i_rasterOp (pIPState, pS,	/* dest ptr */
			wS-cFrameBits,	/* dest x */
			cFrameBits,	/* dest y */
			rightCols,	/* width of dest in pixels */
			hS-2*cFrameBits, /* height of dest */
			2,		/* dest depth */
			sBpl,		/* dest bpl */
			PIX_CLR,	/* operation */
			NULL,		/* src pointer */
			0, 0,		/* src x, src y */
			0,		/* src depth */
			0);		/* src bytes/line */
	}
	break;

    case 4:
	if (((rightCols + wS)/8) == (wS / 8))
	    ip_clearRightFrame (pIPState, pInsideFrame,
				wS-2*cFrameBits, hS-2*cFrameBits,
			        sBpl, 4);
	else
	{
		/* cleared area crosses word boundaries right frame */
	    i_rasterOp (pIPState, pS,	/* dest ptr */
			wS-cFrameBits,	/* dest x */
			cFrameBits,	/* dest y */
			rightCols,	/* width of dest in pixels */
			hS-2*cFrameBits, /* height of dest */
			4,		/* dest depth */
			sBpl,		/* dest bpl */
			PIX_CLR,	/* operation */
			NULL,		/* src pointer */
			0, 0,		/* src x, src y */
			0,		/* src depth */
			0);		/* src bytes/line */
	}
	break;

    case 8:
	if (((rightCols + wS)/4) == (wS / 4))
	    ip_clearRightFrame (pIPState, pInsideFrame,
				wS-2*cFrameBits, hS-2*cFrameBits,
				sBpl, 8);
	else
	{
		/* cleared area crosses word boundaries right frame */
	    i_rasterOp (pIPState, pS,	/* dest ptr */
			wS-cFrameBits,	/* dest x */
			cFrameBits,	/* dest y */
			rightCols,	/* width of dest in pixels */
			hS-2*cFrameBits,/* height of dest */
			8,		/* dest depth */
			sBpl,		/* dest bpl */
			PIX_CLR,	/* operation */
			NULL,		/* src pointer */
			0, 0,		/* src x, src y */
			0,		/* src depth */
			0);		/* src bytes/line */
	}
	break;
    case 16:
	if (((rightCols + wS)/2) == (wS / 2))
	    ip_clearRightFrame (pIPState, pInsideFrame,
				wS-2*cFrameBits, hS-2*cFrameBits,
				sBpl, 16);
	else
	{
		/* cleared area crosses word boundaries right frame */
	    i_rasterOp (pIPState, pS,	/* dest ptr */
			wS-cFrameBits,	/* dest x */
			cFrameBits,	/* dest y */
			rightCols,	/* width of dest in pixels */
			hS-2*cFrameBits,/* height of dest */
			16,		/* dest depth */
			sBpl,		/* dest bpl */
			PIX_CLR,	/* operation */
			NULL,		/* src pointer */
			0, 0,		/* src x, src y */
			0,		/* src depth */
			0);		/* src bytes/line */
	}
	break;
    default:
	return ia_invalidParm;
    }

	/* clear bottom frame */
    i_rasterOp (pIPState, pS,	/* dest ptr */
		0,		/* dest x */
		hS-cFrameBits,	/* dest y */
		wS,		/* width of dest in pixels */
		bottomRows,	/* height of dest */
		depth,		/* dest depth */
		sBpl,		/* dest bpl */
		PIX_CLR,	/* operation */
		NULL,		/* src pointer */
		0, 0,		/* src x, src y */
		0,		/* src depth */
		0);		/* src bytes/line */

    NORMAL_RETURN;
}


/* PUBLIC */
/********************************************************************
 *  i_calcImageProperties()  
 *	Input:
 *		width =		the width in pixels of the image (before frame)
 *		height =	the height of the image (before frame)
 *		depth =		depth of source image.  must be 1, 2, 4, or 8
 *		frameWidth = 	width of frame in pixels (almost always equals
 *				cFrameBits)
 *		pBpl =		ptr to returned bytes per line of image
 *		pBufSize = 	ptr to returned image buffer size
 *
 *	On return: 
 *		Image bytes/line, buffer size and the offset from the
 *		beginning of the image buffer (*pBufSize) to the
 *		upper left pixel of the image.  This last quantity
 *		is not necessarily pBuffer + frameWidthInBytes +
 *		frameHeight * bpl because alignment issues may cause
 *		us to move the image over in the buffer a little bit.
 *
 *	Return code:
 *		ia_successful	successful completion
 *		ia_invalidParm	one of the pointers is NULL.
 *
 ********************************************************************/
Int32   CDECL 
i_calcImageProperties (
    IPState    *pIPState,
    Int32	width,
    Int32	height,
    Int32	depth,
    Int32	frameWidth,
    Int32      *pBpl,
    Int32      *pBufSize)
{
    PROLOGUE;
    Int32	 bpl;
    Int32	 fullHeight;	/* real height of image.  Accounts for
				 * padding on bottom that's needed if we
				 * do rotation in place.
				 */
    Int32	 accessLen;	/* length of longest access on this
				 * processor */

	/* Complain if any of the pointers are NULL */
    if ((pBpl == NULL) || (pBufSize == NULL))
	return ia_invalidParm;

	/* Figure out the longest access for this processor */
    ip_getLongestAccess (pIPState,  &accessLen);

	/* Calculate the bytes/line.  The first clause calculates the
	 * number of bytes in enough 32-bit words to hold all the
	 * image data.  The second clause calculates the number of
	 * bytes needed to hold both the left and right frames.
	 */
    if (accessLen == 4)
    {
	bpl = (((width * depth + 31) / 32) * 4) +
	      ((2 * frameWidth * depth) / 8);

	    /* Calculate the real height of the image.  This includes any
	     * padding on the bottom that will be needed if we rotate
	     * the image by 90 degrees in place.
	     */
	fullHeight = ((((height * depth) + 31) & ~0x1F) / depth) +
			2 * frameWidth;
    }
    else
    if (accessLen == 8)
    {
	bpl = (((width * depth + 63) / 64) * 8) +
	      ((2 * frameWidth * depth) / 8);

	    /* Calculate the real height of the image.  This includes any
	     * padding on the bottom that will be needed if we rotate
	     * the image by 90 degrees in place.
	     */
	fullHeight = ((((height * depth) + 63) & ~0x3F) / depth) +
			2 * frameWidth;
    }
    else
	    /* If it wasn't one of these, we're in trouble */
	return ia_internal;

	/* It all worked.  Update the return vals and be done.  Note that
	 * We always add an extra 4 bytes so that we can adjust the position
	 * of the image by 4 bytes to guarantee that the leftmost byte of
	 * the real image (inside the frame) is always on an 8 byte
	 * boundary.
	 */
    *pBpl = bpl;
    *pBufSize = (bpl * fullHeight) + 4;

    NORMAL_RETURN;
}

/* PUBLIC */
/********************************************************************
 *  i_calcImageOffset()  
 *	Input:
 *		pImageBuf =	pointer to buffer containing image.
 *		bpl =		image's bytes/line
 *		depth =		depth of source image.  must be 1, 2, 4, or 8
 *		frameWidth = 	width of frame in pixels (almost always equals
 *				cFrameBits)
 *		pOffsetTo00 =	ptr to returned offset from beginning of
 *				image buffer to location (0, 0) in bytes.
 *
 *	On return: 
 *		The offset from the beginning of the image buffer
 *		(pImageBuf) to the upper left pixel of the image is
 *		calculated.  This is not necessarily pImageBuf +
 *		frameWidthInBytes + frameHeight * bpl because alignment
 *		issues may cause us to move the image over in the
 *		buffer a little bit.
 *
 *	Return code:
 *		ia_successful	successful completion
 *		ia_invalidParm	one of the pointers is NULL.
 *
 ********************************************************************/
Int32   CDECL 
i_calcImageOffset (
    IPState    *pIPState,
    UInt8      *pImageBuf,
    Int32	bpl,
    Int32	depth,
    Int32	frameWidth,
    Int32      *pOffsetTo00)
{
    PROLOGUE;
    Int32	 accessLen;
    Int32	 offsetTo00;

	/* Complain if any of the pointers are NULL */
    if (pOffsetTo00 == NULL)
	return ia_invalidParm;

	/* Figure out the longest access for this processor */
    ip_getLongestAccess (pIPState,  &accessLen);

	/* calculate the offset to (0,0) from the image buffer ptr.
	 * Clients can calculate backwards from this to get the
	 * location of the upper left corner of the frame that's
	 * required by ipshared routines.  For 32-bit accesses, this is simply
	 * move over by depth * cFrameBytes and down by cFrameBits lines.
	 * For 64-bit accesses, we want to make sure that
	 * pImageBuf + *pOffsetTo00 is a multiple of 8.  So,
	 * we calculate the offset first.  If the sum is not a
	 * multiple of 8, we add 4 to it.
	 */
    offsetTo00 = (frameWidth * depth)/8 + (frameWidth * bpl);
    if ( (accessLen == 8) &&
	 ((((UInt32)pImageBuf + offsetTo00) & 0x4) != 0) )
	    /* The first word would be on an odd word boundary,
	     * so move the offset over by 4.
	     */
	offsetTo00 += 4;

    *pOffsetTo00 = offsetTo00;

    NORMAL_RETURN;
}


/*

 $Log: i_frame.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 Imported sources

 Revision 1.12  1998/02/10 19:47:02  ddavies
 i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v

 Add switch cases to the frame clearing routines to handle 16-bit/pixel
 images.

 Revision 1.11  1998/01/21 21:24:02  ddavies
 i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v

 Fix indentation.
 Remove global declaration for ipVar_enableLongAccesses.  It's now part
 of the IPState structure.

 Revision 1.10  1998/01/13 23:47:53  ddavies
 i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.9  1996/09/24 02:38:16  ddavies
 ddavies on Mon Sep 23 19:34:31 PDT 1996

 i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v

 change i_calcImageProperties so that it deals with MMX (64-bit) processors.
 It now determines the bytes/line and fullHeight by looking at the processor
 type.
 Change i_calcImageOffset so that it also deals with 64-bit processors.
 It now uses the processor type and access length (from ip_getLongestAccess)
 to determine the offset from the beginning of the memory buffer containing
 the image to the beginning of the image.
 These are important changes in that older code that used to assume that
 it could calculate bytes/line or assumed that the start of the memory
 buffer was equal to the start of the image will no longer work correctly
 (sometimes).

 Revision 1.8  1996/09/13 18:41:19  ddavies
 ddavies on Fri Sep 13 11:40:20 PDT 1996

 i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v

 Add first implementations for i_calcImageProperties and i_calcImageOffset.
 The first routine takes an image width, height, depth and frame.  It
 returns the bytes/line and size of memory buffer required hold the image.
 These may be a function of the type of processor you're running on.
 i_calcImageOffset decides how much to shift the image so that the left
 column inside the frame is aligned on a byte boundary allowing the longest
 memory access the current processor can handle.

 Revision 1.7  1995/12/13 02:43:55  ejaquith
 i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/09/13  22:50:21  ddavies
 * ddavies on Wed Sep 13 15:50:03 PDT 1995
 *
 * i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v
 *
 * Add implementation for i_clearRightBottomFrames.  This routine is used to
 * clear out the frames on the right and bottom side of an image after
 * an operation used a dest that was slightly too small.
 *
 * Revision 1.5  1994/08/01  21:06:13  ddavies
 * i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v
 *
 * Change "static const" to "static" to work around buggy gcc/Purify interaction.
 *
 * Revision 1.4  1994/07/11  23:06:55  ddavies
 * i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v
 *
 * Pass rightWordMask down to ip_addFrameToLine so that all of the frame
 * bits in the rightmost word containing image are cleared.  The previous version
 * of the code assumed that frameless images were padded to byte boundaries
 * with 0's.  This is not true when we're adding a frame to an image that has
 * been rotated in place.
 *
 * Revision 1.3  1994/02/13  00:45:25  ddavies
 * ddavies on Sat Feb 12 16:45:15 PST 1994
 *
 * i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v
 *
 * Added implementation of i_clearAllFrames.  This procedure sets the frames
 * of an image to 0.
 *
 * Revision 1.2  1993/12/22  04:57:03  ddavies
 * ddavies on Tue Dec 21 20:55:59 PST 1993
 *
 * i_frame.c -> /product/ipcore/ipshared/src/RCS/i_frame.c,v
 *
 * Change i_removeFrame, i_removeFrameInPlace and i_addFrame so they can all
 * deal with frameless images whose byte/line is not a multiple of 4.  In the case
 * of i_removeFrameInPlace, this means adding a dBpl (dest bytes/line)
 * parmeter to the interface.
 *
 * Revision 1.1  1993/12/14  21:24:53  ddavies
 * Initial revision
 *

 */

