/*************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  i_glyph.c:  contains common setup for glyph convolution and 
 *		thresholding routines.
 *
 *          Contains subroutines
 *
 *                Int32         i_convolve()
 *                Int32         i_checkArea()
 *
 *                Int32         i_compileGlyphTables()
 *                Int32         i_extractGlyphRect()
 *		  Int32		i_convolveWhenLikePixels()
 *
 *		  Int32		i_createIAKernel()
 *		  Int32		i_destroyIAKernel()
 *		  Int32		i_getIAKernelDepth()
 *		  Int32		i_getIAKernelHeight()
 *		  Int32		i_getIAKernelIs3x3()
 *		  Int32		i_getIAKernelLowerThreshold()
 *		  Int32		i_getIAKernelName()
 *		  Int32		i_getIAKernelUpperThreshold()
 *		  Int32		i_getIAKernelWeights()
 *		  Int32		i_getIAKernelWidth()
 *		  Int32		i_setIAKernelUpperThreshold()
 *		  Int32		i_setIAKernelLowerThreshold()
 *
 *		  Int32		i_createIASlope()
 *		  Int32		i_destroyIASlope()
 *		  Int32		i_getIASlopeHorizDx()
 *		  Int32		i_getIASlopeHorizDy()
 *		  Int32		i_getIASlopeVertDx()
 *		  Int32		i_getIASlopeVertDy()
 *
 *		  Int32		i_glyphThresh()
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "lmask.prv"
#include "kernel.prv"
#include "frames.pub"
#include "memory.pub"

    /*  prototypes */
    /* shrglyph.prv MUST BE FIRST since it defines the private structures */
#include "shrglyph.prv"
#include "shrglyph.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_glyph.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");
#define  DEBUG_I_GLYPH    0
#ifdef DEBUG_ON				/* needed for WATCOM */
#if  DEBUG_ON
#undef	 DEBUG_I_GLYPH
#define  DEBUG_I_GLYPH    1		/* used to print array of centers */
#endif  /* DEBUG_ON */
#endif  /* DEBUG_ON */


/* An important component of AODS technology is the ability to embed data
 * in document.  We do this with arrays of tiny "glyphs," each representing
 * a 0 or a 1.  This is a misuse of the term since "glyph" really 
 * means any connected component.  Here, we mean one of a small set of 
 * marks used to represent data.
 *
 * We currently use two marks, a diagonal line that slopes from lower left
 * to upper right at 45 degrees and a line that slopes from upper left to
 * lower right at 45 degrees.  Each mark represents one bit of encoded data.
 * The line that slopes to the upper right represents a 0.  The line that
 * slopes towards the upper left represents a 1.  People have discussed
 * glyphs with four orientations, blobs on one end (another way to get four
 * orientations) and colored glyphs, but no real work has been put into them.
 *
 * Putting glyphs down on paper is relatively straightforward.  Yeah,
 * there's lots of stuff about encoding standards, ECC generation, byte
 * folding and sync lattices, but after all the formatting is done,
 * you just print the glyphs.
 *
 * Decoding is harder.  One must find the glyphs, figure out their size,
 * look for sync lattices, figure out how they were encoded, read the
 * data, compute ECC and correct if necessary, unfold the data and put it
 * somewhere.  This is all harder in the presence of errors since sync
 * lattices or data may be illegible.  It may be necessary to try reading
 * in different directions to get to all the legible glyphs.
 *
 * In our first implementations, the bulk of the time was spent finding the
 * center of each glyph, then determining whether it represented a 0 or a 1.
 * Both of these tasks are done using threshold convolution with
 * a matched filter.  What's that?  Say we want to find the center of a glyph.
 * Using the positions of the sync lattice glyphs (and possibly the
 * positions of neighboring glyphs), we estimate the position of the 
 * glyph's center.  We position the center-finding filter at each pixel
 * in a 3x3 neighborhood around the estimated center.  The filter is a
 * square array with the same dimensions as the glyph.  Each array 
 * position contains an integer.  We multiply all the weights in the filter
 * with the image pixels under them to form a convolution sum.  We do this
 * at each of the 3x3 filter positions, obtaining 9 sums.  The position with
 * the highest sum that's less than an upper threshold (we don't want to
 * be fooled by big black things) represents our best guess at the real
 * center of the glyph.  The center-finding filter has high
 * weights on both center diagonals so gets a high score when placed on
 * the center of either type of glyph.
 *
 * For awhile, the upper threshold on glyph sums was helping us track
 * the glyphs, but now it seems to be used less (is given a pretty high
 * value).  We'll keep it around in case it comes into fashion again.
 *
 * To decide whether the glyph represents a 1 or a 0, we use different
 * convolution kernels.  These kernels are symmetric in that all the
 * weights sum to 0.  The weights on one diagonal are positive, the weights
 * on the other diagonal are negative.  The convolution filter is placed
 * only at the center of the glyph.  In the current code, we use the sign
 * of the result to decide whether the glyph represented a 1 or a 0.  We
 * also require that the absolute value of the sum be above a threshold.
 * 
 * How can we improve convolution performance?  The first implementations
 * read each pixel individually, combined it with its corresponding kernel
 * weight, and added the product to the running sum.  Say we deal with
 * horizontal blocks of pixels instead of individual pixels.  We use each
 * block as an index into a table.  The table contains the inner product 
 * for that block, that is, the sum of the products of each pixel in the 
 * block with the corresponding weight.  This clearly improves performance.
 *
 * Terminology gets confusing now.  There is an array of input image
 * pixels.  There is an array of output convolved sums.  There is an
 * array of kernel weights.  Laying the kernel array on top of the input
 * image at exactly one position produces exactly one convolved sum.
 * We move the kernel array around to produce an array of convolved
 * sums.  When finding glyphs, the output array is either 3x3 (for 
 * finding centers) or 1x1 (distinguishing 0 from 1).
 *
 * Let's call the input the *image* array.  Let's call the output 
 * the *convolved* array.  The array of weights is called the convolution
 * kernel, or just the kernel.  The job of i_extractGlyphRect is to jump from 
 * estimated glyph position to estimated glyph position in the image 
 * array producing and discarding 3x3 and 1x1 convolved arrays.  
 * The convolved arrays are used to find glyphs and determine their 
 * orientation.  i_extractGlyphRect fills in two output arrays.  These
 * array have one bit position for each glyph.  In one array, we put a
 * 1 if we found a valid glyph at the corresponding position in the image.
 * In the other, we put a 1 if the glyph represented a 1.
 *
 * How wide should the blocks of image pixels be?  Ideally, we'd like
 * to grab blocks as wide as the kernel since all of those pixels will
 * involved in the calculation of the result.  If we are supposed to
 * be doing convolutions for 3x3 neighborhoods of destinations, it
 * would be nice if the horizontal blocks included all the pixels needed
 * for a row of 3 convolved sums in parallel.  Well, ok, it would be
 * swell if we could generate 3, 6, or 9 convolved sums in parallel from
 * a horizontal block of pixels, but we can't fit more than 3 sums of
 * reasonable size into 32 bits.
 *
 * If the convolution kernel is of width w, then the center sum in an 
 * convolved array row will use w input pixels from the image array.  
 * The block of w pixels required by the sum to the left will be shifted
 * left by 1 relative to the w pixels required by the center sum.  That 
 * is, the two sets of pixels overlap in w-1 positions.  We only need one
 * new pixel to compute both convolved results.  Similarly, 
 * the w pixels needed for the sum to the right overlaps the center's 
 * image pixels in w-1 places.  The total number of pixels needed to 
 * compute all three sums in parallel is w + 1 (on the left) + 
 * 1 (on the right) or w + 2.
 *
 * The biggest problem with table lookup schemes is that the table indices
 * often get too big.  Say we wanted to use a 5x5 kernel on 4 bit/pixel
 * image data.  To do the 3x3 convolution, we'd need 5+2 = 7 pixels to index
 * each table entry.  That's 7 pix * 4 b/pix = 28 bits of index.  
 * The table would have 256M entries, each with 4 bytes, so would 
 * consume 1GB.  Too, too much.
 *
 * Well, how many image bits do we really need to worry about?  Dan Bloomberg
 * said these are the glyph sizes and image depths we should consider:
 *
 *	Glyph Size   Depths (b/pix)	1x1 widths    3x3 widths
 *	 2 x  2		   2*,4*	 		     8*, 16*
 *	 3 x  3		1*,2, 4		 6, 12		 5*, 10, 20
 *	 4 x  4		1*,2, 4		 8, 16		 6*, 12, 24
 *	 5 x  5		1, 2, 4		 5, 10, 20	 7,  14, 28
 *	 6 x  6		1, 2		 6, 12		 8,  16
 *	 7 x  7		1, 2		 7, 14		 9,  18
 *	 8 x  8		1		 8		10
 *	 9 x  9		1		 9		11
 *	10 x 10		1		10		12
 *	11 x 11		1		11		13
 *	12 x 12		1		12		14
 *	13 x 13		1		13		15
 *	14 x 14		1		14		16
 *
 * The * values are there because the center-finding kernels for the
 * larger sizes have blank rows and columns around the outside, reducing
 * their effective sizes.  NOTE that these extra sizes apply only to the
 * center-finding (3x3) convolution widths.
 *
 * The sizes in the 1x1 column are obtained by multiplying the glyph width
 * by the number of bits/pixel.  The 3x3 column multiplies (glyph width + 2)
 * by bits/pixel.
 *
 * The convolution additions need not be done in any particular order.
 * Therefore, we can split the blocks of bits listed above into groups,
 * obtain sums for each group and combine the sums.  This puts us somewhere
 * between treating each pixel individually and handling the entire row
 * in one operation.
 *
 * How big should the groups be?  As a going-in position, let's assume no
 * group contains more than 8 bits of image.  That's eight binary pixels,
 * four 2-bit pixels or two 4-bit pixels.  We may decide that this is too
 * simple at some point, but it feels reasonable now.  It keeps the lookup
 * tables fairly small.
 *
 * What about the other bits in a table index?  The subtables for each group
 * of image bits will be different.  There will really be subtables for
 * each horizontal group in each kernel row.  For example, when convolving
 * with a 5x5 kernel to produce a 3x3 convolved array using 4 bit/pixel
 * images, we divide the 28 image bits/row into three 8-bit groups and a
 * 4-bit group.  Since there are 5 rows in the kernel, there are a total 
 * of 5*4 = 20 subtables.  Does that mean we need at least 5 extra bits 
 * of table index?  If so, the tables would have 2**(5+8) = 8K entries.  
 * Each entry in the 3x3 tables has 4 bytes (see table entries below) 
 * for a total of 32KB.  Yuck.
 *
 * To get around this problem, we'll cheat.  The code making the table will
 * know about the code that reads the table.  Another way of saying this
 * is that changes to one may require changes to the other.  Not ideal
 * software engineering, but it could reduce the total table size by a 
 * factor of 32.  The code that makes the table simulates the sequence 
 * in which the table will be used.  For example, it could know that 
 * the horizontal groups of the topmost row will be used first and will 
 * be accessed in right to left order.  It could put the subtables into 
 * the main table in this order.  The code using the table adjusts its 
 * table pointer as it goes along knowing where the next subtable will be.
 *
 * Even this requires a little thought.  When producing a 3x3 convolved
 * array, a horizontal group of image pixels might be used to calculate
 * 1, 2 or 3 different convolved rows.  Huh?  Say we're using a 5x5 kernel
 * and are producing a 3x3 convolved array.  Just as we need to read 5+2
 * pixels horizontally in each row, we need to read a total of 5+2 image
 * rows to produce all three convolved array rows.  The topmost image row
 * will be used only to produce the topmost convolution row via the topmost
 * row of the convolution kernel.  The second image row will be used to
 * produce the topmost convolved row using the second kernel row
 * and to produce the second convolved array row using the topmost kernel
 * row and so on.  A mapping of this is shown below.
 *
 *	Image Row	Kernel Row  ->	   Convolved Row
 *	  0		  0			0
 *
 *	  1		  1			0
 *	  1		  0			1
 *
 *	  2		  2			0
 *	  2		  1			1
 *	  2		  0			2
 *
 *	  3		  3			0
 *	  3		  2			1
 *	  3		  1			2
 *
 *	  4		  4			0
 *	  4		  3			1
 *	  4		  2			2
 *
 *	  5		  4			1
 *	  5		  3			2
 *
 *	  6		  4			2

 * How should the subtables be ordered?  If the big table contains exactly
 * one copy of each subtable, it makes sense to put all the subtables for
 * the horizontal groups making up one row next to each other.  The
 * subtables for the next row would follow and so on.  However, that would mean
 * that the 3x3 convolution code would maintain as many as three pointers to
 * subtables.  For example when processing the leftmost horizontal group
 * of image row 2, it needs to access the subtables for the leftmost
 * horizontal groups in rows 0, 1 and 2.  That's a bit annoying, especially
 * in the PC where registers are scarce.
 *
 * Ok, say we group the horizontal subtables vertically instead of
 * horizontally.  That is, we put all the leftmost tables together, then
 * the next-to-leftmost and so on.  Well, this means the convolution code
 * has to do a fair amount of jumping forward over unused subtables as it
 * moves from left to right within any one row, then backtrack for the
 * next row.  Yeah, this could be made to work, but seems unnecessarily
 * complex.
 *
 * It would be nice to relax the constraint that only one copy of each
 * subtable appear in the big table.  Say the code making the table
 * puts in the subtables for row 0, then the subtables for rows 1 and 0,
 * then the subtables for rows 2, 1, and 0, then the subtables for rows
 * 3, 2 and 1 (see table above) and so on.  This increases the total table
 * size by a factor of 3, but makes the inner loop simple.  The convolution 
 * code would simply advance its table pointer each time a new result 
 * was needed.  One variable, nice and easy.
 *
 * How big does that make the tables?  In the 3x3 convolved array with
 * 5x5 kernel on 4 bits/pixel, we'd have:
 *	15 rows (number of rows in the table above) *
 *	(3 subtables * 256 entries/subtable + 1 subtable * 16 entries) *
 *	4 bytes/entry
 *		= 47040 bytes.
 *
 * That's awful.  On the other hand, how big are the tables at 1 bit/pixel
 * for the same kernel and convolved array?
 *	15 rows (number of rows in the table above) *
 *	(1 subtable * 128 entries/subtable) *
 *	4 bytes/entry
 *		= 7680 bytes.
 *
 * Not as bad, but not swell either.  Putting a copy of every
 * subtable in every time its needed increases the total table size by
 * 3 (the number of rows in the output convolved array).
 *
 * Agonize, moan.  Ok, I surrender.  These are all big tables and increasing
 * their sizes by 3x is a worse thing that maintaining a couple of extra
 * pointers.  Note that we end up with at least 3 registers no matter now
 * we do it (ptr,ptr,ptr; ptr,offset,temp; ptr;offset;offset;temp), so
 * we might as well use three pointers.
 * 
 * THE RESOLUTION: For each output array, we identify a small portion of
 * input array.  This is the portion that affects the output.  We process
 * this input subarray from right to left (see below), top to bottom in
 * blocks of 8 bits.
 * These blocks correspond to 8, 4 or 2 pixels.  Each block of input bits 
 * is used as an index into 1, 2 or 3 compiled subtables of kernel weights 
 * depending on the number of rows of convolved array to be produced from 
 * the particular input row (see table above).
 * The compiled table will contain the subtables for the first kernel
 * row from right to left followed by those for the second row from right
 * to left and so on.  Encoded in the kernel data structure will be the
 * number of bytes taken by all the subtables in a row.  We already know that
 * each subtable to the left of the rightmost one in a row requires
 * 256 * entryWidth bytes (because we index each table with 8 image bits).
 * This reduces the size of the tables above by a factor of 3, to 15580 bytes
 * and 2560 bytes respectively.
 *
 * The other benefit of this solution is that the table layout is
 * straightforward and reasonably independent of the code using the tables.
 *
 * But why do we process from right to left within a row?  Because it
 * simplifies processing.  We will hold all the image bits in a single 32-bit
 * variable.  The bits must be at the LS end of an index variable to be
 * used as a table index.  We could process the image bits from left to
 * right, shifting each 8-bit block of the image down to the low end
 * independently, but that takes an extra variable to 
 * remember the changing shift counts.  If we consume the bits
 * from the LS end of the register to the MS end, we simply shift the
 * right by 8 after each step.  It's simpler and eliminates a variable.
 *
 * What do table entries look like?  In the 1x1 case, life is easy.  Each
 * table entry simply contains a positive or negative number which is the
 * algebraic sum of the weights and image elements for its horizontal group.
 * Since we know that the absolute value of convolution sums is less than
 * 512 (see below), each entry would be 2 bytes.  For the sake of
 * having simple interfaces, table entries will always be 4 bytes.
 *
 * The 3x3 case is a bit trickier since we'd like each table entry to have
 * the inner products for 3 convolved array elements.  Dan Bloomberg and
 * Rob Weltman have agreed that the maximum absolute value for any possible
 * convolution is 512.  Therefore, we can use 9 bits plus a sign bit to
 * represent each sum.  Say we divide each 32 bit table entry as follows:
 *
 *	Bit number:	31	    22 | 21	     11 | 10		0
 *	Value:		Sum 0 (10 bits)	 Sum 1 (11 bits)  Sum 2 (11 bits)
 *
 * What about negative numbers?  Don't the carries from addition on negative
 * numbers affect the more significant sums in the word?  They would
 * if we let them.  There are two sources of negative numbers, the values 
 * we add from the tables and the current sum values.  We'll deal
 * with them independently.
 *
 * We bias the current sums by 512 initally.  That means their range of
 * -512:511 is translated to 0:1023.  This fits in 10 bits.  When subtract
 * off the bias before returning the sum to a calling routine.
 *
 * The table entries must be both positive and negative because it must
 * be possible to both increase and decrease each sum.  To handle this, 
 * we inspect each set of three entries after it is formed from right (least 
 * significant, Sum 2 above) to left (most significant, Sum 0 above).
 * If the value under inspection is negative, we subtract 1 from *the
 * next more significant value*.  Why?  Because we know that when the
 * negative value is added to the current sum, there will be a carry into
 * the next most significant pixel.  By subtracting 1 from the next most
 * significant entry in the table word now, we will end up with the 
 * corrent value when the table value is used.  Since the value of one 
 * sum can cause the next most significant sum to become negative, we 
 * must inspect them from least significant to most significant.
 *
 * This gets us to the issue of where these tables are kept.  It would be
 * a pain to regenerate them every time i_extractGlyphRect was called.  The
 * current glyph software defines a Kernel data structure.  Since these
 * tables are nothing but "compiled" versions of the kernel weights, it
 * makes sense to put them in there, so we will.
 *
 * Do we need two sets of tables for each kernel, one for 1x1 and one for
 * 3x3?  Not today.  For glyph work, we only have two types of kernels.
 * Center-finding kernels are always used to produce 3x3 convolved
 * arrays and 0/1 kernels (the ones used to distinguish glyphs representing
 * 1s from those representing 0s) are always used to produce 1x1 convolved
 * arrays.  Knowing the type of the kernel tells us what type of tables
 * to make.
 *
 * Since we may wish to tune kernels for different image depths (1, 2 or
 * 4 bits/pixel), it makes sense to have one kernel for each combination 
 * of glyph size, function (center-finding, 0/1 discrimination) and depth.  
 * We'll do that.  The kernels are currently kept in the file glyconv.se.
 * Today, it's stored in /product/strel.
 *
 * In addition to the kernel weights, we also need to know the threshold
 * for each kernel.  Actually, there are at most two thresholds per kernel.
 * For center-finding kernels, we want to make sure that the convolution
 * sum is between two values.  If it's above the upper threshold, the
 * region we're looking at has too many dots to be a glyph.  If the sum 
 * is below the lower threshold, it has too few.  The 0/1  discrimination
 * kernels only need 1 threshold.  If the absolute value of the sum is
 * greater than this value, we can feel reasonably certain that this is
 * a glyph.  The threshold values are recorded in each kernel data structure.
 *
 * Now that we've groveled in the low level convolution code, what should
 * extractGlyphRect do?  It is used to find a rectangular region of glyphs.
 * We were going to do a linear region, but the calling overhead got too
 * large.  i_extractGlyphRect is given an image pointer to the center of either 
 * the upper left glyph in the image or the lower right one.  A "direction"
 * parm is equal to -1 if the rectangle is to be processed right-to-left,
 * bottom-to-top.  Otherwise, direction = +1 and the glyph rectangle 
 * is processed left-to-right, top-to-bottom.  The routine
 * returns two bit arrays.  One specifies the value of each glyph.  The
 * other says whether the value in the first array is valid.  Both arrays
 * are packed tightly; that is, if the rectangle to be interpreted has n
 * rows of n columns each, each bit array requires exactly n*m bits.  The
 * m bits of one row are adjacent to the m bits of the next one.
 *
 *	1. If the "direction" parameter is negative, negate the values used
 *	   to move vertically and horizontally.
 *	   Also point to the end of both result bit vectors.
 *	2. Convolve the area with the given center-finding kernel, producing
 *	   a 3x3 result.
 *	3. Find the sum with the greatest value less than the upper threshold.
 *	   If this sum is less than the lower threshold, mark this bit
 *	   position as invalid and proceed to step 8 - there's no
 *	   glyph here.
 *	4. Using the point found (the original center's value is 
 *	   preferred among equals) convolve with the 0/1 discrimination 
 *	   kernel producing a 1x1 result.
 *	5. If the absolute value of the result is less than the lower
 *	   threshold, mark this position as invalid.  In any case,
 *	   use the sign of the result to determine whether this glyph is a
 *	   0 or a 1.  Why?  In some cases, the glyph data is repeated over
 *	   several partially occluded blocks.  Higher level software will
 *	   normally disregard weak glyphs, but in that case it will try to
 *	   do some sort of average or voting to retrieve the signal.
 *	6. Concatenate the result bit with the existing the bit vector.
 *	   Mark the bit as valid in the validity bit vector.
 *	   Add the direction parameter to the bit pointers to both bit vectors.
 *	7. If the self_clocking parm is cTrue, use the center position we found
 *	   to find the next center position as described in the next step.  
 *	   Otherwise, use the originally estimated center position.
 *	8. Add horizDx and horixDy to this position to go to the next
 *	   horizontal glyph.
 *	   horizDx and horixDy are scaled integers with 16 bits of integer and
 *	   16 bits of fraction.  They are in units of pixels.
 *	9. Repeat steps 2 through 8 for each glyph in the current row.
 *     10. Add vertDx and vertDy to the initial row position to move 
 *	   up/down a row.
 *     11. Sometimes, alternate rows of glyphs are offset by 1/2 of the
 *	   glyph width.  This tends to keep the glyph ends from intersecting.
 *	   If this is true, the xAlternatingOffset parm.  If the glyphs rows
 *	   are offset, this equals glyphWidth/2, otherwise
 *	   it's 0.  We always add xAlternatingOffset to the next center's
 *	   position and negate xAlternatingOffset.
 *     12. Repeat steps 2 through 12 until the specified number of
 *	   glyph positions have been investigated.  Return.
 *
 * There are a few other performance considerations.  First, some
 * kernels have all 0's in the top and/or bottom rows.  Convolving with
 * these would be a waste of time.  Encoded in the Kernel data structure
 * is the number of the initial row of non-zero kernel weights and the
 * number of kernel rows to be processed.
 *
 * Second, some kernels don't use the leftmost and/or rightmost columns.
 * Ignoring these allows us to reduce the number of horizontal blocks and/or
 * the number of entries per table.  Encoded in the Kernel data structure
 * will be the number of the leftmost active column (relative to 0) and 
 * the number of active columns.
 *
 * In computing the number of active rows and columns, we make no attempt
 * to skip all-zero rows or column that are sandwiched between active rows
 * and columns.  We only chop off all-zero rows and column around the
 * outside of the kernel.
 *
 * Software engineering considerations:  What's done where?  It has been
 * our general policy to put accelerated functions into ipshared and
 * everything else outside.  ipshared functions normally don't take 
 * structures as arguments, just integers and pointers to arrays.
 *
 * However, since there are a LOT of parameters for i_extractGlyphRect
 * and many of them do not change for all the glyphs on the page, we create 
 * and maintain two private structures.  One, IAKernel, has all the
 * information about a convolution kernel.  i_extractGlyphRect needs two
 * kernels, so these structures are handy.  The other, IASlope, has the
 * spacing information needed to move from glyph to glyph.  This is a function
 * of the glyph size, the scanner resolution and the skew angle of the
 * image.  These are constant for a given region of glyphs.
 *
 * Applications code reads kernel information from a file like glyconv.se
 * and passes it to i_createIAKernel.  This routine creates a private IAKernel
 * structure and passes back an opaque handle.  Applicaions, being compiled
 * with only shrglyph.pub, see the handle, a IAKernelPtr, as a pointer to
 * void.  They can only pass it back to the ipshared glyph routines.  This
 * code, compiled also with shrglyph.prv, can see inside the structure.
 *
 * Having created the structure, applications code then calls
 * i_compileGlyphTables.  This routine fills in lots of other private
 * information including the tables full of inner products described
 * above and, if possible, a pointer to machine code that will do the
 * convolution required by this kernel.
 *
 * i_compileGlyphTables takes as args a pointer to the kernel struct,
 * a pointer to a memory allocation routine and a pointer to a memory
 * freeing routine.  The memory allocation routine is used to allocate
 * the tables.  The freeing routine is used when errors are encountered
 * after some memory has been allocated.  Since the tables exist for
 * awhile, the caller might be picky about how they are allocated and freed.
 *
 * i_compileGlyphTables returns with all of the IAKernel parameters filled
 * in if successful.  If it is not successful, it frees any memory it 
 * allocated.  The standard error codes are used.  The resulting tables 
 * are meaningful only to i_extractGlyphRect.
 *
 * i_extractGlyphRect operates generally as defined above.  We need to discuss
 * the convolution code in a little more depth.
 *
 * There are 22 combinations of depth and kernel size listed above.  For
 * each of the 1 bit/pixel combinations, there are 32 possible starting
 * points within a word.  For each of the 2 bit/pixel images, there are
 * 16 possible starting points.  Each of the 4 bit/pixel images have
 * 8 possible starting points.  The starting points are important because
 * they govern the sequence of shifts and masks required to use image data
 * as table indices.  This is a total of (12 * 32) + (6 * 16) + (4 * 8)
 * or 512 cases (a coincidence! Honest!).  I don't want to write out anything
 * like 512 little blocks of code.  Yet, we don't want to spend a lot of
 * time in each convolution messing with loop counts, arbitrary shifts 
 * and tests.
 *
 * The solution is to have two sets of code, slow and fast.  The slow code
 * will be written in C.  It will indeed mess with loop counts, arbitrary
 * shifts and tests.  There will actually be two sets of this code, ip_convolve3x3
 * does center-finding and returns the highest convolution sum less than
 * the upper threshold and the location of the the pixel giving that sum.
 * ip_convolve1x1 does a simple convolution and returns the sum.  This routine
 * can also be used for other convolution tasks.
 *
 * The fast code will be machine code compiled for each kernel by a
 * C program.  There will be one such "compiler" for each processor platform.
 * This is similar to the way we addressed morphology.  The file generating
 * code will be named ssgencon.c for SPARCStation and pcgencon.c for PC's.
 * Those modules will have lots of comments.
 *
 * How are the fast and slow modules built?  This file, i_cglyph.c, is common
 * to all.  The pure C implementation adds sglyph.c which contains C
 * implementations for ip_convolve1x1 and ip_convolve3x3.  It also contains
 * a stub for ip_genConvAsm, the routine that generates assembly code.  The
 * fast code for SPARCs is built from this file, fglyph.c and ssgencon.c.
 * The latter file contains an implementation of ip_genConvAsm for SPARCs.
 * Similarly, fast PC code is built from this file, fglyph.c and pcgencon.c.
 * fglyph.c contains only stubs for ip_convolve1x1 and ip_convolve3x3.  To avoid
 * an extra level of subroutine call, i_compileGlyphTables puts a pointer
 * to the machine code version of convolution into the kernel's struct
 * if machine code exists.  If there is no machine code for this platform,
 * it has the struct point to ip_convolve1x1 or ip_convolve3x3 as appropriate.
 *
 * We should also discuss x,y glyph coordinates.  The image address
 * passed to i_extractGlyphRect specifies the upper left corner of the
 * first glyph.  To account for the fact that the first active row and 
 * column might not be the kernel's row 0, col 0; we then add 
 * (firstActiveCol, firstActiveRow) to get the coordinates of the
 * upper left pixel.
 *
 * The easiest way to translate these coordinates into memory addresses
 * is to multiply the y coordinate by bytes/line and the word portion of
 * x.  The caller of i_extractGlyphRect does this for the first glyph.
 * Doing multiplies for every glyph is too slow.  Instead, we'll set up 
 * an table with multiples of words/line.  We use the table to supply 
 * the small vertical image offsets required to move "horizontally" to 
 * the next glyph.  When moving relatively large vertical distances
 * between rows of glyphs, we'll really multiply the y distance by words/line.
 * We assume that vertical moves don't happen very often compared to horizontal
 * moves - so we use the small table for horizontal moves and real multiply
 * ops for vertical moves.  This allows us to keep the table small and 
 * independent of glyph spacing.
 *
 * Higher level code occasionally needs to do straight convolutions.  This
 * is done when trying to find the exact coordinates of glyphs.
 * The routine i_convolve is a wrapper for ip_convolve1x1.  i_convolve takes
 * x,y coordinates from the upper left corner of an image instead of a 
 * pointer to the glyph itself.
 *
 * i_checkArea is a wrapper for ip_convolve3x3.  It takes the addresses
 * of x, y coordinates.  It uses ip_convolve3x3 to do a convolution in a 3x3
 * space and choose the best convolution sum.  i_checkArea translates the
 * modified image pointer and centerX parms that ip_convolve3x3 returns
 * into regular x,y coordinates and returns these.
 *
 * *************************** GRAY IMAGES ****************************
 *
 * The first implementation attempted to handle gray images of glyphs, but
 * failed.  I often equated "pixel" and "bit," especially when navigating
 * around the image.
 *
 * We must do the slope calculations in units of pixels since we always want
 * to end up on a pixel boundary.  On the other hand, when we access the
 * image, we need to know about bits.  The lowest level routines can almost
 * ignore the image depth since it's built into the tables.  However, the
 * 3x3 convolution routine needs to know about depth just so it can move
 * the initial convolution point up and *to the left* by the proper number
 * of bits.
 *
 * We will make the following changes:
 *	1. Add a field called firstActiveColBit to the kernel structure.
 *	   It contains firstActiveCol * depth.
 *	2. In i_extractGlyphRect, we keep curX in units of pixels, but
 *	   centerXBit (replaces centerX) is in units of bits.  This bit
 *	   value is handed to the lower level routines.  It is not a
 *	   split integer.
 *	3. The lower level routines add pKernel->firstActiveColBit to the
 *	   incoming centerXBit arg instead of adding firstActiveCol.
 *	   They also subtract pKernel->depth instead of 1 to move to the left.
 *	4. If self clocking is enabled, the new value of curX is calculated
 *	   as ( (centerXBit << (16 - logDepth) | (curX & 0xFFFF) ) + horizDx.
 *	   I.e. the integer part of curX is replaced with the integer
 *	   pixel number from centerXBit before horizDx is added.
 *
 * *************************** STRUCTURES *****************************
 *
 * In an attempt to avoid passing a bunch of parameters that don't change
 * each time we call i_extractGlyphRect, we have created two structures.
 * An IAKernel represents all the information one needs about convolution
 * kernels.  An IASlope specifies the slope and spacing of a block of glyphs.
 * In particular, it holds the x and y increments needed to move horizontally
 * and vertically by units of 1 glyph.
 *
 * To insulate applications code from changes in these structures, we
 * define a set of routines used to create them, destroy them and to
 * read and write their public fields.  These are i_createIAKernel,
 * i_destroyIAKernel, i_getIAKernelName, i_getIAKernelWidth,
 * i_getIAKernelHeight, i_getIAKernelDepth, i_getIAKernelWeights, 
 * i_getIAKernelUpperThreshold, i_getIAKernelLowerThreshold;
 * i_createIASlope, i_destroyIASlope, i_getIASlopeHorizDx,
 * i_getIASlopeHorizDy, i_getIASlopeVertDx and i_getIASlopeVertDy.
 */ 



/* *************** IAKernel Routines ********************* */



/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_createIAKernel():
 *        Creates an instance of an IAKernel data structure, returning
 *	  a pointer to via a location passed as a parameter.
 *
 *        This routine allocates space using an allocator passed in
 *	  by the caller.  It initializes all the caller-visible fields
 *	  with values passed in as parameters.
 *
 *	ppKernel:	Ptr to ptr to kernel structure.  The ptr to the
 *			kernel will be filled in by this routine.
 *	name:		Ptr to string of chars containing ASCII name
 *			of the kernel.
 *	width:		Width of kernel array
 *	height:		Height of kernel array
 *	depth:		Depth of image on which kernel is to be used.
 *	weights:	Width x height array of INT32's, each holding
 *			one kernel weight.  These are listed in left to
 *			right, top to bottom order as shown in
 *			/product/strel/glyconv.se.
 *	upperThreshold:	If the result of a convolution with the 
 *			center-finding kernel is bigger than this, 
 *			there's no glyph here.
 *	lowerThreshold: Used to decide if valid glyph position (for
 *			center-finding kernel) or valid glyph (for
 *			0/1 kernel) has been found.
 *	is3x3:		cTrue if kernel will be used to produce a 3x3
 *			convolved array.
 *			cFalse if kernel will be use to produce a 1x1
 *			convolved array.
 *	allocator:	Ptr to function that is used to allocate memory
 *			for the table.  Function's single argument is
 *			the number of BYTES requested.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of malloc
 *			provided by WATCOM on the PC!!
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem 		could not allocate memory
 */
Int32 CDECL 
i_createIAKernel (
    IPState    *pIPState,
    IAKernelPtr *ppKernel,
    char       *name,
    UInt32	width,
    UInt32	height,
    UInt32	depth,
    Int32      *weights,
    Int32	upperThreshold,
    Int32	lowerThreshold,
    UInt32	is3x3,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	))
{
    PROLOGUE;
    IAKernel	*pKernel;	/* local ptr to kernel */

	/* create space for the kernel */
    pKernel = (IAKernel *) allocator(pIPState, sizeof(IAKernel));
    if (pKernel == NULL)
	return ia_nomem;

	/* set the returned kernel pointer and fill in the fields */
    *ppKernel = pKernel;
    pKernel->name = name;
    pKernel->width = width;
    pKernel->height = height;
    pKernel->depth = depth;
    pKernel->weights = weights;
    pKernel->upperThreshold = upperThreshold;
    pKernel->lowerThreshold = lowerThreshold;
    pKernel->is3x3 = is3x3;

	/* initialize the private fields */
    pKernel->numActiveRows = 0;
    pKernel->firstActiveRow = 0;
    pKernel->numActiveCols = 0;
    pKernel->firstActiveCol = 0;
    pKernel->firstActiveColBit = 0;
    pKernel->bytesPerTableRow = 0;	/* no tables yet */
    pKernel->finalTablePtrVal = 0;
    pKernel->compWeights = NULL;
    pKernel->conv1Code = NULL;
    pKernel->conv3Code = NULL;

    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_destroyIAKernel():
 *        frees an instance of an IAKernel data structure using
 *	  a routine passed in by the caller.
 *
 *	pKernel:	Ptr to kernel structure.
 *	freer:		ptr to function that is used to free memory.
 *			This function is used to free all the strings,
 *			code and tables associated with the kernel.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of free
 *			provided by WATCOM on the PC!!
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_destroyIAKernel (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *))
{

	/* free the strings, code and tables attached to the kernel */
    PROLOGUE;
    freer(pIPState, (void *) pKernel->compWeights);
	/* only free the pointer to the convolution code if it was
	 * space filled with assembly code.  if this points to C code
	 * DO NOT free it.
	 * If this is Visual C++, turn off the warning that says we're
	 * casting a pointer to a function to a pointer to data.
	 * As if (void *) were a pointer to data!
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    if ((pKernel->conv3Code != ip_convolve3x3) &&
        (pKernel->conv3Code != NULL))
	freer(pIPState, (void *) (pKernel->conv3Code));

    if ((pKernel->conv1Code != ip_convolve1x1) &&
	(pKernel->conv1Code != NULL))
	freer(pIPState, (void *) (pKernel->conv1Code));
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

	/* free the structure itself */
    freer(pIPState, pKernel);

    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIAKernelName():
 *        returns a pointer to the string containing the kernel's name
 *	  in variable that was passed in.  THIS POINTER POINTS TO TRASH
 *	  IF THE CALLER DESTROYS THE KERNEL STRUCTURE!!!
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIAKernelName (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    char      **ppName)
{
    PROLOGUE;
	/* do the useful work */
    *ppName = pKernel->name;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIAKernelHeight():
 *        returns the kernel's height in variable that was passed in.
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIAKernelHeight (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    UInt32     *pHeight)
{
    PROLOGUE;
	/* do the useful work */
    *pHeight = pKernel->height;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIAKernelWidth():
 *        returns the kernel's width in variable that was passed in.
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIAKernelWidth (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    UInt32     *pWidth)
{
    PROLOGUE;
	/* do the useful work */
    *pWidth = pKernel->width;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIAKernelDepth():
 *        returns the kernel's depth in variable that was passed in.
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIAKernelDepth (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    UInt32     *pDepth)
{
    PROLOGUE;
	/* do the useful work */
    *pDepth = pKernel->depth;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIAKernelWeights():
 *        returns a pointer to the array containing the kernel's weights
 *	  in variable that was passed in.  THIS POINTER POINTS TO TRASH
 *	  IF THE CALLER DESTROYS THE KERNEL STRUCTURE!!!
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIAKernelWeights (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    Int32     **ppWeights)
{
    PROLOGUE;
	/* do the useful work */
    *ppWeights = pKernel->weights;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIAKernelUpperThreshold():
 *        returns the kernel's upper threshold in variable that was passed in.
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIAKernelUpperThreshold (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    UInt32     *pUpperThreshold)
{
    PROLOGUE;
	/* do the useful work */
    *pUpperThreshold = pKernel->upperThreshold;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIAKernelLowerThreshold():
 *        returns the kernel's lower threshold in variable that was passed in.
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIAKernelLowerThreshold (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    UInt32     *pLowerThreshold)
{
    PROLOGUE;
	/* do the useful work */
    *pLowerThreshold = pKernel->lowerThreshold;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIAKernelIs3x3():
 *        returns flag saying whether kernel is to be used to produce
 *	  a 3x3 array of outputs in arg.
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIAKernelIs3x3 (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    UInt32     *pIs3x3)
{
    PROLOGUE;
	/* do the useful work */
    *pIs3x3 = pKernel->is3x3;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_setIAKernelLowerThreshold():
 *        sets the kernel's lower threshold to given value.
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_setIAKernelLowerThreshold (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    UInt32	lowerThreshold)
{
	/* do the useful work */
    PROLOGUE;
    pKernel->lowerThreshold = lowerThreshold;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_setIAKernelUpperThreshold():
 *        sets the kernel's upper threshold to the given value.
 *
 *	pKernel:	Ptr to kernel structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_setIAKernelUpperThreshold (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    UInt32	upperThreshold)
{
	/* do the useful work */
    PROLOGUE;
    pKernel->upperThreshold = upperThreshold;
    NORMAL_RETURN;
}



/* *************** IASlope Routines ********************* */


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_createIASlope():
 *        Creates an instance of an IASlope data structure, returning
 *	  a pointer to via a location passed as a parameter.
 *
 *        This routine allocates space using an allocator passed in
 *	  by the caller.  It initializes all the caller-visible fields
 *	  with values passed in as parameters.
 *
 *	ppSlope:	Ptr to ptr to slope structure.  The ptr to the
 *			slope will be filled in by this routine.
 *	horizDx:	Amount one moves left in the image to move
 *			to the left by one glyph.
 *	horizDy:	Amount one moves down in the image to move to the
 *			left by one glyph.
 *	vertDx:		Amount one moves left in the image to move down
 *			by one glyph.
 *	vertDy:		Amount one moves down in the image to move down
 *			by one glyph.
 *	allocator:	Ptr to function that is used to allocate memory
 *			for the table.  Function's single argument is
 *			the number of BYTES requested.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of malloc
 *			provided by WATCOM on the PC!!
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem 		could not allocate memory
 */
Int32 CDECL 
i_createIASlope (
    IPState    *pIPState,
    IASlopePtr *ppSlope,
    ScaledInt1616 horizDx,
    ScaledInt1616 horizDy,
    ScaledInt1616 vertDx,
    ScaledInt1616 vertDy,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	))
{
    PROLOGUE;
    IASlope	*pSlope;	/* local ptr to slope */

	/* create space for the slope */
    pSlope = (IASlope *) allocator(pIPState, sizeof(IASlope));
    if (pSlope == NULL)
	return ia_nomem;

	/* set the returned slope pointer and fill in the fields */
    *ppSlope = pSlope;
    pSlope->horizDx = horizDx;
    pSlope->horizDy = horizDy;
    pSlope->vertDx = vertDx;
    pSlope->vertDy = vertDy;

    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_destroyIASlope():
 *        frees an instance of an IASlope data structure using
 *	  a routine passed in by the caller.
 *
 *	pSlope:	Ptr to slope structure.
 *	freer:		ptr to function that is used to free memory.
 *			This function is used to free all the strings,
 *			code and tables associated with the slope.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of free
 *			provided by WATCOM on the PC!!
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_destroyIASlope (
    IPState    *pIPState,
    IASlopePtr	pSlope,
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *))
{

	/* free the structure itself */
    PROLOGUE;
    freer(pIPState, pSlope);

    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIASlopeHorizDx():
 *        returns the slopes's horizDx entry
 *	  in variable that was passed in.
 *
 *	pSlope:	Ptr to slope structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIASlopeHorizDx (
    IPState    *pIPState,
    IASlopePtr	pSlope,
    ScaledInt1616 *pHorizDx)
{
    PROLOGUE;
	/* do the useful work */
    *pHorizDx = pSlope->horizDx;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIASlopeHorizDy():
 *        returns the slopes's horizDy entry
 *	  in variable that was passed in.
 *
 *	pSlope:	Ptr to slope structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIASlopeHorizDy (
    IPState    *pIPState,
    IASlopePtr	pSlope,
    ScaledInt1616 *pHorizDy)
{
    PROLOGUE;
	/* do the useful work */
    *pHorizDy = pSlope->horizDy;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIASlopeVertDx():
 *        returns the slopes's vertDx entry
 *	  in variable that was passed in.
 *
 *	pSlope:	Ptr to slope structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIASlopeVertDx (
    IPState    *pIPState,
    IASlopePtr	pSlope,
    ScaledInt1616 *pVertDx)
{
    PROLOGUE;
	/* do the useful work */
    *pVertDx = pSlope->vertDx;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_getIASlopeVertDy():
 *        returns the slopes's vertDy entry
 *	  in variable that was passed in.
 *
 *	pSlope:	Ptr to slope structure.
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 */
Int32 CDECL 
i_getIASlopeVertDy (
    IPState    *pIPState,
    IASlopePtr	pSlope,
    ScaledInt1616 *pVertDy)
{
    PROLOGUE;
	/* do the useful work */
    *pVertDy = pSlope->vertDy;
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_convolve():
 *        Does convolution with the given kernel at exactly one point.
 *	  Returns the convolution sum.
 *
 *	pImage:		Pointer to upper left corner of image INSIDE FRAME.
 *			This is a byte pointer, but is word aligned.
 *	bpl:		Image's bytes/line
 *	x:		x position (in pixels) of upper left corner of
 *			area to be convolved.
 *	y:		y position (in pixels) of upper left corner of
 *			area to be convolved.
 *	pKernel:	Pointer to convolution kernel
 *	pResult:	Pointer to Int32 that will contain the result
 *	
 *
 *	Return code:
 *		ia_depthNotSupported	bad kernel depth
 *		ia_invalidParm		the kernel supplied cannot be used
 *					to do a convolution at a single point
 *					(it's used for doing convolutions at
 *					an array of 3x3 points).
 *		ia_successful		good completion
 */
Int32 CDECL 
i_convolve (
    IPState    *pIPState,
    UInt8      *pImage,
    UInt32	bpl,
    Int32	x,
    Int32	y,
    IAKernelPtr	pKernel,
    Int32      *pResult)
{
	/* can we do a single convolution with this kernel? */
    PROLOGUE;
    if (pKernel->is3x3)
	return ia_invalidParm;

	/* compute the real image pointer */
    switch(pKernel->depth)
    {
        case 1:				/* 32 pixels per word */
	    pImage = pImage + y * bpl + ((x / 32) * 4);
	    x = x & 0x1F;		/* bit position within 32-bit word */
	    break;
        case 2:
	    pImage = pImage + y * bpl + ((x / 16) * 4);
	    x = (x & 0xF) << 1;		/* bit position within 32-bit word */
	    break;
        case 4:
	    pImage = pImage + y * bpl + ((x / 8) * 4);
	    x = (x & 0x7) << 2;		/* bit position within 32-bit word */
	    break;
	default:
	    return ia_depthNotSupported;
    }

	/* call the convolution routine for this kernel.  The args
	 * are a pointer to the word containing the upper left corner
	 * of the kernel, the x bit position, the number
	 * of words/line and a pointer to the kernel. */
    *pResult = pKernel->conv1Code((UInt32 *)pImage,
				  x, bpl/4, pKernel);

    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_checkArea():
 *        Convolves with the given kernel in at a 3x3 range of points (does
 *	  nine convolutions).  Returns the point at which the convolution
 *	  sum was the highest, but not higher than the kernel's upperThreshold.
 *	  Also returns the value of this sum.
 *
 *	pImage:		Pointer to upper left corner of image INSIDE FRAME.
 *			This is a byte pointer, but is word aligned.
 *	bpl:		Image's bytes/line
 *	pX:		Pointer to x position (in pixels) of upper left
 *			corner of center of area to be convolved.
 *			The upper left corner of the convolution kernel
 *			will be placed in a 3x3 set of points centered at
 *			this x position.  The original x position will be
 *			modified to point to the position at which the
 *			kernel found the biggest legal (less than
 *			pKernel->upperThreshold) sum.
 *	pY:		Pointer to incoming and outgoing y coordinate, 
 *			similar in function to pX.
 *	pKernel:	Pointer to convolution kernel
 *	pResult:	Pointer to Int32 that will contain the result
 *	
 *
 *	Return code:
 *		ia_depthNotSupported	bad kernel depth
 *		ia_invalidParm		the kernel supplied cannot be used
 *					to do a convolution at an array
 *					of 3x3 points (it's only good for
 *					a convolution at a single point).
 *		ia_successful		good completion
 */
Int32 CDECL 
i_checkArea (
    IPState    *pIPState,
    UInt8      *pImage,
    UInt32	bpl,
    Int32      *pX,
    Int32      *pY,
    IAKernelPtr	pKernel,
    Int32      *pResult)
{
    PROLOGUE;
    Int32	 x, y;			/* local versions of coords */
    UInt8	*pImageSaved;		/* value of pointer passed to
					 * the convolution code.  The
					 * difference between this pointer
					 * val and the one returned tells
					 * us where the best sum was. */
    Int32	 xSaved;		/* ditto for the x coordinate. */
    UInt32	 logPixPerByte;		/* the shift distance used to
					 * simulate division when computing
					 * number of pixels/byte */
    Int32	 ptrDiff;		/* holds the difference between
					 * pImage and pImageSaved so we can
					 * figure out what the convolution
					 * routine did to the pointer. */
	/* can we do a 3x3 convolution with this kernel? */
    if (!pKernel->is3x3)
	return ia_invalidParm;

	/* initialize the local versions of x and y */
    x = *pX;
    y = *pY;

	/* compute the real image pointer */
    switch(pKernel->depth)
    {
        case 1:				/* 32 pixels per word */
	    pImage = pImage + y * bpl + ((x / 32) * 4);
	    x = x & 0x1F;		/* bit position within 32-bit word */
	    logPixPerByte = 3;
	    break;
        case 2:
	    pImage = pImage + y * bpl + ((x / 16) * 4);
	    x = (x & 0xF) << 1;		/* bit position within 32-bit word */
	    logPixPerByte = 2;
	    break;
        case 4:
	    pImage = pImage + y * bpl + ((x / 8) * 4);
	    x = (x & 0x7) << 2;		/* bit position within 32-bit word */
	    logPixPerByte = 1;
	    break;
	default:
	    return ia_depthNotSupported;
    }

	/* save the original values of the image pointer and the x
	 * coordinate.  These will be adjusted by the convolution routine
	 * to tell us where the best point is.
	 */
    pImageSaved = pImage;
    xSaved = x;

	/* call the convolution routine for this kernel.  The args
	 * are the address of the pointer to the word containing the
	 * upper left corner
	 * of the kernel, the x coordinate as a ScaledInt1616, the number
	 * of words/line and a pointer to the kernel.  The last arg
	 * says to scan the 3x3 resulting sums from right-to-left,
	 * bottom to top.  This is slightly quicker than the alternative. */
    *pResult = pKernel->conv3Code((UInt32 **)&pImage,
				  &x, bpl/4, pKernel, 1);

	/* How to figure out where the best point is from the current values
	 * of pImage and x?  We first compute the difference between the
	 * new image pointer and the old one.  If the absolute value of
	 * this difference is bigger than 4, we must be on a new line.
	 */
    ptrDiff = (Int32) pImage - (Int32) pImageSaved;	/* new val - old val */
    if (ABS(ptrDiff) > 4)
    {
	if (ptrDiff < 0)		/* went up a line */
	{
	    *pY -= 1;
	    ptrDiff += bpl;		/* toss that difference */
	}
	else
	{				/* went down a line */
	    *pY += 1;
	    ptrDiff -= bpl;
	}

    }	/* end of accounting for differences in pImage */

	/* now, ptrDiff should be -4, 0, or +4.  Multiply this value
	 * by the number of pixels/byte and add it to *pX since it
	 * is one component of the number of pixels we've moved left
	 * or right */
    *pX += (ptrDiff << logPixPerByte);

	/* now divide the difference between x and xSaved by the number of
	 * bits/pixel and add that to *pX too.  That says how far over we
	 * moved in the word. 
	 */
/*
    *pX += shiftRightArithmetic((x - xSaved, 3-logPixPerByte);
*/
	/* What's this trash??  Well, acc issues warnings when you shift
	 * an integer right by a variable shift amount.  This is how we
	 * avoid doing that.  WEIRD!!!!
	 */
    switch (logPixPerByte)
    {
	case 1:
	    *pX += shiftRightArithmetic(((x & cCenterXBitMask) - xSaved), 3-1);
	    break;
	case 2:
	    *pX += shiftRightArithmetic(((x & cCenterXBitMask) - xSaved), 3-2);
	    break;
	case 3:
	    *pX += (x & cCenterXBitMask) - xSaved;
	    break;
    }

    NORMAL_RETURN;
}



/* *************** Convolution  Routines ********************* */



/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_compileGlyphTables():
 *        Compiles tables of combined kernel weights for use by 
 *	  i_extractGlyphRect.
 *
 *        Returns the number of bytes per set of tables
 *	  representing one row of kernel weights, the number of the
 *	  first active row, the number of active rows, the number of
 *	  the first active column, the number of active columns;
 *	  allocates and fills the compiled set of weights and sets a
 *	  pointer to point to the table.  Asks ip_genConvAsm to compile
 *	  the kernel into a set of machine code.  Only the dimensions
 *	  of the kernel really matter for the machine code generation.
 *
 *	pKernel:	Ptr to kernel structure
 *	allocator:	Ptr to function that is used to allocate memory
 *			for the table.  Function's single argument is
 *			the number of BYTES requested.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of malloc
 *			provided by WATCOM on the PC!!
 *	freer:		Ptr to function used to free memory in case of
 *			error.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of free
 *			provided by WATCOM on the PC!!
 *	
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	Can't generate tables for this
 *					image depth
 *		ia_nomem 		could not allocate memory
 *		ia_invalidParm		kernel weights added up to illegal
 *					values (too big)
 */
Int32 CDECL 
i_compileGlyphTables (
    IPState    *pIPState,
    IAKernelPtr	pKernel,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	),
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *))
{
    PROLOGUE;
    Int32	 kRow, kCol, kIndex;	/* kernel iteration vars */
    UInt32	 blockData;		/* non-shifted input data */
    UInt8	 data;			/* shifted input data */
    Int32	*pTableLocal;		/* ptr into table being created */
    Int32	 firstActiveRow, lastActiveRow, numActiveRows;
    Int32	 firstActiveCol, lastActiveCol, numActiveCols;
    Int32	 totalHorizBits;	/* total number of image bits required
					 * to compute the convolution
					 * information
					 * for one row of the kernel. */
    Int32	 tableEntriesPerRow;	/* number of table entries allocated to
					 * process one kernel row */
    Int32	 blockBits;		/* number of bits in the current
					 * horiz blk */
    Int32	 blockBitPos, curPixPos; /* position of MS bit of current block
					  * and pos of pix currently
					  * being inspected */
    Int32	 sum0, sum1, sum2;	/* table vals being composed */
    Int32	 pixMask;		/* used to inspect individual pixels */
    Int32	 totalPosSum, totalNegSum; /* sums used to check total
					    * kernel weights */

	/* initialize pointers so the right thing will happen in case
	 * of error.  We can't do all of these on the same line because
	 * the compiler complains about type mismatches. */
    pKernel->conv1Code = NULL;
    pKernel->conv3Code = NULL;
    pTableLocal = NULL;
    pKernel->compWeights = NULL;

	/* Check the image depth and other parms */
    if ((pKernel->depth != 1) && (pKernel->depth != 2) && (pKernel->depth != 4))
    {
        __error__ = ia_depthNotSupported;
	goto error;
    }

	/* first, figure out how many active rows and cols the
	 * kernel has.  Also check to see if the convolution
	 * sums could get too big.  Note that all image pixel values
	 * are positive, so the only way we can get big numbers
	 * is to have either all the positive kernel values have
	 * big image values OR all the negative kernel values correspond
	 * to big image values.  If both positive and negative kernel
	 * values correspond to non-zero image pixels, this tends to
	 * bring the total convolution value back towards 0. */
    firstActiveRow = pKernel->height;	/* greater than any row */
    firstActiveCol = pKernel->width;	/* greater than any column */
    lastActiveRow = lastActiveCol = -1;	/* less than any row, col */
    totalPosSum = totalNegSum = 0;
    for (kRow = 0, kIndex = 0; kRow < pKernel->height; kRow++)
    {
	for (kCol = 0; kCol < pKernel->width; kCol++, kIndex++)
	{
	    if (pKernel->weights[kIndex] != 0)
	    {
		if (kRow < firstActiveRow)
		    firstActiveRow = kRow;
		if (kCol < firstActiveCol)
		    firstActiveCol = kCol;
		if (kRow > lastActiveRow)
		    lastActiveRow = kRow;
		if (kCol > lastActiveCol)
		    lastActiveCol = kCol;
		if (pKernel->weights[kIndex] > 0)
		    totalPosSum += ((1 << pKernel->depth) - 1) *
				    pKernel->weights[kIndex];
		else
		    totalNegSum += ((1 << pKernel->depth) - 1) *
				    pKernel->weights[kIndex];
	    }
	}
    }

	/* if there's nothing in this kernel or the weights are too big, quit */
    if ((firstActiveRow == pKernel->height) || (totalPosSum >= 512) ||
	(totalNegSum <= -512))
    {
	__error__ = ia_invalidParm;
	goto error;
    }

	/* store these */
    pKernel->firstActiveRow = firstActiveRow;
    pKernel->numActiveRows = numActiveRows =
				lastActiveRow - firstActiveRow + 1;
    pKernel->firstActiveCol = firstActiveCol;
    pKernel->firstActiveColBit = firstActiveCol * pKernel->depth;
    pKernel->numActiveCols = numActiveCols =
				lastActiveCol - firstActiveCol + 1;

	/* How many image bits will be operated on per kernel row?
	 * This is numActiveCols * pKernel->depth if we're doing a 1x1
	 * convolution and (numActiveCols+2) * pKernel->depth if we're doing
	 * a 3x3 convolution.
	 */
    if (pKernel->is3x3)
	totalHorizBits = (numActiveCols+2) * pKernel->depth;
    else
	totalHorizBits = (numActiveCols) * pKernel->depth;

	/* Fill in the entries concerned with the total number of bits
	 * in each horizontal row.  The rowBitsMask convers all the input
	 * bits, but they are aligned with the LS bit at bit 2 of the
	 * word so they can be used to address words. */
    pKernel->numBitsPerRow = totalHorizBits;
    pKernel->rowBitsMask = leftMask[totalHorizBits-1] >>
				(32 - totalHorizBits - 2);

	/* Decide how big each row of the table will be.  There
	 * will be 256 four byte entries for every horizontal group of
	 * 8 horiz bits, then 4 *  2**(number of bits remaining)
	 * bytes for the rest.  If totalHorizBits is a multiple of
	 * 8, this will put 1 extra entry into each set of row tables.
	 * Big deal.
	 */
    tableEntriesPerRow = (totalHorizBits / 8) * 256 +
			 (1 << (totalHorizBits % 8));
    pKernel->bytesPerTableRow = 4 * tableEntriesPerRow;

	/* Allocate space for the big table */
    pTableLocal = (Int32 *) allocator(pIPState, numActiveRows *
				      (pKernel->bytesPerTableRow));
    if (pTableLocal == NULL)
    {
	__error__ = ia_nomem;
	goto error;
    }

	/* the pointer's ok, so give it to the returnee and set the
	 * value the table pointer will have at the bottom of the
	 * convolution loop */
    pKernel->compWeights = pTableLocal;
    pKernel->finalTablePtrVal = (UInt32) pTableLocal + 
				 (pKernel->bytesPerTableRow *
					pKernel->numActiveRows);

	/* compute the mask used to look at individual pixels.  This
	 * is the mask we use to look at pixels within a block of
	 * 1 pixels.  While the convolution of groups of 8-bit
	 * blocks will proceed from right to left, the order in
	 * which we look at pixels within a block when building
	 * the table doesn't really matter.  However, for the
	 * same reason that convolving from right to left is easier,
	 * inspecting the pixels within a block from right to
	 * left is easier too.
	 * Depth	Mask
	 *   1		0x01
	 *   2		0x03
	 *   4		0x0F
	 */
    pixMask = 0xFF >> (8 - pKernel->depth);

	/* Before we get into table making, it's time to say what's
	 * going on.  As stated above, the convolution routine will
	 * divide the input bit vector into groups of 8 bits starting
	 * from the right end with the remaining bits in a group at
	 * the left end.  Each group of 8 bits will be used as an 
	 * index into a table of pre-computed convolution
	 * values.  Well, this is trickier than it looks.  The
	 * last, leftmost, group of bits will be right-aligned
	 * in its byte when used as an index.  That means the last
	 * table is smaller than the others.
	 *
	 * There's another complication.  In the 3x3
	 * case, the rightmost input bit only figures into the
	 * rightmost sum.  The next bit figures in the rightmost two
	 * sums and so on.  In general, the choice of kernel values
	 * used to compute sums is a function of both the number
	 * of the input bit (relative to the first input bit in
	 * whatever group it was in) and which sum is being
	 * computed.
	 *
	 * We will structure the table-building code in four
	 * levels of loop.  The outermost loop goes through the
	 * rows of the kernel.  The next inner loop counts in units of
	 * horizontal blocks of inputs *from right to left*.
	 * The next loop inside counts through all possible input 
	 * values for that block given its size.  The innermost
	 * loop inspects the bits of each input value, combines them
	 * with the appropriate kernel weights and places them into
	 * table entries.
	 *
	 * So, we need some variables:
	 *  blockBitPos:	The number of the LS bit of the current
	 *			data group relative to the leftmost
	 *			bit (0) in the entire data group.  For example,
	 *			if there are 16 bits in the current data
	 *			group, blockBitPos = 15 during the entire
	 *			time the table for the first (rightmost)
	 *			byte is being generated.
	 *  curPixPos:		The number of the current input pixel
	 *			relative to the leftmost input pixel in
	 *			the entire horizontal group.  Used to
	 *			select kernel weights in the row.
	 *  blockBits:		The number of input bits in the current
	 *			block.  This ranges from 1 though 8.
	 *  blockData:		The current input value.
	 *  data:		Shifted version of current input value.
	 *  kIndex:		Index of leftmost weight in current kernel row
	 *  kRow:		Current kernel row
	 *  sum0, sum1, sum2:	The three sums that will go into a
	 *			table entry.  It's easier to compute them
	 *			individually, then insert the final sums
	 *			into the right places in the table entry.
	 *			If we're generating tables for a 1x1 kernel,
	 *			we only use sum0.
	 *  pixMask:		Used to inspect individual pixels
	 */

	/* skip over the top vacant kernel rows, if any. */
    kIndex = firstActiveRow * pKernel->width;

	/* skip over the leftmost vacant kernel columns, if any. */
    kIndex += firstActiveCol;

	/* loop through kernel rows */
    for (kRow = firstActiveRow;
	 kRow <= lastActiveRow;
	 kRow++, kIndex += pKernel->width)
    {
	    /* loop in units of blocks from right to left */
	for (blockBitPos = totalHorizBits-1;
	     blockBitPos >= 0;
	     blockBitPos -= 8)
	{
		/* loop through data values in this block.
		 * Left-align the data so we always start with
		 * the MS bit or pixel. */
	    blockBits = MIN(8, blockBitPos+1);
	    for (blockData = 0;
		 blockData <  (1UL << blockBits);
		 blockData++ )
	    {
		    /* inspect each data bit.  This lets us inspect the
		     * bits from right to left starting with the LS bit
		     * for any data width.  After each step, we shift data
		     * right.  When there are no more 1s left, we stop. */
		for (sum0 = sum1 = sum2 = 0,
			data = (UInt8) blockData,
			curPixPos = blockBitPos / pKernel->depth;
		     data != 0;
		     data >>= pKernel->depth, curPixPos--)
		{
			/* if there's no bit here, go to next pos */
		    if (!(data & pixMask))
			continue;

			/* Compute sum 0 for this pixel */
		    if (curPixPos < numActiveCols)
			sum0 += pKernel->weights[kIndex+curPixPos] *
				(Int32)(data & pixMask);

		    if (pKernel->is3x3)
		    {
			    /* Compute sum 1 for this pixel */
			if ((curPixPos >= 1) &&
			    (curPixPos < numActiveCols+1))
			    sum1 += pKernel->weights[kIndex+curPixPos-1] *
				    (Int32)(data & pixMask);

			    /* Compute sum 2 for this pixel */
			if (curPixPos >= 2)
			    sum2 += pKernel->weights[kIndex+curPixPos-2] *
				    (Int32)(data & pixMask);
		    }
		}  /* have computed sums for this block */

		if (pKernel->is3x3)
		    *pTableLocal = ( sum0	      << 2*cSumSpacing) |
				   ((sum1 & cSumMask) << 1*cSumSpacing) |
				   ( sum2 & cSumMask);
		else
		    *pTableLocal = sum0;

		pTableLocal++;	/* pt to next table entry */
	    } /* end of loop through vals in this block */
	} /* end of blocks of input data */
    }  /* end of kernel rows */

	/* if we get here, the tables must have been filled correctly.
	 * Call the routine that compiles kernel shapes into assembly
	 * language.  If there's no machine code for this platform,
	 * the function will put in a pointer to the C implementation. */
    __error__ = ip_genConvAsm (pIPState, pKernel, allocator, freer);

    if (__error__ == ia_successful)
	NORMAL_RETURN;
    else
        goto error;

	/* get here if something went wrong */
error:
    pKernel->firstActiveRow = 0xFFFFFFFFUL;
    pKernel->numActiveRows = 0UL;
    pKernel->firstActiveCol = 0xFFFFFFFFUL;
    pKernel->firstActiveColBit = 0xFFFFFFFFUL;
    pKernel->numActiveCols = 0UL;
    pKernel->bytesPerTableRow = 0UL;
	/* If this is Visual C++, turn off the warning that says we're
	 * casting a function pointer to a data pointer, though why
	 * Visual C++ thinks void * is a data pointer I'll never know.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    if (pKernel->conv1Code != NULL)
	freer(pIPState, (void *) pKernel->conv1Code);
    pKernel->conv1Code = NULL;
    if (pKernel->conv3Code != NULL)
	freer(pIPState, (void *) pKernel->conv3Code);
    pKernel->conv3Code = NULL;
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif
    if (pTableLocal != NULL)
	freer(pIPState, (void *) pTableLocal);
    pKernel->compWeights = NULL;
    return __error__;

}

#if 0
Int32	 errorRoutineCount;
#endif

/* 
 * i_convolveWhenLikePixels()
 *	This routine computes the summation of the convolution results of
 * 	the directional kernel at all places in the glyph whose pixels are
 *	greater than or equal to the center glyph (actually 4 bit/pixel
 *	images look for pixels greater than or equal to one less than
 *	the center pixel.  The routine is used when the 
 *	center convolution found a valid glyph but the directional
 *	was too low.
 *
 * Input:
 *	pImageUpperLeft		point to the upper left corner of the
 *				glyph
 *	wpl			words per line in the image
 *	pOneKernel		the directional kernel
 *	upperLeftX		bit offset to upper left pixel of the glyph
 *	convResult		pointer to (sum of) convolution result
 *
 * Returns:
 *	ia_successful		good completion
 *  
 * This procedure is called from i_extractGlyphRect when the center
 * finding kernel has found a glyph but the directional kernel was unable
 * to determine the direction because the value was below the threshold.
 * This seems to happen when we have a "boxy" glyph.  For example,
 *	0 0 0 0 0	or	0 0 0 0 0
 *	0 0 0 1 0		0 0 0 1 0
 *	0 0 1 1 0		0 0 1 1 1
 *	0 0 1 1 0		0 0 1 1 0
 *	0 0 0 0 0 		0 0 0 0 0
 * are example of 1 bit/pixel boxy glyphs.  The problem is that 
 * placing the center at several locations (usually 3) will yield the
 * same convolution value.  If we pick the "wrong" one the directional
 * result will be too low.
 *
 * We've tried several algorithms to find the direction.  Some don't 
 * work at all while others work for some image depths but not all depths.
 * The methods that we have tried are:
 * 1) Convolve with the directional kernel at all locations in the 
 * 3x3 area.  This method doesn't give a clear winner.  For example,
 * consider the convolution results from positions a and b:
 *	0 0 0 0 0
 *	0 a 0 1 0
 *	0 0 1 1 0
 *	0 b 1 1 0
 *	0 0 0 0 0
 * where a and b pixel values are 0.  Clearly, the absolute value
 * of the convolution results at both places are the same.  Unfortunately,
 * one indicates a \ glyph and the other indicates a / glyph.
 *
 * 2) Use a center kernel with different weights.  Emphasizes 
 * 8-connectedness.  Center value is largest with all other values 
 * around it the same.  This method still does not give one clear
 * center.  In the above glyph, placing the center on both on pixels along
 * the center row yield the same convolution result so which one do you
 * choose?
 *
 * 3) Use a center kernel that emphasizes 4-connectedness and then
 * use the directional kernel on that new center.  Experience with (2) 
 * led us to try a 4-connected kernel.  It works for the 1 bit/pixel
 * cases that we encounted but failed in the 2 bit/pixel cases.
 * Consider:
 *	0 0 0 1 1
 *	0 1 1 1 0
 *	1 2 2 1 0
 *	1 2 2 2 1
 *	0 0 2 2 1
 *	0 0 0 1 1 
 * Again, placing the center on more than one pixel will yield the
 * same convolution result so which one do you use?
 *
 * 4) Use the directional kernel at only the center locations that 
 * have the same value as the center that we've chosen.  I have no clue
 * why this one didn't work.  How embarassing.  I should have written this
 * comment as I was trying all the different algorithms.....
 *
 * 5)  At all pixels on the glyph whose values are greater than or
 * equal to the center that was chosen, use the directional kernel.
 * These kernels vote (adding together) to determine the center.
 *
 * 10-Jun-95.  The following comments describe the attempts to get  
 * 4x4 glyphs to derender.  Since 4x4 glyphs are separated by one white
 * space (instead of 2 like 5x5s), they are more likely to touch and
 * create black blobs as big as the glyphs themselves.  These black
 * blobs wreck havoc with the tracking code.  Sometimes it confuses the
 * code enough to pull it into the row above or below.  At other times
 * it just keeps us from finding the optimal center.  In addition, even
 * if we are on the optimal center, the extra black blobs may cause 
 * the direction to be found incorrectly.
 *
 * In the first attempt, the glyphs were left at 4x4 and various kernels
 * of size 4x4 and 5x5 were attempted.  For example, since there was so
 * much black, I tried a kernel which looked for the white border that is
 * supposed to be around each glyph.  This did not help the tracking.  I
 * also tried to cleanup the image by removing some of the black blobs.
 * This had a tendency to destroy some of the glyphs themselves.  However,
 * the biggest problem was the tracking.  With the glyphs at 4x4, a little
 * mistake was quickly magnified and we were off into another row.  
 *
 * The easiest way to solve the above tracking problem is to expand the
 * image, resulting in 8x8 glyphs.  The larger glyphs allow better tracking
 * since it takes more mistakes before the tracking is off into another
 * row.  This change helped derendering more than any others.  Other 
 * things that were tried:
 *
 * 1) 8x8 vs 9x9 kernels.  9x9 kernels work best.  Since these are really
 * 8x8 glyphs with a center that is spread out among several pixels, it
 * might make sense to have the kernel reflect this.  In practice, this
 * makes things worse.  It seems to alllow the center to move around too
 * much and then the directionals are incorrect.
 *
 * 2) Cleanup the image by removing the black blobs that connect the
 * glyphs.  This does not seem to help and in fact makes things worse
 * because it destroys some glyphs.  For example, let's say that you
 * try to find all the places where there is a diagonal of length 7 and
 * break that in the center to form 2 small diagonals.  This works as 
 * long as the longest diagonal is of length 7.  However, diagonals of
 * length greater than 7 eliminate several pixels and if the diagonal is
 * long enough (say it connects 3 glyphs) then huge hunks of the middle
 * glyph will be destroyed.
 *
 * 3) Try to constrain the tracking in i_extractGlyphRect based on the
 * slope.  Once the image has been expanded, tracking still sometimes
 * totally messes up and goes up or down into another row.  We tried to
 * eliminate this by constraining the amount that the center could move
 * up or down.  First, we needed a better slope than the original slope
 * found before the synch was located.  Using the synch we can find a 
 * better slope by averaging the slope from all the rows of the
 * rectangle (and then all the columns).  However, this seems to constain
 * the center movement too much.  There is still local movement that
 * needs to take place and by keeping the center within +/- of what the
 * slope would allow, that movement does not take place and we do not
 * set down on very good centers at all.  I also tried computing the 
 * slope more locally - based on the corners of the frame being derender.
 * This didn't help, however.
 *
 * 4) While the movement to the row above or below is a problem, the
 * more important problem was the single bit errors that were occurring
 * when a single glyph was incorrectly derendered.  Tracking was
 * staying on the correct row but a single bit in each row was incorrect.
 * These incorrect decisions were usually made when applying the
 * directional to the center yielded a low theshold and then 
 * i_convolveWhenLikePixels to try to fix the discrepancy.  The centers
 * in many of the cases were off (from the optimal center) by +/-1 and/or
 * +/- column.  That doesn't sound so bad but because the other glyphs 
 * were so close, the directional kernel frequently "saw" multiple 
 * glyphs when it was making a decision.  Two things help this a little.
 * First, make the kernels smaller.  The kernels remain at 9x9 but the
 * active elements in the kernel are pulled in slightly leaving 1 
 * rows and columns of 0 pad around the outside instead of the usual zero.
 * Second, make the diagonals slightly more blocky.  For example, the
 * kernel should look more like:
 *	0  0  0  0  0  0  0  0  0
 *	0 -2 -1 -1  0  1  1  2  0
 *	0 -1 -4 -2  0  2  4  1  0
 *	0 -1 -2 -3  0  3  2  1  0
 *	0  0  0  0  0  0  0  0  0
 *	0  1  2  3  0 -3 -2 -1  0
 *	0  1  4  2  0 -2 -4 -1  0
 *	0  2  1  1  0 -1 -1 -2  0
 *	0  0  0  0  0  0  0  0  0
 * Notice how the active elements in each of the quarters is a square -
 * slightly more blocky than the original.  Note that the current code
 * does not make these fixes.  They are important for these glyphs but
 * probably not for glyphs that were originally this size.
 *
 * 5) Another method used to fix the problem described in #4 above, 
 * was to modify i_convolveWhenLikePixels.  Many times even after 
 * executing the routine, the directional was still too low.  This occurred
 * because the center was slightly off and adjacent glyphs were still
 * getting noticed by the directional kernel.  Three things were tried 
 * to fix the problem.  First, instead of counting in the directional
 * result at all points in the 3x3 area that match the center pixel, 
 * use only the results in which the center kernel is between the
 * upper and lower thresholds.  This didn't help.  Second, this routine
 * is called with the revised center.  Try calling it with the original
 * center.  As I recall, this actually made things worse.  Third,
 * if the directional was still too low at the end of the routine, take
 * 2 averages.  One average of all the positive results and one average 
 * of all the negative results.  The higest average is the direction
 * of the glyph.  This didn't help either. In fact, as I recall it actually
 * made things worse.  The adjacent glyphs continue to influence the
 * directional kernel too much.
 */

/* PUBLIC */
Int32 CDECL 
i_convolveWhenLikePixels (
    IPState    *pIPState,
    UInt32     *pImageUpperLeft,
    UInt32	wpl,
    IAKernelPtr	pOneKernel,
    Int32	upperLeftX,
    Int32      *convResult)
{
PROLOGUE;
 UInt32		 centerPixel;	/* the value of the center pixel */
UInt32		 centerWord;	/* contains the center pixel of the 
				 * current convolution. */
Int32		 centerX;	/* bit position of pixel which is
				 * the center of the current convolution */
Int32		 currY;		/* loop counter */
UInt32		 mask;		/* current pixel mask */
UInt32		*pCenterWord;	/* pointer to word that contains the 
				 * center. 
				 */
UInt32		*pImageTop;  	/* top of image for convolution */
Int32	 	 trialX;	/* bit position of leftmost pixel for
				 * convolution. */

#if 0
    errorRoutineCount++;
#endif

	/* first get center pixel */
    pCenterWord = pImageUpperLeft + (pOneKernel->height / 2) * wpl;
    centerX = upperLeftX + (pOneKernel->width / 2) * pOneKernel->depth;
    if (centerX & 0x20)
	centerWord = pCenterWord[1];
    else centerWord = *pCenterWord;
    mask = (1 << pOneKernel->depth) - 1;
    centerPixel = (centerWord >> (32 - ((centerX & 0x1F) +
					pOneKernel->depth))) & mask;
	/* if this is a 4 bit/pixel image, allow the center pixel to
	 * vary a little more.
	 */
    if (pOneKernel->depth == 4)
	centerPixel = MAX(centerPixel - 1, 0);

	/* point to the top row of the 3x3 block */
    pCenterWord -= wpl;
    pImageTop = pImageUpperLeft - wpl;

    *convResult = 0;

	/* the starting pixel for the rightmost pixel in the block */
    centerX +=  pOneKernel->depth;
	/* the upper left corner of the glyph if the center is
	 * the rightmost pixel of the 3x3 block.
	 */
    trialX = upperLeftX + pOneKernel->depth;
    for (currY = 3; currY > 0; currY--)
    {
	    /* process pixels from right to left so that the first 
	     * center is the rightmost pixel of the 3x3 block.
	     */


	    /* get the word containing the current center.  hopefully,
	     * we'll only need to read it once.  shift it so the
	     * rightmost pixel starts at bit 0.
	     */
	if (centerX & 0x20)
	{
	    pCenterWord++;
	    centerX &= 0x1F;
	}
	centerWord = *pCenterWord >> (32 - (centerX + pOneKernel->depth));

	if (trialX & 0x20)
	{
            pImageTop++;
            trialX &= 0x1F;
	}
	    /* if this center has the same value as the real center
	     * it's convolution sum should count.
	     */
	if ((centerWord & mask) >= centerPixel)
	    *convResult += pOneKernel->conv1Code(pImageTop,
						 trialX, wpl,
					         pOneKernel);

	    /* now process the middle pixel of the 3x3 block */
	trialX -= pOneKernel->depth;
	if (trialX & 0x20)
	{
            pImageTop--;
            trialX &= 0x1F;
	}
	centerX -= pOneKernel->depth;
	if (centerX & 0x20)
	{
	    pCenterWord--;
	    centerX &= 0x1F;
	    centerWord = *pCenterWord >> 
			 (32 - (centerX + pOneKernel->depth));
	}
	else centerWord = centerWord >> pOneKernel->depth;
	if ((centerWord & mask) >= centerPixel)
	    *convResult += pOneKernel->conv1Code(pImageTop,
						 trialX, wpl,
					         pOneKernel);

	    /* now process the leftmost pixel of the block */
	trialX -= pOneKernel->depth;
	if (trialX & 0x20)
	{
            pImageTop--;
            trialX &= 0x1F;
	}
	centerX -= pOneKernel->depth;
	if (centerX & 0x20)
	{
	    pCenterWord--;
	    centerX &= 0x1F;
	    centerWord = *pCenterWord >> 
			 (32 - (centerX + pOneKernel->depth));
	}
	else centerWord = centerWord >> pOneKernel->depth;
	if ((centerWord & mask) >= centerPixel)
	    *convResult += pOneKernel->conv1Code(pImageTop,
						 trialX, wpl,
					         pOneKernel);

	    /* get ready for the next row */
	centerX += pOneKernel->depth + pOneKernel->depth;
	trialX += pOneKernel->depth + pOneKernel->depth;
	pCenterWord += wpl;
	pImageTop += wpl;
    }
    NORMAL_RETURN;
}


/* This is a publically defined routine */
/* PUBLIC */
/*
 *  i_extractGlyphRect():
 *        Decodes a rectangular area of glyphs.  Stops when the entire
 *	  rectangle has been explored.
 *
 *        Returns the bit vectors in *pOutWords and *pValidWords.  The bits
 *	  in *pOutWords are the polarities of the glyphs.  The bits in
 *	  *pValidWords indicate which bits in *pOutWords are valid (1 => valid).
 *
 *	pImageRow:	Pointer to bitmap containing glyphs.  Points to the
 *			32 bit word containing the upper left corner of the
 *			first glyph.  NOTE: if direction = -1,
 *			this points to the word containing the upper left
 *			corner of the lower right glyph in the rectangle.  
 *			We assume that the application has gone to some 
 *			effort to find the exact starting point every time 
 *			this routine is called.
 *	rowX:		A ScaledInt1616, this is the X coordinate of the 
 *			upper left corner of the first
 *			glyph relative to left end of 32-bit
 *			word pointed to by pImageRow.  As with pImageRow,
 *			this parameter always points to the first glyph
 *			to be decoded, independent of direction.
 *	rowY:		A ScaledInt1616, This is a fractional value only.
 *			The integer portion of rowY MUST be 0 since
 *			pImageRow points to the first glyph to be 
 *			decoded.  This is used to contain fractional
 *			amount that we over or under shot by picking
 *			the row and is really only important if 
 *			self-clocking is on.
 *	wpl:		Number of 32-bit words/line in image
 *	pSlope:		Ptr to IASlope structure containing the slope of the
 *			glyph region
 *	glyphRectX:	X dimension of glyph rect in units of glyphs
 *	glyphRectY:	Y dimension of glyph rect in units of glyphs
 *	direction:	When equal to -1, decode glyphs from right-to-left,
 *			bottom-to-top.  When equal to +1, decode glyphs
 *			left-to-right, top-to-bottom.  These are the only
 *			legal values.
 *	pCenterKernel:	Ptr to IAKernel structure.  This is the kernel
 *			used to find the centers of glyphs.
 *	pOneKernel:	Ptr to IAKernel structure.  This is the kernel
 *			used to decide whether the glyph represents a
 *			0 or a 1.
 *	xAlternatingOffset: Sometimes alternating rows of glyphs are offset
 *			from each other by glyphWidth/2.  This parameter
 *			contains +glyphWidth/2 if the rows are offset and 
 *			the first movement up or down implies a positive 
 *			(to the right) x movement.
 *			It contains -glyphWidth/2 if the first vertical 
 *			movement requires a jag to the left.  If the rows 
 *			are not offset, this parameter must contain a 0.
 *			It is always used to form the next glyph position.
 *			Note that this parameter is independent of the
 *			value of direction.  Moving up or down in the first
 *			move requires the same x offset.
 *	selfClocking:	If this BOOLEAN is true, the next glyph position is
 *			calculated relative to the center position found
 *			for the current glyph.  Otherwise, the next glyph's
 *			position is calculated relative to the originally
 *			estimated position.
 *	pOutWords:	Ptr to vector of words containing the results
 *			of glyph convolution.  Each word contains 32 single
 *			bit results.  The results are recorded in the words
 *			from the MS to LS end if direction is +1.
 *			They are recorded the other way if direction
 *			is -1.  The vector can contain bits in positions
 *			where the validity vector says there are no glyphs.
 *			This happens when a valid glyph center was found,
 *			but the 0/1 kernel didn't find a glyph of sufficient
 *			strength.  The bit recorded is our best guess as
 *			to the sign of the glyph.  Higher level software
 *			may want to resort to these guesses.
 *	pValidWords:	Ptr to vector of words containing the results
 *			of glyph convolution.  A bit in this vector is set (1)
 *			if the corresponding bit in pOutWords is valid.
 *			If the corresponding bit in pOutWords is invalid
 *			(no glyph was found at the location), the bit in
 *			pValidWords is 0.
 *			Each word contains 32 single bit booleans.  The
 *			results are recorded in the words from the MS to LS
 *			end if direction is +1.
 *			They are recorded the other way if direction
 *			is -1.
 *	pDebugImageRow:	Pointer to an image where we will write a pixel for
 *			each center position if DEBUG_I_GLYPH is set.  This
 *			image must have the same number of bytes/line as
 *			the real image.  We will write a single bit in the
 *			MS bit of the pixel containing the center.
 *	nextXKeep	used when keeping the (x, y) coordinates of a
 *			glyph into pXYCoords.  this variable indicates
 *			how often to put the x value into the array.
 *	nextYKeep	used when keeping the (x, y) coordinates of a
 *			glyph into pXYCoords.  this variable indicates
 *			how often to put the y value into the array.
 *			nextXKeep and nextYKeep can be different values.
 *			either BOTH nextXKeep and nextYKeep can be 0, in
 *			which case the coordinates are not stored (and
 *			pXYCoords can be NULL, or both nextXKeep and
 *			nextYKeep are non-zero.
 *	pXYCoords	pointer to the array which will contain the
 *			(x, y) coordinates of certain glyphs.  the 
 *			coordinates are stored first x and then y.  the
 *			first glyph in the rectangle is the first glyph
 *			stored into the array.	
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		couldn't allocate internal buffers
 *		ia_angleTooLarge	horizDy was too large (slope of
 *					image containing glyph was too big)
 *					
 */
Int32 CDECL 
i_extractGlyphRect (
    IPState    *pIPState,
    UInt32     *pImageRow,
    ScaledInt1616 rowX,
    ScaledInt1616 rowY,
    UInt32	wpl,
    IASlopePtr	pSlope,
    Int32	glyphRectX,
    Int32	glyphRectY,
    Int32	direction,
    IAKernelPtr	pCenterKernel,
    IAKernelPtr	pOneKernel,
    Int32	xAlternatingOffset,
    UInt32	selfClocking,
    UInt32     *pOutWords,
    UInt32     *pValidWords,
    UInt32     *pDebugImageRow,
    Int32	keepEachX,
    Int32	keepEachY,
    Int32      *pXYCoords)
{

#define cMaxYOffset	8		/* can't move up/down more than
					 * this many lines when moving
					 * horizontally from one glyph
					 * to the next. */
#define	cLogMaxYOffset 3		/* used in lieu of a multiply */

	/* defines a macro that generates a single bit mask */
#define BITMASK(bitnum)	((UInt32) 0x80000000 >> bitnum)

    PROLOGUE;
    UInt32	 nextOutBit;		/* number of next output bit in
					 * outputWord. */
    ScaledInt1616 curX, curY;		/* current value of x and y within
					 * a row of glyphs. */
    UInt32	*pImageCol;		/* ptr to current glyph position
					 * within a row. */
    UInt32	*pImageCenter;		/* like pImageCol only modified to
					 * account for the difference between
					 * the estimated center glyph position
					 * and the real center position */
    Int32	 centerXBit;		/* holds offset into word pointed to
					 * by pImageCenter. */
    Int32	 xToBitShift;		/* used to convert x position
					 * to centerXBit and back */
    UInt32	 xToBitMask;		/* used to clear lower bits when
					 * converting curX to centerXBit */
    UInt32	 curXModMask;		/* used to clear upper bits when
					 * setting curX back to the legal
					 * range of values (ones that point
					 * within a 32-bit word) */
    Int32	 yAddrOffset[(2*cMaxYOffset)+1]; /* array of y addr offsets.
					 * This is used to move up/down when
					 * moving "horizontally" to the next
					 * glyph when the block of glyphs
					 * is slightly skewed.  The address
					 * offset for moving down by n lines is
					 * found in array location
					 * n + cMaxYOffset */
    Int32	 col;			/* current glyph column */
    Int32	 convResult;		/* Holds the sum returned by the
					 * convolution routines. */
    ScaledInt1616 horizDx, horizDy, vertDx, vertDy; /* negatable versions
					 * of glyph spacing values */
    Int32	 rowYOffset;		/* number of lines from estimate to
					 * where the first glyph in the
					 * row was found.  used for 
					 * vertical self clocking.
					 */
    Int32	 rowXOffset;		/* number of pixels from estimate 
					 * to where the first glyph in the
					 * row was found.  used for 
					 * vertical self clocking.
					 */
    Int32	 xCoordChange;		/* indicates how much the current
					 * glyph has changed x position
					 * from the start of the rect.
					 * in pixel quantities.
					 */
    Int32	 yCoordChange;		/* indicates how much the current
					 * glyph has changed y position
					 * from the start of the rect.
					 * in pixel quantities.
					 */
    Int32	 xRowChange;		/* indicates how much the start of 
					 * the current row has changed x
					 * position from the start of the
					 * rect.  in pixel quantities.
					 */
    Int32	 yRowChange;		/* indicates how much the start of 
					 * the current row has changed y
					 * position from the start of the
					 * rect.  in pixel quantities.
					 */
    Int32	 nextXKeep;		/* when col becomes this value
					 * coordinates will be stored
					 * in pXYCoords.
					 */
    Int32	 nextYKeep;		/* when glyphRectY becomes this
					 * value coordinates will be stored
					 * in pXYCoords.
					 */
#if DEBUG_I_GLYPH
    Int32	debugCenterXBit;	/* like centerXBit, only points to
					 * the center of glyph instead of the
					 * upper left corner */
    UInt32	*pDebugImageCol;	/* like pImageCol in debug image */
    UInt32	*pDebugImageCenter;	/* like pImageCenter in debug image */

	/* Since we want to write dots in the glyph centers, move
	 * pDebugImageRow down by the right number of rows */
    for (col = (pCenterKernel->height)/2; col > 0; col--)
	pDebugImageRow += wpl;
#else
    pDebugImageRow = pDebugImageRow;	/* so compiler doesn't complain */
#endif

	/* if the slope of the glyph region is too big, quit.  We get
	 * the absolute value of the integer part of pSlope->horizDy, then add
	 * 2 since we could move up/down by 1 if we're self-clocking and
	 * round-off could move us by 1 more. */
    if ( (ABS(shiftRightArithmetic(pSlope->horizDy, 16)) + 2) > cMaxYOffset )
	return ia_angleTooLarge;

	/* fill in yAddrOffsets.  These are used when we're moving
	 * horizontally (mostly).  They're not used to move vertically
	 * in the glyph array.  That means this array can be fairly small
	 * and not a function of the vertical glyph spacing.
	 * We're using convResult as a temp so the compiler won't be 
	 * tempted to make another variable just for this loop.
	 */
    for (col = 2*cMaxYOffset, convResult = (wpl << cLogMaxYOffset);
	 col >= 0;
	 col--, convResult -= wpl)
    {
	yAddrOffset[col] = convResult;
    }

	/* set up constants that depend on the depth.  These are used to
	 * convert between x displacements expressed in pixels and those
	 * expressed in bits.  We can use the depth in either pCenterKernel or
	 * pOneKernel, they should be the same.
	 * xToBitShift is used to convert the integer part of curX to
	 * the equivalent bit offset.  For 1 bit/pixel, we shift off all
	 * of the fraction bits since the integer part of curX equals the
	 * number of bits.  For 2 bits/pixel, the number of bits = 2 * integer
	 * part of curX, so we shift right by 15 and mask off the bottom bit
	 * (xToBitMask has a 0 in the lsb).  Also, when we test to see if
	 * curX has gotten too big or small when hopping from glyph to glyph,
	 * we must remember that the range of legal values of curX is a
	 * function of depth.  For example, at 4 bits/pixel, the integer
	 * part of curX must between 0 and 7.  If it isn't, we should adjust
	 * the current addresss pointer and lop off the upper bits of curX.
	 * The value of curXModMask covers the bits of curX that are allowed
	 * to be set.
	 */
    if (pCenterKernel->depth == 1)
    {
	xToBitShift = 16;
	xToBitMask = 0xFFFFFFFF;
	curXModMask = 0x1FFFFF;		/* integer part of curX is in [0..31] */
    }
    else
    if (pCenterKernel->depth == 2)
    {
	xToBitShift = 15;
	xToBitMask = 0xFFFFFFFE;
	curXModMask = 0xFFFFF;		/* integer part of curX is in [0..15] */
    }
    else
    {
	xToBitShift = 14;
	xToBitMask = 0xFFFFFFFC;
	curXModMask = 0x7FFFF;		/* integer part of curX is in [0..7] */
    }

	/* if the "direction" parameter is negative, negate horizDx, horizDy,
	 * vertDx and vertDy.  Also point to the end of the output
	 * and valid bit vectors.
	 */
    if (direction < 0)
    {
	horizDx = -pSlope->horizDx;
	horizDy = -pSlope->horizDy;
	vertDx = -pSlope->vertDx;
	vertDy = -pSlope->vertDy;
	    /* add the correct number of words to pOutWords and pValidWords
	     * to point to the last bit that we will add.  we must 
	     * subtract 1, to get an offset which starts at 0.
	     */
	pOutWords += ((glyphRectX * glyphRectY - 1)/32);
	pValidWords += ((glyphRectX * glyphRectY - 1)/32);
	nextOutBit = (glyphRectX * glyphRectY - 1) % 32;
    }
    else
    {
	horizDx = pSlope->horizDx;
	horizDy = pSlope->horizDy;
	vertDx = pSlope->vertDx;
	vertDy = pSlope->vertDy;
	nextOutBit = 0;			/* point to MS output bit */
    }

    rowXOffset = rowYOffset = 0;

	/* initialize the first word of the valid and output bit vectors. */
    *pOutWords = *pValidWords = 0;

	/* These variables keep the x and y pixel offset from the beginning
	 * of the image.  initialize them now.  eventually they will be
	 * written into the pXYCoords array. We need 2 variables, 
	 * yRowChange and yCoordChange.  yCoordChange keeps track of the
	 * y movement along a row.  however, we still need something that
	 * resets at the beginning of the row since how we ended the
	 * previous row is no help to the location at which we start the
	 * next row; hence, yRowChange and xRowChange;
	 */
    xRowChange = xCoordChange = shiftRightArithmetic(rowX, 16);
    yRowChange = yCoordChange = 0;
 	/* when should we write into the pXYCoords array next? */
    if ((keepEachX != 0) && (keepEachY != 0))
    {
    	nextYKeep = glyphRectY;
    	nextXKeep = glyphRectX;
	    /* if we are going from bottom to top, left to right we need
	     * to point to the end of xyCoords array to be consistent
	     * with pOutWords and pValidWords.  note that we always
	     * keep the first coordinate in a row.  the amount we keep
	     * then depends on keepEachX; hence, the calculation below.
	     */
	if (direction < 0)
	    pXYCoords += ((((glyphRectX - 1) / keepEachX + 1) *
			   ((glyphRectY - 1) / keepEachY + 1)) * 2) - 2;
    }
    else
    {
	nextYKeep = -1;
	nextXKeep = -1;
    }

	/* loop through the rows and columns of the glyph rectangle */
    for (; glyphRectY > 0; glyphRectY--)
    {

#if DEBUG_I_GLYPH
	pDebugImageCol = pDebugImageRow;
#endif

	for (col = glyphRectX, pImageCol = pImageRow, curX = rowX, curY = rowY;
	     col > 0;
	     col--, nextOutBit += direction)
	{

		/* If we've reached a word boundary in the output bit
		 * vectors, store the local versions and pt to the 
		 * next word in each. */
	    if (nextOutBit & 0x20)	/* catches both +32 and -1 */
	    {
		pOutWords += direction;	/* advance the pointers */
		pValidWords += direction;
		nextOutBit &= 0x1F;	/* if nextOutBit was -1, it is now
					 * set to 31. If it was 32, it's
					 * now set to 0. */
		*pOutWords = *pValidWords = 0;
	    }

		/* find the best center location for this glyph
		 * within a 3x3 area centered at the estimated position.
		 * Pass in the values of the estimated center position
		 * in pImageCenter and centerXBit.  These will be updated
		 * to point to the word and offset of the real center
		 * position by the convolution code. */
	    pImageCenter = pImageCol;
	    centerXBit = (curX >> xToBitShift) & xToBitMask;
	    convResult = pCenterKernel->conv3Code(&pImageCenter, &centerXBit,
						  wpl, pCenterKernel,
						  nextOutBit);

		/* if the score wasn't high enough, skip this glyph */
	    if (convResult < pCenterKernel->lowerThreshold)
	    {
	    	if ((nextYKeep == glyphRectY) && (nextXKeep == col))
	    	{
		    *pXYCoords = xCoordChange;
		    *(pXYCoords+1) = yCoordChange;
		    pXYCoords += direction + direction;
		    nextXKeep -= keepEachX;
	    	}
		xCoordChange += shiftRightArithmetic((curX & 0xFFFF) + horizDx, 16);
		curX += horizDx;
	    }
	    else
	    {

		    /* for vertical self clocking. keep offset of "real" 
		     * start of row.
		     */
	        if (col == glyphRectX)
		{
		    rowYOffset = shiftRightArithmetic(centerXBit,
						      cVerticalRightShift);
		    rowXOffset = 
			shiftRightArithmetic((centerXBit <<
                                              cHorizontalLeftShift),
                                             cHorizontalRightShift);
		}


#if DEBUG_I_GLYPH	/* debugging code to print where it found the centers */
		pDebugImageCenter = pDebugImageCol +
				    (((UInt32) pImageCenter-(UInt32) pImageCol)/4);
		debugCenterXBit = (centerXBit & cCenterXBitMask) +
			(((pCenterKernel->width)/2)<<(16-xToBitShift));
		if (debugCenterXBit < 32)
		    pDebugImageCenter[0] |= BITMASK(debugCenterXBit);
		else
		    pDebugImageCenter[1] |= BITMASK((debugCenterXBit-32));
#endif
    
		    /* there's a valid center.  Figure out whether it's a
		     * 0 or a 1.  The center's convolution code has already updated
		     * pImageCenter and centerXBit to point to the best position
		     * to decode this glyph.
		     */
		convResult =
		    pOneKernel->conv1Code(pImageCenter,
					  (centerXBit & cCenterXBitMask),
					  wpl, pOneKernel);
    
		    /* if the absolute value of the result is big enough,
		     * mark it valid.  As stated above, we'll use the glyph's
		     * sign to mark the data bit in any case.  This will
		     * let higher level software try to guess glyph polarities
		     * if it has to.
		     */
		if (ABS(convResult) >= pOneKernel->lowerThreshold)
		    *pValidWords |= BITMASK(nextOutBit);
		else
		{
		    i_convolveWhenLikePixels (pIPState, pImageCenter, 
					      wpl, pOneKernel,
					      (centerXBit & cCenterXBitMask),
					      &convResult);
		    if (ABS(convResult) >= pOneKernel->lowerThreshold)
			*pValidWords |= BITMASK(nextOutBit);
		}
    
		    /* Mark the glyph's polarity.  Yes, we do an average
		     * of 1.5 memory writes for EVERY glyph, but how could 
		     * we do less?  If we add local variables to hold words
		     * to be written to memory, the compiler will keep them
		     * in memory anyway.  Besides, the convolution ops are
		     * probably doing some local var writes.
		     */
		if (convResult > 0)
		    *pOutWords |= BITMASK(nextOutBit);
    
		    /* if the selfClocking parm was set, advance from the
		     * position of the glyph found.  pImageCenter has already
		     * been adjusted for the change in y position, if any.
		     * Since curY is only adjusted relative to the position
		     * of pImageCol, we don't need to change it to account
		     * for changes already reflected in pImageCol.
		     * If we're not selfClocking, advance from the estimated
		     * position.
		     * We replace the integer part of curX with an integer
		     * derived from centerXBit since that's where curX should
		     * have been pointing.
		     */
		if (selfClocking)
		{
		    pImageCol = pImageCenter;
#if DEBUG_I_GLYPH
		    pDebugImageCol = pDebugImageCenter;
#endif
			/* add in the amount due to selfclocking before
			 * writing the coordinates for the current glyph.
			 */
		    yCoordChange += 
			shiftRightArithmetic(centerXBit,
				 	     cVerticalRightShift);
		    xCoordChange += 
			shiftRightArithmetic((centerXBit <<
					      cHorizontalLeftShift),
					     cHorizontalRightShift);
	    	    if ((nextYKeep == glyphRectY) && (nextXKeep == col))
	    	    {
		        *pXYCoords = xCoordChange;
		        *(pXYCoords+1) = yCoordChange;
		        pXYCoords += direction + direction;
		        nextXKeep -= keepEachX;
	    	    }
		    xCoordChange += shiftRightArithmetic((curX & 0xFFFF) + 
							  horizDx,
				                         16);
		    curX = (((centerXBit & cCenterXBitMask) << xToBitShift) 
			    | (curX & 0xFFFF))
			   + horizDx;
		}
		else
		{
		    curX += horizDx;
	    	    if ((nextYKeep == glyphRectY) && (nextXKeep == col))
	    	    {
		        *pXYCoords = xCoordChange;
		        *(pXYCoords+1) = yCoordChange;
		        pXYCoords += direction + direction;
		        nextXKeep -= keepEachX;
	    	    }
		    xCoordChange += shiftRightArithmetic(curX, 16);
		}
	    }
	    curY += horizDy;

		/* Advance the pointer by the amount of the y displacement
		 * and the x displacement.  There's often no change in the
		 * y displacement, so we skip changing the y if it's not
		 * necessary.
		 */
	    if (curY & 0xFFFF0000)
	    {
		pImageCol += yAddrOffset[(shiftRightArithmetic(curY, 16))+
					 cMaxYOffset];
#if DEBUG_I_GLYPH
		pDebugImageCol += yAddrOffset[(shiftRightArithmetic(curY, 16))+
						cMaxYOffset];
#endif
		yCoordChange += shiftRightArithmetic(curY, 16);
		curY &= 0xFFFF;		/* get rid of integer portion */
	    }

		/* if we crossed a word boundary, change the image pointer.
		 * The adjustment amount equals the number of 32-bit words
		 * crossed.  This equals curX right shifted by xToBitShift
		 * (to get rid of the fraction bits) + 5 (to get rid
		 * of the bits within a word).
		 */
	    if (curX & ~curXModMask)
	    {
		pImageCol += shiftRightArithmetic(curX, (xToBitShift+5));
#if DEBUG_I_GLYPH
		pDebugImageCol += shiftRightArithmetic(curX, (xToBitShift+5));
#endif
		    /* lop off the bits above the word position */
		curX &= curXModMask;
	    }
	}	/* end of loop through glyph columns */

	if (nextYKeep == glyphRectY)
	{
#if 0
	    if (nextXKeep == col)
	    {
	    	*pXYCoords = xCoordChange;
	    	*(pXYCoords+1) = yCoordChange;
	   	 pXYCoords += direction + direction;
	    }
#endif
	    nextYKeep -= keepEachY;
    	    nextXKeep = glyphRectX;
	}

	    /* move up/down one row */
	if (selfClocking)
	{
	    rowY += vertDy + (rowYOffset << 16);
	    pImageRow += (shiftRightArithmetic(rowY, 16)) * wpl;
#if DEBUG_I_GLYPH
	    pDebugImageRow += (shiftRightArithmetic(rowY, 16)) * wpl;
#endif
	    yRowChange += shiftRightArithmetic(rowY, 16);
	    yCoordChange = yRowChange;

	    xRowChange += shiftRightArithmetic((rowX & 0xFFFF) + vertDx +
					        xAlternatingOffset, 16) +
			  rowXOffset;
	    xCoordChange = xRowChange;
	    rowX += vertDx + xAlternatingOffset + (rowXOffset << 16);
	    xAlternatingOffset = -xAlternatingOffset;
	    pImageRow += shiftRightArithmetic(rowX, (xToBitShift+5));
#if DEBUG_I_GLYPH
	    pDebugImageRow += shiftRightArithmetic(rowX, (xToBitShift+5));
#endif
	    rowYOffset = rowXOffset = 0;
	}
	else
	{
	    rowY += vertDy;
	    pImageRow += shiftRightArithmetic(rowY, 16) * wpl;
#if DEBUG_I_GLYPH
	    pDebugImageRow += shiftRightArithmetic(rowY, 16) * wpl;
#endif
	    yRowChange += shiftRightArithmetic(rowY, 16);
	    yCoordChange = yRowChange;

	    xRowChange += shiftRightArithmetic((rowX & 0xFFFF) + vertDx +
					        xAlternatingOffset, 16);
	    xCoordChange = xRowChange;
	    rowX += vertDx + xAlternatingOffset;
	    xAlternatingOffset = -xAlternatingOffset;
	    pImageRow += shiftRightArithmetic(rowX, (xToBitShift+5));
#if DEBUG_I_GLYPH
	    pDebugImageRow += shiftRightArithmetic(rowX, (xToBitShift+5));
#endif
	}

	rowY &= 0xFFFF;			/* get rid of integer part */
	rowX &= curXModMask;

    }	/* of going through all the glyph rows */

    NORMAL_RETURN;
}

/*
 * i_glyphThresh()
 *	This procedure is used to threshold an image that is believed
 *	to contain glyphs from 8bpp to 1bpp show that the rectangle 
 * 	finding code can be ran over the resulting image.  This procedure
 *	uses an adaptive threshold and was intended to be used on
 * 	600dpi images which are frequently too light or too dark to
 * 	use a straight 128 threshold.
 *
 * Parameters:
 *	pGray		pointer to the 8bpp gray source image
 *	w		width of the source/dest image in pixels
 *	h		height of the source/dest image
 *	gBpl		number of bytes per line of source
 *	pDest		pointer to the destination image
 *	dBpl		bytes per line in the destination image
 *
 * Returns:
 *	ia_successful	if threshold succeeded
 *	ia_noMem	if could not allocate memory
 */
Int32  CDECL 
i_glyphThresh (
    IPState    *pIPState,
    UInt8      *pGray,
    Int32	w,
    Int32	h,
    Int32	gBpl,
    UInt8      *pDest,
    Int32	dBpl)
{
PROLOGUE;
 UInt8		 currThresh;
UInt32		 lastWordMask[32] =     
    {0xFFFFFFFF, 0x80000000, 0xC0000000, 0xE0000000, 0xF0000000,
                 0xF8000000, 0xFC000000, 0xFE000000, 0xFF000000,
                 0xFF800000, 0xFFC00000, 0xFFE00000, 0xFFF00000,
                 0xFFF80000, 0xFFFC0000, 0xFFFE0000, 0xFFFF0000,
                 0xFFFF8000, 0xFFFFC000, 0xFFFFE000, 0xFFFFF000,
                 0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00, 0xFFFFFF00,
                 0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0, 0xFFFFFFF0,
                 0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE};
UInt32		 mask;
UInt32		*repThreshTable;
Int32            row;
Int32		 bytesInSource;

    /* get rid of the frame */
    pGray += (gBpl << cLogFrameBits) + 8*cFrameBytes;
    pDest += (dBpl << cLogFrameBits) + cFrameBytes;
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;

    /* this table is used whenever we are not adjusting the threshold
     * for each pixel, just using the same threshold for all the pixels
     * in the source word.  we would like to do all these at once so we 
     * need a buffer bit between the pixels.  hence, we repeat the 
     * threshold in all pixels, shift it right by 1, and put a one in 
     * the buffer bit (ms bit of each pixel) so that a subtract will
     * not influence the adjacent pixel.
     */
    if ((repThreshTable = (UInt32 *)MALLOC(256*sizeof(UInt32))) == NULL)
	return ia_nomem;
    for (row = 0; row < 256; row++)
    {
	repThreshTable[row] = (row << 8) | row;
	repThreshTable[row] = (repThreshTable[row] << 16) | 
			       repThreshTable[row];
	repThreshTable[row] = (repThreshTable[row] >> 1) | 0x80808080;
    }

    /* we are going to ignore the last 2 bits of the source gray image.
     * why 2?  well, we want to compute the average of 4 source pixels
     * at a time so we going to have to loose some bits or we'll overflow.
     * 4 pixels are going into the average, so we'll need 2 extra bits
     * to hold the sum.
     */
    currThresh = 128;
    mask = lastWordMask[w & 0x1F];
    /* the number of bytes in each source line given that the width is
     * padded to a 32 bit boundary.  remember this is an 8 bpp image....
     */
    bytesInSource = ((w + 3) / 4) * 4;
    /* for each source line */
    for (row = 0; row < h; row++, pGray += gBpl, pDest += dBpl)
    {
	currThresh = ip_glyphThreshLine (pIPState, pGray, w,
					 gBpl, pGray+bytesInSource,
				         pDest, mask, repThreshTable, 
				         currThresh);
    }

    FREE(repThreshTable);
    NORMAL_RETURN;
}





/*

$Log: i_glyph.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.11  1998/01/21 21:24:21  ddavies
i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v

Remove cast in FREE invocation.

Revision 1.10  1998/01/13 23:47:55  ddavies
i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.9  1997/07/11 01:48:44  ddavies
i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.8  1997/05/10 01:26:41  jcraig
jcraig on Fri May 9 18:25:58 PDT 1997

i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v

remove (#if 0) use of errorRoutineCount which was only used for
debugging purposed and should never have remained in the code.

Revision 1.7  1996/06/10 22:36:25  jcraig
jcraig on Mon Jun 10 15:35:40 PDT 1996

i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v

add comments to describe the work on i_extractGlyphRect to get 4x4
glyphs to derender correctly.

Revision 1.6  1995/12/13 02:43:59  ejaquith
i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/10/18  03:53:08  ddavies
 * i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.4  1995/03/27  23:58:14  jcraig
 * jcraig on Mon Mar 27 15:57:34 PST 1995
 *
 * i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v
 *
 * add new procedure, i_glyphThresh, which is used to threshold an 8bpp
 * 600dpi image which may contain glyphs.
 *
 * Revision 1.3  1994/04/12  04:22:52  jcraig
 * jcraig on Mon Apr 11 21:21:35 PDT 1994
 *
 * i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v
 *
 * when doing vertical selfclocking (at the start of a row) use the
 * x change of the first glyph in the previous row as well as the y
 * change.
 *
 * Revision 1.2  1994/03/10  01:24:08  jcraig
 * jcraig on Wed Mar 9 17:22:01 PST 1994
 *
 * i_glyph.c -> /product/ipcore/ipshared/src/RCS/i_glyph.c,v
 *
 * add new paramters to i_extractGlyphRect: nextXKeep, nextYKeep and
 * pXYCoords.  if nextXKeep and nextYKeep are not zero, each
 * nextXKeep in the horizontal direction and nextYKeep in the vertical
 * direction - the coordinates of the glyph is stored in pXYCoords.
 *
 * Revision 1.1  1994/01/22  03:47:31  ddavies
 * Initial revision
 *
 * Revision 1.31  1993/12/02  19:20:22  ddavies
 * cglyph.c -> /product/ipcore/ipshared/src/RCS/cglyph.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.30  1993/11/05  01:23:19  ddavies
 * cglyph.c -> /product/ipcore/ipshared/src/RCS/cglyph.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

