/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	Geodesic propagation ("Skeleton by Influence Zones")
 *			
 *		i_propagateGray4 ()     - 4-connected geodesic propagation
 *		i_propagateGray4Down () - downward sweep of 4-connected prop.
 *		i_propagateGray4Up ()   - upward sweep of 4-connected prop.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

#include "shros.pub"
#include "shrip.pub"
#include "shrrast.pub"

/*
IP_RCSINFO(RCSInfo, "$RCSfile: i_gprop.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");
*/
/* PUBLIC */
/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Geodesic propagation of non-zero gray values to neighborhood, 4-connected
 *
 * LongDescription:::
 *  Given a grayscale image, propagate all non-zero pixels to (zero)
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly, i.e. that a given zero pixel is
 *  filled with the closest pixel, where "closest" is the distance 
 *  along the four-connected grid.
 *  <BR>
 *  We need a temporary "distance" image, which is the same dimensions
 *  as the input image, but 16 bits/pixel. This image should be passed
 *  in by the caller. Right now, since none of our tools support
 *  16 bit images, we're going to just pass in a double-wide 8-bit
 *  image. 
 *  <BR>
 *  The operation is done in-place.
 *
 * PrivateDescription:::
 *  We use a two-pass sequential algorithm, which is a modification
 *  of the sequential distance transform. The basic idea is to do
 *  a downward pass where we look at the up and left neighbors, and
 *  an upward pass where we look at the down and right neighbors.
 *  <BR>
 *  In the downward pass, we look at the pixel value. If the pixel is non-zero,
 *  we do nothing. If the pixel is zero, we look at the two neighbors, and
 *  select the non-zero neighbor with the lowest distance and fill that into
 *  the current pixel. We fill the current distance with the corresponding
 *  neighbor's distance plus one. 
 *  <BR>
 *  In the upward pass, we do something similar, but look at the distance
 *  for the current pixel as well. We take the non-zero pixel with the
 *  minimum distance (current one and its two neighbors, the neighbor's
 *  distances having been incremented) and set the  current pixel's value 
 *  to the minimum-distance pixel's value. We also update the current
 *  distance.
 *  <BR>
 *  This algorithm is faster (at least on a SPARC) than the previous
 *  one used (a FIFO algorithm) but could probably be made quite a 
 *  bit faster by working with longwords.
 *
 * Input:::
 *    pS::      Pointer to upper-left corner of frame of source image 
 *    wS::      Width (pixels) of source image without frame 
 *    hS::      Height (pixels) of source image without frame
 *    bplS::    Bytes between lines of image
 *    depthS::  Depth (bits/pixel) of image
 *    pDist::   Pointer to upper-left corner of frame of distance image.
 *              The distance image is not cleared by this routine, in
 *              case one wants to seed the first row with previous data.
 *    bplDist:: Bytes between lines of distance image (double-wide)
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *      morphology
 *
 * Keywords:::
 *	gray, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in-progress
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
i_propagateGray4 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    Int32	depthS,
    UInt8      *pDist,
    Int32	bplDist)
{
    PROLOGUE;
    

    /* Only 8bpp at the moment */
    if (depthS != 8)
	return ia_noImpl;

    /* Move past the frame */
    pS += (bplS << cLogFrameBits) + cFrameBits;
    pDist += (bplDist << cLogFrameBits) + cFrameBits;

    /* Do the downward pass */
    __error__ = i_propagateGray4Down_8bpp (pIPState, pS, wS, hS,
					 bplS, pDist, bplDist);

    if (__error__ != ia_successful)
	NORMAL_RETURN;

    /* Do the upward pass */
    __error__ = i_propagateGray4Up_8bpp (pIPState, pS, wS, hS,
				       bplS, pDist, bplDist);

    NORMAL_RETURN;
}

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Downward pass of geodesic propagation
 *
 * LongDescription:::
 *  Given a grayscale image, propagate all non-zero pixels to (zero)
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly, i.e. that a given zero pixel is
 *  filled with the closest pixel, where "closest" is the distance 
 *  along the four-connected grid.
 *  <BR>
 *  We need a temporary "distance" image, which is the same dimensions
 *  as the input image, but 16 bits/pixel. This image should be passed
 *  in by the caller. Right now, since none of our tools support
 *  16 bit images, we're going to just pass in a double-wide 8-bit
 *  image. 
 *  <BR>
 *  The operation is done in-place. The algorithm used is a two-pass
 *  sequential algorithm. This routine does one pass, and is used for
 *  each of the two passes. Note that the offset values can be positive
 *  or negative depending on scanning direction.
 *  <BR>
 *  The distance image is passed as an 8 bits/pixel image, but 
 *  is accessed as a 16 bits/pixel image, so the image passed should
 *  be twice as wide as the source image.
 *
 * PrivateDescription:::
 *  We use a two-pass sequential algorithm, which is a modification
 *  of the sequential distance transform. The basic idea is to do
 *  a downward pass where we look at the up and left neighbors, and
 *  an upward pass where we look at the down and right neighbors.
 *  <BR>
 *  In the downward pass, we look at the pixel value. If the pixel is non-zero,
 *  we do nothing. If the pixel is zero, we look at the two neighbors, and
 *  select the non-zero neighbor with the lowest distance and fill that into
 *  the current pixel. We fill the current distance with the corresponding
 *  neighbor's distance plus one. 
 *
 * Input:::
 *    pS::      Pointer to upper-left corner of frame of source image 
 *    wS::      Width (pixels) of source image without frame 
 *    hS::      Height (pixels) of source image without frame
 *    bplS::    Bytes between lines of image
 *    pDist::   Pointer to upper-left corner of frame of distance image
 *    bplDist:: Bytes between lines of distance image (double-wide)
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *      morphology
 *
 * Keywords:::
 *	gray, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in-progress
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
i_propagateGray4Down_8bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    UInt8      *pDist,
    Int32	bplDist)
{
    PROLOGUE;
    Int32 i, pixel, dist1, dist2, val1, val2;
    UInt8 *lineP, *distP;

    /* Advance one row and column */
    pS += bplS + 1;
    pDist += bplDist + 2;

    /* do h-1 lines, because we skipped the first one */
    while (hS-- > 1)
    {
	lineP = pS;
	distP = pDist;

	/* Do w-1 columns, because we skipped the first one */
	for (i = 1; i < wS; i++)
	{
	    pixel = UCHAR_ACCESS (lineP);
	    if (pixel == 0)
	    {
		val1 = UCHAR_ACCESS (lineP - bplS);
		if (val1)
		    dist1 = USHORT_ACCESS (distP - bplDist) + 1;
		else
		    dist1 = 30000;

		val2 = UCHAR_ACCESS (lineP - 1);
		if (val2)
		    dist2 = USHORT_ACCESS (distP - 2) + 1;
		else
		    dist2 = 30000;

		if (val1 || val2)
		{
		    if (dist1 < dist2)
		    {
			UCHAR_ACCESS (lineP) = (UInt8) val1;
			USHORT_ACCESS (distP) = (UInt16) dist1;
		    }
		    else
		    {
			UCHAR_ACCESS (lineP) = (UInt8) val2;
			USHORT_ACCESS (distP) = (UInt16) dist2;
		    }
		}
	    }

	    ++ lineP;
	    distP += 2;
	}

	pS += bplS;
	pDist += bplDist;
    }

    NORMAL_RETURN;
}

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Upward pass of geodesic propagation
 *
 * LongDescription:::
 *  Given a grayscale image, propagate all non-zero pixels to (zero)
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly, i.e. that a given zero pixel is
 *  filled with the closest pixel, where "closest" is the distance 
 *  along the four-connected grid.
 *  <BR>
 *  We need a temporary "distance" image, which is the same dimensions
 *  as the input image, but 16 bits/pixel. This image should be passed
 *  in by the caller. Right now, since none of our tools support
 *  16 bit images, we're going to just pass in a double-wide 8-bit
 *  image. 
 *  <BR>
 *  The operation is done in-place. The algorithm used is a two-pass
 *  sequential algorithm. This routine does one pass, and is used for
 *  each of the two passes. Note that the offset values can be positive
 *  or negative depending on scanning direction.
 *  <BR>
 *  The distance image is passed as an 8 bits/pixel image, but 
 *  is accessed as a 16 bits/pixel image, so the image passed should
 *  be twice as wide as the source image.
 *
 * PrivateDescription:::
 *  We use a two-pass sequential algorithm, which is a modification
 *  of the sequential distance transform. The basic idea is to do
 *  a downward pass where we look at the up and left neighbors, and
 *  an upward pass where we look at the down and right neighbors.
 *  <BR>
 *  In the upward pass, we take the non-zero pixel with the
 *  minimum distance (current one and its two neighbors, the neighbor's
 *  distances having been incremented) and set the current pixel's value 
 *  to the minimum-distance pixel's value. We also update the current
 *  distance.
 *
 * Input:::
 *    pS::      Pointer to upper-left corner of frame of source image 
 *    wS::      Width (pixels) of source image without frame 
 *    hS::      Height (pixels) of source image without frame
 *    bplS::    Bytes between lines of image
 *    pDist::   Pointer to upper-left corner of frame of distance image
 *    bplDist:: Bytes between lines of distance image (double-wide)
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *      morphology
 *
 * Keywords:::
 *	gray, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in-progress
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
i_propagateGray4Up_8bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    UInt8      *pDist,
    Int32	bplDist)
{
    PROLOGUE;
    Int32 i, pixel, dist, dist1, pixel1;
    UInt8 *lineP, *distP;

    /* Advance to lower-right minus one column and row*/
    pS += bplS * (hS - 1) + (wS - 1);
    pDist += bplDist * (hS - 1) + (wS - 1) * 2;

    /* do h-1 lines, because we skipped the first one */
    while (hS-- > 1)
    {
	lineP = pS;
	distP = pDist;

	/* Do w-1 columns, because we skipped the first one */
	for (i = 1; i < wS; i++)
	{
	    pixel = UCHAR_ACCESS (lineP);
	    if (pixel)
		dist = USHORT_ACCESS (distP);
	    else
		dist = 30000;
         /* dist1 >= 1, if dist = 0, pixel will not be changed */
	    if( dist != 0 )
	    {
		pixel1 = UCHAR_ACCESS (lineP + bplS);
		if (pixel1)
		{
		    dist1 = USHORT_ACCESS (distP + bplDist) + 1;
		    if (dist1 < dist)
		    {
			dist = dist1;
			pixel = pixel1;
		    }
		}

		pixel1 = UCHAR_ACCESS (lineP + 1);
		if (pixel1)
		{
		    dist1 = USHORT_ACCESS (distP + 2) + 1;
		    if (dist1 < dist)
		    {
			dist = dist1;
			pixel = pixel1;
		    }
		}

		if (pixel)
		{
		    UCHAR_ACCESS (lineP) = (UInt8) pixel;
		    USHORT_ACCESS (distP) = (UInt16) dist;
		}
	    }
	    -- lineP;
	    distP -= 2;
	}

	pS -= bplS;
	pDist -= bplDist;
    }

    NORMAL_RETURN;
}


/*
 * $Log: i_gprop.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 * Imported sources
 *
 * Revision 1.3  1998/01/13 23:47:58  ddavies
 * i_gprop.c -> /product/ipcore/ipshared/src/RCS/i_gprop.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.2  1997/07/09 16:56:13  xzhu
 * xzhu on Wed Jul 9 09:55:21 PDT 1997
 *
 * i_gprop.c -> /product/ipcore/ipshared/src/RCS/i_gprop.c,v
 *
 * Speeded up prpagateGray4UP_8bpp.
 *
 * Revision 1.1  1996/07/19 15:17:46  macleod
 * Initial revision
 *
 */


