/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1998, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	Geodesic propagation ("Skeleton by Influence Zones") using
 *      mask.
 *			
 *	i_propagateColor4Mask()   - 4-connected geodesic propagation
 *
 *	ip_propagateColor4Mask8Down()  - downward sweep of 4-connected prop,
 *					 8-bit distance values.
 *	ip_propagateColor4Mask16Down() - downward sweep of 4-connected prop,
 *					 16-bit distance values.
 *	ip_propagateColor4Mask8Up()    - upward sweep of 4-connected prop,
 *					 8-bit distance values.
 *	ip_propagateColor4Mask16Up()   - upward sweep of 4-connected prop,
 *					 16-bit distance values.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

#include "shros.pub"
#include "shrip.pub"
#include "shrip.prv"
#include "shrrast.pub"

/*
IP_RCSINFO(RCSInfo, "$RCSfile: i_gprop2.c,v $; $Revision: 1.1.1.1 $");
*/
/* $Date: 1999/09/20 18:18:11 $ */

#define U16_ACC(UShort) *((UInt16 *)(&UShort))

/* Filling an image with gray or color values starting with a known set of
 * seeds is a very useful operation.  In this operation, we fill each
 * destination pixel with the value of the seed pixel that's closest to
 * the destination pixel.
 *
 * In our current system, this operation is used in three places.  It's
 * used to spread thresholds in a gray image.  That is, there are places
 * where a good threshold value is known (at the edge of a character) and
 * places where no good value is known.  The gray propagate operation
 * spreads the known thresholds throughout the image.  The fact that the
 * threshold at a pixel is set equal to the nearest good threshold is
 * important.
 *
 * The gray propagation operation is also used to smooth our foreground
 * and background images.  A foreground image holds the colors of text.
 * A background image holds the color of the stuff behind text.  We would
 * like to compress both using very lossy compression, so would like both
 * to be smooth.  We use a color version of gray propagate to fill in
 * text holes in the background image (propagating the nearest background
 * pixels) and to spread out the colors of text the foreground region.
 * Ideally, in a region of red text on green background, the background
 * image is all green and the foreground image is all red.
 *
 * The original implementation of gray propagation used a fancy, and slow,
 * stack scheme.  The second implementation was much faster, using a two-pass
 * scheme.  The first pass spread known pixels and distances from those pixels
 * from the upper left to the lower right.  The second pass spread known
 * pixel and their distances back to the upper right.  It also set the final
 * value for every propagated pixel to the value of the pixel having the
 * smallest distance after both passes.
 *
 * The third implementation is also a two pass scheme, but with some important
 * changes in details.  Let's formally define "seed pixels" as those pixels
 * in the image whose value is known and which should not change.  We'll
 * the remaining pixels "non-seed pixels."  The purpose of gray propagate
 * is to fill in the values of the non-seed pixels from the seed pixels.
 * In particular, it should set each non-seed pixel to the value of the
 * nearest seed pixel.
 *
 * There are several ways of defining "nearest."  One could mean real
 * Euclidean distance measured using square roots of sums of squares.  One
 * could simply define each horizontal and vertical step as having length
 * 1 and each diagonal step as having some other length (1.5 is pretty close).
 * For our purposes, we use manhattan distance - horizontal and vertical
 * steps have length 1 and diagonal steps have length 2.  This means we
 * really only have to look at the horizontal and vertical neighbors of
 * each pixel.
 *
 * In the previous implementation, the caller passed in an initialized
 * pixel array and an initialized distance array.  Before calling the
 * routine, the caller set the value of non-seed pixels to 0.  This was
 * mildly annoying both because is took time and because you weren't allowed
 * to have any seed pixels that really had a value of 0.  The distance
 * array was initialized to 0.  It used adjacent 8-bit pixels as
 * single 16-bit distances.
 *
 * The former gray propagate code did a clever dance with the distances
 * and tagged pixels.  On the first pass, it checked the pixels value to
 * make sure that a) either the value of pixel above or the pixel to the
 * left was known, and b) the value of the current pixel is not known.
 * If these conditions were true, it wrote a pixels value and distance.
 * On the second pass, since most, but not all, of the pixel tags had
 * been erased, it had to look at the distances which might be either
 * non-zero (initialized) or zero (either not initialized or corresponds
 * to a seed pixel) and pixel values (to distinguish those cases) to
 * decide what to do.
 *
 * In our implementation, the caller at this level passes in a pixel array
 * in which there are no tags and a distance array in which all the
 * seed pixels correspond to distances of 0 and all non-seed pixels
 * have distances of all 1's.  An interesting advantage is that the caller
 * need not mark up the pixel array and need not guarantee that any pixels
 * have non-tag values.
 *
 * The distance array may have 8 bit values or 16-bit values.
 * The latter guarantees that exactly the right thing happens very far
 * away from seed pixels at the cost of being twice as big as of using
 * slow (on a PC) 16-bit operations.
 *
 * In our first pass, we simply zip through distances until we find one
 * that's non-zero.  This can be done a word at a time and is much easier than
 * finding isolated pixels or distances that equal 0 as was required by
 * the previous scheme.  When a non-zero distance is found, we compare the
 * distance corresponding to the pixel to the left.  If that distance is
 * strictly less than the current distance (note that uninitialized non-seed
 * pixel distances fail this test because they are equal), we set the
 * current distance to the left pixel's distance plus 1 and the value of
 * the current pixel to be the value of the left pixel.  We repeat this
 * process for the pixel above the current pixel.  That's it.
 *
 * In the second pass, we do exactly the same thing, except we compare each
 * pixel with the pixel to the right and the pixel below.
 *
 * We write out different inner loops for the cases of 8-bit distances and
 * 16-bit distances since the accesses are different.
 *
 * We also have separate cases for color and gray images since it is MUCH
 * faster to simply set two more color samples than to recompute the
 * distances for each sample.
 *
 * 				*	*	*
 *
 * Modern processors, like the Pentium II or UltraSPARC, have multiple
 * processing units and fancy pipelines.  They work best if you can
 * maintain multiple, independent computation streams.  We can't do that
 * completely, but we can try to move store-load dependencies apart.
 *
 * But don't we have to process the line strictly from left to right or
 * strictly from right to left?  Only sort of.  Let's look at the left
 * to right scan case.  We have to do two comparisons for each pixel.
 * One is with the pixel to the left.  The other is with the pixel above.
 * I maintain that we can do the comparisons with the pixels above in
 * any order.  In particular, we need not do both comparisons for the
 * leftmost pixel, then both for the next pixel and so on.
 *
 * How can this be true?  Say we propagate the distance of pixel p(i,j) to 
 * p(i+1,j) horizontally, then later discover that p(i,j) is closer to
 * a seed pixel above it than it was to whatever seed pixel was to its
 * left.  Won't p(i+1,j) have the wrong value?  Maybe it will momentarily,
 * but it will eventually be corrected as seen below.
 *
 * There are two main cases.
 * 1. p(i,j) is a seed pixel.
 * 2. p(i,j) is not a seed pixel.
 *
 * If p(i,j) is a seed pixel, then when we propagate its distance+1 to
 * p(i+1,j), we're guaranteed to get the right distance.  Since the distance
 * for p(i,j) will never be made smaller (it's already 0), it doesn't matter
 * when you attempt to update pixel i's distance by looking vertically.
 *
 * If p(i,j) is not a seed pixel, then we may indeed update its distance
 * after we have used its previous distance to update p(i+1,j)'s distance.
 * However, we already have the correct values in the entire row above
 * pixels i and i+1.  Say the new value of p(i,j)'s distance is D, which is
 * less than its previous distance.  Then, we know that p(i,j-1)'s distance
 * is D-1 and p(i+1,j-1)'s distance is no larger than (D-1) + 1 = D.
 * Therefore, whether p(i+1,j)'s distance is updated vertically before or
 * after consulting p(i,j)'s distance, p(i+1,j) will get a distance no
 * larger than D+1, which is correct.
 *
 * Ok, but what about p(i,j)'s actual pixel value as opposed to its distance?
 * Well, since the pixel values are carried along with their distances,
 * having the correct distance guarantees having the correct pixel value.
 *
 *
 * What does it all mean?
 * In big endian processors when the distances are 8-bit things, we
 * test 4 distances at a time to see if any are non-zero.  If any are,
 * we process the four corresponding pixels (each in a byte which is
 * offset from the pointer which always points to an address that's 0
 * mod 4) in this order:
 *	dest	source
 *	0	-1
 *	3	prevLine + 3
 *	1	0
 *	0	prevLine + 0
 *	2	1
 *	1	prevLine + 1
 *	3	2
 *	2	prevLine + 2
 *
 * Note that in each case, we read a value immediately *before* updating
 * it instead of the other way around.
 *
 * For little endian machine, we do the same thing in this order:
 *	dest	source
 *	3	-4
 *	0	prevLine + 0
 *	2	3
 *	3	prevLine + 3
 *	1	2
 *	2	prevLine + 2
 *	0	1
 *	1	prevLine + 1
 *
 *
 * Things are uglier when distances are 2-byte things.  In a big endian
 * machine, four consecutive distances have the following offsets relative
 * to a pointer:
 *	-4 -3 -2 -1 0 1 2 3  4 5 6 7
 *	            ^   ^    ^   ^		; or 0, 2, 4, 6.  Very nice
 * These distances correspond to 4 bytes at offsets
 *      -4 -3 -2 -1 0 1 2 3
 *	            ^ ^ ^ ^			; or 0, 1, 2, 3.
 *
 * In a little endian machine, the corresponding offsets are
 *      -1 -2 -3 -4 3 2 1 0  7 6 5 4
 *                    ^   ^    ^   ^		; or 2, 0, 6, 4.  Ugly.
 * These distances correspond to 4 bytes at offsets
 *      -1 -2 -3 -4 3 2 1 0
 *                  ^ ^ ^ ^			; or 3, 2, 1, 0
 *
 * The other complication is that we can only test two distances at a time
 * to see if one of them is non-zero.  That means we really want to
 * update only the two distances we tested.  We can't just map the previous
 * 8-bit order because of that little problem.  The new order for big
 * endian machines is (destOffset,pixOffset):
 *
 * test destword at 0ffset 0.  If either dest is non-zero, process dest's and
 * pixels in this order:
 *	dest	source
 *	(0,0)	(-2, -1)
 *	(2,1)	(prevLine+2, prevLine+1)
 *	(0,0)	(prevLine+0, prevLine+0)
 *	(2,1)	(0,0)
 * test destword at 0ffset 4.  If either dest is non-zero, process dest's and
 * pixels in this order:
 *	dest	source
 *	(4,2)	(2,1)
 *	(6,3)	(prevLine+6, prevLine+3)
 *	(4,2)	(prevLine+4, prevLine+2)
 *	(6,3)	(4,2)
 *
 * Here's the corresponding action for little endian machines:
 * test destword at 0ffset 0.  If either dest is non-zero, process dest's and
 * pixels in this order:
 *	dest	source
 *	(2,3)	(-4,-4)
 *	(0,2)	(prevLine+0, prevLine+2)
 *	(2,3)	(prevLine+2, prevLine+3)
 *	(0,2)	(2, 3)
 *
 *	(6,1)	(0, 2)
 *	(4,0)	(prevLine+4, prevLine+0)
 *	(6,1)	(prevLine+6, prevLine+1)
 *	(4,0)	(6, 1)
 *
 *				*	*	*
 *
 * The color version of this routine is the same as the gray version
 * except that it takes three pointers to images and update three pixels
 * on each move instead of one.
 */


/* PUBLIC */
/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Geodesic propagation of seed color values using mask, 4-connected
 *
 * LongDescription:::
 *  Given a color image, propagate all seed pixels to non-seed
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly, i.e. that a given non-seed pixel is
 *  filled with the closest seed pixel, where "closest" is the distance 
 *  along the four-connected grid.
 *  <BR>
 *  We need a temporary "distance" image, which is the same dimensions
 *  as the input image, but 8 or 16 bits/pixel. This image should be passed
 *  in by the caller.
 *  <BR>
 *  The caller must initialize the distance image with all 0's corresponding
 *  to each seed pixel and all 1's corresponding to each non-seed pixel.
 *  This may be done using i_depth1To8 or i_depth1To16 on a binary mask
 *  that contains 1's for all *non-seed* pixels.
 *  <BR>
 *  The operation is done in-place.
 *
 * PrivateDescription:::
 *  We use a two-pass sequential algorithm, which is a modification
 *  of the sequential distance transform. The basic idea is to do
 *  a downward pass where we look at the up and left neighbors, and
 *  an upward pass where we look at the down and right neighbors.
 *  <BR>
 *  In the downward pass, we look at the pixel's distance. If the
 *  distance is already zero, we do nothing since that pixel cannot be
 *  replaced with a pixel that is closer to a seed pixel (it *is* a seed
 *  pixel).  If the distance is non-zero, we look at the two neighbors, and
 *  select the neighbor with the lowest distance and fill that pixel value into
 *  the current pixel. We fill the current distance with the corresponding
 *  neighbor's distance plus one. 
 *  <BR>
 *  In the upward pass, we do the same thing, but look at the neighbors
 *  that are to the right and below.
 *
 * Input:::
 *    pIPState:: Pointer to core state.
 *    pR::      Pointer to upper-left corner of frame of one plane
 *		of source image.  This is also the corresponding plane
 *		of the destination image.
 *    pG::	Pointer to upper-left corner of frame of second plane of
 *		source/dest image.
 *    pB::	Pointer to upper-left corner of frame of third plane of
 *		source/dest image.
 *    wS::      Width (pixels) of source image without frame 
 *    hS::      Height (pixels) of source image without frame
 *    bplS::    Bytes between lines of image.  This must apply to all three
 *		source images.
 *    depthS::  Depth (bits/pixel) of each plane of the image.  All planes
 *		must have the same depth.
 *    pDist::   Pointer to upper-left corner of frame of distance image.
 *              Distances in this image may be either 8-bits or 16-bits
 *		(see depthD below).  The values corresponding to
 *		seed pixels (the ones that spread into the rest of the
 *		image) must be preset to 0 by the caller.  The values
 *		corresponding to non-seed pixels must be set to 0xFF or
 *		0xFFFF depending on the number of bits per distance.
 *    bplDist:: Bytes between lines of distance image (double-wide)
 *    depthD::	The number of bits per distance value.  Must be either
 *		8 or 16.
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *      morphology
 *
 * Keywords:::
 *	gray, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in-progress
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
i_propagateColor4Mask (
    IPState    *pIPState,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    Int32	depthS,
    UInt8      *pDist,
    Int32	bplDist,
    Int32	depthD)
{
    PROLOGUE;

	/* Check depthD */
    ERR_GEN(((depthD != 8) && (depthD != 16)), ia_invalidParm);

	/* Only 8bpp images at the moment */
    ERR_GEN ((depthS != 8), ia_noImpl);

	/* adjust base pointers to avoid the frame.  Note that the
	 * Frame around the distance image is still in units of 8-bit
	 * pixels.  That's because we really don't have 16-bit pixel
	 * images yet so the caller has to cheat a tiny bit.
	 */
    pR += (bplS << cLogFrameBits) + depthS * cFrameBytes;
    pG += (bplS << cLogFrameBits) + depthS * cFrameBytes;
    pB += (bplS << cLogFrameBits) + depthS * cFrameBytes;
    pDist += (bplDist << cLogFrameBits) + depthD * cFrameBytes;

    /* Do the downward pass */
    if (depthD == 8)
    {
	ERR_PASS( ip_propagateColor4Mask8Down (pIPState, pR, pG, pB, wS, hS,
					       bplS, pDist, bplDist));
    }
    else
    {
	ERR_PASS( ip_propagateColor4Mask16Down (pIPState, pR, pG, pB, wS, hS,
					        bplS, pDist, bplDist));
    }

    /* Do the upward pass */
    if (depthD == 8)
    {
	ERR_PASS( ip_propagateColor4Mask8Up (pIPState, pR, pG, pB, wS, hS,
					     bplS, pDist, bplDist));
    }
    else
    {
	ERR_PASS( ip_propagateColor4Mask16Up (pIPState, pR, pG, pB, wS, hS,
					      bplS, pDist, bplDist));
    }

    NORMAL_RETURN;
}


/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Downward pass of color geodesic propagation, 8-bit distance values
 *
 * LongDescription:::
 *  Given a color image, propagate all seed pixels to non-seed
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly, i.e. that a given non-seed pixel is
 *  filled with the closest seed pixel, where "closest" is the distance 
 *  along the four-connected grid.
 *  <BR>
 *  We need a temporary "distance" image, which is the same dimensions
 *  as the input image, and is 8 bits/pixel. This image should be passed
 *  in by the caller.  The caller must set the distances of all seed
 *  pixels to 0 and the distances of all non-seed pixels to 0xFF.
 *  This can be done using i_depth1To8.
 *  <BR>
 *  The operation is done in-place.
 *
 * PrivateDescription:::
 *  We use a two-pass sequential algorithm, which is a modification
 *  of the sequential distance transform. The basic idea is to do
 *  a downward pass where we look at the up and left neighbors, and
 *  an upward pass where we look at the down and right neighbors.
 *  <BR>
 *  In the downward pass, we look at a pixel's distance. If the distance
 *  is zero, we do nothing since the pixel can't be replaced with a value
 *  that's closer to a seed pixel (the pixel *is* already a seed pixel).
 *  If the distance is non-zero, we look at the two neighbors, and
 *  select the neighbor with the lowest distance and fill that pixel value into
 *  the current pixel. We fill the current distance with the corresponding
 *  neighbor's distance plus one. 
 *
 * Input:::
 *    pIPState:: Pointer to core state.
 *    pR::      Pointer to upper-left corner of one plane of source/dest
 *		image inside frame
 *    pG::	Pointer to upper-left corner of second plane of source/dest
 *		image inside frame
 *    pB::	Pointer to upper-left corner of third plane of source/dest
 *		image inside frame
 *    wS::      Width (pixels) of source image without frame 
 *    hS::      Height (pixels) of source image without frame
 *    bplS::    Bytes between lines of image
 *    pDist::   Pointer to upper-left corner of frame of distance image
 *    bplDist:: Bytes between lines of distance image (double-wide)
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *      morphology
 *
 * Keywords:::
 *	gray, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in-progress
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
ip_propagateColor4Mask8Down (
    IPState    *pIPState,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    UInt8      *pDist,
    Int32	bplDist)
{
    PROLOGUE;
    UInt8      *pRLine, *pGLine, *pBLine;
    UInt8      *pDistLine, *pDistEndLine;

	/* Do a special thing for the first line since we don't
	 * compare it to the line above.  In every line, we treat
	 * the first column separately since we don't compare it
	 * to the column in the frame.
	 */
    pRLine = pR;
    pGLine = pG;
    pBLine = pB;
    pDistLine = pDist;
    pDistEndLine = pDistLine + wS;
    bplS = -bplS;		/* we use these because all the machines
				 * really want to ADD address offsets, not
				 * subtract them.
				 */
    bplDist = -bplDist;

	/* Do the first word in the line */
    if (ULONG_ACCESS(pDistLine) != 0)
    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* This is a little endian machine so access the bytes in
	     * the 3, 2, 1, 0 order.
	     */
	if (pDistLine[2] > pDistLine[3])
	{
	    pDistLine[2] = pDistLine[3] + 1;
	    pRLine[2] = pRLine[3];
	    pGLine[2] = pGLine[3];
	    pBLine[2] = pBLine[3];
	}
	if (pDistLine[1] > pDistLine[2])
	{
	    pDistLine[1] = pDistLine[2] + 1;
	    pRLine[1] = pRLine[2];
	    pGLine[1] = pGLine[2];
	    pBLine[1] = pBLine[2];
	}
	if (pDistLine[0] > pDistLine[1])
	{
	    pDistLine[0] = pDistLine[1] + 1;
	    pRLine[0] = pRLine[1];
	    pGLine[0] = pGLine[1];
	    pBLine[0] = pBLine[1];
	}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* This is a big endian machine so access the bytes in
	     * the 0, 1, 2, 3 order.
	     */
	if (pDistLine[1] > pDistLine[0])
	{
	    pDistLine[1] = pDistLine[0] + 1;
	    pRLine[1] = pRLine[0];
	    pGLine[1] = pGLine[0];
	    pBLine[1] = pBLine[0];
	}
	if (pDistLine[2] > pDistLine[1])
	{
	    pDistLine[2] = pDistLine[1] + 1;
	    pRLine[2] = pRLine[1];
	    pGLine[2] = pGLine[1];
	    pBLine[2] = pBLine[1];
	}
	if (pDistLine[3] > pDistLine[2])
	{
	    pDistLine[3] = pDistLine[2] + 1;
	    pRLine[3] = pRLine[2];
	    pGLine[3] = pGLine[2];
	    pBLine[3] = pBLine[2];
	}
#endif
    }

	/* go to next word on first line */
    pRLine += 4;
    pGLine += 4;
    pBLine += 4;
    pDistLine += 4;
    while (pDistLine < pDistEndLine)
    {
	if (ULONG_ACCESS(pDistLine) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* This is a little endian machine so access the bytes in
		 * the 3, 2, 1, 0 order.
		 */
	    if (pDistLine[3] > pDistLine[-4])
	    {
		pDistLine[3] = pDistLine[-4] + 1;
		pRLine[3] = pRLine[-4];
		pGLine[3] = pGLine[-4];
		pBLine[3] = pBLine[-4];
	    }
	    if (pDistLine[2] > pDistLine[3])
	    {
		pDistLine[2] = pDistLine[3] + 1;
		pRLine[2] = pRLine[3];
		pGLine[2] = pGLine[3];
		pBLine[2] = pBLine[3];
	    }
	    if (pDistLine[1] > pDistLine[2])
	    {
		pDistLine[1] = pDistLine[2] + 1;
		pRLine[1] = pRLine[2];
		pGLine[1] = pGLine[2];
		pBLine[1] = pBLine[2];
	    }
	    if (pDistLine[0] > pDistLine[1])
	    {
		pDistLine[0] = pDistLine[1] + 1;
		pRLine[0] = pRLine[1];
		pGLine[0] = pGLine[1];
		pBLine[0] = pBLine[1];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		/* This is a big endian machine so access the bytes in
		 * the 0, 1, 2, 3 order.
		 */
	    if (pDistLine[0] > pDistLine[-1])
	    {
		pDistLine[0] = pDistLine[-1] + 1;
		pRLine[0] = pRLine[-1];
		pGLine[0] = pGLine[-1];
		pBLine[0] = pBLine[-1];
	    }
	    if (pDistLine[1] > pDistLine[0])
	    {
		pDistLine[1] = pDistLine[0] + 1;
		pRLine[1] = pRLine[0];
		pGLine[1] = pGLine[0];
		pBLine[1] = pBLine[0];
	    }
	    if (pDistLine[2] > pDistLine[1])
	    {
		pDistLine[2] = pDistLine[1] + 1;
		pRLine[2] = pRLine[1];
		pGLine[2] = pGLine[1];
		pBLine[2] = pBLine[1];
	    }
	    if (pDistLine[3] > pDistLine[2])
	    {
		pDistLine[3] = pDistLine[2] + 1;
		pRLine[3] = pRLine[2];
		pGLine[3] = pGLine[2];
		pBLine[3] = pBLine[2];
	    }
#endif
	}
	pRLine += 4;
	pGLine += 4;
	pBLine += 4;
	pDistLine += 4;
    }


	/* So much for the first line.  Now do all the other lines.
	 * In these tests, we compare each pixel with the pixel to
	 * the left AND the pixel above.
	 * Remember that bplS and bplDist are the offsets to the
	 * line above, not below so we subtract them to move down.
	 */
    pR -= bplS;
    pG -= bplS;
    pB -= bplS;
    pDist -= bplDist;
    hS--;	/* already did first line */

    while (hS-- > 0)
    {
	pRLine = pR;
	pGLine = pG;
	pBLine = pB;
	pDistLine = pDist;
	pDistEndLine = pDistLine + wS;
    
	    /* Do the first word in the line */
	if (ULONG_ACCESS(pDistLine) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* This is a little endian machine so access the bytes in
		 * the 3, 2, 1, 0 order.
		 * Note that the order in which we do the vertical
		 * compares doesn't matter with respect to the
		 * other horizontal compares or with respect to
		 * each other.  Therefore, we'll reorder them in
		 * an attempt to break the dependency chains.
		 */
	    if (pDistLine[0] > pDistLine[bplDist + 0])
	    {
		pDistLine[0] = pDistLine[bplDist + 0] + 1;
		pRLine[0] = pRLine[bplS + 0];
		pGLine[0] = pGLine[bplS + 0];
		pBLine[0] = pBLine[bplS + 0];
	    }
	    if (pDistLine[2] > pDistLine[3])
	    {
		pDistLine[2] = pDistLine[3];
		pRLine[2] = pRLine[3];
		pGLine[2] = pGLine[3];
		pBLine[2] = pBLine[3];
	    }
	    if (pDistLine[3] > pDistLine[bplDist + 3])
	    {
		pDistLine[3] = pDistLine[bplDist + 3] + 1;
		pRLine[3] = pRLine[bplS + 3];
		pGLine[3] = pGLine[bplS + 3];
		pBLine[3] = pBLine[bplS + 3];
	    }
	    if (pDistLine[1] > pDistLine[2])
	    {
		pDistLine[1] = pDistLine[2];
		pRLine[1] = pRLine[2];
		pGLine[1] = pGLine[2];
		pBLine[1] = pBLine[2];
	    }
	    if (pDistLine[2] > pDistLine[bplDist + 2])
	    {
		pDistLine[2] = pDistLine[bplDist + 2] + 1;
		pRLine[2] = pRLine[bplS + 2];
		pGLine[2] = pGLine[bplS + 2];
		pBLine[2] = pBLine[bplS + 2];
	    }
	    if (pDistLine[0] > pDistLine[1])
	    {
		pDistLine[0] = pDistLine[1];
		pRLine[0] = pRLine[1];
		pGLine[0] = pGLine[1];
		pBLine[0] = pBLine[1];
	    }
	    if (pDistLine[1] > pDistLine[bplDist + 1])
	    {
		pDistLine[1] = pDistLine[bplDist + 1] + 1;
		pRLine[1] = pRLine[bplS + 1];
		pGLine[1] = pGLine[bplS + 1];
		pBLine[1] = pBLine[bplS + 1];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		/* This is a big endian machine so access the bytes in
		 * the 0, 1, 2, 3 order.
		 * Note that the order in which we do the vertical
		 * compares doesn't matter with respect to the
		 * other horizontal compares or with respect to
		 * each other.  Therefore, we'll reorder them in
		 * an attempt to break the dependency chains.
		 */
	    if (pDistLine[3] > pDistLine[bplDist + 3])
	    {
		pDistLine[3] = pDistLine[bplDist + 3] + 1;
		pRLine[3] = pRLine[bplS + 3];
		pGLine[3] = pGLine[bplS + 3];
		pBLine[3] = pBLine[bplS + 3];
	    }
	    if (pDistLine[1] > pDistLine[0])
	    {
		pDistLine[1] = pDistLine[0];
		pRLine[1] = pRLine[0];
		pGLine[1] = pGLine[0];
		pBLine[1] = pBLine[0];
	    }
	    if (pDistLine[0] > pDistLine[bplDist + 0])
	    {
		pDistLine[0] = pDistLine[bplDist + 0] + 1;
		pRLine[0] = pRLine[bplS + 0];
		pGLine[0] = pGLine[bplS + 0];
		pBLine[0] = pBLine[bplS + 0];
	    }
	    if (pDistLine[2] > pDistLine[1])
	    {
		pDistLine[2] = pDistLine[1];
		pRLine[2] = pRLine[1];
		pGLine[2] = pGLine[1];
		pBLine[2] = pBLine[1];
	    }
	    if (pDistLine[1] > pDistLine[bplDist + 1])
	    {
		pDistLine[1] = pDistLine[bplDist + 1] + 1;
		pRLine[1] = pRLine[bplS + 1];
		pGLine[1] = pGLine[bplS + 1];
		pBLine[1] = pBLine[bplS + 1];
	    }
	    if (pDistLine[3] > pDistLine[2])
	    {
		pDistLine[3] = pDistLine[2];
		pRLine[3] = pRLine[2];
		pGLine[3] = pGLine[2];
		pBLine[3] = pBLine[2];
	    }
	    if (pDistLine[2] > pDistLine[bplDist + 2])
	    {
		pDistLine[2] = pDistLine[bplDist + 2] + 1;
		pRLine[2] = pRLine[bplS + 2];
		pGLine[2] = pGLine[bplS + 2];
		pBLine[2] = pBLine[bplS + 2];
	    }
#endif
	}

	    /* do the remaining words on the second through bottom lines */
	pRLine += 4;
	pGLine += 4;
	pBLine += 4;
	pDistLine += 4;
	while (pDistLine < pDistEndLine)
	{
	    if (ULONG_ACCESS(pDistLine) != 0)
	    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		    /* This is a little endian machine so access the bytes in
		     * the 3, 2, 1, 0 order.
		     * Note that the order in which we do the vertical
		     * compares doesn't matter with respect to the
		     * other horizontal compares or with respect to
		     * each other.  Therefore, we'll reorder them in
		     * an attempt to break the dependency chains.
		     */
		if (pDistLine[3] > pDistLine[-4])
		{
		    pDistLine[3] = pDistLine[-4] + 1;
		    pRLine[3] = pRLine[-4];
		    pGLine[3] = pGLine[-4];
		    pBLine[3] = pBLine[-4];
		}
		if (pDistLine[0] > pDistLine[bplDist + 0])
		{
		    pDistLine[0] = pDistLine[bplDist + 0] + 1;
		    pRLine[0] = pRLine[bplS + 0];
		    pGLine[0] = pGLine[bplS + 0];
		    pBLine[0] = pBLine[bplS + 0];
		}
		if (pDistLine[2] > pDistLine[3])
		{
		    pDistLine[2] = pDistLine[3] + 1;
		    pRLine[2] = pRLine[3];
		    pGLine[2] = pGLine[3];
		    pBLine[2] = pBLine[3];
		}
		if (pDistLine[3] > pDistLine[bplDist + 3])
		{
		    pDistLine[3] = pDistLine[bplDist + 3] + 1;
		    pRLine[3] = pRLine[bplS + 3];
		    pGLine[3] = pGLine[bplS + 3];
		    pBLine[3] = pBLine[bplS + 3];
		}
		if (pDistLine[1] > pDistLine[2])
		{
		    pDistLine[1] = pDistLine[2] + 1;
		    pRLine[1] = pRLine[2];
		    pGLine[1] = pGLine[2];
		    pBLine[1] = pBLine[2];
		}
		if (pDistLine[2] > pDistLine[bplDist + 2])
		{
		    pDistLine[2] = pDistLine[bplDist + 2] + 1;
		    pRLine[2] = pRLine[bplS + 2];
		    pGLine[2] = pGLine[bplS + 2];
		    pBLine[2] = pBLine[bplS + 2];
		}
		if (pDistLine[0] > pDistLine[1])
		{
		    pDistLine[0] = pDistLine[1] + 1;
		    pRLine[0] = pRLine[1];
		    pGLine[0] = pGLine[1];
		    pBLine[0] = pBLine[1];
		}
		if (pDistLine[1] > pDistLine[bplDist + 1])
		{
		    pDistLine[1] = pDistLine[bplDist + 1] + 1;
		    pRLine[1] = pRLine[bplS + 1];
		    pGLine[1] = pGLine[bplS + 1];
		    pBLine[1] = pBLine[bplS + 1];
		}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		    /* This is a big endian machine so access the bytes in
		     * the 0, 1, 2, 3 order.
		     * Note that the order in which we do the vertical
		     * compares doesn't matter with respect to the
		     * other horizontal compares or with respect to
		     * each other.  Therefore, we'll reorder them in
		     * an attempt to break the dependency chains.
		     */
		if (pDistLine[0] > pDistLine[-1])
		{
		    pDistLine[0] = pDistLine[-1] + 1;
		    pRLine[0] = pRLine[-1];
		    pGLine[0] = pGLine[-1];
		    pBLine[0] = pBLine[-1];
		}
		if (pDistLine[3] > pDistLine[bplDist + 3])
		{
		    pDistLine[3] = pDistLine[bplDist + 3] + 1;
		    pRLine[3] = pRLine[bplS + 3];
		    pGLine[3] = pGLine[bplS + 3];
		    pBLine[3] = pBLine[bplS + 3];
		}
		if (pDistLine[1] > pDistLine[0])
		{
		    pDistLine[1] = pDistLine[0] + 1;
		    pRLine[1] = pRLine[0];
		    pGLine[1] = pGLine[0];
		    pBLine[1] = pBLine[0];
		}
		if (pDistLine[0] > pDistLine[bplDist + 0])
		{
		    pDistLine[0] = pDistLine[bplDist + 0] + 1;
		    pRLine[0] = pRLine[bplS + 0];
		    pGLine[0] = pGLine[bplS + 0];
		    pBLine[0] = pBLine[bplS + 0];
		}
		if (pDistLine[2] > pDistLine[1])
		{
		    pDistLine[2] = pDistLine[1] + 1;
		    pRLine[2] = pRLine[1];
		    pGLine[2] = pGLine[1];
		    pBLine[2] = pBLine[1];
		}
		if (pDistLine[1] > pDistLine[bplDist + 1])
		{
		    pDistLine[1] = pDistLine[bplDist + 1] + 1;
		    pRLine[1] = pRLine[bplS + 1];
		    pGLine[1] = pGLine[bplS + 1];
		    pBLine[1] = pBLine[bplS + 1];
		}
		if (pDistLine[3] > pDistLine[2])
		{
		    pDistLine[3] = pDistLine[2] + 1;
		    pRLine[3] = pRLine[2];
		    pGLine[3] = pGLine[2];
		    pBLine[3] = pBLine[2];
		}
		if (pDistLine[2] > pDistLine[bplDist + 2])
		{
		    pDistLine[2] = pDistLine[bplDist + 2] + 1;
		    pRLine[2] = pRLine[bplS + 2];
		    pGLine[2] = pGLine[bplS + 2];
		    pBLine[2] = pBLine[bplS + 2];
		}
#endif
	    }
	    pRLine += 4;
	    pGLine += 4;
	    pBLine += 4;
	    pDistLine += 4;
	}
	/* Remember that bplS and bplDist are the offsets
	 * to the line ABOVE, so we subtract them to
	 * move down.
	 */
    pR -= bplS;
    pG -= bplS;
    pB -= bplS;
    pDist -= bplDist;
    }

    NORMAL_RETURN;
}


/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Downward pass of color geodesic propagation, 16-bit distance values
 *
 * LongDescription:::
 *  Given a color image, propagate all seed pixels to non-seed
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly, i.e. that a given non-seed pixel is
 *  filled with the closest seed pixel, where "closest" is the distance 
 *  along the four-connected grid.
 *  <BR>
 *  We need a temporary "distance" image, which is the same dimensions
 *  as the input image, and is 16 bits/pixel. This image should be passed
 *  in by the caller.  The caller must set the distances of all seed
 *  pixels to 0 and the distances of all non-seed pixels to 0xFFFF.
 *  This can be done using i_depth1To16.
 *  <BR>
 *  The operation is done in-place.
 *
 * PrivateDescription:::
 *  We use a two-pass sequential algorithm, which is a modification
 *  of the sequential distance transform. The basic idea is to do
 *  a downward pass where we look at the up and left neighbors, and
 *  an upward pass where we look at the down and right neighbors.
 *  <BR>
 *  In the downward pass, we look at a pixel's distance. If the distance
 *  is zero, we do nothing since the pixel can't be replaced with a value
 *  that's closer to a seed pixel (the pixel *is* already a seed pixel).
 *  If the distance is non-zero, we look at the two neighbors, and
 *  select the neighbor with the lowest distance and fill that pixel value into
 *  the current pixel. We fill the current distance with the corresponding
 *  neighbor's distance plus one. 
 *
 * Input:::
 *    pIPState:: Pointer to core state.
 *    pR::      Pointer to upper-left corner of one plane of source/dest
 *		image inside frame
 *    pG::	Pointer to upper-left corner of second plane of source/dest
 *		image inside frame
 *    pB::	Pointer to upper-left corner of third plane of source/dest
 *		image inside frame
 *    wS::      Width (pixels) of source image without frame 
 *    hS::      Height (pixels) of source image without frame
 *    bplS::    Bytes between lines of image
 *    pDist::   Pointer to upper-left corner of frame of distance image
 *    bplDist:: Bytes between lines of distance image (double-wide)
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *      morphology
 *
 * Keywords:::
 *	gray, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in-progress
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
ip_propagateColor4Mask16Down (
    IPState    *pIPState,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    UInt8      *pDist,
    Int32	bplDist)
{
    PROLOGUE;
    UInt8      *pRLine, *pGLine, *pBLine;
    UInt8      *pDistLine, *pDistEndLine;

	/* Do a special thing for the first line since we don't
	 * compare it to the line above.  In every line, we treat
	 * the first column separately since we don't compare it
	 * to the column in the frame.
	 */
    pRLine = pR;
    pGLine = pG;
    pBLine = pB;
    pDistLine = pDist;
    pDistEndLine = pDistLine + 2*wS;
    bplS = -bplS;		/* we use these because all the machines
				 * really want to ADD address offsets, not
				 * subtract them.
				 */
    bplDist = -bplDist;


	/* Do the first word in the line */
    if (ULONG_ACCESS(&pDistLine[0]) != 0)
    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* This is a little endian machine so access the bytes in
	     * the 3, 2, 1, 0 order.
	     */
	if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[2]))
	{
	    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[2]) + 1;
	    pRLine[2] = pRLine[3];
	    pGLine[2] = pGLine[3];
	    pBLine[2] = pBLine[3];
	}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* This is a big endian machine so access the bytes in
	     * the 0, 1, 2, 3 order.
	     */
	if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[0]))
	{
	    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[0]) + 1;
	    pRLine[1] = pRLine[0];
	    pGLine[1] = pGLine[0];
	    pBLine[1] = pBLine[0];
	}
#endif
    }
    if (ULONG_ACCESS(&pDistLine[4]) != 0)
    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[0]))
	{
	    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[0]) + 1;
	    pRLine[1] = pRLine[2];
	    pGLine[1] = pGLine[2];
	    pBLine[1] = pBLine[2];
	}
	if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[6]))
	{
	    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[6]) + 1;
	    pRLine[0] = pRLine[1];
	    pGLine[0] = pGLine[1];
	    pBLine[0] = pBLine[1];
	}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[2]))
	{
	    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[2]) + 1;
	    pRLine[2] = pRLine[1];
	    pGLine[2] = pGLine[1];
	    pBLine[2] = pBLine[1];
	}
	if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[4]))
	{
	    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[4]) + 1;
	    pRLine[3] = pRLine[2];
	    pGLine[3] = pGLine[2];
	    pBLine[3] = pBLine[2];
	}
#endif
    }

	/* go to next word on first line */
    pRLine += 4;
    pGLine += 4;
    pBLine += 4;
    pDistLine += 8;
    while (pDistLine < pDistEndLine)
    {
	if (ULONG_ACCESS(&pDistLine[0]) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* This is a little endian machine so access the bytes in
		 * the 3, 2, 1, 0 order.
		 */
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[-4]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[-4]) + 1;
		pRLine[3] = pRLine[-4];
		pGLine[3] = pGLine[-4];
		pBLine[3] = pBLine[-4];
	    }
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[2]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[2]) + 1;
		pRLine[2] = pRLine[3];
		pGLine[2] = pGLine[3];
		pBLine[2] = pBLine[3];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[-2]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[-2]) + 1;
		pRLine[0] = pRLine[-1];
		pGLine[0] = pGLine[-1];
		pBLine[0] = pBLine[-1];
	    }
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[0]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[0]) + 1;
		pRLine[1] = pRLine[0];
		pGLine[1] = pGLine[0];
		pBLine[1] = pBLine[0];
	    }
#endif
	}
	if (ULONG_ACCESS(&pDistLine[4]) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[0]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[0]) + 1;
		pRLine[1] = pRLine[2];
		pGLine[1] = pGLine[2];
		pBLine[1] = pBLine[2];
	    }
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[6]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[6]) + 1;
		pRLine[0] = pRLine[1];
		pGLine[0] = pGLine[1];
		pBLine[0] = pBLine[1];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		/* This is a big endian machine so access the bytes in
		 * the 0, 1, 2, 3 order.
		 */
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[2]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[2]) + 1;
		pRLine[2] = pRLine[1];
		pGLine[2] = pGLine[1];
		pBLine[2] = pBLine[1];
	    }
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[4]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[4]) + 1;
		pRLine[3] = pRLine[2];
		pGLine[3] = pGLine[2];
		pBLine[3] = pBLine[2];
	    }
#endif
	}
	pRLine += 4;
	pGLine += 4;
	pBLine += 4;
	pDistLine += 8;
    }


	/* So much for the first line.  Now do all the other lines.
	 * In these tests, we compare each pixel with the pixel to
	 * the left AND the pixel above.
	 * Remember that bplS and bplDist are the offsets to the
	 * line above, not below so we subtract them to move down.
	 */
    pR -= bplS;
    pG -= bplS;
    pB -= bplS;
    pDist -= bplDist;
    hS--;	/* already did first line */

    while (hS-- > 0)
    {
	pRLine = pR;
	pGLine = pG;
	pBLine = pB;
	pDistLine = pDist;
	pDistEndLine = pDist + 2*wS;
    
	    /* Do the first word in the line */
	if (ULONG_ACCESS(&pDistLine[0]) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* This is a little endian machine so access the bytes in
		 * the 3, 2, 1, 0 order.
		 * Note that the order in which we do the vertical
		 * compares doesn't matter with respect to the
		 * other horizontal compares or with respect to
		 * each other.  Therefore, we'll reorder them in
		 * an attempt to break the dependency chains.
		 */
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[bplDist + 0]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[bplDist + 0]) + 1;
		pRLine[2] = pRLine[bplS + 2];
		pGLine[2] = pGLine[bplS + 2];
		pBLine[2] = pBLine[bplS + 2];
	    }
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[bplDist + 2]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[bplDist + 2]) + 1;
		pRLine[3] = pRLine[bplS + 3];
		pGLine[3] = pGLine[bplS + 3];
		pBLine[3] = pBLine[bplS + 3];
	    }
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[2]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[2]) + 1;
		pRLine[2] = pRLine[3];
		pGLine[2] = pGLine[3];
		pBLine[2] = pBLine[3];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		/* This is a big endian machine so access the bytes in
		 * the 0, 1, 2, 3 order.
		 * Note that the order in which we do the vertical
		 * compares doesn't matter with respect to the
		 * other horizontal compares or with respect to
		 * each other.  Therefore, we'll reorder them in
		 * an attempt to break the dependency chains.
		 */
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[bplDist + 2]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[bplDist + 2]) + 1;
		pRLine[1] = pRLine[bplS + 1];
		pGLine[1] = pGLine[bplS + 1];
		pBLine[1] = pBLine[bplS + 1];
	    }
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[bplDist + 0]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[bplDist + 0]) + 1;
		pRLine[0] = pRLine[bplS + 0];
		pGLine[0] = pGLine[bplS + 0];
		pBLine[0] = pBLine[bplS + 0];
	    }
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[0]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[0]) + 1;
		pRLine[1] = pRLine[0];
		pGLine[1] = pGLine[0];
		pBLine[1] = pBLine[0];
	    }
#endif
	}

	if (ULONG_ACCESS(&pDistLine[4]) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[0]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[0]) + 1;
		pRLine[1] = pRLine[2];
		pGLine[1] = pGLine[2];
		pBLine[1] = pBLine[2];
	    }
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[bplDist + 4]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[bplDist + 4]) + 1;
		pRLine[0] = pRLine[bplS + 0];
		pGLine[0] = pGLine[bplS + 0];
		pBLine[0] = pBLine[bplS + 0];
	    }
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[bplDist + 6]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[bplDist + 6]) + 1;
		pRLine[1] = pRLine[bplS + 1];
		pGLine[1] = pGLine[bplS + 1];
		pBLine[1] = pBLine[bplS + 1];
	    }
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[6]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[6]) + 1;
		pRLine[0] = pRLine[1];
		pGLine[0] = pGLine[1];
		pBLine[0] = pBLine[1];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[2]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[2]) + 1;
		pRLine[2] = pRLine[1];
		pGLine[2] = pGLine[1];
		pBLine[2] = pBLine[1];
	    }
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[bplDist + 6]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[bplDist + 6]) + 1;
		pRLine[3] = pRLine[bplS + 3];
		pGLine[3] = pGLine[bplS + 3];
		pBLine[3] = pBLine[bplS + 3];
	    }
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[bplDist + 4]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[bplDist + 4]) + 1;
		pRLine[2] = pRLine[bplS + 2];
		pGLine[2] = pGLine[bplS + 2];
		pBLine[2] = pBLine[bplS + 2];
	    }
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[4]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[4]) + 1;
		pRLine[3] = pRLine[2];
		pGLine[3] = pGLine[2];
		pBLine[3] = pBLine[2];
	    }
#endif
	}

	    /* do the remaining words on the second through bottom lines */
	pRLine += 4;
	pGLine += 4;
	pBLine += 4;
	pDistLine += 8;
	while (pDistLine < pDistEndLine)
	{
	    if (ULONG_ACCESS(&pDistLine[0]) != 0)
	    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		    /* This is a little endian machine so access the bytes in
		     * the 3, 2, 1, 0 order.
		     * Note that the order in which we do the vertical
		     * compares doesn't matter with respect to the
		     * other horizontal compares or with respect to
		     * each other.  Therefore, we'll reorder them in
		     * an attempt to break the dependency chains.
		     */
		if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[-4]))
		{
		    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[-4]) + 1;
		    pRLine[3] = pRLine[-4];
		    pGLine[3] = pGLine[-4];
		    pBLine[3] = pBLine[-4];
		}
		if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[bplDist + 0]))
		{
		    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[bplDist + 0]) + 1;
		    pRLine[2] = pRLine[bplS + 2];
		    pGLine[2] = pGLine[bplS + 2];
		    pBLine[2] = pBLine[bplS + 2];
		}
		if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[bplDist + 2]))
		{
		    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[bplDist + 2]) + 1;
		    pRLine[3] = pRLine[bplS + 3];
		    pGLine[3] = pGLine[bplS + 3];
		    pBLine[3] = pBLine[bplS + 3];
		}
		if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[2]))
		{
		    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[2]) + 1;
		    pRLine[2] = pRLine[3];
		    pGLine[2] = pGLine[3];
		    pBLine[2] = pBLine[3];
		}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		    /* This is a big endian machine so access the bytes in
		     * the 0, 1, 2, 3 order.
		     * Note that the order in which we do the vertical
		     * compares doesn't matter with respect to the
		     * other horizontal compares or with respect to
		     * each other.  Therefore, we'll reorder them in
		     * an attempt to break the dependency chains.
		     */
		if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[-2]))
		{
		    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[-2]) + 1;
		    pRLine[0] = pRLine[-1];
		    pGLine[0] = pGLine[-1];
		    pBLine[0] = pBLine[-1];
		}
		if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[bplDist + 2]))
		{
		    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[bplDist + 2]) + 1;
		    pRLine[1] = pRLine[bplS + 1];
		    pGLine[1] = pGLine[bplS + 1];
		    pBLine[1] = pBLine[bplS + 1];
		}
		if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[bplDist + 0]))
		{
		    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[bplDist + 0]) + 1;
		    pRLine[0] = pRLine[bplS + 0];
		    pGLine[0] = pGLine[bplS + 0];
		    pBLine[0] = pBLine[bplS + 0];
		}
		if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[0]))
		{
		    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[0]) + 1;
		    pRLine[1] = pRLine[0];
		    pGLine[1] = pGLine[0];
		    pBLine[1] = pBLine[0];
		}
#endif
	    }

	    if (ULONG_ACCESS(&pDistLine[4]) != 0)
	    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[0]))
		{
		    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[0]) + 1;
		    pRLine[1] = pRLine[2];
		    pGLine[1] = pGLine[2];
		    pBLine[1] = pBLine[2];
		}
		if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[bplDist + 4]))
		{
		    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[bplDist + 4]) + 1;
		    pRLine[0] = pRLine[bplS + 0];
		    pGLine[0] = pGLine[bplS + 0];
		    pBLine[0] = pBLine[bplS + 0];
		}
		if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[bplDist + 6]))
		{
		    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[bplDist + 6]) + 1;
		    pRLine[1] = pRLine[bplS + 1];
		    pGLine[1] = pGLine[bplS + 1];
		    pBLine[1] = pBLine[bplS + 1];
		}
		if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[6]))
		{
		    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[6]) + 1;
		    pRLine[0] = pRLine[1];
		    pGLine[0] = pGLine[1];
		    pBLine[0] = pBLine[1];
		}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[2]))
		{
		    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[2]) + 1;
		    pRLine[2] = pRLine[1];
		    pGLine[2] = pGLine[1];
		    pBLine[2] = pBLine[1];
		}
		if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[bplDist + 6]))
		{
		    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[bplDist + 6]) + 1;
		    pRLine[3] = pRLine[bplS + 3];
		    pGLine[3] = pGLine[bplS + 3];
		    pBLine[3] = pBLine[bplS + 3];
		}
		if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[bplDist + 4]))
		{
		    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[bplDist + 4]) + 1;
		    pRLine[2] = pRLine[bplS + 2];
		    pGLine[2] = pGLine[bplS + 2];
		    pBLine[2] = pBLine[bplS + 2];
		}
		if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[4]))
		{
		    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[4]) + 1;
		    pRLine[3] = pRLine[2];
		    pGLine[3] = pGLine[2];
		    pBLine[3] = pBLine[2];
		}
#endif
	    }
	    pRLine += 4;
	    pGLine += 4;
	    pBLine += 4;
	    pDistLine += 8;
	}
	/* Remember that bplS and bplDist are the offsets
	 * to the line ABOVE, so we subtract them to
	 * move down.
	 */
    pR -= bplS;
    pG -= bplS;
    pB -= bplS;
    pDist -= bplDist;
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Upward pass of color geodesic propagation, 8-bit distance values
 *
 * LongDescription:::
 *  Given a color image, propagate all seed pixels to non-seed
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly, i.e. that a given non-seed pixel is
 *  filled with the closest seed pixel, where "closest" is the distance 
 *  along the four-connected grid.
 *  <BR>
 *  We need a temporary "distance" image, which is the same dimensions
 *  as the input image, and is 8 bits/pixel. This image should be passed
 *  in by the caller.  The caller must set the distances of all seed
 *  pixels to 0 and the distances of all non-seed pixels to 0xFF.
 *  This can be done using i_depth1To8.
 *  <BR>
 *  The operation is done in-place.
 *
 * PrivateDescription:::
 *  We use a two-pass sequential algorithm, which is a modification
 *  of the sequential distance transform. The basic idea is to do
 *  a downward pass where we look at the up and left neighbors, and
 *  an upward pass where we look at the down and right neighbors.
 *  <BR>
 *  In the upward pass, we look at a pixel's distance. If the distance
 *  is zero, we do nothing since the pixel can't be replaced with a value
 *  that's closer to a seed pixel (the pixel *is* already a seed pixel).
 *  If the distance is non-zero, we look at the two neighbors, and
 *  select the neighbor with the lowest distance and fill that pixel value into
 *  the current pixel. We fill the current distance with the corresponding
 *  neighbor's distance plus one. 
 *
 * Input:::
 *    pIPState:: Pointer to core state.
 *    pR::      Pointer to upper-left corner of one plane of source/dest
 *		image inside frame
 *    pG::	Pointer to upper-left corner of second plane of source/dest
 *		image inside frame
 *    pB::	Pointer to upper-left corner of third plane of source/dest
 *		image inside frame
 *    wS::      Width (pixels) of source image without frame 
 *    hS::      Height (pixels) of source image without frame
 *    bplS::    Bytes between lines of image
 *    pDist::   Pointer to upper-left corner of distance image inside frame
 *    bplDist:: Bytes between lines of distance image (double-wide)
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *      morphology
 *
 * Keywords:::
 *	gray, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in-progress
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
ip_propagateColor4Mask8Up (
    IPState    *pIPState,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    UInt8      *pDist,
    Int32	bplDist)
{
    PROLOGUE;
    UInt8      *pRLine, *pGLine, *pBLine;
    UInt8      *pDistLine;

	/* Point to the bottom rows of the pixel and distance
	 * images.
	 */
    pR += bplS * (hS-1);
    pG += bplS * (hS-1);
    pB += bplS * (hS-1);
    pDist += bplDist * (hS-1);
    
	/* Do a special thing for the last line since we don't
	 * compare it to the line below.  In every line, we treat
	 * the last column separately since we don't compare it
	 * to the column in the frame.
	 */
    wS--;		/* we really want wS-1 in all our calculations. */
    pRLine = pR + (wS & ~0x3);
    pGLine = pG + (wS & ~0x3);
    pBLine = pB + (wS & ~0x3);
    pDistLine = pDist + (wS & ~0x3);

	/* Do the last word in the last line */
    if (ULONG_ACCESS(&pDistLine[0]) != 0)
    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* This is a little endian machine so access the bytes in
	     * the 3, 2, 1, 0 order.
	     */
	if (((wS & 0x3) == 3) && (pDistLine[1] > pDistLine[0]))
	{
	    pDistLine[1] = pDistLine[0] + 1;
	    pRLine[1] = pRLine[0];
	    pGLine[1] = pGLine[0];
	    pBLine[1] = pBLine[0];
	}
	if (((wS & 0x3) >= 2) && (pDistLine[2] > pDistLine[1]))
	{
	    pDistLine[2] = pDistLine[1] + 1;
	    pRLine[2] = pRLine[1];
	    pGLine[2] = pGLine[1];
	    pBLine[2] = pBLine[1];
	}
	if (((wS & 0x3) >= 1) && (pDistLine[3] > pDistLine[2]))
	{
	    pDistLine[3] = pDistLine[2] + 1;
	    pRLine[3] = pRLine[2];
	    pGLine[3] = pGLine[2];
	    pBLine[3] = pBLine[2];
	}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* This is a big endian machine so access the bytes in
	     * the 0, 1, 2, 3 order.
	     */
	if (((wS & 0x3) == 3) && (pDistLine[2] > pDistLine[3]))
	{
	    pDistLine[2] = pDistLine[3] + 1;
	    pRLine[2] = pRLine[3];
	    pGLine[2] = pGLine[3];
	    pBLine[2] = pBLine[3];
	}
	if (((wS & 0x3) >= 2) && (pDistLine[1] > pDistLine[2]))
	{
	    pDistLine[1] = pDistLine[2] + 1;
	    pRLine[1] = pRLine[2];
	    pGLine[1] = pGLine[2];
	    pBLine[1] = pBLine[2];
	}
	if (((wS & 0x3) >= 1) && (pDistLine[0] > pDistLine[1]))
	{
	    pDistLine[0] = pDistLine[1] + 1;
	    pRLine[0] = pRLine[1];
	    pGLine[0] = pGLine[1];
	    pBLine[0] = pBLine[1];
	}
#endif
    }

	/* go to next-to-last word on last line */
    pRLine -= 4;
    pGLine -= 4;
    pBLine -= 4;
    pDistLine -= 4;
    while (pDistLine >= pDist)
    {
	if (ULONG_ACCESS(pDistLine) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* This is a little endian machine so access the bytes in
		 * the 3, 2, 1, 0 order.
		 */
	    if (pDistLine[0] > pDistLine[7])
	    {
		pDistLine[0] = pDistLine[7] + 1;
		pRLine[0] = pRLine[7];
		pGLine[0] = pGLine[7];
		pBLine[0] = pBLine[7];
	    }
	    if (pDistLine[1] > pDistLine[0])
	    {
		pDistLine[1] = pDistLine[0] + 1;
		pRLine[1] = pRLine[0];
		pGLine[1] = pGLine[0];
		pBLine[1] = pBLine[0];
	    }
	    if (pDistLine[2] > pDistLine[1])
	    {
		pDistLine[2] = pDistLine[1] + 1;
		pRLine[2] = pRLine[1];
		pGLine[2] = pGLine[1];
		pBLine[2] = pBLine[1];
	    }
	    if (pDistLine[3] > pDistLine[2])
	    {
		pDistLine[3] = pDistLine[2] + 1;
		pRLine[3] = pRLine[2];
		pGLine[3] = pGLine[2];
		pBLine[3] = pBLine[2];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		/* This is a big endian machine so access the bytes in
		 * the 0, 1, 2, 3 order.
		 */
	    if (pDistLine[3] > pDistLine[4])
	    {
		pDistLine[3] = pDistLine[4] + 1;
		pRLine[3] = pRLine[4];
		pGLine[3] = pGLine[4];
		pBLine[3] = pBLine[4];
	    }
	    if (pDistLine[2] > pDistLine[3])
	    {
		pDistLine[2] = pDistLine[3] + 1;
		pRLine[2] = pRLine[3];
		pGLine[2] = pGLine[3];
		pBLine[2] = pBLine[3];
	    }
	    if (pDistLine[1] > pDistLine[2])
	    {
		pDistLine[1] = pDistLine[2] + 1;
		pRLine[1] = pRLine[2];
		pGLine[1] = pGLine[2];
		pBLine[1] = pBLine[2];
	    }
	    if (pDistLine[0] > pDistLine[1])
	    {
		pDistLine[0] = pDistLine[1] + 1;
		pRLine[0] = pRLine[1];
		pGLine[0] = pGLine[1];
		pBLine[0] = pBLine[1];
	    }
#endif
	}
	pRLine -= 4;
	pGLine -= 4;
	pBLine -= 4;
	pDistLine -= 4;
    }


	/* So much for the last line.  Now do all the other lines.
	 * In these tests, we compare each pixel with the pixel to
	 * the right AND the pixel below.
	 * Remember that bplS and bplDist are the offsets to the
	 * line below, not above so we subtract them to move up.
	 */
    pR -= bplS;
    pG -= bplS;
    pB -= bplS;
    pDist -= bplDist;
    hS--;	/* already did last line */

    while (hS-- > 0)
    {
	pRLine = pR + (wS & ~0x3);
	pGLine = pG + (wS & ~0x3);
	pBLine = pB + (wS & ~0x3);
	pDistLine = pDist + (wS & ~0x3);

	    /* Do the last word in the line */
	if (ULONG_ACCESS(pDistLine) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* This is a little endian machine so access the bytes in
	     * the 3, 2, 1, 0 order.
	     * Note that the order in which we do the vertical
	     * compares doesn't matter with respect to the
	     * other horizontal compares or with respect to
	     * each other.  Therefore, we'll reorder them in
	     * an attempt to break the dependency chains.
	     */
	    if (pDistLine[3] > pDistLine[bplDist + 3])
	    {
		pDistLine[3] = pDistLine[bplDist + 3] + 1;
		pRLine[3] = pRLine[bplS + 3];
		pGLine[3] = pGLine[bplS + 3];
		pBLine[3] = pBLine[bplS + 3];
	    }
	    if (((wS & 0x3) == 3) && (pDistLine[1] > pDistLine[0]))
	    {
		pDistLine[1] = pDistLine[0] + 1;
		pRLine[1] = pRLine[0];
		pGLine[1] = pGLine[0];
		pBLine[1] = pBLine[0];
	    }
	    if (pDistLine[0] > pDistLine[bplDist + 0])
	    {
		pDistLine[0] = pDistLine[bplDist + 0] + 1;
		pRLine[0] = pRLine[bplS + 0];
		pGLine[0] = pGLine[bplS + 0];
		pBLine[0] = pBLine[bplS + 0];
	    }
	    if (((wS & 0x3) >= 2) && (pDistLine[2] > pDistLine[1]))
	    {
		pDistLine[2] = pDistLine[1] + 1;
		pRLine[2] = pRLine[1];
		pGLine[2] = pGLine[1];
		pBLine[2] = pBLine[1];
	    }
	    if (pDistLine[1] > pDistLine[bplDist + 1])
	    {
		pDistLine[1] = pDistLine[bplDist + 1] + 1;
		pRLine[1] = pRLine[bplS + 1];
		pGLine[1] = pGLine[bplS + 1];
		pBLine[1] = pBLine[bplS + 1];
	    }
	    if (((wS & 0x3) >= 1) && (pDistLine[3] > pDistLine[2]))
	    {
		pDistLine[3] = pDistLine[2] + 1;
		pRLine[3] = pRLine[2];
		pGLine[3] = pGLine[2];
		pBLine[3] = pBLine[2];
	    }
	    if (pDistLine[2] > pDistLine[bplDist + 2])
	    {
		pDistLine[2] = pDistLine[bplDist + 2] + 1;
		pRLine[2] = pRLine[bplS + 2];
		pGLine[2] = pGLine[bplS + 2];
		pBLine[2] = pBLine[bplS + 2];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* This is a big endian machine so access the bytes in
	     * the 0, 1, 2, 3 order.
	     * Note that the order in which we do the vertical
	     * compares doesn't matter with respect to the
	     * other horizontal compares or with respect to
	     * each other.  Therefore, we'll reorder them in
	     * an attempt to break the dependency chains.
	     */
	    if (pDistLine[0] > pDistLine[bplDist + 0])
	    {
		pDistLine[0] = pDistLine[bplDist + 0] + 1;
		pRLine[0] = pRLine[bplS + 0];
		pGLine[0] = pGLine[bplS + 0];
		pBLine[0] = pBLine[bplS + 0];
	    }
	    if (((wS & 0x3) == 3) && (pDistLine[2] > pDistLine[3]))
	    {
		pDistLine[2] = pDistLine[3] + 1;
		pRLine[2] = pRLine[3];
		pGLine[2] = pGLine[3];
		pBLine[2] = pBLine[3];
	    }
	    if (pDistLine[3] > pDistLine[bplDist + 3])
	    {
		pDistLine[3] = pDistLine[bplDist + 3] + 1;
		pRLine[3] = pRLine[bplS + 3];
		pGLine[3] = pGLine[bplS + 3];
		pBLine[3] = pBLine[bplS + 3];
	    }
	    if (((wS & 0x3) >= 2) && (pDistLine[1] > pDistLine[2]))
	    {
		pDistLine[1] = pDistLine[2] + 1;
		pRLine[1] = pRLine[2];
		pGLine[1] = pGLine[2];
		pBLine[1] = pBLine[2];
	    }
	    if (pDistLine[2] > pDistLine[bplDist + 2])
	    {
		pDistLine[2] = pDistLine[bplDist + 2] + 1;
		pRLine[2] = pRLine[bplS + 2];
		pGLine[2] = pGLine[bplS + 2];
		pBLine[2] = pBLine[bplS + 2];
	    }
	    if (((wS & 0x3) >= 1) && (pDistLine[0] > pDistLine[1]))
	    {
		pDistLine[0] = pDistLine[1] + 1;
		pRLine[0] = pRLine[1];
		pGLine[0] = pGLine[1];
		pBLine[0] = pBLine[1];
	    }
	    if (pDistLine[1] > pDistLine[bplDist + 1])
	    {
		pDistLine[1] = pDistLine[bplDist + 1] + 1;
		pRLine[1] = pRLine[bplS + 1];
		pGLine[1] = pGLine[bplS + 1];
		pBLine[1] = pBLine[bplS + 1];
	    }
#endif
	}

	    /* do the remaining words on the second through bottom lines */
	pRLine -= 4;
	pGLine -= 4;
	pBLine -= 4;
	pDistLine -= 4;
	while (pDistLine >= pDist)
	{
	    if (ULONG_ACCESS(pDistLine) != 0)
	    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		    /* This is a little endian machine so access the bytes in
		     * the 3, 2, 1, 0 order.
		     * Note that the order in which we do the vertical
		     * compares doesn't matter with respect to the
		     * other horizontal compares or with respect to
		     * each other.  Therefore, we'll reorder them in
		     * an attempt to break the dependency chains.
		     */
		if (pDistLine[0] > pDistLine[7])
		{
		    pDistLine[0] = pDistLine[7] + 1;
		    pRLine[0] = pRLine[7];
		    pGLine[0] = pGLine[7];
		    pBLine[0] = pBLine[7];
		}
		if (pDistLine[3] > pDistLine[bplDist + 3])
		{
		    pDistLine[3] = pDistLine[bplDist + 3] + 1;
		    pRLine[3] = pRLine[bplS + 3];
		    pGLine[3] = pGLine[bplS + 3];
		    pBLine[3] = pBLine[bplS + 3];
		}
		if (pDistLine[1] > pDistLine[0])
		{
		    pDistLine[1] = pDistLine[0] + 1;
		    pRLine[1] = pRLine[0];
		    pGLine[1] = pGLine[0];
		    pBLine[1] = pBLine[0];
		}
		if (pDistLine[0] > pDistLine[bplDist + 0])
		{
		    pDistLine[0] = pDistLine[bplDist + 0] + 1;
		    pRLine[0] = pRLine[bplS + 0];
		    pGLine[0] = pGLine[bplS + 0];
		    pBLine[0] = pBLine[bplS + 0];
		}
		if (pDistLine[2] > pDistLine[1])
		{
		    pDistLine[2] = pDistLine[1] + 1;
		    pRLine[2] = pRLine[1];
		    pGLine[2] = pGLine[1];
		    pBLine[2] = pBLine[1];
		}
		if (pDistLine[1] > pDistLine[bplDist + 1])
		{
		    pDistLine[1] = pDistLine[bplDist + 1] + 1;
		    pRLine[1] = pRLine[bplS + 1];
		    pGLine[1] = pGLine[bplS + 1];
		    pBLine[1] = pBLine[bplS + 1];
		}
		if (pDistLine[3] > pDistLine[2])
		{
		    pDistLine[3] = pDistLine[2] + 1;
		    pRLine[3] = pRLine[2];
		    pGLine[3] = pGLine[2];
		    pBLine[3] = pBLine[2];
		}
		if (pDistLine[2] > pDistLine[bplDist + 2])
		{
		    pDistLine[2] = pDistLine[bplDist + 2] + 1;
		    pRLine[2] = pRLine[bplS + 2];
		    pGLine[2] = pGLine[bplS + 2];
		    pBLine[2] = pBLine[bplS + 2];
		}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		    /* This is a big endian machine so access the bytes in
		     * the 0, 1, 2, 3 order.
		     * Note that the order in which we do the vertical
		     * compares doesn't matter with respect to the
		     * other horizontal compares or with respect to
		     * each other.  Therefore, we'll reorder them in
		     * an attempt to break the dependency chains.
		     */
		if (pDistLine[3] > pDistLine[4])
		{
		    pDistLine[3] = pDistLine[4] + 1;
		    pRLine[3] = pRLine[4];
		    pGLine[3] = pGLine[4];
		    pBLine[3] = pBLine[4];
		}
		if (pDistLine[0] > pDistLine[bplDist + 0])
		{
		    pDistLine[0] = pDistLine[bplDist + 0] + 1;
		    pRLine[0] = pRLine[bplS + 0];
		    pGLine[0] = pGLine[bplS + 0];
		    pBLine[0] = pBLine[bplS + 0];
		}
		if (pDistLine[2] > pDistLine[3])
		{
		    pDistLine[2] = pDistLine[3] + 1;
		    pRLine[2] = pRLine[3];
		    pGLine[2] = pGLine[3];
		    pBLine[2] = pBLine[3];
		}
		if (pDistLine[3] > pDistLine[bplDist + 3])
		{
		    pDistLine[3] = pDistLine[bplDist + 3] + 1;
		    pRLine[3] = pRLine[bplS + 3];
		    pGLine[3] = pGLine[bplS + 3];
		    pBLine[3] = pBLine[bplS + 3];
		}
		if (pDistLine[1] > pDistLine[2])
		{
		    pDistLine[1] = pDistLine[2] + 1;
		    pRLine[1] = pRLine[2];
		    pGLine[1] = pGLine[2];
		    pBLine[1] = pBLine[2];
		}
		if (pDistLine[2] > pDistLine[bplDist + 2])
		{
		    pDistLine[2] = pDistLine[bplDist + 2] + 1;
		    pRLine[2] = pRLine[bplS + 2];
		    pGLine[2] = pGLine[bplS + 2];
		    pBLine[2] = pBLine[bplS + 2];
		}
		if (pDistLine[0] > pDistLine[1])
		{
		    pDistLine[0] = pDistLine[1] + 1;
		    pRLine[0] = pRLine[1];
		    pGLine[0] = pGLine[1];
		    pBLine[0] = pBLine[1];
		}
		if (pDistLine[1] > pDistLine[bplDist + 1])
		{
		    pDistLine[1] = pDistLine[bplDist + 1] + 1;
		    pRLine[1] = pRLine[bplS + 1];
		    pGLine[1] = pGLine[bplS + 1];
		    pBLine[1] = pBLine[bplS + 1];
		}
#endif
	    }
	    pRLine -= 4;
	    pGLine -= 4;
	    pBLine -= 4;
	    pDistLine  -= 4;
	}
	/* Remember that bplS and bplDist are the offsets
	 * to the line below, so we subtract them to
	 * move up.
	 */
    pR -= bplS;
    pG -= bplS;
    pB -= bplS;
    pDist -= bplDist;
    }

    NORMAL_RETURN;
}




/* PRIVATE */
/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Upward pass of color geodesic propagation, 16-bit distance values
 *
 * LongDescription:::
 *  Given a color image, propagate all seed pixels to non-seed
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly, i.e. that a given non-seed pixel is
 *  filled with the closest seed pixel, where "closest" is the distance 
 *  along the four-connected grid.
 *  <BR>
 *  We need a temporary "distance" image, which is the same dimensions
 *  as the input image, and is 16 bits/pixel. This image should be passed
 *  in by the caller.  The caller must set the distances of all seed
 *  pixels to 0 and the distances of all non-seed pixels to 0xFFFF.
 *  This can be done using i_depth1To16.
 *  <BR>
 *  The operation is done in-place.
 *
 * PrivateDescription:::
 *  We use a two-pass sequential algorithm, which is a modification
 *  of the sequential distance transform. The basic idea is to do
 *  a downward pass where we look at the up and left neighbors, and
 *  an upward pass where we look at the down and right neighbors.
 *  <BR>
 *  In the upward pass, we look at a pixel's distance. If the distance
 *  is zero, we do nothing since the pixel can't be replaced with a value
 *  that's closer to a seed pixel (the pixel *is* already a seed pixel).
 *  If the distance is non-zero, we look at the two neighbors, and
 *  select the neighbor with the lowest distance and fill that pixel value into
 *  the current pixel.  We fill the current distance with the corresponding
 *  neighbor's distance plus one. 
 *
 * Input:::
 *    pIPState:: Pointer to core state.
 *    pR::      Pointer to upper-left corner of one plane of source/dest
 *		image inside frame
 *    pG::	Pointer to upper-left corner of second plane of source/dest
 *		image inside frame
 *    pB::	Pointer to upper-left corner of third plane of source/dest
 *		image inside frame
 *    wS::      Width (pixels) of source image without frame 
 *    hS::      Height (pixels) of source image without frame
 *    bplS::    Bytes between lines of image
 *    pDist::   Pointer to upper-left corner of distance image inside frame
 *    bplDist:: Bytes between lines of distance image (double-wide)
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *      morphology
 *
 * Keywords:::
 *	gray, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in-progress
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
ip_propagateColor4Mask16Up (
    IPState    *pIPState,
    UInt8      *pR,
    UInt8      *pG,
    UInt8      *pB,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    UInt8      *pDist,
    Int32	bplDist)
{
    PROLOGUE;
    UInt8      *pRLine, *pGLine, *pBLine;
    UInt8      *pDistLine;

	/* Point to the bottom rows of the pixel and distance
	 * images.
	 */
    pR += bplS * (hS-1);
    pG += bplS * (hS-1);
    pB += bplS * (hS-1);
    pDist += bplDist * (hS-1);
    
	/* Do a special thing for the last line since we don't
	 * compare it to the line below.  In every line, we treat
	 * the last column separately since we don't compare it
	 * to the column in the frame.
	 */
    wS--;		/* we really want wS-1 in all our calculations. */
    pRLine = pR + (wS & ~0x3);
    pGLine = pG + (wS & ~0x3);
    pBLine = pB + (wS & ~0x3);
    pDistLine = pDist + (2*wS & ~0x7);

	/* Do the last word in the line */
    if (ULONG_ACCESS(&pDistLine[4]) != 0)
    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* This is a little endian machine so access the bytes in
	     * the 3, 2, 1, 0 order.
	     */
	if (((wS & 0x3) == 3) &&
		(U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[4])))
	{
	    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[4]) + 1;
	    pRLine[1] = pRLine[0];
	    pGLine[1] = pGLine[0];
	    pBLine[1] = pBLine[0];
	}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* This is a big endian machine so access the bytes in
	     * the 0, 1, 2, 3 order.
	     */
	if (((wS & 0x3) == 3) &&
		(U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[6])))
	{
	    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[6]) + 1;
	    pRLine[2] = pRLine[3];
	    pGLine[2] = pGLine[3];
	    pBLine[2] = pBLine[3];
	}
#endif
    }
    if (ULONG_ACCESS(&pDistLine[0]) != 0)
    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	if (((wS & 0x3) >= 2) &&
		(U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[6])))
	{
	    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[6]) + 1;
	    pRLine[2] = pRLine[1];
	    pGLine[2] = pGLine[1];
	    pBLine[2] = pBLine[1];
	}
	if (((wS & 0x3) >= 1) &&
		(U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[0])))
	{
	    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[0]) + 1;
	    pRLine[3] = pRLine[2];
	    pGLine[3] = pGLine[2];
	    pBLine[3] = pBLine[2];
	}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	if (((wS & 0x3) >= 2) &&
		(U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[4])))
	{
	    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[4]) + 1;
	    pRLine[1] = pRLine[2];
	    pGLine[1] = pGLine[2];
	    pBLine[1] = pBLine[2];
	}
	if (((wS & 0x3) >= 1) &&
		(U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[2])))
	{
	    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[2]) + 1;
	    pRLine[0] = pRLine[1];
	    pGLine[0] = pGLine[1];
	    pBLine[0] = pBLine[1];
	}
#endif
    }

	/* go to next word on first line */
    pRLine -= 4;
    pGLine -= 4;
    pBLine -= 4;
    pDistLine -= 8;
    while (pDistLine >= pDist)
    {
	if (ULONG_ACCESS(&pDistLine[4]) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* This is a little endian machine so access the bytes in
		 * the 3, 2, 1, 0 order.
		 */
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[10]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[10]) + 1;
		pRLine[0] = pRLine[7];
		pGLine[0] = pGLine[7];
		pBLine[0] = pBLine[7];
	    }
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[4]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[4]) + 1;
		pRLine[1] = pRLine[0];
		pGLine[1] = pGLine[0];
		pBLine[1] = pBLine[0];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[8]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[8]) + 1;
		pRLine[3] = pRLine[4];
		pGLine[3] = pGLine[4];
		pBLine[3] = pBLine[4];
	    }
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[6]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[6]) + 1;
		pRLine[2] = pRLine[3];
		pGLine[2] = pGLine[3];
		pBLine[2] = pBLine[3];
	    }
#endif
	}
	if (ULONG_ACCESS(&pDistLine[0]) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[6]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[6]) + 1;
		pRLine[2] = pRLine[1];
		pGLine[2] = pGLine[1];
		pBLine[2] = pBLine[1];
	    }
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[0]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[0]) + 1;
		pRLine[3] = pRLine[2];
		pGLine[3] = pGLine[2];
		pBLine[3] = pBLine[2];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		/* This is a big endian machine so access the bytes in
		 * the 0, 1, 2, 3 order.
		 */
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[4]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[4]) + 1;
		pRLine[1] = pRLine[2];
		pGLine[1] = pGLine[2];
		pBLine[1] = pBLine[2];
	    }
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[2]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[2]) + 1;
		pRLine[0] = pRLine[1];
		pGLine[0] = pGLine[1];
		pBLine[0] = pBLine[1];
	    }
#endif
	}
	pRLine -= 4;
	pGLine -= 4;
	pBLine -= 4;
	pDistLine -= 8;
    }


	/* So much for the last line.  Now do all the other lines.
	 * In these tests, we compare each pixel with the pixel to
	 * the right AND the pixel below.
	 * Remember that bplS and bplDist are the offsets to the
	 * line below, so we subtract them to move up.
	 */
    pR -= bplS;
    pG -= bplS;
    pB -= bplS;
    pDist -= bplDist;
    hS--;	/* already did last line */

    while (hS-- > 0)
    {
	pRLine = pR + (wS & ~0x3);
	pGLine = pG + (wS & ~0x3);
	pBLine = pB + (wS & ~0x3);		/* remember that wS equals
						 * real width - 1. */
	pDistLine = pDist + (2*wS & ~0x7);

	    /* Do the last word in the line */
	if (ULONG_ACCESS(&pDistLine[4]) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* This is a little endian machine so access the bytes in
		 * the 3, 2, 1, 0 order.
		 * Note that the order in which we do the vertical
		 * compares doesn't matter with respect to the
		 * other horizontal compares or with respect to
		 * each other.  Therefore, we'll reorder them in
		 * an attempt to break the dependency chains.
		 */
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[bplDist + 6]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[bplDist + 6]) + 1;
		pRLine[1] = pRLine[bplS + 1];
		pGLine[1] = pGLine[bplS + 1];
		pBLine[1] = pBLine[bplS + 1];
	    }
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[bplDist + 4]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[bplDist + 4]) + 1;
		pRLine[0] = pRLine[bplS + 0];
		pGLine[0] = pGLine[bplS + 0];
		pBLine[0] = pBLine[bplS + 0];
	    }
	    if (((wS & 0x3) == 3) &&
		    (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[4])))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[4]) + 1;
		pRLine[1] = pRLine[0];
		pGLine[1] = pGLine[0];
		pBLine[1] = pBLine[0];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		/* This is a big endian machine so access the bytes in
		 * the 0, 1, 2, 3 order.
		 * Note that the order in which we do the vertical
		 * compares doesn't matter with respect to the
		 * other horizontal compares or with respect to
		 * each other.  Therefore, we'll reorder them in
		 * an attempt to break the dependency chains.
		 */
	    if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[bplDist + 4]))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[bplDist + 4]) + 1;
		pRLine[2] = pRLine[bplS + 2];
		pGLine[2] = pGLine[bplS + 2];
		pBLine[2] = pBLine[bplS + 2];
	    }
	    if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[bplDist + 6]))
	    {
		U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[bplDist + 6]) + 1;
		pRLine[3] = pRLine[bplS + 3];
		pGLine[3] = pGLine[bplS + 3];
		pBLine[3] = pBLine[bplS + 3];
	    }
	    if (((wS & 0x3) == 3) &&
		    (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[6])))
	    {
		U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[6]) + 1;
		pRLine[2] = pRLine[3];
		pGLine[2] = pGLine[3];
		pBLine[2] = pBLine[3];
	    }
#endif
	}

	if (ULONG_ACCESS(&pDistLine[0]) != 0)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    if (((wS & 0x3) >= 2) &&
		    (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[6])))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[6]) + 1;
		pRLine[2] = pRLine[1];
		pGLine[2] = pGLine[1];
		pBLine[2] = pBLine[1];
	    }
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[bplDist + 2]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[bplDist + 2]) + 1;
		pRLine[3] = pRLine[bplS + 3];
		pGLine[3] = pGLine[bplS + 3];
		pBLine[3] = pBLine[bplS + 3];
	    }
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[bplDist + 0]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[bplDist + 0]) + 1;
		pRLine[2] = pRLine[bplS + 2];
		pGLine[2] = pGLine[bplS + 2];
		pBLine[2] = pBLine[bplS + 2];
	    }
	    if (((wS & 0x3) >= 1) &&
		    (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[0])))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[0]) + 1;
		pRLine[3] = pRLine[2];
		pGLine[3] = pGLine[2];
		pBLine[3] = pBLine[2];
	    }
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    if (((wS & 0x3) >= 2) &&
		    (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[4])))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[4]) + 1;
		pRLine[1] = pRLine[2];
		pGLine[1] = pGLine[2];
		pBLine[1] = pBLine[2];
	    }
	    if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[bplDist + 0]))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[bplDist + 0]) + 1;
		pRLine[0] = pRLine[bplS + 0];
		pGLine[0] = pGLine[bplS + 0];
		pBLine[0] = pBLine[bplS + 0];
	    }
	    if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[bplDist + 2]))
	    {
		U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[bplDist + 2]) + 1;
		pRLine[1] = pRLine[bplS + 1];
		pGLine[1] = pGLine[bplS + 1];
		pBLine[1] = pBLine[bplS + 1];
	    }
	    if (((wS & 0x3) >= 1) &&
		    (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[2])))
	    {
		U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[2]) + 1;
		pRLine[0] = pRLine[1];
		pGLine[0] = pGLine[1];
		pBLine[0] = pBLine[1];
	    }
#endif
	}

	    /* do the remaining words on the second-to-last through
	     * top lines, moving upwards */
	pRLine -= 4;
	pGLine -= 4;
	pBLine -= 4;
	pDistLine -= 8;
	while (pDistLine >= pDist)
	{
	    if (ULONG_ACCESS(&pDistLine[4]) != 0)
	    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		    /* This is a little endian machine so access the bytes in
		     * the 3, 2, 1, 0 order.
		     * Note that the order in which we do the vertical
		     * compares doesn't matter with respect to the
		     * other horizontal compares or with respect to
		     * each other.  Therefore, we'll reorder them in
		     * an attempt to break the dependency chains.
		     */
		if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[10]))
		{
		    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[10]) + 1;
		    pRLine[0] = pRLine[7];
		    pGLine[0] = pGLine[7];
		    pBLine[0] = pBLine[7];
		}
		if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[bplDist + 6]))
		{
		    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[bplDist + 6]) + 1;
		    pRLine[1] = pRLine[bplS + 1];
		    pGLine[1] = pGLine[bplS + 1];
		    pBLine[1] = pBLine[bplS + 1];
		}
		if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[bplDist + 4]))
		{
		    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[bplDist + 4]) + 1;
		    pRLine[0] = pRLine[bplS + 0];
		    pGLine[0] = pGLine[bplS + 0];
		    pBLine[0] = pBLine[bplS + 0];
		}
		if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[4]))
		{
		    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[4]) + 1;
		    pRLine[1] = pRLine[0];
		    pGLine[1] = pGLine[0];
		    pBLine[1] = pBLine[0];
		}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		    /* This is a big endian machine so access the bytes in
		     * the 0, 1, 2, 3 order.
		     * Note that the order in which we do the vertical
		     * compares doesn't matter with respect to the
		     * other horizontal compares or with respect to
		     * each other.  Therefore, we'll reorder them in
		     * an attempt to break the dependency chains.
		     */
		if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[8]))
		{
		    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[8]) + 1;
		    pRLine[3] = pRLine[4];
		    pGLine[3] = pGLine[4];
		    pBLine[3] = pBLine[4];
		}
		if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[bplDist + 4]))
		{
		    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[bplDist + 4]) + 1;
		    pRLine[2] = pRLine[bplS + 2];
		    pGLine[2] = pGLine[bplS + 2];
		    pBLine[2] = pBLine[bplS + 2];
		}
		if (U16_ACC(pDistLine[6]) > U16_ACC(pDistLine[bplDist + 6]))
		{
		    U16_ACC(pDistLine[6]) = U16_ACC(pDistLine[bplDist + 6]) + 1;
		    pRLine[3] = pRLine[bplS + 3];
		    pGLine[3] = pGLine[bplS + 3];
		    pBLine[3] = pBLine[bplS + 3];
		}
		if (U16_ACC(pDistLine[4]) > U16_ACC(pDistLine[6]))
		{
		    U16_ACC(pDistLine[4]) = U16_ACC(pDistLine[6]) + 1;
		    pRLine[2] = pRLine[3];
		    pGLine[2] = pGLine[3];
		    pBLine[2] = pBLine[3];
		}
#endif
	    }

	    if (ULONG_ACCESS(&pDistLine[0]) != 0)
	    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[6]))
		{
		    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[6]) + 1;
		    pRLine[2] = pRLine[1];
		    pGLine[2] = pGLine[1];
		    pBLine[2] = pBLine[1];
		}
		if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[bplDist + 2]))
		{
		    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[bplDist + 2]) + 1;
		    pRLine[3] = pRLine[bplS + 3];
		    pGLine[3] = pGLine[bplS + 3];
		    pBLine[3] = pBLine[bplS + 3];
		}
		if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[bplDist + 0]))
		{
		    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[bplDist + 0]) + 1;
		    pRLine[2] = pRLine[bplS + 2];
		    pGLine[2] = pGLine[bplS + 2];
		    pBLine[2] = pBLine[bplS + 2];
		}
		if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[0]))
		{
		    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[0]) + 1;
		    pRLine[3] = pRLine[2];
		    pGLine[3] = pGLine[2];
		    pBLine[3] = pBLine[2];
		}
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[4]))
		{
		    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[4]) + 1;
		    pRLine[1] = pRLine[2];
		    pGLine[1] = pGLine[2];
		    pBLine[1] = pBLine[2];
		}
		if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[bplDist + 0]))
		{
		    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[bplDist + 0]) + 1;
		    pRLine[0] = pRLine[bplS + 0];
		    pGLine[0] = pGLine[bplS + 0];
		    pBLine[0] = pBLine[bplS + 0];
		}
		if (U16_ACC(pDistLine[2]) > U16_ACC(pDistLine[bplDist + 2]))
		{
		    U16_ACC(pDistLine[2]) = U16_ACC(pDistLine[bplDist + 2]) + 1;
		    pRLine[1] = pRLine[bplS + 1];
		    pGLine[1] = pGLine[bplS + 1];
		    pBLine[1] = pBLine[bplS + 1];
		}
		if (U16_ACC(pDistLine[0]) > U16_ACC(pDistLine[2]))
		{
		    U16_ACC(pDistLine[0]) = U16_ACC(pDistLine[2]) + 1;
		    pRLine[0] = pRLine[1];
		    pGLine[0] = pGLine[1];
		    pBLine[0] = pBLine[1];
		}
#endif
	    }
	    pRLine -= 4;
	    pGLine -= 4;
	    pBLine -= 4;
	    pDistLine -= 8;
	}
	/* Remember that bplS and bplDist are the offsets
	 * to the line below, so we subtract them to
	 * move up.
	 */
    pR -= bplS;
    pG -= bplS;
    pB -= bplS;
    pDist -= bplDist;
    }

    NORMAL_RETURN;
}


/*
 * $Log: i_gprop2.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 * Imported sources
 *
 * Revision 1.1  1998/02/10 19:53:09  ddavies
 * Initial revision
 *
 */


