/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	8-bit gray gradient finder	
 *			
 *		i_findMagnitudeOfGradient8bpp ()	
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

#include "shros.pub"
#include "shrip.pub"
#include "grad.prv"

/*
IP_RCSINFO(RCSInfo, "$RCSfile: i_grad8.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");
*/

/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *	Computes the magnitude of the gradient of an 8bpp gray image.
 *
 * LongDescription:::
 *	The opType and neighborhood parameters define the shape and
 *	mode of the gradient operation (see defines in sgrad.pub).
 *
 * PrivateDescription:::
 *
 * Input:::
 *	is::	Pointer to source image including all frames
 *	ws::	Width of source image including frames
 *	hs::	Height of source image including frames
 *	bpls::	Bytes per line of the source image (must be %4)
 *	opType::	Type of gradient computation to perform
 *	neighborhood::	Shape of neighborhood to use for computation
 *	id::	Pointer to dest image including all frames
 *	wd::	Width of dest image including frames
 *	hd::	Height of dest iamge including frames
 *	bpld::	Bytes per line of the dest image (must be %4)
 *
 * Output:::
 *	id::	Pointer to dest image including all frames
 *
 * Returns:::
 *	ia_successful
 *	ia_invalidParm
 *
 * Notes:::
 *	Source and dest images must be the same width and height.
 *
 * FunctionCategory:::
 *	filtering
 *
 * Keywords:::
 *	gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
i_findMagnitudeOfGradient8bpp (
    IPState    *pIPState,
    UInt8      *is,
    Int32	ws,
    Int32	hs,
    Int32	bpls,
    Int32	opType,
    Int32	neighborhood,
    UInt8      *id,
    Int32	wd,
    Int32	hd,
    Int32	bpld)
{
    PROLOGUE;
    

    if ((ws != wd) || (hs != hd))
        return (ia_invalidParm);

    if ((is == NULL) || (id == NULL))
        return (ia_invalidParm);

    /* Currently this is the only supported neighborhood shape */
    if (neighborhood != cGradNeighborhoodDownRight)
        return (ia_noImpl);

    switch (opType)
    {
	/* for single plane images, these are all equivalent */
        case cGradOpMaxL1FromCenter:
	case cGradOpMaxL2FromCenter:
	case cGradOpMaxLInfFromCenter:
            __error__ =
		ip_downRightMaxDiffFromCenterGradient8bpp (pIPState, is,
							   ws, hs,
							   bpls, id, bpld);
	    break;

	default:
	    __error__ = ia_invalidParm;
	    break;
    }

    NORMAL_RETURN;

}  /* i_findMagnitudeOfGradient8bpp () */


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *	Simple 2 neighbor gradient finder for 8 bit images.
 *
 * LongDescription:::
 *	For each source pixel, sets the corresponding destination pixel
 *	to the maximum of (the absolute value of the intensity difference
 *	looking right one pixel) and (the absolute value of the intensity
 *	difference looking down one pixel).
 *
 * PrivateDescription:::
 *
 * Input:::
 *	is::	Pointer to source image including all frames
 *	w::	Width of the source and dest images (including frames)
 *	h::	Height of the source and dest images (including frames)
 *	bpls::	Bytes per line of the source image (must be %4)
 *	id::	Pointer to dest image including all frames
 *	bpld::	Bytes per line of the dest image (must be %4)
 *
 * Output:::
 *	id::	Pointer to dest image including all frames
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *	For a single plane, L1, L2, and L-infinity distance measures
 *	are equivalent.
 *
 * FunctionCategory:::
 *	filtering
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL 
ip_downRightMaxDiffFromCenterGradient8bpp (
    IPState    *pIPState,
    UInt8      *is,
    Int32	w,
    Int32	h,
    Int32	bpls,
    UInt8      *id,
    Int32	bpld)
{
    PROLOGUE;
    Int32  row, col, wpls,
           widthInWords;      /* width in FULL words */
    register UInt32 topLeft, topRight,
           botLeft,
	   s1, s2, x,
           diffRight, diffDown,
	   gradient;
    UInt32 *isp, *idp;

    /* point to the image data inside the frames */
    is += (bpls << cLogFrameBits) + (8 * cFrameBytes);
    id += (bpld << cLogFrameBits) + (8 * cFrameBytes);

    /* get the unframed width & height */
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

    /* since we're going to read and write only full words, calculate
     * the number of words in a row which have any real (non-frame)
     * pixels in them as well as the stride in words.
     */
    wpls = bpls / 4;
    widthInWords = (w + 3) / 4;

    for (row = 0; row < h; row++)
    {
        /* increment the source and dest pointers to the next row */
	isp = (UInt32 *)(is + (row * bpls));
	idp = (UInt32 *)(id + (row * bpld));

        /* we only need to explicitly read this part of our neighborhood
	 * at the start of a new row.
	 */
	topLeft = ULONG_ACCESS (isp);

        for (col = 0; col < widthInWords; col++)
	{
	    gradient = 0;

	    /* read a word of data to the right of the current "center" */
	    topRight = ULONG_ACCESS (isp + 1);

	    /* read a word of data from directly below the current "center" */
	    botLeft = ULONG_ACCESS (isp + wpls);

            /* We'll make heavy use of the neat block of code adapted
	     * from ip_absValDiffLine8 () which simultaneously computes
	     * the difference of 4 8-bit pixels at a time.  See the
	     * comments in i_absdif.c for more info on this hack.
	     */

            /* compute the right (->) absolute differences */
            s1 = topLeft;

	    /* This makes a word of pixels with an offset of one to
	     * the right.
	     */
            s2 = (topLeft << 8) | (topRight >> 24);
	    x = s1 ^ s2;
            diffRight = (s1 | 0x80808080) -
	                (s2 & 0x7f7f7f7f);
	    diffRight = ((~x) & diffRight) | (s1 & (~s2));
	    diffRight &= 0x80808080;
	    diffRight = 0x80808080 - (diffRight >> 7);
	    diffRight ^= 0x80808080;
	    diffRight &= x;
	    s1 ^= diffRight;
	    s2 ^= diffRight;
	    diffRight = (s2 - s1);

            /* compute the down absolute differences */
	    x = topLeft ^ botLeft;
            diffDown = (topLeft | 0x80808080) -
	               (botLeft & 0x7f7f7f7f);
	    diffDown = ((~x) & diffDown) | (topLeft & (~botLeft));
	    diffDown &= 0x80808080;
	    diffDown = 0x80808080 - (diffDown >> 7);
	    diffDown ^= 0x80808080;
	    diffDown &= x;
	    topLeft ^= diffDown;
	    botLeft ^= diffDown;
	    diffDown = (botLeft - topLeft);

            /* The same cool hack used above can be used to find the
	     * pixel-wise max (or min) a word at a time by stopping at the
	     * point where one word contains all the max pixels.  This
	     * is the gradient!
	     */

            /* find the max of diffDown and diffRight, this is the gradient */
	    x = diffDown ^ diffRight;
            gradient = (diffDown | 0x80808080) -
	               (diffRight & 0x7f7f7f7f);
	    gradient = ((~x) & gradient) | (diffDown & (~diffRight));
	    gradient &= 0x80808080;
	    gradient = 0x80808080 - (gradient >> 7);
	    gradient ^= 0x80808080;
	    gradient &= x;
	    gradient ^= diffRight;

            /* write the completed gradient word */
	    ULONG_ACCESS (idp) = gradient;

	    /* don't read twice if you don't have too! */
	    topLeft = topRight;

	    idp++;
	    isp++;

	}  /* for each column */

	IP_YIELD (aborted);

    }  /* for each row */

    /* either we bomb or it worked */
    NORMAL_RETURN;

aborted:
    return (ia_aborted);

}  /* ip_downRightL1Gradient8bppFast () */



/*
 * $Log: i_grad8.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 * Imported sources
 *
 * Revision 1.4  1998/01/13 23:48:01  ddavies
 * i_grad8.c -> /product/ipcore/ipshared/src/RCS/i_grad8.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.3  1996/07/25 16:48:18  dferrell
 * dferrell on Thu Jul 25 09:47:46 PDT 1996
 *
 * i_grad8.c -> /product/ipcore/ipshared/src/RCS/i_grad8.c,v
 *
 * Changed "operator" parameter name in gradient code to "opType" since
 * in C++ "operator" is a reserved word.
 *
 * Revision 1.2  1996/07/02 18:37:10  dferrell
 * dferrell on Tue Jul 2 11:36:46 PDT 1996
 *
 * i_grad8.c -> /product/ipcore/ipshared/src/RCS/i_grad8.c,v
 *
 * Added keyword for nag scripts.
 *
 * Revision 1.1  1996/07/01 17:09:10  dferrell
 * Initial revision
 *
 *
 */

