/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
** File:        i_granul.c
**
** Exports:
** 
**	Grayscale linear granulometry functions:
**	    i_grayLinearGranulFunctionLine16 ()
**	    i_grayLinearGranulFunctionLine32 ()
**
**	Grayscale Granulometry Computation:
**	    i_granulometryByOpeningLine
**
**
** Author: Luc Vincent
** Hacked into IPSHARED by P. MacLeod.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#include "except.prv"

	/* prototypes */
#include "shrip.prv"
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_granul.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/*
**  Exported Functions:
**************************
*
*  i_grayLinearGranulFunctionLine16() Grayscale granulfct of INT16 line
*  i_grayLinearGranulFunctionLine32() Same, for INT32
*
*/

/* PUBLIC */
/* ===================================================================
**  i_grayLinearGranulFunctionLine16()  --  Grayscale linear granulometry function
**					    for one line of Int16
**
**  ARGUMENTS:  lineS - input line on which to operate
**              length - number of pixels on this line
**              border_mode - cGeodesic or cEuclidean
**              lineD - output line
**
**  OPERATION/PRINCIPLE:
**
** This routine takes an input line that contains only positive values.
** For each pixel p of this line, we first go left until a pixel of lower
** value is found, in which case this value becomes the new current
** value. This is iterated until the beginning of the line is found.
** The same is done towards the right, until the end of line is reached.
** The integral of the current value at each location is the output
** value of pixel p (value given to pixel p in lineD).
** 
** What does this do in fact? If a pixel p has a high value in the
** original line, AND there is a wide area of pixels with high value
** in its neighborhood, then the output value of p will be very high.
** If however, p is high, but is kind of an isolated peak, its output
** value will be relatively low. Low input pixels result in low output
** pixels.
** 
** Using this routine, we can extract the "widest area of high value"
** in a histogram, without having to use any threshold of height nor
** any parameter characterizing the width. Width and height are
** automatically combiend here, and the global maximum of the resulting
** histogram is typically what a user of this routine would eb interested
** in extracting.
** 
** The routine is very fast and works in O(N), N being the length
** of the input line.
**
**  RETURNS:	nothing
**
**  NOTES:      There MUST be allocated space (at least one bin) on
**              either side of lineS and lineD, otherwise we will be addressing
**              pixels outside allocated bounds (abr)
=================================================================== */
Int32 CDECL 
i_grayLinearGranulFunctionLine16 (
    IPState    *pIPState,
    Int16      *lineS,
    Int16	length,
    Int16	border_mode,
    Int32      *lineD)
{
    PROLOGUE;
    register Int16  *endlineS;
    register Int16  *pixS, *pixSleft, *pixSright;
    register Int32  *pixD, *pixA;
    register Int16 **pixP;
    Int32           *lineA_array,  *lineA;
    Int16          **lineP_array, **lineP;

    ExceptRaise (length <= 0, ia_invalidParm);
    ExceptRaise (border_mode != cEuclidean && border_mode != cGeodesic,
		 ia_invalidParm);

        /*----- Auxiliary line buffer for result of second scan ----*/

    if ((lineA_array = (Int32 *)
	 MALLOC ((length + 2) * sizeof(Int32))) == NULL)
	return ia_nomem;
    lineA = lineA_array + 1;

        /*----- Array of pixel pointers -----*/

    if ((lineP_array = (Int16 **)
	 MALLOC ((length + 2) * sizeof(Int16 *))) == NULL)
	return ia_nomem;
    lineP = lineP_array + 1;

        /*----- Initializations -----*/

    endlineS = lineS + length;

        /*----- Propagation from left to right -----*/

    *(lineS - 1) = 0;
    *(lineD - 1) = 0;
    *(lineP - 1) = lineS - 1;

    for (pixS = lineS, pixD = lineD, pixP = lineP;
	 pixS < endlineS; pixS++, pixD++, pixP++) {

	if (*(pixS - 1) < *pixS) {
	    *pixD = *(pixD - 1) + ((Int32) *pixS);
	    *pixP = pixS - 1;
	}

	else if (*(pixS - 1) == *pixS) {
	    if (border_mode == cEuclidean ||
		*(pixP - 1) >= lineS)
		*pixD = *(pixD - 1) + ((Int32) *pixS);
	    else
		*pixD = *(pixD - 1) + (((Int32) *pixS) << 1);
	    *pixP = *(pixP - 1);
	}

	else {
	    pixSleft = *(pixP - 1);
	    while (*pixSleft > *pixS)
		pixSleft = *(lineP + (pixSleft - lineS));

	    if (pixSleft >= lineS)
		*pixD = *(lineD + (pixSleft - lineS)) +
		    ((pixS - pixSleft) * ((Int32) *pixS));
	    else {
		*pixD = (pixS - pixSleft) * ((Int32) *pixS);
		if (border_mode == cGeodesic)
		    *pixD <<= 2;
	    }

	    *pixP = pixSleft;
	}
    }

        /*----- Propagation from right to left -----*/

    pixA  = lineA + length;

    *pixS = 0;
    *pixA = 0;
    *pixP = pixS;

    while (--pixS >= lineS) {

	pixA--;
	pixP--;

	if (*pixS > *(pixS + 1)) {
	    *pixA = *(pixA + 1) + ((Int32) *pixS);
	    *pixP = pixS + 1;
	}

	else if (*pixS == *(pixS + 1)) {
	    if (border_mode == cEuclidean ||
		*(pixP + 1) < endlineS)
		*pixA = *(pixA + 1) + ((Int32) *pixS);
	    else
		*pixA = *(pixA + 1) + (((Int32) *pixS) << 1);
	    *pixP = *(pixP + 1);
	}

	else {
	    pixSright = *(pixP + 1);
	    while (*pixSright > *pixS)
		pixSright = *(lineP + (pixSright - lineS));

	    if (pixSright < endlineS)
		*pixA = *(lineA + (pixSright - lineS)) +
		    ((pixSright - pixS) * ((Int32) *pixS));
	    else {
		*pixA = (pixSright - pixS) *((Int32) *pixS);
		if (border_mode == cGeodesic)
		    *pixA <<= 2;
	    }

	    *pixP = pixSright;
	}
    }

        /*----- Addition of these two "distance functions" -----*/

    for (pixS = lineS, pixD = lineD, pixA = lineA;
	 pixS < endlineS; pixS++, pixD++, pixA++)
	*pixD += (*pixA - ((Int32) *pixS));

        /*----- Frees up memory -----*/

    FREE(lineP_array);
    FREE(lineA_array);

    NORMAL_RETURN;
} /* GrayLinearGranulFunctionLine16 */



/* ===================================================================
**  i_grayLinearGranulFunctionLine32()  Grayscale linear granulometry function
**                                      for one line of Int32
**
**  ARGUMENTS:  lineS - input line on which to operate
**              length - number of pixels on this line
**              border_mode - cGeodesic or cEuclidean
**              lineD - output line
**
**  OPERATION:	As explained in i_grayLinearGranulFunctionLine16
**
**  RETURNS:	error code
**  NOTES:      There MUST be allocated space (at least one bin) on
**              either side of lineS and lineD, otherwise we will be addressing
**              pixels outside allocated bounds (abr)
=================================================================== */
Int32 CDECL 
i_grayLinearGranulFunctionLine32 (
    IPState    *pIPState,
    Int32      *lineS,
    Int16	length,
    Int16	border_mode,
    Int32      *lineD)
{
    PROLOGUE;
    register Int32  *endlineS;
    register Int32  *pixS, *pixSleft, *pixSright;
    register Int32  *pixD, *pixA;
    register Int32 **pixP;
    Int32           *lineA_array,  *lineA;
    Int32          **lineP_array, **lineP;

    ExceptRaise (length <= 0, ia_invalidParm);
    ExceptRaise (border_mode != cEuclidean && border_mode != cGeodesic,
		 ia_invalidParm);
    ExceptRaise (lineS == lineD, ia_invalidParm);

        /*----- Auxiliary line buffer for result of second scan ----*/

    if ((lineA_array = (Int32 *)
	 MALLOC ((length + 2) * sizeof(Int32))) == NULL)
	return ia_nomem;
    lineA = lineA_array + 1;

        /*----- Array of pixel pointers -----*/

    if ((lineP_array = (Int32 **)
	 MALLOC ((length + 2) * sizeof(Int32 *))) == NULL)
	return ia_nomem;
    lineP = lineP_array + 1;

        /*----- Initializations -----*/

    endlineS = lineS + length;

        /*----- Propagation from left to right -----*/

    *(lineS - 1) = 0;
    *(lineD - 1) = 0;
    *(lineP - 1) = lineS - 1;

    for (pixS = lineS, pixD = lineD, pixP = lineP;
	 pixS < endlineS; pixS++, pixD++, pixP++) {

	if (*(pixS - 1) < *pixS) {
	    *pixD = *(pixD - 1) + ((Int32) *pixS);
	    *pixP = pixS - 1;
	}

	else if (*(pixS - 1) == *pixS) {
	    if (border_mode == cEuclidean ||
		*(pixP - 1) >= lineS)
		*pixD = *(pixD - 1) + ((Int32) *pixS);
	    else
		*pixD = *(pixD - 1) + (((Int32) *pixS) << 1);
	    *pixP = *(pixP - 1);
	}

	else {
	    pixSleft = *(pixP - 1);
	    while (*pixSleft > *pixS)
		pixSleft = *(lineP + (pixSleft - lineS));

	    if (pixSleft >= lineS)
		*pixD = *(lineD + (pixSleft - lineS)) +
		    ((pixS - pixSleft) * ((Int32) *pixS));
	    else {
		*pixD = (pixS - pixSleft) * ((Int32) *pixS);
		if (border_mode == cGeodesic)
		    *pixD <<= 2;
	    }

	    *pixP = pixSleft;
	}
    }

        /*----- Propagation from right to left -----*/

    pixA  = lineA + length;

    *pixS = 0;
    *pixA = 0;
    *pixP = pixS;

    while (--pixS >= lineS) {

	pixA--;
	pixP--;

	if (*pixS > *(pixS + 1)) {
	    *pixA = *(pixA + 1) + ((Int32) *pixS);
	    *pixP = pixS + 1;
	}

	else if (*pixS == *(pixS + 1)) {
	    if (border_mode == cEuclidean ||
		*(pixP + 1) < endlineS)
		*pixA = *(pixA + 1) + ((Int32) *pixS);
	    else
		*pixA = *(pixA + 1) + (((Int32) *pixS) << 1);
	    *pixP = *(pixP + 1);
	}

	else {
	    pixSright = *(pixP + 1);
	    while (*pixSright > *pixS)
		pixSright = *(lineP + (pixSright - lineS));

	    if (pixSright < endlineS)
		*pixA = *(lineA + (pixSright - lineS)) +
		    ((pixSright - pixS) * ((Int32) *pixS));
	    else {
		*pixA = (pixSright - pixS) *((Int32) *pixS);
		if (border_mode == cGeodesic)
		    *pixA <<= 2;
	    }

	    *pixP = pixSright;
	}
    }

        /*----- Addition of these two "distance functions" -----*/

    for (pixS = lineS, pixD = lineD, pixA = lineA;
	 pixS < endlineS; pixS++, pixD++, pixA++)
	*pixD += (*pixA - ((Int32) *pixS));

        /*----- Frees up memory -----*/

    FREE(lineP_array);
    FREE(lineA_array);

    NORMAL_RETURN;
} /* ip_grayLinearGranulFunctionLine32 */


/* THIS DOESN'T COMPILE YET, but I'm checking the file in so that iadef.prv
   is not included */
#if 0
/* ===================================================================
**  GranulometryByOpeningLine()  --  Contributes to computing the input
**				     granulometric curve, on a given line
**
**  ARGUMENTS:  lineS - input line on which to operate
**              length - number of pixels on this line
**              curve - granulometric curve being computed
**              border_mode - cGeodesic or cEuclidean
**
**  OPERATION:	Contributes to the computation of differenciated
**              granulometric curve by openings curve.
**
**  RETURNS:	nothing
**  NOTES:      This is the core of the linear granulometry algorithms.
**              lineS is not modified by routine. LineL, which can
**              be anything as input, can also contain any value
**              after a call to this routine
**  SEE ALSO:   linearGranulometryByOpening in ipwrap
=================================================================== */
EXPORT void CDECL
i_granulometryByOpeningLine(
    IPState    *pIPState,
    UInt8      *lineS,
    Int32	length,
    Int32	curve,
    Int32	border_mode)
PIX8U   *lineS;
DIMENSION length;
GRANUL   *curve;
FLAG      border_mode;
{
    PROLOGUE;
    INT16           label, lbl;
    REGISTER PIX8U *pixS, *pix1S, *pix2S, *pix1b, *pix2b;
    PIX8U          *pix_init;
    REGISTER PIX8U *endlineS;
    REGISTER PIX8U  val, val1, val2, newval;
    INT32           width;
    INT32           size_max;
    GRANDATA       *granul_data;
    INT16          *lineL_array, *lineL;
    INT16          *label_width;

    PROCNAME("GranulometryByOpeningLine16");

    CheckParameters(length > 0);
    CheckParameters(border_mode == EUCLIDEAN || border_mode == GEODESIC);

        /*----- Allocate line buffer where labels
	             of "extended" maxima are stored -----*/

    if ((lineL_array = (INT16 *)
	 CALLOC ((length + 1), sizeof(INT16))) == NULL)
	ERROR(ERR_OUT_OF_MEMORY);
    lineL = lineL_array + 1;

         /*----- Array where width of these maxima is stored -----*/

    if ((label_width = (INT16 *)
	 MALLOC (((length / 2) + 2) * sizeof(INT16))) == NULL)
	ERROR(ERR_OUT_OF_MEMORY);
    
        /*----- Initializations -----*/

    label       = 0;
    size_max    = MIN((INT32) length - 1, (INT32) GRANUL_GET_MAXVAL(curve));
    granul_data = GRANUL_GET_DATA(curve);

    *(lineS - 1) = *(lineS + length) =
	(border_mode == EUCLIDEAN ? MINPIX8U : MAXPIX8U);


        /*----- Label rightmost pixel of each regional maximum -----
	  ----- then propagates the labels and build the curve -----
	  ----- contribution of each maximum                   -----*/

    endlineS = lineS + length;
    pixS = lineS;

    while (1) {

            /* Goes right down as much as possible until potential
	       maximum is extracted */

	while (*(pixS-1) >= *pixS && pixS < endlineS)
	    pixS++;

	if (pixS >= endlineS)
	    break;

	    /* The current pixel is > than its left neighbor */

	val = *pixS++;

	    /* Perhaps a maximum, check further */

	if (*pixS <= val) {
	    pix_init = pixS;
	    while (*pixS == val && pixS < endlineS)
		pixS++;

	        /* This is indeed a maximum: */

	    if (*pixS < val) {

		++label;

		    /* now goes down around this maximum until
		       an unlabelled pixel of higher value
		       than current value is found, filling the
		       curve structure along the way */

		pix1S = pix_init - 2;
		pix2S = pixS;

		    /* Now, pix1S is the first pixel on the left of
		       the plateau and pix2S is the first pixel on the
		       right. Start going down... */

		val1 = *pix1S;
		val2 = *pix2S;
		while (1) {

		        /* newval = current value of local threshold
			   around considered maximum */

		    newval = MAX(val1, val2);
		    width = (pix2S - pix1S) - 1;

		        /* Adds contribution of this maximum to the
			   granulometric array, or exits if already
			   too wide. */

		    if (width > size_max) {
			pixS = pix2S;
			break;
		    }
		    granul_data[width] += (GRANDATA) (width * (val - newval));

		        /* Go left of pix1 and/or right of pix2,
			   depending on which of val1 or val2 is
			   largest. If they are equal, go left and right */

		    val = newval;
		    pix1b = pix1S;
		    pix2b = pix2S;
		    if (val1 <= val2) {

			    /* Must go right as long as we are on
			       a  plateau at altitude val2 */

			while (pix2b < endlineS && *pix2b == val2)
			    pix2b++;

			    /* If pixel encountered at the end of this
			       exercise is greater than val2, it means
			       we are going up again, so we are out of the
			       "influence" of the current maximum; in this
			       case, breaks out of the loop */

			if (*pix2b >= val2) {
			    pixS = pix2b;
			    break;
			}
		    }
		    if (val2 <= val1) {
			pix1b--;

			    /* Here, we must go left as long as we
			       are at level val1, *or*, we encounter
			       a previously "filled" (i.e. labelled)
			       "influence zone" of a maximum */

			while (*pix1b == val1 || (lbl = lineL[pix1b - lineS])) {
			    if (*pix1b == val1)
				pix1b--;

			        /* If lbl is set, we can skip over a
				   previously labelled maximum area... */

			    else
				pix1b -= label_width[lbl];
			}

			    /* This case cannot happen if size_max is
			       equal to full line width and we are in
			       Euclidean mode, because in this
			       latter case, all the maxima before current
			       one would have been filled! */

			if (*pix1b > val1) {
			    pixS = pix2b;
			    break;
			}
		    }

		        /* Updates val1 and val2, pix1S and pix2S */

		    val1 = *pix1b;
		    val2 = *pix2b;
		    pix1S = pix1b;
		    pix2S = pix2b;
		}

		    /* We are done with the propagation around the
		       current maximum, set its width in label_width[]
		       (useful to skip over it later if needed) */

		lineL[pix2S - lineS - 1] = label;
		label_width[label] = (INT16) (pix2S - pix1S) - 1;
	    }
	}
    }

        /*----- Cleans up memory -----*/

    MFREE(label_width);
    CFREE(lineL_array);

} /* GranulometryByOpeningLine8U */
#endif


