 /*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_greduc.c: Grayscale min/max reduction
 *
 * Exports:
 *	i_reduce2Min	    -	    2x reduction, minimum
 *	i_reduce2Max	    -	    2x reduction, maximum
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "shrip.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrrast.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_greduc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* PUBLIC */
/*
 * i_reduce2Min8bpp()
 *
 *	Grayscale min reduction. Each output pixel is taken from the lowest 
 *	of 4 input pixels.
 *
 * Arguments:
 *	pD	    -	    Destination image, including frame
 *	wD	    -	    Width of destination image, including frame
 *	hD	    -	    Height of destination image, including frame
 *	dBpl	    -	    distance between line of destination image
 *	pS	    -	    Source image, including frame
 *	sBpl	    -	    distance between line of source image
 *	
 * Returns:
 *	ia_successful or error code 
 */

Int32 CDECL 
i_reduce2Min8bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    Int32 line;
    /* Get rid of the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBits;
    pD += (dBpl << cLogFrameBits) + cFrameBits;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    /* Loop through the destination image, generating lines */
    for (line = 0; line < hD; line++)
    {
	ip_reduce2Min8bpp (pIPState, pD, wD, pS, sBpl);
	pD += dBpl;
	pS += sBpl << 1;
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/*
 * i_reduce2Max8bpp()
 *
 *	Grayscale max reduction. Each output pixel is taken from the highest
 *	of 4 input pixels. Works on 8 bit/pixel images only, at the moment.
 *
 * Arguments:
 *	pD	    -	    Destination image, including frame
 *	wD	    -	    Width of destination image, including frame
 *	hD	    -	    Height of destination image, including frame
 *	dBpl	    -	    distance between line of destination image
 *	pS	    -	    Source image, including frame
 *	sBpl	    -	    distance between line of source image
 *	
 * Returns:
 *	ia_successful or error code 
 */

Int32 CDECL 
i_reduce2Max8bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    Int32 line;
    /* Get rid of the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBits;
    pD += (dBpl << cLogFrameBits) + cFrameBits;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    /* Loop through the destination image, generating lines */
    for (line = 0; line < hD; line++)
    {
	ip_reduce2Max8bpp (pIPState, pD, wD, pS, sBpl);
	pD += dBpl;
	pS += sBpl << 1;
    }

    NORMAL_RETURN;
}

/*
 * i_reduce2Min4bpp()
 *
 *	Grayscale min reduction. Each output pixel is taken from the lowest 
 *	of 4 input pixels.
 *
 * Arguments:
 *	pD	    -	    Destination image, including frame
 *	wD	    -	    Width of destination image, including frame
 *	hD	    -	    Height of destination image, including frame
 *	dBpl	    -	    distance between line of destination image
 *	pS	    -	    Source image, including frame
 *	sBpl	    -	    distance between line of source image
 *	
 * Returns:
 *	ia_successful or error code 
 */

Int32 CDECL 
i_reduce2Min4bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    Int32 line;
    /* Get rid of the frame */
    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    /* Loop through the destination image, generating lines */
    for (line = 0; line < hD; line++)
    {
	ip_reduce2Min4bpp (pIPState, pD, wD, pS, sBpl);
	pD += dBpl;
	pS += sBpl << 1;
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/*
 * i_reduce2Max4bpp()
 *
 *	Grayscale max reduction. Each output pixel is taken from the highest
 *	of 4 input pixels. Works on 8 bit/pixel images only, at the moment.
 *
 * Arguments:
 *	pD	    -	    Destination image, including frame
 *	wD	    -	    Width of destination image, including frame
 *	hD	    -	    Height of destination image, including frame
 *	dBpl	    -	    distance between line of destination image
 *	pS	    -	    Source image, including frame
 *	sBpl	    -	    distance between line of source image
 *	
 * Returns:
 *	ia_successful or error code 
 */

Int32 CDECL 
i_reduce2Max4bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    Int32 line;
    /* Get rid of the frame */
    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    /* Loop through the destination image, generating lines */
    for (line = 0; line < hD; line++)
    {
	ip_reduce2Max4bpp (pIPState, pD, wD, pS, sBpl);
	pD += dBpl;
	pS += sBpl << 1;
    }

    NORMAL_RETURN;
}


/* 
 * $Log: i_greduc.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 * Imported sources
 *
 * Revision 1.6  1998/01/13 23:48:04  ddavies
 * i_greduc.c -> /product/ipcore/ipshared/src/RCS/i_greduc.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.5  1995/12/13 02:44:08  ejaquith
 * i_greduc.c -> /product/ipcore/ipshared/src/RCS/i_greduc.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.4  1995/10/18  03:53:15  ddavies
 * i_greduc.c -> /product/ipcore/ipshared/src/RCS/i_greduc.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.3  1995/02/14  02:58:47  macleod
 * macleod on Mon Feb 13 18:58:18 PST 1995
 *
 * i_greduc.c -> /product/ipcore/ipshared/src/RCS/i_greduc.c,v
 *
 * Added 4-bit/pixel versions of min-of-four and max-of-four x2 reductions
 *
 * Revision 1.2  1994/07/12  23:55:43  ejaquith
 * i_greduc.c -> /product/ipcore/ipshared/src/RCS/i_greduc.c,v
 *
 * removed #include "iadef.prv".
 *
 * Revision 1.1  1994/06/06  23:26:31  macleod
 * Initial revision
 *
 */
