/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_grybin.c: contains ops combining gray and binary pixrects
 *
 *         Combine Gray Pixrect with Binary pixrect:
 *                  Int32     	i_opGrayBin() accelerated with assembly code
 *                  Int32	i_opGrayBinIP() accelerated with assembly code
 *
 *         Select between two gray Pixrect using binary pixrect:
 *                  Int32       i_graySelect()
 *
 */


#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrscal.pub"
#include "shrrast.pub"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_grybin.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* This file contains the upper level code used to logically combine
 * binary pixrects with gray pixrects.  The primitives are implemented in
 * migrybin.c (C implementation) and ssgrybin.c (SPARC Assembly).
 *
 * These routines combine binary and "gray" images of the same width and
 * height.  The gray is in quotes because the gray images may also be binary.
 * Two operations are currently supported: mask and paint.  The operation
 * selectors are the equal to the mask and paint selectors used by the
 * Sun Pixrect library routines (mask = PIX_SRC & PIX_DST, paint =
 * PIX_SRC | PIX_DST).
 *
 * A paint operations makes the gray image black everywhere there's a 1 in
 * the binary image.
 *
 * A mask operation makes the gray image white everywhere there's a 0 in the
 * binary image.
 *
 * If both images are binary, the mask operation returns an image
 * that's the AND of the two source images.
 *
 * If both images are binary, the paint operation returns a images that's
 * the OR of the two source images.
 *
 * Unfortunately for image processing brains everywhere, the mapping between
 * [white <-> black] and [small pixel values <-> big pixel values] works
 * one way for binary images and the opposite way for gray images.  In the
 * binary world, 0 = white; 1 = black.  It came from computer people drawing
 * pictures on screens and putting 1's where they wanted something to appear
 * (all real computers use white blackgrounds, right?).  In the gray world,
 * 0 = black and 1111...11 = white.  The value represent the number of photons
 * hitting a sensor.
 *
 * The mask operation wants to force white to appear in the gray image
 * everywhere there's a 0 in the binary image.  We do this by inverting the
 * binary image, expanding it to cover the gray image, and ORing the
 * results.  Every 0 in the binary image forces the corresponding gray pixel
 * to be all 1's.
 *
 * The paint operation is complementary - we invert the binary image, expand
 * it and AND the results.  Every 1 in the binary image forces the
 * corresponding gray pixel to be 0 (black).
 *
 * The same expansion tables used in the fast expansion procedures (see
 * fexpand.c) are used to expand the binary pixrect to cover the gray pixrect.
 * Note that our expansion is purely horizontal, not vertical.
 *
 * opGrayBinPr combines two source pixrects into a new gray pixrect.
 * opGraybinPrIP combines two source pixrects.  The result is placed in the
 * original gray pixrect.
 */

/* The difference between the accelerated and non-accelerated versions is
 * that each line is processed with assembly code in the accelerated version.
 * We don't bother accelerating the 1 bit/pixel version since it's using the
 * standard rasterop function already (and is pretty fast anyway).
 */



/*********************************************************************
 *************************      Procedures     ***********************
 *********************************************************************/

/* PUBLIC */
/**********************************************************************
 *  i_opGrayBin()
 *	Input: 
 *		pG = source gray image (includes frame)
 *		w = width in pixels (including frame)
 *		h = height (including frame)
 *		gBpl = gray image bytes/line
 *		depth = bits/pixel of source gray image
 *		pB = source binary image
 *		bBpl = binary bytes/line
 *		operation = paint (PIX_SRC | PIX_DST or mask (PIX_SRC & PIX_DST)
 *		pD = destination image (including frame)
 *		dBpl = destination bytes/line

 *	On return: 
 *		destination gray image equals either graySrc masked by
 *		binary image (has white where binary image has 0's)
 *		or graySrc with binary image painted into it (is black
 *		where binary image has 1's)..
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 **********************************************************************/
Int32   CDECL 
i_opGrayBin (
    IPState    *pIPState,
    UInt8      *pG,
    UInt32	w,
    UInt32	h,
    UInt32	gBpl,
    UInt32	depth,
    UInt8      *pB,
    UInt32	bBpl,
    Int32	operation,
    UInt8      *pD,
    UInt32	dBpl)
{
PROLOGUE;
 UInt32      	 wS, hS;
UInt32      	 line;			/* iteration vars */
Int32		 doPaint;		/* boolean */

UInt16          *pTabExp2;              /* 2:1 expansion table byte */
UInt32          *pTabExp4;              /* 4:1 expansion table byte */
UInt32          *pTabExp8;              /* 8:1 expansion table byte */
UInt8		*pDOrig;		/* remember the start of the
					 * dest image so we can clear the
					 * right frame.
					 */
    wS = w - 2 * cFrameBits;
    hS = h - 2 * cFrameBits;

	/* If pG is a binary image, copy pG into pD and do the specified
	 * operation with the binary image */
    if (depth == 1)
    {
	i_rasterOp (pIPState, pD, 0, 0, w, h, depth, dBpl, PIX_SRC,
		    pG, 0, 0, depth, gBpl);
        i_rasterOp (pIPState, pD, cFrameBits, cFrameBits, wS, hS,
		    depth, dBpl, operation, pB, cFrameBits, cFrameBits,
		    depth, bBpl);
	NORMAL_RETURN;
    }

	/* The source gray image is in fact gray, so we need to expand
	 * the binary image to match it and carefully decide what
	 * "white" and "black" mean in the gray world.
	 * First, adjust the source and dest pointers to avoid the pads */
    pD += (dBpl << cLogFrameBits) + depth*cFrameBytes;
    pG += (gBpl << cLogFrameBits) + depth*cFrameBytes;
    pB += (bBpl << cLogFrameBits) +   cFrameBytes;
    pDOrig = pD;			/* need this to clear right frame */

    doPaint = (operation == (PIX_SRC | PIX_DST));

    if (depth == 2)
    {
	    /* make the 2:1 expansion tables */
	pTabExp2 = NULL;
	if ((__error__ = i_makeExp2LUT (pIPState, 1, &pTabExp2)) != ia_successful)
	    NORMAL_RETURN;

	    /* Loop through all lines.  With each line, read and
	     * expand each binary byte.  Combine the expanded shorts into
	     * a word.  AND or OR the inverted word with the current gray
	     * source data and store the result.  We pass the number of
	     * bytes in the gray pixrect's line to the assembly routine.
	     * This can be used to derive the width of the binary pixrect,
	     * but the width of the binary pixrect cannot be used to
	     * derive the width of the gray pixrect (can't replace lost
	     * LS bits). */
	for (line = 0; line < hS; line++, pD += dBpl, pG += gBpl, pB += bBpl)
	{
	    if (doPaint) /* do the paint operation */
		ip_paint2bppLine (pIPState, pG, pB, pD, wS/4, pTabExp2);
	    else	/* do the mask operation */
		ip_mask2bppLine (pIPState, pG, pB, pD, wS/4, pTabExp2);
	}
	if (!doPaint) ip_clearRightFrame (pIPState, pDOrig, wS, hS, dBpl, 2);
	FREE(pTabExp2);
    }
    else
    if (depth == 4)
    {
	    /* make the 4:1 expansion tables */
	pTabExp4 = NULL;
	if ((__error__ = i_makeExp4LUT (pIPState, 1, &pTabExp4)) != ia_successful)
	    NORMAL_RETURN;

	    /* Loop through all lines.  With each line, read and
	     * expand each binary byte into a word.  OR the inverted word
	     * with the current gray source data and store the result */
	for (line = 0; line < hS; line++, pD += dBpl, pG += gBpl, pB += bBpl)
	{
	    if (doPaint) /* do the paint operation */
		ip_paint4bppLine (pIPState, pG, pB, pD, wS/2, pTabExp4);
	    else	/* do the mask operation */
		ip_mask4bppLine (pIPState, pG, pB, pD, wS/2, pTabExp4);
	}
	if (!doPaint) ip_clearRightFrame (pIPState, pDOrig, wS, hS, dBpl, 4);
	FREE(pTabExp4);
    }
    else  /* depth == 8 */
    {
	    /* make the 8:1 expansion tables */
	pTabExp8 = NULL;
	if ((__error__ = i_makeExp8LUT (pIPState, 1, &pTabExp8)) != ia_successful)
	    NORMAL_RETURN;

	    /* Loop through all lines.  With each line, read and
	     * expand each binary nibble into a word.  OR the inverted word
	     * with the current gray source data and store the result */
	for (line = 0; line < hS; line++, pD += dBpl, pG += gBpl, pB += bBpl)
	{
	    if (doPaint) /* do the paint operation */
		ip_paint8bppLine (pIPState, pG, pB, pD, wS, pTabExp8);
	    else	/* do the mask operation */
		ip_mask8bppLine (pIPState, pG, pB, pD, wS, pTabExp8);
	}
	if (!doPaint) ip_clearRightFrame (pIPState, pDOrig, wS, hS, dBpl, 8);
	FREE(pTabExp8);
    }

    NORMAL_RETURN;
}



/* PUBLIC */
/**********************************************************************
 *    i_opGrayBinIP()
 *	Input: 
 *		pG = source gray image (includes frame)
 *		w = width in pixels (including frame)
 *		h = height (including frame)
 *		gBpl = gray image bytes/line
 *		depth = bits/pixel of source gray image
 *		pB = source binary image
 *		bBpl = binary bytes/line
 *		operation = paint (PIX_SRC | PIX_DST) or
 *			    mask (PIX_SRC & PIX_DST)
 *
 *	On return: 
 *		source gray image is replaced by image that equals either 
 *		graySrc masked by binary image or binary image painted into
 *		graySrc.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 **********************************************************************/
Int32  CDECL 
i_opGrayBinIP (
    IPState    *pIPState,
    UInt8      *pG,
    UInt32	w,
    UInt32	h,
    UInt32	gBpl,
    UInt32	depth,
    UInt8      *pB,
    UInt32	bBpl,
    Int32	operation)
{
PROLOGUE;
 UInt32      	 wS, hS;
UInt32      	 line;			/* iteration vars */
Int32		 doPaint;		/* boolean */

UInt16          *pTabExp2;              /* 2:1 expansion table byte */
UInt32          *pTabExp4;              /* 4:1 expansion table byte */
UInt32          *pTabExp8;              /* 8:1 expansion table byte */
UInt8		*pDOrig;		/* remember the start of the
					 * dest image so we can clear the
					 * right frame.
					 */
    wS = w - 2 * cFrameBits;
    hS = h - 2 * cFrameBits;

	/* If prG is a binary image, just and do the specified
	 * operation with the binary image */
    if (depth == 1)
    {
	i_rasterOp (pIPState, pG, cFrameBits, cFrameBits,
		    wS, hS, depth, gBpl, operation, pB,
		    cFrameBits, cFrameBits, depth, bBpl);
	NORMAL_RETURN;
    }
	/* if this is a gray operation, expand each binary
	 * byte by a factor of "depth", invert it and OR the results.
	 */
	/* adjust the source and dest pointers to avoid the pads */
    pG += (gBpl << cLogFrameBits) + depth*cFrameBytes;
    pB += (bBpl << cLogFrameBits) +   cFrameBytes;
    pDOrig = pG;			/* need this to clear right frame */

    doPaint = (operation == (PIX_SRC | PIX_DST));

    if (depth == 2)
    {
	    /* make the 2:1 expansion tables */
	pTabExp2 = NULL;
	if ((__error__ = i_makeExp2LUT (pIPState, 1, &pTabExp2)) != ia_successful)
	    NORMAL_RETURN;

	    /* Loop through all lines.  With each line, read and
	     * expand each binary byte.  Combine the expanded shorts into
	     * a word.  AND or OR the inverted word with the current gray
	     * source data and store the result.  We pass the number of
	     * bytes in the gray pixrect's line to the assembly routine.
	     * This can be used to derive the width of the binary pixrect,
	     * but the width of the binary pixrect cannot be used to
	     * derive the width of the gray pixrect (can't replace lost
	     * LS bits). */
	for (line = 0; line < hS; line++, pG += gBpl, pB += bBpl)
	{
	    if (doPaint) /* do the paint operation */
		ip_paint2bppLine (pIPState, pG, pB, pG, wS/4, pTabExp2);
	    else	/* do the mask operation */
		ip_mask2bppLine (pIPState, pG, pB, pG, wS/4, pTabExp2);
	}
	if (!doPaint) ip_clearRightFrame (pIPState, pDOrig, wS, hS, gBpl, 2);
	FREE(pTabExp2);
    }
    else
    if (depth == 4)
    {
	    /* make the 4:1 expansion tables */
	pTabExp4 = NULL;
	if ((__error__ = i_makeExp4LUT (pIPState, 1, &pTabExp4)) != ia_successful)
	    NORMAL_RETURN;

	    /* Loop through all lines.  With each line, read and
	     * expand each binary byte into a word.  OR the inverted word
	     * with the current gray source data and store the result */
	for (line = 0; line < hS; line++, pG += gBpl, pB += bBpl)
	{
	    if (doPaint) /* do the paint operation */
		ip_paint4bppLine (pIPState, pG, pB, pG, wS/2, pTabExp4);
	    else	/* do the mask operation */
		ip_mask4bppLine (pIPState, pG, pB, pG, wS/2, pTabExp4);
	}
	if (!doPaint) ip_clearRightFrame (pIPState, pDOrig, wS, hS, gBpl, 4);
	FREE(pTabExp4);
    }
    else  /* depth == 8 */
    {
	    /* make the 8:1 expansion tables */
	pTabExp8 = NULL;
	if ((__error__ = i_makeExp8LUT (pIPState, 1, &pTabExp8)) != ia_successful)
	    NORMAL_RETURN;

	    /* Loop through all lines.  With each line, read and
	     * expand each binary nibble into a word.  OR the inverted word
	     * with the current gray source data and store the result */
	for (line = 0; line < hS; line++, pG += gBpl, pB += bBpl)
	{
	    if (doPaint) /* do the paint operation */
		ip_paint8bppLine (pIPState, pG, pB, pG, wS, pTabExp8);
	    else	/* do the mask operation */
		ip_mask8bppLine (pIPState, pG, pB, pG, wS, pTabExp8);
	}
	if (!doPaint) ip_clearRightFrame (pIPState, pDOrig, wS, hS, gBpl, 8);
	FREE(pTabExp8);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/**********************************************************************
 *  i_graySelect()
 * 
 *	Toggle between foreground and background gray images based
 *	on a binary or 4-bit selection image.
 *	Either or the foreground or the background or both may be
 *	missing.  If the foreground is missing, assume it's black.
 *	If the background is missing, assume it's white.
 *	The gray input images (if they exist) and the selector image
 *	and the gray output image must all be the same size.
 *	The two source gray images must be 8 bits/pixel.  The two gray
 *	images must also have the same bytes/line. 
 *	The toggle image must be the same size as all the other images, and 
 *	can be either binary or 4 bits/pixel.
 *
 *	Note that since the 4-bit gray selector was probably produced from
 *	a binary image, the sense of the pixels in the 4-bit image is
 *	inverted.  In binary images, 0 = white.  In gray images,
 *	0 = black.  Therefore, the conversion routine inverted the pixels.
 *	Even though the selector pixels don't represent black and white,
 *	the convention was maintained.
 *
 *	Therefore, a binary selector pixel value of 0 selects the
 *	background image and a 1 selects the foreground image.
 *	A 4-bit selector pixel value of 0 selects the foreground image and
 *	a 4-bit selector pixel value of 0xF selects the background image.
 *
 *	Input: 
 *		pBack = first source gray image (includes frame)
 *		pFore = second source gray image (includes frame)
 *		width = width of all images in pixels (including frame)
 *		height = height of all images in pixels (including frame)
 *		gBpl = gray image bytes/line
 *		sDepth = bits/pixel of source gray images and dest gray image
 *		pT = toggle image, 0 = first image, 1 = second
 *                   at each pixel. If the toggle image is 4-bit,
 *                   0 = second image and 0xF = first image.
 *		     For other values, the images are blended.
 *              tBpl = binary image bytes/line
 *              tDepth = depth (bits/pixel) of toggle image, 1 or 4
 *              pDest = Destination gray image (includes frame)
 *              dBpl = Destination image bytes/line
 *
 *	On return: 
 *		Each pixel in the destination image contains either
 *              a corresponding pixel from the first source image, or 
 *              a pixel from the second image, or a blended combination
 *		of both depending in the value of the corresponding
 *		mask pixel.
 *              
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 * 
 *
 **********************************************************************/
Int32   CDECL 
i_graySelect (
    IPState    *pIPState,
    UInt8      *pBack,
    UInt8      *pFore,
    UInt32	width,
    UInt32	height,
    UInt32	gBpl,
    UInt32	sDepth,
    UInt8      *pT,
    UInt32	tBpl,
    UInt32	tDepth,
    UInt8      *pD,
    UInt32	dBpl)
{
    PROLOGUE;
    UInt32  *mask = NULL;
    Int32   i;
    UInt8  *pDOrig;
    Int32   longestAccess1, longestAccess2;
					/* the longest memory access
					 * allowed for the current processor
					 * and source and dest images. */
    

    if ((sDepth != 8) || ((tDepth != 1) && (tDepth != 4)))
	return ia_invalidParm;

    /* Move the pointers to the image, adjust the size */
    if (pBack != NULL)
	pBack += (gBpl << cLogFrameBits) + sDepth*cFrameBytes;
    if (pFore != NULL)
	pFore += (gBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pDOrig = pD;
    pT += (tBpl << cLogFrameBits) +  tDepth*cFrameBytes;
    width = width - 2 * cFrameBits;
    height = height - 2 * cFrameBits;


	/* Figure out whether the lower level routines will do long
	 * (8-byte) or short (4-byte) accesses.
	 */
    ip_checkImageAccess (pIPState, pFore, gBpl, pD, dBpl, &longestAccess1);
    if ((longestAccess1 != 4) && (longestAccess1 != 8))
	return ia_internal;

    ip_checkImageAccess (pIPState, pBack, gBpl, pD, dBpl, &longestAccess2);
    if ((longestAccess2 != 4) && (longestAccess2 != 8))
	return ia_internal;

	/* Get the real longest access */
    longestAccess1 = MIN(longestAccess1, longestAccess2);

    /* 1-bit selection mask */
    if (tDepth == 1)
    {
	if ((longestAccess1 == 4) || (!ip_longOpIPLibLinked (pIPState)))
	{
		/* The long ops don't need this table, so only
		 * make it if we're using short ops.
		 */
	    mask = NULL;
	    __error__ = i_makeExp8LUT (pIPState, 1, &mask);
	    if (__error__ != ia_successful)
		NORMAL_RETURN;
    
		/* It turns out that we really want the inverse of this
		 * table (1's for the background) for the cases where
		 * either the foreground or background is missing, so
		 * instead of complementing each entry as it's used,
		 * we just complement the table once.
		 */
	    for (i = 0; i < 512; i++)
		mask[i] = ~mask[i];

		/* Branch to one of the four cases:
		 * 1. Both background and foreground supplied;
		 * 2. Only background supplied;
		 * 3. Only foreground supplied;
		 * 4. Neither background nor foreground supplied.
		 */
	    if ((pBack != NULL) && (pFore != NULL))
	    {
		/* Crank through the lines */
		for (i = height; i > 0; i--)
		{
		    ip_grayBinarySelLine (pIPState,
					  (UInt32 *)pBack, (UInt32 *)pFore, pT,
					  (UInt32 *)pD, width, mask);
    
		    pBack += gBpl;
		    pFore += gBpl;
		    pD += dBpl;
		    pT += tBpl;
		}
	    }
	    else
	    if (pBack != NULL)	/* only supplied a background image */
	    {
		/* Crank through the lines */
		for (i = height; i > 0; i--)
		{
		    ip_grayBinarySelLineNoFG (pIPState, (UInt32 *)pBack, pT,
					      (UInt32 *)pD, width, mask);
    
		    pBack += gBpl;
		    pD += dBpl;
		    pT += tBpl;
		}
	    }
	    else
	    if (pFore != NULL)	/* only supplied a foreground image */
	    {
		/* Crank through the lines */
		for (i = height; i > 0; i--)
		{
		    ip_grayBinarySelLineNoBG (pIPState, (UInt32 *)pFore, pT,
					      (UInt32 *)pD, width, mask);
    
		    pFore += gBpl;
		    pD += dBpl;
		    pT += tBpl;
		}
	    }
	    else	/* supplied neither background nor foreground */
	    {
		/* Crank through the lines */
		for (i = height; i > 0; i--)
		{
		    ip_grayBinarySelLineNoFGNoBG (pIPState, pT,
						  (UInt32 *)pD, width, mask);
    
		    pD += dBpl;
		    pT += tBpl;
		}
	    }
	}
	else	/* use long ops */
	{
		/* do whatever's necessary to enter long operation mode. */
	    ip_initLongOps (pIPState);

	    if ((pBack != NULL) && (pFore != NULL))
	    {
		/* Crank through the lines */
		for (i = height; i > 0; i--)
		{
		    ip_grayBinarySelLineLong (pIPState, (UInt32 *)pBack,
					      (UInt32 *)pFore,
					      (UInt32 *)pD, pT, width);
    
		    pBack += gBpl;
		    pFore += gBpl;
		    pD += dBpl;
		    pT += tBpl;
		}
	    }
	    else
	    if (pBack != NULL)	/* only supplied a background image */
	    {
		/* Crank through the lines */
		for (i = height; i > 0; i--)
		{
		    ip_grayBinarySelLineLongNoFG (pIPState, (UInt32 *)pBack,
						  (UInt32 *)pD, pT, width);
    
		    pBack += gBpl;
		    pD += dBpl;
		    pT += tBpl;
		}
	    }
	    else
	    if (pFore != NULL)	/* only supplied a foreground image */
	    {
		/* Crank through the lines */
		for (i = height; i > 0; i--)
		{
		    ip_grayBinarySelLineLongNoBG (pIPState, (UInt32 *)pFore,
						  (UInt32 *)pD, pT, width);
    
		    pFore += gBpl;
		    pD += dBpl;
		    pT += tBpl;
		}
	    }
	    else	/* supplied neither background nor foreground */
	    {
		/* Crank through the lines */
		for (i = height; i > 0; i--)
		{
		    ip_grayBinarySelLineLongNoFGNoBG (pIPState,
						      (UInt32 *)pD, pT, width);
    
		    pD += dBpl;
		    pT += tBpl;
		}
	    }	/* end of using no foreground, no background */

		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}	/* end of case using long ops */
    }		/* end of case using binary selector */
    else	/* The selector plane has 4 bits/pixel */
    {
	    /* See if we can use long operations */
	if ((longestAccess1 == 4) || (!ip_longOpIPLibLinked (pIPState)))
	{
		/* This is the case where we use ordinary (not 64-bit)
		 * operations.
		 * The interpolation and expansion tables used here
		 * are declared in graysel.prv.
		 */
    
		/* Go do the real work */
		/* Branch to one of the four cases:
		 * 1. Both background and foreground supplied;
		 * 2. Only background supplied;
		 * 3. Only foreground supplied;
		 * 4. Neither background nor foreground supplied.
		 */
	    if ((pBack != NULL) && (pFore != NULL))
	    {
		for (i = height;
		     i > 0;
		     i--, pBack += gBpl, pFore += gBpl, pD += dBpl, pT += tBpl)
		{
		    ip_gray4bppSelLine (pIPState, pBack, pFore,
					pT, (UInt32 *)pD, width);
		}
	    }
	    else
	    if (pBack != NULL)	/* Only background supplied, foreground is
				     * black */
	    {
		for (i = height;
		     i > 0;
		     i--, pBack += gBpl, pD += dBpl, pT += tBpl)
		{
		    ip_gray4bppSelLineNoFG (pIPState, pBack,
					    pT, (UInt32 *)pD, width);
		}
	    }
	    else
	    if (pFore != NULL)	/* only foreground image supplied,
				     * background is white.
				     */
	    {
		for (i = height;
		     i > 0;
		     i--, pFore += gBpl, pD += dBpl, pT += tBpl)
		{
		    ip_gray4bppSelLineNoBG (pIPState, pFore,
					    pT, (UInt32 *)pD, width);
		}
	    }
	    else	/* neither foreground nor background supplied.  Assume
		     * foreground is black and background is white.
		     */
	    {
		for (i = height;
		     i > 0;
		     i--, pD += dBpl, pT += tBpl)
		{
		    ip_gray4bppSelLineNoFGNoBG (pIPState,
						pT, (UInt32 *)pD, width);
		}	/* end of loop down rows */
	    }	/* end of no-foreground, no-background case */
	}
	else	/* use long ops */
	{
		/* This is the case where we use ordinary (not 64-bit)
		 * operations.  First, do whatever's necessary to enter long
		 * operation mode.
		 */
	    ip_initLongOps (pIPState);

		/* Go do the real work */
		/* Branch to one of the four cases:
		 * 1. Both background and foreground supplied;
		 * 2. Only background supplied;
		 * 3. Only foreground supplied;
		 * 4. Neither background nor foreground supplied.
		 */
	    if ((pBack != NULL) && (pFore != NULL))
	    {
		for (i = height;
		     i > 0;
		     i--, pBack += gBpl, pFore += gBpl, pD += dBpl, pT += tBpl)
		{
		    ip_gray4bppSelLineLong (pIPState, pBack, pFore, pT,
					    (UInt32 *)pD, width);
		}
	    }
	    else
	    if (pBack != NULL)	/* Only background supplied, foreground is
				     * black */
	    {
		for (i = height;
		     i > 0;
		     i--, pBack += gBpl, pD += dBpl, pT += tBpl)
		{
		    ip_gray4bppSelLineLongNoFG (pIPState, pBack, pT,
						(UInt32 *)pD, width);
		}
	    }
	    else
	    if (pFore != NULL)	/* only foreground image supplied,
				     * background is white.
				     */
	    {
		for (i = height;
		     i > 0;
		     i--, pFore += gBpl, pD += dBpl, pT += tBpl)
		{
		    ip_gray4bppSelLineLongNoBG (pIPState, pFore, pT,
						(UInt32 *)pD, width);
		}
	    }
	    else	/* neither foreground nor background supplied.  Assume
		     * foreground is black and background is white.
		     */
	    {
		for (i = height;
		     i > 0;
		     i--, pD += dBpl, pT += tBpl)
		{
		    ip_gray4bppSelLineLongNoFGNoBG (pIPState,
						    pT, (UInt32 *)pD, width);
		}	/* end of loop down rows */
	    }	/* end of no-foreground, no-background case */
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}	/* end of case using long ops */
    }	    /* end case of using gray selector */

    /* We've now trashed the right frame, so clear it */
    ip_clearRightFrame (pIPState, pDOrig, width, height, dBpl, 8);
    if (mask != NULL) FREE(mask);

    NORMAL_RETURN;
}

/*

$Log: i_grybin.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.18  1998/01/13 23:48:05  ddavies
i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.17  1997/06/06 04:19:06  ddavies
ddavies on Thu Jun 5 21:18:23 PDT 1997

i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v

Replace explicit inner loops (code that processes one scan line) with
calls to private routines that do the same thing.  Add MMX versions of
all these routines.  The MMX versions, like ip_gray4bppSelLineLong instead
of ip_gray4bppSelLine, are about twice as fast.

Revision 1.16  1997/05/16 23:31:36  ddavies
ddavies on Fri May 16 16:29:56 PDT 1997

i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v

A couple of performance improvements for the 4bpp selector case.
First, we expand the selector a byte at a time instead of a nibble at at
time using a different table.
Second, the tables for the 4bpp selector case are now staticly declared
instead of being generated on each call.  This should help especially
when this routine is called on strips.

Revision 1.15  1997/05/16 18:06:21  ddavies
ddavies on Fri May 16 11:01:45 PDT 1997

i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v

Big changes in i_graySelect.  It no longer scales the input gray images.  It can
take either foreground or background or both or neither.  It is now
consistent in how it treats the selector pixel values (white always selects
background instead of 0 always selecting background).
The position of indices used for the look up tables has been modified in
hopes of making it faster.  The contents of the lookup tables are now
symmetric and have values that should give the same results as an MMX
implementation (happening any day now).  The input args have also
been renamed from G1 and G2 to Fore and Back.  The routine now only takes
one set of width and height parameters since all input images must be the
same size.  The C code is about 2x-2.5x faster.  The options that don't use
foreground or background or either one are even faster.

Revision 1.14  1997/01/20 23:01:49  xzhu
xzhu on Mon Jan 20 14:59:43 PST 1997

i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v

In i_greySelect(), fixed bytes swap bug (4 bits mask).

Revision 1.13  1996/02/07 19:27:53  dferrell
dferrell on Wed Feb 7 11:25:50 PST 1996

i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v

Changed the sizes of the 4 to 1 and 4 to 2 LUTs in i_graySelect ()
from UInt32 to Uint8 so that the stack frame for this function does
not exceed 32K (a limit on the Mac).

 * Revision 1.12  1996/01/09  17:52:30  macleod
 * macleod on Tue Jan 9 09:52:08 PST 1996
 *
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * Typecast to silence the NT compiler's whining about converting
 * UInt32 to UInt16.
 *
 * Revision 1.11  1996/01/08  15:49:07  macleod
 * macleod on Mon Jan 8 07:46:44 PST 1996
 *
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 *  i_graySelect can now take a 4-bit selection "mask", in addition to the
 * 1-bit mask it could take before. If given a 4-bit mask, the value at
 * each pixel is a weighting between the "foreground" and "background"
 * images, weighted linearly, i.e. mask*fg + (1-mask)*bg.
 * This required changing the interface slightly.
 *
 * Revision 1.9  1995/12/13  18:54:37  macleod
 * macleod on Wed Dec 13 10:53:35 PST 1995
 *
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * added i_graySelect, which allows one to select between two
 * gray (8-bit) images based on a binary mask. The gray images may
 * be smaller than the binary image, in which case they will be
 * nearest-neighbor sampled to produce the resulting image.
 *
 * Revision 1.8  1995/12/13  02:44:11  ejaquith
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.7  1995/10/18  03:53:19  ddavies
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.6  1994/09/17  17:56:40  ddavies
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * Change i_opGrayBin and i_opGrayBinIP so they use the the newly public procedures
 * i_makeExp2LUT, i_makeExp4LUT and i_expandAndMask8 instead of the old
 * ip_ counterparts.
 *
 * Revision 1.5  1994/07/27  15:59:06  ejaquith
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Made functions use local pointers to expansion tables instead of globals.
 *
 * Revision 1.4  1994/07/13  01:18:42  ddavies
 * ddavies on Tue Jul 12 18:18:11 PDT 1994
 *
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * Fixed bug in clearing right frame when masking 8 bit/pixel images.  Was
 * using h instead of hS as the image height.
 *
 * Revision 1.3  1994/07/11  23:08:53  ddavies
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * Supply the correct args to ip_clearRightFrame.  The old code tried to
 * point to the outside of the frame.  One should point to the inside of the
 * frame.
 *
 * Revision 1.2  1994/02/04  20:15:22  dferrell
 * i_grybin.c -> /product/ipcore/ipshared/src/RCS/i_grybin.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/14  04:29:16  ddavies
 * Initial revision
 *
 * Revision 1.12  1993/12/02  19:21:17  ddavies
 * fgrbinop.c -> /product/ipcore/ipshared/src/RCS/fgrbinop.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.11  1993/11/05  01:24:55  ddavies
 * fgrbinop.c -> /product/ipcore/ipshared/src/RCS/fgrbinop.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


