/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_grydif.c: Grayscale image subtraction
 *
 *                  Int32     	i_subtractGray()
 *		    Int32	i_subtractGrayNoClip ()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /*  prototyes */
#include "shrrast.pub"
#include "shrip.pub"
#include "shrip.prv"

    /* Constants */
#define cHBits8	    0x80808080UL
#define cLBits8	    0x7f7f7f7fUL
#define cHBits4	    0x88888888UL
#define cLBits4	    0x77777777UL

IP_RCSINFO(RCSInfo, "$RCSfile: i_grydif.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");
 
/* PUBLIC */
/***********************************************************************
 *  i_subtractGray ()
 *	Subtract two grayscale images. Clips at zero.
 *
 *	Input: 
 *	pS1 =	 ptr to source image 1 (includes frame), image from which to
 *               subtract
 *	sBpl1 =	 source image 1 bytes/line
 *	pS2 =	 ptr to source image 2 (includes frame), image which is 
 *		 subtracted from pS1
 *	sBpl2 =	 source gray image 2 bytes/line
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	depth =	 bits/pixel of all gray images
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *	dBpl =	 dest gray image bytes/line
 *
 *	On return: 
 *		destination gray image that equals the clipped-at-zero
 *		difference between source gray images, pS1 - pS2
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	image depth was not 4 or 8
 ***********************************************************************/
Int32   CDECL 
i_subtractGray (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt32	sBpl1,
    UInt8      *pS2,
    UInt32	sBpl2,
    UInt32	w,
    UInt32	h,
    UInt32	depth,
    UInt8      *pD,
    UInt32	dBpl)
{

	/* point to the images inside the frames */
    PROLOGUE;
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pS1 += sBpl1 << cLogFrameBits;	/* do left frame adjustment below */
    pS2 += sBpl2 << cLogFrameBits;
    pD += dBpl << cLogFrameBits;

	/* perform the correct operation given the image depth */
    switch (depth)
    {
    case 4:
	pS1 += 4 * cFrameBytes;
	pS2 += 4 * cFrameBytes;
	pD += 4 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
	    ip_subtractLine4 (pIPState, pS1, pS2, w, pD); /* compute difference
							   * in one line of
							   * 4 bit pixels */
	}
	break;

    case 8:
	pS1 += 8 * cFrameBytes;
	pS2 += 8 * cFrameBytes;
	pD += 8 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
	    ip_subtractLine8 (pIPState, pS1, pS2, w, pD); /* compute difference
							   * in one line of
							   * 8 bit pixels */
	}
	break;

    default:
	return ia_depthNotSupported;
    }

    NORMAL_RETURN;
}
/* PUBLIC */
/***********************************************************************
 *  i_subtractGrayNoClip ()
 *	Subtract two grayscale images. Does no clipping. If a pixel
 *	in pS2 is greater than the corresponding pixel in pS1,
 *	the result is some nasty rollover, including possible borrowing
 *	from neighbor pixels! This is probably only useful in cases 
 *	where the caller knows that pixels in pS1 are always greater
 *	than the ones in pS2, such as when doing a top-hat transform.
 *
 *	Input: 
 *	pS1 =	 ptr to source image 1 (includes frame), image from which to
 *               subtract
 *	sBpl1 =	 source image 1 bytes/line
 *	pS2 =	 ptr to source image 2 (includes frame), image which is 
 *		 subtracted from pS1
 *	sBpl2 =	 source gray image 2 bytes/line
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	depth =	 bits/pixel of all gray images
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *	dBpl =	 dest gray image bytes/line
 *
 *	On return: 
 *		destination gray image that equals the
 *		difference between source gray images, pS1 - pS2
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	image depth was not 4 or 8
 ***********************************************************************/
Int32   CDECL 
i_subtractGrayNoClip (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt32	sBpl1,
    UInt8      *pS2,
    UInt32	sBpl2,
    UInt32	w,
    UInt32	h,
    UInt32	depth,
    UInt8      *pD,
    UInt32	dBpl)
{

	/* point to the images inside the frames */
    PROLOGUE;
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pS1 += sBpl1 << cLogFrameBits;	/* do left frame adjustment below */
    pS2 += sBpl2 << cLogFrameBits;
    pD += dBpl << cLogFrameBits;

	/* perform the correct operation given the image depth */
    switch (depth)
    {
    case 4:
	pS1 += 4 * cFrameBytes;
	pS2 += 4 * cFrameBytes;
	pD += 4 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
		/* compute difference in one line of 4 bit pixels */
	    ip_subtractLine4NC (pIPState, pS1, pS2, w, pD); 
	}
	break;

    case 8:
	pS1 += 8 * cFrameBytes;
	pS2 += 8 * cFrameBytes;
	pD += 8 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
		/* compute difference in one line of 8 bit pixels */
	    ip_subtractLine8NC (pIPState, pS1, pS2, w, pD);    
	}
	break;

    default:
	return ia_depthNotSupported;
    }

    NORMAL_RETURN;
}

/*
 * Internal routines. Try to get it a bit faster by avoiding doing a byte
 * at a time, and by clipping without using branches.
 * We can do the same thing, with more economy of scale, for 4-bit.
 * 
 * What we do is as follows. Assume we're computing p1-p2 or 0 if p2>p1.
 * First, set the high bit on each pixel in w1, clear the high bit on
 * each pixel in w2, and subtract. What we have left in the high bits
 * is a "didn't borrow" bit, i.e. the low bits of w1 >= the low bits of
 * w2. Then, make a mask bit for each byte. There are two cases where
 * the mask bit for a pixel should be set:
 *
 *  1. The high bits are equal (either 0 or 1) and we didn't borrow
 * -or-
 *  2. The high bit of p1 is 1 and the high bit of p2 is 0
 * 
 */

/* PRIVATE */
/***********************************************************************
 *  ip_subtractLine8 ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 8 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals clipped 
 *		difference between source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_subtractLine8 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	wMask, sWord1, sWord2, hWord1, hWord2, notBorrowed;
    UInt8      *pDEnd;

	/* keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 3)/4) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);
	hWord1 = sWord1 & cHBits8;
	hWord2 = sWord2 & cHBits8;

	/* notBorrowed has 1 in each high bit if the low bits did
	   not cause a borrow, i.e. lowbits1 >= lowbits2 for each pixel */
	notBorrowed = ((sWord1 | cHBits8) - (sWord2 & cLBits8)) & cHBits8;

	/* Create high mask bits. A 1 means that we do the subtract, a zero
	   means we clip */
	wMask = (~(hWord1 ^ hWord2) & notBorrowed) |
		(hWord1 & ~hWord2);

	/* Replicate the high bit, to make a mask */
	wMask |= wMask >> 1;
	wMask |= wMask >> 2;
	wMask |= wMask >> 4;

	/* Mask each one so that we don't borrow from the adjacent pixel */
	ULONG_ACCESS(pD) = (sWord1 & wMask) - (sWord2 & wMask);
    }
}

/* PRIVATE */
/***********************************************************************
 *  ip_subtractLine4 ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 4 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals clipped 
 *		difference between source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_subtractLine4 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    register UInt32	 wMask, sWord1, sWord2, hWord1, hWord2, notBorrowed;
	     UInt8	*pDEnd;	

	/* keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 7)/8) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);
	hWord1 = sWord1 & cHBits4;
	hWord2 = sWord2 & cHBits4;

	/* notBorrowed has 1 in each high bit if the low bits did
	   not cause a borrow, i.e. lowbits1 >= lowbits2 for each pixel */
	notBorrowed = ((sWord1 | cHBits4) - (sWord2 & cLBits4)) & cHBits4;

	/* Create high mask bits. A 1 means that we do the subtract, a zero
	   means we clip */
	wMask = (~(hWord1 ^ hWord2) & notBorrowed) |
		(hWord1 & ~hWord2);

	/* Replicate the high bit, to make a mask */
	wMask |= wMask >> 1;
	wMask |= wMask >> 2;

	/* Mask each one so that we don't borrow from the adjacent pixel */
	ULONG_ACCESS(pD) = (sWord1 & wMask) - (sWord2 & wMask);
    }
}

/*
 * The no-clip versions are easy; just subtract and hope for the best. 
 */

/* PRIVATE */
/***********************************************************************
 *  ip_subtractLine8NC ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 8 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals NON-CLIPPED
 *		difference between source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_subtractLine8NC (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	 sWord1, sWord2;
    UInt8	*pDEnd;		         /* pts to last dest word + 1 */

	/* keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 3)/4) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);
	ULONG_ACCESS (pD) = sWord1 - sWord2;
    }
}

/* PRIVATE */
/***********************************************************************
 *  ip_subtractLine4NC ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 4 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals NON-CLIPPED 
 *		difference between source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_subtractLine4NC (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	 sWord1, sWord2;
    UInt8	*pDEnd;		         /* pts to last dest word + 1 */

	/* keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 7)/8) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);
	ULONG_ACCESS (pD) = sWord1 - sWord2;
    }
}

/* 
 * $Log: i_grydif.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 * Imported sources
 *
 * Revision 1.3  1998/01/13 23:48:07  ddavies
 * i_grydif.c -> /product/ipcore/ipshared/src/RCS/i_grydif.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.2  1995/12/13 02:44:15  ejaquith
 * i_grydif.c -> /product/ipcore/ipshared/src/RCS/i_grydif.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.1  1994/07/29  01:33:08  macleod
 * Initial revision
 *
 */
