/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  i_gryprj.c
 *
 *
 *
 *
 *
 */

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

#include "gryprj.prv"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: i_gryprj.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Eight bit gray horizontal projection of intensity.
 *
 * LongDescription:::
 *	Projects (sums) rows of 8 bit intensity values into an
 *	array with "height" entries (one per row).
 *
 * Input:::
 *	iS::		source image pointer (including frames)
 *	w::		framed width of image
 *	h::		framed height of image
 *	bpl::		bytes per line of image (including frame)
 *	invert::	TRUE = invert intensity while projecting,
 *			FALSE = project intensity normally
 *
 * Output:::
 *	projection::	pointer to resulting projection array
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_projectHGrayPixels8 (
    IPState    *pIPState,
    UInt8      *iS,
    Int32	w,
    Int32	h,
    Int32	bpl,
    Bool	invert,
    Int32      *projection)
{
    PROLOGUE;
    Int32    row;
    UInt32   thisWord;
    UInt8   *last, *this;
    Int32   *p;


    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    iS += (bpl << cLogFrameBits) + (8 * cFrameBytes);

    p = projection;

    if (invert)
    {
        /* NOTE: this will now add 255 for frame pixels in partial end words */
	for (row = 0; row < h; row++)
	{
	    this = iS;
	    last = this + w;
    
	    while (this < last)
	    {
		thisWord = ~(ULONG_ACCESS (this));
		*p += (thisWord & 0x000000ff) + 
		     ((thisWord >> 8) & 0x000000ff) +
		     ((thisWord >> 16) & 0x000000ff) + 
		     ((thisWord >> 24));
		this += 4;
	    }

            /* subtract out inverted frame pixels which may have been added */
            if ((w % 4) > 0)
	        *p -= 255 * (4 - (w % 4));

	    iS += bpl;
	    p++;
	}
    }
    else
    {
	for (row = 0; row < h; row++)
	{
	    this = iS;
	    last = this + w;
    
	    while (this < last)
	    {
		thisWord = ULONG_ACCESS (this);
		*p += (thisWord & 0x000000ff) + 
		     ((thisWord >> 8) & 0x000000ff) +
		     ((thisWord >> 16) & 0x000000ff) + 
		     ((thisWord >> 24));
		this += 4;
	    }
	    iS += bpl;
	    p++;
	}
    }

    NORMAL_RETURN;

}  /* i_projectHGrayPixels8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Eight bit gray vertical projection of intensity.
 *
 * LongDescription:::
 *	Projects (sums) columns of 8 bit intensity values into an
 *	array with "width" entries (one per column).
 *
 * Input:::
 *	iS::		source image pointer (including frames)
 *	w::		framed width of image
 *	h::		framed height of image
 *	bpl::		bytes per line of image (including frame)
 *	invert::	TRUE = invert intensity while projecting,
 *			FALSE = project intensity normally
 *
 * Output:::
 *	projection::	pointer to resulting projection array
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_projectVGrayPixels8 (
    IPState    *pIPState,
    UInt8      *iS,
    Int32	w,
    Int32	h,
    Int32	bpl,
    Bool	invert,
    Int32      *projection)
{
    PROLOGUE;
    Int32    row, wW, sW, b;
    UInt32   thisWord1, thisWord2, thisWord3, thisWord4;
    UInt8   *last, *this;
    Int32   *p, hOver4;


    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    iS += (bpl << cLogFrameBits) + (8 * cFrameBytes);


    wW = ((w / 4) * 4);
    sW = w % 4;

    hOver4 = h / 4;

    if (invert)
    {
	for (row = 0; row < hOver4; row++)
	{
	    this = iS;
	    last = this + wW;
	    p = projection;
    
	    while (this < last)
	    {
		thisWord1 = ~(ULONG_ACCESS (this));
		thisWord2 = ~(ULONG_ACCESS (this + bpl));
		thisWord3 = ~(ULONG_ACCESS (this + bpl + bpl));
		thisWord4 = ~(ULONG_ACCESS (this + bpl + bpl+ bpl));
		*p += (thisWord1 >> 24);
		*p += (thisWord2 >> 24);
		*p += (thisWord3 >> 24);
		*p += (thisWord4 >> 24);
		p++;
		*p += ((thisWord1 & 0x00ff0000) +
		       (thisWord2 & 0x00ff0000) +
		       (thisWord3 & 0x00ff0000) +
		       (thisWord4 & 0x00ff0000)) >> 16;
		p++;
		*p += ((thisWord1 & 0x0000ff00) +
		       (thisWord2 & 0x0000ff00) +
		       (thisWord3 & 0x0000ff00) +
		       (thisWord4 & 0x0000ff00)) >> 8;
		p++;
		*p += (thisWord1 & 0x000000ff) +
		      (thisWord2 & 0x000000ff) +
		      (thisWord3 & 0x000000ff) +
		      (thisWord4 & 0x000000ff);
		p++; 
    
		this += 4;
	    }
    
	    thisWord1 = ~(ULONG_ACCESS (this));
	    thisWord2 = ~(ULONG_ACCESS (this + bpl));
	    thisWord3 = ~(ULONG_ACCESS (this + bpl + bpl));
	    thisWord4 = ~(ULONG_ACCESS (this + bpl + bpl+ bpl));
	    for (b = sW; b > 0; b--)
	    {
		*p += (thisWord1 >> 24) +
		      (thisWord2 >> 24) +
		      (thisWord3 >> 24) +
		      (thisWord4 >> 24);
		thisWord1 = thisWord1 << 8;
		thisWord2 = thisWord2 << 8;
		thisWord2 = thisWord3 << 8;
		thisWord2 = thisWord4 << 8;
		p++;
	    }
	    iS += bpl << 2;
	}

        /* compute remaining rows one at a time */
	hOver4 = h % 4;
	for (row = 0; row < hOver4; row++)
	{
	    this = iS;
	    last = this + wW;
	    p = projection;
    
	    while (this < last)
	    {
		thisWord1 = ~(ULONG_ACCESS (this));
		*p += (thisWord1 >> 24);
		p++;
		*p += (thisWord1 & 0x00ff0000) >> 16;
		p++;
		*p += (thisWord1 & 0x0000ff00) >> 8;
		p++;
		*p += (thisWord1 & 0x000000ff);
		p++; 
    
		this += 4;
	    }
    
	    thisWord1 = ~(ULONG_ACCESS (this));
	    for (b = sW; b > 0; b--)
	    {
		*p += thisWord1 >> 24;
		thisWord1 = thisWord1 << 8;
		p++;
	    }
	    iS += bpl;
	}

    }
    else
    {
	for (row = 0; row < hOver4; row++)
	{
	    this = iS;
	    last = this + wW;
	    p = projection;
    
	    while (this < last)
	    {
		thisWord1 = ULONG_ACCESS (this);
		thisWord2 = ULONG_ACCESS (this + bpl);
		thisWord3 = ULONG_ACCESS (this + bpl + bpl);
		thisWord4 = ULONG_ACCESS (this + bpl + bpl + bpl);
		*p += (thisWord1 >> 24);
		*p += (thisWord2 >> 24);
		*p += (thisWord3 >> 24);
		*p += (thisWord4 >> 24);
		p++;
		*p += ((thisWord1 & 0x00ff0000) +
		       (thisWord2 & 0x00ff0000) +
		       (thisWord3 & 0x00ff0000) +
		       (thisWord4 & 0x00ff0000)) >> 16;
		p++;
		*p += ((thisWord1 & 0x0000ff00) +
		       (thisWord2 & 0x0000ff00) +
		       (thisWord3 & 0x0000ff00) +
		       (thisWord4 & 0x0000ff00)) >> 8;
		p++;
		*p += (thisWord1 & 0x000000ff) +
		      (thisWord2 & 0x000000ff) +
		      (thisWord3 & 0x000000ff) +
		      (thisWord4 & 0x000000ff);
		p++; 

		this += 4;
	    }
    
	    thisWord1 = ULONG_ACCESS (this);
	    thisWord2 = ULONG_ACCESS (this + bpl);
	    thisWord3 = ULONG_ACCESS (this + bpl + bpl);
	    thisWord4 = ULONG_ACCESS (this + bpl + bpl + bpl);
	    for (b = sW; b > 0; b--)
	    {
		*p += (thisWord1 >> 24) +
		      (thisWord2 >> 24) +
		      (thisWord3 >> 24) +
		      (thisWord4 >> 24);
		thisWord1 = thisWord1 << 8;
		thisWord2 = thisWord2 << 8;
		thisWord2 = thisWord3 << 8;
		thisWord2 = thisWord4 << 8;
		p++;
	    }
	    iS += bpl << 2;
	}

        /* compute remaining rows one at a time */
	hOver4 = h % 4;
	for (row = 0; row < hOver4; row++)
	{
	    this = iS;
	    last = this + wW;
	    p = projection;
    
	    while (this < last)
	    {
		thisWord1 = ULONG_ACCESS (this);
		*p += (thisWord1 >> 24);
		p++;
		*p += (thisWord1 & 0x00ff0000) >> 16;
		p++;
		*p += (thisWord1 & 0x0000ff00) >> 8;
		p++;
		*p += (thisWord1 & 0x000000ff);
		p++; 

		this += 4;
	    }
    
	    thisWord1 = ULONG_ACCESS (this);
	    for (b = sW; b > 0; b--)
	    {
		*p += thisWord1 >> 24;
		thisWord1 = thisWord1 << 8;
		p++;
	    }
	    iS += bpl;
	}
    }

    NORMAL_RETURN;

}  /* i_projectVGrayPixels8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Redistributes two orthogonal projections into an image using MIN.
 *
 * LongDescription:::
 *	The value of the result image pixel at location (x, y) is: 
 *	MIN (vProjection[x], hProjection[y]).  Note that the projection 
 *	arrays should be normalized before this operation.
 *
 * Input:::
 *	invert::	If cTrue, invert the intensity when projecting
 *	normalize::	If cTrue, normalize before backprojecting
 *	hProjection::	The horizontal projection array
 *	vProjection::	The vertical projection array
 *	w::		Framed width of destination image
 *	h::		Framed height of destination image
 *	bpl::		Bytes per line of destination image
 *
 * Output:::
 *	iD::		Pointer to destination image data including frames
 *
 * Returns:::
 *
 * Notes:::
 *	iD must be 8bpp.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_backprojectGrayPixelsMin8 (
    IPState    *pIPState,
    Int32	invert,
    Int32	normalize,
    Int32      *hProjection,
    Int32      *vProjection,
    Int32	w,
    Int32	h,
    Int32	bpl,
    UInt8      *iD)
{
    PROLOGUE;
    Int32   r,  rowV;
    UInt8  *vProjRow;

    ERR_GEN( ((hProjection == NULL) || (vProjection == NULL)), ia_invalidParm);

    ERR_GEN( (iD == NULL), ia_invalidParm);

    iD += (cFrameBits * bpl) + cFrameBits;
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

    if (normalize || invert)
    {
	ERR_PASS( i_normalizeProjection (pIPState, hProjection, h, 8, invert, 
					 (normalize ? w : 1)));

	ERR_PASS( i_normalizeProjection (pIPState, vProjection, w, 8, invert, 
					 (normalize ? h : 1)));
    }

    vProjRow = (UInt8 *)CALLOC (((w + 3) / 4) * 4, sizeof (UInt8));
    if (vProjRow == NULL)
        return (ia_nomem);

    /* convert our projection histogram into a fake image row */
    ERR_PASS_CLEANUP( ip_projectionToImageRow8 (pIPState, vProjection,
						w, vProjRow));

    for (r = 0; r < h; r++)
    {
        rowV = hProjection[r];

	ERR_PASS_CLEANUP(
		ip_backprojectGrayPixelsMin8Row (pIPState, rowV, vProjRow,
						 w, iD));
	iD += bpl;
    }

    FREE (vProjRow);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (vProjRow != NULL)
        FREE (vProjRow);
    ERR_RETURN;

}  /* i_backprojectGrayPixelsMin8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Redistributes two orthogonal projections into an image using MAX.
 *
 * LongDescription:::
 *	The value of the result image pixel at location (x, y) is: 
 *	MAX (vProjection[x], hProjection[y]).  Note that the projection 
 *	arrays should be normalized before this operation.
 *
 * Input:::
 *	invert::	If cTrue, invert the intensity when projecting
 *	normalize::	If cTrue, normalize before backprojecting
 *	hProjection::	The horizontal projection array
 *	vProjection::	The vertical projection array
 *	w::		Framed width of destination image
 *	h::		Framed height of destination image
 *	bpl::		Bytes per line of destination image
 *
 * Output:::
 *	iD::		Pointer to destination image data including frames
 *
 * Returns:::
 *	ia_successful	if completes successfully
 *	ia_nomem	if unable to allocate intermediate row buffer
 *	ia_invalidParm	if one or more parameters are bogus
 *
 * Notes:::
 *	iD must be 8bpp.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_backprojectGrayPixelsMax8 (
    IPState    *pIPState,
    Int32	invert,
    Int32	normalize,
    Int32      *hProjection,
    Int32      *vProjection,
    Int32	w,
    Int32	h,
    Int32	bpl,
    UInt8      *iD)
{
    PROLOGUE;
    Int32   r,  rowV;
    UInt8  *vProjRow;

    ERR_GEN( ((hProjection == NULL) || (vProjection == NULL)), ia_invalidParm);

    ERR_GEN( (iD == NULL), ia_invalidParm);

    iD += (cFrameBits * bpl) + cFrameBits;
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

    if (normalize || invert)
    {
	ERR_PASS( i_normalizeProjection (pIPState, hProjection, h, 8, invert, 
					 (normalize ? w : 1)));

	ERR_PASS( i_normalizeProjection (pIPState, vProjection, w, 8, invert, 
					 (normalize ? h : 1)));
    }

    vProjRow = (UInt8 *)CALLOC (((w + 3) / 4) * 4, sizeof (UInt8));
    ERR_GEN((vProjRow == NULL),ia_nomem);

    /* convert our projection histogram into a fake image row */
    ERR_PASS_CLEANUP( ip_projectionToImageRow8 (pIPState, vProjection,
						w, vProjRow));

    for (r = 0; r < h; r++)
    {
        rowV = hProjection[r];

	ERR_PASS_CLEANUP( ip_backprojectGrayPixelsMax8Row (pIPState, rowV,
							   vProjRow, w, iD));
	iD += bpl;
    }

    FREE (vProjRow);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (vProjRow != NULL)
        FREE (vProjRow);

    ERR_RETURN;

}  /* i_backprojectGrayPixelsMax8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Redistributes two orthogonal projections into an image using DIFF.
 *
 * LongDescription:::
 *	The value of the result image pixel at location (x, y) is: 
 *	DIFF (vProjection[x], hProjection[y]) where DIFF is absolute value
 *	of difference.  Note that the projection arrays should be 
 *	normalized before this operation.
 *
 * Input:::
 *	invert::	If cTrue, invert the intensity when projecting
 *	normalize::	If cTrue, normalize before backprojecting
 *	hProjection::	The horizontal projection array
 *	vProjection::	The vertical projection array
 *	w::		Framed width of destination image
 *	h::		Framed height of destination image
 *	bpl::		Bytes per line of destination image
 *
 * Output:::
 *	iD::		Pointer to destination image data including frames
 *
 * Returns:::
 *	ia_successful	if completes successfully
 *	ia_nomem	if unable to allocate intermediate row buffer
 *	ia_invalidParm	if one or more parameters are bogus
 *
 * Notes:::
 *	iD must be 8bpp.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_backprojectGrayPixelsDiff8 (
    IPState    *pIPState,
    Int32	invert,
    Int32	normalize,
    Int32      *hProjection,
    Int32      *vProjection,
    Int32	w,
    Int32	h,
    Int32	bpl,
    UInt8      *iD)
{
    PROLOGUE;
    Int32   r,  rowV;
    UInt8  *vProjRow;

    ERR_GEN( ((hProjection == NULL) || (vProjection == NULL)), ia_invalidParm);

    ERR_GEN( (iD == NULL), ia_invalidParm);

    iD += (cFrameBits * bpl) + cFrameBits;
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

    if (normalize || invert)
    {
	ERR_PASS( i_normalizeProjection (pIPState, hProjection, h, 8, invert, 
					 (normalize ? w : 1)));

	ERR_PASS( i_normalizeProjection (pIPState, vProjection, w, 8, invert, 
					 (normalize ? h : 1)));
    }

    vProjRow = (UInt8 *)CALLOC (((w + 3) / 4) * 4, sizeof (UInt8));
    ERR_GEN((vProjRow == NULL), ia_nomem);

    /* convert our projection histogram into a fake image row */
    ERR_PASS_CLEANUP( ip_projectionToImageRow8 (pIPState, vProjection,
						w, vProjRow));

    for (r = 0; r < h; r++)
    {
        rowV = hProjection[r];

	ERR_PASS_CLEANUP( ip_backprojectGrayPixelsDiff8Row (pIPState, rowV,
							    vProjRow, w, iD));
	iD += bpl;
    }

    FREE (vProjRow);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (vProjRow != NULL)
        FREE (vProjRow);

    ERR_RETURN;

}  /* i_backprojectGrayPixelsDiff8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Redistributes two orthogonal projections into an image using AVG.
 *
 * LongDescription:::
 *	The value of the result image pixel at location (x, y) is: 
 *	AVG (vProjection[x], hProjection[y]) where AVG is the mean (to
 *	within 1).  Note that the projection arrays should be 
 *	normalized before this operation.
 *
 * Input:::
 *	invert::	If cTrue, invert the intensity when projecting
 *	normalize::	If cTrue, normalize before backprojecting
 *	hProjection::	The horizontal projection array
 *	vProjection::	The vertical projection array
 *	w::		Framed width of destination image
 *	h::		Framed height of destination image
 *	bpl::		Bytes per line of destination image
 *
 * Output:::
 *	iD::		Pointer to destination image data including frames
 *
 * Returns:::
 *	ia_successful	if completes successfully
 *	ia_nomem	if unable to allocate intermediate row buffer
 *	ia_invalidParm	if one or more parameters are bogus
 *
 * Notes:::
 *	iD must be 8bpp.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_backprojectGrayPixelsAvg8 (
    IPState    *pIPState,
    Int32	invert,
    Int32	normalize,
    Int32      *hProjection,
    Int32      *vProjection,
    Int32	w,
    Int32	h,
    Int32	bpl,
    UInt8      *iD)
{
    PROLOGUE;
    Int32   r,  rowV;
    UInt8  *vProjRow;

    ERR_GEN( ((hProjection == NULL) || (vProjection == NULL)), ia_invalidParm);

    ERR_GEN( (iD == NULL), ia_invalidParm);

    iD += (cFrameBits * bpl) + cFrameBits;
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

    if (normalize || invert)
    {
	ERR_PASS( i_normalizeProjection (pIPState, hProjection, h, 8, invert, 
					 (normalize ? w : 1)));

	ERR_PASS( i_normalizeProjection (pIPState, vProjection, w, 8, invert, 
					 (normalize ? h : 1)));
    }

    vProjRow = (UInt8 *)CALLOC (((w + 3) / 4) * 4, sizeof (UInt8));
    ERR_GEN((vProjRow == NULL), ia_nomem);

    /* convert our projection histogram into a fake image row */
    ERR_PASS_CLEANUP( ip_projectionToImageRow8 (pIPState, vProjection,
						w, vProjRow));

    for (r = 0; r < h; r++)
    {
        rowV = hProjection[r];

	ERR_PASS_CLEANUP( ip_backprojectGrayPixelsAvg8Row (pIPState, rowV,
							   vProjRow, w, iD));
	iD += bpl;
    }

    FREE (vProjRow);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (vProjRow != NULL)
        FREE (vProjRow);
    ERR_RETURN;

}  /* i_backprojectGrayPixelsAvg8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Creates a single image from projections using MIN.
 *
 * LongDescription:::
 *	Pixel x in the row will be assigned the value MIN (rowV, vProj[x])
 *
 * Input:::
 *	rowV::		Horizontal projection value for this row
 *	vProj::		Image row buffer containing vertical projection
 *	w::		Width of destination image in pixels
 *
 * Output:::
 *	iD::		Pointer to row of destination image (no frames)
 *
 * Returns:::
 *  ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_backprojectGrayPixelsMin8Row (
    IPState    *pIPState,
    Int32	rowV,
    UInt8      *vProj,
    Int32	w,
    UInt8      *iD)
{
    PROLOGUE;
    UInt32 w1, w2, x, d;
    UInt8 *last;

    last = iD + (((w + 3) / 4) * 4);

    w1 = (UInt32)rowV;
    w1 |= w1 << 16;
    w1 |= w1 << 8;

    while (iD < last)
    {
        /* this computes a min for 4 pixels at a time.  See i_absdif.c 
	 * for comments. 
	 */
        w2 = ULONG_ACCESS (vProj);
	x = w1 ^ w2;
	d = (w1 | 0x80808080) - (w2 & 0x7f7f7f7f);
	d = ((~x) & d) + (w1 & (~w2));
	d &= 0x80808080;
	d = 0x80808080 - (d >> 7);
	d ^= 0x80808080;
	d &= x;
	w2 = w1 ^ d;
	ULONG_ACCESS (iD) = w2;

	iD += 4;
	vProj += 4;
    }

    /* fix up the right frame if necessary */
    w = w % 4;
    if (w > 0)
    {
        w = 3 - w;
	iD -= 4;
	ULONG_ACCESS (iD) &= (0xffffff00 << w);
    }

    NORMAL_RETURN;

}  /* ip_backprojectGrayPixelsMin8Row () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Creates a single image from projections using MAX.
 *
 * LongDescription:::
 *	Pixel x in the row will be assigned the value MAX (rowV, vProj[x])
 *
 * Input:::
 *	rowV::		Horizontal projection value for this row
 *	vProj::		Image row buffer containing vertical projection
 *	w::		Width of destination image in pixels
 *
 * Output:::
 *	iD::		Pointer to row of destination image (no frames)
 *
 * Returns:::
 *  ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_backprojectGrayPixelsMax8Row (
    IPState    *pIPState,
    Int32	rowV,
    UInt8      *vProj,
    Int32	w,
    UInt8      *iD)
{
    PROLOGUE;
    UInt32 w1, w2, x, d;
    UInt8 *last;

    last = iD + (((w + 3) / 4) * 4);

    w1 = (UInt32)rowV;
    w1 |= w1 << 16;
    w1 |= w1 << 8;

    while (iD < last)
    {
        /* this computes a max for 4 pixels at a time.  See i_absdif.c 
	 * for comments. 
	 */
        w2 = ULONG_ACCESS (vProj);
	x = w1 ^ w2;
	d = (w1 | 0x80808080) - (w2 & 0x7f7f7f7f);
	d = ((~x) & d) + (w1 & (~w2));
	d &= 0x80808080;
	d = 0x80808080 - (d >> 7);
	d ^= 0x80808080;
	d &= x;
	w2 ^= d;
	ULONG_ACCESS (iD) = w2;

	iD += 4;
	vProj += 4;
    }

    /* fix up the right frame if necessary */
    w = w % 4;
    if (w > 0)
    {
        w = 3 - w;
	iD -= 4;
	ULONG_ACCESS (iD) &= (0xffffff00 << w);
    }

    NORMAL_RETURN;

}  /* ip_backprojectGrayPixelsMax8Row () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Creates a single image from projections using DIFF.
 *
 * LongDescription:::
 *	Pixel x in the row will be assigned the value DIFF (rowV, vProj[x])
 *	where DIFF is absolute value of difference.
 *
 * Input:::
 *	rowV::		Horizontal projection value for this row
 *	vProj::		Image row buffer containing vertical projection
 *	w::		Width of destination image in pixels
 *
 * Output:::
 *	iD::		Pointer to row of destination image (no frames)
 *
 * Returns:::
 *  ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_backprojectGrayPixelsDiff8Row (
    IPState    *pIPState,
    Int32	rowV,
    UInt8      *vProj,
    Int32	w,
    UInt8      *iD)
{
    PROLOGUE;
    UInt32 w1, w2, x, d;
    UInt8 *last;

    last = iD + (((w + 3) / 4) * 4);

    w1 = (UInt32)rowV;
    w1 |= w1 << 16;
    w1 |= w1 << 8;

    while (iD < last)
    {
        /* this computes a 4 pixel absolute value of difference.  See 
	 * i_absdif.c for comments. 
	 */
        w2 = ULONG_ACCESS (vProj);
	x = w1 ^ w2;
	d = (w1 | 0x80808080) - (w2 & 0x7f7f7f7f);
	d = ((~x) & d) + (w1 & (~w2));
	d &= 0x80808080;
	d = 0x80808080 - (d >> 7);
	d ^= 0x80808080;
	d &= x;
	ULONG_ACCESS (iD) = (w2 ^ d) - (w1 ^ d);

	iD += 4;
	vProj += 4;
    }

    /* fix up the right frame if necessary */
    w = w % 4;
    if (w > 0)
    {
        w = 3 - w;
	iD -= 4;
	ULONG_ACCESS (iD) &= (0xffffff00 << w);
    }

    NORMAL_RETURN;

}  /* ip_backprojectGrayPixelsDiff8Row () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Creates a single image from projections using AVG.
 *
 * LongDescription:::
 *	Pixel x in the row will be assigned the value AVG (rowV, vProj[x])
 *	where AVG is the mean (to within 1).
 *
 * Input:::
 *	rowV::		Horizontal projection value for this row
 *	vProj::		Image row buffer containing vertical projection
 *	w::		Width of destination image in pixels
 *
 * Output:::
 *	iD::		Pointer to row of destination image (no frames)
 *
 * Returns:::
 *  ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_backprojectGrayPixelsAvg8Row (
    IPState    *pIPState,
    Int32	rowV,
    UInt8      *vProj,
    Int32	w,
    UInt8      *iD)
{
    PROLOGUE;
    UInt32 w1, w2, x, d;
    UInt8 *last;

    last = iD + (((w + 3) / 4) * 4);

    w1 = (UInt32)rowV;
    w1 |= w1 << 16;
    w1 |= w1 << 8;

    while (iD < last)
    {
        /* this computes a 4 pixel mean.  See 
	 * i_absdif.c for comments. 
	 */
        w2 = ULONG_ACCESS (vProj);
	x = w1 ^ w2;
	d = (w1 | 0x80808080) - (w2 & 0x7f7f7f7f);
	d = ((~x) & d) + (w1 & (~w2));
	d &= 0x80808080;
	d = 0x80808080 - (d >> 7);
	d ^= 0x80808080;
	d &= x;
	w2 ^= d;
	w2 -= w1 ^ d;
	w2 = (w2 & 0xfefefefe) >> 1;
	ULONG_ACCESS (iD) = w2 + (w1 ^ d);

	iD += 4;
	vProj += 4;
    }

    /* fix up the right frame if necessary */
    w = w % 4;
    if (w > 0)
    {
        w = 3 - w;
	iD -= 4;
	ULONG_ACCESS (iD) &= (0xffffff00 << w);
    }

    NORMAL_RETURN;

}  /* ip_backprojectGrayPixelsAvg8Row () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Converts a projection array into a PIXR row with pad, but no frame.
 *
 * LongDescription:::
 *
 * Input:::
 *	vProjection::	The vertical projection array to convert
 *	w::		Size of (number of entries in) projection array
 *
 * Output:::
 *	vProjRow::	Pointer to image row (not allocated here)
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_projectionToImageRow8 (
    IPState    *pIPState,
    Int32      *vProjection,
    Int32	w,
    UInt8      *vProjRow)
{
    PROLOGUE;
    UInt32 dW;
    Int32 *last;

    last = vProjection + w - 1;

    while (cTrue)
    {
        dW = (*vProjection) << 24;
	vProjection++;
	if (vProjection > last)
	    break;

	dW |= ((*vProjection) << 16) & 0x00ff0000;
	vProjection++;
	if (vProjection > last)
	    break;

	dW |= ((*vProjection) << 8) & 0x0000ff00;
	vProjection++;
	if (vProjection > last)
	    break;

	dW |= (*vProjection) & 0x000000ff;
	vProjection++;
	if (vProjection > last)
	    break;

	ULONG_ACCESS (vProjRow) = dW;

	vProjRow += 4;
    }

    ULONG_ACCESS (vProjRow) = dW;

    NORMAL_RETURN;

}  /* ip_projectionToImageRow8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Normalize a projection histogram (divide each entry by a factor)
 *
 * LongDescription:::
 *
 * Input:::
 *	projection::	Pointer to projection array to be normalized
 *	length::	Length of the projection array
 *	depth::		Depth of the image projected into the array
 *	invert::	if cTRUE, invert the intensity sense while
 *			normalizing, else normalize normally
 *	normalizationFactor:: The number by which to divide each projection
 *			array value
 *
 * Output:::
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_normalizeProjection (
    IPState    *pIPState,
    Int32      *projection,
    Int32	length,
    Int32	depth,
    Int32	invert,
    Int32	normalizationFactor)
{
    PROLOGUE;
    Int32  i, max;
    Int32 *p;

    p = projection;

    if (depth == 4)
        max = 16;
    else
        max = 255;

    if (invert)
	for (i = 0; i < length; i++, p++)
	    *p = max - (*p / normalizationFactor);
    else
	for (i = 0; i < length; i++, p++)
	    *p /= normalizationFactor;

    NORMAL_RETURN;

}  /* i_normalizeProjection () */


/*
 * $Log: i_gryprj.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 * Imported sources
 *
 * Revision 1.2  1998/01/13 23:48:08  ddavies
 * i_gryprj.c -> /product/ipcore/ipshared/src/RCS/i_gryprj.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.1  1997/03/03 22:26:59  dferrell
 * Initial revision
 *
 */


