/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_grysum.c: Grayscale image addition
 *
 *                  Int32     	i_addGray()
 *		    Int32	i_addGrayNoClip ()
 *		    Int32	i_averageGray ()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /*  prototyes */
#include "shrip.pub"
#include "shrip.prv"

    /* Constants */
#define cHBits8		0x80808080UL
#define cLBits8		0x7f7f7f7fUL
#define cLSBits8	0x01010101UL
#define cHBits4		0x88888888UL
#define cLBits4		0x77777777UL
#define cLSBits4	0x11111111UL

IP_RCSINFO(RCSInfo, "$RCSfile: i_grysum.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* PUBLIC */
/***********************************************************************
 *  i_addGray ()
 *	Add two grayscale images. Clips at 255 or 15.
 *
 *	Input: 
 *	pS1 =	 ptr to source image 1 (includes frame)
 *	sBpl1 =	 source image 1 bytes/line
 *	pS2 =	 ptr to source image 2 (includes frame), image which is 
 *		 added to pS1
 *	sBpl2 =	 source gray image 2 bytes/line
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	d =	 depth of all images
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *	dBpl =	 dest gray image bytes/line
 *
 *	On return: 
 *		destination gray image that equals the clipped-at-MAX
 *		sum of source gray images, pS1 + pS2
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	image depth was not 4 or 8
 ***********************************************************************/
Int32   CDECL 
i_addGray (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt32	sBpl1,
    UInt8      *pS2,
    UInt32	sBpl2,
    UInt32	w,
    UInt32	h,
    UInt32	depth,
    UInt8      *pD,
    UInt32	dBpl)
{
	/* point to the images inside the frames */
    PROLOGUE;
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pS1 += sBpl1 << cLogFrameBits;	/* do left frame adjustment below */
    pS2 += sBpl2 << cLogFrameBits;
    pD += dBpl << cLogFrameBits;

	/* perform the correct operation given the image depth */
    switch (depth)
    {
    case 4:
	pS1 += 4 * cFrameBytes;
	pS2 += 4 * cFrameBytes;
	pD += 4 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
	    ip_addLine4 (pIPState, pS1, pS2, w, pD);    /* compute sum
							 * in one line of
							 * 4 bit pixels */
	}
	break;

    case 8:
	pS1 += 8 * cFrameBytes;
	pS2 += 8 * cFrameBytes;
	pD += 8 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
	    ip_addLine8 (pIPState, pS1, pS2, w, pD);    /* compute sum
							 * in one line of
							 * 8 bit pixels */
	}
	break;

    default:
	return ia_depthNotSupported;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/***********************************************************************
 *  i_addGrayNoClip ()
 *	Add two grayscale images. Does no clipping.  If the sum
 *	of pS1[x,y] + pS2[x,y] is greater than 255 or 15,
 *	the result is some nasty rollover, including possible adding
 *	to neighboring pixels! This is probably only useful in cases 
 *	where the caller knows that every pixel sum is less than
 *	or equal to the maximum value for the depth.
 *
 *	Input: 
 *	pS1 =	 ptr to source image 1 (includes frame)
 *	sBpl1 =	 source image 1 bytes/line
 *	pS2 =	 ptr to source image 2 (includes frame), image which is 
 *		 added to pS1
 *	sBpl2 =	 source gray image 2 bytes/line
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	depth =	 bits/pixel of all gray images
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *	dBpl =	 dest gray image bytes/line
 *
 *	On return: 
 *		destination gray image that equals the
 *		pixel-by-pixel sum of source gray images, pS1 + pS2
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	image depth was not 4 or 8
 ***********************************************************************/
Int32   CDECL 
i_addGrayNoClip (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt32	sBpl1,
    UInt8      *pS2,
    UInt32	sBpl2,
    UInt32	w,
    UInt32	h,
    UInt32	depth,
    UInt8      *pD,
    UInt32	dBpl)
{

	/* point to the images inside the frames */
    PROLOGUE;
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pS1 += sBpl1 << cLogFrameBits;	/* do left frame adjustment below */
    pS2 += sBpl2 << cLogFrameBits;
    pD += dBpl << cLogFrameBits;

	/* perform the correct operation given the image depth */
    switch (depth)
    {
    case 4:
	pS1 += 4 * cFrameBytes;
	pS2 += 4 * cFrameBytes;
	pD += 4 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
	    ip_addLine4NC (pIPState, pS1, pS2, w, pD);    /* compute sum
							   * in one line of
							   * 4 bit pixels */
	}
	break;

    case 8:
	pS1 += 8 * cFrameBytes;
	pS2 += 8 * cFrameBytes;
	pD += 8 * cFrameBytes;
	for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
	{
	    ip_addLine8NC (pIPState, pS1, pS2, w, pD);    /* compute sum
							   * in one line of
							   * 8 bit pixels */
	}
	break;

    default:
	return ia_depthNotSupported;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/***********************************************************************
 *  i_averageGray ()
 *	Average two grayscale images.  Works only on 8-bit images.
 *	The LS bit of each source pixel is ignored.
 *
 *	Input: 
 *	pS1 =	 ptr to source image 1 (includes frame)
 *	sBpl1 =	 source image 1 bytes/line
 *	pS2 =	 ptr to source image 2 (includes frame), image which is 
 *		 added to pS1
 *	sBpl2 =	 source gray image 2 bytes/line
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	depth =	 bits/pixel of all gray images, currently must be 8
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *	dBpl =	 dest gray image bytes/line
 *
 *	On return: 
 *		destination gray image that equals the
 *		pixel-by-pixel average of source gray images, (pS1 + pS2)/2
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	image depth was not 8
 ***********************************************************************/
Int32   CDECL 
i_averageGray (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt32	sBpl1,
    UInt8      *pS2,
    UInt32	sBpl2,
    UInt32	w,
    UInt32	h,
    UInt32	depth,
    UInt8      *pD,
    UInt32	dBpl)
{

	/* point to the images inside the frames */
    PROLOGUE;
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pS1 += sBpl1 << cLogFrameBits;	/* do left frame adjustment below */
    pS2 += sBpl2 << cLogFrameBits;
    pD += dBpl << cLogFrameBits;

	/* check the depth */
    if (depth != 8)
	return ia_depthNotSupported;

	/* perform the operation */
    pS1 += 8 * cFrameBytes;
    pS2 += 8 * cFrameBytes;
    pD += 8 * cFrameBytes;
    for (; h > 0; h--, pS1 += sBpl1, pS2 += sBpl2, pD += dBpl)
    {
	ip_averageLine8 (pIPState, pS1, pS2, w, pD);  /* compute average
						       * in one line of
						       * 4 bit pixels */
    }

    NORMAL_RETURN;
}


/*
 * Internal routines. Try to get it a bit faster by avoiding doing a byte
 * at a time, and by clipping without using branches.
 *
 * What we do is as follows. Assume we're computing p1+p2 or
 * MAXVALUE if p1+p1 > MAXVALUE.
 *
 * First, clear the high bits on each pixel in p1 and in p2,
 * and add.  What we have left in the high bits
 * is a "carry" bit, i.e. the low bits of w1 + the low bits of
 * w2 caused a carry to the high bits.  Then, make a conditional
 * mask for each byte.  The mask for each byte will contain either
 * all 0's or all 1's.  The mask will be ORed with the byte to produce
 * The proper sum of the bytes or 255 (0xFF).
 * 
 * There are two cases where the mask byte for a pixel should be set:
 *
 *  1. The high bits of both pixels are set;
 * -or-
 *  2. The high bit of either pixel is set and the carry bit is set.
 * 
 */

/* PRIVATE */
/***********************************************************************
 *  ip_addLine4 ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 4 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals clipped 
 *		difference between source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_addLine4 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	 wMask, sWord1, sWord2;
    UInt8	*pDEnd;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 7)/8) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);

	/* wMask has 1 in each high bit of a byte if there was a carry
	 * from adding the two constituent bytes.  It has trash in the
	 * lower 7 bits of each byte.
	 */
	wMask = (((sWord1 & cLBits4) + (sWord2 & cLBits4)) &
		 (sWord1 | sWord2)) | (sWord1 & sWord2);

	/* Create mask bits.  A 0xF in a byte means that we clip the
	 * result to 15, a 0x0 means we use the real sum.
	 */
	wMask = (cHBits4 - ((wMask >> (4-1)) & cLSBits4)) ^ cHBits4;

	/* One way to compute the sums without any carries is to use
	 * (((sWord1 & cLBits4) + (sWord2 & cLBits4)) |
	 *  ((sWord1 | sWord2) & cHBits4)) | wMask.
	 * That's 7 operations.  There's a
	 * dumber way to do it (((sWord1 & ~wMask) + sWord2) | wMask),
	 * which works because ~wMask turns any byte that will cause
	 * a carry to 0 (hence eliminates the carry), then replaces
	 * the sum with 0xF.
	 */
	ULONG_ACCESS(pD) = (((sWord1 & ~wMask) + sWord2) | wMask);
    }
}


/* PRIVATE */
/***********************************************************************
 *  ip_addLine8 ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 8 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals clipped 
 *		difference between source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_addLine8 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	 wMask, sWord1, sWord2;
    UInt8	*pDEnd;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 3)/4) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	sWord1 = ULONG_ACCESS(pS1);
	sWord2 = ULONG_ACCESS(pS2);

	/* wMask has 1 in each high bit of a byte if there was a carry
	 * from adding the two constituent bytes.  It has trash in the
	 * lower 7 bits of each byte.
	 */
	wMask = (((sWord1 & cLBits8) + (sWord2 & cLBits8)) &
		 (sWord1 | sWord2)) | (sWord1 & sWord2);

	/* Create mask bits.  A 0xFF in a byte means that we clip the
	 * result to 255, a 0x00 means we use the real sum.
	 */
	wMask = (cHBits8 - ((wMask >> (8-1)) & cLSBits8)) ^ cHBits8;

	/* One way to compute the sums without any carries is to use
	 * (((sWord1 & cLBits8) + (sWord2 & cLBits8)) |
	 *  ((sWord1 | sWord2) & 0x80808080)) | wMask.
	 * That's 7 operations.  There's a
	 * dumber way to do it (((sWord1 & ~wMask) + sWord2) | wMask),
	 * which works because ~wMask turns any byte that will cause
	 * a carry to 0 (hence eliminates the carry), then replaces
	 * the sum with 0xFF.
	 */
	ULONG_ACCESS(pD) = (((sWord1 & ~wMask) + sWord2) | wMask);
    }
}


/*
 * The no-clip versions are easy; just add and hope for the best. 
 */

/* PRIVATE */
/***********************************************************************
 *  ip_addLine4NC ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 4 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals NON-CLIPPED 
 *		sum of source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_addLine4NC (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt8	*pDEnd;		         /* pts to last dest word + 1 */

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 7)/8) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	ULONG_ACCESS (pD) = ULONG_ACCESS(pS1) + ULONG_ACCESS(pS2);
    }
}


/* PRIVATE */
/***********************************************************************
 *  ip_addLine8NC ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 8 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals NON-CLIPPED
 *		sum of source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_addLine8NC (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt8	*pDEnd;		         /* pts to last dest word + 1 */

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 3)/4) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	ULONG_ACCESS (pD) = ULONG_ACCESS(pS1) + ULONG_ACCESS(pS2);
    }
}


/* PRIVATE */
/***********************************************************************
 *  ip_averageLine8 ()
 *	Input: 
 *	pS1 =	 ptr to source image 1 (without frame)
 *	pS2 =	 ptr to source image 2 (without frame)
 *	w =	 width in pixels of all images (without frame)
 *	pD =	 ptr to dest image (can be same as one of the
 *		 source images)
 *
 *	All image depths must be 8 bits/pixel.
 *
 *	On return: 
 *		give line of dest gray image that equals
 *		average of source gray images.
 *
 ***********************************************************************/
void CDECL 
ip_averageLine8 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt32	w,
    UInt8      *pD)
{
    PROLOGUE;
    UInt8	*pDEnd;		         /* pts to last dest word + 1 */

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* initialize termination address.  We use an address instead of
	 * decrementing w to avoid a memory write */
    pDEnd = pD + (((w + 3)/4) * 4);

	/* loop through all the words in the line, transforming each one */
    for (; pD != pDEnd; pS1 += 4, pS2 += 4, pD += 4)
    {
	ULONG_ACCESS (pD) = ((ULONG_ACCESS(pS1) >> 1) & 0x7F7F7F7F) +
			    ((ULONG_ACCESS(pS2) >> 1) & 0x7F7F7F7F);
    }
}

/*

 $Log: i_grysum.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:48:10  ddavies
 i_grysum.c -> /product/ipcore/ipshared/src/RCS/i_grysum.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1997/07/31 00:49:51  ddavies
 Initial revision


 */
