/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_gsrfl1.c: Gray Scale Rotation, FLoating point interface #1
 *
 *	    Upper level rotation routines that do gray scale rotation,
 *		SMALL ANGLES ONLY! (< 10 degrees)
 *		These routines take the rotation angle in floating point
 *		radians.
 *	    Lower level routines implemented in C or assembler.
 *
 *          Gray Scale rotation routines:
 *                  Int32     	i_rotateGray()
 *		    Int32	i_rotateGrayIP()
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"

    /*  prototypes */
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_gsrfl1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");
/* This is just an interface to the fixed point version of gray scale
 * rotation.  Please see the EXTENSIVE comments in i_gsrfx1.c for a
 * description of the rotation algorithm.
 *
 * The two routines here take the rotation angle in units of radians
 * expressed as Float64's.  They convert the radians to ScaledAngle format
 * and call i_rotateGrayFixedPoint or
 * i_rotateGrayIPFixedPoint.  Those routines do the real work.
 */


/* PUBLIC */
/**************************************************************************
 *  i_rotateGray()
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		depth = depth of gray image.  4 and 8 bpp supported.
 *		rotAngle = rotation angle in RADIANS
 *		pD = destination gray image
 *		dBpl = destination bytes/line
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *
 *	On return:
 *		Destination image contains the source image rotated by
 *		the given angle.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		unable to allocate memory
 *		ia_nopoint		could not find next point
 *		ia_depthNotSupported	wrong number of bits per pixel
 *		ia_angleTooLarge     	abs value of rotation angle is
 *					greater than 10 degrees (0.17 radians)
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	Frame in destination image is assumed to already be cleared.
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateGray (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	wS,
    UInt32	hS,
    UInt32	sBpl,
    UInt32	depth,
    Float64	rotAngle,
    UInt8      *pD,
    UInt32	dBpl,
    Int16	xRes,
    Int16	yRes)
{
    PROLOGUE;
    ERR_PASS (i_rotateGrayFixedPoint (pIPState, pS, wS, hS, sBpl, depth,
				      RADIAN_TO_ScaledAngle(rotAngle),
				      pD, dBpl, xRes, yRes));
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_rotateGrayIP()
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		depth = depth of gray image.  4 and 8 bpp supported.
 *		rotAngle = rotation angle in RADIANS
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *
 *	On return:
 *		Source image contains the source image rotated by
 *		the given angle.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *		ia_nopoint	could not find next point
 *		ia_depthNotSupported wrong number of bits per pixel
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	Frame in destination image is assumed to already be cleared.
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateGrayIP (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	wS,
    UInt32	hS,
    UInt32	sBpl,
    UInt32	depth,
    Float64	rotAngle,
    Int16	xRes,
    Int16	yRes)
{
    PROLOGUE;
    ERR_PASS(i_rotateGrayIPFixedPoint (pIPState, pS, wS, hS, sBpl, depth,
				       RADIAN_TO_ScaledAngle(rotAngle),
				       xRes, yRes));
    NORMAL_RETURN;
}

/*

$Log: i_gsrfl1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.3  1998/01/13 23:48:12  ddavies
i_gsrfl1.c -> /product/ipcore/ipshared/src/RCS/i_gsrfl1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1995/12/13 02:44:18  ejaquith
i_gsrfl1.c -> /product/ipcore/ipshared/src/RCS/i_gsrfl1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1995/11/06  18:16:11  ddavies
 * Initial revision
 *
 * Revision 1.5  1995/02/03  19:45:31  ddavies
 * ddavies on Fri Feb 3 11:45:22 PST 1995
 *
 * i_gsrflt.c -> /product/ipcore/ipshared/src/RCS/i_gsrflt.c,v
 *
 * Change xRes and yRes from UInt16's to Int16's to keep acc happy.
 *
 * Revision 1.4  1995/02/03  01:16:04  ddavies
 * ddavies on Thu Feb 2 17:14:47 PST 1995
 *
 * i_gsrflt.c -> /product/ipcore/ipshared/src/RCS/i_gsrflt.c,v
 *
 * Add xRes and yRes args to small angle rotations.  These are used to apply
 * a horizontal shear when rotating anisotropic images.  The shear is needed
 * to account for the shear naturally present in such images.
 *
 * Revision 1.3  1993/12/02  19:23:04  ddavies
 * i_gsrflt.c -> /product/ipcore/ipshared/src/RCS/i_gsrflt.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:28:10  ddavies
 * i_gsrflt.c -> /product/ipcore/ipshared/src/RCS/i_gsrflt.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


