/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_gsrfl2.c: Gray Scale Rotation, FLoating point interface #2
 *
 *	    Upper level rotation routines that do gray scale rotation,
 *		in strips.  SMALL ANGLES ONLY! (< 10 degrees)
 *		These routines take the rotation angle in floating point
 *		radians.
 *	    Lower level routines implemented in C or assembler.
 *
 *          Gray Scale rotation routines:
 *                  Int32     	i_getInputLinesForRotatedOutputLines()
 *                  Int32     	i_rotateGrayInit()
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"

    /*  prototypes */
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_gsrfl2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* This is just an interface to the fixed point version of gray scale
 * rotation by strips.  Please see the EXTENSIVE comments in i_gsrfx2.c for a
 * description of the rotation algorithm.
 *
 * The routine encoded here takes the rotation angle in units of radians
 * expressed as Float64's.  It converts the radians to ScaledAngle format
 * and call i_rotateGrayInit.  Since the rotation angle is not an argument
 * for i_rotateGrayProcess and i_rotateGrayFinish, there are no special
 * versions of these routines that take the angle in floating point radians.
 */

/* PUBLIC */
/**************************************************************************
 *  i_getInputLinesForRotatedOutputLines()
 *	Input:
 *		width = width of gray image.
 *		height = total height of source image.
 *		destHeight = height of dest gray strip.
 *		rotAngle = rotation angle in radians.
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *		pSrcHeight = pointer to returned size in bytes.
 *
 *	On return:
 *		The size of the rotation block (in bytes) is stored
 *		into *pSize.
 *
 *	Return code:
 *		ia_successful		good completion
 *
 **************************************************************************/
Int32   CDECL 
i_getInputLinesForRotatedOutputLines (
    IPState    *pIPState,
    Int32	width,
    Int32	height,
    Int32	destHeight,
    Float64	rotAngle,
    Int32	xRes,
    Int32	yRes,
    Int32      *pSrcHeight)
{
    PROLOGUE;
    ScaledAngle	fxpRotAngle = RADIAN_TO_ScaledAngle(rotAngle);

    ERR_PASS( i_getInputLinesForRotatedOutputLinesFixedPoint (pIPState, width,
							      height,
							      destHeight,
							      fxpRotAngle,
							      xRes, yRes,
							      pSrcHeight));
    NORMAL_RETURN;
}

/* PUBLIC */
/**************************************************************************
 *  i_rotateGrayInit()
 *	Input:
 *		depth = depth of gray image.  4 and 8 bpp supported.
 *		rotAngle = rotation angle in RADIANS
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *		rotateInfoSize = size of the private structure
 *				 being passed in.  If this is not the
 *				 correct size, we return ia_invalidParm.
 *		pRotateInfo = pointer to existing rotation structure
 *			      that will be initialized in here.
 *		
 *
 *	On return:
 *		The contents of the rotation structure will be correctly
 *		initialized.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	wrong number of bits per pixel
 *		ia_angleTooLarge     	abs value of rotation angle is
 *					greater than 10 degrees (0.17 radians)
 *		ia_invalidParm		pRotateInfo is NULL or rotateInfoSize
 *					is incorrect.
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	Frame in destination image is assumed to already be cleared.
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateGrayInit (
    IPState    *pIPState,
    UInt32	depth,
    Float64	rotAngle,
    Int16	xRes,
    Int16	yRes,
    Int32	rotateInfoSize,
    void       *pRotateInfo)
{
    PROLOGUE;
    ERR_PASS( i_rotateGrayFixedPointInit (pIPState, depth,
				          RADIAN_TO_ScaledAngle(rotAngle),
				          xRes, yRes, rotateInfoSize,
				          pRotateInfo));
    NORMAL_RETURN;
}


/*

 $Log: i_gsrfl2.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 Imported sources

 Revision 1.4  1998/01/13 23:48:13  ddavies
 i_gsrfl2.c -> /product/ipcore/ipshared/src/RCS/i_gsrfl2.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1996/10/05 02:10:09  ddavies
 ddavies on Fri Oct 4 19:08:06 PDT 1996

 i_gsrfl2.c -> /product/ipcore/ipshared/src/RCS/i_gsrfl2.c,v

 Add height parameter to i_getInputLinesForRotatedOutputLines.  You really must
 know the total X excursion to calculate the number of input lines needed.
 To know this, you need to know how far over in X the first sampling path starts.
 This sin(rotAngle) * height/2, so you need height.

 Revision 1.2  1995/12/13 03:20:23  ddavies
 ddavies on Tue Dec 12 19:20:23 PST 1995

 i_gsrfl2.c -> /product/ipcore/ipshared/src/RCS/i_gsrfl2.c,v

 CommonCheck: Added semicolon after IP_RCSINFO macro.

 * Revision 1.1  1995/12/13  01:15:13  ddavies
 * Initial revision
 *

 */

