/*****************************************************************
 *  Copyright (c) 1998, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_gsrfl3.c: Gray Scale Rotation, FLoating point interface #3
 *
 *	    Full floating-point interpolated grayscale rotation.
 *		These routines take the rotation angle in floating point
 *		radians.
 *
 *          Gray Scale rotation routines:
 *                  Int32     	i_rotateInterpolated()
 *
 */


#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrscal.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "fxpmath.pub"
#include "shros.pub"
#include "shrscgr.prv"
#include "errgen.pub"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_gsrfl3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* This is the floating-point interpolating rotation routine.
 *
 * The routine coded here takes the rotation angle in units of radians
 * expressed as Float64's.
 *
 * Currently, the xRes and yRes are restricted to be equal -- no attempt to
 * account for non-square pixels is made here.
 */

/* PUBLIC */
/**************************************************************************
 *  i_rotateInterpolated()
 *	Input:
 *		pS = source gray image (not including frame)
 *		wS = width of source image in pixels
 *		hS = height of source image
 *		sBpl = source bytes/line
 *		xResS = starting resolution of image in x direction
 *		yResS = starting resolution of image in y direction
 *		depth = bits/pixel in source and dest images.
 *		fill = fill pixel value (when input image is missing)
 *		rotAngle = rotation angle in RADIANS
 *		pD = destination gray image (not including frame)
 *		wD = width of destination image in pixels
 *		hD = height of destination image
 *		dBpl = destination bytes/line
 *		pxResD = resulting resolution of image in x direction
 *		pyResD = resulting resolution of image in y direction
 *
 *	On return:
 *		Destination image contains the source image rotated by
 *		the given angle.  The rotation uses interpolation from
 *		1, 2, or 4 pixels, and uses floating-point to do the
 *		calculation. There is no restriction on the rotation angle,
 *		and the destination image can be any size. If no destination
 *		image is provided, an image with the same size is created.
 *		The rotation occurs about the center of the input image.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	could not rotate source image
 *					of specified depth
 *		ia_typeNotSupported	only square resolutions supported
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	Frame in destination image is assumed to already be cleared.
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *
 *	Authors:
 *		Luc Vincent (taken from rotate.c)
 *		Dave Abercrombie (modified for arbitrary images)
 *
 *	Notes:
 *		Code works but is still rather preliminary. Speed could be
 *		improved a lot. Float64 are probably not needed. Routine
 *		is now designed for unsigned char, but was originally designed
 *		to work with short int images, (there is no reason it could
 *		not be expanded for 12/16-bit images).
 *
 *		Note that this is not the right rotation technique for
 *		binary images, the results would be pretty bad, especially
 *		for one-pixel thick lines.
 *
 *		The adjustment due to non-square pixels is not accounted for
 *		in this routine. It is probably best to adjust for it outside
 *		this routine.
 *
 **************************************************************************/
Int32   CDECL 
i_rotateInterpolated (
    IPState    *pIPState,
    UInt8Ptr	pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Float32	xResS,
    Float32	yResS,
    Int32	depth,
    Int32	fill,
    Float64	rotAngle,
    UInt8Ptr	pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    Float32	*pxResD,
    Float32	*pyResD)
{
    PROLOGUE;
    Float64	cosinus, sinus;
    Float64	xac, yac;		/* Center of source image */
    Float64	xbc, ybc;		/* Center of destination image */
    Float64	xa, ya;			/* Coordinate in source image */
    Int32	xb, yb;			/* Coordinate in destination image */
    Float64	partxa, partya;
    Float64	value2;
    UInt8	*pixelsS, *pixelsD, *lineD;
    UInt8	*pixS, *pixD;

    /* Check for support */
    ERR_GEN(depth != 8, ia_depthNotSupported);
    ERR_GEN(xResS != yResS, ia_typeNotSupported);   /* If computed, remove this XXXXXXXXXX */

    /*----- Initializations -----*/
    *pxResD = xResS;
    *pyResD = yResS;
    

    /* point to the upper left corner of the source and dest image. */ 
    pixelsS = pS;
    pixelsD = pD;

    cosinus = cos(-rotAngle/* * M_PI / 180.0 */);
    sinus   = sin(-rotAngle/* * M_PI / 180.0 */);

    /* position of center of rotation in both spaces */
    xac = (Float64) wS / 2.0;
    yac = (Float64) hS / 2.0;
    xbc = (Float64) wD / 2.0;
    ybc = (Float64) hD / 2.0;


    /*----- Starts actual rotation -----*/

    /* Loop through each of the output pixels */

    for (yb = 0, lineD = pixelsD; yb < hD; yb++, lineD += dBpl) {

	for (xb = 0, pixD = lineD; xb < wD; xb++, pixD++) {

	    /* Compute the X/Y position that this output pixel has in the
	     * input image space */
	    xa = xac + cosinus * (((Float64) xb) - xbc) -
		sinus * (((Float64) yb) - ybc);
	    ya = yac + sinus * (((Float64) xb) - xbc) +
		cosinus * (((Float64) yb) - ybc);

	    /* If the corresponding input pixel is within the image
	     * dimensions, then there is a input pixel that contributes
	     * to this output pixel */
	    if (xa >= 0.0 && xa <= (Float64) (wS - 1) &&
		ya >= 0.0 && ya <= (Float64) (hS - 1)) {

		/* Compute the upper-left-most pixel of the 1, 2 or 4 pixels
		 * that will contribute to the output pixel. (Note that no
		 * round-up occurs here, since the actual pixel locations are
		 * computed below.) */
		pixS = pixelsS + (((Int32) ya) * sBpl + (Int32) xa);

		/* If the input X dimension is exactly on an input pixel
		 * boundary, then either one or two input pixels contribute */
		if (xa == (Float64) ((Int32) xa)) {

		    /* If the input Y dimension is exactly on an input pixel
		     * boundary, then only one input pixel contributes */
		    if (ya == (Float64) ((Int32) ya))
			value2 = UCHAR_ACCESS(pixS);

		    /* ... otherwise two input pixels contribute */
		    else {
			partya = ya - ((Float64) ((Int32) ya));
			value2 = (UCHAR_ACCESS(pixS)) * (1.0 - partya) +
			    (UCHAR_ACCESS(pixS + sBpl)) * partya;
		    }
		}

		/* ... otherwise the output pixel straddles 2 or 4 input
		 * pixels */
		else {
		    partxa = xa - ((Float64) ((Int32) xa));

		    /* If the input Y dimension is exactly on an input pixel
		     * boundary, then only two input pixels contribute */
		    if (ya == (Float64) ((Int32) ya))
			value2 = (UCHAR_ACCESS(pixS)) * (1.0 - partxa) +
			    (UCHAR_ACCESS(pixS + 1)) * partxa;

		    /* ... otherwise four input pixels contribute */
		    else {
			partya = ya - ((Float64) ((Int32) ya));
			value2 = (1.0 - partya) *
			    ((UCHAR_ACCESS(pixS)) * (1.0 - partxa)
			     + (UCHAR_ACCESS(pixS+1)) * partxa)
			    + partya * ((UCHAR_ACCESS(pixS+sBpl)) *
					(1.0 - partxa) +
					(UCHAR_ACCESS(pixS+sBpl+1)) * partxa);
		    }
		}

		/* Round the output */
		UCHAR_ACCESS(pixD) = (value2 - ((Float64) ((Int32) value2))
		    >= 0.5 ? (((UInt8) value2) + 1) : ((UInt8) value2));

	    } else {
		UCHAR_ACCESS(pixD) = (UInt8)fill;
	    }
	}
    }

    /*----- Clean-up and return -----*/

    NORMAL_RETURN;

} /* i_rotateInterpolated() */






/*
 $Log: i_gsrfl3.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 Imported sources

 Revision 1.1  1998/01/16 00:32:21  aber
 Initial revision

 */

