/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_gsrfx1.c: Gray Scale Rotation, FiXed point interface #1
 *
 *          Upper level gray scale rotation routines, SMALL ANGLES ONLY!
 *		(< 10 degrees)
 *	    These routine require no floating point support, they use
 *		fixed point calculations only.  They take angles in
 *		ScaledAngle format.
 *
 *          Gray Scale rotation routines:
 *                  Int32     	i_rotateGrayFixedPoint()
 *		    Int32	i_rotateGrayIPFixedPoint()
 * 
 */


#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrscal.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "fxpmath.pub"
#include "shros.pub"
#include "shrrast.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_gsrfx1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");
/* People don't always put documents on copiers straight.  The ability
 * make these skewed images look nice by rotating them a few degrees is
 * thought to be a valuable feature.
 *
 * Rotation is an ugly operation.  We have to move all the pixels in the
 * image by distances that appear straightforward to image procesing
 * grad students, but arbitary to microcoders.  WE DON'T LIKE TRANCENDENTAL
 * FUNCTIONS!!  Heck, we don't even like mutiplication.
 *
 * Image processing experts are to assembly programmers as Newtonian
 * mechanics is to quantum mechanics.  There are no continuous functions
 * down here.
 *
 * Naturally, we'd like to do all this pixel moving on binary images.  Dan
 * Bloomberg has some very nice rotation algorithms that shear binary
 * images.  Unfortunately, the results have icky little artifacts.  Since
 * the rotation op only has binary pixels to work with, it can do no better
 * than to simply move each pixel in the input image to some reasonable
 * approximation of the correct location in the output image.  Well, ok, it
 * could try to smooth the input image into a gray image and interpolate.
 * However, if the binary image contained any dithered pictures, we'd have
 * to dither them again to get binary.  Yuck.  We'll assume that binary
 * images are rotated by simply moving the existing pixels around.
 *
 * The input image contained slanted lines (that's why we're rotating it)
 * and curves; so had jaggies.  A binary rotation operation introduces more
 * jaggies by shearing sections of the image.  There's nothing in the
 * rotation operation that tends to force the rotation jaggies to cancel
 * the inherent jaggies.  In fact, both sets end up in the final image.
 * This is most noticable in long lines - they get 1 pixel dips and bumps
 * along their edges.  Tiny text gets noticeably bent.  Unfortunately, people
 * are pretty sensitive to these imperfections.  Mr. Z will also point out
 * that the jaggies on each side of a nearly vertical line aren't horizontally
 * aligned, so a series of horizontal samples through a slanted line see
 * the width of the line vary with position.  After being rotated, these
 * variations result in hourglass bumps along the line.
 *
 * One could try to remove all the dips and bumps.  Unfortunately, we don't
 * really know how to characterize them.  Images are filled with 1 pixel
 * jaggies; how do we tell the good ones from the bad ones?  If we find a
 * bad jaggie, do we trim one side or fatten the other?  At this point, we
 * don't quite know what the original looked like.  If we had a robust theory
 * of image content, perhaps we could derive the original.  Another opportunity
 * for breakthrough, principled AI technology.  Surely the NEA would
 * be interested.
 *
 * In the meantime, we'll just try not to screw up in the first place.  The
 * problem was in rotating binarized images.  If we have a gray scale image,
 * we can interpolate between pixels.  This is important because the pixels
 * in the destination image are not aligned with those in the source image.
 * Imagine holding two pieces of superimposed graph paper up to the light, then
 * rotating one slightly.  The vertices no longer line up.  Bummer.  All
 * of the other implementations in this library assume that input pixels
 * and output pixels are exactly aligned.  This means we can't just blindly
 * move over one pixel in the source image to produce a new pixel in the
 * destination image.
 *
 * Just for fun, how far off do things get?  Say the rotation angle is 5 deg.
 * Say the image is 3300 pixels tall.  COS(5 degrees) = 0.996195.  If we
 * are rotating around the upper left corner, the pixel in the lower right
 * corner of the destination image is found by moving down by 3287.4425
 * (3300 * cos(5)) rows and over by 287.614 (3300 * sin(5)) columns.
 *
 * The whole trick is to find the right source pixels for each destination
 * pixel and figure out how to combine them.  Good, clear, straightforward
 * implementations loop through all the destination pixels - computing
 * locations of the source pixels that affect each dest pixel, computing
 * a set of weights and combining the source pixels using the weights.
 * Good, clear implementations of straightforward algorithms take
 * over 6 MINUTES to rotate a 15 Mpixel image.  A much more clever
 * implementation cut this down to 21 seconds.  That's nice, but 21 seconds
 * is a long time for a copier to do anything.  We'll try to do a bit better.
 *
 * We'll apply the same sorts of techniques used in other alplib accelerators.
 * Writes should be minimized.  Try not to read a pixel more than once.
 * Hold partial results in registers.  Use table lookup for any vaguely
 * complex functions (like anything that takes more than a few cycles to
 * compute).  Pack several pixels into a register and operate on all of them
 * at once.
 *
 * Since we will be combining weighted pixel values with their neighbors,
 * this function can be viewed as a kind of convolution.  That means we want
 * some big routine filled with mostly straight-line code that processes
 * horizontal slices of the image and keeps intermediate results in registers.
 * The height of the slice is determined by the number of results required and
 * the number of registers available.
 *
 * The thing that makes this convolution Special (as in Special Olympics)
 * is how we load the input pixels.  Since the input and output pixels
 * don't line up, we can't just grab one input pixel for every output pixel.
 * Even worse, the input pixels are skewed with respect to the outputs.
 * We have a choice of having a nice linear input sampling and a weird
 * output stream or a weird input stream and a nice output stream.  By "weird"
 * we mean a stream that's usually straight, but occasionally turns a
 * corner or touches a pixel twice.  On the SPARCStation 2, writes are much
 * more expensive than reads, so we want to pack lots of pixels together
 * on every write.  It is particularly important to move data efficiently
 * since gray scale images are so big.  That leads us to choose a nice
 * output stream and a weird input stream.  We'll read bytes and write
 * double words.
 *
 * Another source of inefficiency is redundant read operations.  In the
 * classic algorithm, each pixel is read about 4 times, once for each of
 * the destination pixels it affects.  Each of these four samples is weighted
 * differently.  We would like to read each input pixel exactly once.  But,
 * life is hard.  Requiring each input pixel be read exactly once and each
 * destination pixel be written exactly once requires a huge amount of
 * intermediate storage (registers).  Say you decide to compute the value of
 * a destination pixel.  You read the surrounding source pixels.  Since you've
 * now read them, you must compute the destination pixels surrounding all
 * the source pixels.  That requires reading all the source pixels around them
 * and so on.  Since the number of registers is limited, we choose to
 * process the source image in diagonal stripes - reading the pixels on the
 * edges of the stripes twice.
 *
 * This business of taking diagonal paths through the source image is trickier
 * than it looks.  At least it seems that way to those of us with double digit
 * IQ's.  Say the basic algorithm is: A) move diagonally to find a new
 * source pixel; B) combine it with all affected destination pixels; C) shift
 * out any complete destination pixels.  Seems simple, right?  Not quite.  To
 * move diagonally, we add cos(rotAngle) to the x coordinate and sin(rotAngle)
 * to the y coordinate.  Both of these values are less than 1.  Therefore,
 * we will occasionally sample an input pixel twice.  If nothing is done,
 * we'll spread that input pixel's value into the output image *twice*.
 * This would probably result in bright streaks in the output image.
 *
 * The criterion for correctness is that each source pixel effecting a
 * destination pixel should be combined with the dest pixel exactly once.  I
 * tried to find all the tricky ways this simple algorithm could combine
 * a source pixel with a given dest pixel more than once.  Several fixes were
 * proposed, none of them nice.  And what about source pixels that are skipped
 * entirely because the diagonal motion avoided them?  None of this worked
 * well.
 *
 * If going through the source pixels at an angle is hard, why not
 * process the source image in simple left-to-right, top-to-bottom
 * and emit whatever destination pixels we happen to produce?  This is a real
 * mess.  First, we need some method of figuring out when a destination pixel
 * has been fully formed.  Next, we need some sort of large buffer for
 * partially formed destination pixels.  This buffer holds dest pixels that
 * were touched by a sweep through the image, but must be passed on to
 * the next sweep for further processing.  Third, we need some way to decide
 * when destination pixels should be shifted relative to the source.
 * Finally, memory writes become a real speed problem since each horizontal
 * pass through the source image will write ragged sequences of pixels in many
 * lines of the destination image.  Aligning and merging these sequences
 * efficiently is hard.
 *
 * However, some sort of scheme that goes through the source pixels in a
 * non-repetitive manner is desireable.  The current method is a compromise.
 * Start with the idea that we'll compute several lines of destination data
 * in each diagonal pass through the source image.  This maintains a clean
 * output data stream.  Say we produce h lines of destination data on each
 * pass.  In the previous scheme, we'd produce one new h pixel tall column of
 * output bits by trying to read exactly the correct skewed column of h+2 input
 * bits.  We got into trouble attempting to guarantee that the union of these
 * little skewed columns covered every input pixel exactly once.
 *
 * Our compromise is to get rid of the skew in the input columns.
 * Picture a little box containing partially calculated output pixels moving
 * diagonally across the input image.  The output pixels in the box are
 * aligned with the output image, hence skewed with respect to the input
 * image.  In the previous scheme, we tried to feed the box with little
 * columns of data that were aligned with the box, hence skewed with respect
 * to the input.  In this scheme, we move the box across the input image by
 * feeding it successive vertical (with respect to the input image) columns
 * of data.  We adjust the vertical position of each column so it covers
 * the box.  However, since the input columns are vertical with respect to
 * the input image, they're skewed with respect to the data being computed.
 * That's ok so long as the box is wide enough (holds enough horizontal output
 * pixels) so all the pixels affected by each input column fit inside the box.
 *
 * TIME OUT!  It's important to note that normal algorithms choose an output
 * pixel, find all the relevant input pixels, combine them somehow, and write
 * the output.  We're proposing doing the reverse - read an input pixel,
 * distribute it among all relevant output pixels, then toss it. TIME IN!
 *
 * An example is in order.  Say we want to handle rotations of up to 10
 * degrees.  The relative horizontal skew of input pixels vs output pixels
 * as we move down through the box will be cos(10) = 0.17 pixels/line.  If
 * we want the box to be 5 destination lines tall, we must expect that the
 * column of input pixels will appear to move horizontally (with respect to
 * the output columns) by (5 + 2) * 0.17 = 1.2 pixels from the square above
 * the top line to the square below the bottom line.  Since each input pixel
 * can affect the output pixels on both left and right, the box should be
 * 1.2 + 1 + 1 = 3.2 pixels wide.  Four pixels is a nice round number.  It
 * also fits into a register nicely.
 *
 * Intuitively, how does this scheme work?  It has a funny mix of unconditional
 * and conditional steps.  Say we're finished with a column of input data and
 * wish to start on the next h pixel tall column.  We always move one input
 * column to the right.  If the bottom edge of the uppermost pixel in the
 * new input column does not overlap the top pixels in the box, we move the
 * input column up or down by one pixel (up for clockwise rotations, down for
 * counterclockwise).  If all of the output pixels affected by the new column
 * fit in the box, we leave it alone; otherwise the pixels in the box are
 * shifted left and a new column of h output pixels is emitted.  We then proceed
 * down the column, splitting each input pixel among several output pixels.
 * Moving down by one input line normally implies moving down by one output
 * line, but not always.  Occasionally the output line is unchanged.
 * We move down the input column until none of the pixels in the box will
 * be affected by lower input pixels.
 *
 * The conditional steps are 1) whether to adjust the choice of upper input
 * pixel; 2) whether to shift the pixels in the box; and 3) whether each new
 * vertical input pixel should be combined with a new row of destination pixels
 * or the same row of dest pixels as the previous input pixel.
 *
 * How do we make all these decisions?  They are dependent on the relative
 * positions of the input and output pixels.  Since we measure pixels relative
 * to the upper left corner of the image, let's assume each pixel's coordinates
 * refer to the upper left corner of the pixel.
 *
 * What is our frame of reference?  It will turn out that we'll use both the
 * input and output images as reference frames.  As we march along in unit
 * steps in the input image, we need to know our position in the output's
 * frame of reference so we can decide which output pixels receive portions
 * of the input pixel's value.  We'll also need to know our position relative
 * to the input frame so we can know when we're off the edge of the input
 * image.  In discusions below, all references to fractional pixel position
 * refer to the position of a given input pixel in the output's frame of
 * reference.
 *
 * Two types of decision depend on coordinates of destination pixels.
 * We use the position of an input pixel in the output image to decide whether
 * to move columns up or down.  We also use this position to decide whether
 * to shift the output pixels (essentially deciding whether to move the
 * output box to the right or not).  Since we wish to do very careful rotations
 * (we should be able to distinguish between slopes differing in 1 part in
 * 1000), these decisions require a great deal of precision.  The other
 * position-dependent decision allocates an input pixel's value among several
 * output pixels.  Since intensity values are limited to [0, 256] ([0, 64]
 * for today's scanners), much less precision is required.  We'll do the
 * intensity spreading with a table lookup based on intensity and x,y offsets.
 *
 * Theoretically, these offsets are continuous values.  Using floating point
 * numbers as table indices doesn't work.  Using large integers isn't a very
 * good idea either - the tables get too large.  However, we only need a few
 * bits of x,y offset for the table lookup.  We'll maintain exact position
 * information using split integers containing at least 11 bits of fraction
 * and 16 bits of integer.  Moving horizontally by 1 pixel means moving
 * horizontally by cos(rotAngle) and vertically by sin(rotAngle) in the
 * destination's coordinate system.  Both of these values can be calculated
 * once at the beginning of the rotation operation and converted to split
 * integers.
 *
 * We'll divide each pixel in the source image into an 8x8 square.  The box
 * of partially formed destination pixels will be 4 pixels wide and
 * RBOX_HEIGHT lines tall (which depends on number of registers available).
 * In the C implementation, RBOX_HEIGHT will be fairly short, say 3, because
 * we can't control the compiler's register usage and don't want to tempt it
 * to store partial results in local variables.  The assembly code
 * implementation (on the SPARC anyway) will use a taller block because we
 * can control the registers exactly.  The taller the block, the fewer
 * redundant image reads.
 *
 * The box will contain four columns of pixels, each taking up one byte in
 * the stack of 32-bit registers.  If the input column slants
 * towards the lower right, we start it in the upper left.  In fact, the
 * topmost pixel of the input column is the one whose lower left corner
 * intersects the box's upper left pixel.
 *
 * If the input column slants to the lower left, the topmost input pixel
 * is the one whose lower right corner intersects the box's upper right
 * pixel.  Sorry about the lack of graphics here.  I encourage you to draw
 * in a notebook.  I've got pages of little boxes and arrows in mine.
 *
 * These positions are calculated with full accuracy.  What about the
 * value-spreading calculations?  These will use table lookup.  Since we'll
 * get 4 bytes (the width of the box) from the table, we'll calculate the
 * position of the input pixel relative to the left edge of the box.  Since
 * the table only returns spreading data for one row of output, we'll
 * calculate y position of an input pixel relative to the row
 * containing the top edge of the input pixel.
 *
 * I'm not exactly sure how much accuracy is required for the value spreading
 * calculations.  For a first attempt, let's divide each each pixel into an
 * 8x8 array.  X position can therefore range from 0 to 31 (2**5) and 
 * position from 0 through 7 (2**3).  Say we only use the upper 6 bits of
 * each pixel's intensity.  That means 5 + 3 + 6 = 14 bits of table lookup
 * address.  Each table entry contains 4 bytes.  That's 16K * 4B = 64KB per
 * table.  Ouch.
 *
 * How many tables do we need?  Since we calculate each pixel's position
 * relative to the destination row containing the upper edge of the input pixel,
 * each input pixel is divided between one dest line and the dest line below
 * that.  Sounds like two tables.  However, the fractions of pixel value
 * applied to the two lines are closely related.  There are 8 possible vertical
 * positions for a pixel of given horizontal position and intensity.  Since the
 * lookup table entry for (x,,y,,pixel) gives the amounts of the pixel lying
 * above the boundary to the next lower dest line in each column, the lookup
 * table entry for (x,,(8-y),,pixel) gives the fractions of pixel value lying
 * below the line in each column.  Therefore, we can get away with only 1
 * lookup table.  However, this does require a lookup table with 9 y
 * entries (0 through 8).  Nine isn't nearly as nice as eight.  We'll fix
 * that below.
 * 
 * We thought about spreading the source pixel values out just a smidgen
 * farther.  There's a little bit of rotation we're ignoring and the relative
 * grid displacements won't be exact.  We could adjust the lookup tables to
 * spread a few crumbs of pixel value to pixels that just barely missed out.
 * But, how much should be spread out, and where?  Any adjustment will help
 * half of the time and hurt half of the time.  Say we assume that an offset
 * of 0 relative to the top of a row really means an average offset of 1/16
 * of a row.  The y offsets and proportions of pixel value applied to the
 * current row are:
 *
 *	Y Offset from	Pixel Fraction
 *	Top of Row	in Current Row
 *	  0		  15/16
 *	  1		  13/16
 *	  2		  11/16
 *	  3		   9/16
 *	  4		   7/16
 *	  5		   5/16
 *	  6		   3/16
 *	  7		   1/16
 *
 * This assignment also has the nice property of simplifying the calculation
 * of "fraction of pixel lying in the next row".  Indexing the lookup table with
 * (x,,y,,pixel) and (x,,(y XOR 7),,pixel) gives two results that add to
 * the pixel value.  The lookup table can be reduced to eight y entries again.
 *
 * We'll make a similar assignment of pixel value to x offset.
 *
 * We should be able to sketch out the inner loop by now.  This is the inner
 * loop for counterclockwise rotation.  The differences for clockwise
 * rotation are discussed below.  In counterclockwise rotation, the input
 * columns slant towards the lower right of the box.  The loop starts by
 * finding the topmost pixel of a new input column.
 *
 *   1.	Advance the address of the topmost input pixel right by 1.
 *   2.	Add cos(rotAngle) and sin(rotAngle) to the destination-relative
 *	x and y coordinates of the topmost input pixel respectively.
 *   3.	If the x coordinate is now to the right of the leftmost box column
 *	(there was a carry from the fraction to the integer of the x coord),
 *	shift one column of destination data left into holding registers.
 *   4.	If the holding registers are now full, write them all to a four-pixel
 *	column in the destination image.
 *   5. If the y coordinate of the topmost pixel is now above the row of
 *	dest pixels *above* the box, adjust the topmost pixel down by 1
 *	place and add {sin(rotAngle), cos(rotAngle)} to it's {x,y} coordinate
 *	pair.
 *   6.	Read the next ("topmost" if this is the first time through this
 *	line for a new input column) pixel from the input image.
 *   7.	Add GSRotTable(x,,(y XOR 7),,pixel) to the top row of the image.
 *   8.	Adjust the input image location down by 1 row.
 *   9.	Add {sin(rotAngle), cos(rotAngle)} to it's {x,y} coordinate
 *	pair.
 *  10.	If the y coordinate did not cross an integer boundary, repeat steps
 *	6 through 10 with the new pixel value.
 *  11. Read the next pixel from the input image.
 *  12. Add GSRotTable(y,,x,,pixel) to the current image row.
 *  13. Add GSRotTable((y XOR 7),,x,,pixel) to the image row below this one.
 *  14. Adjust the input image location down by 1 row.
 *  15.	Add {sin(rotAngle), cos(rotAngle)} to it's {x,y} coordinate
 *	pair.
 *  16.	If the y coordinate did not cross an integer boundary, repeat steps
 *	11 through 16 with the new pixel *on the same output rows*.
 *  17.	Advance down one output row.
 *  18. Repeat steps 11 through 17 until the bottom row of the box is reached.
 *  19. Repeat steps 11, 12, 14, 16 and the check in step 16 until the bottom
 *	row is completed.
 *  20. Repeat steps 1 through 19 until either the destination row is complete
 *	or all valid input pixels have been read.
 *
 * To do clockwise rotation, we shift the box pixels as before.  We move the
 * topmost pixel up until it's upper left corner lies in the pixel above the
 * THIRD column of the box (i.e. overlaps the pixel above the third and
 * fourth columns).
 *
 * Concatenating 5 bits of x coord, 3 bits of y coord and 6 bits of pixel value
 * is a little messy.  To make things a bit easier, we'll assume that the
 * value of the y coord seldom changes.  Well, maybe it's more correct to say
 * that the integer portion of the y coord almost always changes by 1, but
 * the top 3 bits of the fractional part of the y coord seldom change.  Those
 * are the bits we use in the table lookup.  Why don't they change?
 * We add cos(rotAngle) to the y coord.  This is a number that is very close
 * to 1, i.e. has 1's in the three most significant bits of the fractional
 * part.  A sequence of y coordinates as we go down the input column
 * will be something like 17.62, 18.61, 19.60, 20.59, 21.58, etc.  Note
 * that there are lots of .6x's then lots of .5x's and so on.  Since we're
 * only using the top few bits of the fractional part in the table lookup,
 * we can write the code so that it's very fast when these bits do not
 * change (and somewhat slower when they do).  We will have two pointers to
 * the lookup table, one with the current value of y and the other with the
 * current value of (y XOR 7).  When we add cos(rotAngle) to the y coord, we
 * check to see if there was a carry into the 3 MS fraction bits.  If so, we
 * need only combine the upper fraction bits of the X coord (which usually
 * change) and the pixel value.  If there was no carry into these bits, we
 * go fix up the table pointers and the y coord.  This isn't a huge savings,
 * but is in the inner loop so is important.
 *
 * What about the boundary conditions?  They're untidy.  Some portions of
 * the input image are left out of the output.  Some portions of the output
 * come from outside of the input.  We could have made the output image bigger
 * so that it would hold a (roughly) trapezoidal copy of the input.  After a
 * little canvassing, we decided to keep the input and output the same size.
 * If we're about to print the silly image, it has to fit on a piece of
 * standard paper anyway, so we might as well retain the image size.
 *
 * We also decided to rotate the input around its center.  This is what Dan
 * Bloomberg's most popular rotation routines do.  Dan admits that if we
 * happen to know that one of the corners is special (as might be the case
 * if we knew that there was a bar along the left side of the platen and
 * customers often put only one corner of the paper against the bar), we could
 * rotate about that corner.  But we don't know that, so rotation about the
 * center is safest.
 *
 * Can we depend on the image frames to supply the pixels that are rotated
 * from off the sides?  No, for two reasons.  First, the frames aren't wide
 * enough.  Second, the frames are filled with 0's (black).  We want the
 * fabricated image pixels to be white; like paper or the white background
 * of document feeders.  Yes, in the glorious future the fill in color would
 * be a parameter.  Let's get the basic stuff to work, then see if there's
 * any demand for this feature.
 *
 * A diagonal path may cause an entire input column to cross a vertical
 * input edge.  The path may also cause pieces of an input column
 * to fall off the top or bottom of the image.  We could maintain separate
 * X and Y coordinates for each pixel read, check them against the image
 * boundaries and supply 0xFF's appropriately.  This sounds awful, especially
 * since the vast majority of the time is spent inside the boundaries of
 * the input image.
 *
 * Any diagonal sweep across the input image may start (paritally) off the
 * image, go across the image, then might finish (partially) off the image.
 * The sweep may also start or finish inside the input image.  Let's
 * write two loops: one that checks every input pixel and one that trusts
 * all input pixels to be on the input image.  The first loop will
 * be executed near or beyond the edges and the second, much faster, loop
 * would process the bulk of the data in the middle.
 *
 * How do we decide when to execute which loop, and when is the decision made?
 * Calculating the exact intersection point of the sweep with an edge is a bit
 * tricky.  Oh, the X coord is ok since we only read vertical columns.  It's
 * the y coord that's messy.  We'd really like to check every input address
 * before trying to read the pixel.
 *
 * A cheap way to do that is to write two loops as described above, but choose
 * which loop to execute every time we access a new input column.  If the
 * column is off either edge horizontally, we do the checking loop.  If either
 * the top or bottom ends of the column are above or below the image, do the
 * checking loop.  If all of that's ok, do the fast loop.  Putting one check
 * at the top of the pass through a column is pretty lightweight.
 *
 * We can make this check a bit faster by noting that it's not necessary to
 * check to see if the X value is both to the right of the left input image
 * edge and to the left of the right edge.  Because the source and dest images
 * are the same size, any given sweep across the image crosses at most one
 * horizontal edge and at most one vertical edge.  To cross both the
 * left and right vertical edges, the cosine of the rotation angle would have
 * to be greater than 1.  The same reasoning applies to the Y coordinate.
 * This means higher level C code (rotateGrayPr) can figure out the distance
 * from the starting point of the sweep to the nearest edge in both X and Y.
 *
 * Instead of passing the real distances to the inner loop code, we'll pass
 * related counts.  These X and Y counts are maintained by the code doing the
 * sweep - we add 1 to the X count every time we move right and 1 to the Y
 * count every time we move down.  The rotation routines set the xCount
 * to xICoord and yCount to yICoord.  The lower level routines do unsigned
 * compares of xCount >= wS and yCount >= hS.  These return "true" if
 * xCount is negative (off the image to the left) or xCount is off the
 * image to the right.  The same trick works for yCount.
 *
 * There's something implicit here that ought to be explicit.  At least it
 * confused me for awhile.  The algorithm will be maintaining 3 sets of
 * coordinates:
 *
 *	1. Input pixel position relative to input image
 *		These are the counts used to detect the boundaries of the
 *		input image.
 *
 *	2. Input pixel position relative to the output image
 *		These are used to adjust the relative positions of the input
 *		columns and the box.  They're also used to distribute input
 *		data among output pixels.
 *
 *	3. Output pixel position relative to the output image
 *		Used to decide where to write output bits.  Also used for
 *		the loop counter.
 *
 * The important thing to remember is that when we change our input focus, we
 * need to adjust *both* of the input pixel positions.  We also need to adjust
 * the pointer into the input image.
 *
 * Here's a really annoying detail.  It's of interest only because I got it
 * wrong the first time.  When we move horizontally in the input image, 
 * *exactly* how do we adjust the coordinates of the input pixel in the output's
 * frame of reference?  Our convention is that positive rotation angles imply
 * clockwise rotation.  The cosine function is symmetric and positive around
 * 0.  The sine function crosses through (0, 0) with a positive slope.  To
 * rotate an image clockwise, we sweep across it in paths that go from lower
 * left to upper right.  These paths put out horizontal pixels in the rotated
 * image.  Therefore, relative the *rotated*, or output, image, a move to the
 * left in the input image is a move towards the *lower* right in the output
 * image.  I always get confused by inversions.  Here's a table:
 *
 *	Direction in		Coord. mods in
 *	Input Image		Output Image
 *	  +X (to the right)	X = X + cosRotAngle; Y = Y + sinRotAngle
 *	  -X (to the left)	X = X - cosRotAngle; Y = Y - sinRotAngle
 *	  +Y (down)		X = X - sinRotAngle; Y = Y + cosRotAngle
 *	  -Y (up)		X = X + sinRotAngle; Y = Y - cosRotAngle
 *
 * Note that these formulas work for both positive and negative rotation
 * angles.  In these formulas, the cosine term is the major change of
 * direction and the sine just specifies the tilt.
 *
 * We've spent a lot of time discussing how the inner loop code traverses
 * across the image.  How do these traverses start?  How do they end?
 * Remember that the input columns are slanted with respect to the box.  We
 * can't just start with the input column whose upper pixel is immediately
 * to the left of the upper box pixel.  If the input column slants to the
 * lower right, the box pixels in the lower left won't be touched, hence will
 * be incorrect.  The same sort of reasoning applies at the end of the scan -
 * we must keep going until all the pixels in the last column are properly
 * calculated and written out.  This means we must start about 3 pixels to
 * the left of the desired input location and finish whenever the last column
 * has finally been written.
 *
 * How do we find the leftmost pixel in each sweep of the input EXACTLY?
 * If we miscalculate, adjacent sweeps of pixels in the output image will be
 * misaligned.  After a ridiculous amount of whining about this, the answer
 * became obvious.  The inner loop traces a diagonal path through the input
 * image to follow a horizontal path in the output image.  To find the
 * starting position of the next sweep, we trace a diagonal path in the input
 * image following a vertical path in the output image.  In the inner loop,
 * we always move horizontally by 1 column and conditionally move up or down
 * by 1.  To trace the left edge, we move vertically down the input and
 * conditionally move right or left by 1.  In this process we use the
 * left/right movement to stay in the input pixels that should be distributed
 * to column -3 of the output.
 *
 * There is one funny difference between the horizontal and vertical edge
 * tracing routines.  In horizontal edge tracing, we *always* move one input
 * column to the right.  We never stay in an input column and never backtrack.
 * That's how we guarantee that every pixel is used exactly once.  In the
 * vertical case, we want to make sure we don't miss a pixel at the very
 * beginning.  Once we start on a horizontal path, we won't miss anything, but
 * we must guarantee that the very first pixel chosen was correct.
 *
 * Ok, how could we possibly miss a relevant pixel?  After all, we stay in the
 * correct input column and stop moving down when the coordinates of the chosen
 * input pixel in the output frame of reference overlap the box.
 * Say we're advancing down a column of input pixels and are forced to take
 * a right or left jog.  Say the coordinates of the bottom of the resulting
 * pixel in the output frame of reference are below the top of the box.  We
 * should go no lower.  However, since that last horizontal movement also
 * entailed a vertical movement in the output coordinate frame of reference,
 * we might have gone too far.  We must check to see if the bottom of the
 * input pixel immediately above the one chosen is also below the top of the
 * box.  If so, move up one input pixel.  Yes, moving up one pixel might take
 * us out of column (-3), but that pixel is affects our output pixels, so we
 * should use it.
 *
 * Yeah, but doesn't something similar happen when moving horizontally?
 * The analogous situation would have us move over one input column then
 * up/down one input row and have the total horizontal movement in output
 * space appear to skip an output row entirely.  It doesn't make sense to
 * back up one input column, we've already distributed those inputs into the
 * output image.  Should we shift the output box by 2 pixels to catch up?
 *
 * Yes.  This drove me nuts for awhile.  Since we add cos(rotAngle) to the
 * X coord of the input column in the output space, and cos(anything) < 1,
 * doesn't that mean we need to traverse more than n input columns to produce
 * n output columns?  If that's so, why would we ever shift more than one
 * output column for one input column?  The conceptual roadblock here is that
 * I assumed we *always* add cos(rotAngle) to the output coord.  We don't.
 * When there's a jog in the vertical position of the input column, we also
 * add sin(rotAngle).  It turns out that we actually traverse *fewer* input
 * columns than output columns.  Imagine the diagonal path the output box
 * traces across the input.  At each step, we advance 1 input column and
 * some number of output columns.  If we end up traversing n output columns
 * We'll only have traversed n*cos(rotAngle) input columns.  Therefore, we
 * occasionally need to shift 2 output columns for one input column.  However,
 * there may still be cases where a shift of one input column doesn't cross
 * an output column boundary, and would result in no output shifts.  We
 * must be able to handle 0, 1, or 2 output shifts.
 *
 * Finally, there's the issue of smashing the bottom frame in the destination
 * image.  Since we process a few lines at once, there's a good chance that
 * the last sweep across the gray image will overlap the bottom frame.  It's
 * easier to restore the frame than to put in a bunch of checks.  We'll do
 * that.
 *
 * *********************** FIRST REWORK ********************************
 *
 * After lots of little bug fixes, there were some annoying artifacts.
 * First, pixel values overflowed.  This caused pixels that should have been
 * extra bright to be dark.  This was fixed by using 7 bits of pixel value
 * and an overflow bit.
 *
 * A more serious, as in more difficult to remove, artifact was that at certain
 * rotation angles, 7 or 8 degrees in particular, rotated images had thin
 * lighter crosshatching superimposed on the real data.  The crosshatching
 * wasn't white, just noticeably lighter.  The organized pattern of the light
 * pixels made it stand out much more than the few, random light pixels
 * that appeared at smaller rotation angles.
 *
 * As we move down a column or across the image, we add cos(rotAngle) to the
 * y or x coordinate.  As discussed above, cos(rotAngle) is just slightly
 * less than 1 for our angles.  Occasionally, the fractional part of the
 * y or x coordinate would "back up."  For example, it might go from 1.7 to
 * 2.6.  So what?  Well, the lookup table makes the not-quite-correct
 * assumption that each individual input pixel is aligned with the output
 * image.  By knowing only the upper left corner of the input pixel, we can
 * determine, with less than perfect accuracy, how a pixel's value should
 * be distributed among the output pixels in the box.  There is a critical
 * assumption here that input pixels and output pixels are the same size.
 * When a fractional y or x coordinate backs up, two input pixels appear
 * to overlap in the output's frame of reference.  That is, we apply two
 * input pixels to the output, but don't separate them by the height or
 * width of a full pixel.  The result is that one or more output pixels
 * receive extra contributions from the inputs, so get too bright.
 *
 * What to do?  We can't just pretend the cosine terms are equal to 1, there
 * would be no rotation.
 *
 * We try two solutions, one each for the Y and X coordinates.  Both are based
 * on the observation that having the input and output frames of reference
 * occasionally misaligned by 1/8 of a pixel is far less noticeable than
 * allowing pixels to overlap.  It's ok to pretend the cosine term is 1 for
 * a little while, so long as we either start over or fix the error before
 * long.
 *
 * It's fairly obvious that our box isn't very tall.  We naturally start over
 * on every column, so here's an opportunity to cheat.  As we go down an
 * input column, we simply use the same fractional y coordinate for all input
 * pixels in that column.  This actually speeds things up significantly since
 * we don't have to maintain the y coordinate.  The box is so short that
 * we'll be off by at most 1/8 of a pixel in the y dimension by the time we
 * get to the bottom.  This also eliminates all the code that checks to see
 * if we want to apply a new input pixel to the same output line.  The result
 * of this fix was to eliminate one dimension of the crosshatching.  We were
 * left with an image with a bunch of parallel slanted lines - presumably
 * due to overlaps in the X dimension.
 *
 * It's a real pain to start over in the X dimension.  Every pixel affects
 * two output pixels in the box.  The box is actually a little pipeline.
 * Starting and stopping it is slow and difficult to get exactly right.
 * So, is there some opportunity to pretend X movement is by 1 in the output
 * dimension and fix things up occasionally?  Fixing things up at some fixed
 * interval would be a disaster, we'd just have overlap at that interval.
 * ideally, we'd like to do the fixup (subtracting a little bit from the
 * X position of the input in the output's frame of reference) at some time
 * when that position was going to change anyway.
 *
 * So I tried increasing box's X by 1 between jogs and making up for it at
 * the jog.  Didn't work - the fractional parts changed randomly at the jogs.
 * Tried not changing the box's X at the jogs, just allowing it to catch up
 * when the real X position got ahead.  Didn't work - the integer part of the
 * box's x position would jump.  And, a real heartbreak, if you do not
 * change the box's X at a jog, you get a guaranteed overlap.  That's because
 * we want to move at an angle at the jog (in the output's frame of ref).
 * If we only change the y position, the next sweep through the box overlaps
 * the entire new column with an old one.  Total bummer - you can't change
 * the box's X when you want, but you must change it at every jog.
 * 
 * So I got crude.  ONLY change the box's X position at the jogs.  There are
 * no overlaps between jogs.  There are no overlaps at the jogs.  Sure enough,
 * all the streaks went away.
 *
 * Yeah, but aren't the results all distorted?  Isn't it really cheating to
 * increased the X position by 1 instead of cos(rotAngle) between jogs and
 * by the full sin(rotAngle) at the jogs?  Ummm.  Well it seems like that to
 * me, but all the images, even the really big ones, seem to look just fine.
 *
 * ************************ SECOND REWORK *********************************
 *
 * Lovely stuff, but too slow.  It took over 14 seconds to rotate
 * an 8 MByte gray image.  We'd like something about 50% faster, even in the
 * C implementation.  Careful study revealed that we were only spending
 * about 30% of our time in the code that combined pixels.  The rest of the
 * time was spend setting things up and writing results out.  Ugh.
 *
 * The algorithm described above is too complex.  The real inner loop only
 * retires RBOX_HEIGHT pixels before messing around with column adjustment.
 * There are too many variables - the compiler spends too much time loading
 * and reloading.  We spend too much time shifting each pixel out of
 * its evaluation register into its holding register before actually writing
 * it to memory.
 *
 * One way to simplify things is to forget about trying to compute several
 * lines at once.  In this case, computing several lines at once does
 * little to enhance peformance of the inner loop, but substantially
 * complicates life.  While it's true that an input pixel generally affects
 * two output lines, each output line requires its own read of the
 * grayRotationTable.  There are no partial results that can be reduced by
 * producing multiple lines.  On the down side, maintaining multiple lines
 * of output pixels means we'll never be able to do double writes.  These are
 * pretty important when writing huge images.
 *
 * If we're going to produce one line at a time, our inner loop will have to
 * move horizontally instead of vertically.  That's ok.  We discovered above
 * that it's legal to move horizontally by one input and one output pixel at
 * each step, which makes things easier.  However, just one pass across the
 * output line (really line segment) won't work.  Since the input and output
 * are tilted with respect to each other, several slanted input lines will
 * affect our line segment.
 *
 * How do we find the beginning and end of each input line segment?  Let's use
 * clockwise rotation as an example.  Here, input lines slope from upper left
 * to lower right relative to the output line.
 *
 * Say we're going to compute all the pixels in a short line segment, say
 * 8 pixels long.  They can be held in 2 registers (and written using a double
 * write in assembly code).  We start with the input pixel whose Y output
 * coordinate is at or immediately below the line of outputs being computed.
 * We follow that line of input pixels to the right, distributing them among
 * the output pixels, until the Y coordinate in the output's frame of
 * reference is below the line being computed.
 *
 * We then move up one line in the input space.  We may also wish to move to
 * the left by one pixel.  The input pixel finally chosen should be the first
 * one above and to the left of the leftmost output pixel.  We follow this
 * line of input pixels until it crosses the y coordinate of the output line.
 * In this section of the traverse, we use the portion of the grayRotationTable
 * that distributes pixel values to the line below the actual pixel.  This
 * is done by inverting the Y coord before adding it to the lookup table index
 * (see comments above).  When the Y coord of the input line becomes equal
 * to the Y coord of the output line, two actions are taken.  First, we record
 * the X coordinate (relative to the input f.o.r.) where the crossover
 * happened.  This let's us remember where to restart.  Second, we start using
 * the section of the grayRotationTable that distributes pixel values to the
 * current line (do *not* XOR the Y coord before adding to lookup table index).
 * We continue until either the Y coord crosses another line boundary or
 * the X coord crosses the end of the line segment.
 *
 * If the input's Y coord crossed from above the output line to the output
 * line, we have more work to do.  We go back to the crossover position and
 * move up one line in the input space.  This adjusts both the X and Y coords
 * in the output space.  We then proceed as above.
 *
 * When finished with a line segment, we look at the last position.  If it's
 * still above the Y coord of the output line, move down one input line and
 * possibly move to the right one pixel.  The move to the right will be taken
 * if we need to move over to intersect the leftmost output pixel in the next
 * line segment.  When the adjustment is done, start over.  Continue this
 * process until all pixels in an output line have been written.
 *
 * Counterclockwise rotation is similar except input lines to towards the
 * upper right in the output frame of reference.  To keep things tractable,
 * we'll introduce another trick here.  When doing counterclockwise rotation,
 * a pixel's coordinates will refer to it's *lower* left hand corner.  That
 * way, we can use the same inner loop code for both clockwise and
 * counterclockwise rotation by simply flipping the polarity of sinRotAngle
 * and source Byte/Line on entry for counterclockwise rotation.  Saves code.
 *
 * This approach has another nice feature.  Since we know our X position will
 * only change by integer amounts in any one sweep, we can freeze the X portion
 * of the table lookup index.  The grayRotationTable will now return 16 bit
 * quantities that are shifted into position before being combined with the
 * outputs.  This cuts the size of the grayRotationTable down to 2 ** (3 x bits
 * + 3 y bits + 7 pixel bits) * 2 bytes/entry = 16KB.
 *
 * Another feature is that the X sweeps can be unrolled.  Since we read one
 * input pixel for every output pixel, we know exactly how many processing
 * steps there are in each sweep.  We can also compile in the amounts by which
 * the 16 bit grayRotation entries will be shifted before being used.
 *
 * Horizontal input sweeps go towards the lower right in clockwise rotation
 * or the upper right in counterclockwise rotation.  It would be unfortunate
 * to have two inner loops for this silliness.  However, if we start from the
 * lower right corner of the image and declare that a pixel's coordinates refer
 * to it's lower right corner for counterclockwise rotation, we can use one
 * set of code for both rotation directions.  We'll do that.  We also have to
 * invert the sin(rotAngle), source byte/line and dest bytes/line.
 * 
 */

/* So why do an accelerated version anyway?  The code generated by the Sun
 * compiler looks pretty good.  Our problem was that we couldn't convince
 * the compiler to quit using so many local variables.  It appeared in our
 * performance tests that local variable reads at the bottom of
 * ip_rotateGrayLine's inner loop was causing the operation as a whole to go
 * at 50% to 70% of theoretical speed.  For a 2x performance gain, we'll do
 * custom assembly code.
 *
 *
 * **************************** THIRD REWORK *******************************
 *
 * As explained in ssgsrot.s, we were spending too much time in overhead.
 * Instead of doing a sufficient number of pure horizontal input sweeps to
 * cover the output, we decided to do one double-height sweep.  To make this
 * work, the top pixel of this sweep must be adjusted up each time it enters
 * into row 0.  That's pretty easy.  Many more comments on the changes are in
 * ssgsrot.s.
 *
 * The effect here is to slightly change the strategy for moving down one line.
 * Before, we wanted to stay in column -1 and either row -1 or row 0 (row above
 * output or same row as output).  Since we are switching to a scheme that
 * always samples two pixels vertically, we must always be in row -1.  We'd
 * prefer to be in col -1, but col 0 is ok.
 *
 * ************************************************************************
 *
 * 10/7/92: This routine writes the destination gray image in units of 4 bytes.
 * If the destination image width is not a multiple of 4 bytes, we've written
 * some frame pixels.  Fix this by calling ip_clearRightFrame after the rotate.
 *
 * *****************   ROTATION IN PLACE  *********************************
 *
 * This code give very nice results, but uses a lot of memory.  Gray
 * rotation is one of the only (the only?) situation in normal processing
 * that has both source and destination gray images.  At 300 spi, thats
 * two 8 MB images.  At 400 spi, that's two 16MB images.  Yuck.
 *
 * Z Smith, that font of creativity, suggested that we allocate more memory
 * than actually required for the image data in the image pixrect; then
 * rotate the image to the empty destination space.  We'd roll along, so
 * at the end would have the rotated image in some displaced position relative
 * to the source image.
 *
 * That idea was ok, but our image data structures didn't know about blank
 * spots.  So, we'll try a different approach.  i_rotateGrayIP will allocate
 * a ring buffer.  The optimized routine ip_rotateGrayLine will put its
 * rotated data into the ring buffer.
 *
 * Rotation will be done in three stages:
 *  1.	Rotate enough lines to fill the ring buffer.  At this point, we
 *	should no longer need to use the first line of the original image
 *	as input data.
 *  2.	There are two steps for each rotated line:
 *	a. Copy the next line from the ring buffer to the image.
 *	b. Fill the empty space in the ring buffer with a new rotated line.
 *	This continues until we have produced enough rotated lines.
 *  3.	In the last stage, we flush the rest of the lines in the ring buffer
 *	to the image.  This should end when we have overwritten the last
 *	image line.
 *
 * So, how big is the ring buffer?  Obviously, it must be wide enough to
 * hold rotated lines.  For the sake of simplicity, we'll make it sBpl bytes
 * wide.  How tall should it be?  We can compute the maximum vertical
 * distance a diagonal path across the input image will cross.  This is
 * equal to wS * sin(rotAngle).  The ring buffer must be at least this tall
 * because when we have just stopped touching the top of the image, the
 * lower edge of the diagonal path will be this distance below the top.  We
 * assume that each source line produces about one dest line, so there must
 * be at least this many dest lines in the ring buffer.
 *
 * When we move down one line in the dest, we only move down cos(rotAngle)
 * lines in the source.  This is very close to 1, but not quite.  Therefore,
 * When we move down hS lines in the dest, we'll only move down
 * hS * cos(rotAngle) lines in the source.  To fill a 4400 line image at an
 * angle of 10 degrees, we only cross 4333 lines of source image.  This means
 * the distance between the source image's yCoord and the dest image's
 * yCoord keeps growing as we move down the image.  If the ring buffer were
 * only tall enough to cover the number of lines needed to span the
 * vertical displacement of one path across the source image, we would
 * start writing dest lines on top of source lines that had not yet been
 * retired by the time we reached the bottom of the pile.  To account
 * for this, we look at the difference between cos(rotAngle) and 1.  At
 * a maximum rotation angle of 10 degrees, this is 0.0152, or about
 * 1 part in 65.  To be safe, we'll add hS/32 lines to the ring buffer
 * height.
 *
 * Are there any other considerations?  Do we need to put in extra lines
 * because of starting or ending conditions?  If we started n lines above
 * the top of the source image, we'd need to add n/cos(rotAngle) extra
 * lines to the ring buffer.  The current code always rotates around the
 * middle of the image, so we never start above the top of the image.  I
 * can't think of any ending conditions that would make a difference.
 *
 * Therefore the ring buffer height should be (wS * sin(rotAngle)) +
 * hS / 32 lines tall.
 *
 ************************* FIXED POINT IMPLEMENTATION ********************
 *
 * XIS supports platforms that have no floating point support.  They have
 * supplied fixed point implementations of sin, cos, atan and sqrt.  These
 * functions are mainly used to generate sinRotAngle and to find the
 * input point that corresponds to output point (0,0).  We substitute
 * the fixed point functions for their floating equvalents in a
 * straightforward way.
 *
 ********************** FIX FOR ANISOTROPIC RESOLUTIONS *****************
 *
 * Standard Fax images have anisotropic resolutions, like 200 x 100 dpi.
 * If you take a skewed image with square pixels and take out ever other
 * vertical row you introduce a horizontal shear.  Huh?  Well, reducing
 * the image in the y direction makes everything more horizontal.  If the
 * image contained (slightly skewed) horizontal and vertical lines, the
 * skewed horizontal lines become more horizontal and so do the vertical
 * lines.  For the vertical lines, this means they become less vertical.
 * the result is that the angle between the formerly orthogonal lines is
 * no longer 90 degrees.
 *
 * What to do?  There is no rotation angle which will cause the lines
 * to appear to be orthogonal.  Obviously, the reverse situation is also
 * true - if the lines were orthogonal in the anisotropic image and we
 * rotate it, then make it isotropic, they will no longer be orthogonal.
 * Our solution is to introduce a horizontal shear into the rotation
 * algorithm.  The client should specify the rotation angle that makes
 * the horizontal components of the rotated image look correct (i.e. the
 * skewed horizontal lines become really horizontal) and we will shear
 * the image so that, if one were to make the image isotropic, it would
 * look right.
 *
 * How much shear is that?  Well, say you have an isotropic image that's
 * rotated by theta.  We then make the image anisotropic with horizontal
 * resolution xRes and vertical resolution yRes.  If we run the skew
 * detect code on the new image, we get a skew angle of phi such that
 * sin(phi) = (yRes/xRes) sin(theta).  Since we calculated the skew
 * by looking at horizontal things, phi is the angle we send to this
 * routine.  It will make the skewed horizontal things really horizontal.
 *
 * In the rotation routines, we convert diagonal lines of image input
 * into horizontal lines of output.  Therefore, we happily move up or
 * down by sin(phi) for every horizontal step.  The shear comes in 
 * when we choose the left end of each diagonal line in the input image.
 * In the isotropic image, we would have moved over by sin(theta) for
 * each step down.  In the anisotropic image, we have to move over by
 * (xRes/yRes) sin(theta) = (xRes/yRes)*(xRes/yRes)*sin(phi).  Therefore,
 * we multiply the horizontal motion by (xRes/yRes) squared.
 *
 * Since we're moving around in the image in a weird way, how do we figure
 * out where the center is now?  We figure out where the center would
 * be in the isotropic image and multiply coordinate corresponding to the
 * smaller resolution by smallRes/largeRes.  For example, if xRes = 2*yRes,
 * We multiply the image height by 2 when initially figuring the center,
 * then divide the resulting y coord by 2 to get the new center.
 *
******************* SECOND ANISOTROPIC IMAGE REWORK *********************
 *
 * The alert reader (not me) will notice that the explanation for the
 * behaviour of anisotropic rotation depends on the fact that the naive
 * skew detect code returned an angle which depended on the ratio beteen
 * the x resolution and the y resolution, i.e. (yRes/xRes) * (skew of
 * the original sheet of paper).  Knowing this, the rotation routine multiplied
 * the angle (the sine of a small angle is almost equal to the angle)
 * by (yRes/xRes) before using it.  Well, this is ducky if the rotation
 * angle came from a naive skew detection routine.  What if it came from
 * somewhere else?  Or, what if we scale the image anistropically before
 * calling the rotation routine?  In either case, the "correction" supplied
 * by the rotation routines will not do the right thing.
 *
 * Our solution is to demand that angles supplied to rotation routines
 * must be independent of x and y resolution.  What does this mean?  Let's
 * introduce the "isotropic cousin" of an anisotropic image  The isotropic
 * cousin of an image is the smallest image which is isotropic
 * (x resolution == y resolution) and can be generated by scaling the
 * original image in either x or y where the scaling factor is >= 1.0.
 * For example, a standard fax image has an x resolution of 200 dpi and
 * a y resolution of 100 dpi.  The nearest isotropic image is a version of
 * the original image that has been scaled by 2x in the vertical direction
 * only, therefore has xRes = yRes = 200 dpi.  The isotropic cousin of
 * an isotropic image is the image itself.
 *
 * We require that the angles that are passed around correspond to the
 * rotation to be performed on the isotropic cousin of the actual
 * image provided.  We perform an operation on the real image such
 * that the result of our rotation-like operation followed by a scaling
 * operation that makes the result isotropic is equivalent to simply
 * rotating the isotropic cousin of the original image.
 *
 * Why this convoluted definition?  Well, when someone looks at an image,
 * it will almost surely be using isotropic resolutions.  We want to make
 * sure that the rotation-like operation we perform will, when scaled for
 * viewing, look like a real rotation.
 *
 * So, what's involved?  First, the skew detection routines must scale
 * the vertical displacements they find by (xRes/yRes) before computing
 * the arctan.  That's because the naive computation found a vert/horiz
 * ratio that was (yRes/xRes) * (vert/horiz for the isotropic cousin of
 * the image).
 *
 * Next, we look at how the rotation code uses the angle and the
 * x and y resolutions.  For small angles, cos(rotAngle) is always very
 * close to 1 even if rotAngle moves around a litte.  Therefore, we leave
 * the cosine terms alone.  However, we compute two versions
 * of sin(rotAngle).  vertSinRotAngle is used to modify X coordinates
 * when moving vertically.  horizSinRotAngle is used to modify Y coords
 * when moving horizontally.  Why do we need both?  Well, think of converting
 * an isotropic image to an anisotropic image by squishing all the individual
 * pixels, then laying the result out over an isotropic grid.  Say we make
 * the input pixels short and wide.  Then, each time we move down in the
 * input image, we have to remember to move over extra far in the output
 * image to get to the right place.  Similarly, each time we move over in
 * the input image, we have to remember that only a small movement is
 * implied vertically.
 *
 * That's basically the deal.  We must pass both horizSinRotAngle and
 * vertSinRotAngle to the lower level routines since they move both
 * horizontally and vertically so need both versions.  When moving down
 * vertically to find the beginning of the next scan line, we use
 * vertSinRotAngle as before.  When finding the upper left corner of
 * the image, we use both horizSinRotAngle and vertSinRotAngle as appropriate.
 *
 */

/* PUBLIC */
/**************************************************************************
 *  i_rotateGrayFixedPoint()
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		depth = depth of gray image.  4 and 8 bpp supported.
 *		rotAngle = rotation angle in ScaledAngle format
 *		pD = destination gray image
 *		dBpl = destination bytes/line
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *
 *	On return:
 *		Destination image contains the source image rotated by
 *		the given angle.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		unable to allocate memory
 *		ia_nopoint		could not find next point
 *		ia_depthNotSupported	wrong number of bits per pixel
 *		ia_angleTooLarge     	abs value of rotation angle is
 *					greater than 10 degrees
 *		ia_invalidParm		xRes != yRes and either xRes
 *					or yRes is 0.
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	Frame in destination image is assumed to already be cleared.
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateGrayFixedPoint (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	depth,
    ScaledAngle	rotAngle,
    UInt8      *pD,
    Int32	dBpl,
    Int16	xRes,
    Int16	yRes)
{
    PROLOGUE;
    Int32	line;			/* interation variables */
    Int32	xICoord, yICoord;	/* integers containing the
 					 * X and Y coords of the current input
					 * pixel in the input image's frame
					 * of reference.
					 */
    ScaledInt1616 xIOCoord, yIOCoord;	/* split integers containing the
					 * X and Y coords of the current
					 * input pixel in the output image's
					 * frame of reference.
					 */
    ScaledInt1616 sinRotAngle, cosRotAngle; /* split integers containing the
					   * sign and cosine of the rotation
					   * angle respectively.  These have
					   * 16 bit Int32 and frac parts */
    ScaledInt1616 vertSinRotAngle;	/* The sine of the rotation angle
					 * that's been modified to account
					 * for shears in anisotropic images.
					 * This is used to modify the X
					 * coord when moving vertically. */
    ScaledInt1616 horizSinRotAngle;	/* The sine of the rotation angle
					 * that's been modified to account
					 * for shears in anisotropic images.
					 * This is used to modify the Y
					 * coord when moving horizontally. */
    UInt32  	 xCount, yCount;	/* distances to vertical and horiz
					 * edges passed to ip_rotateGrayLine.
					 * equivalent to input coords relative
					 * to input image. */
    UInt8	*pDOrig;		/* used to remember addr of upper left
					 * corner of image so we can clear its
					 * right frame when done. */
    UInt16	*p4bppGrayRotationTable = NULL; /* table for 4bpp code. */
    UInt16      *pGrayRotationTable = NULL;     /* table for gray rotation */
#define cXCoordMask	0xF800		/* selects x bits for table
					 * index formation */
#define	cYCoordMask	0x700		/* selects y bits for
					 * table index formation */
#define	cYAndPixMask	0x7FF		/* selects all bits we want
					 * to be unaffected when X
					 * bits are inserted into table
					 * lookup.  This is a SPARC
					 * constant, so maybe the
					 * compiler will be smart enough
					 * to do an andn with this */
#define	cScaledPi	16384		/* 32768 corresponds to 2pi in
					 * scaled angle notation */
#define	cMaxRotAngle	((10 * cScaledPi) / 180) /* corresponds to 10 degrees */

	/* can we handle this depth image? */
    if ((depth != 4) && (depth != 8))
	return ia_depthNotSupported;

    if (ABS(rotAngle) > cMaxRotAngle)
	return ia_angleTooLarge;

	/* Are the resolutions reasonable? */
    if ((xRes != yRes) && ((xRes == 0) || (yRes == 0)))
	return ia_invalidParm;

	/* get the width and height of the source without the frame */
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + depth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + depth*cFrameBytes;	/* skip the gray pad */

	/* if the rotation angle is 0, just blit src to dest */
    if (rotAngle == 0)
    {
	__error__ = i_rasterOp (pIPState, pD, 0, 0, wS, hS, depth, dBpl, PIX_SRC,
			      pS, 0, 0, depth, sBpl);
	NORMAL_RETURN;
    }

	/* Remember address of upper right corner of dest image
	 * so we can clear its right frame when done. */
    pDOrig = pD;

	/* Make the rotation tables if they don't exist */
    if (depth == 4)
    {
	ERR_PASS(ip_make4bppGrayRotationTable (pIPState,
					       c4bppGrayRotTableSize,
					       &p4bppGrayRotationTable));
    }
    else
    {
	ERR_PASS(ip_makeGrayRotationTable (pIPState,
					   c8bppGrayRotTableSize,
					   &pGrayRotationTable));
    }

	/* We'd like to use one set of code for both clockwise and counter
	 * clockwise rotation.  If sinRotAngle < 0, we're doing counter
	 * clockwise rotation so should process the image starting from the
	 * lower left corner instead of the upper left corner.  Reposition
	 * pS and pD.  Invert sinRotAngle, dBpl and sBpl.
	 */
    if (rotAngle < 0)
    {
	pS += (hS-1) * sBpl;	/* point to lower right corner */
	pD += (hS-1) * dBpl;
	sBpl = -sBpl;
	dBpl = -dBpl;
	rotAngle = -rotAngle;
    }

	/* compute the sine and costine terms we'll need to relate the
	 * source and dest coordinate systems. */
    ERR_PASS(fract_sin(pIPState, rotAngle, (ScaledInt1715 *) &sinRotAngle));
    ERR_PASS(fract_cos(pIPState, rotAngle, (ScaledInt1715 *) &cosRotAngle));

	/* Since we really want ScaledInt1616's instead of ScaledInt1715's,
	 * we need to shift these left by 1.
	 */
    sinRotAngle <<= 1;
    cosRotAngle <<= 1;
    if (ABS(xRes - yRes) > 1)
    {
	vertSinRotAngle = (xRes * sinRotAngle)/yRes;
	horizSinRotAngle = (yRes * sinRotAngle)/xRes;
    }
    else
    {
	vertSinRotAngle = sinRotAngle;
	horizSinRotAngle = sinRotAngle;
    }

	/* Ron Danisewicv at XIS found a much simpler way to derive the
	 * coordinates of the first input pixel.  You just walk from the
	 * center to the upper left corner, remembering to account for
	 * the changes to the y coordinate for each horizontal move
	 * and the changes to the x coordinate for each vertical move.
	 * Let's do the horizontal moves first:
	 *	old input coords:	(wS/2, hS/2)
	 *	input x coord after moving left by wS/2 in the output:
	 *				(wS/2 - (wS/2) * cos(rotAngle))
	 *				= ((wS/2) * (1-cos(rotAngle));
	 *	input y coord after moving left by wS/2 in the output:
	 *				(hS/2 + (wS/2) * sin(rotAngle));
	 *
	 * Now move up by hS/2 in the output and trace where the input goes.
	 *	current input coord:	((wS/2) * (1-cos(rotAngle)),
	 *				 hS/2 + (wS/2) * sin(rotAngle))
	 *	final x coord after moving up by hS/2
	 *				(wS/2) * (1-cos(rotAngle)) +
	 *				  (hS/2) * -sin(rotAngle)
	 *	final y coord after moving up by hS/2
	 *				hS/2 + (wS/2) * sin(rotAngle) -
	 *					hS/2 * cos(rotAngle)
	 *				= (hS/2) * (1-cos(rotAngle)) +
	 *				   (wS/2) * sin(rotAngle)
	 *
	 * This works great for square resolutions.  What if xRes != yRes?
	 * In this case, we use the versions of sinRotAngle that account for
	 * the shears involved, namely vertSinRotAngle (used to modify X
	 * coords when moving vertically) and horizSinRotAngle (used to
	 * modify Y coords when moving horizontally).  Since sinRotAngle =
	 * vertSinRotAngle = horizSinRotAngle when xRes == yRes, we can
	 * use vertSinRotAngle and horizSinRotAngle all the time.
	 */
    xICoord = (wS/2) * (0x10000 - cosRotAngle) -
	      ((hS/2) * vertSinRotAngle);
    xICoord = ScaledInt1616_TO_Int32(xICoord);

    yICoord = ((hS/2) * (0x10000 - cosRotAngle)) +
	      ((wS/2) * horizSinRotAngle);
    yICoord = ScaledInt1616_TO_Int32(yICoord);


	/* This is the input coordinate which we decree corresponds exactly
	 * to coordinate (0,0) in the output frame of reference.  What do we
	 * do with it?  First, we adjust the input address to point to this
	 * place.  This will almost certainly point to something outside
	 * the input image, but that's ok.  We won't do any memory references
	 * with this address until we enter the input image.
	 */
    if (depth == 4)
        pS += yICoord * sBpl + (xICoord >> 1);
    else
	pS += yICoord * sBpl + xICoord;

	/* By decree, the location of the this input pixel relative to the
	 * output image is (-1.0, -1.0)
	 */
    xIOCoord = (-1<<16);
    yIOCoord = (-1<<16);

	/* Have ip_rotateGrayLine actually do the work on each line. */
    for (line = 0; line < hS; line++, pD += dBpl)
    {
	    /* At this point, xIOCoord and yIOCoord are the output coordinates
	     * of an input pixel whose upper (or lower if counter clockwise)
	     * left corner is in the row to be rotated.  If possible, the
	     * pixel is also in column -1.  If that pixel can't be reached,
	     * our first pixel is in column 0.
	     */

	    /* Calculate the distance to the first horizontal and vertical
	     * edges that we'll cross.  Lower level code checks to see if
	     * (UInt32)xCount >= (UInt32)wS.  This will be true if
	     * xCount is negative or if xCount is to the right of the image.
	     * Similarly, the lower level code does an unsigned comparison
	     * between yCount and hS.
	     */
	xCount = xICoord;
	yCount = yICoord;		/* starting near top of image */

	if (depth == 4)
	    ip_rotate4bppGrayLine (pIPState, pS, sBpl, pD, wS, vertSinRotAngle,
				   horizSinRotAngle,
		                   xIOCoord, yIOCoord, xCount, yCount,
			           xICoord&1, p4bppGrayRotationTable, hS);
	else 
	    ip_rotateGrayLine (pIPState, pS, sBpl, pD, wS, vertSinRotAngle,
			       horizSinRotAngle,
		               xIOCoord, yIOCoord, xCount, yCount,
			       pGrayRotationTable, hS);

	    /* We would like to adjust the position of the first pixel in
	     * the input row so that it is above the next output row (row -1)
	     * and col -1 or col 0.  Sometimes, we can't get there from here
	     * because moving down 1 column gets us to col -2.  If that
	     * happens, we have to think about what to do next.  Since we're
	     * always apparently rotating clockwise (remember we reversed
	     * a bunch of polarities above), moving down moves to the left
	     * in the output space.  If moving down takes us into column -2,
	     * we try moving to the right and moving diagonally.  If either of
	     * those gets us to [row, col] = [-1, -1] or [0, -1] move there.
	     * If neither of those lands in the right location, stay here.
	     * We'll start the next sweep in row -1.
	     * What happens if we show up here having done nothing in the
	     * previous pass?  If we did nothing the previous pass, we jump
	     * to the diagonal now.
	     *
	     * But, that's not all.  A single step down may not be far enough
	     * to reach the next line.  If not, it may or may not cross out
	     * of column -1.  This adds two more possible input points to the
	     * search.  We also have to look at the point 2 rows down and the
	     * one 2 rows down and 1 col to the right.
	     *
	     * Since we may be introducing a horizontal shear at this point,
	     * use vertSinRotAngle instead of sinRotAngle to move sideways
	     * and horizSinRotAngle to move up/down.
	     */
	yIOCoord -= (1<<16);		/* we're above the new line */
		
	    /* look straight down.  If we're in row -1, we might as
	     * well stay here since we're right next to the row 0 boundary. */
	if (((yIOCoord + cosRotAngle) >> 16 == -1) &&
	    ((xIOCoord - vertSinRotAngle) >> 16 > -2))	/* the common case */
	{
	    yIOCoord += cosRotAngle;
	    xIOCoord -= vertSinRotAngle;
	    yICoord += 1;
	    pS += sBpl;
	}
	else
	    /* look to the right */
	if (((yIOCoord + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += horizSinRotAngle;
	    xIOCoord += cosRotAngle;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += 1;	    
	    }
	    else pS += 1;
	}
	else
	    /* look straight down and to the right */
	if (((yIOCoord + cosRotAngle + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord - vertSinRotAngle + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += cosRotAngle + horizSinRotAngle;
	    xIOCoord += -vertSinRotAngle + cosRotAngle;
	    yICoord += 1;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += sBpl + 1;
		else pS += sBpl;	    
	    }
	    else pS += sBpl + 1;
	}
	else
	    /* look down 2 */
	if (((yIOCoord + 2*cosRotAngle) >> 16 == -1) &&
	    ((xIOCoord - 2*vertSinRotAngle) >> 16 > -2))
	{
	    yIOCoord += 2*cosRotAngle;
	    xIOCoord -= 2*vertSinRotAngle;
	    yICoord += 2;
	    pS += 2*sBpl;
	}
	else
	    /* look down 2 and to the right */
	if (((yIOCoord + 2*cosRotAngle + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord - 2*vertSinRotAngle + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += 2*cosRotAngle + horizSinRotAngle;
	    xIOCoord += -2*vertSinRotAngle + cosRotAngle;
	    yICoord += 2;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += 2*sBpl + 1;
		else pS += 2*sBpl;	    
	    }
	    else pS += 2*sBpl + 1;
	}
	else
	{
		/* Couldn't find the next point, so give up. */
    	    if (depth == 4)
		FREE(p4bppGrayRotationTable);
	    else 
		FREE(pGrayRotationTable);
	    return ia_nopoint;
	}

	if ((line & 0xFF) == 0)
	    IP_YIELD(errorFinish)

    } /* end of loop that rotates all the lines */

	/* clear the right frame of the rotated image.  It was probably
	 * written mistakenly by the rotation code. */
    if (dBpl > 0)
	ip_clearRightFrame (pIPState, pDOrig, wS, hS, dBpl, depth);
    else
	ip_clearRightFrame (pIPState, pDOrig, wS, hS, -dBpl, depth);


    if (depth == 4)
	FREE(p4bppGrayRotationTable);
    else 
	FREE(pGrayRotationTable);
    NORMAL_RETURN;

errorFinish:
    if (depth == 4)
	FREE(p4bppGrayRotationTable);
    else 
	FREE(pGrayRotationTable);
    return ia_aborted;
}


/* PUBLIC */
/**************************************************************************
 *  i_rotateGrayIPFixedPoint()
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		depth = depth of gray image.  4 and 8 bpp supported.
 *		rotAngle = rotation angle in ScaledAngle format
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *
 *	On return:
 *		Source image contains the source image rotated by
 *		the given angle.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_angleTooLarge     	abs value of rotation angle is
 *					greater than 10 degrees
 *		ia_nomem		unable to allocate memory
 *		ia_nopoint		could not find next point
 *		ia_depthNotSupported	wrong number of bits per pixel
 *		ia_invalidParm		xRes != yRes and either xRes
 *					or yRes is 0.
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateGrayIPFixedPoint (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	depth,
    ScaledAngle	rotAngle,
    Int16	xRes,
    Int16	yRes)
{
PROLOGUE;
Int32		line, ringLine;		/* interation variables */
Int32	 	xICoord, yICoord;	/* integers containing the
 					 * X and Y coords of the current input
					 * pixel in the input image's frame
					 * of reference.
					 */
Int32	 	xIOCoord, yIOCoord;	/* split integers containing the
					 * X and Y coords of the current
					 * input pixel in the output image's
					 * frame of reference.
					 */
ScaledInt1616	sinRotAngle, cosRotAngle; /* split integers containing the
					   * sign and cosine of the rotation
					   * angle respectively.  These have
					   * 16 bit Int32 and frac parts */
ScaledInt1616	vertSinRotAngle;	/* The sine of the rotation angle
					 * that's been modified to account
					 * for shears in anisotropic images.
					 * This is used to modify the X
					 * coord when moving vertically. */
ScaledInt1616	horizSinRotAngle;	/* The sine of the rotation angle
					 * that's been modified to account
					 * for shears in anisotropic images.
					 * This is used to modify the Y
					 * coord when moving horizontally. */
UInt32      	xCount, yCount;		/* distances to vertical and horiz
					 * edges passed to ip_rotateGrayLine.
					 * equivalent to input coords relative
					 * to input image. */
UInt8	       *pDOrig;			/* used to remember addr of upper left
					 * corner of image so we can clear its
					 * right frame when done. */

Int32		maxVertDisp;		/* maximum vertical distance covered
					 * by a diagonal path across the
					 * source image */
Int32		extraBufLines;		/* the number of lines added to
					 * the ring buffer to allow for the
					 * fact that we'll proceed down the
					 * dest image faster than we'll
					 * proceed down the source image.
					 */
Int32		ringBufLines;		/* total number of lines in the
					 * ring buffer */
UInt8	       *pD, *pRing, *pLine;	/* pointer to dest loc in source/dest
					 * image and ptr to ring buffer and
					 * current line in ring buffer. */
Int32		absSBpl;		/* absolute value of sBpl.  Used
					 * to cycle through ring buffer */
UInt16	       *p4bppGrayRotationTable = NULL; /* table for 4bpp code. */
UInt16         *pGrayRotationTable = NULL;     /* table for gray rotation */

#define cXCoordMask	0xF800		/* selects x bits for table
					 * index formation */
#define	cYCoordMask	0x700		/* selects y bits for
					 * table index formation */
#define	cYAndPixMask	0x7FF		/* selects all bits we want
					 * to be unaffected when X
					 * bits are inserted into table
					 * lookup.  This is a SPARC
					 * constant, so maybe the
					 * compiler will be smart enough
					 * to do an andn with this */

/* see definition of cMaxRotAngle above */

	/* can we handle this depth image? */
    if ((depth != 4) && (depth != 8))
	return ia_depthNotSupported;

    if (ABS(rotAngle) > cMaxRotAngle)
	return ia_angleTooLarge;

	/* Are the resolutions reasonable? */
    if ((xRes != yRes) && ((xRes == 0) || (yRes == 0)))
	return ia_invalidParm;

	/* get the width and height of the source without the frame */
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + depth*cFrameBytes;
    pD = pS;				/* initial dest is start of source. */

	/* if the rotation angle is 0, we're done */
    if (rotAngle == 0)
    {
	NORMAL_RETURN;
    }

	/* Remember address of upper right corner of dest image
	 * so we can clear its right frame when done. */
    pDOrig = pD;

	/* allocate the ring buffer to be used for destination lines.
	 * The buffer is sBpl bytes wide and maxVertDisp + extraBufLines
	 * lines tall.  maxVertDisp equals the number of source lines
	 * a diagonal path across the source will cross.  When we're
	 * maxVertDisp lines below the top of the source, we shouldn't
	 * be touching the top line anymore.  extraBufLines allows
	 * for the fact that we'll proceed down the source image slower
	 * than we'll proceed down the dest image (because cos(rotAngle)
	 * is less than 1).  We approximate the amount of slop needed
	 * by hS / 32.  This should be more than enough for a 10 degree
	 * rotation.  Note that maxVertDisp
	 * is a real number of lines, not a ScaledInt1616.  We use it to
	 * set the size of the ring buffer and fill it.  We don't compare
	 * it to another ScaledInt1616 as is done in
	 * i_rotateNoInterpolationFixedPoint.  sinRotAngle
	 * should always be positive.  
	 * Note the way we multply sinRotAngle by wS.  When does
	 * this overflow?  We are concerned because the result of fract_sin
	 * is the real sine multiplied by 32K.  Well, 0xFFFFFFFF/(32K * sin(10)
	 * is about 754,000.  So this works so long as wS < 754,000.  I think
	 * we're safe.
	 */
    ERR_PASS( fract_sin(pIPState, rotAngle, (ScaledInt1715 *) &sinRotAngle));
    sinRotAngle =  sinRotAngle << 1;	/* convert to ScaledInt1616 */
    sinRotAngle = ABS(sinRotAngle);
    if (ABS(xRes - yRes) > 1)
    {
	vertSinRotAngle = (xRes * sinRotAngle)/yRes;
	horizSinRotAngle = (yRes * sinRotAngle)/xRes;
    }
    else
    {
	vertSinRotAngle = sinRotAngle;
	horizSinRotAngle = sinRotAngle;
    }
    maxVertDisp = ScaledInt1616_TO_Int32(wS * horizSinRotAngle);
    extraBufLines = hS / 32;
    ringBufLines = maxVertDisp + extraBufLines;

    pRing = (UInt8 *) MALLOC(sBpl * ringBufLines);
    pLine = pRing;			/* point to start of buffer */
    absSBpl = sBpl;

	/* Make the rotation tables if they don't exist */
    if (depth == 4)
    {
	ERR_PASS( ip_make4bppGrayRotationTable (pIPState,
						c4bppGrayRotTableSize,
						&p4bppGrayRotationTable));
    }
    else
    {
	ERR_PASS( ip_makeGrayRotationTable (pIPState,
					    c8bppGrayRotTableSize,
					    &pGrayRotationTable));
    }

	/* We'd like to use one set of code for both clockwise and counter
	 * clockwise rotation.  If sinRotAngle < 0, we're doing counter
	 * clockwise rotation so should process the image starting from the
	 * lower left corner instead of the upper left corner.  Reposition
	 * pS and pD.  Invert sBpl.
	 */
    if (rotAngle < 0)
    {
	pS += (hS-1) * sBpl;	/* point to lower right corner */
	pD = pS;
	sBpl = -sBpl;
	rotAngle = -rotAngle;
    }

	/* compute the cosine term we'll need to relate the
	 * source and dest coordinate systems. */
    ERR_PASS( fract_cos(pIPState, rotAngle, (ScaledInt1715 *) &cosRotAngle));
    cosRotAngle = cosRotAngle << 1;

	/* Ron Danisewicv at XIS found a much simpler way to derive the
	 * coordinates of the first input pixel.  You just walk from the
	 * center to the upper left corner, remembering to account for
	 * the changes to the y coordinate for each horizontal move
	 * and the changes to the x coordinate for each vertical move.
	 * Let's do the horizontal moves first:
	 *	old input coords:	(wS/2, hS/2)
	 *	input x coord after moving left by wS/2 in the output:
	 *				(wS/2 - (wS/2) * cos(rotAngle))
	 *				= ((wS/2) * (1-cos(rotAngle));
	 *	input y coord after moving left by wS/2 in the output:
	 *				(hS/2 + (wS/2) * sin(rotAngle));
	 *
	 * Now move up by hS/2 in the output and trace where the input goes.
	 *	current input coord:	((wS/2) * (1-cos(rotAngle)),
	 *				 hS/2 + (wS/2) * sin(rotAngle))
	 *	final x coord after moving up by hS/2
	 *				(wS/2) * (1-cos(rotAngle)) +
	 *				  (hS/2) * -sin(rotAngle)
	 *	final y coord after moving up by hS/2
	 *				hS/2 + (wS/2) * sin(rotAngle) -
	 *					hS/2 * cos(rotAngle)
	 *				= (hS/2) * (1-cos(rotAngle)) +
	 *				   (wS/2) * sin(rotAngle)
	 *
	 * This works great for square resolutions.  What if xRes != yRes?
	 * In this case, we use the versions of sinRotAngle that account for
	 * the shears involved, namely vertSinRotAngle (used to modify X
	 * coords when moving vertically) and horizSinRotAngle (used to
	 * modify Y coords when moving horizontally).  Since sinRotAngle =
	 * vertSinRotAngle = horizSinRotAngle when xRes == yRes, we can
	 * use vertSinRotAngle and horizSinRotAngle all the time.
	 */
    xICoord = (wS/2) * (0x10000 - cosRotAngle) -
	      ((hS/2) * vertSinRotAngle);
    xICoord = ScaledInt1616_TO_Int32(xICoord);

    yICoord = ((hS/2) * (0x10000 - cosRotAngle)) +
	      ((wS/2) * horizSinRotAngle);
    yICoord = ScaledInt1616_TO_Int32(yICoord);

	/* This is the input coordinate which we decree corresponds exactly
	 * to coordinate (0,0) in the output frame of reference.  What do we
	 * do with it?  First, we adjust the input address to point to this
	 * place.  This will almost certainly point to something outside
	 * the input image, but that's ok.  We won't do any memory references
	 * with this address until we enter the input image.
	 */
    if (depth == 4)
        pS += yICoord * sBpl + (xICoord >> 1);
    else
	pS += yICoord * sBpl + xICoord;

	/* By decree, the location of the this input pixel relative to the
	 * output image is (-1.0, -1.0)
	 */
    xIOCoord = (-1<<16);
    yIOCoord = (-1<<16);

	/* ************* FILL RING BUFFER ******************** */

	/* Have ip_rotateGrayLine actually do the work on each line.
	 * Start by just filling the ring Buffer */
    for (line = 0, pLine = pRing;
	 (line < hS) && (line < ringBufLines);
	 line++, pLine += absSBpl)
    {
	    /* At this point, xIOCoord and yIOCoord are the output coordinates
	     * of an input pixel whose upper (or lower if counter clockwise)
	     * left corner is in the row to be rotated.  If possible, the
	     * pixel is also in column -1.  If that pixel can't be reached,
	     * our first pixel is in column 0.
	     */

	    /* Calculate the distance to the first horizontal and vertical
	     * edges that we'll cross.  Lower level code does an unsigned
	     * comparison of xCount >= wS to see if we're off the
	     * image horizontally.  This comparison returns true if xCount
	     * is positive or to the right of the image.  Similarly, the
	     * lower level code does an unsigned check for yCount >= hS.
	     */
	xCount = xICoord;
	yCount = yICoord;

	if (depth == 4)
	    ip_rotate4bppGrayLine (pIPState, pS, sBpl, pLine,
				   wS, vertSinRotAngle,
				   horizSinRotAngle,
		                   xIOCoord, yIOCoord, xCount, yCount,
			           xICoord&1, p4bppGrayRotationTable, hS);
	else 
	    ip_rotateGrayLine (pIPState, pS, sBpl, pLine, wS, vertSinRotAngle,
			       horizSinRotAngle,
		               xIOCoord, yIOCoord, xCount, yCount,
			       pGrayRotationTable, hS);

	    /* We would like to adjust the position of the first pixel in
	     * the input row so that it is in the next output row
	     * and col -1.  Sometimes, we can't get there from here
	     * because moving down 1 column gets us to col -2.  If that
	     * happens, we have to think about what to do next.  Since we're
	     * always apparently rotating clockwise (remember we reversed
	     * a bunch of polarities above), moving down moves to the left
	     * in the output space.  If moving down takes us into column -2,
	     * we try moving to the right and moving diagonally.  If either of
	     * those gets us to [row, col] = [-1, -1] or [0, -1] move there.
	     * If neither of those lands in the right location, stay here.
	     * We'll start the next sweep in row -1.
	     * What happens if we show up here having done nothing in the
	     * previous pass?  If we did nothing the previous pass, we jump
	     * to the diagonal now.
	     *
	     * But, that's not all.  A single step down may not be far enough
	     * to reach the next line.  If not, it may or may not cross out
	     * of column -1.  This adds two more possible input points to the
	     * search.  We also have to look at the point 2 rows down and the
	     * one 2 rows down and 1 col to the right.
	     *
	     * Since we may be introducing a horizontal shear at this point,
	     * use vertSinRotAngle instead of sinRotAngle to move sideways.
	     */
	yIOCoord -= (1<<16);		/* we're above the new line */
		
	    /* look straight down.  If we're in row -1, we might as
	     * well stay here since we're right next to the row 0 boundary. */
	if (((yIOCoord + cosRotAngle) >> 16 == -1) &&
	    ((xIOCoord - vertSinRotAngle) >> 16 > -2))	/* the common case */
	{
	    yIOCoord += cosRotAngle;
	    xIOCoord -= vertSinRotAngle;
	    yICoord += 1;
	    pS += sBpl;
	}
	else
	    /* look to the right */
	if (((yIOCoord + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += horizSinRotAngle;
	    xIOCoord += cosRotAngle;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += 1;	    
	    }
	    else pS += 1;
	}
	else
	    /* look straight down and to the right */
	if (((yIOCoord + cosRotAngle + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord - vertSinRotAngle + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += cosRotAngle + horizSinRotAngle;
	    xIOCoord += -vertSinRotAngle + cosRotAngle;
	    yICoord += 1;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += sBpl + 1;
		else pS += sBpl;	    
	    }
	    else pS += sBpl + 1;
	}
	else
	    /* look down 2 */
	if (((yIOCoord + 2*cosRotAngle) >> 16 == -1) &&
	    ((xIOCoord - 2*vertSinRotAngle) >> 16 > -2))
	{
	    yIOCoord += 2*cosRotAngle;
	    xIOCoord -= 2*vertSinRotAngle;
	    yICoord += 2;
	    pS += 2*sBpl;
	}
	else
	    /* look down 2 and to the right */
	if (((yIOCoord + 2*cosRotAngle + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord - 2*vertSinRotAngle + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += 2*cosRotAngle + horizSinRotAngle;
	    xIOCoord += -2*vertSinRotAngle + cosRotAngle;
	    yICoord += 2;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += 2*sBpl + 1;
		else pS += 2*sBpl;	    
	    }
	    else pS += 2*sBpl + 1;
	}
	else
	{
	        /* Couldn't find the next starting point, so quit. */
	    if (depth == 4)
		FREE(p4bppGrayRotationTable);
	    else 
		FREE(pGrayRotationTable);
	    FREE(pRing);
	    return ia_nopoint;
	}

	if (line & 0xFF)
	    IP_YIELD(errorFinish)

    } /* end of loop that initially fills the ring buffer */

	/* ******* TRANSFER FROM RING BUFFER TO SOURCE/DEST & ROTATE ***** */

	/* Have ip_rotateGrayLine actually do the work on each line.
	 * Before rotating each line into the ring buffer, move the
	 * previously rotated contents from the ring buffer to the
	 * source/dest image.  At the entry to the loop, pRing points
	 * one line below the ring buffer as it would if we wrapped
	 * were about to wrap around (which is the very first
	 * thing we'll do!). */
    for (ringLine = ringBufLines;
	 line < hS;
	 line++, ringLine++, pLine += absSBpl)
    {

	    /* If we have wrapped around, adjust ringLine and pLine. */
	if (ringLine == ringBufLines)
	{
	    pLine = pRing;		/* point to top of buffer */
	    ringLine = 0;
	}

	    /* Transfer the previously rotated line in this position in
	     * the ring buffer to the dest image location pointed to by
	     * pD.  Round the transfer length up to the next word boundary
	     * to avoid byte order problems in little endian platforms.
	     */
	if (depth == 4)	/* 8 pix/word */
	    memcpy((void *)pD, (void *)pLine, ((wS+7)/8) * 4);
	else		/* 4 pix/word */
	    memcpy((void *)pD, (void *)pLine, ((wS+3)/4) * 4);
	pD += sBpl;			/* pt to next line in source/dest
					 * image */

	    /* At this point, xIOCoord and yIOCoord are the output coordinates
	     * of an input pixel whose upper (or lower if counter clockwise)
	     * left corner is in the row to be rotated.  If possible, the
	     * pixel is also in column -1.  If that pixel can't be reached,
	     * our first pixel is in column 0.
	     */

	    /* Calculate the distance to the first horizontal and vertical
	     * edges that we'll cross.  Lower level code does an unsigned
	     * comparison of xCount >= wS to see if we're off the
	     * image horizontally.  This comparison returns true if xCount
	     * is positive or to the right of the image.  Similarly, the
	     * lower level code does an unsigned check for yCount >= hS.
	     */
	xCount = xICoord;
	yCount = yICoord;

	if (depth == 4)
	    ip_rotate4bppGrayLine (pIPState, pS, sBpl, pLine,
				   wS, vertSinRotAngle,
				   horizSinRotAngle,
		                   xIOCoord, yIOCoord, xCount, yCount,
			           xICoord&1, p4bppGrayRotationTable, hS);
	else 
	    ip_rotateGrayLine (pIPState, pS, sBpl, pLine, wS, vertSinRotAngle,
			       horizSinRotAngle,
		               xIOCoord, yIOCoord, xCount, yCount,
			       pGrayRotationTable, hS);

	    /* We would like to adjust the position of the first pixel in
	     * the input row so that it is in the next output row
	     * and col -1.  Sometimes, we can't get there from here
	     * because moving down 1 column gets us to col -2.  If that
	     * happens, we have to think about what to do next.  Since we're
	     * always apparently rotating clockwise (remember we reversed
	     * a bunch of polarities above), moving down moves to the left
	     * in the output space.  If moving down takes us into column -2,
	     * we try moving to the right and moving diagonally.  If either of
	     * those gets us to [row, col] = [-1, -1] or [0, -1] move there.
	     * If neither of those lands in the right location, stay here.
	     * We'll start the next sweep in row -1.
	     * What happens if we show up here having done nothing in the
	     * previous pass?  If we did nothing the previous pass, we jump
	     * to the diagonal now.
	     *
	     * But, that's not all.  A single step down may not be far enough
	     * to reach the next line.  If not, it may or may not cross out
	     * of column -1.  This adds two more possible input points to the
	     * search.  We also have to look at the point 2 rows down and the
	     * one 2 rows down and 1 col to the right.
	     */
	yIOCoord -= (1<<16);		/* we're above the new line */
		
	    /* look straight down.  If we're in row -1, we might as
	     * well stay here since we're right next to the row 0 boundary. */
	if (((yIOCoord + cosRotAngle) >> 16 == -1) &&
	    ((xIOCoord - vertSinRotAngle) >> 16 > -2))	/* the common case */
	{
	    yIOCoord += cosRotAngle;
	    xIOCoord -= vertSinRotAngle;
	    yICoord += 1;
	    pS += sBpl;
	}
	else
	    /* look to the right */
	if (((yIOCoord + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += horizSinRotAngle;
	    xIOCoord += cosRotAngle;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += 1;	    
	    }
	    else pS += 1;
	}
	else
	    /* look straight down and to the right */
	if (((yIOCoord + cosRotAngle + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord - vertSinRotAngle + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += cosRotAngle + horizSinRotAngle;
	    xIOCoord += -vertSinRotAngle + cosRotAngle;
	    yICoord += 1;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += sBpl + 1;
		else pS += sBpl;	    
	    }
	    else pS += sBpl + 1;
	}
	else
	    /* look down 2 */
	if (((yIOCoord + 2*cosRotAngle) >> 16 == -1) &&
	    ((xIOCoord - 2*vertSinRotAngle) >> 16 > -2))
	{
	    yIOCoord += 2*cosRotAngle;
	    xIOCoord -= 2*vertSinRotAngle;
	    yICoord += 2;
	    pS += 2*sBpl;
	}
	else
	    /* look down 2 and to the right */
	if (((yIOCoord + 2*cosRotAngle + horizSinRotAngle) >> 16 == -1) &&
	    ((xIOCoord - 2*vertSinRotAngle + cosRotAngle) >> 16 > -2))
	{
	    yIOCoord += 2*cosRotAngle + horizSinRotAngle;
	    xIOCoord += -2*vertSinRotAngle + cosRotAngle;
	    yICoord += 2;
	    xICoord += 1;
	    if (depth == 4)
	    {
                    /* only need to bump the source pointer if we moved to
                     * new byte.  we could have just moved over in the same
                     * byte.
                     */
                if ((xICoord & 1) == 0)
                    pS += 2*sBpl + 1;
		else pS += 2*sBpl;	    
	    }
	    else pS += 2*sBpl + 1;
	}
	else
	{
		/* Couldn't find starting point for next scan line, so quit. */
	    if (depth == 4)
		FREE(p4bppGrayRotationTable);
	    else 
		FREE(pGrayRotationTable);
	    FREE(pRing);
	    return ia_nopoint;
	}

	if (line & 0xFF)
	    IP_YIELD(errorFinish)

    } /* end of loop that both rotates and transfer rotated data from
       * the ring buffer to the source/dest image. */

	/* At this point, the ring buffer is full of rotated lines.
	 * Flush it to the image. */
    for (line = MIN(ringBufLines, hS);
	 line > 0;
	 line--, ringLine++, pLine += absSBpl)
    {
	    /* If we have wrapped around, adjust ringLine and pLine. */
	if (ringLine == ringBufLines)
	{
	    pLine = pRing;		/* point to top of buffer */
	    ringLine = 0;
	}

	    /* Transfer the previously rotated line in this position in
	     * the ring buffer to the dest image location pointed to by
	     * pD.  Note that we have to round the transfer up to the
	     * next word boundary to account for byte order differences
	     * between little endian and big endian machines.  The overflow
	     * will be cleared when the frame is cleared.
	     */
	if (depth == 4)	/* 8 pix/word */
	    memcpy((void *)pD, (void *)pLine, ((wS+7)/8) * 4);
	else		/* 4 pix/word */
	    memcpy((void *)pD, (void *)pLine, ((wS+3)/4) * 4);

	pD += sBpl;			/* pt to next line in source/dest
					 * image */

    }

	/* clear the right frame of the rotated image.  It was probably
	 * written mistakenly by the rotation code. */
    ip_clearRightFrame (pIPState, pDOrig, wS, hS, absSBpl, depth);

    if (depth == 4)
	FREE(p4bppGrayRotationTable);
    else 
	FREE(pGrayRotationTable);
    FREE(pRing);
    NORMAL_RETURN;

errorFinish:
    if (depth == 4)
	FREE(p4bppGrayRotationTable);
    else 
	FREE(pGrayRotationTable);
    FREE(pRing);
    return ia_aborted;
}

/*

$Log: i_gsrfx1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.7  1998/01/13 23:48:14  ddavies
i_gsrfx1.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1997/05/07 22:20:58  ddavies
ddavies on Wed May 7 15:20:16 PDT 1997

i_gsrfx1.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx1.c,v

Check for rotate angles of 0.  In this case, just copy source image or
leave source image alone (in place rotation).

Revision 1.5  1996/10/05 02:13:39  ddavies
i_gsrfx1.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx1.c,v

Largely rework the way xCount and yCount are calculated and used.  They now
just reflect xICoord and yICoord.  The lower level code does unsigned compares
for xCount >= width and yCount >= height instead of depending on them to be
negative on both sides of the image.  Performing the negative trick can't be
done on very narrow images since there's no chance to reverse the polarity
of the counts.
Also, in the in-place rotation, copy 32-bit words, not bytes, from the rotation
buffer to the image.  This gets around byte-order bugs that have apparently
been there forever.  These would result in trash on the right side of an image.

Revision 1.4  1996/02/10 01:26:04  dferrell
dferrell on Fri Feb 9 17:25:15 PST 1996

i_gsrfx1.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx1.c,v

Added extra parens around & operations so that gcc 2.7.2 won't complain.

Revision 1.3  1995/12/13 02:44:20  ejaquith
i_gsrfx1.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/12/13  01:33:06  ddavies
 * i_gsrfx1.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx1.c,v
 *
 * Rework the way anisotropic images are rotated.  In the old code, we assumed that
 * the angle passed in was calculated in a naive way and was really equal to
 * (real angle) * (yResolution / xResolution).  That's because if the y resolution
 * was small, the image would be squished vertically so all vertical displacements
 * would look small to the skew detect code.  The old code multipled the incoming
 * angle by (xResolution/yResolution) before use (sort of).  Ok, what it really
 * did was use two versions of sin(rotAngle): vertSinRotAngle = (xRes*xRes)/(yRes*yRes)
 * times sinRotAngle and just plain sinRotAngle.
 * In the new code, we assume that the angle is resolution independent so we
 * can get angles from anywhere.  The cosine of the rotation angle is not really
 * affected by this since the angles are so small, so we don't bother with them.
 * Ok, I really meant that we use one version of cosRotAngle.  However, we
 * still use two versions of sin(rotAngle): vertSinRotAngle = (xRes/yRes) *
 * sinRotAngle and horizSinRotAngle = (yRes/xRes) * sinRotAngle.  We use
 * vertSinRotAngle in EVERY instance where vertical motion requires adjusting
 * the x coordinate.  We use horizSinRotAngle in EVERY instanced where horizontal
 * motion requires adjusting the y coordinate.  These are used when calculating
 * the initial input position, on every call to the lower level rotation routines
 * and when adjusting the input position in preparation for rotating the next
 * scan line.  Actually, by making things more consistent, it's easier to
 * figure how what's going on.
 * These changes apply to both the 8 bit/pixel and 4 bit/pixel rotation routines.
 *
 * Revision 1.1  1995/11/06  18:16:43  ddavies
 * Initial revision
 *
 * Revision 1.13  1995/10/18  03:53:26  ddavies
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.12  1995/02/14  01:37:30  dferrell
 * dferrell on Mon Feb 13 17:35:47 PST 1995
 *
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Added typecaste in comparison between signed & unsigned values so
 * that the evaluation was signed.  This caused the 0th output row to
 * be incorrect at 0 degrees of rotation on some platforms.
 *
 * Revision 1.11  1995/02/03  19:45:39  ddavies
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Change xRes and yRes from UInt16's to Int16's to keep acc happy.
 *
 * Revision 1.10  1995/02/03  01:20:30  ddavies
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Add xRes and yRes args to small angle rotations.  These are used to apply
 * a horizontal shear when rotating anisotropic images.  The shear is needed
 * to account for the shear naturally present in such images.
 * Since this entailed changing the calculation of the coordinates of the
 * point in the input image that corresponds to (0,0) in the output image,
 * I also changed the method used to do the calculation from the bizarre
 * one using arctan's and right triangles to one that just steps from the
 * center of the output image to the upper left corner and applies the
 * appropriate transforms to the input coordinates when taking the corresponding
 * steps in the input image.  This is how the upper left corner of the input
 * image is calculated in the non-interpolated rotations.  The result is that
 * the new output images are shifted by a pixel or so relative to the old
 * ones - not a big deal.
 *
 * Revision 1.9  1994/11/07  22:05:28  ddavies
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.8  1994/08/05  21:17:45  ddavies
 * ddavies on Fri Aug 5 14:17:22 PDT 1994
 *
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Reduce the frequency of IP_YIELD calls by 256x.
 *
 * Revision 1.7  1994/07/27  16:01:19  ejaquith
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Made functions use local pointers to gray rotation tables instead of
 * globals.
 *
 * Revision 1.6  1994/07/11  23:10:45  ddavies
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Change bytes/line from unsigned ints to int.  These vars are set to negative
 * numbers inside the routine when doing counter-clockwise rotations.  The symptom
 * of this bug is that we were clearing random words above the image in memory
 * after a rotate when attempting to clear the right frame.  Ouch.
 *
 * Revision 1.5  1994/03/02  03:50:47  ddavies
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Add IP_YIELD to give up control during gray rotates.
 *
 * Revision 1.4  1994/02/04  20:15:28  dferrell
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:23:08  ddavies
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:28:18  ddavies
 * i_gsrfxp.c -> /product/ipcore/ipshared/src/RCS/i_gsrfxp.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

