/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_gsrfx2.c: Gray Scale Rotation, FiXed point interface #2
 *
 *          Upper level gray scale rotation routines, SMALL ANGLES ONLY!
 *		(< 10 degrees)
 *	    These routines rotate the image in strips.
 *	    These routine require no floating point support, they use
 *		fixed point calculations only.  They take angles in
 *		ScaledAngle format.
 *
 *          Gray Scale rotation routines:
 *		    Int32	i_getInputLinesForRotatedOutputLinesFixedPoint()
 *		    Int32	i_getSizeOfGrayRotateInfo()
 *                  Int32     	i_rotateGrayFixedPointInit()
 *		    Int32	i_rotateGrayProcess()
 * 
 */


#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrscal.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "fxpmath.pub"
#include "shros.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_gsrfx2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* There is a huge comment in i_gsrfx1.c.  It describes how rotation
 * is really implemented.  Please read it for insight into how we
 * attack small angle rotation.
 *
 * This module addresses basically the same problem, but we allow the
 * client to supply the image in strips.  The first application for this
 * is in deskewing the text plane of a pre-scanned image.  Since people
 * are picky about text, we should scan it at full resolution and do the
 * skew correction with a grayscale, interpolated, small angle rotation.
 * We already have the skew angle from the low resolution pre-scanned image.
 * The full resolution gray image is huge, so it uses too much memory
 * even if we rotate it in place.  The solution is to rotate each strip
 * as it's read from the scanner.  The deskewed strips are then thresholded,
 * but that's irrevelent here.
 *
 * We'll use the same low level rotation routines as were used in the
 * ordinary interpolated rotation.  The new code will be used to decide
 * how to start and stop each strip.
 *
 * Our rotation method requires a lookup table.  It would clearly be
 * ridiculous to recreate this table for each strip.  We can also use
 * some cached state.  Some of the state does not change, like the
 * sine and cosine of the rotation angle.  Some does change, like the
 * number of the next output line and a couple of sets of coordinates.
 * These will be stored as hints, not as real true facts.  This way,
 * we can use them when the strip we're given is immediately below the
 * previous strip, and recalculate them when the new strip either belongs
 * to another image or is out of order in the same image.
 *
 * More on recalculating state here.
 *
 * Time for some software engineering.  The nicest way to acquire the private
 * state block that holds the rotation table and other state is to
 * have the client allocate (and free) it.  This gives the client complete
 * control over memory management and so on.  The DAE has strict restrictions
 * on pointers, so the block will contain none.  A structure containing
 * saved hints will be prepended to the rotation table in one big block.
 *
 * To allow the client to allocate the private state block, we supply a
 * routine called i_getSizeOfGrayRotateInfo.  This routine returns the
 * size of the block to be used.
 *
 * The next routine, i_rotateGrayFixedPointInit fills the already-allocated
 * block with whatever is needed to start off.  This includes at least
 * the rotation table.
 *
 * The main routine is i_rotateGrayProcess.  This routine takes the private
 * state block, pointers to the input and output strips, the numbers of the
 * first lines in each, the dimensions of the strips and so on.  If the
 * strips passed immediately follow the strips passed in the last time
 * this routine was called, the saved hints are used to determine the
 * rotation coordinates.  If not, the hints are recalculated from scratch
 * as discussed above.  In either case, the routine rotates until it either
 * a) runs out of input lines, b) runs out of space for output lines or
 * c) has produced the bottom output line.  This routine
 * returns the number of the input line that should be provided at the
 * beginning of the next buffer and the number of the output line to be
 * put out next.
 *
 * When the rotation operation has been completed (or perhaps aborted), the
 * client simply frees the state block.
 *
 * Another utility routine is provided to help the client estimate the number
 * of input lines that should be provided.  The greater the angle and the
 * wider the image, the more input lines are needed to create a single output.
 * The routine i_getNumInputLines returns the number of input lines to
 * be provided to produce a specified number of output lines.  Note that
 * i_rotateGrayProcess does not expect the client to supply lines that are
 * either above or below the input image even though that region theoretically
 * becomes part of the output image.
 *
 * One of the interesting changes between this version of the rotation code
 * and the version that rotates an entire image in one step is that this
 * version ALWAYS processes the input image from top to bottom.  The
 * standard version processes the image from top to bottom when the rotation
 * angle is positive (doing a clockwise rotation) and from bottom to top
 * when the rotation angle is negative (counterclockwise rotation).
 * That trick was done to simplify things, so they are a bit more complicated
 * here.
 *
 * How so?  Well, the lowest level primitives always move from left to right
 * across the image.  They always add the sine of the rotate angle to
 * the current yIOCoord (the coordinates of the current input point in the
 * output's frame of reference) each time we move to the right in the
 * input.  They then detect whether the new yIOCoord is >= 0.  This indicates
 * the source position has gotten too low in the output space, so we're
 * no longer starting on the source pixel that would be bordering the
 * top edge of the output pixel.  When the new yIOCoord >= 0, we subtract 
 * the source bytes/line from the source address and subtract 1.0 from 
 * the yIOCoord (and move the xIOCoord over).
 *
 * From the sense of these tests and adjustments, we can fill in the table
 * below.
 *			CW/High/Low	CCW/Low		CCW/High
 * sinRotAngle		sin(angle)	-sin(angle)	sin(angle)
 *
 * src bytes/line	> 0		< 0		> 0
 *
 * yIOCoord		-1.x		-1.x		1.x
 *
 * direction in	 	DOWN		UP		DOWN
 * which yIOCoord
 * grows
 *
 * direction in which	to right	N/A		to left
 * X is adjusted when
 * we move down in
 * source image
 *
 * locations of UL	(-1, -1), (0, -1),		(-1, 1), (0, 1),
 * corner of first src	(-1, -2)			(-1, 0)
 * pixel relative to
 * UL corner of
 * first output pixel
 *
 * Do do we handle this?  I had lots of fancy schemes, but the easiest
 * way to understand it is to have two sections of upper level code.
 * The clockwise section looks almost exactly like the original
 * rotation code.  The counterclockwise section inverts the direction of
 * the change to xIOCoord when moving down and changes the tests made
 * to find out if we're in a legal position and inverts sinRotAngle
 * before passing it to the lower level routines and complements yIOCoord
 * before passing it down.
 *
 ************************ SEARCH ORDER REWORK *************************
 *
 * After lots of work, the code finally worked.  It got the same answer
 * for top-to-bottom rotation whether it was done in one slice or in
 * multiple slices.  However, it got a very slightly different answer when
 * rotated from bottom to top, one line at a time.  Why?
 *
 * Well, on some scan lines, there are two possible starting positions.
 * They are usually (always?) in diagonally opposite corners of a pixel.
 * The lower level code attempts to fix this by moving to the leftmost
 * legal coordinate, but it does so with reduced precision.  This results
 * in slightly different values of xIOCoord and yIOCoord in the lower
 * level routines, hence in slightly different pixel values.
 *
 * My attempt to fix this is to revise the search order used to find the
 * input pixel used to start the next output line.  This is the order
 * in which input pixels are inspected to see if they qualify.  As soon
 * as a qualifying pixel is found, we stop.  The current ("old search
 * order" by the time you read this comment) is:
 *
 *	Clockwise Rotation		Counterclockwise Rotation
 *		*					*
 *			2			2
 *
 *	     1						    1
 *		      3				    3
 *
 *	 4						       4
 *		   5					5
 *
 * Where the asterisk is the current position in the input pixel space.
 * The reason the dead reckoning method looks at 5 points instead of 3
 * is because it tries to get to the right location in one step.
 * The new search orders for the dead reckoning method will be:
 *	Clockwise Rotation		Counterclockwise Rotation
 *		*					*
 *			3			3
 *
 *	     1						    1
 *		      4				    4
 *
 *	 2						       2
 *		   5					5
 *
 *
 * The new searching order used by the searching method is:
 *	Clockwise Rotation		Counterclockwise Rotation
 *		*					*
 *			2			2
 *
 *	     1						    1
 *		      3				    3
 *
 * This code need only investigate three points since it's happy to
 * move down in multiple steps.
 *
 * The goal is to try to have the dead reckoning method used when finding
 * the next line in a top-to-bottom slice pick the same pixel as the
 * search method that is used when given an arbitrary output line.  The search
 * method has two main stages.  In the first, it estimates the position of
 * the correct input pixel using arithmetic.  In the second, it adjusts
 * the position to find the real input pixel.
 * In the estimation state, it moves down in Y first, then over in X.
 * When it moves down in Y, it moves to approximately the correct input
 * y coord, which means its y coord in the output space is too small.
 * Then it moves over in X, which means the y coord in output space grows.
 * It approaches from the "uphill" side (from the left in clockwise
 * rotation, from the right in counterclockwise rotation) in the diagrams
 * above.
 * I used to think that it would never overshoot at this step, but I
 * was wrong.  Our first adjustment step is to move up if we originally
 * went too low.  Our second step is to move to the left (clockwise
 * rotation) or right (counterclockwise) if we overshot in X.  Our third
 * step is to carefully move to the correct X position from the "uphill"
 * side.  This entails moving down the image.  If we go to low, we
 * move back up and to the "uphill" side again.  Finally, we carefully
 * move down using the search pattern described above.
 *
 */

#define	cScaledPi	16384		/* 32768 corresponds to 2pi in
					 * scaled angle notation */
#define	cMaxRotAngle	((10 * cScaledPi) / 180) /* corresponds to 10 degrees */

/* PUBLIC */
/**************************************************************************
 *  i_getInputLinesForRotatedOutputLinesFixedPoint()
 *	Input:
 *		width = width of source and dest gray strip.
 *		height = total height of source image.
 *		destHeight = height of dest gray strip.
 *		rotAngle = rotation angle in ScaledAngle units.
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *		pSrcHeight = pointer to number of source lines required
 *			     to produce hD dest lines.
 *
 *	On return:
 *		The number of input lines required to produce the
 *		specified number of output lines is returned.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_invalidParm		xRes != yRes and either xRes
 *					or yRes is 0.
 *
 **************************************************************************/
Int32   CDECL 
i_getInputLinesForRotatedOutputLinesFixedPoint (
    IPState    *pIPState,
    Int32	width,
    Int32	height,
    Int32	destHeight,
    ScaledAngle	rotAngle,
    Int32	xRes,
    Int32	yRes,
    Int32      *pSrcHeight)
{
    PROLOGUE;
    
    ScaledInt1715	horizSinRotAngle, vertSinRotAngle, cosRotAngle;
    Int32		apparentWidth;
    Int32		linesNeeded;

	/* Are the resolutions reasonable? */
    if ((xRes != yRes) && ((xRes == 0) || (yRes == 0)))
	return ia_invalidParm;

	/* The number of input lines needed equals the distance up or down
	 * the slanted path across the input will take.  This is the
	 * width of the image times the sine of the rotation angle.
	 *
	 * Whoops, well, almost.  Say you have a very thin, tall image.
	 * Say you're rotating clockwise.  The input samples for the
	 * first output line may be quite a ways to the left of the
	 * input, and quite a bit below the top of the input.  The
	 * poor unsuspecting client will initially give us the number
	 * of lines calculated by this procedures starting at input
	 * line 0.  That may not extend down to the first pseudo input
	 * line.  Ok, we're really reading off the side of the image,
	 * but it's even harder to tell the client that they need to
	 * give no lines for awhile (while all the input is off the
	 * side of the image) then some lines, then some more lines,
	 * then fewer lines, then no lines for awhile.  So, we just
	 * say that the client should give us enough lines to cover
	 * our needs.
	 *
	 * How many is that?  Well, the total width of x positions we'll
	 * reach is really width * sinRotAngle.  But to account for the
	 * fact that we may be starting too low, we'll add a fake term
	 * to the width to make it appear wider.  The fake term is how
	 * far over the first sample will be, or sinRotAngle * (height/2).
	 * So the total number number of lines we pretend to cross in
	 * a traverse of the image will be
	 * sinRotAngle * (width + sinRotAngle*(height/2)).
	 *
	 * We must add the number of lines the client wants to produce
	 * to this.  What's the weird (cosRotAngle/2) term for?  It's to round
	 * the division instead of truncating.  What we'd do if we
	 * had floating point (and it were portable), is calculate
	 * (Int32)((destHeight/cos(rotAngle)) + 0.5).  This is
	 * equal to destHeight/cosRotAngle + (cosRotAngle/2)/cosRotAngle, or
	 * (destHeight + (cosRotAngle/2)) / cosRotAngle.
	 * Note that as always, we want to scale sinRotAngle before
	 * using it to modify the y coords when moving horizontally.
	 * As usual, we really want to use (yRes/xRes)*sinRotAngle.
	 * Where does the fudge factor come from?
	 * 1. We start reading both the current line and the line below
	 *	(clockwise) or the line above (counterclockwise), so
	 *	we need at least one extra line.
	 * 2. Sometimes, the code that adjusts the initial line position
	 *	moves down (clockwise) or up (counterclockwise) by an
	 *	extra line before starting, so we need one more extra
	 *	line, for a total of 2 so far.
	 * 3. Sometimes, we start right at the top (clockwise) or
	 *	bottom (counterclockwise) edge of a scan line,
	 *	so that if the width*sin(rotAngle) term equals n,
	 *	we really access n+1 lines in the course of our journey
	 *	across the image, so we need yet another line, for a
	 *	total of 3.
	 */

    ERR_PASS( fract_sin(pIPState, rotAngle,
			(ScaledInt1715 *) &horizSinRotAngle));

    if (ABS(xRes - yRes) > 1)	/* they're really different */
    {
	vertSinRotAngle = (xRes * horizSinRotAngle) / yRes;
	horizSinRotAngle = (yRes * horizSinRotAngle) / xRes;
    }
    else
    {
	vertSinRotAngle = horizSinRotAngle;
    }

    ERR_PASS( fract_cos(pIPState, rotAngle,
			(ScaledInt1715 *) &cosRotAngle));

	/* calculate the apparent width of the image.  This is equal
	 * to the real width plus a fudge factor that accounts for the
	 * fact that in very tall, thin images, we might be starting
	 * quite a bit to the side of the image.
	 */
    apparentWidth = width +
		    ABS(ScaledInt1715_TO_Int32((height/2) * vertSinRotAngle));

    linesNeeded =
	ABS(ScaledInt1715_TO_Int32(apparentWidth * horizSinRotAngle)) +
	(Int32_TO_ScaledInt1715(destHeight) + (cosRotAngle/2)) / cosRotAngle +
	3;	/* for good measure */

	/* if this is greater than the source height, chop it down. */
    if (linesNeeded > height)
	linesNeeded = height;

	/* return the value */
    *pSrcHeight = linesNeeded;

    NORMAL_RETURN;
}

/* PUBLIC */
/**************************************************************************
 *  i_getSizeOfGrayRotateInfo()
 *	Input:
 *		depth = depth of gray image.  4 and 8 bpp supported.
 *		pSize = pointer to returned size in bytes.
 *
 *	On return:
 *		The size of the rotation block (in bytes) is stored
 *		into *pSize.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_invalidParm		pSize is NULL
 *		ia_depthNotSupported	depth not equal to 4 or 8
 *
 **************************************************************************/
Int32   CDECL 
i_getSizeOfGrayRotateInfo (
    IPState    *pIPState,
    Int32	depth,
    Int32      *pSize)
{
    PROLOGUE;
    if (pSize == NULL)
	return ia_invalidParm;

    switch (depth)
    {
    case 4:
	*pSize = sizeof(GrayRotateInfo) + c4bppGrayRotTableSize;
	break;
    case 8:
	*pSize = sizeof(GrayRotateInfo) + c8bppGrayRotTableSize;
	break;
    default:
	return ia_depthNotSupported;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_rotateGrayFixedPointInit()
 *	Input:
 *		depth = depth of gray image.  4 and 8 bpp supported.
 *		rotAngle = rotation angle in ScaledAngle format
 *		rotateInfoSize = size of the private structure
 *				 being passed in.  If this is too small,
 *				 we return ia_invalidParm.
 *		pRotateInfo = pointer to existing rotation structure
 *			      that will be initialized in here.
 *		
 *
 *	On return:
 *		The contents of the rotation structure will be correctly
 *		initialized.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_depthNotSupported	wrong number of bits per pixel
 *		ia_angleTooLarge     	abs value of rotation angle is
 *					greater than 10 degrees (0.17 radians)
 *		ia_invalidParm		pRotateInfo is NULL or rotateInfoSize
 *					is incorrect.
 *		ia_invalidParm		xRes != yRes and either xRes
 *					or yRes is 0.
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateGrayFixedPointInit (
    IPState    *pIPState,
    Int32	depth,
    ScaledAngle	rotAngle,
    Int16	xRes,
    Int16	yRes,
    Int32	rotateInfoSize,
    void       *pRotateInfoIn)
{
    PROLOGUE;
    GrayRotateInfo	*pRotateInfo;
    
    UInt16		*locTable;

	/* convert the info pointer to one we can use */
    pRotateInfo = (GrayRotateInfo *) pRotateInfoIn;

	/* can we handle this depth image? */
    if ((depth != 4) && (depth != 8))
	return ia_depthNotSupported;

	/* is the angle too big? */
    if (ABS(rotAngle) > cMaxRotAngle)
	return ia_angleTooLarge;

	/* Are the resolutions reasonable? */
    if ((xRes != yRes) && ((xRes == 0) || (yRes == 0)))
	return ia_invalidParm;

	/* is the pre-allocated structure there and is it the right size? */
    if (pRotateInfo == NULL)
	return ia_invalidParm;
    if (depth == 4)
    {
	if (rotateInfoSize < (Int32)(sizeof(GrayRotateInfo) +
						c4bppGrayRotTableSize))
	    return ia_invalidParm;
    }
    else	/* depth == 8 */
    {
	if (rotateInfoSize < (Int32)(sizeof(GrayRotateInfo) +
						c8bppGrayRotTableSize))
	    return ia_invalidParm;
    }

	/* everything seems to be ok.  Initialize the GrayRotateInfo.
	 */
    ERR_PASS( fract_sin(pIPState, rotAngle,
			(ScaledInt1715 *) &(pRotateInfo->sinRotAngle)));

	/* convert from ScaledInt1715 to ScaledInt1616 */
    pRotateInfo->sinRotAngle <<= 1;

    ERR_PASS( fract_cos(pIPState, rotAngle,
			(ScaledInt1715 *) &(pRotateInfo->cosRotAngle)));

	/* convert from ScaledInt1715 to ScaledInt1616 */
    pRotateInfo->cosRotAngle <<= 1;

	/* calculate the amount that we move over in the image when
	 * we move down one pixel (vertSinRotAngle).
	 * This is different than sinRotAngle
	 * so we can account for a horizontal shear when rotating
	 * anamorphic images (xRes != yRes).  There's also a version
	 * for how far up/down we move in y when moving over 1
	 * pixel in X (horizSinRotAngle).
	 */
    if (xRes != yRes)
    {
	pRotateInfo->vertSinRotAngle =
	    (xRes * pRotateInfo->sinRotAngle)/yRes;
	pRotateInfo->horizSinRotAngle =
	    (yRes * pRotateInfo->sinRotAngle)/xRes;
    }
    else
    {
	pRotateInfo->vertSinRotAngle = pRotateInfo->sinRotAngle;
	pRotateInfo->horizSinRotAngle = pRotateInfo->sinRotAngle;
    }

	/* We will initialize the rest of the variables inside
	 * i_rotateGrayProcess.  This allows us to use the same
	 * block if (for some bizarre reason), we wish to rotate
	 * several images by the same angle or submit the strips
	 * for one image out of order.  The way we tell
	 * i_rotateGrayProcess to do the job is to set the
	 * dimensions to 0.
	 */
    pRotateInfo->width = 0;
    pRotateInfo->height = 0;
    pRotateInfo->depth = depth;
    pRotateInfo->xICoord = 0;
    pRotateInfo->yICoord = 0;
    pRotateInfo->xIOCoord = 0;
    pRotateInfo->yIOCoord = 0;
    pRotateInfo->firstInputLine = 0;
    pRotateInfo->firstOutputLine = 0;
    pRotateInfo->inputsPerOutput = 0;

	/* Now go fill in the rotation table */
    locTable = (UInt16 *)((UInt8 *)pRotateInfo + sizeof(GrayRotateInfo));

    if (depth == 4)
	__error__ = ip_make4bppGrayRotationTable (pIPState, c4bppGrayRotTableSize,
					        &locTable);
    else	/* depth == 8 */
	__error__ = ip_makeGrayRotationTable (pIPState, c8bppGrayRotTableSize,
					    &locTable);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* That's it.  Things have apparently gone well, so say so. */
    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *  i_rotateGrayProcess()
 *	Input:
 *		pGrayRotInfo =	pointer to private data block used for
 *			      	interpolated rotation in strips.  This
 *				block is allocated by the user using
 *				a size obtained from i_getSizeOfGrayRotateInfo
 *				and initialized by i_rotateGrayInit or
 *				i_rotateGrayFixedPointInit.
 *		pS = pointer to strip of source gray image (NOT including frame)
 *		wS = total width in pixels (NOT including frame) of
 *		     both source and dest images.
 *		hS = total image height (NOT including frame) of
 *		     both source and dest images.
 *		sBpl = source bytes/line (should be multiple of 4)
 *		depth = depth of gray image.  4 and 8 bpp supported.
 *		pD = pointer to strip of destination gray image
 *		dBpl = destination bytes/line (should be multiple of 4)
 *		firstInputLine = number of first input line in strip
 *				 passed in.  0 <= firstInputLine < hS.
 *		numInputLines = number of input lines passed in
 *		firstOutputLine = number of first output line expected
 *		numOutputLines = there's room in the output buffer for
 *				 this many output lines.
 *		pNumLinesProduced = pointer to number of lines actually
 *				    rotated.  This number is returned
 *				    by this routine.
 *		pNextInputLine = Pointer to number of the first input line
 *				 expected in next input buffer.  This
 *				 assumes that the output strip immediately
 *				 below the current one will be rotated next.
 *
 *	On return:
 *		This routine produces output lines until a) it runs
 *		out of input lines, or b) it runs out of space for
 *		output lines.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_invalidParm		the depth of the image does
 *					not match the depth of the
 *					image for with the rotation block
 *					was created, OR one of the pointers
 *					was NULL.
 *		ia_nopoint		could not find next point
 *		ia_depthNotSupported	wrong number of bits per pixel
 *		ia_bufferExceeded	not enough input lines provided
 *					for even one output line.
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	Since this routine operates on strips, neither the source
 *	nor the destination are assumed to be framed.  However,
 *	both source and dest scanlines must be padded to word boundaries.
 *
 *	Best performance is achieved when rotating vertically adjacent
 *	strips from the top of one image to the bottom.
 *	If the width or height (but NOT the depth) of the
 *	image has changed, private coordinates are recomputed.
 *	The same is true if strips are provided out of order.
 *	Thus, it is possible to use the same private data block to
 *	rotate several images (having a common depth) by the same
 *	angle or to rotate one image whose strips are out of order.
 *
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateGrayProcess (
    IPState    *pIPState,
    void       *pGrayRotInfoIn,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	depth,
    UInt8      *pD,
    Int32	dBpl,
    Int32	firstInputLine,
    Int32	numInputLines,
    Int32	firstOutputLine,
    Int32	numOutputLines,
    Int32      *pNumLinesProduced,
    Int32      *pNextInputLine)
{
    PROLOGUE;
    GrayRotateInfo     *pGrayRotInfo;	    /* useful ptr to private data */
    Int32		line;		    /* interation variables */
    Int32	 	xICoord, yICoord;   /* integers containing the
					     * X and Y coords of the current
					     * input pixel in the input
					     * image's frame of reference.
					     */
    ScaledInt1616	xIOCoord, yIOCoord; /* split integers containing the
					     * X and Y coords of the current
					     * input pixel in the output image's
					     * frame of reference.
					     */
    ScaledInt1616	sinRotAngle, cosRotAngle; /* split integers containing
					     * the sign and cosine of the
					     * rotation angle respectively.
					     * These have
					     * 16 bit integer and frac parts */
    ScaledInt1616	vertSinRotAngle;    /* The sine of the rotation angle
					     * that's been modified to account
					     * for shears in anisotropic images.
					     * This is used to modify the X
					     * coord when moving in Y. */
    ScaledInt1616	horizSinRotAngle;   /* The sine of the rotation angle
					     * that's been modified to account
					     * for shears in anisotropic images.
					     * This is used to modify the Y
					     * coord when moving in X. */
    Int32		xCount, yCount;	    /* distances to vertical and horiz
					     * edges passed to
					     * ip_rotateGrayLine. These are
					     * equivalent to input coords
					     * relative to input image. */
    Int32		inputsPerOutput;    /* number of input lines required
					     * to form one output line. */
    Int32		lastInputLine;	    /* the lowest value for yICoord
					     * at which a line can be rotated
					     */
    UInt8	       *pDOrig;		    /* used to remember addr of
					     * upper left corner of image so
					     * we can clear its right frame
					     * when done. */
    UInt16	       *pRotationTable;	    /* ptr to rotation table. */

#define cXCoordMask	0xF800		/* selects x bits for table
					 * index formation */
#define	cYCoordMask	0x700		/* selects y bits for
					 * table index formation */
#define	cYAndPixMask	0x7FF		/* selects all bits we want
					 * to be unaffected when X
					 * bits are inserted into table
					 * lookup.  This is a SPARC
					 * constant, so maybe the
					 * compiler will be smart enough
					 * to do an andn with this */
	/* can we handle this depth image? */
    if ((depth != 4) && (depth != 8))
	return ia_depthNotSupported;

    if (pGrayRotInfoIn == NULL) 
	return ia_invalidParm;

	/* the width, height and image pointers are already frameless,
	 * so they don't need adjusting.
	 * Remember address of upper right corner of dest image
	 * so we can clear its right frame when done. */
    pDOrig = pD;

	/* load local pointers to all the stuff in the private data
	 * block.
	 */
    pGrayRotInfo = (GrayRotateInfo *)pGrayRotInfoIn;
    if (depth != pGrayRotInfo->depth)
	return ia_invalidParm;

	/* load the values that are only a function of rotation angle
	 * and/or depth.  These cannot be recalculated here.
	 */
    sinRotAngle = pGrayRotInfo->sinRotAngle;
    cosRotAngle = pGrayRotInfo->cosRotAngle;
    vertSinRotAngle = pGrayRotInfo->vertSinRotAngle;
    horizSinRotAngle = pGrayRotInfo->horizSinRotAngle;
    pRotationTable = (UInt16 *)((UInt8 *)pGrayRotInfo +
				sizeof(GrayRotateInfo));

	/* Now see if we have to recalculate the remaining
	 * internal coordinates.
	 */
    if ((wS != pGrayRotInfo->width) || (hS != pGrayRotInfo->height) ||
	(firstInputLine != pGrayRotInfo->firstInputLine) ||
	(firstOutputLine != pGrayRotInfo->firstOutputLine))
    {
	    /* we must recalculate the remaining coordinates.
	     * Start with the number of input lines per output line.
	     * Note that every output line requires at least 2 input
	     * lines, hence the "+ 2" term.  And, ooops, we need
	     * to round the wS * horizSinRotAngle UP, not down, so
	     * we add another 1, for a total of + 3.
	     * Ooops.  Sometimes, we can't start in the position we'd
	     * like to, so we're forced to start in the line below
	     * counterclockwise rotation) or above (counterclockwise
	     * rotation).  This adds a potential extra line.  HOWEVER,
	     * When this happens, the Y coord is adjusted down in the
	     * lower level routines, so we don't see it here.  Therefore,
	     * we pretend at this point that this never happens.
	     * There's a subtle trick here.  When rotating clockwise,
	     * we initially need the current input line and the one
	     * BELOW it.  However, as we move across the image, we move
	     * UP across the image.  Therefore, when figuring which
	     * lines we need, we need to look from the height we'll
	     * reach at the rightmost end of the run down to the
	     * line below the initial run.  Similarly, when rotating
	     * counterclockwise, we need to look from the line above
	     * the initial run down to the position we'll reach
	     * at the rightmost end of the run.
	     */
	inputsPerOutput =
	    ScaledInt1616_TO_Int32(ABS((Int32)wS * horizSinRotAngle)) + 3;

	    /* Ron Danisewicv at XIS found a much simpler way to derive the
	     * coordinates of the first input pixel.  You just walk from the
	     * center to the upper left corner, remembering to account for
	     * the changes to the y coordinate for each horizontal move
	     * and the changes to the x coordinate for each vertical move.
	     * Let's do the horizontal moves first:
	     *	old input coords:	(wS/2, hS/2)
	     *	input x coord after moving left by wS/2 in the output:
	     *				(wS/2 - (wS/2) * cos(rotAngle))
	     *				= ((wS/2) * (1-cos(rotAngle));
	     *	input y coord after moving left by wS/2 in the output:
	     *				(hS/2 + (wS/2) * sin(rotAngle));
	     *
	     * Now move up by hS/2 in the output and trace where the input goes.
	     *	current input coord:	((wS/2) * (1-cos(rotAngle)),
	     *				 hS/2 + (wS/2) * sin(rotAngle))
	     *	final x coord after moving up by hS/2
	     *				(wS/2) * (1-cos(rotAngle)) +
	     *				  (hS/2) * -sin(rotAngle)
	     *	final y coord after moving up by hS/2
	     *				hS/2 + (wS/2) * sin(rotAngle) -
	     *					hS/2 * cos(rotAngle)
	     *				= (hS/2) * (1-cos(rotAngle)) +
	     *				   (wS/2) * sin(rotAngle)
	     *
	     * This works great for square resolutions.  What if xRes != yRes?
	     * In this case, we use the versions of sinRotAngle that account for
	     * the shears involved, namely vertSinRotAngle (used to modify X
	     * coords when moving vertically) and horizSinRotAngle (used to
	     * modify Y coords when moving horizontally).  Since sinRotAngle =
	     * vertSinRotAngle = horizSinRotAngle when xRes == yRes, we can
	     * use vertSinRotAngle and horizSinRotAngle all the time.
	     */
	xICoord = (wS/2) * (0x10000 - cosRotAngle) -
		  ((hS/2) * vertSinRotAngle);
	xICoord = ScaledInt1616_TO_Int32(xICoord);
    
	yICoord = ((hS/2) * (0x10000 - cosRotAngle)) +
		  ((wS/2) * horizSinRotAngle);
	yICoord = ScaledInt1616_TO_Int32(yICoord);


	    /* By decree, the location of the this input pixel relative to the
	     * output image is (-1.0, -1.0) if we're rotating clockwise
	     * and (-1.0, 1.0) if we're rotating counterclockwise.
	     * That's because the lower level routines look at the input
	     * pixel below for clockwise rotation and at the input pixel
	     * above for counterclockwise rotation (because we set
	     * sBpl to be negative there).
	     * Well, we really set them to (-0.99999.., -0.9999999) or
	     * (-0.999999..., 0.99999....).  That's because we'll
	     * fine-tune their positions just below this and we don't
	     * want the initial positions to change at the top of
	     * the image.
	     */
	xIOCoord = Int32_TO_ScaledInt1616(-1) + 1;
	if (sinRotAngle >= 0)
	    yIOCoord = Int32_TO_ScaledInt1616(-1) + 1;
	else
	    yIOCoord = Int32_TO_ScaledInt1616(1) - 1;

	    /* Now we take care of the fact that we might not be at
	     * the top of the image.  Why didn't we just adjust the
	     * y coord calculations above?  Because we have to do special
	     * stuff to the xIOCoord and yIOCoord numbers relative to the
	     * mythical coordinates they had at the top of the image
	     * and I'm too stupid to put it all into one ball of wax
	     * and have a hope of figuring it all out again.
	     * This is weird.  We want to figure out how far down to
	     * move in the input space to put us as close as we can
	     * to the place where we want to be in the output space.
	     * What we're really doing is trying to emulate the regular
	     * loop that moves down the image without having to
	     * go through each step of the loop.
	     * Well, we mostly want to move down some number of steps
	     * in the input image.  Call this dY.  As a result of moving
	     * down dY steps in the input image, we will have moved
	     * down dY*cos(rotAngle) steps in the output image AND
	     * over dY*vertSin(rotAngle) steps.  Since the loop tries to
	     * keep us moving straight down in the output image,
	     * it will move back by dY*vertSin(rotAngle), which moves us
	     * down by an additional dY*vertSin(rotAngle) * horizSin(rotAngle)
	     * steps.  The total should be the distance we really want
	     * to move vertically in the output image.
	     * Therefore, the vertical distance we move down in the
	     * input image is
	     * dY = (firstOutputLine) /
	     *	     (cos(rotAngle) + horizSin(rotAngle)*vertSin(rotAngle))
	     */
	yCount = Int32_TO_ScaledInt1616(firstOutputLine) /
		    (cosRotAngle +
		     ((vertSinRotAngle * horizSinRotAngle) / 0x10000));

	yICoord += yCount;

	    /* Having moved down yCount lines in the input, that moved
	     * us over by yCount * sin(rotAngle) columns in the output.
	     * The transversal loop would have compensated by moving
	     * back by yCount * sin(rotAngle).  Note that we divide
	     * instead of shift to produce xCount.  This rounds towards
	     * 0 instead of negative infinity.
	     */
	xCount = (yCount * vertSinRotAngle) / 0x10000;
	xICoord += xCount;

	    /* the division has the effect of getting rid of all of the
	     * fraction bits in the number added to xICoord and yICoord,
	     * which is what we want anyway.
	     * Ok, having moved to this new spot in the input, we must
	     * adjust xIOCoord and yIOCoord by the change.  This is the
	     * trick.  xIOCoord and yIOCoord correspond to the location
	     * of the upper left corner of the input pixel we're really
	     * going to use.  They're not the location of some output
	     * pixel, they're the location of an input pixel in output
	     * coords.
	     */
	xIOCoord += -(yCount * vertSinRotAngle) + (xCount * cosRotAngle);
	if (sinRotAngle >= 0)
	    yIOCoord += (yCount * cosRotAngle) + (xCount * horizSinRotAngle) -
			Int32_TO_ScaledInt1616(MAX(0, firstOutputLine));
	else
	    yIOCoord += (yCount * cosRotAngle) + (xCount * horizSinRotAngle) -
			Int32_TO_ScaledInt1616(MIN(hS-1, firstOutputLine));

	    /* Now, sadly, we might still not be on the input pixel we want.
	     * We use a four step adjustment process, each step getting
	     * more careful.
	     * 1. If we're too low, move up.
	     * 2. If we've overshot in X, move back (entails moving up).
	     * 3. If we're too far to the left (cw) or right (ccw),
	     *	  move right or left as required.  This entails moving
	     *    down so be careful not to go too low.  If this
	     *    happens, move diagonally up and back the way we
	     *    came.  This is always safe.
	     * 4. Finally, the X position is correct.  Move down using
	     *    the search pattern described above.
	     *
	     * Note that the position adjustments are different for
	     * clockwise and counterclockwise rotations, so we use
	     * two complete bodies of code.
	     *
	     * Note also that we do NOT change pS during these updates,
	     * pS is initialized after all this stuff.
	     */
	if (sinRotAngle >= 0)
	{
		/* 1. If we're too low, move up.  If we're already
		 * in the correct row, back out and reenter
		 * properly below.
		 */
	    while (yIOCoord >> 16 >= -1)
	    {
		yIOCoord += -cosRotAngle;
		xIOCoord += vertSinRotAngle;
		yICoord -= 1;
	    }

		/* 2. If we're too far to the right, move left.
		 * We want to land on the leftmost legal point.
		 */
	    while ((xIOCoord - cosRotAngle) >> 16 > -2)
	    {
		yIOCoord -= horizSinRotAngle;
		xIOCoord -= cosRotAngle;
		xICoord -= 1;
	    }

		/* 3. If we're too far to the left, move right.  If
		 * we cross below the Y coord, move up and to
		 * the right.  The loop will move back to the
		 * left if necessary in the next pass. */
	    while (xIOCoord >> 16 <= -2)
	    {
		yIOCoord += horizSinRotAngle;
		xIOCoord += cosRotAngle;
		xICoord += 1;

		    /* Are we too low now?  If so, move up and to
		     * the left.
		     */
		if (yIOCoord >> 16 > -1)
		{
		    yIOCoord += -cosRotAngle - horizSinRotAngle;
		    xIOCoord += vertSinRotAngle - cosRotAngle;
		    yICoord -= 1;
		    xICoord -= 1;
		}
	    }

		/* 4. If we're above the correct y coord, move down */
	    while (yIOCoord >> 16 < -1)
	    {
		    /* look down 1 */
		if (((yIOCoord + cosRotAngle) >> 16 <= -1) &&
		    ((xIOCoord - vertSinRotAngle) >> 16 > -2))
		{
		    yIOCoord += cosRotAngle;
		    xIOCoord -= vertSinRotAngle;
		    yICoord += 1;
		}
		else
		    /* look to the right */
		if (((yIOCoord + horizSinRotAngle) >> 16 <= -1) &&
		    ((xIOCoord + cosRotAngle) >> 16 > -2))
		{
		    yIOCoord += horizSinRotAngle;
		    xIOCoord += cosRotAngle;
		    xICoord += 1;
		}
		else
		    /* look straight down and to the right */
		if (((yIOCoord + cosRotAngle + horizSinRotAngle) >> 16 <= -1) &&
		    ((xIOCoord - vertSinRotAngle + cosRotAngle) >> 16 > -2))
		{
		    yIOCoord += cosRotAngle + horizSinRotAngle;
		    xIOCoord += -vertSinRotAngle + cosRotAngle;
		    yICoord += 1;
		    xICoord += 1;
		}
	    }


		/* Did we get to a legal spot? */
	    if (((yIOCoord >> 16) != -1) || ((xIOCoord >> 16) <= -2) ||
		((xIOCoord >> 16) > 1))
		return ia_nopoint;
	}
	else	/* rotating counter clockwise */
	{
		/* 1. If we're too low, move up  If we're already
		 * in the correct row, back out and reenter
		 * properly below.
		  */
	    while (yIOCoord >> 16 >= 0)
	    {
		yIOCoord += -cosRotAngle;
		xIOCoord += vertSinRotAngle;
		yICoord -= 1;
	    }

		/* 2. If we're too far to the left, move right.
		 * We want to land on the rightmost legal point.
		 */
	    while ((xIOCoord + cosRotAngle) >> 16 < 1)
	    {
		yIOCoord += horizSinRotAngle;
		xIOCoord += cosRotAngle;
		xICoord += 1;
	    }

		/* 3.  See if we're too far to the right.  If
		 * we cross below the Y coord, move up and to
		 * the right.  The loop will move back to the
		 * left if necessary in the next pass.
		 */
	    while (xIOCoord >> 16 >= 1)
	    {
		yIOCoord -= horizSinRotAngle;
		xIOCoord -= cosRotAngle;
		xICoord -= 1;

		    /* Are we too low now?  If so, move up and to
		     * the right.
		     */
		if (yIOCoord >> 16 > 0)
		{
		    yIOCoord += -cosRotAngle + horizSinRotAngle;
		    xIOCoord += vertSinRotAngle + cosRotAngle;
		    yICoord -= 1;
		    xICoord += 1;
		}
	    }

		/* 4. If we're above the correct y coord, move down */
	    while (yIOCoord >> 16 < 0)
	    {
		     /* look straight down */
		if (((yIOCoord + cosRotAngle) >> 16 <= 0) &&
		    ((xIOCoord - vertSinRotAngle) >> 16 < 1))
		{
		    yIOCoord += cosRotAngle;
		    xIOCoord -= vertSinRotAngle;
		    yICoord += 1;
		}
		else
		    /* look to the left */
		if (((yIOCoord - horizSinRotAngle) >> 16 <= 0) &&
		    ((xIOCoord - cosRotAngle) >> 16 < 1))
		{
		    yIOCoord -= horizSinRotAngle;
		    xIOCoord -= cosRotAngle;
		    xICoord -= 1;
		}
		else
		    /* look straight down and to the left */
		if (((yIOCoord + cosRotAngle - horizSinRotAngle) >> 16 <= 0) &&
		    ((xIOCoord - vertSinRotAngle - cosRotAngle) >> 16 < 1))
		{
		    yIOCoord += cosRotAngle - horizSinRotAngle;
		    xIOCoord += -vertSinRotAngle - cosRotAngle;
		    yICoord += 1;
		    xICoord -= 1;
		}
	    }

		/* Did we get to a legal spot? */
	    if (((yIOCoord >> 16) != 0) || ((xIOCoord >> 16) >= 1) ||
	        ((xIOCoord >> 16) <= -2))
		return ia_nopoint;
	}
    } 
    else 	/* the saved values were fine */
    {
	xICoord = pGrayRotInfo->xICoord;
	yICoord = pGrayRotInfo->yICoord;
	xIOCoord = pGrayRotInfo->xIOCoord;
	yIOCoord = pGrayRotInfo->yIOCoord;
	inputsPerOutput = pGrayRotInfo->inputsPerOutput;
    }

	/* This is the input coordinate which we decree corresponds
	 * exactly to coordinate (0,firstOutputLine) in the output
	 * frame of reference.
	 * What do we do with it?  First, we adjust the input address
	 * to point to this place.  This will almost certainly point
	 * to something outside the input strip, but that's ok.  We
	 * won't do any memory references with this address until we
	 * enter the input image.
	 */
    if (depth == 4)
	pS += (yICoord - firstInputLine) * sBpl + (xICoord >> 1);
    else
	pS += (yICoord - firstInputLine) * sBpl + xICoord;

    line = firstOutputLine;	/* init so we know we haven't done anything */

	/* Here's where most of the real work happens */
    if (sinRotAngle >= 0)
    {
	/* Now see if we have enough input lines to produce the
	 * first output line.  If sinRotAngle >= 0, the input scans
	 * initially use lines yICoord and yICoord+1 and will move
	 * UP as they go across the page.  Ooops, except sometimes
	 * they look at lines yICoord+1 and yICoord+2 at the beginning
	 * when they couldn't start where they wanted to. Therefore, 
	 * we must check to see if the left end of the scan (yICoord+2) is
	 * too low or the right end is too high.
	 * There are two cases on each end of each scan line.  If
	 * the end is "high", we check to see if it's above the top
	 * of the image.  If not, we make sure that the line it
	 * reads has been supplied.  If so, we check to see if line 0
	 * has been supplied.  If the end of the line is "low", we check
	 * to see if it's off the bottom of the image.  If not, we
	 * check to see if the line it references is there.  If so, we
	 * check to see if the last image line has been supplied.
	 * Remember that inputsPerOutput counts the line below the
	 * yICoord as well as the line at yICoord and all lines needed
	 * above that.  Therefore, the number of the uppermost input
	 * line we'll access is yICoord - (inputsPerOutput-2).
	 */
	if ((((yICoord + 2) < hS) &&
	     ((yICoord + 2) >= firstInputLine + numInputLines)) ||
	    (((yICoord + 2) >= hS) &&
	     (firstInputLine + numInputLines < hS)) ||
	    (((yICoord-inputsPerOutput+2) >= 0) &&
	     ((yICoord-inputsPerOutput+2) < firstInputLine)) ||
	    (((yICoord-inputsPerOutput+2) < 0) &&
	     (firstInputLine > 0)))
	{
	    goto finishGrayRotProcess;
	}
	else	/* calculate how low we can let yICoord go in this strip. */
	{
	    if (firstInputLine + numInputLines < hS)
		lastInputLine = firstInputLine + numInputLines - 2;
	    else	/* doesn't really matter since we'll just read
			 * white off the bottom of the image anyway.
			 */
		lastInputLine = hS + inputsPerOutput;
	}

	    /* Have ip_rotateGrayLine actually do the work on each line. */
	for (;
	     (line < (firstOutputLine + numOutputLines)) &&
	       (yICoord <= lastInputLine);
	     line++, pD += dBpl)
	{

		/* At this point, xIOCoord and yIOCoord are the output
		 * coordinates of an input pixel whose upper
		 * left corner is in the row to be rotated.  If possible, the
		 * pixel is also in column -1.  If that pixel can't be reached,
		 * our first pixel is in column 0.
		 */
    
		/* Calculate the distance to the first horizontal and vertical
		 * edges that we'll cross.  Lower level code does an unsigned
		 * comparison of xCount >= wS to see if we're off the
		 * image horizontally.  This comparison returns true if xCount
		 * is positive or to the right of the image.  Similarly, the
		 * lower level code does an unsigned check for yCount >= hS.
		 */
	    xCount = xICoord;
	    yCount = yICoord;

	    if (depth == 4)
		ip_rotate4bppGrayLine (pIPState, pS, sBpl,
				       pD, wS, vertSinRotAngle,
				       horizSinRotAngle,
				       xIOCoord, yIOCoord, xCount, yCount,
				       xICoord&1, pRotationTable, hS);
	    else 
		ip_rotateGrayLine (pIPState, pS, sBpl, pD, wS, vertSinRotAngle,
				   horizSinRotAngle,
				   xIOCoord, yIOCoord, xCount, yCount,
				   pRotationTable, hS);
    
		/* We would like to adjust the position of the first pixel in
		 * the input row so that it is above the next output row
		 * (row -1) and col -1 or col 0.  Sometimes, we can't get
		 * there from here because moving down 1 column gets us to
		 * col -2.  If that happens, we have to think about what to
		 * do next.  Since we're rotating clockwise in this case,
		 * moving down moves to the left in the output space.
		 * If moving down takes us into column -2, we try moving
		 * to the right and moving diagonally.  If either of
		 * those gets us to [col, row] = [-1, -1] or [0, -1] move there.
		 * If neither of those lands in the right location, stay here.
		 * We'll start the next sweep in row -1.
		 * What happens if we show up here having done nothing in the
		 * previous pass?  If we did nothing the previous pass, we jump
		 * to the diagonal now.
		 *
		 * But, that's not all.  A single step down may not be
		 * far enough to reach the next line.  If not, it may or
		 * may not cross out of column -1.  This adds two more
		 * possible input points to the search.  We also have to
		 * look at the point 2 rows down and the one 2 rows down
		 * and 1 col to the right.
		 *
		 * Since we may be introducing a horizontal shear at this point,
		 * use vertSinRotAngle instead of sinRotAngle to move sideways
		 * and horizSinRotAngle instead of sinRotAngle to move up/down.
		 */
	    yIOCoord -= (1<<16);		/* we're above the new line */
		    
		/* Try the common case, moving straight down in the input
		 * image.  If we end up in row -1, we might as
		 * well stay here since we're right next to the row 0 boundary.
		 */
	    if (((yIOCoord + cosRotAngle) >> 16 == -1) &&
		((xIOCoord - vertSinRotAngle) >> 16 > -2))
	    {
		yIOCoord += cosRotAngle;
		xIOCoord -= vertSinRotAngle;
		yICoord += 1;
		pS += sBpl;
	    }
	    else
		/* look down 2 */
	    if (((yIOCoord + 2*cosRotAngle) >> 16 == -1) &&
		((xIOCoord - 2*vertSinRotAngle) >> 16 > -2))
	    {
		yIOCoord += 2*cosRotAngle;
		xIOCoord -= 2*vertSinRotAngle;
		yICoord += 2;
		pS += 2*sBpl;
	    }
	    else
		/* look to the right */
	    if (((yIOCoord + horizSinRotAngle) >> 16 == -1) &&
		((xIOCoord + cosRotAngle) >> 16 > -2))
	    {
		yIOCoord += horizSinRotAngle;
		xIOCoord += cosRotAngle;
		xICoord += 1;
		if (depth == 4)
		{
			/* only need to bump the source pointer if we moved to
			 * new byte.  we could have just moved over in the same
			 * byte.
			 */
		    if ((xICoord & 1) == 0)
			pS += 1;	    
		}
		else pS += 1;
	    }
	    else
		/* look straight down and to the right */
	    if (((yIOCoord + cosRotAngle + horizSinRotAngle) >> 16 == -1) &&
		((xIOCoord - vertSinRotAngle + cosRotAngle) >> 16 > -2))
	    {
		yIOCoord += cosRotAngle + horizSinRotAngle;
		xIOCoord += -vertSinRotAngle + cosRotAngle;
		yICoord += 1;
		xICoord += 1;
		if (depth == 4)
		{
			/* only need to bump the source pointer if we moved to
			 * new byte.  we could have just moved over in the same
			 * byte.
			 */
		    if ((xICoord & 1) == 0)
			pS += sBpl + 1;
		    else pS += sBpl;	    
		}
		else pS += sBpl + 1;
	    }
	    else
		/* look down 2 and to the right */
	    if (((yIOCoord + 2*cosRotAngle + horizSinRotAngle) >> 16 == -1) &&
		((xIOCoord - 2*vertSinRotAngle + cosRotAngle) >> 16 > -2))
	    {
		yIOCoord += 2*cosRotAngle + horizSinRotAngle;
		xIOCoord += -2*vertSinRotAngle + cosRotAngle;
		yICoord += 2;
		xICoord += 1;
		if (depth == 4)
		{
			/* only need to bump the source pointer if we moved to
			 * new byte.  we could have just moved over in the same
			 * byte.
			 */
		    if ((xICoord & 1) == 0)
			pS += 2*sBpl + 1;
		    else pS += 2*sBpl;	    
		}
		else pS += 2*sBpl + 1;
	    }
	    else
	    {
		    /* Couldn't find the next point, so give up. */
		return ia_nopoint;
	    }
	}
    }
    else	/* doing counterclockwise rotation */
    {
	/* Now see if we have enough input lines to produce the
	 * first output line.  Since sinRotAngle < 0, the input scans
	 * initially use lines yICoord and yICoord-1 and will move
	 * DOWN as they go across the page.  Ooops, except sometimes
	 * we start with yICoord-1 and yICoord-2 when we can't start
	 * where we want to.  Therefore, we must check to see if the
	 * left end of the scan (yICoord-2) is too high or the right
	 * end is too low.
	 * There are two cases on each end of each scan line.  If
	 * the end is "high", we check to see if it's above the top
	 * of the image.  If not, we make sure that the line it
	 * reads has been supplied.  If so, we check to see if line 0
	 * has been supplied.  If the end of the line is "low", we check
	 * to see if it's off the bottom of the image.  If not, we
	 * check to see if the line it references is there.  If so, we
	 * check to see if the last image line has been supplied.
	 * Remember that inputsPerOutput counts the line above the
	 * yICoord as well as the line at yICoord and all lines needed
	 * below that.  Therefore, the number of the lowest input
	 * line we'll access is yICoord + (inputsPerOutput-2).
	 */
	if ((((yICoord-2) >= 0) && ((yICoord-2) < firstInputLine)) ||
	    (((yICoord-2) < 0) && (firstInputLine > 0)) ||
	    (((yICoord+inputsPerOutput-2) <= (hS-1)) &&
	     ((yICoord+inputsPerOutput-2) > firstInputLine+numInputLines-1)) ||
	    (((yICoord+inputsPerOutput-2) > (hS-1)) &&
	     (firstInputLine + numInputLines < hS)))
	{
	    goto finishGrayRotProcess;
	}
	else	/* calculate how low we can let yICoord go in this strip. */
	{
	    if (firstInputLine + numInputLines < hS)
		lastInputLine = firstInputLine + numInputLines - 1 -
				(inputsPerOutput - 2);
	    else	/* doesn't really matter since we'll just read
			 * white off the bottom of the image anyway.
			 */
		lastInputLine = hS + inputsPerOutput;
	}

	    /* Have ip_rotateGrayLine actually do the work on each line. */
	for (;
	     (line < (firstOutputLine + numOutputLines)) &&
	       (yICoord <= lastInputLine);
	     line++, pD += dBpl)
	{

		/* Calculate the distance to the first horizontal and vertical
		 * edges that we'll cross.  Lower level code does an unsigned
		 * comparison of xCount >= wS to see if we're off the
		 * image horizontally.  This comparison returns true if xCount
		 * is positive or to the right of the image.  Similarly, the
		 * lower level code does an unsigned check for yCount >= hS.
		 * Note that here we also invert the sense of yICoord.
		 * That's because at the top of the image, we start
		 * with a number that's bigger than hS that when
		 * decremented sufficiently becomes less than hS and
		 * therefore signals that we're on the image.
		 * This trick also works at the bottom of the image,
		 * where we want coords that are a little smaller than
		 * hS-1 to be small positive numbers.  As we move across
		 * the image and down, these numbers are decremented
		 * and become negative when we walk off the bottom of
		 * the image.
		 */
	    xCount = xICoord;
	    yCount = (hS-1) - yICoord;

		/* the first weird trick is the way we alter the
		 * arguments to the routines.  We negate the sinRotAngle
		 * since it should always be positive.  We negate
		 * sBpl since we need to slope up as we go across the
		 * image instead of down.
		 * We also negate yIOCoord because the routine thinks
		 * it's getting a number that looks like -1.x which
		 * grow larger as we approach the top of the image.
		 * yIOCoord in this case looks like +1.x and grows larger
		 * as we approach the bottom of the image.
		 */
	    if (depth == 4)
		ip_rotate4bppGrayLine (pIPState, pS, -sBpl,
				       pD, wS, -vertSinRotAngle,
				       -horizSinRotAngle,
				       xIOCoord, -yIOCoord, xCount, yCount,
				       xICoord&1, pRotationTable, hS);
	    else 
		ip_rotateGrayLine (pIPState, pS, -sBpl,
				   pD, wS, -vertSinRotAngle,
				   -horizSinRotAngle,
				   xIOCoord, -yIOCoord, xCount, yCount,
				   pRotationTable, hS);
    
		/* We would like to adjust the position of the first pixel in
		 * the input row so that it is in the next output row
		 * (row 0) and col -1 or col 0.  Sometimes, we can't get
		 * there from here because moving down 1 column gets us to
		 * col 1.  If that happens, we have to think about what to
		 * do next.  Since we're rotating counterclockwise in this case,
		 * moving down moves to the right in the output space.
		 * If moving down takes us into column 1, we try moving
		 * to the left and moving diagonally.  If either of
		 * those gets us to [col, row] = [-1, 0] or [0, 0] move there.
		 * If neither of those lands in the right location, stay here.
		 * We'll start the next sweep in row 0.
		 * What happens if we show up here having done nothing in the
		 * previous pass?  If we did nothing the previous pass, we jump
		 * to the diagonal now.
		 *
		 * But, that's not all.  A single step down may not be
		 * far enough to reach the next line.  If not, it may or
		 * may not cross out of column -1.  This adds two more
		 * possible input points to the search.  We also have to
		 * look at the point 2 rows down and the one 2 rows down
		 * and 1 col to the left.
		 *
		 * Since we may be introducing a horizontal shear at this point,
		 * use vertSinRotAngle instead of sinRotAngle to move sideways
		 * and horizSinRotAngle instead of sinRotAngle to move
		 * up/down.
		 *
		 * AND FINALLY, we are really looking at the position of the
		 * lower left corner of the input pixel (since the y dimension
		 * has been mirrored).  We need this to be between 0.0 and 1.0.
		 * Therefore, when the fraction bits are shifted off, the
		 * integer part should be 0, not 1.  During clockwise rotation,
		 * we wanted the upper left corner to be between -1.0 and 0.0.
		 * There, shifting off the fraction part leaves an integer
		 * part of -1.
		 */
	    yIOCoord -= (1<<16);		/* we're above the new line */
		    
		/* Try the common case, moving straight down in the input
		 * image.  If we end up in row 0, we might as
		 * well stay here since we're right next to the row 1 boundary.
		 */
	    if (((yIOCoord + cosRotAngle) >> 16 == 0) &&
		((xIOCoord - vertSinRotAngle) >> 16 < 1))
	    {
		yIOCoord += cosRotAngle;
		xIOCoord -= vertSinRotAngle;
		yICoord += 1;
		pS += sBpl;
	    }
	    else
		/* look down 2 */
	    if (((yIOCoord + 2*cosRotAngle) >> 16 == 0) &&
		((xIOCoord - 2*vertSinRotAngle) >> 16 < 1))
	    {
		yIOCoord += 2*cosRotAngle;
		xIOCoord -= 2*vertSinRotAngle;
		yICoord += 2;
		pS += 2*sBpl;
	    }
	    else
		/* look to the left */
	    if (((yIOCoord - horizSinRotAngle) >> 16 == 0) &&
		((xIOCoord - cosRotAngle) >> 16 < 1))
	    {
		yIOCoord -= horizSinRotAngle;
		xIOCoord -= cosRotAngle;
		xICoord -= 1;
		if (depth == 4)
		{
			/* only need to bump the source pointer if we moved to
			 * new byte.  we could have just moved over in the same
			 * byte.
			 */
		    if ((xICoord & 1) == 1)
			pS -= 1;	    
		}
		else pS -= 1;
	    }
	    else
		/* look straight down and to the left */
	    if (((yIOCoord + cosRotAngle - horizSinRotAngle) >> 16 == 0) &&
		((xIOCoord - vertSinRotAngle - cosRotAngle) >> 16 < 1))
	    {
		yIOCoord += cosRotAngle - horizSinRotAngle;
		xIOCoord += -vertSinRotAngle - cosRotAngle;
		yICoord += 1;
		xICoord -= 1;
		if (depth == 4)
		{
			/* only need to bump the source pointer if we moved to
			 * new byte.  we could have just moved over in the same
			 * byte.
			 */
		    if ((xICoord & 1) == 1)
			pS += sBpl - 1;
		    else pS += sBpl;	    
		}
		else pS += sBpl - 1;
	    }
	    else
		/* look down 2 and to the left */
	    if (((yIOCoord + 2*cosRotAngle - horizSinRotAngle) >> 16 == 0) &&
		((xIOCoord - 2*vertSinRotAngle - cosRotAngle) >> 16 < 1))
	    {
		yIOCoord += 2*cosRotAngle - horizSinRotAngle;
		xIOCoord += -2*vertSinRotAngle - cosRotAngle;
		yICoord += 2;
		xICoord -= 1;
		if (depth == 4)
		{
			/* only need to bump the source pointer if we moved to
			 * new byte.  we could have just moved over in the same
			 * byte.
			 */
		    if ((xICoord & 1) == 1)
			pS += 2*sBpl - 1;
		    else pS += 2*sBpl;	    
		}
		else pS += 2*sBpl - 1;
	    }
	    else
	    {
		    /* Couldn't find the next point, so give up. */
		return ia_nopoint;
	    }
	}
    }	/* end of counterclockwise rotation */

	/* clear the right frame of the rotated image strip.  It was probably
	 * written mistakenly by the rotation code. */
    if (line > firstOutputLine)
	ip_clearRightFrame (pIPState, pDOrig,
			    wS, line - firstOutputLine, dBpl, depth);

    /* we get here when we're all done.  Update the numbers saved in
     * the private data block and return the number of lines processed
     * and the number of the next input line expected.
     * Also update the private data structure.
     */
finishGrayRotProcess:
    *pNumLinesProduced = line - firstOutputLine;
    pGrayRotInfo->width = wS;
    pGrayRotInfo->height = hS;
    pGrayRotInfo->firstOutputLine = line;
    pGrayRotInfo->inputsPerOutput = inputsPerOutput;
    pGrayRotInfo->xICoord = xICoord;
    pGrayRotInfo->yICoord = yICoord;
    pGrayRotInfo->xIOCoord = xIOCoord;
    pGrayRotInfo->yIOCoord = yIOCoord;
    if (sinRotAngle >= 0)
    {
	    /* since inputsPerOutput counts the line below yICoord
	     * as well as yICoord's line itself, the topmost line
	     * we'll want is yICoord - (inputsPerOutput - 2).
	     */
	*pNextInputLine = MAX(yICoord - (inputsPerOutput - 2), 0);
    }
    else
    {
	    /* Remember that the first two lines that figure into the next
	     * output line are yICoord and yICoord-1, except when they're
	     * yICoord-1 and yICoord-2.  Assume the worst case: yICoord - 2.
	     */
	*pNextInputLine = MAX(yICoord - 2, 0);
    }
    pGrayRotInfo->firstInputLine = *pNextInputLine;

	/* if there wasn't enough input to produce any rotated
	 * lines, the user should notice anyway by checking
	 * *pNumLinesProduced.  When rotating slices out of order,
	 * we recommend that the user call this routine twice for
	 * each slice, once to get the starting input line for
	 * each slice and again to do the actualy rotation.
	 * Therefore, returning 0 lines rotated is perfectly legal.
	 */
    NORMAL_RETURN;
}


/*

 $Log: i_gsrfx2.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 Imported sources

 Revision 1.5  1998/01/13 23:48:18  ddavies
 i_gsrfx2.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx2.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1996/10/05 02:16:21  ddavies
 i_gsrfx2.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx2.c,v

 Change how the number of lines needed for a set of output lines is calculated.
 We now account for the fact that on thin images, the sampling path may be
 quite a ways outside the image.  Therefore, it may be lower than you'd think.
 Therefore, we need more input lines for the process routine to be happy.
 We now add a sin(rotAngle) * height/2 term.  height is a new parmaeter.
 Also largely rework the way xCount and yCount are calculated and used.  They now
 just reflect xICoord and yICoord.  The lower level code does unsigned compares
 for xCount >= width and yCount >= height instead of depending on them to be
 negative on both sides of the image.  Performing the negative trick can't be
 >> done on very narrow images since there's no chance to reverse the polarity
 of the counts.

 Revision 1.3  1995/12/13 20:02:24  ddavies
 ddavies on Wed Dec 13 12:02:15 PST 1995

 i_gsrfx2.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx2.c,v

 Fix bug in which the angle passed to i_rotateGrayFixedPointInit was in
 floating point instead of in ScaledAngle units.

 * Revision 1.2  1995/12/13  03:20:26  ddavies
 * i_gsrfx2.c -> /product/ipcore/ipshared/src/RCS/i_gsrfx2.c,v
 *
 * CommonCheck: Added semicolon after IP_RCSINFO macro.
 *
 * Revision 1.1  1995/12/13  01:11:44  ddavies
 * Initial revision
 *

 */



