/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * i_hc1ms.c:	    Hausdorff binary image comparator
 *  	    	    when a model set is represented as the original binary
 *  	    	    bitmap that is passed in
 *  	    	    (allows almost no new memory allocation)
 *
 *  Public routines:
 *
 *  [none, all routines called from corresponding default routines in
 *   i_hc32ms.c and from search routines in i_hcsrch.c]
 *
 *
 * Extensive comments concerning the Hausdorff comparator method can be found
 * at the top of i_hc32ms.c, the default method for use.
 *
 */


#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"
#include "defines.pub"
#include "memory.pub"
#include "shrip.pub"
#include "bitcount.pub"

/* prototypes */
#include "shrpatt.prv"


IP_RCSINFO(RCSInfo, "$RCSfile: i_hc1ms.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");


/* Needed for masking off model row end bits that overhang the natural end of
 * each row of the imagery when overlapping the two regions.
 * (note how this differs slightly from lmask.prv)
 */
static UInt32 myLeftMask[32] =
{   0xFFFFFFFF, 0x80000000, 0xC0000000, 0xE0000000,
    0xF0000000, 0xF8000000, 0xFC000000, 0xFE000000,
    0xFF000000, 0xFF800000, 0xFFC00000, 0xFFE00000,
    0xFFF00000, 0xFFF80000, 0xFFFC0000, 0xFFFE0000,
    0xFFFF0000, 0xFFFF8000, 0xFFFFC000, 0xFFFFE000,
    0xFFFFF000, 0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00,
    0xFFFFFF00, 0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0,
    0xFFFFFFF0, 0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE };


/* PRIVATE */
/******************************************************************************
 * Count the number of "on" bits in a region
 *
 * Description:
 *  Given a pointer to a word in an image (model or imagery) bitmap, and
 *  some bounds to define a specified region, count the number of "on"
 *  bits in this region.
 *  Shift image right by some amount before actually counting bits.
 *
 * Returns:
 *  The number of "on" bits in the region.
 *****************************************************************************/
static Int32 CDECL 
ip_countOverlapWithShifts (
    IPState    *pIPState,
    Int8	rshift,		   /* The number of bits to right shift the
				    * image before we count */
    UInt32     *pM,		   /* Ptr to first word of image region */
    Int32	heightRegion,	   /* Number of rows in the region */
    Int32	wplRegion,	   /* Number of words per line in the region,
				    * minus one */
    UInt32	rowLastWordMask,   /* A left mask for the last overlapping word
				    * of each image row, to chop off excess
				    * bits of the last row word that aren't
				    * in the region. */
    UInt32	wplM)              /* # words per line, including padding and
				    * frame */
{	     
    PROLOGUE;
    UInt32    word;
    Int32     ovrlpCnt = 0, column;
    Int8      lshift = WORD_SIZE - rshift;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for ( ; heightRegion > 0; pM += wplM, heightRegion--)
    {
	for (column = 0; column < wplRegion; column++)
	{
	    /* Count the number of on bits in the current model word */
	    word = (pM[column] >> rshift) | (pM[column-1] << lshift);
	    ovrlpCnt += bitCount[word & 0xFF];
	    word >>= 8;
	    ovrlpCnt += bitCount[word & 0xFF];
	    word >>= 8;
	    ovrlpCnt += bitCount[word & 0xFF];
	    word >>= 8;
	    ovrlpCnt += bitCount[word];
	}

	/* Count the (special) last overlapping word in the row */
	word = ((pM[column] >> rshift) | (pM[column-1] << lshift))
	    & rowLastWordMask;
	ovrlpCnt += bitCount[word & 0xFF];
	word >>= 8;
	ovrlpCnt += bitCount[word & 0xFF];
	word >>= 8;
	ovrlpCnt += bitCount[word & 0xFF];
	word >>= 8;
	ovrlpCnt += bitCount[word];
    }

    return ovrlpCnt;
}


/* PRIVATE */
/****************************************************************************
 * Return space needed to make model set from image
 *
 * Description:
 *  Given the dimensions of the image to make a model set from, determine
 *  the space that will be needed in an allocator call.
 *  The model set will be filled in i_makeModelSet()
 *
 *  Returns:
 *  ia_successful	invocation done and okay
 *
 * Notes:
 *  blockHeight: (should always be 0 when creating a _B_ model set)
 *   If 0, it means don't order the model rows, and later always sequentially
 *     compare the model row by row. (slowest)
 *   If >= sHeight, it means order all the model rows densest first.
 *   Otherwise, divide the model into blocks of blockHeight rows.  Within each
 *     block, order the rows densest first.  Compare the densest blocks first.
 *     This is to help avoid cache misses.  Thus, blockHeight should be set
 *     with the cache size and the width of the model bitmap AND the imagery
 *     bitmap in mind.
 *****************************************************************************/
Int32 CDECL 
ip_getModelSetSizeOnePtr (
    IPState    *pIPState,
    Int32	sHeight,     /* frame, excluding padding */
    Int32	blockHeight, /* Indicates how to divide up and order the model
			      * bitmap (see Notes above) */
    Int32      *pModelSetSize) /* Ptr to returned value of model set size. */
{
    PROLOGUE;

    *pModelSetSize = (Int32) (sizeof(modelSet) +
			      (blockHeight > 0 ?
				3 * (sHeight - 2*cFrameBits) * sizeof(Int32) :
				0));

    NORMAL_RETURN;
}


/* PRIVATE */
/****************************************************************************
 * Make a model set from an image bitmap
 *
 * Description:
 *  Given an image and its dimensions, make a model set from it.
 *  This model set will be used in the Hausdorff comparator functions.
 *  NOTICE it may be desirable for the image to have the ADDITIONAL
 *  'dilateRadius' zeroed border around the image.
 *  (see note on 'dilateRadius' at top of ipshared/i_hc32ms.c)
 *
 * Returns:
 *  ia_successful	invocation done and okay
 *  ia_nomem    	could not allocate required space
 *  ia_imageTooWide	too big to make a model set with current implementation
 *
 * Notes:
 *  pM:
 *   *pM != NULL, the model set has been allocated by a client and it will
 *   simply get filled in here.
 *  blockHeight: (should always be 0 when creating a _B_ model set)
 *   If 0, it means don't order the model rows, and later always sequentially
 *     compare the model row by row. (slowest)
 *   If >= sHeight, it means order all the model rows densest first.
 *   Otherwise, divide the model into blocks of blockHeight rows.  Within each
 *     block, order the rows densest first.  Compare the densest blocks first.
 *     This is to help avoid cache misses.  Thus, blockHeight should be set
 *     with the cache size and the width of the model bitmap AND the imagery
 *     bitmap in mind.
 *****************************************************************************/
Int32 CDECL 
ip_makeModelSetOnePtr (
    IPState    *pIPState,
    UInt8      *pS,	     /* Ptr to upper left corner of image to make
			      * model set from, including frame */
    Int32	sBpl,	     /* Number of bytes per line, including padding
			      * and frame */
    Int32	sWidth,	     /* Width and height of image, including */
    Int32	sHeight,     /* frame, excluding padding */
    Int32	blockHeight, /* Indicates how to divide up and order the model
			      * bitmap (see Notes above) */
    void       *pM)           /* Ptr to internal representation of model
			      * set (see Notes above) */
{
    PROLOGUE;
    
    modelSet  *set;
    
    /* Make dimens unframed */
    sWidth -= 2*cFrameBits;
    sHeight -= 2*cFrameBits;

    /* Point to the set struct portion of the model set memory chunk
     * and fill in the non-pointer fields of this struct portion */
    set = (modelSet *) pM;
    set->nBitmaps = 1;
    set->widthM = sWidth;
    set->wplM = (sBpl >> 2);
    set->heightM = sHeight;

    /* This is sneaky and rotten and probably unprofessional, but this
     * is how we'll dereference the rest of the memory chunk AND
     * assign all model pointers to locations in the memory chunk.
     */
    set->pOffsetsM = (blockHeight > 0 ? (Int32 *)set : NULL);
    set = ip_derefModelSetVoid (pIPState, (void *)pM);
    
    /* Skip binary frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;

    
    if (blockHeight > 0)
    {
	/* When blockHeight > 0, it means to set up the model
	 * to access the rows when comparing later in order of densest row
	 * first.
	 * The set memory chunk has 3 more pieces, all are Int32s arrays
	 * the length of the number model rows.
	 * The first piece is where pOffsetsM points to and is the model
	 * bitmap offsets to various rows.
	 * The second piece is where the imagery bitmap offsets to various
	 * rows is stored, and gets filled in later when a search is begun.
	 * The third piece is the array that contains the order of the
	 * actual row accesses from which the above offsets are created.
	 * For example, this array would be {0,3,1,2} if the model was
	 * 4 rows high, its second row was the densenst, the first row
	 * the next most dense, the third row the next most dense, and the
	 * zeroth row was the least dense.
	 */
	Int32 *pTmp;
	
	pTmp = (Int32 *)CALLOC(sHeight, sizeof(Int32));
	if (!pTmp)
	    return ia_nomem;

	/* (set->pOffsetsM + 2 * sHeight) is where in the set memory chunk
	 *  the row order information is stored */

	if (ip_assignModelRowOrder (pIPState, pS, sBpl, sWidth, sHeight,
				    blockHeight, pTmp,
				    set->pOffsetsM + 2*sHeight,
				    &(set->nptsM)) != ia_successful)
	    return ia_nomem;
	
	if ((__error__ = ip_rowMakeOffsetsFromOrder (pIPState,
						   set->pOffsetsM + 2*sHeight,
						   pTmp,
						   sHeight,
						   set->wplM,
						   set->pOffsetsM))
	    != ia_successful)
	    NORMAL_RETURN;
	    
	FREE(pTmp);
    }
    else
    {
	set->pOffsetsM = NULL;
	
	/* The number of points (bits that are 1/on) in each model */
	i_countBits (pIPState, pS, 0, 0,
		     sWidth, sHeight, 1, sBpl, &(set->nptsM));
    }

    NORMAL_RETURN;
}    


/* PRIVATE */
/****************************************************************************
 * Hausdorff compares model and imagery
 *
 * Description:
 *  Take the model and the imagery and compare them to see if they match, at
 *  the appropriate passed in offset of the model over the imagery.  Do this by
 *  taking the model bits and ANDing them with the imagery bits.
 *
 * Returns:
 *  The number of model bits ON in the region.
 *
 * Notes:
 *  minAndedCnt is used for early termination.  It could be calculated inside
 *  this routine, but we unrolled it for speed.
 *****************************************************************************/
Int32 CDECL 
ip_hausdorffCompareOvrlpOneModel (
    IPState    *pIPState,
    void       *pModelData,	/* Ptr to single model bitmap in words,
				 * cast to a void * */
    Int32	wplM,		/* # words per line, including padding*/
    Int32	widthM,		/* # model columns (excluding padding/frame) */
    Int32	heightM,	/* # model rows (excluding padding/frame) */
    Int32	nptsM,		/* # on bits in entire model bitmap */
    Int32      *pOffsetsM,	/* Ptr to ordered offsets table */
    Int32	x,		/* Offset coordinates to place the model's */
    Int32	y,		/* upper left corner w.r.t. the upper left
				 * corner of the imagery */
    UInt32     *pI,		/* Ptr to upper left corner of imagery,
				 * EXCLUDING frame */
    Int32	wplI,		/* # words per line, including padding/frame */
    Int32	widthI,		/* # imagery columns(excluding padding/frame)*/
    Int32	heightI,	/* # imagery rows (excluding padding/frame)*/
    Float64	minPercent,	/* The imagery and model must correlate with
				 * a Hausdorff percent of at least this much
				 * to constitute a match, [0..1] */
    Int32	minAndedCnt,	/* Pass this in as CEIL(nptsM * minPercent) */
    Int32      *pAndedCnt)  	/* Return the number of total model bits
				 * that AND over imagery bits in the overlap
				 * ( >= minAndedCnt or -1) */
{
    PROLOGUE;
    UInt32  *pM = (UInt32 *)pModelData;
    Int32    wplOverlap, heightOverlap;
    UInt32   rowLastWordMask;
    Int8     rshift;
#ifdef pbmDEBUG
    DumpPBM("Model.pbm", (UInt8 *)pM, wplM << 2, widthM + MAX(x,0), heightM);
#endif

    /* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;
    widthM = widthM;
    pOffsetsM = pOffsetsM;
    
    /* Figure the right shift to add given this x position */
    rshift = x & WORD_SIZE_MINUS1;
    
    if (ip_getOverlapDimens (pIPState, &pM, wplM, heightM, x >> WORD_OFFSET, y,
			     &pI, wplI, heightI,
			     &wplOverlap, &heightOverlap))
    {
	/* Means no portion of the imagery and model overlapped! */
	*pAndedCnt = -1;
	return nptsM;
    }
    
    /* In THIS case of using the original framed image bitmap as the model's
     * bitmap, we notice there are always 2 words of frame at the end of each
     * row.  While we DO need to compare the 1st word of the frame (so the
     * last real word of the bitmap can be shifted into this word!), we DON'T
     * need to compare the 2nd word of the frame (it will always be zero).
     */
    if (wplM == wplOverlap)
	/* We will just care about cases where the entire model row fits in
	 * the imagery (because it's easier).
	 * The 2nd word in the frame overlays the imagery, subtract it! */
	wplOverlap--;

    
    /* If the _last_ word of an imagery row is overlapping some portion of the
     * model, we have to (potentially) mask off the extra model word bits that
     * would overhang the natural imagery's end, because the imagery is of
     * course, padded to a full word.
     */
    rowLastWordMask = ((MAX(0, x >> WORD_OFFSET) + wplOverlap) >=
		       (wplI - 2*cFrameWords) ?
		       myLeftMask[widthI & WORD_SIZE_MINUS1] :
		       myLeftMask[0] /* The whole word */);

    /* Count the number of total model points that are "on" in this
     * overlapping area since we aren't trying to match the whole model */
    if (rshift)
    {
	nptsM = ip_countOverlapWithShifts (pIPState, rshift,
					   pM,
					   heightOverlap,
					   wplOverlap-1,
					   rowLastWordMask,
					   wplM);

	/* Rescale the minAndedCnt appropriately */
	minAndedCnt = (Int32)(ceil(minPercent * nptsM));

	ip_correlateWithShifts (pIPState, rshift,
			        pI,
			        pM,
			        heightOverlap,
			        wplOverlap-1,
			        rowLastWordMask,
			        wplI,
			        wplM,
			        (const UInt8 *) &(bitCount[0]),
			        nptsM - minAndedCnt,
			        pAndedCnt);
    }
    else
    {
	nptsM = ip_countOverlap (pIPState, pM,
				 heightOverlap,
				 wplOverlap-1,
				 rowLastWordMask,
				 wplM);

	/* Rescale the minAndedCnt appropriately */
	minAndedCnt = (Int32)(ceil(minPercent * nptsM));

	ip_correlate (pIPState, pI,
		      pM,
		      heightOverlap,
		      wplOverlap-1,
		      rowLastWordMask,
		      wplI,
		      wplM,
		      (const UInt8 *) &(bitCount[0]),
		      nptsM - minAndedCnt,
		      pAndedCnt);
    }

    if (nptsM - minAndedCnt <* pAndedCnt)
	/* Means the above routine terminated early */
	*pAndedCnt = -1;
    else 
	/* NOTE that *pAndedCnt is actually the number of MISSES, so make it
	 * into what it should be, the number of not MISSES */
	*pAndedCnt = nptsM - *pAndedCnt;

    return (nptsM); /* The number of ON points in the model */
}


/* PRIVATE */
/****************************************************************************
 * Hausdorff compares model and imagery
 *
 * Description:
 *  Take the model and the imagery and compare them to see if they match, at
 *  the appropriate passed in offset of the model over the imagery.  Do this by
 *  taking the model bits and ANDing them with the imagery bits.
 *
 * Returns:
 *  The number of model bits ON in the entire model.
 *
 * Notes:
 *  minAndedCnt is used for early termination.  It could be calculated inside
 *  this routine, but we unrolled it for speed.
 *****************************************************************************/
Int32 CDECL 
ip_hausdorffCompareOneModel (
    IPState    *pIPState,
    void       *pModelData,	/* Ptr to single model bitmap in words,
				 * cast to a void * */
    Int32	wplM,		/* # words per line, including padding*/
    Int32	widthM,		/* # model columns (excluding padding/frame) */
    Int32	heightM,	/* # model rows (excluding padding/frame) */
    Int32	nptsM,		/* # on bits in entire model bitmap */
    Int32      *pOffsetsM,	/* Ptr to ordered offsets table */
    Int32	x,		/* Offset coordinates to place the model's */
    Int32	y,		/* upper left corner w.r.t. the upper left
				 * corner of the imagery */
    UInt32     *pI,		/* Ptr to upper left corner of imagery,
				 * EXCLUDING frame */
    Int32	wplI,		/* # words per line, including padding/frame */
    Int32	widthI,		/* # imagery columns(excluding padding/frame)*/
    Int32	heightI,	/* # imagery rows (excluding padding/frame)*/
    Float64	minPercent,	/* The imagery and model must correlate with
				 * a Hausdorff percent of at least this much
				 * to constitute a match, [0..1] */
    Int32	minAndedCnt,	/* Pass this in as CEIL(nptsM * minPercent) */
    Int32      *pAndedCnt)  	/* Return the number of total model bits
				 * that AND over imagery bits in the overlap
				 * ( >= minAndedCnt or -1) */
{
    PROLOGUE;
    UInt32  *pM = (UInt32 *)pModelData;
    Int32    wplOverlap, heightOverlap;
    UInt32   rowLastWordMask;
    Int8     rshift;
#ifdef pbmDEBUG
    DumpPBM("Model.pbm", (UInt8 *)pM, wplM << 2, widthM + MAX(x,0), heightM);
#endif
    
    /* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;
    minPercent = minPercent;

    /* Figure the right shift to add given this x position */
    rshift = x & WORD_SIZE_MINUS1;
    
    if (ip_getOverlapDimens (pIPState, &pM, wplM, heightM, x >> WORD_OFFSET, y,
			     &pI, wplI, heightI,
			     &wplOverlap, &heightOverlap) || nptsM == 0)
    {
	/* Means no portion of the imagery and model overlapped! */
	*pAndedCnt = -1;
	return nptsM;
    }

    /* In THIS case of using the original framed image bitmap as the model's
     * bitmap, we notice there are always 2 words of frame at the end of each
     * row.  While we DO need to compare the 1st word of the frame (so the
     * last real word of the bitmap can be shifted into this word!), we DON'T
     * need to compare the 2nd word of the frame (it will always be zero).
     */
    if (wplM == wplOverlap)
	/* We will just care about cases where the entire model row fits in
	 * the imagery (because it's easier).
	 * The 2nd word in the frame overlays the imagery, subtract it! */
	wplOverlap--;
    

    /* If the _last_ word of an imagery row is overlapping some portion of the
     * model, we have to (potentially) mask off the extra model word bits that
     * would overhang the natural imagery's end, because the imagery is of
     * course, padded to a full word.
     */
    rowLastWordMask = ((MAX(0, x >> WORD_OFFSET) + wplOverlap) >=
		       (wplI - 2*cFrameWords) ?
		       myLeftMask[widthI & WORD_SIZE_MINUS1] :
		       myLeftMask[0] /* The whole word */);

    if (pOffsetsM && heightOverlap == heightM)
    {
	if (rshift)
	    ip_correlateOrderedWithShifts (pIPState, rshift,
					   pI,
					   pM,
					   heightM,
					   wplOverlap-1,
					   rowLastWordMask,
					   pOffsetsM + heightM,
					   pOffsetsM,
					   (const UInt8 *) &(bitCount[0]),
					   nptsM - minAndedCnt,
					   pAndedCnt);
	else
	    ip_correlateOrdered (pIPState, pI,
				 pM,
				 heightM,
				 wplOverlap-1,
				 rowLastWordMask,
				 pOffsetsM + heightM,
				 pOffsetsM,
				 (const UInt8 *) &(bitCount[0]),
				 nptsM - minAndedCnt,
				 pAndedCnt);
    }
    else
    {
	if (rshift)
	    ip_correlateWithShifts (pIPState, rshift,
				    pI,
				    pM,
				    heightOverlap,
				    wplOverlap-1,
				    rowLastWordMask,
				    wplI,
				    wplM,
				    (const UInt8 *) &(bitCount[0]),
				    nptsM - minAndedCnt,
				    pAndedCnt);
	else
	    ip_correlate (pIPState, pI,
			  pM,
			  heightOverlap,
			  wplOverlap-1,
			  rowLastWordMask,
			  wplI,
			  wplM,
			  (const UInt8 *) &(bitCount[0]),
			  nptsM - minAndedCnt,
			  pAndedCnt);
    }

    if (nptsM - minAndedCnt < *pAndedCnt)
	/* Means the above routine terminated early */
	*pAndedCnt = -1;
    else
    {
	/* NOTE that *pAndedCnt is actually the number of MISSES, so make it
	 * into what it should be, the number of not MISSES */

	if ((heightOverlap == heightM) && (x >= 0) && (x + widthM <= widthI))
	{
	    /* The entire model fits in the image, the number of ANDs is thus
	     * the number of total model points minus the number of MISSes*/
	    *pAndedCnt = nptsM - *pAndedCnt;
	}
	else
	{
	    /* The number of ANDs is the number of model points in the
	     * overlapping region minus the number of MISSes */
	    *pAndedCnt *= -1;

	    *pAndedCnt += rshift ?
		ip_countOverlapWithShifts (pIPState, rshift,
					   pM,
					   heightOverlap,
					   wplOverlap-1,
					   rowLastWordMask,
					   wplM) :
		ip_countOverlap (pIPState, pM,
				 heightOverlap,
				 wplOverlap-1,
				 rowLastWordMask,
				 wplM);
	}
    }

    return (nptsM); /* The number of ON points in the model */
}    


/*

$Log: i_hc1ms.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.9  1998/01/13 23:48:19  ddavies
i_hc1ms.c -> /product/ipcore/ipshared/src/RCS/i_hc1ms.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1995/12/13 02:44:24  ejaquith
i_hc1ms.c -> /product/ipcore/ipshared/src/RCS/i_hc1ms.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.7  1995/09/14  18:59:39  ejaquith
 * ejaquith on Thu Sep 14 11:59:38 PDT 1995
 *
 * i_hc1ms.c -> /product/ipcore/ipshared/src/RCS/i_hc1ms.c,v
 *
 * CommonCheck: Converted all routines that were private in files to be of form
 * ip_*(), which they weren't.  No functionality/noticeable changes.
 *
 * Revision 1.6  1995/04/22  20:45:28  ejaquith
 * i_hc1ms.c -> /product/ipcore/ipshared/src/RCS/i_hc1ms.c,v
 *
 * Please see general (large!) module comments for same date in i_hc32ms.c
 * Removed i_imagePutModelOneModel(), i_getModelSetSizeOneModel(), and
 * i_makeModelSetOneModel().
 * Added i_getModelSetSizeOnePtr() and i_makeModelSetOnePtr() which will
 * get the size for and fill a model using the single bitmap pointer
 * model set data type.
 *
 * Revision 1.5  1995/03/29  01:57:51  ejaquith
 * i_hc1ms.c -> /product/ipcore/ipshared/src/RCS/i_hc1ms.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.4  1995/03/15  01:07:38  ejaquith
 * i_hc1ms.c -> /product/ipcore/ipshared/src/RCS/i_hc1ms.c,v
 *
 * Updated i_getModelSetSizeOneModel() and i_makeModelSetOneModel()
 * to work with row ordered model new feature and use blockHeight
 * information.
 * Added calls to i_correlateOrdered() and
 * i_correlateOrderedWithShifts()
 *
 * Revision 1.3  1995/01/04  18:39:33  ejaquith
 * ejaquith on Wed Jan 4 10:35:12 PST 1995
 *
 * i_hc1ms.c -> /product/ipcore/ipshared/src/RCS/i_hc1ms.c,v
 *
 * Abandoned use of "unit" (variable sized word for storage of models) for
 * use of 32 bit words.  Made bit counting safer.  Changed all public routines
 * to use image byte pointers and bytes per line (from image word pointers
 * and words per line).  Fixed bug in correlation percent calculation (minor).
 * Changed correlation loop calls appropriately to use number of MISSes when
 * bitwise ANDing imagery and model.
 *
 * Revision 1.2  1994/12/07  22:59:35  ejaquith
 * i_hc1ms.c -> /product/ipcore/ipshared/src/RCS/i_hc1ms.c,v
 *
 * Ref'd unused parameters to keep compiler happy.
 *
 * Revision 1.1  1994/12/06  23:41:14  ejaquith
 * Initial revision
 *

*/
