/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * i_hc32ms.c:  Hausdorff binary image comparator
 *  	    	when a model set is represented as WORD_SIZE bitmaps,
 *  	    	each a shift of the original model binary bitmap
 *
 *  Public routines:
 *
 *  	    Code to construct a set of models
 *		i_getModelSetSize()
 *  	    	i_makeModelSet()
 *
 *	    Code to sort a list of elements
 *		i_orderBySort()
 */

#include <stdlib.h>  /* for rand() */
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"
#include "memory.pub"
#include "bitcount.pub"
#include "defines.pub"

    /* prototypes */
#include "shrpatt.prv"
#include "shrpatt.pub"
#include "shrip.pub"  /* for i_countBits() */
#include "shrip.prv"  /* for ip_countBitsLine() */

IP_RCSINFO(RCSInfo, "$RCSfile: i_hc32ms.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");


/*____________________________________________________________________________
 *
 *
 * This purpose of this code is to implement comparison of two binary
 * bitmaps using the Hausdorff metric method.  The comparison can be either
 * 1-way or 2-way.  The method can be thought of as a "dilate and correlate"
 * method.
 *
 * To try it out, run the "haustest" program in ipshared/testsrc.
 *
 * NOTE:  This entire module is really simply a normalized binary correlation
 * package in disguise.  If you pass in dilated images (as you should), it's
 * Hausdorff binary correlation.  However, if you do not dilate the images
 * (tsk, tsk), everything becomes straightforward normalized binary
 * correlation (typically an inferior binary correlation tool, but alas
 * a faster one as dilation is not done).
 *
 *____________________________________________________________________________
 *__________________HIGH-LEVEL DESCRIPTION of technique_______________________
 *
 * Given two binary images, AS and BS, and their dilations AD and BD,
 * we can compare them:
 *
 * 1-way:
 *      Take AS and place it over BD.  Count the number of ANDed points, #AND.
 *      Find the highest value of #AND given all possible translations of AS
 *      over BD.  Return the Hausdorff percent as this highest
 *      #AND / #points in AS.  This is the percent correlation between
 *      AS and BS.
 *
 * 2-way:
 *      Take AS and place it over BD.  Count the number of ANDed points, #AND1.
 *      HP1 = #AND1 / #points in AS.
 *      At the corresponding location, place BS over AD.  Count the number of
 *      ANDed points, #AND2.
 *      HP2 = #AND2 / #points in BS.
 *      The Hausdorff percent at this location is MIN(HP1, HP2).
 *      Find the highest Hausdorff percent given all possible translations of
 *      AS over BD, and return this as the perecent correlation between
 *      AS and BS.
 *    
 * The 2-way comparison is generally used.  Here's why:
 * Let's say we are trying to match a bitmap of the character "z" to some
 * other image.  If the image is all black, the 1-way comparison will return
 * 100%!
 *
 *____________________________________________________________________________
 *______________Why Hausdorff?_______What can I do with it?___________________
 *
 * This method is typically much more accurate than normalized binary
 * correlation because it incorporates dilating to adjust for quantization
 * error and noise.  The dilation allows points to be "very close" but
 * not necessarily directly overlapping.
 *
 * It can be used to:
 *   - Determine if 2 images are similar
 *   - Find the location of some subimage(s) in another image
 *   - Use a multi-resolutional searching technique for searching for matches
 *     during image comparison
 *
 *____________________________________________________________________________
 *___________PHILOSOPHY and IMPLEMENTATION of Hausdorff comparator____________
 *
 * The purpose of this module is to compare two binary images and determine
 * some good measure of the percentage correlation between the two.
 *
 * 
 * Here is a description of the underlying math.
 *   [modeled after a document by Dan Huttenlocher from PARC,
 *    http://www.cs.cornell.edu/Info/People/dph/hausdorff/hausdorff.html]
 * Given two sets of points
 *     AS={a1,...,am}, BS={b1,...,bn}
 * the Hausdorff distance is defined as
 *     H(AS,BS) = max(h(AS,BS), h(BS,AS))
 * where
 *     h(AS,BS) = max (AB1,...,ABm)
 *     h(BS,AS) = max (BA1,...,BAn)
 * where
 *     ABi = distance between ai and the closest point in BS      i = 1...m
 *     BAj = distance between bj and the closest point in AS      j = 1...n
 *
 * The function h(AS,BS) is called the _directed_ Hausdorff 'distance' from
 * AS to BS (this function is not symmetric and thus is not a true distance).
 * It identifies the point, a* in AS, that is farthest from any point in BS,
 * and measures the distance from a* to its nearest neighbor in BS.  Thus the
 * Hausdorff distance, H(AS,BS), measures the degree of mismatch between two
 * sets, as it reflects the distance of the point of AS this is farthest from
 * any point of BS and vice versa.  Intuitively, if the Hausdorff distance
 * is 'd', then every point of AS must be within a distance 'd' of some point
 * of BS and vice versa.  It is well known that the Hausdorff distance is a
 * metric over the set of all closed, bounded sets -- it obeys the properties
 * of identity, symmetry, and triangle inequality.
 *
 * One way to understand the directed Hausdorff measure is to think of it in
 * terms of set containment.  Let BD = BS + Cd, where Cd is a disk of radius d
 * and + is the Minkowski sum
 *   (for two point sets P and Q, P + Q = {p + q | p in P, q in Q})
 * Intuitively, BD is the set obtained by replacing each point of BS with a
 * disk of radius d, and taking the union of all of these disks.
 * Thus, BD is simply the dilated by disc radius d version of BS.
 * By definition
 *         h(AS,BS) <= d     if and only if   AS is a subset of BD
 * because in order for every point of AS to be within distance d of BS, it
 * must be contained in BD.
 *
 * The Hausdorff distance is very sensitive to even a single "outlying" point
 * of AS or BS.  For example, consider AS = BS union {x}, where the point x is
 * some large distance D from any point of AS. In this case H(AS,BS) = D -- it
 * is determined solely by the point x. Thus, rather than using H(AS,BS) we
 * use a generalization of the Hausdorff distance.  We ensure that a large
 * fraction of the number of points in AS fall within distance d of some point
 * in BS, instead of 100% of the points in AS.
 *____________________________________________________________________________
 *_______________KEY IDEA if you don't like the above math!___________________
 *
 * Thus, we may use the Hausdorff distance to tell us given AS and BS,
 *  "What percent of the points of AS fall within distance 'd' of any point
 *   in BS?"
 * We simply dilate BS by 'd' to make BD, and then AND the point sets together.
 * This one-way (directed) percent from AS to BS is thus the number of ANDed
 * points over the number of points of AS.  We can then take BS and AD
 * (AS dilated by disc of radius 'd') and determine the corresponding
 * directed percent from BS to AS.
 * The overall corelation percent is thus the min of the two percents.
 *
 *____________________________________________________________________________
 *_____________________________NOTATION_______________________________________
 *
 *.We refer to the two point sets we are comparing _overall_ as A and B.
 *
 * This comparator will compare two binary images AS and BS. 
 *
 * A "model set" is made (internally) from an original source image, S.
 *
 * An "imagery" is an actual binary image, the dilation of S.
 *
 * A is made up of a "model set", AM, and an "imagery", AD.
 * B is made up of a "model set", BM, and an "imagery", BD.
 *
 * Or, if you like, quasi-mathematically
 *   AS -> AM, AD
 *   BS -> BM, BD
 *   A  = {AM, AD}
 *   B  = {BM, BD}
 *
 * Thus a 1-way comparison of A and B would be done by:
 *   placing AM over BD at all translations of AM over BD.
 *
 * And a 2-way comparison of A and B by:
 *   placing AM over BD and BM over AD, at all translations of AM over BD.
 *
 *____________________________________________________________________________
 *                  IMPORTANT IMPLEMENTATION Details
 *_________________________ALGORITHM description______________________________
 *
 *METHODS:
 * We provide three methods of matching (once the model set(s) have been made)
 *
 *   i)  i_hausdorffRectSearchSize()
 *       i_hausdorffRectSearchInit()
 *       i_hausdorffSearch()
 *
 *  ii)  i_hausdorffListSearchSize()
 *       i_hausdorffListSearchInit()
 *       i_hausdorffSearch()
 *
 * iii)  i_hausdorffMultResFilterSearchInit()
 *       i_hausdorffSearch()
 * 
 * i) This method is to compare the two binary images, A and B.  We place
 * A at all locations in some subrect of B and return all locations where the
 * images match with a percent greater than or equal to some passed in percent.
 * Each match is returned one at a time with iterative calls to
 * i_hausdorffSearch().
 *
 * ii) This method is like i) except it allows an unconstrained list of
 * translations to try instead of a subrect to search.
 *
 * iii) This method is identical in function to i) but typically much faster
 * (on the order of 10 to 40 times faster) because it filters out translations
 * in the subrect that aren't matches in a multi-resolutional approach.
 * (see note below).
 *
 * The methods return the calculated Hausdorf match percents as well as
 * match locations.
 *
 *DILATION, 'DILATERADIUS'
 * When the original binary image point sets AS and BS are previously created
 *   by the client, it may be desirable to have made them with an additional
 *   border (left, right, top, and bottom) of 'dilateRadius' zeroed pixels
 *   around the binary images.
 *   'dilateRadius' is determined by how much the AS and BS images are dilated
 *   in making AD and BD.  'dilateRadius' is exactly how much the image would
 *   need to be swollen on all four borders to contain the _full_ dilation!
 *   For example, let's say we use STREL
 *      1 1 1
 *      1 1c1  (what we typically suggest)
 *      1 1 1
 *   then 'dilateRadius' is equal to 1.  Using this additional border is
 *   beneficial because every dilated pixel and original pixel in A is
 *   important and must be recorded.  If some pixel 'a' was 1 on the left
 *   hand edge of AS, then AD would not contain the dilated pixel to the
 *   left of 'a' unless there was a 1 pixel border of zeros in AS _before_
 *   it was dilated to make AD.
 *   If this is not done, comparator results may be slightly less accurate.
 *
 * If using the above 'dilateRadius' border, it must be used in BOTH AS and
 *   BS, and it must be the_same_.  AD and BD must be created by the same
 *   corresponding dilation STREL.
 *
 * The dilation STREL must of course be symmetric in X and Y for it to make
 *   sense.
 * Note that as the size of the dilation STREL increases, so does the returned
 *   Hausdorff percent.  However, if increased too much, information (for
 *   example, topology) will become lost, and the likelihood of a mismatch
 *   will also increase.  That is why we suggest the above STREL.
 *
 *MINIMUM HAUSDORFF PERCENTS:
 * Use of minimum Hausdorff percents is useful in i_hausdorffSearch().  If
 * the percent(s) is set to 0%, it will return back _all_ locations tested
 * with their associated percent correlations.  This may be useful if you
 * wanted to have your own "best match" critique.  Futher, if the 2-Way
 * comparison is used, the B to A percent may be set to 0% and the A to B to >
 * 0%.  This would allow you to select the best A,B matches yourself after you
 * knew the A to B match was above some certain percent.
 * The passed in percents may change for each invoked iteration of
 * i_hausdorffSearch().  This is typically useful when making the percents
 * increase or remain the same at each call to find a single "best" match.
 * For example, you could start looking for 60%,60% matches.   If the first
 * match returned was 70%,70%, you should increase the minimum percents to
 * 70%,70% so all remaining iterative calls only report matches at least this
 * good.  As pointed out below, the higher these minimum percents, the
 * faster the search can be done.
 *
 * For i_hausdorffSearch() using 2-Way comparison, two hausdorff percents are
 * used as minimum thresholds.  All things considered, you should _definitely_
 * arrange your calls to this routine to have the A to B percent greater than
 * the B to A percent (if they are not equal).  This is because the matcher
 * always searches for acceptable A to B match percents first, and then B to A
 * matches.  The higher your percents are, the faster the inner comparison
 * loop will speed up with early termination. (see below)
 *
 * If the Hausdorff comparator terminates early, it will return -1.0%
 * (to differentiate it from other possible legal values [0..1] and since
 *  no actual percent gets calculated).
 *
 *SUBRECT, SPARSE FILTERING:
 * Often it is desirable to find A _in_ B, where A is most likely contained in
 * some known region, B.  Subrect searching is the most common interface.
 * This will try matching A at/over every location offset in some subrect of
 * B, and return the best match.  Using subrects allows use of a technique
 * for calculating densities before it actually calls a Hausdorff routine,
 * to see if it can avoid calling the Hausdorff comparator routine
 * altogether.  If we place A at some location in B, and we notice that the
 * overlapping area in B has too little points to meet some target match
 * percent, then we don't call the Hausdorff comparator routine.  We refer to
 * this as "sparse filtering".  Note that as the target match percent is
 * always nondecreasing, when it increases, more locations will become sparse.
 *
 *MULTI-RESOLUTION SEARCHING, TRANSLATION AREAS:
 * Another way of searching for A _in_ B is by trying to determine where large
 * areas of translations will not yield a match first (checking at a coarse
 * resolution) before doing the final match checking (at the finest
 * resolution).  The idea is pretty simple.  Suppose we are trying to match
 * AS in subrect BD.  Divide the subrect up into translation areas, squares
 * of some dimension, say 16.  Now take BD and dilate it by a 16x16 lower-right
 * corner dilation (that is to say, for every "on" pixel in BD, turn on the
 * 15 pixels above it, the 15 pixels to the left, and the pixels in the square
 * that the two lines share).  Now we may take the top left corner of every
 * translation area and try matching AS over (the new) BD at this top left
 * corner location.  If it doesn't match, we know that none of the remaining
 * translations in this 16x16 translation area can possibly yield a match
 * (shifting AS anywhere else inside this 16x16 area cannot do better).
 * Thus, we may toss out 255 translations without ever checking.
 * If it does match, we need to determine what our next translation area
 * square size is, say 8.  We then divide up our 16x16 square into four 8x8
 * squares.  We dilate the original BD by a 8x8 lower-right corner dilation,
 * and check the new 4 translation areas.  And so on, until we get to the
 * final desired translation area square size (> 1).  Here, each translation
 * area that matches is divided into the full number of translations it
 * contains, and AS gets placed at each location over the original BD.
 * Finally, anything that matches is a true match.
 *
 * We do breadth-first matching on account of the dilations that need to be
 * done.  That is to say, we keep a list of all locations that match at 16x16,
 * then go to the next level and update the list to all locations that match
 * at 8x8, and so on.  Thus, BD gets dilated once for each "level", only.
 *
 * The translation areas are up to the client, providing they are all
 * multiples of 2, and in strict decreasing order.  The last area size is
 * always 1.  Areas of 16,8,4,2,1 seem pretty fast on a 300 dpi text image.
 * As does 8,4,2,1.  The largest area allowed is 128, and would be
 * desirable with images containing very large sparse areas.
 *
 * On a typical page, multi-resolutional searching is 10 to 30 times faster
 * than the normal sparse filtering search techniques!
 *
 *SKIPX, SKIPY:
 * Routine calls allow for restricting trying matches at every 1 in n
 * translations.  If the values are 1 and 1 respectively, the routines will
 * search every translation in the region.  Otherwise, it searches 1 in
 * (skipx * skipy) translations.  If the values are 2 and 3 respectively,
 * it will try every 2 columns and every 3 rows in the translation space,
 * or 1 in 6.  The routines will be approximately the same factor faster, and
 * this could be used instead of downsampling.  A coarse-fine search could be
 * undertaken, where a first pass would be done, searching every 1 in 4 or 1
 * in 9 translations, and then a smaller subrect could be fit around any
 * resulting match to get more precise.
 *
 *TRYING TO MATCH ONE IMAGE -IN- ANOTHER NOTE:
 * If you are searching for the best location of one full image inside some
 * subregion (including its entirety) of another image, make the first image
 * A and the second image B.  To see why, see the USE_FULL_B_FOR_BTOA_PERCENT
 * note below.
 *
 *MATCH ALIGNMENT:
 * If you have a hard time being sure your matches are returning the correct
 * locations, try "calibrating" by clipping a subrect of an image to be "A"
 * and comparing it to the whole image "B", and make sure you get the match
 * in the place the subrect was clipped from!
 *
 *SEARCH METHOD:
 * When doing a non multi-resolutional subrect search, you pick one of
 * these specific search methods: (from ipshared/include/hausconf.pub)
 *  SPARSE_FILTER
 *  EVERY_TRANS
 *  FLY_FILTER
 * and may OR it with USE_FULL_B_FOR_BTOA_PERCENT.
 * When doing a multi-resolutional subrect search or doing a list of
 * translations search, you pick 0 or USE_FULL_B_FOR_BTOA_PERCENT as the
 * specific search method.
 * SPARSE_FILTER:
 *  Most common method that does sparse filtering of subrect areas out
 *  and processes the search area in strips.
 * FLY_FILTER:
 *  The above sparse filtering search searches in strips, and filters
 *  by allocating some memory.  If memory allocation is undesirable, use
 *  this flag to sparse filter on-the-fly without memory allocation.
 *  However, this will not search in strips like the default method,
 *  which is likely to be about 25-30% faster.
 * EVERY_TRANS:
 *  Sometimes, one would like to search certain subrects where the sparse
 *  filtering or multi-resolutional approaches will be slower (for example,
 *  a subrect that will allow 50 translations of the model in X, and just
 *  the single 1 translation in Y).  Use this flag to try Hausdorff matching
 *  at EVERY translation, without bothering to sparse filter out or to filter
 *  out using multi-resolutional searching certain translations in the subrect.
 * USE_FULL_B_FOR_BTOA_PERCENT
 *  This module assumes A is smaller than the other image B.  Presumably,
 *  we'd then like to find the best location of A that overlaps B somewhere.
 *  Our A to B Hausdorff correlation percent would return at every specified
 *  location the #pts of AS that overlap BD divided by the total #pts of AS.
 *  However, once we did the B to A correlation percent, we'd notice the
 *  percent is very low _even if the bitmap portions that overlap are
 *  identical_.  This is because the B to A percent is the #pts of BS that
 *  overlap AD at the corresponding location divided by the _total_ #pts of
 *  BS.  The total #pts BS is presumably much greater than the total #pts AS,
 *  since A is bigger than B.  Thus, we'd like to say given the area of BD
 *  that AS overlapped it at, calculate the percent correlation between the
 *  _same sized corresponding area_ of BS and AD.  Thus, while the #pts of BS
 *  that overlap AD will be the same, the denominator will become the #pts in
 *  BS that overlap AD, a smaller value than the total #pts in BS
 *  (presumably).  This will make the Hausdorff percent from B to A as it is
 *  intuitively desired in this case.  We do this by default.
 *  HOWEVER, if A and B are approximately the same size, you would want to
 *  explicitly tell the comparator to use the full #pts BS for a more
 *  accurate B to A percent, and would do so by ORing the passed in search
 *  method with USE_FULL_B_FOR_BTOA_PERCENT.
 *  Thus, one thing to remember is that you always make B the larger image,
 *  when you are looking for an A image _in_ a B image.
 *  
 *NUMBER OF BITMAPS, SIZE/MEMORY CONSIDERATIONS
 * The Hausdorff comparator compares the bits of binary images.  It can do
 * so presently in three ways, using 32 bitmaps of an image for fastest
 * comparison, using 1 bitmap of an image for less memory, or not using
 * bitmaps at all for sparse images.  Model's number of bitmaps:
 * 32:
 *  The model set will be constructed with 32 pointers to full image bitmaps
 *  similar to the original passed in image, only shifted.  This allows no
 *  further shifting when later comparing images, and thus is fast.  The 0th
 *  bitmap will be the unshifted image, the 1st bitmap the image right
 *  shifted 1, ...
 * 1:
 *  The model set will instead use 1 bitmap pointer to the original image
 *  bitmap.  This requires almost no additional memory whatsoever, but
 *  is a little slower, as shifting must be done at each comparison.
 * 0:
 *  The model set will be constructed as a list of x offsets, one offset for
 *  each on point, plus one "flag" offset per row of the image signifying the
 *  end of the row.  This may be significantly smaller in size than using 32
 *  models like above, but not necessarily as fast.  For example, for a 1MB
 *  image that is 25% dense (1/4 of its pixels are "on"), it will need
 *  approximately 2 bytes per pixel, and thus use about 4MB of memory to make
 *  a model set.  However, for very sparse images, this may be the fastest
 *  model storage type.
 *
 * The nice thing about this set up is that the model set data is internally
 * represented without the client needing to know its exact representation.
 * Further, one may use 2 different methods for 2 different model sets and
 * still compare them.
 * For example, if you were comparing small bitmaps in subrects of a 1MB
 * image, you could create the model sets for the small bitmaps using the
 * 32 bitmap method, and create the model set for the 1MB image using the
 * 1 bitmap method, for a nearly optimal mix of speed and memory allocation.
 *
 *MODEL BLOCKHEIGHT:  (not useable when number of bitmaps is 0)
 *  When a model set is created, a nonsequential ordering of the bitmap rows
 *  is usually much faster for comparing with other bitmaps (imagery).
 *  This is because the bitmap comparison routines use early termination
 *  (when too many model points are not lying over imagery points) and thus
 *  will terminate faster if the densest model rows are compared first.
 *  This scheme would be optimal with an unlimited cache size.  However, with
 *  a finite cache size, if the model and imagery row pointers are jumping
 *  all over the place, cache misses may occur and slow things down.
 *  Thus, when creating a model, a parameter "blockHeight" is passed by the
 *  client to control model row ordering:
 *   If blockHeight is 0, it means don't order the model rows, and always
 *   sequentially compare the model row by row. (slowest)
 *   If blockHeight >= the model height, it means order all the model rows
 *   densest first.
 *   Otherwise, divide the model into blocks of blockHeight rows.  Within each
 *   block, order the rows densest first.  Compare the densest blocks first.
 *   This is to help avoid cache misses, and blockHeight should thus be set
 *   with the width of the model bitmap AND the imagery bitmap in mind.
 *  Further, blockHeight should only be nonzero when creating an A model set.
 *
 *MODEL CREATION:
 * The model set gets set up in i_makeModelSet()/makeModelSet().  The client
 * allocates space for this model previous to the call using hir own allocater
 * with the size needed determined by i_getModelSetSize().  The client later
 * frees this model set by simply doing hir own free to the void * model set
 * pointer.
 * If it is desired to compare one image with many other images, keep the one
 * image's model set until it has been compared with all the other images.
 *
 *COMPARATOR ALGORITHM:
 * Once we have two binary images to compare and a location to compare them
 * at, the comparison routines internally make one call to ipshared's
 * i_hc32ms.c: i_hausdorffCompare() if we want to do a 1-Way comparison,
 * and two calls to it if we want to do a 2-Way comparison.
 * This main comparator function, i_hausdorffCompare(), gets called with:
 *   a model
 *   a location offset (to place the model over the imagery) 
 *   an imagery (the dilated version of the original non-shifted other model)
 *   a desired percent (that will constitute a match in this routine)
 *   the way to calculate the percent (see the description in the arg list)
 *   some ptrs for the returned information
 * It proceeds as follows:
 *   1. Skip the binary frame of the imagery
 *   2. Calculate how many points of mismatch can be tolerated before we
 *      know the model and imagery cannot possible have at least the desired
 *      percent of correlated points
 *   3. Calculate the portion of the model that overlaps the imagery at the
 *      desired offsets and adjust the word pointers to their beginning
 *      locations for bitwise ANDing
 *   4. AND the model and imagery, proceeding row by row.
 *      Count the number of model points on for this row.
 *      Count the number of ANDed points for this row.
 *        * At the end of each row, check if the mismatch (from 2. above) is
 *          enough that we may terminate ANDing early.
 *  	    If so, return -1.0% as the percent correlation.
 *   6. Calculate and return the percent of correlating points.  Calculate by:
 *        * The number of ANDed points / the number of total model points
 *          if we are checking the full model and imagery correlation.
 *        * The number of ANDed points / the number of _overlapping_ model
 *  	    points if we are checking the overlapping model and imagery
 *          correlation. (if USE_FULL_B_FOR_BTOA_PERCENT not used).
 *
 * Some changes to the above algorithm that were entered 12/1/94:
 *  Unrolled the imagery frame skipping (1), the calculation of the number of
 *  mismatch (2), and the calculation of the actual Hausdorff percent (6).
 *  Now:
 *  	*   the calling routines pass in the pointer to the imagery as the
 *  	    first word in the actual bitmap, and skips the frame
 *  	*   the calling routines decide what the number of mismatch points
 *  	    should be and pass it in as a parameter (maxMissCnt) because this
 *  	    value sometimes will not change at all (and saves a multiply)
 *  	*   i_hausdorffCompare() returns the number of ANDed model points
 *  	    that actually overlap points in the imagery, and the number of
 *  	    model points in the relevant area (The whole model if
 *  	    USE_FULL_B_FOR_BTOA_PERCENT, or just the overlap region o.w.)
 *  	    Calling routines may calculate the Hausdorff percent then as the
 *          #ANDed/#model points, these two returned values.
 * Just these simple unrolled changes and a few others dropped timings by a
 * good degree (about 10-20% for a quarter of a million translations!)
 *____________________________________________________________________________
 */


/* Needed for masking off model row end bits that overhang the natural end of
 * each row of the imagery when overlapping the two regions.
 * (note how this differs slightly from lmask.prv)
 */
static UInt32 myLeftMask[32] =
{   0xFFFFFFFF, 0x80000000, 0xC0000000, 0xE0000000,
    0xF0000000, 0xF8000000, 0xFC000000, 0xFE000000,
    0xFF000000, 0xFF800000, 0xFFC00000, 0xFFE00000,
    0xFFF00000, 0xFFF80000, 0xFFFC0000, 0xFFFE0000,
    0xFFFF0000, 0xFFFF8000, 0xFFFFC000, 0xFFFFE000,
    0xFFFFF000, 0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00,
    0xFFFFFF00, 0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0,
    0xFFFFFFF0, 0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE };


#ifdef pbmDEBUG
    /* When this is defined, it makes three images get dumped in the Hausdorff
     * comparator loop below, in PNM form.  The 3 files are binary bitmaps,
     * one corresponding to the model passed in, one to the portion of this
     * model that overlapped the imagery, and one to the actual overlapped and
     * ANDed bits.
     */

void CDECL 
ip_dumpPBM (
    IPState    *pIPState,
    const char *fn, /* file name */
    UInt8      *ptr, /* Image ptr (packed bytes) */
    Int32	bpl, /* Image bytes per line */
    Int32	w, /* Width of image in bits */
    Int32	h)      /* Height of image in bits */
{
    PROLOGUE;
    FILE *fp;
    Int32 pbm_bpl, i;

    if ((fp = fopen(fn, "w")) == NULL)
    {
	fprintf(stderr, "couldn't open debugging file '%s'.\n", fn);
	return;
    }
    pbm_bpl = w / 8 + (w & 7 ? 1 : 0);

    fprintf(fp, "P4\n#overlap.pbm\n%ld %ld\n", w, h);
    for (i = 0; i < h; i++)
    {
	if (fwrite(ptr, sizeof(UInt8), pbm_bpl, fp) != pbm_bpl)
	    fprintf(stderr, "couldn't write debugging file '%s'.\n", fn);

	/* Go to next image row */
	ptr += bpl;
    }

    fclose(fp);
}

#endif /* pbmDEBUG */

    
/* PRIVATE */
/****************************************************************************
 * getOverlapDimens():
 *  	Given the size of the model and the model's desired position with
 *  	respect to the imagery, figure the number of overlapping rows and
 *  	columns in the model that will be ANDed over some portion of the
 *  	imagery.
 *  	Move the imagery and model pointers to their starting positions.
 *
 *  Input:
 *  	ppM: 	    	Ptr to ptr to internal representation of ONE model
 *  	wplM:	    	Number of words in a model line/row.
 *  	heightM:    	Number of rows in model
 *  	x, y:  	    	Offset coordinates to place the model's upper left
 *  	    	    	corner with respect to the upper left corner of the
 *  	    	    	imagery, and with respect to WORDs.  (That is, the
 *  	    	    	x value is the x word offset desired between the two,
 *  	    	    	and the y value is the row offset between the two.
 *  	    	    	(0,0) will align the model's and imagery's
 *  	    	    	upper left word corners when ANDing words.
 *  	ppI: 	    	Ptr to ptr to upper left corner of imagery's frame,
 *			excluding frame and padding
 *  	wplI:	    	Number of words in imagery line/row, including frame
 *  	    	    	and padding.
 *  	heightI:    	Number of rows in imagery excluding frame and padding.
 *  	pWplOverlap:	Ptr to returned number of words per row that overlap
 *  	pHeightOverlap:	Ptr to returned number of rows that overlap
 *
 *  Returns:
 *  	0   model bitmap overlaps imagery
 *  	1   model bitmap does not overlap imagery
 *****************************************************************************/
Int32 CDECL 
ip_getOverlapDimens (
    IPState    *pIPState,
    UInt32    **ppM,
    Int32	wplM,
    Int32	heightM,
    Int32	x,
    Int32	y,
    UInt32    **ppI,
    Int32	wplI,
    Int32	heightI,
    Int32      *pWplOverlap,
    Int32      *pHeightOverlap)
{
    PROLOGUE;

	/* keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    if (y < 0)
    {
	/* Advance model pointer to first row to AND */
	*ppM -= y * wplM;
	*pHeightOverlap = MIN(heightM + y, heightI);
    }
    else
    {
	/* Advance imagery pointer to first row to AND */
	*ppI += y * wplI;
	*pHeightOverlap = MIN(heightI - y, heightM);
    }

    /* Now make the imagery width in words be what it ACTUALLY is (without
     * the frame) */
    wplI -= 2*cFrameWords; 
    
    if (x < 0)
    {
	/* Advance model pointer to first model's long to AND */
	*ppM -= x;
	*pWplOverlap = MIN(wplM + x, wplI);
    }
    else
    {
	/* Advance imagery pointer to first imagery's long to AND */
	*ppI += x;
	*pWplOverlap = MIN(wplI - x, wplM);
    }

    return (*pHeightOverlap <= 0 || *pWplOverlap <= 0 ? 1 : 0);
}


/* PRIVATE */
/****************************************************************************
 * countModelPoints():
 *
 *  Input:
 *  	pM: 	    	Ptr to internal representation of ONE model
 *  	wordsPerModel:	Number of words in this model
 *
 *  Returns:
 *  	Number of "on" points(pixels) in this model
 *****************************************************************************/
static UInt32 CDECL 
ip_countModelPoints (
    IPState    *pIPState,
    UInt32     *pM,
    Int32	wordsPerModel)
{
    PROLOGUE;
    UInt32  cnt = 0;
    UInt32  word;

	/* keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for (--wordsPerModel; wordsPerModel >= 0; wordsPerModel--)
    {
	word = pM[wordsPerModel];
	cnt += bitCount[word & 0xFF];
	word >>= 8;
	cnt += bitCount[word & 0xFF];
	word >>= 8;
	cnt += bitCount[word & 0xFF];
	word >>= 8;
	cnt += bitCount[word];
    }
    
    return cnt;
}


/* PRIVATE */
/******************************************************************************
 * Count the number of "on" bits in a region
 *
 * Description:
 *  Given a pointer to a word in an image (model or imagery) bitmap, and
 *  some bounds to define a specified region, count the number of "on"
 *  bits in this region.
 *
 * Returns:
 *  The number of "on" bits in the region.
 *****************************************************************************/
Int32 CDECL 
ip_countOverlap (
    IPState    *pIPState,
    UInt32     *pM,		   /* Ptr to first word of image region */
    Int32	heightRegion,	   /* Number of rows in the region */
    Int32	wplRegion,	   /* Number of words per line in the region,
				    * minus one */
    UInt32	rowLastWordMask,   /* A left mask for the last overlapping word
				    * of each image row, to chop off excess
				    * bits of the last row word that aren't
				    * in the region. */
    UInt32	wplM)              /* # words per line, including padding and
				    * frame */
{	     
    PROLOGUE;
    UInt32    word;
    Int32     ovrlpCnt = 0, column;

	/* keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for ( ; heightRegion > 0; pM += wplM, heightRegion--)
    {
	for (column = 0; column < wplRegion; column++)
	{
	    /* Count the number of on bits in the current model word */
	    word = pM[column];
	    ovrlpCnt += bitCount[word & 0xFF];
	    word >>= 8;
	    ovrlpCnt += bitCount[word & 0xFF];
	    word >>= 8;
	    ovrlpCnt += bitCount[word & 0xFF];
	    word >>= 8;
	    ovrlpCnt += bitCount[word];
	}

	/* Count the (special) last overlapping word in the row */
	word = pM[column] & rowLastWordMask;
	ovrlpCnt += bitCount[word & 0xFF];
	word >>= 8;
	ovrlpCnt += bitCount[word & 0xFF];
	word >>= 8;
	ovrlpCnt += bitCount[word & 0xFF];
	word >>= 8;
	ovrlpCnt += bitCount[word];
    }

    return ovrlpCnt;
}


/* PRIVATE */
/*****************************************************************************
 * Simple insertion sort
 *
 * Description:
 *  Sort in order of increasing values, with the minimum valued array
 *  element's index in the first position of the returned array.
 *
 * Results:
 *  order[0] as _index_ of smallest pArr value,
 *  order[1] as _index_ of 2nd smallest pArr value, ...
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *****************************************************************************/
static Int32 CDECL 
ip_orderByInsertSort (
    IPState    *pIPState,
    Int32      *pArr,  /* Array values to sort */
    Int32	l,     /* Leftmost element index (inclusive) to sort */
    Int32	r,     /* Rightmost element index (inclusive) to sort */
    Int32      *pOrder)  /* Ptr to resulting array of pArr's index values,
			* ordered nondecreasingly by pArr's values */
{
    PROLOGUE;
    Int32 i, j, t;
    
    /* Insertion sort */
    for (i = l; i < r; i++)
    {
	for (j = i+1; j <= r; j++)
	{
	    if (pArr[pOrder[i]] > pArr[pOrder[j]])
	    {
		/* Swap out of order elements */
		t = pOrder[i];
		pOrder[i] = pOrder[j];
		pOrder[j] = t;
	    }
	}
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/*****************************************************************************
 * Simple Quicksort
 *
 * Description:
 *  Sort in order of increasing values, with the minimum valued array
 *  element's index in the first position of the returned array.
 *
 * Results:
 *  order[0] as _index_ of smallest pArr value,
 *  order[1] as _index_ of 2nd smallest pArr value, ...
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *****************************************************************************/
static Int32 CDECL 
ip_orderByQuickSort (
    IPState    *pIPState,
    Int32      *pArr,  /* Array values to sort */
    Int32	l,     /* Leftmost element index (inclusive) to sort */
    Int32	r,     /* Rightmost element index (inclusive) to sort */
    Int32      *pOrder)  /* Ptr to resulting array of pArr's index values,
			* ordered nondecreasingly by pArr's values */
{
    PROLOGUE;
    Int32  i, j, k, pivotVal;

    /* Find a pivot element for Quicksort partition:
     *  make pivotVal the value of the larger of the leftmost two different
     *  values
     */
    for (pivotVal = pArr[pOrder[l]], i = l + 1; i <= r; i++)
    {
	if (pArr[pOrder[i]] > pivotVal)
	{
	    pivotVal = pArr[pOrder[i]];
	    break;
	}
	else if (pArr[pOrder[i]] < pivotVal)
	{
	    break;
	}
    }

    if (i <= r)
    {
	/* Partition pOrder[l]..pOrder[r] into two groups
	 * so pArr[pOrder[x]] < pivotVal are at the left and
	 * so pArr[pOrder[x]] >= pivotVal are on the right.
	 */
	
	i = l;  /* Left index moving right */
	j = r;  /* Right index moving left */

	do
	{
	    /* Swap */
	    k = pOrder[i];
	    pOrder[i] = pOrder[j];
	    pOrder[j] = k;

	    while (pArr[pOrder[i]] < pivotVal)
		i++;

	    while (pArr[pOrder[j]] >= pivotVal)
		j--;
	} while (i <= j);

	/* "i" is now the first (leftmost) index in the group on the right */

	/* Now recursively sort the two pieces of the partition */
	ip_orderByQuickSort (pIPState, pArr, l, i-1, pOrder);
	ip_orderByQuickSort (pIPState, pArr, i, r,   pOrder);
    }
    /* else all values were equal, do nothing */

    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * Order an array by Sorting
 *
 * Description:
 *  Sort in order of increasing values, with the minimum valued array
 *  element's index in the first position of the returned array.
 *  Use either insertion sorting or quicksort, depending on number elements.
 *
 * Results:
 *  order[0] as _index_ of smallest pArr value,
 *  order[1] as _index_ of 2nd smallest pArr value, ...
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *****************************************************************************/
Int32 CDECL 
i_orderBySort (
    IPState    *pIPState,
    Int32      *pArr,  /* Array values to sort */
    Int32	l,     /* Leftmost element index (inclusive) to sort */
    Int32	r,     /* Rightmost element index (inclusive) to sort */
    Int32      *pOrder)  /* Ptr to resulting array of pArr's index values,
			* ordered nondecreasingly by pArr's values */
{
    PROLOGUE;
    Int32  i;
    
    /* Initialize the array ordering */
    for (i = l; i <= r; i++)
	pOrder[i] = i;

    if (r-l < 100)
    {
	ip_orderByInsertSort (pIPState, pArr,l,r,pOrder);
    }
    else
    {
	Int32 j, tmp;
	
	/* Before we do a Quicksort, we want to make sure the values are
	 * randomly shuffled for finding good pivot values later */
	for (i = l; i < r; i++) {
	    /* Select a random index val j from range [i..r] */
	    j = (rand() % (r-i+1)) + i;

	    /* Swap order i,j */
	    tmp = pOrder[i];
	    pOrder[i] = pOrder[j];
	    pOrder[j] = tmp;
	}

	ip_orderByQuickSort (pIPState, pArr,l,r,pOrder);
    }

    NORMAL_RETURN;
}
    

/* PRIVATE */
/****************************************************************************
 * Compute the offsets to access image rows
 *
 * Description:
 *  In order to compare model rows in a better ordering computed in
 *  assignModelRowOrder(), we need to convert from a row ordering to word-per-
 *  line offsets to add to some image pointer.  Create these offsets.
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_nomem	    	couldn't allocate needed memory
 ***************************************************************************/
Int32 CDECL 
ip_rowMakeOffsetsFromOrder (
    IPState    *pIPState,
    Int32      *pRowOrder,  /* Ptr to model row order array */
    Int32      *pTmp,	    /* Ptr to same sized array as pRowOrder for
			     * temporary internal use, or NULL (will create
			     * and free array internally then) */
    Int32	nRows,	    /* The number of model rows, also the size of
			     * the two above arrays */
    Int32	wplImage,   /* Words per line in the image we want to convert
			     * to offsets to add to an image pointer to move
			     * from row to row in order in pRowOrder. */
    Int32      *pOffsetsImage)/* Ptr to offsets array that gets filled (not
			     * created!) herein */
{
    PROLOGUE;
    Int32  i, j;
    Int8   freeTmp = 0;

    if (!pTmp)
    {
	pTmp = (Int32 *)CALLOC(nRows, sizeof(Int32));
	if (!pTmp)
	    return ia_nomem;
	
	freeTmp = 1;
    }
    
    /* pOffsetsImage[x] = x * wplM */
    for (i = j = 0; i < nRows; j += wplImage, i++)
	pOffsetsImage[i] = j;
    
    /* Put in pTmp[x] array the value of pRowOrder[x] * wplM */
    for (i = 0; i < nRows; i++)
	pTmp[i] = pOffsetsImage[pRowOrder[i]];
    
    /* Put in pOffsetsImage[x] array the difference which will be what
     * we add to the current model/image row that was just compared to
     * get to the next model/image row, when bit comparing the model and
     * image later */
    for (i = 0; i < nRows-1; i++)
	pOffsetsImage[i] = pTmp[i] - pTmp[i+1];
    pOffsetsImage[nRows-1] = pTmp[nRows-1];

    if (freeTmp == 1)
	FREE(pTmp);

    NORMAL_RETURN;
}


#ifdef hDEBUG
#define BLOCK_PRINT 1
#else /*hDEBUG*/
#define BLOCK_PRINT 0
#endif /*hDEBUG*/

/* PRIVATE */
/****************************************************************************
 * Compute the order to access model rows
 *
 * Description:
 *  We will be comparing the image bitmap of this model with some other
 *  image bitmap later.  We would like to compare the model in a "smart"
 *  way, namely comparing the densest model row first, the second most
 *  dense second, etc.  Store the densest model row number in the last
 *  pOrder element, the second densest in the second to last pOrder element,
 *  etc.
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_nomem	    	couldn't allocate needed memory
 *
 * Notes:
 *  pOrder[0..Z] may all be the same row number, which indicates that the
 *  sparsest Z rows had no ON bits in them at all.
 ***************************************************************************/
Int32 CDECL 
ip_assignModelRowOrder (
    IPState    *pIPState,
    UInt8      *pS,	     /* Ptr to binary image's upper left corner,
			      * without frame */
    Int32	sBpl,	     /* Image bytes per line, including padding/frame*/
    Int32	sWidth,	     /* Image width and height, */
    Int32	sHeight,     /* excluding frame and padding */
    Int32	blockHeight, /* Number of rows per block, [2..sHeight] */
    Int32      *pTmp,	     /* Ptr to array, same size as pOrder, used
			      * internally for ordering */
    Int32      *pOrder,	     /* Ptr to model row access order array, which
			      * gets filled in in this routine and contains
			      * sHeight elements */
    Int32      *pNptsM)      /* Ptr to returned number of ON bits in the
			      * model */
{
    PROLOGUE;
    Int32       i, tmp, cnt = 0, nBlocks = 0;
    Int32	lineCount;

    /* Count the number of ON bits in each image row and store them */
    for (i = 0; i < sHeight; i++)
    {
	ERR_PASS(ip_countBitsLine (pIPState, pS, (sWidth + 0x1F) >> 5,
				   0xFFFFFFFF, 0xFFFFFFFF,
				   (const UInt8 *) &(bitCount[0]),
				   &lineCount));
	pTmp[i] = lineCount;
	cnt += lineCount;
	pS += sBpl;
    }
    /* The number of ON bits/points in the entire model */
    *pNptsM = (UInt32)cnt;

    /* Get a pseudo-random seed to start pseudo-random number generation
     * (Used in Quicksorting) */
    srand((UInt32)cnt);

    /* Sort each block, making the densest row the last row in each block */
    for (i = 0; i < sHeight; i += blockHeight) {
	/* Sort the number of ON bits per row array, in increasing order */
	tmp = MIN(i+blockHeight-1,sHeight-1);
	i_orderBySort (pIPState, pTmp, i, tmp, pOrder);
	nBlocks++;
    }

    /* Now we want to order the blocks, in turn, with the densest block
     * last in the list */
    if (nBlocks > 1)
    {
	Int32  *pBlockSums, *pBlockOrder, *pTmpBlock, *pCurBlkOrdr, j, block;

	pBlockSums = (Int32 *)CALLOC(3*nBlocks+blockHeight, sizeof(Int32));
	pBlockOrder= pBlockSums + nBlocks;  /* Array [0..nBlocks-1] */
	pCurBlkOrdr= pBlockOrder + nBlocks; /* Array [0..nBlocks-1] */
	pTmpBlock  = pCurBlkOrdr + nBlocks; /* Array [0..blockHeight-1] */

	if (!pBlockSums)
	    return ia_nomem;
	
	for (block = i = 0; i < sHeight; block++, i += blockHeight)
	{
	    cnt = 0;
	    for (j = i; j < MIN(i+blockHeight, sHeight); j++)
		cnt += pTmp[pOrder[j]];

	    /* The sum of the number of ON bits in this block */
	    pBlockSums[block] = cnt;
	    pCurBlkOrdr[block] = block;
	}

	/* How many rows were included in the last block */
	cnt = sHeight % blockHeight;
	if (cnt != 0)
	    /* Scale the last block that is only partially full of rows to what
	     * it would be if it was full */
	    pBlockSums[nBlocks-1] = pBlockSums[nBlocks-1] * blockHeight / cnt;

	/* Sort the block sums, making the densest block the last block in
	 * the order */
	tmp = (cnt == 0 ? nBlocks-1 : nBlocks-2);
	i_orderBySort (pIPState, pBlockSums, 0, tmp, pBlockOrder);

	
	/* Now take the pOrder[] array, arranged in blocks, and swap
	 * the array around block by block until we have all the FULL
	 * blocks in their proper order.
	 * NOTE that this scheme moves blocks efficiently, if a block
	 * is moved, it is always moved to its final position (other
	 * than "paging" in and out to a single temporary block
	 * once during a block moving "round") */
	if (cnt != 0)
	    nBlocks--;
	
	for (block = 0; block < nBlocks; block++)
	{
	    Int32  toBlock, fromBlock;
	    
	    if (pCurBlkOrdr[block] != pBlockOrder[block])
	    {
		/* Start another "round" of block moving to make
		 * the order correct up to block "block" */
	       
		/* tmp will carry the block number that pTmpBlock
		 * contains */
		toBlock = tmp = pCurBlkOrdr[block];
		fromBlock = pBlockOrder[block];

		/* Move this block to the pTmpBlock */
		i = blockHeight * toBlock;
		for (cnt = 0; cnt < blockHeight; i++, cnt++)
		    pTmpBlock[cnt] = pOrder[i];

		while (fromBlock != tmp)
		{
		    /* Block to move into (vacated above) */
		    i = blockHeight * toBlock;
		    
		    /* Block to move from (into vacated block) */
		    j = blockHeight * fromBlock;

		    /* Move block in (vacating another block) */
		    for (cnt = i + blockHeight; i < cnt; i++, j++)
			pOrder[i] = pOrder[j];

		    /* The block we moved into is now the correct block
		     * according to the sorted order */
		    pCurBlkOrdr[toBlock] = pBlockOrder[toBlock];
		    
		    toBlock = fromBlock;
		    fromBlock = pBlockOrder[fromBlock];
		}

		/* Now fromBlock is the block we initially moved
		 * into pTmpBlock */

		/* Block to move into (vacated above) */
		i = blockHeight * toBlock;
		    
		/* Block to move from is pTmpBlock */

		/* Move saved out pTmpBlock back in */
		for (cnt = 0; cnt < blockHeight; i++, cnt++)
		    pOrder[i] = pTmpBlock[cnt];

		/* The block we moved into is now the correct block
		 * according to the sorted order */
		pCurBlkOrdr[toBlock] = pBlockOrder[toBlock];
	    }
	}

	
	/* Finally, move the pOrder[] array until we have the last
	 * partially full block in its correct place */
	cnt = sHeight % blockHeight;
	if (cnt != 0)
	{
	    /* Find the block location where the partially full block should
	     * move to.  This block and any block denser than this block
	     * need to then shift "cnt" rows "down" the pOrder array.
	     */
	    for (block = 0; block < nBlocks; block++)
		if (pBlockSums[nBlocks] < pBlockSums[pBlockOrder[block]])
		    break;

	    if (block < nBlocks)
	    {
		/* Save out the last partial block */
		j = blockHeight * nBlocks;
		for (i = 0; i < cnt; i++, j++)
		    pTmpBlock[i] = pOrder[j];
		
		/* Shift the pOrder array "cnt" rows down. */
		j = blockHeight * block + cnt;
		for (i = sHeight-1; i >= j; i--)
		    pOrder[i] = pOrder[i-cnt];

		/* Now insert/write in the partial block */
		j -= cnt;
		for (i = 0; i < cnt; j++, i++)
		    pOrder[j] = pTmpBlock[i];
	    }
	    
	    /* Restore this to the proper number of blocks */
	    nBlocks++;
	}

#ifdef hDEBUG

	if (1)
	{
	    Int32  prevBlockSum = 0, prevRowSum;

	    cnt = sHeight % blockHeight;
	    if (cnt == 0)
		/* This means that the image ended on a full block, thus
		 * there is not the single partial height block floating
		 * in the blocks somewhere. */
		block = nBlocks-1;

	    /* Now make the pCurBlkOrdr[] array store the index location in
	     * pOrder where each block begins */
	    for (i = 0; i <= block; i++)
		pCurBlkOrdr[i] = blockHeight*i;
	    for ( ; i < nBlocks; i++)
		/* All of these blocks are after the partial block, so adjust
		 * their start indices */
		pCurBlkOrdr[i] = blockHeight*i - (blockHeight - cnt);

#if BLOCK_PRINT
	    fprintf(stderr, "Check note: the last block should be the densest, and the last row of each\n   block should be the densest row in each block\n");
	    fprintf(stderr,
		    "Each row is entered as (image row #: # ON bits)\n");
#endif /*BLOCK_PRINT*/
	    
	    for (i = 0; i < nBlocks; i++)
	    {
#if BLOCK_PRINT
		fprintf(stderr, "Block #%3ld: ", i);
#endif /*BLOCK_PRINT*/

		/* This is first index for the next block,
		 * marking the end of this block */
		tmp = (i < nBlocks-1 ? pCurBlkOrdr[i+1] : sHeight);
		prevRowSum = cnt = 0;

		for (j = pCurBlkOrdr[i]; j < tmp; j++)
		{

#if BLOCK_PRINT
		    fprintf(stderr, "(%3ld: %4ld) ", pOrder[j],
			    pTmp[pOrder[j]]);
#endif /*BLOCK_PRINT*/
		    
		    hCHECK(pTmp[pOrder[j]] < prevRowSum,
			   "rows not in correct (increasing) order");
		    prevRowSum = pTmp[pOrder[j]];
		    
		    cnt += pTmp[pOrder[j]];
		}
		
		if (tmp - pCurBlkOrdr[i] != blockHeight)
		{
		    /* This means this block is the single partially
		     * full block in the array of blocks, and this
		     * sum was scaled previously so we need to
		     * scale it here...*/
		    cnt = cnt * blockHeight / (sHeight % blockHeight);

#if BLOCK_PRINT
		    fprintf(stderr,
			    "\n  (Scaled) Block #%3ld total ON bits: *[%ld]*\n",
			    i, cnt);
#endif /*BLOCK_PRINT*/
		    
		}

		else
		{
#if BLOCK_PRINT
		    fprintf(stderr,
			    "\n  Block #%3ld total ON bits: *[%ld]*\n",
			    i, cnt);
#endif /*BLOCK_PRINT*/
		}
		
		
		hCHECK(cnt < prevBlockSum,
		       "blocks not in correct (increasing) order");
		prevBlockSum = cnt;
	    }

#if BLOCK_PRINT
	    fprintf(stderr,
		    "Each block total should also appear in this list: {");
	    for (i = 0; i < nBlocks; i++)
		fprintf(stderr, "%ld ", pBlockSums[i]);
	    fprintf(stderr, "}\n\n");
#endif /*BLOCK_PRINT*/
	    
	}
	
	
#endif /*hDEBUG*/
	    
	FREE(pBlockSums);
    }
    

    if (pOrder[sHeight-1] == 0)
    {
	/* If the most dense row is the very first row in the image, we must
	 * switch the two most dense rows with each other, because we later
	 * compute offsets by the difference from one row to the next and make
	 * the difference of 0 to be a flag indicating that the bitmap row
	 * is all zeros.  Thus, if row 0 is the first row, its offset will be
	 * 0 (because we begin at the top of the model and move from there and
	 *  the top is where we want to be in this case, hence a 0 offset)
	 * but this would make the 0 offset flag invalid.  Thus, the solution
	 * is to permute the two densest rows so this can't happen.
	 */
	tmp = pOrder[sHeight-1];
	pOrder[sHeight-1] = pOrder[sHeight-2];
	pOrder[sHeight-2] = tmp;
    }

    
    
    /* Now find all locations where pTmp[X] is 0, rows where there were no
     * ON bits in the entire row.  Make the corresponding pOrder[X] be same
     * as the previous pOrder[X+1] in those instances.
     */

    /* We know the row correpsonding to this index (the first row to search --
     * and if not the densest, then the second most dense) cannot be the
     * very first row of the image from above */
    cnt = pOrder[sHeight-1];
    
    for (tmp = sHeight-2; tmp >= 0; tmp--)
	if (pTmp[pOrder[tmp]] == 0)
	    /* No ON bits in this row, make it the index of a previous nonzero
	     * row */
	    pOrder[tmp] = cnt;
	else
	    /* Get the newest previous nonzero row index */
	    cnt = pOrder[tmp];
    
    NORMAL_RETURN;
}


/* PRIVATE */
/****************************************************************************
 * getModelDimens():
 *  	Given the frameless width and height of the source image for the
 *  	model set, return the words-per-line and number of rows for the
 *  	resulting model set.
 ***************************************************************************/
static void CDECL 
ip_getModelDimens (
    IPState    *pIPState,
    Int32	sWidth,
    Int32	sHeight,
    Int32     *pWplModel,
    Int32     *pHeightModel)
{
    /* Keep compiler happy */
    pIPState = pIPState;

    /* The words per row for each of the model bitmaps will be the same as
     * the unframed passed in image, plus one to accomodate all the shifted
     * variants.
     * BUG: ddavies 4/2/97; You can't calculate the width of an image from
     * its bytes/line.  This is especially true when 64 bit operations
     * are allowed.  This procedure is changed to pass in sWidth instead
     * of sBpl and calculate WplModel from that.
     */
    *pWplModel = ((sWidth + 31) / 32) + 1;
    *pHeightModel = sHeight;
}


/* PRIVATE */
/****************************************************************************
 * derefModelSetVoid():
 *
 * 	Given a previously allocated chunk of memory (referred to herein as
 *	the 'set memory chunk') pointed to by this here void *, cast it to the
 *	properly set up modelSet pointer.  Because the modelSet pointer has
 *	pointers in it to other places in this set memory chunk, garbage
 *	collection normally can cause problems.  Thus, keeping this in
 *	mind (if the set memory chunk has somehow moved) still return the
 *	properly set up cast modelSet struct pointer.
 *  	Returns NULL on error.
 *****************************************************************************/
modelSet * CDECL 
ip_derefModelSetVoid (
    IPState    *pIPState,
    void       *pM)
{
    PROLOGUE;
    modelSet *set;

    /* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Point to the set struct portion of the model set memory chunk */ 
    set = (modelSet *) pM;

    /* This is unexpected.  It may mean we haven't initially set the
     * model pointers yet to locations in the set memory chunk.
     * More likely, this could mean that the pointer to void *pm has
     * changed since when we last looked at it on account of garbage
     * collection!
     * Either way, we need to reassign the pointers in this
     * modelSet to point to the appropriate areas in the set memory.
     */
    switch (set->nBitmaps)
    {
    case 32:
    case 1:
	if (set->pOffsetsM != NULL &&
	    set->pOffsetsM != (Int32 *)(set + 1))
	    set->pOffsetsM = (Int32 *)(set + 1);
	break;
    case 0:
	if (set->pX != (UInt16 *)(set + 1))
	    set->pX = (UInt16 *)(set + 1);
	break;
    default:
	return (modelSet *)NULL;
    }

    /* The pointers are set in the block properly, return the model set
     * pointer
     */
    return set;
}


/* PUBLIC */
/****************************************************************************
 * Return space needed to make model set from image
 *
 * Description:::
 *  Given the dimensions of the image to make a model set from, determine
 *  the space that will be needed in an allocator call.
 *  The model set will be filled in in i_makeModelSet()
 *
 * Returns:::
 *  ia_successful	invocation done and okay
 *  ia_imageTooWide	too big to make a model set with current implementation
 *  ia_invalidParm 	bad model set representation
 *
 * Notes:::
 *  pBitmapSize::  When "nBitmaps" is 32, client will have to allocate space
 *                 for 32 binary bitmaps (with pointers) when calling
 *                 i_makeModelSet().  This will point to the size that each
 *                 bitmap must be, in bytes.
 *  blockHeight: (should always be 0 when creating a _B_ model set)
 *   If 0, it means don't order the model rows, and later always sequentially
 *     compare the model row by row. (slowest)
 *   If >= sHeight, it means order all the model rows densest first.
 *   Otherwise, divide the model into blocks of blockHeight rows.  Within each
 *     block, order the rows densest first.  Compare the densest blocks first.
 *     This is to help avoid cache misses.  Thus, blockHeight should be set
 *     with the cache size and the width of the model bitmap AND the imagery
 *     bitmap in mind.
 *  This routine requires/assumes the standard frame.
 *****************************************************************************/
Int32 CDECL 
i_getModelSetSize (
    IPState    *pIPState,
    UInt8      *pS,	     /* Ptr to upper left corner of image to make
			      * model set from, including frame */
    Int32	sBpl,	     /* Number of bytes per line, including padding
			      * and frame */
    Int32	sWidth,	     /* Width and height of image, including */
    Int32	sHeight,     /* frame, excluding padding */
    Int32	blockHeight, /* Indicates how to divide up and order the model
			      * bitmap (see Notes above) */
    UInt8	nBitmaps,    /* 0, 1, 32 (see i_makeModelSet())*/
    Int32      *pBitmapSize, /* Ptr to returned #bytes needed for each bitmap
			      * when "nBitmaps"==32;  O.w. ptr to 0 */
    Int32      *pModelSetSize)/* Ptr to returned value of model set size */
{
    PROLOGUE;
    Int32	wplModel, heightModel;

    switch (nBitmaps)
    {
    case 32:
	break;
    case 0:
	__error__ = ip_getModelSetSizePtList (pIPState, pS, sBpl,
					      sWidth, sHeight,
					      pModelSetSize);
	if (__error__ == ia_successful)
	    *pBitmapSize = 0;
	NORMAL_RETURN;
	break;
    case 1:
	__error__ = ip_getModelSetSizeOnePtr (pIPState, sHeight, blockHeight,
					      pModelSetSize);
	if (__error__ == ia_successful)
	    *pBitmapSize = 0;
	NORMAL_RETURN;
	break;
    default:
	return ia_invalidParm;
    }

    ip_getModelDimens (pIPState, sWidth - 2*cFrameBits, sHeight - 2*cFrameBits,
		       &wplModel, &heightModel);

    /* Each of the 32 bitmaps will then be in size:
     *   (4 bytes per word) * (#words in line) * (height)
     */
    *pBitmapSize = (wplModel << 2) * heightModel;
    

    /* The model set will be in a chunk of memory comprised of one model
     * set structure and the bits for the WORD_SIZE models (in words)
     * and an array of row ordering info (if blockHeight != 0)
     */
    *pModelSetSize = sizeof(modelSet)
	+ (blockHeight > 0 ? 3 * heightModel * sizeof(Int32) : 0);

    /* WARNING!  Since I am using a chunk of memory with a struct casted
     * to the top of the chunk with pointers that point to areas further
     * in the memory chunk (and I always use this kind of usage when using
     * a memory chunk), I have to be careful of word/double boundaries.
     * It so happens here that the (struct casted top of the chunk) + 1
     * will ALWAYS start on a word boundary.
     * Why?  Well because sizeof(any struct with a long field (or any field
     * that takes up 4 bytes)) will ALWAYS return a multiple of 4.
     * We know MALLOC returns a double word aligned chunk of memory
     * (it has to!! think about this:  ptr = (double *)MALLOC(sizeof(double)))
     * so a ptr to the casted struct plus 1 will always be word aligned.
     * Thus, so long as I organize my pointers in the struct to fall in the
     * memory chunk largest pointer FIRST (sizeof(pointer) == 4), I am
     * guaranteed to not have any memory fragmentation and the size
     * computed above is acurrate AND the absolute least amount of space
     * I could possible use!
     */
    
    NORMAL_RETURN;
}


/* PUBLIC */
/****************************************************************************
 * Make a model set from an image bitmap
 *
 * Description:
 *  Given an image and its dimensions, make a model set from it.
 *  This model set will be used in the Hausdorff comparator functions.
 *  NOTICE it may be desirable for the image to have the ADDITIONAL
 *  'dilateRadius' zeroed border around the image.
 *  (see note on 'dilateRadius' at top of ipshared/i_hc32ms.c)
 *
 * Returns:
 *  ia_successful	invocation done and okay
 *  ia_nomem    	could not allocate required space
 *  ia_imageTooWide	too big to make a model set with current implementation
 *  ia_invalidParm    	bad parameter passed in
 *  ia_internal	    	something unexpected happened that shouldn't
 *
 * Notes:
 *  pM:
 *   *pM != NULL, the model set has been allocated by a client and it will
 *   simply get filled in here.
 *  nBitmaps, ppBitmaps:
 *   If 0, means make the model set data a list of points
 *     (instead of using bitmap(s) in comparisons.  Pass in NULL for ppBitmaps)
 *   If 1, means use the ptr to pS image excluding frame and padding later
 *     (ppBitmaps should be an array of 1 ptr to the upper left corner of
 *      the pS image excluding frame and padding, but will be ignored in
 *      this routine).
 *   If 32, means copy the pS image 32 times (one copy per right shift 0..31),
 *     filling the ppBitmaps bitmaps.
 *     Each ppBitmaps ptr must point to an empty binary bitmap of the size
 *     returned in i_getModelSetSize() param "pBitmapSize".
 *     Bitmaps are unframed;  each row begins at a new longword.
 *     Upon completion ppBitmaps[0] will point to the unshifted image bitmap,
 *     ppBitmaps[1] will point to the image right shifted 1 bitmap, etc...
 *  blockHeight: (should always be 0 when creating a _B_ model set)
 *   If 0, it means don't order the model rows, and later always sequentially
 *     compare the model row by row. (slowest)
 *   If >= sHeight, it means order all the model rows densest first.
 *   Otherwise, divide the model into blocks of blockHeight rows.  Within each
 *     block, order the rows densest first.  Compare the densest blocks first.
 *     This is to help avoid cache misses.  Thus, blockHeight should be set
 *     with the cache size and the width of the model bitmap AND the imagery
 *     bitmap in mind.
 *  This routine requires/assumes pS has the standard frame.
 *****************************************************************************/
Int32 CDECL 
i_makeModelSet (
    IPState    *pIPState,
    UInt8      *pS,	     /* Ptr to upper left corner of image to make
			      * model set from, including frame */
    Int32	sBpl,	     /* Number of bytes per line, including padding
			      * and frame */
    Int32	sWidth,	     /* Width and height of image, including */
    Int32	sHeight,     /* frame, excluding padding */
    Int32	blockHeight, /* Indicates how to divide up and order the model
			      * bitmap (see Notes above) */
    UInt8     **ppBitmaps,   /* If nBitmaps == 32, array of 32 ptrs to upper
			      * left corner of bitmaps (without frame or
			      * padding).  O.w. it is ignored here. */
    UInt8	nBitmaps,    /* 0, 1, or 32 */
    void       *pM)           /* Ptr to internal representation of model
			      * set memory chunk (see Notes above) */
{
    PROLOGUE;
    UInt32    *pCurBitmap, *pPrevBitmap, *pSWord;
    Int32     wordsPerModel, i, j;
    modelSet  *set;
    

    if (!pM)
	return ia_invalidParm;

    switch (nBitmaps)
    {
    case 32:
	break;
    case 0:
	return (ip_makeModelSetPtList (pIPState, pS, sBpl,
				       sWidth, sHeight, pM));
	break;
    case 1:
	return (ip_makeModelSetOnePtr (pIPState, pS, sBpl,
				       sWidth, sHeight, blockHeight,
				       pM));	
	break;
    default:
	return ia_invalidParm;
    }
    
    /* Make dimens unframed */
    sWidth -= 2*cFrameBits;
    sHeight -= 2*cFrameBits;

    /* Point to the set struct portion of the model set memory chunk
     * and fill in the non-pointer fields of this struct portion */ 
    set = (modelSet *) pM;
    set->nBitmaps = 32;
    set->widthM = sWidth;
    ip_getModelDimens (pIPState, sWidth, sHeight,
		       &(set->wplM), &(set->heightM));

    /* This just to avoid another level of indirection below */
    sHeight = set->heightM;

    /* This is sneaky and rotten and probably unprofessional, but this
     * is how we'll dereference the rest of the memory chunk AND
     * assign all model pointers to locations in the memory chunk.
     */
    set->pOffsetsM = (blockHeight > 0 ? (Int32 *)set : NULL);
    set = ip_derefModelSetVoid (pIPState, (void *)pM);
    if (set == NULL)
	return ia_internal;

    /* Skip binary frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pSWord = (UInt32 *)pS;
    
    /* Make the 0th bitmap from the image.  Note that we can't
     * assume that the entire word following the rightmost
     * word containing source bits on a line is filled with 0's.
     * So, we just fill the extra word on each line with 0.
     */
    pCurBitmap = (UInt32 *)ppBitmaps[0];
    for (j = 0; j < sHeight; j++)
    {
	for (i = 0; i < set->wplM-1; i++)
	    pCurBitmap[i] = pSWord[i];
	pCurBitmap[i] = 0;		/* fill extra word on each line */

	/* Advance the image pointers to the next rows */
	pSWord += sBpl >> 2;
	pCurBitmap += set->wplM;
    }

    /* Make the 1st..31st bitmaps */
    wordsPerModel = set->wplM * sHeight;

    for (i = 1; i < WORD_SIZE; i++)
    {
	/* Point to the previous made bitmap's last word */
	pPrevBitmap = (UInt32 *)ppBitmaps[i-1] + (wordsPerModel-1);
	
	/* Point to the to be filled bitmap's last word */
	pCurBitmap  = (UInt32 *)ppBitmaps[i] + (wordsPerModel-1);
	
	/* Make the bitmap "backwards", last bitmap word first */
	while (pCurBitmap > (UInt32 *)ppBitmaps[i])
	{
	    *pCurBitmap = (*(pPrevBitmap--) >> 1);
	    
	    /* Now carry the bit (or do nothing if no carry) */
	    *(pCurBitmap--) |= (*pPrevBitmap << WORD_SIZE_MINUS1);
	}
	/* Make the first shifted word of the new bitmap now, and we
	 * will be done (of course, there's nothing to carry) */
	*pCurBitmap = (*pPrevBitmap >> 1);
    }

    
    if (blockHeight > 0)
    {
	/* When blockHeight > 0, it means to set up the model
	 * to access the rows when comparing later in order of densest row
	 * first.
	 * The set memory chunk has 3 more pieces, all are Int32s arrays
	 * the length of the number model rows.
	 * The first piece is where pOffsetsM points to and is the model
	 * bitmap offsets to various rows.
	 * The second piece is where the imagery bitmap offsets to various
	 * rows is stored, and gets filled in later when a search is begun.
	 * The third piece is the array that contains the order of the
	 * actual row accesses from which the above offsets are created.
	 * For example, this array would be {0,3,1,2} if the model was
	 * 4 rows high, its second row was the densenst, the first row
	 * the next most dense, the third row the next most dense, and the
	 * zeroth row was the least dense.
	 */
	Int32 *pTmp;
	
	pTmp = (Int32 *)CALLOC(sHeight, sizeof(Int32));
	if (!pTmp)
	    return ia_nomem;

	/* (set->pOffsetsM + 2 * sHeight) is where in the set memory chunk
	 *  the row order information is stored */

	if (ip_assignModelRowOrder (pIPState, pS, sBpl, sWidth, sHeight,
				    blockHeight, pTmp,
				    set->pOffsetsM + 2*sHeight,
				    &(set->nptsM)) != ia_successful)
	    return ia_nomem;
	
	if ((__error__ = ip_rowMakeOffsetsFromOrder (pIPState,
						   set->pOffsetsM + 2*sHeight,
						   pTmp,
						   sHeight,
						   set->wplM,
						   set->pOffsetsM))
	    != ia_successful)
	    NORMAL_RETURN;
	    
	FREE(pTmp);
    }
    else
    {
	set->pOffsetsM = NULL;
	
	/* The number of points (bits that are 1/on) in each model */
	set->nptsM = ip_countModelPoints (pIPState, (UInt32 *)ppBitmaps[0],
					  wordsPerModel);
    }
    
    NORMAL_RETURN;
}


/* PRIVATE */
/****************************************************************************
 * Hausdorff compares model and imagery
 *
 * Description:
 *  Take the model and the imagery and compare them to see if they match, at
 *  the appropriate passed in offset of the model over the imagery.  Do this by
 *  taking the model bits and ANDing them with the imagery bits.
 *
 * Returns:
 *  The number of model bits ON in the region.
 *
 * Notes:
 *  minAndedCnt is used for early termination.  It could be calculated inside
 *  this routine, but we unrolled it for speed.
 *****************************************************************************/
Int32 CDECL 
ip_hausdorffCompareOvrlp (
    IPState    *pIPState,
    void       *pModelData,	/* Ptr to first model bitmap in words, of the
				 * 32 in the set, cast to a void * */
    Int32	wplM,		/* # words per line, including padding*/
    Int32	widthM,		/* # model columns (excluding padding/frame) */
    Int32	heightM,	/* # model rows (excluding padding/frame) */
    Int32	nptsM,		/* # on bits in entire model bitmap */
    Int32      *pOffsetsM,	/* Ptr to ordered offsets table */
    Int32	x,		/* Offset coordinates to place the model's */
    Int32	y,		/* upper left corner w.r.t. the upper left
				 * corner of the imagery */
    UInt32     *pI,		/* Ptr to upper left corner of imagery,
				 * excluding frame */
    Int32	wplI,		/* # words per line, including padding/frame */
    Int32	widthI,		/* # imagery columns(excluding padding/frame)*/
    Int32	heightI,	/* # imagery rows (excluding padding/frame)*/
    Float64	minPercent,	/* The imagery and model must correlate with
				 * a Hausdorff percent of at least this much
				 * to constitute a match, [0..1] */
    Int32	minAndedCnt,	/* Pass this in as CEIL(nptsM * minPercent) */
    Int32      *pAndedCnt)  	/* Return the number of total model bits
				 * that AND over imagery bits in the overlap
				 * ( >= minAndedCnt or -1) */
{
    PROLOGUE;
    UInt32  *pM = (UInt32 *)pModelData;
    Int32    wplOverlap, heightOverlap;
    UInt32   rowLastWordMask;
#ifdef pbmDEBUG
    DumpPBM("HD_Model.pbm", (UInt8 *)pM, wplM << 2, widthM + MAX(x,0), heightM);
#endif /* pbmDEBUG */

    /* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;
    widthM = widthM;
    pOffsetsM = pOffsetsM;


    if (ip_getOverlapDimens (pIPState, &pM, wplM,
			     heightM, x >> WORD_OFFSET, y, &pI,
			     wplI, heightI,
			     &wplOverlap, &heightOverlap))
    {
	/* Means no portion of the imagery and model overlapped! */
	*pAndedCnt = -1;
	return nptsM;
    }
    
    /* If the _last_ word of an imagery row is overlapping some portion of the
     * model, we have to (potentially) mask off the extra model word bits that
     * would overhang the natural imagery's end, because the imagery is of
     * course, padded to a full word.
     */
    rowLastWordMask = ((MAX(0, x >> WORD_OFFSET) + wplOverlap) >=
		       (wplI - 2*cFrameWords) ?
		       myLeftMask[widthI & WORD_SIZE_MINUS1] :
		       myLeftMask[0] /* The whole word */);

    /* Count the number of total model points that are "on" in this
     * overlapping area since we aren't trying to match the whole model */
    nptsM = ip_countOverlap (pIPState, pM,
			     heightOverlap,
			     wplOverlap-1,
			     rowLastWordMask,
			     wplM);
    
    /* Rescale the minAndedCnt appropriately */
    minAndedCnt = (Int32)(ceil(minPercent * nptsM));

    ip_correlate (pIPState, pI,
		  pM,
		  heightOverlap,
		  wplOverlap-1,
		  rowLastWordMask,
		  wplI,
		  wplM,
		  (const UInt8 *) &(bitCount[0]),
		  nptsM - minAndedCnt,
		  pAndedCnt);
    
    if (nptsM - minAndedCnt < *pAndedCnt)
	/* Means the above routine terminated early */
	*pAndedCnt = -1;
    else 
	/* NOTE that *pAndedCnt is actually the number of MISSES, so make it
	 * into what it should be, the number of not MISSES */
	*pAndedCnt = nptsM - *pAndedCnt;

    return (nptsM); /* The number of ON points in the model */
}


/* PRIVATE */
/****************************************************************************
 * Hausdorff compares model and imagery
 *
 * Description:
 *  Take the model and the imagery and compare them to see if they match, at
 *  the appropriate passed in offset of the model over the imagery.  Do this by
 *  taking the model bits and ANDing them with the imagery bits.
 *
 * Returns:
 *  The number of model bits ON in the entire model.
 *
 * Notes:
 *  minAndedCnt is used for early termination.  It could be calculated inside
 *  this routine, but we unrolled it for speed.
 *****************************************************************************/
Int32 CDECL 
ip_hausdorffCompare (
    IPState    *pIPState,
    void       *pModelData,	/* Ptr to first model bitmap in words, of the
				 * 32 in the set, cast to a void * */
    Int32	wplM,		/* # words per line, including padding*/
    Int32	widthM,		/* # model columns (excluding padding/frame) */
    Int32	heightM,	/* # model rows (excluding padding/frame) */
    Int32	nptsM,		/* # on bits in entire model bitmap */
    Int32      *pOffsetsM,	/* Ptr to ordered offsets table */
    Int32	x,		/* Offset coordinates to place the model's */
    Int32	y,		/* upper left corner w.r.t. the upper left
				 * corner of the imagery */
    UInt32     *pI,		/* Ptr to upper left corner of imagery,
				 * excluding frame */
    Int32	wplI,		/* # words per line, including padding/frame */
    Int32	widthI,		/* # imagery columns(excluding padding/frame)*/
    Int32	heightI,	/* # imagery rows (excluding padding/frame)*/
    Float64	minPercent,	/* The imagery and model must correlate with
				 * a Hausdorff percent of at least this much
				 * to constitute a match, [0..1] */
    Int32	minAndedCnt,	/* Pass this in as CEIL(nptsM * minPercent) */
    Int32      *pAndedCnt)  	/* Return the number of total model bits
				 * that AND over imagery bits in the overlap
				 * ( >= minAndedCnt or -1) */
{
    PROLOGUE;
    UInt32  *pM = (UInt32 *)pModelData;
    Int32    wplOverlap, heightOverlap;
    UInt32   rowLastWordMask;
#ifdef pbmDEBUG
    DumpPBM("HD_Model.pbm", (UInt8 *)pM, wplM << 2, widthM + MAX(x,0), heightM);
#endif /* pbmDEBUG */

    /* Keep compiler happy */
    minPercent = minPercent;
    pIPState = pIPState;
    __error__ = __error__;

    if (ip_getOverlapDimens (pIPState, &pM, wplM, heightM, x >> WORD_OFFSET, y,
			     &pI, wplI, heightI,
			     &wplOverlap, &heightOverlap) || nptsM == 0)
    {
	/* Means no portion of the imagery and model overlapped! */
	*pAndedCnt = -1; 
	return nptsM;
    }

    /* If the _last_ word of an imagery row is overlapping some portion of the
     * model, we have to (potentially) mask off the extra model word bits that
     * would overhang the natural imagery's end, because the imagery is of
     * course, padded to a full word. */
    rowLastWordMask = ((MAX(0, x >> WORD_OFFSET) + wplOverlap) >=
		       (wplI - (cFrameWords << 1)) ?
		       myLeftMask[widthI & WORD_SIZE_MINUS1] :
		       myLeftMask[0] /* The whole word */);


    if (pOffsetsM && heightOverlap == heightM)
    {
	ip_correlateOrdered (pIPState, pI,
			     pM,
			     heightM,
			     wplOverlap-1,
			     rowLastWordMask,
			     pOffsetsM + heightM,
			     pOffsetsM,
			     (const UInt8 *) &(bitCount[0]),
			     nptsM - minAndedCnt,
			     pAndedCnt);
    }
    else
    {
	ip_correlate (pIPState, pI,
		      pM,
		      heightOverlap,
		      wplOverlap-1,
		      rowLastWordMask,
		      wplI,
		      wplM,
		      (const UInt8 *) &(bitCount[0]),
		      nptsM - minAndedCnt,
		      pAndedCnt);
    }

    if (nptsM - minAndedCnt < *pAndedCnt)
	/* Means the above routine terminated early */
	*pAndedCnt = -1;
    else
    {
	/* NOTE that *pAndedCnt is actually be the number of MISSES, so make it
	 * into what it should be, the number of not MISSES */

	if ((heightOverlap == heightM) && (x >= 0) && (x + widthM <= widthI))
	{
	    /* The entire model fits in the image, the number of ANDs is thus
	     * the number of total model points minus the number of MISSes*/
	    *pAndedCnt = nptsM - *pAndedCnt;
	}
	else
	{
	    /* The number of ANDs is the number of model points in the
	     * overlapping region minus the number of MISSes */
	    *pAndedCnt *= -1;
	    *pAndedCnt += ip_countOverlap (pIPState, pM,
					   heightOverlap,
					   wplOverlap-1,
					   rowLastWordMask,
					   wplM);
	}
    }
    
    return (nptsM); /* The number of ON points in the model */
}


/*

$Log: i_hc32ms.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.18  1998/01/13 23:48:21  ddavies
i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.17  1997/04/15 19:34:46  ddavies
i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v

Fixed bug in ip_getModelDimens.  It was attempting to calculate the
bytes/line of an image from its width.  That doesn't work, especially
on MMX machines.
Ooops, I meant to say that it tried to calculate the width from the
bytes/line, which also doesn't work.  We now pass width as an argument,
replacing the bytes/line arg.
Also fix bug used when filling 0th bitmap in model set.  The old code
had assumed that the word to the right of the rightmost pixel in the
input image was filled entirely with 0's.  This often not true, it's
usually right frame plus uninitialized pad.

Revision 1.16  1996/02/29 03:03:44  ejaquith
ejaquith on Wed Feb 28 19:01:18 PST 1996

i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v

Added parameter pBitmapSize to i_getModelSize() to tell client how
many bytes each bitmap must be in the 32-bitmap model case (DAE request)
This is a much better way of doing things, and less error-prone.
Updated comments in i_getModelSetSize() (oops that first routine reference
is missing "Set") and i_makeModelSet() to support this change.
Made ip_getModelDimens() a static private routine since no one else calls it.

Revision 1.15  1995/12/13 02:44:27  ejaquith
i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.14  1995/09/14  18:59:43  ejaquith
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * CommonCheck: Converted all routines that were private in files to be of form
 * ip_*(), which they weren't.  No functionality/noticeable changes.
 *
 * Revision 1.13  1995/04/22  20:42:48  ejaquith
 * ejaquith on Sat Apr 22 13:03:38 PDT 1995
 *
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * New Hausdorff module with MASSIVE restructuring and changes.
 * This comment block will cover the general changes for all i_hc*.c
 * Hausdorff source files.
 * General changes:
 *  -ipshared level now uses an iterative method of region searching and
 *   reporting matches instead of a callback method.  This is for the DAE
 *   environment because callbacks don't make much sense in their world.
 *   The timings of this new method vs. the old are comparable.
 *   Before the module would search a subrect for matches, invoking a
 *   client supplied callback routine to invoke when a match was found.
 *   Now, the search is split into an init routine and an iteritively
 *   called routine.  The latter returns every time a match is found, saving
 *   the current state so searching may continue with another call.  Thus,
 *   essentially the module is now more like the callback, and the client
 *   (DAE) has more control.
 *  -ALL persistent memory management is done by the client now.  Anything
 *   persisent must be interfaced with a call to a ...Size() routine, a
 *   client allocation of this size, a filling routine, and finally, a
 *   client free of this data when done.
 *  -A new search method interface has been added: searching a list of
 *   translations.
 *  -Removed model set bitmaps from internal model set struct.  Now, the
 *   client is responsible for allocating them, the module for filling
 *   them, and the client for passing them in to the search routines.
 *   This allows the bitmaps to be compressed for example and used outside
 *   the Hausdorff module if desired.
 *  -Added ability to pass in just a image/pixr pointer to a bitmap
 *   instead in a much safer way.  Previously, since the bitmaps were
 *   hidden, DAE couldn't use this method.  Now, since they pass in pointers
 *   to bitmaps now, they can.  Client is responsible for freeing any and
 *   all bitmaps.
 *  -Dropped Hausdorff flags interface.  After feedback from clients that
 *   this was silly/confusing, I've separated that info up.  Now client
 *   passes in the number of stored bitmaps desired per model set, 0 1 or
 *   32.  Client then also passes in a MACRO for the desired search method.
 *  -Removed i_imagePutModel() and associate routines because if client
 *   really wants to do something like this, they can just access the bitmaps
 *   that are now exposed.
 *  -Removed i_deleteModelSet() and other i_delete() routines because we
 *   no longer do memory management.
 *  -Checked and optimized way memory chunks were being built.  Went through
 *   each memory chunk set up and ensured all struct field pointers were
 *   pointing to _properly aligned_ data.
 *  -Similarly, reordered all structs in module for optimal space usage.
 *   Since sizeof() determines the bytes needed to tile structs in arrays,
 *   I order the struct fields in largest field first to guarantee optimal
 *   packing/allocation!
 *  -This restructuring allows added flexibility to client to change
 *   threshold Hausdorff percents to ANY valid values at each iterative
 *   call.
 *  -This restructuring still allows callback interface at the ipwrap level.
 * Specific to this file:
 * Updated module comments.
 * Removed i_deleteModelSet(), i_imagePutModel()
 * Changed i_makeModelSet() interface to require client to have already
 * allocated space for model set.
 * Modified (most) routines to comply with new changes.
 *
 * Revision 1.12  1995/03/31  22:15:31  ejaquith
 * ejaquith on Fri Mar 31 14:12:39 PST 1995
 *
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Yanked i_probFalseMatch() routine and supplemental routine probRandomMatch()
 * since i_probFalseMatch() routine is not returning good confidence values
 * 100% of the time and I am unaware of if it is fixable in the near future.
 * Changed orderBySort() to i_orderBySort()
 *
 * Revision 1.11  1995/03/30  19:06:21  ejaquith
 * ejaquith on Thu Mar 30 11:05:04 PST 1995
 *
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Fixed i_probFalseMatch() ACC warning concerning erfc()
 * by commenting the call out and returning ia_noImpl (sour grapes?)
 * This routine is unstable and not working anyway (one of my bug fixes
 * sur la stack).
 *
 * Revision 1.10  1995/03/29  01:58:22  ejaquith
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.9  1995/03/21  19:09:59  ejaquith
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Updated Hausdorff module comments.
 *
 * Revision 1.8  1995/03/15  00:59:35  ejaquith
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Added new functionality speed up for ordering model rows.
 * The logic is simple as follows:
 *  Since the underlying correlation routines can terminate early when
 *  comparing (when too few model points overlap imagery points to match)
 *  we would like to compare the densest rows of the model first.
 * Towards this end, the following new routines were made:
 *  orderByInsertSort(), orderByQuickSort() and orderBySort():
 *   Will sort elements in an array in the fastest way.
 *  assignModelRowOrder():
 *   Given a image bitmap to make a model from, and a parameter "blockHeight"
 *   for number of rows per block, make an ordering of the rows as follows:
 *    Cut the image into blocks of blockHeight rows.
 *    Order each of the blocks, densest row last.
 *    Order the blocks, densest block last.
 *    (The correlation loops work backwards through the row order array).
 *  rowMakeOffsetsFromOrder():
 *   Given the above row order, create a list of offsets in an image
 *   to jump to the indicated row from the present row.
 * i_getModelSetSize() and i_makeModelSet() were modified to include
 * this new blockHeight parameter, and to set up the ordering.
 * Added FREE_DATA_1_MODEL hausdorff bit flag support to allow destruction
 * of a model in this data format to free the bitmap it points to.
 * Added calls to i_correlateOrdered() when the model data is ordered.
 * All the combined changes for this new Hausdorff module checkin made
 * the code 10%-75% faster (up to four times faster!) with most code
 * 50% faster using non multi-resolutional searching and 25% faster
 * using multi-resolutional searching.
 * For Smartpaper, code is about twice as fast, a needed speedup.
 * Oops, looking at my notes again, the speedups are 75% faster
 * using non multi-resolutional searching and about 40% faster using
 * multi-resolutional searching.  (First set of figures was for just
 * the row ordered model changes alone).
 *
 * Revision 1.7  1995/03/01  21:47:08  ejaquith
 * ejaquith on Wed Mar 1 13:45:35 PST 1995
 *
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Updated comments for new searching strategies.
 *
 * Revision 1.6  1995/02/14  01:43:33  ejaquith
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Updated library comments to include new flag SEARCH_EVERY_TRANS.
 *
 * Revision 1.5  1995/01/08  02:31:12  ejaquith
 * ejaquith on Sat Jan 7 18:29:33 PST 1995
 *
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Changed probability of a random match routine to cap the Hausdorff percent
 * to 0.999, because values of true 1. throw the calculations off and
 * will always then return probabilities of 0 (not desirable).
 *
 * Revision 1.4  1995/01/07  21:51:11  ejaquith
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Made calculating where the imagery and model overlap a bit faster and
 * cleaner.
 * Added probFalseMatch() routine which will take a model bitmap that matched
 * some imagery bitmap at some location and some Hausdorff percent(s) and
 * return the probability that this match is false, that it could have occurred
 * at random, and is therefore unreliable.  This isn't a final version of
 * this function, as it is not performing well.
 * NOTE: will return nonsensical answers on the PC as I can't find where
 * the erfc() routine is located.,
 *
 * Revision 1.3  1995/01/04  18:42:40  ejaquith
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Update general Hausdorff module comments to current state.
 * Abandoned use of "unit" (variable sized word for storage of models) for
 * use of 32 bit words.  Made bit counting safer.  Changed all public routines
 * to use image byte pointers and bytes per line (from image word pointers
 * and words per line).  Fixed bug in correlation percent calculation (minor).
 * Changed correlation loop calls appropriately to use number of MISSes when
 * bitwise ANDing imagery and model.
 *
 * Revision 1.2  1994/12/07  22:59:56  ejaquith
 * i_hc32ms.c -> /product/ipcore/ipshared/src/RCS/i_hc32ms.c,v
 *
 * Ref'd unused parameters to keep compiler happy.
 *
 * Revision 1.1  1994/12/06  23:36:24  ejaquith
 * Initial revision
 *

*/
