/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * i_hcmres.c: 	Hausdorff multi-resolutional search strategies for matches
 *
 *  Public routines:
 *
 *  	    Code to construct multi-resolutional information:
 *  	    	Int32	i_getMultResInfoSize()
 *  	    	Int32	i_makeMultResInfo()
 *
 *  	    Code to multi-resolution filter unmatchable subrect translations
 *  	    and set up a search (for remaining translations)
 *  	    	Int32	i_hausdorffMultResFilterSearchInit()
 *
 *
 * Extensive comments concerning the Hausdorff comparator method can be found
 * at the top of i_hc32ms.c, the default method for use.
 *
 */


#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"
#include "shrpatt.prv"
#include "memory.pub"
#include "defines.pub"

/* prototypes */
#include "shrpatt.pub"
#include "shrcmprs.pub" /* for block memory management */
#include "shrmorph.pub"    /* for i_fastDilate() */

IP_RCSINFO(RCSInfo, "$RCSfile: i_hcmres.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");


#define ASSEMBLY_DILATE 0
#define C_FAST_DILATE   1
#define C_RECT_DILATE   2
    
typedef struct myStrel {
    Int32    n;  /* Size of the square strel (height and width),
		  * The center will always be in the lower right corner */
    Int8     dilateRoutine;
} myStrel;

typedef struct mrInfo 
{
    myStrel    *pStrels;
    
    UInt8      *pData;
    Int32       nDataBytes;
    
    Int32       blockWidth2;
    DilateProc  dilateOp2;
    Int32      sizeOp2;
    
    Int32       blockWidth4;
    DilateProc	dilateOp4;
    Int32      sizeOp4;

    Int8        nStrels;
} mrInfo;


typedef struct filterType
{
    Int32        *pListX;
    Int32        *pListY;
    Int32         nTrans;
    mrInfo       *pMRInfo;
    Int32         level;
} filterType;



#define BIG_USE_RECT_DILATE    8



typedef struct node {
    struct node  *pNext;
    Int32         x;  /* The Translation Size sized area whose upper left
			  * corner is (x,y) */
    Int32         y;
} node;



/* Forward declarations */
static Int32 CDECL 
ip_multResFilterFirst (
    IPState    *pIPState,
    filterType *pFilter,	   /* Ptr to filtering info */
    modelSet   *pAMS,		   /* Ptr to model set "A" */
    UInt8     **ppBitmapsA,	   /* Array of stored A bitmaps */
    UInt32     *pBD,		   /* Ptr to upper left corner of the dilated
				    * imagery of "B", excluding frame */
    UInt32     *pBDTemp,	   /* As pBD, and same dimens.
				    * Used for calculations inside routine */
    Int32	wplBD,		   /* Number of words in imagery "B" line/row,
				    * including padding and frame */
    Int32	widthBD,	   /* imagery "B" width and height, */
    Int32	heightBD,	   /* excluding frame and padding */
    Int32	x,		   /* Rectangular area in B where we search */
    Int32	y,		   /* for A to match. */
    Int32	w,
    Int32	h,
    Float64	minPercentAtoB,	   /* A to B Hausdorff percent to use as
				    * target filter */
    Int32 (CDECL *hausdorffRoutineAtoB) ( /* Ptr to routine to use to compare */
		    IPState    *pIPState,
		    void       *pModelData, /* A to B */
		    Int32	wplM,
		    Int32	widthM,
		    Int32	heightM,
		    Int32	nptsM,
		    Int32      *pOffsetsM,
		    Int32	x,
		    Int32	y,
		    UInt32     *pI,
		    Int32	wplI,
		    Int32	widthI,
		    Int32	heightI,
		    Float64	minPercent,
		    Int32	minAndedCnt,
		    Int32      *pAndedCnt));

static Int32 CDECL 
ip_multResFilterNext (
    IPState    *pIPState,
    filterType *pFilter,	   /* Ptr to filtering info */
    modelSet   *pAMS,		   /* Ptr to model set "A" */
    UInt8     **ppBitmapsA,	   /* Array of stored A bitmaps */
    UInt32     *pBD,		   /* Ptr to upper left corner of the dilated
				    * imagery of "B", excluding frame */
    UInt32     *pBDTemp,	   /* As pBD, and same dimens.
				    * Used for calculations inside routine */
    Int32	wplBD,		   /* Number of words in imagery "B" line/row,
				    * including padding and frame */
    Int32	widthBD,	   /* imagery "B" width and height, */
    Int32	heightBD,	   /* excluding frame and padding */
    Float64	minPercentAtoB,	   /* A to B Hausdorff percent to use as
				    * target filter */
    Int32 (CDECL *hausdorffRoutineAtoB) ( /* Ptr to routine to use to compare */
		    IPState    *pIPState,
		    void       *pModelData, /* A to B */
		    Int32	wplM,
		    Int32	widthM,
		    Int32	heightM,
		    Int32	nptsM,
		    Int32      *pOffsetsM,
		    Int32	x,
		    Int32	y,
		    UInt32     *pI,
		    Int32	wplI,
		    Int32	widthI,
		    Int32	heightI,
		    Float64	minPercent,
		    Int32	minAndedCnt,
		    Int32      *pAndedCnt),
    TCBlockHeader *pBlox,	   /* Ptr to memory block structure for
				    * storing nodes */
    node       *pHeadNode,	   /* First node to begin checking in list of
				    * possible translation areas that may
				    * later contain a match */
    node       *pFootNode,	   /* Last node in list */
    Int32	maxx,		   /* The leftmost,bottommost (x,y) location*/
    Int32	maxy);             /* where a match may be found */
    
static Int32 CDECL 
ip_multResFilterLast (
    IPState    *pIPState,
    filterType *pFilter,	    /* Ptr to filtering info */
    node       *pHeadNode,	    /* First node to begin checking in list of
				     * possible translation areas that may
				     * later contain a match */
    node       *pFootNode,	    /* Last node in list */
    Int32	maxx,		    /* The leftmost,bottommost (x,y) location*/
    Int32	maxy);              /* where a match may be found */



/* PRIVATE */
/****************************************************************************
 * Count how many dilations we will need.
 * Also check that the Translation Area Sizes array is okay.
 * Figure which of the dilations will be done in assembly and the size
 * the assembly code will take.
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_invalidParm    	bad parameter passed in
 *  ia_internal	    	something unexpected happened that shouldn't
 *****************************************************************************/
static Int32 CDECL 
ip_getDilationsInfo (
    IPState    *pIPState,
    Int32      *pTransAreaSizes, /* Ptr to array of Translation Area Sizes */
    Int8       *pNDilates,     /* Ptr to resulting number of dilations */
    Int32      *pSizeOp2,      /* Ptr to assembly 2x2 dilation code size in
				* bytes (0 if could not generate assembly)*/
    Int32      *pSizeOp4)      /* Ptr to assembly 4x4 dilation code size in
				* bytes (0 if could not generate assembly)*/
{
    PROLOGUE;
    Int32 n, i;
    

    n = 0;
    while (pTransAreaSizes[n] != 1)
	n++;

    *pNDilates = (Int8) ++n;

    /* Initialize these */
    *pSizeOp2 = *pSizeOp4 = 0;

    if (n < 2)
	/* Don't waste my multi-resolutional time */
	return ia_invalidParm;

    if (pTransAreaSizes[0] > 128)
	return ia_invalidParm;

    for (i = 0; i < n; i++)
    {
	/* Ensure decreasing list of transArea sizes */
	if (i != 0)
	    if (pTransAreaSizes[i] >= pTransAreaSizes[i-1])
		return ia_invalidParm;
	/* Ensure powers of 2 */
	if ((pTransAreaSizes[i] & (pTransAreaSizes[i]-1)) != 0)
	    return ia_invalidParm;

	if (pTransAreaSizes[i] < BIG_USE_RECT_DILATE)
	{
	    /* If we can't use the rect dilate, then see if we can
	     * use an assembly code dilate that we later will generate.
	     * If so, return a nonzero size.
	     */
	    switch (pTransAreaSizes[i])
	    {
	    case 2:
	        __error__ = i_getDilateAsmSize (pIPState, 2, 2, 1, 1, pSizeOp2);
		if ((__error__ != ia_successful) && (__error__ != ia_noImpl))
		    return ia_invalidParm;
		break;
	    case 4:
	        __error__ = i_getDilateAsmSize (pIPState, 4, 4, 3, 3, pSizeOp4);
		if ((__error__ != ia_successful) && (__error__ != ia_noImpl))
		    return ia_invalidParm;
		break;
	    case 1:
		break;
	    default:
		/* This won't happen unless someone gets sloppy when editing
		 * this code.  tsk tsk!
		 */
		return ia_internal;
	    }
	}
    }

    __error__ = ia_successful;	/* we're happy */
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determine space needed for multi-resolution info allocation
 *
 * LongDescription:::
 *  Given an array of Translation Area Sizes, determine the space needed in
 *  an allocator call to make the multi-resolution info.
 *  The info will be filled in in i_makeMultResInfo().
 *
 * Input:::
 *	pTransAreaSizes::	Ptr to array of Translation Area Sizes
 *
 * Output:::
 *	pMultResInfoSize::	Ptr to returned value of multi-res info size
 *
 * Returns:::
 *  ia_successful::   	invocation done and okay
 *  ia_invalidParm::   	bad parameter passed in
 *  ia_internal::    	something unexpected happened that shouldn't
 *
 * Notes:::
 *  pTransAreaSizes array (of integers) must::
 *   contain at least 2 elements
 *   contain elements in strictly decreasing order of multiples of "2"
 *   have no element greater than 128
 *   end with element "1"
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, multi-resolution, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_getMultResInfoSize (
    IPState    *pIPState,
    Int32      *pTransAreaSizes,
    UInt32     *pMultResInfoSize)
{
    PROLOGUE;
    Int8   nDilates;
    
    Int32 sizeOp2, sizeOp4;

    __error__ = ip_getDilationsInfo (pIPState, pTransAreaSizes, &nDilates, 
				     &sizeOp2, &sizeOp4);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    *pMultResInfoSize =
	sizeof(mrInfo) + 
	nDilates * sizeof(myStrel) +
	pTransAreaSizes[0] * pTransAreaSizes[0] * sizeof(UInt8) +
	sizeOp2 +
	sizeOp4 +
	3 * 4;

    /* The last +3 is to add 3 more bytes on as a precautionary measure
     * for each pointer in the struct because in a worse case scenario,
     * something in the the modelSet memory chunk could end on the first
     * byte of a longword, and a casted pointer could START at the next
     * longword, thus skipping three bytes!
     */

    NORMAL_RETURN;
}


/* PRIVATE */
/*****************************************************************************
 * Return a properly casted pMRInfo pointer
 *
 * Description:
 *  This will dereference the memory block and reset any pointers in the
 *  memory block that may have moved during garbage collection, etc.
 *****************************************************************************/
static mrInfo * CDECL 
ip_derefMultResInfoVoid (
    IPState    *pIPState,
    void       *pMultResInfo)
{
    PROLOGUE;
    mrInfo *pMRInfo = (mrInfo *)pMultResInfo;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    if (pMRInfo->pStrels != (myStrel *)(pMRInfo + 1))
    {
	/* This is unexpected.  It may mean we haven't initially set the
	 * pointers yet to locations in the memory chunk.
	 * More likely, this could mean that the pointer to void *pMultResInfo
	 * has changed since when we last looked at it on account of garbage
	 * collection!
	 * Either way, we need to reassign the pointers in this
	 * to point to the appropriate areas in the memory chunk.
	 */

	/* myStrel is size of a multiple of words
	 * pMRInfo is as well.
	 * Thus this pointer will be casting to a word location as it should.
	 * The logic can follow for the rest of the pointers to show that
	 * they all are correctly set and in such a way so as to optimize
	 * the smallest memory chunk needed.
	 * (For example, think about if I wasn't careful and casted a
	 * UInt32 pointer to the third byte in a word! --> seg fault!)
	 * Also, if this is the Visual C++ compiler warning about
	 * casting a data pointer to a function pointer since that's
	 * exactly what's happening.
	 * Also turn off the warning about casting a function pointer to
	 * a data pointer since we have to do that to get the address
	 * of the second dilation procedure.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#pragma warning ( disable : 4054 )
#endif
	pMRInfo->pStrels = (myStrel *)(pMRInfo + 1);
	pMRInfo->pData   = (UInt8 *)(pMRInfo->pStrels + pMRInfo->nStrels);
	if (pMRInfo->sizeOp2)
	    pMRInfo->dilateOp2 =
		(DilateProc) (pMRInfo->pData + pMRInfo->nDataBytes);
	else
	    pMRInfo->dilateOp2 = NULL;

	if (pMRInfo->sizeOp4)
	{
	    if (pMRInfo->sizeOp2)
		pMRInfo->dilateOp4 =
		    (DilateProc) ((UInt8 *)pMRInfo->dilateOp2 +
					 pMRInfo->sizeOp2);
	    else
		pMRInfo->dilateOp4 =
		    (DilateProc) (pMRInfo->pData + pMRInfo->nDataBytes);
	}
	else
	    pMRInfo->dilateOp4 = NULL;
    }
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#pragma warning ( default : 4054 )
#endif

    return pMRInfo;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make the multi-resolution info
 *
 * PrivateDescription:::
 *  Given an array of Translation Area Sizes, make the multi-resolution info
 *  from it, for use in the multi-resolutional searching routines.
 *  The size of the array determines the number of levels to search.
 *  Each level searches a finer resolution than the previous, and rules out
 *  areas equal to the array elements squared.  For example, if 8, 4, 2, 1 is
 *  used, later search routines remove 8x8 translation regions (from a
 *  subrect of possible translations) first, remove 4x4 regions next, remove
 *  2x2 regions lastly, and finally check all remaining subrect translations.
 *
 * Input:::
 *   pTransAreaSizes::	Ptr to array of Translation Area Sizes
 *
 * Output:::
 *   pMultResInfo::	Ptr to multi-res info (see Note above)
 *
 * Returns:::
 *  ia_successful::   	invocation done and okay
 *  ia_nomem::	    	could not allocate required space
 *  ia_invalidParm::   	bad parameter passed in
 *  ia_internal::	something unexpected happened that shouldn't
 *
 * Notes:::
 *  pMultResInfo::
 *   The space for the info has been allocated by a client and this routine
 *   simply fills it in.
 *  pTransAreaSizes array (of integers) must::
 *   contain at least 2 elements <BR>
 *   contain elements in strictly decreasing order of multiples of "2" <BR>
 *   have no element greater than 128 <BR>
 *   end with element "1" <BR>
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching, multi-resolution
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_makeMultResInfo (
    IPState    *pIPState,
    Int32      *pTransAreaSizes,
    void       *pMultResInfo)
{
    PROLOGUE;
    Int32      i;
    Int32      tmp;
    mrInfo    *pMRInfo;
    myStrel   *pStrel;

    if (!pMultResInfo)
	return ia_invalidParm;

    pMRInfo = (mrInfo *) pMultResInfo;

    tmp = ip_getDilationsInfo (pIPState, pTransAreaSizes, &(pMRInfo->nStrels),
			       &(pMRInfo->sizeOp2), &(pMRInfo->sizeOp4));

    /* Since pTransAreaSizes[0] is the biggest STREL we need, and ALL of the
     * STRELS are square, we can get away with just making one array of
     * bytes all composed of 1s that is this max size squared! */
    pMRInfo->nDataBytes = pTransAreaSizes[0] * pTransAreaSizes[0];

    /* Set these deliberately in bad locations in the memory chunk so
     * the deref routine will (re)set them!
     */
    pMRInfo->pStrels   = (myStrel *)pMRInfo;
    pMRInfo->pData     = (UInt8 *)pMRInfo;

    /* Now finally reset all the pointers to their proper location */
    pMRInfo = ip_derefMultResInfoVoid (pIPState, (void *)pMultResInfo);
    
    /* Set up the ALL hits STREL data */
    for (i = 0; i < pMRInfo->nDataBytes; i++)
	pMRInfo->pData[i] = 1;	/* Hit */
    
    for (i = 0; i < pMRInfo->nStrels; i++)
    {
	pStrel = &(pMRInfo->pStrels[i]);
	
	/* Number of rows and columns in structure element */
	pStrel->n = pTransAreaSizes[i];
	if (pStrel->n >= BIG_USE_RECT_DILATE)
	    pStrel->dilateRoutine = C_RECT_DILATE; /* Use i_rectDilate() */
	else
	{
	    switch (pStrel->n)
	    {
	    case 2:
		if (pMRInfo->dilateOp2 &&
		    (i_genDilateAsm (pIPState, pStrel->n, pStrel->n,
				     pStrel->n - 1, pStrel->n - 1,
				     pMRInfo->pData,
				     &(pMRInfo->blockWidth2),
				     &(pMRInfo->dilateOp2),
				     &tmp) == ia_successful))
		    pStrel->dilateRoutine = ASSEMBLY_DILATE;
		else
		{
		    /* Couldn't generate an assembly code dilation
		     * routine, make sure routine pointer is reset to
		     * NULL and use normal C (slower) routine later.
		     */
		    pMRInfo->dilateOp2 = NULL;
		    pStrel->dilateRoutine = C_FAST_DILATE;
		}
		break;
	    case 4:
		if (pMRInfo->dilateOp4 &&
		    (i_genDilateAsm (pIPState, pStrel->n, pStrel->n,
				     pStrel->n - 1, pStrel->n - 1,
				     pMRInfo->pData,
				     &(pMRInfo->blockWidth4),
				     &(pMRInfo->dilateOp4),
				     &tmp) == ia_successful))
		    pStrel->dilateRoutine = ASSEMBLY_DILATE;
		else
		{
		    /* Couldn't generate an assembly code dilation
		     * routine, make sure routine pointer is reset to
		     * NULL and use normal C (slower) routine later.
		     */
		    pMRInfo->dilateOp4 = NULL;
		    pStrel->dilateRoutine = C_FAST_DILATE;
		}
		break;
	    default:
		pStrel->dilateRoutine = C_FAST_DILATE;
		break;
	    }
	}
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/****************************************************************************
 * Dilate an image for multi-resolutional use
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_internal	    	something unexpected happened that shouldn't
 *****************************************************************************/
static Int32 CDECL 
ip_myDilate (
    IPState    *pIPState,
    UInt8      *pD,	  /* Ptr to upper left corner of dest image,
			   * excluding frame */
    UInt8      *pS,	  /* Ptr to upper left corner of source image,
			   * excluding frame */
    Int32	width,	  /* Width of images, excluding frame and padding */
    Int32	height,	  /* Height of images, excluding frame and padding */
    Int32	bpl,	  /* Bytes per line of images, INCLUDING frame and
			   * padding */
    mrInfo     *pMRInfo,  /* Ptr to STREL information to dilate pS by */
    Int32	level)    /* Which element of STREL array to use */
{
    PROLOGUE;
    Int32 backFrame = (bpl << cLogFrameBits) + cFrameBytes;

    switch (pMRInfo->pStrels[level].dilateRoutine)
    {
    case ASSEMBLY_DILATE:
	/* Do the normal square assembly dilation with i_fastDilate() */
	switch (pMRInfo->pStrels[level].n)
	{
	case 2:
	    if (i_fastDilate (pIPState,  pD - backFrame,
			      pS - backFrame,
			      pMRInfo->dilateOp2,
			      NULL,
			      width + 2 * cFrameBits,
			      height + 2 * cFrameBits,
			      bpl,
			      bpl,
			      pMRInfo->blockWidth2,
			      0,
			      2, 2, 1, 1, NULL) != ia_successful)
		return ia_internal;
	    break;
	case 4:
	    if (i_fastDilate (pIPState,  pD - backFrame,
			      pS - backFrame,
			      pMRInfo->dilateOp4,
			      NULL,
			      width + 2 * cFrameBits,
			      height + 2 * cFrameBits,
			      bpl,
			      bpl,
			      pMRInfo->blockWidth4,
			      0,
			      4, 4, 3, 3, NULL) != ia_successful)
		return ia_internal;
	    break;
	default:
	    return ia_internal;
	}
	    break;
    case C_FAST_DILATE:
	/* Do the normal square C dilation with i_fastDilate() */
	if (i_fastDilate (pIPState,  pD - backFrame,
			  pS - backFrame,
			  NULL, NULL,
			  width + 2 * cFrameBits,
			  height + 2 * cFrameBits,
			  bpl,
			  bpl,
			  0, 0,
			  /* square dilation */
			  pMRInfo->pStrels[level].n,
			  pMRInfo->pStrels[level].n,
			  /* center==lower right corner */
			  pMRInfo->pStrels[level].n - 1,
			  pMRInfo->pStrels[level].n - 1,
			  /* bytes data */
			  pMRInfo->pData) != ia_successful)
	    return ia_internal;
	break;
    case C_RECT_DILATE:
	/* Do the square dilation with i_rectDilate() */
	if (i_rectDilate (pIPState,  pD - backFrame,
			  pS - backFrame,
			  width + 2 * cFrameBits,
			  height + 2 * cFrameBits,
			  bpl,
			  bpl,
			  pMRInfo->pStrels[level].n,
			  pMRInfo->pStrels[level].n) != ia_successful)
	    return ia_internal;
	break;
    default:
	return ia_internal;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Filter unmatchable translations and set up a search
 *
 * LongDescription:::
 *  Given information about "A" and "B" set up a search
 *  to find Hausdorff matches of A and B.
 *  Given a subrect of B, a model set and dilated imagery, to search in for
 *  matches of A, another model set and dilated imagery, filter out as many
 *  translation locations in the subrect that have A to B percents less
 *  than a target value as possible.  This will not filter absolutely every
 *  such translation out and will not filter out any translations that
 *  have A to B percents at least the target value.  Thus, we set up a final
 *  search info struct to use with i_hausdorffSearch() to check all the
 *  subrect translations that remain.  This filtering and i_hausdorffSearch()
 *  is much faster (typically 10-40 times faster) than simply running
 *  i_hausdorffSearch() alone.
 *
 *  Each location that has Hausdorff correlation percents that are at least
 *  the client-specified values will be reported as a match in each iterative
 *  call to i_hausdorffSearch().
 *
 * Input:::
 *	pAM::		Ptr to internal representation of model
 *			set "A" 
 *	ppBitmapsA::	Array of ptrs to stored A bitmaps.  Each
 *			ptr points to upper left corner of a
 * 			bitmap, excluding any frame and padding
 *	nBitmapsA::     Number of stored A bitmaps, 0,1,32 
 *	pBD::		Ptr to the upper left corner of the
 *			dilated imagery of "B", including frame
 *	pBDTemp::	As pBD, and same dimens as pBD image.
 *			Used for calculations inside routine 
 * 	bplBD::		Number of bytes in imagery "B" line/row,
 *			including padding and frame 
 *	widthBD::	imagery "B" width and height, 
 *	heightBD::	including frame, excluding padding 
 *	pBM::		Ptr to internal representation of model
 *			set "B" 
 *	pAD::		Ptr to the upper left corner of the
 *			dilated imagery of "A", including frame
 *	bplAD::		Number of bytes in imagery "A" line/row,
 *			including padding and frame 
 *	widthAD::	imagery "A" width and height, 
 *	heightAD::	including frame, excluding padding 
 *
 *	pMultResInfo::	Ptr to multi-resolution info 
 *	x::		Rectangular area in B where we search 
 *	y::		for A to match. 
 *	w::		May extend outside of B 
 *	h::
 *	searchMeth::	Desired search method.  Either 0 or
 * 			USE_FULL_B_FOR_BTOA_PERCENT 
 *	filterPercentAtoB::	A to B Hausdorff percent to use
 * 				as target filter.  Remove as many
 * 				subrect locations with A to B percents
 * 				less than this as possible. 
 *	allocator::	Allocator for ppSearchInfo
 *	allocInfo::	Private data for allocator
 *
 * Output:::
 *	ppSearchInfo::	Ptr to ptr to search info struct that
 * 			gets allocated (with passed in allocator)
 * 			and filled in with remaining subrect
 * 			translations after filtering. 
 *
 * Returns:::
 *  ia_successful::   	invocation done and okay
 *  ia_internal::	something unexpected happened that shouldn't
 *  ia_invalidParm::  	one or more parameters are invalid
 *  	    	    	OR a dilation error occurred
 *  	    	    	OR something wrong with model set(s) passed in
 *  ia_nomem::	    	couldn't allocate needed memory
 *
 * Notes:::
 *  Translation locations::
 *    These refer to the location of A's upper left corner
 *    with respect to B.  i.e. (0,0) means align upper left corners, (-1,0)
 *    means place A so its upper left corner is 1 pixel to the left of B's.
 *    Match locations will be in the frame of the full B image.
 *  1-Way/2-Way matching::
 *    In 1-Way matching AM is compared to BD to return the directed Hausdorff
 *    percent from A to B.
 *    In 2-Way matching, additionally BM is compared to AD to return the
 *    directed Hausdorff percent from B to A.
 *    If the client will never desire the B to A percent to be calculated
 *    and returned in the search, then BM may be passed in NULL, and the
 *    AD parameters will be ignored.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, multi-resolution, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_hausdorffMultResFilterSearchInit (
    IPState    *pIPState,
    void       *pAM,
    UInt8     **ppBitmapsA,
    UInt8	nBitmapsA,
    UInt8      *pBD,
    UInt8      *pBDTemp,
    Int32	bplBD,
    Int32	widthBD,
    Int32	heightBD,
    void       *pBM,
    UInt8      *pAD,
    Int32	bplAD,
    Int32	widthAD,
    Int32	heightAD,
    void       *pMultResInfo,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8	searchMeth,
    Float64	filterPercentAtoB,
    MallocType	allocator,
    void       *allocInfo,
    void      **ppSearchInfo)
{
    PROLOGUE;
    UInt32         searchInfoSize;
    
    filterType     filter;
    modelSet      *pAMS;
    Int32  (CDECL *hausdorffRoutineAtoB) (
	IPState	   *pIPState,
	void	   *pModelData,
	Int32	    wplM,
	Int32	    widthM,
	Int32	    heightM,
	Int32	    nptsM,
	Int32	   *pOffsetsM,
	Int32	    x,
	Int32	    y,
	UInt32	   *pI,
	Int32	    wplI,
	Int32	    widthI,
	Int32	    heightI,
	Float64	    minPercent,
	Int32	    minAndedCnt,
	Int32	   *pAndedCnt);

    pAMS = ip_derefModelSetVoid (pIPState, pAM);
    if (pAMS == NULL)
	return ia_invalidParm;

    filter.pMRInfo = ip_derefMultResInfoVoid (pIPState, pMultResInfo);
    filter.level = 0;

    if (pAMS->nBitmaps != nBitmapsA)
	return ia_invalidParm;
    
    switch (pAMS->nBitmaps)
    {
    case 32:
	hausdorffRoutineAtoB = ip_hausdorffCompare;
	break;
    case 0:
	hausdorffRoutineAtoB = ip_hausdorffComparePtList;
	break;
    case 1:
	hausdorffRoutineAtoB = ip_hausdorffCompareOneModel;
	break;
    default:
	return ia_invalidParm;
	break;
    }

    if (pAMS->pOffsetsM)
	ip_rowMakeOffsetsFromOrder (pIPState, pAMS->pOffsetsM + 2*pAMS->heightM,
				    NULL, pAMS->heightM, bplBD >> 2,
				    pAMS->pOffsetsM + pAMS->heightM);
    
    /* Do the real work */
    __error__ = ip_multResFilterFirst (pIPState, &filter,
				     pAMS,
				     ppBitmapsA,
				     (UInt32 *)(pBD + (bplBD << cLogFrameBits)
					    + cFrameBytes),
				     (UInt32 *)(pBDTemp +
						(bplBD << cLogFrameBits) +
						cFrameBytes),
				     bplBD >> 2,
				     widthBD - 2*cFrameBits,
				     heightBD - 2*cFrameBits,
				     x,
				     y,
				     w,
				     h,
				     filterPercentAtoB,
				     hausdorffRoutineAtoB);
    
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    /* Now transfer the remaining list of translations that passed the filter
     * into a search info structure.
     */

    /* Note:  filter.nTrans may be 0! */
    if (i_hausdorffListSearchSize (pIPState, filter.nTrans, &searchInfoSize))
	return ia_invalidParm;

#ifdef DEBUG_HEAPS
    *ppSearchInfo = (void *)allocator(searchInfoSize, 0, allocInfo,
					__FILE__, __LINE__);
#else
    *ppSearchInfo = (void *)allocator(searchInfoSize, 0, allocInfo);
#endif
    if (!*ppSearchInfo)
	return ia_nomem;

    if (i_hausdorffListSearchInit (pIPState, pAM, pBD, bplBD, widthBD, heightBD,
				   pBM, pAD, bplAD, widthAD, heightAD,
				   filter.pListX, filter.pListY, filter.nTrans,
				   searchMeth, *ppSearchInfo))
	return ia_invalidParm;

    if (filter.nTrans)
    {
	FREE(filter.pListX);
	FREE(filter.pListY);
    }

    NORMAL_RETURN;
}


/* Nice little macro to add another sub translation area to the list of areas
 * to check at the next level. */
#define ADD_NODE(pBloxP, pPrevNodeP, pNewNodeP, xP, yP, errLabel) \
  if (i_allocateStruct (pIPState, pBloxP, \
		        (void **)&pNewNodeP) != ia_successful) \
    goto errLabel; \
  pNewNodeP->x = xP; \
  pNewNodeP->y = yP; \
  pPrevNodeP->pNext = pNewNodeP; \
  pPrevNodeP = pNewNodeP;


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Do first level of multi-resolution filtering
 *
 * Input:::
 *	pFilter::	Ptr to filtering info
 *	pAMS::		Ptr to model set "A"
 *	ppBitmapsA::	Array of stored A bitmaps
 *	pBD::		Ptr to upper left corner of the dilated
 *			imagery of "B", excluding frame
 *	pBDTemp::	As pBD, and same dimens.
 *			Used for calculations inside routine
 *	wplBD::		Number of words in imagery "B" line/row,
 *			including padding and frame
 *	widthBD::	imagery "B" width and height,
 *	heightBD::	excluding frame and padding
 *	x::		Rectangular area in B where we search
 *	y::		for A to match.
 *	w::		
 *	h::		
 *	minPercentAtoB::	A to B Hausdorff percent to use as target
 *				filter
 *	hausdorffRoutineAtoB::	Ptr to routine to use to compare A to B
 *
 * Returns:::
 *  ia_successful::   	invocation done and okay
 *  ia_nomem::	    	couldn't allocate needed memory
 *  ia_internal::	something unexpected happened that shouldn't
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching,
 *  multi-resolution
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static Int32 CDECL 
ip_multResFilterFirst (
    IPState    *pIPState,
    filterType *pFilter,
    modelSet   *pAMS,
    UInt8     **ppBitmapsA,
    UInt32     *pBD,
    UInt32     *pBDTemp,
    Int32	wplBD,
    Int32	widthBD,
    Int32	heightBD,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Float64	minPercentAtoB,
    Int32 (CDECL *hausdorffRoutineAtoB) (
		    IPState    *pIPState,
		    void       *pModelData,
		    Int32	wplM,
		    Int32	widthM,
		    Int32	heightM,
		    Int32	nptsM,
		    Int32      *pOffsetsM,
		    Int32	x,
		    Int32	y,
		    UInt32     *pI,
		    Int32	wplI,
		    Int32	widthI,
		    Int32	heightI,
		    Float64	minPercent,
		    Int32	minAndedCnt,
		    Int32      *pAndedCnt))
{
    PROLOGUE;
    Int32          andedCntAtoB, tmpInt32;
    Int32          minAndedCntAtoB;
    Int32          firstx, nx, ny, transAreaX;
    Int32          transAreaSize, splitSize, nSplits;
    TCBlockHeader *pBlox = NULL;
    node          *pHeadNode, *pNewNode, *pPrevNode;

    /* Size of the (square) translation area on a side (thus its the sqrt) */
    transAreaSize = pFilter->pMRInfo->pStrels[0].n;
    /* Split each translation area that matches into nSplits * nSplits
     * smaller areas (dividing up with a grid) */
    nSplits = transAreaSize / pFilter->pMRInfo->pStrels[1].n;
    /* The increment from one subdivided area to another */
    splitSize = transAreaSize / nSplits;

    /* Dilate the BD image at this new level according to the size of this
     * new STREL.
     * The STREL is a pStrel[0].n (or nSplits * splitSize) squared
     * STREL with the center element in the lower right corner. */
    __error__ = ip_myDilate (pIPState,  (UInt8 *)pBDTemp, (UInt8 *)pBD,
			   widthBD, heightBD, wplBD << 2,
			   pFilter->pMRInfo, 0);
    if (__error__ != ia_successful)
	goto filterFirstErrFin;

    /* We'll use the memory block allocation system to keep track of
     * allocating and freeing "nodes" (representing translation areas).
     * We need to start with some number of nodes in the first block, so we
     * choose to start with the number of translation areas that there are
     * in this first level */
    __error__ = i_allocateMemoryBlock (pIPState,  &pBlox, sizeof(node),
				     MAX(w / transAreaSize * h / transAreaSize,
					 10),
				     cFalse,
				     NULL, NULL,
				     i_compressAlloc, i_compressFree );
    if (__error__ != ia_successful)
	goto filterFirstErrFin;

    /* Set the minAndedCnt counters for early termination */
    minAndedCntAtoB = (Int32)(ceil(minPercentAtoB * pAMS->nptsM));

    /* The first node is always a header node and not used or freed, just
     * used as a marker */
    __error__ = i_allocateStruct (pIPState, pBlox, (void **)&pHeadNode );
    if (__error__ != ia_successful)
	goto filterFirstErrFin;
    pPrevNode = pHeadNode;
    
    pFilter->nTrans = 0;

    /* Go through the entire subrect of B, breaking it up into pieces of
     * translation areas.  Try A to B matching at the upper left corner of
     * each translation area.
     */
    /* The upper left location of the node which represents some
     * translation area.  This area is exactly
     * nodex,nodey  s.t. x <= nodex < x + nSplits*splitSize
     *                   y <= nodey < y + nSplits*splitSize
     */
    
    /* Last x,y values to check at */
    w += (x - pAMS->widthM);
    h += (y - pAMS->heightM);
    firstx = x;
    
    __error__ = ia_nomem; /* Setup for Error breaking out of loop */

    /* An incidental note:
     * Here the call to the checker really only wants to know "Does A
     * match B at least this well at this location?" and not the exact
     * match percent, because this is simply a Yes/No criterion to
     * decide "do we need to expand this translation area into finer
     * parts?"  However, in practice, upon making a function that would
     * take advantage of this, by offering both early termination from
     * the image comparing by once we found either enough HITs (A model
     * points overlapping BD imagery points) or enough MISSes to
     * determine whether A could match BD here, it needed two counters.
     * When put to the test, it was slower except when the Hausdorff
     * percent was very low (30%-50%) (thus more early terminations
     * due to sufficient HITs overshadowed the extra counter),
     * and only faster by a little.  Net: not worth it!
     */

    if (nSplits == 2)
    {
	for ( ; y <= h; y += transAreaSize)
	{
	    for (x = firstx ; x <= w; x += transAreaSize)
	    {
		/* This will just give us the 1Way answer to
		 * "Could my A image match B at this location at this level?"
		 */

		hausdorffRoutineAtoB(pIPState,
				     ip_getModelData (pIPState,
						      pAMS,
						      ppBitmapsA,
						      x),
				     pAMS->wplM,
				     pAMS->widthM,
				     pAMS->heightM,
				     pAMS->nptsM,
				     pAMS->pOffsetsM,
				     x,
				     y,
				     pBDTemp,
				     wplBD,
				     widthBD,
				     heightBD,
				     minPercentAtoB,
				     minAndedCntAtoB,
				     &andedCntAtoB);
		
		if (andedCntAtoB >= minAndedCntAtoB)
		{
		    /* A matched BD dilated by the STREL at this level, so this
		     * means that this translation area may be matchable in the
		     * future */
		    
		    /* We need to divide this current transArea into 4
		     * sub transAreas */
		    nx = x + splitSize;
		    ny = y + splitSize;
		    
		    ADD_NODE(pBlox, pPrevNode, pNewNode,  x, y,
			     filterFirstErrFin);
		    ADD_NODE(pBlox, pPrevNode, pNewNode, nx, y,
			     filterFirstErrFin);
		    ADD_NODE(pBlox, pPrevNode, pNewNode,  x,ny,
			     filterFirstErrFin);
		    ADD_NODE(pBlox, pPrevNode, pNewNode, nx,ny,
			     filterFirstErrFin);

		    pFilter->nTrans += 4;
		}
	    }
	}
    }
    else
    {
        for ( ; y <= h; y += transAreaSize)
	{
	    for (x = firstx ; x <= w; x += transAreaSize)
	    {
		/* This will just give us the 1Way answer of to
		 * "Could my A image match B at this location at this level?"
		 */
		
		hausdorffRoutineAtoB(pIPState,
				     ip_getModelData (pIPState,
						      pAMS,
						      ppBitmapsA,
						      x),
				     pAMS->wplM,
				     pAMS->widthM,
				     pAMS->heightM,
				     pAMS->nptsM,
				     pAMS->pOffsetsM,
				     x,
				     y,
				     pBDTemp,
				     wplBD,
				     widthBD,
				     heightBD,
				     minPercentAtoB,
				     minAndedCntAtoB,
				     &andedCntAtoB);
		
		if (andedCntAtoB >= minAndedCntAtoB)
		{
		    
		    /* A matched BD dilated by the STREL at this level, so this
		     * means that this translation area may be matchable in the
		     * future */
		    
		    /* We need to divide this current transArea into more
		     * than 4 sub transAreas */
		    transAreaX = x;
		    tmpInt32 = y;
		    nx = x + transAreaSize;
		    ny = y + transAreaSize;

		    for ( ; y < ny; y += splitSize)
		    {
			for (x = transAreaX; x < nx; x += splitSize)
			{
			    ADD_NODE(pBlox, pPrevNode, pNewNode, x, y,
				     filterFirstErrFin);
			}
		    }
		    
		    /* Restore the upper left hand corner of the subdivided
		     * translation area */
		    x = transAreaX;
		    y = tmpInt32;

		    pFilter->nTrans += (nSplits*nSplits);
		}
	    }
	}
    }
    

    /* Make a footer node which will be a dummy node that will never be used
     * or freed like the header node, and just a marker */
    __error__ = i_allocateStruct (pIPState, pBlox, (void **)&pNewNode);
    if (__error__ != ia_successful)
	goto filterFirstErrFin;
    pPrevNode->pNext = pNewNode;
    /* pNewNode is now this footer node */

    /* Now do the next multi-resolutional level */
    pFilter->level = 1;

    if (pFilter->pMRInfo->pStrels[pFilter->level].n != 1)
	__error__ = ip_multResFilterNext (pIPState,  pFilter, pAMS, ppBitmapsA,
					pBD, pBDTemp, wplBD, widthBD, heightBD,
					minPercentAtoB, hausdorffRoutineAtoB,
					pBlox, pHeadNode, pNewNode,
					w, h );
    else
	__error__ = ip_multResFilterLast (pIPState,  pFilter,
					pHeadNode, pNewNode, w, h );
    

filterFirstErrFin:
    if (pBlox != NULL)
	i_freeAllMemoryBlocks (pIPState, &pBlox);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do middle level of multi-resolution filtering
 *
 * Input:::
 *	pFilter::	Ptr to filtering info
 *	pAMS::		Ptr to model set "A"
 *	ppBitmapsA::	Array of stored A bitmaps
 *	pBD::		Ptr to upper left corner of the dilated
 *			imagery of "B", excluding frame
 *	pBDTemp::	As pBD, and same dimens.
 *			Used for calculations inside routine
 *	wplBD::		Number of words in imagery "B" line/row,
 *			including padding and frame
 *	widthBD::	imagery "B" width and height,
 *	heightBD::	excluding frame and padding
 *	minPercentAtoB::	A to B Hausdorff percent to use as target
 *				filter
 *	hausdorffRoutineAtoB::	Ptr to routine to use to compare A to B
 *	pBlox::		Ptr to memory block structure for
 *			storing nodes
 *	pHeadNode::	First node to begin checking in list of
 *			possible translation areas that may
 *			later contain a match
 *	pFootNode::	Last node in list
 *	maxx::		The leftmost,bottommost (x,y) location
 *	maxy::		where a match may be found
 *
 * Returns:::
 *  ia_successful::   	invocation done and okay
 *  ia_nomem::	    	couldn't allocate needed memory
 *  ia_internal::    	something unexpected happened that shouldn't
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, multi-resolution, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static Int32 CDECL 
ip_multResFilterNext (
    IPState    *pIPState,
    filterType *pFilter,
    modelSet   *pAMS,
    UInt8     **ppBitmapsA,
    UInt32     *pBD,
    UInt32     *pBDTemp,
    Int32	wplBD,
    Int32	widthBD,
    Int32	heightBD,
    Float64	minPercentAtoB,
    Int32 (CDECL *hausdorffRoutineAtoB) (
		    IPState    *pIPState,
		    void       *pModelData,
		    Int32	wplM,
		    Int32	widthM,
		    Int32	heightM,
		    Int32	nptsM,
		    Int32      *pOffsetsM,
		    Int32	x,
		    Int32	y,
		    UInt32     *pI,
		    Int32	wplI,
		    Int32	widthI,
		    Int32	heightI,
		    Float64	minPercent,
		    Int32	minAndedCnt,
		    Int32      *pAndedCnt),
    TCBlockHeader *pBlox,
    node       *pHeadNode,
    node       *pFootNode,
    Int32	maxx,
    Int32	maxy)
{
    PROLOGUE;
    node   *pCurNode;
    Int32   andedCntAtoB, tmpInt32;
    Int32   x, y, nSplits, splitSize, nx, ny, minAndedCntAtoB;
    node   *pNewNode, *pPrevNode, *pNextNode;
    
    /* The first node is always a header node and not used or freed, just
     * used as a marker */
    pPrevNode = pHeadNode;
    pCurNode  = pHeadNode->pNext;

    /* Set the minAndedCnt counters for early termination */
    minAndedCntAtoB = (Int32)(ceil(minPercentAtoB * pAMS->nptsM));
	
    /* Split each translation area that matches into nSplits * nSplits
     * smaller areas (dividing up with a grid) */
    nSplits = pFilter->pMRInfo->pStrels[pFilter->level].n /
	pFilter->pMRInfo->pStrels[pFilter->level + 1].n;
    /* The increment from one subdivided area to another */
    splitSize = pFilter->pMRInfo->pStrels[pFilter->level].n / nSplits;
	
    /* Dilate the BD image at this new level according to the size of this
     * new STREL.
     * The STREL is a pStrels[level].n (or nSplits * splitSize) squared
     * STREL with the center element in the lower right corner. */
    __error__ = ip_myDilate (pIPState,  (UInt8 *)pBDTemp, (UInt8 *)pBD,
			   widthBD, heightBD, wplBD << 2,
			   pFilter->pMRInfo, pFilter->level);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    /* This will just give us the 1Way answer of to
     * "Could my A image match B at this location at this level?"
     */

    if (nSplits == 2)
    {
	while (pCurNode != pFootNode)
	{ 
	    pNextNode = pCurNode->pNext; 
		
	    /* The upper left location of the node which represents some
	     * translation area.  This area is exactly
	     * nodex,nodey  s.t. x <= nodex < x + nSplits*splitSize
	     *                   y <= nodey < y + nSplits*splitSize
	     */
	    x = pCurNode->x;
	    y = pCurNode->y;
		
	    /* Parse out x,y locations that may have stepped over the
	     * right and bottom subrect bounds when a upper left
	     * identified translation area got split at the previous
	     * level, and parse out non matches
	     */
	    if (x <= maxx && y <= maxy)
	    {
		hausdorffRoutineAtoB(pIPState,
				     ip_getModelData (pIPState,
						      pAMS,
						      ppBitmapsA,
						      x),
				     pAMS->wplM,
				     pAMS->widthM,
				     pAMS->heightM,
				     pAMS->nptsM,
				     pAMS->pOffsetsM,
				     x,
				     y,
				     pBDTemp,
				     wplBD,
				     widthBD,
				     heightBD,
				     minPercentAtoB,
				     minAndedCntAtoB,
				     &andedCntAtoB);
		    
		if (andedCntAtoB >= minAndedCntAtoB)
		{
		    /* A matched BD dilated by the STREL at this level, so
		     * this means that this translation area may later on
		     * be matchable */

		    /* Now add 3 nodes to the node list to subdivide this
		     * transArea into 4.  (3 because the current node is
		     * still the same!) */
		    nx = x + splitSize;
		    ny = y + splitSize;
			
		    ADD_NODE(pBlox, pCurNode, pNewNode, nx, y,
			     filterNextErrFin);
		    ADD_NODE(pBlox, pCurNode, pNewNode,  x,ny,
			     filterNextErrFin);
		    ADD_NODE(pBlox, pCurNode, pNewNode, nx,ny,
			     filterNextErrFin);

		    pFilter->nTrans += 3;
			
		    pNewNode->pNext = pNextNode;
		}
		else /* didn't match */
		{
		    /* Kill this node and rule out the entire translation
		     * area that it represents!! (A can never match BD at
		     * any of these translations contained in this area) */
		    pPrevNode->pNext = pNextNode;
		    i_freeStruct (pIPState, pBlox, (void **)&pCurNode);
			
		    pCurNode = pPrevNode;
			
		    (pFilter->nTrans)--;
		}
	    }
	    else {
		/* Translation area out of bounds */
		pPrevNode->pNext = pNextNode;
		i_freeStruct (pIPState, pBlox, (void **)&pCurNode);
		    
		pCurNode = pPrevNode;
		    
		(pFilter->nTrans)--;
	    }
		
		
	    pPrevNode = pCurNode;
	    pCurNode = pNextNode;
	}
    }
    else {
	Int32 nsize = pFilter->pMRInfo->pStrels[pFilter->level].n;
	    
	while (pCurNode != pFootNode)
	{ 
	    pNextNode = pCurNode->pNext; 
		
	    /* The upper left location of the node which represents some
	     * translation area.  This area is exactly
	     * nodex,nodey  s.t. x <= nodex < x + nSplits*splitSize
	     *                   y <= nodey < y + nSplits*splitSize
	     */
	    x = pCurNode->x;
	    y = pCurNode->y;
		
	    /* Parse out x,y locations that may have stepped over the
	     * right and bottom subrect bounds when a upper left
	     * identified translation area got split at the previous
	     * level, and parse out non matches
	     */
	    if (x <= maxx && y <= maxy)
	    {
		hausdorffRoutineAtoB(pIPState,
				     ip_getModelData (pIPState,
						      pAMS,
						      ppBitmapsA,
						      x),
				     pAMS->wplM,
				     pAMS->widthM,
				     pAMS->heightM,
				     pAMS->nptsM,
				     pAMS->pOffsetsM,
				     x,
				     y,
				     pBDTemp,
				     wplBD,
				     widthBD,
				     heightBD,
				     minPercentAtoB,
				     minAndedCntAtoB,
				     &andedCntAtoB);
		    
		if (andedCntAtoB >= minAndedCntAtoB)
		{
		    /* A matched BD dilated by the STREL at this level, so
		     * this means that this translation area may later on
		     * be matchable */
		    
		    /* We need to divide this current transArea into more
		     * than 4 sub transAreas */
		    nx = x + nsize;
		    ny = y + nsize;
		    tmpInt32 = x;
			
		    /* Again, the current node is still the same, so do the
		     * first row separately to exclude the top left
		     * corner */
		    for (x += splitSize; x < nx; x += splitSize)
		    {
			ADD_NODE(pBlox, pCurNode, pNewNode, x, y,
				 filterNextErrFin);
		    }
			
		    /* Now do the rest of the rows */
		    for (y += splitSize; y < ny; y += splitSize)
		    {
			for (x = tmpInt32; x < nx; x += splitSize)
			{
			    ADD_NODE(pBlox, pCurNode, pNewNode, x, y,
				     filterNextErrFin);
			}
		    }

		    pFilter->nTrans += (nSplits * nSplits - 1);
			
		    pNewNode->pNext = pNextNode;
		}
		else /* didn't match */
		{
		    /* Kill this node and rule out the entire translation
		     * area that it represents!! (A can never match BD at
		     * any of these translations contained in this area) */
		    pPrevNode->pNext = pNextNode;
		    i_freeStruct (pIPState, pBlox, (void **)&pCurNode);
			
		    pCurNode = pPrevNode;
			
		    (pFilter->nTrans)--;
		}
	    }
	    else
	    {
		/* Translation area out of bounds */
		pPrevNode->pNext = pNextNode;
		i_freeStruct (pIPState, pBlox, (void **)&pCurNode);
		    
		pCurNode = pPrevNode;
		    
		(pFilter->nTrans)--;
	    }
		
	    pPrevNode = pCurNode;
	    pCurNode = pNextNode;
	}
    }

    /* Now do the next multi-resolutional level */
    pFilter->level++;

    if (pFilter->pMRInfo->pStrels[pFilter->level].n != 1)
	return ( ip_multResFilterNext (pIPState,  pFilter, pAMS, ppBitmapsA,
				       pBD, pBDTemp, wplBD, widthBD, heightBD,
				       minPercentAtoB, hausdorffRoutineAtoB,
				       pBlox, pHeadNode, pFootNode,
				       maxx, maxy ));
    else
	return ( ip_multResFilterLast (pIPState,  pFilter, pHeadNode, pFootNode,
				       maxx, maxy ));
    
filterNextErrFin:
    /* Ran out of memory while allocating new nodes */
    return ia_nomem;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do last level of multi-resolution filtering
 *
 * Input:::
 *  pFilter::		Ptr to filtering info
 *  pHeadNode::		First node to begin checking in list of
 *			possible translation areas that may
 *			later contain a match
 *  maxx::		The leftmost,bottommost (x,y) location
 *  maxy::		where a match may be found
 *
 * Returns:::
 *  ia_successful::   	invocation done and okay
 *  ia_nomem::	    	couldn't allocate needed memory
 *  ia_internal::    	something unexpected happened that shouldn't
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching, multi-resolution
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static Int32 CDECL 
ip_multResFilterLast (
    IPState    *pIPState,
    filterType *pFilter,
    node       *pHeadNode,
    node       *pFootNode,
    Int32	maxx,
    Int32	maxy)
{
    PROLOGUE;
    Int32          i = 0;
    
    /* The first node is always a header node and not used or freed, just
     * used as a marker */
    pHeadNode  = pHeadNode->pNext;

    if (pFilter->nTrans <= 0)
    {
	/* No translations left at this level; make the translation lists 
	 * empty.
	 */
	pFilter->pListX = NULL;
	pFilter->pListY = NULL;
	pFilter->nTrans = 0;
	NORMAL_RETURN;
    }

    /* Make a translation list (separate x/y component arrays) that can hold
     * as many translations as there are left.
     */
    pFilter->pListX = (Int32 *)CALLOC(pFilter->nTrans, sizeof(Int32));
    pFilter->pListY = (Int32 *)CALLOC(pFilter->nTrans, sizeof(Int32));
    if (!pFilter->pListX || !pFilter->pListY)
	return ia_nomem;

    while (pHeadNode != pFootNode)
    { 
	/* Parse out x,y locations that may have stepped over the right and
	 * bottom subrect bounds when an upper left identified translation
	 * area got split at the previous level.
	 */
	if (pHeadNode->x <= maxx && pHeadNode->y <= maxy)
	{
	    /* The upper left location of the node represents exactly
	     * that one single translation now!
	     */
	    pFilter->pListX[i] = pHeadNode->x;
	    pFilter->pListY[i] = pHeadNode->y;
		
	    i++;
	}
	else {
	    /* This remaining x,y location is out of bounds, so don't record
	     * it in the above list AND readjust the passed back number of
	     * translations that will be present in each component array.
	     */
	    pFilter->nTrans--;
	}
	    
	pHeadNode = pHeadNode->pNext;
    }

    if (i != pFilter->nTrans)
	/* If this happens, then the recording scheme for the number of
	 * translations in the heap is not equal to then number of
	 * translations ACTUALLY in the heap.  A bug/error, report this!
	 */
	return ia_internal;

    /* all done */
    NORMAL_RETURN;
}



/*

$Log: i_hcmres.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.26  1998/01/13 23:48:24  ddavies
i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.25  1997/07/11 01:48:51  ddavies
i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.24  1996/12/20 22:00:02  ddavies
ddavies on Fri Dec 20 13:59:17 PST 1996

i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v

Have calls to i_getDilateAsmSize not crash if an ia_noImpl code is returned.
This happens when the slow code is linked now.

Revision 1.23  1996/12/18 23:09:35  ddavies
ddavies on Wed Dec 18 15:08:43 PST 1996

i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v

Add dummy parameters to calls to i_fastDilate because it takes
pointers to long machine code now.  If we want this stuff to
work fast, someone might want to rearrange things so that MMX
code can be used.

Revision 1.22  1996/04/17 23:25:10  ejaquith
ejaquith on Wed Apr 17 16:24:49 PDT 1996

i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v

Fixed auto-doc stuff that is lingering there for demo of IPShared routines.

Revision 1.21  1996/03/26 22:38:46  ddavies
ddavies on Tue Mar 26 14:37:25 PST 1996

i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v

Change the type of the allocator passed to i_hausdorffMultResFilterSearchInit
to be MallocType.  This type of allocator takes two variables, the number
of bytes to allocate and a void * supplied by whoever supplied the allocator
Not only is this a nicer type of allocator, it allows us to get rid of
a plain MALLOC that was passed down from ipwrap.

Revision 1.20  1995/12/13 02:44:30  ejaquith
i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.19  1995/12/08  00:48:02  ejaquith
 * ejaquith on Thu Dec 7 16:46:49 PST 1995
 *
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Updated some of the function comment blocks to privatize some info
 * that will be going out soon to iplib.
 *
 * Revision 1.18  1995/10/30  18:12:04  ejaquith
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Updated HTML secret flags.
 *
 * Revision 1.17  1995/10/18  03:57:20  ddavies
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.16  1995/10/03  21:12:17  ddavies
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Include shrmorph.pub instead of shrip.pub.  Remove casts of &(pMRInfo->dilateOp2) and &(pMRInfo->dilateOp4) since they are no longer needed.
 *
 * Revision 1.15  1995/10/03  00:29:59  ddavies
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Replace the funky old declarations for runtime machine code procedures
 * void CDECL (*dilateOp)() with types.  The types are conditionally defined
 * in shrip.pub so they vary with platform.
 *
 * Revision 1.14  1995/09/14  18:59:47  ejaquith
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * CommonCheck: Converted all routines that were private in files to be of form
 * ip_*(), which they weren't.  No functionality/noticeable changes.
 *
 * Revision 1.13  1995/07/14  20:29:44  ejaquith
 * ejaquith on Fri Jul 14 13:29:04 PDT 1995
 *
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Jumped the gun and added some new function header comments to test
 * the pre-natal auto-doc system.
 *
 * Revision 1.12  1995/07/10  23:15:38  ddavies
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Change args to i_allocateMemoryBlock to take args saying that we should NOT
 * be notified when the heap blocks move.  This is done because the heap is
 * NOT persistent (it worked before pointer fixup was introduced, right?)
 * Also change the i_allocateStruct casts from (TCBlockHeader **) to
 * (void **).  This is because I was cheating in trying to use a TCBlockHeader
 * as a stand-in for a random structure when I should have been using a void.
 *
 * Revision 1.11  1995/04/22  20:58:38  ejaquith
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Please see general (large!) module comments for same date in i_hc32ms.c
 * Modified routines to comply with these new changes.
 * Removed i_hausdorff1WayMatchesMR() and i_hausdorff2WayMatchesMR().
 * Interface is now via use of new routine i_hausdorffMultResFilterSearchInit()
 * which filters out most translations in a subrect.  Remaining translations
 * are searched through standard iterative calls to i_hausdorffSearch().
 * This new routine does multi-resolution matching, dropping from the first
 * desired level to the last, parsing out areas of translations in the
 * normal way.  These lower level routines have all been changed/modified
 * to use the new interface.  The last level is now a separate routine
 * which takes all remaining translations, and stores them into x,y
 * component arrays and returns.  Upon return to i_hausdorff...SearchInit()
 * it takes these arrays and (internally) calls i_hausdorffListSearchSize()
 * and uses the client's passed in allocator to allocate this space for
 * a persistent memory searching structure.  Then it calls
 * i_hausdorffListSearchInit() to transfer the lists to the searching
 * structure, and returns.
 * At this time, the client would call i_hausdorffSearch() using 1-Way or
 * 2-Way matching to sort through all translations that remain.
 * Verified all memory chunks were setting their pointer fields to
 * properly aligned areas in the chunk.  Optimized all used structs to
 * save space.
 * Removed i_deleteMultResInfo(), client has to do this now.
 * What multi-resolution searching does is now more intuitive with this
 * interface -- it is just a filter really that will not remove any actual
 * match and conservatively removes all mismatches.
 *
 * Revision 1.10  1995/03/29  01:58:45  ejaquith
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.9  1995/03/24  02:18:51  ejaquith
 * ejaquith on Thu Mar 23 18:16:27 PST 1995
 *
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Since the dilations by 2x2 and 4x4 STRELs were getting done in assembly,
 * I had to make where they got written be inside the pointer to the
 * multi-resolution info memory chunk.
 * Revamped my STREL structure and memory block and made it more robust.
 *
 * Revision 1.8  1995/03/21  19:09:37  ejaquith
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Consolidated some code and made remaining code more flexible.
 * (And faster!)
 * emoved "best" searching techniques, routines, etc.
 * This type of searching is now done easily with the client's callback
 * routine (depending on the client's return value).
 * Now, both 1Way matching and 2Way matching have the same parametered
 * callback routine (to allow greater flexibility even though an extra
 * dummy Hausdorff percent value will now be returned to the callback
 * when only doing 1Way searching).
 * Removed i_hausdorff1WayBestMatchMR() since the normal 1Way matching
 * can now handle this.
 * Invocations of matchCheck1Way() and matchCheck2Way() now handle invoking
 * the client's callback routine and depending on the callback return value,
 * whether or not any search information should be updated.
 * This removes a whole level of routines for "best" searching.
 *
 * Revision 1.7  1995/03/15  00:42:36  ejaquith
 * ejaquith on Tue Mar 14 16:33:23 PST 1995
 *
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Updated multi-resolutional searching to work with new feature
 * "row ordered models".
 * Reorganized code listing in multResNextLevel() to speed up searching
 * by unrolling a check to be checked now once per invocation of routine.
 * (Before it was checked once per node in the list of nodes).
 * Surprisingly this sped up code a lot.
 *
 * Revision 1.6  1995/03/01  22:04:39  ejaquith
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Updated myDilate() routine and i_makeMultResInfo() routine to set up and
 * use i_rectDilate() routine instead of i_fastDilate() for all dilations
 * of 8x8 and greater (up to 128x128).
 * Reorganized structure of multResNextLevel() to increase speed a tad
 * and make it more readable, by splitting it into two logical pieces --
 * when the call is at the last level and when it is not.
 * Doesn't change functionality here.
 * myDilate() is now much much faster at larger dilations, and the usage
 * of 16x16, 32x32, and 64x64 are now completely viable at 300dpi.
 *
 * Revision 1.5  1995/01/31  03:38:15  ejaquith
 * ejaquith on Mon Jan 30 19:36:52 PST 1995
 *
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Fixed bug in allocation of blocks for list of translation area cells
 * Really rare cases could trigger allocating 0 structs in the block
 * and resulted in a complimentary seg fault.
 *
 * Revision 1.4  1995/01/17  19:09:42  ejaquith
 * ejaquith on Tue Jan 17 11:07:57 PST 1995
 *
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Changed interface to i_hausdorff1WayBestMatchMR()
 * to take a ptr to a "verify routine" when finding the best
 * match instead of ptrs to return the best
 * Hausdorff location and percent.  This shrinks the arg list.  Additionally,
 * it allows the client to decide if each found best match will be acceptable
 * (as the routine progresses through the rect) as the final best match.
 * (I'm using it to rule out certain sub-areas of the sub-rect where I don't
 * want the final match to be contained in).
 *
 * Revision 1.3  1995/01/09  17:38:08  lperry
 * lperry on Mon Jan 9 09:37:05 PST 1995
 *
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Fixed assignment of NULL ptr to strel dilateOp to eliminate
 * acc warning message.
 *
 * Revision 1.2  1995/01/07  21:47:53  ejaquith
 * i_hcmres.c -> /product/ipcore/ipshared/src/RCS/i_hcmres.c,v
 *
 * Made dilations of image at each level in the multi-level resolution
 * searching use the faster generated Assembly dilations (when possible).
 *
 * Revision 1.1  1995/01/04  18:55:17  ejaquith
 * Initial revision
 *

*/
