/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * i_hcptsms.c:	    Hausdorff binary image comparator
 *  	    	    when a model set is represented as a list of points.
 *
 *  Public routines:
 *
 *  [none, all routines called from corresponding default routines in
 *   i_hc32ms.c and from search routines in i_hcsrch.c]
 *
 *
 * Extensive comments concerning the Hausdorff comparator method can be found
 * at the top of i_hc32ms.c, the default method for use.
 *
 */


#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"
#include "memory.pub"
#include "defines.pub"

/* prototypes */
#include "shrpatt.prv"
#include "shrpatt.pub"
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_hcptms.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

static UInt32 bit32[32] = {
    0x80000000, 0x40000000, 0x20000000, 0x10000000,
    0x08000000, 0x04000000, 0x02000000, 0x01000000,
    0x00800000, 0x00400000, 0x00200000, 0x00100000,
    0x00080000, 0x00040000, 0x00020000, 0x00010000,
    0x00008000, 0x00004000, 0x00002000, 0x00001000,
    0x00000800, 0x00000400, 0x00000200, 0x00000100,
    0x00000080, 0x00000040, 0x00000020, 0x00000010,
    0x00000008, 0x00000004, 0x00000002, 0x00000001};

#ifdef pbmDEBUG
static UInt8 bit8[8] = {
    0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
#endif /*pbmDEBUG*/


/* PRIVATE */    
/******************************************************************************
 * Count the number of "on" bits in a "pt list" region
 *
 * Description:
 *  Given a pointer to an image bitmap represented as a list of
 *  x offsets in a row, and some bounds to define a specified region,
 *  count the number of "on" bits in this region.
 *
 * Returns:
 *  The number of "on" bits in the region.
 *****************************************************************************/
static Int32 CDECL 
ip_countOverlapPtList (
    IPState    *pIPState,
    const UInt16 *pX,	    /* Ptr to array of x offsets */
    UInt32	index,	    /* pX Index to first x offset of first row that is
			     * in region */
    UInt32	firstX,	    /* x offset of first column in region */
    UInt32	lastX,	    /* x offset of last column in region */
    UInt32	nRows)      /* number of rows in region */
{
    PROLOGUE;
    Int32 ovrlpCnt = 0;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    while (nRows > 0)
    {
	while (pX[index] < firstX)
	    index++;

	/* Overlapping portion of row, bound by the column offsets */
	while (pX[index] <= lastX)
	{
	    index++;
	    ovrlpCnt++;
	}

	/* Go until you hit the flag signifying end of the row */ 
	while (pX[index] != ROW_END)
	    index++;

	nRows--;
	index++;
    }

    return ovrlpCnt;
}


/* PRIVATE */
/****************************************************************************
 * Return space needed to make model set from image
 *
 * Description:
 *  Given the dimensions of the image to make a model set from, determine
 *  the space that will be needed in an allocator call.
 *  The model set will be filled in i_makeModelSet()
 *
 *  Returns:
 *  ia_successful	invocation done and okay
 *  ia_imageTooWide	too big to make a model set with current implementation
 *****************************************************************************/
Int32 CDECL 
ip_getModelSetSizePtList (
    IPState    *pIPState,
    UInt8      *pS,	     /* Ptr to upper left corner of image to make
			      * model set from, including frame */
    Int32	sBpl,	     /* Number of bytes per line, including padding
			      * and frame */
    Int32	sWidth,	     /* Width and height of image, including */
    Int32	sHeight,     /* frame, excluding padding */
    Int32      *pModelSetSize) /* Ptr to returned value of model set size. */
{
    PROLOGUE;
    Int32 npts;

    /* Make dimens unframed */
    sWidth -= 2*cFrameBits;
    sHeight -= 2*cFrameBits;

    if (sWidth > 65534 /* == 2^16 */)
 	return ia_imageTooWide;

    /* Skip binary frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    
    i_countBits (pIPState, pS, 0, 0, sWidth, sHeight, 1, sBpl, &npts);
     
    *pModelSetSize = sizeof(modelSet) + ((npts+sHeight)*sizeof(UInt16));
    
    NORMAL_RETURN;
}
    

/* PRIVATE */
/****************************************************************************
 * Make a model set from an image bitmap
 *
 * Description:
 *  Given an image and its dimensions, make a model set from it.
 *  This model set will be used in the Hausdorff comparator functions.
 *  NOTICE it may be desirable for the image to have the ADDITIONAL
 *  'dilateRadius' zeroed border around the image.
 *  (see note on 'dilateRadius' at top of ipshared/i_hc32ms.c)
 *
 * Returns:
 *  ia_successful	invocation done and okay
 *  ia_nomem    	could not allocate required space
 *  ia_imageTooWide	too big to make a model set with current implementation
 *
 * Notes:
 *  If the images are large, suggest downsampling them first.
 *  pM:
 *   *pM != NULL, the model set has been allocated by a client and it will
 *   simply get filled in here.
 *****************************************************************************/
Int32 CDECL 
ip_makeModelSetPtList (
    IPState    *pIPState,
    UInt8      *pS,	     /* Ptr to upper left corner of image to make
			      * model set from, including frame */
    Int32	sBpl,	     /* Number of bytes per line, including padding
			      * and frame */
    Int32	sWidth,	     /* Width and height of image, including */
    Int32	sHeight,     /* frame, excluding padding */
    void       *pM)           /* Ptr to internal representation of model
			      * set (see Notes above) */
{
    PROLOGUE;
    Int32     x, y, n;
    UInt16    *pX;
    modelSet  *set;
    UInt32    *pSWord;
    
    /* Skip binary frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pSWord = (UInt32 *)pS;

    /* Make dimens unframed */
    sWidth -= 2*cFrameBits;
    sHeight -= 2*cFrameBits;

    /* Point to the set struct portion of the model set memory chunk
     * and fill in the non-pointer fields of this struct portion */
    set = (modelSet *) pM;
    set->nBitmaps = 0;
    set->wplM = (sBpl >> 2);
    set->widthM = sWidth;
    set->heightM = sHeight;
    i_countBits (pIPState, pS, 0, 0, sWidth, sHeight, 1, sBpl, &(set->nptsM));

    /* This is sneaky and rotten and probably unprofessional, but this
     * is how we'll dereference the rest of the memory chunk AND
     * assign the model pointer to a location in the memory chunk.
     */
    set->pX = (UInt16 *)(set);
    set->pOffsetsM = NULL;
    set = ip_derefModelSetVoid (pIPState, (void *)pM);

    /* Make the point list from the image.
     * This will consist of a sequential stream of X offsets.
     * At the end of every row, a row end flag will be written instead of
     * a valid X offset to later let us know we are at the end of the row.
     */
    pX = set->pX;
    for (n = 0, y = 0; y < sHeight; y++)
    {
	/* Search the entire image row left to right for ON pixels */
	for (x = 0; x < sWidth; x++)
	{
	    if ((pSWord[x >> 5] & bit32[x & 31]) != 0)
	    {
		/* if pSWord[x][y] == 1*/
		pX[n] = (UInt16) x;
		n++;
	    }
	}
	
	/* At end of row, store row end flag */
	pX[n] = ROW_END;
	n++;
	
	pSWord += (sBpl >> 2); /* Move pSWord to the next row! */
    }

    hCHECK(n != set->nptsM + set->heightM,
	   "Moved past the allocated point list end");

    NORMAL_RETURN;
}    


/* PRIVATE */
/****************************************************************************
 * Hausdorff compares model and imagery
 *
 * Description:
 *  Take the model and the imagery and compare them to see if they match, at
 *  the appropriate passed in offset of the model over the imagery.  Do this by
 *  taking the model bits and ANDing them with the imagery bits.
 *
 * Returns:
 *  The number of model bits ON in the region.
 *
 * Notes:
 *  minAndedCnt is used for early termination.  It could be calculated inside
 *  this routine, but we unrolled it for speed.
 *****************************************************************************/
Int32 CDECL 
ip_hausdorffCompareOvrlpPtList (
    IPState    *pIPState,
    void       *pModelData,	/* Ptr to single model bitmap in point list
				 * representation, cast to a void * */
    Int32	wplM,		/* # words per line, including padding*/
    Int32	widthM,		/* # model columns (excluding padding/frame) */
    Int32	heightM,	/* # model rows (excluding padding/frame) */
    Int32	nptsM,		/* # on bits in entire model bitmap */
    Int32      *pOffsetsM,	/* Ptr to ordered offsets table */
    Int32	x,		/* Offset coordinates to place the model's */
    Int32	y,		/* upper left corner w.r.t. the upper left
				 * corner of the imagery */
    UInt32     *pI,		/* Ptr to upper left corner of imagery,
				 * EXCLUDING frame */
    Int32	wplI,		/* # words per line, including padding/frame */
    Int32	widthI,		/* # imagery columns(excluding padding/frame)*/
    Int32	heightI,	/* # imagery rows (excluding padding/frame)*/
    Float64	minPercent,	/* The imagery and model must correlate with
				 * a Hausdorff percent of at least this much
				 * to constitute a match, [0..1] */
    Int32	minAndedCnt,	/* Pass this in as CEIL(nptsM * minPercent) */
    Int32      *pAndedCnt)  	/* Return the number of total model bits
				 * that AND over imagery bits in the overlap
				 * ( >= minAndedCnt or -1) */
{
    PROLOGUE;
    Int32   index, firstX, lastX, firstY, lastY;
    UInt16  *pX = (UInt16 *)pModelData;
#ifdef pbmDEBUG
    UInt8   *MODEL, *pMODEL;
    Int32   rowCnt, MODELbpl = widthM / 8 + (widthM & 7 ? 1 : 0);
    MODEL = pMODEL = (UInt8 *)CALLOC(MODELbpl * heightM, sizeof(UInt8));

    index = 0;
    rowCnt = 0;
    while (rowCnt < heightM)
    {
	if (pX[index] == ROW_END)
	{
	    rowCnt++;
	    pMODEL += MODELbpl;
	}
	else
	    *(pMODEL + (pX[index] >> 3)) |= bit8[pX[index] & 7];

	index++;
    }
    DumpPBM("HD_Model.pbm", MODEL, MODELbpl, widthM + MAX(x,0), heightM);
    FREE(MODEL);
#endif /*pbmDEBUG*/

    /* Keep compiler happy */
    pOffsetsM = pOffsetsM;
    pIPState = pIPState;
    __error__ = __error__;


    /* Check here to avoid out of bounds (negative val for unsigned) below */
    if (x >= widthI || y >= heightI)
    {
	/* Means no portion of the imagery and model overlapped! */
	*pAndedCnt = -1;
	return nptsM;
    }
    
    /* Setup the bounds of the model data that we should check that overlaps
     * some part of the imagery
     * firstX and lastX are the X column indices (inclusive) of the model
     * that overlaps
     * firstY and lastY are the Y row indices (inclusive)
     */
    firstX = MAX(-x, 0);
    firstY = MAX(-y, 0);
    lastX  = MIN(widthM,  widthI  - x) - 1;
    lastY  = MIN(heightM, heightI - y) - 1;

    if (firstY > lastY || firstX > lastX)
    {
	/* Means no portion of the imagery and model overlapped! */
	*pAndedCnt = -1;
	return nptsM;
    }

    if (y > 0)
	pI += wplI * y;

    /* Advance to the first row of the Model that overlaps the imagery,
     * temporarily using the unused wplM parameter */
    index = 0;
    if (firstY != 0)
    {
	wplM = 0;
	while (wplM < firstY)
	{
	    if (pX[index] == ROW_END)
		(wplM)++;
	    index++;
	}
    }

    /* Count the number of total model points that are "on" in this
     * overlapping area since we aren't trying to match the whole model */
    nptsM = ip_countOverlapPtList (pIPState, pX,
				   index,
				   firstX,
				   lastX,
				   lastY - firstY + 1);

    /* Rescale the minAndedCnt appropriately */
    minAndedCnt = (Int32)(ceil(minPercent * nptsM));

    ip_correlatePtList (pIPState, pI,
		        lastY - firstY + 1,
		        wplI,
		        x,
		        pX,
		        index,
		        firstX,
		        lastX,
		        nptsM - minAndedCnt,
		        pAndedCnt);

    if (nptsM - minAndedCnt < *pAndedCnt)
	/* Means the above routine terminated early */
	*pAndedCnt = -1;
    else 
	/* NOTE that *pAndedCnt is actually the number of MISSES, so make it
	 * into what it should be, the number of not MISSES */
	*pAndedCnt = nptsM - *pAndedCnt;

    return (nptsM); /* The number of ON points in the model */
}


/* PRIVATE */
/****************************************************************************
 * Hausdorff compares model and imagery
 *
 * Description:
 *  Take the model and the imagery and compare them to see if they match, at
 *  the appropriate passed in offset of the model over the imagery.  Do this by
 *  taking the model bits and ANDing them with the imagery bits.
 *
 * Returns:
 *  The number of model bits ON in the entire model.
 *
 * Notes:
 *  minAndedCnt is used for early termination.  It could be calculated inside
 *  this routine, but we unrolled it for speed.
 *****************************************************************************/
Int32 CDECL 
ip_hausdorffComparePtList (
    IPState    *pIPState,
    void       *pModelData,	/* Ptr to single model bitmap in point list
				 * representation, cast to a void * */
    Int32	wplM,		/* # words per line, including padding*/
    Int32	widthM,		/* # model columns (excluding padding/frame) */
    Int32	heightM,	/* # model rows (excluding padding/frame) */
    Int32	nptsM,		/* # on bits in entire model bitmap */
    Int32      *pOffsetsM,	/* Ptr to ordered offsets table */
    Int32	x,		/* Offset coordinates to place the model's */
    Int32	y,		/* upper left corner w.r.t. the upper left
				 * corner of the imagery */
    UInt32     *pI,		/* Ptr to upper left corner of imagery,
				 * EXCLUDING frame */
    Int32	wplI,		/* # words per line, including padding/frame */
    Int32	widthI,		/* # imagery columns(excluding padding/frame)*/
    Int32	heightI,	/* # imagery rows (excluding padding/frame)*/
    Float64	minPercent,	/* The imagery and model must correlate with
				 * a Hausdorff percent of at least this much
				 * to constitute a match, [0..1] */
    Int32	minAndedCnt,	/* Pass this in as CEIL(nptsM * minPercent) */
    Int32      *pAndedCnt)  	/* Return the number of total model bits
				 * that AND over imagery bits in the overlap
				 * ( >= minAndedCnt or -1) */
{
    PROLOGUE;
    Int32   index, firstX, lastX, firstY, lastY;
    UInt16  *pX = (UInt16 *)pModelData;
#ifdef pbmDEBUG
    UInt8   *MODEL, *pMODEL;
    Int32   rowCnt, MODELbpl = widthM / 8 + (widthM & 7 ? 1 : 0);
    MODEL = pMODEL = (UInt8 *)CALLOC(MODELbpl * heightM, sizeof(UInt8));

    index = 0;
    rowCnt = 0;
    while (rowCnt < heightM)
    {
	if (pX[index] == ROW_END)
	{
	    rowCnt++;
	    pMODEL += MODELbpl;
	}
	else
	    *(pMODEL + (pX[index] >> 3)) |= bit8[pX[index] & 7];

	index++;
    }
    DumpPBM("HD_Model.pbm", MODEL, MODELbpl, widthM + MAX(x,0), heightM);
    FREE(MODEL);
#endif /*pbmDEBUG*/

    /* Keep compiler happy */
    minPercent = minPercent;
    pOffsetsM = pOffsetsM;
    pIPState = pIPState;
    __error__ = __error__;


    /* Check here to avoid out of bounds (negative val for unsigned) below */
    if (x >= widthI || y >= heightI)
    {
	/* Means no portion of the imagery and model overlapped! */
	*pAndedCnt = -1;
	return nptsM;
    }

    /* Setup the bounds of the model data that we should check that overlaps
     * some part of the imagery
     * firstX and lastX are the X column indices (inclusive) of the model
     * that overlaps
     * firstY and lastY are the Y row indices (inclusive)
     */
    firstX = MAX(-x, 0);
    firstY = MAX(-y, 0);
    lastX  = MIN(widthM,  widthI  - x) - 1;
    lastY  = MIN(heightM, heightI - y) - 1;

    if (firstY > lastY || firstX > lastX || nptsM == 0)
    {
	/* Means no portion of the imagery and model overlapped! */
	*pAndedCnt = -1;
	return nptsM;
    }

    if (y > 0)
	pI += wplI * y;

    index = 0;

    if (firstX == 0 && firstY == 0 &&
	lastX == widthM-1 && lastY == heightM-1)
    {
	ip_correlateFullPtList (pIPState, pI,
			        heightM,
			        wplI,
			        x,
			        pX,
			        nptsM - minAndedCnt,
			        pAndedCnt);
    }
    else
    {
	/* Advance to the first row of the Model that overlaps the imagery,
	 * temporarily using the unused wplM parameter*/
	if (firstY != 0)
	{
	    wplM = 0;
	    while (wplM < firstY)
	    {
		if (pX[index] == ROW_END)
		    (wplM)++;
		index++;
	    }
	}
	
	ip_correlatePtList (pIPState, pI,
			    lastY - firstY + 1,
			    wplI,
			    x,
			    pX,
			    index,
			    firstX,
			    lastX,
			    nptsM - minAndedCnt,
			    pAndedCnt);
    }
	
    if (nptsM - minAndedCnt < *pAndedCnt)
	/* Means the above routine terminated early */
	*pAndedCnt = -1;
    else 
	/* NOTE that *pAndedCnt is actually the number of MISSES, so make it
	 * into what it should be, the number of not MISSES */

	if (firstX == 0 && firstY == 0 &&
	    lastX == widthM-1 && lastY == heightM-1)
	{
	    /* The entire model fits in the image, the number of ANDs is thus
	     * the number of total model points minus the number of MISSes*/
	    *pAndedCnt = nptsM - *pAndedCnt;
	}
	else
	{
	    /* The number of ANDs is the number of model points in the
	     * overlapping region minus the number of MISSes */
	    *pAndedCnt *= -1;
	    *pAndedCnt += ip_countOverlapPtList (pIPState, pX,
						 index,
						 firstX,
						 lastX,
						 lastY - firstY + 1); 
	}
    
    return (nptsM); /* The number of ON points in the model */
}


/*

$Log: i_hcptms.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.11  1998/01/13 23:48:27  ddavies
i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.10  1995/12/13 02:44:33  ejaquith
i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.9  1995/10/18  03:57:27  ddavies
 * i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.8  1995/09/14  18:59:51  ejaquith
 * i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v
 *
 * CommonCheck: Converted all routines that were private in files to be of form
 * ip_*(), which they weren't.  No functionality/noticeable changes.
 *
 * Revision 1.7  1995/04/22  20:47:15  ejaquith
 * i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v
 *
 * Removed i_imagePutModelPtList().
 * Changed i_makeModelSetPtList() interface to require client to have
 * already allocated space for model set.
 * Please see general (large!) module comments for same date in i_hc32ms.c
 * Modified routines to comply with these new changes.
 *
 * Revision 1.6  1995/03/29  01:59:02  ejaquith
 * i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.5  1995/03/15  01:05:43  ejaquith
 * i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v
 *
 * Updated indention.
 * Addded calls to i_correlateFullPtList()
 *
 * Revision 1.4  1995/01/07  21:45:46  ejaquith
 * ejaquith on Sat Jan 7 13:44:57 PST 1995
 *
 * i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v
 *
 * Made calculating where the imagery and model overlap a bit faster and
 * cleaner.
 *
 * Revision 1.3  1995/01/04  18:40:56  ejaquith
 * i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v
 *
 * Abandoned use of "unit" (variable sized word for storage of models) for
 * use of 32 bit words.  Made bit counting safer.  Changed all public routines
 * o use image byte pointers and bytes per line (from image word pointers
 * and words per line).  Fixed bug in correlation percent calculation (minor).
 * Changed correlation loop calls appropriately to use number of MISSes when
 * bitwise ANDing imagery and model.
 *
 * Revision 1.2  1994/12/07  23:00:03  ejaquith
 * i_hcptms.c -> /product/ipcore/ipshared/src/RCS/i_hcptms.c,v
 *
 * Ref'd unused parameters to keep compiler happy.
 *
 * Revision 1.1  1994/12/06  23:39:22  ejaquith
 * Initial revision
 *

*/
