/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_hcsrch.c: 	Hausdorff search strategies for matches
 *
 *  Public routines:
 *
 *  	Code to init a subrect search
 *  	    i_hausdorffRectSearchSize()
 *  	    i_hausdorffRectSearchInit()
 *
 *  	Code to init a search through a list of translations 
 *  	    i_hausdorffListSearchSize()
 *  	    i_hausdorffListSearchInit()
 *
 *  	Code to search
 *  	    i_hausdorffSearch()
 *
 *
 *  Other extra miscellaneous public routines:
 *
 *  	Code to count the number of ON bits in a window at each row location
 *  	    i_makeRowXSums()
 *
 *  	Code to count the number of ON bits in a window at each image location
 *  	    i_countFilter()
 *
 *
 * Extensive comments concerning the Hausdorff comparator method can be found
 * at the top of i_hc32ms.c, the default method for use.
 *
 *
 */


#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"
#include "lmask.prv"
#include "rmask.prv"
#include "defines.pub"
#include "bitcount.pub"
#include "memory.pub"

/* prototypes */
#include "shrpatt.pub"
#include "shrpatt.prv"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_hcsrch.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

#ifdef hDEBUG
#include "shrip.pub"  /* for i_countBits() */
#endif /*hDEBUG*/



/* Forward declarations */
static Int32 CDECL 
ip_searchSparseFilter (
    IPState    *pIPState,
    searchType *pAB,		  /* Ptr to A and B info */
    modelSet   *pAMS,		  /* Ptr to A model set */
    UInt8     **ppBitmapsA,	  /* Array of stored A bitmaps */
    UInt32     *pAD,		  /* Ptr to upper left corner of the dilated
				   * imagery of A ("AD"), excluding frame */
    modelSet   *pBMS,		  /* Ptr to B model set */
    UInt8     **ppBitmapsB,	  /* Array of stored B bitmaps */
    UInt32     *pBD,		  /* Ptr to upper left corner of the dilated
				   * imagery of B ("BD"), excluding frame */
    Float64	minPercentAtoB,	  /* The A to B and B to A Hausdorff */
    Float64	minPercentBtoA,	  /* percents must be at least these values
				   * to match, [0..1] */
    Int32      *pX,		  /* Ptrs to returned match location (x,y) */
    Int32      *pY,
    Float64    *pPercentAtoB,	  /* Ptrs to returned Hausdorff percents at */
    Float64    *pPercentBtoA); 	  /* this location. */

static Int32 CDECL 
ip_searchEveryTrans (
    IPState    *pIPState,
    searchType *pAB,		  /* Ptr to A and B info */
    modelSet   *pAMS,		  /* Ptr to A model set */
    UInt8     **ppBitmapsA,	  /* Array of stored A bitmaps */
    UInt32     *pAD,		  /* Ptr to upper left corner of the dilated
				   * imagery of A ("AD"), excluding frame */
    modelSet   *pBMS,		  /* Ptr to B model set */
    UInt8     **ppBitmapsB,	  /* Array of stored B bitmaps */
    UInt32     *pBD,		  /* Ptr to upper left corner of the dilated
				   * imagery of B ("BD"), excluding frame */
    Float64	minPercentAtoB,	  /* The A to B and B to A Hausdorff */
    Float64	minPercentBtoA,	  /* percents must be at least these values
				   * to match, [0..1] */
    Int32      *pX,		  /* Ptrs to returned match location (x,y) */
    Int32      *pY,
    Float64    *pPercentAtoB,	  /* Ptrs to returned Hausdorff percents at */
    Float64    *pPercentBtoA); 	  /* this location. */

static Int32 CDECL 
ip_searchTransList (
    IPState    *pIPState,
    searchType *pAB,		  /* Ptr to A and B info */
    modelSet   *pAMS,		  /* Ptr to A model set */
    UInt8     **ppBitmapsA,	  /* Array of stored A bitmaps */
    UInt32     *pAD,		  /* Ptr to upper left corner of the dilated
				   * imagery of A ("AD"), excluding frame */
    modelSet   *pBMS,		  /* Ptr to B model set */
    UInt8     **ppBitmapsB,	  /* Array of stored B bitmaps */
    UInt32     *pBD,		  /* Ptr to upper left corner of the dilated
				   * imagery of B ("BD"), excluding frame */
    Float64	minPercentAtoB,	  /* The A to B and B to A Hausdorff */
    Float64	minPercentBtoA,	  /* percents must be at least these values
				   * to match, [0..1] */
    Int32      *pX,		  /* Ptrs to returned match location (x,y) */
    Int32      *pY,
    Float64    *pPercentAtoB,	  /* Ptrs to returned Hausdorff percents at */
    Float64    *pPercentBtoA); 	  /* this location. */

static Int32 CDECL 
ip_searchFlyFilter (
    IPState    *pIPState,
    searchType *pAB,		  /* Ptr to A and B info */
    modelSet   *pAMS,		  /* Ptr to A model set */
    UInt8     **ppBitmapsA,	  /* Array of stored A bitmaps */
    UInt32     *pAD,		  /* Ptr to upper left corner of the dilated
				   * imagery of A ("AD"), excluding frame */
    modelSet   *pBMS,		  /* Ptr to B model set */
    UInt8     **ppBitmapsB,	  /* Array of stored B bitmaps */
    UInt32     *pBD,		  /* Ptr to upper left corner of the dilated
				   * imagery of B ("BD"), excluding frame */
    Float64	minPercentAtoB,	  /* The A to B and B to A Hausdorff */
    Float64	minPercentBtoA,	  /* percents must be at least these values
				   * to match, [0..1] */
    Int32      *pX,		  /* Ptrs to returned match location (x,y) */
    Int32      *pY,
    Float64    *pPercentAtoB,	  /* Ptrs to returned Hausdorff percents at */
    Float64    *pPercentBtoA); 	  /* this location. */


/* Warning!  Make sure this doesn't clash with any search method in
 * hausconf.pub.
 */
#define TRANS_LIST         0x03


#define BIT_COUNT(word, bct) \
   bct[ word >> 24        ] + \
   bct[(word >> 16) & 0xff] + \
   bct[(word >>  8) & 0xff] + \
   bct[ word        & 0xff]


/* PUBLIC */
/*****************************************************************************
 * Count the number of bits in a window at each image location
 *
 * Description:
 *  Counts the number of ON bits in a certain width and height window in an
 *  image at every passed in location of the image range.
 *
 * Results:
 *  pppCntIm will point to a bitmap (w) x (h), and each each entry
 *  (*pppCntIm)[y][x] will contain the number of ON pixels in the rectangle
 *  x,y,(wFilter),(hFilter) in the original passed in pIm bitmap.
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_nomem	    	couldn't allocate needed memory
 *  ia_invalidParm    	filter size is too large
 *
 * Notes:
 *  The product of wFilter and hFilter must be less than 65,536
 *  To free *pppCntIm, free((*pppCntIm)[0]) and then free(*pppCntIm)
 *  This could be made a good deal faster, done in strips, and to allocate
 *  much less memory.  I'm using this mainly for debugging, so let me know
 *  if you use it and want it sped up!
 ****************************************************************************/
Int32 CDECL 
i_countFilter (
    IPState    *pIPState,
    Int32	wFilter,      /* Width and Height of window/filter we'd like */
    Int32	hFilter,      /* to count the number of ON bits inside of at
			       * every image location */
    UInt8      *pIm,	      /* Ptr to binary image bitmap upper left corner,
			       * excluding padding and frame */
    Int32	bpl,	      /* Number of bytes per line in image */
    Int32	w,	      /* Image width and height, excluding */
    Int32	h,	      /* padding and frame */
    const UInt8 *pBitCount,   /* Ptr to bit counting table, containing the
			       * number of 1's in a byte */
    UInt16   ***pppCntIm)     /* Ptr to bitmap of returned counting results */
{
    PROLOGUE;
    UInt16 **ppCntIm;
    UInt16 **ppTmp;
    UInt16  *pRow1, *pRow2, *pRow3, *pRow4, *pRowE;
    UInt16   tmp;
    Int32    x, addIndex, subIndex;
    Int8     div, mod;
    UInt32   subWord = 0, addWord, mask, *pImage;

    if (wFilter * hFilter >= 65536)
	return ia_invalidParm;

    /* This is a two pass algorithm that only needs to access longwords in
     * the original bitmap three times.
     * The first pass adds the number of bits ON in 1 high x (wFilter) windows
     * in the image.
     * The second pass adds (hFilter) first pass numbers together to make
     * the number of ON bits in the (hFilter) x (wFilter) window.
     */

    /* Make a temporary bitmap (w)x(h) that will contain (h) rows.
     * Each index, X, in a row corresponding to row y in the original image,
     * will store the number of ON bits in the original image bitmap from
     * [(X,y)...(X+wFilter-1,y)].
     *
     * Make a count bitmap (w)x(h) that will contain (h) rows.
     * Each index, X, in a row corresponding to row y in the original image,
     * will store the number of ON bits in the original image bitmap in the
     * subrectangle (X,y,wFilter,hFilter), a summation of hFilter appropriate
     * indexed temporary bitmap values,
     * ppTmp[y][X] + ppTmp[y+1][X] + ... + ppTmp[y+hFilter-1][X]
     */
    ppTmp   = (UInt16 **)CALLOC(h, sizeof(UInt16 *));
    ppCntIm = (UInt16 **)CALLOC(h, sizeof(UInt16 *));
    
    if (!ppTmp || !ppCntIm)
	return ia_nomem;
    
    ppTmp[0]   = (UInt16 *)CALLOC(w*h, sizeof(UInt16));
    ppCntIm[0] = (UInt16 *)CALLOC(w*h, sizeof(UInt16));

    if (!ppTmp[0] || !ppCntIm[0])
	return ia_nomem;

    /* Assign the row pointers to the temporary and count bitmaps */
    for (tmp = 1; tmp < (UInt16)h; tmp++)
    {
	ppTmp[tmp]   = ppTmp[tmp-1]   + w;
	ppCntIm[tmp] = ppCntIm[tmp-1] + w;
    }

    pImage = (UInt32 *)pIm; /* Start row, of binary image to sum */
    pRow1  = ppTmp[0];      /* Start row, of partial sums image  */
    pRowE  = ppTmp[h-1];    /* End row, of partial sums image */

    /* Convert bytes per line to words per line */
    bpl >>= 2;

    mod  = (Int8) (wFilter & 0x1F);
    div  = (Int8) (wFilter >> 5);
    /* We'll mask the last word of the words that lie under the first
     * wFilter pixels with this mask to get a valid count */
    mask = (mod == 0 ? 0x00000000 : leftMask[mod-1]);
    
    for ( ; pRow1 <= pRowE; pRow1 += w)
    {
	/* Sum the number of ON bits in the first wFilter pixels into the
	 * first tmp count */
	for (tmp = 0, x = 0; x < div; x++)
	{
	    addWord = pImage[x];
	    tmp += BIT_COUNT(addWord, pBitCount);
	}
	addWord = pImage[x] & mask;
	tmp += BIT_COUNT(addWord, pBitCount);
	
	/* The MSB of each of these words will be the interesting bits to
	 * subtract and add, respectively */
	subIndex = 0;        /* Subtract Word location */
	addIndex = div;      /* Add Word location */
	
	/* Do the middle of the row */
	for (x = 0; x <= w - wFilter; x++)
	{
	    if ((x & 0x1F) == 0)
	    {
		/* Time to re/load addWord and subWord */
		subWord  = pImage[subIndex++];
		addWord  = (mod != 0 ? (pImage[addIndex] << mod) |
			    (pImage[addIndex+1] >> (32 - mod)) :
			    pImage[addIndex]);
		addIndex++;
	    }

	    /* This is the current correct sum, from [x..x+wFilter-1] */
	    pRow1[x] = tmp;

	    /* Subtract the old bit that is no longer in the wFilter window,
	     * and add the new bit that has moved into this window, as we
	     * progress one in X down the row */
	    tmp -= (UInt16) ((subWord & 0x80000000) >> 0x1F);
	    tmp += (UInt16) ((addWord & 0x80000000) >> 0x1F);

	    /* Move out the old used MSBs and in with the new */
	    addWord <<= 1;
	    subWord <<= 1;
	}
	
	/* Do the end of the row, note no new pixels get added as the
	 * wFilter window is now partially over the padding/frame of the
	 * image */
	for ( ; x < w; x++)
	{
	    if ((x & 0x1F) == 0)
		/* Time to re/load the subWord */
		subWord  = pImage[subIndex++];
	    
	    pRow1[x] = tmp;

	    /* Subtract the old bit; there is no valid new bit to add */
	    tmp -= (UInt16) ((subWord & 0x80000000) >> 0x1F);
	    
	    subWord <<= 1;
	}

	/* Move down another row in the image to process it */
	pImage += bpl;
    }

    
    /* Now add up the hFilter of the partial sums rows for each full sum
     * row value to get the final number of ON bits in the
     * (wFilter) x (hFilter) window. */

    /* Sum the first hFilter pixels for the first output count */
    pRow1 = ppCntIm[0];
    for (x = 0; x < w; x++)
    {
	pRow2 = ppTmp[0];       /* Start row */
	pRow3 = ppTmp[hFilter-1]; /* End row */
	for (tmp = 0; pRow2 <= pRow3; pRow2 += w)
	    tmp += pRow2[x];
	pRow1[x] = tmp;
    }

    /* Do the middle of the column */
    pRow1 = ppCntIm[1];     /* Current row's sum */
    pRowE = ppCntIm[h-hFilter/*-1*/]; /* End row */ 
    pRow2 = ppCntIm[0];     /* Previous row's sum */
    pRow3 = ppTmp[0];       /* Old intermediate sum to subtract */
    pRow4 = ppTmp[hFilter]; /* New intermediate sum to add to current */
    while (pRow1 <= pRowE)
    {
	for (x = 0; x < w; x++)
	    /* Make the new full sum at this location by taking the previous
	     * (in Y!) location's sum and adding the new partial sum that
	     * is now in the hFilter window, and subtracting the old partial
	     * sum that has moved out of this window */
	    pRow1[x] = pRow2[x] - pRow3[x] + pRow4[x];
	
	/* Advance to the next rows */
	pRow1 += w;
	pRow2 += w;
	pRow3 += w;
	pRow4 += w;
    }
    
    /* Do the end of the column */
    /* Note that at these column locations, the hFilter high window now is
     * partially lying over the padding/frame, and thus there are no new
     * valid bits to add */
    pRowE = ppCntIm[h-1]; /* End Row */
    while (pRow1 <= pRowE)
    {
	for (x = 0; x < w; x++)
	    pRow1[x] = pRow2[x] - pRow3[x];

	/* Advance to the next rows */
	pRow1 += w;
	pRow2 += w;
	pRow3 += w;
    }
    
    FREE(ppTmp[0]);
    FREE(ppTmp);

    *pppCntIm = ppCntIm;
    NORMAL_RETURN;
}


/* PRIVATE */
/****************************************************************************
 * getModelData():
 *  	Return a void * pointer to the model data in the representation it is
 *  	in, given the appropriate x value if necessary
 *  	Returns NULL on error.
 ***************************************************************************/
void * CDECL 
ip_getModelData (
    IPState    *pIPState,
    modelSet   *pSet,
    UInt8     **ppBitmaps,
    Int32	x)
{
    PROLOGUE;

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    switch (pSet->nBitmaps)
    {
    case 32:
	return (void *)(ppBitmaps[x & WORD_SIZE_MINUS1]);
	break;
    case 0:
	return (void *)(pSet->pX);
	break;
    case 1:
	return (void *)(ppBitmaps[0]);
	break;
    default:
	return (void *)NULL;
    }
}


/* PRIVATE */
/****************************************************************************
 * countBitsBinary():
 *  Count the number of 1's in a rectangle of a binary image, starting
 *  at the passed in row pointer.
 *
 *  Input:
 *  	pS: 	    	Pointer to row source image data WITHOUT frame
 *  	leftSideMask:	Mask used to discard bits on the left side of the rect
 *  	rightSideMask:	Mask used to discard bits on the right side of the rect
 *  	leftBitNum: 	Index of leftmost bit in rect
 *  	rightBitNum: 	Index of rightmost bit in rect
 *  	h:  	    	height of source rectangle
 *  	bpl:	    	source image bytes/line.
 *  	pCount:	    	pointer to UInt32 that will contain the result
 *
 *      Returns:
 *              ia_successful            it worked
 *
 ************************************************************************/
static Int32 CDECL 
ip_countBitsBinary (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	leftSideMask,
    UInt32	rightSideMask,
    UInt32	leftBitNum,
    UInt32	rightBitNum,
    UInt32	h,
    UInt32	bpl,
    Int32      *pCount)
{
    PROLOGUE;
    Int32 total, w;
    
    /* Calculate the number of interesting words/line.  We always
     * start and stop on word boundaries to keep from being
     * confused by the little-endian / big endian issue.
     */
    w = ((rightBitNum - leftBitNum) / 32) + 1;
    total = 0;

    /* Adjust the starting address to word containing leftmost bit
     * bit of the rectangle.
     */
    pS += (leftBitNum / 8);

    /* Loop through the lines */
    for ( ; h > 0; h--, pS += bpl)
    {
	Int32 curTotal;
	ERR_PASS(ip_countBitsLine (pIPState, pS,
				   w,
				   leftSideMask,
				   rightSideMask,
				   (const UInt8 *) &(bitCount[0]),
				   &curTotal));
	total += curTotal;
    }
    
    /* return the result */
    *pCount = total;

    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * Count the number of bits in a window at each row location
 *
 * Description:
 *  Counts the number of ON bits in a certain width window in an image row
 *  at every passed in location of the image row range.
 *
 * Results:
 *  Will assign
 *  pRow[0]:            number ON bits in pIm[firstx  ..firstx+wFilter-1],
 *  pRow[1]:            number ON bits in pIm[firstx+1..firstx+wFilter], ...
 *  pRow[lastx-firstx]: number ON bits in pIm[lastx   ..lastx+wFilter-1]
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 ****************************************************************************/
Int32 CDECL 
i_makeRowXSums (
    IPState    *pIPState,
    UInt8      *pIm,	    /* Ptr to binary image row to count, without
			     * frame */
    Int32	firstx,	    /* First x position to count */
    Int32	lastx,	    /* Last x position to count */
    Int32	wFilter,    /* Width of counting window */
    UInt16     *pRow)       /* Ptr to resulting count row */
{
    PROLOGUE;
    Int8       mod;
    UInt16     cnt;
    UInt32     subWord = 0, addWord = 0;
    UInt32    *pIm32;
    Int32      addIndex, subIndex, origCnt;
    
    /* Temporarily the index of the last bit to count in this row */
    subWord = lastx + wFilter - 1;
    
    /* The first arg adjusts the starting address to word containing
     * leftmost bit of the rectangle.
     * The second arg is the number of interesting words/line.
     * We always start and stop on word boundaries to keep from being
     * confused by the little-endian / big endian issue.
     * The third and fourth arguments are the left side mask and the
     * right side mask to mask the leftmost and rightmost words in the
     * counting region.
     * But what the hay does this do, you say?
     * Counts the number of bits ON in pIm[lastx...lastx+wFilter-1].
     */
    ERR_PASS(ip_countBitsLine (pIPState, pIm + ((lastx & ~0x1F) / 8),
			       (((subWord & ~0x1F) -
				       (lastx & ~0x1F)) / 32) + 1,
			       rightMask[(~lastx) & 0x1F],
			       leftMask[subWord & 0x1F],
			       (const UInt8 *) &(bitCount[0]),
			       &origCnt));

    cnt = (UInt16) origCnt;
    pIm32 = (UInt32 *)pIm;

#ifdef ATTEMPT_AT_CLEVERNESS_WAS_SLOWER
    if (cnt == 0)
    {
	/* There were no ON bits in this last wFilter window.  Make a check
	 * to see if there are ANY ON bits at all in this row.  If not,
	 * return 1, leaving pRow _untouched_
	 */

	/* Check if all the words from firstx to lastx are 0 */
	addIndex = firstx >> 5;
	subIndex = lastx >> 5;

	if (addIndex == subIndex)
	{
	    /* There is only one common word in first..lastx range */
	    if (rightMask[(~firstx) & 0x1F] &
		leftMask[lastx & 0x1F] & pIm32[addIndex] == 0)
		/* And it's all zeros */
		return 1;
	}
	else if (rightMask[(~firstx) & 0x1F] & pIm32[addIndex++] == 0 &&
		 leftMask[lastx & 0x1F] & pIm32[subIndex--] == 0)
	{
	    while (addIndex <= subIndex)
	    {
		if (pIm32[addIndex] != 0 || pIm32[subIndex] != 0)
		    break;
		subIndex--;
		addIndex++;
	    }
	    
	    if (addIndex > subIndex)
		/* Only zeros in range */
		return 1;
	}
    }
#endif    
    
    /* The MSB of each of the words subWord and addWord will be the interesting
     * bits to subtract and add, respectively */

    /* Offset the pRow pointer so writing to pRow[firstx]..pRow[lastx]
     * below will actually be writing to pRow[0]..pRow[lastx-firstx]
     */
    pRow -= firstx;

    /* Write the last count for this row */
    pRow[lastx--] = cnt;
    

    /* Add Word location */
    addIndex = lastx >> 5;

    /* Load addWord and subWord. */
    mod = 0x1F - (lastx & 0x1F);
    if (mod != 0)
    {
	addWord  = pIm32[addIndex--] >> mod;
	addWord |= pIm32[addIndex]   << (32 - mod);

	mod = (Int8) (0x1F - (subWord & 0x1F));
	if (mod != 0)
	{
	    /* Subtract Word location */
	    subIndex = subWord >> 5;
	    
	    subWord  = pIm32[subIndex]   >> mod;
	    subWord |= pIm32[subIndex-1] << (32 - mod);
	}
	else
	    subWord = pIm32[subWord >> 5];
    }
    /* else mod == 0, means lastx is pointing at the LSB of pIm32[addIndex]
     * word, so we will load it first thing in the loop below */

    /* move subIndex to the appropriate word so the correct one will
     * be reloaded below */
    subIndex = addIndex + ((wFilter + 0x1F) >> 5);
	
    /* How far to right shift the subWord when it gets reloaded below */
    mod = (32 - (wFilter & 0x1F)) & 0x1F;

    /* Do the rest of the row */
    if (mod != 0)
    {
	for ( ; lastx >= firstx; lastx--)
	{
	    if ((lastx & 0x1F) == 0x1F)
	    {
		/* Time to re/load addWord and subWord,
		 * all of their valid bits have been left shifted out */
		addWord  = pIm32[addIndex--];
		subWord  = pIm32[subIndex--] >> mod;
		subWord |= pIm32[subIndex]   << (32 - mod);
	    }

	    /* Update the current count for the next x location */
	    cnt -= (UInt16) (subWord & 0x00000001); /* Subtract current LSB
						     * (0 or 1)*/
	    cnt += (UInt16) (addWord & 0x00000001); /* Add current LSB
						     * (0 or 1)*/

	    /* Shift out the old LSBs and shift in the new interesting bits */
	    addWord >>= 1;
	    subWord >>= 1;

	    /* Assign the current count, which is the number of ON bits from
	     * pIm[lastx..lastx+wFilter-1] */
	    pRow[lastx] = cnt;
	}
    }
    else
    {
	for ( ; lastx >= firstx; lastx--)
	{
	    if ((lastx & 0x1F) == 0x1F)
	    {
		addWord  = pIm32[addIndex--];
		subWord  = pIm32[subIndex--];
	    }

	    cnt -= (UInt16) (subWord & 0x00000001); /* Subtract current LSB
						     * (0 or 1)*/
	    cnt += (UInt16) (addWord & 0x00000001); /* Add current LSB
						     * (0 or 1)*/

	    addWord >>= 1;
	    subWord >>= 1;

	    pRow[lastx] = cnt;
	}
    }
    
    NORMAL_RETURN;
}


/* PRIVATE */
/*****************************************************************************
 * Dereference a void * to a search info struct
 *
 * Description:
 *  Dereference the void * and reset any pointers in the struct that may
 *  no longer be valid
 *
 * Returns:
 *  Properly casted searchType struct pointer.
 *****************************************************************************/
static searchType * CDECL 
ip_derefSearchInfoVoid (
    IPState    *pIPState,
    void       *pSearchInfo)
{
    PROLOGUE;
    searchType  *pInfo;

    /* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Point to the search info struct portion of the search info memory
     * chunk.
     */
    pInfo = (searchType *)pSearchInfo;

    switch (pInfo->searchMeth)
    {
    case SPARSE_FILTER:
	if (pInfo->pSumsXY != (UInt32 *)(pInfo + 1))
	{
	    Int32 i;
	    /* This is unexpected.  It may mean we haven't initially set the
	     * struct pointers yet to locations in the memory chunk.
	     * More likely, this could mean that the pointer to the void * has
	     * changed since when we last looked at it on account of garbage
	     * collection!
	     * Either way, we need to reassign the pointers in this
	     * struct to point to the approapriate areas in the memory chunk.
	     */
	    
	    /* This array contains wSums elements */
	    pInfo->pSumsXY = (UInt32 *)(pInfo + 1);

	    pInfo->ppSumsX = (UInt16 **)(pInfo->pSumsXY + pInfo->wSums);
	    
	    /* Assign the ppSumsX pointers.
	     * Each of the hSums pointers points to a row of wSums elements */
	    pInfo->ppSumsX[0] =	(UInt16 *)
		(pInfo->ppSumsX + pInfo->hSums);
	    
	    for (i = 1; i < pInfo->hSums; i++)
		pInfo->ppSumsX[i] =
		    pInfo->ppSumsX[i-1] + pInfo->wSums;
	}
	break;
    case TRANS_LIST:
	if (pInfo->pListX != (Int32 *)(pInfo + 1))
	{
	    /* Unexpected like above case */
	    pInfo->pListX = (Int32 *)(pInfo + 1);
	    pInfo->pListY = (Int32 *)(pInfo->pListX + pInfo->nTrans);
	}
	break;
    default:
	break;
    }

    return pInfo;
}


/* PUBLIC */
/****************************************************************************
 * Return space needed to make search info
 *
 * Description:
 *  Given the number of translations the client wants to search, return
 *  the space that will be needed in a client allocator call to hold a search
 *  info struct.
 *  Client then fills allocated search info struct with
 *  i_hausdorffListSearchInit().
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *
 * Notes:
 *  So long as the input parameters do not change, the size will not change.
 *  Thus any allocated search structure may be re-initialized if both searches
 *  would have identical calls to this routine.
 *****************************************************************************/
Int32 CDECL 
i_hausdorffListSearchSize (
    IPState    *pIPState,
    Int32	nTrans,		   /* Number of translations to search in
				    * translation list */
    UInt32     *pSearchInfoSize)   /* Ptr to returned value of search info
				    * size */
{
    PROLOGUE;

    *pSearchInfoSize = sizeof(searchType) + (2 * nTrans * sizeof(Int32));
    
    NORMAL_RETURN;
}


/* PUBLIC */
/****************************************************************************
 * Return space needed to make search info
 *
 * Description:
 *  Given some information about the search for matches of A in a subrect of
 *  B that we would like to invoke, return the space that will be needed in
 *  a client allocator call to hold a search info struct.
 *  Client then fills allocated search info struct with
 *  i_hausdorffRectSearchInit().
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_invalidParm  	bad desired search method passed in
 *  	    	    	OR something wrong with model set passed in
 *
 * Notes:
 *  So long as the input parameters do not change, the size will not change.
 *  Thus any allocated search structure may be re-initialized if both searches
 *  would have identical calls to this routine.
 *****************************************************************************/
Int32 CDECL 
i_hausdorffRectSearchSize (
    IPState    *pIPState,
    void       *pAM,		   /* Ptr to internal representation of model
				    * set "A" */
    Int32	w,		   /* Width of subrectangle of B we want to
				    * search */
    UInt8	searchMeth,	   /* Desired search method. One of:
				    * {SPARSE_FILTER, FLY_FILTER, EVERY_TRANS}
				    * May be OR-ed with
				    * USE_FULL_B_FOR_BTOA_PERCENT */
    UInt32     *pSearchInfoSize)   /* Ptr to returned value of search info
				    * size */
{
    PROLOGUE;
    modelSet   *pAMS;
    
    if (!(pAMS = ip_derefModelSetVoid (pIPState, pAM)))
	return ia_invalidParm;
    
    switch (searchMeth & ~((UInt8)USE_FULL_B_FOR_BTOA_PERCENT))
    {
    case FLY_FILTER:
    case EVERY_TRANS:
	*pSearchInfoSize = sizeof(searchType);
	break;
    case SPARSE_FILTER:
	/* The UInt32s are for pSumsXY 1D array
	 * The UInt16s are for ppSumsX 2D array
	 * The UInt16 ptrs are for the ptrs ppSumsX[..]
	 */
	*pSearchInfoSize =
	    sizeof(searchType) +
	    (w - pAMS->widthM + 1) * sizeof(UInt32) +
	    (pAMS->heightM + 1) * (w - pAMS->widthM + 1) * sizeof(UInt16) +
	    (pAMS->heightM + 1) * sizeof(UInt16 *);
	break;
    default:
	return ia_invalidParm;
    }
    
    NORMAL_RETURN;
}


/* PUBLIC */
/****************************************************************************
 * Set up a search; fill in a search info struct
 *
 * Description:
 *  Given information about "A" and "B" set up a search
 *  to find Hausdorff matches of A and B.
 *  The search will take A, a model set and its dilated imagery, and
 *  place it over/at B, another model set and dilated imagery,
 *  at each desired location in a list of translations.
 *  Each location that has Hausdorff correlation percents that are at least
 *  the client-specified values will be reported as a match in each iterative
 *  call to i_hausdorffSearch().
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_invalidParm  	ptr to search info isn't allocated and valid
 *  	    	    	OR something wrong with model set(s) passed in
 *
 * Notes:
 *  Translation locations:
 *    These refer to the location of A's upper left corner
 *    with respect to B.  i.e. (0,0) means align upper left corners, (-1,0)
 *    means place A so its upper left corner is 1 pixel to the left of B's.
 *    Translation values are unconstrained;  A may be placed anywhere relative
 *    to B including places where A only overlaps B partially.
 *    The passed in array of locations may be discarded upon return.
 *  1-Way/2-Way matching:
 *    In 1-Way matching AM is compared to BD to return the directed Hausdorff
 *    percent from A to B.
 *    In 2-Way matching, additionally BM is compared to AD to return the
 *    directed Hausdorff percent from B to A.
 *    If the client will never desire the B to A percent to be calculated
 *    and returned in the search, then BM may be passed in NULL, and the
 *    AD parameters will be ignored.
 *****************************************************************************/
Int32 CDECL 
i_hausdorffListSearchInit (
    IPState    *pIPState,
    void       *pAM,		   /* Ptr to internal representation of model
				    * set "A" */
    UInt8      *pBD,		   /* Ptr to the upper left corner of the
				    * dilated imagery of "B", including frame*/
    Int32	bplBD,		   /* Number of bytes in imagery "B" line/row,
				    * including padding and frame */
    Int32	widthBD,	   /* imagery "B" width and height, */
    Int32	heightBD,	   /* including frame, excluding padding */
    void       *pBM,		   /* Ptr to internal representation of model
				    * set "B" */
    UInt8      *pAD,		   /* Ptr to the upper left corner of the
				    * dilated imagery of "A", including frame*/
    Int32	bplAD,		   /* Number of bytes in imagery "A" line/row,
				    * including padding and frame */
    Int32	widthAD,	   /* imagery "A" width and height, */
    Int32	heightAD,	   /* including frame, excluding padding */
    Int32      *pListX,		   /* Array of X and Y components of desired */
    Int32      *pListY,		   /* translations to search */
    Int32	nTrans,		   /* Size of above arrays; number of */
				   /* translations to search */
    UInt8	searchMeth,	   /* Desired search method. Either 0 or
				    * USE_FULL_B_FOR_BTOA_PERCENT */
    void       *pSearchInfo)       /* Ptr to client allocated search info
				    * that will be filled in here
				    * (see i_hausdorffListSearchSize()) */
{
    PROLOGUE;
    modelSet    *pAMS, *pBMS;
    searchType  *pAB;
    Int32        i;
    
    /* Keep anal compilers happy */
    pBD = pBD;
    pAD = pAD;

    if (!pSearchInfo)
	return ia_invalidParm;

    if (!(pAMS = ip_derefModelSetVoid (pIPState, pAM)))
	return ia_invalidParm;

    pAB = (searchType *)pSearchInfo;
    
    pAB->searchMeth = TRANS_LIST;
    pAB->nTrans = nTrans;
    pAB->pListX = NULL;
    pAB->pListY = NULL;

    /* Properly cast this and set up the pointers in the searchType struct
     * that we will use for this search method. */
    pAB = ip_derefSearchInfoVoid (pIPState, pSearchInfo);
    
    pAB->ip_searchRoutine = ip_searchTransList;


    /* Set up the translation list */
    pAB->nTrans = nTrans;
    for (i = 0; i < nTrans; i++)
    {
	pAB->pListX[i] = pListX[i];
	pAB->pListY[i] = pListY[i];
    }

    pAB->wplAD    = bplAD >> 2;
    pAB->widthAD  = widthAD - cFrameBits*2;
    pAB->heightAD = heightAD - cFrameBits*2;

    pAB->wplBD    = bplBD >> 2;
    pAB->widthBD  = widthBD - cFrameBits*2;
    pAB->heightBD = heightBD - cFrameBits*2;

    /* Determine which underlying image comparison routine we will use
     * based on the way the model set's data is stored
     */
    switch (pAMS->nBitmaps)
    {
    case 32:
	pAB->hausdorffRoutineAtoB = ip_hausdorffCompare;
	break;
    case 0:
	pAB->hausdorffRoutineAtoB = ip_hausdorffComparePtList;
	break;
    case 1:
	pAB->hausdorffRoutineAtoB = ip_hausdorffCompareOneModel;
	break;
    default:
	return ia_invalidParm;
	break;
    }

    if (pBM != NULL)
    {
	if (!(pBMS = ip_derefModelSetVoid (pIPState, pBM)))
	    return ia_invalidParm;

	switch (pBMS->nBitmaps)
	{
	case 32:
	    pAB->hausdorffRoutineBtoA = (searchMeth &
					 USE_FULL_B_FOR_BTOA_PERCENT ?
					 ip_hausdorffCompare :
					 ip_hausdorffCompareOvrlp);
	    break; 
	case 0:
	    pAB->hausdorffRoutineBtoA = (searchMeth & 
					 USE_FULL_B_FOR_BTOA_PERCENT ?
					 ip_hausdorffComparePtList :
					 ip_hausdorffCompareOvrlpPtList);
	    break; 
	case 1:
	    pAB->hausdorffRoutineBtoA = (searchMeth & 
					 USE_FULL_B_FOR_BTOA_PERCENT ?
					 ip_hausdorffCompareOneModel :
					 ip_hausdorffCompareOvrlpOneModel);
	    break;
	default:
	    return ia_invalidParm;
	    break;
	}
    }
    else
	/* Client will never want to do 2-Way matching */
	pAB->hausdorffRoutineBtoA = NULL;
    

    /* If we ordered the A model rows for faster comparison, set up the
     * row offsets for this B image
     */
    if (pAMS->pOffsetsM)
	ip_rowMakeOffsetsFromOrder (pIPState, pAMS->pOffsetsM + 2*pAMS->heightM,
				    NULL, pAMS->heightM, pAB->wplBD,
				    pAMS->pOffsetsM + pAMS->heightM);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set up a search; fill in a search info struct
 *
 * LongDescription:::
 *  Given information about "A" and "B" set up a search
 *  to find Hausdorff matches of A and B.
 *  The search will take A, a model set and its dilated imagery, and
 *  place it over/at B, another model set and dilated imagery,
 *  at each desired location in a subrect of B.
 *  Each location that has Hausdorff correlation percents that are at least
 *  the client-specified values will be reported as a match in each iterative
 *  call to i_hausdorffSearch().
 *
 * Input:::
 *	pAM::		Ptr to internal representation of model  
 *			set "A"
 *	pBD::		Ptr to the upper left corner of the      
 *			dilated imagery of "B", including frame
 *	bplBD::		Number of bytes in imagery "B" line/row, 
 *			including padding and frame
 *	widthBD::	imagery "B" width and height,
 *	heightBD::	including frame, excluding padding
 *	pBM::		Ptr to internal representation of model  
  *			set "B"
 *	pAD::		Ptr to the upper left corner of the      
 *			dilated imagery of "A", including frame
 *	bplAD::		Number of bytes in imagery "A" line/row
 *			including padding and frame
 *	widthAD::	imagery "A" width and height,
 *	heightAD::	including frame, excluding padding
 *	x::		Rectangular area in B where we search
 *	y::		for A to match.
 *	w::		May extend outside of B (only) when
 *	h::		using EVERY_TRANS
 *	skipx::		Amounts to skip.  Check for a match
 *	skipy::		every skipx pixel in X, every skipy pixel
 *			in Y in subrect translation area
 *	searchMeth::	Desired search method. One of:      
 *			{SPARSE_FILTER, FLY_FILTER, EVERY_TRANS}
 *			May be OR-ed with                   
 *			USE_FULL_B_FOR_BTOA_PERCENT
 *
 * Output:::
 *	pSearchInfo::	Ptr to client allocated search info 
 *			that will be filled in here         
 *			(see i_hausdorffRectSearchSize())
 *
 * Returns:::
 *  ia_successful::   	invocation done and okay
 *  ia_invalidParm::  	ptr to search info isn't allocated and valid
 *  	    	    	OR invalid search method
 *  	    	    	OR invalid subrect and search method combination
 *			OR no valid A translations in subrect (too small)
 *  	    	    	OR something wrong with model set(s) passed in
 *
 * Notes:::
 *  Translation locations::
 *    These refer to the location of A's upper left corner
 *    with respect to B.  i.e. (0,0) means align upper left corners, (-1,0)
 *    means place A so its upper left corner is 1 pixel to the left of B's.
 *    The subrect of B may extend outside of B if (and only if) the search
 *    method EVERY_TRANS is used.
 *    Match locations will be in the frame of the full B image.
 *  Skipx, skipy::
 *    If skipx == skipy == 1, will try every translation in subrect
 *  1-Way/2-Way matching::
 *    In 1-Way matching AM is compared to BD to return the directed Hausdorff
 *    percent from A to B.
 *    In 2-Way matching, additionally BM is compared to AD to return the
 *    directed Hausdorff percent from B to A.
 *    If the client will never desire the B to A percent to be calculated
 *    and returned in the search, then BM may be passed in NULL, and the
 *    AD parameters will be ignored.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_hausdorffRectSearchInit (
    IPState    *pIPState,
    void       *pAM,
    UInt8      *pBD,
    Int32	bplBD,
    Int32	widthBD,
    Int32	heightBD,
    void       *pBM,
    UInt8      *pAD,
    Int32	bplAD,
    Int32	widthAD,
    Int32	heightAD,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	skipx,
    Int32	skipy,
    UInt8	searchMeth,
    void       *pSearchInfo) 
{
    PROLOGUE;
    modelSet    *pAMS, *pBMS;
    searchType  *pAB;

    /* Keep anal compilers happy */
    pAD = pAD;

    if (!pSearchInfo)
	return ia_invalidParm;

    if (!(pAMS = ip_derefModelSetVoid (pIPState, pAM)))
	return ia_invalidParm;

    pAB = (searchType *)pSearchInfo;

    switch (searchMeth & ~((UInt8)USE_FULL_B_FOR_BTOA_PERCENT))
    {
    case EVERY_TRANS:
	pAB->searchMeth = EVERY_TRANS;
	
	pAB = ip_derefSearchInfoVoid (pIPState, pSearchInfo);

	pAB->ip_searchRoutine = ip_searchEveryTrans;

	/* Set up to make our first translation to check (x,y)! */
	pAB->tx = x - skipx;
	pAB->ty = y;
	break;
    case FLY_FILTER:
	pAB->searchMeth = FLY_FILTER;
	
	pAB = ip_derefSearchInfoVoid (pIPState, pSearchInfo);

	pAB->ip_searchRoutine = ip_searchFlyFilter;

	/* Set up to make our first translation to check (x,y)! */
	pAB->tx = x - skipx;
	pAB->ty = y;
	pAB->flyDirec = -1;
	break;
    case SPARSE_FILTER:
    {
	Int32    i, col, hSumsM1;
	UInt16  *pRow;
	UInt8   *pIm;
	
	/* Assign these two pointers values that will make deref..()
	 * think we have garbage collected and will reset the pointers
	 * to where they SHOULD be!
	 */
	pAB->pSumsXY = NULL;
	pAB->ppSumsX = NULL;
	pAB->wSums = w - pAMS->widthM + 1;
	pAB->hSums = pAMS->heightM + 1;
	pAB->searchMeth = SPARSE_FILTER;

	pAB = ip_derefSearchInfoVoid (pIPState, pSearchInfo);

	pAB->ip_searchRoutine = ip_searchSparseFilter;

	/* Set up to make our first translation to check (x,y)! */
	pAB->tx = x - skipx;
	pAB->ty = y;
	
	/* Set up the box filtering summing information for the first row in
	 * the search subrect.
	 * ppSumsX is be a stack of (hSums) rows, each (wSums) wide.
	 * pSumsXY will be a single row (wSums) wide, which for the first row,
	 * is the sum of each of the rows in the range
	 * ppSumsX[0]..ppSumsX[hSums-2].
	 * For the first row, ppSumsX[hSums-1] will be the extra row which we
	 * write to when we move down one row.
	 * Thus, the value pSumsXY[0] = ppSumsX[0][0] + ppSumsX[1][0] + ... +
	 * ppSumsX[hSums-2][0], equivalently the number of ON bits
	 * in the rectangle x, y, wSums, hSums-1.
	 */
	/* Skip binary frame */
	pIm = pBD + (bplBD << cLogFrameBits) + cFrameBytes;

	hSumsM1 = pAB->hSums - 1;
	for (i = 0; i < hSumsM1; i++)
	{
	    i_makeRowXSums (pIPState, pIm,
			    x,
			    x + pAB->wSums - 1,
			    pAMS->widthM,
			    pAB->ppSumsX[i]);
	    pIm += bplBD;
	}
	/* Extra row index which we will write to when we move down one row. */
	pAB->addRow = hSumsM1;

	/* Now add up the hSums-1 sumsX in each column of pSumsXY and store
	 * the totals in that column of pSumsXY
	 */
	for (col = 0; col < pAB->wSums; col++)
	    pAB->pSumsXY[col] = 0;
	
	for (i = 0; i < hSumsM1; i++)
	{
	    pRow = pAB->ppSumsX[i];
	    for (col = 0; col < pAB->wSums; col++)
		pAB->pSumsXY[col] += pRow[col];
	}
	
	break;
    }
    default:
	return ia_invalidParm;
    }
    
    pAB->wplAD    = bplAD >> 2;
    pAB->widthAD  = widthAD - cFrameBits*2;
    pAB->heightAD = heightAD - cFrameBits*2;

    pAB->wplBD    = bplBD >> 2;
    pAB->widthBD  = widthBD - cFrameBits*2;
    pAB->heightBD = heightBD - cFrameBits*2;

    pAB->firstx = x;
    pAB->firsty = y;
    pAB->lastx = x + w - pAMS->widthM;
    pAB->lasty = y + h - pAMS->heightM;
    pAB->skipx = skipx;
    pAB->skipy = skipy;

    /* Check to make sure there is at least one translation to check in the
     * desired subrect
     */
    if (pAB->firstx > pAB->lastx  ||  pAB->firsty > pAB->lasty)
	return ia_invalidParm;

    /* Check to make sure subrect only extends outside of B with allowable
     * search method EVERY_TRANS, and o.w. does not
     */
    if (((searchMeth & ~(UInt8)USE_FULL_B_FOR_BTOA_PERCENT) != EVERY_TRANS) &&
	(pAB->firstx < 0 || pAB->firsty < 0 ||
	 pAB->lastx + pAMS->widthM  > pAB->widthBD ||
	 pAB->lasty + pAMS->heightM > pAB->heightBD))
	return ia_invalidParm;

    /* Determine which underlying image comparison routine we will use
     * based on the way the model set's data is stored
     */
    switch (pAMS->nBitmaps)
    {
    case 32:
	pAB->hausdorffRoutineAtoB = ip_hausdorffCompare;
	break;
    case 0:
	pAB->hausdorffRoutineAtoB = ip_hausdorffComparePtList;
	break;
    case 1:
	pAB->hausdorffRoutineAtoB = ip_hausdorffCompareOneModel;
	break;
    default:
	return ia_invalidParm;
	break;
    }

    if (pBM != NULL)
    {
	if (!(pBMS = ip_derefModelSetVoid (pIPState, pBM)))
	    return ia_invalidParm;

	switch (pBMS->nBitmaps)
	{
	case 32:
	    pAB->hausdorffRoutineBtoA = (searchMeth &
					 USE_FULL_B_FOR_BTOA_PERCENT ?
					 ip_hausdorffCompare :
					 ip_hausdorffCompareOvrlp);
	    break; 
	case 0:
	    pAB->hausdorffRoutineBtoA = (searchMeth & 
					 USE_FULL_B_FOR_BTOA_PERCENT ?
					 ip_hausdorffComparePtList :
					 ip_hausdorffCompareOvrlpPtList);
	    break; 
	case 1:
	    pAB->hausdorffRoutineBtoA = (searchMeth & 
					 USE_FULL_B_FOR_BTOA_PERCENT ?
					 ip_hausdorffCompareOneModel :
					 ip_hausdorffCompareOvrlpOneModel);
	    break;
	default:
	    return ia_invalidParm;
	    break;
	}
    }
    else
	/* Client will never want to do 2-Way matching */
	pAB->hausdorffRoutineBtoA = NULL;
    

    /* If we ordered the A model rows for faster comparison, set up the
     * row offsets for this B image
     */
    if (pAMS->pOffsetsM)
	ip_rowMakeOffsetsFromOrder (pIPState, pAMS->pOffsetsM + 2*pAMS->heightM,
				    NULL, pAMS->heightM, pAB->wplBD,
				    pAMS->pOffsetsM + pAMS->heightM);
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find a(nother) match of A in B search area
 *
 * LongDescription:::
 *  This is an iterative routine that will take A and search for a match
 *  through all desired locations of A over B.  The locations to try are
 *  set up and the search gets initialized using
 *  [i_hausdorffRectSearchSize() and i_hausdorffRectSearchInit()] or
 *  [i_hausdorffListSearchSize() and i_hausdorffListSearchInit()].<BR>
 *
 *  The match of A and B is based on the passed in percents, which may
 *  change at each iterative invocation if desired.
 *
 * Input:::
 *  pAM::            Ptr to internal representation of model
 *                   set "A"
 *  ppBitmapsA::     Array of ptrs to stored A bitmaps.  Each
 *                   ptr points to upper left corner of a
 *                   bitmap, excluding any frame and padding
 *  nBitmapsA::      Number of stored A bitmaps, 0,1,32 
 *  pBD::            Ptr to the upper left corner of the
 *                   dilated imagery of "B", including frame
 *  pBM::            Ptr to internal representation of model
 *                   set "B" 
 *  ppBitmapsB::     Array of ptrs to stored B bitmaps.  Each
 *                   ptr points to upper left corner of a
 *                   bitmap, excluding any frame and padding
 *  nBitmapsB::      Number of stored B bitmaps, 0,1,32 
 *  pAD::            Ptr to the upper left corner of the
 *                   dilated imagery of "A", including frame
 *  pSearchInfo::    Ptr to search info 
 *  minPercentAtoB:: Minimum Hausdorff A to B percent that will 
 *                   define a match of A and B for any
 *                   remaining translations in the search
 *                   whose percents are at least these 
 *  minPercentBtoA:: Minimum Hausdorff B to A percent that will 
 *                   define a match of A and B for any
 *                   remaining translations in the search
 *                   whose percents are at least these 
 *
 * Output:::
 *  pX::             Ptr to returned match X location
 *  pY::             Ptr to returned match Y location
 *  pPercentAtoB::   Ptr to returned Hausdorff A to B percent at this location 
 *  pPercentBtoA::   Ptr to returned Hausdorff B to A percent at this location.
 *  pDoneSearch::    Ptr to 1 on return if search area is
 *                   exhausted (no valid match returned).
 *                   O.w., ptr to 0 (valid match returned) 
 *
 * Returns:::
 *  ia_successful::     invocation done and okay
 *  ia_invalidParm::    wanted 2-Way matching info but previously initted
 *                      to 1-Way only
 *                      OR invalid nBitmapsA OR invalid nBitmapsB
 *                      OR something wrong with model set(s) passed in
 *
 * Notes:::
 *  If minPercentBtoA < 0::
 *    - the routine won't calculate the *pPercentBtoA 2-Way Hausdorff percent,
 *      and will just return -1. instead. (1-Way matching)
 *    - parameters pBM, ppBitmapsB, nBitmapsB, and pAD will be ignored.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_hausdorffSearch (
    IPState    *pIPState,
    void       *pAM,
    UInt8     **ppBitmapsA,
    UInt8	nBitmapsA,
    UInt8      *pBD,
    void       *pBM,
    UInt8     **ppBitmapsB,
    UInt8	nBitmapsB,
    UInt8      *pAD,
    void       *pSearchInfo,
    Float64	minPercentAtoB,
    Float64	minPercentBtoA,
    Int32      *pX,
    Int32      *pY,
    Float64    *pPercentAtoB,
    Float64    *pPercentBtoA,
    UInt8      *pDoneSearch)
{
    PROLOGUE;
    searchType  *pAB;
    modelSet    *pAMS;

    /* Deref this pointer to get our match info struct and reassign any
     * struct pointers in this memory block that may have moved
     */
    pAB = ip_derefSearchInfoVoid (pIPState, pSearchInfo);

    if (!(pAMS = ip_derefModelSetVoid (pIPState, pAM)))
	return ia_invalidParm;

    if (pAMS->nBitmaps != nBitmapsA)
	return ia_invalidParm;

    if (minPercentBtoA >= 0.)
    {
	/* Do 2-Way matching */
	modelSet *pBMS;
    
	if (!pBM || !pAD)
	    /* Client wanted B to A percent to be returned but didn't follow
	     * rules of passing in B to A info when client called
	     * i_hausdorffSearchInit()!
	     */
	    return ia_invalidParm;
    
	if (!(pBMS = ip_derefModelSetVoid (pIPState, pBM)))
	    return ia_invalidParm;

	if (pBMS->nBitmaps != nBitmapsB)
	    return ia_invalidParm;
	
	pAB->ip_searchRoutine (pIPState, pAB,
			       pAMS,
			       ppBitmapsA,
			       (UInt32 *)(pAD
					+ (pAB->wplAD << (2 + cLogFrameBits))
					+ cFrameBytes),
			       pBMS,
			       ppBitmapsB,
			       (UInt32 *)(pBD
					+ (pAB->wplBD << (2 + cLogFrameBits))
					+ cFrameBytes),
			       minPercentAtoB,
			       minPercentBtoA,
			       pX,
			       pY,
			       pPercentAtoB,
			       pPercentBtoA);
    }
    else
    {
	/* Do 1-Way matching */
	pAB->ip_searchRoutine (pIPState, pAB,
			       pAMS,
			       ppBitmapsA,
			       NULL,
			       NULL,
			       NULL,
			       (UInt32 *)(pBD
					+ (pAB->wplBD << (2 + cLogFrameBits))
					+ cFrameBytes),
			       minPercentAtoB,
			       -1.,
			       pX,
			       pY,
			       pPercentAtoB,
			       pPercentBtoA);
    }
	
    *pDoneSearch = (*pPercentAtoB < 0. ? 1 : 0);

    NORMAL_RETURN;
}


/* Why bother with a routine when you can use a macro?  =)
 * No really, this ACTUALLY adds to the readability of this file greatly
 * because it allows the searching strategies to be compared.
 * Also, as all the variables are already set up in the search routines,
 * I couldn't see a reason to make this a routine and add extra overhead.
 * Thus, we have this macro, which allows editting, viewing, and comparing
 * code pieces in this file much easier.
 *
 * Used in:
 *  i_searchSparseFilter()
 *  i_searchEveryTrans()
 *  i_searchTransList()
 *  i_searchFlyFilter()
 */
#define HAUS_CHECK_MATCH_MACRO() \
{\
    pAB->hausdorffRoutineAtoB(pIPState,\
			      ip_getModelData (pIPState, pAMS, ppBitmapsA, x),\
			      pAMS->wplM, \
			      pAMS->widthM, \
			      pAMS->heightM, \
			      pAMS->nptsM, \
			      pAMS->pOffsetsM, \
			      x, \
			      y, \
			      pBD, \
			      pAB->wplBD, \
			      pAB->widthBD, \
			      pAB->heightBD, \
			      minPercentAtoB, \
			      minAndedCntAtoB, \
			      &andedCntAtoB); \
    if (andedCntAtoB >= minAndedCntAtoB) \
    { \
	/* A to B percent is good enough */ \
	if (minAndedCntBtoA < 0) \
	{ \
	    /* 1-Way matching only, so we want to report this match \
	     * and return */ \
	    pAB->tx = *pX = x; \
	    pAB->ty = *pY = y; \
	    *pPercentAtoB = HAUS_PERCENT(andedCntAtoB, pAMS->nptsM); \
	    *pPercentBtoA = -1.; \
	} \
	else \
	{ \
	    /* 2-Way matching, so check B to A percent */ \
	    Int32  ovrlpCntBtoA; \
	     \
	    ovrlpCntBtoA = \
		pAB->hausdorffRoutineBtoA(pIPState,\
					  ip_getModelData (pIPState, \
							   pBMS, \
							   ppBitmapsB, -x), \
					  pBMS->wplM, \
					  pBMS->widthM, \
					  pBMS->heightM, \
					  pBMS->nptsM, \
					  pBMS->pOffsetsM, \
					  -x, \
					  -y, \
					  pAD, \
					  pAB->wplAD, \
					  pAB->widthAD, \
					  pAB->heightAD, \
					  minPercentBtoA, \
					  minAndedCntBtoA, \
					  &andedCntBtoA); \
	     \
	    /* Catch divide by 0s */ \
	    if (ovrlpCntBtoA != 0) \
	    { \
		*pPercentBtoA = HAUS_PERCENT(andedCntBtoA, \
					     ovrlpCntBtoA); \
		if (*pPercentBtoA >= minPercentBtoA) \
		{ \
		    /* B to A percent is also good enough, we want to \
		     * report this match and return */ \
		    pAB->tx = *pX = x; \
		    pAB->ty = *pY = y; \
		    *pPercentAtoB = HAUS_PERCENT(andedCntAtoB, \
						 pAMS->nptsM); \
		} \
	    } \
	} \
    } \
}


/* PRIVATE */
/*****************************************************************************
 * Find a(nother) match of A in B search area
 *
 * Description:
 *  This is an iterative routine that will take A and search for a match
 *  through all desired locations of A over B.  The locations to try are
 *  set up and the search gets initialized using
 *  i_hausdorff..SearchSize() and i_hausdorff..SearchInit().
 *  The match of A and B is based on the passed in percents, which may
 *  change at each iterative invocation if desired.
 *  Search simply trying every translation in a list.
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *
 * Notes:
 *  If minPercentBtoA < 0., the routine won't calculate the *pPercentBtoA
 *  2-Way Hausdorff percent, and will just return -1. instead.
 *  If *pPercentAtoB is returned < 0., it means we are done the search.
 ****************************************************************************/
static Int32 CDECL 
ip_searchTransList (
    IPState    *pIPState,
    searchType *pAB,		  /* Ptr to A and B info */
    modelSet   *pAMS,		  /* Ptr to A model set */
    UInt8     **ppBitmapsA,	  /* Array of stored A bitmaps */
    UInt32     *pAD,		  /* Ptr to upper left corner of the dilated
				   * imagery of A ("AD"), excluding frame */
    modelSet   *pBMS,		  /* Ptr to B model set */
    UInt8     **ppBitmapsB,	  /* Array of stored B bitmaps */
    UInt32     *pBD,		  /* Ptr to upper left corner of the dilated
				   * imagery of B ("BD"), excluding frame */
    Float64	minPercentAtoB,	  /* The A to B and B to A Hausdorff */
    Float64	minPercentBtoA,	  /* percents must be at least these values
				   * to match, [0..1] */
    Int32      *pX,		  /* Ptrs to returned match location (x,y) */
    Int32      *pY,
    Float64    *pPercentAtoB,	  /* Ptrs to returned Hausdorff percents at */
    Float64    *pPercentBtoA)  	  /* this location. */
{
    PROLOGUE;
    Int32       x, y;
    Int32       andedCntAtoB, minAndedCntAtoB;
    Int32       andedCntBtoA, minAndedCntBtoA;
    

    /* These will be the threshold number of bits that need to be overlapping
     * to constitute a match */
    minAndedCntAtoB = (Int32)(ceil(minPercentAtoB * pAMS->nptsM));
    minAndedCntBtoA = (minPercentBtoA >= 0. ?
		       (Int32)(ceil(minPercentBtoA * pBMS->nptsM)) :
		       -1);

    while (pAB->nTrans > 0)
    {
	pAB->nTrans--;

	x = pAB->pListX[pAB->nTrans];
	y = pAB->pListY[pAB->nTrans];
	
	/* Init for use with match macro below */
	*pX = x - 1;
	
	/* Check if it does match (if the percent(s) here are at least
	 * what our thresholds are).
	 */
	HAUS_CHECK_MATCH_MACRO();
	if (*pX == x)
	    /* Means macro check matched!
	     * Report match location and percent(s).
	     */
	    NORMAL_RETURN;
    }

    /* All done search!!  Return flag value to indicate we are done */
    *pPercentAtoB = -1.;
    NORMAL_RETURN;
}


/* PRIVATE */
/*****************************************************************************
 * Find a(nother) match of A in B search area
 *
 * Description:
 *  This is an iterative routine that will take A and search for a match
 *  through all desired locations of A over B.  The locations to try are
 *  set up and the search gets initialized using
 *  i_hausdorff..SearchSize() and i_hausdorff..SearchInit().
 *  The match of A and B is based on the passed in percents, which may
 *  change at each iterative invocation if desired.
 *  Search simply trying every translation in subrect.
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *
 * Notes:
 *  If minPercentBtoA < 0., the routine won't calculate the *pPercentBtoA
 *  2-Way Hausdorff percent, and will just return -1. instead.
 *  If *pPercentAtoB is returned < 0., it means we are done the search.
 ****************************************************************************/
static Int32 CDECL 
ip_searchEveryTrans (
    IPState    *pIPState,
    searchType *pAB,		  /* Ptr to A and B info */
    modelSet   *pAMS,		  /* Ptr to A model set */
    UInt8     **ppBitmapsA,	  /* Array of stored A bitmaps */
    UInt32     *pAD,		  /* Ptr to upper left corner of the dilated
				   * imagery of A ("AD"), excluding frame */
    modelSet   *pBMS,		  /* Ptr to B model set */
    UInt8     **ppBitmapsB,	  /* Array of stored B bitmaps */
    UInt32     *pBD,		  /* Ptr to upper left corner of the dilated
				   * imagery of B ("BD"), excluding frame */
    Float64	minPercentAtoB,	  /* The A to B and B to A Hausdorff */
    Float64	minPercentBtoA,	  /* percents must be at least these values
				   * to match, [0..1] */
    Int32      *pX,		  /* Ptrs to returned match location (x,y) */
    Int32      *pY,
    Float64    *pPercentAtoB,	  /* Ptrs to returned Hausdorff percents at */
    Float64    *pPercentBtoA)  	  /* this location. */
{
    PROLOGUE;
    Int32      x, y;
    Int32      andedCntAtoB, minAndedCntAtoB;
    Int32      andedCntBtoA, minAndedCntBtoA;

    /* These will be the threshold number of bits that need to be overlapping
     * to constitute a match */
    minAndedCntAtoB = (Int32)(ceil(minPercentAtoB * pAMS->nptsM));
    minAndedCntBtoA = (minPercentBtoA >= 0. ?
		       (Int32)(ceil(minPercentBtoA * pBMS->nptsM)) :
		       -1);

    /* Init for use with match macro below */
    *pX = pAB->firstx - 1;

    /* Go to the next translation from the last one that was reported as
     * a match.
     */
    x = pAB->tx + pAB->skipx;
    y = pAB->ty;

    for ( ; y <= pAB->lasty; y += pAB->skipy)
    {
	for ( ; x <= pAB->lastx; x += pAB->skipx)
	{
	    /* Check if it does match (if the percent(s) here are at least
	     * what our thresholds are).
	     */
	    HAUS_CHECK_MATCH_MACRO();
	    if (*pX == x)
		/* Means macro check matched!
		 * Report match location and percent(s).
		 */
		NORMAL_RETURN;
	}

	/* We want to move to the first x translation in the next row now */
	x = pAB->firstx;
    }


    /* All done search!!  Return flag value to indicate we are done */
    *pPercentAtoB = -1.;
    NORMAL_RETURN;
}


/* PRIVATE */
/*****************************************************************************
 * Find a(nother) match of A in B search area
 *
 * Description:
 *  This is an iterative routine that will take A and search for a match
 *  through all desired locations of A over B.  The locations to try are
 *  set up and the search gets initialized using
 *  i_hausdorff..SearchSize() and i_hausdorff..SearchInit().
 *  The match of A and B is based on the passed in percents, which may
 *  change at each iterative invocation if desired.
 *  Search the subrect by eliminating sparse areas of B that need not be
 *  checked in a method that is slower but uses less memory used than
 *  i_searchSparseFilter().
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *
 * Notes:
 *  If minPercentBtoA < 0., the routine won't calculate the *pPercentBtoA
 *  2-Way Hausdorff percent, and will just return -1. instead.
 *  If *pPercentAtoB is returned < 0., it means we are done the search.
 ****************************************************************************/
static Int32 CDECL 
ip_searchFlyFilter (
    IPState    *pIPState,
    searchType *pAB,		  /* Ptr to A and B info */
    modelSet   *pAMS,		  /* Ptr to A model set */
    UInt8     **ppBitmapsA,	  /* Array of stored A bitmaps */
    UInt32     *pAD,		  /* Ptr to upper left corner of the dilated
				   * imagery of A ("AD"), excluding frame */
    modelSet   *pBMS,		  /* Ptr to B model set */
    UInt8     **ppBitmapsB,	  /* Array of stored B bitmaps */
    UInt32     *pBD,		  /* Ptr to upper left corner of the dilated
				   * imagery of B ("BD"), excluding frame */
    Float64	minPercentAtoB,	  /* The A to B and B to A Hausdorff */
    Float64	minPercentBtoA,	  /* percents must be at least these values
				   * to match, [0..1] */
    Int32      *pX,		  /* Ptrs to returned match location (x,y) */
    Int32      *pY,
    Float64    *pPercentAtoB,	  /* Ptrs to returned Hausdorff percents at */
    Float64    *pPercentBtoA)  	  /* this location. */
{
    PROLOGUE;
    Int32      x, y, skipy;
    Int32      andedCntAtoB, minAndedCntAtoB;
    Int32      andedCntBtoA, minAndedCntBtoA;
    Int32      yModSkip;
    UInt32     leftSideMask = 0, rightSideMask = 0;
    Int32     leftBitNum = 0, rightBitNum = 0;
    UInt8     *pOldRowBD, *pNewRowBD;
    Int32      pitchBD, columnDoneRow;
    Int32     cnt;

    /*  Search Scheme:
     *       __________
     *      |x  _      |
     *      || | |     |
     *      || | | ... |- subrect to search in
     *      ||_| v     |  searches starting at the "x", moving towards the "v"
     *      |__________|
     *
     */

    /* These will be the threshold number of bits that need to be overlapping
     * to constitute a match */
    minAndedCntAtoB = (Int32)(ceil(minPercentAtoB * pAMS->nptsM));
    minAndedCntBtoA = (minPercentBtoA >= 0. ?
		       (Int32)(ceil(minPercentBtoA * pBMS->nptsM)) : -1);

    /* Init for use with match macro below */
    *pX = pAB->firstx - 1;
    
    /* Go to the next translation from the last one that was reported as
     * a match.
     * (Up or down one from the last one as indicated by direction we were
     * moving last).
     */
    x = pAB->tx;
    y = pAB->ty + pAB->flyDirec;

    /* This allows checking at: y, y + skipy, y + 2*skipy, ... */
    skipy = pAB->skipy - 1;
    yModSkip = pAB->firsty & skipy;
    
    pitchBD = pAB->wplBD << 2;
    
    /* move to the beginning of the first image row, skipping binary frame */
    if (pAB->flyDirec == 1)
    {
	/* Means we are moving down in Y */
	pOldRowBD = ((UInt8 *)pBD) + (y-1) * pitchBD;
	pNewRowBD = pOldRowBD + pAMS->heightM * pitchBD;
	columnDoneRow = pAB->lasty + 1;
    }
    else
    {
	/* Means we are moving up in Y */
	pNewRowBD = ((UInt8 *)pBD) + y * pitchBD;
	pOldRowBD = pNewRowBD + pAMS->heightM * pitchBD;
	columnDoneRow = pAB->firsty - 1;
	pitchBD *= -1;
    }


    /* If we are at the end of a column the masks will get reloaded *
     * automatically below, so only do when still in column
     */
    if (y != columnDoneRow)
    {
	/* Calculate the left and right masks.  The static array
	 * leftMask[x] contains x+1 1's at the left side of the
	 * word.  This is exactly what we want to get rid of bits
	 * on the right side of the rightmost word.
	 * The funny calculation says "get the index of the leftmost bit
	 * of the rightmost pixel in the line (x + w - 1).
	 */
	rightBitNum   = x + pAMS->widthM - 1;
	rightSideMask = leftMask[rightBitNum & 0x1F];
	rightBitNum  &= ~0x1F; /* move to word boundary */
	
	/* The static array rightMask[x] contains x+1 1's on the right
	 * side of the word.  We really need all 1's when the leftmost
	 * bit offset is 0, 30 1's on the right when the index is 1 and
	 * so on.  This can be done by simply complementing the index.
	 */
	leftBitNum   = x;
	leftSideMask = rightMask[(~leftBitNum) & 0x1F];
	leftBitNum  &= ~0x1F; /* move to word boundary */
    }


    while (x <= pAB->lastx)
    {
	if (y != columnDoneRow)
	{
	    /* Window has moved UP or DOWN one in Y so recalculate the window
	     * sum.
	     */

	    /* Subtract the number of bits on in the old row that
	     * disappeared when the subrect window moved in Y.
	     */
	    ip_countBitsBinary (pIPState,  pOldRowBD,
				leftSideMask, rightSideMask,
				leftBitNum, rightBitNum,
				1, ABS(pitchBD), &cnt );
	    pAB->winSumBD -= cnt;
	    
	    /* Add the number of bits on in the new row that
	     * has moved in when the subrect window moved in Y.
	     */
	    ip_countBitsBinary (pIPState,  pNewRowBD,
				leftSideMask, rightSideMask,
				leftBitNum, rightBitNum,
				1, ABS(pitchBD), &cnt );
	    pAB->winSumBD += cnt;
	}
	else
	{
	    /* We are now done with the searching the current column.
	     * Now move the subrect window RIGHT (one location) from where
	     * it was at the last comparison and reverse the direction of the
	     * column search.
	     */
	    x += pAB->skipx;

	    if (x > pAB->lastx)
	    {
		/* This happens very rarely, but it can happen so we have
		 * to check for it.  All done search!!
		 * Return flag value to indicate we are done
		 */
		*pPercentAtoB = -1.;
		NORMAL_RETURN;
	    }

	    pAB->flyDirec *= -1;
	    pitchBD *= -1;
	    
	    y += pAB->flyDirec;

	    /* Swap the row pointers */
	    {
		UInt8  *tmp;
		tmp = pOldRowBD;
		pOldRowBD = pNewRowBD;
		pNewRowBD = tmp;
	    }
	    
	    columnDoneRow = (pAB->flyDirec == 1
			     ? pAB->lasty + 1 : pAB->firsty - 1);

	    /* Calculate the left and right masks.  The static array
	     * leftMask[x] contains x+1 1's at the left side of the
	     * word.  This is exactly what we want to get rid of bits
	     * on the right side of the rightmost word.
	     * The funny calculation says "get the index of the leftmost bit
	     * of the rightmost pixel in the line (x + w - 1).
	     */
	    rightBitNum   = x + pAMS->widthM - 1;
	    rightSideMask = leftMask[rightBitNum & 0x1F];
	    rightBitNum  &= ~0x1F; /* move to word boundary */
	
	    /* The static array rightMask[x] contains x+1 1's on the right
	     * side of the word.  We really need all 1's when the leftmost
	     * bit offset is 0, 30 1's on the right when the index is 1 and
	     * so on.  This can be done by simply complementing the index.
	     */
	    leftBitNum   = x;
	    leftSideMask = rightMask[(~leftBitNum) & 0x1F];
	    leftBitNum  &= ~0x1F; /* move to word boundary */
	
	    /* This new window is also one column over.  Count all the points
	     * in this window, discarding the number of points information for
	     * the last window
	     */
	    ip_countBitsBinary (pIPState,  (pAB->flyDirec == 1 ?
				pOldRowBD + pitchBD : pNewRowBD),
				leftSideMask, rightSideMask,
				leftBitNum, rightBitNum,
				pAMS->heightM, ABS(pitchBD),
				&(pAB->winSumBD) );
	}


#ifdef hDEBUG
	/* This simply counts the number of points on in the subrect
	 * window the slow, stupid, yet simpler way, and then ensures
	 * that my people are working easier and just as well as your
	 * people.
	 */
	i_countBits (pIPState,  (UInt8 *)pBD, x, y, pAMS->widthM,
		     pAMS->heightM, 1, ABS(pitchBD), &cnt );
	hCHECK(cnt != pAB->winSumBD, "--Count checking error!");
	if (cnt != pAB->winSumBD)
	    fprintf(stderr, "[%3ld,%3ld] %3lu v. %3lu\n",
		    x, y, cnt, pAB->winSumBD);
#endif
	    
	    
	if (pAB->winSumBD >= minAndedCntAtoB &&
	    ((y & skipy) == yModSkip))
	{
	    /* This location in BD is dense enough to possibly match!
	     * and at a desirable y offset */

	    /* Check if it does match (if the percent(s) here are at least
	     * what our thresholds are).
	     */
	    HAUS_CHECK_MATCH_MACRO();
	    if (*pX == x)
		/* Means macro check matched!
		 * Report match location and percent(s).
		 */
		NORMAL_RETURN;
	}

	/* Move UP or DOWN one row in Y */
	y += pAB->flyDirec;
	pOldRowBD += pitchBD;
	pNewRowBD += pitchBD;
    }

    /* All done search!!  Return flag value to indicate we are done */
    *pPercentAtoB = -1.;
    NORMAL_RETURN;
}


/*****************************************************************************
 * Find a(nother) match of A in B search area
 *
 * Description:
 *  This is an iterative routine that will take A and search for a match
 *  through all desired locations of A over B.  The locations to try are
 *  set up and the search gets initialized using
 *  i_hausdorff..SearchSize() and i_hausdorff..SearchInit().
 *  The match of A and B is based on the passed in percents, which may
 *  change at each iterative invocation if desired.
 *  Search the subrect by eliminating sparse areas of B that need not be
 *  checked, processing the area in strips.
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *
 * Notes:
 *  If minPercentBtoA < 0., the routine won't calculate the *pPercentBtoA
 *  2-Way Hausdorff percent, and will just return -1. instead.
 *  If *pPercentAtoB is returned < 0., it means we are done the search.
 ****************************************************************************/
static Int32 CDECL 
ip_searchSparseFilter (
    IPState    *pIPState,
    searchType *pAB,		  /* Ptr to A and B info */
    modelSet   *pAMS,		  /* Ptr to A model set */
    UInt8     **ppBitmapsA,	  /* Array of stored A bitmaps */
    UInt32     *pAD,		  /* Ptr to upper left corner of the dilated
				   * imagery of A ("AD"), excluding frame */
    modelSet   *pBMS,		  /* Ptr to B model set */
    UInt8     **ppBitmapsB,	  /* Array of stored B bitmaps */
    UInt32     *pBD,		  /* Ptr to upper left corner of the dilated
				   * imagery of B ("BD"), excluding frame */
    Float64	minPercentAtoB,	  /* The A to B and B to A Hausdorff */
    Float64	minPercentBtoA,	  /* percents must be at least these values
				   * to match, [0..1] */
    Int32      *pX,		  /* Ptrs to returned match location (x,y) */
    Int32      *pY,
    Float64    *pPercentAtoB,	  /* Ptrs to returned Hausdorff percents at */
    Float64    *pPercentBtoA)  	  /* this location. */
{
    PROLOGUE;
    Int32      x, y, skipy;
    Int32      andedCntAtoB, minAndedCntAtoB;
    Int32      andedCntBtoA, minAndedCntBtoA;
    UInt16    *pAddRow, *pSubRow;
    UInt8     *pImAddRow;
    Int32      tmp, yModSkip;
    
    /* These will be the threshold number of bits that need to be overlapping
     * to constitute a match */
    minAndedCntAtoB = (Int32)(ceil(minPercentAtoB * pAMS->nptsM));
    minAndedCntBtoA = (minPercentBtoA >= 0. ?
		       (Int32)(ceil(minPercentBtoA * pBMS->nptsM)) : -1);

    /* Init for use with match macro below */
    *pX = pAB->firstx - 1;

    /* Go to the next translation from the last one that was reported as
     * a match.
     */
    x = pAB->tx + pAB->skipx;
    y = pAB->ty;

    /* This allows checking at: y, y + skipy, y + 2*skipy, ... */
    skipy = pAB->skipy - 1;
    yModSkip = pAB->firsty & skipy;

    /* This is a pointer to the row in the image that will be added next
     * to the window summing info when this row moves into the window
     * (when we move down a row!)
     */
    pImAddRow = (UInt8 *)pBD +
	(y + pAB->hSums - 1) * (pAB->wplBD << 2);

    for ( ; y <= pAB->lasty; y++)
    {
	for ( ; x <= pAB->lastx; x += pAB->skipx)
	{

#ifdef hDEBUG
	    if (1)
	    {
		Int32  temp;
		/* This simply counts the number of points on in the subrect
		 * window the slow, stupid, yet simpler way, and then ensures
		 * that my people are working easier and just as well as your
		 * people.
		 */
		i_countBits (pIPState,  (UInt8 *)pBD, x, y,
			     /* Crop the width, height of the window based
			      * on the nonnegative (x,y) position! */
			     MIN(pAMS->widthM,  pAB->widthBD  - x),
			     MIN(pAMS->heightM, pAB->heightBD - y),
			     1, (pAB->wplBD << 2), &temp );
		hCHECK(pAB->pSumsXY[x - pAB->firstx] != temp,
		       "--Count checking error!");
		if (pAB->pSumsXY[x - pAB->firstx] != temp)
		    fprintf(stderr, "[%3ld,%3ld] %3lu v. %3lu\n",
			    x, y, pAB->pSumsXY[x - pAB->firstx],
			    temp);
	    }
#endif /*hDEBUG*/
	    
	    if (((y & skipy) == yModSkip) &&
		pAB->pSumsXY[x - pAB->firstx] >= (UInt32)minAndedCntAtoB)
	    {
		/* This location in BD is dense enough to possibly match!
		 * and at a desirable y offset */

		/* Check if it does match (if the percent(s) here are at least
		 * what our thresholds are).
		 */
		HAUS_CHECK_MATCH_MACRO();
		if (*pX == x)
		    /* Means macro check matched!
		     * Report match location and percent(s).
		     */
		    NORMAL_RETURN;
	    }
	}
	

	/* Now update the sumsX and sumsXY rows */
	
	/* Where to add new row, and add to sumsXY */
	pAddRow = pAB->ppSumsX[pAB->addRow];
	
	/* Overwrite the row that got subtracted before with the new row that
	 * will get added */
	i_makeRowXSums (pIPState, pImAddRow,
			pAB->firstx, pAB->lastx, pAMS->widthM,
		        pAddRow);
	pImAddRow += (pAB->wplBD << 2);
	
	pAB->addRow = (pAB->addRow == pAB->hSums-1
		       ? 0 : pAB->addRow+1);
	
	/* Old row to subtract from sumsXY */
	pSubRow = pAB->ppSumsX[pAB->addRow];
	
	/* Subtract and add the appropriate rows from the sumsXY,
	 * to update it for the next row */
	for (tmp = 0; tmp < pAB->lastx - pAB->firstx + 1; tmp++)
	    pAB->pSumsXY[tmp] += pAddRow[tmp] - pSubRow[tmp];
	
	/* We want to move to the first x translation in the next row now */
	x = pAB->firstx;
    }
    
    
    /* All done search!!  Return flag value to indicate we are done */
    *pPercentAtoB = -1.;
    NORMAL_RETURN;
}


/*

$Log: i_hcsrch.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.22  1998/01/13 23:48:29  ddavies
i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.21  1997/07/11 01:48:55  ddavies
i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.20  1996/04/17 23:25:20  ejaquith
i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v

Fixed auto-doc stuff that is lingering there for demo of IPShared routines.

Revision 1.19  1995/12/13 02:44:36  ejaquith
i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.18  1995/10/30  18:12:12  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Updated HTML secret flags.
 *
 * Revision 1.17  1995/10/18  03:57:32  ddavies
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.16  1995/09/14  18:59:54  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * CommonCheck: Converted all routines that were private in files to be of form
 * ip_*(), which they weren't.  No functionality/noticeable changes.
 *
 * Revision 1.15  1995/07/24  22:04:23  ejaquith
 * ejaquith on Mon Jul 24 15:02:34 PDT 1995
 *
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Bug Fix.
 * In i_searchFlyFilter(), at the end of the searching strategy, we get
 * to a translation that is one to the right of the rightmost subrect
 * edge.  However, instead of aborting immediately, the routine was
 * actually checking this translation -- no no!  Added check to abort
 * once we get the rightmost edge.
 *
 * Revision 1.14  1995/07/14  20:29:59  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Jumped the gun and added some new function header comments to test
 * the pre-natal auto-doc system.
 *
 * Revision 1.13  1995/04/22  21:14:40  ejaquith
 * ejaquith on Sat Apr 22 14:14:38 PDT 1995
 *
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * (Checkin log aborted me without my comments, here they are)
 * Removed i_hausdorff1WayMatches(), i_hausdorff2WayMatches().
 * Added new interface to searching subrects, routines
 * i_hausdorffRectSearchSize() and i_hausdorffRectSearchInit().
 * Added complete new way of searching a list of (arbitrary) translations,
 * routines i_hausdorffListSearchSize() and i_hausdorffListSearchSize().
 * Both pairs of routines get invoked once to initialize a structure for
 * searching for matches, and then the client iteratively calls
 * new routine i_hausdorffSearch().  This last routine will return once
 * it finds every single match and save out the current state to continue
 * the search upon another invocation.
 * There are now four methods of searching, three for subrects plus one
 * for a list of translations.
 * Each search method has its own (new -- modified from any previous
 * routine that did something similar in old module) routine.  Each routine
 * now, in turn calls a nifty little comparison MACRO (instead of a routine)
 * that allows faster comparison and standardized use.
 * i_searchSparseFilter(), the modifed i_hausdorffSparseFilter() as an
 * example has been updated to use the same searching structure for whether
 * (oops thats i_searchFlyFilter() and i_hausdorffFlyFilter() above!)
 * or not the searcher is moving up or down in the subrect for code
 * consolidation.  All routines had to add ability to save state every
 * time a match was found.  The initializing routines had to be able to start
 * up and allocate and fill needed info for each of these routines.
 * For example, when sparse filtering, the top N rows (where N is the height
 * of the A bitmap) of the search subrect are window summed and saved
 * for the top row of searching usage.  As the search proceeds, it updates
 * these window summed rows whenever a new row is reached.
 * All in all, the changes are many and too dense/boring to describe in
 * detail.  However, the module is about the same speed with all these
 * new changes and a bit more flexible in the search.  Clients get much more
 * control and can write interface code much easier since they need not
 * use callbacks by using ipshared interface.
 *
 * Revision 1.12  1995/04/22  20:48:01  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Please see general (large!) module comments for same date in i_hc32ms.c
 *
 * Revision 1.11  1995/04/03  20:44:02  ejaquith
 * ejaquith on Mon Apr 3 13:39:31 PDT 1995
 *
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Bug Fix.
 * In i_makeRowXSums(), the routine that is used for parsing out potential
 * match translations because they are too "sparse", had an instance where
 * a word could be right shifted by 32.  On suns, this doesn't do the logical
 * thing of producing a 0x0 word and instead does a no op, which through of
 * (threw off) bit counting for filter widths of 32.
 *
 * Revision 1.10  1995/03/29  01:59:20  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.9  1995/03/21  19:07:25  ejaquith
 * ejaquith on Tue Mar 21 11:02:12 PST 1995
 *
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Consolidated some code and made remaining code more flexible.
 * (And faster!)
 * Removed "best" searching techniques, routines, etc.
 * This type of searching is now done easily with the client's callback
 * routine (depending on the client's return value).
 * Now, both 1Way matching and 2Way matching have the same parametered
 * callback routine (to allow greater flexibility even though an extra
 * dummy Hausdorff percent value will now be returned to the callback
 * when only doing 1Way searching).
 * Removed routines best1WayMatch(), matchCheckBest1Way(),
 * i_hausdorff1WayBestMatch() since the normal 1Way matching can now handle
 * this.
 * Invocations of matchCheck1Way() and matchCheck2Way() now handle invoking
 * the client's callback routine and depending on the callback return value,
 * whether or not any search information should be updated.
 * This removes a whole level of routines for "best" searching.
 *
 * Revision 1.8  1995/03/15  01:03:56  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Updated (non multi-resolutional) searching to work with new feature
 * "row ordered models".
 * Fixed indentation style (in here and entire Hausdorff module)
 * to conform with ipcore's ideal style (basically moved all beginning
 * braces "{" to newlines).
 * Attempted to speed up i_makeRowXSums() with no luck and #ifdef'd out
 * (with ATTEMPT_AT_CLEVERNESS_WAS_SLOWER)
 *
 * Revision 1.7  1995/03/01  21:58:56  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Made new searching strategy, which is the new default method when
 * not calling multi-resolutional methods.
 * The old method is now callable through the use of the new Hausdorff
 * bit flag "FLY_FILTER_SEARCH".
 * The new search strategy sparse filters like the old one.
 * However, through allocating some minimal memory, it now processes the
 * search area in strips, which the old technique did not.
 * The new technique is about 25-33% faster.
 * The new technique sums the number of BD points on in a subrect the
 * size of the A image for each pixel in a BD subrect row.  It does so
 * by making calls to the new i_makeRowXSums() routine, which counts the
 * number of on points in a "window" of some width, then moves the window
 * over one, subtracts the old pixel and adds the new, and assigns the
 * next number of on points.
 * The search technique then has a certain number of this row sums and
 * adds them for the final sum the size of the A window.  As it moves to
 * the next row, it subtracts the old row sum and adds the new row sum.
 * Since this is done in all strips and buffers, added routine
 * i_countFilter() which is used primarily for sanity-check debugging
 * and does the above scheme over the entire BD subrect at once, not in
 * strips.
 * Added checks in i_hausdorffFlyFilter() at the end of a column of
 * searching to ensure we do not move off an image if it's frameless.
 * Sped up i_hausdorffEvery(), the technique used when the Hausdorff
 * bit flag SEARCH_EVERY_TRANS is set.
 *
 * Revision 1.6  1995/02/14  01:43:10  ejaquith
 * ejaquith on Mon Feb 13 17:40:50 PST 1995
 *
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Added Hausdorff flag SEARCH_EVERY_TRANS to library.
 * The 3 calling interfaces in this file now check for it and will
 * call the new routine i_hausdorffEvery() if the flag is set.
 * This is for instances where filtering overhead outweighs the
 * fewer calls to the comparator.
 *
 * Revision 1.5  1995/01/19  22:41:55  ejaquith
 * ejaquith on Thu Jan 19 14:40:32 PST 1995
 *
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Fixed bug with new Hausdorff Best matching routines.  It was invoking
 * the client's passed in verify routine at the wrong time and corrupted a
 * few inner state variables as a result.
 *
 * Revision 1.4  1995/01/17  19:10:44  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Changed interface to i_hausdorff1WayBestMatch()
 * to take a ptr to a "verify routine" when finding the best
 * match instead of ptrs to return the best
 * Hausdorff location and percent.  This shrinks the arg list.  Additionally,
 * it allows the client to decide if each found best match will be acceptable
 * (as the routine progresses through the rect) as the final best match.
 * (I'm using it to rule out certain sub-areas of the sub-rect where I don't
 * want the final match to be contained in).
 *
 * Revision 1.3  1995/01/04  18:47:45  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Abandoned use of "unit" (variable sized word for storage of models) for
 * use of 32 bit words.  Changed all public routines to use image byte
 * pointers and bytes per line (from image word pointers and words per line).
 * Completely rewrote way module works for searching.  Now, the specific
 * search style call (1WayBest, 1WayMatches, 2WayMatches) will set up a
 * Hausdorff calling structure (complete with appropriate "did it match?"
 * routine pointer) and then call the lower level i_hausdorff().
 * Added best1WayMatch() to update when searching for best match.
 * Added matchCheck1Way(), matchCheckBest1Way(), and matchCheck2Way() as
 * these specific "did it match?" routines.
 * Thus, approximately halved the file size because the subrect (sparse
 * filtering) searching overall method had been replicated 3 full times (one
 * for each search style) and now is just one.
 *
 * Revision 1.2  1994/12/07  23:00:35  ejaquith
 * i_hcsrch.c -> /product/ipcore/ipshared/src/RCS/i_hcsrch.c,v
 *
 * Ref'd unused parameters to keep compiler happy.
 * Fixed type check bug in assigning Hausdorff routine pointer.
 *
 * Revision 1.1  1994/12/06  23:31:08  ejaquith
 * Initial revision
 *
 */
