/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_heap.c: Memory allocation/deallocation procedures used to manage
 *		heaps of structures.  All structures in a given heap
 *		are of the same type.
 *
 *	Procedures for allocating and freeing blocks of structures.
 *		Int32	i_allocateMemoryBlock()
 *		Int32	i_freeAllMemoryBlocks()
 *
 *	Procedures for allocating and freeing structures within memory blocks.
 *		Int32	i_allocateStruct()
 *		Int32	i_freeStruct()
 *
 *	Procedures for doing local allocation and freeing of memory.
 *	These are used when the client isn't picky about the memory
 *	allocator.
 *		void *	i_cmpressAlloc
 *		void *	i_cmpressFree
 *
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "memory.pub"
#include "ptrfixup.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcc.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_heap.c,v $; $Revision: 1.2 $; $Date: 2001/05/08 06:43:36 $");

/* This file is used to manage large blocks of structures.  These procedures
 * allocate memory blocks, each containing a fixed number of structures
 * of a fixed length,  The client can then allocate and free individual
 * structures from these blocks.  When finished, the client can free
 * all the large blocks in one step.  These routines are intended for
 * clients that repeatedly allocate and free many structures of the same
 * type.  The symbol based compression is one such client.
 */

/* We could just use malloc and free to allocate and deallocate all the
 * different data structures.  This would likely lead to awful fragmentation.
 * It could be pretty slow too as we have to maintain the chains of
 * allocated and free memory.
 *
 * Instead, we'll allocate one or more large blocks for each type of data
 * structure.  We allocate and deallocate from those.  When we're done,
 * we simply free the large blocks.
 *
 * Instead of writing a slightly different, fairly large, procedure for
 * allocating and freeing each type of memory block, we'll just have one
 * set that uses variable size structures.  There will be space for a
 * structure at the beginning of each of these blocks.  That structure
 * contains, among other things, a pointer that points to the next block.
 * The structure also contains the size of the structures
 * comprising the block, the number of structures and the procedures
 * used to allocate and free memory blocks.  This information
 * is used to allocate new blocks.  The first block also contains a pointer
 * to the top of the heap.  With all this stuff stored in each big
 * memory block, we can just pass a pointer to the current block when
 * we want to allocate a new structure.  Otherwise, we'd have to pass all
 * the info required to allocate a new memory block just in case it was needed.
 *
 * When the heap is initialized, it consists of a single memory block
 * containing a bunch of structures.  All the structures are linked together
 * in the list of free blocks.  The free list is maintained by smashing
 * a pointer to the next free block over the first four bytes of
 * each structure.
 *
 * Each time a block is needed, it is taken from the head of the free list
 * and the head pointer is moved to the next block in the free list.
 * Each time a block is returned, it is put onto the head of the free list.
 *
 * If allocating a structure exhausts the free list, we immediately allocate
 * another block, initialize it with free structures and make the free
 * list pointer point to that block.  Since structures may be allocated and
 * freed in any order, the free list may snake arbitrarily between all
 * the memory blocks in the heap.
 *
 * GARBAGE COLLECTION:  That was all fine.  However, sometimes heaps
 * persist after control returns to a client that does garbage collection,
 * like the DAE.  The DAE's garbage collection know nothing about C structures,
 * or our heaps.  However, since we politely used the DAE's allocator to
 * allocate each memory block, the blocks may be moved around.  We trust
 * the DAE folks have figured out how to guarantee that the heaps don't
 * get garbage collected (freed).  Anyway, the heap memory blocks always
 * contain pointers.  Even if the structures inside the heaps have no
 * pointers, there's always the list of free structures.
 *
 * What to do?  Well, the DAE folks can call an ipshared routine every time
 * a block is moved.  They can call it immediately after each block is moved
 * (before any other block is moved).  They can know the old block address
 * and the new block address.  We have a pointer fixup scheme that depends
 * on this fact.
 *
 * At the top of the TCBlockHeader is a PointerFixup structure, see
 * ptrfixup.pub in ipshared/include.  This structure contains
 * pointers to other memory blocks that might contain pointers to this one
 * and other memory blocks to which this block might point.  It also contains
 * a pointer to a routine that can be used to fix up pointers inside this
 * memory block.  Finally, it contains the size of this block.
 *
 * To make this work, i_allocateMemoryBlock must take the arguments
 * used to initialize the pointer fixup block.  It's argument list
 * has been expanded to do so.
 *
 * We also have to be a little less cavalier about smashing the free
 * list structure pointers.  In the old design, we just pretended that
 * each structure was a TCBlockHeader because the first thing in the
 * TCBlockHeader was a pointer. Ouch.  Now we'll define a TCFreeStructure
 * type which contains only a pointer.
 */


/************************************************************************/
/************************ Generic Block routines ************************/
/************************************************************************/

/* PUBLIC */
/**************************************************************************
 *   i_allocateMemoryBlock()
 *
 *	This procedure allocates and initializes a block containing
 *	space for numStructs copies of a struct, each containing
 *	sizeOfStruct bytes (at least 4).
 *
 *	After the space is allocated, the first four bytes of each struct
 *	are set to point to the next struct so the simple heap allocator
 *	will work.
 *
 *	This routine is used for memory blocks containing all different
 *	kinds of structs.  We specify the kind of struct with the size
 *	of the struct and the number of them in the block.
 *
 *	The memory blocks allocated by this routine become members of two
 *	rings.  One is the set of all memory blocks which might contain
 *	pointers to anything in the block and all memory blocks to which
 *	things in the block might point.  This ring is used for pointer
 *	fixup after garbage collection.  A pointer to some block in
 *	the garbage collection ring must be supplied to this routine.
 *
 *	The second type of ring connects all memory blocks that are
 *	part of one heap.  This routine does not attempt to add
 *	this block to a heap, that's the responsibility of the caller.
 *	We do this so the routine can be used to either initialize
 *	a heap or add a block to a heap.
 *
 *	Input:
 *		pp1stMemoryBlock:	Pointer to pointer to first
 *					Memory block for this type of struct.
 *		structSize:		The size of the structure to be
 *					allocated in bytes.
 *		numStructs:		The number of structs assigned to
 *					each memory block.
 *		fixupPointers:		If equal to cTrue, the memory
 *					blocks allocated will persist after
 *					the calling ipshared routine
 *					returns to the client AND they will
 *					contain pointers that should be
 *					corrected after the DAE moves
 *					memory blocks around.  Otherwise,
 *					set this to cFalse.
 *		pAttachPt:		If fixupPointer equals cTrue, this
 *					word contains a pointer to some
 *					PointerFixup struct in the ring
 *					of blocks we inspect when some 
 *					block moves as a result of garbage 
 *					collection.  This ring includes 
 *					all blocks that contain pointers 
 *					to this block as well as all blocks
 *					to which this block might point.
 *		fixupProc:		If fixupPointer equals cTrue, this
 *					word contains a pointer to the
 *					procedure used to fix ptrs
 *					inside this block after garbage
 *					collection has moved some block.
 *		allocator:		Ptr to proc we will use to
 *					allocate the large memory blocks.
 *					In addition to the size of the
 *					block in bytes, this allocator takes
 *					a second parameter which is set
 *					equal to fixupPointers when the
 *					allocator is called.
 *		freer:			ptr to proc used to free the
 *					large memory blocks.  This
 *					routine takes the length of the
 *					block to be freed to maintain
 *					compatibility with Ghostscript.
 *
 *	Note:	This routine initializes a heap.  The heap starts with
 *		one memory block, but may grow to include several.
 *		If the heap is persistent (lasts after control returns
 *		to an environment using garbage collection, like the DAE),
 *		then at least the free pointers which are written over
 *		the first four bytes of each structure contained in a
 *		heap memory block will have to be updated after the
 *		garbage collector moves ANY of the heap memory blocks.
 *		This has two consequences.  First, ALL persistent heaps
 *		must be garbage collected.  The client must set the
 *		fixupPointers, pAttachPoint and fixupProc pointers
 *		appropriately.  The client must also supply a fixupProc, 
 *		possibly using an existing one.
 *
 *		Second, there is nothing built into the heap manager that
 *		can be used to easily tell whether a memory block is free
 *		or not.  Yes, we could follow the free list down, but how
 *		would we enumerate all remaining the structures, the ones
 *		in use?  Since the free list pointers are written over the
 *		first four bytes of EACH structure, the client defining the
 *		structure (that's you, right?) has only a few choices.
 *		    1.	Don't use the first four bytes.  This leaves the
 *			old free list pointer there, but that's ok.
 *		    2.	Put a pointer into the first four bytes of the
 *			structure.  That way, there's always a pointer there.
 *			The fixup routine blindly correct it if it points
 *			to the block that moved.
 *		    3.	Put something that couldn't possible look like it
 *			pointed at the moved memory block in the first four
 *			bytes.  This is kinda tricky, watch out.
 *
 *	Returns:
 *		ia_successful		block was allocated
 *		ia_nomem		ran out of memory
 *
 **************************************************************************/
Int32  CDECL 
i_allocateMemoryBlock (
    IPState    *pIPState,
    TCBlockHeader **pp1stMemoryBlock,
    UInt32	structSize,
    UInt32	numOfStructs,
    UInt32	fixupPointers,
    PointerFixup *pAttachPt,
    Int32 (CDECL *fixupProc) (
		    IPState    *pIPState,
		    void       *pThisBlock,
		    void       *pOldBlock,
		    void       *pNewBlock,
		    UInt32	blkLength),
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	bufSizeInBytes,
		    UInt32	fixupPointers),
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *pBlock,
		    UInt32	length))
{
    PROLOGUE;
    UInt8	*pCurrent;	/* points to the current struct */
    UInt8	*pLast;		/* pointer to last struct in the block. */
    UInt32	 blockSize;	/* size of this block in bytes */


	/* allocate the block */
    blockSize = sizeof(TCBlockHeader) + (numOfStructs * structSize);
    pCurrent = (UInt8 *) allocator(pIPState, blockSize, fixupPointers);

	/* if we ran out of memory, say so */
    if (pCurrent == NULL)
	return ia_nomem;

	/* Return a pointer to the new block. */
    *pp1stMemoryBlock = (TCBlockHeader *) pCurrent;

	/* Initialize the other elements of the struct at the
	 * head of the block.  If we're not supposed to fixup
	 * the pointers, set the blocksize to 0.  That's the
	 * flag i_allocateStruct uses to decide how to set
	 * fixupPointers when it calls i_allocateMemoryBlock
	 * when the heap runs dry.
	 */
    __error__ = ip_initPointerFixup (pIPState,
				     &(((TCBlockHeader *)pCurrent)->fixup),
				     (fixupPointers ? blockSize : 0),
				     pAttachPt,
				     fixupProc);
    if (__error__ != ia_successful)
    {
	(*freer)(pIPState, pCurrent, blockSize);
	NORMAL_RETURN;
    }

    ((TCBlockHeader *)pCurrent)->pNext = NULL;
    ((TCBlockHeader *)pCurrent)->structSize = structSize;
    ((TCBlockHeader *)pCurrent)->numOfStructs = numOfStructs;
    ((TCBlockHeader *)pCurrent)->ipcAllocator = allocator;
    ((TCBlockHeader *)pCurrent)->freer = freer;
    ((TCBlockHeader *)pCurrent)->pTop = (TCBlockHeader *) (pCurrent +
							sizeof(TCBlockHeader));

	/* Clear all the remaining bytes of the block.  There are
	 * later times when we'll want to look at all blocks of
	 * a given type without worrying about whether they've
	 * been initialized.
	 */
    pCurrent += sizeof(TCBlockHeader);			/* skip header */
    pLast = pCurrent + ((numOfStructs-1)*structSize);	/* pt to last run */
    MEMSET(pCurrent, 0, numOfStructs * structSize);

	/* Initialize each block in the run to point to the next one.
	 * This is the initial list of free blocks.
	 * We have already skipped the TCBlockHeader, so are pointing
	 * at the first word of the first structure.  End before the last
	 * structure.  We'll set its free list pointer to NULL.
	 */
    for (; pCurrent != pLast; pCurrent += structSize)
    {
	*((void **)pCurrent) = (void *) (pCurrent + structSize);
    }

	/* Initialize the next pointer in the last element to NULL */
    *((void **)pCurrent) = NULL;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_freeAllMemoryBlocks()
 *
 *	This procedure frees all large memory blocks in the chain pointed
 *	to by *pp1stBlock.
 *
 *	Input:
 *		pp1stBlock:	Pointer to pointer to first block to be
 *				freed.  The pointed to pointer is set to
 *				NULL.
 *
 *	Returns:
 *		ia_successful		blocks were freed
 *
 **************************************************************************/
Int32  CDECL 
i_freeAllMemoryBlocks (
    IPState    *pIPState,
    TCBlockHeader **pp1stBlock)
{
PROLOGUE;
 TCBlockHeader	 *pCurrent, *pNext;	/* points to the current, next blocks */
void	  (CDECL *freer) (
	IPState	   *pIPState,
	void	   *pBlock,
	UInt32	    length);
UInt32		  blockSize;		/* Ghostscript memory freeing
					 * routines want the length of the
					 * block being freed
					 */

	/* walk the chain of blocks, freeing each one */
    pCurrent = *pp1stBlock;	/* point to first block */
    if (pCurrent == NULL)
	NORMAL_RETURN;	/* nothing to free */

    freer = pCurrent->freer;	/* get ptr to freeing routine */
    blockSize = (pCurrent->numOfStructs * pCurrent->structSize) +
		sizeof(TCBlockHeader);

    while (pCurrent != NULL)
    {
	ip_detachBlockFromPtrFixupRing (pIPState, &(pCurrent->fixup));

	pNext = pCurrent->pNext;		/* point to next one */
	freer(pIPState, pCurrent, blockSize);
	pCurrent = pNext;
    }

	/* set the original ptr to NULL too. */
    *pp1stBlock = NULL;

    NORMAL_RETURN;
}



/************************************************************************/
/************************ Generic Struct Routines ***********************/
/************************************************************************/


/* PUBLIC */
/**************************************************************************
 *   i_allocateStruct()
 *
 *	This procedure allocates a single struct from a heap.
 *	The heap is composed of a chain of large memory blocks, each
 *	containing a collection of structs.  The size of each struct
 *	and the number of structs and a pointer to the top of the
 *	heap (a struct) are also put at the beginning of each
 *	memory block.
 *	For the sake of speed, we assume that pTop (in the memory block's
 *	header) always points to a valid structure.  If that's structure's
 *	pNext pointer is NULL, we allocate a new memory block, insert it
 *	into the chain of memory blocks and set our pTop pointer to point
 *	to the first struct in that block.
 *
 *	Input:
 *		pMemoryBlock:	Pointer to memory block containing
 *				pointer to top of heap.
 *		ppNewStruct:	Pointer to a pointer to the
 *				struct we've just allocated.
 *
 *	Returns:
 *		ia_successful		TCScanLineRun was allocated
 *		ia_nomem		ran out of memory
 *
 **************************************************************************/
Int32  CDECL 
i_allocateStruct (
    IPState    *pIPState,
    TCBlockHeader *pMemoryBlock,
    void      **ppNewStruct)
{
PROLOGUE;
 void		**pTop;		/* points to first free struct, which
				 * contains a pointer to the second one */
TCBlockHeader	 *pNewBlock;	/* points to new block */


	/* The first thing in each structure is a pointer to the next
	 * struct on the free list.  If the struct on top of the heap
	 * isn't the last free struct in the heap, allocate it and
	 * point to the next free struct.
	 */
    pTop = (void **) pMemoryBlock->pTop;
    *ppNewStruct = pTop;		/* return ptr to this struct */
    if (*pTop != NULL)
    {
	    /* point to the next one */
	pMemoryBlock->pTop =  *pTop;
	NORMAL_RETURN;
    }
    else	/* we need to allocated a new block */
    {
	__error__ = i_allocateMemoryBlock (pIPState, &pNewBlock,
					 pMemoryBlock->structSize,
					 pMemoryBlock->numOfStructs,
					 (pMemoryBlock->fixup.blockSize != 0),
					 &(pMemoryBlock->fixup),
					 pMemoryBlock->fixup.fixupProc,
					 pMemoryBlock->ipcAllocator,
					 pMemoryBlock->freer);

	if (__error__ != ia_successful)
	{
	    *ppNewStruct = NULL;
	    NORMAL_RETURN;
	}

	    /* We got the new block.  Note that at this point, pNewBlock
	     * points to the new memory block, not a structure within
	     * either memory block.
	     * Insert the new block into the chain of memory
	     * blocks.  Note that since we aren't passed a pointer to 
	     * a pointer to memory blocks, we can't change the caller's
	     * idea of where the first memory block is.  Therefore, we
	     * have to insert the new block right after the current one.
	     */
	pNewBlock->pNext = pMemoryBlock->pNext;
	pMemoryBlock->pNext = pNewBlock;

	    /* Make the top of the heap point to the first block in the
	     * new block.  That's where it will point after we return the
	     * block that used to be the last one in the heap.
	     */
	pMemoryBlock->pTop = pNewBlock->pTop;
    }

	/* that worked, so return.  Putting a real return here keeps
	 * both gcc and acc happy.
	 */
    NORMAL_RETURN;

}


/* PUBLIC */
/**************************************************************************
 *   i_freeStruct()
 *
 *	This procedure frees a single TCScanLineRun to a heap.
 *
 *	Input:
 *		pMemoryBlock:	Pointer to memory block containing
 *				pointer to top of heap.
 *		ppFreedStruct:	Pointer to Ptr to freed structure.
 *
 *	Action:
 *	The structure is returned to the large block of similar structures.
 *	*ppFreedStruct is set to NULL.
 *
 *	Returns:
 *		ia_successful		TCScanLineRun was freed
 *
 **************************************************************************/
Int32  CDECL 
i_freeStruct (
    IPState    *pIPState,
    TCBlockHeader *pMemoryBlock,
    void      **ppFreedStruct)
{
PROLOGUE;
 void	**pFreedStruct;		/* pointer to freed struct, which will
				 * contain a pointer the the next struct
				 * on the free list. */

	/* It's easier to use this */
    pFreedStruct = (void **) *ppFreedStruct;

	/* Smash the first 4 bytes of the struct to be a pointer to the
	 * next struct.
	 */
    *pFreedStruct = pMemoryBlock->pTop;
	/* make the freed thing the top thing on the heap */
    pMemoryBlock->pTop = (void *) pFreedStruct;
	/* set the incoming pointer to NULL and return */
    *ppFreedStruct = NULL;
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_compressAlloc()
 *
 *	This procedure allocates memory.
 *
 *	Input:
 *		blockSizeInBytes:	block size
 *		pointerFixup:		cTrue if pointers are to be
 *					adjusted after garbage collection.
 *		
 *	Returns:
 *		ptr to block, NULL if no block allocated.
 *
 **************************************************************************/
void *  CDECL 
i_compressAlloc (
    IPState    *pIPState,
    UInt32	blockSizeInBytes,
    UInt32	pointerFixup)
{
    PROLOGUE;
    pointerFixup = pointerFixup;	/* this is a dummy.  This routine
					 * is only specified when there is
					 * no fixup to be done.
					 */
	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    return MALLOC(blockSizeInBytes);
}


/* PUBLIC */
/**************************************************************************
 *   i_compressFree()
 *
 *	This procedure frees memory allocated by i_compressAlloc.
 *
 *	Input:
 *		pBlock:		Pointer to block to be freed
 *		length:		length of block in bytes.  This routine takes
 *				the length of the block to be freed
 *				to maintain compatibility with Ghostscript.
 *		
 *	Returns:
 *		nothing
 *
 **************************************************************************/
void  CDECL 
i_compressFree (
    IPState    *pIPState,
    void       *pBlock,
    UInt32	length)
{
    PROLOGUE;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;
    length = length;

    FREE(pBlock);
}





/*

 $Log: i_heap.c,v $
 Revision 1.2  2001/05/08 06:43:36  jlin
 change ->allocator to ->ipcAllocator

 * Revision 1.1.1.1  1999/09/20  18:18:11  jlin
 * Imported sources
 *
 Revision 1.6  1998/01/13 23:48:32  ddavies
 i_heap.c -> /product/ipcore/ipshared/src/RCS/i_heap.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.5  1997/07/11 01:49:00  ddavies
 i_heap.c -> /product/ipcore/ipshared/src/RCS/i_heap.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.4  1995/12/13 02:44:40  ejaquith
 i_heap.c -> /product/ipcore/ipshared/src/RCS/i_heap.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/07/10  23:41:14  ddavies
 * i_heap.c -> /product/ipcore/ipshared/src/RCS/i_heap.c,v
 *
 * Lotsa stuff having to do with adding pointer fixup inside persisten heaps.
 * Change inclusion of shrcmprs.* to inclusion of shrcc.* since the heap stuff
 * moved there.  Added inclusion of ptrfixup.pub to get definition of PointerFixup.
 * Added lots of comments about fixing pointers.
 * Greatly expanded the comments for i_allcateMemoryBlock to explain about
 * pointer fixup in persistent heaps and how to set the three new arguments:
 * fixupPointers, pAttachPt and fixupProc.  Added an argument to the
 * supplied allocator so that it can be signalled when we should be
 * notified when an allocated block is moved by the DAE garbage collector.
 * Change implementation of i_allocateMemoryBlock to use the new info.
 * Changed implementation of i_allocateStruct to both pass new args to
 * i_allocateMemoryBlock as needed and to decide that the first four bytes
 * of each structure in a heap is a (void *) instead of a
 * (TCBlockHeader *).  This is more correct, especially since a TCBlockHeader
 * no longer begins with a pNext pointer.  Did similar changes in
 * i_freeStruct.
 * Added an arg to i_compressAlloc so that it can be used as the allocator
 * supplied to the heap routines.  It ignore the pointerFixup arg since
 *
 * Revision 1.2  1994/07/14  21:01:50  ddavies
 * ddavies on Thu Jul 14 14:01:21 PDT 1994
 *
 * i_heap.c -> /product/ipcore/ipshared/src/RCS/i_heap.c,v
 *
 * Changed "memset" to "MEMSET" to get around casting problem caught by acc.
 *
 * Revision 1.1  1994/06/06  19:33:33  ddavies
 * Initial revision
 *
 * Revision 1.6  1994/05/28  00:10:41  ddavies
 * micmpmem.c -> /product/ipcore/ipshared/src/RCS/micmpmem.c,v
 *
 * Change ip_allocateMemoryBlock so that it clears all bytes in the block
 * allocated.  Sometimes we want to do something to all blocks of a given
 * type, this allows us to do that without worrying about uninitialized memory.
 *
 * Revision 1.5  1994/03/24  20:38:32  ddavies
 * micmpmem.c -> /product/ipcore/ipshared/src/RCS/micmpmem.c,v
 *
 * Change the procs that allocate and free blocks of structures so they use
 * an allocator and a free routine that are passed in instead of using
 * malloc and free.  This is needed for the symbol-based compression routines
 * where the structures may be long-lasting.  Also implement ip_compressAlloc
 * and ip_compressFree.  These are used for private structures.
 *
 * Revision 1.4  1994/02/09  01:28:48  ddavies
 * ddavies on Tue Feb 8 17:26:32 PST 1994
 *
 * micmpmem.c -> /product/ipcore/ipshared/src/RCS/micmpmem.c,v
 *
 * Fixed a bug in ip_allocateStruct.  When allocating a new memory block,
 * The code was using pMemoryBlock->pTop->structSize instead of
 * pMemoryBlock->structSize.  Ditto for numStructs.  Also added a new variable
 * called pNewBlock so that the code didn't use pTop to point to both the
 * structure and the new block of structures.
 *
 * Revision 1.3  1994/02/04  20:18:04  dferrell
 * micmpmem.c -> /product/ipcore/ipshared/src/RCS/micmpmem.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.2  1994/01/21  21:24:35  ddavies
 * micmpmem.c -> /product/ipcore/ipshared/src/RCS/micmpmem.c,v
 *
 * Fixed bug in ip_freeAllMemoryBlocks - replace FREE(&pCurrent) with
 * FREE(pCurrent).  Thanks to Ron Danisewicz at XIS for finding this!!
 *
 * Revision 1.1  1993/12/17  22:14:12  ddavies
 * Initial revision
 *

 */

