/*************************************************************
 *  Copyright (c) 1992-1996 Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * i_histeq.c	Histogram equalization functions
 *
 */

#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "memory.pub"
#include "iaerror.pub"
#include "frames.pub"

#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_histeq.c,v $");
/* $Date: 1999/09/20 18:18:11 $ */

/* PUBLIC */

/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Histogram equalizes a given intensity band of an 8-bit image
 *
 * LongDescription:::
 *  
 *
 * PrivateDescription:::
 *
 * Input:::
 *  lo::		lowest gray value in band to be equalized [0-254]
 *  hi::		highest gray value in band to be equalized [1-255]
 *  pS::		pointer to 8-bit input image (including frames)
 *  w::			framed width of input image
 *  h::			framed height of input image
 *  sBpl::		bytes per line of input image
 *  dBpl::		bytes per line of destination image
 *
 * Output:::
 *  pD::		pointer to 8-bit destination image (including frames)
 *
 * Returns:::
 *  ia_successful
 *	
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram analysis
 *
 * Keywords:::
 *  automatic enhancement, histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *  Digital Image Processing, Gonzalez and Woods, Addison-Wesley, 1992.
 *  pp 173-180.  ISBN 0-201-50803-6.
 *
 * EndDocumentation:::
 *
 **************************************************************************/
Int32 CDECL 
i_histogramEqualizeBand8 (
    IPState    *pIPState,
    Int32	lo,
    Int32	hi,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    Int32 i, j, c, range, count;
    Int32 *inHist = NULL;
    UInt8 *EQmap = NULL;

    ERR_GEN( ((lo >= hi) || (lo < 0) || (hi > 255)), ia_invalidParm );

    /* allocate the histogram table and the map */
    inHist = (Int32 *)CALLOC (256, sizeof (Int32));
    ERR_GEN_CLEANUP( (inHist == NULL), ia_nomem);

    EQmap = (UInt8 *)MALLOC (256 * sizeof (UInt8));
    ERR_GEN_CLEANUP ( (EQmap == NULL), ia_nomem);

    /* get the histogram for the input image */
    ERR_PASS_CLEANUP( i_histoRect (pIPState, pS, sBpl, cFrameBits, cFrameBits, 
				   w - 2*cFrameBits, h - 2*cFrameBits, inHist));

    /* set the parts of the map which aren't affected */
    for (i = 0; i < lo; i++)
        EQmap[i] = (UInt8)i;
    for (i = hi + 1; i < 256; i++)
        EQmap[i] = (UInt8)i;

    /* count the pixels in the specified region */
    for (i = lo, count = 0; i <= hi; i++)
    {
        count += inHist[i];
    }

    range = hi - lo + 1;

    /* compute the equalized section of the map */
    for (i = lo, j = lo, c = 0; i <= hi; i++)
    {
	c += inHist[i];
	EQmap[i] = (UInt8)j;
	/* this is slower but prevents roundoff errors */
	j = lo + (c * range / count);
    }

    /* apply the map to create the result */
    ERR_PASS_CLEANUP( i_map8To8 (pIPState, pS, w, h, sBpl, pD, dBpl, EQmap));

ERR_CLEANUP:
    /* cleanup all the tables */
    if (inHist != NULL)
        FREE (inHist);
    if (EQmap != NULL)
        FREE (EQmap);

    NORMAL_RETURN;

}  /* i_histogramEqualizeBand8 () */


/*
 * $Log: i_histeq.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 * Imported sources
 *
 * Revision 1.4  1998/01/13 23:48:34  ddavies
 * i_histeq.c -> /product/ipcore/ipshared/src/RCS/i_histeq.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.3  1997/07/03 21:54:03  jcraig
 * i_histeq.c -> /product/ipcore/ipshared/src/RCS/i_histeq.c,v
 *
 * use i_histoRect instead of i_histo.
 *
 * Revision 1.2  1996/08/06 18:05:57  dferrell
 * dferrell on Tue Aug 6 11:05:20 PDT 1996
 *
 * i_histeq.c -> /product/ipcore/ipshared/src/RCS/i_histeq.c,v
 *
 * Added typecastes to fix warning on NT build.
 *
 * Revision 1.1  1996/08/05 21:11:05  dferrell
 * Initial revision
 *
 *
 */


