/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_huff.c:	Utility procedures used generate Huffman codes.
 *
 *
 *	i_genHuffmanCodes()
 *		Generate a set of Huffman codes for a set of data items.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "memory.pub"

    /*  prototyes */
#include "shrcmprs.pub"
#include "shrcmprs.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_huff.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* It is often useful to generate names for a bunch of items when
 * compressing them.  For example, one might want to fill a file with
 * a series of coordinates.  If the distribution of the probability of
 * occurence of coordinate values is skewed (like 0 occurs a lot and 154
 * happens very seldom), it's often a good idea to generate Huffman codes
 * for all the items.  When the Huffman codes are substituted for
 * the items, the file will become shorter.  Often, much shorter.
 *
 * There are classical methods of generating Huffman codes.  We'll use
 * one (mostly).  Say you want to build codes for N items.  Say the count
 * associated with item i is Ci.  We want items with higher counts
 * to have shorter codes.  This means there's some sorting involved.
 * The count is the "primary" sort key. What do we do about items
 * with the same count?  Well, the user should supply a "secondary"
 * sorting key that can be used to consistently break ties.  No
 * two items with the same primary sorting key can have the same
 * secondary sorting key.  The secondary key might be based on the
 * order in which the items were first encountered, or some other
 * unique characteristic.
 *
 * The client should allocate an array of 2*N Huffman nodes.  Because
 * of the size of the values in the Huffman nodes, we can assign codes to
 * no more than 32768 items (N <= 32768).
 *
 * The client then sorts the original N items using the primary and
 * secondary keys as determinants.  Item 1 should be nearer the head
 * of the list than Item 2 if Item1.primaryKey > Item2.primaryKey
 * OR ((Item1.primaryKey == Item2.primaryKey) AND
 *     (Item1.secondaryKey < Item2.secondaryKey))
 * Note the switch in the sign of the comparisons.  The first primary
 * key must be BIGGER than the second or the primary keys must be equal
 * and the first secondary key must be SMALLER than the second.  This
 * is done so that items with big counts (primary keys) go to the head
 * of the list while the secondary key is used normally.
 *
 * The standard ANSI qsort routine can be used to sort the items.
 *
 * The "primaryKey" and "secondaryKey" entries of the first
 * N Huffman nodes are to be filled by the client, Huffman node i
 * corresponding to sorted client item i.  Array entry i should be
 * initialized to:
 *	count: number of times the item i occurs.
 *	index: the value of the
 *	All other structure entries: 0
 *
 * The classic algorithm:
 *   1.	Scan the array for the two lowest-value items as determined
 *	by the primary and secondary keys as used above.  Initially, these
 *	are the first two items.
 *   2.	Create a new Huffman tree node pointing to the two items.
 *	Use the next node in the Huffman tree array.  Its primaryKey is set
 *	to the sum of the two input primaryKeys.  Its secondaryKey is the
 *	minimum of the two input secondaryKeys.
 *   3. Insert the new tree node into the array such that the order of
 *	all elements is preserved.  In general, this means moving down
 *	the array until the right spot is found.
 *   4.	Loop back to step 1 until the array contains only a single node.
 *	In this process, we will build a Huffman tree since tree nodes are
 *	built on top of tree nodes.  This tree grows from the bottom
 *	towards the root which connections being made when needed.  Note
 *	that the items with the highest counts appear nearer the root
 *	of the tree since they're encountered last.
 *   5. Use this tree to find the length of the Huffman code for each
 *	item.  Traverse the tree down to each node.  The length of this
 *	path is the number of bits in the code for that symbol.  The
 *	number of bits for the root of the tree is is 0 (makes sense -
 *	if there's only one thing to encode, we don't need any bits to
 *	distinguish it from anything).
 *   6.	Assign the codes themselves using the "cononical Huffman code"
 *	asisgnment algorithm:
 *	a. Let c[len] be the number of codes of length len.
 *	b. Assume the maximum possible code length is 32.
 *	c. We will calculate f[len], the lowest code value for
 *	   all items having code length len as follows:
 *	f[32] = 0;
 *	for (len = 31; len >= 0; len--)
 *	    f[len] = (f[len+1] + c[len+1]) / 2;
 *
 * This is where we stop, though the classic algorithm goes on to actually
 * assign a Huffman code to each item as follows:
 *
 *   7.	Now assign all codes of length len by starting from f[len] and
 *	counting up.  The first code of length len gets a code value
 *	of f[len].  The next one gets a code value of f[len]+1 and
 *	so on.  It works!  Weird, huh?
 *
 * The reason we don't perform step 7 is that we don't know the order
 * in which the client wants to go through the nodes.  Most clients
 * will probably want to traverse items having the same code length in
 * the order determined by the primary and secondary keys.  However,
 * at least one client wants to use only the secondary key.  We return
 * the c[] and f[] arrays and let the client assign the codes as specified
 * in step 7 above.  We do record the code length for each item in
 * the item's Huffman structure (the first N in the array).  The client
 * can associate the Huffman nodes with whatever the client wanted to
 * assign codes to.  That brings up the other reason for letting the
 * client do the work. The codes will really go into some client-defined
 * structure and we don't want to have to know about it.
 * 
 *
 * The only trick we introduce is the method of keeping a representation
 * of the total order of the remaining items.  This is needed in step 1
 * and referred to in step 3.  At each pass through the loop, we remove
 * two items and add one, so the representation should be easy to update.
 * Some folks use a priority queue for this.  A priority queue is a very
 * clever thing.  Basically, it's a full, balanced (as much as possible)
 * binary tree in which the thing at each node is guaranteed to be
 * "smaller" (nearer the head of the list in our case) than all items
 * below it.  When a new item is added, it's added at the next vacant
 * position at the bottom of the tree.  Each row of the tree is filled
 * from left to right before a new row is started.  Once a node is added,
 * one swaps value of the new node and its parent, then that node and
 * its parent, etc, until the new value bubbles up high enough to be
 * smaller than everything below it.  When a node (usually the top node)
 * is removed, we move the node that was last inserted to the location of
 * the removed node, then bubble the moved thing up or down until order is
 * restored.
 *
 * This queue is very efficient since we only need to do at most log n
 * swaps to insert or remove an item from a queue of length n.  There are
 * a couple of inefficient things from our point of view however:
 * 1. The swaps: it would be nicer to avoid the memory writes.
 * 2. We always remove the top item.  When we remove it, we replace
 *    it with the bottom item.  This usually results in the maximum
 *    number of swaps.  We do this twice per iteration.
 * 3. When we add an item, it's usually a relatively small count since
 *    it was the sum of two smallest counts.  Since we add it to the bottom,
 *    it has to bubble to near the top.
 *
 * So how can we improve on this?  My thought is to notice that since the
 * items are originally sorted (and we keep the list sorted at each step),
 * i.e. A < B < C < D, then A+B < C+D.  That is, the new item produced
 * in each iteration is always strictly greater than the item produced
 * in the previous iteration.  This means we can move the insertion point
 * in the queue strictly upwards at each step.
 *
 * Say we allocate an array of a new type of structure.  This is the
 * ListPos structure.  There will be N ListPos structures in the array,
 * one for each new tree node.  Each ListPos structure is used to
 * mark the position of its tree node in the evolving list.  It
 * contains the array indices of the original Huffman tree nodes (i.e.
 * only includes nodes initialized by the client, not ones inserted here)
 * between it and the previous ListPos structure. These indices
 * are revised as the original tree nodes are processed.
 * Here's our detailed algorithm:
 *
 *   1.	Have the client fill the Huffman Tree array as specified above.
 *   2.	Allocate a ListPos array containing N ListPos structures.
 *   3. We're going to initialize the first Huffman and ListPos nodes.
 *	Set the next Huffman tree node by computing the sum of
 *	the first two Huffman tree node primaryKeys and the minimum of their
 *	secondaryKeys.  Also set the left and right pointers to these two
 *	child nodes.
 *   4. Do a binary search on the remaining Huffman tree positions to
 *	find where the first new node should go.  Say it fits between
 *	nodes k and k+1.
 *   5.	Initialize the first ListPos struct by setting its minLeft
 *	index to 2 (past 0 and 1) and its maxLeft index to
 *	k.  Set its HuffmanTree index to N (the node initialized in step 3).
 *   6.	Set the current insertion index to k, the current ListPos
 *	index to 0, the next ListPos index to 1, the current New Huffman
 *	tree pointer to index N+1 and the current Input Huffman Tree
 *	index to N.
 *   7. Here's the start of the loop.  Read the next two Huffman tree
 *	nodes specified by the current ListPos node.  These might be:
 *	a. two original Huffman tree nodes to the left of the ListPos node;
 *	b. one original Huffman node (there was only one) to the left
 *	   of the ListPos node plus its Huffman tree node;
 *	c. if there are no Huffman tree nodes in the list to the left
 *	   of the current ListPos's Huffman tree node, then one of the
 *	   new Huffman tree children is the current ListPos's Huffman tree
 *	   node.
 *	d. If there is another ListPos entry to the right of the current
 *	   one, use it's leftmost original Huffman tree node as the
 *	   other new child Huffman tree node.  If it doesn't have any
 *	   original Huffman tree nodes to the left of its Huffman Tree
 *	   node, use its Huffman tree node as the other new child node;
 *	e. If there was no other ListPos entry, we're done.  Exit the loop.
 *   8.	Initialize the next Huffman tree node by pointing at the two
 *	new child nodes.  Initialize its primaryKey and secondaryKey
 *	entries as in step 3.
 *   9.	If there are any remaining original Huffman tree nodes to the right
 *	of the current insertion point, do a binary search to find where
 *	to insert this new node.  We might speed things up here by first
 *	checking to see if inserting immediately to the right of
 *	the previous insertion point was correct.  Say the new
 *	insertion point is between nodes l and l+1.
 *  10. Initialize a new ListPos structure with a left range of k through l
 *	(set indices to -1 if k==l) and the index of the new Huffman tree node.
 *	Set the new range of original symbols to the right of the latest
 *	ListPos to be l+1 to N-1 (or [-1, -1] if l+1 > N-1).
 *  11.	Increment the next new Huffman and ListPos pointers.
 *  12. Loop back to step 7 until the 7e clause breaks us out of the loop.
 *
 * This algorithm creates the Huffman tree with a minimum of fuss (we hope).
 * From there, the algorithm is the same as the classical one specified
 * above.
 *
 * BUG FIX / FEATURE ADDITION: some knuckleheads (like me), find it
 * convenient to supply this routine with items having primary keys of 0.
 * This causes all sort of trouble.  As a first step, we will now skip
 * over all the nodes having a primary key of 0.  We assign them
 * code lengths of 31.
 */

/* PUBLIC */
/**************************************************************************
 *   i_genHuffmanCodeLengths()
 *
 *	This procedure assigns Huffman code lengths, in bits, to an array
 *	of HuffTreeNodes supplied by the client.  There are 32 possible
 *	code length classes (we assume code lengths range from 0 to 31).
 *	This routine also fills in an array containing the first
 *	Huffman code for each code length class.  We leave it to the client
 *	to assign the codes to the items in whatever order is appropriate.
 *	Codes are assigned by simply counting through all items in a code
 *	length class, assigning consecutive codes starting from the first
 *	code for that class.  We leave this task to the client because
 *	different clients may wish to assign codes in different orders.
 *	The order is often mandated by a compression specification.
 *
 *	This array also fills in an array containg the number of items
 *	in each code length class.  This is often useful when creating
 *	tables used to decode the Huffman codes.
 *
 *	Input:
 *		huffTreeArray:	ptr to first entry in array of
 *				2*N + 1 HuffTreeNodes.  This is used to
 *				assign code lengths to N items.
 *				Before calling this routine, the client
 *				must allocate this array and set the
 *				primaryKey and secondaryKey entries in
 *				each of the first N entries.  The primaryKey
 *				is the basic determinant of the code length
 *				of an item.  Items with larger primaryKeys
 *				will generally have shorter code lengths.
 *				The secondaryKey entries are used to break
 *				ties among items having unique primaryKeys.
 *				ALL secondary keys must be unique.
 *				IMPORTANT: the first N items in the array
 *				must already have been sorted (using qsort)
 *				using the primary and secondary keys.  Items
 *				with the smallest primaryKey entries
 *				(in case of tie, smallest secondaryKey
 *				entries) should appear at the head of the
 *				list.
 *		huffTreeArrayLen: Number of items (= 2*N + 1) in huffTreeArray.
 *		firstCodeArray:	ptr to first entry in array of 32 (0..31)
 *				first code values.  The contents of the
 *				array are filled in by this routine.
 *				firstCodeArray[i] will be set to the
 *				first Huffman code value for the class of
 *				items having Huffman code length i.
 *		numItemsArray:	ptr to first entry in array of 32 (0..31)
 *				values containing the number of items
 *				having Huffman code lengths.
 *				If this pointer is not NULL, this routine
 *				will set numItemsArray[i] to
 *				the number of items having Huffman codes
 *				of length i.  This is useful when preparing
 *				decoding tables.
 *
 *	Action:
 *		This routine build a Huffman tree from the original
 *		items and uses it to assign code lengths.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_nomem		ran out of memory for internal structs.
 *		ia_invalidParm		huffTreeArrayLen > 65534,
 *					huffTreeArrayLen == 0,
 *					huffTreeArrayLen is even
 *					
 *
 *
 **************************************************************************/
Int32  CDECL 
i_genHuffmanCodeLengths (
    IPState    *pIPState,
    HuffTreeNode *huffTreeArray,
    Int32	huffTreeArrayLen,
    UInt32     *firstCodeArray,
    Int32     *numItemsArray)
{
    PROLOGUE;
    HuffListPos		*listPosArray = NULL;
					/* array of HuffListPos structs
					 * used to remember locations of
					 * new HuffTreeNodes in the queue
					 * of remaining HuffTreeNodes.
					 */
    HuffListPos		*pSrcListPos;	/* ptr to current source HuffListPos
					 * entry, i.e. the one that will supply
					 * the next two HuffTreeNodes that
					 * will become children of a new node.
					 */
    HuffListPos		*pDestListPos;	/* ptr to next empty HuffListPos
					 * node */
    HuffTreeNode	*pSrcTreeNode0;	/* first child tree node */
    HuffTreeNode	*pSrcTreeNode1;	/* second child tree node */
    HuffTreeNode	*pDestTreeNode;	/* ptr to next empty HuffTreeNode */
    Int32		 i, di;		/* indices used to search for
					 * the position of a new HuffTreeNode
					 * in the queue. */
    UInt32		 curRight;	/* index of leftmost original
					 * dest tree node that's to the right
					 * of all new tree nodes. */
    UInt32		 maxRight;	/* index of rightmost
					 * original tree node. */
    HuffTreeNode	*pCurLeft;	/* ptr to leftmost unprocessed
					 * original node. */
    UInt32		locItemsArray[32];

	/* First, is the number of items ok? */
    if (((huffTreeArrayLen & 0x1) == 0) ||
	(huffTreeArrayLen > 0xFFFF) ||
	(huffTreeArrayLen == 0) || (huffTreeArrayLen == 1))
	return ia_invalidParm;

	/* initialize the output code length and first val arrays */
    for (i = 31; i >= 0; i--)
    {
       firstCodeArray[i] = 0;
       locItemsArray[i] = 0;
    }
	/* Take care of bozo case, some of the items in the array have
	 * a primary key of 0.  We skip these, assigning them code
	 * lengths of 31.  That ought to be safe, right?  We don't
	 * expect anyone to actually issue these codes.
	 */
    while (huffTreeArray->primaryKey == 0)
    {
	huffTreeArray->codeLength = 31;
	huffTreeArrayLen -= 2;
	huffTreeArray++;
    }

	/* If we've seen all the useful stuff, return after filling
	 * the output arrays with 0's.
	 */
    if (huffTreeArrayLen < 3)
    {
	if (numItemsArray != NULL)
	    memset (numItemsArray, 0, 32 * sizeof(UInt32));
	memset (firstCodeArray, 0, 32 * sizeof(UInt32));
	NORMAL_RETURN;
    }

	/* Take care of a pathological case.  What happens if we only have
	 * one node?  When we were pure, we'd assign it a code length of
	 * 0 and return.  This was done on the theory that no Huffman
	 * code is required to discriminate among a class of 1.
	 * Well, yeah, but what happens if you really do have a class
	 * of 1?  Rather than write a bunch of special code that knows
	 * how to produce and consume the magic code of length 0, we
	 * just pretend that there are exactly two codes, 0 and 1.
	 * This will happily fill all the tables.
	 */
    if (huffTreeArrayLen == 3)
    {
	huffTreeArray[0].codeLength = 1;	/* pretend there are two */
	huffTreeArray[1].codeLength = 1;	/* codes of length 1 */
	memset (firstCodeArray, 0, 32 * sizeof(UInt32));
	if (numItemsArray != NULL)
	{
	    memset (numItemsArray, 0, 32 * sizeof(UInt32));
	    numItemsArray[1] = 2;		/* pretend there are two
						 * codes of length 1 (0 and 1).
						 */
	}
	NORMAL_RETURN;
    }

	/* Set the left entry of the first N entries to
	 * 0xFFFF so we know they're leaves.  Note that all tree nodes have
	 * either 2 descendents or 0, so we only need to set one
	 * coord to -1.
	 */
    for (i = huffTreeArrayLen/2; i >= 0 ; i--)
	huffTreeArray[i].left = 0xFFFF;

	/* Also initialize the HuffTree node above the list client
	 * node.  We sometimes find it during the binary search.
	 */
    huffTreeArray[(huffTreeArrayLen - 1)/2].primaryKey = 0xFFFFFFFF;

	/* Allocate the matching array of HuffListPos items.  Note that
	 * we only need N of these since we only need 1 for each
	 * non-leaf node in the tree.
	 */
    listPosArray = (HuffListPos *) MALLOC ((huffTreeArrayLen / 2) *
					   sizeof(HuffListPos));
    if (listPosArray == NULL)
	return ia_nomem;

	/* Initialize the first HuffListPos entry */
    pDestListPos = &listPosArray[0];
	/* pt to first non-leaf node, skipping the dummy node at the
	 * end of the array.  Also pointer to the first two
	 * source nodes.
	 */
    pDestTreeNode = &huffTreeArray[(huffTreeArrayLen / 2) + 1];
    pSrcTreeNode0 = &huffTreeArray[0];
    pSrcTreeNode1 = &huffTreeArray[1];

    pDestTreeNode->primaryKey =
		pSrcTreeNode0->primaryKey + pSrcTreeNode1->primaryKey;
    pDestTreeNode->secondaryKey =
		MIN(pSrcTreeNode0->secondaryKey,
		    pSrcTreeNode1->secondaryKey);
    pDestTreeNode->left = (UInt16) 0;
    pDestTreeNode->right = (UInt16) 1;
    pSrcTreeNode1->parent = pSrcTreeNode0->parent =
				(UInt16) (pDestTreeNode - huffTreeArray);

	/* figure out where this new node goes in the remaining list
	 * of original nodes.  We want to do a binary search, so want
	 * to start in the middle of the remaining nodes.
	 * The index beyond the rightmost original
	 * node is at (huffTreeArrayLen / 2).  The index just to the
	 * right of the leftmost remaining node is at 2.  Therefore
	 * the new index is at ((huffTreeArray/2) + 2)/2.
	 *
	 * BUG (5/8/98).  We bias di to be a tiny bit too big so that
	 * we always get to the spot we want before it runs down to 1.
	 * This might mean a little oscillating at the end of the search,
	 * which is ok.  However, the problem comes when the search moves
	 * relentlessly to one end of the array or the other.  In that
	 * case, we want to make sure we don't go off the beginning or
	 * end of the array.
	 */
    curRight = 2;
    maxRight = huffTreeArrayLen/2;
    i = (maxRight + curRight)/2;
    di = (maxRight - curRight)/4 + 1;

    while (cTrue)
    {
	if ((pDestTreeNode->primaryKey > huffTreeArray[i].primaryKey) ||
	    ((pDestTreeNode->primaryKey == huffTreeArray[i].primaryKey) &&
	     (pDestTreeNode->secondaryKey > huffTreeArray[i].secondaryKey)))
	    i += di;
	else
	    i -= di;

	if (i < 0)
	    i = 0;
	if (i > (Int32) maxRight)
	    i = maxRight;

	if (di > 1)
	    di = (di + 1)/2;
	else
	    break;
    }

	/* Now we know that the first node should be inserted on one
	 * side or the other of node i.  Figure out which.
	 */
    if ((pDestTreeNode->primaryKey > huffTreeArray[i].primaryKey) ||
	((pDestTreeNode->primaryKey == huffTreeArray[i].primaryKey) &&
	 (pDestTreeNode->secondaryKey > huffTreeArray[i].secondaryKey)))
    {
	pDestListPos->pNextLeft = &huffTreeArray[i];
	curRight = i+1;
    }
    else
    {
	pDestListPos->pNextLeft = &huffTreeArray[i-1];
	curRight = i;
    }
    pDestListPos->pHuffNode = pDestTreeNode;

	/* Point to the new HuffListPos node as the place to get the
	 * next two source nodes.  Then increment the dest node
	 * pointers to the next nodes in the arrays since they will
	 * be the next places to put new tree nodes.
	 */
    pCurLeft = &huffTreeArray[2];
    pSrcListPos = pDestListPos;
    pDestListPos++;
    pDestTreeNode++;

	/* Here's the main loop.  This is where we build the Huffman
	 * tree.  In each pass through the loop, we remove the leftmost
	 * (smallest primaryKey) nodes, create a new node and insert
	 * it into the queue as shown above.  When there's only one
	 * node left in the queue, we quit.  That's the root of the
	 * Huffman tree.
	 */
    while (cTrue)
    {
	    /* Go get two tree nodes */
	if (pCurLeft < pSrcListPos->pNextLeft)
	{
	    pSrcTreeNode0 = pCurLeft;
	    pSrcTreeNode1 = &pCurLeft[1];
	    pCurLeft += 2;
	}
	else
	if (pCurLeft == pSrcListPos->pNextLeft)
	{
	    pSrcTreeNode0 = pCurLeft;
	    pCurLeft++;
	    pSrcTreeNode1 = pSrcListPos->pHuffNode;
	    pSrcListPos++;	/* pt to next list node */
	}
	else	/* pCurLeft > pSrcListPos->nextLeft, i.e.
		 * there aren't any original nodes left to the left
		 * of pSrcListPos. */
	{
	    pSrcTreeNode0 = pSrcListPos->pHuffNode;
	    pSrcListPos++;

	    if (pSrcListPos == pDestListPos)
	    {
		    /* Are there any more original nodes to read?
		     * If so, they must all be above us, so
		     * pCurLeft must equal &huffTreeArray[curRight].
		     * We can get into this state
		     * when the new node is inserted immediately
		     * above one of its constituent nodes.
		     */
		if (curRight < maxRight)
		{
		    pSrcTreeNode1 = pCurLeft;
		    pCurLeft++;
		    curRight++;
		}
		else
		{
			/* We're trying to read a node that hasn't been
			 * created yet and there are no more original
			 * nodes, so we must be done.
			 */
		    break;
		}
	    }

		/* We're not done.  If there are any original nodes to
		 * the left of the next inserted node, get one.  Otherwise
		 * get the next HuffTreeNode
		 */
	    else
	    if (pCurLeft <= pSrcListPos->pNextLeft)
	    {
		pSrcTreeNode1 = pCurLeft;
		pCurLeft++;
	    }
	    else
	    {
		pSrcTreeNode1 = pSrcListPos->pHuffNode;
		pSrcListPos++;
	    }
	}

	    /* We've got the two source HuffTreeNodes.  Update the new
	     * HuffTree Parent Node.
	     */
	pDestTreeNode->primaryKey =
		    pSrcTreeNode0->primaryKey + pSrcTreeNode1->primaryKey;
	pDestTreeNode->secondaryKey =
		    MIN(pSrcTreeNode0->secondaryKey,
			pSrcTreeNode1->secondaryKey);
	pDestTreeNode->left = (UInt16) (pSrcTreeNode0 - huffTreeArray);
	pDestTreeNode->right = (UInt16) (pSrcTreeNode1 - huffTreeArray);
	pSrcTreeNode1->parent = pSrcTreeNode0->parent =
				(UInt16) (pDestTreeNode - huffTreeArray);

	   /* Figure out where the new node goes */
	i = (maxRight + curRight)/2;
	di = (maxRight - curRight)/4 + 1;
	while (cTrue)
	{
	    if ((pDestTreeNode->primaryKey > huffTreeArray[i].primaryKey) ||
		((pDestTreeNode->primaryKey == huffTreeArray[i].primaryKey) &&
		 (pDestTreeNode->secondaryKey > huffTreeArray[i].secondaryKey)))
		i += di;
	    else
		i -= di;

		/* Bug fix (5/8/98):  We make di a little big just so that
		 * we won't run out of di before we find the right spot.
		 * That means we may oscillate a little before settling down,
		 * but that's ok.  The bug is that if we move relentlessly
		 * towards one end of the array, we may run off the end.
		 * Therefore, need to prevent di from going too far.
		 */
	    if (i < 0)
	        i = 0;
	    if (i > (Int32) maxRight)
	        i = maxRight;

	    if (di > 1)
		di = (di + 1)/2;
	    else
		break;
	}

	    /* Now we know that the first node should be inserted on one
	     * side or the other of node i.  Figure out which.
	     */
	if ((pDestTreeNode->primaryKey > huffTreeArray[i].primaryKey) ||
	    ((pDestTreeNode->primaryKey == huffTreeArray[i].primaryKey) &&
	     (pDestTreeNode->secondaryKey > huffTreeArray[i].secondaryKey)))
	{
	    pDestListPos->pNextLeft = &huffTreeArray[i];
	    curRight = i+1;
	}
	else
	{
	    pDestListPos->pNextLeft = &huffTreeArray[i-1];
	    curRight = i;
	}
	pDestListPos->pHuffNode = pDestTreeNode;

	    /* Point to the next dest tree nodes */
	pDestTreeNode++;
	pDestListPos++;
    }

	/* Groovy.  We have a Huffman tree.  Traverse it, setting code
	 * lengths and counting the number of items of a given length.
	 * First, set the parent of the root to NULL.  Falling off
	 * that end is how we know we've traversed the entire tree.
	 * There's a tiny trick we use a couple of times.  Every
	 * node in this tree either has 2 children or 0 children.
	 * We can tag the left child index of a node with 0xFFFF
	 * to say it's a leaf and 0xFFFE to say we're done with
	 * this node.  Once we see one of these tags, we know
	 * we don't have to look at the right child.
	 */
    pSrcTreeNode0->parent = 0xFFFF;
    i = pSrcTreeNode0 - huffTreeArray;		/* the current index */
    di = 0;					/* the current level */
    while (cTrue)
    {
	    /* can we go down to the left?  If so, go down */
	if ((pSrcTreeNode0->left & (UInt16) 0xFFFE) != (UInt16) 0xFFFE)
	{
	    i = pSrcTreeNode0->left;
	    pSrcTreeNode0 = &huffTreeArray[i];
	    di++;
	}
	else
	{
		/* We can't go down this path, at least to the left,
		 * any farther.  See if this is a leaf node.
		 * If so, mark this node's code length and
		 * increment the count of things at this length.
		 */
	    if (pSrcTreeNode0->left == (UInt16) 0xFFFF)
	    {
		pSrcTreeNode0->codeLength = (UInt16) di;
		locItemsArray[di]++;
	    }

		/* if we're at the root, quit */
	    if (pSrcTreeNode0->parent == (UInt16) 0xFFFF)
	        break;

		/* If we're not at the root.  We logically want to go
		 * up in the tree.  There are two choices.  If
		 * we haven't visited our parent's right child yet,
		 * go there.  If we've visited that, go up to
		 * our parent, after marking it so we won't bounce
		 * back down here.
		 */
	    if (huffTreeArray[pSrcTreeNode0->parent].left == (UInt16) i)
	    {
		i = huffTreeArray[pSrcTreeNode0->parent].right;
		pSrcTreeNode0 = &huffTreeArray[i];
	    }
	    else	/* we've already explored the right
			 * branch.  Mark the parent as being
			 * done and go back up to it.
			 */
	    {
		i = pSrcTreeNode0->parent;
		pSrcTreeNode0 = &huffTreeArray[i];
		pSrcTreeNode0->left = (UInt16) 0xFFFE;	/* seen but not leaf */
		di--;
	    }
	}
    }

	/* We've now marked all the client leaf nodes with code lengths
	 * and computed the number of nodes having each length.
	 * Go compute the first code value for each length.  This
	 * loop is copied directly from my Huffman spec.  It just
	 * works out!
	 */
    firstCodeArray[31] = 0;		/* = (0 + 0) / 2 */
    for (i = 30; i >= 0; i--)
    {
	firstCodeArray[i] = (firstCodeArray[i+1] + locItemsArray[i+1]) / 2;
    }

	/* that's it! return happily.  If we were passed a numItemsArray,
	 * fill it in.
	 */

    if (numItemsArray != NULL)
    {
	memcpy( (void *) numItemsArray, (void *) &locItemsArray[0],
		32 * sizeof(UInt32));
    }

    if (listPosArray != NULL)
	FREE(listPosArray);

    NORMAL_RETURN;
}





