/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_itree.c: 	Interpretation tree-based search strategies for matching
 *              generic "model" features with generic "data" features.
 */


#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "ipserrs.pub"
#include "errmacs.pub"
#include "defines.pub"
#include "memory.pub"

/* prototypes */
#include "shrpatt.pub"
#include "shrutils.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: i_itree.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Match arbitrary data and model features via an interpretation tree matcher
 *
 * LongDescription:::
 *  Give two lists of "features", this routine will attempt to match them
 *  pairwise.  This is mainly only the searching logic;  the rules of the
 *  matching are entirely up to the client.  For example, the client may
 *  decide that two data feature may match the same model feature.<BR><BR>
 *
 *  An interpretation tree matcher's goal is to match N data features up
 *  to M model features.  Two lists of rectangles, where one is some "ideal"
 *  (model) set and the other is another (say results from a scan of the
 *  page of the model rects) set referred to as data features.  We want to
 *  try to match each data feature to some model feature.  The search technique
 *  looks at each new data feature to consider, and tries to match it with
 *  every model feature, including "no model feature", which is referred to
 *  as a match with the "null branch" in the search tree.  Thus, this is
 *  an order(N to the M) algorithm.<BR><BR>
 *
 *  The way the algorithm tries to be clever is in the matching criterion.
 *  <BR><BR>
 *
 *  Constraints::
 *   There are three ways to constrain the matching.  The first is the
 *   simplest, unary constraint checking.  This asks, "can this one data
 *   feature match this one model feautre?" in essentially a context-free
 *   environment.  A rectangle-matching example might just check to
 *   make sure that the features' widths are similar and that their heights
 *   are as well.  After passing the first, the second matching constraint
 *   is the binary constraint.  This asks, "given this unary match of data
 *   feature d and model feature m, is the relationship between d and the
 *   previously matched d-i data feature 'consistent' with the relationship
 *   between m and the model feature which matched the d-i data feature,
 *   for i in [1..d]?".  For rectangles, we might compare the distances
 *   between the upper left corners of features, and deem "consistent" to
 *   mean that the two distances are within so many pixels of each other.
 *   The last constraint is the model test.  This is only done at the
 *   "leaf" level, when all the data features have matched a model feature
 *   (or the null branch).  Typically, this is where the more expensive
 *   consistency testing should be done.  Additionally, the client is to
 *   use this routine call to record data-model matchings.
 *
 *  Match checking order::
 *   Once a new node is reached, we always begin checking this new data
 *   feature for a match with the 0th model feature, the 1st model
 *   feature, ..., the last model feature, the null branch (no model feature).
 *   When a data feature matches the null branch, the unary and binary 
 *   constraint checks are always considered to be true (and therfore the
 *   checks are skipped), but note that the model test will be invoked
 *   if we are at a leaf data feature in the search tree.
 *
 *  Speed::
 *   What makes this fast is by making the unary and binary constraints
 *   fast checks!  It is often the case that much of the necessary information
 *   can be pre-computed.  For example, if we were matching rectangles, we
 *   might create arrays with minimum and maximum widths for each model
 *   feature for the unary constraint comparison.  Thus, we just check if
 *   the data feature's width is in the model feature's width/height range.
 *
 *  Pruning::
 *   Using minNMatchesDtoM, we instruct at what point in the search tree
 *   to prune any further search.  For example, if we are at depth 7 in a
 *   depth 10 tree, have 3 data feature matches so far, and minNMatchesDtoM
 *   is > 7, we can prune the search.  Notice that we will consider
 *   minNMatchesDtoM-1 data feature matches to meet this criterion
 *   and will invoke the modelTest() routine passed as a parameter
 *   iff we are at a leaf node.  [This is because minNMatchesDtoM only tells us
 *   when to _prune_.  If we had minNMatchesDtoM-1 data features at depth
 *   n-1 in a depth n tree, then we would not have pruned the search and would
 *   have searched down into depth n.  Whether or not this leaf node matches
 *   will determine whether we get minNMatchesDtoM or minNMatchesDtoM-1!]
 *
 * Input:::
 *  nModelFeatures::        Number of model features that we may match data 
 *                          features to.
 *  nDataFeatures::         Number of data features, each of which we will
 *                          attempt to find a match with a model feature.
 *  minNMatchesDtoM::       This is the number of data features to model
 *                          features that we will find minimally acceptable
 *                          to continue traverse the tree.  In other words,
 *                          it is the number for PRUNING the search.
 *                          This number must be in range [1..nDataFeatures].
 *                          [A null branch match (when a data feature matches
 *                          no model features) is not considered a data to
 *                          model feature match.]
 *  unaryConstraints::      Ptr to the client's routine to perform the
 *                          data and model feature match test.
 *                          Passed ptr to client's info and the data and the
 *                          model feature we are considering for a match.
 *                          Routine passes back whether or not test passed.
 *  binaryConstraints::     Ptr to the client's routine to perform the
 *                          data and model feature pairings consistency
 *                          test.
 *                          Passed ptr to client's info; the data feature
 *                          number we just matched, d, where all data features
 *                          [0..d-1] have matched; the array, elements [0..d]
 *                          filled, indicating that the "index"th data feature
 *                          matched the "array[index]"th model feature;
 *                          the array value that indicates that the "index"th
 *                          data feature matched no model feature
 *                          (null branch).
 *                          Routine passes back whether or not test passed.
 *  modelTest::             Ptr to client's routine to perform the "model
 *                          test" check and report leaf match information.
 *                          Passed number of data and model features; the
 *                          number of data features that matched actual model
 *                          features; matchesDtoM and pITStuff as above.
 *                          Routine passes back whether or not to increase
 *                          the pruning value so only leafs at least this good
 *                          get considered in the future.
 *  pITStuff::              Ptr to client's information, passed into each
 *                          invocation of the client's passed routines.
 *
 * Output:::
 *  pPassed::               Ptr to cTrue if a data-model feature pairing
 *                          with client-specified "minNMatchesDtoM"
 *                          made it to an invocation of "modelTest".
 *                          Ptr to cFalse, o.w.
 *
 * Returns:::
 *
 * Notes:::
 *  minNMatchesDtoM::
 *   Note the "modelTest" routine may instruct this value to be updated
 *   as the routine proceeds.
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_interpTreeMatcher (
    IPState    *pIPState,
    Int32	nModelFeatures,
    Int32	nDataFeatures,
    Int32	minNMatchesDtoM,
    Int32 (CDECL *unaryConstraints) (
		    IPState    *pIPState,
		    void       *pITStuff,
		    Int32	dataFeatureNum,
		    Int32	modelFeatureNum,
		    Bool       *pPassed),
    Int32 (CDECL *binaryConstraints) (
		    IPState    *pIPState,
		    void       *pITStuff,
		    Int32	d,
		    Int32      *matchesDtoM,
		    Bool       *pPassed),
    Int32 (CDECL *modelTest) (
		    IPState    *pIPState,
		    Int32	nDataFeatures,
		    Int32	nModelFeatures,
		    Int32      *matchesDtoM,
		    Int32	nMatchesDtoM,
		    void       *pITStuff,
		    Bool       *pNextNMatchesDtoMAsGood),
    void       *pITStuff,
    Bool       *pPassed)
{
    PROLOGUE;
    typedef enum { searchDown, searchAcross, backtrack, done } modeType;
    Int32        i;
    modeType     mode;
    Int32        modelNum = 0, nMatchesDtoM;
    Int32        d = -1;
    Bool         passed, nextNMatchesDtoMAsGood, gotMatching = cFalse;
    Int32       *matchesDtoM = NULL;

    /* There must be at least this many data features that DON'T match the
     * null branch (no model feature) to be considered for the overall
     * (typically much more expensive and exhaustive than just pairwise
     * consistency checks) model test as a potential data-model pairing.
     */
    ERR_GEN(minNMatchesDtoM < 0  ||  minNMatchesDtoM > nDataFeatures,
	    cErrInvalidParm);


    /* We'll be filling in each index of this array with values
     * {0..nModelFeatures-1} (inclusive) or
     * NO_FEATURE_MATCHED to indicate the NULL branch match.
     */
    matchesDtoM = (Int32 *)IP_CALLOC(nDataFeatures, sizeof(Int32));
    
    ERR_GEN(matchesDtoM == NULL,  cErrNomem);

    for (i = 0; i < nDataFeatures; i++)
	matchesDtoM[i] = NO_FEATURE_MATCHED;
    
    /* nMatchesDtoM will be a running counter to indicate the number of
     * Data features fitted to Model features that are "real", ie not the
     * Null branch model feature.
     */
    nMatchesDtoM = 0;
    


    /* Here's the logic behind the interpretation tree searching.
     * I'll present it using Finite State Automata states, without the
     * snazzy little diagrams from state to state (the graph is left
     * as an exercise for the reader! ;-)
     *
     * There are essentially five basic states:
     *   begin, searchDown, searchAcross, backtrack, end.
     *
     * Begin:
     *  ==> searchDown
     *    always
     *
     * searchDown:
     *  ==> searchAcross
     *    Move (down in the tree) to the next data feature to be considered.
     *    Always, except when
     *  ==> backtrack
     *    We decide to prune the search tree.
     *
     * searchAcross:
     *  ==> searchAcross
     *    The model feature we just considered did not match the current
     *    data feature.  (So check the next model feature in the tree).
     *  ==> searchDown
     *    The model feature we just considered did match the current data
     *    feature.  If the model feature is the special case null branch
     *    feature (or "wild card"), then it always matches!
     *  ==> backtrack
     *    The data feature/depth we just considered is the last data feature
     *    in the tree.
     *
     * backtrack:
     *  ==> searchAcross
     *    The model feature we just considered was NOT the special null branch
     *    feature.
     *  ==> backtrack
     *    The model feature was the special null branch feature, and thus
     *    it means that we have tried matching the data feature at this depth
     *    in the tree with every possible model feature (including the null
     *    branch), so we should move up a level in the tree and try again.
     *    Always, except when
     *  ==> end
     *    We are at the first data feature/the top node in the tree!
     *
     * end:
     *  smile, we are done.
     *
     */


    /* We'll use the variable "d" to be both the level in the tree that
     * we are at [0..nDataFeatures-1], 
     * and to simultaneously be the data feature array index (ie the data
     * feature number we are trying to match).
     * Hence, this is why
     *   when we are at the current tree depth "d"
     *   we are trying to match data feature "d".
     */


    mode = searchDown;

    /* stay in tree control blocks, either backtracking or searching,
     * until tree is finished being searched for possible solutions
     */
    while (mode != done)
    {
	while (mode == backtrack)
	{
	    /* We got here one of 3 ways:
	     * (1) at present depth, we just matched a feature (either real
	     *     or null branch) but as we tried to search down, we decided
	     *     to prune the searching.
	     * (2) the present depth is the last level (ie. we are at a leaf)
	     *     in the tree
	     * (3) we are still backtracking
	     *
	     * Once here, we want to check any other remaining model features
	     * for matches (including null branch) at this depth/data feature.
	     * There's no more to check obviously, if we are trying to 
	     * backtrack from a null branch matching, so we'll have to keep
	     * backtracking UP a level (choice 3 above) and try this again...
	     */

	    /* to backtrack in tree, if we can still try another model
	     * feature at a given node, continue search through this
	     * feature 
	     */
	    if (modelNum < nModelFeatures)
	    {
		modelNum++;

		/* decrement the number of (non null branch) matches */
		nMatchesDtoM--;
		mode = searchAcross;
	    }
	    else
	    {
		/* This means we just tried the null branch and want to
		 * backtrack.  That mean's we're done attempting to match
		 * the current data feature with all the model features
		 * (including null branch).
		 *
		 * Since done with current node and wish to backtrack, 
		 * go up one
		 * node by decreasing depth and starting at where
		 * we left off at
		 * this node (i.e. which model feature is to be checked next)
		 */
		matchesDtoM[d] = NO_FEATURE_MATCHED;

		d--;
		if (d < 0)
		    /* have finished searching tree, root is done */
		    mode = done;
		else
		    /* keep mode = backtrack */
		    modelNum = (matchesDtoM[d] != NO_FEATURE_MATCHED ?
				matchesDtoM[d] : nModelFeatures);
	    }
	}

	while (mode == searchAcross || mode == searchDown)
	{
	    if (mode == searchDown)
	    {
		if ((nMatchesDtoM + nDataFeatures-1 - d) >= minNMatchesDtoM)
		{
		    /* add one to depth and start at first model feature
		     * in this
		     * new node 
		     */
		    d++;
		    modelNum = 0;
		    mode = searchAcross;
		}
		else
		    /* Prune the search here.  Can't ever get path
		     * with at least minNMatchesDtoM
		     * non null branch matches below this node.  
		     * Go back one node.
		     */
		    mode = backtrack;
	    }

	    if (mode == searchAcross)
	    {
	
		if (modelNum >= nModelFeatures)
		{
		    /* wild card match, the null branch*/

		    matchesDtoM[d] = NO_FEATURE_MATCHED;

		    if (d == (nDataFeatures-1))
		    {
			/* At leaf, so save/check possible final matching */
			ERR_PASS_CLEANUP( modelTest (pIPState,
						     nDataFeatures,
						     nModelFeatures,
						     matchesDtoM,
						     nMatchesDtoM,
						     pITStuff, 
						     &nextNMatchesDtoMAsGood));
			gotMatching = cTrue;
			
			if (nextNMatchesDtoMAsGood)
			    minNMatchesDtoM = MAX(minNMatchesDtoM,
						  nMatchesDtoM);

			/* This means we just checked the last feature, the
			 * null branch, at a leaf.  Thus, we'll want to go
			 * up one in the backtracking mode.
			 */
			mode = backtrack;
		    }
		    else
			/* Consistent by virtue of this being the null
			 * branch "feature" that always successfully matches,
			 * so continue search down tree.
			 */
			mode = searchDown;
		}
		else
		{
		    /* real model feature */

		    matchesDtoM[d] = modelNum;

		    ERR_PASS_CLEANUP( unaryConstraints (pIPState, pITStuff,
							d, modelNum, 
							&passed));
		    
		    if (passed) {

			/* Can use below for debugging information */
#if 0
			fprintf(fp_out,
				"---Unary: Data #%ld and Model #%ld features---\n",
				d, modelNum);
#endif

			ERR_PASS_CLEANUP( binaryConstraints (pIPState,
							     pITStuff, 
							     d, 
							     matchesDtoM,
							     &passed));
			
			if (passed)
			{
			    /* node is consistent, so increase number of
			     * nonnull matches */
			    nMatchesDtoM++;
			    if (d == (nDataFeatures-1))
			    {
				/* At leaf, so save/check possible final
				 * matching
				 */
				ERR_PASS_CLEANUP( modelTest (
						     pIPState,
						     nDataFeatures,
						     nModelFeatures,
						     matchesDtoM, 
						     nMatchesDtoM, 
						     pITStuff, 
						     &nextNMatchesDtoMAsGood));
				gotMatching = cTrue;

				if (nextNMatchesDtoMAsGood)
				    minNMatchesDtoM = MAX(minNMatchesDtoM,
							  nMatchesDtoM);
				
				/* We'll go to the backtracking state.  Notice
				 * that there will still be at least one
				 * model feature to check for a match at/with
				 * this depth/data feature, so in the
				 * backtracking state, we'll go to the
				 * searchAcross state!
				 */
				mode = backtrack;
			    }
			    else
				/* not at leaf, so continue search down tree
				 * to next node */
				mode = searchDown;
			}
		    }
		    
		    if (!passed)
		    {
			modelNum++;
			mode = searchAcross;
		    }
		    
		} /* end of real model feature */

	    }
	}
    }

    IP_FREE(matchesDtoM);
    *pPassed = gotMatching;

    NORMAL_RETURN;

ERR_CLEANUP:
    IP_FREE(matchesDtoM);
    ERR_RETURN;
}


/*

$Log: i_itree.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.4  1998/01/13 23:48:40  ddavies
i_itree.c -> /product/ipcore/ipshared/src/RCS/i_itree.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1997/07/11 01:49:03  ddavies
i_itree.c -> /product/ipcore/ipshared/src/RCS/i_itree.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.2  1996/02/22 01:34:21  ejaquith
ejaquith on Wed Feb 21 17:34:02 PST 1996

i_itree.c -> /product/ipcore/ipshared/src/RCS/i_itree.c,v

Updated comments.
Changed variable names to be more intuitive as to what they are/do.

 * Revision 1.1  1996/01/31  02:51:17  ejaquith
 * Initial revision
 *

*/
