/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_ldc.c: High level code used in Lossy Dither compression.
 *
 *   	Int32	i_compressRowsLD()
 *			Compresses a block of scan lines using
 *			the Lossy Dither algorithm.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#define _NEED_LDCLASS_
#include "ldc.prv"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_ldc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* This file contains the routine used to compress many rows of data using
 * Lossy Dither encoding.  This method is novel to Xerox.
 *
 * There is a big comment all about the low level mechanics of our
 * implementation of LD compression in mildc.c.  This comment will
 * deal with software engineering and interfaces on the next higher
 * level.  The basic style of interfaces is shared by the CCITT
 * compression routines and the PackBits compression routine.
 * 
 * Lossy Dither encoding uses strings on one scan line to encode the
 * data on the next scan line.  The "coding" line is the line being
 * encoded.  The "reference" line is the one above the coding line.
 * There are two byte arrays in the encoder state block.  We ping-pong
 * between them; filling one and using it for the coding line, then
 * filling the other and using the first for the reference line,
 * then filling the first and using the second for the reference line
 * and so on.  The line1IsCoding boolean in the compression data
 * structure indicates whether the second byte array is the coding
 * line (cTrue) or the reference line (cFalse).
 *
 * This routine has two jobs at the start of each scan line.  It must
 * fill the coding array with equivalence class numbers (bytes are
 * put into equivalence classes using a table).  But first, it must set up
 * data structures that we use to search the reference line.
 *
 * Filling the coding line is easy.  We just look up each image
 * byte (reading left to right in BYTE order) in the ldClass table
 * and put the result into the coding line array.  Note that we read
 * 4-byte words from the image and write bytes to the coding array.
 * Putting the coding array in byte order helps to encode strings of
 * bytes.
 *
 * Setting up the reference line data structures is a little trickier.
 * There is one structure.  Remember that we encode the coding line
 * bytes using strings that are copied direclty down from the reference
 * line alternating with strings of literals that are supplied by the
 * encoder.  The data structure records the ending positions of the
 * copy strings that made up the reference line.  We'll try to use them
 * to encode the ends of copy strings that make up the coding line.
 *
 * To reduce checking overhead, the two byte buffers each have two dummy bytes
 * at the end.  The first dummy bytes have values that are greater than
 * the biggest equivalence class number *and* are different from each
 * other.  This guarantees that they won't match anything including
 * each other, so all string matches will terminate at the dummies.  The
 * second dummy bytes are both equal to 0.  This guarantees that our
 * search for literal strings that follow copy strings always terminates.
 *
 * To initialize the structure, we inspect the reference line and the
 * to-be-coding line.  The to-be-coding line holds the bytes from the
 * line above the reference line.  We execute a simple loop that contains
 * two sequential inner loops.  The first looks for strings of matching
 * bytes.  It enters an offset of the last pair of matching bytes into
 * the marker array.  The second inner loop then skips over any non-matching
 * bytes.  The result of executing the outer loop is that the marker array
 * is filled with locations where the copy strings end.
 *
 * After the bytes have been classified into equivalence class numbers
 * and the reference line data structures have been set up, we call
 * ip_compLDRow to encode the equivalence class numbers in the
 * coding line into LDC codes.
 *
 * Note that the encoded data stream is always written out in the
 * platform's natural byte order.  It will have to show up in that
 * order at some point and (theoretically) there aren't very many
 * encoded bytes, so this is no problem, right?
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which this procedure will stop.  It might actually
 * finish encoding the entire scan line.  When this happens, we update the
 * pointer to the compressed data that was passed in, set the state
 * to cLDCompNewLine and store the current accumulator and vacantBits
 * and return happily.
 *
 * The other reason for returning is that we run out of room for
 * compressed data.  ip_compLDRow detects this by comparing pCurCompData to a
 * pointer to the end of the buffer just before writing encoded data.
 * Since we want to pretend that we have a callback to fill the buffer,
 * we need a restart location next to every check.  There are
 * 5 checks, hence 5 restart locations.
 *
 *			*	*	*
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * and the continueImageBlock and continueCompDataStream args.
 * The caller also supplies pointers to words used to return the
 * exhaustedImageData, exhaustedCompData, bytesWritten and percentDone args.
 * The bytesWritten arg returns the number of compressed bytes written so far.
 *
 * We do not supply a flushPipeline argument.  We'll make LDC strips end
 * in the way as MMR strips.  That means the compressed data for
 * every LDC scan line should be followed immediately by the data for
 * the next scanline or by two EOLs (the EndOfFacsimileBlock code).
 * It's possible that the buffer may fill in the midst of writing these.
 * It's a real pain to figure out how far we've gotten when trying to
 * restart, so we just don't bother.  We recommend that the client use
 * i_writeLDCEOL and i_flushLDC, with proper attention paid to
 * detecting a full compression buffer.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	This means we compressed the desired number of
 *	image lines without filling the compressed data buffer.
 *	The routine can be restarted with new startLine (and
 *	numLines), continueImageBlock = cFalse.  
 *	Some of the compressed data may still be in the
 *	pipeline and the saved compressed data state is valid.
 *	This routine may be called again with
 *	continueCompDataStream = cTrue.  The client must supply
 *	a new value for pCompData and compDataSize.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to fill the
 *	compressed data buffer at the same time the compressed
 *	data came to an end.  As before, there is probably be more
 *	compressed data in the pipeline.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of room for compressed data before writing the last
 *	byte of compressed data requested.  Some of the data to be written
 *	is in the pipeline.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */

/* PUBLIC */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress rows of a binary image using the Lossy Dither technique.
 *
 * LongDescription:::
 *  Compress some rows of an image using the Lossy Dither algorithm.
 *  We will stop when either an error is encountered, we run out of image rows
 *  or we fill the compressed data buffer.
 *
 * Input:::
 *  pLDCompInfoIn::         Pointer to private LDC compression state block
 *  pImage::                Pointer to the upper left corner of the frame
 *			    of the uncompressed image.
 *  w::             	    Width of uncompressed image including frame.
 *  startLine::             First image line to be compressed
 *			    relative to the top of frame.
 *  numLines::              Number of lines to be compressed.
 *  bpl::		    Bytes/line of uncompressed data.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    written.  If several calls
 *			    to this routine are needed to create
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer.
 *  continueImageBlock::    cTrue => we are compressing
 *			    from the same slice of image as was
 *			    specified in the previous call and the
 *			    saved state in pLDCompInfoIn is
 *			    valid.<BR>
 *			    This arg is set to cFalse if either
 *			    we are starting with all new data
 *			    or the previous call returned because it
 *			    ran out of image data.
 *			    When this arg is true, startLine and numLines
 *			    should be set to the same value as they were
 *			    on the previous call.  The image should also
 *			    be the same.
 *  continueCompDataStream:: cTrue => The saved compression state
 *			    in pLDCompInfoIn is valid.  If
 *			    the previous call to this routine
 *			    filled the compressed data buffer
 *			    before the strip, tile or image was
 *			    complete, a new buffer should be
 *			    supplied.  This arg is set to 
 *			    cFalse when the block of compressed
 *			    data is unrelated to the block used
 *			    in the previous call to this routine
 *			    as recorded in pLDCompInfoIn.
 *			    SETTING THIS ARG TO cFalse CAUSES
 *			    DATA IN THE PIPELINE TO BE DISCARDED!
 *
 * Output:::
 *  pExhaustedImageData::   Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because specified
 *			    number of lines have been processed.
 *			    This may require more than one
 *			    call to this routine.
 *  pExhaustedCompData::    Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because it filled
 *			    the compressed data buffer.  This
 *			    routine may be called again with
 *			    a new buffer.
 *  pPercentDone::          Pointer to returned arg.  When
 *			    routine returns with *pExhaustedCompData
 *			    set, this arg specifies the rough
 *			    percentage of the image block that
 *			    has been compressed.  It is
 *			    intended to be used to estimate the
 *			    amount of compressed data needed
 *			    to complete the task.  The result
 *			    ranges from 0 to 100.
 *  pBytesWritten::         Number of compressed bytes actually
 *			    written to the compressed data
 *			    block.  Some bytes required to complete
 *			    a scan line may still be in the
 *			    pipeline and are not included in
 *			    this count.
 *
 * Notes:::
 *  This routine compresses a horizontal slice of a binary image using the
 *  Lossy Dither technique.  There are several completion conditions:
 *
 *	__error__	!= cSuccessful::
 *	*pExhaustedImageData == don't care::
 *	*pExhaustedCompData == don't care::
 *		This means something catastrophic happened.
 *
 *	__error__	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cFalse::
 *		This means we compressed the desired number of
 *		image lines without filling the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.  Some of the
 *		compressed data may still be in the
 *		pipeline and the saved compressed data state is valid.
 *		This routine may be called again with
 *		continueCompDataStream = cTrue and the new compressed
 *		data will be seamlessly concatenated with the old
 *		compressed data (on any bit boundary).  The
 *		client must supply a new value for pCompData and
 *		compDataSize.
 *
 *	__error__	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cTrue::
 *		This means we filled the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is unusual, but not impossible.  Some data may
 *		remain in the pipeline.
 *
 *	__error__	== cSuccessful::
 *	*pExhaustedImageData == cFalse::
 *	*pExhaustedCompData == cTrue::
 *		The routine filled the compressed data buffer before completing
 *		the specified number of lines.  One way this can
 *		happen is if all image data has been processed and some
 *		data remains in the pipeline.
 *		When this condition does occur, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.<BR> 
 *		The image may be moved around as a block after this
 *		condition has been returned, but the internal
 *		composition of the image (and startLine and numLines) 
 *		should not be changed.
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *<BR><BR>		
 *		The image data is always in PIXR order.
 *		
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_compressRowsLD (
    IPState    *pIPState,
    void       *pLDCompInfoIn,
    UInt8      *pImage,
    Int32	w,
    Int32	startLine,
    Int32	numLines,
    Int32	bpl,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesWritten)
{
    PROLOGUE;
    LDCompressionInfo	*pLDCompInfo;
    UInt8		*pCompDataEnd;	/* address 1 beyond the last
				     * byte of compressed data */
    UInt8		*pCurCompData;	/* current compressed data
				     * pointer */
    UInt8		*pCodingBytes;	/* Pointer to bytes in coding
				     * line. */
    UInt8		*pRefBytes; 	/* pointer to runs for reference
					 * line in array of runs. */
    Int32		*markerTable;	/* pointer to first word in the
					 * table that contains offsets of
					 * copy strings in the ref line. */
    UInt8		*pTemp;		/* utility ptr to bytes */
    UInt8		*pTempImage;	/* pts to bytes inside image */
    Int32		 index;		/* used to index coding and ref runs */
    Int32		 markerIndex;	/* indexes marker table */
    Int32		 remainingLines;
    

	/* check for null pointers */
    if ((pLDCompInfoIn == NULL) || (pCompData == NULL) ||
	(pImage == NULL) || (compDataBufSize == 0) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesWritten == NULL))
	return ia_invalidParm;

	/* Strip off the frame */
    pImage += (bpl << cLogFrameBits) + cFrameBytes;
    startLine -= cFrameBits;
    w -= 2*cFrameBits;
    if (w <= 0)
	NORMAL_RETURN;	/* there's no image here */

	/* load a useful copy of the opaque private data */
    pLDCompInfo = (LDCompressionInfo *) pLDCompInfoIn;
    markerTable = (Int32 *)((UInt8 *)pLDCompInfo + pLDCompInfo->markerOffset);

	/* Initialize the state from the saved state passed in *pLDCompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	pImage += startLine * bpl;
	pLDCompInfo->lineNumber = 0;
	pLDCompInfo->line1IsCoding = cFalse;
	pLDCompInfo->dataIndex = 0;
	pLDCompInfo->stringIndex = 0;
	pLDCompInfo->litIndex = 0;
	pLDCompInfo->markerIndex = 0;
	pLDCompInfo->remainingLength = 0;
	pLDCompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pLDCompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pLDCompInfo->lineNumber;
	pImage += (startLine + pLDCompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pLDCompInfo->compState = cLDCompNewLine;
	pLDCompInfo->accumBits = 0;
	pLDCompInfo->vacantBit = 0;
	pLDCompInfo->compInfoIsValid = cTrue;
    }
    else
    if (!pLDCompInfo->compInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Set the run pointers to the first byte in each line. */
    if (pLDCompInfo->line1IsCoding)
    {
	pCodingBytes = (UInt8 *)(pLDCompInfo) + pLDCompInfo->line1Offset;
	pRefBytes = (UInt8 *)(pLDCompInfo) + pLDCompInfo->line0Offset;
    }
    else
    {
	pCodingBytes = (UInt8 *)(pLDCompInfo) + pLDCompInfo->line0Offset;
	pRefBytes = (UInt8 *)(pLDCompInfo) + pLDCompInfo->line1Offset;
    }

	/* If we're not continuing from a previous image block,
	 * set the reference and coding lines to be "blank" by filling it with
	 * the equivalence class for white bytes.  We fill both byte arrays
	 * because later we'll create the list of "marker" points by
	 * comparing them before replacing the old coding bytes.
	 */
    if (!continueImageBlock)
    {
	for (index = pLDCompInfo->widthInBytes - 1;
	     index >= 0;
	     index--)
	{
	    pCodingBytes[index] = ldClass[0];	/* the class for white */
	    pRefBytes[index] = ldClass[0];	/* the class for white */
	}
    }

	/* Compress each scan line.
	 * This is done in several steps:
	 *   1. If the compression state does not equal cLDCompNewLine,
	 *	jump to step 3.
	 *   2. We did not quit in the middle of putting out this scan
	 *	line.  Use the ldClass table to classify image bytes and
	 *	put them into the coding line.  Initialize the reference
	 *	line data structures.
	 *   3. Call ip_compLDRow to put out the codes for the runs
	 *	of the coding line.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--, pImage += bpl)
    {
	if (pLDCompInfo->compState == cLDCompNewLine)
	{
		/* Fill in the marker table.  This contains the locations
		 * where the current reference line is different from
		 * the previous previous reference line (current coding
		 * line).  These will be used to encode the runs of
		 * similar bytes and different bytes.  First, make sure
		 * the dummies are in place so we don't run off the end.
		 */
	    index = pLDCompInfo->widthInBytes;
	    if (pLDCompInfo->line1IsCoding)
	    {
		pCodingBytes[index] = cLine1DummyVal;
		pCodingBytes[index+1] = 0;
		pRefBytes[index] = cLine0DummyVal;
		pRefBytes[index+1] = 0;
	    }
	    else
	    {
		pCodingBytes[index] = cLine0DummyVal;
		pCodingBytes[index+1] = 0;
		pRefBytes[index] = cLine1DummyVal;
		pRefBytes[index+1] = 0;
	    }

		/* remember the ends of the copy strings in the ref line */
	    for (markerIndex = 0, index = 0;
		 index < pLDCompInfo->widthInBytes;
		 )
	    {
		    /* Skip over the bytes where ref line == prev ref line */
		while (pCodingBytes[index] == pRefBytes[index])
		    index++;

		    /* Mark this first difference byte as a marker point */
		markerTable[markerIndex] = index - 1;
		markerIndex++;

		    /* skip over bytes where they're different */
		while (pCodingBytes[index] != pRefBytes[index])
		    index++;
	    }
		/* Make sure the last marker is off the edge of the image.
		 * Note that if the line ends in different bytes, we'll
		 * stop AFTER the last dummy byte, which is too far.
		 */
	    markerTable[markerIndex] = pLDCompInfo->widthInBytes;
	    markerTable[markerIndex + 1] = pLDCompInfo->widthInBytes;


		/* Fill the coding line with equivalence class numbers
		 * for the image bytes.  Also put the bytes in natural
		 * byte order inside the coding line.  They're easier
		 * to deal with as bytes this way.
		 */
	    for (pTemp = pCodingBytes + ((pLDCompInfo->widthInBytes - 1)/4) * 4,
		 pTempImage = pImage + ((pLDCompInfo->widthInBytes - 1)/4) * 4;
		 pTemp >= pCodingBytes;
		 pTemp -= 4, pTempImage -= 4)
	    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		pTemp[0] = ldClass[pTempImage[3]];
		pTemp[1] = ldClass[pTempImage[2]];
		pTemp[2] = ldClass[pTempImage[1]];
		pTemp[3] = ldClass[pTempImage[0]];
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		pTemp[0] = ldClass[pTempImage[0]];
		pTemp[1] = ldClass[pTempImage[1]];
		pTemp[2] = ldClass[pTempImage[2]];
		pTemp[3] = ldClass[pTempImage[3]];
#endif
	    }

		/* restore the dummies at the end of the coding line */
	    if (pLDCompInfo->line1IsCoding)
		pCodingBytes[pLDCompInfo->widthInBytes] = cLine1DummyVal;
	    else
		pCodingBytes[pLDCompInfo->widthInBytes] = cLine0DummyVal;
	    pCodingBytes[pLDCompInfo->widthInBytes + 1] = 0;


		/* Just to be extra safe, reset the various indices
		 * and lengths in the compression info so that we start
		 * compressing the very first byte and not some
		 * byte left over from some previous stoppage.
		 * I'm not really sure when this does anything
		 * that wasn't already done, but it seems tidy.
		 */
	    pLDCompInfo->dataIndex = 0;
	    pLDCompInfo->litIndex = -1;
	}

	    /* Ask the compression code to translate runs into the
	     * compressed data stream.  The width sent down is used
	     * to decide when we've coded the whole line.
	     */
	__error__ = ip_compLDRow (pIPState, pLDCompInfo, w,
			        &pCurCompData, pCompDataEnd);
	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;

	    /* We have now attempted to compress one more scan line,
	     * possibly not the last one.  Should we stop now?
	     * If we stopped in the middle of a line because the
	     * compressed data buffer was full and we had more stuff
	     * to write, or we finished the line but just happened
	     * to fill the compressed data buffer and there are
	     * more lines to go, stop now.
	     */
	if ((__error__ == ia_bufferExceeded) ||
	    ((pCurCompData == pCompDataEnd) && (remainingLines > 1)))
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * ((pLDCompInfo->lineNumber) * w)) /
			    (numLines * w);
	    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

		/* if we happened to stop on an image scan line boundary,
		 * increment the saved line number so we don't do it again.
		 * Also swap the coding and reference runs.
		 */
	    if (pLDCompInfo->compState == cLDCompNewLine)
	    {
		pLDCompInfo->lineNumber++;
		pLDCompInfo->line1IsCoding ^= cTrue;
	    }
	    __error__ = ia_successful;	/* this is the convention if we get this
					 * far.  It means we didn't encounter a
					 * real error, though we may have filled
					 * output buffer.  The caller detects
					 * that by looking at
					 * *pExhaustedCompData.
					 */
	    NORMAL_RETURN;
	}
	    /* increment the current line number in case we need
	     * to stop in the next line.  Also swap the coding
	     * and reference runs.
	     */
	pLDCompInfo->lineNumber++;
	pLDCompInfo->line1IsCoding ^= cTrue;
	pTemp = pCodingBytes;
	pCodingBytes = pRefBytes;
	pRefBytes = pTemp;
    }

	/* We've done all the scan lines.  So, just
	 * say things were successful and return.
	 */
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;

	/* Return the number of bytes written, including any
	 * bytes flushed.
	 */
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

    NORMAL_RETURN;

}


/*

 $Log: i_ldc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:48:41  ddavies
 i_ldc.c -> /product/ipcore/ipshared/src/RCS/i_ldc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1997/03/17 19:46:28  ddavies
 Initial revision


 */

