/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_ldd.c: High level code used in Lossy Dither decompression.
 *
 *   	Int32	i_decompressRowsLD()
 *			Decompresses a block of scan lines using
 *			the Lossy Dither algorithm.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#define _NEED_LDCLASS_
#define _NEED_LDDITHER_
#include "ldc.prv"

    /*  prototyes */
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_ldd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* There is a big comment all about the low level mechanics of our
 * implementation of Lossy Dither decompression in mildd.c.  This comment will
 * deal with software engineering and interfaces on the next higher
 * level.  The basic style of interfaces is shared by the CCITT
 * compression routines and the PackBits compression routine.
 * 
 * Lossy Dither encoding uses strings on one scan line to encode the
 * data on the next scan line.  The "coding" line is the line being
 * encoded.  The "reference" line is the one above the coding line.
 * There are two byte arrays in the encoder state block.  We ping-pong
 * between them; filling one and using it for the coding line, then
 * filling the other and using the first for the reference line,
 * then filling the first and using the second for the reference line
 * and so on.  The line1IsCoding boolean in the compression data
 * structure indicates whether the second byte array is the coding
 * line (cTrue) or the reference line (cFalse).
 *
 * Swell.  So how does the decoding work?  It's pretty much the inverse
 * of the encoding routine.  For each line, we call ip_decompLDRow
 * to get a new line of equivalence class numbers using one of byte
 * arrays as the reference line and the other as the new coding line.
 * Then we shove the new coding line bytes found into the bitmap.
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which the lower level procedure will stop.  It
 * might actually finish decoding one complete scan line.  When this
 * happens, update the pointer to the compressed data that was passed
 * in, store the internal decoding state into the state structure
 * and return happily.
 *
 * The other reason for returning is that we run out of compressed data.
 * The ip_decompLDRow routine detects this by comparing
 * its pCurCompData to a pointer to the end of the buffer just before
 * reading compressed data. Since we want to pretend that we have a
 * callback to fill the buffer, we need a restart location next to
 * every check.  There are **** checks when reading codes.
 *
 *			*	*	*
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * and the continueImageBlock, continueCompDataStream and flushPipeline args
 * discussed above.  The caller also supplies pointers to words used to
 * return the exhaustedImageData, exhaustedCompData, bytesWritten and
 * percentDone args.  The bytesRead arg returns the number of compressed
 * bytes read so far.  If flushPipeline is not set, some of the compressed
 * bytes may have been read, but not yet decoded.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.  One possibly catastrophe is
 *	that flushPipeline was set but insufficient compressed data
 *	was supplied.  In this case, we return ia_bufferExceeded.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	We have processed all the image data requested.  The compression
 *	buffer is not empty.  Some of the compressed data bytes are
 *	still in the pipeline regardless of the state of flushPipeline.
 *	The number of bytes actually read from the buffer
 *	(including any bytes still in the pipeline) are returned.
 *	When the decompression routine is called next, the client must
 *	pass in a pointer to the byte holding the next compressed byte.
 *	This may be just beyond the end of the current compressed data,
 *	at the beginning of the same buffer, or in another buffer entirely.
 *	The size of the compressed data buffer must also be (re)calculated
 *	by the client.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to read the last
 *	byte of the compressed data buffer at the same time the compressed
 *	data came to an end.  As before, if flushPipeline was false, there
 *	may be more compressed data in the pipeline.  Normally, this
 *	case occurs when exactly the right amount of compressed data was
 *	supplied and flushPipeline was true.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of compressed data before reading the last
 *	byte of compressed data requested.  Some of the data read
 *	is in the pipeline.  This can only happen if flushPipeline was
 *	false.  If flushPipeline is true, we keep going until we either
 *	finish all the scan lines or get an ia_bufferExceeded.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */

/* PUBLIC */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress rows of a binary image using the Lossy Dither technique.
 *
 * LongDescription:::
 *  Decompress some rows of an image using the LD algorithm.  We will
 *  stop when either an error is encountered, we run out of image rows
 *  or we empty the compressed data buffer.
 *
 * Input:::
 *  pLDDecompInfoIn::       Pointer to private LD decompression state block
 *  pD::		    Pointer to upper left corner of frame on dest image
 *  startLine::             First image line to be decompressed
 *			    relative to the top of the image, not frame.
 *  numLines::              Number of lines to be compressed.
 *  bpl::		    Dest image bytes/line
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    read.  If several calls
 *			    to this routine are needed to read
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *  continueImageBlock::    cTrue => we are decompressing
 *			    into the same slice of image as was
 *			    specified in the previous call and the
 *			    saved state in pLDDecompInfoIn is
 *			    valid.<BR>
 *			    This arg is set to cFalse if either
 *			    we are starting with all new data
 *			    or the previous call returned because it
 *			    ran out of image data.
 *			    When this arg is true, startLine and numLines
 *			    should be set to the same value as they were
 *			    on the previous call.  The image should also
 *			    be the same.
 *  continueCompDataStream:: cTrue => The saved compression state
 *			    in pLDDecompInfoIn is valid.  If
 *			    the previous call to this routine
 *			    emptied the compressed data buffer
 *			    before the strip, tile or image was
 *			    complete, a new buffer should be
 *			    supplied.  This arg is set to 
 *			    cFalse when the block of compressed
 *			    data is unrelated to the block used
 *			    in the previous call to this routine
 *			    as recorded in pLDDecompInfoIn.
 *			    SETTING THIS ARG TO cFalse CAUSES
 *			    DATA IN THE PIPELINE TO BE DISCARDED!
 *  flushPipeline::         If this boolean is set, this routine
 *			    will not stop when all data has been
 *			    loaded into its pipeline.  It will attempt
 *			    to continue until all the scan lines have
 *			    been decompressed.  The client should not
 *			    set flushPipeline if compressed
 *			    data from subsequent scan lines is
 *			    to be seamlessly concatenated to
 *			    the current compressed data stream.
 *			    The client *should* set flushPipeline
 *			    when supplying exactly the correct amount
 *			    of compressed data for the scan lines
 *			    requested.
 *			    When flushPipeline is set and the
 *			    decompression task suceeds in
 *			    processing all image lines, all
 *			    saved state is marked invalid.
 *			    Subsequent calls with either continueImageBlock or
 *			    continueCompDataStream will return
 *			    cErrInvalidParm.
 *
 * Output:::
 *  pExhaustedImageData::   Pointer to returned arg.  When the
 *			    arg is returned with value cTrue,
 *			    routine stopped because specified
 *			    number of lines have been processed.
 *			    This may require more than one
 *			    call to this routine.
 *  pExhaustedCompData::    Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because it emptied
 *			    the compressed data buffer.  This
 *			    routine may be called again with
 *			    a new buffer.  This arg will not be 
 *			    set if flushPipeline is set and we 
 *			    failed to decompress the specified 
 *			    number of scan lines (routine returns
 *			    cErrBufferExceeded instead).
 *  pPercentDone::          Pointer to returned arg.  When
 *			    routine returns with *pExhaustedCompData
 *			    set, this arg specifies the rough
 *			    percentage of the image block that
 *			    has been decompressed.  It is
 *			    intended to be used to estimate the
 *			    amount of compressed data needed
 *			    to complete the task.  The result
 *			    ranges from 0 to 100.
 *  pBytesRead::            Number of compressed bytes actually
 *			    read from the compressed data
 *			    block.  If flushPipeline is false,
 *			    some bytes may still be in the
 *			    pipeline.  These are included in
 *			    this count.
 *			    The client may safely dispose of
 *			    these bytes.
 *
 * Returns:::
 *  Normally, this routine thinks that emptying the compressed data
 *  buffer is ok.  However, if flushPipeline is set to cTrue and we empty
 *  the compressed data buffer, the routine returns cErrBufferExceeded.
 *
 * Notes:::
 *  This routine decompresses a horizontal slice of a binary image using the
 *  Lossy Dither technique.  There are several completion conditions:
 *
 *	__error__	!= cSuccessful::
 *	*pExhaustedImageData == don't care::
 *	*pExhaustedCompData == don't care::
 *		This means something catastrophic happened.  One possible
 *		catastrophe (cErrBufferExceeded) happens if the client
 *		set flushPipeline and failed to supply enough compressed
 *		data.
 *
 *	__error__	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cFalse::
 *		This means we decompressed the desired number of
 *		image lines without emptying the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.
 *		Some of the compressed data is still in the
 *		pipeline and the saved compressed data state is valid.
 *		This routine may be called again with
 *		continueCompDataStream = cTrue and the new compressed
 *		data will be seamlessly concatenated with the old
 *		compressed data (on any bit boundary).  The
 *		client must supply a new value for pCompData and
 *		compDataSize.
 *
 *	__error__	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cTrue::
 *		This means we emptied the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is normal when exactly the correct number of compressed
 *		data bytes were supplied.  It is very unusual for this
 *		condition to occur if flushPipeline was not set.
 *		It is not possible to continue from this state.
 *
 *	__error__	== cSuccessful::
 *	*pExhaustedImageData == cFalse::
 *	*pExhaustedCompData == cTrue::
 *		The routine emptied the compressed data buffer before completing
 *		the specified number of lines.  This can only happen if
 *		flushPipeline was not set, so some data remains in the pipeline.
 *		When this condition occurs, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.<BR>
 *		Note that the first byte of the new compressed
 *		data must be the one immediately following the last
 *		byte of old compressed data.<BR>
 *		The image may be moved around as a block after this
 *		condition has been returned, but the internal
 *		composition of the image (and startLine and numLines) 
 *		should not be changed.<BR>
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *<BR><BR>		
 *		The image data is always in PIXR order.
 *		
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_decompressRowsLD (
    IPState    *pIPState,
    void       *pLDDecompInfoIn,
    UInt8      *pD,
    Int32	startLine,
    Int32	numLines,
    Int32	bpl,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Bool	flushPipeline,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesRead)
{
    PROLOGUE;
    LDDecompressionInfo	*pLDDecompInfo;
    UInt8		*pCompDataEnd;	/* address 1 beyond the last
					 * byte of compressed data */
    UInt8		*pCurCompData;	/* current compressed data
					 * pointer */
    UInt8		*pCodingBytes;	/* Pointer to bytes in coding
					 * line. */
    UInt8		*pRefBytes; 	/* pointer to runs for reference
					 * line in array of runs. */
    Int32		*markerTable;	/* pointer to first word in the
					 * table that contains offsets of
					 * copy strings in the ref line. */
    Int32		 markerIndex;	/* indexes marker table */
    UInt8		*pTemp;		/* utility ptr to bytes */
    UInt8		*pTempImage;	/* pts to bytes inside image */
    Int32		 index;		/* used to index coding and ref runs */
    Int32		 remainingLines;
    Int32		 widthInBytes;	/* width of unframed image */
    Int32		 curLine;	/* current line number mod 8 */
    UInt8		*pDOrig;	/* used in ip_clearRightFrame */
    

	/* check for null pointers */
    if ((pLDDecompInfoIn == NULL) || (pCompData == NULL) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesRead == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pLDDecompInfo = (LDDecompressionInfo *) pLDDecompInfoIn;
    markerTable = (Int32 *)
			((UInt8 *)pLDDecompInfo + pLDDecompInfo->markerOffset);

	/* Strip off the frame */
    pD += (bpl << cLogFrameBits) + cFrameBytes;
    pDOrig = pD;			/* used in clearRightFrame */
    startLine -= cFrameBits;
    widthInBytes = (pLDDecompInfo->widthInPixels + 7) / 8;

	/* Initialize the state from the saved state passed in *pLDDecompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	pD += startLine * bpl;
	curLine = startLine % 8;
	pLDDecompInfo->lineNumber = 0;
	pLDDecompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pLDDecompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pLDDecompInfo->lineNumber;
	pD += (startLine + pLDDecompInfo->lineNumber) * bpl;
	curLine = (startLine + pLDDecompInfo->lineNumber) % 8;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pLDDecompInfo->decompState = cLDDecompNewLine;
	pLDDecompInfo->line1IsCoding = cFalse;
	pLDDecompInfo->dataIndex = 0;
	pLDDecompInfo->markerIndex = 0;
	pLDDecompInfo->accumBits = 0;
	pLDDecompInfo->vacantBit = 0;
	pLDDecompInfo->codeLen = 0;
	pLDDecompInfo->remainingLength = 0;
	pLDDecompInfo->decompInfoIsValid = cTrue;
    }
    else
    if (!pLDDecompInfo->decompInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Set the run pointers to the second runs.  The first is a dummy. */
    if (pLDDecompInfo->line1IsCoding)
    {
	pCodingBytes = (UInt8 *)(pLDDecompInfo) + pLDDecompInfo->line1Offset;
	pRefBytes = (UInt8 *)(pLDDecompInfo) + pLDDecompInfo->line0Offset;
    }
    else
    {
	pCodingBytes = (UInt8 *)(pLDDecompInfo) + pLDDecompInfo->line0Offset;
	pRefBytes = (UInt8 *)(pLDDecompInfo) + pLDDecompInfo->line1Offset;
    }

	/* If we're not continuing from a previous image block,
	 * set the reference and coding lines to be "blank" by filling it with
	 * the equivalence class for white bytes.  We fill both byte arrays
	 * because later we'll create the list of "marker" points by
	 * comparing them before replacing the old coding bytes.
	 */
    if (!continueImageBlock)
    {
	for (index = widthInBytes - 1;
	     index >= 0;
	     index--)
	{
	    pCodingBytes[index] = ldClass[0];	/* the class for white */
	    pRefBytes[index] = ldClass[0];	/* the class for white */
	}
    }

	/* Decompress each scan line.
	 * This is done in several steps:
	 *   1. If we are not restarting after a failed decode,
	 *	refill the array of markers used by ip_decompLDRow.
	 *   2. Call ip_decompLDRow to fill codingBytes from
	 *	compressed tokens.  We both start here and
	 *	restart here if we stopped in the previous pass.
	 *   3. Dither the symbol IDs back into the image.
	 *	We cannot abort in the middle of this step.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--, pD += bpl, curLine = (curLine + 1) % 8)
    {
	if (pLDDecompInfo->decompState == cLDDecompNewLine)
	{
		/* Fill in the marker table.  This contains the locations
		 * where the current reference line is different from
		 * the previous previous reference line (current coding
		 * line).  These will be used to encode the runs of
		 * similar bytes and different bytes.  First, make sure
		 * the dummies are in place so we don't run off the end.
		 */
	    index = widthInBytes;
	    if (pLDDecompInfo->line1IsCoding)
	    {
		pCodingBytes[index] = cLine1DummyVal;
		pCodingBytes[index+1] = 0;
		pRefBytes[index] = cLine0DummyVal;
		pRefBytes[index+1] = 0;
	    }
	    else
	    {
		pCodingBytes[index] = cLine0DummyVal;
		pCodingBytes[index+1] = 0;
		pRefBytes[index] = cLine1DummyVal;
		pRefBytes[index+1] = 0;
	    }

		/* Put the locations where the ref line's copy strings end
		 * inot the marker table.
		 */
	    for (markerIndex = 0, index = 0;
		 index < widthInBytes;
		 )
	    {
		    /* Skip over the bytes where ref line == prev ref line */
		while (pCodingBytes[index] == pRefBytes[index])
		    index++;

		    /* Mark this first difference byte as a marker point */
		markerTable[markerIndex] = index - 1;
		markerIndex++;

		    /* skip over bytes where they're different */
		while (pCodingBytes[index] != pRefBytes[index])
		    index++;
	    }
		/* Make sure the last marker is off the edge of the image.
		 * Note that if the line ends in different bytes, we'll
		 * stop AFTER the last dummy byte, which is too far.
		 */
	    markerTable[markerIndex] = widthInBytes;
	    markerTable[markerIndex + 1] = widthInBytes;

		/* Just to be extra safe, reset the various indices
		 * and lengths in the compression info so that we start
		 * compressing the very first byte and not some
		 * byte left over from some previous stoppage.
		 * I'm not really sure when this does anything
		 * that wasn't already done, but it seems tidy.
		 */
	    pLDDecompInfo->dataIndex = 0;
	}

	    /* decode the bytes for the next line. */
	__error__ = ip_decompLDRow (pIPState, pLDDecompInfo,
				  widthInBytes, flushPipeline,
			          &pCurCompData, pCompDataEnd);

	    /* If there was a disaster, return now */
	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;

	    /* If we emptied the buffer after the client promised
	     * that this was the very last buffer we were going
	     * to get, go whine about it.
	     */
	if ((__error__ == ia_bufferExceeded) && (flushPipeline))
	{
	    pLDDecompInfo->imageInfoIsValid = cFalse;
	    pLDDecompInfo->decompInfoIsValid = cFalse;
	    NORMAL_RETURN;
	}

	    /* If we didn't get to the end of a scan line, return now
	     * instead of redithering all the bytes to the bitmap.  We
	     * don't redither the bytes because we don't have all of them.
	     * Since flushPipeline was not set, this is all ok.
	     */
	if (__error__ == ia_bufferExceeded)
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * pLDDecompInfo->lineNumber) /
			    (numLines);
	    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

	    NORMAL_RETURN;
	}

	    /* Great, we got a set of equivalence class IDs.  Let's
	     * redither them back into the image.
	     */
	for (pTemp = pCodingBytes + ((widthInBytes - 1)/4) * 4,
		pTempImage = pD + ((widthInBytes - 1)/4) * 4;
	     pTemp >= pCodingBytes;
	     pTemp -= 4, pTempImage -= 4)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		pTempImage[3] = ldClassDither[pTemp[0]][curLine];
		pTempImage[2] = ldClassDither[pTemp[1]][curLine];
		pTempImage[1] = ldClassDither[pTemp[2]][curLine];
		pTempImage[0] = ldClassDither[pTemp[3]][curLine];
#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		pTempImage[0] = ldClassDither[pTemp[0]][curLine];
		pTempImage[1] = ldClassDither[pTemp[1]][curLine];
		pTempImage[2] = ldClassDither[pTemp[2]][curLine];
		pTempImage[3] = ldClassDither[pTemp[3]][curLine];
#endif
	}

	    /* increment the current line number just in case we need
	     * to stop in the next line.  Also swap the coding
	     * and reference lines.
	     */
	pLDDecompInfo->lineNumber++;
	pLDDecompInfo->line1IsCoding ^= cTrue;
	pTemp = pCodingBytes;
	pCodingBytes = pRefBytes;
	pRefBytes = pTemp;
    }

	/* We may have dithered into the right frame, so clear it. */
    ip_clearRightFrame (pIPState, pDOrig,
			pLDDecompInfo->widthInPixels, numLines, bpl, 1);

	/* Great! we completed the right number of scan lines.
	 * Say that things worked and return.
	 */
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;
    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;


    NORMAL_RETURN;
}


/*

 $Log: i_ldd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:11  jlin
 Imported sources

 Revision 1.3  1998/01/13 23:48:43  ddavies
 i_ldd.c -> /product/ipcore/ipshared/src/RCS/i_ldd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1997/03/18 20:10:53  ddavies
 ddavies on Tue Mar 18 12:10:35 PST 1997

 i_ldd.c -> /product/ipcore/ipshared/src/RCS/i_ldd.c,v

 Remove combineResults parameter from i_decompressLDRows.  The method in which
 decompressed lossy dither images are recombined with the original image is
 more complex than the combineResults flag can express.

 Revision 1.1  1997/03/17 19:47:25  ddavies
 Initial revision


 */

