/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_ldut.c: High level utilities used in Lossy Dither compression.
 *
 *   	Int32	i_getSizeOfLDCompressionInfo()
 *			Returns size of opaque data structure used in
 *			LD compression.
 *
 *   	Int32	i_getSizeOfLDDecompressionInfo()
 *			Returns size of opaque data structure used in
 *			LD compression.
 *
 *	Int32	i_initializeLDCompressionInfo()
 *			This routine initializes the opaque data structure
 *			used in LD compression.
 *
 *	Int32	i_initializeLDDecompressionInfo()
 *			This routine initializes the opaque data structure
 *			used in LD decompression.
 *
 *	Int32	i_flushLD()
 *			This routine flushes the pipeline.
 *
 *	Int32	i_writeLDEOL()
 *			This routine writes and EndOfLine code.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "ldc.prv"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* PUBLIC */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return size of Lossy Dither Compression block.
 *
 * LongDescription:::
 *  This procedure returns the size of the memory block that will
 *  be needed to hold the state information used to perform Lossy
 *  Dither compression.
 *
 * Input:::
 *  w::			    width of binary image including frame
 *
 * Output:::
 *  pSize::		    pointer to UInt32 that will contain the size.
 *			    of the buffer required to hold the symbol
 *			    information in bytes
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
i_getSizeOfLDCompressionInfo (
    IPState    *pIPState,
    Int32	w,
    UInt32     *pSize)
{
    PROLOGUE;
    Int32	widthInBytes;

	/* remove the frame */
    w -= 2 * cFrameBits;

	/* The width of the image in bytes rounded up to the next
	 * multiple of 4 is the number of bytes to hold w bits
	 * plus one for the dummy all rounded up to the next multiple
	 * of 4.
	 */
    widthInBytes = ((((w+7)/8) + 2 + 3)/4) * 4;

	/* Return size of LD compression structure
	 * plus three arrays.  Two hold the equivalence class numbers,
	 * one byte per image byte, and one holds the offset to the
	 * next occurance of a particular equivalence class number,
	 * one Int32 per image byte.  There's also one dummy byte at
	 * the end of the two equivalence class arrays.  We add 8
	 * just to make sure there's plenty of room.
	 */
    *pSize = sizeof(LDCompressionInfo) +
	     2 * ((widthInBytes + 8) * sizeof(UInt8)) +
	     ((widthInBytes/2 + 1) * sizeof(Int32));

    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return size of Lossy Dither Decompression block.
 *
 * LongDescription:::
 *  This procedure returns the size of the memory block that will
 *  be needed to hold the state information used to perform Lossy
 *  Dither decompression.
 *
 * Input:::
 *  w::			    width of binary image including frame
 *
 * Output:::
 *  pSize::		    pointer to UInt32 that will contain the size.
 *			    of the buffer required to hold the symbol
 *			    information in bytes
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
i_getSizeOfLDDecompressionInfo (
    IPState    *pIPState,
    Int32	w,
    UInt32     *pSize)
{
    PROLOGUE;
    Int32	 widthInBytes;

	/* remove the frame */
    w -= 2 * cFrameBits;

	/* The width of the image in bytes rounded up to the next
	 * multiple of 4 is the number of bytes to hold w bits
	 * plus one for the dummy all rounded up to the next multiple
	 * of 4.
	 */
    widthInBytes = ((((w+7)/8) + 2 + 3)/4) * 4;

	/* Return size of LD decompression structure
	 * plus three arrays.  They hold the equivalence class numbers,
	 * one byte per image byte.  There are also two dummy bytes at
	 * the ends of the two equivalence class arrays.  We add 8
	 * just to make sure there's plenty of room.
	 */
    *pSize = sizeof(LDDecompressionInfo) +
	     2 * ((widthInBytes + 8) * sizeof(UInt8)) +
	     ((widthInBytes/2 + 1) * sizeof(Int32));

    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize Lossy Dither Compression block.
 *
 * LongDescription:::
 *  This procedure initializes the private data block used by Lossy
 *  Dither compression.
 *
 * Input:::
 *  pLDCompInfoIn::	    pointer to LD compression info data block
 *  w::			    width of binary image including frame
 *  compInfoBufSize::	    the number of bytes in the compression info block.
 *
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
i_initializeLDCompressionInfo (
    IPState    *pIPState,
    void       *pLDCompInfoIn,
    Int32	w,
    UInt32	compInfoBufSize)
{
    PROLOGUE;
    LDCompressionInfo	*pLDCompInfo;
    UInt32		 realCompInfoBufSize;
    Int32		 arrayWidth;
    

	/* initialize a useful pointer */
    pLDCompInfo = (LDCompressionInfo *)pLDCompInfoIn;

	/* Get the size the block should be */
    __error__ = i_getSizeOfLDCompressionInfo (pIPState, w, &realCompInfoBufSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* See if any of the parameters were illegal */
    if ((pLDCompInfo == NULL) ||
	(realCompInfoBufSize > compInfoBufSize))
	return ia_invalidParm;

	/* Remove frame from the width */
    w -= 2*cFrameBits;

	/* calculate width of arrays.  These include one entry
	 * for each byte in an input line plus two dummy bytes
	 * all rounded up to the next multiple of 4.
	 */
    arrayWidth = (((w + 7) / 8 + 2 + 3)/4) * 4;

	/* Initialize all the rest of the default stuff */
    pLDCompInfo->widthInBytes = (w + 7) / 8;
    pLDCompInfo->imageInfoIsValid = 0;
    pLDCompInfo->lineNumber = 0;
    pLDCompInfo->line0Offset = sizeof(LDCompressionInfo);
    pLDCompInfo->line1Offset = pLDCompInfo->line0Offset + arrayWidth;
    pLDCompInfo->line1IsCoding = cFalse;
    pLDCompInfo->markerOffset = pLDCompInfo->line1Offset + arrayWidth;
    pLDCompInfo->dataIndex = 0;
    pLDCompInfo->stringIndex = 0;
    pLDCompInfo->litIndex = 0;
    pLDCompInfo->markerIndex = 0;
    pLDCompInfo->remainingLength = 0;
    pLDCompInfo->compInfoIsValid = 0;
    pLDCompInfo->compState = cLDCompNewLine;
    pLDCompInfo->accumBits = 0;
    pLDCompInfo->vacantBit = 0;

	/* Now we initialize a lot of stuff that really should be
	 * set before it's ever used, but setting it to a known
	 * value should reduce the chance of some uninitialized
	 * entry causing different behavior on different platforms.
	 */

	/* Initialize the reference run data structures.
	 */
    memset ((void *)((UInt8 *)pLDCompInfo + pLDCompInfo->markerOffset), 0,
	    (arrayWidth/2 + 1) * sizeof(Int32));

	/* Initialize the coding and reference byte arrays just to
	 * be predictable.
	 */
    memset ((void *)((UInt8 *)pLDCompInfo + pLDCompInfo->line0Offset), 0,
	    arrayWidth * sizeof(UInt8));
    memset ((void *)((UInt8 *)pLDCompInfo + pLDCompInfo->line1Offset), 0,
	    arrayWidth * sizeof(UInt8));

    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize Lossy Dither Decompression block.
 *
 * LongDescription:::
 *  This procedure initializes the private data block used by Lossy
 *  Dither decompression.
 *
 * Input:::
 *  pLDDecompInfoIn::	    pointer to LD decompression info data block
 *  w::			    width of binary image including frame
 *  decompInfoBufSize::	    the number of bytes in the decompression info block.
 *
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
i_initializeLDDecompressionInfo (
    IPState    *pIPState,
    void       *pLDDecompInfoIn,
    Int32	w,
    UInt32	decompInfoBufSize)
{
    PROLOGUE;
    LDDecompressionInfo *pLDDecompInfo;
    UInt32		 realDecompInfoBufSize;
    Int32		 arrayWidth;
    

	/* initialize a useful pointer */
    pLDDecompInfo = (LDDecompressionInfo *)pLDDecompInfoIn;

	/* Get the size the block should be */
    __error__ = i_getSizeOfLDDecompressionInfo (pIPState, w,
					      &realDecompInfoBufSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* See if any of the parameters were illegal */
    if ((pLDDecompInfo == NULL) ||
	(realDecompInfoBufSize > decompInfoBufSize))
	return ia_invalidParm;

	/* Remove frame from the width */
    w -= 2*cFrameBits;

	/* calculate width of arrays.  These include one entry
	 * for each byte in an input line plus two dummy bytes
	 * all rounded up to the next multiple of 4.
	 */
    arrayWidth = (((w + 7) / 8 + 2 + 3)/4) * 4;

	/* Initialize all the rest of the default stuff */
    pLDDecompInfo->widthInPixels = w;
    pLDDecompInfo->imageInfoIsValid = 0;
    pLDDecompInfo->lineNumber = 0;
    pLDDecompInfo->line0Offset = sizeof(LDDecompressionInfo);
    pLDDecompInfo->line1Offset = pLDDecompInfo->line0Offset + arrayWidth;
    pLDDecompInfo->markerOffset = pLDDecompInfo->line1Offset + arrayWidth;
    pLDDecompInfo->line1IsCoding = cFalse;
    pLDDecompInfo->dataIndex = 0;
    pLDDecompInfo->markerIndex = 0;
    pLDDecompInfo->codeLen = 0;
    pLDDecompInfo->remainingLength = 0;
    pLDDecompInfo->decompInfoIsValid = 0;
    pLDDecompInfo->decompState = cLDDecompNewLine;
    pLDDecompInfo->accumBits = 0;
    pLDDecompInfo->vacantBit = 0;


	/* Now we initialize a lot of stuff that really should be
	 * set before it's ever used, but setting it to a known
	 * value should reduce the chance of some uninitialized
	 * entry causing different behavior on different platforms.
	 */

	/* Initialize the reference run data structures.
	 */
    memset ((void *)((UInt8 *)pLDDecompInfo + pLDDecompInfo->markerOffset), 0,
	    (arrayWidth/2 + 1) * sizeof(Int32));

	/* Initialize the coding and reference byte arrays just to
	 * be predictable.
	 */
    memset ((void *)((UInt8 *)pLDDecompInfo + pLDDecompInfo->line0Offset), 0,
	    arrayWidth * sizeof(UInt8));
    memset ((void *)((UInt8 *)pLDDecompInfo + pLDDecompInfo->line1Offset), 0,
	    arrayWidth * sizeof(UInt8));

    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Flush Lossy Dither compression pipeline to output data stream.
 *
 * LongDescription:::
 *  This procedure flushes the compression pipeline to the output
 *  data stream.  It should be called at the end of each strip.
 *
 * Input:::
 *  pLDCompInfoIn::	    pointer to LD compression info data block
 *  pCompData::		    Pointer to first vacant byte
 *			    of buffer into which compressed
 *			    data will be placed.
 *  compDataBufSize::	    the number of vacant bytes remaining in the
 *			    compressed data buffer.
 *
 * Output:::
 *  pBytesWritten::	    the number of bytes written to the compressed
 *			    data block.
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid,
 *  ia_bufferExceeded => unable to complete writing since compressed data
 *  block filled up.
 *
 * Notes:::
 *  This routine should only be called at the end of strips, tiles or images.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
i_flushLD (
    IPState    *pIPState,
    void       *pLDCompInfoIn,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Int32      *pBytesWritten)
{
    PROLOGUE;
    LDCompressionInfo	*pLDCompInfo;
    UInt8		*pCompDataEnd;	/* address 1 beyond the last
					 * byte of compressed data */
    UInt8		*pCurCompData;	/* current compressed data
					 * pointer */
    

	/* check for null pointers */
    if ((pLDCompInfoIn == NULL) || (pCompData == NULL) ||
	(pBytesWritten == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pLDCompInfo = (LDCompressionInfo *) pLDCompInfoIn;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Call the routine that does the flushing.
	 */
    __error__ = ip_flushLD (pIPState, (LDCompressionInfo *)pLDCompInfoIn,
			  &pCurCompData, pCompDataEnd);

	/* Record the number of bytes written and the return code.
	 * If __error__ == ia_bufferExceeded, the client should
	 * supply us with more data and try again.
	 */
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;
    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write EndOfFacsimile block code at end of Lossy Dither block.
 *
 * LongDescription:::
 *  This procedure writes two End of Line codes back to back, which
 *  form one EndOfFacsimile block code, at the end of a compressed
 *  Lossy Dither block.
 *
 * Input:::
 *  pLDCompInfoIn::	    pointer to LD compression info data block
 *  pCompData::		    Pointer to first vacant byte
 *			    of buffer into which compressed
 *			    data will be placed.
 *  compDataBufSize::	    the number of vacant bytes remaining in the
 *			    compressed data buffer.
 *
 * Output:::
 *  pBytesWritten::	    the number of bytes written to the compressed
 *			    data block.
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid,
 *  ia_bufferExceeded => unable to complete writing since compressed data
 *  block filled up.
 *
 * Notes:::
 *  This routine should only be called at the end of strips, tiles or images.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
i_writeLDEOL (
    IPState    *pIPState,
    void       *pLDCompInfoIn,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Int32      *pBytesWritten)
{
    PROLOGUE;
    LDCompressionInfo	*pLDCompInfo;
    UInt8		*pCompDataEnd;	/* address 1 beyond the last
					 * byte of compressed data */
    UInt8		*pCurCompData;	/* current compressed data
					 * pointer */
    

	/* check for null pointers */
    if ((pLDCompInfoIn == NULL) || (pCompData == NULL) ||
	(pBytesWritten == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pLDCompInfo = (LDCompressionInfo *) pLDCompInfoIn;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Call the routine that does the flushing.
	 */
    __error__ = ip_writeLDEOL (pIPState, (LDCompressionInfo *)pLDCompInfoIn,
			     &pCurCompData, pCompDataEnd);

	/* Record the number of bytes written and the return code.
	 * If __error__ == ia_bufferExceeded, the client should
	 * supply us with more data and try again.
	 */
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;
    NORMAL_RETURN;
}


/*

 $Log: i_ldut.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:48:44  ddavies
 i_ldut.c -> /product/ipcore/ipshared/src/RCS/i_ldut.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1997/03/17 19:51:12  ddavies
 Initial revision


 */

