/*****************************************************************
 *  Copyright (c) 1996 Xerox Corporation.  All rights reserved.  *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_matvec.c: matrix / vector multiply routines
 *
 *                  Int32	i_matvec_mult_8BA()
 *
 */


#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"

/* prototypes */
#include "fxpmath.pub"
#include "fxpmath.prv"
#include "shrpixr.pub"	/* needed for shrpixr.prv */
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_matvec.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");


/* These routines were written by Rich Smith at XIS.  There are a
 * few changes in format and type names to adapt them to ipshared.
 * These routines were imported into ipshared for the express purpose
 * of porting them to MMX.  The C version of the routine is used on
 * all non-MMX (and non-long-operation) machines and when the slow
 * math libraries are linked.
 */


/*************************************  matvec_mult_8BA   */
/* Multiply ax=b (a=matrix, b,x=vectors)                  */
/* i.e., b[i] = sum over j of a[i][j]*x[j]                */
/* Code for 8 Byte Aligned input, output, and matrix rows */
/* This will be replaced by an MMX routine                */
/**********************************************************/

/* PUBLIC */
/*********************************************************************
 *  i_matvec_mult_8BA()  
 *		This routine multiplies a matrix by a vector, producing
 *		another vector.  The matrix of weights is composed of
 *		8 bit signed integers.  The matrix of weights is composed
 *		of 16 bit (well, really 9 bits of useful stuff) integers
 *		that have been sign extended to 32 bits.  The outputs
 *		are signed 32-bit integers.  Before the multiplication,
 *		all numbers are sign extended to be the same length.
 *		In the MMX implemenation, the 32-bit inputs are packed
 *		down to 16 bits before the multplication, so don't
 *		use this routine to multiple a matrix by big ( > 64K )
 *		integers.
 *
 *		To accomodate the alignment needs of machines with
 *		long memory accesses, each row of the weight matrix
 *		has been aligned on a 8-byte boundary.  Similarly,
 *		the input and output vectors are aligned on 8-byte
 *		boundaries.
 *
 *		OutVector(r) = Matrix(r,0) * InVector(0) +
 *			       Matrix(r,1) * InVector(1) +
 *				...
 *			       Matrix(r,numCols-1) * InVector(numCols-1)
 *
 *	Input:
 *		pMatrix =	pointer to upper left corner of weight matrix
 *		pInVector =	ptr to first element of input vector.
 *		pOutVector =	ptr to first element of output vector.
 *		numRows =	number of rows in weight matrix.  This
 *				is also the number of outputs to produce.
 *		numCols = 	number of columns in the weight matrix.
 *				This is also the length of the input vector
 *				in units of integers.
 *
 *	Return code:
 *		ia_successful	good completion
 *********************************************************************/
Int32 CDECL
i_matvec_mult_8BA (
    IPState	*pIPState,
    Int8	*pMatrix,
    Int32	*pInVector,
    Int32	*pOutVector,
    Int32	 numRows,
    Int32	 numCols)
{
    PROLOGUE;
    Int32	 row;
    Int32	 sum;
    Int32	*pLocalInVec, *pEndInVec;
    Int32	 longestAccess;		/* the longest memory access
					 * allowed for the current processor
					 * and source and dest images. */
	/* the matrix is laid out in memory as a bunch of rows,
	 * each padded to an 8-byte boundary.  Since we only
	 * visit each matrix element once, we just march through
	 * it linearly.  colsPad is the number we add at the end
	 * of each row to get to the beginning of the next row.
	 */
    Int32	 colsPad = (numCols & 0x7) ? (8 - (numCols & 0x7)) : 0;

    if (numRows <= 0)
    NORMAL_RETURN;
    
    if (numCols <= 0)
    NORMAL_RETURN;

	/* in the inner loop, we can reduce the number of active
	 * variables by comparing one of the advancing pointers
	 * with a constant instead of decrementing a column number
	 * and seeing if it hits 0.
	 */
    pEndInVec = pInVector + numCols;

	/* See if there's an MMX implementation we're allowed to use.
	 */
    ip_checkImageAccess (pIPState, (UInt8 *) pMatrix,
			 numCols+colsPad,
			 (UInt8 *) pInVector,
			 (Int32)   pOutVector,
			 &longestAccess);
    if ((longestAccess != 4) && (longestAccess != 8))
	return ia_internal;

	/* If there's no MMX code, do the multiplies with C */
    if ((longestAccess == 4) || (!ip_longOpMathLibLinked (pIPState)))
    {
	for (row = numRows; row > 0; row--)
	{
	    sum = 0;
	    pLocalInVec = pInVector;
	    while (pLocalInVec < pEndInVec)
	    {
		sum += (Int32)(*pMatrix) * (*pLocalInVec);
		pMatrix++;
		pLocalInVec++;
	    }
    
		/* move to next row starting on an 8 byte boundary */
	    pMatrix += colsPad;
	    *pOutVector++ = sum;
	}
    }
    else	/* There's MMX or other long operation code present */
    {
	    /* do whatever's necessary to enter long operation mode. */
	ip_initLongOps (pIPState);

	ip_multMatrix8Vector32 (pIPState, pMatrix, pInVector, pOutVector,
			        numRows, numCols);

	    /* do whatever's necessary to exit long operation mode. */
	ip_finishLongOps (pIPState);
    }

    NORMAL_RETURN;
}


/*

$Log: i_matvec.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.2  1998/01/13 23:48:46  ddavies
i_matvec.c -> /product/ipcore/ipshared/src/RCS/i_matvec.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.1  1996/10/30 05:51:48  ddavies
Initial revision


*/

