/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_mhc.c: High level code used in Modified Huffman compression.
 *
 *   	Int32	i_compressMHRows()
 *			Compresses a block of scan lines using
 *			the Modified Huffman algorithm.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#define	cDummyMarker	0xFFFFFFFF	/* use to mark dummy run beyond end of
					 * valid runs */

IP_RCSINFO(RCSInfo, "$RCSfile: i_mhc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* This is the BIG COMMENT that tells the overall story of our approach
 * to Modified Huffman compression.  There's a similar comment about
 * decompression in i_mhd.c.
 */

/* The basic style of interface we use here was worked out when the
 * PackBits code was written.  Lots of the comments are cribbed from
 * i_packc.c.
 *
 * The specifications for MH and MR encoding are written up in the
 * TIFF 6.0 spec, sections 10 and 11.  The CCITT specifications for the
 * FAX standards these are based on can be found in Fascicle VII.3 -
 * Recommendation T.4, "Standardization of Group 3 Facsimile Apparatus
 * for Document Transmission."
 *
 * The basic idea is that an image line is divided into runs of white
 * pixels and runs of black pixels.  The runs alternate.  One should
 * encode all the runs in each line from left to right.  For the
 * sake of convention, one assumes that the first run in every line is
 * white.  If it's really black, a white run of length 0 is encoded.
 *
 * The excitement here is that there are fancy codes used to encode the
 * runs.  There is one code used to specify blocks of white space
 * (used to cover the distance from a right endpoint of one run
 * to the left endpoint of the run to the right) and another
 * code used to specify runs of black pixels.  Why two codes?
 * It's probably because black runs tend to be shorter than white runs.
 * The black code has fewer bits/distance at the small distances.
 * The white code has fewer bits/distance at the larger distances.
 *
 * Each of the codes has individual entries for all distances from
 * 0 through 63.  They also have individual entries for each
 * multiple of 64 up through 1728.  There is one common code for
 * spanning pixels in blocks of 1792, 1856 (units of 64), ..., 2560.
 * For distances greater than 2560, one concatenates 2560 codes
 * together.  Therefore, a single distances has the form:
 *	(<some number of 2560 codes>)
 *	(<black or white code for 64 through 2496 in blocks of 64))
 *	<terminating black or white code specifying a distance
 *	 from 0 through 63>
 *
 * Only the last code is required.  If the block length is an
 * exact multiple of 64, the terminating code specifies a distance
 * of 0.  If the block is an exact multiple of 2560, the middle
 * code is left out and the last code specifies 0.
 * If n * 2560 < distance < n * 2560 + 64, the middle code is also
 * left out and the terminating code specifies the offset from
 * the last multiple of 2560.
 *
 * As stated above, the black and white codes are different.
 * The white codes are nicely arranged so that using an 8 bit
 * lookup table will return the following info:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code).
 * If the code value is greater than 63, it is a non-terminating
 * code.
 * If the code length is 9 (the length of the longest white code),
 * it is a 64x code and the final value should be found by adding
 * the ninth bit * 64 to the code value.  If the answer is 1664,
 * Then the answer is really 1728 (weird - is 1664 so common that
 * it needs a special code?).  If the code length is > 9, this is
 * a 64x code shared with the black codes.  Use the black lookup
 * method as described next.
 *
 * The black codes have a nice feature.  Either the first four bits
 * are 0000, or they're not.  If they are not 0000, we can use a
 * lookup table indexed by 6 bits that encodes the 9 interesting
 * entries. These represent the distances 1 through 9.  Each entry has
 *	1. code value
 *	2. code length in bits.
 *
 * If the first four bits are 0000, we get rid of them and use an 8-bit
 * lookup table on the remaining bits.  The output of this
 * table is the same as the output of the white table described above:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code after discarding the first 4 bits).
 * There are 13-bit black codes, but they are nicely arranged.  So,
 * if the code length after discarding the first 4 bits is 9, this
 * is a 64x code and the final value should be found by adding the
 * ninth bit * 64 to the code value.
 *
 * To encode a distance, we put out one 2560 code for each multiple
 * of 2560 in the distance, then the correct 64x code (if any) for
 * any non-zero multiples of 64 remaining, then the terminating code
 * for the remainder.  Only the terminating code is required.  To
 * decode a number, we read in codes, adding their values, until
 * a terminating code (one whose length is less than 64) is reached.
 *
 *			*	*	*
 *
 * Swell.  So how does the encoding work?  For each line, we call
 * i_findAllRunsInLineArray to find all the run endpoints.  We assume
 * that all pixels are in pixr order so i_findAllRunsInLineArray works
 * right.
 *
 * The run array is in the encoder state block.  It starts with a dummy
 * element that has left and right coords at -1.  We then just work through
 * the array.  If it starts with a black run, we encoded a white run
 * of length 0 first.  Otherwise, we encode the length of the first
 * white run (distance to first black run endpoint).  Following this
 * is the length of the next black run and so on.
 *
 * Let's say something about the mechanics of composing and decoding
 * bitstreams.  In both cases, we keep the bitstream in a 32-bit
 * register.  When composing, we keep a pointer to the first vacant
 * bit relative to the LS (right) end of the word.  New codes are shifted
 * left by this much and ORed into the accumulator - to the left
 * of the older codes.  The pointer to the first vacant bit is
 * incremented by the length of the code.  Each time we insert
 * a new code, we look at the new value of the pointer.  If it's
 * less than 16, we just keep going.  If there are at least
 * 16 bits to be written, then we write out the bottom two bytes
 * in the accumulator.
 *.
 * Note that since the bits are being encoded in the
 * register from right to left and the encoded stream is to be
 * written from left to right, we need to reverse the order of the
 * bits in each byte before writing them to memory.  We use the
 * reverseByte table for this.
 *
 * At the bottom of the encoding loop (the end of one scan line),
 * we have to decide whether to flush the accumulator.  In MH,
 * the accumulator is flushed on every line.  In MR and MMR,
 * there is no flush.  Flushing is done by shifting and writing
 * until the vacant bit pointer is less than or equal to 0.
 * Since the logical right shifts insert 0's at the top of the
 * accumulator, the last byte will be padded with 0's.  This is fine.
 *
 * Note that the encoded data stream is always written out in the
 * platform's natural byte order.  It will have to show up in that
 * order at some point and (theoretically) there aren't very many
 * encoded bytes, so this is no problem, right?
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which this procedure will stop.  It might actually
 * finish encoding the entire scan line.  When this happens, we update the
 * pointer to the compressed data that was passed in, set the state
 * to cCCITTCompNewLine and store the current accumulator and vacantBits
 * and return happily.  Note that if we've just flushed the accumulator,
 * vacantBits will be 0.
 *
 * The other reason for returning is that we run out of room for
 * compressed data.  We detect this by comparing pCurCompData to a
 * pointer to the end of the buffer just before writing encoded data.
 * Since we want to pretend that we have a callback to fill the buffer,
 * we need a restart location next to every check.  There are
 * three checks when writing codes for long white runs (remember
 * that long codes can have a sequence of 2560s plus another non-terminating
 * code plus a terminating code and each of these offers an opportunity
 * to write accumulated bits to the buffer), one check when writing
 * short white runs, three checks when writing long black runs,
 * one check for short black runs and one chech when attempting to
 * flush the accumulator at the end of the line.  That's 10 states
 * in all: 1 (start of line) + 3 (long white) + 1 (short white) +
 * 3 (long black) + 1 (short black) + 1 (flush accum)..
 *
 * We always start the run array with a dummy element that says there is
 * a black run whose left and right endpoints are at -1.  We know we can
 * always point at this run.  The last run is marked by having a left
 * endpoint of imageWidth and a right endpoint of cVeryLargeNum.
 *
 *			*	*	*
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * and the continueImageBlock, continueCompDataStream and flushPipeline args
 * discussed above.  The caller also supplies pointers to words used to
 * return the exhaustedImageData, exhaustedCompData, bytesWritten and
 * percentDone args.  The bytesWritten arg returns the number of compressed
 * bytes written so far.  If flushPipeline is not set, some of the compressed
 * bytes may have been computed, but not yet written.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	We have processed all the image data requested.  The compression
 *	buffer still has room.  If flushPipeline was true, all the compressed
 *	data has been flushed to the output.  Otherwise, some may still be in
 *	the compression pipeline so it can be concatenated with the next
 *	set of compressed data generated.  In either case, the number of
 *	bytes actually written to the buffer (not including any bytes
 *	still in the pipeline) are returned.
 *	When the compression routine is called next, the client must
 *	pass in a pointer to the byte at which the next compressed byte
 *	should be placed.  This may be just beyond the end of the
 *	current compressed data, at the beginning of the same buffer, 
 *	or in another buffer entirely.  The size of the compressed data
 *	buffer must also be (re)calculated by the client.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to write the last
 *	byte of the compressed data buffer at the same time the compressed
 *	data came to an end.  As before, if flushPipeline was false, there
 *	may be more compressed data in the pipeline.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of room for compressed data before writing the last
 *	byte of compressed data requested.  Some of the data to be written
 *	is in the pipeline.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */

/* PUBLIC */
/**************************************************************************
 *   i_compressRowsMH()
 *
 *	This procedure compresses images in units of scan lines
 *	using the Modified Huffman algorithm.
 *
 *	Input:
 *		pCCITTCompInfoIn:	Pointer to CCITT compression info
 *		pImage:			Pointer to the upper left corner
 *					of the frame of the uncompressed
 *					image
 *		w:			Width of uncompressed image including
 *					frame.
 *		startLine:		First image line to be compressed
 *					relative to the top of the frame
 *		numLines:		Number of lines to be compressed
 *		bpl:			bytes/line of uncompressed data.
 *		pCompData:		Pointer to first vacant byte
 *					of buffer into which compressed
 *					data will be placed.
 *		compDataBufSize:	Number of vacant bytes remaining
 *					in the compressed data buffer.
 *		continueImageBlock:	cTrue => we are compressing
 *					from the same slice of image as was
 *					specified in the previous call and the
 *					saved state in pCCITTCompInfoIn is
 *					valid.
 *					This arg is set to cFalse if either
 *					we are starting with all new data
 *					or the previous call returned because it
 *					ran out of image data.
 *		continueCompDataStream:	cTrue => The saved compression state
 *					in pCCITTCompInfoIn is valid.  If
 *					the previous call to this routine
 *					filled the compressed data buffer
 *					before the strip, tile or image was
 *					complete, a new buffer should be
 *					supplied.  This arg is set to 
 *					cFalse when the block of compressed
 *					data is unrelated to the block used
 *					in the previous call to this routine
 *					as recorded in pCCITTCompInfoIn.
 *					SETTING THIS ARG TO cFalse CAUSES
 *					DATA IN THE PIPELINE TO BE DISCARDED!
 *		flushPipeline:		If this arg is set, this routine
 *					will flush all the compressed data
 *					in the pipeline to the compressed
 *					data buffer when the specified
 *					number of scan lines have been
 *					compressed.  Note that in MH encoding,
 *					the pipeline is flushed at the end
 *					of each scanline anyway.
 *					However this parameter
 *					can be used by the client to
 *					indicate when we are supposed to
 *					be done.  If this routine returns
 *					with *pExhaustedImageData true when
 *					flushPipeline was set, subsequent
 *					calls with either continueImageBlock or
 *					continueCompDataStream will return
 *					ia_invalidParm.
 *		pExhaustedImageData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because specified
 *					number of lines have been processed.
 *					This may require more than one
 *					call to this routine.
 *		pExhaustedCompData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because it filled
 *					the compressed data buffer.  This
 *					routine may be called again with
 *					a new buffer.
 *		pPercentDone:		Pointer to returned arg.  When
 *					routine returns with *pExhaustedCompData
 *					set, this arg specifies the rough
 *					percentage of the image block that
 *					has been compressed.  It is
 *					intended to be used to estimate the
 *					amount of compressed data needed
 *					to complete the task.  The result
 *					ranges from 0 to 100.
 *		pBytesWritten:		Number of compressed bytes actually
 *					written to the compressed data
 *					block.  If flushPipeline is false or
 *					(*pExhaustedImageData == cFalse and
 *					*pExhaustedCompData == cTrue),
 *					some bytes required to complete
 *					a scan line may still be in the
 *					pipeline and are not included in
 *					this count.
 *
 *	Action:
 *		Compresses a horizontal slice image using the
 *		Modified Huffman method.  See TIFF 6.0 Specification,
 *		Section 10.
 *
 *		There are several completion conditions:
 *
 *	__error__	!= ia_successful
 *	*pExhaustedImageData == don't care
 *	*pExhaustedCompData == don't care
 *		This means something catastrophic happened.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cFalse
 *		This means we compressed the desired number of
 *		image lines without filling the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.  If flushPipeline
 *		was set, all compressed data has been written to
 *		the compressed data buffer and the saved compressed state
 *		was marked invalid.  If flushPipeline was not set,
 *		some of the compressed data may still be in the
 *		pipeline and the saved compressed data state is valid.
 *		In the latter case, this routine may be called again with
 *		continueCompDataStream = cTrue.  In either case, the
 *		client must supply a new value for pCompData and
 *		compDataSize.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cTrue
 *		This means we filled the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is unusual, but not impossible.  If flushPipeline
 *		was set, all compressed data has been written to the
 *		compressed data buffer, the saved state is marked
 *		invalid and it is not possible to continue from this state.
 *		Otherwise, some data may remain in the pipeline.
 *		This return state cannot occur of flushPipeline was set
 *		and the pipeline still contains data (see next state).
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cFalse
 *	*pExhaustedCompData == cTrue
 *		The routine filled the compression data buffer before completing
 *		the specified number of lines.  One way this can
 *		happen is if all image data has been processed, but
 *		flushPipeline was set and some data remains in the pipeline.
 *		When this condition does occur, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  
 *		The image may be moved around as a block after this
 *		condition has been returned, but the internal
 *		composition of the image (and startLine and numLines) 
 *		should not be changed.
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *		
 *		The image data is always in PIXR order.
 *		
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		a pointer was NULL or 
 *					(continueImageBlock && the saved image
 *					info is not valid) or
 *					(continueCompDataStream && the saved
 *					compression info is not valid)
 *					or pCCITTCompInfo->padType is invalid.
 *					It must be cPadToByte or cPadToWord.
 *		ia_bufferExceeded	ran out of room for compressed data
 *					when flushPipeline was set.
 *
 **************************************************************************/
Int32  CDECL 
i_compressRowsMH (
    IPState    *pIPState,
    void       *pCCITTCompInfoIn,
    UInt8      *pImage,
    Int32	w,
    Int32	startLine,
    Int32	numLines,
    Int32	bpl,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool  	continueImageBlock,
    Bool  	continueCompDataStream,
    Bool	flushPipeline,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesWritten)
{
    PROLOGUE;
    CCITTCompressionInfo	*pCCITTCompInfo;
    UInt8			*pCompDataEnd;	/* address 1 beyond the last
						 * byte of compressed data */
    UInt8			*pCurCompData;	/* current compressed data
						 * pointer */
    Run32			*pRuns;		/* pointer to first run in
						 * array of runs. */
    Run32			*pDummy;	/* returned ptr to element
						 * just beyond last valid run
						 * in the array */
    UInt32			 remainingLines;
    


	/* check for null pointers */
    if ((pCCITTCompInfoIn == NULL) || (pCompData == NULL) ||
	(pImage == NULL) || (compDataBufSize == 0) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesWritten == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pCCITTCompInfo = (CCITTCompressionInfo *) pCCITTCompInfoIn;

	/* If the padding is wrong, complain */
    if ((pCCITTCompInfo->padType != cPadToByte) &&
	(pCCITTCompInfo->padType != cPadToWord))
	return ia_invalidParm;

	/* Strip off the frame */
    pImage += (bpl << cLogFrameBits) + cFrameBytes;
    startLine -= cFrameBits;
    w -= 2*cFrameBits;
    if (w <= 0)
	NORMAL_RETURN;	/* there's no image here */

	/* Initialize the state from the saved state passed in *pCCITTCompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	pImage += startLine * bpl;
	pCCITTCompInfo->lineNumber = 0;
	pCCITTCompInfo->currentRun = 0;
	pCCITTCompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pCCITTCompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pCCITTCompInfo->lineNumber;
	pImage += (startLine + pCCITTCompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pCCITTCompInfo->compState = cCCITTCompNewLine;
	pCCITTCompInfo->accumBits = 0;
	pCCITTCompInfo->vacantBit = 0;
	pCCITTCompInfo->remainingLength = 0;
	pCCITTCompInfo->compInfoIsValid = cTrue;
    }
    else
    if (!pCCITTCompInfo->compInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* get a pointer to the second run.  The first is a dummy. */
    pRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
			        pCCITTCompInfo->runArrayOffset +
			        sizeof(Run32));


	/* Compress each scan line.
	 * This is done in several steps:
	 *   1. If the compression state does not equal cCCITTCompNewLine,
	 *	jump to step 3.
	 *   2. We did not quit in the middle of putting out this scan
	 *	line.  Find all the runs.
	 *   3. Call ip_compressMHRow to put out the codes for the runs
	 *	of the scan line.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--, pImage += bpl)
    {
	if (pCCITTCompInfo->compState == cCCITTCompNewLine)
	{
		/* Find the run boundaries and put them into the array
		 * of runs.
		 */
	    __error__ =
	        i_findAllRunsInLineArray (pIPState,
					  (UInt8 *)pImage, /* inside "frame" */
				          w-1,		 /* right limit */
				          0,		 /* left limit */
				          pRuns,	 /* dest array */
				          (w+1)/2,	 /* # of Run32s */
				          0,		 /* add this to each
							  * run coord */
					  &pDummy);	 /* Address of run
							  * just past last
							  * valid run */
	    if (__error__ != ia_successful)
		NORMAL_RETURN;

	        /* Mark the dummy element at the end.  It's left coord
		 * should be just beyond the right edge of the image.  The
		 * right coord will be set to cDummyMarker, since that's
		 * not a legal address.
		 */
	    pDummy->runLeft = w;
	    pDummy->runRight = cDummyMarker;

		/* Set the run offset in the compression info to 0
		 * so that we start interpreting the very first run
		 * and not some run left over from some previous
		 * stoppage.
		 */
	    pCCITTCompInfo->currentRun = 0;
	}

	    /* Ask the compression code to translate runs into the
	     * compressed data stream.  The second arg is the value used to
	     * mark the right coord of the dummy element at the end
	     * of the array.
	     */
	__error__ = ip_compMHRow (pIPState, pCCITTCompInfo, cDummyMarker,
			        pCCITTCompInfo->padType,
			        &pCurCompData, pCompDataEnd);
	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;

	    /* We have now attempted to compress one more scan line,
	     * possibly not the last one.  Should we stop now?
	     * If we stopped in the middle of a line because the
	     * compressed data buffer was full and we had more stuff
	     * to write, or we finished the line but just happened
	     * to fill the compressed data buffer and there are
	     * more lines to go, stop now.
	     */
	if ((__error__ == ia_bufferExceeded) ||
	    ((pCurCompData == pCompDataEnd) && (remainingLines > 1)))
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * ((pCCITTCompInfo->lineNumber) * w)) /
			    (numLines * w);
	    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

		/* if we happened to stop on an image scan line boundary,
		 * increment the saved line number so we don't do it again.
		 */
	    if (pCCITTCompInfo->compState == cCCITTCompNewLine)
	    {
		pCCITTCompInfo->lineNumber++;
	    }
		/* Set __error__ to ia_successful.  This is the convention
		 * at this point.  There wasn't a real error, we just
		 * filled up the output buffer.  The caller detects this
		 * by looking at *pExhaustedCompData.
		 */
	    __error__ = ia_successful;
	    NORMAL_RETURN;
	}
	    /* increment the current line number in case we need
	     * to stop in the next line.
	     */
	pCCITTCompInfo->lineNumber++;
    }

	/* We've done all the scan lines.  So, just
	 * say things were successful and return.
	 */
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

	/* if flushPipeline is true, declare all the saved state to be
	 * invalid.  This helps the client catch bugs in which continuation
	 * was attempted, but the saved state is invalid.
	 */
    if (flushPipeline)
    {
	pCCITTCompInfo->imageInfoIsValid = cFalse;
	pCCITTCompInfo->compInfoIsValid = cFalse;
    }

    NORMAL_RETURN;

}


/*

 $Log: i_mhc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.4  1998/01/13 23:48:47  ddavies
 i_mhc.c -> /product/ipcore/ipshared/src/RCS/i_mhc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1996/05/11 01:00:59  ddavies
 i_mhc.c -> /product/ipcore/ipshared/src/RCS/i_mhc.c,v

 Fix lots of comments. Move the loading of pCCITTCompInfo up so we can
 use it to check the padType inside the structure.

 Revision 1.2  1996/04/25 18:41:28  ddavies
 i_mhc.c -> /product/ipcore/ipshared/src/RCS/i_mhc.c,v

 Fix comments.  Abort if the real width of the image is negative.

 Revision 1.1  1996/04/10 22:01:15  ddavies
 Initial revision


 */
