/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_mhd.c: High level code used in Modified Huffman decompression.
 *
 *   	Int32	i_decompressMHRows()
 *			Decompresses a block of scan lines using
 *			the Modified Huffman algorithm.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_mhd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* This is the BIG COMMENT that tells the overall story of our approach
 * to Modified Huffman decompression.  There's a similar comment about
 * compression in i_mhc.c.
 */

/* The basic style of interface we use here was worked out when the
 * PackBits code was written.  Lots of the comments are cribbed from
 * i_packd.c.
 *
 * The specifications for MH and MR encoding are written up in the
 * TIFF 6.0 spec, sections 10 and 11.  The CCITT specifications for the
 * FAX standards these are based on can be found in Fascicle VII.3 -
 * Recommendation T.4, "Standardization of Group 3 Facsimile Apparatus
 * for Document Transmission."
 *
 * The basic idea is that an image line is divided into runs of white
 * pixels and runs of black pixels.  The runs alternate.  One should
 * encode all the runs in each line from left to right.  For the
 * sake of convention, one assumes that the first run in every line is
 * white.  If it's really black, a white run of length 0 is encoded.
 *
 * The excitement here is that there are fancy codes used to encode the
 * runs.  There is one code used to specify blocks of white space
 * (used to cover the distance from a right endpoint of one run
 * to the left endpoint of the run to the right) and another
 * code used to specify runs of black pixels.  Why two codes?
 * It's probably because black runs tend to be shorter than white runs.
 * The black code has fewer bits/distance at the small distances.
 * The white code has fewer bits/distance at the larger distances.
 *
 * Each of the codes has individual entries for all distances from
 * 0 through 63.  They also have individual entries for each
 * multiple of 64 up through 1728.  There is one common code for
 * spanning pixels in blocks of 1792, 1856 (units of 64), ..., 2560.
 * For distances greater than 2560, one concatenates 2560 codes
 * together.  Therefore, a single distances has the form:
 *	(<some number of 2560 codes>)
 *	(<black or white code for 64 through 2496 in blocks of 64))
 *	<terminating black or white code specifying a distance
 *	 from 0 through 63>
 *
 * Only the last code is required.  If the block length is an
 * exact multiple of 64, the terminating code specifies a distance
 * of 0.  If the block is an exact multiple of 2560, the middle
 * code is left out and the last code specifies 0.
 * If n * 2560 < distance < n * 2560 + 64, the middle code is also
 * left out and the terminating code specifies the offset from
 * the last multiple of 2560.
 *
 * As stated above, the black and white codes are different.
 * The white codes are nicely arranged so that using an 8 bit
 * lookup table will return the following info:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code).
 * If the code value is greater than 63, it is a non-terminating
 * code.
 * If the code length is 9 (the length of the longest white code),
 * it is a 64x code and the final value should be found by adding
 * the ninth bit * 64 to the code value.  If the answer is 1664,
 * Then the answer is really 1728 (weird - is 1664 so common that
 * it needs a special code?).  If the code length is > 9, this is
 * a 64x code shared with the black codes.  Use the black lookup
 * method as described next.
 *
 * The black codes have a nice feature.  Either the first four bits
 * are 0000, or they're not.  If they are not 0000, we can use a
 * lookup table indexed by 6 bits that encodes the 9 interesting
 * entries. These represent the distances 1 through 9.  Each entry has
 *	1. code value
 *	2. code length in bits.
 *
 * If the first four bits are 0000, we get rid of them and use an 8-bit
 * lookup table on the remaining bits.  The output of this
 * table is the same as the output of the white table described above:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code after discarding the first 4 bits).
 * There are 13-bit black codes, but they are nicely arranged.  So,
 * if the code length after discarding the first 4 bits is 9, this
 * is a 64x code and the final value should be found by adding the
 * ninth bit * 64 to the code value.
 *
 * To encode a distance, we put out one 2560 code for each multiple
 * of 2560 in the distance, then the correct 64x code (if any) for
 * any non-zero multiples of 64 remaining, then the terminating code
 * for the remainder.  Only the terminating code is required.  To
 * decode a number, we read in codes, adding their values, until
 * a terminating code (one whose length is less than 64) is reached.
 *
 *			*	*	*
 *
 * Swell.  So how does the decoding work?  It's pretty much the inverse
 * of the encoding routine.  For each line, we call ip_decompMHRow
 * to find all the run endpoints.  Then we call ip_paintRunsLine to
 * paint the runs found into the bitmap.  The run array is in the
 * decoder state block.
 *
 * Let's say something about the mechanics of composing and decoding
 * bitstreams.  In both cases, we keep the bitstream in a 32-bit
 * register.  When decoding, we keep a pointer to the first vacant
 * bit relative to the LS (right) end of the word.  New bytes from
 * the compresssed data stream are shifted left by this much and ORed
 * into the accumulator - to the left of the older data.
 * The pointer to the first vacant bit is incremented by the length
 * of a byte (8).
 *
 * Data is removed from the right end of the register as it is decoded.
 * Each code has a value (enocded runlength) and a length in
 * bits.  Each time we remove a code, removed the indicated number
 * of bits from the accumulator.
 *
 * Since a single run may be encoded as several tokens, we need a loop
 * to decode each kind of run (white vs black).  The lower level routine
 * is composed of an outer loop that contains two inner loops in sequence.
 * The first inner loop decodes a white code.  The second one decodes a
 * black code.  One trip through the outer loop fills in the left and right
 * edges one run.
 *
 * At the top of each inner loop, we look at the number of bits in
 * the bitstream accumulator.  If it's greater than than 16, we just
 * keep going.  If there are fewer than 16 bits in the accumulator,
 * then we read in two new bytes - shifting them into position just to the
 * left of the leftmost currently valid bits.
 *
 * Note that since the bits are being decoded in the register from
 * right to left and the encoded stream is to be read from left to
 * right, we need to reverse the order of the bits in each byte before
 * inserting them into the accumulator.  We use the reverseByte table for this.
 *
 * At the bottom of the decoding loop (the end of one scan line),
 * we have to decide whether to skip over bits in the accumulator to
 * get to the next byte or word boundary.  In MH, every line is padded
 * to byte boundaries.  Leffler also allows one to pad to word boundaries,
 * though this is not in the spec.  In MR and MMR, there is no padding.
 * Flushing is done by discarding bits on the right end of the accumulator
 * until the number of bits remaining is a multiple of 8 (or 16 if flushing
 * to the next word boundary).
 *
 * Note that the compressed data stream is always read in the
 * platform's natural byte order.  That's just the way life is.  I can't
 * think of any practical way to read lots of bytes at once and there`
 * aren't very many compressed bytes, right?
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which the lower level procedure will stop.  It
 * might actually finish decoding one complete scan line.  When this
 * happens, flush to the next byte or word boundary if necessary,
 * update the pointer to the compressed data that was passed in, set the
 * state to cCCITTDecompNewLine, store the current accumulator and
 * vacantBits and return happily.
 *
 * The other reason for returning is that we run out of compressed data.
 * We detect this by comparing pCurDecompData to a pointer to the end of
 * the buffer just before reading compressed data. Since we want to
 * pretend that we have a callback to fill the buffer, we need a restart
 * location next to every check.  There are two checks when reading
 * codes, one at the top of the loop that reads white run tokens and
 * one at the top of the loop that reads black run tokens.
 *
 *			*	*	*
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * and the continueImageBlock, continueCompDataStream and flushPipeline args
 * discussed above.  The caller also supplies pointers to words used to
 * return the exhaustedImageData, exhaustedCompData, bytesWritten and
 * percentDone args.  The bytesRead arg returns the number of compressed
 * bytes read so far.  If flushPipeline is not set, some of the compressed
 * bytes may have been read, but not yet decoded.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.  One possibly catastrophe is
 *	that flushPipeline was set but insufficient compressed data
 *	was supplied.  In this case, we return ia_bufferExceeded.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	We have processed all the image data requested.  The compression
 *	buffer is not empty.  Some of the compressed data bytes are
 *	still in the pipeline regardless of the state of flushPipeline.
 *	The number of bytes actually read from the buffer
 *	(including any bytes still in the pipeline) are returned.
 *	When the compression routine is called next, the client must
 *	pass in a pointer to the byte holding the next compressed byte.
 *	This may be just beyond the end of the current compressed data,
 *	at the beginning of the same buffer, or in another buffer entirely.
 *	The size of the compressed data buffer must also be (re)calculated
 *	by the client.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to read the last
 *	byte of the compressed data buffer at the same time the compressed
 *	data came to an end.  As before, if flushPipeline was false, there
 *	may be more compressed data in the pipeline.  Normally, this
 *	case occurs when exactly the right amount of compressed data was
 *	supplied and flushPipeline was true.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of compressed data before reading the last
 *	byte of compressed data requested.  Some of the data read
 *	is in the pipeline.  This can only happen if flushPipeline was
 *	false.  If flushPipeline is true, we keep going until we either
 *	finish all the scan lines or get an ia_bufferExceeded.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */


/* PUBLIC */
/**************************************************************************
 *   i_decompressRowsMH()
 *
 *	This procedure decompresses Modified Huffman encoded images in
 *	units of scan lines.
 *
 *	Input:
 *		pCCITTDecompInfoIn:	Pointer to CCITT decompression info
 *		pImage:			Pointer to the upper left corner
 *					of the frame of the uncompressed
 *					image.  If NULL, no runs are painted.
 *		startLine:		First image line to be decompressed
 *					relative to the top of the frame
 *		numLines:		Number of lines to be decompressed
 *		bpl:			bytes/line of uncompressed data.
 *		combineResults:		If set to cFalse, the image is assumed
 *					to be initially clear.  This is most
 *					efficient.  If set to cTrue, the
 *					decoded image is ORed into the original. *		pCompData:		Pointer to next byte of compressed
 *					data to be read.  If several calls
 *					to this routine are needed to process
 *					a block of compressed data, this
 *					pointer must be updated on each call.
 *		compDataBufSize:	Number of bytes of compressed data
 *					remaining in the buffer.
 *		continueImageBlock:	cTrue => we are decompressing
 *					into the same slice of image as was
 *					specified in the previous call and the
 *					saved state in pCCITTDecompInfoIn is
 *					valid.
 *					This arg is set to cFalse if either
 *					we are starting with all new data
 *					or the previous call returned because it
 *					ran out of image data.
 *		continueCompDataStream:	cTrue => The saved decompression state
 *					in pCCITTDecompInfoIn is valid.  This
 *					allows the procedure to treat several
 *					buffers of compressed data as though
 *					it were one contiguous buffer.
 *					If set to cFalse, this routine will
 *					discard all previous compression
 *					state.
 *		flushPipeline:		If this arg is set, this routine
 *					will attempt to complete the
 *					specified number of scan lines
 *					WITHOUT stopping when it starts to
 *					run out of data; i.e. it continues
 *					until its pipeline is completely
 *					empty.  If this is not possible,
 *					an ia_bufferExceeded error is raised.
 *		pExhaustedImageData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because specified
 *					number of lines have been processed.
 *					This may require more than one
 *					call to this routine.
 *		pExhaustedCompData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because it reached
 *					the end of the compressed data buffer
 *					as specified by compDataSize.  This
 *					routine may be called again with 
 *					new compression data and the the 
 *					same image.  This arg will not be 
 *					set if flushPipeline is set and we 
 *					failed to decompress the specified 
 *					number of scan lines (routine returns
 *					ia_bufferExceeded instead).
 *		pPercentDone:		Pointer to returned arg.  When
 *					routine returns with *pExhaustedCompData
 *					set, this arg specifies the rough
 *					percentage of the image block that
 *					has been decompressed.  It is
 *					intended to be used to estimate the
 *					amount of compressed data needed
 *					to complete the task.  The result
 *					ranges from 0 to 100.
 *		pBytesRead:		Pointer to returned arg.  Contains
 *					number of compressed data bytes
 *					read by the routine before it returned.
 *					The client may safely dispose of
 *					these bytes.
 *
 *	Action:
 *		Decompresses a horizontal slice of MH data.
 *		There are several completion conditions:
 *
 *	__error__	!= ia_successful
 *	*pExhaustedImageData == don't care
 *	*pExhaustedCompData == don't care
 *		This means something catastrophic happened.  One possible
 *		catastrophe (ia_bufferExceeded) happens if the client
 *		set flushPipeline and failed to supply enough compressed
 *		data.		
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cFalse
 *		This means we produced the desired number of uncompressed
 *		image lines without running out of compressed data.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse and
 *		continueCompDataStream = cTrue.  pCompData and compDataSize
 *		should be adjusted to account for the number of 
 *		compressed data bytes that have already been read (returned
 *		in *pBytesRead).
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cTrue
 *		This means we ran out of compressed data at the same time
 *		we completed the correct number of image lines.  This
 *		is the normal condition at the bottom of a TIFF slice or
 *		the bottom of the image.  It is not possible to restart
 *		from this state, though it is possible to decompress
 *		with new image and compression data.  When the client
 *		expects to run out of compressed data, the flushPipeline
 *		argument should be set.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cFalse
 *	*pExhaustedCompData == cTrue
 *		The routine ran out of compressed data before completing
 *		the specified number of lines.  This condition will not
 *		occur if flushPipeline was set.  If flushPipeline was set
 *		and we ran out of compressed data before completing the
 *		the task, an ia_bufferExceeded error will be returned.
 *		When this condition does occur, the routine may be
 *		restarted by supplying more compressed data, setting
 *		BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  Note that the first byte of the new compressed
 *		data must be the one immediately following the last
 *		byte of old compressed data.
 *		The image may be moved around as a block if necessary, but
 *		the internal composition of the image (and startLine
 *		and numLines) should not be changed.  When this
 *		condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much compression data is 
 *		needed to complete the slice.
 *		
 *		The image data is always in PIXR order.
 *		
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		a pointer was NULL or 
 *					(continueImageBlock && the saved image
 *					info is not valid) or
 *					(continueCompDataStream && the saved
 *					compression info is not valid)
 *		ia_bufferExceeded	ran out of compressed data when
 *					flushPipeline was set.
 *		ia_decodeFailure	an illegal code was enountered.
 *
 **************************************************************************/
Int32  CDECL 
i_decompressRowsMH (
    IPState    *pIPState,
    void       *pCCITTDecompInfoIn,
    UInt8      *pImage,
    Int32	startLine,
    Int32	numLines,
    Int32	bpl,
    Bool	combineResults,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Bool	flushPipeline,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesRead)
{
    PROLOGUE;
    CCITTDecompressionInfo	*pCCITTDecompInfo;
    UInt8			*pCompDataEnd;	/* address 1 beyond the last
						 * byte of compressed data */
    UInt8			*pCurCompData;	/* current compressed data
						 * pointer */
    Run32			*pRuns;		/* pointer to first run in
						 * pCCITTDecompInfo */
    Run32			*pDummy;	/* returned ptr to element
						 * just beyond last valid
						 * run in the array */
    Int32			 remainingLines;
    Int32			 imWidth;	/* width of unframed image */
    

	/* check for null pointers */
    if ((pCCITTDecompInfoIn == NULL) || (pCompData == NULL) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesRead == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pCCITTDecompInfo = (CCITTDecompressionInfo *) pCCITTDecompInfoIn;

	/* If the padding is wrong, complain */
    if ((pCCITTDecompInfo->padType != cPadToByte) &&
	(pCCITTDecompInfo->padType != cPadToWord))
	return ia_invalidParm;

	/* Strip off the frame */
    if (pImage != NULL)
	pImage += (bpl << cLogFrameBits) + cFrameBytes;
    startLine -= cFrameBits;
    imWidth = pCCITTDecompInfo->width;

	/* Initialize the state from the saved state passed in *pCCITTCompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	if (pImage != NULL)
	    pImage += startLine * bpl;
	pCCITTDecompInfo->lineNumber = 0;
	pCCITTDecompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pCCITTDecompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pCCITTDecompInfo->lineNumber;
	if (pImage != NULL)
	    pImage += (startLine + pCCITTDecompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
	pCCITTDecompInfo->currentRun = 0;
	pCCITTDecompInfo->accumBits = 0;
	pCCITTDecompInfo->vacantBit = 0;
	pCCITTDecompInfo->decompInfoIsValid = cTrue;
    }
    else
    if (!pCCITTDecompInfo->decompInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;


	/* get a pointer to the second run.  The first is a dummy
	 * filled with [-1, -1].  This dummy is used by the
	 * lower level code to turn the first run length into
	 * a real coordinate.  All run lengths are converted to coords
	 * by adding the previous coord.
	 */
    pRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
			        pCCITTDecompInfo->runArrayOffset +
			        sizeof(Run32));

	/* Decompress each scan line.
	 * This is done in several steps:
	 *   1. We might have quit in the middle of decoding all the
	 *	runs in this scan line.  That's the only place we
	 *	might have quit.  Since we're starting with that step
	 *	in any case, just do it using ip_decompMHRow.
	 *   2. Call ip_paintRunsLine to draw the runs into the bitmap.
	 *	We cannot abort in the middle of this step.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--)
    {
	    /* decode the runs for the next line. */
	__error__ = ip_decompMHRow (pIPState, pCCITTDecompInfo,
				  pCCITTDecompInfo->padType, imWidth,
				  flushPipeline,
			          &pCurCompData, pCompDataEnd,
				  &pDummy);
	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;
	if ((__error__ == ia_bufferExceeded) && (flushPipeline))
	{
	    pCCITTDecompInfo->imageInfoIsValid = cFalse;
	    pCCITTDecompInfo->decompInfoIsValid = cFalse;
	    NORMAL_RETURN;
	}


	    /* If we didn't get to the end of a scan line, return now
	     * instead of painting all the runs to the bitmap.  We
	     * don't paint the runs because we don't have all of them.
	     * Since flushPipeline was not set, this is all ok.
	     */
	if (__error__ == ia_bufferExceeded)
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * pCCITTDecompInfo->lineNumber) /
			    (numLines);
	    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

	    NORMAL_RETURN;
	}

	    /* Great, we got a set of runs.  Go paint them into the
	     * image.  If pImage is NULL, someone just wants the runs
	     * so don't paint them.
	     */
	if (pImage != NULL)
	{
	    __error__ = ip_paintRunsLine (pIPState, pImage, imWidth,
				          pRuns, pDummy, combineResults);
	    if (__error__ != ia_successful)
		NORMAL_RETURN;
	}

	    /* increment the current line number just in case we need
	     * to stop in the next line.
	     */
	pCCITTDecompInfo->lineNumber++;

	if (pImage != NULL)
	    pImage += bpl;
    }

	/* Great! we completed the right number of scan lines.
	 * Say that things worked and return.  But first,
	 * mark the last run as the last run so that we can
	 * find it if we happen to really only want the runs.
	 */
    pDummy->runLeft = imWidth;
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;
    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

    NORMAL_RETURN;
}


/*

 $Log: i_mhd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.7  1998/01/13 23:48:49  ddavies
 i_mhd.c -> /product/ipcore/ipshared/src/RCS/i_mhd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.6  1996/07/18 23:37:37  ddavies
 i_mhd.c -> /product/ipcore/ipshared/src/RCS/i_mhd.c,v

 Change initImage parameter to combineResults parameter.  When combineResults
 is set, the decompressed image is ORed into the original dest image.
 When combineResults is not set, the original dest image is assumed to be
 white (this is often true and is faster).

 Revision 1.5  1996/05/21 18:46:18  ddavies
 i_mhd.c -> /product/ipcore/ipshared/src/RCS/i_mhd.c,v

 Do not paint runs into a bitmap if the bitmap pointer is NULL.  This allows
 us to simply extract the runs from a MH file.  Also mark the last run
 as the dummy so we can find it when we're analyzing the run array.

 Revision 1.4  1996/05/16 17:06:21  ddavies
 ddavies on Thu May 16 10:04:45 PDT 1996

 i_mhd.c -> /product/ipcore/ipshared/src/RCS/i_mhd.c,v

 Fix several bugs that are encountered when you repeatedly read a single scan
 line from a large block of compressed data.
 First, delete the code that invalidates the saved state when flushPipeline
 is set.  It is very difficult to determine whether a particular scan line
 is the last scanline of a strip at the level where flushPipeline is set.
 Second, move the code that initializes pCCITTDecompInfo->currentRun from
 the block controlled by continueImageData to the block controlled by
 continueCompDataStream.  In decompression, the runs are associated with the
 compressed data, not with the image.  Also move some code around for
 more uniformity with the other CCITT decompression routines.

 Revision 1.3  1996/05/11 01:02:14  ddavies
 i_mhd.c -> /product/ipcore/ipshared/src/RCS/i_mhd.c,v

 Remove padType as an argument for i_decompressRowsMH.  Move the loading
 of pCCITTDecompInfo up so we can used it to check the version of padType
 now passed the structure.  Use the type from the structure.

 Revision 1.2  1996/05/06 18:45:48  ddavies
 i_mhd.c -> /product/ipcore/ipshared/src/RCS/i_mhd.c,v

 Improve comments.  Set imageInfoIsValid and decompInfoIsValid to cFalse
 if we're about to return ia_bufferExceeded when flushPipeline is true.

 Revision 1.1  1996/04/25 18:43:57  ddavies
 Initial revision


 */


