/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_minmax.c: compare two gray images, producing the min. or max.
 *	        also, functions for computing range, etc.
 *
 *                  Int32     	i_min()
 *		    Int32	i_max()
 *		    Int32	i_range()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrip.pub"
#include "shrip.prv"

/*
 * i_min
 *
 * Take the pixel-wise minimum of two images. This is "approximate", i.e. 
 * it might give the wrong answer on 1-bit differences.
 *
 * Arguments:
 *	pS1	    -	    First image
 *	sBpl1	    -	    Bytes/line of first image
 *	pS2	    -	    Second image
 *	sBpl2	    -	    Bytes/line of second image
 *	w	    -	    Width of image (w/frame)
 *	h	    -	    heigth of image (w/frame)
 *	depth	    -	    Depth of images
 *	pD	    -	    Destination
 *	dBpl	    -	    Destination bytes/line
 *
 * Returns:
 *	ia_successful, ...
 */

Int32 CDECL 
i_min (
    IPState    *pIPState,
    UInt8      *pS1,
    Int32	sBpl1,
    UInt8      *pS2,
    Int32	sBpl2,
    Int32	w,
    Int32	h,
    Int32	depth,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    Int32 i, j;
    register UInt32  wMask, w1, w2, h1, h2, notBorrowed, hBits, lBits;
	    
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

    switch (depth)
    {
    case 8:
	hBits = 0x80808080;
	lBits = 0x7f7f7f7f;
	break;

    case 4:	
	hBits = 0x88888888;
	lBits = 0x77777777;
	w >>= 1;
	break;

    case 2:
	hBits = 0xaaaaaaaa;
	lBits = 0x33333333;
	w >>= 2;
	break;

    default:
	return ia_depthNotSupported;
    }

    pS1 += (sBpl1 << cLogFrameBits) + cFrameBytes * depth;
    pS2 += (sBpl2 << cLogFrameBits) + cFrameBytes * depth;
    pD += (dBpl << cLogFrameBits) + cFrameBytes * depth;

    for (i = 0; i < h; i++)
    {
	for (j = 0; j < w; j += 4)
	{
	    w1 = ULONG_ACCESS (pS1 + j);
	    w2 = ULONG_ACCESS (pS2 + j);
	    h1 = w1 & hBits;
	    h2 = w2 & hBits;

	    /* notBorrowed has 1 in each high bit if the low bits did
	       not cause a borrow, i.e. lowbits1 >= lowbits2 for each pixel */
	    notBorrowed = ((w1 | hBits) - (w2 & lBits)) & hBits;

	    /* Create high mask bits. A 1 in the high bit means pix1 > pix2 */
	    wMask = (~(h1 ^ h2) & notBorrowed) | (h1 & ~h2);

	    /* Replicate the high bit, to make a mask */
	    wMask |= wMask >> 1;
	    wMask |= wMask >> 2;
	    wMask |= wMask >> 4;

	    /* Mask out the pixels for the result.*/
	    ULONG_ACCESS (pD + j) = (w1 & (~wMask)) | (w2 & wMask);
	}

	pS1 += sBpl1;
	pS2 += sBpl2;
	pD += dBpl;
    }

    NORMAL_RETURN;
}

/*
 * i_max
 *
 * Take the pixel-wise maximum of two images. This is "approximate", 
 * i.e. it might give the wrong answer on 1-bit differences.
 *
 * Arguments:
 *	pS1	    -	    First image
 *	sBpl1	    -	    Bytes/line of first image
 *	pS2	    -	    Second image
 *	sBpl2	    -	    Bytes/line of second image
 *	w	    -	    Width of image (w/frame)
 *	h	    -	    heigth of image (w/frame)
 *	depth	    -	    Depth of images
 *	pD	    -	    Destination
 *	dBpl	    -	    Destination bytes/line
 *
 * Returns:
 *	ia_successful, ...
 */

Int32 CDECL 
i_max (
    IPState    *pIPState,
    UInt8      *pS1,
    Int32	sBpl1,
    UInt8      *pS2,
    Int32	sBpl2,
    Int32	w,
    Int32	h,
    Int32	depth,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    Int32 i, j;
    register UInt32  wMask, w1, w2, h1, h2, notBorrowed, hBits, lBits;
	    
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

    switch (depth)
    {
    case 8:
	hBits = 0x80808080;
	lBits = 0x7f7f7f7f;
	break;

    case 4:	
	hBits = 0x88888888;
	lBits = 0x77777777;
	w >>= 1;
	break;

    case 2:
	hBits = 0xaaaaaaaa;
	lBits = 0x33333333;
	w >>= 2;
	break;

    default:
	return ia_depthNotSupported;
    }

    pS1 += (sBpl1 << cLogFrameBits) + cFrameBytes * depth;
    pS2 += (sBpl2 << cLogFrameBits) + cFrameBytes * depth;
    pD += (dBpl << cLogFrameBits) + cFrameBytes * depth;

    for (i = 0; i < h; i++)
    {
	for (j = 0; j < w; j += 4)
	{
	    w1 = ULONG_ACCESS (pS1 + j);
	    w2 = ULONG_ACCESS (pS2 + j);
	    h1 = w1 & hBits;
	    h2 = w2 & hBits;

	    /* notBorrowed has 1 in each high bit if the low bits did
	       not cause a borrow, i.e. lowbits1 >= lowbits2 for each pixel */
	    notBorrowed = ((w1 | hBits) - (w2 & lBits)) & hBits;

	    /* Create high mask bits. A 1 in the high bit means pix1 > pix2 */
	    wMask = (~(h1 ^ h2) & notBorrowed) | (h1 & ~h2);

	    /* Replicate the high bit, to make a mask */
	    wMask |= wMask >> 1;
	    wMask |= wMask >> 2;
	    wMask |= wMask >> 4;

	    /* Mask out the pixels for the result.*/
	    ULONG_ACCESS (pD + j) = (w2 & (~wMask)) | (w1 & wMask);
	}

	pS1 += sBpl1;
	pS2 += sBpl2;
	pD += dBpl;
    }

    NORMAL_RETURN;
}

/***********************************************************************
 *
 * i_range 
 * 
 *	Compute the lowest and highest values in an image 
 *
 * Arguments:
 *	pImage		-	Pointer to frame of image
 *	bpl		-	Bytes/line
 *	w, h		-	Size of image, including frame
 *	depth		-	Depth of image data
 *	pLowest		-	[output] results in lowest value
 *	pHighest	-	[output] results in highest value
 *
 * Returns:
 *	ia_successful or error code
 *
 ***********************************************************************/

Int32 CDECL 
i_range (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	bpl,
    Int32	w,
    Int32	h,
    Int32	depth,
    UInt8      *pLowest,
    UInt8      *pHighest)
{
    PROLOGUE;
    UInt8   lowest, highest, pixel;
    Int32   i, j;

    if (depth != 4 && depth != 8)
	return ia_depthNotSupported;

    /* Move past the frame */
    pImage += (bpl << cLogFrameBits) + cFrameBytes * depth;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

#define RANGE_PIXEL(p, lowest, highest)\
	if (p < lowest) lowest = p;\
	if (p > highest) highest = p;

    if (depth == 8)
    {
	lowest = 255;
	highest = 0;
	for (i = 0; i < h; i++)
	{
	    /* Do all but the ones at the end in either byte order */
	    for (j = 0; j < (w & ~3); j++)
	    {  
		pixel = pImage[j];
		RANGE_PIXEL (pixel, lowest, highest);
	    }

	    /* Do the rest */
	    for ( ; j < w; j++)
	    {
		pixel = UCHAR_ACCESS (pImage + j);
		RANGE_PIXEL (pixel, lowest, highest);

	    }

	    pImage += bpl;
	}
    }
    else
    {
	Int32 byte, byteWidth = w / 2;
	lowest = 15;
	highest = 0;

	for (i = 0; i < h; i++)
	{
	    /* Do the ones inside words */
	    for (j = 0; j < (byteWidth & ~3); j++)
	    {
		byte = pImage[j];
		pixel = (UInt8)(byte >> 4);
		RANGE_PIXEL (pixel, lowest, highest);
		pixel = byte & 0xf;
		RANGE_PIXEL (pixel, lowest, highest);
	    }

	    /* Do the rest of the bytes */
	    for (; j < byteWidth; j++)
	    {
		byte = UCHAR_ACCESS (pImage + j);
		pixel = (UInt8)(byte >> 4);
		RANGE_PIXEL (pixel, lowest, highest);
		/* Only do the last one if the width is even */
		if ((w & 1) == 0)
		{
		    pixel = byte & 0xf;
		    RANGE_PIXEL (pixel, lowest, highest);
		}
	    }

	    pImage += bpl;
	}
    }

    *pLowest = lowest;
    *pHighest = highest;

    NORMAL_RETURN;
}

IP_RCSINFO(RCSInfo, "$RCSfile: i_minmax.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");
/*
 * $Log: i_minmax.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 * Imported sources
 *
 * Revision 1.7  1998/01/13 23:48:50  ddavies
 * i_minmax.c -> /product/ipcore/ipshared/src/RCS/i_minmax.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.6  1996/02/10 01:43:06  dferrell
 * dferrell on Fri Feb 9 17:42:45 PST 1996
 *
 * i_minmax.c -> /product/ipcore/ipshared/src/RCS/i_minmax.c,v
 *
 * Added parens around & operation so that gcc 2.7.2 won't whine.
 *
 * Revision 1.5  1995/12/13 02:44:47  ejaquith
 * i_minmax.c -> /product/ipcore/ipshared/src/RCS/i_minmax.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.4  1995/10/13  23:50:59  dferrell
 * i_minmax.c -> /product/ipcore/ipshared/src/RCS/i_minmax.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.3  1994/11/04  03:20:23  macleod
 * macleod on Thu Nov 3 19:20:13 PST 1994
 *
 * i_minmax.c -> /product/ipcore/ipshared/src/RCS/i_minmax.c,v
 *
 * Added i_range, for computing the (min,max) range of image data
 *
 * Revision 1.2  1994/07/29  01:34:36  macleod
 * i_minmax.c -> /product/ipcore/ipshared/src/RCS/i_minmax.c,v
 *
 * Changed i_min and i_max. First, the low bit is no longer trashed,
 * and second, support was added for 4- and 2-bit/pixel images.
 * I haven't actually tested it at 2-bit/pixel (all depths are done
 * by the same routine)
 *
 * Revision 1.1  1994/07/27  18:22:23  macleod
 * Initial revision
 *  
 */
