/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_mmrc.c: High level code used in Modified Modified Read
 *	      (CCITT Group 4) compression.
 *
 *   	Int32	i_compressMMRRows()
 *			Compresses a block of scan lines using
 *			the Modified Modified Read algorithm; also known as
 *			the CCITT Group 4, or CCITT Fascicle VII.3,
 *			Recommendation T.6 algorithm.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_mmrc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* There is a big comment all about the low level mechanics of our
 * implementation of MMR compression in mimmrc.c.  This comment will
 * deal with software engineering and interfaces on the next higher
 * level.  The basic style of interfaces is shared by the other CCITT
 * compression routines and the PackBits compression routine.
 * 
 * MMR encoding uses endpoints on two scan lines to do the encoding.
 * The "coding" line is the line being encoded.  The "reference" line is
 * the one above the coding line.  There are two run arrays in the
 * encoder state block.  We ping-pong between them; filling one and
 * using it for the coding line, then filling the other and using the
 * first for the reference line, then filling the first and using the
 * second for the reference line and so on.  The run2ArrayIsCoding
 * boolean in the compression data structure indicates whether the
 * second run array is the coding line (cTrue) or the reference line
 * (cFalse).  In MH compression, it's always false.
 *
 * The procedure i_findAllRunsInLineArray is used to find all the
 * run endpoints for the new coding line.  We assume that all pixels
 * are in pixr order so i_findAllRunsInLineArray works right.
 * Each run array starts with a dummy element that has left and right
 * coords at -1.
 *
 * After the run endpoints for a line are found, we call ip_compMMRRow
 * to encode the run endpoints into MMR codes.
 *
 * Note that the encoded data stream is always written out in the
 * platform's natural byte order.  It will have to show up in that
 * order at some point and (theoretically) there aren't very many
 * encoded bytes, so this is no problem, right?
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which this procedure will stop.  It might actually
 * finish encoding the entire scan line.  When this happens, we update the
 * pointer to the compressed data that was passed in, set the state
 * to cCCITTCompNewLine and store the current accumulator and vacantBits
 * and return happily.
 *
 * The other reason for returning is that we run out of room for
 * compressed data.  ip_compMMRRow detects this by comparing pCurCompData to a
 * pointer to the end of the buffer just before writing encoded data.
 * Since we want to pretend that we have a callback to fill the buffer,
 * we need a restart location next to every check.  There are
 * three checks when writing codes for white runs, and three checks
 * when writing black runs.  That's 6 states in all.
 *
 * We always start the run array with a dummy element that says there is
 * a black run whose left and right endpoints are at -1.  We know we can
 * always point at this run.  We quit encoding a line when the current
 * position is off the right edge of the image.
 *
 *			*	*	*
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * and the continueImageBlock and continueCompDataStream args.
 * The caller also supplies pointers to words used to return the
 * exhaustedImageData, exhaustedCompData, bytesWritten and percentDone args.
 * The bytesWritten arg returns the number of compressed bytes written so far.
 *
 * We do not supply a flushPipeline argument.  That's because the
 * compressed data for every MMR scan line should be followed
 * immediately by the data for the next scanline or by two EOLs (the
 * EndOfFacsimileBlock code).  It's possible that the buffer may
 * fill in the midst of writing these.  It's a real pain to figure out
 * how far we've gotten when trying to restart, so we just don't bother.
 * We recommend that the client use i_writeCCITTEOL and i_flushCCITT,
 * with proper attention paid to detecting a full compression buffer.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	This means we compressed the desired number of
 *	image lines without filling the compressed data buffer.
 *	The routine can be restarted with new startLine (and
 *	numLines), continueImageBlock = cFalse.  
 *	Some of the compressed data may still be in the
 *	pipeline and the saved compressed data state is valid.
 *	This routine may be called again with
 *	continueCompDataStream = cTrue.  The client must supply
 *	a new value for pCompData and compDataSize.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to fill the
 *	compressed data buffer at the same time the compressed
 *	data came to an end.  As before, there is probably be more
 *	compressed data in the pipeline.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of room for compressed data before writing the last
 *	byte of compressed data requested.  Some of the data to be written
 *	is in the pipeline.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */

/* This is a big number used to mark the right coord of the second dummy
 * run at the end of each scan line.
 */
#define	cDummyMarker	0x7FFFFFFF

/* PUBLIC */
/**************************************************************************
 *   i_compressRowsMMR()
 *
 *	This procedure compresses images in units of scan lines
 *	using the Modified Modified Read (CCITT Group 4) algorithm.
 *
 *	Input:
 *		pCCITTCompInfoIn:	Pointer to CCITT compression info
 *		pImage:			Pointer to the upper left corner
 *					of the frame of the uncompressed
 *					image
 *		w:			Width of uncompressed image including
 *					frame.
 *		startLine:		First image line to be compressed
 *					relative to the top of the frame
 *		numLines:		Number of lines to be compressed
 *		bpl:			bytes/line of uncompressed data.
 *		pCompData:		Pointer to first vacant byte
 *					of buffer into which compressed
 *					data will be placed.
 *		compDataBufSize:	Number of vacant bytes remaining
 *					in the compressed data buffer.
 *		continueImageBlock:	cTrue => we are compressing
 *					from the same slice of image as was
 *					specified in the previous call and the
 *					saved state in pCCITTCompInfoIn is
 *					valid.
 *					This arg is set to cFalse if either
 *					we are starting with all new data
 *					or the previous call returned because it
 *					ran out of image data.
 *		continueCompDataStream:	cTrue => The saved compression state
 *					in pCCITTCompInfoIn is valid.  If
 *					the previous call to this routine
 *					filled the compressed data buffer
 *					before the strip, tile or image was
 *					complete, a new buffer should be
 *					supplied.  This arg is set to 
 *					cFalse when the block of compressed
 *					data is unrelated to the block used
 *					in the previous call to this routine
 *					as recorded in pCCITTCompInfoIn.
 *					SETTING THIS ARG TO cFalse CAUSES
 *					DATA IN THE PIPELINE TO BE DISCARDED!
 *		pExhaustedImageData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because specified
 *					number of lines have been processed.
 *					This may require more than one
 *					call to this routine.
 *		pExhaustedCompData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because it filled
 *					the compressed data buffer.  This
 *					routine may be called again with
 *					a new buffer.
 *		pPercentDone:		Pointer to returned arg.  When
 *					routine returns with *pExhaustedCompData
 *					set, this arg specifies the rough
 *					percentage of the image block that
 *					has been compressed.  It is
 *					intended to be used to estimate the
 *					amount of compressed data needed
 *					to complete the task.  The result
 *					ranges from 0 to 100.
 *		pBytesWritten:		Number of compressed bytes actually
 *					written to the compressed data
 *					block.  If (*pExhaustedImageData ==
 *					cFalse and *pExhaustedCompData == 
 *					cTrue), some bytes required to complete
 *					a scan line may still be in the
 *					pipeline and are not included in
 *					this count.
 *
 *	Action:
 *		Compresses a horizontal slice image using the
 *		Modified Modified Read (CCITT Group 4) method.
 *		See TIFF 6.0 Specification, Section 11.
 *
 *		There are several completion conditions:
 *
 *	__error__	!= ia_successful
 *	*pExhaustedImageData == don't care
 *	*pExhaustedCompData == don't care
 *		This means something catastrophic happened.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cFalse
 *		This means we compressed the desired number of
 *		image lines without filling the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.  Some of the
 *		compressed data may still be in the pipeline and the 
 *		saved compressed data state is valid.
 *		This routine may be called again with
 *		continueCompDataStream = cTrue.  The client must supply 
 *		a new value for pCompData and compDataSize.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cTrue
 *		This means we filled the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is unusual, but not impossible.  Some data probably
 *		remains in the pipeline.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cFalse
 *	*pExhaustedCompData == cTrue
 *		The routine filled the compression data buffer before completing
 *		the specified number of lines.  One way this can
 *		happen is if all image data has been processed and
 *		some data remains in the pipeline.
 *		When this condition does occur, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  
 *		The image may be moved around as a block after this
 *		condition has been returned, but the internal
 *		composition of the image (and startLine and numLines) 
 *		should not be changed.
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *		
 *		The image data is always in PIXR order.
 *		
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		a pointer was NULL or 
 *					(continueImageBlock && the saved image
 *					info is not valid) or
 *					(continueCompDataStream && the saved
 *					compression info is not valid)
 *					or padType is invalid
 *
 **************************************************************************/
Int32  CDECL 
i_compressRowsMMR (
    IPState    *pIPState,
    void       *pCCITTCompInfoIn,
    UInt8      *pImage,
    Int32	w,
    Int32	startLine,
    Int32	numLines,
    Int32	bpl,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesWritten)
{
    PROLOGUE;
    CCITTCompressionInfo	*pCCITTCompInfo;
    UInt8			*pCompDataEnd;	/* address 1 beyond the last
						 * byte of compressed data */
    UInt8			*pCurCompData;	/* current compressed data
						 * pointer */
    Run32			*pCodingRuns;	/* Pointer to runs for coding
						 * line. */
    Run32			*pRefRuns; 	/* pointer to runs for reference
						 * line in array of runs. */
    Run32			*pDummy;	/* returned ptr to element
						 * just beyond last valid
						 * run in the array */
    UInt32			 remainingLines;
    

	/* check for null pointers */
    if ((pCCITTCompInfoIn == NULL) || (pCompData == NULL) ||
	(pImage == NULL) || (compDataBufSize == 0) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesWritten == NULL))
	return ia_invalidParm;

	/* Strip off the frame */
    pImage += (bpl << cLogFrameBits) + cFrameBytes;
    startLine -= cFrameBits;
    w -= 2*cFrameBits;
    if (w <= 0)
	NORMAL_RETURN;	/* there's no image here */

	/* load the local copies of the remote pointers */
    pCCITTCompInfo = (CCITTCompressionInfo *) pCCITTCompInfoIn;

	/* Initialize the state from the saved state passed in *pCCITTCompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	pImage += startLine * bpl;
	pCCITTCompInfo->lineNumber = 0;
	pCCITTCompInfo->run2ArrayIsCoding = cFalse;
	pCCITTCompInfo->currentRun = 0;
	pCCITTCompInfo->refRun = 0;
	pCCITTCompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pCCITTCompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pCCITTCompInfo->lineNumber;
	pImage += (startLine + pCCITTCompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pCCITTCompInfo->compState = cCCITTCompNewLine;
	pCCITTCompInfo->accumBits = 0;
	pCCITTCompInfo->vacantBit = 0;
	pCCITTCompInfo->remainingLength = 0;
	pCCITTCompInfo->compInfoIsValid = cTrue;
    }
    else
    if (!pCCITTCompInfo->compInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Set the run pointers to the second runs.  The first is a dummy. */
    if (pCCITTCompInfo->run2ArrayIsCoding)
    {
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
					  pCCITTCompInfo->run2ArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				       pCCITTCompInfo->runArrayOffset +
				       sizeof(Run32));
    }
    else
    {
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
					  pCCITTCompInfo->runArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				       pCCITTCompInfo->run2ArrayOffset +
				       sizeof(Run32));
    }

	/* If we're not continuing from a previous image block,
	 * set the reference line to be "blank" by just putting
	 * in the two dummies at the end of its line.
	 */
    if (!continueImageBlock)
    {
	pRefRuns[0].runLeft = w;
	pRefRuns[0].runRight = w-1;
	pRefRuns[1].runLeft = w;
	pRefRuns[1].runRight = cDummyMarker;
    }

	/* Compress each scan line.
	 * This is done in several steps:
	 *   1. If the compression state does not equal cCCITTCompNewLine,
	 *	jump to step 3.
	 *   2. We did not quit in the middle of putting out this scan
	 *	line.  Find all the runs for the current coding line.
	 *   3. Call ip_compressMMRRow to put out the codes for the runs
	 *	of the scan line.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--, pImage += bpl)
    {
	if (pCCITTCompInfo->compState == cCCITTCompNewLine)
	{
		/* Find the run boundaries and put them into the array
		 * of runs.
		 */
	    __error__ =
	        i_findAllRunsInLineArray (pIPState,
					  (UInt8 *)pImage, /* inside "frame" */
				          w-1,		 /* right limit */
				          0,		 /* left limit */
				          pCodingRuns,	 /* dest array */
				          (w+1)/2,	 /* # of Run32s */
				          0,		 /* add this to each
							  * run coord */
					  &pDummy);	 /* Address of run
							  * just past last
							  * valid run */
	    if (__error__ != ia_successful)
		NORMAL_RETURN;

	        /* Mark the two dummy elements at the end.  The left coord
		 * of the first dummy should be just beyond the right
		 * edge of the image.  The right coord of the first
		 * dummy is back at the right edge of the image.  This
		 * very strange setting turns out to make things work
		 * out at the right side of the image (white runs end
		 * at the edge properly, black runs end with a vertical
		 * offset of 0 properly, the dummy run has a width of
		 * 0 properly).
		 * The second dummy has a left coord of w again, just
		 * to make sure we're off the image, and a right coord
		 * of a very big number.
		 */
	    pDummy[0].runLeft = w;
	    pDummy[0].runRight = w-1;
	    pDummy[1].runLeft = w;
	    pDummy[1].runRight = cDummyMarker;

		/* Just to be extra safe, set the run offset in
		 * the compression info to 0 so that we start
		 * interpreting the very first run and not some
		 * run left over from some previous stoppage.
		 * I'm not really sure when this does anything
		 * that wasn't already done, but it seems tidy.
		 */
	    pCCITTCompInfo->currentRun = 0;
	    pCCITTCompInfo->refRun = 0;
	}

	    /* Ask the compression code to translate runs into the
	     * compressed data stream.  The width sent down is used
	     * to decide when we've coded the whole line.
	     */
	__error__ = ip_compMMRRow (pIPState, pCCITTCompInfo, w,
			         &pCurCompData, pCompDataEnd);
	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;

	    /* We have now attempted to compress one more scan line,
	     * possibly not the last one.  Should we stop now?
	     * If we stopped in the middle of a line because the
	     * compressed data buffer was full and we had more stuff
	     * to write, or we finished the line but just happened
	     * to fill the compressed data buffer and there are
	     * more lines to go, stop now.
	     */
	if ((__error__ == ia_bufferExceeded) ||
	    ((pCurCompData == pCompDataEnd) && (remainingLines > 1)))
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * ((pCCITTCompInfo->lineNumber) * w)) /
			    (numLines * w);
	    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

		/* if we happened to stop on an image scan line boundary,
		 * increment the saved line number so we don't do it again.
		 * Also swap the coding and reference runs.
		 */
	    if (pCCITTCompInfo->compState == cCCITTCompNewLine)
	    {
		pCCITTCompInfo->lineNumber++;
		pCCITTCompInfo->run2ArrayIsCoding ^= cTrue;
	    }
	    __error__ = ia_successful;	/* this is the convention if we get this
					 * far.  It means we didn't encounter a
					 * real error, though we may have filled
					 * output buffer.  The caller detects
					 * that by looking at
					 * *pExhaustedCompData.
					 */
	    NORMAL_RETURN;
	}
	    /* increment the current line number in case we need
	     * to stop in the next line.  Also swap the coding
	     * and reference runs.
	     */
	pCCITTCompInfo->lineNumber++;
	pCCITTCompInfo->run2ArrayIsCoding ^= cTrue;
	pDummy = pCodingRuns;
	pCodingRuns = pRefRuns;
	pRefRuns = pDummy;
    }

	/* We've done all the scan lines.  So, just
	 * say things were successful and return.
	 */
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;

	/* Return the number of bytes written, including any
	 * bytes flushed.
	 */
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

    NORMAL_RETURN;

}


/*

 $Log: i_mmrc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.4  1998/01/13 23:48:51  ddavies
 i_mmrc.c -> /product/ipcore/ipshared/src/RCS/i_mmrc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1996/05/06 18:48:03  ddavies
 i_mmrc.c -> /product/ipcore/ipshared/src/RCS/i_mmrc.c,v

 Improve lots of comments.  Remove the flushPipeline argument from
 i_compressRowsMMR.  After some
 consideration, I decided that you never want to just flush the pipeline
 after compressing an MMR row.  The compressed data for a one of these
 rows is *always* immediately followed by another row, an EndOfLine code
 or an EndOfFacsimile code.
 Also remove the half-hearted attempt to take action when flushPipeline
 was set.

 Revision 1.2  1996/05/01 23:25:06  ddavies
 i_mmrc.c -> /product/ipcore/ipshared/src/RCS/i_mmrc.c,v

 Yow, those original checkin messages were what happens when you hit the
 "paste" key by accident.  This is a new file which contains the procedure
 i_compressRowsMMR.  This routine compresses some scan lines of an
 ipshared-style image using MMR (CCITT Group 4).  It's about 3.5x faster
 than the Leffler equivalent (on the SPARC, ~2x-3x faster on PC).
 This uses the same style of interface used by i_compressRowsMH and
 i_compressRowsPackBits (and soon - i_compressRowsMR).  The routine
 stops when either it completes the compression task or it fills the
 output buffer.  It can be restarted in either case.

 Revision 1.1  1996/05/01 23:15:27  ddavies
 Initial revision


 */


