/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_mmrd.c: High level code used in Modified Modifed Read decompression.
 *
 *   	Int32	i_decompressRowsMMR()
 *			Decompresses a block of scan lines using
 *			the Modified Modified Read algorithm; also known as
 *			the CCITT Group 4, or CCITT Fascicle VII.3,
 *			Recommendation T.6 algorithm.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_mmrd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* There is a big comment all about the low level mechanics of our
 * implementation of MMR decompression in mimmrd.c.  This comment will
 * deal with software engineering and interfaces on the next higher
 * level.  The basic style of interfaces is shared by the other CCITT
 * compression routines and the PackBits compression routine.
 * 
 * MMR encoding uses endpoints on two scan lines to do the encoding.
 * The "coding" line is the line being encoded.  The "reference" line is
 * the one above the coding line.  There are two run arrays in the
 * encoder state block.  We ping-pong between them; filling one and
 * using it for the coding line, then filling the other and using the
 * first for the reference line, then filling the first and using the
 * second for the reference line and so on.  The run2ArrayIsCoding
 * boolean in the compression data structure indicates whether the
 * second run array is the coding line (cTrue) or the reference line
 * (cFalse).  In MH decompression, it's always false.
 *
 * Swell.  So how does the decoding work?  It's pretty much the inverse
 * of the encoding routine.  For each line, we call ip_decompMMRRow
 * to find all the run endpoints using one of run arrays as the reference
 * line and the other as the new coding line.  Then we call ip_paintRunsLine
 * to paint the new coding line runs found into the bitmap.
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which the lower level procedure will stop.  It
 * might actually finish decoding one complete scan line.  When this
 * happens, update the pointer to the compressed data that was passed
 * in, store the internal decoding state into the state structure
 * and return happily.
 *
 * The other reason for returning is that we run out of compressed data.
 * The ip_decompMMRRow routine detects this by comparing
 * its pCurCompData to a pointer to the end of the buffer just before
 * reading compressed data. Since we want to pretend that we have a
 * callback to fill the buffer, we need a restart location next to
 * every check.  There are six checks when reading codes, one at the
 * top of the loop that decodes left endpoints, one at the top of the
 * loop that decodes right endpoints and two inside each of these loops.
 * The inner checks are done when decoding horizontal codes.
 *
 *			*	*	*
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * and the continueImageBlock, continueCompDataStream and flushPipeline args
 * discussed above.  The caller also supplies pointers to words used to
 * return the exhaustedImageData, exhaustedCompData, bytesWritten and
 * percentDone args.  The bytesRead arg returns the number of compressed
 * bytes read so far.  If flushPipeline is not set, some of the compressed
 * bytes may have been read, but not yet decoded.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.  One possibly catastrophe is
 *	that flushPipeline was set but insufficient compressed data
 *	was supplied.  In this case, we return ia_bufferExceeded.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	We have processed all the image data requested.  The compression
 *	buffer is not empty.  Some of the compressed data bytes are
 *	still in the pipeline regardless of the state of flushPipeline.
 *	The number of bytes actually read from the buffer
 *	(including any bytes still in the pipeline) are returned.
 *	When the decompression routine is called next, the client must
 *	pass in a pointer to the byte holding the next compressed byte.
 *	This may be just beyond the end of the current compressed data,
 *	at the beginning of the same buffer, or in another buffer entirely.
 *	The size of the compressed data buffer must also be (re)calculated
 *	by the client.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to read the last
 *	byte of the compressed data buffer at the same time the compressed
 *	data came to an end.  As before, if flushPipeline was false, there
 *	may be more compressed data in the pipeline.  Normally, this
 *	case occurs when exactly the right amount of compressed data was
 *	supplied and flushPipeline was true.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of compressed data before reading the last
 *	byte of compressed data requested.  Some of the data read
 *	is in the pipeline.  This can only happen if flushPipeline was
 *	false.  If flushPipeline is true, we keep going until we either
 *	finish all the scan lines or get an ia_bufferExceeded.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */

/* This is a big number used to mark the right coord of the second dummy
 * run at the end of each scan line.
 */
#define	cDummyMarker	0x7FFFFFFF

/* PUBLIC */
/**************************************************************************
 *   i_decompressRowsMMR()
 *
 *	This procedure decompresses Modified Huffman encoded images in
 *	units of scan lines.
 *
 *	Input:
 *		pCCITTDecompInfoIn:	Pointer to CCITT decompression info
 *		pImage:			Pointer to the upper left corner
 *					of the frame of the uncompressed
 *					image.  If NULL, no runs are painted.
 *		startLine:		First image line to be decompressed
 *					relative to the top of the frame
 *		numLines:		Number of lines to be decompressed
 *		bpl:			bytes/line of uncompressed data.
 *		combineResults:		If set to cFalse, the image is assumed
 *					to be initially clear.  This is most
 *					efficient.  If set to cTrue, the
 *					decoded image is ORed into the original.
 *		pCompData:		Pointer to next byte of compressed
 *					data to be read.  If several calls
 *					to this routine are needed to process
 *					a block of compressed data, this
 *					pointer must be updated on each call.
 *		compDataBufSize:	Number of bytes of compressed data
 *					remaining in the buffer.
 *		continueImageBlock:	cTrue => we are decompressing
 *					into the same slice of image as was
 *					specified in the previous call and the
 *					saved state in pCCITTDecompInfoIn is
 *					valid.
 *					This arg is set to cFalse if either
 *					we are starting with all new data
 *					or the previous call returned because it
 *					ran out of image data.
 *		continueCompDataStream:	cTrue => The saved decompression state
 *					in pCCITTDecompInfoIn is valid.  This
 *					allows the procedure to treat several
 *					buffers of compressed data as though
 *					it were one contiguous buffer.
 *					If set to cFalse, this routine will
 *					discard all previous compression
 *					state.
 *		flushPipeline:		If this arg is set, this routine
 *					will attempt to complete the
 *					specified number of scan lines
 *					WITHOUT stopping when it starts to
 *					run out of data; i.e. it continues
 *					until its pipeline is completely
 *					empty.  If this is not possible,
 *					an ia_bufferExceeded error is raised.
 *		pExhaustedImageData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because specified
 *					number of lines have been processed.
 *					This may require more than one
 *					call to this routine.
 *		pExhaustedCompData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because it reached
 *					the end of the compressed data buffer
 *					as specified by compDataSize.  This
 *					routine may be called again with 
 *					new compression data and the the 
 *					same image.  This arg will not be 
 *					set if flushPipeline is set and we 
 *					failed to decompress the specified 
 *					number of scan lines (routine returns
 *					ia_bufferExceeded instead).
 *		pPercentDone:		Pointer to returned arg.  When
 *					routine returns with *pExhaustedCompData
 *					set, this arg specifies the rough
 *					percentage of the image block that
 *					has been decompressed.  It is
 *					intended to be used to estimate the
 *					amount of compressed data needed
 *					to complete the task.  The result
 *					ranges from 0 to 100.
 *		pBytesRead:		Pointer to returned arg.  Contains
 *					number of compressed data bytes
 *					read by the routine before it returned.
 *					The client may safely dispose of
 *					these bytes.
 *
 *	Action:
 *		Decompresses a horizontal slice of MMR data.
 *		There are several completion conditions:
 *
 *	__error__	!= ia_successful
 *	*pExhaustedImageData == don't care
 *	*pExhaustedCompData == don't care
 *		This means something catastrophic happened.  One possible
 *		catastrophe (ia_bufferExceeded) happens if the client
 *		set flushPipeline and failed to supply enough compressed
 *		data.		
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cFalse
 *		This means we produced the desired number of uncompressed
 *		image lines without running out of compressed data.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse and
 *		continueCompDataStream = cTrue.  pCompData and compDataSize
 *		should be adjusted to account for the number of 
 *		compressed data bytes that have already been read (returned
 *		in *pBytesRead).
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cTrue
 *		This means we ran out of compressed data at the same time
 *		we completed the correct number of image lines.  This
 *		is the normal condition at the bottom of a TIFF slice or
 *		the bottom of the image.  It is not possible to restart
 *		from this state, though it is possible to decompress
 *		with new image and compression data.  When the client
 *		expects to run out of compressed data, the flushPipeline
 *		argument should be set.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cFalse
 *	*pExhaustedCompData == cTrue
 *		The routine ran out of compressed data before completing
 *		the specified number of lines.  This condition will not
 *		occur if flushPipeline was set.  If flushPipeline was set
 *		and we ran out of compressed data before completing the
 *		the task, an ia_bufferExceeded error will be returned.
 *		When this condition does occur, the routine may be
 *		restarted by supplying more compressed data, setting
 *		BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  Note that the first byte of the new compressed
 *		data must be the one immediately following the last
 *		byte of old compressed data.
 *		The image may be moved around as a block if necessary, but
 *		the internal composition of the image (and startLine
 *		and numLines) should not be changed.  When this
 *		condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much compression data is 
 *		needed to complete the slice.
 *		
 *		The image data is always in PIXR order.
 *		
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		a pointer was NULL or 
 *					(continueImageBlock && the saved image
 *					info is not valid) or
 *					(continueCompDataStream && the saved
 *					compression info is not valid)
 *		ia_bufferExceeded	ran out of compressed data when
 *					flushPipeline was set.
 *		ia_decodeFailure	an illegal code was enountered.
 *
 **************************************************************************/
Int32  CDECL 
i_decompressRowsMMR (
    IPState    *pIPState,
    void       *pCCITTDecompInfoIn,
    UInt8      *pImage,
    Int32	startLine,
    Int32	numLines,
    Int32	bpl,
    Bool	combineResults,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Bool	flushPipeline,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesRead)
{
    PROLOGUE;
    CCITTDecompressionInfo	*pCCITTDecompInfo;
    UInt8			*pCompDataEnd;	/* address 1 beyond the last
						 * byte of compressed data */
    UInt8			*pCurCompData;	/* current compressed data
						 * pointer */
    Run32			*pCodingRuns;	/* pointer to first
						 * coding run in
						 * pCCITTDecompInfo */
    Run32			*pRefRuns;	/* pointer to first
						 * reference run in
						 * pCCITTDecompInfo */
    Run32			*pDummy;	/* returned ptr to element
						 * just beyond last valid
						 * run in the array */
    Int32			 remainingLines;
    Int32			 imWidth;	/* width of unframed image */
    

	/* check for null pointers */
    if ((pCCITTDecompInfoIn == NULL) || (pCompData == NULL) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesRead == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pCCITTDecompInfo = (CCITTDecompressionInfo *) pCCITTDecompInfoIn;

	/* Strip off the frame */
    if (pImage != NULL)
	pImage += (bpl << cLogFrameBits) + cFrameBytes;
    startLine -= cFrameBits;
    imWidth = pCCITTDecompInfo->width;

	/* Initialize the state from the saved state passed in *pCCITTDecompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	if (pImage != NULL)
	    pImage += startLine * bpl;
	pCCITTDecompInfo->lineNumber = 0;
	pCCITTDecompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pCCITTDecompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pCCITTDecompInfo->lineNumber;
	if (pImage != NULL)
	    pImage += (startLine + pCCITTDecompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
	pCCITTDecompInfo->run2ArrayIsCoding = cFalse;
	pCCITTDecompInfo->currentRun = 0;
	pCCITTDecompInfo->refRun = 0;
	pCCITTDecompInfo->accumBits = 0;
	pCCITTDecompInfo->vacantBit = 0;
	pCCITTDecompInfo->remainingLength = 0;
	pCCITTDecompInfo->decompInfoIsValid = cTrue;
    }
    else
    if (!pCCITTDecompInfo->decompInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Set the run pointers to the second runs.  The first is a dummy. */
    if (pCCITTDecompInfo->run2ArrayIsCoding)
    {
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
					  pCCITTDecompInfo->run2ArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				       pCCITTDecompInfo->runArrayOffset +
				       sizeof(Run32));
    }
    else
    {
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
					  pCCITTDecompInfo->runArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				       pCCITTDecompInfo->run2ArrayOffset +
				       sizeof(Run32));
    }

	/* If we're not continuing from a previous image block,
	 * set the reference line to be "blank" by just putting
	 * in the two dummies at the end of its line.
	 */
    if (!continueCompDataStream)
    {
	pRefRuns[0].runLeft = imWidth;
	pRefRuns[0].runRight = imWidth-1;
	pRefRuns[1].runLeft = imWidth;
	pRefRuns[1].runRight = cDummyMarker;
    }

	/* Decompress each scan line.
	 * This is done in several steps:
	 *   1. We might have quit in the middle of decoding all the
	 *	runs in this scan line.  That's the only place we
	 *	might have quit.  Since we're starting with that step
	 *	in any case, just do it using ip_decompMMRRow.
	 *   2. Call ip_paintRunsLine to draw the runs into the bitmap.
	 *	We cannot abort in the middle of this step.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--)
    {
	    /* decode the runs for the next line. */
	__error__ = ip_decompMMRRow (pIPState, pCCITTDecompInfo,
				   imWidth, flushPipeline,
			           &pCurCompData, pCompDataEnd,
				   &pDummy);

	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;
	if ((__error__ == ia_bufferExceeded) && (flushPipeline))
	{
	    pCCITTDecompInfo->imageInfoIsValid = cFalse;
	    pCCITTDecompInfo->decompInfoIsValid = cFalse;
	    NORMAL_RETURN;
	}


	    /* If we didn't get to the end of a scan line, return now
	     * instead of painting all the runs to the bitmap.  We
	     * don't paint the runs because we don't have all of them.
	     * Since flushPipeline was not set, this is all ok.
	     */
	if (__error__ == ia_bufferExceeded)
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * pCCITTDecompInfo->lineNumber) /
			    (numLines);
	    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

	    NORMAL_RETURN;
	}
	else
	{
		/* Mark the two runs beyond the last run in the same
		 * way they were marked during encoding.  This guarantees
		 * that we'll get the same answer when we're searching
		 * through these runs when they become the reference row.
		 */
	    pDummy[0].runLeft = imWidth;
	    pDummy[0].runRight = imWidth-1;
	    pDummy[1].runLeft = imWidth;
	    pDummy[1].runRight = cDummyMarker;
	}

	    /* Great, we got a set of runs.  Go paint them into the
	     * image.  If pImage is NULL, someone just wants the runs
	     * so don't paint them.
	     */
	if (pImage != NULL)
	{
	    __error__ = ip_paintRunsLine (pIPState, pImage, imWidth, pCodingRuns,
				        pDummy, combineResults);
	    if (__error__ != ia_successful)
		NORMAL_RETURN;
	}

	    /* increment the current line number just in case we need
	     * to stop in the next line.  Also swap the coding
	     * and reference runs.
	     */
	pCCITTDecompInfo->lineNumber++;
	pCCITTDecompInfo->run2ArrayIsCoding ^= cTrue;
	pDummy = pCodingRuns;
	pCodingRuns = pRefRuns;
	pRefRuns = pDummy;
	if (pImage != NULL)
	    pImage += bpl;
    }

	/* Great! we completed the right number of scan lines.
	 * Say that things worked and return.
	 */
    pCCITTDecompInfo->run2ArrayMostRecent = 
			    pCCITTDecompInfo->run2ArrayIsCoding ^ cTrue;
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;
    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;


    NORMAL_RETURN;
}



/*

 $Log: i_mmrd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.7  1998/01/13 23:48:53  ddavies
 i_mmrd.c -> /product/ipcore/ipshared/src/RCS/i_mmrd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.6  1997/08/20 20:23:46  ddavies
 i_mmrd.c -> /product/ipcore/ipshared/src/RCS/i_mmrd.c,v

 Remove ia_depthNotSupported in comment, it can't happen.

 Revision 1.5  1996/07/18 23:38:42  ddavies
 i_mmrd.c -> /product/ipcore/ipshared/src/RCS/i_mmrd.c,v

 Change initImage parameter to combineResults parameter.  When combineResults
 is set, the decompressed image is ORed into the original dest image.
 When combineResults is not set, the original dest image is assumed to be
 white (this is often true and is faster).

 Revision 1.4  1996/05/21 18:50:01  ddavies
 i_mmrd.c -> /product/ipcore/ipshared/src/RCS/i_mmrd.c,v

 Remove width parameter, it's passed in the CCITTDecompressionInfo struct.
 Rename of internal width parameter from w to imWidth for readability.
 Do not paint decoded runs into a bitmap if the bitmap pointer is NULL.  This
 allows us to simply extract the runs from a MMR file.  Also update the
 run2ArrayMostRecent var on leaving the routine.

 Revision 1.3  1996/05/16 17:09:01  ddavies
 i_mmrd.c -> /product/ipcore/ipshared/src/RCS/i_mmrd.c,v

 Fix several bugs that are encountered when you repeatedly read a single scan
 line from a large block of compressed data.
 First, delete the code that invalidates the saved state when flushPipeline
 is set.  It is very difficult to determine whether a particular scan line
 is the last scanline of a strip at the level where flushPipeline is set.
 Second, move the code that initializes state variables controlling the run
 data structures from the block controlled by continueImageBlock
 to the block controlled by continueCompDataStream.  In decompression,
 the runs are associated with the compressed data, not with the image.
 Also move some code around for more uniformity with the other CCITT decompression routines.
 Also initialize the runs in the reference row when we're starting a new
 compressed data stream, not when we're starting a new image block.

 Revision 1.2  1996/05/11 01:02:41  ddavies
 i_mmrd.c -> /product/ipcore/ipshared/src/RCS/i_mmrd.c,v

 Fix comments.  Remove debugging code.

 Revision 1.1  1996/05/06 18:41:26  ddavies
 Initial revision


 */


