/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_mrc.c: High level code used in Modified Read
 *	      (CCITT Group 3) compression.
 *
 *   	Int32	i_compressRowsMR()
 *			Compresses a block of scan lines using
 *			the Modified Read algorithm; also known as
 *			the CCITT Group 3, or CCITT Fascicle VII.3,
 *			Recommendation T.4 algorithm.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_mrc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* The wonderful/horrible thing about MR compression is that it has both a
 * pure 1D mode and a hybrid 1D/2D mode.  In 1D mode, all lines are
 * compressed using a 1D compression technique using standard Huffman
 * tables for encoding lengths of alternating white and black runs.
 * In 2D mode, the position of each run endpoint is encoded relative to the
 * endpoint of the corresponding run in the line above it if possible.
 * When there is no reference endpoint close enough, we encode the lengths
 * of two runs.
 *
 * The CCITT standard says that when using hybrid 1D/2D coding, the
 * image should be encoded in blocks of 2 scan lines or 4 scan lines.
 * These correspond to k = 2 and k = 4.  In "k = 2" mode, there's
 * one 1D scan line for each 2D scan line.  In the "k = 4" mode,
 * there's one 1D scan line followed by three 2D scan lines.
 * The mode is selected based on resolution.  For "standard"
 * vertical resolution (97.79 lines per vertical inch), one
 * should use the first mode.  For optional higher vertical
 * resolution(s) (195.58 lines/vertical inch), one should
 * use the second mode.  Leffler's code decided that the first
 * mode was used for resolutions less than or equal to 150 dpi and
 * the second mode for larger vertical resolutions.  The point of
 * all this is that if a receiver gets lost because of transmission
 * errors, it has a hope of resynchronizing before too many scan
 * lines have passed.
 *
 * In MR, each line ends with an EndOfLine code.  There's also one
 * EOL before the first line, so maybe these are BeginningOfLine
 * codes.  In fact, it's easier to think of them this way for
 * two reasons.  First, any padding required to make the EOL end on
 * a byte boundary (an option) is inserted between the EOL and the
 * scan line data *before* it.  This associates the EOL more closely
 * with the following scan line.  The other reason has to do with the
 * extra flag immediately following the EOL.
 *
 * In pure 1D mode, the EndOfLine codes are simply inserted.  In hybrid
 * 1D/2D mode, each EOL is followed by a flag bit indicating whether the
 * next line is encoded using 1D codes or 2D codes.  The bit is a 1
 * when the next line is coded using 1D codes and a 0 when the next line uses
 * 2D codes.  This also makes the EOL act like a BeginningOfLine code.
 *
 * Our routine will start each scan line by encoding optional padding,
 * the EOL code and the optional 1D/2D flag bit.  It then uses a
 * routine to encode the 1D or 2D scan line.
 *
 * In ipshared, 1D compression of a scan line's data is performed
 * by ip_compMHRow.  This is the same primitive routine used in
 * MH compression.  Our 2D compression is supplied by ip_compMMRRow,
 * which is also used in MMR compression.
 *
 *			*	*	*
 *
 * There is a big comment all about the low level mechanics of our
 * implementation of MMR compression in mimmrc.c.  Another big comment
 * in mimhc.c talks about the mechanics of 1D compression.  Since that's
 * all taken care of there, this comment will deal with software
 * engineering and interfaces on the next higher level.  The basic
 * style of interfaces is shared by the other CCITT compression
 * routines and the PackBits compression routine.
 * 
 * MH encoding uses the endpoints of the runs of one scan line to
 * do compression.  MMR encoding uses endpoints on two scan lines.
 * The "coding" line is the line being encoded.  The "reference" line is
 * the one above the coding line.  There are two run arrays in the
 * encoder state block.  We ping-pong between them; filling one and
 * using it for the coding line, then filling the other and using the
 * first for the reference line, then filling the first and using the
 * second for the reference line and so on.  The run2ArrayIsCoding
 * boolean in the compression data structure indicates whether the
 * second run array is the coding line (cTrue) or the reference line
 * (cFalse).  When doing MH compression, we always start over with
 * the first run array, so run2ArrayIsCoding is always set to cFalse.
 *
 * The procedure i_findAllRunsInLineArray is used to find all the
 * run endpoints for the new coding line.  We assume that all pixels
 * are in pixr order so i_findAllRunsInLineArray works right.
 * Each run array starts with a dummy element that has left and right
 * coords at -1.
 *
 * After the run endpoints for a line are found, we write the optional
 * pad, write the EOL and write the optional 1D/2D flag bit, then
 * call ip_compMHRow or ip_compMMRRow to encode the run endpoints.
 *
 * Note that the encoded data stream is always written out in the
 * platform's natural byte order.  It will have to show up in that
 * order at some point and (theoretically) there aren't very many
 * encoded bytes, so this is no problem, right?
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which this procedure will stop.  It might actually
 * finish encoding the entire scan line.  When this happens, we update the
 * pointer to the compressed data that was passed in, set the state
 * to cCCITTCompNewLine and store the current accumulator and vacantBits
 * and return happily.
 *
 * The other reason for returning is that we run out of room for
 * compressed data.  ip_writeCCITTEOL, ip_compMMRRow and ip_compMMRRow
 * all detect this by comparing pCurCompData to a pointer to the end of
 * the buffer just before writing encoded data.  Since we want to
 * pretend that we have a callback to fill the buffer, we need a
 * restart location next to every check.  There are several checks
 * in each routine.  The numbers indicating the restart state are
 * left in the CCITTCompInfo state block.  These numbers are unique for
 * each routine.  When restarting, we first see if we're doing a 1D
 * line or a 2D line.  We then always call the routine that writes an
 * EOL.  If the restart state doesn't belong to it, it just returns.
 * If the restart state does belong to it, it takes up where it left
 * off and sets the restart state to cCCITTCompNewLine.  We then call
 * ip_compMHRow or ip_compMMRRow.  These routines work the same way -
 * if they own the restart state, they restart in the proper spot.
 * All routines recognize cCCITTCompNewLine as the "normal" starting
 * state.
 *
 * We always start the run array with a dummy element that says there is
 * a black run whose left and right endpoints are at -1.  We know we can
 * always point at this run.  We quit encoding a line when the current
 * position is off the right edge of the image.
 *
 *			*	*	*
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * and the continueImageBlock and continueCompDataStream args.
 * The caller also supplies pointers to words used to return the
 * exhaustedImageData, exhaustedCompData, bytesWritten and percentDone args.
 * The bytesWritten arg returns the number of compressed bytes written so far.
 *
 * We do not supply a flushPipeline argument.  That's because the
 * compressed data for every MR scan line should be followed
 * immediately by optional padding and one or six repetitions of
 * EOL:<optional flag bit>.  It's possible that the buffer may
 * fill in the midst of writing these.  It's a real pain to figure out
 * how far we've gotten when trying to restart, so we just don't bother.
 * We recommend that the client use i_writeCCITTEOL and i_flushCCITT,
 * with proper attention paid to detecting a full compression buffer.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	This means we compressed the desired number of
 *	image lines without filling the compressed data buffer.
 *	The routine can be restarted with new startLine (and
 *	numLines), continueImageBlock = cFalse.  
 *	Some of the compressed data may still be in the
 *	pipeline and the saved compressed data state is valid.
 *	This routine may be called again with
 *	continueCompDataStream = cTrue.  The client must supply
 *	a new value for pCompData and compDataSize.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to fill the
 *	compressed data buffer at the same time the compressed
 *	data came to an end.  As before, there is probably be more
 *	compressed data in the pipeline.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of room for compressed data before writing the last
 *	byte of compressed data requested.  Some of the data to be written
 *	is in the pipeline.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */

/* This is a big number used to mark the right coord of the second dummy
 * run at the end of each scan line.
 */
#define	cDummyMarker	0x7FFFFFFF


/* PUBLIC */
/**************************************************************************
 *   i_compressRowsMR()
 *
 *	This procedure compresses images in units of scan lines
 *	using the Modified Read (CCITT Group 3) algorithm.
 *
 *	Input:
 *		pCCITTCompInfoIn:	Pointer to CCITT compression info
 *		pImage:			Pointer to the upper left corner
 *					of the frame of the uncompressed
 *					image
 *		w:			Width of uncompressed image including
 *					frame.
 *		startLine:		First image line to be compressed
 *					relative to the top of the frame
 *		numLines:		Number of lines to be compressed
 *		bpl:			bytes/line of uncompressed data.
 *		pCompData:		Pointer to first vacant byte
 *					of buffer into which compressed
 *					data will be placed.
 *		compDataBufSize:	Number of vacant bytes remaining
 *					in the compressed data buffer.
 *		continueImageBlock:	cTrue => we are compressing
 *					from the same slice of image as was
 *					specified in the previous call and the
 *					saved state in pCCITTCompInfoIn is
 *					valid.
 *					This arg is set to cFalse if either
 *					we are starting with all new data
 *					or the previous call returned because it
 *					ran out of image data.
 *		continueCompDataStream:	cTrue => The saved compression state
 *					in pCCITTCompInfoIn is valid.  If
 *					the previous call to this routine
 *					filled the compressed data buffer
 *					before the strip, tile or image was
 *					complete, a new buffer should be
 *					supplied.  This arg is set to 
 *					cFalse when the block of compressed
 *					data is unrelated to the block used
 *					in the previous call to this routine
 *					as recorded in pCCITTCompInfoIn.
 *					SETTING THIS ARG TO cFalse CAUSES
 *					DATA IN THE PIPELINE TO BE DISCARDED!
 *		pExhaustedImageData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because specified
 *					number of lines have been processed.
 *					This may require more than one
 *					call to this routine.
 *		pExhaustedCompData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because it filled
 *					the compressed data buffer.  This
 *					routine may be called again with
 *					a new buffer.
 *		pPercentDone:		Pointer to returned arg.  When
 *					routine returns with *pExhaustedCompData
 *					set, this arg specifies the rough
 *					percentage of the image block that
 *					has been compressed.  It is
 *					intended to be used to estimate the
 *					amount of compressed data needed
 *					to complete the task.  The result
 *					ranges from 0 to 100.
 *		pBytesWritten:		Number of compressed bytes actually
 *					written to the compressed data
 *					block.  If (*pExhaustedImageData ==
 *					cFalse and *pExhaustedCompData == 
 *					cTrue), some bytes required to complete
 *					a scan line may still be in the
 *					pipeline and are not included in
 *					this count.
 *
 *	Action:
 *		Compresses a horizontal slice image using the
 *		Modified Read (CCITT Group 3) method.
 *		See TIFF 6.0 Specification, Section 11
 *		or CCITT Fascicle VII.3, Recommendation T.4.
 *
 *		There are several completion conditions:
 *
 *	__error__	!= ia_successful
 *	*pExhaustedImageData == don't care
 *	*pExhaustedCompData == don't care
 *		This means something catastrophic happened.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cFalse
 *		This means we compressed the desired number of
 *		image lines without filling the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.  Some of the
 *		compressed data may still be in the pipeline and the 
 *		saved compressed data state is valid.
 *		This routine may be called again with
 *		continueCompDataStream = cTrue.  The client must supply 
 *		a new value for pCompData and compDataSize.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cTrue
 *		This means we filled the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is unusual, but not impossible.  Some data probably
 *		remains in the pipeline.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cFalse
 *	*pExhaustedCompData == cTrue
 *		The routine filled the compression data buffer before completing
 *		the specified number of lines.  One way this can
 *		happen is if all image data has been processed and
 *		some data remains in the pipeline.
 *		When this condition does occur, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  
 *		The image may be moved around as a block after this
 *		condition has been returned, but the internal
 *		composition of the image (and startLine and numLines) 
 *		should not be changed.
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *		
 *		The image data is always in PIXR order.
 *		
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		a pointer was NULL or 
 *					(continueImageBlock && the saved image
 *					info is not valid) or
 *					(continueCompDataStream && the saved
 *					compression info is not valid)
 *					or pCCITTCompInfo->padType is invalid
 *
 **************************************************************************/
Int32  CDECL 
i_compressRowsMR (
    IPState    *pIPState,
    void       *pCCITTCompInfoIn,
    UInt8      *pImage,
    Int32	w,
    Int32	startLine,
    Int32	numLines,
    Int32	bpl,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesWritten)
{
    PROLOGUE;
    CCITTCompressionInfo	*pCCITTCompInfo;
    UInt8			*pCompDataEnd;	/* address 1 beyond the last
						 * byte of compressed data */
    UInt8			*pCurCompData;	/* current compressed data
						 * pointer */
    Run32			*pCodingRuns;	/* Pointer to runs for coding
						 * line. */
    Run32			*pRefRuns; 	/* pointer to runs for reference
						 * line in array of runs. */
    Run32			*pDummy;	/* returned ptr to element
						 * just beyond last valid
						 * run in the array */
    Bool			 encode1DLine;	/* true if we're going to encode
						 * a 1D line */
    UInt32			 remainingLines;
    

	/* check for null pointers */
    if ((pCCITTCompInfoIn == NULL) || (pCompData == NULL) ||
	(pImage == NULL) || (compDataBufSize == 0) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesWritten == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pCCITTCompInfo = (CCITTCompressionInfo *) pCCITTCompInfoIn;

	/* Check the pad type in the structure */
    if ((pCCITTCompInfo->padType != cNoPad) &&
	(pCCITTCompInfo->padType != cPadToByte))
	return ia_invalidParm;

	/* Strip off the frame */
    pImage += (bpl << cLogFrameBits) + cFrameBytes;
    startLine -= cFrameBits;
    w -= 2*cFrameBits;
    if (w <= 0)
	NORMAL_RETURN;	/* there's no image here */

	/* Initialize the state from the saved state passed in *pCCITTCompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	pImage += startLine * bpl;
	pCCITTCompInfo->lineNumber = 0;
	pCCITTCompInfo->run2ArrayIsCoding = cFalse;
	pCCITTCompInfo->currentRun = 0;
	pCCITTCompInfo->refRun = 0;
	pCCITTCompInfo->k = pCCITTCompInfo->maxK;
	pCCITTCompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pCCITTCompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pCCITTCompInfo->lineNumber;
	pImage += (startLine + pCCITTCompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pCCITTCompInfo->compState = cCCITTCompNewLine;
	pCCITTCompInfo->accumBits = 0;
	pCCITTCompInfo->vacantBit = 0;
	pCCITTCompInfo->remainingLength = 0;
	pCCITTCompInfo->compInfoIsValid = cTrue;
    }
    else
    if (!pCCITTCompInfo->compInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Set the run pointers to the second runs.  The first is a dummy. */
    if (pCCITTCompInfo->run2ArrayIsCoding)
    {
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
					  pCCITTCompInfo->run2ArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				       pCCITTCompInfo->runArrayOffset +
				       sizeof(Run32));
    }
    else
    {
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
					  pCCITTCompInfo->runArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				       pCCITTCompInfo->run2ArrayOffset +
				       sizeof(Run32));
    }

	/* If we're not continuing from a previous image block,
	 * set the reference line to be "blank" by just putting
	 * in the two dummies at the end of its line.
	 */
    if (!continueImageBlock)
    {
	pRefRuns[0].runLeft = w;
	pRefRuns[0].runRight = w-1;
	pRefRuns[1].runLeft = w;
	pRefRuns[1].runRight = cDummyMarker;
    }

	/* Compress each scan line.
	 * This is done in several steps:
	 *   1. If the compression state does not equal cCCITTCompNewLine,
	 *	jump to step 3.
	 *   2. We did not quit in the middle of putting out this scan
	 *	line.  Find all the runs for the current coding line.
	 *   3. Call ip_compressMMRRow to put out the codes for the runs
	 *	of the scan line.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--, pImage += bpl)
    {
	    /* this variable is set of we're going to encode a 1D line
	     * this time.
	     */
	encode1DLine = !(pCCITTCompInfo->use2D) ||
			(pCCITTCompInfo->k == pCCITTCompInfo->maxK);

	if (pCCITTCompInfo->compState == cCCITTCompNewLine)
	{
		/* Find the run boundaries and put them into the array
		 * of runs.
		 */
	    __error__ =
	        i_findAllRunsInLineArray (pIPState,
					  (UInt8 *)pImage, /* inside "frame" */
				          w-1,		 /* right limit */
				          0,		 /* left limit */
				          pCodingRuns,	 /* dest array */
				          (w+1)/2,	 /* # of Run32s */
				          0,		 /* add this to each
							  * run coord */
					  &pDummy);	 /* Address of run
							  * just past last
							  * valid run */
	    if (__error__ != ia_successful)
		NORMAL_RETURN;

	        /* Mark either one or two dummy elements at the end.
		 * If this is a 1D run, mark one dummy with
		 * [w,cDummyMarker].  After the line is encoded,
		 * if we're in hybrid 1D/2D mode, mark the last
		 * two elements as though this were a 2D line.
		 * This keeps both lower level routines, ip_compMHRow
		 * and ip_compMMRRow, happy.
		 * If this scan line will be encoded using 2D mode,
		 * mark two dummy runs at the end with [w,w-1] and
		 * [w,cDummyMarker].  This weird combination guarantees
		 * that white runs end at w-1 and are encoded as having
		 * vertical offsets of 0, black runs end at w-1 and
		 * are also encoded as having vertical offsets of 0
		 * and horizontal runs can skip past the [w,w-1] run,
		 * correctly encoded a run length of 0, and still
		 * have a valid run to look at.
		 */
	    if (encode1DLine)
	    {
		pDummy[0].runLeft = w;
		pDummy[0].runRight = cDummyMarker;
	    }
	    else	/* This is a 2D encoding line */
	    {
		pDummy[0].runLeft = w;
		pDummy[0].runRight = w-1;
		pDummy[1].runLeft = w;
		pDummy[1].runRight = cDummyMarker;
	    }

		/* Just to be extra safe, set the run offset in
		 * the compression info to 0 so that we start
		 * interpreting the very first run and not some
		 * run left over from some previous stoppage.
		 * I'm not really sure when this does anything
		 * that wasn't already done, but it seems tidy.
		 */
	    pCCITTCompInfo->currentRun = 0;
	    pCCITTCompInfo->refRun = 0;

	}
	else	/* we're restarting after a fault.  If we are
		 * going to encode a 1D line, we have to find the position
		 * of the dummy run at the end so we can
		 * convert this into the type of reference run
		 * that the 2D code wants to see.
		 */
	if (encode1DLine)
	{
	    for (pDummy = pCodingRuns;
		 pDummy->runRight != cDummyMarker;
		 pDummy++)
		 ;
	}

	    /* Write out the proper EOL code.  If this is
	     * hybrid 1D/2D mode, write out a flag bit
	     * indicating which type of line follows.  This
	     * call and check is out here so that we can
	     * restart here if this call filled the compression
	     * buffer.  If we filled the compression buffer
	     * while in ip_writeCCITTEOL, the state recorded
	     * in pCCITTCompInfo->compState belongs to
	     * ip_writeCCITTEOL.  When we get back here, this
	     * routine will take up where it left off.  If
	     * we stop in ip_compMHRow or ip_compMMRRow, the
	     * restart state will not belong to ip_writeCCITTEOL
	     * and it will simply return after being called.
	     */
	if (pCCITTCompInfo->use2D)
	{
	    __error__ = ip_writeCCITTEOL (pIPState, 
					pCCITTCompInfo, pCCITTCompInfo->padType,
				        (encode1DLine? cEOL1D : cEOL2D),
				        &pCurCompData, pCompDataEnd);
	}
	else	/* pure 1D encoding, so no flag bit */
	{
	    __error__ = ip_writeCCITTEOL (pIPState,
					pCCITTCompInfo, pCCITTCompInfo->padType,
					cPlainEOL,
					&pCurCompData, pCompDataEnd);
	}

	    /* Should we stop after writing the EOL?
	     * If there was a catastrophe, stop now.
	     * If we couldn't finish ip_writeCCITTEOL because the
	     * compression buffer filled up, stop now.  If we
	     * happened to fill it while happily completing the
	     * ip_writeCCITTEOL routine, __error__ will be ia_successful
	     * but pCurCompData will be equal to pCompDataEnd. This
	     * condition will be detected in ip_compMHRow or
	     * ip_compMMRRow below.  Since they'll set the restart
	     * state, that's where we'll restart.  If we stopped
	     * here just because pCurCompData == pCompDataEnd,
	     * the restart state would be cCCITTCompNewLine, which
	     * looks like a perfectly fine state to ip_writeCCITTEOL
	     * and the restart would produce *another* EOL code.
	     */
	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;	/* catastrophe */

	if (__error__ == ia_bufferExceeded)
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * ((pCCITTCompInfo->lineNumber) * w)) /
			    (numLines * w);
	    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

		/* Set __error__ to ia_successful.  This is the convention
		 * at this point.  There wasn't a real error, we just
		 * filled up the output buffer.  The caller detects this
		 * by looking at *pExhaustedCompData.
		 */
	    __error__ = ia_successful;
	    NORMAL_RETURN;
	}

	    /* Ask the compression code to translate runs into the
	     * compressed data stream.  The width sent down is used
	     * to decide when we've coded the whole line.
	     * If we are either in pure 1D mode or this is the
	     * first line of a hybrid 1D/2D block, use ip_compMHRow
	     * to do 1D encoding.  Otherwise, use 2D encoding.
	     * Note that the real reason that ip_compMHRow has a
	     * separte padType argument is so that we can
	     * set it to cNoPad here.  We never want to flush
	     * the 1D scan lines in MR encoding.
	     */
	if (encode1DLine)
	{
	    __error__ = ip_compMHRow (pIPState, pCCITTCompInfo, cDummyMarker,
				    cNoPad,
				    &pCurCompData, pCompDataEnd);
	}
	else	/* use 2D encoding */
	{
	    __error__ = ip_compMMRRow (pIPState, pCCITTCompInfo, w,
				     &pCurCompData, pCompDataEnd);
	}

	    /* did a catastrophe happen during encoding the line? */
	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;

	    /* We have now attempted to compress one more scan line,
	     * possibly not the last one.  If we completed it,
	     * update the lineNumber, the coding pointers, the
	     * dummy runs and all sort of other stuff.  This should
	     * happen whether or not we happened to run out of
	     * compression buffer at the same time.
	     */
	if (pCCITTCompInfo->compState == cCCITTCompNewLine)
	{
	    pCCITTCompInfo->lineNumber++;

	    if (pCCITTCompInfo->use2D)
	    {
		if (encode1DLine)	/* need to fix dummy ptrs on the end */
		{
		    pDummy[0].runLeft = w;
		    pDummy[0].runRight = w-1;
		    pDummy[1].runLeft = w;
		    pDummy[1].runRight = cDummyMarker;
		}
		pCCITTCompInfo->k--;	/* decrement number of lines left in
					 * the current block */

		    /* Figure out whether to reset the reference and coding
		     * run pointers.  This happens when we're going to start
		     * a new block in hybrid 1D/2D mode.
		     */
		if (pCCITTCompInfo->k == 0)
		{
		    pCodingRuns = (Run32 *)
			    ((UInt8 *)(pCCITTCompInfo) +
			     pCCITTCompInfo->runArrayOffset +
			     sizeof(Run32));
		    pRefRuns = (Run32 *)
			    ((UInt8 *)(pCCITTCompInfo) +
			     pCCITTCompInfo->run2ArrayOffset +
			     sizeof(Run32));
		    pCCITTCompInfo->k = pCCITTCompInfo->maxK;
		    pCCITTCompInfo->run2ArrayIsCoding = cFalse;
		}
		else
		{
			/* Switch the coding and reference runs */
		    pCCITTCompInfo->run2ArrayIsCoding ^= cTrue;
		    pDummy = pCodingRuns;
		    pCodingRuns = pRefRuns;
		    pRefRuns = pDummy;
		}
	    }
	}

	    /* Should we stop after encoding the line?
	     * If we stopped in the middle of a line because the
	     * compressed data buffer was full and we had more stuff
	     * to write, or we finished the line but just happened
	     * to fill the compressed data buffer and there are
	     * more lines to go, stop now.
	     */
	if ((__error__ == ia_bufferExceeded) ||
	    ((pCurCompData == pCompDataEnd) && (remainingLines > 1)))
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * ((pCCITTCompInfo->lineNumber) * w)) /
			    (numLines * w);
	    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

		/* Set __error__ to ia_successful.  This is the convention
		 * at this point.  There wasn't a real error, we just
		 * filled up the output buffer.  The caller detects this
		 * by looking at *pExhaustedCompData.
		 */
	    __error__ = ia_successful;
	    NORMAL_RETURN;
	}
    }

	/* We've done all the scan lines.  So, just
	 * say things were successful and return.
	 */
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;

	/* Return the number of bytes written, including any
	 * bytes flushed.
	 */
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

    NORMAL_RETURN;

}


/*

 $Log: i_mrc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:48:54  ddavies
 i_mrc.c -> /product/ipcore/ipshared/src/RCS/i_mrc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1996/05/11 00:52:55  ddavies
 Initial revision


 */

