/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_mrd.c: High level code used in Modified Read decompression.
 *
 *   	Int32	i_decompressRowsMR()
 *			Decompresses a block of scan lines using
 *			the Modified Read algorithm; also known as
 *			the CCITT Group 3, or CCITT Fascicle VII.3,
 *			Recommendation T.4 algorithm.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_mrd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* The wonderful/horrible thing about MR decompression is that it has both a
 * pure 1D mode and a hybrid 1D/2D mode.  In 1D mode, all lines are
 * compressed using a 1D compression technique using standard Huffman
 * tables for encoding lengths of alternating white and black runs.
 * In 2D mode, the position of each run endpoint is encoded relative to the
 * endpoint of the corresponding run in the line above it if possible.
 * When there is no reference endpoint close enough, we encode the lengths
 * of two runs.
 *
 * The CCITT standard says that when using hybrid 1D/2D coding, the
 * image should be encoded in blocks of 2 scan lines or 4 scan lines.
 * These correspond to k = 2 and k = 4.  In "k = 2" mode, there's
 * one 1D scan line for each 2D scan line.  In the "k = 4" mode,
 * there's one 1D scan line followed by three 2D scan lines.
 * The mode is selected based on resolution.  For "standard"
 * vertical resolution (97.79 lines per vertical inch), one
 * should use the first mode.  For optional higher vertical
 * resolution(s) (195.58 lines/vertical inch), one should
 * use the second mode.  Leffler's code decided that the first
 * mode was used for resolutions less than or equal to 150 dpi and
 * the second mode for larger vertical resolutions.  The point of
 * all this is that if a receiver gets lost because of transmission
 * errors, it has a hope of resynchronizing before too many scan
 * lines have passed.
 *
 * In MR, each line ends with an EndOfLine code.  There's also one
 * EOL before the first line, so maybe these are BeginningOfLine
 * codes.  In fact, it's easier to think of them this way for
 * two reasons.  First, any padding required to make the EOL end on
 * a byte boundary (an option) is inserted between the EOL and the
 * scan line data *before* it.  This associates the EOL more closely
 * with the following scan line.  The other reason has to do with the
 * extra flag immediately following the EOL.
 *
 * In pure 1D mode, the EndOfLine codes are simply inserted.  In hybrid
 * 1D/2D mode, each EOL is followed by a flag bit indicating whether the
 * next line is encoded using 1D codes or 2D codes.  The bit is a 1
 * when the next line is coded using 1D codes and a 0 when the next line uses
 * 2D codes.  This also makes the EOL act like a BeginningOfLine code.
 *
 * Our routine will start each scan line by skipping optional padding,
 * finding the EOL code and the optional 1D/2D flag bit.  It then uses a
 * routine to decode the 1D or 2D scan line.
 *
 * In ipshared, 1D decompression of a scan line's data is performed
 * by ip_decompMHRow.  This is the same primitive routine used in
 * MH decompression.  Our 2D compression is supplied by ip_decompMMRRow,
 * which is also used in MMR decompression.
 *
 *			*	*	*
 *
 * There is a big comment all about the low level mechanics of our
 * implementation of MMR decompression in mimmrd.c.  Another big comment
 * in mimhd.c talks about the mechanics of 1D decompression.  Since that's
 * all taken care of there, this comment will deal with software
 * engineering and interfaces on the next higher level.  The basic
 * style of interfaces is shared by the other CCITT decompression
 * routines and the PackBits decompression routine.
 * 
 * MH decoding just produces the endpoints of the runs of one scan line.
 * MMR decoding uses endpoints on one scan line and the incoming codes
 * to produce the runs on the new scan line.
 * The "coding" line is the line being encoded.  The "reference" line is
 * the one above the coding line.  There are two run arrays in the
 * decoder state block.  We ping-pong between them; filling one by
 * using the other as the reference line, then filling the other and using the
 * first for the reference line and so on.  The run2ArrayIsCoding
 * boolean in the decompression data structure indicates whether the
 * second run array is the coding line (cTrue) or the reference line
 * (cFalse).  When doing MH decompression, we always start over with
 * the first run array, so run2ArrayIsCoding is always set to cFalse.
 *
 * After skipping the optional padding and EOL and the optional 1D/2D flag,
 * either ip_decompMHRow or ip_decompMMRRow is used to find all the
 * run endpoints for the new coding line.  Each run array starts with
 * a dummy element that has left and right coords at -1.  We insert
 * two dummy elements after the last valid run so it will look right
 * if it's used as a reference line later.
 *
 * After the run endpoints for a line are found, we paint the runs
 * into the bitmap using ip_paintRunsLine.
 *
 * Note that the compressed data stream is always read in the
 * platform's natural byte order.  It will have to show up in that
 * order at some point and (theoretically) there aren't very many
 * compressed bytes, so this is no problem, right?
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which this procedure will stop.  It might actually
 * finish decoding the entire scan line.  When this happens, we update the
 * pointer to the compressed data that was passed in, set the state
 * to cCCITTDecompNewLine and store the current accumulator and vacantBits
 * and return happily.
 *
 * The other reason for returning is that we run out of
 * compressed data.  ip_skipCCITTEOL, ip_decompMMRRow and ip_decompMMRRow
 * all detect this by comparing pCurCompData to a pointer to the end of
 * the buffer just before reading encoded data.  Since we want to
 * pretend that we have a callback to fill the buffer, we need a
 * restart location next to every check.  There are several checks
 * in each routine.  The numbers indicating the restart state are
 * left in the CCITTDecompInfo state block.  These numbers are unique for
 * each routine.  When restarting, we first try to skip the EOL again.
 * If the restart state doesn't belong to it, it just returns.
 * If the restart state does belong to it, it takes up where it left
 * off and sets the restart state to cCCITTCompNewLine.  We then call
 * ip_decompMHRow or ip_decompMMRRow.  These routines work the same way -
 * if they own the restart state, they restart in the proper spot.
 * All routines recognize cCCITTDecompNewLine as the "normal" starting
 * state.
 *
 * We always start the run array with a dummy element that says there is
 * a black run whose left and right endpoints are at -1.  We know we can
 * always point at this run.  We quit decoding a line when the current
 * position is off the right edge of the image.
 *
 *			*	*	*
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * and the continueImageBlock, continueCompDataStream and flushPipeline args
 * discussed above.  The caller also supplies pointers to words used to
 * return the exhaustedImageData, exhaustedCompData, bytesWritten and
 * percentDone args.  The bytesRead arg returns the number of compressed
 * bytes read so far.  If flushPipeline is not set, some of the compressed
 * bytes may have been read, but not yet decoded.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.  One possibly catastrophe is
 *	that flushPipeline was set but insufficient compressed data
 *	was supplied.  In this case, we return ia_bufferExceeded.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	We have processed all the image data requested.  The compression
 *	buffer is not empty.  Some of the compressed data bytes are
 *	still in the pipeline regardless of the state of flushPipeline.
 *	The number of bytes actually read from the buffer
 *	(including any bytes still in the pipeline) are returned.
 *	When the decompression routine is called next, the client must
 *	pass in a pointer to the byte holding the next compressed byte.
 *	This may be just beyond the end of the current compressed data,
 *	at the beginning of the same buffer, or in another buffer entirely.
 *	The size of the compressed data buffer must also be (re)calculated
 *	by the client.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to read the last
 *	byte of the compressed data buffer at the same time the compressed
 *	data came to an end.  As before, if flushPipeline was false, there
 *	may be more compressed data in the pipeline.  Normally, this
 *	case occurs when exactly the right amount of compressed data was
 *	supplied and flushPipeline was true.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of compressed data before reading the last
 *	byte of compressed data requested.  Some of the data read
 *	is in the pipeline.  This can only happen if flushPipeline was
 *	false.  If flushPipeline is true, we keep going until we either
 *	finish all the scan lines or get an ia_bufferExceeded.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */

/* This is a big number used to mark the right coord of the second dummy
 * run at the end of each scan line.
 */
#define	cDummyMarker	0x7FFFFFFF

/* PUBLIC */
/**************************************************************************
 *   i_decompressRowsMR()
 *
 *	This procedure decompresses Modified Huffman encoded images in
 *	units of scan lines.
 *
 *	Input:
 *		pCCITTDecompInfoIn:	Pointer to CCITT decompression info
 *		pImage:			Pointer to the upper left corner
 *					of the frame of the uncompressed
 *					image.  If NULL, no runs are painted.
 *		startLine:		First image line to be decompressed
 *					relative to the top of the frame
 *		numLines:		Number of lines to be decompressed
 *		bpl:			bytes/line of uncompressed data.
 *		combineResults:		If set to cFalse, the image is assumed
 *					to be initially clear.  This is most
 *					efficient.  If set to cTrue, the
 *					decoded image is ORed into the original.
 *		pCompData:		Pointer to next byte of compressed
 *					data to be read.  If several calls
 *					to this routine are needed to process
 *					a block of compressed data, this
 *					pointer must be updated on each call.
 *		compDataBufSize:	Number of bytes of compressed data
 *					remaining in the buffer.
 *		continueImageBlock:	cTrue => we are decompressing
 *					into the same slice of image as was
 *					specified in the previous call and the
 *					saved state in pCCITTDecompInfoIn is
 *					valid.
 *					This arg is set to cFalse if either
 *					we are starting with all new data
 *					or the previous call returned because it
 *					ran out of image data.
 *		continueCompDataStream:	cTrue => The saved decompression state
 *					in pCCITTDecompInfoIn is valid.  This
 *					allows the procedure to treat several
 *					buffers of compressed data as though
 *					it were one contiguous buffer.
 *					If set to cFalse, this routine will
 *					discard all previous compression
 *					state.
 *		flushPipeline:		If this arg is set, this routine
 *					will attempt to complete the
 *					specified number of scan lines
 *					WITHOUT stopping when it starts to
 *					run out of data; i.e. it continues
 *					until its pipeline is completely
 *					empty.  If this is not possible,
 *					an ia_bufferExceeded error is raised.
 *		pExhaustedImageData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because specified
 *					number of lines have been processed.
 *					This may require more than one
 *					call to this routine.
 *		pExhaustedCompData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because it reached
 *					the end of the compressed data buffer
 *					as specified by compDataSize.  This
 *					routine may be called again with 
 *					new compression data and the the 
 *					same image.  This arg will not be 
 *					set if flushPipeline is set and we 
 *					failed to decompress the specified 
 *					number of scan lines (routine returns
 *					ia_bufferExceeded instead).
 *		pPercentDone:		Pointer to returned arg.  When
 *					routine returns with *pExhaustedCompData
 *					set, this arg specifies the rough
 *					percentage of the image block that
 *					has been decompressed.  It is
 *					intended to be used to estimate the
 *					amount of compressed data needed
 *					to complete the task.  The result
 *					ranges from 0 to 100.
 *		pBytesRead:		Pointer to returned arg.  Contains
 *					number of compressed data bytes
 *					read by the routine before it returned.
 *					The client may safely dispose of
 *					these bytes.
 *
 *	Action:
 *		Decompresses a horizontal slice of MR data.
 *		There are several completion conditions:
 *
 *	__error__	!= ia_successful
 *	*pExhaustedImageData == don't care
 *	*pExhaustedCompData == don't care
 *		This means something catastrophic happened.  One possible
 *		catastrophe (ia_bufferExceeded) happens if the client
 *		set flushPipeline and failed to supply enough compressed
 *		data.		
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cFalse
 *		This means we produced the desired number of uncompressed
 *		image lines without running out of compressed data.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse and
 *		continueCompDataStream = cTrue.  pCompData and compDataSize
 *		should be adjusted to account for the number of 
 *		compressed data bytes that have already been read (returned
 *		in *pBytesRead).
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cTrue
 *		This means we ran out of compressed data at the same time
 *		we completed the correct number of image lines.  This
 *		is the normal condition at the bottom of a TIFF slice or
 *		the bottom of the image.  It is not possible to restart
 *		from this state, though it is possible to decompress
 *		with new image and compression data.  When the client
 *		expects to run out of compressed data, the flushPipeline
 *		argument should be set.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cFalse
 *	*pExhaustedCompData == cTrue
 *		The routine ran out of compressed data before completing
 *		the specified number of lines.  This condition will not
 *		occur if flushPipeline was set.  If flushPipeline was set
 *		and we ran out of compressed data before completing the
 *		the task, an ia_bufferExceeded error will be returned.
 *		When this condition does occur, the routine may be
 *		restarted by supplying more compressed data, setting
 *		BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  Note that the first byte of the new compressed
 *		data must be the one immediately following the last
 *		byte of old compressed data.
 *		The image may be moved around as a block if necessary, but
 *		the internal composition of the image (and startLine
 *		and numLines) should not be changed.  When this
 *		condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much compression data is 
 *		needed to complete the slice.
 *		
 *		The image data is always in PIXR order.
 *		
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		a pointer was NULL or 
 *					(continueImageBlock && the saved image
 *					info is not valid) or
 *					(continueCompDataStream && the saved
 *					compression info is not valid)
 *		ia_bufferExceeded	ran out of compressed data when
 *					flushPipeline was set.
 *		ia_decodeFailure	an illegal code was enountered.
 *
 **************************************************************************/
Int32  CDECL 
i_decompressRowsMR (
    IPState    *pIPState,
    void       *pCCITTDecompInfoIn,
    UInt8      *pImage,
    Int32	startLine,
    Int32	numLines,
    Int32	bpl,
    Bool	combineResults,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Bool	flushPipeline,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesRead)
{
    PROLOGUE;
    CCITTDecompressionInfo	*pCCITTDecompInfo;
    UInt8			*pCompDataEnd;	/* address 1 beyond the last
						 * byte of compressed data */
    UInt8			*pCurCompData;	/* current compressed data
						 * pointer */
    Run32			*pCodingRuns;	/* pointer to first
						 * coding run in
						 * pCCITTDecompInfo */
    Run32			*pRefRuns;	/* pointer to first
						 * reference run in
						 * pCCITTDecompInfo */
    Run32			*pDummy;	/* returned ptr to
						 * element just beyond
						 * last valid run in
						 * the array */
    UInt32			 remainingLines;
    UInt32			 imWidth;	/* width of unframed image */
    

	/* check for null pointers */
    if ((pCCITTDecompInfoIn == NULL) || (pCompData == NULL) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesRead == NULL))
	return ia_invalidParm;

	/* load the local copies of the remote pointers */
    pCCITTDecompInfo = (CCITTDecompressionInfo *) pCCITTDecompInfoIn;

	/* Strip off the frame */
    if (pImage != NULL)
	pImage += (bpl << cLogFrameBits) + cFrameBytes;
    startLine -= cFrameBits;
    imWidth = pCCITTDecompInfo->width;

	/* Initialize the state from the saved state passed in *pCCITTDecompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	if (pImage != NULL)
	    pImage += startLine * bpl;
	pCCITTDecompInfo->lineNumber = 0;
	pCCITTDecompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pCCITTDecompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pCCITTDecompInfo->lineNumber;
	if (pImage != NULL)
	    pImage += (startLine + pCCITTDecompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
	pCCITTDecompInfo->run2ArrayIsCoding = cFalse;
	pCCITTDecompInfo->currentRun = 0;
	pCCITTDecompInfo->refRun = 0;
	pCCITTDecompInfo->EOLType = cEOL1D;
	pCCITTDecompInfo->accumBits = 0;
	pCCITTDecompInfo->vacantBit = 0;
	pCCITTDecompInfo->remainingLength = 0;
	pCCITTDecompInfo->decompInfoIsValid = cTrue;
    }
    else
    if (!pCCITTDecompInfo->decompInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* Set the run pointers to the second runs.  The first is a dummy. */
    if (pCCITTDecompInfo->run2ArrayIsCoding)
    {
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
					  pCCITTDecompInfo->run2ArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				       pCCITTDecompInfo->runArrayOffset +
				       sizeof(Run32));
    }
    else
    {
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
					  pCCITTDecompInfo->runArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				       pCCITTDecompInfo->run2ArrayOffset +
				       sizeof(Run32));
    }

	/* If we're not continuing from a previous image block,
	 * set the reference line to be "blank" by just putting
	 * in the two dummies at the end of its line.  This is
	 * useful if we're about to decode a 2D line and doesn't
	 * hurt if we're going to decode a 1D line.
	 */
    if (!continueCompDataStream)
    {
	pRefRuns[0].runLeft = imWidth;
	pRefRuns[0].runRight = imWidth-1;
	pRefRuns[1].runLeft = imWidth;
	pRefRuns[1].runRight = cDummyMarker;
    }

	/* Decompress each scan line.
	 * This is done in several steps:
	 *   1. Skip until and EndOfLine code is found.  If we
	 *	are using the 2D option the type of EOL found is
	 *	recorded in pCCITTDecompInfo->EOLType,
	 *	otherwise it's left alone.
	 *	If we ran out of data in this step, this is
	 *	a swell place to restart.  If we ran out of data
	 *	decoding a scan line's codes, this routine will
	 *	simply return without changing pCCITTDecompInfo->EOLType.
	 *   2. We might have quit in the middle of decoding all the
	 *	runs in this scan line.  We choose which scan line
	 *	decoding routine to use based on pCCITTDecompInfo->use2D
	 *	and pCCITTDecompInfo->EOLType and execute it.
	 *   2. Call ip_paintRunsLine to draw the runs into the bitmap.
	 *	We cannot abort in the middle of this step.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--)
    {

	    /* Skip over the next EOL.  If we're using 2D encoding,
	     * record the EOL type.
	     */
	__error__ = ip_skipCCITTEOL (pIPState, pCCITTDecompInfo,
				   pCCITTDecompInfo->use2D,
				   flushPipeline,
				   &pCurCompData, pCompDataEnd,
				   &(pCCITTDecompInfo->EOLType));

	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;	/* disaster */
	if ((__error__ == ia_bufferExceeded) && (flushPipeline))
	{
	    pCCITTDecompInfo->imageInfoIsValid = cFalse;
	    pCCITTDecompInfo->decompInfoIsValid = cFalse;
	    NORMAL_RETURN;
	}

	    /* If we didn't get to the end of an EOL, return now
	     * instead of trying to decode a scan line.
	     * Since flushPipeline was not set, this is all ok.
	     */
	if (__error__ == ia_bufferExceeded)
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * pCCITTDecompInfo->lineNumber) /
			    (numLines);
	    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

	    NORMAL_RETURN;
	}

	    /* Ok, we found the EOL successfully.
	     * Decode the scan line codes that come next. */
	if (pCCITTDecompInfo->use2D)
	{
	    if (pCCITTDecompInfo->EOLType == cEOL2D)
	    {
		__error__ = ip_decompMMRRow (pIPState, pCCITTDecompInfo, imWidth,
					   flushPipeline,
					   &pCurCompData, pCompDataEnd,
					   &pDummy);
	    }
	    else
	    {
		/* We're restarting, so reset the coding and reference
		 * runs.  I'm not sure that this really makes a difference,
		 * but it's what the encoding side does and doing the
		 * same thing in both places is nice.
		 */
		pCodingRuns = (Run32 *)
				((UInt8 *)(pCCITTDecompInfo) +
				pCCITTDecompInfo->runArrayOffset +
				sizeof(Run32));
		pRefRuns = (Run32 *)
				((UInt8 *)(pCCITTDecompInfo) +
				pCCITTDecompInfo->run2ArrayOffset +
				sizeof(Run32));

		__error__ = ip_decompMHRow (pIPState, pCCITTDecompInfo,
					  cNoPad, imWidth, flushPipeline,
					  &pCurCompData, pCompDataEnd,
						&pDummy);
	    }
	}
	else	/* this is pure 1D mode so we always leave the coding
		 * and reference runs alone.
		 */
	{
	    __error__ = ip_decompMHRow (pIPState, pCCITTDecompInfo,
				      cNoPad, imWidth, flushPipeline,
				      &pCurCompData, pCompDataEnd,
				      &pDummy);
	}

	if ((__error__ != ia_successful) && (__error__ != ia_bufferExceeded))
	    NORMAL_RETURN;
	if ((__error__ == ia_bufferExceeded) && (flushPipeline))
	{
	    pCCITTDecompInfo->imageInfoIsValid = cFalse;
	    pCCITTDecompInfo->decompInfoIsValid = cFalse;
	    NORMAL_RETURN;
	}

	    /* If we didn't get to the end of a scan line, return now
	     * instead of painting all the runs to the bitmap.  We
	     * don't paint the runs because we don't have all of them.
	     * Since flushPipeline was not set, this is all ok.
	     */
	if (__error__ == ia_bufferExceeded)
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 * pCCITTDecompInfo->lineNumber) /
			    (numLines);
	    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

	    NORMAL_RETURN;
	}
	else
	{
		/* Mark the two runs beyond the last run in the same
		 * way they were marked during encoding.  This guarantees
		 * that we'll get the same answer when we're searching
		 * through these runs when they become the reference row.
		 * Note that we can always use the dummies used for 2D
		 * reference lines since the 2D encoding routines are the
		 * only ones that care.
		 */
	    pDummy[0].runLeft = imWidth;
	    pDummy[0].runRight = imWidth-1;
	    pDummy[1].runLeft = imWidth;
	    pDummy[1].runRight = cDummyMarker;
	}

	    /* Great, we got a set of runs.  Go paint them into the
	     * image.  If pImage is NULL, someone just wants the runs
	     * so don't paint them.
	     */
	if (pImage != NULL)
	{
	    __error__ = ip_paintRunsLine (pIPState, pImage, imWidth, pCodingRuns,
				        pDummy, combineResults);
	    if (__error__ != ia_successful)
		NORMAL_RETURN;
	}

	    /* increment the current line number just in case we need
	     * to stop in the next line.  Also swap the coding
	     * and reference runs if we're doing 2D encoding.
	     */
	pCCITTDecompInfo->lineNumber++;
	if (pCCITTDecompInfo->use2D)
	{
	    pCCITTDecompInfo->run2ArrayIsCoding ^= cTrue;
	    pDummy = pCodingRuns;
	    pCodingRuns = pRefRuns;
	    pRefRuns = pDummy;
	}
	if (pImage != NULL)
	    pImage += bpl;
    }

	/* Great! we completed the right number of scan lines.
	 * Say that things worked and return.
	 */
    if (pCCITTDecompInfo->use2D)
	pCCITTDecompInfo->run2ArrayMostRecent = 
				pCCITTDecompInfo->run2ArrayIsCoding ^ cTrue;
    else
	pCCITTDecompInfo->run2ArrayMostRecent = cFalse;

    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;
    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

    NORMAL_RETURN;
}


/*

 $Log: i_mrd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.5  1998/01/13 23:48:56  ddavies
 i_mrd.c -> /product/ipcore/ipshared/src/RCS/i_mrd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1996/07/18 23:39:12  ddavies
 i_mrd.c -> /product/ipcore/ipshared/src/RCS/i_mrd.c,v

 Change initImage parameter to combineResults parameter.  When combineResults
 is set, the decompressed image is ORed into the original dest image.
 When combineResults is not set, the original dest image is assumed to be
 white (this is often true and is faster).

 Revision 1.3  1996/05/21 18:48:09  ddavies
 i_mrd.c -> /product/ipcore/ipshared/src/RCS/i_mrd.c,v

 Change name of width variable from w to imWidth for readability.
 Do not paint decoded runs into a bitmap if the bitmap pointer is NULL.  This allows
 us to simply extract the runs from a MR file.  Also update the
 run2ArrayMostRecent var on leaving the routine.

 Revision 1.2  1996/05/16 17:08:19  ddavies
 i_mrd.c -> /product/ipcore/ipshared/src/RCS/i_mrd.c,v

 Fix several bugs that are encountered when you repeatedly read a single scan
 line from a large block of compressed data.
 First, delete the code that invalidates the saved state when flushPipeline
 is set.  It is very difficult to determine whether a particular scan line
 is the last scanline of a strip at the level where flushPipeline is set.
 Second, move the code that initializes state variables controlling the run
 data structures from the block controlled by continueImageBlock
 to the block controlled by continueCompDataStream.  In decompression,
 the runs are associated with the compressed data, not with the image.
 Also move some code around for more uniformity with the other CCITT decompression routines.
 Also initialize the runs in the reference row when we're starting a new
 compressed data stream, not when we're starting a new image block.

 Revision 1.1  1996/05/11 00:53:53  ddavies
 Initial revision


 */

