/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_nirfl1.c: NonInterpolatedRotation, FLoaTing point interface
 *
 *	    Upper level rotation routine that does NOT use interpolation to
 *		form output pixels, SMALL ANGLES ONLY! (< 10 degrees)
 *		This routine takes the rotation angle in floating point
 *		radians.
 *
 *          Binary/gray rotation routine:
 *		Int32     	i_rotateNoInterpolation()	Assembler
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"

    /*  prototypes */
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_nirfl1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* This is just an interface to the fixed point version of non-interpolated
 * rotation.  Please see the EXTENSIVE comments in i_nirfx1.c for a
 * description of the rotation algorithm.
 *
 * This routine takes the rotation angle in units of radians
 * expressed as a Float64.  It converts the radians to ScaledAngle format
 * and calls i_rotateNoInterpolationFixedPoint.  That routine does
 * the real work.
 */


/* PUBLIC */
/**************************************************************************
 *  i_rotateNoInterpolation()
 *	Input:
 *		pS = source binary or gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		rotAngle = rotation angle in RADIANS
 *		pD = destination binary or gray image
 *		dBpl = destination bytes/line
 *		depth = bits/pixel in source and dest images.
 *		nice = cTrue if we should do the slower version
 *			of rotation that makes an attempt to
 *			reduce the number of jaggies.
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *
 *	On return:
 *		Destination image contains the source image rotated by
 *		the given angle.  Unlike i_rotateGray, this routine
 *		does not form rotated pixels by averaging the values
 *		of the input pixels that overlap the output pixel.
 *		In this routine, each output pixel is assigned the
 *		value of the input pixel with the greatest overlap.
 *		If nice = cTrue, the rotation routine makes some attempt
 *		to avoid horizontal shears of vertical strokes by adjusting
 *		the horizontal shift amount by 1 if vertical strokes
 *		don't line up.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		unable to allocate memory for tables
 *		ia_nopoint		could not find next point
 *		ia_depthNotSupported	could not rotate source image
 *					of specified depth
 *		ia_angleTooLarge	can not rotate by this angle
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	Frame in destination image is assumed to already be cleared.
 *	Frame in destination image is assumed to already be cleared.
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateNoInterpolation (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	wS,
    UInt32	hS,
    Int32	sBpl,
    Float64	rotAngle,
    UInt8      *pD,
    Int32	dBpl,
    Int32	depth,
    Bool	nice,
    Int16	xRes,
    Int16	yRes)
{
    PROLOGUE;
    ERR_PASS(i_rotateNoInterpolationFixedPoint (pIPState, pS, wS, hS, sBpl,
					        RADIAN_TO_ScaledAngle(rotAngle),
					        pD, dBpl, depth, nice,
					        xRes, yRes));
    NORMAL_RETURN;
}


/*

$Log: i_nirfl1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.8  1998/01/13 23:48:57  ddavies
i_nirfl1.c -> /product/ipcore/ipshared/src/RCS/i_nirfl1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1997/08/20 20:26:57  ddavies
i_nirfl1.c -> /product/ipcore/ipshared/src/RCS/i_nirfl1.c,v

Change Int8 nice to Bool nice in arg list.

Revision 1.6  1997/06/13 21:17:29  jcraig
jcraig on Fri Jun 13 14:16:31 PDT 1997

i_nirfl1.c -> /product/ipcore/ipshared/src/RCS/i_nirfl1.c,v

yet another comment change.  this comment indicated that
i_rotateNoInterpolation took only binary images (when it takes either
binary OR gray).

Revision 1.5  1997/06/13 21:11:25  jcraig
jcraig on Fri Jun 13 14:10:00 PDT 1997

i_nirfl1.c -> /product/ipcore/ipshared/src/RCS/i_nirfl1.c,v

change comments to indicate that i_rotateNoInterpolation will take
both binary and gray images (not just gray).

Revision 1.4  1995/12/13 02:44:49  ejaquith
i_nirfl1.c -> /product/ipcore/ipshared/src/RCS/i_nirfl1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/02/03  19:45:47  ddavies
 * i_nirfl1.c -> /product/ipcore/ipshared/src/RCS/i_nirfl1.c,v
 *
 * Change xRes and yRes from UInt16's to Int16's to keep acc happy.
 *
 * Revision 1.2  1995/02/03  01:21:24  ddavies
 * i_nirfl1.c -> /product/ipcore/ipshared/src/RCS/i_nirfl1.c,v
 *
 * Add xRes and yRes args to small angle rotations.  These are used to apply
 * a horizontal shear when rotating anisotropic images.  The shear is needed
 * to account for the shear naturally present in such images.
 * Also modified the calculation of the upper left corner of anisotropic
 * input images.
 *
 * Revision 1.1  1995/01/12  03:09:02  ddavies
 * Initial revision
 *
 * Revision 1.4  1993/12/02  19:23:15  ddavies
 * i_nirflt.c -> /product/ipcore/ipshared/src/RCS/i_nirflt.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:28:28  ddavies
 * i_nirflt.c -> /product/ipcore/ipshared/src/RCS/i_nirflt.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


