/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_nirfx1.c: NonInterpolatedRotation, FiXed Point interface
 *
 *	    Upper level rotation routines that do NOT use interpolation to
 *		form output pixels, SMALL ANGLES ONLY! (< 10 degrees)
 *		These routines take the rotation angle in fixed point
 *		ScaledAngle format.
 *	    Lower level routines implemented in C or assembler.
 *
 *          Binary/gray rotation routine:
 *		Int32     	i_rotateNoInterpolationFixedPoint()
 * 
 *
 */


#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"
#include "fxpmath.pub"

    /*  prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrscal.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrrast.pub"
#include "shros.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_nirfx1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* People don't always put documents on copiers straight.  The ability
 * make these skewed images look nice by rotating them a few degrees is
 * thought to be a valuable feature.
 *
 * Rotation is an ugly operation.  We have to move all the pixels in the
 * image by distances that appear straightforward to image procesing
 * grad students, but arbitary to microcoders.  WE DON'T LIKE TRANCENDENTAL
 * FUNCTIONS!!  Heck, we don't even like mutiplication.
 *
 * Image processing experts are to assembly programmers as Newtonian
 * mechanics is to quantum mechanics.  There are no continuous functions
 * down here.
 *
 * The "rotateGray" family of routines attack this problem by doing
 * interpolated rotation of 8-bit gray images.  Each pixel is formed from
 * the weighted sum of the corresponding pixels in the input image.
 * Interpolation works well because there are lots of gray levels in both
 * input and output images.  That is, the gray image does a very good job of
 * representing a diagonal line as a group of pixels whose values vary
 * smoothly.  There are no input "jaggies."  Similarly, it makes perfect
 * sense to compose an output pixel from a weighted sum of continuous values,
 * since output pixels also take on gray values.
 *
 * While gray rotation gives very nice results (after working on it for a
 * month!), it's slow.  There's too much data and the operations are too
 * complex.  For performance reasons, we'd like to rotate images after they've
 * been converted to binary.
 *
 * Can we make these images look nice?  It's really hard.  First, consider the
 * problem of jaggies.  All binary images have them since diagonal lines must
 * be represented by a "staircase" of binary pixels.  One's first inclination
 * is to simply smooth the jaggies back into gray.  Well, if you round off
 * the corners of a staircase, you end up with a staircase with rounded corners;
 * not at all the same as a straight line.  What if you smooth things even
 * more?  The problem is that there are edges in the binary image that you
 * want to keep nice and sharp.  Very thin lines and characters should be
 * retained.  If the image was converted to binary using error diffusion, there
 * are lots of isolated dots and precise grids of dots that you can't afford
 * to lose or blur into nothing.  How do you tell the difference betweeen
 * nice clean sharp edges and ugly jaggies?
 *
 * Take this one step further.  Say a line which should be horizontal is
 * represented as a staircase because the image was skewed.  We're trying to
 * deskew the image, so will be following a similar staircase across the
 * input image to produce the output image.  The problem is that the staircase
 * we follow will inevitably be out of phase with the staircase in the drawing.
 * This is because staircases in the drawing have arbitrary phase with respect
 * to each other.  Any fixed phase we choose will be wrong.  Big deal, so
 * we're out of phase, so what?  This means that at the edge of the line,
 * we'll dip down into black pixels for awhile, that line of black pixels will
 * end so we'll go through some white pixels then we dip down into black
 * again and so on.  Any stateless method that computes output pixels 
 * based solely on input pixels in the neighborhood of the output 
 * (and is capable of producing both 0's and 1's) will see neighborhoods 
 * dominated by alternately by 0's and 1's, so will produce an output 
 * line with bumps along the edges.  This is the biggest, though not the 
 * only, problem we're trying to avoid.
 *
 * At this point, we could start messing around with algorithms that have
 * state, perhaps using some sort of hysteresis to smooth the bumps.  Neat.
 * We'll go into that in a later set of routines in different files.
 * For now, lets see if we can just do a plain rotation very quickly.
 * This type of rotation will just move each pixel to some other location 
 * without trying to smooth the bumps.  It will add a new set of jaggies.  
 * However, at sufficiently high resolution, this looks sort of ok.  
 * Anyway, it's the basis for the more complicated methods.
 *
 * This job is currently done in the Alpaca world by the procedure "rotateArb".
 * The procedure use two or three shears of the image to get the results.
 * Each shear writes the entire image using a bunch of separate bitblts.  We'd
 * like to get improved performance by avoiding shears and only writing the
 * output image once.
 *
 * How?  We'll form each output line by tracing a diagonal path across the
 * input image.  For example, if we're supposed to be rotating clockwise,
 * the input is a staircase that rises as it goes across the image from
 * left to right.
 *
 * The tricky part is assigning an exact position to each input bit.  The
 * most straightforward thing to do is to concatenate strings of bits read
 * from the input image.  For example, we might read n bits from line x,
 * another n bits from line x-1 and so on.  The problem with this scheme is
 * that we're assuming that a horizontal move of n bits in the input image
 * exactly equals a horizontal move of n bits in the output image.  This is
 * equivalent to assuming the the cosine of our small angle equals 1.
 *
 * Well, how bad is this assumption?  Say we're rotating by 5 degrees.
 * The cosine of 5 degrees is 0.996195.  Assume we're rotating an 11 x 17
 * image across the 17 inch dimension.  Tracing a diagonal path, we 
 * ought to see 17/0.996195 = 17.06 inches of image.  Instead, we see 
 *  exactly 17 inches since every horizontal bit in the source image 
 * equals one horizontal bit in the dest image.  So we're 0.06 inches, 
 * or 0.4% off.  Big Deal.  This seems like a sufficiently good approximation.
 * Putting the thing through a copier adds around 2% distortion anyway..
 *
 * For comparison, say we're scanning across 8.5 inches.  We'll be off by
 * 8.5/0.996195 - 8.5 = 0.03 inches.  Seems ok to me.  This buys several
 * very nice simplifications in the rotation algorithm as we'll see below..
 *
 * Ok, each output bit will be composed of the data from exactly one input
 * bit.  The input bits will be read in stairstep fashion from the input image
 * (read a run of bits, move up/down one line, read another run of bits, etc).
 * How many input bits should we read from any given scan line?
 *
 * Think of the input and output images as x,y grids - each grid square
 * containing one pixel.  Put the two grids exactly on top of each other.
 * Every output pixel will correspond to exactly one input pixel.  Put a pin
 * through the center of both grids.  Rotate the input by a few degrees
 * with respect to the output around the pin.  The grid square of any output
 * pixel will now intersect four input pixels.  Ideally, the value of the
 * output pixel is equal to the weighted sum of these four pixels.  The
 * weight for each input pixel equals the percentage of output pixel covered
 * by that input pixel (which equals the percentage of input pixel
 * covered by the output pixel).
 *
 * However, in the interests of speed, we will not be computing weighted sums.
 * The next most logical thing to do is to assign the output pixel the value
 * of the input pixel with the greatest overlap.  We'll do that.
 *
 * To do this exactly, we'd need to account for the fact that the individual
 * input pixels are tilted with respect to the output pixels.  The overlapped
 * areas have between 3 and 8 sides.  Yuck.  In our first approximation, we'll
 * pretend that the input grid is really a sheared version of the original
 * input.  That is, the individual input squares have the same orientation as
 * the output squares.  This means the input squares no longer form a
 * rectangular grid, but that's easier to handle, trust me. Think of an
 * arrangement where the upper left corner of each input pixel is in exactly
 * the right place, then we rotate all the input pixels individually around
 * their upper left corners so all the individual input pixels are aligned
 * with all the individual output pixels.
 *
 * In simplest version of the algorithm, we cruise across the output image
 * considering each output pixel.  For each one, we look at the four input
 * pixels, choosing the one with the greatest overlap.  If we're doing
 * clockwise rotation, moving over 1 pixel in output space means we're moving
 * over a little less than one pixel and up some fraction of a pixel in input
 * space.  Our cos(rotAngle) = 1 approximation says that we'll assume
 * moving over by 1 in output space corresponds to moving over by exactly 1
 * pixel in input space, though we still move up by sin(rotAngle).
 *
 * To make a faster algorithm, we don't really want to carefully inspect every
 * output pixel.  We really want to assign runs of input pixels to runs of
 * output pixels.  How are the run lengths computed?  For each x,y position
 * *within a pixel* at a given angle, we get back a run length.  Say the
 * input run slants from lower left to upper right relative to the output.
 * The top half of the input pixels at the left edge of the run overlap the
 * bottom half of the leftmost output pixels.  At the right edge of the run,
 * the bottom half of the input pixels overlap the top half of the output
 * pixels.

 * Given this view of things, run lengths could be found by indexing a
 * table with the fractional portions of the current (x,y) position.  At
 * this point in the narrative, we should have a grasp of the basic problem
 * and assumptions.  From here, we'll start exploring the general idea
 * of using position to determine run lengths.  We can make substantial
 * improvements in the algorithm.
 *
 * It would be nice if the table lookup also retured the next y coordinate.
 * This equals (initial y coord) + (runLength*sin(rotAngle)).  To what accuracy
 * do we keep the overall y coordinate?  Well, 10-12 fractional bits ought
 * to do it.  That will let us compute our trajectory across thousands of
 * pixels with good accuracy.  But, indexing a table with 10-12 y bits plus
 * a few x bits and retrieving 4 bytes each time quickly adds up to a lot
 * of table bits.  A big table produces a lot of costly cache misses.  Is
 * there some way we can cheat?
 *
 * Yes.  We'll take a trick from the gray scale rotation code.  The angles
 * for which changes in y coord result in the largest changes in run length
 * are very small angles (run length is proportional to
 * (y coord)/sin(rotAngle)).  Say we decide that rotating images by less
 * than 1 degree is silly because the jaggies introduced are more
 * objectional than the original skew.  A one degree angle gives a rise of
 * 1 bit in a span of 58 bits.  If we want to determine run lengths to
 * within 1 bit, we need to determine y position to within 1/58 bits.  We
 * will put 6 bits of fractional y coordinate for an accuracy of 1/64 bit
 * into the table.  (This turns out not to be a wonderful idea, but it lets
 * us make progress towards our ultimate goal.)
 *
 * Run length is not nearly so sensitive to fractional x coordinate.  In fact,
 * given our assumption that cos(rotAngle) = 1, run length is independent of
 * fractional X coordinate.    This is not true in normal geometry.  If we
 * can prove run length is independent of fractional X coordinate, we can
 * ignore it in the inner loop.
 *
 * In the proof, we'll concentrate on the path we claim
 * input pixels take across the output array with this assumption.
 *
 * The coordinates of a pixel will be measured relative to the upper left
 * corner of the pixel.  Fractional x and y coordinates range between 0 and 1.
 * An output pixel, O, is assigned the value of an input pixel, I, if the
 * intersection of O with I is greater than the intersection of O with any
 * other pixel.  Define a run of pixels to be {Xn}.  This contains n pixels
 * numbered 0 through n-1 from left to right.  The ith pixel in the run is
 * called Xi.  A run of output pixels {On} is assigned the values of a run
 * of input pixels {In} if and only if for every value of i, 0 <= i < n, Oi
 * is assigned the value of Ii.
 *
 * Say O0 is assigned the value of I0.  This can only happen if both I0's
 * fractional x coordinate, xf, satisfies -1/2 < xf < +1/2 relative to O0's
 * x coordinate.  If xf were less than -1/2, then the input to the right of
 * I0 would have a greater overlap with O0 than I0.  Similarly, if xf were
 * greater than 1/2, the input pixel to the left of I0 would have the greater
 * overlap. This is true for any value of y coordinate.
 *
 * Similarly, yf (the fractional y coordinate) must be in -1/2 < yf < 1/2.
 *
 * Without loss of generality, say we're rotating clockwise.  That means
 * the y coordinate of the input pixels in the output's frame of reference
 * is gradually decreasing (moving towards the top of the image).  Since we
 * assume that cos(rotAngle) = 1, the fractional x coordinate does not change
 * as we move across the input and output images.
 *
 * The run starts when xf is in its legal range and yf is just barely less
 * than 1/2.  As we move along the run, xf does not change and yf decreases
 * towards -1/2.  The run ends at I(n-1) whose fractional x coordinate is still
 * xf and whose fractional y coordinate is almost -1/2 (relative to the
 * output line).
 * The length of this run is
 *  - ([final yf] - [original yf])/sin(rotAngle) 
 * = ([original yf] - [final yf])/sin(rotAngle).
 * The original value of xf and the final value of xf are the same.  That
 * value does not appear here. Therefore, the length of the run is
 * independent of xf. Q.E.D.
 *
 * What do we do with this information?  Well, first it tells us that we
 * can ignore the fractional x coord in the inner loop.  That's a nice
 * simplification.  What else happens in the inner loop?
 *
 * Several conditions can cause us to break a contiguous run of bits:
 *  1. Reach the end of an input word
 *  2. Reach the end of an output word
 *  3. Reach the end of the current run
 *  4. Reach the left edge of the input image.
 *  5. Reach the right edge of the input image.
 *  6. Reach the right edge of the output image.
 *  7. Reach the top edge of the input image.
 *  8. Reach the bottom edge of the input image.
 *
 * Luckily, the last two conditions can only occur when we move up or down
 * in the input image, which only happens on a run boundary, thus are
 * subcases of case 3.
 *
 * The remaining 6 cases are bad enough.
 *
 * Furthermore, the input staircase may be shifted arbitrarily in a
 * horizontal direction with respect to the output line.  Since
 * cos(rotAngle) = 1, the horizontal distance between an input bit and
 * its output position will be constant for all input bits, but will
 * vary from one output line to the next.  Most of the shift distance will
 * be handled by the difference in input and output image pointers.
 * However, composition of output words will usually require shifting.
 *
 * Finally, runs may begin and end at arbitrary bit boundaries.  That means
 * the inner loop must keep track of shift amounts and use masks to
 * merge adjacent input runs into contiguous output words.  It must also
 * decide whether the next set of input words comes from the word to the
 * left, the word above, or the word above and to the left of the previous
 * input word.  Having decided this, it must decide whether the new input bits
 * are to be shifted right or left and how many of them should be combined
 * with the output word.
 *
 * How do we make sense of this mess?  Clearly, our code should spend most
 * of its time moving pixels and not so much time figuring out what to
 * do next.  We need some way of organizing things so that deciding what
 * to do next is relatively simple.
 *
 * Using constants instead of variables is usually a good idea.  Say that
 * instead of organizing the code so that each loop iteration transfers a
 * run of bits, we organize it so that each iteration computes a 32 bit
 * output word.  The x coordinate will always change by 32 in both worlds
 * (because of our assumption that cos(rotAngle) = 1).  The input's y
 * coordinate in the output's frame of reference always changes by
 * (32 * sin(rotAngle)).  That's a constant for any given rotation.  Things
 * are already looking up, we don't need to find the new y coordinate
 * in a table.
 * 
 * Any given output word is formed from contiguous sets of bits.  Each
 * set comes from one input word.  What do we know about the input words
 * and the sets?  Say the input words are stacked vertically so that in every
 * input word, the set of bits that went to the output word is aligned
 * with its final position in the output word.  This is a graphic depiction
 * of the staircase of inputs being mapped to an output.  One interesting
 * feature of this stack of input words is that all the bit 0 positions
 * line up. That's because every horizontal move of n bits in output space
 * corresponds to a move of exactly n bits in input space.  In fact, for
 * any given scan line, the position of bit 0 of the input words relative
 * to the output words will be constant across *all* the output words.
 *
 * What can we tell about the runs?  A really interesting feature is that
 * for any given rotation, there are only two possible run lengths.  For
 * example, say we wish to rotate by 3 degrees.  This means we should move
 * up one row in input space for every 19.1 bits we move horizontally.
 * The two possible run lengths are therefore 19 and 20.  Why is this
 * interesting?  It means that we don't (usually) need to determine the
 * the length of the current run from a fancy table, we can have the table
 * tell us whether to use run length A or run length B.
 *
 * Why do we say "usually"?  The very first run of input bits put into an
 * output word is often the remains of the last run put into the previous
 * output word.  That means it may range in length from 0 to MIN(max run
 * length, 32).  So, for each word, we need to know how long the first run
 * is exactly, then know whether each of the succeeding runs is of length
 * A or length B.  How many runs can fit into a 32 bit word?  We'll
 * assume that this code is only called for rotation angles of less than
 * 10 degrees.  That corresponds to runs of length 5.75 degrees.  5.55
 * such runs can fit into a 32 bit word, so we'll say 6.  That's not the
 * end of it.  Say the very first run only has 1 bit in it.  This is
 * followed by 5 runs, each of length 5.75 bits, for a total of 28.75 bits.
 * We've produced 1 + 28.75 = 29.75 bits.  There's clearly room for one
 * more run, for a total of 7 runs.  Therefore, run information for any 
 * given 32 bit word should contain at least 5 bits to specify the length 
 * of the first run plus 6 more bits to specify the lengths of the 
 * succeeding runs, for a total of 11 bits.  (Unless we find another way
 * to encode runs, which we will.  But this keep us moving down the path.)
 * 
 * Say bit 0 in the input words corresponds to bit n in the output words.
 * Let's call all the bits to the left of bit n in the output words the "left"
 * side of the output word.  Similarly, all the bits from bit n to the
 * right end of the output word are called the "right" side of the
 * output word.
 * 
 * Filling the output word can be done in four stages:
 *   1. Fill in first run.  This goes into the left side of the output
 *	word, though it may overlap into the right side.
 *   2. Fill in any remaining runs that start in the left side.
 *   3. If it exists, fill in the remaining part of the last left side run
 *	into the left portion of the right side of the output word.
 *   4. Fill in any remaining portions of the right side of the output
 *	word.
 * 
 * One pleasant feature of this arrangement is that execution proceed fairly
 * smoothly from one state to the next.  States 1 and 3 may be skipped if
 * the preceeding runs ended exactly on word boundaries.  This is pretty
 * easy to determine.  In any case, we don't need a fancy lookup table with
 * multi-way dispatch to decide what to do next.
 * 
 * That's all fine for the inner loop.  What is the basic structure of
 * the code?  We'll have an outer loop that calls this inner loop once
 * per output line.  The outer loop will be responsible for finding the
 * initial position, in both x and y, of the first input bit.  What does
 * this entail?
 *
 * By our assumption that cos(rotAngle) = 1, moving down one line in the
 * output image requires us to move down exactly one line in the input
 * image.  That's easy.  We add sin(rotAngle) to the current x coordinate
 * (relative to the input image).  That tells us when to move over by a
 * bit in the input.  Do we use exactly the same fractional y coordinate
 * for the start of all input lines?  Our assumption that cos(rotAngle)
 * equals 1 says we should.  We'll try it and see what the results look
 * like.  Using exactly the same fractional y coordinate at the start of
 * all input lines means we'd get exactly the same sequence of run lengths
 * in every line.  Is this is good idea?  Well, if the lines started with
 * the same integral x coordinate, it seems swell.  The stairsteps of the
 * two paths through the input will dovetail perfectly.  Every input bit
 * will appear in exactly one output pixel which seems nice.
 *
 * However, say the starting point of one input line is both down 1
 * and over 1 from the starting point of the previous input line.  Say
 * the rotation direction is clockwise, so the stairsteps rise gradually
 * as we move to the right (this diagonal path is reproduced as a
 * horizontal line in the output image, rotating it in a clockwise direction).
 * The vertical path taken by the left ends of the staircase moves gradually
 * to the right as we go down the image.  Relative to the staircase that
 * nests into the ones above it, our new staircase will be displaced 1 bit
 * to the right.  This means it will ignore 1 input bit at every transition
 * between runs.  How bad is this?  Well, I really don't know, but here's
 * a rationalization for deciding it's ok.  Say the pixel skipped is a
 * black pixel in the midst of black pixels or a white pixel in the midst
 * of white pixels.  Clearly, the effect will not be noticeable.  Say the
 * pixel skipped was part of a big edge.  The edge will have a jaggie at
 * the point where the pixel was skipped.  This is what you expect when
 * you rotate an image, so I can't get too excited about it.  Say the pixel
 * was black, but surrounded by white pixels or vice versa.  It seems to me
 * that there are two subcases.  If the pixel was noise, removing it is
 * a fine thing.  If the pixel is part of a carefully arranged set of dots
 * intended to portray a gray level, we are essentially sampling a random
 * set of data.  Say there are b black dots and w white dots in the region
 * of interest.  Say we ignore some fraction i of the dots.  The final
 * ratio of black to white dots will be b*i / w*i = b/w just as before.
 *
 * Clearly, we'll get in trouble if there is an input pattern that matches
 * the pattern of dots being ignored.  I guess we'll hope this hardly
 * ever happens.
 *
 * All this brings up another nice simplification.  If we start *all* lines
 * with the same fractional y position so we get exactly the same sequence
 * of run lengths for all output lines. we can create a lookup table that is
 * indexed simply by horizontal position in the output image.  This is both
 * much smaller and easier to index than a fully general lookup table.
 * Furthermore, since the very first fractional y coordinate is probably
 * arbitrary, we can choose any one we like; 0 for example.  This means we
 * can dispense with keeping the fractional y coordinate around at all!
 *
 * Well, gee, if the table is just a list of run lengths, why not have 1
 * byte per run length?  Because some runs are longer than 255 bits.  Ok
 * why not have one 16-bit thing per run length?  Because we'd still need
 * to generate a mask for each run length.  That will take time.  If we
 * branch on 1 bit specifying short/long run, we can go to code that has
 * both the run length and the appropriate mask in registers all ready to go.
 *
 * Ok, then why not just have a single long bit array that contains
 * nothing but run length specifiers?  Because the code would have to
 * compute how much run was left over from the previous word on every
 * new word.  That takes longer than just looking up the answer (right?).
 * The real trick is to simulate processing a line and record all the
 * information that would be useful at each step.
 *
 * From that point of view, say the preprocessing step outputs a sequence
 * of properly shifted masks.  Each mask is correctly positioned for the
 * output word on which it will be used.  There's one entry in the sequence
 * for each run plus on entry for each output word crossing (the masks
 * necessarily end at word boundaries).  The code tests each mask to
 * see if it should change states.  It can do this by ANDing the
 * boundary bits for both input a output words with the masks to see if
 * the boundary bits have just been filled.  The boundary bit for
 * output words is always the LSbit.  If a mask has that bit turned on,
 * we're done with this output word.  Similarly, if bit 0 (LS Bit) of the input
 * word corresponds to bit n of the output words and bit n is a 1 in the
 * latest mask, we're done with the left side of the output word.
 * Further inputs will be shifted right instead of left to align them.
 * If bit n-1 is also turned on, we should use the input word to the 
 * right to complete the current run.  Otherwise, move up/down a row in 
 * the input image.
 *
 * There's still a problem with this.  How do we tell the difference
 * between a run that *ends on* an output word boundary and a run that
 * *crosses* an output word boundary just by looking at the masks?  After
 * all, masks must end on output word boundaries.  Since masks may be
 * 32 bits wide, we can't hide secret bits inside the masks.  It seems like
 * the easiest thing to do is to follow each output-word-ending mask
 * with the number we should add to the current source address to
 * get the next source word.  Normally, runs cross output word boundaries
 * so this number is 0.  If the run ends on an output word boundary,
 * this number is +- sourcePitch.  This seems simpler than having yet
 * another table (with its own base and index).
 *
 * Yes, we'll need a duplicate set of code that's executed when the input
 * words are exactly aligned with the output words.  This set of code will
 * just ignore the "right side" processing.
 *
 * What happens at the edges of the input image?  The rotated image will
 * have triangles of new stuff at its edges corresponding to areas that
 * were off the edge of the input image.  By fiat, this stuff should be
 * white (like the platen).  In binary images, it should therefore be
 * 0's (1's in gray images).  Frames are filled with 0's, so we're done,
 * right?  No.  We will often be much further off the edge of the image
 * than can be handled by the frames.  A nice way to decide whether the
 * next input word is off the edge is to maintain two counters, xCount 
 * and yCount.  The xCount is incremented as we move left, yCount is 
 * incremented as we move down.  The counters are initialized
 * at the beginning of each line with the distance to the next image edge,
 * sort of.  We would like it to be true that when either counter is
 * negative, we're off the input image.  Therefore, when we're to the left
 * of the input image, xCount contains the x coordinate, a negative number.
 * The same is true of yCount when we're above the image.  However, when
 * we're in the middle of the image, we'd still like to use xCount and yCount
 * to detect the right and bottom edges.  How?  We use a dirty trick.  We
 * set them to 0x80000000 - (distance to edge).  We increment these as we
 * move.  When we reach the edge, they overflow from the largest positive
 * number, 0x7FFFFFFF, to the largest negative number, 0x80000000.  Yeah, it's
 * overflow, but it's still a negative number.  This way, we can always
 * blindly increment the counters and test for negative values.
 *
 * As we learned in the gray rotation case, it is just barely possible to
 * encounter both the left and right edges of the input image in a single
 * sweep.  That's because we'll read a few more input bits than output bits.
 * So, when xCount goes from negative to positive on the left side of the
 * image (xCount < 32), we change it to 0x80000000 - image width - xCount.
 *
 * Let's get back to the structure of the code.  There will be one routine
 * that produces one output line.  It will be called in a loop that locates
 * the leftmost input bit.  Having found the leftmost input bit, we set
 * up xCount, yCount, the original source address and a mask containing
 * a bit in the output bit position corresponding to input bit 0.  Together
 * with address of the sequence of masks, these are passed to the routine
 * that produces one output line.
 *
 * Here's a really annoying detail.  It's of interest only because I got it
 * wrong the first time.  When we move horizontally in the input image, 
 * *exactly* how do we adjust the coordinates of the input pixel in the output's
 * frame of reference?  Our convention is that positive rotation angles imply
 * clockwise rotation.  The cosine function is symmetric and positive around
 * 0.  As agreed, we will approximate cos(rotAngle) with 1. The sine
 * function crosses through (0, 0) with a positive slope.  To
 * rotate an image clockwise, we sweep across it in paths that go from lower
 * left to upper right.  These paths put out horizontal pixels in the rotated
 * image.  Therefore, relative the *rotated*, or output, image, a move to the
 * left in the input image is a move towards the *lower* right in the output
 * image.  I always get confused by inversions.  Here's a table:
 *
 *	Direction in		Coord. mods in
 *	Input Image		Output Image
 *	  +X (to the right)	X = X + 1; Y = Y + sin(rotAngle)
 *	  -X (to the left)	X = X - 1; Y = Y - sin(rotAngle)
 *	  +Y (down)		X = X - sin(rotAngle); Y = Y + 1
 *	  -Y (up)		X = X + sin(rotAngle); Y = Y - 1
 *
 * We can go the other way by reversing the sign of rotAngle, which
 * reverses the sign of sin(rotAngle):
 *
 *	Direction in		Coord. mods in
 *	Output Image		Input Image
 *	  +X (to the right)	X = X + 1; Y = Y - sin(rotAngle)
 *	  -X (to the left)	X = X - 1; Y = Y + sin(rotAngle)
 *	  +Y (down)		X = X + sin(rotAngle); Y = Y + 1
 *	  -Y (up)		X = X - sin(rotAngle); Y = Y - 1
 *
 *
 * Note that these formulas work for both positive and negative rotation
 * angles.  In these formulas, the cosine term is the major change of
 * direction and the sine just specifies the tilt.
 *
 * How do we find the very first input point?  This is a function of both
 * the rotation angle and the point about which the image should be rotated.
 * for example, if we rotate around the upper left corner, the first input
 * point might always be the upper left corner!
 *
 * In the gray rotate code, we always rotated about the center of the image.
 * That was usually ok, but there have been grumblings.  For example, if 
 * the scanner returned an image that contained a smaller, tilted, piece
 * of paper, it would be nice to rotate just the piece of paper and move
 * the rotated image to the upper left corner of the result.  The only
 * problem with this scenario is that we don't know how to find the piece
 * of paper inside the image.  For now, we'll just rotate around the center
 * as before.  If someone wants some other sort of rotation, they can
 * ask.
 *
 * Yeah, but how do you find the first input point?  Ron Danisewicv at
 * XIS found a MUCH simpler way of deriving this, and it's correct too!
 * The new way keep the point about which rotation happens in the center
 * of the picture.  It used to be a function of the rotation angle.
 *
 * To do this we find the equation that lets us compute the coordinates
 * of the source pixel that will appear at a given destination pixel.
 * These equations are based two things.  First is the direction
 * equations given above that tells us how to relate output moves to
 * input moves.  Second is the choice of one pixel that has the same
 * coordinates in both the input and output frames of reference.  If
 * we start from this point in both frames and move to (0,0) in the
 * output, we can calculate where we'll end up in the input.
 *
 * The pixel that has the same coordinates in both frames of reference
 * is clearly the center of rotation.  Let's always put this at (wS/2, hS/2).
 *  
 * Now, using the update equations we derive the source coordinate that
 * we reach when we move wS/2 pixels to the left and hS/2 pixels up
 * in the output.  Remember that each step to the left in the output
 * is a set of (-1, sin(rotAngle)) in the input.  Similarly, each
 * step up in the output is a step of (-sin(rotAngle), -1) in the input.
 *
 * Therefore, starting at (wS/2, hS/2) in both coordinate systems and moving
 * left by wS/2 in the output results in a new set of coordinates in the
 * input as follows:
 *	old input coords:	(wS/2, hS/2)
 *	input x coord after moving left by wS/2 in the output:
 *				wS/2 + (wS/2) * (-1) = 0
 *	input y coord after moving left by wS/2 in the output:
 *				hS/2 + (wS/2) * sin(rotAngle)
 *
 * Now move up by hS/2 in the output and trace where the input goes.
 *	current input coord:	(0, hS/2 + (wS/2) * sin(rotAngle))
 *	final x coord after moving up by hS/2
 *				0 + (hS/2) * -sin(rotAngle)
 *				= -(hS/2) * sin(rotAngle);
 *	final y coord after moving up by hS/2
 *				hS/2 + (wS/2) * sin(rotAngle) + hS/2 * (-1)
 *				= (wS/2) * sin(rotAngle)
 *
 * So hooray, the coordinates in input space of the pixel that corresponds
 * to (0,0) in the output space are (-(hS/2) * sin(rotAngle),
 *				     (wS/2) * sin(rotAngle))
 *
 * Too easy, huh?
 *
 * That's about it.  The basic steps of this routine are:
 *   1. Boilerplate (strip off frame bits, etc).
 *   2. Make the table containing sequences of masks and source address
 *	offsets.
 *   3. Find the first (upper left) input pixel using the calculations
 *	described above.
 *   4. Loop through all the output rows, calling the proc that produces one
 *	row at a time for each row.  After the proc returns, calculate
 *	the position of the next input pixel.
 *
 * *********************** FIRST REWORK ***************************
 *
 * That was very nice.  It produced nice images and was pretty fast.  But
 * not fast enough.  One (minor) way to speed things up is to rework the
 * routine that creates the masks.  In the first version, it found each
 * run length by dividing the distance from the current vertical position
 * to the top of the line of pixels by sinRotAngle.  We don't need to
 * do all those divides (and corresponding multiplies to advance the
 * current position).
 *
 * There are really only two possible run lengths.  Say the exact run
 * length is 29.3.  There are only two possible run lengths, 29 and 30.
 * each time we use 29, we're 0.3 under the perfect amount.  Every time we
 * use 30, we're 0.7 over the perfect amount.  Say we keep a running sum
 * of these differences (-0.3 when we use the short run, +0.7 when we use
 * the long run).  Whenever the sum is negative, use a long run, otherwise
 * use the short run.  Too easy, right?  All this happens in
 * ip_makeNoInterpRotTable.
 *
 * Implementing that idea that made me feel better, but didn't have any
 * perceptible effect on performance.
 *
 * The next point to note is that our code checks every input word to see
 * if it's off the input image.  Almost all words are not off the input
 * image.  Isn't there some way to skip the test most of the time?
 *
 * Our current solution is to divide each code for procesing lines into
 * three sections.  The first section checks (and maintains) both xCount
 * and yCount.  It is used near the top and bottom of the image and at
 * the left edge of lines that start off the left side of the input image.
 *
 * The second section of code maintains xCount, but fetches input words
 * without checking xCount.  This section of code is executed
 * either after crossing the left edge of the image (so we know we won't be
 * crossing the right edge) or when we start the input line in the input
 * image.  We should be spending by far the most time here.
 *
 * The third section of code maintains xCount and checks it when fetching
 * input words.  This section of code is used when we're getting ready to
 * cross the right side of the image, or have already crossed it.  This
 * section of code ignores yCount.
 *
 * How do we get from section to section?  We pass two boundaries down to
 * ip_rotateNoInterpLine.  These boundaries refer to the number of output
 * words put out.  ip_rotateNoInterpLine executes the first section of code
 * until is has put out enough words to pass the first boundary, then (if
 * any words remain on the line) executes the second section until is 
 * passes the second boundary, then, if any output words remain, it 
 * finishes the line with the third section.
 *
 * In practice, the boundaries will be set as follows:
 * 1. Near top or bottom of image: execute first section of code only.
 * 2. Starting to left of input image: execute first and second sections only.
 * 3. Starting inside image: execute second and third sections only.
 *
 * I toyed with the idea of generating lines either from left to right (if
 * line goes off the left side of the input image) or right to left.  This
 * was more complicated in that we had to remember both the left and right
 * endpoints of the lines.
 *
 ***************************** NICE ROTATION **************************`
 *
 * That was all really cool, it worked fast and generated nice output.
 * But isn't there a way to reduced the more apparent jaggies?
 *
 * As discussed above, this isn't an easy problem.  Looking at an image
 * that's been rotated by a very small angle, as happens when the paper
 * moves a tiny bit when the top of the scanner is opened or closed, one of
 * the most obvious problems is that characters become skewed.  The top
 * part of the characters is moved over by a pixel relative to the bottom
 * of the characters.  This results in unintentional italics.
 *
 * It would be cute to do all the horizontal shears in the white space
 * between the lines of text, wouldn't it?  Neato.  What about tall vertical
 * lines?  Well, we have to shear them because that's what rotation is
 * all about.  Tough.  But avoiding shearing the text would be cool.  This
 * does point out the fact that we need some way to avoid shearing sometimes,
 * but eventually do the shearing when it has to be done.
 *
 * How?  Say we AND each input word with the word directly above it.  That
 * identifies the bits that form vertical lines.  Say we shift both the
 * real input and the ANDed inputs into position relative to the new output
 * We then AND the shifted vertical bits with the complement of the
 * destination word written directly above where this one goes.  If any
 * of the input bits that used to line up vertically are not aligned with
 * the output vertically, the result will be non-zero.  This means that
 * bits that were aligned in the input would not be aligned in the output
 * if this version of the input were inserted into the output.  If this is
 * true we blindly shift the input back by 1 before inserting it into
 * the output.  If this is a tall vertical line, one shift will only fix
 * the problem for a few more output lines (those lines where the difference
 * between the ideal output and the current output is 1 pixel).  That's ok,
 * This will just cause the jaggies to move down on the page.
 *
 *
 * There's another, more informative, block of comments just above
 * ip_rotateNoInterpLineNice.
 *
 ************************************
 * Applying ip_rotateNoInterpLineNice to gray images is a bit more challenging.
 * We have to shift and open things by depth, not by 1.  All the masks
 * become functions of the depth.  The value of "white" changes. It's
 * generally a pain.
 *
 * The gray algorithm will require variables where the binary rotation
 * algorithm has constants.  This uses more registers, generally slowing
 * things down.  The binary version of rotation will be useful for all
 * fax applications and most applications where speed and storage are
 * concerns.  That includes almost everything on the PC.  The gray
 * rotation will be used primarily for 2 bit/pixel images in Towada.
 * Having thought about this, we'll make two routines
 * ip_rotateNoInterpLineNiceBinary and rotateNoInterLineNiceGray.
 *
 *************************************************************************
 *
 * All of the preceeding comments came from the C implementation.  They're
 * true here too.  The difference between the pure C code and this code is
 * that ip_rotateNoInterpLine is implemented in assembly here.
 *
 * *****************   ROTATION IN PLACE  *********************************
 *
 * This code give very nice results, but uses a lot of memory.  We
 * require full size images in both source and destination..  At 300 spi,
 * that's two 1 MB images.  At 400 spi, 11"x17", that's two 4MB images.
 * We haven't been thinking of these memory demands as extraordinary, but
 * PC users and XIS think they're big.
 *
 * We've already implemented gray rotation in place.  We'll use the same
 * algorithm.  We'll repeat the edited comments here.
 *
 * Z Smith, that font of creativity, suggested that we allocate more memory
 * than actually required for the image data in the image pixrect; then
 * rotate the image to the empty destination space.  We'd roll along, so
 * at the end would have the rotated image in some displaced position relative
 * to the source image.
 *
 * That idea was ok, but our image data structures didn't know about blank
 * spots.  So, we'll try a different approach.  i_rotateNoInterpolationIP
 * will allocate a ring buffer.  The optimized routines ip_rotateNoInterpLine,
 * ip_rotateNoInterpLineNiceBinary and ip_rotateNoInterpLineNiceGray will put
 * rotated data into the ring buffer.
 *
 * Rotation will be done in three stages:
 *  1.	Rotate enough lines to fill the ring buffer.  At this point, we
 *	should no longer need to use the first line of the original image
 *	as input data.
 *  2.	There are two steps for each rotated line:
 *	a. Copy the next line from the ring buffer to the image.
 *	b. Fill the empty space in the ring buffer with a new rotated line.
 *	This continues until we have produced enough rotated lines.
 *  3.	In the last stage, we flush the rest of the lines in the ring buffer
 *	to the image.  This should end when we have overwritten the last
 *	image line.
 *
 * So, how big is the ring buffer?  Obviously, it must be wide enough to
 * hold rotated lines.  For the sake of simplicity, we'll make it sBpl bytes
 * wide.  How tall should it be?  We can compute the maximum vertical
 * distance a diagonal path across the input image will cross.  This is
 * equal to wS * sin(rotAngle).  The ring buffer must be at least this tall
 * because when we have just stopped touching the top of the image, the
 * lower edge of the diagonal path will be this distance below the top.  We
 * assume that each source line produces about one dest line, so there must
 * be at least this many dest lines in the ring buffer.
 *
 * When we move down one line in the dest, we only move down cos(rotAngle)
 * lines in the source.  This is very close to 1, but not quite.  Therefore,
 * When we move down hS lines in the dest, we'll only move down
 * hS * cos(rotAngle) lines in the source.  To fill a 4400 line image at an
 * angle of 10 degrees, we only cross 4333 lines of source image.  This means
 * the distance between the source image's yCoord and the dest image's
 * yCoord keeps growing as we move down the image.  If the ring buffer were
 * only tall enough to cover the number of lines needed to span the
 * vertical displacement of one path across the source image, we would
 * start writing dest lines on top of source lines that had not yet been
 * retired by the time we reached the bottom of the pile.  To account
 * for this, we look at the difference between cos(rotAngle) and 1.  At
 * a maximum rotation angle of 10 degrees, this is 0.0152, or about
 * 1 part in 65.  To be safe, we'll add hS/32 lines to the ring buffer
 * height.
 *
 * Are there any other considerations?  Do we need to put in extra lines
 * because of starting or ending conditions?  If we started n lines above
 * the top of the source image, we'd need to add n/cos(rotAngle) extra
 * lines to the ring buffer.  The current code always rotates around the
 * middle of the image, so we never start above the top of the image.  I
 * can't think of any ending conditions that would make a difference.
 *
 * Therefore the ring buffer height should be (wS * sin(rotAngle)) +
 * hS / 32 lines tall.
 *
 *********************** ROTATION IN PLACE REWORK ***********************
 *
 * Really close.  Except for "nice" rotation.  This flavor of rotation
 * compares each destination line with the line above it.  Since rotation
 * in place uses a ring buffer, this fails when we wrap around since the
 * line above the new first line is at the end of the ring buffer.  Ok,
 * so the results would be a bit off.  The real problem is that the
 * algorithm tries to reference data that may or may not have been allocated
 * since it's outside the ring buffer.
 *
 * How to fix?  An obvious idea is to copy the last line up to the first
 * line each time we wrap.  This works so long as there is never good data
 * in the first line to destroy.  Since we already have some slop in the
 * hS / 32 factor, we'll just start initially and wrap around to the second
 * line of the ring buffer.
 *
 ************************* FIXED POINT IMPLEMENTATION ********************
 *
 * XIS supports platforms that have no floating point support.  They have
 * supplied fixed point implementations of sin, cos, atan and sqrt.  These
 * functions are mainly used to generate sinRotAngle and to find the
 * input point that corresponds to output point (0,0).  We substitute
 * the fixed point functions for their floating equvalents in a
 * straightforward way.
 *
 ********************** FIX FOR ANISOTROPIC RESOLUTIONS *****************
 *
 * Standard Fax images have anisotropic resolutions, like 200 x 100 dpi.
 * If you take a skewed image with square pixels and take out ever other
 * vertical row you introduce a horizontal shear.  Huh?  Well, reducing
 * the image in the y direction makes everything more horizontal.  If the
 * image contained (slightly skewed) horizontal and vertical lines, the
 * skewed horizontal lines become more horizontal and so do the vertical
 * lines.  For the vertical lines, this means they become less vertical.
 * the result is that the angle between the formerly orthogonal lines is
 * no longer 90 degrees.
 *
 * What to do?  There is no rotation angle which will cause the lines
 * to appear to be orthogonal.  Obviously, the reverse situation is also
 * true - if the lines were orthogonal in the anisotropic image and we
 * rotate it, then make it isotropic, they will no longer be orthogonal.
 * Our solution is to introduce a horizontal shear into the rotation
 * algorithm.  The client should specify the rotation angle that makes
 * the horizontal components of the rotated image look correct (i.e. the
 * skewed horizontal lines become really horizontal) and we will shear
 * the image so that, if one were to make the image isotropic, it would
 * look right.
 *
 * How much shear is that?  Well, say you have an isotropic image that's
 * rotated by theta.  We then make the image anisotropic with horizontal
 * resolution xRes and vertical resolution yRes.  If we run the skew
 * detect code on the new image, we get a skew angle of phi such that
 * sin(phi) = (yRes/xRes) sin(theta).  Since we calculated the skew
 * by looking at horizontal things, phi is the angle we send to this
 * routine.  It will make the skewed horizontal things really horizontal.
 *
 * In the rotation routines, we convert diagonal lines of image input
 * into horizontal lines of output.  Therefore, we happily move up or
 * down by sin(phi) for every horizontal step.  The shear comes in 
 * when we choose the left end of each diagonal line in the input image.
 * In the isotropic image, we would have moved over by sin(theta) for
 * each step down.  In the anisotropic image, we have to move over by
 * (xRes/yRes) sin(theta) = (xRes/yRes)*(xRes/yRes)*sin(phi).  Therefore,
 * we multiply the horizontal motion by (xRes/yRes) squared.
 *
 * Since we're moving around in the image in a weird way, how do we figure
 * out where the center is now?  We figure out where the center would
 * be in the isotropic image and multiply coordinate corresponding to the
 * smaller resolution by smallRes/largeRes.  For example, if xRes = 2*yRes,
 * We multiply the image height by 2 when initially figuring the center,
 * then divide the resulting y coord by 2 to get the new center.
 *
 * ***********************************************************************
 * BUG FIX, 9/20/95:  Richard Mlynarik found that rotating gray images
 * sometimes produced artifacts on the right side.  After further
 * investigation, I found that there's a generic problem with the way
 * we handle the rightmost word of an image.  If the right boundary of
 * an input image does not end on a word boundary, we sometimes use the
 * entire word (which doesn't work when the frame is not white) and sometimes
 * ignore the rightmost word entirely.  No good.
 *
 * The problem has the do with confusion about xCount.  In the gray rotate
 * code, we read one pixel at a time and updated xCount each time.  In
 * this code we read 32 bits (some number of pixels) each time.  xCount
 * indicated the distance from some pixel inside the word to the edge
 * of the input image.  This uncertaintly threw us off.
 *
 * In the new scheme, we truncate xCount to the nearest multiple of 32
 * towards -infinity.  This makes it calculate the distance from the leftmost
 * bit in the word being read to the edge of the input image.  We then
 * check the value of xCount in the third section of code that is used when
 * we're approaching the right edge of the image.  If xCount is < 0 (greater
 * than 0x80000000), we know the entire word being read is off the right
 * side of the image and just use whiteval.  This is the common case
 * there.  If xCount is positive, we now have two cases where before we
 * had one.  If xCount + 32 is also postive, we read an entire new word
 * from the source image.  If xCount+32 is negative ( >= 0x80000000 ),
 * we merge the word read from the image with whiteval using a new mask.
 * The mask is a constant computed from the width of the image in bits.
 * We can use the constant since the width of the source image (the one
 * from which we are reading the data) is constant.  In fact, since we
 * know that the stuff off the edge of the source image is frame, therefore
 * filled with 0's, we can just supply a rightWhite mask which is filled
 * with whiteVal bits in the frame section of the input word.  We just
 * OR that with the partial input word read.  The rightWhite mask is
 * computed in the lower level routines since not all of them take a
 * whiteVal anyway.
 *
 ******************* SECOND ANISOTROPIC IMAGE REWORK *********************
 *
 * The alert reader (not me) will notice that the explanation for the
 * behaviour of anisotropic rotation depends on the fact that the naive
 * skew detect code returned an angle which depended on the ratio beteen
 * the x resolution and the y resolution, i.e. (yRes/xRes) * (skew of
 * the original sheet of paper).  Knowing this, the rotation routine multiplied
 * the angle (the sine of a small angle is almost equal to the angle)
 * by (yRes/xRes) before using it.  Well, this is ducky if the rotation
 * angle came from a naive skew detection routine.  What if it came from
 * somewhere else?  Or, what if we scale the image anistropically before
 * calling the rotation routine?  In either case, the "correction" supplied
 * by the rotation routines will not do the right thing.
 *
 * Our solution is to demand that angles supplied to rotation routines
 * must be independent of x and y resolution.  What does this mean?  Let's
 * introduce the "isotropic cousin" of an anisotropic image  The isotropic
 * cousin of an image is the smallest image which is isotropic
 * (x resolution == y resolution) and can be generated by scaling the
 * original image in either x or y where the scaling factor is >= 1.0.
 * For example, a standard fax image has an x resolution of 200 dpi and
 * a y resolution of 100 dpi.  The nearest isotropic image is a version of
 * the original image that has been scaled by 2x in the vertical direction
 * only, therefore has xRes = yRes = 200 dpi.  The isotropic cousin of
 * an isotropic image is the image itself.
 *
 * We require that the angles that are passed around correspond to the
 * rotation to be performed on the isotropic cousin of the actual
 * image provided.  We perform an operation on the real image such
 * that the result of our rotation-like operation followed by a scaling
 * operation that makes the result isotropic is equivalent to simply
 * rotating the isotropic cousin of the original image.
 *
 * Why this convoluted definition?  Well, when someone looks at an image,
 * it will almost surely be using isotropic resolutions.  We want to make
 * sure that the rotation-like operation we perform will, when scaled for
 * viewing, look like a real rotation.
 *
 * So, what's involved?  First, the skew detection routines must scale
 * the vertical displacements they find by (xRes/yRes) before computing
 * the arctan.  That's because the naive computation found a vert/horiz
 * ratio that was (yRes/xRes) * (vert/horiz for the isotropic cousin of
 * the image).
 *
 * Next, we look at how the rotation code uses the angle and the
 * x and y resolutions.  In this code, we assume the cos(rotAngle) = 1
 * anyway, so we don't change that.  However, we compute two versions
 * of sin(rotAngle).  vertSinRotAngle is used to modify the X coordinates
 * when moving vertically.  horizSinRotAngle is used to modify the Y coord
 * when moving horizontally.  Why do we need both?  Well, think of converting
 * an isotropic image to an anisotropic image by squishing all the individual
 * pixels, then laying the result out over an isotropic grid.  Say we make
 * the input pixels short and wide.  Then, each time we move down in the
 * input image, we have to remember to move over extra far in the output
 * image to get to the right place.  Similarly, each time we move over in
 * the input image, we have to remember that only a small movement is
 * implied vertically.
 *
 * That's basically the deal.  The length of the runs is computed by
 * ip_makeNoInterpRotTable.  A run ends when the center of the next
 * horizontal input pixel goes above the current output row.  Since we
 * want to know about vertical distances when moving horizontally, we
 * pass horizSinRotAngle to ip_makeNoInterpRotTable.  When moving down
 * vertically to find the beginning of the next scan line, we use
 * vertSinRotAngle as before.  When finding the upper left corner of
 * the image, we use both horizSinRotAngle and vertSinRotAngle as appropriate.
 *
 * ************************* FLAT IMAGE BUG FIX **************************
 *
 * In the previous code, I always assumed that the diagonal track would
 * either go from below the image into the image; or into the image and
 * stay in the image; or from into the image to above the image.  In very
 * flat images (wide, but not tall), a single diagonal track might go
 * from below the image, through the image and out the top.  This means
 * that we have to be careful about setting yCount - we check to see if
 * it's below the bottom before checking to see if it's near the top.
 * For flat images, it can be both near the top and below the bottom, but
 * we were only looking at the near the top case.
 * The lower level code will now also have to watch yCount as it's decremented.
 * If it goes from 0x80000000 to 0x7fffffff, it should be set to the
 * source image height - 1.  That means we have to send the image height
 * to the lower level routines, which is unfortunately a new parameter.
 */


/* PUBLIC */
/**************************************************************************
 *  i_rotateNoInterpolationFixedPoint()
 *	Input:
 *		pS = source binary or gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		rotAngle = rotation angle in ScaledAngle's
 *		pD = destination binary or gray image
 *		dBpl = destination bytes/line
 *		depth = bits/pixel in source and dest images.
 *		nice = cTrue if we should do the slower version
 *			of rotation that makes an attempt to
 *			reduce the number of jaggies.
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *
 *	On return:
 *		Destination image contains the source image rotated by
 *		the given angle.  Unlike i_rotateGray, this routine
 *		does not form rotated pixels by averaging the values
 *		of the input pixels that overlap the output pixel.
 *		In this routine, each output pixel is assigned the
 *		value of the input pixel with the greatest overlap.
 *		If nice = cTrue, the rotation routine makes some attempt
 *		to avoid horizontal shears of vertical strokes by adjusting
 *		the horizontal shift amount by 1 if vertical strokes
 *		don't line up.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		unable to allocate memory for tables
 *		ia_nopoint		could not find next point
 *		ia_depthNotSupported	could not rotate source image
 *					of specified depth
 *		ia_angleTooLarge	can not rotate by this angle
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	Frame in destination image is assumed to already be cleared.
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateNoInterpolationFixedPoint (
    IPState    *pIPState,
    UInt8Ptr	pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    ScaledAngle	rotAngle,
    UInt8Ptr	pD,
    Int32	dBpl,
    Int32	depth,
    Bool	nice,
    Int16	xRes,
    Int16	yRes)
{
PROLOGUE;
Int32		line;			/* interation variables */
Int32		xICoord, yICoord;	/* integers containing the
					 * X and Y coords of the current input
					 * pixel in the input image's frame
					 * of reference.  To keep things
					 * simple, xICoord is maintained in
					 * units of bits, not pixels.
					 */
Int32		leftmostBitPosition;	/* keeps track of leftmost source
					 * image bit within a word */
ScaledInt1616	xIOCoord;		/* split integer containing the
					 * X coord of the current
					 * input pixel in the output image's
					 * frame of reference.
					 */
ScaledInt1616	sinRotAngle;		/* split integer containing the
					 * sine of the rotation angle.
					 * These have 16 bit int and frac
					 * parts */
ScaledInt1616	vertSinRotAngle;	/* The sine of the rotation angle
					 * that's been modified to account
					 * for shears in anisotropic images.
					 * This is used to modify the X
					 * coord when moving vertically. */
ScaledInt1616	horizSinRotAngle;	/* The sine of the rotation angle
					 * that's been modified to account
					 * for shears in anisotropic images.
					 * This is used to modify the Y
					 * coord when moving horizontally. */
UInt32		xCount, yCount;		/* distances to vertical and horiz
					 * edges passed to rotateNoInterpLine.
					 * equivalent to input coords relative
					 * to input image. */
UInt8Ptr	pDOrig;			/* used to remember addr of upper left
					 * corner of image so we can clear its
					 * right frame when done. */
UInt32Ptr	pMaskTable;		/* table of masks used to rotate
					 * a single line */
Int32		pixelsPerWord, logPixelsPerByte; /* used to adjust addresses */
UInt32		widthInBits;		/* used to adjust xCount */

UInt32		whiteVal;		/* a 32 bit word holding white
					 * pixels.  These are 0 for binary
					 * images and 1's for gray images */
UInt32		rightWhiteVal;		/* white pixels inserted into rightmost
					 * partial image word. */
Int32		boundary1, boundary2;	/* args passed to line rotation
					 * routine telling it when to check
					 * for the edge of the input image. */
Int32		maxVertDisp;		/* holds the distance a line may
					 * travel vertically in the input
					 * image as it crosses the input
					 * image. */

/* here are the floating point versions of the defines given below */
/*
#define cMaxRotAngle	10.0 * 3.14159 / 180.0
#define cMinRotAngle	-10.0 * 3.14159 / 180.0
*/
#define	cScaledPi	((Int32)16384)		/* 32768 corresponds to 2pi in
										 * scaled angle notation */
#define	cMaxRotAngle	((10 * cScaledPi) / 180)
#define cMinRotAngle	((-10 * cScaledPi) / 180)

    pMaskTable = NULL;

	/* can we handle this depth image? */
    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8))
	return ia_depthNotSupported;

	/* can we handle this angle? */
    if ((rotAngle < cMinRotAngle) || (rotAngle > cMaxRotAngle))
	return ia_angleTooLarge;

	/* Are the resolutions ok? */
    if ((ABS(xRes - yRes) > 1) && ((xRes == 0) || (yRes == 0)))
	return ia_invalidParm;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + depth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + depth*cFrameBytes;	/* skip the gray pad */

	/* Remember address of upper right corner of dest image
	 * so we can clear its right frame when done. */
    pDOrig = pD;

	/* get the width and height of the source without the frame */
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;

	/* set up depth-dependent parameters */
    switch (depth) {
    case 1: /* depth == 1 */
        pixelsPerWord = 32;
	logPixelsPerByte = 3;		/* log of 8 */
	widthInBits = wS;
	whiteVal = 0;
	break;
    case 2: /* depth == 2 */
        pixelsPerWord = 16;
	logPixelsPerByte = 2;		/* log of 4 */
	widthInBits = wS*2;
	whiteVal = 0xFFFFFFFF;
	break;
    case 4: /* depth == 4 */
        pixelsPerWord = 8;
	logPixelsPerByte = 1;		/* log of 2 */
	widthInBits = wS*4;
	whiteVal = 0xFFFFFFFF;
	break;
    default: /* depth == 8 */
        pixelsPerWord = 4;
	logPixelsPerByte = 0;		/* log of 1 */
	widthInBits = wS*8;
	whiteVal = 0xFFFFFFFF;
	break;
    }

	/* If the image does not end on a word boundary on the right,
	 * insert some white pixels.  Otherwise, insert nothing.
	 */
    if ((widthInBits & 0x1F) != 0)
	rightWhiteVal = whiteVal >> (widthInBits & 0x1F);
    else
	rightWhiteVal = 0;

	/* We'd like to use one set of code for both clockwise and counter
	 * clockwise rotation.  If rotAngle < 0, we're doing counter-
	 * clockwise rotation so should process the image starting from the
	 * lower left corner instead of the upper left corner.  Reposition
	 * pS and pD.  Invert rotAngle, dBpl and sBpl.
	 */
    if (rotAngle < 0)
    {
	pS += (hS-1) * sBpl;	/* point to lower left corner */
	pD += (hS-1) * dBpl;
	sBpl = -sBpl;
	dBpl = -dBpl;
	rotAngle = -rotAngle;
    }

	/* compute the sine term we'll need to relate the
	 * source and dest coordinate systems.  Note that fract_sin
	 * returns numbers as ScaledInt1715's and we want ScaledInt1616.
	 * Therefore, we shift the result left by 1 to conver it. */
    if ((__error__ = fract_sin(pIPState,
			       rotAngle, (ScaledInt1715Ptr) &sinRotAngle))
	    != ia_successful)
	NORMAL_RETURN;

    sinRotAngle <<= 1;				/* convert to ScaledInt1616 */
    if (ABS(xRes - yRes) > 1)
    {
	vertSinRotAngle = (xRes * sinRotAngle)/yRes;
	horizSinRotAngle = (yRes * sinRotAngle)/xRes;
    }
    else
    {
	vertSinRotAngle = sinRotAngle;
	horizSinRotAngle = sinRotAngle;
    }
	/* how far up will line go? */
    maxVertDisp = ScaledInt1616_TO_Int32(wS * horizSinRotAngle);

	/* if sinRotAngle <= 2 (sin(rotAngle) <= (0.0000000000000010),
	 * this angle is too small to bother with.  Just copy the source
	 * to the dest unchanged.  There are several places below where
	 * we assume sinRotAngle is non-zero. */
    if (MIN(vertSinRotAngle, horizSinRotAngle) <= 2)
    {
	i_rasterOp (pIPState, pD, 0, 0, wS, hS, depth, dBpl,
		    PIX_SRC, pS, 0, 0, depth, sBpl);
	NORMAL_RETURN;
    }

	/* Make the sequence of masks using modified vals.  Set var
	 * setting type of table to be made according to whether
	 * assembly code is present.  This is done
	 * since even in nice rotation, the very first line is done
	 * by the assembly routine.
	 */
    if ((__error__ =
	    ip_makeNoInterpRotTable (
			pIPState,
			horizSinRotAngle, &pMaskTable, wS,
			sBpl / 4, depth,
			(Int8)ip_rotateAssemblyCodeIsPresent (pIPState)))
			!= ia_successful)
	NORMAL_RETURN;
    IP_YIELD(errorFinish)

	/* compute the initial input position using the algorithm
	 * described in the comments above.  Note that xICoord is
	 * calculated in units of BITS, not pixels.  This allows us
	 * to ignore the depth later on.  The coordinates of the
	 * initial input point are ((-hS/2) sinRotAngle, (wS/2) sinRotAngle)
	 * Since sinRotAngle is currently kept as an Int1616, we have to
	 * shift the result right by 16 after doing the multiply.
	 * Adding 0x8000 is to get the rounding correct.
	 * Remember that if the image is anisotropic (non-square resolution),
	 * we have to use horizSinRotAngle to modify yICoord when moving
	 * horizontally and vertSinRotAngle to modify xICoord when
	 * moving vertically.
	 */

    xICoord = -1*(((hS/2) * vertSinRotAngle + 0x8000) >> 16) * depth;  
    yICoord = ((wS/2) * horizSinRotAngle + 0x8000) >> 16;

	/* This is the input coordinate which we decree corresponds exactly
	 * to coordinate (0,0) in the output frame of reference.  What do we
	 * do with it?  First, we adjust the input address to point to this
	 * place.  This will almost certainly point to something outside
	 * the input image, but that's ok.  We won't do any memory references
	 * with this address until we enter the input image.
	 * Note that if xICoord is a positive number, the (xICoord >>
	 * logPixelsPerByte) rounds down.  This is good since we want
	 * to move only left one word for every complete 32 bits.  However
	 * if xICoord is negative, the (xICoord >> logPixelsPerByte) term
	 * rounds up.  This is also good.  We want to move left one
	 * word for every fragment of a word to the left of us.
	 *
	 * Well, that was all very clever, but shifting integers right in
	 * ANSI C is undefined (although the Sun ANSI Compiler seems pretty
	 * haphazard about issuing warnings).  The processor might
	 * duplicate the sign bit, or it might not.  So, we had to use
	 * "/ 32". If xICoord is positive, we want
	 * (xICoord/32.  If xICoord is negative, we really 
	 * want ((xICoord/32)-1);
	 * NOTE: both xICoord and 32 must be SIGNED or the
	 * dumb Sun ANSI compiler will do an UNSIGNED divide!!!
	 * At this point, rotAngle is always positive, so cos(aspectAngle
	 * + rotAngle) is always less than cos(aspectAngle), so xICoord
	 * is always negative (or 0).
	 */
    pS += yICoord * sBpl +
		( ((xICoord - 31) / 32) * 4);

	/* Set the position of the leftmost input bit in the first word.
	 * Here, leftmostBitPosition is measured relative to the left end
	 * of the word.  So, for example, if xICoord = -1,
	 * leftmostBitPosition is set to 32+(-1) = 31.
	 * If xICoord is more negative that -32, the "OR (-32)" term will
	 * blast the more significant bits.  If we have gray pixels,
	 * xICoord has already been multiplied by depth, so we'll
	 * still get the right answer.
	 */
    leftmostBitPosition = 32 + (xICoord  | (-32));

	/* By decree, the location of this input pixel relative to the
	 * output image is (0.0, 0.0).  Since we assume cos(rotAngle) = 1,
	 * we really only care about xIOCoord.  It will be used to decide
	 * when we've crossed a bit (or word) boundary.
	 */
    xIOCoord = 0;

    xCount = yCount = 0;	/* initialization to make gcc happy.  This
				 * is not really necessary since yCount is
				 * initialized before it is actually used.
				 */

	/* Have ip_rotateNoInterpLine actually do the work on each line. */
    for (line = hS; line > 0; line--)
    {
	    /* Calculate the distance to the first horizontal and vertical
	     * edges that we'll cross.  These are set up so that the count
	     * will be positive inside the input image and negative outside.
	     * The boundaryX values are expressed in units of bits from
	     * the RIGHT image boundary.  That's because the width parm
	     * is decremented by ip_rotateNoInterpLine as it moves from
	     * left to right across the image.
	     */
	if (xICoord < 0)
	{
	    xCount = xICoord - leftmostBitPosition; /* compute negative
						 * of distance from
						 * left edge of word to
						 * left boundary of input
						 * image */
		/* Stop the first section just a bit after starting.  If
		 * the image is very narrow, widthInBits + xICoord might
		 * be (very) negative so we might do the first section
		 * too many times.  we clip boundary1 at 0 so we
		 * always stop soon enough.
		 */
	    boundary1 = MAX((Int32)widthInBits + xICoord, 0);
	    boundary2 = MAX(xICoord + 32, 0);	/* near right edge of image.
						 * make sure we get into
						 * the checking code BEFORE
						 * arriving at the right edge
						 */
	}
	else
	{
	    xCount = 0x80000000 -
		     (widthInBits - xICoord + leftmostBitPosition);
						/* dist from left edge of
						 * word to right side of input
						 * image. */
	    boundary1 = widthInBits;		/* at the left edge (i.e.
						 * skip first code section) */
	    boundary2 = xICoord + 32;		/* near right edge of image.
						 * make sure we get into
						 * the checking code BEFORE
						 * arriving at the right edge.
						 */
	}

	    /* if we're near the top or bottom of the image, check
	     * xCount and yCount all the way across.  Usually, we're not
	     * near the top or bottom, so this stuff doesn't happen.
	     * If neither of these is true, initialize yCount to
	     * something sensible so it doesn't cause trouble in
	     * the first code section.
	     * Check to see if we're near the bottom of the image
	     * first to account for the case where the diagonal
	     * path starts on the bottom, goes through the middle
	     * and comes out the top.
	     */
	yCount = yICoord;		/* the default number of lines
					 * until we hit the top row */
	if (yICoord >= hS)		/* starting near bottom of image */
	{
	    yCount = 0x80000000 - (hS - yICoord); /* starting near bottom of
						   * image */
	    boundary1 = 0;		/* check xCount and yCount across
					 * entire image */
	    boundary2 = 0;		/* doesn't matter */
	}
	else
	if (yICoord < maxVertDisp)
	{
	    boundary1 = 0;		/* check xCount and yCount across
					 * entire image */
	    boundary2 = 0;		/* doesn't matter */
	}

	if (!nice || (line == hS))	/* do it the fast way */
	{
	    ip_rotateNoInterpLine (pIPState, pS, sBpl, pD, widthInBits, xCount,
			           yCount, leftmostBitPosition, pMaskTable,
			           whiteVal, rightWhiteVal,
				   boundary1, boundary2, hS);

	    if (nice && ip_rotateAssemblyCodeIsPresent (pIPState))
		    /* Make a new mask table for the rest of the lines.
		     * The first line was done by the assembly routine,
		     * but the rest of them will be done by C routines
		     * so we need the C version of the table. */
	    {
		FREE(pMaskTable);
		__error__ = ip_makeNoInterpRotTable (pIPState, horizSinRotAngle,
						   &pMaskTable, wS,
						   sBpl / 4, depth, cFalse);
		if (__error__ != ia_successful)
		    NORMAL_RETURN;
	    }
	}
	else				/* try to avoid sheared text */
	if (depth == 1)
		/* ip_rotateNoInterpLineNiceBinary negates what it sees as
		 * sBpl and dBpl when accessing the word that it is
		 * trying to "line up" with to eliminate jaggies.  This
		 * means that if the rotation angle is negative, and
		 * hence we are going from the bottom to the top of the
		 * image, that sBpl and dBpl must have different signs.
		 * As a result we use sBpl and absSBpl for the sBpl and
		 * dBpl respectively.
		 */
	    ip_rotateNoInterpLineNiceBinary (pIPState, pS, sBpl,
					     pD, dBpl, widthInBits,
					     xCount, yCount,
					     leftmostBitPosition,
					     pMaskTable, boundary1, boundary2,
					     hS);
	else
	    ip_rotateNoInterpLineNiceGray (pIPState, pS, sBpl,
					   pD, dBpl, widthInBits,
				           xCount, yCount,
					   leftmostBitPosition,
				           pMaskTable, depth, rightWhiteVal,
					   boundary1, boundary2, hS);

	    /* now find the starting position of the next line */
	pD += dBpl;			/* move down 1 in output space */
	pS += sBpl;			/* and in input space */
	yICoord++;			/* moving down one line in output
					 * space always corresponds to
					 * moving down 1 line in input space.
					 */
	xIOCoord += vertSinRotAngle;	/* and move over in input space. */
	if ((xIOCoord >> 16) != 0)
	{				/* we need to move over in x */
	    xIOCoord &= 0xFFFF;		/* get rid of integer part again */
	    xICoord += depth;
	    leftmostBitPosition += depth;
	    if (leftmostBitPosition > 31)
	    {
		pS += 4;			/* move to next word boundary */
		leftmostBitPosition &=  0x1F;
	    }
	}

	    /* I don't know what's going on here.  The fast, optimized
	     * version of this code fails on a SPARCStation with a bus error.
	     * The unoptimized code works.
	     * The slow code (linked with C implementation of lower level
	     * rotate ops) works.
	     * The PC version of the code works.
	     * Adding enough fprintf's in this module makes it work.
	     * Purifying either the optimized or unoptimized code makes
	     * it work.
	     * The assembly code put out by the compiler for this
	     * module looks ok to me.
	     *
	     * So try inserting random stuff to see if it goes away.
	     */

	if ((line & 0x1FF) == 0)
	    IP_YIELD(errorFinish)

    } /* end of loop that rotates all the lines */

	/* clear the right frame of the rotated image.  It was probably
	 * written mistakenly by the rotation code. */
    if (dBpl > 0)
	ip_clearRightFrame (pIPState, pDOrig, wS, hS, dBpl, depth);
    else
	ip_clearRightFrame (pIPState, pDOrig, wS, hS, -dBpl, depth);

    FREE(pMaskTable);
    NORMAL_RETURN;

errorFinish:
    FREE(pMaskTable);
    return ia_aborted;
}



/*

$Log: i_nirfx1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.14  1998/01/13 23:49:00  ddavies
i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.13  1997/08/20 20:28:37  ddavies
i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v

Change Int8 nice to Bool nice in arg list.

Revision 1.12  1997/06/13 21:23:36  jcraig
jcraig on Fri Jun 13 14:21:32 PDT 1997

i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v

changed comment at the beginning of the module that claimed that
it contained code for i_rotateNoInterpolation when in fact it is
i_rotateNoInterpolationFixedPoint.  change the comment from binary
rotatation routine to Binary/gray rotation routine.  (Really
this is the last checkin for comment changes about this routine that
I'm gonna do.....)

Revision 1.11  1997/06/13 21:12:10  jcraig
i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v

change comments to indicate that i_rotateNoInterpolationFixedPoint
will take both binary and gray images (not just gray).

Revision 1.10  1996/10/05 02:18:31  ddavies
i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v

Change the order in which we check to see if we're near the top or bottom
of an image, since in flat images, we could be near both at the same time.
We now check to see if we're near the bottom first.
Also pass the image height down to the lower level routines so they can
update yCount correctly.

Revision 1.9  1996/02/10 01:33:03  dferrell
dferrell on Fri Feb 9 17:32:26 PST 1996

i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v

Added parens to & operations so that gcc 2.7.2 won't whine.

Revision 1.8  1995/12/13 02:44:54  ejaquith
i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.7  1995/12/13  01:37:15  ddavies
 * i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v
 *
 * Rework the way anisotropic images are rotated.  In the old code, we assumed that
 * the angle passed in was calculated in a naive way and was really equal to
 * (real angle) * (yResolution / xResolution).  That's because if the y resolution
 * was small, the image would be squished vertically so all vertical displacements
 * would look small to the skew detect code.  The old code multipled the incoming
 * angle by (xResolution/yResolution) before use (sort of).  Ok, what it really
 * did was use two versions of sin(rotAngle): vertSinRotAngle = (xRes*xRes)/(yRes*yRes)
 * times sinRotAngle and just plain sinRotAngle.
 * In the new code, we assume that the angle is resolution independent so we
 * can get angles from anywhere.  The cosine of the rotation angle is not really
 * affected by this since the angles are so small (in fact, we always assume that
 * it's equal to 1 anyway).  However, we
 * still use two versions of sin(rotAngle): vertSinRotAngle = (xRes/yRes) *
 * sinRotAngle and horizSinRotAngle = (yRes/xRes) * sinRotAngle.  We use
 * vertSinRotAngle in EVERY instance where vertical motion requires adjusting
 * the x coordinate.  We use horizSinRotAngle in EVERY instanced where horizontal
 * motion requires adjusting the y coordinate.  These are used when calculating
 * the initial input position, when building the rotation tables and when
 * adjusting the input position in preparation for rotating the next
 * scan line.  Actually, by making things more consistent, it's easier to
 * figure how what's going on.
 * These changes apply to both nice and non-nice (fast) non-interpolated
 * rotation.
 *
 * Revision 1.6  1995/10/18  03:57:43  ddavies
 * i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.5  1995/09/21  02:53:07  ddavies
 * ddavies on Wed Sep 20 19:48:47 PDT 1995
 *
 * i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v
 *
 * Fixed a bug in i_rotateNoInterpolationFixedPoint.  This routine was either
 * ignoring the rightmost word of an image or inserting the entire word
 * (including frame) into the non-framed portion of an output image.  Both of
 * these produced artifacts, sometimes black stuff, sometimes white stuff.
 * xCount now is the distance from the left end of each input word being
 * read to a boundary in the input image.  Formerly, it had been the distance
 * from some pixel in that word to the boundary.  This lets us see where we
 * really are.  Also added an arg called rightWhiteVal, which is filled with
 * whatever the value of white is in the pixels that would otherwise be in
 * the right frame when reading the rightmost image word.  rightWhiteVal is
 * inserted into this rightmost image word as it is read.  The other method
 * of doing this would be to write the right frame with white, do the rotate,
 * then restore it.  This seems pretty crude.
 *
 * Revision 1.4  1995/02/17  22:21:36  ddavies
 * ddavies on Fri Feb 17 14:21:14 PST 1995
 *
 * i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v
 *
 * Clarify comments in code that computes location of first input pixel to
 * be rotated.
 *
 * Revision 1.3  1995/02/03  19:45:56  ddavies
 * i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v
 *
 * Change xRes and yRes from UInt16's to Int16's to keep acc happy.
 *
 * Revision 1.2  1995/02/03  01:23:52  ddavies
 * i_nirfx1.c -> /product/ipcore/ipshared/src/RCS/i_nirfx1.c,v
 *
 * Add xRes and yRes args to small angle rotations.  These are used to apply
 * a horizontal shear when rotating anisotropic images.  The shear is needed
 * to account for the shear naturally present in such images.
 * Also modified the calculation of the upper left corner of anisotropic
 * input images.  We now expand the dimension corresponding to the smallest
 * resolution so that the two resolutions are the same, do the corner-finding
 * calculation, then scale the modified dimension back down.  This should
 * make the image look correct if it's converted back to square pixels.
 *
 * Revision 1.1  1995/01/12  03:10:07  ddavies
 * Initial revision
 *
 * Revision 1.10  1994/11/10  01:16:35  ddavies
 * ddavies on Wed Nov 9 17:14:17 PST 1994
 *
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Change calculation of point in input coords that corresponds to (0,o) in
 * the output space.  Ron Danisewicv actually figured out what the rotation
 * code did, then figured out where the center of rotation should actually be,
 * then came up with a MUCH simpler formula we can use to calculate the
 * first input point.  Thanks Ron!
 *
 * Revision 1.9  1994/11/07  22:05:33  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.8  1994/08/05  21:18:06  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Reduce the frequency of IP_YIELD calls by 512x.
 *
 * Revision 1.7  1994/04/03  02:04:53  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Incorporate pointer types for use in SantaFe project.
 *
 * Revision 1.6  1994/03/02  03:51:09  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Add IP_YIELD to give up control during non-interpolated rotates.
 *
 * Revision 1.5  1994/02/04  20:15:35  dferrell
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:23:20  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:28:36  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

