/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	i_nirfx2.c: NonInterpolatedRotation, FiXed Point interface
 *
 *	    Upper level rotation routines that do NOT use interpolation to
 *		form output pixels, SMALL ANGLES ONLY! (< 10 degrees)
 *		These routines take the rotation angle in fixed point
 *		ScaledAngle format.
 *	    Lower level routines implemented in C or assembler.
 *
 *          Binary rotation routine:
 *		Int32     	i_rotateNoInterpolationIP()
 * 
 * 
 *
 */


#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"
#include "fxpmath.pub"

    /*  prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrscal.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrrast.pub"
#include "shros.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_nirfx2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* There is an enormous comment in i_nirfx1.c.  Please read that to get
 * an idea of how rotation is done.  We do include here the comments on
 * rotation in place since that's the routine in this file.
 *
 * *****************   ROTATION IN PLACE  *********************************
 *
 * This code give very nice results, but uses a lot of memory.  We
 * require full size images in both source and destination..  At 300 spi,
 * that's two 1 MB images.  At 400 spi, 11"x17", that's two 4MB images.
 * We haven't been thinking of these memory demands as extraordinary, but
 * PC users and XIS think they're big.
 *
 * We've already implemented gray rotation in place.  We'll use the same
 * algorithm.  We'll repeat the edited comments here.
 *
 * Z Smith, that font of creativity, suggested that we allocate more memory
 * than actually required for the image data in the image pixrect; then
 * rotate the image to the empty destination space.  We'd roll along, so
 * at the end would have the rotated image in some displaced position relative
 * to the source image.
 *
 * That idea was ok, but our image data structures didn't know about blank
 * spots.  So, we'll try a different approach.  i_rotateNoInterpolationIP
 * will allocate a ring buffer.  The optimized routines ip_rotateNoInterpLine,
 * ip_rotateNoInterpLineNiceBinary and ip_rotateNoInterpLineNiceGray will put
 * rotated data into the ring buffer.
 *
 * Rotation will be done in three stages:
 *  1.	Rotate enough lines to fill the ring buffer.  At this point, we
 *	should no longer need to use the first line of the original image
 *	as input data.
 *  2.	There are two steps for each rotated line:
 *	a. Copy the next line from the ring buffer to the image.
 *	b. Fill the empty space in the ring buffer with a new rotated line.
 *	This continues until we have produced enough rotated lines.
 *  3.	In the last stage, we flush the rest of the lines in the ring buffer
 *	to the image.  This should end when we have overwritten the last
 *	image line.
 *
 * So, how big is the ring buffer?  Obviously, it must be wide enough to
 * hold rotated lines.  For the sake of simplicity, we'll make it sBpl bytes
 * wide.  How tall should it be?  We can compute the maximum vertical
 * distance a diagonal path across the input image will cross.  This is
 * equal to wS * sin(rotAngle).  The ring buffer must be at least this tall
 * because when we have just stopped touching the top of the image, the
 * lower edge of the diagonal path will be this distance below the top.  We
 * assume that each source line produces about one dest line, so there must
 * be at least this many dest lines in the ring buffer.
 *
 * When we move down one line in the dest, we only move down cos(rotAngle)
 * lines in the source.  This is very close to 1, but not quite.  Therefore,
 * When we move down hS lines in the dest, we'll only move down
 * hS * cos(rotAngle) lines in the source.  To fill a 4400 line image at an
 * angle of 10 degrees, we only cross 4333 lines of source image.  This means
 * the distance between the source image's yCoord and the dest image's
 * yCoord keeps growing as we move down the image.  If the ring buffer were
 * only tall enough to cover the number of lines needed to span the
 * vertical displacement of one path across the source image, we would
 * start writing dest lines on top of source lines that had not yet been
 * retired by the time we reached the bottom of the pile.  To account
 * for this, we look at the difference between cos(rotAngle) and 1.  At
 * a maximum rotation angle of 10 degrees, this is 0.0152, or about
 * 1 part in 65.  To be safe, we'll add hS/32 lines to the ring buffer
 * height.
 *
 * Are there any other considerations?  Do we need to put in extra lines
 * because of starting or ending conditions?  If we started n lines above
 * the top of the source image, we'd need to add n/cos(rotAngle) extra
 * lines to the ring buffer.  The current code always rotates around the
 * middle of the image, so we never start above the top of the image.  I
 * can't think of any ending conditions that would make a difference.
 *
 * Therefore the ring buffer height should be (wS * sin(rotAngle)) +
 * hS / 32 lines tall.
 *
 *********************** ROTATION IN PLACE REWORK ***********************
 *
 * Really close.  Except for "nice" rotation.  This flavor of rotation
 * compares each destination line with the line above it.  Since rotation
 * in place uses a ring buffer, this fails when we wrap around since the
 * line above the new first line is at the end of the ring buffer.  Ok,
 * so the results would be a bit off.  The real problem is that the
 * algorithm tries to reference data that may or may not have been allocated
 * since it's outside the ring buffer.
 *
 * How to fix?  An obvious idea is to copy the last line up to the first
 * line each time we wrap.  This works so long as there is never good data
 * in the first line to destroy.  Since we already have some slop in the
 * hS / 32 factor, we'll just start initially and wrap around to the second
 * line of the ring buffer.
 *
 ************************* FIXED POINT IMPLEMENTATION ********************
 *
 * XIS supports platforms that have no floating point support.  They have
 * supplied fixed point implementations of sin, cos, atan and sqrt.  These
 * functions are mainly used to generate sinRotAngle and to find the
 * input point that corresponds to output point (0,0).  We substitute
 * the fixed point functions for their floating equvalents in a
 * straightforward way.
 *
 * ***********************************************************************
 * BUG FIX, 9/20/95:  Richard Mlynarik found that rotating gray images
 * sometimes produced artifacts on the right side.  After further
 * investigation, I found that there's a generic problem with the way
 * we handle the rightmost word of an image.  If the right boundary of
 * an input image does not end on a word boundary, we sometimes use the
 * entire word (which doesn't work when the frame is not white) and sometimes
 * ignore the rightmost word entirely.  No good.
 *
 * The problem has the do with confusion about xCount.  In the gray rotate
 * code, we read one pixel at a time and updated xCount each time.  In
 * this code we read 32 bits (some number of pixels) each time.  xCount
 * indicated the distance from some pixel inside the word to the edge
 * of the input image.  This uncertaintly threw us off.
 *
 * In the new scheme, we truncate xCount to the nearest multiple of 32
 * towards -infinity.  This makes it calculate the distance from the leftmost
 * bit in the word being read to the edge of the input image.  We then
 * check the value of xCount in the third section of code that is used when
 * we're approaching the right edge of the image.  If xCount is < 0 (greater
 * than 0x80000000), we know the entire word being read is off the right
 * side of the image and just use whiteval.  This is the common case
 * there.  If xCount is positive, we now have two cases where before we
 * had one.  If xCount + 32 is also postive, we read an entire new word
 * from the source image.  If xCount+32 is negative ( >= 0x80000000 ),
 * we merge the word read from the image with whiteval using a new mask.
 * The mask is a constant computed from the width of the image in bits.
 * We can use the constant since the width of the source image (the one
 * from which we are reading the data) is constant.  In fact, since we
 * know that the stuff off the edge of the source image is frame, therefore
 * filled with 0's, we can just supply a rightWhite mask which is filled
 * with whiteVal bits in the frame section of the input word.  We just
 * OR that with the partial input word read.  The rightWhite mask is
 * computed in the lower level routines since not all of them take a
 * whiteVal anyway.
 *
 ******************* SECOND ANISOTROPIC IMAGE REWORK *********************
 *
 * The alert reader (not me) will notice that the explanation for the
 * behaviour of anisotropic rotation depends on the fact that the naive
 * skew detect code returned an angle which depended on the ratio beteen
 * the x resolution and the y resolution, i.e. (yRes/xRes) * (skew of
 * the original sheet of paper).  Knowing this, the rotation routine multiplied
 * the angle (the sine of a small angle is almost equal to the angle)
 * by (yRes/xRes) before using it.  Well, this is ducky if the rotation
 * angle came from a naive skew detection routine.  What if it came from
 * somewhere else?  Or, what if we scale the image anistropically before
 * calling the rotation routine?  In either case, the "correction" supplied
 * by the rotation routines will not do the right thing.
 *
 * Our solution is to demand that angles supplied to rotation routines
 * must be independent of x and y resolution.  What does this mean?  Let's
 * introduce the "isotropic cousin" of an anisotropic image  The isotropic
 * cousin of an image is the smallest image which is isotropic
 * (x resolution == y resolution) and can be generated by scaling the
 * original image in either x or y where the scaling factor is >= 1.0.
 * For example, a standard fax image has an x resolution of 200 dpi and
 * a y resolution of 100 dpi.  The nearest isotropic image is a version of
 * the original image that has been scaled by 2x in the vertical direction
 * only, therefore has xRes = yRes = 200 dpi.  The isotropic cousin of
 * an isotropic image is the image itself.
 *
 * We require that the angles that are passed around correspond to the
 * rotation to be performed on the isotropic cousin of the actual
 * image provided.  We perform an operation on the real image such
 * that the result of our rotation-like operation followed by a scaling
 * operation that makes the result isotropic is equivalent to simply
 * rotating the isotropic cousin of the original image.
 *
 * Why this convoluted definition?  Well, when someone looks at an image,
 * it will almost surely be using isotropic resolutions.  We want to make
 * sure that the rotation-like operation we perform will, when scaled for
 * viewing, look like a real rotation.
 *
 * So, what's involved?  First, the skew detection routines must scale
 * the vertical displacements they find by (xRes/yRes) before computing
 * the arctan.  That's because the naive computation found a vert/horiz
 * ratio that was (yRes/xRes) * (vert/horiz for the isotropic cousin of
 * the image).
 *
 * Next, we look at how the rotation code uses the angle and the
 * x and y resolutions.  In this code, we assume the cos(rotAngle) = 1
 * anyway, so we don't change that.  However, we compute two versions
 * of sin(rotAngle).  vertSinRotAngle is used to modify the X coordinates
 * when moving vertically.  horizSinRotAngle is used to modify the Y coord
 * when moving horizontally.  Why do we need both?  Well, think of converting
 * an isotropic image to an anisotropic image by squishing all the individual
 * pixels, then laying the result out over an isotropic grid.  Say we make
 * the input pixels short and wide.  Then, each time we move down in the
 * input image, we have to remember to move over extra far in the output
 * image to get to the right place.  Similarly, each time we move over in
 * the input image, we have to remember that only a small movement is
 * implied vertically.
 *
 * That's basically the deal.  The length of the runs is computed by
 * ip_makeNoInterpRotTable.  A run ends when the center of the next
 * horizontal input pixel goes above the current output row.  Since we
 * want to know about vertical distances when moving horizontally, we
 * pass horizSinRotAngle to ip_makeNoInterpRotTable.  When moving down
 * vertically to find the beginning of the next scan line, we use
 * vertSinRotAngle as before.  When finding the upper left corner of
 * the image, we use both horizSinRotAngle and vertSinRotAngle as appropriate.
 *
 * ************************* FLAT IMAGE BUG FIX **************************
 *
 * In the previous code, I always assumed that the diagonal track would
 * either go from below the image into the image; or into the image and
 * stay in the image; or from into the image to above the image.  In very
 * flat images (wide, but not tall), a single diagonal track might go
 * from below the image, through the image and out the top.  This means
 * that we have to be careful about setting yCount - we check to see if
 * it's below the bottom before checking to see if it's near the top.
 * For flat images, it can be both near the top and below the bottom, but
 * we were only looking at the near the top case.
 * The lower level code will now also have to watch yCount as it's decremented.
 * If it goes from 0x80000000 to 0x7fffffff, it should be set to the
 * source image height - 1.  That means we have to send the image height
 * to the lower level routines, which is unfortunately a new parameter.
 */


/* PUBLIC */
/**************************************************************************
 *  i_rotateNoInterpolationIPFixedPoint()  (rotation in place)
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		rotAngle = rotation angle in ScaledAngle's
 *		depth = bits/pixel in source and dest images.
 *		nice = cTrue if we should do the slower version
 *			of rotation that makes an attempt to
 *			reduce the number of jaggies.
 *		xRes = resolution of image in x direction
 *		yRes = resolution of image in y direction
 *
 *	On return:
 *		if successful, SOURCE image contains the source image
 *		rotated by the given angle.  Unlike i_rotateGray, this routine
 *		does not form rotated pixels by averaging the values
 *		of the input pixels that overlap the output pixel.
 *		In this routine, each output pixel is assigned the
 *		value of the input pixel with the greatest overlap.
 *		If nice = cTrue, the rotation routine makes some attempt
 *		to avoid horizontal shears of vertical strokes by adjusting
 *		the horizontal shift amount by 1 if vertical strokes
 *		don't line up.
 *
 *		This routine allocates and frees a buffer used to hold
 *		a band of the rotated image.  Rotated bands are written
 *		back to the source image.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		unable to allocate memory for tables
 *		ia_depthNotSupported	could not rotate source image
 *					of specified depth
 *		ia_angleTooLarge	can not rotate by this angle
 *
 *	POSITIVE ROTATION ANGLE => CLOCKWISE ROTATION
 *	When rotating an image with non-square resolution, rotAngle should
 *	be specified so that the rotated image will be correct horizontally.
 *	The rotation op will shear the image horizontally so that if it
 *	were to be converted to a square resolution, the correct image
 *	will appear.
 **************************************************************************/
Int32   CDECL 
i_rotateNoInterpolationIPFixedPoint (
    IPState    *pIPState,
    UInt8Ptr	pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    ScaledAngle	rotAngle,
    Int32	depth,
    Bool	nice,
    Int16	xRes,
    Int16	yRes)
{
PROLOGUE;
Int32		line, ringLine;		/* interation variables */
Int32		xICoord, yICoord;	/* integers containing the
					 * X and Y coords of the current input
					 * pixel in the input image's frame
					 * of reference.  To keep things
					 * simple, xICoord is maintained in
					 * units of bits, not pixels.
					 */
Int32		leftmostBitPosition;	/* keeps track of leftmost source
					 * image bit within a word */
ScaledInt1616	xIOCoord;		/* split integer containing the
					 * X coord of the current
					 * input pixel in the output image's
					 * frame of reference.
					 */
ScaledInt1616	sinRotAngle;		/* split integer containing the
					 * sine of the rotation angle.
					 * These have 16 bit int and frac
					 * parts */
ScaledInt1616	vertSinRotAngle;	/* The sine of the rotation angle
					 * that's been modified to account
					 * for shears in anisotropic images.
					 * This is used to modify the X
					 * coord when moving vertically. */
ScaledInt1616	horizSinRotAngle;	/* The sine of the rotation angle
					 * that's been modified to account
					 * for shears in anisotropic images.
					 * This is used to modify the Y
					 * coord when moving horizontally. */
UInt32		xCount, yCount;		/* distances to vertical and horiz
					 * edges passed to rotateGrayLine.
					 * equivalent to input coords relative
					 * to input image. */
UInt8Ptr	pDOrig;			/* used to remember addr of upper left
					 * corner of image so we can clear its
					 * right frame when done. */
UInt32Ptr	pMaskTable;		/* table of masks used to rotate
					 * a single line */
Int32		pixelsPerWord, logPixelsPerByte; /* used to adjust addresses */
UInt32		widthInBits;		/* used to adjust xCount */

UInt32		whiteVal;		/* a 32 bit word holding white
					 * pixels.  These are 0 for binary
					 * images and 1's for gray images */
UInt32		rightWhiteVal;		/* white pixels inserted into rightmost
					 * partial image word. */
Int32		boundary1, boundary2;	/* args passed to line rotation
					 * routine telling it when to check
					 * for the edge of the input image. */
Int32		maxVertDisp;		/* holds the distance a line may
					 * travel vertically in the input
					 * image as it crosses the input
					 * image. */
Int32		extraBufLines;		/* the number of lines added to
					 * the ring buffer to allow for the
					 * fact that we'll proceed down the
					 * dest image faster than we'll
					 * proceed down the source image.
					 */
Int32		ringBufLines;		/* total number of lines in the
					 * ring buffer */
UInt8Ptr	pD, pRing, pLine;	/* pointer to dest loc in source/dest
					 * image and ptr to ring buffer and
					 * current line in ring buffer. */
Int32		absSBpl;		/* absolute value of sBpl.  Used
					 * to cycle through ring buffer */
UInt32		widthInBytes;		/* number of bytes to transfer from
					 * ring buffer to src image */

/* see definitions of cMaxRotAngle and cMinRotAngle above */
/*
#define cMaxRotAngle	10.0 * 3.14159 / 180.0
#define cMinRotAngle	-10.0 * 3.14159 / 180.0
*/
#define	cScaledPi	((Int32)16384)		/* 32768 corresponds to 2pi in
										 * scaled angle notation */
#define	cMaxRotAngle	((10 * cScaledPi) / 180)
#define cMinRotAngle	((-10 * cScaledPi) / 180)

    pMaskTable = NULL;
    pD = pRing = pLine = NULL;

	/* can we handle this depth image? */
    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8))
	return ia_depthNotSupported;

	/* can we handle this angle? */
    if ((rotAngle < cMinRotAngle) || (rotAngle > cMaxRotAngle))
	return ia_angleTooLarge;

	/* Are the resolutions ok? */
    if ((ABS(xRes - yRes) > 1) && ((xRes == 0) || (yRes == 0)))
	return ia_invalidParm;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + depth*cFrameBytes;

	/* Remember address of upper right corner of dest image
	 * so we can clear its right frame when done. */
    pDOrig = pS;

	/* get the width and height of the source without the frame */
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;

	/* set up depth-dependent parameters */
    switch (depth) {
    case 1: /* depth == 1 */
        pixelsPerWord = 32;
	logPixelsPerByte = 3;		/* log of 8 */
	widthInBits = wS;
	whiteVal = 0;
	break;
    case 2: /* depth == 2 */
        pixelsPerWord = 16;
	logPixelsPerByte = 2;		/* log of 4 */
	widthInBits = wS*2;
	whiteVal = 0xFFFFFFFF;
	break;
    case 4: /* depth == 4 */
        pixelsPerWord = 8;
	logPixelsPerByte = 1;		/* log of 2 */
	widthInBits = wS*4;
	whiteVal = 0xFFFFFFFF;
	break;
    default: /* depth == 8 */
        pixelsPerWord = 4;
	logPixelsPerByte = 0;		/* log of 1 */
	widthInBits = wS*8;
	whiteVal = 0xFFFFFFFF;
	break;
    }

	/* If the image does not end on a word boundary on the right,
	 * insert some white pixels.  Otherwise, insert nothing.
	 */
    if ((widthInBits & 0x1F) != 0)
	rightWhiteVal = whiteVal >> (widthInBits & 0x1F);
    else
	rightWhiteVal = 0;

	/* Calculate the number of bytes we need to move to the
	 * dest image from the ring buffer.  We want to round the
	 * number of bits up to the next multiple of a word.  This
	 * will make things consistent with the non-InPlace case.
	 */
    widthInBytes = ((widthInBits+31)/32) * 4;

	/* allocate the ring buffer to be used for destination lines.
	 * The buffer is sBpl bytes wide and maxVertDisp + extraBufLines
	 * lines tall.  maxVertDisp equals the number of source lines
	 * a diagonal path across the source will cross.  When we're
	 * maxVertDisp lines below the top of the source, we shouldn't
	 * be touching the top line anymore.
	 * extraBufLines was originally installed because I thought
	 * that we'd go down the input image faster than we'd go
	 * down the output image (because cos(rotAngle) < 1.0).  However,
	 * We always go down one input line for each output line, so that
	 * doesn't make any sense.  I suppose it still exists because
	 * it doesn't hurt anything.
	 *
	 * sinRotAngle should always be positive.
	 * 
	 * Note the way we multply horizSinRotAngle by wS.  When does
	 * this overflow?  We are concerned because the result of fract_sin
	 * is the real sine multiplied by 32K.  Well, 0xFFFFFFFF/(64K * sin(10)
	 * is about 377,000.  So this works so long as wS < 377,000.  I think
	 * we're safe.
	 */
    if ((__error__ = fract_sin(pIPState,
			       rotAngle, (ScaledInt1715Ptr) &sinRotAngle))
	!= ia_successful)
	NORMAL_RETURN;
    sinRotAngle =  sinRotAngle << 1;	/* convert to ScaledInt1616 */
    sinRotAngle = ABS(sinRotAngle);
    if (ABS(xRes - yRes) > 1)
    {
	vertSinRotAngle = (xRes * sinRotAngle)/yRes;
	horizSinRotAngle = (yRes * sinRotAngle)/xRes;
    }
    else
    {
	vertSinRotAngle = sinRotAngle;
	horizSinRotAngle = sinRotAngle;
    }
    maxVertDisp = ScaledInt1616_TO_Int32(wS * horizSinRotAngle);
    extraBufLines = hS / 32;	/* just for slop */
    ringBufLines = maxVertDisp + extraBufLines;

    pRing = (UInt8Ptr) MALLOC(sBpl * ringBufLines);
    if (pRing == NULL)
	return ia_nomem;
    absSBpl = sBpl;

	/* We'd like to use one set of code for both clockwise and counter
	 * clockwise rotation.  If rotAngle < 0, we're doing counter-
	 * clockwise rotation so should process the image starting from the
	 * lower left corner instead of the upper left corner.  Reposition
	 * pS.  Invert sBpl.
	 */
    if (rotAngle < 0)
    {
	pS += (hS-1) * sBpl;		/* point to lower right corner */
	sBpl = -sBpl;
	rotAngle = -rotAngle;
    }

	/* set the dest pointer once we've really figured out where
	 * the source pointer should start */
    pD = pS;				/* initial dest is start of source. */

	/* if sinRotAngle <= 2 (sin(rotAngle) <= (0.0000000000000010),
	 * this angle is too small to bother with.  Just leave the
	 * source unchanged.  There are several places below where
	 * we assume sinRotAngle is non-zero. */
    if (MIN(vertSinRotAngle, horizSinRotAngle) <= 2)
    {
	FREE(pRing);
	NORMAL_RETURN;
    }

	/* Make the sequence of masks using modified vals.  Set the
	 * forAssembly parameter according to whether there is assembly
	 * code present.  This is done since even in nice rotation, the
	 * very first line is done by the assembly routine. */
    if ((__error__ = ip_makeNoInterpRotTable (
			pIPState, 
			horizSinRotAngle, &pMaskTable, wS,
			sBpl / 4, depth, 
			(Int8)ip_rotateAssemblyCodeIsPresent (pIPState)))
	!= ia_successful)
	    NORMAL_RETURN;
    IP_YIELD(errorFinish)

	/* compute the initial input position using the algorithm
	 * described in the comments above.  Note that xICoord is
	 * calculated in units of BITS, not pixels.  This allows us
	 * to ignore the depth later on.  The coordinates of the
	 * initial input point are ((-hS/2) sinRotAngle, (wS/2) sinRotAngle)
	 * Since sinRotAngle is currently kept as an Int1616, we have to
	 * shift the result right by 16 after doing the multiply.
	 * Adding 0x8000 is to get the rounding correct.
	 * Remember that if the image is anisotropic (non-square resolution),
	 * we have to use horizSinRotAngle to modify yICoord when moving
	 * horizontally and vertSinRotAngle to modify xICoord when
	 * moving vertically.
	 */

    xICoord = -1*(((hS/2) * vertSinRotAngle + 0x8000) >> 16) * depth;  
    yICoord = ((wS/2) * horizSinRotAngle + 0x8000) >> 16;

	/* This is the input coordinate which we decree corresponds exactly
	 * to coordinate (0,0) in the output frame of reference.  What do we
	 * do with it?  First, we adjust the input address to point to this
	 * place.  This will almost certainly point to something outside
	 * the input image, but that's ok.  We won't do any memory references
	 * with this address until we enter the input image.
	 * Note that if xICoord is a positive number, the (xICoord >>
	 * logPixelsPerByte) rounds down.  This is good since we want
	 * to move only left one word for every complete 32 bits.  However
	 * if xICoord is negative, the (xICoord >> logPixelsPerByte) term
	 * rounds up.  This is also good.  We want to move left one
	 * word for every fragment of a word to the left of us.
	 *
	 * Well, that was all very clever, but shifting integers right in
	 * ANSI C is undefined (although the Sun ANSI Compiler seems pretty
	 * haphazard about issuing warnings).  The processor might
	 * duplicate the sign bit, or it might not.  So, we had to use
	 * "/ 32". If xICoord is positive, we want
	 * (xICoord/32.  If xICoord is negative, we really 
	 * want ((xICoord/32)-1);
	 * NOTE: both xICoord and 32 must be SIGNED or the
	 * dumb Sun ANSI compiler will do an UNSIGNED divide!!!
	 * At this point, rotAngle is always positive, so cos(aspectAngle
	 * + rotAngle) is always less than cos(aspectAngle), so xICoord
	 * is always negative (or 0).
	 */
    pS += yICoord * sBpl +
		( ((xICoord - 31) / 32) * 4);

	/* Set the position of the leftmost input bit in the first word.
	 * Here, leftmostBitPosition is measured relative to the left end
	 * of the word.  So, for example, if xICoord = -1,
	 * leftmostBitPosition is set to 32+(-1) = 31.
	 * If xICoord is more negative that -32, the "OR (-32)" term will
	 * blast the more significant bits.  If we have 2 bit pixels,
	 * xICoord (already multiplied by depth) will equal 2, so we'll
	 * still get the right answer.
	 */
    leftmostBitPosition = 32 + (xICoord  | (-32));

	/* By decree, the location of the this input pixel relative to the
	 * output image is (0.0, 0.0).  Since we assume cos(rotAngle) = 1,
	 * we really only care about xIOCoord.  It will be used to decide
	 * when we've crossed a bit (or word) boundary.
	 */
    xIOCoord = 0;

    xCount = yCount = 0;	/* initialization to make gcc happy.  This
				 * is not really necessary since yCount is
				 * initialized before it is actually used.
				 */


	/* ************* FILL RING BUFFER ******************** */


    boundary1 = 0;			/* check xCount and yCount across
				         * entire image while filling
					 * the ring buffer since we're
					 * near the top of the image. */
    boundary2 = 0;			/* doesn't matter */

	/* Have ip_rotateNoInterpLine actually do the work on each line.
	 * Start by just filling the ring Buffer.
	 * BUG FIX for nice rotation: put new data into the second line
	 * of the ring buffer so that we can use the first line for
	 * data copied from the end when we wrap around.  This allows
	 * the nice algorithm to always look up one line when computing
	 * where to do the shifts. */
    pLine = pRing + absSBpl;
    for (line = 0;
	 (line < hS) && (line < (ringBufLines-1));
	 line++, pLine += absSBpl)
    {
	    /* Calculate the distance from left end of first word read
	     * to the first vertical edges that we'll cross.  These are
	     * set up so that the count will be positive inside the input
	     * image and negative outside.  Note that if xICoord is negative,
	     * we want xCount to be the negative of the distance to the
	     * left edge of the input image (so it goes pos when we get
	     * there).
	     */
	if (xICoord < 0)
	    xCount = xICoord - leftmostBitPosition;
	else
	    xCount = 0x80000000 -
		     (widthInBits - xICoord + leftmostBitPosition);

	    /* Since we're near the top of the image, check
	     * xCount and yCount all the way across.  In the loop below,
	     * this checking is only done if we're near the bottom.
	     * Ooops, if the image is flat, we might be near the top
	     * and bottom at the same time.  Better check.
	     */
	if (yICoord >= hS)
	    yCount = 0x80000000 - (hS - yICoord); /* starting near bottom of
						   * image */
	else	/* the image is not flat */
	    yCount = yICoord;		/* starting near top of image, so
					 * we're likely to cross the top
					 * boundary.*/

	if (!nice || (line == 0))	/* do it the fast way */
	{
	    ip_rotateNoInterpLine (pIPState, pS, sBpl,
				   pLine, widthInBits, xCount,
			           yCount, leftmostBitPosition, pMaskTable,
			           whiteVal, rightWhiteVal,
				   boundary1, boundary2, hS);
	    if (nice && ip_rotateAssemblyCodeIsPresent (pIPState))
		    /* make a new mask table for the rest of the lines.
		     * The first line was done by the assembly routine,
		     * but the rest of them will be done by C routines
		     * so we need the C version of the table. */
	    {
		FREE(pMaskTable);
		ip_makeNoInterpRotTable (pIPState,
					 horizSinRotAngle, &pMaskTable,
					 wS, sBpl / 4, depth, cFalse);
		if (__error__ != ia_successful)
		    NORMAL_RETURN;
	    }
	}
	else				/* try to avoid sheared text */
	if (depth == 1)
		/* ip_rotateNoInterpLineNiceBinary negates what it sees as
		 * sBpl and dBpl when accessing the word that it is
		 * trying to "line up" with to eliminate jaggies.  This
		 * means that if the rotation angle is negative, and
		 * hence we are going from the bottom to the top of the
		 * image, that sBpl and dBpl must have different signs.
		 * As a result we use sBpl and absSBpl for the sBpl and
		 * dBpl respectively.
		 */
	    ip_rotateNoInterpLineNiceBinary (pIPState, pS, sBpl, pLine, absSBpl,
					     widthInBits, xCount, yCount,
					     leftmostBitPosition,
					     pMaskTable, boundary1, boundary2,
					     hS);
	else
	    ip_rotateNoInterpLineNiceGray (pIPState, pS, sBpl,
					   pLine, absSBpl, widthInBits,
				           xCount, yCount, leftmostBitPosition,
				           pMaskTable, depth, rightWhiteVal,
					   boundary1, boundary2, hS);

	    /* now find the starting position of the next line */
	pS += sBpl;			/* and in input space */
	yICoord++;			/* moving down one line in output
					 * space always corresponds to
					 * moving down 1 line in input space.
					 */
	xIOCoord += vertSinRotAngle;	/* and move over in input space. */
	if ((xIOCoord >> 16) != 0)
	{				/* we need to move over in x */
	    xIOCoord &= 0xFFFF;		/* get rid of integer part again */
	    xICoord += depth;
	    leftmostBitPosition += depth;
	    if (leftmostBitPosition > 31)
	    {
		pS += 4;			/* move to next word boundary */
		leftmostBitPosition &=  0x1F;
	    }
	}

	if ((line & 0x1FF) == 0)
	    IP_YIELD(errorFinish)

    } /* end of loop that initially fills the ring buffer */


	/* ******* TRANSFER FROM RING BUFFER TO SOURCE/DEST & ROTATE ***** */


	/* Have ip_rotateNoInterpLine actually do the work on each line.
	 * Before rotating each line into the ring buffer, move the
	 * previously rotated contents from the ring buffer to the
	 * source/dest image.  At the entry to the loop, pRing points
	 * one line below the ring buffer as it would if we wrapped
	 * were about to wrap around (which is the very first
	 * thing we'll do!).
	 * BUG FIX for nice rotation.  When we wrap around, copy the
	 * last line in the ring buffer up to the first line so that
	 * the nice rotation routines will always be able to loop "up"
	 * one line when comparing potential destination lines.
	 */
    for (ringLine = ringBufLines;
	 line < hS;
	 line++, ringLine++, pLine += absSBpl)
    {

	    /* If we have wrapped around, copy the last line in the
	     * ring buffer to the first line and adjust ringLine and pLine
	     * to point to the second line.  This allows the nice rotation
	     * algorithms to always look up one line.  Note that we
	     * round up to the next word boundary. */
	if (ringLine == ringBufLines)
	{
	    MEMCPY(pRing, pLine - absSBpl, widthInBytes);
	    pLine = pRing + absSBpl;		/* point to 2nd line of buf */
	    ringLine = 1;
	}

	    /* Transfer the previously rotated line in this position in
	     * the ring buffer to the dest image location pointed to by
	     * pD.
	     */
	MEMCPY(pD, pLine, widthInBytes);
	pD += sBpl;			/* pt to next line in source/dest
					 * image */

	    /* Calculate the distance to the first horizontal and vertical
	     * edges that we'll cross.  These are set up so that the count
	     * will be positive inside the input image and negative outside.
	     * The boundaryX values are expressed in units of bits from
	     * the RIGHT image boundary.  That's because the width parm
	     * is decremented by ip_rotateNoInterpLine as it moves from
	     * left to right across the image.
	     */
	if (xICoord < 0)
	{
	    xCount = xICoord - leftmostBitPosition;
		/* Stop the first section just a bit after starting.  If
		 * the image is very narrow, widthInBits + xICoord might
		 * be (very) negative so we might do the first section
		 * too many times.  we clip boundary1 at 0 so we
		 * always stop soon enough.
		 */
	    boundary1 = MAX((Int32)widthInBits + xICoord, 0);
	    boundary2 = MAX(xICoord + 32, 0);	/* near right edge of image.
						 * make sure we get into
						 * the checking code BEFORE
						 * arriving at the right edge
						 */
	}
	else
	{
	    xCount = 0x80000000 -
		     (widthInBits - xICoord + leftmostBitPosition);
						/* dist from left edge of
						 * word to right side of input
						 * image. */
	    boundary1 = widthInBits;		/* at the left edge (i.e.
						 * skip first code section) */
	    boundary2 = xICoord + 32;		/* near right edge of image.
						 * make sure we get into
						 * the checking code BEFORE
						 * arriving at the right edge.
						 */
	}

	    /* If we're near the bottom of the image, check
	     * xCount and yCount all the way across.
	     * If we're in the middle, initialize yCount to
	     * something sensible so it doesn't cause trouble in
	     * the first code section.
	     */
	yCount = yICoord;		/* default, equals number of lines
					 * until top row. */
	if (yICoord >= hS)		/* starting near bottom of image */
	{
	    yCount = 0x80000000 - (hS - yICoord); /* starting near bottom of
						   * image */
	    boundary1 = 0;		/* check xCount and yCount across
					 * entire image */
	    boundary2 = 0;		/* doesn't matter */
	}

	if (!nice)			/* do it the fast way */
	    ip_rotateNoInterpLine (pIPState, pS, sBpl,
				   pLine, widthInBits, xCount,
			           yCount, leftmostBitPosition, pMaskTable,
			           whiteVal, rightWhiteVal,
				   boundary1, boundary2, hS);
	else				/* try to avoid sheared text */
	if (depth == 1)
	    ip_rotateNoInterpLineNiceBinary (pIPState, pS, sBpl, pLine, absSBpl,
					     widthInBits, xCount, yCount,
					     leftmostBitPosition,
					     pMaskTable, boundary1, boundary2,
					     hS);
	else
	    ip_rotateNoInterpLineNiceGray (pIPState, pS, sBpl,
					   pLine, absSBpl, widthInBits,
				           xCount, yCount, leftmostBitPosition,
				           pMaskTable, depth, rightWhiteVal,
					   boundary1, boundary2, hS);

	    /* now find the starting position of the next line */
	pS += sBpl;			/* and in input space */
	yICoord++;			/* moving down one line in output
					 * space always corresponds to
					 * moving down 1 line in input space.
					 */
	xIOCoord += vertSinRotAngle;	/* and move over in input space. */
	if ((xIOCoord >> 16) != 0)
	{				/* we need to move over in x */
	    xIOCoord &= 0xFFFF;		/* get rid of integer part again */
	    xICoord += depth;
	    leftmostBitPosition += depth;
	    if (leftmostBitPosition > 31)
	    {
		pS += 4;		/* move to next word boundary */
		leftmostBitPosition &=  0x1F;
	    }
	}

	if ((line & 0xFF) == 0)
	    IP_YIELD(errorFinish)

    } /* end of loop that reads all of the source image */


	/* At this point, the ring buffer is full of rotated lines.
	 * Flush it to the image. */
    for (line = MIN(ringBufLines-1, hS);
	 line > 0;
	 line--, ringLine++, pLine += absSBpl, pD += sBpl)
    {
	    /* If we have wrapped around, adjust ringLine and pLine. */
	if (ringLine == ringBufLines)
	{
	    pLine = pRing + absSBpl;	/* point to 2nd line of buf */
	    ringLine = 0;
	}

	    /* Transfer the previously rotated line in this position in
	     * the ring buffer to the dest image location pointed to by
	     * pD.
	     */
	MEMCPY(pD, pLine, widthInBytes);

    }

	/* clear the right frame of the rotated image.  It was probably
	 * written mistakenly by the rotation code. */
    if (sBpl > 0)
	ip_clearRightFrame (pIPState, pDOrig, wS, hS, sBpl, depth);
    else
	ip_clearRightFrame (pIPState, pDOrig, wS, hS, -sBpl, depth);

    FREE(pRing);
    FREE(pMaskTable);
    NORMAL_RETURN;

errorFinish:
    FREE(pRing);
    FREE(pMaskTable);
    return ia_aborted;

}



/*

$Log: i_nirfx2.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.11  1998/01/13 23:49:03  ddavies
i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.10  1997/08/20 20:28:43  ddavies
i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v

Change Int8 nice to Bool nice in arg list.

Revision 1.9  1996/10/05 02:19:26  ddavies
i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v

Change the order in which we check to see if we're near the top or bottom
of an image, since in flat images, we could be near both at the same time.
We now check to see if we're near the bottom first.
Also pass the image height down to the lower level routines so they can
update yCount correctly.

Revision 1.8  1996/02/10 01:33:11  dferrell
i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v

Added parens to & operations so that gcc 2.7.2 won't whine.

Revision 1.7  1995/12/13 02:44:57  ejaquith
i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/12/13  01:41:58  ddavies
 * i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v
 *
 * Rework the way anisotropic images are rotated.  In the old code, we assumed that
 * the angle passed in was calculated in a naive way and was really equal to
 * (real angle) * (yResolution / xResolution).  That's because if the y resolution
 * was small, the image would be squished vertically so all vertical displacements
 * would look small to the skew detect code.  The old code multipled the incoming
 * angle by (xResolution/yResolution) before use (sort of).  Ok, what it really
 * did was use two versions of sin(rotAngle): vertSinRotAngle = (xRes*xRes)/(yRes*yRes)
 * times sinRotAngle and just plain sinRotAngle.
 * In the new code, we assume that the angle is resolution independent so we
 * can get angles from anywhere.  The cosine of the rotation angle is not really
 * affected by this since the angles are so small (in fact, we always assume that
 * it's equal to 1 anyway).  However, we
 * still use two versions of sin(rotAngle): vertSinRotAngle = (xRes/yRes) *
 * sinRotAngle and horizSinRotAngle = (yRes/xRes) * sinRotAngle.  We use
 * vertSinRotAngle in EVERY instance where vertical motion requires adjusting
 * the x coordinate.  We use horizSinRotAngle in EVERY instanced where horizontal
 * motion requires adjusting the y coordinate.  These are used when calculating
 * the initial input position, when building the rotation tables and when
 * adjusting the input position in preparation for rotating the next
 * scan line.  Actually, by making things more consistent, it's easier to
 * figure how what's going on.
 * These changes apply to the non-interpolated rotation in place.
 *
 * Revision 1.5  1995/10/18  03:57:50  ddavies
 * i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.4  1995/09/21  02:54:59  ddavies
 * i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v
 *
 * Fixed a bug in i_rotateNoInterpolationIPFixedPoint.  This is the same bug
 * that was in ip_rotateNoInterpolationFixedPoint..  This routine was either
 * ignoring the rightmost word of an image or inserting the entire word
 * (including frame) into the non-framed portion of an output image.  Both of
 * these produced artifacts, sometimes black stuff, sometimes white stuff.
 * xCount now is the distance from the left end of each input word being
 * read to a boundary in the input image.  Formerly, it had been the distance
 * from some pixel in that word to the boundary.  This lets us see where we
 * really are.  Also added an arg called rightWhiteVal, which is filled with
 * whatever the value of white is in the pixels that would otherwise be in
 * the right frame when reading the rightmost image word.  rightWhiteVal is
 * inserted into this rightmost image word as it is read.  The other method
 * of doing this would be to write the right frame with white, do the rotate,
 * then restore it.  This seems pretty crude.
 *
 * Revision 1.3  1995/02/03  19:46:02  ddavies
 * i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v
 *
 * Change xRes and yRes from UInt16's to Int16's to keep acc happy.
 *
 * Revision 1.2  1995/02/03  01:24:30  ddavies
 * i_nirfx2.c -> /product/ipcore/ipshared/src/RCS/i_nirfx2.c,v
 *
 * Add xRes and yRes args to small angle rotations.  These are used to apply
 * a horizontal shear when rotating anisotropic images.  The shear is needed
 * to account for the shear naturally present in such images.
 * Also modified the calculation of the upper left corner of anisotropic
 * input images.  We now expand the dimension corresponding to the smallest
 * resolution so that the two resolutions are the same, do the corner-finding
 * calculation, then scale the modified dimension back down.  This should
 * make the image look correct if it's converted back to square pixels.
 *
 * Revision 1.1  1995/01/12  03:10:23  ddavies
 * Initial revision
 *
 * Revision 1.10  1994/11/10  01:16:35  ddavies
 * ddavies on Wed Nov 9 17:14:17 PST 1994
 *
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Change calculation of point in input coords that corresponds to (0,o) in
 * the output space.  Ron Danisewicv actually figured out what the rotation
 * code did, then figured out where the center of rotation should actually be,
 * then came up with a MUCH simpler formula we can use to calculate the
 * first input point.  Thanks Ron!
 *
 * Revision 1.9  1994/11/07  22:05:33  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.8  1994/08/05  21:18:06  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Reduce the frequency of IP_YIELD calls by 512x.
 *
 * Revision 1.7  1994/04/03  02:04:53  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Incorporate pointer types for use in SantaFe project.
 *
 * Revision 1.6  1994/03/02  03:51:09  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Add IP_YIELD to give up control during non-interpolated rotates.
 *
 * Revision 1.5  1994/02/04  20:15:35  dferrell
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:23:20  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:28:36  ddavies
 * i_nirfxp.c -> /product/ipcore/ipshared/src/RCS/i_nirfxp.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

