/*******************************************************************
 *  Copyright (c) 1993-95, Xerox Corporation.  All rights reserved.*
 *  Copyright protection claimed includes all forms and matters    *
 *  of copyrightable material and information now allowed by       *
 *  statutory or judicial law or hereafter granted, including      *
 *  without limitation, material generated from the software       *
 *  programs which are displayed on the screen such as icons,      *
 *  screen display looks, etc.                                     *
 *******************************************************************/

/*	
 *  i_nxtbt1.c: routines to find the next 1 in the scan line.
 *
 *		Int32	i_nextBitToTheLeft()
 *		Int32	i_nextBitToTheRight()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "imageref.pub"
#include "defines.pub"
#define	_USE_FIRST_BIT_TABLE_		/* want both tables from iopbyte.prv */
#define	_USE_LAST_BIT_TABLE_
#include "iopbyte.prv"
#include "lmask.prv"			/* contains leftMask[] */
#include "rmask.prv"			/* contains rightMask[] */


/* prototypes */
#include "shrcc.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_nxtbt1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* i_findNextBitToTheLeft and i_findNextBitToTheRight find the first 1
 * to the left or right respectively.  They do not continue to find the
 * end of the run.
 *
 * All routines take the word address of the leftmost word in the scan
 * line INSIDE the frame, the offset into the scan line and the width
 * of the scan line in pixels.  All routines work only on binary images.
 *
 * They return ia_successful if they find a run before running out of
 * scan line or ia_noOnPixel if no 1's are found before the end of the line.
 */

/* PUBLIC */
/**************************************************************************
 *   i_nextBitToTheLeft()
 *
 *	This procedure starts from the given offset in a scan line and
 *	looks to the left for 1's.  If it finds some, it returns
 *	the offset of the first 1 found.  The offset is relative
 *	to the left end of the scan line (like the initial offset).
 *
 *	THIS PROCEDURE ONLY WORKS ON BINARY IMAGES.  Since we expect
 *	to be calling it zillions of times, it doesn't check.
 *
 *	If there is a 1 at the initialOffset, this routine will return
 *	the initialOffset as the first 1.
 *
 *	If no 1 is found in bounds, *pRunRight is not changed.  This allows
 *	callers to use the same variable for initialOffset and
 *	*pRunRight.
 *
 *	Input:
 *		pS:		Pointer to leftmost 32-bit word in the
 *				scan line INSIDE the frame.
 *		minX:		Don't look to the left of this offset
 *		initialOffset:	starting offset in line relative to 
 *				left edge of image INSIDE frame.
 *		pRunRight:	pointer to Int32 that will contain the
 *				x offset of the rightmost 1 in the next
 *				run to the left.
 *
 *	Returns:
 *		ia_successful		found a 1
 *		ia_noOnPixel		didn't find a 1.
 *
 **************************************************************************/
Int32  CDECL 
i_nextBitToTheLeft (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	minX,
    Int32	initialOffset,
    Int32     *pRunRight)
{
    PROLOGUE;
    Int32	 byteOffset;		/* byte offset into the source line */
    Int32	 bitOffset;		/* the bit offset of the 1 found */
    UInt32	 inputWord;		/* the word we read from the image */
    Int32	 minXByteOffset;	/* the offset at which we should stop */

	/* calculate the byte offset of the word holding the initial offset */
    byteOffset = (initialOffset/8) & ~0x3;
    minXByteOffset = (minX/8) & ~0x3;

    inputWord = ULONG_ACCESS(&pS[byteOffset]);	/* get the first word */

	/* mask off the bits on the right that we should ignore */
    inputWord &= leftMask[initialOffset & 0x1F];

	/* if there are no 1's in this word, go find a word with some 1's */
    if (inputWord == 0)
    {
	byteOffset -= 4;			/* pt to next word */
	while ( (byteOffset > minXByteOffset) &&
		(ULONG_ACCESS(&pS[byteOffset]) == 0) )
	    byteOffset -= 4;

	    /* At this point, we're either pointing at a non-zero word or
	     * at the first word of the scan line (or both).  Put this
	     * word into inputWord.  If it's 0, return ia_noOnPixel.
	     */
	inputWord = ULONG_ACCESS(&pS[byteOffset]);
	if (inputWord == 0)
	{
	    return ia_noOnPixel;
	}
    }

	/* find the rightmost 1.  Do this by locating the 1 down to
	 * the byte, then using tables to find it in a byte */
    bitOffset = byteOffset * 8;
    if ((inputWord & 0xFFFF) != 0)
    {
	if ((inputWord & 0xFF) != 0)
	{
	    bitOffset += 24 + lastBitTable[inputWord & 0xFF];
	}
	else
	{
	    bitOffset += 16 + lastBitTable[(inputWord >> 8) & 0xFF];
	}
    }
    else	/* there's something in the upper half of the word */
    {
	inputWord = inputWord >> 16;	/* get the upper half */
	if ((inputWord & 0xFF) != 0)
	{
	    bitOffset += 8 + lastBitTable[inputWord & 0xFF];
	}
	else
	{
	    bitOffset += lastBitTable[(inputWord >> 8)];
	}
    }

	/* Don't return a value if it's too small */
    if (bitOffset >= minX)
    {
	*pRunRight = bitOffset;
	NORMAL_RETURN;
    }
    else
	return ia_noOnPixel;
}



/* PUBLIC */
/**************************************************************************
 *   i_nextBitToTheRight()
 *
 *	This procedure starts from the given offset in a scan line and
 *	looks to the right for 1's.  If it finds some, it returns
 *	the offset of the first 1 found.  The offset is relative
 *	to the left end of the scan line (like the initial offset).
 *
 *	THIS PROCEDURE ONLY WORKS ON BINARY IMAGES.  Since we expect
 *	to be calling it zillions of times, it doesn't check.
 *
 *	If there is a 1 at the initialOffset, this routine will return
 *	the initialOffset as the first 1.
 *
 *	If no 1 is found in bounds, *pRunLeft is not changed.  This allows
 *	callers to use the same variable for initialOffset and
 *	*pRunLeft.
 *
 *	Input:
 *		pS:		Pointer to leftmost 32-bit word in the
 *				scan line INSIDE the frame.
 *		w:		width of image in bits WITHOUT frame
 *		initialOffset:	starting offset in line relative to 
 *				left edge of image INSIDE frame.
 *		pRunLeft:	pointer to Int32 that will contain the
 *				x offset of the leftmost 1 in the next
 *				run to the right.
 *
 *	Returns:
 *		ia_successful		found a 1
 *		ia_noOnPixel		didn't find a 1.
 *
 **************************************************************************/
Int32  CDECL 
i_nextBitToTheRight (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	initialOffset,
    Int32     *pRunLeft)
{
    PROLOGUE;
    Int32	 byteOffset;		/* byte offset into the source line */
    Int32	 bitOffset;		/* the bit offset of the 0 found */
    UInt32	 inputWord;		/* the word we read from the image */
    Int32	 maxXByteOffset;	/* the offset at which we should stop */

	/* calculate the byte offset of the word holding the initial offset */
    byteOffset = (initialOffset/8) & ~0x3;
	/* turn w into the offset of the word holding the last byte */
    maxXByteOffset = (w/8) & ~0x3;

    inputWord = ULONG_ACCESS(&pS[byteOffset]);	/* get the first word */

	/* mask off the bits on the left that we should ignore */
    inputWord &= rightMask[(~initialOffset) & 0x1F];

	/* if there are no 1's in this word, go find a word with some 1's */
    if (inputWord == 0)
    {
        byteOffset += 4;
	while ( (byteOffset < maxXByteOffset) &&
		(ULONG_ACCESS(&pS[byteOffset]) == 0) )
	    byteOffset += 4;

	    /* At this point, we're either pointing at a non-zero word or
	     * at the last word of the scan line (or both).  Put this
	     * word into inputWord.  If it's 0, return ia_noOnPixel.
	     */
	inputWord = ULONG_ACCESS(&pS[byteOffset]);
	if (inputWord == 0)
	{
	    return ia_noOnPixel;
	}
    }

	/* find the leftmost 1.  Do this by locating the 1 down to
	 * the byte, then using tables to find it in a byte */
    bitOffset = byteOffset * 8;
    if ((inputWord & 0xFFFF0000) != 0)
    {
	inputWord = inputWord >> 16;	/* get the upper half */
	if ((inputWord & 0xFF00) != 0)
	{
	    bitOffset += firstBitTable[inputWord >> 8];
	}
	else
	{
	    bitOffset += 8 + firstBitTable[inputWord & 0xFF];
	}
    }
    else	/* there's something in the lower half of the word */
    {
	if ((inputWord & 0xFF00) != 0)
	{
	    bitOffset += 16 + firstBitTable[inputWord >> 8];
	}
	else
	{
	    bitOffset += 24 + firstBitTable[inputWord & 0xFF];
	}
    }

	/* Don't return a value if it's too big */
    if (bitOffset <= w)
    {
	*pRunLeft = bitOffset;
	NORMAL_RETURN;
    }
    else
	return ia_noOnPixel;
}


/*

 $Log: i_nxtbt1.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.8  1998/01/13 23:49:06  ddavies
 i_nxtbt1.c -> /product/ipcore/ipshared/src/RCS/i_nxtbt1.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.7  1995/12/13 02:45:02  ejaquith
 i_nxtbt1.c -> /product/ipcore/ipshared/src/RCS/i_nxtbt1.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/07/27  02:15:27  ddavies
 * i_nxtbt1.c -> /product/ipcore/ipshared/src/RCS/i_nxtbt1.c,v
 *
 * Make maxXByteOffset and minXByteOffset into Int32's instead of UInt32s
 * so that the routines don't wander off the image when using very thin
 * images.  This could happen in i_next1ToTheLeft if the initial byteoffset
 * was 0.
 *
 * Revision 1.5  1995/01/11  22:26:26  dferrell
 * i_nxtbt1.c -> /product/ipcore/ipshared/src/RCS/i_nxtbt1.c,v
 *
 * Changed include of shrcmprs.pub to shrcc.pub.
 *
 * Revision 1.4  1994/01/04  03:08:22  ddavies
 * i_nxtbt1.c -> /product/ipcore/ipshared/src/RCS/i_nxtbt1.c,v
 *
 * Change i_nextBitToTheLeft and i_nextBitToTheRight so they modify the returned
 * positions (*pRunRight and *pRunLeft respectively) only if a 1 is found
 * inside the bounds given.  i_nextBitToTheRight was also modified so that it
 * does not return values that are within a 32 bit word that straddles the
 * rightmost boundary and are to the right of the boundary.
 *
 * Revision 1.3  1993/12/17  22:15:47  ddavies
 * i_nxtbt1.c -> /product/ipcore/ipshared/src/RCS/i_nxtbt1.c,v
 *
 * Changed comment in i_nextBitToTheLeft.  pRunRight contains the number of the
 * rightmost bit in the next run to the left, not right.
 *
 * Revision 1.2  1993/12/02  19:23:29  ddavies
 * i_nxtbt1.c -> /product/ipcore/ipshared/src/RCS/i_nxtbt1.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.1  1993/12/01  04:09:26  ddavies
 * Initial revision
 *

 */

