/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_nxtrun.c: routines to find the next run in a scan line.
 *
 *		Int32	i_nextRunToTheRight()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "imageref.pub"
#include "defines.pub"
#define	_USE_FIRST_BIT_TABLE_		/* want both tables from iopbyte.prv */
#include "iopbyte.prv"
#include "lmask.prv"			/* contains leftMask[] */
#include "rmask.prv"			/* contains rightMask[] */


/* prototypes */
#include "shrcc.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_nxtrun.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* PUBLIC */
/**************************************************************************
 *   i_nextRunToTheRight()
 *
 *	This procedure starts from the given offset in a scan line and
 *	looks to the right for 1's.  If it finds some, it returns
 *	the offset of the first and last 1 found.  The offset is relative
 *	to the left end of the scan line (like the initial offset).
 *
 *	THIS PROCEDURE ONLY WORKS ON BINARY IMAGES.  Since we expect
 *	to be calling it zillions of times, it doesn't check.
 *
 *	If there is a 1 at the initialOffset, this routine will return
 *	the initialOffset as the first 1.
 *
 *	Input:
 *		pS:		Pointer to leftmost 32-bit word in the
 *				scan line INSIDE the frame.
 *		rightLimit:	rightmost pixel at which we should look
 *				for bits.  By definition, no run can
 *				begin or end to the right of this offset.
 *				Runs ending beyond this offset will be
 *				chopped off here.
 *		initialOffset:	starting offset in line relative to 
 *				left edge of image INSIDE frame.
 *		blackrun:	true if a black run, def: white run
 *		pRunLeft:	pointer to UInt32 that will contain the
 *				x offset of the leftmost 1 in the next
 *				run to the right.
 *		pRunRight:	pointer to UInt32 that will contain the
 *				x offset of the rightmost 1 in the next
 *				run to the right.
 *
 *	Returns:
 *		ia_successful		found a run
 *		ia_noOnPixel		didn't find a run.
 *
 **************************************************************************/
Int32  CDECL 
i_nextRunToTheRight (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	rightLimit,
    Int32	initialOffset,
    Bool        blackrun,
    Int32     *pRunLeft,
    Int32     *pRunRight)
{
    PROLOGUE;
    Int32	 byteOffset;		/* byte offset into the source line */
    Int32	 rightByteLimit;	/* byte offset of the word holding the
					 * rightmost pixel */
    UInt32	 inputWord;		/* the word we read from the image */
    Int32	 leftEnd, rightEnd;	/* the left and right ends of the run
					 */
    UInt32	 color = blackrun? 0x00000000: 0xffffffff;


	/* calculate the byte offset of the word holding the initial offset */
    byteOffset = (initialOffset/8) & ~0x3;
	/* turn rightLimit into the offset of the word holding the last byte */
    rightByteLimit = (rightLimit/8) & ~0x3;

    inputWord = ULONG_ACCESS(&pS[byteOffset])^color; /* get the first word */

	/* mask off the bits on the left that we should ignore */
    inputWord &= rightMask[(~initialOffset) & 0x1F];

	/* if there are no 1's in this word, go find a word with some 1's.
	 * BUG FIX: even if inputWord contains only 0's, we might
	 * still be at the right edge of the image.  Don't look forwards
	 * in that case.
	 */
    if (inputWord == 0)
    {
	if (byteOffset < rightByteLimit)
	{
	    byteOffset += 4;
	    while ( (byteOffset < rightByteLimit) &&
		    ( (ULONG_ACCESS(&pS[byteOffset])^color) == 0) )
		byteOffset += 4;
    
		/* At this point, we're either pointing at a non-zero word or
		 * at the last word of the scan line (or both).  Put this
		 * word into inputWord.  If it's 0, return ia_noOnPixel.
		 */
	    inputWord = ULONG_ACCESS(&pS[byteOffset])^color;
	}
	if (inputWord == 0)
	{
	    *pRunLeft = *pRunRight = 0;
	    return ia_noOnPixel;
	}
    }

	/* find the leftmost 1.  Do this by locating the 1 down to
	 * the byte, then using tables to find it in a byte.
	 * Note that we also set the 0's to the left of the leftmost
	 * 1 to 1.  This prevents us from finding them when we
	 * look for the first 0.
	 */
    if ((inputWord & 0xFFFF0000) != 0)
    {
	leftEnd = inputWord >> 16;	/* get the upper half.  We use
				     * leftEnd so we don't mess up
				     * inputWord.  We'll need it
				     * below to find the right
				     * end of the run
				     */
	if ((leftEnd & 0xFF00) != 0)
	{
		/* Left End of Run in Byte 0 */

	    leftEnd = firstBitTable[leftEnd >> 8];	/* get offset within
							 * word */
	    inputWord |= leftMask[leftEnd];
	    leftEnd += (byteOffset * 8);
		/* if that was off the right end of the area we're allowed to
		 * look at, return ia_noOnPixel
		 */
	    if (leftEnd > rightLimit)
	    {
		*pRunLeft = *pRunRight = 0;
		return ia_noOnPixel;
	    }
	    else
		*pRunLeft = leftEnd;
	}
	else
	{
		/* Left End of Run in Byte 1 */

	    leftEnd = 8 + firstBitTable[leftEnd & 0xFF];
	    inputWord |= leftMask[leftEnd];
	    leftEnd += (byteOffset * 8);
		/* if that was off the right end of the area we're allowed to
		 * look at, return ia_noOnPixel
		 */
	    if (leftEnd > rightLimit)
	    {
		*pRunLeft = *pRunRight = 0;
		return ia_noOnPixel;
	    }
	    else
		*pRunLeft = leftEnd;

		/* if inputWord is now not all 1's, find the other end
		 * of the run.  We put this short-cut in because runs
		 * are often short and because we know here that the
		 * run can't be in the top byte (so can save a test)
		 */
	    if (inputWord != 0xFFFFFFFF)
	    {
		inputWord = ~inputWord;
		if ((inputWord & 0xFFFF0000) != 0)	/* end is in byte 1 */
		{
		    rightEnd = (byteOffset * 8) + 8 +
				    firstBitTable[inputWord >> 16] - 1;
		    *pRunRight = MIN(rightLimit, rightEnd);
		    NORMAL_RETURN;
		}
		else
		if ((inputWord & 0xFF00) != 0)
		{
		    rightEnd = (byteOffset * 8) + 16 +
				    firstBitTable[inputWord >> 8] - 1;
		    *pRunRight = MIN(rightLimit, rightEnd);
		    NORMAL_RETURN;
		}
		else
		{
		    rightEnd = (byteOffset * 8) + 24 +
				    firstBitTable[inputWord & 0xFF] - 1;
		    *pRunRight = MIN(rightLimit, rightEnd);
		    NORMAL_RETURN;
		}
	    }
	}
    }
    else	/* The left end of the run starts in the lower
		 * half of the word */
    {
	if ((inputWord & 0xFF00) != 0)
	{
		/* Left End of Run in Byte 2 */

	    leftEnd = 16 + firstBitTable[inputWord >> 8];
	    inputWord |= leftMask[leftEnd];
	    leftEnd += (byteOffset * 8);
		/* if that was off the right end of the area we're allowed to
		 * look at, return ia_noOnPixel
		 */
	    if (leftEnd > rightLimit)
	    {
		*pRunLeft = *pRunRight = 0;
		return ia_noOnPixel;
	    }
	    else
		*pRunLeft = leftEnd;

		/* if inputWord is now not all 1's, find the other end
		 * of the run.  We put this short-cut in because runs
		 * are often short and because we know here that the
		 * run can't be in the top halfword (so can save a test)
		 */
	    if (inputWord != 0xFFFFFFFF)
	    {
		inputWord = ~inputWord;
		if ((inputWord & 0xFF00) != 0)
		{
		    rightEnd = (byteOffset * 8) + 16 +
				    firstBitTable[inputWord >> 8] - 1;
		    *pRunRight = MIN(rightLimit, rightEnd);
		    NORMAL_RETURN;
		}
		else
		{
		    rightEnd = (byteOffset * 8) + 24 +
				    firstBitTable[inputWord & 0xFF] - 1;
		    *pRunRight = MIN(rightLimit, rightEnd);
		    NORMAL_RETURN;
		}
	    }

	}
	else
	{
		/* Left End of Run in Byte 3 */

	    leftEnd = 24 + firstBitTable[inputWord & 0xFF];
	    inputWord |= leftMask[leftEnd];
	    leftEnd += (byteOffset * 8);
		/* if that was off the right end of the area we're allowed to
		 * look at, return ia_noOnPixel
		 */
	    if (leftEnd > rightLimit)
	    {
		*pRunLeft = *pRunRight = 0;
		return ia_noOnPixel;
	    }
	    else
		*pRunLeft = leftEnd;

		/* if inputWord is now not all 1's, find the other end
		 * of the run.  We put this short-cut in because runs
		 * are often short and because we know here that the
		 * run can't be in the top 3 bytes.
		 */
	    if (inputWord != 0xFFFFFFFF)
	    {
		inputWord = ~inputWord;
		rightEnd = (byteOffset * 8) + 24 +
				firstBitTable[inputWord & 0xFF] - 1;
		*pRunRight = MIN(rightLimit, rightEnd);
		NORMAL_RETURN;
	    }

	}
    }


	/* At this point, we've found the left end of a valid run.
	 * If the run started in bytes 1, 2, or 3 and the ended in
	 * bytes 1, 2, or 3, the run has already been returned.
	 * Therefore, either the run begins in byte 0, or it ends
	 * in some word to the right.
	 */

    inputWord = ~inputWord;
	/* if this word is filled with 1's (after filling in the left
	 * end), go find a word with some 0's.
	 * BUG FIX: even if inputWord contains only 1's, we might
	 * still be at the right edge of the image.  We don't
	 * want to read off the right edge.  On the other hand,
	 * we can't stop just because the memory location
	 * at byteOffset doesn't hold all 1's, because it
	 * might not.  Remember that the current value of inputWord
	 * has been modified, so may not reflect what's in memory.
	 * On the third hand, we have to do something to detect the
	 * case when the last run really goes to the right boundary,
	 * In my first bug "fix", inputWord contained all 1's
	 * if we started at the right boundary and it contained
	 * all 0's if we didn't.  Hmmm.
	 * Ok, so we always advance byteOffset and just don't
	 * do the read if byteOffset >= rightByteLimit.  Then we
	 * check to see if inputWord == 0 OR byteOffset >= rightByteLimit.
	 * In either case, the run ends at the right boundary.
	 */
    if (inputWord == 0)		
    {
	byteOffset += 4;
	while ( (byteOffset < rightByteLimit) &&
		( (ULONG_ACCESS(&pS[byteOffset])^color) == 0xFFFFFFFF) )
	    byteOffset += 4;
    
	    /* At this point, we're either pointing at a non-1's word or
	     * at the last word of the scan line (or both) OR we're
	     * pointing one word beyond the last word if we started
	     * in the last word.  If there's a word here, put it's
	     * inverse into inputWord.  If that inverse is all
	     * 0's or we're now beyond the rightLimit, return rightLimit
	     * as the right end of the run.
	     */
	if (byteOffset <= rightByteLimit)
	{
	    inputWord = ~( (ULONG_ACCESS(&pS[byteOffset]))^color);
	}
	else
	{
	    inputWord = 0;	/* our way of saying that the run ended
				 * at the rightLimit.
				 */
	}

	if (inputWord == 0)
	{
	    *pRunRight = rightLimit;
	    NORMAL_RETURN;
	}
    }

	/* find the leftmost 1 in the complemented word.  The rightmost
	 * bit of the run in 1 bit to the left of that.  Do this
	 * by locating the 1 down to the byte, then using tables to
	 * find it in a byte.
	 */
    if ((inputWord & 0xFFFF0000) != 0)
    {
	inputWord = inputWord >> 16;		/* get the upper half. */
	if ((inputWord & 0xFF00) != 0)
	{
	    rightEnd = (byteOffset * 8) +
			    firstBitTable[inputWord >> 8] - 1;
	    *pRunRight = MIN(rightLimit, rightEnd);
	    NORMAL_RETURN;
	}
	else
	{
	    rightEnd = (byteOffset * 8) + 8 +
			    firstBitTable[inputWord & 0xFF] - 1;
	    *pRunRight = MIN(rightLimit, rightEnd);
	    NORMAL_RETURN;
	}
    }
    else	/* there's something in the lower half of the word */
    {
	if ((inputWord & 0xFF00) != 0)
	{
	    rightEnd = (byteOffset * 8) + 16 +
			    firstBitTable[inputWord >> 8] - 1;
	    *pRunRight = MIN(rightLimit, rightEnd);
	    NORMAL_RETURN;
	}
	else
	{
	    rightEnd = (byteOffset * 8) + 24 +
			    firstBitTable[inputWord & 0xFF] - 1;
	    *pRunRight = MIN(rightLimit, rightEnd);
	    NORMAL_RETURN;
	}
    }

#ifdef __GNUC__	/* knuckleheaded gcc needs to see this */
    NORMAL_RETURN;
#endif /* __GNUC__ */
}


/*

 $Log: i_nxtrun.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.9  1998/04/08 16:39:05  aber
 aber on Wed Apr 8 09:38:29 PDT 1998

 i_nxtrun.c -> /product/ipcore/ipshared/src/RCS/i_nxtrun.c,v

 New parameter added, to check for black runs; changes via Ron D.

 Revision 1.1  1998/04/07 21:45:49  aber
 Initial revision

 Revision 1.8  1998/01/13 23:49:08  ddavies
 i_nxtrun.c -> /product/ipcore/ipshared/src/RCS/i_nxtrun.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.7  1996/05/21 18:50:42  ddavies
 i_nxtrun.c -> /product/ipcore/ipshared/src/RCS/i_nxtrun.c,v

 Fix comments.  Fix bug introduced by last change.  We did the right thing
 when looking for the left endpoint of runs about not running off the right
 edge of the image, but introduced a bug when looking for right endpoints.
 We now always increment the byteOffset, only look at the word if it's in
 bounds, load the inverse of the ending word if its in bounds or load 0
 otherwise.  This has the effect of correctly stopping all runs that end
 on word boundaries and at the right edge of the image.

 Revision 1.6  1996/05/18 00:23:08  ddavies
 i_nxtrun.c -> /product/ipcore/ipshared/src/RCS/i_nxtrun.c,v

 Fix bug in i_findNextRunToTheRight found by William Rucklidge.  It was
 possible to read off the right edge of frameless images.  We now check
 position before moving right to skip over words containing all 1's or all 0's.

 Revision 1.5  1995/12/13 02:45:05  ejaquith
 i_nxtrun.c -> /product/ipcore/ipshared/src/RCS/i_nxtrun.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1995/01/11  22:26:54  dferrell
 * i_nxtrun.c -> /product/ipcore/ipshared/src/RCS/i_nxtrun.c,v
 *
 * Changed include of shrcmprs.pub to shrcc.pub.
 *
 * Revision 1.3  1993/12/02  19:23:34  ddavies
 * i_nxtrun.c -> /product/ipcore/ipshared/src/RCS/i_nxtrun.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/12/01  04:03:23  ddavies
 * i_nxtrun.c -> /product/ipcore/ipshared/src/RCS/i_nxtrun.c,v
 *
 * Added i_nextRunToTheRight and moved i_nextBitToTheRight and i_nextBitToTheLeft
 * to i_nxtbt1.c.
 *
 * Revision 1.1  1993/11/20  02:55:36  ddavies
 * Initial revision
 *

 */

