/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_orthrt.c--includes
 *
 *	i_fastRot90():		high level code to rotate images 90 degrees
 *				clockwise
 *	i_fastRot90InPlace():	high level code to rotate images 90 degrees
 *				clockwise, leaving the result in the
 *				same image buffer
 *
 *	i_fastRot180():		high level code to rotate images 180 degrees.
 *				If the source and dest image pointers are
 *				equal, this routine does the rotation in place.
 *
 *	i_fastRot270():		high level code to rotate images 270 degrees
 *				clockwise
 *	i_fastRot270InPlace():	high level code to rotate images 270 degrees
 *				clockwise, leaving the result in the
 *				same image buffer
 */


#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#define	_NEED_REVERSEBYTE_
#define _NEED_REVERSEBYTE2_
#define _NEED_REVERSEBYTE4_
#define _NEED_REVERSEBYTE8_
#include "revbyte.prv"
#include "memory.pub"

    /* prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrpixr.pub"
#include "shrrast.pub"
#include "shros.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_orthrt.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* A reasonable way to rotate an image by 90 degrees is to divide the
 * image into rectangles, rotate the rectangles, and write them.  
 * The nice feature of this scheme is that we always write full 32-bit 
 * words.  The dimensions of the rectangles in pixels depend on the 
 * number of bits/pixel.
 *
 * For example, if we want to rotate a 1 bit/pixel image, the destination
 * rectangle are 32 bits wide, so the source rectangles are 32 bits tall.
 * The width of the source rectangles (height of the dest rectangles) is set
 * by the number of registers available for rotation.  Our compilers are
 * stupid enough to have trouble dealing with more than 8 registers, so
 * let's make the rectangles 8 columns wide.  Ok, so this is totally
 * bogus for the PC.  Well, too bad.  We tried a few different schemes and
 * the general plan of distributing the bits of one source thinger
 * (byte, halfword or word) among lots of destinations won every time.
 * In the C implemenation, we'll do source rectangles 8 bits wide and
 * 32 bits tall.
 *
 * How to rotate the squares?  One way is to test and branch on each bit,
 * conditionally ORing a mask into one of 8 dest registers.  Nice, but
 * this takes about three opcodes per bit.  Can't we do better?
 *
 * The C implementation used a lookup table to spread the input bits so
 * we could accumulate runs of output bits in a register.  It was all
 * very clever.  However, the SPARC has a really cool multiply step opcode
 * that we can use to do double shifts.  See ssorthrt.s for the cutsey
 * implementation.  The x86 has no such thing, so we'll just use the
 * the C implementation there.
 *
 * The files ssorthrt.s and miorthrt.c contain SPARC and machine independent
 * versions of code to rotate and write one rectangle.
 * We pass it the address of the lower or upper (depending on the
 * implementation) left corner of the source rect,
 * the upper left corner of the dest rect, the source and dest bytes/line.
 * Because the C implementation needs it, we also pass down the address
 * of the rotation table.
 *
 * In what order do we address the source and destination?  The 90 degree
 * rotation is defined to be clockwise, so the lower left corner becomes
 * the upper left corner.  The first block rotated will be in the lower
 * left corner and will be stored in the upper left corner of the destination
 * image.  To take advantage of the processor's cache, the second block
 * read will be just to the right of the first block read and will be stored
 * just below the first block stored.  Both the SS2 and the 486 have
 * write-through caches that don't fetch locations being written into the
 * cache implicitly, so the order in which we write dest words doesn't matter.
 * This may all change if we switch to processing images in bands.  There,
 * we will probably find that it's easier to read a column than to write a
 * column.
 *
 * What about images whose widths are not even multiples of 8 or whose
 * heights are not even mulitples of 32?  Look at the height first.  We
 * want the bottommost line of the source image to become the leftmost bits
 * of the leftmost words of the destination image.  That means we should
 * start the bottom rectangle exactly on the bottommost line of the
 * source image.  The topmost rectangles in the source image may well include
 * some of the upper frame.  Big deal, that upper frame will become the
 * right frame in the destination image.
 *
 * What about the right edge of the source image?  The source image will
 * likely end somewhere inside the rightmost 8 bit source words.  Again,
 * big deal.  The extra 0's to the right of the real image will appear
 * as part of the lower frame in the destination image.
 *
 * In both cases, we can safely rotate entire rectangles without worrying
 * about edge conditions.
 *
 *********************
 *
 * That was all fine for binary images.  Worked great.  Now, we have to
 * generalize to 2 bit/pixel images.
 *
 * The C implementation got slower, but more efficient for 2 bits/pixel.
 * Each byte contained 4 pixels.  When four such bytes were spread out
 * into a 32 bit word, there were 4 complete bytes without any shifting
 * and combining.  Cool.  That makes the table lookup algorithm harder
 * to beat.
 *
 * We'll try a very slightly modified version of the assembly code that
 * reads 32 bits (16 pixels) at a time, rotating them into 16 registers
 * using the SPARC's mulscc opcode.  Since we rotate larger blocks (in
 * units of bits), it should be slighly more efficient than the code for
 * 1 bit/pixel image, though still slighly slower than 1 cycle/bit.  We'll
 * see if this is a good tradeoff. (It was, barely)
 *
 *********************
 *
 * PC's don't have very much memory.  We want to be able to rotate gray images
 * by 90, 180 and 270 degrees.  This means that we need to add the ability
 * to rotate 4 and 8 bit images.  It also means we have to introduce rotate
 * in place functions.
 *
 * Adding 4 and 8 bit rotation is relatively straightforward.  We'll use
 * the C implementation only - no SPARC assembly since the multiply step
 * trick gets less efficient for larger pixels.  The primitives for
 * 4 and 8 bit rotation will be in miorth48.c.
 *
 * Rotation in place is more tricky.  The big problem arises in reading
 * using the source bytes/line and writing using the destination bytes/line
 * without smashing something you'll need to read later.  Just rotating
 * tiles doesn't solve the problem - where do you put the rotated tiles?
 * With what bytes/line?  There are cutesy tricks for running around an
 * image picking up pixels and depositing previous pixels.  Yuck.  The
 * trick I'll use came from Webster Research Center.  There are several
 * versions of the trick that vary only in whether you read columns and
 * write rows vs reading rows and writing columns.  I'll pick a version
 * that both fits with the primitives we have and seems to access the
 * processor caches in a nice way.
 *
 * We have the client supply us with a framed image whose unframed height
 * equals the unframed height of the rotated destination image
 * (width of the source image) and whose unframed width is a multiple
 * of 32 PIXELS (NOT BITS).  We'll call this image the "transfer buffer."
 * The area used for both source and final destination is called the
 * "image buffer."  The wider the transfer buffer, the faster the rotation.
 * Since the rotation primitives assume they can write into the frame,
 * we need a frame on the transfer buffer.
 *
 * Start with 90 degree clockwise rotation.  The process used for 270
 * degree rotation is different in only minor ways.
 * We divide the image buffer into two sections, one above the other
 * ("above" in the source's frame of reference).  The source bits that
 * haven't been rotated are in the top rectangle.  The destination bits
 * that have been rotated are on the bottom.  The trick is to read out
 * rows of source and write columns of destination.  How do we keep
 * the dest columns in the bottom part of the image buffer?  Doesn't
 * each column extend to the top of the image?  The trick is to write
 * columns of the correct height, but scale the number of bytes/line
 * of the dest bits so that the total rotated image fits below the
 * untouched source bits.
 *
 * To do the rotation, we repeat the following steps:
 *
 *   1.	Identify the next group of rows of source blocks that,
 *	when rotated, just fills the transfer buffer.
 *   2. Rotate that source rows into the transfer buffer.  The source
 *	rows should be immediately *above* the bottom of the image, which
 *	is immediately above the previously written destination bits in
 *	the image buffer.  Note that the bytes/line of the image
 *	in the transfer buffer is related only to the width of the transfer
 *	buffer, not the width of the source or dest images.
 *   3. We now have a vacant space above the dest bits already in the
 *	image (if any).  Rescale the dest image so that each line has
 *	enough bytes/line to hold the newly-rotated data.  This means
 *	moving the top existing dest line up to the top of the vacant space,
 *	skipping the right number of bytes/line, moving the second line
 *	up and so on.  Since we're going from top to bottom, we can use
 *	i_rasterOp.  Note that when we're rotating 90 degrees clockwise,
 *	we move the previously rotated bits to the left end of each
 *	dest line.  That's because things at the bottom of the image
 *	should end up at the left side of the rotated image.  When rotating
 *	by 270 degrees, we move the previously rotated data to the right
 *	side of each line so the latest stuff can appear on the left.
 *   4.	Move the bits from the transfer buffer into the now-vacant
 *	column on the right or left side of the destination.
 *
 * The last dest column may not fill the entire buffer, so the amounts
 * by which the destination image is moved around may be smaller.
 *
 * Things to worry about:
 *   1. Will the rotated image always fit into the memory allocated for the
 *	original image?  For example, say we have an image that's 1 pixel
 *	tall and 1024 pixels wide.  This takes up exactly 32 memory
 *	words (32 bits each).  When rotated, we have an image thats
 *	1 pixel wide and 1024 pixels tall.  This takes up 1024 words.
 *
 *	createPixr rounds the height of images up when creating images.
 *	It rounds the height up to (((h * d + 31)/32) * 32) / d.  Say
 *	d = 1.  This turns into (((1 * 1 + 31)/32) * 32) / 1 = 32 lines.
 *	The total number of words is then w * h = 32 * 32 = 1024.
 *	Say d = 8.  The equation is (((1 * 8 + 31)/32) * 32) / 8 = 32,
 *	so we're still ok.
 *
 *   2. Is the number of words allocated for frame still correct?
 *
 *	For every row in the original image, there are 32 * depth bits
 *	on either side for horizontal frame.  For every column of the
 *	image, there are 32 * depth bits on the top and the bottom.
 *	When we exchange rows for columns, the all the frame bits allocated
 *	for rows exchange neatly for those allocated for columns and
 *	the totals stay the same.
 *
 *   3. How do we make sure the dest image ends up in the right place
 *	without a lot of expensive shifting?
 *
 *	This is one of those things you do for awhile before discovering
 *	the principle behind it.  Shifting vertically in an image is
 *	easy, you just change the horizontal line pointer.  Shifting
 *	horizontally is relatively hard, you actually have to move bits
 *	around inside words.  We want to avoid horizontal shifting.
 *
 *	When doing 90 degree rotation, horizontal motion in the source
 *	becomes vertical motion in the destination and vice versa.  Therefore,
 *	we need to adjust our vertical position in the source image
 *	to guarantee no shifting is needed in the destination and adjust
 *	our vertical position in the destination to eliminate horizontal
 *	shifting in the source.
 *
 *	When doing 90 degree rotation, we want the bottom line of the
 *	source image to become the left line of the dest image.  Therefore,
 *	the rotation blocks should be aligned with the bottom line of the
 *	image.  Furthermore, we want to align them with the left line of
 *	the source image so we don't have to do any source shifting.
 *	This line will become the top line of the destination, so we start
 *	writing the destination buffer at the top image line.
 *
 *	When doing 270 degree rotation, the *top* line of the source will
 *	become the left line of the dest.  What happens if we start at
 *	the top of the source image?  Well, the area allocated for the
 *	destination grows down instead of up.  That means we have to
 *	move from bottom to top when we make space in each line for newly
 *	rotated data.  If we went through the thing from top to bottom,
 *	we'd end up copying the first scan line on top of the second, then
 *	copying that on top of the third and so on - replicating the first
 *	scan line everywhere.  i_rasterOp only understands top to bottom
 *	transfers.  But wait!  Say we just give i_rasterOp a negative number
 *	for the dest bytes/line and started at the bottom of the dest area?
 *	Since the source and dest area for no scan line overlap during
 *	the enlargement process until we get to the top scan line, this
 * 	will work!  The top scan line will move over to the space
 *	occupied by the second scan line, but when that happens, the
 *	second scan line will be safely out of the way.  So, we'll start
 *	at the top of the image when do ing 270 degree rotations.
 *
 *	The left line of the source will become the bottom line of the dest
 *	when doing 270 degree rotations.  Therefore, we will start
 *	writing the transfer buffer at the bottom image line.  When
 *	rotating the rightmost image block, we will probably rotate some
 *	frame bits into the top frame of the transfer buffer, but that's
 *	ok.
 *
 *	NOT!  Say we keep pointing to the left ends of lines but set the
 *	number of bytes/line negative.  We happily move the image from
 *	bottom to top, but we go through each line from left to right.
 *	Say the difference between the length of the old lines and the
 *	new lines is less than the length of the old lines.  You can
 *	see that the very top old line overlaps the very top new line
 *	because we didn't move the start of the second line over
 *	far enough.  Bummer, part of the old lines gets duplicated.
 *
 *	So, we try approach number 2.  We read the source from bottom
 *	to top, just as we do in 90 degree rotates.  This lets us handle
 *	the growing dest bits in the same way.  However, we must remember
 *	that the array of rectangles must still be vertically aligned with
 *	top source line, since it becomes the left dest line.  Therefore
 *	the first source address is computed from the corner of the
 *	the lowest rectangle in the image, not the bottom line of the
 *	image.
 *
 *	Furthermore, when we rotate these rows into the dest, we fill the
 *	dest from right to left, not left to right.  When stretching the
 *	existing dest bits, we must leave room on the left for new stuff,
 *	not on the right.  Finally, when the transfer buffer is only
 *	partly filled (because we reached the top of the source image),
 *	we read the part on the right, not the part on the left.
 *
 * Why go through the image in this fashion (through the source by rows)?
 * Going through the source by rows uses the processor caches well.
 * Since the dest buffer is probably not very wide, going through it
 * vertically isn't very different from going through it horizontally
 * in terms of cache accesses - things are still pretty local.  Making
 * room in the existing dest bits is a i_rasterOp sort of operation,
 * so it uses cache lines pretty well.
 *
 ********************************
 *
 * BUG BUG BUG.  Say we read a strip from the middle of the image.  It
 * has a relativly small number of frame bytes associated with it - only
 * the ones on the side.  Say we want to write the rotated strip into
 * the same space.  The rotated strip has a relatively large number of
 * frame bits since the height of the rotated image is usually a lot
 * larger than the height of the strip (strip is flat and wide, rotated
 * strip is tall and thin - both have frame down the sides).  Therefore,
 * the rotated, framed thing won't fit into the space left by the
 * unrotated framed thing.  Bummer.
 *
 * We could remove the frame from the source image before rotation and
 * add it on after rotation.  We can get a similar effect by simply
 * not bothering to include the frame when writing dest bits to the space
 * left free by rotated source bits.  When we're done rotating, we move
 * the frameless rotated bits up to the upper left corner of the framed
 * area and add the frame.
 *
 * What about differences in the space taken up by pad bits in the two
 * images?  If there are more pad bits in the rotated image than there are
 * in the unrotated image, it will take up more room.  For example, in the
 * case of 90 degree rotation, we repeatedly stretch the destination image
 * towards the upper left, leaving the lower right corner fixed in memory.  If
 * the rotated image were bigger, it would extend above the upper left
 * corner of the unrotated image.  Can it extent above the top frame of the
 * unrotated image?  No.  The dest image is padded to the nearest 32-bit
 * boundary.  The amount of padding is proportional to the height of the
 * rotated image.  This is equal to the width the unrotated image, hence
 * the width of the top frame of the unrotated image.  There are at most
 * 31 bits of padding on each line of the rotated image.  There are 
 * 32 pixels in the top frame of the unrotated image.  Since 32 is bigger
 * than 31, we can't run out of top frame.  The same goes for the bottom
 * frame in 270 degree rotation.
 *
 *******************************************
 *
 * After that all worked, we added rotation of 4 and 8 bit images.  The
 * cute algorithm that used tables to spread pixels out worked fine for that,
 * the multiply-step cheat has no particular advantage when the bits come
 * in big clumps.  So, we only have a C implementation of this algorithm.
 * Where do we put it?  The C and asm implementations of ip_rotateBlock90 and
 * ip_rotateBlock270 already branch on the depth of the image.  We could
 * modify both so they branch to common procedure when the depth is 4 or 8.
 * That seems pretty silly.  Instead, we just call the proper procedure,
 * ip_rotate48Block90 or ip_rotate48Block270 from the higher level
 * procedures directly.
 *
 */

 /* This is a public procedure */
/* PUBLIC */
/**************************************************************************
 *  i_fastRot90():  Performs a clockwise 90 degree rotation of an image.
 *
 *	destImage:		Ptr to upper left corner of dest image's frame
 *	srcImage:		Ptr to upper left corner of src image's frame
 *	srcWidth:		Width of source image in pixels including frame
 *				BUT NOT pad.  This is the normal
 *				definition of width passed to all the "i_"
 *				routines.
 *	srcHeight:		Height of source image in pixels including
 *				frame.
 *	srcPitchInBytes:	Number of bytes/line in source image including
 *				image, border, frame and pad.  This equals the
 *				number you add to a source address to move
 *				down one line.
 *	destPitchInBytes:	Ditto for dest image.
 *	depth:			Number of bits/pixel in src and dest images.
 *
 *	Return code:
 *	 ia_successful		good completion
 *	 ia_depthNotSupported	cannot rotate image with given number of
 *				bits/pixel (depth).
 *
 *
 **************************************************************************/
Int32  CDECL 
i_fastRot90 (
    IPState    *pIPState,
    UInt8      *destImage,
    UInt8      *srcImage,
    Int32	srcWidth,
    Int32	srcHeight,
    Int32	srcPitchInBytes,
    Int32	destPitchInBytes,
    Int32	depth)
{
PROLOGUE;
 UInt32	*pRotationTable;	/* pointer to lookup table used
				 * to expand bits of each byte */
Int32	 widthInBlocks;		/* number of blocks processed by
				 * the primitive routine in each
				 * horizontal pass through the image.
				 */
Int32	 srcHeightInBlocks;	/* number of rows of blocks in the
				 * image
				 */
Int32	 interBlockSrcFixup;	/* number to be added to source
				 * address to go from bottom of one
				 * completed source rect to top
				 * of next new source rect. */
Int32	 interRowSrcFixup;	/* number to be added to source
				 * address to go from one row of
				 * blocks to the next.
				 */
Int32	 dummyVar;		/* used in place of parms that the
				 * setup procs supply to the in-place
				 * versions of rotation
				 */


	/* is this an image depth we support? */
    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8))
	return ia_depthNotSupported;

    pRotationTable = NULL;

	/* strip off the frames */
    srcImage += (srcPitchInBytes << cLogFrameBits) + depth*cFrameBytes;
    destImage += (destPitchInBytes << cLogFrameBits) + depth*cFrameBytes;
    srcWidth -= 2 * cFrameBits;
    srcHeight -= 2 * cFrameBits;

	/* set up all necessary parameters */
    __error__ = ip_fastRot90Setup (pIPState, &srcImage,
				 srcWidth, srcHeight, depth,
				 srcPitchInBytes, &destImage, destPitchInBytes,
				 &pRotationTable, &widthInBlocks,
				 &srcHeightInBlocks,
				 &interBlockSrcFixup, &interRowSrcFixup,
				 &dummyVar);

    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* loop through rows of rectangles of source image from the
	 * bottom to the top. */
    for (;
	 srcHeightInBlocks > 0;
	 srcImage -= interRowSrcFixup, destImage += 4, srcHeightInBlocks--)
    {
	if ((depth & 3))	/* depth == 1 or depth == 2 */
	{
	    ip_rotateBlock90 (pIPState, srcImage, srcPitchInBytes, destImage,
			      destPitchInBytes, pRotationTable,
			      widthInBlocks, interBlockSrcFixup, depth);
	}	/* end of rotating all rows of rectangles */
	else 	/* depth == 4 or depth == 8 */
	{
	    ip_rotate48Block90 (pIPState, srcImage, srcPitchInBytes, destImage,
			        destPitchInBytes, pRotationTable,
			        widthInBlocks, interBlockSrcFixup, depth);
	}	/* end of rotating all rows of rectangles */
	IP_YIELD(errorFinish)
    }


    if (pRotationTable != NULL)
	FREE(pRotationTable);

    NORMAL_RETURN;

errorFinish:
    if (pRotationTable != NULL)
	FREE(pRotationTable);

    return ia_aborted;

}


/* PUBLIC */
/**************************************************************************
 *  i_fastRot90InPlace():  Performs a clockwise 90 degree rotation of an
 *			   image, leaving the result bits in the same
 *			   image buffer.
 *
 *	pBuff:			Ptr to upper left corner of the transfer
 *				buffer.  This is a framed image that
 *				will hold image bits temporarily during
 *				rotation.  The unframed height of this
 *				buffer equals the unframed height of the
 *				rotated image (width of source image) and
 *				its unframed width in PIXELS (not bits)
 *				must be a multiple of 32.
 *				It must have the same depth as the source
 *				image.
 *				Other things being equal, the wider the
 *				buffer, the faster the rotation operation.
 *				The frames of this buffer will be
 *				TRASHED on return.  We expect users to
 *				destroy this image immediately.
 *	srcImage:		Ptr to upper left corner of src image's frame
 *				NOTE!!  The space allocated for this
 *				image buffer must have had enough pad
 *				on the BOTTOM.  This is done by rounding
 *				allowing enough memory for a full height of
 *				(((((height * depth) + 31)/32) * 32) / depth) +
 *				2 * frame.  This is done by ipwrap's
 *				createPixr.
 *	srcWidth:		Width of source image in pixels including frame
 *				BUT NOT pad.  This is the normal
 *				definition of width passed to all the "i_"
 *				routines.
 *	srcHeight:		Height of source image in pixels including
 *				frame.
 *	bufWidth:		Width of transfer buffer including frame.
 *	srcPitchInBytes:	Number of bytes/line in source image including
 *				image, border, frame and pad.  This equals the
 *				number you add to a source address to move
 *				down one line.
 *	bufPitchInBytes:	Transfer Buffer's bytes/line.
 *	depth:			Number of bits/pixel in src and dest images.
 *
 *	Return code:
 *	 ia_successful		good completion
 *	 ia_depthNotSupported	cannot rotate image with given number of
 *				bits/pixel (depth).
 *	 ia_invalidParm:	width of transfer buffer is not a multiple
 *				of 32 pixels.
 *
 *
 **************************************************************************/
Int32  CDECL 
i_fastRot90InPlace (
    IPState    *pIPState,
    UInt8      *pBuff,
    UInt8      *srcImage,
    Int32	srcWidth,
    Int32	srcHeight,
    Int32	bufWidth,
    Int32	srcPitchInBytes,
    Int32	bufPitchInBytes,
    Int32	depth)
{
PROLOGUE;
 UInt32	*pRotationTable;	/* pointer to lookup table used
				 * to expand bits of each byte */
Int32	 widthInBlocks;		/* number of blocks processed by
				 * the primitive routine in each
				 * horizontal pass through the image.
				 */
Int32	 srcHeightInBlocks;	/* number of rows of blocks in the
				 * source image
				 */
Int32	 currentLinesInBuf;	/* the number of source lines that have
				 * been rotated into the buffer
				 */
Int32	 srcBlockHeight;	/* height of rotation block in
				 * units of source scan lines */
Int32	 interBlockSrcFixup;	/* number to be added to source
				 * address to go from bottom of one
				 * completed source rect to top
				 * of next new source rect. */
Int32	 interRowSrcFixup;	/* number to be added to source
				 * address to go from one row of
				 * blocks to the next.
				 */
UInt8	*pOldDestTop, *pNewDestTop; /* These pointers are used to keep
				 * track of the rotated data in the
				 * image buffer. pOldDestTop points to
				 * the existing rotated data.  pNewDestTop
				 * points to the top of the space currently
				 * available for dest data.
				 */
Int32	 oldDestBpl, newDestBpl; /* number of bytes/line in portion of
				 * dest image that's been transferred to
				 * the image buffer.  oldDestBpl is
				 * correct before the making room for the
				 * newest column from the transfer buffer
				 * and newDestBpl is what the bytes/line
				 * will be after making room.
				 */
Int32	 finalDestBpl;		/* the bytes/line of the rotated image.
				 * This must be calculated explicitly
				 * because it may depend on the environment
				 * (processor type, etc).
				 */
Int32	 totalDestWidth;	/* the width of the portion of the dest
				 * image that's been transferred to
				 * the image buffer.  We need to know
				 * this to know how much data to move
				 * when we're making room for data that's
				 * been rotated into the transfer buffer.
				 */
UInt8	*pS;			/* utility ptr to source image.  We save
				 * srcImage for clearing out the frame.
				 */
UInt8	*pModifiedBuf;		/* like pBuff, except this is modified
				 * by the setup routine in some
				 * implementation-dependent way.  This
				 * points to wherever the rotation routine
				 * wants to write the first rotated block.
				 */
UInt8	*pCurrentBuf;		/* like pModifiedBuf, but this is the
				 * pointer that gets modified each time
				 * we write a source row to a dest column
				 * in the buffer.  This pointer is
				 * reset to pModifiedBuf each time we
				 * start over in the buffer.
				 */


	/* is this an image depth we support? */
    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8))
	return ia_depthNotSupported;

    if (bufWidth & 0x1F)
	return ia_invalidParm;

    pRotationTable = NULL;

	/* strip off the frames */
    pS = srcImage + (srcPitchInBytes << cLogFrameBits) + depth*cFrameBytes;
    srcWidth -= 2 * cFrameBits;
    srcHeight -= 2 * cFrameBits;

	/* To be more efficient, we'll use the side frames of the
	 * transfer buffer as well as the middle.  Therefore,
	 * keep bufWidth the same and only skip the top frame.
	 */
    pBuff += (bufPitchInBytes << cLogFrameBits);
    pModifiedBuf = pBuff;

	/* Set pOldDestTop to point to the area below the source
	 * bits.  Since the bottom of the grid of rotation blocks is
	 * aligned with the bottom of the image, we used to just point
	 * to the top line of the bottom frame.  However, in the very
	 * last step of the rotation process, we will adjust the position
	 * of the final image, moving it either up or down in the memory
	 * block.  Moving up takes one copy, moving down takes two.
	 * To increase the chances that we'll be moving up (slightly),
	 * we'll starting writing on the bottom line of the unrotated
	 * source images's bottom frame..
	 */
    oldDestBpl = newDestBpl = 0;		/* start no data bits */
    pOldDestTop = pS + (srcHeight + cFrameBits - 1) * srcPitchInBytes -
			depth*cFrameBytes;
    pNewDestTop = pOldDestTop;

	/* set up all necessary parameters */
    __error__ = ip_fastRot90Setup (pIPState, &pS, srcWidth, srcHeight, depth,
				 srcPitchInBytes, &pModifiedBuf,
				 bufPitchInBytes, &pRotationTable,
				 &widthInBlocks, &srcHeightInBlocks,
				 &interBlockSrcFixup, &interRowSrcFixup,
				 &srcBlockHeight);

    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* This is a little bit complicated.  We want to loop through
	 * the entire source filling the transfer buffer with columns
	 * of destination bits, making room for these bits in the
	 * dest portion of the image buffer, then copying them into
	 * the image buffer.  This loop contains another loop, the
	 * one used to fill the transfer buffer from the source image.
	 * We are not guaranteed that the transfer buffer will always be
	 * filled.  It will probably not be filled when we reach the
	 * top of the source image.  Therefore, we must count the
	 * number of source rows (dest columns) we actually put into it
	 * so we can make room for, and copy, the right amount.
	 * The setup procedure returned srcBlockHeight, which tells us
	 * how many source rows (dest cols) there are in a rotated block.
	 * Since we fill the transfer buffer in units of these blocks, we
	 * know the number of cols in the transfer buffer equals the number
	 * of times we called ip_rotateBlock90 times srcBlockHeight.
	 * 
	 * Fine, but how do we know where the top of the image is?
	 * We need to keep track of two things - the number of source block
	 * rows that fit into the transfer buffer and the number of source block
	 * rows that fit in the source image.  We'll do that.
	 *  Variables:
	 *	currentBlocksInBuf:	The number of rows of source blocks
	 *				that have been rotated into the
	 *				transfer buffer.
	 *	pModifiedBuf:		Pointer to the first address in
	 *				the transfer buffer
	 *	pCurrentBuf:		Pointer to the next address in the
	 *				transfer buffer
	 *	srcHeightInBlocks:	The number of source rows in the
	 *				source image.
	 *	pS:			(probably modified) pointer to first
	 *				word to be rotated in next source
	 *				row inside the image buffer.
	 *	pOldDestTop:		Pointer to left word in top line
	 *				of dest pixels that have already been
	 *				rotated and put into the image
	 *				buffer.
	 *	pNewDestTop:		Points to left word of top line
	 *				of space available for dest pixels
	 *				after they have been expanded to
	 *				make room for newly rotated pixels.
	 *	totalDestWidth:		The width (in pixels) of the rotated
	 *				data that's currently in the image
	 *				buffer.  We use this when moving
	 *				that data to make room for new
	 *				stuff from the transfer buffer.
	 */
    totalDestWidth = 0;

    if (srcHeightInBlocks <= 0)		/* if there's nothing to do, just */
	goto successFinish;		/* quit */

    while (srcHeightInBlocks > 0)
    {
	    /* fill up the buffer, or stop if we run out of source image */
	for (currentLinesInBuf = 0, pCurrentBuf = pModifiedBuf;
	     (currentLinesInBuf < bufWidth) && (srcHeightInBlocks > 0);
	     currentLinesInBuf += srcBlockHeight, srcHeightInBlocks--,
		pS -= interRowSrcFixup, pCurrentBuf += 4)
	{
	    if (depth & 3)	/* depth == 1 or depth == 2 */
	    {
		ip_rotateBlock90 (pIPState, pS, srcPitchInBytes, pCurrentBuf,
				  bufPitchInBytes, pRotationTable,
				  widthInBlocks, interBlockSrcFixup, depth);
	    }
	    else		/* depth == 4 or depth == 8 */
	    {
		ip_rotate48Block90 (pIPState, pS, srcPitchInBytes, pCurrentBuf,
				    bufPitchInBytes, pRotationTable,
				    widthInBlocks, interBlockSrcFixup, depth);
	    }
	}

	IP_YIELD(errorFinish)

	    /* At this point, the transfer is normally full, but might
	     * be only partially filled.  That happens when we get to
	     * the top of the source image and run out of source bits
	     * before filling the buffer.
	     * First, make room for the bits that are actually in
	     * the buffer.  We do this by moving the rectangle of existing
	     * destination bits that's totalDestWidth x srcWidth
	     * (totalDestWidth x destHeight) pixels from an area of the
	     * image buffer where it is packed together with oldDestBpl
	     * bytes/line to one that has newDestBpl bytes/line.
	     * Note that we copy the existing dest pixels to the left
	     * side of the image so we can put the new stuff on the
	     * right.
	     * The very first time we do this, totalDestWidth == 0,
	     * so nothing happens.  That's correct since there are
	     * no dest bits to move.
	     */
	newDestBpl = oldDestBpl + (currentLinesInBuf * depth)/8;

	    /* set pNewDestTop.  The distance between the pOldDestTop
	     * and the bottom of the old rotated image is
	     * oldDestBpl*srcWidth.  The distance between the new dest
	     * top and bottom of the rotated image is newDestBpl*srcWidth.
	     * Therefore, we need to adjust pNewDestTop by
	     * (newDestBpl - oldDestBpl)*srcWidth.
	     */
	pNewDestTop -= (newDestBpl-oldDestBpl) * srcWidth;

	i_rasterOp (pIPState,
		    pNewDestTop,	/* dest ptr */
		    0, 0,		/* dest x, dest y */
		    totalDestWidth,	/* width of dest in pixels */
		    srcWidth,		/* height of dest */
		    depth,		/* dest depth */
		    newDestBpl,		/* dest bpl */
		    PIX_SRC,		/* operation */
		    pOldDestTop,	/* src pointer */
		    0, 0,		/* src x, src y */
		    depth,		/* src depth */
		    oldDestBpl);	/* src bytes/line */


	    /* Insert the rotated bits from the transfer buffer into
	     * the image buffer.
	     */
	i_rasterOp (pIPState,
		    pNewDestTop,	/* dest ptr */
		    totalDestWidth, 0,	/* dest x, dest y */
		    currentLinesInBuf,	/* width of dest in pixels */
		    srcWidth,		/* height of dest */
		    depth,		/* dest depth */
		    newDestBpl,		/* dest bpl */
		    PIX_SRC,		/* operation */
		    pBuff,		/* src pointer */
		    0, 0,		/* src x, src y */
		    depth,		/* src depth */
		    bufPitchInBytes);	/* src bytes/line */

	    /* Adjust the pointers to where the dest bits are in the
	     * image buffer.
	     */
	oldDestBpl = newDestBpl;
	pOldDestTop = pNewDestTop;
	totalDestWidth += currentLinesInBuf;

	IP_YIELD(errorFinish)

    }

	/* At this point, the entire image has been rotated.  It has
	 * no frame as yet.  We have two choices.  We can move the
	 * image up to the top of the framed area and call
	 * i_addFrameInPlace, or we can use i_rasterOp to add the
	 * area for the frame to the image and clear the frame
	 * explicitly.  The first method copies the image twice,
	 * so it would be nice to avoid it.  If making room for the frame
	 * using i_rasterOp means that any image words will move
	 * to higher memory addresses, i_rasterOp will copy some
	 * image words on top of others.  That's because it moves
	 * from low memory addresses to high memory addresses,
	 * if the source and dest overlap, it can write over
	 * source bits that haven't been read yet.
	 * i_addFrameInPlace is the only reliable
	 * means we have of copying an image from right-to-left,
	 * bottom-to-top and I don't want to invent another routine
	 * for this operation.  Therefore, we can use i_rasterOp if
	 * all the words are moved towards lower memory addresses.
	 * We can test this by seeing if the last image line moves to a
	 * lower address since the top of the image will
	 * always be moving farther up relative to the bottom of
	 * the image when we add the frame.
	 * UPDATE: Sometimes the bytes/line of a final image is
	 * padded to something larger than a 4 byte boundary (like
	 * 8 byte boundary).  We need to call i_calcImageProperties
	 * to get the bytes/line of the final image.  We used
	 * to just use newDestBpl + 2*depth*cFrameBytes.
	 * We use widthInBlocks as a convenient place to throw
	 * away the total image size needed.
	 */
    __error__ = i_calcImageProperties (pIPState, srcHeight, srcWidth,
				     depth, cFrameBits,
				     &finalDestBpl, &widthInBlocks);

    if ((srcImage + (cFrameBits + srcWidth - 1) * finalDestBpl) <
	(pNewDestTop + (srcWidth - 1) * newDestBpl))
    {
	    /* everything is moving, so do the one-copy method
	     * since i_rasterOp works.
	     */
	i_rasterOp (pIPState,
		    srcImage,		/* dest ptr */
		    cFrameBits,		/* dest x */
		    cFrameBits,		/* dest y */
		    srcHeight,		/* width of dest in pixels */
		    srcWidth,		/* height of dest */
		    depth,		/* dest depth */
		    finalDestBpl,	/* dest bpl */
		    PIX_SRC,		/* operation */
		    pNewDestTop,	/* src pointer */
		    0, 0,		/* src x, src y */
		    depth,		/* src depth */
		    newDestBpl);	/* src bytes/line */

	i_clearAllFrames (pIPState, srcImage,
			  srcHeight + 2*cFrameBits, /* width of image */
			  srcWidth + 2*cFrameBits,  /* height of image */
			  finalDestBpl,		   /* bytes/line */
			  depth);
    }
    else
    {
	/* Can't move the rotated image directly into position because
	 * i_rasterOp would write on source bits before reading them.
	 * Bummer.  So copy the source image up the the upper corner
	 * and add the frame explicitly.
	 */
	i_rasterOp (pIPState, srcImage,	/* dest ptr */
		    0, 0,		/* dest x, dest y */
		    srcHeight,		/* width of dest in pixels */
		    srcWidth,		/* height of dest */
		    depth,		/* dest depth */
		    newDestBpl,		/* dest bpl */
		    PIX_SRC,		/* operation */
		    pNewDestTop,	/* src pointer */
		    0, 0,		/* src x, src y */
		    depth,		/* src depth */
		    newDestBpl);	/* src bytes/line */
    
	    /* Add the frame to the image */
	i_addFrameInPlace (pIPState, srcImage, srcHeight, srcWidth,
			   depth, newDestBpl);
    }

successFinish:
    if (pRotationTable != NULL)
	FREE(pRotationTable);

    NORMAL_RETURN;

errorFinish:
    if (pRotationTable != NULL)
	FREE(pRotationTable);

    return ia_aborted;

}


 /* This is a public procedure */
/* PUBLIC */
/**************************************************************************
 *  i_fastRot270():  Performs a clockwise 270 degree rotation of an image.
 *
 *	destImage:		Ptr to upper left corner of dest image's frame
 *	srcImage:		Ptr to upper left corner of src image's frame
 *	srcWidth:		Width of source image in pixels including frame
 *				BUT NOT pad.  This is the normal
 *				definition of width passed to all the "i_"
 *				routines.
 *	srcHeight:		Height of source image in pixels including
 *				frame.
 *	srcPitchInBytes:	Number of bytes/line in source image including
 *				image, frame and pad.  This equals the
 *				number you add to a source address to move
 *				down one line.
 *	destPitchInBytes:	Ditto for dest image.
 *	depth:			Number of bits/pixel in src and dest images.
 *
 *	Return code:
 *	 ia_successful		good completion
 *	 ia_depthNotSupported	cannot rotate image with given number of
 *				bits/pixel (depth).
 *
 *
 **************************************************************************/
Int32  CDECL 
i_fastRot270 (
    IPState    *pIPState,
    UInt8      *destImage,
    UInt8      *srcImage,
    Int32	srcWidth,
    Int32	srcHeight,
    Int32	srcPitchInBytes,
    Int32	destPitchInBytes,
    Int32	depth)
{
PROLOGUE;
 UInt32		*pRotationTable;	/* pointer to lookup table used
					 * to expand bits of each byte */
Int32		 widthInBlocks;		/* number of blocks processed by
					 * the primitive routine in each
					 * horizontal pass through the image.
					 */
Int32		 srcHeightInBlocks;	/* number of rows of blocks in the
					 * image
					 */
Int32		 interBlockSrcFixup;	/* number to be added to source
					 * address to go from bottom of one
					 * completed source rect to top
					 * of next new source rect. */
Int32		 interRowSrcFixup;	/* number to be added to source
					 * address to go from one row of
					 * blocks to the next.
					 */
Int32		 dummyVar;		/* replaces variable that's only
					 * used by the in-place code.
					 */


	/* is this an image depth we support? */
    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8))
	return ia_depthNotSupported;

    pRotationTable = NULL;

	/* strip off the frames */
    srcImage += (srcPitchInBytes << cLogFrameBits) + depth*cFrameBytes;
    destImage += (destPitchInBytes << cLogFrameBits) + depth*cFrameBytes;
    srcWidth -= 2 * cFrameBits;
    srcHeight -= 2 * cFrameBits;

	/* set up all necessary parameters */
    __error__ = ip_fastRot270Setup (pIPState, &srcImage,
				  srcWidth, srcHeight, depth,
				  srcPitchInBytes, &destImage, srcHeight,
				  destPitchInBytes,
				  &pRotationTable, &widthInBlocks,
				  &srcHeightInBlocks,
				  &interBlockSrcFixup, &interRowSrcFixup,
				  &dummyVar);

    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* loop through rows of rectangles of source image from the
	 * bottom to the top. */
    for (;
	 srcHeightInBlocks > 0;
	 srcImage -= interRowSrcFixup, destImage -= 4, srcHeightInBlocks--)
    {
	if (depth & 3)	/* depth == 1 or depth == 2 */
	{
	    ip_rotateBlock270 (pIPState, srcImage, srcPitchInBytes, destImage,
			       destPitchInBytes, pRotationTable,
			       widthInBlocks, interBlockSrcFixup, depth);
	}	/* end of rotating all rows of rectangles */
	else		/* depth == 4 or depth == 8 */
	{
	    ip_rotate48Block270 (pIPState, srcImage, srcPitchInBytes, destImage,
			         destPitchInBytes, pRotationTable,
			         widthInBlocks, interBlockSrcFixup, depth);
	}
	IP_YIELD(errorFinish)

    }	/* end of rotating all rows of rectangles */

    if (pRotationTable != NULL)
	FREE(pRotationTable);

    NORMAL_RETURN;

errorFinish:
    if (pRotationTable != NULL)
	FREE(pRotationTable);

    return ia_aborted;

}


/* PUBLIC */
/**************************************************************************
 *  i_fastRot270InPlace():  Performs a clockwise 270 degree rotation of an
 *			    image, leaving the result bits in the same
 *			    image buffer.
 *
 *	pBuff:			Ptr to upper left corner of the transfer
 *				buffer.  This is a framed image that
 *				will hold image bits temporarily during
 *				rotation.  The unframed height of this
 *				buffer equals the unframed height of the
 *				rotated image (width of source image) and
 *				its unframed width in PIXELS (not bits)
 *				must be a multiple of 32.
 *				It must have the same depth as the source
 *				image.
 *				Other things being equal, the wider the
 *				buffer, the faster the rotation operation.
 *				The frames of this buffer will be
 *				TRASHED on return.  We expect users to
 *				destroy this image immediately.
 *	srcImage:		Ptr to upper left corner of src image's frame
 *				NOTE!!  The space allocated for this
 *				image buffer must have had enough pad
 *				on the BOTTOM.  This is done by rounding
 *				allowing enough memory for a full height of
 *				(((((height * depth) + 31)/32) * 32) / depth) +
 *				2 * frame.  This is done by ipwrap's
 *				createPixr.
 *	srcWidth:		Width of source image in pixels including frame
 *	srcHeight:		Height of source image in pixels including
 *				frame.
 *	bufWidth:		Width of transfer buffer including frame.
 *	srcPitchInBytes:	Number of bytes/line in source image including
 *				image, border, frame and pad.  This equals the
 *				number you add to a source address to move
 *				down one line.
 *	bufPitchInBytes:	Transfer Buffer's bytes/line.
 *	depth:			Number of bits/pixel in src and dest images.
 *
 *	Return code:
 *	 ia_successful		good completion
 *	 ia_depthNotSupported	cannot rotate image with given number of
 *				bits/pixel (depth).
 *	 ia_invalidParm:	width of transfer buffer is not a multiple
 *				of 32 pixels.
 *
 *
 **************************************************************************/
Int32  CDECL 
i_fastRot270InPlace (
    IPState    *pIPState,
    UInt8      *pBuff,
    UInt8      *srcImage,
    Int32	srcWidth,
    Int32	srcHeight,
    Int32	bufWidth,
    Int32	srcPitchInBytes,
    Int32	bufPitchInBytes,
    Int32	depth)
{
PROLOGUE;
 UInt32	*pRotationTable;	/* pointer to lookup table used
				 * to expand bits of each byte */
Int32	 widthInBlocks;		/* number of blocks processed by
				 * the primitive routine in each
				 * horizontal pass through the image.
				 */
Int32	 srcHeightInBlocks;	/* number of rows of blocks in the
				 * source image
				 */
Int32	 currentLinesInBuf;	/* the number of source lines that have
				 * been rotated into the buffer
				 */
Int32	 srcBlockHeight;	/* height of rotation block in
				 * units of source scan lines */
Int32	 interBlockSrcFixup;	/* number to be added to source
				 * address to go from bottom of one
				 * completed source rect to top
				 * of next new source rect. */
Int32	 interRowSrcFixup;	/* number to be added to source
				 * address to go from one row of
				 * blocks to the next.
				 */
UInt8	*pOldDestTop, *pNewDestTop; /* These pointers are used to keep
				 * track of the rotated data in the
				 * image buffer. pOldDestTop points to
				 * the existing rotated data.  pNewDestTop
				 * points to the top of the space currently
				 * available for dest data.
				 */
Int32	 oldDestBpl, newDestBpl; /* number of bytes/line in portion of
				 * dest image that's been transferred to
				 * the image buffer.  oldDestBpl is
				 * correct before the making room for the
				 * newest column from the transfer buffer
				 * and newDestBpl is what the bytes/line
				 * will be after making room.
				 */
Int32	 finalDestBpl;		/* the bytes/line of the rotated image.
				 * This must be calculated explicitly
				 * because it may depend on the environment
				 * (processor type, etc).
				 */
Int32	 totalDestWidth;	/* the width of the portion of the dest
				 * image that's been transferred to
				 * the image buffer.  We need to know
				 * this to know how much data to move
				 * when we're making room for data that's
				 * been rotated into the transfer buffer.
				 */
UInt8	*pS;			/* utility ptr to source image.  We save
				 * srcImage for clearing out the frame.
				 */
UInt8	*pModifiedBuf;		/* like pBuff, except this is modified
				 * by the setup routine in some
				 * implementation-dependent way.  This
				 * points to wherever the rotation routine
				 * wants to write the first rotated block.
				 */
UInt8	*pCurrentBuf;		/* like pModifiedBuf, but this is the
				 * pointer that gets modified each time
				 * we write a source row to a dest column
				 * in the buffer.  This pointer is
				 * reset to pModifiedBuf each time we
				 * start over in the buffer.
				 */


	/* is this an image depth we support? */
    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8))
	return ia_depthNotSupported;

    if (bufWidth & 0x1F)
	return ia_invalidParm;

    pRotationTable = NULL;

	/* strip off the frames */
    pS = srcImage + (srcPitchInBytes << cLogFrameBits) + depth*cFrameBytes;
    srcWidth -= 2 * cFrameBits;
    srcHeight -= 2 * cFrameBits;

	/* To be more efficient, we'll use the side frames of the
	 * transfer buffer as well as the middle.  Therefore,
	 * keep bufWidth the same and only skip the top frame.
	 */
    pBuff += (bufPitchInBytes << cLogFrameBits);
    pModifiedBuf = pBuff;

	/* Set pOldDestTop to point to the area below the source
	 * bits.  There were several buggy attempts to do this.
	 * First, we set it to point to the first line of the bottom
	 * frame as is done in 90 degree rotation.  That failed here
	 * because we often read below the top of the frame.  That's
	 * because the grid of rotation rectangles is aligned with the
	 * TOP of the unrotated image, not the bottom, so some part of
	 * the bottom row of rectangles usually hangs over into
	 * the frame.
	 * Then we tried just setting pOldDestTop to point to the line
	 * below pS after calling ip_fastRot270Setup.  Closer, but still
	 * no good.  When doing setup for the SPARC assembly implementation,
	 * pS does indeed point to the lower left corner of the first block
	 * so setting pOldDestTop below that always works.  However, when
	 * using the machine independent rotation code, pS is set to point
	 * to the TOP of the bottom row of blocks, so pOldDestTop
	 * would point inside the bottom row instead of below it.
	 * Finally, we try brute force.  We just make pOldDestTop point
	 * to the start of the bottom row of the frame.  The tallest rotation
	 * block takes 32 lines of source image, therefore can only overhang
	 * the top 31 lines of the bottom frame.  Since the frame is 32
	 * lines tall, we can always safely point at the bottom line.
	 */
    oldDestBpl = newDestBpl = 0;		/* start no data bits */
    pOldDestTop = pS +(srcHeight + cFrameBits - 1) * srcPitchInBytes -
			depth*cFrameBytes;
    pNewDestTop = pOldDestTop;

	/* set up all necessary parameters */
    __error__ = ip_fastRot270Setup (pIPState, &pS, srcWidth, srcHeight, depth,
				  srcPitchInBytes, &pModifiedBuf, bufWidth,
				  bufPitchInBytes,
				  &pRotationTable, &widthInBlocks,
				  &srcHeightInBlocks,
				  &interBlockSrcFixup, &interRowSrcFixup,
				  &srcBlockHeight);

    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* This is a little bit complicated.  We want to loop through
	 * the entire source filling the transfer buffer with columns
	 * of destination bits, making room for these bits in the
	 * dest portion of the image buffer, then copying them into
	 * the image buffer.  This loop contains another loop, the
	 * one used to fill the transfer buffer from the source image.
	 * We are not guaranteed that the transfer buffer will always be
	 * filled.  It will probably not be filled when we reach the
	 * top of the source image.  Therefore, we must count the
	 * number of source rows (dest columns) we actually put into it
	 * so we can make room for, and copy, the right amount.
	 * The setup procedure returned srcBlockHeight, which tells us
	 * how many source rows (dest cols) there are in a rotated block.
	 * Since we fill the transfer buffer in units of these blocks, we
	 * know the number of cols in the transfer buffer equals the number
	 * of times we called ip_rotateBlock270 times srcBlockHeight.
	 * 
	 * Fine, but how do we know where the top of the image is?
	 * We need to keep track of two things - the number of source block
	 * rows that fit into the transfer buffer and the number of source block
	 * rows that fit in the source image.  We'll do that.
	 *  Variables:
	 *	currentBlocksInBuf:	The number of rows of source blocks
	 *				that have been rotated into the
	 *				transfer buffer.
	 *	pModifiedBuf:		Pointer to the first address in
	 *				the transfer buffer
	 *	pCurrentBuf:		Pointer to the next address in the
	 *				transfer buffer
	 *	srcHeightInBlocks:	The number of source rows in the
	 *				source image.
	 *	pS:			Pointer to first word to be
	 *				rotated in next source row inside
	 *				the image buffer.
	 *	pOldDestTop:		Pointer to left word in top line
	 *				of dest pixels that have already been
	 *				rotated and put into the image
	 *				buffer.
	 *	pNewDestTop:		Points to left word of top line
	 *				of space available for dest pixels
	 *				after they have been expanded to
	 *				make room for newly rotated pixels.
	 *	totalDestWidth:		The width (in pixels) of the rotated
	 *				data that's currently in the image
	 *				buffer.  We use this when moving
	 *				that data to make room for new
	 *				stuff from the transfer buffer.
	 */
    totalDestWidth = 0;

    if (srcHeightInBlocks <= 0)		/* if there's nothing to do, just */
	goto successFinish;		/* quit */

    while (srcHeightInBlocks > 0)
    {
	    /* fill up the buffer, or stop if we run out of source image */
	for (currentLinesInBuf = 0, pCurrentBuf = pModifiedBuf;
	     (currentLinesInBuf < bufWidth) && (srcHeightInBlocks > 0);
	     currentLinesInBuf += srcBlockHeight, srcHeightInBlocks--,
		pS -= interRowSrcFixup, pCurrentBuf -= 4)
	{
	    if (depth & 3)	/* depth == 1 or depth == 2 */
	    {
		ip_rotateBlock270 (pIPState, pS, srcPitchInBytes, pCurrentBuf,
				   bufPitchInBytes, pRotationTable,
				   widthInBlocks, interBlockSrcFixup, depth);
	    }
	    else		/* depth == 4 or depth == 8 */
	    {
		ip_rotate48Block270 (pIPState, pS, srcPitchInBytes, pCurrentBuf,
				     bufPitchInBytes, pRotationTable,
				     widthInBlocks, interBlockSrcFixup, depth);
	    }
	}
	IP_YIELD(errorFinish)

	    /* At this point, the transfer buffer is normally full, but might
	     * be only partially filled.  That happens when we get to
	     * the top of the source image and run out of source bits
	     * before filling the buffer.
	     * First, make room for the bits that are actually in
	     * the buffer.  We do this by moving the rectangle of existing
	     * destination bits that's totalDestWidth x srcWidth
	     * (totalDestWidth x destHeight) pixels from an area of the
	     * image buffer where it is packed together with oldDestBpl
	     * bytes/line to one that has newDestBpl bytes/line.
	     * Note that we copy the existing dest pixels to the left
	     * side of the image so we can put the new stuff on the
	     * right.
	     * The very first time we do this, totalDestWidth == 0,
	     * so nothing happens.  That's correct since there are
	     * no dest bits to move.
	     */
	newDestBpl = oldDestBpl + (currentLinesInBuf * depth)/8;

	    /* set pNewDestTop.  The distance between the pOldDestTop
	     * and the bottom of the old rotated image is
	     * oldDestBpl*srcWidth.  The distance between the new dest
	     * top and bottom of the rotated image is newDestBpl*srcWidth.
	     * Therefore, we need to adjust pNewDestTop by
	     * (newDestBpl - oldDestBpl)*srcWidth.
	     */
	pNewDestTop -= (newDestBpl-oldDestBpl) * srcWidth;

	    /* Remember that we need to make room for new data on the
	     * left, not right.
	     */
	i_rasterOp (pIPState,
		    pNewDestTop,	/* dest ptr */
		    currentLinesInBuf,	/* dest x, leave room for buffer
					 * data on the left */
		    0,			/* dest y */
		    totalDestWidth,	/* width of dest in pixels */
		    srcWidth,		/* height of dest */
		    depth,		/* dest depth */
		    newDestBpl,		/* dest bpl */
		    PIX_SRC,		/* operation */
		    pOldDestTop,	/* src pointer */
		    0, 0,		/* src x, src y */
		    depth,		/* src depth */
		    oldDestBpl);	/* src bytes/line */


	    /* Insert the rotated bits from the transfer buffer into
	     * the image buffer.
	     */
	i_rasterOp (pIPState,
		    pNewDestTop,	/* dest ptr */
		    0,			/* dest x, insert new data at left */
		    0,			/* dest y */
		    currentLinesInBuf,	/* width of dest in pixels */
		    srcWidth,		/* height of dest */
		    depth,		/* dest depth */
		    newDestBpl,		/* dest bpl */
		    PIX_SRC,		/* operation */
		    pBuff,		/* src pointer */
		    bufWidth-currentLinesInBuf, /* fetch stuff from
					 * right side of buffer */
		    0,			/* src y */
		    depth,		/* src depth */
		    bufPitchInBytes);	/* src bytes/line */

	    /* Adjust the pointers to where the dest bits are in the
	     * image buffer.
	     */
	oldDestBpl = newDestBpl;
	pOldDestTop = pNewDestTop;
	totalDestWidth += currentLinesInBuf;
	IP_YIELD(errorFinish)

    }

	/* At this point, the entire image has been rotated.  It has
	 * no frame as yet.  We have two choices.  We can move the
	 * image up to the top of the framed area and call
	 * i_addFrameInPlace, or we can use i_rasterOp to add the
	 * area for the frame to the image and clear the frame
	 * explicitly.  The first method copies the image twice,
	 * so it would be nice to avoid it.  If making room for the frame
	 * using i_rasterOp means that any image words will move
	 * to higher memory addresses, i_rasterOp will copy some
	 * image words on top of others.  That's because it moves
	 * from low memory addresses to high memory addresses,
	 * if the source and dest overlap, it can write over
	 * source bits that haven't been read yet.
	 * i_addFrameInPlace is the only reliable
	 * means we have of copying an image from right-to-left,
	 * bottom-to-top and I don't want to invent another routine
	 * for this operation.  Therefore, we can use i_rasterOp if
	 * all the words are moved towards lower memory addresses.
	 * We can test this by seeing if the last image line moves to a
	 * lower address since the top of the image will
	 * always be moving farther up relative to the bottom of
	 * the image when we add the frame.
	 * UPDATE: Sometimes the bytes/line of a final image is
	 * padded to something larger than a 4 byte boundary (like
	 * 8 byte boundary).  We need to call i_calcImageProperties
	 * to get the bytes/line of the final image.  We used
	 * to just use newDestBpl + 2*depth*cFrameBytes.
	 * We use widthInBlocks as a convenient place to throw
	 * away the total image size needed.
	 */
    __error__ = i_calcImageProperties (pIPState, srcHeight, srcWidth,
				     depth, cFrameBits,
				     &finalDestBpl, &widthInBlocks);

    if ((srcImage + (cFrameBits + srcWidth - 1) * finalDestBpl) <
	(pNewDestTop + (srcWidth - 1) * newDestBpl))
    {
	    /* everything is moving, so do the one-copy method
	     * since i_rasterOp works.
	     */
	i_rasterOp (pIPState,
		    srcImage,		/* dest ptr */
		    cFrameBits,		/* dest x */
		    cFrameBits,		/* dest y */
		    srcHeight,		/* width of dest in pixels */
		    srcWidth,		/* height of dest */
		    depth,		/* dest depth */
		    finalDestBpl,	/* dest bpl */
		    PIX_SRC,		/* operation */
		    pNewDestTop,	/* src pointer */
		    0, 0,		/* src x, src y */
		    depth,		/* src depth */
		    newDestBpl);	/* src bytes/line */

	i_clearAllFrames (pIPState, srcImage,
			  srcHeight + 2*cFrameBits, /* width of image */
			  srcWidth + 2*cFrameBits,  /* height of image */
			  finalDestBpl,		   /* bytes/line */
			  depth);
    }
    else
    {
	/* Can't move the rotated image directly into position because
	 * i_rasterOp would write on source bits before reading them.
	 * Bummer.  So copy the source image up the the upper corner
	 * and add the frame explicitly.
	 */
	i_rasterOp (pIPState, srcImage,		/* dest ptr */
		    0, 0,		/* dest x, dest y */
		    srcHeight,		/* width of dest in pixels */
		    srcWidth,		/* height of dest */
		    depth,		/* dest depth */
		    newDestBpl,		/* dest bpl */
		    PIX_SRC,		/* operation */
		    pNewDestTop,		/* src pointer */
		    0, 0,		/* src x, src y */
		    depth,		/* src depth */
		    newDestBpl);		/* src bytes/line */
    
	    /* Add the frame to the image */
	i_addFrameInPlace (pIPState, srcImage, srcHeight,
			   srcWidth, depth, newDestBpl);
    }

successFinish:
    if (pRotationTable != NULL)
	FREE(pRotationTable);

    NORMAL_RETURN;

errorFinish:
    if (pRotationTable != NULL)
	FREE(pRotationTable);

    return ia_aborted;

}


/* 180 degree rotation is quite different from 90 or 270 degree rotation.
 * All we have to do is reverse the order of the bits in each line and
 * reverse the order of the lines.
 *
 * We'll scan the source image from left to right, top to bottom.  There is
 * some small chance that this will make the cache circuitry happier (there
 * is no more conventional way to scan memory).  We'll scan the destination
 * image from right to left, bottom to top.
 *
 * In the inner loop, we read four bytes of source word, reverse the order
 * of the bits within the bytes using a lookup table, reverse the order
 * of the bytes within the source word, and write the reversed dest word.
 * Because it's easiest to read bytes into the bottom of a register, we'll
 * shift the dest left to make room for newly reversed source bytes.  This
 * means the first byte read will end up on the left end of the dest word.
 * Therefore, we should read the rightmost byte of the source word first.
 *
 * Well, that's all very nice, but it only works if the read image bits
 * exactly fit into 32 bit words.  Not likely.  If the rightmost 32-bit word
 * of each row has n frame bits on the right end, the rotated image will
 * be shifted right by these very same n bits.  This is no good.
 *
 * To fix this, we'll have two cases: the fast one where everything happens
 * to fit in exactly 32 bits and the somewhat slower one where some shifting
 * is required.  If shifting is required, we maintain two accumulator words,
 * newWord and oldWord.  We do the byte reversing as described above to
 * fill newWord with the contents of one reversed word.  Since we're writing
 * the destination line from right to left, oldWord is trailing along to the
 * right of newWord from the destination's point of view.  Therefore, the
 * word written to memory is generated by shifting oldWord right by 32-padBits
 * and shifting newWord left by padBits, then ORing them.  After the write,
 * we stuff newWord into oldWord and continue.
 *
 * How do we start and how many times do we go through the loop?  Initially,
 * we fill oldWord with 0's.  This is used to generate the frame bits on the
 * right end of the destination line.
 * We go through the loop enough times to complete writing an entire dest
 * line.
 *
 * That worked fine for 1 bit/pixel.  How do we extend it to 2 bits/pixel?
 * In the case where the width is an exact multiple of 32, everything's
 * too easy.  We just use a different table to reverse the order of pixels
 * within bytes.  That is, pixels are moved in units of 2 bits instead of
 * moving them in units of 1 bit.  For example, byte 0000 0001 goes to
 * 0100 0000 instead of 1000 0000.
 *
 * For the cases that require shifting, the only other thing we modify is
 * calculation of the position of the rightmost bit in the memory image.
 * Instead of simply taking srcWidth & 0x1F (equals srcWidth REM 32), we
 * calculate (srcWidth * depth) & 0x1F.  This is done because these images
 * are "depth" times as wide as the 1 bit/pixel images.
 *
 * Ok, there's one more tweak.  "White" pixels in binary images are 0.
 * White pixels in gray images are 1's.  When we check to see if we can
 * skip the reversing step, we check the incoming word against a variable
 * that's set as a function of depth (0 for binary, 0xFFFFFFFF) for gray).
 * If we're being picky, words full of all 0's or all 1's don't need to be
 * reversed in any images, but testing for all 0's OR all 1's requires an
 * extra opcode, and there's not too much black, so we won't bother
 * trying to detect black.
 *
 */

/* This is a public procedure */
/* PUBLIC */
/****************************************************************************
 *  i_fastRot180():  Performs a clockwise 180 degree rotation of an image.
 *
 *     Inputs:
 *	destImage:		Ptr to upper left corner of dest image's frame
 *	srcImage:		Ptr to upper left corner of src image's frame
 *	srcWidth:		Width of source image in pixels including frame
 *				BUT NOT pad.  This is the normal
 *				definition of width passed to all the "i_"
 *				routines.
 *	srcHeight:		Height of source image in pixels including
 *				frame.
 *	srcPitchInBytes:	Number of bytes/line in source image including
 *				image, frame and pad.  This equals the
 *				number you add to a source address to move
 *				down one line.
 *	destPitchInBytes:	Ditto for dest image.
 *	depth:			Number of bits/pixel in src and dest images.
 *
 *	If destImage = srcImage, then the rotation happens in place, i.e.
 *	the source image itself is rotated.
 *
 *	Return code:
 *	 ia_successful		good completion
 *	 ia_depthNotSupported	cannot rotate image with given number of
 *				bits/pixel (depth).
 *	 ia_nomem		not enough memory to allocate the buffer
 *				used for rotation in place
 *
 ***************************************************************************/
Int32  CDECL 
i_fastRot180 (
    IPState    *pIPState,
    UInt8      *destImage,
    UInt8      *srcImage,
    Int32	srcWidth,
    Int32	srcHeight,
    Int32	srcPitchInBytes,
    Int32	destPitchInBytes,
    Int32	depth)
{
PROLOGUE;
 UInt32		 widthInWords;		/* width of image in words */
UInt32		 rightShift;		/* distance by which dest line
					 * is shifted after being reversed. */
const UInt8	*pRevTable;		/* ptr to table used to reverse
					 * pixels within bytes */
UInt8		*pOneLine;		/* a buffer that holds 1 line of
					 * reversed pixels.  Used when doing
					 * rotation in place
					 */
UInt32		 frameInBytes;		/* width of frame in bytes */

	/* is this an image depth we support? */
    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8))
	return ia_depthNotSupported;

	/* strip off the frames */
    srcImage += (srcPitchInBytes << cLogFrameBits) + depth*cFrameBytes;
    destImage += (destPitchInBytes << cLogFrameBits) + depth*cFrameBytes;
    srcWidth -= 2 * cFrameBits;
    srcHeight -= 2 * cFrameBits;

	/* set up useful vars */
    switch (depth) {
    case 1:
	widthInWords = (srcWidth + 31) >> 5;
	rightShift = srcWidth & 0x1F;		/* get shift distance */
	pRevTable = (const UInt8 *) reverseByte; /* use table for reversing
						 * binary image bytes. */
	break;
    case 2:
	widthInWords = ((srcWidth << 1) + 31) >> 5;
	rightShift = (srcWidth << 1) & 0x1F;	/* get shift distance */
	pRevTable = (const UInt8 *) reverseByte2; /* use table for reversing
						 * 2 bit/pixel image bytes. */
	break;
    case 4:
	widthInWords = ((srcWidth << 2) + 31) >> 5;
	rightShift = (srcWidth << 2) & 0x1F;	/* get shift distance */
	pRevTable = (const UInt8 *) reverseByte4; /* use table for reversing
						 * 4 bit/pixel image bytes. */
	break;
    default:	/* depth == 8 */
	widthInWords = ((srcWidth << 3) + 31) >> 5;
	rightShift = (srcWidth << 3) & 0x1F;	/* get shift distance */
	pRevTable = (const UInt8 *) reverseByte8; /* use table for reversing
						 * 4 bit/pixel image bytes. */
	break;
    }

	/* if srcImage = destImage, we do the rotation in place, which
	 * takes longer.  Check to see if this is the easy case.
	 */
    if (srcImage != destImage)
    {
	    /* Set dest image pointer to word in lower left corner
	     * of image */
	destImage += ((srcHeight-1) * destPitchInBytes);
    
	    /* call the primitive routine for each line. */
	for (;
	     srcHeight > 0;
	     srcImage += srcPitchInBytes, destImage -= destPitchInBytes,
		    srcHeight--)
	{
	    ip_reverseLine (pIPState, srcImage, destImage,
			    widthInWords, rightShift, pRevTable);
	    IP_YIELD(errorFinish)
	}
    }
    else
    {
	    /* Set dest image pointer to first word in lower left corner
	     * of image */
	destImage += ((srcHeight-1) * destPitchInBytes);
    
	    /* Allocate a 1-line buffer we can use to hold reversed lines
	     */
	frameInBytes = depth * cFrameBytes;
	pOneLine = (UInt8 *)CALLOC(widthInWords*4+frameInBytes, sizeof(UInt8));
	if (pOneLine == NULL)
	    return ia_nomem;
	pOneLine += frameInBytes;	/* we always point inside the frame */

	    /* call the primitive routine for each pair of lines.
	     * We reverse the upper line into the buffer, then 
	     * reverse the lower line into the upper line, finally,
	     * copy the reversed upper line into the lower line.
	     *
	     * CHEAT!  We need to remember to do something special
	     * if the number of image lines is odd.  In that case, we'd need
	     * to reverse the middle line to pOneLine, then copy
	     * it back to the middle line (instead of doing two
	     * calls to ip_reverseLine).  We could do this first,
	     * but then we'd have to calculate the address of the
	     * middle line.  So, we need some way to remember to do
	     * the special thing after flipping all the regular lines.
	     * We appropriate destPitchInBytes, putting srcHeight in
	     * it.  When we get to the end of the regular loop, we
	     * test the LS bit of destPitchInBytes.  If it's set,
	     * we do the middle line.  At that point, srcImage will
	     * be pointing to it.
	     */
	destPitchInBytes = srcHeight;
	for (srcHeight = srcHeight/2;
	     srcHeight > 0;
	     srcImage += srcPitchInBytes, destImage -= srcPitchInBytes,
		    srcHeight--)
	{
	    ip_reverseLine (pIPState, srcImage, pOneLine,
			    widthInWords, rightShift, pRevTable);
	    ip_reverseLine (pIPState, destImage, srcImage,
			    widthInWords, rightShift, pRevTable);
	    memcpy((void *)destImage, (void *)pOneLine, widthInWords*4);
	    IP_YIELD(errorFinish)
	}
	    /* if there's a middle line, reverse it too */
	if (destPitchInBytes & 1)
	{
	    ip_reverseLine (pIPState, srcImage, pOneLine,
			    widthInWords, rightShift, pRevTable);
	    memcpy((void *)destImage, (void *)pOneLine, widthInWords*4);
	}

	    /* We're done with the image, free the transfer buffer */
	pOneLine -= frameInBytes;	/* correct the buffer pointer */
	FREE(pOneLine);
    }

    NORMAL_RETURN;

errorFinish:
    if (pOneLine != NULL)
	FREE(pOneLine);
    return ia_aborted;
}




/*

$Log: i_orthrt.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.13  1998/01/13 23:49:09  ddavies
i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.12  1997/02/06 01:34:24  ddavies
ddavies on Wed Feb 5 17:33:53 PST 1997

i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v

Fix gcc uninitialized var warning (not really a problem) on freebsd.

Revision 1.11  1996/09/24 02:39:49  ddavies
i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v

Fix indentation in comments.  Use i_calcImageProperties to get the bytes/line
of destination images that are rotated in place instead of trying to
calculate it here.  This is done for both 90 and 270 rotates in place.

Revision 1.10  1996/03/21 21:13:19  ddavies
ddavies on Thu Mar 21 13:11:49 PST 1996

i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v

Defined _NEED_REVERSEBYTE_, _NEED_REVERSEBYTE2_, _NEED_REVERSEBYTE4_ and
_NEED_REVERSEBYTE8_ before including revbyte.prv so that all tables would
be used.

Revision 1.9  1995/12/13 02:45:10  ejaquith
i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.8  1994/12/14  01:56:35  ddavies
 * ddavies on Tue Dec 13 17:53:54 PST 1994
 *
 * i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v
 *
 * Fix bug in i_fastRot270InPlace.  With some aspect ratios and with small
 * enough transfer buffers, the code was reading a portion of the memory
 * buffer after it had been written with rotated data instead of always
 * reading source bits before they're overwritten.  This caused vertical
 * strips in the output image.  Thanks to Macduff for finding this!
 *
 * Revision 1.7  1994/11/07  22:05:41  ddavies
 * i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.6  1994/05/05  21:48:49  ddavies
 * ddavies on Thu May 5 14:47:43 PDT 1994
 *
 * i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v
 *
 * Fix bug in the version of i_fastRot180 used when rotating in place.
 * The old code didn't touch the middle line of an image when the
 * number of lines in the image was odd.
 *
 * Revision 1.5  1994/04/05  02:54:47  ddavies
 * i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v
 *
 * Fix bug detected by gcc in the 90 and 270 degree rotate in place routines.
 * If the height of the image was 0, we would skip the rotate, but then try
 * to blit the (un)rotated bits into position in the top of the image anyway.
 *
 * Revision 1.4  1994/03/02  03:51:30  ddavies
 * i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v
 *
 * Add IP_YIELD to give up control during 90, 180 and 270 degree rotates.
 *
 * Revision 1.3  1994/02/13  00:50:07  ddavies
 * i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v
 *
 * Cleaned up comments on all routines.  Added wonderfully clever (!)
 * implementations of i_fastRot90InPlace and i_fastRot270InPlace.  These two
 * routines rotate a 1, 2, 4 or 8 bpp image by 90 degrees or 270 degrees
 * respectively.  Added LOTS of comments about how this is done.  Callers
 * of these new routines should read the interfaces carefully.
 * Also modified i_fastRot180 so that it can handle 4 bit and 8 bit images.
 * It was already able to rotate images in place if necessary.
 * Did I say that the InPlace routines return the image in the same buffer
 * in which it was passed in?  Well, you probably guessed that.  This
 * places new restrictions on the way image buffers are allocated (see
 * interfaces).  We have to make sure we have enough memory in the
 * image buffer to hold either orientation.
 *
 * Revision 1.2  1994/02/04  20:15:45  dferrell
 * i_orthrt.c -> /product/ipcore/ipshared/src/RCS/i_orthrt.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/21  21:14:56  ddavies
 * Initial revision
 *
 * Revision 1.13  1993/12/02  19:21:26  ddavies
 * forthrt.c -> /product/ipcore/ipshared/src/RCS/forthrt.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.12  1993/11/05  01:25:12  ddavies
 * forthrt.c -> /product/ipcore/ipshared/src/RCS/forthrt.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


