/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_packc.c: High level code used in PackBits compression.
 *
 *   	Int32	i_getSizeOfPackBitsCompressionInfo()
 *			Returns size of opaque data structure used in
 *			PackBits compression.
 *
 *	Int32	i_initializePackBitsCompressionInfo()
 *			This routine initializes the opaque data structure
 *			used in PackBits compression.
 *
 *	Int32	i_compressRowsPackBits()
 *			Compress specified number of rows using packbits
 *			decompression.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrscal.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#define cByteMask	0xFF
#define	cDummyMarker	0	/* use to mark dummy run beyond end of
				 * valid runs */

IP_RCSINFO(RCSInfo, "$RCSfile: i_packc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* This is the BIG COMMENT that tells the overall story of our approach
 * to packbits compression.  There's a similar comment about decompression
 * in i_packd.c.
 */

/* First, here is a description of the PackBits format from section 9 of
 * the TIFF 6.0 specification:
 *
 * -----------------------------------------------------------------
 *
 * In choosing a simple byte-oriented run-length compression scheme, we
 * arbitrarily chose the Apple Macintosh PackBits scheme.  It has a good
 * worst case behavior (at most 1 extra byte for every 128 input bytes).
 * For Macintosh users, the toolbox utilities PackBits and UnPackBits will
 * do the work for you, but it is easy to implement your own routines.
 *
 * A pseudo-code fragment to unpack might look like this:
 *
 *    Loop until you get the number of unpacked bytes you are expecting:
 *
 *	Read the next source byte into n.
 *
 *	If n is between 0 and 127 inclusive, copy the next n+1 bytes literally.
 *
 *	Else if n is between -127 and -1 inclusive, copy the next byte -n+1
 *	times.
 *
 *	Else if n == -128, noop.
 *
 *   Endloop
 *
 * In the inverse routine, it is best to encode a 2-byte repeat run as
 * a replicate run except when preceded and followed by a literal run.
 * In that case, it is best to merge the three runs into one literal run.
 * Always encode 3-byte repeats as replicate runs.
 *
 * That is the essence of the algorithm.  Here are some additional rules:
 *
 * * Pack each row separately.  Do not compress across row boundaries.
 *
 * * The number of uncompressed bytes per row is defined to be
 *   (ImageWidth + 7)/8.  If the uncompressed bitmap is required to have
 *   an even number of bytes per row, decompress into word-aligned buffers.
 *
 * * If a run is larger than 128 bytes, encode the remainder of the run as
 *   one or more additional replicate runs.
 *
 * -------------------------------------------------------------------------
 *
 * ALMOST!! There's one more hint that's not in the TIFF documentation.
 *
 * -------------------------------------------------------------------------
 *
 * If the length of a repeat run is greater than 128 and equals 1 MOD 128
 * (129, 257, etc), then it should be shortened by 1 byte before being
 * encoded.  The worst thing that can happen is that the last byte will
 * be encoded as a 1-byte literal run.  Note that the code for a 1 byte
 * literal is identical to the code for a 1-byte replica (count of 0
 * followed by byte).  However, if a literal run follows the repeat
 * run, the leftover repeat byte will be absorbed into the literal run,
 * saving a byte.  Similarly, if the long repeat run is followed by a
 * 2-byte repeat run which is followed in turn by a literal run, the
 * literal run will correctly absorb the leftover repeat byte and the
 * 2-byte repeat run.  This also saves a byte.
 *
 * -------------------------------------------------------------------------
 *
 * Simple enough.  I looked at the XIS and Leffler implementations.  They
 * compress and decompress in the simple ways.  The compression code uses
 * a state machine to remember whether it's compressing runs of bytes or
 * passing through literal strings (hereinafter called "literals").
 * The decompression code can just use memcpy and memset.
 *
 * But hey, why not live a little?  While the only thing you can really
 * worry about in the decompression code is writing words instead of bytes,
 * the compression routine is fertile ground for foolishness.  Before
 * we get into method, let's discuss interfaces.
 *
 * We'd like to make routines that can be used by the DAE, SmartPaper,
 * PerfectScan and the PerfectScan Basic Viewer.  The DAE says we're not
 * allowed to have pointers, so any extra arrays or tables we need should
 * be attached to one pointer-free object.  PerfectScan already has a TIFF
 * package, so just wants to plug in a new body of compression code.
 * SmartPaper wants to use the Leffler TIFF code found in IPWrap.  We'd like
 * to use the Leffler code as a testbed anyway.
 *
 * Here's an interface that fits into all worlds:
 *   1.	The client asks for the size of the opaque data object that will
 *	retain state and hold tables and buffers between scan lines.
 *	This is done using i_getSizeOfPackBitsCompressionInfo.  There
 *	is one type of object for compression, another for decompression.
 *   2.	The client allocates a block of this size.
 *   3.	The client calls i_initializePackBitsCompressionInfo.  This fills
 *	in the info that you wouldn't want to recompute on every
 *	compression or decompression call.
 *   4.	The client calls i_compressRowsPackBits.  Image bytes are
 *	compressed in units of scan lines.  The routine returns when either the
 *	desired number of lines has been processed or it has filled
 *	the compression buffer.  Clients can restart in either case.
 *   5.	When it's done with a scan line, the compression code
 *	will normally keep the last fragment of the compressed data
 *	inside the info block instead of writing it to memory.  This
 *	is a reasonably efficient way to concatenate the compressed
 *	data for several lines together.  However, on the last line,
 *	the client needs some way to get the last fragment of compressed
 *	data.  When compressing the last line, the client should call
 *	the compression routine with the flushPipeline argument set to true.
 *	This causes the last portion of compressed data to be flushed
 *	to memory. 
 *   6. When complete, the client frees the compression info block.
 *
 * Yeah, a callback routine would be a nice way to get a new the compression
 * buffer instead of returning and restarting.  Unfortunately, callback
 * routines in the DAE world are pretty much out of touch with important
 * things like I/O.  They'd have to knock on the interpreter's back door
 * to ask for access.  This is very hard.  So, we just return to the
 * client when necessary and let the client sort it all out.
 *
 * There's still some difficulty with byte order.  We will declare that
 * all uncompressed data is in PIXR order.  In fact, it's almost certainly
 * in PIXRs.  We'll try to avoid copying data when possible.  However,
 * the PackBits standard says that the compressed bytestream should be
 * read one byte at a time in the order that's natural to the machine.
 * In fact (as I found out to my embarassment), that's the only way
 * to read and write a bytestream.  That's great except that a
 * significant component of a PackBits file is strings of literals
 * taken directly from the image.  On little-endian platforms, these
 * would be in the wrong order if copied directly from PIXRs.
 *
 * We have three choices.  We could byteswap the compressed data.  We
 * could byteswap the uncompressed image data.  We could write fancy code that
 * did reordered bytes while it was compressing or decompressing.  What to do?
 *
 *
 * An aside.  I also thought about (ok, well, maybe I wrote a lot of code)
 * storing compressed data in PIXR order to avoid byteswaps when using
 * PackBits as an internal cache formt.  The PerfectScan basic viewer
 * uses PackBits in this way.  Storing unaligned data in PIXR order on
 * a little endian machine is nonsense.  Ooops.  Say a PIXR line contains
 * the image bytes:
 *	A B C D   D D E F   G H I J
 *
 * If we tried to compress this without doing byteswapping, we'd get
 *	2 A B C  -2 D 5 E   F G H I  J x x x
 *
 * On a little endian machine, the bytes at the bottoms of the words
 * appear first in the data stream, so the bytestream looks like (where
 * left to right indicates order of the bytes in memory).
 *	C B A 2 E 5 D -2 I H G F x x x J
 *
 * Now say the stream is shifted by one byte relative to word boundaries
 * when it's read back.  This is to be expected since the position of
 * the bytestream relative to word boundaries is random.  Writing the
 * stream starting at a word boundary, we get:
 *	y C B A 2 E 5 D -2 I H G F x x x J z z z
 *
 * If this is read in units of words, which is the whole point of storing
 * the thing in PIXR order, we'd get
 *	A B C y   D 5 E 2  G H I -2  x x x F z z z J
 *
 * This bears little similarity to the original data. Therefore, since
 * the compressed data will be treated as a bytestream, we MUST store it
 * in the natural byte order of the platform we're on.
 *
 *
 * In all cases, we need code that blits bytes between the compressed data
 * stream and the image.  If the uncompressed and compressed data are both
 * in the platform's natural byte order, we could use memcpy.  This
 * is a point in favor of converting the uncompressed data into the
 * processor's natural byte order.
 *
 * But, there are points against that too.  For one, there's a lot of
 * uncompressed data, so byteswapping it is not very efficient.  Second,
 * most runs are fairly short, so making zillions of procedure calls
 * (memset & memcpy) is not very efficient.
 *
 * Byteswapping the compressed data stream doesn't work because we
 * can't guarantee anything about word alignment of the data stream.
 * Say byte A was supposed to be the first byte processed.  If the
 * data stream were going appear on a word boundary, we'd put byte A
 * in the fourth byte of the stream so it would appear as the first
 * byte after byteswapping.  However, if the bytestream were going
 * to appear in memory with it's first byte in byte 1 of a word (relative
 * to the left end of the word), then we're in trouble.  Yeah, we
 * could have put byte A in byte 2 (if we were prescient), but the trash
 * in byte 0 of the word would appear as the third byte of the swapped
 * bytestream, so we'd be forever confused.
 *
 * Our last choice is to build a byte-blit function that deals with
 * image data in PIXR order and the compressed bytestream in natural
 * byte order. If we do the byteswap in the byteblt function, then we have the
 * minimum number of memory transfers (don't write a block, then write
 * it again to byteswap it) and the best cache usage, so that's what we'll do.
 * Since this method doesn't require any preprocessing or postprocssing
 * of either the image or compressed data stream, it's the easiest one
 * to interface with.
 *
 * Here's where we should touch on a little software engineering.  Both the
 * replication and copy steps could use a byteblt routine.  We'd probably
 * want to write this routine in assembly on the PC so we could use the
 * very nice 486 byteswap opcode.  However, calling a routine many times
 * for short runs is a real performance sink, especially on glorious
 * Intel processors.
 *
 * One alternative is to have an assembly routine that decodes an entire
 * lines.  Since there's really not a lot to this routine in addition to
 * the byteblt, we'll do that.  Naturally, there will also be a C version
 * that does byteswaps the hard way if necessary.  I'm not sure if it's
 * worthwhile to do a SPARC assembly version - who would use it?
 *
 * The real trick in compression is in finding the runs of literals
 * and the runs of replicated bytes ("replicas").  Standard implementations
 * of PackBits compression merge the run-finding and output steps by
 * having a state machine operating in a loop.  Some implementations just
 * read one byte and go to the next state for each byte read.  Some try
 * to find a string of identical bytes and use that as input to the next
 * state.  This is better, but fooey.  We're still contemplating each
 * byte individually.
 *
 * Much of the success of ipshared routines has come from processing
 * multiple bytes concurrently.  With this is mind, what can we do?
 * Well, say we start by reading a word, shifting it left by 8 and XORing
 * with the original.  Ok, we also read the word to the left, shift it
 * right by 24 and XOR it with the original too.  The result has a 1
 * wherever a bit in one byte doesn't match the corresponding bit in
 * the next byte.  This is a nice start.  How do we collapse this down
 * to one difference bit per byte?  The difference bit equals 0 when
 * a byte equals the next byte and 1 when it doesn't.
 *
 * We start by saving the MS difference bit in each byte in another
 * word.  We clear this bit in the original difference.  We add 0x7F
 * to each byte.  This propagates any differences into the MS bit of each
 * byte.  We OR in the saved MS difference bits.  This leaves us with a
 * word where the MS bit of each byte is 0 if the byte is the same as
 * the next byte and 1 if not.
 *
 * We then mask all the other bits off.  Cool, now we have four bits
 * that tell us what we want to know instead of 32 bits.  They are in bits
 * 7, 15, 23 and 31 of the word.  But, this would be even more useful
 * if we could pack 32 consecutive difference bits (one for each of 32
 * bytes) into a word instead of just having 4.  So, just as is done
 * in the binarization code (converting 8- 4- or 2-bit images to binary
 * images), we repeat the process described above 7 more times,
 * accumulating the results in a word.  The difference bits
 * are then all in one place, but are out of order.  The difference bits
 * for consecutive bytes are separated by 8 bits instead of being next
 * to each other.  We solve this problem using the same un-scambling
 * tables used in binarization, resulting in a word containing 32 consecutive
 * difference bits.
 *
 * Is this really faster that looking at each byte individually?  Well,
 * I think so.  On a PC, the individual byte inspection requires
 * a read, a compare, a branch, then increment the address, another read,
 * another compare and another conditional branch.  Finally,
 * there's the branch at the bottom of the loop.
 * On a 486, that's about 1+1+1+1+1+2+1+3 = 11 cycles/byte not counting
 * the time taken to do anything interesting once you know that two bytes
 * are the same or different (assume we're accumulating a literal string
 * or a run of sequential bytes).  The glorious new algorithm described
 * above is 2 reads, a move, 2 shifts, 2 XORs, move, AND, ADD, OR, AND,
 * shift = 2+1+4+2+1+1+1+1+1+2 = 16 cycles/4 bytes = 4 cycles/byte.
 * There is more rearrangement after we accumulate 32 difference bits,
 * but we get to amortize that over 32 bytes, so it's tiny.  Fine so far,
 * but isn't there more to do?
 *
 * Yes, indeed.  The PackBits spec says that the sequence [literal run :
 * replica run of length 2 : literal run] should be combined into one
 * long literal run.  I had a glorious morphology-based scheme for
 * identifying isolated replica runs of length 2.  The problem was all
 * of the double shifting required to execute the morphology.  This is
 * all slow and messy.  This situation doesn't come up very often anyway,
 * so a solution that involves moderately complex processing of all the
 * difference bits seems inefficient.  Instead, we'll look at the runs
 * described below.
 *
 * We use i_findAllRunsInLineArray to produce an array containing
 * the coordinates of the run endpoints.  As before, this takes some time,
 * but we can skip over long runs very quickly, so it's pretty efficient.
 * PackBits compressed images generally contain shortish strings of literals
 * and fairly long strings of white stuff, so being able to skip over
 * long difference strings easily is a good trick.
 *
 * We then process the run endpoints.  Each run of 1's actually represents
 * the start of a real run of literals.  We look for the end by advancing
 * to the next run so long as a) both the current and next runs have length
 * greater than 1, and b) there is only a single 0 between them.
 * This allows us to identify consolidated runs of literals.
 *
 * Each run of 0's causes us to copy the -(length of the run + 1
 * (because the 1 at the end of the run is part of this run) - 1
 * (because of PackBits encoding rules)) into the data stream followed
 * by the byte at the beginning of the run.  If the run is longer than
 * 128 bytes, we break it into runs of length 128 plus a remainder run.
 * Each consolidated literal run (at least two adjacent 1's) causes us to
 * copy the length of the run - 1 (because the literal string is one shorter
 * than the number of 1's in a row) - 1 (because of the PackBits encoding
 * rules) to the output stream followed by the string of input bytes
 * starting at the position signified by the second 1 in the run.  We
 * stop when we've handled all the runs in a scan line.
 *
 *
 * So much for the fancy compression scheme.  Time for a little more
 * software engineering.  Clients can't really predict the amount of
 * compressed data required to represent a given slice of an image.
 * What happens when the compression buffer is full or empty?
 * Originally, we were going to let the client supply a callback procedure.
 * We would call it whenever the compressed buffer needed attention.
 * Unfortunately, this doesn't work well for the DAE.  The callback
 * procedure would be down very low.  It would have a very hard time
 * knocking on the back door of the PostScript interpreter asking for
 * more memory or some I/O to be done.
 *
 * Similarly, a client may wish to stop compressing or decompressing
 * at a non-strip or non-tile boundary with the expectation of starting
 * up again later.  This makes sense when the strips or tiles are very
 * large and there's not much memory on the platform.
 *
 * Our new scheme is to simply return if we either run out of compressed
 * data (decompression) or room for compressed data (compression) or
 * have processed the specified number of scan lines.
 * The compression and decompression routines return two flags.  One
 * says the routine stopped because the compressed data is finished.
 * The other says the routine stopped because the image data is finished.
 * Occasionally, both may be set.  This is common at the bottom of
 * an image.  Additionally, the compression routine returns the number
 * of image bytes written to the compressed data buffer.  If the
 * flushPipeline argument was set on the compression call, all of the
 * compressed data calculated has been written to memory.  If flushPipeline
 * was not set, some compressed data may remain in the internal state.
 *
 * Whenever a routine stops, it retains enough state to restart.
 * If it did not run out of image data, the client may move the current
 * image block around as a unit, but should not attempt to guess which
 * part is no longer needed and remove it.
 *
 * We treat the compressed data a little differently.  The client can
 * know exactly how much memory is required for uncompressed image lines.
 * The amount of compressed data needed to represent an image line is
 * usually unknown, especially during compression.  In spite of this ignorance,
 * we would like to present an interface that allows clients
 * to use one buffer/strip, or a buffer for a sub-strip, or double
 * buffers for alternate sub-strips or a ring buffer or whatever.
 *
 * To acheive this flexibility, the compression routine always
 * returns the number of bytes of compressed data written.  The client
 * may safely dispose of these bytes.  It is the client's job to advance
 * the pointer to the compressed data buffer and recompute the amount of
 * buffer space remaining before each call.
 *
 * In bit-oriented compression protocols, scan lines may not end
 * on byte boundaries, so the number of bytes written only roughly
 * corresponds to the number of bits required to represent the
 * scan lines.  In PackBits compression (a byte-oriented protocol),
 * the number of bytes written by a call to i_compressRowsPackBits should
 * exactly equal the number of bytes required to represent these lines.
 *
 * Clients supply two arguments to the compression routines.
 * One specifies whether the routine should continue with existing
 * image data.  The other specifies whether the the routine
 * should continue with the current compression data *stream*.  This
 * is set if we ran out of compression data last time and a new
 * compression data buffer is passed in to hold a seamless continuation
 * of the old compressed data stream.  For example, if we filled the
 * compressed data buffer after writing a bytecount but before writing
 * any data, this flag says to treat the first byte of the new buffer
 * as space for the first data byte of the run specified by that last
 * bytecount.  Otherwise, any saved state is discarded.
 *
 * If the image data is new, the compression routine discards its
 * saved image state.  The same goes for new compression
 * data and saved compression state.  If told to continue on with
 * existing image data or an existing compression stream, and there
 * is no corresponding saved state, an error is returned.
 *
 * When they have exhausted the compression info, the compression
 * routines also return a parameter that vaguely specifies the amount
 * of the original request that has been processed.  Clients are free
 * to ignore this parameter.  It is intended to allow clients guess
 * how much compressed data is needed to complete the slice.
 * Before restarting, the client should supply more room for compressed
 * data.
 *
 * Clients supply one more argument that tells the compression
 * routines what to do with the pipeline of compressed data when the
 * specified number of scan lines has been processed.  If flushPipeline
 * is set on compression, the valid bytes in the final word of compressed data
 * are flushed to the compressed data buffer before returning.  The total
 * number of valid bytes is also returned.  If flushPipeline is not set,
 * the final partial word is held back so that further compressed data
 * may be appended to it without the padding.  The count of compressed
 * bytes does not include this partial word in this case.
 *
 * There's one more hack.  When using PackBits as an internal cache
 * format, we'd like to have random access to the compressed image lines.
 * The compression routine will fill an array allocated by the client
 * with the offsets from the beginning of the compressed data buffer
 * to the beginning of each scan line.  If the array pointer passed in by
 * the client is NULL, this is not done.
 *
 * Time for more detail on the internal workings of compression.
 * The image is compressed one line at a time.  The basic process for
 * each line is:
 *   1. Compute the runs of replicas and literals as explained above.
 *	After this, we still need the uncompressed image lines as
 *	a source of replica and literal values, so we can't just
 *	ignore that line.  Note that we can't run out of resources
 *	in this step, so we don't need to retain restart state here.
 *   2. Process each run of replicas and literals.  Since this is the
 *	stage at which we are filling the output compression buffer,
 *	this is the stage at which we could either run out of room
 *	in the compressed buffer or finish with the specified number
 *	of image lines.  Therefore we need to be able to restart in 
 *	the middle of this stage.
 *
 *	Each pass through the main loop puts out either a run of
 *	literals followed by a run of replicas or just a run of
 *	replicas.  A run of literals is indicated by a run of
 *	more than one 1	in the difference bits.  A run of replicas
 *	is indicated by a runs of 0's in the difference bits PLUS 
 *	THE NEXT 1.  Therefore, each block of literals or replicas 
 *	starts on the SECOND 1 of a run of 1's in the difference bits.
 *	In each pass through the main loop, we look at a new run 
 *	of difference bits.  If there is more than one 1 in it, 
 *	we start a run of literals that extends through the last 1.  
 *	We then tack on a run of replicas that includes the first 1 
 *	of the next run.  If that next run is off the edge of the 
 *	image, we stop at the edge.
 *
 *	Oh, yeah, one more point.  Runs of longer than 128 must be
 *	broken into subruns.  We have to remember both that we're
 *	doing a run and that we may be doing a subrun within a run.
 *
 *	How do we start the left end of the scan line?  We'd like to
 *	not have much special code for this case.  We allocate an
 *	extra run there.  If the scan line starts with a bunch of
 *	literals, we disregard the extra run and set the left coord
 *	of the first literal run to -1.  This allows us to start the
 *	literals at (runLeft + 1) or 0.  If the scan line starts with
 *	a bunch of replicas, we use the extra run.  It's coords are
 *	set to (-1, -1).  It signifies the end of the frame so the
 *	replicas can start at (runRight + 1) or 0.
 *
 * Whoops, another point.  The compression buffer may not start or stop
 * on a word boundary.  This is certainly true of the DAE world where
 * there are no alignment constraints on strings and I/O can only be done
 * with strings.  It's also true of most compressed files - the strips
 * are simply jammed together on byte boundaries.
 *
 * It would be pleasant to do writes in units of words, not bytes.
 * However, to do this, we have to compose words that have bytes in the
 * natural byte order of the platform.  That means not byteswapping them
 * on the big-endian platforms and byteswapping (or some reordering) on
 * the little endian platforms. That really means two routines.  Yuck.
 * Our C code is supposed to be nice and simple and run on all platforms
 * with only minor platform variances.
 *
 * Well, how many bytes are there in a PackBits file?  At 300 dpi, even
 * fairly complex files have less than 400K bytes.  On a SPARC IPX (about
 * 0.25 microseconds/write), that's only 0.1 seconds for all the bytes.
 * On a 486 PC with faster writes, its faster.  So, what the heck, read
 * the image a word at a time and write it to the compressed data stream
 * a byte at a time.  This is a low overhead approach which works well
 * with lots of short literal strings.  That's what we usually see in
 * these files.
 *
 * We'll also write x86 assembly code to do this job.  It should be faster
 * in that it can avoid several of the shifts and extra writes.
 *
 * The code that computes runs of difference bits will be written in
 * C and common to all platforms.  The code that converts runs of difference
 * bits into a compressed file is platform-dependent.  This is the code
 * that does byteswaps if necessary.  We'll write x86 assembly code to
 * do this job.  It should be faster in that it can avoid several of
 * the shifts and extra writes by using the 486's byteswap opcode and
 * misaligned writes.
 *
 *
 * We need a clean way of stopping when the compression buffer is
 * full.  We do the check for the buffer being full immediately
 * before writing another byte into the buffer.  
 * All states record what action we would have taken had the check
 * completed successfully.  There is one extra state used to start things off.
 *
 * What's in the compression state block?  Also see shrcmprs.prv for the
 * real declaration.
 *   1. Flag saying whether saved image data is valid.
 *   2. Number of current uncompressed line (offset from top line).
 *   3. Array of difference bits (bit n is set if byte n != byte n+1
 *	in the current line).
 *   4. Array of runs of difference bits (start and stop of each array
 *	of 1's in the difference bit array).  Last run starts at
 *	cVeryLargeNum, which is off the image.  This array is allocated
 *	once based on the maximum number of runs that could fit in an
 *	image of the specified width.
 *   5. Index of current run being processed or next run to be processed
 *	if we are between runs.
 *   6.	Flag saying whether saved compressed state is valid.
 *   7.	Current compression state
 *	a. Start new line (used to kick off line compression routine.  Other
 *	   states used to restart line compression routine).
 *	b. Write literal subrun byte count
 *	c. Complete literal subrun.
 *	d. Write replica subrun byte count
 *	e. Complete replica subrun.
 *	f. Complete pipeline flush.
 *   8. Count of subrun bytes still to be processed in states 7 b, c, d and e.
 *   9.	Count of all run bytes still to be processed after completing
 *	subruns in states 7 b, c, d, and e.
 *  10. Index of next element of the offset array used to record
 *	offsets of left ends of scan lines in the compressed data block.
 *  11.	Pointer to the routine used to compress on row.
 *
 * The caller supplies the address of the start of the block of uncompressed
 * data, the uncompressed data's width and height in pixels and bytes/line,
 * a pointer to a block for compressed data, the length of that block
 * (must be 4-byte aligned and length must be a multiple of 4 bytes),
 * and the continueImageBlock, continueCompDataStream and flushPipeline args
 * discussed above.  The caller also supplies pointers to words used to
 * return the exhaustedImageData, exhaustedCompData, bytesWritten and
 * percentDone args.  The bytesWritten arg returns the number of compressed
 * bytes written so far.  If flushPipeline is not set, some of the compressed
 * bytes may have been computed, but not yet written.
 *
 * What are the states in which the compression routine can return?
 *
 *   1.   Return code != ia_successful
 *	Something catastrophic happened.
 *
 *   2.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cFalse
 *	We have processed all the image data requested.  The compression
 *	buffer still has room.  If flushPipeline was true, all the compressed
 *	data has been flushed to the output.  Otherwise, some may still be in
 *	the compression pipeline so it can be concatenated with the next
 *	set of compressed data generated.  In either case, the number of
 *	bytes actually written to the buffer (not including any bytes
 *	still in the pipeline) are returned.
 *	When the compression routine is called next, the client must
 *	pass in a pointer to the byte at which the next compressed byte
 *	should be placed.  This may be just beyond the end of the
 *	current compressed data, at the beginning of the same buffer, 
 *	or in another buffer entirely.  The size of the compressed data
 *	buffer must also be (re)calculated by the client.
 *
 *   3.	  __error__ == ia_successful
 *	  exhaustedImageData == cTrue
 *	  exhaustedCompData == cTrue
 *	Same as case 2 above except we just happened to write the last
 *	byte of the compressed data buffer at the same time the compressed
 *	data came to an end.  As before, if flushPipeline was false, there
 *	may be more compressed data in the pipeline.
 *
 *   4.	  __error__ == ia_successful
 *	  exhaustedImageData = cFalse
 *	  exhaustedCompData = cTrue
 *	We ran out of room for compressed data before writing the last
 *	byte of compressed data requested.  Some of the data to be written
 *	is in the pipeline, the rest is specified by the runs of difference
 *	bits.
 *
 *   5.	  __error__ == ia_successful
 *	  exhaustedImageData == cFalse
 *	  exhaustedCompData == cFalse
 *	Shouldn't happen - why stop if we have both more work to do
 *	(exhaustedImageData is false) and more room to put the compressed
 *	data (exhaustedCompData is false) and nothing went wrong
 *	(__error__ is ia_successful).
 */

/* PUBLIC */
/**************************************************************************
 *   i_getSizeOfPackBitsCompressionInfo()
 *
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform PackBits
 *	compression.
 *
 *	Input:
 *		w:	width of image including frame
 *		d:	depth of image
 *		pSize:	pointer to Int32 that will contain the size.
 *			of the buffer required to hold the symbol
 *			information in bytes.
 *
 *	Returns:
 *		ia_successful		 it worked
 *		ia_depthNotSupported	 depth was illegal.
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfPackBitsCompressionInfo (
    IPState    *pIPState,
    Int32	w,
    Int32	d,
    Int32     *pSize)
{
    PROLOGUE;
    Int32	bytesPerLine, diffBytesPerLine;

	/* if the depth isn't supported, return an error */
    if ((d != 1) && (d != 2) && (d != 4) && (d != 8))
	return ia_depthNotSupported;

	/* remove the frame */
    w -= 2 * cFrameBits;

	/* calculate number of bytes of uncompressed data in a
	 * scan line.
	 */
    bytesPerLine = ((w*d + 31)/32) * 4;

	/* Calculate the number of bytes of difference bits.
	 */
    diffBytesPerLine = ((bytesPerLine + 31)/32) * 4;

	/* Return size of PackBits compression structure
	 * plus an array big enough to hold all the difference bits
	 * (1 bit/byte) plus an array big enough to hold the endpoints
	 * of the runs of difference bits (a Run for every 2 bits
	 * in the difference bits).
	 */
    *pSize = sizeof(PackBitsCompressionInfo) +
	     diffBytesPerLine +	/* holds array of diff bits */
	     ((bytesPerLine/2) + 1) * sizeof(Run32); /* holds max number of Run
						      * structs plus one
						      * dummy struct at the
						      * beginning of the
						      * scan line. */

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_initializePackBitsCompressionInfo()
 *
 *	This procedure fills a block of PackBits compression info
 *	with necessary information.
 *
 *	Input:
 *		pPBCompInfoIn:		pointer to PackBits compression info
 *		w:			width of uncompressed image including
 *					frame.
 *		d:			depth of uncompressed image.
 *		compInfoBufSize:	the number of bytes in the
 *					compression info.
 *
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		pPBCompInfoIn was NULL, buffer was
 *					wrong size
 *
 **************************************************************************/
Int32  CDECL 
i_initializePackBitsCompressionInfo (
    IPState    *pIPState,
    void       *pPBCompInfoIn,
    Int32	w,
    Int32	d,
    Int32	compInfoBufSize)
{
    PROLOGUE;
    PackBitsCompressionInfo	*pPBCompInfo;
    Int32			 realCompInfoBufSize;
    Int32			 bytesPerLine, diffBytesPerLine;
    Int32			 run;
    UInt32			*pByte0Tab, *pByte1Tab, *pByte2Tab, *pByte3Tab;
    Run32			*pRun;


    pPBCompInfo = (PackBitsCompressionInfo *)pPBCompInfoIn;

	/* Get the size the block should be */
    __error__ = i_getSizeOfPackBitsCompressionInfo (pIPState, w, d,
						    &realCompInfoBufSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* See if any of the parameters were illegal */
    if ((pPBCompInfo == NULL) ||
	(realCompInfoBufSize > compInfoBufSize))
	return ia_invalidParm;

    if ((d != 1) && (d != 2) && (d != 4) && (d != 8))
	return ia_depthNotSupported;

	/* Initialize all the rest of the default stuff */
    pPBCompInfo->lineNumber = 0;
    pPBCompInfo->runStart = 0;
    pPBCompInfo->runEnd = 0;
    pPBCompInfo->compState = cPBCompNewLine;
    pPBCompInfo->subRunByteCount = 0;
    pPBCompInfo->runByteCount = 0;
    pPBCompInfo->lineArrayIndex = 0;

	/* Fill in the tables used to unscramble the arrays of difference
	 * bits.  These are also used when converting bytes to bits,
	 * which is what we do with difference bits.  Note that this
	 * routine expects a pointers to pointers to UInt32's as args,
	 * so we have to make them up.
	 */
    pByte0Tab = &(pPBCompInfo->byte0Table[0]);
    pByte1Tab = &(pPBCompInfo->byte1Table[0]);
    pByte2Tab = &(pPBCompInfo->byte2Table[0]);
    pByte3Tab = &(pPBCompInfo->byte3Table[0]);
    __error__ = i_makeBinarizationTables (pIPState, &pByte0Tab, &pByte1Tab,
				        &pByte2Tab, &pByte3Tab);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* Get the addresses of the routines used to compress lines.
	 */
    ip_getPackBitsCompProc (pIPState, &(pPBCompInfo->compRowProc));

	/* record offset to the beginning of the array of difference bits */
    pPBCompInfo->differenceBitArrayOffset = sizeof(PackBitsCompressionInfo);

	/* calculate the length of the difference bit array in bytes.  To
	 * do this, we first calculate the bytes in the image.
	 */
    bytesPerLine = (((w - 2*cFrameBits)*d + 31)/32) * 4;

	/* Calculate the number of bytes of difference bits */
    diffBytesPerLine = ((bytesPerLine + 31)/32) * 4;

	/* Point to beginning of endpoint array.  This is after
	 * the array of difference bits, which holds enough difference
	 * bits for one scan line.
	 */
    pPBCompInfo->runArrayOffset = pPBCompInfo->differenceBitArrayOffset +
				  diffBytesPerLine;

	/* Initialize both arrays.  The difference bit array is set to 0,
	 * meaning a line entirely full of literals.  The Run32 arrays is
	 * set to all cVeryLargeNums, meaning a bunch of runs off the
	 * end of the page.
	 */
    memset((void *)((UInt8 *)pPBCompInfo +
		    pPBCompInfo->differenceBitArrayOffset),
	   0, diffBytesPerLine);

    for (pRun = (Run32 *)((UInt8 *)pPBCompInfo + pPBCompInfo->runArrayOffset),
	    run = 0;
	 run < bytesPerLine/2;
	 run++)
    {
	pRun[run].runLeft = 0;
	pRun[run].runRight = 0;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_compressRowsPackBits()
 *
 *	This procedure compresses images in units of scan lines
 *	using the PackBits algorithm.
 *
 *	Input:
 *		pPBCompInfoIn:		Pointer to PackBits compression info
 *		pImage:			Pointer to the upper left corner
 *					of the frame of the uncompressed
 *					image
 *		w:			Width of uncompressed image including
 *					frame.
 *		startLine:		First image line to be compressed
 *					relative to the top of the frame
 *		numLines:		Number of lines to be compressed
 *		d:			Depth of uncompressed image.
 *		bpl:			bytes/line of uncompressed data.
 *		pCompData:		Pointer to first vacant byte
 *					of buffer into which compressed
 *					data will be placed.
 *		compDataBufSize:	Number of vacant bytes remaining
 *					in the compressed data buffer.
 *		continueImageBlock:	cTrue => we are compressing
 *					from the same slice of image as was
 *					specified in the previous call and the
 *					saved state in pPBCompInfoIn is
 *					valid.
 *					This arg is set to cFalse if either
 *					we are starting with all new data
 *					or the previous call returned because it
 *					ran out of image data.
 *		continueCompDataStream:	cTrue => The saved compression state
 *					in pPBCompInfoIn is valid.  If
 *					the previous call to this routine
 *					filled the compressed data buffer
 *					before the strip, tile or image was
 *					complete, a new buffer should be
 *					supplied.  This arg is set to 
 *					cFalse when the block of compressed
 *					data is unrelated to the block used
 *					in the previous call to this routine
 *					as recorded in pPBCompInfoIn.
 *					SETTING THIS ARG TO cFalse CAUSES
 *					DATA IN THE PIPELINE TO BE DISCARDED!
 *		continueLineOffsets:	We compute offsets into the
 *					compressed data that point to the
 *					beginning of each scan line.  When
 *					this arg is true, next offset should
 *					be computed from the previous one,
 *					i.e. the current block of compressed
 *					data is an extension of the block
 *					produced by the previous call.
 *					When this is cFalse, the array of
 *					offsets is restarted from 0.
 *		flushPipeline:		If this arg is set, this routine
 *					will flush all the compressed data
 *					in the pipeline to the compressed
 *					data buffer when the specified
 *					number of scan lines have been
 *					compressed.  The client should not
 *					set flushPipeline if compressed
 *					data from subsequent scan lines is
 *					to be seamlessly concatenated to
 *					the current compressed data stream.
 *					When flushPipeline is set and the
 *					compression task suceeds in
 *					processing all image lines, all
 *					saved state is marked invalid.
 *					Subsequent calls with either
 *					continueImageBlock or
 *					continueCompDataStream will return
 *					ia_invalidParm.
 *		pExhaustedImageData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because specified
 *					number of lines have been processed.
 *					This may require more than one
 *					call to this routine.
 *		pExhaustedCompData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because it filled
 *					the compressed data buffer.  This
 *					routine may be called again with
 *					a new buffer.
 *		pPercentDone:		Pointer to returned arg.  When
 *					routine returns with *pExhaustedCompData
 *					set, this arg specifies the rough
 *					percentage of the image block that
 *					has been compressed.  It is
 *					intended to be used to estimate the
 *					amount of compressed data needed
 *					to complete the task.  The result
 *					ranges from 0 to 100.
 *		pBytesWritten:		Number of compressed bytes actually
 *					written to the compressed data
 *					block.  If flushPipeline is false,
 *					some bytes required to complete
 *					a scan line may still be in the
 *					pipeline and are not included in
 *					this count.
 *		lineArray:		Array of offsets to beginnings
 *					of compresses scan lines in compressed
 *					data block.  After line n has been
 *					successfully compressed,
 *					lineArray[n+1] = lineArray[n] +
 *					{number of compressed bytes used
 *					to represent line n}.  This prepares
 *					us to compress line n+1.  The client
 *					should allocated a array that's one
 *					line too long to account for the
 *					extra write after the last scan line.
 *
 *	Action:
 *		Compresses a horizontal slice image using the
 *		PackBits method.
 *		There are several completion conditions:
 *
 *	__error__	!= ia_successful
 *	*pExhaustedImageData == don't care
 *	*pExhaustedCompData == don't care
 *		This means something catastrophic happened.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cFalse
 *		This means we compressed the desired number of
 *		image lines without filling the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.  If flushPipeline
 *		was set, all compressed data has been written to
 *		the compressed data buffer and the saved compressed state
 *		is marked invalid.  If flushPipeline was not set,
 *		some of the compressed data may still be in the
 *		pipeline and the saved compressed data state is valid.
 *		In the latter case, this routine may be called again with
 *		continueCompDataStream = cTrue.  In either case, the
 *		client must supply a new value for pCompData and
 *		compDataSize.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cTrue
 *		This means we filled the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is unusual, but not impossible.  If flushPipeline
 *		was set, all compressed data has been written to the
 *		compressed data buffer and the saved state is marked
 *		invalid.  Otherwise, some data may remain in the pipeline.
 *		This return state cannot occur of flushPipeline was set
 *		and the pipeline still contains data (see next state).
 *		If flushPipeline was set, it is not possible to continue
 *		from this state as all saved state is invalid.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cFalse
 *	*pExhaustedCompData == cTrue
 *		The routine filled the compression data buffer before completing
 *		the specified number of lines.  One way this can
 *		happen is if all image data has been processed, but
 *		flushPipeline was set and some data remains in the pipeline.
 *		When this condition does occur, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  
 *		The image may be moved around as a block if necessary, but
 *		the internal composition of the image (and startLine
 *		and numLines) should not be changed.
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *		
 *		The image data is always in PIXR order.  This routine
 *		will byteswap strings of literals in the compressed data
 *		on the way to the image if necessary.  The byte order
 *		of the compressed data was specified in the call to
 *		i_initializePackBitsCompressionInfo that initialized
 *		the opaque compression info pointed to by pPBCompInfoIn.
 *		
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		a pointer was NULL or 
 *					(continueImageBlock && the saved image
 *					info is not valid) or
 *					(continueCompDataStream && the saved
 *					compression info is not valid)
 *		ia_bufferExceeded	ran out of room for compressed data
 *					when flushPipeline was set.
 *
 **************************************************************************/
Int32  CDECL 
i_compressRowsPackBits (
    IPState    *pIPState,
    void       *pPBCompInfoIn,
    UInt8      *pImage,
    Int32	w,
    Int32	startLine,
    Int32	numLines,
    Int32	d,
    Int32	bpl,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Int32	continueLineOffsets,
    Bool	flushPipeline,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesWritten,
    Int32      *lineArray)
{
    PROLOGUE;
    PackBitsCompressionInfo	*pPBCompInfo;
    Int32	 widthInBytes;	/* number of bytes to decode
				 * per line */
    Int32	 rightmostWordOffset;	/* Offset to the rightmost
					 * word in the scanline in units
					 * of bytes. */
    UInt8	*pCurImage;	/* temporary image pointer */
    UInt8	*pImageEnd;	/* points to end of current image
				 * block */
    UInt8	*pCompDataEnd;	/* address 1 beyond the last
				 * byte of compressed data */
    UInt8	*pCurCompData;	/* current compressed data
				 * pointer */
    UInt32	*pDifferenceBits; /* pointer to array of difference
				   * bits */
    Run32	*pRuns;		/* pointer to first run in array
				 * of runs. */
    Run32	*pDummy;	/* returned ptr to element just beyond
				 * last valid run in the array */
    UInt32	 diffBitAccum;	/* word to hold accumulating
				 * difference bits. */
    UInt32	 input;		/* holds input word */
    Int32	 temp;		/* general scratch word */
    Int32	 remainingLines;


	/* check for null pointers */
    if ((pPBCompInfoIn == NULL) || (pCompData == NULL) ||
	(pImage == NULL) || (compDataBufSize == 0) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesWritten == NULL))
	return ia_invalidParm;

	/* Strip off the frame */
    pImage += (bpl << cLogFrameBits) + d * cFrameBytes;
    startLine -= cFrameBits;
    w -= 2*cFrameBits;

	/* load the local copies of the remote pointers */
    pPBCompInfo = (PackBitsCompressionInfo *) pPBCompInfoIn;

	/* widthInBytes is the real number of bytes in a scan line.
	 * byteWidthInWords is widthInBytes rounded up to the next
	 * multiple of 4 so it always points to a word boundary.
	 */
    widthInBytes = (w*d + 7)/8;
    if (widthInBytes == 0)	/* didn't ask for anything, so don't do
				 * anything. */
	NORMAL_RETURN;

	/* get offset to rightmost word in the line.  We calculate
	 * difference bits from right to left.
	 */
    rightmostWordOffset = (widthInBytes - 1) & ~0x3;

	/* If this is a new block of compressed data and there is
	 * an array of offsets, set lineArray[0] to 0
	 * and point to the beginning of the new array we were just given.
	 */
    if ((!continueLineOffsets) && (lineArray != NULL))
    {
	lineArray[0] = 0;
	pPBCompInfo->lineArrayIndex = 0;
    }

	/* Initialize the state from the saved state passed in *pPBCompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	pImage += startLine * bpl;
	pPBCompInfo->lineNumber = 0;
	pPBCompInfo->runStart = 0;
	pPBCompInfo->runEnd = 0;
	pPBCompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pPBCompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pPBCompInfo->lineNumber;
	pImage += (startLine + pPBCompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pPBCompInfo->compState = cPBCompNewLine;
	pPBCompInfo->runByteCount = 0;
	pPBCompInfo->subRunByteCount = 0;
	pPBCompInfo->compInfoIsValid = cTrue;
    }
    else
    if (!pPBCompInfo->compInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;

	/* get a pointer to the second run.  The first is a dummy. */
    pRuns = (Run32 *)((UInt8 *)(pPBCompInfo) +
			        pPBCompInfo->runArrayOffset +
			        sizeof(Run32));

	/* Compress each scan line.
	 * This is done in several steps:
	 *   1. If the compression state does not equal cPBCompNewLine,
	 *	jump to step 4.
	 *   2. We did not quit in the middle of putting out this scan
	 *	line.  Compute the raw difference bits for this scan line.
	 *	Bit n of the array is a 1 if byte n equals byte n+1.
	 *   3. Find all the runs in the arrays of difference bits.
	 *	Jump to step 5.
	 *   4.	We must be operating on a set of runs that have already
	 *	been computed.  Adjust the addresses recorded in the
	 *	runs if the address in the first run does not equal
	 *	pImage - 1 (the value of pImage for which the runs
	 *	were computed)
	 *   5. Call ip_compressRunsPackBits (or equivalent) to put out
	 *	the codes for the runs of the scan line.  This routine
	 *	combines isolated replica runs of length 2 with the adjoining
	 *	literal runs.
	 */
    for (;
	 remainingLines > 0;
	 remainingLines--, pImage += bpl)
    {
	if (pPBCompInfo->compState == cPBCompNewLine)
	{
		/* get a pointer to the rightmost word in the array
		 * of difference bits */
	    pDifferenceBits = (UInt32 *)((UInt8 *)(pPBCompInfo) +
					 pPBCompInfo->differenceBitArrayOffset +
					 (((widthInBytes - 1) / 32) * 4));
    
		/* Compute the array of difference bits from right to left.
		 * Compute the address of the rightmost word relative to the
		 * left end of the scan line.  Also compute the address of
		 * the rightmost word in the scan line whose offset from
		 * the scan line is a multple of 32.  We know diffBitAccum
		 * has either a multiple of 32 bits or the remainder of
		 * (scan line width) / 32 when we're pointing to this input
		 * word..
		 */
	    pCurImage = pImage + rightmostWordOffset;
	    pImageEnd = pImage + (rightmostWordOffset & ~0x1F);
	    diffBitAccum = 0;
    
		/* Compute the difference bits from right to left.
		 * in all cases, difference bit n is set if byte n
		 * does not equal byte n+1
		 */
	    while (pCurImage >= pImage)
	    {
		input = ULONG_ACCESS(pCurImage);	/* get 4 bytes */
		input ^= (input << 8);			/* mark diffs in upper
							 * 3 bytes */
		    /* mark diffs in lower byte */
		input ^= (ULONG_ACCESS(&pCurImage[4]) >> 24);
		    /* Accumulate all differences to MS bits of each byte */
		temp = input & 0x80808080;		/* save MS bits */
		input &= 0x7F7F7F7F;			/* delete MS bits
							 * in input */
		input += 0x7F7F7F7F;			/* accumulate diffs
							 * in lower 7 bits into
							 * MS bits. */
		input |= temp;				/* insert diffs in
							 * MS bits */
		diffBitAccum >>= 1;			/* make room fo new
							 * diff bits */
		diffBitAccum |= input & 0x80808080;	/* save diff bits */
    
		    /* If diffBitAccum is now full, unscramble the bits
		     * so they make sense */
		if (pCurImage == pImageEnd)
		{
			/* If diffBitAccum is all 0 (white areas), just
			 * store it as is */
		    if (diffBitAccum == 0)
			*pDifferenceBits-- = 0;
		    else
		    {
			*pDifferenceBits-- =
			    pPBCompInfo->byte3Table[diffBitAccum  & cByteMask] |
			    pPBCompInfo->byte2Table[(diffBitAccum >> 8) &
								    cByteMask] |
			    pPBCompInfo->byte1Table[(diffBitAccum >> 16) &
								    cByteMask] |
			    pPBCompInfo->byte0Table[diffBitAccum >> 24];
			diffBitAccum = 0;
		    }
		    pImageEnd -= 32;			/* point to next bdry */
		}
		pCurImage -= 4;
	    }
    
		/* At this point, pDifferenceBits points one word to the
		 * left of the real difference bit array.  We really
		 * want to have it pointing to the first difference
		 * byte, so move it back.
		 */
	    pDifferenceBits++;

		/* Fill dummy first run with [-1,-1].  This is a dummy
		 * used to help compute the first run of replicas,
		 * if any.  pRuns already points to the second run,
		 * so we have to step back to point to the first run.
		 */
	    pRuns[-1].runLeft = (Int32)pImage - 1;
	    pRuns[-1].runRight = (Int32)pImage - 1;

		/* Find the run boundaries and put them into the array
		 * of runs.  Note that we add the address of the beginning
		 * of the line to each run coordinate so that the numbers
		 * recorded in the run structures are actually pointers to
		 * the bytes.
		 * Also note that the coords returned by this routine are
		 * relative to the starting location of pDifferenceBits.
		 * Since we're starting with a pointer to the first word
		 * of the difference bit word, we get the correct coords.
		 */
	    __error__ =
	        i_findAllRunsInLineArray (pIPState,
					 (UInt8 *)pDifferenceBits,
							  /* inside "frame" */
				          widthInBytes,
				          0,		  /* left limit */
				          pRuns,	  /* dest array */
				          widthInBytes/2, /* # of Run32s */
				          (UInt32)pImage, /* add this to each
							   * run coord */
					  &pDummy);	  /* no valid coord
							   * will equal NULL */

	    if (__error__ != ia_successful)
		NORMAL_RETURN;

	        /* Mark the dummy element at the end.  It's left coord
		 * should be the address of the rightmost byte in the line
		 * since that's the proper end of any replica run.  The
		 * right coord will be set to cDummyMarker, since that's
		 * not a legal address.
		 */
	    pDummy->runLeft = (Int32)pImage + widthInBytes - 1;
	    pDummy->runRight = cDummyMarker;

		/* if there is a single 0 byte at the right end of the
		 * scan line, include it in the rightmost literal.  It
		 * shouldn't look like a replica since we shouldn't
		 * include the frame in the replica calculations.
		 */
	    if (pDummy[-1].runRight == (Int32)pImage + widthInBytes - 2)
		pDummy[-1].runRight++;

		/* Set the run offsets in the compression info to 0
		 * so that we start interpreting the very first run
		 * and not some run left over from some previous
		 * stoppage.
		 */
	    pPBCompInfo->runStart = 0;
	    pPBCompInfo->runEnd = 0;
	}
	else	/* we already had a set of runs.  Do we need to adjust the
		 * addresses in the runs?  This happens if we quit in
		 * the middle of a line and the client moves the image
		 * block around before calling us again.
		 */
	{
	    if (pRuns[-1].runRight != (Int32)pImage - 1)
	    {
		    /* get amount by which all runs should be
		     * adjusted.
		     */
		temp = (Int32)pImage - (pRuns[-1].runLeft + 1);

		    /* fix all the runs */
		for (pRuns--;
		     pRuns->runLeft != 0;
		     pRuns++)
		{
		    pRuns->runLeft += temp;
		    pRuns->runRight += temp;
		}
		    /* Restore pRuns. */
		pRuns = (Run32 *)((UInt8 *)(pPBCompInfo) +
					  pPBCompInfo->runArrayOffset +
					  sizeof(Run32));
	    }
	}

	    /* Ask the compression code to translate runs into the
	     * compressed data stream.  Since the addresses of the
	     * runs are encoded directly in the runs, we don't need
	     * pointers to them.  The second arg is the value used to
	     * mark the right coord of the dummy element at the end
	     * of the array.
	     */
	ERR_PASS( pPBCompInfo->compRowProc(pIPState,
					   pPBCompInfo,
					   cDummyMarker,
					   &pCurCompData,
					   pCompDataEnd));

	    /* If we didn't get to the end of a scan line, or
	     * filled the compressed data buffer and are not out of scan
	     * lines, ask for more room for compressed data.
	     * We need to check to see if we stopped in mid-scanline
	     * to detect the case of running out of data in the middle
	     * of the last scan line of the strip.
	     */
	if ((pPBCompInfo->compState != cPBCompNewLine) ||
	    ((pCurCompData == pCompDataEnd) && (remainingLines > 1)))
	{
	    *pExhaustedImageData = cFalse;
	    *pExhaustedCompData = cTrue;
	    *pPercentDone = (100 *
			     ((pPBCompInfo->lineNumber) * widthInBytes)) /
			    (numLines * widthInBytes);
	    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

		/* if we happened to stop on an image scan line boundary,
		 * increment the saved line number so we don't do it again.
		 * Also update the offset index.
		 */
	    if (pPBCompInfo->compState == cPBCompNewLine)
	    {
		pPBCompInfo->lineNumber++;
		if (lineArray != NULL)
		{
		    temp = pPBCompInfo->lineArrayIndex;
		    lineArray[temp + 1] = lineArray[temp] + *pBytesWritten;
		    pPBCompInfo->lineArrayIndex = temp + 1;
		}
	    }
		/* Set __error__ to ia_successful.  This is the convention
		 * at this point.  There wasn't a real error, we just
		 * filled up the output buffer.  The caller detects this
		 * by looking at *pExhaustedCompData.
		 */
	    __error__ = ia_successful;
	    NORMAL_RETURN;
	}
	    /* increment the current line number in case we need
	     * to stop in the next line.  Also update the index of
	     * offsets to starting locations of scan lines in the
	     * compressed data block.
	     */
	pPBCompInfo->lineNumber++;
	if (lineArray != NULL)
	{
	    temp = pPBCompInfo->lineArrayIndex;
	    lineArray[temp + 1] = lineArray[temp] + *pBytesWritten;
	    pPBCompInfo->lineArrayIndex = temp + 1;
	}
    }

	/* We've done all the scan lines.  So, just
	 * say things were successful and return.
	 */
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;
    *pBytesWritten = (UInt32)pCurCompData - (UInt32)pCompData;

	/* if flushPipeline is true, declare all the saved state to be
	 * invalid.  This helps the client catch bugs in which continuation
	 * was attempted, but the saved state is invalid.
	 */
    if (flushPipeline)
    {
	pPBCompInfo->imageInfoIsValid = cFalse;
	pPBCompInfo->compInfoIsValid = cFalse;
    }

    NORMAL_RETURN;

}



/*

 $Log: i_packc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.5  1998/01/13 23:49:13  ddavies
 i_packc.c -> /product/ipcore/ipshared/src/RCS/i_packc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1995/12/13 02:45:14  ejaquith
 i_packc.c -> /product/ipcore/ipshared/src/RCS/i_packc.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/07/10  23:42:32  ddavies
 * i_packc.c -> /product/ipcore/ipshared/src/RCS/i_packc.c,v
 *
 * Ooops, well that last message ended abruptly.  Anyway, change all references
 * to CCRuns to references to Run32's.
 *
 * Revision 1.2  1995/05/10  02:00:37  ddavies
 * ddavies on Tue May 9 18:58:56 PDT 1995
 *
 * i_packc.c -> /product/ipcore/ipshared/src/RCS/i_packc.c,v
 *
 * Fix typecast to memset is happy in acc.  Set pPBCompInfo->runStart and
 * pPBCompInfo->runEnd to 0 when starting a new line so we don't use some
 * random runs when starting a new line after the previous line was interrupted
 * by filling the compression buffer.
 *
 * Revision 1.1  1995/05/09  02:52:26  ddavies
 * Initial revision
 *

 */

