/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_packd.c: High level code used in PackBits decompression.
 *
 *   	Int32	i_getSizeOfPackBitsDecompressionInfo()
 *			Returns size of opaque data structure used in
 *			PackBits decompression.
 *
 *	Int32	i_initializePackBitsDecompressionInfo()
 *			This routine initializes the opaque data structure
 *			used in PackBits decompression.
 *
 *	Int32	i_decompressRowsPackBits()
 *			Decompress specified number of rows using packbits
 *			decompression.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"

    /*  prototyes */
#include "shrcmprs.prv"
#include "shrcmprs.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_packd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* Here is a duplicate of the big comment that describes our general
 * approach to PackBits decompression.  There's a similar comment about
 * compression in i_packc.c.
 */

/* First, here is a description of the PackBits format from section 9 of
 * the TIFF 6.0 specification:
 *
 * -----------------------------------------------------------------
 *
 * In choosing a simple byte-oriented run-length compression scheme, we
 * arbitrarily chose the Apple Macintosh PackBits scheme.  It has a good
 * worst case behavior (at most 1 extra byte for every 128 input bytes).
 * For Macintosh users, the toolbox utilities PackBits and UnPackBits will
 * do the work for you, but it is easy to implement your own routines.
 *
 * A pseudo-code fragment to unpack might look like this:
 *
 *    Loop until you get the number of unpacked bytes you are expecting:
 *
 *	Read the next source byte into n.
 *
 *	If n is between 0 and 127 inclusive, copy the next n+1 bytes literally.
 *
 *	Else if n is between -127 and -1 inclusive, copy the next byte -n+1
 *	times.
 *
 *	Else if n == -128, noop.
 *
 *   Endloop
 *
 * In the inverse routine, it is best to encode a 2-byte repeat run as
 * a replicate run except when preceded and followed by a literal run.
 * In that case, it is best to merge the three runs into one literal run.
 * Always encode 3-byte repeats as replicate runs.
 *
 * That is the essence of the algorithm.  Here are some additional rules:
 *
 * * Pack each row separately.  Do not compress across row boundaries.
 *
 * * The number of uncompressed bytes per row is defined to be
 *   (ImageWidth + 7)/8.  If the uncompressed bitmap is required to have
 *   an even number of bytes per row, decompress into word-aligned buffers.
 *
 * * If a run is larger than 128 bytes, encode the remainder of the run as
 *   one or more additional replicate runs.
 *
 * -------------------------------------------------------------------------
 *
 * Simple enough.  I looked at the XIS and Leffler implementations.  They
 * compress and decompress in the simple ways.  The compression code uses
 * a state machine to remember whether it's compressing runs of bytes or
 * passing through literal strings (hereinafter called "literals").
 * The decompression code can just use memcpy and memset.
 *
 * But hey, why not live a little?  The compression code has more room
 * for foolishness, but there are thing we can do to make the decompression
 * code fast.  Before we get into method, let's discuss interfaces.
 *
 * We'd like to make routines that can be used by the DAE, SmartPaper,
 * PerfectScan and the PerfectScan Basic Viewer.  The DAE says we're not
 * allowed to have pointers, so any extra arrays or tables we need should
 * be attached to one pointer-free object.  PerfectScan already has a TIFF
 * package, so just wants to plug in a new body of compression code.
 * SmartPaper wants to use the Leffler TIFF code found in IPWrap.  We'd like
 * to use the Leffler code as a testbed anyway.
 *
 * Here's an interface that fits into all worlds:
 *   1.	The client asks for the size of the opaque data object that will
 *	retain state and hold tables and buffers between scan lines.
 *	This is done using i_getSizeOfPackBitsDecompressionInfo.  There
 *	is one type of object for compression, another for decompression.
 *   2.	The client allocates a block of this size.
 *   3.	The client calls i_initializePackBitsDecompressionInfo.  This fills
 *	in the info that you wouldn't want to recompute on every
 *	compression or decompression call.
 *   4.	The client calls i_decompressRowsPackBits.  Image bytes are
 *	decompressed in units of scan lines.  The routine returns when
 *	either the desired number of lines has been processed or it has
 *	emptied the compression buffer.  Clients can restart in either
 *	case.
 *   5. When complete, the client frees the compression block.
 *
 * Yeah, a callback routine would be a nice way to refill the compression
 * buffer instead of returning and restarting.  Unfortunately, callback
 * routines in the DAE world are pretty much out of touch with important
 * things like I/O.  They'd have to knock on the interpreter's back door
 * to ask for access.  This is very hard.  So, we just return to the
 * client when necessary and let the client sort it all out.
 *
 * There's still some difficulty with byte order.  We will declare that
 * all uncompressed data is in PIXR order.  In fact, it's almost certainly
 * in PIXRs.  We'll try to avoid copying data when possible.  However,
 * the PackBits standard says that the compressed bytestream should be
 * read one byte at a time in the order that's natural to the machine.
 * That's great, except that a significant component of a PackBits file
 * is strings of literals taken directly from the image.  On little-endian
 * platforms, these would be in the wrong order if copied directly to PIXRs.
 *
 * We have three choices.  We could byteswap the compressed data.  We
 * could byteswap the uncompressed data.  We could write fancy code that
 * did byteswapping while it was compressing or decompressing.  What to do?
 *
 *
 * An aside.  I also thought about (ok, well, maybe I wrote a lot of code)
 * storing compressed data in PIXR order to avoid byteswaps when using
 * PackBits as an internal cache formt.  The PerfectScan basic viewer
 * uses PackBits in this way.  Storing unaligned data in PIXR order on
 * a little endian machine is nonsense.  Ooops.  Say a PIXR line contains
 * the image bytes:
 *	A B C D   D D E F   G H I J
 *
 * If we tried to compress this without doing byteswapping, we'd get
 *	2 A B C  -2 D 5 E   F G H I  J x x x
 *
 * On a little endian machine, the bytes at the bottoms of the words
 * appear first in the data stream, so the bytestream looks like
 *	C B A 2 E 5 D -2 I H G F x x x J
 *
 * Now say the stream is shifted by one byte relative to word boundaries
 * when it's read back.  This is to be expected since the position of
 * the bytestream relative to word boundaries is random.  Writing the
 * stream starting at a word boundary, we get:
 *	y C B A 2 E 5 D -2 I H G F x x x J z z z
 *
 * If this is read in units of words, which is the whole point of storing
 * the thing in PIXR order, we'd get
 *	A B C y   D 5 E 2  G H I -2  x x x F z z z J
 *
 * This bears little similarity to what we started with; it's trash.
 * Therefore, since the compressed data will be treated as a bytestream,
 * we MUST store it in the natural byte order of the platform we're on.
 *
 *
 * In all cases, we need code that blits bytes from the compressed data
 * stream to the image.  If the uncompressed and compressed data are both
 * in the platform's natural byte order, we could use memcpy.  This
 * is a point in favor of converting the uncompressed data into the
 * processor's natural byte order.
 *
 * But, there are points against that too.  For one, there's a lot of
 * uncompressed data, so byteswapping it is not very efficient.  Second,
 * most runs are fairly short, so making zillions of procedure calls
 * (memset & memcpy) is not very efficient.
 *
 * Byteswapping the compressed data stream doesn't work because we
 * can't guarantee anything about word alignment of the data stream.
 * Say byte A was supposed to be the first byte processed.  If the
 * data stream were going appear on a word boundary, we'd put byte A
 * in the fourth byte of the stream so it would appear as the first
 * byte after byteswapping.  However, if the bytestream were going
 * to appear in memory with it's first byte in byte 1 of a word (relative
 * to the left end of the word), then we're in trouble.  Yeah, we
 * could have put byte A in byte 2 (if we were prescient), but the trash
 * in byte 0 of the word would appear as the third byte of the swapped
 * bytestream, so we'd be forever confused.
 *
 * Our last choice is to build a byte-blit function that deals with
 * image data in PIXR order and the compressed bytestream in natural
 * byte order. If we do the byteswap in the byteblt function, then we have the
 * minimum number of memory transfers (don't write a block, then write
 * it again to byteswap it) and the best cache usage, so that's what we'll do.
 * Since this method doesn't require any preprocessing or postprocssing
 * of either the image or compressed data stream, it's the easiest one
 * to interface with.
 *
 * This is the decompression routine, so let's talk about that.  While
 * we could fool around with fancier things, it seems like the simplest,
 * and fastest, way to do decompression is to simply read a count byte,
 * then decide whether to ignore it (noop), replicate the following byte
 * or copy the next string of literals to the output.
 *
 * Here's where we should touch on a little software engineering.  Both the
 * replication and copy steps could use a byteblt routine.  We'd probably
 * want to write this routine in assembly on the PC so we could use the
 * very nice 486 byteswap opcode.  However, calling a routine many times
 * for short runs is a real performance sink, especially on glorious
 * Intel processors.
 *
 * One alternative is to have an assembly routine that decodes an entire
 * line.  Since there's really not a lot to this routine
 * in addition to the byteblt, we'll do that.  Naturally, there will also
 * be a C version that does byteswaps the hard way if necessary.  I'm
 * not sure if it's worthwhile to do a SPARC assembly version - who would
 * use it?
 *
 * In fact, the routine that expands a compressed data stream into an
 * image is a lot like the routine that turns an image plus
 * information about where the runs of literals and replicas start and
 * stop (see below) into a compressed data stream.  Both basically blit
 * (and byteswap) bytes.  Their control info comes from different places.
 * The decompression routine gets its info from the counts embedded in
 * the file.  The compression routine gets its control info from the
 * descriptions of runs (see below).  The decompression routine
 * byteswaps immediately after reading the compressed data.  The
 * compression routine byteswaps immediately before writing the
 * compressed data.
 *
 *
 * Time for a little more software engineering.  Clients can't really
 * predict the amount of compressed data required to represent a given slice of
 * an image.  What happens when the compression buffer is full or empty?
 * Originally, we were going to let the client supply a callback procedure.
 * We would call it whenever the compressed buffer needed attention.
 * Unfortunately, this doesn't work well for the DAE.  The callback
 * procedure would be down very low.  It would have a very hard time
 * knocking on the back door of the PostScript interpreter asking for
 * more memory or some I/O to be done.
 *
 * Similarly, a client may wish to stop compressing or decompressing
 * at a non-strip or non-tile boundary with the expectation of starting
 * up again later.  This makes sense when the strips or tiles are very
 * large and there's not much memory on the platform.
 *
 * Our new scheme is to simply return if we either run out of compressed
 * data (decompression) or room for compressed data (compression) or
 * have processed the specified number of scan lines.
 * The compression and decompression routines return two flags.  One
 * says the routine stopped because the compressed data is finished.
 * The other says the routine stopped because the image data is finished.
 * Occasionally, both may be set.  This is common at the bottom of
 * an image.
 *
 * Whenever a routine stops, it retains enough state to restart.
 * If it did not run out of image data, the client may move the current
 * image block around as a unit, but should not attempt to guess which
 * part is no longer needed and remove it.
 *
 * We treat the compressed data a little differently.  The client can
 * know exactly how much memory is required for uncompressed image lines.
 * The amount of compressed data needed to represent an image line is
 * usually unknown, though the client may know the number of bytes
 * used to represent a group of lines, or strip.  In spite of this ignorance,
 * we would like to present an interface that allows clients
 * to use one buffer/strip, or a buffer for a sub-strip, or double
 * buffers for alternate sub-strips or a ring buffer or whatever.
 *
 * To acheive this flexibility, the decompression routine always
 * returns the number of bytes of compressed data read.  The client
 * may safely dispose of these bytes.  It is the client's job to advance
 * the pointer to compressed data and recompute the number of compressed
 * bytes remaining on each call.
 *
 * In bit-oriented compression protocols, scan lines may not end
 * on byte boundaries, so the number of bytes read only roughly
 * corresponds to the number of bits required to represent the
 * scan lines.  In PackBits compression (a byte-oriented protocol),
 * the number of bytes read by a call to i_decompressRowsPackBits should
 * exactly equal the number of bytes required to represent these lines.
 *
 * Clients supply two arguments to the decompression routine.
 * One specifies whether the routine should continue with
 * existing image data.  The other specifies whether the the routine
 * should continue with the current compression data *stream*.  This
 * is set if we ran out of compression data last time and the new
 * compression data should be considered an extension to the previous
 * block of data.  For example, if we ran out of compressed data after
 * reading a bytecount but before reading any data, this flag says to
 * treat the first byte of the new buffer as the first data byte
 * of the run specified by the last bytecount read.  Otherwise, any
 * saved state is discarded.
 *
 * If the image data is new, the compression or decompression routine
 * discards its saved image state.  The same goes for new compression
 * data and saved compression state.  If told to continue on with
 * existing image data or an existing compression stream, and there
 * is no corresponding saved state, an error is returned.
 *
 * When they have exhausted the compression info, the decompression
 * routines also return a parameter that vaguely specifies the amount
 * of the original request that has been processed.
 * Clients are free to ignore this parameter.  It is intended to allow
 * clients guess how much compressed data is needed to complete the slice.
 * Before restarting, the client should supply more compressed data.
 *
 * Clients supply one more argument that tells the decompression
 * routines what to do with the pipeline of compressed data when the
 * specified number of scan lines has been processed.
 * If flushPipeline is set during decompression, it means the decompression
 * routine should not stop when it has no more compressed data to
 * put into the pipeline.  The data already in the pipeline, padded with
 * 0's, should suffice to reach the end of the last scan line.
 * Furthermore, if the pipeline empties without reaching the end of
 * the specified image line, an error should be returned.
 *
 * Another way of looking at flushPipeline is that it is an indicator
 * from the client that we should be reaching the end of a contiguous
 * block of compressed data.  In addition to flushing the pipeline,
 * the internal routines discard all saved state when the specified
 * number of scan lines have been successfully decompressed.
 *
 * Time for more detail on the internal workings of decompression.
 * We do things a line at a time.  We read each byte count from
 * the compressed data and either copy literals or fill in replicas
 * as indicated.  If flushPipeline is true, we expect to finish a
 * scan line just as the compressed data runs out.  The decompState
 * records what we were doing when we failed to read the next compressed
 * byte.  The decompression state block contains:
 *   1. Flag saying whether saved image data is valid.
 *   3. Number of current uncompressed line relative to top line of
 *	current image slice.
 *   4. Byte offset into current scan line where next uncompressed
 *	word will be written.
 *   5. Partial uncompressed word.
 *   6.	Number of valid bytes in partial uncompressed word.
 *   7.	Flag saying whether compressed stream state is valid.
 *   8. Current decompression state:
 *	a. Preparing to read a new byte count
 *	b. Preparing to read data. This is also used to complete
 *	   interrupted strings of literals.
 *   9. Current byte count for state 6c.
 *  10. Byte offset of next compressed byte to read relative to beginning
 *	of the compressed data block.
 *  11.	Pointer to routine used to decompress one line.
 *
 * Note that when we run out of compressed data, BOTH the image information
 * and the decompression information are valid.  The image information
 * holds the partially formed next output word and where to write it.
 * The decompression state says how many bytes remain in the current run
 * (if any) and what state we should restart in when new compression info
 * arrives.
 */


/* PUBLIC */
/**************************************************************************
 *   i_getSizeOfPackBitsDecompressionInfo()
 *
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform PackBits
 *	decomression.
 *
 *	Input:
 *		pSize:	pointer to Int32 that will contain the size.
 *			of the buffer required to hold the symbol
 *			information in bytes.
 *
 *	Returns:
 *		ia_successful		 it worked
 *		ia_depthNotSupported	 depth was illegal.
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfPackBitsDecompressionInfo (
    IPState    *pIPState,
    Int32     *pSize)
{
    PROLOGUE;

	/* Return size of PackBits decompression structure
	 */
    *pSize = sizeof(PackBitsDecompressionInfo);

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_initializePackBitsDecompressionInfo()
 *
 *	This procedure fills a block of PackBits compression info
 *	with necessary information.
 *
 *	Input:
 *		pPBDecompInfoIn:	pointer to PackBits compression info
 *		decompInfoBufSize:	the number of bytes in the
 *					compression info.
 *
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		pPBDecompInfoIn was NULL, buffer was
 *					wrong size, the byte order
 *					was illegal
 *
 **************************************************************************/
Int32  CDECL 
i_initializePackBitsDecompressionInfo (
    IPState    *pIPState,
    void       *pPBDecompInfoIn,
    Int32	decompInfoBufSize)
{
    PROLOGUE;
    PackBitsDecompressionInfo	*pPBDecompInfo;
    Int32	realDecompInfoBufSize;


    pPBDecompInfo = (PackBitsDecompressionInfo *)pPBDecompInfoIn;

	/* Get the size the block should be */
    __error__ = i_getSizeOfPackBitsDecompressionInfo (pIPState,
						      &realDecompInfoBufSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* See if any of the parameters were illegal */
    if ((pPBDecompInfo == NULL) ||
	(realDecompInfoBufSize > decompInfoBufSize))
	return ia_invalidParm;

	/* Get the addresses of the routines used to compress lines.
	 */
    ip_getPackBitsDecompProc (pIPState, &(pPBDecompInfo->decompRowProc));

	/* Initialize all the rest of the default stuff.  Yeah,
	 * initializing things and declaring they they are invalid
	 * is sort of silly, but I like to initialize everying
	 * so things are deterministic.
	 */
    pPBDecompInfo->lineNumber = 0;
    pPBDecompInfo->imageInfoIsValid = cFalse;
    pPBDecompInfo->imageWord = 0;
    pPBDecompInfo->nextVacantByte = 0;
    pPBDecompInfo->imageWordOffset = 0;
    pPBDecompInfo->decompInfoIsValid = cFalse;
    pPBDecompInfo->decompState = cPBDecompNewLine;
    pPBDecompInfo->runByteCount = 0;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_decompressRowsPackBits()
 *
 *	This procedure decompresses PackBits-encoded images in units of
 *	scan lines.
 *
 *	Input:
 *		pPBDecompInfoIn:	Pointer to PackBits decompression info
 *		pImage:			Pointer to the upper left corner
 *					of the frame of the uncompressed
 *					image
 *		w:			Width of uncompressed image including
 *					frame.
 *		startLine:		First image line to be decompressed
 *					relative to the top of the frame
 *		numLines:		Number of lines to be decompressed
 *		d:			Depth of uncompressed image.
 *		bpl:			bytes/line of uncompressed data.
 *		pCompData:		Pointer to next byte of compressed
 *					data to be read.  If several calls
 *					to this routine are needed to process
 *					a block of compressed data, this
 *					pointer must be updated on each call.
 *		compDataBufSize:	Number of bytes of compressed data
 *					remaining in the buffer.
 *		continueImageBlock:	cTrue => we are decompressing
 *					into the same slice of image as was
 *					specified in the previous call and the
 *					saved state in pPBDecompInfoIn is
 *					valid.
 *					This arg is set to cFalse if either
 *					we are starting with all new data
 *					or the previous call returned because it
 *					ran out of image data.
 *		continueCompDataStream:	cTrue => The saved decompression state
 *					in pPBDecompInfoIn is valid.  This
 *					allows the procedure to treat several
 *					buffers of compressed data as though
 *					it were one contiguous buffer.
 *					If set to cFalse, this routine will
 *					discard all previous compression
 *					state..
 *		flushPipeline:		If this arg is set, this routine
 *					will attempt to complete the
 *					specified number of scan lines
 *					WITHOUT stopping when it starts to
 *					run out of data; i.e. it continues
 *					until its pipeline is completely
 *					empty.  If this is not possible,
 *					an error is raised.
 *					When flushPipeline is set and the
 *					decompression task suceeds in
 *					processing all image lines, all
 *					saved state is marked invalid.
 *					Subsequent calls with either
 *					continueImageBlock or
 *					continueCompDataStream set will return
 *					ia_invalidParm.
 *		pExhaustedImageData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because specified
 *					number of lines have been processed.
 *					This may require more than one
 *					call to this routine.
 *		pExhaustedCompData:	Pointer to returned arg.  When
 *					arg is returned with value cTrue,
 *					routine stopped because it reached
 *					the end of the compressed data buffer
 *					as specified by compDataSize.  This
 *					routine may be called again with 
 *					new compression data and the the 
 *					same image.  This arg will not be 
 *					set if flushPipeline is set and we 
 *					failed to decompress the specified 
 *					number of scan lines (routine returns
 *					ia_bufferExceeded instead).
 *		pPercentDone:		Pointer to returned arg.  When
 *					routine returns with *pExhaustedCompData
 *					set, this arg specifies the rough
 *					percentage of the image block that
 *					has been decompressed.  It is
 *					intended to be used to estimate the
 *					amount of compressed data needed
 *					to complete the task.  The result
 *					ranges from 0 to 100.
 *		pBytesRead:		Pointer to returned arg.  Contains
 *					number of compressed data bytes
 *					read by the routine before it returned.
 *					The client may safely dispose of
 *					these bytes.
 *
 *	Action:
 *		Decompresses a horizontal slice of PackBits data.
 *		There are several completion conditions:
 *
 *	__error__	!= ia_successful
 *	*pExhaustedImageData == don't care
 *	*pExhaustedCompData == don't care
 *		This means something catastrophic happened.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cFalse
 *		This means we produced the desired number of uncompressed
 *		image lines without running out of compressed data.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse and
 *		continueCompDataStream = cTrue.  pCompData and compDataSize
 *		should be adjusted to account for the number of 
 *		compressed data bytes that have already been read (returned
 *		in *pBytesRead).
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cTrue
 *	*pExhaustedCompData == cTrue
 *		This means we ran out of compressed data at the same time
 *		we completed the correct number of image lines.  This
 *		is the normal condition at the bottom of a TIFF slice or
 *		the bottom of the image.  It is not possible to restart
 *		from this state, though it is possible to decompress
 *		with new image and compression data.  When the client
 *		expects to run out of compressed data, the flushPipeline
 *		argument should be set.
 *
 *	__error__	== ia_successful
 *	*pExhaustedImageData == cFalse
 *	*pExhaustedCompData == cTrue
 *		The routine ran out of compression data before completing
 *		the specified number of lines.  This condition will not
 *		occur if flushPipeline was set.  If flushPipeline was set
 *		and we ran out of compressed data before completing the
 *		the task, an ia_bufferExceeded error will be returned.
 *		When this condition does occur, the routine may be
 *		restarted by supplying more compressed data, setting
 *		BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  Note that the first byte of the new compressed
 *		data must be the one immediately following the last
 *		byte of old compressed data.
 *		The image may be moved around as a block if necessary, but
 *		the internal composition of the image (and startLine
 *		and numLines) should not be changed.  When this
 *		condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much compression data is 
 *		needed to complete the slice.
 *		
 *		The image data is always in PIXR order.  This routine
 *		will byteswap strings of literals in the compressed data
 *		on the way to the image if necessary.
 *		
 *	Returns:
 *		ia_successful		it worked
 *		ia_depthNotSupported	depth was illegal.
 *		ia_invalidParm		a pointer was NULL or 
 *					(continueImageBlock && the saved image
 *					info is not valid) or
 *					(continueCompDataStream && the saved
 *					compression info is not valid)
 *		ia_bufferExceeded	ran out of compressed data when
 *					flushPipeline was set.
 *		ia_decodeFailure	an illegal code was enountered.
 *
 **************************************************************************/
Int32  CDECL 
i_decompressRowsPackBits (
    IPState    *pIPState,
    void       *pPBDecompInfoIn,
    UInt8      *pImage,
    Int32	w,
    Int32	startLine,
    Int32	numLines,
    Int32	d,
    Int32	bpl,
    UInt8      *pCompData,
    Int32	compDataBufSize,
    Bool	continueImageBlock,
    Bool	continueCompDataStream,
    Bool	flushPipeline,
    Bool       *pExhaustedImageData,
    Bool       *pExhaustedCompData,
    Int32      *pPercentDone,
    Int32      *pBytesRead)
{
    PROLOGUE;
    PackBitsDecompressionInfo	*pPBDecompInfo;
    Int32	widthInBytes;	/* number of bytes to decode
				 * per line */
    UInt8      *pCompDataEnd;	/* address 1 beyond the last
				 * byte of compressed data */
    UInt8      *pCurCompData;	/* current compressed data
				 * pointer */
    Int32	remainingLines;


	/* check for null pointers */
    if ((pPBDecompInfoIn == NULL) || (pCompData == NULL) ||
	(pImage == NULL) || (compDataBufSize == 0) ||
	(pExhaustedImageData == NULL) || (pExhaustedCompData == NULL) ||
	(pPercentDone == NULL) || (pBytesRead == NULL))
	return ia_invalidParm;

	/* Strip off the frame */
    pImage += (bpl << cLogFrameBits) + d * cFrameBytes;
    startLine -= cFrameBits;
    w -= 2*cFrameBits;

	/* load the local copies of the remote pointers */
    pPBDecompInfo = (PackBitsDecompressionInfo *) pPBDecompInfoIn;
    widthInBytes = (w*d + 7)/8;

	/* Initialize the state from the saved state passed in *pPBDecompInfo
	 * by using continueImageBlock and continueCompDataStream.
	 */
    if (!continueImageBlock)
    {
	    /* Initialize all the saved state about the image */
	pImage += startLine * bpl;
	pPBDecompInfo->lineNumber = 0;
	pPBDecompInfo->imageWord = 0;
	pPBDecompInfo->nextVacantByte = 0;
	pPBDecompInfo->imageWordOffset = 0;
	pPBDecompInfo->imageInfoIsValid = cTrue;
	remainingLines = numLines;
    }
    else	/* supposed to use the saved state */
    if (!(pPBDecompInfo->imageInfoIsValid))
	return ia_invalidParm;			/* saved state is invalid */
    else
    {
	    /* it's ok to use the saved state */
	remainingLines = numLines - pPBDecompInfo->lineNumber;
	pImage += (startLine + pPBDecompInfo->lineNumber) * bpl;
    }

    if (!continueCompDataStream)	/* start with new compressed data */
    {
	    /* next state is "start a new line" */
	pPBDecompInfo->decompState = cPBDecompNewLine;
	pPBDecompInfo->runByteCount = 0;
	pPBDecompInfo->decompInfoIsValid = cTrue;
    }
    else
    if (!pPBDecompInfo->decompInfoIsValid)
	return ia_invalidParm;

	/* calculate the address just beyond the last compressed
	 * data byte.
	 */
    pCurCompData = pCompData;
    pCompDataEnd = pCompData + compDataBufSize;


	/* We now just decompress each line.  The line decompression
	 * routine fixes the state each time it ends.  This allows
	 * us to simply restart on the next line.
	 * It's silly to recalculate the address of the image
	 * and compression buffers each time, so we do that here.
	 * There are two ways this decompression routine can
	 * complete.  It can actually decompress the right number
	 * of lines.  That's detected in this loop as we decrement
	 * the number of lines remaining.  It can also run out
	 * of compressed data before it runs out of image data.
	 * There are two ways that can happen.  We can run out of
	 * compressed data in the middle of a scan line.  When this
	 * happens, pPBDecompInfo->imageWordOffset != 0 on return.
	 * We can get to the end of a scan line, but not the end of
	 * the last scan line.  This is indicated by
	 * pPBDecompInfo->imageWordOffset == 0 and remainingLines > 1
	 * (before the decrement at the bottom of the loop).  If
	 * either of these occurs, we set the return values and
	 * return immediately.
	 */

	/* Loop through the scan lines */
    for (;
	 remainingLines > 0;
	 remainingLines--, pImage += bpl)
    {
		/* decompress one line of image data */
	ERR_PASS( pPBDecompInfo->decompRowProc(pIPState,
					       pPBDecompInfo,
					       pImage,
					       pImage+widthInBytes,
					       &pCurCompData,
					       pCompDataEnd));

	    /* If we didn't get to the end of a scan line, or
	     * ran out of compression data and are not out of scan
	     * lines, decide whether it was ok that we ran out of
	     * compressed data.  If flushPipeline was set, then it
	     * was not ok.
	     * We need to check to see if we stopped in mid-scanline
	     * to detect the case of running out of data in the middle
	     * of the last scan line of the strip.
	     */
	if ((pPBDecompInfo->imageWordOffset != 0) ||
	    (pPBDecompInfo->nextVacantByte != 0) ||
	    ((pCurCompData == pCompDataEnd) && (remainingLines > 1)))
	{
	    if (flushPipeline)	/* should have been able to get to the
				 * end of the image, but we didn't.
				 */
	    {
		return ia_bufferExceeded;
	    }
	    else		/* We're stopped somewhere in the middle
				 * of the image strip we were supposed to
				 * complete.  But the client said this
				 * might happen (flushPipeline == cFalse),
				 * so it's ok.  Just tell the client that
				 * it happened.
				 */
	    {
		*pExhaustedImageData = cFalse;
		*pExhaustedCompData = cTrue;
		*pPercentDone = (100 *
				 ((pPBDecompInfo->lineNumber) * widthInBytes +
				     pPBDecompInfo->imageWordOffset)) /
				(numLines * widthInBytes);
		*pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

		    /* if we happened to stop on an image scan line boundary,
		     * increment the saved line number so we don't do it again.
		     * The test for "at beginning of line~ is a little
		     * tricky.  The word offset into the line must be 0,
		     * the offset into the first word must be 0 AND the
		     * decomposition state must be cPBDecompNewLine.  The
		     * last condition guarantees that we didn't stop
		     * just after finding a bytecount but haven't read
		     * the first data byte yet.
		     */
		if ((pPBDecompInfo->imageWordOffset == 0) &&
		    (pPBDecompInfo->nextVacantByte == 0) &&
		    (pPBDecompInfo->decompState == cPBDecompNewLine))
		    pPBDecompInfo->lineNumber++;

		NORMAL_RETURN;
	    }
	}
	    /* increment the current line number just in case we need
	     * to stop in the next line.
	     */
	pPBDecompInfo->lineNumber++;
    }

	/* We've done all the scan lines.  So, just
	 * say things were successful and return.
	 */
    *pExhaustedImageData = cTrue;
    if (pCurCompData != pCompDataEnd)
	*pExhaustedCompData = cFalse;
    else
	*pExhaustedCompData = cTrue;
    *pBytesRead = (UInt32)pCurCompData - (UInt32)pCompData;

	/* if flushPipeline is true, declare all the saved state to be
	 * invalid.  This helps the client catch bugs in which continuation
	 * was attempted, but the saved state is invalid.
	 */
    if (flushPipeline)
    {
	pPBDecompInfo->imageInfoIsValid = cFalse;
	pPBDecompInfo->decompInfoIsValid = cFalse;
    }

    NORMAL_RETURN;
}



/*

 $Log: i_packd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.4  1998/01/13 23:49:14  ddavies
 i_packd.c -> /product/ipcore/ipshared/src/RCS/i_packd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1995/12/13 02:45:17  ejaquith
 i_packd.c -> /product/ipcore/ipshared/src/RCS/i_packd.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/05/19  02:52:24  ddavies
 * ddavies on Thu May 18 19:51:45 PDT 1995
 *
 * i_packd.c -> /product/ipcore/ipshared/src/RCS/i_packd.c,v
 *
 * Don't check the return code from ip_getPackBitsDecompProc since it's
 * been turned into void.
 *
 * Revision 1.1  1995/05/09  02:54:53  ddavies
 * Initial revision
 *

 */

