/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_ptrfix.c:	Utility routines used to fix up pointers
 *		after DAE garbage collection.
 *
 *   	Int32	i_fixupPointersAfterGarbageCollection()
 *			Loop through all blocks in a garbage collection
 *			ring having all pointers adjusted after a block
 *			has moved.
 *
 *   	Int32	ip_initPointerFixup()
 *			Initialize a PointerFixup structure when a block
 *			is allocated
 *
 *   	Int32	ip_detachBlockFromPtrFixupRing()
 *			Remove a block from the ring of blocks used
 *			by garbage collection.
 *
 *	Int32	ip_isolateAllBlocksInPtrFixupRing()
 *			Disconnect all blocks in a pointer fixup ring
 *			from each other so that they may be deleted
 *			in any order without worrying about pointers
 *			to deleted stuff..
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcc.prv"
#include "ptrfixup.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_ptrfix.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");


/* PUBLIC */
/**************************************************************************
 *   i_fixupPointersAfterGarbageCollection()
 *
 *	This is the procedure that's called immediately after a
 *	client's garbage collector has moved a single block of memory.
 *	It goes through the ring of memory blocks that might have pointers
 *	to the moved block (including the moved block itself) and has
 *	all pointers adjusted.
 *
 *	Input:
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *
 *	Returns:
 *		ia_successful	-------------------------------	it worked
 *		something else (from individual fixup routines)	it didn't
 *
 **************************************************************************/
Int32  CDECL 
i_fixupPointersAfterGarbageCollection (
    IPState    *pIPState,
    void       *pOldAddr,
    void       *pNewAddr)
{
    PROLOGUE;
    PointerFixup	*pFixup;	/* ptr to PointerFixup blocks */
    UInt32		 blkLength;	/* length of block that moved */

	/* Use a pointer of the correct type and get the mask for
	 * the structures in this type of block.
	 */
    pFixup = (PointerFixup *) pNewAddr;
    blkLength = pFixup->blockSize;

	/* Adjust the pointers from the blocks on either side of this
	 * one back to this block.  If there's only one block in the
	 * ring (this one), just fix its pointers.
	 */
    if (pFixup->pNext != pOldAddr)	/* was NOT pointing to itself */
    {
	pFixup->pNext->pPrev = pFixup;
	pFixup->pPrev->pNext = pFixup;
    }
    else	/* was pointing to itself */
    {
	pFixup->pNext = pFixup;
	pFixup->pPrev = pFixup;
    }

	/* if blkLength is 0, the client has guaranteed that nothing
	 * should be done when this block is moved.  Well, it does seem
	 * like a good idea to update the ring pointers anyway,
	 * so we did that.  Anyway, quit now if we're not supposed to
	 * look in other blocks for pointers.
	 */
    if (blkLength == 0)
	NORMAL_RETURN;

	/* Follow the pointers around the ring starting with this block
	 * until we get back here again.  We blindly call the fixup
	 * proc on every block because of the problems inherent in trying
	 * to remember which types of structures and structure pointers
	 * are in each block.  As each structure is modified (a pointer
	 * added), we would have to manually update some sort of
	 * list of all structure pointers.  This is too prone to error
	 * to think about.
	 */
    do
    {
	if (pFixup->fixupProc != NULL)
	{
	    __error__ = (pFixup->fixupProc)(pIPState, pFixup, pOldAddr,
					    pNewAddr, blkLength);
	    if (__error__ != ia_successful)
		NORMAL_RETURN;	/* this is a disaster, so just stop */
	}
	    /* point to the next block */
	pFixup = pFixup->pNext;
    } while (pFixup != (PointerFixup *)pNewAddr);

	/* We must have finished successfully, so return */
    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_initPointerFixup()
 *
 *	This procedure initializes the entries in a PointerFixup block
 *	with its arguments.
 *
 *	Input:
 *		pNewBlock:	Pointer to the PointerFixup block to be
 *				initialized.
 *		blkLength:	Length of the entire memory block containing
 *				the new PointerFixup block in bytes.
 *		pAttachPt:	Pointer to PointerFixup routine next to which
 *				this new block should be inserted in the
 *				ring of blocks to be inspected during
 *				pointer fixup operations.
 *		fixupProc:	Pointer to procedure used to fix up
 *				pointers inside this entire memory block.
 *
 *	Returns:
 *		ia_successful	it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_initPointerFixup (
    IPState    *pIPState,
    PointerFixup *pNewBlock,
    UInt32	blkLength,
    PointerFixup *pAttachPt,
    Int32 (CDECL *fixupProc) (
		    IPState    *pIPState,
		    void       *pThisBlock,
		    void       *pOldBlock,
		    void       *pNewBlock,
		    UInt32	blkLength))
{
	/* just initialize the block */
    PROLOGUE;
    pNewBlock->blockSize = blkLength;
    pNewBlock->fixupProc = fixupProc;

	/* insert this new block into a ring.  If there is no existing
	 * ring (pAttachPt == NULL), make it point to itself.
	 */
    if (pAttachPt != NULL)
    {
	pNewBlock->pNext = pAttachPt->pNext;
	pAttachPt->pNext->pPrev = pNewBlock;
	pAttachPt->pNext = pNewBlock;
	pNewBlock->pPrev = pAttachPt;
    }
    else	/* start a new ring */
    {
	pNewBlock->pNext = pNewBlock->pPrev = pNewBlock;
    }

    NORMAL_RETURN;
}



/* PRIVATE */
/**************************************************************************
 *   ip_detachBlockFromPtrFixupRing()
 *
 *	This procedure removes a block from the ring of blocks that is
 *	inspected for pointer fixup when a block is moved.
 *
 *	Input:
 *		pOldBlock:	Pointer to the PointerFixup block to be
 *				initialized.
 *
 *	Returns:
 *		ia_successful	it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_detachBlockFromPtrFixupRing (
    IPState    *pIPState,
    PointerFixup *pOldBlock)
{

	/* If the ring has more than one block in it, take this one
	 * out, otherwise, leave it alone since it will soon be
	 * deleted anyway.
	 */
    PROLOGUE;
    if (pOldBlock->pNext != pOldBlock)
    {
	pOldBlock->pNext->pPrev = pOldBlock->pPrev;
	pOldBlock->pPrev->pNext = pOldBlock->pNext;
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_isolateAllBlocksInPtrFixupRing()
 *
 *	This procedure isolates all blocks in a pointer fixup ring.
 *	It is used just before freeing all the blocks.  The freeing
 *	operation might otherwise try to detach a block from something
 *	that was already free, causing chaos.
 *
 *	Input:
 *		pOldBlock:	Pointer to the PointerFixup block to be
 *				initialized.
 *
 *	Returns:
 *		ia_successful	it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_isolateAllBlocksInPtrFixupRing (
    IPState    *pIPState,
    PointerFixup *pOldBlock)
{
    PROLOGUE;
    PointerFixup	*pCurBlk, *pNextBlk;

	/* Go around the ring, turning each block's pointers inwards. */
    for (pCurBlk = pOldBlock->pNext; pCurBlk != pOldBlock; pCurBlk = pNextBlk)
    {
	pNextBlk = pCurBlk->pNext;
	pCurBlk->pNext = pCurBlk;
	pCurBlk->pPrev = pCurBlk;
    }

	/* Fix the pointers in the original block too */
    pOldBlock->pNext = pOldBlock;
    pOldBlock->pPrev = pOldBlock;

    NORMAL_RETURN;
}



/*

 $Log: i_ptrfix.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 Imported sources

 Revision 1.5  1998/01/13 23:49:17  ddavies
 i_ptrfix.c -> /product/ipcore/ipshared/src/RCS/i_ptrfix.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1997/07/11 01:49:06  ddavies
 i_ptrfix.c -> /product/ipcore/ipshared/src/RCS/i_ptrfix.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.3  1996/07/18 23:42:22  ddavies
 i_ptrfix.c -> /product/ipcore/ipshared/src/RCS/i_ptrfix.c,v

 Added implementation of ip_isolateAllBlcoksInPtrFixupRing.  This proc
 is used to disconnect all the memory blocks in a pointer fixup ring
 from each other before deleting them all.  Some deletion procedures
 would otherwise try to reference other (possibly already deleted) blocks
 in the ring to detach themselves.
 Fuss with indenting and comments too.

 * Revision 1.2  1995/12/13  02:45:22  ejaquith
 * i_ptrfix.c -> /product/ipcore/ipshared/src/RCS/i_ptrfix.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.1  1995/07/10  23:03:52  ddavies
 * Initial revision
 *

 */

