/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_rastln.c: C implementation of raster line drawing routine
 *
 *          rasterize line (bitblt) for 32 bit words:
 *                  Int32	i_rasterLine()
 *
 *
 */


#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "lmask.prv"

    /* prototypes */
#include "shrrast.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_rastln.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");



/* PUBLIC */
/*****************************************************************************
 * Draw line in a binary image row [left...right]
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *****************************************************************************/
Int32 CDECL 
i_rasterHorizLine (
    IPState    *pIPState,
    UInt32     *pRow, /* Ptr to first word in binary image row */
    Int32	left, /* Left and right inclusive bounds to draw line */
    Int32	right,
    Int8	val)   /* Value to write, 0 or 1 */
{
    PROLOGUE;
    Int32  lword = left >> 5;
    Int32  rword = right >> 5;

    if (val)
    {
	if (lword == rword)
	{
	    /* Line is contained in one word,
	     * Make a 0..0 word with 1s [(left MOD 32)..(right MOD 32)]
	     */
	    pRow[lword] |=
		((UInt32)0xFFFFFFFF >> (left & 0x1F)) &
		leftMask[right & 0x1F];
	}
	else {
	    pRow[lword] |= (UInt32)0xFFFFFFFF >> (left & 0x1F);
	    pRow[rword] |= leftMask[right & 0x1F];
	    
	    for (++lword; lword < rword; lword++)
		pRow[lword] = 0xFFFFFFFF;
	}
    }
    else
    {
	if (lword == rword)
	{
	    /* Line is contained in one word
	     * Make a 1..1 word mask with 0s [(left MOD 32)..(right MOD 32)]
	     */
	    pRow[lword] &=
		(leftMask[left & 0x1F] << 1) |
		(0x7FFFFFFF >> (right & 0x1F));
	}
	else {
	    pRow[lword] &= (leftMask[left & 0x1F] << 1);
	    pRow[rword] &= 0x7FFFFFFF >> (right & 0x1F);
	    
	    for (++lword; lword < rword; lword++)
		pRow[lword] = 0x00000000;
	}
    }
    
    NORMAL_RETURN;
}


/* PUBLIC */
/*****************************************************************************
 * Draw line in image from (x0,y0) to (x1,y1)
 *
 * Returns:
 *  ia_successful   	invocation done and okay
 *  ia_invalidParm  	unsupported depth (wasn't 1,2,4, or 8)
 *  ia_internal	    	routine failed, if happens report as bug!
 *
 * Notes:
 *  Caller MUST have already ensured that the points (x0,y0) and (x1,y1) are
 *  valid points in the image (as no checking is done here)
 *  Incrementing both x values (or both y values) and then calling this
 *  routine again will nicely not introduce any gaps in the 2 thick line.
 *  Code for binary images has been optimized.
 *****************************************************************************/
Int32 CDECL 
i_rasterLine (
    IPState    *pIPState,
    UInt8      *pr,    /* Ptr to image upper left corner, including frame */
    Int32	w,     /* Width of image, including frame, excluding pad */
    Int32	h,     /* Height of image, including frame, excluding pad */
    Int32	d,     /* Depth of image in bits */
    Int32	pitch, /* Bytes per line of image, including frame and pad */
    Int32	offx,  /* x offset value of the image */
    Int32	offy,  /* y offset value of the image */
    Int32	x0,    /* First (x,y) location to write */
    Int32	y0,
    Int32	x1,    /* Last (x,y) location to write */
    Int32	y1,
    Int32	val)   /* Value to write at each (x,y) location */
{
    PROLOGUE;
    Int32     a, b, decide, count, xinc, yinc, minorThresh;
    UInt32   *pRow;
    
    if (d != 1 && d != 2 && d != 4 && d != 8)
	return ia_invalidParm;
    
    /* Make pitch in terms of words */
    pitch >>= 2;

    
    /* Handle strictly horizontal or vertical line cases */
    if (y0 == y1 || x0 == x1)
    {
	if (d == 1)
	{
	    i_rasterOpFull (pIPState, pr, w, h, offx, offy,
			    MIN(x0,x1), MIN(y0,y1), ABS(x1-x0)+1, ABS(y1-y0)+1,
			    d, pitch << 2, val ? PIX_SET : PIX_CLR,
			    NULL, 0, 0, 0, 0, 0, 0, 0, 0);
	}
	else
	{
	    pRow = (UInt32 *)(pr) + (MIN(y0,y1) + offy) * pitch;
	    if (y0 == y1)
	    {
		/* Horizontal line */
		switch (d)
		{
		case 2:
		    for (a = x0 + offx; a < x0 + offx + ABS(x1-x0)+1; a++)
			SET_IMAGE_2BITS(pRow, a, val);
		    break;
		case 4:
		    for (a = x0 + offx; a < x0 + offx + ABS(x1-x0)+1; a++)
			SET_IMAGE_4BITS(pRow, a, val);
		    break;
		case 8:
		    for (a = x0 + offx; a < x0 + offx + ABS(x1-x0)+1; a++)
			SET_IMAGE_BYTE(pRow, a, (UInt8) val);
		    break;
		}
	    }
	    else
	    {
		/* Vertical line */
		x0 += offx;
		
		switch (d)
		{
		case 2:
		    for (a = 0; a < ABS(y1-y0)+1; pRow += pitch, a++)
			SET_IMAGE_2BITS(pRow, x0, val);
		    break;
		case 4:
		    for (a = 0; a < ABS(y1-y0)+1; pRow += pitch, a++)
			SET_IMAGE_4BITS(pRow, x0, val);
		    break;
		case 8:
		    for (a = 0; a < ABS(y1-y0)+1; pRow += pitch, a++)
			SET_IMAGE_BYTE(pRow, x0, (UInt8) val);
		    break;
		}
	    }
	}

	NORMAL_RETURN;
    }


    /* Handle diagonal lines */

    /* We'll use Bresenham's line-drawing algorithm here.
     * This type of line drawing plots a series of points (x,y)
     * either diagonally or laterally with respect to the previous point.
     * [Note: This means lines are 4-connected].
     * Movement from point pixel to point pixel alwasy proceeds in a
     * dominant lateral direction (parellel to the axis of the larger
     * dimension).
     * The algorithm (since the lines can't be horizonal or vertical at this
     * point) also increments (at least once) in the direction with the
     * smaller dimension, producing a diagonal (minor direction) movement
     */

    /* Make x0 always be < x1 */
    if (x0 > x1)
    {
	/* Swap the endpoints */
	a  = x1;
	x1 = x0;
	x0 = a;

	a  = y1;
	y1 = y0;
	y0 = a;
    }
    

    /* Make the axis with the largest extent have dimension 'a'
     * Make the axis with the least extent have dimension 'b'
     */
    if (ABS(y1 - y0) > ABS(x1 - x0))
    {
	a = ABS(y1 - y0); /* Major dimension (in Y) */
	b = ABS(x1 - x0); /* Minor dimension (in X) */

	/* Dominant (major) direction move */
	xinc = 0;
	yinc = (y1 > y0 ? pitch : -pitch);
    }
    else
    {
	a = ABS(x1 - x0); /* Major dimension (in X) */
	b = ABS(y1 - y0); /* Minor dimension (in Y) */

	/* Dominant (major) direction move */
	xinc = 1;
	yinc = 0;
    }
    

    /* Count register, the number of pixels we will draw in making the line */
    count = a + 1;

    /* Advance to image row which the first pixel's y value is on */
    x0 += offx;
    pRow = (UInt32 *)(pr) + (y0 + offy) * pitch;

    if (y1 < y0)
    {
	/* If y is decreasing then modify the pitch to make us move UP
	 * a row when moving in the diagonal (minor) direction
	 * (If y is increasing, we move DOWN a row)
	 */
	pitch *= -1;
	minorThresh = 1;
    }
    else
	minorThresh = 0;
	
    /* Initialize decision variable
     * Whenever this is >= minorThresh (0 when y increasing, 1 o.w.)
     * we will move in the diagonal (minor) direction.
     * This acts as the error term for where we are moving pixel-wise
     * vs. where we are relative to the true line
     */
    decide = 2*b - a;

    /* Now double the a and b values and set them to the needed increment
     * values below
     */
    a <<= 1;
    b <<= 1;
    a = b - a;

    if (d == 1 && xinc == 1)
    {
	/* Special optimization for binary image when major dimension is
	 * along X axis
	 */
	
	/* Ensure value is 0 or 1 */
	val = val ? 1 : 0;

	/* Make xinc now be the right endpoint in a horizontal line segment
	 * that we will draw at the current row whenever we get to a move
	 * in the diagonal (minor) direction
	 */
	for (xinc = x0; count > 0; count--)
	{
	    if (decide >= minorThresh)
	    {
		/* Move in the diagonal (minor) direction
		 * Move right in x one pixel and UP/DOWN one row in y
		 */
		i_rasterHorizLine (pIPState, pRow, x0, xinc, (Int8) val);
		
		decide += a;
		x0 = ++xinc;
		pRow += pitch;
	    }
	    else
	    {
		/* Move in the dominant (major) direction
		 * Move right in x one pixel
		 */
		decide += b;
		xinc++;
	    }
	}

	/* Now let's make sure we are actually where we should be and
	 * double check to be sure.  These checks will ensure we wrote
	 * (x1,y1)
	 */
	xinc--;
	if (xinc != x1 + offx)
	    return ia_internal;

	if (x0-1 != x1 + offx)
	{
	    /* Means we have to flush the last horizontal line segment left
	     * since we didn't have a diagonal move at the end
	     */
	    i_rasterHorizLine (pIPState, pRow, x0, xinc, (Int8) val);
	    x0 = xinc;

	    if (x0 != x1 + offx)
		return ia_internal;
	}
    }
    else
    {
	for ( ; count > 0; count--)
	{
	    switch (d)
	    {
	    case 1:
		SET_IMAGE_1BIT(pRow, x0, val);
		break;
	    case 2:
		SET_IMAGE_2BITS(pRow, x0, val);
		break;
	    case 4:
		SET_IMAGE_4BITS(pRow, x0, val);
		break;
	    case 8:
		SET_IMAGE_BYTE(pRow, x0, (UInt8) val);
		break;
	    }

	    if (decide >= minorThresh)
	    {
		/* Move in the diagonal (minor) direction
		 * Move right in x one pixel and UP/DOWN one row in y
		 */
		decide += a;
		x0++;
		pRow += pitch;
	    }
	    else
	    {
		/* Move in the dominant (major) direction
		 * Move right in x one pixel OR UP/DOWN one row in y
		 */
		decide += b;
		x0 += xinc;
		pRow += yinc;
	    }
	}
    }

    NORMAL_RETURN;
}


/*

$Log: i_rastln.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.6  1998/01/13 23:49:18  ddavies
i_rastln.c -> /product/ipcore/ipshared/src/RCS/i_rastln.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1996/01/03 20:12:48  ejaquith
ejaquith on Wed Jan 3 12:11:39 PST 1996

i_rastln.c -> /product/ipcore/ipshared/src/RCS/i_rastln.c,v

Update private and static routine "drawHorizLine()" to
public routine i_rasterHorizLine() for drawing line segments in binary
images quickly.  This already gets indirectly regression tested by
being called from i_rasterLine() on binary images.

 * Revision 1.4  1995/12/13  02:45:24  ejaquith
 * i_rastln.c -> /product/ipcore/ipshared/src/RCS/i_rastln.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.3  1995/10/18  03:58:07  ddavies
 * i_rastln.c -> /product/ipcore/ipshared/src/RCS/i_rastln.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.2  1995/04/22  19:25:37  ejaquith
 * ejaquith on Sat Apr 22 12:23:55 PDT 1995
 *
 * i_rastln.c -> /product/ipcore/ipshared/src/RCS/i_rastln.c,v
 *
 * Fixed bug in drawing line where non-binary images with the first
 * endpoint's y coord > second endpoint's y coord was drawing the wrong
 * way (oops) on account of a positive pitch.
 *
 * Revision 1.1  1995/04/15  01:24:55  ejaquith
 * Initial revision
 *

*/
