/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_rastop.c: C implementation of rasterOp procedures
 *
 *          rasterOp (bitblt) for 32 bit words:
 *                  Int32	i_rasterOpFull()
 *
 *
 */

/*
 * This file contains homegrown routines that implement the raster operations
 * supported by Sun's pixrect library.
 * 
 * The main routine in this file is i_rasterOpFull().  It contains many
 * subroutines for processsing the various operations that can be requested.
 *
 * Source and destination pixrects don't need to begin on cROpUnit boundaries.
 * The routines in this file generally process the leftmost cROpUnit or partial
 * cROpUnit of a destination line, then process the full cROpUnits of the
 * destination line, and finally process the final partial cROpUnit of
 * the destination.  They shift and mask source bits as necessary to bring
 * them to the cROpUnit boundaries required for the destination.
 * 
 * To improve execution speed, each operation is implemented with several
 * routines to deal with the special cases where no shifting and masking
 * are necessary.
 *
 * 6/5/92: Make halfhearted attempt to deal with non-binary images.  Multiply
 *	  incoming dprw, dproffx, dx and dw by dd (depth).  Similarly,
 *	  multiply sprw, sx and sproffx by dd.  Ignore sd.
 *	  This won't handle source and dest images of different depths.
 *
 * DANGER! DANGER! DANGER!
 * A great many variables are passed between the high level procedures and
 * the lower level procs using a struct called rasterOpData.  This struct
 * lives on the stack of the high level procedure.  Passing pointers to things
 * on a procedure's stack can be a source of errors.  However, in this case
 * the called procs will return before the higher level proc returns and the
 * struct vanishes.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "memory.pub"

    /* prototypes */
#include "shrrast.pub"
#include "shrrast.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_rastop.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");


/* index of middle entry in masks array.  Note that there are extra 0 entries
 * before the real mask entries in both halves of the table.
 */
#define	cMaskOffset	cROpMaxBitsPerUnit + 1

#define	cOpMask		0x1e	/* mask for operation code in op arg't */

#define	EDGE_OP( d, s, m)	( (d & ~m) | (s & m) )

/* The code in word, ip_doCopySprNoRemPart2
 * ip_doCopySprNoRemFull2 and ip_doCopySprNoRem3Words has only been
 * tested with cROpUnit set to UInt32.  I think that it might work
 * if cROpUnit is changed but I wouldn't guarantee it.  As a result,
 * the code is conditionally compiled so that if cROpUnitSIZE is not 32,
 * those 4 procedures are not called and ip_doCopySprNoRem is called
 * instead just as the original author wrote it.
 */
#define cSizeOfUInt32		32
#define cROpUnit		UInt32
#define cROpUnitSIZE		cSizeOfUInt32
#define cROpBitsPerUnit		cROpUnitSIZE
#define cROpBytesPerUnit	( cROpBitsPerUnit / 8 )
#define cROpUnitAddrMask	(~(cROpBytesPerUnit - 1) )
#define cROpMaxBitsPerUnit	32


/* LOCAL */
/* NOTE: we cannot declare this table to be "const" because of a bad
 * interaction between gcc and Purify.  gcc generates relocation records
 * relative to the beginning of the table, in particular it can generate
 * negative ones.  The stuff before this table is some dreck.  Purify
 * moves that thing around, possibly changing the distance between that
 * thing and the beginning of the table.  It then modifies the opcodes
 * that reference the table to reference the dreck.
 * By eliminating "const" we are trying to move this table out of the
 * code segment so Purify won't mess with it.
 */
static UInt32 masks[] = {
  0,
  0x80000000, 0xc0000000, 0xe0000000, 0xf0000000,
  0xf8000000, 0xfc000000, 0xfe000000, 0xff000000,
  0xff800000, 0xffc00000, 0xffe00000, 0xfff00000,
  0xfff80000, 0xfffc0000, 0xfffe0000, 0xffff0000,
  0xffff8000, 0xffffc000, 0xffffe000, 0xfffff000,
  0xfffff800, 0xfffffc00, 0xfffffe00, 0xffffff00,
  0xffffff80, 0xffffffc0, 0xffffffe0, 0xfffffff0,
  0xfffffff8, 0xfffffffc, 0xfffffffe, 0xffffffff,
  0,
  0x00000001, 0x00000003, 0x00000007, 0x0000000f,
  0x0000001f, 0x0000003f, 0x0000007f, 0x000000ff,
  0x000001ff, 0x000003ff, 0x000007ff, 0x00000fff,
  0x00001fff, 0x00003fff, 0x00007fff, 0x0000ffff,
  0x0001ffff, 0x0003ffff, 0x0007ffff, 0x000fffff,
  0x001fffff, 0x003fffff, 0x007fffff, 0x00ffffff,
  0x01ffffff, 0x03ffffff, 0x07ffffff, 0x0fffffff,
  0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff
  };


/*
 * BmProduct is used when doing multiplications involving pixrects,
 * and casts its arguments so that the compiler will use 16 by 16 multiplies.
 */
#ifndef BM_PRODUCT
#if defined(sun) && !defined(sparc)
#define   BM_PRODUCT(a, b)   ((Int16)(a) * (Int16)(b))
#else
#define   BM_PRODUCT(a, b)   ((a) * (b))
#endif
#endif


/* PUBLIC */
/* i_rasterOpFull()
 *
 *	Input:
 *		dpr =	pointer to desination image
 *		dprw =	width of the entire destination image in pixels
 *		dprh = 	height of the entire destination image
 *		dproffx = x offset value of the destination image
 *		dproffy = y offset value of the destination image
 *		dx =	x-coordinate of leftmost pixel in destination rect
 *		dy =	y-coordinate of topmost  pel in destination rect
 *		dw =	width of destination rectangle in pixels
 *		dh = 	height of destination rectangle
 *		dd =	bits/pixel of destination
 *		dLinebytes = bytes/line of destination
 *		operation = operation code.  see defines.h
 *		spr =	pointer to source image
 *		sprw = width of the entire source image in pixels
 *		sprh =	height of the entire source image
 *		sx =	x-coordinate of leftmost pixel in source rect
 *		sy = 	y-coordinate of leftmost pixel source image
 *		sproffx = x offset value of the source image
 *		sproffy = y offset value of the source image
 *		sd =	bits/pixel of source
 *		sLinebytes = bytes/line of source
 *
 *	Returns:
 *	ia_successful:		It worked
 *	ia_depthNotSupported:	The source and dest depth parameters
 *				either had the same illegal value or
 *				were not equal to each other
 *	ia_invalidParm:		dw or dh were negative
 */
Int32 CDECL 
i_rasterOpFull (
    IPState    *pIPState,
    UInt8      *dpr,		/* desination image */
    Int32	dprw,		/* width of the entire destination image */
    Int32	dprh,		/* height of the entire destination image  */
    Int32	dproffx,	/* x offset value of the destination image */
    Int32	dproffy,	/* y offset value of the destination image */
    Int32	dx,		/* x-coordinate of leftmost pel in dest rect */
    Int32	dy,		/* y-coordinate of topmost  pel in dest rect */
    Int32	dw,		/* width of destination rectangle (in pels) */
    Int32	dh,		/* height of destination rectangle (in pels) */
    Int32	dd,		/* depth of destination in bits */
    Int32	dLinebytes,
    Int32	operation,	/* operation code */
    UInt8      *spr,		/* source image */
    Int32	sprw,
    Int32	sprh,
    Int32	sx,
    Int32	sy,
    Int32	sproffx,
    Int32	sproffy,
    Int32	sd,		/* depth of source in bits  */
    Int32	sLinebytes)

{

  PROLOGUE;
  Int32		 basex;
  Int32		 basey;
  RasterData	 rasterOpData;	/* holds all data for a particular call
				 * to i_rasterOpFull.  This allows this
				 * routine to be reentrant
				 */

  /* If the depth of the source and the destination are unequal or illegal,
   * we have no idea what the caller wanted, so return an error code.
   * if the source was null, 0 was passed in for the depth, etc. so
   * check for this case.
   */
  if ( ((spr != NULL) && (sd != dd)) ||
       ((dd != 1) && (dd != 2) && (dd != 4) && (dd != 8) && (dd != 16)) )
      return ia_depthNotSupported;

  /* Make a halfhearted attempt to deal with gray images.  Multiply
   * the incoming dprw, dproffx, dx and dw by dd (the dest depth).  Also
   * multiply sprw, sx and sproffx by sd.  Ignore sd.
   * This won't handle source and dest images of different depths.
   */
  if (dd != 1)
  {
    switch (dd)
    {
    case 2:
      dprw <<= 1;
      dproffx <<= 1;
      dx <<= 1;
      dw <<= 1;
      sprw <<= 1;
      sx <<= 1;
      sproffx <<= 1;
      break;

    case 4:
      dprw <<= 2;
      dproffx <<= 2;
      dx <<= 2;
      dw <<= 2;
      sprw <<= 2;
      sx <<= 2;
      sproffx <<= 2;
      break;

    case 8:
      dprw <<= 3;
      dproffx <<= 3;
      dx <<= 3;
      dw <<= 3;
      sprw <<= 3;
      sx <<= 3;
      sproffx <<= 3;
      break;

    case 16:
      dprw <<= 4;
      dproffx <<= 4;
      dx <<= 4;
      dw <<= 4;
      sprw <<= 4;
      sx <<= 4;
      sproffx <<= 4;
      break;

    default:
      dprw *= dd;
      dproffx *= dd;
      dx *= dd;
      dw *= dd;
      sprw *= dd;
      sx *= dd;
      sproffx *= dd;
      break;
    }   /* end of depth switch */
  }	/* end of depth adjustment */


  /* If dx or dy is negative, we adjust sx and dw or
     sy and dh, and set dx or dy to 0.
     If sx or sy is negative, we adjust dx and dw or
     dy and dh, and set sx or sy to 0.
     (In other words, we make sure the top left corner
     of the rectangle is within both pixrects.)
     This is ok for the case where we want to clip;
     for the other case, we must adjust the pointers
     to the data.
     
     This code clips to the indicated pixrect if clipping
     isn't inhibited.  If clipping is inhibited, it clips
     to the primary pixrect's top and left boundary.
     */
  if ( !( operation & PIX_DONTCLIP ) ) {
    if ( dx < 0 ) {
      sx -= dx;
      dw += dx;
      dx  = 0;
    }
    if ( dy < 0 ) {
      sy -= dy;
      dh += dy;
      dy  = 0;
    }
    if ( sx < 0 ) {
      dx -= sx;
      dw += sx;
      sx  = 0;
    }
    if ( sy < 0 ) {
      dy -= sy;
      dh += sy;
      sy  = 0;
    }
    basex = dproffx + dx;
    rasterOpData.dstInitOverhang = cROpUnitSIZE - ( basex % cROpUnitSIZE );
  }
  else {
    if ( (basex = dx + dproffx) < 0 ) {
      sx -= basex;
      dw += basex;
      dx -= basex;
      basex = 0;
    }
    if ( (basey = dy + dproffy) < 0 ) {
      sy -= basey;
      dh += basey;
      dy -= basey;
      basey = 0;
    }

    rasterOpData.dstInitOverhang = cROpUnitSIZE - ( basex % cROpUnitSIZE );
  }

    /* Move the checks to here from above.  We want to check the
     * width and height AFTER they're adjusted by the clipping
     * process.
     */
  if ((dw == 0) || (dh == 0))
	NORMAL_RETURN;		/* nothing to do */

  if ((dw < 0) || (dh < 0))
	return ia_invalidParm;		/* meaningless */

  if ( sx >= 0) rasterOpData.srcInitOverhang = cROpUnitSIZE - ( sx % cROpUnitSIZE );
  else {
    if ( sx % cROpUnitSIZE ) 
      rasterOpData.srcInitOverhang = cROpUnitSIZE - 
	( cROpUnitSIZE +  sx % cROpUnitSIZE );
    else	rasterOpData.srcInitOverhang = cROpUnitSIZE;
  }

  if ( rasterOpData.dstInitOverhang > rasterOpData.srcInitOverhang ) {
    rasterOpData.srcRemainder = rasterOpData.dstInitOverhang - rasterOpData.srcInitOverhang;
    
    rasterOpData.srcMiddleOverhang = cROpUnitSIZE - rasterOpData.srcRemainder;
  }
  else if ( rasterOpData.dstInitOverhang < rasterOpData.srcInitOverhang ) {
    rasterOpData.srcMiddleOverhang = rasterOpData.srcInitOverhang - rasterOpData.dstInitOverhang;
    rasterOpData.srcRemainder       = cROpUnitSIZE - rasterOpData.srcMiddleOverhang;
  }
  else {
    rasterOpData.srcRemainder = 0;
    rasterOpData.srcMiddleOverhang = cROpUnitSIZE;
  }

  if ( !( operation & PIX_DONTCLIP ) ) {
    /* make sure the bottom right corner of the rectangle
       is within both pixrects.
       */
    rasterOpData.width = dprw - dx;
    if ( rasterOpData.width > dw ) rasterOpData.width = dw;
    if ( spr && (rasterOpData.width > sprw - sx) ) 
      rasterOpData.width = sprw - sx;

    rasterOpData.lineLimit = dprh - dy;
    if ( rasterOpData.lineLimit > dh ) rasterOpData.lineLimit = dh;
    if ( spr && (rasterOpData.lineLimit > sprh - sy) ) 
      rasterOpData.lineLimit = sprh - sy;
  }
  else	{
    rasterOpData.width = dw;
    rasterOpData.lineLimit = dh;
  }

  if ( rasterOpData.width >= rasterOpData.dstInitOverhang ) {
    rasterOpData.fullUnits    = ( rasterOpData.width -
				  rasterOpData.dstInitOverhang ) / cROpUnitSIZE;
    rasterOpData.leftoverBits = ( rasterOpData.width -
				  rasterOpData.dstInitOverhang ) % cROpUnitSIZE;

    rasterOpData.dstInitMask   = masks[ cMaskOffset +
					rasterOpData.dstInitOverhang ];
    rasterOpData.dstFinalMask  = masks[ cMaskOffset - cROpUnitSIZE +
					rasterOpData.leftoverBits - 1 ];
  }
  else {
    rasterOpData.fullUnits    = 0;
    rasterOpData.leftoverBits = 0;
    
    rasterOpData.dstInitMask   = masks[ cMaskOffset +
					rasterOpData.dstInitOverhang ] &
				~masks[ cMaskOffset +
					rasterOpData.dstInitOverhang -
					rasterOpData.width];
    rasterOpData.dstFinalMask  = rasterOpData.dstInitMask;
  }
  
  rasterOpData.srcLptr = ( spr ) ? (spr + BM_PRODUCT(sLinebytes, sproffy + sy) + (((sproffx + sx) >> 3) & cROpUnitAddrMask)) : NULL;
  rasterOpData.dstLptr = dpr + BM_PRODUCT(dLinebytes, dproffy + dy) + (((dproffx + dx) >> 3) & cROpUnitAddrMask);
                       /* mprdAddr( mprD( dpr ), dx, dy ); */
  rasterOpData.srcLinebytes = ( spr ) ? sLinebytes : 0;
  rasterOpData.dstLinebytes = dLinebytes;
  
  if ( spr == NULL ) 
    rasterOpData.srcValue = ( PIX_OPCOLOR( operation ) ) ? 
      masks[ cMaskOffset + cROpUnitSIZE ] : 0;
  
    /* Set the offset of the rightmost word in a scanline.  This
     * is used to prevent us from reading one word too many.
     */
  rasterOpData.srcRWordOffset = ((rasterOpData.width - 1 -
				  rasterOpData.srcInitOverhang) >> 5) + 1;
  

  switch( operation & cOpMask ) {
  case PIX_CLR:	ip_doClr (pIPState, &rasterOpData); break;
    
  case PIX_NOT(PIX_DST | PIX_SRC):
    ip_doOp1 (pIPState,  &rasterOpData ); break;
    
  case PIX_DST & PIX_NOT(PIX_SRC):
    ip_doOp2 (pIPState,  &rasterOpData ); break;
    
  case PIX_NOT(PIX_SRC):
    ip_doOp3 (pIPState,  &rasterOpData ); break;

  case PIX_NOT(PIX_DST) & PIX_SRC:
    ip_doOp4 (pIPState,  &rasterOpData ); break;

  case PIX_NOT(PIX_DST):
    ip_doOp5 (pIPState,  &rasterOpData ); break;
    
  case PIX_DST ^ PIX_SRC:
    ip_doXor (pIPState,  &rasterOpData ); break;
    
  case PIX_NOT(PIX_DST & PIX_SRC):
    ip_doOp7 (pIPState,  &rasterOpData ); break;

  case PIX_DST & PIX_SRC:
    ip_doAnd (pIPState,  &rasterOpData ); break;
    
  case PIX_DST ^ PIX_NOT( PIX_SRC ): 
    ip_doOp9 (pIPState,  &rasterOpData ); break;
    
  case PIX_DST: 
    break;
    
  case PIX_DST | PIX_NOT( PIX_SRC ):
    ip_doOp11 (pIPState,  &rasterOpData ); break;

  case PIX_SRC:
    ip_doCopy (pIPState,  &rasterOpData ); break;

  case PIX_NOT( PIX_DST ) | PIX_SRC:
    ip_doOp13 (pIPState,  &rasterOpData ); break;
    
  case PIX_DST | PIX_SRC:
    ip_doOr (pIPState,  &rasterOpData ); break;

  case PIX_SET:
    ip_doSet (pIPState,  &rasterOpData );
    break;

    default:
    break;
  }

  NORMAL_RETURN;
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doAnd (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

  /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doAndSprRem (pIPState, rasterOpData);
    else
	ip_doAndSprNoRem (pIPState, rasterOpData);
  }
  else ip_doAndNullSpr (pIPState, rasterOpData);
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doAndNullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  

  /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  sv = rasterOpData->srcValue;
  for ( line = 0; line <rasterOpData->lineLimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP8( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    {
        register UInt32 *pDst, *pEnd;

        /* 
          Calculate end addr explicitly and assign to a local
          variable. Otherwise SPARC insists on reloading fullUnits
          and recalculating end address every time thru loop.
        */
        pEnd = dstptr + rasterOpData->fullUnits;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP8( *pDst, sv);
        }
        dstptr = pDst; /* Update dtsptr to proper position */
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if (rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP8( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doAndSprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  cROpUnit	 sv;
  Int32		 ls;
  Int32		 rs;
  

  /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls = rasterOpData->srcRemainder;
  rs = rasterOpData->srcMiddleOverhang;
  
  for ( line = 0; line < rasterOpData->lineLimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP8( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + rasterOpData->fullUnits;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROP8( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/

    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROP8( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doAndSprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  
  /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  for ( line = 0; line < rasterOpData->lineLimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    rasterOpData->srcValue = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP8( *dstptr, rasterOpData->srcValue), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + rasterOpData->fullUnits;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP8( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP8( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doClr (
    IPState    *pIPState,
    RasterData *rasterOpData) 
{
  PROLOGUE;
  cROpUnit	*dstptr;
  Int32		 line;
  
  /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  for ( line = 0; line < rasterOpData->lineLimit; ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, 0, rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    memset( (char *)dstptr, 0, rasterOpData->fullUnits * cROpBytesPerUnit );
    
    dstptr += rasterOpData->fullUnits;
    
    /* finally process the final (partial) unit in the destination*/

    if ( rasterOpData->leftoverBits ) 
      *dstptr = EDGE_OP( *dstptr, 0, rasterOpData->dstFinalMask );
    
  }
}


/*--------------------------------------------------------------------------*/

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doCopy (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder ) 
	ip_doCopySprRem (pIPState,  rasterOpData );
#if cROpUnitSIZE == cSizeOfUInt32
    else if ( rasterOpData ->fullUnits > 1)
	ip_doCopySprNoRem (pIPState,  rasterOpData );
    else if ( rasterOpData ->leftoverBits == 0 )
	{
	if ( rasterOpData ->fullUnits == 0 )
	    ip_doCopySprNoRemOneWord (pIPState,  rasterOpData );
	else
	    ip_doCopySprNoRemFull2 (pIPState,  rasterOpData );
	}
    else
	{
	if ( rasterOpData ->fullUnits == 0 )
	    ip_doCopySprNoRemPart2 (pIPState,  rasterOpData );
	else
	    ip_doCopySprNoRem3Words (pIPState,  rasterOpData );
	}
#else
    else ip_doCopySprNoRem (pIPState,  rasterOpData );
#endif
  }
  else ip_doCopyNullSpr (pIPState,  rasterOpData );
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doCopyNullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData) 
{

  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  sv = rasterOpData->srcValue;
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, sv, rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = sv;
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, sv, rasterOpData->dstFinalMask );
    }
  }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doCopySprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  cROpUnit	 sv;
  Int32		 ls;
  Int32		 rs;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, sv, rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
        sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, sv, rasterOpData->dstFinalMask );
    }
  }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doCopySprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 llimit;
  cROpUnit	 sv;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, sv, rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    memcpy( (char *)dstptr, (char *)srcptr, rasterOpData->fullUnits * cROpBytesPerUnit );
    dstptr += rasterOpData->fullUnits;
    srcptr += rasterOpData->fullUnits;
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, *srcptr, rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doCopySprNoRemOneWord (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 llimit;
  Int32		 srcLinebytes;
  Int32		 dstLinebytes;
  Int32		 mask;

    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  llimit = rasterOpData->lineLimit;
  srcptr = (cROpUnit *)rasterOpData->srcLptr;
  dstptr = (cROpUnit *)rasterOpData->dstLptr;
  srcLinebytes = rasterOpData->srcLinebytes;
  dstLinebytes = rasterOpData->dstLinebytes;
  mask = rasterOpData->dstInitMask;
  
  for ( line = llimit; line > 0; 
       --line, srcptr = (cROpUnit *) (((UInt8 *)srcptr) + srcLinebytes), dstptr = (cROpUnit *) (((UInt8 *)dstptr) + dstLinebytes) ) 
   {
    
    *dstptr = EDGE_OP( *dstptr, *srcptr, mask );
    
   }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doCopySprNoRemPart2 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 llimit;
  Int32		 srcLinebytes;
  Int32		 dstLinebytes;
  Int32		 initMask;
  Int32		 finalMask;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  llimit = rasterOpData->lineLimit;
  srcptr = (cROpUnit *)rasterOpData->srcLptr;
  dstptr = (cROpUnit *)rasterOpData->dstLptr;
  srcLinebytes = rasterOpData->srcLinebytes;
  dstLinebytes = rasterOpData->dstLinebytes;
  initMask = rasterOpData->dstInitMask;
  finalMask = rasterOpData->dstFinalMask;
 
  for ( line = llimit; line > 0; 
       --line, srcptr = (cROpUnit *) (((UInt8 *)srcptr) + srcLinebytes), dstptr = (cROpUnit *) (((UInt8 *)dstptr) + dstLinebytes) ) 
    {
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, *srcptr, initMask );
    
    /* finally process the final (partial) unit in the destination*/
    
    *(dstptr+1) = EDGE_OP( *(dstptr+1), *(srcptr+1), finalMask );
    }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doCopySprNoRemFull2 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 llimit;
  Int32		 srcLinebytes;
  Int32		 dstLinebytes;
  Int32		 mask;

    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  llimit = rasterOpData->lineLimit;
  srcptr = (cROpUnit *)rasterOpData->srcLptr;
  dstptr = (cROpUnit *)rasterOpData->dstLptr;
  srcLinebytes = rasterOpData->srcLinebytes;
  dstLinebytes = rasterOpData->dstLinebytes;
  mask = rasterOpData->dstInitMask;
  
  for ( line = llimit; line > 0; 
       --line, srcptr = (cROpUnit *) (((UInt8 *)srcptr) + srcLinebytes), dstptr = (cROpUnit *) (((UInt8 *)dstptr) + dstLinebytes) ) 
   {
    
    *dstptr = EDGE_OP( *dstptr, *srcptr, mask );
    *(dstptr+1) = *(srcptr+1);
    
   }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doCopySprNoRem3Words (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 llimit;
  Int32		 srcLinebytes;
  Int32		 dstLinebytes;
  Int32		 initMask;
  Int32		 finalMask;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  llimit = rasterOpData->lineLimit;
  srcptr = (cROpUnit *)rasterOpData->srcLptr;
  dstptr = (cROpUnit *)rasterOpData->dstLptr;
  srcLinebytes = rasterOpData->srcLinebytes;
  dstLinebytes = rasterOpData->dstLinebytes;
  initMask = rasterOpData->dstInitMask;
  finalMask = rasterOpData->dstFinalMask;
 
  for ( line = llimit; line > 0; 
       --line, srcptr = (cROpUnit *) (((UInt8 *)srcptr) + srcLinebytes), dstptr = (cROpUnit *) (((UInt8 *)dstptr) + dstLinebytes) ) 
    {
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, *srcptr, initMask );

    /* process the one full unit */
    *(dstptr+1) = *(srcptr+1);
    
    /* finally process the final (partial) unit in the destination*/
    
    *(dstptr+2) = EDGE_OP( *(dstptr+2), *(srcptr+2), finalMask );
    }
}
/*--------------------------------------------------------------------------*/

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOrSprRem (pIPState,  rasterOpData );
    else
	ip_doOrSprNoRem (pIPState,  rasterOpData );
  }
  else ip_doOrNullSpr (pIPState,  rasterOpData );
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOrNullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < rasterOpData->lineLimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, (sv | *dstptr), rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst |= sv;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if (rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, (sv | *dstptr), rasterOpData->dstFinalMask );
    }
  }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOrSprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  cROpUnit	 sv;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  
  for ( line = 0; line < rasterOpData->lineLimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, (sv | *dstptr), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst |= ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, (sv | *dstptr),
			rasterOpData->dstFinalMask );
    }
  }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOrSprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  cROpUnit	 sv;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, (sv | *dstptr), rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst |= *srcptr++;
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, (*srcptr | *dstptr),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doSet (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, 0xffffffff, rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the dest'n */
    
    memset( (char *)dstptr, 0xff, rasterOpData->fullUnits * cROpBytesPerUnit );
    
    dstptr += rasterOpData->fullUnits;
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
      *dstptr = EDGE_OP( *dstptr, 0xffffffff, 
			rasterOpData->dstFinalMask);
  }
}


/*--------------------------------------------------------------------------*/

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doXor (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doXorSprRem (pIPState,  rasterOpData );
    else
	ip_doXorSprNoRem (pIPState,  rasterOpData );
  }
  else ip_doXorNullSpr (pIPState,  rasterOpData );
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doXorNullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  sv     = rasterOpData->srcValue;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP6( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP6( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP6( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doXorSprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ls;
  Int32		 rs;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROP6( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROP6( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROP6( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}

/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doXorSprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  cROpUnit	 sv;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP6( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP6( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP6( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp1 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp1SprRem (pIPState,  rasterOpData );
    else
	ip_doOp1SprNoRem (pIPState,  rasterOpData );
  }
  else ip_doOp1NullSpr (pIPState,  rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp1NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32	 	 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  sv     = rasterOpData->srcValue;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP1( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP1( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP1( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp1SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROP1( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROP1( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROP1( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp1SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP1( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP1( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP1( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}

/*--------------------------------------------------------------------------*/


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp2 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

      /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp2SprRem (pIPState,  rasterOpData );
    else
	ip_doOp2SprNoRem (pIPState,  rasterOpData );
  }
  else ip_doOp2NullSpr (pIPState,  rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp2NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP2( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP2( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP2( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp2SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  cROpUnit	 sv;
  Int32	llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROP2( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROP2( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROP2( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp2SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP2( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP2( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP2( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/



/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp3 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

      /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp3SprRem (pIPState,  rasterOpData );
    else
	ip_doOp3SprNoRem (pIPState,  rasterOpData );
  }
  else ip_doOp3NullSpr (pIPState,  rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp3NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP3( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP3( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP3( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp3SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  cROpUnit	 sv;
  Int32	llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROP3( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROP3( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROP3( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp3SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{

  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP3( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP3( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP3( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}



/*--------------------------------------------------------------------------*/



/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp4 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

      /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp4SprRem (pIPState, rasterOpData );
    else
	ip_doOp4SprNoRem (pIPState, rasterOpData );
  }
  else ip_doOp4NullSpr (pIPState, rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp4NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP4( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP4( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP4( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}



/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp4SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROP4( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROP4( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROP4( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp4SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP4( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP4( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP4( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/



/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp5 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

      /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp5SprRem (pIPState, rasterOpData );
    else
	ip_doOp5SprNoRem (pIPState, rasterOpData );
  }
  else ip_doOp5NullSpr (pIPState, rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp5NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP5( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP5( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP5( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp5SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = 0;		/* unimportant value, since source is 
			   ignored on this operation ( ~dst) */
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP5( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP5( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP5( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp5SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = 0;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP5( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP5( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP5( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/



/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp7 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

      /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp7SprRem (pIPState, rasterOpData );
    else
	ip_doOp7SprNoRem (pIPState, rasterOpData );
  }
  else ip_doOp7NullSpr (pIPState, rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp7NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP7( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP7( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP7( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp7SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROP7( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROP7( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROP7( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp7SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP7( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP7( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP7( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/



/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp9 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp9SprRem (pIPState, rasterOpData );
    else
	ip_doOp9SprNoRem (pIPState, rasterOpData );
  }
  else ip_doOp9NullSpr (pIPState, rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp9NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP9( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP9( *pDst, sv);
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP9( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp9SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROP9( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROP9( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROP9( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp9SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROP9( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROP9( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROP9( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/



/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp11 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

      /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp11SprRem (pIPState, rasterOpData );
    else
	ip_doOp11SprNoRem (pIPState, rasterOpData );
  }
  else ip_doOp11NullSpr (pIPState, rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp11NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROPB( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROPB( *pDst, sv);
        }
        dstptr = pDst;
    }
    

    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROPB( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp11SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROPB( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROPB( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROPB( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp11SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    sv = *srcptr++;
    
    *dstptr = EDGE_OP( *dstptr, OP_UROPB( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROPB( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROPB( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}


/*--------------------------------------------------------------------------*/



/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp13 (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;

    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  if ( rasterOpData->srcLptr ) {
    if ( rasterOpData->srcRemainder )
	ip_doOp13SprRem (pIPState, rasterOpData );
    else
	ip_doOp13SprNoRem (pIPState, rasterOpData );
  }
  else ip_doOp13NullSpr (pIPState, rasterOpData );
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp13NullSpr (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	 sv;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  sv     = rasterOpData->srcValue;
  
  for ( line = 0; line < llimit; ++line, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    *dstptr = EDGE_OP( *dstptr, OP_UROPD( *dstptr, sv),
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROPD( *pDst, sv);
        }
        dstptr = pDst;
    }
    

    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROPD( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp13SprRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*srcRptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  Int32		 ls;
  Int32		 rs;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ls     = rasterOpData->srcRemainder;
  rs     = rasterOpData->srcMiddleOverhang;
  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    srcRptr = srcptr + rasterOpData->srcRWordOffset;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */
    
    if ( rasterOpData->dstInitOverhang > rasterOpData->srcInitOverhang )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      ++srcptr;
    }
    else
    if ( rasterOpData->dstInitOverhang < rasterOpData->srcInitOverhang )
    {
      sv = *srcptr >>
        (rasterOpData->srcInitOverhang - rasterOpData->dstInitOverhang);
    }
    else sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROPD( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
            *pDst = OP_UROPD( *pDst, sv);
            ++srcptr;
        }
        dstptr = pDst;
    }
    

    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits )
    {
      if (srcRptr > srcptr)
	sv = ( *srcptr << ls ) | ( *(srcptr+1) >> rs);
      else
	sv = ( *srcptr << ls );
      *dstptr = EDGE_OP( *dstptr, OP_UROPD( *dstptr, sv),
			rasterOpData->dstFinalMask );
    }
  }
}


/* PRIVATE */
/* Internal i_rasterOp procedure */
static void CDECL 
ip_doOp13SprNoRem (
    IPState    *pIPState,
    RasterData *rasterOpData)
{
  PROLOGUE;
  cROpUnit	*srcptr;
  cROpUnit	*dstptr;
  Int32		 line;
  Int32		 ilimit;
  cROpUnit	 sv;
  Int32		 llimit;
  
    /* Keep compiler happy */
  pIPState = pIPState;
  __error__ = __error__;

  ilimit = rasterOpData->fullUnits;
  llimit = rasterOpData->lineLimit;
  
  for ( line = 0; line < llimit; 
       ++line, rasterOpData->srcLptr += rasterOpData->srcLinebytes, rasterOpData->dstLptr += rasterOpData->dstLinebytes ) {
    
    srcptr = (cROpUnit *)rasterOpData->srcLptr;
    dstptr = (cROpUnit *)rasterOpData->dstLptr;
    
    /* process the 1st (partial) unit in the destination */

    sv = *srcptr++;

    *dstptr = EDGE_OP( *dstptr, OP_UROPD( *dstptr, sv), 
			rasterOpData->dstInitMask );
    dstptr++;
    
    /* now process the middle full units in the destination */
    
    {
        register UInt32 *pDst, *pEnd;
        pEnd = dstptr + ilimit;
        for (pDst=dstptr; pDst<pEnd; pDst++)
        {
            *pDst = OP_UROPD( *pDst, *srcptr);
            ++srcptr;
        }
        dstptr = pDst;
    }
    
    
    /* finally process the final (partial) unit in the destination*/
    
    if ( rasterOpData->leftoverBits ) {
      *dstptr = EDGE_OP( *dstptr, OP_UROPD( *dstptr, *srcptr),
			rasterOpData->dstFinalMask );
    }
  }
}

/*

$Log: i_rastop.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.12  1998/02/10 19:48:06  ddavies
i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v

Add 16-bit/pixel to list of legal depths.  Also add a 16bpp case to the
switch statement that adjusts widths to account for depths.

Revision 1.11  1998/01/13 23:49:20  ddavies
i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.10  1997/08/02 02:16:00  ddavies
ddavies on Fri Aug 1 19:14:53 PDT 1997

i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v

Moved the checks for width and height BACK down to below the clipping
code.  I did this on June 25, but that change mysteriously vanished.
Grumble.

Revision 1.9  1997/08/01 01:25:47  ddavies
ddavies on Thu Jul 31 18:24:16 PDT 1997

i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v

Add one entry to the beginning of the masks table and change cMaskOffset
to account for it.  Our code was generating an offset of -1 into this
table.  Purify 4.0.1 caught it, but ONLY in the non-debug version of
code, which made it MUCH harder to find.

Revision 1.8  1995/12/21 04:34:45  ddavies
ddavies on Wed Dec 20 20:32:35 PST 1995

i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v

Use the new srcRWordOffset entry in the RasterData structure to check
to see if we're at the right edge of a scan line.  In routines that compose
each output word from two input words, we were always reading *srcptr and
*(srcptr+1).  If srcptr pointed to the last scan line word, this could
cause a seg fault or GP fault.  We now check before reading srcptr+1.

 * Revision 1.7  1995/12/13  02:45:26  ejaquith
 * i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.6  1994/11/07  22:05:49  ddavies
 * i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.5  1994/10/28  01:59:26  lperry
 * lperry on Thu Oct 27 18:57:36 PDT 1994
 *
 * i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v
 *
 * Oops. Blew it prev checkin. This is the real one.
 * Minor code tightening on the rasterop inner loops.
 * They now use ptr values for loop end tests rather
 * than a separate loop counter. Also moved
 * function args to local vars before using in loops.
 *
 * Revision 1.4  1994/07/30  00:19:39  ddavies
 * ddavies on Fri Jul 29 17:19:01 PDT 1994
 *
 * i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v
 *
 * Remove "const" from declaration of the array masks.  This is a workaround
 * for a buggy interaction between gcc and Purify.
 *
 * Revision 1.3  1994/02/13  00:51:18  ddavies
 * i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v
 *
 * Added tests for width and height being 0 or negative.  If negative,
 * i_rasterOpFull returns with ia_invalidParm.  If either is 0, it immediately
 * returns with ia_successful - it did what you asked, which was nothing.
 *
 * Revision 1.2  1994/02/04  20:15:51  dferrell
 * i_rastop.c -> /product/ipcore/ipshared/src/RCS/i_rastop.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/02/03  00:11:10  ddavies
 * Initial revision
 *
 * Revision 1.20  1993/12/02  19:26:05  ddavies
 * rasterop.c -> /product/ipcore/ipshared/src/RCS/rasterop.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.19  1993/11/05  01:53:22  ddavies
 * rasterop.c -> /product/ipcore/ipshared/src/RCS/rasterop.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


