/*****************************************************************
 *  Copyright (c) 1988, 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_reduce.c: contains calls to assembler-level subroutines for
 *
 *          2x threshold reduction (and cascades):
 *                  Int32	i_reduce16()
 *                  Int32	i_reduce8()
 *                  Int32	i_reduce4()
 *                  Int32	i_reduce2()
 *
 *	    IO threshold reduction (and cascades):
 *                  Int32	i_ioReduce16()
 *                  Int32	i_ioReduce8()
 *                  Int32	i_ioReduce4()
 *                  Int32	i_ioReduce2()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "shrip.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrrast.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_reduce.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");


/* This is the upper level of threshold reduction code.  It calls
 * assembly-level primitives (which might be written in C) to produce
 * each output line.
 *
 * There are four versions of assembly code, one for each legal threshold
 * value.  There are also assembly-level routines for 16x, 8x and 4x
 * reduction with thresholds of 1 and 4.
 *
 * Original implementations of the assembly-level primitives reduced the
 * entire image in one step.  The new implementations will produce a single
 * output line.  This allows us to easily use them on banded or tiled images
 * as well as on entire images.
 *
 * Column reductions will be done using a combination of mask, shift and
 * table lookup operations.  Each 32-bit word will be masked so that only
 * the even numbered bits remain.  The even bits in the upper half of the
 * 32-bit word will be shifted down and ORed into the holes left by the mask.
 * This puts the 16 output bits into the low word, but in the wrong order.
 * Instead of 15,14,13, ..., 2,1,0; we would have 15,7,14,6,...,9,1,8,0.
 * The table tabPermute is used to put the bits into the correct order.
 * tabPermute is constructed in bintable.c.
 *
 * *************  I/O REDUCTION  ***************
 *
 * These operations were added at the request of XIS.  They have very tight
 * memory requirements.  In fact, there is no guarantee that they can fit
 * even one full scale image in memory.  Therefore, they wish to reduce the
 * input image in bands as it arrives from the scanner or from disk.
 *
 * Images arriving from I/O devices generally don't have frames around them.
 * Images going to I/O devices generally don't have frames around them.
 * For the sake of generality, these routines will make no assumptions about
 * whether the source and/or destination images have frames.
 *
 * The source and destination image pointers will point to the upper
 * left corner of the frameless source and destination respectively.
 *
 * To make this really work, we must add a parameter to the ip_reduceTx
 * routines.  This parameter is the width of the source in pixels.
 * In previous versions, we derived this from the source bytes/line
 * assuming there was a frame.  Since we can't assume there's a frame,
 * we must specify the width directly.  This will also help if someone
 * gives us an image with a bigger frame than we would have otherwise expected.
 *
 * Since we ignore the frames, a user can use these routines to reduce
 * an image by bands.  That is, each band of an image can either come from
 * a slice of a source image or go to a slice of a destination image or both.
 *
 * In actual fact, the differences from the other reduction routines are
 * tiny.  We just don't adjust the source or dest pointers to account for the
 * frame.
 *
 * There is an obvious similarity between this code and normal thresholding
 * code insofar as we want to select every 2nd, 4th, 8th or 16th bit from
 * a word after all the ANDs and ORs have been done.  We'll use the same
 * code structure as is used in i_depth to work our way across a scan line,
 * so will need to pass the procedures the same args - offset of the rightmost
 * dest word, offset of the leftmost byte in the group of source words
 * that will be mapped to the rightmost dest word and offset of the rightmost
 * source word.
 *
 **************************************************************************
 *
 * There are obvious similarities between thresholding an image from n bits
 * to 1 bit and doing an n:1 reduction.  In the reduction, we need to look
 * at n rows as well as n columns and need to compute a function between
 * all the bits in the columns.  Once those steps are complete, we select
 * 1 out of n bits (the bit that has the correct reduced value).  This
 * last 1-from-n step is also used in thresholding an image.
 *
 * Originally, the part of the primitive algorithm used for the 1-from-n
 * step in reduction and the one used for thresholding were different.
 * We will now make them the same.  This allows us to get rid of one
 * set of tables and table-making routines (except that it's also used
 * in ssubsampl.c for now).
 *
 * There's another issue.  When doing 2:1 reduction, we reduce each 2x2
 * square to 2 bits in a word, one of which is meaningful.  A "calculated"
 * word contains a useful result in every other bit.  The only thing that
 * really distinguishes the four threshold reduction routines is the
 * calculation done to produce these bits.  When we have them, we combine
 * the meaningful bits in one word with the meaningful bits in another
 * to produce a full word.  That's fairly straightforward - two simple
 * calculation steps and a combination step..  When doing 16:1 reduction,
 * we combine 1/16 of the bits in a calculated word with 1/16 of the bits
 * in the next calculated word, and so on 16 times.  The calculation is
 * big because we have to get rid of so many bits.  The number of times
 * we have to do the calculation is also large.  When doing 2:1 reduction,
 * putting the calculation right into the main loop is reasonable.
 * When doing 16:1 reduction, it makes more sense to put it into a
 * subroutine.  Where do we draw the line?  Arbitrarily,
 * I'll put 2:1 and 4:1 calculations into the main loops and 8:1 and 16:1
 * calculations into subroutines.
 *
 * There are 4 types of 2:1 calculations, one for each threshold value.
 * There are only 2 types of calculations for 4:1, 8:1 and 16:1 reduction
 * because we only accelerate the cases for thresholds of 1 and 4.  The
 * other cases are done as cascades of 2:1 reductions.  As we said above,
 * the only thing that distinguishes R:1 reduction by threshold n from
 * R:1 reduction the threshold m is the type of calculation done to
 * compute a single meaningful bit from each RxR square.  In 8:1 and 16:1
 * reduction, this calculation will be done by subroutine.  In those
 * two cases, there will only be one overall body of line-at-a-time
 * primitive code for each reduction factor.  That code will be passed
 * a pointer to the subroutine corresponding to the threshold (1 or 4)
 * we need.
 *
 *********************************************************************
 *
 * Historically, the i_reduceX routines were written first, the io_ReduceX
 * routines later on.  When revising this file, it was noted that the
 * only difference between the i_reduceX and i_ioReduceX routines were
 * that the i_reduceX routines strip off the frame first.  Instead of
 * having two almost-copies of each routine, we just have the i_reduceX
 * routines strip the frame from an incoming image (adjust the pointers
 * widths and heights) and call the i_ioReduceX routines.
 *
 */


/*********************************************************************
 ************       Threshold reduction procedures       *************
 *********************************************************************/

/* PUBLIC */
/*********************************************************************
 *  i_reduce16()  
 *		Each 16x16 square in the source image is reduced to
 *		one pixel in the destination image according to the
 *		given level.  Conceptually 4 2x2 reductions are done
 *		with as given by the 4 levels.
 *
 *	NOTE: The source (plus frame) must be at least 16 times as wide
 *		and 16 times as tall as the dest.  If this condition
 *		does not hold, we will read off the end of the source.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(includes frame)
 *		wD =	width in pixels of dest (including frame)
 *		hD =	height of dest (including frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(includes frame)
 *		wS=	width in pixels of source (including frame)
 *		hS =	height of source (including frame)
 *		sBpl =	source image bytes/line
 *		level1 = level of first 2x2 reduction 1, 2, 3, 4
 *		level2 = level of second 2x2 reduction 1, 2, 3, 4
 *		level3 = level of third 2x2 reduction 1, 2, 3, 4
 *		level4 = level of fourth 2x2 reduction 1, 2, 3, 4
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *********************************************************************/
Int32  CDECL 
i_reduce16 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	level1,
    Int32	level2,
    Int32	level3,
    Int32	level4)
{
	/* This routine is exactly the same as i_ioReduce16 except that
	 * the images called here are assumed to have frames.
	 * We'll just strip off the frame and call i_ioReduce16 to save
	 * duplication of code.
	 *
	 * NOTE: in io_Reduce16, we will read from the source image
	 * as though it's of width 16 * wD and height 16 * hD.
	 * If the source + frame is less than 16 * dest in either dimension,
	 * we'll read off the end of the source image.
	 */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wS -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;
    hS -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    ERR_PASS( i_ioReduce16 (pIPState, pD, wD, hD, dBpl, pS, wS, hS, sBpl,
			    level1, level2, level3, level4));
    NORMAL_RETURN;
}


/* PUBLIC */
/*********************************************************************
 *  i_reduce8()
 *		Each 8x8 square in the source image is reduced to
 *		one pixel in the destination image according to the
 *		given level.  Conceptually 3 2x2 reductions are done
 *		with as given by the 3 levels.
 *
 *	NOTE: The source (plus frame) must be at least 8 times as wide
 *		and 8 times as tall as the dest.  If this condition
 *		does not hold, we will read off the end of the source.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(includes frame)
 *		wD =	width in pixels of dest (including frame)
 *		hD =	height of dest (including frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(includes frame)
 *		wS=	width in pixels of source (including frame)
 *		hS =	height of source (including frame)
 *		sBpl =	source image bytes/line
 *		level1 = level of first 2x2 reduction 1, 2, 3, 4
 *		level2 = level of second 2x2 reduction 1, 2, 3, 4
 *		level3 = level of third 2x2 reduction 1, 2, 3, 4
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *********************************************************************/
Int32  CDECL 
i_reduce8 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	level1,
    Int32	level2,
    Int32	level3)
{
	/* This routine is exactly the same as i_ioReduce8 except that
	 * the images called here are assumed to have frames.
	 * We'll just strip off the frame and call i_ioReduce8 to save
	 * duplication of code.
	 *
	 * NOTE: in io_Reduce8, we will read from the source image
	 * as though it's of width 8 * wD and height 8 * hD.
	 * If the source + frame is less than 8 * dest in either dimension,
	 * we'll read off the end of the source image.
	 */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wS -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;
    hS -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    ERR_PASS( i_ioReduce8 (pIPState, pD, wD, hD, dBpl, pS, wS, hS, sBpl,
			   level1, level2, level3));
    NORMAL_RETURN;
}


/* PUBLIC */
/*********************************************************************
 *  i_reduce4()
 *		Each 4x4 square in the source image is reduced to
 *		one pixel in the destination image according to the
 *		given level.  Conceptually 2 2x2 reductions are done
 *		with as given by the 2 levels.
 *
 *	NOTE: The source (plus frame) must be at least 4 times as wide
 *		and 4 times as tall as the dest.  If this condition
 *		does not hold, we will read off the end of the source.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(includes frame)
 *		wD =	width in pixels of dest (including frame)
 *		hD =	height of dest (including frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(includes frame)
 *		wS=	width in pixels of source (including frame)
 *		hS =	height of source (including frame)
 *		sBpl =	source image bytes/line
 *		level1 = level of first 2x2 reduction 1, 2, 3, 4
 *		level2 = level of second 2x2 reduction 1, 2, 3, 4
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *********************************************************************/
Int32  CDECL 
i_reduce4 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	level1,
    Int32	level2)
{
	/* This routine is exactly the same as i_ioReduce4 except that
	 * the images called here are assumed to have frames.
	 * We'll just strip off the frame and call i_ioReduce4 to save
	 * duplication of code.
	 *
	 * NOTE: in io_Reduce4, we will read from the source image
	 * as though it's of width 4 * wD and height 4 * hD.
	 * If the source + frame is less than 4 * dest in either dimension,
	 * we'll read off the end of the source image.
	 */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wS -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;
    hS -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    ERR_PASS( i_ioReduce4 (pIPState, pD, wD, hD, dBpl, pS, wS, hS, sBpl,
			   level1, level2));
    NORMAL_RETURN;
}


/* PUBLIC */
/*********************************************************************
 *  i_reduce2()
 *		Each 2x2 square in the source image is reduced to
 *		one pixel in the destination image according to the
 *		given level.  The reduced image has an ON pixel whenever
 *		at least 'level' of the four corresponding pixels in
 *		the source image are ON.
 *
 *	NOTE: The source (plus frame) must be at least 2 times as wide
 *		and 2 times as tall as the dest.  If this condition
 *		does not hold, we will read off the end of the source.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(includes frame)
 *		wD =	width in pixels of dest (including frame)
 *		hD =	height of dest (including frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(includes frame)
 *		sBpl =	source image bytes/line
 *		level1 = level of first 2x2 reduction 1, 2, 3, 4
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *********************************************************************/
Int32  CDECL 
i_reduce2 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	level)
{
	/* This routine is exactly the same as i_ioReduce2 except that
	 * the images called here are assumed to have frames.
	 * We'll just strip off the frame and call i_ioReduce2 to save
	 * duplication of code.
	 *
	 * NOTE: in io_Reduce2, we will read from the source image
	 * as though it's of width 2 * wD and height 2 * hD.
	 * If the source + frame is less than 2 * dest in either dimension,
	 * we'll read off the end of the source image.
	 */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    ERR_PASS( i_ioReduce2 (pIPState, pD, wD, hD, dBpl, pS, 2*wD, sBpl, level));
    NORMAL_RETURN;
}




/*********************************************************************
 ************       I/O reduction procedures       *************
 *********************************************************************/

/* PUBLIC */
/*********************************************************************
 *  i_ioReduce16()  
 *		Each 16x16 square in the source image is reduced to
 *		one pixel in the destination image according to the
 *		given level.  Conceptually 4 2x2 reductions are done
 *		with as given by the 4 levels.
 *
 *		This routine makes no assumption about whether the source
 *		and/or dest images have frames.  It can be used to
 *		reduce image by bands.  All pointers are to the upper
 *		left corner of the real image (inside the frame if one
 *		exists).
 *
 *	NOTE: The source (NO frame) must be at least 16 times as wide
 *		and 16 times as tall as the dest.  If this condition
 *		does not hold, we will read off the end of the source.
 *		When this routine is called from i_reduce16, there
 *		is an implicit frame around the outside of the
 *		image we're given.  That adds some leeway.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(No frame or inside frame)
 *		wD =	width in pixels of dest (No frame or inside frame)
 *		hD =	height of dest (No frame or inside frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(No frame or inside frame)
 *		wS=	width in pixels of source (No frame or inside frame)
 *		hS =	height of source (No frame or inside frame)
 *		sBpl =	source image bytes/line
 *		level1 = level of first 2x2 reduction 1, 2, 3, 4
 *		level2 = level of second 2x2 reduction 1, 2, 3, 4
 *		level3 = level of third 2x2 reduction 1, 2, 3, 4
 *		level4 = level of fourth 2x2 reduction 1, 2, 3, 4
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *********************************************************************/
Int32  CDECL 
i_ioReduce16 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	level1,
    Int32	level2,
    Int32	level3,
    Int32	level4)
{
PROLOGUE;
 Int32	 wT, hT, tBpl, row;
UInt8	*pT;
Int32	 srcLimit;			/* max bytes in source line */
Int32	 srcInitOffset;			/* initial source line offset */
Int32	 dstInitOffset;			/* initial dst line offset */
UInt8	*pDOrig;			/* used to clear the right frame
					 * when we're done */

    if ((level1 == level2) && (level1 == level3) && (level1 == level4) &&
        ((level1 == 1) || (level1 == 4)))               /* do special case */
    {
	    /* if w == 0, just return instead of blowing up below */
	if (wD <= 0)
	    NORMAL_RETURN;
    
	    /* this points to the last dest word. */
	dstInitOffset = ((wD - 1)/32) * 4 ;
	    /* it takes 16 source words to make one destination word */
	srcInitOffset = dstInitOffset * 16;	
    
	    /* we want a pointer to the last word in the source line since
	     * this might not be a multiple of 4.
	     * There's a funny case here.  Say wS = 527.  Then the caller
	     * should have set wD = wS/16 = 32.  Therefore, dstInitOffset
	     * will be set to (32 - 1)/32 * 4 = 0.  srcInitOffset will be
	     * dstInitOffset*16 = 0.  This is fine, we'll only write one
	     * dest word and will only need 16 source words.  BUT, if we
	     * just set srcLimit to ((wS - 1)/32 * 4, we'd get (527-1)/32 * 4,
	     * or 64.  We want an initial offset of 60 bytes, not 64.
	     * So, we toss the LS 4 bits of wS before doing anything else.
	     */
	srcLimit = (((wS & ~0xF) - 1)/32) * 4;

	pDOrig = pD;			/* remember for clearRightFrame */

	if (level1 == 1)
	{
	    for (row = hD; row > 0; row--, pS += 16*sBpl, pD += dBpl)
	    {
		ip_reduce16Line (pIPState, pS, sBpl, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 &ip_reduce16xT1);
	    }
	}
        else
	{
	    for (row = hD; row > 0; row--, pS += 16*sBpl, pD += dBpl)
	    {
		ip_reduce16Line (pIPState, pS, sBpl, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 &ip_reduce16xT4);
	    }
	}
	    /* clear the right frame.  We need to do this when there was
	     * the source was wider than 2*wD.  If it was only 1 column
	     * wider, we could have bits there from thresholds 1 or 2.
	     * If the caller was trying to do a crude clipping operation,
	     * the source might be even wider than that, so we'll just
	     * clear the right frame all the time.
	     */
	ip_clearRightFrame (pIPState, pDOrig, wD, hD, dBpl, 1);

    }
    else /* no special case */
    {
            /* create the intermediate quarter-width, quarter-height pr.
	     * Since the DAE folks like to supply a dest image that's
	     * too big and i_ioReduce4 will read 4 source lines for
	     * each dest line, we'll create the temp image as 4x the
	     * dest image instead of 1/4x the source image.  That way,
	     * we'll read too much from the original source image.
	     * If we were called from i_reduce16, there's a frame, so
	     * it's ok.  If we were called directly, there's a
	     * problem.  The DAE folks say they're never call this
	     * routine directly.
             */
        wT = wD * 4;
        hT = hD * 4;
	tBpl = ((wT + 31) / 32) * 4;	/* implict depth of 1 */
	pT = (UInt8 *)MALLOC(tBpl*hT);
	if (!pT)
	    return ia_nomem;

        i_ioReduce4 (pIPState, pT, wT, hT, tBpl, pS, wS, hS, sBpl,
		     level1, level2);
        i_ioReduce4 (pIPState, pD, wD, hD, dBpl, pT, wT, hT, tBpl,
		     level3, level4);
        
        FREE(pT);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*********************************************************************
 *  i_ioReduce8()
 *		Each 8x8 square in the source image is reduced to
 *		one pixel in the destination image according to the
 *		given level.  Conceptually 3 2x2 reductions are done
 *		with as given by the 3 levels.
 *
 *		This routine makes no assumption about whether the source
 *		and/or dest images have frames.  It can be used to
 *		reduce image by bands.  All pointers are to the upper
 *		left corner of the real image (inside the frame if one
 *		exists).
 *
 *	NOTE: The source (NO frame) must be at least 8 times as wide
 *		and 8 times as tall as the dest.  If this condition
 *		does not hold, we will read off the end of the source.
 *		When this routine is called from i_reduce8, there
 *		is an implicit frame around the outside of the
 *		image we're given.  That adds some leeway.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(No frame or inside frame)
 *		wD =	width in pixels of dest (No frame or inside frame)
 *		hD =	height of dest (No frame or inside frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(No frame or inside frame)
 *		wS=	width in pixels of source (No frame or inside frame)
 *		hS =	height of source (No frame or inside frame)
 *		sBpl =	source image bytes/line
 *		level1 = level of first 2x2 reduction 1, 2, 3, 4
 *		level2 = level of second 2x2 reduction 1, 2, 3, 4
 *		level3 = level of third 2x2 reduction 1, 2, 3, 4
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *********************************************************************/
Int32  CDECL 
i_ioReduce8 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	level1,
    Int32	level2,
    Int32	level3)
{
    PROLOGUE;
    Int32	wT, hT, tBpl, row;
    UInt8      *pT;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    UInt8      *pDOrig;			/* used to clear the right frame
					 * when we're done */
    Int32	longestAccess;		/* the longest memory access
					 * allowed for the current processor
					 * and source and dest images. */
    UInt32     *pTabBinarizeByte0;      /* permute table bytes */
    UInt32     *pTabBinarizeByte1;
    UInt32     *pTabBinarizeByte2;
    UInt32     *pTabBinarizeByte3;


    if ((level1 == level2) && (level1 == level3) &&
        ((level1 == 1) || (level1 == 4)))               /* do special case */
    {

	    /* if w == 0, just return instead of blowing up below */
	if (wD <= 0)
	    NORMAL_RETURN;
    
	    /* Figure out whether the lower level routines will do long
	     * (8-byte) or short (4-byte) accesses.
	     */
	ip_checkImageAccess (pIPState, pS, sBpl, pD, dBpl, &longestAccess);
	if ((longestAccess != 4) && (longestAccess != 8))
	    return ia_internal;

	    /* make the binarization tables */
	pTabBinarizeByte0 = NULL;
	pTabBinarizeByte1 = NULL;
	pTabBinarizeByte2 = NULL;
	pTabBinarizeByte3 = NULL;

	pDOrig = pD;		/* save for clearRightFrame below */
    
	    /* Set the line offsets and initialize the tables according
	     * to whether we're processing 4-byte things or 8-byte things.
	     * We always process 4-byte things if the long operations aren't
	     * here.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
	    if ((__error__ = i_makeBinarizationTables (pIPState,
						     &pTabBinarizeByte0,
						     &pTabBinarizeByte1,
						     &pTabBinarizeByte2,
						     &pTabBinarizeByte3))
		!= ia_successful)
	      NORMAL_RETURN;
    
		/* this points to the last dest word. */
	    dstInitOffset = ((wD - 1)/32) * 4 ;
		/* it takes 8 source words to make one destination word */
	    srcInitOffset = dstInitOffset * 8;	
	
		/* we want a pointer to the last word in the source line since
		 * this might not be a multiple of 4.
		 * There's a funny case here.  Say wS = 263.  Then the caller
		 * should have set wD = wS/8 = 32.  Therefore, dstInitOffset
		 * will be set to (32 - 1)/32 * 4 = 0.  srcInitOffset will be
		 * dstInitOffset*8 = 0.  This is fine, we'll only write one
		 * dest word and will only need 8 source words.  BUT, if we
		 * just set srcLimit to ((wS - 1)/32 * 4, we'd get
		 * (263-1)/32 * 4, or 32.  We want an initial offset of
		 * 28 bytes, not 32.  So, we toss the LS 3 bits of wS
		 * before doing anything else.
		 */
	    srcLimit = (((wS & ~0x7) - 1)/32) * 4;
    
	    pDOrig = pD;		/* remember for clearRightFrame */
    
	    if (level1 == 1)
	    {
		for (row = hD; row > 0; row--, pS += 8*sBpl, pD += dBpl)
		{
		    ip_reduce8Line (pIPState, pS, sBpl, pD,
				    dstInitOffset, srcInitOffset, srcLimit,
				    pTabBinarizeByte3,
				    ip_reduce8xT1);
		}
	    }
	    else  /* level1 == 4 */
	    {
		for (row = hD; row > 0; row--, pS += 8*sBpl, pD += dBpl)
		{
		    ip_reduce8Line (pIPState, pS, sBpl, pD,
				    dstInitOffset, srcInitOffset, srcLimit,
				    pTabBinarizeByte3,
				    ip_reduce8xT4);
		}
	    }
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* this points to the last dest dword. */
	    dstInitOffset = ((wD - 1)/64) * 8 ;
		/* it takes 8 source dwords to make one destination dword */
	    srcInitOffset = dstInitOffset * 8;	
    
		/* we want a pointer to the last dword in the source line since
		 * this might not be a multiple of 8.
		 * There's a funny case here.  Say wS = 519.  Then the caller
		 * should have set wD = wS/8 = 64.  Therefore, dstInitOffset
		 * will be set to (64 - 1)/64 * 8 = 0.  srcInitOffset will be
		 * dstInitOffset*4 = 0.  This is fine, we'll only write one
		 * dest dword and will only need 8 source words.  BUT, if we
		 * just set srcLimit to ((wS - 1)/64 * 8, we'd get
		 * (519-1)/64 * 8, or 64.  We want an initial offset of
		 * 56 bytes, not 64.  So, we toss the LS three bits of
		 * wS before doing anything else.
		 */
	    srcLimit = (((wS & ~0x7) - 1)/64) * 8;

		/* do whatever's necessary to enter long operation mode. */
	    ip_initLongOps (pIPState);

	    if (level1 == 1)
	    {
		for (row = hD; row > 0; row--, pS += 8*sBpl, pD += dBpl)
		{
		    ip_reduce8T1LineLong (pIPState, pS, wS, sBpl, pD,
					  dstInitOffset, srcInitOffset,
					  srcLimit,
					  NULL);
		}
	    }
	    else
	    {
		for (row = hD; row > 0; row--, pS += 8*sBpl, pD += dBpl)
		{
		    ip_reduce8T4LineLong (pIPState, pS, wS, sBpl, pD,
					  dstInitOffset, srcInitOffset,
					  srcLimit,
					  NULL);
		}
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}

	    /* clear the right frame.  We need to do this when there was
	     * the source was wider than 2*wD.  If it was only 1 column
	     * wider, we could have bits there from thresholds 1 or 2.
	     * If the caller was trying to do a crude clipping operation,
	     * the source might be even wider than that, so we'll just
	     * clear the right frame all the time.
	     */
	ip_clearRightFrame (pIPState, pDOrig, wD, hD, dBpl, 1);

	if (pTabBinarizeByte0 != NULL)
	    FREE(pTabBinarizeByte0);
	if (pTabBinarizeByte1 != NULL)
	    FREE(pTabBinarizeByte1);
	if (pTabBinarizeByte2 != NULL)
	    FREE(pTabBinarizeByte2);
	if (pTabBinarizeByte3 != NULL)
	    FREE(pTabBinarizeByte3);
    }
    else /* no special case */
    {
            /* create the intermediate quarter-width, quarter-height pr.
	     * Since the DAE folks like to supply a dest image that's
	     * too big and i_reduceX will read X source lines for
	     * each dest line, we'll create the temp image as 2x the
	     * dest image instead of 1/4x the source image.
	     * If we were called from i_reduce8, there's a frame, so
	     * it's ok.  If we were called directly, there's a
	     * problem.  The DAE folks say they're never call this
	     * routine directly.
             */
        wT = wD * 2;
        hT = hD * 2;
	tBpl = ((wT + 31) / 32) * 4;	/* implicit depth of 1 */
	pT = (UInt8 *)MALLOC(tBpl*hT);
	if (!pT)
	    return ia_nomem;

        i_ioReduce4 (pIPState, pT, wT, hT, tBpl, pS, wS, hS, sBpl,
		     level1, level2);
        i_ioReduce2 (pIPState, pD, wD, hD, dBpl, pT, wT, tBpl,
		     level3);
    
        FREE(pT);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*********************************************************************
 *  i_ioReduce4()
 *		Each 4x4 square in the source image is reduced to
 *		one pixel in the destination image according to the
 *		given level.  Conceptually 2 2x2 reductions are done
 *		with as given by the 2 levels.
 *
 *		This routine makes no assumption about whether the source
 *		and/or dest images have frames.  It can be used to
 *		reduce image by bands.  All pointers are to the upper
 *		left corner of the real image (inside the frame if one
 *		exists).
 *
 *	NOTE: The source (NO frame) must be at least 4 times as wide
 *		and 4 times as tall as the dest.  If this condition
 *		does not hold, we will read off the end of the source.
 *		When this routine is called from i_reduce4, there
 *		is an implicit frame around the outside of the
 *		image we're given.  That adds some leeway.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(No frame or inside frame)
 *		wD =	width in pixels of dest (No frame or inside frame)
 *		hD =	height of dest (No frame or inside frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(No frame or inside frame)
 *		wS=	width in pixels of source (No frame or inside frame)
 *		hS =	height of source (No frame or inside frame)
 *		sBpl =	source image bytes/line
 *		level1 = level of first 2x2 reduction 1, 2, 3, 4
 *		level2 = level of second 2x2 reduction 1, 2, 3, 4
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *********************************************************************/
Int32  CDECL 
i_ioReduce4 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	level1,
    Int32	level2)
{
    PROLOGUE;
    Int32	wT, hT, tBpl, row;
    UInt8      *pT;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    UInt8      *pDOrig;			/* used to clear the right frame
					 * when we're done */
    Int32	longestAccess;		/* the longest memory access
					 * allowed for the current processor
					 * and source and dest images. */
    UInt32     *pTab4To1bppByte0;	/* permute table bytes */
    UInt32     *pTab4To1bppByte1;
    UInt32     *pTab4To1bppByte2;
    UInt32     *pTab4To1bppByte3;


	/* Keep the compiler happy */
    hS = hS;

    if ((level1 == level2) &&
        ((level1 == 1) || (level1 == 4)))               /* do special case */
    {
	    /* if w == 0, just return instead of blowing up below */
	if (wD <= 0)
	    NORMAL_RETURN;
    
	    /* Figure out whether the lower level routines will do long
	     * (8-byte) or short (4-byte) accesses.
	     */
	ip_checkImageAccess (pIPState, pS, sBpl, pD, dBpl, &longestAccess);
	if ((longestAccess != 4) && (longestAccess != 8))
	    return ia_internal;

	    /* initialize bin table pointers */
	pTab4To1bppByte0 = NULL;
	pTab4To1bppByte1 = NULL;
	pTab4To1bppByte2 = NULL;
	pTab4To1bppByte3 = NULL;

	pDOrig = pD;		/* save for clearRightFrame below */
    
	    /* Set the line offsets and initialize the tables according
	     * to whether we're processing 4-byte things or 8-byte things.
	     * We always process 4-byte things if the long operations aren't
	     * here.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
	    if ((__error__ = i_make4To1BinarizationTables (pIPState,
							 &pTab4To1bppByte0,
							 &pTab4To1bppByte1,
							 &pTab4To1bppByte2,
							 &pTab4To1bppByte3)
		 ) != ia_successful)
		NORMAL_RETURN;
    
		/* this points to the last dest word. */
	    dstInitOffset = ((wD - 1)/32) * 4 ;
		/* it takes 4 source words to make one destination word */
	    srcInitOffset = dstInitOffset * 4;	
    
		/* we want a pointer to the last word in the source line since
		 * this might not be a multiple of 4.
		 * There's a funny case here.  Say wS = 131.  Then the caller
		 * should have set wD = wS/4 = 32.  Therefore, dstInitOffset
		 * will be set to (32 - 1)/32 * 4 = 0.  srcInitOffset will be
		 * dstInitOffset*4 = 0.  This is fine, we'll only write one
		 * dest word and will only need 4 source words.  BUT, if we
		 * just set srcLimit to ((wS - 1)/32 * 4, we'd get
		 * (131-1)/32 * 4, or 16.  We want an initial offset of
		 * 12 bytes, not 16.  So, we toss the LS two bits of
		 * wS before doing anything else.
		 */
	    srcLimit = (((wS & ~0x3) - 1)/32) * 4;
    
	    if (level1 == 1)
	    {
		for (row = hD; row > 0; row--, pS += 4*sBpl, pD += dBpl)
		{
		    ip_reduce4T1Line (pIPState, pS, sBpl, pD,
				      dstInitOffset, srcInitOffset, srcLimit,
				      pTab4To1bppByte3);
		}
	    }
	    else
	    {
		for (row = hD; row > 0; row--, pS += 4*sBpl, pD += dBpl)
		{
		    ip_reduce4T4Line (pIPState, pS, sBpl, pD,
				      dstInitOffset, srcInitOffset, srcLimit,
				      pTab4To1bppByte3);
		}
	    }
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* this points to the last dest dword. */
	    dstInitOffset = ((wD - 1)/64) * 8 ;
		/* it takes 4 source dwords to make one destination dword */
	    srcInitOffset = dstInitOffset * 4;	
    
		/* we want a pointer to the last dword in the source line since
		 * this might not be a multiple of 8.
		 * There's a funny case here.  Say wS = 259.  Then the caller
		 * should have set wD = wS/4 = 64.  Therefore, dstInitOffset
		 * will be set to (64 - 1)/64 * 8 = 0.  srcInitOffset will be
		 * dstInitOffset*4 = 0.  This is fine, we'll only write one
		 * dest dword and will only need 4 source words.  BUT, if we
		 * just set srcLimit to ((wS - 1)/64 * 8, we'd get
		 * (259-1)/64 * 8, or 32.  We want an initial offset of
		 * 24 bytes, not 32.  So, we toss the LS two bits of
		 * wS before doing anything else.
		 */
	    srcLimit = (((wS & ~0x3) - 1)/64) * 8;

		/* do whatever's necessary to enter long operation mode. */
	    ip_initLongOps (pIPState);	

	    if (level1 == 1)
	    {
		for (row = hD; row > 0; row--, pS += 4*sBpl, pD += dBpl)
		{
		    ip_reduce4T1LineLong (pIPState, pS, wS, sBpl, pD,
					  dstInitOffset, srcInitOffset,
					  srcLimit,
					  NULL);
		}
	    }
	    else
	    {
		for (row = hD; row > 0; row--, pS += 4*sBpl, pD += dBpl)
		{
		    ip_reduce4T4LineLong (pIPState, pS, wS, sBpl, pD,
					  dstInitOffset, srcInitOffset,
					  srcLimit,
					  NULL);
		}
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}

	    /* clear the right frame.  We need to do this when there was
	     * the source was wider than 2*wD.  If it was only 1 column
	     * wider, we could have bits there from thresholds 1 or 2.
	     * If the caller was trying to do a crude clipping operation,
	     * the source might be even wider than that, so we'll just
	     * clear the right frame all the time.
	     */
	ip_clearRightFrame (pIPState, pDOrig, wD, hD, dBpl, 1);

	if (pTab4To1bppByte0 != NULL)
	    FREE(pTab4To1bppByte0);
	if (pTab4To1bppByte1 != NULL)
	    FREE(pTab4To1bppByte1);
	if (pTab4To1bppByte2 != NULL)
	    FREE(pTab4To1bppByte2);
	if (pTab4To1bppByte3 != NULL)
	    FREE(pTab4To1bppByte3);
    }
    else /* no special case */
    {
            /* create the intermediate half-width, half-height pr.
	     * Since the DAE folks like to supply a dest image that's
	     * too big and i_reduceX will read X source lines for
	     * each dest line, we'll create the temp image as 2x the
	     * dest image instead of 1/2x the source image.
	     * If we were called from i_reduce4, there's a frame, so
	     * it's ok.  If we were called directly, there's a
	     * problem.  The DAE folks say they're never call this
	     * routine directly.
             */
        wT = wD * 2;
        hT = hD * 2;
	tBpl = ((wT + 31) / 32) * 4;	/* implicit depth of 1 */
	pT = (UInt8 *)MALLOC(tBpl*hT);
	if (!pT)
	    return ia_nomem;
    
        i_ioReduce2 (pIPState, pT, wT, hT, tBpl, pS, wS, sBpl, level1);
        i_ioReduce2 (pIPState, pD, wD, hD, dBpl, pT, wT, tBpl, level2);
    
        FREE(pT);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*********************************************************************
 *  i_ioReduce2()
 *		Each 2x2 square in the source image is reduced to
 *		one pixel in the destination image according to the
 *		given level.  The reduced image has an ON pixel whenever
 *		at least 'level' of the four corresponding pixels in
 *		the source image are ON.
 *
 *		This routine makes no assumption about whether the source
 *		and/or dest images have frames.  It can be used to
 *		reduce image by bands.  All pointers are to the upper
 *		left corner of the real image (inside the frame if one
 *		exists).
 *
 *		Since it makes no assumption about frame size, it must
 *		take an additional parameter, the width of the of the
 *		source image.  This is used to calculate the address of
 *		the rightmost word in each source line.
 *
 *	NOTE: The source (NO frame) must be at least 2 times as wide
 *		and 2 times as tall as the dest.  If this condition
 *		does not hold, we will read off the end of the source.
 *		When this routine is called from i_reduce2, there
 *		is an implicit frame around the outside of the
 *		image we're given.  That adds some leeway.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(No frame or inside frame)
 *		wD =	width in pixels of dest (No frame or inside frame)
 *		hD =	height of dest (No frame or inside frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(No frame or inside frame)
 *		wS=	width in pixels of source (No frame or inside frame)
 *		sBpl =	source image bytes/line
 *		level1 = level of first 2x2 reduction 1, 2, 3, 4
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *********************************************************************/
Int32  CDECL 
i_ioReduce2 (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	wS,
    Int32	sBpl,
    Int32	level)
{
    PROLOGUE;
     
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    UInt8      *pDOrig;			/* used to clear the right frame
					 * when we're done */
    Int32	row;			/* iteration var */
    Int32	longestAccess;		/* the longest memory access
					 * allowed for the current processor
					 * and source and dest images. */
    UInt32     *pTab2To1bppByte0;	/* permute table bytes */
    UInt32     *pTab2To1bppByte1;
    UInt32     *pTab2To1bppByte2;
    UInt32     *pTab2To1bppByte3;

	/* if wD == 0, just return instead of blowing up below */
    if (wD <= 0)
	NORMAL_RETURN;

	/* Figure out whether the lower level routines will do long
	 * (8-byte) or short (4-byte) accesses.
	 */
    ip_checkImageAccess (pIPState, pS, sBpl, pD, dBpl, &longestAccess);
    if ((longestAccess != 4) && (longestAccess != 8))
	return ia_internal;

	/* initialize bin table pointers */
    pTab2To1bppByte0 = NULL;
    pTab2To1bppByte1 = NULL;
    pTab2To1bppByte2 = NULL;
    pTab2To1bppByte3 = NULL;

    pDOrig = pD;			/* save to clear right frame */

	/* Set the line offsets and initialize the tables according
	 * to whether we're processing 4-byte things or 8-byte things.
	 * We always process 4-byte things if the long operations aren't
	 * here.
	 */
    if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
    {
	if ((__error__ = i_make2To1BinarizationTables (pIPState,
						     &pTab2To1bppByte0,
						     &pTab2To1bppByte1,
						     &pTab2To1bppByte2,
						     &pTab2To1bppByte3))
	    != ia_successful)
	NORMAL_RETURN;

	    /* this points to the last dest word.  The dest has 1-bit pixels. */
	dstInitOffset = ((wD - 1)/32) * 4 ;
	    /* it takes 2 source words to make one destination word */
	srcInitOffset = dstInitOffset * 2;	
    
	    /* we want a pointer to the last word in the source line since
	     * this might not be a multiple of 4.  We use 32 since there are
	     * 32 1-bit pixels per 32-bit word.
	     * There's a funny case here.  Say wS = 65.  Then the caller
	     * should have set wD = wS/2 = 32.  Therefore, dstInitOffset
	     * will be set to (32 - 1)/32 * 4 = 0.  srcInitOffset will be
	     * dstInitOffset*2 = 0.  This is fine, we'll only write one
	     * dest word and will only need two source words.  BUT, if we
	     * just set srcLimit to ((wS - 1)/32 * 4, we'd get (65-1)/32 * 4,
	     * or 8.  We want an initial offset of 4 bytes, not 8.
	     * So, we toss the LS bit of wS before doing anything else.
	     */
	srcLimit = (((wS & ~0x1) - 1)/32) * 4;

	switch (level) {
	case 1:
	    for (row = hD; row > 0; row--, pS += 2*sBpl, pD += dBpl)
	    {
		ip_reduceT1Line (pIPState, pS, sBpl, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 pTab2To1bppByte3);
	    }
	    break;
	case 2:
	    for (row = hD; row > 0; row--, pS += 2*sBpl, pD += dBpl)
	    {
		ip_reduceT2Line (pIPState, pS, sBpl, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 pTab2To1bppByte3);
	    }
	    break;
	case 3:
	    for (row = hD; row > 0; row--, pS += 2*sBpl, pD += dBpl)
	    {
		ip_reduceT3Line (pIPState, pS, sBpl, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 pTab2To1bppByte3);
	    }
	    break;
	default:	 /* case 4 */
	    for (row = hD; row > 0; row--, pS += 2*sBpl, pD += dBpl)
	    {
		ip_reduceT4Line (pIPState, pS, sBpl, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 pTab2To1bppByte3);
	    }
	    break;
	}

    }
    else	/* we're going to process in units of 64-bit words */
    {
	    /* this points to the last dest doubleword.
	       The dest has 1-bit pixels. */
	dstInitOffset = ((wD - 1)/64) * 8 ;
	    /* it takes 2 source doublewords to make one dest doubleword */
	srcInitOffset = dstInitOffset * 2;	
    
	    /* we want a pointer to the last doubleword in the source line
	     * since this might not be a multiple of 8.  We use 64
	     * since there are 64 1-bit pixels per 64-bit doubleword.
	     * There's a funny case here.  Say wS = 129.  Then the caller
	     * should have set wD = wS/2 = 64.  Therefore, dstInitOffset
	     * will be set to (64 - 1)/64 * 4 = 0.  srcInitOffset will be
	     * dstInitOffset*2 = 0.  This is fine, we'll only write one
	     * dest doubleword and will only need two source doublewords.
	     * BUT, if we just set srcLimit to ((wS - 1)/64 * 4, we'd get
	     * (129-1)/64 * 8, or 16.  We want an initial offset of
	     * 8 bytes, not 16.
	     * So, we toss the LS bit of wS before doing anything else.
	     */
	srcLimit = (((wS & ~0x1) - 1)/64) * 8;

	    /* do whatever's necessary to enter long operation mode. */
	ip_initLongOps (pIPState);

	switch (level) {
	case 1:
	    for (row = hD; row > 0; row--, pS += 2*sBpl, pD += dBpl)
	    {
		ip_reduceT1LineLong (pIPState, pS, wS, sBpl, pD,
				     dstInitOffset, srcInitOffset, srcLimit,
				     pTab2To1bppByte3);
	    }
	    break;
	case 2:
	    for (row = hD; row > 0; row--, pS += 2*sBpl, pD += dBpl)
	    {
		ip_reduceT2LineLong (pIPState, pS, wS, sBpl, pD,
				     dstInitOffset, srcInitOffset, srcLimit,
				     pTab2To1bppByte3);
	    }
	    break;
	case 3:
	    for (row = hD; row > 0; row--, pS += 2*sBpl, pD += dBpl)
	    {
		ip_reduceT3LineLong (pIPState, pS, wS, sBpl, pD,
				     dstInitOffset, srcInitOffset, srcLimit,
				     pTab2To1bppByte3);
	    }
	    break;
	default:	 /* case 4 */
	    for (row = hD; row > 0; row--, pS += 2*sBpl, pD += dBpl)
	    {
		ip_reduceT4LineLong (pIPState, pS, wS, sBpl, pD,
				     dstInitOffset, srcInitOffset, srcLimit,
				     pTab2To1bppByte3);
	    }
	    break;
	}
	    /* issue any opcodes necessary to exit from long access mode */
	ip_finishLongOps (pIPState);
    }

	/* clear the right frame.  We need to do this when there was
	 * the source was wider than 2*wD.  If it was only 1 column
	 * wider, we could have bits there from thresholds 1 or 2.
	 * If the caller was trying to do a crude clipping operation,
	 * the source might be even wider than that, so we'll just
	 * clear the right frame all the time.
	 */
    ip_clearRightFrame (pIPState, pDOrig, wD, hD, dBpl, 1);

    if (pTab2To1bppByte0 != NULL)
	FREE(pTab2To1bppByte0);
    if (pTab2To1bppByte1 != NULL)
	FREE(pTab2To1bppByte1);
    if (pTab2To1bppByte2 != NULL)
	FREE(pTab2To1bppByte2);
    if (pTab2To1bppByte3 != NULL)
	FREE(pTab2To1bppByte3);
    NORMAL_RETURN;
}



/*

$Log: i_reduce.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.8  1998/01/13 23:49:22  ddavies
i_reduce.c -> /product/ipcore/ipshared/src/RCS/i_reduce.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1996/10/24 19:12:34  ddavies
ddavies on Thu Oct 24 12:08:15 PDT 1996

i_reduce.c -> /product/ipcore/ipshared/src/RCS/i_reduce.c,v

Modified i_ioReduce2, i_ioReduce4 and i_ioReduce8 so they decide whether
MMX code can be used and call it if so.  ip_checkImageAccess and
ip_longOpScalLibLinked tell it whether MMX operations are present and legal
on the current image AND whether code using MMX opcodes is linked in.
If all of these are true, the MMX routines ip_reduceT1LineLong,
ip_reduceT2LineLong, etc are used.  When MMX is used, we also call
ip_initLongOps and ip_finishLongOps around the calls to the MMX code.

Revision 1.6  1995/12/13 02:45:29  ejaquith
i_reduce.c -> /product/ipcore/ipshared/src/RCS/i_reduce.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/10/18  03:58:15  ddavies
 * i_reduce.c -> /product/ipcore/ipshared/src/RCS/i_reduce.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.4  1994/09/17  17:58:59  ddavies
 * i_reduce.c -> /product/ipcore/ipshared/src/RCS/i_reduce.c,v
 *
 * Change i_ioReduce2, i_ioReduce4 and i_ioReduce8 to use the newly public
 * table making procedures i_make2To1BinarizationTables,
 * i_make4To1BinarizationTables and i_makeBinarizationTables instead of their
 * ip_ (private) counterparts.
 *
 * Revision 1.3  1994/07/27  16:06:21  ejaquith
 * ejaquith on Wed Jul 27 09:04:10 PDT 1994
 *
 * i_reduce.c -> /product/ipcore/ipshared/src/RCS/i_reduce.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Made functions use local pointers to tables instead of globals.
 *
 * Revision 1.2  1994/02/04  20:15:58  dferrell
 * i_reduce.c -> /product/ipcore/ipshared/src/RCS/i_reduce.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/02/03  00:00:38  ddavies
 * Initial revision
 *
 * Revision 1.20  1994/01/27  03:49:47  ddavies
 * freduce.c -> /product/ipcore/ipshared/src/RCS/freduce.c,v
 *
 * Quick fix of the previous fix:  The code that attempted to clear the right
 * and bottom frames of the intermediate images actually cleared some area below
 * the bottom of the image.  I forgot that the frames were already included
 * in wT and hT.  Following Richard Mylnarik's suggestion, the temporary
 * images are now allocated relative to the dest, not the source.  This allows
 * us to skip initializing the bottom and right areas of the temp images
 * even when the DAE folks give us a dest image that's too big.  When this
 * happens, we'll read into the frame of the source image, except in io_reduceX
 * where the source might not have a frame.  The DAE folks assure us they
 * don't use io_reduce and never will.
 * In the next day or so, sreduce.c and freduce.c will become i_reduce.c and
 * mireduce.c.
 *
 * Revision 1.19  1994/01/14  23:28:19  ddavies
 * ddavies on Fri Jan 14 15:27:46 PST 1994
 *
 * freduce.c -> /product/ipcore/ipshared/src/RCS/freduce.c,v
 *
 * Fix bugs caused by uninitialized frames when making temporary images.
 * Thanks for finding these Richard!!
 *
 * Revision 1.18  1993/12/02  19:21:43  ddavies
 * freduce.c -> /product/ipcore/ipshared/src/RCS/freduce.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.17  1993/11/05  01:25:36  ddavies
 * freduce.c -> /product/ipcore/ipshared/src/RCS/freduce.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


