/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_rescal.c
 *
 * Public rescaling routines:
 *		Int32	i_rescaleArb ()
 *		Int32	i_rescaleAreaMap ()
 *		Int32	i_rescaleNoInterp ()
 *		Int32	i_getScaleToSizeTableSize ()
 *		Int32	i_initScaleToSizeTables ()
 *		Int32	i_rescaleNoInterpSubRegion ()
 *
 * Private rescaling routines:
 *		Int32	ip_getAllScaleToSizeTableSizes ()
 */

#include <stdio.h>

#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "shrscal.pub"
#include "shrrast.pub"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrscal.prv"
#include "memory.pub"
#include "shrip.pub"
#include "shros.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_rescal.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

#define	cResFracPos	16
#define cResFracRnd	0x8000


/**************************************************************************
 * i_rescaleArb()
 * 	Rescale the given image into the destination using a
 *	linear weighted interpolation scheme.
 *
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		sDepth = depth of source image
 *		pD = destination gray image
 *		wD = width in pixels (including frame)
 *		hD = height including frame
 *		dBpl = source bytes/line
 *		dDepth = depth of destination image
 *
 *	Current implementation assumes 8 bits/pixel.   
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *		ia_depthNotSupported not an 8 bit/pixel image
 *************************************************************************/
Int32   CDECL 
i_rescaleArb (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	sDepth,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    Int32	dDepth)
{
    PROLOGUE;
    Int32	 line;			/* interation variables */
    UInt32	 xOICoord, yOICoord;	/* split integers containing the
					 * X and Y coords of the current
					 * input pixel in the output 
					 * images's frame of reference.
					 */
    UInt8      *pDOrig;			/* used to remember addr of upper 
					 * left corner of image so we can 
					 * clear its right frame when done.
					 */
    UInt8	*pSOrig;		/* used to remember addor of upper
					 * left corner of source so we
					 * can clear the right and bottom
					 * frame when done.
					 */
    
    UInt16      *pRescaleTable; 	/* rescale table */
    UInt8	*pStateTable;		/* state table */
    UInt32	 xOIIncrement, yOIIncrement;
					/* split integer containing the amount
					 * to increment to get to the next
					 * input pixel in the output image's
					 * frame of reference.
					 */
    UInt32	 fourYOIIncrement;	/* 4 * yOIIncrement */
    UInt32	 threeYOIIncrement;	/* 3 * yOIIncrement */
    UInt16	*pTabExp2;		/* ptr to expansion table used
					     * to convert 4bpp to 8bpp */

    if (!((sDepth == 8 && dDepth == 8) || (sDepth == 4 && dDepth == 4)))
	return ia_depthNotSupported;

	/* get the width and height of the source without the frame*/
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    wD -= 2*cFrameBits;
    hD -= 2*cFrameBits;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + dDepth*cFrameBytes;

	/* Remember address of upper right corner of dest image
	 * so we can clear its right frame when done. */
    pDOrig = pD;
	/* Remember address of upper right corner of source image
	 * so we can clear its right and bottom frame when done. */
    pSOrig = pS;

    xOIIncrement = (UInt32) ((((Float32) wS) / ((Float32) wD)) * (1 << cResXIPos));
    yOIIncrement = (UInt32) ((((Float32) hS) / ((Float32) hD)) * (1 << cResYIPos));
    threeYOIIncrement = yOIIncrement + yOIIncrement + yOIIncrement;
    fourYOIIncrement = yOIIncrement << 2;

	/* By decree we will look at the left corner of the pixel */
    xOICoord = 0;
    yOICoord = 0;

    if ((__error__ = ip_make8bppRescaleTable (pIPState,
					    &pRescaleTable, &pStateTable,
					    wD, xOIIncrement)) != ia_successful)
	NORMAL_RETURN;


	/* The inner loop does not check to make sure that it always 
	 * references on the image.  As a result, the last few
	 * destination pixels may use the last source pixel and the 
	 * first frame pixel.  This may result in a black line along the
	 * edge.  This may happen on both the right edge and along the
	 * bottom.  In order to keep the inner loop from checking for
	 * the edge, we will have the frame contain the last source pixel.
	 * The time this adds should be small in comparison to the time
	 * that it takes to rescale the image....  
	 */

	/* the right frame */
    i_rasterOp (pIPState, pS, wS, 0, 1, hS, sDepth, sBpl, PIX_SRC,
	        pS, wS - 1, 0, sDepth, sBpl);
	/* the bottom frame */
    i_rasterOp (pIPState, pS + (hS * sBpl), 0, 0,
		wS + 1, 1, sDepth, sBpl, PIX_SRC,
	        pS + ((hS - 1) * sBpl), 0, 0, sDepth, sBpl);

    if (sDepth == 4)
    {
	/* Fun: for the 4-bit case, rather than re-write that whole state
	 * table and look-up table business, we're going to just convert
	 * the two lines of interest to 8 bit, process an output line
	 * one at a time, and then convert it back.
	 * Note that we potentially convert each line to 8 bit twice.
	 */
	UInt8 *srcLines, *destLine;
	Int32 siFrameOffset, 
	      diFrameOffset,
	      sLinesBpl,
	      dLineBpl,
	      fs;

	fs = 2*cFrameBits;

        /* calculate Bpl for two 8-bit temp buffers */
        sLinesBpl = ((wS + fs) + 3) & 0xfffffffc;
	dLineBpl = ((wD + fs) + 3) & 0xfffffffc;

        /* allocate 2 lines of 8-bit data w/ L & R frames which
	   will hold the result of 4-8 depth conversions */
/*	if (!(srcLines = (UInt8 *) MALLOC ((wS + fs)*2))) */
	if (!(srcLines = (UInt8 *) CALLOC (1, sLinesBpl * 2))) 
	    return ia_nomem;

        /* allocate 1 line of 8-bit data w/ L & R frames which
	   will hold the result of scaling the source lines above */
/*	if (!(destLine = (UInt8 *) MALLOC (wD + fs))) */
	if (!(destLine = (UInt8 *) CALLOC (1, dLineBpl))) 
	{
	    FREE (srcLines);
	    return ia_nomem;
	}

/*	diFrameOffset = ((wD + fs) << cLogFrameBits) + 8*cFrameBytes;
	siFrameOffset = ((wS + fs) << cLogFrameBits) + 8*cFrameBytes;
*/
	siFrameOffset = ((sLinesBpl) << cLogFrameBits) + 8*cFrameBytes;
	diFrameOffset = ((dLineBpl) << cLogFrameBits) + 8*cFrameBytes;


	/* Move past the left frame in the src, dest lines. 
	   Note that there is no top, bottom frame. */
	srcLines += 8*cFrameBytes;
	destLine += 8*cFrameBytes;

	/* Since the 4to8 and 8to4 routines take frame-relative coordinates, we're going to move
	   the pointers back.. */
	pS -= (sBpl << cLogFrameBits) + 4*cFrameBytes;
	pD -= (dBpl << cLogFrameBits) + 4*cFrameBytes;

	    /* get the table used to convert 4 bpp to 8bpp.  We'll free
	     * this at the bottom of the loop.
	     */
	pTabExp2 = NULL;
	if ((__error__ = i_makeExp2LUT (pIPState, 4, &pTabExp2)) != ia_successful)
	{
	    FREE (srcLines);
	    FREE (destLine);
	    NORMAL_RETURN;
	}

	for (line = hD; line > 0; line--)
	{
	    /* Copy two lines out to srcLines, expanding to 8 bits/pixel */
	    i_depth4To8bpp (pIPState, pS, wS + fs, 2 + fs, sBpl, 
			    srcLines - siFrameOffset,
			    sLinesBpl, pTabExp2); 

/*			    wD + fs); */

/*			    ((wD + fs) + 3) & 0xfffffffc); */

	    /* Generate an output line */
/*	    ip_arbRescale8bppOneLine (pIPState, srcLines,
				      wS + fs, destLine, wD, */
	    ip_arbRescale8bppOneLine (pIPState, srcLines,
				      sLinesBpl, destLine, wD, 
	                              xOICoord, yOICoord,
				      xOIIncrement, pRescaleTable, pStateTable);

	    /* Convert the line back and stuff it into the destination */
/*	    i_threshold8To4bpp (pIPState, destLine - diFrameOffset,
				wD + fs, 1 + fs, */
	    i_threshold8To4bpp (pIPState, destLine - diFrameOffset,
				dLineBpl, 1 + fs, 
	                        dLineBpl, pD, dBpl);

	    yOICoord = (yOICoord & cResYFAMask) + yOIIncrement;
	    pD += dBpl;
	    pS += (yOICoord >> cResYIPos) * sBpl;

	    IP_YIELD (aborted);
	}

	srcLines -= 8*cFrameBytes;
	destLine -= 8*cFrameBytes;

	FREE (pTabExp2);
	FREE (srcLines);
	FREE (destLine);

	/* Move the pointers up again, so that the clear-frame business doesn't blow up */
	pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
	pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;

    }
    else
    {

	/* we need to make sure that we do not write into the bottom
	 * frame of the destination.  if we allowed line to start at
	 * hD, we could conceivably write into the frame if we take
	 * the ip_arbRescale8bppFourLines path through the loop and
	 * we don't have exactly 4 lines left at the bottom of the image.
	 * to keep this from happening we start out with less than
	 * the full height to allow some pad at the bottom.  at the
	 * end, line may be negative.  this is ok since we couldn't have
	 * written into the frame.  the next loop will only do the
	 * amount that it needs, one destination line at a time.
	 * if we were doing one destination line at a time anyway,
	 * we'll just do all but three of them in the first loop and
	 * then the remaining 3 in the last loop.
	 */
	for (line = hD-3; line > 0; )
	{
	    if ((Int32)(((yOICoord & cResYFAMask) + threeYOIIncrement) >> cResYIPos) > 0)
	    {	   /* then can only do 1 destination line since we
		    * are using more than 2 source lines indicated by the
		    * integer portion being greater than 0.
		    */
		ip_arbRescale8bppOneLine (pIPState, pS, sBpl, pD, wD,
					  xOICoord, yOICoord,
					  xOIIncrement, pRescaleTable,
					  pStateTable);
		yOICoord = (yOICoord & cResYFAMask) + yOIIncrement;
		line--;
		pD += dBpl;
	    }
	    else
	    {
		ip_arbRescale8bppFourLines (pIPState, pS, sBpl, pD, wD,
					    dBpl, xOICoord, 
					    yOICoord, xOIIncrement,
					    yOIIncrement, 
					    pRescaleTable, pStateTable);
		yOICoord = (yOICoord & cResYFAMask) + fourYOIIncrement;
		line -= 4;
		pD += dBpl << 2;
	    }
	    pS += (yOICoord >> cResYIPos) * sBpl;
	    }
	for (line += 3; line > 0; line--)
	{
	    ip_arbRescale8bppOneLine (pIPState, pS, sBpl, pD, wD,
				      xOICoord, yOICoord,
				      xOIIncrement, pRescaleTable, pStateTable);
	    yOICoord = (yOICoord & cResYFAMask) + yOIIncrement;
	    pD += dBpl;
	    pS += (yOICoord >> cResYIPos) * sBpl;
	}

	IP_YIELD (aborted);
    }

	/* clear the right destination frame since we always write
	 * a 32-bit quantity in each loop iteration.
	 */
    ip_clearRightFrame (pIPState, pDOrig, wD, hD, dBpl, dDepth);
	/* Now clear the right and bottom source frames that we
	 * wrote into above.  Can't use ip_clearRightFrame because
	 * if the rightmost word of the image contained only image and
	 * NO frame bits, the frame is not cleared.
	 */

    IP_YIELD (aborted);

    i_rasterOp (pIPState, pSOrig, wS, 0, 1, hS, sDepth, sBpl, PIX_CLR,
	        NULL, 0, 0, 0, 0);
    i_rasterOp (pIPState, pSOrig + (hS * sBpl), 0, 0, 
	        wS + 1, 1, sDepth, sBpl, PIX_CLR,
	        NULL, 0, 0, 0, 0);

    FREE(pRescaleTable);
    FREE(pStateTable);
    NORMAL_RETURN;

aborted:
    FREE(pRescaleTable);
    FREE(pStateTable);
    
    return ia_aborted;
}


/**************************************************************************
 * i_rescaleAreaMap()
 * 	Magnify the given image into the destination using area mapping.
 *
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		sDepth = depth of source image
 *		pD = destination gray image
 *		wD = width in pixels (including frame)
 *		hD = height including frame
 *		dBpl = source bytes/line
 *		dDepth = depth of destination image
 *
 *	Current implementation assumes 8 bits/pixel. 
 *	This implementation words for [100%..200%) scaling  but the best
 *	image is really only in the [100%..150%] range.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *		ia_depthNotSupported not an 8 bit/pixel image
 *************************************************************************/
Int32   CDECL 
i_rescaleAreaMap (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	wS,
    UInt32	hS,
    UInt32	sBpl,
    UInt32	sDepth,
    UInt8      *pD,
    UInt32	wD,
    UInt32	hD,
    UInt32	dBpl,
    UInt32	dDepth)
{
PROLOGUE;
 UInt32		line;			/* interation variables */
UInt32	 	xIOCoord, yIOCoord;	/* split integers containing the
					 * X and Y coords of the current
					 * input pixel in the output 
					 * images's frame of reference.
					 */
UInt8	       *pDOrig;			/* used to remember addr of upper 
					 * left corner of image so we can 
					 * clear its right frame when done.
					 */

UInt16         *pRescaleTable; 		/* rescale table */
UInt8          *pStateTable; 		/* state table */
UInt32		xCount, yCount;	/* split integer containing the number of
					 * X and Y pixels to process before
					 * adding an extra destination pixel
					 */
UInt32		xIOIncrement, yIOIncrement;
					/* split integer containing the amount
					 * to increment to get to the next
					 * input pixel in the output image's
					 * frame of reference.
					 */
UInt32 	 	extraLine;		/* is it time to add the extra dest line? */

#define cResSIntMask	0xFFFF0000	/* integer portion of
					 * split integer.
					 */
#define cResSFracMask	0xFFFF		/* fraction portion of split
					 * integer
					 */

    if ((sDepth != 8) || (dDepth != 8))
	return ia_depthNotSupported;

	/* get the width and height of the source without the frame*/
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    wD -= 2*cFrameBits;
    hD -= 2*cFrameBits;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + dDepth*cFrameBytes;

	/* Remember address of upper right corner of dest image
	 * so we can clear its right frame when done. */
    pDOrig = pD;

	/* By decree the upper left hand corners of the input pixel
	 * and the output pixel are the same.
	 */
    xIOCoord = 0;
    yIOCoord = 0;

	/* For every xCount pixels in the horizontal direction we produce
	 * xCount+1 pixels in the destination image.
	 * For every yCount pixels in the vertical direction we produce
	 * yCount+1 lines in the destination image.
	 * It would be easier if we could always assume that these were
	 * integers.  Unfortunately, as the magnification factor approaches
	 * 2x, *Count would truncate to 1 which would add too many 
	 * pixels into the image.  Hence, we will keep these as 
	 * split integers.  We will only look at the integer part when
	 * deciding to add a new pixel (or line if this is yCount) but
	 * will add the fractional part to the next count in the hopes
	 * that eventually the fractional part will cause an increment
	 * into the integer part and so we won't add as many pixels.
	 */
    if (wD == wS)
	xCount = wS * (1 << 16);
    else xCount = (Int32)(((Float32)wS) / ((Float32)(wD - wS)) * (1 << 16));
    if (hS == hD)
	yCount = hS;
    else yCount = (Int32)(((Float32)hS) / ((Float32)(hD - hS)) * (1 << 16));

    xIOIncrement = (Int32) ((((Float32) wD) / ((Float32) wS)) * (1 << 16));
    yIOIncrement = (Int32) ((((Float32) hD) / ((Float32) hS)) * (1 << 16));

    if ((__error__ = ip_make8bppAreaMapTable (pIPState,
					    &pRescaleTable, &pStateTable,
					    wD, hD,
					    xIOIncrement, yIOIncrement,
					    xCount)) != ia_successful)
	NORMAL_RETURN;

    extraLine = yCount;
    for (line = 0; line < hS; line++)
    {
	ip_areaMap8bppLine (pIPState, pS, sBpl, pD, wD, xIOCoord, yIOCoord,
			    xIOIncrement, xCount, pRescaleTable, pStateTable);
	pD += dBpl;
	yIOCoord += yIOIncrement;
	    /* do we need to process an extra line this time? */
	extraLine -= 0x10000;
	if ((extraLine & cResSIntMask) == 0)
	{
	    yIOCoord &= cResSIntMask;
	    extraLine = (extraLine & cResSFracMask) + yCount;
	    ip_areaMap8bppLine (pIPState, pS, sBpl, pD, wD, xIOCoord, 
				yIOCoord, xIOIncrement, xCount,
				pRescaleTable, pStateTable);
	    pD += dBpl;
	}
	pS += sBpl;
	IP_YIELD (aborted);
    }

	/* clear the right destination frame since we always write
	 * a 32-bit quantity in each loop iteration.
	 */
    ip_clearRightFrame (pIPState, pDOrig, wD, hD, dBpl, dDepth);

    FREE(pRescaleTable);
    FREE(pStateTable);
    NORMAL_RETURN;

aborted:
    FREE(pRescaleTable);
    FREE(pStateTable);
    return ia_aborted;
}



/***************************************************************************
 *			NON-INTERPOLATED SCALING
 ***************************************************************************/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Arbitrary rescaling by sub/super sampling 
 *
 * LongDescription:::
 *	Handles any isomorphic or anamorphic cases.  Images must be 1, 4, 
 *	or 8 bits in depth.
 *
 * PrivateDescription:::
 *
 * Input:::
 *	pS::		Pointer to source image data including frames
 *	wS::		Width of source image to be rescaled
 *	hS::		Height of source image to be rescaled
 *	sBpl::		Bytes per line of source image
 *	sDepth::	Depth of source image (1, 4, or 8)
 *	wD::		Width of dest image to be rescaled
 *	hD::		Height of dest image to be rescaled
 *	dBpl::		Bytes per line of dest image
 *	dDepth::	Depth of dest image (1, 4, or 8).  Must be same as src.
 *
 * Output:::
 *	pD::		Pointer to dest image data including frames
 *
 * Returns:::
 *	ia_successful	if completes successfully
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32   CDECL 
i_rescaleNoInterp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	sDepth,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    Int32	dDepth)
{
    PROLOGUE;
    Int32      tableSize;
    void      *pTable = NULL;

    if ((sDepth != 1) && (sDepth != 4) && (sDepth != 8))
	return (ia_depthNotSupported);

    if (sDepth != dDepth)
	return (ia_depthNotSupported);

    __error__ = i_getScaleToSizeTableSize (pIPState, sDepth, 
					 wD - (2 * cFrameBits),
					 hD - (2 * cFrameBits),
					 &tableSize);
    if (__error__ != ia_successful)
        goto cleanup;

    /* Allocate the index tables */
    /* each entry in the array indicates which source pixel (offset
     * from the left side of the line) to get that destination
     * pixel.
     */
    if ((pTable = (void *)MALLOC (tableSize)) == NULL)
    {
        __error__ = ia_nomem;
	goto cleanup;
    }

    __error__ = i_initScaleToSizeTables (pIPState, wS - (2 * cFrameBits),
                                       hS - (2 * cFrameBits),
				       sBpl, sDepth,
                                       wD - (2 * cFrameBits),
                                       hD - (2 * cFrameBits),
				       tableSize,
				       pTable);
    if (__error__ != ia_successful)
        goto cleanup;

    __error__ = i_rescaleNoInterpSubRegion (pIPState, pS, wS, hS, sBpl, sDepth,
                                          pTable,
					  pD, 0, 0, wD, hD, dBpl, dDepth);

cleanup:
    if (pTable != NULL)
	FREE (pTable);

    NORMAL_RETURN;

}  /* i_rescaleNoInterp () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Computes table size for arbitrary sub/super sampling rescaling
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *	depth::		Depth of image being scaled
 *	destW::		Desired destination width for scaling op
 *	destH::		Desired destination height for scaling op
 *
 * Output:::
 *	pTableSize::	Pointer to total size of all tables.  These
 *			should be allocated in one block.
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_getScaleToSizeTableSize (
    IPState    *pIPState,
    Int32	depth,
    Int32	destW,
    Int32	destH,
    Int32      *pTableSize)
{
    PROLOGUE;
    Int32  xIndexSize, yOffsetSize, xShiftSize;
    

	/* check parameters */
    if ((pTableSize == NULL) || (destW <= 0) || (destH <= 0))
	return ia_invalidParm;

    if ((depth != 1) && (depth != 4) && (depth != 8))
	return ia_depthNotSupported;

	/* get the sizes of the constitutient tables */
    __error__ = ip_getAllScaleToSizeTableSizes (pIPState, depth, destW, destH,
					      &xIndexSize, &yOffsetSize,
					      &xShiftSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* It worked, so set the total table size and return */
    *pTableSize = sizeof(NonInterpScalingInfo) +
		  xIndexSize + yOffsetSize + xShiftSize;

    NORMAL_RETURN;

}  /* i_getScaleToSizeTableSize () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Computes all three table sizes for arbitrary sub/super sampling rescaling
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *	depth::		Depth of image being scaled
 *	destW::		Desired destination width for scaling op
 *	destH::		Desired destination height for scaling op
 *
 * Output:::
 *	xIndexSize::	Horizontal index table size in bytes
 *	yOffsetSize::	Vertical index table size in bytes
 *	xShiftSize::	Horizontal shift table size in bytes (may be zero)
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_getAllScaleToSizeTableSizes (
    IPState    *pIPState,
    Int32	depth,
    Int32	destW,
    Int32	destH,
    Int32      *xIndexSize,
    Int32      *yOffsetSize,
    Int32      *xShiftSize)
{
    PROLOGUE;

    *xIndexSize = destW * sizeof (Int32);
    *yOffsetSize = destH * sizeof (Int32);

    if (depth == 4)
        *xShiftSize = destW;
    else
        *xShiftSize = 0;

    NORMAL_RETURN;

}  /* ip_getAllScaleToSizeTableSizes () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Initializes tables for arbitrary sub/super sampling rescaling
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *	srcW::		Width of entire source image to be rescaled
 *	srcH::		Height of entire source image to be rescaled
 *	srcBpl::	bytes per line of source image
 *	depth::		Bit depth of source and destination images (1, 4, or 8)
 *	destW::		Width of entire scaled depth image.  This is the
 *			width you would get if you scaled the entire
 *			source image.
 *	destH::		Height of entire scaled depth image.  This is the
 *			height you would get if you scaled the entire
 *			source image.
 *	tableSize::	Number of bytes in composite table passed in.
 *
 * Output:::
 *	pTable::	Pointer to composite table to be initialized.
 *			Must be nonNULL.
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_initScaleToSizeTables (
    IPState    *pIPState,
    Int32	srcW,
    Int32	srcH,
    Int32	srcBpl,
    Int32	depth,
    Int32	destW,
    Int32	destH,
    Int32	tableSize,
    void       *pTable)
{
    PROLOGUE;
    NonInterpScalingInfo *pScaleTables;
    UInt32	wFactor, hFactor, fracIndex;
    Int32	*xIndex, *yOffset;
    UInt8	*xShift;
    Int32	xIndexSize, yOffsetSize, xShiftSize;
    Int32	i;

    if (pTable == NULL)
        return (ia_invalidParm);

	/* get the sizes of the constitutient tables */
    ERR_PASS( ip_getAllScaleToSizeTableSizes (pIPState, depth, destW, destH,
					      &xIndexSize, &yOffsetSize,
					      &xShiftSize));

	/* check the sizes against the total table size */
    if (xIndexSize + yOffsetSize + xShiftSize +
			(Int32)sizeof(NonInterpScalingInfo) > tableSize)
	return ia_invalidParm;

	/* Setup the structure entries. */
    pScaleTables = (NonInterpScalingInfo *)pTable;
    pScaleTables->numXIndexEntries = xIndexSize / sizeof(Int32);
    pScaleTables->xIndexSize = xIndexSize;
    pScaleTables->numYOffsetEntries = yOffsetSize / sizeof(Int32);
    pScaleTables->yOffsetSize = yOffsetSize;
    pScaleTables->numXShiftEntries = xShiftSize / sizeof(UInt8);
    pScaleTables->xShiftSize = xShiftSize;

	/* set the individual table pointers. */
    xIndex = (Int32 *) ((UInt8 *) pScaleTables + sizeof(NonInterpScalingInfo));
    yOffset = (Int32 *) ((UInt8 *)xIndex + xIndexSize);
    xShift = (UInt8 *) ((UInt8 *)yOffset + yOffsetSize);

    if ((depth == 1) || (depth == 4) || (depth == 8))
    {
	/* use split integers instead of messing with floating point */
	wFactor = Int32_TO_ScaledInt1616(srcW) / destW;
	hFactor = Int32_TO_ScaledInt1616(srcH) / destH;

	for (i = 0, fracIndex = 0; 
	     i < destW; i++)
	{
		/* note that we truncate, not round, here */
	    xIndex[i] = fracIndex >> cResFracPos;
	    if (depth == 4)
	    {
		xShift[i] = (xIndex[i] & 1) ? 0 : 4;
		xIndex[i] >>= 1;
	    }
    
	    fracIndex += wFactor;
	}
	for (i = 0, fracIndex = 0; 
	     i < destH; i++)
	{
		/* note that we truncate, not round, here */
	    yOffset[i] = srcBpl * (fracIndex >> cResFracPos);
	    fracIndex += hFactor;
	}
    }
    else
	return (ia_invalidParm);

    NORMAL_RETURN;

}  /* i_initScaleToSizeTables () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Return dimensions encoded in scaling tables.
 *
 * LongDescription:::
 *	A client may create a set of scaling tables, then forget the
 *	dimensions of the virtual dest image that's encoded in them.
 *	This routine reads the dimensions encoded in the tables.
 *
 * PrivateDescription:::
 *
 * Input:::
 *	pTable::	Pointer to scaling tables
 *
 * Output:::
 *	pWidth::	Pointer to width of virtual dest image.
 *	pHeight::	Pointer to height of virtual dest image.
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
i_getDimensionsFromScalingTables (
    IPState    *pIPState,
    void       *pTable,
    Int32      *pWidth,
    Int32      *pHeight)
{
    PROLOGUE;
    NonInterpScalingInfo *pScaleTables;

    if ((pTable == NULL) || (pWidth == NULL) || (pHeight == NULL))
	return ia_invalidParm;

    pScaleTables = (NonInterpScalingInfo *) pTable;

    *pWidth = pScaleTables->numXIndexEntries;
    *pHeight = pScaleTables->numYOffsetEntries;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Arbitrary rescaling by sub/super sampling.
 *
 * LongDescription:::
 *	Expanded interface for selecting a subregion of the source image
 *	to be rescaled.  Handles any isomorphic or anamorphic case.
 *	Images must be 1, 4, or 8 bits in depth.
 *	<BR><BR>
 *	This routine takes pointers to an entire source image.  A rectangle
 *	of this source image will be scaled.  One might like to specify
 *	the coordinates of this rectangle relative to the source image,
 *	but the client can't know the exact coordinates in the source
 *	space since they depend on the vagaries of the scaling function.
 *	That is, the choice of exactly which source pixel will be chosen for
 *	a given dest pixel is not public information.
 *	<BR><BR>
 *	Therefore, we adapt the strategy used by i_scaleToGray4bppArb.
 *	The client supplies an entire source image and a dest image that's
 *	just large enough to hold the scaled rectangle.  The coordinates
 *	of the rectangle are given relative to a virtual destination image,
 *	i.e., the image one would get if one scaled the entire source
 *	image.  [xD, yD] are the coordinates of the upper left corner of
 *	the dest rectangle inside the virtual dest image.
 *	<BR><BR>
 *	Since there are frames on both the real dest image and the
 *	virtual dest image, it doesn't matter if you think of these
 *	coordinates as being relative to the insides of both frames
 *	(offset of first real non-frame pixel relative to first virtual image
 *	non-frame pixel) or the outsides of both frames (offset of upper
 *	left corner of real frame relative to upper left corner of virtual
 *	image frame) so long as you're consistent.  For the sake of consistency
 *	with the rest of IPShared, we'll assume specify them relative
 *	to the outsides of the frames.
 *
 * PrivateDescription:::
 *
 * Input:::
 *	pS::		Pointer to source image data including frames
 *	wS::		Width of entire source image to be rescaled
 *			including frame.
 *	hS::		Height of entire source image to be rescaled
 *			including frame.
 *	sBpl::		Bytes per line of source image
 *	sDepth::	Depth of source image (1, 4, or 8)
 *	pTable::	Pointer to composite table.  Must be nonNULL.
 *	xD::		X offset into framed virtual dest image of
 *			frame of real dest image.
 *	yD::		Y offset into framed virtual dest image of
 *			frame of real dest image.
 *	wD::		Width of real dest image including frame.
 *	hD::		Height of real dest image including frame.
 *	dBpl::		Bytes per line of dest image
 *	dDepth::	Depth of dest image (1, 4, or 8).  Must be same as src.
 *
 * Output:::
 *	pD::		Pointer to dest image data including frames
 *
 * Returns:::
 *	ia_successful	if completes successfully
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32   CDECL 
i_rescaleNoInterpSubRegion (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	sDepth,
    void       *pTable,
    UInt8      *pD,
    Int32	xD,
    Int32	yD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    Int32	dDepth)
{
    PROLOGUE;
    NonInterpScalingInfo *pScaleTables;
    Int32	row, col;
    Int32	i;
    UInt8      *sPtr, *dPtr;
    Int32      *xIndex, *yOffset;
    UInt8      *xShift;
    Int32      *pXIndex, *pYOffset, ix;
    UInt32	grayOut;
    Int32	currYOffset, prevYOffset;
    UInt32	leftoverDestWidth;	/* number of bytes left
					 * after all the 4-byte
					 * words have been written.
					 */
    Int32	fullWordWidth; 		/* width if only full
					 * 4-byte words are 
					 * considered. */
    
    UInt8	grayOutByte;


    if ((sDepth != 1) && (sDepth != 4) && (sDepth != 8))
	return ia_depthNotSupported;

    if (sDepth != dDepth)
	return ia_depthNotSupported;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + dDepth*cFrameBytes;

	/* get the width and height of the source without the frame*/
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    wD -= 2*cFrameBits;
    hD -= 2*cFrameBits;

	/* get the sizes of the portions of the tables that we'll
	 * use in this procedure.  We want to make sure we won't
	 * run off the end.
	 */
    pScaleTables = (NonInterpScalingInfo *) pTable;

	/* set the individual table pointers. */
    xIndex = (Int32 *) ((UInt8 *)pScaleTables + sizeof(NonInterpScalingInfo));
    yOffset = (Int32 *) ((UInt8 *)xIndex + pScaleTables->xIndexSize);
    xShift = (UInt8 *) ((UInt8 *)yOffset + pScaleTables->yOffsetSize);

	/* See if the indices given run off the source or dest images. */
    if ((xD < 0) || (xD + wD - 1 > pScaleTables->numXIndexEntries) ||
	(yD < 0) || (yD + hD - 1 > pScaleTables->numYOffsetEntries) ||
	((dDepth == 4) && (xD + wD - 1 > pScaleTables->numXShiftEntries)) ||
	(xIndex[xD + wD - 1] > wS - 1) ||
	(yOffset[yD + hD - 1] > (hS - 1) * sBpl))
	return ia_invalidParm;

    if (sDepth == 1)
    {
	UInt8   *lineS, *prevLineS;
	UInt8   *lineD;
        UInt32  wordS;           /* current source word */
	UInt32  wordD;           /* current destination word */
	Int32   wordIndexS;      /* 32-bit source word index */
	Int32   prevWordIndexS;  /* last source word index */
	Int32   wplD;            /* whole words/line in dest pixr */
	Int32   jMod;            /* remainder bits in last dest word */
	Int32  *xIndexP;         /* current x LUT index */
	UInt32 *wpD;             /* pointer to current dest word */
	UInt32 *wpDE;            /* pointer to last dest word on a row */
        Int32   j, bitS;
	UInt8  *pdbyte, *pdbyteLast;

	wplD = wD / 32;
	wordS = 0;


	/* march over the new dimensions */
	prevLineS = NULL;		/* y position of previous source line */
	lineD     = pD;			/* ptr to beginning of dest line */
	for (i = yD; i < yD + hD; i++)
	{
	    lineS = pS + yOffset[i];	/* pt to beginning of src line */
	    pdbyte = lineD;
	    if (lineS != prevLineS) 
	    {  /* compute new dest line */

                /* process all the complete words in the destination */
	        wpDE = ((UInt32 *)lineD) + wplD;
		xIndexP = &(xIndex[xD]);
		wordS = 0;
		prevWordIndexS = -1L;

                /* process the destination image rows a word at a time */
		for (wpD = (UInt32 *)lineD; wpD != wpDE; wpD++)
		{
		    wordD = 0L;

                    /* cobble together the destination word a bit at a time */
                    for (j = 0; j < 32; j++)
		    {
			bitS = *xIndexP;
			xIndexP++;
			wordIndexS = bitS >> 5;

                        /* Stepped over source word boundary? */
			if (wordIndexS != prevWordIndexS)
			{
			    /* read new source word */
			    wordS = *(UInt32 *)(((UInt32 *)lineS) + wordIndexS);
			    prevWordIndexS = wordIndexS;
			}

                        /* update the destination word */
			wordD = (wordD + wordD) + 
			        ((wordS >> (~bitS & 31)) & 1L);

		    }

                    /* write out the newly created destination word */
		    *wpD = wordD;
		}

                /* get the last word in if it is incomplete */
		if ((jMod = wD & 0x001f))  /* YES, it is supposed to be "=" */
		{
		    wordD = 0L;
		    prevWordIndexS = -1L;

		    for (j = 0; j < jMod; j++)
		    {
			bitS = *xIndexP;
			xIndexP++;
			wordIndexS = bitS >> 5;

			if (wordIndexS != prevWordIndexS)
			{
			    /* read new source word */
			    wordS = *(UInt32 *)(((UInt32 *)lineS) + wordIndexS);
			    prevWordIndexS = wordIndexS;
			}

                        /* update the destination word */
			wordD = (wordD + wordD) + 
			        ((wordS >> (~bitS & 31)) & 1L);

		    }

		    /* shift the word by the remaining amount */
		    wordD = wordD << (32 - jMod);

		    *wpD = wordD;
		}

	    }
	    else  /* repeat previous dest line */
	    {
		pdbyteLast = pdbyte - dBpl;
		memcpy((char*)pdbyte, (char*)pdbyteLast, dBpl);
		pdbyte += dBpl;
	    }
	    prevLineS = lineS;
	    lineD += dBpl;
	}    
      
    }
    else  /* 4 or 8 bit depth */
    {
	pYOffset = &(yOffset[yD]);
	prevYOffset = -1;

	    /* the inner loop will write 4-byte quantities.  width should
	     * reflect this count.  in order to keep from writing into the
	     * frame, the remainder bytes will be written one byte at a time
	     * instead as of a 4-byte quantity.
	     * BUG: when doing the memcpy, we need to write in units of 4
	     * bytes to make it work on little endian platforms.  This
	     * copies some of the frame from the line above, but that's ok.
	     */
	if (sDepth == 8)
	{
	    /* The 8 bit/pixel case */
	    leftoverDestWidth = wD & 0x3;
	    fullWordWidth = wD & ~0x3;
	    for (row = hD; row > 0; row--, pD += dBpl)
	    {
		currYOffset = *pYOffset++;
		if (prevYOffset == currYOffset)
			/* This is line is the same as the last one, so
			 * copy it.
			 */
    
		    memcpy((void *)pD, (void *)(pD-dBpl), ((wD + 3) & ~0x3));
		else 
		{				/* calculate the new line */
		    prevYOffset = currYOffset;
		    sPtr = pS + currYOffset;
		    dPtr = pD;
		    pXIndex = &(xIndex[xD]);
			/* write one destination word for each loop iteration */
		    for (col = fullWordWidth; col > 0; col -= 4)
		    {
			grayOut = UCHAR_ACCESS(&sPtr[pXIndex[0]]) << 24;
			grayOut |= UCHAR_ACCESS(&sPtr[pXIndex[1]]) << 16;
			grayOut |= UCHAR_ACCESS(&sPtr[pXIndex[2]]) << 8;
			grayOut |= UCHAR_ACCESS(&sPtr[pXIndex[3]]);
			pXIndex += 4;
			ULONG_ACCESS(dPtr) = grayOut;
			dPtr += 4;
		    }
		    for (col = leftoverDestWidth; col > 0; col--)
		    {
			UCHAR_ACCESS(dPtr++) = UCHAR_ACCESS(&sPtr[*pXIndex++]);
		    }
		}
		IP_YIELD (aborted);
	    }
	}
	else
	{
	    /* The 4 bit/pixel case */
	    leftoverDestWidth = wD & 0x7;
	    fullWordWidth = wD & ~0x7;
	    for (row = hD; row > 0; row--, pD += dBpl)
	    {
		currYOffset = *pYOffset++;
		if (prevYOffset == currYOffset)
		    /* duplicate the previous line. */
		    memcpy((void *)pD, (void *)(pD-dBpl), (((wD + 7)/2) & ~0x3));
		else 
		{				/* calculate the new line */
		    prevYOffset = currYOffset;
		    sPtr = pS + currYOffset;
		    dPtr = pD;
		    ix = xD;
			/* write one destination word for each loop iteration */
			/* I suppose that one could have some scheme for dealing
			 * with the case where two output pixels come from the
			 * same input byte, but it hardly seems worth it.
			 */
		    for (col = fullWordWidth; col > 0; col -= 8)
		    {
			grayOut = ((UCHAR_ACCESS(&sPtr[xIndex[ix]]) >> xShift[ix]) 
				   & 0xf) << 28;
			ix++;
			grayOut |=((UCHAR_ACCESS(&sPtr[xIndex[ix]]) >> xShift[ix]) 
				   & 0xf) << 24;
			ix++;
			grayOut |=((UCHAR_ACCESS(&sPtr[xIndex[ix]]) >> xShift[ix]) 
				   & 0xf) << 20;
			ix++;
			grayOut |=((UCHAR_ACCESS(&sPtr[xIndex[ix]]) >> xShift[ix]) 
				   & 0xf) << 16;
			ix++;
			grayOut |=((UCHAR_ACCESS(&sPtr[xIndex[ix]]) >> xShift[ix]) 
				   & 0xf) << 12;
			ix++;
			grayOut |=((UCHAR_ACCESS(&sPtr[xIndex[ix]]) >> xShift[ix]) 
				   & 0xf) << 8;
			ix++;
			grayOut |=((UCHAR_ACCESS(&sPtr[xIndex[ix]]) >> xShift[ix]) 
				   & 0xf) << 4;
			ix++;
			grayOut |=((UCHAR_ACCESS(&sPtr[xIndex[ix]]) >> xShift[ix]) 
				   & 0xf);
			ix++;
    
			ULONG_ACCESS(dPtr) = grayOut;
			dPtr += 4;
		    }
    
		    /* do it 2-at-a-time */
		    for (col = leftoverDestWidth; col > 0; col -= 2)
		    {
			grayOutByte = ((UCHAR_ACCESS(&sPtr[xIndex[ix]]) 
					>> xShift[ix])  & 0xf) << 4;
			ix++;
    
			if (col > 1)
			    grayOutByte |= ((UCHAR_ACCESS(&sPtr[xIndex[ix]]) 
					     >> xShift[ix])  & 0xf);
			ix++;
    
			UCHAR_ACCESS(dPtr++) = grayOutByte;
		    }
		}
    
		IP_YIELD (aborted);
	    }
	}
    }
    NORMAL_RETURN;

aborted:
    return ia_aborted;

}  /* i_rescaleNoInterpSubRegion () */



/*

$Log: i_rescal.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.32  1998/09/03 01:12:37  aber
aber on Wed Sep 2 18:11:58 PDT 1998

i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

Revert back to original, until I can resolve differences in
regression tests and/or file system problems.

Revision 1.31  1998/08/31 23:44:39  aber
aber on Mon Aug 31 16:43:06 PDT 1998

i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

Round to the neares line number instead of truncating. This causes real
noticable affects when used as the scaling function for a masked image.
You'll see a 0 pixel ring on one side, but a 2 pixel ring on the opposite.

Revision 1.30  1998/02/24 03:45:25  ddavies
i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

Moved NonInterpScalingInfo definition to shrscal.prv.  There's no reason
why other scaling functions should reinvent this.

Revision 1.29  1998/01/13 23:49:24  ddavies
i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.28  1997/05/07 23:53:10  ddavies
ddavies on Wed May 7 16:41:38 PDT 1997

i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

Big rework of non-interpolated scaling of partial images.  In the old code,
the tables had to be recomputed for each window into the source or dest.
In the new code, the tables depend only on the size and depth of the
source image and the scale factor.  This means we pass this info to
i_initScaleToSizeTables.  We also have to pass the x,y offset inside a
virtual dest image (the one you'd get by scaling the entire source) to
i_rescaleNoInterpSubRegion instead of having this offset implicitly
encoded in the tables.
A new structure, NonInterpScalingInfo now holds all the tables and misc
info associated with the tables.  This means we no longer have to pass
this stuff explicitly (like dimensions of the virtual dest image).
i_getScaleToSizeTableSizes has been replaced by i_getScaleToSizeTableSize,
which returns the size of the NonIinterptScalingInfo structure that holds
all the tables.  That replaces three things with one thing.  The contents
of the tables are no longer visible.
The yIndex table has been replaced with the yOffset table, so we don't have
to multiply by sourceBytes/line each time a yIndex entry is read (it's
pre-multiplied in the table).
The one floating pt calculation has been replaced with a fixed point
calculation for the sake of repeatability across platforms.
Added i_getDimensionsFromScalingTables which returns the width and height
of the virtual dest image.  This relieves the client of having to remember
this between the time the tables are created and when they're used.
Lots of rework of i_rescaleNoInterpSubRegion.  It now takes a rectangle
in the virtual dest image instead of in the source image.  It also takes
whoops, I already said that.  It indexes into the scaling tables (which
start at 0) to decide how to produce dest pixel x,y instead of having the
tables remade for each window into the source/dest.  It uses yOffset
(has bytes/line premultipled by yIndex) instead of doing the multiply
on each line access.

Revision 1.27  1997/04/17 22:48:19  dferrell
dferrell on Thu Apr 17 15:46:24 PDT 1997

i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

Removed two unused parameters from i_rescaleNoInterpSubRegion declaration.
Changed offset table creation for binary rescaling from using floating
point computation to split int (fixed point) to eliminate platform
dependent disagreements due to roundoff.

Revision 1.26  1997/04/16 18:21:33  dferrell
dferrell on Wed Apr 16 11:20:42 PDT 1997

i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

Added new functions for subrect arbitrary noninterpolated scaling.

Revision 1.25  1997/03/26 18:24:24  jcraig
jcraig on Wed Mar 26 10:23:16 PST 1997

i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

in i_rescaleArb, check to make sure that both the source and destination
depth are 8  - don't check the source depth twice and ignore the
destination depth.

Revision 1.24  1995/12/13 02:45:33  ejaquith
i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.23  1995/10/18  03:58:21  ddavies
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.22  1995/10/03  22:05:52  dferrell
 * dferrell on Tue Oct 3 15:01:54 PDT 1995
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * fixed bug in i_rescaleArb () which caused variable results on the right
 * and/or bottom edge(s) of 4-bit results.  The problem was that the
 * buffers used to up and down convert to and from 8 bit data were MALLOCed
 * and the uninitialized frames were read during the scaling operation
 * causing small intermittent errors at the edges of the result image.
 *
 * Revision 1.21  1995/03/01  19:54:12  dferrell
 * dferrell on Wed Mar 1 11:53:23 PST 1995
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Moved binary sub/super-sampling rescale code from ipwrap into
 * i_rescaleNoInterp ().
 *
 * Revision 1.20  1994/12/17  01:56:01  ddavies
 * ddavies on Fri Dec 16 17:54:20 PST 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Fix bug in i_rescaleNoInterp.  It was using memcpy to copy destination
 * lines that were the same.  Unfortunately, it wasn't always doing the
 * copies in units of 4 bytes, so it did the wrong thing on the right ends
 * of lines in little endian platforms.  This caused at least some of the
 * artifacts seen in PerfectScan around images.
 *
 * Revision 1.19  1994/10/11  16:37:40  lperry
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Added include of shros.pub to get IP_YIELD def
 *
 * Revision 1.18  1994/09/17  18:01:45  ddavies
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Change i_rescaleArb so it creates a table for i_depth4To8bpp and passes
 * this table in.  i_depth4to8bpp is called to convert strips of image and
 * creating and destroying the table once is better than creating and
 * destroying it on every strip as we were forced to do with the old
 * interface.
 *
 * Revision 1.17  1994/07/12  23:55:49  ejaquith
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * removed #include "iadef.prv".
 *
 * Revision 1.16  1994/07/11  23:11:59  ddavies
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Free the array xShift in i_rescaleNoInterp.  This fixes a memory leak that
 * occured when rescaling 4 bit/pixel images.
 *
 * Revision 1.15  1994/06/29  16:42:41  macleod
 * macleod on Wed Jun 29 09:42:24 PDT 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Added some IP_YIELD calls
 *
 * Revision 1.14  1994/05/19  23:50:40  macleod
 * macleod on Thu May 19 16:49:42 PDT 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * i_rescaleNoInterp failed on odd-width 4-bit images because the
 * final byte loop did two pixels at a time, but didn't check for only one
 * pixel left.
 *
 * Revision 1.13  1994/05/19  23:25:59  macleod
 * macleod on Thu May 19 16:25:20 PDT 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Modified i_rescaleNoInterp() to work on 4 bit/pixel images
 *
 * Revision 1.12  1994/05/11  01:05:01  dferrell
 * dferrell on Tue May 10 17:57:29 PDT 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Didn't quite kill it the first time...now I think it's fixed.
 * Basically, bpl calculations for the 8-bit buffers created when
 * scaling 4-bit PIXRs were incorrect.
 * Removed rounding in non-interpolated rescaling which was causing
 * the result to use one column and on row of the source's frame when
 * computing the bottom row and right column pixels.  This was really
 * Peter MacLeod's fix.
 *
 * Revision 1.11  1994/05/10  01:15:54  dferrell
 * dferrell on Mon May 9 18:14:26 PDT 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Fixed bug in i_rescaleArb for 4-bit images.  The call to i_depth4To8bpp
 * did not account for padding in odd sized images.
 *
 * Revision 1.10  1994/03/09  17:51:22  macleod
 * macleod on Wed Mar 9 09:50:32 PST 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Fixed two bugs in 4-bit rescale.
 * The first caused it to stomp on memory past a temporary buffer.
 * The second caused FREE to be called on something not MALLOCed.
 *
 * Revision 1.9  1994/03/01  02:49:03  dferrell
 * dferrell on Mon Feb 28 18:47:30 PST 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Added typecaste to _signed_int_ in expression with shift and evaluation
 * "> 0" which caused an incorrect branch opcode to be generated by MPW C.
 * Arbitrary rescale now works on the Mac.
 *
 * Revision 1.8  1994/02/09  22:32:09  macleod
 * macleod on Wed Feb 9 14:31:59 PST 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Chnaged i_rescaleArb to support 4-bit-per-pixel rescaling.
 *
 * Revision 1.7  1994/02/04  20:16:31  dferrell
 * dferrell on Fri Feb 4 12:16:22 PST 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.6  1994/01/26  04:14:12  jcraig
 * jcraig on Tue Jan 25 20:11:52 PST 1994
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * check to make sure that depth is 8 in i_rescaleArb,
 * i_rescaleAreaMap and i_rescaleNoInterp.  if not, return
 * ia_depthNotSupported.
 *
 * Revision 1.5  1993/12/02  19:23:38  ddavies
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.4  1993/11/18  00:29:48  jcraig
 * jcraig on Wed Nov 17 16:28:49 PST 1993
 *
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * keep from writing into the bottom frame (since sometimes we write
 * 4 destination lines at once) in i_rescaleArb
 *
 * Revision 1.3  1993/11/16  02:57:27  ddavies
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Added a call to ip_clearRightFrame in i_rescaleAreaMap.  It looked like
 * the author had intended to put one here and it was writing into the right
 * frame.  Don't you hate it when you can't back up a line?  I meant that
 * i_rescaleAreaMap was writing into the right frame, so needed to
 * have ip_clearRightFrame called.
 *
 * Revision 1.2  1993/11/05  01:28:48  ddavies
 * i_rescal.c -> /product/ipcore/ipshared/src/RCS/i_rescal.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


