/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_rescon.c
 *
 * Public resolution conversion routines:
 *		Int32	i_rescon2To3()
 *		Int32	i_rescon1To2()
 *		Int32   i_rescon3To1()
 *		Int32	i_rescon2To1()
 *
 */


#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "shrscal.pub"
#include "shrrast.pub"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrscal.prv"
#include "memory.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_rescon.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* The EZScan folks would like to store gray images at 1/1.5x or 1/2x
 * the resolution of binary images.  For example, they'd like to store
 * 200 spi gray images and 300 spi binary images; or 150 spi gray and
 * 300 spi binary.  In fact, it might be nice to store 100 spi gray and
 * 300 spi binary, then use 1.5x followed by 2x expansion to make the
 * resolutions the same.
 *
 * We already have lots of scaling and expansion routines.  The fast
 * ones just copy pixels.  This is fast, but the results aren't pretty.
 * The completely general routines that look good only work on 8 bit/pixel
 * images and are pretty slow.  The routines implemented here try to do
 * relatively fast 1.5x and 2x expansion with smoothed pixel values.
 *
 * We could write four fairly complicated routines, one each for 4 bpp/1.5x,
 * 4 bpp/2x, 8bpp/1.5x and 8bpp/2x.  These routines would produce a block of
 * output lines in each pass through the image.
 *
 * The EZScan folks will use these routines on bands of pixels.  Each
 * band will then be given to Windows, which will do all sorts of unspeakable
 * acts on the way to the screen.  The point is that our code will be
 * swapped into and out of the cache (probably) for each band.  This makes
 * it reasonably important that the code be small and fast.  Therefore,
 * we will try to avoid the complicated methods.
 *
 * We will attack the problem by implementing several lower level procedures
 * and have higher level code string them together:
 *
 *   Procedures the produce a single line of output:
 *	ip_rescon1_5x4bppLine
 *	ip_rescon2x4bppLine
 *	ip_rescon1_5x8bppLine	(if scaleLinear isn't fast enough)
 *	ip_rescon2x8bppLine	(if scaleLinear isn't fast enough)
 *
 *   Procedures that average two output lines:
 *	ip_average4bppLines
 *	ip_average8bppLines	(if scaleLinear isn't fast enough)
 *
 * The first set of procedures expand each line of input pixels into a
 * line of output pixels with the proper width and pixel values.  The
 * second set average two existing output lines to produce a third line.
 *
 * For 2.0x expansion, weights look like this:
 *
 *	A    B	  C		A	(A+B)/2		B	C
 *				(A+D)/2	(A+B+D+E)/4	(B+E)/2	(C+F)/2
 *	D    E	  F	=>	D	(D+E)/2		E	F
 *				
 *	G    H    I		G	(G+H)/2		H	I
 *
 * How do we use the routines?  As we see, every two pixels horizontally
 * produces three pixels horizontally.  The first and third pixels have
 * the same gray values as the original two pixels.  The pixel in the
 * middle has the average value of the two originals.
 *
 * Similarly, at 1.5x expansion, every two input rows produces three
 * output rows with the same {1.0, 0.5+0.5, 1.0} weights.  We use the
 * ip_rescon1_5x procedures to produces the output lines with 1.0 weights
 * directly.  We use the ip_averageXbppLines procedures to produce the
 * intermediate lines.
 *
 * For 2.0x expansion, weights look like this:
 *
 *	A    B	  C		A	(A+B)/2		B	(B+C)/2
 *				(A+D)/2	(A+B+D+E)/4	(B+E)/2	(B+C+E+F)/4
 *	D    E	  F	=>	D	(D+E)/2		E	(E+F)/2
 *				(D+G)/2 (D+E+G+H)/4	(E+H)/2	(E+F+H+I)/4
 *	G    H    I		G	(G+H)/2		H	(H+I)/2
 *
 * Again, we use the ip_rescon2x procedures to produce every other line
 * directly and the same averageXbppLine procedures to produce the
 * intermediate lines.  Each time we produce a new direct line, we produce
 * the intermediate line above it as well.
 *
 * We hope that by using simple procedures with a relatively small number
 * of variables, we will allow the C compilers to keep the important
 * variables in registers.  This is especially hard on the PC.
 *
 * Interface change: The EasyScan folks would like to convert 200x200x4
 * images to 300x300x1 without going to 300x300x4 on the way.  That uses
 * too much memory.  So, we'll have an ipwrap routine that does the
 * operation in strips - converting a strip of 200x200x4 into a strip
 * of 300x300x4, then thresholding the 300x300x4 strip into a 300x300x1
 * strip.  To make this vaguely efficient, we add a table pointer arg
 * to the resolution conversion routines so that they can take a table
 * pointer.  If the pointer is NULL, the table is allocated and freed
 * inside the i_resconXX routine.
 *
 * Slight bug fix: We were averaging the frame on the right side of the
 * image into the pixels on the right side of the output.  This isn't really
 * very nice.  Instead, we should replicate the right column of pixels
 * as necessary to fill the gap.  To make this slightly more efficient,
 * the lower level routines would like to know that they will never
 * encounter images of width 0.  We check for those at the top level now.
 */

/* PUBLIC */
/**************************************************************************
 * i_rescon2To3()
 * 	Convert a 200x200 image into a 300x300 image.
 *	Currently works only with 4 bit and 8 bit images.
 *
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		sDepth = depth of source image
 *		pD = destination gray image
 *		wD = width in pixels (including frame)
 *		hD = height including frame
 *		dBpl = source bytes/line
 *		dDepth = depth of destination image
 *		pResconTable = ptr to table used during conversion.
 *				if NULL, table will be allocated.
 *				table can be obtained by calling
 *				i_make2To3ResconTable.
 *
 *	Current implementation assumes 4 bits/pixel.
 *	wD should equal wS * 3 /2
 *	hD should equal hD * 3 / 2
 **************************************************************************/
Int32   CDECL 
i_rescon2To3 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	sDepth,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    Int32	dDepth,
    UInt32     *pResconTable)
{
    PROLOGUE;
    Int32	 sLine;			/* interation variables */
    
    Bool	 tableShouldBeFreed;
    UInt8	*pDOrig;

	/* get the width and height without the frame*/
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    wD -= 2*cFrameBits;
    hD -= 2*cFrameBits;

    if (((sDepth != 4) && (sDepth != 8)) || (sDepth != dDepth))
	return ia_depthNotSupported;

    if ((wS == 0) || (hS == 0))
	NORMAL_RETURN;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + dDepth*cFrameBytes;	
    pDOrig = pD;

	/* allocate a table if necessary */
    if (pResconTable == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make2To3ResconTable (pIPState, &pResconTable, dDepth))
		!= ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* expand each pair of lines into three lines.  We assume that
	 * the height of the dest image is (3 * hS)/2 and this has been
	 * truncated.  Therefore, the dest image may end after either
	 * the first of the pair of source lines has been written
	 * (hS is odd) or after the second of the pair of source lines
	 * has been written and the middle line has been interpolated
	 * (hS is even).
	 */
    for (sLine = 0; 
	 sLine < hS; 
	 sLine++, pS += sBpl, pD += dBpl)
    {
	if (dDepth == 4)
	    ip_rescon1_5x4bppLine (pIPState, pS, wS, pD, pResconTable);
	else
	{
	    ip_rescon1_5x8bppLine (pIPState, pS, wS, pD);
	}

	sLine++;	/* advance down one source line */
	if (sLine < hS)
	{
		/* go down 1 source line and *2* dest lines for
		 * the next direct output line */
	    pD += 2*dBpl;
	    pS += sBpl;
	    if (dDepth == 4)
	    {
		ip_rescon1_5x4bppLine (pIPState, pS, wS, pD, pResconTable);
		ip_average4bppLines (pIPState, pD - 2*dBpl, dBpl, wD);
	    }
	    else	/* depth == 8 */
	    {
		ip_rescon1_5x8bppLine (pIPState, pS, wS, pD);
		ip_average8bppLines (pIPState, pD - 2*dBpl, dBpl, wD);
	    }
	}
    }

	/* clean up the right frame */
    ip_clearRightFrame (pIPState, pDOrig, wD, hD, dBpl, dDepth);

    if ((tableShouldBeFreed) && (pResconTable != NULL))
	FREE(pResconTable);

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 * i_rescon1To2()
 * 	Expand an image by 2x, smoothing the pixels.
 *	Currently works only with 4 bit and 8 bit images.
 *
 *	Input:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		sDepth = depth of source image
 *		pD = destination gray image
 *		wD = width in pixels (including frame)
 *		hD = height including frame
 *		dBpl = source bytes/line
 *		dDepth = depth of destination image
 *		pResconTable = ptr to table used during conversion.
 *				if NULL, table will be allocated.
 *				table can be obtained by calling
 *				i_make1To2ResconTable.
 *
 *	Current implementation assumes 4 bits/pixel.
 *	wD should equal wS * 2
 *	hD should equal hD * 2
 **************************************************************************/
Int32   CDECL 
i_rescon1To2 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	sDepth,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    Int32	dDepth,
    UInt32     *pResconTable)
{
    PROLOGUE;
    Int32	 sLine;			/* interation variables */
    
    Bool	 tableShouldBeFreed;
    UInt8	*pDOrig;

	/* get the width and height without the frame*/
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    wD -= 2*cFrameBits;
    hD -= 2*cFrameBits;

    if (((sDepth != 4) && (sDepth != 8)) || (sDepth != dDepth))
	return ia_depthNotSupported;

    if ((wS == 0) || (hS == 0))
	NORMAL_RETURN;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + dDepth*cFrameBytes;	
    pDOrig = pD;

	/* allocate a table if necessary */
    if (pResconTable == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make1To2ResconTable (pIPState, &pResconTable, dDepth))
		!= ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* expand each source lines into two lines.  The very last line
	 * of the dest image is just copied from the directly computed
	 * line above it.
	 * Do the first line separately since we won't be computing
	 * an averaged line above it.
	 */
    if (dDepth == 4)
	ip_rescon2x4bppLine (pIPState, pS, wS, pD, pResconTable);
    else
    {
	ip_rescon2x8bppLine (pIPState, pS, wS, pD);
    }

	/* process the second through last lines of the images */
    for (sLine = 1, pS += sBpl, pD += 2*dBpl; 
	 sLine < hS; 
	 sLine++, pS += sBpl, pD += 2*dBpl)
    {
	if (dDepth == 4)
	{
	    ip_rescon2x4bppLine (pIPState, pS, wS, pD, pResconTable);
	    ip_average4bppLines (pIPState, pD - 2*dBpl, dBpl, wD);
	}
	else	/* depth == 8 */
	{
	    ip_rescon2x8bppLine (pIPState, pS, wS, pD);
	    ip_average8bppLines (pIPState, pD - 2*dBpl, dBpl, wD);
	}

	if (sLine == hS - 1) 	/* there's no source line below this one
				 * so copy the last valid dest line down one */
	{
	    memcpy((void *)(pD+dBpl), (const void *) pD, dBpl);
	}
    }

	/* clean up the right frame */
    ip_clearRightFrame (pIPState, pDOrig, wD, hD, dBpl, dDepth);

    if ((tableShouldBeFreed) && (pResconTable != NULL))
	FREE(pResconTable);

    NORMAL_RETURN;
}

/* PUBLIC */
/****************************************************************************
 * i_rescon2To1
 *
 *	This is going to be used, among other things, by the "image spigot"
 *	code, which is taking a 200 or 300dpi stream and downconverting on the
 *	fly. 
 *
 *	The conversion is going to be a straightforward "box" filter,
 *	i.e. add up four input pixels to get one:
 *
 *	1 2 -> 1 
 *	3 4
 *
 *	One question is what to do when not given an even number
 *	of lines. In this case, we're going to leave it to higher-level
 *	routines--we always expect an even number of lines, and an
 *	even number of pixels on a line. If this isn't true, it'll
 *	just go into the frame to get the edge pixels, and thus make
 *	the edges darker, so the caller is responsible for replicating
 *	the last column and row if needed.
 *
 * Arguments:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		sDepth = depth of source image
 *		pD = destination gray image
 *		wD = width in pixels (including frame)
 *		hD = height including frame
 *		dBpl = source bytes/line
 *		dDepth = depth of destination image
 *
 *	Current implementation assumes 4 or 8 bits/pixel.
 *	wD should equal wS / 2
 *	hD should equal hD / 2
 * 
 * Returns:
 *     error status
 ****************************************************************************/

Int32 CDECL 
i_rescon2To1 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	sDepth,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    Int32	dDepth)
{
    PROLOGUE;
    Int32   line;
    UInt32  srcDelta;

	/* get the width and height without the frame*/
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    wD -= 2*cFrameBits;
    hD -= 2*cFrameBits;

    if (((sDepth != 4) && (sDepth != 8)) || (sDepth != dDepth))
	return ia_depthNotSupported;

    if ((wS == 0) || (hS == 0))
	NORMAL_RETURN;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + dDepth*cFrameBytes;	
    srcDelta = sBpl * 2;

    /* Crank through the output lines */
    for (line = 0; line < hD; line++)
    {
	if (sDepth == 4)
	    ip_rescon2To1Line4bpp (pIPState, pS, wS, sBpl, pD, wD);
	else
	    ip_rescon2To1Line8bpp (pIPState, pS, wS, sBpl, pD, wD);

	pS += srcDelta;
	pD += dBpl;
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/****************************************************************************
 * i_rescon3To1
 *
 *	This is going to be used, among other things, by the "image spigot"
 *	code, which is taking a 300dpi stream and downconverting on the
 *	fly. 
 *
 *	The conversion is going to be a straightforward "box" filter,
 *	i.e. add up nine input pixels to get one:
 *
 *	1 2 3 
 *	4 5 6 -> 1
 *	7 8 9
 *
 *	One question is what to do when not given a multiple of three
 *	input lines. In this case, we're going to leave it to higher-level
 *	routines--we always expect an even number of lines, and an
 *	even number of pixels on a line. If this isn't true, it'll
 *	just go into the frame to get the edge pixels, and thus make
 *	the edges darker, so the caller is responsible for replicating
 *	the last columns and rows if needed.
 *
 * Arguments:
 *		pS = source gray image (including frame)
 *		wS = width in pixels (including frame)
 *		hS = height including frame
 *		sBpl = source bytes/line
 *		sDepth = depth of source image
 *		pD = destination gray image
 *		wD = width in pixels (including frame)
 *		hD = height including frame
 *		dBpl = source bytes/line
 *		dDepth = depth of destination image
 *
 *	Current implementation assumes 4 or 8 bits/pixel.
 *	wD should equal wS / 3
 *	hD should equal hD / 3
 * 
 * Returns:
 *     error status
 ****************************************************************************/

Int32 CDECL 
i_rescon3To1 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    Int32	sDepth,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    Int32	dDepth)
{
    PROLOGUE;
    Int32   line;
    Int32  srcDelta;

	/* get the width and height without the frame*/
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    wD -= 2*cFrameBits;
    hD -= 2*cFrameBits;

    if (((sDepth != 4) && (sDepth != 8)) || (sDepth != dDepth))
	return ia_depthNotSupported;

    if ((wS == 0) || (hS == 0))
	NORMAL_RETURN;

	/* point to the upper left corner of the source and dest image. */
    pS += (sBpl << cLogFrameBits) + sDepth*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + dDepth*cFrameBytes;	
    srcDelta = sBpl * 3;

    /* Crank through the output lines */
    for (line = 0; line < hD; line++)
    {
	if (sDepth == 4)
	    ip_rescon3To1Line4bpp (pIPState, pS, wS, sBpl, pD, wD);
	else
	    ip_rescon3To1Line8bpp (pIPState, pS, wS, sBpl, pD, wD);

	pS += srcDelta;
	pD += dBpl;
    }

    NORMAL_RETURN;
}


/*

$Log: i_rescon.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.11  1998/01/13 23:49:27  ddavies
i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.10  1995/12/13 02:45:35  ejaquith
i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.9  1995/10/18  03:58:26  ddavies
 * i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.8  1995/03/08  04:26:13  ddavies
 * ddavies on Tue Mar 7 20:24:42 PST 1995
 *
 * i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v
 *
 * Add checks to see if source image has width or height of 0.  This means
 * that the lower level routines don't have to check on every line.
 *
 * Revision 1.7  1994/09/17  18:03:47  ddavies
 * i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v
 *
 * Change i_rescon1To2 and i_rescon2to3 so they can either use tables passed
 * in by clients or can create and destroy their own tables.  This makes using
 * the to process images in strips more efficient.  The table making procedures
 * are the newly-public i_make1To2ResonTable and i_make2To3RescontTable instead
 * of the old, private ip_ versions.
 *
 * Revision 1.6  1994/05/26  00:21:01  macleod
 * macleod on Wed May 25 17:16:48 PDT 1994
 *
 * i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v
 *
 * added i_rescon3To1 and i_rescon2To1, for downconverting gray images.
 * i_rescon2To1 uses a box filter, i.e. adds up four input pixels for each
 * output pixel. i_rescon3To1 uses a center-weighted filter, i.e.
 * 1/16 1/8 1/16
 * 1/8  1/4 1/8
 * 1/16 1/8 1/16
 * Which has the two advantages over a box filter that it doesn't lose quite
 * as much contrast, and it can be done with only shifts. While it's not
 * the ideal filter by any means, it does produce a decent image and is
 * reasonably quick (.82 s on a 8.5x11" 300dpi image on a sparc 2).
 *
 * Revision 1.5  1994/02/04  20:16:36  dferrell
 * i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:23:42  ddavies
 * i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:28:57  ddavies
 * i_rescon.c -> /product/ipcore/ipshared/src/RCS/i_rescon.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


