/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_rnhst1.c
 *
 *	A "25 cent" frequency transform: make a "run-length histogram"
 *	for a bitmap. Unlike runHistogram, these routines have
 *	lots of bells and whistles.
 */


#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "memory.pub"
#include "lmask.prv"
#include "rmask.prv"

/* Prototypes */
#include "shrcc.pub"
#include "shrcc.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_rnhst1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* The code that finds blocks of glyphs on a page uses histograms of run
 * lengths.  XIS uses them extensively for segmentation.  In one sense,
 * they are a cheap alternative to computing fourier transforms.
 *
 * The module i_runhst.c also contains code to compute histograms of
 * horizontal runs.  If this code works better, the code in i_runhst.c
 * will become obsolete.
 *
 * XIS asked for a routine that has a bunch of arguments.  We've added
 * a couple more.  The arguments are:
 *
 *	pointers to image and rectangle within the image;
 *	pHisto: pointer to histogram to be returned;
 *	numHistoBins: equals the number of bins in the histogram
 *	direction: samples are to be taken horizontally or vertically;
 *	sampleFreq: skip sampleFreq-1 lines between each sampled row
 *		or column;
 *	runType: whether to count black runs, white runs or cycles (lengths
 *		of black/white pairs);
 *	useEndRuns: all runs are terminated at the boundary of the
 *		sampling rectangle.  If useEndRuns is true, we should
 *		count the runs that touch the boundary.  If it is
 *		false, we should ignore these runs.  This is a pain,
 *		but is useful.
 *		Well, in fact, NOT ignoring cycles that abut the
 *		borders is not a well defined operation.  Besides,
 *		it's hard to implement.  So we'll say that if we are
 *		measuring cycles, we always ignore cycles that abut
 *		the edges of the rectangle.
 *	countLongRuns:	if false, we ignore runs that have no bins in
 *		the histogram.  If true, all the counts for the large
 *		runs are put into to the last histogram bin.
 *	weightByLength: if false, finding a run of length x causes us to
 *		add 1 to bin x.  If true, finding a run of length x causes
 *		us to add x to bin x.  What is this for anyway?
 *	
 * This is quite a monster - the swiss army knife approach to writing code.
 * Unfortunately, I think we actually use this stuff.
 *
 * We'd really like to avoid making a zillion decisions for every run found.
 * One way to avoid some decisions is to create a temporary histogram,
 * then process it to produce the histogram we really want.  We will create a
 * temporary histogram containing a bin for every possible run length.
 * If we're looking for horizontal runs, the temporary histogram will have
 * bins for 1 through rectangleWidth.  If we're tracing vertical runs,
 * we need bins for 1 through rectangleHeight.  When we've filled the
 * temporary histogram with run info from the image, we start post processing.
 *
 * If weightByLength is true, we multiply the number in each bin by the
 * bin number.
 *
 * We then add the contents of the first numHistoBins bins from the
 * temporary histogram to the contents of the output histogram.
 *
 * If countLargeRuns is true, we add the contents of the remaining bins
 * in the temporary histogram to the pHisto[numHistoBins-1].
 *
 * That takes care of several of the parameters.  There are several
 * significant problems remaining.  The first one is the horizontal/vertical
 * direction.  Calculating histograms of horizontal runs is certainly
 * easier since you get to see several bits in a run in each sample.
 * Most of the overhead is encountered in detecting that a run has ended
 * and taking appropriate action.  In the vertical direction, you have
 * a choice of maintaining several runs concurrently or processing each
 * pixel in the image individually - yuck.  Let's describe the horizontal
 * run code first, getting back to the vertical code later.
 *
 * Even with just horizontal histograms, the choice of counting runs of
 * 1's, runs of 0's or cycles is not one we'd like to make on every
 * image sample.  We'll have different code for each of these cases.
 * What are cycles?  When finding areas containing periodic data, like
 * glyphs or halftones, it's best to count not just the width of each
 * mark or the width of the space between them, but the period of
 * white/black pairs.  This is the thing that's most likely to be constant.
 * Another way to think of this is that we count the distance between
 * 0-to-1 transitions.
 *
 * The really messy remaining parameter is useEndRuns.  It is much more
 * convenient to count all the runs.  When useEndRuns is false, we must
 * efficiently trace the leftmost and rightmost runs on each scan line,
 * then ignore them.  If we were tracing each run serially, this would
 * be trivial.  Since we're trying to process as much data as we can
 * on each sample, this will be messy.  We will have separate verions of
 * run counting code for these too.
 *
 * The routine at this level, i_runHistogram, will call lower routines
 * to process each line.  These routines will have the real inner loops.
 * What do the inner loops of these routines look like?  The inner loops
 * are similar for all the routines.  The cycle routine looks for somewhat
 * different things than the routines that look for 1's or 0's, but
 * the principle is the same.  The SkipEnds and UseEnds routines differ
 * only in how they start and stop.
 *
 * We clearly want to look at more than one pixel at a time.  The inner
 * loops should be able to skip over runs of 0's and runs of 1's in large
 * jumps.  This is especially true since these are fairly common.  On the
 * other hand, a large chunk of bits might contain many internal runs,
 * so we can't just deal with large chunks all the time.
 *
 * One cute observation is that the inner loops have two basic jobs.  One
 * is to track runs that bridge chunks.  The other is to track runs that
 * start and stop inside individual chunks.  Tracking runs that start and
 * stop purely within chunks need not be done as each chunk is encountered.
 * We can create yet another histogram, this one of chunk values seen.
 * When we get done with the image, we calculate the number of runs that
 * started and stopped inside each chunk and add that into the temporary
 * histogram.  For example, say chunks are bytes and we find an 0xAA.  We
 * increment the 0xAA bin.  At the end of the image, say we've found
 * 70,000 0xAA's.  Say we're tracking runs of 1's.  This byte (1010 1010)
 * has three internal runs (plus one on the left that bridges to the next
 * byte, each of length 1.  We would add 3 * 70,000 to the
 * temporary histogram bin for runs of length 1.
 *
 * Using this trick, we only have to track runs between chunks when
 * we're actually going through the image - the runs inside chunks are
 * done later.  This greatly simplifies the code done for each byte.
 *
 * We use tables to compute things about chunks.  To keep table sizes
 * small, we'd like the chunk size to be fairly small.
 * To be able to skip over runs of black or white efficiently, we'd like
 * the chunk size to be large.  We'll use a combination.  We process the
 * image in units of 32-bit words.  If a word is entirely black or entirely
 * white, we skip over it easily.  If it contains both black and white,
 * we process its four bytes individually.  At the end of the four bytes,
 * we look at the next word.
 *
 * There are two ways we can do the calculations on each byte: we can use
 * a state machine that uses each byte as input or we can derive both
 * the state and output from the region around each byte.  Huh?  The
 * state machine approach is straightforward.  Say we're looking for runs
 * of 1's.  The current state is set by the previous byte - we're either
 * in a run of 0's or a run of 1's:
 *	State	byte starts with	Action
 *	0's	0			keep looking for 1's
 *	0's	1			start new run
 *	1's	0			end current run, start looking for
 *					another
 *	1's	1			continue current run.
 *
 * This is lovely, but state machines are sort of complicated what with
 * all the testing a branching that goes on.
 *
 * Note that the state from the previous byte is equal to the last bit of
 * that byte together with the length of the run we're now in (if any).
 * We could eliminate the state machine stuff by just concatenating that
 * bit with the others in the byte when doing our table lookups.
 *
 * What table lookups?  We use three tables for each byte:
 *   1.	Incremental run length
 *	This table performs two functions.  It indicates when a run from
 *	the previous byte ends and gives the distance by which that run
 *	extends into this byte.  The run ends when the value in this
 *	table is negative.  The trick is that we need two values of 0.
 *	When a run ends at the beginning of this byte, we should store
 *	negative 0.  When there is no run coming into this byte, we store
 *	positive 0.  Since we need both positive and negative values for
 *	0, we use one's complement to store negative values.
 *   2.	Byte Value
 *	This is a histogram of byte values.  When we come to the end of
 *	a run, we increment the bin corresponding to this byte.  We use
 *	these counts later to add in the runs that were purely internal
 *	to the bytes.
 *   3.	New Run Length
 *	This is the number of bits in the run that includes the last
 *	bit of the byte.  We use it to set the run Length counter before
 *	encountering the next byte.
 * 
 * One nice feature of organizing things this way is that the inner loops
 * of the code that finds cycles and the code that finds runs of 1's (or 0's)
 * is all the same.  The only difference is the contents of the tables.
 * But, we use different routines anyway.  If we use different routines, we
 * can address the tables directly in each routine.  If we used one
 * routine, we'd have to pass a pointer to the table, which would use
 * a register.  There aren't a lot of registers in a x86 processor, so
 * we'll use different routines.
 *
 * Do we scan left to right or right to left within a line?  When we
 * encounter a new word, we would like to perform a quick check to see
 * if its the entension to an existing run or contains the end of that
 * run.  We can do this by looking at the last bit of the previous word
 * and the value of the new word:
 *	last bit of prev word	new word	Action
 *		0		00...00		Add 32 to runLength unless
 *						we're looking for runs of 1's
 *		1		11...11		Add 32 to runLength unless
 *						we're looking for runs of 0's
 *		all other combinations		Find end of run in word
 * It's easy to branch on the last bit of the previous word if that bit
 * is in the sign bit position (MSBit).  Therefore, we will go from right
 * to left within a scan line.
 *
 * Do we read each byte individually or do we read entire words into a
 * register and do shifting and masking to form the table lookup values?
 * Reading bytes individually is generally a good idea, but maybe not in
 * this case.  First, there is the annoyance of dealing with different
 * byte orders.  If we read by words, this goes away.  That's not a
 * big deal.  Second, at least the byte value histogram entries are not
 * bytes.  Therefore, the machine will do some shifting anyway.  The
 * current run length will not be a byte, so the value read from the
 * Incremental run length table will either be expanded into a word
 * (an operation that our compilers do very stupidly on PCs) or the
 * incremental run length table entries will be words already.  Ditto for
 * the new run length table.  We can save some shifting of address
 * offsets by making all the tables have the same entry size, so pick
 * 32-bit words.  This also eliminates the operations used to expand
 * smaller entries to words when they're read.
 *
 * Next, we're not addressing at least the incremental run length table
 * with bytes.  We use 9 bits - 8 from the current byte and 1 from the
 * previous byte.  We'd have to compose these entries anyway, so we might
 * as well do this from a register that hold all the bytes at once.  So
 * long as one of the tables is addressed using 9 bits, we might as well
 * address all of them that way - again to avoid extra address computation.
 * Reading the image in units of words and processing it in units of runs
 * of bits from these words has another advantage when processing the
 * end runs as seen below.
 *
 * What do we do with the end runs anyway?  When we are keeping the end
 * runs and looking for runs of 1's (or 0's), we just smash the portions of the
 * works on the far left and right that are outside our rectangle with
 * 0's (or 1's).  When looking for cycles, we smash the portion of the
 * word on the right with 0's (since that cannot start a new run) and
 * the portion of the leftmost word outside the rectangle with 1's
 * (since that cannot end a run).  When ignoring the end runs, we'll
 * duplicate the inner loop.  The first copy is executed only for the
 * very first, rightmost, run.  It does the normal things, except that
 * it doesn't update the temporary histogram for the first run length.
 * It then falls through to the ordinary loop.
 *
 * To ignore the leftmost run, we simply make the inner loop stop before
 * processing the word containing the left end of the rectangle.  The
 * portion of the this word that's outside the rectangle is filled with 1's
 * (0's if we're looking for runs of 0's).  When we get to the end of this
 * word, we simply neglect to update the temporary histogram for the last run.
 *
 * We can combine the routine that ignores ends with the routine that
 * pays attention to ends by choosing the appropriate starting and ending
 * code.  Since this branch only happens twice per line, the performance
 * penalty for this branch should be negligible.
 *
 * How do we actually compute the internal runs inside each byte?  The
 * first thing to notice is that there are only 512 different values.
 * No matter what we do, it won't take very long.  We just take each byte
 * and look up the length of the run at the right end.  We shift right
 * by that amount plus 1, since we were supposed to ignore the run's bit
 * in the word plus the extra bit that belonged to the previous byte.
 * We also calculate the number of bits in the byte that are left over
 * when we subtract off the bits belonging to the runs on the left and
 * right ends.  In a loop, we then look for the length of the run on
 * the right end of the byte using the regular tables.  If the length
 * is negative, the run terminates in the byte.  Increment the temp
 * histogram bin corresponding to (runLength + 1) (to account for the
 * bit in the LS bit of the index - that bit is part of our byte but is
 * not counted by the table) by the number in the byte's entry in the
 * byte histogram.  Let me try that again.  The number in the byte's bin
 * in the byte histogram corresponds to the number of times this byte
 * was seen in the image.  Therefore, each run we find inside this byte
 * should be weighted by this occurance count before we put it into
 * the final histogram of run lengths.
 * Anyway, we shift the remaining byte bits right by runLength+1 and
 * subtract runLength+1 from the number of bits remaining in the byte
 * and go to the top of the loop.  The loop terminates when the number
 * of bits remaining in the byte is 0.
 *
 * The tables are in the include file runhist.prv.  To utilize processor
 * caches efficiently, we have folded the four tables described above into
 * two.  Almost every time we get the distance a run extends into the byte
 * from the right, we also want the distance the next run extends into
 * the same byte from the left.  So, we interleave the two values so that
 * when one appears in the caches, the other is probably also brought in.
 * The table onesOnEnds holds the number of 1's on the left and right ends
 * of each byte.  The table cyclesOnEnds holds the lengths of the cycles
 * on each end of a byte.
 *
 * Yeah, but what about vertical runs?  Well, the routine that looks at
 * vertical runs will be entirely different than the one that looks at
 * horizontal runs, so we'll just make it a different routine altogether.
 * Here's the horizontal routine.
 ****************************************************************************
 *
 * SPACE SAVING, 1/4/95.
 * These routines worked fine, but were too big for XIS's Reading Edge
 * product.  In an effort to save space, we have taken the following
 * actions:
 *   1.	We have combined the routines that detect runs of 1's and runs of
 *	0's into a single routine.  This routine, ip_horizRunHistoLine,
 *	takes an argument that's used to conditionally invert every image
 *	word read.  We do the invert when looking for runs of 0's, so we're
 *	always looking for runs of 1's.  This makes a little more
 *	intuitive sense when debugging (I think).
 *   2.	Since we really only look for runs of 1's now, the table used to
 *	find runs of 0's is eliminated.
 *   3.	The run length table entries have been changed from 32-bit words
 *	to 8-bit bytes.  This saves quite a bit of space, so should improve
 *	cache performance.  However, we do insert an extra sign-extension
 *	when reading the run-length table (takes longer on x86 machines)
 *	and an extra shift when incrementing the histogram of byte values.
 */


/* PUBLIC */
/*****************************************************************************
 * i_horizRunHistogram
 *
 *	Make a histogram of the length of horizontal runs of 1's or lengths of
 *	pairs of concatenated horizontal runs of 0's followed by runs of 1's.
 *
 *	This routine computes a temporary histogram, then folds it into
 *	the client's desired histogray according to the parameters given
 *	below.
 *
 *	The client's histograms should be zeroed by the client before entry.
 *	This routine just adds to the existing histogram.
 *
 * Arguments:
 *	pImage =	Pointer to the image in question, including frame
 *	bpl =		Bytes per line of image, including frame
 *	x, y =		(x,y) offset into image, from outside of frame
 *	w, h =		width and height of rectangle in which histogram
 *			is computed
 *	pHistos =	Pointer to client's histogram.
 *	numHistoBins =	Number of Int32's in the array pointed to by
 *			pHistos.
 *	sampleFreq =	skip sampleFreq-1 lines between samples
 *	runType	=	If equal to cRunOfZeros, look for runs of 0's.
 *			If equal to cRunOfOnes, look at runs of 1's.  If
 *			equal to cRunOfCycles, look at adjacent runs of
 *			0's and 1's.
 *	useEndRuns =	if equal to cTrue, count the runs that touch the
 *			borders of the rectangle, otherwise ignore them.
 *			However, if we are measuring cycles, we ALWAYS
 *			ignore the cycles that abut the edges of the
 *			rectangle.  Not ignoring them is a poorly-defined
 *			operation.
 *	countLongRuns =	if cTrue, put the counts for all runs that have
 *			no bins in the client's histogram into the
 *			last bin.  If cFalse, ignore those runs
 *	weightByLength = if cTrue, weight the number in each bin by the
 *			bin number.  If there are 17 runs in bin 33, set
 *			the number in bin 33 to 33 * 17.  If cFalse, each
 *			run counts as 1.
 *
 * Returns:
 *	ia_successful		everything worked
 *	ia_nomem		not enough memory for tables
 *	ia_invalidParm		something wrong with parameters
 *
 *****************************************************************************/
Int32 CDECL 
i_horizRunHistogram (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32      *pHistos,
    Int32	numHistoBins,
    Int32	sampleFreq,
    Int32	runType,
    Bool	useEndRuns,
    Bool	countLongRuns,
    Bool	weightByLength)
{
    PROLOGUE;
    UInt32	*pTempHisto;	    /* temporary histogram used to
				     * compute new values */
    UInt32	*pByteHisto;	    /* keeps track of the number of times
				     * we've seen each byte that might
				     * contain an internal run */
    Int32	 leftWordMask;	    /* used to delete areas outside
				     * of rectangle in left word of scan
				     * line. */
    UInt32	 rightWordMask;	    /* used to delete areas outside
				     * of rectangle in right word of
				     * scan line. */
    Int32	 rightOffset;	    /* the byte offset of the right word of the
				     * rectangle.
				     */
    
	/* The following are used to count the runs in the byte histogram */
    Int32	 index;		    /* used for histograms */
    UInt32	 trialByte;	    /* used to hold values of bytes when looking
				     * through the byte histogram */
    Int32	 byteRunLength;	    /* used to calculate the length of runs
				     * inside bytes */
    Int32	 bitsRemaining;	    /* number of bits not yet explored
				     * in a byte */
    Int8	*pRunTable;	    /* points to tables used to calculate run
				     * lengths */
    UInt32	 runOfZeros;	    /* When calling ip_horizRunHistoLine,
				     * this equals 0x00000000 when looking for
				     * runs of 1's and 0xFFFFFFFF when looking
				     * for runs of 0's.
				     */

	/* set these so we'll know these are well defined no matter
	 * when we quit.
	 */
    pTempHisto = pByteHisto = NULL;

	/* if the client told us to count cycles and pay attention to
	 * the cycles that abut the borders of the rectangle,
	 * complain.  We don't know how to do that.
	 */
    if (useEndRuns && (runType == cRunOfCycles))
    {
	__error__ = ia_invalidParm;
	goto finHorizRunHisto;
    }

	/* allocate the temporary histogram.  We want one entry for every
	 * possible run length.
	 */
    pTempHisto = (UInt32 *) CALLOC(w+1, sizeof(UInt32));
    if (pTempHisto == NULL)
    {
	__error__ = ia_nomem;
	goto finHorizRunHisto;
    }

	/* allocate the byte histogram.  We want one entry for every
	 * possible byte value.  Since we really index this thing with
	 * 9 bit things and there are two entries per thing,
	 * we need 512*2 entries.
	 */
    pByteHisto = (UInt32 *) CALLOC(512*2, sizeof(UInt32));
    if (pByteHisto == NULL)
    {
	__error__ = ia_nomem;
	goto finHorizRunHisto;
    }

	/* make pImage point to the upper left word we'll sample */
    pImage += y * bpl + (x / 32) * 4;

	/* how far away is the right word of the rectangle relative to
	 * pImage?
	 */
    rightOffset = ((x + w - 1)/32 - x/32) * 4;

	/* convert the sample frequency to the number of bytes to skip
	 * between lines.
	 */
    bpl *= sampleFreq;

	/* compute masks used to delete left and right ends of the rectangle.
	 * There are 5 cases:
	 * 1. runs of 1's; use runs on the ends.
	 *	the areas outside the rectangles are both filled with 0's
	 *	so the runs are truncated at the boundaries.
	 * 2. runs of 1's; discard runs on the ends.
	 *	the areas outside the rectangles are both filled with 1's.
	 *	This allows us to track the end runs without worrying about
	 *	exactly where the boundary is.
	 * 3. run of 0's; use runs on ends.
	 *	the areas outside of the rectangle are both filled with 1's
	 *	so the runs are truncated at the boundaries.
	 * 4. runs of 0's; discard runs on the ends.
	 *	the areas outside the rectangles are both filled with 0's.
	 *	This allows us to track the end runs without worrying about
	 *	exactly where the boundary is.
	 * 5. runs of cycles; we always discard the cycles on the ends.
	 *	The area to the right of the rectangle should be filled
	 *	with 0's so it acts as an extension of the cycle on the
	 *	right.  The area to the left of the rectangle should
	 *	be filled with 1's so it acts as an extension of the
	 *	leftmost cycle.
	 * We pass both masks and useEndRuns down to the routine
	 * that processes lines.  runType and useEndRuns tell the routine
	 * what to do with the masks.
	 *
	 * So long as we're switching on runType, we might as well
	 * do all the primary data gathering here too.  The lower
	 * level routine called depends on the run type.
	 *
	 ****************************************************************
	 *
	 * SPACE SAVING, 1/4/95.  Since we've combined the runs of 1's
	 * and the runs of 0's cases, we calculate the same masks for
	 * both.  The only difference is the runsOfZeros parameter passed
	 * down to ip_horizRunHistoLine.
	 */
    switch (runType)
    {
    case cRunOfZeros:
    case cRunOfOnes:
	if (useEndRuns)	/* set areas outside of rect to 0 */
	{
	    leftWordMask = rightMask[31 - (x % 32)];
	    rightWordMask = leftMask[(x+w-1) % 32];
	}
	else	/* set areas outside of rect to 1 */
	{
	    leftWordMask = ~(rightMask[31 - (x % 32)]);
	    rightWordMask = ~(leftMask[(x+w-1) % 32]);
	}
	if (runType == cRunOfZeros)
	    runOfZeros = 0xFFFFFFFF;
	else
	    runOfZeros = 0x00000000;

	for (__error__ = ia_successful; h > 0; pImage += bpl, h -= sampleFreq)
	{
	    __error__ = ip_horizRunHistoLine (pIPState, pImage,
					    rightOffset,
					    leftWordMask,
					    rightWordMask,
					    useEndRuns,
					    pTempHisto,
					    pByteHisto,
					    runOfZeros);
	    if (__error__ != ia_successful)
		goto finHorizRunHisto;
	}
	break;

    case cRunOfCycles:
	    /* Set area to the right of the rect to 0 and the area to
	     * to the left of the rect to 1's.  This will extend the
	     * last runs to both word boundaries.
	     */
	leftWordMask = ~(rightMask[31 - (x % 32)]);
	rightWordMask = leftMask[(x+w-1) % 32];
	    /* go get the data from the image a line at a time */
	for (__error__ = ia_successful; h > 0; pImage += bpl, h -= sampleFreq)
	{
	    __error__ = ip_horizRunOfCyclesHistoLine (pIPState, pImage,
						    rightOffset,
						    leftWordMask,
						    rightWordMask,
						    pTempHisto,
						    pByteHisto);
	    if (__error__ != ia_successful)
		goto finHorizRunHisto;
	}
	break;

    default:
	__error__ = ia_invalidParm;
	goto finHorizRunHisto;
    }


	/* At this point, the combination of pTempHisto and pByteHisto
	 * have the relevant run info.  Compute the internal runs from
	 * pByteHisto and put them into pTempHisto.  First, get the
	 * pointer to the table we'll use to decipher the runs.
	 */
    __error__ = ip_getHorizRunTablePtr (pIPState, runType, &pRunTable);
    if (__error__ != ia_successful)
	goto finHorizRunHisto;

    for (index = 0; index < 512*2; index += 2)
    {
	if (pByteHisto[index] == 0)
	    continue;				/* no image bytes had
						 * this value */
	trialByte = index;			/* there are two entries
						 * in the table for each
						 * index val */
	byteRunLength = pRunTable[trialByte];
	if (byteRunLength < 0)			/* a run enters from the
						 * right and ends in
						 * this byte */
	    byteRunLength = ~byteRunLength;	/* get the real count */

	    /* find the number of bits in this byte that aren't part
	     * of the runs on either end.  The bits in the byte
	     * that aren't part of the runs on either end may
	     * contain a new run.
	     */
	bitsRemaining = 8 - pRunTable[trialByte+1] - byteRunLength;

	    /* go through the internal runs of this byte */
	while (bitsRemaining > 0)
	{
	    trialByte >>= byteRunLength + 1;	/* shift out the bits of the
						 * run plus the LS bit from
						 * the "previous" run. */
	    trialByte &= ~1;			/* create a new address mask */
	    byteRunLength = pRunTable[trialByte];
	    if (byteRunLength < 0)
	    {
		    /* We've found an internal run.  Record its length
		     * in the temporary histogram.  Note that the length
		     * of this run includes the bit that would normally
		     * be part of the "previous byte."  We shifted a
		     * bit from our current byte into that position.
		     * The table entries don't count that bit as part
		     * of the run in the current byte, so we have to
		     * add it in.
		     */
		byteRunLength = ~byteRunLength;
		pTempHisto[byteRunLength + 1] += pByteHisto[index];
	    }
	    bitsRemaining -= byteRunLength + 1;	/* get rid of this run */
	}
    }

	/* At this point, the temporary histogram contains all the runs
	 * in the image.  If we are supposed to weight them by length,
	 * do so.  To avoid extra handling, do this while putting them
	 * into the client's histogram.
	 */
    if (weightByLength)
    {
	for (index = 0; index < MIN(numHistoBins, w+1); index++)
	    pHistos[index] += pTempHisto[index] * index;

	if (countLongRuns)	/* put the rest of the entries in pTempHisto
				 * into the client's last bin */
	{
	    for (; index <= w; index++)
		pHistos[numHistoBins-1] += pTempHisto[index] * index;
	}
    }
    else	/* don't weight by each entry by length */
    {
	for (index = 0; index < MIN(numHistoBins, w+1); index++)
	    pHistos[index] += pTempHisto[index];

	if (countLongRuns)	/* put the rest of the entries in pTempHisto
				 * into the client's last bin */
	{
	    for (; index <= w; index++)
		pHistos[numHistoBins-1] += pTempHisto[index];
	}
    }

	/* that's it.  We just free the stuff we've allocated and return */
finHorizRunHisto:
    if (pByteHisto != NULL)
	FREE(pByteHisto);
    if (pTempHisto != NULL)
	FREE(pTempHisto);
    NORMAL_RETURN;
}

/*

$Log: i_rnhst1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.8  1998/01/13 23:49:28  ddavies
i_rnhst1.c -> /product/ipcore/ipshared/src/RCS/i_rnhst1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1997/04/24 23:29:10  ddavies
i_rnhst1.c -> /product/ipcore/ipshared/src/RCS/i_rnhst1.c,v

Change types of useEndRuns countLongRuns and weightByLength parameters
from Int32 to Bool.

Revision 1.6  1995/12/13 02:45:38  ejaquith
i_rnhst1.c -> /product/ipcore/ipshared/src/RCS/i_rnhst1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/01/11  22:28:59  dferrell
 * dferrell on Wed Jan 11 14:28:22 PST 1995
 *
 * i_rnhst1.c -> /product/ipcore/ipshared/src/RCS/i_rnhst1.c,v
 *
 * Changed include of shrip.pub and shrip.prv to shrcc.pub and shrcc.prv,
 * respectively.
 *
 * Revision 1.4  1995/01/05  03:58:08  ddavies
 * ddavies on Wed Jan 4 19:54:40 PST 1995
 *
 * i_rnhst1.c -> /product/ipcore/ipshared/src/RCS/i_rnhst1.c,v
 *
 * Add comments about new space-saving measures.  ip_horizRunOf0sHistoLine
 * and ip_horizRunOf1sHistoLine have been consolidated into a single
 * routine, ip_horizRunHistoLine.  This routine has a new arg, runOfZeros,
 * which is set to 0 when looking for runs of 1's and 0xFFFFFFFF when looking
 * for runs of 0's.  The routine XORs runOfZeros with every word read from
 * the image.  So, the cases used to call these old routines are also
 * consolidated.  pByteHisto is no longer cast to be a pointer to UInt8 and
 * ooops, that's it.
 *
 * Revision 1.3  1994/07/06  20:10:43  ddavies
 * i_rnhst1.c -> /product/ipcore/ipshared/src/RCS/i_rnhst1.c,v
 *
 * Make pByteHisto a pointer to UInt8's instead of Int8's in the
 * calls to ip_horizRunOf0sHistoLine, ip_horizRunOf1sHistoLine and
 * ip_horizRunofCyclesHistoLine.
 *
 * Revision 1.2  1994/06/28  22:06:18  ddavies
 * ddavies on Tue Jun 28 15:05:25 PDT 1994
 *
 * i_rnhst1.c -> /product/ipcore/ipshared/src/RCS/i_rnhst1.c,v
 *
 * Add comments about how the masks used on the left and right ends of lines
 * when counting cycle lengths are derived.
 *
 * Revision 1.1  1994/06/27  22:46:56  ddavies
 * Initial revision
 *

*/
