/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_rnhst2.c
 *
 *	A "25 cent" frequency transform: make a "run-length histogram"
 *	for a bitmap. Unlike runHistogram, these routines have
 *	lots of bells and whistles.
 *	And, unlike the routines in i_rnhst1.c, these routines take
 *	histograms of vertical run lengths.
 */


#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "memory.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* Prototypes */
#include "shrcc.pub"
#include "shrcc.prv"
#include "shrip.prv"  /* for ip_makeRotTable prototype */

IP_RCSINFO(RCSInfo, "$RCSfile: i_rnhst2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* The code that finds blocks of glyphs on a page uses histograms of run
 * lengths.  XIS uses them extensively for segmentation.  In one sense,
 * they are a cheap alternative to computing fourier transforms.
 *
 * The module i_runhst.c also contains code to compute histograms of
 * horizontal runs.  If this code works better, the code in i_runhst.c
 * will become obsolete.
 *
 * The module i_rnhst1.c contains code that calculates histograms of
 * horizontal run lengths.  This modules contains code to calculate
 * histograms of vertical run lengths.  While there are several
 * similarities, there are many differences.
 *
 * Because the horizontal and vertical implementations are similar, we'll
 * include the bulk of the comments from i_rnhst1.c.
 *
 ***************************************************************************
 *
 * XIS asked for a routine that has a bunch of arguments.  We've added
 * a couple more.  The arguments are:
 *
 *	pointers to image and rectangle within the image;
 *	pHisto: pointer to histogram to be returned;
 *	numHistoBins: equals the number of bins in the histogram
 *	sampleFreq: skip sampleFreq-1 lines between each sampled row
 *		or column (our case, skip in units of 8 columns);
 *	runType: whether to count black runs, white runs or cycles (lengths
 *		of black/white pairs);
 *	useEndRuns: all runs are terminated at the boundary of the
 *		sampling rectangle.  If useEndRuns is true, we should
 *		count the runs that touch the boundary.  If it is
 *		false, we should ignore these runs.  This is a pain,
 *		but is useful.
 *		Well, in fact, NOT ignoring cycles that abut the
 *		borders is not a well defined operation.  Besides,
 *		it's hard to implement.  So we'll say that if we are
 *		measuring cycles, we always ignore cycles that abut
 *		the edges of the rectangle.
 *	countLongRuns:	if false, we ignore runs that have no bins in
 *		the histogram.  If true, all the counts for the large
 *		runs are put into to the last histogram bin.
 *	weightByLength: if false, finding a run of length x causes us to
 *		add 1 to bin x.  If true, finding a run of length x causes
 *		us to add x to bin x.  What is this for anyway?
 *	
 * This is quite a monster - the swiss army knife approach to writing code.
 * Unfortunately, I think we actually use this stuff.
 *
 * We'd really like to avoid making a zillion decisions for every run found.
 * One way to avoid some decisions is to create a temporary histogram,
 * then process it to produce the histogram we really want.  We will create a
 * temporary histogram containing a bin for every possible run length.
 * If we're looking for horizontal runs, the temporary histogram will have
 * bins for 1 through rectangleWidth.  If we're tracing vertical runs,
 * we need bins for 1 through rectangleHeight.  When we've filled the
 * temporary histogram with run info from the image, we start post processing.
 *
 * If weightByLength is true, we multiply the number in each bin by the
 * bin number.
 *
 * We then add the contents of the first numHistoBins bins from the
 * temporary histogram to the contents of the output histogram.
 *
 * If countLargeRuns is true, we add the contents of the remaining bins
 * in the temporary histogram to the pHisto[numHistoBins-1].
 *
 * Let's review how horizontal histograms are counted.  We will rotate each
 * byte in the image, then use pretend we're counting horizontally
 * on the rotated bytes.
 *
 * Even with just horizontal histograms, the choice of counting runs of
 * 1's, runs of 0's or cycles is not one we'd like to make on every
 * image sample.  We'll have different code for each of these cases.
 * What are cycles?  When finding areas containing periodic data, like
 * glyphs or halftones, it's best to count not just the width of each
 * mark or the width of the space between them, but the period of
 * white/black pairs.  This is the thing that's most likely to be constant.
 * Another way to think of this is that we count the distance between
 * 0-to-1 transitions.
 *
 * The really messy remaining parameter is useEndRuns.  It is much more
 * convenient to count all the runs.  When useEndRuns is false, we must
 * efficiently trace the leftmost and rightmost runs on each scan line,
 * then ignore them.  If we were tracing each run serially, this would
 * be trivial.  Since we're trying to process as much data as we can
 * on each sample, this will be messy.  We will have separate verions of
 * run counting code for these too.
 *
 * The routine at this level, i_horizRunHistogram, will call lower routines
 * to process each line.  These routines will have the real inner loops.
 * What do the inner loops of these routines look like?  The inner loops
 * are similar for all the routines.  The cycle routine looks for somewhat
 * different things than the routines that look for 1's or 0's, but
 * the principle is the same.  The SkipEnds and UseEnds routines differ
 * only in how they start and stop.
 *
 * We clearly want to look at more than one pixel at a time.  The inner
 * loops should be able to skip over runs of 0's and runs of 1's in large
 * jumps.  This is especially true since these are fairly common.  On the
 * other hand, a large chunk of bits might contain many internal runs,
 * so we can't just deal with large chunks all the time.
 *
 * One cute observation is that the inner loops have two basic jobs.  One
 * is to track runs that bridge chunks.  The other is to track runs that
 * start and stop inside individual chunks.  Tracking runs that start and
 * stop purely within chunks need not be done as each chunk is encountered.
 * We can create yet another histogram, this one of chunk values seen.
 * When we get done with the image, we calculate the number of runs that
 * started and stopped inside each chunk and add that into the temporary
 * histogram.  For example, say chunks are bytes and we find an 0xAA.  We
 * increment the 0xAA bin.  At the end of the image, say we've found
 * 70,000 0xAA's.  Say we're tracking runs of 1's.  This byte (1010 1010)
 * has three internal runs (plus one on the left that bridges to the next
 * byte, each of length 1.  We would add 3 * 70,000 to the
 * temporary histogram bin for runs of length 1.
 *
 * Using this trick, we only have to track runs between chunks when
 * we're actually going through the image - the runs inside chunks are
 * done later.  This greatly simplifies the code done for each byte.
 *
 * We use tables to compute things about chunks.  To keep table sizes
 * small, we'd like the chunk size to be fairly small.
 * To be able to skip over runs of black or white efficiently, we'd like
 * the chunk size to be large.  We'll use a combination.  We process the
 * image in units of 32-bit words.  If a word is entirely black or entirely
 * white, we skip over it easily.  If it contains both black and white,
 * we process its four bytes individually.  At the end of the four bytes,
 * we look at the next word.
 *
 * There are two ways we can do the calculations on each byte: we can use
 * a state machine that uses each byte as input or we can derive both
 * the state and output from the region around each byte.  Huh?  The
 * state machine approach is straightforward.  Say we're looking for runs
 * of 1's.  The current state is set by the previous byte - we're either
 * in a run of 0's or a run of 1's:
 *	State	byte starts with	Action
 *	0's	0			keep looking for 1's
 *	0's	1			start new run
 *	1's	0			end current run, start looking for
 *					another
 *	1's	1			continue current run.
 *
 * This is lovely, but state machines are sort of complicated what with
 * all the testing a branching that goes on.
 *
 * Note that the state from the previous byte is equal to the last bit of
 * that byte together with the length of the run we're now in (if any).
 * We could eliminate the state machine stuff by just concatenating that
 * bit with the others in the byte when doing our table lookups.
 *
 * What table lookups?  We use three tables for each byte:
 *   1.	Incremental run length
 *	This table performs two functions.  It indicates when a run from
 *	the previous byte ends and gives the distance by which that run
 *	extends into this byte.  The run ends when the value in this
 *	table is negative.  The trick is that we need two values of 0.
 *	When a run ends at the beginning of this byte, we should store
 *	negative 0.  When there is no run coming into this byte, we store
 *	positive 0.  Since we need both positive and negative values for
 *	0, we use one's complement to store negative values.
 *   2.	Byte Value
 *	This is a histogram of byte values.  When we come to the end of
 *	a run, we increment the bin corresponding to this byte.  We use
 *	these counts later to add in the runs that were purely internal
 *	to the bytes.
 *   3.	New Run Length
 *	This is the number of bits in the run that includes the last
 *	bit of the byte.  We use it to set the run Length counter before
 *	encountering the next byte.
 * 
 * One nice feature of organizing things this way is that the inner loops
 * of the code that finds cycles and the code that finds runs of 1's (or 0's)
 * is all the same.  The only difference is the contents of the tables.
 * But, we use different routines anyway.  If we use different routines, we
 * can address the tables directly in each routine.  If we used one
 * routine, we'd have to pass a pointer to the table, which would use
 * a register.  There aren't a lot of registers in a x86 processor, so
 * we'll use different routines.
 *
 * Do we scan left to right or right to left within a line?  When we
 * encounter a new word, we would like to perform a quick check to see
 * if its the entension to an existing run or contains the end of that
 * run.  We can do this by looking at the last bit of the previous word
 * and the value of the new word:
 *	last bit of prev word	new word	Action
 *		0		00...00		Add 32 to runLength unless
 *						we're looking for runs of 1's
 *		1		11...11		Add 32 to runLength unless
 *						we're looking for runs of 0's
 *		all other combinations		Find end of run in word
 * It's easy to branch on the last bit of the previous word if that bit
 * is in the sign bit position (MSBit).  Therefore, we will go from right
 * to left within a scan line.
 *
 * Do we read each byte individually or do we read entire words into a
 * register and do shifting and masking to form the table lookup values?
 * Reading bytes individually is generally a good idea, but maybe not in
 * this case.  First, there is the annoyance of dealing with different
 * byte orders.  If we read by words, this goes away.  That's not a
 * big deal.  Second, at least the byte value histogram entries are not
 * bytes.  Therefore, the machine will do some shifting anyway.  The
 * current run length will not be a byte, so the value read from the
 * Incremental run length table will either be expanded into a word
 * (an operation that our compilers do very stupidly on PCs) or the
 * incremental run length table entries will be words already.  Ditto for
 * the new run length table.  We can save some shifting of address
 * offsets by making all the tables have the same entry size, so pick
 * 32-bit words.  This also eliminates the operations used to expand
 * smaller entries to words when they're read.
 *
 * Next, we're not addressing at least the incremental run length table
 * with bytes.  We use 9 bits - 8 from the current byte and 1 from the
 * previous byte.  We'd have to compose these entries anyway, so we might
 * as well do this from a register that hold all the bytes at once.  So
 * long as one of the tables is addressed using 9 bits, we might as well
 * address all of them that way - again to avoid extra address computation.
 * Reading the image in units of words and processing it in units of runs
 * of bits from these words has another advantage when processing the
 * end runs as seen below.
 *
 * What do we do with the end runs anyway?  When we are keeping the end
 * runs and looking for runs of 1's (or 0's), we just smash the portions of the
 * works on the far left and right that are outside our rectangle with
 * 0's (or 1's).  When looking for cycles, we smash the portion of the
 * word on the right with 0's (since that cannot start a new run) and
 * the portion of the leftmost word outside the rectangle with 1's
 * (since that cannot end a run).  When ignoring the end runs, we'll
 * duplicate the inner loop.  The first copy is executed only for the
 * very first, rightmost, run.  It does the normal things, except that
 * it doesn't update the temporary histogram for the first run length.
 * It then falls through to the ordinary loop.
 *
 * To ignore the leftmost run, we simply make the inner loop stop before
 * processing the word containing the left end of the rectangle.  The
 * portion of the this word that's outside the rectangle is filled with 1's
 * (0's if we're looking for runs of 0's).  When we get to the end of this
 * word, we simply neglect to update the temporary histogram for the last run.
 *
 * We can combine the routine that ignores ends with the routine that
 * pays attention to ends by choosing the appropriate starting and ending
 * code.  Since this branch only happens twice per line, the performance
 * penalty for this branch should be negligible.
 *
 * How do we actually compute the internal runs inside each byte?  The
 * first thing to notice is that there are only 512 different values.
 * No matter what we do, it won't take very long.  We just take each byte
 * and look up the length of the run at the right end.  We shift right
 * by that amount plus 1, since we were supposed to ignore the run's bit
 * in the word plus the extra bit that belonged to the previous byte.
 * We also calculate the number of bits in the byte that are left over
 * when we subtract off the bits belonging to the runs on the left and
 * right ends.  In a loop, we then look for the length of the run on
 * the right end of the byte using the regular tables.  If the length
 * is negative, the run terminates in the byte.  Increment the temp
 * histogram bin corresponding to (runLength + 1) (to account for the
 * bit in the LS bit of the index - that bit is part of our byte but is
 * not counted by the table) by the number in the byte's entry in the
 * byte histogram.  Let me try that again.  The number in the byte's bin
 * in the byte histogram corresponds to the number of times this byte
 * was seen in the image.  Therefore, each run we find inside this byte
 * should be weighted by this occurance count before we put it into
 * the final histogram of run lengths.
 * Anyway, we shift the remaining byte bits right by runLength+1 and
 * subtract runLength+1 from the number of bits remaining in the byte
 * and go to the top of the loop.  The loop terminates when the number
 * of bits remaining in the byte is 0.
 *
 * The tables are in the include file runhist.prv.  To utilize processor
 * caches efficiently, we have folded the four tables described above into
 * two.  Almost every time we get the distance a run extends into the byte
 * from the right, we also want the distance the next run extends into
 * the same byte from the left.  So, we interleave the two values so that
 * when one appears in the caches, the other is probably also brought in.
 * The table onesOnEnds holds the number of 1's on the left and right ends
 * of each byte.  The table cyclesOnEnds holds the lengths of the cycles
 * on each end of a byte.
 *
 * Yeah, but what about vertical runs?  Well, the routine that looks at
 * vertical runs will be entirely different than the one that looks at
 * horizontal runs, so we'll just make it a different routine altogether.
 * Here's the horizontal routine.
 *
 ************************************************************************
 *
 * So much for histograms of horizontal runs.  How does this apply to
 * vertical runs?
 *
 * Let's say more about this rotation stuff.  Clearly, it would be easy
 * to rotate the entire image, then do horizontal histograms.  Clients
 * could do that easily.  The only reason this routine exists is because
 * the clients ran out of memory - they couldn't afford the space required
 * for the rotated copy.  Therefore, we have to do something clever.
 *
 * We could just read down entire columns of the image.  Unfortunately,
 * the image is stored in memory by horizontal scan lines.  If we move
 * vertically through the image, we access a slice of every scan line.
 * This means we fault every level of the memory heirarchy (processor primary
 * cache, secondary cache and main memory) on every major access.
 * Any possible performance gains we could see by accessing the image
 * vertically are overwhelmed by all this faulting.  We really want
 * to read the image (roughly) horizontally.
 *
 * However, we don't want to read individual scan lines either.  Since we're
 * really computing vertical run lengths, we'd have to look at each
 * pixel individually - incrementing the corresponding run length count
 * if the pixel has the right polarity.  We could try keeping the run
 * length counts in some weird vertical order (one byte holds bit 0 of
 * 8 different counts, the next one holds bit 1 of the eight counts and
 * so on).  Since we need 16 bit counters, we need 32 operations to increment
 * the 8 counts AND we have to decide if we've reached the end of a run
 * in any of the columns anyway.  So, we really don't want to do something
 * with histogram data every time we see a scan line.
 *
 * We'll compromise.  We read the image in units of 8 scan lines - 7 new
 * ones and 1 old one.  In the horizontal code described above, we used
 * 8 new ones and 1 old one.  Why change?  Well, we really read columns
 * of bytes, rotating them into the 8 nibbles of a word as we go.  Each
 * nibble represents the bits in one column of data.  We do this twice,
 * the combine the nibbles to produce 8 bytes.  The 8 bytes are used in
 * a table lookup (as described above) to update the current run length
 * for each column as well as decide when a run is complete.  Yes, this
 * means using different tables.  Bummer.
 *
 * What do we do about the edge conditions?  We would like to base this code
 * on the orthogonal rotation code that blithely ignore the edges - it rotates
 * the frame as needed.  We have to stop exactly on the edges.
 *
 * First, we decide which type of rotation we want.  We choose 270 degree
 * (as opposed to 90 degree) rotation.  This has two advantages.  The
 * existing code works from the bottom of the source image towards the
 * top.  This corresponds to working from right to left in the rotated
 * image, which is what we want.
 *
 * Both 90 and 270 rotation divide the image into blocks of width 8 and
 * height 32.  The rotation operation puts together 8 32-bit words and
 * writes them to memory.  We could also produces 8 32-bit words, then
 * compute histograms from each.  This allows us to closely follow the
 * logic of the horizontal histogram code.  On the other hand, it forces
 * us to spend a bunch of cycles putting the words together, then take
 * them apart to compute the histograms.  We'll just compute the histograms
 * on each block of 8 bytes, then throw them away.
 *
 * The simplest thing to do it process and 8x8 block, then move to the
 * next block on to the right.  Do we gain anything interesting from
 * processing several 8x8 blocks vertically before moving horizontally?
 * We could keep using the same run length counters for a longer time
 * (there's one run length counter for each column of bits).  Well, the
 * run length counters are going to be kept in an array anyway.  There
 * are at least 8 of them per 8x8 block.  That's too many to keep in
 * x86 registers - probably too many for the SPARC compiler to handle
 * as well.  So the compiler will almost certainly keep the things in
 * memory anyway.  Since we don't eliminate any memory reads or writes,
 * we might as well keep the loop simple and just go from right to left.
 *
 * Back to edge conditions.  In 270 degree rotation, the upper right corner
 * of the source image becomes the upper left corner of the dest.
 * Therefore, the grid of blocks is aligned over the source image so
 * that the upper right corner of the grid corresponds to the upper
 * right corner of the image.  The top row of blocks is exactly aligned
 * with the top row of the image and the right row of blocks lines up
 * with the right row of the image.  Therefore, the left and bottom sides
 * either happen to also line up or the blocks hang off the edges of the
 * image.
 *
 * In our case, we will align the top row of the source rectangle with
 * the top row of blocks.  The right side of the column of bytes
 * containing the right edge of the rectangle will be aligned with the
 * right edge of the blocks.  In our case, there can be extra bits on
 * the left, right and bottom.  In fact, even if there are no extra
 * bits on the bottom, we have to make one up for the very bottom row
 * since we need to make up a bit from the "previous" row.
 *
 * So, the things we can do when we read a blocks are:
 *	1. apply the left mask.
 *	2. apply the right mask.
 *	3. apply the left and right masks.
 *	4. apply the bottom mask.
 *	5. apply the left and bottom masks.
 *	6. apply the right and bottom masks.
 *	7. apply the left, right and bottom masks.
 *	8. apply no mask.
 *
 * The last condition happens almost all the time, except in very small
 * images where performance isn't a problem.  We'll have two
 * versions of the code that processes 8x8 blocks.  One applies all the
 * masks.  The other applies none of them.  We have to set up all the masks
 * before calling the first block of code.  We call it all along the bottom
 * of the rectangle and at both left and right edges of each line.  Just
 * to keep things tidy, we'll make it a subroutine.  We only call the
 * no-mask processing code from one place, so we just write it out.
 *
 * The inner loops are implemented in mirnhst2.c
 * 
 */


/* PUBLIC */
/*****************************************************************************
 * i_vertRunHistogram
 *
 *	Make a histogram of the length of vertical runs of 1's or vertical
 *	distances between 1->0 transitions.
 *
 *	This routine computes a temporary histogram, then folds it into
 *	the client's desired histogray according to the parameters given
 *	below.
 *
 *	The client's histograms should be zeroed by the client before entry.
 *	This routine just adds to the existing histogram.
 *
 * Arguments:
 *	pImage =	Pointer to the image in question, including frame
 *	bpl =		Bytes per line of image, including frame
 *	x, y =		(x,y) offset into image, from outside of frame
 *	w, h =		width and height of rectangle in which histogram
 *			is computed
 *	pHistos =	Pointer to client's histogram.
 *	numHistoBins =	Number of Int32's in the array pointed to by
 *			pHistos.
 *	sampleFreq =	the image is sampled in units of 8 columns.
 *			Skip sampleFreq-1 columns between samples
 *	runType	=	If equal to cRunOfZeros, look for runs of 0's.
 *			If equal to cRunOfOnes, look at runs of 1's.  If
 *			equal to cRunOfCycles, look at adjacent runs of
 *			0's and 1's.
 *	useEndRuns =	if equal to cTrue, count the runs that touch the
 *			borders of the rectangle, otherwise ignore them.
 *			However, if we are measuring cycles, we ALWAYS
 *			ignore the cycles that abut the edges of the
 *			rectangle.  Not ignoring them is a poorly-defined
 *			operation.
 *	countLongRuns =	if cTrue, put the counts for all runs that have
 *			no bins in the client's histogram into the
 *			last bin.  If cFalse, ignore those runs.
 *	weightByLength = if cTrue, weight each run by its length when
 *			recording it it its bin.  For example, finding
 *			a run of length n would cause n to be added to bin n.
 *
 * Returns:
 *	ia_successful		everything worked
 *	ia_nomem		not enough memory for tables
 *	ia_invalidParm		something wrong with parameters
 *
 *****************************************************************************/
Int32 CDECL 
i_vertRunHistogram (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32      *pHistos,
    Int32	numHistoBins,
    Int32	sampleFreq,
    Int32	runType,
    Bool	useEndRuns,
    Bool	countLongRuns,
    Bool	weightByLength)
{
    PROLOGUE;
    UInt32	*pTempHisto;	    /* temporary histogram used to
				     * compute new values */
    UInt32	*pByteHisto;	    /* keeps track of the number of times
				     * we've seen each byte that might
				     * contain an internal run */
    Int32	*pRunLengths;	    /* array of run length counts.  We need
				     * one count per column of bits we
				     * actually count.
				     */
    UInt8	 leftByteAndMask;   /* used to delete areas outside */
    UInt8	 leftByteOrMask;    /* of rectangle in left word of scan line */
    UInt8	 rightByteAndMask;  /* used to delete areas outside */
    UInt8	 rightByteOrMask;   /* of rect in right word of scan line. */
    Int32	 widthInBlocks;	    /* width of rectangle in units
				     * of 8x8 blocks */
    Int32	 heightInBlocks;    /* height of rect in blocks */
    Int32	 rightCoord;	    /* This is the rightmost pixel that we
				     * actually touch in a scan line.  If
				     * sampleFreq != 1, this may not be the
				     * rightmost pixel in the rect.
				     */
    UInt32	 blockAndMask;	    /* specifies the number of bits to retain
				     * from each rotated byte.  This has 0's
				     * in it only when we're looking at the
				     * bottom row of blocks in the rect.
				     */
    UInt32	 blockOrMask;	    /* contains 1's in bits that should be set
				     * at in the bytes from the bottom block. */
    UInt32	*pRotationTable;    /* pointer to table used to do rotations */
    Int8	*pRunTable;	    /* points to tables used to calculate run
				     * lengths */
    
	/* The following are used to count the runs in the byte histogram */
    Int32	 index;		    /* used for histograms */
    UInt32	 trialByte;	    /* used to hold values of bytes when looking
				     * through the byte histogram */
    Int32	 byteRunLength;	    /* used to calculate the length of runs
				     * inside bytes */
    Int32	 bitsRemaining;	    /* number of bits not yet explored
				     * in a byte */

	/* set these so we'll know these are well defined no matter
	 * when we quit.
	 */
    pTempHisto = pByteHisto = NULL;
    pRunLengths = NULL;
    pRotationTable = NULL;

	/* if the client told us to count cycles and pay attention to
	 * the cycles that abut the borders of the rectangle,
	 * complain.  We don't know how to do that.
	 */
    if (useEndRuns && (runType == cRunOfCycles))
    {
	__error__ = ia_invalidParm;
	goto finVertRunHisto;
    }

	/* allocate the temporary histogram.  We want one entry for every
	 * possible run length.
	 */
    pTempHisto = (UInt32 *) CALLOC(h+1, sizeof(UInt32));
    if (pTempHisto == NULL)
    {
	__error__ = ia_nomem;
	goto finVertRunHisto;
    }

	/* allocate the byte histogram.  We want one entry for every
	 * possible byte value.  Since we really index this thing with
	 * 8 bit things and there are two entries per thing,
	 * we need 512*2 entries.
	 */
    pByteHisto = (UInt32 *) CALLOC(256*2, sizeof(UInt32));
    if (pByteHisto == NULL)
    {
	__error__ = ia_nomem;
	goto finVertRunHisto;
    }

	/* make pImage point to the upper left byte we'll sample */
    pImage += y * bpl + x/8;

	/* Compute the table we'll use to expand bytes into 32-bit things */
    __error__ = ip_makeRotTable (pIPState, &pRotationTable, 1);
    if (__error__ != ia_successful)
    {
	goto finVertRunHisto;
    }

	/* set up all parameters we use to do a fake rotation. */
    widthInBlocks = (((x+w-1)/8 - x/8) / sampleFreq) + 1;
    heightInBlocks = (h-1)/7 + 1;

	/* Find the rightmost pixel we actually touch.  This is
	 * is used to set the right byte masks.
	 */
    rightCoord = x + MIN(w, widthInBlocks*sampleFreq*8) - 1;

	/* since we look at 8 lines in each block, even though we
	 * only advance up 7 lines in each iteration, the number
	 * of lines at the bottom of the bottom block is one
	 * larger than would normally be the case.
	 * This mask has 1's for the *rotated* bits we want
	 * to keep in each byte.  There will be four rotated
	 * bytes in each word, so we replicate the mask into
	 * each byte.
	 */
    blockAndMask = 0xFF << ((heightInBlocks * 7) - h + 1);
    blockAndMask &= 0xFF;			/* we only want what's left
						 * in the bottom byte */
    blockAndMask |= blockAndMask << 8;		/* put mask in two bytes */
    blockAndMask |= blockAndMask << 16;		/* put mask into all bytes */

	/* Set the source so it points to the upper left corner of
	 * the rectangle in the lower left of the image.  The rects
	 * are aligned with the top line of the image.  We do this
	 * weird thing so we can apply the bottom mask to the last
	 * bits in the rect.
	 */
    pImage += ((heightInBlocks - 1) * 7) * bpl;

	/* allocate the array of run lengths.  We need one for every
	 * column of bits that's counted.  We count every column in
	 * every block we look at, including the ones off the left and
	 * right ends of the rectangle.
	 */
    pRunLengths = (Int32 *) CALLOC(widthInBlocks*8, sizeof(Int32));
    if (pRunLengths == NULL)
    {
	__error__ = ia_nomem;
	goto finVertRunHisto;
    }

	/* If we're supposed to count the end runs, set all the run
	 * length counts to 0.  If we're supposed to ignore the
	 * end runs, set them all to big negative numbers (0x80808080).
	 * The inner loops look at the run length counts to decide
	 * whether to store the first run lengths.  If they're
	 * negative, we skip the store.
	 */
    if (useEndRuns)
	MEMSET(pRunLengths, 0, widthInBlocks*8*sizeof(Int32));
    else
	MEMSET(pRunLengths, 0x80, widthInBlocks*8*sizeof(Int32));

	/* get the address of the table used to compute runLengths.
	 * In the horizontal run length code, we had three separate
	 * lower level routines.  Here, we only have one because
	 * it's so complicated.  The percentage of time it spends
	 * using this pointer is fairly low and is localized
	 * in the code, so loading it from memory isn't awful.
	 */
    __error__ = ip_getVertRunTablePtr (pIPState, runType, &pRunTable);
    if (__error__ != ia_successful)
	goto finVertRunHisto;


	/* compute masks used to delete left and right ends of the rectangle.
	 * There are 5 cases:
	 * 1. runs of 1's; use runs on the ends.
	 *	the areas outside the rectangles are both filled with 0's
	 *	so the runs are truncated at the boundaries.
	 * 2. runs of 1's; discard runs on the ends.
	 *	the areas outside the rectangles are both filled with 1's.
	 *	This allows us to track the end runs without worrying about
	 *	exactly where the boundary is.
	 * 3. run of 0's; use runs on ends.
	 *	the areas outside of the rectangle are both filled with 1's
	 *	so the runs are truncated at the boundaries.
	 * 4. runs of 0's; discard runs on the ends.
	 *	the areas outside the rectangles are both filled with 0's.
	 *	This allows us to track the end runs without worrying about
	 *	exactly where the boundary is.
	 * 5. runs of cycles; we always discard the cycles on the ends.
	 *	The area to the right of the rectangle should be filled
	 *	with 0's so it acts as an extension of the cycle on the
	 *	right.  The area to the left of the rectangle should
	 *	be filled with 1's so it acts as an extension of the
	 *	leftmost cycle.
	 * We pass both masks and useEndRuns down to the routine
	 * that processes lines.  runType and useEndRuns tell the routine
	 * what to do with the masks.
	 *
	 * So long as we're switching on runType, we might as well
	 * do all the primary data gathering here too.  The lower
	 * level routine called depends on the run type.
	 */
    switch (runType)
    {
    case cRunOfZeros:
	if (useEndRuns)	/* set areas outside of rect to 1 */
	{
	    leftByteOrMask = ~(0xFF >> (x % 8));
	    leftByteAndMask = 0xFF;
	    rightByteOrMask = 0x7F >> (rightCoord % 8);
	    rightByteAndMask = 0xFF;
	    blockOrMask = ~blockAndMask;
	}
	else	/* set areas outside of rect to 0 */
	{
	    leftByteOrMask = 0;
	    leftByteAndMask = 0xFF >> (x % 8);
	    rightByteOrMask = 0;
	    rightByteAndMask = ~(0x7F >> (rightCoord % 8));
	    blockOrMask = 0;
	}
	break;

    case cRunOfOnes:
	if (useEndRuns)	/* set areas outside of rect to 0 */
	{
	    leftByteOrMask = 0;
	    leftByteAndMask = 0xFF >> (x % 8);
	    rightByteOrMask = 0;
	    rightByteAndMask = ~(0x7F >> (rightCoord % 8));
	    blockOrMask = 0;
	}
	else	/* set areas outside of rect to 1 */
	{
	    leftByteOrMask = ~(0xFF >> (x % 8));
	    leftByteAndMask = 0xFF;
	    rightByteOrMask = 0x7F >> (rightCoord % 8);
	    rightByteAndMask = 0xFF;
	    blockOrMask = ~blockAndMask;
	}
	break;

    case cRunOfCycles:
	    /* Set area to either side of the rect to 0's (either
	     * polarity will do).
	     */
	    leftByteOrMask = 0;
	    leftByteAndMask = 0xFF >> (x % 8);
	    rightByteOrMask = 0;
	    rightByteAndMask = ~(0x7F >> (rightCoord % 8));
	    blockOrMask = 0;
	break;

    default:
	__error__ = ia_invalidParm;
	goto finVertRunHisto;
    }


	/* loop through rows of rectangles of source image from the
	 * bottom to the top.  Note that since we need to look at the
	 * bit from the previous line each time, we need only step
	 * up 7 lines each iteration instead of 8.
	 */
    for (__error__ = ia_successful;
	 heightInBlocks > 0;
	 pImage -= (bpl*8 - bpl), heightInBlocks--)
    {
	__error__ = ip_vertRunHistoLine (pIPState, pImage,
				       bpl,
				       leftByteAndMask,
				       leftByteOrMask,
				       rightByteAndMask,
				       rightByteOrMask,
				       pTempHisto,
				       pByteHisto,
				       pRotationTable,
				       pRunTable,
				       pRunLengths,
				       widthInBlocks,
				       sampleFreq,
				       blockAndMask,
				       blockOrMask);
	if (__error__ != ia_successful)
	    goto finVertRunHisto;
	blockAndMask = 0xFFFFFFFF;	/* there are no spare lines
					 * above the bottom row */
    }

	/* If we're using the runs that abut the rect boundaries,
	 * add the last run lengths into the histogram.
	 */
    if (useEndRuns)
    {
	for (index = 0; index < widthInBlocks*8; index++)
	{
	    pTempHisto[pRunLengths[index]]++;
	}
	pTempHisto[0] = 0;		/* toss all the zero counts */
    }

	/* At this point, the combination of pTempHisto and pByteHisto
	 * have the relevant run info.  Compute the internal runs from
	 * pByteHisto and put them into pTempHisto.  First, get the
	 * pointer to the table we'll use to decipher the runs.
	 */
    for (index = 0; index < 256*2; index += 2)
    {
	if (pByteHisto[index] == 0)
	    continue;				/* no image bytes had
						 * this value */
	trialByte = index;			/* there are two entries
						 * in the table for each
						 * index val */
	byteRunLength = pRunTable[trialByte];
	if (byteRunLength < 0)			/* a run enters from the
						 * right and ends in
						 * this byte */
	    byteRunLength = ~byteRunLength;	/* get the real count */

	    /* find the number of bits in this byte that aren't part
	     * of the runs on either end.  The bits in the byte
	     * that aren't part of the runs on either end may
	     * contain a new run.
	     */
	bitsRemaining = 7 - pRunTable[trialByte+1] - byteRunLength;

	    /* go through the internal runs of this byte */
	while (bitsRemaining > 0)
	{
	    trialByte >>= byteRunLength + 1;	/* shift out the bits of the
						 * run plus the LS bit from
						 * the "previous" run. */
	    trialByte &= ~1;			/* create a new address mask */
	    byteRunLength = pRunTable[trialByte];
	    if (byteRunLength < 0)
	    {
		    /* We've found an internal run.  Record its length
		     * in the temporary histogram.  Note that the length
		     * of this run includes the bit that would normally
		     * be part of the "previous byte."  We shifted a
		     * bit from our current byte into that position.
		     * The table entries don't count that bit as part
		     * of the run in the current byte, so we have to
		     * add it in.
		     */
		byteRunLength = ~byteRunLength;
		pTempHisto[byteRunLength + 1] += pByteHisto[index];
	    }
	    bitsRemaining -= byteRunLength + 1;	/* get rid of this run */
	}
    }

	/* At this point, the temporary histogram contains all the runs
	 * in the image.  If we are supposed to weight them by length,
	 * do so.  To avoid extra handling, do this while putting them
	 * into the client's histogram.
	 */
    if (weightByLength)
    {
	for (index = 0; index < MIN(numHistoBins, h+1); index++)
	    pHistos[index] += pTempHisto[index] * index;

	if (countLongRuns)	/* put the rest of the entries in pTempHisto
				 * into the client's last bin */
	{
	    for (; index <= h; index++)
		pHistos[numHistoBins-1] += pTempHisto[index] * index;
	}
    }
    else	/* don't weight by each entry by length */
    {
	for (index = 0; index < MIN(numHistoBins, h+1); index++)
	    pHistos[index] += pTempHisto[index];

	if (countLongRuns)	/* put the rest of the entries in pTempHisto
				 * into the client's last bin */
	{
	    for (; index <= h; index++)
		pHistos[numHistoBins-1] += pTempHisto[index];
	}
    }

	/* that's it.  We just free the stuff we've allocated and return */
finVertRunHisto:
    if (pRotationTable != NULL)
	FREE(pRotationTable);
    if (pByteHisto != NULL)
	FREE(pByteHisto);
    if (pTempHisto != NULL)
	FREE(pTempHisto);
    if (pRunLengths != NULL)
	FREE(pRunLengths);
    NORMAL_RETURN;
}

/*

$Log: i_rnhst2.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.9  1998/01/13 23:49:30  ddavies
i_rnhst2.c -> /product/ipcore/ipshared/src/RCS/i_rnhst2.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1997/04/24 23:29:28  ddavies
i_rnhst2.c -> /product/ipcore/ipshared/src/RCS/i_rnhst2.c,v

Change types of useEndRuns countLongRuns and weightByLength parameters
from Int32 to Bool.

Revision 1.7  1995/12/13 02:45:41  ejaquith
i_rnhst2.c -> /product/ipcore/ipshared/src/RCS/i_rnhst2.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/01/11  22:29:41  dferrell
 * i_rnhst2.c -> /product/ipcore/ipshared/src/RCS/i_rnhst2.c,v
 *
 * Removed include of shrip.pub and added shrcc.pub and shrcc.prv.
 *
 * Revision 1.5  1995/01/05  03:59:21  ddavies
 * i_rnhst2.c -> /product/ipcore/ipshared/src/RCS/i_rnhst2.c,v
 *
 * Small changes to use run tables composed of bytes, not words.  The type
 * of pRunTable is set pointer to Int8 and pByteHisto is no longer cast to
 * be a pointer to UInt8.
 *
 * Revision 1.4  1994/07/14  16:21:02  lperry
 * lperry on Thu Jul 14 09:15:23 PDT 1994
 *
 * i_rnhst2.c -> /product/ipcore/ipshared/src/RCS/i_rnhst2.c,v
 *
 * Fixed args to memset by using a casting macro MEMSET, added to
 * memory.pub . GCC didn't complain because it had a proper prototype
 * in ansiprot.h. This was only a problem with acc because it doesn't
 * have proper prototypes.
 *
 * Revision 1.3  1994/07/11  23:12:41  ddavies
 * i_rnhst2.c -> /product/ipcore/ipshared/src/RCS/i_rnhst2.c,v
 *
 * Free pRotationTable at the end of i_vertRunHistogram.  This fixes a memory
 * leak.
 *
 * Revision 1.2  1994/07/08  20:29:25  ddavies
 * ddavies on Fri Jul 8 13:28:58 PDT 1994
 *
 * i_rnhst2.c -> /product/ipcore/ipshared/src/RCS/i_rnhst2.c,v
 *
 * Fix cast of pByteHisto in call to ip_vertRunHistoLine.  It was an Int8,
 * should have been a UInt8.
 *
 * Revision 1.1  1994/07/01  04:25:26  ddavies
 * Initial revision
 *

*/
