/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_rnmrph.c--include
 *
 *	i_dilateHorizontalRuns ()
 *	i_dilateVerticalRuns ()
 *	i_erodeHorizontalRuns ()
 *	i_erodeVerticalRuns ()
 *
 */

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"
#include "shrip.pub"
#include "shrcc.pub"
#include "shrrast.pub"   /* for i_rasterOpFull () */
#include "runflt.prv"    /* for run blitting tables */

IP_RCSINFO(RCSInfo, "$RCSfile: i_rnmrph.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");


/**************************************************************************
 * i_dilateHorizontalRuns ()
 *
 *	Dilate horizontal black runs.
 *
 * Arguments:
 *	runBuf :	Pointer to run buffer of the appropriate size
 *			as determined by i_getRunFilterBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	runBufferSize :	If the run buffer is pre-allocated (preferred),
 *			the size of the buffer (in bytes).  If a NULL
 *			pointer is passed for the run buffer this value
 *			is ignored.
 *	srcImage :	Pointer to source image data including frame.
 *	width :		Width of source and dest images (MUST be the same).
 *	height :	Height of source and dest images (MUST be the same).
 *	srcBpl :	Bpl of source image.
 *	destBpl :	Bpl of dest image.
 *	strelWidth :	Width of dilation "strel."  The strel is a
 *			solid horizontal line of hits.
 *	strelCenter :	Center location of "strel."  0 -> left end
 *			and strelWidth - 1 -> right end.
 *	destImage :	Pointer to dest image data including frame.
 *
 * Returns:
 *	ia_successful	if filter operation completed normally,
 *	error codes	if bad things happened during filtering.
 *
 * Notes:
 *	If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be ORed with the results produced
 *	by this function.  Be sure to set the output image
 *	to white before calling i_dilateHorizontalRuns () if  
 *	you only want the result of a single filter operation.
 *
 **************************************************************************/
Int32 CDECL 
i_dilateHorizontalRuns (
    IPState    *pIPState,
    UInt8      *runBuf,
    Int32	runBufferSize,
    UInt8      *srcImage,
    Int32	width,
    Int32	height,
    Int32	srcBpl,
    Int32	destBpl,
    Int32	strelWidth,
    Int32	strelCenter,
    UInt8      *destImage)
{
    PROLOGUE;
    Int32   row,            /* current row of input image */
	    runCount,       /* number of runs found in current row */
	    runCountD,      /* run count after dilation */
	    runBufAlloced = cFalse,
	    inPlace;        /* TRUE if destImage == srcImage */
    register Int32 length;         /* length of current run */
    register Int32 cRunWord;
    register Int32 lRunWord;
    register Int32 i;
    register Int32 wc;
    register Int32 wm;
    register Int32 w0;
    register Int32 w1;
    register Int32  left;
    register Int32  right;    /* left and right offsets of current run */
    register UInt32 outWord;        /* output word being constructed */
    Run32  *runBuffer,      /* array of runs to pass to run finder */
           *dummyRun,
	   *outRun;
    register UInt32 *rowPtr;


    if (runBuf == NULL)
    {
        runBufferSize = (((width - (2 * cFrameBits)) / 2) + 3);
        runBuffer = (Run32 *)MALLOC (runBufferSize * sizeof (Run32));
	if (runBuffer == NULL)
            return (ia_nomem);
	else
	    runBufAlloced = cTrue;
    }
    else
    {
        runBuffer = (Run32 *)runBuf;
	/* this is passed in as size in bytes */
	runBufferSize /= sizeof (Run32);
    }

    if (srcImage == destImage)
        inPlace = cTrue;
    else
        inPlace = cFalse;

    srcImage += (srcBpl << cLogFrameBits) + cFrameBytes;
    destImage += (destBpl << cLogFrameBits) + cFrameBytes;

    width -= 2 * cFrameBits;
    height -= 2 * cFrameBits;

    /* Get the left and right sub-width parameters for the "strel" */
    w0 = strelCenter;
    w1 = strelWidth - strelCenter - 1;
    wm = width - 1;


    /* For now we treat in place and in to out the same.  It is not
     * immediately obvious that is better to write only the differences
     * in the in place case.
     */
    /* filter runs one row at a whack */
    for (row = 0; row < height; row++)
    {
	/* get all the runs in the current row */
	__error__ = i_findAllRunsInLineArray (pIPState, srcImage,
					    width,
					    0,
					    runBuffer,
					    runBufferSize,
					    0,
					    &dummyRun);

	if (__error__ != ia_successful)
	    break;

	runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
			   sizeof (Run32));

	if (runCount > 0)
	{
	    runCountD = runCount;
	    /* dilate each run and merge overlapping/abutting runs */
	    outRun = runBuffer;
	    outRun->runLeft = MAX (0, outRun->runLeft - w0);
	    outRun->runRight = MIN (wm, outRun->runRight + w1);
	    for (i = 1; i < runCount; i++)
	    {
		left = MAX (0, runBuffer[i].runLeft - w0);
		right = MIN (wm, runBuffer[i].runRight + w1);

		/* does this run merge with the previous one? */
		if (left <= outRun->runRight + 1)
		{
		    /* merge runs */
		    outRun->runRight = right;
		    runCountD--;
		}
		else
		{
		    outRun++;
		    outRun->runLeft = left;
		    outRun->runRight = right;
		}
	    }

	    /* write out the new runs */
	    outWord = 0x00000000;
	    lRunWord = 0;

	    /* go thru the modified run array for the current row */
	    for (i = 0; i < runCountD; i++)
	    {
		left = runBuffer[i].runLeft;
		right = runBuffer[i].runRight;
		length = right - left + 1;

		/* construct the run in the output image */
		cRunWord = left / 32;
		wc = (right / 32) - cRunWord + 1;

		/* if we moved to a new word, write the old one */
		if (cRunWord != lRunWord)
		{
		    /* write the output word then clear it */
		    ULONG_ACCESS ((UInt32 *)destImage + lRunWord) |=
					       outWord;
		    outWord = 0x00000000;
		    lRunWord = cRunWord;
		}

		/* is this a single or multiple word run? */
		if (wc == 1)
		{
		    outWord |= (endRunPosLUT[(left % 32)]) &
			       (endRunNegLUT[(right % 32)]);
		}
		else
		{
		    rowPtr = (UInt32 *)destImage + cRunWord;
		    ULONG_ACCESS (rowPtr) |= (outWord |
			    endRunPosLUT[(left % 32)]);
		    wc--;
		    rowPtr++;
		    outWord = 0x00000000;

		    /* write all of the full middle words */
		    for (; wc > 1; wc--)
		    {
			ULONG_ACCESS (rowPtr) = 0xffffffff;
			rowPtr++;
		    }

		    /* start the next output word with the tail
		     * end of this run.
		     */
		    outWord |= (endRunNegLUT[(right % 32)]);

		    lRunWord = right / 32;
		}

	    }  /* for each dilated run */

	    /* Be sure to write out the last word we worked on! */
	    ULONG_ACCESS ((UInt32 *)destImage + lRunWord) |= outWord;

	}  /* if the row had any runs to dilate */

	srcImage += srcBpl;
	destImage += destBpl;

    }  /* for each image row */

    /* if we allocated this locally, free it */
    if (runBufAlloced)
        FREE (runBuffer);

    NORMAL_RETURN;

}  /* i_dilateHorizontalRuns () */


/**************************************************************************
 * i_dilateVerticalRuns ()
 *
 * Dilate vertical black runs. 
 *
 * Arguments:
 *	runBuf :	Pointer to run buffer of the appropriate size
 *			as determined by i_getRunDilateBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	runBufferSize :	If the run buffer is pre-allocated (preferred),
 *			the size of the buffer (in bytes).  If a NULL
 *			pointer is passed for the run buffer this value
 *			is ignored.
 *	rotBuf :	Pointer to rotation buffer of the appropriate
 *			size as determined by i_getRunDilateBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	rotBufBpl :	Bytes per line of the rotation buffer as
 *			determined by i_getRunDilateBufferSizes ().  If
 *			a NULL pointer is passed for the rotation buffer
 *			this value is ignored.
 *	srcImage :	Pointer to source image data including frame.
 *	width :		Width of source and dest images (MUST be the same).
 *	height :	Height of source and dest images (MUST be the same).
 *	srcBpl :	Bpl of source image.
 *	destBpl :	Bpl of dest image.
 *	strelHeight :	Height of dilation "strel."  The strel is a
 *			solid horizontal column of hits.
 *	strelCenter :	Center location of "strel."  0 -> top end
 *			and strelWidth - 1 -> bottom end.
 *	destImage :	Pointer to dest image data including frame.
 *
 * Returns:
 *	ia_successful	if dilate operation completed normally,
 *	error codes	if bad things happened during dilation.
 *
 * Notes:
 *
 *	The vertical run dilate is necessarily slower than
 *	its horizontal counterpart since it performs two
 *	rotations of each column (actually it works on
 *	blocks of columns) in addition to the run finding/
 *	dilation operations.
 *
 *	If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be replaced with the results
 *	produced by this function.  This differs from
 *	i_dilateHorizontalRuns ().  To combine results from
 *	a vertical and a horizontal run dilate, perform the
 *	the vertical dilate first, then the horizontal using the
 *	same destination image for both operations.
 *
 **************************************************************************/
Int32 CDECL 
i_dilateVerticalRuns (
    IPState    *pIPState,
    UInt8      *runBuf,
    Int32	runBufferSize,
    UInt8      *rotBuf,
    Int32	rotBufBpl,
    UInt8      *srcImage,
    Int32	width,
    Int32	height,
    Int32	srcBpl,
    Int32	destBpl,
    Int32	strelHeight,
    Int32	strelCenter,
    UInt8      *destImage)
{
    PROLOGUE;
    Int32   col,            /* current column of input image */
            wu,             /* width unframed */
	    lastSlice,      /* widht of last vertical slice processed */
	    runBufAlloced = cFalse,
	    rotBufAlloced = cFalse;
    UInt8  *oDestImage,     /* Original dest image pointer */
           *tmpImage;       /* tmp image data pointer */


    __error__ = ia_successful;

    if (runBuf == NULL)
    {
        runBufferSize = (((height - (2 * cFrameBits)) / 2) + 3);
        runBuf = (UInt8 *)MALLOC (runBufferSize * sizeof (Run32));
	if (runBuf == NULL)
            return (ia_nomem);
	else
	    runBufAlloced = cTrue;
    }



    /* save the original dest image pointer for later frame clearing */
    oDestImage = destImage;

    if (rotBuf == NULL)
    {
        rotBufBpl = ((((height + 31) / 32) * 4));
        tmpImage = (UInt8 *)CALLOC (1, rotBufBpl * 
	                           (COL_BLOCK_SIZE + (2 * cFrameBits)));
	if (tmpImage == NULL)
	{
	    if (runBufAlloced)
	        FREE (runBuf);

	    return (ia_nomem);
	}
	else
	    rotBufAlloced = cTrue;
    }
    else
	tmpImage = rotBuf;


    lastSlice = width % COL_BLOCK_SIZE;
    if (lastSlice == 0)
        lastSlice = COL_BLOCK_SIZE;

    wu = width - (2 * cFrameBits);

    /* process the image in 32 column vertical bands */
    for (col = 0; col < wu; col += COL_BLOCK_SIZE)
    {
	/* rotate COL_BLOCK_SIZE columns of the original image 90 into 
	 * the temp image.
	 */
	__error__ = i_fastRot90 (pIPState, tmpImage, srcImage, 
	                       (COL_BLOCK_SIZE + (2 * cFrameBits)), 
	                       (height), 
			       srcBpl, 
			       rotBufBpl, 1);

        if (__error__ != ia_successful)
	    break;

	/* dilate horizontal runs in place in the temp image */
        __error__ = i_dilateHorizontalRuns (
			pIPState, runBuf, runBufferSize,
			tmpImage,
			height,
			(((col + COL_BLOCK_SIZE) < wu) ? 
				(COL_BLOCK_SIZE + (2 * cFrameBits)) : 
				(lastSlice + (2 * cFrameBits))),
			rotBufBpl, rotBufBpl, 
			strelHeight,
			strelHeight - strelCenter - 1,
			tmpImage);

        if (__error__ != ia_successful)
	    break;

	/* rotate the temp image 270 degrees into either the output image
	 * or back into the input image if in place.
	 */
	__error__ = i_fastRot270 (pIPState, destImage, tmpImage, 
	                        height, 
		                COL_BLOCK_SIZE + (2 * cFrameBits),
			        rotBufBpl, destBpl, 1);

        if (__error__ != ia_successful)
	    break;

        /* increment the pointers to the next column block */
        srcImage += COL_BLOCK_SIZE / 8;
	destImage += COL_BLOCK_SIZE / 8;

    }  /* for each column block */

    /* Clear the right frame if necessary since the rotate may have
     * written trash there on the last iteration.
     */
    if ((lastSlice != 0) && (__error__ == ia_successful))
    {
        __error__ = i_rasterOpFull (pIPState, oDestImage, 
	                          width,
	                          height,
				  0,
				  0,
				  width - cFrameBits,
				  0,
				  cFrameBits,
				  height,
				  1,
				  destBpl,
				  PIX_CLR,
				  NULL, 0, 0, 0, 0, 0, 0, 0, 0);
    }

    /* if we allocated this locally, free it */
    if (runBufAlloced)
        FREE (runBuf);
    if (rotBufAlloced)
        FREE (tmpImage);

    NORMAL_RETURN;

}  /* i_dilateVerticalRuns () */


/**************************************************************************
 * i_erodeHorizontalRuns ()
 *
 *	Erode horizontal black runs.
 *
 * Arguments:
 *	runBuf :	Pointer to run buffer of the appropriate size
 *			as determined by i_getRunFilterBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	runBufferSize :	If the run buffer is pre-allocated (preferred),
 *			the size of the buffer (in bytes).  If a NULL
 *			pointer is passed for the run buffer this value
 *			is ignored.
 *	srcImage :	Pointer to source image data including frame.
 *	width :		Width of source and dest images (MUST be the same).
 *	height :	Height of source and dest images (MUST be the same).
 *	srcBpl :	Bpl of source image.
 *	destBpl :	Bpl of dest image.
 *      borderValue :   "Color" of border beyond the image (0 = white,
 *                      1 = black).  The border is assumed to be infinite
 *                      for run length calculations (e.g. a black run which
 *			abuts a black border is assumed to be infinitely
 *			long and is eroded appropriately).
 *	strelWidth :	Width of erosion "strel."  The strel is a
 *			solid horizontal line of hits.
 *	strelCenter :	Center location of "strel."  0 -> left end
 *			and strelWidth - 1 -> right end.
 *	destImage :	Pointer to dest image data including frame.
 *
 * Returns:
 *	ia_successful	if filter operation completed normally,
 *	error codes	if bad things happened during filtering.
 *
 * Notes:
 *	If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be ORed with the results produced
 *	by this function.  Be sure to set the output image
 *	to white before calling i_erodeHorizontalRuns () if  
 *	you only want the result of a single filter operation.
 *
 **************************************************************************/
Int32 CDECL 
i_erodeHorizontalRuns (
    IPState    *pIPState,
    UInt8      *runBuf,
    Int32	runBufferSize,
    UInt8      *srcImage,
    Int32	width,
    Int32	height,
    Int32	srcBpl,
    Int32	destBpl,
    Int32	borderValue,
    Int32	strelWidth,
    Int32	strelCenter,
    UInt8      *destImage)
{
    PROLOGUE;
    Int32   row,            /* current row of input image */
	    runCount,       /* number of runs found in current row */
	    runCountE,      /* run count after erosion */
	    runBufAlloced = cFalse,
	    inPlace;        /* TRUE if destImage == srcImage */
    register Int32 length;         /* length of current run */
    register Int32 cRunWord;
    register Int32 lRunWord;
    register Int32 i;
    register Int32 wt;
    register Int32 wc;
    register Int32 wm;
    register Int32 w0;
    register Int32 w1;
    register Int32  left;
    register Int32  right;    /* left and right offsets of current run */
    register UInt32 outWord;        /* output word being constructed */
    Run32  *runBuffer,      /* array of runs to pass to run finder */
           *dummyRun,
	   *outRun;
    register UInt32 *rowPtr;


    if (runBuf == NULL)
    {
        runBufferSize = (((width - (2 * cFrameBits)) / 2) + 3);
        runBuffer = (Run32 *)MALLOC (runBufferSize * sizeof (Run32));
	if (runBuffer == NULL)
            return (ia_nomem);
	else
	    runBufAlloced = cTrue;
    }
    else
    {
        runBuffer = (Run32 *)runBuf;
	/* this is passed in as size in bytes */
	runBufferSize /= sizeof (Run32);
    }

    if (srcImage == destImage)
        inPlace = cTrue;
    else
        inPlace = cFalse;

    srcImage += (srcBpl << cLogFrameBits) + cFrameBytes;
    destImage += (destBpl << cLogFrameBits) + cFrameBytes;

    width -= 2 * cFrameBits;
    height -= 2 * cFrameBits;

    /* Get the left and right sub-width parameters for the "strel" */
    wt = strelWidth - 1;
    w0 = strelCenter;
    w1 = wt - strelCenter;
    wm = width - 1;


    if (inPlace)
    {
        if (borderValue == 1)
	{
	    /* filter runs one row at a whack */
	    for (row = 0; row < height; row++)
	    {
		/* get all the runs in the current row */
		__error__ = i_findAllRunsInLineArray (pIPState, srcImage,
						    width,
						    0,
						    runBuffer,
						    runBufferSize,
						    0,
						    &dummyRun);
	
		if (__error__ != ia_successful)
		    break;
	
		runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
				   sizeof (Run32));
	
		if (runCount > 0)
		{
		    /* handle the edge cases */
		    if (runBuffer[0].runLeft == 0)
			runBuffer[0].runLeft -= strelWidth;

		    if (runBuffer[runCount - 1].runRight == wm)
			runBuffer[runCount - 1].runRight += strelWidth;

		    outWord = 0xffffffff;
		    lRunWord = 0;
    
		    /* go thru the run array for the current row */
		    for (i = 0; i < runCount; i++)
		    {
			/* check to see if the run is completely removed */
			if (((runBuffer[i].runRight - runBuffer[i].runLeft) <
			      wt) )
			{
			    left = MAX (0, runBuffer[i].runLeft);
			    right = MIN (wm, runBuffer[i].runRight);
    
			    /* erase the run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;
    
			    /* if we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
				/* write the output word then clear it */
				ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) &=
							   outWord;
				outWord = 0xffffffff;
				lRunWord = cRunWord;
			    }
    
			    /* is this a single or multiple word run? */
			    if (wc == 1)
			    {
				outWord &= ~((endRunPosLUT[(left % 32)]) &
					     (endRunNegLUT[(right % 32)]));
			    }
			    else
			    {
				rowPtr = (UInt32 *)srcImage + cRunWord;
				ULONG_ACCESS (rowPtr) &= (outWord &
					~(endRunPosLUT[(left % 32)]));
				wc--;
				rowPtr++;
				outWord = 0xffffffff;
    
				/* write all of the empty middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0x00000000;
				    rowPtr++;
				}
    
				/* start the next output word by erasing
				 * the tail end of this run.
				 */
				outWord &= ~(endRunNegLUT[(right % 32)]);
    
				lRunWord = right / 32;
			    }
			}
			else
			{
			    /* check for erasure at the left end of the run */
			    if ((w0 > 0) && (runBuffer[i].runLeft >= 0))
			    {
				left = MAX (0, runBuffer[i].runLeft);
				right = MIN (wm, runBuffer[i].runLeft + w0 - 1);

				/* erase the run in the output image */
				cRunWord = left / 32;
				wc = (right / 32) - cRunWord + 1;
	
				/* if we moved to a new word, 
				 * write the old one.
				 */
				if (cRunWord != lRunWord)
				{
				    /* write the output word then clear it */
				    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord)
				                  &= outWord;
				    outWord = 0xffffffff;
				    lRunWord = cRunWord;
				}
	
				/* is this a single or multiple word run? */
				if (wc == 1)
				{
				    outWord &= ~((endRunPosLUT[(left % 32)]) &
						 (endRunNegLUT[(right % 32)]));
				}
				else
				{
				    rowPtr = (UInt32 *)srcImage + cRunWord;
				    ULONG_ACCESS (rowPtr) &= (outWord &
					    ~(endRunPosLUT[(left % 32)]));
				    wc--;
				    rowPtr++;
				    outWord = 0xffffffff;
	
				    /* write all of the empty middle words */
				    for (; wc > 1; wc--)
				    {
					ULONG_ACCESS (rowPtr) = 0x00000000;
					rowPtr++;
				    }
	
				    /* start the next output word by erasing
				     * the tail end of this run.
				     */
				    outWord &= ~(endRunNegLUT[(right % 32)]);
	
				    lRunWord = right / 32;
				}
			    }

			    /* check for erasure at the right end of the run */
			    if ((w1 > 0) && (runBuffer[i].runRight <= wm))
			    {
				left = MAX (0, runBuffer[i].runRight - w1 + 1);
				right = MIN (wm, runBuffer[i].runRight);

				/* erase the run in the output image */
				cRunWord = left / 32;
				wc = (right / 32) - cRunWord + 1;
	
				/* if we moved to a new word, 
				* write the old one.
				*/
				if (cRunWord != lRunWord)
				{
				    /* write the output word then clear it */
				    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord)
				                  &= outWord;
				    outWord = 0xffffffff;
				    lRunWord = cRunWord;
				}
	
				/* is this a single or multiple word run? */
				if (wc == 1)
				{
				    outWord &= ~((endRunPosLUT[(left % 32)]) &
						 (endRunNegLUT[(right % 32)]));
				}
				else
				{
				    rowPtr = (UInt32 *)srcImage + cRunWord;
				    ULONG_ACCESS (rowPtr) &= (outWord &
					    ~(endRunPosLUT[(left % 32)]));
				    wc--;
				    rowPtr++;
				    outWord = 0xffffffff;
	
				    /* write all of the empty middle words */
				    for (; wc > 1; wc--)
				    {
					ULONG_ACCESS (rowPtr) = 0x00000000;
					rowPtr++;
				    }
	
				    /* start the next output word by erasing
				     * the tail end of this run.
				     */
				    outWord &= ~(endRunNegLUT[(right % 32)]);
	
				    lRunWord = right / 32;
				}
			    }
			}
		    }
    
		    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) &= outWord;
	
		}  /* if the row had any runs to erode */
    
		srcImage += srcBpl;
    
	    }  /* for each image row */
	}
	else  /* white border case */
	{
	    /* filter runs one row at a whack */
	    for (row = 0; row < height; row++)
	    {
		/* get all the runs in the current row */
		__error__ = i_findAllRunsInLineArray (pIPState, srcImage,
						    width,
						    0,
						    runBuffer,
						    runBufferSize,
						    0,
						    &dummyRun);
	
		if (__error__ != ia_successful)
		    break;
	
		runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
				   sizeof (Run32));
	
		if (runCount > 0)
		{
		    outWord = 0xffffffff;
		    lRunWord = 0;
    
		    /* go thru the run array for the current row */
		    for (i = 0; i < runCount; i++)
		    {
			/* check to see if the run is completely removed */
			if (((runBuffer[i].runRight - runBuffer[i].runLeft) <
			      wt) )
			{
			    left = runBuffer[i].runLeft;
			    right = runBuffer[i].runRight;
    
			    /* erase the run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;
    
			    /* if we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
				/* write the output word then clear it */
				ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) &=
							   outWord;
				outWord = 0xffffffff;
				lRunWord = cRunWord;
			    }
    
			    /* is this a single or multiple word run? */
			    if (wc == 1)
			    {
				outWord &= ~((endRunPosLUT[(left % 32)]) &
					     (endRunNegLUT[(right % 32)]));
			    }
			    else
			    {
				rowPtr = (UInt32 *)srcImage + cRunWord;
				ULONG_ACCESS (rowPtr) &= (outWord &
					~(endRunPosLUT[(left % 32)]));
				wc--;
				rowPtr++;
				outWord = 0xffffffff;
    
				/* write all of the empty middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0x00000000;
				    rowPtr++;
				}
    
				/* start the next output word by erasing
				 * the tail end of this run.
				 */
				outWord &= ~(endRunNegLUT[(right % 32)]);
    
				lRunWord = right / 32;
			    }
			}
			else
			{
			    if (w0 > 1)
			    {
				left = runBuffer[i].runLeft;
				right = runBuffer[i].runLeft + w0 - 1;
    
				/* erase the run in the output image */
				cRunWord = left / 32;
				wc = (right / 32) - cRunWord + 1;
	
				/* if we moved to a new word, 
				 * write the old one.
				 */
				if (cRunWord != lRunWord)
				{
				    /* write the output word then clear it */
				    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord)
				               &= outWord;
				    outWord = 0xffffffff;
				    lRunWord = cRunWord;
				}
	
				/* is this a single or multiple word run? */
				if (wc == 1)
				{
				    outWord &= ~((endRunPosLUT[(left % 32)]) &
						 (endRunNegLUT[(right % 32)]));
				}
				else
				{
				    rowPtr = (UInt32 *)srcImage + cRunWord;
				    ULONG_ACCESS (rowPtr) &= (outWord &
					    ~(endRunPosLUT[(left % 32)]));
				    wc--;
				    rowPtr++;
				    outWord = 0xffffffff;
	
				    /* write all of the empty middle words */
				    for (; wc > 1; wc--)
				    {
					ULONG_ACCESS (rowPtr) = 0x00000000;
					rowPtr++;
				    }
	
				    /* start the next output word by erasing
				     * the tail end of this run.
				     */
				    outWord &= ~(endRunNegLUT[(right % 32)]);
	
				    lRunWord = right / 32;
				}
			    }
	
			    if (w1 > 1)
			    {
				left = runBuffer[i].runRight - w1 + 1;
				right = runBuffer[i].runRight;
				/* erase the run in the output image */
				cRunWord = left / 32;
				wc = (right / 32) - cRunWord + 1;
	
				/* if we moved to a new word, 
				 * write the old one.
				 */
				if (cRunWord != lRunWord)
				{
				    /* write the output word then clear it */
				    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord)
				                  &= outWord;
				    outWord = 0xffffffff;
				    lRunWord = cRunWord;
				}
	
				/* is this a single or multiple word run? */
				if (wc == 1)
				{
				    outWord &= ~((endRunPosLUT[(left % 32)]) &
						 (endRunNegLUT[(right % 32)]));
				}
				else
				{
				    rowPtr = (UInt32 *)srcImage + cRunWord;
				    ULONG_ACCESS (rowPtr) &= (outWord &
					    ~(endRunPosLUT[(left % 32)]));
				    wc--;
				    rowPtr++;
				    outWord = 0xffffffff;
	
				    /* write all of the empty middle words */
				    for (; wc > 1; wc--)
				    {
					ULONG_ACCESS (rowPtr) = 0x00000000;
					rowPtr++;
				    }
	
				    /* start the next output word by erasing
				     * the tail end of this run.
				     */
				    outWord &= ~(endRunNegLUT[(right % 32)]);
	
				    lRunWord = right / 32;
				}
			    }
			}
		    }
    
		    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) &= outWord;
	
		}  /* if the row had any runs to erode */
    
		srcImage += srcBpl;
    
	    }  /* for each image row */

	}  /* else border was white */

    }  /* in-place */
    else
    {
        /* In-to-out case */

	/* filter runs one row at a whack */
	for (row = 0; row < height; row++)
	{
	    /* get all the runs in the current row */
	    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
						width,
						0,
						runBuffer,
						runBufferSize,
						0,
						&dummyRun);
    
	    if (__error__ != ia_successful)
		break;
    
	    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
			       sizeof (Run32));
    
	    if (runCount > 0)
	    {
	        if (borderValue == 1)
		{
		    /* handle the edge cases */
		    if (runBuffer[0].runLeft == 0)
			runBuffer[0].runLeft -= w0;
    
		    if (runBuffer[runCount - 1].runRight == wm)
			runBuffer[runCount - 1].runRight += w1;
	        }

		runCountE = runCount;

		/* erode each run removing too short runs */
		outRun = runBuffer;
		for  (i = 0; i < runCount; i++)
		{
		    left = runBuffer[i].runLeft + w0;
		    right = runBuffer[i].runRight - w1;
                    if (left > right)
		    {
		        /* run is completely eroded away */
			runCountE--;
		    }
		    else
		    {
		        outRun->runLeft = MIN (wm, left);
			outRun->runRight = MAX (0, right);
			outRun++;
		    }
		}


		/* write out the new runs */
		outWord = 0x00000000;
		lRunWord = 0;
    
		/* go thru the modified run array for the current row */
		for (i = 0; i < runCountE; i++)
		{
		    left = runBuffer[i].runLeft;
		    right = runBuffer[i].runRight;
		    length = right - left + 1;
    
		    /* construct the run in the output image */
		    cRunWord = left / 32;
		    wc = (right / 32) - cRunWord + 1;
    
		    /* if we moved to a new word, write the old one */
		    if (cRunWord != lRunWord)
		    {
			/* write the output word then clear it */
			ULONG_ACCESS ((UInt32 *)destImage + lRunWord) |=
						   outWord;
			outWord = 0x00000000;
			lRunWord = cRunWord;
		    }
    
		    /* is this a single or multiple word run? */
		    if (wc == 1)
		    {
			outWord |= (endRunPosLUT[(left % 32)]) &
				   (endRunNegLUT[(right % 32)]);
		    }
		    else
		    {
			rowPtr = (UInt32 *)destImage + cRunWord;
			ULONG_ACCESS (rowPtr) |= (outWord |
				endRunPosLUT[(left % 32)]);
			wc--;
			rowPtr++;
			outWord = 0x00000000;
    
			/* write all of the full middle words */
			for (; wc > 1; wc--)
			{
			    ULONG_ACCESS (rowPtr) = 0xffffffff;
			    rowPtr++;
			}
    
			/* start the next output word with the tail
			 * end of this run.
			 */
			outWord |= (endRunNegLUT[(right % 32)]);
    
			lRunWord = right / 32;
		    }
    
		}  /* for each eroded run */
    
		/* Be sure to write out the last word we worked on! */
		ULONG_ACCESS ((UInt32 *)destImage + lRunWord) |= outWord;
    
	    }  /* if the row had any runs to erode */
    
	    srcImage += srcBpl;
	    destImage += destBpl;
    
	}  /* for each image row */
    }

    /* if we allocated this locally, free it */
    if (runBufAlloced)
        FREE (runBuffer);

    NORMAL_RETURN;

}  /* i_erodeHorizontalRuns () */


/**************************************************************************
 * i_erodeVerticalRuns ()
 *
 * Erode vertical black runs. 
 *
 * Arguments:
 *	runBuf :	Pointer to run buffer of the appropriate size
 *			as determined by i_getRunErodeBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	runBufferSize :	If the run buffer is pre-allocated (preferred),
 *			the size of the buffer (in bytes).  If a NULL
 *			pointer is passed for the run buffer this value
 *			is ignored.
 *	rotBuf :	Pointer to rotation buffer of the appropriate
 *			size as determined by i_getRunErodeBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	rotBufBpl :	Bytes per line of the rotation buffer as
 *			determined by i_getRunErodeBufferSizes ().  If
 *			a NULL pointer is passed for the rotation buffer
 *			this value is ignored.
 *	srcImage :	Pointer to source image data including frame.
 *	width :		Width of source and dest images (MUST be the same).
 *	height :	Height of source and dest images (MUST be the same).
 *	srcBpl :	Bpl of source image.
 *	destBpl :	Bpl of dest image.
 *      borderValue :   "Color" of border beyond the image (0 = white,
 *                      1 = black).  The border is assumed to be infinite
 *                      for run length calculations (e.g. a black run which
 *			abuts a black border is assumed to be infinitely
 *			long and is eroded appropriately).
 *	strelHeight :	Height of erosion "strel."  The strel is a
 *			solid horizontal column of hits.
 *	strelCenter :	Center location of "strel."  0 -> top end
 *			and strelWidth - 1 -> bottom end.
 *	destImage :	Pointer to dest image data including frame.
 *
 * Returns:
 *	ia_successful	if erode operation completed normally,
 *	error codes	if bad things happened during erosion.
 *
 * Notes:
 *
 *	Vertical run erosion is necessarily slower than
 *	its horizontal counterpart since it performs two
 *	rotations of each column (actually it works on
 *	blocks of columns) in addition to the run finding/
 *	erosion operations.
 *
 *	If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be replaced with the results
 *	produced by this function.  This differs from
 *	i_erodeHorizontalRuns ().  To combine results from
 *	a vertical and a horizontal run erode, perform the
 *	the vertical erode first, then the horizontal using 
 *	the same destination image for both operations.
 *
 **************************************************************************/
Int32 CDECL 
i_erodeVerticalRuns (
    IPState    *pIPState,
    UInt8      *runBuf,
    Int32	runBufferSize,
    UInt8      *rotBuf,
    Int32	rotBufBpl,
    UInt8      *srcImage,
    Int32	width,
    Int32	height,
    Int32	srcBpl,
    Int32	destBpl,
    Int32	borderValue,
    Int32	strelHeight,
    Int32	strelCenter,
    UInt8      *destImage)
{
    PROLOGUE;
    Int32   col,            /* current column of input image */
            wu,             /* width unframed */
	    lastSlice,      /* width of last vertical slice processed */
	    runBufAlloced = cFalse,
	    rotBufAlloced = cFalse;
    UInt8  *oDestImage,     /* Original dest image pointer */
           *tmpImage;       /* tmp image data pointer */


    __error__ = ia_successful;

    if (runBuf == NULL)
    {
        runBufferSize = (((height - (2 * cFrameBits)) / 2) + 3);
        runBuf = (UInt8 *)MALLOC (runBufferSize * sizeof (Run32));
	if (runBuf == NULL)
            return (ia_nomem);
	else
	    runBufAlloced = cTrue;
    }



    /* save the original dest image pointer for later frame clearing */
    oDestImage = destImage;

    if (rotBuf == NULL)
    {
        rotBufBpl = ((((height + 31) / 32) * 4));
        tmpImage = (UInt8 *)CALLOC (1, rotBufBpl * 
	                           (COL_BLOCK_SIZE + (2 * cFrameBits)));
	if (tmpImage == NULL)
	{
	    if (runBufAlloced)
	        FREE (runBuf);

	    return (ia_nomem);
	}
	else
	    rotBufAlloced = cTrue;
    }
    else
	tmpImage = rotBuf;


    /* figure out the width of the last slice */
    lastSlice = width % COL_BLOCK_SIZE;
    if (lastSlice == 0)  /* it must be a full slice */
        lastSlice = COL_BLOCK_SIZE;

    wu = width - (2 * cFrameBits);

    /* process the image in 32 column vertical bands */
    for (col = 0; col < wu; col += COL_BLOCK_SIZE)
    {
	/* rotate COL_BLOCK_SIZE columns of the original image 90 into 
	 * the temp image.
	 */
	__error__ = i_fastRot90 (pIPState, tmpImage, srcImage, 
	                       (COL_BLOCK_SIZE + (2 * cFrameBits)), 
	                       height, 
			       srcBpl, 
			       rotBufBpl, 1);

        if (__error__ != ia_successful)
	    break;

	/* erode horizontal runs in place in the temp image */
        __error__ = i_erodeHorizontalRuns (
			pIPState, runBuf, runBufferSize,
			tmpImage,
			height,
			(((col + COL_BLOCK_SIZE) < wu) ?
				(COL_BLOCK_SIZE + (2 * cFrameBits)) :
				(lastSlice + (2 * cFrameBits))),
			rotBufBpl, rotBufBpl, 
			borderValue,
			strelHeight,
			strelHeight - strelCenter - 1,
			tmpImage);

        if (__error__ != ia_successful)
	    break;

	/* rotate the temp image 270 degrees into either the output image
	 * or back into the input image if in place.
	 */
	__error__ = i_fastRot270 (pIPState, destImage, tmpImage, 
	                        height, 
		                COL_BLOCK_SIZE + (2 * cFrameBits),
			        rotBufBpl, destBpl, 1);

        if (__error__ != ia_successful)
	    break;

        /* increment the pointers to the next column block */
        srcImage += COL_BLOCK_SIZE / 8;
	destImage += COL_BLOCK_SIZE / 8;

    }  /* for each column block */

    /* Clear the right frame if necessary since the rotate may have
     * written trash there on the last iteration.
     */
    if ((lastSlice != 0) && (__error__ == ia_successful))
    {
        __error__ = i_rasterOpFull (pIPState, oDestImage, 
	                          width,
	                          height,
				  0,
				  0,
				  width - cFrameBits,
				  0,
				  cFrameBits,
				  height,
				  1,
				  destBpl,
				  PIX_CLR,
				  NULL, 0, 0, 0, 0, 0, 0, 0, 0);
    }

    /* if we allocated this locally, free it */
    if (runBufAlloced)
        FREE (runBuf);
    if (rotBufAlloced)
        FREE (tmpImage);

    NORMAL_RETURN;

}  /* i_erodeVerticalRuns () */


/*
 * $Log: i_rnmrph.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 * Imported sources
 *
 * Revision 1.5  1998/01/21 21:42:11  ddavies
 * i_rnmrph.c -> /product/ipcore/ipshared/src/RCS/i_rnmrph.c,v
 *
 * Fix Doug's weird syntax for FREE - don't decrement the thing freed in the
 * macro invocation.  Especially, don't decrement the pointer AFTER the thing
 * it's pointing to is free.
 *
 * Revision 1.4  1998/01/13 23:49:31  ddavies
 * i_rnmrph.c -> /product/ipcore/ipshared/src/RCS/i_rnmrph.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.3  1996/05/29 20:13:56  dferrell
 * dferrell on Wed May 29 10:20:58 PDT 1996
 *
 * i_rnmrph.c -> /product/ipcore/ipshared/src/RCS/i_rnmrph.c,v
 *
 * Fixed bugs related to frames and framed width/height which were non-fatal.
 *
 * Revision 1.2  1996/02/26 20:06:21  dferrell
 * i_rnmrph.c -> /product/ipcore/ipshared/src/RCS/i_rnmrph.c,v
 *
 * Fixed bug in vertical erosion and dilation whereby the last 32 columns
 * were not processed for images with widths of exact multiples of 32.
 *
 * Revision 1.1  1996/02/20 23:52:47  dferrell
 * Initial revision
 *
 *
 */

