/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_runflt.c--includes
 *
 *	i_initRunFilter ()
 *	i_finishRunFilter ()
 *	i_filterHorizontalRuns ()	ipshared interface level code
 *					for horizontal run filtering.
 *
 *	i_filterVerticalRuns ()		ipshared interface level code
 *					for vertical run filtering.
 *
 */

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"
#include "shrip.pub"
#include "shrcc.pub"
#include "shrrast.pub"   /* for i_rasterOpFull () */
#include "runflt.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_runflt.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");


/* Often in document image processing using morphology we use "linear"
 * strels (1 x n or n x 1) for speed.  Unfortunately, above a certain
 * size (around n = 32) speed is greatly compromised.  For a typical
 * document image of 300 or 400 dpi resolution a 32 pixel strel is
 * very small in comparison to the features being processed.  This is
 * particularly apparent when the features to be matched or fitlered
 * are long vertical and/or horizontal lines (e.g. rules, computer
 * generated line art, etc.).  In these cases, rather than using
 * morphological operations alone, we would like to filter pixels by
 * their membership in runs of a certain length.
 *
 * For example, we might collect vertical and horizontal run length
 * histograms for a document.  By analyzing these we could detect the
 * probable presence of rules which we might want to remove.  In order
 * to guarantee removal of only rules using morphology we would want
 * to erode with very large strels (100s of pixels).  Using smaller
 * strels risks removing other important document features (like human
 * drawn gestures).  In this case we could employ filtering by run
 * length to remove only the very long runs from the image.  Or we
 * could extract only the long runs into another image for further
 * processing using morphology and then use this image in a rasterOp
 * operation with the original.  
 *
 * Basically, these functions provide an alternative to morphology for
 * filtering large, linear features.  In fact, when called with the
 * proper borderValue/runValue combinations, these functions perform
 * morphological opens and closes with linear, centered, strels.
 *
 */

/* some pretty big numbers used to make edge abutting runs
 * pseudo-infinite.  Since the run end points are signed
 * 32-bit integers, these should really be bigger...
 */
#define cBigLeftEnd  -65536L  
#define cBigRightEnd  65536L


/* PUBLIC */
/**************************************************************************
 * i_getRunFilterBufferSizes
 *
 * Given image width, height, and filter type calculate the run buffer
 * and rotation buffer sizes necessary to run the filter(s) by calling
 * i_filterHorizontalRuns () and/or i_filterVerticalRuns ().
 *
 * Arguments:
 *	w:	Width of the image which will be run filtered 
 *		(including frame)
 *	h:	Height of the image which will be run filtered
 *		(including frame)
 *	filters: What type of filtering will be done to the
 *		image.  Can be one of cVerticalRuns, cHorizontalRuns,
 *		or cBothRuns
 *
 * Output:
 *	runBufferSize: Size (in bytes) required for run buffer
 *	rotBufferSize: Size (in bytes) required for rotation buffer
 *		(this may be zero)
 *	rotBufferBpl:  Bytes per line parameter for the rotation
 *		buffer (this may also be zero)
 *
 * Returns:
 *	ia_successful	if filter operation completed normally,
 *	error codes	if invalid parameters
 *
 * Notes:
 *	Call this function to determine how large the buffers need to
 * 	be for run filtering.  Allocate the buffers and pass them to
 *	i_filterHorizontalRuns () and/or i_filterVerticalRuns ().  If
 *	multiple filter operations are to be performed on the same
 *	image SIZE, these buffers may re-used for each operation.
 *
 **************************************************************************/
Int32 CDECL 
i_getRunFilterBufferSizes (
    IPState    *pIPState,
    Int32	w,
    Int32	h,
    Int32	filters,
    Int32      *runBufferSize,
    Int32      *rotBufferSize,
    Int32      *rotBufferBpl)
{

    /* The max number of black runs on a row is (w / 2) + 1 and on a column
     * is (h / 2) + 1.  We allocate an extra run at the beginning and end 
     * for use by the white run filter, hence the + 3 term.
     */
    PROLOGUE;
    switch (filters)
    {
        case cVerticalRuns:
            *rotBufferBpl = (((h + 31) / 32) * 4);

            /* the rotation temp image buffer which is COL_BLOCK_SIZE x h 
	     * bytes with a frame of 32 pixels.
             */
	    *rotBufferSize = *rotBufferBpl * 
	                     (COL_BLOCK_SIZE + (2 * cFrameBits));
	    h -= 2 * cFrameBits;
	    *runBufferSize = ((h / 2) + 3) * sizeof (Run32);
	    break;

	case cHorizontalRuns:
	    /* Don't need a rotation buffer */
	    *rotBufferSize = 0;
	    *rotBufferBpl = 0;
	    w -= 2 * cFrameBits;
	    *runBufferSize = ((w / 2) + 3) * sizeof (Run32);
	    break;

	case cBothRuns:
            *rotBufferBpl = (((h + 31) / 32) * 4);

            /* the rotation temp image buffer which is COL_BLOCK_SIZE x h 
	     * bytes with a frame of 32 pixels.
             */
	    *rotBufferSize = *rotBufferBpl * 
	                     (COL_BLOCK_SIZE + (2 * cFrameBits));
	    w -= 2 * cFrameBits;
	    h -= 2 * cFrameBits;
	    *runBufferSize = ((MAX (w, h) / 2) + 3) * sizeof (Run32);
	    break;

	default:
	    return (ia_invalidParm);
	    break;
    }

    NORMAL_RETURN;

}  /* i_getRunFilterBufferSizes () */


/**************************************************************************
 * i_filterHorizontalRuns ()
 *
 * Filters horizontal black or white runs by their length.  Provides
 * the canonical set of filter operations: HP, LP, BP, BR.
 *
 * Arguments:
 *	runBuf :	Pointer to run buffer of the appropriate size
 *			as determined by i_getRunFilterBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	runBufferSize :	If the run buffer is pre-allocated (preferred),
 *			the size of the buffer (in bytes).  If a NULL
 *			pointer is passed for the run buffer this value
 *			is ignored.
 *	srcImage :	Pointer to source image data including frame.
 *	width :		Width of source and dest images (MUST be the same).
 *	height :	Height of source and dest images (MUST be the same).
 *	srcBpl :	Bpl of source image.
 *	destBpl :	Bpl of dest image.
 *      runValue :	"Color" of runs to filter (0 = white, 1 = black)
 *      borderValue :   "Color" of border beyond the image (0 = white,
 *                      1 = black).  The border is assumed to be infinite
 *                      for run length calculations (e.g. a black run which
 *			abuts a black border is assumed to be infinitely
 *			long and is filtered appropriately).
 *	minRunLength :	Minimum run length in band (inclusive).
 *	maxRunLength :  Maximum run length in band (inclusive).
 *	polarity :	Filter polarity can be
 *			cPassRuns - Runs which are within the band
 *			are kept, others are discarded, or
 *			cRejectRuns - Runs which are within the band
 *			are discarded, others are kept.
 *	destImage :	Pointer to dest image data including frame.
 *
 * Returns:
 *	ia_successful	if filter operation completed normally,
 *	error codes	if bad things happened during filtering.
 *
 * Notes:
 *	This function is capable of performing all the basic
 *	filter functions: highpass, lowpass, bandpass, band
 *	reject (or notch) based on horizontal run length.
 *
 *	If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be combined with the results
 *	produced by this function (black runs are ORed into the
 *	destination, white runs are ANDed).  Be sure to set the
 *	output image (to white for black runs or to black for
 *	white runs) before calling i_filterHorizontalRuns if you 
 *	only want the result of a single filter operation.
 *
 **************************************************************************/
Int32 CDECL 
i_filterHorizontalRuns (
    IPState    *pIPState,
    UInt8      *runBuf,
    Int32	runBufferSize,
    UInt8      *srcImage,
    Int32	width,
    Int32	height,
    Int32	srcBpl,
    Int32	destBpl,
    Int32	runValue,
    Int32	borderValue,
    Int32	minRunLength,
    Int32	maxRunLength,
    Int32	polarity,
    UInt8      *destImage)
{
    PROLOGUE;
    Int32   row,            /* current row of input image */
	    runCount,       /* number of runs found in current row */
	    runBufAlloced = cFalse,
	    inPlace;        /* cTrue if destImage == srcImage */
    register Int32 length;         /* length of current run */
    register Int32 cRunWord;
    register Int32 lRunWord;
    register Int32 i;
    register Int32 wc;
    register Int32  left;
    register Int32  right;    /* left and right offsets of current run */
    register UInt32 outWord;        /* output word being constructed */
    Run32  *runBuffer,      /* array of runs to pass to run finder */
           *dummyRun;
    register UInt32 *rowPtr;


    if (runBuf == NULL)
    {
        runBufferSize = (((width - (2 * cFrameBits)) / 2) + 3);
        runBuffer = (Run32 *)MALLOC (runBufferSize * sizeof (Run32));
	if (runBuffer == NULL)
            return (ia_nomem);
	else
	    runBufAlloced = cTrue;
    }
    else
    {
        runBuffer = (Run32 *)runBuf;
	/* this is passed in as size in bytes */
	runBufferSize /= sizeof (Run32);
    }

    /* Remember that we want to keep an extra run on either end
     * for white run filtering, so offset the start by 1 and
     * only let the run finder use the middle part.
     */
    runBuffer++;

    if (srcImage == destImage)
        inPlace = cTrue;
    else
        inPlace = cFalse;

    srcImage += (srcBpl << cLogFrameBits) + cFrameBytes;
    destImage += (destBpl << cLogFrameBits) + cFrameBytes;

    width -= 2 * cFrameBits;
    height -= 2 * cFrameBits;

    /* There are four cases for the inner loop:
     *
     * - BP filter in place, {white runs | black runs} & {white border | black border}
     * - BR filter in place,
     * - BP filter to output pixr, and
     * - BR filter to output pixr.
     *
     * Each has been coded seperately to avoid excessive branching
     * inside the inner loop.
     */

    if (runValue == borderValue)  /* border abutting runs are "infinite" */
    {
        if (runValue == 0)  /* filtering white runs */
	{
	    Run32 *whiteRunBuffer;
	    Int32  whiteRunCount;

	    if (inPlace)
	    {
	        /* filter white runs in->in (in-place) */

		/* filter runs one row at a whack */
		for (row = 0; row < height; row++)
		{
		    /* get all the runs in the current row */
		    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
		                                        width,
							0,
							runBuffer,
							runBufferSize,
							0,  
							&dummyRun);

                    if (__error__ != ia_successful)
		        break;

		    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
		                       sizeof (Run32));
                    whiteRunCount = runCount + 1;

		    if (whiteRunCount == 1)
		    {
			/* all white row special case */
		        whiteRunBuffer = runBuffer;

			/* check for infinite edge run case */
			if (borderValue == 0)
			{
			    whiteRunBuffer[0].runLeft = cBigLeftEnd;
			    whiteRunBuffer[0].runRight = cBigRightEnd;
			}
			else
			{
			    whiteRunBuffer[0].runLeft = -1;
			    whiteRunBuffer[0].runRight = width;
			}
		    }
		    else  /* multiple white runs */
		    {
			/* We want to shift the run array left or right one
			 * half place to make it an array of white runs. 
			 * The endpoints of the white runs will be 1 off.
			 */
			if (runBuffer[0].runLeft == 0)
			{
			    /* There's a black run at the left edge so
			     * shift the array right.
			     */
			    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer + 1);
			    whiteRunCount--;
			}
			else
			{
			    /* There's a white run at the left edge so
			     * shift the array left and add a dummy start
			     * value for the first white run.
			     */
			    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer - 1);
    
			    /* put the dummy value in for the left edge of the
			     * white run.
			     */
			    /* check for infinite edge run case */
			    if (borderValue == 0)
				whiteRunBuffer[0].runLeft = cBigLeftEnd;
			    else
				whiteRunBuffer[0].runLeft = -1;
			}
    
			/* Now we need to fix up the right edge */
			if (runBuffer[runCount - 1].runRight == width - 1)
			{
			    /* There's a black run at the right edge so
			     * the end of the white run array is shifted
			     * left.
			     */
			    whiteRunCount--;
			}
			else
			{
			    /* There's a white run at the right edge
			     * so add its dummy right end point.
			     */
			    /* check for infinite edge run case */
			    if (borderValue == 0)
				whiteRunBuffer[whiteRunCount - 1].runRight = 
				                         cBigRightEnd;
			    else
				whiteRunBuffer[whiteRunCount - 1].runRight = 
				                         width;
			}
		    }

		    /* We'll be erasing white runs which means writing
		     * black so start off with a white word.
		     */
		    outWord = 0x00000000;
		    lRunWord = 0;

		    /* go thru the run array for the current row */
		    for (i = 0; i < whiteRunCount; i++)
		    {
		        /* Since this was originally an array of black
			 * runs, the ends of the white runs are one off.
			 */
		        left = whiteRunBuffer[i].runLeft + 1;
			right = whiteRunBuffer[i].runRight - 1;
			length = right - left + 1;

                        /* does the run NOT pass the filter? */
			if (((polarity == cPassRuns) && 
			    ((length < minRunLength) || 
			     (length > maxRunLength))) ||
			    ((polarity == cRejectRuns) &&
			    ((length >= minRunLength) && 
			     (length <= maxRunLength))))
			{
			    /* fix the left and right */
			    left = MAX (0, left);
			    right = MIN (width - 1, right);

			    /* Erase the run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;

                            /* If we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
			        /* write the output word then clear it */
                                ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) |=
				                           outWord;
				outWord = 0x00000000;
				lRunWord = cRunWord;
                            }

                            /* Is this a single or multiple word run? */
                            if (wc == 1)
			    {
			        outWord |= (endRunPosLUT[(left % 32)]) &
	                                   (endRunNegLUT[(right % 32)]);
			    }
			    else
			    {
                                rowPtr = (UInt32 *)srcImage + cRunWord;
				ULONG_ACCESS (rowPtr) |= (outWord |
				        endRunPosLUT[(left % 32)]);
				outWord = 0x00000000;
				wc--;
				rowPtr++;

                                /* Write all of the black middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0xffffffff;
	                            rowPtr++;
				}

                                /* Start the next output word by erasing
				 * the tail end of this run
				 */
				outWord |= endRunNegLUT[(right % 32)];

				lRunWord = right / 32;
			    }

			}  /* erase the run? */

                    }  /* for each white run */

                    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) |= outWord;

		    srcImage += srcBpl;

                }  /* for each row */
            }
	    else
	    {
	        /* filter white runs in->out */

		/* filter runs one row at a whack */
		for (row = 0; row < height; row++)
		{
		    /* get all the runs in the current row */
		    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
		                                        width,
							0,
							runBuffer,
							runBufferSize,
							0,
							&dummyRun);

                    if (__error__ != ia_successful)
		        break;

		    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
		                       sizeof (Run32));
                    whiteRunCount = runCount + 1;

		    if (whiteRunCount == 1)
		    {
			/* all white row special case */
		        whiteRunBuffer = runBuffer;

			/* check for infinite edge run case */
			if (borderValue == 0)
			{
			    whiteRunBuffer[0].runLeft = cBigLeftEnd;
			    whiteRunBuffer[0].runRight = cBigRightEnd;
			}
			else
			{
			    whiteRunBuffer[0].runLeft = -1;
			    whiteRunBuffer[0].runRight = width;
			}
		    }
		    else  /* multiple white runs */
		    {
			/* We want to shift the run array left or right one
			 * half place to make it an array of white runs. 
			 * The endpoints of the white runs will be 1 off.
			 */
			if (runBuffer[0].runLeft == 0)
			{
			    /* There's a black run at the left edge so
			     * shift the array right.
			     */
			    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer + 1);
			    whiteRunCount--;
			}
			else
			{
			    /* There's a white run at the left edge so
			     * shift the array left and add a dummy start
			     * value for the first white run.
			     */
			    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer - 1);
    
			    /* put the dummy value in for the left edge of the
			     * white run.
			     */
			    /* check for infinite edge run case */
			    if (borderValue == 0)
				whiteRunBuffer[0].runLeft = cBigLeftEnd;
			    else
				whiteRunBuffer[0].runLeft = -1;
			}
    
			/* Now we need to fix up the right edge */
			if (runBuffer[runCount - 1].runRight == width - 1)
			{
			    /* There's a black run at the right edge so
			     * the end of the white run array is shifted
			     * left.
			     */
			    whiteRunCount--;
			}
			else
			{
			    /* There's a white run at the right edge
			     * so add its dummy right end point.
			     */
			    /* check for infinite edge run case */
			    if (borderValue == 0)
				whiteRunBuffer[whiteRunCount - 1].runRight = 
				                    cBigRightEnd;
			    else
				whiteRunBuffer[whiteRunCount - 1].runRight = 
				                    width;
			}
		    }

		    /* We'll be writing white runs which means erasing
		     * black so start off with a black word.
		     */
		    outWord = 0xffffffff;
		    lRunWord = 0;

		    /* go thru the run array for the current row */
		    for (i = 0; i < whiteRunCount; i++)
		    {
		        /* Since this was originally an array of black
			 * runs, the ends of the white runs are one off.
			 */
		        left = whiteRunBuffer[i].runLeft + 1;
			right = whiteRunBuffer[i].runRight - 1;
			length = right - left + 1;

                        /* does the run pass the filter? */
			if (((polarity == cPassRuns) && 
			    ((length >= minRunLength) && 
			     (length <= maxRunLength))) ||
			    ((polarity == cRejectRuns) &&
			    ((length < minRunLength) || 
			     (length > maxRunLength))))
			{
			    /* fix the left and right */
			    left = MAX (0, left);
			    right = MIN (width - 1, right);

			    /* Write the white run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;

                            /* If we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
			        /* write the output word then clear it */
                                ULONG_ACCESS ((UInt32 *)destImage + lRunWord) &=
				                           outWord;
				outWord = 0xffffffff;
				lRunWord = cRunWord;
                            }

                            /* Is this a single or multiple word run? */
                            if (wc == 1)
			    {
			        outWord &= ~((endRunPosLUT[(left % 32)]) &
	                                     (endRunNegLUT[(right % 32)]));
			    }
			    else
			    {
                                rowPtr = (UInt32 *)destImage + cRunWord;
				ULONG_ACCESS (rowPtr) &= (outWord &
				        ~(endRunPosLUT[(left % 32)]));
				outWord = 0xffffffff;
				wc--;
				rowPtr++;

                                /* Write all of the white middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0x00000000;
	                            rowPtr++;
				}

                                /* Start the next output word by erasing
				 * the tail end of this run
				 */
				outWord &= ~(endRunNegLUT[(right % 32)]);

				lRunWord = right / 32;
			    }

			}  /* erase the run? */

                    }  /* for each white run */

                    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)destImage + lRunWord) &= outWord;

		    srcImage += srcBpl;
                    destImage += destBpl;

                }  /* for each row */

	    }  /* filtering white runs in->out */
	}
	else  /* filtering black runs */
	{
	    if (inPlace)
	    {
	        /* filter in->in (in-place) */

		/* filter runs one row at a whack */
		for (row = 0; row < height; row++)
		{
		    /* get all the runs in the current row */
		    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
		                                        width,
							0,
							runBuffer,
							runBufferSize,
							0,
							&dummyRun);

                    if (__error__ != ia_successful)
		        break;

		    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
		                       sizeof (Run32));

                    if (runCount > 0)
		    {
			/* Check for infinite edge run case */
			if (borderValue == 1)
			{
			    /* Is the first run abutting the left edge? */
			    if (runBuffer[0].runLeft == 0)
				runBuffer[0].runLeft = cBigLeftEnd;
    
			    if (runBuffer[runCount - 1].runRight == width - 1)
				runBuffer[runCount - 1].runRight = cBigRightEnd;
			}
		    }

		    outWord = 0xffffffff;
		    lRunWord = 0;

		    /* go thru the run array for the current row */
		    for (i = 0; i < runCount; i++)
		    {
		        left = runBuffer[i].runLeft;
			right = runBuffer[i].runRight;
			length = right - left + 1;

                        /* does the run NOT pass the filter? */
			if (((polarity == cPassRuns) && 
			    ((length < minRunLength) || 
			     (length > maxRunLength))) ||
			    ((polarity == cRejectRuns) &&
			    ((length >= minRunLength) && 
			     (length <= maxRunLength))))
			{
			    /* fix the left and right */
			    left = MAX (0, left);
			    right = MIN (width - 1, right);

			    /* erase the run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;

                            /* if we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
			        /* write the output word then clear it */
                                ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) &=
				                           outWord;
				outWord = 0xffffffff;
				lRunWord = cRunWord;
                            }

                            /* is this a single or multiple word run? */
                            if (wc == 1)
			    {
                                outWord &= ~((endRunPosLUT[(left % 32)]) &
	                                     (endRunNegLUT[(right % 32)]));
			    }
			    else
			    {
                                rowPtr = (UInt32 *)srcImage + cRunWord;
				ULONG_ACCESS (rowPtr) &= (outWord &
				        ~(endRunPosLUT[(left % 32)]));
				wc--;
				rowPtr++;
				outWord = 0xffffffff;

                                /* write all of the empty middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0x00000000;
	                            rowPtr++;
				}

                                /* start the next output word by erasing
				 * the tail end of this run.
				 */
				outWord &= ~(endRunNegLUT[(right % 32)]);

				lRunWord = right / 32;
			    }
			}
		    }

                    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) &= outWord;

		    srcImage += srcBpl;
		}
	    }
	    else
	    {
	        /* filter in->out */

		/* filter runs one row at a whack */
		for (row = 0; row < height; row++)
		{
		    /* get all the runs in the current row */
		    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
		                                        width,
							0,
							runBuffer,
							runBufferSize,
							0,
							&dummyRun);

                    if (__error__ != ia_successful)
		        break;

		    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
		                       sizeof (Run32));

                    if (runCount > 0)
		    {
			/* Check for infinite edge run case */
			if (borderValue == 1)
			{
			    /* Is the first run abutting the left edge? */
			    if (runBuffer[0].runLeft == 0)
				runBuffer[0].runLeft = cBigLeftEnd;
    
			    if (runBuffer[runCount - 1].runRight == width - 1)
				runBuffer[runCount - 1].runRight = cBigRightEnd;
			}
                    }

		    outWord = 0x00000000;
		    lRunWord = 0;

		    /* go thru the run array for the current row */
		    for (i = 0; i < runCount; i++)
		    {
		        left = runBuffer[i].runLeft;
			right = runBuffer[i].runRight;
			length = right - left + 1;

                        /* does the run pass the filter? */
			if (((polarity == cPassRuns) && 
			    ((length >= minRunLength) && 
			     (length <= maxRunLength))) ||
			    ((polarity == cRejectRuns) &&
			    ((length < minRunLength) || 
			     (length > maxRunLength))))
			{
			    /* fix the left and right */
			    left = MAX (0, left);
			    right = MIN (width - 1, right);

			    /* construct the run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;

                            /* if we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
			        /* write the output word then clear it */
                                ULONG_ACCESS ((UInt32 *)destImage + lRunWord) |=
				                           outWord;
				outWord = 0x00000000;
				lRunWord = cRunWord;
                            }

                            /* is this a single or multiple word run? */
                            if (wc == 1)
			    {
			        outWord |= (endRunPosLUT[(left % 32)]) &
	                                   (endRunNegLUT[(right % 32)]);
			    }
			    else
			    {
                                rowPtr = (UInt32 *)destImage + cRunWord;
				ULONG_ACCESS (rowPtr) |= (outWord |
				        endRunPosLUT[(left % 32)]);
				wc--;
				rowPtr++;
				outWord = 0x00000000;

                                /* write all of the full middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0xffffffff;
	                            rowPtr++;
				}

                                /* start the next output word with the tail
				 * end of this run.
				 */
                                outWord |= (endRunNegLUT[(right % 32)]);

				lRunWord = right / 32;
			    }
			}
		    }

                    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)destImage + lRunWord) |= outWord;

		    srcImage += srcBpl;
                    destImage += destBpl;
		}
	    }

	}  /* normal black run filtering */

    }
    else  /* border abutting runs end at the edge (normal case) */
    {
        if (runValue == 0)  /* filtering white runs */
	{
	    Run32 *whiteRunBuffer;
	    Int32  whiteRunCount;

	    if (inPlace)
	    {
	        /* filter white runs in->in (in-place) */

		/* filter runs one row at a whack */
		for (row = 0; row < height; row++)
		{
		    /* get all the runs in the current row */
		    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
		                                        width,
							0,
							runBuffer,
							runBufferSize,
							0,
							&dummyRun);

                    if (__error__ != ia_successful)
		        break;

		    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
		                       sizeof (Run32));
                    whiteRunCount = runCount + 1;

		    if (whiteRunCount == 1)
		    {
			/* all white row special case */
		        whiteRunBuffer = runBuffer;
			whiteRunBuffer[0].runLeft = -1;
			whiteRunBuffer[0].runRight = width;
		    }
		    else  /* multiple white runs */
		    {
			/* We want to shift the run array left or right one
			 * half place to make it an array of white runs. 
			 * The endpoints of the white runs will be 1 off.
			 */
			if (runBuffer[0].runLeft == 0)
			{
			    /* There's a black run at the left edge so
			     * shift the array right.
			     */
			    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer + 1);
			    whiteRunCount--;
			}
			else
			{
			    /* There's a white run at the left edge so
			     * shift the array left and add a dummy start
			     * value for the first white run.
			     */
			    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer - 1);
    
			    /* put the dummy value in for the left edge of the
			     * white run.
			     */
			    whiteRunBuffer[0].runLeft = -1;
			}
    
			/* Now we need to fix up the right edge */
			if (runBuffer[runCount - 1].runRight == width - 1)
			{
			    /* There's a black run at the right edge so
			     * the end of the white run array is shifted
			     * left.
			     */
			    whiteRunCount--;
			}
			else
			{
			    /* There's a white run at the right edge
			     * so add its dummy right end point.
			     */
			    whiteRunBuffer[whiteRunCount - 1].runRight = width;
			}
		    }

		    /* We'll be erasing white runs which means writing
		     * black so start off with a white word.
		     */
		    outWord = 0x00000000;
		    lRunWord = 0;

		    /* go thru the run array for the current row */
		    for (i = 0; i < whiteRunCount; i++)
		    {
		        /* Since this was originally an array of black
			 * runs, the ends of the white runs are one off.
			 */
		        left = whiteRunBuffer[i].runLeft + 1;
			right = whiteRunBuffer[i].runRight - 1;
			length = right - left + 1;

                        /* does the run NOT pass the filter? */
			if (((polarity == cPassRuns) && 
			    ((length < minRunLength) || 
			     (length > maxRunLength))) ||
			    ((polarity == cRejectRuns) &&
			    ((length >= minRunLength) && 
			     (length <= maxRunLength))))
			{
			    /* Erase the run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;

                            /* If we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
			        /* write the output word then clear it */
                                ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) |=
				                           outWord;
				outWord = 0x00000000;
				lRunWord = cRunWord;
                            }

                            /* Is this a single or multiple word run? */
                            if (wc == 1)
			    {
			        outWord |= (endRunPosLUT[(left % 32)]) &
	                                   (endRunNegLUT[(right % 32)]);
			    }
			    else
			    {
                                rowPtr = (UInt32 *)srcImage + cRunWord;
				ULONG_ACCESS (rowPtr) |= (outWord |
				        endRunPosLUT[(left % 32)]);
				outWord = 0x00000000;
				wc--;
				rowPtr++;

                                /* Write all of the black middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0xffffffff;
	                            rowPtr++;
				}

                                /* Start the next output word by erasing
				 * the tail end of this run
				 */
				outWord |= endRunNegLUT[(right % 32)];

				lRunWord = right / 32;
			    }

			}  /* erase the run? */

                    }  /* for each white run */

                    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) |= outWord;

		    srcImage += srcBpl;

                }  /* for each row */
            }
	    else
	    {
	        /* filter white runs in->out */

		/* filter runs one row at a whack */
		for (row = 0; row < height; row++)
		{
		    /* get all the runs in the current row */
		    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
		                                        width,
							0,
							runBuffer,
							runBufferSize,
							0,
							&dummyRun);

                    if (__error__ != ia_successful)
		        break;

		    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
		                       sizeof (Run32));
                    whiteRunCount = runCount + 1;

		    if (whiteRunCount == 1)
		    {
			/* all white row special case */
		        whiteRunBuffer = runBuffer;
			whiteRunBuffer[0].runLeft = -1;
			whiteRunBuffer[0].runRight = width;
		    }
		    else  /* multiple white runs */
		    {
			/* We want to shift the run array left or right one
			 * half place to make it an array of white runs. 
			 * The endpoints of the white runs will be 1 off.
			 */
			if (runBuffer[0].runLeft == 0)
			{
			    /* There's a black run at the left edge so
			     * shift the array right.
			     */
			    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer + 1);
			    whiteRunCount--;
			}
			else
			{
			    /* There's a white run at the left edge so
			     * shift the array left and add a dummy start
			     * value for the first white run.
			     */
			    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer - 1);
    
			    /* put the dummy value in for the left edge of the
			     * white run.
			     */
			    whiteRunBuffer[0].runLeft = -1;
			}
    
			/* Now we need to fix up the right edge */
			if (runBuffer[runCount - 1].runRight == width - 1)
			{
			    /* There's a black run at the right edge so
			     * the end of the white run array is shifted
			     * left.
			     */
			    whiteRunCount--;
			}
			else
			{
			    /* There's a white run at the right edge
			     * so add its dummy right end point.
			     */
			    whiteRunBuffer[whiteRunCount - 1].runRight = width;
			}
		    }

		    /* We'll be writing white runs which means erasing
		     * black so start off with a black word.
		     */
		    outWord = 0xffffffff;
		    lRunWord = 0;

		    /* go thru the run array for the current row */
		    for (i = 0; i < whiteRunCount; i++)
		    {
		        /* Since this was originally an array of black
			 * runs, the ends of the white runs are one off.
			 */
		        left = whiteRunBuffer[i].runLeft + 1;
			right = whiteRunBuffer[i].runRight - 1;
			length = right - left + 1;

                        /* does the run pass the filter? */
			if (((polarity == cPassRuns) && 
			    ((length >= minRunLength) && 
			     (length <= maxRunLength))) ||
			    ((polarity == cRejectRuns) &&
			    ((length < minRunLength) || 
			     (length > maxRunLength))))
			{
			    /* Write the white run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;

                            /* If we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
			        /* write the output word then clear it */
                                ULONG_ACCESS ((UInt32 *)destImage + lRunWord) &=
				                           outWord;
				outWord = 0xffffffff;
				lRunWord = cRunWord;
                            }

                            /* Is this a single or multiple word run? */
                            if (wc == 1)
			    {
			        outWord &= ~((endRunPosLUT[(left % 32)]) &
	                                     (endRunNegLUT[(right % 32)]));
			    }
			    else
			    {
                                rowPtr = (UInt32 *)destImage + cRunWord;
				ULONG_ACCESS (rowPtr) &= (outWord &
				        ~(endRunPosLUT[(left % 32)]));
				outWord = 0xffffffff;
				wc--;
				rowPtr++;

                                /* Write all of the white middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0x00000000;
	                            rowPtr++;
				}

                                /* Start the next output word by erasing
				 * the tail end of this run
				 */
				outWord &= ~(endRunNegLUT[(right % 32)]);

				lRunWord = right / 32;
			    }

			}  /* erase the run? */

                    }  /* for each white run */

                    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)destImage + lRunWord) &= outWord;

		    srcImage += srcBpl;
                    destImage += destBpl;

                }  /* for each row */

	    }  /* filtering white runs in->out */
	}
	else  /* filtering black runs */
	{
	    if (inPlace)
	    {
	        /* filter in->in (in-place) */

		/* filter runs one row at a whack */
		for (row = 0; row < height; row++)
		{
		    /* get all the runs in the current row */
		    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
		                                        width,
							0,
							runBuffer,
							runBufferSize,
							0,  /* cFrameBits? */
							&dummyRun);

                    if (__error__ != ia_successful)
		        break;

		    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
		                       sizeof (Run32));

		    outWord = 0xffffffff;
		    lRunWord = 0;

		    /* go thru the run array for the current row */
		    for (i = 0; i < runCount; i++)
		    {
		        left = runBuffer[i].runLeft;
			right = runBuffer[i].runRight;
			length = right - left + 1;

                        /* does the run NOT pass the filter? */
			if (((polarity == cPassRuns) && 
			    ((length < minRunLength) || 
			     (length > maxRunLength))) ||
			    ((polarity == cRejectRuns) &&
			    ((length >= minRunLength) && 
			     (length <= maxRunLength))))
			{
			    /* erase the run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;

                            /* if we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
			        /* write the output word then clear it */
                                ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) &=
				                           outWord;
				outWord = 0xffffffff;
				lRunWord = cRunWord;
                            }

                            /* is this a single or multiple word run? */
                            if (wc == 1)
			    {
                                outWord &= ~((endRunPosLUT[(left % 32)]) &
	                                     (endRunNegLUT[(right % 32)]));
			    }
			    else
			    {
                                rowPtr = (UInt32 *)srcImage + cRunWord;
				ULONG_ACCESS (rowPtr) &= (outWord &
				        ~(endRunPosLUT[(left % 32)]));
				wc--;
				rowPtr++;
				outWord = 0xffffffff;

                                /* write all of the empty middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0x00000000;
	                            rowPtr++;
				}

                                /* start the next output word by erasing
				 * the tail end of this run.
				 */
				outWord &= ~(endRunNegLUT[(right % 32)]);

				lRunWord = right / 32;
			    }
			}
		    }

                    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)srcImage + lRunWord) &= outWord;

		    srcImage += srcBpl;
		}
	    }
	    else
	    {
	        /* filter in->out */

		/* filter runs one row at a whack */
		for (row = 0; row < height; row++)
		{
		    /* get all the runs in the current row */
		    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
		                                        width,
							0,
							runBuffer,
							runBufferSize,
							0,
							&dummyRun);

                    if (__error__ != ia_successful)
		        break;

		    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
		                       sizeof (Run32));

		    outWord = 0x00000000;
		    lRunWord = 0;

		    /* go thru the run array for the current row */
		    for (i = 0; i < runCount; i++)
		    {
		        left = runBuffer[i].runLeft;
			right = runBuffer[i].runRight;
			length = right - left + 1;

                        /* does the run pass the filter? */
			if (((polarity == cPassRuns) && 
			    ((length >= minRunLength) && 
			     (length <= maxRunLength))) ||
			    ((polarity == cRejectRuns) &&
			    ((length < minRunLength) || 
			     (length > maxRunLength))))
			{
			    /* construct the run in the output image */
			    cRunWord = left / 32;
			    wc = (right / 32) - cRunWord + 1;

                            /* if we moved to a new word, write the old one */
			    if (cRunWord != lRunWord)
			    {
			        /* write the output word then clear it */
                                ULONG_ACCESS ((UInt32 *)destImage + lRunWord) |=
				                           outWord;
				outWord = 0x00000000;
				lRunWord = cRunWord;
                            }

                            /* is this a single or multiple word run? */
                            if (wc == 1)
			    {
			        outWord |= (endRunPosLUT[(left % 32)]) &
	                                   (endRunNegLUT[(right % 32)]);
			    }
			    else
			    {
                                rowPtr = (UInt32 *)destImage + cRunWord;
				ULONG_ACCESS (rowPtr) |= (outWord |
				        endRunPosLUT[(left % 32)]);
				wc--;
				rowPtr++;
				outWord = 0x00000000;

                                /* write all of the full middle words */
				for (; wc > 1; wc--)
				{
				    ULONG_ACCESS (rowPtr) = 0xffffffff;
	                            rowPtr++;
				}

                                /* start the next output word with the tail
				 * end of this run.
				 */
                                outWord |= (endRunNegLUT[(right % 32)]);

				lRunWord = right / 32;
			    }
			}
		    }

                    /* Be sure to write out the last word we worked on! */
		    ULONG_ACCESS ((UInt32 *)destImage + lRunWord) |= outWord;

		    srcImage += srcBpl;
                    destImage += destBpl;
		}
	    }

	}  /* normal black run filtering */

    }  /* normal run filtering */

    /* if we allocated this locally, free it */
    if (runBufAlloced)
    {
	runBuffer--;	/* back up to beginning of run buffer */
        FREE (runBuffer);
    }

    NORMAL_RETURN;

}  /* i_filterHorizontalRuns () */


/**************************************************************************
 * i_filterVerticalRuns ()
 *
 * Filters vertical black or white runs by their length.  Provides
 * the canonical set of filter operations: HP, LP, BP, BR.
 *
 * Arguments:
 *	runBuf :	Pointer to run buffer of the appropriate size
 *			as determined by i_getRunFilterBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	runBufferSize :	If the run buffer is pre-allocated (preferred),
 *			the size of the buffer (in bytes).  If a NULL
 *			pointer is passed for the run buffer this value
 *			is ignored.
 *	rotBuf :	Pointer to rotation buffer of the appropriate
 *			size as determined by i_getRunFilterBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	rotBufBpl :	Bytes per line of the rotation buffer as
 *			determined by i_getRunFilterBufferSizes ().  If
 *			a NULL pointer is passed for the rotation buffer
 *			this value is ignored.
 *	srcImage :	Pointer to source image data including frame.
 *	width :		Width of source and dest images (MUST be the same).
 *	height :	Height of source and dest images (MUST be the same).
 *	srcBpl :	Bpl of source image.
 *	destBpl :	Bpl of dest image.
 *      runValue :	"Color" of runs to filter (0 = white, 1 = black)
 *      borderValue :   "Color" of border beyond the image (0 = white,
 *                      1 = black).  The border is assumed to be infinite
 *                      for run length calculations (e.g. a black run which
 *			abuts a black border is assumed to be infinitely
 *			long and is filtered appropriately).
 *	minRunLength :	Minimum run length in band.
 *	maxRunLength :  Maximum run length in band.
 *	polarity :	Filter polarity can be
 *			cPassRuns - Runs which are within the band
 *			are kept, others are discarded, or
 *			cRejectRuns - Runs which are within the band
 *			are discarded, others are kept.
 *	destImage :	Pointer to dest image data including frame.
 *
 * Returns:
 *	ia_successful	if filter operation completed normally,
 *	error codes	if bad things happened during filtering.
 *
 * Notes:
 *	This function is capable of performing all the basic
 *	filter functions: highpass, lowpass, bandpass, band
 *	reject (or notch) based on vertical run length.
 *
 *	The vertical run filter is necessarily slower than
 *	its horizontal counterpart since it performs two
 *	rotations of each column (actually it works on
 *	blocks of columns) in addition to the run finding/
 *	filtering operations.
 *
 *	If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be replaced with the results
 *	produced by this function.  This differs from
 *	i_filterHorizontalRuns ().  To combine results from
 *	a vertical and a horizontal run filter, perform the
 *	the vertical filter first, then the horizontal using the
 *	same destination image for both operations.
 *
 **************************************************************************/
Int32 CDECL 
i_filterVerticalRuns (
    IPState    *pIPState,
    UInt8      *runBuf,
    Int32	runBufferSize,
    UInt8      *rotBuf,
    Int32	rotBufBpl,
    UInt8      *srcImage,
    Int32	width,
    Int32	height,
    Int32	srcBpl,
    Int32	destBpl,
    Int32	runValue,
    Int32	borderValue,
    Int32	minRunLength,
    Int32	maxRunLength,
    Int32	polarity,
    UInt8      *destImage)
{
    PROLOGUE;
    Int32   col,            /* current column of input image */
            wu,             /* width unframed */
	    lastSlice,      /* width of last vertical slice processed */
	    runBufAlloced = cFalse,
	    rotBufAlloced = cFalse;
    UInt8  *oDestImage,     /* Original dest image pointer */
           *tmpImage;       /* tmp image data pointer */


    __error__ = ia_successful;

    if (runBuf == NULL)
    {
        runBufferSize = (((height - (2 * cFrameBits)) / 2) + 3);
        runBuf = (UInt8 *)MALLOC (runBufferSize * sizeof (Run32));
	if (runBuf == NULL)
            return (ia_nomem);
	else
	    runBufAlloced = cTrue;
    }



    /* save the original dest image pointer for later frame clearing */
    oDestImage = destImage;

    if (rotBuf == NULL)
    {
        rotBufBpl = (((height + 31) / 32) * 4);
        tmpImage = (UInt8 *)CALLOC (1, rotBufBpl * 
	                           (COL_BLOCK_SIZE + (2 * cFrameBits)));
	if (tmpImage == NULL)
	{
	    if (runBufAlloced)
	        FREE (runBuf);

	    return (ia_nomem);
	}
	else
	    rotBufAlloced = cTrue;
    }
    else
	tmpImage = rotBuf;


    /* figure out the width of the last slice */
    lastSlice = width % COL_BLOCK_SIZE;
    if (lastSlice == 0)  /* it must be a full slice */
        lastSlice = COL_BLOCK_SIZE;

    wu = width - (2 * cFrameBits);

    /* process the image in 32 column vertical bands */
    for (col = 0; col < wu; col += COL_BLOCK_SIZE)
    {
	/* rotate COL_BLOCK_SIZE columns of the original image 90 into 
	 * the temp image.
	 */
	__error__ = i_fastRot90 (pIPState, tmpImage, srcImage, 
	                       (COL_BLOCK_SIZE + (2 * cFrameBits)), 
	                       height, 
			       srcBpl, 
			       rotBufBpl, 1);

        if (__error__ != ia_successful)
	    break;

	/* filter horizontal runs in place in the temp image */
        __error__ = i_filterHorizontalRuns (
			pIPState, runBuf, runBufferSize,
			tmpImage,
			height,
			(((col + COL_BLOCK_SIZE) < wu) ? 
				(COL_BLOCK_SIZE + (2 * cFrameBits)) : 
				(lastSlice + (2 * cFrameBits))),
			rotBufBpl, rotBufBpl,
			runValue, borderValue,
			minRunLength, maxRunLength,
			polarity, tmpImage);

        if (__error__ != ia_successful)
	    break;

	/* rotate the temp image 270 degrees into either the output image
	 * or back into the input image if in place.
	 */
	__error__ = i_fastRot270 (pIPState, destImage, tmpImage, 
	                        height, 
		                COL_BLOCK_SIZE + (2 * cFrameBits),
			        rotBufBpl, destBpl, 1);

        if (__error__ != ia_successful)
	    break;

        /* increment the pointers to the next column block */
        srcImage += COL_BLOCK_SIZE / 8;
	destImage += COL_BLOCK_SIZE / 8;

    }  /* for each column block */

    /* Clear the right frame if necessary since the rotate may have
     * written trash there on the last iteration.
     */
    if ((lastSlice != 0) && (__error__ == ia_successful))
    {
        __error__ = i_rasterOpFull (pIPState, oDestImage, 
	                          width,
	                          height,
				  0,
				  0,
				  width - cFrameBits,
				  0,
				  cFrameBits,
				  height,
				  1,
				  destBpl,
				  PIX_CLR,
				  NULL, 0, 0, 0, 0, 0, 0, 0, 0);
/*
	__error__ = i_rasterOpFull (pIPState, oDestImage, 
	                          destBpl << 3,
	                          height + (2 * cFrameBits),
				  0,
				  0,
				  width + cFrameBits,
				  0,
				  cFrameBits,
				  height + (2 * cFrameBits),
				  1,
				  destBpl,
				  PIX_CLR,
				  NULL, 0, 0, 0, 0, 0, 0, 0, 0);
*/
    }

    /* if we allocated this locally, free it */
    if (runBufAlloced)
        FREE (runBuf);
    if (rotBufAlloced)
        FREE (tmpImage);

    NORMAL_RETURN;

}  /* i_filterVerticalRuns () */



/*
 * $Log: i_runflt.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 * Imported sources
 *
 * Revision 1.17  1998/01/21 21:42:51  ddavies
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fix Doug's weird syntax for FREE - don't decrement the thing freed in the
 * macro invocation.  Especially, don't decrement the pointer AFTER the thing
 * it's pointing to has been freed.
 *
 * Revision 1.16  1998/01/13 23:49:33  ddavies
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.15  1996/05/29 20:14:10  dferrell
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fixed bugs related to frames and framed width/height which were non-fatal.
 *
 * Revision 1.14  1996/05/22 22:15:53  dferrell
 * dferrell on Wed May 22 15:14:48 PDT 1996
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fixed typo in comment.
 *
 * Revision 1.13  1996/02/26 20:05:25  dferrell
 * dferrell on Mon Feb 26 12:04:11 PST 1996
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fixed bug in vertical filtering whereby the last 32 columns were not
 * processed for images with widths of exact multiples of 32.
 *
 * Revision 1.12  1996/02/20 23:52:00  dferrell
 * dferrell on Tue Feb 20 15:51:02 PST 1996
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Made bpl parameter order consistent in vertical and horizontal filter
 * functions.
 *
 * Revision 1.11  1996/02/09 18:05:24  dferrell
 * dferrell on Fri Feb 9 10:04:52 PST 1996
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fixed pair of bugs caused by last bug fix.  Should not core dump now (doh!).
 *
 * Revision 1.10  1996/02/08 22:58:58  dferrell
 * dferrell on Thu Feb 8 14:57:57 PST 1996
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fixed bug when filtering black runs and a row is all white.
 *
 * Revision 1.9  1996/02/08 00:34:47  dferrell
 * dferrell on Wed Feb 7 16:32:11 PST 1996
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Removed state init and finish functions and replaced with "get buffer size"
 * function.  Modified interfaces for run filter functions to include
 * seperate parameters rather than a simple state block.  Added auto-alloc
 * of buffers if none passed in.
 *
 * Revision 1.8  1996/01/17  00:55:07  dferrell
 * dferrell on Tue Jan 16 16:51:54 PST 1996
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fixed bug in i_filterHorizontalRuns () where I had left out a typecaste
 * in an rfState offset calculation.  This caused us to write beyond the
 * run array.
 *
 * Revision 1.7  1995/12/13  02:45:44  ejaquith
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.6  1995/12/12  20:18:38  dferrell
 * dferrell on Tue Dec 12 12:13:57 PST 1995
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Complete rewrite of run filtering code.  Now able to filter black or
 * white runs; specify "border" color (for special edge-abutting run
 * handling).  These changes make run filtering equivalent to morphological
 * opening and closing with linear strels.  For large "strels" run filtering
 * is faster (where large is platform dependent).
 *
 * Revision 1.5  1995/01/16  22:46:48  dferrell
 * dferrell on Mon Jan 16 14:45:32 PST 1995
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fixed miscalculation of bpl in temporary rotation buffer use by
 * i_filterVerticalRuns.  This resulted in non-word aligned row addresses
 * being generated in the rotation code and !Bus Error!
 *
 * Revision 1.4  1995/01/12  22:57:02  dferrell
 * dferrell on Thu Jan 12 14:56:36 PST 1995
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Improved comments in function description block.
 *
 * Revision 1.3  1995/01/12  19:38:35  dferrell
 * dferrell on Thu Jan 12 11:37:38 PST 1995
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Fixed bug where the tmpImage buffer for rotation in the vertical run
 * filter was not large enough.  Added more comments.
 *
 * Revision 1.2  1995/01/12  18:11:23  lperry
 * lperry on Thu Jan 12 10:09:25 PST 1995
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * Added type cast to calloc to silence acc complaint.
 * Added RCS log macro.
 *
 * Revision 1.1  1995/01/11  14:34:14  dferrell
 * dferrell on Wed Jan 11 14:34:14 PST 1995
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * New file containing public interface for vertical and horizontal run
 * filtering.
 *
 */

