/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_runprj.c
 *
 *	i_projectHorizontalRuns ()
 *	i_projectVerticalRuns ()
 */

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#include "shrip.pub"
#include "shrcc.pub"
/*
#include "imageref.pub"
*/

IP_RCSINFO(RCSInfo, "$RCSfile: i_runprj.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* These functions provide simple and fast orthogonal projection
 * capability with built-in run filtering.  
 *
 */

#define ROT_BLOCK_SIZE 32


/* PUBLIC */
/**************************************************************************
 * i_projectHorizontalRuns ()
 *
 * Creates a horizontal projection histogram of a binary image.
 * Can project white or black runs.  Includes horizontal run pre-filtering.
 *
 * Arguments:
 *	runBuf :	Pointer to run buffer of the appropriate size
 *			as determined by i_getRunFilterBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	runBufferSize :	If the run buffer is pre-allocated (preferred),
 *			the size of the buffer (in bytes).  If a NULL
 *			pointer is passed for the run buffer this value
 *			is ignored.
 *	srcImage :	Pointer to source image data including frame.
 *	w :		Framed width of source image.
 *	h :		Framed height of source image.
 *	bpl :		Bpl of source image.
 *      runValue :	"Color" of runs to filter (0 = white, 1 = black)
 *      includeEndRuns:	If cTrue, project all runs, if cFalse, ignore
 *                      runs which begin or end at the left and/or right
 *			edge of the image.
 *	minRunLength :	Minimum run length in band (inclusive).
 *	maxRunLength :  Maximum run length in band (inclusive).
 *	polarity :	Run filter polarity can be
 *			cPassRuns - Runs which are within the band
 *			are kept, others are discarded, or
 *			cRejectRuns - Runs which are within the band
 *			are discarded, others are kept.
 *	projectionOffset: Index offset into the projection array (normally
 *			should be 0).  Used when called from
 *			i_projectVerticalRuns ().
 *  	projection:	Array of projection data.  Must be cleared between
 *			calls if re-used.
 *
 * Returns:
 *	ia_successful	if filter operation completed normally,
 *	error codes	if bad things happened during filtering.
 *
 * Notes:
 *
 **************************************************************************/
Int32 CDECL 
i_projectHorizontalRuns (
    IPState    *pIPState,
    UInt8      *runBuf,
    Int32	runBufferSize,
    UInt8      *srcImage,
    Int32	w,
    Int32	h,
    Int32	d,
    Int32	bpl,
    Int32	runValue,
    Int32	includeEndRuns,
    Int32	minRunLength,
    Int32	maxRunLength,
    Int32	polarity,
    Int32	projectionOffset,
    Int32      *projection)
{
    PROLOGUE;
    Int32   row,            /* current row of input image */
	    runCount,       /* number of runs found in current row */
	    whiteRunCount,  /* number of white runs in current row */
	    runBufAlloced = cFalse;
    register Int32  i;
    register Int32  left;
    register Int32  right;    /* left and right offsets of current run */
    register Int32  length;   /* length of current run */
    Run32  *runBuffer,        /* array of runs to pass to run finder */
           *whiteRunBuffer,   /* array of white runs */
	   *dummyRun;

    if (d != 1)
        return (ia_depthNotSupported);

    if (runBuf == NULL)
    {
        runBufferSize = (((w - (2 * cFrameBits)) / 2) + 3);
        runBuffer = (Run32 *)MALLOC (runBufferSize * sizeof (Run32));
	if (runBuffer == NULL)
            return (ia_nomem);
	else
	    runBufAlloced = cTrue;
    }
    else
    {
        runBuffer = (Run32 *)runBuf;
	/* this is passed in as size in bytes */
	runBufferSize /= sizeof (Run32);
    }

    /* "Unframe" the sizes */
    w -= (2 * cFrameBits);
    h -= (2 * cFrameBits);

    srcImage += (bpl << cLogFrameBits) + cFrameBytes;

    /* Remember that we want to keep an extra run on either end
     * for white run filtering, so offset the start by 1 and
     * only let the run finder use the middle part.
     */
    runBuffer++;

    projection += projectionOffset;

    if (runValue == 1)  /* black run projection */
    {
	/* process runs one row at a whack */
	for (row = 0; row < h; row++)
	{
	    /* get all the runs in the current row */
	    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
						w,
						0,
						runBuffer,
						runBufferSize,
						0,
						&dummyRun);

	    if (__error__ != ia_successful)
		break;

	    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
			       sizeof (Run32));

	    /* go thru the run array for the current row */
	    for (i = 0; i < runCount; i++)
	    {
		left = runBuffer[i].runLeft;
		right = runBuffer[i].runRight;
		length = right - left + 1;

		/* does the run pass the filter? */
		if (((polarity == cPassRuns) && 
		    ((length >= minRunLength) && 
		     (length <= maxRunLength))) ||
		    ((polarity == cRejectRuns) &&
		    ((length < minRunLength) || 
		     (length > maxRunLength))))
		{
		    if ((left == 0) || (right == w - 1))
		    {
		        if (includeEndRuns)
		            *projection += length;
		    }
		    else
		        *projection += length;
		}

            }  /* for each run */

            srcImage += bpl;
            projection++;

        }  /* for each row */
    }
    else  /* white run projection */
    {
	/* process runs one row at a whack */
	for (row = 0; row < h; row++)
	{
	    /* get all the runs in the current row */
	    __error__ = i_findAllRunsInLineArray (pIPState, srcImage,
						w,
						0,
						runBuffer,
						runBufferSize,
						0,
						&dummyRun);

	    if (__error__ != ia_successful)
		break;

	    runCount = (Int32)(((Int32)dummyRun - (Int32)runBuffer) / 
			       sizeof (Run32));
	    whiteRunCount = runCount + 1;

	    if (whiteRunCount == 1)
	    {
		/* all white row special case */
		if (includeEndRuns && (polarity == cPassRuns))
		    *projection += w;
	    }
	    else
	    {
		if (runBuffer[0].runLeft == 0)
		{
		    /* There's a black run at the left edge so
		     * shift the array right.
		     */
		    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer + 1);
		    whiteRunCount--;
		}
		else
		{
		    /* There's a white run at the left edge so
		     * shift the array left and add a dummy start
		     * value for the first white run.
		     */
		    whiteRunBuffer = (Run32 *)((Int32 *)runBuffer - 1);

		    /* put the dummy value in for the left edge of the
		     * white run.
		     */
		    whiteRunBuffer[0].runLeft = -1;
		}

		/* Now we need to fix up the right edge */
		if (runBuffer[runCount - 1].runRight == w - 1)
		{
		    /* There's a black run at the right edge so
		     * the end of the white run array is shifted
		     * left.
		     */
		    whiteRunCount--;
		}
		else
		{
		    /* There's a white run at the right edge
		     * so add its dummy right end point.
		     */
		    whiteRunBuffer[whiteRunCount - 1].runRight = w;
		}

		/* go thru the run array for the current row */
		for (i = 0; i < whiteRunCount; i++)
		{
		    left = whiteRunBuffer[i].runLeft + 1;
		    right = whiteRunBuffer[i].runRight - 1;
		    length = right - left + 1;

		    /* does the run pass the filter? */
		    if (((polarity == cPassRuns) && 
			((length >= minRunLength) && 
			 (length <= maxRunLength))) ||
			((polarity == cRejectRuns) &&
			((length < minRunLength) || 
			 (length > maxRunLength))))
		    {
			if ((left == 0) || (right == w - 1))
			{
			    if (includeEndRuns)
				*projection += length;
			}
			else
			    *projection += length;
		    }
    
		}  /* for each run */

            }  /* multiple white runs */

	    srcImage += bpl;
	    projection++;

        }  /* for each row */
    }

    /* if we allocated this locally, free it */
    if (runBufAlloced)
    {
	runBuffer--;	/* back up to beginning of buffer */
        FREE (runBuffer);
    }

    NORMAL_RETURN;

}  /* i_horizontalProjectRuns () */


/**************************************************************************
 * i_projectVerticalRuns ()
 *
 * Creates a vertical projection histogram of a binary image.
 * Can project white or black runs.  Includes vertical run pre-filtering.
 *
 * Arguments:
 *	runBuf :	Pointer to run buffer of the appropriate size
 *			as determined by i_getRunFilterBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	runBufferSize :	If the run buffer is pre-allocated (preferred),
 *			the size of the buffer (in bytes).  If a NULL
 *			pointer is passed for the run buffer this value
 *			is ignored.
 *	rotBuf :	Pointer to rotation buffer of the appropriate
 *			size as determined by i_getRunFilterBufferSizes (),
 *			or NULL.  If NULL, the buffer will be allocated
 *			and freed internally.
 *	rotBufBpl :	Bytes per line of the rotation buffer as
 *			determined by i_getRunFilterBufferSizes ().  If
 *			a NULL pointer is passed for the rotation buffer
 *			this value is ignored.
 *	srcImage :	Pointer to source image data including frame.
 *	w :		Framed width of source image.
 *	h :		Framed height of source image.
 *	bpl :		Bpl of source image.
 *      runValue :	"Color" of runs to filter (0 = white, 1 = black)
 *      includeEndRuns:	If cTrue, project all runs, if cFalse, ignore
 *                      runs which begin or end at the top and/or bottom
 *			edge of the image.
 *	minRunLength :	Minimum run length in band (inclusive).
 *	maxRunLength :  Maximum run length in band (inclusive).
 *	polarity :	Run filter polarity can be
 *			cPassRuns - Runs which are within the band
 *			are kept, others are discarded, or
 *			cRejectRuns - Runs which are within the band
 *			are discarded, others are kept.
 *	projectionOffset: Index offset into the projection array (normally
 *			should be 0).
 *  	projection:	Array of projection data.  Must be cleared between
 *			calls if re-used.
 *
 * Returns:
 *	ia_successful	if filter operation completed normally,
 *	error codes	if bad things happened during filtering.
 *
 * Notes:
 *
 **************************************************************************/
Int32 CDECL 
i_projectVerticalRuns (
    IPState    *pIPState,
    UInt8      *runBuf,
    Int32	runBufferSize,
    UInt8      *rotBuf,
    Int32	rotBufBpl,
    UInt8      *srcImage,
    Int32	w,
    Int32	h,
    Int32	d,
    Int32	bpl,
    Int32	runValue,
    Int32	includeEndRuns,
    Int32	minRunLength,
    Int32	maxRunLength,
    Int32	polarity,
    Int32	projectionOffset,
    Int32      *projection)
{
    PROLOGUE;
    Int32   col,            /* current column of input image */
            wu,             /* width unframed */
	    lastSlice,      /* width of last vertical slice processed */
	    runBufAlloced = cFalse,
	    rotBufAlloced = cFalse;
    UInt8  *tmpImage;       /* tmp image data pointer */

    if (d != 1)
        return (ia_depthNotSupported);

    __error__ = ia_successful;

    if (runBuf == NULL)
    {
        runBufferSize = (((h - (2 * cFrameBits)) / 2) + 3);
        runBuf = (UInt8 *)MALLOC (runBufferSize * sizeof (Run32));
	if (runBuf == NULL)
            return (ia_nomem);
	else
	    runBufAlloced = cTrue;
    }

    if (rotBuf == NULL)
    {
        rotBufBpl = (((h + 31) / 32) * 4);
        tmpImage = (UInt8 *)CALLOC (1, rotBufBpl * 
	                           (ROT_BLOCK_SIZE + (2 * cFrameBits)));
	if (tmpImage == NULL)
	{
	    if (runBufAlloced)
	        FREE (runBuf);

	    return (ia_nomem);
	}
	else
	    rotBufAlloced = cTrue;
    }
    else
	tmpImage = rotBuf;


    /* figure out the width of the last slice */
    lastSlice = w % ROT_BLOCK_SIZE;
    if (lastSlice == 0)  /* it must be a full slice */
        lastSlice = ROT_BLOCK_SIZE;

    projection += projectionOffset;

    wu = w - (2 * cFrameBits);

    /* process the image in 32 column vertical bands */
    for (col = 0; col < wu; col += ROT_BLOCK_SIZE)
    {
	/* rotate ROT_BLOCK_SIZE columns of the original image 90 into 
	 * the temp image.
	 */
	__error__ = i_fastRot90 (pIPState, tmpImage, srcImage, 
	                       (ROT_BLOCK_SIZE + (2 * cFrameBits)), 
	                       h, 
			       bpl, 
			       rotBufBpl, 1);

        if (__error__ != ia_successful)
	    break;

	/* filter horizontal runs in place in the temp image */
        __error__ = i_projectHorizontalRuns (
			pIPState, runBuf, runBufferSize,
			tmpImage,
			h,
			(((col + ROT_BLOCK_SIZE) < wu) ?
				(ROT_BLOCK_SIZE + (2 * cFrameBits)) :
				(lastSlice + (2 * cFrameBits))),
			d,
			rotBufBpl,
			runValue, includeEndRuns,
			minRunLength, maxRunLength,
			polarity, projectionOffset,
			projection);

        if (__error__ != ia_successful)
	    break;

        /* increment the pointers to the next column block */
        srcImage += ROT_BLOCK_SIZE / 8;

	/* increment the projection offset by the column block width */
	projectionOffset += ROT_BLOCK_SIZE;

    }  /* for each column block */

    /* if we allocated this locally, free it */
    if (runBufAlloced)
        FREE (runBuf);
    if (rotBufAlloced)
        FREE (tmpImage);

    NORMAL_RETURN;


}  /* i_projectVerticalRuns () */

/*
 * $Log: i_runprj.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:12  jlin
 * Imported sources
 *
 * Revision 1.3  1998/01/21 21:50:19  ddavies
 * ddavies on Wed Jan 21 13:49:51 PST 1998
 *
 * i_runprj.c -> /product/ipcore/ipshared/src/RCS/i_runprj.c,v
 *
 * Fix Doug's weird syntax for FREE - don't decrement the thing freed in the
 * macro invocation.  Especially, don't decrement the pointer AFTER the thing
 * it's pointing to has been freed.
 *
 * Revision 1.2  1998/01/13 23:49:35  ddavies
 * i_runprj.c -> /product/ipcore/ipshared/src/RCS/i_runprj.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.1  1996/05/29 20:14:30  dferrell
 * Initial revision
 *
 */

