/*****************************************************************
 *  Copyright (c) 1993.                              		 *
 *  Xerox Corporation.  All rights reserved.                     *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_subsmp1.c: contains subroutines simple subsampled reduction. 
 *
 *		    Int32	 i_subsample1bpp()
 *		    Int32	 i_subsample4bpp()
 *		    Int32	 i_subsample8bpp()
 *
 * Internal procedures:
 *
 *          void     ip_subsample8bpp1_5xLine()
 *          void     ip_subsample4bpp1_5xLine()
 */


#include <stdio.h>
#include <stdlib.h>

#ifndef _TYPES_PUB_INCLUDED
#include "types.pub"
#include "ipscore.prv"
#endif

#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"


    /* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_sbsmp1.c,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:18:12 $ */

/* PUBLIC */
/**************************************************************************
 *  i_subsample1bpp()
 *  	Computes simple subsampled reduction for 1 bit/pixel images
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_subsample1bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
PROLOGUE;
 register UInt8	*sPtr, *dPtr;
register Int32   row, col;
Int32		 sInc;
UInt8		*redTable;

UInt8		 oWord, srcMask;
    
	/* adjust base pointers to avoid the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    switch (factor)
    {
    case 2:
	sInc = sBpl * 2;
	srcMask = 0xFF << ((8 - (wD & 0x7)) & 0x7);
	if ((__error__ = ip_makeRedLUT2 (pIPState, &redTable)) != ia_successful)
	    NORMAL_RETURN;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
        {
            dPtr = pD;
            sPtr = pS;

            for (col = wD; col > 0; col -= 8, dPtr++, sPtr += 2)
            {
                UCHAR_ACCESS(dPtr) = 
		    (redTable[UCHAR_ACCESS(sPtr)] << 4) | 
		    redTable[UCHAR_ACCESS(&sPtr[1])];
            }
		/* get rid of any extra bits in the last byte.  this
		 * would have occurred if the src was an odd number of
		 * bits.
		 */
	    UCHAR_ACCESS(dPtr-1) = UCHAR_ACCESS(dPtr-1) & srcMask;
        }
	FREE(redTable);
	break;

    case 4:
	sInc = sBpl * 4;
	srcMask = 0xFF << ((8 - (wD & 0x7)) & 0x7);
	if ((__error__ = ip_makeRedLUT4 (pIPState, &redTable)) != ia_successful)
	    NORMAL_RETURN;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
        {
            dPtr = pD;
            sPtr = pS;

            for (col = wD; col > 0; col -= 8, dPtr++, sPtr += 4)
            {
		UCHAR_ACCESS(dPtr) = 
		    (redTable[UCHAR_ACCESS(sPtr)] << 6) |
		    (redTable[UCHAR_ACCESS(&sPtr[1])] << 4) | 
		    (redTable[UCHAR_ACCESS(&sPtr[2])] << 2) | 
 		    redTable[UCHAR_ACCESS(&sPtr[3])]; 
            }
	    UCHAR_ACCESS(dPtr-1) = UCHAR_ACCESS(dPtr-1) & srcMask;
        }
	FREE(redTable);
	break;

    case 8:
	sInc = sBpl * 8;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
        {
            dPtr = pD;
            sPtr = pS;

		/* for each full 8-bit destination quantity....
		 * the last 8-bit quantity may or may not really
		 * be composed of 8 bits and if we reference the
		 * source as if all 8 bits are valid we will reference
		 * past the end of the frame.   Ooops.  So, we
		 * will do the last 8-bit quantity separately.
		 */
            for (col = (wD & ~0x7); col > 0; col -= 8, dPtr++, sPtr += 8)
            {
		UCHAR_ACCESS(dPtr) =
		    (UCHAR_ACCESS(sPtr) & 0x80) |
		    ((UCHAR_ACCESS(&sPtr[1]) >> 1) & 0x40) |
		    ((UCHAR_ACCESS(&sPtr[2]) >> 2) & 0x20) |
		    ((UCHAR_ACCESS(&sPtr[3]) >> 3) & 0x10) |
		    ((UCHAR_ACCESS(&sPtr[4]) >> 4) & 0x08) |
		    ((UCHAR_ACCESS(&sPtr[5]) >> 5) & 0x04) |
		    ((UCHAR_ACCESS(&sPtr[6]) >> 6) & 0x02) |
		    ((UCHAR_ACCESS(&sPtr[7]) >> 7) & 0x01);
            }
		/* do the final, perhaps partial, 8-bit destination
		 * quantity.  do it one destination bit at a time,
		 * from left to right within the 8-bit quantity.
		 * continue picking the source bit as we did in the 
		 * inner loop.
		 */
	    if (wD & 7)
	    {
		oWord = 0;
		srcMask = wD & 0x7;
		for (col = srcMask; col > 0; col--, sPtr++)
		{
		    oWord |= (UCHAR_ACCESS(sPtr) & 0x80U) >> (srcMask - col);
		}
		UCHAR_ACCESS(dPtr) = oWord;
	    }
        }  
	break;     

    case 16:
	sInc = sBpl * 16;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
        {
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x7); col > 0; col -= 8, dPtr++, sPtr += 16)
            {
		UCHAR_ACCESS(dPtr) = (UInt8)
		    ((USHORT_ACCESS(sPtr) & 0x80) |
		     (USHORT_ACCESS(&sPtr[2]) & 0x40) |
		     (USHORT_ACCESS(&sPtr[4]) & 0x20) |
		     (USHORT_ACCESS(&sPtr[6]) & 0x10) |
		     (USHORT_ACCESS(&sPtr[8]) & 0x08) |
		     (USHORT_ACCESS(&sPtr[10]) & 0x04) |
		     (USHORT_ACCESS(&sPtr[12]) & 0x02) |
		     (USHORT_ACCESS(&sPtr[14]) & 0x01));
            }
	    if (wD & 7)
	    {
		oWord = 0;
		srcMask = 0x80;
		for (col = (wD & 0x7); col > 0; col--, sPtr += 2)
		{
		    oWord |= UCHAR_ACCESS(sPtr) & srcMask;
		    srcMask = srcMask >> 1;
		}
		UCHAR_ACCESS(dPtr) = oWord;
	    }
        }   
	break;    

    case 32:
	sInc = sBpl * 32;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
        {
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x7); col > 0; col -= 8, dPtr++, sPtr += 32)
            {
		UCHAR_ACCESS(dPtr) = (UInt8)
		    ((ULONG_ACCESS(sPtr) & 0x80) |
		     (ULONG_ACCESS(&sPtr[4]) & 0x40) |
		     (ULONG_ACCESS(&sPtr[8]) & 0x20) |
		     (ULONG_ACCESS(&sPtr[12]) & 0x10) |
		     (ULONG_ACCESS(&sPtr[16]) & 0x08) |
		     (ULONG_ACCESS(&sPtr[20]) & 0x04) |
		     (ULONG_ACCESS(&sPtr[24]) & 0x02) |
		     (ULONG_ACCESS(&sPtr[28]) & 0x01));
            }
	    if (wD & 7)
	    {
		oWord = 0;
		srcMask = 0x80;
		for (col = (wD & 0x7); col > 0; col--, sPtr += 4)
		{
		    oWord |= UCHAR_ACCESS(sPtr) & srcMask;
		    srcMask = srcMask >> 1;
		}
		UCHAR_ACCESS(dPtr) = oWord;
	    }
        } 
	break;
    }      

    NORMAL_RETURN;
}

/* PUBLIC */
/**************************************************************************
 *  i_subsample4bpp()
 *  	Computes simple subsampled reduction for 4 bit/pixel images
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_subsample4bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
PROLOGUE;
 register UInt8	*sPtr, *dPtr;
register Int32   row, col;
Int32		 sInc;
register UInt32	 iWord, oWord, oMask;
UInt32		 rightMask = 0xFFFFFFFF << (4*((8 - (wD & 0x7)) & 0x7));

	/* THIS IS RIDICULOUS!!!  gcc gives a warning:
	 * "warning: `oWord' may be used uninitialized in this function"
	 * when compiled with optimization turned on.  It doesn't give
	 * the warning when compiled in debugging mode.  Neither of the
	 * other compilers give the warning.  There is no reason for
	 * the warning.  BLAH!  But, just to keep things nice, we'll
	 * set oWord to 0 here.
	 */
    oWord = 0;

	/* adjust base pointers to avoid the frame */
    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    switch (factor)
    {
    case 1:	/* reduction by 1.5x in both X and Y (see comments for
		 * ip_subsample4bpp1_5xLine.) */
        for (row = hD; row > 0; row--, pD += dBpl, pS += sBpl)
	{
	    ip_subsample4bpp1_5xLine (pIPState, pD, wD, pS, sBpl, rightMask);
	    row--;
	    if (row > 0)    /* there's another line to do.  Point to the third
			     * input line as the major input and the line
			     * *above* it as the minor inout.
			     */
	    {
		pS += 2*sBpl;
		pD += dBpl;
		ip_subsample4bpp1_5xLine (pIPState, pD, wD, pS,
					  -sBpl, rightMask);
	    }
	}
	break;

    case 2:

	    /* In the algorithms below, we consolidate the nibbles in each
	     * 32-bit word read by repeated mask, shift and OR operations.
	     * Notation: X = desired nibble, t = trash, z = zero.
	     * Since we can't read beyond the frame with only 2x reduction,
	     * we just compute the result in units of full 32-bit words,
	     * then mask off the last word and write it again.  The mask
	     * equals 0xFFFFFFFF if wD MOD 8 = 0, 0xF0000000 if wD MOD 8 = 1
	     * 0xFF000000 if wD MOD 8 = 2, ... 0xFFFFFFF0 if wD MOD 8 = 7.
	     *
	     * NOTE: All these algorithms would be somewhat faster if we could
	     * just choose the most convenient set of pixels out of each
	     * word read.  For example, we might choose pixels 0 and 3
	     * from one word and pixels 2 and 5 from the next one.
	     * Unfortunately, this makes the result images look bad. It
	     * appears to be important to maintain a consistent distance
	     * between the input pixels that are chosen to be output pixels.
	     */
	sInc = sBpl * 2;
	oMask = 0xFFFFFFFF << (4 * ((8 - (wD & 0x7)) & 0x7));
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;

            for (col = wD; col > 0; col -= 8, dPtr += 4, sPtr += 8)
	    {
	        iWord = ULONG_ACCESS(sPtr) & 0xF0F0F0F0; /* produce XzXzXzXz */
		iWord |= iWord << 4;			 /* XXttXXtt */
		iWord &= 0xFF00FF00;			 /* XXzzXXzz */
		oWord = (iWord | (iWord << 8)) & 0xFFFF0000; /* XXXXzzzz */
		iWord = (ULONG_ACCESS(&sPtr[4]) >> 4)
				& 0x0F0F0F0F;		/* zXzXzXzX */
		iWord |= iWord >> 4;			/* ttXXttXX */
		iWord &= 0x00FF00FF;			/* zzXXzzXX */
		oWord |= (iWord | (iWord >> 8)) & 0x0000FFFF;
		ULONG_ACCESS(dPtr) = oWord;
	    }
		ULONG_ACCESS(&dPtr[-4]) = oWord & oMask; /* fix up last word */
	}
	break;

    case 4:
	sInc = sBpl * 4;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x7); col > 0; col -= 8, dPtr += 4, sPtr += 16)
	    {
		oWord = ULONG_ACCESS(sPtr) & 0xF000F000;  /* X1zzz X2zzz */
		oWord |= (ULONG_ACCESS(&sPtr[4]) >> 8)
				& 0x00F000F0; /* X1zX3zX2zX4z */

		oWord |= oWord << 12;		/* X1 X2 X3 X4..t z t z */
		oWord &= 0xFFFF0000;
		iWord = (ULONG_ACCESS(&sPtr[8]) >> 4)
				& 0x0F000F00; /* zX5zz zX6zz */
		iWord |= (ULONG_ACCESS(&sPtr[12]) >> 12)
				& 0x000F000F; /* zX5zX7zX6zX8 */
		iWord |= iWord >> 12;		/* z t z t X5 X6 X7 X8 */
		iWord &= 0x0000FFFF;
		ULONG_ACCESS(dPtr) = oWord | iWord;
	    }
		/* do the last word carefully */
	    if (wD & 0x7)
	    {
		for (col = (wD & 0x7), iWord = 0; col > 0; col--, sPtr += 2)
		{
		    iWord = (iWord << 4) |
			    ((UInt32) (USHORT_ACCESS(sPtr) & 0xF));
		}
		ULONG_ACCESS(dPtr) = iWord << (4 * (8 - (wD & 0x7)));
	    }
	}
	break;

    case 8:
	sInc = sBpl * 8;
	oMask = 0xF0000000;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x7); col > 0; col -= 8, dPtr += 4, sPtr += 32)
	    {
		oWord  =  ULONG_ACCESS(sPtr)      & oMask;
		oWord |= (ULONG_ACCESS(&sPtr[4])  & oMask) >>  4;
		oWord |= (ULONG_ACCESS(&sPtr[8])  & oMask) >>  8;
		oWord |= (ULONG_ACCESS(&sPtr[12]) & oMask) >> 12;
		oWord |= (ULONG_ACCESS(&sPtr[16]) & oMask) >> 16;
		oWord |= (ULONG_ACCESS(&sPtr[20]) & oMask) >> 20;
		oWord |= (ULONG_ACCESS(&sPtr[24]) & oMask) >> 24;
		oWord |= (ULONG_ACCESS(&sPtr[28]) & oMask) >> 28;
		ULONG_ACCESS(dPtr) = oWord;
	    }
		/* do the last word carefully */
	    if (wD & 0x7)
	    {
		for (col = (wD & 0x7), iWord = 0; col > 0; col--, sPtr += 4)
		{
		    iWord = (iWord << 4) | (ULONG_ACCESS(sPtr) & 0xF);
		}
		    ULONG_ACCESS(dPtr) = iWord << (4 * (8 - (wD & 0x7)));
	    }
	}
	break;

    case 16:
	sInc = sBpl * 16;
	oMask = 0xF0000000;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x7); col > 0; col -= 8, dPtr += 4, sPtr += 64)
	    {
		oWord  =  ULONG_ACCESS(sPtr)      & oMask;
		oWord |= (ULONG_ACCESS(&sPtr[8])  & oMask) >>  4;
		oWord |= (ULONG_ACCESS(&sPtr[16]) & oMask) >>  8;
		oWord |= (ULONG_ACCESS(&sPtr[24]) & oMask) >> 12;
		oWord |= (ULONG_ACCESS(&sPtr[32]) & oMask) >> 16;
		oWord |= (ULONG_ACCESS(&sPtr[40]) & oMask) >> 20;
		oWord |= (ULONG_ACCESS(&sPtr[48]) & oMask) >> 24;
		oWord |= (ULONG_ACCESS(&sPtr[56]) & oMask) >> 28;
		ULONG_ACCESS(dPtr) = oWord;
	    }
		/* do the last word carefully */
	    if (wD & 0x7)
	    {
		for (col = (wD & 0x7), iWord = 0; col > 0; col--, sPtr += 8)
		{
		    iWord = (iWord << 4) | (ULONG_ACCESS(sPtr) & 0xF);
		}
		ULONG_ACCESS(dPtr) = iWord << (4 * (8 - (wD & 0x7)));
	    }
	}
	break;

    case 32:
	sInc = sBpl * 32;
	oMask = 0xF0000000;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;


            for (col = (wD & ~0x7); col > 0; col -= 8, dPtr += 4, sPtr += 128)
	    {
		oWord  =  ULONG_ACCESS(sPtr)       & oMask;
		oWord |= (ULONG_ACCESS(&sPtr[16])  & oMask) >>  4;
		oWord |= (ULONG_ACCESS(&sPtr[32])  & oMask) >>  8;
		oWord |= (ULONG_ACCESS(&sPtr[48])  & oMask) >> 12;
		oWord |= (ULONG_ACCESS(&sPtr[64])  & oMask) >> 16;
		oWord |= (ULONG_ACCESS(&sPtr[80])  & oMask) >> 20;
		oWord |= (ULONG_ACCESS(&sPtr[96])  & oMask) >> 24;
		oWord |= (ULONG_ACCESS(&sPtr[112]) & oMask) >> 28;
		ULONG_ACCESS(dPtr) = oWord;
	    }
		/* do the last word carefully */
	    if (wD & 0x7)
	    {
		for (col = (wD & 0x7), iWord = 0; col > 0; col--, sPtr += 16)
		{
		    iWord = (iWord << 4) | (ULONG_ACCESS(sPtr) & 0xF);
		}
		ULONG_ACCESS(dPtr) = iWord << (4 * (8 - (wD & 0x7)));
	    }
	}
	break;
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/**************************************************************************
 *  i_subsample8bpp()
 *  	Computes simple subsampled reduction for 8 bit/pixel images
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 1 (which represents 1.5)
 *			 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_subsample8bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
PROLOGUE;
 register UInt8	*sPtr, *dPtr;
register Int32   row, col;
Int32		 sInc;
register UInt32	 oWord;
UInt32		 rightMask = 0xFFFFFFFF << (8*((4 - (wD & 0x3)) & 0x3));

	/* adjust base pointers to avoid the frame */
    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 8*cFrameBytes;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    switch (factor)
    {
    case 1:	/* reduction by 1.5x in both X and Y (see comments above) */
        for (row = hD; row > 0; row--, pD += dBpl, pS += sBpl)
	{
	    ip_subsample8bpp1_5xLine (pIPState, pD, wD, pS, sBpl, rightMask);
	    row--;
	    if (row > 0)    /* there's another line to do.  Point to the third
			     * input line as the major input and the line
			     * *above* it as the minor inout.
			     */
	    {
		pS += 2*sBpl;
		pD += dBpl;
		ip_subsample8bpp1_5xLine (pIPState, pD, wD, pS,
					  -sBpl, rightMask);
	    }
	}
	break;

    case 2:
	sInc = sBpl * 2;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x3); col > 0; col -= 4, dPtr += 4, sPtr += 8)
	    {
		oWord = ULONG_ACCESS(sPtr) & 0xFF000000;
		oWord |= (ULONG_ACCESS(sPtr) & 0x0000FF00) << 8;
		oWord |= (ULONG_ACCESS(&sPtr[4]) & 0xFF000000) >> 16;
		oWord |= (ULONG_ACCESS(&sPtr[4]) & 0x0000FF00) >> 8;
		ULONG_ACCESS(dPtr) = oWord;
	    }
	    for (col = (wD & 0x3); col > 0; col--, dPtr++, sPtr += 2)
	    {
		UCHAR_ACCESS(dPtr) = UCHAR_ACCESS(sPtr);
	    }
	}
	break;

    case 4:
	sInc = sBpl * 4;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x3); col > 0; col -= 4, dPtr += 4, sPtr += 16)
	    {
		oWord = ULONG_ACCESS(sPtr) & 0xFF000000;
		oWord |= (ULONG_ACCESS(&sPtr[4]) & 0xFF000000) >> 8;
		oWord |= (ULONG_ACCESS(&sPtr[8]) & 0xFF000000) >> 16;
		oWord |= (ULONG_ACCESS(&sPtr[12]) & 0xFF000000) >> 24;
		ULONG_ACCESS(dPtr) = oWord;
	    }
	    for (col = (wD & 0x3); col > 0; col--, dPtr++, sPtr += 4)
	    {
		UCHAR_ACCESS(dPtr) = UCHAR_ACCESS(sPtr);
	    }
	}
	break;

    case 8:
	sInc = sBpl * 8;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x3); col > 0; col -= 4, dPtr += 4, sPtr += 32)
	    {
		oWord = ULONG_ACCESS(sPtr) & 0xFF000000;
		oWord |= (ULONG_ACCESS(&sPtr[8]) & 0xFF000000) >> 8;
		oWord |= (ULONG_ACCESS(&sPtr[16]) & 0xFF000000) >> 16;
		oWord |= (ULONG_ACCESS(&sPtr[24]) & 0xFF000000) >> 24;
		ULONG_ACCESS(dPtr) = oWord;
	    }
	    for (col = (wD & 0x3); col > 0; col--, dPtr++, sPtr += 8)
	    {
		UCHAR_ACCESS(dPtr) = UCHAR_ACCESS(sPtr);
	    }
	}
	break;

    case 16:
	sInc = sBpl * 16;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~0x3); col > 0; col -= 4, dPtr += 4, sPtr += 64)
	    {
		oWord = ULONG_ACCESS(sPtr) & 0xFF000000;
		oWord |= (ULONG_ACCESS(&sPtr[16]) & 0xFF000000) >> 8;
		oWord |= (ULONG_ACCESS(&sPtr[32]) & 0xFF000000) >> 16;
		oWord |= (ULONG_ACCESS(&sPtr[48]) & 0xFF000000) >> 24;
		ULONG_ACCESS(dPtr) = oWord;
	    }
	    for (col = (wD & 0x3); col > 0; col--, dPtr++, sPtr += 16)
	    {
		UCHAR_ACCESS(dPtr) = UCHAR_ACCESS(sPtr);
	    }
	}
	break;

    case 32:
	sInc = sBpl * 32;
        for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	{
            dPtr = pD;
            sPtr = pS;


            for (col = (wD & ~0x3); col > 0; col -= 4, dPtr += 4, sPtr += 128)
	    {
		oWord = ULONG_ACCESS(sPtr) & 0xFF000000;
		oWord |= (ULONG_ACCESS(&sPtr[32]) & 0xFF000000) >> 8;
		oWord |= (ULONG_ACCESS(&sPtr[64]) & 0xFF000000) >> 16;
		oWord |= (ULONG_ACCESS(&sPtr[96]) & 0xFF000000) >> 24;
		ULONG_ACCESS(dPtr) = oWord;
	    }
	    for (col = (wD & 0x3); col > 0; col--, dPtr++, sPtr += 32)
	    {
		UCHAR_ACCESS(dPtr) = UCHAR_ACCESS(sPtr);
	    }
	}
	break;
    }

    NORMAL_RETURN;
}


/**************************************************************************
 * ip_subsample4bpp1_5xLine
 *
 * To reduce an image by a factor of 1.5 in both X and Y, you have to
 * turn a 3x3 square of pixels into a 2x2 square of pixels.  There is
 * no consistent set of 4 pixels you can pick out of every 3x3 square
 * that produces nice results.  For example, say you always take the
 * corner pixels of the 3x3 square.  You'll end up taking 2 pixels,
 * then skipping 1, taking two, skipping one, etc.  The sampling is not
 * regular.  This has a particularly bad effect on text where strokes
 * that were supposed to be the same width either get fat (when both
 * pixels chosen were in the stroke) or thin (when one of the stroke
 * pixels was skipped).
 *
 * However, the fact we're trying to do subsampling means that the emphasis
 * is on speed, not exquisite quality.  Say the following transformation
 * takes place:
 *
 *	In0  In1  In2	  Out0	Out1
 *	In3  In4  In5 =>
 *	In6  In7  In8	  Out2	Out3
 *
 * Say calculate the output pixels as follows:
 *	Out0 = In0/2 + In1/4 + In3/4
 *	Out1 = In2/2 + In1/4 + In5/4
 *	Out2 = In6/2 + In7/2 + In3/4
 *	Out3 = In8/2 + In7/2 + In5/4
 *
 * This is fairly easy and doesn't result in overflow.  The sampling is
 * sufficiently regular to produce reasonable results.
 * If all the InX pixels are equal to 0, the resulting output pixel is 0.
 * If all the InX pixels are equal to 0xF, the output pixel is 0xF.  However,
 * this is true ONLY if we retain all the bits from the inputs, converting
 * them to fraction bits.  Bummer.  Since dividing by 4 implies the use of
 * two fraction bits, we'd need 6-bit pixels.  The 4-bit pixels are packed
 * in words with no space between them.
 *
 * What do do?  Spreading the pixels in a word apart, doing the calculation,
 * then putting them back together is takes a long time.  Keeping the fraction
 * bits in another register is better, but this means we have to do everything
 * twice, once for the integer part and once for the fraction part.  The
 * best idea I've been able to think of is to put the even numbered pixels in
 * one register and the odd pixels in another.  The space around these
 * isolated pixels is used for fraction bits and carries.  We don't have to do
 * everything twice since pixels that are shifted right are only combined with
 * the pixels in one register and those shifted left are only combined with
 * the other register.
 *
 * As we work our way from left to right across a scan line, we will
 * access 32-bit words in groups of 3.  Say we tag the nibbles in the 
 * corners of the 3x3 squares with the numbers of the *output* nibbles:
 *
 * Word 0:	| N0  |     | N1  | N2  ||     | N3  | N4  |     | 
 * Word 1:	| N5  | N6  |     | N7  || N8  |     | N9  | N10 |
 * Word 2:	|     | N11 | N12 |     || N13 | N14 |     | N15 |
 *
 * The next word repeats the pattern started by Word 0.
 *
 * We read word 0, shift it right 1 and put nibbles 1, 2 and 4 into evenNibs
 * and nibbles 1 and 3 into oddNibs.  This gives us space around each nibble
 * for fraction bits to the right and carries to the left.  The resulting
 * nibbles will end up in the correct spots since the adds will move everything
 * one place to the left.
 *
 * evenNibs:	| N0  |     |     | N2  ||     |     | N4  |     |
 * oddNibs:	|     |     | N1  |     ||     | N3  |     |     | 
 *
 * We select out the 3 nibbles represented by blanks above from the original
 * word read.  These are the pixels that are between the corner pixels, or
 * In1 in the figure above.
 * We shift the result left by 2 and add it to evenNibs.  This corresponds
 * to In0/2 + In1/4 step.  We shift the same register right by 6 and add it
 * to oddNibs.  This corresponds to the In1/4 + In2/2 step above.  We're
 * almost done with the horizontal stuff.  However, pixel N0 needs a
 * contribution from its left.  We get this by reading the byte to its left,
 * masking off the upper nibble, shifting it into position (left by 26) and
 * adding it to evenNibs.
 *
 * To add in the vertical contributions, we access the word below (or
 * above is sBpl is negative) and get rid of all the nibbles represented by
 * blanks in Word 0.  The result is shifted right by 2 because each nibble 
 * represents In3/4 (or In5/4) above.  We then select only even nibbles and
 * add it to evenNibs and select only the odd nibbles and add them to
 * oddNibs.
 *
 * At this point, the nibbles are fully formed, though they have fraction
 * bits hanging off to the right.  I can't find any marvelously clever way
 * to collapse all the nibbles into the string N0 | N1 | N2 | N3 | N4.
 * We'll just strip the fraction bits off both sets of nibbles and combine
 * them.  Then use two shift-mask-mask-or sequences to shift the nibbles
 * into position.
 *
 * Word 1 is handled in a similar way, except evenNibs is shifted right by 1
 * *and* oddNibs is shifted left by 1 so the nibble on the end have room.
 * When we're done, we shift N5 and N6 into position in Word0, then insert
 * N8.  We then consolidate the remaining 3 nibbles in the top of the word.
 *
 * Finally, we process Word 2 similarly.  Here evenNibs is not shifted and
 * oddNibs is shifted left by 1.  After processing, we put the 5 nibbles into
 * the 5 LS nibbles of the word, combine it with the 3 nibbles found
 * previously and write out the whole thing.
 *
 * This whole process seems complicated because it is.  Sorry. every time I
 * tried a nice simple scheme that read a byte and produce a pixel or two,
 * the computation time shot up.  The advantage of this scheme is that we
 * process 5 or 6 pixels at a time.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_subsample4bpp1_5xLine()  
 *		Each 3x3 square of the source binary image is mapped
 *		to a 2x2 square of 4-bit output pixels.  The values
 *		of the output pixels are a linear function of
 *		the corner pixels and the pixels that are horizontally
 *		and vertically adjacent to the corner pixels.
 *
 *	Input:
 *		pD =		pointer to upper left corner of dest image
 *				(includes frame)
 *		wD =		width in pixels of dest (including frame)
 *		pS =		pointer to upper left corner of source image
 *				(includes frame)
 *		sBpl =		source image bytes/line
 *		rightMask =	used to set unused pixel on the right
 *				end of the line to zero.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_subsample4bpp1_5xLine (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    UInt8      *pS,
    Int32	sBpl,
    UInt32	rightMask)
{
    PROLOGUE;
    UInt32	oWord;			/* holds the nibbles to be put out */
    UInt32	iWord, evenNibs, oddNibs; /* input word and nibbles being
					   * computed */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Prepare to loop over the words in the output line.
	 * The images is always left aligned.  The last word on the 
	 * right is masked with rightMask.  We start by reading the
	 * first 5 pixels of the row.
	 * Each output pixel is composed of inputs from the major and
	 * minor lines as described in the comment above.
	 * We'll start with the following pattern of input Nibbles:
	 * | N0  |      | N1  | N2  ||     | N3  | N4  |     |
	 */
    iWord = ULONG_ACCESS(pS);		/* can't shift yet, need to keep
					 * the LS bit so it can be combined
					 * with N4. */
    oddNibs = iWord >> 1;		/* make room for carries */
    evenNibs = oddNibs & 0x78078078;	/* select N0, N2, N4 */
    oddNibs = oddNibs & 0x00780780;	/* select N1, N3 */

	/* evenNibs: 0xxx | x000 | 0000 | 0xxx | x000 | 0000 | 0xxx | x000
	 * oddNibs : 0000 | 0000 | 0xxx | x000 | 0000 | 0xxx | x000 | 0000
	 * Get the nibbles in the blanks and combine them with the
	 * even and and odd nibbles.  They are combined with the even
	 * nibbles on the left and the odd nibbles on the right.
	 * Note that we need to shift left by 4 (to get to next nibble) -
	 * 1 (nibs have already been shifted right by 1) - 1 (the
	 * correction nibbles should be shifted right by 1 relative to the
	 * sum nibbles) = 2.  Similarly, we need to shift right by
	 * 4 + 1 + 1 = 6.
	 */
    iWord &= 0x0F00F00F;		/* get the blanks */
    evenNibs += iWord << 2;
    oddNibs += iWord >> 6;

	/* get the contribution from below next.  As before, these pixels
	 * should be shifted right by *2* so they can represent In3/4 and
	 * In5/4 in the comment above.
	 */
    iWord = ULONG_ACCESS(&pS[sBpl]) >> 2;
    evenNibs += (iWord & 0x3C03C03C);
    oddNibs += (iWord & 0x003C03C0);

	/* N0 needs no contribution from a pixel to the left.
	 * The pixels are fully formed.  Compact them into the upper
	 * 5 nibbles of oWord.
	 */
    evenNibs &= 0xF00F00F0;		/* strip off fraction bits */
    oddNibs  &= 0x00F00F00;
    oWord = evenNibs | oddNibs;		/* get 0xN0NN0NN0 */
    oWord = (oWord & 0xF0000000) | ((oWord & 0x00FF0000) << 4) |
	    ((oWord & 0x00000FF0) << 8);


	/* Get the second set of nibbles in Word 2.  These are patterned as
	 * | N5  | N6  |     | N7  || N8  |     | N9  | N10 |
	 */
    pS += 4;
    iWord = ULONG_ACCESS(pS);		/* can't shift yet, need to keep
					 * the LS bit so it can be combined
					 * with N4. */
    oddNibs = (iWord & 0xF00F00F0) >> 1; /* select the odd nibbles */
    evenNibs = (iWord & 0x0F00F00F) << 1; /* select the even nibbles, leave
					 * space for 2 fraction nibbles.
					 * But didn't we only shift left by
					 * 1?  Yes, but remember that puts the
					 * binary point is above bit 1 since
					 * we are pretending the original
					 * pixel values have already been
					 * divided by 2, i.e.
					 * the final sums will appear
					 * shifted to the left of the
					 * original positions of the
					 * bits in oddNibs and evenNibs.
					 */

	/* generate the contribution from the two blank nibbles.
	 * Since the evenNibs have been shifted left by 1 instead of right
	 * by 1 as above, we need to shift the blank nibs left by 4 + 1 - 1.
	 * Since the odd Nibs have been shifted right by 1 as before,
	 * we still shift right by 6.
	 */
    iWord &= 0x00F00F00;
    evenNibs += iWord << 4;
    oddNibs += iWord >> 6;

	/* get contribution from the word to the left for N5. */
    iWord = UCHAR_ACCESS(&pS[-1]);
    oddNibs += (iWord & 0xF) << 26;

	/* get contribution from word below.  Since the
	 * evenNibs have already been shifted left by 1, we can combine the
	 * pixels from below without shifting them.  Since the odd nibs have
	 * been shifted right by 1, we must shift the pixels from below
	 * right by 2.
	 */
    iWord = ULONG_ACCESS(&pS[sBpl]);
    evenNibs += (iWord & 0x0F00F00F);
    oddNibs += (iWord & 0xF00F00F0) >> 2;

	/* Get contribution from byte to the right of N10. */
    iWord = UCHAR_ACCESS(&pS[4]);
    evenNibs += (iWord & 0xF0) >> 4;

	/* These six pixels are fully formed.  Strip off the fraction bits
	 * and insert N5, N6 and N7 into oWord.
	 */
    oddNibs  &= 0xF00F00F0;
    evenNibs &= 0x3C03C03C;	/* remember evenNibs was shifted left by 1,
				 * so the result pixels are shifted left
				 * by 2 after all this addition */
    oddNibs |= (evenNibs >> 2);
    oWord |= (oddNibs >> 20) | ((oddNibs >> 16) & 0xF);


	/* Now do the real loop.  We start by writing the word we
	 * already have so the last time through the loop we can
	 * fall out the bottom with the last word.  If this is the last
	 * time through the loop, we'll mask it
	 * before writing it.  The loop has enough sections so that
	 * when we get back up to the top, oWord contains a complete
	 * output word, oddNibs contains three output nibbles.
	 */
    for (; ; wD -= 16)
    {
	if (wD <= 8)
	{
	    ULONG_ACCESS(pD) = oWord & rightMask;
	    break;
	}

	ULONG_ACCESS(pD) = oWord;
	pD += 4;

	    /* Consolidate the 3 remaining output nibbles in oWord */
	oWord = ((oddNibs & 0x0000F000) << 16) | (oddNibs << 20);

	    /* Get the last set of nibbles in Word 3.  These are patterned as
	     * |     | N11 | N12 |     || N13 | N14 |     | N15 |
	     */
	pS += 4;
	iWord = ULONG_ACCESS(pS);
	oddNibs = iWord << 1;		    /* make room for frac bits
					     * below N15 */
	evenNibs = (oddNibs & 0x01E01E00);  /* select the even nibbles*/
	oddNibs =  (oddNibs & 0x1E01E01E);  /* select the odd nibbles */

	    /* generate the contribution from the three blank nibbles.
	     * Since the both evenNibs and oddNibs have been shifted left by 1,
	     * we need to shift the blank nibs left by 4 + 1 - 1 for evenNibs
	     * and right by 4 - 1 + 1
	     */
	iWord &= 0xF00F00F0;
	evenNibs += iWord << 4;
	oddNibs += iWord >> 4;
    
	    /* get the contribution from below next.  As before, these pixels
	     * need not be shifted since the sums have already been shifted
	     * left by 1.  That leaves these pixels properly aligned.
	     */
	iWord = ULONG_ACCESS(&pS[sBpl]);
	evenNibs += (iWord & 0x00F00F00);
	oddNibs += (iWord & 0x0F00F00F);
    
	    /* The byte to the right of N15 doesn't contribute to N15, so
	     * ignore it.  At this point, all the pixels are fully formed.
	     * Strip their fraction bits and consolidate them into the
	     * bottom 5 nibbles of oWord.
	     */
	evenNibs &= 0x03C03C00;		/* answers are shifte left by 2 */
	oddNibs  &= 0x3C03C03C;
	oddNibs |= evenNibs;
	oddNibs >>= 2;			/* Align it.  We need three shifts
					 * no matter how we do it. */ 

	    /* The nibbles in bits 27:20 go to 19:12, the nibbles in
	     * 15:8 go to 11:4; the LS nibble stays where it is.
	     */
	oWord |= ((oddNibs >> 8) & 0x000FF000) | ((oddNibs >> 4) & 0x0FF0) |
		 (oddNibs & 0xF);

	    /* if that's the last word in the line, quit now.  Otherwise,
	     * write it to the output and continue.
	     */
	if (wD <= 16)
	{
	    ULONG_ACCESS(pD) = oWord & rightMask;
	    break;
	}

	ULONG_ACCESS(pD) = oWord;
	pD += 4;

	    /* Repeat the code used to set up the loop.  We start by reading
	     * the first 5 pixels of the row.
	     * Each output pixel is composed of inputs from the major and
	     * minor lines as described in the comment above.
	     * So, we'll start with the following pattern of input Nibbles:
	     * | N0  |      | N1  | N2  ||     | N3  | N4  |     |
	     */
	pS += 4;
	iWord = ULONG_ACCESS(pS);	    /* can't shift yet, need to keep
					     * the LS bit so it can be combined
					     * with N4. */
	oddNibs = iWord >> 1;		    /* make room for carries */
	evenNibs = oddNibs & 0x78078078;    /* select N0, N2, N4 */
	oddNibs = oddNibs & 0x00780780;	    /* select N1, N3 */
    
	    /* evenNibs: 0xxx | x000 | 0000 | 0xxx | x000 | 0000 | 0xxx | x000
	     * oddNibs : 0000 | 0000 | 0xxx | x000 | 0000 | 0xxx | x000 | 0000
	     * Get the nibbles in the blanks and combine them with the
	     * even and and odd nibbles.  They are combined with the even
	     * nibbles on the left and the odd nibbles on the right.
	     * Note that we need to shift left by 4 (to get to next nibble) -
	     * 1 (nibs have already been shifted right by 1) - 1 (the
	     * correction nibbles should be shifted right by 1 relative to the
	     * sum nibbles) = 2.  Similarly, we need to shift right by
	     * 4 + 1 + 1 = 6.
	     */
	iWord &= 0x0F00F00F;		/* get the blanks */
	evenNibs += iWord << 2;
	oddNibs += iWord >> 6;
    
	    /* get the contribution from below next.  As before, these pixels
	     * should be shifted right by *2* so they can represent In3/4 and
	     * In5/4 in the comment above.
	     */
	iWord = ULONG_ACCESS(&pS[sBpl]) >> 2;
	evenNibs += (iWord & 0x3C03C03C);
	oddNibs += (iWord & 0x003C03C0);
    
	    /* N0 needs no contribution from the pixel to the left.
	     * The pixels are fully formed.  Compact them into the upper
	     * 5 nibbles of oWord.
	     */
	evenNibs &= 0xF00F00F0;		/* strip off fraction bits */
	oddNibs  &= 0x00F00F00;
	oWord = evenNibs | oddNibs;		/* get 0xN0NN0NN0 */
	oWord = (oWord & 0xF0000000) | ((oWord & 0x00FF0000) << 4) |
		((oWord & 0x00000FF0) << 8);
    
    
	    /* Get the second set of nibbles in Word 2.  These are patterned as
	     * | N5  | N6  |     | N7  || N8  |     | N9  | N10 |
	     */
	pS += 4;
	iWord = ULONG_ACCESS(pS);	    /* can't shift yet, need to keep
					     * the LS bit so it can be combined
					     * with N4. */
	oddNibs = (iWord & 0xF00F00F0) >> 1; /* select the odd nibbles */
	evenNibs = (iWord & 0x0F00F00F) << 1; /* select the even nibbles, leave
					     * space for 2 fraction nibbles.
					     * But didn't we only shift left
					     * by 1?  Yes, but remember that
					     * puts the binary point is above
					     * bit 1 since we are pretending
					     * the original pixel values have
					     * already been divided by 2, i.e.
					     * the final sums will appear
					     * shifted to the left of the
					     * original positions of the
					     * bits in oddNibs and evenNibs.
					     */
    
	    /* generate the contribution from the two blank nibbles.
	     * Since the evenNibs have been shifted left by 1 instead of right
	     * by 1 as above, we need to shift the blank nibs left by 4 + 1 - 1.
	     * Since the odd Nibs have been shifted right by 1 as before,
	     * we still shift right by 6.
	     */
	iWord &= 0x00F00F00;
	evenNibs += iWord << 4;
	oddNibs += iWord >> 6;
    
	    /* get contribution from the word to the left for N5. */
	iWord = UCHAR_ACCESS(&pS[-1]);
	oddNibs += (iWord & 0xF) << 26;
    
	    /* get contribution from word below.  Since the
	     * evenNibs have already been shifted left by 1, we can combine the
	     * pixels from below without shifting them.  Since the odd nibs have
	     * been shifted right by 1, we must shift the pixels from below
	     * right by 2.
	     */
	iWord = ULONG_ACCESS(&pS[sBpl]);
	evenNibs += (iWord & 0x0F00F00F);
	oddNibs += (iWord & 0xF00F00F0) >> 2;
    
	    /* Get contribution from byte to the right of N10.
	     */
	iWord = UCHAR_ACCESS(&pS[4]);
	evenNibs += (iWord & 0xF0) >> 4;
    
	    /* These six pixels are fully formed.  Strip off the fraction bits
	     * and insert N5, N6 and N7 into oWord.
	     */
	oddNibs  &= 0xF00F00F0;
	evenNibs &= 0x3C03C03C;	    /* remember evenNibs was shifted left by 1,
				     * so the result pixels are shifted left
				     * by 2 after all this addition
				     */
	oddNibs |= (evenNibs >> 2);
	oWord |= (oddNibs >> 20) | ((oddNibs >> 16) & 0xF);
    }
}



/**************************************************************************
 * ip_subsample8bpp1_5xLine
 *
 * To reduce an image by a factor of 1.5 in both X and Y, you have to
 * turn a 3x3 square of pixels into a 2x2 square of pixels.  There is
 * no consistent set of 4 pixels you can pick out of every 3x3 square
 * that produces nice results.  For example, say you always take the
 * corner pixels of the 3x3 square.  You'll end up taking 2 pixels,
 * then skipping 1, taking two, skipping one, etc.  The sampling is not
 * regular.  This has a particularly bad effect on text where strokes
 * that were supposed to be the same width either get fat (when both
 * pixels chosen were in the stroke) or thin (when one of the stroke
 * pixels was skipped).
 *
 * However, the fact we're trying to do subsampling means that the emphasis
 * is on speed, not exquisite quality.  Say the following transformation
 * takes place:
 *
 *	In0  In1  In2	  Out0	Out1
 *	In3  In4  In5 =>
 *	In6  In7  In8	  Out2	Out3
 *
 * Say calculate the output pixels as follows:
 *	Out0 = In0/2 + In1/4 + In3/4
 *	Out1 = In2/2 + In1/4 + In5/4
 *	Out2 = In6/2 + In7/2 + In3/4
 *	Out3 = In8/2 + In7/2 + In5/4
 *
 * This is fairly easy and doesn't result in overflow.  The sampling is
 * sufficiently regular to produce reasonable results.
 * If all the InX pixels are equal to 0, the resulting output pixel is 0.
 * If all the InX pixels are equal to 0xFF, the output pixel is 0xFD.  This
 * is a little bit off, but the error is small compared to 0xFF.
 *
 * As we work our way from left to right across a scan line, we will
 * access 32-bit words in groups of 3.  Say we tag the bytes in the 
 * corners of the 3x3 squares with the numbers of the *output* bytes:
 *
 * Word 0:	Byte0 | Byte1 |       | Byte2
 * Word 1:	Byte3 |       | Byte4 | Byte5
 * Word 2:	      | Byte6 | Byte7 |
 *
 * The next word repeats the pattern started by Word 0.
 *
 * We read word 0 and the word below it.  We shift the word below it right
 * by 2 (/4), mask the result and add it to Word 0.  This completes the
 * vertical contribution.  We then shift the unused byte in word0 right by 9
 * add it in, shift the shifted byte left by 16 and add it it.  Finally, we
 * mask the byte to the right of Word0  by 0xFC, shift it left by 22 (/4
 * and aligned with Byte0) and add it in.  This completes the calcalation of
 * Word 0.  Byte2 is shifted up into the blank spot.
 *
 * We proceed with a similar plan for Word 1.  When complete, we put Byte 3
 * into Word 0 and write it to memory.  We then shift Bytes 4:5 to the top
 * of the word.
 *
 * Finally, we perform roughly the same algorithm on Word 2, shift the
 * results right by 8, combine it with Word 2 and write the result.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_subsample8bpp1_5xLine()  
 *		Each 3x3 square of the source binary image is mapped
 *		to a 2x2 square of 4-bit output pixels.  The values
 *		of the output pixels are a linear function of
 *		the corner pixels and the pixels that are horizontally
 *		and vertically adjacent to the corner pixels.
 *
 *	Input:
 *		pD =		pointer to upper left corner of dest image
 *				(includes frame)
 *		wD =		width in pixels of dest (including frame)
 *		pS =		pointer to upper left corner of source image
 *				(includes frame)
 *		sBpl =		source image bytes/line
 *		rightMask =	used to set unused pixel on the right
 *				end of the line to zero.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_subsample8bpp1_5xLine (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    UInt8      *pS,
    Int32	sBpl,
    UInt32	rightMask)
{
    PROLOGUE;
    UInt32	oWord1, oWord2, iWord;	/* the output and input words */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Prepare to loop over the words in the output line.
	 * The images is always left aligned.  The last word on the 
	 * right is masked with rightMask.  We start by reading the
	 * first 6 pixels of the row.
	 * Each output pixel is composed of inputs from the major and
	 * minor lines as described in the comment above.
	 * Since we're starting at the left edge of the line, we can't
	 * ask for contributions from the left.  So, we'll start with
	 * the following pattern of input bytes:
	 * | Byte0 |       | Byte1 | Byte2 |
	 */
    oWord1 = (ULONG_ACCESS(pS) >> 1) & 0x7F007F7F;	/* get corner pixels
							 * /2 */
    iWord = (ULONG_ACCESS(&pS[sBpl]) >> 2) & 0x3F003F3F; /* get word below /4 */
    oWord1 += iWord;

	/* Distribute 25% of byte1 left and right */
    iWord = UCHAR_ACCESS(&pS[1]) & 0xFC;		/* do premask */
    oWord1 += (iWord << 22) + (iWord << 6);		/* add to bytes 0, 2 */
	/* Distribute 25% of byte4 left */
    iWord = UCHAR_ACCESS(&pS[4]) & 0xFC;
    oWord1 += iWord >> 2;

	/* Consolidate the three bytes into the top three bytes of
	 * the word */
    oWord1 = (oWord1 & 0xFF000000) | (oWord1 << 8);

	/* get two more bytes so we can form a complete output word.
	 * The pattern of useful stuff in the next word is
	 * |       | Byte3 | Byte4 |       |
	 */
    pS += 4;
    oWord2 = (ULONG_ACCESS(pS) >> 1) & 0x007F7F00;
    oWord2 += (iWord << 14);				/* add byte 0 */
    iWord = (ULONG_ACCESS(&pS[sBpl]) >> 2) & 0x003F3F00;
    oWord2 += iWord;

	/* Distribute 25% of byte 3 left */
    iWord = UCHAR_ACCESS(&pS[3]) & 0xFC;
    oWord2 += (iWord << 6);
    oWord1 |= (oWord2 >> 16);		/* insert byte 3 into oWord1,
					 * completing oWord1 */
    oWord2 <<= 16;			/* align remaining byte */


	/* Now do the real loop.  We start by writing the word we
	 * already have so the last time through the loop we can
	 * fall out the bottom with the last word.  If this is the last
	 * time through the loop, we'll mask it
	 * before writing it.  The loop has enough sections so that
	 * when we get back up to the top, oWord1 contains a complete
	 * output word, oWord2 contains one output byte in the MS byte
	 * and iWord contains a masked version of the byte at pS[3].
	 */
    for (; ; wD -= 8)
    {
	if (wD <= 4)
	{
	    ULONG_ACCESS(pD) = oWord1 & rightMask;
	    break;
	}

	ULONG_ACCESS(pD) = oWord1;
	pD += 4;

	    /* The pattern of data in this word is:
	     * | Byte5 | Byte6 |       | Byte7 |
	     */
	pS += 4;
	oWord1 = (ULONG_ACCESS(pS) >> 1) & 0x7F7F007F;
	oWord1 += (iWord << 22);	/* apply byte from the left */
	iWord = (ULONG_ACCESS(&pS[sBpl]) >> 2) & 0x3F3F003F;
	oWord1 += iWord;

	    /* Distribute 25% of byte 2 left and right */
	iWord = UCHAR_ACCESS(&pS[2]) & 0xFC;
	oWord1 += (iWord << 14) + (iWord >> 2);
	    /* insert the 3 bytes into the 3 LS bytes of oWord2 */
	oWord2 |= (oWord1 >> 8) | (oWord1 & 0xFF);


	    /* if that's the last word in the line, quit now.  Otherwise,
	     * write it to the output and continue.
	     */
	if (wD <= 8)
	{
	    ULONG_ACCESS(pD) = oWord2 & rightMask;
	    break;
	}

	ULONG_ACCESS(pD) = oWord2;
	pD += 4;

	    /* Repeat the code used to set up the loop.  The pattern is:
	     * | Byte0 |       | Byte1 | Byte2 |
	     */
	pS += 4;
	oWord1 = (ULONG_ACCESS(pS) >> 1) & 0x7F007F7F;	/* get corner pixels
							 * /2 */
	iWord = (ULONG_ACCESS(&pS[sBpl]) >> 2) & 0x3F003F3F; /* get word below
							 * /4 */
	oWord1 += iWord;
    
	    /* Distribute 25% of byte1 left and right */
	iWord = UCHAR_ACCESS(&pS[1]) & 0xFC;		/* do premask */
	oWord1 += (iWord << 22) + (iWord << 6);		/* add to bytes 0, 2 */

	    /* Distribute 25% of byte4 left */
	iWord = UCHAR_ACCESS(&pS[4]) & 0xFC;
	oWord1 += iWord >> 2;

	    /* Consolidate the three bytes into the top three bytes of
	     * the word */
	oWord1 = (oWord1 & 0xFF000000) | (oWord1 << 8);

	    /* get two more bytes so we can form a complete output word.
	     * The pattern of useful stuff in the next word is
	     * |	| Byte3 | Byte4 |       |
	     */
	pS += 4;
	oWord2 = (ULONG_ACCESS(pS) >> 1) & 0x007F7F00;
	oWord2 += (iWord << 14);			/* add byte 0 */
	iWord = (ULONG_ACCESS(&pS[sBpl]) >> 2) & 0x003F3F00;
	oWord2 += iWord;
    
	    /* Distribute 25% of byte 3 left */
	iWord = UCHAR_ACCESS(&pS[3]) & 0xFC;
	oWord2 += (iWord << 6);
	oWord1 |= (oWord2 >> 16);		/* insert byte 3 into oWord1,
						 * completing oWord1 */
	oWord2 <<= 16;				/* align remaining byte */
    }

}

/*

$Log: i_sbsmp1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.4  1998/01/13 23:49:36  ddavies
i_sbsmp1.c -> /product/ipcore/ipshared/src/RCS/i_sbsmp1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1995/12/13 02:45:50  ejaquith
i_sbsmp1.c -> /product/ipcore/ipshared/src/RCS/i_sbsmp1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/10/18  03:58:34  ddavies
 * i_sbsmp1.c -> /product/ipcore/ipshared/src/RCS/i_sbsmp1.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.1  1994/09/09  00:18:02  lperry
 * Initial revision
 *
 * Revision 1.5  1994/08/01  21:06:28  ddavies
 * i_subsmp.c -> /product/ipcore/ipshared/src/RCS/i_subsmp.c,v
 *
 * Change "static const" to "static" to work around buggy gcc/Purify interaction.
 *
 * Revision 1.4  1994/07/27  16:07:14  ejaquith
 * i_subsmp.c -> /product/ipcore/ipshared/src/RCS/i_subsmp.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Made functions use local pointers to tables instead of globals.
 *
 * Revision 1.3  1994/07/09  20:36:31  jcraig
 * jcraig on Sat Jul 9 13:26:40 PDT 1994
 *
 * i_subsmp.c -> /product/ipcore/ipshared/src/RCS/i_subsmp.c,v
 *
 * in i_subsample1bpp when reducing by 2 or 4, mask the last byte of
 * the line so that no extra bits will end up in the frame if the source
 * width isn't a multiple of the reduction factor.
 * in i_reduceVOrHAnd and i_reduceVAndHOr use CALLOC instead of MALLOC
 * so that purify won't complain about accessing uninitialized memory
 * (even though the word is never used).
 * 2624
 *
 * Revision 1.2  1994/02/04  20:17:53  dferrell
 * dferrell on Fri Feb 4 12:17:43 PST 1994
 *
 * i_subsmp.c -> /product/ipcore/ipshared/src/RCS/i_subsmp.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/02/03  00:11:46  ddavies
 * Initial revision
 *
 * Revision 1.19  1993/12/21  23:41:57  jcraig
 * jcraig on Tue Dec 21 15:41:16 PST 1993
 *
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * in i_reduceRowsAnd, i_reduceRowsSample and i_reduceRowsOr use
 * dBpl in the call to ip_compressRows* instead of sBpl
 *
 * Revision 1.18  1993/12/02  19:30:14  ddavies
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.17  1993/11/17  03:01:32  ddavies
 * ddavies on Tue Nov 16 18:59:18 PST 1993
 *
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * Fixed bug in ip_compressRowsArb (called from i_reduceRowsSample).  This bug
 * caused us to write each row 8 times as often as necessary.  By a process
 * that I still can't figure out, it also wrote trash into the right frame.
 * If the source and dest are both the same bytes/line, don't the frames always
 * line up, even if we copy many source lines to the dest in a block?  Well,
 * anyway, it works now.
 *
 * Revision 1.16  1993/11/16  03:10:05  ddavies
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * The 1bit/pixel, 8x subsample code was putting the rightmost byte together
 * incorrectly - the bits were being packed into the right end of the byte
 * instead of into the left end.
 * This caused it to write into the right frame.
 * The 4 bit/pixel subsampling at 8x, 16x and 32x was also writing into the
 * right frame.  These routines were producing 2 lines of output each time,
 * then shifting over by one line and doing it again - basically doing the
 * entire image twice!  They were also carefully composing the last word
 * of the line, then writing it whether it was needed or not.
 *
 * Revision 1.15  1993/11/05  02:26:04  ddavies
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


