/*****************************************************************
 *  Copyright (c) 1993.                              		 *
 *  Xerox Corporation.  All rights reserved.                     *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_subsmp2.c: contains subroutines for fast textured reduction
 *
 *                  Int32        i_reduceHOr()
 *                  Int32        i_reduceHAnd()
 *                  Int32        i_reduceVOr()
 *                  Int32        i_reduceVAnd()
 *		    Int32	 i_reduceVOrHAnd()
 *		    Int32	 i_reduceVAndHOr()
 *                  Int32        i_reduceRowsOr()
 *                  Int32        i_reduceRowsAnd()
 *                  Int32        i_reduceRowsSample()
 *                  Int32        i_reduceColsOr()
 *                  Int32        i_reduceColsAnd()
 *                  Int32        i_reduceColsSample()
 *
 * Internal procedures:
 *		    void	 ip_compressRowsAnd()
 *		    void	 ip_compressRowsOr()
 *		    void	 ip_compressRowsArb()
 *		    Int32	 ip_compressColsArb()
 *		    Int32	 ip_compressColsOr()
 *		    Int32	 ip_compressColsAnd()
 *
 * The textured reduction code could be rewritten to run faster.  There are
 * some obvious modifications that would help.  The inner loops do not 
 * write 32 bits of destination....  Since the execution time is so small
 * already (at least starting with 8.5x11 300spi images), this didn't 
 * seem worth it.  In addition, i_reduceV* could combine the row
 * compression with the picking of a column.  Since the row compression
 * could also be shared with i_reduce2D* and the times were so 
 * small, this was not done.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef _TYPES_PUB_INCLUDED
#include "types.pub"
#include "ipscore.prv"
#endif

#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"

#include "memory.pub"

    /* prototypes */
#include "shrpixr.pub"		/* for shrpixr.prv */
#include "shrpixr.prv"
#include "shrscal.pub"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_sbsmp2.c,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:18:13 $ */


/* PUBLIC */
/**************************************************************************
 *  i_reduceHOr()
 *  	Computes "horizontal (OR)-textured" reduction.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceHOr (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
	/* adjust base pointers to avoid the frame */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    ERR_PASS( ip_compressColsOr (pIPState, pD, wD, hD,
			         dBpl, pS, sBpl*factor, factor));

    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *  i_reduceHAnd()
 *  	Computes "horizontal (AND)-textured" reduction.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceHAnd (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    
	/* adjust base pointers to avoid the frame */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wD = wD - 2 * cFrameBits;
    hD = hD - 2 * cFrameBits;

    ERR_PASS( ip_compressColsAnd (pIPState, pD, wD, hD,
			          dBpl, pS, sBpl*factor, factor));
    NORMAL_RETURN;
}

/* PUBLIC */
/**************************************************************************
 *  i_reduceVAnd()
 *  	Computes "vertical (AND)-textured" reduction.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceVAnd (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    UInt8      *pF, *pFOrig;


        /* make intermediate (flat) pixr */
    pF = (UInt8 *)MALLOC(sBpl*hD);
    if (pF == NULL)
	return ia_nomem;
    pFOrig = pF;

	/* adjust base pointers to avoid the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    pF += (sBpl << cLogFrameBits) + cFrameBytes;

    hD -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;

    ip_compressRowsAnd (pIPState, pF, wD*factor, hD, sBpl, pS, sBpl, factor);

    ERR_PASS_CLEANUP( ip_compressColsArb (pIPState, pD, wD, hD,
					  dBpl, pF, sBpl, factor));

    FREE(pFOrig);
    NORMAL_RETURN;

ERR_CLEANUP:
    FREE(pFOrig);
    ERR_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_reduceVOr()
 *  	Computes "vertical (OR)-textured" reduction.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceVOr (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    UInt8      *pF, *pFOrig;


        /* get space for intermediate image */
    pF = (UInt8 *)MALLOC(sBpl*hD);
    if (pF == NULL)
	return ia_nomem;
    pFOrig = pF;

	/* adjust base pointers to avoid the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    pF += (sBpl << cLogFrameBits) + cFrameBytes;

    hD -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;

    ip_compressRowsOr (pIPState, pF, wD*factor, hD, sBpl, pS, sBpl, factor);

    ERR_PASS_CLEANUP( ip_compressColsArb (pIPState, pD, wD, hD,
					  dBpl, pF, sBpl, factor));

    FREE(pFOrig);
    NORMAL_RETURN;

ERR_CLEANUP:
    FREE(pFOrig);
    ERR_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_reduceVOrHAnd()
 *  	Computes reduction, vertically by or'ing and horizontally by
 *	and'ing.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceVOrHAnd (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    UInt8      *pF, *pFOrig;


        /* get space for intermediate image.  this used to be a malloc
	 * but purify complains about an unitialized memory read in
	 * ip_compressColsAnd.  that memory is never actually used but
	 * in order to get rid of the warning we'll change this to a
	 * calloc.  this is not the fastest way to fix the problem but
	 * then if speed ever becomes a problem there are other things
	 * in these routines that need to be changed as well.
	 * why does the unitialized memory read happen?  well, say you
	 * have an image 515 pixels wide and are reducing by 4.  then
	 * the destination width is 128 which only uses 512 source
	 * pixels.  we allocate enough for the entire source size but
	 * ip_compressRowsOr only actually fills the first 512 pixels of
	 * each line.  the inner loop of ip_compressColsAnd reads ahead
	 * by one source word.  when we are on the last destination word
	 * this causes that extra source word to be read but we never use
	 * it since we fall out of the loop first.
	 */
    pF = (UInt8 *)CALLOC(sBpl*hD, sizeof(UInt8));
    if (pF == NULL)
	return ia_nomem;
    pFOrig = pF;

	/* adjust base pointers to avoid the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    pF += (sBpl << cLogFrameBits) + cFrameBytes;

    hD -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;

    ip_compressRowsOr (pIPState, pF, wD*factor, hD, sBpl, pS, sBpl, factor);

    ERR_PASS_CLEANUP( ip_compressColsAnd (pIPState, pD, wD, hD,
					  dBpl, pF, sBpl, factor));

    FREE(pFOrig);
    NORMAL_RETURN;

ERR_CLEANUP:
    FREE(pFOrig);
    ERR_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_reduceVAndHOr()
 *  	Computes reduction, vertically by and'ing and horizontally by
 *	or'ing.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceVAndHOr (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    UInt8      *pF, *pFOrig;


        /* get space for intermediate image.  see comments on
	 * i_reduceVOrHAnd for the reason that calloc is used
	 */
    pF = (UInt8 *)CALLOC(sBpl*hD, sizeof(UInt8));
    if (pF == NULL)
	return ia_nomem;
    pFOrig = pF;

	/* adjust base pointers to avoid the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    pF += (sBpl << cLogFrameBits) + cFrameBytes;

    hD -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;

    ip_compressRowsAnd (pIPState, pF, wD*factor, hD, sBpl, pS, sBpl, factor);

    ERR_PASS_CLEANUP( ip_compressColsOr (pIPState, pD, wD, hD,
					 dBpl, pF, sBpl, factor));

    FREE(pFOrig);
    NORMAL_RETURN;

ERR_CLEANUP:
    FREE(pFOrig);
    ERR_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_reduceColsOr()
 *	Compress columns by or'ing factor columns together.  Rows are
 *	unchanged.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceColsOr (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
	/* adjust base pointers to avoid the frame */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;

    ERR_PASS( ip_compressColsOr (pIPState, pD, wD, hD, dBpl, pS, sBpl, factor));
    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *  i_reduceColsAnd()
 *	Compress columns by and'ing factor columns together.  Rows are
 *	unchanged.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceColsAnd (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    
	/* adjust base pointers to avoid the frame */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wD = wD - 2 * cFrameBits;
    hD = hD - 2 * cFrameBits;

    ERR_PASS( ip_compressColsAnd (pIPState, pD, wD, hD,
				  dBpl, pS, sBpl, factor));
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_reduceColsSample()
 *	Compress columns by subsampling.  Rows are not reduced.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceColsSample (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    
	/* adjust base pointers to avoid the frame */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    wD = wD - 2 * cFrameBits;
    hD = hD - 2 * cFrameBits;

    ERR_PASS( ip_compressColsArb (pIPState, pD, wD, hD,
				  dBpl, pS, sBpl, factor));
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_reduceRowsAnd()
 *	Compress rows by and'ing factor rows together.  Columns are
 *	unchanged.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *************************************************************************/
Int32  CDECL 
i_reduceRowsAnd (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{

	/* adjust base pointers to avoid the frame */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    hD -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;

    ip_compressRowsAnd (pIPState, pD, wD, hD, dBpl, pS, sBpl, factor);

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_reduceRowsSample()
 *	Compress rows by subsampling.  Columns are not reduced.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *************************************************************************/
Int32  CDECL 
i_reduceRowsSample (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{

	/* adjust base pointers to avoid the frame */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    hD -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;

    ip_compressRowsArb (pIPState, pD, wD, hD, dBpl, pS, sBpl*factor);

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_reduceRowsOr()
 *	Compress rows by and'ing factor rows together.  Columns are
 *	unchanged.
 *
 *	Input:
 *		pD = destination image (pointer to image including frame)
 *		wD = width of dest in pixels including frame
 *		hD = height of dest including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *************************************************************************/
Int32  CDECL 
i_reduceRowsOr (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{

	/* adjust base pointers to avoid the frame */
    PROLOGUE;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;

    hD -= 2 * cFrameBits;
    wD -= 2 * cFrameBits;

    ip_compressRowsOr (pIPState, pD, wD, hD, dBpl, pS, sBpl, factor);

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_compressRowsArb():   low-level code to implement row compression
 *                          by subsampling.
 *	Input:
 *		pD = destination image (pointer to image NOT including frame)
 *		wD = width of dest in pixels NOT including frame
 *		hD = height of dest NOT including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image NOT including frame)
 * 		sInc = amount to increment pS to get to next source line
 *
 **************************************************************************/
void  CDECL 
ip_compressRowsArb (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sInc)
{
    PROLOGUE;
    Int32       rows;
    UInt32	lineWidthInBytes;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Note that we must transfer rows in units of words for the
	 * little endian and big endian machines to get the
	 * same answer.  Since the rightmost word of each scan line
	 * is filled with 0's in the frame, we can always transfer
	 * the whole thing.  We avoid using dBpl just in case the
	 * source and dest images don't have the same number of
	 * bytes/line.
	 */
    lineWidthInBytes = ((wD + 31)/8) & ~0x3;
    for (rows = hD; rows > 0; rows--, pD += dBpl, pS += sInc)
    {
	memcpy((void *)pD, (void *)pS, lineWidthInBytes);
    }
}


/* PRIVATE */
/**************************************************************************
 *  ip_compressRowsOr():   low-level code to implement row compression
 *                         with logical or operation.
 *                      Computes each line in the dest by ORing or
 *                         "factor" rows together.
 *	Input:
 *		pD = destination image (pointer to image NOT including frame)
 *		wD = width of dest in pixels NOT including frame
 *		hD = height of dest NOT including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image NOT including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 **************************************************************************/
void  CDECL 
ip_compressRowsOr (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    Int32	row;
    Int32	col, words, fRow;
    UInt8      *sPtr, *dPtr, *nextRow;
    Int32	sInc;
    UInt32	oWord;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    words = (wD+31)/32;   /* 32-bit words */

    switch (factor)
    {
    case 2:
        sInc = sBpl << 1;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		ULONG_ACCESS(dPtr) = 
		    ULONG_ACCESS(sPtr) | ULONG_ACCESS(&sPtr[sBpl]);
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;

    case 4:
        sInc = sBpl << 2;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		ULONG_ACCESS(dPtr) = 
		    ULONG_ACCESS(sPtr) | ULONG_ACCESS(&sPtr[sBpl]) |
		    ULONG_ACCESS(&sPtr[sBpl+sBpl]) |
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl]);
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;

    case 8:
        sInc = sBpl << 3;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		ULONG_ACCESS(dPtr) = 
		    ULONG_ACCESS(sPtr) | ULONG_ACCESS(&sPtr[sBpl]) |
		    ULONG_ACCESS(&sPtr[sBpl+sBpl]) |
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl]) |
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl+sBpl]) |
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl+sBpl+sBpl]) |
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl+sBpl+sBpl+sBpl]) |
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl+sBpl+sBpl+sBpl+sBpl]);
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;

    case 16:
        sInc = sBpl << 4;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

            for (col = words; col > 0; col--)    
	    {
		oWord = ULONG_ACCESS(sPtr);
		nextRow = sPtr + sBpl;
		for (fRow = 15; fRow > 0; fRow--, nextRow += sBpl)
		    oWord |= ULONG_ACCESS(nextRow);
		ULONG_ACCESS(dPtr) = oWord;
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;

    case 32:
        sInc = sBpl << 5;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		oWord = ULONG_ACCESS(sPtr);
		nextRow = sPtr + sBpl;
		for (fRow = 31; fRow > 0; fRow--, nextRow += sBpl)
		    oWord |= ULONG_ACCESS(nextRow);
		ULONG_ACCESS(dPtr) = oWord;
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;
    }
}




/* PRIVATE */
/**************************************************************************
 *  ip_compressRowsAnd():   low-level code to implement row compression
 *                          with logical and operation.
 *                      Computes each line in the dest by ANDing or
 *                         "factor" rows together.
 *	Input:
 *		pD = destination image (pointer to image NOT including frame)
 *		wD = width of dest in pixels NOT including frame
 *		hD = height of dest NOT including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image NOT including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 **************************************************************************/
void  CDECL 
ip_compressRowsAnd (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    Int32	row;
    Int32	col, words, fRow;
    UInt8      *sPtr, *dPtr, *nextRow;
    UInt32	oWord;
    Int32	sInc;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    words = (wD+31)/32;   /* 32-bit words */

    switch (factor)
    {
    case 2:
        sInc = sBpl << 1;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		ULONG_ACCESS(dPtr) = 
		    ULONG_ACCESS(sPtr) & ULONG_ACCESS(&sPtr[sBpl]);
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;

    case 4:
        sInc = sBpl << 2;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		ULONG_ACCESS(dPtr) = 
		    ULONG_ACCESS(sPtr) & ULONG_ACCESS(&sPtr[sBpl]) &
		    ULONG_ACCESS(&sPtr[sBpl+sBpl]) &
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl]);
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;

    case 8:
        sInc = sBpl << 3;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		ULONG_ACCESS(dPtr) = 
		    ULONG_ACCESS(sPtr) & ULONG_ACCESS(&sPtr[sBpl]) &
		    ULONG_ACCESS(&sPtr[sBpl+sBpl]) &
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl]) &
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl+sBpl]) &
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl+sBpl+sBpl]) &
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl+sBpl+sBpl+sBpl]) &
		    ULONG_ACCESS(&sPtr[sBpl+sBpl+sBpl+sBpl+sBpl+sBpl+sBpl]);
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;

    case 16:
        sInc = sBpl << 4;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		oWord = ULONG_ACCESS(sPtr);
		nextRow = sPtr + sBpl;
		for (fRow = 15; fRow > 0; fRow--, nextRow += sBpl)
		    oWord &= ULONG_ACCESS(nextRow);
		ULONG_ACCESS(dPtr) = oWord;
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;

    case 32:
        sInc = sBpl << 5;
        for (row = hD; row > 0; row--, pS += sInc, pD += dBpl)
        {
	    sPtr = pS;
	    dPtr = pD;

		/* copy the first line */
            for (col = words; col > 0; col--)    
	    {
		oWord = ULONG_ACCESS(sPtr);
		nextRow = sPtr + sBpl;
		for (fRow = 31; fRow > 0; fRow--, nextRow += sBpl)
		    oWord &= ULONG_ACCESS(nextRow);
		ULONG_ACCESS(dPtr) = oWord;
		dPtr += 4;
		sPtr += 4;
	    }
        }
	break;
    }
}


/* PRIVATE */
/**************************************************************************
 *  ip_compressColsArb():  low-level code to implement the column filtering 
 *                         and subsampling part of i_reduceV*() by picking an
 *			   arbitrary column.
 *	Input:
 *		pD = destination image (pointer to image NOT including frame)
 *		wD = width of dest in pixels NOT including frame
 *		hD = height of dest NOT including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image NOT including frame)
 * 		sBpl = bytes per line of source (including frame)
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 **************************************************************************/
Int32  CDECL 
ip_compressColsArb (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    Int32	factor)
{
    PROLOGUE;
    Int32	row;		/* iteration variable */
    Int32	col;		/* iteration variable */
    UInt8      *sPtr, *dPtr;	/* current source and dest pointers */
    UInt8      *table;		/* table used by factor == 2 */
    UInt32	oWord;		/* output word being built */
    UInt32	dstMask;	/* which source bit to use for dest bit */
    UInt16     *pTabPermute;   /* permute byte tables */
    UInt8      *pTab4Permute;
    static UInt16 bit16Mask[] =
	{0xFFFF, 0x8000, 0xC000, 0xE000, 0xF000,
		 0xF800, 0xFC00, 0xFE00, 0xFF00,
		 0xFF80, 0xFFC0, 0xFFE0, 0xFFF0,
		 0xFFF8, 0xFFFC, 0xFFFE};
    static UInt8 bit8Mask[] = {0xFF, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE};
    UInt16	mask16;		/* mask for last dest */
    UInt8	mask8;

    
    switch (factor)
    {
    case 2:
	if ((__error__ = ip_makePermuteLUT (pIPState, &pTabPermute))
							!= ia_successful)
	    NORMAL_RETURN;
	table = (UInt8 *) pTabPermute;
        mask16 = bit16Mask[wD&0xF];
        for (row = hD; row > 0; row--, pD += dBpl, pS += sBpl)
        {
            dPtr = pD;
            sPtr = pS;

            for ( col = wD >> 4; 
                  col >= 0; 
                  col--, dPtr += 2, sPtr += 4 ) 
	    {

		oWord = ULONG_ACCESS(sPtr);
		    /* OR the data horizonatlly positions.
		     */
		oWord &= 0xAAAAAAAA;
		oWord = oWord | (oWord >> 15);
	        USHORT_ACCESS(dPtr) =
		    ((*(UInt16 *) &table[(oWord >> 8) & 0x1FE]) << 4) |
		    ((*(UInt16 *) &table[(oWord     ) & 0x1FE])     );

            }
		/* get rid of any bits in the frame */
	    USHORT_ACCESS(&dPtr[-2]) &= mask16;
        }
	FREE(pTabPermute);
	break;

    case 4:
        mask8 = bit8Mask[wD&0x7];
	if ((__error__ = ip_make4PermuteLUT (pIPState, &pTab4Permute))
							!= ia_successful)
	    NORMAL_RETURN;

        for (row = hD; row > 0; row--, pD += dBpl, pS += sBpl)
        {
            dPtr = pD;
            sPtr = pS;

            for (col = 0; col < wD; col += 8, dPtr++, sPtr += 4)
            {
		oWord = ULONG_ACCESS(sPtr);
		oWord &= 0x88888888;
		oWord = oWord >> 3;
		oWord |= oWord >> 15;
		oWord |= oWord >> 6;
	        UCHAR_ACCESS(dPtr) =
		    *(UInt8 *) &pTab4Permute[oWord & 0xFF];
            }
	    UCHAR_ACCESS(&dPtr[-1]) &= mask8;
        }
	FREE(pTab4Permute);
	break;

    case 8:
        mask8 = bit8Mask[wD&0x7];
        for (row = hD; row > 0; row--, pD += dBpl, pS += sBpl)
        {
            dPtr = pD;
            sPtr = pS;

            for (col = wD; col > 0; col -= 4)
            {
                oWord = (UCHAR_ACCESS(sPtr++) & 0x80);
                oWord += (UCHAR_ACCESS(sPtr++) >> 1) & 0x40;
                oWord += (UCHAR_ACCESS(sPtr++) >> 2) & 0x20;
                oWord += (UCHAR_ACCESS(sPtr++) >> 3) & 0x10;
		if ((col -= 4) <= 0)
		{
		    UCHAR_ACCESS(dPtr++) = (UInt8) oWord;
		    break;
		}
                oWord += (UCHAR_ACCESS(sPtr++) >> 4) & 0x08;
                oWord += (UCHAR_ACCESS(sPtr++) >> 5) & 0x04;
                oWord += (UCHAR_ACCESS(sPtr++) >> 6) & 0x02;
                oWord += (UCHAR_ACCESS(sPtr++) >> 7) & 0x01;
		UCHAR_ACCESS(dPtr++) = (UInt8) oWord;
            }
	    UCHAR_ACCESS(&dPtr[-1]) &= mask8;
        }       
	break;

    case 16:
        for (row = hD; row > 0; row--, pD += dBpl, pS += sBpl)
        {
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~7); col > 0; col -= 8)
            {
                oWord = (USHORT_ACCESS(sPtr) & 0x80);
		sPtr += 2;
                oWord += (USHORT_ACCESS(sPtr) & 0x40);
		sPtr += 2;
                oWord += (USHORT_ACCESS(sPtr) & 0x20);
		sPtr += 2;
                oWord += (USHORT_ACCESS(sPtr) & 0x10);
		sPtr += 2;
                oWord += (USHORT_ACCESS(sPtr) & 0x08);
		sPtr += 2;
                oWord += (USHORT_ACCESS(sPtr) & 0x04);
		sPtr += 2;
                oWord += (USHORT_ACCESS(sPtr) & 0x02);
 		sPtr += 2;
                oWord += (USHORT_ACCESS(sPtr) & 0x01);
		sPtr += 2;
		UCHAR_ACCESS(dPtr++) = (UInt8) oWord;
            }
	    if (wD & 7)
	    {
		oWord = 0;
		dstMask = 0x80;
		for (col = (wD & 7); col > 0; col--)
		{
                    oWord += (USHORT_ACCESS(sPtr) & dstMask);
 		    sPtr += 2;
		    dstMask = dstMask >> 1;
		}
	    UCHAR_ACCESS(dPtr) = (UInt8) oWord;
	    }
        }       
	break;

    case 32:
        for (row = hD; row > 0; row--, pD += dBpl, pS += sBpl)
        {
            dPtr = pD;
            sPtr = pS;

            for (col = (wD & ~7); col > 0; col -= 8)
            {
                oWord = (ULONG_ACCESS(sPtr) & 0x80);
		sPtr += 4;
                oWord += (ULONG_ACCESS(sPtr) & 0x40);
		sPtr += 4;
                oWord += (ULONG_ACCESS(sPtr) & 0x20);
		sPtr += 4;
                oWord += (ULONG_ACCESS(sPtr) & 0x10);
		sPtr += 4;
                oWord += (ULONG_ACCESS(sPtr) & 0x08);
		sPtr += 4;
                oWord += (ULONG_ACCESS(sPtr) & 0x04);
		sPtr += 4;
                oWord += (ULONG_ACCESS(sPtr) & 0x02);
		sPtr += 4;
                oWord += (ULONG_ACCESS(sPtr) & 0x01);
		sPtr += 4;
		UCHAR_ACCESS(dPtr++) = (UInt8) oWord;
            }
	    if (wD & 7)
	    {
		oWord = 0;
		dstMask = 0x80;
		for (col = (wD & 7); col > 0; col--)
		{
                    oWord += (ULONG_ACCESS(sPtr) & dstMask);
 		    sPtr += 4;
		    dstMask = dstMask >> 1;
		}
	    UCHAR_ACCESS(dPtr) = (UInt8) oWord;
	    }
        }       

    }
    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_compressColsOr():   low-level code to implement column compression 
 *			   or'ing all the columns together.
 *	Input:
 *		pD = destination image (pointer to image NOT including frame)
 *		wD = width of dest in pixels NOT including frame
 *		hD = height of dest NOT including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image NOT including frame)
 * 		sInc = amount to increment pS to get to next source line
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *  Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 **************************************************************************/
Int32  CDECL 
ip_compressColsOr (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sInc,
    Int32	factor)
{
    PROLOGUE;
    Int32	row;
    Int32	col;
    UInt8      *sPtr;    	/* pointer into source pixrect */
    UInt8      *dPtr;    	/* pointer into dest'n pixrect */
    UInt32	srcMask; 	/* mask of src pels being considered */
    UInt32	destMask; 	/* mask of dest pels being set */
    UInt8      *table;		/* table used by factor == 2 */
    UInt32	oWord;		/* output word being built */
    UInt32	source;		/* current source word */
    UInt16     *pTabPermute;	/* permute table byte */
    Int32	srcLimit;	/* max bytes in source line */
    Int32	srcLimit2;	/* used for 2 level reduction */
    Int32	srcInitOffset;	/* initial source line offset */
    Int32	srcInitOffset2; /* used for 2 level reduction */
    Int32	dstInitOffset;	/* initial dst line offset */
    Int32	dstInitOffset2; /* used for 2 level reduction */
    UInt8      *pDOrig;	/* used to clear right frame */
    Int32	longestAccess;	/* the longest memory access
				 * allowed for the current processor
				 * and source and dest images. */
    UInt8      *pLineBuf;	/* holds one reduced image line, used
				 * for 16x and 32x reduction. */
    static UInt16 endMask[32] =
    {0xFFFF, 0x8000, 0xC000, 0xE000, 0xF000,
	     0xF800, 0xFC00, 0xFE00, 0xFF00,
	     0xFF80, 0xFFC0, 0xFFE0, 0xFFF0,
	     0xFFF8, 0xFFFC, 0xFFFE};
    UInt16	mask;
    
	/* if wD == 0, just return instead of blowing up below */
    if (wD <= 0)
	NORMAL_RETURN;

	/* Figure out whether the lower level routines will do long
	 * (8-byte) or short (4-byte) accesses.
	 */
    ip_checkImageAccess(pIPState, pS, 0, pD, dBpl, &longestAccess);
    if ((longestAccess != 4) && (longestAccess != 8))
	return ia_internal;

    pDOrig = pD;			/* save to clear right frame */

    switch (factor)
    {
    case 2:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
	    ERR_PASS( ip_makePermuteLUT (pIPState, &pTabPermute));
	    table = (UInt8 *) pTabPermute;
	    mask = endMask[wD&0xF];
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		dPtr = pD;
		sPtr = pS;
    
		for ( col = wD >> 4; 
		      col >= 0; 
		      col--, dPtr += 2, sPtr += 4 ) 
		{
    
		    oWord = ULONG_ACCESS(sPtr);
			/* OR the data horizontally.
			 */
		    oWord = (oWord | (oWord << 1)) & 0xAAAAAAAA;
		    oWord = oWord | (oWord >> 15);
		    USHORT_ACCESS(dPtr) =
			((*(UInt16 *) &table[(oWord >> 8) & 0x1FE]) << 4) |
			((*(UInt16 *) &table[(oWord     ) & 0x1FE])     );
    
		}
		    /* get rid of any bits in the frame */
		USHORT_ACCESS(&dPtr[-2]) &= mask;
	    }
	    FREE(pTabPermute);
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* this points to the last dest doubleword.
		 * The dest has 1-bit pixels. */
	    dstInitOffset = ((wD - 1)/64) * 8 ;
		/* it takes 2 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 2;	
	
		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 2 * wD.
		 */
	    srcLimit = (((wD * 2) - 1)/64) * 8;
    
	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
					* operation mode.
					*/
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compressColsOrLineLong (pIPState, pS, 2 * wD, pD,
				           dstInitOffset, srcInitOffset,
					   srcLimit);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}
	break;

    case 4:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
		/* Do it the old way */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		oWord = 0;
		dPtr = pD;
		sPtr = pS;
    
		source = ULONG_ACCESS(sPtr);
		for ( col = wD, destMask = 0x80000000, srcMask = 0xf0000000;
		      col > 0;
		      col-- )
		{
    
		    if ( source & srcMask )
			oWord |= destMask;
    
		    destMask >>= 1;
		    srcMask >>= 4;
		    if ( srcMask == 0 )
		    {
			srcMask = 0xf0000000;
			sPtr += 4;
			source = ULONG_ACCESS(sPtr);
    
			if ( destMask == 0 )
			{
			    destMask = 0x80000000;
			    ULONG_ACCESS(dPtr) = oWord;
			    dPtr += 4;
			    oWord = 0;
			}
		    }    
		}
		if (destMask != 0x80000000)
		    ULONG_ACCESS(dPtr) = oWord;    
	    }
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* this points to the last dest doubleword.
		 * The dest has 1-bit pixels. */
	    dstInitOffset = ((wD - 1)/64) * 8 ;
		/* it takes 4 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 4;	
	
		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 4 * wD.
		 */
	    srcLimit = (((wD * 4) - 1)/64) * 8;
    
	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compress4ColsOrLineLong (pIPState, pS, 4 * wD, pD,
				            dstInitOffset, srcInitOffset,
					    srcLimit);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}
	break;

    case 8:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
		/* Do it the old way */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		oWord = 0;
		dPtr = pD;
		sPtr = pS;
    
		source = ULONG_ACCESS(sPtr);
		for ( col = wD, destMask = 0x80000000, srcMask = 0xff000000;
		      col > 0;
		      col-- ) {
    
		    if ( source & srcMask )
			oWord |= destMask;
    
		    destMask >>= 1;
		    srcMask >>= 8;
		    if ( srcMask == 0 ) {
			srcMask = 0xff000000;
			sPtr += 4;
			source = ULONG_ACCESS(sPtr);
    
			if ( destMask == 0 ) {
			    destMask = 0x80000000;
			    ULONG_ACCESS(dPtr) = oWord;
			    dPtr += 4;
			    oWord = 0;
			}
		    } 
		}   
		if (destMask != 0x80000000)
		    ULONG_ACCESS(dPtr) = oWord;    
	    }
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* this points to the last dest doubleword.
		 * The dest has 1-bit pixels. */
	    dstInitOffset = ((wD - 1)/64) * 8 ;
		/* it takes 8 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 8;	

		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 8 * wD.
		 */
	    srcLimit = (((wD * 8) - 1)/64) * 8;
    
	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compress8ColsOrLineLong (pIPState, pS, 8 * wD, pD,
				            dstInitOffset, srcInitOffset,
					    srcLimit);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}
	break;

    case 16:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		oWord = 0;
		dPtr = pD;
		sPtr = pS;
    
		source = ULONG_ACCESS(sPtr);
		for ( col = wD, destMask = 0x80000000, srcMask = 0xffff0000;
		      col > 0;
		      col-- )
		{
     
		    if ( source & srcMask )
			oWord |= destMask;
     
		    destMask >>= 1;
		    srcMask >>= 16;
		    if ( srcMask == 0 )
		    {
			srcMask = 0xffff0000;
			sPtr += 4;
			source = ULONG_ACCESS(sPtr);
     
			if ( destMask == 0 )
			{
			    destMask = 0x80000000;
			    ULONG_ACCESS(dPtr) = oWord;
			    dPtr += 4;
			    oWord = 0;
			}
		    }
		}
		if (destMask != 0x80000000)
		    ULONG_ACCESS(dPtr) = oWord;    
	    }    	
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* We'll do this in two steps, first reduction by
		 * 8, then reduction by 2.  First allocate the intermediate
		 * line buffer.  It's width in bits will be (wD * 2).
		 * However, we want to round this up to the next multiple
		 * of 64 bits, so we do the (+63)/64 trick.  We need to
		 * convert this back into bytes, so we multiply by
		 * 8 bytes/dword.  We also use CALLOC to make sure that
		 * the glop in the last byte is cleared.
		 * And well, just to make sure there's some pad on the
		 * end, we add another doubleword.
		 */
	    pLineBuf = (UInt8 *) CALLOC(1, ((((wD * 2) + 63) / 64) + 1) * 8);
	    if (pLineBuf == NULL)
		return ia_nomem;

		/* Set up offsets for the first 8x reduction.
		 * dstInitOffset points to the last dest doubleword.
		 * The dest has 1-bit pixels.  Note that the width of
		 * the dest here is wD * 2.
		 */
	    dstInitOffset = (((wD * 2) - 1)/64) * 8 ;
		/* it takes 8 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 8;	

		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 8 * (wD * 2).
		 */
	    srcLimit = ((((wD * 2) * 8) - 1)/64) * 8;


		/* Set up the corresponding offsets for the second stage
		 * 2x reduction.
		 */
	    dstInitOffset2 = ((wD - 1)/64) * 8 ;
		/* it takes 2 source doublewords to make one dest doubleword */
	    srcInitOffset2 = dstInitOffset2 * 2;	
	    srcLimit2 = (((wD * 2) - 1)/64) * 8;

	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compress8ColsOrLineLong (pIPState, pS, 16 * wD, pLineBuf,
				            dstInitOffset, srcInitOffset,
					    srcLimit);
		ip_compressColsOrLineLong (pIPState, pLineBuf, 2 * wD, pD,
				           dstInitOffset2, srcInitOffset2,
					   srcLimit2);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);

		/* We're done.  Free the intermediate line buffer */
	    FREE(pLineBuf);
	}
        break;       

    case 32:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		oWord = 0;
		dPtr = pD;
		sPtr = pS;
     
		for ( col = wD, destMask = 0x80000000;
		      col > 0;
		      col-- )
		{
     
		    if ( ULONG_ACCESS(sPtr) )
			oWord |= destMask;
		    sPtr += 4;
     
		    destMask >>= 1;
		    if ( destMask == 0 )
		    {
			destMask = 0x80000000;
			ULONG_ACCESS(dPtr) = oWord;
			dPtr += 4;
			oWord = 0;
		    }
		}
		if (destMask != 0x80000000)
		    ULONG_ACCESS(dPtr) = oWord;    
	    }
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* We'll do this in two steps, first reduction by
		 * 8, then reduction by 4.  First allocate the intermediate
		 * line buffer.  It's width in bits will be (wD * 4).
		 * However, we want to round this up to the next multiple
		 * of 64 bits, so we do the (+63)/64 trick.  We need to
		 * convert this back into bytes, so we multiply by
		 * 8 bytes/dword.  We also use CALLOC to make sure that
		 * the glop in the last byte is cleared.
		 * And well, just to make sure there's some pad on the
		 * end, we add another doubleword.
		 */
	    pLineBuf = (UInt8 *) CALLOC(1, ((((wD * 4) + 63) / 64) + 1) * 8);
	    if (pLineBuf == NULL)
		return ia_nomem;

		/* Set up offsets for the first 8x reduction.
		 * dstInitOffset points to the last dest doubleword.
		 * The dest has 1-bit pixels.  Note that the width of
		 * the dest here is wD * 4.
		 */
	    dstInitOffset = (((wD * 4) - 1)/64) * 8 ;
		/* it takes 8 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 8;	

		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 8 * (wD * 4).
		 */
	    srcLimit = ((((wD * 4) * 8) - 1)/64) * 8;


		/* Set up the corresponding offsets for the second stage
		 * 4x reduction.
		 */
	    dstInitOffset2 = ((wD - 1)/64) * 8 ;
		/* it takes 4 source doublewords to make one dest doubleword */
	    srcInitOffset2 = dstInitOffset2 * 4;	
	    srcLimit2 = (((wD * 4) - 1)/64) * 8;

	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compress8ColsOrLineLong (pIPState, pS, 32 * wD, pLineBuf,
				            dstInitOffset, srcInitOffset,
					    srcLimit);
		ip_compress4ColsOrLineLong (pIPState, pLineBuf, 4 * wD, pD,
				            dstInitOffset2, srcInitOffset2,
					    srcLimit2);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);

		/* We're done.  Free the intermediate line buffer */
	    FREE(pLineBuf);
	}
        break;    
    }      

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_compressColsAnd():   low-level code to implement column compression 
 *			    and'ing all the columns together.
 *	Input:
 *		pD = destination image (pointer to image NOT including frame)
 *		wD = width of dest in pixels NOT including frame
 *		hD = height of dest NOT including frame
 *		dBpl = bytes per line of destination (including frame)
 *		pS = source image (pointer to image NOT including frame)
 * 		sInc = amount to increment pS to get to next source line
 *		factor = reduction factor. must be 2, 4, 8, 16, or 32.
 *
 *  Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 **************************************************************************/
Int32  CDECL 
ip_compressColsAnd (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sInc,
    Int32	factor)
{
    PROLOGUE;
    Int32   	row;
    Int32	col;
    UInt8      *sPtr;    	/* pointer into source pixrect */
    UInt8      *dPtr;    	/* pointer into dest'n pixrect */
    UInt32	srcMask; 	/* mask of src pels being considered */
    UInt32	destMask; 	/* mask of dest pels being set */
    UInt8      *table;		/* table used by factor == 2 */
    UInt32	oWord;		/* output word being built */
    UInt32	source;		/* current source word */
    UInt16     *pTabPermute;	/* permute table byte */
    Int32	srcLimit;	/* max bytes in source line */
    Int32	srcLimit2;	/* used for 2 level reduction */
    Int32	srcInitOffset;	/* initial source line offset */
    Int32	srcInitOffset2; /* used for 2 level reduction */
    Int32	dstInitOffset;	/* initial dst line offset */
    Int32	dstInitOffset2; /* used for 2 level reduction */
    UInt8      *pDOrig;		/* used to clear right frame */
    Int32	longestAccess;	/* the longest memory access
				 * allowed for the current processor
				 * and source and dest images. */
    UInt8      *pLineBuf;	/* holds one reduced image line, used
				 * for 16x and 32x reduction. */
    static UInt16 endMask[32] =
	{0xFFFF, 0x8000, 0xC000, 0xE000, 0xF000,
		 0xF800, 0xFC00, 0xFE00, 0xFF00,
		 0xFF80, 0xFFC0, 0xFFE0, 0xFFF0,
		 0xFFF8, 0xFFFC, 0xFFFE};
    UInt16	mask;

	/* if wD == 0, just return instead of blowing up below */
    if (wD <= 0)
	NORMAL_RETURN;

	/* Figure out whether the lower level routines will do long
	 * (8-byte) or short (4-byte) accesses.
	 */
    ip_checkImageAccess (pIPState, pS, 0, pD, dBpl, &longestAccess);
    if ((longestAccess != 4) && (longestAccess != 8))
	return ia_internal;

    pDOrig = pD;			/* save to clear right frame */

    switch (factor)
    {
    case 2:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
	    ERR_PASS( ip_makePermuteLUT (pIPState, &pTabPermute));
	    table = (UInt8 *) pTabPermute;
	    mask = endMask[wD&0xF];
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		dPtr = pD;
		sPtr = pS;
    
		for ( col = wD >> 4; 
		      col >= 0; 
		      col--, dPtr += 2, sPtr += 4 ) 
		{
    
		    oWord = ULONG_ACCESS(sPtr);
			/* AND the data horizontally and put the good bits
			 * in the odd numbered positions.
			 */
		    oWord = (oWord & (oWord << 1)) & 0xAAAAAAAA;
		    oWord = oWord | (oWord >> 15);
		    USHORT_ACCESS(dPtr) =
			((*(UInt16 *) &table[(oWord >> 8) & 0x1FE]) << 4) |
			((*(UInt16 *) &table[(oWord     ) & 0x1FE])     );
    
		}
		    /* get rid of any bits in the frame */
		USHORT_ACCESS(&dPtr[-2]) &= mask;
	    }
	    FREE(pTabPermute);
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* this points to the last dest doubleword.
		   The dest has 1-bit pixels. */
	    dstInitOffset = ((wD - 1)/64) * 8 ;
		/* it takes 2 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 2;	
	
		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 2 * wD.
		 */
	    srcLimit = (((wD * 2) - 1)/64) * 8;
    
	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compressColsAndLineLong (pIPState, pS, 2 * wD, pD,
				            dstInitOffset, srcInitOffset,
					    srcLimit);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}
	break;

    case 4:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
		/* Do it the old way */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		oWord = 0;
		dPtr = pD;
		sPtr = pS;
    
		source = ULONG_ACCESS(sPtr);
		for ( col = wD, destMask = 0x80000000, srcMask = 0xf0000000; 
		      col > 0; 
		      col-- ) 
		{
    
		    if ( (source & srcMask) == srcMask )
			oWord |= destMask;
    
		    destMask >>= 1;
		    srcMask >>= 4;
		    if ( srcMask == 0 ) 
		    {
			srcMask = 0xf0000000;
			sPtr += 4;
			source = ULONG_ACCESS(sPtr);
    
			if ( destMask == 0 ) 
			{
			    destMask = 0x80000000;
			    ULONG_ACCESS(dPtr) = oWord;
			    dPtr += 4;
			    oWord = 0;
			}
		    }
		}
		if (destMask != 0x80000000)
		    ULONG_ACCESS(dPtr) = oWord;    
	    }
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* this points to the last dest doubleword.
		   The dest has 1-bit pixels. */
	    dstInitOffset = ((wD - 1)/64) * 8 ;
		/* it takes 4 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 4;	
	
		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 4 * wD.
		 */
	    srcLimit = (((wD * 4) - 1)/64) * 8;
    
	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compress4ColsAndLineLong (pIPState, pS, 4 * wD, pD,
				             dstInitOffset, srcInitOffset,
					     srcLimit);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}
	break;

    case 8:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
		/* Do it the old way */
	    for (row = 0; row < hD; row++, pD += dBpl, pS += sInc)
	    {
		oWord = 0;
		dPtr = pD;
		sPtr = pS;
    
		source = ULONG_ACCESS(sPtr);
    
		for ( col = 0, destMask = 0x80000000, srcMask = 0xff000000; 
		      col < wD; 
		      ++col ) 
		{
    
		    if ( (source & srcMask) == srcMask ) 
			oWord |= destMask;
    
		    destMask >>= 1;
		    srcMask >>= 8;
		    if ( srcMask == 0 ) 
		    {
			srcMask = 0xff000000;
			sPtr += 4;
			source = ULONG_ACCESS(sPtr);
    
			if ( destMask == 0 ) 
			{
			    destMask = 0x80000000;
			    ULONG_ACCESS(dPtr) = oWord;
			    dPtr += 4;
			    oWord = 0;
			}
		    }
		}
		if (destMask != 0x80000000)
		    ULONG_ACCESS(dPtr) = oWord;    
	    }       
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* this points to the last dest doubleword.
		   The dest has 1-bit pixels. */
	    dstInitOffset = ((wD - 1)/64) * 8 ;
		/* it takes 8 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 8;	
	
		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 8 * wD.
		 */
	    srcLimit = (((wD * 8) - 1)/64) * 8;
    
	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compress8ColsAndLineLong (pIPState, pS, 8 * wD, pD,
				             dstInitOffset, srcInitOffset,
					     srcLimit);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);
	}
	break;

    case 16:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
	    for (row = 0; row < hD; row++, pD += dBpl, pS += sInc)
	    {
		oWord = 0;
		dPtr = pD;
		sPtr = pS;
    
		source = ULONG_ACCESS(sPtr);
		for ( col = 0, destMask = 0x80000000, srcMask = 0xffff0000; 
		      col < wD; 
		      ++col ) {
    
		    if ( (source & srcMask) == srcMask )
			oWord |= destMask;
    
		    destMask >>= 1;
		    srcMask >>= 16;
		    if ( srcMask == 0 ) 
		    {
			srcMask = 0xffff0000;
			sPtr += 4;
			source = ULONG_ACCESS(sPtr);
    
			if ( destMask == 0 ) 
			{
			    destMask = 0x80000000;
			    ULONG_ACCESS(dPtr) = oWord;
			    dPtr += 4;
			    oWord = 0;
			}
		    }
		}
		if (destMask != 0x80000000)
		    ULONG_ACCESS(dPtr) = oWord;    
	    }       
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* We'll do this in two steps, first reduction by
		 * 8, then reduction by 2.  First allocate the intermediate
		 * line buffer.  It's width in bits will be (wD * 2).
		 * However, we want to round this up to the next multiple
		 * of 64 bits, so we do the (+63)/64 trick.  We need to
		 * convert this back into bytes, so we multiply by
		 * 8 bytes/dword.  We also use CALLOC to make sure that
		 * the glop in the last byte is cleared.
		 * And well, just to make sure there's some pad on the
		 * end, we add another doubleword.
		 */
	    pLineBuf = (UInt8 *) CALLOC(1, ((((wD * 2) + 63) / 64) + 1) * 8);
	    if (pLineBuf == NULL)
		return ia_nomem;

		/* Set up offsets for the first 8x reduction.
		 * dstInitOffset points to the last dest doubleword.
		 * The dest has 1-bit pixels.  Note that the width of
		 * the dest here is wD * 2.
		 */
	    dstInitOffset = (((wD * 2) - 1)/64) * 8 ;
		/* it takes 8 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 8;	

		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 8 * wD.
		 */
	    srcLimit = ((((wD * 2) * 8) - 1)/64) * 8;


		/* Set up the corresponding offsets for the second stage
		 * 2x reduction.
		 */
	    dstInitOffset2 = ((wD - 1)/64) * 8 ;
		/* it takes 2 source doublewords to make one dest doubleword */
	    srcInitOffset2 = dstInitOffset2 * 2;	
	    srcLimit2 = (((wD * 2) - 1)/64) * 8;

	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compress8ColsAndLineLong (pIPState, pS, 16 * wD, pLineBuf,
				             dstInitOffset, srcInitOffset,
					     srcLimit);
		ip_compressColsAndLineLong (pIPState, pLineBuf, 2 * wD, pD,
				            dstInitOffset2, srcInitOffset2,
					    srcLimit2);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);

		/* We're done.  Free the intermediate line buffer */
	    FREE(pLineBuf);
	}
        break;       

    case 32:
	    /* See whether we're going to use fast code or do it the
	     * old way.
	     */
	if ((longestAccess == 4) || (!ip_longOpScaleLibLinked (pIPState)))
	{
	    for (row = 0; row < hD; row++, pD += dBpl, pS += sInc)
	    {
		oWord = 0;
		dPtr = pD;
		sPtr = pS;
    
		for ( col = 0, destMask = 0x80000000; 
		      col < wD; 
		      ++col ) 
		{
    
		    if ( ULONG_ACCESS(sPtr) == 0xffffffff )
			oWord |= destMask;
		    sPtr += 4;
    
		    destMask >>= 1;
		    if ( destMask == 0 ) 
		    {
			destMask = 0x80000000;
			ULONG_ACCESS(dPtr) = oWord;
			dPtr += 4;
			oWord = 0;
		    }
		}
		if (destMask != 0x80000000)
		    ULONG_ACCESS(dPtr) = oWord;    
	    } 
	}
	else	/* we're going to process in units of 64-bit words */
	{
		/* We'll do this in two steps, first reduction by
		 * 8, then reduction by 4.  First allocate the intermediate
		 * line buffer.  It's width in bits will be (wD * 4).
		 * However, we want to round this up to the next multiple
		 * of 64 bits, so we do the (+63)/64 trick.  We need to
		 * convert this back into bytes, so we multiply by
		 * 8 bytes/dword.  We also use CALLOC to make sure that
		 * the glop in the last byte is cleared.
		 * And well, just to make sure there's some pad on the
		 * end, we add another doubleword.
		 */
	    pLineBuf = (UInt8 *) CALLOC(1, ((((wD * 4) + 63) / 64) + 1) * 8);
	    if (pLineBuf == NULL)
		return ia_nomem;

		/* Set up offsets for the first 8x reduction.
		 * dstInitOffset points to the last dest doubleword.
		 * The dest has 1-bit pixels.  Note that the width of
		 * the dest here is wD * 4.
		 */
	    dstInitOffset = (((wD * 4) - 1)/64) * 8 ;
		/* it takes 8 source doublewords to make one dest doubleword */
	    srcInitOffset = dstInitOffset * 8;	

		/* we want a pointer to the last doubleword in the source line
		 * since this might not be a multiple of 8.  We use 64
		 * since there are 64 1-bit pixels per 64-bit doubleword.
		 * Since we don't have the real source width, we just assume
		 * it's 8 * (wD * 4).
		 */
	    srcLimit = ((((wD * 4) * 8) - 1)/64) * 8;


		/* Set up the corresponding offsets for the second stage
		 * 4x reduction.
		 */
	    dstInitOffset2 = ((wD - 1)/64) * 8 ;
		/* it takes 4 source doublewords to make one dest doubleword */
	    srcInitOffset2 = dstInitOffset2 * 4;	
	    srcLimit2 = (((wD * 4) - 1)/64) * 8;

	    ip_initLongOps (pIPState); /* do whatever's necessary to enter long
				        * operation mode.
				        */
	    for (row = hD; row > 0; row--, pD += dBpl, pS += sInc)
	    {
		ip_compress8ColsAndLineLong (pIPState, pS, 32 * wD, pLineBuf,
				             dstInitOffset, srcInitOffset,
					     srcLimit);
		ip_compress4ColsAndLineLong (pIPState, pLineBuf, 4 * wD, pD,
				             dstInitOffset2, srcInitOffset2,
					     srcLimit2);
	    }
		/* issue any opcodes necessary to exit from long access mode */
	    ip_finishLongOps (pIPState);

		/* We're done.  Free the intermediate line buffer */
	    FREE(pLineBuf);
	}
	break;
    }      

    NORMAL_RETURN;
}

/*

$Log: i_sbsmp2.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.6  1998/01/13 23:49:38  ddavies
i_sbsmp2.c -> /product/ipcore/ipshared/src/RCS/i_sbsmp2.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1996/11/04 21:46:40  ddavies
ddavies on Mon Nov 4 13:44:00 PST 1996

i_sbsmp2.c -> /product/ipcore/ipshared/src/RCS/i_sbsmp2.c,v

Modified ip_compressColsOr to use long operations (64-bit or longer) when they're
available.  It does this using ip_compressColsOrLineLing, ip_compress4ColsOrLineLong
and ip_compress8ColsOrLineLong.  Also use concatenation of these for 16x and 32x
reductions.
Also modified ip_compressColsAnd to use ip_compressColsAndLineLong,
ip_compress4ColsAndLineLong and ip_compress8ColsAndLineLong in the same way.

 * Revision 1.4  1995/12/13  02:45:52  ejaquith
 * i_sbsmp2.c -> /product/ipcore/ipshared/src/RCS/i_sbsmp2.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.3  1995/10/18  03:58:40  ddavies
 * i_sbsmp2.c -> /product/ipcore/ipshared/src/RCS/i_sbsmp2.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.2  1994/11/07  22:05:56  ddavies
 * i_sbsmp2.c -> /product/ipcore/ipshared/src/RCS/i_sbsmp2.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.1  1994/09/09  00:19:45  lperry
 * Initial revision
 *
 * Revision 1.5  1994/08/01  21:06:28  ddavies
 * i_subsmp.c -> /product/ipcore/ipshared/src/RCS/i_subsmp.c,v
 *
 * Change "static const" to "static" to work around buggy gcc/Purify interaction.
 *
 * Revision 1.4  1994/07/27  16:07:14  ejaquith
 * i_subsmp.c -> /product/ipcore/ipshared/src/RCS/i_subsmp.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Made functions use local pointers to tables instead of globals.
 *
 * Revision 1.3  1994/07/09  20:36:31  jcraig
 * jcraig on Sat Jul 9 13:26:40 PDT 1994
 *
 * i_subsmp.c -> /product/ipcore/ipshared/src/RCS/i_subsmp.c,v
 *
 * in i_subsample1bpp when reducing by 2 or 4, mask the last byte of
 * the line so that no extra bits will end up in the frame if the source
 * width isn't a multiple of the reduction factor.
 * in i_reduceVOrHAnd and i_reduceVAndHOr use CALLOC instead of MALLOC
 * so that purify won't complain about accessing uninitialized memory
 * (even though the word is never used).
 * 2624
 *
 * Revision 1.2  1994/02/04  20:17:53  dferrell
 * dferrell on Fri Feb 4 12:17:43 PST 1994
 *
 * i_subsmp.c -> /product/ipcore/ipshared/src/RCS/i_subsmp.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/02/03  00:11:46  ddavies
 * Initial revision
 *
 * Revision 1.19  1993/12/21  23:41:57  jcraig
 * jcraig on Tue Dec 21 15:41:16 PST 1993
 *
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * in i_reduceRowsAnd, i_reduceRowsSample and i_reduceRowsOr use
 * dBpl in the call to ip_compressRows* instead of sBpl
 *
 * Revision 1.18  1993/12/02  19:30:14  ddavies
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.17  1993/11/17  03:01:32  ddavies
 * ddavies on Tue Nov 16 18:59:18 PST 1993
 *
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * Fixed bug in ip_compressRowsArb (called from i_reduceRowsSample).  This bug
 * caused us to write each row 8 times as often as necessary.  By a process
 * that I still can't figure out, it also wrote trash into the right frame.
 * If the source and dest are both the same bytes/line, don't the frames always
 * line up, even if we copy many source lines to the dest in a block?  Well,
 * anyway, it works now.
 *
 * Revision 1.16  1993/11/16  03:10:05  ddavies
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * The 1bit/pixel, 8x subsample code was putting the rightmost byte together
 * incorrectly - the bits were being packed into the right end of the byte
 * instead of into the left end.
 * This caused it to write into the right frame.
 * The 4 bit/pixel subsampling at 8x, 16x and 32x was also writing into the
 * right frame.  These routines were producing 2 lines of output each time,
 * then shifting over by one line and doing it again - basically doing the
 * entire image twice!  They were also carefully composing the last word
 * of the line, then writing it whether it was needed or not.
 *
 * Revision 1.15  1993/11/05  02:26:04  ddavies
 * ssubsmpl.c -> /product/ipcore/ipshared/src/RCS/ssubsmpl.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


