/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * i_scgarb:	Scale-to-gray, arbitrary downsampling
 *
 * Exports:
 *	Int32	    i_scaleToGray4bppArb()
 *	Int32	    i_makeScaleToGrayArbFilters()
 *	Int32	    i_makeScaleToGrayArbLineTab()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

	/* prototypes */
#include "shrscgr.pub"
#include "shrscgr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_scgarb.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");

/*
 * This code is used to do an arbitrary scaling (always scaling down)
 * from a binary image to a 4-bit gray image. The reason for the existance
 * of this code is that when displaying high-resolution bitmaps on the
 * screen, one often needs to do a scale-to-width, which won't result in
 * a nice scaling ratio, and scaling to the nearest ratio and then 
 * doing a non-interpolated resample results in really ugly images.
 *
 * We'd also like to implement a slightly better filter than in the
 * current scale-to-gray code, which uses a box filter. The problem
 * with a box filter is that it lets through too much of the stopband,
 * resulting in quite a bit of aliasing. There are many filters we could
 * use, the idea one being a sin(x)/x filter. The problem with such filters
 * is that they require huge kernels. For example, a 2-lobe Lanczos-windowed
 * sin(x)/x filter to decimate 4x requires a 16x16 kernel. So perhaps we'll do a
 * simple tent filter (i.e. center weighted) to decrease the passband
 * a bit. We should observe that since we're going to implement the
 * summation of filter coefficients by a look-up, doing a look-up of
 * a larger number of bits obviously doesn't take any more time than
 * a look-up of a smaller number of bits, only more space in the table.
 * So one thing we might try is doing a wider filter in the horizontal
 * direction, say a sin(x)/x or some such (depending on the reduction
 * factor--it gets prohibitive at large reduction factors) and do
 * a smaller filter in the vertical direction, which might be 
 * reasonable, since it has been alleged that we perceive frequency
 * better horizontally. 
 *
 * Certainly, a simple tent filter over the size that we currently
 * use or perhaps slightly wider would be better than a box filter.
 * Since speed is quite important in this case, I'm going to opt for the
 * tent filter, at least in an initial implementation. The code should
 * be written so that we can vary the max filter size and shape
 * fairly easily, so that we can experiment.
 *
 * Since this code will often be used for displaying bitmaps on the screen,
 * including zoomed views, it would probably be useful to support
 * the scaling of arbitrary rectangles out of the image, rather than
 * requiring the caller to cut out a piece and then scale that.
 *
 * Unlike the fixed scale-to-gray stuff, we're going to need to compute
 * offsets into each line for each set of input pixels needed for
 * each output pixel. In addition, we should probably have filters
 * corresponding to more than one phase, maybe 0-1 cycle in 1/4 
 * cycle increments, which would yield 16 filters (4 increments in 
 * x, 4 in y). 
 *
 * For each output line, we can compute the starting input line and 
 * y-phase in the code. We'll encode the x-phase and the x-offset into input
 * lines for each output pixel in a table. The contents of this table is:
 * (for each pixel in output width) 
 *
 *	Longword offset for starting word
 *	shift count for first word's bits
 *	mask for first word's bits
 *	shift count for second word's bits
 *	mask for second word's bits
 *	x-phase, 0-3
 *
 * For a tent filter, we'll usually need ceil[2x the subsampling factor]
 * pixels in each direction. We should pin the max at some reasonable
 * value, at least in Y. The X max is limited by how large one wants the
 * table to grow.
 *
 * Our algorithm, therefore, is similar to the existing fixed-ratio
 * scale-to-gray algorithms, except that for each output pixel,
 * we compute an offset into the input line(s) and get one or two words,
 * shifting and masking each into a value to look up in the table 
 * (also indexed by the x-phase, the set of x-phases for a given
 * y-phase should already have been looked up at the start of the line)
 * which is summed into our output value. The values in the table will
 * be fixed-point integers, with three bytes of fraction and one byte
 * of whole number.
 *
 * Note that there are two calls to create look-up tables. One creates
 * the filter look-up tables for each of the 16 x-y phases given a 
 * scale factor, and one creates a line offset-shift-mask table 
 * for a given scale factor, image size, and rectangle size and offset.
 * (Actually, the latter is only dependent on the image width, and
 * can be used over, for example when doing banded processing.)
 *
 * It would also, I suppose, be nice to be able to specify the type
 * of filter to use, or at least some "quality" knob passed in by 
 * caller in order to make a speed/quality trade-off.
 *
 * WARNING: I'm going to use FLOATING POINT in places where it won't
 * affect speed. I don't think any products that can't use floating point
 * (i.e. Reading Edge) will want to use this code.
 */

/*****************************************************************
 *
 * i_scaleToGray4bppArb ()
 *
 *	Do an arbitrary scale-to-gray from a binary image to a 
 *	4 bit/pixel image. See description above for some
 *	explanation of the general scheme. 
 *
 *      4/5/95: The original implementation took a source window, a source
 *      rectangle, and a destination image. That worked fine unless
 *      one was, for example, scrolling an image through a window.
 *      In that case, the scale factor is constant, but the source
 *      offset might not lie on a pixel boundary.  We've now changed
 *      it so that we take a source image, and a "virtual"
 *      image size, which is used to determine the scale factor, and then
 *      we take a destination offset and size, which is relative to the
 *      "virtual" image size. This avoids having the caller need to
 *      keep track of fractional pixels.
 *
 * Arguments:
 *	pD	-	Pointer to frame of destination image
 *      yD      -       Y-Offset into "virtual" image (without frame) 
 *	wD, hD	-	Width and height of destination image
 *	dBpl	-	Destination image distance between lines in bytes
 *	pS	-	Pointer to frame of source image 
 *	sBpl	-	Source image distance between lines, in bytes
 *	filter	-	Filter returned by i_makeScaleToGray4bppArbFilters
 *	lineTab	-	Line table returned by i_makeScaleToGray4bppArbLineTab
 *
 * Returns:
 *	ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
i_scaleToGray4bppArb (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	yD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    void       *filter,
    void       *lineTab)
{
    PROLOGUE;
    FilterDesc *fDesc = (FilterDesc *) filter;
    Int32	row, filterOffset, rowStart, filterHeight,
		yPhase, *filterData, *filterRow, filterNextRow, sWpl, 
		sourceOffset;
    UInt32	*sourceP, *destP;
    Float64	sourceRow, rowInc;

    /* Move the dest. pointer past the frame */
    pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;

    rowInc = 1.0 / fDesc->ratioY;
    sourceRow = yD * rowInc + cFrameBits;
    sWpl = sBpl / 4;
    filterOffset = fDesc->width / 2;
    filterData = &fDesc->filters[0];
    filterHeight = fDesc->height;
    filterNextRow = (1 << fDesc->width);
    
    for (row = 0; row < hD; row++)
    {
	rowStart = (Int32)((sourceRow - filterOffset) + cEpsilon);
	yPhase = (Int32)((sourceRow - floor (sourceRow)) * fDesc->phases);

	/* Move to the correct row of filters */
	filterRow = filterData + (yPhase * fDesc->phases * fDesc->filterSize);

	/* Move to the right row of source image */
	sourceOffset = ((Int32) rowStart) * sBpl;
	sourceP = (UInt32 *) (pS + sourceOffset);
	destP = (UInt32 *) pD;
	ip_scaleToGrayArb4bppLine (pIPState, destP, sourceP,
				   sWpl, fDesc->width, 
				   filterHeight,
				   wD, lineTab, filterRow, filterNextRow);
			       
	pD += dBpl;
	sourceRow += rowInc;
    }
		    
    NORMAL_RETURN;
    
}

/***********************************************************************
 *
 * i_makeScaleToGray4bppArbFilters()
 *
 * Make filters for arbitrary scaling-to-gray. Since we're scaling
 * binary images, the implementation of filter coefficient multiply
 * and row addition is implemented by a look-up for the bits of the
 * row. 
 *
 * The user passes in a function to generate coefficients for the filter
 * at a given (x,y) phase, and a function to compute the maximum size
 * of the filter for a given reduction factor. This routine turns
 * the filter coefficients for each phase into a series of look-up
 * tables for the bit patterns of each row in the source image corresponding
 * to an output pixel. 
 *  
 * ----
 * This used to take a count of phases, but I've changed it so that
 * it always does a zero phase filter only. This means that unless one
 * is doing integer scale ratios, one should probably only use box filters.
 *
 * Arguments:
 *	wD, wS,      
 *      hD, hS      -	    width and  height of dest. and source images,
 *			    or a rational representation of the scale
 *			    factor, if you prefer.
 *	filterGen   -	    Pointer to filter generation function
 *	filterInfo  -	    Pointer to filter info function
 *			    There are standard functions to implement
 *			    box, tent, and Lanczos-windowed sinc filters.
 *	filterParm  -	    A tunable parameter for the filter, the
 *			    interpretation of which is filter-dependent
 *	maxWidth    -	    Maximum width of a filter. A scale factor
 *			    which calls for a very huge filter will
 *			    get truncated to this width, and we'll
 *			    attempt to make do in the filter implementation,
 *			    though it will alias.
 *	maxHeight   -	    Max height of the filter. This is separate from 
 *			    width so that we can support unsquare filters.
 *	pFilter    -	    Filter structure handle, allocated by the caller 
 *			    (after asking i_scaleToGrayArbFilterBytes)
 *			    and filled in here.
 *
 * Returns:
 *	ia_successful or error code
 *
 ***********************************************************************/

Int32 CDECL 
i_makeScaleToGrayArbFilters (
    IPState    *pIPState,
    Int32	wD,
    Int32	wS,
    Int32	hD,
    Int32	hS,
    Int32	depth,
    FilterGen	filterGen,
    FilterInfo	filterInfo,
    Float64	filterParm,
    Int32	maxWidth,
    Int32	maxHeight,
    void       *pFilter)
{
    PROLOGUE;
    FilterDesc	*fDesc;
    Int32	*filterData, maxVal;

    /* Check parameters. Make more descriptive errors for these &&& */
    if (wD > wS)
	return ia_invalidParm;

    if (depth != 4 && depth != 8)
	return ia_invalidParm;

    fDesc = (FilterDesc *) pFilter;
    fDesc->phases = /* phases */ 1;
    maxVal = (1 << depth) - 1;

    /* Compute our decimation ratio */
    fDesc->ratioX = (Float64) wD / (Float64) wS;
    fDesc->ratioY = (Float64) hD / (Float64) hS;

    /* How wide should the filter be? */
    (*filterInfo)(pIPState, fDesc->ratioX, fDesc->ratioY, filterParm, 
		  &fDesc->idealWidth, &fDesc->idealHeight);

    /* Pin the actual size */
    fDesc->width = MIN(fDesc->idealWidth, maxWidth);
    fDesc->height = MIN(fDesc->idealHeight, maxHeight);
    /* Compute the size of the filter, in longwords */
    fDesc->filterSize = fDesc->height * (1 << fDesc->width);

    /* Fill in the filter info */
    filterData = &fDesc->filters[0];

    /* Fill the LUT given the filter description */
    ip_generateFilterLUT (pIPState, fDesc, 0, 0, maxVal, filterGen, 
			  filterParm, filterData);
    NORMAL_RETURN;
}

    
/***********************************************************************
 *
 * i_scaleToGrayArbFilterBytes
 *
 *	Computes the size, in bytes, of block that must be allocated for
 *	i_makeScaleToGray4bppArbFilters. (It's done this way to support
 *	applications like the DAE which must do their own allocation)
 * 
 *	wD, wS	    
 *      hD, hS      -	    width and  height of dest. and source images,
 *			    or a rational representation of the scale
 *			    factor, if you prefer 
 *	filterInfo  -	    Pointer to filter info function
 *	filterParm  -	    A tunable parameter for the filter, the
 *			    interpretation of which is filter-dependent
 *	maxWidth    -	    Maximum width of a filter. A scale factor
 *			    which calls for a very huge filter will
 *			    get truncated to this width, and we'll
 *			    attempt to make do in the filter implementation,
 *			    though it will alias.
 *	maxHeight   -	    Max height of the filter. This is separate from 
 *			    width so that we can support unsquare filters.
 *
 * Returns:
 *	number of bytes, 0 = error 
 *
 ***********************************************************************/

Int32 CDECL 
i_makeScaleToGrayArbFilterBytes (
    IPState    *pIPState,
    Int32	wD,
    Int32	wS,
    Int32	hD,
    Int32	hS,
    FilterInfo	filterInfo,
    Float64	filterParm,
    Int32	maxWidth,
    Int32	maxHeight)
{
    PROLOGUE;
    Int32	 bytes, bytesPer;
    Int32	 width, height;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    if (wD > wS)
	return ia_invalidParm;

    (*filterInfo)(pIPState, (Float64) wD / (Float64) wS,
		  (Float64) hD / (Float64) hS, 
		  filterParm, &width, &height);

    width = MIN (width, maxWidth);
    height = MIN (height, maxHeight);
    bytesPer = sizeof(Int32) * height * (1 << width);
    bytes = sizeof (FilterDesc) + bytesPer;
    return bytes;
}

/***********************************************************************
 *
 * i_makeScaleToGrayArbLineTabBytes
 *
 *	Computes the size, in bytes, of block that must be allocated for
 *	i_makeScaleToGray4bppArbLineTAb. (It's done this way to support
 *	applications like the DAE which must do their own allocation)
 * 
 *	filter	-	    The filter being used, created by above
 *      xD      -           X-Offset into "virtual" image for dest.
 *	wD	-	    width (or height) of dest image
 *      wVirt   -           Width of "virtual" image
 *	wS	-	    Width of source rectangle
 *	sBpl	-	    inter-line distance, in bytes, in source image
 *
 * Returns:
 *	number of bytes, 0 = error 
 *
 ***********************************************************************/

Int32 CDECL 
i_makeScaleToGrayArbLineTabBytes (
    IPState    *pIPState,
    void       *filter,
    Int32	xD,
    Int32	wD,
    Int32	wVirt,
    Int32	wS,
    Int32	sBpl)
{
    PROLOGUE;
	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    return ip_scaleToGrayArbLineTabBytes (pIPState, filter,
					  xD, wD, wVirt, wS, sBpl);
}

/***********************************************************************
 * 
 * i_makeScaleToGrayArbLineTab
 *
 *	This either makes a "line table" with offset and shifts,
 *	or code to produce a line, depending on whether we're running 
 *	machine-independent code or not. In any case, the result should be
 *	passed to i_scaleToGray4bppArb ().
 *
 * Arguments:
 *	filter  -	filter LUT returned by i_makeScaleToGrayArbFilter
 *      xD      -       X-offset into "virtual" image (without frame)
 *	wD	-	Width of destination image
 *      wDVirt  -       Width of "virtual" destination image
 *	wS	-	Width of source rectangle
 *	sBpl	-	Source image distance between lines, in bytes
 *	lineTab	-	Line table structure, allocated by caller,
 *			size returned by i_makeScaleToGrayArbLineTabBytes
 *	
 * Returns:
 *	ia_successful or error code
 *
 ***********************************************************************/

Int32 CDECL 
i_makeScaleToGrayArbLineTab (
    IPState    *pIPState,
    void       *filter,
    Int32	xD,
    Int32	wD,
    Int32	wDVirt,
    Int32	wS,
    Int32	sBpl,
    void       *lTab)
{
    PROLOGUE;
    ERR_PASS( ip_makeScaleToGrayArbLineTab (pIPState, filter, xD, wD,
					    wDVirt, wS, sBpl, lTab)); 
    NORMAL_RETURN;
}

/***********************************************************************
 * 
 * ip_generateFilterLUT
 *
 *	We are passed phase for x and y, and a filter-description function
 *	which gives us the floating-point coefficients for the filter
 *	at the given reduction factor and phases. Given that filter
 *	descriptor, we need to make a look-up table for each row of
 *	the filter, which will map bits in the source image to the
 *	corresponding sum of weights that one would get, scaled
 *	properly so that 0=0 and 1.0=maxVal.
 *	
 * Arguments:
 *	fDesc	    -	    Filter descriptor
 *	rowPhase    -	    y-phase, [0, phases-1]
 *	colPhase    -	    x-phase, [0, phases-1]
 *	maxVal	    -	    Maximum value of a pixel
 *	filterGen   -	    Filter generation function 
 *	filterParm  -	    A tunable parameter for the filter, the
 *			    interpretation of which is filter-dependent
 *	filterData  -	    Place to stuff filter data
 *
 * Returns:
 *	ia_successful or error status
 ***********************************************************************/

Int32 CDECL 
ip_generateFilterLUT (
    IPState    *pIPState,
    FilterDesc *fDesc,
    Int32	rowPhase,
    Int32	colPhase,
    Int32	maxVal,
    FilterGen	filterGen,
    Float64	filterParm,
    Int32      *filterData)
{
    PROLOGUE;
    Float64	*coefs, rowPhi, colPhi, rowSum, *coefP, rowScaleFactor;
    Int32	 row, bits, i, colMax, scaledValue;
    UInt32	 bitMask;

    coefs = (Float64 *) MALLOC (sizeof(Float64) * fDesc->width * fDesc->height);
    if (!coefs)
	return ia_nomem;

    colMax = (1 << fDesc->width) - 1;

    /* Compute the phases [0-1) instead of [0,phases-1] */
    rowPhi = (Float64) rowPhase / (Float64) fDesc->phases ;
    colPhi = (Float64) colPhase / (Float64) fDesc->phases ;

    /* 
     * Generate the coefficients. Note that if the phase is zero, thereby
     * aligning the center with a pixel, the filter might be smaller
     * than the width/height by 1. In this case, the filter is always
     * left-aligned. We don't need to worry about that here, just
     * in the code that generates the line table.
     *
     */
    (*filterGen)(pIPState, fDesc->ratioX, fDesc->width, fDesc->height, 
		 fDesc->idealWidth, fDesc->idealHeight,
		 rowPhi, colPhi, filterParm, coefs);

    /* Generate the rows */
    coefP = &coefs[0];
    /* Or is it maxVal+1...? */
    rowScaleFactor = 16777216.0 * (Float64) maxVal;

    for (row = 0; row < fDesc->height; row++)
    {
	/* Iterate through possible bit combinations in this row */

	for (bits = 0; bits <= colMax; bits++)
	{
	    rowSum = 0.0;
	    bitMask = 1 << (fDesc->width - 1);

	    for (i = 0; i < fDesc->width; i++)
	    {
		/* If this bit is on, add in its coefficient. */
		if (bits & bitMask)
		    rowSum += coefP[i];

		bitMask >>= 1;
	    }

	    /*
	     * Scale that sum and stuff it into the look-up table.
	     * Rounding might become an issue here--we want to avoid never
	     * being able to get maxVal. The format of the scaled value
	     * is xx . yy yy yy, where "yyyyyy" are the six hex digits
	     * of fraction. Note that the rowSum, and therefore scaledValue,
	     * CAN be negative. 
	     */

	    scaledValue = (Int32)(rowSum * rowScaleFactor + 0.9);

	    /* Stuff that value into the LUT */
	    filterData[(~bits) & colMax] = scaledValue;
	}

	filterData += (colMax + 1);

	/* Move on to the next row of coefficients */
	coefP += fDesc->width;
    }

    FREE (coefs);

    NORMAL_RETURN;
    
}

/*
$Log: i_scgarb.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.8  1998/01/13 23:49:40  ddavies
i_scgarb.c -> /product/ipcore/ipshared/src/RCS/i_scgarb.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1997/11/12 18:03:17  xzhu
xzhu on Wed Nov 12 09:52:58 PST 1997

i_scgarb.c -> /product/ipcore/ipshared/src/RCS/i_scgarb.c,v

Fixed crash when reductions is different in X and Y. Use reductionY to scale down lines. Filter size is estimated by both reductionX and reductionY. If want scale down image with one retio, just input reductionX (reductionX, reductionX, ) instead of (reductionX, reductionY,  ).

Revision 1.6  1995/12/13 02:45:55  ejaquith
i_scgarb.c -> /product/ipcore/ipshared/src/RCS/i_scgarb.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/10/16  17:51:13  dferrell
 * i_scgarb.c -> /product/ipcore/ipshared/src/RCS/i_scgarb.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.4  1995/04/06  17:09:21  macleod
 * macleod on Thu Apr 6 10:05:25 PDT 1995
 *
 * i_scgarb.c -> /product/ipcore/ipshared/src/RCS/i_scgarb.c,v
 *
 * Changed the interfaces to i_scaleToGray4bpArb, i_makeScaleToGrayArbLineTab,
 * and i_makeScaleToGrayArbLineTabBytes.
 * The original version supported a source window and a destination image.
 * The interface was changed to take a source image, a "virtual" destination
 * width, which determines the scaling factor, and destination x,y offset
 * into the "virtual" image, allowing one to land on integer pixel boundaries
 * in the scaled image, which may cause pixels to land on fractional
 * boundaries in the source image.
 * Note that the line table must be re-computed every time the
 * x-offset or destination width changes, i.e. you can use the same line
 * table when scrolling vertically, but not when scrolling horizontally.
 * Changed a couple of places where floats were truncated to add in an
 * epsilon beforehand, so that one lands on the same pixel whether doing the
 * whole image or taking a small piece.
 *
 * Revision 1.3  1995/01/13  02:01:59  macleod
 * macleod on Thu Jan 12 17:55:03 PST 1995
 *
 * i_scgarb.c -> /product/ipcore/ipshared/src/RCS/i_scgarb.c,v
 *
 * Got rid of the "phases" parameter to the filter-creation functions.
 * This parameter allowed one to generate filters with more than one
 * interpolated phase, so that a filter landing on a fractional pixel
 * is weighted correctly. It was taken out so that the function could
 * run faster. The consequence is that, in general, one should only use
 * a box filter rather than the fancier filters. In fact, even with the
 * box filter, having the filter not weighted between source pixels does
 * introduce some error, but it's probably not worth worrying about.
 *
 * Revision 1.2  1995/01/04  18:45:35  macleod
 * macleod on Wed Jan 4 10:44:59 PST 1995
 *
 * i_scgarb.c -> /product/ipcore/ipshared/src/RCS/i_scgarb.c,v
 *
 * forgot the #ifdef __GNUC__ around the #include "ansiprot.h"
 *
 * Revision 1.1  1995/01/04  01:12:49  macleod
 * Initial revision
 *
*/
