/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_scgr41.c: wrapper for scale to 4bpp gray, 1.5x reduction
 *
 *          1.5x scale to gray reduction:
 *                  Int32	i_scaleToGray4bpp1_5x()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_scgr41.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* This is the wrapper for code doing 1.5x reduction from binary to gray image.
 * The input is a binary image.  The output is a 4 bit/pixel gray image.
 * The output image is 1/1.5 the width and 1/1.5 the height of the input image.
 *
 * The basic method is to use  6-bit block of input bits as an index
 * into one of two tables.  The table entries for two vertically adjacent
 * rows are summed, producing the gray output pixels.
 *
 * Unfortunately, the details are a little harder.  We want to repeatedly
 * map 9 binary pixels (3x3) in the input to 4 gray pixels (2x2) in the output.
 * This is equivalent to mulitplying the width and height by 2/3, or 1/1.5.
 * Here's a picture of 9 binary input pixels mapped to 4 output pixels:
 *
 * 	0  1  2		0    1
 *	3  4  5     =>
 *	6  7  8		2    3
 *
 * The four gray output pixels would ideally be calculated as follows:
 * 
 *	Out0 = (In0 + In1/2 + In3/2 + In4/4) * 15/2.25
 *	Out1 = (In2 + In1/2 + In5/2 + In4/4) * 15/2.25
 *	Out2 = (In6 + In7/2 + In3/2 + In4/4) * 15/2.25
 *	Out3 = (In8 + In7/2 + In5/2 + In4/4) * 15/2.25
 *
 * This mapping is perfectly linear and covers the entire range of outputs.
 * How do we approximate it efficiently?
 *
 * We use two tables, one for the first and third input rows, another for the
 * second input row.  Each table is addressed with 6 bits, so we will map
 * two groups of 3 pixels into 2 gray pixels.  Let's just look at the
 * weights used to translate one set of 3 pixels into 2 gray pixels.  In
 * fact, lets just look at the weights used to calculate pixel Out0 above;
 * the other weights in the table are calculated symmetrically.
 *
 * If we can meet these goals, we'll be happy:
 *	1. When none of the pixels are active, the result is 0.
 *	2. When all of the pixels are active, the result is 15.
 *	3. There is no need to check for rollover.
 *	4. The mapping from number of input pixels to output value is
 *	   roughly linear (within roundoff error) and monotonic.
 *	5. The mapping is only a function of the number of pixels active,
 *	   not their position.  For example, pixels In1 and In3 should
 *	   count equally to the sum.
 *	6. Several sums can be computed in parallel.
 *	7. Any fixup effort required at the end of the summing process,
 *	   like to add in carries from fractional parts, is minimized.
 *
 * One way to look at the problem is that we'd like to weight each pixel
 * by (4/9) * 15.  That's hard to calculate, but 4/8 is trivial.  However, then
 * the sum of the weights is too big.  To compensate, we drop the LS bit
 * of the 15, making it (4/8) * 14.  This happens to work as seen below
 * The weights we'll use:
 *  Pixel	weighted sum	In binary
 *   In0		7.0	0111.00
 *   In1		3.5	0011.10
 *   In3		3.5	0011.10
 *   In4	        1.75	0001.11
 *
 * Another set of mappings that doesn't use any fraction bits is:
 *  Pixel	weighted sum	In binary
 *   In0		7	0111
 *   In1		3	0011
 *   In3		3	0011
 *   In4	        2	0010
 *
 *
 * In the table below, we calculate the perfect sum of pixel values, the
 * sum found using the first set of weights then truncating and the sum
 * found using the second set of weights (no truncation required):
 *
 * In0 In1 In3 In4	Ideal	Sum1	Sum2
 *   0   0   0   0	0.00	0	0
 *   0   0   0   1	1.67	1	2
 *   0   0   1   0	3.33	3	3
 *   0   0   1   1	5.00	5	5
 *   0   1   0   0	3.33	3	3
 *   0   1   0   1	5.00	5	5
 *   0   1   1   0	6.67	7	6
 *   0   1   1   1	8.34	8	8
 *   1   0   0   0	6.67	7	7
 *   1   0   0   1	8.34	8	9
 *   1   0   1   0	10.00	10	10
 *   1   0   1   1	11.67	12	12
 *   1   1   0   0	10.00	10	10
 *   1   1   0   1	11.67	12	12
 *   1   1   1   0	13.33	14	13
 *   1   1   1   1	15.00	15	15
 *
 * The second mapping seems fine to me and doesn't involve any fraction
 * bits, so is faster to compute.  We'll use it.
 *
 * How does the algorithm work?
 *
 *	1. Read an input image 6-bit block.  The byte is divided into 2
 *	   triples of input pixels.
 *	2. Use it to index an array of 32-bit words.  The lower half of
 *	   each word contains four 4-bit numbers which correspond to the
 *	   integer parts of summed weights associated with each pair of
 *	   input pixels.
 *	3. Read the input byte directly below the one read first.
 *	4. Look it up in the the other half of the table since it
 *	   corresponds to the pixels that are weighted by 1/2 and 1/4
 *	   instead of 1.0 and 1/2.
 *	5. Add the two table-lookup entries.  This produces the summed
 *	   weights for 1 pair of output pixels.
 *
 * Now things start getting funny.  We produce 4 output pixels each time
 * we do steps 1 through 5.  We want to write 8 output pixels at a time.
 * On the other hand, we need groups of 6 input pixels.  32 is not a
 * multiple of 6.  So we perform steps 1..5 in the following sequence:
 *	Read bits [7:2] from byte 0, produce 4 pixels
 *	Combine byte0.[1:0] and byte1.[7:4], produce 4 more, write 1 word
 *	Combine byte1.[3:0] and byte2.[7:6}, produce 4
 *	Use byte2.[5:0] to produce 4 more, write 1 word.
 *
 * We use this sequence to produce a line of outputs.  We then advance
 * down *1* line in both the input and output images; and negate the
 * offset used to get to the next input line when performing step 3
 * above.  This allows us to use the second line of input image
 * twice.  After producing another line of output, we skip a line of input
 * entirely.
 *
 * As a matter of software engineering, this file contains the piece of code
 * that accumulates some image and calls the lower level routine that
 * produces one output line.  When we switch between banded images and
 * whole images, we can just use different routines here while using the
 * same real routine below.
 *
 * The first (only?) user of this function is EasyScan.  They will want to
 * apply the function to small images (extracted from large images) many
 * times a second.  In this world, it's a bad idea to create and destroy
 * the table every time we're called.  A separate procedure,
 * i_createScaleToGrayTab4bpp2x is called by the user to create the table
 * beforehand.  A pointer to this table is passed down.  If the pointer is
 * NULL, the table is created, used and destroyed before returning.
 */

/*********************************************************************
 ************       Scale To 4 bpp Gray, 1.5x Reduction    *************
 *********************************************************************/

/* PUBLIC */
/**************************************************************************
 *  i_scaleToGray4bpp1_5x()  
 *		Each 3x3 square of the source binary image is mapped
 *		to a 2x2 square of 4-bit output pixels.  The value of
 *		the output pixels is a (roughly) linear function of the
 *		number of bits turned on in the input square.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(includes frame)
 *		wD =	width in pixels of dest (including frame)
 *		hD =	height of dest (including frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(includes frame)
 *		sBpl =	source image bytes/line
 *		pScaleTable = ptr to table used to do scaling
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	not enough memory for the table
 *		ia_internal	internal error during table construction
 **************************************************************************/
Int32  CDECL 
i_scaleToGray4bpp1_5x (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    UInt32     *pScaleTable)
{
PROLOGUE;
 UInt32	 destroyTable = cFalse;	/* Boolean, cTrue when we should
				 * destroy the table before returning */
	/* compute the mask used to get rid of the bits on the right
	 * end of the line. */
UInt32	 rightMask = 0xFFFFFFFF << (4*((8 - (wD & 0x7)) & 0x7));
UInt32	 tableSizeInBytes;

	/* Unframe the image and pointers */
    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;

	/* Do we have a conversion table? */
    if (pScaleTable == NULL)
    {
	destroyTable = cTrue;
	if (i_getSizeOfScaleToGray4bpp1_5xTab (pIPState, &tableSizeInBytes)
		!= ia_successful)
	    return ia_internal;
	pScaleTable = (UInt32 *) MALLOC(tableSizeInBytes);
	if (pScaleTable == NULL)
	    return ia_nomem;
	if (i_fillScaleToGray4bpp1_5xTab (pIPState, pScaleTable,
					  tableSizeInBytes)
		!= ia_successful)
	    return ia_internal;
    }

	/* loop through the rows of the images, calling
	 * ip_scaleToGray4bpp1_5xLine to reduce each triple of input
	 * lines to a pair of output lines.
	 */
    for (; hD > 0; hD--, pD += dBpl, pS += sBpl)
    {
	ip_scaleToGray4bpp1_5xLine (pIPState, pD, wD, pS, sBpl,
				    pScaleTable, rightMask);
	hD--;
	if (hD > 0)	/* there's another line to do.  Point to the second
			 * input line as the major input and the line
			 * *above* it as the minor input.
			 */
	{
	    pS += 2*sBpl;
	    pD += dBpl;
	    ip_scaleToGray4bpp1_5xLine (pIPState, pD, wD, pS, -sBpl,
				        pScaleTable, rightMask);
	}
    }

    if (destroyTable)
	FREE(pScaleTable);

    NORMAL_RETURN;
}

/*

$Log: i_scgr41.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.8  1998/01/13 23:49:41  ddavies
i_scgr41.c -> /product/ipcore/ipshared/src/RCS/i_scgr41.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1996/03/26 22:42:16  ddavies
i_scgr41.c -> /product/ipcore/ipshared/src/RCS/i_scgr41.c,v

Change i_scaleToGray4bpp1_5x so that it allocates its table differently
if needed.  If the table pointers is NULL, it gets the size of the
table using i_getSizeOfScaleToGray4bpp1_5x, allocates it using MALLOC
(as before) and fills it using i_fillScaleToGray4bpp1_5xTab.  This
reflects the demise of i_makeScaleToGray4bpp1_5xTab.

Revision 1.6  1995/12/13 02:45:57  ejaquith
i_scgr41.c -> /product/ipcore/ipshared/src/RCS/i_scgr41.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1994/02/04  20:16:48  dferrell
 * i_scgr41.c -> /product/ipcore/ipshared/src/RCS/i_scgr41.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:23:46  ddavies
 * i_scgr41.c -> /product/ipcore/ipshared/src/RCS/i_scgr41.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:29:06  ddavies
 * i_scgr41.c -> /product/ipcore/ipshared/src/RCS/i_scgr41.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

