/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  i_scgr42.c: wrapper for scale to 4bpp gray, 2x reduction
 *
 *          2x scale to gray reduction:
 *                  Int32	i_scaleToGray4bpp2x()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_scgr42.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* This is the wrapper for code doing 2x reduction from binary to gray image.
 * The input is a binary image.  The output is a 4 bit/pixel gray image.
 * The output image is 1/2 the width and 1/2 the height of the input image.
 *
 * The basic method is to use each byte of input as an index into a table.
 * The table entries for two vertically adjacent rows are summed, producing
 * the gray output pixels.
 *
 * Unfortunately, the details are a little harder.  We want to repeatedly
 * map 4 binary pixels in the input to 1 gray pixel in the output.
 * The group of 4 binary pixels can take one of *5* values: 0, 1, 2, 3, or 
 * 4 pixels turned on.  The output pixel has a range from 0 to 15.  To the
 * extent possible, we want the input range to linearly to the output range.
 * The mapping should depend only on the number of input 1's, not on their
 * position within the 2x2 square.  A nice mapping is:
 *
 * Number of input pixels	Output value
 *		0		0
 *		1		3
 *		2		7
 *		3		11
 *		4		15
 *
 * This mapping is reasonably linear and covers the entire range of output.
 *
 * For the sake of efficiency, we would like to process as many pixels at
 * once as possible.  Table lookup is a good way to space out the pixels and
 * do some computing.  On the other hand, we don't like to make huge tables,
 * so we like to restrict the number of pixels used to index the tables.
 * We often use 8 as a nice table index size.
 *
 * After thinking about several schemes, we settled on one that weights each
 * 1 in the input by 15/4.  The weights are:
 *  # of pix		weighted sum	In binary
 *	0		0		0000.00
 *	1		3.75		0011.11
 *	2		7.5		0111.10
 *	3		11.25		1011.01
 *	4		15.0		1111.00
 *
 * We use a table to compute the summed weights.
 * It would be conceptually simple to keep each pixel as a fixed point, 6 bit
 * binary number.  However, after computing the fixed point sum, we would need
 * to discard the 2 fraction bits.  Consolidating a vector of 6 bit pixels
 * into a vector of 4 bit pixels is a pain.  Instead, we'll keep each
 * quantity in two separate 4 bit numbers.  One vector of packed 4-bit
 * numbers will contain the integer part of the sums  Say this fills the
 * lower 16 bits of a register..  The other vector of 4-bit numbers
 * will each contain 2 fraction bits and 2 integer bits.  These last two
 * integer bits are used to hold the carries when the fraction bits are
 * summed.  Ok, in this case we really only need 1 bit for that, but we use
 * the other bit for spacing so we can align all the carries with the
 * integer parts in the lower half of the register with a single shift.
 *
 * How does the algorithm work?
 *
 *	1. Read an input image byte.  The byte is divided into 4 pairs
 *	   of input pixels.
 *	2. Use it to index an array of 32-bit words.  The lower half of
 *	   each word contains four 4-bit numbers which correspond to the
 *	   integer parts of  summed weights associated with each pair of
 *	   input pixels.  The upper half of the table contains the
 *	   four fractional parts of the summed weights.
 *	3. Read the input byte directly below the one read first.
 *	4. Look it up in the same table.
 *	5. Add the two table-lookup entries.  This produces the summed
 *	   weights for each 2x2 square of input pixels.
 *	6. Shift the sum word right by 16 and mask out the fractional bits
 *	   leaving only the carries.
 *	7. Add the carries into the integer parts.  This produces a vector
 *	   of 4 pixels.
 *	8. Repeat steps 1 through 7 for the next vertical stack of 2 bytes
 *	   on the right.
 *	   Shift the first sum of inputs left by 16 and add in the second
 *	   set, then write the result.
 *
 * As a matter of software engineering, this file contains the piece of code
 * that accumulates some image and calls the lower level routine that
 * produces one output line.  When we switch between banded images and
 * whole images, we can just use different routines here while using the
 * same real routine below.
 *
 * The first (only?) user of this function is EasyScan.  They will want to
 * apply the function to small images (extracted from large images) many
 * times a second.  In this world, it's a bad idea to create and destroy
 * the table every time we're called.  A separate procedure,
 * i_createScaleToGrayTab4bpp2x is called by the user to create the table
 * beforehand.  A pointer to this table is passed down.  If the pointer is
 * NULL, the table is created, used and destroyed before returning.
 */

/*********************************************************************
 ************       Scale To 4 bpp Gray, 2x Reduction    *************
 *********************************************************************/

/* PUBLIC */
/**************************************************************************
 *  i_scaleToGray4bpp2x()  
 *		Each 2x2 square of the source binary image is mapped
 *		to a 4-bit output pixel.  The value of the output
 *		pixel is a (roughly) linear function of the number of
 *		bits turned on in the 2x2 input square.
 *
 *	Input:
 *		pD =	pointer to upper left corner of dest image
 *			(includes frame)
 *		wD =	width in pixels of dest (including frame)
 *		hD =	height of dest (including frame)
 *		dBpl =	destination image bytes/line
 *		pS =	pointer to upper left corner of source image
 *			(includes frame)
 *		sBpl =	source image bytes/line
 *		pScaleTable = ptr to table used to do scaling
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	not enough memory for the table
 *		ia_internal	internal error during table construction
 **************************************************************************/
Int32  CDECL 
i_scaleToGray4bpp2x (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    Int32	hD,
    Int32	dBpl,
    UInt8      *pS,
    Int32	sBpl,
    UInt32     *pScaleTable)
{
PROLOGUE;
 UInt32	 destroyTable = cFalse;	/* Boolean, cTrue when we should
				 * destroy the table before returning */
	/* compute the mask used to get rid of the bits on the right
	 * end of the line. */
UInt32	 rightMask = 0xFFFFFFFF << (4*((8 - (wD & 0x7)) & 0x7));
UInt32	 tableSizeInBytes;

	/* Unframe the image and pointers */
    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;
    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;

	/* Do we have a conversion table? */
    if (pScaleTable == NULL)
    {
	destroyTable = cTrue;
	if (i_getSizeOfScaleToGray4bpp2xTab (pIPState, &tableSizeInBytes)
		!= ia_successful)
	    return ia_internal;
	pScaleTable = (UInt32 *) MALLOC(tableSizeInBytes);
	if (pScaleTable == NULL)
	    return ia_nomem;
	if (i_fillScaleToGray4bpp2xTab (pIPState, pScaleTable, tableSizeInBytes)
		!= ia_successful)
	    return ia_internal;
    }

	/* loop through the rows of the images, calling ip_scaleToGray4bpp2xLine
	 * to reduce each pair of input lines to one output line.
	 */
    for (; hD > 0; hD--, pD += dBpl, pS += 2*sBpl)
	ip_scaleToGray4bpp2xLine (pIPState, pD, wD,
				  pS, sBpl, pScaleTable, rightMask);

    if (destroyTable)
	FREE(pScaleTable);

    NORMAL_RETURN;
}

/*

$Log: i_scgr42.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.8  1998/01/13 23:49:43  ddavies
i_scgr42.c -> /product/ipcore/ipshared/src/RCS/i_scgr42.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1996/03/26 22:43:04  ddavies
i_scgr42.c -> /product/ipcore/ipshared/src/RCS/i_scgr42.c,v

Change i_scaleToGray4bpp2x so that it allocates its table differently
if needed.  If the table pointers is NULL, it gets the size of the
table using i_getSizeOfScaleToGray4bpp2x, allocates it using MALLOC
(as before) and fills it using i_fillScaleToGray4bpp2xTab.  This
reflects the demise of i_makeScaleToGray4bpp2xTab.

Revision 1.6  1995/12/13 02:46:00  ejaquith
i_scgr42.c -> /product/ipcore/ipshared/src/RCS/i_scgr42.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1994/02/04  20:16:54  dferrell
 * i_scgr42.c -> /product/ipcore/ipshared/src/RCS/i_scgr42.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:23:50  ddavies
 * i_scgr42.c -> /product/ipcore/ipshared/src/RCS/i_scgr42.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:29:14  ddavies
 * i_scgr42.c -> /product/ipcore/ipshared/src/RCS/i_scgr42.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


