/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters	 *
 *  of copyrightable material and information now allowed by	 *
 *  statutory or judicial law or hereafter granted, including	 *
 *  without limitation, material generated from the software	 *
 *  programs which are displayed on the screen such as icons,	 *
 *  screen display looks, etc.					 *
 *****************************************************************/

/*
 * i_sharpen
 *
 *	Sharpen / Smooth
 *
 * Exports:
 *
 *	i_sharpen	-	Apply a sharpening filter to the image
 *	i_smooth	-	Apply a smoothing filter to the image
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"

/* Prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shros.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_sharp.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");

/*******************************************************************************
 * i_sharpen
 * 
 *	Apply a sharpening filter to the 8-bit grayscale image passed. The 
 *	sharpening parameter specifies how much to sharpen. The scheme used
 *	is to do a 3x3 filter using power-of-two fractions for the outer ring
 *	coefficients, which can be done with only shifts. The center
 *	pixel will usually need a shift and an add, e.g. if the outer pixels
 *	are 1/4, the sum is 2, so the inner pixel needs to be *3, = << 1 + self.
 *	There is a tradeoff between doing, say, all 8 outer pixels, or doing 
 *	just four, i.e. the two vertical and two horizontal. For the purposes
 *	of the "rush" this week, I'm going to implement the latter.
 *
 * Arguments:
 *	pImage =	Image to sharpen, pointer to upper left corner of frame
 *	sBpl =		Bytes per line of source image 
 *	w, h =		Size of image, including frame
 *	pResult = 	Place to put result.
 *	sharpness =	How much to sharpen, 0 - 3. A sharpness of 0 is not very
 *			much sharpening; a sharpness of 3 is more sharp.
 * 
 * Returns:
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_sharpen (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	sBpl,
    UInt8      *pResult,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    Int32	sharpness)
{
    PROLOGUE;
    UInt8      *pImg;
    UInt32 	outerSum, shiftOuter, right, center, left;
    Int32	pix;
    Int32 	i, j, shiftCenter;

    /* Move past the frame */
    pImg = pImage +  (sBpl << cLogFrameBits) + cFrameBits;
    pResult = pResult +  (dBpl << cLogFrameBits) + cFrameBits;
    h -= 2 * cFrameBits;
    w -= 2 * cFrameBits;

    /* Get a shift count */
    shiftOuter = 3 - sharpness;
    shiftCenter = sharpness;

    for (i = 0; i < h; i++)
    {
	right = 0; 
	center = 0;

	for (j = 0; j < w; j++)
	{
	    /* Move on to the next pixel */
	    left = center;
	    center = right;
	    right = UCHAR_ACCESS(pImg + j + 1);

	    /* Sum the outer ring */
	    outerSum = ((UInt32) UCHAR_ACCESS(pImg + j-sBpl-1)) +
		       ((UInt32) UCHAR_ACCESS(pImg + j+sBpl-1)) +
		       ((UInt32) UCHAR_ACCESS(pImg + j-sBpl  )) + 
		       ((UInt32) UCHAR_ACCESS(pImg + j+sBpl  )) +
		       ((UInt32) UCHAR_ACCESS(pImg + j-sBpl+1)) + 
		       ((UInt32) UCHAR_ACCESS(pImg + j+sBpl+1)) +
		       right + left;

	    /* Subtract it from the center */
	    pix = ((center << shiftCenter) + center) - (outerSum >> shiftOuter);

	    /* Clamp it and store */
	    UCHAR_ACCESS(pResult + j) = (UInt8)((pix < 0) ? 0 : (UInt8)((pix > 0xff ? 0xff : pix)));
	}

	IP_YIELD (sharp_finish);

	pImg += sBpl;
	pResult += dBpl;
    }

sharp_finish:
    NORMAL_RETURN;
}
/*******************************************************************************
 * i_smooth
 * 
 *	Apply a smoothing filter to the 8-bit grayscale image passed. The 
 *	smoothing parameter specifies how much to smooth. The scheme used
 *	is to do a 3x3 filter using power-of-two fractions for the outer ring
 *	coefficients, which can be done with only shifts.
 *
 * Arguments:
 *	pImage =	Image to smooth, pointer to upper left corner of frame
 *	sBpl =		Bytes per line of source image 
 *	w, h =		Size of image, including frame
 *	pResult = 	Place to put result.
 *	smoothness =	How much to smooth, 0 - 2. A smoothness of 0 is not very
 *			much smoothing; a smoothness of 2 is more blurry
 * 
 * Returns:
 *	ia_successful, ...
 *******************************************************************************/

Int32 CDECL 
i_smooth (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	sBpl,
    UInt8      *pResult,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    Int32	smoothness)
{
    PROLOGUE;
    UInt8      *pImg;
    UInt32	outerSum, right, center, left;
    Int32	pix;
    Int32	i, j, shiftResult;
    static Int32 mults[] = { 0, 8, 24, 56 },
		 shifts[] = { 3, 4, 5, 6 };
    Int32       centerLut[256]; 	/* To make code reentrant, this
					 * can't be declared statically! */

    /* Move past the frame */
    pImg = pImage +  (sBpl << cLogFrameBits) + cFrameBits;
    pResult = pResult +  (dBpl << cLogFrameBits) + cFrameBits;
    h -= 2 * cFrameBits;
    w -= 2 * cFrameBits;

    /* Make a multiplication table. The idea is to make a fraction which,
       when added to 8/1 (the outer ring) adds up to some power of two,
       so we can shift down to get our result */
    smoothness = 3 - smoothness;
   
    for (i = 0; i < 256; i++)
    {
	centerLut[i] = mults[smoothness] * i;
    }

    shiftResult = shifts[smoothness];

    for (i = 0; i < h; i++)
    {
	right = ((UInt32) UCHAR_ACCESS(pImg)); 
	center = 0;
	    /* Initialize outerSum as the sum of the pixels in
	     * the first column.
	     */
	outerSum = ((UInt32) UCHAR_ACCESS(pImg + sBpl)) +
		   ((UInt32) UCHAR_ACCESS(pImg + -sBpl));

	for (j = 0; j < w; j++)
	{
	    /* Move on to the next pixel */
	    left = center;
	    center = right;
	    right = UCHAR_ACCESS(pImg + j + 1);

	    /* Add in the new pixels to the right of our
	     * current position.
	     */
	    outerSum += ((UInt32) UCHAR_ACCESS(pImg + j-sBpl+1)) + 
		        ((UInt32) UCHAR_ACCESS(pImg + j+sBpl+1));

	    pix = (centerLut[center] + outerSum + left + right) >> shiftResult;

		/* subtract off the pixels in the left column. */
	    outerSum -= ((UInt32) UCHAR_ACCESS(pImg + j-sBpl-1)) +
		        ((UInt32) UCHAR_ACCESS(pImg + j+sBpl-1));


#if 0
	    /* Clamp it and store */
	    UCHAR_ACCESS(pResult + j) = (UInt8)((pix < 0) ? 0 : (UInt8)((pix > 0xff ? 0xff : pix)));
#else
	    UCHAR_ACCESS(pResult + j) = (UInt8) pix;
#endif
	}

	IP_YIELD(smooth_finish);
	pImg += sBpl;
	pResult += dBpl;
    }

smooth_finish:
    NORMAL_RETURN;
}

/*

$Log: i_sharp.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.12  1998/01/13 23:49:49  ddavies
i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.11  1997/07/11 01:49:09  ddavies
i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.10  1995/12/13 02:46:10  ejaquith
i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.9  1995/10/13  23:51:34  dferrell
 * i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.8  1994/10/11  16:37:58  lperry
 * i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v
 *
 * Added include of shros.pub to get IP_YIELD def
 *
 * Revision 1.7  1994/07/13  16:42:12  ejaquith
 * ejaquith on Wed Jul 13 09:39:59 PDT 1994
 *
 * i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v
 *
 * Made centerLut variable non-register since need address.
 *
 * Revision 1.6  1994/07/13  00:00:43  ejaquith
 * ejaquith on Tue Jul 12 16:59:40 PDT 1994
 *
 * i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v
 *
 * Removed #include "iadef.prv"
 * Changed statically declared variable centerLut to register variable
 *
 * Revision 1.5  1994/06/23  05:59:52  macleod
 * i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v
 *
 * Added some IP_YIELD calls
 *
 * Revision 1.4  1994/02/04  20:17:09  dferrell
 * i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:24:03  ddavies
 * i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:29:39  ddavies
 * i_sharp.c -> /product/ipcore/ipshared/src/RCS/i_sharp.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

