/*************************************************************
 *  Copyright (c) 1991, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  i_skwflt.c:  contains floating point wrappers for skew detection.
 *
 *          Contains subroutines
 *
 *                Int32         i_detectSkew()
 *
 *    
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"

    /*  prototypes */
#include "shrskew.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_skwflt.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");
/* For a description of the skew detection algorithms, see i_skwfxp.c or
 * miskwfxp.c.  This module, being only a wrapper is pretty simple.  It
 * translates the skewAngle returned by i_detectSkewFixedPoint from our
 * fixed point format to ordinary floating point radians.
 */ 


/* This is a publically defined routine */
/* PUBLIC */
/****************************************************************************
 *  i_detectSkew():
 *        Uses i_detectSkewFixedPoint to detect the skew.  This routine
 *		converts the results to radians in floating point.
 *        Allowable reduction factors are 1, 2, 4, 8, 16, and 32.
 *
 *        Returns the skew angle in radians.  On error, returns
 *		 angle of 0 and confidence of 0.
 *
 *        Typical redFactors for 8.5x11 image at 200 to 300 pixels/inch are:
 *                4 to 8.
 *
 *	pImage is pointer to the image (including frame)
 *	imWidth is width of image in pixels including frame
 *	imHeight is height of image including frame
 *	imBpl is bytes/line of image (including frame)
 *	redFactor is the subsampling factor.  The higher it is, the
 *	faster the routine runs but the lower the accuracy.
 *	skewAngle is a pointer to the angle to be returned by this
 *	function.  The angle is in radians.
 *	confidence is a pointer to value returned that indicates the
 *		confidence the application should have in the answer.
 *		The larger the confidence value, the better.
 *	xRes is the resolution in the x direction.
 *	yRes is the resolution in the y direction.  The units of xRes
 *		and yRes are irrevelant so long as both xRes and
 *		yRes have the same units.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem 	could not allocate memory
 *		ia_noLines	could not find enough horizontal edges
 *		ia_invalidParm	xRes != yRes and either xRes == 0 or
 *				yRes == 0.
 ****************************************************************************/
Int32 CDECL 
i_detectSkew (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	imWidth,
    Int32	imHeight,
    Int32	imBpl,
    Int32	redFactor,
    Float64    *skewAngle,
    UInt32     *confidence,
    Int32	xRes,
    Int32	yRes)
{
    PROLOGUE;
    
    ScaledAngle	fixedPtSkewAngle;

	/* Have i_detectSkewFixedPoint do the real work */
    __error__ = i_detectSkewFixedPoint (pIPState, pImage,
				      imWidth,
				      imHeight,
				      imBpl,
				      redFactor,
				      &fixedPtSkewAngle,
				      confidence,
				      xRes, yRes);

	/* convert the angle and return it */
    if (__error__ == ia_successful)
	*skewAngle = ScaledAngle_TO_RADIAN64(fixedPtSkewAngle);
    else
    {
	*confidence = 0;
	*skewAngle = 0.0;	/* for compatibility */
    }

    NORMAL_RETURN;
}

/*

$Log: i_skwflt.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.8  1998/01/13 23:49:51  ddavies
i_skwflt.c -> /product/ipcore/ipshared/src/RCS/i_skwflt.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1997/06/25 03:43:57  ddavies
ddavies on Tue Jun 24 20:42:38 PDT 1997

i_skwflt.c -> /product/ipcore/ipshared/src/RCS/i_skwflt.c,v

At the risk of creating a self referential checkin message, fixed the spelling
of the following word: irrevelant.

Revision 1.6  1996/05/22 01:16:25  ddavies
ddavies on Tue May 21 18:15:58 PDT 1996

i_skwflt.c -> /product/ipcore/ipshared/src/RCS/i_skwflt.c,v

i_detectSkew no longer returns beastAngle. It returns an angle of 0 with
a confidence of 0 when it couldn't find an angle.

Revision 1.5  1995/12/13 02:46:13  ejaquith
i_skwflt.c -> /product/ipcore/ipshared/src/RCS/i_skwflt.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1995/12/13  01:43:11  ddavies
 * i_skwflt.c -> /product/ipcore/ipshared/src/RCS/i_skwflt.c,v
 *
 * Pass xResolution and yResolution down to the real skew detection routines
 * so they can remove the native resolution-dependence of angles calculated
 * there.
 *
 * Revision 1.3  1993/12/02  19:24:07  ddavies
 * i_skwflt.c -> /product/ipcore/ipshared/src/RCS/i_skwflt.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:29:47  ddavies
 * i_skwflt.c -> /product/ipcore/ipshared/src/RCS/i_skwflt.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


