/*************************************************************
 *  Copyright (c) 1991, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  i_skwfxp.c:  contains common wrappers for skew detection.  Uses
 *		 only fixed point calculations.
 *
 *          Contains subroutines
 *
 *                Int32         i_detectSkewFixedPoint()
 *                Int32		ip_detectSkewFine()
 *                Int32	      	ip_detectSkewRobot()
 *
 *    
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#include "fxpmath.pub"

    /*  prototypes */
#include "shrskew.pub"
#include "shrskew.prv"
#include "shros.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_skwfxp.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");

/* This version of skew code is a variant of the original skew.c.  The
 * job of skew.c is to detect and correct image skew.  It depends on
 * images having a lot of horizontal stuff (lines, text).
 *
 * The original skew.c accepted binary images as input and produced them
 * as output.  We would like to improve the speed and quality of skew
 * detection.  We would also like to explore grey scale rotation for skew
 * correction.  We'll work on skew detection first.
 *
 * Skew.c`s detection algorithm had three major steps:
 *	1. Reduce image to a tractable size.
 *	2. Coarsely rotate the reduced image over a range of angles (say
 *	   -10 degrees to +10 degrees in 1 degree increments) to try to
 *	   find the position in which the skew is minimized.
 *	3. Use a bifurcation algorithm around the best angle found in the
 *	   sweep to fine tune the results.  Stop looking when the amount by
 *	   which the image will be rotated for the next step reaches a
 *	   value passed in as a parameter.
 *
 * The method used to obtain a "skew score" for a particular angle is to
 * vertically shear the reduced image by the angle and count the number of
 * pixels in each horizontal line.  The score equals the sum of the squares
 * of the differences in pixel counts across all adjacent lines.  In effect,
 * we're looking for skew angles at which we find the steepest horizontal
 * edges.  The squaring operation sharpens our sensitivity to steep edges.
 *
 * We'd like to explore improving this algorithm on two fronts.  First,
 * information is lost when the image is reduced.  This degrades our ability
 * to get good answers even with an ideal detection algorithm.  We'd like
 * to avoid reducing the image.
 *
 * Second, the process of continually shearing the image is relatively slow,
 * especially for a full scale image.  We'd like to avoid shearing.
 *
 * Why is the image sheared?  Because it's easier to count pixels
 * horizontally than diagonally.  However, if we could count diagonally,
 * we could dispense with the shearing, hence with the reduction.  We will
 * do so.
 *
 * We also note that it's not necessary to count every pixel in the image for
 * every potential shear angle.  Skew detection is a matter of statistics.
 * The more image samples we make, the better the statistics.  However, there
 * is probably some point of diminishing returns, but we don't know where
 * that is now.  We'll start by sampling roughly every other long word
 * horizontally and a pair of lines out of every 2*reductionFactor lines
 * (see below) vertically.  For a full scale 2300 x 3300 image, that's
 * about 32 horizontal samples and (reduction factor of 8), 400 samples
 * vertically, for a total of 12,800 samples.  One of the interesting jobs
 * will be to see how much this can be reduced while retaining sufficient
 * accuracy.
 *
 * Not only is it not necessary to count every pixel in a line, it's really
 * not necessary to count every pixel in a word.  If we think of looking for
 * horizontal rivers of white space, we can get by with saying the value of
 * a particular 32 bit sample is "1" if there are any bits on and 0 otherwise.
 * This improves counting speed.
 *
 * The original skew.c had three levels of code hierarchy.  The top level,
 * deskew(), used detectSkew to find the skew angle and rotateArb to
 * correct the image.  detectSkew used detectSkewSweep and detectSkewBif to
 * do the sweep and bifurcation steps respectively.  Each of those reduced
 * the image (!!) and used shearC and findDiffPowerSum to compute the scores.
 *
 * The sweep stage is used because the bifurcate step can otherwise choose a
 * local, but not global, maximum in skew score.  The number of false peaks
 * and amount of general noise of a hypothethical skew score vs angle curve
 * appears to be strong function of reduction factor.  This is because when
 * the program effectively looks at every 4th or 8th line, it's resolution
 * for edge discrimination is reduced.  Naturally, you say.  Edges can only
 * be determined to within 4 or 8 lines.  However, in addition to this
 * effect, the algorithm starts to lose its ability to distinguish between
 * one long edge and two separate horizontal edges which are slightly
 * displaced relative to each other.  For example, take a page with three
 * columns of text.  Say the inter-line spacing in all columns is identical,
 * but the columns are verticaly shifted relative to each other.  This is
 * reasonably common.
 *
 * Say the skew detection algorithm samples along a slanted line that goes
 * through all the black pixels of every column even though it is misaligned
 * with each column individually.  If the next sample line is far enough down
 * (due to reduction), it can go through all the white pixels of every column
 * even though it is also misaligned with each column.  The difference in
 * pixel counts between these two lines will be high, giving a misleading
 * high score.  I have an image where this happens - the "deskewed" image is
 * consistently skewed by about 0.5 degrees.
 *
 * One way to fix this is to sample every line of the image.  That's too
 * expensive.  A cheaper way is to sample adjacent pairs of lines of the
 * original image.  That is, instead of sampling every 4th or 8th line, as
 * is effectively done by reduction, sample 2 lines, then skip 6 or 14.
 * The skew score is calculated only between these pairs.  This is what we'll
 * do.
 *
 * Sampling adjacent pairs of lines has the effect of giving the skew score
 * curve a VERY thin, spikey peak.  This gives wonderful final resolution if
 * the algorithm can find it.  If the curve has no slope except in a very
 * thin region around the peak, the peak becomes a needle in the haystack.
 *
 * The spike will be widened by the fact that text lines, a prominent
 * horizontal feature do not present perfect horizontal surfaces.  They're
 * a bit noisy, so will contribute non-zero skew scores when they are close
 * to aligned.
 *
 * Unfortunately, even with this wonderful technique, we cannot eliminate
 * false peaks in the curve.  Even if the skew evaluation were perfect, pages
 * with slanted areas (almost any page with a picture on it) will have
 * blips in their skew curves.  These blips can potentially fool a pure
 * bifurcation algorithm.
 * 
 * The brute force way to identify the real peak is to sweep across the entire
 * band of angles.  The sweep can be a bit coarse, using bifurcation to home
 * in on the exact value.  But it's boring to be brutes.  Slow too.
 *
 * Another way to use two methods of generating skew scores.  One isn't very
 * exact, but produces a wide curve with enough amplitude to swamp out most
 * of the noise.  The other produces a thin, spikey curve for good resolution.
 * The second method is the one described above.  What function would produce
 * a big, wide curve?  Simply performing the previous algorithm with fewer
 * samples flattens the peak of the curve, but also decreases the signal to
 * noise ratio.
 *
 * A property of text lines is that they are very long, thin, horizontal bars.
 * We could construct a set of slanted paths across the image, counting
 * ON pixels encountered by these paths.  Higher counts would tend to
 * indicate that the paths are stay inside text lines.  However, this measure
 * is sensitve to both local orientation of text lines (which is good) and
 * global placement of lines and graphics across the page (which is bad).
 * A line's score can be high simply because it goes through many
 * independent text lines or through two or three text lines in different
 * columns that are displaced relative to each other.
 *
 * Instead, we sample pixels along linear paths in the image, remembering for
 * each path the *maximum number of consecutive pixels* encountered.  A path
 * that stays inside one text line will get a high score.  A path that goes
 * through a large number of text lines will get a low score since it will not
 * see a large section of any one of the lines.
 *
 * This scoring scheme doesn't give very good resolution since it can't
 * distinguish any two paths that follow a text line across the page.  Its
 * minimum resolution is the height of a text line.  However, it does produce
 * a nice broad curve that's reasonably insensitive to noise.  As the angle
 * of the paths relative to the text lines increases, the skew scores fall off
 * almost linearly.
 *
 * We could use bifurcation with the MaxPixels curve to find the skew angle
 * to within a degree, then use the HorizontalEdge method to fine tune. 
 * This is probably what we'll end up doing.  However, just to be adventurous,
 * we'll try another scheme first.  We assume the broad curve is pretty
 * linear.  We find a couple of points on each slope (four samples in all),
 * compute the angle at which the slopes intersect and use the bifurcation
 * method starting from there.  If that doesn't give a sufficiently accurate
 * first guess, we'll go back to bifurcation with the wide curve.
 *
 * How do we do bifurcation?  The original skew.c used 5 points, choosing one
 * of the interior three as the new center, then halving the angular distance
 * between them.  The number of samples required for this method is
 * 5 + 2 * number of steps.  Initially, we'll try a method that relies on the
 * the curves being fairly smooth.  We pick three inital points separated by
 * the maximum initial separation.  While the middle sample is not the highest,
 * take one more sample at the maximum separation on the other side of the
 * highest endpoint.  Once the middle is the highest, take a sample half way
 * between the middle and the next highest endpoint.  If everything were
 * smooth and symmetric, we'd know that the highest point was somewhere
 * in the range between the two highest samples, so we could just keep taking
 * a new sample between the two highest values and find it.  This gives a
 * total number of samples of 3 + number of samples.  If the curves for real
 * images are too noisy, we'll just repeat the first part of the algorithm -
 * taking samples on the far side of the highest endpoint until the middle
 * sample is the highest, then halving the separation and taking a sample
 * between the two highest samples.
 *
 * Efficiently tracing diagonal paths is more complex than I originally
 * thought.  Moving horizontally is easy, figuring out when to move
 * vertically and by how much is tricky.  To get sufficient accuracy,
 * we need to move vertically by fractional pixels.  Well, at least we need
 * to approximate fractional pixel movement.  This can be done pretty well
 * by moving up or down in units of whole lines at selected intervals.
 *
 * The job of plotting a diagonal path through the image is reduced then to
 * determining the size and frequency of vertical movement as a function of
 * angle.  One way to approach the problem is to pick an angle,
 * compute its tangent, multiply that by the width of the image (giving the
 * total rise at the end of the path) and divide the result by the number of
 * horizontal samples to be taken.  Yikes!  A trig function, a multiply and
 * a divide for every sample line!  Simple for scientists, but scary
 * for engineers.
 *
 * Instead of using angles as the space through which we search or sweep,
 * say we use vertical path displacement.  For small angles, the vertical
 * displacement is very nearly linear with angle anyway.  We can compute the
 * maximum positive and negative vertical displacments of the initial
 * search space as well as the vertical displacement associated with the
 * initial deltaAngle.  We then search or sweep through these displacements
 * skipping the number of lines associated with deltaAngle at each step.
 * To do bifurcation, we can just divide the incremental number of displacment
 * lines in half.  All this eliminates the trig function and the multiply
 * in the original algorithm.
 *
 * This is still a little messy because the width of the image still figures
 * into the equations.  We note that the angle of a path is (obviously)
 * independent of the width of the image.  In fact, we could use a hypothetical
 * image of any convenient width to compute vertical displacements so long as
 * the code that's traversing the real image checks to see when it's reached
 * an edge.  Convenient image widths are powers of 2 so divides and multiplies
 * turn into shifts.  We will follow this approach - a hypothetical image
 * used to compute vertical displacements that are used in turn to traverse
 * the real image.
 *
 * We discussed simulating fractional vertical line movement by occasional
 * full line movement.  For example, if the path rises 10 lines in
 * 20 samples, we move up one line every other sample.  Rather than have the
 * sampling code compute some marvelous error function, we'll just define
 * when vertical movements should be taken in a set of bit vectors.  We will
 * not produce a set of bit vectors for every possible image width.  Instead,
 * we will produce sets of fairly short bit vectors (16 bits) that can be
 * reused if the code takes more than 16 horizontal samples in a line.
 *
 * Nice theory, but what really happens?  Here are the steps in calculating
 * the necessary vertical displacement and choosing a bit vector corresponding
 * to the angle of the path.
 * 	1. Decide on the spacing between horizontal samples.  One algorithm
 *	   might want a 4 byte sample every 4 bytes.  Another might take a
 *	   4 byte sample every 12 or 16 bytes.
 *	2. Since our traversal algorithm will be periodic every 16 samples,
 *	   multiply the sample distance chosen in step 1 by 16 (shift left 4).
 *	   This gives the width of the hypothetical image.
 *	3. To determine the vertical offset in the hypothetical image for
 *	   maxAngle, multiply tan(maxAngle) by the width of the hypothetical
 *	   image computed in step 2.  Repeat for minAngle and delAngle.
 *	4. Sweep or search through the space of vertical offsets between
 *	   maxAngle's offset and minAngle's offset using an (initial)
 *	   increment of delAngle's offset.
 *	5. The intra-sample vertical offset for a particular path comes in
 * 	   two parts, integer and fraction.  The integer part is added to
 *	   the address on every sample.  This is simply equal to the
 *	   hypothetical vertical offset (the things we're searching through)
 *	   divided by 16 (shifted right 4).  The fractional part equals the
 *	   hypothetical vertical offset MOD 16 (AND with 0xF).  This is used
 *	   to choose a bit vector.  The bit vector is shifted left one place
 *	   after taking each sample.  If the bit that fell off the end was a
 *	   1, the bytes/line is added to (subtracted from) the next sample
 *	   address.
 *
 ***************************************************************************
 *
 * That worked, but we can do better in the coarse search.  Cycling through
 * (and reloading) the bit vectors is harder than we'd like.  We still end
 * up computing too many trig functions, doing too many multiplies and
 * divides, etc.  There are also too many cache refills.
 *
 * It turns out that a rise of 0.5 bits across 32 bits is 0.9 degrees.  That's
 * pretty close to 1 degree, which is the traditional value of deltaAngle
 * for coarse skew detection.
 * Anyway, it's good enough for a coarse search.  We'll divide the range from
 * minAngle to maxAngle into vertical displacements that change in units of
 * 8 bits.
 *
 * The assembly code that actually computes scores does so in a slightly
 * unrolled loop.  On every other read, it adds rSrcPitch to the source
 * address.  The code passes 0, +sourcePitch or -sourcePitch in this argument
 * depending on whether the current vertical displacment/32 horiz bits
 * is an integer (0), a positive non-integer (+sourcePitch) or a
 * negative non-integer (-sourcePitch).  Things become simpler, hence faster.
 *
 * We can improve the cache hit ratio on sweep detection algorithms by
 * dividing the image into blocks and doing the entire sweep on each block,
 * then moving on to the next block.  After a few passes, most of the bits
 * in a block will already be in the cache, so the remaining passes will go
 * faster.  However, if the blocks aren't wide enough, we end up adding lots
 * of little maximums to the skew score instead of discarding them in favor
 * of fewer, bigger maximums.  This flattens the score curves.
 *
 * Since the slant calculation for the coarse algorithm is independent of
 * the width of the block or image, we might as well make blocks cover the
 * entire width of the image.  This will cause more cache misses than we'd
 * like, but will improve the accuracy of the result.
 *
 ***************************************************************************
 *
 * That was successful and still pretty fast.  Now, we change from a sweep
 * to a binary search.  In the search algorithm, we sample both extreme grid
 * points and the middle.  We then continue to sample half way between the
 * two highest grid points until we reach a sample interval of 1 degree.
 *
 * Well, almost.  Say the image is almost centered, the skew angle is just
 * a bit positive (for example).  Noise in the skew function can cause
 * minAngle's score to be a little bigger than maxAngle's score.  This has us
 * eliminating all the positive angles in the first step, a mistake.  Instead
 * we don't adjust the trial angle quite so far in the first couple of steps.
 * For example, instead of curMaxAngle = (curMaxAngle+curMinAngle)/2, we say
 * curMaxAngle = (curMaxAngle+curMinAngle)/2 + (curMaxAngle-curMinAngle)/8.
 * The second term shortens the first couple of steps.
 *
 ***************************************************************************
 *
 * The fine skew detection algorithm will also use a binary search.  Instead
 * of using angles, we deal in terms of vertical displacment of slanted lines
 * over a hypothetical 4k bit horizontal span.  Yeah, but real images aren't
 * 4K bits wide, right?  This just means that if the slanted line would have
 * risen n bits over a 4Kb stride, it rises n*(imageWidth/4K) bits over a
 * imageWidth stride.
 *
 * Since everything is proportional, why not choose a smaller hypothetical
 * stride?  If we choose a stride smaller than the width of a real image, we
 * get insufficient angular resolution.  In an extreme case, say the
 * hypothetical image were 32 bits wide and the smallest angle we could
 * handle had 1 bit rise in 32 bit stride.  This corresponds to a rise of
 * 80 lines in a 2560 bit real image.  We'd like a resolution of +- 1 line
 * in 2560 bits, so need to choose a hypothetical image wider than the real
 * one.  The next larger power of 2 is 4K.
 *
 * The fine skew detection algorithms is based on adding up differences
 * vertically adjacant samples.  It's not necessary to sample every point
 * along each grid line, so we'll spread the samples out a bit.  Initially,
 * we'll try sampling 32 bits (8 bits?) out of every 128.  That would give
 * 4K/128 = 32 samples in the hypothetical image.  Real images are less wide,
 * so we have to stop when we reach the edge.
 *
 * Since we'll need good accuracy, we can't get away with the 1 degree trick
 * used by the coarse routine.  In fact, we need to be able to rise or fall
 * 0 to 31 lines smoothly over the course of 32 samples.  There is a
 * cute way to do this using one counter setting the interval at which
 * the pitch is added to the address and another counter setting the interval
 * for subtracting the pitch (error correction).  However, since we know
 * there will never be more than 32 samples, we'll just use a bit vector
 * to decide when to add the (positive or negative) pitch to the current
 * address when moving "horizontally".
 *
 *********************
 * The original version of this just got noise, no peak to the curve.  More
 * samples are probably needed.  But that brings up the question, "Why are
 * so many samples needed?"  Because very few of them actually find edges
 * to detect.  We need to take lots of samples so that some interesting
 * number actually find edges to measure.
 *
 * That sounds awfully inefficient.  What if we looked for edges instead of
 * simply hoping that they'd show up sometimes?  This is the idea behind
 * ip_detectSkewRobot.
 *
 * Say we view an image as a maze of text and data also having white and
 * black deserts.  The text and data (especially text) have nice horizontal
 * edges we'd like to find and whose angle we'd like to record.  The deserts
 * are big vacant areas.
 *
 * We'll pick a bunch of points (not at random, that takes too long).  An easy
 * set of points forms a big X across the image.  At each image point, we put
 * down a little "robot."  If the little fellow lands in a white space
 * (all 32 bits are white), it looks vertically down for some black.  It
 * samples every 16th line.  If no black is found in a few samples, the
 * robot concludes it's in a desert and gives up (returns).
 *
 * If the robot finds some black, or lands in black to begin with, it
 * looks below for white.  It samples every 8th line (white space being
 * generally thin between text lines).  If no white is found in a few samples
 * the robot concludes it's in a black desert and gives up.
 *
 * If white space is found, we have a black/white edge.  These tend to be
 * the best kind since text is usually darker than the background (not
 * always, but tough) and the underside of a text line isn't as jagged as
 * the top side.  We do a short binary search until we have found the edge
 * exactly.
 *
 * We now attempt to follow the edge to the right.  Starting on the white
 * line under the black line, we sample the next word to the right.  If it's
 * also white, we look up to see if the black stuff is still there.  If
 * so, we record that one horizontal step was taken and no vertical steps
 * were taken.  If not, we keep looking up for 5 lines.  If the black stuff
 * is found, we record one horizontal step and as many vertical steps
 * (a negative number) as were required to find the black stuff.  We also
 * move up to the line immediately below the black stuff.  If no black
 * stuff was found, we conclude that the edge has ended (or was too steep
 * to consider) and quit.
 *
 * If the step to the right found black stuff, we look down for up to 5 lines.
 * If white stuff is found again, we record one horizontal step and as many
 * the number of lines (this time a positive number) as required to find white.
 * We also move down.  If no white is found, we just quit.
 *
 * The previous two steps are repeated so long as the edge is seen.
 *
 * The purpose of all this is to record the slope of the edges.  We pick some
 * number of total horizontal steps to record and keep putting down robots
 * until that number is found.  We then get the slope by dividing the total
 * number of vertical steps by the total number of horizontal steps.
 *
 * Clearly, we need to traverse the image several times to get rid of noise.
 * The number of horizontal steps required will be determined by
 * experimentation.
 *
 * A more complex version of this code could maintain several vertical
 * counters for different slopes.  We'd then pick the biggest one as the
 * dominant slope.  As it is now, slanted lines in the image contribute to
 * the average slope calculation.
 *
 **************************************************************
 *
 * We had some problems.  First, there weren`t enough samples.  We now
 * take sample points in columns.  The sample points within a column are
 * close enough to catch the majority of text lines.
 *
 * Second, the samples were so wide that we could bridge across slanted
 * text rows.  The sample width is now 16 bits instead of 32.
 *
 *************************************************************
 *
 * All this worked well on most documents.  Z Smith noticed that it got
 * confused with normal subsampling factors in line drawings.  We think
 * this is because the lines are so thin.  To get some indication about
 * when the answer is confused, we now return a ratio of the maximum
 * power found to the minimum power found.  Applications can use this
 * heuristically to decide when to change the subsample factor.  A pointer
 * to the ratio, called "confidence" is passed to i_detectSkewFixedPoint
 * and then to ip_detectSkewFine.  ip_detectSkewFine fills in the value
 * pointed to when it returns.
 *
 *************************************************************
 *
 * That was nice, but wasn't a very nice measure of confidence.  We'd like
 * something that depends only only on the relative height of the peak
 * (the previous measure depended only on the difference between the
 * highest and lowest points so could be confused by noise) but also on
 * the peak's width.
 *
 * Say the total power over all samples is sumOfPowers.  The sum of the
 * absolute values of angular distances from the final skew angle to the
 * sample points is sumOfAngles.  The new confidence measure returned by
 * ip_detectSkewFine is
 *	confidence = ( Sum( ABS(finalAngle - angle) * power(angle) ) /
 *			  (sumOfPowers) )
 * That is, we weight each other angle searched by the fraction of the total
 * power assigned to that angle.  This weighted sum is the average angular
 * distance from the result.
 *
 * i_detectSkewFixedPoint computes the fraction of this number compared
 * to 1 degree.  If the number is greater than 1 degree, the returned 
 * confidence is 0.
 * If the average offset is 0, the returned number is 100.  From a few
 * trials on images, confidences of more than 65 seem to indicate
 * reasonable certainty.  The caller may also wish to call this
 * routine with a smaller subsampling factor and compare answers.  If
 * the confidence stays sort of low (50's, lower 60's) and the answers
 * vary by more than 0.2 degrees from one subsampling interval to the
 * next, we're probably confused.
 *
 *************************************************************
 *
 * That was better, but as Dan Bloomberg noted, rotating an image in which
 * the confidence was ok (but not great) by 90 degrees and computing
 * the skew angle again gave about the same confidence.  Basically, it
 * does fine on images that have lots of long horizontal edges and not
 * much else.
 *
 * We'll try looking out +- 2 degrees and +- 3 degrees from the angle found
 * The powers we get from these will be used in the confidence study.  The
 * idea is that a lowish, noise peak should be emphasized if we look farther
 * away if it's good and be shown as noise if it's bad.
 *
 * Post change: adding this change and fixing a bug further increased the
 * utility of the confidence measure.  However, if you run the skew detect
 * code on normal magazine images that have been rotated 90 degrees, you
 * still get a pretty high confidence.  Why?  Well, there are lots of
 * vertical strokes in text.  These become horizontal strokes.  Also, the
 * page often contains vertical rules that become horizontal rules.  Finally,
 * text is usually in columns with justified edges.  These edges are also
 * found.  However, I did find that the presence of a slightly skewed line
 * or two can result in a skew measure that's almost a degree off, yet still
 * reports high confidence.  I recommend that users use the landscape/portrait
 * heuristic to orient the image, then run this code to find the really
 * correct angle.
 *
 * In this iteration, we also started having ip_detectSkewFine look +- 2
 * degrees around the offset returned by ip_detectSkewRobot.  This helps
 * for fairly large skews (6-8 degrees).
 *
 * IMPROVEMENT 4/19/94: For a very few images (one in many, many attempts),
 * ip_detectSkewFine jumps over the peak in the skew curve.  In an
 * attempt to get around this, we'll make the amount by which the trial
 * angle changes a function of the reduction factor:
 *	reduction factor	amount of change
 *		1		(max - min)/16
 *		2		(max - min)/8
 *		4		(max - min)/4
 *		> 4		(max - min)/4
 *
 * ANOTHER IMPROVEMENT, 12/14/94:  After many more images, we found a form
 * in which the fine skew detector jumps right over a very sharp peak that's
 * exactly at the place detected by the robot detector.  To get around this
 * without messing up the confidence measures too much, we simply don't let
 * either us skip over this point.  That's seems pretty brute force, but
 * it does have the nice property that it doesn't do much for almost all
 * images and does prevent us from making the occasional big mistake.
 *
 * BUG FIX: 2/21/95: Ron Danisewicv found an image in which the scores
 * computed by ip_findDiffPixelsSum overflowed.  In his case, it made a
 * big positive number into a big negative number.  There are several
 * possible solutions.  The most sure one seems to be to turn the number
 * into a double integer.  Since we can't detect the carry out of an
 * add, we use the sign bit of the lower 32-bit integer as the carry flag.
 * When it's set, we clear it and increment the upper 32-bit integer.
 * This also means we have to do two word comparisions, but tough.  This
 * seems like a better solution than always scaling the results down
 * (almost all images work fine now, we don't want to mess that up) or
 * changing the method of calculating scores when we find an extra-big
 * one.  We can't just change the method of calculating big scores, we
 * have to go back and redo all previously calculated scores too.  Yuck.
 * Besides, we have no way of absolutely guaranteeing that that would
 * fix the problem anyway.  So, just use more bits to hold the score.
 *
 * And that's not all.  As we go along a diagonal path, four difference
 * sums are calculated in each 32-bit word.  The four sums are added
 * together, producing the final difference sum for the line.  This
 * sum is shifted left by a distance specified by (sum/16).  There are
 * lots of overflow possibilities here.
 *
 * To restrain the overflow, we increase horizontal distance between
 * samples so that the total number of samples in a horizontal pass
 * is less than 128.  This means the sum of two samples is less than
 * 256 (so we can just add adjacent counts without masking anything)
 * and the sum of all four counts is less than 512.  511/16 = 31, so
 * the shift distance will be less than 32.  This means that any shifted
 * sum will fit into a two-word total sum.
 *
 *************************** ANISOTROPIC IMAGES *************************
 *
 * In our original encounter with anisotropic images, we attempted to
 * handle all changes in the rotation code.  This turned out to be a problem.
 * Say we ignore x resolution and y resolution here.  The skew angle we
 * find will be (skew angle of isotropic image) * (yRes/xRes).  That's
 * because the vertical dimensions will have been flattened by (yRes/xRes)
 * and we calculate the skew angle by looking at arctan((vertDisplacement)/
 * (horizDisplacement)).
 *
 * Knowing that this happened, the small angle rotation code essentially
 * multiplied its incoming angle by xRes/yRes before rotating.  This was
 * fine so long as the angle came from skew detection code, but incorrect
 * if the angle came from somewhere else.
 *
 * Our fix is to multiply the vertical displacement by xRes/yRes explicitly
 * before computing the arctan.  This returns a "resolution independent"
 * angle instead of one that depends on the ratio of the x and y resolutions.
 * Similarly, the small angle rotation code will assume the given angle
 * is always resolution independent and will need to compensate only
 * for the resolution of the image its given.
 */ 

/* default: suppress all debugging */
#define  DEBUG_PRINT_SKEW     0
#define  DEBUG_PLOT_SKEW      0

/* Enable debugging info */
#ifdef  ALP_DEBUG_ON
#define  DEBUG_PRINT_SKEW     1
#define  DEBUG_PLOT_SKEW      1
#endif  /* ALP_DEBUG_ON */

#define  cBadOffset	    999	    /* error condition */
#define  cMinSkewOffset     -16    /* lines; used in bifurcation */
#define  cMaxSkewOffset      16    /* lines; used in bifurcation */
#define	 cReasonableDivergence 16   /* about 1 degree */


#define	 cNumTrialLines     10	    /* number of "full width lines" the
				       edge follower robot must find to
				       get an adequate estimate of
				       the slope */


/* Bit vectors used for "fractional" vertical line displacements as
 * sampling proceeds across a line.  Ones indicate when we should move
 * one more vertical line.
 */
static UInt32 frac32LineVectors[32] = {
				0x00000000,	/* frac = 0/32 */
				0x00010000,	/* frac = 1/32 */
				0x00400800,	/* frac = 2/32 */
				0x01010100,	/* frac = 3/32 */
				0x04082040,	/* frac = 4/32 */
				0x08210820,	/* frac = 5/32 */
				0x10884220,	/* frac = 6/32 */
				0x11111110,	/* frac = 7/32 */
				0x22444890,	/* frac = 8/32 */
				0x24892448,	/* frac = 9/32 */
				0x24922492,	/* frac = 10/32 */
				0x49494948,	/* frac = 11/32 */
				0x4A5294A4,	/* frac = 12/32 */
				0x52A552A4,	/* frac = 13/32 */
				0x552AA554,	/* frac = 14/32 */
				0x55555554,	/* frac = 15/32 */
				0xAAAAAAAA,	/* frac = 16/32 */
				0xAAD55AAA,	/* frac = 17/32 */
				0xAD5AAD5A,	/* frac = 18/32 */
				0xB5AD6B5A,	/* frac = 19/32 */
				0xB6B6B6B6,	/* frac = 20/32 */
				0xDB6DDB6C,	/* frac = 21/32 */
				0xDB76DBB6,	/* frac = 22/32 */
				0xDDBBB76E,	/* frac = 23/32 */
				0xEEEEEEEE,	/* frac = 24/32 */
				0xEF77BDDE,	/* frac = 25/32 */
				0xF7DEF7DE,	/* frac = 26/32 */
				0xFBF7DFBE,	/* frac = 27/32 */
				0xFEFEFEFE,	/* frac = 28/32 */
				0xFFBFF7FE,	/* frac = 29/32 */
				0xFFFEFFFE,	/* frac = 30/32 */
				0xFFFFFFFE};	/* frac = 31/32 */

#if 0
/* here are 16 bit versions of the same things.  They're left shifted
 * in a 32 bit word for ease of use.  We're not presently using them,
 * but they were so hard to make that I can't bear to discard them. 
 */
static UInt32 frac16LineVectors[16] = {
				0x00000000,	/* frac = 0/16 */
				0x01000000,	/* frac = 1/32 */
				0x08200800,	/* frac = 2/32 */
				0x11110000,	/* frac = 3/32 */
				0x24480000,	/* frac = 4/32 */
				0x25240000,	/* frac = 5/32 */
				0x52520000,	/* frac = 6/32 */
				0x55540000,	/* frac = 7/32 */
				0xAAAA0000,	/* frac = 8/32 */
				0xAD5A0000,	/* frac = 9/32 */
				0xDADA0000,	/* frac = 10/32 */
				0xDBB60000,	/* frac = 11/32 */
				0xEEEE0000,	/* frac = 12/32 */
				0xF7DE0000,	/* frac = 13/32 */
				0xFEFE0000,	/* frac = 14/32 */
				0xFFFE0000};	/* frac = 15/32 */
#endif  /* 0 */


/* This is a publically defined routine */
/* PUBLIC */
/****************************************************************************
 *  i_detectSkewFixedPoint():
 *        Uses MaxPixels and difference functions to determine skew (see
 *	  large comment above).
 *        Allowable reduction factors are 1, 2, 4, 8, 16, and 32.
 *
 *        Returns the skew angle in ScaledAngle format in *skewAngle.
 *	  On error, returns the proper error code.
 *
 *        Typical redFactors for 8.5x11 image at 200 to 300 pixels/inch are:
 *                4 to 8.
 *
 *	pImage is pointer to the image (including frame)
 *	imWidth is width of image in pixels including frame
 *	imHeight is height of image including frame
 *	imBpl is bytes/line of image (including frame)
 *	confidence is a pointer to value returned that indicates the
 *		confidence the application should have in the answer.
 *		The larger the confidence value, the better.
 *	xRes is the resolution in the X direction.
 *	yRes is the resolution in the Y direction.  The units of xRes
 *		and yRes are irrevelent so long as both xRes and
 *		yRes have the same units.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem 	could not allocate memory
 *		ia_noLines	could not find enough horizontal edges
 *		ia_invalidParm	xRes != yRes and either xRes == 0 or
 *				yRes == 0.
 ****************************************************************************/
Int32 CDECL 
i_detectSkewFixedPoint (
    IPState    *pIPState,
    UInt8Ptr	pImage,
    Int32	imWidth,
    Int32	imHeight,
    Int32	imBpl,
    Int32	redFactor,
    ScaledAnglePtr skewAngle,
    UInt32Ptr	confidence,
    Int32	xRes,
    Int32	yRes)
{
    PROLOGUE;
    Int32	minA, maxA, bestOffset1;
    Int32	numLines;
    

	/* See if the x and y resolutions are resonable */
    if ((xRes != yRes) && ((yRes == 0) || (xRes == 0)))
	return ia_invalidParm;

	/* do coarse search */
    numLines = cNumTrialLines;
    bestOffset1 = cBadOffset;
    __error__ = ia_successful;	/* keep gcc happy */
    while ((numLines > 0) && (bestOffset1 == cBadOffset))
    {
	__error__ = ip_detectSkewRobot (pIPState, pImage, imWidth - 2*cFrameBits,
				      imHeight -2*cFrameBits, imBpl,
				      numLines, &bestOffset1);
/* DEBUG ONLY!!!!!!!!!!!!!!!!!!*/
/*
	break;
*/
/* END DEBUG */

	numLines = numLines/2;
    }
    if (__error__ != ia_successful)
    {
	*skewAngle = 0;		/* skew of 0 radians is a good guess */
	*confidence = 0;	/* but not a guess we believe */
	if (__error__ == ia_noLines) 	/* this is not an error */
	{
	    __error__ = ia_successful;
	    NORMAL_RETURN;	/* not finding an angle is ok */
	}
	else
	    NORMAL_RETURN;	/* this is an error */
    }

    IP_YIELD(errorFinish)

	/* get offset range (in lines) and do bifurcation step.
	 * Note that if this is a very wide, flat thing, we'll
	 * try a smaller angular offset so that we won't quit
	 * just because the sampling paths run off the image.
	 */
    if ( (imHeight - 2*cFrameBits) > ((imWidth - 2*cFrameBits) / 4) )
    {
	minA = bestOffset1 + 2*cMinSkewOffset;
	maxA = bestOffset1 + 2*cMaxSkewOffset;
    }
    else	/* try only searching in a 1 degree range instead.
		 * If we can't do that, we're probably going to miss
		 * the correct answer.
		 */
    {
	minA = bestOffset1 + cMinSkewOffset;
	maxA = bestOffset1 + cMaxSkewOffset;
    }
    __error__ = ip_detectSkewFine (pIPState, pImage, imWidth - 2*cFrameBits,
				 imHeight - 2*cFrameBits, imBpl, redFactor,
				 minA, maxA, skewAngle, confidence,
				 xRes, yRes);

    NORMAL_RETURN;

errorFinish:
    *skewAngle = 0;		/* try to be deterministic */
    *confidence = 0;		/* but don't believe it */
    return ia_aborted;
}


/* This procedure is private */
/* PRIVATE */
/******************************************************************************
 *  ip_detectSkewFine():
 *
 *        Uses subsampling (NOT reduction!) on the image, and
 *	    differential line count function (the difference between the
 *          number of ON pixels in the current scanline and the number in
 *          the previous scanline).
 *        The search algorithm tries different sets of parallel paths across
 *	     the image, calculating scores for each set based on the
 *	     difference in the number of ON pixels found in each path.  We
 *	     perform a highly non-linear function on the scores so that
 *	     paths having a very large difference (horizontal edges) result
 *	     in much larger scores than paths having moderate differences
 *	     (random edges).
 *        Allowable subsampling factors are 1, 2, 4, 8, 16, and 32.
 *	     They are used decide how many vertical lines to skip
 *           between samples.
 *        The range over which a maximum is searched spans, in semi-normalized
 *           offset units, (minVertDisp, maxVertDisp).
 *           Where cw and ccw skew are expected
 *           to occur symmetrically, one can choose
 *                  minVertDisp = - maxVertDisp
 *	  The arguments to this routine are in "offset" units for the sake of
 *	     speed.  The former interfaces used radians, but they were
 *	     a pain to manipulate.
 *	     We return ScaledAngles, a fixed point version of radians.
 *	     ScaledAngle = (32768 * (theta in radians)/(2 * pi)).
 *        ip_detectSkewRobot() returns the lines of offset, relative
 *           to horizontal.  ip_detectSkewFine() does a binary search
 *           centered around the offset "angle" returned by
 *	     ip_detectSkewRobot().
 *           At the end, ip_detectSkewFine() converts lines of offset to
 *           ScaledAngles.
 *        Computation stops when the offset increment between samples
 *           becomes less than a rise (fall) of 1 line in 1024.
 *        Algorithm takes the power values at two points
 *                power1               power2
 *                angMin               angMax
 *	  The offset with the smaller power value is moved approx.
 *          half way towards the other offset and the new power value is taken. 
 *          This process repeats until they are next to each other.
 *
 *	pImage: 	pointer to upper left corner of frame
 *	imWidth:	width of image in pixels including frame
 *	imHeight:	width of image in pixels including frame
 *	imBpl:		image's bytes/line
 *	redFactor:	subsampling factor (1, 2, 4, 8, 16, 32 are legal)
 *			The larger redFactor, the slower and more accurate
 *			the routine is.
 *	minVertDisp, maxVertDisp:
 *			Range over which search for the image's real slope
 *			takes place.  These are specified in units of lines
 *			one moves up (neg) or down (pos) as one crosses the
 *			entire image to achieve the specified slope.  For
 *			example, if the image were 1000 pixels wide and we
 *			wanted to search from +3 degrees to -2 degrees,
 *			minVertDisp = tan(-2)*1000 = -35, maxVertDisp =
 *			tan(3)*1000 = 52.
 *			It is intended that ip_detectSkewFine be called
 *			immediately after ip_detectSkewRobot, which returns its
 *			estimate of the skew in these units.
 *	skewAngle	Pointer to returned angle in ScaledAngle format
 *	confidence	Pointer to returned confidence measure.  If this
 *			value is small, it is unlikely that the skew angle
 *			was accurately found.  The value returned is a
 *			measure of the average angular distance from the
 *			chosen angle in the rest of the angles.  To gain
 *			resolution, this value has already been multiplied
 *			by 100.  It is in units of lines we move up or
 *			down out of cHypoImWidth lines.
 *	xRes		The resolution in the X direction.
 *	yRes		The resolution in the Y direction.  The units of
 *			xRes and yRes are irrevelent so long as both xRes and
 *			yRes have the same units.
 ******************************************************************************/
Int32 CDECL 
ip_detectSkewFine (
    IPState    *pIPState,
    UInt8Ptr	pImage,
    Int32	imWidth,
    Int32	imHeight,
    Int32	imBpl,
    Int32	redFactor,
    Int32	minVertDisp,
    Int32	maxVertDisp,
    ScaledAnglePtr skewAngle,
    UInt32Ptr	confidence,
    Int32	xRes,
    Int32	yRes)
{
    PROLOGUE;
    Int32	midVertDisp;		/* original center of region
					 * searched.  Don't cross this without
					 * sampling here. */
    Int32	maxAngles;
    Int32	maxAngleIndex, minAngleIndex, nextAngleIndex;
    Int32	numVSamp, numHSamp;
    Int32	hAddrDisp;		/* distance between horiz samples
					 * in bytes. */
    Int32	hSubSampleFactor; /* normally is 1.  we don't take
					 * as many samples in a wide image,
					 * subsampling every 2**n words.
					 */
    Int32	sampleHeight, topLine, bottomLine;
    Int32	slantAddrDisp, fracBitVector, vAddrDisp, absVertDisp;
    Int32Ptr	power;			/* power[evenNum] is the low order
					 * 31 bits of the sum, power[evenNum+1]
					 * is the upper 32 bits of the number
					 */
    Int32Ptr	angle;
    ScaledAngle	bestAngle;
    UInt32	totalPower[2], avgAngle[2]; /* these keep the high word
					 * in [1] and the low word in [0]
					 */
    Int32	avgAngleShiftCount;	/* amount we shift avgAngle right
					 * to normalize it.
					 */
    Int32	totalPowerShiftCount;	/* amount we shift totalPower right
					 * to normalize it.
					 */
    UInt32	localConfidence;
    Int32	tempOffset, finalOffset, index;
    UInt32	logAngleStepSize;	/* equals log of step size between
					 * offsets.
					 */

/* DO NOT CHANGE THESE VALUES WITHOUT MODIFYING cMinSkewOffset,
 * cMaxSkewOffset and cReasonableDivergence ABOVE!!!.  Those quantities
 * are set up so 16 offset lines is about 1 degree, which works when
 * cHypoImWidth is 1024.
 */
#define	cHypoImWidth		1024	/* width of image used to compute
					   vertical displacements */
#define	cLogHypoImWidth		10	/* used for various conversions */
#define	cHAddrDisp		4	/* horiz distance between samples
					   in bytes */
#define	cLogHAddrDisp		2	/* used for conversions */
#define cMaxHSamples		127	/* not allowed to have more horizontal
					 * samples than this.
					 */
#define	cCoarseDegToDisp	cHypoImWidth/64 /* factor used to convert
					 * vertical rise to degrees quickly */
/* on completion, we need to convert the vertical displacement to radians */
#define	cNumSamples		32	/* # of horiz samples across
					   hypothetical 1Kbit image */
#define	cLogNumSamples		5	/* log2(32), used to find amount of
					   vertical disp to be added on
					   every sample */
#define cNumSamplesMask		0x1F	/* used to find amount of vert disp
					   spread out over all samples */


	/* initialize */
    power = NULL;
    angle = NULL;
    midVertDisp = (minVertDisp + maxVertDisp)/2;

    imWidth -= 2*cFrameBits;
    imHeight -= 2*cFrameBits;

	/* The slanted lines for a set of samples start from fixed points
	 * on the left side of the image and slope up or down as they go
	 * to the right side.  It's illegal to have the lines slope up
	 * off the top or down off the bottom of the image.  We account for
	 * slopes by restricting the area on the left side of the image
	 * in which lines may start.
	 * Remembering that the Y coordinate increases as we go DOWN the
	 * image (hence "negative" slopes go towards the top of the image)
	 * we calculate the highest starting position by saying that the
	 * the most slanted line will move minVertDisp bits up for every
	 * cHypoImWidth bits it moves over.  The total vertical displacement is
	 * (minVertDisp/cHypoImWidth)*imWidth.  If minVertDisp is a positive
	 * number, we use 0 instead, so topLine doesn't move.  If maxVertDisp
	 * is negative, we leave bottomLine at imHeight.
	 * BUG FIX - 10/6/93: The confidence measure is done by looking +- 3
	 * degrees (roughly) from the final angle determined.  This final
	 * angle could be anywhere between minVertDisp and maxVertDisp.
	 * We must take this extra offset into account when fixing the
	 * starting points or we'll read off the top or bottom of the
	 * image when computing the confidence.
	 */

    topLine = (MAX(0, -(minVertDisp + 3*cMinSkewOffset)) * imWidth) >>
		cLogHypoImWidth;
    bottomLine = imHeight -
		   ((MAX(0, (maxVertDisp + 3*cMaxSkewOffset)) * imWidth) >>
			cLogHypoImWidth);
    sampleHeight = bottomLine - topLine + 1;

	/* If we've just avoided the entire image, quit.  We'll return
	 * the sample angle with a confidence of 0.
	 * BUG FIX 10/21/96: We should compare sampleHeight with redFactor
	 * instead of 0 since we're going to divide it by redFactor to
	 * produce the number of samples to be taken.  Taking one sample
	 * is silly, so give a confidence of 0 for sampleHeight == redFactor
	 * too.
	 */
    if (sampleHeight <= redFactor)
    {
	*skewAngle = 0;
	*confidence = 0;
	return ia_noLines;
    }

	/* skip over top pad */
    pImage += (imBpl << cLogFrameBits) + cFrameBytes;
	/* move down far enough so most skewed path will still be inside the
	   image */
    pImage += imBpl * topLine;

	/* Initialize power and angle arrays.  The power array uses
	 * two words to hold each power value.
	 */
    maxAngles = 100;  /* this is a safely large number!! */
    if ((power = (Int32Ptr) CALLOC(maxAngles, 2*sizeof(Int32))) == NULL)
    {
	return ia_nomem;
    }
    if ((angle = (Int32Ptr) CALLOC(maxAngles, sizeof(Int32))) == NULL)
    {
	return ia_nomem;
    }

	/* Calculate # of bytes between vert grid lines */
    vAddrDisp = imBpl * redFactor;

	/* We use a binary search to find the correct angle.  However,
	 * we don't cut the distance between the angles in half
	 * at each step.  In an earlier version of the code, we
	 * moved one of the endpoints in by 1/4 of the distance
	 * between them.  That was usually ok, but sometimes too
	 * big.  We now set the step size as a function of the
	 * reduction factor.
	 */
    switch (redFactor)
    {
    case 1:
	logAngleStepSize = 4;		/* move by (max-min)/16 */
	break;
    case 2:
	logAngleStepSize = 3;		/* move by (max-min)/8 */
	break;
    default:
	logAngleStepSize = 2;		/* move by (max-min)/4 */
	break;
    }

	/* number of horizontal samples/grid line equals width of image
	 * /8 {convert to bytes} /(bytes/sample).  The -1 is for safety.
	 * Note that we're not allowed to take more than 127 horizontal
	 * samples in any one line.  If we're trying to, we increase the
	 * distance between samples.  To make this work, we have to
	 * make the slope between samples steeper, so we increase
	 * absVertDisp by the same factor.  absVertDisp is really the
	 * distance we move vertically in 32 horizontal samples.
	 */
    numHSamp = (imWidth >> (3 + cLogHAddrDisp)) - 1;
    hAddrDisp = cHAddrDisp;
    hSubSampleFactor = 0;
    while (numHSamp > cMaxHSamples)
    {
	numHSamp /= 2;
	hAddrDisp *= 2;
	hSubSampleFactor++;
    }
    numVSamp = sampleHeight/(redFactor);

        /* get the first two points */
    absVertDisp = ABS(minVertDisp) << hSubSampleFactor;
    slantAddrDisp = (absVertDisp >> cLogNumSamples) * imBpl;
    if (minVertDisp < 0) slantAddrDisp = -slantAddrDisp;
    fracBitVector = frac32LineVectors[absVertDisp & cNumSamplesMask];
    minAngleIndex = 0;
    angle[minAngleIndex] = minVertDisp;
    ERR_PASS_CLEANUP( ip_findDiffPixelsSum (pIPState, pImage,
					    numVSamp, numHSamp,
					    vAddrDisp,
					    (hAddrDisp + slantAddrDisp),
					    (minVertDisp>=0 ? imBpl:-imBpl),
					    fracBitVector,
					    &(power[2*minAngleIndex]),
					    &(power[2*minAngleIndex+1])));

    absVertDisp = ABS(maxVertDisp) << hSubSampleFactor;
    slantAddrDisp = (absVertDisp >> cLogNumSamples) * imBpl;
    if (maxVertDisp < 0) slantAddrDisp = -slantAddrDisp;
    fracBitVector = frac32LineVectors[absVertDisp & cNumSamplesMask];
    maxAngleIndex = 1;
    angle[maxAngleIndex] = maxVertDisp;
    ERR_PASS_CLEANUP( ip_findDiffPixelsSum (pIPState, pImage,
					    numVSamp, numHSamp,
					    vAddrDisp,
					    (hAddrDisp + slantAddrDisp),
					    (maxVertDisp>=0 ? imBpl:-imBpl),
					    fracBitVector,
					    &(power[2*maxAngleIndex]),
					    &(power[2*maxAngleIndex+1])));

	/* begin the binary search loop.  At each step, we compare the
	 * scores of maxAngle and minAngle.  If maxAngle's is
	 * smaller, we assume the peak is closer to minAngle, so we
	 * move maxAngle to half way between and evaluate it.  If
	 * minAngle's score was smaller, we move minAngle and
	 * evaluate.  We stop when the distance between maxVertDisp
	 * and minVertDisp is equal to 1.  Since this is a pretty peaky
	 * function, don't take such big steps at first.  The
	 * (maxVertDisp - minVertDisp)/4 modify the first steps.
	*/
    nextAngleIndex = 2;
    while (maxVertDisp-minVertDisp > 1)
    {
	if ((power[2*minAngleIndex+1] < power[2*maxAngleIndex+1]) ||
	    ((power[2*minAngleIndex+1] == power[2*maxAngleIndex+1]) &&
	     (power[2*minAngleIndex] <= power[2*maxAngleIndex])))
	{	/* move minAngle up.  If we're near the end,
		 * make only very fine adjustments.  Don't cross
		 * the midpoint of the span without sampling there.
		 */
	    if (((UInt32)(maxVertDisp - minVertDisp) >> logAngleStepSize) == 0)
		minVertDisp++;
	    else  /* move the bottom point toward the top point */
	    if (minVertDisp < midVertDisp)
	    {
		minVertDisp +=
		    (UInt32)(maxVertDisp - minVertDisp) >> logAngleStepSize;
		if (minVertDisp > midVertDisp)
		    minVertDisp = midVertDisp;
	    }
	    else
		minVertDisp +=
		    (UInt32)(maxVertDisp - minVertDisp) >> logAngleStepSize;
	    absVertDisp = ABS(minVertDisp) << hSubSampleFactor;
	    slantAddrDisp = (absVertDisp >> cLogNumSamples) * imBpl;
	    if (minVertDisp < 0) slantAddrDisp = -slantAddrDisp;
	    fracBitVector = frac32LineVectors[absVertDisp & cNumSamplesMask];
	    minAngleIndex = nextAngleIndex++;
	    angle[minAngleIndex] = minVertDisp;
	    ERR_PASS_CLEANUP (
		ip_findDiffPixelsSum (pIPState, pImage,
				      numVSamp, numHSamp,
				      vAddrDisp,
				      (hAddrDisp + slantAddrDisp),
				      (minVertDisp >=0 ? imBpl : -imBpl),
				      fracBitVector,
				      &(power[2*minAngleIndex]),
				      &(power[2*minAngleIndex+1])));
	}
	else
	{	/* move maxAngle down */
	    if (((UInt32)(maxVertDisp - minVertDisp) >> logAngleStepSize) == 0)
		maxVertDisp--;
	    else
	    if (maxVertDisp > midVertDisp)
	    {
		maxVertDisp -= 
		    (UInt32)(maxVertDisp - minVertDisp) >> logAngleStepSize;
		if (maxVertDisp < midVertDisp)
		    maxVertDisp = midVertDisp;
	    }
	    else
		maxVertDisp -= 
		    (UInt32)(maxVertDisp - minVertDisp) >> logAngleStepSize;

	    absVertDisp = ABS(maxVertDisp) << hSubSampleFactor;
	    slantAddrDisp = (absVertDisp >> cLogNumSamples) * imBpl;
	    if (maxVertDisp < 0) slantAddrDisp = -slantAddrDisp;
	    fracBitVector = frac32LineVectors[absVertDisp & cNumSamplesMask];
	    maxAngleIndex = nextAngleIndex++;
	    angle[maxAngleIndex] = maxVertDisp;
	    ERR_PASS_CLEANUP (
		ip_findDiffPixelsSum (pIPState, pImage,
				      numVSamp, numHSamp,
				      vAddrDisp,
				      (hAddrDisp + slantAddrDisp),
				      (maxVertDisp >=0 ? imBpl : -imBpl),
				      fracBitVector,
				      &(power[2*maxAngleIndex]),
				      &(power[2*maxAngleIndex+1])));
	}
    } /* of binary search loop */

    IP_YIELD(ERR_CLEANUP)		/* let other tasks run */

	/* Return result in units of scaled angles (32768 * (theta/2pi)).
	 *
	 * Note that if we didn't pay attention to x and y resolutions
	 * at this step, then the angle returned would be a function
	 * of x resolution, y resolution and skew.  Let's introduce
	 * the concept of the nearest isotropic image.  That's the
	 * smallest image corresponding to the current image in which
	 * a) xRes = yRes and b) the x and y scale factors used to
	 * obtain this image from the current image are both greater than
	 * or equal to 1.0.  To compute it, we look at xRes and yRes.
	 * If they're already equal, stop.  If one is smaller, scale
	 * the image in that dimension ONLY so that xRes == yRes..
	 * For example, if we're given a standard fax image with
	 * xRes = 200 and yRes = 100, scale the image by 2x vertically
	 * only.
	 * The vertical displacement we've calculated so far is the
	 * vertical displacement for the nearest isotropic image 
	 * times (yRes/xRes).  That is, we can pretend that all the
	 * vertical distances were scaled by this factor.  We'd like
	 * to return a skew angle for the nearest isotropic image so
	 * that it ws independent of the resolutions of the image used
	 * to compute it.  To remove the x and y resolution factors
	 * (i.e. compute the skew angle for the nearest isotropic image),
	 * we multiply the vertical displacement by (xRes/yRes).
	 */
    if ((power[2*maxAngleIndex+1] > power[2*minAngleIndex+1]) ||
	((power[2*maxAngleIndex+1] == power[2*minAngleIndex+1]) &&
	 (power[2*maxAngleIndex] >= power[2*minAngleIndex])))
    {
        if (ABS(xRes - yRes) > 1)
	    __error__ = iatan2(pIPState, cHypoImWidth,
			       (-angle[maxAngleIndex] * xRes)/yRes,
			       &bestAngle);
	else	/* xRes == yRes */
	    __error__ = iatan2(pIPState, cHypoImWidth,
			       -angle[maxAngleIndex],
			       &bestAngle);

	if (__error__ != ia_successful)
	{
	    *skewAngle = 0;
	    *confidence = 0;
	    NORMAL_RETURN;
	}
	finalOffset = angle[maxAngleIndex];
    }
    else
    {
        if (ABS(xRes - yRes) > 1)
	    __error__ = iatan2(pIPState, cHypoImWidth,
			       (-angle[minAngleIndex] * xRes)/yRes,
			       &bestAngle);
	else	/* xRes == yRes */
	    __error__ = iatan2(pIPState, cHypoImWidth,
			       -angle[minAngleIndex],
			       &bestAngle);

	if (__error__ != ia_successful)
	{
	    *skewAngle = 0;
	    *confidence = 0;
	    NORMAL_RETURN;
	}
	finalOffset = angle[minAngleIndex];
    }

	/* compute confidence (see comment block at beginning of file).
	 * First, look +- 2 degrees and +- 3 degrees from the peak
	 * we've found.  This should give us a much better idea of whether
	 * this is real peak or just noise.
	 */

	/* Look at finalOffset - 2 degrees */
    maxVertDisp = finalOffset + 2*cMinSkewOffset;
    absVertDisp = ABS(maxVertDisp) << hSubSampleFactor;
    slantAddrDisp = (absVertDisp >> cLogNumSamples) * imBpl;
    if (maxVertDisp < 0) slantAddrDisp = -slantAddrDisp;
    fracBitVector = frac32LineVectors[absVertDisp & cNumSamplesMask];
    maxAngleIndex = nextAngleIndex++;
    angle[maxAngleIndex] = maxVertDisp;
    ERR_PASS_CLEANUP (
	ip_findDiffPixelsSum (pIPState, pImage,
			      numVSamp, numHSamp,
			      vAddrDisp,
			      (hAddrDisp + slantAddrDisp),
			     ( maxVertDisp >=0 ? imBpl : -imBpl),
			      fracBitVector,
			      &(power[2*maxAngleIndex]),
			      &(power[2*maxAngleIndex+1])));

	/* Look at finalOffset + 2 degrees */
    maxVertDisp = finalOffset + 2*cMaxSkewOffset;
    absVertDisp = ABS(maxVertDisp) << hSubSampleFactor;
    slantAddrDisp = (absVertDisp >> cLogNumSamples) * imBpl;
    if (maxVertDisp < 0) slantAddrDisp = -slantAddrDisp;
    fracBitVector = frac32LineVectors[absVertDisp & cNumSamplesMask];
    maxAngleIndex = nextAngleIndex++;
    angle[maxAngleIndex] = maxVertDisp;
    ERR_PASS_CLEANUP (
	ip_findDiffPixelsSum (pIPState, pImage,
			      numVSamp, numHSamp,
			      vAddrDisp,
			      (hAddrDisp + slantAddrDisp),
			      (maxVertDisp >=0 ? imBpl : -imBpl),
			      fracBitVector,
			      &(power[2*maxAngleIndex]),
			      &(power[2*maxAngleIndex+1])));

	/* Look at finalOffset - 3 degrees */
    maxVertDisp = finalOffset + 3*cMinSkewOffset;
    absVertDisp = ABS(maxVertDisp) << hSubSampleFactor;
    slantAddrDisp = (absVertDisp >> cLogNumSamples) * imBpl;
    if (maxVertDisp < 0) slantAddrDisp = -slantAddrDisp;
    fracBitVector = frac32LineVectors[absVertDisp & cNumSamplesMask];
    maxAngleIndex = nextAngleIndex++;
    angle[maxAngleIndex] = maxVertDisp;
    ERR_PASS_CLEANUP (
	ip_findDiffPixelsSum (pIPState, pImage,
			      numVSamp, numHSamp,
			      vAddrDisp,
			      (hAddrDisp + slantAddrDisp),
			      (maxVertDisp >=0 ? imBpl : -imBpl),
			      fracBitVector,
			      &(power[2*maxAngleIndex]),
			      &(power[2*maxAngleIndex+1])));

	/* Look at finalOffset + 3 degrees */
    maxVertDisp = finalOffset + 3*cMaxSkewOffset;
    absVertDisp = ABS(maxVertDisp) << hSubSampleFactor;
    slantAddrDisp = (absVertDisp >> cLogNumSamples) * imBpl;
    if (maxVertDisp < 0) slantAddrDisp = -slantAddrDisp;
    fracBitVector = frac32LineVectors[absVertDisp & cNumSamplesMask];
    maxAngleIndex = nextAngleIndex++;
    angle[maxAngleIndex] = maxVertDisp;
    ERR_PASS_CLEANUP (
	ip_findDiffPixelsSum (pIPState, pImage,
			      numVSamp, numHSamp,
			      vAddrDisp,
			      (hAddrDisp + slantAddrDisp),
			      (maxVertDisp >=0 ? imBpl : -imBpl),
			      fracBitVector,
			      &(power[2*maxAngleIndex]),
			      &(power[2*maxAngleIndex+1])));

	/* Here's where we compute the confidence.  The power numbers
	 * can be quite large.  To avoid overflow without resorting to
	 * floating point, we divide them by 32 before multiplying them
	 * by the angular offset.  We correct the result in the
	 * final division.
	 */
    totalPower[0] = totalPower[1] = 0;
    avgAngle[0] = avgAngle[1] = 0;
    for (index = 0; index < nextAngleIndex; index++)
    {
	totalPower[0] += power[2*index];
	if (totalPower[0] & 0x80000000)	/* wrapped around to neg number */
	{
	    totalPower[0] &= 0x7FFFFFFF;
	    totalPower[1] += power[2*index+1] + 1;
	}
	else
	{
	    totalPower[1] += power[2*index+1];
	}
	    /* now we have to do a double-precision multiply.
	     * We first calculate tempOffset, known to be a fairly
	     * small number.  We then separate the power into
	     * three sections, bits [15:0], which when multiplied by
	     * tempOffset always results in a single word; bits
	     * [31:16], which may result in a carry to the upper word
	     * when multiplied by tempOffset; and bits [47:32].
	     * All of the product bits from the last set end up
	     * in the upper word.  We shift the bits [31:16] down
	     * to the bottom of the word before multiplying, then
	     * shift the product up before adding it into the
	     * lower word and down (by 16) before adding it to the
	     * upper word.
	     */
	tempOffset = ABS(angle[index] - finalOffset);
	avgAngle[0] += (power[2*index] & 0x0000FFFF) * tempOffset;
	avgAngle[1] += power[2*index+1] * tempOffset;
	    /* calc cross product.  Bits that should go into
	     * avgAngle[1] end up in tempOffset[31:15], NOT
	     * tempOffset[31:16].  To be consistent, we want
	     * to keep avgAngle[0].31 clear all the time.
	     */
	tempOffset = ((UInt32)(power[2*index]) >> 16) * tempOffset;
	avgAngle[0] += (tempOffset << 16) & 0x7FFFFFFF;
	avgAngle[1] += tempOffset >> 15;
    }

	/* Now for a double-precision divide.  Since there is no such
	 * operation in C, we cheat.  We shift avgAngle right until there
	 * are no more bits in avgAngle[1].  For each shift, we increment
	 * shiftCount so we remember the shifts.  We also shift totalPower
	 * right by 1 until there are no bits in totalPower[1].  For each
	 * of these shifts, we decrement shiftCount because the quotient
	 * avgAngle/totalPower should be shifted right for each shift of
	 * totalPower.  We then compute the quotient and shift it
	 * left by shiftCount (or right by -shiftCount if shiftCount is
	 * negative).
	 * This step causes a divide by zero error when there is very
	 * little in the image, i.e. totalPower < 32.  In this case,
	 * we know that (power[index]/32) was also zero for all values
	 * of index, therefore, avgAngle == 0 also.  Rather than say
	 * that 0/0 is small, we'll say it's big.  This means we
	 * have no confidence in the answer.
	 */
	/* First, close the hole in avgAngle.  It's at avgAngle[0].31.
	 */
    avgAngleShiftCount = 0;
    avgAngle[0] |= (avgAngle[1] << 31);
    avgAngle[1] >>= 1;
    while ((1UL << avgAngleShiftCount) <= avgAngle[1])
	avgAngleShiftCount++;

	/* avgAngleShiftCount is now equal to the amount by which we
	 * should shift avgAngle right to make it fit into avgAngle[0].
	 * Note that shifting by (32 - 0) is the same as shifting by 0
	 * on most machines!  If avgAngleShiftCount == 0, we really don't
	 * want to adjust either avgAngle[0] or avgAngle[1], so skip
	 * these ops in that case.
	 */
    if (avgAngleShiftCount != 0)
    {
	avgAngle[0] >>= avgAngleShiftCount;
	avgAngle[0] |= avgAngle[1] << (32 - avgAngleShiftCount);
    }
	/* make sure avgAngle[0] is positive.  The sign bit could have
	 * been set in either of the last two shifts.
	 */
    if (avgAngle[0] & 0x80000000)
    {
	avgAngleShiftCount++;
	avgAngle[0] >>= 1;
    }

	/* Now do the same for totalPower. */
    totalPowerShiftCount = 0;
    totalPower[0] |= (totalPower[1] << 31);
    totalPower[1] >>= 1;
    while ((1UL << totalPowerShiftCount) <= totalPower[1])
	totalPowerShiftCount++;

    if (totalPowerShiftCount != 0)
    {
	totalPower[0] >>= totalPowerShiftCount;
	totalPower[0] |= totalPower[1] << (32 - totalPowerShiftCount);
    }

	/* make sure totalPower[0] is positive.  The sign bit could have
	 * been set in either of the last two shifts.
	 */
    if (totalPower[0] & 0x80000000)
    {
	totalPowerShiftCount++;
	totalPower[0] >>= 1;
    }

	/* The interesting bits are in totalPower[0] and avgAngle[0].  If
	 * totalPower[0] > 0, do the divide.  Otherwise, just say
	 * the angle is too big.
	 */
    if (totalPower[0] > 0)
    {
	localConfidence = avgAngle[0] / totalPower[0];
	avgAngleShiftCount = avgAngleShiftCount - totalPowerShiftCount;
	if (avgAngleShiftCount > 0)
	    localConfidence <<= avgAngleShiftCount;
	else
	    localConfidence >>= -avgAngleShiftCount;
    }
    else
	localConfidence = cReasonableDivergence;

	/* It's nice if the confidence is a unitless number.  We arbitrarily
	 * decide to compute this as the fraction of 1 degree taken by the
	 * average offset computed by ip_detectSkewFine.  If the average offset
	 * is more than a degree, we set the confidence to 0.  We scale the
	 * result so it will be a number from 0 to 100.
	 */
    if (localConfidence >= cReasonableDivergence)
        *confidence = 0;
    else
	*confidence = (100 * (cReasonableDivergence - localConfidence)) /
		       cReasonableDivergence;

#if  DEBUG_PRINT_SKEW
    fprintf(stderr, " Bifurcation: Skew angle is %8.3f deg.\n",
		    (2 * 3.1415926 * bestAngle) / 32768.0);
    fprintf(stderr, " Confidence factor is %d\n", *confidence);
#endif  /* DEBUG_PRINT_SKEW */

#if  DEBUG_PLOT_SKEW
    debugPlotSkew(nextAngleIndex, power, angle, (57.29/1024.));
#endif  /* DEBUG_PLOT_SKEW */

	/* clean up */
    FREE(angle);
    FREE(power);

    *skewAngle = bestAngle;
    NORMAL_RETURN;

ERR_CLEANUP:
    FREE(angle);
    FREE(power);

    *skewAngle = 0;		/* try to be deterministic */
    ERR_RETURN;
}



/* This procedure is private */
/* PRIVATE */
/******************************************************************************
 *  ip_detectSkewRobot():
 *
 *        Uses probes to find and follow nearly-horizontal edges.  The
 *	    probes record the slopes required to follow the edges.
 *	  See large comment in sskew.c for a complete explanation.
 *	  Basically, this routine drops a bunch of little probes into
 *	    the image.  Each probe tries to find and follow a horizontal
 *	    edge as would appear at the bottom of a line of text.
 *	    The result is the cumulative slope found by all of the
 *	    successful probes.
 *        Returns the lines of vertical offset relative to horizontal.
 *	pImage: 	pointer to upper left corner of frame
 *	imWidth:	width of image in pixels including frame
 *	imHeight:	width of image in pixels including frame
 *	imBpl:		image's bytes/line
 *	numLines:	Quit when the probes have seen a total of at
 *			least (numLines * pixels per line) pixels
 *	offset:		The resulting vertical displacement
 ******************************************************************************/
Int32 CDECL 
ip_detectSkewRobot (
    IPState    *pIPState,
    UInt8Ptr	pImage,
    Int32	imWidth,
    Int32	imHeight,
    Int32	imBpl,
    Int32	numLines,
    Int32Ptr	offset)
{
    PROLOGUE;
    Slope	imSlope;
    Int32	numTests;
    Int32	imWidthInBytes;
    Int32	numYSamples, vAddrDisp;
    Int32	colAddrOffset, row, rowAddrOffset;
    Int32	minLegalAddr, maxLegalAddr;
    Float64	conversion;
#if  DEBUG_PRINT_SKEW
    Float64	bestAngle;
#endif /* DEBUG_PRINT_SKEW */

#define	cXSampleSpacing		256	/* sample pt every 256 bits horiz */
#define	cXSampleSpacingInBytes	32	/* sample pt every 32 bytes horiz */
#define	cYSampleSpacing		32	/* 32 lines between samples */
#define	cLogYSampleSpacing	5
#define	cLogBitsPerHalfWord	4	/* log2(16) */
#define	cLogBitsPerWord		5	/* log2(32) */
#define	cMaxVertSearch		16	/* ip_doEdgeSample will look no more
					   than 16 lines down before giving
					   up.  Used to set size of sample
					   area. */


	/* initialize */
    conversion = 3.1415926535 / 180.;
    imWidthInBytes = imWidth >> 3;
    if (imWidthInBytes < cXSampleSpacingInBytes)
    {
	*offset = cBadOffset;
	return ia_noLines;
    }

	 /* skip over top pad */
    pImage += (imBpl << cLogFrameBits) + cFrameBytes;
    	/* skip to first sample point */
    pImage += cXSampleSpacingInBytes/2;

	/* compute the number of samples in each column */
    numYSamples = imHeight/cYSampleSpacing - 2;
    if (numYSamples <= 0)
    {
	*offset = cBadOffset;
	return ia_noLines;
    }

    vAddrDisp = imBpl << cLogYSampleSpacing;
    minLegalAddr = (Int32) pImage;
    maxLegalAddr = minLegalAddr + vAddrDisp*numYSamples;

	/* If we were trying to be really exact, we'd do a better job
	   of determining how much white space the edge tracer is allowed
	   to skip.  Measuring slope over wide lines is generally a good
	   thing, unless the lines are in different columns.  However, since
	   this is only being used as an approximation of slope, we'll
	   just let the thing skip over big amounts of white space.
	   JUST IN CASE: the way to decide how much white space to use is
	   to measure the number of white space blocks/edge when allowing
	   the edge follower to skip lots of white space.  If the number of
	   blocks is less than 2, we're probably skipping over columns and
	   don't need to skip white space much anyway, so set the amount
	   of white space we're allowed to skip to 0.  "Lots" of white
	   space is about 4*16 samples. 
	*/
    numTests = numLines * (imWidth >> cLogBitsPerHalfWord);
    if (numTests == 0)
    {
	*offset = cBadOffset;
	return ia_noLines;
    }

    imSlope.hCount = imSlope.vCount = 0;
	/* loop through columns of samples */
    for (colAddrOffset = 0;
	 (imSlope.hCount < numTests) && (colAddrOffset < imWidthInBytes);
	 colAddrOffset += cXSampleSpacingInBytes)
    {
	for (row = 0, rowAddrOffset = 0;
	     (imSlope.hCount < numTests) &&
	         (row < imHeight - cMaxVertSearch);
	     row += cYSampleSpacing, rowAddrOffset += vAddrDisp)
	{
	    ip_doEdgeSample (pIPState, pImage+colAddrOffset+rowAddrOffset,
			     imBpl, &imSlope,
			     cMaxVertSearch, minLegalAddr, maxLegalAddr);
/* DEBUG ONLY!!!!!!!!!!!!!!!! */
/*
	    if (imSlope.vCount != 0)
	    {
		*offset = cBadOffset;
		printf("imSlope.hCount: %d, imSlope.vCount: %d\n", imSlope.hCount, imSlope.vCount);
		return ia_noLines;
	    }
*/
/* END DEBUG */
	}
    }

	/* At this point, we either saw enough edges or ran through all the
	 * sample points.  If we didn't find enough edges, complain.
	*/
    if (imSlope.hCount < numTests)
    {
	*offset = cBadOffset;
	return ia_noLines;
    }

	/* return result in lines of offset from 0 */
    	
#if  DEBUG_PRINT_SKEW
    	/* print result in RADIANS */
    bestAngle = (Float64) (-atan((Float64) (imSlope.vCount) /
    		              (Float64) (imSlope.hCount << cLogBitsPerHalfWord)));

    fprintf(stderr, " Robot: Skew angle is %8.3f deg.\n",
		    bestAngle / (3.1415926535 / 180.));
#endif  /* DEBUG_PRINT_SKEW */

	/* returns the number of lines of offset from 0 */
    *offset = (imSlope.vCount << cLogHypoImWidth) /
	   (imSlope.hCount << cLogBitsPerHalfWord);
    NORMAL_RETURN;
}

/*

$Log: i_skwfxp.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.19  1998/01/13 23:49:52  ddavies
i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.18  1996/10/21 19:27:49  ddavies
ddavies on Mon Oct 21 12:19:40 PDT 1996

i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v

Fix bug found by CEO of Visioneer in Core 10 version of skew detect.
ip_detectSkewFine traces diagonal paths across the image.  It doesn't want
to run off the top or bottom of the image.  It computes how far down it
has to start so it won't run off the top and how far up from the bottom
it has to start so it won't run off the bottom.  The old code checked to
see if there was any image left between these two boundaries.  This was
wrong because we compute the number of vertical samples as
sampleHeight (the difference between the two limits) / redFactor.  If
the sampleHeight was greater than 0 and less than redFactor (usually 4),
we would call lower level routines with a number of vertical samples
being 0.  The assembly routines didn't like that, they tried to decrement
it to get to 0, so kept going essentially forever.  The new code
exits if sampleHeight <= redFactor.
Since the image on which this test failed was a fairly nice image that
you'd probably want to work, we also changed the conditions under which
ip_detectSkewFine is called.  If the image is wide and flat, so the diagonal
paths are likely to cross the bottom, try with a band of 1 degree around
the angle returned by ip_detectSkewRobot instead of a band of +- 2 degrees.

Revision 1.17  1996/10/05 02:20:26  ddavies
i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v

Check to see if an image is too narrow or too flat to have any samples.  Without
these checks, it got a divide by 0 check.

Revision 1.16  1996/05/22 01:16:48  ddavies
i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v

i_detectSkewFixedPoint no longer returns beastAngle. It returns an angle of 0 with
a confidence of 0 when it couldn't find an angle.

Revision 1.15  1995/12/13 02:46:15  ejaquith
i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.14  1995/12/13  01:46:41  ddavies
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Use the xResolution and yResolution to remove the native resolution dependence
 * of angle calculated here.  For example, if yRes < xRes, all vertical features
 * will be relatively small.  Therefore, the angle calculated from arctan(dy/dx)
 * will be too small.  In the old world, the rotation code knew this and
 * multiplied sin(rotAngle) by xRes/yRes before using it.  This created an
 * unpleasant linkage between the skew detection code and the small angle rotation
 * code.
 * We now have ip_detectSkewFine multiply the y displacement by (xRes/yRes)
 * before calculating the arctan.  To do this, we must pass yRes and xRes down
 * to it via i_detectSkew.
 *
 * Revision 1.13  1995/02/24  02:30:46  ddavies
 * ddavies on Thu Feb 23 18:29:51 PST 1995
 *
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Ron Danisewicv at XIS found an image which caused ip_detectSkewFine to
 * overflow its counts at several places.  The code has been changed in
 * several ways.  First, the scores for different angles are held in 64-bit
 * words so we have more bits.  Second, we take at most 127 samples across
 * an image in any one horizontal pass.  If the image is wider than 128*32 =
 * 4K pixels, we start subsampling horizontally.  To do this correctly, the
 * vertical displacement (which is really the vertical displacement reached
 * after 32 horizontal samples) must be multiplied by the subsampling factor.
 * Furthermore, the horizontal displacement must also be multiplied by the
 * subsampling factor.  Third, when we compare scores, we must do
 * double-precision compares.  Finally, when computing the confidence,
 * we have to do double-precision multiplies and divides with the scores.
 *
 * Revision 1.12  1994/12/15  02:45:52  ddavies
 * ddavies on Wed Dec 14 18:41:10 PST 1994
 *
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Fix one bug, make one improvement in the heuristic.  In almost-blank
 * images, we could have a divide-by-0 fault when computing the confidence.
 * The confidence was the average angle divided by the total of all the
 * scores divided by 32.  When that total was less than 32, we got a divide
 * check.
 * The improvement in the heuristic doesn't allow ip_detectSkewFine to step
 * over the angle found by ip_detectSkewRobot without taking a sample.  This
 * fixed a problem found by David Hirsch when deskewing a form that produced
 * a VERY sharp peak in the skew scores and a general hump at -1 degrees from
 * that.  ip_detectSkewRobot found the right angle, but ip_detectSkewFine
 * had been stepping around the thin peak and finding the broad, but low, hump.
 *
 * Revision 1.11  1994/10/14  20:26:29  ddavies
 * ddavies on Fri Oct 14 13:24:47 PDT 1994
 *
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Fix bug that could cause code to read off the image.  It carefully
 * spaced itself down from the top so it wouldn't read off the top.  It
 * carefully did the same thing going up from the bottom.  It didn't check
 * to see if the number of lines remaining in the image was positive.  It
 * does now.  Hope this fixes PerfectScan's seg fault.
 *
 * Revision 1.10  1994/10/04  01:31:17  ddavies
 * ddavies on Mon Oct 3 18:30:29 PDT 1994
 *
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Initialize __error__ so gcc doesn't complain.  This is a new, somewhat spurious,
 * warning with version 2.6.0.
 *
 * Revision 1.9  1994/08/01  21:06:21  ddavies
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Change "static const" to "static" to work around buggy gcc/Purify interaction.
 *
 * Revision 1.8  1994/04/19  22:21:22  ddavies
 * ddavies on Tue Apr 19 15:19:44 PDT 1994
 *
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Make the size of the angular step taken in ip_detectSkewFine a function of
 * the reduction factor.  Smaller "reduction factors" generally mean the code
 * should work harder to find the skew, so we take smaller steps.  This fixed
 * a problem found by the Paperware team - a form that wouldn't deskew.
 *
 * Revision 1.7  1994/04/03  02:05:16  ddavies
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Incorporate pointer types for use in SantaFe project.
 *
 * Revision 1.6  1994/03/02  03:51:52  ddavies
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Add IP_YIELD to give up control during skew detection.
 *
 * Revision 1.5  1994/02/04  20:17:15  dferrell
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:24:11  ddavies
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:29:56  ddavies
 * i_skwfxp.c -> /product/ipcore/ipshared/src/RCS/i_skwfxp.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


