/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tagrun.c: tags each pixel with the length of the run containing it.
 *
 *                  Int32     	i_tagHorizontalRuns()
 *                  Int32     	i_tagVerticalRuns()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

/* prototyes */
#include "shrcc.pub"
#include "shrcc.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tagrun.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");
/* During page segmentation it is useful to identify the white gutters
 * separating areas of text.  Every pixel in a gutter is in the middle of
 * both a horizontal run and a vertical run.  If the gutter is mainly
 * vertical, say separating two columns of text, the vertical run is
 * significantly longer than the horizontal run.  We would like to be
 * able to tag each pixel with the lengths of the horizontal and vertical
 * runs containing it so we can decide whether the gutters containing them
 * are horizontal or vertical.
 *
 *************************************************************************
 ***********                Horizontal Runs                  *************
 *************************************************************************
 *
 * i_tagHorizontalRuns tags each pixel with the length of the horizontal run
 * containing it.  How do we decide whether we're talking about white runs
 * or black runs?  I don't know, probably invert the incoming binary image.
 * This routine performs the same function as HorizGranulometry (the version
 * that's not commented out) in XIS's giputils.c.  That routine doesn't
 * conditionally invert the input image, so we won't either.
 *
 * The inputs to this routine are a binary and a gray image of the same
 * dimensions.  The binary image contains the pixels of interest.  When we
 * return, the gray image holds the output tag values.  This is a bit
 * different than HorizGranulometry in that that routine takes a single
 * gray image as both input and output.  On input, each gray pixel was
 * set to 0x00 or 0x01, depending on the value of the corresponding binary
 * pixel.  That operation was done by giputil's BitmapToImage routine.  That
 * routine also ignores black/white polarity (does no inversion).
 *
 * The gray image always has 8 bits/pixel.  What do we do with runs longer
 * than 255 pixels?  The XIS routine simply records the run length as
 * 255.  We'll follow the same convention.
 *
 * The XIS routine did two passes through each line of their gray image.
 * In the first one, each pixel was tagged with the number of contiguous
 * non-zero pixels to its left + 1.  In the second pass, the sum at the
 * right end of each run was copied back to all the pixels in the run.
 *
 * This is a very nice, clear algorithm.  We will try to eliminate
 * some steps.  First, we don't translate the binary image into a gray image
 * containing 1 interesting bit in each pixel before starting.  This would
 * cause us to write the image 1 extra time.  Second, we do each line in
 * one pass, not two.  The combination saves us two passes through the
 * gray image.
 *
 * The details of how a line is filled are described in mitagrun.c.  I will
 * not duplicate them here to avoid letting these get out of date as that
 * algorithm is modified.
 *
 *************************************************************************
 ***********                 Vertical Runs                   *************
 *************************************************************************
 *
 * Theoretically, vertical runs differ from horizontal runs in only
 * trivial ways.  In our world, however, the differences are significant.
 * First, every word read from the image contains several horizontal pixels.
 * This means its relatively easy to look right or left, but harder to
 * look up and down.  When we look up and down, we really ought to keep
 * track of several columns of pixels at once so as to avoid reading the
 * same word more than once.  This adds complexity.
 *
 * Another aspect that's sometimes ignored is cache efficiency.  In general,
 * the bits in a block of the cache correspond to a horizontal block of
 * bits in the image.  Moving through the image horizontally means we will
 * be using the cache efficiently.  Moving through the image vertically
 * means we will be cache faulting on every word (byte, pixel) read, then
 * ignoring most of the cache bytes brought in.  It can take a great
 * deal longer to take a cache fault on each word than to do an extra
 * write of the word to memory.  If saving some state avoids a cache fault,
 * it's probably a good idea.
 *
 * An issue we'll have in the future is image banding.  In almost everyone's
 * banding scheme, the bands are horizontal.  Accessing vertical bands is
 * hard.
 *
 * Image compression is normally done horizontally.  Accessing a vertical
 * column of pixels is hard in a compressed image, i.e. it takes a long time
 * or a lot of state.
 *
 * For all these reasons, we will violate a cardinal rule of ipshared code.
 * We will write the output gray image twice.  In the first pass, we record
 * the number of ON pixels in the current run above the pixel we're writing
 * + 1.  At the end of this pass, the lowest pixel in each run contains
 * the length of the run.  In the second pass, we copy this number to all
 * pixels in the run.
 *
 * So far, this is a description of XIS's VertGranularity.  The main
 * difference between that code and this is that we'll calculate
 * all the pixels in a scan line before moving down to the
 * next scan line instead of processing each column individually.  This
 * should greatly improve cache statistics.  We will also process 32 bits
 * of binary and gray image at a time instead of processing one gray pixel
 * at a time.
 *
 * In the first pass, we go from the top of the image to the bottom, calling
 * ip_calculateVerticalRunLengths for each line.  In the second pass, we go
 * through the image from bottom to top, calling ip_tagVerticalPixels for
 * each line.  These procedures are described in the comments to
 * mitagvrn.c.  There may or may not be SPARC assembly code - we'll see whether
 * double reads and writes help much on the SPARC.  The chances of there
 * being PC assembly code are even less - I'm not planning to do anything
 * that the C compiler shouldn't be able to do.
 */


/*********************************************************************
 *************************      Procedures     ***********************
 *********************************************************************/

/* PUBLIC */
/***********************************************************************
 *  i_tagHorizontalRuns()
 *	Input: 
 *	pB =	 ptr to binary source image (includes frame)
 *	bBpl =	 binary source image bytes/line
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	pG =	 ptr to gray dest image (includes frame)
 *	gBpl =	 dest gray image bytes/line
 *
 *	On return: 
 *		Every pixel in the gray image is tagged with the
 *		length of the horizontal run containing that pixel
 *		in the binary image.
 *
 *	Return code:
 *		ia_successful		good completion
 ***********************************************************************/
Int32   CDECL 
i_tagHorizontalRuns (
    IPState    *pIPState,
    UInt8      *pB,
    UInt32	bBpl,
    UInt32	w,
    UInt32	h,
    UInt8      *pG,
    UInt32	gBpl)
{

	/* point to the images inside the frames */
    PROLOGUE;
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pB += bBpl << cLogFrameBits;
    pB += cFrameBytes;
    pG += gBpl << cLogFrameBits;
    pG += 8 * cFrameBytes;

	/* do the real work 1 line at a time comparing
	 * *pS1 with scaleTable[*pS2]  (or p1 with (mult2/mult1 * p2))
	 */
    for (; h > 0; h--, pB += bBpl, pG += gBpl)
    {
	ip_tagHorizontalRunsLine (pIPState, pB, pG, w);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/***********************************************************************
 *  i_tagVerticalRuns()
 *	Input: 
 *	pB =	 ptr to binary source image (includes frame)
 *	bBpl =	 binary source image bytes/line
 *	w =	 width in pixels of all images (including frame)
 *	h =	 height of all images (including frame)
 *	pG =	 ptr to gray dest image (includes frame)
 *	gBpl =	 dest gray image bytes/line
 *
 *	On return: 
 *		Every pixel in the gray image is tagged with the
 *		length of the vertical run containing that pixel
 *		in the binary image.
 *
 *	Return code:
 *		ia_successful		good completion
 ***********************************************************************/
Int32   CDECL 
i_tagVerticalRuns (
    IPState    *pIPState,
    UInt8      *pB,
    UInt32	bBpl,
    UInt32	w,
    UInt32	h,
    UInt8      *pG,
    UInt32	gBpl)
{
    PROLOGUE;
    UInt32	line;

	/* point to the images inside the frames */
    w = w - 2 * cFrameBits;
    h = h - 2 * cFrameBits;
    pB += bBpl << cLogFrameBits;
    pB += cFrameBytes;
    pG += gBpl << cLogFrameBits;
    pG += 8 * cFrameBytes;

	/* Calculate the lengths of the vertical runs by calling
	 * ip_calculateVerticalRunLengths once per line.
	 */
    for (line = h; line > 0; line--, pB += bBpl, pG += gBpl)
    {
	ip_calculateVerticalRunLengths (pIPState, pB, pG, gBpl, w);
    }

	/* tag each pixel with the length of the run it's in by
	 * copying these lengths from the bottom pixel in each
	 * run.  Note that pB and pG start out pointing one line
	 * below the bottom of the image.  We want to start with
	 * the line that's just above the bottom line since this
	 * is the first one that could possibly change.
	 * Every pixel in the bottom line has either the
	 * length of the run it's in or 0, both of which are ok.
	 */
    pG -= 2*gBpl;
    pB -= 2*bBpl;
    for (line = h; line > 0; line--, pB -= bBpl, pG -= gBpl)
    {
	ip_tagVerticalPixels (pIPState, pB, bBpl, pG, gBpl, w);
    }

    NORMAL_RETURN;
}

/*

$Log: i_tagrun.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.7  1998/01/13 23:49:55  ddavies
i_tagrun.c -> /product/ipcore/ipshared/src/RCS/i_tagrun.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1995/12/13 02:46:18  ejaquith
i_tagrun.c -> /product/ipcore/ipshared/src/RCS/i_tagrun.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/01/11  22:30:55  dferrell
 * i_tagrun.c -> /product/ipcore/ipshared/src/RCS/i_tagrun.c,v
 *
 * Changed include of shrip.pub and shrip.prv to shrcc.pub and shrcc.prv,
 * respectively.
 *
 * Revision 1.4  1994/07/12  23:55:59  ejaquith
 * i_tagrun.c -> /product/ipcore/ipshared/src/RCS/i_tagrun.c,v
 *
 * removed #include "iadef.prv".
 *
 * Revision 1.3  1993/12/02  19:24:18  ddavies
 * i_tagrun.c -> /product/ipcore/ipshared/src/RCS/i_tagrun.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:44:46  ddavies
 * i_tagrun.c -> /product/ipcore/ipshared/src/RCS/i_tagrun.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


