/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tcdbug.c: Contains procedures used to debug text compression
 *
 *		Int32	i_createDictionaryPositions()
 *			Creates a position block containing an entry
 *			for each symbol in the dictionary.  Used to
 *			print out the dictionary.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#include "memory.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /*  prototyes */
#include "shrcmprs.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tcdbug.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");


/* The process we use to compress images by finding repeated symbols is
 * well documented (at least there's lots of documentation) in the
 * file i_tcompr.c.
 *
 * This file contains routines used to debug the process of text compression.
 * i_createDictionaryPositions creates a block of position info that,
 * when decoded, prints each symbol in the dictionary in an image.
 * The client supplies the image dimensions and the size of the box used
 * for each symbol.  If a symbol fits in the box, position data is put out
 * to place the lower right corner of the symbol in the lower right
 * corner of the symbol's assigned box on the page and the symbol ID
 * follows.
 *
 */

/* PUBLIC */
/**************************************************************************
 *   i_createDictionaryPositions()
 *
 *	This procedure produces a block of position data that, when
 *	decoded, causes the dictionary symbols to be printed in a grid
 *	in an image.  This is used to inspect the dictionary symbols
 *	for debugging purposes.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to structure containing
 *					symbol info
 *		firstSymbol:		symbol ID of symbol to be placed
 *					in the upper left box of the image
 *		imgWidth:		image width including frame
 *		imgHeight:		image height including frame
 *		boxWidth:		width of grid box in image
 *		boxHeight:		height of grid box in image
 *		pPositionBuf:		pointer to buffer that will contain
 *					position and symbol ID's.
 *		positionBufSize:	size of that buffer in bytes.
 *		pResultSize:		ptr to Int32 that will hold the
 *					number of bytes actually used in
 *					the position buffer
 *
 *	Action:
 *		Goes through the symbols in order of symbol ID, advancing
 *		one box position (left to right, top to bottom) for each
 *		symbol.  If a symbol fits in its box, puts out the x and y
 *		offset info required to place lower right corner of symbol
 *		in lower right corner of box followed by symbol's ID.
 *		Stops when either the symbols or boxes in one image are
 *		used up.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	position and symbol ID info didn't
 *					fit into the position buffer.
 *
 **************************************************************************/
Int32  CDECL 
i_createDictionaryPositions (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Int32	firstSymbol,
    Int32	imgWidth,
    Int32	imgHeight,
    Int32	boxWidth,
    Int32	boxHeight,
    UInt8      *pPositionBuf,
    Int32	positionBufSize,
    Int32     *pResultSize)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo;	/* ptr to compression info */
    TCSymbol   *pSymbol;
    Int32	posBufIndex;	/* index into the position buffer */
    Int32	curX, curY;	/* current x and y position */
    Int32	newX, newY;	/* position of next symbol on the page */
    

	/* make sure all the pointers are ok */
    if ((pTextCompInfoIn == NULL) || (pPositionBuf == NULL) ||
	(pResultSize == NULL) || (positionBufSize == 0))
	return ia_invalidParm;

	/* convert the pointer to the correct type */
    pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;

	/* is firstSymbol ok? */
    if ((firstSymbol < 0) || (firstSymbol >= pTextCompInfo->nextSymbolNum))
	return ia_invalidParm;

	/* We want the real image width and height after this, so subtract
	 * off the frames now.
	 */
    imgWidth -= 2*cFrameBits;
    imgHeight -= 2*cFrameBits;

	/* Start the position info with a magic number and version number.
	 */
    pPositionBuf[cTCPosMagicNum0] =
		(UInt8)(cTextCompPositionInfoMagicNum);
    pPositionBuf[cTCPosMagicNum1] = 
		(UInt8)(cTextCompPositionInfoMagicNum >> 8);
    pPositionBuf[cTCPosMagicNum2] = 
		(UInt8)(cTextCompPositionInfoMagicNum >> 16);
    pPositionBuf[cTCPosMagicNum3] = 
		(UInt8)(cTextCompPositionInfoMagicNum >> 24);

    pPositionBuf[cTCPosVer0] = (UInt8)(cTextCompPositionInfoVersionNum);
    pPositionBuf[cTCPosVer1] = (UInt8)(cTextCompPositionInfoVersionNum >> 8);

	/* Add image width and height, 4 bytes each */
    pPositionBuf[cTCPosWidth0] = (UInt8) imgWidth;
    pPositionBuf[cTCPosWidth1] = (UInt8) (imgWidth >> 8);

    pPositionBuf[cTCPosHeight0] = (UInt8) imgHeight;
    pPositionBuf[cTCPosHeight1] = (UInt8) (imgHeight >> 8);

	/* Say the stats are invalid and clear them */
    memset((void *)&(pPositionBuf[cTCPosStatsValid]), 0, cTCPosStatsLen);

	/* the first data bytes will be written after this stuff */
    posBufIndex = cTCPosHeaderVer2Len;

	/* Start at the top of the page */
    curX = 0;
    curY = -1;

	/* start at the bottom of the first box */
    newX = boxWidth-1;
    newY = boxHeight-1;

	/* Every optimistic, assume everything will turn out ok. */
    __error__ = ia_successful;

	/* if there were no symbols, the firstSymbol check should have
	 * failed, so we assume there are symbols.  Go find the first one.
	 */
    pSymbol = pTextCompInfo->p1stSymbol;

	/* If we somehow ran off the end of the list without finding the
	 * first symbol, complain.
	 */
    if (pSymbol == NULL)
	return ia_internal;

	/* Here's the main loop where we go through the symbols */
    for (;
	 pSymbol != NULL;
	 pSymbol = pSymbol->pNextNum)
    {
	if (((pSymbol->maxY - pSymbol->minY)/2 < boxHeight) &&
	    ((pSymbol->maxX - pSymbol->minX)/2 < boxWidth))
	{
		/* Actually put out the position info and symbol ID for
		 * this box.  First, do these bytes fit into the buffer?
		 * Since this is only a debugging routine, we'll make the
		 * crude assumption that the maximum number of bytes is
		 * required.  That's 1 for x, 2 for y, 2 more for the
		 * real x and 2 for the symbol ID = 7 bytes.
		 */
	    if (posBufIndex + 7 > positionBufSize)
	    {
		__error__ = ia_bufferExceeded;
		goto debugPosBufCleanup;
	    }

		/* do we need to change the y coord?
		 * In the new version of internal position info,
		 * we start the buffer with a Y coord relative to
		 * -1, there is no prelimary X coord of 0.
		 */
	    if (newY != curY)
	    {
		curX = 0;
		if ((newY - curY) > 127)
		{
		    pPositionBuf[posBufIndex] = (UInt8) ((newY - curY) | 0x80);
		    pPositionBuf[posBufIndex+1] = (UInt8) ((newY - curY) >> 7);
		    posBufIndex += 2;
		}
		else
		{
		    pPositionBuf[posBufIndex] = (UInt8) (newY - curY);
		    posBufIndex++;
		}
		curY = newY;
	    }
		/* now put out the x position.  Note that since this is
		 * the coord of the lower RIGHT corner of the box, it
		 * will never be on the left edge of the page, so we don't
		 * have to worry about the x = 0 case (which means there's
		 * another y coord coming).
		 */
	    if ((newX - curX) > 127)
	    {
		pPositionBuf[posBufIndex] = (UInt8) ((newX - curX) | 0x80);
		pPositionBuf[posBufIndex+1] = (UInt8) ((newX - curX) >> 7);
		posBufIndex += 2;
	    }
	    else
	    {
		pPositionBuf[posBufIndex] = (UInt8) (newX - curX);
		posBufIndex++;
	    }
	    curX = newX;

		/* Now put out the symbol ID */
	    if (pSymbol->symbolID > 127)
	    {
		pPositionBuf[posBufIndex] = (UInt8) (pSymbol->symbolID | 0x80);
		pPositionBuf[posBufIndex+1] = (UInt8) (pSymbol->symbolID >> 7);
		posBufIndex += 2;
	    }
	    else
	    {
		pPositionBuf[posBufIndex] = (UInt8) (pSymbol->symbolID);
		posBufIndex++;
	    }
	}
		/* having put out a position and symbol ID if necessary,
		 * update newX,newY
		 */
	newX += boxWidth;
	if (newX >= imgWidth)
	{
		/* Put out a deltaX of 0 marking the end of the line */
	    pPositionBuf[posBufIndex] = 0;
	    posBufIndex++;
	    newX = boxWidth;
	    newY += boxHeight;
	    if (newY >= imgHeight)
	    {
		    /* there are no more grid positions in the image,
		     * so we're done.
		     */
		break;
	    }
	}

    }

	/* That's it!  We've put out all the symbols we're going to.
	 * Set the number of bytes we actually used.  Also append
	 * two zeros for x offset of 0 (=> y offset coming up) and 
	 * y offset 0f 0 (=> end of position buf info)
	 */
    pPositionBuf[posBufIndex] = 0;
    pPositionBuf[posBufIndex+1] = 0;
    *pResultSize = posBufIndex + 2;

	/* return success or failure */
debugPosBufCleanup:
    NORMAL_RETURN;
}


/* It's often interesting to know the number of a particular symbol.
 * One way to do this is to go through the position info and find the
 * relation between coordinates and symbol IDs.  This routine turns
 * each coordinate in the position info into the format
 * yyyy: xxxx/ssss xxxx/ssss .... xxxx/ssss\n
 * where xxxx is a four digit x position, yyyy is the four digit y position
 * and ssss is the four digit symbol ID.
 * Like all other ipshared procs, this one reads a memory block and
 * fills a memory block.  The size of the output block can be estimated
 * by multiplying the total number of components on the page by about 12.
 */
/* PUBLIC */
/**************************************************************************
 *   i_decodePositionInfo()
 *
 *	This procedure converts a block of encoded position info into
 *	a block of ASCII text.  It is only used during debugging of
 *	symbol-based compression.  Each scan line containing a symbol
 *	is represented in the form:
 *	yyyy: xxxx/ssss xxxx/ssss ... xxxx/ssss\n
 *	where yyyy = 4 digit y coord, xxxx = 4 digit x coord and
 *	ssss = 4 digit symbol ID.
 *
 *	Input:
 *		pPositionBuf:		pointer to buffer that contains
 *					position and symbol ID's.
 *		pCharBuf:		ptr to buffer that will hold
 *					ASCII info.
 *		charBufSize:		size of character buffer
 *		pCharsWritten		ptr to number of chars that were
 *					written into the buffer by this
 *					routine.
 *
 *	Action:
 *		Goes through the position info, converting each change
 *		in y position to a new line of output and each x position -
 *		symbol ID pair to a pair of coordinates in the line.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL or
 *					position buffer wasn't full.
 *		ia_bufferExceeded	decoded position and symbol ID
 *					info didn't fit into the character
 *					buffer.
 *
 **************************************************************************/
Int32  CDECL 
i_decodePositionInfo (
    IPState    *pIPState,
    UInt8      *pPositionBuf,
    char       *pCharBuf,
    UInt32	charBufSize,
    UInt32     *pCharsWritten)
{
    PROLOGUE;
    Int32	curXPos, curYPos;	/* current x and y positions
					 * in the image */
    UInt32	charIndex;		/* index into character buf */
    Int32	temp;			/* general variable */
    Int32	coordsInLine;		/* number of coords written
					 * in current line */
    UInt32	version;		/* the version number of the
					 * position buf */
    UInt32	magicNum;		/* the magic number of the
					 * position buf */
#define	cMaxCoordsInLine	6	/* put in line break after
					 * writing this many coords
					 * in a single line */

    if ((pPositionBuf == NULL) || (pCharBuf == NULL))
	return ia_invalidParm;

	/* Skip over, and check, the magic number and version at the
	 * beginning of the position info.
	 */
    magicNum = ((UInt32) pPositionBuf[cTCPosMagicNum3] << 24) |
		((UInt32) pPositionBuf[cTCPosMagicNum2] << 16) |
		((UInt32) pPositionBuf[cTCPosMagicNum1] << 8) |
		((UInt32) pPositionBuf[cTCPosMagicNum0]);

    if (magicNum != cTextCompPositionInfoMagicNum)
	return ia_invalidParm;

    version = ((UInt32) pPositionBuf[cTCPosVer1] << 8) +
	      ((UInt32) pPositionBuf[cTCPosVer0]);

    if (version > cTextCompPositionInfoVersionNum)
	return ia_invalidParm;

	/* if that worked, either skip over just skip over the
	 * old or new version's header length.
	 */
    if (version == 1)
	pPositionBuf += cTCPosHeaderVer1Len;
    else	/* version == 2 */
	pPositionBuf += cTCPosHeaderVer2Len;

	/* point to the beginning of the character buffer */
    charIndex = 0;

	/* Start decoding the position info.
	 *
	 * I should say something about how to parse a position block.
	 * A position block contains a sequence of delta X's and symbol IDs.
	 * If a delta X == 0, there's a delta Y following instead of a
	 * symbol ID.  If that delta Y is also 0, this is the end of
	 * the position buffer.  If the delta Y is -1, the real delta Y
	 * is 0 (stay on the same line).  Following each delta Y (except
	 * the one that signifies the end of the file) is a delta X that
	 * gives the position relative to the left edge of the image of
	 * the first thing.  The delta X following a delta Y of -1 always
	 * brings us back to the same spot (it's used to say that two symbol
	 * bounding boxes share the same lower right corner), so we can
	 * just skip over them.
	 *
	 * In the original position buffer format, the first byte in a file
	 * was a deltaX, usually 0 and usually followed by a delta Y.
	 * If there was a symbol that ended on line 0, I thought the
	 * first deltaX would be non-zero.  BUT, what if the first symbol
	 * in the file is a single dot in the upper left corner?  The
	 * deltaX for that would be 0 and we'd be confused.
	 *
	 * In the modern format, the first byte in the file is a delta Y
	 * relative to -1.  If the position buffer is empty (blank page),
	 * this is 0.  Otherwise, it's non-zero.  Following every deltaY,
	 * including this first one, is a firstX relative to 0.  Zero
	 * is a legal value for this first X since a symbol may occupy only
	 * column 0.  We always read this firstX.  A line ends when
	 * a deltaX (follows a symbolID) is zero.
	 *
	 * Delta X, Delta Y and symbol IDs can be one or two bytes.
	 * Single byte quantities have a 0 in the MSbit.  Two-byte
	 * quantities have the LS byte first.
	 * Here's a review:
	 * The position info can be of several forms:
	 *	1 byte x offset; 1 or 2 byte symbol ID.
	 *	2 byte x offset; 1 or 2 byte symbol ID.
	 *	1 byte x offset (= 0), 1 byte y offset (= 0) => end of file.
	 *	1 byte x offset (= 0), 1 byte y position (dist to new line), 
	 *		1 or 2 byte x offset, 1 or 2 byte sym ID.
	 *	1 byte x offset (= 0), 2 byte y position (dist to new line), 
	 *		1 or 2 byte x offset, 1 or 2 byte sym ID.
	 *      1 byte x offset (= 0), 2 byte y position (= -1),
	 *		1 or 2 byte x offset (equal to original X position),
	 *		1 or 2 byte symbol ID.
	 *
	 * Note that there are no lines that contain only x offset of 0
	 * and a non-zero y offset.  That's because we can jump as far
	 * as necessary with any 2-byte y offset, so don't need any blank
	 * lines that contain only y distances.  We can also end any
	 * scan line with an "end of position block" indicator, so won't
	 * have any scan lines that contain only one of these.
	 */

	/* Read the first y position, x position and symbol.
	 * If I weren't such a principled person, I'd just jump
	 * into the middle of the loop.  Instead, we'll just
	 * copy the relevant portions of the loop up here..
	 */
    curYPos = -1;
    temp = *pPositionBuf++;
    if (temp & 0x80)
	temp = (temp & 0x7F) + ((*pPositionBuf++) << 7);
    curYPos += temp;		/* This is the first real Y position */

    sprintf(&pCharBuf[charIndex], "%4ld: ", curYPos);	/* start first line */
    coordsInLine = 0;
    charIndex += 6;

	/* Get the x coordinate for the first line */
    temp = *pPositionBuf++;
    if (temp & 0x80)
	curXPos = (temp & 0x7F) | ((*pPositionBuf++) << 7);
    else
	curXPos = temp;

    sprintf(&pCharBuf[charIndex], "%4ld/", curXPos);
    charIndex += 5;

	/* get the symbol ID */
    temp = *pPositionBuf++;		/* get first byte of symbol ID */

    if (temp & 0x80)			/* the next byte is needed too */
    {
	temp = (temp & 0x7F) | ((*pPositionBuf++) << 7);
    }
    sprintf(&pCharBuf[charIndex], "%4ld ", temp);
    charIndex += 5;
    coordsInLine = 1;			/* first coordinate written in
					 * first line.
					 */

    while (cTrue)
    {
	    /* if there's not enough space for the longest set of
	     * characters in the character buffer, quit.
	     */
	if (charIndex + 16 >= charBufSize)
	    return ia_bufferExceeded;

	temp = *pPositionBuf++;		/* get first byte of x offset */

	if (temp & 0x80)		/* the next byte is needed too */
	{
	    temp = (temp & 0x7F) | ((*pPositionBuf++) << 7);
	    if (temp & 0x4000)		/* the x offset is negative */
		curXPos += temp | 0xFFFF8000;	/* sign extend x */
	    else
		curXPos += temp;
	    sprintf(&pCharBuf[charIndex], "%4ld/", curXPos);
	    charIndex += 5;
	}
	else
	if (temp != 0)			/* it's a 1 byte x coord */
	{
	    curXPos += temp;
	    sprintf(&pCharBuf[charIndex], "%4ld/", curXPos);
	    charIndex += 5;
	}
	else				/* there's a y coordinate */
	{
	    temp = *pPositionBuf++;	/* get first byte of y offset */
    
	    if (temp & 0x80)		/* the next byte is needed too */
	    {
		if (*pPositionBuf & 0x80)   /* neg Y offset says that
					     * the bounding box is not
					     * supposed to change at all
					     */
		    pPositionBuf++;
		else
		{
		    curYPos += (temp & 0x7F) | ((*pPositionBuf++) << 7);
		    sprintf(&pCharBuf[charIndex], "\n%4ld: ", curYPos);
		    charIndex += 7;
		    coordsInLine = 0;
		}
	    }
	    else
	    if (temp != 0)
	    {
		curYPos += temp;
		sprintf(&pCharBuf[charIndex], "\n%4ld: ", curYPos);
		charIndex += 7;
		coordsInLine = 0;
	    }
	    else			/* We're at the end of the file */
	    {
		break;
	    }

		/* Get the x coordinate for the new line */
	    temp = *pPositionBuf++;
	    if (temp & 0x80)
		curXPos = (temp & 0x7F) | ((*pPositionBuf++) << 7);
	    else
		curXPos = temp;

	    sprintf(&pCharBuf[charIndex], "%4ld/", curXPos);
	    charIndex += 5;
	}

	    /* get the symbol ID */
	temp = *pPositionBuf++;		/* get first byte of symbol ID */

	if (temp & 0x80)		/* the next byte is needed too */
	{
	    temp = (temp & 0x7F) | ((*pPositionBuf++) << 7);
	}
	sprintf(&pCharBuf[charIndex], "%4ld ", temp);
	charIndex += 5;
	coordsInLine++;			/* one more coordinate written in
					 * current line.
					 */
	    /* if this line is full, go to a new line */
	if (coordsInLine >= cMaxCoordsInLine)
	{
	    sprintf(&pCharBuf[charIndex], "\n      ");
	    charIndex += 7;
	    coordsInLine = 0;
	}
    }

	/* end the last symbol */
    if (charIndex < charBufSize)
    {
	sprintf(&pCharBuf[charIndex], "\n");
	*pCharsWritten = charIndex + 1;
    }
    else
	return ia_bufferExceeded;	/* didn't quite fit */

	/* If we get here, it all worked, so return */
    NORMAL_RETURN;
}

/*

 $Log: i_tcdbug.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:13  jlin
 Imported sources

 Revision 1.9  1998/01/13 23:49:57  ddavies
 i_tcdbug.c -> /product/ipcore/ipshared/src/RCS/i_tcdbug.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.8  1996/11/08 22:03:07  ddavies
 ddavies on Fri Nov 8 14:01:27 PST 1996

 i_tcdbug.c -> /product/ipcore/ipshared/src/RCS/i_tcdbug.c,v

 Fix bug introduced during Tundra format conversion.  The internal format
  of a position block changed (starts with y position now) and
 i_decodePositionInfo had to be updated to match the new format.

 Revision 1.7  1996/08/23 02:53:04  ddavies
 ddavies on Thu Aug 22 19:52:06 PDT 1996

 i_tcdbug.c -> /product/ipcore/ipshared/src/RCS/i_tcdbug.c,v

 Modify parsing of internal position info so the block begins with a deltaY
 relative to -1 instead of a deltaX.

 Revision 1.6  1996/07/18 23:45:27  ddavies
 i_tcdbug.c -> /product/ipcore/ipshared/src/RCS/i_tcdbug.c,v

 Change all the references to the internal position buffers so they use
 the new cTCPos constants to reference items in the header.  Slightly
 improve the code used to find the magic number and version number.
 Reduce the number of bytes dedicated to the image width and height
 in this header from 4 to 2.
 Fuss with indenting.

 Revision 1.5  1996/07/04 02:56:08  ddavies
 ddavies on Wed Jul 3 19:54:41 PDT 1996

 i_tcdbug.c -> /product/ipcore/ipshared/src/RCS/i_tcdbug.c,v

 Change order of symbol linked list, so p1stSymbol points to symbol with
 lowest symbol ID instead of symbol with the highest symbol ID.
 This also means that pSymbol->pNextNum points to the symbol with the next
 higher symbolID instead of the next lower one.

 Revision 1.4  1995/12/13 02:46:21  ejaquith
 i_tcdbug.c -> /product/ipcore/ipshared/src/RCS/i_tcdbug.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/10/18  03:58:53  ddavies
 * i_tcdbug.c -> /product/ipcore/ipshared/src/RCS/i_tcdbug.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.2  1995/03/15  02:10:48  ddavies
 * ddavies on Tue Mar 14 18:10:31 PST 1995
 *
 * i_tcdbug.c -> /product/ipcore/ipshared/src/RCS/i_tcdbug.c,v
 *
 * Added i_decodePositionInfo.  This routine converts a block of encoded position
 * info into a block of ASCII text.  It is used during debugging of symbol-based
 * compression.  It prints out each y coord, then all the x coords and symbol
 * IDs on that scan line.
 *
 * Revision 1.1  1994/10/26  01:04:05  ddavies
 * Initial revision
 *

 */


