/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tcdec.c: Contains procedures used to turn a list of symbols and
 *		position data about the symbols into an image.  This
 *		code does the reverse of the code in i_tcenc.c.
 *
 *		Int32	i_getCompressedImageDimensions()
 *			Return the width and height of an image from
 *			the image's position info.
 *		Int32	i_convertSymbolsToImage()
 *			Turn each symbol into a bitmap and place the bitmap
 *			on the image.  If the bitmap is small enough,
 *			it is kept, otherwise the expansion is done
 *			directly onto the image.
 *
 *		Int32	ip_convertSymbolToBitmap()
 *			Produce the bitmap for a symbol.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#include "lmask.prv"		/* contains leftmask */
#include "memory.pub"
#include "imageref.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /*  prototyes */
#include "shrcmprs.pub"
#include "shrpixr.pub"
#include "shrrast.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tcdec.c,v $; $Revision: 1.2 $; $Date: 2001/05/08 06:43:36 $");

/* This code is used to turn a list of symbols and a block of position data
 * about the symbols into an image.  The concept of encoding images containing
 * text as symbols and positions of those symbols is discussed extensively
 * in the comments to i_tcompr.c.  You should read that to understand this.
 *
 * Broading speaking, the image information exists in one of three forms:
 * the original image; a set of positions plus internal data structures
 * specifying symbol shapes; and the same set of positions plus a compact
 * bytestream specifying symbol shapes.  We have routines to do each
 * stage of conversion:
 *
 *	Routine				Type of conversion
 *	i_convertImageToSymbols		image -> positions + internal data
 *					structures describing symbols and
 *					their shapes
 *
 *	i_consolidateSymbols		internal data structures describing
 *					symbols and their shapes ->
 *					similar internal data structures
 *					with a more refined set of shapes.
 *
 *	i_remapSymbolIDs		position info -> position info with
 *					old symbol IDs converted to new
 *					symbol IDs that were produced by
 *					i_consolidateSymbols.
 *
 *	i_serializeCompressionInfo	internal data structures describing
 *					symbols and their shapes ->
 *					compact bytestream describing
 *					symbols and their shapes.
 *
 *	i_deserializeCompressionInfo	compact bytestream describing
 *					symbols and their shapes ->
 *					internal data structures describing
 *					symbols and their shapes
 *
 *	i_convertSymbolsToImage		position info + internal data
 *					structures describing symbols and
 *					their shapes -> image
 *
 * Why all these little blocks?  The main reason is flexibility.  We can
 * use these to encode and decode images in a reasonable way.  We can
 * also use them to encode a bunch of images, write the symbols out to a
 * compact file, then use the same symbols to encode more images.  It's
 * also nice to be able to shortcut the compression -> decompression
 * cycle at several places to isolate bugs.  For example, we can
 * call i_convertSymbolsToImage immediately after calling
 * i_convertImageToSymbols, cutting out the symbol optimization and
 * serialization/deserialization steps.  With the addition of synthetic
 * position info, you can even make an image out of the set of symbols.
 *
 * The routines in this file carry out the conversion of a list of
 * symbols and position information into an image.  There are two
 * fundamental algorithms we could use to compose an image.
 *   1.	Go through the position info, remembering all the positions for
 *	every image; then go through the symbols.  For each symbol,
 *	create a bitmap of the symbol, blit it to each of its destinations
 *	and discard the bitmap.
 *   2. For each position, see if the symbol already has a bitmap.  If so,
 *	just blit the symbol to the position.  If not, create and save
 *	a bitmap for the symbol, then blit it into position.
 *
 * If we are only planning on using the symbol info for a single page, we
 * would probably choose the first option.  It saves memory by not caching
 * a bunch of bitmaps.  However, if we're going to decode a number of
 * pages, the second option looks good since we don't have to regenerate
 * the bitmaps for every page.  Assuming most documents are multi-page and
 * users with viewers will want to flip back and forth, we'll choose
 * the second option.
 *
 * Well, mostly.  Some bitmaps are huge - like the one for a box that goes
 * around the page.  If a bitmap is small enough, we'll cache it.  Otherwise,
 * we'll just expand it onto the image directly.
 *
 * There will be two types of expansion routines.  One is for images that
 * are small enough and only 1 word wide.  This is a very common case.
 * The second type of expansion routine is for images that are wider than
 * one word and/or are too big to keep around.
 *
 * How do we do the expansion?  Attached to each symbols is a list of runs.
 * The runs start with the rightmost run of the bottom line of the symbol.
 * They work from right to left along scan lines.  Scan lines go from
 * bottom to top.  We can tell when we get to the end of a scan line because
 * the right endpoint of the next run is greater than, or equal to, the
 * left endpoint of the current run minus 1.  This condition works because
 * symbols are connected components so our horizontal runs must have some
 * overlap.  A run on a previous scan line that violates this condition must
 * not be attached to the runs on the current scan line, hence is not part
 * of this connected component.
 *
 * The expansion process is fairly simple.  Let's start with the routine that
 * only deals with symbols that fit into a single word horizontally.
 *   1. Find a right run endpoint.
 *   2.	Generate a mask that has 1's in the bit specified by the right endpoint
 *	and all the bits to the left.
 *   3. Find the run's corresponding left endpoint.
 *   4.	Generate a mask that has 1's in the bit specified by the left endpoint
 *	and all the bits to the right.
 *   5.	AND the masks found in steps 2 and 4 together.  This represents the
 *	new run.  OR it into the accumulating word.
 *   6. If that was the last (leftmost) run of the scan line, write the
 *	accumulated word to memory, clear the accumulator and go back to
 *	step 1.
 *   7.	Repeat steps 1 through 6 until all the symbol's scan lines have been
 *	processed.
 *
 * We can improving on this procedure slightly by ORing the mask found in
 * step 2 into the accumulating word directly and ANDing the mask found
 * in step 4 with this word directly.
 *
 * The differences in the more general expansion task are that we must check
 * for word boundaries before processing every endpoint and we must OR the
 * accumulating word into the output instead of simply writing it.
 *
 * When a word boundary is found between runs, we just write the current
 * accumulator word and move on.  When a word boundary is in the middle
 * of a run, we do the same thing.  ORing the mask found in step 2 above
 * will cause the run to extend to the end of the word.  However, the
 * new accumulator word is set equal to the mask found in step 4 instead
 * of doing the AND step.
 *
 * What about the x coordinates of the runs?  They are stored relative to the
 * origin of the symbol, which is the middle of its bounding box.  To
 * calculate runs, we need endpoints that are relative to the left edge
 * of the bounding box.  We just add the symbol's minX entry to every
 * run endpoint.
 *
 */

/* PUBLIC */
/**************************************************************************
 *   i_getCompressedImageDimensions()
 *
 *	Return an image's width and height from the image's position info.
 *
 *	Input:
 *		pPositionBuf:		pointer to buffer that contains
 *					position and symbol ID's.
 *		pWidth:			pointer to integer that will contain
 *					image width.
 *		pHeight:		pointer to integer that will contain
 *					image height.
 *
 *	Action:
 *		Read image width and height from the position information.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL or
 *					the position info was invalid
 *
 **************************************************************************/
Int32  CDECL 
i_getCompressedImageDimensions (
    IPState    *pIPState,
    UInt8      *pPositionBuf,
    UInt32     *pWidth,
    UInt32     *pHeight)
{
    PROLOGUE;
    UInt32	version;	/* the version number of the
				 * position buf */
    UInt32	magicNum;	/* the magic number of the
				 * position buf */


	/* make sure the pointers are ok */
    if ((pPositionBuf == NULL) || (pWidth == NULL) ||
	(pHeight == NULL))
	return ia_invalidParm;

	/* make sure the position info starts with magic number and
	 * proper version */
    magicNum = ((UInt32) pPositionBuf[cTCPosMagicNum3] << 24) |
		((UInt32) pPositionBuf[cTCPosMagicNum2] << 16) |
		((UInt32) pPositionBuf[cTCPosMagicNum1] << 8) |
		((UInt32) pPositionBuf[cTCPosMagicNum0]);

    if (magicNum != cTextCompPositionInfoMagicNum)
	return ia_invalidParm;

    version = ((UInt32) pPositionBuf[cTCPosVer1] << 8) +
	      ((UInt32) pPositionBuf[cTCPosVer0]);

    if (version > cTextCompPositionInfoVersionNum)
	return ia_invalidParm;

	/* read the width and height.  The positions and lengths
	 * of the encoded values depend on the version.
	 */
    if (version == 1)
    {
	*pWidth = ((UInt32) pPositionBuf[cTCPosV1Width0] +
		    ((UInt32) pPositionBuf[cTCPosV1Width1] << 8) +
		    ((UInt32) pPositionBuf[cTCPosV1Width2] << 16) +
		    ((UInt32) pPositionBuf[cTCPosV1Width3] << 24));
    
	*pHeight = ((UInt32) pPositionBuf[cTCPosV1Height0] +
		    ((UInt32) pPositionBuf[cTCPosV1Height1] << 8) +
		    ((UInt32) pPositionBuf[cTCPosV1Height2] << 16) +
		    ((UInt32) pPositionBuf[cTCPosV1Height3] << 24));
    }
    else	/* this is version 2 */
    {
	*pWidth = ((UInt32) pPositionBuf[cTCPosWidth0] +
		    ((UInt32) pPositionBuf[cTCPosWidth1] << 8));
    
	*pHeight = ((UInt32) pPositionBuf[cTCPosHeight0] +
		    ((UInt32) pPositionBuf[cTCPosHeight1] << 8));
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_convertSymbolsToImage()
 *
 *	This procedure converts a set of positions, symbol ID's and a list
 *	of symbol shapes into a binary image.
 *
 *	Input:
 *		pImage:			Pointer to upper left corner of image
 *					INCLUDING frame.
 *		wS:			width of image including frame
 *		hS:			height of image including frame
 *		bpl:			The image's bytes/line
 *		pTextCompInfoIn:	pointer to structure containing
 *					information about symbols
 *		pPositionBuf:		pointer to buffer that contains
 *					position and symbol ID's.
 *
 *	Action:
 *		As each symbol position is encountered, the bitmap for
 *		the corresponding symbol is blitted into position.
 *		Symbol bitmaps that are small enough are retained in a
 *		cache.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers, including
 *					the pointer to the head of the list
 *					of symbols, was NULL
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *
 **************************************************************************/
Int32  CDECL 
i_convertSymbolsToImage (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	wS,
    Int32	hS,
    UInt32	bpl,
    void       *pTextCompInfoIn,
    UInt8      *pPositionBuf)
{
    PROLOGUE;
    TCCompressionInfoPtr  pTextCompInfo; /* ptr to compression info */
    TCSymbol   *pSymbol;		/* pointer to current symbol */
    TCSymbol  **pSymbolTable;		/* a table used to find a
					 * symbol from a symbol number.
					 */
    Int32	temp;			/* general variable */
    Int32	symHeight;		/* height of a symbol */
    Int32	symBpl;		/* symbol bitmap bytes/line */
    Int32	curXPos, curYPos;	/* current x and y positions
					 * in the image */
    Int32	bitmapBlockSize;	/* size of memory block
					 * that will hold all
					 * the cached bitmaps.
					 */
    UInt32	version;		/* the version number of the
					 * position buf */
    UInt32	magicNum;		/* the magic number of the
					 * position buf */
    

	/* make sure all the pointers are ok */
    if ((pImage == NULL) || (pTextCompInfoIn == NULL) ||
	(pPositionBuf == NULL))
	return ia_invalidParm;

	/* convert the pointer to the correct type */
    pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;

	/* If there aren't any symbols, complain */
    if (pTextCompInfo->nextSymbolNum == 0)
	return ia_invalidParm;

	/* check to see if the position info begins with the right
	 * magic number and version number.
	 */
    magicNum = ((UInt32) pPositionBuf[cTCPosMagicNum3] << 24) |
		((UInt32) pPositionBuf[cTCPosMagicNum2] << 16) |
		((UInt32) pPositionBuf[cTCPosMagicNum1] << 8) |
		((UInt32) pPositionBuf[cTCPosMagicNum0]);

    if (magicNum != cTextCompPositionInfoMagicNum)
	return ia_invalidParm;

    version = ((UInt32) pPositionBuf[cTCPosVer1] << 8) +
	      ((UInt32) pPositionBuf[cTCPosVer0]);

    if (version > cTextCompPositionInfoVersionNum)
	return ia_invalidParm;

	/* if that worked, skip over the version 1 or version2 header.
	 */
    if (version == 1)
	pPositionBuf += cTCPosHeaderVer1Len;
    else	/* version == 2 */
	pPositionBuf += cTCPosHeaderVer2Len;

	/* remove the frames from the image */
    pImage += (bpl << cLogFrameBits) + cFrameBytes;

	/* See if there's already a symbol table, i.e. a table used
	 * to associate symbol IDs with the symbol data structures.
	 * If so, use it as is. If not, allocate one and save it
	 * in the TCCompressionInfo structure.
	 */
    if ((pTextCompInfo->pSymTable == NULL) ||
	(pTextCompInfo->pSymTable->numPtrs != pTextCompInfo->nextSymbolNum))
    {	/* allocate and initialize a new table.  Delete the old one
	 * if it exists.
	 */
	__error__ = ip_createTCSymbolTable (pIPState, pTextCompInfo);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
    }
    pSymbolTable = &(pTextCompInfo->pSymTable->table[0]);

	/* The table is already filled with pointers to the symbols.
	 * However, we do need to count the number of bytes required
	 * to hold all the bitmaps that we'll cache, so we'll go
	 * through all the symbols again.  Note that the symbol
	 * with the lowest number appears at the head of the list.
	 * Also note that we leave room at the start of the bitmap for
	 * a PointerFixup block.  There aren't any pointers in the
	 * bitmap, but we need to chain the bitmap onto the other structures
	 * that have pointers to the bitmap.  This chain is used when
	 * the bitmap moves.
	 */
    bitmapBlockSize = sizeof(PointerFixup);
    for (pSymbol = pTextCompInfo->p1stSymbol;
	 pSymbol != NULL;
	 pSymbol = pSymbol->pNextNum)
    {
	symBpl = ((((pSymbol->maxX - pSymbol->minX)/2 + 1) + 31)/32) * 4;
	symHeight = (pSymbol->maxY - pSymbol->minY)/2 + 1;
	symHeight *= symBpl;		/* we really want the total bytes */
	    /* if the cached bitmap is small enough, remember the
	     * symbol's bytes/line and offset into the block.  We'll
	     * put the bitmap into the block once the block is
	     * allocated.  Don't fuss with the pointers if
	     * we already have a set of cached bitmaps.
	     */
	if ((symHeight < cMaxCachedBitmapSize) &&
	    (pTextCompInfo->pBitmapBlock == NULL))
	{
	    pSymbol->bytesPerLine = symBpl;
	    pSymbol->pBitmap = (UInt8 *) bitmapBlockSize;
	    bitmapBlockSize += symHeight;
	}
    }

	/* make bitmapBlockSize a little bigger because we were
	 * referencing off the end for some bizarre reason.
	 */
    bitmapBlockSize += 8;

	/* If we don't already have a set of cached bitmaps, allocate
	 * a block big enough to hold all of them.
	 * We use the client's memory allocator since this block will
	 * last a long time.
	 */
    if (pTextCompInfo->pBitmapBlock == NULL)
    {
	pTextCompInfo->pBitmapBlock =
			pTextCompInfo->ipcAllocator(pIPState,
						 bitmapBlockSize, cTrue);
	if (pTextCompInfo->pBitmapBlock == NULL)
	{
	    __error__ = ia_nomem;
	    goto tcSymToImgCleanup;
	}
	else
	{
		/* clear the bitmap so we don't have to carefully write
		 * every word with white stuff when expanding the symbols.
		 * We do this first because it also clears the PointerFixup
		 * structure at the top of the bitmap.  We will initialize
		 * this structure next.
		 */
	    memset((void *)pTextCompInfo->pBitmapBlock, 0, bitmapBlockSize);
    
	    pTextCompInfo->bitmapBlockSize = bitmapBlockSize;
	    __error__ = ip_initPointerFixup (pIPState, (PointerFixup *)
					     pTextCompInfo->pBitmapBlock,
					   bitmapBlockSize,
					   &(pTextCompInfo->fixup),
					   NULL); /* no pointer fixup routine */
	    if (__error__ != ia_successful)
	    {
		(pTextCompInfo->freer)(pIPState, pTextCompInfo->pBitmapBlock,
				       bitmapBlockSize);
		goto tcSymToImgCleanup;
	    }
	}

	    /* Decompress all the selected symbol bitmaps into the cache.
	     * This is the lazy way to do it, but it's simple.
	     */
	for (pSymbol = pTextCompInfo->p1stSymbol;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextNum)
	{
	    if (pSymbol->bytesPerLine != 0)	/* this symbol's bitmap
						 * should be cached.
						 */
	    {
		    /* point to the bitmap's location in the block of bitmaps */
		pSymbol->pBitmap = (UInt8 *) (pTextCompInfo->pBitmapBlock +
					       (UInt32) pSymbol->pBitmap);
		    /* put the bitmap there */
		__error__ = ip_symbolToBitmap (pIPState,
					     pSymbol->pBitmap,	/* pImage */
					     pSymbol->bytesPerLine, /* bpl */
					     0,		  	/* x offset */
					     0,			/* y offset */
					     pSymbol);	  /* ptr to symbol */
		if (__error__ != ia_successful)
		{	/* free the bitmap block.  A bunch of pointers may
			 * now be pointing into space, but this is the only
			 * routine that uses these pointers and this routine
			 * fixes up the pointers before using them, so
			 * it should be ok.
			 */
		    pTextCompInfo->freer(pIPState, pTextCompInfo->pBitmapBlock,
					 pTextCompInfo->bitmapBlockSize);
		    pTextCompInfo->pBitmapBlock = NULL;
		    pTextCompInfo->bitmapBlockSize = 0;
		    goto tcSymToImgCleanup;
		}
	    }
	}
    }	/* end of making a new block of cached bitmaps */

	/* Go through the position info, blitting symbols as we go.
	 * The position info can be of several forms:
	 *	1 byte x offset; 1 or 2 byte symbol ID.
	 *	2 byte x offset; 1 or 2 byte symbol ID.
	 *	1 byte x offset (= 0), 1 byte y position, 1 or 2 byte sym ID.
	 *	1 byte x offset (= 0), 2 byte y position, 1 or 2 byte sym ID.
	 *	1 byte x offset (= 0), 1 byte y offset (= 0) => end of file.
	 * Bit 7 of all 1-byte quantities is 0, Bit 7 of all 2 byte
	 * quantities is 1 - that's how we tell the difference.
	 * The x and y positions we find are the positions of the lower
	 * right corner of the character's bounding box.  We have to
	 * adjust by the bounding box width and height to get the
	 * coordinates used for rasterOp.
	 * The x coordinate may be negative.  If this happens, it
	 * is in the two-byte format.
	 * NEW VERSION: In the old code, the first thing in the
	 * buffer was a deltaX, which was usually 0 followed by
	 * a delta Y relative to 0.  If there was actually a symbol
	 * in line 0, the deltaX was non-zero, I thought.  But,
	 * what if the first symbol occupies only the upper left
	 * dot of the image?  Then the first deltaX would be 0 and
	 * we'd be confused.
	 * In the new version, the first thing in the position info
	 * is a deltaY relative to -1.  If the position buffer is
	 * empty, this deltaY is zero, otherwise it's non-zero.
	 * Following EVERY deltaY, including the first one, is a
	 * firstX.  This firstX is occasionally 0, which means the
	 * first symbol is completely contained in column 0.  That's
	 * fine.  Subsequent symbols are followed by deltaX coords.
	 * If one of these is 0, there's a deltaY next.
	 */

    curYPos = -1;
    __error__ = ia_successful;		/* hope it stays this way */
    curXPos = 0;
	/* pretend to have read an X position of 0 so we'll start
	 * by reading the first deltaY coord.
	 */
    temp = 0;

    while (cTrue)
    {
	if (temp & 0x80)		/* the next byte is needed too */
	{
	    temp = (temp & 0x7F) | ((*pPositionBuf++) << 7);
	    if (temp & 0x4000)		/* the x offset is negative */
		curXPos += temp | 0xFFFF8000;	/* sign extend x */
	    else
		curXPos += temp;
	}
	else
	if (temp != 0)			/* it's a 1 byte x coord */
	{
	    curXPos += temp;
	}
	else				/* there's a y coordinate */
	{
	    temp = *pPositionBuf++;	/* get first byte of y offset */
    
	    if (temp & 0x80)		/* the next byte is needed too */
	    {
		if (*pPositionBuf & 0x80)	/* neg Y offset says that
						 * the bounding box is not
						 * supposed to change at all
						 */
		    pPositionBuf++;
		else
		    curYPos += (temp & 0x7F) | ((*pPositionBuf++) << 7);
	    }
	    else
	    if (temp != 0)
	    {
		curYPos += temp;
	    }
	    else			/* We're at the end of the file */
	    {
		break;
	    }

		/* Get the x coordinate for the new line */
	    temp = *pPositionBuf++;
	    if (temp & 0x80)
		curXPos = (temp & 0x7F) | ((*pPositionBuf++) << 7);
	    else
		curXPos = temp;
	}

	    /* get the symbol ID */
	temp = *pPositionBuf++;		/* get first byte of symbol ID */

	if (temp & 0x80)		/* the next byte is needed too */
	{
	    temp = (temp & 0x7F) | ((*pPositionBuf++) << 7);
	}

	    /* point to the symbol in question */
	pSymbol = pSymbolTable[temp];
	temp = (pSymbol->maxX - pSymbol->minX)/2 + 1;	   /* symbol width */
	symHeight = (pSymbol->maxY - pSymbol->minY)/2 + 1; /* symbol height */

	    /* if we've got a cached bitmap, blit it into position.
	     * Otherwise, expand it in the right position */
	if (pSymbol->pBitmap != NULL)
	{
	    i_rasterOp (pIPState, pImage,	/* dest ptr */
			curXPos - temp + 1,	/* dx */
			curYPos - symHeight + 1, /* dy */
			temp,			/* dw */
			symHeight,		/* dh */
			1,			/* dd */
			bpl,			/* dLinebytes */
			PIX_SRC | PIX_DST,	/* op */
			pSymbol->pBitmap,	/* src ptr */
			0, 0,			/* sx, sy */
			1,			/* sd */
			pSymbol->bytesPerLine);	/* sLineBytes */
	}
	else
	{
	    __error__ = ip_symbolToBitmap (pIPState, pImage,
					 bpl,
					 curXPos - temp + 1,
					 curYPos - symHeight + 1,
					 pSymbol);
	    if (__error__ != ia_successful)
		goto tcSymToImgCleanup;
	}

	temp = *pPositionBuf++;		/* get first byte of next x offset */
    }

	/* the new image has been created.  However, since the exemplar
	 * symbols don't necessarily quite line up with the original symbols,
	 * little pieces of exemplar might have spilled over into the
	 * frames.  Clean this up by clearing all the frames just in case.
	 */
	/* Put the frame back on for i_clearAllFrames */
    pImage -= (bpl << cLogFrameBits) + cFrameBytes;
    __error__ = i_clearAllFrames (pIPState, pImage, wS, hS, bpl, 1);

tcSymToImgCleanup:

    NORMAL_RETURN;

}


/* PRIVATE */
/**************************************************************************
 *   ip_symbolToBitmap()
 *
 *	This procedure converts a list of runs into a picture of a
 *	character in a bitmap.
 *
 *	Input:
 *		pImage:		Ptr to destination bitmap
 *		bpl:		bitmap's bytes/line
 *		x:		x offset into bitmap of upper left corner
 *				of symbol bounding box.
 *		x:		y offset into bitmap of upper left corner
 *				of symbol bounding box.
 *		pSymbol:	Ptr to symbol
 *
 *	Action:
 *		The runs of the symbol are expanded into a bitmap.
 *		the bits are ORed into the image at the offset given.
 *
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32 CDECL 
ip_symbolToBitmap (
    IPState    *pIPState,
    UInt8      *pImage,
    UInt32	bpl,
    Int32	x,
    Int32	y,
    TCSymbol   *pSymbol)
{
    PROLOGUE;
    UInt32	oWord;			/* contains image words */
    Int32	runLeft, curOffset;	/* x coordinate of right end of
					 * run and current x coord */
    TCScanLineRun *pRun, *pPrevRun;	/* pointers to the symbol's runs */
    /* do not declare this array "const" to workaround a buggy interaction
     * between gcc and Purify.
     */
    static UInt32   rightMask[32] =
       {0xFFFFFFFF, 0x7FFFFFFF, 0x3FFFFFFF, 0x1FFFFFFF,
	0x0FFFFFFF, 0x07FFFFFF, 0x03FFFFFF, 0x01FFFFFF,
	0x00FFFFFF, 0x007FFFFF, 0x003FFFFF, 0x001FFFFF,
	0x000FFFFF, 0x0007FFFF, 0x0003FFFF, 0x0001FFFF,
	0x0000FFFF, 0x00007FFF, 0x00003FFF, 0x00001FFF,
	0x00000FFF, 0x000007FF, 0x000003FF, 0x000001FF,
	0x000000FF, 0x0000007F, 0x0000003F, 0x0000001F,
	0x0000000F, 0x00000007, 0x00000003, 0x00000001 };

	/* We need to move pImage down to the start of the image scan
	 * line containing the bottom of the image.  That's because
	 * the runs appear from bottom to top.  We do this in two
	 * steps: move from the top of the image to the top of the
	 * symbol, then move to the bottom of the symbol.
	 * HOWEVER, we do the real movement after we test bpl.
	 * If bpl is equal to 4 (the image fits in a single word),
	 * we can eliminate a multiply.  Since this is the common
	 * case, it's worth the trouble.
	 */

	/* There are two cases - one in which the image is one word wide
	 * and one where it's wider.  In the one-word case, we
	 * don't have to check for word boundaries.  This case covers
	 * the vast majority of symbols scanned at reasonable
	 * resolutions.
	 * Each case has an outer loop and an inner loop.  The
	 * outer loop increments once per scan line.  The inner
	 * loop increments once per run.
	 */
    if ((bpl == 4) && (x == 0))			/* do the common case */
    {
	    /* make pImage point to the line at the bottom of
	     * the symbol.  The leftmost bit in the word
	     * pointed to by pImage has x coordinate 0.  We
	     * find run coordinates relative to this.
	     */
	pImage += (y + (pSymbol->maxY - pSymbol->minY)/2) * 4;
	curOffset = -(pSymbol->minX);		/* gets symbolWidth/2
						 * NOTE: pSymbol->minX equals
						 * the symbol width because
						 * all coords have an extra
						 * bit on the LS end */
	pPrevRun = pSymbol->pRunTail;
	oWord = leftMask[(pPrevRun->runRight + curOffset)/2] &
			rightMask[(pPrevRun->runLeft + curOffset)/2];
	pRun = pPrevRun->pNext;
	while (cTrue)				/* loop over scan lines */
	{
	    while ((pRun != NULL) &&
		   (pRun->runRight < pPrevRun->runLeft - 2))
	    {
		oWord |= leftMask[(pRun->runRight + curOffset)/2];
		oWord &= rightMask[(pRun->runLeft + curOffset)/2];
		pPrevRun = pRun;
		pRun = pRun->pNext;
	    }

	    ULONG_ACCESS(pImage) = oWord;	/* we assume that no one is
						 * expanding a symbol into such
						 * a small bitmap that already
						 * contains something else.
						 */
	    pImage -= 4;			/* point to line above */

	    if (pRun == NULL)
		break;
	    else
	    {
		oWord = leftMask[(pRun->runRight + curOffset)/2] &
				rightMask[(pRun->runLeft + curOffset)/2];
		pPrevRun = pRun;
		pRun = pRun->pNext;
	    }

	}
    }
    else	/* the image is wider, so we have to watch
		 * out for word boundaries and pay attention
		 * to the x coordinates.
		 */
    {
	pImage += (y + (pSymbol->maxY - pSymbol->minY)/2) * bpl;
	pRun = pSymbol->pRunTail;
	pPrevRun = NULL;
	x = 2*x - pSymbol->minX;		/* get offset of left edge
						 * of the symbol.  Remember
						 * that all x coords should
						 * have an extra bit of
						 * precision, so insert one
						 * here. */
	    /* Loop through the runs.  The outer loop does one
	     * scan line at a time, the inner loop deals with
	     * individual runs.
	     * REMEMBER: the x (and y) coordinates have an extra bit
	     * of precision.  That's why we're always dividing by 2.
	     * We keep the extra bit of precision around so that
	     * all the bits end up in the right place.
	     */
	while (cTrue)				/* loop over scan lines */
	{
	    while ((pRun != NULL) &&
		   ((pPrevRun == NULL) ||
		    (pRun->runRight < pPrevRun->runLeft - 2)))
	    {
		curOffset = x + pRun->runRight;
		runLeft = x + pRun->runLeft;
		oWord = leftMask[(curOffset/2) % 32];
		curOffset = curOffset & ~0x3F;		/* point to beginning
							 * of the word from
							 * here on */
	
		if (runLeft < curOffset)		/* fill in middle words								 * with 1's */
		{
		    ULONG_ACCESS(&pImage[curOffset / 16]) =
			ULONG_ACCESS(&pImage[curOffset / 16]) | oWord;
		    oWord = 0xFFFFFFFF;
		    curOffset -= 64;
			/* fill in the intermediate words, stopping before
			 * we get to the word containing the left endpoint
			 * The test works because curOffset always points
			 * to the beginning of a word.
			 */
		    while (runLeft < curOffset)
		    {
			ULONG_ACCESS(&pImage[curOffset / 16]) = oWord;
			curOffset -= 64;
		    }
		}
		oWord &= rightMask[(runLeft/2) % 32];
		ULONG_ACCESS(&pImage[curOffset / 16]) =
		    ULONG_ACCESS(&pImage[curOffset / 16]) | oWord;

		pPrevRun = pRun;
		pRun = pRun->pNext;
	    }
	    pImage -= bpl;				/* point to next line */
	    pPrevRun = NULL;

	    if (pRun == NULL)
		break;
	}
    }	/* end of case where bpl != 4 */

    NORMAL_RETURN;
}

/*

 $Log: i_tcdec.c,v $
 Revision 1.2  2001/05/08 06:43:36  jlin
 change ->allocator to ->ipcAllocator

 * Revision 1.1.1.1  1999/09/20  18:18:13  jlin
 * Imported sources
 *
 Revision 1.11  1998/01/13 23:49:58  ddavies
 i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.10  1996/08/23 02:53:26  ddavies
 i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v

 Modify parsing of internal position info so the block begins with a deltaY
 relative to -1 instead of a deltaX.

 Revision 1.9  1996/07/18 23:49:09  ddavies
 i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v

 Change all the references to the internal position buffers so they use
 the new cTCPos constants to reference items in the header.  Slightly
 improve the code used to find the magic number and version number.
 Reduce the number of bytes dedicated to the image width and height
 in this header from 4 to 2.
 Fuss with indenting.
 Use the new symbol table created for the TCCompressionInfo to associated
 symbol IDs with symbols instead of creating such a table.  Don't free
 the local table since it doesn't exist anymore.

 Revision 1.8  1996/07/04 02:56:33  ddavies
 i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v

 Change order of symbol linked list, so p1stSymbol points to symbol with
 lowest symbol ID instead of symbol with the highest symbol ID.
 This also means that pSymbol->pNextNum points to the symbol with the next
 higher symbolID instead of the next lower one.

 Revision 1.7  1995/12/13 02:46:23  ejaquith
 i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/07/10  23:43:52  ddavies
 * i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v
 *
 * Add a PointerFixup structure at the top of the memory block used fo
 * cached bitmaps.  Initialize it after the cache bitmap block is allocated.
 *
 * Revision 1.5  1994/11/23  05:53:28  ddavies
 * ddavies on Tue Nov 22 21:51:20 PST 1994
 *
 * i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v
 *
 * Several changes to i_convertSymbolsToImages.  Add image width and height
 * parmeters so that the frames can be cleared after the image is decompressed.
 * This is needed because the exemplar symbols don't exactly line up with the
 * instance symbols, so may hang off into the frames.  Also make the
 * bitmapBlockSize a little bigger so that i_rasterOpFull doesn't read off
 * the end (why does it do than anyway?).
 *
 * Revision 1.4  1994/10/17  22:56:25  ddavies
 * ddavies on Mon Oct 17 15:56:13 PDT 1994
 *
 * i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v
 *
 * Added (void *) cast to first arg of memset.
 *
 * Revision 1.3  1994/10/14  01:18:07  ddavies
 * ddavies on Thu Oct 13 18:17:47 PDT 1994
 *
 * i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v
 *
 * Add <string.h> to get prototype for memset.
 *
 * Revision 1.2  1994/09/08  00:30:24  ddavies
 * ddavies on Wed Sep 7 17:29:54 PDT 1994
 *
 * i_tcdec.c -> /product/ipcore/ipshared/src/RCS/i_tcdec.c,v
 *
 * Changes to support the new x coordinates.  These have an extra bit of
 * precision in the LSB.
 *
 * Revision 1.1  1994/08/03  19:18:46  ddavies
 * Initial revision
 *

 */


