/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tcdser.c: Contains procedures to turn a compact bytestream into
 *		a list of symbols.
 *
 *		Int32	i_deserializeCompressionInfo()
 *			Convert a compact bytestream into a
 *			list of symbols and their runs.
 *
 *		Int32	ip_decodeLengthCode()
 *			Decode black or write code for a horizontal
 *			distance.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#define _NEED_TOPLEVEL_CODES_
#define _NEED_HORIZONTAL_CODES_
#include "mmrdec.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /*  prototyes */
#include "shrcmprs.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tcdser.c,v $; $Revision: 1.2 $; $Date: 2001/05/08 06:43:36 $");

/* PUBLIC */
/**************************************************************************
 *   i_deserializeCompressionInfo()
 *
 *	This procedure turns a serial bytestream produced by
 *	i_serializeCompressionInfo into the obscure internal representation
 *	of the information used to describe connected components.
 *
 *	The serial bytestream is not affected by this process,
 *	i.e. you can try it again if the procedure runs out of
 *	memory.  HOWEVER, you cannot combine serialized bytestreams
 *	into a single block of shape information by calling this
 *	procedure on each one.  This procedure returns with ia_invalidParm
 *	if called when pTextCompInfoIn already represents some set of
 *	symbols.
 *
 *	The memory used by the various linked lists is allocated here
 *	using the allocator contained in pTextCompInfo.
 *	If the procedure fails, all allocated memory is freed.
 *
 *
 *	Input:
 *		pTextCompInfoIn:	ptr to text compression info.
 *		pShapeInfoBuf:		ptr to buffer that holds the serialized
 *					data.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one or both of the input pointers
 *					was NULL or pTextCompInfo was not
 *					set up correctly.
 *		ia_nomem		ran out of memory
 *		ia_decodeFailure	encountered an error while attempting
 *					to decode the bytestream, probably
 *					an illegal code.
 *
 **************************************************************************/
Int32  CDECL 
i_deserializeCompressionInfo (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pShapeInfoBuf)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo;
    TCScanLineRun *pRefLeftEndRun, *pRefRightEndRun;
    TCScanLineRun *pRunHead, *pRunTail;
    TCScanLineRun *pTempRun;
    Int32	symWidth, symHeight;
    Int32	symInstanceCount;
    Int32	temp, w;	/* general temp variable */
    
    UInt32	accum;		/* used to accumulate bytestream. */
    Int32	vacantBit;	/* number of 1st vacant bit in accum,
				 * 0 at LS end. */
    UInt8	codeIndex;	/* used to index code tables */
    UInt8	codeValByte;	/* used to hold values returned from
				 * 1 byte code tables. */
    Int32	lastX;		/* last X position */
    UInt32	lastLine;	/* tags for line on which lastX is
				 * valid */
    UInt32	state;		/* cLeftEndpoint => ready to generate
				 * a left endpoint.
				 * cRightEndpoint => ready to generate
				 * a right endpoint.
				 */
    TCInstance	instance;	/* used to pass info to
				 * ip_exactSymbolIsComplete. */
    

	/* do this to avoid casting all the time */
    pTextCompInfo = (TCCompressionInfoPtr) pTextCompInfoIn;

	/* if there's no input buffer or text info, complain.
	 * If there are already symbols in the text info, complain
	 */
    if ((pTextCompInfo == NULL) || (pShapeInfoBuf == NULL) ||
	(pTextCompInfo->ipcAllocator == NULL) || (pTextCompInfo->freer == NULL) ||
	(pTextCompInfo->p1stSymbol != NULL))
	return ia_invalidParm;

	/* There's no need to destroy an existing symbol table.  If
	 * there were any existing symbols, the check above would
	 * have failed.
	 */

	/* initialize lots if text info variables just for cleanliness */
    for (w = 0; w <= cMaxSymIndexDim; w++)
	for (temp = 0; temp <= cMaxSymIndexDim; temp++)
	    (pTextCompInfo->pSymIndex->array)[w][temp] = NULL;
    pTextCompInfo->pBigSymbols = NULL;
    pTextCompInfo->imageWidth = 0;
    pTextCompInfo->imageHeight = 0;
    pTextCompInfo->pPositionBuf = NULL;
    pTextCompInfo->positionBufSize = 0;
    pTextCompInfo->resultSize = 0;		/* used no bytes so far */
    pTextCompInfo->lastY = 0;
    pTextCompInfo->classifyOnly = cFalse;
    pTextCompInfo->useResidual = cTrue;
    pTextCompInfo->p1stSymbol = NULL;
    pTextCompInfo->pLastSymbol = NULL;
    pTextCompInfo->totalSymbols = 0;
    pTextCompInfo->nextSymbolNum = 0;

	/* if we haven't already allocated starter blocks for the
	 * various data structures, do so now.
	 */
    if (pTextCompInfo->p1stScanLineBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stScanLineBlock),
					 sizeof(TCScanLineRun),
					 cTCScanLineRunBlockSize,
					 cTrue,		/* adjust ptrs after
							 * garbage collection */
					 &(pTextCompInfo->fixup),
					 &ip_fixupScanLinePtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

    if (pTextCompInfo->p1stGroupBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stGroupBlock),
					 sizeof(TCGroupRun),
					 cTCGroupRunBlockSize,
					 cTrue,
					 &(pTextCompInfo->fixup),
					 &ip_fixupTCGroupRunPtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

    if (pTextCompInfo->p1stSymbolBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stSymbolBlock),
					 sizeof(TCSymbol),
					 cTCSymbolBlockSize,
					 cTrue,
					 &(pTextCompInfo->fixup),
					 &ip_fixupTCSymbolPtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

	/* initialize the instance structure we use to pass info to
	 * ip_exactSymbolIsComplete.  A default of 0 is safe for the stuff
	 * we really haven't computed yet.
	 */
    instance.minWidth = 0;
    instance.maxWidth = 0;
    instance.minHeight = 0;
    instance.maxHeight = 0;
    instance.x = 0;
    instance.y = 0;
    instance.w = 0;
    instance.h = 0;
    instance.leftProfile = 0;
    instance.rightProfile = 0;
    instance.strokeWidth = 0;
    instance.mass = 0;
    instance.numRuns = 0;
    instance.checkHVolume = cFalse;
    instance.checkVVolume = cFalse;
    instance.hVolume = 0;
    instance.vVolume = 0;
    instance.maxDimension = 0;
    instance.slant = 0;
    instance.hSAsymmetry = 0;
    instance.vSAsymmetry = 0;
    instance.pCurSymbol = NULL;
    instance.pHorizRuns = NULL;
    instance.pDupHorizRuns = NULL;
    instance.pVertRuns = NULL;
    instance.pDupVertRuns = NULL;

	/* The process of decoding symbols has two parts.  The big
	 * part is turning the MMR encoded runs in the bytestream
	 * into TCScanLineRuns.  The smaller part is turning a bunch
	 * of runs into a real symbol with all the symbol's metrics
	 * and groups.  In one sense, the process of extracting runs
	 * from the bytestream is much like the process of extracting
	 * run from an image in that they both extract the runs of
	 * one connected component from some icky data structure.
	 * When encoding connected components, we try to match the
	 * new symbol against existing symbols.  If it doesn't match,
	 * we make up a new symbol.  Here, we don't bother with the
	 * matching process, we know that the new symbol is unique.
	 * So, this routine looks sort of like a combination of the
	 * symbol classification code in i_tcenc.c and an inverted version
	 * of the encoding code in i_tcdser.c.  Since most of the work
	 * is in decoding, we'll repeat the comments (slightly modified
	 * so they work for decoding instead of encoding) from
	 * i_tcser.c.
	 */

	/* This routine decodes the symbols and their shapes from a
	 * bytestream.  As explained in i_tcompr.c, we encode the
	 * symbols in order of symbol ID.
	 *
	 * Before we describe the content of the input, let's say
	 * a word about format.  In order to make translation between
	 * little endian and big endian machines painless, the compressed
	 * info is put out as a stream of bytes.  So long as the
	 * order of the bytes in the stream is maintained between platforms,
	 * information will be preserved.  When putting out multi-byte
	 * quantities, we put out the least significant byte first.
	 *
	 * It's useful to describe encoding and decoding at the same time.
	 * At least it helps me.  So what happens first?  We put out / take in
	 * a bunch of boilerplate at the beginning of the compression
	 * info.  This includes a magic number and a version number.
	 * This is analogous to what`s done in the position info.
	 *
	 * Next, we go to the end of the list of symbols.  The list is
	 * kept in order of reverse symbol ID (larger symbol IDs at the
	 * front).  That means we find the symbol with the lowest ID.  As
	 * we go, we'll work our way to the beginning of the list (highest
	 * symbol ID).
	 *
	 * We begin each symbol with the symbol's width and height.
	 *
	 * For each symbol, we record:
	 *	symbolWidth	2 bytes
	 *	symbolHeight	2 bytes
	 *	usage count	2 bytes
	 *	pseudo-MMR encoding of the symbol shape.
	 *
	 * What are our deviations from real MMR?
	 *   1.	Since the runs are listed from right to left, bottom to
	 *	top, our encoding is also given in this order.
	 *   2. To make encoding and decoding a little faster, we reverse
	 *	the bit order of the codes.  That is, the LS bit of each
	 *	code is the "first" one.  The small benefit comes from
	 *	the fact that machines do reads and writes and table indexing
	 *	operations using the LS byte of a word, not the MS byte.
	 *	Therefore, we want the action to take place at the LS end
	 *	of a word.  For this to be true, we have to read the bits
	 *	from the LS end towards the MS end.
	 *   3.	Since we are encoding connected components, we know that
	 *	there are no blank lines.  Since the connected components
	 *	are almost always small, we know there are very few runs,
	 *	usually 1 or 2 per scan line.
	 *	In standard MMR, there's a fake run off the end of each
	 *	scan line.  One marks the end of the scan line by issuing
	 *	a vertical code saying that you line up with the fake run.
	 *	In normal MMR, this extra bit is an insignificant part
	 *	of the data.  In our tiny images, it's significant.
	 *	Therefore, we will eliminate it.  So, what do we do instead?
	 *	a. We maintain pointers to right and left endpoints on
	 *	   the previous line.  All the runs are in a single list,
	 *	   so the pointers can cross from the previous line to the
	 *	   current line unexpectedly.  When this happens, we
	 *	   ignore the endpoints they point to.
	 *	b. To make the previous line endpoints active again, the
	 *	   current line (the one being decoded) must cross a line
	 *	   boundary.  It can do this by 1) issuing a vertical
	 *	   match code to a run on the next line, or 2) issuing
	 *	   a horizontal code that wraps around to the next line.
	 *	   A PASS code never causes us to move to the next line
	 *	   since a PASS code implies that there's a run on the current
	 *	   line beyond the next run on the previous line.  Therefore,
	 *	   we can`t jump beyond the last run in the current line.
	 *   4.	The end of the symbol is signalled by the "extension" code
	 *	instead of the "end of facsimile block" code.  The extension
	 *	code is shorter and we won't use any extensions.
	 *
	 * The encoding loop is a state machine, much like the decoding
	 * loop.  In fact, we'll try to make them as similar as possible
	 * to reduce the chance of error.  We go around the loop once for
	 * each complete code put out or taken in.  I'd better explain
	 * MMR coding in order for that to make sense.  At least my
	 * interpretation of MMR coding.
	 *
	 * The standard draws several little pictures.  The "Reference Line"
	 * is the line above the one being processed.  The "Coding Line" is the
	 * image line being encoded/decoded.  At each stage, we have a starting
	 * position, lastX, in the coding line.  That is obvious, but
	 * important.  The starting position in the coding line is
	 * immediately to the left (since we encode right to left) of
	 * the last pixel that was encoded.  An important point is that both
	 * the encoder and the decoder can identify the starting position
	 * since it reflects only things that have happened in the past.
	 * A similar value, lastLine holds the tags of the line from which
	 * lastX was taken.  In the first first line, lastX,lastLine
	 * points to the lower right corner of the image's bounding box.
	 *
	 * Both the encoder and decoder can see the reference line, so
	 * let's start with that.  Starting at lastX in the reference
	 * line, the encoder looks to the left for the first pixel that a)
	 * has a different value than the pixel to its right and b) has the
	 * opposite value than the pixel at lastX in the codeing line.
	 * Even though we haven't encoded lastX, we can deduce its
	 * polarity from the sequence of coding state.
	 *
	 * For example, say we're about to encode a right endpoint.
	 * Then we look to the left for the first right endpoint.
	 * If we're going to encode a left endpoint, we look to
	 * the left in the reference line for the nearest left endpoint.
	 * The CCITT standard calls this first endpoint on the reference
	 * line b1.  The location of the next pixel to the left of b1 of
	 * opposite color from b1 is called b2.
	 *
	 * It's in the coding line that encoding and decoding become
	 * different.  In encoding, we look to the left for the first
	 * endpoint of proper left/right polarity.  Call this
	 * location a1.  We then look to the left of a1 for the next pixel
	 * of opposite color.  Call the location of this pixel a2.
	 * Several things can happen at this point.
	 *   1. There may be no pixels on the coding line under the 
	 *	reference line run (b1,b2).  This is indicated by b2
	 *	(the end of the run in the reference line) being to the
	 *	right of a1 (the beginning of the run in the coding line).
	 *	In fact, we require a difference of at least one pixel
	 *	between b2 and a1 to decide the run has really
	 *	ended.  In this case, we issue a PASS code.  This tells
	 *	both the encoder and decoder to skip the black or white run
	 *	on the reference line by advancing lastX to b2.  Note that
	 *	insisting on the 1-pixel buffer between b2 and a1 means
	 *	that the color of the pixel under the new lastX will be
	 *	the same as the color of the pixel under old lastX.  If
	 *	we were looking for a left endpoint originally, we'll
	 *	still be looking for one.  Ditto for right endpoints.
	 *   2. If we didn't issue a PASS code, the next point on the
	 *	reference line might be within distance 3 of the next
	 *	point on the coding line.  That is, | b1 - a1 | <= 3.
	 *	If this is true, we use a vertical code to specify the 
	 *	relative distance between them and move lastX up to a1's 
	 *	position.  Since we're only passing one endpoint, we swap 
	 *	the polarity of the next endpoint to be encoded or decoded.
	 *	Group 4 encoding is based on the notion that this is
	 *	the common case.
	 *   3. If neither of those work, we encode a horizontal code
	 *	followed by two distances.  One goes to the next endpoint, 
	 *	the other goes to the endpoint after that.  Why two distances?
	 *	Because we often want two insert two endpoints to get to
	 *	the next vertical match.  A brand new run or a fork inserts
	 *	two new endpoints.
	 *
	 * Here's a point that confused me.  Why do we always look to the
	 * left from lastX?  Say we have the following case:
	 *
	 *			X
	 *		      X X X
	 *
	 * It would be nice to encode the second line using 6 vertical
	 * codes (2 for each symbol), all relative to the bit on the first
	 * line.  However, the standard says that the pixel on the left
	 * will be encoded relative to some other pixel on the reference
	 * line because the obvious reference pixel isn't to the left of
	 * the coding pixel.  Why did they do that?
	 *
	 * Take the following example:
	 *		X	X
	 *		     X  X  X
	 *
	 * Since we didn't change the coding line, we should get the same
	 * code for the coding line, right?  It would be +2, +2, 0, 0, -2, -2.
	 * Now say we have the following image:
	 *		X	X
	 *	      X		X X
	 *
	 * The encoding would be the SAME: +2, +2, 0, 0, -2, -2.  Ooops.
	 * That's why they say you always have to look left (they say right
	 * since they're encoding left to right) for the reference run.
	 *
	 * Ok, but why does the standard say *strictly* to the left?
	 * Well, you have to decide on some convention to distinguish
	 *		X	X
	 *	      X		X X
	 * from
	 *		X	X
	 *	      X	X	  X
	 *
	 * and they decided that having a meaningful endpoint immediately
	 * above was much less likely than having it farther along the line.
	 *
	 * Well, here's where we make another little change from the
	 * standard.  What happens if the reference and coding line runs
	 * both touch the right end of the image?  If we don't start looking
	 * from a position off the edge of the image, we won't see the
	 * proper reference line endpoint because it won't be strictly to the
	 * left.  If we do start witn an imaginary pixel off to the right,
	 * we have to remember not to include this imaginary pixel in the
	 * first code.  Well, we could do all that but it would be pain.
	 * Instead, we look up and to the right for reference line pixels
	 * when encoding.  That means that when we must look up and to the
	 * the left when decoding also.
	 *
	 *			*	*	*
	 *
	 * There are several layers of encoding.  At the top, is a layer that
	 * can find or generate one of the following codes (listed in bit
	 * reversed order relative to the standard documents - remember that
	 * one should read these from right to left, i.e. <--<--<--< ):
	 *
	 *	Code	Meaning
	 *	      1	Next coding line endpoint (left or right) is
	 *		immediately below the next reference line endpoint.
	 *	    110	Next coding line endpoint is 1 bit to the right of
	 *		next reference line endpoint.
	 *	    010	Next coding line endpoint is 1 bit to the left of
	 *		next reference line the endpoint.
	 *	    100	Horizontal distances to next two endpoints in
	 *		coding line relative to previous point in *coding*
	 *		line follow (see below)
	 *	   1000	Pass code - move lastX to end of next run in
	 *		reference line.
	 *	 110000	Next coding line endpoint is 2 bits to the right of
	 *		next reference line endpoint.
	 *	 010000	Next coding line endpoint is 2 bits to the left of
	 *		next reference line endpoint.
	 *	1100000	Next coding line endpoint is 3 bits to the right of
	 *		next reference line endpoint.
	 *	0100000	Next coding line endpoint is 3 bits to the left of
	 *		next reference line endpoint.
	 *	1000000	End of string.  This is the normal G4 extension code
	 *		but it's a lot shorter than the normal G4 "end of
	 *		block" code, so we use it for end of block.
	 *		The last MMR byte is padded with 0's so the next
	 *		symbol's width begins on a byte boundary.
	 *
	 * Each pass through the encoding loop generates one of these.
	 * Similarly, each pass through the decoding loop decodes one of
	 * these.  If the horizontal code is generated, we also generate /
	 * decode distances to the next two endpoints before returning
	 * to the top of the loop.
	 *
	 * We can be in either of two states at the top of the loop,
	 * preparing to encode/decode a right endpoint or preparing to
	 * encode/decode a left endpoint.  Unfortunately, we need two 
	 * reference run pointers.  One points to the run having the next
	 * reference line left endpoint and one points to the run having
	 * the next reference line right endpoint.  Don't these always
	 * belong to the same run?  No.  Look at this example:
	 *
	 *	SSSSSSSSSSSSS		  RRRRRRRRRRRRRR
	 *		      DD	CCCCC	BBBBB  AA
	 *
	 * Once again, remember we're going from right to left.
	 * The reference right endpoint for run A is the right endpoint of
	 * R.  The reference right endpoint for runs B and C is the right
	 * endpoint of S.  The reference left endpoints of A and B is the
	 * left endpoint of R.  The reference left endpoint of runs C and D
	 * is the left endpoint of S.  So, for runs B and C, the reference
	 * left endpoint is to the right of the reference right endpoint.
	 * For runs A and D, it's the other way around.  The best way I've
	 * found to represent this is to have independent endpoint pointers.
	 * These pointers are set at the top of the loop given the current
	 * x coordinate, lastX.
	 * 
	 * If we get a horizontal code when preparing to generate a
	 * right endpoint, the next coordinates are for the right
	 * endpoint and the next left endpoint (specifying an entire run).
	 * If we're ready to generate a left endpoint, the two coordinates
	 * finish off the left end of the current run and start the right
	 * end of the next one.
	 *
	 * As stated above, if a horizontal code pushes us beyond the left
	 * edge of the bitmap, we wrap around to the right end on the
	 * next line.  The bitmap width equals the width of the symbol exactly.
	 * Since we're only translating to and from runlength data structures,
	 * we don't have to worry about padding symbol width to a multiple
	 * of 32 bits.
	 *
	 * The standard G3/G4 horizontal distance codes are a little weird.
	 * There is one code used to specify blocks of white space
	 * (used to cover the distance from a left endpoint of one run
	 * back to the right endpoint of the run to the left) and another
	 * code used to specify blocks of black pixels.  Why two codes?
	 * It's probably because black runs tend to be shorter than white runs.
	 * The black code has fewer bits/distance at the small distances.
	 * The white code has fewer bits/distance at the larger distances.
	 *
	 * Each of the codes has individual entries for all distances from
	 * 0 through 63.  They also have individual entries for each
	 * multiple of 64 up through 1728.  There is one common code for
	 * spanning pixels in blocks of 1792, 1856 (units of 64), ..., 2560.
	 * For distances greater than 2560, one concatenates 2560 codes
	 * together.  Therefore, a single distances has the form:
	 *	(<some number of 2560 codes>)
	 *	(<black or white code for 64 through 2496 in blocks of 64))
	 *	<terminating black or white code specifying a distance
	 *	 from 0 through 63>
	 *
	 * Only the last code is required.  If the block length is an
	 * exact multiple of 64, the terminating code specifies a distance
	 * of 0.  If the block is an exact multiple of 2560, the middle
	 * code is left out and the last code specifies 0.
	 * If n * 2560 < distance < n * 2560 + 64, the middle code is also
	 * left out and the terminating code specifies the offset from
	 * the last multiple of 2560.
	 *
	 * As stated above, the black and white codes are different.
	 * The white codes are nicely arranged so that using an 8 bit
	 * lookup table will return the following info:
	 *	1. code value (distance represented)
	 *	2. code length in bits (amount by which we should advance
	 *	   to get to the next code).
	 * If the code value is greater than 63, it is a non-terminating
	 * code.  If the code length is 9 (the length of the longest
	 * white code), it is a 64x code and the final value should be
	 * found by adding the ninth bit * 64 to the code value.  If the
	 * answer is 1664, Then the answer is really 1728 (weird - is 1664
	 * so common that it needs a special code?).  If the code length
	 * is > 9, this is a 64x code shared with the black codes.  Use
	 * the black lookup method as described next.
	 *
	 * The black codes have a nice feature.  Either the first four bits
	 * are 0000, or they're not.  If they are not 0000, we can use a
	 * lookup table indexed by 6 bits that encodes the 9 interesting
	 * entries. These represent the distances 1 through 9.  Each entry has
	 *	1. code value
	 *	2. code length in bits.
	 *
	 * If the first four bits are 0000, we get rid of them and use an 8-bit
	 * lookup table on the remaining bits.  The output of this
	 * table is the same as the output of the white table described above:
	 *	1. code value (distance represented)
	 *	2. code length in bits (amount by which we should advance
	 *	   to get to the next code after discarding the first 4 bits).
	 * There are 13-bit black codes, but they are nicely arranged.  So,
	 * if the code length after discarding the first 4 bits is 9, this
	 * is a 64x code and the final value should be found by adding the
	 * ninth bit * 64 to the code value.
	 *
	 * To encode a distance, we put out one 2560 code for each multiple
	 * of 2560 in the distance, then the correct 64x code (if any) for
	 * any non-zero multiples of 64 remaining, then the terminating code
	 * for the remainder.  Only the terminating code is required.  To
	 * decode a number, we read in codes, adding their values, until
	 * a terminating code (one whose length is less than 64) is reached.
	 *
	 *			*	*	*
	 *
	 * So much for reading and writing numbers in horizontal codes.  How
	 * does the encoding/decoding loop really work?  The encoding loop
	 * turns runs on the current and preceeding lines into a sequence of
	 * codes.  The decoding loop does the opposite.  The encoding loop
	 * tests its current position with respect to runs on the preceeding
	 * line to decode what codes to generate.  The decoding loop uses the
	 * tokens read to branch to branch to blocks of code to generate runs.
	 *
	 * Both loops have two main states: generating/expecting a left
	 * endpoint and generating/expecting a right endpoint.  These govern
	 * interpretation of input runs during encoding and generation of
	 * output runs during decoding.  They also govern the generation/
	 * interpretation of horizontal codes - either white then black if
	 * looking for a left endpoint or black then white if looking for a
	 * right endpoint.
	 *
	 * All the runs for a symbol are in a single linked list.  When either
	 * encoding and decoding, the pointer to the runs on the reference
	 * line can be advanced to point to the coding line.  In fact, it
	 * happens all the time.  How do we know when the reference line
	 * pointer is a) invalid, and b) valid again?
	 *
	 * When encoding, we have a set of well-formed runs.  The pClient
	 * pointers of these runs contain, not pointers, but alternating
	 * even/odd line tags.  All runs on one line may contain even tags
	 * and all runs on the next line contain odd tags.  We decide that
	 * the reference line run pointer is valid when the tags of the
	 * run it points to are not the same as the tags of the current run.
	 * This automatically takes care of making the reference line pointers
	 * invalid when they get too close and valid again when the coding
	 * line moves ahead.
	 *
	 * When decoding, things are a little more complicated.  We
	 * need to maintain the current line tags in an independent
	 * variable instead of looking at the line tags of the current
	 * run.  That's because when we get done generating the left
	 * endpoint of a run, there is no current run.  The next run may
	 * be on the same line or on a new line, we don't know yet.
	 *
	 * The previous line run pointer still becomes invalid when its
	 * line tags become the same as the current line tags.  It becomes
	 * valid again when the current line tags flip.  This happens
	 * when a) a horizontal code goes off the left edge of the image
	 * and wraps around; b) the reference line was invalid and a
	 * PASS code is found (we must be on a new line if we're now
	 * crossing over a run on the previous line);  c) a vertical code
	 * is found (we must be on a new line if we're directly over a
	 * run on the previous line).
	 *
	 * Isn't this really annoying?  Every time I'm about to tell you
	 * how it all works, I go back and give more background.  Well, we`re
	 * getting close now.
	 *
	 * We start encoding by issuing the horizontal codes for the first
	 * run and letting both reference line endpoint pointers point to it.
	 * This happens prior to entering the loop to set things up.  Decoding
	 * does a similar thing - setting up the first run from the first
	 * horizontal code and setting the reference line pointers to the new
	 * run.
	 *
	 * The loop terminates when we run out of runs, so we don't have
	 * to check for that explicitly inside.  Also note that the reference
	 * run pointers are behind the coding run pointer, so they can't
	 * be NULL either.
	 *
	 * Let's look a what happens when we want to encode a new right
	 * endpoint.
	 *   1. Skip right endpoints on the reference line until we either
	 *	find one that's to the left of the current position or have
	 *	wandered into the coding line.
	 *   2. If the reference right endpoint is valid (really points
	 *	to a run on the reference line) and the left endpoint of
	 *	that run is to the right of the next right endpoint on the
	 *	coding line + 1, issue a PASS code, set the current
	 *	horizontal position equal to the position of the reference
	 *	line's left endpoint - 1 (so we're back in white space)
	 *	and go back to the top of the loop with the state still 
	 *	set to right endpoint.
	 *   3. If a PASS code wasn't issued and the next coding line
	 *	right endpoint is within 3 pixels of the reference line
	 *	right endpoint, issue the corresponding vertical code, 
	 *	update the current coding line position to be the position 
	 *	of the coding line right endpoint, change the state to 
	 *	left endpoint and go back to the top of the loop.
	 *   4. If all else fails, issue a horizontal code specifying the
	 *	number of white pixels to the next right endpoint,
	 *	wrapping around the image if necessary, and the
	 *	number of pixels in the black run.  Set the lastX
	 *	to the position of the new left endpoint - 1 (so we're
	 *	back in white space) and go back to the top of the loop 
	 *	with the state still set to right endpoint.
	 *
	 * When generating left endpoints,
	 *   1. Skip left endpoints on the reference line until we either
	 *	find one that's to the left of the current position or have
	 *	wandered into the coding line.
	 *   2. If the reference line ptr is valid (really points to a run
	 *	on the reference line) and the right endpoint of the
	 *	*next* run on the reference line is to the right of the
	 *	next left endpoint on the coding line, issue a PASS code.
	 *	Also set the current x position to the position of the
	 *	new right endpoint (so we're still pointing to a black
	 *	pixel) and go back to the top of the loop with
	 *	the state still set to left endpoint.
	 *   3. If a PASS code wasn't issued and the next reference line
	 *	left endpoint is within 3 pixels of the next coding line
	 *	left endpoint, issue the corresponding vertical code, 
	 *	change the state to right endpoint and go back to the 
	 *	top of the loop.
	 *   4. If all else fails, issue a horizontal code specifying the
	 *	number of pixels to the next endpoint (a run of width 1
	 *	generates a 1, not a 0) and the number of white pixels
	 *	before the next right endpoint, wrapping around the image
	 *	if necessary.  Set the current horizontal position to
	 *	the position of the new right endpoint and go back to
	 *	the top of the loop with the state still set to left endpoint.
	 *
	 * When decoding, step 1 is the same in both cases.  We decode
	 * the next token in the bitstream to see which decision was made
	 * by the encoder.  We mirror the decision in the decoder, creating
	 * runs as necessary.
	 *
	 * There's a subtle feature here.  We encode the length of
	 * black and white runs, not the distance required to get from one
	 * endpoint to the next.  For example,  There are many 1-pixel
	 * black runs.  The left and right endpoints of these runs are
	 * equal.  However, we encode 1, not 0 for the black length.
	 * Originally, I was planning to encoding differences in endpoint
	 * coordinates instead of lengths, but the codes for 0 and 1 are
	 * too long.
	 *
	 * Let's say something about the mechanics of composing and decoding
	 * bitstreams.  In both cases, we keep the bitstream in a 32-bit
	 * register.  When composing, we keep a pointer to the first vacant
	 * bit relative to the LS end of the word.  New codes are shifted
	 * left by this much and ORed into the accumulator - to the right
	 * of the older codes.  The pointer to the first vacant bit is
	 * incremented by the length of the code.  Each time we insert
	 * a new code, we look at the new value of the pointer.  While
	 * it's greater than or equal to 8, we write out the bottom byte,
	 * shift the word right by 8 and subtract 8 from the pointer.
	 *
	 * At the bottom of the encoding loop, we flush the accumulator by
	 * shifting and writing until the vacant bit pointer is less than
	 * or equal to 0.  Since the logical right shifts insert 0's at
	 * the top of the accumulator, the last byte will be padded with
	 * 0's.  Then we write another 4 bytes so that we won't run off
	 * the end of the buffer when decoding.
	 *
	 * When decoding the bitstream, we also keep a pointer to the first
	 * vacant bit in the accumulator.  Before doing a decode, we look at
	 * the vacant bit pointer.  While it's less than or equal to 24, we
	 * shift the next input byte left to align it with the next vacant
	 * bit and OR it into the accumulator.  To decode words, we mask the
	 * lower bits of the accumulator and use the masked bits as an
	 * index into the tables described above.
	 *
	 * The decoded runs contain the distance from the right edge of
	 * the symbol, then distances from right to left.  This means
	 * that lastX starts at symWidth has has each distance subtracted
	 * from it as we go.  We know when we've gone off the left side
	 * of the image when lastX goes negative.
	 *
	 * And one more thing.  Geez.  The endpoint coordinates inside the
	 * TCScanLineRun structs range from -w+1 to w-1 in steps of 2.
	 * This allows an extra bit of precision so we can tolerate a 
	 * little change in origin due to noise.  We'd don't want to 
	 * encode these 2x coordinates because the bigger numbers take
	 * a lot of extra bits.  Furthermore, we'd like it if the
	 * bitmap coordintes went from 0 to w-1 so we could easily tell
	 * when the coordinates wrapped around the left side (i.e. went
	 * negative).  So, we want a new set of coordinates.  We could
	 * either change them all in place and change them back, or
	 * make up a new set of runs for each symbol.  We'll change them all,
	 * then change them back.  This avoids the need for allocating
	 * extra memory at a time when we're probably trying to get
	 * rid of the memory occupied by the internal coding data structures.
	 * It should also optimize cache usage.
	 *
	 * When decoding, we generate the raw runs, then process them to
	 * produce 2x coordinates just as is done during classification.
	 * At the same time, we calculate the mass (number of
	 * black pixels) just as the classification procedure does.
	 * Finally, we calculate a set of vertical runs and the slant
	 * of the symbol.  When we're done, we call ip_exactSymbolIsComplete
	 * with all of the info.  It inserts the new symbol into the
	 * list of symbols.  We set the insertOnly arg of
	 * ip_exactSymbolIsComplete to cTrue so it doesn't do anything else.
	 *
	 * BUG FIX: When we've seen the extension code that terminates a
	 * symbol, we have the next few bytes in the accumulator.  These
	 * are the next symbol's width and height.  We need to back up so
	 * these are read again properly when decoding the next symbol.
	 * Since we know that the extension code is padded out to byte
	 * boundaries, we just decrement pShapeInfoBuf by (vacantBit/8)
	 * after vacantBit has been adjusted to account for the extension
	 * code.  This works because vacantBit equals the number of bits
	 * in the accumulator.  We want to re-read every complete byte
	 * that was put into the accumulator "by mistake".
	 */

	/* Check the boilerplate on the shape info */
    temp = pShapeInfoBuf[0] |
	   (pShapeInfoBuf[1] << 8) |
	   (pShapeInfoBuf[2] << 16) |
	   (pShapeInfoBuf[3] << 24);

    if (temp != cTextCompShapeInfoMagicNum)
    {
	__error__ = ia_decodeFailure;
	goto tcDeserializeError;
    }

    temp = pShapeInfoBuf[4] |
	   (pShapeInfoBuf[5] << 8);

    if (temp != cTextCompShapeInfoVersionNum)
    {
	__error__ = ia_decodeFailure;
	goto tcDeserializeError;
    }

	/* skip the boilerplate */
    pShapeInfoBuf += 6;


	/* *************  MAIN SYMBOL DECODE LOOP  ***************** */
    while (cTrue)
    {
	    /* Read the width of this symbol.  If it's zero, we quit the loop.
	     */
	symWidth = pShapeInfoBuf[0] |
		    (pShapeInfoBuf[1] << 8);

	if (symWidth == 0)
	    break;

	    /* since we mostly use symWidth-1 from here, just decrement
	     * symWidth here.
	     */
	symWidth--;

	    /* There's are real symbol here, read its height and
	     * instanceCount too */
	symHeight = pShapeInfoBuf[2] |
		    (pShapeInfoBuf[3] << 8);
	symInstanceCount = pShapeInfoBuf[4] |
			   (pShapeInfoBuf[5] << 8);
	pShapeInfoBuf += 6;


	    /* initialize position vars and run ptr for this symbol.
	     * Then allocate the first run for this symbol.  The
	     * run-decoding loop assumes that there's already at
	     * least one run in existance.
	     *
	     * Here's a slightly strange feature.  When we found the
	     * runs for a symbol in an image, we found them in order
	     * from left to right, top to bottom.  They were put
	     * on the linked list with the last run at the beginning
	     * and all the pointers pointing back to the first run.
	     * Therefore, the first run on the list is the rightmost
	     * one in the bottom line.  That's also the order in which
	     * they were encoded.  We must preserve this order when
	     * we return the list.
	     * We use two pointers, pRunHead and pRunTail.  pRunHead
	     * points to the leftmost, topmost run and pRunTail points
	     * to the rightmost, bottommost run in the list.  Therefore,
	     * pRunTail remains pointing to the very first run decoded
	     * while pRunHead advances along the chain as we go.
	     */
	lastX = symWidth;
	__error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
				    (void **)&pRunTail);
	if (__error__ != ia_successful)
	{
	    __error__ = ia_nomem;
	    goto tcDeserializeError;
	}

	pRunHead = pRunTail;

	    /* Fill up the accumulator with bytes from the input */
	accum = pShapeInfoBuf[0] |
		(pShapeInfoBuf[1] << 8) |
		(pShapeInfoBuf[2] << 16) |
		(pShapeInfoBuf[3] << 24);
	pShapeInfoBuf += 4;
	vacantBit = 32;			/* first open position, i.e. there
					 * are no openings */

	    /* take out the first code, which should be a horizontal code */
	codeValByte = mmrRevTopLevelDecodes[accum & cTopLevelCodeMask];
	if (codeValByte != cHorizCode)
	{
	    __error__ = ia_decodeFailure;
	    goto tcDeserializeError;
	}
	accum >>= cMMRHorizLen;
	vacantBit -= cMMRHorizLen;

	    /* now for the fun stuff - reading the first white code and
	     * the the first black code.  These are used to fill in the
	     * endpoints of the first run.  We find all white codes by
	     * trying to look them up in an 8 bit table.  If we get
	     * a terminating code, great.  If not, we call
	     * ip_decodeLengthCode to compose the number.
	     */
	codeIndex = (UInt8) (accum & cWhiteTermDecodeMask);
	codeValByte = mmrRevWhiteTermDecodes[codeIndex];
	if (codeValByte <= 63)
	{
	    lastX -= codeValByte;
		/* get bit length of code */
	    codeValByte = mmrRevWhiteTermDecodes[cTermDecodeTableLengthOffset +
						 codeIndex];
	    accum >>= codeValByte;
	    vacantBit -= codeValByte;
	    while (vacantBit <= 24)
	    {
		accum |= pShapeInfoBuf[0] << vacantBit;
		pShapeInfoBuf++;
		vacantBit += 8;
	    }
	}
	else	/* the code was a long one */
	{
	    __error__ = ip_decodeLengthCode (pIPState, &accum,
					   &vacantBit, &pShapeInfoBuf,
					   &lastX, cWhiteCode);
	    if (__error__ != ia_successful)
		goto tcDeserializeError;
	}

	    /* Set the run's right coord to whatever lastX is after skipping
	     * white space on the right */
	pRunHead->runRight = (Int16) lastX;
	pRunHead->pClient = (void *)cEvenLineTag;

	    /* get the number of black bits in the run.  First, see if
	     * it's a short code.
	     */
	if ((accum & cBlackLSBitsMask) != 0)
	{
		/* get code length and index using the short code table */
	    codeIndex = (UInt8) (accum & cShortBlackTermDecodeMask);
	    codeValByte = mmrRevBlackShortTermDecodes[codeIndex];
	    codeIndex =
		mmrRevBlackShortTermDecodes[cShortTermDecodeTableLengthOffset +
					    codeIndex];
	}
	else
	{
		/* skip the LS 4 bits when decoding, they're all 0, and use
		 * the long code table.
		 */
	    accum >>= 4;
	    vacantBit -= 4;
	    codeIndex = (UInt8) (accum & cBlackTermDecodeMask);
	    codeValByte = mmrRevBlackLongTermDecodes[codeIndex];
	    codeIndex =
		mmrRevBlackLongTermDecodes[cTermDecodeTableLengthOffset +
					   codeIndex];
	}
	    /* codeValByte holds the number of pixels represented by this
	     * code.  codeIndex holds the length of the code in bits.
	     */
	if (codeValByte <= 63)
	{
	    lastX -= codeValByte;
	    accum >>= codeIndex;
	    vacantBit -= codeIndex;
	    while (vacantBit <= 24)
	    {
		accum |= pShapeInfoBuf[0] << vacantBit;
		pShapeInfoBuf++;
		vacantBit += 8;
	    }
	}
	else	/* the code was a long one */
	{
	    __error__ = ip_decodeLengthCode (pIPState, &accum,
					   &vacantBit, &pShapeInfoBuf,
					   &lastX, cBlackCode);
	    if (__error__ != ia_successful)
		goto tcDeserializeError;
	}

	    /* Set the run's right coord to whatever lastX is after skipping
	     * the first black run */
	pRunHead->runLeft = (Int16) (lastX + 1);


	    /* At this point, we've retired the first symbol run.
	     * Let pRefLeftEndRun and pRefRightEndRun point to it
	     * and go to the next run.
	     * Set the last x position and the line flags corresponding
	     * to this x position.
	     */
	if (lastX < 0)
	{
		/* the first run ended at the left edge of the symbol,
		 * so wrap around to the rightmost pixel in the symbol
		 * just as the encoder did.  We're expecting to
		 * find a right endpoint next.  If the run on this line
		 * of the symbol butts up against the right edge of the
		 * symbol, we'll put out 0 white pixels as we're
		 * supposed to.
		 */
	    lastX = symWidth;
	    lastLine = cOddLineTag;
	}
	else
	    lastLine = cEvenLineTag;

	pRefLeftEndRun = pRefRightEndRun = pRunHead;
	state = cRightEndpoint;

	    /* Enter the loop that decodes all the rest of the runs
	     * for this symbol.
	     */
	while (state != cDone)
	{
	    switch (state)
	    {
	    case cRightEndpoint:
		    /* go find the the right endpoint on the reference
		     * line whose coordinate is less than or equal to
		     * the current X coordinate, lastX.
		     */
		while (((UInt32)pRefRightEndRun->pClient != lastLine) &&
		       (pRefRightEndRun->runRight > lastX))
		    pRefRightEndRun = pRefRightEndRun->pNext;

		    /* if pRefRightEndRun is now on the current line,
		     * make sure pRefLeftEndRun is there too.  Neither
		     * will be used right now, but that's ok.  We had
		     * a bug in which one of the pointers would advance
		     * to the next line, but the other wouldn't.
		     */
		if ((UInt32)pRefRightEndRun->pClient == lastLine)
		    pRefLeftEndRun = pRefRightEndRun;

		    /* Now, go find the code and dispatch on it.  This
		     * tells us what the encoding process decided to do
		     * at this point.
		     */
		codeValByte = mmrRevTopLevelDecodes[accum & cTopLevelCodeMask];

		switch (codeValByte)
		{
		case cVertM3Offset:
		case cVertM2Offset:
		case cVertM1Offset:
		case cVert0Offset:
		case cVertP1Offset:
		case cVertP2Offset:
		case cVertP3Offset:

			/* first, get rid of the code */
		    accum >>= mmrRevVertCodeLengths[codeValByte];
		    vacantBit -= mmrRevVertCodeLengths[codeValByte];
		    while (vacantBit <= 24)
		    {
			accum |= pShapeInfoBuf[0] << vacantBit;
			pShapeInfoBuf++;
			vacantBit += 8;
		    }
			/* we need to allocate another run since a new run
			 * has appeared.
			 */
		    __error__ =
			i_allocateStruct (pIPState,
					  pTextCompInfo->p1stScanLineBlock,
					  (void **)(&pRunHead->pNext));
		    if (__error__ != ia_successful)
		    {
			__error__ = ia_nomem;
			goto tcDeserializeError;
		    }
		    pRunHead = pRunHead->pNext;
    
			/* It doesn't matter whether pRefRightEndRun thought
			 * it was pointing to the reference line or had
			 * slipped into the coding line, it's pointing to
			 * the reference line relative to the current run
			 * now.  So just set the right endpoint of the new
			 * run relative to the reference right endpoint.
			 */
		    pRunHead->runRight = pRefRightEndRun->runRight +
					 codeValByte - cVert0Offset;
			/* if this run is the first one on a new line,
			 * flip the line tags. */
		    if (lastLine == (UInt32) pRefRightEndRun->pClient)
			lastLine ^= cLineTagMask;
		    pRunHead->pClient = (void *)lastLine;
		    lastX = pRunHead->runRight;
		    state = cLeftEndpoint;
		    break;
    
		case cPassCode:
			/* In this case, we're just advancing lastX and
			 * emptying the code out of the accumulator.  We
			 * advance it to just past the right endpoint of
			 * pRefRightEndRun.  This works whether or not
			 * pRefRightEndRun was on the same line as the
			 * *previous* coding line run because that's what
			 * the encoding process did.
			 */
		    vacantBit -= cMMRPassLen;
		    accum >>= cMMRPassLen;
		    while (vacantBit <= 24)
		    {
			accum |= pShapeInfoBuf[0] << vacantBit;
			pShapeInfoBuf++;
			vacantBit += 8;
		    }
			/* if we are skipping a run on the previous line,
			 * we must be on a new line, so flip the line tags. */
		    if (lastLine == (UInt32) pRefRightEndRun->pClient)
			lastLine ^= cLineTagMask;
		    lastX = pRefRightEndRun->runLeft - 1;
			/* we don't have to check to see if lastX went
			 * negative here because a reference run that's
			 * up against the left edge of the symbol can't
			 * be PASS'd because there can't be a coding
			 * run to its left.
			 */
		    break;

		case cHorizCode:
			/* get rid of the horizontal code itself */
		    vacantBit -= cMMRHorizLen;
		    accum >>= cMMRHorizLen;
		    while (vacantBit <= 24)
		    {
			accum |= pShapeInfoBuf[0] << vacantBit;
			pShapeInfoBuf++;
			vacantBit += 8;
		    }
    
			/* Oh, boy.  Now we get to decode a run of white pixels
			 * followed by a run of black pixels.  First, allocate
			 * a new run to put this info into.
			 */
		    __error__ =
			i_allocateStruct (pIPState,
					  pTextCompInfo->p1stScanLineBlock,
					  (void **) (&pRunHead->pNext));
		    if (__error__ != ia_successful)
		    {
			__error__ = ia_nomem;
			goto tcDeserializeError;
		    }
		    pRunHead = pRunHead->pNext;
    
			/* See if the white code is a terminating code. */
		    codeIndex = (UInt8) (accum & cWhiteTermDecodeMask);
		    codeValByte = mmrRevWhiteTermDecodes[codeIndex];
		    if (codeValByte <= 63)
		    {
			lastX -= codeValByte;
			    /* get bit length of code */
			codeValByte =
			    mmrRevWhiteTermDecodes[
						cTermDecodeTableLengthOffset +
						codeIndex];
			accum >>= codeValByte;
			vacantBit -= codeValByte;
			while (vacantBit <= 24)
			{
			    accum |= pShapeInfoBuf[0] << vacantBit;
			    pShapeInfoBuf++;
			    vacantBit += 8;
			}
		    }
		    else	/* the code was a long one */
		    {
			__error__ = ip_decodeLengthCode (pIPState, &accum,
						       &vacantBit,
						       &pShapeInfoBuf,
						       &lastX, cWhiteCode);
			if (__error__ != ia_successful)
			    goto tcDeserializeError;
		    }

			/* If the new run is on a new line, fix up lastX
			 * and lastLine.  Before advancing to this line,
			 * advance cRefRightEndRun and cRefLeftEndRun
			 * so they're really pointing to the line above
			 * the new one.
			 */
		    if (lastX < 0)
		    {
			lastX += symWidth + 1;
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			    pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			lastLine ^= cLineTagMask;
		    }

			/* Set the run's right coord to whatever lastX is
			 * after skipping white space on the right */
		    pRunHead->runRight = (Int16) lastX;
		    pRunHead->pClient = (void *)lastLine;
	    
			/* get the number of black bits in the run.
			 * First, see if it's a short code.
			 */
		    if ((accum & cBlackLSBitsMask) != 0)	/* it's short */
		    {
			    /* get code length and index using the short code
			     * table */
			codeIndex = (UInt8) (accum & cShortBlackTermDecodeMask);
			codeValByte = mmrRevBlackShortTermDecodes[codeIndex];
			codeIndex =
			    mmrRevBlackShortTermDecodes[
					    cShortTermDecodeTableLengthOffset +
					    codeIndex];
		    }
		    else
		    {
			    /* skip the LS 4 bits when decoding, they're all 0,
			     * and use the long code table.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (UInt8) (accum & cBlackTermDecodeMask);
			codeValByte = mmrRevBlackLongTermDecodes[codeIndex];
			codeIndex =
			    mmrRevBlackLongTermDecodes[
						cTermDecodeTableLengthOffset +
						codeIndex];
		    }
			/* codeValByte holds the number of pixels represented
			 * by this code.  codeIndex holds the length of the
			 * code in bits.
			 */
		    if (codeValByte <= 63)	/* it was a terminating code */
		    {
			lastX -= codeValByte;
			accum >>= codeIndex;
			vacantBit -= codeIndex;
			while (vacantBit <= 24)
			{
			    accum |= pShapeInfoBuf[0] << vacantBit;
			    pShapeInfoBuf++;
			    vacantBit += 8;
			}
		    }
		    else	/* the code was a long one */
		    {
			__error__ = ip_decodeLengthCode (pIPState, &accum,
						       &vacantBit,
						       &pShapeInfoBuf,
						       &lastX, cBlackCode);
			if (__error__ != ia_successful)
			    goto tcDeserializeError;
		    }
	    
			/* Set the run's right coord to whatever lastX is
			 * after skipping the first black run */
		    pRunHead->runLeft = lastX + 1;
    
			/* if that run abutted the left edge of the symbol,
			 * set lastX to the rightmost pixel on the previous
			 * line of the symbol.
			 * If that the symbol touches the right edge at
			 * that point, we'll happily say there are 0 white
			 * pixels before you get to black pixels.
			 * Also make sure that the reference run pointers
			 * are advanced so they really point to the line
			 * above the one we're about to encode.  If we
			 * didn't do this, the flags for the current
			 * reference run (line n-1) would become the
			 * same as the new value for lastLine (line n+1)
			 * even though they do not point to the same line.
			 */
		    if (lastX < 0)
		    {
			lastX = symWidth;
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			   pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			lastLine ^= cLineTagMask;
		    }
		    break;	/* end of finding a horizontal code when looking
			     * for a right endpoint */
    
		case cExtensionCode:
			/* We're done.  Set the state to say so.  Also set the
			 * pNext pointer of the last real run to NULL.
			 * Also back up pShapeInfoBuf so that we can
			 * start the next symbol by reading its width
			 * and height - which are now in accum.
			 */
		    pRunHead->pNext = NULL;
		    state = cDone;

		    vacantBit -= cMMRExtensionLen;	/* adjust for length
							 * of ext code */
		    pShapeInfoBuf -= (vacantBit / 8);	/* re-read all complete
							 * bytes */
		    break;
    
		default:	/* found an illegal encoding */
		    __error__ = ia_decodeFailure;
		    goto tcDeserializeError;
		}	/* end of switch on top level decode when looking
			 * for a right endpoint */
		break;	/* end of decoding a right endpoint */

	    case cLeftEndpoint:
		    /* go find the the left endpoint on the reference
		     * line whose coordinate is less than or equal to
		     * the current X coordinate, lastX.
		     */
		while (((UInt32)pRefLeftEndRun->pClient != lastLine) &&
		       (pRefLeftEndRun->runLeft > lastX))
		    pRefLeftEndRun = pRefLeftEndRun->pNext;

		    /* if pRefLeftEndRun is now on the current line,
		     * make sure pRefRightEndRun is there too.  Neither
		     * will be used right now, but that's ok.  We had
		     * a bug in which one of the pointers would advance
		     * to the next line, but the other wouldn't.
		     */
		if ((UInt32)pRefLeftEndRun->pClient == lastLine)
		    pRefRightEndRun = pRefLeftEndRun;

		    /* Now, go find the code and dispatch on it.  This
		     * tells us what the encoding process decided to do
		     * at this point.
		     */
		codeValByte = mmrRevTopLevelDecodes[accum & cTopLevelCodeMask];

		switch (codeValByte)
		{
		case cVertM3Offset:
		case cVertM2Offset:
		case cVertM1Offset:
		case cVert0Offset:
		case cVertP1Offset:
		case cVertP2Offset:
		case cVertP3Offset:
    
			/* first, get rid of the code */
		    accum >>= mmrRevVertCodeLengths[codeValByte];
		    vacantBit -= mmrRevVertCodeLengths[codeValByte];
		    while (vacantBit <= 24)
		    {
			accum |= pShapeInfoBuf[0] << vacantBit;
			pShapeInfoBuf++;
			vacantBit += 8;
		    }

			/* It doesn't matter whether pRefLeftEndRun thought
			 * it was pointing to the reference line or had
			 * slipped into the coding line, it's pointing to
			 * the reference line relative to the current run
			 * now.  So just set the left endpoint of the new
			 * run relative to the reference left endpoint.
			 */
		    pRunHead->runLeft = pRefLeftEndRun->runLeft +
					 codeValByte - cVert0Offset;
		    lastX = pRunHead->runLeft - 1;
			/* if that run abutted the left edge of the symbol,
			 * set lastX to the rightmost pixel on the previous
			 * line of the symbol.
			 * If that the symbol touches the right edge at
			 * that point, we'll happily say there are 0 white
			 * pixels before you get to black pixels.
			 * Also make sure that the reference run pointers
			 * are advanced so they really point to the line
			 * above the one we're about to encode.  If we
			 * didn't do this, the flags for the current
			 * reference run (line n-1) would become the
			 * same as the new value for lastLine (line n+1)
			 * even though they do not point to the same line.
			 */
		    if (lastX < 0)
		    {
			lastX = symWidth;
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			   pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			lastLine ^= cLineTagMask;
		    }
		    state = cRightEndpoint;
		    break;
    
		case cPassCode:
			/* In this case, we're just advancing lastX and
			 * emptying the code out of the accumulator.  We
			 * advance it to the right endpoint of
			 * pRefLeftEndRun->pNext.  This works whether or not
			 * pRefRightEndRun was on the same line as the
			 * *previous* coding line run because that's what
			 * the encoding process did.
			 */
		    vacantBit -= cMMRPassLen;
		    accum >>= cMMRPassLen;
		    while (vacantBit <= 24)
		    {
			accum |= pShapeInfoBuf[0] << vacantBit;
			pShapeInfoBuf++;
			vacantBit += 8;
		    }
			/* if we are skipping a run on the previous line,
			 * we must be on a new line, so flip the line tags. */
		    if (lastLine == (UInt32) pRefLeftEndRun->pClient)
			lastLine ^= cLineTagMask;
		    pRefLeftEndRun = pRefLeftEndRun->pNext;
		    lastX = pRefLeftEndRun->runRight;
		    break;

		case cHorizCode:
			/* get rid of the horizontal code itself */
		    vacantBit -= cMMRHorizLen;
		    accum >>= cMMRHorizLen;
		    while (vacantBit <= 24)
		    {
			accum |= pShapeInfoBuf[0] << vacantBit;
			pShapeInfoBuf++;
			vacantBit += 8;
		    }

			/* Oh, boy.  Now we get to decode a run of black pixels
			 * followed by a run of white pixels.  First,
			 * see if it's a short black code.
			 */
		    if ((accum & cBlackLSBitsMask) != 0)	/* it's short */
		    {
			    /* get code length and index using the short code
			     * table */
			codeIndex = (UInt8) (accum & cShortBlackTermDecodeMask);
			codeValByte = mmrRevBlackShortTermDecodes[codeIndex];
			codeIndex =
			    mmrRevBlackShortTermDecodes[
					    cShortTermDecodeTableLengthOffset +
					    codeIndex];
		    }
		    else
		    {
			    /* skip the LS 4 bits when decoding, they're all 0,
			     * and use the long code table.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (UInt8) (accum & cBlackTermDecodeMask);
			codeValByte = mmrRevBlackLongTermDecodes[codeIndex];
			codeIndex =
			    mmrRevBlackLongTermDecodes[
						cTermDecodeTableLengthOffset +
						codeIndex];
		    }
			/* codeValByte holds the number of pixels represented
			 * by this code.  codeIndex holds the length of the
			 * code in bits.
			 */
		    if (codeValByte <= 63)	/* it was a terminating code */
		    {
			lastX -= codeValByte;
			accum >>= codeIndex;
			vacantBit -= codeIndex;
			while (vacantBit <= 24)
			{
			    accum |= pShapeInfoBuf[0] << vacantBit;
			    pShapeInfoBuf++;
			    vacantBit += 8;
			}
		    }
		    else	/* the code was a long one */
		    {
			__error__ = ip_decodeLengthCode (pIPState, &accum,
						       &vacantBit,
						       &pShapeInfoBuf,
						       &lastX, cBlackCode);
			if (__error__ != ia_successful)
			    goto tcDeserializeError;
		    }
	    
			/* Set the run's right coord to whatever lastX is
			 * after skipping the black run */
		    pRunHead->runLeft = lastX + 1;
    
			/* if that run abutted the left edge of the symbol,
			 * set lastX to the rightmost pixel on the previous
			 * line of the symbol.
			 * If that the symbol touches the right edge at
			 * that point, we'll happily say there are 0 white
			 * pixels before you get to black pixels.
			 * Also make sure that the reference run pointers
			 * are advanced so they really point to the line
			 * above the one we're about to encode.  If we
			 * didn't do this, the flags for the current
			 * reference run (line n-1) would become the
			 * same as the new value for lastLine (line n+1)
			 * even though they do not point to the same line.
			 */
		    if (lastX < 0)
		    {
			lastX = symWidth;
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			   pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			lastLine ^= cLineTagMask;
		    }

			/* Allocate a new run for the new right endpoint we're
			 * about to find.
			 */
		    __error__ =
			i_allocateStruct (pIPState,
					  pTextCompInfo->p1stScanLineBlock,
					  (void **) (&pRunHead->pNext));
		    if (__error__ != ia_successful)
		    {
			__error__ = ia_nomem;
			goto tcDeserializeError;
		    }
		    pRunHead = pRunHead->pNext;
    
			/* See if the white code is a terminating code. */
		    codeIndex = (UInt8) (accum & cWhiteTermDecodeMask);
		    codeValByte = mmrRevWhiteTermDecodes[codeIndex];
		    if (codeValByte <= 63)
		    {
			lastX -= codeValByte;
			    /* get bit length of code */
			codeValByte =
			    mmrRevWhiteTermDecodes[
						cTermDecodeTableLengthOffset +
						codeIndex];
			accum >>= codeValByte;
			vacantBit -= codeValByte;
			while (vacantBit <= 24)
			{
			    accum |= pShapeInfoBuf[0] << vacantBit;
			    pShapeInfoBuf++;
			    vacantBit += 8;
			}
		    }
		    else	/* the code was a long one */
		    {
			__error__ = ip_decodeLengthCode (pIPState,
						       &accum, &vacantBit,
						       &pShapeInfoBuf,
						       &lastX, cWhiteCode);
			if (__error__ != ia_successful)
			    goto tcDeserializeError;
		    }

			/* If the new run is on a new line, fix up lastX
			 * and lastLine.  Before advancing to this line,
			 * advance cRefRightEndRun and cRefLeftEndRun
			 * so they're really pointing to the line above
			 * the new one.
			 */
		    if (lastX < 0)
		    {
			lastX += symWidth + 1;
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			    pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			lastLine ^= cLineTagMask;
		    }

			/* Set the run's right coord to whatever lastX is
			 * after skipping white space on the right */
		    pRunHead->runRight = (Int16) lastX;
		    pRunHead->pClient = (void *)lastLine;
		    break;

		case cExtensionCode:
			/* We're done.  Since we were looking for a
			 * left endpoint at the time, the last horizontal
			 * code must have run out of runs in the middle.
			 * Deallocate the last bogus run now.  Also set the
			 * pNext pointer of the last real run to NULL.
			 * Also back up pShapeInfoBuf so that we can
			 * start the next symbol by reading its width
			 * and height - which are now in accum.
			 */
		    while (pRefLeftEndRun->pNext != pRunHead)
			pRefLeftEndRun = pRefLeftEndRun->pNext;
		    pRefLeftEndRun->pNext = NULL;
		    i_freeStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
				  (void **)&pRunHead);
		    state = cDone;

		    vacantBit -= cMMRExtensionLen;	/* adjust for length
							 * of ext code */
		    pShapeInfoBuf -= (vacantBit / 8);	/* re-read all complete
							 * bytes */
		    break;
    
		default:	/* found an illegal encoding */
		    __error__ = ia_decodeFailure;
		    goto tcDeserializeError;
		}	/* end of switch on top level decode when looking
			 * for a left endpoint */
		break;	/* end of decoding a left endpoint */
	    }	/* end of left/right endpoint switch inside decoding loop */
	}	/* end of loop that decodes MMR runs for a symbol */


	    /* At this point, we have the width, height and runs for
	     * a new symbol.  Adjust the run coords (multiplying by 2
	     * and making them span the range -(w-1) to (w-1)).
	     * Note that the pClient members of each list element have
	     * already had the correct line tags inserted.  These were used
	     * during the decoding process.
	     */
	pTempRun = pRunTail;
	pTempRun->runLeft = (Int16) (2 * pTempRun->runLeft - symWidth);
	pTempRun->runRight = (Int16) (2 * pTempRun->runRight - symWidth);
	for (pRunHead = pTempRun->pNext;
	     pRunHead != NULL;
	     pRunHead = pRunHead->pNext)
	{
	    pRunHead->runLeft = (Int16) (2 * pRunHead->runLeft - symWidth);
	    pRunHead->runRight = (Int16) (2 * pRunHead->runRight - symWidth);
	    pTempRun = pRunHead;
	}

	    /* go insert this symbol into the list of symbols.  Remember
	     * that symWidth has already been decremented, so it
	     * equals (width of symbol - 1).
	     */
	instance.x = -(symWidth);
	instance.y = -(symHeight-1);
	instance.w = symWidth+1;	/* it was decremented above */
	instance.h = symHeight;
	instance.pHorizRuns = pRunTail;
	__error__ = ip_exactSymbolIsComplete (pIPState, pTextCompInfo,
					    NULL,
					    &instance,
					    cTrue);	/* just insert
							 * this symbol into
							 * the list */
	if (__error__ != ia_successful)
	{
		/* the vertical runs will be freed when we wipe out the
		 * scan line heap, so just pop down to the error return.
		 */
	    goto tcDeserializeError;
	}
	    /* The symbol was given the latest symbol ID.  That means
	     * pTextCompInfo->pLastSymbol points to it.  We need to
	     * set the instance count for the symbol.
	     */
	pTextCompInfo->pLastSymbol->instanceCount = symInstanceCount;

    }	/* end of loop that decodes all symbols */

	/* At this point, all the symbols have been decoded.  Return
	 * successfully with all the data structures filled up.
	 */
    NORMAL_RETURN;


	/* We get here when something went wrong.  This routine deallocates
	 * all the memory blocks and sets the pTextCompInfo pointers
	 * to NULL as needed.
	 */
tcDeserializeError:
	/* Free the arrays of pointers to vertical runs. */
    if (pTextCompInfo->pVertRunHeadArray != NULL)
	pTextCompInfo->freer(pIPState, pTextCompInfo->pVertRunHeadArray,
			     cSizeOfTCSymbolArray + 
			     (pTextCompInfo->pVertRunHeadArray->numPtrs - 1) *
						sizeof(TCScanLineRun *));
    if (pTextCompInfo->pVertRunTailArray != NULL)
	pTextCompInfo->freer(pIPState, pTextCompInfo->pVertRunTailArray,
			     cSizeOfTCSymbolArray + 
			     (pTextCompInfo->pVertRunTailArray->numPtrs - 1) *
						sizeof(TCScanLineRun *));

	/* free the Symbol runs */
    if (pTextCompInfo->p1stScanLineBlock != NULL)
	i_freeAllMemoryBlocks (pIPState, &(pTextCompInfo->p1stScanLineBlock));
    if (pTextCompInfo->p1stGroupBlock != NULL)
	i_freeAllMemoryBlocks (pIPState, &(pTextCompInfo->p1stGroupBlock));

	/* free the Scanned Symbols */
    if (pTextCompInfo->p1stSymbolBlock != NULL)
	i_freeAllMemoryBlocks (pIPState, &(pTextCompInfo->p1stSymbolBlock));

	/* restore all of this in case the caller wants to try again */
    for (w = 0; w <= cMaxSymIndexDim; w++)
	for (temp = 0; temp <= cMaxSymIndexDim; temp++)
	    (pTextCompInfo->pSymIndex->array)[w][temp] = NULL;
    pTextCompInfo->pBigSymbols = NULL;
    pTextCompInfo->p1stSymbol = NULL;
    pTextCompInfo->pLastSymbol = NULL;
    pTextCompInfo->pVertRunHeadArray = NULL;
    pTextCompInfo->pVertRunTailArray = NULL;
    pTextCompInfo->totalSymbols = 0;
    pTextCompInfo->nextSymbolNum = 0;

    NORMAL_RETURN;
}


/**************************************************************************
 *   ip_decodeLengthCode()
 *
 *	This procedure is called to decode numbers representing large
 *	pixel blocks.
 *	"Large" means greater than or equal to 64 since these numbers
 *	require non-terminating codes.
 *	This is the bulk of the code involved in decoding lengths,
 *	but isn't use very often.  That makes it ideal for putting in
 *	a procedure.  Pixel counts that require only terminating
 *	codes are decoded inline by the caller.
 *
 *	Input:
 *		pAccum:			pointer to word holding current
 *					block of bitstream bits.
 *		pVacantBit:		ptr to number of first vacant bit in
 *					accumulator.
 *		ppBuffer:		ptr to ptr to input buffer.
 *		pLastX:			ptr last X position in the line.
 *		codeType:		cWhiteCode => code represents a
 *					run of white pixels;
 *					cBlackCode => code represents a run
 *					of black pixels.
 *
 *	Action:
 *		Decodes the block of bits representing a block of pixels
 *		in the specified color.  The accmumulator is refilled
 *		with new bytes; the accumulator, vacantBit and *pBuffer
 *		are updated on return.  Finally, the x position is
 *		updated by the length of the block.  This may go negative,
 *		that's taken care of by the caller after this routine
 *		returns.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_decodeFailure	ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_decodeLengthCode (
    IPState    *pIPState,
    UInt32     *pAccum,
    Int32      *pVacantBit,
    UInt8     **ppBuffer,
    Int32      *pLastX,
    UInt32	codeType)
{
    PROLOGUE;
    UInt8      *pCurBuf = *ppBuffer;
    UInt32	accum = *pAccum;
    Int32	vacantBit = *pVacantBit;
    mmrTableEntry *pNonTermTable;	/* points to second level tables
					 * of non-terminating codes.
					 */
    UInt32	tokenVal;		/* the number represented by each
					 * terminating or non-terminating
					 * code value found in the loop */
    UInt32	codeIndex;
    UInt32	codeVal;
    UInt32	foundTerminatingCode;	/* set to true when a terminating
					 * code is found */

    foundTerminatingCode = cFalse;

	/* decide whether to use the black or white methods for decoding.
	 */
    if (codeType == cWhiteCode)
    {
	while (foundTerminatingCode == cFalse)
	{
		/* In the white world, we always use 8 bits to index a table.
		 * First, see if this is a terminating code yet.  Yes, this
		 * is a waste of time in the first iteration of the loop,
		 * but tough.
		 */
	    codeIndex = accum & cWhiteTermDecodeMask;
	    codeVal = mmrRevWhiteTermDecodes[codeIndex];
	    if (codeVal <= 63)
	    {
		foundTerminatingCode = cTrue;
		*pLastX -= codeVal;	/* move along the scan line */
		    /* get bit length of code */
		codeVal = mmrRevWhiteTermDecodes[cTermDecodeTableLengthOffset +
						 codeIndex];
		accum >>= codeVal;
		vacantBit -= codeVal;
		while (vacantBit <= 24)
		{
		    accum |= (*pCurBuf) << vacantBit;
		    pCurBuf++;
		    vacantBit += 8;
		}
	    }
	    else	/* look up this token using the
			 * non-terminating tables */
	    {
		pNonTermTable = mmrRevWhite1stNibbleNonTermDecodes[accum & 0xF];
		if (pNonTermTable == NULL)
		    return ia_decodeFailure;
	
		    /* We got a valid table pointer using the bottom 4 bits.
		     * advance past them and decode the next 4 bits.
		     */
		accum >>= 4;
		vacantBit -= 4;
		codeIndex = accum & 0xF;
		if (pNonTermTable[accum & 0xF].codeLength > 5)
		{
			/* The code in accum is one of the common codes used
			 * to skip really big blocks of pixels.  To decode
			 * these, we look it up in the black non-terminating
			 * tables.
			 */
		    pNonTermTable =
			mmrRevBlack1stNibbleNonTermDecodes[accum & 0xF];
		    if (pNonTermTable == NULL)
			return ia_decodeFailure;
	
			/* Now we really have a pointer to the lower level
			 * non-terminating code table, so toss the nibble we
			 * just used.
			 */
		    accum >>= 4;
		    vacantBit -= 4;
		    codeIndex = accum & 0xF;
		}
	
		    /* and look up the code length */
		codeVal = pNonTermTable[codeIndex].codeLength;
		if (codeVal == 0)
		    return ia_decodeFailure;
    
		tokenVal = pNonTermTable[codeIndex].codeBits;
		if (codeVal == 5)	/* add 64 x value of 5th bit */
		{
		    tokenVal += (accum * 4) & 64;
		    if (tokenVal == cWeirdWhiteCode)
			tokenVal += 64;
		} 
		accum >>= codeVal;
		vacantBit -= codeVal;

		    /* now move to the left in the current scan line by
		     * the number of pixels just decoded in tokenVal.
		     * We keep doing this until we find a terminating code.
		     */
		*pLastX -= tokenVal;

		    /* And replenish the accumulator with fresh bits from
		     * the input buffer.
		     */
		while (vacantBit <= 24)
		{
		    accum |= (*pCurBuf) << vacantBit;
		    pCurBuf++;
		    vacantBit += 8;
		}
	    }	/* end of non-terminating white case */
	}	/* end of loop that decodes white values */
    }	/* end of decoding runs of white pixels */
    else	/* decode runs of black pixels */
    {
	accum <<= 4;				/* restore the 4 bits that
						 * were deleted before we
						 * got here */
	vacantBit += 4;
	while (foundTerminatingCode == cFalse)
	{
		/* In the black world, terminating codes can be short
		 * or long.  First, see if we have a short terminating
		 * code.  If that doesn't work, try a long terminating
		 * code.  If that doesn't work, we'll try a non-terminating
		 * code.  Yes, looking for a terminating code
		 * is a waste of time in the first iteration of the loop,
		 * but tough.
		 */
	    if ((accum & cBlackLSBitsMask) != 0)	/* it's short */
	    {
		    /* get code length and index using the short code
		     * table */
		codeIndex = accum & cShortBlackTermDecodeMask;
		codeVal = mmrRevBlackShortTermDecodes[codeIndex];
		codeIndex =
		    mmrRevBlackShortTermDecodes[
				    cShortTermDecodeTableLengthOffset +
				    codeIndex];
	    }
	    else
	    {
		    /* delete the LS four bits since they're all 0, and use
		     * the long terminating code table. */
		accum >>= 4;
		vacantBit -= 4;
		codeIndex = accum & cBlackTermDecodeMask;
		codeVal = mmrRevBlackLongTermDecodes[codeIndex];
		codeIndex =
		    mmrRevBlackLongTermDecodes[
					cTermDecodeTableLengthOffset +
					codeIndex];
	    }

	    if (codeVal <= 63)	/* it was a terminating code */
	    {
		foundTerminatingCode = cTrue;
		    /* move to the left in the scan line for the last time
		     * in this procedure, remove these code bits from the
		     * accumulator and replenish the accumulator with fresh
		     * bits.
		     */
		*pLastX -= codeVal;
		accum >>= codeIndex;
		vacantBit -= codeIndex;
		while (vacantBit <= 24)
		{
		    accum |= (*pCurBuf) << vacantBit;
		    pCurBuf++;
		    vacantBit += 8;
		}
	    }
	    else	/* look up this token using the
			 * non-terminating tables.  At this point, we've
			 * already deleted the LS four bits, so we have
			 * the right nibble for this table
			 */
	    {
		pNonTermTable = mmrRevBlack1stNibbleNonTermDecodes[accum & 0xF];
		if (pNonTermTable == NULL)
		    return ia_decodeFailure;
	
		    /* We got a valid table pointer using the bottom 4 bits.
		     * advance past them and decode the next 4 bits.
		     */
		accum >>= 4;
		vacantBit -= 4;
		codeIndex = accum & 0xF;

		    /* and look up the code length */
		codeVal = pNonTermTable[codeIndex].codeLength;
		if (codeVal == 0)
		    return ia_decodeFailure;
    
		tokenVal = pNonTermTable[codeIndex].codeBits;
		if (codeVal == 5)	/* add 64 x value of 5th bit */
		    tokenVal += (accum * 4) & 64;
		accum >>= codeVal;
		vacantBit -= codeVal;

		    /* now move to the left in the current scan line by
		     * the number of pixels just decoded in tokenVal.
		     * We keep doing this until we find a terminating code.
		     */
		*pLastX -= tokenVal;

		    /* and replenish the accumulator with fresh bits from
		     * the input buffer.
		     */
		while (vacantBit <= 24)
		{
		    accum |= (*pCurBuf) << vacantBit;
		    pCurBuf++;
		    vacantBit += 8;
		}
	    }	/* end of non-terminating black case */
	}	/* end of loop that decodes black values */
    }		/* end of decoding black values */

	/* restore accum, vacantBit and *pBuffer */
    *pAccum = accum;
    *pVacantBit = vacantBit;
    *ppBuffer = pCurBuf;

    NORMAL_RETURN;
}



/*

 $Log: i_tcdser.c,v $
 Revision 1.2  2001/05/08 06:43:36  jlin
 change ->allocator to ->ipcAllocator

 * Revision 1.1.1.1  1999/09/20  18:18:13  jlin
 * Imported sources
 *
 Revision 1.20  1998/01/13 23:50:00  ddavies
 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.19  1997/08/27 02:41:21  ddavies
 ddavies on Tue Aug 26 19:41:05 PDT 1997

 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Initialize the new useResidual flag in the TextCompInfo.

 Revision 1.18  1997/05/08 02:33:47  ddavies
 ddavies on Wed May 7 19:32:50 PDT 1997

 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Use ip_exactSymbolIsComplete instead of ip_symbolIsComplete to insert
 decoded symbols into symbol tree.  This lets us separate lossless stuff
 from lossy stuff more easily.

 Revision 1.17  1997/04/29 02:41:49  ddavies
 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Delete code that (re)alloccates the vertical run arrays during
 dictionary deserialization.  This will be done by i_convertImageToSymbols
 if necessary.
 Initialize the numRuns entry of the instance to 0.

 Revision 1.16  1997/04/24 23:29:55  ddavies
 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Change ip_fixupSymbolPtrs to ip_fixupTCSymbolPtrs.

 Revision 1.15  1997/02/12 02:53:38  ddavies
 ddavies on Tue Feb 11 18:52:53 PST 1997

 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Initialize leftProfile, rightProfile and maxDimension in the place where
 a TCInstance is initialized.

 Revision 1.14  1997/01/04 00:25:10  ddavies
 ddavies on Fri Jan 3 16:24:15 PST 1997

 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Change instance.interior to instance.strokeWidth.  Fix comments so we
 talk about using strokeWidth instead of interior as a discriminating measure.

 Revision 1.13  1996/08/22 01:21:02  ddavies
 ddavies on Wed Aug 21 18:19:28 PDT 1996

 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Add usageCount to info recorded for old serialized dictionaries.  The
 Tundra code will need it.  Yeah, the Tundra code will only use these
 serialized dictionaries while it's being created, but that's ok.

 Revision 1.12  1996/07/22 18:33:14  ddavies
 ddavies on Mon Jul 22 11:32:35 PDT 1996

 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Remove initialization of obsolete TCCompressionInfo entry called lastX.

 Revision 1.11  1996/07/18 23:50:34  ddavies
 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Add comment saying that there's no need to destroy and existing symbol
 table since there can't be any existing symbols.  Normally, we would
 destroy an existing symbol table before adding symbols in
 i_convertSymbolsToImage.

 Revision 1.10  1996/07/04 02:57:10  ddavies
 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Initialize new TCCompressionInfo element, pLastSymbol.

 Revision 1.9  1996/04/25 18:49:23  ddavies
 ddavies on Thu Apr 25 11:48:20 PDT 1996

 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 Define _NEED_TOPLEVEL_CODES_ and _NEED_HORIZONTAL_CODES_ before including
 mmrdec.prv so that both tables in that file will be included.

 Revision 1.8  1995/12/13 02:46:26  ejaquith
 i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.7  1995/10/18  03:58:59  ddavies
 * i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.6  1995/07/17  22:01:19  ddavies
 * ddavies on Mon Jul 17 15:01:08 PDT 1995
 *
 * i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v
 *
 * Change all occurences of sizeof(TCSymbolArray) to cSizeOfTCSymbolArray.
 * This is part of the changes for the Mac MPW compiler.
 *
 * Revision 1.5  1995/07/10  23:50:24  ddavies
 * i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v
 *
 * Change the pointer to an array of TCSymbol pointers in a TCCompressionInfo
 * structure to a pointer to a TCSymbolArray.  This structure has a PointerFixup
 * struct at the beginning.  Do the same for the two arrays of pointers to
 * TCScanLineRun's used to convert horizontal to vertical runs.  Add args to
 * calls to i_allocateMemoryBlock to say we should be notified if the
 * blocks move.  Initialize the PointerFixup structures at the beginning of
 * the arrays of pointers.  Use the number of pointers inside the
 * TCScanLineArray structures instead of the width in the TCCompressionInfo
 * (that's gone).  Initialize the pointer arrays so Purify won't complain when
 * we check them after pointer fixup.  Change the casts in calls to
 * i_allocateStruct from (TCBlockHeader **) to (void **).  This is more
 * correct than pretending every structure in every heap is a
 * TCBlockHeader just because that structure began with a pNext pointer.  And,
 * a TCBlockHeader no longer begins with a pNext pointer.
 *
 * Revision 1.4  1995/04/01  03:20:08  ddavies
 * ddavies on Fri Mar 31 19:18:57 PST 1995
 *
 * i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v
 *
 * Change references to pTextCompInfo->symIndex to (*(pTextCompInfo->pSymIndex)).
 * That's because the TCCompressionInfo structure now has a pointer to a 2D
 * array of symbol pointers instead of the array itself.
 *
 * Revision 1.3  1995/03/15  02:14:13  ddavies
 * i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v
 *
 * Replace a bunch of interal variables with a TCInstance data structure in
 * i_deserializeCompressionInfo.  Pass the TCInstance to ip_symbolIsComplete.
 * Clear the 2D array of pointers to symbols in the TCCompressionInfo structure
 * Instead of clearing a 1D array that was allocated separately.
 *
 * Revision 1.2  1994/11/10  21:44:44  ddavies
 * ddavies on Thu Nov 10 13:44:06 PST 1994
 *
 * i_tcdser.c -> /product/ipcore/ipshared/src/RCS/i_tcdser.c,v
 *
 * Add null volume arg to call to ip_symbolIsComplete.
 *
 * Revision 1.1  1994/10/14  19:23:20  ddavies
 * Initial revision
 *

 */



