/********************************************************************
 *  Copyright (c) 1995-2000, Xerox Corporation.All rights reserved. *
 *  Copyright protection claimed includes all forms and matters     *
 *  of copyrightable material and information now allowed by        *
 *  statutory or judicial law or hereafter granted, including       *
 *  without limitation, material generated from the software        *
 *  programs which are displayed on the screen such as icons,       *
 *  screen display looks, etc.                                      *
 ********************************************************************/

/*	
 *  i_tcenc.c: Contains procedures used to turn an image into a
 *		set of symbols and positions of symbols.  The procedure
 *		used to do the reverse transform is in i_tcdec.c
 *
 *		Int32	i_convertImageToSymbols()
 *			Classify the symbols in an image into a set of
 *			equivalence classes.  Mark the position of each
 *			symbol in a data block.  Build the data structures
 *			used to do the classification.
 *
 *		Int32	ip_classifySymbol()
 *			This procedure is called with a list of all the
 *			runs comprising a symbol.  It classifies the
 *			symbol and puts out position info.
 *
 *		Int32	ip_compareSymbols()
 *			This procedure is used by ip_classifySymbol to
 *			compare an instance with a list of exemplars.
 *			It returns the first match.
 *
 *		Int32	i_consolidateShapeInfo()
 *			Merges classes of symbols that differ only in
 *			noise.
 *
 *		Int32	i_remapSymbolIDs()
 *			Changes symbol IDs in position buf to account
 *			for new symbol IDs produced by i_consolidateShapeInfo.
 *
 */

#ifndef _WIN32_WCE
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#include "memory.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /*  prototyes */
#include "shrcc.pub"      /* for i_findConnectedComponentRuns prototype */
#include "shrcmprs.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tcenc.c,v $; $Revision: 1.5 $; $Date: 2001/05/08 06:43:37 $");

#define	TC_DEBUG	0		/* used to put out debugging info */

/* The process we use to compress images by finding repeated symbols is
 * well documented (at least there's lots of documentation) in the
 * file i_tcompr.c.  You'll have to read that to understand this.
 *
 * This file contains the routine used to turn a page containing connected
 * components into a list of positions and symbol IDs and a set of
 * data structures we can use to do further classification.  The format
 * of the list of positions is given in i_tcompr.c.  The data
 * structures are defined in shrcmprs.prv.
 *
 * This routine calls i_findConnectedComponentRuns, giving it a callback
 * proc that actually classifies each symbol as it is found.
 *
 */

/* PUBLIC */
/**************************************************************************
 *   i_convertImageToSymbols()
 *
 *	This procedure converts a binary image of repeated symbols, usually
 *	characters, into a set of positions, symbol ID's and a data
 *	structure used to classify symbols.
 *
 *	Input:
 *		pImage:			Pointer to upper left corner of image
 *					INCLUDING frame.
 *		width:			image width including frame
 *		height:			image height including frame
 *		bpl:			The image's bytes/line
 *		pTextCompInfo:		pointer to structure containing
 *					information about shape, stroke and
 *					symbol trees
 *		pageID:			Page ID assigned to parent image
 *		pPositionBuf:		pointer to buffer that will contain
 *					position and symbol ID's.
 *		positionBufSize:	size of that buffer in bytes.
 *		pResultSize:		ptr to UInt32 that will hold the
 *					number of bytes actually used in
 *					the position buffer
 *
 *	Action:
 *		Builds classification data structures and writes position
 *		and symbol ID info to *pPositionBuf.  If an error is
 *		encountered, the classification data structures are NOT
 *		freed.  This allows the client to try again with more
 *		memory.  The client may explicitly free them.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *		ia_bufferExceeded	position and symbol ID info didn't
 *					fit into the position buffer.
 *
 **************************************************************************/
Int32  CDECL 
i_convertImageToSymbols (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	width,
    Int32	height,
    UInt32	bpl,
    void       *pTextCompInfoIn,
    UInt32	pageID,
    UInt8      *pPositionBuf,
    UInt32	positionBufSize,
    UInt32     *pResultSize)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo;	/* ptr to compression info */
    TCSymbol   *pSymbol, *pNextSym;
    TCScanLineRun *pNextRun;		/* used to free vertical runs */
    Int32	index;
    Int32	symWidth;		/* of an existing symbol */
    Int32	symHeight;		/* of an existing symbol */
    Int32	classifySymRetCode;	/* return code left
					 * by ip_classifySymbol when something
					 * went wrong */
    Int32	delta;			/* amount by which symbol measures
					 * can vary */
    TCInstance	instance;		/* used to calculate data for syms */
    UInt32	arraySize;		/* size of pointer arrays */
    

	/* make sure all the pointers are ok */
    if ((pImage == NULL) || (pTextCompInfoIn == NULL) ||
	(pPositionBuf == NULL) || (pResultSize == NULL) ||
	(positionBufSize == 0))
	return ia_invalidParm;

	/* convert the pointer to the correct type */
    pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;

	/* If we haven't done any matching yet, set the exactMatch flags.
	 */
    if (!pTextCompInfo->exactMatchSet)
    {
	pTextCompInfo->exactMatchOnly = cFalse;	/* we're doing lossy matches */
	pTextCompInfo->exactMatchSet = cTrue;
    }
    else	/* were we doing lossy matches? */
    if (pTextCompInfo->exactMatchOnly != cFalse)
	return ia_invalidParm;			/* can't switch back */

	/* If the internal data structure contains symbols that
	 * have already been put out in a Tundra dictionary or
	 * residual image, delete them.  Leaving them around will
	 * only create confusion.  Also delete any run-less
	 * symbols that were put here by i_readTundraDictionaryBlock.
	 */
    if (pTextCompInfo->encodeState & cTCSymsHaveBeenEncoded)
    {
	for (pSymbol = pTextCompInfo->p1stSymbol;
	     pSymbol != NULL;
	     pSymbol = pNextSym)
	{
	    pNextSym = pSymbol->pNextNum;		/* pt to next symbol */
	    if ((pSymbol->runsPtToTail) ||
		((pSymbol->pRunTail == NULL) && (pSymbol->pRunHead == NULL)))
		ip_deleteTCDictionarySymbol (pIPState, pTextCompInfo, pSymbol);
	}
	    /* Go reassign the symbol IDs */
	for (index = 0, pSymbol = pTextCompInfo->p1stSymbol;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextNum, index++)
	{
	    pSymbol->symbolID = index;
	}
	    /* Set the next symbol ID for this dictionary and
	     * declare that the encode state is null.
	     */
	pTextCompInfo->nextSymbolNum = index;
	pTextCompInfo->encodeState = 0;
    }

	/* If there's already a cached table to pointers to symbols,
	 * delete it.  The table will be invalid after we add
	 * some symbols.
	 */
    __error__ = ip_destroyTCSymbolTable (pIPState, pTextCompInfo);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* i_findConnectedComponentRuns will do the clipping, so
	 * don't do any here.*/

	/* set the temporary vars in *pTextCompInfo that hold data about
	 * the position buffer.  We put them in here because it makes
	 * it easy to pass them around this way.  We want the real image
	 * width and height after this, so subtract off the frames now.
	 * Note that we initialize lastY to -1 so that the first
	 * deltaY offset recorded for any non-empty image is non-zero
	 * and the first deltaY offset recorded for a blank image is zero.
	 */
    width -= 2*cFrameBits;
    height -= 2*cFrameBits;
    pTextCompInfo->imageWidth = (Int16) width;
    pTextCompInfo->imageHeight = (Int16) height;
    pTextCompInfo->pPositionBuf = pPositionBuf;
    pTextCompInfo->positionBufSize = positionBufSize;
    pTextCompInfo->resultSize = 0;		/* used no bytes so far */
    pTextCompInfo->lastX = 0;
    pTextCompInfo->lastY = -1;

	/* nextSymbolNum is initialized when the text compression
	 * info structure is allocated.  That way, we can continue
	 * to compress pages with the same dictionary of symbols.
	 * Make sure we build the shape tree by turning off
	 * the classifyOnly flag.
	 */
    pTextCompInfo->classifyOnly = cFalse;

	/* Check each entry in the textCompInfo.  If an entry is NULL,
	 * allocate the corresponding initial heap or memory block
	 * and update the textCompInfo entry.  These heaps  and blocks
	 * are freed when we free the text compression info.  This
	 * allows us to compress several pages in a row without fussing
	 * with the heaps all the time.
	 */

	/* if the current vertical run arrays aren't wide enough,
	 * free them.
	 */
    if ((pTextCompInfo->pVertRunHeadArray != NULL) &&
	(pTextCompInfo->pVertRunHeadArray->numPtrs < width))
    {
	arraySize = sizeof(TCScanLineArray) +
		    ((pTextCompInfo->pVertRunHeadArray->numPtrs - 1) *
						sizeof(TCScanLineRun *));
	__error__ = ip_detachBlockFromPtrFixupRing (
				pIPState, 
				&(pTextCompInfo->pVertRunHeadArray->fixup));
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
	pTextCompInfo->freer(pIPState,
			     pTextCompInfo->pVertRunHeadArray, arraySize);
	pTextCompInfo->pVertRunHeadArray = NULL;

	__error__ = ip_detachBlockFromPtrFixupRing (
				pIPState, 
				&(pTextCompInfo->pVertRunTailArray->fixup));
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
	pTextCompInfo->freer(pIPState,
			     pTextCompInfo->pVertRunTailArray, arraySize);
	pTextCompInfo->pVertRunTailArray = NULL;
    }


	/* if there aren't any pointer arrays now, possible because
	 * we just freed them, make some of the right width.
	 */
    if (pTextCompInfo->pVertRunHeadArray == NULL)
    {
	arraySize = sizeof(TCScanLineArray) +
		    ((width-1) * sizeof(TCScanLineRun *));
	pTextCompInfo->pVertRunHeadArray = 
		pTextCompInfo->ipcAllocator(pIPState, arraySize, cTrue);
	if (pTextCompInfo->pVertRunHeadArray == NULL)	/* if it didn't work,
							 * just quit */
	    return ia_nomem;

	    /* initialize the scan line arrays so Purify doesn't complain
	     * when we read the pointers when fixing them after garbage
	     * collection.
	     */
	memset((void *)(pTextCompInfo->pVertRunHeadArray), 0, arraySize);

	pTextCompInfo->pVertRunHeadArray->numPtrs = width;
	    /* Initialize the pointer fixup stuff */
	__error__ =
		ip_initPointerFixup (pIPState,
				     &(pTextCompInfo->pVertRunHeadArray->fixup),
				     arraySize,
				     (PointerFixup *)pTextCompInfo,
				     ip_fixupTCScanLineArrayPtrs);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;

	pTextCompInfo->pVertRunTailArray = 
		pTextCompInfo->ipcAllocator(pIPState, arraySize, cTrue);
	if (pTextCompInfo->pVertRunTailArray == NULL)	/* if it didn't work,
							 * just quit */
	    return ia_nomem;
	    /* initialize the scan line arrays so Purify doesn't complain
	     * when we read the pointers when fixing them after garbage
	     * collection.
	     */
	memset((void *)(pTextCompInfo->pVertRunTailArray), 0, arraySize);

	pTextCompInfo->pVertRunTailArray->numPtrs = width;
	    /* Initialize the pointer fixup stuff */
	__error__ =
		ip_initPointerFixup (pIPState,
				     &(pTextCompInfo->pVertRunTailArray->fixup),
				     arraySize,
				     (PointerFixup *)pTextCompInfo,
				     ip_fixupTCScanLineArrayPtrs);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
    }

    if (pTextCompInfo->p1stScanLineBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stScanLineBlock),
					 sizeof(TCScanLineRun),
					 cTCScanLineRunBlockSize,
					 cTrue,		/* adjust ptrs after
							 * garbage collection */
					 &(pTextCompInfo->fixup),
					 &ip_fixupScanLinePtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

    if (pTextCompInfo->p1stGroupBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stGroupBlock),
					 sizeof(TCGroupRun),
					 cTCGroupRunBlockSize,
					 cTrue,
					 &(pTextCompInfo->fixup),
					 &ip_fixupTCGroupRunPtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

    if (pTextCompInfo->p1stSymbolBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stSymbolBlock),
					 sizeof(TCSymbol),
					 cTCSymbolBlockSize,
					 cTrue,
					 &(pTextCompInfo->fixup),
					 &ip_fixupTCSymbolPtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

	/* If we already have a list of symbols, but they don't have the
	 * classification structures, add the classification structures
	 * to each symbol.
	 */
    pSymbol = pTextCompInfo->p1stSymbol;
    if ((pSymbol != NULL) && (pSymbol->p1stHorizGroup == NULL))
    {
	    /* We'll need to build information for the symbols.
	     * Initialize the instance data structure we'll use below.
	     */
	instance.minWidth = 0;
	instance.maxWidth = 0;
	instance.minHeight = 0;
	instance.maxHeight = 0;
	instance.x = 0;
	instance.y = 0;
	instance.w = 0;
	instance.h = 0;
	instance.leftProfile = 0;
	instance.rightProfile = 0;
	instance.maxDimension = 0;
	instance.strokeWidth = 0;
	instance.mass = 0;
	instance.numRuns = 0;
	instance.checkHVolume = cFalse;
	instance.checkVVolume = cFalse;
	instance.hVolume = 0;
	instance.vVolume = 0;
	instance.slant = 0;
	instance.hSAsymmetry = 0;
	instance.vSAsymmetry = 0;
	instance.pCurSymbol = NULL;
	instance.pHorizRuns = NULL;
	instance.pDupHorizRuns = NULL;
	instance.pVertRuns = NULL;
	instance.pDupVertRuns = NULL;

	    /* build classification data for all the symbols */
	for (;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextNum)
	{
	    symWidth = pSymbol->maxX + 1;
	    symHeight = pSymbol->maxY + 1;
	    delta = MAX(symWidth, symHeight);
	    delta = MIN((delta - cDimensionOffset)/cDimensionDivisor,
			cMaxDimensionSlop);
	    delta = MAX(delta, 0);
	    pSymbol->minWidth = symWidth - delta;
	    pSymbol->maxWidth = symWidth + delta;
	    pSymbol->minHeight = symHeight - delta;
	    pSymbol->maxHeight = symHeight + delta;

		/* compute the mass for this symbol */
	    for (index = 0, pNextRun = pSymbol->pRunTail;
		 pNextRun != NULL;
		 pNextRun = pNextRun->pNext)
	    {
		index += pNextRun->runRight - pNextRun->runLeft + 2;
	    }
	    pSymbol->mass = index;

	    pSymbol->numActiveRuns = 0;	/* nothing merged with this yet */
	    pSymbol->instanceCount = 1;	/* it would be nice if we kept
					 * this around during the bytestream
					 * processing, but we didn't.  So
					 * we just assume there's at least
					 * one instance of this symbol
					 * somewhere.
					 */

	    instance.pHorizRuns = pSymbol->pRunTail,
	    instance.w = symWidth;
	    instance.h = symHeight;
	    instance.mass = pSymbol->mass;

		/* make the vertical runs */
	    __error__ = ip_makeVerticalRuns (pIPState, pTextCompInfo,
					   symWidth, symHeight,
					   pSymbol->pRunTail,
					   &(instance.pVertRuns));
	    if (__error__ != ia_successful)
		NORMAL_RETURN;

	    __error__ = ip_makeGroupsFromRuns (pIPState, pTextCompInfo,
					     &instance,
					     &(pSymbol->p1stHorizGroup),
					     &(pSymbol->pDupHorizGroup),
					     cTrue);
	    if (__error__ != ia_successful)
		NORMAL_RETURN;

	    __error__ = ip_makeGroupsFromRuns (pIPState, pTextCompInfo,
					     &instance,
					     &(pSymbol->p1stVertGroup),
					     &(pSymbol->pDupVertGroup),
					     cFalse);
	    if (__error__ != ia_successful)
		NORMAL_RETURN;

		/* Also calculate the slant for each symbol */
	    __error__ = ip_calculateSlant (pIPState, &instance);

		/* Record all the info calculated by ip_calculateSlant
		 * in the symbol.
		 */
	    pSymbol->slant = instance.slant;
	    pSymbol->maxDimension= instance.maxDimension;
	    pSymbol->checkHVolume = instance.checkHVolume;
	    pSymbol->checkVVolume = instance.checkVVolume;
	    pSymbol->hVolume = instance.hVolume;
	    pSymbol->vVolume = instance.vVolume;
	    pSymbol->hSAsymmetry = instance.hSAsymmetry;
	    pSymbol->vSAsymmetry = instance.vSAsymmetry;
	    pSymbol->pClassSymbol = NULL;
	    pSymbol->pNewSymbol = NULL;

		/* free the vertical runs */
	    while (instance.pVertRuns != NULL)
	    {
		pNextRun = instance.pVertRuns->pNext;
		i_freeStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
			      (void **)&instance.pVertRuns);
		instance.pVertRuns = pNextRun;
	    }
	}
	if (__error__ != ia_successful)
	    NORMAL_RETURN;

    }

	/* Start the position info with a magic number and version number.
	 */
    pTextCompInfo->pPositionBuf[cTCPosMagicNum0] =
				(UInt8)(cTextCompPositionInfoMagicNum);
    pTextCompInfo->pPositionBuf[cTCPosMagicNum1] = 
				(UInt8)(cTextCompPositionInfoMagicNum >> 8);
    pTextCompInfo->pPositionBuf[cTCPosMagicNum2] = 
				(UInt8)(cTextCompPositionInfoMagicNum >> 16);
    pTextCompInfo->pPositionBuf[cTCPosMagicNum3] = 
				(UInt8)(cTextCompPositionInfoMagicNum >> 24);

    pTextCompInfo->pPositionBuf[cTCPosVer0] = 
				(UInt8)(cTextCompPositionInfoVersionNum);
    pTextCompInfo->pPositionBuf[cTCPosVer1] = 
				(UInt8)(cTextCompPositionInfoVersionNum >> 8);

	/* Add image width and height, 2 bytes each */
    pTextCompInfo->pPositionBuf[cTCPosWidth0] = (UInt8) width;
    pTextCompInfo->pPositionBuf[cTCPosWidth1] = (UInt8) (width >> 8);

    pTextCompInfo->pPositionBuf[cTCPosHeight0] = (UInt8) height;
    pTextCompInfo->pPositionBuf[cTCPosHeight1] = (UInt8) (height >> 8);

	/* Set the rest of the statistics to 0 (and invalid) */
    memset(&(pPositionBuf[cTCPosStatsValid]), 0, cTCPosStatsLen);

    pTextCompInfo->pPositionBuf += cTCPosHeaderVer2Len;
    pTextCompInfo->resultSize += cTCPosHeaderVer2Len;

	/* Initialize the first and last dummy elements in the list
	 * of uncommitted positions.
	 * We try to allow bottom line of the recorded symbol positions
	 * to vary up or down a line so that the replacement symbol has
	 * the same horizontal centerline as the instance symbol.
	 * This means we can't be sure what's in a scan line until the
	 * symbols in the line above, the same line, and the line below
	 * have all been found.  To do this, we keep a list sorted by
	 * x position.  The list elements contain x pos, y pos, symID
	 * and ptr to next element.  We initialize two dummy elements
	 * here so that we don't have to worry about inserting elements
	 * at the beginning or end of the list.
	 * Note that we don't have to worry about garbage collection inside
	 * this list since it only exists during i_convertImageToSymbols.
	 * We set the dummy's Y position to -2 so it won't be
	 * confused with the initial -1 for the mythical first row.
	 * This lets us just always flush rows starting with
	 * lowUncomY.
	 */
    pTextCompInfo->dum1stLoc.pClient = (void *) 0xFFFFFFFF; /* no sym */
    pTextCompInfo->dum1stLoc.runLeft = -1;	    /* x pos = -1 */
    pTextCompInfo->dum1stLoc.runRight = -2;	    /* y pos = -2 */
    pTextCompInfo->dum1stLoc.pNext = &(pTextCompInfo->dumLastLoc);

    pTextCompInfo->dumLastLoc.pClient = (void *) 0xFFFFFFFF; /* no sym */
    pTextCompInfo->dumLastLoc.runLeft = 0x7FFF;	    /* x pos = big number */
    pTextCompInfo->dumLastLoc.runRight = -2;        /* y pos = -2 */
    pTextCompInfo->dumLastLoc.pNext = NULL;	    /* nothing to the right */

	/* This first buffer may look something like this.  It assumes
	 * the first symbols may start on line 0.
	 */
    pTextCompInfo->lowUncomY = -1;
    pTextCompInfo->hiUncomY = 1;
    pTextCompInfo->numUncomPos = 0;		    /* the list is empty */
    pTextCompInfo->pLastUncom = &(pTextCompInfo->dum1stLoc);

	/* Allow the page ID to propogate once a symbol has been added to the
	 * dictionary */
    pTextCompInfo->curPageID = pageID;

	/* Call i_findConnectedComponentRuns using ip_classifySymbol as
	 * the callback proc.  ip_classifySymbol will be called once for
	 * each connected component found in the image.
	 */
    __error__ = i_findConnectedComponentRuns(
				pIPWState,
				pImage,
				bpl,
				cFrameBits,		/* clipX */
				cFrameBits,		/* clipY */
				width,			/* clipW */
				height,			/* clipH */
				1,			/* minWidth */
				1,			/* minHeight */
				1,			/* 8 connected */
				(void *)pTextCompInfo,	/* pRectData */
				&ip_classifySymbol,	/* rectReturnProc */
				&classifySymRetCode);	/* pRectReturnCode */

    if (__error__ != ia_successful)	/* something bad happened */
    {
	if (__error__ == ia_callbackError)	/* ip_classifySym complained */
	    __error__ = classifySymRetCode;
	goto finConvertImageToSym;
    }

	/* That's ALMOST it.  Most of the real action happened
	 * in ip_classifySymbol.  We need to flush out the last few
	 * uncommitted positions in the uncommitted list if the
	 * list isn't empty now (which I don't think can happen)  
	 * Set the number of bytes we actually used.  Also set the append
	 * one zero for y offset 0f 0 (=> end of position buf info)
	 */
    if (pTextCompInfo->numUncomPos > 0)
    {
	for (index = pTextCompInfo->lowUncomY;
	     index <= pTextCompInfo->hiUncomY;
	     index++)
	{
	    __error__ = ip_commitTCPositions (pIPState, pTextCompInfo, index);
	    if (__error__ != ia_successful)
		goto finConvertImageToSym;	/* don't bother continuing */
	}
    }

    if (pTextCompInfo->positionBufSize - pTextCompInfo->resultSize >= 1)
    {
	pTextCompInfo->pPositionBuf[0] = 0;
	*pResultSize = pTextCompInfo->resultSize + 1;
    }
    else
	__error__ = ia_bufferExceeded;	/* ran out of space */

finConvertImageToSym:
	/* return the final return code */
    NORMAL_RETURN;
}




/* PRIVATE */
/**************************************************************************
 *   ip_classifySymbol()
 *
 *	This procedure is called by i_findConnectedComponentRuns for
 *	each connected component in the image.
 *
 *	Input:
 *		pTextCompInfoIn:	Pointer to image's textCompInfo.
 *		x, y, w, h:		symbols's bounding box
 *		symMass:		mass of symbol
 *		nRuns:			number of runs in symbol
 *		pRuns:			Pointer list of runs associated
 *					with the symbol.
 *
 *	Action:
 *		Attempts to classify the symbol passed in as one of the
 *		existing symbols.  If this is not possible, expands the
 *		shape tree as needed and creates a new symbol.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *		ia_bufferExceeded	Ran off the end of the position buf
 *
 **************************************************************************/
Int32  CDECL 
ip_classifySymbol (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	symMass,
    Int32	nRuns,
    TCScanLineRun *pRuns)
{
PROLOGUE;
 TCCompressionInfoPtr	 pTextCompInfo;	/* pointer to image's information */
TCScanLineRun		*pCurRun, *pTempRun;	/* temporary run pointers */
TCInstance		 instance;	/* struct containing instance's info */
TCSymbol		*pExemSymbol;	/* pointer to current exemplar */
TCSymbol		*pMatchedSymbol; /* pointer to symbol that matches
					 * the instance. */
Int32			 xOrigin;	/* x origin of symbol */
UInt32			 lineFlags;	/* flags used to tell the runs in
					 * one line from the runs in
					 * adjacent lines */
Int32			 searchedBigs;	/* Boolean, cTrue => we've already
					 * looked at all the big symbols,
					 * no need to look again. */
Int32			 index;		/* index into searchW and searchH */
    /* The following arrays are used to generate the offsets for the search
     * through the 2D array of pointers to exemplars.  The arrays is
     * indexed by [width][height].  It has a maximum dimension of
     * cMaxSymIndexDim in either dimension.  Pointers to symbols having
     * either a width > cMaxSymIndexDim or height > cMaxSymIndexDim are
     * stored in
     * (pTextCompInfo->pSymIndex->array)[cMaxSymIndexDim][cMaxSymIndexDim].
     * Isn't that a mouthful?  It means pSymIndex is a member of the
     * structure pointed to by pTextCompInfo.  pSymIndex is itself a pointer
     * to a structure containing a 2D array of pointers to TCSymbols.
     * So, (pTextCompInfo->pSymIndex->array) is the thing we want to index.
     * We attempt to search for exemplars with the same width and height
     * as the instance first, then spiral out from there.  The order is
     * illustrated in the following 5x5 array.  The central coordinate
     * is (0,0).  We look for exemplars within +-2 in width and height.
     *
     *    w  -2 -1  0  1  2
     *   h ----------------
     *	-2 | 22 18 12 19 23
     *  -1 | 14  6  4  7 15
     *   0 | 10  2  1  3 11
     *   1 | 16  8  5  9 17
     *   2 | 24 20 13 21 25
     */
static Int32		 searchW[25] = { 0,	/* step 1 */
					-1,	/* step 2 */
					+1,	/* step 3 */
					 0,	/* step 4 */
					 0,	/* step 5 */
					-1,	/* step 6 */
					+1,	/* step 7 */
					-1,	/* step 8 */
					+1,	/* step 9 */
					-2,	/* step 10 */
					+2,	/* step 11 */
					 0,	/* step 12 */
					 0,	/* step 13 */
					-2,	/* step 14 */
					+2,	/* step 15 */
					-2,	/* step 16 */
					+2,	/* step 17 */
					-1,	/* step 18 */
					+1,	/* step 19 */
					-1,	/* step 20 */
					+1,	/* step 21 */
					-2,	/* step 22 */
					+2,	/* step 23 */
					-2,	/* step 24 */
					+2};	/* step 25 */
static Int32		 searchH[25] = { 0,	/* step 1 */
					 0,	/* step 2 */
					 0,	/* step 3 */
					-1,	/* step 4 */
					+1,	/* step 5 */
					-1,	/* step 6 */
					-1,	/* step 7 */
					+1,	/* step 8 */
					+1,	/* step 9 */
					 0,	/* step 10 */
					 0,	/* step 11 */
					-2,	/* step 12 */
					+2,	/* step 13 */
					-1,	/* step 14 */
					-1,	/* step 15 */
					+1,	/* step 16 */
					+1,	/* step 17 */
					-2,	/* step 18 */
					-2,	/* step 19 */
					+2,	/* step 20 */
					+2,	/* step 21 */
					-2,	/* step 22 */
					-2,	/* step 23 */
					+2,	/* step 24 */
					+2};	/* step 25 */
    /* Variables used during symbol consolidation. */
TCSymbol		*pCurSymbol;	/* pointer to current symbol */


	/* Keep the compiler happy */
    nRuns = nRuns;

	/* point to the real text compression info */
    pTextCompInfo = (TCCompressionInfoPtr) pTextCompInfoIn;

	/* Initialize these pointers so we know if we have to free
	 * stuff they point to.
	 */
    instance.pVertRuns = instance.pDupHorizRuns = instance.pDupVertRuns = NULL;

	/* find the x origin of the symbol */
    xOrigin = 2*x + w - 1;		/* get the "average" x position.
					 * The width is the distance from
					 * the left edge of the leftmost
					 * pixel to the right edge of the
					 * rightmost pixel.  Since we want
					 * the distance to the middle of
					 * the character, we want x + w/2 - 1/2.
					 * Since all coordinates keep an
					 * extra bit of precision, we use
					 * 2x + w - 1.
					 */

	/* subtract off the frame to reduce my confusion.  We have to do
	 * this after the stuff above this so that the values of x and
	 * xOrigin are consistent.
	 */
    x -= cFrameBits;
    y -= cFrameBits;

	/* At this point, we want to find the *real* width and height of
	 * the symbol.  The w and h parmeters passed in match the runs
	 * passed in.  If we are consolidating symbols, the symbol we're
	 * looking at might have had a run added to the bottom or left.
	 * This will confuse us when we're trying to apply the width
	 * and height criteria for matching symbols.
	 * Remember that symbols coords are symmetric around 0, so
	 * pSymbol->maxX = -pSymbol->minX = w - 1 and pSymbol->maxY =
	 * -pSymbol->minY = h - 1.
	 */
    pCurSymbol = NULL;
    if (pTextCompInfo->classifyOnly)
    {
	pCurSymbol = pTextCompInfo->pClassifiedSymbol;
	if (pCurSymbol->pClassSymbol != NULL)
	{
		/* This symbol is part of an equivalence class.  Use
		 * the recorded minimum and maximum widths and heights
		 * of all symbols in the class.
		 */
	    instance.minWidth = pCurSymbol->pClassSymbol->minX;
	    instance.maxWidth = pCurSymbol->pClassSymbol->maxX;
	    instance.minHeight = pCurSymbol->pClassSymbol->minY;
	    instance.maxHeight = pCurSymbol->pClassSymbol->maxY;
	}
	else	/* this symbols isn't part of an equivalence class */
	{
	    instance.minWidth = instance.maxWidth = pCurSymbol->maxX + 1;
	    instance.minHeight = instance.maxHeight = pCurSymbol->maxY + 1;
	}
    }
    else	/* This is the first time we've seen this symbol,
		 * so minWidth and maxWidth are both equal to w.
		 * Ditto for minHeight, maxHeight and h.
		 */
    {
	instance.minWidth = instance.maxWidth = w;
	instance.minHeight = instance.maxHeight = h;
    }

	/* Record the instance's position and dimensions as passed in */
    instance.x = x;
    instance.y = y;
    instance.w = w;
    instance.h = h;

	/* Adjust all the x coordinates inside the runs so they're
	 * relative to the xOrigin.  Process the
	 * very first real run outside the loop so that
	 * pPrev will point to something meaningful inside the loop.
	 * Note that we keep an extra bit of significance below
	 * the binary point, i.e. a binary coordinate of 0x2 is really
	 * +1, a binary coordinate of 0x3 is really +1.5.
	 * Also mark the runs of each scan line with a 2 bit tag - 10 or 01.
	 * The bits alternate each time we cross a scan line boundary
	 * We use a 2 bit tag so a dummy having neither bit set can be
	 * used to terminate either type of run.
	 * Also calculate the mass of the symbol.  This is another
	 * discrimination criteria.
	 * Note that if we're classifying a known symbols, we'll use
	 * its mass, slant, strokeWidth, volume and signed asymmetries, so
	 * we don't need to compute them.
	 */
    instance.pCurSymbol = pCurSymbol;	/* record ptr to existing symbol,
					 * if any, for ip_compareSymbols */
    if (pCurSymbol == NULL)
    {
	    /* there is no known symbol, so calculate all this stuff
	     * from the runs.
	     */
	pTempRun = pRuns;
	lineFlags = cEvenLineTag;
	pTempRun->runLeft = (Int16) (2 * pTempRun->runLeft - xOrigin);
	pTempRun->runRight = (Int16) (2 * pTempRun->runRight - xOrigin);
	instance.mass = symMass * 2;
	pTempRun->pClient = (void *)lineFlags;
	for (pCurRun = pTempRun->pNext;
	     pCurRun != NULL;
	     pCurRun = pCurRun->pNext)
	{
	    pCurRun->runLeft = (Int16) (2 * pCurRun->runLeft - xOrigin);
	    pCurRun->runRight = (Int16) (2 * pCurRun->runRight - xOrigin);
	    if (pCurRun->runRight >= pTempRun->runLeft - 2)
		lineFlags ^= cLineTagMask;	/* swap the flags */
	    pCurRun->pClient = (void *)lineFlags;
	    pTempRun = pCurRun;
	}

    }
    else
    {
	    /* The symbol is known.  The runs are already correct, so we
	     * don't have to change them.  All we have to do is read
	     * the mass at this point.
	     */
	instance.mass = pCurSymbol->mass;
    }

	/* In either case, duplicate the horizontal runs in the first
	 * scan line.  Actually, put one bit in the middle of each run
	 * so we get the maximum tolerance of noise.  These extra runs
	 * are used to move the symbol up one line.
	 */
    __error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
				(void **)&(instance.pDupHorizRuns));
    if (__error__ != ia_successful)
	goto tcClassifySymCleanup;

    instance.pDupHorizRuns->runLeft = (pRuns->runLeft + pRuns->runRight)/2;
    instance.pDupHorizRuns->runRight = instance.pDupHorizRuns->runLeft;
    instance.pDupHorizRuns->pClient = (void *) cOddLineTag;

	/* that duplicated the first run, duplicate any more in the
	 * first scan line.
	 */
    pCurRun = pRuns->pNext;
    pTempRun = instance.pDupHorizRuns;
    if (pCurRun != NULL)
    {
	for (;
	     pCurRun->pClient == (void *) cEvenLineTag;
	     pCurRun = pCurRun->pNext)
	{
	    __error__ = i_allocateStruct (pIPState,
					pTextCompInfo->p1stScanLineBlock,
					(void **)&(pTempRun->pNext));
	    if (__error__ != ia_successful)
		goto tcClassifySymCleanup;
    
	    pTempRun = pTempRun->pNext;
	    pTempRun->runLeft = (pCurRun->runLeft + pCurRun->runRight)/2;
	    pTempRun->runRight = pTempRun->runLeft;
	    pTempRun->pClient = (void *) cOddLineTag;
	}
    }

	/* Make the last duplicate horiz run point to the first real
	 * horizontal run.
	 */
    pTempRun->pNext = pRuns;

	/* Record the location of the horizontal runs in the instance
	 * structure.  This keeps everything tidy.
	 */
    instance.pHorizRuns = pRuns;


	/* We also ask ip_calculateSlant to get the slant of the symbol.
	 * The slant is a funny measure.  It's main purpose is to
	 * distinguish italic round things from non-italic round things
	 * (like o's and e's) because the other tests don't do a very
	 * good job at that task.  The slant attempts to represent
	 * the extent of outside edges that slope to the right
	 * versus outside edges that slope to the left.  Italic characters
	 * tend to have lines sloping to the upper right.
	 * A few things make the slant measure tricky.  There aren't many
	 * pixels in the tiny symbols in which this measure is supposed
	 * to work, so we don't have a lot of leeway.  We don't want to
	 * have to figure out if we're dealing with little italic round
	 * things when we apply the measure, we'd like to apply it on
	 * every character.  Therefore, it can't be affected (much) by
	 * noise that appears on any character.  Measures like the number
	 * of pixels in the four quadrants of the image don't work.
	 * Raw distances from the corners of the bounding box of the
	 * symbol to the nearest pixel aren't very discrminating either.
	 * The measure we've chosen keeps track of sloping outside edges.
	 * At the endpoints of the edge (where the character ends or the
	 * edge changes direction), we add the minimum of the deltaX and
	 * deltaY to the slant for positive edges and subtract it for
	 * negative edges.  This seems fairly tolerant of noise while
	 * catching real slopes.
	 * ip_calculateSlant also calculates volume and signed asymmetries.
	 * Volume is used to distinguish very small characters with
	 * small bits of noise from each other.  It is the sum of
	 * the sampled widths and heights of the symbol taken at each
	 * row and column.  It uses squarishness to distinguish 8 from B.
	 * The signed asymmetries are used to find the center of gravity
	 * of a symbol in the horizontal and vertical dimensions.  These
	 * are used to align symbols when making comparisions.
	 */
	/* Create the vertical runs too. */
    __error__ = ip_makeVerticalRuns (pIPState, pTextCompInfo,
				   w, h,
				   pRuns,
				   &(instance.pVertRuns));
    if (__error__ != ia_successful)
	goto tcClassifySymCleanup;

	/* Now, duplicate the vertical runs in the first
	 * scan column.  Actually, put one bit in the middle of each run
	 * so we get the maximum tolerance of noise.  These extra runs
	 * are used to move the symbol to the right by one line.
	 */
    __error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
				(void **)&(instance.pDupVertRuns));
    if (__error__ != ia_successful)
	goto tcClassifySymCleanup;

    instance.pDupVertRuns->runLeft =
	(instance.pVertRuns->runLeft + instance.pVertRuns->runRight)/2;
    instance.pDupVertRuns->runRight = instance.pDupVertRuns->runLeft;
    instance.pDupVertRuns->pClient = (void *) cOddLineTag;

	/* that duplicated the first run, duplicate any more in the
	 * first scan column.
	 */
    pCurRun = instance.pVertRuns->pNext;
    pTempRun = instance.pDupVertRuns;
    if (pCurRun != NULL)
    {
	for (;
	     pCurRun->pClient == (void *) cEvenLineTag;
	     pCurRun = pCurRun->pNext)
	{
	    __error__ = i_allocateStruct (pIPState,
					pTextCompInfo->p1stScanLineBlock,
					(void **)&(pTempRun->pNext));
	    if (__error__ != ia_successful)
		goto tcClassifySymCleanup;

	    pTempRun = pTempRun->pNext;
	    pTempRun->runLeft = (pCurRun->runLeft + pCurRun->runRight)/2;
	    pTempRun->runRight = pTempRun->runLeft;
	    pTempRun->pClient = (void *) cOddLineTag;
	}
    }

	/* Make the last duplicate vertical run point to the first real
	 * vertical run.
	 */
    pTempRun->pNext = instance.pVertRuns;

	/* If necessary, calculate the slant, volume and signed asymmetries.
	 * If we're dealing with a known symbol, just read them directly.
	 */
    if (pCurSymbol == NULL)
    {
	__error__ = ip_calculateSlant (pIPState, &instance);
	if (__error__ != ia_successful)
	    goto tcClassifySymCleanup;
    }
    else	/* get get them */
    {
	instance.slant = pCurSymbol->slant;
	instance.maxDimension = pCurSymbol->maxDimension;
	instance.checkHVolume = pCurSymbol->checkHVolume;
	instance.checkVVolume = pCurSymbol->checkVVolume;
	instance.hVolume = pCurSymbol->hVolume;
	instance.vVolume = pCurSymbol->vVolume;
	instance.hSAsymmetry = pCurSymbol->hSAsymmetry;
	instance.vSAsymmetry = pCurSymbol->vSAsymmetry;
	if (pCurSymbol->pClassSymbol == NULL)
	    instance.strokeWidth = pCurSymbol->strokeWidth;
	else
	    instance.strokeWidth = pCurSymbol->pClassSymbol->strokeWidth;
    }

	/* Start comparing the instance symbol against the exemplars.  Note
	 * that pExemSymbol only gives us a starting point, we need to
	 * decide when to stop on our own.  We also stop if we found a match.
	 * Note that all symbols whose width or height > cMaxSymIndexDim
	 * are in the list pointed to by pTextCompInfo->pBigSymbols.
	 * The variable searchedBigs tells us whether we've already
	 * looked at that list.
	 */
    searchedBigs = cFalse;
    pMatchedSymbol = NULL;

    for (index = 0; index < 25; index++)
    {
	if ((searchW[index]+w <= cMaxSymIndexDim) &&
	    (searchW[index]+w > 0) &&
	    (searchH[index]+h <= cMaxSymIndexDim) &&
	    (searchH[index]+h > 0))
	{
	    pExemSymbol =
		(pTextCompInfo->pSymIndex->array)
					[searchW[index]+w][searchH[index]+h];
	    if (pExemSymbol != NULL)
	    {
		__error__ = ip_compareSymbols (pIPState, pTextCompInfo,
					     &instance,
					     pExemSymbol,
					     &pMatchedSymbol);
		if (__error__ != ia_successful)
		    goto tcClassifySymCleanup;
	
		if (pMatchedSymbol != NULL)
		    break;	/* found a match, so quit now */
	    }
	}
	else	/* one of the indices was too big, so see if we can
		 * use the pointer to big symbols */
	if ((!searchedBigs) && (pTextCompInfo->pBigSymbols != NULL))
	{
	    __error__ = ip_compareSymbols (pIPState, pTextCompInfo,
					 &instance,
					 pTextCompInfo->pBigSymbols,
					 &pMatchedSymbol);
	    if (__error__ != ia_successful)
		goto tcClassifySymCleanup;
    
	    if (pMatchedSymbol != NULL)
		break;		/* found a match, so quit now */

		/* Didn't find the exemplar here, so don't look again */
	    searchedBigs = cTrue;
	}
    }

	/* If we found a match, record the symbol's ID in the position
	 * info and leave.  Otherwise, go make up a new symbol.
	 * Which choice is taken depends on whether pMatchedSymbol
	 * is NULL (no matching symbol) or not (found a match).
	 */
    __error__ = ip_symbolIsComplete (pIPState, pTextCompInfo,
				   pMatchedSymbol,
				   &instance,
				   cFalse);

	/* if things went fine, __error__ equals ia_successful at this
	 * point.  If not, it doesn't.  In either case, free the
	 * runs allocated before returning.
	 */
tcClassifySymCleanup:
	/* First, get rid of the duplicate horizontal runs */
    pCurRun = instance.pDupHorizRuns;
    while ((pCurRun != NULL) && (pCurRun != pRuns))
    {
	pTempRun = pCurRun->pNext;
	i_freeStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
		      (void **) &pCurRun);
	pCurRun = pTempRun;
    }
	/* now, get rid of the vertical runs.  If we've already
	 * made the duplicate vertical runs, start there.  Otherwise,
	 * start with the standard vertical runs.  The list of
	 * duplicate vertical runs includes the standard vert runs.
	 * The runs pointed to by pRuns will be freed by the caller,
	 * so don't bother with that.
	 * We haven't allocated a symbol here, so there's no need to free
	 * that.
	 */
    if (instance.pDupVertRuns != NULL)
	pCurRun = instance.pDupVertRuns;
    else
	pCurRun = instance.pVertRuns;

    while (pCurRun != NULL)
    {
	pTempRun = pCurRun->pNext;
	i_freeStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
		      (void **) &pCurRun);
	pCurRun = pTempRun;
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_compareSymbols()
 *
 *	This procedure is called by ip_classifySymbols to match an instance
 *	against a list of exemplars.  It does lossy matches with symbols
 *	already in the dictionary.
 *
 *	Input:
 *		pTextCompInfo:		Pointer to text compression info
 *		pInstance:		Pointer to info about instance.
 *		pExemSymbol:		Pointer to first potential exemplar
 *		ppMatchedSymbol:	Pointer to pointer to exemplar that
 *					matches the instance.
 *
 *	Action:
 *		Compares the instance against each of the exemplars in the
 *		list.  If one matches, it's pointer is put into
 *		*ppMatchedSymbol.  Otherwise, *ppMatchedSymbol is unaffected.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_internal		internal failure
 *
 **************************************************************************/
Int32  CDECL 
ip_compareSymbols (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCInstance *pInstance,
    TCSymbol   *pExemSymbol,
    TCSymbol  **ppMatchedSymbol)
{
    PROLOGUE;
    TCScanLineRun *pCurRun;	/* points to current scan line run */
    TCScanLineRun *pTempRun;	/* points to previous scan line run */
    TCGroupRun *pExemGroup;	/* points to current group from the
				 * exemplar. */
    Int32	exemHeight;
    Int32	exemWidth;
    UInt32	lineTagMask;	/* used to invert line tags when
				 * matching mis-aligned instances and
				 * exemplars.
				 */
    Int32	matchCount;	/* = 0 when matching horizontal runs
				 *     and groups;
				 * = 1 when matching vertical runs
				 *     and groups. */
	/* Variables used during symbol consolidation. */
    TCSymbol   *pTempSymbol;	/* used to see if exemplar is already
				 * equivalent to existing symbol */
    Int32	lastDiffRight;	/* last difference of right edge
				 * coords. */
    Int32	lastDiffLeft;	/* last difference of left edge
				 * coords. */
    Int32	lastDiffChangeLeft;	/* the last non-zero change
					 * in differences we added to
					 * totalDiffLeft */
    Int32	lastDiffChangeRight;	/* the last non-zero change
					 * in differences we added to
					 * totalDiffRight */
    Int32	totalDiffLeft;	/* Sum of difference changes on
				 * left side of sym (sort of). */
    Int32	totalDiffRight; /* Sum of difference changes on
				 * right side of sym (sort of). */
    

    pTextCompInfo = pTextCompInfo;	/* in case TC_DEBUG is turned off */
    __error__ = ia_successful;		/* assume the best */
    while (pExemSymbol != NULL)
    {
	    /* It's been determined that OCR accuracy improves a lot if
	     * we don't allow very small things to get bigger.
	     * The definition of "small" is any symbol with either
	     * dimension less than cNonGrowingSymbolSize.
	     */
	if (MIN (pInstance->w, pInstance->h) < cNonGrowingSymbolSize)
	    {
		    /* What to do if we're consolidating symbols?
		     * How to we choose between a symbol with a smaller
		     * width and a larger height?  When merging symbol
		     * classes during symbol consolidation,
		     * we don't yet know which symbol in a merged
		     * equivalence class will become the exemplar.
		     * In particular, pExemSymbol may not be the exemplar
		     * for the merged class, just a member of it.
		     * The way we solve this dilemma is to insist
		     * that all members of all merged classes have
		     * the same dimensions.  This seems suboptimal,
		     * and from a pure compression point of view, it
		     * is, but the tiny compression loss seems to
		     * be offset by a not-quite-tiny improvment in
		     * OCR accuracy.
		     */
		if (pInstance->pCurSymbol != NULL)
		{
		if ( (pInstance->w != pExemSymbol->width) ||
		     (pInstance->h != pExemSymbol->height) )
			{
#if TC_DEBUG
			    pTextCompInfo->sizeCheck++;
#endif
			    goto tcTryNextSym;
			}
		}
		else 	/* This is not symbol consolidation */
		{
		    if (((pInstance->w < cNonGrowingSymbolSize) &&
			 (pInstance->w < pExemSymbol->width)) ||
			((pInstance->h < cNonGrowingSymbolSize) &&
			 (pInstance->h < pExemSymbol->height)))
		    {
#if TC_DEBUG
			pTextCompInfo->sizeCheck++;
#endif
			goto tcTryNextSym;
		    }
		}
	    }


	    /* Test to see if the exemplar is too wide or too thin.
	     * Notice that we're only checking to see if the range of
	     * symbols represented by the new symbol (normally only
	     * itself, but possibly an entire equivalence class) fits
	     * inside the range allowed by pExemSymbol (or its class).
	     * Should we also check that all the symbols in pExemSymbol's
	     * class fit inside the new symbol's class?  I think not.
	     * If all of the symbols represented by the new symbol fit
	     * inside the dimension range of all of pExemSymbol's class,
	     * then the only way all of pExemSymbol's symbols wouldn't
	     * fit the other way would be if the new symbol's class had
	     * a smaller range of acceptable dimensions.  The difference
	     * in ranges of symbols with almost the same (within 2)
	     * dimensions is at most 1.  So, it could happen.  For
	     * example, a symbol of width 19 might match only symbols
	     * of width 19 while a symbol of width 20 might match symbols
	     * of width 19, 20 or 21.  While this is theoretically a problem,
	     * I figure that it's not a big problem.
	     */
	if (pExemSymbol->pClassSymbol == NULL)
	{
	    if ((pInstance->minWidth < pExemSymbol->minWidth) ||
		(pInstance->maxWidth > pExemSymbol->maxWidth) ||
		(pInstance->minHeight < pExemSymbol->minHeight) ||
		(pInstance->maxHeight > pExemSymbol->maxHeight))
	    {
#if TC_DEBUG
		pTextCompInfo->sizeCheck++;
#endif
		goto tcTryNextSym;
	    }
		/* Check the stroke width */
	    if (ABS(pInstance->strokeWidth - pExemSymbol->strokeWidth) >
		cStrokeWidthTolerance)
	    {
    #if TC_DEBUG
		pTextCompInfo->strokeWidthCheck++;
    #endif
		goto tcTryNextSym;
	    }
	}
	else
	{
	    if ((pInstance->minWidth < pExemSymbol->pClassSymbol->minWidth) ||
		(pInstance->maxWidth > pExemSymbol->pClassSymbol->maxWidth) ||
		(pInstance->minHeight < pExemSymbol->pClassSymbol->minHeight) ||
		(pInstance->maxHeight > pExemSymbol->pClassSymbol->maxHeight))
	    {
#if TC_DEBUG
		pTextCompInfo->sizeCheck++;
#endif
		goto tcTryNextSym;
	    }

		/* Check the stroke width */
	    if (ABS(pInstance->strokeWidth -
				pExemSymbol->pClassSymbol->strokeWidth) >
		cStrokeWidthTolerance)
	    {
    #if TC_DEBUG
		pTextCompInfo->strokeWidthCheck++;
    #endif
		goto tcTryNextSym;
	    }
	}

	exemHeight = pExemSymbol->maxY + 1;
	exemWidth = pExemSymbol->maxX + 1;

	    /* If the slant isn't in range, try another symbol.  Here, we
	     * compare the different in slants against a fraction of the
	     * larger of the width and height.  This catches errors on the
	     * small round things where we were having trouble without
	     * spuriously rejecting comparisons between tall thin things.
	     *
	     * I thought that slant was a picky little measure, but I
	     * started having problems with larger italic characters,
	     * so we test it all the time now.
	     */
	if (pExemSymbol->slant > pInstance->slant)
	{
	    if (pExemSymbol->slant - pInstance->slant >
				MAX(exemHeight, exemWidth)/cSlantSlopFactor)
	    {
#if TC_DEBUG
		pTextCompInfo->slantCheck++;
#endif
		goto tcTryNextSym;
	    }
	}
	else
	{
	    if (pInstance->slant - pExemSymbol->slant >
				MAX(exemHeight, exemWidth)/cSlantSlopFactor)
	    {
#if TC_DEBUG
		pTextCompInfo->slantCheck++;
#endif
		goto tcTryNextSym;
	    }
	}

	    /* Only test the picky little measures for little symbols.
	     */
	if (pInstance->maxDimension < cLittleSymbolSize)
	{

		/* Check the horizontal and vertical volumes only if
		 * each is a sufficiently large fraction of the total area.
		 * Note that either the instance or the symbol can demand
		 * a check of the volume.
		 */
	    if (pInstance->checkHVolume | pExemSymbol->checkHVolume)
	    {
		    /* The type of check is that a change in volume should
		     * follow a change in dimension.  That is, if the
		     * instance is wider, it's horizontal volume should
		     * also be larger.  Similarly, if the widths are the
		     * same, the difference in horizontal volumes should
		     * be small, say the larger of the two heights.
		     */
		if (((pInstance->w < exemWidth) &&
		     (pInstance->hVolume > pExemSymbol->hVolume)) ||
		    ((pInstance->w > exemWidth) &&
		     (pInstance->hVolume < pExemSymbol->hVolume)))
		{
#if TC_DEBUG
		    pTextCompInfo->volumeCheck++;
#endif
		    goto tcTryNextSym;
		}
	    }

		/* Same for the vertical volume */
	    if (pInstance->checkVVolume | pExemSymbol->checkVVolume)
	    {
		if (((pInstance->h < exemHeight) &&
		     (pInstance->vVolume > pExemSymbol->vVolume)) ||
		    ((pInstance->h > exemHeight) &&
		     (pInstance->vVolume < pExemSymbol->vVolume)))
		{
#if TC_DEBUG
		    pTextCompInfo->volumeCheck++;
#endif
		    goto tcTryNextSym;
		}
    
	    }

	}
	    /* Now that we've tried all the easy ways to disqualify
	     * a symbol, try the harder ones.  First, if we're
	     * consolidating symbols, don't look at two symbols
	     * if they're already in the same equivalence class.
	     * One way they could be in the same class is to be
	     * the same symbol, so check that first.
	     */
	if (pInstance->pCurSymbol != NULL)
	{
	    if (pInstance->pCurSymbol == pExemSymbol)
		goto tcTryNextSym;
	    else
	    if (pInstance->pCurSymbol->pNewSymbol != NULL)
	    {
		for (pTempSymbol = pInstance->pCurSymbol->pNewSymbol;
		     pTempSymbol != pInstance->pCurSymbol;
		     pTempSymbol = pTempSymbol->pNewSymbol)
		{
		    if (pTempSymbol == pExemSymbol)
			goto tcTryNextSym;
		}
	    }
	}

	    /* Try matching both the horizontal and vertical runs and groups.
	     * Note that in the code inside the loop, we talk about scan
	     * lines and runs.  When matching vertical stuff, scan lines
	     * correspond to image columns, but everything is logically
	     * the same.
	     *
	     * Also try to align the instance rows with the exemplar
	     * rows.  Let's illustrate this with horizontal runs, but
	     * the same principle applies to vertical runs.
	     * If the exemplar is as tall or taller than the instance
	     * and the exemplar is shifted up in its bounding box with
	     * respect to the instance, use the duplicate version of
	     * the first row of the instance.  This shifts it up so it's
	     * aligned with the exemplar.  Why not just ignore the first
	     * exemplar line?  I'm afraid of deleting any useful info
	     * - it seems less dangerous to add stuff than to take it away.
	     *
	     * Similarly, if the instance is as tall or taller than the
	     * exemplar and is shifted up in its bounding box, use the
	     * duplicate group of the exemplar to shift it up too.
	     * Remember that in the vertical runs, the positive coords
	     * are at the bottom of the symbol, so a symbol is shifted
	     * up if its vertical asymmetry is more negative.
	     *
	     * Here's the tricky part.  How do we decide if something
	     * is shifted?  I started by just using the signed asymmetries,
	     * but that didn't work.  A symbol can appear to be shifted
	     * to the left inside its bounding box either because it's
	     * missing a column of stuff on the left or because it has
	     * extra noise on the right.  All shifts are done by adding
	     * columns on the left.  We only want to add a column on the
	     * left if it was missing a column on the left, not if it
	     * had extra stuff on the right.
	     *
	     * Therefore, we calculate the distance from the symbol's
	     * central vertical axis (as determined by the symbol's
	     * envelope, not all its pixels to avoid some influence of
	     * stroke width) to its left side.  We look at the difference
	     * in these distances between instance and exemplar when
	     * deciding if one should be shifted relative to the other.
	     * A possibly more exact way would be to break the calculation
	     * of the central vertical axis into two independent parts -
	     * the average postion of the left edge and the average position
	     * of the right edge.  We would record the both distances
	     * to the far left column of the symbol.  We'd only shift a
	     * symbol if both the right and left edges looked like they
	     * should be shifted by the same direction and by the same
	     * amount.  This is more complex, so we'll try to avoid
	     * it for now.
	     *
	     * There's a little funny business about the representation
	     * of the distances.  We want to know if we should move over
	     * by 1 column.  If the distances were calculated so that
	     * they changed by 1 for each scan column:
	     *
	     *	exemDist-instanceDiff	Action
	     *	< -2			fail, symbols can't be aligned
	     *	[-1.99, -1]		shift exemplar right by 1 col
	     *	[-0.99, +0.99]		instance and exemplar are aligned
	     *	[1, 1.99]		shift instance right by 1 col
	     *	> 2			fail, symbols can'be be aligned
	     *
	     * However, it's not quite this simple.  First, the central
	     * axes are calculated by averaging the endpoints of the
	     * runs expressed in 2x coords, so all of the numbers above
	     * should be multiplied by 2.  We also kept three extra bits
	     * of precision, so multiply all the numbers by another factor
	     * of 2, a total multiplier of 4.  Also, round down to the
	     * nearest integer.  The real table is:
	     *
	     *	exemDist-instanceDiff	Action
	     *	< -8			fail, symbols can't be aligned
	     *	[-7, -4]		shift exemplar right by 1 col
	     *	[-3, +3]		instance and exemplar are aligned
	     *	[4, 7]		shift instance right by 1 col
	     *	> 8			fail, symbols can'be be aligned
	     *
	     * Whenever we use either of the shifts, we have to invert
	     * the sense of the test used to match instance lines with
	     * exemplar line.  That's because we've just mis-aligned
	     * the instance with the exemplar.  This variable is
	     * lineTagMask.
	     *
	     * Almost.  Noise which displaces a symbol horizontally can
	     * greatly contribute to the signed vertical asymmetry.  Ditto
	     * for vertical noise and the signed horizontal asymmetry.
	     * For example, if the noise is at the upper left or right
	     * corner of a symbol, the signed vertical asymmetry goes
	     * negative by a bunch.  If the noise is at either or both of
	     * the lower corners, the signed vertical asymmetry goes positive.
	     * To help lower the effect of this, we compare the difference
	     * in asymmetries to w+h instead of just w or h.  We really
	     * installed this test to align tall, thin things horizontally,
	     * so adding a small w should still be ok for that case.
	     *
	     * Well, hey.  Even if it looks like the symbols can't be
	     * aligned, try anyway.
	     */
	totalDiffLeft = totalDiffRight = 0;	/* no differences yet */
	for (matchCount = 0; matchCount < 2; matchCount++)
	{
	    lineTagMask = 0;	/* assume instance and exemplar are aligned */
	    lastDiffChangeLeft = lastDiffChangeRight = 0;
	    lastDiffRight = lastDiffLeft = 0;	/* start with no diff in
						 * endpoints */

		/* assume the instance and exemplar are aligned. */
	    if (matchCount == 0)
	    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "Starting Horiz compare for sym %ld at %ld, %ld\n",
	pExemSymbol->symbolID, pInstance->x, pInstance->y);
#endif

#if 0
		if (ABS(pInstance->vSAsymmetry - pExemSymbol->vSAsymmetry) >= 8)
		{
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "Symbol %ld is too far out of vertical alignment, symbol's vert asymmetry = %ld, instance's vert asymmetry = %ld\n",
	pExemSymbol->symbolID, pExemSymbol->vSAsymmetry, vSAsymmetry);
#endif
		    goto tcTryNextSym;
		}
#endif

		if ((pInstance->h >= exemHeight) &&
		    (pInstance->vSAsymmetry - pExemSymbol->vSAsymmetry >= 4))
		{
			/* the exemplar should be moved up */
		    pExemGroup = pExemSymbol->pDupHorizGroup;
		    pCurRun = pInstance->pHorizRuns;
		    lineTagMask = cLineTagMask;
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "Moving exemplar up by one line\n");
#endif
		}
		else
		if ((exemHeight >= pInstance->h) &&
		    (pExemSymbol->vSAsymmetry - pInstance->vSAsymmetry >= 4))
		{
			/* the instance should be moved up */
		    pExemGroup = pExemSymbol->p1stHorizGroup;
		    pCurRun = pInstance->pDupHorizRuns;
		    lineTagMask = cLineTagMask;
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "Moving instance up by one line\n");
#endif
		}
		else	/* they are already aligned */
		{
		    pExemGroup = pExemSymbol->p1stHorizGroup;
		    pCurRun = pInstance->pHorizRuns;
		}
	    }
	    else
	    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "Starting Vert compare for sym at %ld, %ld\n", pInstance->x, pInstance->y);
#endif

#if 0
		if (ABS(pInstance->hSAsymmetry - pExemSymbol->hSAsymmetry) >= 8)
		{
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "Symbol %ld is too far out of horizontal alignment, symbol's horiz asymmetry = %ld, instance's horiz asymmetry = %ld\n",
	pExemSymbol->symbolID, pExemSymbol->hSAsymmetry, hSAsymmetry);
#endif
		    goto tcTryNextSym;
		}
#endif

		if ((pInstance->w >= exemWidth) &&
		    (pInstance->hSAsymmetry - pExemSymbol->hSAsymmetry >= 4))
		{
			/* the exemplar should be moved right */
		    pExemGroup = pExemSymbol->pDupVertGroup;
		    pCurRun = pInstance->pVertRuns;
		    lineTagMask = cLineTagMask;
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "Moving exemplar right by one line\n");
#endif
		}
		else
		if ((exemWidth >= pInstance->w) &&
		    (pExemSymbol->hSAsymmetry - pInstance->hSAsymmetry >= 4))
		{
			/* the instance should be moved right */
		    pExemGroup = pExemSymbol->p1stVertGroup;
		    pCurRun = pInstance->pDupVertRuns;
		    lineTagMask = cLineTagMask;
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "Moving instance right by one line\n");
#endif
		}
		else	/* they are already aligned */
		{
		    pExemGroup = pExemSymbol->p1stVertGroup;
		    pCurRun = pInstance->pVertRuns;
		}
	    }

		/* pCurRun is now pointing to the first instance run 
		 * we want to inspect.  pExemGroup is now pointing to
		 * the first group in the first exemplar line.  Go through
		 * the lines, comparing instance runs with exemplar groups.
		 * Each exemplar group has information about its neighbors
		 * vertically and allowable endpoints.  If there's a mismatch,
		 * quit this loop.
		 * Note that we don't have to worry about scan line boundaries
		 * in the instance or the exemplar.  Each run and group is
		 * tagged with line tags.  These will match in all valid
		 * comparisons.  Ideally, we cruise through both lists,
		 * matching them as we go along.
		 * How does the matching really work?  There are several
		 * types of groups.  Each one does a particular type of
		 * matching against runs and consumes one or more runs.
		 * The types are:
		 *
		 *    * cCheckNoise
		 *	At the edges of symbols, we allow noise.  These
		 *	groups are used to make sure that if noise does
		 *	exist there, it is confined to legal areas.  To
		 *	allow us to use this type of run for two purposes,
		 *	only consume runs that are entirely to the right
		 *	of the group's discardBdry.
		 *
		 *    * cCheck1Run
		 *	Most groups represent isolated vertical strokes.
		 *	This type of group expects to find exactly one run
		 *	in this range.  This group checks the endpoint of
		 *	the next run.  If they match, we consume the run
		 *	and continue.  If they fail, fall through to
		 *	cCheckEndpoints.  Huh?  Well, we don't really
		 *	mind if there happens to be a gap in the stroke,
		 *	it`s just faster to check for exactly 1 run.
		 *	If we fail, cCheckEndpoint will try a little
		 *	harder.
		 *
		 *    *	cCheckEndpoints
		 *	Near forks and joins, we allow the scan lines at
		 *	which the fork or join happens to move up or down.
		 *	This type of group expects to find a run matching
		 *	the right endpoint range, some number of runs in
		 *	the middle and a run (possibly the same run as the
		 *	first one) matching the left endpoint range.  All
		 *	runs are consumed.
		 *
		 *    * cCheck1RunNoAdvance
		 *	Sometimes, the run on the top or bottom of a symbol
		 *	is the only one connecting two strokes.  We use
		 *	this type of group to check to make sure the gap is
		 *	bridged.  The gap is usually made up of some space
		 *	under the stroke on the right, the gap, then some
		 *	space under the stroke on the left.  We use a
		 *	cCheckNoise group to consume any runs contained
		 *	entirely in the space to the right.  A 
		 *	cCheck1RunNoAdvance group checks the gap and another
		 *	cCheckNoise checks for noise on the left.  This
		 *	continues for all gaps.
		 *
		 * There are actually two groups for each type mentioned
		 * above.  One consumes scan lines on even runs and one
		 * consumes scan lines on odd runs.  Yeah, we could just
		 * use a variable to check the scan line polarity, but
		 * I think this may be faster.
		 *
		 * How does the matching process for each type of group
		 * really work?
		 *
		 * cCheckNoise:
		 *   1.	If the current run pointer is NULL or the run it
		 *	points to is on another scan line or is on the
		 *	same scan line but doesn't overlap this group,
		 *	Pass.  We just go on to the next group.
		 *   2.	If the runs overlaps this group, make sure it
		 *	is covered by the group.
		 *   3.	Consume all runs to the right of discardBdry.
		 *
		 * cCheck1Run:
		 *   1.	If the current run pointer is not NULL and the run it
		 *	points to is on the right scan line and has
		 *	endpoints inside the range of legal endpoints,
		 *	go on to the next group and next run.  Otherwise,
		 *	fall through to cCheckEndpoints.
		 *
		 * cCheckEndpoints:
		 *   1.	If the current run pointer is NULL or the run it
		 *	points to is on another scan line or is on the
		 *	same scan line but has a right endpoint not inside the
		 *	range of legal right endpoints, Fail.  That means
		 *	we try another symbol.
		 *   2.	Consume all runs that are on the same scan line
		 *	and are to the right of the leftmost run that's
		 *	to the right of the minimum left endpoint.  This
		 *	weirdness allows us to skip the maximum amount of
		 *	noise.
		 *   3. If the remaining run's left endpoint falls outside the
		 *	range of legal left endpoints, Fail.
		 *   4.	If that check succeeded, go on to the next group
		 *	and run.
		 *
		 * cCheck1RunNoAdvance:
		 *   1.	If the current run pointer is NULL or the run it
		 *	points to is on another scan line or is on the
		 *	same scan line but has endpoints not inside the
		 *	range of legal endpoints, Fail.  That means
		 *	we try another symbol.
		 *   2. If those checks succeeded, go on to the next group
		 *	BUT NOT the next run.
		 *
		 * What happens if we run out of groups before running out of
		 * instance runs?  There are enough groups to cover noise on
		 * the top of the exemplar.  If we run out of them before
		 * running out of instance runs, the instance is too tall,
		 * so we look for another symbol.
		 *
		 * What happens if we run out of instance runs before running
		 * out of groups?  Since there are extra groups, this is likely
		 * to happen.  All of the groups on the top of the exemplar
		 * are cCheckNoise groups.  If there are no instance runs
		 * there, we happily skip them.
		 *
		 * On the setting of lastDiffLeft and lastDiffRight:
		 * We really only need these on little, dense, round symbols.
		 * We assume these only have noise groups on the very
		 * top and bottom, so when we find a noise group, we just
		 * set them to the current difference, but don't update
		 * totalDiff.
		 */
	    while (pExemGroup != NULL)
	    {
		    /* Dispatch on the type of exemplar group */
		switch (pExemGroup->groupType & cGroupTypeDispMask)
		{
		case cCheckNoise:
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
							cOddLineTag) ||
			(pCurRun->runLeft > pExemGroup->rightEndMax) ||
			(pCurRun->runRight < pExemGroup->leftEndMin))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	CheckNoise: No overlap between null run and even line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	CheckNoise: No overlap between run at [%d, %d], tag %p and even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
}
#endif
			    /* There's no overlap, so just skip this group */
			pExemGroup = pExemGroup->pNext;
			break;
		    }
			/* There's overlap.  Does the run fall entirely
			 * in the correct range?
			 */
		    if (pExemGroup->leftEndMin > pCurRun->runLeft)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckNoise: Run at [%d, %d] falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

			/* Check the right endpoint */
		    if (pCurRun->runRight > pExemGroup->rightEndMax)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckNoise: Run at [%d, %d] falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }
			/* The first run is ok.  Update the difference in
			 * right endpoints.
			 */
		    if (pExemGroup->groupType & cAtRightEnd)
		    {
			if (lastDiffRight !=
			    (pCurRun->runRight - pExemGroup->rightEnd))
			{
			    lastDiffChangeRight = lastDiffRight -
						  (pCurRun->runRight -
						   pExemGroup->rightEnd);
			}
			lastDiffRight = (pCurRun->runRight -
						   pExemGroup->rightEnd);
		    }

			/* Discard any more runs to
			 * the right of discardBdry, if any.
			 */
		    if (pCurRun->runLeft >= pExemGroup->discardBdry)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckNoise: Discard run at [%d, %d] to right of boundary at %d\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->discardBdry);
#endif
			    /* if we're at the left end of the scan line,
			     * remember the difference in coords */
			if (((pCurRun->pNext == NULL) ||
			     (pCurRun->pClient != pCurRun->pNext->pClient)) &&
			    pExemGroup->groupType & cAtLeftEnd)
			{
			        /* found run at left end of scan line */
			    if (lastDiffLeft !=
				(pCurRun->runLeft - pExemGroup->leftEnd))
			    {
				lastDiffChangeLeft = lastDiffLeft -
						      (pCurRun->runLeft -
						       pExemGroup->leftEnd);
			    }
			    lastDiffLeft = (pCurRun->runLeft -
						       pExemGroup->leftEnd);
			}

			pCurRun = pCurRun->pNext;
			while ((pCurRun != NULL) &&
				(((UInt32)pCurRun->pClient ^ lineTagMask) ==
							cEvenLineTag) &&
				(pCurRun->runLeft >= pExemGroup->discardBdry))
			{
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckNoise: Discard run at [%d, %d] to right of boundary at %d\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->discardBdry);
#endif
				/* if we're at the left end of the scan line,
				 * remember the difference in coords */
			    if (((pCurRun->pNext == NULL) ||
				 (pCurRun->pClient !=
						pCurRun->pNext->pClient)) &&
				pExemGroup->groupType & cAtLeftEnd)
			    {
				    /* found run at left end of scan line */
				if (lastDiffLeft !=
				    (pCurRun->runLeft - pExemGroup->leftEnd))
				{
				    lastDiffChangeLeft = lastDiffLeft -
							  (pCurRun->runLeft -
							   pExemGroup->leftEnd);
				}
				lastDiffLeft = (pCurRun->runLeft -
							   pExemGroup->leftEnd);
			    }

			    pCurRun = pCurRun->pNext;
			}
		    }
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	CheckNoise: On NULL run after even line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	CheckNoise: On run at [%d, %d] after even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
}
#endif
		    pExemGroup = pExemGroup->pNext;
		    break;

		case cCheck1Run:
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
							cOddLineTag))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	Check1Run: NULL run falls outside even line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	Check1Run: Run at [%d, %d], tag %p, should be even, is odd\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient);
}

#endif
			goto tcTryNextSym;
		    }
		    if (pExemGroup->leftEndMin > pCurRun->runLeft)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1Run: Left endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    if (pCurRun->runRight > pExemGroup->rightEndMax)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1Run: Right endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    if ((pCurRun->runLeft <= pExemGroup->leftEndMax) &&
			(pCurRun->runRight >= pExemGroup->rightEndMin))
		    {	/* hooray, it matched, so try the next
			 * group and run after recording the difference
			 * in endpoints.
			 */
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1Run: Pass run at [%d, %d] relative to even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			if (pExemGroup->groupType & cAtRightEnd)
			{
			    if (lastDiffRight !=
				(pCurRun->runRight - pExemGroup->rightEnd))
			    {
				totalDiffRight += ABS(lastDiffChangeRight +
						      lastDiffRight -
						      (pCurRun->runRight -
						       pExemGroup->rightEnd));
				lastDiffChangeRight = lastDiffRight -
						      (pCurRun->runRight -
						       pExemGroup->rightEnd);
			    }
			    lastDiffRight = (pCurRun->runRight -
						       pExemGroup->rightEnd);
			}
			    /* if we're at the left end of the scan line,
			     * remember the difference in coords */
			if (pExemGroup->groupType & cAtLeftEnd)
			{
			        /* found run at left end of scan line */
			    if (lastDiffLeft !=
				(pCurRun->runLeft - pExemGroup->leftEnd))
			    {
				totalDiffLeft += ABS(lastDiffChangeLeft +
						      lastDiffLeft -
						      (pCurRun->runLeft -
						       pExemGroup->leftEnd));
				lastDiffChangeLeft = lastDiffLeft -
						      (pCurRun->runLeft -
						       pExemGroup->leftEnd);
			    }
			    lastDiffLeft = (pCurRun->runLeft -
						       pExemGroup->leftEnd);
			}
			pExemGroup = pExemGroup->pNext;
			pCurRun = pCurRun->pNext;
			break;
		    }
		    /* if the inside endpoints of the run didn't match,
		     * there might have been more than one run here.
		     * So fall through to cCheckEndpoints, which will
		     * look for that case.
		     */

		case cCheckEndpoints:
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
								cOddLineTag) ||
			(pCurRun->runRight < pExemGroup->rightEndMin))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	CheckEndpoints: NULL run falls outside even line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	CheckEndpoints: Right endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
}
#endif
			goto tcTryNextSym;
		    }

		    if (pCurRun->runRight > pExemGroup->rightEndMax)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckEndpoints: Right endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

			    /* the right endpoint is ok, remember the
			     * difference in right endpoints.
			     */
			if (pExemGroup->groupType & cAtRightEnd)
			{
			    if (lastDiffRight !=
				(pCurRun->runRight - pExemGroup->rightEnd))
			    {
				totalDiffRight += ABS(lastDiffChangeRight +
						      lastDiffRight -
						      (pCurRun->runRight -
						       pExemGroup->rightEnd));
				lastDiffChangeRight = lastDiffRight -
						      (pCurRun->runRight -
						       pExemGroup->rightEnd);
			    }
			    lastDiffRight = (pCurRun->runRight -
						       pExemGroup->rightEnd);
			}
			/* Consume any runs to the right of the leftmost
			 * run in this region.
			 */
		    pTempRun = pCurRun->pNext;
		    if ((pTempRun != NULL) &&
			(((UInt32)pTempRun->pClient ^ lineTagMask) ==
							cEvenLineTag) &&
			(pTempRun->runLeft >= pExemGroup->leftEndMin))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckEndpoints: Discard run at [%d, %d] for even group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			pCurRun = pTempRun;
			pTempRun = pTempRun->pNext;
			while ((pTempRun != NULL) &&
			       (((UInt32)pTempRun->pClient ^ lineTagMask) ==
							cEvenLineTag) &&
				(pTempRun->runLeft >= pExemGroup->leftEndMin))
			{
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckEndpoints: Discard run at [%d, %d] to right of boundary at %d\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMax);
#endif
			    pCurRun = pTempRun;
			    pTempRun = pTempRun->pNext;
			}
		    }
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
							cOddLineTag) ||
			(pCurRun->runLeft > pExemGroup->leftEndMax))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	CheckEndpoints: NULL run falls outside even line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	CheckEndpoints: Left endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
}
#endif
			goto tcTryNextSym;
		    }
		    if (pExemGroup->leftEndMin > pCurRun->runLeft)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1Run: Left endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

			/* hooray, it matched, so try the next group and run */
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckEndpoints: Pass run at [%d, %d] relative to even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			/* if we're at the left end of the scan line,
			 * remember the difference in coords */
		    if (pExemGroup->groupType & cAtLeftEnd)
		    {
			    /* found run at left end of scan line */
			if (lastDiffLeft !=
			    (pCurRun->runLeft - pExemGroup->leftEnd))
			{
			    totalDiffLeft += ABS(lastDiffChangeLeft +
						  lastDiffLeft -
						  (pCurRun->runLeft -
						   pExemGroup->leftEnd));
			    lastDiffChangeLeft = lastDiffLeft -
						  (pCurRun->runLeft -
						   pExemGroup->leftEnd);
			}
			lastDiffLeft = (pCurRun->runLeft -
						   pExemGroup->leftEnd);
		    }
		    pExemGroup = pExemGroup->pNext;
		    pCurRun = pCurRun->pNext;
		    break;

		case cCheck1RunNoAdvance:
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
								cOddLineTag))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	Check1RunNoAdvance: NULL run falls outside even line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	Check1RunNoAdvance: Run at [%d, %d], tag %p, should be even, is odd\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient);
}
#endif
			goto tcTryNextSym;
		    }
		    if (pExemGroup->leftEndMin > pCurRun->runLeft)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1RunNoAdvance: Left endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    if (pCurRun->runRight > pExemGroup->rightEndMax)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1RunNoAdvance: Right endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    if ((pCurRun->runLeft <= pExemGroup->leftEndMax) &&
			(pCurRun->runRight >= pExemGroup->rightEndMin))
		    {	/* hooray, it matched, so try the next
			 * group BUT NOT run */
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1RunNoAdvance: Pass run at [%d, %d] relative to even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			pExemGroup = pExemGroup->pNext;
		    }
		    else	/* failed to match, so try next sym */
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1RunNoAdvance: Run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }
		    break;


		    /* ************* CHECK ODD LINES ***************** */

		case (cOddLineTag + cCheckNoise):
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
							cEvenLineTag) ||
			(pCurRun->runLeft > pExemGroup->rightEndMax) ||
			(pCurRun->runRight < pExemGroup->leftEndMin))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	CheckNoise: No overlap between NULL run and odd line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	CheckNoise: No overlap between run at [%d, %d], tag %p and odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
}
#endif
			    /* There's no overlap, so just skip this group */
			pExemGroup = pExemGroup->pNext;
			break;
		    }
			/* There's overlap.  Does the run fall entirely
			 * in the correct range?
			 */
		    if (pExemGroup->leftEndMin > pCurRun->runLeft)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckNoise: Run at [%d, %d] falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

			/* Check the right endpoint */
		    if (pCurRun->runRight > pExemGroup->rightEndMax)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckNoise: Run at [%d, %d] falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

			/* The first run is ok.  Update the difference in
			 * right endpoints.
			 */
		    if (pExemGroup->groupType & cAtRightEnd)
		    {
			if (lastDiffRight !=
			    (pCurRun->runRight - pExemGroup->rightEnd))
			{
			    lastDiffChangeRight = lastDiffRight -
						  (pCurRun->runRight -
						   pExemGroup->rightEnd);
			}
			lastDiffRight = (pCurRun->runRight -
						   pExemGroup->rightEnd);
		    }

			/* Discard any more runs to
			 * the right of discardBdry, if any.
			 */
		    if (pCurRun->runLeft >= pExemGroup->discardBdry)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckNoise: Discard run at [%d, %d] to right of boundary at %d\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->discardBdry);
#endif
			    /* if we're at the left end of the scan line,
			     * remember the difference in coords */
			if (((pCurRun->pNext == NULL) ||
			     (pCurRun->pClient != pCurRun->pNext->pClient)) &&
			    pExemGroup->groupType & cAtLeftEnd)
			{
			        /* found run at left end of scan line */
			    if (lastDiffLeft !=
				(pCurRun->runLeft - pExemGroup->leftEnd))
			    {
				lastDiffChangeLeft = lastDiffLeft -
						      (pCurRun->runLeft -
						       pExemGroup->leftEnd);
			    }
			    lastDiffLeft = (pCurRun->runLeft -
						       pExemGroup->leftEnd);
			}

			pCurRun = pCurRun->pNext;
			while ((pCurRun != NULL) &&
				(((UInt32)pCurRun->pClient ^ lineTagMask) ==
								cOddLineTag) &&
				(pCurRun->runLeft >= pExemGroup->discardBdry))
			{
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckNoise: Discard run at [%d, %d] to right of boundary at %d\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->discardBdry);
#endif
				/* if we're at the left end of the scan line,
				 * remember the difference in coords */
			    if (((pCurRun->pNext == NULL) ||
				 (pCurRun->pClient !=
						pCurRun->pNext->pClient)) &&
				pExemGroup->groupType & cAtLeftEnd)
			    {
				    /* found run at left end of scan line */
				if (lastDiffLeft !=
				    (pCurRun->runLeft - pExemGroup->leftEnd))
				{
				    lastDiffChangeLeft = lastDiffLeft -
							  (pCurRun->runLeft -
							   pExemGroup->leftEnd);
				}
				lastDiffLeft = (pCurRun->runLeft -
							   pExemGroup->leftEnd);
			    }

			    pCurRun = pCurRun->pNext;
			}
		    }
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	CheckNoise: On NULL run after odd line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	CheckNoise: On run at [%d, %d] after odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
}
#endif
		    pExemGroup = pExemGroup->pNext;
		    break;

		case (cOddLineTag + cCheck1Run):
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
								cEvenLineTag))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	Check1Run: NULL run falls outside odd line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	Check1Run: Run at [%d, %d], tag %p, should be odd, is even\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient);
}

#endif
			goto tcTryNextSym;
		    }
		    if (pExemGroup->leftEndMin > pCurRun->runLeft)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1Run: Left endpoint of run at [%d, %d], tag %p, falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    if (pCurRun->runRight > pExemGroup->rightEndMax)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1Run: Right endpoint of run at [%d, %d], tag %p, falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    if ((pCurRun->runLeft <= pExemGroup->leftEndMax) &&
			(pCurRun->runRight >= pExemGroup->rightEndMin))
		    {	/* hooray, it matched, so try the next
			 * group and run after recording the difference
			 * in endpoints. */
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1Run: Pass run at [%d, %d] relative to odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			if (pExemGroup->groupType & cAtRightEnd)
			{
			    if (lastDiffRight !=
				(pCurRun->runRight - pExemGroup->rightEnd))
			    {
				totalDiffRight += ABS(lastDiffChangeRight +
						      lastDiffRight -
						      (pCurRun->runRight -
						       pExemGroup->rightEnd));
				lastDiffChangeRight = lastDiffRight -
						      (pCurRun->runRight -
						       pExemGroup->rightEnd);
			    }
			    lastDiffRight = (pCurRun->runRight -
						       pExemGroup->rightEnd);
			}
			    /* if we're at the left end of the scan line,
			     * remember the difference in coords */
			if (pExemGroup->groupType & cAtLeftEnd)
			{
			        /* found run at left end of scan line */
			    if (lastDiffLeft !=
				(pCurRun->runLeft - pExemGroup->leftEnd))
			    {
				totalDiffLeft += ABS(lastDiffChangeLeft +
						      lastDiffLeft -
						      (pCurRun->runLeft -
						       pExemGroup->leftEnd));
				lastDiffChangeLeft = lastDiffLeft -
						      (pCurRun->runLeft -
						       pExemGroup->leftEnd);
			    }
			    lastDiffLeft = (pCurRun->runLeft -
						       pExemGroup->leftEnd);
			}
			pExemGroup = pExemGroup->pNext;
			pCurRun = pCurRun->pNext;
			break;
		    }
		    /* if the run didn't match, fall through to
		     * cCheckEndpoints, which will try a little harder.
		     */

		case (cOddLineTag + cCheckEndpoints):
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
							cEvenLineTag) ||
			(pCurRun->runRight < pExemGroup->rightEndMin))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	CheckEndpoints: NULL run falls outside odd line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	CheckEndpoints: Right endpoint of run at [%d, %d], tag %p, falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
}
#endif
			goto tcTryNextSym;
		    }

		    if (pCurRun->runRight > pExemGroup->rightEndMax)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckEndpoints: Right endpoint of run at [%d, %d], tag %p, falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

			    /* the right endpoint is ok, remember the
			     * difference in right endpoints.
			     */
			if (pExemGroup->groupType & cAtRightEnd)
			{
			    if (lastDiffRight !=
				(pCurRun->runRight - pExemGroup->rightEnd))
			    {
				totalDiffRight += ABS(lastDiffChangeRight +
						      lastDiffRight -
						      (pCurRun->runRight -
						       pExemGroup->rightEnd));
				lastDiffChangeRight = lastDiffRight -
						      (pCurRun->runRight -
						       pExemGroup->rightEnd);
			    }
			    lastDiffRight = (pCurRun->runRight -
						       pExemGroup->rightEnd);
			}

			/* Consume any runs to the right of the leftmost
			 * run in this region.
			 */
		    pTempRun = pCurRun->pNext;
		    if ((pTempRun != NULL) &&
			(((UInt32)pTempRun->pClient ^ lineTagMask) ==
							cOddLineTag) &&
			(pTempRun->runLeft >= pExemGroup->leftEndMin))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckEndpoints: Discard run at [%d, %d] for odd group at  [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			pCurRun = pTempRun;
			pTempRun = pTempRun->pNext;
			while ((pTempRun != NULL) &&
				(((UInt32)pTempRun->pClient ^ lineTagMask) ==
								cOddLineTag) &&
				(pTempRun->runLeft >= pExemGroup->leftEndMin))
			{
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckEndpoints: Discard run at [%d, %d] to right of boundary at %d\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMax);
#endif
			    pCurRun = pTempRun;
			    pTempRun = pTempRun->pNext;
			}
		    }
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
							cEvenLineTag) ||
			(pCurRun->runLeft > pExemGroup->leftEndMax))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	CheckEndpoints: NULL run falls outside odd line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	CheckEndpoints: Left endpoint of run at [%d, %d], tag %p, falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
}
#endif
			goto tcTryNextSym;
		    }
		    if (pExemGroup->leftEndMin > pCurRun->runLeft)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1Run: Left endpoint of run at [%d, %d], tag %p, falls outside even line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    /* hooray, it matched, so try the next group and run */
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckEndpoints: Pass run at [%d, %d] relative to odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			/* if we're at the left end of the scan line,
			 * remember the difference in coords */
		    if (pExemGroup->groupType & cAtLeftEnd)
		    {
			    /* found run at left end of scan line */
			if (lastDiffLeft !=
			    (pCurRun->runLeft - pExemGroup->leftEnd))
			{
			    totalDiffLeft += ABS(lastDiffChangeLeft +
						  lastDiffLeft -
						  (pCurRun->runLeft -
						   pExemGroup->leftEnd));
			    lastDiffChangeLeft = lastDiffLeft -
						  (pCurRun->runLeft -
						   pExemGroup->leftEnd);
			}
			lastDiffLeft = (pCurRun->runLeft -
						   pExemGroup->leftEnd);
		    }
		    pExemGroup = pExemGroup->pNext;
		    pCurRun = pCurRun->pNext;
		    break;

		case (cOddLineTag + cCheck1RunNoAdvance):
		    if ((pCurRun == NULL) ||
			(((UInt32)pCurRun->pClient ^ lineTagMask) ==
								cEvenLineTag))
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
{
if (pCurRun == NULL)
fprintf(stderr, "	Check1RunNoAdvance: NULL run falls outside odd line group at [%d, %d]\n", pExemGroup->leftEndMin, pExemGroup->rightEndMax);
else
fprintf(stderr, "	Check1RunNoAdvance: Run at [%d, %d], tag %p, should be odd, is even\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient);
}
#endif
			goto tcTryNextSym;
		    }
		    if (pExemGroup->leftEndMin > pCurRun->runLeft)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1RunNoAdvance: Left endpoint of run at [%d, %d], tag %p, falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    if (pCurRun->runRight > pExemGroup->rightEndMax)
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1RunNoAdvance: Right endpoint of run at [%d, %d], tag %p, falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }

		    if ((pCurRun->runLeft <= pExemGroup->leftEndMax) &&
			(pCurRun->runRight >= pExemGroup->rightEndMin))
		    {	/* hooray, it matched, so try the next
			 * group BUT NOT run */
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1RunNoAdvance: Pass run at [%d, %d] relative to odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			pExemGroup = pExemGroup->pNext;
		    }
		    else	/* failed to match, so try next sym */
		    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	Check1RunNoAdvance: Run at [%d, %d], tag %p, falls outside odd line group at [%d, %d]\n", pCurRun->runLeft, pCurRun->runRight, pCurRun->pClient, pExemGroup->leftEndMin, pExemGroup->rightEndMax);
#endif
			goto tcTryNextSym;
		    }
		    break;

		default:
		    __error__ = ia_internal;
		    goto tcCompareSymCleanup;
		}
	    }	/* End of loop that matches either horizontal runs or
		 * vertical runs.
		 * All the instance runs should have been consumed by the
		 * exemplar groups.  If there are any left, go find another
		 * symbol.
		 */
	    if (pCurRun != NULL)
		goto tcTryNextSym;

		/* if the noise waivered around too much, discard this
		 * symbol */
	    if (matchCount == 0)
		lastDiffLeft =
		    ((pInstance->h * cDiffSlopFactor) - pInstance->h) /
							cDiffSlopFactor;
	    else
		lastDiffLeft =
		    ((pInstance->w * cDiffSlopFactor) - pInstance->w) /
							cDiffSlopFactor;

	    if ((totalDiffLeft > lastDiffLeft) ||
		(totalDiffRight > lastDiffLeft))
	    {
#if TC_DEBUG
/* TC_DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	DifferenceCheck: Symbol with totalDiffLeft = %ld and totalDiffRight = %ld fails.  Largest diff allowed is %ld\n", totalDiffLeft, totalDiffRight, lastDiffLeft);
		pTextCompInfo->diffCheck++;
#endif
		goto tcTryNextSym;
	    }

	}	/* end of loop that matches both horizontally and vertically */

	    /* If we fall out of this loop, everything matched, so
	     * declare success and leave.
	     */
	*ppMatchedSymbol = pExemSymbol;
	break;

	/* we get here if there was a mismatch.  Go try the next symbol */
tcTryNextSym:
	pExemSymbol = pExemSymbol->pNext;
    }

tcCompareSymCleanup:
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getNumberOfSymbols()
 *
 *	This procedure returns the number of distinct symbols represented
 *	in the shape info.  It is used when allocating a table used to
 *	map old symbol IDs to new ones.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to structure containing
 *					information about shape tree
 *		symbolSet:		cSymbolsAll: all in dictionary
 *					cSymbolsMoreThanOnce: counts > 1
 *					cSymbolsOnceOnPage: counts = 1,
 *						same page ID
 *		pageRequest:		Page request: 0 = all pages, counts > 1
 *						>0 = specific page, counts = 1
 *
 *	Output:
 *		pNumOfSymbols:		Pointer to word that will contain
 *					the result.
 *		
 *
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getNumberOfSymbols (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Int32	symbolSet,
    UInt32	pageRequest,
    Int32      *pNumberOfSymbols)
{
    PROLOGUE;
    Int32	symbolCount = 0;
    TCSymbol   *pSymbol;
    TCCompressionInfoPtr pTextCompInfo = (TCCompressionInfoPtr) pTextCompInfoIn;

    if (pTextCompInfo == NULL)
	return ia_invalidParm;

    switch (symbolSet) {
    case cSymbolsAll:
	*pNumberOfSymbols = pTextCompInfo->nextSymbolNum;
	break;
    case cSymbolsMoreThanOnce:
	if (pTextCompInfo->encodeState & cTCSymsAreInDictOrder) {
	    for (pSymbol = pTextCompInfo->p1stDictSymbol;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextDictSym)
	    {
		if (pSymbol->instanceCount > 1)
		    symbolCount++;
	    }
	}
	else
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextNum)
	    {
		if (pSymbol->instanceCount > 1)
		    symbolCount++;
	    }
	}
	*pNumberOfSymbols = symbolCount;
	break;
    case cSymbolsOnceOnPage:
	/* Single symbols are not in the dictionary list, but in the
	 * entire list */
	if (pTextCompInfo->encodeState & cTCSymsAreInDictOrder) {
	    for (pSymbol = pTextCompInfo->p1stDictSymbol;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextDictSym)
	    {
		if (pageRequest == pSymbol->pageID &&
		    pSymbol->instanceCount == 1)
		    symbolCount++;
	    }
	}
	else
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextNum)
	    {
		if (pageRequest == pSymbol->pageID &&
		    pSymbol->instanceCount == 1)
		    symbolCount++;
	    }
	}
	*pNumberOfSymbols = symbolCount;
	break;
    case cSymbolsUsedByPage:
	if (pTextCompInfo->encodeState & cTCSymsAreInDictOrder) {
	    for (pSymbol = pTextCompInfo->p1stDictSymbol;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextDictSym)
	    {
		if (pageRequest == pSymbol->pageID ||
		    pSymbol->instanceCount > 1)
		    symbolCount++;
	    }
	}
	else
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextNum)
	    {
		if (pageRequest == pSymbol->pageID ||
		    pSymbol->instanceCount > 1)
		    symbolCount++;
	    }
	}
	*pNumberOfSymbols = symbolCount;
	break;
    default:
	return ia_invalidParm;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getSizeOfSymbolRemapArray()
 *
 *	This procedure returns the size in bytes of an array that is
 *	filled in by i_consolidateShapeInfo and used by i_remapSymbolIDs.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to structure containing
 *					information about shape tree
 *		pSymRemapArraySize:	Pointer to returned size of
 *					remap table in bytes.
 *		
 *	Action:
 *		Calculates and returns size of remap array.  Client should
 *		allocate and free this array.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfSymbolRemapArray (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Int32     *pSymRemapArraySize)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo = (TCCompressionInfoPtr) pTextCompInfoIn;

	/* This is only useful during lossy matches */
    if (pTextCompInfo->exactMatchSet && pTextCompInfo->exactMatchOnly)
	return ia_invalidParm;

    *pSymRemapArraySize = pTextCompInfo->nextSymbolNum * sizeof(TCRemapSym);
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_consolidateShapeInfo()
 *
 *	This procedure is used after a set of symbols is originally produced.
 *	It attempts to consolidate symbols with other symbols by adding
 *	noise to the outside of each and reclassifying.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to structure containing
 *					information about shape tree
 *		pSymbolIDMapIn:		Pointer to map that will become
 *					the translation table between
 *					old symbol IDs and new symbol IDs.
 *					This maps size is found using
 *					i_getTCSymbolIDMapSize and is
 *					filled in by this routine.
 *		
 *	Action:
 *		Finds symbols that differ only in edge noise.  These are
 *		grouped into equivalence classes.  One member of each class is
 *		retained, the rest discarded.  Discarded symbols are
 *		removed from the symbol list.  For all old symbol ID's,
 *		pSymbolIDMap[oldSymbolID] is filled with the information
 *		that i_remapSymbolIDs will need to modify the position
 *		buffers.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_consolidateShapeInfo (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    void       *pSymbolIDMapIn)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo = (TCCompressionInfoPtr) pTextCompInfoIn;
    TCRemapSym		*pRemapSymArray = (TCRemapSym *)pSymbolIDMapIn;
    TCSymbol		*pCurSymbol, *pTempSymbol, *pBestSymbol;
    Int32		 remainingSyms, i;
    TCScanLineRun	*pNewRunHead, *pNewRunTail; /* for list of runs that
						     * we modify to move the
						     * symbol around a little */
    TCScanLineRun	*pRestOfNewRuns;	    /* points to TCScanLineRun
						     * structs that aren't used
						     * for some symbol. */
    TCScanLineRun	*pNewRunTemp;		    /* scratch ptr */
    TCScanLineRun	 extraRunLeft, extraRunRight;
						    /* used to add noise */
    Int32		 w, h;			    /* sym's width and height */
    TCSymbol	       **symPtrArray;		    /* ptr to array of symbol
						     * ptrs. used by qsort */
    

    if (pTextCompInfo == NULL)
	return ia_invalidParm;

	/* This is only useful during lossy matches */
    if (pTextCompInfo->exactMatchSet && pTextCompInfo->exactMatchOnly)
	return ia_invalidParm;

	/* Prepare structure for future use */
    extraRunLeft.pClient = extraRunRight.pClient = NULL;
    pNewRunHead = NULL;
    pRestOfNewRuns = NULL;

	/* If there's already a cached table to pointers to symbols,
	 * delete it.  The table will be invalid after we rearrange
	 * the symbols.
	 */
    __error__ = ip_destroyTCSymbolTable (pIPState, pTextCompInfo);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* Declare that the symbols are not in any order.  We're about
	 * to rearrange them.
	 */
    pTextCompInfo->encodeState &= ~(cTCSymsAreInCountAndDictOrder |
				    cTCSymsAreInDictOrder);

	/* As explained in the comments in i_tcompr.c, this procedure
	 * is constructed from a number of passes through the symbol
	 * list.  In the first pass, we try adding noise to the bottom
	 * and right edges of each symbol.  We submit the adjusted
	 * list of runs to ip_classifySymbol and look at the result.
	 * We set the classifyOnly flag in pTextCompInfo so ip_classifySymbol
	 * won't attempt to make a new symbol or put out position info.
	 * We also put the symbol's pointer into pTextCompInfo's
	 * pClassifiedSymbol entry so we don't try to match the symbol
	 * against itself.  The results of the classification are
	 * returned in pTextCompInfo->pClassifiedSymbol.
	 * If it is a pointer to the original symbol, we didn't find
	 * an equivalent symbol.
	 *
	 * To add noise, we want to modify the symbol's runs.  When
	 * consolidating symbols, ip_classifySymbol thinks that the
	 * runs passed in are already correct.  That is, they have
	 * 2x coordinates relative to the center of the symbol and have
	 * correct line tags.  We need to make a new list
	 * for each symbol that has the right characteristics.
	 * It's wasteful to allocate and free a list for every symbol, so
	 * we reuse the same one.  pNewRunHead points to the first run
	 * in the list.  pNewRunTail points to the last one for the
	 * particular symbol.  pRestOfNewRuns points to the runs that
	 * have been allocated, but are not used for the current symbol.
	 * When we start classifying each new symbol, pNewRunHead points
	 * to the first run in the list and pNewRunTail and pRestOfNewRuns
	 * are both invalid.
	 */
    pTextCompInfo->classifyOnly = cTrue;
    pTempSymbol = NULL;				/* keep compiler happy */

	/* allocate the first run in the new run list so that we don't
	 * have to test to see if the list is empty all the time.
	 */
    __error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
				(void **)&pNewRunHead);
    if (__error__ != ia_successful)
	goto tcConsolidateSymCleanup;

    pNewRunHead->pNext = NULL;

    if (pTextCompInfo->p1stSymbol == NULL)
	NORMAL_RETURN;

	/* Why do we go through the symbols from last found to first found?
	 * Well, umm, to maintain the order in which previous versions
	 * processed the symbols.  It's a historical accident.
	 */
    for (pCurSymbol = pTextCompInfo->pLastSymbol;
	 pCurSymbol != NULL;
	 pCurSymbol = pCurSymbol->pPrevNum)
    {
	    /* set up a fake bounding box for this symbol.  The runs
	     * are already positive and negative numbers around the
	     * origin.  If we wanted to get exactly the same symbol,
	     * we would call ip_classifySymbol with these values for
	     * x, y, w and h.  Remember that all coordinates have an
	     * extra bit of significance, so 
	     */
	w = pCurSymbol->maxX + 1;
	h = pCurSymbol->maxY + 1;

	    /* if the symbol is too tall or too wide, forget it.  There
	     * can't be very many versions of big things and all the tests
	     * below take a long time on these versions.
	     */
	if ((w > 300) || (h > 300))
	    continue;

	    /* ip_classifySymbol expects all the run coordinates to be
	     * relative to the real image.  At this point, the coordinates
	     * associated with symbols are relative to the middle of the
	     * symbol AND have an extra bit of precision at the bottom.
	     * Create a new set of runs from the symbol's runs.  We'll
	     * use these for classification.
	     */
	pNewRunTail = pNewRunHead;
	for (pNewRunTemp = pCurSymbol->pRunTail; pNewRunTemp != NULL; )
	{
		/* Make sure there will be a run after this one is used up.
		 * The coordinates contained in the run go from -(w-1) to
		 * (w-1).  
		 */
	    pNewRunTail->runLeft = pNewRunTemp->runLeft;
	    pNewRunTail->runRight = pNewRunTemp->runRight;
	    pNewRunTail->pClient = pNewRunTemp->pClient;
	    pNewRunTemp = pNewRunTemp->pNext;
	    if (pNewRunTemp != NULL)
	    {
		if (pNewRunTail->pNext == NULL)
		{
		    __error__ =
			i_allocateStruct (pIPState,
					  pTextCompInfo->p1stScanLineBlock,
					  (void **)&pNewRunTail->pNext);
		    if (__error__ != ia_successful)
			goto tcConsolidateSymCleanup;
		    pNewRunTail->pNext->pNext = NULL;
	        }
		pNewRunTail = pNewRunTail->pNext;
	    }
	}

	    /* At this point, pNewRunTail points to the last valid
	     * run.  End the list here.
	     */
	pRestOfNewRuns = pNewRunTail->pNext;
	pNewRunTail->pNext = NULL;

	    /* First, try classifying the symbol while preventing the
	     * classification process from finding the symbol or any
	     * symbol in this symbol's equivalence class.
	     */
	    /* we know we've found an equivalent symbol when
	     * pTextCompInfo->pClassifiedSymbol comes back pointing
	     * to some other symbol.
	     */
	pTextCompInfo->pClassifiedSymbol = pCurSymbol;
	__error__ = ip_classifySymbol (pIPState, pTextCompInfo,
				     0, 0,	/* x, y don't matter */
				     w, h,
				     0, 0,	/* mass, numRuns are ignored */
				     pNewRunHead);
	if (__error__ != ia_successful)
	    goto tcConsolidateSymCleanup;

	pTempSymbol = pTextCompInfo->pClassifiedSymbol;
	if (pTempSymbol != pCurSymbol)
	{
#if TC_DEBUG
	    pTextCompInfo->noOffsetConsolidate++;
#endif
	    __error__ = ip_foundEquivalentSymbol (pIPState, pTextCompInfo,
					        pTempSymbol, pCurSymbol);

	    if (__error__ != ia_successful)
		goto tcConsolidateSymCleanup;
	}

	    /* Don't allow small symbols to change size. */
	if (MIN (w, h) < cNonGrowingSymbolSize)
	    continue;

	    /* Try adding noise to the bottom of the symbol.  We do this
	     * by adding a run to the middle of the rightmost
	     * run on the bottom (first) scan line.  We a add narrow run
	     * to avoid losing because the noise was too wide.
	     * BUT FIRST, fix up the run list.  All the pClient line tags
	     * should be inverted to account for the fact that all
	     * the lines will be moved up one line.
	     */
	;
	for (pNewRunTemp = pNewRunHead;
	     pNewRunTemp != NULL;
	     pNewRunTemp = pNewRunTemp->pNext)
	{
	    pNewRunTemp->pClient =
		(void *) ((Int32) (pNewRunTemp->pClient) ^ cLineTagMask);
	}
	extraRunRight.runLeft = (pNewRunHead->runLeft +
				 pNewRunHead->runRight)/2;
	extraRunRight.runRight = extraRunRight.runLeft;
	extraRunRight.pClient = (void *) cEvenLineTag;
	extraRunRight.pNext = pNewRunHead;

	    /* we know we've found an equivalent symbol when
	     * pTextCompInfo->pClassifiedSymbol comes back pointing
	     * to some other symbol.
	     */
	pTextCompInfo->pClassifiedSymbol = pCurSymbol;
	__error__ = ip_classifySymbol (pIPState, pTextCompInfo,
				     0, 0,	/* x, y, don't matter */
				     w, h + 1,
				     0, 0,	/* mass, numRuns are ignored */
				     &extraRunRight);
	if (__error__ != ia_successful)
	    goto tcConsolidateSymCleanup;

	pTempSymbol = pTextCompInfo->pClassifiedSymbol;
	if (pTempSymbol != pCurSymbol)
	{
#if TC_DEBUG
	    pTextCompInfo->vertOffsetConsolidate++;
#endif
	    __error__ = ip_foundEquivalentSymbol (pIPState, pTextCompInfo,
					        pTempSymbol, pCurSymbol);

	    if (__error__ != ia_successful)
		goto tcConsolidateSymCleanup;
	}


	    /* try moving the symbol to the right by 1.  Do this by
	     * decrementing every coordinate on the right edge of the
	     * symbol and incrementing all the others.  We have to do
	     * it this way so that there will be an extra vertical run
	     * on the left.  Also, put the line tags back the way they
	     * were.
	     */
	for (pNewRunTemp = pNewRunHead;
	     pNewRunTemp != NULL;
	     pNewRunTemp = pNewRunTemp->pNext)
	{
	    if (pNewRunTemp->runLeft == -(w-1))
		pNewRunTemp->runLeft--;
	    else
		pNewRunTemp->runLeft++;
	    pNewRunTemp->runRight++;
	    pNewRunTemp->pClient =
		(void *) ((Int32) (pNewRunTemp->pClient) ^ cLineTagMask);
	}

	pTextCompInfo->pClassifiedSymbol = pCurSymbol;
	__error__ = ip_classifySymbol (pIPState, pTextCompInfo,
				     0, 0,	/* x, y, don't matter */
				     w + 1, h,
				     0, 0,	/* mass, numRuns are ignored */
				     pNewRunHead);
	if (__error__ != ia_successful)
	    goto tcConsolidateSymCleanup;

	pTempSymbol = pTextCompInfo->pClassifiedSymbol;
	if ((pTempSymbol != NULL) && (pTempSymbol != pCurSymbol))
	{
#if TC_DEBUG
	    pTextCompInfo->horizOffsetConsolidate++;
#endif
	    __error__ = ip_foundEquivalentSymbol (pIPState, pTextCompInfo,
					        pTempSymbol, pCurSymbol);

	    if (__error__ != ia_successful)
		goto tcConsolidateSymCleanup;
	}

	    /* put the list of runs back together again */
	pNewRunTail->pNext = pRestOfNewRuns;
	pRestOfNewRuns = NULL;
    }

	/************************* SORT SYMBOLS ***********************/

	/* All the equivalence classes have been formed.  Go through the
	 * symbol list again.  In this pass, decide which symbol to
	 * retain in each equivalence class.
	 */

    for (pCurSymbol = pTextCompInfo->p1stSymbol, remainingSyms = 0;
	 pCurSymbol != NULL;
	 pCurSymbol = pCurSymbol->pNextNum)
    {
	    /* If this symbol is not equivalent to some other symbol,
	     * no further work is needed since it obviously is the
	     * definitive element of its class.  The only thing we
	     * want to do is remember that this symbol will be
	     * one of the surviving symbols.
	     */
	if (pCurSymbol->pNewSymbol == NULL)
	{
	    remainingSyms++;
	    continue;
	}

	    /* If this symbol's instanceCount is 0 or the next symbol
	     * in this symbol's equivalence class is 0 (this is the
	     * surviving member of its equivalence class), then it's
	     * an element of an equivalence class that we've already
	     * seen, so just skip to the next symbol.
	     */
	if ((pCurSymbol->instanceCount == 0) ||
	    (pCurSymbol->pNewSymbol->instanceCount == 0))
	{
	    continue;	/* we've already dealt with this equivalence class */
	}

	    /* initialize the pointer to the surviving symbol and look
	     * at all the symbols in the loop.  The one with the highest
	     * total number of instances wins.  In case of a tie, the
	     * symbol with the largest number of originals wins.
	     */
	remainingSyms++;		/* record number of surviving symbols */
	pBestSymbol = pCurSymbol;
	for (pTempSymbol = pCurSymbol->pNewSymbol;
	     pTempSymbol != pCurSymbol;
	     pTempSymbol = pTempSymbol->pNewSymbol)
	{
	    if ((pTempSymbol->numActiveRuns > pBestSymbol->numActiveRuns) ||
		((pTempSymbol->numActiveRuns ==
					pBestSymbol->numActiveRuns) &&
		 (pTempSymbol->instanceCount > pBestSymbol->instanceCount)))
		pBestSymbol = pTempSymbol;
	}

	    /* having chosen the surviving symbol, set the original
	     * instance counts of all the others to 0.  This is a
	     * tag for the symbols to be deleted.  Also free all the
	     * scan line runs and groups associated with the
	     * symbol.  But first, delete the symbol that holds
	     * data that's common to the class.  This data was only
	     * used during formation of the class.
	     */
	i_freeStruct (pIPState, pTextCompInfo->p1stSymbolBlock,
		      (void **)&(pBestSymbol->pClassSymbol));

	for (pTempSymbol = pBestSymbol->pNewSymbol;
	     pTempSymbol != pBestSymbol;
	     pTempSymbol = pTempSymbol->pNewSymbol)
	{
	    pBestSymbol->instanceCount += pTempSymbol->instanceCount;
	    pTempSymbol->instanceCount = 0;
	}
    }

	/* Here's some new code.  In the old code, we tried to make
	 * sure that the new symbol ID for every symbol was no longer
	 * than the old symbol ID.  In the new code, we sort the
	 * symbols by instanceCount so that the symbols with the
	 * most usages get the lowest symbol numbers.  This means that
	 * some symbols that were found early (had low symbol IDs), but
	 * not used often, will get longer IDs.  However, (number of
	 * symbols whose IDs get shorter) * (number of times those symbols
	 * are used) should be greater than (number of symbols whose
	 * IDs get longer) * (number of times those symbols are used).
	 * We will use qsort to do the sorting.  It takes an array
	 * of thingies to sort.  In our case, this will be pointers
	 * to symbols.  We also supplied the routine ip_compareSymCounts to
	 * compare two symbols (see above).  The symbol with the largest
	 * instance count will end up at the beginning of the sorted array.
	 * We allocate two extra symbol ptrs, the first and last ones.
	 * These are set to NULL so we can always set
	 * symPtrArray[i]->pPrev = symPtrArray[i-1] and
	 * symPtrArray[i]->pNext = symPtrArray[i+1].
	 */
    symPtrArray = (TCSymbol **) MALLOC((remainingSyms + 2) *
							sizeof(TCSymbol **));
    if (symPtrArray == NULL)
	goto tcConsolidateSymCleanup;	/* Ooops, you're screwed. */

    for (pCurSymbol = pTextCompInfo->p1stSymbol, i = 1;
	 pCurSymbol != NULL;
	 pCurSymbol = pCurSymbol->pNextNum)
    {
	if (pCurSymbol->instanceCount != 0)	/* this is a surviving sym */
	{
	    symPtrArray[i] = pCurSymbol;
	    i++;
	}
    }
	/* set up dummy elements */
    symPtrArray[0] = NULL;
    symPtrArray[remainingSyms + 1] = NULL;

	/* First, set the order of the symbols as they'll be written
	 * out in the dictionary.  Even if some subset of all the
	 * internal symbols is written into a dictionary, the relative
	 * order will be preserved.  We use this order as the secondary
	 * key when sorting the symbols.
	 */
    __error__ = ip_sortTCSymbolsInTundraDictOrder (pIPState, pTextCompInfo,
						   cFalse);

	/* sort the symbol pointers by instance counts of associated syms.
	 * GUESS WHAT!!  The order of the output produced by qsort is
	 * platform dependent!!!  May they all rot!  I think they
	 * decide different things about how to treat the things that
	 * are "equal".  Our comparison routine, ip_compareSymCounts, no
	 * longer says that symbols are equal; using the original
	 * symbol ID to break ties.
	 */
    qsort((void *)(&symPtrArray[1]), remainingSyms, sizeof(TCSymbol *),
	  ip_compareSymCounts);

	/* Now number the symbols in the order given by the array.
	 * Also set the map entries for all symbols in an equivalence
	 * class to the new symbol ID.
	 */
    for (i = 1; i <= remainingSyms; i++)
    {
	pCurSymbol = symPtrArray[i];
#ifdef DD_DEBUG
	printf("%d: %dx%d, count %d\n", i, pCurSymbol->width, pCurSymbol->height,
	    pCurSymbol->instanceCount);
#endif

	    /* Since there are dummy elements at the beginning and
	     * the end of the array (both set to NULL), we can
	     * always do the following two list ptr updates.
	     */
	pCurSymbol->pPrevNum = symPtrArray[i-1];
	pCurSymbol->pNextNum = symPtrArray[i+1];

	/* Set the entries for the other symbols in the this
	 * equivalence class in the symbol ID table.  Delete those
	 * symbols.
	 */
	if (pCurSymbol->pNewSymbol != NULL)
	{
	    for (pTempSymbol = pCurSymbol->pNewSymbol;
		 pTempSymbol != pCurSymbol;
		 pTempSymbol = pBestSymbol)
	    {
		    /* Remember that i is offset by 1, so the
		     * current symbol number is i-1.
		     */
		pRemapSymArray[pTempSymbol->symbolID].width =
							pTempSymbol->width;
		pRemapSymArray[pTempSymbol->symbolID].height =
							pTempSymbol->height;
		pRemapSymArray[pTempSymbol->symbolID].leftProfile =
						pTempSymbol->leftProfile;
		pRemapSymArray[pTempSymbol->symbolID].rightProfile =
						pTempSymbol->rightProfile;
		pRemapSymArray[pTempSymbol->symbolID].newSymbolID = i-1;

		    /* This symbol should be deleted.  Set it's
		     * now-inoperatove pPrevNum and pNextNum
		     * pointers to NULL so we don't fuss with
		     * any other symbol's pointers.
		     */
		pTempSymbol->pPrevNum = NULL;
		pTempSymbol->pNextNum = NULL;
		pBestSymbol = pTempSymbol->pNewSymbol;
		__error__ = ip_deleteTCDictionarySymbol (pIPState, pTextCompInfo,
						       pTempSymbol);
	    }
	}
	    /* Remember that i is offset by 1, so the
	     * current symbol number is i-1.
	     */
	pRemapSymArray[pCurSymbol->symbolID].width =
						pCurSymbol->width;
	pRemapSymArray[pCurSymbol->symbolID].height =
						pCurSymbol->height;
	pRemapSymArray[pCurSymbol->symbolID].leftProfile =
						pCurSymbol->leftProfile;
	pRemapSymArray[pCurSymbol->symbolID].rightProfile =
						pCurSymbol->rightProfile;
	pRemapSymArray[pCurSymbol->symbolID].newSymbolID = i-1;
	pCurSymbol->symbolID = i-1;
    }

	/* Since we fussed with the symbol order, and destroyed lots
	 * of pNextDictSym pointers, declare that the symbols are no
	 * longer sorted in dictionary order.
	 */
    pTextCompInfo->encodeState &= ~cTCSymsAreInDictOrder;

	/* Set the pointers to the first and last symbols to point to
	 * the symbols with the lowest and highest symbol IDs, as required.
	 */
    pTextCompInfo->p1stSymbol = symPtrArray[1];
    pTextCompInfo->pLastSymbol = symPtrArray[remainingSyms];

	/* free the sorted array of symbol pointers */
    FREE(symPtrArray);

	/* record the new next symbol number and declare victory. */
    pTextCompInfo->nextSymbolNum = remainingSyms;
    __error__ = ia_successful;

	/* We've found the exemplars of each equivalence class and removed
	 * the others.  The remapping table has been made.  Free the list
	 * of runs before returning.
	 */
tcConsolidateSymCleanup:
    while (pNewRunHead != NULL)
    {
	pNewRunTail = pNewRunHead->pNext;
	i_freeStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
		      (void **)&pNewRunHead);
	pNewRunHead = pNewRunTail;
    }
    while (pRestOfNewRuns != NULL)
    {
	pNewRunTail = pRestOfNewRuns->pNext;
	i_freeStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
		      (void **)&pRestOfNewRuns);
	pRestOfNewRuns = pNewRunTail;
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_remapSymbolIDs()
 *
 *	Takes the symbol ID map produced by i_consolidateShapeInfo and
 *	pointers to old and new position info buffers.  It copies the
 *	position information from the old buffer to the new one, turning
 *	old symbol ID numbers to new ones as it goes.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to structure containing
 *					information about shape tree
 *		pSymbolIDMapIn:		Pointer to map that is
 *					the translation table between
 *					old symbol IDs and new symbol IDs.
 *		pOldPositionBuf:	Pointer to old position buffer.
 *		oldPositionBufSize:	Number of bytes in old position buf.
 *		pNewPositionBuf:	Pointer to new position buffer.  If
 *					remapping should be done in place,
 *					this pointer is equal to
 *					pOldPositionBuf.  If remapping is not
 *					done in place, this should be at
 *					least cTCPositionBufferSlop bytes
 *					bigger than the old buffer.
 *		newPositionBufSize:	Number of bytes in new position buffer
 *		pBytesUsed:		Pointer to word that will be filled
 *					width the number of bytes used in
 *					the new position buffer.
 *		
 *	Action:
 *		Creates a new position buffer from an old one by replacing
 *		old symbol positions and IDs with new ones specified by
 *		pSymbolIDMapIn.  Symbol positions are modified in an
 *		attempt make the old and new symbols line up even if
 *		one or the other has noise on the edges.
 *
 *		Note that if the client wants to reuse the space taken
 *		by the original position buffer, pNewPositionBuf should
 *		be passed in pointing to pOldPositionBuf.  In the first
 *		version of the code, we knew that the new symbol ID for
 *		every symbol would be no longer than the old symbol ID.
 *		Since we now sort the symbol IDs by use, a symbol that
 *		was found early (had a low ID), but was seldom used,
 *		might get a longer ID.  However, the net buffer length
 *		should go down since there should be more changes from
 *		long IDs to short IDs than vice versa.
 *		When a client specifies the same buffer as the input
 *		and output, this routine allocates a temporary buffer
 *		and calls itself with the temporary as the new output.
 *		Upon return, the contents of the temporary buffer are
 *		copied into the original buffer.
 *		BUG FIX: Sometimes, the new buffer is really longer
 *		than the old one.  In this case, we return ia_bufferExceeded.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		unable to allocate temp buffer
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	the data didn't fit into the
 *					new buffer.
 *
 **************************************************************************/
Int32  CDECL 
i_remapSymbolIDs (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    void       *pSymbolIDMapIn,
    UInt8      *pOldPositionBuf,
    UInt32	oldPositionBufSize,
    UInt8      *pNewPositionBuf,
    UInt32	newPositionBufSize,
    UInt32     *pBytesUsed)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo = (TCCompressionInfoPtr) pTextCompInfoIn;
    TCRemapSym *pRemapSymArray = (TCRemapSym *)pSymbolIDMapIn;
    UInt32	oldSymID;
    UInt8	positionInfo;
    Int32	curX, curY;
    UInt8      *pTempBuf = NULL;	/* used when input buf == output buf */
    UInt32	tempBufSize;
    UInt32	version;		/* the version number of the
					 * position buf */
    UInt32	magicNum;		/* the magic number of the
					 * position buf */
    TCInstance	instance;		/* used to pass info to
					 * ip_createUncommittedTCPosition */
    TCSymbol  **pSymbolTable;		/* a table used to find a
					 * symbol from a symbol number.
					 */
    TCSymbol   *pNewSym;		/* pointers to old and new symbols */
    

	/* make sure things are ok */
    if ((pSymbolIDMapIn == NULL) || (pOldPositionBuf == NULL) ||
	(pNewPositionBuf == NULL) || (pBytesUsed == NULL))
	return ia_invalidParm;

	/* This is only useful during lossy matches */
    if (pTextCompInfo->exactMatchSet && pTextCompInfo->exactMatchOnly)
	return ia_invalidParm;

	/* if the input and output buffers are the same, make an
	 * intermediate buffer and call this routine with it.
	 */
    if (pOldPositionBuf == pNewPositionBuf)
    {
	tempBufSize = newPositionBufSize + cTCPositionBufferSlop;
	pTempBuf = (UInt8 *)MALLOC(tempBufSize);
	if (pTempBuf == NULL)
	    return ia_nomem;

	__error__ = i_remapSymbolIDs (pIPState, pTextCompInfoIn,
				    pSymbolIDMapIn, pOldPositionBuf,
				    oldPositionBufSize, pTempBuf,
				    tempBufSize, pBytesUsed);
	if (__error__ != ia_successful)
	{
	    FREE(pTempBuf);
	    goto finRemapSyms;
	}

	    /* copy the new position stuff to the input/output buffer if
	     * it fits.  If it doesn't fit, complain.
	     */
	if (*pBytesUsed > newPositionBufSize)
	{
	    FREE(pTempBuf);
	    return ia_bufferExceeded;
	}
	else
	{
	    memcpy((void *)pNewPositionBuf, (void *)pTempBuf, *pBytesUsed);
	    FREE(pTempBuf);
	    NORMAL_RETURN;
	}
    }
    else	/* if the new buffer length is shorter than the
		 * old one, complain now */
    {
	if (newPositionBufSize < oldPositionBufSize + cTCPositionBufferSlop)
	    return ia_bufferExceeded;
    }

	/* initialize the instance parts so we don't have to worry
	 * about weird behavior with uninitialized variables.
	 */
    instance.minWidth = 0;
    instance.maxWidth = 0;
    instance.minHeight = 0;
    instance.maxHeight = 0;
    instance.x = 0;
    instance.y = 0;
    instance.w = 0;
    instance.h = 0;
    instance.leftProfile = 0;
    instance.rightProfile = 0;
    instance.maxDimension = 0;
    instance.strokeWidth = 0;
    instance.mass = 0;
    instance.numRuns = 0;
    instance.checkHVolume = cFalse;
    instance.checkVVolume = cFalse;
    instance.hVolume = 0;
    instance.vVolume = 0;
    instance.slant = 0;
    instance.hSAsymmetry = 0;
    instance.vSAsymmetry = 0;
    instance.pCurSymbol = NULL;
    instance.pHorizRuns = NULL;
    instance.pDupHorizRuns = NULL;
    instance.pVertRuns = NULL;
    instance.pDupVertRuns = NULL;

	/* Initialize the first and last dummy elements in the list
	 * of uncommitted positions.
	 * We try to allow bottom line of the recorded symbol positions
	 * to vary up or down a line so that the replacement symbol has
	 * the same horizontal centerline as the instance symbol.
	 * This means we can't be sure what's in a scan line until the
	 * symbols in the line above, the same line, and the line below
	 * have all been found.  To do this, we keep a list sorted by
	 * x position.  The list elements contain x pos, y pos, symID
	 * and ptr to next element.  We initialize two dummy elements
	 * here so that we don't have to worry about inserting elements
	 * at the beginning or end of the list.
	 * Note that we don't have to worry about garbage collection inside
	 * this list since it only exists during i_convertImageToSymbols.
	 * We set the dummy's Y position to -2 so it won't be
	 * confused with the initial -1 for the mythical first row.
	 * This lets us just always flush rows starting with
	 * lowUncomY.
	 */
    pTextCompInfo->dum1stLoc.pClient = (void *) 0xFFFFFFFF; /* no sym */
    pTextCompInfo->dum1stLoc.runLeft = -1;	    /* x pos = -1 */
    pTextCompInfo->dum1stLoc.runRight = -2;	    /* y pos = -2 */
    pTextCompInfo->dum1stLoc.pNext = &(pTextCompInfo->dumLastLoc);

    pTextCompInfo->dumLastLoc.pClient = (void *) 0xFFFFFFFF; /* no sym */
    pTextCompInfo->dumLastLoc.runLeft = 0x7FFF;	    /* x pos = big number */
    pTextCompInfo->dumLastLoc.runRight = -2;        /* y pos = -2 */
    pTextCompInfo->dumLastLoc.pNext = NULL;	    /* nothing to the right */

	/* This first buffer may look something like this.  It assumes
	 * the first symbols may start on line 0.
	 */
    pTextCompInfo->lowUncomY = -1;
    pTextCompInfo->hiUncomY = 1;
    pTextCompInfo->numUncomPos = 0;		    /* the list is empty */
    pTextCompInfo->pLastUncom = &(pTextCompInfo->dum1stLoc);

	/* The input and output position buffers are different.
	 * First, copy all the magic number and version stuff
	 */
    magicNum = ((UInt32) pOldPositionBuf[cTCPosMagicNum3] << 24) |
		((UInt32) pOldPositionBuf[cTCPosMagicNum2] << 16) |
		((UInt32) pOldPositionBuf[cTCPosMagicNum1] << 8) |
		((UInt32) pOldPositionBuf[cTCPosMagicNum0]);

    if (magicNum != cTextCompPositionInfoMagicNum)
	return ia_invalidParm;

    version = ((UInt32) pOldPositionBuf[cTCPosVer1] << 8) +
	      ((UInt32) pOldPositionBuf[cTCPosVer0]);

    if (version != cTextCompPositionInfoVersionNum)
	return ia_invalidParm;

	/* If we get here, it must all have been ok, so copy the first
	 * cTCPosHeaderVer2Len bytes over.  This includes the width
	 * and height as well */
    memcpy((void *)pNewPositionBuf,
	   (void *) pOldPositionBuf,
	   cTCPosHeaderVer2Len);

	/* Set the stats in the new buffer to invalid */
    memset((void *) &(pNewPositionBuf[cTCPosStatsValid]), 0,
	   cTCPosStatsLen);

    pOldPositionBuf += cTCPosHeaderVer2Len;
	/* Record information about the new position buffer
	 * in pTextCompInfo so we can use the subroutines to
	 * insert symbols into it.
	 */
    pTextCompInfo->pPositionBuf = pNewPositionBuf + cTCPosHeaderVer2Len;
    pTextCompInfo->positionBufSize = newPositionBufSize;
    pTextCompInfo->resultSize = cTCPosHeaderVer2Len;	/* so far */
    pTextCompInfo->lastY = -1;				/* initial coord */

	/* Now, see if the table that associates Symbol IDs with
	 * symbols exists.  If not, make it.
	 */
    if ((pTextCompInfo->pSymTable == NULL) ||
	(pTextCompInfo->pSymTable->numPtrs != pTextCompInfo->nextSymbolNum))
    {	/* allocate and initialize a new table.  Delete the old one
	 * if it exists.
	 */
	__error__ = ip_createTCSymbolTable (pIPState, pTextCompInfo);
	if (__error__ != ia_successful)
	    goto finRemapSyms;
    }
    pSymbolTable = &(pTextCompInfo->pSymTable->table[0]);

	/* Now start copying blocks of distance info, eliminating all symbols
	 * whose instance counts = 1. ??????????????????????????????????????????????
	 * In the new position buf format, the first byte in the
	 * position buffer is the deltaY offset relative to -1.
	 * If this is 0, the position buffer is empty.  Thereafter
	 * all non-zero deltaY offsets are followed by firstX offsets.
	 * The firstX offsets may be 0 if the leftmost symbol is fully
	 * contained in column 0 of the image (is 1 pixel wide).  We
	 * do NOT use a 0 value of the first X offset to indicate end
	 * of line.
	 * The position info can be of several forms:
	 *	1 byte x offset; 1 or 2 byte symbol ID.
	 *	2 byte x offset; 1 or 2 byte symbol ID.
	 *	1 byte x offset (= 0), 1 byte y offset,
	 *		1 or 2 byte x position, 1 or 2 byte sym ID.
	 *	1 byte x offset (= 0), 2 byte y offset,
	 *		1 or 2 byte x position, 1 or 2 byte sym ID.
	 *	1 byte x offset (= 0), 1 byte y offset (= 0) => end of file.
	 *	1 byte x offset (= 0), 2 byte y offset (= -1),
	 *		1 or 2 byte x position (to same pos as original
	 *		x position), 1 or 2 byte symbol ID.
	 * Bit 7 of all 1-byte quantities is 0, Bit 7 of all 2 byte
	 * quantities is 1 - that's how we tell the difference.
	 * To replace the old symbol with the new one, we remap the
	 * symbol ID using the table and adjust its position on the
	 * page so the old and new symbols line up.
	 */

	/* To start the loop in the proper phase, pretend we just read
	 * a deltaX of 0 so we'll first really read the first delta Y.
	 */
    positionInfo = 0;
    __error__ = ia_successful;		/* get ready for loop */
    curX = 0;
    curY = -1;

	/* We execute this loop until either we get close to the end
	 * of the output buffer or we're done.  If we exit with
	 * __error__ == ia_successful, we're done.  If we exit with
	 * __error__ == ia_bufferExceeded, we should have been given
	 * a bigger buffer.
	 */
    while (cTrue)
    {
	if (positionInfo & 0x80)		/* parse 2nd byte of x offset */
	{
	    curX += (positionInfo & 0x7F) + (*pOldPositionBuf++ << 7);
	}
	else
	if (positionInfo == 0)			/* there's a y offset */
	{
	    positionInfo = *pOldPositionBuf++;	/* get first byte of y offset */

	    if (positionInfo & 0x80)		/* copy next byte of y offset */
	    {
		if ((positionInfo & *pOldPositionBuf) != 0xFF)
		{
			/* the normal case */
		    curY += (positionInfo & 0x7F) + (*pOldPositionBuf++ << 7);
		}
		else
		{
			/* This is the special code used when the lower right
			 * corners of two symbols have exactly the same lower
			 * right corner.  Just eat the following byte or two
			 * of x offset since it always brings us back to the
			 * same spot.
			 */
		    pOldPositionBuf++;		/* skip 2nd byte of y offset */
		}
	    }
	    else
	    if (positionInfo == 0)	/* that was the end of the file */
	    {
		break;
	    }
	    else	/* a 1 byte y offset */
		curY += positionInfo;

		/* get the x offset at the beginning of the new line */
	    positionInfo = *pOldPositionBuf++;

	    if (positionInfo & 0x80)	/* copy next byte of x offset */
	    {
		curX = (positionInfo & 0x7F) + (*pOldPositionBuf++ << 7);
	    }
	    else
		curX = positionInfo;
	}
	else
	    curX += positionInfo;

	    /* At this point, the next byte or two is the symbol ID, read
	     * it.  Note that the MS bit of the first byte is just
	     * a flag bit.  If it's set, we need to ignore it when
	     * computing the real symbol ID.
	     */
	oldSymID = *pOldPositionBuf++;
	if (oldSymID & 0x80)
	{
	    oldSymID &= 0x7F;
	    oldSymID |= (*pOldPositionBuf++) << 7;
	}

	    /* Get pointers to the old and new symbols */
	pNewSym = pSymbolTable[pRemapSymArray[oldSymID].newSymbolID];
#ifdef DD_DEBUG
	printf("ID %d has count %d\n", pRemapSymArray[oldSymID].newSymbolID, pNewSym->instanceCount);
#endif


	    /* update the instance entries that will be used to
	     * replace the old symbol by pNewSym.  Yeah, if we were really
	     * good people, we'd fill out the entire instance thing,
	     * but forget it.
	     */
	instance.w = pRemapSymArray[oldSymID].width;
	instance.h = pRemapSymArray[oldSymID].height;
	instance.x = curX - instance.w + 1;
	instance.y = curY - instance.h + 1;
	instance.leftProfile = pRemapSymArray[oldSymID].leftProfile;
	instance.rightProfile = pRemapSymArray[oldSymID].rightProfile;

	    /* Insert the new symbol, with it's position adjusted
	     * relative to the old one, into the output position
	     * buffer.
	     */
	__error__ = ip_createUncommittedTCPosition (pIPState, pTextCompInfo,
						  &instance, pNewSym);
	if (__error__ != ia_successful)
	    goto finRemapSyms;

	    /* Go read/copy the next delta X in this line */
	positionInfo = *pOldPositionBuf++;
    }

	/* That's ALMOST it.  We need to flush out the last few
	 * uncommitted positions in the uncommitted list if the
	 * list isn't empty now (which I don't think can happen)  
	 * Set the number of bytes we actually used.  Also set the append
	 * one zero for y offset 0f 0 (=> end of position buf info)
	 */
    if (pTextCompInfo->numUncomPos > 0)
    {
	for (curY = pTextCompInfo->lowUncomY;
	     curY <= pTextCompInfo->hiUncomY;
	     curY++)
	{
	    __error__ = ip_commitTCPositions (pIPState, pTextCompInfo, curY);
	    if (__error__ != ia_successful)
		goto finRemapSyms;	/* don't bother continuing */
	}
    }

    if (pTextCompInfo->positionBufSize - pTextCompInfo->resultSize >= 1)
    {
	pTextCompInfo->pPositionBuf[0] = 0;
	*pBytesUsed = pTextCompInfo->resultSize + 1;
    }
    else
	__error__ = ia_bufferExceeded;	/* ran out of space */

finRemapSyms:
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_initializeTextColorCallback()
 *
 *	Initialize a routine to call back when connected components are
 *	found, which will install a color ID and either a color to the
 *	color list, or an image to the image list.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to structure containing
 *					information about symbols
 *		func:			callback function
 *		pPrivData:		private data to be passed to func
 *
 *	Action:
 *		As each connected component is found, this callback function
 *		will be called to determine the CCs color. It is the job of
 *		the callback function to add the color to the symbol, via
 *		i_attachTextColor() or i_attachTextImage().
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers, including
 *					the pointer to the head of the list
 *					of symbols, was NULL
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *
 **************************************************************************/
Int32  CDECL 
i_initializeTextColorCallback (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Int32 (CDECL *func)(IPState *pIPState, void *pTextCompInfo,
			void *pPrivData),
    void       *pPrivData)
{
    TCCompressionInfoPtr pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;

	/* This is only useful during lossy matches */
    if (func == NULL)
	return ia_invalidParm;

    pTextCompInfo->textColorCallback = func;
    pTextCompInfo->textColorPrivData = pPrivData;

	/* This implies that we will not be generating residual images, but
	 * instead, output only tokenized imagery. */
    return i_initializeTextNoResidual(pIPState, pTextCompInfoIn);
}


/**************************************************************************
 *   i_initializeTextNoResidual()
 *
 *	Initialize a routine to call back when connected components are
 *	found, which will install a color ID and either a color to the
 *	color list, or an image to the image list.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to structure containing
 *					information about symbols
 *
 *	Action:
 *		Signal no residual image to be output (multiple dictaries
 *		and one symbol compressed image).
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers, including
 *					the pointer to the head of the list
 *					of symbols, was NULL
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *
 **************************************************************************/
Int32  CDECL 
i_initializeTextNoResidual (
    IPState    *pIPState,
    void       *pTextCompInfoIn
    )
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;

	/* This signals that we will not be generating residual images, but
	 * instead, output only tokenized imagery (multiple dictionaries). */
    pTextCompInfo->useResidual = 0;

    NORMAL_RETURN;
}


/*

 $Log: i_tcenc.c,v $
 Revision 1.5  2001/05/08 06:43:37  jlin
 change ->allocator to ->ipcAllocator

 * Revision 1.4  2000/08/01  21:46:02  dabercro
 * Rename DEBUG to TC_DEBUG
 *
 Revision 1.3  2000/06/27 00:20:12  jlin
 remove PROLOGUE

 Revision 1.2  2000/02/01 21:43:36  dabercro
 Additional routine to identify singleton dictionaries (no residual selector image).

 Revision 1.1.1.1  1999/09/20 18:18:13  jlin
 Imported sources

 Revision 1.39  1998/08/04 01:04:00  aber
 aber on Mon Aug 3 18:03:09 PDT 1998

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Support multiple dictionaries, separated by single-use on page, and
 shared across one or more page.

 Revision 1.38  1998/06/23 21:45:37  aber
 aber on Tue Jun 23 14:45:36 PDT 1998

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 CommonCheck: Fixes to force Tundra dictionary reading order, esp. for multiple
 dictionaries per page. Some cosmetic comment enhancements. A couple of
 extra error conditions caught.

 Revision 1.37  1998/01/15 01:44:58  ddavies
 ddavies on Wed Jan 14 17:42:52 PST 1998

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Fix bug found by Andrew Miller.  When discarding symbols that have already
 been written out in a Tundra dictionary (discarding happens in
 i_convertImageToSymbols), use pNextSym = pSymbol->pNextNum, not
 pNextSym = pSymbol.

 Revision 1.36  1998/01/13 23:50:03  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.35  1997/05/08 02:34:53  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Remove change that allowed big symbols to change size by a litte during
 symbol consolidation.  The change improved compression a little and degraded
 OCR a little, so it wasn't worth it.

 Revision 1.34  1997/04/29 02:45:08  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Delete a bunch of smelly, obsolete comments that talked about matching using
 projections (it didn't work - why are these still around?)
 Generally fix up other obsolete comments.
 Check and set the exact match flags in i_convertImageToSymbols.
 Simplify the code that calls ip_symbolIsComplete, let value of pMatchedSymbol
 indicate whether a matching symbol was found.
 Check the exactMatch flags in routines that shouldn't be called when
 doing exact matching (symbol consolidation routines).

 Revision 1.33  1997/04/24 23:30:13  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Change ip_fixupSymbolPtrs to ip_fixupTCSymbolPtrs.

 Revision 1.32  1997/03/26 00:30:57  ddavies
 ddavies on Tue Mar 25 16:29:45 PST 1997

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Big old bug fix: take the vertRunHeadArray and vertRunTailArray out of
 the pointer fixup ring when freeing them in i_convertImageToSymbols.
 Otherwise, we end up with pointers to freed memory, which turns into a
 disaster later.

 Revision 1.31  1997/03/25 02:38:30  ddavies
 ddavies on Mon Mar 24 18:36:49 PST 1997

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Replace the tentative code used to keep small symbols small with permanent
 code.  There is now only one definition of small - either dimension is
 small - and one option for what do to with small things - don't let them
 (in whichever dimensions are *small*) get bigger.

 Revision 1.30  1997/02/12 03:03:21  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Many changes in which we generally convert the mass check into the
 strokeWidth check; start using pInstance->maxDimension instead of
 recomputing MAX(w, h) every time, and other changes that aid in
 aligning replacement symbols with the replaced symbols on the page.
 i_getSizeOfSymbolRemapArray is defined.  The array used by
 i_remapSymbolIDs is now filled with TCRemapSym structures instead of
 UInt32's.  These structures carry the information used to properly
 align the replacement symbol with the replaced symbol when position
 buffers are converted after symbol consolidation.
 All instances of pSymbolIDMap (pointer to UInt32) have been converted
 into pSymbolIDMapIn (pointer to void).  This hides the remap table
 from the client.
 i_consolidateShapeInfo now stores relevant info about a soon-to-be-replaced
 symbol in the remap array so that the replacement symbol may be properly
 aligned by i_remapSymbolIDs.
 i_remapSymbolIDs is somewhat fancier in that it allocates a larger
 position buffer when necessary (so the cheap end-of-buffer check works).
 It also uses a new routine, ip_createUncommittedTCPosition to turn an
 old symbol in the old position buffer into a new symbol in the new
 position buffer.  This means it has to initialize the stuff that the
 new routine needs and has to more carefully parse the old position buffer
 to get the x,y coords of each symbol instead of just copying them.

 Revision 1.29  1997/02/06 03:51:14  ddavies
 ddavies on Wed Feb 5 19:50:27 PST 1997

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Fix bug in hack that lets XIS experiment with various ways of dealing with
 very small objects.  Very small symbols were being allowed to change
 size during symbol consolidation.

 Revision 1.28  1997/02/05 01:19:18  ddavies
 ddavies on Tue Feb 4 17:18:10 PST 1997

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Check in experimental hack that XIS will use to determine the best way to
 handle small stuff in images.  When small stuff gets bigger, OCR can
 start mistaking it for punctuation, which is bad.

 Revision 1.27  1997/01/30 00:49:49  ddavies
 ddavies on Wed Jan 29 16:47:25 PST 1997

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Fix bug found when testing effects of symbol-based compression on OCR
 accuracy.  During symbol consolidation, we now require the stroke widths
 of the target symbol and the class's symbol to be very close.  This keeps
 urk, ensures that all all symbols in a class have the same stroke width.
 This, in turn, keeps us from doing violence to very small symbols when
 consolidating them.

 Revision 1.26  1997/01/08 04:28:28  ddavies
 ddavies on Tue Jan 7 20:27:13 PST 1997

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Add newPositionBufSize parameter to i_remapSymbolIDs.  The position buffer
 doesn't always get smaller, so we have to check to see if it overflows.
 The checking code breaks the remapping loop into two loops. The first one
 checks to see if we're getting close once per loop iteration.  The second
 one checks to see if the buffer is full before every byte is stored.

 Revision 1.25  1997/01/04 00:28:29  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Replace the mass and interior measures with the strokeWidth measurement.
 This does a much better job of distinguishing bold from non-bold characters.
 Since the strokeWidth measurment is more stable, we can check all symbols
 for strokeWidth.  We were only checking small (< 20 pix on a side) for mass
 and interior since they made too many spurious differences.
 Remove all references to interior in the Instance and TCSymbol structures.
 Also remove massRange in TCSymbol.

 Revision 1.24  1996/08/29 00:58:00  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Fix comments.  Delete any symbols that don't have runs (symbols read by
 the i_readTundraDictionary routine) before starting on
 i_convertImageToSymbols.
 In i_consolidateShapeInfo, turn off a symbol's pPrevNum and pNextNum pointers
 before deleting it so that the deleting code won't try to follow them.  They're
 pointing to trash at that point.

 Revision 1.23  1996/08/23 02:58:56  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Modify parsing of internal position info so the block begins with a deltaY
 relative to -1 instead of a deltaX.

 Revision 1.22  1996/07/27 02:01:24  ddavies
 ddavies on Fri Jul 26 18:58:55 PDT 1996

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Change TCCompressionInfo element symsHaveBeenEncoded to general encodeState
 structure element.
 Remove ip_compareSymCounts.  It went to mitcenc.c.
 Sort symbols in Tundra dictionary order before sorting them by both
 instanceCount and dictionary order in i_consolidateSymbols.  This more
 like what will happen in the Tundra world.  The new routine
 ip_sortTCSymbolsInTundraDictOrder is used for this.

 Revision 1.21  1996/07/22 18:36:22  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Add mild improvements for vertical symbol placement.
 The lastX entry in TCCompression info is removed and entries that support
 a list of uncommitted symbol positions (dum1stLoc, dumLastLoc, lowUncomY,
 hiUncomY, numUncomPos, pLastUncom) are all added.  i_convertImageToSymbols
 now initializes all these and calls ip_commitTCPositoins to commit the
 positions for the last symbols at the end of the image.
 The general scheme is explained in mitcenc.c.

 Revision 1.20  1996/07/19 00:02:47  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Lots of Tundra changes.
 In i_convertImageToSymbols, see if some of the symbols in the existing
 dictionary have already been encoded into a Tundra dictionary or
 residual block.  If so, find and delete them.  This should prevent a
 client from putting a symbol into two different dictionaries by mistake.
 Convert all reference to position block header so they use the new
 cTCPos... constants.
 Destroy the cached table that associates symbol IDs with symbols at the
 beginning of i_convertImageToSymbols.  This is done because the number
 of symbols is about to change.
 Change the name of the static procedure ip_compareSyms to
 ip_compareSymCounts.  This is to avoid confusion with ip_compareSymbols
 in this module.
 Create and use the symbol table that associates symbol IDs with symbols
 in i_consolidataSymbols.  Ooops, I meant to say that the symbol ID
 table is deleted in i_consolidateShapeInfo since the number of symbols
 is about to change.
 Use the new ip_deleteTCDictionarySymbol procedure to delete symbols
 instead of doing it in an ad hoc fashion.
 Fuss with indenting.
 Fix comments on position block format.

 Revision 1.19  1996/07/09 00:15:37  ddavies
 ddavies on Mon Jul 8 17:14:31 PDT 1996

 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Change the comparison routine passed to qsort so that it never says two
 items are equal.  The result of sorting lists in which some items are equal
 is platform dependent.  We now sort by instanceCount first and original
 symbol ID second.

 Revision 1.18  1996/07/04 03:00:48  ddavies
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 Change order of symbol linked list, so p1stSymbol points to symbol with
 lowest symbol ID instead of symbol with the highest symbol ID.
 This also means that pSymbol->pNextNum points to the symbol with the next
 higher symbolID instead of the next lower one.
 Also, and more important, change i_consolidateShapeInfo so that the
 symbol IDs of the consolidated symbols are ordered by usage, with the
 most used symbols having the smallest IDs.  This can significantly (10%?)
 compress the position information.
 Finally, because some symbol IDs may now get longer, change i_remapSymbolIDs
 so that it allocates a temporary buffer (and calls itself with that) if the
 input and output buffers are the same buffer.  In the old world, we could
 count on symbol IDs never getting longer, so we could always copy down
 inside one buffer.

 Revision 1.17  1995/12/13 02:46:29  ejaquith
 i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.16  1995/10/18  03:59:07  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.15  1995/07/10  23:51:53  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Change the pointer to an array of TCSymbol pointers in a TCCompressionInfo
 * structure to a pointer to a TCSymbolArray.  This structure has a PointerFixup
 * struct at the beginning.  Do the same for the two arrays of pointers to
 * TCScanLineRun's used to convert horizontal to vertical runs.  Add args to
 * calls to i_allocateMemoryBlock to say we should be notified if the
 * blocks move.  Initialize the PointerFixup structures at the beginning of
 * the arrays of pointers.  Use the number of pointers inside the
 * TCScanLineArray structures instead of the width in the TCCompressionInfo
 * (that's gone).  Initialize the pointer arrays so Purify won't complain when
 * we check them after pointer fixup.  Change the casts in calls to
 * i_allocateStruct from (TCBlockHeader **) to (void **).  This is more
 * correct than pretending every structure in every heap is a
 * TCBlockHeader just because that structure began with a pNext pointer.  And,
 * a TCBlockHeader no longer begins with a pNext pointer.
 *
 * Revision 1.14  1995/06/08  20:54:58  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Added symbol mass and number of runs to ip_classifySymbol's parameters.
 * That's because this is a callback for i_findConnectedComponentRuns.
 * Use the mass passed in instead of calculating it from the runs.
 *
 * Revision 1.13  1995/04/01  03:20:37  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Change references to pTextCompInfo->symIndex to (*(pTextCompInfo->pSymIndex)).
 * That's because the TCCompressionInfo structure now has a pointer to a 2D
 * array of symbol pointers instead of the array itself.
 *
 * Revision 1.12  1995/03/15  02:39:29  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 *  Broke ip_classifySymbol into ip_classifySymbol which assembles
 * data on a new instance and call ip_compareSymbols for each set of potential
 * matches of a given width and height.  Introduced the TCInstance data
 * structure which is passed to ip_compareSymbols, ip_calculateSlant and
 * ip_symbolIsComplete instead of a bunch of arguments.  Have ip_classifySymbol
 * look at symbols arranged in a 2D array by [width, height] instead of looking
 * through a 1D array arranged by height.  This gives us more control and tends
 * to find things of the correct size first.
 * Delete the asymmetry test.  Have ip_compareSymbols attempt to align the
 * horizontal and vertical runs of symbols before comparison.  This involves
 * duplicating the runs on the first horizontal scan line, the first vertical
 * scan column, the first horizontal group and the first vertical group.
 * The routine compares hSAsymmetry for the exemplar and instance to decide
 * how far to shift one with respect to the other when preparing horizontal
 * runs and vSAsymmetry when preparing to compare vertical runs.  Added a
 * feature that calculates and checks how much the instance is wobbling around
 * within the legal range of endpoints for small, round, dense symbols.
 * Split the volume checks into horizontal and vertical volume independently.
 * The volume is now calculated using only long runs in an attempt to ignore
 * roundy things.  Add checks for minimum and maximum width and height.  These
 * are carried in each symbol and instance.  The are used to make sure that
 * a consolidated symbol equivalance class contains only symbols in a narrow
 * range of widths and heights.  Change ip_consolidateSymbols so that it
 * just passes the correct horizontal runs to ip_classifySymbol (doesn't keep
 * making and unmaking them on each call).  Also have ip_consolidateSymbols
 * restrict the range of equivalence class widths and heights each time
 * a symbol joins a class.  Added debug counts to ip_consolidateSymbols so we
 * can see how often each type of offset causes symbols to be consolidated.
 * Change the calculation of the extra scan line's runs in ip_consolidateSymbols
 * so that it only puts 1 pixel in the middle of each run being duplicated.
 *
 * Revision 1.11  1995/01/11  22:32:25  dferrell
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Added include of shrcc.pub since run and CC function prototypes were moved.
 *
 * Revision 1.10  1994/11/23  05:56:03  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Turn off debugging checks - gives a modest performance improvement.
 * Fix the process of initializing symbols that have been read in from
 * a bytestream.  The massRange, interior, numActiveRuns and instanceCount
 * parameters weren't being set.  The volumeDelta calculation had a
 * divide by 0 bug.  We were trying to free vertical runs using the group
 * heap instead of the run heap.
 *
 * Revision 1.9  1994/11/10  21:49:00  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Changes to increase robustness and symbol discrimination.  Amount by which
 * mass is allowed to vary is changed.  Volume is a new measure used to
 * distinguish little, round bold blobs.
 * Also changed processing for Check1Run so that it falls through to
 * CheckEndpoints if it fails.  This helps tolerate noise on the edges of
 * characters and in the middles.
 * Change way noise is added onto the bottoms of characters so that we
 * only add 1 dot in the middle of each run instead of replicating the
 * entire run.  This helps when replicating the entire run causes us to fail
 * because the entire run is too long.
 *
 * Revision 1.8  1994/10/14  01:21:12  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Remove frames from image width and height passed into i_convertImageToSymbols.
 * Free the vertical run arrays if they're not wide enough for the current
 * image.  They'll be reallocated immediately below.  Set the new
 * vertRunArrayWidth in pTextCompInfo with the width.  Add new insertOnly
 * flags to calls to ip_classifySymbol.  In i_consolidateSymbols, fix bug
 * that caused us not to delete a redundant symbol if it was the last one
 * in the list after rearrangement.
 *
 * Revision 1.7  1994/09/08  18:57:35  ddavies
 * ddavies on Thu Sep 8 11:57:26 PDT 1994
 *
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Cast arg to memset to keep acc happy.
 *
 * Revision 1.6  1994/09/08  00:32:23  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Zillions of changes.  We now match both horizontal and vertical runs.  The
 * slant calculation is used to find italics.  The interior count separates
 * little commas from little blobs.  We no longer use a tree to do discrimination,
 * the run groups for each symbol are in individual lists.
 *
 * Revision 1.5  1994/08/03  19:17:16  ddavies
 * ddavies on Wed Aug 3 12:16:46 PDT 1994
 *
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Snapshot of version of symbol-based compression code that does a good job
 * of distinguishing symbols.
 *
 * Revision 1.4  1994/07/14  21:03:47  ddavies
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Shapshot of top level module containing the compression code for symbol-based
 * compression.  Almost everything in here changed from last time.  This is
 * a snapshot, the code is still under test.
 *
 * Revision 1.3  1994/05/28  00:08:31  ddavies
 * ddavies on Fri May 27 17:08:11 PDT 1994
 *
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Fixed zillions of bugs, this is a good time for a checkpoint.
 *
 * Revision 1.2  1994/05/03  20:44:47  ddavies
 * ddavies on Tue May 3 13:44:24 PDT 1994
 *
 * i_tcenc.c -> /product/ipcore/ipshared/src/RCS/i_tcenc.c,v
 *
 * Includes general improvements for symbol-based compression.  It's not
 * as good as it will be, but this is a reasonable checkpoint.
 *
 * Revision 1.1  1994/04/26  02:16:13  ddavies
 * Initial revision
 *

 */


