/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tcence.c: Contains procedures used to turn an image into a
 *		set of symbols and positions of symbols using exact
 *		matching between symbols and templates.  The procedure
 *		used to do the reverse transform is in i_tcdec.c
 *
 *		Int32	i_convertImageToSymbolsExactly()
 *			Classify the symbols in an image into a set of
 *			equivalence classes.  All symbols in an
 *			equivalence class should match exactly.
 *			Mark the position of each symbol in a data block.
 *			Build the data structures used to do the
 *			classification.
 *
 *		Int32	ip_classifySymbolExactly()
 *			This procedure is called with a list of all the
 *			runs comprising a symbol.  It classifies the
 *			symbol and puts out position info.
 *
 *		Int32	ip_compareSymbolsExactly()
 *			This procedure is used by ip_classifySymbolExactly to
 *			compare an instance with a list of exemplars.
 *			It returns the first match.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#include "memory.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /*  prototyes */
#include "shrcc.pub"      /* for i_findConnectedComponentRuns prototype */
#include "shrcmprs.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tcence.c,v $; $Revision: 1.2 $; $Date: 2001/05/08 06:43:37 $");

#define	DEBUG	0		/* used to put out debugging info */

/* The process we use to compress images by finding repeated symbols is
 * well documented (at least there's lots of documentation) in the
 * file i_tcompr.c.  But that's all about doing fuzzy matching in the
 * presence of noise.  The routines in this file do exact matching.
 *
 * Who needs exact matching?  We've thought of a few applications.  First,
 * one may do a lossy compression, then a decompression, then want to
 * do a lossless compression.  Why?  Well, umm, if you were editing the
 * document, the bitmap would change, but presumably, the symbols wouldn't
 * (or would all change the same way) and you wouldn't want any more loss.
 * Another, poorer, application is for people who want to make sure
 * there is no loss.  However, plain old CCITT Group 4 will probably do
 * a better job on scanned documents.  The best application is
 * "print to XIFF."  In this one, the print driver produces a series
 * of color images and bitmaps.  The bitmaps contain perfectly formed
 * font symbols.  Losslessly compressing them is not only faster, but it
 * guarantees that there will be no subtle substitutions of even the smallest
 * characters.
 *
 * This file contains the routine used to turn a page containing connected
 * components into a list of positions and symbol IDs and a set of
 * data structures we can use to do further classification.  The format
 * of the list of positions is given in i_tcompr.c.  The data
 * structures are defined in shrcmprs.prv.
 *
 * This routine calls i_findConnectedComponentRuns, giving it a callback
 * proc that actually classifies each symbol as it is found.  The callback
 * proc described here does exact matching instead of fuzzy matching.
 *
 */


/* PUBLIC */
/**************************************************************************
 *   i_convertImageToSymbolsExactly()
 *
 *	This procedure converts a binary image of repeated symbols, usually
 *	characters, into a set of positions, symbol ID's and a data
 *	structure used to classify symbols.
 *
 *	Input:
 *		pImage:			Pointer to upper left corner of image
 *					INCLUDING frame.
 *		width:			image width including frame
 *		height:			image height including frame
 *		bpl:			The image's bytes/line
 *		pTextCompInfoIn:	pointer to structure containing
 *					information about symbol shapes
 *		pageID:			Page ID assigned to parent image
 *		pPositionBuf:		pointer to buffer that will contain
 *					position and symbol ID's.
 *		positionBufSize:	size of that buffer in bytes.
 *		pResultSize:		ptr to UInt32 that will hold the
 *					number of bytes actually used in
 *					the position buffer
 *
 *	Action:
 *		Builds classification data structures and writes position
 *		and symbol ID info to *pPositionBuf.  If an error is
 *		encountered, the classification data structures are NOT
 *		freed.  This allows the client to try again with more
 *		memory.  The client may explicitly free them.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *		ia_bufferExceeded	position and symbol ID info didn't
 *					fit into the position buffer.
 *
 **************************************************************************/
Int32  CDECL 
i_convertImageToSymbolsExactly (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	width,
    Int32	height,
    UInt32	bpl,
    void       *pTextCompInfoIn,
    UInt32	pageID,
    UInt8      *pPositionBuf,
    UInt32	positionBufSize,
    UInt32     *pResultSize)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo;	/* ptr to compression info */
    TCSymbol   *pSymbol, *pNextSym;
    TCScanLineRun *pNextRun;		/* used to free vertical runs */
    Int32	index;
    Int32	classifySymRetCode;	/* return code left
					 * by ip_classifySymbol when something
					 * went wrong */
    

	/* make sure all the pointers are ok */
    if ((pImage == NULL) || (pTextCompInfoIn == NULL) ||
	(pPositionBuf == NULL) || (pResultSize == NULL) ||
	(positionBufSize == 0))
	return ia_invalidParm;

	/* convert the pointer to the correct type */
    pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;

	/* If we haven't done any matching yet, set the exactMatch flags.
	 */
    if (!pTextCompInfo->exactMatchSet)
    {
	pTextCompInfo->exactMatchOnly = cTrue;	/* we're doing exact matches */
	pTextCompInfo->exactMatchSet = cTrue;
    }
    else	/* were we doing lossy matches? */
    if (pTextCompInfo->exactMatchOnly == cFalse)
	return ia_invalidParm;			/* can't switch back */

	/* If the internal data structure contains symbols that
	 * have already been put out in a Tundra dictionary or
	 * residual image, delete them.  Leaving them around will
	 * only create confusion.  Also delete any run-less
	 * symbols that were put here by i_readTundraDictionaryBlock.
	 */
    if (pTextCompInfo->encodeState & cTCSymsHaveBeenEncoded)
    {
	for (pSymbol = pTextCompInfo->p1stSymbol;
	     pSymbol != NULL;
	     pSymbol = pNextSym)
	{
	    pNextSym = pSymbol->pNextNum;		/* pt to next symbol */
	    if ((pSymbol->runsPtToTail) ||
		((pSymbol->pRunTail == NULL) && (pSymbol->pRunHead == NULL)))
		ip_deleteTCDictionarySymbol (pIPState, pTextCompInfo, pSymbol);
	}
	    /* Go reassign the symbol IDs */
	for (index = 0, pSymbol = pTextCompInfo->p1stSymbol;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextNum, index++)
	{
	    pSymbol->symbolID = index;
	}
	    /* Set the next symbol ID for this dictionary and
	     * declare that the encode state is null.
	     */
	pTextCompInfo->nextSymbolNum = index;
	pTextCompInfo->encodeState = 0;
    }

	/* If there's already a cached table to pointers to symbols,
	 * delete it.  The table will be invalid after we add
	 * some symbols.
	 */
    __error__ = ip_destroyTCSymbolTable (pIPState, pTextCompInfo);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* i_findConnectedComponentRuns will do the clipping, so
	 * don't do any here.*/

	/* set the temporary vars in *pTextCompInfo that hold data about
	 * the position buffer.  We put them in here because it makes
	 * it easy to pass them around this way.  We want the real image
	 * width and height after this, so subtract off the frames now.
	 * Note that we initialize lastY to -1 so that the first
	 * deltaY offset recorded for any non-empty image is non-zero
	 * and the first deltaY offset recorded for a blank image is zero.
	 */
    width -= 2*cFrameBits;
    height -= 2*cFrameBits;
    pTextCompInfo->imageWidth = (Int16) width;
    pTextCompInfo->imageHeight = (Int16) height;
    pTextCompInfo->pPositionBuf = pPositionBuf;
    pTextCompInfo->positionBufSize = positionBufSize;
    pTextCompInfo->resultSize = 0;		/* used no bytes so far */
    pTextCompInfo->lastX = 0;
    pTextCompInfo->lastY = -1;

	/* nextSymbolNum is initialized when the text compression
	 * info structure is allocated.  That way, we can continue
	 * to compress pages with the same dictionary of symbols.
	 * Make sure we build the shape tree by turning off
	 * the classifyOnly flag.
	 */
    pTextCompInfo->classifyOnly = cFalse;

	/* Check each entry in the textCompInfo.  If an entry is NULL,
	 * allocate the corresponding initial heap or memory block
	 * and update the textCompInfo entry.  These heaps  and blocks
	 * are freed when we free the text compression info.  This
	 * allows us to compress several pages in a row without fussing
	 * with the heaps all the time.
	 */

    if (pTextCompInfo->p1stScanLineBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stScanLineBlock),
					 sizeof(TCScanLineRun),
					 cTCScanLineRunBlockSize,
					 cTrue,		/* adjust ptrs after
							 * garbage collection */
					 &(pTextCompInfo->fixup),
					 &ip_fixupScanLinePtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

    if (pTextCompInfo->p1stSymbolBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stSymbolBlock),
					 sizeof(TCSymbol),
					 cTCSymbolBlockSize,
					 cTrue,
					 &(pTextCompInfo->fixup),
					 &ip_fixupTCSymbolPtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

	/* If we already have a list of symbols, but they don't have the
	 * classification structures, add the classification structures
	 * to each symbol.
	 */
    pSymbol = pTextCompInfo->p1stSymbol;
    if ((pSymbol != NULL) && (pSymbol->numRuns == 0))
    {
	    /* build classification data for all the symbols */
	for (;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextNum)
	{
		/* compute the mass for this symbol */
	    for (index = 0, pNextRun = pSymbol->pRunTail;
		 pNextRun != NULL;
		 pNextRun = pNextRun->pNext)
	    {
		index += pNextRun->runRight - pNextRun->runLeft + 2;
		pSymbol->numRuns++;
	    }
	    pSymbol->mass = index;

	    pSymbol->numActiveRuns = 0;	/* nothing merged with this yet */
	    pSymbol->instanceCount = 1;	/* it would be nice if we kept
					 * this around during the bytestream
					 * processing, but we didn't.  So
					 * we just assume there's at least
					 * one instance of this symbol
					 * somewhere.
					 */

		/* Record all the non-zero info that would have been
		 * calculated by ip_calculateSlant.
		 */
	    pSymbol->maxDimension= MAX(pSymbol->width, pSymbol->height);
	}
	if (__error__ != ia_successful)
	    NORMAL_RETURN;

    }

	/* Start the position info with a magic number and version number.
	 */
    pTextCompInfo->pPositionBuf[cTCPosMagicNum0] =
				(UInt8)(cTextCompPositionInfoMagicNum);
    pTextCompInfo->pPositionBuf[cTCPosMagicNum1] = 
				(UInt8)(cTextCompPositionInfoMagicNum >> 8);
    pTextCompInfo->pPositionBuf[cTCPosMagicNum2] = 
				(UInt8)(cTextCompPositionInfoMagicNum >> 16);
    pTextCompInfo->pPositionBuf[cTCPosMagicNum3] = 
				(UInt8)(cTextCompPositionInfoMagicNum >> 24);

    pTextCompInfo->pPositionBuf[cTCPosVer0] = 
				(UInt8)(cTextCompPositionInfoVersionNum);
    pTextCompInfo->pPositionBuf[cTCPosVer1] = 
				(UInt8)(cTextCompPositionInfoVersionNum >> 8);

	/* Add image width and height, 2 bytes each */
    pTextCompInfo->pPositionBuf[cTCPosWidth0] = (UInt8) width;
    pTextCompInfo->pPositionBuf[cTCPosWidth1] = (UInt8) (width >> 8);

    pTextCompInfo->pPositionBuf[cTCPosHeight0] = (UInt8) height;
    pTextCompInfo->pPositionBuf[cTCPosHeight1] = (UInt8) (height >> 8);

	/* Set the rest of the statistics to 0 (and invalid) */
    memset(&(pPositionBuf[cTCPosStatsValid]), 0, cTCPosStatsLen);

    pTextCompInfo->pPositionBuf += cTCPosHeaderVer2Len;
    pTextCompInfo->resultSize += cTCPosHeaderVer2Len;

	/* Allow the page ID to propogate once a symbol has been added to the
	 * dictionary */
    pTextCompInfo->curPageID = pageID;

	/* Call i_findConnectedComponentRuns using ip_classifySymbol as
	 * the callback proc.  ip_classifySymbol will be called once for
	 * each connected component found in the image.
	 */
    __error__ = i_findConnectedComponentRuns(
				pIPState,
				pImage,
				bpl,
				cFrameBits,		/* clipX */
				cFrameBits,		/* clipY */
				width,			/* clipW */
				height,			/* clipH */
				1,			/* minWidth */
				1,			/* minHeight */
				1,			/* 8 connected */
				(void *)pTextCompInfo,	/* pRectData */
				&ip_classifySymbolExactly, /* rectReturnProc */
				&classifySymRetCode);	/* pRectReturnCode */

    if (__error__ != ia_successful)	/* something bad happened */
    {
	if (__error__ == ia_callbackError)	/* ip_classifySym complained */
	    __error__ = classifySymRetCode;
	goto finConvertImageToSym;
    }

	/* write the two 0's at the end of the file.  The first one means
	 * delta Y is coming.  The second one says delta Y == 0, which
	 * means end of file.
	 */
    if (pTextCompInfo->positionBufSize - pTextCompInfo->resultSize >= 2)
    {
	pTextCompInfo->pPositionBuf[0] = 0;
	pTextCompInfo->pPositionBuf[1] = 0;
	*pResultSize = pTextCompInfo->resultSize + 2;
    }
    else
	__error__ = ia_bufferExceeded;	/* ran out of space */

finConvertImageToSym:
	/* return the final return code */
    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_classifySymbolExactly()
 *
 *	This procedure is called by i_findConnectedComponentRuns for
 *	each connected component in the image.  It does exact matches
 *	with symbols already in the dictionary.
 *
 *	Input:
 *		pTextCompInfoIn:	Pointer to image's textCompInfo.
 *		x, y, w, h:		symbols's bounding box
 *		symMass:		mass of symbol
 *		nRuns:			number of runs in symbol
 *		pRuns:			Pointer list of runs associated
 *					with the symbol.
 *
 *	Action:
 *		Attempts to classify the symbol passed in as one of the
 *		existing symbols.  If this is not possible, expands the
 *		shape tree as needed and creates a new symbol.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *		ia_bufferExceeded	Ran off the end of the position buf
 *
 **************************************************************************/
Int32  CDECL 
ip_classifySymbolExactly (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	symMass,
    Int32	nRuns,
    TCScanLineRun *pRuns)
{
PROLOGUE;
 TCCompressionInfoPtr	 pTextCompInfo;	/* pointer to image's information */
TCScanLineRun		*pCurRun, *pTempRun;	/* temporary run pointers */
TCInstance		 instance;	/* struct containing instance's info */
TCSymbol		*pExemSymbol;	/* pointer to current exemplar */
TCSymbol		*pMatchedSymbol; /* pointer to symbol that matches
					 * the instance. */
Int32			 xOrigin;	/* x origin of symbol */
UInt32			 lineFlags;	/* flags used to tell the runs in
					 * one line from the runs in
					 * adjacent lines */


	/* point to the real text compression info */
    pTextCompInfo = (TCCompressionInfoPtr) pTextCompInfoIn;

	/* find the x origin of the symbol.  Note that we still use
	 * 2x coords here, even though we're doing exact matching.
	 * This maintains compatibility with the rest of the
	 * encoding and decoding code.
	 */
    xOrigin = 2*x + w - 1;		/* get the "average" x position.
					 * The width is the distance from
					 * the left edge of the leftmost
					 * pixel to the right edge of the
					 * rightmost pixel.  Since we want
					 * the distance to the middle of
					 * the character, we want x + w/2 - 1/2.
					 * Since all coordinates keep an
					 * extra bit of precision, we use
					 * 2x + w - 1.
					 */

	/* subtract off the frame to reduce my confusion.  We have to do
	 * this after the stuff above this so that the values of x and
	 * xOrigin are consistent.
	 */
    x -= cFrameBits;
    y -= cFrameBits;

	/* since we're doing exact matching, don't bother with
	 * minWidth, maxWidth, minHeight and maxHeight.
	 */

	/* Record the instance's position and dimensions as passed in */
    instance.x = x;
    instance.y = y;
    instance.w = w;
    instance.h = h;
    instance.pCurSymbol = NULL;		/* This symbol doesn't correspond
					 * to any other.
					 */

	/* Adjust all the x coordinates inside the runs so they're
	 * relative to the xOrigin.  Process the
	 * very first real run outside the loop so that
	 * pPrev will point to something meaningful inside the loop.
	 * Note that we keep an extra bit of significance below
	 * the binary point, i.e. a binary coordinate of 0x2 is really
	 * +1, a binary coordinate of 0x3 is really +1.5.
	 * Also mark the runs of each scan line with a 2 bit tag - 10 or 01.
	 * The bits alternate each time we cross a scan line boundary
	 * We use a 2 bit tag so a dummy having neither bit set can be
	 * used to terminate either type of run.
	 * Also record the symbol's mass and number of runs.  These are
	 * discrimination criteria.
	 */
    instance.mass = symMass * 2;
    instance.numRuns = nRuns;
    pTempRun = pRuns;
    lineFlags = cEvenLineTag;
    pTempRun->runLeft = (Int16) (2 * pTempRun->runLeft - xOrigin);
    pTempRun->runRight = (Int16) (2 * pTempRun->runRight - xOrigin);
    pTempRun->pClient = (void *)lineFlags;
    for (pCurRun = pTempRun->pNext;
	 pCurRun != NULL;
	 pCurRun = pCurRun->pNext)
    {
	pCurRun->runLeft = (Int16) (2 * pCurRun->runLeft - xOrigin);
	pCurRun->runRight = (Int16) (2 * pCurRun->runRight - xOrigin);
	if (pCurRun->runRight >= pTempRun->runLeft - 2)
	    lineFlags ^= cLineTagMask;	/* swap the flags */
	pCurRun->pClient = (void *)lineFlags;
	pTempRun = pCurRun;
    }

	/* Record the location of the horizontal runs in the instance
	 * structure.  This keeps everything tidy.
	 */
    instance.pHorizRuns = pRuns;

	/* Start comparing the instance symbol against the exemplars.  Note
	 * that pExemSymbol only gives us a starting point, we need to
	 * decide when to stop on our own.  We also stop if we found a match.
	 * Note that all symbols whose width or height > cMaxSymIndexDim
	 * are in the list pointed to by pTextCompInfo->pBigSymbols.
	 */
    pMatchedSymbol = NULL;

    if ((w <= cMaxSymIndexDim) && (h <= cMaxSymIndexDim))
    {
	pExemSymbol = (pTextCompInfo->pSymIndex->array)[w][h];
	if (pExemSymbol != NULL)
	{
	    __error__ = ip_compareSymbolsExactly (pIPState, pTextCompInfo,
					        &instance,
					        pExemSymbol,
					        &pMatchedSymbol);
	    if (__error__ != ia_successful)
		goto tcClassifySymCleanup;
	}
    }
    else	/* one of the indices was too big, so see if we can
		 * use the pointer to big symbols */
    if (pTextCompInfo->pBigSymbols != NULL)
    {
	__error__ = ip_compareSymbolsExactly (pIPState, pTextCompInfo,
					    &instance,
					    pTextCompInfo->pBigSymbols,
					    &pMatchedSymbol);
	if (__error__ != ia_successful)
	    goto tcClassifySymCleanup;
    }

	/* If we found a match, record the symbol's ID in the position
	 * info and leave.  Otherwise, go make up a new symbol.
	 * Which choice is taken depends on whether pMatchedSymbol
	 * is NULL (no matching symbol) or not (found a match).
	 */
    __error__ = ip_exactSymbolIsComplete (pIPState, pTextCompInfo,
				        pMatchedSymbol,
				        &instance,
				        cFalse);

	/* if things went fine, __error__ equals ia_successful at this
	 * point.  If not, it doesn't.  The runs pointed to by pRuns
	 * will be freed by the caller, so don't bother with that.
	 * We haven't allocated a symbol here, so there's no need to free
	 * that.
	 */
tcClassifySymCleanup:
    NORMAL_RETURN;
}

/* PRIVATE */
/**************************************************************************
 *   ip_compareSymbolsExactly()
 *
 *	This procedure is called by ip_classifySymbolsExactly to match
 *	an instance against a list of exemplars.  It does exact matches
 *	with symbols already in the dictionary.
 *
 *	Input:
 *		pTextCompInfo:		Pointer to text compression info
 *		pInstance:		Pointer to info about instance.
 *		pExemSymbol:		Pointer to first potential exemplar
 *		ppMatchedSymbol:	Pointer to pointer to exemplar that
 *					matches the instance.
 *
 *	Action:
 *		Compares the instance against each of the exemplars in the
 *		list.  If one matches, it's pointer is put into
 *		*ppMatchedSymbol.  Otherwise, *ppMatchedSymbol is unaffected.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_internal		internal failure
 *
 **************************************************************************/
Int32  CDECL 
ip_compareSymbolsExactly (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCInstance *pInstance,
    TCSymbol   *pExemSymbol,
    TCSymbol  **ppMatchedSymbol)
{
    PROLOGUE;
    TCScanLineRun	*pInstRun;	/* points to current scan line run */
    TCScanLineRun	*pExemRun;	/* points to previous scan line run */
    

    pTextCompInfo = pTextCompInfo;	/* in case DEBUG is turned off */
    __error__ = ia_successful;		/* assume the best */

	/* Look through the list of exemplars */
    while (pExemSymbol != NULL)
    {
	    /* We know the symbol sizes match exactly or we wouldn't
	     * be here (lists of exexmplars are chosen by size).
	     * Therefore, there's no need to check that.
	     * However, do check the mass and number of runs.
	     * These are the only simple checks.
	     */
	if ((pInstance->mass != pExemSymbol->mass) ||
	    (pInstance->numRuns != pExemSymbol->numRuns))
	{
#if DEBUG
	    pTextCompInfo->slantCheck++;
#endif
	    goto tcTryNextSym;
	}

	/* If that worked, we can just compare runs.  Note that
	 * the run comparison is easy.  We know both the instance
	 * and exemplar have the same number of runs from the
	 * check above.  The linetags don't matter since they're
	 * derived from the left and right endpoints.  All we have
	 * to check are the left and right endpoints.
	 * Obviously, we also don't have to check the vertical runs
	 * since all the info is in the horizontal runs.
	 */
	for (pInstRun = pInstance->pHorizRuns,
		pExemRun = pExemSymbol->pRunTail;
	     pInstRun != NULL;
	     pInstRun = pInstRun->pNext, pExemRun = pExemRun->pNext)
	{
	    if ((pInstRun->runLeft != pExemRun->runLeft) ||
		(pInstRun->runRight != pExemRun->runRight))
	    {
#if DEBUG
/* DEBUG */
if ((topPrintBound < pInstance->y) && (leftPrintBound < pInstance->x) &&
    (bottomPrintBound > pInstance->y) && (rightPrintBound > pInstance->x))
fprintf(stderr, "	CheckRun: Instance run at [%d, %d] does not equal exemplar run at [%d, %d]\n", pInstRun->runLeft, pInstRun->runRight, pExemRun->runLeft, pExemRun->runRight);
#endif
		goto tcTryNextSym;
	    }
	}

	    /* If we fall out of this loop, everything matched, so
	     * declare success and leave.
	     */
	*ppMatchedSymbol = pExemSymbol;
	break;

	/* we get here if there was a mismatch.  Go try the next symbol */
tcTryNextSym:
	pExemSymbol = pExemSymbol->pNext;
    }

    NORMAL_RETURN;
}


/*

 $Log: i_tcence.c,v $
 Revision 1.2  2001/05/08 06:43:37  jlin
 change ->allocator to ->ipcAllocator

 * Revision 1.1.1.1  1999/09/20  18:18:13  jlin
 * Imported sources
 *
 Revision 1.5  1998/08/04 00:17:00  aber
 aber on Mon Aug 3 17:17:00 PDT 1998

 i_tcence.c -> /product/ipcore/ipshared/src/RCS/i_tcence.c,v

 CommonCheck: Retain the page ID such that the dictionary can be split into a
 shared dictionary, and a single-use dictionary for each page.
 Support multiple dictionaries per page.

 Revision 1.4  1998/01/15 01:45:17  ddavies
 i_tcence.c -> /product/ipcore/ipshared/src/RCS/i_tcence.c,v

 Fix bug found by Andrew Miller.  When discarding symbols that have already
 been written out in a Tundra dictionary (discarding happens in
 i_convertImageToSymbols), use pNextSym = pSymbol->pNextNum, not
 pNextSym = pSymbol.

 Revision 1.3  1998/01/13 23:50:08  ddavies
 i_tcence.c -> /product/ipcore/ipshared/src/RCS/i_tcence.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1997/05/08 02:35:25  ddavies
 i_tcence.c -> /product/ipcore/ipshared/src/RCS/i_tcence.c,v

 Add the insertOnly parameter to call to ip_exactSymbolIsComplete.

 Revision 1.1  1997/04/29 02:38:25  ddavies
 Initial revision


 */

