/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tcompr.c: Comments and utility procedures used in symbol-based
 *		compression of text.
 *
 *   	Int32	i_estimateSizeOfPositionInfo()
 *			Estimates the size of the block of memory required to
 *			hold a list of symbol positions for an image.
 *
 *   	Int32	i_estimateSizeOfSerializedCompressionInfo()
 *			Estimates the amount of memory required to hold
 *			serialized versions of the stroke and symbol trees.
 *
 *	Int32	i_allocateTextCompressionInfo()
 *			This routine allocates the structure used to hold
 *			information about the stroke and symbols trees.
 *			This structure is passed to the other routines.
 *
 *	Int32	i_freeTextCompressionInfo()
 *			This routine frees all the text compression info
 *
 *	Int32	ip_createTCSymbolTable()
 *			This routine creates a table that maps symbol IDs
 *			to pointers to symbols.
 *
 *	Int32	ip_createTCSymbolTable()
 *			This routine destroys the table created by
 *			ip_createTCSymbolTable.
 *
 *	Int32	ip_deleteTCDictionarySymbol()
 *			This routine deletes a TCSymbol and all the
 *			data structures associated with it.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"

    /*  prototyes */
#include "shrip.pub"
#include "shrcc.prv"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#include "ptrfixup.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tcompr.c,v $; $Revision: 1.2 $; $Date: 2001/05/08 06:43:37 $");

/* This is the BIG COMMENT that tells the overall story of our approach
 * to symbol-based compression.
 */

/* Images are very big, so there is a lot of interest in compression.
 * There are all sorts of image compression algorithms, some tuned to high
 * performance, some requiring hardware to be practical, some meant for
 * pretty pictures, some for text and line art.  The routines described
 * here are intended to compress machine-printed text and line art.
 * They might work on halftones too, but that would be a happy coincidence.
 * We would like to get significantly better compression than MMR (used in
 * CCITT Group 4 fax machines).  We're willing to go slower, but still want
 * good performance.
 *
 * The standard approach to text compression is to turn a bitmap
 * into a sequence of runs of black and white pixels.  MH encoding does
 * one line at a time.  MR and MMR encode each run relative to a run
 * on the previous line if possible.  Since there are lots of vertical
 * features in text, this can significantly improve compression.
 *
 * These run-length compression schemes are very efficient in images that have
 * long horizontal runs of white and black.  Text, being largely composed
 * of vertical or diagonal strokes, has lots of short runs.  Halftones
 * and dithered documents really drive these things nuts.  Nonetheless,
 * these lossless schemes are the standards to which all others are measured.
 *
 * Another way to approach the problem is to identify unique symbols in
 * the document, then encode one instance of each symbol plus the list of
 * positions at which the symbol appears.  Clearly, this technique depends
 * on the regularity of machine printing - it doesn't do at all well for
 * irregular hand printing.  This technique would be lossless if
 * there were no noise in images and all symbols were ideally rendered and
 * scanned.  Unfortunately, printing is an imperfect process, resulting in
 * a little raggedness around the edges.  Scanning is even less perfect.
 * The general level of brightness can change across a page or across
 * multiple pages.  This causes some instances of a symbol to look big
 * and blobby while others are thin and broken.  If we happen to tolerate
 * this, great.  For safety's sake, we'll assume that symbols whose
 * contours are really different are different symbols.
 *
 * Even among symbols scanned with the same brightness and contrast,
 * there is still noise.  Scanner sensing elements sort of randomly
 * decide whether noisy pixels at the edge of black/white transitions
 * are black or white.  The decision is influenced by the presence of nearby
 * black areas because scanner elements get photons from farther away
 * than we'd like.  The peculiarities of the paper and what's on the back
 * of the page affect edge pixels too.  This results in all scanned
 * symbols being surrounded by a halo of ambiguity.
 *
 * How thick is this halo?  This is a crucial question because it's really
 * the only thing we can really rely on in distinguishing symbols.
 * If we knew the set of symbols before seeing the page, we could find the
 * important areas that distinguish symbols from each other and build a
 * neural net or something.  But, we don't know what symbols will appear.
 * Since we don't know much about the symbol set, we have to make assumptions
 * about the noise.  Well, ok, we can make some basic assumptions about
 * text.  Here's one: well-formed characters are made from strokes, not
 * big blobs (though ill-formed characters have blobs).  That's because
 * text can be written with pencils, pens, brushes and so on without a
 * lot of coloring in big areas.  What does that assumption do for us?
 * A little, not a lot.
 *
 * Back to the thickness of the halo.  We are going to assume it's
 * 1 pixel thick.  This is basically true on images that haven't gone
 * through much copying, faxing, encoding, decoding, etc between the time
 * they're printed and scanned.  Even analog copying isn't too bad since
 * it tends to do the same thing to every character.  This assumption
 * could break down if you print a page, fax it at 100x200 dpi resolution,
 * then scan the result at 600 dpi.  The fax process introduced noise that
 * looks like 6x3 blobs at 600 dpi.  Well, tough.
 * 
 * To preserve losslessness, one would have to encode every variation of
 * every symbol.  This is likely to make for a very large dictionary, roughly
 * one dictionary symbol per character on the page.  This is basically
 * a bad version of MMR compression since we'd end up encoding each symbol
 * in MMR plus encoding the position of each symbol.  If we map all
 * versions of a symbol that differ only in edge noise to the same
 * symbol, we will lose some information, but the image should look fine.
 * It would be annoying to replace every instance of a symbol with
 * the first one found if that symbol was flawed. We will make an
 * attempt to choose a good exemplar for each symbol, but not until
 * a later version of the code.
 *
 * One of our earlier thought-leaders, Z Smith, had an intuitive
 * measure for this amount of degredation.  A compressed and
 * decompressed image should be no different from the original than
 * another scan of the same original would be.
 *
 *			*	*	*
 *
 * There are lots of patents in this area.  It hasn't been used
 * much because the traditional solutions to the problem of merging noisy
 * symbols occasionally merge disjoint symbols.  When this happens,
 * the compression/decompression process changes symbols.  If symbols are
 * letters of the alphabet, this is a very bad thing.  When the symbols
 * are numerals, it's a disaster.
 *
 * Some people use OCR as a symbol classifier.  Since this introduces
 * another level of tricky heuristics into the process, it tends to
 * decrease reliability.  Not only that, it only works for symbols
 * known to the OCR package.  This tends to fail for anything outside
 * the roman character set.  It's considered impolite for a compression
 * routine to quiz the user about the image - compression routines are
 * just supposed to compress.  This increases the chances that an OCR-based
 * package will make a mistake.  Alternatively, an OCR-based compressor
 * can just give up whenever it's not really sure.  This degrades compression.
 * We hope to get better reliability than OCR-based schemes without
 * losing much in compression efficiency.
 *
 * A company called Cartesian Products in Boston has a symbol-based
 * compression scheme.  It works pretty well, at least on clean images.
 * However, we're having trouble establishing a business relationship, their
 * code runs only on a NeXt platform (they want us to port it) and they
 * haven't thought much about gray/color images (?).  Dan Huttenlocher at PARC
 * (and Cornell) started working on a system where the symbols are entire
 * words.  That system was relatively tolerant of noise, but produced pretty
 * large dictionaries of known symbols, so had to amortize the dictionaries
 * over multiple document pages.  Since the pages have different skew, the
 * symbols on different pages are slightly different, which complicates
 * things.  On the other hand, word shapes are really pretty different, so the
 * probability of switching symbols is low.
 *
 * Z Smith got Dan Huttenlocher to try his system on individual characters.
 * It worked pretty well.  Dan had a couple of summer students work on
 * it and it got a lot better.
 * Leigh Klotz has prototyped the system in Docuscript, where it also gets
 * good compression results.
 * The simple version of the algorithm loops through the image, performing
 * five basic steps for each new blob it finds:
 *	1. Find the next connected component.
 *	2. Extract exactly the pixels of that component from the image.
 *	   Leigh thinks it would be ok to get some dreck from the kerned
 *	   characters on either side because it would just make the
 *	   number of symbols in the dictionary bigger.  Dan's system
 *	   improved somewhat when he started using code that does just
 *	   get the cc's pixels.
 *	3. Decide on some subset of the existing dictionary symbols
 *	   to which we will compare the candidate symbol.
 *	4. Compare the candidate with the selected dictionary symbols.  Stop
 *	   when one if them matches.  Record the position of the
 *	   new mark and the identity of the matching dictionary character.
 *	   If none of the characters matches, enter the candidate into the
 *	   dictionary.
 *	   The matching process tolerates edge noise.  It dilates the
 *	   candidate character by 1 and ANDs the result with the dictionary
 *	   character.  The percentage of matching pixels with respect to the
 *	   number of pixels in the dictionary charactyer is found.  Then
 *	   the comparison is done again with the dictionary character being
 *	   dilated and the percentage taken with respect to the undilated
 *	   candidate.
 *	   If both percentages are greater than 96%, we say the character
 *	   matches.  The Cartesian Products patents have a list of tests
 *	   for matching.  At least one of them sounds like this test, but
 *	   Dan H. says they don't do any dilation.
 *	5. It might be necessary to tag the image blob as having already
 *	   been found, maybe by erasing it.  This might be done at the
 *	   time it's extracted from the bitmap.  Whether this is necessary
 *	   depends on how one finds connected image blobs.
 *
 * The newer, more sophisticated, heuristic changes the order of
 * classification.  As symbols are found, their real bitmaps and dilated
 * bitmaps are classified by width and height only.  This results in a
 * bunch of bins, each with a bunch of symbols.  At the end of the first
 * pass, two bitmaps for every symbol instance are stored.  In the second
 * pass, symbols within a bin are compared against each other, generating
 * symbol classes.  The third pass compares symbol classes in neighboring
 * bins, that is, bins whose symbols' widths and heights differ by 1.
 * In the fourth pass, an exemplar is derived for each symbol class.  This
 * is done by finding the average of all the symbols in each class.  The
 * fifth pass (getting optional now), goes through each class, discarding
 * those symbols which do not match the exemplar.  The last pass or two
 * regenerate symbol classes to include the rejects.  Whew.  They tried
 * many variations on this scheme, this was the fastest one.
 *
 * While this scheme seems to give compression results that are comparable
 * to Cartesian's, it does have some annoying features.  Leigh's Docuscript
 * implementation takes about 90 seconds on a page of double spaced, fixed
 * pitch English text and about 2.5 minutes on a dense page of Japanese.
 * On a SPARCStation2.  Ouch.  Dan Huttenlocher's takes about 10-12 seconds
 * on a complex page depending on how many optional steps are performed.
 * Note that it also stores all the bitmaps of all the symbols through the
 * fifth pass.  Say there are 5000 symbols on a complex page.  Say the average
 * symbol has height 30 and width 24.  That's 5000 symbols *
 * 2 bitmaps/symbol * 30 lines/bitmap * 4 bytes/line (each line padded
 * to word boundaries) = 1.2MB.  For a PC, that's still a lot of
 * intermediate memory to consume during compression.  The original image
 * was only 1 MB.
 *
 * I used to think that extracting the symbol from the bitmap would
 * be a real pain.  Well, it's not wonderful, but we do have routines
 * the do this fairly quickly (1 - 2 seconds/page).  However these routines
 * return the symbol in the form of a list of black runs.  Bitmap-based
 * schemes spend time converting these back to bitmaps.
 *
 * Comparing bitmaps is pretty fast.  However, sometimes it's nice to be
 * able to adjust the tolerances in ways that are hard to express in a
 * single bitmap (there must be some pixels in this range, this entire
 * range must be filled with pixels, etc).
 *
 * None of these are absolute showstoppers, but they are annoying.
 *
 *				*	*	*
 *
 * So, my idle daydream was to have some sort of magical state machine into
 * which one fed the relative position and length of each run of black pixels
 * in a symbol.  The classification would pop out.  Cool.  No
 * extracting the character, the third and fourth steps listed above
 * are combined automatically, it works on compressed images since it
 * takes run lengths and it should be pretty fast.
 *
 * I worked on the most straightforward approach to this idea for the
 * better part of a year, on and off.  The basic scheme is to compare
 * individual runs as you find them against templates in one or more
 * discrimination trees.  If we arrive at a tree position where there
 * are no matches, we make up a new tree node.
 *
 * The fundamental problem comes in how the comparisons are done.  Since
 * we are looking at runs of an image, we naturally want to compare the
 * coordinates of the runs we find with the coordinates of the stored runs.
 * Relative to what?  What is the origin relative to which both are
 * measured?
 *
 * This is such a simple, trivial, question that I had no trouble answering
 * it and moving on to more difficult problems.  That happened several
 * times.  I had trouble with relativity in school too.
 *
 * You can't use the absolute origin of the image - that makes every symbol
 * look different than every other symbol since obviously their runs are
 * in different places on the page.  We clearly want to use an
 * origin that's relative to the character.  The approach I tried for the
 * better part of the year attempted to derive the origin as I went along
 * rather than read the entire character and compute the origin relative
 * to the finished character.  The first method requires only one pass, the
 * second method requires two.  I will explain why I gave up on the
 * first method.
 *
 * Before we get started, what if you dispense with a fixed origin entirely?
 * Say you classify each run relative to the one immediately above it?  Cool,
 * except that we must allow for the presence of noise along the edges of
 * characters.  Say we allow 1 pixel of difference between around the
 * edges of characters.  That means we must allow a run to vary left or right
 * by 1 pixel relative to the ideal position.  If we are classifying each
 * run relative to the one above it, we might get a long sequence of
 * "left by 1 pixel" classifications - which would produce a 45 degree
 * line instead of a vertical line.  Since both the 45 degree line and
 * the vertical line fall within the noise margin at each step, we wouldn't
 * be able to tell them apart.  In general, if a relative origin is
 * n lines in back of the current line, we can't distinguish
 * lines with slope 1/n from vertical lines.  This is no good.
 *
 * So, back to fixed origins.  Say we choose one based on the first line or
 * two of a symbol.  Well, we really don't want to use the top line of
 * a symbol, that's just noise.  So, I used the second and third lines.
 * Fine.  I eventually got good discrimination on characters that had
 * one feature that was clearly taller than the rest.  Where I got into
 * trouble was with characters that had two or more independent features
 * that began on (roughly) the same scan line, like "x", "v" and "w".
 * Even serifs can be a pain.
 *
 * The problem is that you don't know that two things that you've encountered
 * in a bitmap are actually all part of a single symbol until you encounter
 * the scan line that joins them.  Therefore, as you encounter the top
 * of each new thing, you give them all individual origins and start them
 * down discrimination trees independently.  When they join, you have to
 * do something about the multiple origins and discrimination tree positions:
 *
 *   1.	We could keep all the origins around as possibilities.  Yuck.
 *	That means you have to try classifying each subsequent run relative
 *	to every origin.  And that's not just until you find some way to
 *	classify the first run either, you have to keep it up all the way
 *	down through the symbol until all but one of the potential origins
 *	drops out (you hope you don't end up with more than one
 *	matching symbol when you're done).  This is complex and slow.
 *	The point of this state machine stuff is to avoid "complex and slow."
 *
 *   2. We could choose one of the origins and discrimination tree
 *	positions.  Based on what?  Well, with several minutes (or months)
 *	of thought, you can convince yourself that any scheme can be
 *	defeated by noise.  For example, say you always choose the origin
 *	from the leftmost incoming stroke.  How tall does something have
 *	to be to qualify as a stroke?  One line is probably noise, so
 *	say two lines tall.  Noise can make a perfectly legal, two line
 *	stroke that was supplying the origin for the joined symbol turn
 *	into a one-line thing that's disregarded as noise.  This would
 *	result in a different origin for the merged symbol, so all of the
 *	subsequent runs would appear to be in a different when only
 *	the origin changed.
 *
 *	Well, say you base your choice of origin on some other physical
 *	property of the candidiate symbols - say height.  For a long time,
 *	I kept the origin belonging to the tallest feature.
 *	That works great on symbols with one tall stroke, like a's, b's, etc.
 *	It is defeated by places where all the incoming strokes are the
 *	same height (m, n, u, v, w, x, etc) so noise causes you
 *	to radically vary the origin and discrimination tree position for
 *	the rest of the symbol.
 *
 *	Any criteria based on position has some sort of boundary.
 *	Noise can cause a potential stroke to cross the boundary.
 *	If it couldn't, you'd be hard pressed to prove that your
 *	system could differentiate symbols that differ in a 
 *	couple of real pixels but is not affected by a couple of
 *	noise pixels.
 *
 *   3.	We could forget about choosing one of the original origins
 *	and combine all of them in some symmetric way.  We no longer
 *	have to remember which of the original origins to use because
 *	we're combining them.  The same goes for the discrimination
 *	tree positions - say all strokes entering a "join" in the 
 *	tree branch to the same next node.
 *	This solution suffers from the same problem as does the solution of 
 *	always choosing the left stroke's origin - how do you know 
 *	if a join happened?  How do you know which of the strokes 
 *	entering the join were tall enough to have an origin that 
 *	should be factored into the calculation?  If noise makes one 
 *	of two incoming strokes too small, we'll decide there was 
 *	no join, hence there was no reason to do the combination of
 *	origins, hence the origin does not change.  This will cause
 *	us to mis-classify the subsequent runs.  We could try to get
 *	around this by trying all the possibilities of modifying the 
 *	origin (by how much, which of the joined strokes was tall 
 *	enough to qualify for an origin?), but we fall back into
 *	the ooze of solution number 1 above.
 *
 *   4. We could base the origin on the bounding box of the portion
 *	of symbol seen so far, changing the origin as the bounding
 *	box changes.  Say we use its midpoint.  That cleverly avoids 
 *	having to know what's going on in the interior of a symbol, 
 *	you just base it on the width of the entire symbol.  Yeah, 
 *	but what's part of the symbol?  If a join in an instance of 
 *	a symbol is delayed by one line relative to the exemplar 
 *	that's in the discrimination tree, the origin used for the 
 *	current run of the symbol may be quite different than the 
 *	origin used to classify the run in the tree, hence we blow it.
 *	Furthermore, it's annoying that a change on one side of a symbol 
 *	can drastically affect the way we classify runs on the other 
 *	side by changing the origin for the symbol as a whole.  AND, 
 *	since the bounding box of a symbol can change as we go across 
 *	a run, should we process each line once to find the bounding 
 *	box and again to classify each run?
 *
 *   5.	There are lots of other crackpot schemes.  We could try to limit
 *	the rate at which the origin changes so as to lessen the affect
 *	of joins happening a line early or a line late.  That means we
 *	have to keep one origin per incoming stroke until they merge.
 *	Back to yucky solution 1 again.
 *
 * So, where are we?
 *   *	You have to have an origin in order to compare run coordinates.
 *   *	You can't base the origin on some point that's relative to the
 *	current run (like the midpoint of the previous line) and still
 *	tolerate noise.
 *   *	If you start a new origin at the top of each new thing, we have
 *	no mechanism for reliably deciding that a join has occured in
 *	all instances of a symbol that differ only in noise, hence no
 *	mechanism for deciding when to change the origin (or what to
 *	change it to).
 *   *	The process of classifying a symbol relative to all its possible
 *	origins is too painful.
 *   *	We can't base the origin on the portion of the entire symbol
 *	we've seen so far.  
 *
 * This seems to leave us with only one choice - reading the symbol once
 * to compute an origin, then going over it again to classify it.  This
 * is not an real proof.  Some smart person may find a one-pass
 * classification method that beats this one, but I have to make progress.
 *
 * The notion of postponing symbol classification until after the symbol
 * has been entirely seen has one nice point, we don't have to keep information
 * on many partially classified symbols.  When a symbol is completely
 * seen, we classify it all at once.  This should save on cache usage
 * if nothing else.
 *
 * Ok, so how do we do the classification?  Another nice feature of
 * finding all the runs of a symbol first is that we're a lot more
 * relaxed about forks and joins and general topology of symbols.  In fact,
 * we could just compare all the bitmaps we find with all the known
 * bitmaps like the scheme described above.  Well, that's boring.
 * The boring part isn't doing comparisons per se, it's having to do so
 * many of them (and keep so many bitmaps around).  We will use a
 * state machine to classify each symbol, but we'll know the origin,
 * and all the runs, ahead of time.
 *
 * Speaking of knowing all the runs, we don't want to just point at the
 * symbol`s bounding box in the image and say "classify that".  Doing
 * so brings back the problem of discerning the symbol we want from
 * whatever portions of other symbols may fit in the bounding box
 * (in some languages, entire other symbols fit inside a bounding box).
 * i_findConnectedComponentRuns finds symbol bounding boxes and the
 * list of the runs that make up the symbol.  XIS needs this anyway for OCR.
 * A client-supplied callback proc (we're the client in this case)
 * is called for each symbol with the bounding box and the list of runs.
 * We'll classify each symbol inside the callback proc.
 *
 *			*	*	*
 *
 * How do we classify the runs?  It turns out we'll base symbol identity
 * on several properties of the runs, but let's start out with the
 * most generally useful classifier - the arrangement of endpoints.
 *
 * First, let's define terms.  A pixel is the atomic unit of the image.
 * In our case, each pixel is represented by a single bit.  0 means
 * white and 1 means black.
 *
 * A "run" of pixels is a contiguous set of black pixels that's one scan
 * line thick.  We only use horizontal and vertical runs.  A horizontal
 * run is one scan line tall.  A vertical run is one scan line wide.
 * Runs are represented by the coordinates of their endpoints.
 *
 * A "stroke" is a single, vertically adjacent stack of horizontal runs.
 * For example, a "T" is made up of a single stroke - it's wide at the
 * top and thin the rest of the way.  An "O" has 4 strokes, one at the
 * top that forks into two others, which rejoin into a stroke at the bottom.
 * There are vertical strokes also.  They are made up of stacks of
 * vertical runs.
 *
 * A "symbol" is another word for a connected component (as are token, figure,
 * glyph, blob, character).  The phrase "connected component" is too long,
 * so we picked one of the alternatives.  A symbol is made up of a
 * collection of strokes where every stroke is directly or indirectly
 * connected to every other stroke.
 *
 * In the original version of this code, we tracked strokes.  That was done
 * because we didn't know which strokes belonged to a symbol until
 * they joined.  Tracking the strokes required us to tolerate not only 
 * small variances in stroke width and offset, but also to tolerate
 * variances in the positions of forks and joins.  For example, two
 * strokes that joined on line n of the exemplar of a symbol might
 * join on line n-1, line n or line n+1 of a symbol instance.
 * Furthermore, stroke widths can vary wildly when they are next to
 * horizontal edges.  That's because noise on horizontal edges can generate
 * long, short or broken horizontal runs.
 *
 * Tolerating all that noise while guaranteeing that symbols that are really
 * different are classified as being different was quite complex.  Since
 * we start out with all the strokes of a symbol, we can simplify things
 * quite a bit here.
 *
 * A word about the runs we get back from i_findConnectedComponentRuns.
 * They are on a single linked list.  The first element of the list
 * is the rightmost run on the lowest line of the symbol.  Subsequent
 * runs proceed from right to left in each line and from the bottom of
 * the symbol towards the top.  Each run element contain only the coordinates
 * of the left and right endpoints and a pointer to the next run.  There's
 * also a pointer that was used by i_findConnectedComponentRuns, we'll
 * reuse the space to hold a tag that flips each time we cross a scan
 * line boundary.  We originally find the end of a scan line by comparing the
 * right and left endpoints of subsequent runs.  If nextRun.right >=
 * prevRun.left - 1, prevRun is on the left boundary of one scan line and
 * nextRun is on the right boundary of the scan line above it.  Once
 * we've used this method to tag all the runs, we'll use the line tags.
 *
 * Dan Huttenlocher's code dilates the exemplar and tries to cover the
 * candidate, then dilates the candidate and tries to cover the
 * exemplar.  If the dilated exemplar covers x% of the candidate's pixels
 * (x is a function of the number of edge pixels in the symbol)
 * and the dilated candidate cover's x% of the exemplar's pixels,
 * the candidate is said to match the exemplar.  This is fine, but
 * it's annoying to dilate every candidate.  Can't we do something with
 * the exemplar to avoid messing with the candidate?  If we can't avoid
 * doing anything to the candidate, can we do something fairly simple
 * and cheap?
 * 
 * In this code, we'll change both the criteria for matching and the
 * heuristics used to do the match.  I'll try to explain them separately.
 *
 * I went through lots of matching heuristics.  An obvious one both
 * eroded and dilated the exemplar - testing to see if the instance
 * covered the eroded exemplar and was covered by the dilated exemplar.
 * This worked fine for large symbols and not at all for small ones.
 * Thin strokes eroded to nothing, so virtually any small symbol matched
 * something of about the same size that was composed of thin strokes.
 *
 * A method that worked better, but not well enough, was to compute endpoint
 * tolerances for each horizontal instance run by looking at the
 * corresponding clump of three vertically adjacent horizontal runs in
 * the exemplar.  These are the line, the line above and the line below.
 * For example, if we want to know the legal values for the endpoints of
 * runs in scan line x, stroke s of a symbol, we look at the runs in
 * scan lines x-1, x and x+1 of stroke s of the exemplar symbol.
 * This gives us some ability to tolerate vertical noise as well as
 * horizontal noise.  This method did a great job on endpoints, but
 * didn't do so well in the middles of runs.  For example, say a symbol
 * has a very small vertical gap between two horizontal strokes.  In order
 * to tolerate small amounts of vertical noise in horizontal runs, the
 * tolerances in the placement of horizontal endpoints around these little
 * gaps is pretty vague.  It generally says, "there may or may not be
 * a gap of some smallish size here."  This wasn't sufficient for
 * distinguishing small characters.
 *
 * The nicest feature of this scheme was that we did virtually nothing
 * to each new symbol instance found in the image.  All the work was
 * done to the exemplars.  Well, too bad.  The only way I could get good
 * compression and discrimination was to do more work to the instances.
 * In Dan Huttenlocher's method, this extra work involves creating
 * (and storing) two bitmaps for each instance.  We'd like to
 * do less than that.
 *
 * The method of checking horizontal runs did a good job of distinguishing
 * symbols whose differences were on the left and right edges.
 * This makes sense since that's where the checking was happening.  In
 * order to check the top and bottom edges of the symbol, we also compose
 * a set of vertical runs for the symbol.  This doesn't take *too* long.
 * We then run the checking routine twice, once on horizontal runs and
 * once on vertical runs.  There is only one checking procedure, so we
 * only have to explain it once.
 *
 * Ok, now I have to try to explain the run checking in depth.  It would
 * be straightforward if it weren't for small characters with fine features.
 *
 * As stated above, the endpoint tolerances for an instance run are
 * found by looking for a) the same stroke in the exemplar, and b) the runs
 * above, at the same scan line and below the instance run in that stroke.
 * For example, say we have a simple symbol composed of a single vertical
 * stroke.  In this class are l, I, T and L.  The last two look like two
 * pen strokes, but in terms of runs, there is no scan line with more than
 * one run in it.  In this example, if we wanted to calculate the legal
 * range for the run in line 5, we'd look at the runs in lines 4, 5 and 6.
 * Call the set of runs in these three lines a "group."
 *
 * How do we handle symbols with multiple strokes?  We generate a group
 * for each scan line of each stroke.  For example, let's look at
 * an "a."
 *
 *  0                    aaaaaaaaaaaaaaa
 *  1                aaaaaaaaaaaaaaaaaaaaaa
 *  2            aaaaaaaaaaaaaaaaaaaaaaaaaaaaa
 *  3          aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
 *  4          aaaaaaa                aaaaaaaaa
 *  5                                 aaaaaaaaa
 *  6                                 aaaaaaaaa
 *  7             aaaaaaaaaaaaaaaaaaaaaaaaaaaaa
 *  8           aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
 *  9         aaaaaaaaaaa             aaaaaaaaa
 * 10        aaaaaaaaaaa              aaaaaaaaa
 * 11       aaaaaaaaaaa               aaaaaaaaa
 * 12        aaaaaaaaaaa              aaaaaaaaa
 * 13         aaaaaaaaaaaaaaa       aaaaaaaaaaa
 * 14            aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
 * 15              aaaaaaaaaaaaaaaaaaaaaaaaaaaa
 * 16               aaaaaaaaaa        aaaaaaaaa
 *
 *
 * Lines 0 through 3 have a single stroke.  Line 4 has two strokes and
 * so on.  Therefore, we have one group in each of lines 0, 1, 2 and 3.
 * There is actually still only one group in line 4 since all of the
 * strokes in line 4 are connected to one run (in line 3).  There is
 * one group in lines 5, 6, 7, 8 and 9 (run in line 8 connects the two
 * strokes in line 9); two groups in lines 10, 11, 12; one group in
 * lines 13, 14, 15 and 16.
 *
 * In the simplest world, each group would have a range for left endpoints,
 * a range for right endpoints and that's it.  The world is not so simple.
 * We actually have four types of groups.  Each group type comes with a
 * set of endpoint ranges and a method of interpreting them.
 *
 * Group type: CheckNoise
 *	At the top and bottom of strokes, we allow noise.  These
 *	groups are used to make sure that if noise does exist there,
 *	it is confined to a legal range.  For example, the run in line
 *	0 may take on many widths in different symbol instances.  In
 *	some, it may not be there at all.  The CheckNoise group simply
 *	says that we should make sure that any pixels in line 0 are
 *	within (roughly) the boundaries of the run in line 1.  See
 *	the discussion of endNoise below.
 *
 * Group type: Check1Run
 *	This is the most common group type.  It applies in the
 *	interior of single strokes.  For example, on line 1 of
 *	the symbol, the run in question has exactly one run above
 *	and one run below it.  In this simple case, we can just set
 *	endpoint ranges calculated from the three runs and demand
 *	that the instance runs have endpoints in those ranges.
 *	In the case of the run on line 2, we (roughly) say that the
 *	left end of the instance run on line 2 must fall between the
 *	range described by MAX(left end of run in line 1, left end
 *	of run in line 2, left end of run in line 3) and MIN(left end
 *	of run in line 1, left end of run in line 2, left end of run
 *	in line 3).  The range for the right end is calculated using
 *	the right ends of the three runs.
 *	This is the main idea, the details are explained below.
 *
 * Group type: CheckEndpoints
 *	This type of group is used to match areas around the tops and
 *	bottoms of strokes, like lines 3, 4, and 5 above.  We must
 *	be able to tolerate a small gap in the run of line 3 (the fork
 *	moves up a line) or a connection between the runs of line 4
 *	(the fork moves down a line or vanishes in this case).  We might
 *	also get extra runs between the two runs on line 4.  To do all
 *	this, we compute a range of endpoints for the left end of all
 *	the runs seen by the group, a range for the right endpoints as
 *	is done for Check1Run.  However, during comparison, we only
 *	check that the right endpoint of the rightmost run on line 4
 *	and the left endpoint of the leftmost run on line 4 are legal.
 *	Any other endpoints or runs between these are ignored.  This
 *	is required to tolerate the kind of noise that commonly appears 
 *	in images, but is awfully vague in the middle.  That's why we 
 *	match both horizontal and vertical runs.
 *
 * Group type: Check1RunNoAdvance
 *	This is seldom used.  When matching runs in the groups above, we
 *	discard each instance run after ensuring that its endpoints
 *	are legal.  Sometimes, two strokes are connected by only a
 *	single run.  This happens on the top or bottom of a small
 *	symbol composed of thin strokes.  To preserve topology, we
 *	want to guarantee that there really are black pixels in the
 *	space between the two strokes.  A CheckEndpoints group would not
 *	do this.  Instead, we use a CheckNoise group to consume any
 *	runs contained entirely under the stroke to the right, a
 *	Check1RunNoAdvance group to make sure there are black bits in
 *	the gap and a CheckNoise group to consume runs to the left of
 *	the left endpoint.  This allows any combination of runs (including
 *	a single run!) on the top or bottom of the strokes that bridges
 *	the gap.
 *
 * Details:
 * There are a few.  The most important one is how much slop we
 * allow.  For example, in the middle of a perfectly vertical line with
 * perfectly vertical edges, we generate Check1Run groups.  Since all
 * the endpoints on the right (and left) are equal to each other, do we
 * say that any matching symbol must have exactly that endpoint?
 * That wouldn't tolerate noise, so we don't do it.  To discuss the
 * tolerances, we have to say something about the causes of endpoint
 * variances.
 *
 * The most obvious cause of endpoint variances is that the instance and the
 * exemplar really don't match.  It's important to keep this in mind as we
 * are tempted to make the tolerances bigger to tolerate more noise.
 *
 * Another cause of variance is plain old noise.  How much noise should we
 * tolerate?  Always tolerating 1 pixel is a nice bold policy, but that's
 * too much for really tiny symbols and not enough for big ones.  This
 * brings up an interesting point - the amount of care we use in matching
 * symbols has something to do with their size.  Big symbols are (generally)
 * made up of big strokes that are fairly far apart.  Therefore, we can
 * tolerate a lot of edge noise and still match well.  Tiny symbols are
 * all edge noise so it's very easy to confuse one with another.  Luckily,
 * tiny symbols don't take much space to encode, so it's not so bad to
 * have several copies of each one in the output file.  This leads us to
 * the general policy that the amount of noise we tolerate is a function
 * of symbol size.
 *
 * What function?  One could use width, height, area or mass (number of
 * black pixels).  Dan Huttenlocher's group uses number of edge pixels.
 * For the sake of simplicity, we use width when generating horizontal
 * run groups and height when generating vertical run groups.  A symbol
 * composed of a large collection of very small strokes would cause
 * a problem here.  But fairly few symbols are assembled this way, so
 * we're ok for at least awhile.  We may need to revisit this issue
 * for Kanji text.
 *
 * Another source of variance is placement of the origin.  We compute the
 * origin as the middle of the symbol's bounding box.  I tried fancier
 * schemes, but this gave results that were as accurate as any in terms
 * of number of symbol classes found and is very fast to compute.  With
 * this scheme, however, noise along the edge of a symbol can move the
 * origin.  We'd like to minimize this effect.  We do this in two ways.
 *
 * The first measure taken to minimize origin shift is to put an extra
 * bit of significance in the coordinates that hold endpoint positions
 * and tolerances.  The endpoint coordinates within a bounding box of
 * width w range from -(w-1) to (w-1).  Adding a new noise column only
 * changes the coordinates of the pixels on the opposite side of a
 * symbol by 1/2 of a pixel (or 1 in our 2x coordinate system).  This
 * 2x coordinate system pervades all of the code, so watch out for it.
 *
 * Say we want to tolerate one pixel of noise.  The tolerance is set to
 * three.  This is the sum of 2 for the noise plus 1 for a possible
 * origin shift.  While this doesn't do a perfect job of tolerating
 * noise, it's certainly an improvement over the normal scheme that shifts
 * the coordinates of a pixel by either 0 pixels or a whole pixel when
 * noise shows up.  This makes an origin shift take up all the tolerance
 * for noise, so origin shift plus any noise looks like a completely
 * different symbol.
 *
 * That's very interesting, but how much noise do we tolerate?  In the
 * current scheme, there are actually three classes of noise, smallNoise
 * bigNoise and endNoise.  The first, smallNoise is generally the
 * amount of slop we add to an endpoint tolerance when computing
 * the tolerance relative to the line above or the line below the
 * current one.  This is done under the general theory that the farther
 * away something is, the smaller its effect should be.  The second class,
 * bigNoise, is when we add to the endpoint tolerances when computing
 * tolerances relative to the current line.  The third class,
 * endNoise, is added to tolerances of endpoints in CheckNoise groups.
 * These are on the tops and bottoms of strokes where a line of
 * pixels may or may not appear at all in an instance.
 *
 * Each of these is computed using a 2-component formula.  The
 * first component is MIN(cMaxSmallNoise, width/cSmallNoiseDivisor) for
 * smallNoise and using the appropriate constants for the other two
 * noise classes.  This says that for very small characters, the amount
 * of noise tolerated goes down as width/constant.  This is meant to
 * distinguish very thin or very flat symbols, like I's, 1's, and l's.
 * Above a fairly small range, the noise tolerated becomes cMaxSmallNoise
 * (or the appropriate constant for the other noise classes).  These
 * constants may change as we find more images.  Today, cMaxSmallNoise
 * is 1, cMaxBigNoise is 3 and cMaxEndNoise is 3.
 *
 * The second component of noise allows us to tolerate larger noise on
 * much larger characters.  It adds (width/cBigSymbolDivisor) to the
 * noise of all three.  Currently, cBigSymbolDivisor is set to 40.
 * Remembering that all coordinates are kept with an extra bit of significance,
 * This means that every increment of 20 pixels in width adds 1/2
 * of a pixel of noise tolerance.
 *
 * Here's another detail.  You can think of the Check1Run case as a
 * special case of CheckEndpoints.  It's common and easy to do the
 * check, so we make a separate case for it.  Most of the time, it
 * passes successfully too.  However, we found cases where symbols
 * were tossed out of the correct class because they had an extra
 * nubbin of noise descending from a run.  The exemplar had no such
 * nubbin, so used the Check1Run group.  Since there were two runs,
 * the test failed even though both runs were legal.
 *
 * Our solution to this was to let the Check1Run test fall through to
 * the CheckEndpoints test when it fails.  This certainly cures the
 * problem mentioned above, but has a side effect.  Symbols with little
 * (less than 3 pixel high) holes become equivalent to symbols without
 * those holes.  That's because a solid exemplar issues Check1Run's
 * for all the scan lines and these match the dual runs in an instance
 * with the holes.
 *
 * Well, is this so bad?  I claim it's ok.  People don't really see
 * tiny holes, they get their info from the overall shape of the
 * character, i.e. the outside.  So long as we're careful about that,
 * we're ok.  In fact, it cleans up documents that have a few versions
 * a character in which the hole is closed.
 *
 * I said there are two ways we work on minimizing the effect of noise
 * on symbols.  The first one was leaving a extra bit in the coordinates.
 * The second is to try to align character templates better.  Originally,
 * we just aligned the bottom scan lines of an instance and exemplar when
 * comparing horizontal runs (left sides when comparing vertical runs).
 * In the new world, we compute the center horizontal and vertical axes
 * by averaging the coordinates of the vertical and horizontal envelopes.
 * Once we have an axis, we compute the distance to the bottom (or left)
 * scan line.  Both distances are recorded with each character.  We
 * insert an extra bit of significance in this number to minimize the
 * effect of rounding.  When we get ready to compare an instance with
 * an exemplar, we compare the distance from the central axis to the 
 * edge of one to the corresponding distance in the other.
 * If the difference between the distances is too big, we add
 * a scan line (duplicate of the bottom line for horiz runs and
 * left col for vertical runs) to the symbol with the smaller distance.
 * This tends to bring the characters into alignment. Doing this had
 * a nice effect on the number of equivalence classes.
 *
 * Here's another relatively new feature.  Tolerating +- 3 units of noise on
 * all pixels is sometimes too lenient.  All small, round, dense things
 * look pretty much the same.  When confronted with them, humans start placing
 * greater weight on smaller differences.  What was a noise tick on a nice
 * well-defined character becomes a significant feature on a blob.  This
 * particularly came up in "e" vs "a".  In small, bold characters, both
 * turn into little round things.
 *
 * The solution is to keep track not only of whether each instance endpoint
 * is within the corresponding group limits, but also whether the instance
 * endpoints are banging back and forth between the exemplar limits.  This
 * test is too fussy to do all the time - we only invoke it in small, round,
 * dense symbols.  How do we distinguish real (but tiny) differences in
 * shape from noise?  For example, if an exemplar edge and an instance
 * edge have uncorrelated noise, it will look like the instance is wandering
 * all over within the exemplar's tolerances.
 *
 * We also can't just add up the endpoint differences.  If the instance is
 * shifted by a pixel relative to the exemplar, we'd get a large total
 * endpoint difference for no real shape difference.
 *
 * Well, we do pseudo calculus.  Remember that the integral of d(f(x))/dx
 * is f(x) + C?  Well, in our world, the C represents the shift constant -
 * we set it to 0 and ignore it.  But how do we get d(f(x))/dx?
 * 
 * Set f(x) equal to the distance between the instance and exemplar endpoints
 * on each scan line x.  The first derivative is the difference between
 * distances on adjoining scan lines.  d(f(x))/dx has the nice property that
 * it only records *changes* in the distances, so we can ignore constant
 * offsets.
 *
 * However, each noise blob on either the instance or the exemplar causes
 * d(f(x))/dx to change twice, once for the beginning of the noise blob
 * and once for the end.  That's annoying since we'd like to ignore
 * noise blobs altogether.  So, we add each non-zero term in d(f(x))/dx
 * to the previous non-zero d(f(x))/dx.  This is a limited integration
 * step, which (sort of) produces f(x) again with the constant offset gone.
 * The absolute value of these sums is added to the total difference.
 * We calculate a total sum of differences for the left side of the symbol
 * and another one for the right side.  If the horizontal runs all match,
 * we look at these total sums.  If both of them are below 3/4 of the
 * character height, we say the runs really matched.  The 3/4 is a magic
 * number that could be adjusted as time goes on.  It seems to give pretty
 * good results now.
 *
 * The same thing is done when comparing vertical runs, except we compare
 * the total sum of differences with the 3/4 of the character width.
 * 
 *				*	*	*
 *
 * So much for checking runs.  Unfortunately, there are tiny symbols that
 * have wonderfully matching endpoints, but are really different symbols.
 * For these, we have four sets of overall measures.  These are also
 * computed for every instance and checked against each exemplar
 * before we do the run matching.  That's because we should do the
 * cheap checks before doing the expensive one.
 *
 * Italics: It's really hard to distinguish small, round italic characters
 * from their non-italic counterparts.  I tried all sorts of measures
 * on real characters, most were startlingly useless.  Many others did
 * a great job of distinguishing italics, but also broke up lots of
 * valid symbol classes.  The measure that seems to distinguish
 * italics without affecting the rest of the symbols is called "slant."
 * It's computed from the number and size of jaggies found in the symbol.
 * You add 1 to the score for every jaggie that goes 1 pixel to the right
 * (as you move from bottom to top in the symbol) and add 2 for every
 * jaggy that moves 2 or more pixels to the right.  Similarly, you
 * subtract 1 pixel for every jaggy that goes 1 to the left and 2 for
 * every jaggie that goes 2 or more to the left.  Italic characters lean
 * to the right, so they tend to get more positive scores than their
 * counterparts.  To increase the effect, we also calculate the slant
 * using the vertical runs and add it in.  This does seem to distinguish
 * small italic "o"s from small, non-italic "o"s.
 *
 * The next problem is distinguishing all those tall thin things, like
 * 1's, l's and I's.  A vertical stroke that's 3 pixels wide looks
 * quite different from one that's 2 pixels wide, but our endpoint matcher
 * is still happy.  Yes, cranking the noise tolerances down helped,
 * but making them too small caused the number of symbol classes to
 * explode.  Instead, we calculate the mass, or the number of black pixels.
 * Things that are generally fatter have a much higher mass, at least
 * as a percentage of total mass.
 *
 * But that didn't work as well as it could have.  It also has the problem
 * of a *really* funky set of mass limits.  We changed to calculating
 * the stroke width, or something close.  To do this, we make a histogram
 * indexed by run length.  We find out how many short runs account for
 * about 1/2 of the total pixels (counting horizontal and vertical runs),
 * then divide the number of pixels accounted for by the number of
 * runs counted.
 *
 * Before introducing the concept of stroke width, we had a problem with
 * tiny blobs matching everything.  Originally, we used a different
 * measure - interior pixels - to solve this.  However, that measure
 * was replace by stroke width.
 *
 * Finally, we got stuck on the difference between small "8"s
 * and "B"s.  With the right bits of noise, their edges are within
 * one pixel of each other everywhere.  Their masses, stroke widths and
 * slants are close.  The measure that seems to (barely?) distinguish
 * them is called volume.  We compute both a horizontal and a vertical
 * volume.  The first is a count of the number of pixels between
 * the far left and right edges of the symbol.  The vertical volume
 * is the number between the top and bottom of the symbol.  The
 * way we check horizontal or vertical volume is to see if the dimension
 * of the symbol and exemplar match.  If so, we ignore volume.  If the
 * instance is wider (thinner) than the exemplar, we require that the
 * horizontal volume also be greater (less).  Ditto for the height vs
 * vertical volume.  This is supposed to prevent us from matching slightly
 * larger round things to smaller square things.  The square things have
 * smaller dimensions, but larger volumes.
 *
 * Well, that almost worked.  Some 8's had larger total volumes than
 * smaller B's anyway.  So, we changed the volume calculation to include
 * only portions of the envelope that were sufficiently wide or tall.
 * That emphasizes square things over round things even more.  That
 * seemed to do it.  The fraction of total width or height a horizontal
 * or vertical runs has to reach to be counted is another magic number.
 * 
 * Since the volume is now a funny measure, we also decide when it means
 * something.  If the horizontal volume is not a sufficiently large
 * fraction of the symbol's area, we ignore it.  Ditto for the vertical
 * volume.  However, either the exemplar's or instance's volume can
 * trigger a check.  This actually does pretty reasonable job on lots
 * of little blobs the differ mainly in being roundish vs squarish.
 *
 *			*	*	*
 *
 * There's another new criterion.  We consolidate symbols (see below),
 * by matching existing symbols against the other existing symbols.
 * We had the problem that symbol A might match symbol B, and B match C;
 * but A might not match C.  The matching tests were pretty tight, but
 * these things happen.  The way we stopped it was to assign minimum and
 * maximum matching widths and heights to each symbol.  The range was
 * set as a function of the symbols's width and height.  Whenever a symbol
 * is found to be equivalent to another symbol, we adjust the range
 * of both symbols - or all symbols in the new equivalence class - to
 * reflect the existing difference between the largest and smallest symbol
 * in the class.  For example, say the initial range is 4 to 8.  That's
 * 6 +- 2.  Now say we match a symbol with a range of 5 to 9 (or 7 +- 2).
 * The new range for both symbols will be restricted.  It will be 5 to 8.
 * This happens when a symbol matches a symbol, a symbol becomes a member
 * of a class or two classes merge.
 *
 * Similarly, when we get ready to match an existing symbol with the other
 * symbols, we record the largest and smallest dimesions of all symbols
 * in the first symbol's equivalence class.  The matching process guarantees
 * that all symbols in the class fit within the range of any matching
 * symbol (or it class).  This means that we're allowed to merge two
 * classes if every symbol from one matches every symbol in the other.  For
 * the sake of convenience, we keep the range of widths and heights for
 * a class in a dummy symbol associated with the class.
 *
 *			*	*	*
 *
 * How do we actually classify a symbol?  We used to use a lovely
 * classification tree.  At various times, we used lots of them.  This
 * version is simpler.  Each exemplar symbol has a set of
 * measures (width, height, mass, slant, stroke width, volume) and pointers
 * to two sets of groups.  One group is used to match horizontal runs,
 * one to match vertical runs.  When we decide that an exemplar doesn't
 * match our current instance, we just go directly to the next exemplar
 * instead of traversing a tree to find the next potential match.  Maybe
 * using a tree would be better, maybe I'll try it sometime.
 *
 * Here's basically what happens now.  ip_classifySymbol contains steps
 * 1 through 3 and 11 through 13.  ip_compareSymbols holds steps 4 through 10.
 *
 *   1. Find the symbol as a set of horizontal runs.  This is done in
 *	i_findConnectedComponentRuns.  That procedure calls our
 *	callback proc, ip_classifySymbol, with the position of the
 *	symbol on the page and the runs.
 *   2. Calculate the vertical runs, mass, slant, stroke width and volume
 *	measures of the instance from these runs.
 *   3. The existing symbols are arranged in a 2D array by (width, height).
 *	Look through the symbols in this order:
 *		a. width and heigth match the instance.
 *		b. symbol width differs by 1 from instance.
 *		c. height differs by 1.
 *		d. width and height differ by 1 from instance.
 *		e. width differs by 2.
 *		f. height differs by 2.
 *		g. width differs by 2 and height by 1
 *		h. height differs by 2 and width by 1
 *		i. both width and height differ by 2.
 *
 *   4. For each of these symbols, check the width and height.  If the
 *	range of widths and height represented by the instance (normally
 *	one width and one height, but can be more of the instance is
 *	already a member of an equivalence class) does not fall entirely
 *	within the range allowed by the symbol (or it's equivalence class),
 *	go on to the next symbol.
 *   5. Check the slant, horizontal and vertical volumes, mass and stroke width.
 *	If any of these fail, go on to the next symbol.  We really only
 *	check the volumes for small symbols since they only make
 *	spurious rejections for larger symbols.
 *   6.	If the instance is a member of an equivalence class, make sure
 *	the symbol is not also a member of the same class.  If so,
 *	go on to the next symbol.
 *   7.	Check the horizontal runs.  If any of these fail, go on to the
 *	next symbol.
 *   8. If this is a small, round, dense symbol, check the bizarrely
 *	calculated (see above) sum of relative endpoint distances on the
 *	left and on the right.  If either is too large, go on to the
 *	next symbol.
 *   9.	Check the vertical runs.  If any of these fail, go on to the next
 *	symbol.
 *  10. If this is a small, round, dense symbol, check the bizarrely
 *	calculated (see above) sum of relative endpoint distances on the
 *	top and on the bottom.  If either is too large, go on to the
 *	next symbol.
 *
 *  11. If the instance matches an exemplar in all of these ways, record
 *	the instance's position and the exemplar's ID in the position
 *	stream (see below).
 *  12. If the instance doesn't match any of the existing exemplar's,
 *	make a new exemplar from this instance.
 *  13. Free the vertical runs and return to i_findConnectedComponentRuns,
 *	which will free the horizontal runs.
 *
 *			*	*	*
 *
 * What was that talk about symbol consolidation?  Doesn't this wonderfully
 * complex process just do the right classification?  No.  We can still get a
 * mismatch solely because the symbol's origin isn't quite lined up with
 * the exemplar's.  This can happen because of noise on the edge.  Yes,
 * our two tricks of putting a extra bit of significance in the coords
 * and attempting to align the instance and symbol axes helped, but they
 * didn't completely solve the problem.
 *
 * Two things can happen.  First, noise on the bottom moves the entire symbol
 * up by an entire line when comparing horizontal runs.  That is, the
 * first run we compare is the noise run instead of the real bottom run.
 * This means we've used up all the vertical tolerance built into the
 * horizontal groups already.  If there's noise on top of this, we don't
 * get a match.  Similarly, noise on the left causes us to shift all the
 * runs to the right when comparing vertical instance runs with vertical
 * exemplar groups.
 *
 * Couldn't we take more care when doing the comparisons to allow for
 * these adjustments?  We already did; we don't want to spend even more
 * time on each instance symbol.  That's where the bulk of the compression
 * time is already.  Rather than slide each instance symbol around trying to
 * find an initial match with an instance, we do the classification in
 * two steps.  The first step classifies instances into a relatively
 * large set of classes.  The number of classes is still much smaller
 * than the number of instances.  The second step consolidates
 * the classes.  Since there are many fewer classes, we can spend more time
 * on each one when trying for a match.  This multi-step refinement is
 * something I learned from Dan Huttenlocher's method.
 *
 * What happens when we reduce the number of redundant classes?  A minor
 * effect is that more of the symbols on the page have symbol ID's less
 * than 128, so they fit into 1 byte (see below).  This makes the length
 * of the position buffer a little shorter.  A more interesting effect
 * is that we encode fewer symbol shapes into the output file.  For 1 page
 * documents, this is somewhere between 1/3 and 2/3 of the total compressed
 * file.  It doesn't matter how many times a shape is used, getting rid
 * of its description is always a good thing.
 *
 * How would we consolidate classes?  Do we really wish to do so?  Any
 * consolidation process increases the probability of putting distinct
 * symbols into the same class.  When do we want to do it?  If we find
 * 3 versions of a character on one page, we may well find the same three
 * on another - it may be faster to do all the encoding for all the pages
 * first, then consolidate at the end rather than consolidating and
 * re-expanding the shape tree after each page.  For these reasons, it
 * would be nice to let the client decide whether and when to do it.
 * For this reason, we'll have a separate routine.
 *
 * i_consolidateShapeInfo takes a pointer to the shape information.  It
 * returns a pointer to a table that's used to map old symbolID's to
 * new ones.  i_remapSymbolIDs takes this table and a pointer to input
 * and output position buffers.  It maps the symbol ID's in the first
 * buffer to the new set and puts the result in the second buffer.
 *
 * How does i_consolidateShapeInfo work?  We'll go down the list of
 * symbols, inspecting each one.  There an equivalence class pointer
 * associated with each symbol.  It's initially null.  There are also
 * two counts associated with each symbol, the number of instances of
 * that symbol found directly in the image(s) and the number of
 * instances found to be equivalent to this symbol.  The first number
 * is updated as i_convertImageToSymbols executes.  The second is updated
 * in this step.
 *
 * We assume that there are two or three versions of each popular symbol.
 * We are trying to merge these versions.  We could try offsetting the
 * origin of each symbol by 1 in all four (8?) directions to see if it matches
 * any other symbol.  We can do better by noticing that there are
 * three reasons symbols might not match.  The first is commutivity.
 * Just because instance A doesn't match the groups of exemplar B doesn't
 * mean that instance B wouldn't match the groups of exemplar A.
 * Endpoint tolerances are produced by differences in vertically adjacent
 * endpoint, which may well be larger in some instances of a symbol than
 * in others.  To catch this case, we simply reclassify each symbol.
 * This guarantees that we compare each symbol with every exemplar
 * that could possibly match it.  In the first pass through the image,
 * we would not test an early version of a symbol to see if it
 * matched some later version.  We do that now.
 *
 * The second reason for mismatches is that the instance and exemplar
 * horizontal runs may be mis-aligned vertically.  The third reason
 * is that the instance and exemplar vertical runs may be misaligned
 * horizontally.  Even though we try to align the central axes of the
 * symbols, there's no guarantee that we did it correctly.  If there's
 * an extra scan line in one or the other, all of the run compares
 * will be off by one scan line. There's tolerance built into the
 * compares that will handle this, but being misaligned by one scan
 * line plus one noise blob will cause a mismatch.
 *
 * It's possible that we have both horiztonal and vertical
 * misalignments, but the amount of time spent detecting this case hasn't
 * yet paid off in reduced symbol dictionary size.
 *
 * In both of the horizontal and vertical cases, we want to shift
 * the entire group of runs.  Since there is no explicit vertical
 * coordinate for horizontal runs (or horizontal coord for vertical
 * runs), we do this by adding another scan line to the beginning
 * of the symbol.  Shouldn't we also try deleting a scan line?  No,
 * that's dangerous.  There may well be interesting information
 * on the bottom scan line that, when deleted, causes two different
 * symbol classes to match.  What do we do instead?  Say symbols
 * A and B should match, but A has an extra line of noise on the bottom.
 * When we get around to looking at B, we'll add a line of noise to it,
 * causing it to line up with A.
 *
 * This scheme has us adding something to the bottom of a symbol to move
 * it up and to the left side of a symbol to move it over.  Shouldn't
 * we try adding stuff to the left and top as well?  No.  The groups
 * tolerate extra noise on the left and top just fine.  What they don't
 * tolerate nearly as well is having all the instance runs offset from
 * the exemplar groups by a scan line.  The way we align instances with
 * exemplars is by aligning the first instance run with the first
 * exemplar scan line.  Both move right to left, bottom to top.  Therefore,
 * the way to slide them around is to add an extra scan line at the
 * beginning of one of the lists.  We choose to play with the instance
 * since it has less stuff.
 *
 * What routine does the matching?  i_consolidateShapeInfo has the job of
 * going through all the symbols.  For each symbol, it recreates the runs
 * that i_findConnectedComponentRuns would have found for that symbol.
 * The coordinates are funny - being centered around the center of the
 * symbol - but that's ok.  i_consolidateShapeInfo also sets a flag
 * in the text compression info structure saying that we're just
 * consolidating shapes, not finding new ones on the page.  It then
 * calls ip_classifySymbol with the runs just as i_findConnectedComponentRuns
 * did.
 *
 * ip_classifySymbol operates on the new symbol's runs in this case
 * much as it did when the symbol was found the first time.  However, there
 * are differences:
 *   1. The symbol ID is also passed by i_consolidateShapeInfo in the
 *	text compression info structure.  ip_classifyShape uses this
 *	info to avoid comparing the symbol with itself or any other symbol
 *	it its equivalence class.  Doing so would probably cause a match,
 *	which would stop us from looking at the rest of the symbols.
 *   2. Interesting info about the symbol, like mass, stroke width, volumes
 *	and so on is just read from the symbol instead of being recomputed
 *	from the runs.
 *   3. No position information is written out for this symbol, nor is a
 *	new symbol made in case of no match.
 *
 * Now you know basically what we want to do and how we want to do it.
 * i_consolidateShapeInfo goes through all the symbols.  For each one,
 * it generates the original runs for that symbol.  It asks ip_classifySymbol
 * for the name of one other symbol it matches.  If one is found,
 * the two symbols are put into the same equivalence class by having
 * them point to each other.  If exactly one of them is already in an
 * equivalence class, that ring of symbols is expanded to include the
 * other symbol.  If both are equivalence classes, the two
 * rings specifying the equivalence classes are joined.  Since
 * ip_compareSymbols won't compare two symbols in the same class, we know
 * that we'll never try to consolidate two symbols that are already in
 * the same class.
 *
 * When two symbols are found to be equivalent, the number of times each
 * symbol is found is added to the number of times the other symbol is
 * found.  We're going to use number of occurences as the measure of
 * which exemplar to preserve from the equivalence class.
 *
 * Having submitted the unchanged runs to ip_classifySymbol, we try adding
 * new runs to the bottom of the symbol, or at the beginning of the
 * list.  What new runs?  I experimented around for awhile.  I finally
 * settled on adding one new pixel to the middle of the first
 * (rightmost) run.  This is the easiest piece of noise to match since
 * it's the smallest.  Again, we ask ip_classifySymbol about this new
 * symbol and adjust equivalence classes accordingly.
 *
 * Finally, we add noise to the left of the symbol.  Ideally, we should
 * also add one pixel to the middle of the leftmost vertical runs.
 * Well, we don't have vertical runs at this point and it's a pain
 * to calculate the leftmost vertical runs, decide which coordinate
 * should be changed and change it at this point.  Instead, we
 * just subtract 1 from the left endpoints (extends them to the left)
 * of all the runs that end on the left boundary.  This seems to
 * work well enough.
 *
 * Doesn't the policy of asking ip_classifySymbol for only the first
 * matching symbol mean that we miss some potential matches?  For example,
 * say A, B and C should all be in the same equivalence class.  We
 * fool around with A and only find that it matches B since B is encountered
 * before C.  Similarly, we fool around with B and only find A.  The
 * thing that's supposed to save us is that when we fool around with C,
 * we should find A or B.  But, isn't it possible that because of
 * commutivity problems, C doesn't match either A or B, even though either
 * A or B would match C?  Yes, it's possible.  This is another one of
 * those areas where the current scheme seems to get good results in
 * a reasonable amount of time.  If we decide the results should be
 * improved and time is less of an issue, we can revisit consolidation
 * strategies.
 *
 * There's another danger here.  Say we discover that symbol A matches
 * symbol B and symbol B matches symbol C.  Is there any guarantee
 * that A matches C?  No.  However, we've already described our solution
 * to this problem.  We limit the range of sizes allowed in each
 * equivalence class to be the intersection of all the size ranges of
 * the elements in the class.  If the symbol sizes are close enough,
 * all the other metrics seem to work well enough to prevent runaway
 * matching trains.
 *
 * Another approach would be to keep all the instances as Dan Huttenlocher
 * does so we can refine existing classes.  Bleah.  We can go
 * faster and take less memory if we can avoid the misclassification
 * in the first place instead of trying to remember how to back out of
 * it later.  That's where all the bizarre measures (slant, mass,
 * stroke width, volume) came from.  We'll try to stay on the path of not
 * making big mistakes in the first place.
 *
 * This is only the first pass the consolidation routine makes through
 * the list of symbols.  It needs a few more.  In the next, we mark all
 * the symbols that won't make it into the new list.  We also free any
 * runs and groups attached to the soon-to-be-discarded symbols.  We
 * leave all the lame duck symbols in the list so that it isn't changing
 * while we're still decided what's in and what's out.
 *
 * Let's talk about this second pass in more detail.  We inspect each
 * symbol in the list.  Symbols with no equivalents (a null equivalence
 * class pointer) are passed over.  They stay in the list.
 * A symbol with equivalents can be in one of three states: a) we
 * haven't seen the equivalence class yet; b) we've seen the class
 * and this symbol is the one that survived; c) we've seen this
 * class and this symbol will be discarded.  In the previous pass,
 * we marked symbols as candidates for deletion by setting the
 * number of instances of that symbol to 0.
 *
 * If a symbol with equivalents has an instance count of 0, we ignore it
 * because it was dealt with previously.  If the next symbol in the
 * equivalence class has an instance count of 0, the current symbol must be
 * the sole remaining representative of the equivalence class, so we
 * also ignore it - it has already been processed.
 *
 * If neither is true, we must find the member of the equivalence class
 * that will survive.  This is the symbol having the largest
 * number of original instances plus merged instances.  In case of a tie
 * (as will happen in all classes having exactly two members),
 * we pick the symbol with the largest number of original instances.
 * The instance counts of all the other members of the class are set to 0.
 *
 * This isn't a very sophisticated selection criterion.  This is probably
 * the place to try methods of making fancier exemplars.  One might try
 * counting the pixels used, weighted by number of matches.  I was
 * also thinking about choosing the exemplar with the fewest horizontal
 * and vertical runs.  While this does tend to select against noise on
 * the edges, it also tends to choose exemplars whose interior holes
 * are filled in.  Maybe later.  Maybe the symbol with the smallest
 * number of checkNoise groups?
 *
 * When we're done with the 2nd pass, we know which symbols we'll keep and
 * which ones we'll toss.  We go through the symbol list a third time.  Each
 * symbol with a non-zero instance count is assigned a new symbol ID,
 * though the old one is still visible.  To make things really work,
 * the info for the new exemplar for an equivalence class is stored
 * in the space occupied by the symbol with the lowest symbol ID.  This
 * guarantees that symbol IDs only get smaller when we do the final
 * mapping.  That's important because it A) moves the most used symbols
 * down to the bottom of the list where they get short IDs; and B)
 * makes the position info only get shorter.  The latter means we can
 * translate the position info, changing symbol IDs, in place without
 * allocating another buffer for the output.
 * 
 * When this pass is finished, we know the old and new symbol IDs of each
 * symbol.  It's (finally) time to delete the old symbols and make the
 * new list.  We go through the list again.  Each symbol with a non-zero
 * instance count is retained in the list.  The mapping from old symbol ID
 * to new ID is updated to map the old symbol number to the new one.
 * We remove the symbols with an instance count of 0 from the symbol list.
 *
 * When we're done, we have new list of symbols and a map that can be
 * used to convert old symbol ID's to new ones.  This is returned.
 *
 *				*	*	*
 *
 * Now we start talking about encoding format.  Originally, I had grand
 * thought about file formats - TIFF tags and stuff.  Then humility was
 * pounded in.  This is a lowly ipshared routine.  It doesn't know
 * anything about files.  It takes blocks of memory as both input
 * and output.  My thought was that I could use a relatively simple
 * output format and clients could perform further compression (LZW?),
 * encapsulate these memory blocks into files and/or put them into
 * a database or whatever.  Our job is only to provide the information.
 * How the information is passed around is the client's job.
 *
 * Well, then, things changed a bit.  An international standards body
 * began debating JBIG-2, a compression format that would be good for
 * images containing text.  AT&T proposed a system based on arithmetic
 * coding.  It does a good job, but arithmetic coding is slow and IBM
 * owns all the patents.  In response, William Rucklidge (who works
 * for Dan Huttenlocher) at PARC proposed his format, which is based
 * on Huffman and CCITT Group 4 codes.  It's a lot faster and there
 * are no patent issues.  The AT&T system does not depend on perfect
 * recognition, it only uses previous symbols as a prediction method
 * for the arithmetic coder.  Both this code and Dan H's code depend
 * on correct matchin of new symbols with existing dictionary symbols.
 * In a display of cooperation, the output of this symbol-based compression
 * package is being converted to William's proposed format.  This is
 * described in "A Proposal for JBIG-2 Coding," by William Ruckledge
 * at PARC (rucklidge@parc).  Maybe by the time you read this, there
 * will be an official JBIG-2 standard!
 *
 * But how?  Originally, I thought about junking all the current
 * output and replacing it with the new stuff.  There is a problem
 * with this method.  Say a client wants to convert several pages to
 * symbols using the same dictionary, then consolidate the dictionary,
 * then write out the dictionary and all the positions.  The symbol
 * IDs in the position info must contain the new symbol numbers.  You
 * can't know those until after the symbol consolidation since you
 * don't know which symbols will survive.  On the other hand, you
 * want a fairly compact way of representing the positions of the
 * symbols for all the pages that have been processed while waiting
 * for symbol consolidation.  I decided to stick with the method of
 * encoding position information described below for the intermediate
 * page descriptions, then convert them to William's format when
 * the client finally decides to write out the file.  The code
 * used to encode the dictionary changes entirely.
 *
 * So this being the policy, how do we encode the symbol ID and position
 * information in the intermediate files?  Let's use the following
 * simple byte code for symbol ID's:
 *	Symbol ID number	Binary Code
 *	0 - 127			0xxxxxxx
 *	128 - 32K+127		1xxxxxxx xxxxxxxx
 *
 * In the latter case, the low order byte appears first.  Since most
 * documents shouldn't have very many symbols and the common ones should
 * be encountered fairly early, we should often be able to encode most
 * symbol IDs in a single byte.
 *
 * What about position?  Many symbol-based compression schemes try to
 * order the stream of positions in the same order that a human would
 * read the characters.  Maybe someday I'll decide that's important,
 * but for now I'll try for small file size.
 *
 * We assume that the images have been deskewed and are oriented correctly.
 * Text will appear in horizontal lines and the bottoms of the characters
 * generally line up.  Therefore, we will have a lot of characters
 * ending on a relatively small number of scan lines, then a gap.  Clearly,
 * there's a lot of redundancy in the Y (vertical) component of the
 * symbol positions.  We will put out position information immediately
 * after finding the rightmost run of the lowest scan line of a character.
 * We will put out a new Y coordinate only if the Y coordinate of the
 * new symbol is different than the Y coordinate for the previous symbol
 * encoded.
 *
 * Here's an example.  Say we have a line of text characters, which
 * end on scan lines as shown below:
 *	character num	last scan line	x coord of last run
 *		1		500		104
 *		2		500		125
 *		3		499		148
 *		4		501		160
 *		5		500		182
 *		6		501		199
 *
 * The approximate encoding in the output file would be:
 *	Y: 499, X: 148, Symbol: 3;
 *	Y: 500, X: 104, Symbol 1; X: 125, Symbol: 2; X: 182, Symbol 5;
 *	Y: 501, X: 160, Symbol 4; X: 199, Symbol: 6;
 *
 *
 * Note that in the proposed JBIG-2 format, one can encode strips of
 * height 1, 2 or 4 rows.  To do this, we'll just keep pointers to
 * 1, 2 or 4 rows of positions, always putting out the next symbol.
 * 
 * The X position is the position of the right side of the symbol's
 * bounding box.  (In the proposed JBIG-2 format, it's the left side.
 * Drat.)  I was going to use the right end of the lowest run,
 * but that's probably noise so there will be little correlation between
 * this coordinate on the symbol instance and the symbol exemplar.
 * Unfortunately, encoding the right side of the bounding box means
 * we may occasionally have to encode negative numbers.  Why?  Well,
 * say we have a large, upside down U with a smaller upside down U
 * inside it.  Say the left sides of the open ends of both figures
 * are the lowest points of both figures and both end on the same
 * scan line.  In this case, we encode the position of the right side
 * of the large figure, then have to back up to record the right side
 * of the small figure.
 *
 * We modify the code given for symbol IDs as shown below.  For the
 * sake of simplicity, we encode the distance from the right edge
 * of one bounding box the right edge of the next.  The fancy business
 * about encoding only the space between the symbols is saved for
 * the final JBIG-2 encoding.
 *
 *	Relative X or Y Position	Binary Code
 *	+0 - +127			0xxxxxxx
 *	+128 - +16K+127			1xxxxxxx 0xxxxxxx
 *	-16K+128 - -1			1xxxxxxx 1xxxxxxx
 *
 * As before, the low order byte is read first, so the MSB of a two-byte
 * symbol is the MS bit of the second byte.  We encode numbers larger than
 * 128 and negative numbers as (0x80 OR (num & 0x7F)):(num >> 7) using an
 * arithmetic shift.  Since there are lots of symbols per scan line
 * (so the distance between them is relatively small) and symbols ID's
 * are often represented in a single byte, the total <position, symbolID>
 * pair can often be represented in two bytes.
 *
 * The end of a line, hence the presence of a new Y coordinate, is
 * signalled by a relative X position of 0.  Every Y coordinate is
 * always followed by an X coordinate.  If the X coordinate is 0, another
 * Y coordinate follows.  This allows us to step down very long pages.
 * Note that if the last run of a symbol abuts the left margin of the
 * page, we will have a non-zero X coordinate since we're encoding the
 * *right* end of the last scan line.
 *
 * Can you see the bug?  I didn't.  Very rarely, two symbols have
 * exactly the same position.  This happens when one reaches around the
 * the other so they end up with same lower right corner.  How do we encode
 * this?  An x offset of 0 is a tag, not a real value.  Since this is
 * so rare, it's ok to have a relatively long code.  I chose to encode
 * x offset of 0, y offset of -1 (which is the tag that we should stay
 * on the same line as opposed to ending the position info file), then
 * the correct X offset from the left side of the page.
 *
 *				*	*	*
 *
 * These intermediate position blocks are moderately compact and easy to
 * parse.  As stated above, we convert them into JBIG-2 blocks before
 * writing them out.  The format of a JBIG-2 position block is described
 * in William's document.
 *
 * An optional feature of the proposed encoding is the residue block.
 * This is a straight Group 4 encoded image of everything that appeared
 * just once over the set of pages compressed.  Therefore, it also
 * appears just once on the page where it is first encountered.  There
 * is one residue block per page.  The effect on the encoded size of
 * the file by using residue blocks is tiny - the same shapes are
 * encoded in the dictionary or in the residue block.  However,
 * using a residue block can dramatically reduce the amount of memory
 * required by the decoder.  An entire residue block is decoded onto
 * the image directly (each run is just painted there).  In some
 * implementations, each symbol in the dictionary is turned into a
 * bitmap and held in memory for a long as the dictionary exists.
 * Since the residue things tend to be big, this is a big waste of
 * space.  We will always use residue blocks.
 *
 * The most straightforward way to encode residue blocks is to allocate
 * a big bitmap, blit all the residue symbols back into it and encode
 * it.  Yuck - the bitmap is probably about a megabyte, which makes it
 * the biggest memory hog in the whole process by far.  One of the features
 * of our method of doing compression is supposed to be a relatively small
 * memory footprint.
 *
 * Our method is a little more complex.  We would like to convert the
 * runs of the image we already have into Group 4 runs fairly directly.
 * This raises several problems.  First, the coordinates in the runs
 * we've got have been multiplied by 2.  They are based relative to the
 * center of the symbol instead of being relative to the left edge of the
 * image.  The runs are ordered from right to left, bottom to top
 * instead of being left to right, top to bottom.
 *
 * A pleasant realization is that we can delete any symbol we find in
 * the residue.  Therefore, we can mess up its data structures a little
 * in the process of converting it to a residue block since we won't
 * be matching anything against this symbol again.
 *
 * Our first step is to figure out whether there are residue symbols
 * on the current page and, if so, what is the bounding box that contains
 * them.  We can save space by not encoding a bunch of blank white
 * space that's outside box.  Well, really, we don't save any space by
 * eliminating the white space on the left and right sides of the
 * bounding box since that compresses just fine.  So we really only
 * find the highest and lowest y coord of all of the residue symbols for
 * the current page.  If there are no residue symbols on this page,
 * we stop here.  In this step, we also link all the residue symbols for
 * this page together into a list that's sorted by the position of the
 * upper left corner of their bounding boxes.  The Y position is the
 * primary determinate and the X position is secondary.  The symbol's
 * position on the page is also recorded into the symbol structure
 * at this time so we don't have to go parse the position info again.
 *
 * Next, we run through the residue symbols, reordering and readjusting
 * each symbol's runs.  We go through the list of runs, resetting
 * each run endpoint coordinate to be relative to the page (instead
 * of the symbol's center) and reverse the order of the list by
 * changing the list pointers.
 *
 * We're now ready to encode each scan line of the residue image.  At
 * the beginning of each scan line, see if any symbols from the residue
 * symbol list (the one created above) can be transferred to our other
 * list, the active list.  A symbol can be transferred if its top scan
 * line is equal to the current scan line.  Symbols in the active list
 * are sorted by X position of their bounding boxes.
 *
 * We're ready to put out the runs for a single scan line.  They'll
 * end up in an array that is fed to the Group 4 encoding routine.
 * Unfortunately, we can't just copy the runs into the array by
 * copying the runs for the current scan line in each symbol in 
 * the order they're found in the active list.  While the runs of a
 * symbol are usually contiguous on the scan line, some complex symbols
 * wrap around other symbols.  This means the runs on the current
 * scan line must be sorted.  Yuck.  Well, I still thinks it's better
 * than blitting them all to the scan line and finding them again.
 *
 * Each run is a TCScanLineRun structure with four components: a pNext
 * pointer to the next run for that symbols (possibly on the next scan
 * line), left and right coords and a pClient pointer that really only
 * holds flags at this point.  We will use the pClient pointer to sort
 * the runs.  The basic idea is to create a sorted linked list of runs, then
 * copy the run coords from the linked list to the run array for
 * Group 4 encoding.
 *
 * During the sorting process, we maintain a pointer to the "current" run
 * in the linked list.  This is the last run inserted.  We think that
 * it is most likely that the next run will be inserted to the right
 * of this one since the symbols are roughly sorted left to right.
 * Each time we find a new symbol, we see if the first run fits between
 * the current run and the run to its right.  If the new run goes to
 * the left of the current run, we go back to the beginning of the
 * scan line and search for the proper insertion point.  This should
 * be rare.  If the new run goes to the right of the run to the right
 * of the current run, we search to the right for the proper insertion
 * point.  When this point is found, we string the new runs together.
 * We continue until a) we run out of runs in the current symbol, or b) the
 * next run from the current symbol should go to the right of the next
 * existing run (in which case we move right to find a new insertion point).
 * We use the pClient pointers to create this new list.  When all the
 * symbols in the active list have been examined, we copy the run coords
 * from the sorted list into the run Array and have the Group 4 encoder
 * do its thing.
 *
 * When we're done encoding the residue, we simply delete all the symbols
 * in the residue list.  By definition, they won't be needed anymore.
 *
 *				*	*	*
 *
 * William's format comprises a header, the symbol dictionary and a
 * set of pages.  Theoretically, there could be multiple dictionaries
 * sprinkled throughout the file, but that isn't supported today.
 * The basic format of a dictionary is a little header stuff
 * followed by a bunch of Group 4 pictures of images.  All symbols
 * in any one picture have the same height, so there is one picture
 * for each symbol height encountered.  Within a picture, the
 * symbols are ordered by width.
 *
 * The symbol "names" are also computed using a Huffman algorithm.  This
 * gives symbols that are used many times short names and symbols that
 * are used seldom long names.  These names are encoded in the
 * position information.
 *
 * We'll encode the symbols by walking through the 2D array of symbol
 * pointers.  All symbols in each row (symbols of the same height)
 * are encoded in one picture.  The actual encoding is done one line
 * at a time.  The runs for each symbol that appear on the current line
 * are copied to an array and offset by the symbol's position in the
 * picture of symbols.  The standard Group 4 code converts these into
 * an encoded data stream.
 *
 *				*	*	*
 *
 * For you historians - I was originally going to encode the stroke from
 * which symbols are made instead of encoding entire symbols.  The theory
 * was that a few strokes would be shared by lots of symbols, so would
 * result in better compression.  Is this true?  Well, each symbol would
 * be composed of a bunch of strokes IDs, each with an [x,y) position.
 * The stroke ID and position info take a bunch of bits, probably 8
 * each for stroke ID, X position and Y position.  If there are an
 * average of 1.5 runs per symbol scan line, we'd probably have about
 * 6 bits/scan line of straight MMR.  Say there are 6 strokes per symbol;
 * remember that strokes are pretty weird little things.  That's
 * 6 * 24 = 144 bits.  Plus we have to encode all the stroke definitions.
 * We'd like to amortize the number of bits required to do this over
 * each symbol, but to do that, we'd have to know how often each stroke
 * was reused.  Gluck.  Say that each symbol has 1 stroke of its own
 * and that stroke is 10 scan lines tall.  We use 4 bits/scan line,
 * so say a total of 40 bits.  Add that to the 144 and we get 184 bits/symbol.
 *
 * Let's do the MMR encoding.  Say that the character is 32 scan lines tall.
 * At an average of 1.5 runs/scan line and 4 bits/run, that's 32*6 = 202 bits.
 * It didn't seem like the speculative 10% reduction in symbol shape
 * size was worth the bother of inventing this new stuff.  Perhaps the
 * story will change in kanji?
 *
 *				*	*	*
 *
 * When this stuff is decompressed, we will get the shape information
 * independently of the position info.  We'll turn all of the
 * shapes into cached bitmaps.  We assume this doesn't take much space
 * because the big shapes (the feared "box around the outside of the page")
 * appear in the residue images for each page.
 * When each position is encountered, the matching shape is blitted
 * into position.
 *
 * As stated above, the dictionary symbol pictures are encoded in
 * strips; one strip for each symbol height encountered.  All symbols
 * of a given height are shown in the strip for that height.  During
 * decoding, one bitmap is allocated for each strip.
 *
 *				*	*	*
 *
 * This leads us to the software engineering questions.  Being a lowly
 * ipshared routine, we can't allocate blocks of memory and hand them
 * back to the client.  Only clients can allocate lasting memory, especially
 * large blocks of lasting memory.  It would be quite nice to give the
 * clients the ability to ask for the size memory blocks they need, let us
 * fill the blocks with data and let them trim the blocks to fits as
 * needed.  That's what we'll do.
 *
 * It's also important to realize that clients may wish to encode several
 * images with the same dictionary.  Most documents have more than one
 * page after all.  Therefore, we have to decouple the process of
 * computing new position information from the process of of putting out
 * shape info.  Second, peculiar clients, like the DAE, like to be able
 * to destroy all active memory between pages, so they need some way to
 * put out the current shape info at the end of each encoded page, then
 * read it back in again and rebuild the encoding data structures from that.
 * Yahoo.  Luckily, this isn't too hard with the current architecture.
 * We write the symbol info as usual when requested to.  When requested,
 * we read the runs for each symbol back in.  Just before classification,
 * we generate all the bizarre measures (slant, volumes, mass, stroke width)
 * and the groups if they don't already exist.
 *
 * How do we estimate sizes of buffers?  We use a weird trick to get an
 * estimate of the position buffer size.  Our goal is to be smaller
 * than a straight MMR encoding of the image.  If we can't do that, why
 * bother with all this tricky stuff?  The size in bytes of an MMR encoding of
 * an image containing roman text is roughly (number of black bits in the
 * image)/8.  Isn't that weird?  Images containing big black blocks encode
 * into fewer bytes.  Images containing lots of high frequency stuff, like
 * kanji, take more bytes.  But this is about right.  So, we just count
 * the black bits, divide by 8 and give that as our estimate.  Experience
 * has shown that clients should be prepared to try again with larger
 * buffers as necessary.
 *
 * The estimate for shape size is harder.  Until I know more, we'll
 * assume 4 bytes/symbol plus 1 byte/run for encoding info.
 * This seems kind of big, but I kept running out of space in the test cases.
 *
 *				*	*	*
 *
 * As Leigh Klotz found out, one can spend most of one's time
 * just allocating and freeing data structures.  We'll have some
 * special private heap management routines.  There will be multiple
 * private heaps, each dedicated to one structure type.  Each heap
 * will be kept as a simple linked list.
 *
 * Associated with each image being compressed are lots of interesting
 * variables.  Included in them are all the private heap pointers,
 * pointers to the list of symbols (really 2 lists, one organized by
 * symbol ID and one organized by size), the next symbol ID and whatever
 * else turns out to be useful.  To make things convenient, we'll
 * package all of this stuff up into an opaque struture.  We have the
 * client call us to allocate it.  The client passes it in to us on
 * every call.  When the client is done with the page, the client asks
 * us to destroy it.
 *
 * This still violates the most stringent requirements of the DAE since
 * this structure contains pointers.  GhostScript can't rearrange the
 * individual memory blocks because not only does our central structure
 * point to them, but they all point to each other (linked lists).  What
 * to do?  Getting rid of linked lists is awful since it means allocating
 * the largest data blocks you could ever need for every document.
 * Our pseudo-solution is to allow the client to turn all of the interesting
 * info into a bytestream at any time and reclaim it later.  The
 * DAE will have to do this before garbage-collection.  Ugly, but it's
 * the best I can think of now.
 *
 * Here's a list of the procedures that will make up the symbol
 * compression API:
 *
 *   i_estimateSizeOfPositionInfo in i_tcompr.c
 *	Estimates the size of the block of memory required to hold
 *	the position information.
 *
 *   i_estimateSizeOfSerializedCompressionInfo in i_tcompr.c
 *	Estimates the amount of memory required to hold serialized
 *	versions of the symbol shape information.
 *
 *   i_allocateTextCompressionInfo() in i_tcompr.c
 *	This routine allocates the central structure used to hold information
 *	about the image being compressed or decompressed.  This is
 *	the thing that's passed to the other routines.
 *	All internal memory is attached to this thing.  It records the
 *	addresses of client-supplied memory allocation and freeing
 *	routines.  This is also not wonderful, but this seems like
 *	the best way of asking clients to manage memory for the
 *	internal stuff.  We'll need to allocate and free this memory
 *	at random, odd times.
 *
 *   i_freeTextCompressionInfo in i_tcompr.c
 *	This routine frees all the text compression info.  It can be
 *	called in case of errors.  It should be called at the end
 *	of either compression or decompression.
 *
 *   i_convertImageToSymbols() in i_tcenc.c
 *	This procedure goes through the input image
 *	and builds, or adds to, the list of known symbols.
 *	It also puts out the symbol positions and ID's into the buffer
 *	whose size was estimated using i_estimateSizeOfPositionInfo.
 *	Along the way, it allocates memory for the symbols as necessary.
 *	Note that the symbol lists are (potentially) lasting data structures,
 *	so we must use a client-supplied allocator for them.
 *	Note that if an error is encountered by the routine, it does NOT
 *	free the symbol lists.  This allows the client to
 *	try again with more, or garbage-collected, memory.
 *	If, on entry, there is already a list of symbols and their runs
 *	but none of the other info needed for classification has been
 *	computed (as would be the case if the symbols have just been
 *	read in from a bytestream), this extra information is generated
 *	here.
 *
 *   i_consolidateShapeInfo() in i_tcenc.c
 *	This procedure consolidates known symbols into equivalence
 *	classes.  It produces a new list of symbols, prunes obsolete symbols
 *	from the symbol list and produces a map from old symbol IDs to
 *	new ones.
 *
 *   i_remapSymbolIDs() in i_tcenc.c
 *	Takes the symbol ID map produced by i_consolidateShapeInfo and
 *	pointers to old and new position info buffers.  It copies the
 *	position information from the old buffer to the new one, turning
 *	old symbol ID numbers to new ones as it goes.  The old and new
 *	position info buffers may be the same.
 *
 *   i_convertSymbolsToImage() in i_tcdec.c
 *	Does the inverse operation to i_convertImageToSymbols.  This
 *	routine takes a pointer to a bunch of position data, a Huffman
 *	decoding table and the symbol dictionary.  It produces an
 *	image with all the symbols in it.
 *	This routine assumes that there are already bitmaps associated
 *	with the symbols.  This ways done using i_loadDictionary
 *	with the forDecompression flag set.
 *
 * The following routines apply to the JBIG-2 encoding
 *
 *   i_storeDictionaryBlock() in i_tcjbe.c
 *	This routine is used to convert the symbol shape information into
 *	a compact bytestream.  It takes an argument of an array of booleans,
 *	one for each input dictionary that was merged to produce the
 *	current set of symbols.  The output block will be the merger
 *	of the input dictionaries whose booleans are TRUE.
 *
 *   i_loadDictionaryBlock() in i_tcjbd.c
 *	This routine merges a dictionary block with the existing
 *	internal shape information.  If the forDecompression flag
 *	is set, bitmaps are created for each symbol.  Otherwise,
 *	the information needed for compression is created.
 *	If the client is going to decompress a page (will call
 *	i_convertSymbolsToImage), the forDecompression flag should
 *	be set when the dictionaries are loaded.  If the client
 *	is going to do more compression (will call i_convertImageToSymbols),
 *	the forDecompression flag should be cleared.
 *
 *   various utility routines in i_tcenc.c, mitcenc.c, i_tcdbug.c
 *	These do things like classify individual symbols, make up groups
 *	and new symbols, create slant and volume measures, write symbol
 *	positions out to the position file.  Routines in i_tcdbug are 
 *	used to create position info for an image that contains
 *	all the reasonably sized symbols.
 *
 * Declarations of private prototypes and structures, like all the structures,
 * are in shrcmprs.prv.  The public structures and prototypes are in
 * shrcmprs.pub.
 *
 *
 * EXAMPLE:  Say you want to compress a document.  Here's some pseudo-code
 *		that shows the control flow.  The order of args in the
 *		routines will change.  Real code should also
 *		check the error codes from each of the routines.
 *
 *	i_allocateTextCompressionInfo(..., &pTextCompInfo, ...)
 *
 *	    * If you want to use an existing set of symbols do this
 *	    * step.
 *	i_loadDictionaryBlock(pTextCompInfo, cFalse, ...);
 *
 *	for (p = 0; p < maxPageNum; p++)
 *	{
 *	    i_estimateSizeOfPositionInfo(<page p image info>, &positionBufSize);
 *	    positionBuf = yourMemoryAllocator(positionBufSize);
 *	    i_convertImageToSymbols(<page p image info>, pTextCompInfo,
 *				    positionBuf, positionBufSize, ...);
 *	    <shrink positionBuf down to size used if desired>
 *	}
 *
 *	     * Now that we've seen all versions of each symbol,
 *	     * consolidate the versions.  Note that even if we started
 *	     * with a set of symbols, the classification process
 *	     * probably added some, so we need to write out the
 *	     * shape info again.
 *	     
 *	<allocate remapTable used to map every old symbol ID to a new ones>
 *	i_consolidateShapeInfo(pTextCompInfo, &remapTable, ...);
 *	for (p = 0; p < maxPageNum; p++)
 *	{
 *		* this step is optional, you can do the remapping in place *
 *	    newPositionBuf = yourMemoryAllocator(size of old position Buf p);
 *	    i_remapSymbolIDs(pTextCompInfo, <page p positionBuf>,
 *				newPositionBuf, remapTable, ...);
 *	    <free page p's old position buf>
 *	    <write new position buf data to file?>
 *	}
 *
 *
 *	i_estimateSizeOfDictionaryBlock(pTextCompInfo,
 *					&compInfoBufSize);
 *      compInfoBuf = yourMemoryAllocator(compInfoSize);
 *	i_storeDictionaryBlock(..., pTextCompInfo, compInfoBuf,
 *			       compInfoBufSize, ...);
 *	<write serialized compression info to file?>
 *
 *	i_freeTextCompressionInfo(pTextCompInfo);
 *
 *
 * EXAMPLE: Say you want to decompress a document.  Here's pseudo code
 *		with the same caveats.
 *
 *	i_allocateTextCompressionInfo(..., &pTextCompInfo, ...);
 *	i_loadDictionaryBlock(<ptr to bytes containing compression info>,
 *				pTextCompInfo, cTrue, ...);
 *
 *	for (p = 0; p < maxPageNum; p++)
 *	{
 *	    <page p info> = yourRoutineForGettingImageSpace(...);
 *	    i_convertSymbolsToImage(<page p info>, pTextCompInfo, ...);
 *	    <do something with image of page p>;
 *	}
 *
 *	i_freeTextCompressionInfo(pTextCompInfo);
 *
 * Here's a little info on bitmaps of symbols.  We only need them when
 * decompressing pages, hence the forDecompression parameter to
 * i_loadDictionaryBlock.  When set, ALL symbols in the dictionary are
 * converted to bitmaps.  We assume that really big, ugly symbols don't
 * make it into the dictionary - they're in the residue image associated
 * with each page.  Therefore, the amount of memory used for these
 * bitmaps isn't too big.
 *
 * If the forDecompression flag is not set, the incoming bitmaps
 * (which may or may not be compressed!) are immediately converted
 * into runs and the data needed for further compression is created.
 * This is done by simply extracting the connected component for each
 * symbol and classifying it.  Since it won't match anything else
 * in the dictionary, a new dictionary entry will be created for it.
 *
 * Many folks have said that it would be even cuter to just remember
 * the location of one (the first?) instance of each symbol in the
 * image where it was first found and blit that around as needed.
 * There are two problems with this.  First, we have no guarantee
 * that a rectangle containing this symbol won't overlap neighboring
 * symbols.  Therefore, if we extract a rectangle, we might get dreck
 * from the neighbors, which wouldn't look good in other instances
 * of the same symbol.  The galling thing is that it DOES work most
 * of the time.  The second problem is that the symbol dictionary
 * becomes the union of residues from all pages. Having to read and
 * retain pages 1 through n-1 to be able print page n is horrible.  
 *
 * If we were only decompressing one page, it might be more sensible to
 * accumulate a list of all the symbol positions first, then decompose
 * each symbol, blit it everywhere it goes and toss it.  This would
 * exchange a bunch of cached bitmaps for a bunch of positions.
 * However, Z Smith's copier statistics say that only about 25% of the
 * pieces of paper that come out of a copier are from a 1 page original.
 * The number of copies should roughly correspond to the number of pages
 * viewed in this world.  Therefore, we shouldn't optimize for the 25%
 * case.
 */



/* This routine is used to estimate the size of the output memory block.
 * Our guess is that the output file shouldn't be bigger than a straight
 * MMR encoding of the same image.  The heuristic we use to guess the size
 * of the MMR equivalent is to simply count the number of bits in the image
 * and divide by 8.  Weird, but it works pretty well for roman text.  Images
 * with very fine text, like kanji are bigger, but we'll see if this is
 * a problem before trying to fix it.
 */

/* PUBLIC */
/**************************************************************************
 *   i_estimateSizeOfPositionInfo()
 *
 *	This procedure estimates the size of the memory block that will
 *	be needed to hold the position information portion of a compressed
 *	image.  The estimate equals the number of 1's in the input
 *	(binary) file divided by 8.  The estimate is in units of bytes.
 *
 *	Input:
 *		pS:	Pointer to source image data including frame
 *		w:	width of image including frame
 *		h:	height of image including frame
 *		d:	depth of image (currently must be 1)
 *		bpl:	image bytes/line.
 *		pSize:	pointer to UInt32 that will contain the size.
 *			of the buffer required to hold the symbol
 *			information in bytes.
 *
 *	Returns:
 *		ia_successful		 it worked
 *		ia_depthNotSupported	 depth was not equal to 1
 *		whatever else i_countBits
 *		   returns		 something else happened.
 *
 **************************************************************************/
Int32  CDECL 
i_estimateSizeOfPositionInfo (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	w,
    UInt32	h,
    UInt32	d,
    UInt32	bpl,
    UInt32     *pSize)
{
    PROLOGUE;
    Int32	numBits;


	/* if the image isn't binary, return an error */
    if (d != 1)
	return ia_depthNotSupported;

	/* remove the frame */
    pS += (bpl << cLogFrameBits) + cFrameBytes;
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

	/* get the number of 1's in the image at x,y location 0,0 */
    __error__ = i_countBits (pIPState, pS, 0, 0, w, h, d, bpl, &numBits);

    if (__error__ == ia_successful)
	*pSize = numBits/8 + cTCPosHeaderVer2Len;
    else
        *pSize = 0;		/* something didn't work */

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_estimateSizeOfSerializedCompressionInfo()
 *
 *	This procedure estimates the size of the memory block that will
 *	be needed to hold the a serialized version of the pictures of
 *	symbols.  This info is NOT a bunch of pictures, we try to be
 *	significantly more compact than that.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pTextCompInfoBufSize:	pointer to UInt32 that will contain
 *					the number of bytes required to
 *					hold the text compression info.
 *					This is an estimate, callers
 *					should be prepared to discard
 *					unused space.
 *
 *	Returns:
 *		ia_successful		 it worked
 *
 **************************************************************************/
Int32  CDECL 
i_estimateSizeOfSerializedCompressionInfo (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt32     *pTextCompInfoBufSize)
{
    PROLOGUE;
    TCSymbol	*pSymbol;
    TCScanLineRun *pRun;
    UInt32	numberOfRuns;
    UInt32	totalBytes;
    TCCompressionInfoPtr pTextCompInfo;

    pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;

	/* The file starts with 6 bytes of magic number and version number */
    totalBytes = 6;

	/* account for the symbols and the number of lines in each one */
    for (pSymbol = pTextCompInfo->p1stSymbol;
	 pSymbol != NULL;
	 pSymbol = pSymbol->pNextNum)
    {
	    /* Allow 4 bytes/symbol.
	     */
	totalBytes += 4;		/* 2 bytes for width, 2 bytes
					 * for height.
					 */
	    /* plus 1 byte/run, which ought to be plenty.
	     */
	numberOfRuns = 0;
	for (pRun = pSymbol->pRunTail; pRun != NULL; pRun = pRun->pNext)
	    numberOfRuns++;
	totalBytes += numberOfRuns;
    }

    *pTextCompInfoBufSize = totalBytes;

    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *   i_allocateTextCompressionInfo()
 *
 *	This procedure allocates and initializes a structure that
 *	is used to hold internal text compression information between
 *	operations and between pages.
 *
 *	Input:
 *		ppTextCompInfo:	ptr to ptr to text compression info.
 *		allocator:	ptr to a procedure passed down by the client to
 *				allocate long-lasting data.  This procedure
 *				takes TWO arguments, the size of the block
 *				to be allocated and a boolean.  If the
 *				boolean equal cTrue, we will be notified
 *				every time a block allocated by this
 *				allocator is moved.  Notification takes
 *				the from of calling i_fixupPointerAfterMove.
 *				with the new and old block addresses.
 *		freer:		ptr to a procedure passed down by the client
 *				to free long-lasting data
 *
 *	Action:
 *		*ppTextCompInfo is set to point to a newly allocated
 *		block of text compression information
 *	Returns:
 *		ia_successful		it worked
 *		ia_nomem		ran out of memory
 *
 **************************************************************************/
Int32  CDECL 
i_allocateTextCompressionInfo (
    IPState    *pIPState,
    void      **ppTextCompInfo,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	bufSizeInBytes,
		    UInt32	fixupPointers),
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *pBlock,
		    UInt32	length))
{
    PROLOGUE;
    
    Int32			w, h;
    TCCompressionInfoPtr	pTextCompInfo;

	/* do the allocation */
    pTextCompInfo = (TCCompressionInfoPtr)
			(*allocator)(pIPState,
				     sizeof(TCCompressionInfo), cTrue);

    if (pTextCompInfo == NULL)
    {
	__error__ = ia_nomem;
	goto finAllocTextCompInfo;
    }

    *ppTextCompInfo = pTextCompInfo;
	/* Initialize the pointer fixup stuff */
    __error__ = ip_initPointerFixup (pIPState, &(pTextCompInfo->fixup),
				   sizeof(TCCompressionInfo),
				   NULL,
				   ip_fixupTCCompressionInfoPtrs);
    if (__error__ != ia_successful)
    {
	(*freer)(pIPState, pTextCompInfo, sizeof(TCCompressionInfo));
	goto finAllocTextCompInfo;
    }

	/* Allocate the 2D array of symbol pointers */
    pTextCompInfo->pSymIndex = (TCSymbolArray *)
				(*allocator)(pIPState,
					     cSizeOfTCSymbolArray, cTrue);
    if (pTextCompInfo->pSymIndex == NULL)
    {
	(*freer)(pIPState, pTextCompInfo, sizeof(TCCompressionInfo));
	__error__ = ia_nomem;
	goto finAllocTextCompInfo;
    }

	/* Initialize the the pointer block's pointer fixup stuff */
    __error__ = ip_initPointerFixup (pIPState, &(pTextCompInfo->pSymIndex->fixup),
				   cSizeOfTCSymbolArray,
				   &(pTextCompInfo->fixup),
				   ip_fixupSymbolArrayPtrs);
    if (__error__ != ia_successful)
    {
	(*freer)(pIPState, pTextCompInfo->pSymIndex, cSizeOfTCSymbolArray);
	(*freer)(pIPState, pTextCompInfo, sizeof(TCCompressionInfo));
	goto finAllocTextCompInfo;
    }

    pTextCompInfo->encodeState = 0;
    for (w = 0; w <= cMaxSymIndexDim; w++)
	for (h = 0; h <= cMaxSymIndexDim; h++)
	    (pTextCompInfo->pSymIndex->array)[w][h] = NULL;
    pTextCompInfo->pSymTable = NULL;
    pTextCompInfo->pBigSymbols = NULL;
    pTextCompInfo->p1stSymbol = NULL;
    pTextCompInfo->pLastSymbol = NULL;
    pTextCompInfo->p1stDictSymbol = NULL;
    pTextCompInfo->currentDictionaryNum = 0;
    pTextCompInfo->maxDictionaryNum = 0;
    pTextCompInfo->pPositionBuf = NULL;
    pTextCompInfo->positionBufSize = 0;
    pTextCompInfo->resultSize = 0;
    pTextCompInfo->lastX = 0;
    pTextCompInfo->lastY = 0;
    pTextCompInfo->exactMatchSet = cFalse;
    pTextCompInfo->exactMatchOnly = cFalse;
    pTextCompInfo->totalSymbols = 0;
    pTextCompInfo->nextSymbolNum = 0;
    pTextCompInfo->pVertRunHeadArray = NULL;
    pTextCompInfo->pVertRunTailArray = NULL;
    pTextCompInfo->ipcAllocator = allocator;
    pTextCompInfo->freer = freer;
    pTextCompInfo->classifyOnly = cFalse;
    pTextCompInfo->pClassifiedSymbol = NULL;
    pTextCompInfo->useResidual = cTrue;
    pTextCompInfo->p1stScanLineBlock = NULL;
    pTextCompInfo->p1stGroupBlock = NULL;
    pTextCompInfo->p1stSymbolBlock = NULL;
    pTextCompInfo->pBitmapBlock = NULL;
    pTextCompInfo->bitmapBlockSize = 0;

	/* Initialize debug counts */
    pTextCompInfo->sizeCheck = 0;
    pTextCompInfo->slantCheck = 0;
    pTextCompInfo->volumeCheck = 0;
    pTextCompInfo->strokeWidthCheck = 0;
    pTextCompInfo->diffCheck = 0;
    pTextCompInfo->noOffsetConsolidate = 0;
    pTextCompInfo->horizOffsetConsolidate = 0;
    pTextCompInfo->vertOffsetConsolidate = 0;

	/* setup the stuff used by the pointer fixup routines */

    __error__ = ia_successful;

finAllocTextCompInfo:
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_freeTextCompressionInfo()
 *
 *	This procedure frees all the information associated with
 *	symbol based text compression.
 *
 *	Input:
 *		ppTextCompInfoIn:	ptr to ptr to text compression info.
 *		freer:			ptr to a procedure passed down by
 *					the client to free this long-lasting
 *					data
 *
 *	Action:
 *		*ppTextCompInfo is set to point to a newly allocated
 *		block of text compression information
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
i_freeTextCompressionInfo (
    IPState    *pIPState,
    void       *pTextCompInfoIn)
{
    PROLOGUE;
    TCCompressionInfoPtr	 pTextCompInfo;
    
    TCSymbol			*pSymbol, *pTempSymbol;
    PointerFixup		*pBitmap;

    pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;

	/* Free all the memory blocks.  Since each of these wants to
	 * detach itself from the ring of blocks, and since that
	 * ring of blocks is vanishing, set all the pointers
	 * in the fixup blocks to point to themselves.  This
	 * tells the detaching proc to keep its hands off.
	 */
    __error__ = ip_isolateAllBlocksInPtrFixupRing (pIPState,
						 &(pTextCompInfo->fixup));
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* free the Symbol runs */
    if (pTextCompInfo->p1stScanLineBlock != NULL)
	i_freeAllMemoryBlocks (pIPState, &(pTextCompInfo->p1stScanLineBlock));
    if (pTextCompInfo->p1stGroupBlock != NULL)
	i_freeAllMemoryBlocks (pIPState, &(pTextCompInfo->p1stGroupBlock));

	/* If there is no common bitmap block, go through the
	 * stored symbols to see if there are any individual
	 * bitmap blocks produced by i_readTundraDictionaryBlock.
	 */
    if (pTextCompInfo->pBitmapBlock == NULL)
    {
	for (pSymbol = pTextCompInfo->p1stSymbol;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextNum)
	{
	    if (pSymbol->pBitmap != NULL)
	    {
		    /* eradicate all further references to this
		     * bitmap so we don't try to delete it again.
		     */
		for (pTempSymbol = pSymbol->pNextNum;
		     pTempSymbol != NULL;
		     pTempSymbol = pTempSymbol->pNextNum)
		{
		    if (pTempSymbol->pBitmap == pSymbol->pBitmap)
			pTempSymbol->pBitmap = NULL;
		}

		    /* free this bitmap.  We don't have to remove
		     * it from the list of PointerFixup blocks first
		     * since we just broke up that list above.
		     */
		pBitmap = (PointerFixup *) (pSymbol->pBitmap -
					    sizeof(PointerFixup));
		pTextCompInfo->freer(pIPState, pBitmap, pBitmap->blockSize);
	    }
	}
    }

	/* free the Scanned Symbols */
    if (pTextCompInfo->p1stSymbolBlock != NULL)
	i_freeAllMemoryBlocks (pIPState, &(pTextCompInfo->p1stSymbolBlock));

	/* Free the 1D array of pointers to all the symbols, arranged
	 * in SymbolID order.  This also detaches itself from the
	 * ring.
	 */
    ip_destroyTCSymbolTable (pIPState, pTextCompInfo);

	/* Free the 2D array of pointers to symbols.  The
	 * symbols will be freed when we free the p1stSymbolBlock
	 * heap
	 */
    if (pTextCompInfo->pSymIndex != NULL)
	pTextCompInfo->freer(pIPState, pTextCompInfo->pSymIndex,
			     cSizeOfTCSymbolArray);

	/* Free the arrays of pointers to vertical runs. */
    if (pTextCompInfo->pVertRunHeadArray != NULL)
	pTextCompInfo->freer(pIPState, pTextCompInfo->pVertRunHeadArray,
			     (pTextCompInfo->pVertRunHeadArray->numPtrs) *
						sizeof(TCScanLineRun *));
    if (pTextCompInfo->pVertRunTailArray != NULL)
	pTextCompInfo->freer(pIPState, pTextCompInfo->pVertRunTailArray,
			     (pTextCompInfo->pVertRunTailArray->numPtrs) *
						sizeof(TCScanLineRun *));

	/* free the bitmaps */
    if (pTextCompInfo->pBitmapBlock != NULL)
	pTextCompInfo->freer(pIPState, pTextCompInfo->pBitmapBlock,
			     pTextCompInfo->bitmapBlockSize);

	/* free the structure itself */
    pTextCompInfo->freer(pIPState, pTextCompInfo, sizeof(TCCompressionInfo));

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_createTCSymbolTable()
 *
 *	This procedure creates a table of symbol pointers.  They are
 *	ordered by symbol ID.  This table is really a cache of pointers
 *	that is created as needed and destroyed when invalidated.
 *
 *	Input:
 *		pTextCompInfo:		Ptr to all text compression info
 *
 *	Action:
 *		If a table already exists, it is freed.  A new table
 *		is allocated and filled with pointers to all the symbols.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_createTCSymbolTable (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo)
{
    PROLOGUE;
    TCSymbol	**pSymTable;
    TCSymbol	 *pCurSymbol;
    UInt32	  symTableSize;
    

	/* Destroy the table if it already exists. */
    __error__ = ip_destroyTCSymbolTable (pIPState, pTextCompInfo);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* Allocate a new table.  Remember that a TCSymbolTable declaration
	 * already contains one pointer, so we only need
	 * pTextCompInfo->nextSymbolNum - 1 more of them.
	 */
    symTableSize = sizeof(TCSymbolTable) +
		   (pTextCompInfo->nextSymbolNum - 1) * sizeof(TCSymbol *);
    pTextCompInfo->pSymTable =
	(TCSymbolTable *) (pTextCompInfo->ipcAllocator)(pIPState,
						     symTableSize, cTrue);
    if (pTextCompInfo->pSymTable == NULL)
    {
	__error__ = ia_nomem;
	goto finCreateSymTable;
    }

	/* Initialize the the symbol table's pointer fixup stuff */
    __error__ = ip_initPointerFixup (pIPState,
				     &(pTextCompInfo->pSymTable->fixup),
				     symTableSize,
				     &(pTextCompInfo->fixup),
				     ip_fixupTCSymbolTablePtrs);
    if (__error__ != ia_successful)
    {
	(pTextCompInfo->freer)(pIPState, pTextCompInfo->pSymTable,
			       symTableSize);
	goto finCreateSymTable;
    }

	/* The new table is safely allocated.
	 * Set the length of the table and initialize our local pointer
	 * to the table.
	 */
    pTextCompInfo->pSymTable->numPtrs = pTextCompInfo->nextSymbolNum;
    pSymTable = &(pTextCompInfo->pSymTable->table[0]);

	/* fill the table with pointers to the symbols */
    for (pCurSymbol = pTextCompInfo->p1stSymbol;
	 pCurSymbol != NULL;
	 pCurSymbol = pCurSymbol->pNextNum, pSymTable++)
    {
	*pSymTable = pCurSymbol;
    }

finCreateSymTable:
    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_destroyTCSymbolTable()
 *
 *	This procedure frees a table of symbol pointers.  They are
 *	ordered by symbol ID.  This table is really a cache of pointers
 *	that is created as needed and destroyed when invalidated.
 *
 *	Input:
 *		pTextCompInfo:		Ptr to all text compression info
 *
 *	Action:
 *		If a table already exists, it is freed.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_destroyTCSymbolTable (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo)
{
    PROLOGUE;
    UInt32	symTableSize;
    

	/* See if the table exists.  Remember that a TCSymbolTable declaration
	 * already contains one pointer, so we only have
	 * pTextCompInfo->pSymTable->symTableSize - 1 more of them.
	 */
    if (pTextCompInfo->pSymTable != NULL)
    {
	__error__ = ip_detachBlockFromPtrFixupRing (
				pIPState, 
				&(pTextCompInfo->pSymTable->fixup));
	if (__error__ != ia_successful)
	    NORMAL_RETURN;

	symTableSize = sizeof(TCSymbolTable) +
		       (pTextCompInfo->pSymTable->numPtrs - 1) *
			  sizeof(TCSymbol *);
	(pTextCompInfo->freer)(pIPState,
			       pTextCompInfo->pSymTable, symTableSize);

	    /* Ooops, almost forgot!  Set the symbol table pointer in the
	     * main structure back to NULL.
	     */
	pTextCompInfo->pSymTable = NULL;
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_deleteTCDictionarySymbol()
 *
 *	This procedure deletes as symbol which was in the text compression
 *	internal dictionary.  It deletes the symbol, as well as any
 *	groups, runs and other pointers to the symbol.
 *
 *	Input:
 *		pTextCompInfo:		Ptr to all text compression info
 *		pSymbol:		Ptr to the symbol to be freed
 *
 *	Returns:
 *		ia_successful		things went well
 *
 **************************************************************************/
Int32 CDECL 
ip_deleteTCDictionarySymbol (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCSymbol   *pSymbol)
{
    PROLOGUE;
    TCScanLineRun	*pRun, *pNextRun;
    TCGroupRun		*pGroup, *pNextGroup;
    Int32		 w, h;			/* for each symbol */
    TCSymbol		*pTempSymbol;
    PointerFixup	*pBitmap;
    UInt32		 bitmapSize;
    

	/* if the symbol ptr is null, just return */
    if (pSymbol == NULL)
	NORMAL_RETURN;

	/* Free the run pointers for this symbol */
    if (pSymbol->runsPtToTail)
	pRun = pSymbol->pRunHead;
    else
	pRun = pSymbol->pRunTail;

    for (;
	 pRun != NULL;
	 pRun = pNextRun)
    {
	pNextRun = pRun->pNext;
	i_freeStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
			(void **)&pRun);
    }

	/* free the horizontal groups in this symbol */
    for (pGroup = pSymbol->p1stHorizGroup;
	 pGroup != NULL;
	 pGroup = pNextGroup)
    {
	pNextGroup = pGroup->pNext;
	i_freeStruct (pIPState, pTextCompInfo->p1stGroupBlock,
		      (void **)&pGroup);
    }
	/* free the vertical groups in this symbol */
    for (pGroup = pSymbol->p1stVertGroup;
	 pGroup != NULL;
	 pGroup = pNextGroup)
    {
	pNextGroup = pGroup->pNext;
	i_freeStruct (pIPState, pTextCompInfo->p1stGroupBlock,
		      (void **)&pGroup);
    }

	/* Disconnect it from the list of symbols
	 * that are sorted by width and height.  Note
	 * that the same routine that set the symbol's
	 * runsPtToTail boolean set the minX, minY, maxX
	 * and maxY coords to be absolute, not relative.
	 * When they're relative, pSymbol->minX = -pSymbol->maxX =
	 * width - 1.
	 */
    if (pSymbol->runsPtToTail)
    {
	h = MIN(pSymbol->maxY - pSymbol->minY + 1, cMaxSymIndexDim);
	w = MIN(pSymbol->maxX - pSymbol->minX + 1, cMaxSymIndexDim);
    }
    else
    {
	h = MIN(pSymbol->maxY + 1, cMaxSymIndexDim);
	w = MIN(pSymbol->maxX + 1, cMaxSymIndexDim);
    }

	/* If the symbol we're removing is pointed to by the symbol
	 * index, update the symbol index.
	 */
    if ((pTextCompInfo->pSymIndex->array)[w][h] == pSymbol)
    {
	(pTextCompInfo->pSymIndex->array)[w][h] =
					    pSymbol->pNext;
    }

    if (pSymbol->pPrev != NULL)
	pSymbol->pPrev->pNext = pSymbol->pNext;
    if (pSymbol->pNext != NULL)
	pSymbol->pNext->pPrev = pSymbol->pPrev;

	/* Declare that the symbols are no longer sorted in dictionar
	 * order.  They probably aren't sorted anyway, but this will
	 * keep anyone from trying to follow the pNextDictSym ptr.
	 */
    pTextCompInfo->encodeState &= ~cTCSymsAreInDictOrder;

	/* If the symbol we're removing is pointed to by the first
	 * and/or last num pointers, update them.
	 */
    if (pTextCompInfo->pBigSymbols == pSymbol)
	pTextCompInfo->pBigSymbols = pSymbol->pNext;
    if (pTextCompInfo->p1stSymbol == pSymbol)
	pTextCompInfo->p1stSymbol = pSymbol->pNext;
    if (pTextCompInfo->pLastSymbol == pSymbol)
	pTextCompInfo->pLastSymbol = pSymbol->pPrev;

	/* Take this symbol out of the symbolID list */
    if (pSymbol->pPrevNum != NULL)
	pSymbol->pPrevNum->pNextNum = pSymbol->pNextNum;
    if (pSymbol->pNextNum != NULL)
	pSymbol->pNextNum->pPrevNum = pSymbol->pPrevNum;

	/* If this symbol has a bitmap attached to it and it`s
	 * not the common bitmap block shared among all symbols
	 * that's created by i_convertSymbolsToImage, see if
	 * that bitmap is used by any other symbol.  If not,
	 * free it.  This is the type of height class bitmap
	 * created by i_readTundraDictionaryBlock.
	 */
    if ((pSymbol->pBitmap != NULL) &&
	(pTextCompInfo->pBitmapBlock == NULL))
    {
	for (pTempSymbol = pTextCompInfo->p1stSymbol;
	     pTempSymbol != NULL;
	     pTempSymbol = pTempSymbol->pNextNum)
	{
	    if (pTempSymbol->pBitmap == pSymbol->pBitmap)
		break;
	}

	    /* If we didn't find another symbol with the
	     * same bitmap (note that pSymbol was already
	     * removed from the list), remove this bitmap.
	     */
	if (pTempSymbol == NULL)
	{
		/* First, fix up the symbol's bitmap pointer so it
		 * points to the pointer fixup part of the bitmap
		 * block.
		 */
	    pBitmap = (PointerFixup *) ((UInt8 *)(pSymbol->pBitmap) -
					sizeof(PointerFixup));
	    bitmapSize = pBitmap->blockSize;
	    __error__ = ip_detachBlockFromPtrFixupRing (pIPState, pBitmap);
	    if (__error__ != ia_successful)
		NORMAL_RETURN;
		/* and free the bitmap */
	    (pTextCompInfo->freer)(pIPState, pBitmap, bitmapSize);
	}
    }

	/* finally, free the symbol structure */
    i_freeStruct (pIPState, pTextCompInfo->p1stSymbolBlock,
		  (void **)&pSymbol);

    NORMAL_RETURN;
}


/*

 $Log: i_tcompr.c,v $
 Revision 1.2  2001/05/08 06:43:37  jlin
 change ->allocator to ->ipcAllocator

 * Revision 1.1.1.1  1999/09/20  18:18:13  jlin
 * Imported sources
 *
 Revision 1.27  1998/06/23 21:45:40  aber
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 CommonCheck: Fixes to force Tundra dictionary reading order, esp. for multiple
 dictionaries per page. Some cosmetic comment enhancements. A couple of
 extra error conditions caught.

 Revision 1.26  1998/01/13 23:50:09  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.25  1997/08/27 02:41:33  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Initialize the new useResidual flag in the TextCompInfo.

 Revision 1.24  1997/07/11 01:49:12  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.23  1997/04/29 02:45:46  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Initialize TCCompressionInfo lastX, exactMatchSet and exactMatchOnly
 parameters.

 Revision 1.22  1997/02/12 03:03:53  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Replace initializing massCheck with initializing strokeWidthCheck.

 Revision 1.21  1997/01/04 00:29:38  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Replace the mass and interior measures with the strokeWidth measurement.
 This does a much better job of distinguishing bold from non-bold characters.
 Fix comments so they talk about strokeWidth instead of interior.  Also
 remove interiorCheck as an element of TextCompressionInfo..

 Revision 1.20  1996/08/29 01:01:49  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Make i_destroyTCCompressionInfo delete any bitmaps associated with symbols
 before deleting the symbols.  Ditto for ip_deleteTCDictionarySymbol.

 Revision 1.19  1996/07/27 02:02:30  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Change TCCompressionInfo element symsHaveBeenEncoded to general encodeState
 >> structure element.
 Initialize new p1stDictSymbol element of TCCompressionInfo.

 Revision 1.18  1996/07/22 18:36:42  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Remove initialization of obsolete TCCompressionInfo entry called lastX.

 Revision 1.17  1996/07/19 00:08:14  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Add position header size to estimated size of position block.
 Fuss with indenting.
 Use the new ip_isolateAllBlocksInPtrFixupRing procedure before
 deleting any blocks when freeing the TCCompressionInfo and all attached
 blocks.  This prevents some of them from trying to reference already
 freed blocks.
 Use the new ip_destroyTCSymbolTable to free the new symbol table attached
 to the TCCompressionInfo.
 Add implementation of ip_createTCSymbolTable, a procedure that creates
 a table used to map symbol IDs to symbol structures.  Since the structure
 created contains pointers, it is put into the pointer fixup ring.
 Add implementation of ip_destroyTCSymbolTable, which detaches the
 table from the ring and frees it.
 Add implementation of ip_deleteTCDictionarySymbol, which carefully
 frees all structures and lists associated with a symbol that's in the
 TCCompressionInfo list and frees the symbol structure itself.

 Revision 1.16  1996/07/04 03:01:21  ddavies
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Initialize new TCCompressionInfo element, pLastSymbol.

 Revision 1.15  1996/07/02 22:52:15  ddavies
 ddavies on Tue Jul 2 15:49:13 PDT 1996

 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 Update comments once again to reflect the coming Tundra format for
 symbol-based compressed images.  This format was invented by William
 Rucklidge at PARC.  Also add inititalization for currentDictionaryNum
 and maxDictionaryNum in the TCCompressionInfo structure.

 Revision 1.14  1995/12/13 02:46:33  ejaquith
 i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.13  1995/07/17  22:01:53  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Change all occurences of sizeof(TCSymbolArray) to cSizeOfTCSymbolArray.
 * This is part of the changes for the Mac MPW compiler.
 *
 * Revision 1.12  1995/07/10  23:57:57  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Add inclusion of ptrfixup.pub to get PointerFixup structure.
 * Added comments about new arg used by allocator in i_allocateTextCompressionInfo.
 * The allocator's new arg specifies whether we should be notified when a
 * block that's allocated is moved by a garbage collector.
 * Add call to ip_initPointerFixup after allocating the
 * TCCompressionInfo block.  Also change the 2D array of symbol pointers
 * into a TCSymbolArray structure and the 2 1D arrays of TCScanLineRun
 * pointers into two TCScanLineArray structs (each with PointerFixup blocks
 * at the tops that must be initialized).
 * In i_freeTextCompressionInfo, free the heaps before freeing the other
 * stuff so that the garbage collection ring is intact when we free the
 * heaps.  They carefully try to maintain the integrity of the ring.
 *
 * Revision 1.11  1995/04/01  03:21:53  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Change references to pTextCompInfo->symIndex to (*(pTextCompInfo->pSymIndex)).
 * Also allocate the array in i_allocateTextCompressionInfo and free it
 * in i_freeTextCompressionInfo.
 * TCCompressionInfo structure now has a pointer to a 2D array of symbol
 * pointers instead of the array itself.
 *
 * Revision 1.10  1995/03/15  02:47:32  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Update the galactic comments describing symbol-based compression.  Lotsa
 * changes to correct the bugs found by XIS in January.
 * Also initialize the new debug counts in the global TCCompressionInfo
 * structure.
 * .,
 *
 * Revision 1.9  1994/11/20  22:58:02  ddavies
 * ddavies on Sun Nov 20 14:57:50 PST 1994
 *
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Redid high-level comment that describes the entire algorithm so that it
 * matches the way the thing works today.  Removed references to trees and
 * such.  Added descriptions of current classifier.
 *
 * Revision 1.8  1994/10/22  02:13:14  ddavies
 * ddavies on Fri Oct 21 19:12:37 PDT 1994
 *
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Fix bug in i_estimateSizeOfSerializedCompressionInfo that caused it to
 * skip a bunch of the symbols when deciding how much space they would require.
 *
 * Revision 1.7  1994/10/14  01:23:12  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Modify the estimated length of serialize info to use number of runs * 1.5 bytes.
 * Omit length of encoding from each symbol, we use the extension code to signal
 * the end of the runs instead.
 * When freeing the vertical run arrays, use the stored vertRunArrayWidth instead
 * of the width of the last image.
 *
 * Revision 1.6  1994/09/08  00:33:34  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Several changes to accomodate new members of the TCTextCompInfo structure.
 * Also changed the guess for the amount of space needed to hold the shape
 * info to be a function of the number of shapes.
 *
 * Revision 1.5  1994/08/03  19:17:31  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Snapshot of version of symbol-based compression code that does a good job
 * of distinguishing symbols.
 *
 * Revision 1.4  1994/07/14  21:02:42  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Snapshot of top level text compression module.  Lots of comments describing
 * the design changed since last time.  Minor changes in functions.  This is
 * still under test.
 *
 * Revision 1.3  1994/05/28  00:08:58  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Fixed zillions of bugs, this is a good time for a checkpoint.
 *
 * Revision 1.2  1994/05/03  20:45:16  ddavies
 * i_tcompr.c -> /product/ipcore/ipshared/src/RCS/i_tcompr.c,v
 *
 * Includes general improvements for symbol-based compression.  It's not
 * as good as it will be, but this is a reasonable checkpoint.
 *
 * Revision 1.1  1994/04/26  02:14:26  ddavies
 * Initial revision
 *

 */

