/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tcser.c: Contains procedures to turn an list of symbols and their
 *		runs into a compact bytestream.
 *
 *		Int32	i_serializeCompressionInfo()
 *			Convert list of symbols and their runs into a
 *			compact bytestream.
 *
 *		Int32	ip_genLongMHCodes()
 *			Write black or write code for a distance.  Check
 *			for buffer overflow.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#define _NEED_HORIZONTAL_CODES_
#define _NEED_VERTICAL_CODES_
#include "mmrenc.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /*  prototyes */
#include "shrcmprs.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tcser.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");


/* PUBLIC */
/**************************************************************************
 *   i_serializeCompressionInfo()
 *
 *	This procedure turns the obscure internal representation
 *	of the information used to describe connected components into
 *	a serial bytestream.
 *
 *	The text compression information is not affected by this process,
 *	i.e. you can try it again with a larger buffer if the first
 *	attempt fails.
 *
 *	The buffer of serialized data is allocated by the caller.  The
 *	size of this buffer is found by calling
 *	i_estimateSizeOfSerializedCompressionInfo.  The real number of
 *	bytes used will be returned in *pBytesUsed.
 *
 *	Input:
 *		pTextCompInfoIn:	ptr to text compression info.
 *		pShapeInfoBuf:		ptr to buffer used to hold the
 *					serialized data.
 *		bufSize:		number of bytes in the buffer as
 *					allocated
 *		deleteResidual:		If cTrue, do not include the
 *					symbols that only occur once in
 *					the output dictionary.
 *		pBytesUsed:		ptr to Int32 holding the number of
 *					bytes actually used in the buffer
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_bufferExceeded	text compression info didn't
 *					fit into the buffer space
 *
 **************************************************************************/
Int32  CDECL 
i_serializeCompressionInfo (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pShapeInfoBuf,
    UInt32	bufSize,
    Bool	deleteResidual,
    UInt32     *pBytesUsed)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo;
    TCScanLineRun *pRun, *pRefLeftEndRun, *pRefRightEndRun;
    TCSymbol   *pSymbol;
    Int32	symWidth;
    Int32	temp;		/* general temp variable */
    UInt8      *pCurBuf;	/* points to next vacant byte in output
				 * buffer. */
    UInt8      *pEndOfBuf;	/* points to byte beyond end of
				 * output buffer, used for bounds
				 * checking */
    UInt32	accum;		/* used to accumulate bytestream. */
    Int32	vacantBit;	/* number of 1st vacant bit in accum,
				 * 0 at LS end. */
    Int32	lastX;		/* last X position */
    UInt32	lastLine;	/* tags for line on which lastX is
				 * valid */
    UInt32	state;		/* cLeftEndpoint => ready to generate
				 * a left endpoint.
				 * cRightEndpoint => ready to generate
				 * a right endpoint.
				 */
    

	/* The routine encodes the symbols and their shapes into a
	 * bytestream.  As explained in i_tcompr.c, we encode the
	 * symbols in order of symbol ID.
	 *
	 * Before we describe the content of the output, let's say
	 * a word about format.  In order to make translation between
	 * little endian and big endian machines painless, this
	 * info is put out as a stream of bytes.  So long as the
	 * order of the bytes in the stream is maintained between platforms,
	 * information will be preserved.  When putting out multi-byte
	 * quantities, we put out the least significant byte first.
	 *
	 * It's useful to describe encoding and decoding at the same time.
	 * At least it helps me.  So what happens first?  We put out / take in
	 * a bunch of boilerplate at the beginning of the block.  This
	 * includes a magic number and a version number.  This is analogous
	 * to what`s done in the position info.
	 *
	 * Next, we go to the end of the list of symbols.  The list is
	 * kept in order of reverse symbol ID (larger symbol IDs at the
	 * front).  That means we find the symbol with the lowest ID.  As
	 * we go, we'll work our way to the beginning of the list.
	 *
	 * We begin each symbol with the symbol's width and height.
	 *
	 * For each symbol, we record:
	 *	symbolWidth	2 bytes
	 *	symbolHeight	2 bytes
	 *	useCount	2 bytes
	 *	pseudo-MMR encoding of the symbol shape.
	 *
	 * What are our deviations from real MMR?
	 *   1.	Since the runs are listed from right to left, bottom to
	 *	top, our encoding is also given in this order.
	 *   2. To make encoding and decoding a little faster, we reverse
	 *	the bit order of the codes.  That is, the LS bit of each
	 *	code is the "first" one.  The small benefit comes from
	 *	the fact that machines do reads and writes and table indexing
	 *	operations using the LS byte of a word, not the MS byte.
	 *	Therefore, we want the action to take place at the LS end
	 *	of a word.  For this to be true, we have to read the bits
	 *	from the LS end towards the MS end.
	 *   3.	Since we are encoding connected components, we know that
	 *	there are no blank lines.  Since the connected components
	 *	are almost always small, we know there are very few runs,
	 *	usually 1 or 2 per scan line.
	 *	In standard MMR, there's a fake run off the end of each
	 *	scan line.  One marks the end of the scan line by issuing
	 *	a vertical code saying that you line up with the fake run.
	 *	In normal MMR, this extra bit is an insignificant part
	 *	of the data.  In our tiny images, it's significant.
	 *	Therefore, we will eliminate it.  So, what do we do instead?
	 *	a. We maintain pointers to right and left endpoints on
	 *	   the previous line.  All the runs are in a single list,
	 *	   so the pointers can cross from the previous line to the
	 *	   current line unexpectedly.  When this happens, we
	 *	   ignore the endpoints they point to.
	 *	b. To make the previous line endpoints active again, the
	 *	   current line (the one being decoded) must cross a line
	 *	   boundary.  It can do this by 1) issuing a vertical
	 *	   match code to a run on the next line, or 2) issuing
	 *	   a horizontal code that wraps around to the next line.
	 *	   A PASS code never causes us to move to the next line
	 *	   since a PASS code implies that there's a run on the current
	 *	   line beyond the next run on the previous line.  Therefore,
	 *	   we can`t jump beyond the last run in the current line.
	 *   4.	The end of the symbol is signalled by the "extension" code
	 *	instead of the "end of facsimile block" code.  The extension
	 *	code is shorter and we won't use any extensions.
	 *
	 * The encoding loop is a state machine, much like the decoding
	 * loop.  In fact, we'll try to make them as similar as possible
	 * to reduce the chance of error.  We go around the loop once for
	 * each complete code put out or taken in.  I'd better explain
	 * MMR coding in order for that to make sense.  At least my
	 * interpretation of MMR coding.
	 *
	 * The standard draws several little pictures.  The "Reference Line"
	 * is the line above the one being coded.  The "Coding Line" is the
	 * image line being coded.  At each stage, we have a starting
	 * position, lastX, in the coding line.  That is obvious, but
	 * important.  The starting position in the coding line is
	 * immediately to the left (since we encode right to left) of
	 * the last pixel that was encoded.  An important point is that both
	 * the encoder and the decoder can identify the starting position
	 * since it reflects only things that have happened in the past.
	 * In the first first line, lastX is just off the right side
	 * of the image.
	 *
	 * Both the encoder and decoder can see the reference line, so
	 * let's start with that.  Starting at lastX in the reference
	 * line, the encoder looks to the left for the first pixel that a)
	 * has a different value than the pixel to its right and b) has the
	 * opposite value than the pixel at lastX in the codeing line.
	 * Even though we haven't encoded lastX, we can deduce its
	 * polarity from the sequence of coding state.
	 *
	 * For example, say we're about to encode a right endpoint.
	 * Then we look to the left for the first right endpoint.
	 * If we're going to encode a left endpoint, we look to
	 * the left in the reference line for the nearest left endpoint.
	 * The CCITT standard calls this first endpoint on the reference
	 * line b1.  The location of the next pixel to the left of b1 of
	 * opposite color from b1 is called b2.
	 *
	 * It's in the coding line that encoding and decoding become
	 * different.  In encoding, we look to the left for the first
	 * endpoint of proper left/right polarity.  Call this
	 * location a1.  We then look to the left of a1 for the next pixel
	 * of opposite color.  Call the location of this pixel a2.
	 * Several things can happen at this point.
	 *   1. There may be no pixels on the coding line under the 
	 *	reference line run (b1,b2).  This is indicated by b2
	 *	(the end of the run in the reference line) being to the
	 *	right of a1 (the beginning of the run in the coding line).
	 *	In fact, we require a difference of at least one pixel
	 *	between b2 and a1 to decide the run has really
	 *	ended.  In this case, we issue a PASS code.  This tells
	 *	both the encoder and decoder to skip the black or white run
	 *	on the reference line by advancing lastX to b2.  Note that
	 *	insisting on the 1-pixel buffer between b2 and a1 means
	 *	that the color of the pixel under the new lastX will be
	 *	the same as the color of the pixel under old lastX.  If
	 *	we were looking for a left endpoint originally, we'll
	 *	still be looking for one.  Ditto for right endpoints.
	 *   2. If we didn't issue a PASS code, the next point on the
	 *	reference line might be within distance 3 of the next
	 *	point on the coding line.  That is, | b1 - a1 | <= 3.
	 *	If this is true, we use a vertical code to specify the 
	 *	relative distance between them and move lastX up to a1's 
	 *	position.  Since we're only passing one endpoint, we swap 
	 *	the polarity of the next endpoint to be encoded or decoded.
	 *	Group 4 encoding is based on the notion that this is
	 *	the common case.
	 *   3. If neither of those work, we encode a horizontal code
	 *	followed by two distances.  One goes to the next endpoint, 
	 *	the other goes to the endpoint after that.  Why two distances?
	 *	Because we often want two insert two endpoints to get to
	 *	the next vertical match.  A brand new run or a fork inserts
	 *	two new endpoints.
	 *
	 * Here's a point that confused me.  Why do we always look to the
	 * left from lastX?  Say we have the following case:
	 *
	 *			X
	 *		      X X X
	 *
	 * It would be nice to encode the second line using 6 vertical
	 * codes (2 for each symbol), all relative to the bit on the first
	 * line.  However, the standard says that the pixel on the left
	 * will be encoded relative to some other pixel on the reference
	 * line because the obvious reference pixel isn't to the left of
	 * the coding pixel.  Why did they do that?
	 *
	 * Take the following example:
	 *		X	X
	 *		     X  X  X
	 *
	 * Since we didn't change the coding line, we should get the same
	 * code for the coding line, right?  It would be +2, +2, 0, 0, -2, -2.
	 * Now say we have the following image:
	 *		X	X
	 *	      X		X X
	 *
	 * The encoding would be the SAME: +2, +2, 0, 0, -2, -2.  Ooops.
	 * That's why they say you always have to look left (they say right
	 * since they're encoding left to right) for the reference run.
	 *
	 * Ok, but why *strictly* to the left?  Well, you have to decide
	 * on some convention to distinguish
	 *		X	X
	 *	      X		X X
	 * from
	 *		X	X
	 *	      X	X	  X
	 *
	 * and they decided that having a meaningful endpoint immediately
	 * above was much less likely than having it farther along the line.
	 *
	 * Well, here's where we make another little change from the
	 * standard.  What happens if the reference and coding line runs
	 * both touch the right end of the image?  If we don't start looking
	 * from a position off the edge of the image, we won't see the
	 * proper reference line endpoint because it won't be strictly to the
	 * left.  If we do start with an imaginary pixel off to the right,
	 * we have to remember not to include this imaginary pixel in the
	 * first code.  Well, we could do all that but it would be pain.
	 * Instead, we look up and to the right for reference line pixels
	 * when encoding.  That means that when we must look up and to the
	 * the left when decoding also.
	 *
	 *			*	*	*
	 *
	 * There are several layers of encoding.  At the top, is a layer that
	 * can find or generate one of the following codes (listed in bit
	 * reversed order relative to the standard documents - remember that
	 * one should read these from right to left, i.e. <--<--<--< ):
	 *
	 *	Code	Meaning
	 *	      1	Next coding line endpoint (left or right) is
	 *		immediately below the next reference line endpoint.
	 *	    110	Next coding line endpoint is 1 bit to the right of
	 *		next reference line endpoint.
	 *	    010	Next coding line endpoint is 1 bit to the left of
	 *		next reference line the endpoint.
	 *	    100	Horizontal distances to next two endpoints in
	 *		coding line relative to previous point in *coding*
	 *		line follow (see below)
	 *	   1000	Pass code - move lastX to end of next run in
	 *		reference line..
	 *	 110000	Next coding line endpoint is 2 bits to the right of
	 *		next reference line endpoint.
	 *	 010000	Next coding line endpoint is 2 bits to the left of
	 *		next reference line endpoint.
	 *	1100000	Next coding line endpoint is 3 bits to the right of
	 *		next reference line endpoint.
	 *	0100000	Next coding line endpoint is 3 bits to the left of
	 *		next reference line endpoint.
	 *	1000000	End of string.  This is the normal G4 extension code
	 *		but it's a lot shorter than the normal G4 "end of
	 *		block" code, so we use it for end of block.
	 *		The last MMR byte is padded with 0's so the next
	 *		symbol's width begins on a byte boundary.
	 *
	 * Each pass through the encoding loop generates one of these.
	 * Similarly, each pass through the decoding loop decodes one of
	 * these.  If the horizontal code is generated, we also generate
	 * distances to the next two endpoints before returning to the top
	 * of the loop.
	 *
	 * We can be in either of two states at the top of the loop,
	 * preparing to encode/decode a right endpoint or preparing to
	 * encode/decode a left endpoint.  Unfortunately, we need two 
	 * reference run pointers.  One points to the run having the next
	 * reference line left endpoint and one points to the run having
	 * the next reference line right endpoint.  Don't these always
	 * belong to the same run?  No.  Look at this example:
	 *
	 *	SSSSSSSSSSSSS		  RRRRRRRRRRRRRR
	 *		      DD	CCCCC	BBBBB  AA
	 *
	 * Once again, remember we're going from right to left.
	 * The reference right endpoint for run A is the right endpoint of
	 * R.  The reference right endpoint for runs B and C is the right
	 * endpoint of S.  The reference left endpoints of A and B is the
	 * left endpoint of R.  The reference left endpoint of runs C and D
	 * is the left endpoint of S.  So, for runs B and C, the reference
	 * left endpoint is to the right of the reference right endpoint.
	 * For runs A and D, it's the other way around.  The best way I've
	 * found to represent this is to have independent endpoint pointers.
	 * These pointers are set at the top of the loop given the current
	 * x coordinate, lastX.
	 * 
	 * If we get a horizontal code when preparing to generate a
	 * right endpoint, the next coordinates are for the right
	 * endpoint and the next left endpoint (specifying an entire run).
	 * If we're ready to generate a left endpoint, the two coordinates
	 * finish off the left end of the current run and start the right
	 * end of the next one.
	 *
	 * As stated above, if a horizontal code pushes us beyond the left
	 * edge of the bitmap, we wrap around to the right end on the
	 * next line.  The bitmap width equals the width of the symbol exactly.
	 * Since we're only translating to and from runlength data structures,
	 * we don't have to worry about padding symbol width to a multiple
	 * of 32 bits.
	 *
	 * The standard G3/G4 horizontal distance codes are a little weird.
	 * There is one code used to specify blocks of white space
	 * (used to cover the distance from a left endpoint of one run
	 * back to the right endpoint of the run to the left) and another
	 * code used to specify blocks of black pixels.  Why two codes?
	 * It's probably because black runs tend to be shorter than white runs.
	 * The black code has fewer bits/distance at the small distances.
	 * The white code has fewer bits/distance at the larger distances.
	 *
	 * Each of the codes has individual entries for all distances from
	 * 0 through 63.  They also have individual entries for each
	 * multiple of 64 up through 1728.  There is one common code for
	 * spanning pixels in blocks of 1792, 1856 (units of 64), ..., 2560.
	 * For distances greater than 2560, one concatenates 2560 codes
	 * together.  Therefore, a single distances has the form:
	 *	(<some number of 2560 codes>)
	 *	(<black or white code for 64 through 2496 in blocks of 64))
	 *	<terminating black or white code specifying a distance
	 *	 from 0 through 63>
	 *
	 * Only the last code is required.  If the block length is an
	 * exact multiple of 64, the terminating code specifies a distance
	 * of 0.  If the block is an exact multiple of 2560, the middle
	 * code is left out and the last code specifies 0.
	 * If n * 2560 < distance < n * 2560 + 64, the middle code is also
	 * left out and the terminating code specifies the offset from
	 * the last multiple of 2560.
	 *
	 * As stated above, the black and white codes are different.
	 * The white codes are nicely arranged so that using an 8 bit
	 * lookup table will return the following info:
	 *	1. code value (distance represented)
	 *	2. code length in bits (amount by which we should advance
	 *	   to get to the next code).
	 * If the code value is greater than 63, it is a non-terminating
	 * code.
	 * If the code length is 9 (the length of the longest white code),
	 * it is a 64x code and the final value should be found by adding
	 * the ninth bit * 64 to the code value.  If the answer is 1664,
	 * Then the answer is really 1728 (weird - is 1664 so common that
	 * it needs a special code?).  If the code length is > 9, this is
	 * a 64x code shared with the black codes.  Use the black lookup
	 * method as described next.
	 *
	 * The black codes have a nice feature.  Either the first four bits
	 * are 0000, or they're not.  If they are not 0000, we can use a
	 * lookup table indexed by 6 bits that encodes the 9 interesting
	 * entries. These represent the distances 1 through 9.  Each entry has
	 *	1. code value
	 *	2. code length in bits.
	 *
	 * If the first four bits are 0000, we get rid of them and use an 8-bit
	 * lookup table on the remaining bits.  The output of this
	 * table is the same as the output of the white table described above:
	 *	1. code value (distance represented)
	 *	2. code length in bits (amount by which we should advance
	 *	   to get to the next code after discarding the first 4 bits).
	 * There are 13-bit black codes, but they are nicely arranged.  So,
	 * if the code length after discarding the first 4 bits is 9, this
	 * is a 64x code and the final value should be found by adding the
	 * ninth bit * 64 to the code value.
	 *
	 * To encode a distance, we put out one 2560 code for each multiple
	 * of 2560 in the distance, then the correct 64x code (if any) for
	 * any non-zero multiples of 64 remaining, then the terminating code
	 * for the remainder.  Only the terminating code is required.  To
	 * decode a number, we read in codes, adding their values, until
	 * a terminating code (one whose length is less than 64) is reached.
	 *
	 *			*	*	*
	 *
	 * So much for reading and writing numbers in horizontal codes.  How
	 * does the encoding loop really work?  The encoding loop turns
	 * runs on the current and preceeding lines into a sequence of
	 * codes.  The decoding loop does the opposite.  The encoding loop
	 * tests its current position with respect to runs on the preceeding
	 * line to decode what codes to generate.  The decoding loop uses the
	 * tokens read to branch to branch to blocks of code to generate runs.
	 *
	 * Both loops have two main states: generating/expecting a left
	 * endpoint and generating/expecting a right endpoint.  These govern
	 * interpretation of input runs during encoding and generation of
	 * output runs during decoding.  They also govern the generation/
	 * interpretation of horizontal codes - either white then black if
	 * looking for a left endpoint or black then white if looking for a
	 * right endpoint.
	 *
	 * All the runs for a symbol are in a single linked list.  When either
	 * encoding and decoding, the pointer to the runs on the reference
	 * line can be advanced to point to the coding line.  In fact, it
	 * happens all the time.  How do we know when the reference line
	 * pointer is a) invalid, and b) valid again?
	 *
	 * When encoding, we have a set of well-formed runs.  The pClient
	 * pointers of these runs contain, not pointers, but alternating
	 * even/odd line tags.  All runs on one line may contain even tags
	 * and all runs on the next line contain odd tags.  We decide that
	 * the reference line run pointer is valid when the tags of the
	 * run it points to are not the same as the tags of the current run.
	 * This automatically takes care of making the reference line pointers
	 * invalid when they get too close and valid again when the coding
	 * line moves ahead.
	 *
	 * When decoding, things are a little more complicated.  We
	 * need to maintain the current line tags in an independent
	 * variable instead of looking at the line tags of the current
	 * run.  That's because when we get done generating the left
	 * endpoint of a run, there is no current run.  The next run may
	 * be on the same line or on a new line, we don't know yet.
	 *
	 * The previous line run pointer still becomes invalid when its
	 * line tags become the same as the current line tags.  It becomes
	 * valid again when the current line tags flip.  This happens
	 * when a) a horizontal code goes off the left edge of the image
	 * and wraps around; b) the reference line was invalid and a
	 * PASS code is found (we must be on a new line if we're now
	 * crossing over a run on the previous line);  c) a vertical code
	 * is found (we must be on a new line if we're directly over a
	 * run on the previous line).
	 *
	 * Isn't this really annoying?  Every time I'm about to tell you
	 * how it all works, I go back and give more background.  Well, we`re
	 * getting close now.
	 *
	 * We start encoding by issuing the horizontal codes for the first
	 * run and letting both reference line endpoint pointers point to it.
	 * This happens prior to entering the loop to set things up.  Decoding
	 * does a similar thing - setting up the first run from the first
	 * horizontal code and setting the reference line pointers to the new
	 * run.
	 *
	 * The loop terminates when we run out of runs, so we don't have
	 * to check for that explicitly inside.  Also note that the reference
	 * run pointers are behind the coding run pointer, so they can't
	 * be NULL either.
	 *
	 * Let's look a what happens when we want to encode a new right
	 * endpoint.
	 *   1. Skip right endpoints on the reference line until we either
	 *	find one that's to the left of the current position or have
	 *	wandered into the coding line.
	 *   2. If the reference right endpoint is valid (really points
	 *	to a run on the reference line) and the left endpoint of
	 *	that run is to the right of the next right endpoint on the
	 *	coding line + 1, issue a PASS code, set the current
	 *	horizontal position equal to the position of the reference
	 *	line's left endpoint - 1 (so we're back in white space)
	 *	and go back to the top of the loop with the state still 
	 *	set to right endpoint.
	 *   3. If a PASS code wasn't issued and the next coding line
	 *	right endpoint is within 3 pixels of the reference line
	 *	right endpoint, issue the corresponding vertical code, 
	 *	update the current coding line position to be the position 
	 *	of the coding line right endpoint, change the state to 
	 *	left endpoint and go back to the top of the loop.
	 *   4. If all else fails, issue a horizontal code specifying the
	 *	number of white pixels to the next right endpoint,
	 *	wrapping around the image if necessary, and the
	 *	number of pixels in the black run.  Set the lastX
	 *	to the position of the new left endpoint - 1 (so we're
	 *	back in white space) and go back to the top of the loop 
	 *	with the state still set to right endpoint.
	 *
	 * When generating left endpoints,
	 *   1. Skip left endpoints on the reference line until we either
	 *	find one that's to the left of the current position or have
	 *	wandered into the coding line.
	 *   2. If the reference line ptr is valid (really points to a run
	 *	on the reference line) and the right endpoint of the
	 *	*next* run on the reference line is to the right of the
	 *	next left endpoint on the coding line, issue a PASS code.
	 *	Also set the current x position to the position of the
	 *	new right endpoint (so we're still pointing to a black
	 *	pixel) and go back to the top of the loop with
	 *	the state still set to left endpoint.
	 *   3. If a PASS code wasn't issued and the next reference line
	 *	left endpoint is within 3 pixels of the next coding line
	 *	left endpoint, issue the corresponding vertical code, 
	 *	change the state to right endpoint and go back to the 
	 *	top of the loop.
	 *   4. If all else fails, issue a horizontal code specifying the
	 *	number of pixels to the next endpoint (a run of width 1
	 *	generates a 1, not a 0) and the number of white pixels
	 *	before the next right endpoint, wrapping around the image
	 *	if necessary.  Set the current horizontal position to
	 *	the position of the new right endpoint and go back to
	 *	the top of the loop with the state still set to left endpoint.
	 *
	 * There's a subtle feature here.  We encode the length of
	 * black and white runs, not the distance required to get from one
	 * endpoint to the next.  For example,  There are many 1-pixel
	 * black runs.  The left and right endpoints of these runs are
	 * equal.  However, we encode 1, not 0 for the black length.
	 * Originally, I was planning to encoding differences in endpoint
	 * coordinates instead of lengths, but the codes for 0 and 1 are
	 * too long.
	 *
	 * Let's say something about the mechanics of composing and decoding
	 * bitstreams.  In both cases, we keep the bitstream in a 32-bit
	 * register.  When composing, we keep a pointer to the first vacant
	 * bit relative to the LS end of the word.  New codes are shifted
	 * left by this much and ORed into the accumulator - to the right
	 * of the older codes.  The pointer to the first vacant bit is
	 * incremented by the length of the code.  Each time we insert
	 * a new code, we look at the new value of the pointer.  While
	 * it's greater than or equal to 8, we write out the bottom byte,
	 * shift the word right by 8 and subtract 8 from the pointer.
	 *
	 * At the bottom of the encoding loop, we flush the accumulator by
	 * shifting and writing until the vacant bit pointer is less than
	 * or equal to 0.  Since the logical right shifts insert 0's at
	 * the top of the accumulator, the last byte will be padded with
	 * 0's.  Then we write another 4 bytes so that we won't run off
	 * the end of the buffer when decoding.
	 *
	 * When decoding the bitstream, we also keep a pointer to the first
	 * vacant bit in the accumulator.  Before doing a decode, we look at
	 * the vacant bit pointer.  While it's less than or equal to 24, we
	 * shift the next input byte left to align it with the next vacant
	 * bit and OR it into the accumulator.  To decode words, we mask the
	 * lower bits of the accumulator and use the masked bits as an
	 * index into the tables described above.
	 *
	 * And one more thing.  Geez.  The endpoint coordinates inside the
	 * TCScanLineRun structs range from -w+1 to w-1 in steps of 2.
	 * This allows an extra bit of precision so we can tolerate a 
	 * little change in origin due to noise.  We'd don't want to 
	 * encode these 2x coordinates because the bigger numbers take
	 * a lot of extra bits.  Furthermore, we'd like it if the
	 * bitmap coordintes went from 0 to w-1 so we could easily tell
	 * when the coordinates wrapped around the left side (i.e. went
	 * negative).  So, we want a new set of coordinates.  We could
	 * either change them all in place and change them back, or
	 * make up a new set of runs for each symbol.  We'll change them all,
	 * then change them back.  This avoids the need for allocating
	 * extra memory at a time when we're probably trying to get
	 * rid of the memory occupied by the internal coding data structures.
	 * It should also optimize cache usage.
	 *
	 * Normally, when someone is creating a serial version of the
	 * symbols, the symbols are about to be thrown away.  This makes
	 * restoring the original coordinates seem sort of silly.
	 * However, we'll keep doing this until we demonstrate that
	 * there's a performance problem.
	 */

	/* Initialize local values of args.  We'd rather not destroy
	 * the originals */
    pTextCompInfo = (TCCompressionInfoPtr)pTextCompInfoIn;
    __error__ = ia_successful;			/* hope it stays this way */
    pCurBuf = pShapeInfoBuf;
    pEndOfBuf = pCurBuf + bufSize - 1;		/* pts to last entry in buf */

	/* Start the shape info with a magic number and version number.
	 */
    if (bufSize < 6)
    {
	__error__ = ia_bufferExceeded;
	goto tcSerializeCleanup;
    }

    pCurBuf[0] = (UInt8)(cTextCompShapeInfoMagicNum);
    pCurBuf[1] = (UInt8)(cTextCompShapeInfoMagicNum >> 8);
    pCurBuf[2] = (UInt8)(cTextCompShapeInfoMagicNum >> 16);
    pCurBuf[3] = (UInt8)(cTextCompShapeInfoMagicNum >> 24);

    pCurBuf[4] = (UInt8)(cTextCompShapeInfoVersionNum);
    pCurBuf[5] = (UInt8)(cTextCompShapeInfoVersionNum >> 8);

    pCurBuf += 6;

	/* Put out all the symbols starting with the symbol
	 * having symbol ID number 1.
	 */
    for (pSymbol = pTextCompInfo->p1stSymbol;
	 pSymbol != NULL;
	 pSymbol = pSymbol->pNextNum)
    {
	    /* If this symbol only occurs once and we're supposed
	     * to delete residual symbols, just get the next
	     * symbol.
	     */
	if ((deleteResidual) && (pSymbol->instanceCount == 1))
	    continue;

	    /* put out the symbol width and height */
	if (pCurBuf + 4 > pEndOfBuf)
	{
	    __error__ = ia_bufferExceeded;
	    goto tcSerializeCleanup;	/* no need to restore this symbol's
					 * runs, they haven't been touched
					 * yet.
					 */
	}

	    /* Write out the symbol's width */
	symWidth = (pSymbol->maxX - pSymbol->minX)/2 + 1;
	pCurBuf[0] = (UInt8)symWidth;
	pCurBuf[1] = (UInt8)(symWidth >> 8);

	    /* since we mostly use symWidth-1 from here, just decrement
	     * symWidth here.
	     */
	symWidth--;

	    /* write out the symbol's height */
	temp = (pSymbol->maxY - pSymbol->minY)/2 + 1;
	pCurBuf[2] = (UInt8)temp;
	pCurBuf[3] = (UInt8)(temp >> 8);

	    /* Put out number of times symbol is used */
	pCurBuf[4] = (UInt8) (pSymbol->instanceCount);
	pCurBuf[5] = (UInt8) (pSymbol->instanceCount >> 8);
	pCurBuf += 6;

	    /* change the run coordinates from the range [-w+1, w-1] to
	     * [0, w-1].  This makes the runs shorter so they can be
	     * encoded in fewer bits.
	     */
	for (pRun = pSymbol->pRunTail; pRun != NULL; pRun = pRun->pNext)
	{
	    pRun->runLeft = (pRun->runLeft + symWidth) / 2;
	    pRun->runRight = (pRun->runRight + symWidth) / 2;
	}

	    /* initialize position vars and run ptr for this symbol */
	lastX = symWidth;
	pRun = pSymbol->pRunTail;	/* start with the first run for
					 * this symbol. */

	    /* Put out a horizontal code for the first run.  The
	     * loop assumes that pReferenceRun points to valid run,
	     * so we have to put out a run so pReferenceRun can
	     * point to it.
	     */
	accum = cMMRHoriz;
	vacantBit = cMMRHorizLen;

	    /* First, the distance from the right edge of
	     * the image to the first right endpoint.  This is the number
	     * of white pixels on the right end of the image.
	     * If this is big number, let ip_genLongMHCodes do it,
	     * otherwise just do it inline.  This is the way we
	     * put out all horizontal codes.  Since there are VERY
	     * few long codes in characters, we normally do things
	     * inline without messing with the weirdness involved
	     * in creating long codes.
	     */
	temp = lastX - pRun->runRight;
	if (temp >= 64)
	{
	    __error__ = ip_genLongMHCodes (pIPState, temp,
					 mmrRevWhiteNonTermCodes,
					 mmrRevWhiteTermCodes,
					 &pCurBuf, pEndOfBuf,
					 &accum, &vacantBit);
	    if (__error__ != ia_successful)
		goto tcSerializeSymbolCleanup;
	}
	else
	{
		/* Write the terminating white run code for the first run */
	    accum |= mmrRevWhiteTermCodes[temp].codeBits << vacantBit;
	    vacantBit += mmrRevWhiteTermCodes[temp].codeLength;
		/* write out any full bytes */
	    while (vacantBit >= 8)
	    {
		if (pCurBuf == pEndOfBuf)
		    goto tcSerializeSymbolCleanup;
    
		    /* write out LS byte of accumulator and increment
		     * pointers */
		*pCurBuf = (UInt8) accum;
		pCurBuf++;
		accum >>= 8;			/* get rid of that byte */
		vacantBit -= 8;
	    }
	}
	    /* Do the same for the length of the black run. */
	temp = pRun->runRight - pRun->runLeft + 1;
	if (temp >= 64)
	{
	    __error__ = ip_genLongMHCodes (pIPState, temp,
					 mmrRevBlackNonTermCodes,
					 mmrRevBlackTermCodes,
					 &pCurBuf, pEndOfBuf,
					 &accum, &vacantBit);
	    if (__error__ != ia_successful)
		goto tcSerializeSymbolCleanup;
	}
	else
	{
		/* Write the terminating black run code for the first run */
	    accum |= mmrRevBlackTermCodes[temp].codeBits << vacantBit;
	    vacantBit += mmrRevBlackTermCodes[temp].codeLength;
		/* write out any full bytes */
	    while (vacantBit >= 8)
	    {
		if (pCurBuf == pEndOfBuf)
		    goto tcSerializeSymbolCleanup;
    
		    /* write out LS byte of accumulator and increment
		     * pointers */
		*pCurBuf = (UInt8) accum;
		pCurBuf++;
		accum >>= 8;			/* get rid of that byte */
		vacantBit -= 8;
	    }
	}

	    /* At this point, we've retired the first symbol run.
	     * Let pRefLeftEndRun and pRefRightEndRun point to it
	     * and go to the next run.
	     * Set the last x position and the line flags corresponding
	     * to this x position.
	     */
	lastX = pRun->runLeft - 1;
	if (lastX < 0)
	{
		/* the first run ended at the left edge of the symbol,
		 * so wrap around to the rightmost pixel in the symbol
		 * just as the decoder will do.  We're expecting to
		 * find a right endpoint next.  If the run on this line
		 * of the symbol butts up against the right edge of the
		 * symbol, we'll put out 0 white pixels as we're
		 * supposed to.
		 */
	    lastX = symWidth;
	    lastLine = cOddLineTag;
	}
	else
	    lastLine = cEvenLineTag;

	pRefLeftEndRun = pRefRightEndRun = pRun;
	pRun = pRun->pNext;
	state = cRightEndpoint;

	    /* Enter the loop that encodes all the rest of the runs
	     * for this symbol.
	     */
	while (pRun != NULL)
	{
	    switch (state)
	    {
	    case cRightEndpoint:
		    /* go find the the right endpoint on the reference
		     * line whose coordinate is less than or equal to
		     * the current X coordinate, lastX.
		     */
		while (((UInt32)pRefRightEndRun->pClient != lastLine) &&
		       (pRefRightEndRun->runRight > lastX))
		    pRefRightEndRun = pRefRightEndRun->pNext;

		    /* if pRefRightEndRun is now on the current line,
		     * make sure pRefLeftEndRun is there too.  Neither
		     * will be used right now, but that's ok.  We had
		     * a bug in which one of the pointers would advance
		     * to the next line, but the other wouldn't.
		     */
		if ((UInt32)pRefRightEndRun->pClient == lastLine)
		    pRefLeftEndRun = pRefRightEndRun;

		    /* First see if we should issue a PASS code.  This
		     * happens if the reference run is really on the
		     * previous line and its left endpoint is more than
		     * 1 bit to the right of the next right endpoint on
		     * the current line.  The "more than 1" weirdness
		     * is so that if we skip over the previous run,
		     * we'll still end up in white space.
		     */
		if ((pRefRightEndRun->pClient != pRun->pClient) &&
		    (pRefRightEndRun->runLeft - pRun->runRight > 1))
		{
		    accum |= cMMRPass << vacantBit;
		    vacantBit += cMMRPassLen;
			/* write out any full bytes */
		    while (vacantBit >= 8)
		    {
			if (pCurBuf == pEndOfBuf)
			    goto tcSerializeSymbolCleanup;
	    
			    /* write out LS byte of accumulator and increment
			     * pointers */
			*pCurBuf = (UInt8) accum;
			pCurBuf++;
			accum >>= 8;		/* get rid of that byte */
			vacantBit -= 8;
		    }
		    lastX = pRefRightEndRun->runLeft - 1;
			/* We don't have to check to see if pRefRightEndRun
			 * should be advanced by a line since we're under
			 * it by definition now.
			 */
		    lastLine = (UInt32)pRun->pClient;
			/* we don't have to check to see if lastX went
			 * negative here because a reference run that's
			 * up against the left edge of the symbol can't
			 * be PASS'd because there can't be a coding
			 * run to its left.
			 */
		}
		else
		    /* See if we should use a vertical code instead.
		     * This happens if the reference run is really on
		     * a different line and the distance from its right
		     * endpoint to the next right endpoint on this
		     * line is less than or equal to 3.
		     */
		if ((pRefRightEndRun->pClient != pRun->pClient) &&
		    ((UInt32)(pRun->runRight - pRefRightEndRun->runRight + 3)
									<= 6))
		{
		    temp = pRun->runRight - pRefRightEndRun->runRight + 3;
		    accum |= mmrRevVerticalCodes[temp] << vacantBit;
		    vacantBit += mmrRevVertCodeLengths[temp];
			/* write out any full bytes */
		    while (vacantBit >= 8)
		    {
			if (pCurBuf == pEndOfBuf)
			    goto tcSerializeSymbolCleanup;
	    
			    /* write out LS byte of accumulator and increment
			     * pointers */
			*pCurBuf = (UInt8) accum;
			pCurBuf++;
			accum >>= 8;		/* get rid of that byte */
			vacantBit -= 8;
		    }
			/* Point past the white stuff we've just skipped */
		    lastX = pRun->runRight;
		    lastLine = (UInt32)pRun->pClient;
		    state = cLeftEndpoint;
		}
		else	/* issue a horizontal code that gets us to the next
			 * right endpoint.  If the next right endpoint is on
			 * the next line, wrap around.
			 */
		{
			/* First, generate the horizontal code itself */
		    accum |= cMMRHoriz << vacantBit;
		    vacantBit += cMMRHorizLen;
		    while (vacantBit >= 8)
		    {
			if (pCurBuf == pEndOfBuf)
			    goto tcSerializeSymbolCleanup;
	    
			    /* write out LS byte of accum and increment
			     * pointers */
			*pCurBuf = (UInt8) accum;
			pCurBuf++;
			accum >>= 8;	/* get rid of that byte */
			vacantBit -= 8;
		    }

			/* lastX is the x coordinate of a white pixel.
			 * Get to the next right endpoint by skipping white
			 * space.  The numbers we calculate are the number
			 * of white pixels between lastX and the next right
			 * endpoint INCLUDING the pixel at lastX.
			 * To get the X coordinate of that right endpoint,
			 * the decoder subtracts the encoded number
			 * from its current X position.  If the value
			 * is negative, the decoder adds the width of
			 * the symbol.
			 */
		    if (lastLine == (UInt32) pRun->pClient)
			temp = lastX - pRun->runRight;
		    else
		    {
			temp = (lastX + 1) + (symWidth - pRun->runRight);
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			   pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			    /* we have to set lastLine here so that if the
			     * run we're encoding is the full width of the
			     * character, we'll update pRefRightEndRun
			     * correctly at the line transition at the
			     * end of this run.  There's no need to update
			     * lastX now since we don't look at it until
			     * after encoding the left endpoint anyway.
			     */
			lastLine = (UInt32) pRun->pClient;
		    }
			/* Long distances are encoded using ip_genLongMHCodes,
			 * short ones (much more common) are done inline.
			 */
		    if (temp >= 64)
		    {
			__error__ = ip_genLongMHCodes (pIPState, temp,
						     mmrRevWhiteNonTermCodes,
						     mmrRevWhiteTermCodes,
						     &pCurBuf, pEndOfBuf,
						     &accum, &vacantBit);
			if (__error__ != ia_successful)
			    goto tcSerializeSymbolCleanup;
		    }
		    else
		    {
			    /* Write the terminating white run code */
			accum |=
			    mmrRevWhiteTermCodes[temp].codeBits << vacantBit;
			vacantBit +=
			    mmrRevWhiteTermCodes[temp].codeLength;
			    /* write out any full bytes */
			while (vacantBit >= 8)
			{
			    if (pCurBuf == pEndOfBuf)
				goto tcSerializeSymbolCleanup;
		
				/* write out LS byte of accum and increment
				 * pointers */
			    *pCurBuf = (UInt8) accum;
			    pCurBuf++;
			    accum >>= 8;	/* get rid of that byte */
			    vacantBit -= 8;
			}
		    }

			/* Do the same for the length of the black run. */
		    temp = pRun->runRight - pRun->runLeft + 1;
		    if (temp >= 64)
		    {
			__error__ = ip_genLongMHCodes (pIPState, temp,
						     mmrRevBlackNonTermCodes,
						     mmrRevBlackTermCodes,
						     &pCurBuf, pEndOfBuf,
						     &accum, &vacantBit);
			if (__error__ != ia_successful)
			    goto tcSerializeSymbolCleanup;
		    }
		    else
		    {
			    /* Write the terminating black run code */
			accum |=
			    mmrRevBlackTermCodes[temp].codeBits << vacantBit;
			vacantBit +=
			    mmrRevBlackTermCodes[temp].codeLength;
			    /* write out any full bytes */
			while (vacantBit >= 8)
			{
			    if (pCurBuf == pEndOfBuf)
				goto tcSerializeSymbolCleanup;
		
				/* write out LS byte of accum and increment
				 * pointers */
			    *pCurBuf = (UInt8) accum;
			    pCurBuf++;
			    accum >>= 8;	/* get rid of that byte */
			    vacantBit -= 8;
			}
		    }

		    lastX = pRun->runLeft - 1;
			/* if that run abutted the left edge of the symbol,
			 * set lastX to the rightmost pixel on the previous
			 * line of the symbol.
			 * If that the symbol touches the right edge at
			 * that point, we'll happily say there are 0 white
			 * pixels before you get to black pixels.
			 * Also make sure that the reference run pointers
			 * are advanced so they really point to the line
			 * above the one we're about to encode.  If we
			 * didn't do this, the flags for the current
			 * reference run (line n-1) would become the
			 * same as the new value for lastLine (line n+1)
			 * even though they do not point to the same line.
			 */
		    if (lastX < 0)
		    {
			lastX = symWidth;
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			   pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			lastLine = (UInt32)pRun->pClient ^ cLineTagMask;
		    }
		    else
			lastLine = (UInt32)pRun->pClient;

			/* We've just skipped over a left endpoint, so
			 * go to the next run. */
		    pRun = pRun->pNext;
		}
		break;


	    case cLeftEndpoint:
		    /* go find the the left endpoint on the reference
		     * line whose coordinate is less than or equal to
		     * the current X coordinate, lastX.
		     */
		while (((UInt32)pRefLeftEndRun->pClient != lastLine) &&
		       (pRefLeftEndRun->runLeft > lastX))
		    pRefLeftEndRun = pRefLeftEndRun->pNext;

		    /* if pRefLeftEndRun is now on the current line,
		     * make sure pRefRightEndRun is there too.  Neither
		     * will be used right now, but that's ok.  We had
		     * a bug in which one of the pointers would advance
		     * to the next line, but the other wouldn't.
		     */
		if ((UInt32)pRefLeftEndRun->pClient == lastLine)
		    pRefRightEndRun = pRefLeftEndRun;

		    /* First see if we should issue a PASS code.  This
		     * happens if the reference run is really on the
		     * previous line and the run to the left of the
		     * reference run is also on the previous line and
		     * that run's right endpoint is at the same x coord
		     * or is to the right of the desired left endpoint.
		     * Note the difference in this test as compared to
		     * the test done for right endpoints.  There the
		     * difference in endpoints had to be greater than
		     * 1.  That's so we could end up in white space
		     * after the pass.  Here the difference must be
		     * greater than -1.  That's so we
		     * end up in black space (since the endpoints always
		     * point to black space).
		     */
		if ((pRefLeftEndRun->pClient != pRun->pClient) &&
		    (pRefLeftEndRun->pNext->pClient != pRun->pClient) &&
		    (pRefLeftEndRun->pNext->runRight >= pRun->runLeft))
		{
		    accum |= cMMRPass << vacantBit;
		    vacantBit += cMMRPassLen;
			/* write out any full bytes */
		    while (vacantBit >= 8)
		    {
			if (pCurBuf == pEndOfBuf)
			    goto tcSerializeSymbolCleanup;
	    
			    /* write out LS byte of accumulator and increment
			     * pointers */
			*pCurBuf = (UInt8) accum;
			pCurBuf++;
			accum >>= 8;		/* get rid of that byte */
			vacantBit -= 8;
		    }
			/* And point to the point immediately beyond the
			 * white stuff on the previous line we've just
			 * skipped over.
			 */
		    pRefLeftEndRun = pRefLeftEndRun->pNext;
		    lastX = pRefLeftEndRun->runRight;
		    lastLine = (UInt32) pRun->pClient;
		}
		else
		    /* See if we should use a vertical code instead.
		     * This happens if the reference run is really on
		     * a different line and the distance from its left
		     * endpoint to the next left endpoint on this
		     * line is less than or equal to 3.
		     */
		if ((pRefLeftEndRun->pClient != pRun->pClient) &&
		    ((UInt32)(pRun->runLeft - pRefLeftEndRun->runLeft + 3)
									<= 6))
		{
		    temp = pRun->runLeft - pRefLeftEndRun->runLeft + 3;
		    accum |= mmrRevVerticalCodes[temp] << vacantBit;
		    vacantBit += mmrRevVertCodeLengths[temp];
			/* write out any full bytes */
		    while (vacantBit >= 8)
		    {
			if (pCurBuf == pEndOfBuf)
			    goto tcSerializeSymbolCleanup;
	    
			    /* write out LS byte of accumulator and increment
			     * pointers */
			*pCurBuf = (UInt8) accum;
			pCurBuf++;
			accum >>= 8;		/* get rid of that byte */
			vacantBit -= 8;
		    }
			/* point just to the left of the black stuff
			 * we've just skipped.
			 */
		    lastX = pRun->runLeft - 1;
			/* if that run abutted the left edge of the symbol,
			 * set lastX to the rightmost pixel on the previous
			 * line of the symbol.
			 * If that the symbol touches the right edge at
			 * that point, we'll happily say there are 0 white
			 * pixels before you get to black pixels.
			 * Also make sure that the reference run pointers
			 * are advanced so they really point to the line
			 * above the one we're about to encode.  If we
			 * didn't do this, the flags for the current
			 * reference run (line n-1) would become the
			 * same as the new value for lastLine (line n+1)
			 * even though they do not point to the same line.
			 */
		    if (lastX < 0)
		    {
			lastX = symWidth;
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			   pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			lastLine = (UInt32)pRun->pClient ^ cLineTagMask;
		    }
		    else
			lastLine = (UInt32)pRun->pClient;
		    state = cRightEndpoint;
			/* We've also just skipped beyond a left endpoint
			 * in the coding line, so go to the next run
			 * in the coding line.
			 */
		    pRun = pRun->pNext; 
		}
		else	/* issue a horizontal code that gets us to the next
			 * left endpoint.  If the next right endpoint is on
			 * the next line, wrap around.
			 */
		{
			/* First, generate the horizontal code itself */
		    accum |= cMMRHoriz << vacantBit;
		    vacantBit += cMMRHorizLen;
		    while (vacantBit >= 8)
		    {
			if (pCurBuf == pEndOfBuf)
			    goto tcSerializeSymbolCleanup;
	    
			    /* write out LS byte of accum and increment
			     * pointers */
			*pCurBuf = (UInt8) accum;
			pCurBuf++;
			accum >>= 8;	/* get rid of that byte */
			vacantBit -= 8;
		    }

			/* Next, encode the length of the black run. */
		    temp = lastX - pRun->runLeft + 1;
		    if (temp >= 64)
		    {
			__error__ = ip_genLongMHCodes (pIPState, temp,
						     mmrRevBlackNonTermCodes,
						     mmrRevBlackTermCodes,
						     &pCurBuf, pEndOfBuf,
						     &accum, &vacantBit);
			if (__error__ != ia_successful)
			    goto tcSerializeSymbolCleanup;
		    }
		    else
		    {
			    /* Write the terminating black run code */
			accum |=
			    mmrRevBlackTermCodes[temp].codeBits << vacantBit;
			vacantBit +=
			    mmrRevBlackTermCodes[temp].codeLength;
			    /* write out any full bytes */
			while (vacantBit >= 8)
			{
			    if (pCurBuf == pEndOfBuf)
				goto tcSerializeSymbolCleanup;
		
				/* write out LS byte of accum and increment
				 * pointers */
			    *pCurBuf = (UInt8) accum;
			    pCurBuf++;
			    accum >>= 8;	/* get rid of that byte */
			    vacantBit -= 8;
			}
		    }
			/* Get to the next right endpoint by skipping white
			 * space.  First, we remember where we are and
			 * go to the next run.
			 * Get to the next right endpoint by skipping white
			 * space.  The numbers we calculate are the number
			 * of white pixels that are to the left of lastX
			 * and to the right of the next right endpoint.
			 * To get the X coordinate of that right endpoint,
			 * the decoder subtracts (encoded number + 1)
			 * from its current X position.  If the value
			 * is negative, it adds the width of the symbol.
			 * Also make sure that the reference run pointers
			 * are advanced so they really point to the line
			 * above the one we're about to encode.  If we
			 * didn't do this, the flags for the current
			 * reference run (line n-1) would become the
			 * same as the new value for lastLine (line n+1)
			 * even though they do not point to the same line.
			 */
		    lastX = pRun->runLeft - 1;
		    if (lastX < 0)
		    {
			lastX = symWidth;
			while ((UInt32)pRefRightEndRun->pClient != lastLine)
			   pRefRightEndRun = pRefRightEndRun->pNext;
			pRefLeftEndRun = pRefRightEndRun;
			lastLine = (UInt32)pRun->pClient ^ cLineTagMask;
		    }

			/* Go find the right endpoint of the next coding
			 * run */
		    pRun = pRun->pNext;
		    if (pRun == NULL)
			/* Ooops, there aren't any more runs.  Well, just
			 * get us to the left end of the current scan line.
			 */
		    {
			temp = lastX;
		    }
		    else
			/* lastX is the x coordinate of a white pixel.
			 * Get to the next right endpoint by skipping white
			 * space.  The numbers we calculate are the number
			 * of white pixels between lastX and the next right
			 * endpoint INCLUDING the pixel at lastX.
			 * To get the X coordinate of that right endpoint,
			 * the decoder subtracts the encoded number
			 * from its current X position.  If the value
			 * is negative, it adds the width of the symbol.
			 */
		    if (lastLine == (UInt32)pRun->pClient)
			temp = lastX - pRun->runRight;
		    else
		    {
			temp = (lastX + 1) + (symWidth - pRun->runRight);
			while ((UInt32)pRefLeftEndRun->pClient != lastLine)
			   pRefLeftEndRun = pRefLeftEndRun->pNext;
			pRefRightEndRun = pRefLeftEndRun;
		    }

			/* if the white code is long, use a procedure
			 * to generate it, otherwise just generate it
			 * inline.
			 */
		    if (temp >= 64)
		    {
			__error__ = ip_genLongMHCodes (pIPState, temp,
						     mmrRevWhiteNonTermCodes,
						     mmrRevWhiteTermCodes,
						     &pCurBuf, pEndOfBuf,
						     &accum, &vacantBit);
			if (__error__ != ia_successful)
			    goto tcSerializeSymbolCleanup;
		    }
		    else
		    {
			    /* Write the terminating white run code */
			accum |=
			    mmrRevWhiteTermCodes[temp].codeBits << vacantBit;
			vacantBit +=
			    mmrRevWhiteTermCodes[temp].codeLength;
			    /* write out any full bytes */
			while (vacantBit >= 8)
			{
			    if (pCurBuf == pEndOfBuf)
				goto tcSerializeSymbolCleanup;
		
				/* write out LS byte of accum and increment
				 * pointers */
			    *pCurBuf = (UInt8) accum;
			    pCurBuf++;
			    accum >>= 8;	/* get rid of that byte */
			    vacantBit -= 8;
			}
		    }

		    if (pRun != NULL)
		    {
			lastX = pRun->runRight;
			lastLine = (UInt32)pRun->pClient;
		    }
		}
		break;
	    }	/* end of switch on state */
	}	/* end of loop that encodes runs for 1 symbol */

	    /* Put out the extension code that signals the end of the
	     * run info for this symbol. Flush out to the next byte
	     * boundry so we can start the next symbol with width
	     * and height byte-aligned.
	     */
	accum |= cMMRExtension << vacantBit;
	vacantBit += cMMRExtensionLen;
	while (vacantBit > 0)
	{
	    if (pCurBuf == pEndOfBuf)
		goto tcSerializeSymbolCleanup;

		/* write out LS byte of accum and increment
		 * pointers */
	    *pCurBuf = (UInt8) accum;
	    pCurBuf++;
	    accum >>= 8;	/* get rid of that byte */
	    vacantBit -= 8;
	}

	    /* restore the original run coordinates. */
	for (pRun =  pSymbol->pRunTail; pRun != NULL; pRun = pRun->pNext)
	    {
		pRun->runLeft =  (Int16) (2 * pRun->runLeft - symWidth);
		pRun->runRight = (Int16) (2 * pRun->runRight - symWidth);
	    }
    }	/* end of loop that encodes all symbols */


	/* Make sure the buffer is big enough to hold 4 more bytes.
	 * These may be consumed by the decoder.  We want to make sure
	 * it doesn't run off the end of the buffer.
	 */
    if (pCurBuf + 4 > pEndOfBuf)
	__error__ = ia_bufferExceeded;
    else
    {
	pCurBuf[0] = 0;
	pCurBuf[1] = 0;
	pCurBuf[2] = 0;
	pCurBuf[3] = 0;
	pCurBuf += 4;
	__error__ = ia_successful;
    }

tcSerializeCleanup:
	/* if everything worked, update the number of bytes used */
    if (__error__ == ia_successful)
	*pBytesUsed = (Int32)pCurBuf - (Int32)pShapeInfoBuf;

    NORMAL_RETURN;


	/* We only get here when the shape info buffer was too small
	 * and we were in the middle of encoding a bunch of runs when
	 * we ran out of space.
	 * Set the proper return code and restore the original run
	 * coords.
	 */
tcSerializeSymbolCleanup:
	/* restore the original run coordinates. */
    for (pRun =  pSymbol->pRunTail; pRun != NULL; pRun = pRun->pNext)
	{
	    pRun->runLeft = (Int16) (2 * pRun->runLeft - symWidth);
	    pRun->runRight = (Int16) (2 * pRun->runRight - symWidth);
	}
    return ia_bufferExceeded;
}



/* PRIVATE */
/**************************************************************************
 *   ip_genLongMHCodes()
 *
 *	This procedure is called to encode large pixel numbers.
 *	"Large" means greater than or equal to 64 since these numbers
 *	require non-terminating codes.
 *	This is the bulk of the code involved in writing out lengths,
 *	but isn't use very often.  That makes it ideal for putting in
 *	a procedure.  Pixel counts that require only terminating
 *	codes are encoded inline by the caller.
 *
 *	Input:
 *		numPix:			number of pixels to be encoded.
 *		pNonTermTable:		ptr to table containing non-terminating
 *					codes.
 *		pTermTable:		ptr to table containing terminating
 *					codes.
 *		ppBuffer:		ptr to ptr to buffer into which code is
 *					to be written.
 *		pEndOfBuf:		ptr to last byte of buffer
 *		pAccum:			ptr to word containing a partial
 *					code byte.
 *		pVacantBit:		ptr to number of first vacant bit in
 *					accumulator.
 *
 *	Action:
 *		Encodes the non-terminating portion of length
 *		into the accumulator.  All complete accumulator bytes
 *		are written to the output buffer.  The buffer pointer is
 *		incremented to point to the next vacant byte.  The
 *		accumulator and vacant bit variables are updated
 *		before returning.  Also, the number of pixels to be
 *		encoded is set to the remainder of numPix divided
 *		by 64.  This selects the terminating code to be
 *		written when we return.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_bufferExceeded	ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_genLongMHCodes (
    IPState    *pIPState,
    UInt32	numPix,
    mmrTableEntry *pNonTermTable,
    mmrTableEntry *pTermTable,
    UInt8     **ppBuffer,
    UInt8      *pEndOfBuf,
    UInt32     *pAccum,
    Int32      *pVacantBit)
{
    PROLOGUE;
    UInt8      *pCurBuf = *ppBuffer;
    UInt32	accum = *pAccum;
    Int32	vacantBit = *pVacantBit;

	/* write out the multiples of 2560 */
	while (numPix >= 2560)
	{
	    accum |= pNonTermTable[2560/64].codeBits << vacantBit;
	    vacantBit += pNonTermTable[2560/64].codeLength;
	    numPix -= 2560;

		/* write out any full bytes */
	    while (vacantBit >= 8)
	    {
		if (pCurBuf == pEndOfBuf)
		    return ia_bufferExceeded;

		    /* write out LS byte of accumulator and increment
		     * pointers */
		*pCurBuf = (UInt8) accum;
		pCurBuf++;
		accum >>= 8;			/* get rid of that byte */
		vacantBit -= 8;
	    }
	}

	    /* Write out codes needed to get us to the next multiple of 64
	     * if necessary */
	if (numPix >= 64)
	{
	    accum |=
		pNonTermTable[numPix/64].codeBits << vacantBit;
	    vacantBit +=
		pNonTermTable[numPix/64].codeLength;
	    numPix &= 0x3F;			/* only len MOD 64 left */

		/* write out any full bytes */
	    while (vacantBit >= 8)
	    {
		if (pCurBuf == pEndOfBuf)
		    return ia_bufferExceeded;

		    /* write out LS byte of accumulator and increment
		     * pointers */
		*pCurBuf = (UInt8) accum;
		pCurBuf++;
		accum >>= 8;			/* get rid of that byte */
		vacantBit -= 8;
	    }
	}

	    /* Write the terminating run code */
	accum |= pTermTable[numPix].codeBits << vacantBit;
	vacantBit += pTermTable[numPix].codeLength;
	    /* write out any full bytes */
	while (vacantBit >= 8)
	{
	    if (pCurBuf == pEndOfBuf)
		return ia_bufferExceeded;

		/* write out LS byte of accum and increment
		 * pointers */
	    *pCurBuf = (UInt8) accum;
	    pCurBuf++;
	    accum >>= 8;	/* get rid of that byte */
	    vacantBit -= 8;
	}

	    /* IF we get to this point, we've successfully written out
	     * the non-terminating codes.  Update the caller's
	     * variables and return.
	     */
    *pAccum = accum;
    *pVacantBit = vacantBit;
    *ppBuffer = pCurBuf;
    NORMAL_RETURN;
}


/*

 $Log: i_tcser.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:13  jlin
 Imported sources

 Revision 1.10  1998/01/13 23:50:13  ddavies
 i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.9  1997/04/29 02:46:26  ddavies
 i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v

 Remove encoding the width of the vertical run arrays in the serialized
 dictionary.  It isn't needed.

 Revision 1.8  1996/08/22 01:21:28  ddavies
 i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v

 Add usageCount to info recorded for old serialized dictionaries.  The
 Tundra code will need it.  Yeah, the Tundra code will only use these
 serialized dictionaries while it's being created, but that's ok.

 Revision 1.7  1996/07/19 00:09:35  ddavies
 i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v

 Add deleteResidual parameter to i_serializeCompressionInfo.  When set,
 all symbols whose instanceCount == 1 are left out of the serialized
 compression info.  This is used when we're also creating residual images
 for each page.

 Revision 1.6  1996/07/04 03:02:13  ddavies
 i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v

 Change order of symbol linked list, so p1stSymbol points to symbol with
 lowest symbol ID instead of symbol with the highest symbol ID.
 This also means that pSymbol->pNextNum points to the symbol with the next
 higher symbolID instead of the next lower one.

 Revision 1.5  1996/04/10 21:57:33  ddavies
 ddavies on Wed Apr 10 14:56:58 PDT 1996

 i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v

 Define _NEED_HORIZONTAL_CODES_ and _NEED_VERTICAL_CODES_ before including
 mmrenc.prv so we get both sets of tables.

 Revision 1.4  1995/12/13 02:46:37  ejaquith
 i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/10/18  03:59:16  ddavies
 * i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.2  1995/07/10  23:59:17  ddavies
 * i_tcser.c -> /product/ipcore/ipshared/src/RCS/i_tcser.c,v
 *
 * When writing out width of the array of vertical run pointers,
 * use the width held in the TCScanLineArray structure instead of the
 * width held in the TCComressionInfo.  That width is gone.
 *
 * Revision 1.1  1994/10/14  01:25:09  ddavies
 * Initial revision
 *

 */

