/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tctnd.c:	Utility procedures used for reading Tundra format data
 *		for either further symbol-based compression or for
 *		decompression.
 *
 *	Int32	i_readTundraHeader()
 *			Checks the validity of a Tundra header.
 *
 *	Int32	i_checkTundraVersion()
 *			Checks whether version is supported.
 *
 *	Int32	i_readTundraNumberOfPagesField()
 *			Reads the number of pages specified in a Tundra
 *			memory block.
 *
 *	Int32	i_readTundraDictionaryBlockLengthField()
 *			Reads the length in bytes of a Tundra dictionary
 *			block
 *
 *	Int32	i_getExistingTundraPageNameSize()
 *			Returns the size of the page name string in
 *			an existing Tundra Page header.
 *			
 *	Int32	i_readTundraPageHeader()
 *			Reads a Tundra Page header
 *
 *	Int32	i_readTundraResidualHeader()
 *			Reads a Tundra Residual block header
 *
 *   	Int32	i_readTundraDictionaryBlock()
 *			Merges symbols in new Tundra dictionary block
 *			with existing symbols.
 *
 *   	Int32	i_readTundraPositionBlock()
 *			Converts Tundra position information into a bitmap.
 *
 *   	Int32	i_readTundraResidualBlock()
 *			Converts Tundra residual block into a bitmap.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "memory.pub"
#include "lmask.prv"
#include "imageref.pub"
#include "tundra.prv"	/* for Tundra constants */

    /*  prototyes */
#include "shrip.pub"
#include "shrpixr.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#include "shrrast.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_tctnd.c,v $; $Revision: 1.2 $; $Date: 2001/05/08 06:43:38 $");

/* There is a big comment describing our approach to symbol based
 * compression in i_tcompr.c.
 *
 * These routines are used to read Tundra memory blocks (read from
 * Tundra files).  i_readTundraDictionaryBlock merges the symbols contained
 * in a Tundra dictionary block with whatever symbols have already been
 * loaded.  The client must say whether the symbols are to be used to
 * decompress a page (symbol shapes kept as bitmaps) or to compress
 * more pages (symbol shapes kept as runs and other classification
 * info is generated).
 *
 * While there is an internal format for position data,
 * i_readTundraPositionBlock does not convert to that format.  It
 * simply turns the position block into a bitmap.
 * While there may be occasions in which one wants to convert an
 * encoded position block back into the internal format used during
 * compression, I haven't thought of any yet.  Even moving a page
 * from one file to another can be done without decompressing the
 * page block.
 */


/* PUBLIC */
/**************************************************************************
 *   i_readTundraHeader()
 *
 *	This procedure parses a Tundra header, checking for correctness.
 *	It also returns the number of bytes read and the version number
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pHeader:		pointer to the Tundra header block
 *		headerSize:		size of the Tundra header passed in
 *		pVersion:		ptr to version recorded in header.
 *		pBytesRead:		number of bytes read in header.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	the block wasn't big enough
 *		ia_decodeFailure	block was not a Tundra header
 *
 **************************************************************************/
Int32  CDECL 
i_readTundraHeader (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pHeader,
    UInt32	headerSize,
    UInt8      *pVersion,
    UInt32     *pBytesRead)
{
    PROLOGUE;
    char	headerIDString[] = {cTundraHeaderIDString};
    Int32	i;
    UInt8	*pLocalBuf = pHeader;

	/* Keep the compiler happy */
    pTextCompInfoIn = pTextCompInfoIn;

	/* check to see if the pointers are ok. */
    if ((pHeader == NULL) || (pBytesRead == NULL) || (pVersion == NULL))
	return ia_invalidParm;

    if (headerSize < cTundraHeaderSize)
	return ia_bufferExceeded;

	/* compare the header bytes with a known good header */
    for (i = 0; i < cTundraHeaderIDSize; i++)
    {
	if ((char)pLocalBuf[0] != headerIDString[i])
	    return ia_decodeFailure;
	pLocalBuf++;
    }

	/* Return the version number */
    *pVersion = *pLocalBuf++;

	/* Set the number of bytes read. */
    *pBytesRead = pLocalBuf - pHeader;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_checkTundraVersion()
 *
 *	This procedure checks to see if the current code can read Tundra
 *	files having the version that was passed in.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		version:		the version in question
 *		pVersionIsOk:		ptr to the answer.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_checkTundraVersion (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8	version,
    Bool       *pVersionIsOk)
{
	/* Keep the compiler happy */
    PROLOGUE;
    pTextCompInfoIn = pTextCompInfoIn;

	/* check to see if the pointers are ok. */
    if ((pVersionIsOk == NULL))
	return ia_invalidParm;

	/* We only support one version now */
    if (version == cTundraCurrentVersion)
	*pVersionIsOk = cTrue;
    else
	*pVersionIsOk = cFalse;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_readTundraNumberOfPagesField()
 *
 *	This procedure reads in the number of pages field in a Tundra
 *	memory block
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pNumOfPagesField:	pointer to the Tundra numberOfPages
 *					field
 *		bufSize:		number of bytes available to
 *					read.
 *		pNumOfPages:		the quantity that was read.
 *		pBytesRead:		number of bytes used.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_readTundraNumberOfPagesField (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pNumOfPagesField,
    UInt32	bufSize,
    UInt32     *pNumOfPages,
    UInt32     *pBytesRead)
{
    PROLOGUE;
    UInt8*pLocalBuf = pNumOfPagesField;

	/* Keep the compiler happy */
    pTextCompInfoIn = pTextCompInfoIn;

	/* check to see if the block exists and is big enough. */
    if ((pNumOfPagesField == NULL) || (pBytesRead == NULL) ||
	(pNumOfPages == NULL))
	return ia_invalidParm;

	/* if the number bytes in the buffer won't hold the field,
	 * complain.
	 */
    if (bufSize < cTundraNumOfPagesSize)
	return ia_bufferExceeded;

	/* Yea, it fits.  Read the page number, MS byte first */
    *pNumOfPages = ((UInt32) pLocalBuf[0] << 8) + (UInt32) pLocalBuf[1];

    *pBytesRead = cTundraNumOfPagesSize;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_readTundraDictionaryBlockLengthField()
 *
 *	This procedure reads in the field containing the number of
 *	bytes in a Tundra dictionary.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pDictBlkLenField:	pointer to the Tundra dictionary
 *					block length field
 *		bufSize:		number of bytes available to
 *					read.
 *		pDictBlkLen:		the quantity that was read.
 *		pBytesRead:		number of bytes used.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_readTundraDictionaryBlockLengthField (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pDictBlkLenField,
    UInt32	bufSize,
    UInt32     *pDictBlkLen,
    UInt32     *pBytesRead)
{
    PROLOGUE;
    UInt8      *pLocalBuf = pDictBlkLenField;

	/* Keep the compiler happy */
    pTextCompInfoIn = pTextCompInfoIn;

	/* check to see if the block exists and is big enough. */
    if ((pDictBlkLenField == NULL) || (pBytesRead == NULL) ||
	(pDictBlkLen == NULL))
	return ia_invalidParm;

	/* if the number bytes in the buffer won't hold the field,
	 * complain.
	 */
    if (bufSize < cTundraDictBlockLengthSize)
	return ia_bufferExceeded;

	/* Yea, it fits.  Read the page number, MS byte first */
    *pDictBlkLen = ((UInt32) pLocalBuf[0] << 24) +
		   ((UInt32) pLocalBuf[1] << 16) +
		   ((UInt32) pLocalBuf[2] << 8) +
		   ((UInt32) pLocalBuf[3]);

    *pBytesRead = cTundraDictBlockLengthSize;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getExistingTundraPageNameSize()
 *
 *	This procedure tells the client how much space is needed for
 *	a string that will hold the page name string contained in an
 *	existing Tundra Page header.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pPageHeader:		pointer to the Tundra Page header
 *		bufSize:		number of bytes available to
 *					read.
 *		pPageNameLen:		length of the string NOT INCLUDING the
 *					null terminator.  This is equivalent
 *					to what strlen returns.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_getExistingTundraPageNameSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pPageHeader,
    UInt32	bufSize,
    UInt32     *pPageNameLen)
{
    PROLOGUE;
    UInt32	i;

    pTextCompInfoIn = pTextCompInfoIn;

	/* go find a null, counting the chars along the way */
    for (i = 0; i < bufSize; i++)
    {
	if (pPageHeader[i] == 0)
	    break;
    }

	/* did we really find it or run off the end of the buffer? */
    if (i < bufSize)
	*pPageNameLen = i;
    else
	return ia_bufferExceeded;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_readTundraPageHeader()
 *
 *	This procedure reads a Tundra Page header, returning all the
 *	components.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pPageHeader:		pointer to an existing Tundra
 *					pageHeader
 *		pageNameBufLen:		length of buffer that will hold
 *					page name.  Must have room for
 *					the NULL at the end.
 *		bufSize:		number of bytes available to
 *					read.
 *		pPageName:		pointer to block of memory that
 *					will hold page name string. The
 *					size of this block can be found using
 *					i_getExistingTundraPageNameSize.
 *		pPageWidth:		ptr to returned page width.
 *		pPageHeight:		ptr to returned page height.
 *		pPosBlockLength:	ptr to returned length of position
 *					block.
 *		pBytesRead:		number of bytes used.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	field is bigger than room available,
 *					also returned if pageNameLen is smaller
 *					than page name string..
 *
 **************************************************************************/
Int32  CDECL 
i_readTundraPageHeader (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pPageHeader,
    UInt32	pageNameBufLen,
    UInt32	bufSize,
    char       *pPageName,
    UInt32     *pPageWidth,
    UInt32     *pPageHeight,
    UInt32     *pPosBlockLength,
    UInt32     *pBytesRead)
{
    PROLOGUE;
    UInt8      *pLocalBuf = pPageHeader;
    UInt32	nameLength;
    

	/* check to see if the block exists and is big enough. */
    if ((pPageHeader == NULL) || (pBytesRead == NULL) ||
	(pPageName == NULL) || (pPageWidth == NULL) ||
	(pPageHeight == NULL) || (pPosBlockLength == NULL))
	return ia_invalidParm;

	/* find out how big the page name string is */
    __error__ = i_getExistingTundraPageNameSize (pIPState, pTextCompInfoIn,
					       pPageHeader,
					       bufSize,
					       &nameLength);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* Is the buffer big enough to hold the header?  Remember
	 * that the nameLength does not include the string
	 * termination character, so we have to add that in.
	 * Also return this code if the name string doesn't
	 * fit in the client's string.
	 */
    if ((nameLength + 1 + cTundraPageHeaderFixedSize > bufSize) ||
	(nameLength + 1 > pageNameBufLen))
	return ia_bufferExceeded;

	/* Great, we can read the page header.  Read the
	 * page name, the width, height and position block length
	 * in that order.
	 */
    strcpy(pPageName, (char *)pLocalBuf);
    pLocalBuf += nameLength + 1;

    *pPageWidth = ((UInt32) pLocalBuf[0] << 8) + (UInt32) pLocalBuf[1];
    *pPageHeight = ((UInt32) pLocalBuf[2] << 8) + (UInt32) pLocalBuf[3];

    *pPosBlockLength = ((UInt32) pLocalBuf[4] << 24) +
		       ((UInt32) pLocalBuf[5] << 16) +
		       ((UInt32) pLocalBuf[6] << 8) +
		        (UInt32) pLocalBuf[7];
    pLocalBuf += 8;

    *pBytesRead = pLocalBuf - pPageHeader;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_readTundraResidualHeader()
 *
 *	This procedure reads a Tundra Residual header, returning all the
 *	components.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pResidualHeader:	pointer to an existing Tundra
 *					residual header
 *		bufSize:		number of bytes available to
 *					read.
 *		pResidualX:		ptr to returned coord of left side
 *					of residual relative to bitmap.  No
 *					frames are used.
 *		pResidualY:		ptr to returned coord of top side
 *					of residual relative to bitmap.  No
 *					frames are used.
 *		pResidualWidth:		ptr to returned residual width.
 *		pResidualHeight:	ptr to returned residual height.
 *		pResidualLength:	ptr to returned length of residual
 *					block.
 *		pBytesRead:		number of bytes used.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_readTundraResidualHeader (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pResidualHeader,
    UInt32	bufSize,
    UInt32     *pResidualX,
    UInt32     *pResidualY,
    UInt32     *pResidualWidth,
    UInt32     *pResidualHeight,
    UInt32     *pResidualLength,
    UInt32     *pBytesRead)
{
    PROLOGUE;
    UInt8      *pLocalBuf = pResidualHeader;
    UInt32	headerLength;
    

	/* check to see if the block exists and is big enough. */
    if ((pResidualHeader == NULL) || (pBytesRead == NULL) ||
	(pResidualX == NULL) || (pResidualY == NULL) ||
	(pResidualWidth == NULL) || (pResidualHeight == NULL) ||
	(pResidualLength == NULL))
	return ia_invalidParm;

	/* see if the entire header is contained in the buffer supplied */
    __error__ = i_getTundraResidualHeaderSize (pIPState,
					     pTextCompInfoIn, &headerLength);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    if (bufSize < headerLength)
	return ia_bufferExceeded;

	/* Yea, it fits.  Read the block's x, y, width, height
	 * and position block length in that order.
	 */
    *pResidualX = ((UInt32) pLocalBuf[0] << 8) + (UInt32) pLocalBuf[1];
    *pResidualY = ((UInt32) pLocalBuf[2] << 8) + (UInt32) pLocalBuf[3];
    *pResidualWidth = ((UInt32) pLocalBuf[4] << 8) + (UInt32) pLocalBuf[5];
    *pResidualHeight = ((UInt32) pLocalBuf[6] << 8) + (UInt32) pLocalBuf[7];

    *pResidualLength = ((UInt32) pLocalBuf[8] << 24) +
		       ((UInt32) pLocalBuf[9] << 16) +
		       ((UInt32) pLocalBuf[10] << 8) +
		        (UInt32) pLocalBuf[11];

	/* return the number of bytes read */
    *pBytesRead = cTundraResidualHeaderSize;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_readTundraDictionaryBlock()
 *
 *	This procedure merges the symbols in a Tundra dictionary block
 *	with the symbols already in an internal format, if any.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pDictBuf:		pointer to buffer containing
 *					Tundra dictionary block
 *		bufSize:		number of bytes available to
 *					read.
 *		dictNumber:		the number to be assigned to
 *					this dictionary.
 *		forDecompression:	boolean that's set if the symbols
 *					will be used only for decompression.
 *					If they will NOT be used for
 *					decompression (perhaps for further
 *					compression), set this to cFalse.
 *					In the current implementation, this
 *					parameter must ALWAYS be passed as
 *					cTrue.  If the client wants to
 *					write out a dictionary, read it back
 *					in and continue compressing, use
 *					i_serializeCompressionInfo and
 *					i_deserializeCompressionInfo.
 *
 *	Note:
 *		The internal and archive (Tundra) formats for images aren't
 *		supposed to mix.  If the client has called
 *		i_convertSymbolsToImage (which converts symbols deserialized
 *		with i_deserializeCompressionInfo and internal position
 *		info into an image) with this pTextCompInfoIn structure and
 *		now calls this routines, we'll return ia_noImpl.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_noImpl		forDecompression was not set or
 *					there was a prior call to
 *					i_convertSymbolsToImage with this
 *					text compression info block.
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_readTundraDictionaryBlock (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pDictBuf,
    UInt32	bufSize,
    UInt32	dictNumber,
    Bool	forDecompression)
{
    PROLOGUE;
    TCCompressionInfo	*pTextCompInfo =
				(TCCompressionInfoPtr) pTextCompInfoIn;
						/* real ptr to text
						 * compression info */
    void		*pCCITTDecompInfo = NULL;
						/* the struct used to decompress
						 * height class bitmaps */
    Int32		 ccittInfoSize;		/* size of
						 * CCITTDecompressionInfo */
    TCSymbol		*p1stHeightSym;		/* ptr to first symbol in
						 * current height class */
    Int32		 temp;			/* scratch var */
    Int32		 maxWidth;		/* width of current height class
						 * bitmap */
    UInt8		*pBitmap;		/* ptr to height class bitmap */
    Int32		 bytesPerLine;		/* bytes/Line of pBitmap */
    Int32		 compBitmapSize;	/* size of compressed height
						 * class bitmap. */
    Int32		 decompBitmapSize;	/* size of decompressed height
						 * class bitmap. */
    Bool		 exhaustedImageData;	/* flags from MMR decomp */
    Bool		 exhaustedCompData;
    Int32		 percentDone;
    UInt8		*pLocIn = pDictBuf;
						/* local ptr to input buf. */
    UInt8		*pLocEnd;		/* pts to end of input buf */
    TCInstance		 instance;		/* used to pass info about new
						 * symbol. */
    UInt32		 accum;			/* bit accumulator */
    Int32		 vacantBits;		/* number of vacant bits on
						 * right end of accum */
    Int32		 numItemsArray[32];	/* number of symbols having
						 * a particular Huffman code
						 * length. */
    TCTundraDecTable	*pWidthHeightDecoder;	/* width/height decoder */
    TCTundraDecTable	*pUseDecoder;		/* use count decoder */
    TCTundraEncClass	*pEncClass;		/* scr encoding class ptr */
    TCTundraDecSelectorArray
			*pWidthHeightSel = NULL; /* ptr to struct containing
						 * array used to decode
						 * symbol width and height
						 * Huffman codes */
    TCTundraDecSelectorArray
			*pUseSel = NULL;	/* ptr to struct containing
						 * array used to decode
						 * symbol use count
						 * Huffman codes */
    

	/* First, check the pointers */
    if (pDictBuf == NULL)
	return ia_invalidParm;

	/* If we're not reading the symbols for decompression
	 * return an error temporarily.  Someday, I'll fix this
	 * by finding the runs for each symbol inside the bitmaps
	 * after creating the symbols.
	 */
    if (!forDecompression)
	return ia_noImpl;

	/* Set the flag saying that a bunch of symbols that aren't good
	 * for further compression will be in the internal data
	 * structures.
	 */
    pTextCompInfo->encodeState |= cTCSymsHaveBeenEncoded;

	/* If the client has already called i_convertSymbolsToImage and
	 * is now calling this routine, we're suffering from a severe
	 * case of cognitive dissonance.  The routines for the
	 * internal dictionary and position info aren't supposed to
	 * mix with the routines for the archive (Tundra) dictionary
	 * and position info.
	 */
    if (pTextCompInfo->pBitmapBlock != NULL)
	return ia_noImpl;

	/* If there's an existing symbol table (associates symbol IDs with
	 * ptrs to symbols), destroy it.  We're about to change the
	 * symbols a lot.
	 */
    __error__ = ip_destroyTCSymbolTable (pIPState, pTextCompInfo);
    if (__error__ != ia_successful)
	NORMAL_RETURN;


	/* if we haven't already allocated starter blocks for the
	 * various data structures, do so now.
	 */
    if (pTextCompInfo->p1stScanLineBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stScanLineBlock),
					 sizeof(TCScanLineRun),
					 cTCScanLineRunBlockSize,
					 cTrue,		/* adjust ptrs after
							 * garbage collection */
					 &(pTextCompInfo->fixup),
					 &ip_fixupScanLinePtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

    if (pTextCompInfo->p1stGroupBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stGroupBlock),
					 sizeof(TCGroupRun),
					 cTCGroupRunBlockSize,
					 cTrue,
					 &(pTextCompInfo->fixup),
					 &ip_fixupTCGroupRunPtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

    if (pTextCompInfo->p1stSymbolBlock == NULL)
    {
	__error__ = i_allocateMemoryBlock (pIPState,
					 &(pTextCompInfo->p1stSymbolBlock),
					 sizeof(TCSymbol),
					 cTCSymbolBlockSize,
					 cTrue,
					 &(pTextCompInfo->fixup),
					 &ip_fixupTCSymbolPtrs,
					 pTextCompInfo->ipcAllocator,
					 pTextCompInfo->freer);
	if (__error__ != ia_successful)	/* if it didn't work, just quit */
	    NORMAL_RETURN;
    }

	/* initialize the instance structure we use to pass info to
	 * ip_exactSymbolIsComplete.  A default of 0 is safe for the stuff
	 * we really haven't computed yet.
	 */
    instance.minWidth = 0;
    instance.maxWidth = 0;
    instance.minHeight = 0;
    instance.maxHeight = 0;
    instance.x = 0;
    instance.y = 0;
    instance.w = 0;
    instance.h = 0;
    instance.leftProfile = 0;
    instance.rightProfile = 0;
    instance.strokeWidth = 0;
    instance.maxDimension = 0;
    instance.mass = 0;
    instance.numRuns = 0;
    instance.checkHVolume = cFalse;
    instance.checkVVolume = cFalse;
    instance.hVolume = 0;
    instance.vVolume = 0;
    instance.slant = 0;
    instance.hSAsymmetry = 0;
    instance.vSAsymmetry = 0;
    instance.pCurSymbol = NULL;
    instance.pHorizRuns = NULL;
    instance.pDupHorizRuns = NULL;
    instance.pVertRuns = NULL;
    instance.pDupVertRuns = NULL;

	/* Set the dictionary number to be assigned to all symbols
	 * being read.
	 */
    __error__ = i_setDictionaryNumber (pIPState, pTextCompInfoIn, dictNumber);
    if (__error__ != ia_successful)
	goto cleanupReadTundraDict;

	/* The process of decoding Tundra symbols is pretty easy.
	 * The definition of a Tundra dictionary block is in
	 * William Rucklidge's spec, "Xerox Proposal for JBIG-2
	 * Coding".
	 *
	 * A dictionary begins with the number of symbols in the
	 * dictionary (0 => dictionary clearing code, which we
	 * don't handle now).
	 *
	 * The next entry is a 1-bit flag saying which use count
	 * table to use.  If table 0 is used, there is no residual
	 * image and counts of 0 should be converted to counts of 1.
	 *
	 * Following this is a sequence of height classes.  Each
	 * height class contains symbols having the same height.
	 * At the beginning of each height class is a code giving
	 * (height of this class - height of previous class - 1).
	 * For the first class, "height of previous class" is set
	 * to 0.  Following this is a sequence of pairs containing
	 * (width of current sym - width of previous sym), (use count
	 * of current sym).  The width of the symbol before the
	 * first sym is set to 0.  The use count is the number of
	 * times the symbol is used (theoretically before the next
	 * dictionary).  That's used to generate huffman codes for
	 * position information.
	 * For each <width, use count> pair, add a symbol to the
	 * internal list of symbols.  We'll add pointers to the symbol
	 * bitmaps later.  We do add the useCount and xOffset within the
	 * not-yet-created bitmap (equals total width of all previous
	 * symbols in this height class) to each symbol as they're
	 * created.
	 *
	 * After the last <deltaWidth, useCount> pair, we get a
	 * count of the number of bytes used to hold the compressed
	 * bitmaps of all symbols in this class.  We use the
	 * total width and height of the class to allocate an uncompressed
	 * bitmap (rounding the bytesPerLine to next multiple of 4).
	 * The bitmap is inserted into the pointer fixup chain so we'll
	 * adjust symbol pointers if it moves.
	 * We then decompress the data into the new bitmap using the
	 * MMR decompression routines.  To do this, we have to allocate
	 * a CCITTDecompression structure with the proper width and
	 * initialize it.  If we've already got a CCITTDecompression
	 * structure and it's big enough, we just initialize it without
	 * throwing it away and getting a new one.
	 *
	 * After the bitmap has been created, we go back through all
	 * symbols in this height class, setting the pBitmap and
	 * bytesPerLine entries in each one to correspond to the
	 * new bitmap.
	 *
	 * The data for the compressed bitmaps starts on the byte
	 * boundary after the encoded symbol count.  The next
	 * height class starts on the byte boundary after the compressed
	 * data (they're not concatenated).
	 *
	 * At the end of the last symbol count is an END code instead
	 * of the next delta height.  The size of the extension block
	 * follows that, but we don't use extension blocks now, so we
	 * ignore that.
	 */

	/* Prepare to start reading stuff in */
    accum = 0;				/* bit accumulator is empty */
    vacantBits = 32;
    pLocEnd = pLocIn + bufSize;		/* used to detect end of buffer */

	/* point to the only decoder available for the width and
	 * height values.
	 */
    pWidthHeightDecoder = &widthHeightDecTable0;

	/* Go read (and discard) the number of symbols in the dictionary. */
    __error__ = ip_decodeTundraInteger (pIPState, &accum, &vacantBits, &pLocIn,
				      pLocEnd, (UInt32 *) &temp);
    if (__error__ != ia_successful)
	goto cleanupReadTundraDict;

	/* If we need to read more bits to get the 1-bit use count table
	 * selector flag, do it.
	 */
    if (vacantBits >= (32-16))
    {
	if (pLocIn < pLocEnd)	/* can read 2 bytes */
	{
	    vacantBits -= 8;
	    accum |= (UInt32) pLocIn[0] << vacantBits;
	    vacantBits -= 8;
	    accum |= (UInt32) pLocIn[1] << vacantBits;
	    pLocIn += 2;
	}
	else
	if (pLocIn == pLocEnd)	/* can read 1 byte */
	{
	    vacantBits -= 8;
	    accum |= (UInt32) pLocIn[0] << vacantBits;
	    pLocIn += 1;
	}
	else	/* can't read any bytes.  Are we empty now? */
	if (vacantBits > (32 - cTundraMinExtensionLenCodeLen))
	{
	    __error__ = ia_bufferExceeded;
	    goto cleanupReadTundraDict;
	}
    }

	/* Look at the use count encoding table flag before using
	 * it to get a use count decoding table.  If this flag
	 * indicates that we're going to use the table that allows
	 * use counts of 0 and 1, clear the flag in our central state
	 * that says that there is a residual image.  We need to
	 * know that later on when generating Huffman codes.
	 */
    temp = accum >> 31;
    accum <<= 1;
    vacantBits += 1;

    if (temp == 0)	/* use table that allows use counts of 0 and 1 */
	pTextCompInfo->useResidual = cFalse;

	/* Initialize the use count decoder */
    pUseDecoder = useDecTableArray[temp];

	/* We need to make Huffman decoder tables for the width/height
	 * and symbol use count headers.  We use the tables to
	 * identify which encoder class to use to get the header
	 * and value lengths and the value offset.
	 * To make a Huffman decoder table, we have to fill in
	 * numItemsArray with the number of classes having specified
	 * header lengths.
	 */
    memset(numItemsArray, 0, 32*sizeof(UInt32));

    for (temp = pWidthHeightDecoder->bigPosNumClass; temp >= 0; temp--)
	numItemsArray[(pWidthHeightDecoder->pClassArray)[temp].bitHeaderLen]++;

    __error__ = ip_calcSizeOfTundraDecodeTable (pIPState, &numItemsArray[0],
					        (Int32 *) &temp);
    if (__error__ != ia_successful)
	goto cleanupReadTundraDict;

	/* Allocate the selector array structure.  Make sure all
	 * pointers are initally NULL. */
    pWidthHeightSel = (TCTundraDecSelectorArray *)
		CALLOC (1, sizeof(TCTundraDecSelectorArray) +
			   temp * sizeof(TCTundraDecSelector));
    if (pWidthHeightSel == NULL)
    {
	__error__ = ia_nomem;
	goto cleanupReadTundraDict;
    }
    pWidthHeightSel->arraySize = temp;
    pWidthHeightSel->nextTable = 256;

	/* Go through the encoder classes, inserting their headers into
	 * the decoder array.
	 */
    for (temp = pWidthHeightDecoder->bigPosNumClass; temp >= 0; temp--)
    {
	__error__ = ip_putValIntoTundraDecodeTable(
			pIPState,
			pWidthHeightSel,
			(pWidthHeightDecoder->pClassArray)[temp].bitHeader,
			(pWidthHeightDecoder->pClassArray)[temp].bitHeaderLen,
			(void *) &((pWidthHeightDecoder->pClassArray)[temp]));
	if (__error__ != ia_successful)
	    goto cleanupReadTundraDict;
    }

	/* Repeat this process for the Use Count tables */
    memset(numItemsArray, 0, 32*sizeof(UInt32));

    for (temp = pUseDecoder->bigPosNumClass; temp >= 0; temp--)
	numItemsArray[(pUseDecoder->pClassArray)[temp].bitHeaderLen]++;

    __error__ = ip_calcSizeOfTundraDecodeTable (pIPState, &numItemsArray[0],
					        (Int32 *) &temp);
    if (__error__ != ia_successful)
	goto cleanupReadTundraDict;

	/* Allocate the selector array structure.  Make sure all
	 * pointers are initally NULL. */
    pUseSel = (TCTundraDecSelectorArray *)
		CALLOC (1, sizeof(TCTundraDecSelectorArray) +
			   temp * sizeof(TCTundraDecSelector));
    if (pUseSel == NULL)
    {
	__error__ = ia_nomem;
	goto cleanupReadTundraDict;
    }
    pUseSel->arraySize = temp;
    pUseSel->nextTable = 256;

	/* Go through the encoder classes, inserting their headers into
	 * the decoder array.
	 */
    for (temp = pUseDecoder->bigPosNumClass; temp >= 0; temp--)
    {
	__error__ = ip_putValIntoTundraDecodeTable(
			pIPState,
			pUseSel,
			(pUseDecoder->pClassArray)[temp].bitHeader,
			(pUseDecoder->pClassArray)[temp].bitHeaderLen,
			(void *) &((pUseDecoder->pClassArray)[temp]));
	if (__error__ != ia_successful)
	    goto cleanupReadTundraDict;
    }


	/************************** DECODE LOOP ************************/

	/* Here's the loop that goes through the height classes */
    instance.h = 0;
    ccittInfoSize = 0;
    while (cTrue)
    {
	    /* Fill up the accumulator if necessary.  The
	     * longest height code is 7 bits, so the
	     * canonical 16 will be fine (and more efficient
	     * that just fetching another 8).
	     */
	if (vacantBits > (32-16))
	{
	    if (pLocIn < pLocEnd)	/* can read 2 bytes */
	    {
		vacantBits -= 8;
		accum |= (UInt32) pLocIn[0] << vacantBits;
		vacantBits -= 8;
		accum |= (UInt32) pLocIn[1] << vacantBits;
		pLocIn += 2;
	    }
	    else
	    if (pLocIn == pLocEnd)	/* can read 1 byte */
	    {
		vacantBits -= 8;
		accum |= (UInt32) pLocIn[0] << vacantBits;
		pLocIn += 1;
	    }
	    else	/* can't read any bytes.  Are we empty now? */
	    if (vacantBits > (32 - cTundraMinExtensionLenCodeLen))
	    {
		__error__ = ia_bufferExceeded;
		goto cleanupReadTundraDict;
	    }
	}

	    /* Go read the delta Height for this class.  First, see
	     * if this is the end code.
	     */
	if ((accum & cTundraWidthHeightEndCodeMask) ==
	    (UInt32)(cTundraWidthHeightEndCode <<
					(32 - cTundraWidthHeightEndCodeLen)))
	{
	    accum <<= cTundraWidthHeightEndCodeLen;
	    vacantBits += cTundraWidthHeightEndCodeLen;
	    break;	/* there are no more height classes. */
	}

	    /* It's a real height class, decode the delta height */
	if ((accum & pWidthHeightDecoder->headerMask) ==
					pWidthHeightDecoder->headerRoot)
	{
	    __error__ = ip_decodeTundraBigVal (pIPState, pWidthHeightDecoder,
					     &accum, &vacantBits,
					     &pLocIn, pLocEnd,
					     (UInt32 *) &temp);
	    if (__error__ != ia_successful)
		goto cleanupReadTundraDict;
	}
	else
	{
		/* Since the longest width/height header is 4 bits, we can
		 * decode it in a single table reference.  No need
		 * to follow lower levels of tables.
		 */
	    pEncClass = (TCTundraEncClass *)
			    (pWidthHeightSel->array[accum >> (32-8)].pEntry);
	    accum <<= pEncClass->bitHeaderLen;

		/* Since the longest bounded width/height code is 7 bits
		 * and we had at least 16 bits, just decode the value now
		 * too.
		 */
	    if (pEncClass->encValLen > 0)
		temp = (accum >> (32 - pEncClass->encValLen)) +
							pEncClass->offset;
	    else
		temp = pEncClass->offset;
	    accum <<= pEncClass->encValLen;
	    vacantBits += pEncClass->encValLen + pEncClass->bitHeaderLen;
	}
	instance.h += temp + 1;			/* we encoded deltaH - 1 */
	maxWidth = 0;				/* current bitmap width */
	instance.w = 0;				/* last symbol width */
	p1stHeightSym = NULL;

	    /* In this loop, we go find all the <deltaW, UseCount>
	     * pairs for this height class.
	     */
	while (cTrue)
	{
		/* fill the accumulator before getting the
		 * next width */
	    if (vacantBits > (32-16))
	    {
		if (pLocIn < pLocEnd)	/* can read 2 bytes */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocIn[0] << vacantBits;
		    vacantBits -= 8;
		    accum |= (UInt32) pLocIn[1] << vacantBits;
		    pLocIn += 2;
		}
		else
		if (pLocIn == pLocEnd)	/* can read 1 byte */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocIn[0] << vacantBits;
		    pLocIn += 1;
		}
		else	/* can't read any bytes.  Are we empty now? */
		if (vacantBits > (32 - cTundraMinExtensionLenCodeLen))
		{
		    __error__ = ia_bufferExceeded;
		    goto cleanupReadTundraDict;
		}
	    }

		/* Read the delta width code.  If it's equal to
		 * the end code, leave this loop.
		 */
	    if ((accum & cTundraWidthHeightEndCodeMask) ==
		(UInt32)(cTundraWidthHeightEndCode <<
					(32 - cTundraWidthHeightEndCodeLen)))
		{
		    accum <<= cTundraWidthHeightEndCodeLen;
		    vacantBits += cTundraWidthHeightEndCodeLen;
		    break;	/* end of this height class. */
		}

		/* It's a real symbol, decode the delta width */
	    if ((accum & pWidthHeightDecoder->headerMask) ==
					    pWidthHeightDecoder->headerRoot)
	    {
		__error__ = ip_decodeTundraBigVal (pIPState, pWidthHeightDecoder,
						 &accum, &vacantBits,
						 &pLocIn, pLocEnd,
						 (UInt32 *) &temp);
		if (__error__ != ia_successful)
		    goto cleanupReadTundraDict;
	    }
	    else
	    {
		    /* Since the longest width/height header is 4 bits, we can
		     * decode it in a single table reference.  No need
		     * to follow lower levels of tables.
		     */
		pEncClass = (TCTundraEncClass *)
			    (pWidthHeightSel->array[accum >> (32-8)].pEntry);
		accum <<= pEncClass->bitHeaderLen;
    
		    /* Since the longest bounded width/height code is 7 bits
		     * and we had at least 16 bits, just decode the value now
		     * too.
		     */
		if (pEncClass->encValLen > 0)
		    temp = (accum >> (32 - pEncClass->encValLen)) +
							    pEncClass->offset;
		else
		    temp = pEncClass->offset;
		accum <<= pEncClass->encValLen;
		vacantBits += pEncClass->encValLen + pEncClass->bitHeaderLen;
	    }
		/* Set the width of the current symbol */
	    instance.w += temp;

		/* refill the accumulator before getting the
		 * use count */
	    if (vacantBits > (32-16))
	    {
		if (pLocIn < pLocEnd)	/* can read 2 bytes */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocIn[0] << vacantBits;
		    vacantBits -= 8;
		    accum |= (UInt32) pLocIn[1] << vacantBits;
		    pLocIn += 2;
		}
		else
		if (pLocIn == pLocEnd)	/* can read 1 byte */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocIn[0] << vacantBits;
		    pLocIn += 1;
		}
		else	/* can't read any bytes.  Are we empty now? */
		if (vacantBits > (32 - cTundraMinExtensionLenCodeLen))
		{
		    __error__ = ia_bufferExceeded;
		    goto cleanupReadTundraDict;
		}
	    }

		/* And get the use count */
	    if ((accum & pUseDecoder->headerMask) ==
					    pUseDecoder->headerRoot)
	    {
		__error__ = ip_decodeTundraBigVal (pIPState, pUseDecoder,
						 &accum, &vacantBits,
						 &pLocIn, pLocEnd,
						 (UInt32 *) &temp);
		if (__error__ != ia_successful)
		    goto cleanupReadTundraDict;
	    }
	    else
	    {
		    /* Since the longest use count header is 5 bits, we can
		     * decode it in a single table reference.  No need
		     * to follow lower levels of tables.
		     */
		pEncClass = (TCTundraEncClass *)
				(pUseSel->array[accum >> (32-8)].pEntry);
		accum <<= pEncClass->bitHeaderLen;
    
		    /* Since the longest bounded use count code is 10 bits
		     * and we had at least 16 bits, just decode the value now
		     * too.
		     */
		if (pEncClass->encValLen > 0)
		    temp = (accum >> (32 - pEncClass->encValLen)) +
							    pEncClass->offset;
		else
		    temp = pEncClass->offset;
		accum <<= pEncClass->encValLen;
		vacantBits += pEncClass->encValLen + pEncClass->bitHeaderLen;
	    }

#ifdef DDDEBUG
	    printf("%dx%d count = %d\n", instance.w, instance.h, temp);
#endif

		/* A use count of 0 really is a hint that the use count
		 * is 1 but that the encoder promises that the symbol's
		 * bitmap can be discarded immediately after use.
		 * We don't take advantage of that hint since we
		 * assume residue images (contain all things that only
		 * appear once) are used.  But if someone doesn't
		 * use a residue, fix the 0 use counts.
		 */
	    if (temp == 0)
		temp = 1;

		/* We're ready to insert this symbol into the list of symbols
		 * Note that we don't supply any runs with the symbol.
		 * So far as I can tell, this is ok unless we try to
		 * do further compression with this symbol.  There's code in
		 * i_convertImageToSymbols to delete all symbols that
		 * don't have runs before starting to decode anything.
		 */
	    __error__ = ip_exactSymbolIsComplete (pIPState, pTextCompInfo,
					        NULL,
					        &instance,
					        cTrue);	/* just insert
							 * this symbol into
							 * the list */
	    if (__error__ != ia_successful)
		goto cleanupReadTundraDict;

		/* Set the xOffset and use count of this symbol.  Since
		 * we've just inserted it, it is the last symbol, hence
		 * is pointed to by pTextCompInfo->pLastSymbol.
		 */
	    pTextCompInfo->pLastSymbol->instanceCount = temp;
	    pTextCompInfo->pLastSymbol->xOffset = maxWidth;

		/* increase maxWidth to account for this symbol */
	    maxWidth += instance.w;

		/* If this is the first symbol of the height class,
		 * remember it since we'll have to get back here
		 * to set the pBitmap and bytePerLine entries.
		 */
	    if (p1stHeightSym == NULL)
		p1stHeightSym = pTextCompInfo->pLastSymbol;
	}	/* end of loop that creates symbol structs for a height class */

	    /* At this point, we've just finished reading the end code
	     * for the current height class.  We want to get the
	     * length of the compressed bitmap coming up next.
	     */
	__error__ = ip_decodeTundraInteger (pIPState, &accum,
					  &vacantBits, &pLocIn,
					  pLocEnd, (UInt32 *) &compBitmapSize);
	if (__error__ != ia_successful)
	    goto cleanupReadTundraDict;

	    /* Point to the beginning of the compressed bitmap.  Any
	     * bits that are now in the accumulator belong to some
	     * combination of padding before the start of the
	     * compressed bitmap and the compressed bitmap itself.
	     * We know that bits were put into accum in units of
	     * 8, so the last bit in accum was originally the LS
	     * bit of a byte.
	     * If we repeatedly back pLocIn up and remove 8 bits from
	     * the accumulator for each step until there are fewer
	     * than 8 bits left, we'll be pointing to
	     * the start of the compressed bitmap.
	     */
	while (vacantBits <= (32 - 8))
	{
	    pLocIn--;
	    vacantBits += 8;
	}

	    /* Allocate the bitmap for this height class.  First, compute
	     * the bytes/line.  We can't change maxWidth at this point
	     * since it will be needed to decompress the bitmap.
	     * Since one should always be concerned for deallocating
	     * anything that's allocated, how do we deallocate this
	     * bitmap?  The only pointers to it will be in one or more
	     * symbols.  When a symbol is freed, we see if any other
	     * symbols point to the symbol's bitmap.  If not, the bitmap
	     * is freed.  When the entire text compression info is freed,
	     * we go through all the symbols, freeing any bitmaps
	     * they may have.
	     */
	bytesPerLine = ((maxWidth + 31)/32) * 4;
	decompBitmapSize = sizeof(PointerFixup) + instance.h * bytesPerLine;

	pBitmap = pTextCompInfo->ipcAllocator(pIPState, decompBitmapSize, cTrue);
	if (pBitmap == NULL)
	{
	    __error__ = ia_nomem;
	    goto cleanupReadTundraDict;
	}

	    /* clear the bitmap so we don't have to carefully write
	     * every word with white stuff when expanding the symbols.
	     * We do this first because it also clears the PointerFixup
	     * structure at the top of the bitmap.  We will initialize
	     * that structure next.
	     */
	memset((void *)pBitmap, 0, decompBitmapSize);

	__error__ = ip_initPointerFixup (pIPState, (PointerFixup *) pBitmap,
				         decompBitmapSize,
				         &(pTextCompInfo->fixup),
				         NULL); /* no pointer fixup routine */
	if (__error__ != ia_successful)
	{
	    (pTextCompInfo->freer)(pIPState, pTextCompInfo->pBitmapBlock,
				   decompBitmapSize);
	    goto cleanupReadTundraDict;
	}

	    /* Advance the bitmap pointer beyond the PointerFixup
	     * structure.  Subtract the fake frame.
	     * i_decompressRowsMMR will add it back on.
	     */
	pBitmap += sizeof(PointerFixup);

	    /* See if the bitmap is compressed or not.  */
	if (compBitmapSize == 0)
	{
	    UInt32	 row, rowWidthInBytes, colByte;
	    UInt32	 word = 0;
	    UInt32	 mask = leftMask[(maxWidth - 1) & 0x1F];
	    UInt8	*pLocBitmap = pBitmap;

		/* The uncompressed bitmap begins at pBitmap.  It
		 * has instance.h rows, each (maxWidth + 7)/8 bytes
		 * long.  Read each row, mask off the pad bits and
		 * write the words into the uncompressed bitmap.
		 * First, compute the mask.  (maxWidth - 1) & 0x1F
		 * maps [0, 1, ..., 31] into [31, 0, 1, ..., 30],
		 * which gives us a mask with (maxWidth MOD 32)
		 * bits on the left except when maxWidth MOD 32
		 * equals 0, in which case you get all 1's. The
		 * nice thing about this mask is that you can
		 * ALWAYS use it on the last word of a scan line.
		 */
	    rowWidthInBytes = (maxWidth + 7) / 8;
	    if (pLocIn + instance.h * rowWidthInBytes > pLocEnd)
	    {
		/* Since the data has been refereced via pointer fixup,
		 * make an explicit reference to it, so it will be freed
		 * properly */
		while (p1stHeightSym != NULL)
		{
		    p1stHeightSym->pBitmap = pBitmap;
		    p1stHeightSym->bytesPerLine = bytesPerLine;
		    p1stHeightSym = p1stHeightSym->pNextNum;
		}
		__error__ = ia_bufferExceeded;
		goto cleanupReadTundraDict;
	    }

	    for (row = instance.h;
		 row > 0;
		 row--, pLocBitmap += bytesPerLine, pLocIn += rowWidthInBytes)
	    {
		for (colByte = 0; colByte < rowWidthInBytes; colByte += 4)
		{
		    word = (UInt32)(pLocIn[colByte] << 8) |
			   (UInt32) pLocIn[colByte + 1];
		    word <<= 16;
		    word |= (UInt32) (pLocIn[colByte + 2] << 8) |
			    (UInt32) (pLocIn[colByte + 3]);
		    ULONG_ACCESS(&(pLocBitmap[colByte])) = word;
		}
		    /* mask off bits in the last word */
		ULONG_ACCESS(&(pLocBitmap[colByte - 4])) = word & mask;
	    }
		/* Since every scan line, including the last one, has
		 * rowWidthInBytes in it and pLocIn has just been
		 * advanced beyond it, we're pointing to the start
		 * of the next height class.  When we add compBitmapSize
		 * to pLocIn below, pLocIn won't change since compBitmapSize
		 * is 0 at this point.
		 */
	}
	else	/* we have a real compressed bitmap, the normal case */
	{
		/* Add a fake frame to the bitmap for the i_ routines */
	    pBitmap -= cFrameBytes;
	    pBitmap -= bytesPerLine << cLogFrameBits;

		/* Create a CCITTDecompression structure for this bitmap
		 * if necessary.
		 */
	    __error__ = i_getSizeOfCCITTDecompressionInfo (pIPState,
							 maxWidth +
								2*cFrameBits,
							 (Int32 *) &temp);
	    if (__error__ != ia_successful)
		goto cleanupReadTundraDict;
    
		/* Is the CCITTDecompression struct we've already got
		 * (if any) big enough?  If not, free it and get another.
		 */
	    if ((pCCITTDecompInfo == NULL) ||
		(temp > ccittInfoSize))
	    {
		if (pCCITTDecompInfo != NULL)
		    FREE(pCCITTDecompInfo);
		ccittInfoSize = temp;
		pCCITTDecompInfo = (void *)MALLOC(ccittInfoSize);
		if (pCCITTDecompInfo == NULL)
		{
		    __error__ = ia_nomem;
		    goto cleanupReadTundraDict;
		}
	    }
		/* Initialize the CCITT decompression info, old or new */
	    __error__ = i_initializeCCITTDecompressionInfo (pIPState,
							  pCCITTDecompInfo,
							  maxWidth +
								2*cFrameBits,
							  cFalse,
							  cNoPad,
							  ccittInfoSize);
	    if (__error__ != ia_successful)
		goto cleanupReadTundraDict;
    
		/* Decompress the compressed bitmap into this block */
	    __error__ = i_decompressRowsMMR (pIPState, pCCITTDecompInfo,
					   pBitmap,	/* ptr to bitmap */
					   cFrameBits,	/* start line */
					   instance.h,	/* numLines */
					   bytesPerLine,	/* bpl */
					   cFalse,	/* combineResults */
					   pLocIn,	/* pCompData */
					   compBitmapSize, /* compDataBufSize */
					   cFalse,	/* continueImageBlock */
					   cFalse,	/* continueCompData */
					   cTrue,	/* flushPipeline */
					   &exhaustedImageData,
					   &exhaustedCompData,
					   &percentDone,
					   &temp);	/* bytesRead */
    
	    if (__error__ != ia_successful)
		goto cleanupReadTundraDict;
	    if (!exhaustedImageData)
	    {
		__error__ = ia_internal;
		goto cleanupReadTundraDict;
	    }
	    if (temp != compBitmapSize)
	    {
		__error__ = ia_internal;
		goto cleanupReadTundraDict;
	    }
		/* Take the fake frame back off pBitmap so we can use
		 * it to initialize the symbols.
		 */
	    pBitmap += (bytesPerLine << cLogFrameBits) + cFrameBytes;
	}
	    /* Hooray, we've decompressed the bitmap for this height
	     * class.  Go through the symbols again and set the
	     * pBitmap and bytesPerLine entries in each.  But first,
	     * get rid of the fake frame on pBitmap.
	     */
	while (p1stHeightSym != NULL)
	{
	    p1stHeightSym->pBitmap = pBitmap;
	    p1stHeightSym->bytesPerLine = bytesPerLine;
	    p1stHeightSym = p1stHeightSym->pNextNum;
	}

	    /* pLocIn + compBitmapSize is now pointing to the beginning
	     * of the next height class.  Clear out the accumulator so
	     * we can start decoding it.
	     */
	pLocIn += compBitmapSize;
	accum = 0;
	vacantBits = 32;
    }

	/* At this point, we've read all the height classes.  There's
	 * a extension code length following, but we don't read extensions,
	 * so toss it.  Bye!
	 */
    __error__ = ia_successful;

cleanupReadTundraDict:
    if (pWidthHeightSel != NULL)
	FREE(pWidthHeightSel);
    if (pUseSel != NULL)
	FREE(pUseSel);
    if (pCCITTDecompInfo != NULL)
	FREE(pCCITTDecompInfo);

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_readTundraPositionBlock()
 *
 *	This procedure decompresses a Tundra position block, producing
 *	a bitmap.
 *
 *	NOTE: for the time being, there is no provision to decode the
 *	page using only a subset of the dictionaries in the internal
 *	compression info.  We assume symbols from all dictionaries can
 *	be used.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pPositionBuf:		pointer to buffer containing
 *					ENTIRE Tundra position block.
 *					This code has no provision for
 *					concatenating partial position blocks.
 *		bufSize:		number of bytes available to
 *					read.
 *		pImage:			Pointer to upper left corner of 
 *					binary image INCLUDING frame.
 *		wS:			width of image including frame
 *		hS:			height of image including frame
 *		bpl:			The image's bytes/line
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	field is bigger than room available.
 *		ia_nomem		couldn't get enough memory for
 *					internal data structures.
 *
 **************************************************************************/
Int32  CDECL 
i_readTundraPositionBlock (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pPositionBuf,
    UInt32	bufSize,
    UInt8      *pImage,
    Int32	wS,
    Int32	hS,
    UInt32	bpl)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo;
    UInt8		*pLocalBuf;		/* local position buf ptr */
    UInt8		*pLocEnd;		/* ptr to end of buf */
    Int32		 curX;			/* current X value */
    Int32		 lastFirstX;		/* X offset of first symbol
						 * in previous line */
    Int32		 curY;			/* current Y value */
    Int32		 yPos;			/* intra-strip Y position */
    Int32		 dX;			/* amount by which X changes */
    Int32		 dY;			/* amount by which Y changes */
    TCSymbol		*pSymbol;		/* current symbol */
    Int32		 numSymbolsExpected;	/* number of symbols the
						 * encoder said were here. */
    Int32		 numSymbolsFound;	/* number of symbols we
						 * actually found. */
    Int32		 temp;
    Int32		 numItemsArray[32];	/* number of symbols having
						 * a particular Huffman code
						 * length. */
    UInt32		 accum;			/* bit accumulator */
    Int32		 vacantBits;		/* number of vacant bits on
						 * right end of accum */
    Int32		 stripHeight;		/* height of each strip */
    UInt32		 logStripHeight;	/* avoid multiplies */
    Int32		 modalDeltaX;		/* the most common value
						 * for deltaX on this page */
    Bool		 useTransposition;	/* if true, transpose the
						 * X and Y coords because that's
						 * the way the encoder used
						 * them.  This is for encoding
						 * vertical text. */
    TCTundraDecTable	*pFirstXDecoder;	/* ptr to first X decoder */
    TCTundraDecTable	*pDeltaXDecoder;	/* ptr to delta X decoder */
    TCTundraDecTable	*pDeltaYDecoder;	/* ptr to delta Y decoder */
    TCTundraEncClass	*pEncClass;		/* scr encoding class ptr */
    TCTundraDecSelectorArray
			*pSymSel = NULL;	/* ptr to struct containing
						 * array used to decode
						 * symbol Huffman codes */
    TCTundraDecSelectorArray
			*pFirstXSel = NULL;	/* ptr to struct containing
						 * array used to decode
						 * first X Huffman codes */
    TCTundraDecSelectorArray
			*pDeltaXSel = NULL;	/* ptr to struct containing
						 * array used to decode
						 * delta X Huffman codes */
    TCTundraDecSelectorArray
			*pDeltaYSel = NULL;	/* ptr to struct containing
						 * array used to decode
						 * delta Y Huffman codes */
    TCTundraDecSelector *pDecSelector;		/* scratch selector used
						 * in Huffman decodes. */
#if 0
	/* This is a hack that we use so we can call i_convertSymbolsToImage.
	 * That routine does nothing except create cached bitmaps for the
	 * symbols, then find that the fake position buffer
	 * say "EndOfFile"
	 */
    UInt8		 fakePosBuf[cTCPosHeaderVer2Len + 1];
#endif
    

	/* initialize test compression info pointer */
    pTextCompInfo = (TCCompressionInfo *)pTextCompInfoIn;

	/* check the pointers */
    if ((pPositionBuf == NULL) || (pImage == NULL))
	return ia_invalidParm;

	/* Fix the image pointer and dimensions */
    hS -= 2 * cFrameBits;
    wS -= 2 * cFrameBits;
    pImage += cFrameBytes + (bpl << cLogFrameBits);

#if 0
	/* Initialize the fake position buffer array.  This will be
	 * removed when we replace our old dictionary serialization
	 * with Tundra dictionary initialization.
	 */
    fakePosBuf[0] = (UInt8) (cTextCompPositionInfoMagicNum);
    fakePosBuf[1] = (UInt8) (cTextCompPositionInfoMagicNum >> 8);
    fakePosBuf[2] = (UInt8) (cTextCompPositionInfoMagicNum >> 16);
    fakePosBuf[3] = (UInt8) (cTextCompPositionInfoMagicNum >> 24);

    fakePosBuf[4] = (UInt8) (cTextCompPositionInfoVersionNum);
    fakePosBuf[5] = (UInt8) (cTextCompPositionInfoVersionNum >> 8);

	/* this value tells i_convertSymbolsToImage to stop.
	 * This happens before any blitting of symbols.
	 */
    fakePosBuf[cTCPosHeaderVer2Len] = 0;		/* delta y of 0 */

	/* go create the cached bitmaps.  This will also go when
	 * we convert to the real Tundra dictionary.  We don't
	 * pass pointers to the real image to make sure this
	 * routine can't scribble in it.
	 */
    __error__ = i_convertSymbolsToImage (pIPState, (UInt8 *)1, 0, 0, 0,
				       pTextCompInfo, &fakePosBuf[0]);
#endif


	/* Assign Huffman codes to all the symbols.  We do this
	 * on each page since it's pretty fast and in the future
	 * each page will have a unique set of dictionaries.
	 */
    ERR_PASS_CLEANUP( ip_genSymbolHuffmanCodes (pIPState, pTextCompInfo,
					        &(numItemsArray[0]), 0, cTrue));

	/* Create the decoder tables for the symbols' Huffman codes.
	 * To do this, we first need to get the size of these
	 * tables.  The routine that calculates the size of the
	 * decoder array needs the code size histogram (number
	 * of symbols having each possible code size) returned
	 * by ip_genSymbolHuffmanCodes above.
	 */
    ERR_PASS_CLEANUP(ip_calcSizeOfTundraDecodeTable (pIPState,
						     &(numItemsArray[0]),
						     (Int32 *)&temp));

	/* Allocate the selector array structure.  We use CALLOC
	 * to make sure all the pointers start out NULL.
	 */
    pSymSel = (TCTundraDecSelectorArray *)
		CALLOC (1, sizeof(TCTundraDecSelectorArray) +
			   temp * sizeof(TCTundraDecSelector));
    ERR_GEN_CLEANUP (pSymSel == NULL, ia_nomem);

    pSymSel->arraySize = temp;
    pSymSel->nextTable = 256;

	/* Loop through all the symbols, inserting their Huffman codes
	 * into the symbol decoder array.
	 */
    for (pSymbol = pTextCompInfo->p1stSymbol;
	 pSymbol != NULL;
	 pSymbol = pSymbol->pNextNum)
    {
	ERR_PASS_CLEANUP(
		ip_putValIntoTundraDecodeTable (pIPState, pSymSel,
						pSymbol->huffmanCode,
						pSymbol->huffmanCodeLen,
						(void *) pSymbol));
    }

#ifdef DDDEBUG
	 {
	     TCSymbol *pSym = pTextCompInfo->p1stSymbol;

	     while (pSym) {
		 printf("%3d (%3d): dict %d, %dx%d, count %d, code %d-%d\n",
		     pSym->dictOrder, pSym->symbolID, pSym->dictNum, pSym->width,
		     pSym->height, pSym->instanceCount, pSym->huffmanCode,
		     pSym->huffmanCodeLen);
		 pSym = pSym->pNextDictSym;
	     }
	 }
#endif

	/* Swell, now we can use the decoder tables to point to symbols.
	 * Go read the position block header.
	 */
    accum = 0;
    vacantBits = 32;
    pLocalBuf = pPositionBuf;
    pLocEnd = pLocalBuf + bufSize - 1;		/* ptr to last valid byte */

	/* Discard the number of symbols on this page. */
    ERR_PASS_CLEANUP(ip_decodeTundraInteger (pIPState, &accum, &vacantBits,
					     &pLocalBuf, pLocEnd,
					     (UInt32 *) &numSymbolsExpected));

#ifdef DDEBUG
    printf("%d Tundra position block entries\n", numSymbolsExpected);
#endif
	/* Parse the next 13 bits of position block header.  This includes
	 * 4 bits of modal deltaX, 2 bits each for strip height,
	 * first X table, delta X table, delta Y table and the 1-bit
	 * transposition flag.
	 */
    if (vacantBits > (32-13))
    {
	if (pLocalBuf >= pLocEnd)	/* can't read 2 bytes */
	{
	    __error__ = ia_bufferExceeded;
	    goto ERR_CLEANUP;
	}

	vacantBits -= 8;
	accum |= (UInt32) pLocalBuf[0] << vacantBits;
	vacantBits -= 8;
	accum |= (UInt32) pLocalBuf[1] << vacantBits;
	pLocalBuf += 2;
    }

	/* get the most common delta X value */
    modalDeltaX = accum >> 28;
    accum <<= 4;

	/* get the strip size */
    stripHeight = accum >> 30;
    accum <<= 2;

	/* only values 0, 1, and 3 are defined */
    ERR_GEN_CLEANUP (stripHeight == 2, ia_decodeFailure);

	/* The first Y position is supposed to be at the start of the
	 * first imaginary position above the image.  The number
	 * in stripHeight now is 0, 1, or 3; signifying strip heights
	 * of 1, 2 or 4.  Fix stripHeight.  We also want logStripHeight
	 * to be set to 0, 1 or 2.
	 */
    if (stripHeight > 2)
	logStripHeight = stripHeight-1;
    else
	logStripHeight = stripHeight;
    stripHeight++;

	/* The first imaginary strip is just before the image. */
    curY = -stripHeight;
    yPos = 0;			/* stays at this value if logStripHeight == 0 */

	/* Get the first X table selector */
    pFirstXDecoder = firstXDecTableArray[accum >> 30];
    accum <<= 2;
    ERR_GEN_CLEANUP (pFirstXDecoder == NULL, ia_decodeFailure);

	/* Get the delta X table selector */
    pDeltaXDecoder = deltaXDecTableArray[accum >> 30];
    accum <<= 2;
    ERR_GEN_CLEANUP (pDeltaXDecoder == NULL, ia_decodeFailure);

	/* Get the delta Y table selector */
    pDeltaYDecoder = deltaYDecTableArray[accum >> 30];
    accum <<= 2;
    ERR_GEN_CLEANUP (pDeltaYDecoder == NULL, ia_decodeFailure);

	/* get the transposition flag */
    if (accum & 0x80000000)
	useTransposition = cTrue;
    else
	useTransposition = cFalse;

    accum <<= 1;	/* discard transposition flag */
    vacantBits += 13;	/* account for all the flags just read */

	/* We need to make Huffman decoder tables for the firstX, deltaX and
	 * deltaY headers.  We use the tables to identify which encoder
	 * class to use to get the header and value lengths and
	 * the value offset.
	 * To make a Huffman decoder table, we have to fill in
	 * numItemsArray with the number of classes having specified
	 * header lengths.
	 */
    memset(numItemsArray, 0, 32*sizeof(UInt32));

    for (temp = pFirstXDecoder->bigPosNumClass - 1; temp > 0; temp--)
	numItemsArray[(pFirstXDecoder->pClassArray)[temp].bitHeaderLen]++;

    ERR_PASS_CLEANUP(ip_calcSizeOfTundraDecodeTable (pIPState,
						     &numItemsArray[0],
						     (Int32 *) &temp));

	/* Allocate the selector array structure.  Make sure all
	 * pointers are initally NULL. */
    pFirstXSel = (TCTundraDecSelectorArray *)
		CALLOC (1, sizeof(TCTundraDecSelectorArray) +
			   temp * sizeof(TCTundraDecSelector));
    ERR_GEN_CLEANUP (pFirstXSel == NULL, ia_nomem);
    pFirstXSel->arraySize = temp;
    pFirstXSel->nextTable = 256;

	/* Go through the encoder classes, inserting their headers into
	 * the decoder array.
	 */
    for (temp = pFirstXDecoder->bigPosNumClass - 1; temp > 0; temp--)
    {
	ERR_PASS_CLEANUP(ip_putValIntoTundraDecodeTable(
			 pIPState,
			 pFirstXSel,
			 (pFirstXDecoder->pClassArray)[temp].bitHeader,
			 (pFirstXDecoder->pClassArray)[temp].bitHeaderLen,
			 (void *) &((pFirstXDecoder->pClassArray)[temp])));
    }

	/* Repeat this process for the delta X tables */
    memset(numItemsArray, 0, 32*sizeof(UInt32));

    for (temp = pDeltaXDecoder->bigPosNumClass - 1; temp > 0; temp--)
	numItemsArray[(pDeltaXDecoder->pClassArray)[temp].bitHeaderLen]++;

	/* Remember that the end code for delta X is of length 2, but
	 * it's not in the tables since there's no value associated
	 * with it.  Anyway, to make the calculations come our right,
	 * remember that some table space is occupied by it.
	 */
    numItemsArray[cTundraStripEndCodeLen]++;

    ERR_PASS_CLEANUP(ip_calcSizeOfTundraDecodeTable (pIPState,
						     &numItemsArray[0],
						     (Int32 *) &temp));

	/* Allocate the selector array structure.  Make sure all
	 * pointers are initally NULL. */
    pDeltaXSel = (TCTundraDecSelectorArray *)
		CALLOC (1, sizeof(TCTundraDecSelectorArray) +
			   temp * sizeof(TCTundraDecSelector));
    ERR_GEN_CLEANUP (pDeltaXSel == NULL, ia_nomem);
    pDeltaXSel->arraySize = temp;
    pDeltaXSel->nextTable = 256;

	/* Go through the encoder classes, inserting their headers into
	 * the decoder array.
	 */
    for (temp = pDeltaXDecoder->bigPosNumClass - 1; temp > 0; temp--)
    {
	ERR_PASS_CLEANUP(ip_putValIntoTundraDecodeTable(
			 pIPState,
			 pDeltaXSel,
			 (pDeltaXDecoder->pClassArray)[temp].bitHeader,
			 (pDeltaXDecoder->pClassArray)[temp].bitHeaderLen,
			 (void *) &((pDeltaXDecoder->pClassArray)[temp])));
    }


	/* Repeat this process for the delta Y tables.  Note that
	 * the delta Y class has no unbounded negative class, so
	 * we should look at class 0 when putting codes into the
	 * table.  Note also that the code for the unbounded delta
	 * Y class is not "stolen" from one of the other codes,
	 * it's a real code all by itself.  Therefore, it must
	 * go into the decoding table to make the decoding table
	 * complete.
	 * Ooops, except for deltaY table 2.  There, the header
	 * for the unbounded class is stolen from one of the
	 * previous class.  Weird, huh?
	 */
    memset(numItemsArray, 0, 32*sizeof(UInt32));

    if (pDeltaYDecoder == deltaYDecTableArray[2])
    {
	for (temp = pDeltaYDecoder->bigPosNumClass - 1; temp >= 0; temp--)
	    numItemsArray[(pDeltaYDecoder->pClassArray)[temp].bitHeaderLen]++;
    }
    else
    {
	for (temp = pDeltaYDecoder->bigPosNumClass; temp >= 0; temp--)
	    numItemsArray[(pDeltaYDecoder->pClassArray)[temp].bitHeaderLen]++;
    }

    ERR_PASS_CLEANUP(ip_calcSizeOfTundraDecodeTable (pIPState,
						     &numItemsArray[0],
						     (Int32 *) &temp));

	/* Allocate the selector array structure.  Make sure all
	 * pointers are initally NULL. */
    pDeltaYSel = (TCTundraDecSelectorArray *)
		CALLOC (1, sizeof(TCTundraDecSelectorArray) +
			   temp * sizeof(TCTundraDecSelector));
    ERR_GEN_CLEANUP (pDeltaYSel == NULL, ia_nomem);
    pDeltaYSel->arraySize = temp;
    pDeltaYSel->nextTable = 256;

	/* Go through the encoder classes, inserting their headers into
	 * the decoder array.
	 */
    for (temp = pDeltaYDecoder->bigPosNumClass - 1; temp >= 0; temp--)
    {
	ERR_PASS_CLEANUP(ip_putValIntoTundraDecodeTable(
			 pIPState,
			 pDeltaYSel,
			 (pDeltaYDecoder->pClassArray)[temp].bitHeader,
			 (pDeltaYDecoder->pClassArray)[temp].bitHeaderLen,
			 (void *) &((pDeltaYDecoder->pClassArray)[temp])));
    }

	/************************** DECODE LOOP ************************/

	/* This is the loop where we plow through the input data, decoding
	 * strips composed of:
	 *
	 * deltaY, firstX, yPos, sym1, deltaX, yPos, sym2, ..., symbolN, 00
	 *
	 * The position buffer ends with delta Y that takes us below the
	 * image (curY >= hS).  There is an extension count after that,
	 * but we don't support extensions so we ignore it.
	 *
	 * We format this as an inner and outer loop.  The
	 * inner loop processes one strip.  The outer loop goes
	 * through all the strips.  The outer loop processes a
	 * deltaX and firstX for a line before entering the inner
	 * loop.  The inner loop parses a y position and a symbol code,
	 * then blits the symbol into place, then parses the next delta X
	 * code.  If that deltaX code is 00, the inner loop terminates.
	 *
	 * We need a policy for filling the bit accumulator, accum.
	 * We'd like to do this as seldom as possible (eat big bites),
	 * we'd like to guarantee that we always have enough to
	 * decode the next token and we must guarantee that we never
	 * read past the end of the buffer.  As of now, there is no
	 * need to be able to restart if we run out of bytes.  In that
	 * case, the client should just clear the image, supply the
	 * the correct number of position bytes and start over.
	 * Our refill policy will be to try to have at least 16 bits 
	 * before starting to decode the the delta Y, firstX, and
	 * symbol tokens.  Some symbol tokens may be longer than 16 bits,
	 * so we refill in the middle of decoding a sybmol.
	 * We try to have at least 24 bits before trying to decode
	 * the deltaX token since some deltaX tokens are longer than
	 * 16 bits.
	 *
	 * The longest bounded firstX token is 14 bits and the longest
	 * yPos token is 2 bits, so we don't normally need a refill
	 * before decoding a yPos.  However, if the firstX token was
	 * unbounded, we'll do a refill after decoding it.
	 *
	 * The refill code tries to get as much as it's supposed to.
	 * If fewer bytes than that remain in the buffer, it gets what
	 * it can.  If it can't fetch anything and vacantBits has gotten
	 * bigger than (32 - (length of the shortest extension length
	 * code)), it declares ia_bufferExceeded and quits.
	 * This policy means that if the buffer runs dry, some of the
	 * last tokens may be decoded incorrectly, so some trash may
	 * appear in the image.  Tough.  It's the client's job to give
	 * us the right amount of data.  That amount is recorded in the
	 * compressed file.  We're doing the buffer length check
	 * as a safety/debugging feature.
	 */
    lastFirstX = 0;		/* standard location of first symbol in
				 * imaginary strip above top of image.
				 */
    numSymbolsFound = 0;	/* haven't really found any symbols yet */
    while (cTrue)
    {
	     /* Try to have at least 16 bits before decoding deltaY */
	if (vacantBits >= (32-16))
	{
	    if (pLocalBuf < pLocEnd)	/* can read 2 bytes */
	    {
		vacantBits -= 8;
		accum |= (UInt32) pLocalBuf[0] << vacantBits;
		vacantBits -= 8;
		accum |= (UInt32) pLocalBuf[1] << vacantBits;
		pLocalBuf += 2;
	    }
	    else
	    if (pLocalBuf == pLocEnd)	/* can read 1 byte */
	    {
		vacantBits -= 8;
		accum |= (UInt32) pLocalBuf[0] << vacantBits;
		pLocalBuf += 1;
	    }
	    else	/* can't read any bytes.  Are we empty now? */
	    if (vacantBits > (32 - cTundraMinExtensionLenCodeLen))
	    {
		__error__ = ia_bufferExceeded;
		goto ERR_CLEANUP;
	    }
	}

	    /* we've now got enough for the longest delta Y header, so
	     * decode it.  The first step is to see if it's the
	     * unbounded class.  If so, use ip_decodeTundraBigVal to
	     * decode it.  Otherwise, do it the fast way.
	     */
	if ((accum & pDeltaYDecoder->headerMask) == pDeltaYDecoder->headerRoot)
	{
	    ERR_PASS_CLEANUP(ip_decodeTundraBigVal (pIPState, pDeltaYDecoder,
						    &accum, &vacantBits,
						    &pLocalBuf, pLocEnd,
						    (UInt32 *) &dY));
	}
	else
	{
		/* Since the longest deltaY header is 8 bits, we can
		 * decode it in a single table reference.  No need
		 * to follow lower levels of tables.
		 */
	    pEncClass = (TCTundraEncClass *)
			    (pDeltaYSel->array[accum >> (32-8)].pEntry);
	    ERR_GEN_CLEANUP (pEncClass == NULL, ia_internal);
	    accum <<= pEncClass->bitHeaderLen;

		/* Since the longest bounded deltaY code is 13 bits
		 * and we had at least 16 bits, just decode the value now
		 * too.
		 */
	    if (pEncClass->encValLen > 0)
		dY = (accum >> (32 - pEncClass->encValLen)) + pEncClass->offset;
	    else
		dY = pEncClass->offset;
	    accum <<= pEncClass->encValLen;
	    vacantBits += pEncClass->encValLen + pEncClass->bitHeaderLen;
	}
	curY += dY << logStripHeight;

	    /* If we're off the bottom of the image, we're done. */
	if (((useTransposition) && (curY >= wS)) ||
	    ((!useTransposition) && (curY >= hS)))
	    break;			/* out of the outer loop */

	    /* That was a legal Y position.  Go get the first X.
	     * First, refill the accumulator.
	     */
	if (vacantBits >= (32-16))
	{
	    if (pLocalBuf < pLocEnd)	/* can read 2 bytes */
	    {
		vacantBits -= 8;
		accum |= (UInt32) pLocalBuf[0] << vacantBits;
		vacantBits -= 8;
		accum |= (UInt32) pLocalBuf[1] << vacantBits;
		pLocalBuf += 2;
	    }
	    else
	    if (pLocalBuf == pLocEnd)	/* can read 1 byte */
	    {
		vacantBits -= 8;
		accum |= (UInt32) pLocalBuf[0] << vacantBits;
		pLocalBuf += 1;
	    }
	    else	/* can't read any bytes.  Are we empty now? */
	    if (vacantBits > (32 - cTundraMinExtensionLenCodeLen))
	    {
		__error__ = ia_bufferExceeded;
		goto ERR_CLEANUP;
	    }
	}

	    /* we've now got enough for the longest first X header, so
	     * decode it.  The first step is to see if it's the
	     * unbounded class.  If so, use ip_decodeTundraBigVal to
	     * decode it.  Otherwise, do it the fast way.
	     */
	if ((accum & pFirstXDecoder->headerMask) == pFirstXDecoder->headerRoot)
	{
	    ERR_PASS_CLEANUP(ip_decodeTundraBigVal (pIPState, pFirstXDecoder,
						    &accum, &vacantBits,
						    &pLocalBuf, pLocEnd,
						    (UInt32 *) &dX));

		/* Refill here also since we might not have enough for
		 * getting the y position within the strip if we don't.
		 */
	    if (vacantBits >= (32-16))
	    {
		if (pLocalBuf < pLocEnd)	/* can read 2 bytes */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[0] << vacantBits;
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[1] << vacantBits;
		    pLocalBuf += 2;
		}
		else
		if (pLocalBuf == pLocEnd)	/* can read 1 byte */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[0] << vacantBits;
		    pLocalBuf += 1;
		}
		else	/* can't read any bytes.  Are we empty now? */
		ERR_GEN_CLEANUP (vacantBits >
					(32 - cTundraMinExtensionLenCodeLen),
				 ia_bufferExceeded);
	    }
	}
	else
	{
		/* Since the longest bounded firstX header is 5 bits, we can
		 * decode it in a single table reference.  No need
		 * to follow lower levels of tables.
		 */
	    pEncClass = (TCTundraEncClass *)
			    (pFirstXSel->array[accum >> (32-8)].pEntry);
	    ERR_GEN_CLEANUP (pEncClass == NULL, ia_internal);
	    accum <<= pEncClass->bitHeaderLen;

		/* Since the longest bounded firstX code is 14 bits
		 * and we had at least 16 bits, just decode the value now
		 * too.
		 */
	    dX = (accum >> (32 - pEncClass->encValLen)) + pEncClass->offset;
	    accum <<= pEncClass->encValLen;
	    vacantBits += pEncClass->encValLen + pEncClass->bitHeaderLen;
	}

	    /* Calculate the X position of the left edge of the
	     * first symbol in this strip.  Remember this position
	     * for the next strip.
	     */
	curX = lastFirstX + dX;
	lastFirstX = curX;

	    /********************* REMAINDER OF STRIP LOOP *****************/

	    /* Here's where we read the rest of the strip.  We start by reading
	     * y position, then a symbol, then a deltaX.  If the deltaX
	     * is 00, we're done with the strip.  Otherwise, we loop back
	     * up to read another y and symbol.
	     */
	while (cTrue)
	{
		/* get ready to read y.  Refill the accumulator first. */
	    if (vacantBits >= (32-16))
	    {
		if (pLocalBuf < pLocEnd)	/* can read 2 bytes */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[0] << vacantBits;
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[1] << vacantBits;
		    pLocalBuf += 2;
		}
		else
		if (pLocalBuf == pLocEnd)	/* can read 1 byte */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[0] << vacantBits;
		    pLocalBuf += 1;
		}
		else	/* can't read any bytes.  Are we empty now? */
		ERR_GEN_CLEANUP(vacantBits >
					(32 - cTundraMinExtensionLenCodeLen),
				ia_bufferExceeded);
	    }

		/* Read the intra-strip y position, which has at
		 * most 2 bits.  If logStripHeight == 0, i.e. there
		 * is no Y coord, then yPos stays 0.
		 */
	    if (logStripHeight > 0)
	    {
		yPos = accum >> (32 - logStripHeight);
		accum <<= logStripHeight;
		vacantBits += logStripHeight;
	    }

		/* Parse the first 8 bits of the symbol ID.  Since
		 * we had at least 16 bits and have read at most
		 * 2 at this point, it's safe to read another 8.
		 */
	    temp = accum >> (32 - 8);
		/* If there's more, go into a loop that read the
		 * next 4 bits in each pass.
		 */
	    if (pSymSel->array[temp].bitLength > 8)
	    {
		pDecSelector = (TCTundraDecSelector *)
				pSymSel->array[temp].pEntry;
		accum <<= 8;
		vacantBits += 8;
		    /* pDecSelector now points to the first entry in
		     * a 16 entry table.  The correct entry within
		     * that table is accessed by the next 4 bits
		     * in accum.
		     */
		pDecSelector = &(pDecSelector[accum >> (32-4)]);
		ERR_GEN_CLEANUP (pDecSelector == NULL, ia_internal);

		dX = 12;			/* current code length */
		while (pDecSelector->bitLength > dX)
		{
			/* refill the bit accumulator if necessary */
		    if (vacantBits >= (32-16))
		    {
			if (pLocalBuf < pLocEnd)	/* can read 2 bytes */
			{
			    vacantBits -= 8;
			    accum |= (UInt32) pLocalBuf[0] << vacantBits;
			    vacantBits -= 8;
			    accum |= (UInt32) pLocalBuf[1] << vacantBits;
			    pLocalBuf += 2;
			}
			else
			if (pLocalBuf == pLocEnd)	/* can read 1 byte */
			{
			    vacantBits -= 8;
			    accum |= (UInt32) pLocalBuf[0] << vacantBits;
			    pLocalBuf += 1;
			}
			else	/* can't read any bytes.  Are we empty now? */
			ERR_GEN_CLEANUP(vacantBits >
					(32 - cTundraMinExtensionLenCodeLen),
					ia_bufferExceeded);
		    }
			/* go down one level in the decode tree.  There will
			 * be another 16-entry table there. */
		    pDecSelector = (TCTundraDecSelector *) pDecSelector->pEntry;
		    ERR_GEN_CLEANUP (pDecSelector == NULL, ia_internal);
		    accum <<= 4;
		    vacantBits += 4;
			/* pDecSelector now points to the first entry in
			 * a 16 entry table.  The correct entry within
			 * that table is accessed by the next 4 bits
			 * in accum.
			 */
		    pDecSelector = &(pDecSelector[accum >> (32-4)]);
		    ERR_GEN_CLEANUP (pDecSelector == NULL, ia_internal);
		    dX += 4;
		}
		    /* Set dX to the amount by which accum has been shifted */
		dX -= 4;
	    }
	    else
	    {
		    /* set up dX so it's consistent with the way it
		     * would appear if we had decoded more than one level.
		     */
		dX = 0;
		pDecSelector = &(pSymSel->array[temp]);
		ERR_GEN_CLEANUP (pDecSelector == NULL, ia_internal);
	    }

		/* We've gotten to the end of the symbol code.  At this point,
		 * we've already shifted accum left by dX bits and
		 * added dX to vacantBits.  We need to shift
		 * accum left by the symbolCodelength - dX and
		 * add symbolCodeLength - dX to vacantBits.
		 */
	    pSymbol = (TCSymbol *) (pDecSelector->pEntry);
	    ERR_GEN_CLEANUP(pSymbol == NULL, ia_internal);
	    accum <<= pSymbol->huffmanCodeLen - dX;
	    vacantBits += pSymbol->huffmanCodeLen - dX;

		/* In the real Tundra world, there's always a
		 * symbol bitmap, so paint it.
		 * However, we just look at the transposition flag
		 * to figure out whether to swap the X and Y
		 * coords.  The symbol
		 * bitmap is offset in a height class bitmap.
		 * BUT FIRST, clip the symbol to the image.
		 * Symbols are sometimes matched to other wider
		 * or taller symbols.  Their positions are sometimes
		 * bumped up/down/left/right by a pixel.
		 */
	    if (useTransposition)
	    {
		    /* In this mode, (curX, curY) is the location of the
		     * upper left corner of the symbol.  CurX can
		     * theoretically go negative if a symbol is bumped
		     * up, so we have to check that.  CurX + height can
		     * extend beyond the bottom of the image if the symbol
		     * is bumped down, so we have to check that.
		     * CurY + yPos can't be negative (there's no way to encode
		     * that), so we don't have to check that.  CurY + pPos
		     * + symWidth can be greater than wS, so we need to
		     * check that.
		     * Well, umm, William shifts symbols onto the page
		     * when they fall off instead of just showing the
		     * visible part.  I'll change to do that.  Note that
		     * we can't actually change curX and curY since
		     * subsequent placement will be based on their
		     * current values.
		     */
		    /* Is symbol off top of page? */
		dX = curX;
		if (dX < 0)
		{
		    dX = 0;
		}
		if (dX + pSymbol->height > hS)
		{
		    dX = hS - pSymbol->height;
		}
		dY = curY + yPos;
		if (dY + pSymbol->width > wS) 
		{
		    dY = wS - pSymbol->width;
		}
		ERR_PASS_CLEANUP(i_rasterOp (pIPState,
					     pImage,		/* dest ptr */
					     dY,		/* dx */
					     dX,		/* dy */
					     pSymbol->width,	/* dw */
					     pSymbol->height,	/* dh */
					     1,			/* dd */
					     bpl,		/* dLinebytes */
					     PIX_SRC | PIX_DST,	/* op */
					     pSymbol->pBitmap,	/* src ptr */
					     pSymbol->xOffset, 0, /* sx, sy */
					     1,			/* sd */
					     pSymbol->bytesPerLine));
								/* sLineBytes */
#ifdef DDEBUG
		printf("(%d,%d) %dx%d code = %d-%d\n", dY, dX,
		    pSymbol->width, pSymbol->height,
		    pSymbol->huffmanCode, pSymbol->huffmanCodeLen);
#endif
	    }
	    else	/* no transposition, the normal case */
	    {
		dX = curX;
		if (curX < 0)
		{
		    dX = 0;
		}
		if (dX + pSymbol->width > wS)
		{
		    dX = wS - pSymbol->width;
		}
		dY = curY + yPos;
		if (dY - pSymbol->height + 1 < 0) 
		{
		    dY = pSymbol->height - 1;
		}
		    /* We don't have to check to see if curY is off the
		     * bottom of the image since it can't be.  If it were,
		     * we would have ended the position block.
		     */
		ERR_PASS_CLEANUP(i_rasterOp (pIPState,
					     pImage,		/* dest ptr */
					     dX,		/* dx */
					     dY - pSymbol->height + 1,
								/* dy */
					     pSymbol->width,	/* dw */
					     pSymbol->height,	/* dh */
					     1,			/* dd */
					     bpl,		/* dLinebytes */
					     PIX_SRC | PIX_DST,	/* op */
					     pSymbol->pBitmap,	/* src ptr */
					     pSymbol->xOffset, 0, /* sx, sy */
					     1,			/* sd */
					     pSymbol->bytesPerLine));
								/* sLineBytes */
#ifdef DDEBUG
		printf("(%d,%d) %dx%d code = %d-%d\n", dX, dY,
		    pSymbol->width, pSymbol->height,
		    pSymbol->huffmanCode, pSymbol->huffmanCodeLen);
#endif
	    }

		/* That was lovely, another symbol painted.  Refill the
		 * accumulator again in anticipation of reading a
		 * delta X coord.  This time, we try to have at
		 * least 24 bits so we can decode 17-bit deltaX codes.
		 * Note that if there are fewer than 17 valid bits,
		 * there are at least 16 vacant bits, so we enter
		 * the refill case.  If there were 32 vacant bits,
		 * reading two bytes only gets us 16 valid bits,
		 * which isn't enough.  Therefore, we have to try
		 * to get at least 24 valid bits.
		 */
	    numSymbolsFound++;
	    if (vacantBits >= (32-24))
	    {
		if ((vacantBits >= (32 - 8)) &&
		    (pLocalBuf < pLocEnd - 1))	/* can read 3 bytes */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[0] << vacantBits;
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[1] << vacantBits;
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[2] << vacantBits;
		    pLocalBuf += 3;
		}
		else
		if ((vacantBits >= (32 - 16)) &&
		    (pLocalBuf < pLocEnd))	/* can read 2 bytes */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[0] << vacantBits;
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[1] << vacantBits;
		    pLocalBuf += 2;
		}
		else
		if (pLocalBuf <= pLocEnd)	/* can read 1 byte */
		{
		    vacantBits -= 8;
		    accum |= (UInt32) pLocalBuf[0] << vacantBits;
		    pLocalBuf += 1;
		}
		else	/* can't read any bytes.  Are we empty now? */
		ERR_GEN_CLEANUP (vacantBits >
					(32 - cTundraMinExtensionLenCodeLen),
				 ia_bufferExceeded);
	    }
		/* We've got enough to read the entire deltaX coord.
		 * First, check to see if it's 00.  If so, we're done
		 * with this strip.  There's no entry for 00 in the
		 * decoding tables, so don't try to decode it.
		 */
	    if ((accum & cTundraStripEndCodeMask) ==
			(cTundraStripEndCode << (32 - cTundraStripEndCodeLen)))
	    {
		accum <<= 2;
		vacantBits += 2;
		break;			/* from the strip loop */
	    }

		/* If deltaX is an unbounded coord, read it the hard way.
		 */
	    if ((accum & pDeltaXDecoder->headerMask) ==
						pDeltaXDecoder->headerRoot)
	    {
		ERR_PASS_CLEANUP(ip_decodeTundraBigVal (pIPState,
							pDeltaXDecoder,
						        &accum, &vacantBits,
						        &pLocalBuf, pLocEnd,
						        (UInt32 *) &dX));
	    }
	    else	/* This is the easy way? */
	    {
		    /* There are delta X codes longer than 8 bits, but
		     * none longer than 12 bits, so we can do this in
		     * at most 2 steps.
		     */
		temp = accum >> (32 - 8);
		    /* If there's more, go into a loop that read the
		     * next 4 bits in each pass.
		     */
		if (pDeltaXSel->array[temp].bitLength > 8)
		{
			/* Was a long symbol code.  Therefore, get the
			 * pointer to the second level table.  Then
			 * select the entry in the second level
			 * table and shift off all the first level bits.
			 */
		    pDecSelector = (TCTundraDecSelector *)
					    pDeltaXSel->array[temp].pEntry;
		    ERR_GEN_CLEANUP (pDecSelector == NULL, ia_internal);
		    pDecSelector = &(pDecSelector[(accum >> (32-12)) & 0xF]);
		}
		else	/* was a short symbol code */
		{
		    pDecSelector = &(pDeltaXSel->array[temp]);
		    ERR_GEN_CLEANUP (pDecSelector == NULL, ia_internal);
		}

		    /* We`ve now parsed the header.  Shift it out of the
		     * accumulator and decode the length following.
		     * If the header entirely encodes the length
		     * (encValLen == 0), the offset equals the length
		     * directly.
		     */
		accum <<= pDecSelector->bitLength;
		vacantBits += pDecSelector->bitLength;
		pEncClass = (TCTundraEncClass *)(pDecSelector->pEntry);
		ERR_GEN_CLEANUP (pEncClass == NULL, ia_internal);
		if (pEncClass->encValLen != 0)
		{
		    dX = (accum >> (32 - pEncClass->encValLen)) + 
			 pEncClass->offset;
		    accum <<= pEncClass->encValLen;
		    vacantBits += pEncClass->encValLen;
		}
		else
		    dX = pEncClass->offset;
	    }
		/* adjust the current X position.  Remember that the encoder
		 * encoded dX = (left edge of next symbol) - (right edge of
		 * previous symbol) + 1 - modalDeltaX.
		 * Also remember that in transposed mode,
		 * encoded dX = (top edge of next symbol) - (bottom edge of
		 * previous symbol) + 1 - modalDeltaX
		 */
	    if (useTransposition)
		curX += pSymbol->height + dX + modalDeltaX;
	    else
		curX += pSymbol->width + dX + modalDeltaX;

	}	/* end of loop that reads the rest of one strip */
    }		/* end of loop that reads all strips. */

	/* We've blitted all the symbols.  Make sure we found the
	 * right number.  If not, complain.  If so, be happy and leave */
    if (numSymbolsExpected != numSymbolsFound)
	__error__ = ia_decodeFailure;
    else
	__error__ = ia_successful;

#ifdef DDDEBUG
	 {
	     TCSymbol *pSym = pTextCompInfo->p1stSymbol;

	     printf("Decoded symbols:\n");
	     while (pSym) {
		 printf("%3d (%3d): dict %d, %dx%d, count %d, code %d-%d\n",
		     pSym->dictOrder, pSym->symbolID, pSym->dictNum, pSym->width,
		     pSym->height, pSym->instanceCount, pSym->huffmanCode,
		     pSym->huffmanCodeLen);
		 pSym = pSym->pNextDictSym;
	     }
	 }
#endif


ERR_CLEANUP:
    if (pSymSel != NULL)
	FREE(pSymSel);
    if (pFirstXSel != NULL)
	FREE(pFirstXSel);
    if (pDeltaXSel != NULL)
	FREE(pDeltaXSel);
    if (pDeltaYSel != NULL)
	FREE(pDeltaYSel);

    NORMAL_RETURN;
}


/* This is a big number used to mark the right coord of the second dummy
 * run at the end of each scan line in the CCITT run array.
 */
#define	cDummyMarker	0x7FFFFFFF

/**************************************************************************
 *   i_readTundraResidualBlock()
 *
 *	This procedure decompresses a Tundra residual block, producing
 *	a bitmap.  The residual block, which may be smaller than the
 *	destination image, is placed in the dest image at a specific
 *	offset.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pResidualBuf:		pointer to buffer containing
 *					Tundra residual block
 *		bufSize:		number of bytes available to
 *					read.  If this is 0, the residual
 *					block is uncompressed.  Its size
 *					can be calculated as
 *					resH * ((resW + 7)/8).
 *		combineResults:		If set to cTrue, the residual block
 *					is ORed into an existing image.
 *					If set to cFalse, the original
 *					image is assumed to be initially
 *					white (pix = 0).
 *		pImage:			Pointer to upper left corner of image
 *					INCLUDING frame.
 *		wS:			width of image including frame
 *		hS:			height of image including frame
 *		bpl:			The image's bytes/line
 *		resX:			X coord of residual image's left
 *					edge relative to dest INCLUDING frame.
 *		resY:			Y coord of residual image's top
 *					edge relative to dest INCLUDING frame.
 *		resW:			width of residual image (does not
 *					include hypothetical frame)
 *		resH:			height of residual image (does not
 *					include hypothetical frame)
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL or the
 *					residual image did not fit inside
 *					the dest image.
 *		ia_bufferExceeded	field is bigger than room available.
 *		ia_nomem		couldn't get enough memory for
 *					internal data structures.
 *
 **************************************************************************/
Int32  CDECL 
i_readTundraResidualBlock (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pResidualBuf,
    UInt32	bufSize,
    Bool	combineResults,
    UInt8      *pImage,
    Int32	wS,
    Int32	hS,
    Int32	bpl,
    Int32	resX,
    Int32	resY,
    Int32	resW,
    Int32	resH)
{
    PROLOGUE;
    TCCompressionInfoPtr	 pTextCompInfo;
    CCITTDecompressionInfo	*pCCITTDecompInfo = NULL;
    Int32			 ccittInfoSize;
    UInt8			*pCompDataEnd;	/* ptr beyond end of
						 * pResidualBuf */
    
    Run32			*pCodingRuns;	/* Pointer to runs for coding
						 * line. */
    Run32			*pRefRuns; 	/* pointer to runs for reference
						 * line in array of runs. */
    Run32			*pDummy, *pAdj;	/* scratch ptrs */
    UInt8			*pBuf = NULL;	/* ptr to intermediate buffer */

	/* check the pointers */
    if ((pTextCompInfoIn == NULL) ||
	(pResidualBuf == NULL) || (pImage == NULL))
	return ia_invalidParm;

	/* See if the residual fits */
    if ((resX < cFrameBits) || (resY < cFrameBits) ||
	(resX + resW > wS - cFrameBits) || (resY + resH > hS - cFrameBits))
	return ia_invalidParm;

	/* get a useful pointer to the text compression info structure */
    pTextCompInfo = (TCCompressionInfo *) pTextCompInfoIn;

	/* Remove the frames from the coords. */
    resX -= cFrameBits;
    resY -= cFrameBits;
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    pImage += (bpl << cLogFrameBits) + cFrameBytes;

	/* Move pImage down to the start of the residual image */
    pImage += resY * bpl;

	/* If the file is uncompressed, just write it or OR it into
	 * the image.
	 */
    if (bufSize == 0)
    {
	Int32		 bufBpl;
	GraySwapNearProc swapLineProc = NULL;
	Int32		 row, inputBpl;
	Int32		 op;

	    /* This is a VERY rare case.  In all our test images,
	     * there's only one image that has an uncompressed
	     * residual.  Therefore, we'll write code that isn't
	     * very fast, but only has one path so it can be
	     * tested with our one test case.
	     * The basic algorithm is to copy each scan line
	     * from the input to a buffer, swap the bytes
	     * if necessary, then call bitblit to move the
	     * bits into position.  To try to save some time,
	     * we'll buffer up several lines at a time before
	     * calling bitblit.
	     * First, get a pointer to the byteswap routine.
	     * It will only be used on little endian machines.
	     */
	ERR_PASS_CLEANUP(i_getDIBPixrLineProc (pIPState,
					       cFalse,	    /* invertImage */
					       cPixrToPixr, /* direction */
					       cFalse,	    /* isColor */
					       NULL, NULL,
					       &swapLineProc, NULL));

	    /* figure out what type of operation we should use.
	     * If combineResults is set, we want an OR.  Othewise
	     * we want just transfer.
	     */
	if (combineResults)
	    op = PIX_SRC | PIX_DST;
	else
	    op = PIX_SRC;

	    /* The uncompressed bitmap begins at pResidualBuf.  It
	     * has resH rows, each (resW + 7)/8 bytes
	     * long.  Create a buffer with 4 rows of this stuff,
	     * each row padded out to a 4-byte boundary.  This
	     * will be the input for bitBlit.
	     */
	bufBpl = ((resW + 31) / 32) * 4;
	pBuf = (UInt8 *) MALLOC (4 * bufBpl);
	ERR_GEN_CLEANUP(pBuf == NULL, ia_nomem);

	    /* Copy each row, byteswap if necessary and
	     * write the words into the uncompressed bitmap.
	     * Remember that we've already adjusted pImage to
	     * point to the left edge of the bitmap on row resY.
	     */
	inputBpl = (resW + 7) / 8;
	for (row = 0;
	     row <= resH - 4;
	     row += 4, pImage += 4 * bpl)
	{
	    memcpy(pBuf, pResidualBuf, inputBpl);
	    pResidualBuf += inputBpl;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    (*swapLineProc)(pIPState, (UInt32 *)pBuf,	/* source */
			    (UInt32 *)pBuf,	/* dest */
			    bufBpl/4);		/* words to be swapped */
#endif

	    memcpy(pBuf + bufBpl, pResidualBuf, inputBpl);
	    pResidualBuf += inputBpl;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    (*swapLineProc)(pIPState, (UInt32 *)(pBuf + bufBpl), /* source */
			    (UInt32 *)(pBuf + bufBpl),	/* dest */
			    bufBpl/4);			/* num words */
#endif

	    memcpy(pBuf + 2*bufBpl, pResidualBuf, inputBpl);
	    pResidualBuf += inputBpl;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    (*swapLineProc)(pIPState, (UInt32 *)(pBuf + 2*bufBpl), /* source */
			    (UInt32 *)(pBuf + 2*bufBpl), /* dest */
			    bufBpl/4);			 /* num words */
#endif

	    memcpy(pBuf + 3*bufBpl, pResidualBuf, inputBpl);
	    pResidualBuf += inputBpl;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    (*swapLineProc)(pIPState, (UInt32 *)(pBuf + 3*bufBpl), /* source */
			    (UInt32 *)(pBuf + 3*bufBpl), /* dest */
			    bufBpl/4);			 /* num words */
#endif
	    ERR_PASS_CLEANUP(i_rasterOp (pIPState,
					 pImage,	/* dest ptr */
					 resX,		/* dx */
					 0,		/* dy */
					 resW,		/* dw */
					 4,		/* dh */
					 1,		/* dd */
					 bpl,		/* dLinebytes */
					 op,		/* op */
					 pBuf,		/* src ptr */
					 0, 0,		/* sx, sy */
					 1,		/* sd */
					 bufBpl));	/* sLineBytes */
	}

	    /* Deal with the last lines of the residual, if any */
	for (; row < resH; row++, pImage += bpl)
	{
	    memcpy(pBuf, pResidualBuf, inputBpl);
	    pResidualBuf += inputBpl;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    (*swapLineProc)(pIPState, (UInt32 *)pBuf,	/* source */
			    (UInt32 *)pBuf,	/* dest */
			    bufBpl/4);		/* words to be swapped */
#endif
	    ERR_PASS_CLEANUP(i_rasterOp (pIPState,
					 pImage,		/* dest ptr */
					 resX,			/* dx */
					 0,			/* dy */
					 resW,			/* dw */
					 1,			/* dh */
					 1,			/* dd */
					 bpl,			/* dLinebytes */
					 op,			/* op */
					 pBuf,			/* src ptr */
					 0, 0,			/* sx, sy */
					 1,			/* sd */
					 bufBpl));		/* sLineBytes */
	}

	    /* We've moved the residual into position.  Free
	     * the intermediate buffer.
	     */
	FREE(pBuf);
	pBuf = NULL;

    }	/* end of dealing with an uncompressed residual */
    else	/* the buffer is compressed.  This is the normal case */
    {
	    /* Go get the CCITT decompression structure.  It will be used
	     * by ip_decompMMRRow.
	     */
	ERR_PASS_CLEANUP(i_getSizeOfCCITTDecompressionInfo (pIPState, 
							    resW + 2*cFrameBits,
							    &ccittInfoSize));

	pCCITTDecompInfo = (CCITTDecompressionInfo *) MALLOC(ccittInfoSize);
	ERR_GEN_CLEANUP(pCCITTDecompInfo == NULL, ia_nomem);

	ERR_PASS_CLEANUP(
		i_initializeCCITTDecompressionInfo (pIPState,
						    (void *)pCCITTDecompInfo,
						    resW + 2*cFrameBits,
						    cFalse,
						    cNoPad, ccittInfoSize));

	    /* Set the CCITT run array pointers to the second run in each array.
	     * The first run in each array is a dummy.
	     * Since we know that run2ArrayIsCoding is always set to false,
	     * i.e. the first run array is the initial coding array,
	     * just point to the second array.
	     */
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
					  pCCITTDecompInfo->runArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				       pCCITTDecompInfo->run2ArrayOffset +
				       sizeof(Run32));
    
	    /* Also initialize the reference run to have a blank line.
	     * This is done by putting in only the dummy elements.
	     */
	pRefRuns[0].runLeft = resW;
	pRefRuns[0].runRight = resW-1;
	pRefRuns[1].runLeft = resW;
	pRefRuns[1].runRight = cDummyMarker;
    
	    /* Initialize the end pointer used by ip_decompMMRRow. */
	pCompDataEnd = pResidualBuf + bufSize;
    
	    /* Now, we're going to loop through all the scan lines of
	     * the residual bounding box.  For each line, we decode
	     * the runs of the scan line, adjust their positions by
	     * left edge of the residual bounding box (if resX is
	     * non-zero), then use ip_paintRunsLine to paint them.
	     */
	for (; resH > 0; resH--)
	{
		/* Decode the runs for the next line.
		 * If anything went wrong, quit.  We assume that the
		 * client has supplied the entire residual buffer since
		 * the client knows how big it is.
		 */
	    ERR_PASS_CLEANUP(ip_decompMMRRow (pIPState, pCCITTDecompInfo,
					      resW,	/* encoded width */
					      cTrue,	/* this is the only
							 * buffer, so flush
							 * the pipeline */
					      &pResidualBuf,
					      pCompDataEnd,
					      &pDummy));
    

		/* Mark the two runs beyond the last run in the same
		 * way they were marked during encoding.  This guarantees
		 * that we'll get the same answer when we're searching
		 * through these runs when they become the reference row.
		 */
	    pDummy[0].runLeft = resW;
	    pDummy[0].runRight = resW-1;
	    pDummy[1].runLeft = resW;
	    pDummy[1].runRight = cDummyMarker;
    
		/* If resX is non-zero, adjust the coordinates in all
		 * the runs.
		 */
	    if (resX != 0)
	    {
		for (pAdj = pCodingRuns; pAdj != pDummy; pAdj++)
		{
		    pAdj->runLeft += resX;
		    pAdj->runRight += resX;
		}
	    }
    
		/* paint the runs into the image.  The combineResults flag
		 * says whether to OR them in or just assume the original
		 * image is cleared.
		 */
	    ERR_PASS_CLEANUP(ip_paintRunsLine (pIPState, pImage,
					       wS, pCodingRuns,
					       pDummy, combineResults));

		/* If resX is non-zero, put all the run coords back
		 * so they're function properly as the reference run.
		 */
	    if (resX != 0)
	    {
		for (pAdj = pCodingRuns; pAdj != pDummy; pAdj++)
		{
		    pAdj->runLeft -= resX;
		    pAdj->runRight -= resX;
		}
	    }
    
		/* do the normal boilerplate for the end of a scan line. */
	    pCCITTDecompInfo->run2ArrayIsCoding ^= cTrue;
	    pDummy = pCodingRuns;
	    pCodingRuns = pRefRuns;
	    pRefRuns = pDummy;
	    pImage += bpl;
	}
    }	/* end of dealing with a compressed residual */

	/* Hooray, it all worked!  Say we're happy and return. */
    __error__ = ia_successful;

ERR_CLEANUP:
    if (pBuf != NULL)
	FREE(pBuf);
    if (pCCITTDecompInfo != NULL)
	FREE(pCCITTDecompInfo);

    NORMAL_RETURN;
}


/*

 $Log: i_tctnd.c,v $
 Revision 1.2  2001/05/08 06:43:38  jlin
 change ->allocator to ->ipcAllocator

 * Revision 1.1.1.1  1999/09/20  18:18:13  jlin
 * Imported sources
 *
 Revision 1.21  1998/08/04 00:17:01  aber
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 CommonCheck: Retain the page ID such that the dictionary can be split into a
 shared dictionary, and a single-use dictionary for each page.
 Support multiple dictionaries per page.

 Revision 1.20  1998/06/23 21:45:43  aber
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 CommonCheck: Fixes to force Tundra dictionary reading order, esp. for multiple
 dictionaries per page. Some cosmetic comment enhancements. A couple of
 extra error conditions caught.

 Revision 1.19  1998/01/27 01:47:38  ddavies
 ddavies on Mon Jan 26 17:47:17 PST 1998

 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Add some internal error checking and procedure calling macros.

 Revision 1.18  1998/01/13 23:50:15  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.17  1997/08/27 02:42:43  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Clear the new pTextCompInfo->useResidual flag if a dictionary is found
 whose use counts are encoded using the table that allows use counts of 1.

 Revision 1.16  1997/05/08 02:35:52  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Use ip_exactSymbolIsComplete instead of ip_symbolIsComplete to insert
 decoded symbols into symbol tree.  This lets us separate lossless stuff
 from lossy stuff more easily.

 Revision 1.15  1997/04/29 02:47:21  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Initialize instance.numRuns to 0 before deserializing position info.

 Revision 1.14  1997/04/24 23:36:32  ddavies
 ddavies on Thu Apr 24 16:36:05 PDT 1997

 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Change ip_fixupSymbolPtrs to ip_fixupTCSymbolPtrs.

 Revision 1.13  1997/02/12 03:04:26  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Initialize leftProfile, rightProfile and maxDimension in the declared
 TCInstance.

 Revision 1.12  1997/01/04 00:25:32  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Change instance.interior to instance.strokeWidth.

 Revision 1.11  1996/09/07 01:30:07  ddavies
 ddavies on Fri Sep 6 18:29:29 PDT 1996

 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Fix ability to deal with uncompressed residual images.  The previous fix
 assumed the x offset of the image is always 0.

 Revision 1.10  1996/09/06 04:47:32  ddavies
 ddavies on Thu Sep 5 21:40:46 PDT 1996

 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Hey, I hit the carriage return too soon last time.  Among the changes
 required for interoperability with PARC code: Change the clipping used
 when blitting symbols to shift the entire symbol onto the page instead of
 tossing the part that didn't fit.  This isn't optimal, but its what
 William does; Fix an accumulator refill bug before decoding deltaX codes -
 it was not always exiting with at least 24 valid bits; Handle uncompressed
 symbol bitmaps and residual bitmaps; fix bugs in code that handles
 transposed coords (use symbol height, not symbol width, in several spots);

 Revision 1.9  1996/09/06 04:39:43  ddavies
 ddavies on Thu Sep 5 21:39:09 PDT 1996

 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Fix bugs to allow interoperability with William Rucklidge's code at PARC.

 Revision 1.8  1996/09/04 01:24:18  ddavies
 ddavies on Tue Sep 3 18:18:48 PDT 1996

 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Many little fixes to enable this code to really read the PARC Tundra format.
 Among them: fix bug in reading uncompressed bitmaps so that pBitmap is not
 changed during the read (we need to put it into the symbols); add checks
 for number of symbols found vs number of symbols expected when decoding
 a position block; fix check in which we were always looking at pFirstXDecoder
 being null instead of looking at pFirstXDecoder, pDeltaXDecoder and
 pDeltaYDecoder (when initializing these); recognizing that the header
 for unbounded classes in DeltaY table 2 is taken from one of the other
 classes, unlike the headers in the other Delta Y tables; fix checking for
 end of position block when using transposition;  encode/decode distance
 between symbols, not distance between symbol left edges (!) and fix this
 when using transposition.

 Revision 1.7  1996/08/30 23:47:59  ddavies
 ddavies on Fri Aug 30 16:44:09 PDT 1996

 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Use symbol width and height parameters instead of calculating width and height.
 Adjust symbol positions and widths before blitting so symbols are clipped
 to the image size.  This fixes a frame error.
 Attempt to allow the decoder to handle transposed image coordinates.
 Attempt to allow the decoder to handle uncompressed height class bitmaps.
 These two features will be tested by decoding PARC images (i.e. they might
 not really work yet since we have no way of generating these images).

 Revision 1.6  1996/08/29 01:04:38  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Change i_readTundraDictionary from a worthless stub to a routine that really
 reads in Tundra format dictionaries.  This is an enormous change.
 In i_readTundraPositionBlock, convert adhoc constants into defined constants.
 Also change the rasterOp that displays symbols to use the symbol bitmaps.

 Revision 1.5  1996/08/23 02:59:38  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Fix comment containing name of file.
 Modify parsing of internal position info so the block begins with a deltaY
 relative to -1 instead of a deltaX.

 Revision 1.4  1996/08/22 01:24:59  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Converted i_readTundraPositionBlock from a dummy that read the old
 position info blocks into a routine that reads and writes real
 Tundra position info blocks.  This is an enormous change.
 The new version of this routine assigns Huffman codes to symbols, creates
 decoding tables, parses a Tundra position block and blits symbols onto
 a page.

 Revision 1.3  1996/07/19 00:14:16  ddavies
 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Fix comments (i_decompressTundra... -> i_readTundra...).
 Include memory.pub since we allocate and free tables internally now.
 Change meaning of the combineResults parameter to i_readTundraResidualBlock.
 When not set, we now assume the original image buffer is filled with white
 pixels.
 Change i_readTundraResidualBuffer so it actually reads a residual buffer
 instead of just being a stub.  It does this by calling ip_decompMMRRow for
 each row, adjusting the runs if necessary to make them fit in the residual
 bounding box and having ip_paintRunsLine paint them into the image.
 This procedure looks a lot like i_decompressRowsMMR now.

 Revision 1.2  1996/07/03 21:26:58  ddavies
 ddavies on Wed Jul 3 14:26:50 PDT 1996

 i_tctnd.c -> /product/ipcore/ipshared/src/RCS/i_tctnd.c,v

 Fix unused parameter warnings.

 Revision 1.1  1996/07/02 22:58:26  ddavies
 Initial revision


 */

