/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tctne.c:	Utility procedures used for putting out Tundra
 *		compression format.
 *
 *   	Int32	i_writeTundraHeader()
 *			Fills in a Tundra header block.
 *
 *   	Int32	i_writeTundraNumberOfPagesField()
 *			Fills in the Tundra "Number of Pages" field.
 *
 *   	Int32	i_writeTundraDictBlockLengthField()
 *			Fills in a Tundra dictionary block length field.
 *
 *   	Int32	i_writeTundraPageHeader()
 *			Fills in a Tundra Page Header field.
 *
 *   	Int32	i_writeTundraResidualHeader()
 *			Fills in a Tundra Residual Header field inside a
 *			Page block.
 *
 *	Int32	i_estimateTundraDictionaryBlockSize()
 *			Estimate the size of a Tundra Dictionary block.
 *
 *	Int32	i_writeTundraDictionaryData()
 *			Write data portion of a Tundra Dictionary block.
 *
 *	Int32	i_estimateTundraPositionBlockSize()
 *			Estimate the size of a Tundra Position block.  This
 *			includes only position info.
 *
 *	Int32	i_estimateTundraResidualBlockSize()
 *			Estimate the size of a Tundra Residual block.  This
 *			includes only a Group 4 encoded residual image.
 *
 *	Int32	i_writeTundraPositionBlock()
 *			Write a Tundra position block.  This
 *			includes only position and symbol info.
 *
 *	Int32	i_writeTundraResidualBlock()
 *			Write a Tundra residual block.  This
 *			includes only a Group 4 encoded residual images.
 *
 *	Int32	ip_calcPositionBufStats()
 *			Calculates various useful statistics from a
 *			position buffer, like residual image bounding
 *			box, estimated length of residual image, estimated
 *			length of Tundra position buffer, etc.
 *			Also reverse the order of runs found in residual
 *			symbols and sort the residual symbols by position.
 *
 *	Int32	ip_reverseSymbolRuns()
 *			Reverse the order of TCScanLineRuns associated with
 *			a symbol, and count them.  Also convert them to
 *			1x absolute coordinates.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "tundra.prv"	/* for Tundra constants */
#include "memory.pub"

    /*  prototyes */
#include "shrip.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: i_tctne.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");

/* something that's defined to be off the image */
#define cEOLVal  0x7FFF

/* There is a big comment describing our approach to symbol based
 * compression in i_tcompr.c.
 *
 * During compression and decompression, both the shapes and positions
 * of symbols are kept in an internal format.  When the compressed
 * data is transmitted or stored, it is kept in the proposed Tundra
 * format (see "A Proposal for Tundra Coding," by William Rucklidge).
 * This file contains utility routines used during the conversion
 * process.
 *
 * A Tundra file contains one or more dictionary blocks and one or
 * more page blocks.  The dictionary blocks specify the shapes of
 * the symbols.  The page blocks specify the positions of the symbols
 * on each page plus a residue image.  The residue image has stuff that
 * was was either too large to be classified or was only found once.
 *
 * This file contains routines that help the client figure out how
 * much space to allocate for the various chunks of memory that are
 * used to encode a Tundra file.  i_tcjbut.c holds utilities common
 * to encoding and decoding.  i_tcjbd.c holds decoding routines.
 */

/* PUBLIC */
/**************************************************************************
 *   i_writeTundraHeader()
 *
 *	This procedure fills in a Tundra header, returning the number
 *	of bytes used.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pHeader:		pointer to the Tundra header block
 *		headerSize:		size of the Tundra header passed in
 *		pBytesUsed:		number of bytes used in header.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *		ia_bufferExceeded	the block wasn't big enough
 *
 **************************************************************************/
Int32  CDECL 
i_writeTundraHeader (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pHeader,
    UInt32	headerSize,
    UInt32     *pBytesUsed)
{
    PROLOGUE;
    char	headerIDString[] = {cTundraHeaderIDString};
    Int32	i;
    UInt8	*pLocalBuf = pHeader;

	/* Keep the compiler happy */
    pTextCompInfoIn = pTextCompInfoIn;

	/* check to see if the pointers are ok. */
    if ((pHeader == NULL) || (pBytesUsed == NULL))
	return ia_invalidParm;

    if (headerSize < cTundraHeaderSize)
	return ia_bufferExceeded;

	/* fill in the header */
    for (i = 0; i < cTundraHeaderIDSize; i++)
	*pLocalBuf++ = headerIDString[i];

	/* Write the version number */
    *pLocalBuf++ = cTundraCurrentVersion;

	/* Set the number of bytes written. */
    *pBytesUsed = pLocalBuf - pHeader;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_writeTundraNumberOfPagesField()
 *
 *	This procedure fills in the number of pages field in a Tundra
 *	file.  Well, really into a memory block.  It's the client's
 *	job to do file I/O.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		numOfPages:		the quantity to be written.
 *		pNumOfPagesField:	pointer to the Tundra numberOfPages
 *					field
 *		bufSize:		number of bytes available to
 *					write into.
 *		pBytesUsed:		number of bytes used.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL or
 *					the number of pages is too big.
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_writeTundraNumberOfPagesField (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt32	numOfPages,
    UInt8      *pNumOfPagesField,
    UInt32	bufSize,
    UInt32     *pBytesUsed)
{
    PROLOGUE;
    UInt8      *pLocalBuf = pNumOfPagesField;

	/* Keep the compiler happy */
    pTextCompInfoIn = pTextCompInfoIn;

	/* check to see if the pointers are ok. */
    if ((pNumOfPagesField == NULL) || (pBytesUsed == NULL))
	return ia_invalidParm;

	/* If the number of pages is too big, complain */
    if (numOfPages > cTundraMaxNumOfPages)
	return ia_invalidParm;

	/* if the number bytes in the buffer won't hold the field,
	 * complain.
	 */
    if (bufSize < cTundraNumOfPagesSize)
	return ia_bufferExceeded;

	/* Yea, it fits.  Write the page number, MS byte first */
    pLocalBuf[0] = (UInt8) (numOfPages >> 8);
    pLocalBuf[1] = (UInt8) (numOfPages);

    *pBytesUsed = cTundraNumOfPagesSize;

    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *   i_writeTundraDictBlockLengthField()
 *
 *	This procedure fills in a Tundra dictionary block length field.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		dictBlockLength:	the quantity to be written.
 *		pDictBlockLengthField:	pointer to the Tundra dictionary
 *					block length field
 *		bufSize:		number of bytes available to
 *					write into.
 *		pBytesUsed:		number of bytes used.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL or
 *					the number of pages is too big.
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_writeTundraDictBlockLengthField (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt32	dictBlockLength,
    UInt8      *pDictBlockLengthField,
    UInt32	bufSize,
    UInt32     *pBytesUsed)
{
    PROLOGUE;
    UInt8      *pLocalBuf = pDictBlockLengthField;

	/* Keep the compiler happy */
    pTextCompInfoIn = pTextCompInfoIn;

	/* check to see if the pointers are ok. */
    if ((pDictBlockLengthField == NULL) || (pBytesUsed == NULL))
	return ia_invalidParm;

	/* if the number bytes in the buffer won't hold the field,
	 * complain.
	 */
    if (bufSize < cTundraDictBlockLengthSize)
	return ia_bufferExceeded;

	/* Yea, it fits.  Write the page number, MS byte first */
    pLocalBuf[0] = (UInt8) (dictBlockLength >> 24);
    pLocalBuf[1] = (UInt8) (dictBlockLength >> 16);
    pLocalBuf[2] = (UInt8) (dictBlockLength >> 8);
    pLocalBuf[3] = (UInt8) (dictBlockLength);

    *pBytesUsed = cTundraDictBlockLengthSize;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_writeTundraPageHeader()
 *
 *	This procedure fills in a Tundra Page Header.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pPageName:		ptr to null-terminated string
 *					containing name of page
 *		pageWidth:		page width in pixels (not including
 *					frame)
 *		pageHeight:		page width in pixels (not including
 *					frame)
 *		posBlockLength:		length of position block portion
 *					of page
 *		pPageHeader:		pointer to start of Tundra page
 *		bufSize:		number of bytes available to
 *					write into.
 *		pBytesUsed:		number of bytes used.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL or
 *					the dimensions are too big.
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_writeTundraPageHeader (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    char       *pPageName,
    UInt32	pageWidth,
    UInt32	pageHeight,
    UInt32	posBlockLength,
    UInt8      *pPageHeader,
    UInt32	bufSize,
    UInt32     *pBytesUsed)
{
    PROLOGUE;
    UInt8      *pLocalBuf = pPageHeader;
    UInt32	headerLength;
    

	/* check to see if the pointers are ok. */
    if ((pPageHeader == NULL) || (pBytesUsed == NULL) ||
	(pPageName == NULL))
	return ia_invalidParm;

	/* if width and height are too big, complain */
    if ((pageWidth > cTundraMaxPageWidth) ||
	(pageHeight > cTundraMaxPageHeight))
	return ia_invalidParm;

	/* if the number bytes in the buffer won't hold the field,
	 * complain.
	 */
    __error__ = i_getTundraPageHeaderSize (pIPState, pTextCompInfoIn,
				         pPageName,
				         &headerLength);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    if (bufSize < headerLength)
	return ia_bufferExceeded;

	/* Yea, it fits.  Write the page name, then the width, height
	 * and position block length in that order.
	 */
    strcpy((char *) pLocalBuf, pPageName);
    pLocalBuf += strlen(pPageName) + 1;		/* skip over string including
						 * the null at the end */

    *pLocalBuf++ = (UInt8) (pageWidth >> 8);
    *pLocalBuf++ = (UInt8) (pageWidth);

    *pLocalBuf++ = (UInt8) (pageHeight >> 8);
    *pLocalBuf++ = (UInt8) (pageHeight);

    *pLocalBuf++ = (UInt8) (posBlockLength >> 24);
    *pLocalBuf++ = (UInt8) (posBlockLength >> 16);
    *pLocalBuf++ = (UInt8) (posBlockLength >> 8);
    *pLocalBuf++ = (UInt8) (posBlockLength);

    *pBytesUsed = pLocalBuf - pPageHeader;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_writeTundraResidualHeader()
 *
 *	This procedure fills in a Tundra Residual Header.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		residualX:		x position of left edge of residual
 *					image relative to bitmap (no
 *					coord includes frame).
 *		residualY:		Y position of top edge of residual
 *					image relative to bitmap (no
 *					coord includes frame).
 *		residualWidth:		width of residual image not
 *					including frame.
 *		residualHeight:		height of residual image not
 *					including frame.
 *		residualLength:		length of residual block portion
 *					of page in bytes
 *		pResidualHeader:	pointer to start of Tundra residual
 *					block
 *		bufSize:		number of bytes available to
 *					write into.
 *		pBytesUsed:		number of bytes used.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL or
 *					the dimensions are too big.
 *		ia_bufferExceeded	field is bigger than room available.
 *
 **************************************************************************/
Int32  CDECL 
i_writeTundraResidualHeader (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt32	residualX,
    UInt32	residualY,
    UInt32	residualWidth,
    UInt32	residualHeight,
    UInt32	residualLength,
    UInt8      *pResidualHeader,
    UInt32	bufSize,
    UInt32     *pBytesUsed)
{
    PROLOGUE;
    UInt8      *pLocalBuf = pResidualHeader;
    UInt32	headerLength;
    

	/* check to see if the pointers are ok. */
    if ((pResidualHeader == NULL) || (pBytesUsed == NULL))
	return ia_invalidParm;

	/* if coords or width or height are too big, complain */
    if ((residualX > cTundraMaxResidualLeft) ||
	(residualY > cTundraMaxResidualTop) ||
	(residualWidth > cTundraMaxResidualWidth) ||
	(residualHeight > cTundraMaxResidualHeight))
	return ia_invalidParm;

	/* if the number bytes in the buffer won't hold the field,
	 * complain.
	 */
    __error__ = i_getTundraResidualHeaderSize (pIPState, pTextCompInfoIn,
					     &headerLength);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    if (bufSize < headerLength)
	return ia_bufferExceeded;

	/* Yea, it fits.  Write the block's x, y, width, height
	 * and position block length in that order.
	 */
    *pLocalBuf++ = (UInt8) (residualX >> 8);
    *pLocalBuf++ = (UInt8) (residualX);

    *pLocalBuf++ = (UInt8) (residualY >> 8);
    *pLocalBuf++ = (UInt8) (residualY);

    *pLocalBuf++ = (UInt8) (residualWidth >> 8);
    *pLocalBuf++ = (UInt8) (residualWidth);

    *pLocalBuf++ = (UInt8) (residualHeight >> 8);
    *pLocalBuf++ = (UInt8) (residualHeight);

    *pLocalBuf++ = (UInt8) (residualLength >> 24);
    *pLocalBuf++ = (UInt8) (residualLength >> 16);
    *pLocalBuf++ = (UInt8) (residualLength >> 8);
    *pLocalBuf++ = (UInt8) (residualLength);

    *pBytesUsed = pLocalBuf - pResidualHeader;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_estimateTundraDictionaryBlockSize()
 *
 *	This procedure estimates the number of bytes needed to hold a
 *	Tundra dictionary block.  This is only an estimate, clients should
 *	be prepared to try again with a larger buffer if the dictionary
 *	does not fit in this block.  This routine is called only
 *	during encoding.  During decompression, the dictionary size
 *	is encoded in the Tundra file.  This routine should be called
 *	after all symbols have been found and consolidated (if necessary).
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		dictArray:		pointer to array of booleans, one
 *					per possible dictionary.  If a
 *					dictionary's boolean is set,
 *					symbols from that dictionary are
 *					included in the output dictionary.
 *					If this pointer is NULL, all
 *					symbols are written.
 *		dictArrayLen:		number of booleans in dictArray.
 *		pDictBlockSize:		pointer to UInt32 that will contain
 *					the estimate of the number of
 *					bytes required to hold the
 *					Tundra dictionary block.
 *		pageRequest:		Requested page ID for residual dict
 *					(or 0 for all non-residual symbols)
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_estimateTundraDictionaryBlockSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Bool       *dictArray,
    UInt32	dictArrayLen,
    UInt32	pageRequest,
    UInt32     *pDictBlockSize)
{
	/* This is a hack for now.  We're really putting out
	 * the old fashioned serialized compression info for
	 * the time being.  We'll replace this section as we
	 * go through adding residues, replacing the position
	 * blocks and so on.
	 */
	/* Keep the compiler happy */
    PROLOGUE;
    dictArray = dictArray;
    dictArrayLen = dictArrayLen;
    pageRequest = pageRequest;

    ERR_PASS(i_estimateSizeOfSerializedCompressionInfo (pIPState,
							pTextCompInfoIn,
							pDictBlockSize));

    NORMAL_RETURN;
}


/* This is a big number used to mark the right coord of the second dummy
 * run at the end of each scan line in the CCITT run array.
 */
#define	cDummyMarker	0x7FFFFFFF

/* PUBLIC */
/**************************************************************************
 *   i_writeTundraDictionaryData()
 *
 *	This procedure converts some subset of the current set
 *	of merged dictionaries into the data portion of a single
 *	Tundra dictionary block.
 *	The subset is specified by an array of booleans, one for
 *	each possible dictionary.  If a boolean is set, symbols from
 *	that dictionary are included in the output.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		dictArray:		pointer to array of booleans, one
 *					per possible dictionary.  If a
 *					dictionary's boolean is set,
 *					symbols from that dictionary are
 *					included in the output dictionary.
 *					If this pointer is NULL, all
 *					symbols are written.
 *					If this pointer is not NULL, the
 *					array must have one entry for
 *					every possible dictionary entry as
 *					specified by i_getDictionaryNumbers.
 *		dictArrayLen:		number of booleans in dictArray.
 *		pDictBuf:		pointer to block of memory to be
 *					used for output dictionary.
 *		dictBufSize:		number of bytes in the output
 *					buffer.
 *		pageRequest:		Page request: 0 = all pages, counts > 1
 *						>0 = specific page, counts = 1
 *		pBytesUsed:		pointer to number of bytes actually
 *					filled in the dictionary buffer.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL or
 *					dictArray exists but doesn't cover
 *					all possible dictionaries present
 *					in the internal info.
 *
 **************************************************************************/
Int32  CDECL 
i_writeTundraDictionaryData (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Bool       *dictArray,
    Int32	dictArrayLen,
    UInt8      *pDictBuf,
    Int32	dictBufSize,
    UInt32	pageRequest,
    UInt32     *pBytesUsed)
{
    PROLOGUE;
    TCCompressionInfo	*pTextCompInfo = (TCCompressionInfo *)pTextCompInfoIn;
    CCITTCompressionInfo *pCCITTCompInfo = NULL;
					/* the struct used by ip_compMMRRow */
    Int32		 ccittInfoSize;	/* size of CCITTCompressionInfo */
    Int32		 temp;		/* scratch var */
    Int32		 maxWidth;	/* maximum width of any of the
					 * height class bitmaps */
    Int32		 curWidth;	/* width of current height class
					 * bitmap */
    Int32		 prevHeight;	/* height of previous height class */
    UInt8		*pLocOut = pDictBuf;
					/* local ptr to output buf.  This is
					 * initialized here so we can always
					 * compute the correct number of 
					 * bytes written. */
    UInt8		*pCBitmap;	/* points to location at which
					 * compressed bitmap is written */
    UInt8		*pLocEnd;	/* pts to end of output buf */
    TCSymbol		*p1stHeightSym;	/* pts to first symbol of current
					 * height class */
    TCSymbol		*pNextHeightSym; /* pts to first symbol of next
					  * height class. */
    TCSymbol		*pSymbol;	/* scratch symbol ptr */
    TCScanLineRun	*pSymRun;	/* ptr to next run from symbol */
    Int32		 lineTag;	/* used to identify residual runs
					 * that are on a single scanline */
    Run32		*pCodingRuns;	/* Pointer to runs for coding
					 * line. */
    Run32		*pRefRuns; 	/* pointer to runs for reference
					 * line in array of runs. */
    Run32		*pDummy;	/* scratch ptr */
    UInt32		 accum;		/* bit accumulator */
    Int32		 vacantBits;	/* number of vacant bits on
					 * right end of accum */
    TCTundraEncTable	*pWidthHeightEncoder; /* ptr to width/height encoder */
    TCTundraEncTable	*pUseEncoder;	/* ptr to use count encoder */
    TCTundraEncClass	*pEncClass;	/* scratch encoding class ptr */
    
#ifdef DDEBUG
    Int32	curHeight;
#endif


	/* First, check the arguments */
    if ((pDictBuf == NULL) || (pBytesUsed == NULL))
	return ia_invalidParm;

	/* If the array of dictionaries exists and is too short, complain. */
    if ((dictArray != NULL) &&
	(dictArrayLen <= pTextCompInfo->maxDictionaryNum))
	return ia_invalidParm;

	/* Sort the dictionary entries into proper dictionary order
	 * if they're not already sorted.
	 */
    if (!(pTextCompInfo->encodeState & cTCSymsAreInDictOrder))
    {
	__error__ = ip_sortTCSymbolsInTundraDictOrder (pIPState,
						       pTextCompInfo,
						       cFalse);
	if (__error__ != ia_successful)
	    goto cleanupWriteTundraDict;
    }
#ifdef DDDEBUG
    printf("%d entries in Tundra dict\n", pTextCompInfo->nextSymbolNum);
#endif


	/* Go count the symbols that will be written out and link them
	 * together directly.  This allows us to avoid looking at
	 * symbols we don't need later on.  First, go find the
	 * first symbol we want.
	 */
    if (pageRequest == 0)
    {
	for (pSymbol = pTextCompInfo->p1stDictSymbol;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextDictSym)
	{
	    if ((pSymbol->instanceCount > 1) &&
		((dictArray == NULL) || (dictArray[pSymbol->dictNum])))
		break;
	}
    }
    else
    {
	for (pSymbol = pTextCompInfo->p1stDictSymbol;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextDictSym)
	{
	    if ((pSymbol->instanceCount == 1) &&
		((dictArray == NULL) || (dictArray[pSymbol->dictNum])) &&
		(pSymbol->pageID == pageRequest))
		break;
	}
    }

	/* If there were no symbols to put out, we can skip all the
	 * code that initializes things and reverses the symbol runs.
	 * We can't quit now since we still have to put out the
	 * dictionary boilerplate.
	 */
    if (pSymbol != NULL)
    {
	    /* There is at least one symbol, make it the first symbol */
	p1stHeightSym = pSymbol;
    
	    /* Initialize the count of symbols, the current height class
	     * width and the max height class width.  Also reverse the
	     * runs on the new symbol and convert them to 1x,
	     * absolute coords relative to the start of this height
	     * class.  Doing so means that we can just copy the runs
	     * from the symbol to the CCITT run array directly later
	     * on.
	     */
	temp = 1;			/* num of symbols */
	__error__ = ip_reverseTCSymbolRuns (pIPState, pTextCompInfo,
					  pSymbol,
					  0,	/* xOffset in height class */
					  0,	/* yOffset in height class */
					  &accum);
	if (__error__ != ia_successful)
	    goto cleanupWriteTundraDict;
	curWidth = pSymbol->width;
	maxWidth = curWidth;
	prevHeight = pSymbol->height - 1;
    
	    /* And store the pointer to the first symbol
	     * in pSymbol->pRunTail. */
	pSymbol->pRunTail = pSymbol->pRunHead;
    
	    /* Go find the rest of the symbols.  Reverse the runs on each one
	     * and find the width of the widest heigth class' bitmap.
	     */
	for (pNextHeightSym = pSymbol->pNextDictSym, temp = 1;
	     pNextHeightSym != NULL;
	     pNextHeightSym = pNextHeightSym->pNextDictSym)
	{
	    if (((dictArray == NULL) || (dictArray[pNextHeightSym->dictNum]))
		&&
		((pageRequest == 0 && pNextHeightSym->instanceCount > 1) ||
		(pageRequest > 0 && pageRequest == pNextHeightSym->pageID &&
		 pNextHeightSym->instanceCount == 1)))
	    {
#ifdef DDDEBUG
		printf("ID %d has count %d\n", pSymbol->symbolID, pSymbol->instanceCount);
#endif

		    /* Link this symbol with the previous one and count it */
		pSymbol->pNewSymbol = pNextHeightSym;
		pSymbol = pNextHeightSym;
		temp++;				/* increment symbol count */
		    /* if this symbol is in a new height class, start the
		     * width counter over again.  Note that if the runs
		     * of this symbol have not already been reversed,
		     * pSymbol->maxY = symbolHeight-1 because the 2x coords
		     * of the symbol range from -(symbolHeight-1) to
		     * symbolHeight-1 in steps of 2.  If the runs of
		     * this symbol have already been reversed, we assume
		     * it was done by a process that set the yOffset to
		     * 0 and pSymbol->maxY still equals symbolHeight-1.
		     * This is because the reversed coords range
		     * from 0 to symbolHeight-1 in steps of 1).
		     */
		if (pSymbol->maxY != prevHeight)	/* new height class */
		{
		    if (curWidth > maxWidth)
			maxWidth = curWidth;
		    curWidth = 0;
		    prevHeight = pSymbol->height - 1;
		}
    
		    /* Reverse this symbol's runs with the updated curWidth */
		__error__ = ip_reverseTCSymbolRuns (pIPState, pTextCompInfo,
						  pSymbol,
						  curWidth,	/* xOffset */
						  0,		/* yOffset */
						  &accum);
		if (__error__ != ia_successful)
		    goto cleanupWriteTundraDict;
		    /* update width of this height class.  Note that
		     * since the last symbol went from curWidth to
		     * curWidth + symbolWidth - 1, the new width of
		     * the entire bitmap is pSymbol->maxX (curWidth
		     * + symbolWidth - 1) + 1.
		     */
		pSymbol->pRunTail = pSymbol->pRunHead;
		curWidth = pSymbol->maxX + 1;
	    }
	}
    
	    /* Set the last symbol's pNewSymbol ptr to NULL so we
	     * terminate properly when following the list of
	     * dictionary symbols. */
	pSymbol->pNewSymbol = NULL;
    
	    /* Also see if the last height class was the widest */
	if (curWidth > maxWidth)
	    maxWidth = curWidth;
    
	    /* We're going to copy runs from the symbols directly into the run
	     * array used by the CCITT Group 4 compression code.  This
	     * is a faster way of generating G4 code for the symbols
	     * (and uses a lot less space) than converting the
	     * symbols into bitmaps, then encoding the bitmaps.
	     * Get the size of the CCITT compression info structure
	     * to be used by ip_compMMRRow.  The size of this
	     * structure is a function of the width of the widest
	     * height class bitmap, so we can't just allocate one as an
	     * automatic variable.
	     */
	__error__ = i_getSizeOfCCITTCompressionInfo (pIPState,
						   maxWidth + 2*cFrameBits,
						   &ccittInfoSize);
	if (__error__ != ia_successful)
	    goto cleanupWriteTundraDict;
    
	pCCITTCompInfo = (CCITTCompressionInfo *) MALLOC(ccittInfoSize);
	if (pCCITTCompInfo == NULL)
	{
	    __error__ = ia_nomem;
	    goto cleanupWriteTundraDict;
	}
    
	    /* We will reinitialize the CCITT block for each individual
	     * height class bitmap since they're all different widths.
	     */
    }	/* end of case where there really are symbols to write out */
    else	/* there are no symbols to write out */
    {
	temp = 0;
	p1stHeightSym = NULL;
    }

	/* Prepare to start writing stuff out. */
    accum = 0;				/* bit accumulator is empty */
    vacantBits = 32;
    pLocEnd = pLocOut + dictBufSize;	/* used to detect end of buffer */
	/* The encoder structures are used to encode the width, height
	 * and use counts associated with the symbols.
	 */
    pWidthHeightEncoder = &widthHeightEncTable0;
    if (pageRequest)
    {
	pUseEncoder = &useEncTable0;
    }
    else
    {
	pUseEncoder = &useEncTable1;
    }

#ifdef DDDEBUG
    printf("%d Tundra dict entries\n", temp);
#endif

	/* Write out the number of symbols in this dictionary */
    __error__ = ip_encodeTundraInteger (pIPState, (UInt32) temp,
				      &accum, &vacantBits,
				      &pLocOut, pLocEnd);
    if (__error__ != ia_successful)
	goto cleanupWriteTundraDict;

	/* Make room for the 1-bit flag saying that we're going to
	 * use UseCount encoding table 1.  We use table 1 because
	 * we ALWAYS write out a residual image, right?  Table 1
	 * has no way of encoding use counts of 0 or 1.
	 * First, make room for the bit if necessary.
	 */
    if (vacantBits <= (32-16))
    {
	if (pLocOut >= pLocEnd)
	{
	    __error__ = ia_bufferExceeded;
	    goto cleanupWriteTundraDict;
	}

	    /* Write out 2 bytes */
	pLocOut[0] = (UInt8) (accum >> 24);
	pLocOut[1] = (UInt8) (accum >> 16);
	pLocOut += 2;
	accum <<= 16;
	vacantBits += 16;
    }

	/* insert the use count encoding table flag */
    vacantBits -= 1;
    if (pUseEncoder == &useEncTable1)
    {
        accum |= 1 << vacantBits;	/* use table 1 since we'll put
					 * out a residual image */
    }
    else
    {
        accum &= ~(1 << vacantBits);	/* use table 0 */
    }

	/***********************  ENCODE LOOP  ***********************/

	/* Here's the loop that goes through height classes */
    for (prevHeight = 0;
	 p1stHeightSym != NULL;
	 p1stHeightSym = pNextHeightSym)
    {
	    /* Put out the height difference from the previous height
	     * class.  Since this is always at least 1, we encode
	     * (newHeight - prevHeight) - 1.  The width/height encoding
	     * table has a very short code for the value "0".
	     */
	temp = (p1stHeightSym->maxY + 1) - prevHeight - 1;
	if (temp < 0)
	{
	    __error__ = ia_internal;
	    goto cleanupWriteTundraDict;
	}

#ifdef DDEBUG
	curHeight = prevHeight + temp + 1;
#endif

	if (temp > pWidthHeightEncoder->maxNormValue)
	{
	    __error__ =
		ip_encodeTundraBigVal (pIPState, pWidthHeightEncoder,
				       temp, &accum, &vacantBits,
				       &pLocOut, pLocEnd);
	    if (__error__ != ia_successful)
		goto cleanupWriteTundraDict;
	}
	else
	{
	    if (vacantBits <= 16)
	    {
		if (pLocOut >= pLocEnd)
		{
		    __error__ = ia_bufferExceeded;
		    goto cleanupWriteTundraDict;
		}
		    /* Write out 2 bytes */
		pLocOut[0] = (UInt8) (accum >> 24);
		pLocOut[1] = (UInt8) (accum >> 16);
		pLocOut += 2;
		accum <<= 16;
		vacantBits += 16;
	    }
		/* Each value is encoded by concatenating a header
		 * and (value - offset).
		 * An encoding class is a structure containing the
		 * header (in bits), the length of the header in
		 * bits, the offset and the length of the encoded
		 * (value - offset) in bits.  The classSelector
		 * is an array indexed by the value to be encoded
		 * that returns a class number for the value.  The
		 * classes are arranged in an array pointed to by
		 * pClassArray.
		 */
	    pEncClass =
		&(pWidthHeightEncoder->pClassArray[
		    pWidthHeightEncoder->pClassSelector[temp]]);
	    vacantBits -= pEncClass->bitHeaderLen;
	    accum |= pEncClass->bitHeader << vacantBits;
	    vacantBits -= pEncClass->encValLen;
	    accum |= (temp - pEncClass->offset) << vacantBits;
	}

	    /* First, find all the symbols in this height class.
	     * For each symbol, write out its delta width from
	     * the previous symbol and its use count.
	     * Also keep track of the total of all the widths since
	     * that will become the width of the class's bitmap.
	     */
	for (pNextHeightSym = p1stHeightSym,
		curWidth = 0, maxWidth = 0;
	     (pNextHeightSym != NULL) &&
	         (pNextHeightSym->maxY == p1stHeightSym->maxY);
	     pNextHeightSym = pNextHeightSym->pNewSymbol)
	{
		/* calculate this symbol's width, the
		 * delta width from the previous symbol and the
		 * total width of this height class.
		 * Remember that since we've already adjusted
		 * this symbol's runs in the height class bitmap,
		 * pNextHeightSym->maxX + 1 equals the total width of
		 * the height class bitmap through this symbol.
		 */
	    temp = pNextHeightSym->maxX + 1 - maxWidth;	/* symbol's width */
	    maxWidth += temp;				/* total bitmap width */
	    temp -= curWidth;				/* deltaWidth */
	    curWidth += temp;				/* current width */

		/* Encode the delta width */
	    if (temp > pWidthHeightEncoder->maxNormValue)
	    {
		__error__ =
		    ip_encodeTundraBigVal (pIPState, pWidthHeightEncoder,
					   temp, &accum, &vacantBits,
					   &pLocOut, pLocEnd);
		if (__error__ != ia_successful)
		    goto cleanupWriteTundraDict;
	    }
	    else
	    {
		if (vacantBits <= 16)
		{
		    if (pLocOut >= pLocEnd)
		    {
			__error__ = ia_bufferExceeded;
			goto cleanupWriteTundraDict;
		    }
			/* Write out 2 bytes */
		    pLocOut[0] = (UInt8) (accum >> 24);
		    pLocOut[1] = (UInt8) (accum >> 16);
		    pLocOut += 2;
		    accum <<= 16;
		    vacantBits += 16;
		}
		    /* See comments above on the encoding process */
		pEncClass =
		    &(pWidthHeightEncoder->pClassArray[
			pWidthHeightEncoder->pClassSelector[temp]]);
		vacantBits -= pEncClass->bitHeaderLen;
		accum |= pEncClass->bitHeader << vacantBits;
		vacantBits -= pEncClass->encValLen;
		accum |= (temp - pEncClass->offset) << vacantBits;
	    }

		/* Encode the use count for this symbol */
	    temp = pNextHeightSym->instanceCount;

#ifdef DDDEBUG
	printf("%dx%d count = %d (ID %d)\n", curWidth, curHeight, temp, pNextHeightSym->symbolID);
#endif

	    if (temp > pUseEncoder->maxNormValue)
	    {
		__error__ =
		    ip_encodeTundraBigVal (pIPState, pUseEncoder,
					   temp, &accum, &vacantBits,
					   &pLocOut, pLocEnd);
		if (__error__ != ia_successful)
		    goto cleanupWriteTundraDict;
	    }
	    else
	    {
		if (vacantBits <= 16)
		{
		    if (pLocOut >= pLocEnd)
		    {
			__error__ = ia_bufferExceeded;
			goto cleanupWriteTundraDict;
		    }
			/* Write out 2 bytes */
		    pLocOut[0] = (UInt8) (accum >> 24);
		    pLocOut[1] = (UInt8) (accum >> 16);
		    pLocOut += 2;
		    accum <<= 16;
		    vacantBits += 16;
		}
		if (pUseEncoder == &useEncTable1)
		{
		    /* See comments above on the encoding process */
		    if (temp < 2)
		    {
			__error__ = ia_internal;	/* no residuals! */
			goto cleanupWriteTundraDict;
		    }
		    pEncClass =	&(pUseEncoder->pClassArray[
			pUseEncoder->pClassSelector[temp - 2]]);
		    vacantBits -= pEncClass->bitHeaderLen;
		    accum |= pEncClass->bitHeader << vacantBits;
		    vacantBits -= pEncClass->encValLen;
		    accum |= (temp - pEncClass->offset) << vacantBits;
		}
		else
		{
		    if (temp != 1)
		    {
			__error__ = ia_internal;	/* residuals only! */
			goto cleanupWriteTundraDict;
		    }

		    /* We have the choice of encoding a 0 or a 1, 0 means
		     * that the symbol occurs once, and can be discarded
		     * when first encountered during decode. However, a
		     * 0 encodes with 3 bits, and 1 encodes with only 1.
		     * Use [temp - 1] to use zero. */
		    pEncClass =	&(pUseEncoder->pClassArray[
			pUseEncoder->pClassSelector[temp]]);
		    vacantBits -= pEncClass->bitHeaderLen;
		    accum |= pEncClass->bitHeader << vacantBits;
		    vacantBits -= pEncClass->encValLen;
		    accum |= (temp - pEncClass->offset) << vacantBits;
		}
	    }
	}

	    /* We've gone through all the symbols in the current
	     * height class and written the widths and use counts
	     * of each to the output stream.
	     * At this point, maxWidth has the width of the height
	     * class bitmap and pNextHeightSym points to NULL or
	     * to the first symbol of the next Height class.
	     * Encode the width END code to signal that the
	     * width/useCount section is finished.  first, make
	     * room for it.
	     */
	if (vacantBits <= 16)
	{
	    if (pLocOut >= pLocEnd)
	    {
		__error__ = ia_bufferExceeded;
		goto cleanupWriteTundraDict;
	    }
		/* Write out 2 bytes */
	    pLocOut[0] = (UInt8) (accum >> 24);
	    pLocOut[1] = (UInt8) (accum >> 16);
	    pLocOut += 2;
	    accum <<= 16;
	    vacantBits += 16;
	}
	    /* encode the END code */
	vacantBits -= cTundraWidthHeightEndCodeLen;
	accum |= cTundraWidthHeightEndCode << vacantBits;

	    /* The following is an ugly feature of the encoding
	     * standard.  We're supposed to encode the length
	     * of the compressed bitmap followed by the compressed
	     * height class bitmap.  But we don't know the
	     * size of the compressed bitmap until we do the
	     * compression.  We'd rather not compress the bitmap
	     * to some other buffer, figure out how big it is,
	     * write the length code, then copy the bitmap back
	     * here.  We'll try a trick suggested by William
	     * Rucklidge.  We reserve 39 bits for the count since
	     * all counts will fit in that space, then start encoding
	     * the bitmap on the next byte boundary.  The
	     * next byte boundary starts at
	     * pLocOut +
	     * ((32 - vacantBits) +	; the bits now in accum
	     *   39 +			; bits reserved for bytecount
	     *   7) / 8			; round up to next byte boundary.
	     *
	     * The standard says that if the compressed bitmap is
	     * bigger than the original (can happen with very high
	     * frequency stipple or halftone), we can optionally
	     * just encode the original bitmap at this point.  We'll
	     * ignore that option for now.
	     *
	     * So we'll start encoding at that point.  When we're
	     * done, we'll come back here and write the count the
	     * hard way.
	     */
	pCBitmap = pLocOut + (32 - vacantBits + 39 + 7) / 8;

	    /* But first, we need to initialize the CCITT compression
	     * data block for a bitmap of this width.
	     */
	__error__ = i_initializeCCITTCompressionInfo (pIPState,
						    (void *)pCCITTCompInfo,
						    maxWidth + 2*cFrameBits,
						    cFalse, 0,
						    cNoPad, ccittInfoSize);
	if (__error__ != ia_successful)
	    goto cleanupWriteTundraDict;

	    /* Set the CCITT run array pointers to the second run in each array.
	     * The first run in each array is a dummy.
	     * Since we know that run2ArrayIsCoding is always set to false,
	     * i.e. the first run array is the initial coding array,
	     * just point to the second array.
	     */
	pCodingRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
					  pCCITTCompInfo->runArrayOffset +
					  sizeof(Run32));
	pRefRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				       pCCITTCompInfo->run2ArrayOffset +
				       sizeof(Run32));
	pDummy = pCodingRuns;

	    /* Also initialize the reference run to have a blank line.
	     * This is done by putting in only the dummy elements.
	     */
	pRefRuns[0].runLeft = maxWidth;
	pRefRuns[0].runRight = maxWidth-1;
	pRefRuns[1].runLeft = maxWidth;
	pRefRuns[1].runRight = cDummyMarker;

	    /* This is the loop that goes through the rows of the
	     * symbols in this height class.  Note that when the
	     * loop exits it leaves prevHeight equal to the height
	     * of this height class, which is used to compute the
	     * height delta the next time through the loop.
	     */
	for (prevHeight = 0;
	     prevHeight <= p1stHeightSym->maxY;
	     prevHeight++)
	{
		/* Here's the loop that goes through one row all of the
		 * symbols.  First, we copy all the runs for all the
		 * symbols into the MMR run array, then we end it,
		 * Then we ask ip_compMMRRow to do the encoding.
		 */
	    for (pSymbol = p1stHeightSym;
		 pSymbol != pNextHeightSym;
		 pSymbol = pSymbol->pNewSymbol)
	    {
	        pSymRun = pSymbol->pRunTail;
		lineTag =  (Int32) (pSymRun->pClient);

		    /* Copy the first run.  There's often only
		     * one run, so do this as efficiently as possible.
		     * Note that if this run abuts the previous run,
		     * just increase the right endpoint of the
		     * previous run.
		     */
		if ((pDummy[-1].runRight < 0) ||
		    (pDummy[-1].runRight < pSymRun->runLeft - 1))
		{
		    pDummy->runLeft = pSymRun->runLeft;
		    pDummy->runRight = pSymRun->runRight;
		    pDummy++;
		}
		else
		{
		    pDummy[-1].runRight = pSymRun->runRight;
		}
		pSymRun = pSymRun->pNext;

		    /* if there are any more runs, copy those too.
		     * Since we're past the first run, we don't
		     * have to worry about the left edge of the bitmap
		     */
		while ((pSymRun != NULL) && 
		       (pSymRun->pClient == (void *)lineTag))
		{
		    if (pDummy[-1].runRight < pSymRun->runLeft - 1)
		    {
			pDummy->runLeft = pSymRun->runLeft;
			pDummy->runRight = pSymRun->runRight;
			pDummy++;
		    }
		    else
		    {
			pDummy[-1].runRight = pSymRun->runRight;
		    }
		    pSymRun = pSymRun->pNext;
		}
		    /* remember ptr to first run of next line */
		pSymbol->pRunTail = pSymRun;
	    }

		/* add the dummy elements to the end of the scan line */
	    pDummy[0].runLeft = maxWidth;
	    pDummy[0].runRight = maxWidth-1;
	    pDummy[1].runLeft = maxWidth;
	    pDummy[1].runRight = cDummyMarker;
    
		/* Just to be extra safe, set the run offset in
		 * the compression info to 0 so that we start
		 * interpreting the very first run and not some
		 * run left over from some previous stoppage.
		 * I'm not really sure when this does anything
		 * that wasn't already done, but it seems tidy.
		 */
	    pCCITTCompInfo->currentRun = 0;
	    pCCITTCompInfo->refRun = 0;
    
		/* Ask the compression code to translate runs into the
		 * compressed data stream.  The width sent down is used
		 * to decide when we've coded the whole line.
		 */
	    __error__ = ip_compMMRRow (pIPState, pCCITTCompInfo, maxWidth,
				     &pCBitmap, pLocEnd);
	    if (__error__ != ia_successful)
		goto cleanupWriteTundraDict;
    
		/* Do the normal stuff that gets us to the next scan line.
		 * Note that we leave pDummy pointing to pCodingRuns,
		 * which is where the next runs should go.
		 */
	    pCCITTCompInfo->lineNumber++;
	    pCCITTCompInfo->run2ArrayIsCoding ^= cTrue;
	    pDummy = pRefRuns;
	    pRefRuns = pCodingRuns;
	    pCodingRuns = pDummy;
	}

	    /* So far, so good.  We've touched all the runs of the
	     * symbols in this height class.
	     * Flush the remaining bytes in the
	     * compression pipeline to the output buffer.
	     */
	__error__ = i_flushCCITT (pIPState, (void *)pCCITTCompInfo,
			        pCBitmap,
			        pLocEnd - pCBitmap,
			        (Int32 *)&temp);
	if (__error__ != ia_successful)
	    goto cleanupWriteTundraDict;

	    /* Update pCBitmap to account for the flushed data */
	pCBitmap += temp;

	    /* Swell, we've written out the compressed bitmap for this
	     * height class.  Write out it's length in bytes way back
	     * up at the beginning of the class.  This length
	     * is pCBitmap - (the big pile of glop used to initialize
	     * pCBitmap above).
	     */
	temp = pCBitmap - (pLocOut + (32 - vacantBits + 39 + 7) / 8);

	    /* We never have room in the accumulator for a 39 bit
	     * thing.  First write out all possible bytes */
	while (vacantBits < 24)
	{
	    if (pLocOut > pLocEnd)
	    {
		__error__ = ia_bufferExceeded;
		goto cleanupWriteTundraDict;
	    }

	    *pLocOut++ = (UInt8) (accum >> 24);
	    accum <<= 8;
	    vacantBits += 8;
	}
	    /* Now insert the header */
	vacantBits -= cTundraTable3Class3HeaderLen;
	accum |= (cTundraTable3Class3Header << vacantBits);

	    /* and as much of the 32 bit value as fits */
	accum |= temp >> (32 - vacantBits);

	    /* Try to write out 4 bytes */
	if (pLocOut >= pLocEnd - 4)
	{
	    __error__ = ia_bufferExceeded;
	    goto cleanupWriteTundraDict;
	}

	pLocOut[0] = (UInt8) (accum >> 24);
	pLocOut[1] = (UInt8) (accum >> 16);
	pLocOut[2] = (UInt8) (accum >> 8);
	pLocOut[3] = (UInt8) (accum);
	pLocOut += 4;

	    /* Insert the remainder of the code and flush it out.
	     * remember that we've already written the compressed
	     * data at the end of the count.
	     */
	accum = temp << vacantBits;
	while (vacantBits < 32)
	{
	    if (pLocOut >= pLocEnd)
	    {
		__error__ = ia_bufferExceeded;
		goto cleanupWriteTundraDict;
	    }

	    *pLocOut++ = (UInt8) (accum >> 24);
	    accum <<= 8;
	    vacantBits += 8;
	}

	    /* We've now written out a complete height class.  Update
	     * the pLocOut pointer so it points just past the
	     * compressed data and loop back for another height class.
	     * Note that we start on another byte boundary so
	     * the accumulator should be emptied.
	     */
	pLocOut = pCBitmap;
	accum = 0;
	vacantBits = 32;
    }	 /* end of loop that writes out height classes */

	/* We've written out all the height classes.  Encode the
	 * end code.  Since we've just emptied the accumulator,
	 * we don't have to check to see if we should flush it
	 * now.
	 */
    vacantBits -= cTundraWidthHeightEndCodeLen;
    accum |= cTundraWidthHeightEndCode << vacantBits;

	/* Encode the length of the dictionary extension block, which
	 * is 0.
	 */
    __error__ = ip_encodeTundraInteger (pIPState, 0UL,
				      &accum, &vacantBits,
				      &pLocOut, pLocEnd);
    if (__error__ != ia_successful)
	goto cleanupWriteTundraDict;

	/* flush the last bytes of the accumulator out to the buffer
	 * and return happily.
	 */
    while (vacantBits < 32)
    {
	if (pLocOut >= pLocEnd)
	{
	    __error__ = ia_bufferExceeded;
	    goto cleanupWriteTundraDict;
	}

	*pLocOut++ = (UInt8) (accum >> 24);
	accum <<= 8;
	vacantBits += 8;
    }

    __error__ = ia_successful;

cleanupWriteTundraDict:
    if (pCCITTCompInfo != NULL)
	FREE(pCCITTCompInfo);

	/* If we're returning happily, say how many bytes were used */
    if (__error__ == ia_successful)
	*pBytesUsed = pLocOut - pDictBuf;

    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *   i_estimateTundraPositionBlockSize()
 *
 *	This procedure estimates the number of bytes needed to hold a
 *	Tundra Position block.  This is only an estimate, clients should
 *	be prepared to try again with a larger buffer if the page
 *	does not fit in this block.  A position block includes only
 *	position info, no residual images.  It also does not contain the
 *	page header.  This strange division is made on the assumption
 *	that the format of the position block will not change over
 *	time while the format of the page header and relative
 *	locations of the position and residual blocks will change as
 *	the encapsulation changes.
 *
 *	The estimate of the position buffer size is based on
 *	an internal representation of a particular page's symbols and
 *	their positions, called the input position buffer.  This internal
 *	block was produced by i_convertImagesToSymbols and possibly 
 *	modified by i_remapSymbolIDs.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pInputBuf:		pointer to internal page info
 *		inputBufSize:		size of buffer holding internal
 *					page info in bytes.
 *		pageRequest:		Requested page ID for residual dict
 *					(or 0 for all non-residual symbols)
 *
 *	Output:
 *		pOutputBlockSize:	pointer to UInt32 that will contain
 *					the estimate of the number of
 *					bytes required to hold the
 *					Tundra position block.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_estimateTundraPositionBlockSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pInputBuf,
    UInt32	inputBufSize,
    UInt32	pageRequest,
    UInt32     *pOutputBlockSize)
{

    PROLOGUE;
    TCCompressionInfo	*pTextCompInfo = (TCCompressionInfo *)pTextCompInfoIn;
    

	/* Check the pointers */
    if ((pTextCompInfo == NULL) || (pInputBuf == NULL) ||
	(pOutputBlockSize == NULL))
	return ia_invalidParm;

	/* First, see if the statistics on the position buffer already
	 * exist.  If not, make them.
	 */
    if (pInputBuf[cTCPosStatsValid] == 0)
    {
	__error__ = ip_calcPositionBufStats (pIPState, pTextCompInfo,
					   pInputBuf, inputBufSize,
					   pageRequest);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
    }

	/* Just transfer the calculated residual size. */
    *pOutputBlockSize = ((UInt32) pInputBuf[cTCPosEstPosLen0]) +
			((UInt32) pInputBuf[cTCPosEstPosLen1] << 8) +
			((UInt32) pInputBuf[cTCPosEstPosLen2] << 16) +
			((UInt32) pInputBuf[cTCPosEstPosLen3] << 24);

    NORMAL_RETURN;
}


/**************************************************************************
 *   i_estimateTundraResidualBlockSize()
 *
 *	This procedure estimates the number of bytes needed to hold a
 *	Tundra Residual block.  This is only an estimate, clients should
 *	be prepared to try again with a larger buffer if the page
 *	does not fit in this block.  A residual block does not contain the
 *	residual header.  This strange division is made on the assumption
 *	that the format of the residual block will not change over time
 *	(just Group 4) while the format of the residual header and relative
 *	locations of the position and residual blocks will change as
 *	the encapsulation changes.
 *
 *	The estimate of the residual buffer size is derived from
 *	an internal representation of a particular page's symbols and
 *	their positions, called the input position buffer.  This internal
 *	block was produced by i_convertImagesToSymbols and possibly 
 *	modified by i_remapSymbolIDs.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pInputBuf:		pointer to internal page info
 *		inputBufSize:		size of buffer holding internal
 *					page info in bytes.
 *		pOutputBlockSize:	pointer to UInt32 that will contain
 *					the estimate of the number of
 *					bytes required to hold the
 *					Tundra position block.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_estimateTundraResidualBlockSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pInputBuf,
    UInt32	inputBufSize,
    UInt32     *pOutputBlockSize)
{
    PROLOGUE;
    TCCompressionInfo	*pTextCompInfo = (TCCompressionInfo *)pTextCompInfoIn;
    

	/* Check the pointers */
    if ((pTextCompInfo == NULL) || (pInputBuf == NULL) ||
	(pOutputBlockSize == NULL))
	return ia_invalidParm;

	/* First, see if the statistics on the position buffer already
	 * exist.  If not, make them.
	 */
    if (pInputBuf[cTCPosStatsValid] == 0)
    {
	__error__ = ip_calcPositionBufStats (pIPState, pTextCompInfo,
					   pInputBuf, inputBufSize, 0);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
    }

	/* Just transfer the calculated residual size. */
    *pOutputBlockSize = ((UInt32) pInputBuf[cTCPosEstResLen0]) +
			((UInt32) pInputBuf[cTCPosEstResLen1] << 8) +
			((UInt32) pInputBuf[cTCPosEstResLen2] << 16) +
			((UInt32) pInputBuf[cTCPosEstResLen3] << 24);

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_writeTundraPositionBlock()
 *
 *	This procedure converts an internally-formatted position buffer
 *	into an official Tundra position block,  This does not include
 *	a page header or residual block.  This division is made on the
 *	assumption that the format of the position block will not change  
 *	over time while the format of the page header and relative
 *	locations of the position and residual blocks will change as 
 *	the encapsulation changes.
 *
 *	NOTE: for the time being, there is no provision to encode the
 *	page using only a subset of the dictionaries in the internal
 *	compression info.  We assume symbols from all dictionaries can
 *	be used.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pInputBuf:		pointer to internal position buffer.
 *		inputBufSize:		number of bytes in input pos buffer.
 *		pOutputBuf:		pointer to block of memory to be
 *					used for output position block.
 *		outputBufSize:		number of bytes in the output
 *					buffer.
 *		pRealOutputLength:	pointer to number of bytes actually
 *					used by position block in the output
 *					buffer.
 *		pageRequest:		Requested page ID for residual dict
 *					(or 0 for all non-residual symbols)
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL.
 *		ia_nomem		no memory for internal tables.
 *		ia_bufferExceeded	output didn't fit into pageBuf.
 *
 **************************************************************************/
Int32  CDECL 
i_writeTundraPositionBlock (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pInputBuf,
    UInt32	inputBufSize,
    UInt8      *pOutputBuf,
    UInt32	outputBufSize,
    UInt32	pageRequest,
    UInt32     *pRealOutputLength)
{
    PROLOGUE;
    TCCompressionInfoPtr pTextCompInfo;
    UInt8		*pPos0;			/* pointer to even
						 * position buffer lines. */
    UInt8		*pLocIn;		/* local input buf ptr. */
    UInt8		*pPos;			/* represents either pPos0 or
						 * pLocIn. */
    UInt8		*pLocOut;		/* local ptr to output buf. */
    UInt8		*pLocEnd;		/* ptr to last byte in output
						 * buffer */
    Int32		 temp;			/* universal scratch buf */
    Int32		 lastX;			/* previous X value */
    Int32		 lastFirstX;		/* first X val in previous
						 * strip. */
    Int32		 curX0, curX1;		/* current X position in even
						 * and odd lines */
    Int32		 curX;			/* represents either curX0
						 * or curX1 */
    Int32		 symWidth0, symWidth1;	/* widths of symbols on even
						 * and odd lines.  Well, not
						 * exactly.  These are really
						 * (width-1) since that's
						 * what is normally used here */
    Int32		 symWidth;		/* represents either symWidth0
						 * or symWidth1 */
    Int32		 curY, prevY, nextY;	/* current Y pos, Y pos of
						 * prev strip, Y pos of next
						 * strip */
    TCSymbol		*pSymbol0, *pSymbol1;	/* ptrs to symbols on even
						 * and odd strip lines */
    TCSymbol		*pSymbol;		/* represents either pSymbol0
						 * or pSymbol1 */
    Int32		 numSymbolsEncoded;	/* the number of symbols we
						 * expected to encode */
    TCSymbol	       **pSymbolTable;		/* table associating symbol IDs
						 * with symbols */
    UInt32		 accum;			/* bit accumulator */
    Int32		 vacantBits;		/* number of vacant bits on
						 * right end of accum */
    Int32		 modalDeltaX;		/* the most common value
						 * for deltaX on this page */
    TCTundraEncTable	*pFirstXEncoder;	/* ptr to first X encoder */
    TCTundraEncTable	*pDeltaXEncoder;	/* ptr to delta X encoder */
    TCTundraEncTable	*pDeltaYEncoder;	/* ptr to delta Y encoder */
    TCTundraEncClass	*pEncClass;		/* scratch encoding class ptr */
    Bool		 eofFound;		/* says we got to the end of
						 * the input position buffer */
    

	/* initialize test compression info pointer */
    pTextCompInfo = (TCCompressionInfo *)pTextCompInfoIn;

	/* check for NULL pointers */
    if ((pTextCompInfo == NULL) ||
	(pInputBuf == NULL) || (pOutputBuf == NULL) ||
	(pRealOutputLength == NULL))
	return ia_invalidParm;

	/* make sure the output buffer is big enough */
    if (outputBufSize < inputBufSize)
	return ia_bufferExceeded;

	/* Now, see if the table that associates Symbol IDs with
	 * symbols exists.  If not, make it.
	 */
    if ((pTextCompInfo->pSymTable == NULL) ||
	(pTextCompInfo->pSymTable->numPtrs != pTextCompInfo->nextSymbolNum))
    {	/* allocate and initialize a new table.  Delete the old one
	 * if it exists.
	 */
	__error__ = ip_createTCSymbolTable (pIPState, pTextCompInfo);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
    }
    pSymbolTable = &(pTextCompInfo->pSymTable->table[0]);

	/* If the stats on this position buffer haven't been
	 * accumulated yet, go get them.
	 */
    if (pInputBuf[cTCPosStatsValid] == 0)
    {
	__error__ = ip_calcPositionBufStats (pIPState, pTextCompInfo,
					   pInputBuf, inputBufSize,
					   pageRequest);
	if (__error__ != ia_successful)
	    goto cleanupWrPosBlock;
    }

	/* Assign Huffman codes to all the symbols.  We do this
	 * on each page since it's pretty fast and in the future
	 * each page will have a unique set of dictionaries.
	 */
    __error__ = ip_genSymbolHuffmanCodes (pIPState, pTextCompInfo, NULL,
					  pageRequest, cFalse);
    if (__error__ != ia_successful)
	goto cleanupWrPosBlock;

#ifdef DDEBUG
	 {
	     TCSymbol *pSym = pTextCompInfo->p1stDictSymbol;

	     printf("Dictionary symbols when encoding:\n");
	     while (pSym) {
		 printf("%3d (%3d): dict %d, %dx%d, count %d, code %d-%d\n",
		     pSym->dictOrder, pSym->symbolID, pSym->dictNum, pSym->width,
		     pSym->height, pSym->instanceCount, pSym->huffmanCode,
		     pSym->huffmanCodeLen);
		 pSym = pSym->pNextDictSym;
	     }
	     printf("\n");
	 }
#endif


	/* Parse the input, tossing all symbols with instanceCount = 1.
	 * The output is encoded from the remaining symbols.
	 */
    accum = 0;
    vacantBits = 32;
    pLocIn = pInputBuf;
    pLocOut = pOutputBuf;
    pLocEnd = pOutputBuf + outputBufSize - 1;

	/* First check the header and version. */
    temp = ((UInt32) pLocIn[cTCPosMagicNum3] << 24) |
	   ((UInt32) pLocIn[cTCPosMagicNum2] << 16) |
	   ((UInt32) pLocIn[cTCPosMagicNum1] << 8) |
	   ((UInt32) pLocIn[cTCPosMagicNum0]);

    if (temp != cTextCompPositionInfoMagicNum)
    {
	__error__ = ia_invalidParm;
	goto cleanupWrPosBlock;
    }

    temp = ((UInt32) pLocIn[cTCPosVer1] << 8) +
	   ((UInt32) pLocIn[cTCPosVer0]);

    if (temp != cTextCompPositionInfoVersionNum)
    {
	__error__ = ia_invalidParm;
	goto cleanupWrPosBlock;
    }

	/* Get the most common delta X value.  We'll need it
	 * to encode the other delta X values.
	 */
    modalDeltaX = (Int32) pLocIn[cTCPosModalX];

	/* Put out the header in the output buffer.  It starts
	 * with the number of symbols, so go get that.
	 */
	numSymbolsEncoded = ((UInt32) pLocIn[cTCPosNumSym3] << 24) |
	    ((UInt32) pLocIn[cTCPosNumSym2] << 16) |
	    ((UInt32) pLocIn[cTCPosNumSym1] << 8) |
	    ((UInt32) pLocIn[cTCPosNumSym0]);

#ifdef DDEBUG
    printf("%d Tundra position block entries\n", numSymbolsEncoded);
#endif

	/* Skip over the rest of the position buffer header. */
    pLocIn += cTCPosHeaderVer2Len;

	/* Encode this number using the infamous Table 3 in
	 * William's spec.
	 */
    __error__ = ip_encodeTundraInteger (pIPState, (UInt32) numSymbolsEncoded,
				      &accum, &vacantBits,
				      &pLocOut, pLocEnd);
    if (__error__ != ia_successful)
	goto cleanupWrPosBlock;

	/* See if there's room for 13 bits of position block header.
	 * This includes the 4 bits modal deltaX, 2 bits each for strip
	 * height, first X table, deltaX table, deltaY table and the
	 * 1-bit transposition flag.
	 */
    if (vacantBits < 13)
    {
	if (pLocOut >= pLocEnd)
	{
	    __error__ = ia_bufferExceeded;
	    goto cleanupWrPosBlock;
	}

	    /* Write out 2 bytes */
	pLocOut[0] = (UInt8) (accum >> 24);
	pLocOut[1] = (UInt8) (accum >> 16);
	pLocOut += 2;
	accum <<= 16;
	vacantBits += 16;
    }
	/* insert the new value */
    vacantBits -= 4;
    accum |= modalDeltaX << vacantBits;

	/* insert 2 bit strip size. */
    vacantBits -= 2;
    accum |= cTundraStripHeight2 << vacantBits;

	/* insert 2 bit first x table selector */
    vacantBits -= 2;
    accum |= cTundraFirstXTable0 << vacantBits;
    pFirstXEncoder = &firstXEncTable0;

	/* insert 2 bit delta x table selector */
    vacantBits -= 2;
    accum |= cTundraDeltaXTable0 << vacantBits;
    pDeltaXEncoder = &deltaXEncTable0;

	/* insert 2 bit delta y table selector */
    vacantBits -= 2;
    accum |= cTundraDeltaYTable1 << vacantBits;
    pDeltaYEncoder = &deltaYEncTable1;

	/* insert 1 bit transposition flag */
    vacantBits -= 1;
    accum |= cTundraNoTransposition << vacantBits;


	/************************* STRIP LOOP **************************/

	/* Here's where the real fun starts.  We loop through the strips,
	 * putting out codes for each non-residue symbol in each.
	 * strip.  This code is organized much like the Loop A and
	 * Loop B in ip_calcPositionBufStats below.  Please read those
	 * comments.
	 */

	/* Initialize the y position */
    prevY = -2;				/* pretend we were here */
    lastFirstX = 0;

	/* Initialize the y position.  We do this by reading the
	 * first byte or two of the position buffer.
	 */
    curY = -1;
    nextY = *pLocIn++;
    if (nextY & 0x80)
	nextY = (nextY & 0x7F) + ((*pLocIn++) << 7);

    if (nextY == 0)		/* the position buffer was empty */
	eofFound = cTrue;
    else
	eofFound = cFalse;

    curY += nextY;		/* This is the first real Y position */

	/************************ LOOP A  **************************/

	/* Here's the beginning of the outermost loop, called loop A.
	 * We execute until the end of the position block has been
	 * processed.
	 */
    for (; eofFound != cTrue; curY = nextY)
    {
	    /* We haven't seen any symbols in this strip yet.
	     */
	pSymbol0 = pSymbol1 = NULL;
	pPos0 = NULL;
	symWidth0 = symWidth1 = 0;

	if ((curY & 1) == 0)	/* current line num is even */
	{
	    curX0 = *pLocIn++;	/* get LS byte of first x offset */
	    if (curX0 & 0x80)		/* there's another byte */
	    {
		curX0 = (curX0 & 0x7F) + (*pLocIn++ << 7);
	    }
		/* Since any value of firstX is ok, we don't need to
		 * check it. */

		/* Go read the first even line symbol ID. */
	    temp = *pLocIn++;
	    if (temp & 0x80)		/* it's a 2-byte symbol ID */
	    {
		temp = (temp & 0x7F) | ((*pLocIn++) << 7);
	    }
		/* point to the symbol in question.  Save the symbol's
		 * width - 1;
		 */
	    pSymbol0 = pSymbolTable[temp];
	    symWidth0 = pSymbol0->width - 1;
	    pPos0 = pLocIn;		/* remember where next byte in the
					 * even line is so we can parse it
					 * later. */

		/* Scan to the end of the even
		 * line to see where the next line is.
		 * If it's the next scan line, we'll
		 * process it in this strip.  Otherwise, we'll
		 * process it in the next strip.
		 */
	    while (cTrue)
	    {
		temp = *pLocIn++;		/* get x coord */
		if (temp & 0x80)		/* there's another byte */
		{
		    pLocIn++;			/* skip over it */
		}
		else
		if (temp == 0)			/* reached the end of line
						 * (maybe) */
		{
		    nextY = *pLocIn++;		/* get 1st byte of y offset */
		    if (nextY & 0x80)		/* next byte is needed too */
		    {
			if (*pLocIn & 0x80)	/* BB doesn't move */
			{
				/* skip over the next X coord since it
				 * takes us right back here.
				 */
			    if (pLocIn[1] & 0x80)
				pLocIn += 3;
			    else
				pLocIn += 2;
				/* We want to keep going around the loop.
				 * We set nextY to -1 as a flag that
				 * this should happen, i.e. we didn't
				 * really find the end of the line.
				 * The next symbol ID will also
				 * be skipped at the bottom of the loop.
				 */
			    nextY = -1;
			}
			else
				/* That was a perfectly normal end of
				 * the even scan line.
				 */
			    nextY = (nextY & 0x7F) + ((*pLocIn++) << 7);
		    }
		    else
		    if (nextY == 0)		/* end of the file */
		    {
			eofFound = cTrue;	/* nextY will be updated
						 * below */
		    }

			/* Now, did we really find the end of the
			 * scan line, in which case we should stop
			 * right now, or did we only find one of those
			 * annoying "don't move the bounding box"
			 * sequences?
			 */
		    if (nextY != -1)
		    {
			    /* That was the end of the scan line. */
			nextY += curY;		/* point to next line. */
			break;			/* we found the end of the
						 * even scan line. */
		    }
		}

		    /* If we get here, there's a symbol to skip over before
		     * looping back around to find the next delta X.
		     */
		temp = *pLocIn++;
		if (temp & 0x80)		/* it's a 2-byte symbol ID */
		{
		    pLocIn++;
		}
	    }
		/* We've either found the end of a scan line or the
		 * end of the file.  If it's the end of the file,
		 * nextY == curY.
		 * If nextY == curY + 1, go read the first X offset
		 * for the odd line and its symbol.  Otherwise,
		 * say that there are no odd line symbols.
		 */
	    if (nextY == (curY + 1))
	    {
		curX1 = *pLocIn++;		/* get first x coord */
		if (curX1 & 0x80)		/* there's another byte */
		{
		    curX1 = (curX1 & 0x7F) + ((*pLocIn++) << 7);
		}
		    /* Don't we have to check to see if the odd line
		     * ended here?  No.  Any value of the firstX
		     * coord is ok..
		     */
		temp = *pLocIn++;
		if (temp & 0x80)		/* it's a 2-byte symbol ID */
		{
		    temp = (temp & 0x7F) + ((*pLocIn++) << 7);
		}
		    /* point to the symbol in question */
		pSymbol1 = pSymbolTable[temp];
		symWidth1 = pSymbol1->width - 1;
	    }
	    else	/* the next line isn't in this strip */
		curX1 = cEOLVal;
	}

	else

	{
		/* The first line is odd.  So, just go find the first
		 * symbol in that line.  Also adjust curY so that it
		 * points to the previous even line.  The code below
		 * assumes curY always points to the even line of the
		 * strip.
		 */
	    curY--;
	    curX0 = cEOLVal;
	    curX1 = *pLocIn++;		/* get LS byte of first x offset */
	    if (curX1 & 0x80)		/* there's another byte */
	    {
		curX1 = (curX1 & 0x7F) + (*pLocIn++ << 7);
	    }

		/* There's no need to see if we've reached the end of the
		 * file since any value of firstX is ok. */
	    temp = *pLocIn++;
	    if (temp & 0x80)		/* it's a 2-byte symbol ID */
	    {
		temp = (temp & 0x7F) | ((*pLocIn++) << 7);
	    }
		/* point to the symbol in question  Remember that in
		 * 2x relative coords, pSymbol->maxX = -pSymbol->minX =
		 * width - 1.
		  */
	    pSymbol1 = pSymbolTable[temp];
	    symWidth1 = pSymbol1->width - 1;
	}

	    /********************* LOOP B **************************/

	    /* Don't record distance to first sym */
	lastX = -1;

	    /* Go process the symbols in this strip */
	while (curX0 + curX1 < 2*cEOLVal)
	{
	    if (curX0 - symWidth0 <= curX1 - symWidth1)
	    {
		    /* pSymbol0's left edge is to the left of
		     * pSymbol1's left edge.
		     */
		pSymbol = pSymbol0;
		pPos = pPos0;
		curX = curX0;
		symWidth = symWidth0;
	    }
	    else
	    {
		    /* We get to the symbol on the odd line first */
		pSymbol = pSymbol1;
		pPos = pLocIn;
		curX = curX1;
		symWidth = symWidth1;
	    }

		/* See if pSymbol is a residual or a repeated symbol.
		 * If it's a repeated symbol, write its position and
		 * symbol ID to the dictionary.  Otherwise, just skip
		 * it.
		 */
	    if ((pSymbol->instanceCount > 1) ||
		(pageRequest > 0 && (pageRequest == pSymbol->pageID)))
	    {
		    /* If this is the first non-residue symbol on this
		     * strip, put out the delta Y from the previous strip.
		     * We do this here so that strips that contain no
		     * non-residual symbols get no encoded bits.
		     */
		if (lastX == -1)
		{
			/* record Y distance from previous strip and
			 * update prevY.  Note that this is one of
			 * the only safe places to update prevY since
			 * we only want to update it when we're actually
			 * putting out a new strip.
			 */
		    temp = (curY - prevY)/2;	/* divide by strip height */
		    prevY = curY;

			/* if the value is too big, encoding it using
			 * the routine for encoding big values.
			 * Otherwise, directly encode the value.
			 */
		    if (temp > pDeltaYEncoder->maxNormValue)
		    {
			__error__ =
			    ip_encodeTundraBigVal (pIPState, pDeltaYEncoder,
						   temp, &accum, &vacantBits,
						   &pLocOut, pLocEnd);
			if (__error__ != ia_successful)
			    goto cleanupWrPosBlock;
		    }
		    else
		    {
			if (vacantBits <= 16)
			{
			    if (pLocOut >= pLocEnd)
			    {
				__error__ = ia_bufferExceeded;
				goto cleanupWrPosBlock;
			    }
				/* Write out 2 bytes */
			    pLocOut[0] = (UInt8) (accum >> 24);
			    pLocOut[1] = (UInt8) (accum >> 16);
			    pLocOut += 2;
			    accum <<= 16;
			    vacantBits += 16;
			}
			pEncClass =
			    &(pDeltaYEncoder->pClassArray[
				pDeltaYEncoder->pClassSelector[temp - 1]]);
			vacantBits -= pEncClass->bitHeaderLen;
			accum |= pEncClass->bitHeader << vacantBits;
			vacantBits -= pEncClass->encValLen;
			accum |= (temp - pEncClass->offset) << vacantBits;
		    }

			/* Now encode the distance from the left edge of the
			 * page relative to the previous stripe's first
			 * symbol distance.  However, don't encode a
			 * value that puts us into negative X coords.
			 * William says he intended that there would
			 * never be negative X coords.  Therefore if
			 * curX - symWidth < 0, set temp to
			 * -lastFirstX (putting next X coord at 0).
			 * Otherwise, set temp to (curX - symWidth) -
			 * lastFirstX.  We can't just change curX
			 * because all the subsequent X positions depend
			 * on it.
			 */
		    if (curX - symWidth < 0)
			temp = -lastFirstX;
		    else
			temp = curX - symWidth - lastFirstX;
		    if ((temp > pFirstXEncoder->maxNormValue) ||
			(temp < pFirstXEncoder->minNormValue))
		    {
			__error__ = ip_encodeTundraBigVal (pIPState,
							 pFirstXEncoder,
							 temp, &accum,
							 &vacantBits,
							 &pLocOut, pLocEnd);
			if (__error__ != ia_successful)
			    goto cleanupWrPosBlock;
		    }
		    else
		    {
			if (vacantBits <= 16)
			{
			    if (pLocOut >= pLocEnd)
			    {
				__error__ = ia_bufferExceeded;
				goto cleanupWrPosBlock;
			    }
				/* Write out 2 bytes */
			    pLocOut[0] = (UInt8) (accum >> 24);
			    pLocOut[1] = (UInt8) (accum >> 16);
			    pLocOut += 2;
			    accum <<= 16;
			    vacantBits += 16;
			}
			pEncClass = &(pFirstXEncoder->pClassArray[
				    pFirstXEncoder->pClassSelector[
				     temp - pFirstXEncoder->minNormValue]]);
			vacantBits -= pEncClass->bitHeaderLen;
			accum |= pEncClass->bitHeader << vacantBits;
			vacantBits -= pEncClass->encValLen;
			accum |= (temp - pEncClass->offset) << vacantBits;
		    }
			/* Update the distance to the left edge of the
			 * first symbol in this stripe.  lastX will point
			 * to just beyond right edge of symbol since that's
			 * the kind of deltaX we want to encode.
			 */
		    lastFirstX = temp + lastFirstX;
		    lastX = curX + 1;
		}
		else	/* this is some later symbol on a line */
		{
		    temp = curX - symWidth - lastX - modalDeltaX;
		    lastX = curX + 1;

		    if ((temp > pDeltaXEncoder->maxNormValue) ||
			(temp < pDeltaXEncoder->minNormValue))
		    {
			__error__ = ip_encodeTundraBigVal (pIPState,
							 pDeltaXEncoder,
							 temp, &accum,
							 &vacantBits,
							 &pLocOut, pLocEnd);
			if (__error__ != ia_successful)
			    goto cleanupWrPosBlock;
		    }
		    else
		    {
			    /* Since some of the delta X values are
			     * bigger than 16, treat this case
			     * carefully.  We write out some bits
			     * If there's not room for the longest
			     * code.  If we can write out 2 bytes,
			     * we do.  Otherwise, we write out 1 byte.
			     */
			if (vacantBits < cTundraMaxDeltaXCodeLen)
			{
			    if (vacantBits <= 8)
			    {
				if (pLocOut >= pLocEnd - 1)
				{
				    __error__ = ia_bufferExceeded;
				    goto cleanupWrPosBlock;
				}
				    /* Write out 3 bytes */
				pLocOut[0] = (UInt8) (accum >> 24);
				pLocOut[1] = (UInt8) (accum >> 16);
				pLocOut[2] = (UInt8) (accum >> 8);
				pLocOut += 3;
				accum <<= 24;
				vacantBits += 24;
			    }
			    else	/* we can only write out 2 bytes */
			    {
				if (pLocOut >= pLocEnd)
				{
				    __error__ = ia_bufferExceeded;
				    goto cleanupWrPosBlock;
				}
				    /* Write out 2 bytes */
				pLocOut[0] = (UInt8) (accum >> 24);
				pLocOut[1] = (UInt8) (accum >> 16);
				pLocOut += 2;
				accum <<= 16;
				vacantBits += 16;
			    }
			}
			    /* Use the encoder's selector table to pick
			     * and encoding class.  Use the encoding class
			     * to encode the value.
			     */
			pEncClass = &(pDeltaXEncoder->pClassArray[
				    pDeltaXEncoder->pClassSelector[
				     temp - pDeltaXEncoder->minNormValue]]);
			vacantBits -= pEncClass->bitHeaderLen;
			accum |= pEncClass->bitHeader << vacantBits;
			vacantBits -= pEncClass->encValLen;
			accum |= (temp - pEncClass->offset) << vacantBits;
		    }
		}

		    /* Encode the single bit that gives the y position of this
		     * symbol relative to the top of the stripe.
		     */
		if (vacantBits <= 16)
		{
		    if (pLocOut >= pLocEnd)
		    {
			__error__ = ia_bufferExceeded;
			goto cleanupWrPosBlock;
		    }
			/* Write out 2 bytes */
		    pLocOut[0] = (UInt8) (accum >> 24);
		    pLocOut[1] = (UInt8) (accum >> 16);
		    pLocOut += 2;
		    accum <<= 16;
		    vacantBits += 16;
		}

		vacantBits--;		/* encode a single bit */
		if (pPos == pLocIn)	/* encoding symbol in odd line */
		    accum |= 0x1 << vacantBits;
		    /* If we're encoding a symbol in the even line, we
		     * encode a 0, which is the same as doing nothing.
		     */
    
    
#ifdef DDEBUG
		printf("(%d,%d) code = %d-%d\n", curX, curY,
		    pSymbol->huffmanCode, pSymbol->huffmanCodeLen);
#endif
		    /* Encode the huffman code for this symbol.
		     * Since Huffman codes can be as much as 31 bits
		     * in length, we have to be careful.  It might
		     * not be possible to make enough room in accum
		     * by writing out complete bytes.
		     */
		if (vacantBits < pSymbol->huffmanCodeLen)
		{
		    if (vacantBits == 0)
		    {
			if (pLocOut >= pLocEnd - 2)
			{
			    __error__ = ia_bufferExceeded;
			    goto cleanupWrPosBlock;
			}
			    /* Write out 4 bytes */
			pLocOut[0] = (UInt8) (accum >> 24);
			pLocOut[1] = (UInt8) (accum >> 16);
			pLocOut[2] = (UInt8) (accum >> 8);
			pLocOut[3] = (UInt8) (accum);
			pLocOut += 4;
			accum = 0;
			vacantBits = 32;
		    }
		    else
		    if (vacantBits <= 8)
		    {
			if (pLocOut >= pLocEnd - 1)
			{
			    __error__ = ia_bufferExceeded;
			    goto cleanupWrPosBlock;
			}
			    /* Write out 3 bytes */
			pLocOut[0] = (UInt8) (accum >> 24);
			pLocOut[1] = (UInt8) (accum >> 16);
			pLocOut[2] = (UInt8) (accum >> 8);
			pLocOut += 3;
			accum <<= 24;
			vacantBits += 24;
		    }
		    else
		    if (vacantBits <= 16)
		    {
			if (pLocOut >= pLocEnd)
			{
			    __error__ = ia_bufferExceeded;
			    goto cleanupWrPosBlock;
			}
			    /* Write out 2 bytes */
			pLocOut[0] = (UInt8) (accum >> 24);
			pLocOut[1] = (UInt8) (accum >> 16);
			pLocOut += 2;
			accum <<= 16;
			vacantBits += 16;
		    }
		    else
		    if (vacantBits <= 24)
		    {
			if (pLocOut > pLocEnd)
			{
			    __error__ = ia_bufferExceeded;
			    goto cleanupWrPosBlock;
			}
			    /* Write out 1 byte */
			pLocOut[0] = (UInt8) (accum >> 24);
			pLocOut += 1;
			accum <<= 8;
			vacantBits += 8;
		    }

			/* Now, see if we've made enough room */
		    if (vacantBits < pSymbol->huffmanCodeLen)
		    {
			    /* There's not enough room to insert the entire
			     * huffman code.  Insert it in two pieces.
			     */
			accum |= (pSymbol->huffmanCode) >>
				    (pSymbol->huffmanCodeLen - vacantBits);
			    /* Write out 4 bytes */
			pLocOut[0] = (UInt8) (accum >> 24);
			pLocOut[1] = (UInt8) (accum >> 16);
			pLocOut[2] = (UInt8) (accum >> 8);
			pLocOut[3] = (UInt8) (accum);
			pLocOut += 4;
			vacantBits = 32 - pSymbol->huffmanCodeLen + vacantBits;
			accum = (pSymbol->huffmanCode) >> vacantBits;
		    }
		    else	/* the normal case, the huffman code fits */
		    {
			vacantBits -= pSymbol->huffmanCodeLen;
			accum |= pSymbol->huffmanCode << vacantBits;
		    }
		}
		else	/* the normal case, the huffman code fits */
		{
		    vacantBits -= pSymbol->huffmanCodeLen;
		    accum |= pSymbol->huffmanCode << vacantBits;
		}
		numSymbolsEncoded--;
	    }

		/* Go read next x position in this line. */
	    temp = *pPos++;
	    if (temp & 0x80)
	    {
		temp = (temp & 0x7F) + ((*pPos) << 7);
		if ((*pPos++) & 0x80)	/* x coord is negative */
		    temp |= 0xFFFF8000;
	    }
	    else
	    if (temp == 0)	/* We're either at the end of the
				 * line or we have an offset of 0
				 * (signified by dX = 0, dY = -1 (2
				 * bytes).  See which one it is.
				 * Note that if pPos[0] is non-zero,
				 * we can read pPos[1] since pPos[0]
				 * doesn't contain the EndOfBuffer flag.
				 */
	    {
		if ((pPos[0] & 0x80) && (pPos[1] & 0x80))
		{
			/* The BB doesn't change.  We want to skip
			 * at least the two bogus y coords.  We
			 * also want to skip the next one or
			 * two bytes of X coord since all it does
			 * is bring us back to the same spot.
			 */
		    if (pPos[2] & 0x80)
			pPos += 4;
		    else
			pPos += 3;
		}
		else		/* really was end of line.  If this
				 * is the odd line, calculate the
				 * nextY as well as declaring that
				 * we're done with this line.
				 */
		{
		    if (curX == curX1)
		    {
			nextY = *pPos++;
			if (nextY & 0x80)
			    nextY = (nextY & 0x7F) + ((*pPos++) << 7);
			else
			if (nextY == 0)
			    eofFound = cTrue;

			nextY += curY + 1;	/* get real next Y coord.
						 * Remember that curY points
						 * to the even line above
						 * this one. */
		    }

		    curX = cEOLVal;
		    pSymbol = NULL;
		    symWidth = 0;
		}
	    }
	    curX += temp;

		/* If the curX < cEOLVal, there's a symbol to read. */
	    if (curX < cEOLVal)
	    {
		temp = *pPos++;
		if (temp & 0x80)	/* it's a 2-byte symbol ID */
		{
		    temp = (temp & 0x7F) + ((*pPos++) << 7);
		}
		pSymbol = pSymbolTable[temp];
		symWidth = pSymbol->width - 1;
	    }

		/* We`ve now processed a symbol on the even or odd
		 * scan line.  Update the corresponding variables.
		 */
	    if (pPos > pLocIn)	/* just read stuff from odd line */
	    {
		pSymbol1 = pSymbol;
		pLocIn = pPos;
		curX1 = curX;
		symWidth1 = symWidth;
	    }
	    else
	    {
		pSymbol0 = pSymbol;
		pPos0 = pPos;
		curX0 = curX;
		symWidth0 = symWidth;
	    }
	}	/* end of reading symbols in the strip (loop B) */


	    /* We've encoded the last symbol in the strip.  If there
	     * were any non-residual symbols in this strip, insert the
	     * strip end code for this delta X table.  It turns out
	     * that all the end codes are the same, so we just insert
	     * the one.
	     */
	if (lastX != -1)
	{
	    if (vacantBits < cTundraStripEndCodeLen)
	    {
		if (pLocOut >= pLocEnd)
		{
		    __error__ = ia_bufferExceeded;
		    goto cleanupWrPosBlock;
		}
		    /* Write out 2 bytes */
		pLocOut[0] = (UInt8) (accum >> 24);
		pLocOut[1] = (UInt8) (accum >> 16);
		pLocOut += 2;
		accum <<= 16;
		vacantBits += 16;
	    }
	    vacantBits -= cTundraStripEndCodeLen;
	    accum |= cTundraStripEndCode << vacantBits;
	}
    }		/* end of looping through all the image lines (loop A) */


	/* Hooray!  We've encoded the last symbols in the last strip.
	 * Add a delta Y encoding that puts us off any image.
	 */
    __error__ = ip_encodeTundraBigVal (pIPState, pDeltaYEncoder,
				     0xFFFF, &accum, &vacantBits,
				     &pLocOut, pLocEnd);
    if (__error__ != ia_successful)
	goto cleanupWrPosBlock;

	/* Add the extensions to the position block.  As of today,
	 * the only extension is a length field.  We give it a value
	 * of 0.
	 */
    __error__ = ip_encodeTundraInteger (pIPState, 0, &accum, &vacantBits,
				      &pLocOut, pLocEnd);
    if (__error__ != ia_successful)
	goto cleanupWrPosBlock;

	/* now flush out the accumulated bits */
    while (vacantBits < 32)
    {
	*pLocOut++ = (UInt8) (accum >> 24);
	accum <<= 8;
	vacantBits += 8;
    }

	/* Record the number of bytes used for this section. */
    *pRealOutputLength = pLocOut - pOutputBuf;

	/* If we didn't encode the number of symbols we expected to encode,
	 * complain.  Otherwise, be happy.
	 */
    if (numSymbolsEncoded != 0)
	__error__ = ia_internal;
    else
	__error__ = ia_successful;

cleanupWrPosBlock:
    NORMAL_RETURN;
}


/* This is a big number used to mark the right coord of the second dummy
 * run at the end of each scan line in the CCITT run array.
 */
#define	cDummyMarker	0x7FFFFFFF

/* PUBLIC */
/**************************************************************************
 *   i_writeTundraResidualBlock()
 *
 *	This procedure converts an internally-formatted position buffer
 *	into an official Tundra residual block,  The residual images are
 *	those that only occur once in all the pages being compressed,
 *	therefore do not belong in the symbol dictionary.
 *	The residual block does not include the residual block header.
 *	This division is made on the assumption that the format of the
 *	residual block will not change over time while the format of
 *	the residual header and relative locations of the position
 *	and residual blocks will change as the encapsulation changes.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pInputBuf:		pointer to internal position buffer.
 *		inputBufSize:		number of bytes in input pos buffer.
 *		pOutputBuf:		pointer to block of memory to be
 *					used for output position block.
 *		outputBufSize:		number of bytes in the output
 *					buffer.
 *		pResX:			pointer to returned x position of
 *					residual image.
 *		pResY:			pointer to returned y position of
 *					residual image
 *		pResW:			pointer to returned width of residual
 *					image;
 *		pResH:			pointer to returned height of
 *					residual image.
 *		pRealOutputLength:	pointer to number of bytes actually
 *					used by residual block in the output
 *					buffer.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL.
 *		ia_bufferExceeded	output didn't fit into outputBuf.
 *
 **************************************************************************/
Int32  CDECL 
i_writeTundraResidualBlock (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt8      *pInputBuf,
    UInt32	inputBufSize,
    UInt8      *pOutputBuf,
    UInt32	outputBufSize,
    UInt32     *pResX,
    UInt32     *pResY,
    UInt32     *pResW,
    UInt32     *pResH,
    UInt32     *pRealOutputLength)
{
    PROLOGUE;
    TCCompressionInfo	*pTextCompInfo = (TCCompressionInfo *)pTextCompInfoIn;
    CCITTCompressionInfo *pCCITTCompInfo = NULL;
					/* the struct used by ip_compMMRRow */
    Int32		 ccittInfoSize;	/* size of CCITTCompressionInfo */
    UInt8		*pLocalBuf = pOutputBuf;
    UInt8		*pOutputBufEnd;	/* pts to end of output buf */
    TCSymbol		*pPendingList;	/* list of residual syms whose runs
					 * aren't being put out yet. */
    TCSymbol		dummy1stActive;	/* dummy symbol at beginning of
					 * active list */
    TCSymbol		dummyLastActive; /* last dummy symbol in active list */
    TCSymbol		*pActiveSym;	/* ptr to symbol in Active list */
    TCSymbol		*pSymbol;	/* scratch pointer */
    TCSymbol		**pSymbolTable;	/* a table used to find a
					 * symbol from a symbol ID.
					 */
    TCScanLineRun	 dummy1stRun;	/* dummy run at the head of the
					 * sorted list of runs */
    TCScanLineRun	 dummyLastRun;	/* dummy run at the tail of the
					 * sorted list of runs */
    TCScanLineRun	*pSortRun;	/* ptr to run in sorted list */
    TCScanLineRun	*pNewRun;	/* ptr to run to be put into sorted
					 * list */
    Int32		 lineTag;	/* used to identify residual runs
					 * that are on a single scanline */
    Int32		 symbolID;
    Int32		 resL, resW;	/* left side and width of residual
					 * bounding box. */
    Int32		 curY, resB;	/* current and bottom rows of residual
					 * bounding box. */
    Run32		*pCodingRuns;	/* Pointer to runs for coding
					 * line. */
    Run32		*pRefRuns; 	/* pointer to runs for reference
					 * line in array of runs. */
    Run32		*pDummy;	/* scratch ptr */

    

	/* Check the pointers */
    if ((pTextCompInfo == NULL) || (pInputBuf == NULL) ||
	(pOutputBuf == NULL) || (pResX == NULL) ||
	(pResY == NULL) || (pResW == NULL) || (pResH == NULL) ||
	(pRealOutputLength == NULL))
	return ia_invalidParm;

	/* Point to the end of the output buffer.  ip_compMMRRow needs it */
    pOutputBufEnd = pOutputBuf + outputBufSize;

	/* First, see if the statistics on the position buffer already
	 * exist.  If not, make them.  This also identifies all the
	 * residual symbols on this page, sorts them by upper left
	 * corner and put the symbol ID of the first symbol in
	 * the list into the beginning of the position buffer.
	 */
    if (pInputBuf[cTCPosStatsValid] == 0)
    {
	__error__ = ip_calcPositionBufStats (pIPState, pTextCompInfo,
					   pInputBuf, inputBufSize, 0);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
    }

	/* Next, get the address of the table that maps symbol IDs
	 * to symbols.  It was created as part of ip_calcPositionBufStats.
	 */
    pSymbolTable = &(pTextCompInfo->pSymTable->table[0]);

	/* Record the bounding box and number of bytes used.
	 * If there's no residual, return now.
	 */
    *pResX = ((UInt32) pInputBuf[cTCPosResX0]) +
	     ((UInt32) pInputBuf[cTCPosResX1] << 8);
    *pResY = ((UInt32) pInputBuf[cTCPosResY0]) +
	     ((UInt32) pInputBuf[cTCPosResY1] << 8);
    *pResW = ((UInt32) pInputBuf[cTCPosResWidth0]) +
	     ((UInt32) pInputBuf[cTCPosResWidth1] << 8);
    *pResH = ((UInt32) pInputBuf[cTCPosResHeight0]) +
	     ((UInt32) pInputBuf[cTCPosResHeight1] << 8);

    if (*pResW == 0)
    {
	*pRealOutputLength = 0;
	NORMAL_RETURN;
    }

	/* Set the local version of the residual bounding box */
    resL = *pResX;
    resW = *pResW;
    curY = *pResY;
    resB = curY + *pResH - 1;

	/* Get the symbol ID of the first residual symbol */
    symbolID = ((UInt32) pInputBuf[cTCPos1stResSymID0]) +
	       ((UInt32) pInputBuf[cTCPos1stResSymID1] << 8) +
	       ((UInt32) pInputBuf[cTCPos1stResSymID2] << 16) +
	       ((UInt32) pInputBuf[cTCPos1stResSymID3] << 24);

	/* Set the active and pending lists */
    pPendingList = pSymbolTable[symbolID];

	/* Get the size of the CCITT compression info structure
	 * to be used by ip_compMMRRow.  The size of this
	 * structure is a function of the width of the residual
	 * bounding box, so we can't just allocate one as an
	 * automatic variable.
	 */
    __error__ = i_getSizeOfCCITTCompressionInfo (pIPState, resW + 2*cFrameBits,
					         &ccittInfoSize);
    if (__error__ != ia_successful)
	NORMAL_RETURN;

    pCCITTCompInfo = (CCITTCompressionInfo *) MALLOC(ccittInfoSize);
    if (pCCITTCompInfo == NULL)
	return ia_nomem;

    __error__ = i_initializeCCITTCompressionInfo (pIPState,
						(void *)pCCITTCompInfo,
						resW + 2*cFrameBits,
						cFalse, 0,
						cNoPad, ccittInfoSize);
    if (__error__ != ia_successful)
	goto writeTunResBlkCleanup;

	/* Set the CCITT run array pointers to the second run in each array.
	 * The first run in each array is a dummy.
	 * Since we know that run2ArrayIsCoding is always set to false,
	 * i.e. the first run array is the initial coding array,
	 * just point to the second array.
	 */
    pCodingRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				      pCCITTCompInfo->runArrayOffset +
				      sizeof(Run32));
    pRefRuns = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				   pCCITTCompInfo->run2ArrayOffset +
				   sizeof(Run32));

	/* Also initialize the reference run to have a blank line.
	 * This is done by putting in only the dummy elements.
	 */
    pRefRuns[0].runLeft = resW;
    pRefRuns[0].runRight = resW-1;
    pRefRuns[1].runLeft = resW;
    pRefRuns[1].runRight = cDummyMarker;

	/* Now, we're going to loop through all the scan lines of
	 * the residual bounding box.  For each line, we move
	 * any symbols that start in this line from the pending
	 * to the active list.  The symbols in the active list
	 * are sorted by left edge.
	 *
	 * We then sort all the residual runs in this
	 * line into a list before copying them into the array used
	 * by ip_compMMRRow.  We use the pClient pointers in
	 * the TCScanLine runs to link the list elements.  Prior
	 * to this, the pClient pointers had flags used to
	 * determine which runs were on one scan line (all runs
	 * on even scan lines have one value in pClient, all runs on
	 * odd scan lines have the other possible value in pClient).  We
	 * use these flags to stop at the end of symbol scan
	 * lines.  We also use each symbol's pRunTail pointer
	 * to point to the first run of the next scan line.
	 * We don't use the symbol's pNext pointers for this
	 * job so that the symbol's runs will remain linked
	 * together.  This is needed to delete them later.
	 *
	 * In the sorting procedure, we maintain a sorted list of runs.
	 * There are dummy runs on the ends to so we don't have
	 * to worry about inserting runs on the ends of the list.
	 * We maintain a pointer to the most recently inserted
	 * run.  When we insert a new run, we see if we need
	 * to move backwards in the list.  If so, we start
	 * over at the beginning since there are no backwards pointers.
	 * Since the symbols are sorted, we expect that this won't
	 * happen very often.
	 *
	 * We next check to see if we need to move forwards.  If so,
	 * we move until the new run should be inserted before
	 * the next run in the list.  We then insert the run.
	 * Note that the check for moving backwards need only
	 * happen for the first run of a symbol in a scan line.
	 * All other runs for that symbol in that scan line can
	 * only move forwards.
	 *
	 * When all new runs have been inserted into the list,
	 * we copy them, offsetting them by the X position of the
	 * bounding box if necessary, into the proper run array
	 * held in pCCITTCompInfo.  We then call ip_compMMRRow to
	 * generate the correct MMR codes.
	 */

	/* Initialize the dummy elements on the ends of the sorted
	 * list of runs and for the sorted list of active symbols.
	 * We use dummies in both cases so that we can ignore the
	 * weird cases of empty lists, inserting at the beginning and/or
	 * end, etc.
	 * We use the otherwise unused symbol elements pNewSymbol
	 * and pClassSymbol to link together symbols on the
	 * active list.  This leaves the pNext and pPrev pointers
	 * intact, so we can start down the list again if there's
	 * an error.
	 */
    dummy1stRun.runLeft = dummy1stRun.runRight = -1;
    dummyLastRun.runLeft = dummyLastRun.runRight = cEOLVal;
    dummyLastRun.pClient = NULL;
    dummy1stActive.minX = -1;
    dummyLastActive.minX = cEOLVal;
    dummy1stActive.pNewSymbol = &dummyLastActive;	/* the "next" ptr */
    dummyLastActive.pClassSymbol = &dummy1stActive;	/* the "prev" ptr */
    dummy1stActive.pClassSymbol = NULL;
    dummyLastActive.pNewSymbol = NULL;

	/* process the entire residual bounding box. */
    for (; curY <= resB; curY++)
    {
	    /* See if any symbols in the pending list should be
	     * moved to the active list.  Inside the loop, we
	     * keep pActiveSym pointing to the last inserted
	     * symbol since that's the most
	     * likely location for the next symbol.
	     */
	pActiveSym = dummy1stActive.pNewSymbol;
	while ((pPendingList != NULL) && (pPendingList->minY == curY))
	{
		/* Here's a symbol that should be moved from the
		 * pending list to the active list.  Run through
		 * the Active list to find a place for it.
		 */
	    pSymbol = pPendingList;
	    pPendingList = pPendingList->pNext;	/* remove from pending list */

		/* See if we should move backwards in the list */
	    while (pActiveSym->minX > pSymbol->minX)
		pActiveSym = pActiveSym->pClassSymbol;

		/* see if we have to move forwards.  Since the loop above
		 * will always back us off the last dummy element, we
		 * can always move up one safely.
		 */
	    pActiveSym = pActiveSym->pNewSymbol;
	    while (pActiveSym->minX < pSymbol->minX)
		pActiveSym = pActiveSym->pNewSymbol;

		/* pActiveSym is at the same X position or to the right
		 * of pSymbol.  Insert pSymbol before pActiveSym.
		 */
	    pSymbol->pNewSymbol = pActiveSym;
	    pSymbol->pClassSymbol = pActiveSym->pClassSymbol;
	    pActiveSym->pClassSymbol = pSymbol;
	    pSymbol->pClassSymbol->pNewSymbol = pSymbol;

		/* Make the symbol's pRunTail point to the first run.
		 * We assume each symbol's pRunTail points to the
		 * next run for a scan line.
		 */
	    pSymbol->pRunTail = pSymbol->pRunHead;
	}

	    /* Go through the symbols in the active list, moving
	     * runs to the sorted list.   First, empty the sorted
	     * list.
	     */
	dummy1stRun.pClient = (void *) &dummyLastRun;
	pSortRun = &dummy1stRun;

	for (pSymbol = dummy1stActive.pNewSymbol;
	     pSymbol->minX < cEOLVal;
	     pSymbol = pActiveSym)
	{
		/* Get a pointer to the first run for this symbol
		 * and the lineTag for this line of this symbol.
		 * This is the only place we have to check to see
		 * we have to move backwards for this symbol.
		 * Since the symbol's runs are arranged from left to right,
		 * we never need to move farther to the left for
		 * more runs on this scan line of this symbol.
		 */
	    pNewRun = pSymbol->pRunTail;
	    lineTag = (Int32) (pNewRun->pClient);

	    if (pNewRun->runLeft < pSortRun->runLeft)
		pSortRun = &dummy1stRun;

		/* Process all this symbol's run on this line */
	    for (; 
		 (pNewRun != NULL) && ((Int32)(pNewRun->pClient) == lineTag);
		 pNewRun = pNewRun->pNext)
	    {
		    /* Check to see if we should move forwards in the
		     * sorted list */
		while (((TCScanLineRun *)(pSortRun->pClient))->runLeft <
			pNewRun->runLeft)
		   pSortRun = (TCScanLineRun *)(pSortRun->pClient);

		    /* Insert pNewRun between pSortRun and pSortRun->pClient.
		     */
		pNewRun->pClient = (void *) (pSortRun->pClient);
		pSortRun->pClient = (void *) pNewRun;
	    }

		/* If that was the last run for this symbol, remove it
		 * from the active list */
	    pActiveSym = pSymbol->pNewSymbol;
	    if (pNewRun == NULL)
	    {
		pActiveSym->pClassSymbol = pSymbol->pClassSymbol;
		if (pActiveSym->pClassSymbol != NULL)
		pActiveSym->pClassSymbol->pNewSymbol = pActiveSym;
	    }
	    else	/* there are more runs in this symbol */
		pSymbol->pRunTail = pNewRun;

	}	/* end of sorting all the runs for this scan line */

	    /* Now copy the sorted runs to the correct run array
	     * in the CCITT info.  If the left edge of the bounding
	     * box is 0 (often the case) skip the step of adjusting
	     * the run coords.
	     */
	pDummy = pCodingRuns;		/* point to first coding run */
	if (resL == 0)
	{
	    for (pSortRun = (TCScanLineRun *) (dummy1stRun.pClient);
		 pSortRun->runLeft < cEOLVal;
		 pSortRun = (TCScanLineRun *) (pSortRun->pClient))
	    {
		pDummy->runLeft = pSortRun->runLeft;
		pDummy->runRight = pSortRun->runRight;
		pDummy++;
	    }
	}
	else	/* need to adjust run coords */
	{
	    for (pSortRun = (TCScanLineRun *) (dummy1stRun.pClient);
		 pSortRun->runLeft < cEOLVal;
		 pSortRun = (TCScanLineRun *) (pSortRun->pClient))
	    {
		pDummy->runLeft = pSortRun->runLeft - resL;
		pDummy->runRight = pSortRun->runRight - resL;
		pDummy++;
	    }
	}

	    /* add the dummy elements to the end of the scan line */
	pDummy[0].runLeft = resW;
	pDummy[0].runRight = resW-1;
	pDummy[1].runLeft = resW;
	pDummy[1].runRight = cDummyMarker;

	    /* Just to be extra safe, set the run offset in
	     * the compression info to 0 so that we start
	     * interpreting the very first run and not some
	     * run left over from some previous stoppage.
	     * I'm not really sure when this does anything
	     * that wasn't already done, but it seems tidy.
	     */
	pCCITTCompInfo->currentRun = 0;
	pCCITTCompInfo->refRun = 0;

	    /* Ask the compression code to translate runs into the
	     * compressed data stream.  The width sent down is used
	     * to decide when we've coded the whole line.
	     */
	__error__ = ip_compMMRRow (pIPState, pCCITTCompInfo, resW,
			         &pLocalBuf, pOutputBufEnd);
	if (__error__ != ia_successful)
	    goto writeTunResBlkCleanup;

	    /* Do the normal stuff that gets us to the next scan line */
	pCCITTCompInfo->lineNumber++;
	pCCITTCompInfo->run2ArrayIsCoding ^= cTrue;
	pDummy = pCodingRuns;
	pCodingRuns = pRefRuns;
	pRefRuns = pDummy;
    }	/* End of looping through all lines of the residual image */

	/* So far, so good.  Flush the remaining bytes in the
	 * compression pipeline to the output buffer.
	 */
    __error__ = i_flushCCITT (pIPState, (void *)pCCITTCompInfo,
			      pLocalBuf,
			      pOutputBufEnd - pLocalBuf,
			      &ccittInfoSize);
    if (__error__ != ia_successful)
	goto writeTunResBlkCleanup;

	/* Hooray, it worked!  Compute the number of bytes written and
	 * return happily.
	 */
    *pRealOutputLength = pLocalBuf - pOutputBuf + ccittInfoSize;
    __error__ = ia_successful;

writeTunResBlkCleanup:
	/* If this is an error condition, we might try again.  In this
	 * case, fix up the pClient entries of all the runs of all the
	 * symbols we've touched.
	 */
    if (__error__ != ia_successful)
    {
	for (pSymbol = pSymbolTable[symbolID];
	     pSymbol != pPendingList;
	     pSymbol = pSymbol->pNext)
	{
	    for (pNewRun = pSymbol->pRunHead, lineTag = cEvenLineTag;
		 pNewRun != NULL;
		 pNewRun = pNewRun->pNext)
	    {
		pNewRun->pClient = (void *)lineTag;
		if ((pNewRun->pNext != NULL) &&
		    (pNewRun->runRight + 1 >= pNewRun->pNext->runLeft))
		    lineTag ^= cLineTagMask;
	    }
	}
    }

    if (pCCITTCompInfo != NULL)
	FREE(pCCITTCompInfo);

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_calcPositionBufStats()
 *
 *	This procedure is used by Tundra output routines to calculate
 *	various useful statistics from a position buffer in a single
 *	pass.  This is done instead of having each routine parse the
 *	position buffer for itself.  All the resulting statistics are
 *	stored in reserved spaces at the beginning of the position buffer.
 *
 *	Input:
 *		pTextCompInfo:	Ptr to all text compression info
 *		pInputBuf:	Pointer to input position buffer.
 *		inputBufSize:	length of input position buffer.
 *		pageRequest:	Page ID of single-count items to add, or 0
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_invalidParm		one of the pointers was NULL
 *					or pInputBuf does not point to
 *					a valid position block.
 *
 **************************************************************************/
Int32  CDECL 
ip_calcPositionBufStats (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    UInt8      *pInputBuf,
    UInt32	inputBufSize,
    UInt32	pageRequest)
{
    PROLOGUE;
    UInt8		 *pPos0;		/* pointer to even
						 * position buffer lines. */
    UInt8		 *pLocalBuf;		/* local position buf ptr. */
    Int32		  lastX;		/* last X position */
    Int32		  curX0, curX1;		/* current X position in even
						 * and odd lines */
    Int32		  deltaX;		/* used to calculate modal
						 * value of intra-strip dX */
    Int32		  symWidth0, symWidth1;	/* widths of symbols on even
						 * and odd lines.  Well, not
						 * exactly.  These are really
						 * (width-1) since that's
						 * what is normally used here */
    Int32		  curY, nextY;		/* current and next Y pos */
    TCSymbol		 *pSymbol0, *pSymbol1;	/* current symbols in even and
						 * odd lines. */
    TCSymbol		**pSymbolTable;		/* a table used to find a
						 * symbol from a symbol number.
						 */
    TCSymbol		 *pLastResSym;		/* the last residual symbol
						 * inserted into the list
						 * of residual symbols sorted
						 * by width and height. */
    TCSymbol		  dummy1stRes;		/* dummy symbol, is the default
						 * beginning of the residual
						 * list. */
    TCSymbol		  dummyLastRes;		/* dummy last residual symbol */
    UInt32		  resRunCount;		/* number of residual runs seen
						 * so far. */
    UInt32		  deltaRuns;		/* number of runs found in a
						 * residual symbol */
    Int32		  resL, resR;		/* left and right boundaries
						 * of residual bounding box */
    Int32		  resT, resB;		/* top and bottom of residual
						 * bounding box. */
    Int32		  temp;			/* general variable */
    UInt32		  version;		/* the version number of the
						 * position buf */
    UInt32		  magicNum;		/* the magic number of the
						 * position buf */
    Bool		  eofFound;		/* says whether we've
						 * seen the end of the buffer */
    Int32		  modeHist[cTundraNumModalVals];  /* histogram used
						 * to choose modal delta X
						 * value */
    Int32		  numSymbols;		/* the number of symbols in
						 * this position block */
    

	/* First, check the pointers */
    if ((pTextCompInfo == NULL) || (pInputBuf == NULL))
	return ia_invalidParm;

	/* If the input buffer doesn't hold a header, it's too
	 * short.  If it does, we assume it's a valid input
	 * buffer.
	 */
    if (inputBufSize < cTCPosHeaderVer2Len)
	return ia_invalidParm;

	/* Initialize the local buffer pointer */
    pLocalBuf = pInputBuf;

	/* There are currently several classes of things we want to
	 * calculate:
	 *	1. the estimated Tundra position buffer size,
	 *	2. the estimated Tundra residual buffer size,
	 *	3. The residual buffer bounding box.
	 *	4. The modal value of inter-symbol spacing for
	 *	   a strip height of 2.
	 *	5. Space for ID of first residual symbol on the
	 *	   page (we find the rest by following a list).
	 *	6. The number of non-residual symbols on the page.
	 *
	 * We will probably try strip heights other than 2 in the future.
	 *
	 * Residual symbols are those whose instance count is 1,
	 * i.e. they only occurred once in the set of pages being
	 * compressed.
	 *
	 * The Tundra position buffer size will be estimated
	 * to be the size of this input buffer.  That's easy and
	 * usually errors a little on the high side, which is
	 * desirable.
	 *
	 * The Tundra residual buffer size will be estimated to have
	 * one byte per run in the residual buffer.  We hope that most
	 * runs can be represented by less than a byte.  For this
	 * to be true, the endpoints of most runs must be within
	 * 1 pixel (in X) of the endpoints of the runs above them.
	 * This is a result of the MMR coding scheme.  If there is
	 * lots of jaggy-ness or lots of blank lines, this estimate
	 * will be too low.
	 *
	 * The Tundra residual bounding box is found by simply looking
	 * at the bounding box of each residual symbol.  Since
	 * the position buffer is scanned from left to right, top to
	 * bottom, and the points stored in the position buffer
	 * are the lower right corners of the bounding boxes, we must
	 * consider all four sides of every symbol's bounding box
	 * each time one is encountered.  The bounding box of each
	 * symbol can extend backwards (to the upper left) by any
	 * arbitrary amount.
	 *
	 * The first residual symbol is found by sorting their upper
	 * left corners.  The symbol with the UL corner nearest the
	 * upper left (smallest row, use smallest column in case of
	 * a tie) is this first symbol.  We also sort the residual
	 * symbols by position so that the rest of the list can
	 * be created.
	 *
	 * The number of non-residual symbols on the page is found
	 * just by counting them.
	 *
	 * 
	 *
	 * First, we see if this is a real position buffer by
	 * looking at the header.  If it's not a good header, we
	 * return with ia_invalidParm.  If the header is good,
	 * we assume the entire block is well-formed (not chopped off before
	 * the end).
	 *
	 * We then go through our main loop (loop A).  The code is organized
	 * into two levels of loops.
	 * At the top level (A) is a loop that goes through the stripes
	 * of the position block.
	 * Inside this loop:
	 *  If the first line is even, quickly skip to its end to
	 *  the start of the next line with something in it.  If
	 *  that line is immediately below the even line, keep
	 *  pointing to its beginning.  If the next line is not
	 *  immediately below the current line, set the curX1 value
	 *  to cEOLVal.
	 *  If the first line is odd, set curX0 to cEOLVal since there
	 *  is no even line in this strip.  Start curX1 with the first
	 *  symbol of the odd line.
	 *
	 *  We now begin a loop (B) that continues until both curX0 val
	 *  and curX1 val equal cEOLVal.  Before entering this loop,
	 *  set lastX to -1 to signify that we don't want to record 
	 *  the distance to the previous symbol at the start of the
	 *  line.  Also find the X location and symbol ID of the
	 *  first symbols in the even and odd scan lines (if their
	 *  curX* val isn't already cEOLVal.)  The X position of
	 *  a symbol is the location of the left edge of its bounding
	 *  box.
	 *  This loop contains two clauses:
	 *	If curX0 - symWidth0 <= curX1 - symWidth1
	 *	(the left edge of the current symbol on the even
	 *	line is to the left of, or directly above, the one
	 *	on the odd line), look at the even line's symbol.  If it's
	 *	a residual symbol, record its number of runs and
	 *	bounding box.  If it's a non-residual symbol,
	 *	record its distance from lastX (unless lastX was
	 *	negative) in the histogram of X distances.  In either
	 *	case, update lastX to be the coord of the right edge
	 *	of this symbol + 1.
	 *	Read the next symbol and symbolID.  If we've reached the
	 *	end of the line, set curX0 to cEOLVal.
	 *
	 *	Otherwise (curX0 - symWidth0 > curX1 - symWidth1),
	 *	do the same thing as above with X1 substituted for X0.
	 *	Also, if we reach the end of the line, set nextY to 
	 *	the y coord of the next line.
	 *
	 * I should say something about how to parse a position block.
	 * A position block contains a sequence of delta X's and symbol IDs.
	 * If a delta X == 0, there's a delta Y following instead of a
	 * symbol ID.  If that delta Y is also 0, this is the end of
	 * the position buffer.  If the delta Y is -1, the real delta Y
	 * is 0 (stay on the same line).  Following each delta Y (except
	 * the one that signifies the end of the file) is a delta X that
	 * gives the position relative to the left edge of the image of
	 * the first thing.  The delta X following a delta Y of -1 always
	 * brings us back to the same spot (it's used to say that two symbol
	 * bounding boxes share the same lower right corner), so we can
	 * just skip over them.
	 *
	 * In the original position buffer format, the first byte in a file
	 * was a deltaX, usually 0 and usually followed by a delta Y.
	 * If there was a symbol that ended on line 0, I thought the
	 * first deltaX would be non-zero.  BUT, what if the first symbol
	 * in the file is a single dot in the upper left corner?  The
	 * deltaX for that would be 0 and we'd be confused.
	 *
	 * In the modern format, the first byte in the file is a delta Y
	 * relative to -1.  If the position buffer is empty (blank page),
	 * this is 0.  Otherwise, it's non-zero.  Following every deltaY,
	 * including this first one, is a firstX relative to 0.  Zero
	 * is a legal value for this first X since a symbol may occupy only
	 * column 0.  We always read this firstX.  A line ends when
	 * a deltaX (follows a symbolID) is zero.
	 *
	 * Delta X, Delta Y and symbol IDs can be one or two bytes.
	 * Single byte quantities have a 0 in the MSbit.  Two-byte
	 * quantities have the LS byte first.
	 * Here's a review:
	 * The position info can be of several forms:
	 *	1 byte x offset; 1 or 2 byte symbol ID.
	 *	2 byte x offset; 1 or 2 byte symbol ID.
	 *	1 byte x offset (= 0), 1 byte y offset (= 0) => end of file.
	 *	1 byte x offset (= 0), 1 byte y position (dist to new line), 
	 *		1 or 2 byte x offset, 1 or 2 byte sym ID.
	 *	1 byte x offset (= 0), 2 byte y position (dist to new line), 
	 *		1 or 2 byte x offset, 1 or 2 byte sym ID.
	 *      1 byte x offset (= 0), 2 byte y position (= -1),
	 *		1 or 2 byte x offset (equal to original X position),
	 *		1 or 2 byte symbol ID.
	 *
	 * Note that there are no lines that contain only x offset of 0
	 * and a non-zero y offset.  That's because we can jump as far
	 * as necessary with any 2-byte y offset, so don't need any blank
	 * lines that contain only y distances.  We can also end any
	 * scan line with an "end of position block" indicator, so won't
	 * have any scan lines that contain only one of these.
	 */

	/* So, first check the header and version. */
    magicNum = ((UInt32) pLocalBuf[cTCPosMagicNum3] << 24) |
	       ((UInt32) pLocalBuf[cTCPosMagicNum2] << 16) |
	       ((UInt32) pLocalBuf[cTCPosMagicNum1] << 8) |
	       ((UInt32) pLocalBuf[cTCPosMagicNum0]);

    if (magicNum != cTextCompPositionInfoMagicNum)
	return ia_invalidParm;

    version = ((UInt32) pLocalBuf[cTCPosVer1] << 8) +
	      ((UInt32) pLocalBuf[cTCPosVer0]);

    if (version < cTextCompPositionInfoVersionNum)
	return ia_invalidParm;

	/* advance pLocalBuf beyond the header */
    pLocalBuf += cTCPosHeaderVer2Len;

	/* Now, see if the table that associates Symbol IDs with
	 * symbols exists.  If not, make it.
	 */
    if ((pTextCompInfo->pSymTable == NULL) ||
	(pTextCompInfo->pSymTable->numPtrs != pTextCompInfo->nextSymbolNum))
    {	/* allocate and initialize a new table.  Delete the old one
	 * if it exists.
	 */
	__error__ = ip_createTCSymbolTable (pIPState, pTextCompInfo);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;
    }
    pSymbolTable = &(pTextCompInfo->pSymTable->table[0]);

	/* Clear the histogram used to calculate the modal delta X
	 * value.
	 */
    for (temp = 0; temp < cTundraNumModalVals; temp++)
	modeHist[temp] = 0;

	/* Initialize the variables used to count the residual
	 * runs and remember the residual bounding box */
    resL = resT = cEOLVal;	/* these are off the right and bottom edges */
    resR = resB = 0;		/* these are on the left and top edges. */
    resRunCount = 0;		/* number of residual runs seen so far */
    numSymbols = 0;		/* the number of non-residual symbols seen
				 * so far. */

	/* Initialize the dummy residual symbols.  Having dummies on the
	 * beginning and end of the list makes inserting new symbols
	 * easier since you never have to worry about the end cases.
	 */
    dummy1stRes.minX = -1;
    dummy1stRes.minY = -1;
    dummyLastRes.minX = cEOLVal;
    dummyLastRes.minY = cEOLVal;

	/* Initialize the last residual symbol ptr */
    pLastResSym = &dummy1stRes;
    dummy1stRes.pNext = &dummyLastRes;
    dummy1stRes.pPrev = NULL;
    dummyLastRes.pNext = NULL;
    dummyLastRes.pPrev = &dummy1stRes;

	/* Initialize the y position.  We do this by reading the
	 * first byte or two of the position buffer.
	 */
    curY = -1;
    nextY = *pLocalBuf++;
    if (nextY & 0x80)
	nextY = (nextY & 0x7F) + ((*pLocalBuf++) << 7);

    if (nextY == 0)		/* the position buffer was empty */
	eofFound = cTrue;
    else
	eofFound = cFalse;

    curY += nextY;		/* This is the first real Y position */

	/************************ LOOP A  **************************/

	/* Here's the beginning of the outermost loop, called loop A
	 * above.  We execute until the end of the position block has been
	 * processed.
	 */
    for (; eofFound != cTrue; curY = nextY)
    {

	    /* We haven't seen any symbols in this strip yet.
	     */
	pSymbol0 = pSymbol1 = NULL;
	pPos0 = NULL;
	symWidth0 = symWidth1 = 0;

	    /* We are ready to start processing another strip.
	     * If the first line is even, find the first symbol
	     * and its position with curX0 and pSymbol0.  After
	     * this, skip to the end of this line to
	     * see if the next line with something on it is
	     * immediately after this one (i.e. in the same strip).
	     * If so, read that first symbol into curX1 and
	     * pSymbol1.  If the next line does not immediately
	     * follow this one, set curX1 to cEOLVal.  In either
	     * case, set nextY to the Y coord of the next line.
	     * If the first line is odd, just set curX0 to cEOLVal
	     * and find the first symbol with curX1 and pSymbol1.
	     */
	if ((curY & 1) == 0)	/* current line num is even */
	{
	    curX0 = *pLocalBuf++;	/* get LS byte of first x offset */
	    if (curX0 & 0x80)		/* there's another byte */
	    {
		curX0 = (curX0 & 0x7F) + (*pLocalBuf++ << 7);
	    }
		/* There's no need to check to see if the firstX in
		 * the line is 0 since 0 is a perfectly legal value.
		 */

		/* Go read the first even line symbol ID. */
	    temp = *pLocalBuf++;
	    if (temp & 0x80)		/* it's a 2-byte symbol ID */
	    {
		temp = (temp & 0x7F) | ((*pLocalBuf++) << 7);
	    }
		/* point to the symbol in question.  Save
		 * width - 1.
		 */
	    pSymbol0 = pSymbolTable[temp];
	    symWidth0 = pSymbol0->width - 1;
	    pPos0 = pLocalBuf;		/* remember where next byte in the
					 * even line is so we can parse it
					 * later. */

		/* Scan to the end of the even
		 * line to see where the next line is.
		 * If it's the next scan line, we'll
		 * process it in this strip.  Otherwise, we'll
		 * process it in the next strip.
		 */
	    while (cTrue)
	    {
		temp = *pLocalBuf++;		/* get x coord */
		if (temp & 0x80)		/* there's another byte */
		{
		    pLocalBuf++;		/* skip over it */
		}
		else
		if (temp == 0)			/* reached the end of line
						 * (maybe) */
		{
		    nextY = *pLocalBuf++;	/* get 1st byte of y offset */
		    if (nextY & 0x80)		/* next byte is needed too */
		    {
			if (*pLocalBuf & 0x80)	/* BB doesn't move */
			{
				/* skip over the next X coord since it
				 * takes us right back here.
				 */
			    if (pLocalBuf[1] & 0x80)
				pLocalBuf += 3;
			    else
				pLocalBuf += 2;
				/* We want to keep going around the loop.
				 * We set nextY to -1 as a flag that
				 * this should happen, i.e. we didn't
				 * really find the end of the line.
				 * The next symbol ID will also
				 * be skipped at the bottom of the loop.
				 */
			    nextY = -1;
			}
			else
				/* That was a perfectly normal end of
				 * the even scan line.
				 */
			    nextY = (nextY & 0x7F) + ((*pLocalBuf++) << 7);
		    }
		    else
		    if (nextY == 0)		/* end of the file */
		    {
			eofFound = cTrue;	/* nextY will be updated
						 * below */
		    }
			/* Now, did we really find the end of the
			 * scan line, in which case we should stop
			 * right now, or did we only find one of those
			 * annoying "don't move the bounding box"
			 * sequences?
			 */
		    if (nextY != -1)
		    {
			    /* That was the end of the scan line. */
			nextY += curY;		/* point to next line. */
			break;			/* we found the end of the
						 * even scan line. */
		    }
		}

		    /* If we get here, there's a symbol to skip over before
		     * looping back around to find the next delta X.
		     */
		temp = *pLocalBuf++;
		if (temp & 0x80)		/* it's a 2-byte symbol ID */
		{
		    pLocalBuf++;
		}
	    }
		/* We've either found the end of a scan line or the
		 * end of the file.  If it's the end of the file,
		 * nextY == curY.
		 * If nextY == curY + 1, go read the first X offset
		 * for the odd line and its symbol.  Otherwise,
		 * say that there are no odd line symbols.
		 */
	    if (nextY == (curY + 1))
	    {
		curX1 = *pLocalBuf++;		/* get first x coord */
		if (curX1 & 0x80)		/* there's another byte */
		{
		    curX1 = (curX1 & 0x7F) + ((*pLocalBuf++) << 7);
		}
		    /* Since that was the firstX in a line, any value
		     * is legal - no need to check for 0.  Go get
		     * the first symbol ID.
		     */
		temp = *pLocalBuf++;
		if (temp & 0x80)		/* it's a 2-byte symbol ID */
		{
		    temp = (temp & 0x7F) + ((*pLocalBuf++) << 7);
		}
		    /* point to the symbol in question */
		pSymbol1 = pSymbolTable[temp];
		symWidth1 = pSymbol1->width - 1;
	    }
	    else	/* the next line isn't in this strip */
		curX1 = cEOLVal;
	}

	else

	{
		/* The first line is odd.  So, just go find the first
		 * symbol in that line.  Also adjust curY so that it
		 * points to the previous even line.  The code below
		 * assumes curY always points to the even line of the
		 * strip.
		 */
	    curY--;
	    curX0 = cEOLVal;
	    curX1 = *pLocalBuf++;	/* get LS byte of first x offset */
	    if (curX1 & 0x80)		/* there's another byte */
	    {
		curX1 = (curX1 & 0x7F) + (*pLocalBuf++ << 7);
	    }

		/* Since that was the firstX in a line, any value
		 * is legal - no need to check for 0.  Go get
		 * the first odd line symbol ID. */
	    temp = *pLocalBuf++;
	    if (temp & 0x80)		/* it's a 2-byte symbol ID */
	    {
		temp = (temp & 0x7F) | ((*pLocalBuf++) << 7);
	    }
		/* point to the symbol in question and
		 * save symbol's width - 1.
		  */
	    pSymbol1 = pSymbolTable[temp];
	    symWidth1 = pSymbol1->width - 1;
	}

	    /********************* LOOP B **************************/

	    /* Don't record distance to first sym */
	lastX = -1;

	    /* Go process the symbols in this strip */
	while (curX0 + curX1 < 2*cEOLVal)
	{
	    if (curX0 - symWidth0 <= curX1 - symWidth1)
	    {
		    /* pSymbol0's left edge is to the left of
		     * pSymbol1's left edge.
		     * See if pSymbol0 is a residual or a repeated symbol.
		     */
		if (pSymbol0->instanceCount == 1 && pTextCompInfo->useResidual)
		{
			/* pSymbol0 is a residual symbol.  Check its bounding
			 * box and count its runs
			 */
		    if (curX0 - symWidth0  < resL)
			resL = curX0 - symWidth0;
		    if (curY - pSymbol0->maxY < resT)
			resT = curY - pSymbol0->maxY;
		    if (curX0 > resR)
			resR = curX0;
		    if (curY > resB)
			resB = curY;

			/* Reverse the order of the runs in the
			 * symbol, convert them to 1x, absolute coords,
			 * and count them.  Only the count is used now,
			 * but the other stuff will be needed when
			 * we actually write the residual symbol to
			 * the residual image.
			 */
		    ip_reverseTCSymbolRuns (pIPState, pTextCompInfo, pSymbol0,
					    curX0 - symWidth0,
					    curY - pSymbol0->maxY,
					    &deltaRuns);
		    resRunCount += deltaRuns;

			/* Now insert this run into the list of
			 * residual symbols sorted by upper left
			 * corner.  ip_reverseTCSymbolRuns put the
			 * symbol's UL corner into its minX, minY
			 * entries.  First, disconnect it from
			 * the width and height tables.  Note that
			 * the symbol's min and max entries were
			 * converted to 1x, absolute coords by
			 * ip_reverseTCSymbolRuns.
			 */
		    if ((symWidth0 < cMaxSymIndexDim) &&
			((pSymbol0->height) <= cMaxSymIndexDim))
		    {
			    /* This is a small symbol, so look in the
			     * symbol array. */
			if (pTextCompInfo->pSymIndex->array
					    [symWidth0 + 1]
					    [pSymbol0->height] == pSymbol0)
			{
			    pTextCompInfo->pSymIndex->array
					    [symWidth0 + 1]
					    [pSymbol0->height] =
							    pSymbol0->pNext;
			}
		    }
		    else	/* this is a big symbol */
		    {
			if (pTextCompInfo->pBigSymbols == pSymbol0)
			    pTextCompInfo->pBigSymbols = pSymbol0->pNext;
		    }

		    if (pSymbol0->pNext != NULL)
			pSymbol0->pNext->pPrev = pSymbol0->pPrev;
		    if (pSymbol0->pPrev != NULL)
			pSymbol0->pPrev->pNext = pSymbol0->pNext;
		    pSymbol0->pPrev = pSymbol0->pNext = NULL;

			/* Now, connect it to the list of residual
			 * symbols for this page.  Note that the
			 * presence of dummy elements at either ends of the
			 * the list means that we don't have to
			 * worry about inserting elements at the
			 * end of the list.
			 */
			/* Move backwards in the list if necessary */
		    while ((pLastResSym->minY > pSymbol0->minY) ||
			   ((pLastResSym->minY == pSymbol0->minY) &&
			    (pLastResSym->minX > pSymbol0->minX)))
			pLastResSym = pLastResSym->pPrev;

			    /* Move forwards in the list as necessary.
			     * We can always take at least one step
			     * forwards since if we started on the
			     * last dummy element, the loop above would
			     * have moved us back by 1.
			     */
		    pLastResSym = pLastResSym->pNext;
		    while ((pLastResSym->minY < pSymbol0->minY) ||
			   ((pLastResSym->minY == pSymbol0->minY) &&
			    (pLastResSym->minX < pSymbol0->minX)))
			pLastResSym = pLastResSym->pNext;

			/* Insert new sym before pLastResSym */
		    pSymbol0->pPrev = pLastResSym->pPrev;
		    pLastResSym->pPrev = pSymbol0;
		    pSymbol0->pNext = pLastResSym;
		    pSymbol0->pPrev->pNext = pSymbol0;
		        /* guess that the next symbol will go after
			 * this one. */
		    pLastResSym = pSymbol0;
		}
		else if (pSymbol0->instanceCount > 1 || pageRequest > 0)
		    /* this is a repeated symbol, so just record its
		     * distance from the last repeated symbol in the
		     * mode histogram.
		     */
		{
		    numSymbols++;
		    deltaX = curX0 - symWidth0 - lastX;
		    if ((lastX > 0) &&
			((UInt32) deltaX < cTundraNumModalVals))
			modeHist[deltaX]++;
		    lastX = curX0 + 1;
		}

		    /* Go read next x position in even line. */
		temp = *pPos0++;
		if (temp & 0x80)
		{
		    temp = (temp & 0x7F) + ((*pPos0) << 7);
		    if ((*pPos0++) & 0x80)	/* x coord is negative */
			temp |= 0xFFFF8000;
		}
		else
		if (temp == 0)	    /* We're either at the end of the even
				     * line or we have an offset of 0
				     * (signified by dX = 0, dY = -1 (2
				     * bytes).  See which one it is.
				     * Note that if pPos[0] is non-zero,
				     * we can read pPos[1] since pPos[0]
				     * doesn't contain the EndOfBuffer flag.
				     */
		{
		    if ((pPos0[0] & 0x80) && (pPos0[1] & 0x80))
		    {
			    /* The BB doesn't change.  We want to skip
			     * at least the two bogus y coords.  We
			     * also want to skip the next one or
			     * two bytes of X coord since all it does
			     * is bring us back to the same spot.
			     */
			if (pPos0[2] & 0x80)
			    pPos0 += 4;
			else
			    pPos0 += 3;
		    }
		    else		/* really was end of line */
		    {
			curX0 = cEOLVal;
			pSymbol0 = NULL;
			symWidth0 = 0;
		    }
		}
		curX0 += temp;

		    /* If the curX0 < cEOLVal, there's a symbol to read. */
		if (curX0 < cEOLVal)
		{
		    temp = *pPos0++;
		    if (temp & 0x80)	/* it's a 2-byte symbol ID */
		    {
			temp = (temp & 0x7F) + ((*pPos0++) << 7);
		    }
		    pSymbol0 = pSymbolTable[temp];
		    symWidth0 = pSymbol0->width - 1;
		}
	    }	/* end of reading symbol on even line */
	    else
	    {
		    /* The left edge of the symbol on the odd line was
		     * the leftmost edge.  Go look at it.
		     */
		if (pSymbol1->instanceCount == 1 && pTextCompInfo->useResidual)
		{
			/* pSymbol1 is a residual symbol.  Check its bounding
			 * box and count its runs
			 */
		    if (curX1 - symWidth1  < resL)
			resL = curX1 - symWidth1;
		    if (curY + 1 - pSymbol1->maxY < resT)
			resT = curY + 1 - pSymbol1->maxY;
		    if (curX1 > resR)
			resR = curX1;
		    resB = curY + 1;		/* this is as low as we've gone,
						 * so it must be the bottom
						 * of the residual bounding
						 * box.
						 */

			/* Reverse the order of the runs in the
			 * symbol, convert them to 1x, absolute coords,
			 * and count them.  Only the count is used now,
			 * but the other stuff will be needed when
			 * we actually write the residual symbol to
			 * the residual image.
			 */
		    ip_reverseTCSymbolRuns (pIPState, pTextCompInfo, pSymbol1,
					    curX1 - symWidth1,
					    curY + 1 - pSymbol1->maxY,
					    &deltaRuns);
		    resRunCount += deltaRuns;

			/* Now insert this run into the list of
			 * residual symbols sorted by upper left
			 * corner.  ip_reverseTCSymbolRuns put the
			 * symbol's UL corner into its minX, minY
			 * entries.  First, disconnect it from
			 * the width and height tables.  Note that
			 * the symbol's min and max entries were
			 * converted to 1x, absolute coords by
			 * ip_reverseTCSymbolRuns.
			 */
		    if ((symWidth1 < cMaxSymIndexDim) &&
			((pSymbol1->maxY - pSymbol1->minY) < cMaxSymIndexDim))
		    {
			if (pTextCompInfo->pSymIndex->array
					    [symWidth1 + 1]
					    [pSymbol1->maxY -
					     pSymbol1->minY + 1] == pSymbol1)
			{
			    pTextCompInfo->pSymIndex->array
					    [symWidth1 + 1]
					    [pSymbol1->maxY -
					     pSymbol1->minY + 1] =
							    pSymbol1->pNext;
			}
		    }
		    else	/* this is a big symbol */
		    {
			if (pTextCompInfo->pBigSymbols == pSymbol1)
			    pTextCompInfo->pBigSymbols = pSymbol1->pNext;
		    }

		    if (pSymbol1->pNext != NULL)
			pSymbol1->pNext->pPrev = pSymbol1->pPrev;
		    if (pSymbol1->pPrev != NULL)
			pSymbol1->pPrev->pNext = pSymbol1->pNext;
		    pSymbol1->pPrev = pSymbol1->pNext = NULL;

			/* Now, connect it to the list of residual
			 * symbols for this page.  Note that the
			 * presence of dummy elements at either ends of the
			 * the list means that we don't have to
			 * worry about inserting elements at the
			 * end of the list.
			 */
			/* Move backwards in the list if necessary */
		    while ((pLastResSym->minY > pSymbol1->minY) ||
			   ((pLastResSym->minY == pSymbol1->minY) &&
			    (pLastResSym->minX > pSymbol1->minX)))
			pLastResSym = pLastResSym->pPrev;

			    /* Move forwards in the list as necessary.
			     * We can always take at least one step
			     * forwards since if we started on the
			     * last dummy element, the loop above would
			     * have moved us back by 1.
			     */
		    pLastResSym = pLastResSym->pNext;
		    while ((pLastResSym->minY < pSymbol1->minY) ||
			   ((pLastResSym->minY == pSymbol1->minY) &&
			    (pLastResSym->minX < pSymbol1->minX)))
			pLastResSym = pLastResSym->pNext;

			/* Insert new sym before pLastResSym */
		    pSymbol1->pPrev = pLastResSym->pPrev;
		    pLastResSym->pPrev = pSymbol1;
		    pSymbol1->pNext = pLastResSym;
		    pSymbol1->pPrev->pNext = pSymbol1;
		        /* guess that the next symbol will go after
			 * this one. */
		    pLastResSym = pSymbol1;
		}
		else if (pSymbol1->instanceCount > 1 || pageRequest > 0)
		    /* this is a repeated symbol, so just record its
		     * distance from the last repeated symbol in the
		     * mode histogram.
		     */
		{
		    numSymbols++;
		    deltaX = curX1 - symWidth1 - lastX;
		    if ((lastX > 0) &&
			((UInt32) deltaX < cTundraNumModalVals))
			modeHist[deltaX]++;
		    lastX = curX1 + 1;
		}

		    /* Go read next x position in odd line. */
		temp = *pLocalBuf++;
		if (temp & 0x80)
		{
		    temp = (temp & 0x7F) + ((*pLocalBuf) << 7);
		    if ((*pLocalBuf++) & 0x80)	/* temp is negative */
			temp |= 0xFFFF8000;
		}
		else
		if (temp == 0)	    /* This is either the end of the line
				     * in the odd row or a bounding box
				     * change of 0 (no offset).  See which
				     * one it is.
				     */
		{
		    if ((pLocalBuf[0] & 0x80) && (pLocalBuf[1] & 0x80))
		    {
			    /* We also want to toss the next X coord
			     * since it just takes us back to the
			     * same spot.
			     */
			if (pLocalBuf[2] & 0x80)
			    pLocalBuf += 4;
			else
			    pLocalBuf += 3;
		    }
		    else
		    {
			    /* It's really the end of the odd line.
			     * Compute nextY as well as declaring
			     * that we're done with the odd line.
			     */
			nextY = *pLocalBuf++;
			if (nextY & 0x80)
			    nextY = (nextY & 0x7F) + ((*pLocalBuf++) << 7);
			else
			if (nextY == 0)
			    eofFound = cTrue;

			nextY += curY + 1;	/* get real next Y coord.
						 * Remember that curY points
						 * to the even line above
						 * this one. */
			curX1 = cEOLVal;
			pSymbol1 = NULL;
			symWidth1 = 0;
		    }
		}
		curX1 += temp;

		    /* If curX1 < cEOLVal, there's a symbol to read. */
		if (curX1 < cEOLVal)
		{
		    temp = *pLocalBuf++;
		    if (temp & 0x80)	/* it's a 2-byte symbol ID */
		    {
			temp = (temp & 0x7F) + ((*pLocalBuf++) << 7);
		    }
		    pSymbol1 = pSymbolTable[temp];
		    symWidth1 = pSymbol1->width - 1;
		}
	    }	/* end of reading symbol on odd line */
	}	/* end of reading symbols in the strip (loop B) */
    }		/* end of looping through all the image lines (loop A) */

	/* At this point, we've got the residual bounding box, the
	 * number of residual runs and a histogram that can be used
	 * to compute the modal X value.  Figure out the real modal
	 * X value and write all this stuff into the stats section
	 * of the position buffer.
	 */
    for (lastX = modeHist[0], curX0 = 0, temp = 1;
	 temp < cTundraNumModalVals;
	 temp++)
    {
	if (lastX < modeHist[temp])
	{
	    lastX = modeHist[temp];
	    curX0 = temp;
	}
    }
    pInputBuf[cTCPosModalX] = (UInt8) curX0;	/* write best mode val */
	/* write residual bounding box.  If there were no residual
	 * symbols, write 0's.
	 */
    if (resL > resR)
    {
	resL = resR = resT = resB = 0;
    }
    else	/* calculate the residual width and height */
    {
	resR = resR - resL + 1;
	resB = resB - resT + 1;
    }

    pInputBuf[cTCPosResX0] = (UInt8) resL;
    pInputBuf[cTCPosResX1] = (UInt8) (resL >> 8);

    pInputBuf[cTCPosResY0] = (UInt8) resT;
    pInputBuf[cTCPosResY1] = (UInt8) (resT >> 8);

    pInputBuf[cTCPosResWidth0] = (UInt8) resR;
    pInputBuf[cTCPosResWidth1] = (UInt8) (resR >> 8);

    pInputBuf[cTCPosResHeight0] = (UInt8) resB;
    pInputBuf[cTCPosResHeight1] = (UInt8) (resB >> 8);

	/* Assume one byte / residual run in the Tundra output */
    pInputBuf[cTCPosEstResLen0] = (UInt8) resRunCount;
    pInputBuf[cTCPosEstResLen1] = (UInt8) (resRunCount >> 8);
    pInputBuf[cTCPosEstResLen2] = (UInt8) (resRunCount >> 16);
    pInputBuf[cTCPosEstResLen3] = (UInt8) (resRunCount >> 24);

	/* We assume the Tundra position buffer will be somewhat smaller
	 * that the internal position buffer, so we estimate
	 * the same size.  Yeah, we could
	 * try something fancier, but why bother?
	 */
    pInputBuf[cTCPosEstPosLen0] = (UInt8) inputBufSize;
    pInputBuf[cTCPosEstPosLen1] = (UInt8) (inputBufSize >> 8);
    pInputBuf[cTCPosEstPosLen2] = (UInt8) (inputBufSize >> 16);
    pInputBuf[cTCPosEstPosLen3] = (UInt8) (inputBufSize >> 24);

	/* Write in the symbol ID of the first residual symbol in the
	 * list sorted by position of Upper left corner.
	 */
    if (dummy1stRes.pNext == &dummyLastRes)
	temp = cEOLVal;		/* doesn't matter since we're not supposed
				 * to look at the list if there is no
				 * residual (est length = 0).
				 */
    else
	temp = dummy1stRes.pNext->symbolID;

    pInputBuf[cTCPos1stResSymID0] = (UInt8) temp;
    pInputBuf[cTCPos1stResSymID1] = (UInt8) (temp >> 8);
    pInputBuf[cTCPos1stResSymID2] = (UInt8) (temp >> 16);
    pInputBuf[cTCPos1stResSymID3] = (UInt8) (temp >> 24);

	/* Write out the number of non-residual symbols on this page */
    pInputBuf[cTCPosNumSym0] = (UInt8) numSymbols;
    pInputBuf[cTCPosNumSym1] = (UInt8) (numSymbols >> 8);
    pInputBuf[cTCPosNumSym2] = (UInt8) (numSymbols >> 16);
    pInputBuf[cTCPosNumSym3] = (UInt8) (numSymbols >> 24);


	/* Set the pPrev and pNext ptrs of the elements of the real
	 * end list elements to NULL so we'll know where the end
	 * really is.
	 */
    dummy1stRes.pNext->pPrev = NULL;
    if (dummyLastRes.pPrev != NULL)	/* i.e. there's a symbol in the list */
	dummyLastRes.pPrev->pNext = NULL;

	/* Just say that the stats are valid and we're done. */
    pInputBuf[cTCPosStatsValid] = (UInt8) cTrue;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_reverseTCSymbolRuns()
 *
 *	This procedure is used by Tundra output routines to reverse the
 *	order of the TCScanLineRuns associated with a dictionary symbol.
 *	The final order will be left to right, top to bottom, as required
 *	by normal run-length encoding schemes.  The coordinates of each
 *	symbol are also converted to absolute, 1x coords (from relative
 *	2x coords) by reversing the 1x -> 2x conversion process that was
 *	done when the symbol was encountered and adding an offset to
 *	every coordinate.
 *
 *	Input:
 *		pTextCompInfo:	Ptr to all text compression info
 *		pSymbol:	Pointer to input symbol.
 *		xOffset:	X coord to be added to each run coord.
 *				Equals abs coord of left edge of symbol's
 *				bounding box.
 *		yOffset:	Y coord of top run of symbol.
 *		pRunCount:	pointer to returned count of symbol runs.
 *
 *	Returns:
 *		ia_successful		things went well
 *
 **************************************************************************/
Int32  CDECL 
ip_reverseTCSymbolRuns (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCSymbol   *pSymbol,
    Int32	xOffset,
    Int32	yOffset,
    UInt32     *pRunCount)
{
    PROLOGUE;
    TCScanLineRun	*pCurRun, *pNextRun, *pPrevRun;
    UInt32		 runCount = 0;
    Int32		 scaledXOffset;

	/* Keep the compiler happy */
    pTextCompInfo = pTextCompInfo;

	/* If it's already been done, just count the symbols */
    if (pSymbol->runsPtToTail)
    {
	for (pCurRun = pSymbol->pRunHead;
	     pCurRun != NULL;
	     pCurRun = pCurRun->pNext)
		    runCount++;
	*pRunCount = runCount;
	NORMAL_RETURN;
    }

	/* Go reverse the order of the symbols, convert their coords
	 * and count them.  First, get the symbol width.
	 * We converted the original 1x, absolute, symbol coord to
	 * a 2x relative coord by performing
	 * relCoord = 2 * absCoord - (2 * xOffset + w - 1).
	 * Therefore, we want to convert the relative coords back
	 * to abs coords by calculating:
	 * absCoord = (relCoord + (2 * xOffset + symWidth - 1)) / 2;
	 * To facilitate this convert the xOffset variable into
	 * (2 * xOffset + symWidth - 1).
	 */
    scaledXOffset = 2 * xOffset + pSymbol->width - 1;

    for (pCurRun = pSymbol->pRunTail, pPrevRun = NULL;
	 pCurRun != NULL;
	 pCurRun = pNextRun)
    {
	    /* first, reverse the order of the run pointers */
	pNextRun = pCurRun->pNext;
	pCurRun->pNext = pPrevRun;
	pPrevRun = pCurRun;
	runCount++;

	    /* adjust the coords. */
	pCurRun->runLeft = (pCurRun->runLeft + (Int16) scaledXOffset)/2;
	pCurRun->runRight = (pCurRun->runRight + (Int16) scaledXOffset)/2;
    }

	/* Adjust the minX, minY, maxX and maxY coords so they're
	 * absolute again.  This is used to sort the residual symbols
	 * by position.
	 */
    pSymbol->minX = xOffset;
    pSymbol->maxX = xOffset + pSymbol->maxX;
    pSymbol->minY = yOffset;
    pSymbol->maxY = yOffset + pSymbol->maxY;

	/* make sure the symbol's pRunHead points to the head of the
	 * list.
	 */
    pSymbol->pRunHead = pPrevRun;
    pSymbol->runsPtToTail = cTrue;	/* we've reversed the runs */
    *pRunCount = runCount;		/* and counted the runs */

    NORMAL_RETURN;
}


/**************************************************************************
 *   i_assignDictionaryNumbersForSingleUse()
 *
 *	This procedure assigns dictionary numbers to each of the symbols,
 *	based on later storage of individual dictionaries for: one combined
 *	dictionary shared for all pages, and one individual dictionary for
 *	each page (instanceCount == 1, pageID used to identify the page).
 *
 *	Input:
 *		pTextCompInfo:	Ptr to all text compression info
 *		pageRequest:	Page ID of single-count items to add, or 0
 *
 *	Returns:
 *		ia_successful		things went well
 *
 **************************************************************************/
Int32  CDECL 
i_assignDictionaryNumbersForSingleUse (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo)
{
    TCSymbol	*pSymbol;
    UInt32	num, maxDictNum = 0UL;
    PROLOGUE;

    if (pTextCompInfo->maxDictionaryNum > 0)
	NORMAL_RETURN;

    if (pTextCompInfo->encodeState & cTCSymsAreInDictOrder)
    {
	for (pSymbol = pTextCompInfo->p1stDictSymbol; pSymbol != NULL;
	pSymbol = pSymbol->pNextDictSym)
	{
	    if (pSymbol->instanceCount == 1)
	    {
		pSymbol->dictNum = num = pSymbol->pageID;
		if (num > maxDictNum)
		    maxDictNum = num;
	    }
	    else
	    {
		pSymbol->dictNum = pSymbol->pageID = 0;
	    }
	}
    }
    else
    {
	for (pSymbol = pTextCompInfo->p1stSymbol; pSymbol != NULL;
	pSymbol = pSymbol->pNextNum)
	{
	    if (pSymbol->instanceCount == 1)
	    {
		pSymbol->dictNum = num = pSymbol->pageID;
		if (num > maxDictNum)
		    maxDictNum = num;
	    }
	    else
	    {
		pSymbol->dictNum = pSymbol->pageID = 0;
	    }
	}
    }
    pTextCompInfo->maxDictionaryNum = maxDictNum;

    NORMAL_RETURN;
}


/*
 $Log: i_tctne.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:13  jlin
 Imported sources

 Revision 1.17  1998/08/04 00:17:03  aber
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 CommonCheck: Retain the page ID such that the dictionary can be split into a
 shared dictionary, and a single-use dictionary for each page.
 Support multiple dictionaries per page.

 Revision 1.16  1998/06/23 21:45:46  aber
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 CommonCheck: Fixes to force Tundra dictionary reading order, esp. for multiple
 dictionaries per page. Some cosmetic comment enhancements. A couple of
 extra error conditions caught.

 Revision 1.15  1998/01/13 23:50:19  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.14  1997/10/01 23:32:50  aber
 aber on Wed Oct 1 16:32:08 PDT 1997

 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Add more buffer overflow checking, found a case where overflow wasn't
 reported.

 Revision 1.13  1997/06/25 03:46:56  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Removed bogus check to see if there are any symbols to put out when
 writing a Tundra dictionary.  Instead of simply quitting, go ahead and
 write the dictionary boilerplate even if there are no symbols in the
 dictionary.  This makes all the higher-level code happier.

 Revision 1.12  1997/01/14 02:41:23  ddavies
 ddavies on Mon Jan 13 18:39:11 PST 1997

 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Fix segment fault that occured when image was empty.  ip_calcPositionBufStats
 set dummyLastRes.pPrev->pNext to NULL after indirectly setting
 dummyLastRes.pPrev to NULL in the line before.  This only happens if there
 are no symbols on the page.

 Revision 1.11  1996/09/06 04:48:46  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Change the positioning of encoded symbols so that no symbol has a negative
 x coord.  This isn't explicitly illegal, but it breaks William's code.

 Revision 1.10  1996/09/04 01:27:33  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Several little fixes to enable this code to really write files that the
 PARC Tundra code can read.  Among them: add check to compare number of
 symbols actually written in a position block with the number we say we're
 going to write; encode deltaX as the distance between a symbol's right
 edge and the next symbol's left edge instead of between left edges (!);
 use pSymbol->height instead of calculating it from maxY and minY.

 Revision 1.9  1996/08/30 23:48:49  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Use symbol width and height parameters instead of calculating width and height.
 This is a minor efficiency improvement and eases code maintenance.

 Revision 1.8  1996/08/29 01:06:56  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Change i_writeTundraDictionary from a worthless stub to a routine that really
 writes out Tundra format dictionaries.  This is an enormous change.
 Fix a bunch of bugs in i_writeTundraPositionBlock and ip_calcPositionStats
 so that the symbol width is computed correctly whether or not the runs have
 been converted from 2x coords to 1x coord or not.

 Revision 1.7  1996/08/23 03:00:54  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Modify parsing of internal position info so the block begins with a deltaY
 relative to -1 instead of a deltaX.
 This simplifies the bizarre logic used to parse multi-line strips.

 Revision 1.6  1996/08/22 01:34:07  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Converted i_writeTundraPositionBlock from a dummy that just copied the old
 position info into a routine that really writes Tundra position blocks.
 This is an enormous change.  The new routine assigns Huffman codes to symbols,
 parses the old internal position info and encodes it into a new Tundra
 position block.  Only non-residual symbols are encoded.
 Also updated ip_calcPositoinBufStats so it calculates the number of
 non-residual symbols along with all its other stuff.
 Fixed a bug in ip_calcPositionBufStats that occured when parsing the
 internal format position info and a symbol of width 1 is encountered in
 column 0 of an image.  This has a first X coord of 0, which DOES NOT mean
 end of line as a delta X coord of 0 would.

 Revision 1.5  1996/07/19 20:23:29  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Fix WATCOM whining about unused parameters.

 Revision 1.4  1996/07/19 00:24:31  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Add procedures ip_calcPositionBufStats and ip_reverseSymbolRuns.  The
 first parses an internal position buffer, calculating lots of stuff like
 the residual bounding box, estimated size of Tundra residual and position
 blocks, etc.  The second reverses the order of the TCScanLineRuns associated
 with a symbol and converts their coords from 2x relative to 1x absolute.
 This is needed before converting a symbol to MMR.
 Include memory.pub since we allocate and free the CCITTCompressionInfo now.
 Change deleteResidual arg in temporary call to i_serializeCompressionInfo
 to cTrue since we're making residuals now.
 Replace stub code in i_writeTundraPositionBlock with temporary code that
 copies the internal position block out, deleting references to residual
 symbols.  This was done for testing.  The next major task is to replace
 this temporary code with code that really writes out a Tundra position
 block.
 Replace stub code in i_writeTundraResidualBlock with code that really
 writes out a Tundra residual block.  That's an MMR encoding of all the
 symbols that only occur once.  The block includes only the portion of
 the original image that contained residual symbols.  There's lots of
 sorting of symbols and their runs in this code so that no temporary
 bitmap of the residual symbols is made.  For each scan line, we create
 a sorted list of runs, then have ip_compMMRRow encode them.

 Revision 1.3  1996/07/08 18:00:45  ddavies
 ddavies on Mon Jul 8 11:00:25 PDT 1996

 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Fix warning about unused parameter "outputBufSize".

 Revision 1.2  1996/07/03 21:27:05  ddavies
 i_tctne.c -> /product/ipcore/ipshared/src/RCS/i_tctne.c,v

 Fix unused parameter warnings.

 Revision 1.1  1996/07/02 22:57:03  ddavies
 Initial revision

 */

