/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  i_tctnut.c: Utility procedures used for PARC's Tundra formatting of
 *		symbol-based compression and decompression.
 *
 *   	Int32	i_getTundraHeaderSize()
 *			Returns number of bytes required for a Tundra
 *			file header.
 *
 *   	Int32	i_getTundraNumberOfPagesSize()
 *			Returns number of bytes required for a Tundra
 *			item holding the number of pages in the file.
 *
 *   	Int32	i_getTundraDictBlockLengthSize()
 *			Returns number of bytes required for a Tundra
 *			item holding the number of bytes in a dictionary
 *			block.
 *
 *   	Int32	i_getTundraPageHeaderSize()
 *			Returns number of bytes required for a Tundra
 *			page block header.
 *
 *   	Int32	i_getTundraResidualHeaderSize()
 *			Returns number of bytes required for a Tundra
 *			residue block header.
 *
 *	Int32	ip_compareSymCounts()
 *			Used by qsort to compare two symbols, primarily
 *			for generation of huffman codes.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "tundra.prv"	/* for Tundra constants */

    /*  prototyes */
#include "shrip.pub"
#include "shrcmprs.prv"
#include "shrcmprs.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tctnut.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");

/* There is a big comment describing our approach to symbol based
 * compression in i_tcompr.c.
 *
 * During compression and decompression, both the shapes and positions
 * of symbols are kept in an internal format.  When the compressed
 * data is transmitted or stored, it is kept in the proposed Tundra
 * format (see "A Proposal for Tundra Coding," by William Rucklidge).
 * This file contains utility routines used during the conversion
 * process.
 *
 * A Tundra file contains one or more dictionary blocks and one or
 * more page blocks.  The dictionary blocks specify the shapes of
 * the symbols.  The page blocks specify the positions of the symbols
 * on each page plus a residue image.  The residue image has stuff that
 * was was either too large to be classified or was only found once.
 *
 * This file contains routines that are common to both Tundra encoding
 * and decoding.  i_tcjbe.c contains encoding routines.  i_tcjbd.c
 * contains decoding routines.
 */

/* PUBLIC */
/**************************************************************************
 *   i_getTundraHeaderSize()
 *
 *	This procedure returns the number of bytes needed to hold a
 *	Tundra header.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pHeaderSize:		pointer to UInt32 that will contain
 *					the number of bytes required to
 *					hold the Tundra Header.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getTundraHeaderSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt32     *pHeaderSize)
{
	/* Keep the compiler's happy */
    PROLOGUE;
    pTextCompInfoIn = pTextCompInfoIn;
    if (pHeaderSize == NULL)
	return ia_invalidParm;

    *pHeaderSize = cTundraHeaderSize;
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getTundraNumberOfPagesSize()
 *
 *	This procedure returns the number of bytes needed to hold the
 *	number of pages in a Tundra file.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pNumOfPagesSize:	pointer to UInt32 that will contain
 *					the number of bytes required to
 *					hold the number of pages in a
 *					Tundra file.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getTundraNumberOfPagesSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt32     *pNumOfPagesSize)
{
	/* Keep the compiler's happy */
    PROLOGUE;
    pTextCompInfoIn = pTextCompInfoIn;
    if (pNumOfPagesSize == NULL)
	return ia_invalidParm;

    *pNumOfPagesSize = cTundraNumOfPagesSize;
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getTundraDictBlockLengthSize()
 *
 *	This procedure returns the number of bytes needed to hold the
 *	length of a dictionary block in a Tundra file.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pDictBlkLenSize:	pointer to UInt32 that will contain
 *					the number of bytes required to
 *					hold the length of a dictionary
 *					block in a Tundra file.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getTundraDictBlockLengthSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt32     *pDictBlkLenSize)
{
	/* Keep the compiler's happy */
    PROLOGUE;
    pTextCompInfoIn = pTextCompInfoIn;
    if (pDictBlkLenSize == NULL)
	return ia_invalidParm;

    *pDictBlkLenSize = cTundraDictBlockLengthSize;
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getTundraPageHeaderSize()
 *
 *	This procedure returns the number of bytes needed to hold the
 *	header of a Page block in a Tundra file.  It is used during
 *	encoding.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pPageName:		ptr to null-terminated string
 *					containing name of the page.
 *		pPageHeaderSize:	pointer to UInt32 that will contain
 *					the number of bytes required to
 *					hold the header of a Page
 *					in a Tundra file.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getTundraPageHeaderSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    char       *pPageName,
    UInt32     *pPageHeaderSize)
{
	/* Keep the compiler's happy */
    PROLOGUE;
    pTextCompInfoIn = pTextCompInfoIn;
    if ((pPageHeaderSize == NULL) || (pPageName == NULL))
	return ia_invalidParm;

	/* Don't forget to count the nul at the end of the page name string */
    *pPageHeaderSize = strlen(pPageName) + 1 + cTundraPageHeaderFixedSize;
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *   i_getTundraResidualHeaderSize()
 *
 *	This procedure returns the number of bytes needed to hold the
 *	header of a Residual block inside a Page block in a Tundra file.
 *	It is used during encoding.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pResidualHeaderSize:	pointer to UInt32 that will contain
 *					the number of bytes required to
 *					hold the header of a Residual
 *					in a Tundra file.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getTundraResidualHeaderSize (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    UInt32     *pResidualHeaderSize)
{
	/* Keep the compiler's happy */
    PROLOGUE;
    pTextCompInfoIn = pTextCompInfoIn;
    if (pResidualHeaderSize == NULL)
	return ia_invalidParm;

    *pResidualHeaderSize = cTundraResidualHeaderSize;
    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *   i_setDictionaryNumber()
 *
 *	This procedure sets the current dictionary number in the internal
 *	TextCompressionInfo data structure.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		newDictNumber:		the new dictionary number.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_setDictionaryNumber (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Int32	newDictNumber)
{
    PROLOGUE;
    TCCompressionInfo	*pTextCompInfo;

	/* See if input ptr is null */
    if (pTextCompInfoIn == NULL)
	return ia_invalidParm;

	/* Set the new dictionary number */
    pTextCompInfo = (TCCompressionInfo *) pTextCompInfoIn;
    pTextCompInfo->currentDictionaryNum = newDictNumber;
    if (newDictNumber > pTextCompInfo->maxDictionaryNum)
	pTextCompInfo->maxDictionaryNum = newDictNumber;

    NORMAL_RETURN;
}


/**************************************************************************
 *   i_getDictionaryNumbers()
 *
 *	This procedure retrieves the current and maximum dictionary numbers
 *      from the internal TextCompressionInfo data structure.
 *
 *	Input:
 *		pTextCompInfoIn:	pointer to text compression info
 *		pCurDictNumber:		pointer to the returned current
 *					dictionary number.
 *		pMaxDictNumber:		pointer to the returned maximum
 *					dictionary number.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		one of the pointers was NULL
 *
 **************************************************************************/
Int32  CDECL 
i_getDictionaryNumbers (
    IPState    *pIPState,
    void       *pTextCompInfoIn,
    Int32      *pCurDictNumber,
    Int32      *pMaxDictNumber)
{
    PROLOGUE;
    TCCompressionInfo	*pTextCompInfo;

	/* See if input ptr is null */
    if (pTextCompInfoIn == NULL)
	return ia_invalidParm;

	/* return the two numbers */
    pTextCompInfo = (TCCompressionInfo *) pTextCompInfoIn;
    *pCurDictNumber = pTextCompInfo->currentDictionaryNum;
    *pMaxDictNumber = pTextCompInfo->maxDictionaryNum;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_compareSymCounts()
 *
 *	This procedure is used by qsort to compare two symbols.  We
 *	will sort them by instance count.  Note that since we want
 *	the first symbols to have the *biggest* instance counts
 *	and the biggest dictionary orders, we return "<" if
 *	the first instanceCount is bigger or the instanceCounts
 *	are equal and the first dictOrder is bigger.
 *
 **************************************************************************/
#ifdef __WATCOMC__	/* Watcom doesn't seem to be pure ANSI */
int
#else
int CDECL
#endif
ip_compareSymCounts(
    const void *ppSym1,
    const void *ppSym2)
{
    const TCSymbol *pSym1 = *(TCSymbol * const *)ppSym1;
    const TCSymbol *pSym2 = *(TCSymbol * const *)ppSym2;

    if ((pSym1->instanceCount > pSym2->instanceCount) ||
	((pSym1->instanceCount == pSym2->instanceCount) &&
	 (pSym1->dictOrder > pSym2->dictOrder)))
	return (-1);
    else
	return (1);
}




/*
 $Log: i_tctnut.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:13  jlin
 Imported sources

 Revision 1.5  1998/06/04 16:36:43  aber
 aber on Thu Jun 4 09:36:09 PDT 1998

 i_tctnut.c -> /product/ipcore/ipshared/src/RCS/i_tctnut.c,v

 Allow special case for Watcom, which doesn't seem to adhere to the ANSI
 protocol for using CDECL in the compare function passed to qsort().

 Revision 1.4  1998/05/21 19:25:19  aber
 aber on Thu May 21 12:24:52 PDT 1998

 i_tctnut.c -> /product/ipcore/ipshared/src/RCS/i_tctnut.c,v

 As per Tim's suggestion, use CDECL for compare function passed to qsort.

 Revision 1.3  1998/01/13 23:50:23  ddavies
 i_tctnut.c -> /product/ipcore/ipshared/src/RCS/i_tctnut.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1997/12/05 20:01:41  ddavies
 i_tctnut.c -> /product/ipcore/ipshared/src/RCS/i_tctnut.c,v

 Move ip_compareSymCounts from mitcenc.c.  This procedure is really used
 for tundra compression anyway.

 Revision 1.1  1996/07/02 22:53:10  ddavies
 Initial revision

 */

