/*************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * i_thresh.c -- procedures for computing thresholds
 *   i_findEdgeThreshold()	Compute a threshold, edge-histogram based
 *                          
 */


#include <stdio.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /* prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrpixr.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_thresh.c,v $");

/* Macro used below */
#define ACCUMDIFF(otherPixel) \
    delta = pixel - otherPixel;\
    if (delta)\
    {\
	delta = ABS(delta);\
        if (delta > maxDelta)\
        {\
            sum = pixel + otherPixel;\
            count = 1;\
            maxDelta = delta;\
        }\
        else if (delta == maxDelta)\
        {\
            sum += (pixel + otherPixel);\
            count ++;\
        }\
    }

/* PUBLIC */
/*************************************************************************
 *  i_findEdge2Thresholds()  
 * 
 *    Compute 2 thresholds for a rectangular region, 8-bit case
 *
 *    Here we find threshold one for propagating around and 
 *    threshold two for inside.
 * 
 * 
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *              bpl =   Bytes between lines
 *     		x,y =	upper left corner of rectangle relative to upper 
 *               	left corner of *frame*, not image.
 *		w =     width of rectangle in pixels
 *      	h =     height of rectangle in pixels
 *              minDiff=Minimum distance to consider for a threshold
 *              pThresh1 = Place to stuff computed threhsold.
 *              pThresh2 = Place to stuff computed threshold.
 *
 *	Return code:
 *		ia_successful	good completion
 ************************************************************************/
Int32  CDECL 
i_findEdge2Thresholds (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	minDiff,
    Int32      *pThresh1,
    Int32      *pThresh2)
{
    PROLOGUE;
    Int32 i, j, delta, maxDelta, sum, count;
    UInt8 tmp, pixel;
    UInt32 wordThis, wordNext, wordDL, wordD, wordDR;
    Int32  sum_p, count_p;

    /* Land on an even long-word, and end there */
    pS += (y * bpl) + (x & ~3);
    w += (x & 3);
    if( (w & 3) != 0 )
	w += 4 - (w & 3);

    maxDelta = 0; 
    sum = 0; 
    count = 0;
    count_p = 0;
    sum_p = 0;

    /* This could probably be faster by recognizing that we often
     * shift out the same value more than once. That would require
     * that we have a lot of registers. 
     */

    for (i = 0; i < h - 1; i++, pS += bpl)
    {
	/* Get the pipe going */
	wordThis = ULONG_ACCESS (pS);
	wordD = ULONG_ACCESS (pS + bpl);

	    /* Don't go into the frame */
	if (x < 36)
	    wordDL = wordD;
	else
	    wordDL = ULONG_ACCESS (pS + bpl - 4);


	/* We do 4 pixels at a time */
	for (j = 0; j < w; j += 4, count_p += 4)
	{
	    wordNext = ULONG_ACCESS (pS + j + 4);
	    wordDR = ULONG_ACCESS (pS + j + bpl + 4);

	    /* Get pixel 1 */
	    pixel = (UInt8) (wordThis >> 24);
	    sum_p += (Int32) pixel;

		/* Right */
	    tmp = (UInt8) (wordThis >> 16);
	    ACCUMDIFF (tmp);
		
		/* Down, left */
	    tmp = (UInt8) wordDL;
	    ACCUMDIFF (tmp);
		
		/* Down */
	    tmp = (UInt8) (wordD >> 24);
	    ACCUMDIFF (tmp);
		
		/* Down, right */
	    tmp = (UInt8) (wordD >> 16);
	    ACCUMDIFF (tmp);
		
		/* Get pixel 2 */
	    pixel = (UInt8) (wordThis >> 16);
	    sum_p += (Int32) pixel;

	    /* Right */
	    tmp = (UInt8) (wordThis >> 8);
	    ACCUMDIFF (tmp);

		/* Down, left */
	    tmp = (UInt8) (wordD >> 24);
	    ACCUMDIFF (tmp);
		
		/* Down */
	    tmp = (UInt8) (wordD >> 16);
	    ACCUMDIFF (tmp);
		
		/* Down, right */
	    tmp = (UInt8) (wordD >> 8);
	    ACCUMDIFF (tmp);

		/* Get pixel 3 */
	    pixel = (UInt8) (wordThis >> 8);
	    sum_p += (Int32) pixel;
		
	    /* Right */
	    tmp = (UInt8) wordThis;
	    ACCUMDIFF (tmp);

		/* Down, left */
	    tmp = (UInt8) (wordD >> 16);
	    ACCUMDIFF (tmp);
		
		/* Down */
	    tmp = (UInt8) (wordD >> 8);
	    ACCUMDIFF (tmp);

		/* Down, right */
	    tmp = (UInt8) wordD;
	    ACCUMDIFF (tmp);

		/* Get pixel 4 */
	    pixel = (UInt8) wordThis;
	    sum_p += (Int32) pixel;
		
	    /* Right */
	    tmp = (UInt8) (wordNext >> 24);
	    ACCUMDIFF (tmp);

		/* Down, left */
	    tmp = (UInt8) (wordD >> 8);
	    ACCUMDIFF (tmp);

		/* Down */
	    tmp = (UInt8) wordD;
	    ACCUMDIFF (tmp);

		/* Down, right */
	    tmp = (UInt8) (wordDR >> 24);
	    ACCUMDIFF (tmp);

		/* Roll the values */
	    wordThis = wordNext;
	    wordDL = wordD;
	    wordD = wordDR;

	}
    }

    if (maxDelta < minDiff)
    {
	*pThresh1 = 0;
	*pThresh2 = 0;
    }
    else
    {
	*pThresh1 = *pThresh2 = sum / (count * 2);
	sum_p /= count_p;
	if( sum_p > *pThresh1 )
	{
	    *pThresh2 += (maxDelta / 6); 
	    if( *pThresh2 > sum_p ) *pThresh2 = sum_p;
	}
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/*************************************************************************
 *  i_findEdgeThreshold()  
 * 
 *    Compute a threshold for a rectangular region using an edge model.
 *
 *    This thresholding algorithm works on the following thought:
 *    if one looked at a typical gray edge, one would find that
 *    the black-white transitions started out at a small slope, increased
 *    in the middle, and then returned to a small slope near the end.
 *    The idea is that the point at which the slope is maximized is the
 *    "actual" edge. My theoretical reasoning, simplistic though it may
 *    be, is that if I take an infinite-frequency edge and apply a 
 *    symmetric blurring convolution, the "real" edge would
 *    indeed be at the point of maximum absolute-value slope in the 
 *    blurred result.
 *
 *    How do we implment this in a reasonably robust way? We probably 
 *    can't afford to do anything as fancy as trace the outlines of 
 *    connected components, and assume that edges lie along the normal.
 *    What we do instead is histogram the differences between adjacent
 *    horizontal pixels, and along with the histogram bits, maintain
 *    a parallel set of bins which record the values of both pixels
 *    which comprised the difference, i.e. if pixel1 is 100, and pixel2
 *    is 170, the difference is 70, so add 1 to bin #70 in the histogram,
 *    and add 270 (170 + 100) to bin #70 in the other array. When we're
 *    done, we look backwards down the histogram to find the peak difference.
 *    The idea is that the peak difference is most likely going to be 
 *    found on vertical edges, so the values that we counted will be
 *    valid as endpoints of the maximal-sloped difference. The value
 *    computed is then the value in the sum bin divided by twice the value
 *    in the histogram bin.
 *
 *    This has been tested quite a bit in the four-bit case, where I found
 *    that using 64x64 regions, text regions usually got quite a significant
 *    number of the same peak difference, thus making our statistic more
 *    meaningful. It remains to be seen whether the 8-bit case will
 *    do the same, or whether we'll want to truncate the lower bits in
 *    order to force slightly different values to the same bin to reduce
 *    noise in the result.
 *
 *    I've also used a variation of this code in arbitrary regions,
 *    in the 8-bit case.
 *
 *    This routine has been taken from ipwrap, so that it can be
 *    used by ipshared clients.
 * 
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *              depth = Depth, bits/pixel, of image
 *              bpl =   Bytes between lines
 *     		x,y =	upper left corner of rectangle relative to upper 
 *               	left corner of *frame*, not image.
 *		w =     width of rectangle in pixels
 *      	h =     height of rectangle in pixels
 *              minDiff=Minimum distance to consder for a threshold
 *              pThresh = Place to stuff computed threhsold, or -1 if 
 *                        none found.
 *
 *
 *	Return code:
 *		ia_successful	good completion
 ************************************************************************/
Int32  CDECL 
i_findEdgeThreshold (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	depth,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	minDiff,
    Int32      *pThresh)
{

    PROLOGUE;
    if (depth == 8)
    {
	ip_findEdgeThreshold8 (pIPState, pS, bpl, x, y, w, h, minDiff, pThresh);
    }
    else
    {
	ip_findEdgeThreshold4 (pIPState, pS, bpl, x, y, w, h, minDiff, pThresh);
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/*************************************************************************
 *  ip_findEdgeThreshold8()  
 * 
 *    Compute a threshold for a rectangular region, 8-bit case
 *
 *    This thresholding algorithm works on the following thought:
 *    if one looked at a typical gray edge, one would find that
 *    the black-white transitions started out at a small slope, increased
 *    in the middle, and then returned to a small slope near the end.
 *    The idea is that the point at which the slope is maximized is the
 *    "actual" edge. My theoretical reasoning, simplistic though it may
 *    be, is that if I take an infinite-frequency edge and apply a 
 *    symmetric blurring convolution, the "real" edge would
 *    indeed be at the point of maximum absolute-value slope in the 
 *    blurred result.
 *
 *    This routine has been taken from ipwrap, so that it can be
 *    used by ipshared clients.
 * 
 * 
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *              bpl =   Bytes between lines
 *     		x,y =	upper left corner of rectangle relative to upper 
 *               	left corner of *frame*, not image.
 *		w =     width of rectangle in pixels
 *      	h =     height of rectangle in pixels
 *              minDiff=Minimum distance to consider for a threshold
 *              pThresh = Place to stuff computed threhsold.
 *
 *
 *	Return code:
 *		ia_successful	good completion
 ************************************************************************/
Int32  CDECL 
ip_findEdgeThreshold8 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	minDiff,
    Int32      *pThresh)
{
    PROLOGUE;
    Int32 i, j, delta, maxDelta, sum, count;
    UInt8 tmp, pixel;
    UInt32 wordThis, wordNext, wordDL, wordD, wordDR;


    /* Land on an even long-word, and end there */
    pS += (y * bpl) + (x & ~3);

    /* This is a bug, how could calculate w in this way?
       w += 4 - (x & ~3);
       w += 4 - (w & ~3);
       But the binarization2 results are so good, why? I fixed
       it and another bug at wordNext and wordDR.
    */
    w += (x & 3);
    if( (w & 3) != 0)
	w += 4 - (w & 3);

    maxDelta = 0; 
    sum = 0; 
    count = 0;

    /* This could probably be faster by recognizing that we often
     * shift out the same value more than once. That would require
     * that we have a lot of registers. 
     */

    for (i = 0; i < h - 1; i++, pS += bpl)
    {
	/* Get the pipe going */
	wordThis = ULONG_ACCESS (pS);
	wordD = ULONG_ACCESS (pS + bpl);

	    /* Don't go into the frame */
	if (x < 36)
	    wordDL = wordD;
	else
	    wordDL = ULONG_ACCESS (pS + bpl - 4);

	/* We do 4 pixels at a time */
	for (j = 0; j < w; j += 4)
	{
	    /* Get next */
	    wordNext = ULONG_ACCESS (pS + j + 4);
	    wordDR = ULONG_ACCESS (pS + j + bpl + 4);

	    /* Get pixel 1 */
	    pixel = (UInt8) (wordThis >> 24);

		/* Right */
	    tmp = (UInt8) (wordThis >> 16);
	    ACCUMDIFF (tmp);
		
		/* Down, left */
	    tmp = (UInt8) wordDL;
	    ACCUMDIFF (tmp);
		
		/* Down */
	    tmp = (UInt8) wordD;
	    ACCUMDIFF (tmp);
		
		/* Down, right */
	    tmp = (UInt8) (wordD >> 16);
	    ACCUMDIFF (tmp);
		
		/* Get pixel 2 */
	    pixel = (UInt8) (wordThis >> 16);

	    /* Right */
	    tmp = (UInt8) (wordThis >> 8);
	    ACCUMDIFF (tmp);

		/* Down, left */
	    tmp = (UInt8) (wordD >> 24);
	    ACCUMDIFF (tmp);
		
		/* Down */
	    tmp = (UInt8) (wordD >> 16);
	    ACCUMDIFF (tmp);
		
		/* Down, right */
	    tmp = (UInt8) (wordD >> 8);
	    ACCUMDIFF (tmp);

		/* Get pixel 3 */
	    pixel = (UInt8) (wordThis >> 8);
		
	    /* Right */
	    tmp = (UInt8) wordThis;
	    ACCUMDIFF (tmp);

		/* Down, left */
	    tmp = (UInt8) (wordD >> 8);
	    ACCUMDIFF (tmp);
		
		/* Down */
	    tmp = (UInt8) wordD;
	    ACCUMDIFF (tmp);

		/* Down, right */
	    tmp = (UInt8) (wordDR >> 24);
	    ACCUMDIFF (tmp);

		/* Get pixel 4 */
	    pixel = (UInt8) wordThis;
		
	    /* Right */
	    tmp = (UInt8) (wordNext >> 24);
	    ACCUMDIFF (tmp);

		/* Down, left */
	    tmp = (UInt8) (wordD >> 8);
	    ACCUMDIFF (tmp);

		/* Down */
	    tmp = (UInt8) wordD;
	    ACCUMDIFF (tmp);

		/* Down, right */
	    tmp = (UInt8) (wordDR >> 24);
	    ACCUMDIFF (tmp);

		/* Roll the values */
	    wordThis = wordNext;
	    wordDL = wordD;
	    wordD = wordDR;

	}
    }

    if (maxDelta < minDiff)
    {
	*pThresh = -1;
    }
    else
    {
	*pThresh = sum / (count * 2);
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/*************************************************************************
 *  ip_findEdgeThreshold4()  
 * 
 *    Compute a threshold for a rectangular region, 4-bit case.
 *
 *    This thresholding algorithm works on the following thought:
 *    if one looked at a typical gray edge, one would find that
 *    the black-white transitions started out at a small slope, increased
 *    in the middle, and then returned to a small slope near the end.
 *    The idea is that the point at which the slope is maximized is the
 *    "actual" edge. My theoretical reasoning, simplistic though it may
 *    be, is that if I take an infinite-frequency edge and apply a 
 *    symmetric blurring convolution, the "real" edge would
 *    indeed be at the point of maximum absolute-value slope in the 
 *    blurred result.
 *
 *    This routine has been taken from ipwrap, so that it can be
 *    used by ipshared clients.
 * 
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *              bpl =   Bytes between lines
 *     		x,y =	upper left corner of rectangle relative to upper 
 *               	left corner of *frame*, not image.
 *		w =     width of rectangle in pixels
 *      	h =     height of rectangle in pixels
 *              minDiff=Minimum distance to consider for a threshold
 *              pThresh = Place to stuff computed threhsold.
 *
 *
 *	Return code:
 *		ia_successful	good completion
 ************************************************************************/
Int32  CDECL 
ip_findEdgeThreshold4 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	minDiff,
    Int32      *pThresh)
{
    PROLOGUE;
    Int32 i, j, delta, maxDelta, sum, count;
    UInt8 tmp, pixel;
    UInt8 byteThis, byteNext, byteDL, byteD, byteDR;

    pS += (y * bpl) + x/2;
    sum = 0;
    count = 0;
    maxDelta = 0;

    for (i = 0; i < h - 1; i++, pS += bpl)
    {
	byteThis = UCHAR_ACCESS (pS);
	byteNext = UCHAR_ACCESS (pS + 1);
	byteDL = UCHAR_ACCESS (pS + bpl - 1);
	byteD = UCHAR_ACCESS (pS + bpl);
	byteDR = UCHAR_ACCESS (pS + bpl + 1);
		

	/* Do a pair of pixels */
	for (j = 0; j < w/2; j++)
	{
	    /* Get the ms pixel */
	    pixel = byteThis >> 4;

	    /* Right */
	    tmp = byteThis & 0xf;
	    ACCUMDIFF (tmp);
		
	    /* Down and to the left */
	    tmp = byteDL & 0xf;
	    ACCUMDIFF (tmp);

	    /* Down */
	    tmp = byteD >> 4;
	    ACCUMDIFF (tmp);

	    /* Down and to the right */
	    tmp = byteD & 0xf;
	    ACCUMDIFF (tmp);

	    /* Next pixel. */
	    pixel = byteThis & 0xf;

	    /* Right */
	    tmp = byteNext >> 4;
	    ACCUMDIFF (tmp);

	    /* Down and to the left */
	    tmp = byteD >> 4;
	    ACCUMDIFF (tmp);

	    /* Down */
	    tmp = byteD & 0xf;
	    ACCUMDIFF (tmp);

	    /* Down, right */
	    tmp = byteDR >> 4;
	    ACCUMDIFF (tmp);

	    /* Update everything */
	    byteThis = byteNext;
	    byteNext = UCHAR_ACCESS (pS + j + 1);
	    byteDL = byteD;
	    byteD = byteDR;
	    byteDR = UCHAR_ACCESS (pS + j + bpl + 1);
	}
    }

    if (maxDelta < minDiff)
    {
	*pThresh = -1;
    }
    else
    {
	*pThresh = sum / (count * 2);
    }
    NORMAL_RETURN;
}

/*
$Log: i_thresh.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.6  1998/06/05 20:07:33  xzhu
xzhu on Fri Jun 5 13:06:34 PDT 1998

i_thresh.c -> /product/ipcore/ipshared/src/RCS/i_thresh.c,v

Fixed mistake in calculating width left delta.

Revision 1.5  1998/06/03 00:02:08  xzhu
xzhu on Tue Jun 2 17:01:13 PDT 1998

i_thresh.c -> /product/ipcore/ipshared/src/RCS/i_thresh.c,v

Fixed problem in calculating delta in 2 thresholds. The problems are still in i_findEdgeThreshold.

Revision 1.4  1998/05/21 21:12:01  xzhu
xzhu on Thu May 21 14:10:27 PDT 1998

i_thresh.c -> /product/ipcore/ipshared/src/RCS/i_thresh.c,v

Fixed bugs and added a new function.

Revision 1.3  1998/01/13 23:50:24  ddavies
i_thresh.c -> /product/ipcore/ipshared/src/RCS/i_thresh.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1997/05/02 01:20:29  ddavies
i_thresh.c -> /product/ipcore/ipshared/src/RCS/i_thresh.c,v

Remove comment about histo and hVal, which don't exist.

Revision 1.1  1996/07/19 15:21:42  macleod
Initial revision

*/


