/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


#include "types.pub"
#include "ipscore.prv"
#include "tpdil.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "frames.pub"

/* prototypes */
#include "shrpatt.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_tpdil.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");


/*
 * The file was written at Xerox PARC by Michael Leventon, a summer intern in
 * 1994, and by William Rucklidge, a researcher in the Systems Practices
 * Laboratory at PARC, in the Image Understanding Area.
 * Some additional editing was completed by Brian Freyburger, a summer intern
 * in 1994 and part-time employee during the year.
 * Eric Jaquith did minor cosmetic changes to bring it into ipshared.
 *
 * In some of the macros, the src image is referred to as the "model" and
 * the dest image is referred to as the "image".
 *
 * For notes on the dilation rules, see the file "tpdil.prv".
 */



/*
 * What this macro "MoveAndMask" does:
 * It's called with 3 params:
 *	v - a UInt32 value
 *	b - a bit index specifying a bit in v
 *	nb - a bit index specifying a bit in the output.
 * Its job is to extract bit number b (numbered from the MSB, sigh) from
 * v, and shift it so that it appears in bit nb (numbered from the LSB) of
 * the result.
 *
 * nb must be between 0 and 31 inclusive.
 * b might be between -2 and 31. Yay. b<0 should be treated as (b+32) - whoever
 * called us should have set up the v value so that this is right.
 */

/*
 * How it does what it does:
 * There are two ways to do the basic operation:
 *	1) Shift bit b to the LSB, mask off the LSB, shift the LSB to bit nb.
 *	2) Mask off bit b; shift bit b to bit nb.
 *	(We could also shift b to nb and mask off nb, but that's pretty
 *	 similar).
 *
 */

/* The shift-mask-shift version */
#define	MoveAndMaskSMS(v, b, nb)					\
	((((v) >> (31 - ((b) & 31))) & 1) << (nb))


#if defined(_SPARC_ARCH_) && defined(__GNUC__)
/* It turns out that gcc -O2 generates different instruction sequences
 * for these, if the operands are constants (and they'd better be, or this
 * will generate truly lousy code). So we switch... the tradeoffs here are
 * for the SPARC.
 *
 * Another tweak is that the shift-then-mask method has to shift left sometimes
 * and right other times, so it uses a conditional expression; since the
 * thing the expression tests is constant, one branch will always be dead code.
 * But that's not good enough: the dead branch still contains an illegal
 * operation (shift by a negative number), and the compiler complains. So
 * we stick an ABS() around the shift distance... shuts the compiler up and
 * doesn't change the code. Sigh.
 */
/* The mask-shift version */
#define	MoveAndMaskMS(v, b, nb)						\
	(((31 - ((b) & 31)) > (nb)) ?					\
	 (((v) & (1ul << (31 - ((b) & 31)))) >> ABS((31 - ((b) & 31)) - (nb))) : \
	 (((v) & (1ul << (31 - ((b) & 31)))) << ABS((nb) - (31 - ((b) & 31)))))
/*
 * The hybrid version. The rules here are:
 * SMS produces better code if (31 - ((b) & 31)) >= 12 and nb < 12,
 * or if b == 0, or if nb == 31. MS generates better code in other cases.
 *
 * Where did the magic number 12 come from, you ask? Well, it's because the
 * SPARC has a 12-bit immediate field... numbers over that have to be
 * synthesised into registers before they can be used. If we have enough
 * registers, we can re-use them, but...
 *
 * Of course, this means that on anything but a SPARC, and any compiler
 * but gcc, this may not generate the best code. Life's hard.
 */
#define	MoveAndMask(v, b, nb)					\
	(((b == 0) || ((nb) == 31) ||				\
	  (((31 - ((b) & 31)) >= 12) && ((nb) < 12))) ?		\
	 MoveAndMaskSMS((v), (b), (nb)) :			\
	 MoveAndMaskMS((v), (b), (nb)))
    

#else /* defined(_SPARC_ARCH_) && defined(__GNUC__) */

#define	MoveAndMask(v, b, nb)  MoveAndMaskSMS(v, b, nb)

#endif /* defined(_SPARC_ARCH_) && defined(__GNUC__) */
    


/*
 * Useful things to know:
 * Bits within a word are stored MSB-is-left (lowest x),
 * LSB-is-right (highest x). The leftmost bit is therefore Bit(31), but
 * x == 0 (mod 32).
 *
 * The dilation lookup table is accessed as
 * dilatelookup[env] - 0 or 1 means that is the output.
 * The env index into the table is made up of the current bit and 12 bits
 * surrounding it, packed into a 13-bit value. The locations of the bits
 * in this value are:
 *         0
 *     1   2   3
 * 4   5   6   7   8
 *     9  10  11
 *        12
 * so the bit two rows above the current bit gives the LSB value of env, etc.
 */

/*
 * Some macro definitions to cover up the bit ordering shown above.
 * The bits listed here are the "leading edge" of the mask (see DoBit),
 * and a couple more that show up various places.
 */
#define	BIT_NN	0
#define	BIT_NW	1
#define	BIT_N	2
#define	BIT_WW	4
#define	BIT_W	5
#define	BIT_C	6
#define	BIT_SW	9
#define	BIT_S	10
#define	BIT_SS	12

/*
 * A macro for general bit-within-UInt32 access. Numbering is LSB==0.
 */
#define Bit(b)     (1ul << (b))

/*
 * This macro handles the lookup. bits holds the current environment.
 * m_uu holds the longword two rows up from the current position; m_u is the
 * longword above; m, m_d and m_dd are what you'd expect.
 * outbits holds the result (the dilated result longword), and is
 * built LSB-to-MSB (i.e. right to left) by filling in the MSB and shifting
 * right.
 * bits is assumed to be set up correctly for the position we've just finished
 * with. Shifting it left is equivalent to moving the current position left.
 * (see the bit numbering above), except for bits 0, 1, 4, 9, 12, which
 * need to be reloaded.
 */
#define DoBit(b)							\
    bits <<= 1;								\
    bits  &= 0x1FFF & ~(Bit(BIT_NN) | Bit(BIT_NW) | Bit(BIT_WW) | Bit(BIT_SW) | Bit(BIT_SS));           \
    bits  |= MoveAndMask(m_uu, b,   BIT_NN);				\
    bits  |= MoveAndMask(m_u,  b-1, BIT_NW);				\
    bits  |= MoveAndMask(m,    b-2, BIT_WW);				\
    bits  |= MoveAndMask(m_d,  b-1, BIT_SW);				\
    bits  |= MoveAndMask(m_dd, b,   BIT_SS);				\
    outbits |= lookup[bits] << 31;					\
    /*if (b>0) -- always true! */ outbits >>= 1;


/* This macro MUST be used instead of DoBit(0), because when b==0,
 * we don't want to right shift outbits by one at the end!
 * (This is necessary because the dead branch made in the above macro when
 *  b==0 causes warnings on the PC with WATCOM compilers) */
#define DoBitZero							\
    bits <<= 1;								\
    bits  &= 0x1FFF & ~(Bit(BIT_NN) | Bit(BIT_NW) | Bit(BIT_WW) | Bit(BIT_SW) | Bit(BIT_SS));           \
    bits  |= MoveAndMask(m_uu,   0, BIT_NN);				\
    bits  |= MoveAndMask(m_u,   -1, BIT_NW);				\
    bits  |= MoveAndMask(m,     -2, BIT_WW);				\
    bits  |= MoveAndMask(m_d,   -1, BIT_SW);				\
    bits  |= MoveAndMask(m_dd,   0, BIT_SS);				\
    outbits |= lookup[bits] << 31;


/*
 * This is basically DoBit(1), except that we don't load up BIT_WW;
 * it's used at the left end of the row.
 */
#define DoBit1_zeros							\
    bits <<= 1;								\
    bits  &= 0x1FFF & ~(Bit(BIT_NN) | Bit(BIT_NW) | Bit(BIT_WW) | Bit(BIT_SW) | Bit(BIT_SS));           \
    bits  |= MoveAndMask(m_uu, 1, BIT_NN);				\
    bits  |= MoveAndMask(m_u,  0, BIT_NW);				\
    bits  |= MoveAndMask(m_d,  0, BIT_SW);				\
    bits  |= MoveAndMask(m_dd, 1, BIT_SS);				\
    outbits |= lookup[bits] << 31;					\
    outbits >>= 1;

/*
 * This is basically DoBit(0), except that we don't load up BIT_*W;
 * it's used at the left end of the row.
 */
#define DoBit0_zeros							\
    bits <<= 1;								\
    bits  &= 0x1FFF & ~(Bit(BIT_NN) | Bit(BIT_NW) | Bit(BIT_WW) | Bit(BIT_SW) | Bit(BIT_SS));           \
    bits  |= MoveAndMask(m_uu, 0, BIT_NN);				\
    bits  |= MoveAndMask(m_dd, 0, BIT_SS);				\
    outbits |= lookup[bits] << 31;

/*
 * This macro starts up the shift process as we enter a row from the right.
 * It loads in the first (leftmost) two columns of the environment.
 * Special cases are handled separately. b should be set equal to the
 * width of the image mod 32, and thus refer to a bit just to the right
 * of the rightmost valid bit.
 */
#define SetupBit(b)							\
    bits  = MoveAndMask(m,   b - 2, BIT_WW);				\
    bits |= MoveAndMask(m,   b - 1, BIT_W);				\
    bits |= MoveAndMask(m_u, b - 1, BIT_NW);				\
    bits |= MoveAndMask(m_d, b - 1, BIT_SW);


/* PRIVATE */
/*****************************************************************************
 * Take a row and dilate it
 *
 * Description:
 *  This function handles a row of the dilation. The stuff it needs set up:
 *  mrw, irw: number of longwords in each model, image row.
 *  mp_* need to be loaded to point to the word *just past* the
 *  last longword in the appropriate row of the model; ditto ip.
 *
 * Notes:
 *  This routine handles all the edge case rows. Basically, pass in NULL
 *  for any pointer you don't want to think about; it will be treated as
 *  pointing to a row of zeros.  The edge case rows are the rows in the
 *  image that do not have at least 2 rows above and below them of valid
 *  image bitmap.
 ****************************************************************************/
static void CDECL 
ip_doRow (
    IPState    *pIPState,
    UInt32     *mp_uu,		/* Ptr to longword two rows up
				 * from the current position*/
    UInt32     *mp_u,		/* Ptr to longword one row 
				 * up from the current position*/
    UInt32     *mp,		/* Ptr to longword at 
				 * the current position */
    UInt32     *mp_d,		/* Ptr to longword one row
				 * down from the current position*/
    UInt32     *mp_dd,		/* Ptr to longword two rows
				 * down from the current position*/
    UInt32     *ip,		/* Ptr to longword in dest image */
    UInt32	wordsleft,	/* The number of words left in
				 * the image row that we are processing */
    UInt8	bitsinlastword) /* Number of bits in the last longword of
				 * each model
				 * row (ie without the padding).  Note that the
				 * value "0" means all 32 bits are
				 * in longword. */
{
    PROLOGUE;
    UInt32	m_uu, m_u, m, m_d, m_dd; /* Longs around current pos */
    UInt32	outbits; /* The dilated output */
    UInt32	bits = 0;
    const UInt8 *lookup = dilateLookup;

    /* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Load up the last word. */					
    m = *--mp;								

    m_uu = m_u = m_d = m_dd = 0;
    
    if (mp_u  != (UInt32 *)NULL)  m_u  = *--mp_u;
    if (mp_d  != (UInt32 *)NULL)  m_d  = *--mp_d;
    if (mp_uu != (UInt32 *)NULL)  m_uu = *--mp_uu;
    if (mp_dd != (UInt32 *)NULL)  m_dd = *--mp_dd;
									
    /*									
     * OK - we're set wordsleft to the number of words we haven't	
     * touched yet (we're about to touch the last word, so don't	
     * count it).							
     * Load up outbits with 0s so that the rightmost word is padded	
     * with 0s. These are really					
     * outside the bounds of the image, but still...			
     */									
    wordsleft--;
    outbits = 0;							
									
    /*									
     * This switch loads up "bits" properly. It loads up the		
     * environment with *some* of the bits:    .			
     *                                       X . .			
     *                                     X X . . .			
     *                                       X . .			
     *                                         .			
     * with the other ones (".") set to zero. Why those? So that the	
     * first DoBit shifts appropriately to start on the rightmost	
     * model and image bit.						
     */									
    switch (bitsinlastword)
    {						
    case 0:
    {								
	/*								
	 * 0 really means 32 bits in the last word: it's full.		
	 * We want to set up the state as though we'd just done the bit	
	 * to the right of the last word (which we'll do now).		
	 */								
	SetupBit(32);							
									
	/* Correct for the 0/32 confusion */				
	wordsleft++;							
	break;								
    }								
									
    case 1:
    {								
	/*								
	 * Only one bit left in the rightmost model word.		
	 * Load up the bits as required. We need to start on the	
	 * next word. There had better be a next word...		
	 */								
									
	bits  = MoveAndMask(m,   0, BIT_W);				
	bits |= MoveAndMask(m_u, 0, BIT_NW);				
	bits |= MoveAndMask(m_d, 0, BIT_SW);				
									
	m = *--mp;							
	bits |= MoveAndMask(m,   31, BIT_WW);				
									
	break;								
    }								
									
	/*								
	 * All the remaining cases are simple. Yay. Do a SetupBit on each	
	 * one, which will initialise bits.				
	 */								
    case 2:  SetupBit(2);  break;					
    case 3:  SetupBit(3);  break;					
    case 4:  SetupBit(4);  break;					
    case 5:  SetupBit(5);  break;					
    case 6:  SetupBit(6);  break;					
    case 7:  SetupBit(7);  break;					
    case 8:  SetupBit(8);  break;					
    case 9:  SetupBit(9);  break;					
    case 10: SetupBit(10); break;					
    case 11: SetupBit(11); break;					
    case 12: SetupBit(12); break;					
    case 13: SetupBit(13); break;					
    case 14: SetupBit(14); break;					
    case 15: SetupBit(15); break;					
    case 16: SetupBit(16); break;					
    case 17: SetupBit(17); break;					
    case 18: SetupBit(18); break;					
    case 19: SetupBit(19); break;					
    case 20: SetupBit(20); break;					
    case 21: SetupBit(21); break;					
    case 22: SetupBit(22); break;					
    case 23: SetupBit(23); break;					
    case 24: SetupBit(24); break;					
    case 25: SetupBit(25); break;					
    case 26: SetupBit(26); break;					
    case 27: SetupBit(27); break;					
    case 28: SetupBit(28); break;					
    case 29: SetupBit(29); break;					
    case 30: SetupBit(30); break;					
    case 31: SetupBit(31); break;					
    }									
									
    /* Now finish off the first word. */				
    switch (bitsinlastword)
    {						
	/* Note that these all fall through. */				
									
    case 31: DoBit(30);						
    case 30: DoBit(29);						
    case 29: DoBit(28);						
    case 28: DoBit(27);						
    case 27: DoBit(26);						
    case 26: DoBit(25);						
    case 25: DoBit(24);						
    case 24: DoBit(23);						
    case 23: DoBit(22);						
    case 22: DoBit(21);						
    case 21: DoBit(20);						
    case 20: DoBit(19);						
    case 19: DoBit(18);						
    case 18: DoBit(17);						
    case 17: DoBit(16);						
    case 16: DoBit(15);						
    case 15: DoBit(14);						
    case 14: DoBit(13);						
    case 13: DoBit(12);						
    case 12: DoBit(11);						
    case 11: DoBit(10);						
    case 10: DoBit(9);						
    case 9:  DoBit(8);						
    case 8:  DoBit(7);						
    case 7:  DoBit(6);						
    case 6:  DoBit(5);						
    case 5:  DoBit(4);						
    case 4:  DoBit(3);						
    case 3:  DoBit(2);						
    case 2:
    {								
	/*								
	 * Need to reload m.						
	 */								
	if (wordsleft)
	{						
	    m = *--mp;							
	    DoBit(1);							
	}								
	else
	{								
	    DoBit1_zeros;						
	}								
    }								
									
    case 1:
    {								
	/*								
	 * Need to reload m_u and m_d					
	 */								
	if (wordsleft)
	{						
	    if (mp_u  != (UInt32 *)NULL)  m_u  = *--mp_u;
	    if (mp_d  != (UInt32 *)NULL)  m_d  = *--mp_d;
	    DoBitZero;							
	    if (mp_uu != (UInt32 *)NULL)  m_uu = *--mp_uu;
	    if (mp_dd != (UInt32 *)NULL)  m_dd = *--mp_dd;
	}								
	else
	{								
	    DoBit0_zeros;						
	}								
	*--ip = outbits;						
    }								
	/*								
	 * Note that there's no clause for 0. This switch is just	
	 * to handle partial words, and there aren't any there.		
	 */								
    }								
									
    /*									
     * OK - into the main part.						
     * The state here is that we have wordsleft words left to do.	
     * The m* variables are all loaded up				
     * properly to start a new word, as is bits.			
     */									
    for (; wordsleft; wordsleft--)
    {					
	outbits = 0;							
									
	/* Do all the easy stuff. */					
	DoBit(31); DoBit(30); DoBit(29); DoBit(28); DoBit(27);		
	DoBit(26); DoBit(25); DoBit(24); DoBit(23); DoBit(22);		
	DoBit(21); DoBit(20); DoBit(19); DoBit(18); DoBit(17);		
	DoBit(16); DoBit(15); DoBit(14); DoBit(13); DoBit(12);		
	DoBit(11); DoBit(10); DoBit(9);  DoBit(8);  DoBit(7);		
	DoBit(6);  DoBit(5);  DoBit(4);  DoBit(3);  DoBit(2);		
									
	/* Start reloading, if possible */				
	if (wordsleft > 1)
	{						
	    /* Need to reload m 2 before the end of the word, for WW */	
	    m = *--mp;							
	    DoBit(1);							
									
	    /* m_u and m_d need be reloaded 1 early, for NW and SW */	
	    if (mp_u  != (UInt32 *)NULL)  m_u  = *--mp_u;
	    if (mp_d  != (UInt32 *)NULL)  m_d  = *--mp_d;
	    DoBitZero;							
									
	    /* and finally m_uu and m_dd for the next word */		
	    if (mp_uu != (UInt32 *)NULL)  m_uu = *--mp_uu;
	    if (mp_dd != (UInt32 *)NULL)  m_dd = *--mp_dd;
	}								
	else
	{								
	    DoBit1_zeros;						
	    DoBit0_zeros;						
	}								
	*--ip = outbits;						
    }
}


/* PUBLIC */
/****************************************************************************
 * Dilate an image, preserving its topology
 *
 * Description:
 *  Take a source image and a destination image and dilate the source into
 *  the dest.  Dilate every source pixel by a centered circle of radius 1.0
 *  (in other words, all four neighbors).
 *  Be conservative in dilating all four neighbors, minding the local topology
 *  of the image at each pixel.  Do not turn on neighbors that will "connect"
 *  pieces of the image together.
 *
 * Returns:
 *  ia_successful	invocation done and okay
 *  ia_internal		unable to dilate properly, something is amiss
 *
 * Notes:
 *  This routine doesn't touch the frames.
 *  The source and dest image may not be the same.
 ****************************************************************************/
Int32 CDECL 
i_topologyPreserveDilate (
    IPState    *pIPState,
    UInt8      *dest,  /* Ptr to upper left corner of dest image's frame */
    UInt8      *src,   /* Ptr to upper left corner of src image's frame */
    UInt32	imWidth, /* Width of images in pixels including frame,
			* excluding padding. */
    UInt32	imHeight, /* Height of images in pixels including frame,
			* excluding padding. */
    UInt32	destImBPL, /* Number of bytes per line in dest image including
			* image, frame, and pad. */
    UInt32	srcImBPL) /* As above, for the source image */
{
    PROLOGUE;
    UInt32 *ip;					/* Output pointer */
    UInt32 *mp_uu, *mp_u, *mp, *mp_d, *mp_dd;	/* Input pointers */
    UInt32 wordsPerRow;
    const UInt32 irw = destImBPL / 4; /* The number of longwords in each row
				       * of the "image"/dest */
    const UInt32 mrw = srcImBPL / 4;  /* The number of longwords in each row
				       * of the "model"/src */
    UInt8 bitsinlastword;


    /* Make dimens exclude frame */
    imWidth -= 2*cFrameBits;
    imHeight -= 2*cFrameBits;
    
    /* Number of bits in the last longword of each model row (ie without
     * padding).  Note that the value "0" means all 32 bits are in longword.
     */
    bitsinlastword = (UInt8) imWidth & 31;

    /* The number of words in each image row that are processed */
    wordsPerRow = (imWidth + 31) / 32;

    /* Move past the frames */
    dest += (destImBPL << cLogFrameBits) + cFrameBytes;
    src  += (srcImBPL  << cLogFrameBits) + cFrameBytes;
    
    /* Handle the pathological widths */
    if (imWidth < 2)
    {
	UInt32 y;
	UInt32 bits = 0;	/* The current environment */

	if (imWidth == 0)
	{
	    /* Absolutely no bits are valid. Joy, joy. */
	    NORMAL_RETURN;
	}

	/* I don't care much about speed for this... */
	for (y = 0; y < imHeight; y++)
	{
	    bits = 0;
	    if (y >= 2)
	    {
		bits |= MoveAndMask(*(UInt32 *)(src + srcImBPL * (y - 2)), 0,
				    BIT_NN);
	    }
	    if (y >= 1)
	    {
		bits |= MoveAndMask(*(UInt32 *)(src + srcImBPL * (y - 1)), 0,
				    BIT_N);
	    }
	    bits |= MoveAndMask(*(UInt32 *)(src + srcImBPL * y), 0, BIT_C);
	    if (y <= imHeight - 2)
	    {
		bits |= MoveAndMask(*(UInt32 *)(src + srcImBPL * (y + 1)), 0,
				    BIT_S);
	    }
	    if (y <= imHeight - 3)
	    {
		bits |= MoveAndMask(*(UInt32 *)(src + srcImBPL * (y + 2)), 0,
				    BIT_SS);
	    }
	    *(UInt32 *)(dest + destImBPL * y) = dilateLookup[bits] << 31;
	}

	NORMAL_RETURN;
    }

    /* They'd better be long-aligned. */
    if ((destImBPL & 3) != 0 || (srcImBPL & 3) != 0)
	return ia_internal;

    if (imHeight < 4)
    {
	const UInt32 iskip = destImBPL / 4 - (imWidth + 31) / 32;
	/*
	 * Sigh - tiny src means it's all edge cases. Handle them.
	 */

	switch (imHeight)
	{
	case 3:
	{
	    mp_uu = (UInt32 *)(src) + wordsPerRow;
	    mp_u = mp_uu + mrw;
	    mp = mp_d = mp_dd = mp_u + mrw;
	    ip = (UInt32 *)(dest) + 3 * irw - iskip;
	    
	    /* Bottom row */
	    ip_doRow (pIPState, mp_uu, mp_u, mp, (UInt32 *)NULL, (UInt32 *)NULL,
		      ip, wordsPerRow, bitsinlastword);
	    mp_u -= mrw;
	    mp   -= mrw;
	    ip   -= irw;

	    /* Middle row; row below exists, but row 2 up doesn't. */
	    ip_doRow (pIPState, (UInt32 *)NULL, mp_u, mp, mp_d, (UInt32 *)NULL,
		      ip, wordsPerRow, bitsinlastword);
	    mp   -= mrw;
	    mp_d -= mrw;
	    ip   -= irw;

	    ip_doRow (pIPState, (UInt32 *)NULL, (UInt32 *)NULL, mp, mp_d, mp_dd,
		      ip, wordsPerRow, bitsinlastword);
	    
	    break;
	}

	case 2:
	{
	    mp_u = (UInt32 *)(src) + wordsPerRow;
	    mp = mp_d = mp_u + mrw;
	    ip = (UInt32 *)(dest) + 2 * irw - iskip;
	    
	    /* Bottom row */
	    ip_doRow (pIPState, (UInt32 *)NULL, mp_u, mp, (UInt32 *)NULL,
		      (UInt32 *)NULL, ip, wordsPerRow, bitsinlastword);

	    mp -= mrw;
	    ip -= irw;

	    /* Top row */
	    ip_doRow (pIPState, (UInt32 *)NULL, (UInt32 *)NULL, mp, mp_d,
		      (UInt32 *)NULL, ip, wordsPerRow, bitsinlastword);
	    
	    break;
	}

	case 1:
	{
	    mp = (UInt32 *)(src) + wordsPerRow;
	    ip = (UInt32 *)(dest) + irw - iskip;
	    ip_doRow (pIPState, (UInt32 *)NULL, (UInt32 *)NULL, mp,
		      (UInt32 *)NULL, (UInt32 *)NULL,
		      ip, wordsPerRow, bitsinlastword);
	    
	    /* Only row */
	    break;
	}

	}
    }
    else
    {
	/* General case for a large enough dest,
	 * the images widths are both >= 2 and heights are >= 4 */

	/*
	 * Start up the pointers at the very end; they'll come into
	 * play staggered.
	 */
	mp_uu = (UInt32 *)(src) + (imHeight - 2) * mrw -
	    (srcImBPL / 4 - (imWidth + 31) / 32);
	mp_u = mp_uu + mrw;
	mp = mp_d = mp_dd = mp_u + mrw;
	ip = (UInt32 *)(dest) + imHeight * irw -
	    (destImBPL / 4 - (imWidth + 31) / 32);

	/* Last row; rows below do not exist. */
	ip_doRow (pIPState, mp_uu, mp_u, mp, (UInt32 *)NULL, (UInt32 *)NULL,
		  ip, wordsPerRow, bitsinlastword);
	mp_uu -= mrw;
	mp_u  -= mrw;
	mp    -= mrw;
	ip    -= irw;

	/* Second last row; row below exists. */
	ip_doRow (pIPState, mp_uu, mp_u, mp, mp_d, (UInt32 *)NULL,
		  ip, wordsPerRow, bitsinlastword);
	mp_uu -= mrw;
	mp_u  -= mrw;
	mp    -= mrw;
	mp_d  -= mrw;
	ip    -= irw;

	/*
	 * Middle rows; all rows exist. We're done when the uu pointer
	 * is just about to run off the dest.
	 */
	while (mp_uu > (UInt32 *)(src))

	{
	    ip_doRow (pIPState, mp_uu, mp_u, mp, mp_d, mp_dd, ip, wordsPerRow,
		      bitsinlastword);
	    mp_uu -= mrw;
	    mp_u  -= mrw;
	    mp    -= mrw;
	    mp_d  -= mrw;
	    mp_dd -= mrw;
	    ip    -= irw;
	}

	/* Second from top row - row two above does not exist. */
	ip_doRow (pIPState, (UInt32 *)NULL, mp_u, mp, mp_d, mp_dd,
		  ip, wordsPerRow, bitsinlastword);
	mp    -= mrw;
	mp_d  -= mrw;
	mp_dd -= mrw;
	ip    -= irw;

	/* Top row - row above doesn't exist either */
	ip_doRow (pIPState, (UInt32 *)NULL, (UInt32 *)NULL, mp, mp_d, mp_dd,
		  ip, wordsPerRow, bitsinlastword);
    }

    NORMAL_RETURN;
}


/*

$Log: i_tpdil.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.9  1998/01/13 23:50:26  ddavies
i_tpdil.c -> /product/ipcore/ipshared/src/RCS/i_tpdil.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1995/12/13 02:46:41  ejaquith
i_tpdil.c -> /product/ipcore/ipshared/src/RCS/i_tpdil.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.7  1995/10/18  03:59:22  ddavies
 * i_tpdil.c -> /product/ipcore/ipshared/src/RCS/i_tpdil.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.6  1995/09/14  19:00:00  ejaquith
 * i_tpdil.c -> /product/ipcore/ipshared/src/RCS/i_tpdil.c,v
 *
 * CommonCheck: Converted all routines that were private in files to be of form
 * ip_*(), which they weren't.  No functionality/noticeable changes.
 *
 * Revision 1.5  1995/03/29  01:59:36  ejaquith
 * i_tpdil.c -> /product/ipcore/ipshared/src/RCS/i_tpdil.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.4  1995/03/15  01:09:17  ejaquith
 * i_tpdil.c -> /product/ipcore/ipshared/src/RCS/i_tpdil.c,v
 *
 * Updated indentation.
 *
 * Revision 1.3  1995/02/08  21:50:16  ejaquith
 * ejaquith on Wed Feb 8 13:48:50 PST 1995
 *
 * i_tpdil.c -> /product/ipcore/ipshared/src/RCS/i_tpdil.c,v
 *
 * Made ridiculously long Macro a function for processing rows.
 * This makes both its piggy object code size smaller and it slightly
 * faster.
 * This was giving MPW a hard time.
 * No change in functionality.
 *
 * Revision 1.2  1995/02/02  02:12:50  ejaquith
 * ejaquith on Wed Feb 1 18:11:31 PST 1995
 *
 * i_tpdil.c -> /product/ipcore/ipshared/src/RCS/i_tpdil.c,v
 *
 * Fixed silly bug which resulted from a poor choice for a macro name
 *
 * Revision 1.1  1995/02/01  21:51:45  ejaquith
 * Initial revision
 *

*/
