/***************************************************************
    Copyright (c) 1994, Xerox Corporation.  All rights reserved.
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
***************************************************************/

/*
  This file contains routines to implement a general warp
  capability. This includes, as a subset, the ability to 
  perform rotation and two-axis scaling.

  The public routines in this package are:
    i_resample:            Resample an image to perform
                           rotation and/or magnification

    i_resampleQuad():       Resample an by transforming an
                            arbitrary quadrilateral in the
                            source image into a rectangle
                            in the destination image.

    i_resampleUser():       Resample an image under the control
                            of a user-specified transform function.
                            This version divides the output into
                            small blocks, since there may be
                            enough nonlinearity to require it.


    
*/

/*
GENERAL PURPOSE FAST RESAMPLER

    Resample a block of an image, using a bilinear interpolation
    scheme with 1/4 pixel precision. The scan address generation
    process is bilinear, so this implementation will handle a
    transformation of an arbitrary quadilateral into a rectangle.
    This will accomodate rotation, two-axis scaling, differential
    scaling and keystoning. By partioning the output into small
    blocks, non-linear transformations can be performed in s
    picewise manner.

    No anti-alias filtering is performed by this implementation, so
    if the local x or y magnification gets much less than 0.7 or so,
    there may be some noticeable aliasing artifacts. Better to 
    filter the image down by a factor of two and then upsample to 
    get the smoothest result.

    All x,y coords and steps in this implementation are expressed in 
    16.16 form (16 bits of fractional part). The fractional address
    of a pixel is quantized to 1/4 pixel by using the 2 bits
    immediately to the right of the binary point as the fraction.
    Interpolation is done with simple shifts and adds. This is
    pretty easy with 1/4 pixel precision, but gets progressively
    harder with greater precision.
*/

/*
  Resampling consists of two parts - address generation, in which the 
  source pixel addresses coresponding to an output pixel are generated,
  and interpolation, the calculation of the gray level at a fractional
  pixel location.

BI-LINEAR ADDRESS GENERATION

  The address generation part of the resampler is very simple in
  this implementation. Since we're dealing only with transformations
  which map straight lines to straight lines, all of the address 
  scan paths are straight. This would not be the case with a higher 
  order warp. 
  (It is possible to do higher order warps by piecewise approximation).
  Straight paths let us easily calculate pixel addresses incrementally. 
  For example, the scan address formula for the x and y addresses 
  within a single line are:

          x = x0 + u * dx_u

          y = y0 + v * dy_u

  Where, u and v are coordinates in the output image space.
        x and y are coordinates in the input image space.
        x0 and y0 are the input coords to which u=0 and v=0 will map.
        dx_u and dy_u are the change in x and y for a unit change in u.

  Once we start generating a scan line, each address after x0 and y0
  is easily generated by just adding dx_u and dy_u to x and y. A 
  similar scheme is used to generate the new line start addresses as
  we proceed from one line to another. However, in the bilinear 
  adddress generator, the dx_u and dy_u will also change from line
  to line, giving rise to the ddx_v and ddy_v terms.

  Generating addresses by incremental adds requires that we have enough 
  precision in dx_u and dy_u to avoid roundoff and truncation errors.
  This implementation uses 16 fraction bits. Since were only looking 
  for 2 bits of fractional address (1/4 pixel), we can do 2^14 = 16384 
  incremental additions before running into precision problems. This
  should be more than enough for the size of images we're working with.

  NOTE: This type of address generation is called "backwards mapping".
        We map the desired output image rectangle back into the input 
        image coordinate space. We then walk along each OUTPUT pixel
        position, generating an input image address. This means that,
        with any rotation, we won't be accessing the input image
        along its rows and columns. This will be considerably slower 
        than the multiple shear type of algorithms, but it also 
        doesn't have their limitations, like sampling errors and 
        collapsing problems. Its also very easy to extend the 
        backwards mapping approach to higher-order warps. The same 
        can not be said of shears.
*/

/*
BILINEAR INTERPOLATION

  Bilinear interpolation calculates an interpolated pixel gray value
  for a (possibly) fractional address. Consider a 2x2 set of neighboring
  pixels (the *'s) around the interpolation point X, as shown below.
  The dots represent the fractional pixel addresses to a 1/4 pixel
  precison. In this case, the fractional address is 1/4, 3/4.

        (x,y)=0,0         1,0
               *  .  .  .  *   Top interp value = P00*3/4 + P10*1/4

               .  .  .  .  .

               .  .  .  .  .

               .  X  .  .  .   Final interp value = TOP*1/4 + BOT*3/4

               *  .  .  .  *   Bottom interp value = P01*3/4 + P11*1/4
              0,1         1,1


  The interpolation is done in a separable fashion. Using the x
  fraction, the interpolated value is found between the two top pixels.
  Then the bottom interpolated value is found, using the bottom two
  pixels. The final interpolated value is found by interpolating the
  top and bottom results, based on the y fractional address.

  In order to maintain some semblance of speed, the fractional precision
  is kept to 1/4 pixel. This allows the use of only shifts and adds, 
  since the only coefficients involved are 0, 1/4, 1/2, 3/4 and 1. For
  a fractional address of X, the left pixel gets multiplied by (1-X) 
  and the right pixel gets multiplied by X. Similarly for the top and
  bottom, interpreted same as left and right, respectively.

  The selection of 1/4 pixel was an attempt to get a balance between
  speed and smoothness. Use of 1/2 pixel precision would be faster, but
  not as smooth. Use of 1/8 pixel precision would get into coefficients
  like 6/8, requiring a 4x+2x type of operation.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_warp.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");


/*
  Define this so static procedures defined in shrwarp.prv
  only get included by this source file. This is necessary
  because shrwarp.prv contains the definition of the WarpInfo
  state structure, which must also be included with the
  ipwrap calling functions. Including all the static function
  prototypes generates warning messages from the compiler
  about unreferenced funcions.
*/
#define _I_WARP_C_

#include "shrwarp.prv"   /* This must go first so WarpInfo is defined */
#include "shrwarp.pub"


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Determine a 4 bit code for a point using the Cohen and Sutherland
  line clipping algorithm for clipping a line to a rectangle.
  A point inside the rectangle will have a code of zero
*/
static UInt8 CDECL 
ip_pointCode (
    ResampPoint *pt,
    Float64	xmin,
    Float64	xmax,
    Float64	ymin,
    Float64	ymax)
{
    return (UInt8) ( (pt->x < xmin) << 3 | (pt->x > xmax) << 2 
           | (pt->y < ymin) << 1 | (pt->y > ymax) );
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Assign 4 bit codes to each corner point of the rectangle.
  Test the codes using the logic of the Cohen and Sutherland
  line clipping algorithm for clipping a line to a rectangle.
  A point inside the rectangle will have a code of zero.
  If all four points are zero, the rect is entirely within the image.
  If the bitwise OR of all codes is 1111, the rect contains the image.
  If the bitwise AND of each side's endpoints is non-zero, then
  the rect is entirely outside of the image. Otherwise, its a 
  boundary block.
*/

static Int32 CDECL 
ip_testRect (
    IPState    *pIPState,
    ResampRect *rect,
    UInt32	width,
    UInt32	height)
{

    PROLOGUE;
    ResampCode code;
    Float64 xmin, xmax, ymin, ymax;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* 
	  Make the right/bottom image boundary one pixel inside to account 
	  for the need to access an additional pixel for 2x2 interpolation
	*/
    xmin = 0;
    xmax = width - 2;
    ymin = 0;
    ymax = height - 2;

	/* Assign a code to each of the four rectangle corner points */
    code.ptCode[0] = ip_pointCode(&rect->ulc, xmin, xmax, ymin, ymax);
    code.ptCode[1] = ip_pointCode(&rect->urc, xmin, xmax, ymin, ymax);
    code.ptCode[2] = ip_pointCode(&rect->llc, xmin, xmax, ymin, ymax);
    code.ptCode[3] = ip_pointCode(&rect->lrc, xmin, xmax, ymin, ymax);

	/* If all codes are zero, the rect is entirely within the image */
    if (code.rectCode == 0)
        return RESAMP_INSIDE;

	/* 
	  If the bitwise OR of all the codes is binary 1111, the rect CONTAINS 
	  the image. This can happen with magnifications less than one.
	*/ 
    if ( (code.ptCode[0] | code.ptCode[1] | 
          code.ptCode[2] | code.ptCode[3]) == 0x0f)
        return RESAMP_CONTAIN;
    
	/*
	  If each of the bitwise ANDs of each side's endpoints is non-zero,
	  then the rectangle is completely outside the image.
	*/
    if ( code.ptCode[0] & code.ptCode[1] &&
         code.ptCode[0] & code.ptCode[2] &&
         code.ptCode[1] & code.ptCode[3] &&
         code.ptCode[2] & code.ptCode[3])
        return RESAMP_OUTSIDE;

	/*
	  Only one case left. The block must be a border block with at
	  least some amount overlapping the image area.
	*/

    return RESAMP_BORDER;

}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Calculate the input image address corresponding to the
  four corner points of the output rectangle. This routine uses
  the function pointed to by the global function pointer
  resampTransform.
*/

static void CDECL 
ip_calcCornerCoords (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo,
    Int32	U0,
    Int32	V0,
    UInt32	outWidth,
    UInt32	outHeight,
    ResampRect *inRect)
{
    PROLOGUE;

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    warpInfo->gResampTransform(pIPState, warpInfo, U0, V0, 
			       &(inRect->ulc.x), &(inRect->ulc.y) ); 
    warpInfo->gResampTransform(pIPState, warpInfo, U0+outWidth-1, V0, 
			       &(inRect->urc.x), &(inRect->urc.y) ); 
    warpInfo->gResampTransform(pIPState, warpInfo, U0, V0+outHeight-1, 
			       &(inRect->llc.x), &(inRect->llc.y) ); 
    warpInfo->gResampTransform(pIPState, warpInfo, U0+outWidth-1,
			       V0+outHeight-1, 
			       &(inRect->lrc.x), &(inRect->lrc.y) ); 
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Function to determine source (x,y) addresses given output (u,v)
  addresses. Uses global parameters of a rotation/scale transform.
  NOTE: this procedure must be of type ResampleTransform!!
*/

static void CDECL 
ip_rotateMagTransform (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo,
    Int32	u,
    Int32	v,
    Float64    *x,
    Float64    *y)
{
	/* Keep compiler happy */
    pIPState = pIPState;

    *x =  (u-warpInfo->gUc)*warpInfo->gMx*warpInfo->gCos 
         + (v-warpInfo->gVc)*warpInfo->gMy*warpInfo->gSin 
         + warpInfo->gXc;

    *y =  -(u-warpInfo->gUc)*warpInfo->gMx*warpInfo->gSin 
         + (v-warpInfo->gVc)*warpInfo->gMy*warpInfo->gCos 
         + warpInfo->gYc;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Function to determine source (x,y) addresses given output (u,v)
  addresses. Uses global parameters calculated from the four
  corner points in the source image.
  NOTE: this procedure must be of type ResampleTransform!!
*/

static void CDECL 
ip_quadTransform (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo,
    Int32	u,
    Int32	v,
    Float64    *x,
    Float64    *y)
{
	/* Keep compiler happy */
    pIPState = pIPState;

    *x =  warpInfo->gX0 
         + (u-warpInfo->gU0)*warpInfo->gDx_u 
         + (v-warpInfo->gV0)*warpInfo->gDx_v 
         + (u-warpInfo->gU0)*(v-warpInfo->gV0)*warpInfo->gDdx_v;

    *y =  warpInfo->gY0 
         + (u-warpInfo->gU0)*warpInfo->gDy_u 
         + (v-warpInfo->gV0)*warpInfo->gDy_v 
         + (u-warpInfo->gU0)*(v-warpInfo->gV0)*warpInfo->gDdy_v;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Create a table to map the values of four binary pixels to a 
  resulting binary or gray value, given a fractional interpolation
  address between the group of four. The basic lookup table structure
  is essentially a 3-D lookup table as follows:

      UInt8 lut[yFrac(0:3)][xFrac(0:3)][pixelValues(0:15)]

  The pixelValues field is constructed as a 4 bit value from
  the four binary pixel values, as in
      ULC URC LLC LRC
      MSB         LSB
*/

static ErrCode CDECL 
ip_createInterpLut (
    IPState    *pIPState,
    UInt32	inDepth, /* Bpp in the input data */
    UInt32	outDepth, /* Number of bits in output table */
    UInt8      **ppTable)
{
    PROLOGUE;
    UInt8 *buf, *ptr;
    UInt32 xfrac, yfrac;
    UInt32 top, bottom, gray;
    Int32 ulc, urc, llc, lrc;

/* 
  Allocate memory for the table
  Caller is responsible for freeing it
*/
    buf = (UInt8 *) MALLOC(4 * 4 * (1<<(4*inDepth)) * sizeof(UInt8));
    if (buf == NULL)
    {
	return ia_nomem;
    }

    ptr = buf;
    for (yfrac=0; yfrac<4; yfrac++)
      for (xfrac=0; xfrac<4; xfrac++)
      {
        if (inDepth == 1)
        {
        /*
          For binary images, a zero value is interpreted as white.
          This is just the opposite of grayscale images. So when we
          calculate the interpolation table, give the zero value
          an 8 bpp equivelent value of 255 (white)
        */
            for (ulc=255; ulc>=0; ulc-=255)
              for (urc=255; urc>=0; urc-=255)
                for (llc=255; llc>=0; llc-=255)
                  for (lrc=255; lrc>=0; lrc-=255)
                  {
                    /* Bi-linear Interpolation */
                    top = ulc*(4-xfrac) + urc*xfrac;
                    bottom = llc*(4-xfrac) + lrc*xfrac;

                    /* "gray" contains 8 bit interpolated gray value */
                    gray = top*(4-yfrac) + bottom*yfrac;
                    gray = (gray + 8) / 16;

                    /* Now convert to desired output depth */
                    if (outDepth == 1)
                        gray = 1 - (gray + 128)/256;
                    else if (outDepth == 2)
                        gray = gray >> 6;
                    else if (outDepth == 4)
                        gray = gray >> 4;

                    *ptr++ = (UInt8)gray;
                  }
        }

        else /* inDepth = 2 */
        {

            for (ulc=0; ulc<=3; ulc++)
              for (urc=0; urc<=3; urc++)
                for (llc=0; llc<=3; llc++)
                  for (lrc=0; lrc<=3; lrc++)
                  {
                    if (outDepth == 8)
                    {
                        /* Use Bi-linear Interpolation */
                        top = ulc*(4-xfrac) + urc*xfrac;
                        bottom = llc*(4-xfrac) + lrc*xfrac;

                        /* "gray" contains interpolated gray value in 8 bpp */
                        gray = top*(4-yfrac) + bottom*yfrac;

                        /* Now round to output depth */
                        gray = (gray*255 + 24) / 48;
                    }

                    else
                    {
                        /* 
                          Pick the closest value. For dithered images,
                          with limited output depth, this seems to work 
                          much better than bilinear interpolation. 
                        */
                        if (yfrac<2)
                        {
                            if (xfrac<2)
                                gray = ulc;
                            else
                                gray = urc;
                        }
                        else
                        {
                            if (xfrac<2)
                                gray = llc;
                            else
                                gray = lrc;
                        }

                        /* Now round to output depth */
                        if (outDepth == 1)
                            gray = gray/2;
                        else if (outDepth == 2)
                            gray = gray;
                        else /* if (outDepth == 4) */
                            gray = gray*4 + gray;
                    }

                    *ptr++ = (UInt8)gray;
                  }
        }
      }

	/* It worked, so return the table */
    *ppTable = buf;
    NORMAL_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Calculate resampling coefficients for a bilinear address
  generator from the given rectangle, using forward differences
*/
 
static void CDECL 
ip_calcResampCoeffs (
    IPState    *pIPState,
    ResampRect *rect,	   /* Rectangle of four corner pts */
    UInt32	nu,    /* X size of output image */
    UInt32	nv,    /* Y size of output image */
    Int32      *x0,    /* X coord of 1st sample in input */
    Int32      *y0,    /* Y coord of 1st sample in input */
    Int32      *dx_u,  /* X step size along rows */
    Int32      *dy_u,  /* Y step size along rows */
    Int32      *dx_v,  /* X step size, row to row */
    Int32      *dy_v,  /* Y step size, row to row */
    Int32      *ddx_v, /* Change in X step size, row to row */
    Int32      *ddy_v) /* Change in Y step size, row to row */
{
    PROLOGUE;
    Float64 du, dv;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Calculate the output image dimensions */
    du = nu - 1;
    dv = nv - 1;

	/* Calculate the start address in input space for u=v=0 */
    *x0 = RESAMP_SCALE_16_16(rect->ulc.x);
    *y0 = RESAMP_SCALE_16_16(rect->ulc.y);

	/* Calculate the input image scan step sizes for deltaU and deltaV */
    *dx_u = RESAMP_SCALE_4_28((rect->urc.x - rect->ulc.x)/du);
    *dy_u = RESAMP_SCALE_4_28((rect->urc.y - rect->ulc.y)/du);

    *dx_v = RESAMP_SCALE_16_16((rect->llc.x - rect->ulc.x)/dv);
    *dy_v = RESAMP_SCALE_16_16((rect->llc.y - rect->ulc.y)/dv);

    *ddx_v = RESAMP_SCALE_4_28(((rect->lrc.x - rect->llc.x) 
                        - (rect->urc.x - rect->ulc.x))/du/dv);
    *ddy_v = RESAMP_SCALE_4_28(((rect->lrc.y - rect->llc.y) 
                        - (rect->urc.y - rect->ulc.y))/du/dv);
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  This is the routine that does all the work. It generates
  a single output scan line by following the recipe given
  by the address generation parameters (x0, y0, dx_u, dy_u).
  The ddx_v and ddy_v parameters are not needed within a line,
  since they are only used between lines to update dx_u and dy_u.

  A flag, checkBounds, is passed in to control whether tests
  are made for a pixel being outside the image bounds. This results
  in about a 5% speed penalty, but is preferable to
  maintaining two versions of the code, especially as additions
  are made to deal with other image depths.
*/

static ErrCode CDECL 
ip_resampleLine8Bpp (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo,     /* Included for API consistency */
    UInt8      *inLine,	  /* Input image buffer starting address */
    UInt32     *outLine,   /* Output image buffer starting address */
    Int32      *inBpl,	  /* Array of line stride values */
    Int32	x0,	 /* X coord of 1st sample in input */
    Int32	y0,	 /* Y coord of 1st sample in input */
    Int32	dx_u,	 /* X step size along rows */
    Int32	dy_u,	 /* Y step size along rows */
    Int32	nu,	   /* # of output samples per row */
    Int32	outDepth, /* Depth of output image */
    Int32	inWidth,  /* Width of input image */
    Int32	inHeight,  /* Height of input image */
    Bool	checkBounds) /* True if image bounds check reqd */
{

    PROLOGUE;
    UInt8 *topPtr, *botPtr, *lineBase, *lut;
    Int32 x, y, X, Y, oldY, ydiff;
    Int32 testX, testY;
    UInt32 top, bottom;
    UInt32 ul, ur, ll, lr;
    UInt32 *outPtr;
    UInt32 stride, extraBits, depthShift;
    UInt32 outReg, fillValue, pixPerWord, count, temp;
    UInt32 allOnes = 0xffffffff;

/* This isn't used, but its needed to keep the compiler happy */
    lut = warpInfo->gInterpLut;

/* Set starting point in input space for this output line */
    x = x0;
    y = y0;
    X = RESAMP_INT_PART(x);
    Y = RESAMP_INT_PART(y);
    stride = inBpl[1];
    lineBase = inLine + Y*stride;

/* 
  If bounds checking has been requested, it only means that
  that some part of the block being resampled is outside of
  the image bounds. We still want to test each individual 
  line in the block. If both endpoints are inside the image,
  we won't need to do any further bounds checks. If either
  endpoint is outside the image, then we'll want to turn on
  checking at the pixel level, using the checkBounds flag.

  Note for the future: I considered trying to find the valid
  portion of the line using a clipping test. However, all the
  clipping algorithms I could find either used floating point
  calculations, including divides and multiplies, or used a
  recursive midpoint test. Either way looked like it would use
  more cycles than whats being done now. Since, with a small block
  size like 64x64, most of the blocks will be either completely
  inside or completely outside the image, spending a lot of effort
  on infrequent border cases is probably a waste. However, if anyone 
  comes up with a brilliantly simple clipping method, go for it.
*/ 

    if (checkBounds)  /* This BLOCK requires bounds checks if true */
    {
        checkBounds = cFalse;  /* Reset the flag first */

        /* Test the line starting point */
        if (X<0 || Y<0 || X>(inWidth-2) || Y>(inHeight-2))
            checkBounds = cTrue;
        else /* Calculate and test the line endpoints */
        {
            testX = RESAMP_INT_PART(x + (nu-1)*dx_u);
            if (testX<0 || testX > (inWidth-2))
                checkBounds = cTrue;
            else
            {
                testY = RESAMP_INT_PART(y + (nu-1)*dy_u);
                if (testY<0 || testY > (inHeight-2))
                    checkBounds = cTrue;
            }
        }

    /* CheckBounds now indicates if this LINE requires bounds checks */

    }

/* Generate the proper fill value for the output depth */
    fillValue = allOnes >> (32-outDepth);

    pixPerWord = 32 / outDepth;
    count = pixPerWord;
    extraBits = (nu*outDepth) % 32; /* Bits remaining at line end */
    outReg = 0;
    outPtr = outLine;
    depthShift = 8 - outDepth;

    /* Resample to produce one output row */
    while ( nu-- )
    {

/* 
  If bounds checking is enabled, test if resample point is within image.
  Old-fashioned goto escape seems appropriate here, primarily for
  speed reasons, since I want any of the ORs to cause the IF 
  statement code to be executed immediately without evaluating all 
  of the conditional expressions. Sorry Mr. Dijkstra.
*/

        if (checkBounds)
        {
            if ((X<0) || (Y<0) || (X>inWidth-2) || (Y>inHeight-2) )
            {
                outReg <<= outDepth;
                outReg |= fillValue;
                goto LOOP;
            }
        }

        /* Set ptrs to top and bottom of 2x2 pixel group */
        topPtr = lineBase + X;
        botPtr = topPtr + stride;

        /* Get the four pixel values */
        ul = UCHAR_ACCESS(topPtr);
        ur = UCHAR_ACCESS(topPtr+1);
        ll = UCHAR_ACCESS(botPtr);
        lr = UCHAR_ACCESS(botPtr+1);

        /*
          Do the top and bottom horizontal interpolations
          using a separable approach. Scale each interp
          coefficient by 4, ie. the number of quarters (1/4 th's).
        */
        switch(RESAMP_FRACTION_PART(x))
        {
        case 0: /* left * 4/4 */
#ifdef _386_ARCH_  /* Shifts are slow on the PC, use adds instead */
            top = ul + ul;
            top += top;
            bottom = ll + ll;
            bottom += bottom;
#else
            top = ul<<2;
            bottom = ll<<2;
#endif

            break;

        case 1: /* left * 3/4 + right * 1/4 */
            top = ul + ul + ul + ur; 
            bottom = ll + ll + ll + lr; 
            break;

        case 2: /* left * 2/4 + right * 2/4 */
            top = ul + ur;
            top += top;
            bottom = ll + lr;
            bottom += bottom;
            break;

        case 3: /* left * 1/4 + right * 3/4 */
            top = ul + ur + ur  + ur; 
            bottom = ll + lr + lr + lr; 
            break;

        default: /* This can't happen, he said, optimistically */
            top = bottom = 0;
            break;
        }

        /*
          Now do a vertical interpolation on the top and
          bottom results of the horizontal interpolations.
          Round and descale the result. Since each of the 
          horizontal and vertical ops are scaled by 4, the
          total scaling is 16X. Adding 8 is thus equiv to
          adding 0.5 (8/16). Treat 4/4 and 2/4 cases as special
          cases to save a couple of shifts.
        */
        switch(RESAMP_FRACTION_PART(y))
        {
        case 0: /* top * 1 */
            outReg <<= outDepth;
            outReg |= (top + 2) >> (2+depthShift);
            break;

        case 1: /* top * 3/4 + bottom * 1/4 */
            outReg <<= outDepth;
            outReg |= (top + top + top + bottom + 8) >> (4+depthShift); 
            break;

        case 2: /* top * 1/2 + bottom * 1/2 */
            outReg <<= outDepth;
            outReg |= ( (top + bottom)  + 4) >> (3+depthShift);
            break;

        case 3: /* top * 1/4 + bottom * 3/4 */
            outReg <<= outDepth;
            outReg |= (top + bottom + bottom + bottom + 8) >> (4+depthShift);
            break;

        default: /* This can't happen, he said, optimistically */
            outReg <<= outDepth;
            break;
        }

LOOP:

        /* If enough bits have accumulated write a word */
        count--;
        if (count == 0)
        {
            if (outDepth == 1)
                outReg = ~outReg;
            *outPtr++ = outReg;
            outReg = 0;           /* Clear the accumulation reg */
            count = pixPerWord;   /* Reset the pixel counter */
        }

        /* Save the current value for comparison */
        oldY = Y;

        /* Step along the row scan path to compute the new values */
        x += dx_u;
        y += dy_u;
        X = RESAMP_INT_PART(x);
        Y = RESAMP_INT_PART(y);
        ydiff = Y - oldY;

        /* 
          Use the array of stride values to get to the new
          line offset. Only need to do this if Y has changed.
        */
        if (ydiff != 0)
            lineBase += inBpl[ydiff];

    } /* End main loop */

    /* Write last word if output bit count was not a multiple of 32 */
    if (extraBits != 0)
    {
        if (outDepth == 1)
            outReg = ~outReg;
        temp = *outPtr;
        *outPtr = (temp & (allOnes>>extraBits)) | (outReg << (32 - extraBits));
    }

    NORMAL_RETURN;
}

/*********************************************************************/

static ErrCode CDECL 
ip_resampleLine1Bpp (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo,
    UInt8      *inLine,	  /* Input image buffer starting address */
    UInt32     *outLine,   /* Output image buffer starting address */
    Int32      *inBpl,	  /* Array of line stride values */
    Int32	x0,	 /* X coord of 1st sample in input */
    Int32	y0,	 /* Y coord of 1st sample in input */
    Int32	dx_u,	 /* X step size along rows */
    Int32	dy_u,	 /* Y step size along rows */
    Int32	nu,	   /* # of output samples per row */
    Int32	outDepth, /* Depth of output image */
    Int32	inWidth,  /* Width of input image */
    Int32	inHeight,  /* Height of input image */
    Bool	checkBounds) /* True if image bounds check reqd */
{

    PROLOGUE;
    UInt8 *topPtr, *botPtr, *lineBase;
    Int32 x, y, X, Y, oldY, ydiff;
    Int32 testX, testY;
    UInt32 *outPtr;
    UInt32 index, bitPos, stride, fillValue;
    UInt32 topByte, botByte, outReg, temp;
    UInt32 pixPerWord, count, extraBits;
    UInt32 allOnes = 0xffffffff;
    UInt8 *lut;

/* Set starting point in input space  for this output line */
/* Note that inBpl[1] contains the BytesPerLine value */
    x = x0;
    y = y0;
    X = RESAMP_INT_PART(x);
    Y = RESAMP_INT_PART(y);
    stride = inBpl[1];
    lineBase = inLine + Y*stride;

    if (checkBounds)  /* This BLOCK requires bounds checks if true */
    {
        checkBounds = cFalse;  /* Reset the flag first */

        /* Test the line starting point */
        if (X<0 || Y<0 || X>(inWidth-2) || Y>(inHeight-2))
            checkBounds = cTrue;
        else /* Calculate and test the line endpoints */
        {
            testX = RESAMP_INT_PART(x + (nu-1)*dx_u);
            if (testX<0 || testX > (inWidth-2))
                checkBounds = cTrue;
            else
            {
                testY = RESAMP_INT_PART(y + (nu-1)*dy_u);
                if (testY<0 || testY > (inHeight-2))
                    checkBounds = cTrue;
            }
        }

    /* CheckBounds now indicates if this LINE requires bounds checks */

    }

/* Create the fill value for the desired output depth */
    if (outDepth == 1)
        fillValue = 0;
    else
        fillValue = allOnes >> (32-outDepth);

    pixPerWord = 32 / outDepth;
    count = pixPerWord;
    extraBits = (nu*outDepth)%32;
    outReg = 0;
    outPtr = outLine;
    lut = warpInfo->gInterpLut;

    /* Resample to produce one output row */
    while ( nu-- )
    {

/* Test if resample point is within image */
/* Supply background fill if it isn't */
        if (checkBounds)
        {
            if ((X<0) || (Y<0) || (X>inWidth-2) || (Y>inHeight-2) )
            {
                outReg <<= outDepth;
                outReg |= fillValue;
                goto LOOP;
            }
        }

        /* Set ptrs to top and bottom of 2x2 pixel group */
        topPtr = lineBase + (X>>3);   /* Find the byte addr */
        topByte = UCHAR_ACCESS(topPtr);

        botPtr = topPtr + stride;  /* Move down one stride */
        botByte = UCHAR_ACCESS(botPtr);
        bitPos = X & 7;  /* !!! This counts bits left to right */

        /* 
          If we're not at the LSB, grab two consequtive bits.
          This will happen 7 out of 8 times, so its worth
          special casing. If were at the LSB (bit 7 in this
          left to right counting scheme) grab the LSB and then
          the MSB of the following byte. 

          What's being done here is creating a lookup table index
          from the four binary pixel values, ordered MSB to LSB 
          as (ulc, urc, llc, lrc).
        */

        if (bitPos != 7) 
        {
          index = (topByte >> (6-bitPos)) & 3;
          index = (index<<2) + ((botByte>>(6-bitPos)) & 3);
        }
        else
        {
          index = ((topByte&1)<<1) + ((UCHAR_ACCESS(topPtr+1))>>7);
          index = (index<<2) + ((botByte&1)<<1) +
                  ((UCHAR_ACCESS(botPtr+1))>>7);
        }

        outReg = (outReg<<outDepth) + 
            lut[(RESAMP_FRACTION_PART(y)<<6)
                        + (RESAMP_FRACTION_PART(x)<<4)
                        + index];

LOOP:
/* If enough output values have accumulated, write a word */
        count--;
        if (count == 0)
        {
            *outPtr++ = outReg;
            outReg = 0;          /* Clear the output register again */
            count = pixPerWord;  /* Reset the pixel counter */
        }

        /* Save the current value for comparison */
        oldY = Y;

        /* Step along the row scan path */
        x += dx_u;
        y += dy_u;
        X = RESAMP_INT_PART(x);
        Y = RESAMP_INT_PART(y);
        ydiff = Y - oldY;

        /*
          Use the array of stride values to get to the new
          line offset. Only need to do this if Y has changed.
        */
        if (ydiff != 0)
            lineBase += inBpl[ydiff];

    } /* End main loop */

    /* Write last word if output width was not a multiple of 32 */
    if (extraBits != 0)
    {
        temp = *outPtr;
        *outPtr = (temp & (allOnes>>extraBits)) | (outReg << (32 - extraBits));
    }

    NORMAL_RETURN;
}

/*********************************************************************/

static ErrCode CDECL 
ip_resampleLine2Bpp (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo,
    UInt8      *inLine,	  /* Input image buffer starting address */
    UInt32     *outLine,   /* Output image buffer starting address */
    Int32      *inBpl,	  /* Array of line stride values */
    Int32	x0,	 /* X coord of 1st sample in input */
    Int32	y0,	 /* Y coord of 1st sample in input */
    Int32	dx_u,	 /* X step size along rows */
    Int32	dy_u,	 /* Y step size along rows */
    Int32	nu,	   /* # of output samples per row */
    Int32	outDepth, /* Depth of output image */
    Int32	inWidth,  /* Width of input image */
    Int32	inHeight,  /* Height of input image */
    Bool	checkBounds) /* True if image bounds check reqd */
{

    PROLOGUE;
    Int32 x, y, X, Y, oldY, ydiff, value;
    Int32 testX, testY;
    UInt8 *topPtr, *botPtr, *lineBase;
    UInt32 *outPtr;
    UInt32 index, bitPos, stride, fillValue, extraBits, xfrac, yfrac;
    UInt32 topByte, botByte, outReg, count, pixPerWord, temp;
    UInt32 allOnes = 0xffffffff;
    UInt8 *lut;

/* Set starting point in input space  for this output line */
/* Note that inBpl[1] contains the BytesPerLine value */
    x = x0;
    y = y0;
    X = RESAMP_INT_PART(x);
    Y = RESAMP_INT_PART(y);
    stride = inBpl[1];
    lineBase = inLine + Y*stride;

/* 
  If bounds checking has been requested, it only means that
  that some part of the block being resampled is outside of
  the image bounds. We still want to test each individual 
  line in the block. If both endpoints are inside the image,
  we won't need to do any further bounds checks. If either
  endpoint is outside the image, then we'll want to turn on
  checking at the pixel level, using the checkBounds flag.
*/ 

    if (checkBounds)  /* This BLOCK requires bounds checks if true */
    {
        checkBounds = cFalse;  /* Reset the flag first */

        /* Test the line starting point */
        if (X<0 || Y<0 || X>(inWidth-2) || Y>(inHeight-2))
            checkBounds = cTrue;
        else /* Calculate and test the line endpoints */
        {
            testX = RESAMP_INT_PART(x + (nu-1)*dx_u);
            if (testX<0 || testX > (inWidth-2))
                checkBounds = cTrue;
            else
            {
                testY = RESAMP_INT_PART(y + (nu-1)*dy_u);
                if (testY<0 || testY > (inHeight-2))
                    checkBounds = cTrue;
            }
        }

    /* CheckBounds now indicates if this LINE requires bounds checks */

    }

/* Create the fill value for the desired output depth */
    fillValue = allOnes >> (32-outDepth);

    outPtr = outLine;
    outReg = 0;
    pixPerWord = 32 / outDepth;
    count = pixPerWord;
    extraBits = (nu*outDepth) % 32;
    lut = warpInfo->gInterpLut;

    /* Resample to produce one output row */
    while ( nu-- )
    {

/* Test if resample point is within image */
/* Supply background fill if it isn't */
        if (checkBounds)
        {
            if ((X<0) || (Y<0) || (X>inWidth-2) || (Y>inHeight-2) )
            {
                outReg <<= outDepth;
                outReg |= fillValue;
                goto LOOP;
            }
        }

        /* Set ptrs to top and bottom of 2x2 pixel group */
        topPtr = lineBase + X/4;   /* Find the byte addr */
        topByte = UCHAR_ACCESS(topPtr);

        botPtr = topPtr + stride;  /* Move down one stride */
        botByte = UCHAR_ACCESS(botPtr);
        bitPos = (X & 3);  /* !!! This counts pixels left to right */
        bitPos += bitPos; /* Double it - this is for 2 bpp */

        /* 
          If we're not at bit 6 (counting left to right, or MSB to LSB), 
          grab two consequtive pixels (4 bits).
          This will happen 3 out of 4 times, so its worth
          special casing. If were at the LSB end (bit 6 in this
          left to right counting scheme) grab two bits and then
          the two MSBs of the following byte. 

          What's being done here is creating a lookup table index
          from the four binary pixel values, ordered MSB to LSB 
          as (ulc, urc, llc, lrc).
        */

        if (bitPos != 6) 
        {
          index = (topByte >> (4-bitPos)) & 15;
          index = (index<<4) | ((botByte>>(4-bitPos)) & 15);
        }
        else
        {
          index = ((topByte&3)<<2) | ((UCHAR_ACCESS(topPtr+1))>>6);
          index = (index<<4) | ((botByte&3)<<2) |
                  ((UCHAR_ACCESS(botPtr+1))>>6);
        }

/* 
  Now add the interpolation fraction to the lookup table address,
  as in lut[yfrac][xfrac][ulc][urc][llc][lrc]
     or lut[yfrac][xfrac][index]
*/
        yfrac = RESAMP_FRACTION_PART(y);
        xfrac = RESAMP_FRACTION_PART(x);

        outReg <<= outDepth;

        index = (yfrac<<10) | (xfrac<<8) | index;
        value = lut[index];
        outReg |= (UInt32) value; 

           

LOOP:

/* If enough output values have accumulated, write a word */
        count--;
        if (count == 0)
        {
            if (outDepth == 1)   /* Invert binary images */
                outReg = ~outReg;
            *outPtr++ = outReg;
            outReg = 0;          /* Clear the output register again */
            count = pixPerWord;  /* Reset the pixel counter */
        }

        /* Save the current value for comparison */
        oldY = Y;

        /* Step along the row scan path */
        x += dx_u;
        y += dy_u;
        X = RESAMP_INT_PART(x);
        Y = RESAMP_INT_PART(y);
        ydiff = Y - oldY;

        /*
          Use the array of stride values to get to the new
          line offset. Only need to do this if Y has changed.
        */
        if (ydiff != 0)
            lineBase += inBpl[ydiff];

    }

/* Write last word if output width was not a multiple of 32 */
    if (extraBits != 0)
    {
        if (outDepth == 1)
            outReg = ~outReg;
        temp = *outPtr;
        *outPtr = (temp & (allOnes>>extraBits)) | (outReg << (32 - extraBits));
    }

    NORMAL_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

static ErrCode CDECL 
ip_resampleLine4Bpp (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo,     /* Included for API consistency */
    UInt8      *inLine,	  /* Input image buffer starting address */
    UInt32     *outLine,   /* Output image buffer starting address */
    Int32      *inBpl,	  /* Array of line stride values */
    Int32	x0,	 /* X coord of 1st sample in input */
    Int32	y0,	 /* Y coord of 1st sample in input */
    Int32	dx_u,	 /* X step size along rows */
    Int32	dy_u,	 /* Y step size along rows */
    Int32	nu,	   /* # of output samples per row */
    Int32	outDepth, /* Depth of output image */
    Int32	inWidth,  /* Width of input image */
    Int32	inHeight,  /* Height of input image */
    Bool	checkBounds) /* True if image bounds check reqd */
{

    PROLOGUE;
    UInt8 *topPtr, *botPtr, *lineBase, *lut;
    Int32 x, y, X, Y, oldY, ydiff;
    Int32 testX, testY;
    UInt32 *outPtr;
    UInt32 top, bottom, ul, ur, ll, lr;
    UInt32 bitPos, stride, fillValue, extraBits, depthShift;
    UInt32 topByte, botByte, outReg, count, pixPerWord, temp;
    UInt32 allOnes = 0xffffffff;

/* This isn't used, but its needed to keep the compiler happy */
    lut = warpInfo->gInterpLut;

/* Set starting point in input space  for this output line */
/* Note that inBpl[1] contains the BytesPerLine value */
    x = x0;
    y = y0;
    X = RESAMP_INT_PART(x);
    Y = RESAMP_INT_PART(y);
    stride = inBpl[1];
    lineBase = inLine + Y*stride;

/* 
  If bounds checking has been requested, it only means that
  that some part of the block being resampled is outside of
  the image bounds. We still want to test each individual 
  line in the block. If both endpoints are inside the image,
  we won't need to do any further bounds checks. If either
  endpoint is outside the image, then we'll want to turn on
  checking at the pixel level, using the checkBounds flag.
*/ 

    if (checkBounds)  /* This BLOCK requires bounds checks if true */
    {
        checkBounds = cFalse;  /* Reset the flag first */

        /* Test the line starting point */
        if (X<0 || Y<0 || X>(inWidth-2) || Y>(inHeight-2))
            checkBounds = cTrue;
        else /* Calculate and test the line endpoints */
        {
            testX = RESAMP_INT_PART(x + (nu-1)*dx_u);
            if (testX<0 || testX > (inWidth-2))
                checkBounds = cTrue;
            else
            {
                testY = RESAMP_INT_PART(y + (nu-1)*dy_u);
                if (testY<0 || testY > (inHeight-2))
                    checkBounds = cTrue;
            }
        }

    /* CheckBounds now indicates if this LINE requires bounds checks */

    }

/* Create the fill value for the desired output depth */
    fillValue = allOnes >> (32-outDepth);

    outPtr = outLine;
    outReg = 0;
    pixPerWord = 32 / outDepth;
    count = pixPerWord;
    extraBits = (nu*outDepth) % 32;
    depthShift = 8 - outDepth;

    /* Resample to produce one output row */
    while ( nu-- )
    {

/* Test if resample point is within image */
/* Supply background fill if it isn't */
        if (checkBounds)
        {
            if ((X<0) || (Y<0) || (X>inWidth-2) || (Y>inHeight-2) )
            {
                outReg <<= outDepth;
                outReg |= fillValue;
                goto LOOP;
            }
        }

        /* Set ptrs to top and bottom of 2x2 pixel group */
        topPtr = lineBase + (X>>1);   /* Find the byte addr */
        topByte = UCHAR_ACCESS(topPtr);

        botPtr = topPtr + stride;  /* Move down one stride */
        botByte = UCHAR_ACCESS(botPtr);
        bitPos = (X & 1);  /* !!! This counts nibbles left to right */

        /* 
          If we're not at bit 4 (counting left to right, or MSB to LSB), 
          grab two consequtive pixels (8 bits).
          This will happen half the time. If were at the LSB end 
          (bit 4 in this left to right counting scheme) grab the
          least significant nibble from this byte and the most
          significant nibble from the following byte.

          Adjust the gray values so that they are on an 8 bit scale.
          This will allow a uniform treatment when we get to the
          output depth adjustment.
        */

        if (bitPos == 0) 
        {
          ul = topByte & 0xfffffff0;
          ur = (topByte & 15) << 4;;
          ll = botByte & 0xfffffff0;
          lr = (botByte & 15) << 4;
        }
        else
        {
          ul = (topByte & 15) << 4;
          ur = (UInt32)(UCHAR_ACCESS(topPtr+1)) & 0xfffffff0;
          ll = (botByte & 15) << 4;
          lr = (UInt32)(UCHAR_ACCESS(botPtr+1)) & 0xfffffff0;
        }

        /*
          Do the top and bottom horizontal interpolations
          using a separable approach. Scale each interp
          coefficient by 4, ie. the number of quarters (1/4 th's).
        */
        switch(RESAMP_FRACTION_PART(x))
        {
        case 0: /* left * 4/4 */
#ifdef _386_ARCH_  /* Shifts are slow on the PC, use adds instead */
            top = ul + ul;
            top += top;
            bottom = ll + ll;
            bottom += bottom;
#else
            top = ul<<2;
            bottom = ll<<2;
#endif

            break;

        case 1: /* left * 3/4 + right * 1/4 */
            top = ul + ul + ul + ur; 
            bottom = ll + ll + ll + lr; 
            break;

        case 2: /* left * 2/4 + right * 2/4 */
            top = ul + ur;
            top += top;
            bottom = ll + lr;
            bottom += bottom;
            break;

        case 3: /* left * 1/4 + right * 3/4 */
            top = ul + ur + ur  + ur; 
            bottom = ll + lr + lr + lr; 
            break;

        default: /* This can't happen, he said, optimistically */
            top = bottom = 0;
            break;
        }

        /*
          Now do a vertical interpolation on the top and
          bottom results of the horizontal interpolations.
          Round and descale the result. Since each of the 
          horizontal and vertical ops are scaled by 4, the
          total scaling is 16X. Adding 8 is thus equiv to
          adding 0.5 (8/16). Treat 4/4 and 2/4 cases as special
          cases to save a couple of shifts.
        */
        switch(RESAMP_FRACTION_PART(y))
        {
        case 0: /* top * 1 */
            outReg <<= outDepth;
            outReg |= (top + 2) >> (2+depthShift);
            break;

        case 1: /* top * 3/4 + bottom * 1/4 */
            outReg <<= outDepth;
            outReg |= (top + top + top + bottom + 8) >> (4+depthShift); 
            break;

        case 2: /* top * 1/2 + bottom * 1/2 */
            outReg <<= outDepth;
            outReg |= ( (top + bottom)  + 4) >> (3+depthShift);
            break;

        case 3: /* top * 1/4 + bottom * 3/4 */
            outReg <<= outDepth;
            outReg |= (top + bottom + bottom + bottom + 8) >> (4+depthShift);
            break;

        default: /* This can't happen, he said, optimistically */
            outReg <<= outDepth;
            break;
        }

LOOP:

/* If enough output values have accumulated, write a word */
        count--;
        if (count == 0)
        {
            if (outDepth == 1)
                outReg = ~outReg;
            *outPtr++ = outReg;
            outReg = 0;          /* Clear the output register again */
            count = pixPerWord;  /* Reset the pixel counter */
        }

        /* Save the current value for comparison */
        oldY = Y;

        /* Step along the row scan path */
        x += dx_u;
        y += dy_u;
        X = RESAMP_INT_PART(x);
        Y = RESAMP_INT_PART(y);
        ydiff = Y - oldY;

        /*
          Use the array of stride values to get to the new
          line offset. Only need to do this if Y has changed.
        */
        if (ydiff != 0)
            lineBase += inBpl[ydiff];

    }

/* Write last word if output width was not a multiple of 32 */
    if (extraBits != 0)
    {
        if (outDepth == 1)
            outReg = ~outReg;
        temp = *outPtr;
        *outPtr = (temp & (allOnes>>extraBits)) | (outReg << (32 - extraBits));
    }

    NORMAL_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Resample routine to produce a rectangular block of output
  from an arbitrary quadrilateral area in the source image.
  This is the routine to be called as the core of a
  general purpose warping or rubber sheeting capability.
  Calls resampleLine repeatedly. This will allow progress
  monitoring to be easily accomodated. Breaking things up
  this way keeps the logic within the resampleLine routine
  simpler.
*/

static ErrCode CDECL 
ip_resampleBlock (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo,
    UInt8      *inBlock,  /* Input image buffer starting address */
    Int32	inWidth,  /* Input image width */
    Int32	inHeight, /* Input image height */
    Int32	inDepth, /* Input image depth */
    Int32      *inBpl,	  /* Bytes per line in the input buffer */
    UInt32     *outBlock,  /* Output image buffer starting address */
    Int32	outX,	 /* Output image X origin */
    Int32	outY,	 /* Output image Y origin */
    Int32	outWidth, /* Output image width */
    Int32	outHeight, /* Output image height */
    Int32	outDepth, /* Output image depth */
    Int32	outBpl)    /* Bytes per line in the output buffer */
{
    PROLOGUE;
    UInt32 *outPtr, *ptr;
    ResampRect inRect;
    Int32 x0, y0, dx_u, dy_u, dx_v, dy_v, ddx_v, ddy_v;
    Int32 resampCode;
    Bool checkBounds;
    Int32 newWidth, newHeight, nbits, fillValue;
    UInt32 pixPerWord, mask32, lastWordMask;
    UInt32 allOnes = 0xffffffff;


/* Map output corner coords into input image space */
    ip_calcCornerCoords(pIPState, warpInfo, outX, outY, outWidth,
		        outHeight, &inRect);

/* 
  Calculate the starting address in the output buffer.
  This will always be 32 bit aligned due to the way
  block widths are calculated, but we must account for
  the output depth
*/
    outPtr = outBlock + outY*outBpl/4 + outX*outDepth/32;


/*
  Test the corners of the output block. It all inside the image
  resample normally. If all outside, supply a zero-filled block.
  Otherwise, subdivide the block into quadrants and recursively
  call this function (ip_resampleBlock). Finally, if the output
  block size reaches a specified minimum size, revert to using
  a more general version of the resampler with code to test for 
  boundary conditions.

  Note on coefficient scaling:

  The final X,Y address is in 16.16 form. However, the second
  order coeffs ddx_v and ddy_v can be very small. To avoid 
  loss of significance when ddx_v is addded to dx_u, 
  dx_u is maintained in the same 4.28 form as ddx_u. After the
  increment, a properly scaled version of the new dx_u is 
  passed down to resampleLine.
*/
    resampCode = ip_testRect (pIPState, &inRect, inWidth, inHeight);
    checkBounds = (resampCode != RESAMP_INSIDE);
    if (resampCode == RESAMP_INSIDE ||
        outWidth <= RESAMP_MINBLOCKSIZE || 
        outHeight <= RESAMP_MINBLOCKSIZE)
    {
        ip_calcResampCoeffs (pIPState, &inRect, outWidth, outHeight, &x0, &y0,
                             &dx_u, &dy_u, &dx_v, &dy_v, &ddx_v, &ddy_v);

        while (outHeight--)
        {
            ERR_PASS(warpInfo->gResampleLine (pIPState, warpInfo,
					      inBlock, outPtr, inBpl,
					      x0, y0, dx_u>>12, dy_u>>12, 
					      outWidth, outDepth,
					      inWidth, inHeight,
					      checkBounds));
            outPtr += outBpl/4;
            x0 += dx_v;
            y0 += dy_v;
            dx_u += ddx_v;
            dy_u += ddy_v;
        }
    }

    /* If outside image, supply background fill */
    else if (resampCode == RESAMP_OUTSIDE)
    {

    /* Create a mask to be used to on the last 32 bit word */
        pixPerWord = 32 / outDepth;
        nbits = outDepth * (outWidth%pixPerWord);
        lastWordMask = 0;
        if (nbits > 0)
        {
            if (outDepth == 1)
                /* Shift zeroes in from MSB */
                lastWordMask = allOnes >> nbits;
            else
                /* Shift ones in from MSB */
                lastWordMask =(UInt32) ( ((Int32)0x80000000) 
                                    / (UInt32) (1<<(nbits-1)) );
        }

    /* Use zero fill for binary images, white for all others */
        fillValue = (outDepth == 1) ? 0 : 0xffffffff;

        while (outHeight--)
        {
            for (ptr=outPtr; ptr<outPtr+outWidth/pixPerWord; ptr++ )
                *ptr = fillValue;

            /* Supply fill for the last word, if any */
            if (nbits > 0)
            {
                if (outDepth == 1)
                    /* Insert zeroes in the MSBs */
                    *ptr &= lastWordMask;
                else
                    /* Inset ones in the MSBs */
                    *ptr |= lastWordMask;
            }

            outPtr += outBpl/4;
        }
    }

    /* 
      Divide the output image into four quadrants and resample each one.
      Adjust the width to assure that the output blocks are 32 bit aligned
    */
    else 
    {
        if (outDepth == 1)
            mask32 = (UInt32) 0xffffffe0;
        else if (outDepth == 2)
            mask32 = (UInt32) 0xfffffff0;
        else if (outDepth == 4)
            mask32 = (UInt32) 0xfffffff8;
        else /* outDepth == 8 */
            mask32 = (UInt32) 0xfffffffc;

        newWidth = (outWidth/2) & mask32; 
        newHeight = (outHeight/2) & mask32;

        /* Upper left quadrant */
        ERR_PASS( ip_resampleBlock(pIPState, warpInfo, inBlock,
				   inWidth, inHeight,
				   inDepth, inBpl, outBlock, outX, outY, 
				   newWidth, newHeight, outDepth, outBpl));

        /* Upper right quadrant */
        ERR_PASS( ip_resampleBlock(pIPState, warpInfo, inBlock,
				   inWidth, inHeight,
				   inDepth, inBpl, outBlock,
				   outX+newWidth, outY, 
				   outWidth-newWidth, newHeight, outDepth,
				   outBpl));

        /* Lower left quadrant */
        ERR_PASS( ip_resampleBlock(pIPState, warpInfo, inBlock,
				   inWidth, inHeight,
				   inDepth, inBpl, outBlock,
				   outX, outY+newHeight, 
				   newWidth, outHeight-newHeight, outDepth,
				   outBpl));

        /* Lower right quadrant */
        ERR_PASS( ip_resampleBlock(pIPState, warpInfo, inBlock,
				   inWidth, inHeight,
				   inDepth, inBpl,
				   outBlock, outX+newWidth, outY+newHeight, 
				   outWidth-newWidth, outHeight-newHeight,
				   outDepth, outBpl));

    }

    NORMAL_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Initialize the parameters of a transformation to perform
  rotation and/or magnification. All parameters must be
  supplied. For rotation only, set Mx=My=1.0. For magnification
  only, set angle to 0.0 . The rotation angle refers to the perceived
  effect on the displayed image.

  Get the transform parameters from the application and
  place them in variables global only to this file.
*/

static Int32 CDECL 
i_initRotateMagTransform (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo, /* Ptr to global warp info structure */
    Float64	angle, /* Rotation angle in radians (CW positive) */
    Float64	Mx,  /* Magnification factor along X axis */
    Float64	My,  /* Magnification factor along Y axis */
    Float64	Xc,  /* X coord of displayed image center in source */
    Float64	Yc,  /* Y coord of displayed image center in source */
    Float64	Uc,  /* X display coord about which to rotate */
    Float64	Vc)      /* Y display coord about which to rotate */ 
{
    PROLOGUE;
    warpInfo->gCos = cos(angle);
    warpInfo->gSin = sin(angle);
    warpInfo->gMx = 1.0/Mx;
    warpInfo->gMy = 1.0/My;
    warpInfo->gUc = Uc;
    warpInfo->gVc = Vc;
    warpInfo->gXc = Xc;
    warpInfo->gYc = Yc;
    warpInfo->gResampTransform = ip_rotateMagTransform;

    NORMAL_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Initialize the parameters of a transformation to map
  an arbitrary quadrilateral in the input image into the
  display. Actually the transform does the opposite. It maps
  the display rectangle into the source image.

  Be aware that what appear to be linear transformations will
  not necessarily produce straight lines in the display. This
  is due to the inverse effect of this backwards mapping process.

  Get the transform parameters from the application and
  place them in variables global only to this file.
*/

static Int32 CDECL 
i_initQuadTransform (
    IPState    *pIPState,
    WarpInfoPtr	warpInfo, /* Ptr to global warp info structure */
    UInt32	u0,
    UInt32	v0,
    UInt32	nu,
    UInt32	nv,				 /* Input rectangle*/
    Float64	x0,
    Float64	y0,	       /* Upper left corner in input  */
    Float64	x1,
    Float64	y1,	       /* Upper right corner in input  */
    Float64	x2,
    Float64	y2,	       /* Lower left corner in input  */
    Float64	x3,
    Float64	y3)    	       /* Lower right corner in input  */
{
    PROLOGUE;
    Float64 du, dv;

    du = (nu-1);
    dv = (nv-1);

	/*
	  Test for zero block size - this should never happen, 
	  but let's avoid the possibility of a zero divide
	*/
    ERR_GEN_CLEANUP(((du==0 || dv==0)), ia_invalidParm);

    warpInfo->gX0 = x0;
    warpInfo->gY0 = y0;
    warpInfo->gU0 = u0;
    warpInfo->gV0 = v0;
    warpInfo->gDx_u = (x1 - x0)/du;
    warpInfo->gDy_u = (y1 - y0)/du;
    warpInfo->gDx_v = (x2 - x0)/dv;
    warpInfo->gDy_v = (y2 - y0)/dv;
    warpInfo->gDdx_v = ((x3 - x2) - (x1 - x0))/du/dv;
    warpInfo->gDdy_v = ((y3 - y2) - (y1 - y0))/du/dv;
    warpInfo->gResampTransform = ip_quadTransform;

    NORMAL_RETURN;

ERR_CLEANUP:
    du = dv = RESAMP_MINBLOCKSIZE;     /* Avoid a zero divide */
    ERR_RETURN;

}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Resample by mapping the ENTIRE output block back into the
  source image space. Recursive subdivision is used to find
  sections of the input which are either entirely within or
  entirely outside the image.
*/

Int32 CDECL 
i_resample (
    IPState    *pIPState,
    UInt8      *inBlock,  /* Input image buffer starting address */
    UInt32	inWidth,  /* Input image width */
    UInt32	inHeight, /* Input image height */
    UInt32	inDepth, /* Input image depth */
    Int32	inBpl,	 /* Bytes per line in the input buffer */
    UInt8      *outBlock, /* Output image buffer starting address */
    UInt32	outWidth, /* Output image width */
    UInt32	outHeight, /* Output image height */
    UInt32	outDepth, /* Output image depth */
    UInt32	outBpl,	  /* Bytes per line in the output buffer */
    Float64	angle,	 /* Rotation angle in radians (CW positive) */
    Float64	Mx,	 /* Magnification factor along X axis */
    Float64	My,	 /* Magnification factor along Y axis */
    Float64	Xc,	 /* X coord of displayed image center in source */
    Float64	Yc,	 /* Y coord of displayed image center in source */
    Float64	Uc,	 /* X display coord about which to rotate */
    Float64	Vc)     /* Y display coord about which to rotate */
{
    PROLOGUE;
    Int32	status;
    Int32	bplArray[9];
    Int32      *bpl;
    Int32	i, stride;
    Int32	outX, outY;
    WarpInfoStruct warpInfo;

/* Check magnification -- must be >= 0.25 */
    if (ABS(Mx) < 0.25 || ABS(My) < 0.25)
        return ia_invalidParm;

/* Verify that output is 32 bit aligned */
    if ( ((UInt32)outBlock % 4) != 0)
        return ia_invalidParm;

/* Generate an array of stride values for + or - 4 lines */
    bpl = &bplArray[4];
    bpl[0] = 0;
    stride = inBpl;
    for (i=1; i<=4; i++)
    {
        bpl[i] = stride; 
        bpl[-i] = - stride;
        stride += inBpl;
    }

/* Initialize the rotation/magnification transform */
    status = i_initRotateMagTransform (pIPState, &warpInfo, 
				       angle, Mx, My,
				       Xc, Yc, Uc, Vc);
    if (status != ia_successful)
        return status;

/*
   Set the depth-specific resampling function
   For 1 and 2 bpp, create the resampling lookup table
*/
    switch (inDepth)
    {
    case 1:
        ERR_PASS(ip_createInterpLut(pIPState, inDepth, outDepth,
				    &warpInfo.gInterpLut));
        warpInfo.gResampleLine = ip_resampleLine1Bpp;
        break;
    case 2:
        ERR_PASS(ip_createInterpLut(pIPState, inDepth, outDepth,
				    &warpInfo.gInterpLut));
        warpInfo.gResampleLine = ip_resampleLine2Bpp;
        break;
    case 4:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine4Bpp;
        break;
    case 8:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine8Bpp;
        break;
    default:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine8Bpp;
        break;
    }

/* 
  Call the block resampler, giving it the whole output image
  as the initial target. This will be recursively subdivided as 
  necessary to produce the output image.
*/
    outX = 0;
    outY = 0;
    ERR_PASS_CLEANUP(ip_resampleBlock(pIPState, &warpInfo, inBlock,
				      inWidth, inHeight,
				   inDepth, bpl,
				   (UInt32 *)outBlock, outX, outY, 
				   outWidth, outHeight, outDepth, outBpl));

/* Free the resampling lookup table */
    if (warpInfo.gInterpLut != NULL)
        IP_FREE(warpInfo.gInterpLut);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (warpInfo.gInterpLut != NULL)
        IP_FREE(warpInfo.gInterpLut);

    ERR_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Resample the image under the control of a quadilateral transform.
  The four corners of the destination image rectangle are mapped 
  into the source image. The caller supplies the four source
  coords and the program calculates the transform. The same 
  recursive subdivision algorithm is used as that in i_resample. 
*/

Int32 CDECL 
i_resampleQuad (
    IPState    *pIPState,
    UInt8      *inBlock,	/* Input image buffer starting addr */
    UInt32	inWidth,	/* Input image width */
    UInt32	inHeight,	/* Input image height */
    UInt32	inDepth,	/* Input image depth */
    Int32	inBpl,		/* Bytes per line in the input buffer */
    UInt8      *outBlock,	/* Output image buffer starting addr */
    UInt32	outWidth,	/* Output image width */
    UInt32	outHeight,	/* Output image height */
    UInt32	outDepth,	/* Output image depth */
    UInt32	outBpl,		/* Bytes per line in the output buffer */
    UInt32	u0,
    UInt32	v0,		/* Dest rectangle origin */
    UInt32	nu,
    UInt32	nv,		/* Dest rectangle width, height */
    Float64	x0,
    Float64	y0,		/* Upper left corner coords in input  */
    Float64	x1,
    Float64	y1,		/* Upper right corner coords in input */
    Float64	x2,
    Float64	y2,		/* Lower left corner coords in input  */
    Float64	x3,
    Float64	y3) 		/* Lower right corner coords in input */
{
    PROLOGUE;
    Int32	status;
    Int32	bplArray[9];
    Int32      *bpl;
    Int32	i, stride;
    Int32	outX, outY;
    WarpInfoStruct warpInfo;

/* Verify that output is 32 bit aligned */
    if ( ((UInt32)outBlock % 4) != 0)
        return ia_invalidParm;

/* Generate an array of stride values for + or - 4 lines */
    bpl = &bplArray[4];
    bpl[0] = 0;
    stride = inBpl;
    for (i=1; i<=4; i++)
    {
        bpl[i] = stride; 
        bpl[-i] = - stride;
        stride += inBpl;
    }

/* Initialize the quadrilateral transform */
    status = i_initQuadTransform (pIPState, &warpInfo, 
				  u0, v0, nu, nv,
				  x0, y0, x1, y1, x2, y2, x3, y3);
    if (status != ia_successful)
        return status;

/*
   Set the depth-specific resampling function
   For 1 and 2 bpp, create the resampling lookup table
*/
    switch (inDepth)
    {
    case 1:
        ERR_PASS(ip_createInterpLut(pIPState, inDepth, outDepth,
				    &warpInfo.gInterpLut));
        warpInfo.gResampleLine = ip_resampleLine1Bpp;
        break;
    case 2:
        ERR_PASS(ip_createInterpLut(pIPState, inDepth, outDepth,
				    &warpInfo.gInterpLut));
        warpInfo.gResampleLine = ip_resampleLine2Bpp;
        break;
    case 4:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine4Bpp;
        break;
    case 8:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine8Bpp;
        break;
    default:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine8Bpp;
        break;
    }

/* 
  Call the block resampler, giving it the whole output image
  as the initial target. This will be recursively subdivided as 
  necessary to produce the output image.
*/
    outX = 0;
    outY = 0;
    ERR_PASS_CLEANUP(ip_resampleBlock(pIPState, &warpInfo, inBlock,
				      inWidth, inHeight,
				      inDepth, bpl,
				      (UInt32 *)outBlock, outX, outY, 
				      outWidth, outHeight, outDepth, outBpl));

/* Free the resampling lookup table */
    if (warpInfo.gInterpLut != NULL)
        IP_FREE(warpInfo.gInterpLut);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (warpInfo.gInterpLut != NULL)
        IP_FREE(warpInfo.gInterpLut);

    ERR_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Resample an image under the control of a user-specified transform.
  The resampling is performed by tesselating the output space into a
  set of blocks and mapping each block back into the input
  image using the supplied transform. If the block is larger
  than RESAMP_MINBLOCKSIZE, it may be recursively subdivided
  if it falls on the image edge.
*/
 
Int32 CDECL 
i_resampleUser (
    IPState    *pIPState,
    UInt8      *inBlock,  /* Input image buffer starting addr */
    Int32	inWidth,  /* Input image width */
    Int32	inHeight, /* Input image height */
    Int32	inDepth, /* Input image depth */
    Int32	inBpl,	 /* Bytes per line in the input buffer */
    UInt8      *outBlock, /* Output image buffer starting addr */
    Int32	outWidth, /* Output image width */
    Int32	outHeight, /* Output image height */
    Int32	outDepth, /* Output image depth */
    Int32	outBpl,	   /* Bytes per line in the output buffer */
    void (CDECL *userTransform) (   /* Transform from dst to src */
		    IPState    *pIPState,
		    void       *dummy,
		    Int32	u,
		    Int32	v,
		    Float64    *x,
		    Float64    *y),
    Int32	blksiz)   /* Output block (square) size to use */
{
    PROLOGUE;
    Int32	bplArray[9];
    Int32      *bpl;
    Int32	i, stride;
    Int32	outX, outY;
    WarpInfoStruct warpInfo;

/* Verify that output is 32 bit aligned */
    if ( ((UInt32)outBlock % 4) != 0)
        return ia_invalidParm;

/* Generate an array of stride values for + or - 4 lines */
    bpl = &bplArray[4];
    bpl[0] = 0;
    stride = inBpl;
    for (i=1; i<=4; i++)
    {
        bpl[i] = stride; 
        bpl[-i] = - stride;
        stride += inBpl;
    }

/*
   Set the depth-specific resampling function
   For 1 and 2 bpp, create the resampling lookup table
*/
    switch (inDepth)
    {
    case 1:
        ERR_PASS(ip_createInterpLut(pIPState, inDepth, outDepth,
				    &warpInfo.gInterpLut));
        warpInfo.gResampleLine = ip_resampleLine1Bpp;
        break;
    case 2:
        ERR_PASS(ip_createInterpLut(pIPState, inDepth, outDepth,
				    &warpInfo.gInterpLut));
        warpInfo.gResampleLine = ip_resampleLine2Bpp;
        break;
    case 4:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine4Bpp;
        break;
    case 8:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine8Bpp;
        break;
    default:
        warpInfo.gInterpLut = NULL;
        warpInfo.gResampleLine = ip_resampleLine8Bpp;
        break;
    }

/* 
  Set the global transform function variable 
  to the passed-in routine
*/
    warpInfo.gResampTransform = (ResampleTransform) userTransform;

/* 
  Call the block resampler a minimum sized block at a time. This
  method should be used when it is known that some non-linear
  distortions are present. The piecewise method generally works
  with most any distortion if the block size is made small enough.
*/
    for (outY=0; outY<outHeight; outY+=blksiz)
    {
        for (outX=0; outX<outWidth; outX+=blksiz)
	{
            ERR_PASS_CLEANUP(ip_resampleBlock(pIPState, &warpInfo, inBlock,
					      inWidth, inHeight, inDepth, bpl,
					      (UInt32 *)outBlock, outX, outY, 
					      MIN(blksiz, outWidth-outX), 
					      MIN(blksiz, outHeight-outY), 
					      outDepth,
					      outBpl));
	}
    }

    if (warpInfo.gInterpLut != NULL)
        IP_FREE(warpInfo.gInterpLut);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (warpInfo.gInterpLut != NULL)
        IP_FREE(warpInfo.gInterpLut);

    ERR_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
$Log: i_warp.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.13  1998/01/13 23:50:29  ddavies
i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.12  1997/07/11 01:49:16  ddavies
i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.11  1995/12/13 02:46:47  ejaquith
i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.10  1995/10/21  02:01:56  ddavies
 * ddavies on Fri Oct 20 19:01:23 PDT 1995
 *
 * i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v
 *
 * Change __WATCOMC__ flag to _386_ARCH_ since it's trying to decide whether
 * to shift or add based on the instruction timings.
 *
 * Revision 1.9  1995/10/16  17:50:48  dferrell
 * dferrell on Mon Oct 16 10:50:02 PDT 1995
 *
 * i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.7  1995/10/05  04:21:00  ddavies
 * ddavies on Wed Oct 4 21:20:00 PDT 1995
 *
 * i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v
 *
 * Change the angle passed to i_resample from degrees to radians.  Set the
 * dummy argument of the userTransform callback proc used by i_resampleUser
 * to a void * so that the client doesn't have to have a definition of the
 * private WarpInfo structure.  Have i_resampleUser cast the userTransform
 * to the new ResampleTransform procedure type when it is put into
 * the WarpInfo structure to be used by lower level code.
 *
 * Revision 1.6  1994/09/29  21:47:16  lperry
 * lperry on Thu Sep 29 14:46:47 PDT 1994
 *
 * i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v
 *
 * Change names of included headers from warp.xxx
 * to shrwarp.xxx
 *
 * Revision 1.5  1994/07/15  18:50:28  dferrell
 * dferrell on Fri Jul 15 11:49:45 PDT 1994
 *
 * i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v
 *
 * Changed instances of free () to IP_FREE () so that the Mac doesn't
 * lose its mind.
 *
 * Revision 1.4  1994/05/12  21:40:41  lperry
 * lperry on Thu May 12 14:33:23 PDT 1994
 *
 * i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v
 *
 * Extensive changes. Now handles all bit depths.
 * Takes 1, 2, 4, or 8 bpp image in and outputs
 * any choice of bit depth (1,2,4,8).
 * Changed calling sequences to simplify usage.
 * Previous init call deleted -- all parameters
 * are passed to the warp function itself (rotation
 * angles, magnification, etc).
 * Its also now reentrant -- got rid of statics and globals.
 * Still not happy with quality of results with 2 bpp
 * input dithered images, but its much beter than it was.
 *
 * Revision 1.3  1994/03/24  02:50:17  lperry
 * lperry on Wed Mar 23 18:48:47 PST 1994
 *
 * i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v
 *
 * Made all pixr memory reads and writes use UCHAR_ACCESS
 * so PC code will work correctly
 *
 * Revision 1.2  1994/03/21  20:11:48  lperry
 * lperry on Mon Mar 21 12:10:02 PST 1994
 *
 * i_warp.c -> /product/ipcore/ipshared/src/RCS/i_warp.c,v
 *
 * Oops. Need to have a way for callers to supply their
 *
 * Revision 1.1  1994/03/21  17:52:21  lperry
 * Initial revision
 *
*/
