/*
 * jcapimin.c
 *
 * Copyright (C) 1994-1995, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains application interface code for the compression half
 * of the JPEG library.  These are the "minimum" API routines that may be
 * needed in either the normal full-compression case or the transcoding-only
 * case.
 *
 * Most of the routines intended to be called directly by an application
 * are in this file or in jcapistd.c.  But also see jcparam.c for
 * parameter-setup helper routines, jcomapi.c for routines shared by
 * compression and decompression, and jctrans.c for the transcoding case.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jcapimin.c,v $; $Revision: 1.2 $; $Date: 2001/01/18 00:24:20 $");

/*
 * Initialization of a JPEG compression object.
 * The error manager must already be set up (in case memory manager fails).
 */

GLOBAL ErrCode CDECL 
jpeg_create_compress (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  int i;

  /* For debugging purposes, zero the whole master structure.
   * But error manager pointer is already there, so save and restore it.
   */
  {
    struct jpeg_error_mgr *err = cinfo->err;

    MEMZERO(cinfo, SIZEOF(struct jpeg_compress_struct));

    /* restore known values */
    cinfo->err = err;
  }
  cinfo->is_decompressor = FALSE;

  /* Initialize a memory manager instance for this object */
  ERR_PASS(jinit_memory_mgr (pIPState, (j_common_ptr) cinfo));

  /* Zero out pointers to permanent structures. */
  cinfo->progress = NULL;
  cinfo->dest = NULL;

  cinfo->comp_info = NULL;

  for (i = 0; i < NUM_QUANT_TBLS; i++)
    cinfo->quant_tbl_ptrs[i] = NULL;

  for (i = 0; i < NUM_HUFF_TBLS; i++) {
    cinfo->dc_huff_tbl_ptrs[i] = NULL;
    cinfo->ac_huff_tbl_ptrs[i] = NULL;
  }

  cinfo->input_gamma = 1.0;	/* in case application forgets */

  /* OK, I'm ready */
  cinfo->global_state = CSTATE_START;

  NORMAL_RETURN;
}


/*
 * Destruction of a JPEG compression object
 */

GLOBAL ErrCode CDECL 
jpeg_destroy_compress (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  ERR_PASS(jpeg_destroy (pIPState, (j_common_ptr) cinfo)); /* use common routine */

  NORMAL_RETURN;
}


/*
 * Abort processing of a JPEG compression operation,
 * but don't destroy the object itself.
 */

GLOBAL ErrCode CDECL 
jpeg_abort_compress (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  ERR_PASS(jpeg_abort (pIPState, (j_common_ptr) cinfo)); /* use common routine */

  NORMAL_RETURN;
}


/*
 * Forcibly suppress or un-suppress all quantization and Huffman tables.
 * Marks all currently defined tables as already written (if suppress)
 * or not written (if !suppress).  This will control whether they get emitted
 * by a subsequent jpeg_start_compress call.
 *
 * This routine is exported for use by applications that want to produce
 * abbreviated JPEG datastreams.  It logically belongs in jcparam.c, but
 * since it is called by jpeg_start_compress, we put it here --- otherwise
 * jcparam.o would be linked whether the application used it or not.
 */

GLOBAL ErrCode CDECL 
jpeg_suppress_tables (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    boolean	suppress)
{
  PROLOGUE;

  int i;
  JQUANT_TBL * qtbl;
  JHUFF_TBL * htbl;

  for (i = 0; i < NUM_QUANT_TBLS; i++) {
    if ((qtbl = cinfo->quant_tbl_ptrs[i]) != NULL)
      qtbl->sent_table = suppress;
  }

  for (i = 0; i < NUM_HUFF_TBLS; i++) {
    if ((htbl = cinfo->dc_huff_tbl_ptrs[i]) != NULL)
      htbl->sent_table = suppress;
    if ((htbl = cinfo->ac_huff_tbl_ptrs[i]) != NULL)
      htbl->sent_table = suppress;
  }
  NORMAL_RETURN;
}


/*
 * Finish JPEG compression.
 *
 * If a multipass operating mode was selected, this may do a great deal of
 * work including most of the actual output.
 */

GLOBAL ErrCode CDECL 
jpeg_finish_compress (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  JDIMENSION iMCU_row;
  boolean no_suspend = TRUE;

  if (cinfo->global_state == CSTATE_SCANNING ||
      cinfo->global_state == CSTATE_RAW_OK) {
    /* Terminate first pass */
    if (cinfo->next_scanline < cinfo->image_height)
      ERREXIT(cinfo, JERR_TOO_LITTLE_DATA);
    (*cinfo->master->finish_pass) (pIPState, cinfo);
  } else if (cinfo->global_state != CSTATE_WRCOEFS)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  /* Perform any remaining passes */
  while (! cinfo->master->is_last_pass) {
    ERR_PASS((*cinfo->master->prepare_for_pass) (pIPState, cinfo));
     for (iMCU_row = 0; iMCU_row < cinfo->total_iMCU_rows; iMCU_row++) {
      if (cinfo->progress != NULL) {
	cinfo->progress->pass_counter = (long) iMCU_row;
	cinfo->progress->pass_limit = (long) cinfo->total_iMCU_rows;
	ERR_PASS((*cinfo->progress->progress_monitor) (pIPState, (j_common_ptr) cinfo));
      }
      /* We bypass the main controller and invoke coef controller directly;
       * all work is being done from the coefficient buffer.
       */
      ERR_PASS((*cinfo->coef->compress_data) (pIPState, cinfo,
					      (JSAMPIMAGE) NULL,
					      &no_suspend));
      if( !no_suspend )
	ERREXIT(cinfo, JERR_CANT_SUSPEND);
    }
    ERR_PASS((*cinfo->master->finish_pass) (pIPState, cinfo));
  }
  /* Write EOI, do final cleanup */
  ERR_PASS((*cinfo->marker->write_file_trailer) (pIPState, cinfo));
  ERR_PASS((*cinfo->dest->term_destination) (pIPState, cinfo));
  /* We can use jpeg_abort to release memory and reset global_state */
  ERR_PASS(jpeg_abort (pIPState, (j_common_ptr) cinfo));

  NORMAL_RETURN;
}


/*
 * Write a special marker.
 * This is only recommended for writing COM or APPn markers.
 * Must be called after jpeg_start_compress() and before
 * first call to jpeg_write_scanlines() or jpeg_write_raw_data().
 */

GLOBAL ErrCode CDECL 
jpeg_write_marker (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    int		marker,
    const JOCTET *dataptr,
    unsigned int datalen)
{
  PROLOGUE;

  if (cinfo->next_scanline != 0 ||
      (cinfo->global_state != CSTATE_SCANNING &&
       cinfo->global_state != CSTATE_RAW_OK &&
       cinfo->global_state != CSTATE_WRCOEFS))
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);

  ERR_PASS((*cinfo->marker->write_any_marker) (pIPState, cinfo, marker, dataptr,
  datalen));

  NORMAL_RETURN;
}


/*
 * Alternate compression function: just write an abbreviated table file.
 * Before calling this, all parameters and a data destination must be set up.
 *
 * To produce a pair of files containing abbreviated tables and abbreviated
 * image data, one would proceed as follows:
 *
 *		initialize JPEG object
 *		set JPEG parameters
 *		set destination to table file
 *		jpeg_write_tables(cinfo);
 *		set destination to image file
 *		jpeg_start_compress(cinfo, FALSE);
 *		write data...
 *		jpeg_finish_compress(cinfo);
 *
 * jpeg_write_tables has the side effect of marking all tables written
 * (same as jpeg_suppress_tables(..., TRUE)).  Thus a subsequent start_compress
 * will not re-emit the tables unless it is passed write_all_tables=TRUE.
 */

GLOBAL ErrCode CDECL 
jpeg_write_tables (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  if (cinfo->global_state != CSTATE_START)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);

  /* (Re)initialize error mgr and destination modules */
  ERR_PASS((*cinfo->err->reset_error_mgr) (pIPState, (j_common_ptr) cinfo));
  ERR_PASS((*cinfo->dest->init_destination) (pIPState, cinfo));
  /* Initialize the marker writer ... bit of a crock to do it here. */
  ERR_PASS(jinit_marker_writer (pIPState, cinfo));
  /* Write them tables! */
  ERR_PASS((*cinfo->marker->write_tables_only) (pIPState, cinfo));
  /* And clean up. */
  ERR_PASS((*cinfo->dest->term_destination) (pIPState, cinfo));
  /* We can use jpeg_abort to release memory ... is this necessary? */
  ERR_PASS(jpeg_abort (pIPState, (j_common_ptr) cinfo));

  NORMAL_RETURN;
}

/*
$Log: jcapimin.c,v $
Revision 1.2  2001/01/18 00:24:20  dabercro
Added comment.

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.3  1998/01/13 23:50:37  ddavies
jcapimin.c -> /product/ipcore/ipshared/src/RCS/jcapimin.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1996/11/06 00:36:47  xzhu
xzhu on Tue Nov 5 16:36:46 PST 1996

jcapimin.c -> /product/ipcore/ipshared/src/RCS/jcapimin.c,v

CommonCheck: Did some logical changes to let function return success or error. All outputs go through parameters.

Revision 1.1  1996/04/20 01:28:40  xzhu
Initial revision

*/
