/*
 * jcapistd.c
 *
 * Copyright (C) 1994-1995, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains application interface code for the compression half
 * of the JPEG library.  These are the "standard" API routines that are
 * used in the normal full-compression case.  They are not used by a
 * transcoding-only application.  Note that if an application links in
 * jpeg_start_compress, it will end up linking in the entire compressor.
 * We thus must separate this file from jcapimin.c to avoid linking the
 * whole compression library into a transcoder.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jcapistd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");

/*
 * Compression initialization.
 * Before calling this, all parameters and a data destination must be set up.
 *
 * We require a write_all_tables parameter as a failsafe check when writing
 * multiple datastreams from the same compression object.  Since prior runs
 * will have left all the tables marked sent_table=TRUE, a subsequent run
 * would emit an abbreviated stream (no tables) by default.  This may be what
 * is wanted, but for safety's sake it should not be the default behavior:
 * programmers should have to make a deliberate choice to emit abbreviated
 * images.  Therefore the documentation and examples should encourage people
 * to pass write_all_tables=TRUE; then it will take active thought to do the
 * wrong thing.
 */

GLOBAL ErrCode CDECL 
jpeg_start_compress (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    boolean	write_all_tables)
{
  PROLOGUE;

  if (cinfo->global_state != CSTATE_START)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);

  if (write_all_tables)
    ERR_PASS(jpeg_suppress_tables (pIPState, cinfo, FALSE));  
                                          /* mark all tables to be written */

  /* (Re)initialize error mgr and destination modules */
  ERR_PASS((*cinfo->err->reset_error_mgr) (pIPState, (j_common_ptr) cinfo));
  ERR_PASS((*cinfo->dest->init_destination) (pIPState, cinfo));
  /* Perform master selection of active modules */
  ERR_PASS(jinit_compress_master (pIPState, cinfo));
  /* Set up for the first pass */
  ERR_PASS((*cinfo->master->prepare_for_pass) (pIPState, cinfo));
  /* Ready for application to drive first pass through jpeg_write_scanlines
   * or jpeg_write_raw_data.
   */
  cinfo->next_scanline = 0;
  cinfo->global_state = (cinfo->raw_data_in ? CSTATE_RAW_OK : CSTATE_SCANNING);

  NORMAL_RETURN;
}


/*
 * Write some scanlines of data to the JPEG compressor.
 *
 * The return value will be the number of lines actually written.
 * This should be less than the supplied num_lines only in case that
 * the data destination module has requested suspension of the compressor,
 * or if more than image_height scanlines are passed in.
 *
 * Note: we warn about excess calls to jpeg_write_scanlines() since
 * this likely signals an application programmer error.  However,
 * excess scanlines passed in the last valid call are *silently* ignored,
 * so that the application need not adjust num_lines for end-of-image
 * when using a multiple-scanline buffer.
 */

GLOBAL ErrCode CDECL 
jpeg_write_scanlines (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	scanlines,
    JDIMENSION	num_lines,
    JDIMENSION *num_lines_written)
{
  PROLOGUE;

  JDIMENSION row_ctr, rows_left;

  if (cinfo->global_state != CSTATE_SCANNING)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  if (cinfo->next_scanline >= cinfo->image_height)
    WARNMS(cinfo, JWRN_TOO_MUCH_DATA);

  /* Call progress monitor hook if present */
  if (cinfo->progress != NULL) {
    cinfo->progress->pass_counter = (long) cinfo->next_scanline;
    cinfo->progress->pass_limit = (long) cinfo->image_height;
    ERR_PASS((*cinfo->progress->progress_monitor) (pIPState,
						   (j_common_ptr) cinfo));
  }

  /* Give master control module another chance if this is first call to
   * jpeg_write_scanlines.  This lets output of the frame/scan headers be
   * delayed so that application can write COM, etc, markers between
   * jpeg_start_compress and jpeg_write_scanlines.
   */
  if (cinfo->master->call_pass_startup)
    ERR_PASS((*cinfo->master->pass_startup) (pIPState, cinfo));

  /* Ignore any extra scanlines at bottom of image. */
  rows_left = cinfo->image_height - cinfo->next_scanline;
  if (num_lines > rows_left)
    num_lines = rows_left;

  row_ctr = 0;
  ERR_PASS((*cinfo->main->process_data) (pIPState, cinfo, scanlines, &row_ctr,
  num_lines));
  cinfo->next_scanline += row_ctr;

  *num_lines_written = row_ctr;

  NORMAL_RETURN;
}


/*
 * Alternate entry point to write raw data.
 * Processes exactly one iMCU row per call, unless suspended.
 */

GLOBAL ErrCode CDECL 
jpeg_write_raw_data (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPIMAGE	data,
    JDIMENSION	num_lines,
    JDIMENSION *num_lines_written)
{
  PROLOGUE;

  JDIMENSION lines_per_MCU_row;
  boolean no_suspend = TRUE;

  if (cinfo->global_state != CSTATE_RAW_OK)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  if (cinfo->next_scanline >= cinfo->image_height) {
    WARNMS(cinfo, JWRN_TOO_MUCH_DATA);
  }

  /* Call progress monitor hook if present */
  if (cinfo->progress != NULL) {
    cinfo->progress->pass_counter = (long) cinfo->next_scanline;
    cinfo->progress->pass_limit = (long) cinfo->image_height;
    ERR_PASS((*cinfo->progress->progress_monitor) (pIPState,
						   (j_common_ptr) cinfo));
  }

  /* Give master control module another chance if this is first call to
   * jpeg_write_raw_data.  This lets output of the frame/scan headers be
   * delayed so that application can write COM, etc, markers between
   * jpeg_start_compress and jpeg_write_raw_data.
   */
  if (cinfo->master->call_pass_startup)
    ERR_PASS((*cinfo->master->pass_startup) (pIPState, cinfo));

  /* Verify that at least one iMCU row has been passed. */
  lines_per_MCU_row = cinfo->max_v_samp_factor * DCTSIZE;
  if (num_lines < lines_per_MCU_row)
    ERREXIT(cinfo, JERR_BUFFER_SIZE);

  /* Directly compress the row. */
  ERR_PASS( (*cinfo->coef->compress_data) (pIPState, cinfo, data, &no_suspend));
  if( !no_suspend )
  {
    /* If compressor did not consume the whole row, suspend processing. */
      *num_lines_written = 0;
      NORMAL_RETURN;
  }

  /* OK, we processed one iMCU row. */
  cinfo->next_scanline += lines_per_MCU_row;

  *num_lines_written = lines_per_MCU_row;

  NORMAL_RETURN;
}


/*
$Log: jcapistd.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.3  1998/01/13 23:50:39  ddavies
jcapistd.c -> /product/ipcore/ipshared/src/RCS/jcapistd.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1996/11/06 00:36:51  xzhu
jcapistd.c -> /product/ipcore/ipshared/src/RCS/jcapistd.c,v

CommonCheck: Did some logical changes to let function return success or error. All outputs go through parameters.

Revision 1.1  1996/04/20 01:29:00  xzhu
Initial revision

*/
