/*
*************************************************************************
 *  Copyright (c) 1995-2001, Xerox Corporation.  All rights reserved.   *
 *  Copyright protection claimed includes all forms and matters         *
 *  of copyrightable material and information now allowed by            *
 *  statutory or judicial law or hereafter granted, including           *
 *  without limitation, material generated from the software            *
 *  programs which are displayed on the screen such as icons,           *
 *  screen display looks, etc.                                          *
 ************************************************************************
 *
 * jccolor.c
 *
 * Copyright (C) 1991-1994, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains input colorspace conversion routines.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"


IP_RCSINFO(RCSInfo, "$RCSfile: jccolor.c,v $; $Revision: 1.17 $; $Date: 2001/07/06 18:17:53 $");

/* Private subobject */

typedef struct my_color_converter_s {
  struct jpeg_color_converter pub; /* public fields */

  /* Private state for RGB->YCC conversion */
  Int32 * rgb_ycc_tab;		/* => table for RGB to YCbCr conversion */
  Int32 * rgb_lab_tab;		/* => table for RGB to LAB f() conversion */
} my_color_converter;

typedef my_color_converter *my_cconvert_ptr;


/**************** RGB -> YCbCr conversion: most common case **************/

/*
 * YCbCr is defined per CCIR 601-1, except that Cb and Cr are
 * normalized to the range 0..MAXJSAMPLE rather than -0.5 .. 0.5.
 * The conversion equations to be implemented are therefore
 *	Y  =  0.29900 * R + 0.58700 * G + 0.11400 * B
 *	Cb = -0.16874 * R - 0.33126 * G + 0.50000 * B  + CENTERJSAMPLE
 *	Cr =  0.50000 * R - 0.41869 * G - 0.08131 * B  + CENTERJSAMPLE
 * (These numbers are derived from TIFF 6.0 section 21, dated 3-June-92.)
 * Note: older versions of the IJG code used a zero offset of MAXJSAMPLE/2,
 * rather than CENTERJSAMPLE, for Cb and Cr.  This gave equal positive and
 * negative swings for Cb/Cr, but meant that grayscale values (Cb=Cr=0)
 * were not represented exactly.  Now we sacrifice exact representation of
 * maximum red and maximum blue in order to get exact grayscales.
 *
 * To avoid floating-point arithmetic, we represent the fractional constants
 * as integers scaled up by 2^16 (about 4 digits precision); we have to divide
 * the products by 2^16, with appropriate rounding, to get the correct answer.
 *
 * For even more speed, we avoid doing any multiplications in the inner loop
 * by precalculating the constants times R,G,B for all possible values.
 * For 8-bit JSAMPLEs this is very reasonable (only 256 entries per table);
 * for 12-bit samples it is still acceptable.  It's not very reasonable for
 * 16-bit samples, but if you want lossless storage you shouldn't be changing
 * colorspace anyway.
 * The CENTERJSAMPLE offsets and the rounding fudge-factor of 0.5 are included
 * in the tables to save adding them separately in the inner loop.
 */

#define SCALEBITS	16	/* speediest right-shift on some machines */
#define CBCR_OFFSET	((Int32) CENTERJSAMPLE << SCALEBITS)
#define ONE_HALF	((Int32) 1 << (SCALEBITS-1))
#define FIX(x)		((Int32) ((x) * (1L<<SCALEBITS) + 0.5))

/* We allocate one big table and divide it up into eight parts, instead of
 * doing eight alloc_small requests.  This lets us use a single table base
 * address, which can be held in a register in the inner loops on many
 * machines (more than can hold all eight addresses, anyway).
 */

#define R_Y_OFF		0			/* offset to R => Y section */
#define G_Y_OFF		(1*(MAXJSAMPLE+1))	/* offset to G => Y section */
#define B_Y_OFF		(2*(MAXJSAMPLE+1))	/* etc. */
#define R_CB_OFF	(3*(MAXJSAMPLE+1))
#define G_CB_OFF	(4*(MAXJSAMPLE+1))
#define B_CB_OFF	(5*(MAXJSAMPLE+1))
#define R_CR_OFF	B_CB_OFF		/* B=>Cb, R=>Cr are the same */
#define G_CR_OFF	(6*(MAXJSAMPLE+1))
#define B_CR_OFF	(7*(MAXJSAMPLE+1))
#define TABLE_SIZE	(8*(MAXJSAMPLE+1))


/*
 * Initialize for RGB->YCC colorspace conversion.
 */

METHODDEF ErrCode CDECL 
rgb_ycc_start (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  Int32 * rgb_ycc_tab;
  Int32 i;

  /* Allocate and fill in the conversion tables. */
  ERR_PASS((*cinfo->mem->alloc_small) (pIPState, (j_common_ptr) cinfo,	
				JPOOL_IMAGE,
				(TABLE_SIZE * SIZEOF(Int32)),
                                (void **) &rgb_ycc_tab));
  cconvert->rgb_ycc_tab = rgb_ycc_tab;

  for (i = 0; i <= MAXJSAMPLE; i++) {
    rgb_ycc_tab[i+R_Y_OFF] = FIX(0.29900) * i;
    rgb_ycc_tab[i+G_Y_OFF] = FIX(0.58700) * i;
    rgb_ycc_tab[i+B_Y_OFF] = FIX(0.11400) * i     + ONE_HALF;
    rgb_ycc_tab[i+R_CB_OFF] = (-FIX(0.16874)) * i;
    rgb_ycc_tab[i+G_CB_OFF] = (-FIX(0.33126)) * i;
    /* We use a rounding fudge-factor of 0.5-epsilon for Cb and Cr.
     * This ensures that the maximum output will round to MAXJSAMPLE
     * not MAXJSAMPLE+1, and thus that we don't have to range-limit.
     */
    rgb_ycc_tab[i+B_CB_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
/*  B=>Cb and R=>Cr tables are the same
    rgb_ycc_tab[i+R_CR_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
*/
    rgb_ycc_tab[i+G_CR_OFF] = (-FIX(0.41869)) * i;
    rgb_ycc_tab[i+B_CR_OFF] = (-FIX(0.08131)) * i;
  }

  NORMAL_RETURN;
}


/*
 * Convert some rows of samples to the JPEG colorspace.
 *
 * Note that we change from the application's interleaved-pixel format
 * to our internal noninterleaved, one-plane-per-component format.
 * The input buffer is therefore three times as wide as the output buffer.
 *
 * A starting row offset is provided only for the output buffer.  The caller
 * can easily adjust the passed input_buf value to accommodate any row
 * offset required on that side.
 */

METHODDEF ErrCode CDECL 
rgb_ycc_convert (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
  PROLOGUE;

  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r, g, b;
  register Int32 * ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr0, outptr1, outptr2;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;

  while (--num_rows >= 0) {
    inptr = *input_buf++;
    outptr0 = output_buf[0][output_row];
    outptr1 = output_buf[1][output_row];
    outptr2 = output_buf[2][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++) {
      r = GETJSAMPLE(inptr[RGB_RED]);
      g = GETJSAMPLE(inptr[RGB_GREEN]);
      b = GETJSAMPLE(inptr[RGB_BLUE]);
      inptr += RGB_PIXELSIZE;
      /* If the inputs are 0..MAXJSAMPLE, the outputs of these equations
       * must be too; we do not need an explicit range-limiting operation.
       * Hence the value being shifted is never negative, and we don't
       * need the general RIGHT_SHIFT macro.
       */
      /* Y */
      outptr0[col] = (JSAMPLE)
		((ctab[r+R_Y_OFF] + ctab[g+G_Y_OFF] + ctab[b+B_Y_OFF])
		 >> SCALEBITS);
      /* Cb */
      outptr1[col] = (JSAMPLE)
		((ctab[r+R_CB_OFF] + ctab[g+G_CB_OFF] + ctab[b+B_CB_OFF])
		 >> SCALEBITS);
      /* Cr */
      outptr2[col] = (JSAMPLE)
		((ctab[r+R_CR_OFF] + ctab[g+G_CR_OFF] + ctab[b+B_CR_OFF])
		 >> SCALEBITS);
    }
  }

  NORMAL_RETURN;
}


/**************** Cases other than RGB -> YCbCr **************/


/*
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles RGB->grayscale conversion, which is the same
 * as the RGB->Y portion of RGB->YCbCr.
 * We assume rgb_ycc_start has been called (we only use the Y tables).
 */

METHODDEF ErrCode CDECL 
rgb_gray_convert (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
  PROLOGUE;

  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r, g, b;
  register Int32 * ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;

  while (--num_rows >= 0) {
    inptr = *input_buf++;
    outptr = output_buf[0][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++) {
      r = GETJSAMPLE(inptr[RGB_RED]);
      g = GETJSAMPLE(inptr[RGB_GREEN]);
      b = GETJSAMPLE(inptr[RGB_BLUE]);
      inptr += RGB_PIXELSIZE;
      /* Y */
      outptr[col] = (JSAMPLE)
		((ctab[r+R_Y_OFF] + ctab[g+G_Y_OFF] + ctab[b+B_Y_OFF])
		 >> SCALEBITS);
    }
  }

  NORMAL_RETURN;
}


/*
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles Adobe-style CMYK->YCCK conversion,
 * where we convert R=1-C, G=1-M, and B=1-Y to YCbCr using the same
 * conversion as above, while passing K (black) unchanged.
 * We assume rgb_ycc_start has been called.
 */

METHODDEF ErrCode CDECL 
cmyk_ycck_convert (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
  PROLOGUE;

  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r, g, b;
  register Int32 * ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr0, outptr1, outptr2, outptr3;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;

  while (--num_rows >= 0) {
    inptr = *input_buf++;
    outptr0 = output_buf[0][output_row];
    outptr1 = output_buf[1][output_row];
    outptr2 = output_buf[2][output_row];
    outptr3 = output_buf[3][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++) {
      r = MAXJSAMPLE - GETJSAMPLE(inptr[0]);
      g = MAXJSAMPLE - GETJSAMPLE(inptr[1]);
      b = MAXJSAMPLE - GETJSAMPLE(inptr[2]);
      /* K passes through as-is */
      outptr3[col] = inptr[3];	/* don't need GETJSAMPLE here */
      inptr += 4;
      /* If the inputs are 0..MAXJSAMPLE, the outputs of these equations
       * must be too; we do not need an explicit range-limiting operation.
       * Hence the value being shifted is never negative, and we don't
       * need the general RIGHT_SHIFT macro.
       */
      /* Y */
      outptr0[col] = (JSAMPLE)
		((ctab[r+R_Y_OFF] + ctab[g+G_Y_OFF] + ctab[b+B_Y_OFF])
		 >> SCALEBITS);
      /* Cb */
      outptr1[col] = (JSAMPLE)
		((ctab[r+R_CB_OFF] + ctab[g+G_CB_OFF] + ctab[b+B_CB_OFF])
		 >> SCALEBITS);
      /* Cr */
      outptr2[col] = (JSAMPLE)
		((ctab[r+R_CR_OFF] + ctab[g+G_CR_OFF] + ctab[b+B_CR_OFF])
		 >> SCALEBITS);
    }
  }

  NORMAL_RETURN;
}


/*
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles grayscale output with no conversion.
 */

METHODDEF ErrCode CDECL 
grayscale_convert (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
  PROLOGUE;

  ERR_PASS(jcopy_sample_rows (pIPState, input_buf, 0, output_buf[0], output_row,
		    num_rows, cinfo->image_width));

  NORMAL_RETURN;
}

/* The source can be YCbCr (since Y == gray), changed into gray 
*/

METHODDEF ErrCode CDECL 
YCbCr_gray_convert (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
  PROLOGUE;

  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;
  int instride = cinfo->input_components;

  while (--num_rows >= 0) {
      inptr = *input_buf++;
      outptr = output_buf[0][output_row];
      output_row++;
      for (col = 0; col < num_cols; col++) {
	  *outptr++ = *inptr;	/* don't need GETJSAMPLE() here */
	  inptr += instride;
      }
  }

  NORMAL_RETURN;
}

/*
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles multi-component colorspaces without conversion.
 * We assume input_components == num_components.
 */

METHODDEF ErrCode CDECL 
null_convert (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
  PROLOGUE;

  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  register int ci;
  int nc = cinfo->num_components;
  JDIMENSION num_cols = cinfo->image_width;

  while (--num_rows >= 0) {
    /* It seems fastest to make a separate pass for each component. */
    for (ci = 0; ci < nc; ci++) {
      inptr = *input_buf;
      outptr = output_buf[ci][output_row];
      for (col = 0; col < num_cols; col++) {
	outptr[col] = inptr[ci]; /* don't need GETJSAMPLE() here */
	inptr += nc;
      }
    }
    input_buf++;
    output_row++;
  }

  NORMAL_RETURN;
}


/*
 * Empty method for start_pass.
 */

METHODDEF ErrCode CDECL 
null_method (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  /* Keep compilers happy */
  cinfo = cinfo;

  /* no work needed */

  NORMAL_RETURN;
}


/*
 * Module initialization routine for input colorspace conversion.
 */

GLOBAL ErrCode CDECL 
jinit_color_converter (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  my_cconvert_ptr cconvert;

    ERR_PASS((*cinfo->mem->alloc_small) (pIPState, (j_common_ptr) cinfo,
				JPOOL_IMAGE,
				SIZEOF(my_color_converter),
                                (void **) &cconvert));
  cinfo->cconvert = (struct jpeg_color_converter *) cconvert;
  /* set start_pass to null method until we find out differently */
  cconvert->pub.start_pass = null_method;

  /* Make sure input_components agrees with in_color_space */
  switch (cinfo->in_color_space) {
  case JCS_GRAYSCALE:
    if (cinfo->input_components != 1)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;

  case JCS_RGB:
#if RGB_PIXELSIZE != 3
    if (cinfo->input_components != RGB_PIXELSIZE)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;
#endif /* else share code with YCbCr */

  case JCS_YCbCr:
    if (cinfo->input_components != 3)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;

  case JCS_CMYK:
  case JCS_YCCK:
    if (cinfo->input_components != 4)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;

  default:			/* JCS_UNKNOWN can be anything */
    if (cinfo->input_components < 1)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;
  }

  /* Check num_components, set conversion method based on requested space */
  switch (cinfo->jpeg_color_space) {
  case JCS_GRAYSCALE:
    if (cinfo->num_components != 1)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_GRAYSCALE)
      cconvert->pub.color_convert = grayscale_convert;
    else if (cinfo->in_color_space == JCS_RGB) {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = rgb_gray_convert;
    } else if (cinfo->in_color_space == JCS_YCbCr)
      cconvert->pub.color_convert = YCbCr_gray_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_RGB:
    if (cinfo->num_components != 3)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_RGB && RGB_PIXELSIZE == 3)
      cconvert->pub.color_convert = null_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_YCbCr:
    if (cinfo->num_components != 3)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_RGB) {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = rgb_ycc_convert;
    } else if (cinfo->in_color_space == JCS_YCbCr)
      cconvert->pub.color_convert = null_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_CMYK:
    if (cinfo->num_components != 4)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_CMYK)
      cconvert->pub.color_convert = null_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_YCCK:
    if (cinfo->num_components != 4)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_CMYK) {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = cmyk_ycck_convert;
    } else if (cinfo->in_color_space == JCS_YCCK)
      cconvert->pub.color_convert = null_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  default:			/* allow null conversion of JCS_UNKNOWN */
    if (cinfo->jpeg_color_space != cinfo->in_color_space ||
	cinfo->num_components != cinfo->input_components)
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    cconvert->pub.color_convert = null_convert;
    break;
  }

  NORMAL_RETURN;
}

/*
$Log: jccolor.c,v $
Revision 1.17  2001/07/06 18:17:53  dabercro
Remove unused functions, cleanup

Revision 1.16  2001/06/27 19:12:44  dabercro
Remove colorspace support in JPEG, Lab is supported by identifying the
colorspace as unknown.

Revision 1.15  2001/03/22 17:25:57  dabercro
Fix grayscale ITULAB function when using limited gamut.

Revision 1.14  2001/03/12 23:02:26  dabercro
Use the TIFF-FX default gamut range.

Revision 1.13  2001/03/07 18:34:21  dabercro
Linear RGB replaced by sRGB.  A little code cleanup.

Revision 1.12  2001/01/19 23:59:42  dabercro
Remove editing mishap.
Remove unused function - both ICCLAB and CIELAB share the
same conversion, since the CIELAB was meant to do what
ICCLAB corrected (range was meant to be [-128,127] instead of
[-127,127].

Revision 1.11  2001/01/18 23:29:26  dabercro
Handle 1-component CIELAB/ICCLAB (gray).

Revision 1.10  2001/01/18 00:29:27  dabercro
Fix precision problem.  When quantizing to 8-bit values, I was
scaling to [0,255], rounding, then quantizing.  Problem was that
gave me 255 bins, and the first and last bin corresponded to only
half a pre-quantized value ([0,0.5] went to 0, [0.5,1.5] went to 1,
... [254.5,255.0] went to 255).  Now 256 evenly distributed bins.
Added support for encoding to a non-default gamut.
Added grayscale ITULAB support, and non-default gamut
support for this as well.

Revision 1.9  2000/09/21 21:24:22  dabercro
Avoid unnecessary function call.

Revision 1.8  2000/09/08 22:04:49  dabercro
Slight precision change, to avoid an overflow.

Revision 1.7  2000/09/06 01:28:44  dabercro
Huge optimizations for Lab color space conversion.  12x speed-up
for RGB->LAB, and 4x speed-up for LAB->RGB.

Revision 1.6  2000/08/01 21:46:58  dabercro
Handle grayscale for ITULAB also. Changes from Mark Tobin.

Revision 1.5  2000/04/07 23:59:42  dabercro
Added support for CIELAB and ICCLAB color spaces.

Revision 1.4  2000/02/26 01:28:37  dabercro
Avoid sometime predefined version of fabs().

Revision 1.3  2000/02/26 00:10:54  dabercro
Updated the ITULAB conversion formula, after discussions with Rob Buckley.

Revision 1.2  2000/02/24 00:48:23  dabercro
Changes to support ITULAB color conversion

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.13  1998/01/13 23:50:42  ddavies
jccolor.c -> /product/ipcore/ipshared/src/RCS/jccolor.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.12  1996/04/22 17:30:58  xzhu
xzhu on Mon Apr 22 10:29:55 PDT 1996

jccolor.c -> /product/ipcore/ipshared/src/RCS/jccolor.c,v

Added cinfo = cinfo for no work function to keep compiler happy.

Revision 1.11  1996/04/20 01:33:39  xzhu
xzhu on Fri Apr 19 18:33:15 PDT 1996

jccolor.c -> /product/ipcore/ipshared/src/RCS/jccolor.c,v

new jpeg version

*/
