/*
 * jcinit.c
 *
 * Copyright (C) 1991-1995, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains initialization logic for the JPEG compressor.
 * This routine is in charge of selecting the modules to be executed and
 * making an initialization call to each one.
 *
 * Logically, this code belongs in jcmaster.c.  It's split out because
 * linking this routine implies linking the entire compression library.
 * For a transcoding-only application, we want to be able to use jcmaster.c
 * without linking in the whole library.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jcinit.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");

/*
 * Master selection of compression modules.
 * This is done once at the start of processing an image.  We determine
 * which modules will be used and give them appropriate initialization calls.
 */

GLOBAL ErrCode CDECL 
jinit_compress_master (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  /* Initialize master control (includes parameter checking/processing) */
  ERR_PASS(jinit_c_master_control (pIPState, cinfo,
				   FALSE /* full compression */));

  /* Preprocessing */
  if (! cinfo->raw_data_in) {
    ERR_PASS(jinit_color_converter (pIPState, cinfo));
    ERR_PASS(jinit_downsampler (pIPState, cinfo));
    ERR_PASS(jinit_c_prep_controller (pIPState, cinfo, FALSE)); 
                        /* never need full buffer here */
  }
  /* Forward DCT */
  ERR_PASS(jinit_forward_dct (pIPState, cinfo));
  /* Entropy encoding: either Huffman or arithmetic coding. */
  if (cinfo->arith_code) {
    ERREXIT(cinfo, JERR_ARITH_NOTIMPL);
  } else {
    if (cinfo->progressive_mode) {
#ifdef C_PROGRESSIVE_SUPPORTED
      ERR_PASS(jinit_phuff_encoder (pIPState, cinfo));
#else
      ERREXIT(cinfo, JERR_NOT_COMPILED);
#endif
    } else
      ERR_PASS(jinit_huff_encoder (pIPState, cinfo));
  }

  /* Need a full-image coefficient buffer in any multi-pass mode. */
  ERR_PASS(jinit_c_coef_controller (pIPState, cinfo,
		       (cinfo->num_scans > 1 || cinfo->optimize_coding)));
  ERR_PASS(jinit_c_main_controller (pIPState, cinfo, FALSE));
                                        /* never need full buffer here */

  ERR_PASS(jinit_marker_writer (pIPState, cinfo));

  /* We can now tell the memory manager to allocate virtual arrays. */
  ERR_PASS((*cinfo->mem->realize_virt_arrays) (pIPState, (j_common_ptr) cinfo));

  /* Write the datastream header (SOI) immediately.
   * Frame and scan headers are postponed till later.
   * This lets application insert special markers after the SOI.
   */
  ERR_PASS((*cinfo->marker->write_file_header) (pIPState, cinfo));

  NORMAL_RETURN;
}

/*
$Log: jcinit.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.2  1998/01/13 23:50:47  ddavies
jcinit.c -> /product/ipcore/ipshared/src/RCS/jcinit.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.1  1996/04/20 01:32:00  xzhu
Initial revision

*/
