/*
 * jcmainct.c
 *
 * Copyright (C) 1994-1995, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains the main buffer controller for compression.
 * The main buffer lies between the pre-processor and the JPEG
 * compressor proper; it holds downsampled data in the JPEG colorspace.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jcmainct.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");

/* Note: currently, there is no operating mode in which a full-image buffer
 * is needed at this step.  If there were, that mode could not be used with
 * "raw data" input, since this module is bypassed in that case.  However,
 * we've left the code here for possible use in special applications.
 */
#undef FULL_MAIN_BUFFER_SUPPORTED


/* Private buffer controller object */

typedef struct {
  struct jpeg_c_main_controller pub; /* public fields */

  JDIMENSION cur_iMCU_row;	/* number of current iMCU row */
  JDIMENSION rowgroup_ctr;	/* counts row groups received in iMCU row */
  boolean suspended;		/* remember if we suspended output */
  J_BUF_MODE pass_mode;		/* current operating mode */

  /* If using just a strip buffer, this points to the entire set of buffers
   * (we allocate one for each component).  In the full-image case, this
   * points to the currently accessible strips of the virtual arrays.
   */
  JSAMPARRAY buffer[MAX_COMPONENTS];

#ifdef FULL_MAIN_BUFFER_SUPPORTED
  /* If using full-image storage, this array holds pointers to virtual-array
   * control blocks for each component.  Unused if not full-image storage.
   */
  jvirt_sarray_ptr whole_image[MAX_COMPONENTS];
#endif
} my_main_controller;

typedef my_main_controller * my_main_ptr;


/* Forward declarations */
METHODDEF ErrCode CDECL 
process_data_simple_main (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JDIMENSION *in_row_ctr,
    JDIMENSION	in_rows_avail);
#ifdef FULL_MAIN_BUFFER_SUPPORTED
METHODDEF ErrCode CDECL 
process_data_buffer_main (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JDIMENSION *in_row_ctr,
    JDIMENSION	in_rows_avail);
#endif


/*
 * Initialize for a processing pass.
 */

METHODDEF ErrCode CDECL 
start_pass_main (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    J_BUF_MODE	pass_mode)
{
  PROLOGUE;

  my_main_ptr jmain = (my_main_ptr) cinfo->main;

  /* Do nothing in raw-data mode. */
  if (cinfo->raw_data_in)
      NORMAL_RETURN;

  jmain->cur_iMCU_row = 0;	/* initialize counters */
  jmain->rowgroup_ctr = 0;
  jmain->suspended = FALSE;
  jmain->pass_mode = pass_mode;	/* save mode for use by process_data */

  switch (pass_mode) {
  case JBUF_PASS_THRU:
#ifdef FULL_MAIN_BUFFER_SUPPORTED
    if (jmain->whole_image[0] != NULL)
      ERREXIT(cinfo, JERR_BAD_BUFFER_MODE);
#endif
    jmain->pub.process_data = process_data_simple_main;
    break;
#ifdef FULL_MAIN_BUFFER_SUPPORTED
  case JBUF_SAVE_SOURCE:
  case JBUF_CRANK_DEST:
  case JBUF_SAVE_AND_PASS:
    if (jmain->whole_image[0] == NULL)
      ERREXIT(cinfo, JERR_BAD_BUFFER_MODE);
    jmain->pub.process_data = process_data_buffer_main;
    break;
#endif
  default:
    ERREXIT(cinfo, JERR_BAD_BUFFER_MODE);
    break;
  }

  NORMAL_RETURN;
}


/*
 * Process some data.
 * This routine handles the simple pass-through mode,
 * where we have only a strip buffer.
 */

METHODDEF ErrCode CDECL 
process_data_simple_main (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JDIMENSION *in_row_ctr,
    JDIMENSION	in_rows_avail)
{
  PROLOGUE;

  my_main_ptr jmain = (my_main_ptr) cinfo->main;
  boolean no_suspend = TRUE;

  while (jmain->cur_iMCU_row < cinfo->total_iMCU_rows) {
    /* Read input data if we haven't filled the main buffer yet */
    if (jmain->rowgroup_ctr < DCTSIZE)
      ERR_PASS((*cinfo->prep->pre_process_data) (pIPState, cinfo,
					input_buf, in_row_ctr, in_rows_avail,
					jmain->buffer, &jmain->rowgroup_ctr,
					(JDIMENSION) DCTSIZE));

    /* If we don't have a full iMCU row buffered, return to application for
     * more data. Note that preprocessor will always pad to fill the iMCU row
     * at the bottom of the image.
     */
    if (jmain->rowgroup_ctr != DCTSIZE)
	NORMAL_RETURN;

    /* Send the completed row to the compressor */
    ERR_PASS( (*cinfo->coef->compress_data) (pIPState, cinfo, jmain->buffer,
					     &no_suspend));
    if( !no_suspend )
    {
      /* If compressor did not consume the whole row, then we must need to
       * suspend processing and return to the application.  In this situation
       * we pretend we didn't yet consume the last input row; otherwise, if
       * it happened to be the last row of the image, the application would
       * think we were done.
       */
      if (! jmain->suspended) {
	(*in_row_ctr)--;
	jmain->suspended = TRUE;
      }
      NORMAL_RETURN;
    }
    /* We did finish the row.  Undo our little suspension hack if a previous
     * call suspended; then mark the main buffer empty.
     */
    if (jmain->suspended) {
      (*in_row_ctr)++;
      jmain->suspended = FALSE;
    }
    jmain->rowgroup_ctr = 0;
    jmain->cur_iMCU_row++;
  }

  NORMAL_RETURN;
}


#ifdef FULL_MAIN_BUFFER_SUPPORTED

/*
 * Process some data.
 * This routine handles all of the modes that use a full-size buffer.
 */

METHODDEF ErrCode CDECL 
process_data_buffer_main (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JDIMENSION *in_row_ctr,
    JDIMENSION	in_rows_avail)
{
  PROLOGUE;

  my_main_ptr jmain = (my_main_ptr) cinfo->main;
  int ci;
  jpeg_component_info *compptr;
  boolean writing = (jmain->pass_mode != JBUF_CRANK_DEST);
  boolean no_suspend = TRUE;

  while (jmain->cur_iMCU_row < cinfo->total_iMCU_rows) {
    /* Realign the virtual buffers if at the start of an iMCU row. */
    if (jmain->rowgroup_ctr == 0) {
      for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
	   ci++, compptr++) {
	  ERR_PASS((*cinfo->mem->access_virt_sarray)
	  (pIPState, (j_common_ptr) cinfo, jmain->whole_image[ci],
	   jmain->cur_iMCU_row * (compptr->v_samp_factor * DCTSIZE),
	   (JDIMENSION) (compptr->v_samp_factor * DCTSIZE), writing,
	   &jmain->buffer[ci]));
      }
      /* In a read pass, pretend we just read some source data. */
      if (! writing) {
	*in_row_ctr += cinfo->max_v_samp_factor * DCTSIZE;
	jmain->rowgroup_ctr = DCTSIZE;
      }
    }

    /* If a write pass, read input data until the current iMCU row is full. */
    /* Note: preprocessor will pad if necessary to fill the last iMCU row. */
    if (writing) {
      ERR_PASS((*cinfo->prep->pre_process_data) (pIPState, cinfo,
					input_buf, in_row_ctr, in_rows_avail,
					jmain->buffer, &jmain->rowgroup_ctr,
					(JDIMENSION) DCTSIZE));
      /* Return to application if we need more data to fill the iMCU row. */
      if (jmain->rowgroup_ctr < DCTSIZE)
	  NORMAL_RETURN;
    }

    /* Emit data, unless this is a sink-only pass. */
    if (jmain->pass_mode != JBUF_SAVE_SOURCE) {
	ERR_PASS((*cinfo->coef->compress_data) (pIPState, cinfo, jmain->buffer, 
					       &no_suspend));
	if( !no_suspend ){
 	/* If compressor did not consume the whole row, then we must need to
	 * suspend processing and return to the application.In this situation
	 * we pretend we didn't yet consume the last input row; otherwise, if
	 * it happened to be the last row of the image, the application would
	 * think we were done.
	 */
	if (! jmain->suspended) {
	  (*in_row_ctr)--;
	  jmain->suspended = TRUE;
	}
	NORMAL_RETURN;
      }
      /* We did finish the row. Undo our little suspension hack if a previous
       * call suspended; then mark the main buffer empty.
       */
      if (jmain->suspended) {
	(*in_row_ctr)++;
	jmain->suspended = FALSE;
      }
    }

    /* If get here, we are done with this iMCU row.  Mark buffer empty. */
    jmain->rowgroup_ctr = 0;
    jmain->cur_iMCU_row++;
  }

  NORMAL_RETURN;
}

#endif /* FULL_MAIN_BUFFER_SUPPORTED */


/*
 * Initialize main buffer controller.
 */

GLOBAL ErrCode CDECL 
jinit_c_main_controller (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    boolean	need_full_buffer)
{
  PROLOGUE;

  my_main_ptr jmain;
  int ci;
  jpeg_component_info *compptr;

  ERR_PASS((*cinfo->mem->alloc_small) (pIPState, (j_common_ptr) cinfo,
				JPOOL_IMAGE,
				SIZEOF(my_main_controller), (void **)&jmain));
  cinfo->main = (struct jpeg_c_main_controller *) jmain;
  jmain->pub.start_pass = start_pass_main;

  /* We don't need to create a buffer in raw-data mode. */
  if (cinfo->raw_data_in)
      NORMAL_RETURN;

  /* Create the buffer.  It holds downsampled data, so each component
   * may be of a different size.
   */
  if (need_full_buffer) {
#ifdef FULL_MAIN_BUFFER_SUPPORTED
    /* Allocate a full-image virtual array for each component */
    /* Note we pad the bottom to a multiple of the iMCU height */
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
	 ci++, compptr++) {
       ERR_PASS((*cinfo->mem->request_virt_sarray)
	(pIPState, (j_common_ptr) cinfo, JPOOL_IMAGE, FALSE,
	 compptr->width_in_blocks * DCTSIZE,
	 (JDIMENSION) jround_up((long) compptr->height_in_blocks,
				(long) compptr->v_samp_factor) * DCTSIZE,
	 (JDIMENSION) (compptr->v_samp_factor * DCTSIZE),
	 &jmain->whole_image[ci]));
    }
#else
    ERREXIT(cinfo, JERR_BAD_BUFFER_MODE);
#endif
  } else {
#ifdef FULL_MAIN_BUFFER_SUPPORTED
    jmain->whole_image[0] = NULL; /* flag for no virtual arrays */
#endif
    /* Allocate a strip buffer for each component */
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
	 ci++, compptr++) {
	ERR_PASS((*cinfo->mem->alloc_sarray)
	(pIPState, (j_common_ptr) cinfo, JPOOL_IMAGE,
	 compptr->width_in_blocks * DCTSIZE,
	 (JDIMENSION) (compptr->v_samp_factor * DCTSIZE),
	 (JSAMPARRAY *)&jmain->buffer[ci]));
    }
  }

  NORMAL_RETURN;
}

/*
$Log: jcmainct.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.5  1998/07/28 19:05:30  aber
jcmainct.c -> /product/ipcore/ipshared/src/RCS/jcmainct.c,v

CommonCheck: Rename variable from main to jmain, as main is somewhate reserved.

Revision 1.4  1998/01/13 23:50:48  ddavies
jcmainct.c -> /product/ipcore/ipshared/src/RCS/jcmainct.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1996/11/06 00:37:03  xzhu
jcmainct.c -> /product/ipcore/ipshared/src/RCS/jcmainct.c,v

CommonCheck: Did some logical changes to let function return success or error. All outputs go through parameters.

Revision 1.2  1996/10/31 01:08:43  xzhu
jcmainct.c -> /product/ipcore/ipshared/src/RCS/jcmainct.c,v

CommonCheck: Since asscee_virt and request_virt prototype were changed, we did same change on calling place.

Revision 1.1  1996/04/20 01:32:08  xzhu
Initial revision

*/
