/*
 * jcmcu.c
 *
 * Copyright (C) 1991, 1992, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains MCU extraction routines and quantization scaling.
 * These routines are invoked via the extract_MCUs and
 * extract_init/term methods.
 */

#include "jpeg.h"
#include "jpeg.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jcmcu.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");

/*
 * If this file is compiled with -DDCT_ERR_STATS, it will reverse-DCT each
 * block and sum the total errors across the whole picture.  This provides
 * a convenient method of using real picture data to test the roundoff error
 * of a DCT algorithm.  DCT_ERR_STATS should *not* be defined for a production
 * compression program, since compression is much slower with it defined.
 * Also note that jrevdct.o must be linked into the compressor when this
 * switch is defined.
 */


/* ZAG[i] is the natural-order position of the i'th element of zigzag order. */

static Int16 ZAG[DCTSIZE2] = {
  0,  1,  8, 16,  9,  2,  3, 10,
 17, 24, 32, 25, 18, 11,  4,  5,
 12, 19, 26, 33, 40, 48, 41, 34,
 27, 20, 13,  6,  7, 14, 21, 28,
 35, 42, 49, 56, 57, 50, 43, 36,
 29, 22, 15, 23, 30, 37, 44, 51,
 58, 59, 52, 45, 38, 31, 39, 46,
 53, 60, 61, 54, 47, 55, 62, 63
};


LOCAL Int32 CDECL
extract_block (JSAMPARRAY input_data, Int32 start_row, Int32 start_col,
           JBLOCK output_data, QUANT_TBL_PTR quanttbl)
/* Extract one 8x8 block from the specified location in the sample array; */
/* perform forward DCT, quantization scaling, and zigzag reordering on it. */
{
  /* This routine is heavily used, so it's worth coding it tightly. */
  UInt32 inBlock[16];
  Int16 block[64];

    UInt32 *src;
    UInt32 *dst;
    UInt8 **inrow;

/*
  Convert the char ptrs to 32 bit ptrs, so that we can move the data
  to the DCT array (block) in 4 byte chunks instead of a byte at a time.
  In order to do this, we need to move the subtraction of 128, which
  was previously done here a byte at a time, into the DCT computation.
  Actually, this turns out to be an advantage, since the first
  thing the DCT does is the 8 point butterfly operation. For the 4 adds
  we can just subtract 256. For the subtracts, the biases cancel
  out so we don't need to do them at all. Neat.
*/
    inrow = input_data + start_row;  /* First element of the ptr array */
    dst = inBlock;

/* Row 0 */
    src = (UInt32 *) (inrow[0] + start_col);
    dst[0] = src[0];  /* First 4 pixels in block row */
    dst[1] = src[1];  /* Last 4 pixels in block row */

/* Row 1 */
    src = (UInt32 *) (inrow[1] + start_col);
    dst[2] = src[0];
    dst[3] = src[1];

/* Row 2 */
    src = (UInt32 *) (inrow[2] + start_col);
    dst[4] = src[0];
    dst[5] = src[1];

/* Row 3 */
    src = (UInt32 *) (inrow[3] + start_col);
    dst[6] = src[0];
    dst[7] = src[1];

/* Row 4 */
    src = (UInt32 *) (inrow[4] + start_col);
    dst[8] = src[0];
    dst[9] = src[1];

/* Row 5 */
    src = (UInt32 *) (inrow[5] + start_col);
    dst[10] = src[0];
    dst[11] = src[1];

/* Row 6 */
    src = (UInt32 *) (inrow[6] + start_col);
    dst[12] = src[0];
    dst[13] = src[1];

/* Row 7 */
    src = (UInt32 *) (inrow[7] + start_col);
    dst[14] = src[0];
    dst[15] = src[1];


    j_fwd_dct( (UInt8 *)inBlock, block);

  { 
    register UInt32 i;
    register Int32 temp;
    register UInt32 qdiv;

    for (i = 0; i < DCTSIZE2; i++) {
        temp = (JCOEF) block[ZAG[i]];
	qdiv = *quanttbl++;

      /* divide by *quanttbl, ensuring proper rounding */
        if (temp < 0) 
	{
            temp = -temp;
            temp += qdiv>>1;
	    /* Avoid div if result would be zero or one (most cases) */
	    if (temp < qdiv) 
		*output_data++ = 0;
	    else if (temp < (qdiv+qdiv))
		*output_data++ = -1;
	    else
		*output_data++ = (Int16) (- (temp/(Int32)qdiv));
        } 
	else 
	{
            temp += qdiv>>1;
	    /* Avoid div if result would be zero or one (most cases) */
	    if (temp < qdiv)
		*output_data++ = 0;
	    else if (temp < (qdiv+qdiv))
		*output_data++ = 1;
	    else
		*output_data++ = (Int16) (temp/(Int32)qdiv);
        }
    }
  }


  return ia_successful;
}


/*
 * Extract samples in MCU order, process & hand off to output_method.
 * The input is always exactly N MCU rows worth of data.
 */

METHODDEF Int32 CDECL
extract_MCUs (compress_info_ptr cinfo,
          JSAMPIMAGE image_data,
          Int32 num_mcu_rows,
          MCU_output_method_ptr output_method)
{
  JBLOCK MCU_data[MAX_BLOCKS_IN_MCU];
  Int32 mcurow;
  Int32 mcuindex;
  Int16 blkn, ci, xpos, ypos;
  jpeg_component_info * compptr;
  QUANT_TBL_PTR quant_ptr;
  UInt32 mcuHeight[4], mcuWidth[4];
  Int32 startRow[4];
  Int32 startCol[4];

/* Get the MCU sizes (in units of DCT blocks) for each component */
  for (ci = 0; ci < cinfo->comps_in_scan; ci++) 
  {
    mcuHeight[ci] = cinfo->cur_comp_info[ci]->MCU_height; 
    mcuWidth[ci] = cinfo->cur_comp_info[ci]->MCU_width; 
    startRow[ci] = 0;
    startCol[ci] = 0;
  }

/* Extract data from the image array, DCT it, and quantize it */
  for (mcurow = 0; mcurow < num_mcu_rows; mcurow++) 
  {
    for (mcuindex = 0; mcuindex < cinfo->MCUs_per_row; mcuindex++) 
    {
      blkn = 0;
      for (ci = 0; ci < cinfo->comps_in_scan; ci++) 
      {
        compptr = cinfo->cur_comp_info[ci];
        quant_ptr = cinfo->quant_tbl_ptrs[compptr->quant_tbl_no];
        for (ypos = 0; ypos < mcuHeight[ci]; ypos++) 
	{
          for (xpos = 0; xpos < mcuWidth[ci]; xpos++) 
	  {
            extract_block(image_data[ci],
	      (startRow[ci] + ypos)*DCTSIZE,
	      (startCol[ci] + xpos)*DCTSIZE,
              MCU_data[blkn], quant_ptr);
            blkn++;
          }
        }
      }

/* Send the MCU whereever the pipeline controller wants it to go */
      (*output_method) (cinfo, MCU_data);

      /* Step to next MCU position in row */
      for (ci = 0; ci < cinfo->comps_in_scan; ci++) 
          startCol[ci] += mcuWidth[ci];
    }

    /* Step to next MCU row position */
    for (ci = 0; ci < cinfo->comps_in_scan; ci++) 
        startRow[ci] += mcuHeight[ci];
  }

  return ia_successful;
}


/*
 * Initialize for processing a scan.
 */

METHODDEF Int32 CDECL
extract_init (compress_info_ptr cinfo)
{
    MENTION(cinfo)
  /* no work for now */

  return ia_successful;
}


/*
 * Clean up after a scan.
 */

METHODDEF Int32 CDECL
extract_term (compress_info_ptr cinfo)
{
    MENTION(cinfo)
  /* no work for now */

  return ia_successful;
}



/*
 * The method selection routine for MCU extraction.
 */

GLOBAL Int32 CDECL
jselcmcu (compress_info_ptr cinfo)
{
  /* just one implementation for now */
  cinfo->cmethods->extract_init = extract_init;
  cinfo->cmethods->extract_MCUs = extract_MCUs;
  cinfo->cmethods->extract_term = extract_term;

  return ia_successful;
}
/*
$Log: jcmcu.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.10  1995/12/13 02:47:02  ejaquith
jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.9  1995/10/18  03:59:36  ddavies
 * jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.8  1994/10/21  00:49:20  lperry
 * jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v
 *
 * Get rid of UINT32, etc., so these files can work with Peter's interpreter.
 *
 * Revision 1.7  1994/08/01  21:06:50  ddavies
 * jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v
 *
 * Change "static const" to "static" to work around buggy gcc/Purify interaction.
 *
 * Revision 1.6  1994/07/21  00:15:52  lperry
 * jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v
 *
 * Change error handling from old setjmp/longjmp to function return codes.
 *
 * Revision 1.5  1994/06/27  16:46:56  lperry
 * jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v
 *
 * Moved static variables to the "cinfo" structure to achieve reentrancy.
 *
 * Revision 1.4  1994/06/11  00:19:00  lperry
 * lperry on Fri Jun 10 17:16:12 PDT 1994
 *
 * jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v
 *
 * Change extract_block routine to move data from user's
 * buffer to JPEG DCT buffer in 32 bit chunks instead of 8.
 * Also eliminate gray level bias subtraction here and put
 * it into DC term of DCT function.
 *
 * Revision 1.3  1994/06/09  23:14:03  lperry
 * lperry on Thu Jun 9 16:12:08 PDT 1994
 *
 * jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v
 *
 * Added another check for quantization which would yield a
 * result of +1 or -1, also very frequent cases. Avoid division
 * for these cases as well. Small gain on Sun, no change on PC.
 *
 * Revision 1.2  1994/06/06  18:51:03  lperry
 * lperry on Mon Jun 6 11:46:40 PDT 1994
 *
 * jcmcu.c -> /product/ipcore/ipshared/src/RCS/jcmcu.c,v
 *
 * Change code which performs quantization division so that its
 * tests if the result will be zero before doing the divide. Since
 * as many as three quarters of the quantized coefficients will be
 * zero, this saves a lot of divides. Saves 10-20% of time, higher
 * compression ratios giving more savings.
 *
 * Revision 1.1  1994/02/27  02:20:35  lperry
 * Initial revision
 *
*/
