/*
 * jcomapi.c
 *
 * Copyright (C) 1994, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains application interface routines that are used for both
 * compression and decompression.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jcomapi.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");

/*
 * Abort processing of a JPEG compression or decompression operation,
 * but don't destroy the object itself.
 *
 * For this, we merely clean up all the nonpermanent memory pools.
 * Note that temp files (virtual arrays) are not allowed to belong to
 * the permanent pool, so we will be able to close all temp files here.
 * Closing a data source or destination, if necessary, is the application's
 * responsibility.
 */

GLOBAL ErrCode CDECL 
jpeg_abort (
    IPState    *pIPState,
    j_common_ptr cinfo)
{
  PROLOGUE;

  int pool;

  /* Releasing pools in reverse order might help avoid fragmentation
   * with some (brain-damaged) malloc libraries.
   */
  for (pool = JPOOL_NUMPOOLS-1; pool > JPOOL_PERMANENT; pool--) {
    ERR_PASS((*cinfo->mem->free_pool) (pIPState, cinfo, pool));
  }

  /* Reset overall state for possible reuse of object */
  cinfo->global_state = (cinfo->is_decompressor ? DSTATE_START : CSTATE_START);

  NORMAL_RETURN;
}


/*
 * Destruction of a JPEG object.
 *
 * Everything gets deallocated except the master jpeg_compress_struct itself
 * and the error manager struct.  Both of these are supplied by the application
 * and must be freed, if necessary, by the application.  (Often they are on
 * the stack and so don't need to be freed anyway.)
 * Closing a data source or destination, if necessary, is the application's
 * responsibility.
 */

GLOBAL ErrCode CDECL 
jpeg_destroy (
    IPState    *pIPState,
    j_common_ptr cinfo)
{
  PROLOGUE;

  /* We need only tell the memory manager to release everything. */
  /* NB: mem pointer is NULL if memory mgr failed to initialize. */
  if (cinfo->mem != NULL)
    ERR_PASS((*cinfo->mem->self_destruct) (pIPState, cinfo));
  cinfo->mem = NULL;		/* be safe if jpeg_destroy is called twice */
  cinfo->global_state = 0;	/* mark it destroyed */

  NORMAL_RETURN;
}


/*
 * Convenience routines for allocating quantization and Huffman tables.
 * (Would jutils.c be a more reasonable place to put these?)
 */

GLOBAL ErrCode CDECL 
jpeg_alloc_quant_table (
    IPState    *pIPState,
    j_common_ptr cinfo,
    JQUANT_TBL **tbl)
{
  PROLOGUE;  

  ERR_PASS((*cinfo->mem->alloc_small) (pIPState, cinfo, JPOOL_PERMANENT, 
				SIZEOF(JQUANT_TBL), (void **)tbl));
  (*tbl)->sent_table = FALSE;   
                             /* make sure this is false in any new table */

  NORMAL_RETURN;
}


GLOBAL ErrCode CDECL 
jpeg_alloc_huff_table (
    IPState    *pIPState,
    j_common_ptr cinfo,
    JHUFF_TBL **tbl)
{
  PROLOGUE;  

  ERR_PASS((*cinfo->mem->alloc_small) (pIPState, cinfo, JPOOL_PERMANENT, 
				SIZEOF(JHUFF_TBL), (void **)tbl));
  (*tbl)->sent_table = FALSE;  /* make sure this is false in any new table */

  NORMAL_RETURN;
}

/*
$Log: jcomapi.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.2  1998/01/13 23:50:53  ddavies
jcomapi.c -> /product/ipcore/ipshared/src/RCS/jcomapi.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.1  1996/04/20 01:30:35  xzhu
Initial revision

*/
