/***************************************************************
    Copyright (c) 1994, Xerox Corporation.  All rights reserved. 
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
***************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "types.pub"
#include "memory.pub"

#include "jpeg.h"
#include "jpeg.pub"
#include "jpeg.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jcomprs.c,v $; $Revision: 1.2 $; $Date: 2000/06/27 00:09:38 $");

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

Int32 CDECL
jcInit(CompressInfo *cinfo,
         Int32 imageWidth,
         Int32 imageHeight,
         Int32 numComponents,
         Int32 quality,
         ErrCode (CDECL *WriteCallback) (CompressInfo * cinfo,
                                 UInt8 *JPEGInternalBuffer,
                                 UInt32 nBytes, UInt32 *ncount),
         void *userData,
         void (CDECL *jcSpecialOrders)(CompressInfo * cinfo)
         )
{
    Int32 ci, fullsize_width;

/* Assign image parameters - width, height, color/grayscale */
    cinfo->image_width = imageWidth;
    cinfo->image_height = imageHeight;
    cinfo->input_components = (Int16) numComponents;

    if (numComponents == 1)
    {
        cinfo->in_color_space = cCS_GRAYSCALE;
        cinfo->rowsPerMCU = DCTSIZE;
    }
    else if (numComponents == 3)
    {
        cinfo->in_color_space = cCS_RGB;
        cinfo->rowsPerMCU = DCTSIZE * 2;
    }
    else
        return JERR_TOOMANYCOMPS;

/* Hardwire for 8 bit input */
    cinfo->data_precision = 8;

    cinfo->cur_mcu_row = 0;    /* Initialize MCU row counter to zero */
    cinfo->mcu_rows_output = 0;

    cinfo->whichss = 1;    /* Start sampling with buffer[0] */

/* select std memory alloc routines */
    jselmemmgr(cinfo);    

/* Allocate memory for method listing structures */
/* Initialize the system-dependent method pointers. */
    cinfo->cmethods = (struct Compress_methods_struct *) 
                JALLOC( sizeof(struct Compress_methods_struct) );
    CHECKMEM(cinfo->cmethods)

    cinfo->emethods = (struct External_methods_struct *) 
                JALLOC( sizeof(struct External_methods_struct) );
    CHECKMEM(cinfo->emethods)

/* select std error/trace msg routines */
    jselerror(cinfo);    

/* Register the callback procedure and user buffers */
    cinfo->writeCallback = WriteCallback;
    cinfo->userData = userData;

/*
  Here, set up pointers to your own routines for input data handling
  and post-init parameter selection.
*/

/* 
  Set up default JPEG parameters in the cinfo data structure. 
  The TRUE flag sets force_baseline to true. This will produce
  a JPEG file with no quantization divisors greater than 255.
  The effect of this doesn't kick in until the quality factor 
  is less than 25.
*/
    j_c_defaults(cinfo, quality, TRUE);

/* If the input is gray scale, generate a monochrome JPEG file. */
    if (cinfo->in_color_space == cCS_GRAYSCALE)
        j_monochrome_default(cinfo);

/* For now, always select JFIF output format. */
    jselwjfif(cinfo);

/*
   Allow the UI to set any special options. This routine can be
   NULL or a stub if you don't want to do anthing non-standard
*/
    if (jcSpecialOrders != NULL)
        jcSpecialOrders(cinfo);


/* Now select methods for compression steps. */
    CHECKERR(jc_initial_setup(cinfo))

/* Input image reading method selection is already done. */
/* So is output file header formatting (both are done by user interface). */

/* Gamma and color space conversion */
  CHECKERR(jselccolor(cinfo))

  /* Entropy encoding: either Huffman or arithmetic coding. */
  cinfo->arith_code = FALSE;    /* force Huffman mode */

  jselchuffman(cinfo);

/* Initialize the output file & other modules as needed */
/* (entropy_encoder is inited by pipeline controller) */
    CHECKERR((*cinfo->cmethods->colorin_init) (cinfo))
    CHECKERR((*cinfo->cmethods->write_file_header) (cinfo))

/* Prepare for single scan containing all components */
    if (cinfo->num_components > MAX_COMPS_IN_SCAN)
        ERREXIT(cinfo->emethods, "Too many components for interleaved scan",
        JERR_TOOMANYCOMPS);
    cinfo->comps_in_scan = cinfo->num_components;

    for (ci = 0; ci < cinfo->num_components; ci++) {
        cinfo->cur_comp_info[ci] = &cinfo->comp_info[ci];
    }

    if (cinfo->comps_in_scan == 1) {
        CHECKERR(noninterleaved_scan_setup(cinfo));
    } else {
        CHECKERR(interleaved_scan_setup(cinfo));
    }

/* Compute dimensions of full-size pixel buffers */
/* Note these are the same whether interleaved or not. */
    fullsize_width = jround_up(cinfo->image_width,
                   (Int32) (cinfo->max_h_samp_factor * DCTSIZE));

/* Allocate buffer for sample data before downsampling */
    CHECKERR(jc_alloc_sampling_buffer(cinfo, 
	     cinfo->fullsize_data, fullsize_width))

/* Sampled_data is sample data after downsampling */
    cinfo->sampled_data[0] = (JSAMPIMAGE) 
           JALLOC(cinfo->num_components * SIZEOF(JSAMPARRAY));
    CHECKMEM(cinfo->sampled_data[0])

    for (ci = 0; ci < cinfo->num_components; ci++) {
        cinfo->sampled_data[0][ci] = 
        jalloc_sarray (cinfo, cinfo->comp_info[ci].downsampled_width,
             (Int32) (cinfo->comp_info[ci].v_samp_factor * DCTSIZE));
        CHECKMEM(cinfo->sampled_data[0][ci])
    }

/* Initialize output file & do per-scan object init */

    CHECKERR((*cinfo->cmethods->write_scan_header) (cinfo))
    cinfo->cmethods->entropy_output = cinfo->cmethods->write_jpeg_data;
    CHECKERR((*cinfo->cmethods->entropy_encode_init) (cinfo))
    CHECKERR((*cinfo->cmethods->downsample_init) (cinfo))
    CHECKERR((*cinfo->cmethods->extract_init) (cinfo))

    return ia_successful;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

Int32 CDECL
jcMCURow(compress_info_ptr cinfo,
                JSAMPIMAGE inbuf) 
{

    int rows_in_mem;        /* # of sample rows in full-size buffers */
    int rows_this_time;        /* # of sample rows this time through */
    Int32 fullsize_width;    /* # of samples per row in full-size buffers */
    int mcu_rows_per_loop;    /* # of MCU rows processed per outer loop */

    Int16           i;

    rows_in_mem = cinfo->max_v_samp_factor * DCTSIZE;
    rows_this_time = MIN(cinfo->rowsPerMCU, cinfo->image_height 
                         - cinfo->cur_mcu_row * cinfo->rowsPerMCU);

    if (cinfo->comps_in_scan == 1) {
        /* Vk block rows constitute the same number of MCU rows */
        mcu_rows_per_loop = cinfo->cur_comp_info[0]->v_samp_factor;
    } else {
        /* in an interleaved scan, one MCU row contains Vk block rows */
        mcu_rows_per_loop = 1;
    }

/* Compute dimensions of full-size pixel buffers */
/* Note these are the same whether interleaved or not. */
    fullsize_width = jround_up(cinfo->image_width,
                   (Int32) (cinfo->max_h_samp_factor * DCTSIZE));

    cinfo->whichss ^= 1;    

/*
 * Obtain rows_this_time pixel rows and expand to rows_in_mem rows
   Then we have exactly DCTSIZE row groups for downsampling.
 */
    CHECKERR((*cinfo->cmethods->get_sample_rows) 
            (cinfo, rows_this_time,
             inbuf, cinfo->fullsize_data[cinfo->whichss]))

    CHECKERR((*cinfo->cmethods->edge_expand) 
                 (cinfo, cinfo->image_width, rows_this_time,
                    fullsize_width, rows_in_mem,
                    cinfo->fullsize_data[cinfo->whichss]))

/* Downsample the data (all components) */
/* First time through is a special case */
    if (cinfo->cur_mcu_row == 0) { /* At the image top */

        /* Downsample first row group with dummy above-context */
        downsample(cinfo, cinfo->fullsize_data[cinfo->whichss], 
                   cinfo->sampled_data[0], fullsize_width,
                   (Int16) (-1), (Int16) 0, (Int16) 1,
                   (Int16) 0);

        /* Downsample second through next-to-last row groups of current set */
        for (i = 1; i <= DCTSIZE - 2; i++) {
            downsample(cinfo, cinfo->fullsize_data[cinfo->whichss], 
                       cinfo->sampled_data[0], fullsize_width,
                       (Int16) (i - 1), (Int16) i, (Int16) (i + 1),
                       (Int16) i);
        }
    }


    else {  /* The normal case in the image interior */

        /* Downsample last row group of previous set */
        downsample(cinfo, cinfo->fullsize_data[cinfo->whichss], 
                   cinfo->sampled_data[0], fullsize_width,
                   (Int16) DCTSIZE, (Int16) (DCTSIZE + 1), (Int16) 0,
                    (Int16) (DCTSIZE - 1));

        /* and dump the previous set's downsampled data */
        CHECKERR((*cinfo->cmethods->extract_MCUs) 
                (cinfo, cinfo->sampled_data[0],
                mcu_rows_per_loop,
                cinfo->cmethods->entropy_encode))

        cinfo->mcu_rows_output += mcu_rows_per_loop;

        /* Downsample first row group of this set */
        downsample(cinfo, cinfo->fullsize_data[cinfo->whichss], 
                   cinfo->sampled_data[0], fullsize_width,
                   (Int16) (DCTSIZE + 1), (Int16) 0, (Int16) 1,
                   (Int16) 0);

        /* Downsample second through next-to-last row groups of this set */
        for (i = 1; i <= DCTSIZE - 2; i++) {
            downsample(cinfo, cinfo->fullsize_data[cinfo->whichss], 
                       cinfo->sampled_data[0], fullsize_width,
                       (Int16) (i - 1), (Int16) i, (Int16) (i + 1),
                       (Int16) i);
        }


    } /* End normal case */

   /* At the image bottom only */
   if (cinfo->cur_mcu_row == cinfo->MCU_rows_in_scan - 1) { 

        /* Downsample the last row group with dummy below-context */
        /* Note whichss points to last buffer side used */
        downsample(cinfo, cinfo->fullsize_data[cinfo->whichss], 
                   cinfo->sampled_data[0], fullsize_width,
                   (Int16) (DCTSIZE - 2), (Int16) (DCTSIZE - 1), 
                   (Int16) (-1), (Int16) (DCTSIZE - 1));
        /*
         * Dump the remaining data (may be less than full height if
         * uninterleaved)
         */
        CHECKERR((*cinfo->cmethods->extract_MCUs) 
                (cinfo, cinfo->sampled_data[0],
                (int) (cinfo->MCU_rows_in_scan - cinfo->mcu_rows_output),
                cinfo->cmethods->entropy_encode))
    }

    cinfo->cur_mcu_row++;

    return ia_successful;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

Int32 CDECL
jcTerm(compress_info_ptr cinfo)
{

    CHECKERR((*cinfo->cmethods->extract_term) (cinfo))
    CHECKERR((*cinfo->cmethods->downsample_term) (cinfo))
    CHECKERR((*cinfo->cmethods->entropy_encode_term) (cinfo))
    CHECKERR((*cinfo->cmethods->write_scan_trailer) (cinfo))

    /* Finish output file, release working storage, etc */
    CHECKERR((*cinfo->cmethods->write_file_trailer) (cinfo))
    CHECKERR((*cinfo->cmethods->colorin_term) (cinfo))

    return jfree_all(cinfo);
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/



Int32 CDECL
jcGetImageLinesPerMCU(CompressInfo *cinfo)
{
    return (cinfo->max_v_samp_factor * DCTSIZE);
}

Int32 CDECL
jcGetMCURowsPerImage(CompressInfo *cinfo)
{
    return (cinfo->MCU_rows_in_scan);
}

static Int32 CDECL
jcGetLastMCURowSize(CompressInfo *cinfo)
{
    return 
    ( cinfo->image_height % jcGetImageLinesPerMCU(cinfo) == 0 ) ? 
        jcGetImageLinesPerMCU(cinfo) : 
        cinfo->image_height % jcGetImageLinesPerMCU(cinfo);
}

Int32 CDECL
jcGetLinesToRead(CompressInfo *cinfo)
{
    return
      (cinfo->cur_mcu_row == (jcGetMCURowsPerImage(cinfo)-1)) ? 
       jcGetLastMCURowSize(cinfo): jcGetImageLinesPerMCU(cinfo);
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
  Allocate an array of pointers to line buffers for each channel
  If the allocLineMem flage is TRUE, also allocate memory for
  the line buffers themselves
*/
JSAMPIMAGE CDECL
jcAllocMCURowInputBuffer(CompressInfo *cinfo, Bool allocLineMem)
{

    UInt32 band,row, nlines, nsamps, nbands;
    JSAMPIMAGE buf;

    nlines = jcGetImageLinesPerMCU(cinfo);
    nsamps = cinfo->image_width;
    nbands = cinfo->input_components;

/* Allocate the channel pointers */
    buf = (JSAMPIMAGE) JALLOC(nbands*sizeof(JSAMPARRAY));
    PCHECKMEM(buf)

    for (band=0; band<nbands; band++)
    {
        /* Allocate the row pointers */
        buf[band] = (JSAMPARRAY) JALLOC(nlines * sizeof(JSAMPROW));
        PCHECKMEM(buf[band])

        for (row=0; row<nlines; row++)
        {
            if (allocLineMem) /* Alloc row mem in 32 bit chunks */
            {
                buf[band][row] = (JSAMPROW) JALLOC(((nsamps+3)/4)
                                                    * sizeof(Int32));
                PCHECKMEM(buf[band][row])
            }
            else   /* Just set the row pointers to NULL */
                buf[band][row] = (JSAMPROW) 0;
        }
 
    }

    return buf;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
$Log: jcomprs.c,v $
Revision 1.2  2000/06/27 00:09:38  jlin
change callback

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.16  1997/07/11 01:49:21  ddavies
jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.15  1995/12/13 02:47:05  ejaquith
jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.14  1995/10/18  03:59:41  ddavies
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.13  1995/08/30  20:56:49  ddavies
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Change names of COLOR_SPACE vars from CS_foo to cCS_foo since they're all
 * constants now.
 *
 * Revision 1.12  1994/11/22  19:08:25  lperry
 * lperry on Tue Nov 22 11:05:55 PST 1994
 *
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Changed the force_baseline parameter to j_c_defaults()
 * to be TRUE, instead of false. This has the effect of
 * limiting quantizer values t0 8 bits, which is as much
 * as is supported by "baseline" jpeg implementations.
 * We shouldn't be creating files which might not be able to be
 *
 * Revision 1.11  1994/11/02  01:25:17  ejaquith
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Added inclusion of "types.pub" as memory.pub now no longer includes it.
 *
 * Revision 1.10  1994/10/21  00:49:25  lperry
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Get rid of UINT32, etc., so these files can work with Peter's interpreter.
 *
 * Revision 1.9  1994/07/21  00:16:01  lperry
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Change error handling from old setjmp/longjmp to function return codes.
 *
 * Revision 1.8  1994/07/02  01:37:45  lperry
 * lperry on Fri Jul 1 18:35:07 PDT 1994
 *
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Fixed small memory leak (200 bytes) caused by allocating
 * two structures before the memory manager's linked list
 * was initialized. Since all memory allocations are kept on
 * this list, the first could never be freed.
 *
 * Revision 1.7  1994/06/27  16:47:07  lperry
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Moved static variables to the "cinfo" structure to achieve reentrancy.
 *
 * Revision 1.6  1994/05/24  22:11:24  lperry
 * lperry on Tue May 24 15:10:33 PDT 1994
 *
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Changed interface to jcInit to match that of jdInit. cinfo
 * structure is passed by reference.
 *
 * Revision 1.5  1994/05/24  00:35:26  lperry
 * lperry on Mon May 23 17:32:23 PDT 1994
 *
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Change JPEG buffer alloc routine to allow allocation
 * of either real line buffers or a pointer array
 * This allows data to be read directly from the user's
 * memory buffer when no conversions need to be made.
 *
 * Revision 1.4  1994/04/08  00:02:20  lperry
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Used the JPEG memory allocator exclusively. This lets
 * all memory be freed with a single call.
 * Allocation of the input buffer now creates a set of line
 * pointers, so we can either point at the input data or
 * actually allocate intermediate buffers if desired.
 *
 * Revision 1.3  1994/03/26  00:55:46  ddavies
 * ddavies on Fri Mar 25 16:54:28 PST 1994
 *
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Add extra, superfluous, parens around "if (status = setjum(...))" statements to make gcc happy.
 *
 * Revision 1.2  1994/03/24  20:58:37  lperry
 * lperry on Thu Mar 24 12:57:25 PST 1994
 *
 * jcomprs.c -> /product/ipcore/ipshared/src/RCS/jcomprs.c,v
 *
 * Changed interface for jcInit to allow a procedure to
 * be passed for setting non-standard options
 *
 * Revision 1.1  1994/02/27  02:20:41  lperry
 * Initial revision
 *
*/
