/*********************************************************************
    Copyright (c) 1995-2000, Xerox Corporation.  All rights reserved. 
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
**********************************************************************/

#include "jinclude.h"
#include "jpeg.pub"

IP_RCSINFO(jcxapi_c_RCSInfo, "$RCSfile: jcxapi.c,v $; $Revision: 1.9 $; $Date: 2001/07/06 18:17:42 $");

/*
  After jc_init or jd_init, this state only exists after an abort.
  Checking for this in every API call (except the inits and error handlers) 
  will prevent segmentation violation / GPF errors if the caller (app)
  neglects to check the return codes.
*/
#define JCHECK_STATE \
if (cinfo->global_state == 0) \
{ \
    ERREXIT(cinfo, JERR_BAD_CINFO); \
}


/*================================================================*/
/*================== COMPRESSION API =============================*/
/*================================================================*/

/* jc_extract_tables() */

/*********************************************************************
*
*   Extract the tables, then avoid using them in subsequent compresses
*
*   Description:
*       Output the set of Q-tables and Huffman tables, to be removed
*	from later compresses.
*       Quality is on a scale of 0 to 100, where higher is better.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       quality:
*           Image quality setting (0 to 100)
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is a OPTIONAL call for compression. This is a mechanism
*       to use the JPEGTables feature for TIFF, as documented in TTN2.
*
*********************************************************************/
ErrCode CDECL 
jc_extract_tables (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	quality)
{
    PROLOGUE;

    JCHECK_STATE;

    if (quality < 0 || quality > 100)
    {
        ERREXIT(cinfo, JERR_BAD_QUALITY);
    }
	/* Force baseline JPEG */
    ERR_PASS(jpeg_set_quality (pIPState, cinfo, quality, TRUE));

/* This first pair of functions is how libtiff does it, yet the second seems to do the job */
#if 0
	/* Unsuppress Q-tables */
    ERR_PASS(unsupress_quant_table (pIPState, cinfo, quality, TRUE));

	/* Unsuppress Huffman tables */
    ERR_PASS(unsupress_huff_table (pIPState, cinfo, quality, TRUE));
#else
	/* Unsuppress ALL tables */
    ERR_PASS(jpeg_suppress_tables (pIPState, cinfo, FALSE));

#endif
	/* Generate the tables stream */
    ERR_PASS(jpeg_write_tables (pIPState, cinfo));

	/* Mark tables as separated */
    cinfo->separate_tables = TRUE;

    /* Turn off entropy optimizations (don't know why, libtiff does this)
	 */
    cinfo->optimize_coding = FALSE;

    NORMAL_RETURN;
}


/* jc_set_width() */

/*********************************************************************
*
*   Set the image width parameter for a compression run
*
*   Description:
*       Set the image width parameter for a compression run
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       width:
*           Image width in pixels
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is a MANDATORY call for compression.
*
*********************************************************************/
ErrCode CDECL 
jc_set_width (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	width)
{
    PROLOGUE;

    JCHECK_STATE;

    cinfo->image_width = width;

    NORMAL_RETURN;
}


/* jc_set_height() */

/*********************************************************************
*
*   Set the image height parameter for a compression run
*
*   Description:
*       Set the image height parameter for a compression run
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       height:
*           Image height in lines
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is a MANDATORY call for compression.
*
*********************************************************************/
ErrCode CDECL 
jc_set_height (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	height)
{
    PROLOGUE;

    JCHECK_STATE;

    cinfo->image_height = height;

    NORMAL_RETURN;
}


/* jc_set_components() */

/*********************************************************************
*
*   Set the number of image color bands for a compression run
*
*   Description:
*       Set the number of image components for a compression run.
*       For grayscale images, this is always one. For color images,
*       it will be 3 for RGB images, and 4 for CMYK images.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       num_components:
*           Number of image components. This will be one
*           for grayscale images, 3 for RGB and 4 for CMYK.
*
*   Returns:
*       ia_successful: On success
*       JERR_BAD_IN_COLORSPACE: On invalid component count.
*
*   Notes:
*       This is a MANDATORY call for compression.
*
*********************************************************************/
ErrCode CDECL 
jc_set_components (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	num_components)
{
    PROLOGUE;

    JCHECK_STATE;

    cinfo->input_components = num_components;
    switch (num_components)
    {
        case 1:
            cinfo->in_color_space = JCS_GRAYSCALE;
            break;

        case 3:
            cinfo->in_color_space = JCS_RGB;
            break;

        case 4:
            cinfo->in_color_space = JCS_CMYK;
            break;

        default:
            ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
            break;
    }

    NORMAL_RETURN;

}


/* jc_set_quality() */

/*********************************************************************
*
*   Set the image quality parameter for a compression run
*
*   Description:
*       Set the image quality parameter for a compression run.
*       Quality is on a scale of 0 to 100, where higher is better.
*       A setting of 50 replicates the sample quantization matrices
*       given in the JPEG standard. 
*      
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       quality:
*           Image quality setting (0 to 100)
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is a OPTIONAL call for compression. The API
*       for jc_init() requires a quality parameter, so
*       that quality is explicitly set. jc_set quality()
*       can be used to override this value.
*
*********************************************************************/
ErrCode CDECL 
jc_set_quality (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	quality)
{
    PROLOGUE;

    JCHECK_STATE;

    if (quality < 0 || quality > 100)
    {
        ERREXIT(cinfo, JERR_BAD_QUALITY);
    }
	/* Force baseline JPEG */
    ERR_PASS(jpeg_set_quality (pIPState, cinfo, quality, TRUE));

    NORMAL_RETURN;
}


/* jc_set_sampling_factors() */

/*********************************************************************
*
*   Set the component sampling factors for compression
*
*   Description:
*       Set the component sampling factors. This allows
*       one to vary the sampling ratios of the chrominance
*       components relative to the luminance component.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       Y_horiz:
*           Horizontal samping factor for luminance.
*           (always 1 for grayscale)
*       Y_vert:
*           Vertical samping factor for luminance.
*           (always 1 for grayscale)
*       Cb_horiz:
*           Horizontal samping factor for first chroma component.
*       Cb_vert:
*           Vertical samping factor for first chroma component.
*       Cr_horiz:
*           Horizontal samping factor for second chroma component.
*       Cr_vert:
*           Vertical samping factor for second chroma component.
*
*   Returns:
*       ia_successful:
*
*
*   Notes:
*       This is an OPTIONAL call for compression.
*       Sampling factors are defined in a very non-intuitive way.
*       They represent the ratio of the sampling frequency of a 
*       component relative to the coarsest sampling frequency for
*       all components. For a single component image (grayscale),
*       the normal sampling factors would be 1 in each axis. It would be
*       possible, but highly unusual, to vary this. Color images
*       typically have sampling factors of 2,2 : 1,1 : 1,1 (often 
*       called 4:1:1 sampling, or 2,1 : 1,1 : 1,1 (often called 4:2:2 
*       sampling). The resulting number of DCT blocks in
*       an MCU must not exceed 10. See the JPEG documentation
*       for a more detailed explanation of sampling.
*       
*
*********************************************************************/
ErrCode CDECL 
jc_set_sampling_factors (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	Y_horiz,
    Int32	Y_vert,
    Int32	Cb_horiz,
    Int32	Cb_vert,
    Int32	Cr_horiz,
    Int32	Cr_vert)
{
    PROLOGUE;

    JCHECK_STATE;

    /* Luminance (Y) */
    cinfo->comp_info[0].h_samp_factor = Y_horiz;
    cinfo->comp_info[0].v_samp_factor = Y_vert;

    /* First chroma (Cb) */
    cinfo->comp_info[1].h_samp_factor = Cb_horiz;
    cinfo->comp_info[1].v_samp_factor = Cb_vert;

    /* Second chroma (Cr) */
    cinfo->comp_info[2].h_samp_factor = Cr_horiz;
    cinfo->comp_info[2].v_samp_factor = Cr_vert;

    NORMAL_RETURN;
}


/* jc_set_resolution() */

/*********************************************************************
*
*   Set the image resolution parameter for a compression run
*
*   Description:
*       Set the image resolution parameter for a compression run
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       xDPI:
*           Image resolution in dots/inch along each scanline.
*       yDPI:
*           Image resolution in dots/inch across scanlines.
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is a OPTIONAL call for compression. It has no real
*       effect on the compresion process other than to record
*       the scaning resolution in the JFIF APP0 marker, so that
*       decompression apps can read it. In the absence of this call,
*       The x,y density (resolution) fields are both set to 1, and the 
*       density unit field is set to zero, indicating
*       an aspect ratio of 1:1.
*
*********************************************************************/
ErrCode CDECL 
jc_set_resolution (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	xDPI,
    Int32	yDPI)
{
    PROLOGUE;

    JCHECK_STATE;

    cinfo->density_unit = 1; /* Units are DPI */
    cinfo->X_density = (UInt16) xDPI;
    cinfo->Y_density = (UInt16) yDPI;

    NORMAL_RETURN;
}


/* ZIG[i] is the zigzag-order position of the i'th element of a DCT block */
/* read in natural order (left to right, top to bottom). */
static const int ZIG[DCTSIZE2] = {
    0,  1,  5,  6, 14, 15, 27, 28,
    2,  4,  7, 13, 16, 26, 29, 42,
    3,  8, 12, 17, 25, 30, 41, 43,
    9, 11, 18, 24, 31, 40, 44, 53,
   10, 19, 23, 32, 39, 45, 52, 54,
   20, 22, 33, 38, 46, 51, 55, 60,
   21, 34, 37, 47, 50, 56, 59, 61,
   35, 36, 48, 49, 57, 58, 62, 63
};


/* jc_set_luminance_qtable() */

/*********************************************************************
*
*   Establish a new luminance quantization table.
*
*   Description:
*       Establish a new quantization table for the luminance component,
*       rather than using the default Q table with scaling. This 
*       can be used for special purpose effects or to replicate
*       specific tables to match other compressors.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       qtable:
*           Pointer to a quantization table in natural order.
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is an OPTIONAL call for compression. An example of its
*       use would be to replicate the Q tables used by Adobe in
*       their Postscript DCT filters or Photoshop. Note that the 
*       Q table is supplied in natural order, NOT zigzag order. This
*       call shuffles the data into zigzag order itself, before
*       supplying the table data to the compressor.
*
*********************************************************************/
ErrCode CDECL 
jc_set_luminance_qtable (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32       quality,
    UInt16     *qtable)
{
    PROLOGUE;
    Int32 i, scale;
    unsigned int basic_table[64];
    int which_tbl = 0; /* Luma slot = 0 */

    JCHECK_STATE;

    scale = jpeg_quality_scaling(pIPState, quality);

    for (i=0; i<DCTSIZE2; i++)
    {
	basic_table[ZIG[i]] = qtable[i];
    }
    
    ERR_PASS(jpeg_add_quant_table (pIPState, cinfo, which_tbl,
				   basic_table, scale, FALSE));

    NORMAL_RETURN;
}


/* jc_set_chrominance_qtable() */

/*********************************************************************
*
*   Establish a new chrominance quantization table.
*
*   Description:
*       Establish a new quantization table for the chrominance component,
*       rather than using the default Q table with scaling. This 
*       can be used for special purpose effects or to replicate
*       specific tables to match other compressors.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       quality:
*           Image quality setting (0 to 100)
*       qtable:
*           Pointer to a quantization table in natural order.
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is an OPTIONAL call for compression. An example of its
*       use would be to replicate the Q tables used by Adobe in
*       their Postscript DCT filters or Photoshop. Note that the 
*       Q table is supplied in natural order, NOT zigzag order. This
*       call shuffles the data into zigzag order itself, before
*       supplying the table data to the compressor.
*
*   Bugs:
*       This call loads only one Q table for use by BOTH chroma
*       components. If one wanted to use different quantization for
*       each chroma component, a separate routine would be needed
*       to load it into a different table slot and associate it with
*       the proper component. In as much as ANY usage of these 
*       routines is unlikely, further complicating the interface
*       didn't seem appropriate.
*
*********************************************************************/
ErrCode CDECL 
jc_set_chrominance_qtable (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32       quality,
    UInt16     *qtable)
{
    PROLOGUE;
    Int32 i, scale;
    unsigned int basic_table[64];
    int which_tbl = 1; /* Chroma slot (this is used for both Cb and Cr *) */

    JCHECK_STATE;

    scale = jpeg_quality_scaling(pIPState, quality);

    for (i=0; i<DCTSIZE2; i++)
    {
	basic_table[ZIG[i]] = qtable[i];
    }
    
    ERR_PASS(jpeg_add_quant_table (pIPState, cinfo, which_tbl,
				   basic_table, scale, FALSE));

    NORMAL_RETURN;
}


/* jc_get_error_string() */

/*********************************************************************
*
*   Retrieve a printable error string.
*
*   Description:
*       Allows an application to retrieve a string containing
*       an error string for display in an error message box
*       or other such purpose.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       string:
*           Pointer to a string variable to receive the error string.
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is an OPTIONAL call for compression. Error strings are
*       listed in the file jerror.h . Since they are all in one
*       place, this is convenient for internationalization.
*
*********************************************************************/
ErrCode CDECL 
jc_get_error_string (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    char      **string)
{
    PROLOGUE;

    /* 
      Dont check global_state here, since we're most 
      likely in an abort condition 
    */

    *string = ((j_common_ptr)cinfo)->err->error_string;

    NORMAL_RETURN;
}

/* jc_free_error_string() */

/*********************************************************************
*
*   Free error struct.
*
*   Description:
*       Error struct is allocated in init, but it will not be freed in
*       error exit since the error string will be retrived after error 
*       appear. So use this function to free error struct when error 
*       exit happen.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*
*********************************************************************/
ErrCode CDECL 
jc_free_error_string (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
    PROLOGUE;

    /* 
      Dont check global_state here, since we're most 
      likely in an abort condition 
    */

    if (cinfo->err != NULL)
        IP_FREE(cinfo->err);

    NORMAL_RETURN;
}

/*********************************************************************
*
*   Get current MCU lines.
*
*   Description:
*       Fix MCU lines are samp_fact * DCTSIZE. At the bottom of
*       image (last MCU) the lines are remaining lines.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       nlines:
*           Pointer to a nlines.
*
*   Returns:
*       ia_successful:
*
*
*********************************************************************/
ErrCode CDECL 
jc_get_linesToRead (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32      *nlines)
{
    PROLOGUE;
    Int32 rows, MCURows;

    if (cinfo->global_state == CSTATE_START)
	rows = DCTSIZE << 1;
    else
    {
	MCURows = cinfo->max_v_samp_factor * DCTSIZE;
	rows = cinfo->image_height - cinfo->next_scanline;
	if( rows >  MCURows) rows = MCURows;
    }
    *nlines = rows;

    NORMAL_RETURN;
}

/* jc_init() */

/*********************************************************************
*
*   Initialize for a compression run
*
*   Description:
*       Establish mandatory parameters for a compression run and
*       establish default settings.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       imageWidth:
*           Image width in pixels
*       imageHeight:
*           Image height in scanlines
*       num_components:
*           Number of components (bands) in the image.
*       quality:
*           Quality scaling factor (0 to 100).
*       progressive:
*           True: setup progressive JPEG.
*       XDpi:
*           Horizontal resolution if >0
*       YDpi:
*           Vertical resolution if >0
*       writeCallback:
*           User-supplied callback function to write data to the
*           compressed JPEG file. The compressor will call this
*           function whenever it needs to output data to the
*           compressed file.
*       userState:
*           A pointer to some user state information, for
*           use by the writeCallback function. 
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is an MANDATORY call for compression. 
*       The writeCallback function is patterned after
*       a UNIX write() function. The userState function
*       provides a mechanism for the user to pass in file
*       specific information such as a FILE descriptor.
*
*********************************************************************/
ErrCode CDECL 
jc_init (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	imageWidth,
    Int32	imageHeight,
    Int32	num_components,
    Int32	quality,
    Bool	progressive,
    Int32	XDpi,
    Int32	YDpi,
    jIOCallbackFunc writeCallback,
    void       *userState)
{
    PROLOGUE;

/* Set up the error manager (such as it is) */
    cinfo->err = (struct jpeg_error_mgr *) 
                 IP_MALLOC(sizeof(struct jpeg_error_mgr));
    if (cinfo->err == NULL)
    {
        ERREXIT(cinfo, JERR_ERR_MGR_ALLOC);
    }

    ERR_PASS(jpeg_std_error (pIPState, (j_common_ptr)cinfo, cinfo->err));

    ERR_PASS(jpeg_create_compress (pIPState, cinfo));
    
    ERR_PASS(jpeg_set_destination_mgr (pIPState, cinfo,
				       userState, writeCallback));

/* Mandatory parameters - there can be no defaults for these */
    ERR_PASS(jc_set_width (pIPState, cinfo, imageWidth));
    ERR_PASS(jc_set_height (pIPState, cinfo, imageHeight));
    ERR_PASS(jc_set_components (pIPState, cinfo, num_components));

/* Set defaults - can override later */
    ERR_PASS(jpeg_set_defaults (pIPState, cinfo));

/* Explicitly set quality to (possibly) override default of 75 */
    ERR_PASS(jc_set_quality (pIPState, cinfo, quality));

/*  Set resolution to (possibly) override default 'unknow' */
    if( XDpi > 6 && YDpi > 6 )
	ERR_PASS(jc_set_resolution (pIPState, cinfo, XDpi, YDpi));

/* If select progressive jpeg compression, setup progression */
    if( progressive )
	ERR_PASS( jpeg_simple_progression (pIPState,  cinfo ));

    NORMAL_RETURN;
}

/*********************************************************************
*
*   Set the input colorspace (to something other than JCS_RGB)
*
*   Description:
*       Identify the input colorspace.  This would normally be used
*	when the input is not RGB, as in the case of CIELAB, ICCLAB,
*	and ITULAB.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       cspace:
*           Colorspace, JCS_...
*
*   Returns:
*       ia_successful:
*
*********************************************************************/
ErrCode CDECL 
jc_set_colorspace (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    Int32	cspace
    )
{
    PROLOGUE;

    cinfo->in_color_space = cspace;
    ERR_PASS( jpeg_set_colorspace (pIPState, cinfo, cspace));

    NORMAL_RETURN;
}

/* jc_startCompress() */

/*********************************************************************
*
*   Start for a compression
*
*   Description:
*       Put all of setting in place and ready to do compression.
*       No any setting will be accepted after this called.
*
*   Arguments:
*    pIPState:
*           Point to state
*       cinfo:
*           Pointer to the compression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*    This function is called after jc_init and before compress.
*
*********************************************************************/
ErrCode CDECL 
jc_startCompress (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
    Bool includeTables = TRUE;
    PROLOGUE;

    /* If using separate tables stream, don't include them here */
    if (cinfo->separate_tables) {
	includeTables = FALSE;
    }

/* Prepare for compressing, no any setting can be changed from now on */
    ERR_PASS(jpeg_start_compress (pIPState, cinfo, includeTables));

    NORMAL_RETURN;
}

/* jc_term() */

/*********************************************************************
*
*   Terminate a compression run.
*
*   Description:
*       Terminates the compression run, flushing final data
*       and end-of-image markers to the compressed file.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is an MANDATORY call for compression. 
*       Failure to call this function will result in an
*       incomplete JPEG file which may not be able to be read
*       properly by a decoder.
*
*********************************************************************/
ErrCode CDECL 
jc_term (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
    PROLOGUE;

    JCHECK_STATE;

    ERR_PASS(jpeg_finish_compress (pIPState, cinfo));

    ERR_PASS(jpeg_destroy_compress (pIPState, cinfo));

    if (cinfo->err != NULL)
        IP_FREE(cinfo->err);

    NORMAL_RETURN;
}


/* jc_write_scanline() */

/*********************************************************************
*
*   Write a single scanline to the compressed JPEG file.
*
*   Description:
*       This function is to be called once for each scanline
*       in the input image. The JPEG compressor takes care of 
*       all image buffering necessary to create the DCT blocks
*       and produce the output MCUs.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       srcLine:
*           Pointer to the start of a buffer of data for a single
*           image scanline. For grayscale this is just consecutive
*           UInt8's. For color images, the data must be interleaved
*           as contiguous RGBRGBRGB.... for each scanline. (Or
*           CMYKCMYKCMYK....).
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       The name jc_write_scanline is an abstraction to allow one
*       to think of this in the same way as TIFFWriteScanline().
*       However, all sorts of underlying complexity is being
*       hidden here, including assembling data into DCT blocks,
*       downsampling chroma components, etc. In most cases,
*       data is written only every 8 scanlines for grayscale
*       and every 16 lines for RGB.
*
*********************************************************************/
ErrCode CDECL 
jc_write_scanline (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    JSAMPARRAY	srcLine,
    Int32	nlines)
{
    PROLOGUE;
    JDIMENSION num_lines_written, lines;
    JSAMPARRAY scanlines;
    JCHECK_STATE;

#if 0
/* 
  If this is the first call, make sure that jpeg_start_compress 
  has been called. This way, the API has only an init and a term call.
*/
    if (cinfo->global_state == CSTATE_START)
        ERR_PASS(jpeg_start_compress (pIPState, cinfo, TRUE));
#endif

/* Get a pointer to a line of source image data */
    scanlines = srcLine;

/* JPEG compress the line */
    lines = nlines;
    while(lines >0)
    {
	ERR_PASS(jpeg_write_scanlines (pIPState, cinfo, scanlines, lines,
                                  &num_lines_written));
	lines -= num_lines_written;
	scanlines += num_lines_written;
    }

    NORMAL_RETURN;
}



/*
$Log: jcxapi.c,v $
Revision 1.9  2001/07/06 18:17:42  dabercro
Remove unused functions, cleanup

Revision 1.8  2001/07/05 21:43:55  dabercro
Add an option to extract the tables separatetly for the  JPEGTables TIFF field.

Revision 1.7  2001/06/27 19:12:44  dabercro
Remove colorspace support in JPEG, Lab is supported by identifying the
colorspace as unknown.

Revision 1.6  2001/01/20 00:00:07  dabercro
Comment fixes.

Revision 1.5  2001/01/18 23:29:39  dabercro
Handle 1-component CIELAB/ICCLAB (gray).

Revision 1.4  2000/08/01 21:47:48  dabercro
Handle grayscale for ITULAB also. Changes from Mark Tobin.

Revision 1.3  2000/04/07 23:59:42  dabercro
Added support for CIELAB and ICCLAB color spaces.

Revision 1.2  2000/02/24 00:48:23  dabercro
Changes to support ITULAB color conversion

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.8  1998/07/23 19:54:12  xzhu
xzhu on Thu Jul 23 12:52:00 PDT 1998

jcxapi.c -> /product/ipcore/ipshared/src/RCS/jcxapi.c,v

Moved jpeg_start_compress out from jc_init and made it as a new function jc_startCompress. So we can set something after jc_init. Otherwise we have to set everything in jc_init.

Revision 1.7  1998/07/09 18:47:16  xzhu
xzhu on Thu Jul 9 11:46:52 PDT 1998

jcxapi.c -> /product/ipcore/ipshared/src/RCS/jcxapi.c,v

In set qtable, use quality scaling instead of constant.

Revision 1.6  1998/01/13 23:51:01  ddavies
jcxapi.c -> /product/ipcore/ipshared/src/RCS/jcxapi.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1997/11/19 22:47:01  xzhu
xzhu on Wed Nov 19 14:44:15 PST 1997

jcxapi.c -> /product/ipcore/ipshared/src/RCS/jcxapi.c,v

Added resolution as jc_init's inputs. If XRes YRes >6, override default resolution setting (unknow), otherwise, keep default setting.

Revision 1.4  1997/06/10 20:56:20  xzhu
xzhu on Tue Jun 10 13:54:06 PDT 1997

jcxapi.c -> /product/ipcore/ipshared/src/RCS/jcxapi.c,v

Enabled to setup progressive jpeg compression.

Revision 1.3  1996/10/29 02:13:45  xzhu
xzhu on Mon Oct 28 18:11:23 PST 1996

jcxapi.c -> /product/ipcore/ipshared/src/RCS/jcxapi.c,v

Added jc_free_error_string() for clearning up error message memory.

Revision 1.2  1996/08/19 21:47:44  xzhu
jcxapi.c -> /product/ipcore/ipshared/src/RCS/jcxapi.c,v

CommonCheck: Updated error_string change.

Revision 1.1  1996/04/20 01:25:05  xzhu
Initial revision

*/
