/*
 * jdapimin.c
 *
 * Copyright (C) 1994-1995, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains application interface code for the decompression half
 * of the JPEG library.  These are the "minimum" API routines that may be
 * needed in either the normal full-decompression case or the
 * transcoding-only case.
 *
 * Most of the routines intended to be called directly by an application
 * are in this file or in jdapistd.c.  But also see jcomapi.c for routines
 * shared by compression and decompression, and jdtrans.c for the transcoding
 * case.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jdapimin.c,v $; $Revision: 1.3 $; $Date: 2001/01/18 00:30:38 $");

/*
 * Initialization of a JPEG decompression object.
 * The error manager must already be set up (in case memory manager fails).
 */

GLOBAL ErrCode CDECL 
jpeg_create_decompress (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
  PROLOGUE;

  int i;

  /* 
   * For debugging purposes, zero the whole master structure.
   * But error manager pointer is already there, so save and restore it.
   *
   * The gamut range values could have been set already too, so
   * we need to save/restore them. (mjt - Otherwise, a few functions
   * would have to be changed in order to delay the call to
   * w_jdSetLABGammutRange
   */
  {
    struct jpeg_error_mgr *err = cinfo->err;

    /* clear the memory */
    MEMZERO(cinfo, SIZEOF(struct jpeg_decompress_struct));

    /* restore known values */
    cinfo->err = err;
  }
  cinfo->is_decompressor = TRUE;

  /* Initialize a memory manager instance for this object */
  ERR_PASS(jinit_memory_mgr (pIPState, (j_common_ptr) cinfo));

  /* Zero out pointers to permanent structures. */
  cinfo->progress = NULL;
  cinfo->src = NULL;

  for (i = 0; i < NUM_QUANT_TBLS; i++)
    cinfo->quant_tbl_ptrs[i] = NULL;

  for (i = 0; i < NUM_HUFF_TBLS; i++) {
    cinfo->dc_huff_tbl_ptrs[i] = NULL;
    cinfo->ac_huff_tbl_ptrs[i] = NULL;
  }

  /* Initialize marker processor so application can override methods
   * for COM, APPn markers before calling jpeg_read_header.
   */
  ERR_PASS(jinit_marker_reader (pIPState, cinfo));

  /* And initialize the overall input controller. */
  ERR_PASS(jinit_input_controller (pIPState, cinfo));

  /* OK, I'm ready */
  cinfo->global_state = DSTATE_START;

  NORMAL_RETURN;
}


/*
 * Destruction of a JPEG decompression object
 */

GLOBAL ErrCode CDECL 
jpeg_destroy_decompress (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
  PROLOGUE;

    /* use common routine */
  ERR_PASS(jpeg_destroy (pIPState, (j_common_ptr) cinfo));

  NORMAL_RETURN;
}


/*
 * Abort processing of a JPEG decompression operation,
 * but don't destroy the object itself.
 */

GLOBAL ErrCode CDECL 
jpeg_abort_decompress (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
  PROLOGUE;
    /* use common routine */
  ERR_PASS(jpeg_abort (pIPState, (j_common_ptr) cinfo));

  NORMAL_RETURN;
}


/*
 * Install a special processing method for COM or APPn markers.
 */

GLOBAL ErrCode CDECL 
jpeg_set_marker_processor (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    int		marker_code,
    jpeg_marker_parser_method routine)
{
  PROLOGUE;

  if (marker_code == JPEG_COM)
    cinfo->marker->process_COM = routine;
  else if (marker_code >= JPEG_APP0 && marker_code <= JPEG_APP0+15)
    cinfo->marker->process_APPn[marker_code-JPEG_APP0] = routine;
  else
    ERREXIT1(cinfo, JERR_UNKNOWN_MARKER, marker_code);

  NORMAL_RETURN;
}


/*
 * Set default decompression parameters.
 */

LOCAL ErrCode CDECL 
default_decompress_parms (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
  PROLOGUE;

  /* Guess the input colorspace, and set output colorspace accordingly. */
  /* (Wish JPEG committee had provided a real way to specify this...) */
  /* Note application may override our guesses. */
  switch (cinfo->num_components) {
  case 1:
    cinfo->jpeg_color_space = JCS_GRAYSCALE;
    cinfo->out_color_space = JCS_GRAYSCALE;
    break;
    
  case 3:
    if (cinfo->saw_JFIF_marker) {
      cinfo->jpeg_color_space = JCS_YCbCr; /* JFIF implies YCbCr */
    } else if (cinfo->saw_Adobe_marker) {
      switch (cinfo->Adobe_transform) {
      case 0:
	cinfo->jpeg_color_space = JCS_RGB;
	break;
      case 1:
	cinfo->jpeg_color_space = JCS_YCbCr;
	break;
      default:
	WARNMS1(cinfo, JWRN_ADOBE_XFORM, cinfo->Adobe_transform);
	cinfo->jpeg_color_space = JCS_YCbCr; /* assume it's YCbCr */
	break;
      }
    } else {
      /* Saw no special markers, try to guess from the component IDs */
      int cid0 = cinfo->comp_info[0].component_id;
      int cid1 = cinfo->comp_info[1].component_id;
      int cid2 = cinfo->comp_info[2].component_id;

      if (cid0 == 1 && cid1 == 2 && cid2 == 3)
	cinfo->jpeg_color_space = JCS_YCbCr; /* assume JFIF w/out marker */
      else if (cid0 == 82 && cid1 == 71 && cid2 == 66)
	cinfo->jpeg_color_space = JCS_RGB; /* ASCII 'R', 'G', 'B' */
      else {
	TRACEMS3(cinfo, 1, JTRC_UNKNOWN_IDS, cid0, cid1, cid2);
	cinfo->jpeg_color_space = JCS_YCbCr; /* assume it's YCbCr */
      }
    }
    /* Always guess RGB is proper output colorspace. */
    cinfo->out_color_space = JCS_RGB;
    break;
    
  case 4:
    if (cinfo->saw_Adobe_marker) {
      switch (cinfo->Adobe_transform) {
      case 0:
	cinfo->jpeg_color_space = JCS_CMYK;
	break;
      case 2:
	cinfo->jpeg_color_space = JCS_YCCK;
	break;
      default:
	WARNMS1(cinfo, JWRN_ADOBE_XFORM, cinfo->Adobe_transform);
	cinfo->jpeg_color_space = JCS_YCCK; /* assume it's YCCK */
	break;
      }
    } else {
      /* No special markers, assume straight CMYK. */
      cinfo->jpeg_color_space = JCS_CMYK;
    }
    cinfo->out_color_space = JCS_CMYK;
    break;
    
  default:
    cinfo->jpeg_color_space = JCS_UNKNOWN;
    cinfo->out_color_space = JCS_UNKNOWN;
    break;
  }

  /* Set defaults for other decompression parameters. */
  cinfo->scale_num = 1;		/* 1:1 scaling */
  cinfo->scale_denom = 1;
  cinfo->output_gamma = 1.0;
  cinfo->buffered_image = FALSE;
  cinfo->raw_data_out = FALSE;
  cinfo->dct_method = JDCT_DEFAULT;
  cinfo->do_fancy_upsampling = TRUE;
  cinfo->do_block_smoothing = TRUE;
  cinfo->quantize_colors = FALSE;
  /* We set these in case application only sets quantize_colors. */
  cinfo->dither_mode = JDITHER_FS;
#ifdef QUANT_2PASS_SUPPORTED
  cinfo->two_pass_quantize = TRUE;
#else
  cinfo->two_pass_quantize = FALSE;
#endif
  cinfo->desired_number_of_colors = 256;
  cinfo->colormap = NULL;
  /* Initialize for no mode change in buffered-image mode. */
  cinfo->enable_1pass_quant = FALSE;
  cinfo->enable_external_quant = FALSE;
  cinfo->enable_2pass_quant = FALSE;

  NORMAL_RETURN;
}


/*
 * Decompression startup: read start of JPEG datastream to see what's there.
 * Need only initialize JPEG object and supply a data source before calling.
 *
 * This routine will read as far as the first SOS marker (ie, actual start of
 * compressed data), and will save all tables and parameters in the JPEG
 * object.  It will also initialize the decompression parameters to default
 * values, and finally return JPEG_HEADER_OK.  On return, the application may
 * adjust the decompression parameters and then call jpeg_start_decompress.
 * (Or, if the application only wanted to determine the image parameters,
 * the data need not be decompressed.  In that case, call jpeg_abort or
 * jpeg_destroy to release any temporary space.)
 * If an abbreviated (tables only) datastream is presented, the routine will
 * return JPEG_HEADER_TABLES_ONLY upon reaching EOI.  The application may then
 * re-use the JPEG object to read the abbreviated image datastream(s).
 * It is unnecessary (but OK) to call jpeg_abort in this case.
 * The JPEG_SUSPENDED return code only occurs if the data source module
 * requests suspension of the decompressor.  In this case the application
 * should load more source data and then re-call jpeg_read_header to resume
 * processing.
 * If a non-suspending data source is used and require_image is TRUE, then the
 * return code need not be inspected since only JPEG_HEADER_OK is possible.
 *
 * This routine is now just a front end to jpeg_consume_input, with some
 * extra error checking.
 */

GLOBAL ErrCode CDECL 
jpeg_read_header (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    boolean	require_image,
    int	       *retcode)
{
  PROLOGUE;

  if (cinfo->global_state != DSTATE_START &&
      cinfo->global_state != DSTATE_INHEADER)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);

  ERR_PASS(jpeg_consume_input (pIPState, cinfo, retcode));

  switch (*retcode) {
  case JPEG_REACHED_SOS:
    *retcode = JPEG_HEADER_OK;
    break;
  case JPEG_REACHED_EOI:
    if (require_image)		/* Complain if application wanted an image */
      ERREXIT(cinfo, JERR_NO_IMAGE);
    /* Reset to start state; it would be safer to require the application to
     * call jpeg_abort, but we can't change it now for compatibility reasons.
     * A side effect is to free any temporary memory (there shouldn't be any).
     */
      /* sets state = DSTATE_START */
    ERR_PASS(jpeg_abort (pIPState, (j_common_ptr) cinfo));
    *retcode = JPEG_HEADER_TABLES_ONLY;
    break;
  case JPEG_SUSPENDED:
    /* no work */
    break;
  }

  NORMAL_RETURN;
}


/*
 * Consume data in advance of what the decompressor requires.
 * This can be called at any time once the decompressor object has
 * been created and a data source has been set up.
 *
 * This routine is essentially a state machine that handles a couple
 * of critical state-transition actions, namely initial setup and
 * transition from header scanning to ready-for-start_decompress.
 * All the actual input is done via the input controller's consume_input
 * method.
 */

GLOBAL ErrCode CDECL 
jpeg_consume_input (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    int        *pCalledError)
{
  PROLOGUE;

  *pCalledError = JPEG_SUSPENDED;

  /* NB: every possible DSTATE value should be listed in this switch */
  switch (cinfo->global_state) {
  case DSTATE_START:
    /* Start-of-datastream actions: reset appropriate modules */
    ERR_PASS((*cinfo->inputctl->reset_input_controller) (pIPState, cinfo));
    /* Initialize application's data source module */
    ERR_PASS((*cinfo->src->init_source) (pIPState, cinfo));
    cinfo->global_state = DSTATE_INHEADER;
    /*FALLTHROUGH*/
  case DSTATE_INHEADER:
    ERR_PASS( (*cinfo->inputctl->consume_input) (pIPState, cinfo,
						 pCalledError));
    if (*pCalledError == JPEG_REACHED_SOS) {
      /* Found SOS, prepare to decompress */
      /* Set up default parameters based on header data */
      ERR_PASS(default_decompress_parms (pIPState, cinfo));
      /* Set global state: ready for start_decompress */
      cinfo->global_state = DSTATE_READY;
    }
    break;
  case DSTATE_READY:
    /* Can't advance past first SOS until start_decompress is called */
    *pCalledError = JPEG_REACHED_SOS;
    break;
  case DSTATE_PRELOAD:
  case DSTATE_PRESCAN:
  case DSTATE_SCANNING:
  case DSTATE_RAW_OK:
  case DSTATE_BUFIMAGE:
  case DSTATE_BUFPOST:
  case DSTATE_STOPPING:
    ERR_PASS((*cinfo->inputctl->consume_input) (pIPState, cinfo, pCalledError));
    break;
  default:
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  }

  NORMAL_RETURN;
}


/*
 * Have we finished reading the input file?
 */

GLOBAL ErrCode CDECL 
jpeg_input_complete (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    boolean    *is_compl)
{
  PROLOGUE;

  /* Check for valid jpeg object */
  if (cinfo->global_state < DSTATE_START ||
      cinfo->global_state > DSTATE_STOPPING)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);

   *is_compl = cinfo->inputctl->eoi_reached; 

   NORMAL_RETURN;
}


/*
 * Is there more than one scan?
 */

GLOBAL ErrCode CDECL 
jpeg_has_multiple_scans (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    boolean    *is_mscan)
{
  PROLOGUE;

  /* Only valid after jpeg_read_header completes */
  if (cinfo->global_state < DSTATE_READY ||
      cinfo->global_state > DSTATE_STOPPING)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);

  *is_mscan = cinfo->inputctl->has_multiple_scans; 

  NORMAL_RETURN;
}


/*
 * Finish JPEG decompression.
 *
 * This will normally just verify the file trailer and release temp storage.
 *
 * Returns FALSE if suspended.  The return value need be inspected only if
 * a suspending data source is used.
 */

GLOBAL ErrCode CDECL 
jpeg_finish_decompress (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    boolean    *no_suspend)
{
  PROLOGUE;

  int  calledError;

  *no_suspend = TRUE;
  if ((cinfo->global_state == DSTATE_SCANNING ||
       cinfo->global_state == DSTATE_RAW_OK) && ! cinfo->buffered_image) {
    /* Terminate final pass of non-buffered mode */
    if (cinfo->output_scanline < cinfo->output_height)
	ERREXIT(cinfo, JERR_TOO_LITTLE_DATA);
    ERR_PASS((*cinfo->master->finish_output_pass) (pIPState, cinfo));
    cinfo->global_state = DSTATE_STOPPING;
  } else if (cinfo->global_state == DSTATE_BUFIMAGE) {
    /* Finishing after a buffered-image operation */
    cinfo->global_state = DSTATE_STOPPING;
  } else if (cinfo->global_state != DSTATE_STOPPING) {
    /* STOPPING = repeat call after a suspension, anything else is error */
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  }
  /* Read until EOI */
  while (! cinfo->inputctl->eoi_reached) {
      ERR_PASS((*cinfo->inputctl->consume_input) (pIPState, cinfo,
						  &calledError));
      if ( calledError == JPEG_SUSPENDED)
      {
	  *no_suspend = FALSE;		/* Suspend, come back later */
	  NORMAL_RETURN;
      }
  }
  /* Do final cleanup */
  ERR_PASS((*cinfo->src->term_source) (pIPState, cinfo));
  /* We can use jpeg_abort to release memory and reset global_state */
  ERR_PASS(jpeg_abort (pIPState, (j_common_ptr) cinfo));

  NORMAL_RETURN;
}

/*
$Log: jdapimin.c,v $
Revision 1.3  2001/01/18 00:30:38  dabercro
Removed gammut limit preservation.  Correct way to deal with this
is to set the gammut limits after init.

Revision 1.2  2000/08/01 21:50:09  dabercro
Retain gamut range. Changes from Mark Tobin.

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.3  1998/01/13 23:51:02  ddavies
jdapimin.c -> /product/ipcore/ipshared/src/RCS/jdapimin.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1996/11/06 00:42:51  xzhu
jdapimin.c -> /product/ipcore/ipshared/src/RCS/jdapimin.c,v

CommonCheck:  Did some logical changes to let function return success or error. All outputs go through parameters.

Revision 1.1  1996/04/20 01:29:20  xzhu
Initial revision

*/
