/*
 * jdatadst.c
 *
 * Copyright (C) 1994, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains compression data destination routines for the case of
 * emitting JPEG data to a file (or any stdio stream).  While these routines
 * are sufficient for most applications, some will want to use a different
 * destination manager.
 * IMPORTANT: we assume that fwrite() will correctly transcribe an array of
 * JOCTETs into 8-bit-wide elements on external storage.  If char is wider
 * than 8 bits on your machine, you may need to do some tweaking.
 */

/* this is not a core library module, so it doesn't define JPEG_INTERNALS */
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"
#include "jerror.h"

IP_RCSINFO(RCSInfo, "$RCSfile: jdatadst.c,v $; $Revision: 1.4 $; $Date: 2000/06/09 23:20:17 $");

/* Expanded data destination object for stdio output */

typedef struct {
  struct jpeg_destination_mgr pub; /* public fields */

  void *userState;           /* Users's passed-in state */
  ErrCode 		     /* User's writeMethod */
  (CDECL *writeMethod) (
	IPState	   *pIPState,
	void	   *userState,
	UInt8	   *src,
	UInt32	    nbytes,
	UInt32	   *count);  

  JOCTET * buffer;		/* start of buffer */
} my_destination_mgr;

typedef my_destination_mgr * my_dest_ptr;

#define OUTPUT_BUF_SIZE  4096	/* choose an efficiently fwrite'able size */


/*
 * Initialize destination --- called by jpeg_start_compress
 * before any data is actually written.
 */

METHODDEF ErrCode CDECL 
init_destination (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  PROLOGUE;

  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;

  /* Allocate the output buffer --- it will be released when done with image */
  ERR_PASS((*cinfo->mem->alloc_small) (pIPState, (j_common_ptr) cinfo,
		JPOOL_IMAGE,
		OUTPUT_BUF_SIZE * SIZEOF(JOCTET), (void **)&dest->buffer));

  dest->pub.next_output_byte = dest->buffer;
  dest->pub.free_in_buffer = OUTPUT_BUF_SIZE;

  NORMAL_RETURN;
}


/*
 * Empty the output buffer --- called whenever buffer fills up.
 *
 * In typical applications, this should write the entire output buffer
 * (ignoring the current state of next_output_byte & free_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been dumped.
 *
 * In applications that need to be able to suspend compression due to output
 * overrun, a FALSE return indicates that the buffer cannot be emptied now.
 * In this situation, the compressor will return to its caller (possibly with
 * an indication that it has not accepted all the supplied scanlines).  The
 * application should resume compression after it has made more room in the
 * output buffer.  Note that there are substantial restrictions on the use of
 * suspension --- see the documentation.
 *
 * When suspending, the compressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_output_byte & free_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point will be regenerated after resumption, so do not
 * write it out when emptying the buffer externally.
 */

METHODDEF ErrCode CDECL 
empty_output_buffer (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    boolean    *no_suspend)
{
  PROLOGUE;

  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
  size_t datacoun;

  ERR_PASS(dest->writeMethod(	pIPState, 
				dest->userState,
			        dest->buffer, 
				OUTPUT_BUF_SIZE, 
				(UInt32 *)&datacoun));
  if(datacoun == 0)
  {
      *no_suspend = FALSE;
      NORMAL_RETURN;
  }
  else if( (size_t) datacoun != (size_t) OUTPUT_BUF_SIZE)
    ERREXIT(cinfo, JERR_FILE_WRITE);

  dest->pub.next_output_byte = dest->buffer;
  dest->pub.free_in_buffer = OUTPUT_BUF_SIZE;

  NORMAL_RETURN;
}


/*
 * Terminate destination --- called by jpeg_finish_compress
 * after all data has been written.  Usually needs to flush buffer.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

METHODDEF ErrCode CDECL 
term_destination (
    IPState    *pIPState,
    j_compress_ptr cinfo)
{
  UInt32 count = 0; 
  PROLOGUE;

  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
  size_t datacount = OUTPUT_BUF_SIZE - dest->pub.free_in_buffer;

  /* Write any data remaining in the buffer */
  if (datacount > 0) {
    ERR_PASS(dest->writeMethod(pIPState, dest->userState,
			   dest->buffer, datacount,&count));
    if(count != (UInt32)datacount )
       ERREXIT(cinfo, JERR_FILE_WRITE);
  }

  NORMAL_RETURN;
}


/*
 * Prepare for output to a stdio stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing compression.
 */

GLOBAL ErrCode CDECL 
jpeg_stdio_dest (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    FILE       *outfile)
{
  PROLOGUE;

  my_dest_ptr dest;

  /* The destination object is made permanent so that multiple JPEG images
   * can be written to the same file without re-executing jpeg_stdio_dest.
   * This makes it dangerous to use this manager and a different destination
   * manager serially with the same JPEG object, because their private object
   * sizes may be different.  Caveat programmer.
   */
  if (cinfo->dest == NULL) {	/* first time for this JPEG object? */
    ERR_PASS((*cinfo->mem->alloc_small) (pIPState, (j_common_ptr) cinfo,
			JPOOL_PERMANENT,
			SIZEOF(my_destination_mgr), (void **)&cinfo->dest));
  }

  dest = (my_dest_ptr) cinfo->dest;
  dest->pub.init_destination = init_destination;
  dest->pub.empty_output_buffer = empty_output_buffer;
  dest->pub.term_destination = term_destination;
  dest->userState = (void *)outfile;

  NORMAL_RETURN;
}

GLOBAL ErrCode CDECL 
jpeg_set_destination_mgr (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    void       *jUserState,	   /* handle to pass user state info */
    ErrCode (CDECL *jUserWriteCallback) ( /* user's data output function */
		    IPState    *pIPState,
		    void       *userState,
		    UInt8      *src,
		    UInt32      nbytes,
		    UInt32      *count))
{
  PROLOGUE;

  my_dest_ptr dest;

  /* The destination object is made permanent so that multiple JPEG images
   * can be written to the same file without re-executing jpeg_stdio_dest.
   * This makes it dangerous to use this manager and a different destination
   * manager serially with the same JPEG object, because their private object
   * sizes may be different.  Caveat programmer.
   */
  if (cinfo->dest == NULL) {	/* first time for this JPEG object? */
      ERR_PASS((*cinfo->mem->alloc_small) (pIPState, (j_common_ptr) cinfo,
				JPOOL_PERMANENT,
				SIZEOF(my_destination_mgr),
                                (void **) &(cinfo->dest)));
  }

  dest = (my_dest_ptr) cinfo->dest;
  dest->pub.init_destination = init_destination;
  dest->pub.empty_output_buffer = empty_output_buffer;
  dest->pub.term_destination = term_destination;
  dest->userState = jUserState;
  dest->writeMethod = jUserWriteCallback;

  NORMAL_RETURN;
}

/*
$Log: jdatadst.c,v $
Revision 1.4  2000/06/09 23:20:17  jlin
fix read write callback function return type for jpeg

Revision 1.2  2000/06/07 00:52:04  jlin
change dest->writeMethod to return ErrCode

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.4  1998/01/13 23:51:05  ddavies
jdatadst.c -> /product/ipcore/ipshared/src/RCS/jdatadst.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1997/07/11 01:49:29  ddavies
jdatadst.c -> /product/ipcore/ipshared/src/RCS/jdatadst.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.2  1996/11/06 18:51:58  xzhu
jdatadst.c -> /product/ipcore/ipshared/src/RCS/jdatadst.c,v

CommonCheck: Added data_suspend check in empty_output_buffer() to solve logical conflicts.

Revision 1.1  1996/04/20 01:36:03  xzhu
Initial revision

*/
