/*
 * jddeflts.c
 *
 * Copyright (C) 1991, 1992, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains optional default-setting code for the JPEG decompressor.
 * User interfaces do not have to use this file, but those that don't use it
 * must know more about the innards of the JPEG code.
 */

#include "jpeg.h"
#include "jpeg.pub"
#include "jpeg.prv"


IP_RCSINFO(RCSInfo, "$RCSfile: jddeflts.c,v $; $Revision: 1.2 $; $Date: 2000/06/09 23:23:23 $");

/*
 * Reload the input buffer after it's been emptied, and return the next byte.
 * See the JGETC macro for calling conditions.  Note in particular that
 * read_jpeg_data may NOT return EOF.  If no more data is available, it must
 * exit via ERREXIT, or perhaps synthesize fake data (such as an RST marker).
 * In the present implementation, we insert an EOI marker; this might not be
 * appropriate for non-JFIF file formats, but it usually allows us to handle
 * a truncated JFIF file.
 *
 * This routine can be overridden by the system-dependent user interface,
 * in case the data source is not a stdio stream or some other special
 * condition applies.  Note, however, that this capability only applies for
 * JFIF or similar serial-access JPEG file formats.  The input file control
 * module for a random-access format such as TIFF/JPEG would most likely
 * override the read_jpeg_data method with its own routine.
 */

METHODDEF Int32 CDECL
read_jpeg_data (decompress_info_ptr cinfo)
{
  cinfo->next_input_byte = cinfo->input_buffer + MIN_UNGET;

/*
  Call the user-supplied call back routine to copy from user's 
  compressed data buffer (or file) to JPEG internal buffer.
  Callback routine returns number of bytes read.
*/
  (*cinfo->readCallback) (cinfo, (UInt8 *) cinfo->next_input_byte, 
						  JPEG_BUF_SIZE, (UInt32 *)&cinfo->bytes_in_buffer);

/* The above two lines replace the code commented out below */
#if 0
  cinfo->bytes_in_buffer = (Int32) JFREAD(cinfo->input_file,
					cinfo->next_input_byte,
					JPEG_BUF_SIZE);
#endif
  
  if (cinfo->bytes_in_buffer <= 0) {
    WARNMS(cinfo->emethods, "Premature EOF in JPEG file", JERR_EOF);
    cinfo->next_input_byte[0] = (Int8) 0xFF;
    cinfo->next_input_byte[1] = (Int8) 0xD9; /* EOI marker */
    cinfo->bytes_in_buffer = 2;
  }

  return JGETC(cinfo);
}



/* Default parameter setup for decompression.
 *
 * User interfaces that don't choose to use this routine must do their
 * own setup of all these parameters.  Alternately, you can call this
 * to establish defaults and then alter parameters selectively.  This
 * is the recommended approach since, if we add any new parameters,
 * your code will still work (they'll be set to reasonable defaults).
 *
 * standard_buffering should be TRUE to cause an input buffer to be allocated
 * (the normal case); if FALSE, the user interface must provide a buffer.
 * This option is most useful in the case that the buffer must not be freed
 * at the end of an image.  (For example, when reading a sequence of images
 * from a single file, the remaining data in the buffer represents the
 * start of the next image and mustn't be discarded.)  To handle this,
 * allocate the input buffer yourself at startup, WITHOUT using alloc_small
 * (probably a direct call to malloc() instead).  Then pass FALSE on each
 * call to j_d_defaults to ensure the buffer state is not modified.
 *
 * If the source of the JPEG data is not a stdio stream, override the
 * read_jpeg_data method with your own routine after calling j_d_defaults.
 * You can still use the standard buffer if it's appropriate.
 *
 * CAUTION: if you want to decompress multiple images per run, it's necessary
 * to call j_d_defaults before *each* call to jpeg_decompress, since subsidiary
 * structures like the quantization tables are automatically freed during
 * cleanup.
 */

GLOBAL Int32 CDECL
j_d_defaults (decompress_info_ptr cinfo, boolean standard_buffering)
/* NB: the external methods must already be set up. */
{
  short i;

  /* Initialize pointers as needed to mark stuff unallocated. */
  /* Outer application may fill in default tables for abbreviated files... */
  cinfo->comp_info = NULL;
  for (i = 0; i < NUM_QUANT_TBLS; i++)
    cinfo->quant_tbl_ptrs[i] = NULL;
  for (i = 0; i < NUM_HUFF_TBLS; i++) {
    cinfo->dc_huff_tbl_ptrs[i] = NULL;
    cinfo->ac_huff_tbl_ptrs[i] = NULL;
  }

  /* Default to RGB output */
  /* UI can override by changing out_color_space */
  cinfo->out_color_space = cCS_RGB;
  cinfo->jpeg_color_space = cCS_UNKNOWN;
  /* Setting any other value in jpeg_color_space overrides heuristics in */
  /* jrdjfif.c.  That might be useful when reading non-JFIF JPEG files, */
  /* but ordinarily the UI shouldn't change it. */
  
  /* Default to no smoothing */
  cinfo->do_block_smoothing = FALSE;
  cinfo->do_pixel_smoothing = FALSE;
  
  /* Allocate memory for input buffer, unless outer application provides it. */
  if (standard_buffering) {
    cinfo->input_buffer = (Int8 *) 
			  JALLOC((size_t) (JPEG_BUF_SIZE + MIN_UNGET));
    CHECKMEM(cinfo->input_buffer)
    cinfo->bytes_in_buffer = 0;	/* initialize buffer to empty */
  }

  /* Install standard buffer-reloading method (outer code may override). */
  cinfo->dmethods->read_jpeg_data = read_jpeg_data;


  return ia_successful;
}
/*
$Log: jddeflts.c,v $
Revision 1.2  2000/06/09 23:23:23  jlin
fix read write callback function return type for jpeg

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.7  1995/12/13 02:47:20  ejaquith
jddeflts.c -> /product/ipcore/ipshared/src/RCS/jddeflts.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/08/30  20:57:30  ddavies
 * jddeflts.c -> /product/ipcore/ipshared/src/RCS/jddeflts.c,v
 *
 *  Change names of COLOR_SPACE vars from CS_foo to cCS_foo since they're all
 * constants now.
 *
 * Revision 1.5  1994/10/21  00:49:47  lperry
 * jddeflts.c -> /product/ipcore/ipshared/src/RCS/jddeflts.c,v
 *
 * Get rid of UINT32, etc., so these files can work with Peter's interpreter.
 *
 * Revision 1.4  1994/07/21  00:16:32  lperry
 * jddeflts.c -> /product/ipcore/ipshared/src/RCS/jddeflts.c,v
 *
 * Change error handling from old setjmp/longjmp to function return codes.
 *
 * Revision 1.3  1994/06/27  16:50:01  lperry
 * jddeflts.c -> /product/ipcore/ipshared/src/RCS/jddeflts.c,v
 *
 * Moved static variables to the "cinfo" structure to achieve reentrancy.
 *
 * Revision 1.2  1994/05/24  00:39:44  lperry
 * lperry on Mon May 23 17:38:20 PDT 1994
 *
 * jddeflts.c -> /product/ipcore/ipshared/src/RCS/jddeflts.c,v
 *
 * Change usage of read callback. It now expects it to
 * return the number of bytes read (like Unix read system call)
 *
 * Revision 1.1  1994/02/27  02:21:45  lperry
 * Initial revision
 *
*/
