/*********************************************************************
    Copyright (c) 1995-2000, Xerox Corporation.  All rights reserved. 
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
**********************************************************************/


#include "jinclude.h"
#include "jpeg.pub"

IP_RCSINFO(jdxapi_c_RCSInfo, "$RCSfile: jdxapi.c,v $; $Revision: 1.8 $; $Date: 2001/06/29 17:48:05 $");

/*
  After jc_init or jd_init, this state only exists after an abort.
  Checking for this in every API call (except the inits and error handlers) 
  will prevent segmentation violation / GPF errors if the caller (app)
  neglects to check the return codes.
*/
#define JCHECK_STATE \
if (cinfo->global_state == 0) \
{ \
    ERREXIT(cinfo, JERR_BAD_CINFO); \
}


/*================================================================*/
/*================ DECOMPRESSION API =============================*/
/*================================================================*/

/* jd_init() */

/*********************************************************************
*
*   Initialize for a decompression run
*
*   Description:
*       Establish mandatory parameters for a compression run and
*       establish default settings.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       readCallback:
*           User-supplied callback function to read data from the
*           compressed JPEG file. The decompressor will call this
*           function whenever it needs to read data from the
*           compressed file.
*       userState:
*           A pointer to some user state information, for
*           use by the writeCallback function. 
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is an MANDATORY call for decompression. 
*       The readCallback function is patterned after
*       a UNIX read() function. The userState function
*       provides a mechanism for the user to pass in file
*       specific information such as a FILE descriptor.
*       There are much fewer parameters to this call versus
*       jc_init(), because the JPEG file itself encodes size
*       and component information about the image.
*
*********************************************************************/
ErrCode CDECL 
jd_init (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    jIOCallbackFunc readCallback,
    void       *userState)
{
    PROLOGUE;
    int	calleeError;

/* Set up the error manager (such as it is) */
    cinfo->err = (struct jpeg_error_mgr *) 
                 IP_MALLOC(sizeof(struct jpeg_error_mgr));
    if (cinfo->err == NULL)
    {
        ERREXIT(cinfo, JERR_ERR_MGR_ALLOC);
    }

    ERR_PASS(jpeg_std_error (pIPState, (j_common_ptr)cinfo, cinfo->err));

    ERR_PASS(jpeg_create_decompress (pIPState, cinfo));
    
    ERR_PASS(jpeg_set_source_mgr (pIPState, cinfo, userState, readCallback));

    ERR_PASS(jpeg_read_header (pIPState, cinfo, FALSE, &calleeError));
    if (calleeError == JPEG_HEADER_TABLES_ONLY) {
	ERR_PASS(jpeg_read_header (pIPState, cinfo, TRUE, &calleeError));
    }

	/* The original code threw away calleeError, so I guess we will
	 * also. */

    NORMAL_RETURN;
}

/* jd_set_colorspace() */

/*********************************************************************
*
*   Set the JPEG colorspace (to something other than JCS_RGB)
*
*   Description:
*       Identify the input colorspace.  This would normally be used
*	when the input is not YCbCr, as in the case of CIELAB, ICCLAB,
*	and ITULAB.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       cspace:
*           Colorspace, JCS_...
*
*   Returns:
*       ia_successful:
*
*********************************************************************/
ErrCode CDECL 
jd_set_colorspace (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32 cspace)
{
    PROLOGUE;

    cinfo->out_color_space = cspace;
    cinfo->jpeg_color_space = cspace;
    NORMAL_RETURN;
}

/* jd_start_decompress() */
/*********************************************************************
*
*   Start a decompression run.
*
*   Description:
*       Start the decompression run.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*
*********************************************************************/
ErrCode CDECL 
jd_start_decompress (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
    PROLOGUE;

    ERR_PASS(jpeg_start_decompress (pIPState, cinfo));

    NORMAL_RETURN;
}


/* jd_endQueryJFIF() */
/*********************************************************************
*
*   Terminate a decompression which has been initilized.
*
*   Description:
*       Terminates the decompression which has been initilized
*       for getting image info.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*
*********************************************************************/
ErrCode CDECL 
jd_endQueryJFIF (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
    PROLOGUE;

    ERR_PASS(jpeg_destroy (pIPState, (j_common_ptr)cinfo));

    if (cinfo->err != NULL)
        IP_FREE(cinfo->err);
    cinfo->err = NULL;

    NORMAL_RETURN;
}

/* jd_term() */

/*********************************************************************
*
*   Terminate a decompression run.
*
*   Description:
*       Terminates the decompression run, flushing final data
*       to the decompressed file.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is an MANDATORY call for decompression. 
*       Failure to call this function may result in an
*       incomplete expanded file.
*
*********************************************************************/
ErrCode CDECL 
jd_term (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
    PROLOGUE;

    boolean  no_suspend;

    JCHECK_STATE;

    ERR_PASS( jpeg_finish_decompress (pIPState, cinfo, &no_suspend) );
    if( !no_suspend )
        ERREXIT(cinfo, JERR_CANT_SUSPEND);
  
    ERR_PASS(jpeg_destroy_decompress (pIPState, cinfo));

    if (cinfo->err != NULL)
        IP_FREE(cinfo->err);

    NORMAL_RETURN;
}


/* jd_get_error_string() */

/*********************************************************************
*
*   Retrieve a printable error string.
*
*   Description:
*       Allows an application to retrieve a string containing
*       an error string for display in an error message box
*       or other such purpose.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       string:
*           Pointer to a string variable to receive the error string.
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       This is an OPTIONAL call for decompression. Error strings are
*       listed in the file jerror.h . Since they are all in one
*       place, this is convenient for internationalization.
*
*********************************************************************/
ErrCode CDECL 
jd_get_error_string (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    char      **string)
{
    PROLOGUE;

    /* 
      Dont check global_state here, since we're most 
      likely in an abort condition 
    */

    *string = ((j_common_ptr)cinfo)->err->error_string;

    NORMAL_RETURN;
}

/* jd_free_error_string() */

/*********************************************************************
*
*   Free error struct.
*
*   Description:
*       Error struct is allocated in init, but it will not be freed in
*       error exit since the error string will be retrived after error 
*       appear. So use this function to free error struct when error 
*       exit happen.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*
*********************************************************************/
ErrCode CDECL 
jd_free_error_string (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
    PROLOGUE;

    /* 
      Dont check global_state here, since we're most 
      likely in an abort condition 
    */

    if (cinfo->err != NULL)
        IP_FREE(cinfo->err);

    NORMAL_RETURN;
}

/* jd_read_scanline() */

/*********************************************************************
*
*   Read a single scanline from a compressed JPEG file.
*
*   Description:
*       This function is to be called once for each scanline
*       in the image. The JPEG decompressor takes care of 
*       all image buffering necessary to decode the DCT blocks
*       and produce the output scanlines.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       dstLine:
*           Pointer to the start of a buffer of data to receive a single
*           image scanline. For grayscale this is just consecutive
*           UInt8's. For color images, the data will be interleaved
*           as contiguous RGBRGBRGB.... for each scanline. (Or
*           CMYKCMYKCMYK....).
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       The name jd_read_scanline is an abstraction to allow one
*       to think of this in the same way as TIFFReadScanline().
*       However, all sorts of underlying complexity is being
*       hidden here, including decoding data from DCT blocks,
*       upsampling chroma components, etc. 
*
*********************************************************************/
ErrCode CDECL 
jd_read_scanline (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    JSAMPARRAY	dstLine,
    Int32	inlines)
{
    PROLOGUE;
    JDIMENSION num_lines_read;
    JDIMENSION nlines;
    JSAMPARRAY scanlines;
    JCHECK_STATE;

#if 0
/* 
  If this is the first call, make sure that jpeg_start_decompress 
  has been called. This way, the API has only an init and a term call.
*/
    if (cinfo->global_state == DSTATE_READY)
        ERR_PASS(jpeg_start_decompress (pIPState, cinfo));
#endif

/* Get a pointer to a line of source image data */
    scanlines = dstLine;

/* JPEG compress the line */
    nlines = inlines;
    while( nlines >0)
    {
	ERR_PASS(jpeg_read_scanlines (pIPState, cinfo, scanlines, nlines,
                                  &num_lines_read));
	nlines -= num_lines_read;
	scanlines += num_lines_read;
    }

    NORMAL_RETURN;
}

/* jd_check_progressive */
/*********************************************************************
*
*   Description:
*       If compressed image is progressive jpeg, 
*       enable buffer image.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*  progressive:
*           Pointer to boolean of flag
*  progDisp:
*           Set buffer image or not
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       If don't want progressive display, don't call this function. 
*
*********************************************************************/
ErrCode CDECL 
jd_check_pogressive (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Bool	progDisp,
    Bool       *progressive)
{
   PROLOGUE;

   *progressive = FALSE;
   if( cinfo->progressive_mode &&
       cinfo->input_scan_number > 0)
   {
       if( progDisp )
	   cinfo->buffered_image = TRUE;
       *progressive = TRUE;
   }

   NORMAL_RETURN;
}

/* jd_progStart_output */
/*********************************************************************
*
*   Description:
*       Start progressive pass. It is init of current pass. 
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       pass:
*           Number of pass
*
*   Returns:
*       ia_successful:
*
*   Notes:
*        
*
*********************************************************************/
ErrCode CDECL 
jd_progStart_output (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32	pass)
{
   PROLOGUE;

   if( cinfo->progressive_mode &&
       cinfo->buffered_image)
   {
      ERR_PASS(jpeg_start_output (pIPState, cinfo, pass));
   }

   NORMAL_RETURN;
}

/* jd_progFinish_output */
/*********************************************************************
*
*   Description:
*       Finish one progressive pass. It is term of current pass. 
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*        
*
*********************************************************************/
ErrCode CDECL 
jd_progFinish_output (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
   PROLOGUE;

   if( cinfo->progressive_mode &&
       cinfo->buffered_image)
   {
      ERR_PASS(jpeg_finish_output (pIPState, cinfo));
   }

   NORMAL_RETURN;
}

/* jd_input_complete */
/*********************************************************************
*
*   Description:
*       Check compressed data. If there is no more compressed data,
*     final_pass = true, otherwise, final_pass = false.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       final_pass:
*           Pointer to boolean 
*
*   Returns:
*       ia_successful:
*
*   Notes:
*        
*
*********************************************************************/
ErrCode CDECL 
jd_input_complete (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Bool       *final_pass)
{
   PROLOGUE;

   if( cinfo->progressive_mode &&
       cinfo->buffered_image)
   {
     ERR_PASS (jpeg_input_complete (pIPState, cinfo, (boolean*)final_pass));
      if( !(*final_pass) && 
         cinfo->input_scan_number == cinfo->output_scan_number)
         *final_pass = TRUE;
   }
   else
      *final_pass = TRUE;

   NORMAL_RETURN;
}

/* jd_get_output_pass_lines() */

/*********************************************************************
*
*   Description:
*       Get output number of pass and output lines in current pass. 
*    Normal jpeg has single pass and progressive jpeg has multi-pass.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       pass:
*           Number of output pass ( pass >= 1 )
*       lines:
*           Number of output lines in current pass
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       Number of pass may not include current pass. Output lines
*    may not include current MCU row.
*
*********************************************************************/
ErrCode CDECL 
jd_get_output_pass_lines (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32      *pass,
    Int32      *lines)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state != DSTATE_SCANNING &&
	cinfo->global_state != DSTATE_RAW_OK)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    *pass = (Int32)cinfo->output_scan_number;
    *lines = (Int32)cinfo->output_scanline;

    NORMAL_RETURN;
}

/* jd_get_input_width() */

/*********************************************************************
*
*   Get the width of the image in the compressed JPEG file
*
*   Description:
*       Get the width of the image in the compressed JPEG file
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       width:
*           Image width in pixels
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       In the absence of any exterior knowledge of the image
*       dimensions, this call can be considered MANDATORY, since
*       it will be needed to determine the size of receiving buffers.
*
*   SeeAlso:
*       - jd_get_output_width()
*
*********************************************************************/
ErrCode CDECL 
jd_get_input_width (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32      *width)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state != DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    *width = cinfo->image_width;

    NORMAL_RETURN;
}


/* jd_get_input_height() */

/*********************************************************************
*
*   Get the image height for a compressed JPEG image
*
*   Description:
*       Get the image height for a compressed JPEG image
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       height:
*           Image height in scanlines
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       In the absence of any exterior knowledge of the image
*       dimensions, this call can be considered MANDATORY, since
*       it will be needed to determine the size of receiving buffers.
*
*   SeeAlso:
*       - jd_get_output_height()
*
*********************************************************************/
ErrCode CDECL 
jd_get_input_height (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32      *height)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state != DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    *height = cinfo->image_height;

    NORMAL_RETURN;
}



/* jd_get_output_width() */

/*********************************************************************
*
*   Get the width of the scaled image after scaled decompression
*
*   Description:
*       Get the width of the scaled image after scaled decompression.
*       Calculates and returns the width of the image after scaled
*       decompression. Rounds up to the next highest integer multiple
*      of the reciprocal of the scaling factor.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object.
*       width:
*           Returned value of scaled image width in pixels.
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       Applications should call this function to determine the 
*       output image size rather than trying to calculate it themselves.
*
*   SeeAlso:
*       - jd_get_input_width()
*       - jd_set_output_scale()
*
*********************************************************************/
ErrCode CDECL 
jd_get_output_width (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32      *width)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state != DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    *width = cinfo->output_width;

    NORMAL_RETURN;
}


/* jd_get_output_height() */

/*********************************************************************
*
*   Get the height of the scaled image after scaled decompression
*
*   Description:
*       Get the height of the scaled image after scaled decompression.
*       Calculates and returns the height of the image after scaled
*       decompression. Rounds up to the next highest integer multiple
*      of the reciprocal of the scaling factor.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       height:
*           Returned value of scaled image height in scanlines
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       Applications should call this function to determine the 
*       output image size rather than trying to calculate it themselves.
*
*   SeeAlso:
*       - jd_get_input_height()
*       - jd_set_output_scale()
*
*********************************************************************/
ErrCode CDECL 
jd_get_output_height (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32      *height)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state != DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    *height = cinfo->output_height;

    NORMAL_RETURN;
}


/* jd_get_components() */

/*********************************************************************
*
*   Get the number of color bands in a compressed JPEG file
*
*   Description:
*       Retrieves the number of image components for a decompression run.
*       For grayscale images, this is always one. For color images,
*       it will be 3 for RGB images, and 4 for CMYK images.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       components:
*           Number of image components. This will be one
*           for grayscale images, 3 for RGB and 4 for CMYK.
*
*   Returns:
*       ia_successful: On success
*       JERR_BAD_IN_COLORSPACE: On invalid component count.
*
*   Notes:
*       This is a MANDATORY call for compression.
*       This call gets the number of OUTPUT components,
*       which may be different from the number of input 
*       components if a colorspace change was made. This
*       function should be called AFTER a call to
*       jd_force_grayscale_output().
*
*********************************************************************/
ErrCode CDECL 
jd_get_components (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32      *components)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state != DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    if (cinfo->output_components == 0)  /* No special settings made yet */
    {
        *components = cinfo->num_components;
    }
    else /* Use preset value */
    {
        *components = cinfo->output_components;
    }

    NORMAL_RETURN;
}


/* jd_get_resolution() */

/*********************************************************************
*
*   Gets the image resolution parameter from a compressed JPEG file
*
*   Description:
*       Gets the image resolution parameter from a compressed JPEG file
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       xres:
*           Image resolution along each scanline.
*       yres:
*           Image resolution across scanlines.
*       units:
*           Resolution unit (0 = aspect ratios; 1 = dots/inch;
*           2 = dots/cm).
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       Applications should not assume that this field is
*       filled in correctly for imported JPEG files. For
*       JPEG files produced by THIS code, the values will
*       be correct as long as the call to jc_set resolution was made.
*
*********************************************************************/
ErrCode CDECL 
jd_get_resolution (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32      *xres,
    Int32      *yres,
    Int32      *units)
{
    PROLOGUE;

    JCHECK_STATE;

#if 0
    if (cinfo->global_state != DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }
#endif /* this test seems to be bogus */

    *xres = cinfo->X_density;
    *yres = cinfo->Y_density;
    *units = cinfo->density_unit;

    NORMAL_RETURN;
}


/*********************************************************************
*
*   Get current MCU lines.
*
*   Description:
*       Fix MCU lines are samp_fact * DCTSIZE. At the bottom of
*       image (last MCU) the lines are remaining lines.
*
*   Arguments:
*       cinfo:
*           Pointer to the decompression object
*       nlines:
*           Pointer to a nlines.
*
*   Returns:
*       ia_successful:
*
*
*********************************************************************/
ErrCode CDECL 
jd_get_linesToRead (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32      *nlines)
{
    PROLOGUE;
    Int32 rows, MCURows;

    if (cinfo->global_state == DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    MCURows = cinfo->max_v_samp_factor * DCTSIZE;
    rows = cinfo->output_height - cinfo->output_scanline;
    if( rows >  MCURows) rows = MCURows;
    *nlines = rows;

    NORMAL_RETURN;
}

/* jd_set_output_scale() */

/*********************************************************************
*
*   Set output scaling factor.
*
*   Description:
*       Set the output scaling factor to produce an output
*       image of dimensions 1/2, 1/4 or 1/8 of full scale.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*       scale:
*           Floating point scale factor (see Notes)
*       output_width:
*           Returned value of the scaled width.
*       output_height:
*           Returned value of the scaled height.
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       Currently the only legal values for scale
*       are 1.0, 0.5, 0.25 and 0.125 . Values other than this
*       are forced to the closest legal value which is of higher
*       resolution. Values greater than 1.0 are forced to 1.0.
*       It is expected that this interface will someday be extended
*       to provide interpolated resampling to acheive arbitrary
*       output scaling.
*       The 0.25 scaling factor uses the special Xerox Webster
*       Research Center Inverse DCT which provides extremely
*       fast output.
*
*********************************************************************/
ErrCode CDECL 
jd_set_output_scale (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Float64	scale,
    Int32      *output_width,
    Int32      *output_height)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state != DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    if (scale == 1.0)
    {
        cinfo->scale_num = 1;
        cinfo->scale_denom = 1;
        *output_width = cinfo->image_width;
        *output_height = cinfo->image_height;
    }
    else if (scale == 0.5)
    {
        cinfo->scale_num = 1;
        cinfo->scale_denom = 2;
        *output_width = (cinfo->image_width + 1) / 2;
        *output_height = (cinfo->image_height + 1) / 2;
    }
    else if (scale == 0.25)
    {
        cinfo->scale_num = 1;
        cinfo->scale_denom = 4;
        *output_width = (cinfo->image_width + 3) / 4;
        *output_height = (cinfo->image_height + 3) / 4;
    }
    else if (scale == 0.125)
    {
        cinfo->scale_num = 1;
        cinfo->scale_denom = 8;
        *output_width = (cinfo->image_width + 7) / 8;
        *output_height = (cinfo->image_height + 7) / 8;
    }
    else
    {
        /* 
          Use the closest scale which is of higher resolution than
          that requested.
        */
        if (scale > 1.0)
            scale = 1.0;
        else if (scale < 0.125)
            scale = 0.125;
        else if (scale < 0.25)
            scale = 0.25;
        else if (scale < 0.5)
            scale = 0.5;
        else if (scale < 1.0)
            scale = 1.0;

        /* Call again with the adjusted scale value */
        ERR_PASS(jd_set_output_scale (pIPState, cinfo, scale, 
                                     output_width, output_height));

    }

    NORMAL_RETURN;
}

/* jd_force_grayscale_output() */

/*********************************************************************
*
*   Force grayscale output, even for a color image.
*
*   Description:
*       Sets the output color space to be grayscale, even
*       if the image is RGB or CMYK. The luminance component
*       of the internal representation is output.
*
*   Arguments:
*       cinfo:
*           Pointer to the compression object
*
*   Returns:
*       ia_successful:
*
*   Notes:
*       Works with either gray or color images. Can be
*       combined with output scaling to get an even faster
*       grayscale thumbnail. After this call, a call to
*       jd_get_components() MUST be made to get the proper
*       output component count.
*
*********************************************************************/
ErrCode CDECL 
jd_force_grayscale_output (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state != DSTATE_READY)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    cinfo->out_color_space = JCS_GRAYSCALE;
    cinfo->out_color_components = 1;
    cinfo->output_components = 1;

    NORMAL_RETURN;
}

/* jd_set_smart_idct() */

/*********************************************************************
*
*   Set smart IDCT flag.
*
*   Description:
*       When smart IDCT flag is set, the image data will pass 
*     a filter based post processing after normal IDCT. The speed
*     will slow 4-5 times.
*
*   Arguments:
*       cinfo:       Pointer to the compression object
*        flag:       SIDCT flag, 1-- turn on SIDCT, 0-- turn off SIDCT
*
*   Returns:
*       ia_successful:
*
*   Notes:
*
 *********************************************************************/
ErrCode CDECL 
jd_set_smart_idct (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    Int32	flag)
{
    PROLOGUE;

    JCHECK_STATE;

    if (cinfo->global_state < DSTATE_READY || 
	cinfo->global_state >= DSTATE_STOPPING)
    {
        ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
    }

    cinfo->smart_idct_flag = flag;

    NORMAL_RETURN;
}


/*
$Log: jdxapi.c,v $
Revision 1.8  2001/06/29 17:48:05  dabercro
Allow jpeg_read_header to be called twice, the second time occurs when
the tables appeared in separate segment.

Revision 1.7  2001/06/27 19:12:44  dabercro
Remove colorspace support in JPEG, Lab is supported by identifying the
colorspace as unknown.

Revision 1.6  2001/01/20 00:01:10  dabercro
Remove seemingly unnecessary status check (per Ron
Danisovizch (sp))

Revision 1.5  2001/01/18 23:29:49  dabercro
Handle 1-component CIELAB/ICCLAB (gray).

Revision 1.4  2000/08/01 21:51:47  dabercro
Copyright notice. Changes from Mark Tobin.

Revision 1.3  2000/04/07 23:59:43  dabercro
Added support for CIELAB and ICCLAB color spaces.

Revision 1.2  2000/02/24 00:48:23  dabercro
Changes to support ITULAB color conversion

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.9  1998/07/23 19:17:40  xzhu
xzhu on Thu Jul 23 12:17:25 PDT 1998

jdxapi.c -> /product/ipcore/ipshared/src/RCS/jdxapi.c,v

In check_progressive, the output parameter progressive was not set to default value FALSE and might be unchanged.

Revision 1.8  1998/01/13 23:51:29  ddavies
jdxapi.c -> /product/ipcore/ipshared/src/RCS/jdxapi.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1997/07/01 23:24:01  xzhu
xzhu on Tue Jul 1 16:23:31 PDT 1997

jdxapi.c -> /product/ipcore/ipshared/src/RCS/jdxapi.c,v

Implemented more functions for progressive jpeg decompression.

Revision 1.6  1997/07/01 00:04:12  xzhu
xzhu on Mon Jun 30 17:03:30 PDT 1997

jdxapi.c -> /product/ipcore/ipshared/src/RCS/jdxapi.c,v

Created new functions for progressive jpeg decomposition.

Revision 1.5  1996/11/06 00:42:47  xzhu
xzhu on Tue Nov 5 16:42:47 PST 1996

jdxapi.c -> /product/ipcore/ipshared/src/RCS/jdxapi.c,v

CommonCheck:  Did some logical changes to let function return success or error. All outputs go through parameters.

Revision 1.4  1996/10/29 02:15:32  xzhu
xzhu on Mon Oct 28 18:14:39 PST 1996

jdxapi.c -> /product/ipcore/ipshared/src/RCS/jdxapi.c,v

Added jd_free_error_string to clearning up error message memory.

Revision 1.3  1996/09/20 21:47:55  xzhu
xzhu on Fri Sep 20 14:47:26 PDT 1996

jdxapi.c -> /product/ipcore/ipshared/src/RCS/jdxapi.c,v

Added new interface function to set smart IDCT flag.

Revision 1.2  1996/08/19 21:47:48  xzhu
jdxapi.c -> /product/ipcore/ipshared/src/RCS/jdxapi.c,v

CommonCheck: Updated error_string change.

Revision 1.1  1996/04/20 01:25:43  xzhu
Initial revision

*/
