/*
 * jfdctint.c
 *
 * Copyright (C) 1991-1994, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains a slow-but-accurate integer implementation of the
 * forward DCT (Discrete Cosine Transform).
 *
 * A 2-D DCT can be done by 1-D DCT on each row followed by 1-D DCT
 * on each column.  Direct algorithms are also available, but they are
 * much more complex and seem not to be any faster when reduced to code.
 *
 * This implementation is based on an algorithm described in
 *   C. Loeffler, A. Ligtenberg and G. Moschytz, "Practical Fast 1-D DCT
 *   Algorithms with 11 Multiplications", Proc. Int'l. Conf. on Acoustics,
 *   Speech, and Signal Processing 1989 (ICASSP '89), pp. 988-991.
 * The primary algorithm described there uses 11 multiplies and 29 adds.
 * We use their alternate method with 12 multiplies and 32 adds.
 * The advantage of this method is that no data path contains more than one
 * multiplication; this allows a very simple and accurate implementation in
 * scaled fixed-point arithmetic, with a minimal number of shifts.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "jdct.h"		/* Private declarations for DCT subsystem */

IP_RCSINFO(RCSInfo, "$RCSfile: jfdctint.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");

#ifdef DCT_ISLOW_SUPPORTED

/*
 * This module is specialized to the case DCTSIZE = 8.
 */

#if DCTSIZE != 8
  Sorry, this code only copes with 8x8 DCTs. /* deliberate syntax err */
#endif


/*
 * The poop on this scaling stuff is as follows:
 *
 * Each 1-D DCT step produces outputs which are a factor of sqrt(N)
 * larger than the true DCT outputs.  The final outputs are therefore
 * a factor of N larger than desired; since N=8 this can be cured by
 * a simple right shift at the end of the algorithm.  The advantage of
 * this arrangement is that we save two multiplications per 1-D DCT,
 * because the y0 and y4 outputs need not be divided by sqrt(N).
 * In the IJG code, this factor of 8 is removed by the quantization step
 * (in jcdctmgr.c), NOT in this module.
 *
 * We have to do addition and subtraction of the integer inputs, which
 * is no problem, and multiplication by fractional constants, which is
 * a problem to do in integer arithmetic.  We multiply all the constants
 * by CONST_SCALE and convert them to integer constants (thus retaining
 * CONST_BITS bits of precision in the constants).  After doing a
 * multiplication we have to divide the product by CONST_SCALE, with proper
 * rounding, to produce the correct output.  This division can be done
 * cheaply as a right shift of CONST_BITS bits.  We postpone shifting
 * as long as possible so that partial sums can be added together with
 * full fractional precision.
 *
 * The outputs of the first pass are scaled up by PASS1_BITS bits so that
 * they are represented to better-than-integral precision.  These outputs
 * require BITS_IN_JSAMPLE + PASS1_BITS + 3 bits; this fits in a 16-bit word
 * with the recommended scaling.  (For 12-bit sample data, the intermediate
 * array is Int32 anyway.)
 *
 * To avoid overflow of the 32-bit intermediate results in pass 2, we must
 * have BITS_IN_JSAMPLE + CONST_BITS + PASS1_BITS <= 26.  error analysis
 * shows that the values given below are the most effective.
 */

#if BITS_IN_JSAMPLE == 8
#define CONST_BITS  13
#define PASS1_BITS  2
#define P_CENTERJSAMPLE 1024
#define DESCALE1(x) ( (x+1024)>>11 )
#else
#define CONST_BITS  13
#define PASS1_BITS  1		/* lose a little precision to avoid overflow */
#define P_CENTERJSAMPLE 16384
#define DESCALE1(x) ( (x+2048)>>12 )
#endif

/* Some C compilers fail to reduce "FIX(constant)" at compile time, thus
 * causing a lot of useless floating-point operations at run time.
 * To get around this we use the following pre-calculated constants.
 * If you change CONST_BITS you may want to add appropriate values.
 * (With a reasonable C compiler, you can just rely on the FIX() macro...)
 */

#if CONST_BITS == 13
#define FIX_0_298631336  ((Int32)  2446)	/* FIX(0.298631336) */
#define FIX_0_390180644  ((Int32)  3196)	/* FIX(0.390180644) */
#define FIX_0_541196100  ((Int32)  4433)	/* FIX(0.541196100) */
#define FIX_0_765366865  ((Int32)  6270)	/* FIX(0.765366865) */
#define FIX_0_899976223  ((Int32)  7373)	/* FIX(0.899976223) */
#define FIX_1_175875602  ((Int32)  9633)	/* FIX(1.175875602) */
#define FIX_1_501321110  ((Int32)  12299)	/* FIX(1.501321110) */
#define FIX_1_847759065  ((Int32)  15137)	/* FIX(1.847759065) */
#define FIX_1_961570560  ((Int32)  16069)	/* FIX(1.961570560) */
#define FIX_2_053119869  ((Int32)  16819)	/* FIX(2.053119869) */
#define FIX_2_562915447  ((Int32)  20995)	/* FIX(2.562915447) */
#define FIX_3_072711026  ((Int32)  25172)	/* FIX(3.072711026) */
#else
#define FIX_0_298631336  FIX(0.298631336)
#define FIX_0_390180644  FIX(0.390180644)
#define FIX_0_541196100  FIX(0.541196100)
#define FIX_0_765366865  FIX(0.765366865)
#define FIX_0_899976223  FIX(0.899976223)
#define FIX_1_175875602  FIX(1.175875602)
#define FIX_1_501321110  FIX(1.501321110)
#define FIX_1_847759065  FIX(1.847759065)
#define FIX_1_961570560  FIX(1.961570560)
#define FIX_2_053119869  FIX(2.053119869)
#define FIX_2_562915447  FIX(2.562915447)
#define FIX_3_072711026  FIX(3.072711026)
#endif


/* Multiply an Int32 variable by an Int32 constant to yield an Int32 result.
 * For 8-bit samples with the recommended scaling, all the variable
 * and constant values involved are no more than 16 bits wide, so a
 * 16x16->32 bit multiply can be used instead of a full 32x32 multiply.
 * For 12-bit samples, a full 32-bit multiplication will be needed.
 */

#if BITS_IN_JSAMPLE == 8
#define MULTIPLY(var,const)  MULTIPLY16C16(var,const)
#else
#define MULTIPLY(var,const)  ((var) * (const))
#endif

/*
  Multiplication macros for the specific muliplication coeffs used in JPEG 
  compression. Instead of just directly doing the multiplication,
  an equivalent set of shifts and add/subtracts are used.
  The GNU compiler implements very efficient code in this way whenever
  it sees integer multiplication by compile-time constants. This
  code is modeled after that  extracted from assembly listings. 
  It seems to be even better than that produced using the usual
  Horner's rule polynomial evaluation technique.

  We use a single temp variable mtemp when required, and return the
  result as the last expression of a comma expression.

  On the Sun, doing it explicitly this way makes only about a
  7% improvement. The PC's Watcom compiler puts in actual multiplies,
  however, so this code shows about a 14% improvement there. 
*/

/* Macro to multiply by 2446 */
#define M_0_298631336(x) \
( mtemp = x*8+x , ((mtemp*16+mtemp)*8-x)*2 )

/* Macro to multiply by 3196 */
#define M_0_390180644(x) \
( (((x*2+x)*8+x)*32-x)*4 )

/* Macro to multiply by 4433 */
#define M_0_541196100(x) \
( mtemp = (x*8+x)*16-x , mtemp*32 - mtemp )

/* Macro to multiply by 6270 */
#define M_0_765366865(x) \
( (((x*2+x)*16+x)*64-x)*2 )

/* Macro to multiply by 7373 */
#define M_0_899976223(x) \
( (((((x*8-x)*4+x)*4-x)*4+x)*4-x)*4+x )

/* Macro to multiply by 9633 */
#define M_1_175875602(x) \
( mtemp = x*4+x , ((mtemp*16-mtemp)*4+x)*32+x )

/* Macro to multiply by 12299 */
#define M_1_501321110(x) \
( mtemp = x*2+x , (mtemp*1024+mtemp)*4-x )

/* Macro to multiply by 15137 */
#define M_1_847759065(x) \
( (((x*16-x)*4-x)*8+x)*32+x )

/* Macro to multiply by 16069 */
#define M_1_961570560(x) \
( (((x*64-x)*4-x)*16+x)*4+x )

/* Macro to multiply by 16819 */
#define M_2_053119869(x) \
( ((((x*32+x)*8-x)*4-x)*4+x)*4-x )

/* Macro to multiply by 20995 */
#define M_2_562915447(x) \
( (((x*4+x)*8+x)*128+x)*4-x )

/* Macro to multiply by 25172 */
#define M_3_072711026(x) \
( mtemp = x*2+x , mtemp = (mtemp*16+mtemp)*4-x , (mtemp*32-mtemp)*4 )

/*
 * Perform the forward DCT on one block of samples.
 */

GLOBAL ErrCode CDECL 
jpeg_fdct_islow (
    IPState    *pIPState,
    UInt8      *inData,
    DCTELEM    *outData)
{
  PROLOGUE;

  Int32 tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7;
  Int32 tmp10, tmp11, tmp12, tmp13;
  Int32 z1, z2, z3, z4, z5;
  Int32 mtemp;  /* temp variable for multiply macros */
  DCTELEM *dataptr;
  UInt8 *indataptr;
  int ctr;
  SHIFT_TEMPS

  /* Pass 1: process rows. */
  /* Note results are scaled up by sqrt(8) compared to a true DCT; */
  /* furthermore, we scale the results by 2**PASS1_BITS. */

  indataptr = inData;
  dataptr = outData;
  for (ctr = 0; ctr < DCTSIZE; ctr++) 
  {
    /* Do the 8 point butterfly */
    tmp0 = (Int32) indataptr[0] + indataptr[7];
    tmp7 = (Int32) indataptr[0] - indataptr[7];
    tmp1 = (Int32) indataptr[1] + indataptr[6];
    tmp6 = (Int32) indataptr[1] - indataptr[6];
    tmp2 = (Int32) indataptr[2] + indataptr[5];
    tmp5 = (Int32) indataptr[2] - indataptr[5];
    tmp3 = (Int32) indataptr[3] + indataptr[4];
    tmp4 = (Int32) indataptr[3] - indataptr[4];
    
    /* Even part per LL&M figure 1 --- note that published figure is faulty;
     * rotator "sqrt(2)*c1" should be "sqrt(2)*c6".
     */
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
   
/* 
  Put the bias subtraction into the DC term here.
  Everything else cancels out.
*/
    dataptr[0] = ((tmp10 + tmp11 - P_CENTERJSAMPLE) << PASS1_BITS);
    dataptr[4] = ((tmp10 - tmp11) << PASS1_BITS);
    
    z1 = (tmp12+tmp13);
    z1 = M_0_541196100(z1);

    dataptr[2] = DESCALE1(z1 + M_0_765366865(tmp13) ); 
    dataptr[6] = DESCALE1(z1 - M_1_847759065(tmp12) ); 
    
    /* Odd part per figure 8 --- note paper omits factor of sqrt(2).
     * cK represents cos(K*pi/16).
     * i0..i3 in the paper are tmp4..tmp7 here.
     */
    
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;

    z5 = z3 + z4;
    z5 = M_1_175875602(z5);

    tmp4 = M_0_298631336(tmp4);

    tmp5 = M_2_053119869(tmp5);

    tmp6 = M_3_072711026(tmp6);

    tmp7 = M_1_501321110(tmp7);

    z1 = - M_0_899976223(z1);

    z2 = - M_2_562915447(z2);

    z3 = - M_1_961570560(z3);

    z4 = - M_0_390180644(z4);

    
    z3 += z5;
    z4 += z5;
    
    dataptr[7] = DESCALE1(tmp4 + z1 + z3);
    dataptr[5] = DESCALE1(tmp5 + z2 + z4);
    dataptr[3] = DESCALE1(tmp6 + z2 + z3);
    dataptr[1] = DESCALE1(tmp7 + z1 + z4);
    
    dataptr += DCTSIZE;		/* advance pointer to next row */
    indataptr += DCTSIZE;
  }


  /* Pass 2: process columns. */
  /* Note that we must descale the results by a factor of 8 == 2**3, */
  /* and also undo the PASS1_BITS scaling. */

  dataptr = outData;
  for (ctr = DCTSIZE; ctr > 0; ctr--) 
  {
    tmp0 = dataptr[P_INDEX_0] + dataptr[P_INDEX_7];
    tmp7 = dataptr[P_INDEX_0] - dataptr[P_INDEX_7];
    tmp1 = dataptr[P_INDEX_1] + dataptr[P_INDEX_6];
    tmp6 = dataptr[P_INDEX_1] - dataptr[P_INDEX_6];
    tmp2 = dataptr[P_INDEX_2] + dataptr[P_INDEX_5];
    tmp5 = dataptr[P_INDEX_2] - dataptr[P_INDEX_5];
    tmp3 = dataptr[P_INDEX_3] + dataptr[P_INDEX_4];
    tmp4 = dataptr[P_INDEX_3] - dataptr[P_INDEX_4];
    
    /* Even part per LL&M figure 1 --- note that published figure is faulty;
     * rotator "sqrt(2)*c1" should be "sqrt(2)*c6".
     */
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    dataptr[P_INDEX_0] = DESCALE(tmp10 + tmp11, PASS1_BITS);
    dataptr[P_INDEX_4] = DESCALE(tmp10 - tmp11, PASS1_BITS);
    
    z1 = (tmp12+tmp13);
    z1 = M_0_541196100(z1);

    dataptr[P_INDEX_2] = DESCALE(z1 + M_0_765366865(tmp13) , 
					   CONST_BITS+PASS1_BITS);

    
    dataptr[P_INDEX_6] = DESCALE(z1 - M_1_847759065(tmp12) , 
                                           CONST_BITS+PASS1_BITS);
    
    /* Odd part per figure 8 --- note paper omits factor of sqrt(2).
     * cK represents cos(K*pi/16).
     * i0..i3 in the paper are tmp4..tmp7 here.
     */
    
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;

    z5 = z3 + z4;
    z5 = M_1_175875602(z5);

    tmp4 = M_0_298631336(tmp4);

    tmp5 = M_2_053119869(tmp5);

    tmp6 = M_3_072711026(tmp6);

    tmp7 = M_1_501321110(tmp7);

    z1 = - M_0_899976223(z1);

    z2 = - M_2_562915447(z2);

    z3 = - M_1_961570560(z3);

    z4 = - M_0_390180644(z4);
    
    z3 += z5;
    z4 += z5;
    
    dataptr[P_INDEX_7] = DESCALE(tmp4 + z1 + z3,
					   CONST_BITS+PASS1_BITS);
    dataptr[P_INDEX_5] = DESCALE(tmp5 + z2 + z4,
					   CONST_BITS+PASS1_BITS);
    dataptr[P_INDEX_3] = DESCALE(tmp6 + z2 + z3,
					   CONST_BITS+PASS1_BITS);
    dataptr[P_INDEX_1] = DESCALE(tmp7 + z1 + z4,
					   CONST_BITS+PASS1_BITS);
    
    dataptr++;			/* advance pointer to next column */
  }

  NORMAL_RETURN;
}

#endif /* DCT_ISLOW_SUPPORTED */

/*
$Log: jfdctint.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.2  1998/01/13 23:51:35  ddavies
jfdctint.c -> /product/ipcore/ipshared/src/RCS/jfdctint.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.1  1996/04/20 01:41:16  xzhu
Initial revision

*/
