/*
 * jfwddct.c
 *
 * Copyright (C) 1991, 1992, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains the basic DCT (Discrete Cosine Transform)
 * transformation subroutine.
 *
 * This implementation is based on an algorithm described in
 *   C. Loeffler, A. Ligtenberg and G. Moschytz, "Practical Fast 1-D DCT
 *   Algorithms with 11 Multiplications", Proc. Int'l. Conf. on Acoustics,
 *   Speech, and Signal Processing 1989 (ICASSP '89), pp. 988-991.
 * The primary algorithm described there uses 11 multiplies and 29 adds.
 * We use their alternate method with 12 multiplies and 32 adds.
 * The advantage of this method is that no data path contains more than one
 * multiplication; this allows a very simple and accurate implementation in
 * scaled fixed-point arithmetic, with a minimal number of shifts.
 */

#include "jpeg.h"
#include "jpeg.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jfwddct.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");
/*
 * This routine is specialized to the case DCTSIZE = 8.
 */

#if DCTSIZE != 8
  Sorry, this code only copes with 8x8 DCTs. /* deliberate syntax err */
#endif


/*
 * A 2-D DCT can be done by 1-D DCT on each row followed by 1-D DCT
 * on each column.  Direct algorithms are also available, but they are
 * much more complex and seem not to be any faster when reduced to code.
 *
 * The poop on this scaling stuff is as follows:
 *
 * Each 1-D DCT step produces outputs which are a factor of sqrt(N)
 * larger than the true DCT outputs.  The final outputs are therefore
 * a factor of N larger than desired; since N=8 this can be cured by
 * a simple right shift at the end of the algorithm.  The advantage of
 * this arrangement is that we save two multiplications per 1-D DCT,
 * because the y0 and y4 outputs need not be divided by sqrt(N).
 *
 * We have to do addition and subtraction of the integer inputs, which
 * is no problem, and multiplication by fractional constants, which is
 * a problem to do in integer arithmetic.  We multiply all the constants
 * by CONST_SCALE and convert them to integer constants (thus retaining
 * CONST_BITS bits of precision in the constants).  After doing a
 * multiplication we have to divide the product by CONST_SCALE, with proper
 * rounding, to produce the correct output.  This division can be done
 * cheaply as a right shift of CONST_BITS bits.  We postpone shifting
 * as long as possible so that partial sums can be added together with
 * full fractional precision.
 *
 * The outputs of the first pass are scaled up by PASS1_BITS bits so that
 * they are represented to better-than-integral precision.  These outputs
 * require BITS_IN_JSAMPLE + PASS1_BITS + 3 bits; this fits in a 16-bit word
 * with the recommended scaling.  (To scale up 12-bit sample data, an
 * intermediate Int32 array would be needed.)
 *
 * To avoid overflow of the 32-bit intermediate results in pass 2, we must
 * have BITS_IN_JSAMPLE + CONST_BITS + PASS1_BITS <= 25.  Error analysis
 * shows that the values given below are the most effective.
 */

#ifdef EIGHT_BIT_SAMPLES
#define CONST_BITS  13
#define PASS1_BITS  2
#else
#define CONST_BITS  13
#define PASS1_BITS  0		/* lose a little precision to avoid overflow */
#endif

#define ONE	((Int32) 1)

#define CONST_SCALE (ONE << CONST_BITS)

/* Convert a positive real constant to an integer scaled by CONST_SCALE. */

/*
  Multiplication macros for the specific muliplication coeffs used in JPEG 
  compression. Instead of just directly doing the multiplication,
  an equivalent set of shifts and add/subtracts are used.
  The GNU compiler implements very efficient code in this way whenever
  it sees integer multiplication by compile-time constants. This
  code is modeled after that  extracted from assembly listings. 
  It seems to be even better than that produced using the usual
  Horner's rule polynomial evaluation technique.

  We use a single temp variable mtemp when required, and return the
  result as the last expression of a comma expression.

  On the Sun, doing it explicitly this way makes only about a
  7% improvement. The PC's Watcom compiler puts in actual multiplies,
  however, so this code shows about a 14% improvement there. 
*/

/* Macro to multiply by 2446 */
#define M_0_298631336(x) \
( mtemp = x*8+x , ((mtemp*16+mtemp)*8-x)*2 )

/* Macro to multiply by 3196 */
#define M_0_390180644(x) \
( (((x*2+x)*8+x)*32-x)*4 )

/* Macro to multiply by 4433 */
#define M_0_541196100(x) \
( mtemp = (x*8+x)*16-x , mtemp*32 - mtemp )

/* Macro to multiply by 6270 */
#define M_0_765366865(x) \
( (((x*2+x)*16+x)*64-x)*2 )

/* Macro to multiply by 7373 */
#define M_0_899976223(x) \
( (((((x*8-x)*4+x)*4-x)*4+x)*4-x)*4+x )

/* Macro to multiply by 9633 */
#define M_1_175875602(x) \
( mtemp = x*4+x , ((mtemp*16-mtemp)*4+x)*32+x )

/* Macro to multiply by 12299 */
#define M_1_501321110(x) \
( mtemp = x*2+x , (mtemp*1024+mtemp)*4-x )

/* Macro to multiply by 15137 */
#define M_1_847759065(x) \
( (((x*16-x)*4-x)*8+x)*32+x )

/* Macro to multiply by 16069 */
#define M_1_961570560(x) \
( (((x*64-x)*4-x)*16+x)*4+x )

/* Macro to multiply by 16819 */
#define M_2_053119869(x) \
( ((((x*32+x)*8-x)*4-x)*4+x)*4-x )

/* Macro to multiply by 20995 */
#define M_2_562915447(x) \
( (((x*4+x)*8+x)*128+x)*4-x )

/* Macro to multiply by 25172 */
#define M_3_072711026(x) \
( mtemp = x*2+x , mtemp = (mtemp*16+mtemp)*4-x , (mtemp*32-mtemp)*4 )

/* Descale and correctly round an Int32 value that's scaled by N bits.
 * We assume RIGHT_SHIFT rounds towards minus infinity, so adding
 * the fudge factor is correct for either sign of X.
 */

#define DESCALE(x,n)  RIGHT_SHIFT((x) + (ONE << ((n)-1)), n)


/*
 * Perform the forward DCT on one block of samples.
 */

GLOBAL Int32 CDECL
j_fwd_dct (UInt8 *inBlock, DCTBLOCK data)
{
  Int32 tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7;
  Int32 tmp10, tmp11, tmp12, tmp13;
  Int32 z1, z2, z3, z4, z5, mtemp;
  register DCTELEM *dataptr;
  register UInt8 *indataptr;
  int rowctr;
  SHIFT_TEMPS

  /* Pass 1: process rows. */
  /* Note results are scaled up by sqrt(8) compared to a true DCT; */
  /* furthermore, we scale the results by 2**PASS1_BITS. */

/*
  A word about the input pixel biasing.

  The original implementation of this program biased
  every input pixel by subtracting 128 in the extract_block
  function in jcmcu.c. In order to move the data in 32 bit
  chunks rather then bytes, the bias subtraction was moved into
  this DCT routine. 

  If you look at the first 8 lines of the first pass loop,
  they perform a butterfly operation on a row of 8 pixels.
  The tmp0 ... tmp7 terms are the butterfly output. 
  The additions (tmp0 ... tmp3) would normally require a
  subtraction of 256 (one for each pixel in the add).
  The subtractions (tmp4 ... tmp7) don't require any biasing
  since the two biases wil cancel each other out.

  The additions (tmp0 ... tmp3) only take part in the production
  of the even output terms (0,2,4,6). They feed into a 4 point
  butterfly. The outputs of that butterfly feed a 2 point butterfly
  which produces the DC term and AC term 4. ONLY the DC term requires
  the bias. Since the DC term is adding up the 8 inputs we put the
  bias of 8x128=1024 there.

*/

  dataptr = data;
  indataptr = inBlock;
  for (rowctr = DCTSIZE-1; rowctr >= 0; rowctr--) 
  {
    /* Do the 8 point butterfly */
    tmp0 = indataptr[0] + indataptr[7];
    tmp7 = indataptr[0] - indataptr[7];
    tmp1 = indataptr[1] + indataptr[6];
    tmp6 = indataptr[1] - indataptr[6];
    tmp2 = indataptr[2] + indataptr[5];
    tmp5 = indataptr[2] - indataptr[5];
    tmp3 = indataptr[3] + indataptr[4];
    tmp4 = indataptr[3] - indataptr[4];
    
    /* Even part per LL&M figure 1 --- note that published figure is faulty;
     * rotator "sqrt(2)*c1" should be "sqrt(2)*c6".
     */
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
   
/* 
  Put the bias subtraction into the DC term here.
  Everything else cancels out.
*/
    dataptr[0] = (DCTELEM) ((tmp10 + tmp11 - 1024) << PASS1_BITS);
    dataptr[4] = (DCTELEM) ((tmp10 - tmp11) << PASS1_BITS);
    
    z1 = (tmp12+tmp13);
    z1 = M_0_541196100(z1);

    dataptr[2] = (DCTELEM) DESCALE(z1 + M_0_765366865(tmp13) , 
				   CONST_BITS-PASS1_BITS);

    
    dataptr[6] = (DCTELEM) DESCALE(z1 - M_1_847759065(tmp12) , 
				   CONST_BITS-PASS1_BITS);
    
    /* Odd part per figure 8 --- note paper omits factor of sqrt(2).
     * cK represents cos(K*pi/16).
     * i0..i3 in the paper are tmp4..tmp7 here.
     */
    
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;

    z5 = z3 + z4;
    z5 = M_1_175875602(z5);

    tmp4 = M_0_298631336(tmp4);

    tmp5 = M_2_053119869(tmp5);

    tmp6 = M_3_072711026(tmp6);

    tmp7 = M_1_501321110(tmp7);

    z1 = - M_0_899976223(z1);

    z2 = - M_2_562915447(z2);

    z3 = - M_1_961570560(z3);

    z4 = - M_0_390180644(z4);

    
    z3 += z5;
    z4 += z5;
    
    dataptr[7] = (DCTELEM) DESCALE(tmp4 + z1 + z3, CONST_BITS-PASS1_BITS);
    dataptr[5] = (DCTELEM) DESCALE(tmp5 + z2 + z4, CONST_BITS-PASS1_BITS);
    dataptr[3] = (DCTELEM) DESCALE(tmp6 + z2 + z3, CONST_BITS-PASS1_BITS);
    dataptr[1] = (DCTELEM) DESCALE(tmp7 + z1 + z4, CONST_BITS-PASS1_BITS);
    
    dataptr += DCTSIZE;		/* advance pointer to next row */
    indataptr += DCTSIZE;
  }


  /* Pass 2: process columns. */
  /* Note that we must descale the results by a factor of 8 == 2**3, */
  /* and also undo the PASS1_BITS scaling. */

  dataptr = data;
  for (rowctr = DCTSIZE-1; rowctr >= 0; rowctr--) {
    tmp0 = dataptr[DCTSIZE*0] + dataptr[DCTSIZE*7];
    tmp7 = dataptr[DCTSIZE*0] - dataptr[DCTSIZE*7];
    tmp1 = dataptr[DCTSIZE*1] + dataptr[DCTSIZE*6];
    tmp6 = dataptr[DCTSIZE*1] - dataptr[DCTSIZE*6];
    tmp2 = dataptr[DCTSIZE*2] + dataptr[DCTSIZE*5];
    tmp5 = dataptr[DCTSIZE*2] - dataptr[DCTSIZE*5];
    tmp3 = dataptr[DCTSIZE*3] + dataptr[DCTSIZE*4];
    tmp4 = dataptr[DCTSIZE*3] - dataptr[DCTSIZE*4];
    
    /* Even part per LL&M figure 1 --- note that published figure is faulty;
     * rotator "sqrt(2)*c1" should be "sqrt(2)*c6".
     */
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    dataptr[DCTSIZE*0] = (DCTELEM) DESCALE(tmp10 + tmp11, PASS1_BITS+3);
    dataptr[DCTSIZE*4] = (DCTELEM) DESCALE(tmp10 - tmp11, PASS1_BITS+3);
    
    z1 = (tmp12+tmp13);
    z1 = M_0_541196100(z1);

    dataptr[DCTSIZE*2] = (DCTELEM) DESCALE(z1 + M_0_765366865(tmp13) , 
					   CONST_BITS+PASS1_BITS+3);

    
    dataptr[DCTSIZE*6] = (DCTELEM) DESCALE(z1 - M_1_847759065(tmp12) , 
                                           CONST_BITS+PASS1_BITS+3);
    
    /* Odd part per figure 8 --- note paper omits factor of sqrt(2).
     * cK represents cos(K*pi/16).
     * i0..i3 in the paper are tmp4..tmp7 here.
     */
    
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;

    z5 = z3 + z4;
    z5 = M_1_175875602(z5);

    tmp4 = M_0_298631336(tmp4);

    tmp5 = M_2_053119869(tmp5);

    tmp6 = M_3_072711026(tmp6);

    tmp7 = M_1_501321110(tmp7);

    z1 = - M_0_899976223(z1);

    z2 = - M_2_562915447(z2);

    z3 = - M_1_961570560(z3);

    z4 = - M_0_390180644(z4);
    
    z3 += z5;
    z4 += z5;
    
    dataptr[DCTSIZE*7] = (DCTELEM) DESCALE(tmp4 + z1 + z3,
					   CONST_BITS+PASS1_BITS+3);
    dataptr[DCTSIZE*5] = (DCTELEM) DESCALE(tmp5 + z2 + z4,
					   CONST_BITS+PASS1_BITS+3);
    dataptr[DCTSIZE*3] = (DCTELEM) DESCALE(tmp6 + z2 + z3,
					   CONST_BITS+PASS1_BITS+3);
    dataptr[DCTSIZE*1] = (DCTELEM) DESCALE(tmp7 + z1 + z4,
					   CONST_BITS+PASS1_BITS+3);
    
    dataptr++;			/* advance pointer to next column */
  }


  return ia_successful;
}
/*
$Log: jfwddct.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.7  1995/12/13 02:47:45  ejaquith
jfwddct.c -> /product/ipcore/ipshared/src/RCS/jfwddct.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1994/10/21  00:50:21  lperry
 * jfwddct.c -> /product/ipcore/ipshared/src/RCS/jfwddct.c,v
 *
 * Get rid of UINT32, etc., so these files can work with Peter's interpreter.
 *
 * Revision 1.5  1994/07/21  00:17:22  lperry
 * jfwddct.c -> /product/ipcore/ipshared/src/RCS/jfwddct.c,v
 *
 * Change error handling from old setjmp/longjmp to function return codes.
 *
 * Revision 1.4  1994/06/11  00:22:55  lperry
 * jfwddct.c -> /product/ipcore/ipshared/src/RCS/jfwddct.c,v
 *
 * Move bias subtraction into this routine by subtracting
 * 8 times the bias from the first (horizontal) DCT pass.
 * Also change interface to use separate input and output
 * 8x8 pixel arrays, rather than overwrite input. This was
 * necessary in order to move 8 bit data in and 16 bit
 * DCT coefficients out.
 *
 * Revision 1.3  1994/06/09  23:18:36  lperry
 * jfwddct.c -> /product/ipcore/ipshared/src/RCS/jfwddct.c,v
 *
 * Changed multiplication macros so that they now function
 * as if they were function calls. Used a comma expression
 * to handle intermediate results. Also renamed the macros
 * so that its obvious what constant they use.
 *
 * Revision 1.2  1994/06/09  00:29:42  lperry
 * lperry on Wed Jun 8 17:17:39 PDT 1994
 *
 * jfwddct.c -> /product/ipcore/ipshared/src/RCS/jfwddct.c,v
 *
 * Changed implementation of multiplies by constants to do
 * a series of shifts and adds. This is what gcc does using
 * a factoring algorithm which does better than N shifts and
 * N adds for a constant with N "one" bits.
 *
 * This change has only minimal effect on the Sun version, since
 * we're using gcc there anyway, but it beats Watcom's code by at least
 * 15 percent.
 *
 * Revision 1.1  1994/02/27  02:23:29  lperry
 * Initial revision
 *
*/
