/*
 * jmemnobs.c
 *
 * Copyright (C) 1992-1994, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file provides a really simple implementation of the system-
 * dependent portion of the JPEG memory manager.  This implementation
 * assumes that no backing-store files are needed: all required space
 * can be obtained from malloc().
 * This is very portable in the sense that it'll compile on almost anything,
 * but you'd better have lots of main memory (or virtual memory) if you want
 * to process big images.
 * Note that the max_memory_to_use option is ignored by this implementation.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"
#include "jmemsys.h"		/* import the system-dependent declarations */
#include "memory.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: jmemnobs.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");

#ifndef HAVE_STDLIB_H		/* <stdlib.h> should declare malloc(),free() */
extern void * malloc JPP((size_t size));
extern void free JPP((void *ptr));
#endif


/*
 * Memory allocation and freeing are controlled by the regular library
 * routines malloc() and free().
 *
 * NOTE:
 *   The two base allocation routines
 *     jpeg_get_small and jpeg_get_large
 *   are not error-checked with ERR_PASS() here.
 *   Instead, their return values are checked for
 *   NULL in the routines that call them, which are
 *   all in jmemmgr.c
 *
 */

GLOBAL ErrCode CDECL 
jpeg_get_small (
    IPState    *pIPState,
    j_common_ptr cinfo,
    size_t	sizeofobject,
    void      **memory)
{
  PROLOGUE;

  /* Keep compilers happy */
  cinfo = cinfo;

  *memory = (void *) IP_MALLOC(sizeofobject);

  NORMAL_RETURN;
}

GLOBAL ErrCode CDECL 
jpeg_free_small (
    IPState    *pIPState,
    j_common_ptr cinfo,
    void       *object,
    size_t	sizeofobject)
{
  PROLOGUE;

  /* Keep compilers happy */
  cinfo = cinfo;
  sizeofobject = sizeofobject;

  IP_FREE(object);

  NORMAL_RETURN; 
}


/*
 * "Large" objects are treated the same as "small" ones.
 * NB: although we include FAR keywords in the routine declarations,
 * this file won't actually work in 80x86 small/medium model; at least,
 * you probably won't be able to process useful-size images in only 64KB.
 */

GLOBAL ErrCode CDECL 
jpeg_get_large (
    IPState    *pIPState,
    j_common_ptr cinfo,
    size_t	sizeofobject,
    void      **memory)
{
  PROLOGUE;

  /* Keep compilers happy */
  cinfo = cinfo;

  *memory = (void *) IP_MALLOC(sizeofobject);

  NORMAL_RETURN;
}

GLOBAL ErrCode CDECL 
jpeg_free_large (
    IPState    *pIPState,
    j_common_ptr cinfo,
    void FAR   *object,
    size_t	sizeofobject)
{
  PROLOGUE;

  /* Keep compilers happy */
  cinfo = cinfo;
  sizeofobject = sizeofobject;

  IP_FREE(object);

  NORMAL_RETURN;
}


/*
 * This routine computes the total memory space available for allocation.
 * Here we always say, "we got all you want bud!"
 */

GLOBAL ErrCode CDECL
jpeg_mem_available (
    IPState    *pIPState,
    j_common_ptr cinfo,
    long	min_bytes_needed,
    long	max_bytes_needed,
    long	already_allocated,
    long       *pAvailable)
{
  PROLOGUE;

  /* Keep compilers happy */
  cinfo = cinfo;
  min_bytes_needed = min_bytes_needed;
  already_allocated = already_allocated;

     /* we always pretend to have enough. */
  *pAvailable = max_bytes_needed;

  NORMAL_RETURN;
}


/*
 * Backing store (temporary file) management.
 * Since jpeg_mem_available always promised the moon,
 * this should never be called and we can just error out.
 */
GLOBAL ErrCode CDECL 
jpeg_open_backing_store (
    IPState    *pIPState,
    j_common_ptr cinfo,
    backing_store_ptr info,
    long	total_bytes_needed)
{
  PROLOGUE;

  ERREXIT(cinfo, JERR_NO_BACKING_STORE);

  ERR_GEN(TRUE, JERR_NO_BACKING_STORE);

  /* Keep compilers happy */
  info = info;
  total_bytes_needed = total_bytes_needed;

  NORMAL_RETURN;
}


/*
 * These routines take care of any system-dependent initialization and
 * cleanup required.  Here, there isn't any.
 */

GLOBAL ErrCode CDECL
jpeg_mem_init (
    IPState    *pIPState,
    j_common_ptr cinfo,
    long       *pMax_memory_to_use)
{
  PROLOGUE;

  /* Keep compilers happy */
  cinfo = cinfo;

  *pMax_memory_to_use = 0;  /* just set max_memory_to_use to 0 */

  NORMAL_RETURN;
}

GLOBAL ErrCode CDECL 
jpeg_mem_term (
    IPState    *pIPState,
    j_common_ptr cinfo)
{
  PROLOGUE;

  /* Keep compilers happy */
  cinfo = cinfo;

  /* no work */

  NORMAL_RETURN;
}

/*
$Log: jmemnobs.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.4  1998/01/13 23:51:45  ddavies
jmemnobs.c -> /product/ipcore/ipshared/src/RCS/jmemnobs.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1996/04/23 17:52:41  xzhu
jmemnobs.c -> /product/ipcore/ipshared/src/RCS/jmemnobs.c,v

CommonCheck: Fixed compiler warning " defined, not referenced".

Revision 1.2  1996/04/22 17:31:22  xzhu
jmemnobs.c -> /product/ipcore/ipshared/src/RCS/jmemnobs.c,v

Revision 1.1  1996/04/20 01:42:32  xzhu
Initial revision

*/
