/*
 * jwrjfif.c
 *
 * Copyright (C) 1991, 1992, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains routines to write standard JPEG file headers/markers.
 * The file format created is a raw JPEG data stream with (optionally) an
 * APP0 marker per the JFIF spec.  This will handle baseline and
 * JFIF-convention JPEG files, although there is currently no provision
 * for inserting a thumbnail image in the JFIF header.
 *
 * These routines may need modification for non-Unix environments or
 * specialized applications.  As they stand, they assume output to
 * an ordinary stdio stream.  However, the changes to write to something
 * else are localized in the macros appearing just below.
 *
 * These routines are invoked via the methods write_file_header,
 * write_scan_header, write_jpeg_data, write_scan_trailer, and
 * write_file_trailer.
 */

#include "jpeg.h"
#include "jpeg.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jwrjfif.c,v $; $Revision: 1.2 $; $Date: 2000/06/09 23:23:23 $");

/*
 * To output to something other than a stdio stream, you'd need to redefine
 * these macros.
 */

/* Write a single byte (UInt8 ebyte must be defined in caller) */
#define emit_byte(x)  \
{ebyte = (UInt8) (x); \
*cinfo->writeCallback)(cinfo,&ebyte,(UInt32)1L, ncount);\
if (ncount != 1) \
    return JERR_WRITE; }


typedef enum {			/* JPEG marker codes */
  M_SOF0  = 0xc0,
  M_SOF1  = 0xc1,
  M_SOF2  = 0xc2,
  M_SOF3  = 0xc3,
  
  M_SOF5  = 0xc5,
  M_SOF6  = 0xc6,
  M_SOF7  = 0xc7,
  
  M_JPG   = 0xc8,
  M_SOF9  = 0xc9,
  M_SOF10 = 0xca,
  M_SOF11 = 0xcb,
  
  M_SOF13 = 0xcd,
  M_SOF14 = 0xce,
  M_SOF15 = 0xcf,
  
  M_DHT   = 0xc4,
  
  M_DAC   = 0xcc,
  
  M_RST0  = 0xd0,
  M_RST1  = 0xd1,
  M_RST2  = 0xd2,
  M_RST3  = 0xd3,
  M_RST4  = 0xd4,
  M_RST5  = 0xd5,
  M_RST6  = 0xd6,
  M_RST7  = 0xd7,
  
  M_SOI   = 0xd8,
  M_EOI   = 0xd9,
  M_SOS   = 0xda,
  M_DQT   = 0xdb,
  M_DNL   = 0xdc,
  M_DRI   = 0xdd,
  M_DHP   = 0xde,
  M_EXP   = 0xdf,
  
  M_APP0  = 0xe0,
  M_APP15 = 0xef,
  
  M_JPG0  = 0xf0,
  M_JPG13 = 0xfd,
  M_COM   = 0xfe,
  
  M_TEM   = 0x01,
  
  M_ERROR = 0x100
} JPEG_MARKER;


LOCAL Int32 CDECL
emit_marker (compress_info_ptr cinfo, JPEG_MARKER mark)
/* Emit a marker code */
{
  UInt8 ebyte;
  UInt32 *ncount;

  emit_byte(0xFF);

  emit_byte(mark);

  return ia_successful;
}


LOCAL Int32 CDECL
emit_2bytes(compress_info_ptr cinfo, Int32 value)
/* Emit a 2-byte integer; these are always MSB first in JPEG files */
{
  UInt8 ebyte;
  UInt32 *ncount;

  emit_byte((value >> 8) & 0xFF);
  emit_byte(value & 0xFF);

  return ia_successful;
}


LOCAL Int32 CDECL
emit_dqt (compress_info_ptr cinfo, Int32 index, Int32 *prec)
/* Emit a DQT marker */
/* Returns the precision used (0 = 8bits, 1 = 16bits) for baseline checking */
{
  QUANT_TBL_PTR data = cinfo->quant_tbl_ptrs[index];
  Int32 i;
  UInt8 ebyte;
  UInt32 *ncount;

  *prec = 0;
  
  for (i = 0; i < DCTSIZE2; i++) {
    if (data[i] > 255)
      *prec = 1;
  }

  CHECKERR(emit_marker(cinfo, M_DQT))
  
  CHECKERR(emit_2bytes(cinfo, *prec ? DCTSIZE2*2 + 1 + 2 : DCTSIZE2 + 1 + 2))
  
  emit_byte(index + ((*prec)<<4));
  
  for (i = 0; i < DCTSIZE2; i++) {
    if (*prec)
      emit_byte(data[i] >> 8);

    emit_byte(data[i] & 0xFF);
  }

  return ia_successful;
}


LOCAL Int32 CDECL
emit_dht (compress_info_ptr cinfo, Int32 index, boolean is_ac)
/* Emit a DHT marker */
{
  UInt8 ebyte;
  HUFF_TBL * htbl;
  Int32 length, i;
  
  if (is_ac) {
    htbl = cinfo->ac_huff_tbl_ptrs[index];
    index += 0x10;		/* output index has AC bit set */
  } else {
    htbl = cinfo->dc_huff_tbl_ptrs[index];
  }

  if (htbl == NULL)
    ERREXIT1(cinfo->emethods, "Huffman table 0x%02x was not defined", 
			 JERR_BADHUFF, index);
  
  if (! htbl->sent_table) {
    CHECKERR(emit_marker(cinfo, M_DHT))
    
    length = 0;
    for (i = 1; i <= 16; i++)
      length += htbl->bits[i];
    
    CHECKERR(emit_2bytes(cinfo, length + 2 + 1 + 16))
    emit_byte(index);
    
    for (i = 1; i <= 16; i++)
      emit_byte(htbl->bits[i]);
    
    for (i = 0; i < length; i++)
      emit_byte(htbl->huffval[i]);
    
    htbl->sent_table = TRUE;
  }

  return ia_successful;
}


LOCAL Int32 CDECL
emit_dac (compress_info_ptr cinfo)
/* Emit a DAC marker */
/* Since the useful info is so small, we want to emit all the tables in */
/* one DAC marker.  Therefore this routine does its own scan of the table. */
{
  UInt8 ebyte;
  Int8 dc_in_use[NUM_ARITH_TBLS];
  Int8 ac_in_use[NUM_ARITH_TBLS];
  UInt32 *ncount;
  Int32 length, i;
  
  for (i = 0; i < NUM_ARITH_TBLS; i++)
    dc_in_use[i] = ac_in_use[i] = 0;
  
  for (i = 0; i < cinfo->num_components; i++) {
    dc_in_use[cinfo->comp_info[i].dc_tbl_no] = 1;
    ac_in_use[cinfo->comp_info[i].ac_tbl_no] = 1;
  }
  
  length = 0;
  for (i = 0; i < NUM_ARITH_TBLS; i++)
    length += dc_in_use[i] + ac_in_use[i];
  
  CHECKERR(emit_marker(cinfo, M_DAC))
  
  CHECKERR(emit_2bytes(cinfo, length*2 + 2))
  
  for (i = 0; i < NUM_ARITH_TBLS; i++) {
    if (dc_in_use[i]) {
      emit_byte(i);
      emit_byte(cinfo->arith_dc_L[i] + (cinfo->arith_dc_U[i]<<4));
    }
    if (ac_in_use[i]) {
      emit_byte(i + 0x10);
      emit_byte(cinfo->arith_ac_K[i]);
    }
  }

  return ia_successful;
}


LOCAL Int32 CDECL
emit_dri (compress_info_ptr cinfo)
/* Emit a DRI marker */
{
  UInt32 *ncount;
  CHECKERR(emit_marker(cinfo, M_DRI))
  
  CHECKERR(emit_2bytes(cinfo, 4))	/* fixed length */

  CHECKERR(emit_2bytes(cinfo, (Int32) cinfo->restart_interval))

  return ia_successful;
}


LOCAL Int32 CDECL
emit_sof (compress_info_ptr cinfo, JPEG_MARKER code)
/* Emit a SOF marker */
{
  UInt8 ebyte;
  Int32 i;
  UInt32 *ncount;
  
  CHECKERR(emit_marker(cinfo, code))
  
  CHECKERR(emit_2bytes(cinfo, 3 * cinfo->num_components + 2 + 5 + 1)) /* length */

  if (cinfo->image_height > 65535L || cinfo->image_width > 65535L)
    ERREXIT(cinfo->emethods, 
			"Maximum image dimension for JFIF is 65535 pixels",
			JERR_IMAGETOOBIG);

  emit_byte(cinfo->data_precision);
  CHECKERR(emit_2bytes(cinfo, (Int32) cinfo->image_height))
  CHECKERR(emit_2bytes(cinfo, (Int32) cinfo->image_width))

  emit_byte(cinfo->num_components);

  for (i = 0; i < cinfo->num_components; i++) {
    emit_byte(cinfo->comp_info[i].component_id);
    emit_byte((cinfo->comp_info[i].h_samp_factor << 4)
		     + cinfo->comp_info[i].v_samp_factor);
    emit_byte(cinfo->comp_info[i].quant_tbl_no);
  }

  return ia_successful;
}


LOCAL Int32 CDECL
emit_sos (compress_info_ptr cinfo)
/* Emit a SOS marker */
{
  UInt8 ebyte;
  Int32 i;
  UInt32 *ncount;
  
  CHECKERR(emit_marker(cinfo, M_SOS))
  
  CHECKERR(emit_2bytes(cinfo, 2 * cinfo->comps_in_scan + 2 + 1 + 3)) /* length */
  
  emit_byte(cinfo->comps_in_scan);
  
  for (i = 0; i < cinfo->comps_in_scan; i++) {
    emit_byte(cinfo->cur_comp_info[i]->component_id);
    emit_byte((cinfo->cur_comp_info[i]->dc_tbl_no << 4)
		     + cinfo->cur_comp_info[i]->ac_tbl_no);
  }

  emit_byte(0);		/* Spectral selection start */
  emit_byte(DCTSIZE2-1);	/* Spectral selection end */
  emit_byte(0);		/* Successive approximation */

  return ia_successful;
}


LOCAL Int32 CDECL
emit_jfif_app0 (compress_info_ptr cinfo)
/* Emit a JFIF-compliant APP0 marker */
{
  UInt8 ebyte;
  UInt32 *ncount;
  /*
   * Length of APP0 block	(2 bytes)
   * Block ID			(4 bytes - ASCII "JFIF")
   * Zero byte			(1 byte to terminate the ID string)
   * Version Major, Minor	(2 bytes - 0x01, 0x01)
   * Units			(1 byte - 0x00 = none, 0x01 = inch, 0x02 = cm)
   * Xdpu			(2 bytes - dots per unit horizontal)
   * Ydpu			(2 bytes - dots per unit vertical)
   * Thumbnail X size		(1 byte)
   * Thumbnail Y size		(1 byte)
   */
  
  CHECKERR(emit_marker(cinfo, M_APP0))
  
  CHECKERR(emit_2bytes(cinfo, 2 + 4 + 1 + 2 + 1 + 2 + 2 + 1 + 1)) /* length */

  emit_byte(0x4A);	/* Identifier: ASCII "JFIF" */
  emit_byte(0x46);
  emit_byte(0x49);
  emit_byte(0x46);
  emit_byte(0);
  emit_byte(1);		/* Major version */
  emit_byte(1);		/* Minor version */
  emit_byte(cinfo->density_unit); /* Pixel size information */
  CHECKERR(emit_2bytes(cinfo, (Int32) cinfo->X_density))
  CHECKERR(emit_2bytes(cinfo, (Int32) cinfo->Y_density))
  emit_byte(0);		/* No thumbnail image */
  emit_byte(0);

  return ia_successful;
}


/*
 * Write the file header.
 */


METHODDEF Int32 CDECL
write_file_header (compress_info_ptr cinfo)
{
  Int8 qt_in_use[NUM_QUANT_TBLS];
  Int32 i, prec;
  boolean is_baseline;
  Int32 dqt_prec;
  
  CHECKERR(emit_marker(cinfo, M_SOI))	/* first the SOI */

  if (cinfo->write_JFIF_header)	/* next an optional JFIF APP0 */
    CHECKERR(emit_jfif_app0(cinfo))

  /* Emit DQT for each quantization table. */
  /* Note that doing it here means we can't adjust the QTs on-the-fly. */
  /* If we did want to do that, we'd have a problem with checking precision */
  /* for the is_baseline determination. */

  for (i = 0; i < NUM_QUANT_TBLS; i++)
    qt_in_use[i] = 0;

  for (i = 0; i < cinfo->num_components; i++)
    qt_in_use[cinfo->comp_info[i].quant_tbl_no] = 1;

  prec = 0;
  for (i = 0; i < NUM_QUANT_TBLS; i++) {
    if (qt_in_use[i])
    {
      CHECKERR(emit_dqt(cinfo, i, &dqt_prec))
      prec += dqt_prec;
    }
  }
  /* now prec is nonzero iff there are any 16-bit quant tables. */

  /* Check for a non-baseline specification. */
  /* Note we assume that Huffman table numbers won't be changed later. */
  is_baseline = TRUE;
  if (cinfo->arith_code || (cinfo->data_precision != 8))
    is_baseline = FALSE;
  for (i = 0; i < cinfo->num_components; i++) {
    if (cinfo->comp_info[i].dc_tbl_no > 1 || cinfo->comp_info[i].ac_tbl_no > 1)
      is_baseline = FALSE;
  }
  if (prec && is_baseline) {
    is_baseline = FALSE;
    /* If it's baseline except for quantizer size, warn the user */
    TRACEMS(cinfo->emethods, 0,
	    "Caution: quantization tables are too coarse for baseline JPEG",
		JINFO_BADQT);
  }


  /* Emit the proper SOF marker */
  if (cinfo->arith_code)
    CHECKERR(emit_sof(cinfo, M_SOF9)) /* SOF code for arithmetic coding */
  else if (is_baseline)
    CHECKERR(emit_sof(cinfo, M_SOF0)) /* SOF code for baseline implementation */
  else
    CHECKERR(emit_sof(cinfo, M_SOF1))	/* SOF code for non-baseline Huffman file */

  return ia_successful;
}


/*
 * Write the start of a scan (everything through the SOS marker).
 */

METHODDEF Int32 CDECL
write_scan_header (compress_info_ptr cinfo)
{
  Int32 i;

  if (cinfo->arith_code) {
    /* Emit arith conditioning info.  We will have some duplication
     * if the file has multiple scans, but it's so small it's hardly
     * worth worrying about.
     */
    CHECKERR(emit_dac(cinfo))
  } else {
    /* Emit Huffman tables.  Note that emit_dht takes care of
     * suppressing duplicate tables.
     */
    for (i = 0; i < cinfo->comps_in_scan; i++) {
      CHECKERR(emit_dht(cinfo, cinfo->cur_comp_info[i]->dc_tbl_no, FALSE))
      CHECKERR(emit_dht(cinfo, cinfo->cur_comp_info[i]->ac_tbl_no, TRUE))
    }
  }

  /* Emit DRI if required --- note that DRI value could change for each scan.
   * If it doesn't, a tiny amount of space is wasted in multiple-scan files.
   * We assume DRI will never be nonzero for one scan and zero for a later one.
   */
  if (cinfo->restart_interval)
    CHECKERR(emit_dri(cinfo))

  CHECKERR(emit_sos(cinfo))

  return ia_successful;
}


/*
 * Write some bytes of compressed data within a scan.
 */

METHODDEF Int32 CDECL
write_jpeg_data (compress_info_ptr cinfo, Int8 *dataptr, Int32 datacount)
{
/* Write multiple bytes from a buffer */
if ( (*cinfo->writeCallback)
     (cinfo, (UInt8 *)(dataptr), (Int32)(datacount))
	  != (Int32) (datacount) ) 
    return JERR_WRITE;

    return ia_successful;
}


/*
 * Finish up after a compressed scan (series of write_jpeg_data calls).
 */

METHODDEF Int32 CDECL
write_scan_trailer (compress_info_ptr cinfo)
{
    MENTION(cinfo)
  /* no work needed in this format */
    return ia_successful;
}


/*
 * Finish up at the end of the file.
 */

METHODDEF Int32 CDECL
write_file_trailer (compress_info_ptr cinfo)
{
  CHECKERR(emit_marker(cinfo, M_EOI))

  return ia_successful;
}


/*
 * The method selection routine for standard JPEG header writing.
 * This should be called from c_ui_method_selection if appropriate.
 */

GLOBAL Int32 CDECL
jselwjfif (compress_info_ptr cinfo)
{
  cinfo->cmethods->write_file_header = write_file_header;
  cinfo->cmethods->write_scan_header = write_scan_header;
  cinfo->cmethods->write_jpeg_data = write_jpeg_data;
  cinfo->cmethods->write_scan_trailer = write_scan_trailer;
  cinfo->cmethods->write_file_trailer = write_file_trailer;

  return ia_successful;
}

/*
$Log: jwrjfif.c,v $
Revision 1.2  2000/06/09 23:23:23  jlin
fix read write callback function return type for jpeg

Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.6  1995/12/13 02:47:59  ejaquith
jwrjfif.c -> /product/ipcore/ipshared/src/RCS/jwrjfif.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1994/10/21  00:50:42  lperry
 * jwrjfif.c -> /product/ipcore/ipshared/src/RCS/jwrjfif.c,v
 *
 * Get rid of UINT32, etc., so these files can work with Peter's interpreter.
 *
 * Revision 1.4  1994/07/21  00:17:50  lperry
 * jwrjfif.c -> /product/ipcore/ipshared/src/RCS/jwrjfif.c,v
 *
 * Change error handling from old setjmp/longjmp to function return codes.
 *
 * Revision 1.3  1994/07/16  00:01:37  lperry
 * lperry on Fri Jul 15 17:00:59 PDT 1994
 *
 * jwrjfif.c -> /product/ipcore/ipshared/src/RCS/jwrjfif.c,v
 *
 * Added type casts to args of write callback function.
 *
 * Revision 1.2  1994/06/27  17:02:38  lperry
 * jwrjfif.c -> /product/ipcore/ipshared/src/RCS/jwrjfif.c,v
 *
 * Moved static variables to the "cinfo" structure to achieve reentrancy.
 *
 * Revision 1.1  1994/02/27  02:24:47  lperry
 * Initial revision
 *
*/
