/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  miccptr.c:	Utility routines used to fix up pointers in memory blocks
 *		after DAE garbage collection.
 *
 *	Int32	ip_fixupScanLinePtrs()
 *			Fix pointers inside memory block holding TCScanLineRuns
 *
 *	Int32	ip_fixupCCSymbolPtrs()
 *			Fix pointers inside memory block holding
 *			CCSymbols
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"

    /*  prototyes */
#include "shrcc.pub"
#include "shrcc.prv"
#include "ptrfixup.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: miccptr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");


/* PRIVATE */
/**************************************************************************
 *   ip_fixupScanLinePtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a heap memory block
 *	containing TCScanLineRuns.
 *
 *	Input:
 *		pThisBlock:	address of the TCCompressionInfoBlock.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in every TCScanLineRun struct. to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupScanLinePtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
    PROLOGUE;
    TCScanLineRun	*pRun;
    TCBlockHeader	*pBlock;
    UInt32	numOfStructs;
    UInt32	oldAddr;
    UInt32	addrDiff;

	/* The TCScanLineRun structures are kept in (string of) memory block(s).
	 * There are two ways to look at each structure, either it's on
	 * the free list or it's in use.  In either case, the first four
	 * bytes are a pointer so can be freely corrected.
	 * The number of structures in the memory block is in the
	 * numOfStructs entry of the TCBlockHeader.
	 */
	/* correct the type of the block pointer */
    pBlock = (TCBlockHeader *) pThisBlock;
    oldAddr = (UInt32) pOldAddr;

	/* Fixup pointers in the TCBlockHeader itself */
    addrDiff = ((UInt32)(pBlock->pNext) - oldAddr);
    if (addrDiff < blkLength)
	pBlock->pNext = (TCBlockHeader *) (addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pBlock->pTop) - oldAddr);
    if (addrDiff < blkLength)
	pBlock->pTop = (void *) (addrDiff + (UInt32)pNewAddr);

	/* get the number of structures in this block */
    numOfStructs = pBlock->numOfStructs;
    pRun = (TCScanLineRun *) ((UInt32)pBlock + sizeof(TCBlockHeader));

	/* loop through the array of structures, inspecting each pointer */
    for (; numOfStructs > 0; numOfStructs--, pRun++)
    {
	addrDiff = ((UInt32)(pRun->pNext) - oldAddr);
	if (addrDiff < blkLength)
	    pRun->pNext = (TCScanLineRun *) (addrDiff + (UInt32)pNewAddr);

	/* The pClient pointer is currently used to point to something,
	 * so it should be checked, or to hold a couple of flags.  This
	 * happens during symbol-based compression.  When
	 * it's holding flags, it clearly should NOT be updated.  However,
	 * The only flag values are very small numbers, so the page
	 * portion of the pointer will be 0 in this case.  This is equivalent
	 * to the NULL pointers, so we'll never decide that the flags are
	 * pointing to the block that moved, so we'll never change them.
	 */
	addrDiff = ((UInt32)(pRun->pClient) - oldAddr);
	if (addrDiff < blkLength)
	    pRun->pClient = (void *) (addrDiff + (UInt32)pNewAddr);
    }

    NORMAL_RETURN;
}



/* PRIVATE */
/**************************************************************************
 *   ip_fixupCCSymbolPtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a heap memory block
 *	containing CCSymbols.
 *
 *	Input:
 *		pThisBlock:	address of the TCCompressionInfoBlock.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in every CCSymbols struct. to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupCCSymbolPtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
    PROLOGUE;
     CCSymbol	*pSymbol;
    TCBlockHeader	*pBlock;
    UInt32	numOfStructs;
    UInt32	oldAddr;
    UInt32	addrDiff;

	/* The CCSymbols structures are kept in (string of) memory block(s).
	 * There are two ways to look at each structure, either it's on
	 * the free list or it's in use.  In either case, the first four
	 * bytes are a pointer so can be freely corrected.
	 * The number of structures in the memory block is in the
	 * numOfStructs entry of the TCBlockHeader.
	 */
	/* correct the type of the block pointer */
    pBlock = (TCBlockHeader *) pThisBlock;
    oldAddr = (UInt32) pOldAddr;

	/* Fixup pointers in the TCBlockHeader itself */
    addrDiff = ((UInt32)(pBlock->pNext) - oldAddr);
    if (addrDiff < blkLength)
	pBlock->pNext = (TCBlockHeader *) (addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pBlock->pTop) - oldAddr);
    if (addrDiff < blkLength)
	pBlock->pTop = (void *) (addrDiff + (UInt32)pNewAddr);

	/* get the number of structures in this block */
    numOfStructs = pBlock->numOfStructs;
    pSymbol = (CCSymbol *) ((UInt32)pBlock + sizeof(TCBlockHeader));

	/* loop through the array of structures, inspecting each pointer */
    for (; numOfStructs > 0; numOfStructs--, pSymbol++)
    {
	addrDiff = ((UInt32)(pSymbol->pNewSymbol) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pNewSymbol = (CCSymbol *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pRunHead) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pRunHead = (TCScanLineRun *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pRunTail) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pRunTail = (TCScanLineRun *) (addrDiff + (UInt32)pNewAddr);
    }

    NORMAL_RETURN;
}




/*

 $Log: miccptr.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.7  1998/01/13 23:51:50  ddavies
 miccptr.c -> /product/ipcore/ipshared/src/RCS/miccptr.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.6  1997/04/24 23:33:36  ddavies
 miccptr.c -> /product/ipcore/ipshared/src/RCS/miccptr.c,v

 Change ip_fixupSymbolPtrs into ip_fixupCCSymbolPtrs, which has many fewer
 pointers to fix up.  Change all the TCSymbols to CCSymbols.

 Revision 1.5  1996/07/27 02:03:11  ddavies
 miccptr.c -> /product/ipcore/ipshared/src/RCS/miccptr.c,v

 Add fixup for new pNextDictSym ptr in a TCSymbol.

 Revision 1.4  1996/05/01 23:12:55  ddavies
 miccptr.c -> /product/ipcore/ipshared/src/RCS/miccptr.c,v

 Change includes of shrcmprs.pub and shrcmprs.prv to shrcc.pub and shrcc.prv.
 These are the real files we want to include anyway.  They were included
 indirectly before.

 Revision 1.3  1995/12/13 02:48:13  ejaquith
 miccptr.c -> /product/ipcore/ipshared/src/RCS/miccptr.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/07/14  02:28:47  ddavies
 * ddavies on Thu Jul 13 19:26:54 PDT 1995
 *
 * miccptr.c -> /product/ipcore/ipshared/src/RCS/miccptr.c,v
 *
 * Fixed bugs in pointer fixup routines.  They were supposed to fix all pointers
 * in heap memory blocks, but they didn't fix the heap pointers themselves.  They
 * were only fixing the pointers of the structures in the heaps.
 * This applies to ip_fixupScanLinePtrs and ip_fixupSymbolPtrs.
 *
 * Revision 1.1  1995/07/10  23:05:35  ddavies
 * Initial revision
 *

 */

