/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  miccttut.c: utility routines for Modified Read and Modified Modified
 *		Read compression.
 *
 *	Int32	ip_writeCCITTEOL()
 *			Insert an EndOfLine code into the data stream.
 *	Int32	ip_skipCCITTEOL()
 *			Skip bits until an EndOfLine code has been seen.
 *	Int32	ip_flushCCITT()
 *			Flush encoding pipeline at end of large block.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "mmrenc.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"
#include "shrcmprs.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: miccttut.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* This is a set of utility routines used in CCITT MR and MMR compression.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_writeCCITTEOL()
 *
 *	This procedure writes out the end of line code.  It may follow
 *	the EOL with a bit indicating the encoding type of the next
 *	scan line.
 *
 *	Input:
 *		pCCITTCompInfo:	pointer to compression state.
 *		padType:	If equal to cPadToByte, insert pad before
 *				EOL so that EOL ends on a byte boundary
 *				If equal to cNoPad, do not
 *				insert padding before EOL.
 *		EOLType:	If equal to cEOL1D, follow the EOL with
 *				a bit indicating that the next line will
 *				be encoded with 1D encoding.  If equal to
 *				cEOL2D, append a bit indicating that the
 *				next line will be encoded using 2D
 *				codes.  If cPlainEOL, do not append a bit.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *
 *	Action:
 *		If pCCITTCompInfo->compType == cCCITTCompNewLine (we're really
 *		at the end of a line), attempt to output an EOL code.
 *		Padding may conditionally be inserted
 *		before the beginning of the EOL.  A bit indicating
 *		the encoding type of the next line may be appended.
 *
 *		If we stop because the compressed data buffer fills,
 *		set pCCITTCompInfo->compState to the state which
 *		describes the compressed data byte we're about to write and
 *		put the relevant byte counts and run pointers into
 *		*pCCITTCompInfo.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be written.  Callers can use this
 *		to determine the number of bytes written so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_bufferExceeded	ran out of compressed data buffer
 *
 **************************************************************************/
Int32  CDECL 
ip_writeCCITTEOL (
    IPState    *pIPState,
    CCITTCompressionInfo *pCCITTCompInfo,
    UInt32	padType,
    UInt32	EOLType,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    UInt32	 accum;			/* accumulated encoded bits */
    Int32	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pCCITTCompInfo->accumBits;
    vacantBit = pCCITTCompInfo->vacantBit;

	/* Start off in the proper state */
    switch (pCCITTCompInfo->compState)
    {
    case cCCITTCompNewLine:
	goto compCCITTStartEOL;

    case cCCITTCompFinishEOLPad:
	goto compCCITTStartEOL;

    case cCCITTCompFinishEOL:
	goto compCCITTFlushPad;

    default:	/* restart state belongs to some other routine */
	NORMAL_RETURN;
    }

	/* flush the accumulator to make room for the pad if necessary.
	 */
compCCITTStartEOL:
	if (vacantBit >= 16)
	{
	    if (pCurCompData + 1 >= pCompDataEnd)
	    {
		pCCITTCompInfo->compState = cCCITTCompFinishEOLPad;
		goto compCCITTEOLMightWriteOne;
	    }

		/* write out LS 2 bytes of accumulator and increment
		 * pointers */
	    pCurCompData[0] = reverseByte[(UInt8) accum];
	    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
	    accum >>= 16;			/* get rid of written bytes */
	    pCurCompData += 2;
	    vacantBit -= 16;
	}

	/* See if we're supposed to insert padding before the EOL so that
	 * it ends on a byte boundary.
	 */
    if (padType == cPadToByte)
    {
	    /* The EOL code is 12 bits long.  To make sure it ends on a
	     * multiple of 8 bits, we must start it on an odd multiple
	     * of 4 bits, i.e. at some point where vacantBit == 4.
	     * There are several cases:
	     *	1. 4 >= vacantBit >= 0
	     *		set vacantBit = 4.
	     *	2. 12 >= vacantBit > 4
	     *		set vacantBit = 12.
	     *  3. 15 >= vacantBit > 12
	     *		set vacantBit = 20.
	     * The cute formula below maps numbers in [0..4] to
	     * [3..7], ANDs those to 0 and adds 4, which is the
	     * right answer.  It maps numbers in [5..12] to 8,
	     * adds 4 and gets 12.  It maps [13..15] to 16, adds
	     * 4 and gets 20. 
	     */
	vacantBit = ((vacantBit + 3) & ~0x7) + 4;
    }

compCCITTFlushPad:
    if (vacantBit >= 16)
    {
	if (pCurCompData + 1 >= pCompDataEnd)
	{
	    pCCITTCompInfo->compState = cCCITTCompFinishEOL;
	    goto compCCITTEOLMightWriteOne;
	}

	    /* write out LS 2 bytes of accumulator and increment
	     * pointers */
	pCurCompData[0] = reverseByte[(UInt8) accum];
	pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
	accum >>= 16;			/* get rid of written bytes */
	pCurCompData += 2;
	vacantBit -= 16;
    }

	/* Having inserted the pad if necessary, insert the EOL itself */
    accum |= (cEOLCode) << vacantBit;
    vacantBit += cEOLLength;

	/* are we supposed to follow this with an indicator bit? */
    switch (EOLType)
    {
    case cPlainEOL:
	break;		/* no indicator bit */
    case cEOL1D:
	accum |= cMR1DFlag << vacantBit;	/* and drop through to
						 * record bit's presence */
    case cEOL2D:
	vacantBit++;
	break;
    default:
	return ia_invalidParm;	/* remember, we haven't really changed any
				 * state yet */
    }

	/* Swell, we've inserted the EOL.  Update the saved state do we'll
	 * remember.
	 */
   *ppCompData = pCurCompData;
   pCCITTCompInfo->compState = cCCITTCompNewLine;
   pCCITTCompInfo->accumBits = accum;
   pCCITTCompInfo->vacantBit = vacantBit;

   NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

	/* We get here when we know that two bytes won't fit into the
	 * compressed output buffer.  Try putting one in.
	 */
compCCITTEOLMightWriteOne:
    if (pCurCompData < pCompDataEnd)
    {
	pCurCompData[0] = reverseByte[(UInt8) accum];
	pCurCompData++;
	accum >>= 8;
	vacantBit -= 8;
    }

	/* This is where we end up when we abort because we ran out
	 * of space in the compression buffer.
	 */
    *ppCompData = pCurCompData;
    pCCITTCompInfo->accumBits = accum;
    pCCITTCompInfo->vacantBit = vacantBit;
    
    return ia_bufferExceeded;	/* This is how we say that we weren't able
				 * to finish.
				 */

}


/* PRIVATE */
/**************************************************************************
 *   ip_skipCCITTEOL()
 *
 *	This procedure skips over incoming bits until it has skipped
 *	over an EndOfLine code.  If there's an optional 1D/2D flag,
 *	it skips that too and returns its value.
 *
 *	Input:
 *		pCCITTDecompInfo: pointer to decompression state.
 *		flagExists:	If equal to cTrue, the EOL is
 *				followed by an optional 1D/2D flag bit.
 *				Its state should be returned in EOLType.
 *				Otherwise, there is not flag bit.
 *		flushPipeline:	If cTrue, this routine should be able to
 *				get to the end of the EOL using
 *				only data in the compressed data buffer.
 *				If set to cFalse, it's ok to stop and
 *				ask for more compressed data when we're
 *				starting to run out.  Additional data will
 *				be seamlessly concatenated to current data.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *		pEOLType:	pointer to returned type of EOL found.
 *				This is only used if flagExists is true.
 *
 *	Action:
 *		Attempt to scan the incoming data stream until an
 *		EndOfLine code is found.  If flagExists is set, also
 *		scan the next bit.  If it's set, return *pEOLType =
 *		cEOL1D.  If the next bit is clear, return *pEOLType =
 *		cEOL2D.  If flagExists is not set, do not touch *pEOLType.
 *.
 *		The compressed data is assumed to be in natural byte order and
 *		is arbitrarily aligned, so should be read 1 byte at a
 *		time.
 *
 *		If we stop because the compressed data buffer empties,
 *		set pCCITTDecompInfo->decompState to the state which
 *		describes where we should restart and
 *		put the relevant byte counts and run pointers into
 *		*pCCITTDecompInfo.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be read.  Callers can use this
 *		to determine the number of bytes read so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pProc was NULL
 *		ia_bufferExceeded	stopped because we tried to read
 *					something to the compressed data
 *					buffer but it was empty
 *
 **************************************************************************/
Int32  CDECL 
ip_skipCCITTEOL (
    IPState    *pIPState,
    CCITTDecompressionInfo *pCCITTDecompInfo,
    Bool	flagExists,
    Bool	flushPipeline,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd,
    UInt32     *pEOLType)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    UInt32	 accum;			/* holds bits to be decoded */
    Int8	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */
    Int32        numBitsNeeded;         /* The number of bits we need to find
					 * the EOL code and an optional
					 * flag bit */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pCCITTDecompInfo->accumBits;
    vacantBit = (Int8) pCCITTDecompInfo->vacantBit;

    /* REALLY UGLY WORKAROUND, 11/20/98: The "Class B Writer
     * Version 1.0 Tue Feb 06 14:48:28 1990 (c) Copyright Wang
     * Laboratories, Inc. 1990 " neglects to put
     * EOLs at the beginning of all strips *except* the first one.
     * Since we're being aggressive about skipping over the glop
     * at the beginning of the line, we skip the entire first line
     * until we get to its EOL.  Reading the strip, noting that we're
     * one line short, then going back to try again seems awkward.
     * Well, how do we tell the difference between nice data at
     * the beginning of the line and trash?  Our ugly workaround
     * depends on the fact that our fax packages seem to all put
     * pad bits in front of the EOLs.  We'll say that if this
     * is the first line AND there are no pad bits AND the first
     * byte is non-zero, then just return, pretending to have found
     * the EOL.
     */
    if ((pCCITTDecompInfo->lineNumber == 0) &&
        (vacantBit == 0) &&
        (pCCITTDecompInfo->padType == cNoPad) &&
	(pCurCompData[0] != 0))
	NORMAL_RETURN;

	/* Most of time time, we want to compare to
	 * pCompDataEnd-1, so decrement it here */
    pCompDataEnd--;

        /* If there's no flag bit, we need cEOLLength bits.  Otherwise,
	 * we need cEOLLength + 1 bits.
	 */
    numBitsNeeded = cEOLLength;
    if (flagExists)
	numBitsNeeded++;

	/* see where we should start off.  Yeah, in this case the two
	 * labels (decompEOLNewEOL and decompEOLRestart) could really
	 * be the same.  I was trying for consistency with the
	 * other decoding routines.
	 * We have to have a private abort state (cCCITTDecompEOLAbort
	 * instead of just returning cCCITTDecompNewLine when
	 * we run out of data) to remember that this was the
	 * spot where we ran out of data.  That's how we know
	 * to restart here.
	 */
    switch (pCCITTDecompInfo->decompState)
    {
    case cCCITTDecompNewLine:
	goto decompEOLNewEOL;

    case cCCITTDecompEOLAbort:
	goto decompEOLRestart;

    default:	/* Restart state belongs to some other routine */
	NORMAL_RETURN;
    }

decompEOLNewEOL:

	/* Here's the main loop.  We see if there are at least 16
	 * bits in the accumulator, attempting to reload if not,
	 * the see if the bottom 12 bits equal the EOL code.
	 * If not, shift and try again.  Since we're possibly
	 * going to try to read 13 bits, we can do 3 comparison/shifts
	 * (16 - 13) before trying to refill the accumulator.
	 */
    while (cTrue)
    {
decompEOLRestart:
	if (vacantBit <= 16)
	{
	    if (pCurCompData < pCompDataEnd)
	    {
		accum |= reverseByte[pCurCompData[0]] << vacantBit;
		vacantBit += 8;
		accum |= reverseByte[pCurCompData[1]] << vacantBit;
		vacantBit += 8;
		pCurCompData += 2;
	    }
	    else
		/* We're not allowed to fetch 2 new bytes.
		 * If there's one byte, get it.
		 */
	    if (pCurCompData == pCompDataEnd)
	    {
		accum |= reverseByte[pCurCompData[0]] << vacantBit;
		vacantBit += 8;
		pCurCompData++;
	    }
	    else
	        /* There are no more bytes to fetch.  If this isn't
		 * the end of the data stream, abort now and get them.
		 */
	    if (!flushPipeline)
		goto decompEOLAbort;
	    else
	        /* Hmmm, well, we'd like to get some more bytes, but
		 * there aren't any more.  If we really need them,
		 * die.  Otherwise, press on.
		 */
	    if (vacantBit < numBitsNeeded)
		goto decompEOLDecodeFailure;	    
	}

	    /* Bug Fix, 10/19/98.  We used to die if we found 16 or
	     * more 0's in a row.  It's true that there are no legal
	     * codes containing 16 zeros, BUT, fax machines can insert
	     * an arbitrary number of 0's as pad bits before the EOL.
	     * At this point, we should either have the EOL or some
	     * zeros followed by the EOL.  If we have some other
	     * pattern, we should die now.
	     * Using a scheme from Dave Abercrombie, shift and test the
	     * contents of the accumulator in a loop.  The loop exits
	     * when either we've found some non-zero pattern in the
	     * bits where the EOL should be or we have fewer than
	     * numBitsNeeded in the accumulator.
	     * Next Bug Fix, 10/27/98: Some horrible FAX packages put
	     * dreck at the beginning of their data buffers.  Be more
	     * aggressive about skipping EVERYTHING until an EOL is
	     * found.
	     */
	while ((vacantBit >= numBitsNeeded) &&
	       ((accum & cEOLMask) != cEOLCode))
	{
	    accum >>= 1;
	    vacantBit--;
	}

	    /* If we don't have enough bits, loop back up and get some
	     * more.  Checking for this first is a little subtle.
	     * Say we found some pattern other than 0s before an EOL.
	     * We'll properly fall through this test and die below.
	     * Say we found an EOL followed by a flag bit, and both are in the
	     * accumulator now.  We'll also fall through this test and break
	     * out of the loop happily below.
	     * Say we found an EOL, but the flag bit isn't in the accum
	     * yet.  In this case, vacantBit will equal numBitsNeeded-1.
	     * Then we'll loop back to get the flag bit in the accum-filling
	     * code, fail the test at the top of the shifting loop, and
	     * fall through this test happily without shifting accum.
	     * Say we're still finding a bunch of 0s.  We'll just loop
	     * back to get more bits.
	     */
	if (vacantBit < numBitsNeeded)
	    continue;

	    /* If we found the EOL code, go look for the flag. We're
	     * guaranteed that it's in accum now since we wouldn't be
	     * here if vacantBit were less than numBitsNeeded.  Otherwise,
	     * we found some dreck, so complain.
	     */
	if ((accum & cEOLMask) == cEOLCode)
	    break;
	else
	    goto decompEOLDecodeFailure;
    }

	/* If we get here, we found the EOL.  Yea!  If there's an
	 * optional flag bit, record it's state and discard it.
	 */
    if (flagExists)
    {
	if (accum & (cEOLMask + 1))	/* look above the EOL */
	    *pEOLType = cEOL1D;
	else
	    *pEOLType = cEOL2D;

	    /* Discard the EOL and optional flag */
	accum >>= cEOLLength + 1;
	vacantBit -= cEOLLength + 1;
    }
    else	/* just discard the EOL itself */
    {
	accum >>= cEOLLength;
	vacantBit -= cEOLLength;
    }

	/* it worked, so update the saved state and return */
    *ppCompData = pCurCompData;
    pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
    pCCITTDecompInfo->accumBits = accum;
    pCCITTDecompInfo->vacantBit = vacantBit;

    NORMAL_RETURN;

	/****************** ABORT *******************/

	/* We get here if we were about to have a decode failure.
	 * This only happens when accum == 0.
	 * If flushPipeline is set, then we really just ran out
	 * of data, so return ia_bufferExceeded instead.
	 * If the upper level sees an ia_bufferExceeded
	 * with flushPipeline, it knows that the client didn't
	 * supply enough compressed data and we're doomed.
	 */
decompEOLDecodeFailure:
    if (flushPipeline)
	return ia_bufferExceeded;
    else
	return ia_decodeFailure;

	/* First, we have all the abort target.  We do this weird
	 * thing of setting all the states here instead of above
	 * in the hopes that the compiler will generate better
	 * code in the real loops - just a test a jump down here
	 * where the normal case is straight ahead instead of
	 * a test and jump around the error stuff.
	 */
decompEOLAbort:
    pCCITTDecompInfo->decompState = cCCITTDecompEOLAbort;
    *ppCompData = pCurCompData;
    pCCITTDecompInfo->accumBits = accum;
    pCCITTDecompInfo->vacantBit = vacantBit;
    
    return ia_bufferExceeded;	/* indicates that we stopped because we
				 * ran out of bytes.
				 */

}



/* PRIVATE */
/**************************************************************************
 *   ip_flushCCITT()
 *
 *	This procedure flushes out the CCITT pipeline during encoding.
 *
 *	Input:
 *		pCCITTCompInfo:	pointer to compression state.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *
 *	Action:
 *		This routine simply attempts to empty the bits in
 *		the pipeline to the compressed data stream.  It
 *		is normally called at the end of a large section,
 *		like a tile or strip.
 *
 *		If we stop because the compressed data buffer fills,
 *		set pCCITTCompInfo->compState to the state which
 *		describes the compressed data byte we're about to write and
 *		put the relevant byte counts and run pointers into
 *		*pCCITTCompInfo.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be written.  Callers can use this
 *		to determine the number of bytes written so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_bufferExceeded	wasn't enough room to finish flushing
 *
 **************************************************************************/
Int32  CDECL 
ip_flushCCITT (
    IPState    *pIPState,
    CCITTCompressionInfo *pCCITTCompInfo,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    UInt32	 accum;			/* accumulated encoded bits */
    Int32	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pCCITTCompInfo->accumBits;
    vacantBit = pCCITTCompInfo->vacantBit;

	/* see where we should start off */
    switch (pCCITTCompInfo->compState)
    {
    case cCCITTCompNewLine:
	goto compCCITTSetFlushAmount;

    case cCCITTCompFinishTotalFlush:
	goto compCCITTFinishFlush;

    default:	/* not one of our restart codes */
	NORMAL_RETURN;
    }

compCCITTSetFlushAmount:
	/* Adjust vacantBit so we flush to the next byte boundary. */
    vacantBit = (vacantBit + 7) & ~0x7;

compCCITTFinishFlush:
	    /* flush the EOL */
    while (vacantBit > 0)
    {
	if (pCurCompData >= pCompDataEnd)
	    goto compCCITTAbortFlush;

	    /* write out LS byte of accumulator and increment
	     * pointers */
	pCurCompData[0] = reverseByte[(UInt8) accum];
	accum >>= 8;			/* get rid of written bytes */
	pCurCompData++;
	vacantBit -= 8;
    }

	/* Swell, we've flushed the accumulator.  Update the saved
	 * state do we'll remember.
	 */
   *ppCompData = pCurCompData;
   pCCITTCompInfo->compState = cCCITTCompNewLine;
   pCCITTCompInfo->accumBits = accum;
   pCCITTCompInfo->vacantBit = vacantBit;

   NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

compCCITTAbortFlush:
    pCCITTCompInfo->compState = cCCITTCompFinishTotalFlush;
	/* This is where we end up when we abort because we ran out
	 * of space in the compression buffer.
	 */
    *ppCompData = pCurCompData;
    pCCITTCompInfo->accumBits = accum;
    pCCITTCompInfo->vacantBit = vacantBit;
    
    return ia_bufferExceeded;	/* This is how we say that we weren't able
				 * to finish.
				 */

}


/*

 $Log: miccttut.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.9  1998/11/24 22:29:01  aber
 aber on Tue Nov 24 14:25:33 PST 1998

 miccttut.c -> /product/ipcore/ipshared/src/RCS/miccttut.c,v

 Fix from Dan:
 Really cheesy workaround for bug in Class B Writer Version 1.0 Tue Feb
 06 14:48:28 1990 (c) Copyright Wang Laboratories, Inc. 1990.
 One is supposed to encode every strip of a TIFF file as though it were a
 separate image. That means putting an EOL at the beginning of every strip
 when using Group 3 encoding.  They put an EOL at the beginning of the first
 strip, but not at the beginning of any of the subsequent strips.
 Unfortunately, fax machines use Group 3 and put trash at the beginning of
 strips (sometimes).  We were aggressively discarding the first line of
 all strips after the first one in our quest for the first EOL.  Leffler's
 package does the same thing.
 The really cheesy workaround depends on the observation that fax machines
 (especially the ones that put trash at the beginning of the strip) have
 their EOLs padded so they end on a byte boundary.  The EOLs in this file
 are not padded.  So, when looking for an EOL, we simply return if we're
 at the beginning of the strip and the EOLs are not padded and the first
 byte is non-zero.  We pretend to have found the EOL in this case.
 THIS SHOULD BE TESTED ON MORE FAX IMAGES!!

 Revision 1.8  1998/11/12 00:40:09  aber
 aber on Wed Nov 11 16:40:00 PST 1998

 miccttut.c -> /product/ipcore/ipshared/src/RCS/miccttut.c,v

 gcc doesn't like question marks in the comments.

 Revision 1.7  1998/11/12 00:20:42  aber
 aber on Wed Nov 11 16:19:28 PST 1998

 miccttut.c -> /product/ipcore/ipshared/src/RCS/miccttut.c,v

 Synced fixes from ddavies:
 Since some horrible fax packages put trash before the beginning of the
 first data line (and between lines), get more aggressive about skipping
 all bits until an EOL code is found.  Don't just die if we find something
 besides 0's before the EOL.  I think Dave Abercrombie said to do this
 awhile ago.  It does fix a problem with an image from the DAE fax package.

 Revision 1.6  1998/10/21 20:57:46  aber
 aber on Wed Oct 21 13:53:55 PDT 1998

 miccttut.c -> /product/ipcore/ipshared/src/RCS/miccttut.c,v

 After discussing the 0-padding issue with ddavies, we've decided
 that the fax spec allows for any number of padded zeros, and that the
 TIFF spec doesn't say how many, implying that any number is okay also.
 So we have a solution that allows reading a truely arbitrary number of
 padded zeros before the EOL itself. My last fix really checked the next
 byte or two.

 Revision 1.5  1998/10/14 18:00:34  aber
 aber on Wed Oct 14 10:57:57 PDT 1998

 miccttut.c -> /product/ipcore/ipshared/src/RCS/miccttut.c,v

 Added a kludge to deal with arbitrary lengths of 0-padding just before
 the EOL. This was a case presented by Tim Platt for MR data stored
 directly from a fax device, where the padding exists. He claims that
 0-padding is allowed in the CCITT 3 spec (T.4).

 Revision 1.4  1998/01/13 23:51:52  ddavies
 miccttut.c -> /product/ipcore/ipshared/src/RCS/miccttut.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1996/05/11 01:04:29  ddavies
 miccttut.c -> /product/ipcore/ipshared/src/RCS/miccttut.c,v

 Fix comments.
 Add ip_skipCCITTEOL.  This routine is used to skip over an EndOfLine code
 that preceeds (!) every scan line in CCITT Group 3.  It also returns the
 type of the EOL if we're using the 2D option.  Like other decoding routines,
 it can run out of compressed bytes and be restarted.

 Revision 1.2  1996/05/01 23:15:02  ddavies
 miccttut.c -> /product/ipcore/ipshared/src/RCS/miccttut.c,v

 Change preliminary implementation of ip_writeCCITTEOL so that it checks
 to see if there's enough room in the pipeline for new data before
 putting the new data in.  The old code checked after putting in data to
 see if there was enough data to write out.  The new way is consistent with
 ip_compMHRow and ip_compMMRRow.

 Revision 1.1  1996/04/10 22:06:09  ddavies
 Initial revision


 */

