/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  micmpgry.c: machine independent routine for comparing 2 gray images,
 *		producing a binary image
 *
 *          gray comparison:
 *                  void	ip_compareGrayLineScaleP1()
 *                  void	ip_compareGrayLineScaleP2()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrip.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: micmpgry.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* These procedures are used to compare single scan lines from two gray
 * images, producing a scan line from a binary image.  At every position
 * where the pixel in image1 * multiplier mult1 is strictly greater than
 * the pixel in image2 times mult2, we put a 1 in the binary image.
 *
 * This algorithm called by i_compareGrayImages, which is used for page
 * segmentation.  Pixels are tagged with the length of the horizontal
 * or vertical run they're in by some previous operation.  This routine 
 * is used to find pixels that are mainly in horizontal or mainly in
 * vertical runs.  See pageseg.c in the XIS code.
 *
 * Currently, the only legal depth for the gray images is 8 bit/pixel.  We'll
 * leave a depth argument for the future.
 *
 * The multipliers are clearly a problem.  Without them, we could compare
 * four pixels at once.  Too bad.  Instead, we do one of two comparisons:
 * 	Original comparison: mult1 * p1 > mult2 * p2
 *	New comparison:
 *		if mult1 >= mult2  p1 > mult2/mult1 * p2
 *		if mult1 < mult2   mult1/mult2 * p1 > p2
 *
 * The calling routine has already built a table of either mult1/mult2
 * or mult2/mult1.  The table is indexed with a pixel value to implement
 * the multiply.  The result is another byte, which we use in the comparison.
 * ip_compareGrayLineScaleP1 scales pixel1, ip_compareGrayLineScaleP2 scales
 * pixel2.
 *
 * Note that since we are comparing for mult1*p1 strictly greater than mult2*p2,
 * we can change which pixel indexes the table, but we can't change the
 * order of the operands in the subtraction.  If we change the order of the
 * operands and invert the results, everything works for the cases where
 * mult1*p1 != mult2*p2, but fails when they are equal.  The inverse of
 * (mult1*p1 > mult2*p2) is (mult1*p1 <= mult2*p2), not (mult1*p1 < mult2*p2).
 *
 * The XIS code this routine replaces finds the ends of a runs of 1's in the
 * output image, then fills in the middle.  To deal with each pixel then you'd
 * need a couple of reads to get the pixels, another read for the lookup table,
 * a subtract and a branch.  Say we just insert the MS bit of the subtract into
 * the binary result.  This has the same reads and subtract, followed by two
 * shifts and an OR.  Or, if we process the line from right to left, we can
 * get away with 1 shift, a mask operation to get the MS bit of the result
 * of the subtract, and an OR.  Since Mask ops are cheaper than shifts on a
 * PC, this is slightly preferable.
 *
 * At least it takes this much stuff in C.  In assembly, we can use the carry
 * flag to replace the two shifts and OR with a single ADC (add with carry)
 * operation that does the shift and inserts the new bit all in one glorious
 * (1 cycle) operation.  In C, the extra overhead of inserting
 * the new bit each time is offset by the overhead of calling the procedures
 * that find the beginning and end of each run and actually creating the runs.
 * In Assembly, it's no contest.
 *
 * Ok, ok, the scheme that creates the runs can just skip over the white parts
 * of the image, so it gains something there.  If there is lots of white
 * space in the result, this might count for something.  Of course, we
 * do spend time clearing out the binary image beforehand, so this isn't a
 * complete gain.  We'll stick with the new scheme and tell Luc he can
 * skip the step that clears out the image.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_compareGrayLineScaleP1()  
 *		Takes pointers to two gray source images, a scaling table
 *		and a binary dest image.  For every pixel position, if the
 *		pixel in image1 scaled according to the scale table is
 *		strictly greater than the pixel in image2, write a 1 in the
 *		output image.
 *
 *	Input:
 *		pS1 =		pointer gray source image 1
 *		pS2 =		pointer to gray source image 2
 *		pD =		pointer to binary dest image
 *		pScaleTable =	ptr to table used to do scaling
 *		w =		width of all images in pixels.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_compareGrayLineScaleP1 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt8      *pD,
    UInt8      *pScaleTable,
    UInt32	w)
{
    PROLOGUE;
    UInt32	 oWord;			/* the output word being produced */
    UInt32	 worthlessVar;		/* used to get around WATCOM compiler
					 * bug */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* we'll process from right to left because it's a little easier
	 * to insert the MS bit of the comparison at the MS end of oWord
	 * than at the LS end of oWord.  This means we will process the
	 * rightmost output word fragment first, then do the rest of the
	 * line in units of 32 pixels.  Note that as soon as bitCount
	 * becomes a multiple of 32, the righthand fragment is already
	 * correctly positioned in oWord.
	 * Both input gray images have 8 bit/pixel, so we can address
	 * pixels directly.
	 */
    oWord = 0;

	/* To be correct, we need to read pS1[w-1] and pS2[w-1].  However,
	 * to be efficient, we would like to use w instead of w-1.  If
	 * we predecrement w, we'd need to test to see if (w & 0x1F) != 0x1F
	 * instead of testing to see if (w & 0x1F) != 0.  This test is
	 * a little longer.  Since we're reading the source a byte at at
	 * time anyway, we just pre-decrement the source pointers so
	 * the proper bytes are read.
	 */
    pS1--;
    pS2--;
    while (w  > 0)
    {
	    /* Fill one output word.  Put the decrement and check at
	     * the end of the loop (unlike a "for" statement) so we
	     * won't write a blank word if w was originally a multiple of 32.
	     */
	do
	{
		/* This operation produces a 1 whenever the scaled pixel in pS1
		 * is GREATER than the pixel in pS2.
		 */
	    oWord = (oWord >> 1) | 
		    ( ((UInt32) UCHAR_ACCESS(&pS2[w]) - 
		       (UInt32) pScaleTable[UCHAR_ACCESS(&pS1[w])]) &
		      0x80000000);
		/* put the decrement and check at the end of the loop
		 * (unlike a "for" statement) so we won't write a blank
		 * word if w was originally a multiple of 32
		 */
	    w--;
	    worthlessVar = w & 0x1F;
	} while (worthlessVar != 0);

	ULONG_ACCESS(&pD[w/8]) = oWord;
    }
}



/* PRIVATE */
/**************************************************************************
 *  ip_compareGrayLineScaleP2()  
 *		Takes pointers to two gray source images, a scaling table
 *		and a binary dest image.  For every pixel position, if the
 *		pixel in image1 is strictly greater than the pixel in 
 *		image2 scaled according to the scale table, write a 1 in the
 *		output image.
 *
 *	Input:
 *		pS1 =		pointer gray source image 1
 *		pS2 =		pointer to gray source image 2
 *		pD =		pointer to binary dest image
 *		pScaleTable =	ptr to table used to do scaling
 *		w =		width of all images in pixels.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_compareGrayLineScaleP2 (
    IPState    *pIPState,
    UInt8      *pS1,
    UInt8      *pS2,
    UInt8      *pD,
    UInt8      *pScaleTable,
    UInt32	w)
{
    PROLOGUE;
    UInt32	oWord;			/* the output word being produced */
    UInt32	worthlessVar;		/* used to get around WATCOM compiler
					 * bug */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* we'll process from right to left because it's a little easier
	 * to insert the MS bit of the comparison at the MS end of oWord
	 * than at the LS end of oWord.  This means we will process the
	 * rightmost output word fragment first, then do the rest of the
	 * line in units of 32 pixels.  Note that as soon as bitCount
	 * becomes a multiple of 32, the righthand fragment is already
	 * correctly positioned in oWord.
	 * Both input gray images have 8 bit/pixel, so we can address
	 * pixels directly.
	 */
    oWord = 0;

	/* To be correct, we need to read pS1[w-1] and pS2[w-1].  However,
	 * to be efficient, we would like to use w instead of w-1.  If
	 * we predecrement w, we'd need to test to see if (w & 0x1F) != 0x1F
	 * instead of testing to see if (w & 0x1F) != 0.  This test is
	 * a little longer.  Since we're reading the source a byte at at
	 * time anyway, we just pre-decrement the source pointers so
	 * the proper bytes are read.
	 */
    pS1--;
    pS2--;
    while (w  > 0)
    {
	    /* Fill one output word.  Put the decrement and check at
	     * the end of the loop (unlike a "for" statement) so we
	     * won't write a blank word if w was originally a multiple of 32.
	     * What's the deal with worthlessVar?  In version 9.01e of the
	     * WATCOM compiler, leaving out worthlessVar (just testing w),
	     * causes the compiler to generate bad code.  It tries to test
	     * the loop condition by setting a private variable to
	     * pS2 + 0x1F (which is meaningless), then decrements pS1 and
	     * pS1 each time through the loop, testing to see if
	     * pS2 every equals the stored variable.  Since pS2 is going
	     * down and the variable is higher than its original value,
	     * we crash in flames.  Somehow, adding worthlessVar
	     * causes it to lower the level of optimization to something
	     * it can handle.
	     */
	do
	{
		/* This operation produces a 1 whenever the pixel in pS1
		 * is GREATER than the scaled pixel in pS2.
		 */
	    oWord = (oWord >> 1) | 
		    ( ((UInt32) pScaleTable[UCHAR_ACCESS(&pS2[w])] - 
		       (UInt32) UCHAR_ACCESS(&pS1[w])) &
		      0x80000000);

	    w--;
	    worthlessVar = w & 0x1F;
	} while (worthlessVar != 0);

	ULONG_ACCESS(&pD[w/8]) = oWord;
    }
}

/*

$Log: micmpgry.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.7  1998/01/13 23:51:53  ddavies
micmpgry.c -> /product/ipcore/ipshared/src/RCS/micmpgry.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1995/12/13 02:48:17  ejaquith
micmpgry.c -> /product/ipcore/ipshared/src/RCS/micmpgry.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/10/18  04:02:50  ddavies
 * micmpgry.c -> /product/ipcore/ipshared/src/RCS/micmpgry.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.4  1994/02/04  20:18:00  dferrell
 * micmpgry.c -> /product/ipcore/ipshared/src/RCS/micmpgry.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:24:41  ddavies
 * micmpgry.c -> /product/ipcore/ipshared/src/RCS/micmpgry.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:49:29  ddavies
 * micmpgry.c -> /product/ipcore/ipshared/src/RCS/micmpgry.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


