/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  micntbit.c: counts the number of bits in an image.
 *
 *           Int32     	i_countBitsLine()
 *
 */


#include <stddef.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: micntbit.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/****************************************************************************
 *  ip_countBitsLine():  count the number of 1's in a rectangle of an image.
 *
 *	Input:
 *		pS:		Pointer to source image data
 *		w:		width of source rectangle in words
 *		leftSideMask:	used to mask word on left of rect.
 *		rightSideMask:	used to mask word on right side of rect.
 *		pCountTab:	pointer to UInt8 table containing number
 *				of 1's in a byte.
 *		pCount:		pointer to returned count
 *
 *	Action:
 *		puts number of bits in line into *pCount
 *
 *	Returns:
 *		ia_successful, ...
 *
 ************************************************************************/
ErrCode CDECL
ip_countBitsLine(
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    UInt32	leftSideMask,
    UInt32	rightSideMask,
    const UInt8	*pCountTab,
    Int32       *pCount)
{
    PROLOGUE;
    UInt8	*pSLine;
    UInt32	 inputWord;
    Int32	 total;

	/* We'll process the line from right to left to save incrementing
	 * a variable.
	 */
    pSLine = pS + (w-1)*4;
    inputWord = ULONG_ACCESS(pSLine) & rightSideMask;

	/* if the rectangle is only 1 word wide, mask this word with
	 * the left side mask too.
	 */
    if (w == 1)
    {
	inputWord &= leftSideMask;
	total = pCountTab[inputWord & 0xFF];		/* byte 3 */
	inputWord >>= 8;
	total += pCountTab[inputWord & 0xFF];		/* byte 2 */
	inputWord >>= 8;
	total += pCountTab[inputWord & 0xFF];		/* byte 1 */
	inputWord >>= 8;
	total += pCountTab[inputWord];			/* byte 0 */
	*pCount = total;
	NORMAL_RETURN;
    }

	/* process the right side word */
    total = pCountTab[inputWord & 0xFF];		/* byte 3 */
    inputWord >>= 8;
    total += pCountTab[inputWord & 0xFF];		/* byte 2 */
    inputWord >>= 8;
    total += pCountTab[inputWord & 0xFF];		/* byte 1 */
    inputWord >>= 8;
    total += pCountTab[inputWord];			/* byte 0 */

	/* process all the bytes in the middle of the line.  We
	 * do this a byte at a time just because there's no particular
	 * advantage to reading this stuff a word at a time and
	 * doing all the shifts and masks.
	 */
    for (pSLine -= 4; pSLine > pS; pSLine -= 4)
    {
	total += pCountTab[pSLine[0]];
	total += pCountTab[pSLine[1]];
	total += pCountTab[pSLine[2]];
	total += pCountTab[pSLine[3]];
    }

	/* Do the word on the left */
    inputWord = ULONG_ACCESS(pSLine) & leftSideMask;
    total += pCountTab[inputWord & 0xFF];		/* byte 3 */
    inputWord >>= 8;
    total += pCountTab[inputWord & 0xFF];		/* byte 2 */
    inputWord >>= 8;
    total += pCountTab[inputWord & 0xFF];		/* byte 1 */
    inputWord >>= 8;
    total += pCountTab[inputWord];			/* byte 0 */

    *pCount = total;
    NORMAL_RETURN;
}


/*

 $Log: micntbit.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.4  1998/01/13 23:51:55  ddavies
 micntbit.c -> /product/ipcore/ipshared/src/RCS/micntbit.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1995/12/13 02:48:22  ejaquith
 micntbit.c -> /product/ipcore/ipshared/src/RCS/micntbit.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1993/12/02  19:24:45  ddavies
 * micntbit.c -> /product/ipcore/ipshared/src/RCS/micntbit.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.1  1993/11/16  03:12:20  ddavies
 * Initial revision
 *

 */

