/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  micyclsd.c: machine independent routines for seed placement in
 *		cycles of a given length
 *
 *	Int32	ip_horizCycleSeedLine()
 *
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"

/* prototypes */
#include "shrcc.pub"
#include "shrcc.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: micyclsd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* PRIVATE */
/**************************************************************************
 *  ip_horizCycleSeedLine()  
 *		Seeds cycles of lengths [minCycleThresh..maxCycleThresh]
 *
 *	Input:
 *		sImage:		Pointer to source image
 *		rightOffset:	Byte offset of word containing right end
 *				of scan line.
 *		rightWordMask:	Used to mark right end of rightmost word.
 *		dImage		Pointer to destination image
 *		pByteSeed	seed pixels for current byte if cycle from
 *				previous byte is not included
 *		pByteEndSeed	seed pixels for current byte if cycle from
 *				previous byte must be included
 *		pCyclesOnEndsBytePtr pointer to table with info about
 *				the cycles that span bytes
 *		minCycleThresh	the minimum cycle length to be seeded
 *		maxCycleThresh	the maximum cycle length to be seeded
 *
 *	Return code:
 *		ia_successful	successful comltetion
 *
 **************************************************************************/
Int32  CDECL 
ip_horizCycleSeedLine (
    IPState    *pIPState,
    UInt8      *sImage,
    UInt32	rightOffset,
    UInt32	rightWordMask,
    UInt8      *dImage,
    UInt8      *pByteSeed,
    UInt8      *pByteEndSeed,
    Int8       *pCyclesOnEndsBytePtr,
    Int32	minCycleThresh,
    Int32	maxCycleThresh)
{
    PROLOGUE;
    UInt32	iWord;		/* the binary input word */
    UInt32	prevOWord;	/* previous destination word */
    UInt32	oword = 0;	/* the binary output word */
    UInt8      *pCurSImage;	/* pointer to word being read */
    UInt8      *pCurDImage;	/* pointer to word being written */
    UInt32	trialByte;	/* used to hold each byte being examined */
    Int32	curRunLength;	/* length of the run we're currently
				 * tracking between bytes or words */
    Int32	byteRunLength;	/* amount by which a run extends into a byte */

	/* point to the right end of the scan line */
    pCurSImage = sImage + rightOffset;
    pCurDImage = dImage + rightOffset;

	/* Get this word and extend the rightmost run to the end
	 * since we're always supposed to ignore it.
	 */
    iWord = ULONG_ACCESS(pCurSImage) & rightWordMask;
    trialByte = 0;	/* the word off the right end of the image */
    prevOWord = oword = 0;

	/* We want to test for "end of run"  A run ends where there's
	 * a 0 to the left of a 1.  So we see if (iWord << 1) &
	 * ~iWord is zero.  If this is true, there are NO ends of
	 * runs in this word.
	 */
    while ((((trialByte | (iWord << 1)) & (~iWord)) == 0) &&
	   (pCurSImage > sImage))
    {
    	ULONG_ACCESS(pCurDImage) = 0;
	pCurSImage -= 4;			/* point to the next word */
	pCurDImage -= 4;			/* point to the next word */
	trialByte = iWord >> 31;		/* remember MS bit of
						 * prev word */
	iWord = ULONG_ACCESS(pCurSImage);	/* and fetch next word */
    }

	/* we've now either found a word with the end of the rightmost
	 * run in it, or have found the leftmost word in the scan line,
	 * or both.  Go find all the runs in it.  Since we always ignore
	 * the last cycle in each line, we'll always ignore curRunLength
	 * at the bottom of the loop.  If we find the left word at
	 * this point, we'll skip over the loop, then ignore curRunLength,
	 * so nothing will be recorded when the line contains only one run.
	 */

    curRunLength = -1000000;	/* big negative number */

    if ((trialByte | (iWord << 1)) & (~iWord))
	/* there's something in here, so go find the runs, ignoring
	 * the first one.  We find the first one by initializing
	 * curRunLength to a big negative number, and testing it
	 * before each update to see if it's still negative */
    {
	    /* do byte 3, assume the last bit from the previous word
	     * was a 1.
	     */
	trialByte = ((iWord << (cRunAddrPos+1)) |
		     (trialByte << cRunAddrPos)) & cRunAddrMask;
	byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	if (byteRunLength < 0)	/* we've found the end of the first run */
	{
		/* there might be other runs inside this byte */
	    oword = *((UInt16 *)&(pByteSeed[trialByte])) >> 1;

		/* get length of run extending from this byte */
	    curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	}

	    /* do byte 2 */
	trialByte = (iWord >> (8-cRunAddrPos-1)) & cRunAddrMask;
	byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	if (byteRunLength < 0)	/* we've found the end of a run */
	{
	        /* if this is not the first run (curRunLength > 0) and
		 * the curRunLength is withing than the given threshold,
		 * put a seed bit into the output word and get any other
		 * cycles inside this byte.
		 */
	    if ((curRunLength > 0)  &&
		((curRunLength + (~byteRunLength)) <= maxCycleThresh) &&
		((curRunLength + (~byteRunLength)) >= minCycleThresh))
		oword |= *((UInt16 *)&(pByteEndSeed[trialByte])) << 7;
	    else oword |= *((UInt16 *)&(pByteSeed[trialByte])) << 7;

		/* get length of run on left extending from this byte */
	    curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	}
	else	/* the run just extends through this byte */
	    curRunLength += 8;

	    /* do byte 1 */
	trialByte = (iWord >> (16-cRunAddrPos-1)) & cRunAddrMask;
	byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	if (byteRunLength < 0)	/* we've found the end of a run */
	{
	    if ((curRunLength > 0) && 
		((curRunLength + (~byteRunLength)) <= maxCycleThresh) &&
		((curRunLength + (~byteRunLength)) >= minCycleThresh))
		oword |= *((UInt16 *)&(pByteEndSeed[trialByte])) << 15;
	    else oword |= *((UInt16 *)&(pByteSeed[trialByte])) << 15;

		/* get length of run on left extending from this byte */
	    curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	}
	else	/* the run just extends through this byte */
	    curRunLength += 8;

	    /* do byte 0 */
	trialByte = (iWord >> (24-cRunAddrPos-1)) & cRunAddrMask;
	byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	if (byteRunLength < 0)	/* we've found the end of a run */
	{
	    if ((curRunLength > 0) &&
		((curRunLength + (~byteRunLength)) <= maxCycleThresh) &&
		((curRunLength + (~byteRunLength)) >= minCycleThresh))
		oword |= *((UInt16 *)&(pByteEndSeed[trialByte])) << 23;
	    else oword |= *((UInt16 *)&(pByteSeed[trialByte])) << 23;

		/* get length of run on left extending from this byte */
	    curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	}
	else	/* the run just extends through this byte */
	    curRunLength += 8;
    }	/* finished processing word containing end of rightmost run */

	/* Since we've just processed the data in iWord, get another word */
    pCurSImage -= 4;
    trialByte = iWord >> 31;		/* remember contents of
					 * prev word */
    	/* we cannot write the destination word at this time.  the 
 	 * most significant bit of the previous source word could have 
	 * been the end of the cycle.  we need to be able to right a
	 * seed pixel in the corresponding destination word so we'll 
	 * remain one destination word behind the source when writing.
	 */
    prevOWord = oword;

	/* if there's another word in this scan line, get it */
    if (pCurSImage >= sImage)
	iWord = ULONG_ACCESS(pCurSImage);

	/* At this point, we can be in one of three places.
	 *  1.	iWord contains the word at the right end of the line,
	 *  2.	iWord contains some word to the left of the rightmost
	 *	word, possibly the leftmost word.
	 *  3.	iWord contains the word to the left of the leftmost word.
	 *
	 * In all cases, pCurSImage points to the memory word that
	 * corresponds to iWord and trialByte contains the previous
	 * word in the line.
	 *
	 * We first work our way to the left until we get to the first
	 * word.
	 */
    while (pCurSImage >= sImage)
    {

	    /* Does a cycle end in the current word? If not, just make the
	     * previous run (if any) longer.
	     */
	oword = 0;
	if (((trialByte | (iWord << 1)) & (~iWord)) == 0)
	{
	    curRunLength += 32;		/* extend the run */
    	    ULONG_ACCESS(pCurDImage) = prevOWord;
    	    pCurDImage -= 4;
	}
	else	/* runs start and/or stop in here, so look at all
		 * four bytes.
		 */
	{
		/* do byte 3 */
	    trialByte = ((iWord << (cRunAddrPos + 1)) |
			 (trialByte << cRunAddrPos)) & cRunAddrMask;
	    byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of the first
					 * run */
	    {
	        if (((curRunLength + (~byteRunLength)) <= maxCycleThresh)	&&
		    ((curRunLength + (~byteRunLength)) >= minCycleThresh))
		{
		    oword |= *((UInt16 *)&(pByteEndSeed[trialByte])) >> 1;
		    prevOWord |= *((UInt16 *)&(pByteEndSeed[trialByte])) << 31;
    		    ULONG_ACCESS(pCurDImage) = prevOWord;
    		    pCurDImage -= 4;
		}
	        else 
		{
		    oword |= *((UInt16 *)&(pByteSeed[trialByte])) >> 1;
		    prevOWord |= *((UInt16 *)&(pByteSeed[trialByte])) << 31;
    		    ULONG_ACCESS(pCurDImage) = prevOWord;
    		    pCurDImage -= 4;
		}
		    /* get length of run on left extending from this byte */
		curRunLength = 	
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else		/* the run extends through this byte */
	    {
		ULONG_ACCESS(pCurDImage) = prevOWord;
    		pCurDImage -= 4;
		curRunLength += 8;
	    }

		/* do byte 2 */
	    trialByte = (iWord >> (8-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
	        if (((curRunLength + (~byteRunLength)) <= maxCycleThresh)	&&
		    ((curRunLength + (~byteRunLength)) >= minCycleThresh))
		    oword |= *((UInt16 *)&(pByteEndSeed[trialByte])) << 7;
	        else oword |= *((UInt16 *)&(pByteSeed[trialByte])) << 7;

		    /* get length of run on left extending from this byte */
		curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;

		/* do byte 1 */
	    trialByte = (iWord >> (16-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
	        if (((curRunLength + (~byteRunLength)) <= maxCycleThresh)	&&
		    ((curRunLength + (~byteRunLength)) >= minCycleThresh))
		    oword |= *((UInt16 *)&(pByteEndSeed[trialByte])) << 15;
	        else oword |= *((UInt16 *)&(pByteSeed[trialByte])) << 15;

		    /* get length of run on left extending from this byte */
		curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;

		/* do byte 0 */
	    trialByte = (iWord >> (24-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
	        if (((curRunLength + (~byteRunLength)) <= maxCycleThresh)	&&
		    ((curRunLength + (~byteRunLength)) >= minCycleThresh))
		    oword |= *((UInt16 *)&(pByteEndSeed[trialByte])) << 23;
	        else oword |= *((UInt16 *)&(pByteSeed[trialByte])) << 23;

		    /* get length of run on left extending from this byte */
		curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;
	}	/* finished processing all four bytes of a word */

	    /* Since we've just processed the data in iWord, get another word */
	pCurSImage -= 4;
	trialByte = iWord >> 31;		/* remember contents of
						 * prev word */
	prevOWord = oword;
	    /* if there's another word in this scan line, get it */
	if (pCurSImage >= sImage)
	    iWord = ULONG_ACCESS(pCurSImage);
    }

	/* At this point, we have just processed the leftmost word.
	 * curRunLength holds the length of the leftmost run, if any.
	 * Since we never count the last run when counting cycles,
	 * ignore curRunLength.
	 */

	/* save the current destination word */
    ULONG_ACCESS(pCurDImage) = oword;
    NORMAL_RETURN;

}




/*

$Log: micyclsd.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.3  1998/01/13 23:51:56  ddavies
micyclsd.c -> /product/ipcore/ipshared/src/RCS/micyclsd.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1995/12/13 02:48:25  ejaquith
micyclsd.c -> /product/ipcore/ipshared/src/RCS/micyclsd.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1995/04/14  01:21:45  jcraig
 * Initial revision
 *

*/


