/*************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * midepth1.c -- machine independent primitives for procedures that change
 *		 the depth of images.
 *
 *     Grayscale thresholding:
 *           Int32     ip_toBinary128Line()
 *	     Int32     ip_toBinaryFrom2bppLine()
 *	     Int32     ip_toBinaryFrom4bppLine()
 *           Int32     ip_toTwoBitsPerPixelLine()
 *           Int32     ip_toBinaryThreshLine()
 *	     Int32     ip_toBinaryFrom4bppThreshLine()
 *           Int32     ip_toBinaryBandPassLine()
 */



#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "imageref.pub"

    /* prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: midepth1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

#define cByteMask	0xFF
#define cMSBitMask	0x80808080
#define cMS2BitsMask	0xC0C0C0C0
#define cMSBitMask4bpp  0x88888888
#define cMSBitMask2bpp	0xAAAAAAAA


/*-----------------------------------------------------------------------*
 *             Grayscale Thresholding -- C Version			 *
 *-----------------------------------------------------------------------*/

/* PRIVATE */
/*************************************************************************** 
 *  ip_toBinary128Line()
 *  		Takes an 8-bit grayscale image and returns
 *             	a thresholded binary image, using the most significant
 *             	bit in each 8-bit pixel as the corresponding binary pixel.
 *	Input:
 *		pS =   		Pointer to left end of source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		tabBinByte0 =	Pointer to table used to un-scramble
 *				byte 0 of intermediate binarized 32-bit word.
 *		tabBinByte1 =	Ditto for byte 1.
 *		tabBinByte2 =	Ditto for byte 2.
 *		tabBinByte3 =	Ditto for byte 3.
 *
 *	One can produce inverted binary images by supplying inverted versions
 *	of the tabBinByteX tables.
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void  CDECL 
ip_toBinary128Line (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tabBinByte0,
    UInt32     *tabBinByte1,
    UInt32     *tabBinByte2,
    UInt32     *tabBinByte3)
{
    PROLOGUE;
    UInt32      oword;	/* permuted output word being assembled */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 8 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
	{
	oword = oword >> 1;
	oword |= ULONG_ACCESS(&pS[srcLimit]) & cMSBitMask;
	}
    ULONG_ACCESS(&pD[dstOffset]) = 
	tabBinByte0[(oword >> 24)] |
	tabBinByte1[(oword >> 16) & cByteMask] |
	tabBinByte2[(oword >> 8) & cByteMask] |
	tabBinByte3[(oword & cByteMask)];

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	oword = ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask;
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabBinByte0[(oword >> 24)] |
	    tabBinByte1[(oword >> 16) & cByteMask] |
	    tabBinByte2[(oword >> 8) & cByteMask] |
	    tabBinByte3[(oword & cByteMask)];
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_toBinaryFrom2bppLine()
 *  		Takes a 2-bit grayscale image and returns
 *             	a thresholded binary image, using the most significant
 *             	bit in each 2-bit pixel as the corresponding binary pixel.
 *	Input:
 *		pS =   		Pointer to left end of source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		tab2to1BinByte0: Pointer to table used to un-scramble
 *				 byte 0 of intermediate binarized 32-bit word.
 *		tab2to1BinByte1: Ditto for byte 1.
 *		tab2to1BinByte2: Ditto for byte 2.
 *		tab2to1BinByte3: Ditto for byte 3.
 *
 *	One can produce inverted binary images by supplying inverted versions
 *	of the tab2to1BinByteX tables.
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void  CDECL 
ip_toBinaryFrom2bppLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2to1BinByte0,
    UInt32     *tab2to1BinByte1,
    UInt32     *tab2to1BinByte2,
    UInt32     *tab2to1BinByte3)
{
    PROLOGUE;
    UInt32      oword;	/* permuted output word being assembled */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 2 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
	{
	oword = oword >> 1;
	oword |= ULONG_ACCESS(&pS[srcLimit]) & cMSBitMask2bpp;
	}
    ULONG_ACCESS(&pD[dstOffset]) = 
	tab2to1BinByte0[(oword >> 24)] |
	tab2to1BinByte1[(oword >> 16) & cByteMask] |
	tab2to1BinByte2[(oword >> 8) & cByteMask] |
	tab2to1BinByte3[(oword & cByteMask)];

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	oword = ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask2bpp;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask2bpp;
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tab2to1BinByte0[(oword >> 24)] |
	    tab2to1BinByte1[(oword >> 16) & cByteMask] |
	    tab2to1BinByte2[(oword >> 8) & cByteMask] |
	    tab2to1BinByte3[(oword & cByteMask)];
    }
}

/* PRIVATE */
/*************************************************************************** 
 *  ip_toBinaryFrom4bppLine()
 *  		Takes a 4-bit grayscale image and returns
 *             	a thresholded binary image, using the most significant
 *             	bit in each 4-bit pixel as the corresponding binary pixel.
 *	Input:
 *		pS =   		Pointer to left end of source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		tab4to1BinByte0: Pointer to table used to un-scramble
 *				 byte 0 of intermediate binarized 32-bit word.
 *		tab4to1BinByte1: Ditto for byte 1.
 *		tab4to1BinByte2: Ditto for byte 2.
 *		tab4to1BinByte3: Ditto for byte 3.
 *
 *	One can produce inverted binary images by supplying inverted versions
 *	of the tab4to1BinByteX tables.
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void  CDECL 
ip_toBinaryFrom4bppLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab4To1bppByte0,
    UInt32     *tab4To1bppByte1,
    UInt32     *tab4To1bppByte2,
    UInt32     *tab4To1bppByte3)
{
    PROLOGUE;
    UInt32      oword;	/* permuted output word being assembled */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 4 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;
	oword |= ULONG_ACCESS(&pS[srcLimit]) & cMSBitMask4bpp;
    }
    ULONG_ACCESS(&pD[dstOffset]) = 
	tab4To1bppByte0[(oword >> 24)] |
	tab4To1bppByte1[(oword >> 16) & cByteMask] |
	tab4To1bppByte2[(oword >> 8) & cByteMask] |
	tab4To1bppByte3[(oword & cByteMask)];

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	oword = ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask4bpp;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask4bpp;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask4bpp;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMSBitMask4bpp;
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tab4To1bppByte0[(oword >> 24)] |
	    tab4To1bppByte1[(oword >> 16) & cByteMask] |
	    tab4To1bppByte2[(oword >> 8) & cByteMask] |
	    tab4To1bppByte3[(oword & cByteMask)];
    }
}


/* PRIVATE */
/**************************************************************************
 *  ip_toTwoBitsPerPixelLine()
 *  		Takes an 8-bit grayscale image and returns
 *             	a thresholded 2 bit/pixel image, using the most significant
 *             	2 bits in each 8-bit pixel as the corresponding binary pixel.
 *	Input:
 *		pS =   		Pointer to left end of source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		tabTo2bppByte0: Pointer to table used to un-scramble
 *				 byte 0 of intermediate binarized 32-bit word.
 *		tabTo2bppByte1: Ditto for byte 1.
 *		tabTo2bppByte2: Ditto for byte 2.
 *		tabTo2bppByte3: Ditto for byte 3.
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void  CDECL 
ip_toTwoBitsPerPixelLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tabTo2bppByte0,
    UInt32     *tabTo2bppByte1,
    UInt32     *tabTo2bppByte2,
    UInt32     *tabTo2bppByte3)
{
    PROLOGUE;
    UInt32      oword;	/* permuted output word being assembled */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 4 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 2;
	oword |= ULONG_ACCESS(&pS[srcLimit]) & cMS2BitsMask;
    }
    ULONG_ACCESS(&pD[dstOffset]) = 
	tabTo2bppByte0[(oword >> 24)] |
	tabTo2bppByte1[(oword >> 16) & cByteMask] |
	tabTo2bppByte2[(oword >> 8) & cByteMask] |
	tabTo2bppByte3[(oword & cByteMask)];

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	oword = ULONG_ACCESS(&pS[srcOffset]) & cMS2BitsMask;
	oword = oword >> 2;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMS2BitsMask;
	oword = oword >> 2;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMS2BitsMask;
	oword = oword >> 2;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMS2BitsMask;
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabTo2bppByte0[(oword >> 24)] |
	    tabTo2bppByte1[(oword >> 16) & cByteMask] |
	    tabTo2bppByte2[(oword >> 8) & cByteMask] |
	    tabTo2bppByte3[(oword & cByteMask)];
    }

}


/* PRIVATE */
/**************************************************************************
 *  ip_toBinaryThreshLine()
 *  		Takes an 8-bit grayscale image and returns
 *             	a thresholded binary image.  If a GrayPix(x,y) >= thresh,
 *		BinaryPixel(x,y) = 1.  The binary image can be inverted by
 *		supplying inverted tabBinByteX tables.  These are the same
 *		tabBinByteX tables as are used by toBinary128.
 *	Input:
 *		pS =   		Pointer to left end of source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		thresh =	four copies of the binarization threshold,
 *				each one in the range in [1..255].  There is
 *				one copy in each byte, shifted right by 1.
 *		tabBinByte0 =	Pointer to table used to un-scramble
 *				byte 0 of intermediate binarized 32-bit word.
 *		tabBinByte1 =	Ditto for byte 1.
 *		tabBinByte2 =	Ditto for byte 2.
 *		tabBinByte3 =	Ditto for byte 3.
 *
 *	One can produce inverted binary images by supplying inverted versions
 *	of the tabBinByteX tables.
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void  CDECL 
ip_toBinaryThreshLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tabBinByte0,
    UInt32     *tabBinByte1,
    UInt32     *tabBinByte2,
    UInt32     *tabBinByte3,
    Int32	thresh)
{
    PROLOGUE;
    UInt32      oword;	/* permuted output word being assembled */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 8 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;
	oword |= (((ULONG_ACCESS(&pS[srcLimit]) >> 1) | cMSBitMask) - thresh) & cMSBitMask;
    }
    ULONG_ACCESS(&pD[dstOffset]) = 
	tabBinByte0[(oword >> 24)] |
	tabBinByte1[(oword >> 16) & cByteMask] |
	tabBinByte2[(oword >> 8) & cByteMask] |
	tabBinByte3[(oword & cByteMask)];

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	oword = (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
			thresh) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
			thresh) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
			thresh) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
			thresh) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
			thresh) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
			thresh) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
			thresh) & cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
			thresh) & cMSBitMask;
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabBinByte0[(oword >> 24)] |
	    tabBinByte1[(oword >> 16) & cByteMask] |
	    tabBinByte2[(oword >> 8) & cByteMask] |
	    tabBinByte3[(oword & cByteMask)];
    }
}

/* PRIVATE */
/*************************************************************************** 
 *  ip_toBinaryBandPassLine()
 *  		Takes an 8-bit grayscale image and returns
 *             	a thresholded binary image.  If lower <= GrayPix(x,y) <= upper,
 *		BinaryPixel(x,y) = 1.  The binary image can be inverted by
 *		supplying inverted tabBinByteX tables.  These are the same
 *		tabBinByteX tables as are used by toBinary128.
 *		lower is assumed to be less than or equal to upper.
 *	Input:
 *		pS =   		Pointer to left end of source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		lower =		lower binarization threshold replicated
 *				in all four bytes, right shifted by 1 with
 *				the MS bit of each byte CLEARED.
 *		upperLower =	original upper threshold - original lower
 *				threshold, right shifted by 1 with the MS bit
 *				of each byte SET.
 *		tabBinByte0 =	Pointer to table used to un-scramble
 *				byte 0 of intermediate binarized 32-bit word.
 *		tabBinByte1 =	Ditto for byte 1.
 *		tabBinByte2 =	Ditto for byte 2.
 *		tabBinByte3 =	Ditto for byte 3.
 *
 *	One can produce inverted binary images by supplying inverted versions
 *	of the tabBinByteX tables.
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void  CDECL 
ip_toBinaryBandPassLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tabBinByte0,
    UInt32     *tabBinByte1,
    UInt32     *tabBinByte2,
    UInt32     *tabBinByte3,
    Int32	lower,
    Int32	upperLower)
{
    PROLOGUE;
    UInt32      oword;	/* permuted output word being assembled */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 8 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;
	oword |= ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcLimit]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
    }
    ULONG_ACCESS(&pD[dstOffset]) = 
	tabBinByte0[(oword >> 24)] |
	tabBinByte1[(oword >> 16) & cByteMask] |
	tabBinByte2[(oword >> 8) & cByteMask] |
	tabBinByte3[(oword & cByteMask)];

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	oword = ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= ((upperLower - 
	    ((((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
	    lower) & ~cMSBitMask)) & cMSBitMask);
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabBinByte0[(oword >> 24)] |
	    tabBinByte1[(oword >> 16) & cByteMask] |
	    tabBinByte2[(oword >> 8) & cByteMask] |
	    tabBinByte3[(oword & cByteMask)];
    }
}


/*

$Log: midepth1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.3  1998/01/13 23:51:58  ddavies
midepth1.c -> /product/ipcore/ipshared/src/RCS/midepth1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1995/12/13 02:48:30  ejaquith
midepth1.c -> /product/ipcore/ipshared/src/RCS/midepth1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1994/01/28  04:10:54  ddavies
 * Initial revision
 *
 * Revision 1.25  1993/12/02  19:27:29  ddavies
 * sprxform.c -> /product/ipcore/ipshared/src/RCS/sprxform.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.24  1993/11/24  20:15:20  macleod
 * macleod on Wed Nov 24 12:14:46 PST 1993
 *
 * sprxform.c -> /product/ipcore/ipshared/src/RCS/sprxform.c,v
 *
 * added i_map8To8, which is a straightforward 8-bit LUT
 *
 * Revision 1.23  1993/11/22  22:36:03  jcraig
 * sprxform.c -> /product/ipcore/ipshared/src/RCS/sprxform.c,v
 *
 * in i_map4To4 keep the pointer to the top of the destination image
 * around so that it can be given to ip_clearRightFrame, instead of the
 * destination pointer that has gone down the image and points to the
 * end of the destination.
 *
 * Revision 1.22  1993/11/17  20:32:55  ddavies
 * sprxform.c -> /product/ipcore/ipshared/src/RCS/sprxform.c,v
 *
 * Change _SPARC_ARCH_ to "_ALPACA_IMAGE_FMT_ == cBigEndianFmt" in the code
 * that maps 4 bit images to 4 bit or 8 bit images using tables.
 * Also change i_threshold and i_thresholdI so they use the width to calculate
 * source and dest offsets instead of using source and dest bytes/line.  Don't
 * know why this wasn't done in the previous pass.
 * Also fixed bugs in i_threshol2To1bpp and i_thresholdI2To1bpp.  These
 * procedures produced every output line twice.  They might be fast enough
 * now to allow us to dispense with the assembly code.
 *
 * Revision 1.21  1993/11/16  03:03:49  ddavies
 * sprxform.c -> /product/ipcore/ipshared/src/RCS/sprxform.c,v
 *
 * All of these functions had been writing into the right frame.  All had been
 * using the source bytes/line to find the rightmost source word and the
 * dest bytes/line to find the rightmost dest word.  This doesn't work because
 * bytes/line - 2*frame = image + pad, not just image.  All have been changed
 * to use the image width in pixels to calculate source and destination
 * word offsets.
 *
 * Revision 1.20  1993/11/05  02:23:08  ddavies
 * sprxform.c -> /product/ipcore/ipshared/src/RCS/sprxform.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


