/*************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * midepth2.c -- machine independent primitives for procedures that change
 *		 the depth of images
 *
 *     4bpp Grayscale thresholding:
 *	void	ip_toBinaryFrom4bppThreshLine()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "imageref.pub"

    /* prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: midepth2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

#define cLSBitMask4bpp	0x11111111
#define cMSBitMask4bpp	0x88888888
#define	cLSBitCarry4bpp	0x22222222
#define	cByteMask	0xFF

/* PRIVATE */
/***************************************************************************  
 *  ip_toBinaryFrom4bppThreshLine():  takes a 4-bit grayscale image line and
 *		a threshold value between 1 and 15 and a pointer to
 *		the destination image line.
 *		If GrayPix(x,y) >= thresh, BinaryPix(x,y) = 1 (roughly)
 *
 *	Input:
 *		pS =		Pointer to left end of source line inside frame
 *		pD =		Pointer to left end of dest line inside frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		thresh =	high pass filter threshold in [1,,15].
 *		tabByte0 =	permute table for byte 0
 *		tabByte1 =	permute table for byte 1
 *		tabByte2 =	permute table for byte 2
 *		tabByte3 =	permute table for byte 3
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void CDECL 
ip_toBinaryFrom4bppThreshLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tabByte0,
    UInt32     *tabByte1,
    UInt32     *tabByte2,
    UInt32     *tabByte3,
    Int32	thresh,
    Int32	threshLow)

{
    PROLOGUE;
    UInt32	oword;		/* permuted output word being assembled */
    UInt32	iwordHigh, iwordLow;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* A performance improvement was discovered after the original
	 * code was written.  In the original implementation, we carefully
	 * separated out and subtracted the ls bit of the threshold
	 * from the LS bit of the pixel.  However, there are really only
	 * two possible values of the LS bit of the threshold.  If
	 * it's 0, the borrow into the upper bits (which is all we
	 * care about) is always 0, so we can skip a bunch of processing.
	 * If it's 1, the borrow is always the complement of the
	 * LS pixel bit, so we just use that.  This still saves a bunch
	 * of subtracts and shifts and stuff.
	 */
    if (!(threshLow & 0x1))
    {
	    /* pixel.0 - 0 never generates a borrow, so just ignore it */

	    /* do the last destination word first because it may require
	     * less than 4 source words.  
	     */
	oword = 0;
	for (; srcLimit >= srcOffset; srcLimit -= 4)
	{
	    oword = oword >> 1;
	    iwordHigh = ULONG_ACCESS(&pS[srcLimit]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	}
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabByte0[(oword >> 24)] |
	    tabByte1[(oword >> 16) & cByteMask] |
	    tabByte2[(oword >> 8) & cByteMask] |
	    tabByte3[(oword & cByteMask)];
    
	    /* the loop for each destination word */
	for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
	{
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword = iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    srcOffset -= 4;
    
	    ULONG_ACCESS(&pD[dstOffset]) = 
		tabByte0[(oword >> 24)] |
		tabByte1[(oword >> 16) & cByteMask] |
		tabByte2[(oword >> 8) & cByteMask] |
		tabByte3[(oword & cByteMask)];
	}
    }
    else
    {
	    /* case 1:
	     * pixel.0 - 1 generates a borrow when pixel.0 = 0, i.e. the
	     * borrow is just the complement of pixel.0
	     */

	    /* do the last destination word first because it may require
	     * less than 4 source words.  
	     */
	oword = 0;
	for (; srcLimit >= srcOffset; srcLimit -= 4)
	{
	    oword = oword >> 1;
	    iwordHigh = ULONG_ACCESS(&pS[srcLimit]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	}
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabByte0[(oword >> 24)] |
	    tabByte1[(oword >> 16) & cByteMask] |
	    tabByte2[(oword >> 8) & cByteMask] |
	    tabByte3[(oword & cByteMask)];
    
	    /* the loop for each destination word */
	for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
	{
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword = (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - thresh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    srcOffset -= 4;
    
	    ULONG_ACCESS(&pD[dstOffset]) = 
		tabByte0[(oword >> 24)] |
		tabByte1[(oword >> 16) & cByteMask] |
		tabByte2[(oword >> 8) & cByteMask] |
		tabByte3[(oword & cByteMask)];
	}
    }	/* end of switch */
}

/*

$Log: midepth2.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.4  1998/01/13 23:52:00  ddavies
midepth2.c -> /product/ipcore/ipshared/src/RCS/midepth2.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1995/12/13 02:48:34  ejaquith
midepth2.c -> /product/ipcore/ipshared/src/RCS/midepth2.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1994/03/24  20:40:08  ddavies
 * midepth2.c -> /product/ipcore/ipshared/src/RCS/midepth2.c,v
 *
 * Update implementation of ip_toBinaryFrom4bppThreshLine so it branches
 * on the value of the LS bit of the threshold outside the main loop instead
 * of doing a complicated calculation based on that value inside the loop.
 *
 * Revision 1.1  1994/01/28  04:11:47  ddavies
 * Initial revision
 *
 * Revision 1.9  1993/12/02  19:20:33  ddavies
 * cprxform.c -> /product/ipcore/ipshared/src/RCS/cprxform.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.8  1993/11/16  02:34:33  ddavies
 * cprxform.c -> /product/ipcore/ipshared/src/RCS/cprxform.c,v
 *
 * Modify C implementation of ip_toBinaryFrom4bppThresh so that is uses the
 * width in pixels to calculate source and dest offsets instead of using
 * the source and dest image bytes/line.  This was done to keep the code
 * from writing into the right frame of the destination.
 *
 * Revision 1.7  1993/11/05  01:23:35  ddavies
 * cprxform.c -> /product/ipcore/ipshared/src/RCS/cprxform.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


