/*************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * midepth3.c -- machine independent primitives for procedures that change
 *		 the depth of images
 *
 *	void	ip_map8To8Line()
 *      void    ip_fourToTwoBitsPerPixelLine()
 *	void	ip_threshold8To4Line()
 *      Int32   ip_toBinaryByPixelLine()
 *
 *     4bpp Grayscale bandpass filtering:
 *	void	ip_toBinaryFrom4bppBandpassLine()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "imageref.pub"

    /* prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: midepth3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

#define cNib0Mask	0xF0000000
#define cNib1Mask	0x0F000000
#define cNib2Mask	0x00F00000
#define	cNib3Mask	0x000F0000
#define cNib4Mask	0x0000F000
#define	cNib5Mask	0x00000F00
#define	cNib6Mask	0x000000F0
#define	cNib7Mask	0x0000000F

#define cMSBitMask	0x80808080

#define cLSBitMask4bpp	0x11111111
#define cMSBitMask4bpp	0x88888888
#define	cLSBitCarry4bpp	0x22222222
#define cMS2BitsMask4bpp  0xCCCCCCCC
#define	cByteMask	0xFF


/* PRIVATE */
/*************************************************************************** 
 *  ip_map8To8Line()
 *		Applies a table to all the bytes of one image line, producing
 *		another image line.  This routine is used by both i_map4To4
 *		(maps 2 pixels at once) and by i_map8To8.
 *
 *	Input:
 *		pS:		Pointer to left end of source line
 *				   INSIDE frame
 *		pD:		Pointer to left end of dest line
 *				   INSIDE frame
 *		widthInWords:	Width of source line in 32-bit words
 *		mapTable:	Pointer to table used to map bytes
 *
 **************************************************************************/
void  CDECL 
ip_map8To8Line (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	widthInWords,
    UInt8      *mapTable)
{
    PROLOGUE;
    UInt8      *psrc, *pdst, *pend;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Precalculate the ending ptr value, so we can avoid
     * needing to maintain a loop counter. Process a
     * longword per loop iteration.
     */
    pend = pS + 4*widthInWords; /* longword addr after end */
    for (psrc=pS, pdst=pD; psrc<pend; psrc+=4, pdst+=4)
    {

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

        /* On the SPARC, reads and shifts are fast,
         *  but writes can be slow. So we'll read data
         *  a byte at a time, build the longword and
         *  write it as a 32 bit chunk. The ULONG_ACCESS
         *  macro casts the UInt8 ptr to a UInt32 ptr
         */
	ULONG_ACCESS(pdst) = 
			(mapTable[*(psrc)] << 24) |
			(mapTable[*(psrc+1)] << 16) |
			(mapTable[*(psrc+2)] << 8) |
			(mapTable[*(psrc+3)] ) ;

#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

        /*
           On the PC, shifts are slow, so its faster to
           read and write a byte at a time. UCHAR_ACCESS
           isn't required here, since we're just processing
           a byte at a time, always in full words.
        */
        *(pdst) = mapTable[*(psrc)];
        *(pdst+1) = mapTable[*(psrc+1)];
        *(pdst+2) = mapTable[*(psrc+2)];
        *(pdst+3) = mapTable[*(psrc+3)];

#endif
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_threshold8To4Line()
 *  		Takes an 8-bit grayscale image line and returns
 *             	a thresholded 4-bit image line, using the most significant
 *             	nibble in each byte as the corresponding dest pixel.
 *	Input:
 *		pS =   		Pointer to left end of source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *
 *	One can produce inverted binary images by supplying inverted versions
 *	of the tab2to1BinByteX tables.
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void  CDECL 
ip_threshold8To4Line (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit)
{
    PROLOGUE;
    UInt32	oword;	/* permuted output word being assembled */
    UInt32	iword;	/* word read from source image */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 2 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 16;
	iword = ULONG_ACCESS(&pS[srcLimit]);
	oword |= iword & cNib0Mask;
	oword |= (iword << (27 - 23)) & cNib1Mask;
	oword |= (iword << (23 - 15)) & cNib2Mask;
	oword |= (iword << (19 - 7)) & cNib3Mask;
    }
    ULONG_ACCESS(&pD[dstOffset]) = oword;

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	iword = ULONG_ACCESS(&pS[srcOffset]);	/* get right source word */
	oword = (iword >> (31 - 15)) & cNib4Mask;
	oword |= (iword >> (23 - 11)) & cNib5Mask;
	oword |= (iword >> (15 -  7)) & cNib6Mask;
	oword |= (iword >> ( 7 -  3)) & cNib7Mask;
	srcOffset -= 4;

	iword = ULONG_ACCESS(&pS[srcOffset]);	/* get left source word */
	oword |= iword & cNib0Mask;
	oword |= (iword << (27 - 23)) & cNib1Mask;
	oword |= (iword << (23 - 15)) & cNib2Mask;
	oword |= (iword << (19 - 7)) & cNib3Mask;
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = oword;
    }
}



/* PRIVATE */
/**************************************************************************
 *  ip_fourToTwoBitsPerPixelLine()
 *  		Takes an 4-bit grayscale image and returns
 *             	a thresholded 2 bit/pixel image, using the most significant
 *             	2 bits in each 4-bit pixel as the corresponding binary pixel.
 *	Input:
 *		pS =   		Pointer to left end of source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		tab4To2bpp :    Pointer to table used to un-scramble
 *				 bytes of intermediate binarized 32-bit word.
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void CDECL 
ip_fourToTwoBitsPerPixelLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab4To2bpp)
{
    PROLOGUE;
    UInt32	oword;	/* permuted output word being assembled */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* do the last destination word first because it may require
     * less than 2 source words.  
     */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 2;
	oword |= ULONG_ACCESS(&pS[srcLimit]) & cMS2BitsMask4bpp;
    }
    ULONG_ACCESS(&pD[dstOffset]) = 
	 tab4To2bpp[(oword >> 24)] |
	(tab4To2bpp[(oword >> 16) & cByteMask] >> 4) |
	(tab4To2bpp[(oword >> 8) & cByteMask] >> 8) |
	(tab4To2bpp[(oword & cByteMask)] >> 12);

    /* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	oword = ULONG_ACCESS(&pS[srcOffset]) & cMS2BitsMask4bpp;
	oword = oword >> 2;
	srcOffset -= 4;
	oword |= ULONG_ACCESS(&pS[srcOffset]) & cMS2BitsMask4bpp;
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = 
	     tab4To2bpp[(oword >> 24)] |
	    (tab4To2bpp[(oword >> 16) & cByteMask] >> 4) |
	    (tab4To2bpp[(oword >> 8) & cByteMask] >> 8) |
	    (tab4To2bpp[(oword & cByteMask)] >> 12);
    }

}  /* ip_fourToTwoBitsPerPixelLine () */


/* PRIVATE */
/**************************************************************************
 *  ip_toBinaryByPixelLine()
 *  		Takes two 8-bit grayscale images and returns
 *             	a thresholded binary image.  If GrayPix(x,y) >= ThreshPix(x,y),
 *		BinaryPixel(x,y) = 1.
 *		The binary image can be inverted by supplying inverted
 *		tabBinByteX tables.  Inverting is the normal case,
 *		but we're trying to make this routine compatible with
 *		all the others like it.  These are the same
 *		tabBinByteX tables as are used by toBinary128Line.
 *	Input:
 *		pS =   		Pointer to left end of gray source line
 *				  INSIDE frame
 *		pT =   		Pointer to left end of threshold source line
 *				  INSIDE frame
 *		pD =    	Pointer to left end of dest line
 *				  INSIDE frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		tabBinByte0 =	Pointer to table used to un-scramble
 *				byte 0 of intermediate binarized 32-bit word.
 *		tabBinByte1 =	Ditto for byte 1.
 *		tabBinByte2 =	Ditto for byte 2.
 *		tabBinByte3 =	Ditto for byte 3.
 *
 *	One can produce inverted binary images by supplying inverted versions
 *	of the tabBinByteX tables.
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void  CDECL 
ip_toBinaryByPixelLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pT,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tabBinByte0,
    UInt32     *tabBinByte1,
    UInt32     *tabBinByte2,
    UInt32     *tabBinByte3)
{
    PROLOGUE;
    UInt32      oword;	/* permuted output word being assembled */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 8 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;
	oword |= (((ULONG_ACCESS(&pS[srcLimit]) >> 1) | cMSBitMask) -
		  ((ULONG_ACCESS(&pT[srcLimit]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
    }
    ULONG_ACCESS(&pD[dstOffset]) = 
	tabBinByte0[(oword >> 24)] |
	tabBinByte1[(oword >> 16) & cByteMask] |
	tabBinByte2[(oword >> 8) & cByteMask] |
	tabBinByte3[(oword & cByteMask)];

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	oword = (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
		 ((ULONG_ACCESS(&pT[srcOffset]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
		  ((ULONG_ACCESS(&pT[srcOffset]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
		  ((ULONG_ACCESS(&pT[srcOffset]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
		  ((ULONG_ACCESS(&pT[srcOffset]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
		  ((ULONG_ACCESS(&pT[srcOffset]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
		  ((ULONG_ACCESS(&pT[srcOffset]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
		  ((ULONG_ACCESS(&pT[srcOffset]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
	oword = oword >> 1;
	srcOffset -= 4;
	oword |= (((ULONG_ACCESS(&pS[srcOffset]) >> 1) | cMSBitMask) -
		  ((ULONG_ACCESS(&pT[srcOffset]) >> 1) & ~cMSBitMask)) &
			cMSBitMask;
	srcOffset -= 4;
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabBinByte0[(oword >> 24)] |
	    tabBinByte1[(oword >> 16) & cByteMask] |
	    tabBinByte2[(oword >> 8) & cByteMask] |
	    tabBinByte3[(oword & cByteMask)];
    }
}


/* PRIVATE */
/***************************************************************************  
 *  ip_toBinaryFrom4bppBandpassLine():  takes a 4-bit grayscale image line and
 *		a threshold value between 0 and 15 and a pointer to
 *		the destination image line.
 *		If {lower <= GrayPix(x,y) <= upper}, BinaryPix(x,y) = 1
 *
 *		The values in the permute tables not only permute the bits
 *		we calculate with the sums into the right order, they
 *		can also be used to invert the output so this routine
 *		is called by both i_passfilt4To1bpp and i_passfiltI4To1bpp.
 *
 *	Input:
 *		pS =		Pointer to left end of source line inside frame
 *		pD =		Pointer to left end of dest line inside frame
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		tabByte0 =	permute table for byte 0
 *		tabByte1 =	permute table for byte 1
 *		tabByte2 =	permute table for byte 2
 *		tabByte3 =	permute table for byte 3
 *		upLow =		MS 3 bits of (upper threshold - lower threshold)
 *				replicated in all 8 nibbles and right shifted
 *				by 1.
 *		lower =		MS 3 bits of lower threshold replicated in
 *				all 8 nibbles and right shifted by 1
 *		lowBits =	Contains the LS bit of (upperThreshold -
 *				lower threshold) in bit 1 and the LS bit of
 *				the lower threshold in bit 0.  We use this
 *				in a switch statement to handle the LS bit
 *				computations.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
void CDECL 
ip_toBinaryFrom4bppBandpassLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tabByte0,
    UInt32     *tabByte1,
    UInt32     *tabByte2,
    UInt32     *tabByte3,
    Int32	upLow,
    Int32	lower,
    Int32	lowBits)

{
    PROLOGUE;
    UInt32		oword;	/* permuted output word being assembled */
    UInt32		iwordHigh, iwordLow;

	/* The calculation we do is sort of tricky.  In it's basic form,
	 * we first subtract the lower threshold from the incoming pixels.
	 * This produces a bunch of numbers in the range [0, upper-lower].
	 * We shift these right (remembering the LS bit elsewhere) and clear
	 * the MS bits.  This is subtracted from upLow, producing 1's in the
	 * MS bits of those nibbles that are less than or equal to upLow.
	 * These are the nibbles that were greater than or equal to lower
	 * and less than or equal to upper.
	 *
	 * This process is complicated by the fact that each "subtract"
	 * operation is weird.  First, we extract the LS bits from each
	 * of minuend pixels and OR in the bits just above them (this was
	 * done ahead of time when the minuend is the LS bits of upLow).
	 * We then subtract the LS bits of the subtrahend.  This produces
	 * 1's in the carry bits when no borrow was needed.  Since we
	 * want to compute borrows, we complement the result, select out
	 * the carry bits and shift the result right by 1 so the borrows
	 * are in the right position.
	 *
	 * We then shift the minuend right by 1 and turn on the (now vacated)
	 * MS bits.  We subtract the 3 MS bits of the minuend - subtrahend
	 * and subtract off the borrows calculated above.  So easy!
	 *
	 * Well, not quite.  We need to do two subtracts here.  In the
	 * second step described above, we tossed the LS bit of (pix - lower).
	 * We can't do that before doing the second difference.  In fact,
	 * doing two subtracts in this cumbersome way is doubly
	 * annoying.  Are there any shortcuts?
	 *
	 * Let's do a 4 way branch based on the LS bits of (upper - lower)
	 * and lower.
	 * We notice that if the LS bit of lower is 0, the borrow from
	 * (pix - lower) is equal to the LS bit of pix, so we can just use
	 * that.  The LS bit of (pix - lower) is also just the LS bit of
	 * pix in that case.  If the LS bit of lower is 1, the borrow
	 * from (pix - lower) is always 0 and the LS bit of the difference
	 * is the complement of the LS bit of pix.
	 *
	 * When the LS bit of the real (upper - lower), which is held in
	 * upLowLSBits, is 0, the borrow is the LS bit of (lower - pix).
	 * When the LS bit of (upper - lower) is 1, there is no borrow, so
	 * we can disregard that computation.
	 */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    switch (lowBits)
    {
    case 0:		/* LS bit of upLow = 0, LS bit of lower = 0.
			 * Therefore, the borrow from the LS bit of
			 * (pix - lower) equals 0;
			 * and the borrow from the LS bit of
			 * upLow - (pix - lower) equals the
			 * LS bit of pix.
			 */
	    /* do the last destination word first because it may require
	     * less than 4 source words.  
	     */
	oword = 0;
	for (; srcLimit >= srcOffset; srcLimit -= 4)
	{
	    oword = oword >> 1;
	    iwordHigh = ULONG_ACCESS(&pS[srcLimit]);
		/* calculate (input pixels - lower threshold) */
	    iwordLow = iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	}
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabByte0[(oword >> 24)] |
	    tabByte1[(oword >> 16) & cByteMask] |
	    tabByte2[(oword >> 8) & cByteMask] |
	    tabByte3[(oword & cByteMask)];
    
	    /* the loop for each destination word */
	for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
	{
		/* calculate (input pixels - lower threshold) */
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword = (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    srcOffset -= 4;
    
	    ULONG_ACCESS(&pD[dstOffset]) = 
		tabByte0[(oword >> 24)] |
		tabByte1[(oword >> 16) & cByteMask] |
		tabByte2[(oword >> 8) & cByteMask] |
		tabByte3[(oword & cByteMask)];
	}
	break;	/* end of case where upLow.0 = 0 and lower.0 = 0 */

    case 1:	/* upLow.0 = 0, lower.0 = 1.
		 * in this case, the borrow from the LS bit of (pix - lower)
		 * equals the complement of the LS bit of pix.
		 * The borrow from the LS bit of upLow - (pix-lower)
		 * equals the complement of the LS bit of pix.
		 */
	    /* do the last destination word first because it may require
	     * less than 4 source words.  
	     */
	oword = 0;
	for (; srcLimit >= srcOffset; srcLimit -= 4)
	{
	    oword = oword >> 1;
	    iwordHigh = ULONG_ACCESS(&pS[srcLimit]);
		/* calculate (input pixels - lower threshold) */
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	}
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabByte0[(oword >> 24)] |
	    tabByte1[(oword >> 16) & cByteMask] |
	    tabByte2[(oword >> 8) & cByteMask] |
	    tabByte3[(oword & cByteMask)];
    
	    /* the loop for each destination word */
	for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
	{
		/* calculate (input pixels - lower threshold) */
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword = (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;

	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= (iwordHigh - iwordLow) & cMSBitMask4bpp;
	    srcOffset -= 4;
    
	    ULONG_ACCESS(&pD[dstOffset]) = 
		tabByte0[(oword >> 24)] |
		tabByte1[(oword >> 16) & cByteMask] |
		tabByte2[(oword >> 8) & cByteMask] |
		tabByte3[(oword & cByteMask)];
	}
	break;	/* end of case where upLow.0 = 0 and lower.0 = 1 */

    case 2:	/* upLow.0 = 1, lower.0 = 0.
		 * in this case, the borrow from the LS bit of (pix - lower)
		 * equals 0.  The borrow from the LS bit
		 * of upLow - (pix-lower) equals 0 also.  Therefore,
		 * we can ignore the LS bit entirely.
		 */
	    /* do the last destination word first because it may require
	     * less than 4 source words.  
	     */
	oword = 0;
	for (; srcLimit >= srcOffset; srcLimit -= 4)
	{
	    oword = oword >> 1;
	    iwordHigh = ULONG_ACCESS(&pS[srcLimit]);
		/* calculate (input pixels - lower threshold) */
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	}
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabByte0[(oword >> 24)] |
	    tabByte1[(oword >> 16) & cByteMask] |
	    tabByte2[(oword >> 8) & cByteMask] |
	    tabByte3[(oword & cByteMask)];
    
	    /* the loop for each destination word */
	for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
	{
		/* calculate (input pixels - lower threshold) */
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword = iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = iwordHigh & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    srcOffset -= 4;
    
	    ULONG_ACCESS(&pD[dstOffset]) = 
		tabByte0[(oword >> 24)] |
		tabByte1[(oword >> 16) & cByteMask] |
		tabByte2[(oword >> 8) & cByteMask] |
		tabByte3[(oword & cByteMask)];
	}
	break;	/* end of case where upLow.0 = 1 and lower.0 = 0 */

    default:	/* upLow.0 = 1, lower.0 = 1.
		 * in this case, the borrow from the LS bit of (pix - lower)
		 * equals the complement of the LS bit of pix.
		 * The borrow from the LS bit of upLow - (pix-lower) equals 0.
		 */
	    /* do the last destination word first because it may require
	     * less than 4 source words.  
	     */
	oword = 0;
	for (; srcLimit >= srcOffset; srcLimit -= 4)
	{
	    oword = oword >> 1;
	    iwordHigh = ULONG_ACCESS(&pS[srcLimit]);
		/* calculate (input pixels - lower threshold) */
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	}
	ULONG_ACCESS(&pD[dstOffset]) = 
	    tabByte0[(oword >> 24)] |
	    tabByte1[(oword >> 16) & cByteMask] |
	    tabByte2[(oword >> 8) & cByteMask] |
	    tabByte3[(oword & cByteMask)];
    
	    /* the loop for each destination word */
	for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
	{
		/* calculate (input pixels - lower threshold) */
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword = iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    oword = oword >> 1;
	    srcOffset -= 4;
    
	    iwordHigh = ULONG_ACCESS(&pS[srcOffset]);
	    iwordLow = ~iwordHigh & cLSBitMask4bpp;
	    iwordHigh = ((iwordHigh >> 1) | cMSBitMask4bpp) - lower;
	    iwordHigh = (iwordHigh - iwordLow) & ~cMSBitMask4bpp;
		/* calculate (upLow - (pix - lower)) */
	    iwordHigh = upLow - iwordHigh;
	    oword |= iwordHigh & cMSBitMask4bpp;
	    srcOffset -= 4;
    
	    ULONG_ACCESS(&pD[dstOffset]) = 
		tabByte0[(oword >> 24)] |
		tabByte1[(oword >> 16) & cByteMask] |
		tabByte2[(oword >> 8) & cByteMask] |
		tabByte3[(oword & cByteMask)];
	}
	break;	/* end of case where upLow.0 = 1 and lower.0 = 1 */
    }
}


/*

 $Log: midepth3.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.7  1998/02/24 03:44:32  ddavies
 midepth3.c -> /product/ipcore/ipshared/src/RCS/midepth3.c,v

 Add implementation of ip_toBinaryByPixelLine.  This function converts a
 line of a gray source image to a line of a binary image by thresholding
 each pixel relative to a line of threshold values.

 Revision 1.6  1998/01/13 23:52:01  ddavies
 midepth3.c -> /product/ipcore/ipshared/src/RCS/midepth3.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.5  1995/12/13 02:48:38  ejaquith
 midepth3.c -> /product/ipcore/ipshared/src/RCS/midepth3.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/10/28  01:48:04  lperry
 * lperry on Thu Oct 27 18:43:01 PDT 1994
 *
 * midepth3.c -> /product/ipcore/ipshared/src/RCS/midepth3.c,v
 *
 * Change ip_map8To8Line so that PC specific code is
 * faster. Now does individual byte ops rather than
 * assembling a longword with shifts. New code is
 * about 20% faster.
 * This change was originally incorporated in code within
 * i_aiepro.c . By moving it to midepth3.c, all functions
 * which do 8 bit gray mappings can benefit.
 *
 * Revision 1.3  1994/07/30  00:06:19  dferrell
 * midepth3.c -> /product/ipcore/ipshared/src/RCS/midepth3.c,v
 *
 * Wrote new ip_fourToTwoBitsPerPixelLine () which converts a single row of
 * pixels from 4 to 2 bit depth.
 *
 * Revision 1.2  1994/03/24  20:41:01  ddavies
 * midepth3.c -> /product/ipcore/ipshared/src/RCS/midepth3.c,v
 *
 * Add implementation of ip_toBinaryFrom4bppBandpassLine.  This routine
 * produces a single line of 1 bit pixels from a single line of 4 bit pixels
 * using bandpass filtering.
 *
 * Revision 1.1  1994/01/28  04:13:12  ddavies
 * Initial revision
 *

 */

