/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  i_dtran4.c -- IPShared 4 neighbor distance transform functionality
 *
 *	ip_leftNeighborDTRasRow ()
 *	ip_bottomNeighborDTRasRow ()
 *	ip_rightNeighborDTAntiRasRow ()
 *	ip_topRightNeighborDTAntiRasRow ()
 *	ip_fixRightFrameRow ()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

#include "dtrans.prv"


IP_RCSINFO(RCSInfo, "$RCSfile: midtran4.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");


/* Distance transform black universe cases for top and bottom rows */
static UInt32 DTOnes[] = {
	0x01010101,
	0x01010100,
	0x01010001,
	0x01010000,
	0x01000101,
	0x01000100,
	0x01000001,
	0x01000000,
	0x00010101,
	0x00010100,
	0x00010001,
	0x00010000,
	0x00000101,
	0x00000100,
	0x00000001,
	0x00000000};

/* Distance transform forward ramps */
static UInt32 DTFRamp[] = {
	0x01020304,
	0x01020300,
	0x01020001,
	0x01020000,
	0x01000102,
	0x01000100,
	0x01000001,
	0x01000000,
	0x00010203,
	0x00010100,
	0x00010001,
	0x00010000,
	0x00000102,
	0x00000100,
	0x00000001,
	0x00000000};

/* Distance transform forward ramp masks */
static UInt32 DTFMask[] = {
	0xffffffff,
	0xffffff00,
	0xffff0000,
	0xffff0000,
	0xff000000,
	0xff000000,
	0xff000000,
	0xff000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000};

/* Distance transform reverse ramps */
static UInt32 DTRRamp[] = {
	0x04030201,
	0x03020100,
	0x02010001,
	0x02010000,
	0x01000201,
	0x01000100,
	0x01000001,
	0x01000000,
	0x00030201,
	0x00020100,
	0x00010001,
	0x00010000,
	0x00000201,
	0x00000100,
	0x00000001,
	0x00000000};

/* Distance transform reverse ramp masks */
static UInt32 DTRMask[] = {
	0xffffffff,
	0x00000000,
	0x000000ff,
	0x00000000,
	0x0000ffff,
	0x00000000,
	0x000000ff,
	0x00000000,
	0x00ffffff,
	0x00000000,
	0x000000ff,
	0x00000000,
	0x0000ffff,
	0x00000000,
	0x000000ff,
	0x00000000};



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Left neighbor raster order distance transform for one row
 *
 * LongDescription:::
 *
 * Input:::
 *	bImage::	pointer to binary input image row (w/o frame)
 *	w::		width of image row (w/o frame)
 *	edgeRow::	cTrue if top or bottom image row, otherwise cFalse
 *	fromBlack::	cTrue if computing distance of white pixels from
 *			nearest black pixels, cFalse if computing distance
 *			of black pixels from nearest white pixels
 *	universeColor::	0 for white universe, 1 for black
 *
 * Output:::
 *	gImage::	Pointer to first image byte of the gray destination row
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *	This function essentially builds the left to right distance ramps
 *	from a row of binary input.  If the row is at an edge and the
 *	universe color affects the distance, then all distances are 1 or
 *	0 for this row.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_leftNeighborDTRasRow (
    IPState    *pIPState,
    UInt8      *bImage,
    Int32	w,
    Int32	edgeRow,
    Int32	fromBlack,
    Int32	universeColor,
    UInt8      *gImage)
{
    PROLOGUE;
    UInt32	iWord, oWord, iNibble, fWord, lWord, rWord;
    Int32	bWordCount, sparePix;
    UInt8      *lastGAddr;


    /* do all gray words which are part of a corresponding binary
     * image words (including frame pixels)
     */
    lastGAddr = gImage + (((w + 31) / 32) * 32);

    bWordCount = w / 32;
    sparePix = w % 32;

    if (edgeRow && ((fromBlack && (universeColor == 1)) || 
                    (!fromBlack && (universeColor == 0))))
    {
	/* top or bottom edge where universe color influences distance */

	/* simply put 01 where there was white and 00 where there was black */
	while (gImage < lastGAddr)
	{
	    iWord = ULONG_ACCESS (bImage);
	    if (!fromBlack)
	        iWord = ~iWord;
	    bImage += 4;


	    /***** Gray word 0 *****/
	    iNibble = (iWord >> 28) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 1 *****/
	    iNibble = (iWord >> 24) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 2 *****/
	    iNibble = (iWord >> 20) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 3 *****/
	    iNibble = (iWord >> 16) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 4 *****/
	    iNibble = (iWord >> 12) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 5 *****/
	    iNibble = (iWord >> 8) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 6 *****/
	    iNibble = (iWord >> 4) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 7 *****/
	    iNibble = iWord & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	}  /* for each binary word of 32 pixels */
    }
    else  /* not an edge case where the universe color should affect the ends */
    {
        if (universeColor == 1)
	    /* black universe */
	    lWord = 0x00000000;
	else
	    /* white universe */
	    lWord = 0xffffffff;

        if (!fromBlack)
	    lWord = ~lWord;

	while (gImage < lastGAddr)
	{
	    iWord = ULONG_ACCESS (bImage);
	    if (!fromBlack)
	        iWord = ~iWord;
	    bImage += 4;
    
	    if (bWordCount == 0)
	    {
		if (!fromBlack)
		{
                    if (universeColor == 1)
			iWord &= (0xfffffffe << (31 - sparePix));
		}
		else
		{
		    if (universeColor == 1)
			iWord |= ~(0xfffffffe << (31 - sparePix));
		}
	    }

	    bWordCount--;

	    /***** Gray word 0 *****/
	    iNibble = (iWord >> 28) & 0x000f;
	    rWord = DTFRamp[iNibble];	/* get the ramps */
	    lWord &= DTFMask[iNibble];	/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = rWord + lWord;	/* compute the continued ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    oWord = (lWord - fWord) + rWord;  /* presubtract overflow 
					       * and recalc */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
	    oWord |= fWord;		/* OR in the overflow value */
    
	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0x000000ff;
	    lWord |= lWord << 16;
	    lWord |= lWord << 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 1 *****/
	    iNibble = (iWord >> 24) & 0x000f;
	    rWord = DTFRamp[iNibble];	/* get the ramps */
	    lWord &= DTFMask[iNibble];	/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = rWord + lWord;	/* compute the continued ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    oWord = (lWord - fWord) + rWord;  /* presubtract overflow 
					       * and recalc */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
	    oWord |= fWord;		/* OR in the overflow value */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* set the last value word for the next step */
	    lWord = oWord & 0x000000ff;
	    lWord |= lWord << 16;
	    lWord |= lWord << 8;

	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 2 *****/
	    iNibble = (iWord >> 20) & 0x000f;
	    rWord = DTFRamp[iNibble];	/* get the ramps */
	    lWord &= DTFMask[iNibble];	/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = rWord + lWord;	/* compute the continued ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    oWord = (lWord - fWord) + rWord;  /* presubtract overflow 
					       * and recalc */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
	    oWord |= fWord;		/* OR in the overflow value */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* set the last value word for the next step */
	    lWord = oWord & 0x000000ff;
	    lWord |= lWord << 16;
	    lWord |= lWord << 8;

	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 3 *****/
	    iNibble = (iWord >> 16) & 0x000f;
	    rWord = DTFRamp[iNibble];	/* get the ramps */
	    lWord &= DTFMask[iNibble];	/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = rWord + lWord;	/* compute the continued ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    oWord = (lWord - fWord) + rWord;  /* presubtract overflow 
					       * and recalc */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
	    oWord |= fWord;		/* OR in the overflow value */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* set the last value word for the next step */
	    lWord = oWord & 0x000000ff;
	    lWord |= lWord << 16;
	    lWord |= lWord << 8;

	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 4 *****/
	    iNibble = (iWord >> 12) & 0x000f;
	    rWord = DTFRamp[iNibble];	/* get the ramps */
	    lWord &= DTFMask[iNibble];	/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = rWord + lWord;	/* compute the continued ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    oWord = (lWord - fWord) + rWord;  /* presubtract overflow 
					       * and recalc */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
	    oWord |= fWord;		/* OR in the overflow value */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* set the last value word for the next step */
	    lWord = oWord & 0x000000ff;
	    lWord |= lWord << 16;
	    lWord |= lWord << 8;

	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 5 *****/
	    iNibble = (iWord >> 8) & 0x000f;
	    rWord = DTFRamp[iNibble];	/* get the ramps */
	    lWord &= DTFMask[iNibble];	/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = rWord + lWord;	/* compute the continued ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    oWord = (lWord - fWord) + rWord;  /* presubtract overflow 
					       * and recalc */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
	    oWord |= fWord;		/* OR in the overflow value */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* set the last value word for the next step */
	    lWord = oWord & 0x000000ff;
	    lWord |= lWord << 16;
	    lWord |= lWord << 8;

	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 6 *****/
	    iNibble = (iWord >> 4) & 0x000f;
	    rWord = DTFRamp[iNibble];	/* get the ramps */
	    lWord &= DTFMask[iNibble];	/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = rWord + lWord;	/* compute the continued ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    oWord = (lWord - fWord) + rWord;  /* presubtract overflow 
					       * and recalc */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
	    oWord |= fWord;		/* OR in the overflow value */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* set the last value word for the next step */
	    lWord = oWord & 0x000000ff;
	    lWord |= lWord << 16;
	    lWord |= lWord << 8;

	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	    /***** Gray word 7 *****/
	    iNibble = iWord & 0x000f;
	    rWord = DTFRamp[iNibble];	/* get the ramps */
	    lWord &= DTFMask[iNibble];	/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = rWord + lWord;	/* compute the continued ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    oWord = (lWord - fWord) + rWord;  /* presubtract overflow 
					       * and recalc */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
	    oWord |= fWord;		/* OR in the overflow value */
    
	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0x000000ff;
	    lWord |= lWord << 16;
	    lWord |= lWord << 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage += 4;

	}  /* for each binary word of 32 pixels */
    }

    NORMAL_RETURN;

}  /* ip_leftNeighborDTRasRow () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Computes minimum of current row and row directly below (incremented)
 *
 * LongDescription:::
 *
 * Input:::
 *	w::		width of image row (w/o frame)
 *	gBpl::		bytes oer line of the gray image
 *
 * Output:::
 *	gImage::	Pointer to first image byte of the gray destination row
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *	For the given gray row, replaces the original pixel values with
 *	the minimum of the original value and the value plus one of the
 *	pixel below it.  This is the final step of the distance transform.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_bottomNeighborDTRasRow (
    IPState    *pIPState,
    Int32	w,
    Int32	gBpl,
    UInt8      *gImage)
{
    PROLOGUE;
    UInt32	oWord, tWord, fWord, bWord;
    UInt8      *gBelow, *lastGAddr;


    /* point to the row below this one */
    gBelow = gImage + gBpl;

    /* do all gray words which are part of a corresponding binary
     * image words (including frame pixels)
     */
    lastGAddr = gImage + (((w + 31) / 32) * 32);

    while (gImage < lastGAddr)
    {
	/***** Gray word 0 *****/
        oWord = ULONG_ACCESS (gImage);

        /* take the word-wise MIN + 1 with the pixels directly below */
	/* safely increment the "below" pixels */
        bWord = ULONG_ACCESS (gBelow);
	fWord = bWord + 0x01010101;
	bWord &= ~fWord;
	bWord = (bWord >> 7) & 0x01010101;
        bWord = fWord - bWord;

	/* now take the min */
	fWord = bWord ^ oWord;
	tWord = (bWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	tWord = ((~fWord) & tWord) | (bWord & (~oWord));
	tWord &= 0x80808080;
	tWord = 0x80808080 - (tWord >> 7);
	tWord ^= 0x80808080;
	tWord &= fWord;
	oWord = bWord ^ tWord;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* point to next gray word.  Quit if we're done. */
	gImage += 4;
	gBelow += 4;

	/***** Gray word 1 *****/
        oWord = ULONG_ACCESS (gImage);

        /* take the word-wise MIN + 1 with the pixels directly below */
	/* safely increment the "below" pixels */
        bWord = ULONG_ACCESS (gBelow);
	fWord = bWord + 0x01010101;
	bWord &= ~fWord;
	bWord = (bWord >> 7) & 0x01010101;
        bWord = fWord - bWord;

	/* now take the min */
	fWord = bWord ^ oWord;
	tWord = (bWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	tWord = ((~fWord) & tWord) | (bWord & (~oWord));
	tWord &= 0x80808080;
	tWord = 0x80808080 - (tWord >> 7);
	tWord ^= 0x80808080;
	tWord &= fWord;
	oWord = bWord ^ tWord;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* point to next gray word.  Quit if we're done. */
	gImage += 4;
	gBelow += 4;

	/***** Gray word 2 *****/
        oWord = ULONG_ACCESS (gImage);

        /* take the word-wise MIN + 1 with the pixels directly below */
	/* safely increment the "below" pixels */
        bWord = ULONG_ACCESS (gBelow);
	fWord = bWord + 0x01010101;
	bWord &= ~fWord;
	bWord = (bWord >> 7) & 0x01010101;
        bWord = fWord - bWord;

	/* now take the min */
	fWord = bWord ^ oWord;
	tWord = (bWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	tWord = ((~fWord) & tWord) | (bWord & (~oWord));
	tWord &= 0x80808080;
	tWord = 0x80808080 - (tWord >> 7);
	tWord ^= 0x80808080;
	tWord &= fWord;
	oWord = bWord ^ tWord;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* point to next gray word.  Quit if we're done. */
	gImage += 4;
	gBelow += 4;

	/***** Gray word 3 *****/
        oWord = ULONG_ACCESS (gImage);

        /* take the word-wise MIN + 1 with the pixels directly below */
	/* safely increment the "below" pixels */
        bWord = ULONG_ACCESS (gBelow);
	fWord = bWord + 0x01010101;
	bWord &= ~fWord;
	bWord = (bWord >> 7) & 0x01010101;
        bWord = fWord - bWord;

	/* now take the min */
	fWord = bWord ^ oWord;
	tWord = (bWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	tWord = ((~fWord) & tWord) | (bWord & (~oWord));
	tWord &= 0x80808080;
	tWord = 0x80808080 - (tWord >> 7);
	tWord ^= 0x80808080;
	tWord &= fWord;
	oWord = bWord ^ tWord;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* point to next gray word.  Quit if we're done. */
	gImage += 4;
	gBelow += 4;

	/***** Gray word 4 *****/
        oWord = ULONG_ACCESS (gImage);

        /* take the word-wise MIN + 1 with the pixels directly below */
	/* safely increment the "below" pixels */
        bWord = ULONG_ACCESS (gBelow);
	fWord = bWord + 0x01010101;
	bWord &= ~fWord;
	bWord = (bWord >> 7) & 0x01010101;
        bWord = fWord - bWord;

	/* now take the min */
	fWord = bWord ^ oWord;
	tWord = (bWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	tWord = ((~fWord) & tWord) | (bWord & (~oWord));
	tWord &= 0x80808080;
	tWord = 0x80808080 - (tWord >> 7);
	tWord ^= 0x80808080;
	tWord &= fWord;
	oWord = bWord ^ tWord;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* point to next gray word.  Quit if we're done. */
	gImage += 4;
	gBelow += 4;

	/***** Gray word 5 *****/
        oWord = ULONG_ACCESS (gImage);

        /* take the word-wise MIN + 1 with the pixels directly below */
	/* safely increment the "below" pixels */
        bWord = ULONG_ACCESS (gBelow);
	fWord = bWord + 0x01010101;
	bWord &= ~fWord;
	bWord = (bWord >> 7) & 0x01010101;
        bWord = fWord - bWord;

	/* now take the min */
	fWord = bWord ^ oWord;
	tWord = (bWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	tWord = ((~fWord) & tWord) | (bWord & (~oWord));
	tWord &= 0x80808080;
	tWord = 0x80808080 - (tWord >> 7);
	tWord ^= 0x80808080;
	tWord &= fWord;
	oWord = bWord ^ tWord;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* point to next gray word.  Quit if we're done. */
	gImage += 4;
	gBelow += 4;

	/***** Gray word 6 *****/
        oWord = ULONG_ACCESS (gImage);

        /* take the word-wise MIN + 1 with the pixels directly below */
	/* safely increment the "below" pixels */
        bWord = ULONG_ACCESS (gBelow);
	fWord = bWord + 0x01010101;
	bWord &= ~fWord;
	bWord = (bWord >> 7) & 0x01010101;
        bWord = fWord - bWord;

	/* now take the min */
	fWord = bWord ^ oWord;
	tWord = (bWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	tWord = ((~fWord) & tWord) | (bWord & (~oWord));
	tWord &= 0x80808080;
	tWord = 0x80808080 - (tWord >> 7);
	tWord ^= 0x80808080;
	tWord &= fWord;
	oWord = bWord ^ tWord;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* point to next gray word.  Quit if we're done. */
	gImage += 4;
	gBelow += 4;

	/***** Gray word 7 *****/
        oWord = ULONG_ACCESS (gImage);

        /* take the word-wise MIN + 1 with the pixels directly below */
	/* safely increment the "below" pixels */
        bWord = ULONG_ACCESS (gBelow);
	fWord = bWord + 0x01010101;
	bWord &= ~fWord;
	bWord = (bWord >> 7) & 0x01010101;
        bWord = fWord - bWord;

	/* now take the min */
	fWord = bWord ^ oWord;
	tWord = (bWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	tWord = ((~fWord) & tWord) | (bWord & (~oWord));
	tWord &= 0x80808080;
	tWord = 0x80808080 - (tWord >> 7);
	tWord ^= 0x80808080;
	tWord &= fWord;
	oWord = bWord ^ tWord;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* point to next gray word.  Quit if we're done. */
	gImage += 4;
	gBelow += 4;

    }  /* for each binary word of 32 pixels */

    NORMAL_RETURN;

}  /* ip_bottomNeighborDTRasRow () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Right neighbor anti-raster order distance transform for one row
 *
 * LongDescription:::
 *	
 * Input:::
 *	bImage::	pointer to binary input image row (w/o frame)
 *	w::		width of image row (w/o frame)
 *	fromBlack::	cTrue if computing distance of white pixels from
 *			nearest black pixels, cFalse if computing distance
 *			of black pixels from nearest white pixels
 *	universeColor::	0 for white universe, 1 for black
 *
 * Output:::
 *	gImage::	Pointer to first image byte of the gray destination row
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *	This function essentially builds the left to right distance ramps
 *	from a row of binary input.  If the row is at an edge and the
 *	universe color affects the distance, then all distances are 1 or
 *	0 for this row.  The new result is then compared to the previous
 *	contents in the gray destination image and the minimum of the
 *	two is written out.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_rightNeighborDTAntiRasRow (
    IPState    *pIPState,
    UInt8      *bImage,
    Int32	w,
    Int32	fromBlack,
    Int32	universeColor,
    UInt8      *gImage)
{
    PROLOGUE;
    UInt32	iWord,
		oWord,
		iNibble,
		fWord,
		lWord,
		rWord,
		mWord,
		sparePix,
		first = cTrue;
    UInt8      *lastGAddr;


    lastGAddr = (gImage - 4);  /* ??? */
    gImage += ((((w - 1) / 32) * 8) + 7) * 4;
    bImage += ((w - 1) / 32) * 4;

    sparePix = w % 32;
    if (!sparePix)
	first = cFalse;

    if ((fromBlack && (universeColor == 1)) || 
        (!fromBlack && (universeColor == 0)))
    {
	/* universe color affects distance */
	/* simply put 01 where there was white and 00 where there was black */
        while (gImage > lastGAddr)
	{
	    iWord = ULONG_ACCESS (bImage);
	    if (!fromBlack)
	        iWord = ~iWord;
	    bImage -= 4;

	    if (first)
	    {
	        iWord |= ~(0xfffffffe << (31 - sparePix));
		first = cFalse;
	    }

	    /***** Gray word 7 *****/
	    iNibble = iWord & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 6 *****/
	    iNibble = (iWord >> 4) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 5 *****/
	    iNibble = (iWord >> 8) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 4 *****/
	    iNibble = (iWord >> 12) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 3 *****/
	    iNibble = (iWord >> 16) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 2 *****/
	    iNibble = (iWord >> 20) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 1 *****/
	    iNibble = (iWord >> 24) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 0 *****/
	    iNibble = (iWord >> 28) & 0x000f;
	    oWord = DTOnes[iNibble];	/* get the ones */

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;

	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	}  /* for each binary word of 32 pixels */
    }
    else  
    {
        if (universeColor == 0)
	    lWord = 0xffffffff;
	else
	    lWord = 0x00000000;

        if (!fromBlack)
	    lWord = ~lWord;

        while (gImage > lastGAddr)
	{
	    iWord = ULONG_ACCESS (bImage);
	    bImage -= 4;

	    if (!fromBlack)
	        iWord = ~iWord;

	    if (first)
	    {
		if (!fromBlack)
		{
		    if (universeColor == 1)
			iWord &= (0xfffffffe << (31 - sparePix));
		}
		else
		{
		    if (universeColor == 1)
			iWord |= ~(0xfffffffe << (31 - sparePix));
		}
		first = cFalse;
	    }


	    /***** Gray word 7 *****/
	    iNibble = iWord & 0x000f;
	    rWord = DTRRamp[iNibble];	/* get the ramps */
	    mWord = DTRMask[iNibble];	/* get the continuation mask */
	    lWord &= mWord;		/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
            oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	    oWord |= fWord;		/* overflow pixels */

	    /* now take the min of this result and the raster order result */
	    rWord = ULONG_ACCESS (gImage);
	    fWord = rWord ^ oWord;
	    iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	    iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	    iNibble &= 0x80808080;
	    iNibble = 0x80808080 - (iNibble >> 7);
	    iNibble ^= 0x80808080;
	    iNibble &= fWord;
	    oWord = rWord ^ iNibble;

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0xff000000;
	    lWord |= lWord >> 16;
	    lWord |= lWord >> 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 6 *****/
	    iNibble = (iWord >> 4) & 0x000f;
	    rWord = DTRRamp[iNibble];	/* get the ramps */
	    mWord = DTRMask[iNibble];	/* get the continuation mask */
	    lWord &= mWord;		/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
            oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	    oWord |= fWord;		/* overflow pixels */

	    /* now take the min of this result and the raster order result */
	    rWord = ULONG_ACCESS (gImage);
	    fWord = rWord ^ oWord;
	    iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	    iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	    iNibble &= 0x80808080;
	    iNibble = 0x80808080 - (iNibble >> 7);
	    iNibble ^= 0x80808080;
	    iNibble &= fWord;
	    oWord = rWord ^ iNibble;

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0xff000000;
	    lWord |= lWord >> 16;
	    lWord |= lWord >> 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 5 *****/
	    iNibble = (iWord >> 8) & 0x000f;
	    rWord = DTRRamp[iNibble];	/* get the ramps */
	    mWord = DTRMask[iNibble];	/* get the continuation mask */
	    lWord &= mWord;		/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
            oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	    oWord |= fWord;		/* overflow pixels */

	    /* now take the min of this result and the raster order result */
	    rWord = ULONG_ACCESS (gImage);
	    fWord = rWord ^ oWord;
	    iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	    iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	    iNibble &= 0x80808080;
	    iNibble = 0x80808080 - (iNibble >> 7);
	    iNibble ^= 0x80808080;
	    iNibble &= fWord;
	    oWord = rWord ^ iNibble;

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0xff000000;
	    lWord |= lWord >> 16;
	    lWord |= lWord >> 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 4 *****/
	    iNibble = (iWord >> 12) & 0x000f;
	    rWord = DTRRamp[iNibble];	/* get the ramps */
	    mWord = DTRMask[iNibble];	/* get the continuation mask */
	    lWord &= mWord;		/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
            oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	    oWord |= fWord;		/* overflow pixels */

	    /* now take the min of this result and the raster order result */
	    rWord = ULONG_ACCESS (gImage);
	    fWord = rWord ^ oWord;
	    iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	    iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	    iNibble &= 0x80808080;
	    iNibble = 0x80808080 - (iNibble >> 7);
	    iNibble ^= 0x80808080;
	    iNibble &= fWord;
	    oWord = rWord ^ iNibble;

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0xff000000;
	    lWord |= lWord >> 16;
	    lWord |= lWord >> 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 3 *****/
	    iNibble = (iWord >> 16) & 0x000f;
	    rWord = DTRRamp[iNibble];	/* get the ramps */
	    mWord = DTRMask[iNibble];	/* get the continuation mask */
	    lWord &= mWord;		/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
            oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	    oWord |= fWord;		/* overflow pixels */

	    /* now take the min of this result and the raster order result */
	    rWord = ULONG_ACCESS (gImage);
	    fWord = rWord ^ oWord;
	    iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	    iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	    iNibble &= 0x80808080;
	    iNibble = 0x80808080 - (iNibble >> 7);
	    iNibble ^= 0x80808080;
	    iNibble &= fWord;
	    oWord = rWord ^ iNibble;

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0xff000000;
	    lWord |= lWord >> 16;
	    lWord |= lWord >> 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 2 *****/
	    iNibble = (iWord >> 20) & 0x000f;
	    rWord = DTRRamp[iNibble];	/* get the ramps */
	    mWord = DTRMask[iNibble];	/* get the continuation mask */
	    lWord &= mWord;		/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
            oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	    oWord |= fWord;		/* overflow pixels */

	    /* now take the min of this result and the raster order result */
	    rWord = ULONG_ACCESS (gImage);
	    fWord = rWord ^ oWord;
	    iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	    iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	    iNibble &= 0x80808080;
	    iNibble = 0x80808080 - (iNibble >> 7);
	    iNibble ^= 0x80808080;
	    iNibble &= fWord;
	    oWord = rWord ^ iNibble;

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0xff000000;
	    lWord |= lWord >> 16;
	    lWord |= lWord >> 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	    /***** Gray word 1 *****/
	    iNibble = (iWord >> 24) & 0x000f;
	    rWord = DTRRamp[iNibble];	/* get the ramps */
	    mWord = DTRMask[iNibble];	/* get the continuation mask */
	    lWord &= mWord;		/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
            oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	    oWord |= fWord;		/* overflow pixels */

	    /* now take the min of this result and the raster order result */
	    rWord = ULONG_ACCESS (gImage);
	    fWord = rWord ^ oWord;
	    iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	    iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	    iNibble &= 0x80808080;
	    iNibble = 0x80808080 - (iNibble >> 7);
	    iNibble ^= 0x80808080;
	    iNibble &= fWord;
	    oWord = rWord ^ iNibble;

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0xff000000;
	    lWord |= lWord >> 16;
	    lWord |= lWord >> 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

            /***** Gray word 0 *****/
	    iNibble = (iWord >> 28) & 0x000f;
	    rWord = DTRRamp[iNibble];	/* get the ramps */
	    mWord = DTRMask[iNibble];	/* get the continuation mask */
	    lWord &= mWord;		/* allow only continued ramps to
					 * be affected by the last value */
	    oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	    fWord = ~oWord & lWord;	/* find bits that went 1 -> 0 */
	    fWord = fWord >> 7;		/* only need MSBs per pixel */
	    fWord &= 0x01010101;	/* remove other crap */
	    fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	    fWord ^= 0x80808080;
            oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	    oWord |= fWord;		/* overflow pixels */

	    /* now take the min of this result and the raster order result */
	    rWord = ULONG_ACCESS (gImage);
	    fWord = rWord ^ oWord;
	    iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	    iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	    iNibble &= 0x80808080;
	    iNibble = 0x80808080 - (iNibble >> 7);
	    iNibble ^= 0x80808080;
	    iNibble &= fWord;
	    oWord = rWord ^ iNibble;

	    /* write out the new data */
	    ULONG_ACCESS (gImage) = oWord;
    
	    /* set the last value word for the next step */
	    lWord = oWord & 0xff000000;
	    lWord |= lWord >> 16;
	    lWord |= lWord >> 8;
    
	    /* point to next gray word.  Quit if we're done. */
	    gImage -= 4;

	}  /* for each binary word of 32 pixels */

    }

    NORMAL_RETURN;

}  /* ip_rightNeighborDTAntiRasRow () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Top and right neighbor anti-raster distance transform for one row
 *
 * LongDescription:::
 *
 * Input:::
 *	bImage::	pointer to binary input image row (w/o frame)
 *	w::		width of image row (w/o frame)
 *	fromBlack::	cTrue if computing distance of white pixels from
 *			nearest black pixels, cFalse if computing distance
 *			of black pixels from nearest white pixels
 *	universeColor::	0 for white universe, 1 for black
 *	gBpl::		bytes oer line of the gray image
 *
 * Output:::
 *	gImage::	Pointer to first image byte of the gray destination row
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *	This function produces the right to left ramps, takes the minimum
 *	of this result and the incremented row above, and finally takes
 *	the minimum of this and the previous contents of the gray row
 *	(created by the left to right ramp builder) as the result.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_topRightNeighborDTAntiRasRow (
    IPState    *pIPState,
    UInt8      *bImage,
    Int32	w,
    Int32	fromBlack,
    Int32	universeColor,
    Int32	gBpl,
    UInt8      *gImage)
{
    PROLOGUE;
    UInt32	iWord,
		oWord, 
		iNibble, 
		fWord, 
		lWord, 
		rWord,
		mWord,
		sparePix,
		first = cTrue;
    UInt8      *gAbove, *lastGAddr;


    lastGAddr = gImage - 4;  /* ??? */
    gImage += ((((w - 1) / 32) * 8) + 7) * 4;
    bImage += ((w - 1) / 32) * 4;

    /* point to the row above this one */
    gAbove = gImage - gBpl;

    sparePix = w % 32;
    if (!sparePix)
	first = cFalse;

    if (universeColor == 0)
	/* white universe! */
	lWord = 0xffffffff;
    else
	/* black universe! */
	lWord = 0x00000000;

    if (!fromBlack)
	lWord = ~lWord;

    while (gImage > lastGAddr)
    {
        iWord = ULONG_ACCESS (bImage);
	bImage -= 4;


	if (!fromBlack)
	    iWord = ~iWord;


	if (first)
	{
	    if (!fromBlack)
	    {
		if (universeColor == 1)
		    iWord &= (0xfffffffe << (31 - sparePix));
	    }
	    else
	    {
		if (universeColor == 1)
		    iWord |= ~(0xfffffffe << (31 - sparePix));
	    }
	    first = cFalse;
	}


	/***** Gray word 7 *****/
	iNibble = iWord & 0x000f;
	rWord = DTRRamp[iNibble];	/* get the ramps */
	mWord = DTRMask[iNibble];	/* get the continuation mask */
	lWord &= mWord;			/* allow only continued ramps to
					 * be affected by the last value */
        oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	fWord = ~oWord & lWord;		/* find bits that went 1 -> 0 */
	fWord = fWord >> 7;		/* only need MSBs per pixel */
	fWord &= 0x01010101;		/* remove other crap */
	fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	fWord ^= 0x80808080;
        oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	oWord |= fWord;			/* overflow pixels */

	/* take the word-wise MIN + 1 with the pixels directly above */
	/* safely increment the "above" pixels */
        rWord = ULONG_ACCESS (gAbove);
	fWord = rWord + 0x01010101;
	rWord &= ~fWord;
	rWord = (rWord >> 7) & 0x01010101;
        rWord = fWord - rWord;

	/* now take the min of above + 1 and the reverse ramps */
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* now take the min of this result and the raster order result */
	rWord = ULONG_ACCESS (gImage);
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* set the last value word for the next step */
	lWord = oWord & 0xff000000;
	lWord |= lWord >> 16;
	lWord |= lWord >> 8;

	/* point to next gray word.  Quit if we're done. */
	gImage -= 4;
	gAbove -= 4;

	/***** Gray word 6 *****/
	iNibble = (iWord >> 4) & 0x000f;
	rWord = DTRRamp[iNibble];	/* get the ramps */
	mWord = DTRMask[iNibble];	/* get the continuation mask */
	lWord &= mWord;			/* allow only continued ramps to
					 * be affected by the last value */
        oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	fWord = ~oWord & lWord;		/* find bits that went 1 -> 0 */
	fWord = fWord >> 7;		/* only need MSBs per pixel */
	fWord &= 0x01010101;		/* remove other crap */
	fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	fWord ^= 0x80808080;
        oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	oWord |= fWord;			/* overflow pixels */

	/* take the word-wise MIN + 1 with the pixels directly above */
	/* safely increment the "above" pixels */
        rWord = ULONG_ACCESS (gAbove);
	fWord = rWord + 0x01010101;
	rWord &= ~fWord;
	rWord = (rWord >> 7) & 0x01010101;
        rWord = fWord - rWord;

	/* now take the min of above + 1 and the reverse ramps */
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* now take the min of this result and the raster order result */
	rWord = ULONG_ACCESS (gImage);
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* set the last value word for the next step */
	lWord = oWord & 0xff000000;
	lWord |= lWord >> 16;
	lWord |= lWord >> 8;

	/* point to next gray word.  Quit if we're done. */
	gImage -= 4;
	gAbove -= 4;

	/***** Gray word 5 *****/
	iNibble = (iWord >> 8) & 0x000f;
	rWord = DTRRamp[iNibble];	/* get the ramps */
	mWord = DTRMask[iNibble];	/* get the continuation mask */
	lWord &= mWord;			/* allow only continued ramps to
					 * be affected by the last value */
        oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	fWord = ~oWord & lWord;		/* find bits that went 1 -> 0 */
	fWord = fWord >> 7;		/* only need MSBs per pixel */
	fWord &= 0x01010101;		/* remove other crap */
	fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	fWord ^= 0x80808080;
        oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	oWord |= fWord;			/* overflow pixels */

	/* take the word-wise MIN + 1 with the pixels directly above */
	/* safely increment the "above" pixels */
        rWord = ULONG_ACCESS (gAbove);
	fWord = rWord + 0x01010101;
	rWord &= ~fWord;
	rWord = (rWord >> 7) & 0x01010101;
        rWord = fWord - rWord;

	/* now take the min of above + 1 and the reverse ramps */
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* now take the min of this result and the raster order result */
	rWord = ULONG_ACCESS (gImage);
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* set the last value word for the next step */
	lWord = oWord & 0xff000000;
	lWord |= lWord >> 16;
	lWord |= lWord >> 8;

	/* point to next gray word.  Quit if we're done. */
	gImage -= 4;
	gAbove -= 4;

	/***** Gray word 4 *****/
	iNibble = (iWord >> 12) & 0x000f;
	rWord = DTRRamp[iNibble];	/* get the ramps */
	mWord = DTRMask[iNibble];	/* get the continuation mask */
	lWord &= mWord;			/* allow only continued ramps to
					 * be affected by the last value */
        oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	fWord = ~oWord & lWord;		/* find bits that went 1 -> 0 */
	fWord = fWord >> 7;		/* only need MSBs per pixel */
	fWord &= 0x01010101;		/* remove other crap */
	fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	fWord ^= 0x80808080;
        oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	oWord |= fWord;			/* overflow pixels */

	/* take the word-wise MIN + 1 with the pixels directly above */
	/* safely increment the "above" pixels */
        rWord = ULONG_ACCESS (gAbove);
	fWord = rWord + 0x01010101;
	rWord &= ~fWord;
	rWord = (rWord >> 7) & 0x01010101;
        rWord = fWord - rWord;

	/* now take the min of above + 1 and the reverse ramps */
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* now take the min of this result and the raster order result */
	rWord = ULONG_ACCESS (gImage);
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* set the last value word for the next step */
	lWord = oWord & 0xff000000;
	lWord |= lWord >> 16;
	lWord |= lWord >> 8;

	/* point to next gray word.  Quit if we're done. */
	gImage -= 4;
	gAbove -= 4;

	/***** Gray word 3 *****/
	iNibble = (iWord >> 16) & 0x000f;
	rWord = DTRRamp[iNibble];	/* get the ramps */
	mWord = DTRMask[iNibble];	/* get the continuation mask */
	lWord &= mWord;			/* allow only continued ramps to
					 * be affected by the last value */
        oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	fWord = ~oWord & lWord;		/* find bits that went 1 -> 0 */
	fWord = fWord >> 7;		/* only need MSBs per pixel */
	fWord &= 0x01010101;		/* remove other crap */
	fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	fWord ^= 0x80808080;
        oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	oWord |= fWord;			/* overflow pixels */

	/* take the word-wise MIN + 1 with the pixels directly above */
	/* safely increment the "above" pixels */
        rWord = ULONG_ACCESS (gAbove);
	fWord = rWord + 0x01010101;
	rWord &= ~fWord;
	rWord = (rWord >> 7) & 0x01010101;
        rWord = fWord - rWord;

	/* now take the min of above + 1 and the reverse ramps */
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* now take the min of this result and the raster order result */
	rWord = ULONG_ACCESS (gImage);
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* set the last value word for the next step */
	lWord = oWord & 0xff000000;
	lWord |= lWord >> 16;
	lWord |= lWord >> 8;

	/* point to next gray word.  Quit if we're done. */
	gImage -= 4;
	gAbove -= 4;

	/***** Gray word 2 *****/
	iNibble = (iWord >> 20) & 0x000f;
	rWord = DTRRamp[iNibble];	/* get the ramps */
	mWord = DTRMask[iNibble];	/* get the continuation mask */
	lWord &= mWord;			/* allow only continued ramps to
					 * be affected by the last value */
        oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	fWord = ~oWord & lWord;		/* find bits that went 1 -> 0 */
	fWord = fWord >> 7;		/* only need MSBs per pixel */
	fWord &= 0x01010101;		/* remove other crap */
	fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	fWord ^= 0x80808080;
        oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	oWord |= fWord;			/* overflow pixels */

	/* take the word-wise MIN + 1 with the pixels directly above */
	/* safely increment the "above" pixels */
        rWord = ULONG_ACCESS (gAbove);
	fWord = rWord + 0x01010101;
	rWord &= ~fWord;
	rWord = (rWord >> 7) & 0x01010101;
        rWord = fWord - rWord;

	/* now take the min of above + 1 and the reverse ramps */
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* now take the min of this result and the raster order result */
	rWord = ULONG_ACCESS (gImage);
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* set the last value word for the next step */
	lWord = oWord & 0xff000000;
	lWord |= lWord >> 16;
	lWord |= lWord >> 8;

	/* point to next gray word.  Quit if we're done. */
	gImage -= 4;
	gAbove -= 4;

	/***** Gray word 1 *****/
	iNibble = (iWord >> 24) & 0x000f;
	rWord = DTRRamp[iNibble];	/* get the ramps */
	mWord = DTRMask[iNibble];	/* get the continuation mask */
	lWord &= mWord;			/* allow only continued ramps to
					 * be affected by the last value */
        oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	fWord = ~oWord & lWord;		/* find bits that went 1 -> 0 */
	fWord = fWord >> 7;		/* only need MSBs per pixel */
	fWord &= 0x01010101;		/* remove other crap */
	fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	fWord ^= 0x80808080;
        oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	oWord |= fWord;			/* overflow pixels */

	/* take the word-wise MIN + 1 with the pixels directly above */
	/* safely increment the "above" pixels */
        rWord = ULONG_ACCESS (gAbove);
	fWord = rWord + 0x01010101;
	rWord &= ~fWord;
	rWord = (rWord >> 7) & 0x01010101;
        rWord = fWord - rWord;

	/* now take the min of above + 1 and the reverse ramps */
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* now take the min of this result and the raster order result */
	rWord = ULONG_ACCESS (gImage);
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* set the last value word for the next step */
	lWord = oWord & 0xff000000;
	lWord |= lWord >> 16;
	lWord |= lWord >> 8;

	/* point to next gray word.  Quit if we're done. */
	gImage -= 4;
	gAbove -= 4;

	/***** Gray word 0 *****/
	iNibble = (iWord >> 28) & 0x000f;
	rWord = DTRRamp[iNibble];	/* get the ramps */
	mWord = DTRMask[iNibble];	/* get the continuation mask */
	lWord &= mWord;			/* allow only continued ramps to
					 * be affected by the last value */
        oWord = (lWord + rWord) & mWord; /* compute the continuation ramp */
	fWord = ~oWord & lWord;		/* find bits that went 1 -> 0 */
	fWord = fWord >> 7;		/* only need MSBs per pixel */
	fWord &= 0x01010101;		/* remove other crap */
	fWord = 0x80808080 - fWord;	/* fill overflow pixels with ff */
	fWord ^= 0x80808080;
        oWord = (lWord & ~fWord) + rWord; /* non-overflow pixels */
	oWord |= fWord;			/* overflow pixels */

	/* take the word-wise MIN + 1 with the pixels directly above */
	/* safely increment the "above" pixels */
        rWord = ULONG_ACCESS (gAbove);
	fWord = rWord + 0x01010101;
	rWord &= ~fWord;
	rWord = (rWord >> 7) & 0x01010101;
        rWord = fWord - rWord;

	/* now take the min of above + 1 and the reverse ramps */
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* now take the min of this result and the raster order result */
	rWord = ULONG_ACCESS (gImage);
	fWord = rWord ^ oWord;
	iNibble = (rWord | 0x80808080) - (oWord & 0x7f7f7f7f);
	iNibble = ((~fWord) & iNibble) | (rWord & (~oWord));
	iNibble &= 0x80808080;
	iNibble = 0x80808080 - (iNibble >> 7);
	iNibble ^= 0x80808080;
	iNibble &= fWord;
	oWord = rWord ^ iNibble;

	/* write out the new data */
	ULONG_ACCESS (gImage) = oWord;

	/* set the last value word for the next step */
	lWord = oWord & 0xff000000;
	lWord |= lWord >> 16;
	lWord |= lWord >> 8;

	/* point to next gray word.  Quit if we're done. */
	gImage -= 4;
	gAbove -= 4;

    }  /* for each binary word of 32 pixels */

    NORMAL_RETURN;

}  /* ip_topRightNeighborDTAntiRasRow () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Fixes (zeroes) right frame pixels of 8-bit image row
 *
 * LongDescription:::
 *
 * Input:::
 *	w::		Width of image w/o frame
 *
 * Output:::
 *	gImage::	Pointer to first image byte of the row to be fixed
 *
 * Returns:::
 *	ia_successful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
Int32 CDECL 
ip_fixRightFrameRow (
    IPState    *pIPState,
    UInt8      *gImage,
    Int32	w)
{
    PROLOGUE;
    Int32	sparePix, spareWords, dp;

    if ((w / 32) > 0)
    {
	dp = (((w / 32) + 1) * 32);
    
	gImage += (dp - 4);
    
	dp -= w;
    
	sparePix = (dp % 4);
    
	spareWords = (dp / 4);
    
	for (; spareWords > 0; spareWords--)
	{
	    ULONG_ACCESS (gImage) = 0x00000000;
	    gImage -= 4;
	}
    
	if (sparePix)
	    ULONG_ACCESS (gImage) &= (0xffffff00 << ((sparePix - 1) * 8));
    }

    NORMAL_RETURN;

}  /* ip_fixRightFrameRow () */


/*
 * $Log: midtran4.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 * Imported sources
 *
 * Revision 1.3  1998/01/13 23:52:06  ddavies
 * midtran4.c -> /product/ipcore/ipshared/src/RCS/midtran4.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.2  1997/01/03 17:52:18  dferrell
 * dferrell on Fri Jan 3 09:51:30 PST 1997
 *
 * midtran4.c -> /product/ipcore/ipshared/src/RCS/midtran4.c,v
 *
 * Removed unused parameters from ip_bottomNeighborDTRasRow () interface.
 *
 * Revision 1.1  1996/12/31 22:02:26  dferrell
 * Initial revision
 *
 *
 */

