/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miexpndm.c: machine independent primitives for
 *                expand-and-mask operations.
 *
 *         2x, 4x, 8x expansion procedures:
 *		Int32		ip_expandAndMask2Asm()
 *		Int32		ip_expandAndMask4Asm()
 *		Int32		ip_expandAndMask8Asm()
 *
 *	   No expansions larger than 8x are implemented here because
 *	   no one seems to use them, especially for masking full scale
 *	   images.
 */


#include <stdio.h>
#include <stdlib.h>
#if __WATCOMC__
#include <string.h>
#endif
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototypes */
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miexpndm.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");
/* default: suppress all debugging */
#define  TIMING         0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING         1
#endif  /* DEBUG_ON */

/* The lookup tables used by the expansion code are generated in exptable.c
 *
 * A common operation in our style of image processing is to make a mask
 * at some reduced scale, expand it to full scale and use it to mask the
 * original image.  This requires another full scale image.  In the
 * interest of saving memory, we'll introduce a new set of operations that
 * expand the image piece by piece, mask the original with the expanded
 * pieces, then toss the expanded image.
 *
 * These routines will work a lot like the expansion routines, except the
 * destination pointers point to the image that will be masked.
 *
 * We depend on the fact that reduction always rounds down in image size.
 * This means the number of lines in the expanded image will always be less
 * than or equal to the number of lines in the destination image.  That
 * means we don't need to compute the minimum of these numbers.
 */


/*********************************************************************
 ************       2x Expand and Mask Procedures       **************
 ********************************************************************/

/* PRIVATE */
/*
 *  ip_expandAndMask2Asm()
 *              Each pixel in the source pr is mapped to a 2x2 square
 *              of 4 identical pixels, ANDed with the original dest
 *		pixels and discarded.
 *
 *    
 *	Input:  
 *		ps =	 pointer to upper left corner of source binary image
 *			 (does not include frame)
 *		sBpl =	 source bytes/line
 *		pd =	 pointer to upper left corner of dest binary image
 *			 (does not include frame)
 *		dBpl =	 destination bytes/line
 *		w =	 width of source image ( does not include frame)
 *		h =	 height of source image (does not include frame)
 *              pTabExp2 pointer to expansion table 
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.  This is a C implementation which is
 *	normally an assembly routine - hence the name.
 */
Int32  CDECL 
ip_expandAndMask2Asm (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt16     *pTabExp2)
{

    PROLOGUE;
    UInt8      *pSByte;
    UInt32	expandedWord;
    UInt8      *pDByte;
    Int32	maxDestOffset, maxSrcOffset;

    maxDestOffset = ( (2*w) / 32 ) << 2;	/* addr of rightmost dest wd */
    maxSrcOffset = maxDestOffset / 2;		/* addr of rightmost src wd */

	/* For each source row, we expand and AND its columns with the
	 * next two destination rows.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the next.
	 */
    for (; h > 0; h--)
    {
        pSByte = ps + maxSrcOffset;
        pDByte = pd + maxDestOffset;
        while ((UInt32) pSByte >= (UInt32) ps)
	{
                /* Expand each byte into a halfword.  Do this
		 * twice, then AND the resulting word with the
		 * dest words. */
	    expandedWord = (pTabExp2[UCHAR_ACCESS(&pSByte[0])] << 16) |
			    pTabExp2[UCHAR_ACCESS(&pSByte[1])];
	    pSByte -= 2;
            ULONG_ACCESS(&pDByte[0]) &= expandedWord;
	    ULONG_ACCESS(&pDByte[dBpl]) &= expandedWord;
	    pDByte -= 4;
        }
	ps += sBpl;
	pd += dBpl + dBpl;
    }

    NORMAL_RETURN;
}


/*********************************************************************
 *************      4x  Expand and Mask Procedures       *************
 ********************************************************************/

/* PRIVATE */
/*
 *  ip_expandAndMask4Asm()   
 *              Each pixel in the source pr is mapped to a 4x4 square
 *              of 16 identical pixels, ANDed with the original dest
 *		pixels and discarded.
 *
 *	Input:  
 *		ps =	   pointer to upper left corner of source binary image
 *			   (does not include frame)
 *		sBpl =	   source bytes/line
 *		pd =	   pointer to upper left corner of dest binary image
 *			   (does not include frame)
 *		dBpl =	   destination bytes/line
 *		w =	   width of source image ( does not include frame)
 *		h =	   height of source image (does not include frame)
 *              pTabExp4 = pointer to expansion table
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.  This is a C implementation which is
 *	normally an assembly routine - hence the name.
 */
Int32  CDECL 
ip_expandAndMask4Asm (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt32     *pTabExp4)
     
{
    PROLOGUE;
    UInt8      *pSByte;
    UInt8      *pDByte;
    UInt32	expandedWord;
    UInt32	maxSrcOffset, maxDestOffset;

    maxDestOffset = ( (4*w) / 32 ) << 2;	/* addr of rightmost dest wd */
    maxSrcOffset = maxDestOffset / 4;		/* addr of rightmost src wd */

	/* For each source row, we expand its columns and AND them with the
	 * next four destination rows.
	 */
    for (; h > 0; h--)
    {
        pSByte = ps + maxSrcOffset;
        pDByte = pd + maxDestOffset;
        while ((UInt32) pSByte >= (UInt32) ps)
	{

                /* obtain 8 bits from the source */
                /* get the corresponding 32 bits from the table,
                 * and put it in the destination byte */
	    expandedWord = pTabExp4[UCHAR_ACCESS(&pSByte[0])];
	    pSByte--;
            ULONG_ACCESS(&pDByte[0]) &= expandedWord;
            ULONG_ACCESS(&pDByte[dBpl]) &= expandedWord;
            ULONG_ACCESS(&pDByte[2*dBpl]) &= expandedWord;
            ULONG_ACCESS(&pDByte[3*dBpl]) &= expandedWord;
	    pDByte -= 4;
        }

	ps += sBpl;
	pd += 4 * dBpl;		/* skip four output lines */
    }

    NORMAL_RETURN;
}



/*********************************************************************
 *************      8x  Expand and Mask Procedures       *************
 ********************************************************************/
/* PRIVATE */
/*
 *  ip_expandAndMask8Asm()
 *              Each pixel in the source pr is mapped to a 8x8 square
 *              of 64 identical pixels, ANDed with the original dest
 *		pixels and discarded.
 *
 *    
 *	Input:  
 *		ps =	   pointer to upper left corner of source binary image
 *			   (does not include frame)
 *		sBpl =	   source bytes/line
 *		pd =	   pointer to upper left corner of dest binary image
 *			   (does not include frame)
 *		dBpl =	   destination bytes/line
 *		w =	   width of source image ( does not include frame)
 *		h =	   height of source image (does not include frame)
 *              pTabExp8 = pointer to expansion table
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.  This is a C implementation which is
 *	normally an assembly routine - hence the name.
 */
Int32  CDECL 
ip_expandAndMask8Asm (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt32     *pTabExp8)
{
    PROLOGUE;
    UInt8      *pSByte;
    UInt8      *pDByte;
    UInt32	expandedWord;
    UInt32	maxSrcOffset, maxDestOffset;
    UInt32	lineOffset;
    UInt32      *tabExp8Hi;		/* pts to upper half
						 * to expansion tab */

    maxDestOffset = ( (8*w) / 32 ) << 2;	/* addr of rightmost dest wd */
    maxSrcOffset = maxDestOffset / 8;		/* addr of rightmost src wd */

	/* point to upper half of expansion table */
    tabExp8Hi = pTabExp8 + 256;

	/* For each source row, we expand its columns and AND with the
	 * next 8 destination rows  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (; h > 0; h--)
    {
        pSByte = ps + maxSrcOffset;
        pDByte = pd + maxDestOffset;
	    /* Is the rightmost dest word produced from the LS nibble
	     * or the MS nibble of the source word?  If maxDestOffset
	     * points to an even word, we need the MS nibble.  Otherwise,
	     * we need the LS nibble.  The MS nibble is the special case
	     * since we do only one nibble before falling through.
	     */
	if (!(maxDestOffset & 0x4))
	{
	    expandedWord = tabExp8Hi[UCHAR_ACCESS(&pSByte[0])];
	    pSByte--;
	    ULONG_ACCESS(&pDByte[0]) &= expandedWord;
	    ULONG_ACCESS(&pDByte[dBpl]) &= expandedWord;
	    lineOffset = 2*dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    pDByte -= 4;
	}

        while ((UInt32) pSByte >= (UInt32) ps)
	{
	    expandedWord = pTabExp8[UCHAR_ACCESS(&pSByte[0])];
	    ULONG_ACCESS(&pDByte[0]) &= expandedWord;
	    ULONG_ACCESS(&pDByte[dBpl]) &= expandedWord;
	    lineOffset = 2*dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    pDByte -= 4;

	    expandedWord = tabExp8Hi[UCHAR_ACCESS(&pSByte[0])];
	    pSByte--;
	    ULONG_ACCESS(&pDByte[0]) &= expandedWord;
	    ULONG_ACCESS(&pDByte[dBpl]) &= expandedWord;
	    lineOffset = 2*dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    lineOffset += dBpl;
	    ULONG_ACCESS(&pDByte[lineOffset]) &= expandedWord;
	    pDByte -= 4;
        }

	ps += sBpl;
	pd += 8*dBpl;			/* skip 8 dest lines */
    }

    NORMAL_RETURN;
}

/*

$Log: miexpndm.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.4  1998/01/13 23:52:09  ddavies
miexpndm.c -> /product/ipcore/ipshared/src/RCS/miexpndm.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1995/12/13 02:48:56  ejaquith
miexpndm.c -> /product/ipcore/ipshared/src/RCS/miexpndm.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1994/07/27  15:52:18  ejaquith
 * miexpndm.c -> /product/ipcore/ipshared/src/RCS/miexpndm.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Added pointers to table bytes to all functions in place of globals.
 *
 * Revision 1.1  1994/01/06  03:31:59  ddavies
 * Initial revision
 *
 * Revision 1.8  1993/12/02  19:26:36  ddavies
 * sexpandm.c -> /product/ipcore/ipshared/src/RCS/sexpandm.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.7  1993/11/05  02:21:34  ddavies
 * sexpandm.c -> /product/ipcore/ipshared/src/RCS/sexpandm.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


