/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* 
 *  miframe.c:  C implementation of lower level of frame routines:
 *	   Int32    ip_addFrameToLine()
 *	   Int32    ip_removeFrameFromLine()
 *	   Int32    ip_clearRightFrameBlock()
 */

#include <stddef.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /* prototypes */
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miframe.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");


/* ip_addFrameToLine basically does one line of bitblt where the only
 * possible shift distances are multiples of 8.  Rather than trying to
 * code up a fancy C function, we'll just use memcpy.  Note that we always
 * have to round the amount of data to be transferred up to a multiple of
 * 4 bytes so that the right data will be transferred even on the PC.
 * If we put extra data into the right frame, we'll clear it immediately
 * anyway.
 *
 * NOTE: we assume that the source image was padded with 0's, not 1's or
 * random trash.
 *
 */


/* PRIVATE */
/********************************************************************
 *  ip_addFrameToLine()
 *	Input:
 *		dPtr =		pointer to left end of dest line	
 *		sPtr =		pointer to left end of source line	
 *		imageWidth =	exact width of line in bytes
 *		frameWidth =	Number of bytes in left (or right) frame.
 *				Equals cFrameBytes * depth
 *		rightMask =	mask used to clear bits on right end
 *				of image line.
 *
 *      On return:
 *		dest line has the source line with cFrameBits pixels of  
 *		frame added on left and right sides.
 *
 *
 ********************************************************************/
void   CDECL 
ip_addFrameToLine (
    IPState    *pIPState,
    UInt8Ptr	dPtr,
    UInt8Ptr	sPtr,
    UInt32	imageWidth,
    UInt32	frameWidth,
    UInt32	rightMask) 
{
    PROLOGUE;
    UInt32	leftShift;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* clear the left frame, transfer the data in units of words,
	 * then clear the right frame.  Note that we assume that
	 * any pad bits to the right of the rightmost source image bit
	 * in its byte were 0's.  We clear out the right frame and pad
	 * starting on the next byte boundary.
	 */
    memset((char *)dPtr, 0, frameWidth);
    dPtr += frameWidth;			/* point to beginning of image */

	/* copy the image bytes.  Do using a little rasterOp fragment
	 * that shifts things in units of words.  memcpy gets the
	 * order of the bytes wrong on little endian machines.  It
	 * also shifts the words in the wrong direction.  We could
	 * call rasterOp for each line, but that seems like a real
	 * waste of energy.
	 * Since we know that dPtr points to a word boundary, we know
	 * that the left shift distance is equal to (sPtr & 0x3) * 8.
	 */
    leftShift = ((UInt32)sPtr & 0x3) * 8;

	/* if no shifting is required, or we're on a big endian machine,
	 * do it the fast way */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt 

    memcpy((char *) dPtr, (char *)sPtr, (imageWidth + 3) & ~0x3);

#else	/* Use little endian image format */

    if (leftShift == 0)
    {
	memcpy((char *) dPtr, (char *)sPtr, (imageWidth + 3) & ~0x3);
    }
    else
    {
	    /* process the line from right to left to keep the inner
	     * loop small.  oldInputWord is the source word to the
	     * right of newInputWord.  To form an output word, we
	     * combine <word to the left> << leftShift |
	     * <word to the right> >> rightShift.
	     */
	Int32	 offsetInBytes, rightShift;
	UInt32	 oldInputWord, newInputWord;

	rightShift = 32 - leftShift;
	sPtr -= (UInt32)sPtr & 0x3;		/* point to word boundary */
	offsetInBytes = (imageWidth - 1) & ~0x3;
	oldInputWord = ULONG_ACCESS(&sPtr[offsetInBytes+4]) >> rightShift;

	for ( ; offsetInBytes >= 0; offsetInBytes -= 4)
	{
	    newInputWord = ULONG_ACCESS(&sPtr[offsetInBytes]);
	    ULONG_ACCESS(&dPtr[offsetInBytes]) = oldInputWord |
					       (newInputWord << leftShift);
	    oldInputWord = (newInputWord >> rightShift);
	}
    }
#endif /* _ALPACA_IMAGE_FMT_ */

    dPtr += ((imageWidth + 3) & ~0x3);	/* point to first full frame word */

	/* clear any remaining bits in the rightmost word */
    ULONG_ACCESS(&dPtr[-4]) &= rightMask;

	/* clear remaining right frame and pad */
    memset((char *)dPtr, 0, frameWidth);
}


/* PRIVATE */
/********************************************************************
 *  ip_removeFrameFromLine()
 *		This routine copies one line of source image to
 *		dest image.  The frame in the source image is ignored.
 *		Assumes source line width is a multiple of 32.
 *		This is basically a routine to do an efficient copy
 *		of blocks of memory.  The source starts
 *		inside the frame and ends before the next frame, so
 *		it really doesn't have a lot to do with frames.  The
 *		assembly version on the SPARC tries to use double
 *		word writes for speed.
 *		  
 *	Input:  
 *		dPtr =		Pointer to left edge of dest line
 *		sPtr =		Pointer to left edge of source line
 *				  INSIDE frame.
 *		imageBytes =	Number of bytes to transfer
 *
 *	On return: 
 *		dest line has the frame-less part of the source line
 *		padded with 0's on the right to the next word boundary.
 *		If the next dest line starts before the next word
 *		boundary, (part of) the pad will be convered up next
 *		time.
 *
 ********************************************************************/
void   CDECL 
ip_removeFrameFromLine (
    IPState    *pIPState,
    UInt8      *dPtr,
    UInt8      *sPtr,
    UInt32	imageBytes)
{
    PROLOGUE;
    UInt32	rightShift;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* copy the image bytes.  Do using a little rasterOp fragment
	 * that shifts things in units of words.  memcpy gets the
	 * order of the bytes wrong on little endian machines.  It
	 * also shifts the words in the wrong direction.  We could
	 * call rasterOp for each line, but that seems like a real
	 * waste of energy.
	 * Since we know that sPtr points to a word boundary, we know
	 * that the right shift distance is equal to (dPtr & 0x3) * 8.
	 */
    rightShift = ((UInt32)dPtr & 0x3) * 8;

	/* if no shifting is required, or we're on a big endian machine,
	 * do it the fast way.  Note that since the source frame is
	 * filled with 0's, padding to the next word boundary
	 * happens automatically.
	 */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt 

    memcpy((char *) dPtr, (char *)sPtr, (imageBytes + 3) & ~0x3);

#else	/* Use little endian image format */

    if (rightShift == 0)
	/* Why does this work?  Because we can scribble on a few bytes
	 * beyond the end of the real destination.  Those will be
	 * overwritten in the next pass.  So long as we transfer
	 * things in units of full words, all the bytes we really
	 * meant to transfer will get to the right place.
	 * Note that since the source frame is
	 * filled with 0's, padding to the next word boundary
	 * happens automatically.
	 */
    {
	memcpy((char *) dPtr, (char *)sPtr, (imageBytes + 3) & ~0x3);
    }
    else
    {
	    /* process the line from right to left to keep the inner
	     * loop small.  Since we're generally going to shift
	     * right and we want to use a single offset for both
	     * source and dest words AND we have to do special merging
	     * with the leftmost dest word, make the dest pointer
	     * point to the next higher word boundary.  Each dest
	     * word will be formed from (oldOutputWord |
	     * sourceWord << leftShift) and oldOutputWord =
	     * sourceWord >> rightShift.  Like this:
	     *
	     *	dest |----|----|----|----|----|----|----|----|
	     *		 ^dPtr[0]			  ^dPtr[imageBytes-1]
	     *  src  |----|----|----|----|----|----|----|----|
	     *	      ^sPtr[0]			      ^sPtr[imageBytes-1]
	     * leftShift = 8, rightShift = 24; dPtr is set to point
	     * to the second word in the line called dest and we
	     * process from right to left, i.e. from the dest word
	     * containing dPtr[imageBytes-1] and the source word containing
	     * sPtr[imageBytes-1] down to the left.  We stop before touching
	     * the word containing dPtr[0].  That word is processed by
	     * merging the shifted source word with any previous dest
	     * bytes in that word.
	     * Note that since we start with oldInputWord set to 0,
	     * the rightmost word is padded with 0's.
	     */
	Int32	 offsetInBytes, leftShift;
	UInt32	 newInputWord, oldInputWord;

	leftShift = 32 - rightShift;
	dPtr = (UInt8 *)(((UInt32)dPtr & ~0x3) + 4); /* point to NEXT word boundary */
	offsetInBytes = (imageBytes - 1) & ~0x3;
	oldInputWord = 0;

	for ( ; offsetInBytes >= 0; offsetInBytes -= 4)
	{
	    newInputWord = ULONG_ACCESS(&sPtr[offsetInBytes]);
	    ULONG_ACCESS(&dPtr[offsetInBytes]) = (newInputWord << leftShift) |
						 oldInputWord;
	    oldInputWord = newInputWord >> rightShift;
	}
	    /* merge the last output word */
	newInputWord = ULONG_ACCESS(&dPtr[offsetInBytes]);
	newInputWord &= (0xFFFFFFFF << leftShift); /* preserve the old stuff */
	ULONG_ACCESS(&dPtr[offsetInBytes]) = newInputWord | oldInputWord;
    }
#endif /* _ALPACA_IMAGE_FMT_ */

}


/********************************************************************
 *
 *  ip_clearRightFrameBlock()
 *	Input:
 *      	sPtr = 	     pointer to uppermost word to be cleared
 *		sBpl =	     bytes/line (includes frame)
 *		sH =	     exact height of the block being cleared
 *		imageMask =  used to preserved image bits in rightmost word
 *		frameMask =  used to detect bits in frame portion of
 *			     rightmost word.  Yes, imageMask = ~frameMask
 *			     but parameters are cheap, right?
 *
 * This routine clears the *first 32 bit word* of the right frame independent
 * of the depth of the image.  It doesn't really clear the entire right
 * frame, but it does touch the part of the frame that's most likely to
 * be soiled.
 *
 *	On return:
 *		Zeros have been put into the frame portion of the rightmost
 *		word of the image.
 *
 *	Return code:
 *		0	     always
 *
 ********************************************************************/
Int32   CDECL 
ip_clearRightFrameBlock (
    IPState    *pIPState,
    UInt8      *sPtr,
    Int32	sBpl,
    Int32	sH,
    UInt32	imageMask,
    UInt32	frameMask)
{
    PROLOGUE;
    UInt32	imageWord;
    UInt8      *finalPtr;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* set the stopping condition.  This removes one increment
	 * from the loop (frees a register?)
	 */
    finalPtr = sPtr + sH * sBpl;

	/* let's just do it */
    for ( ; sPtr != finalPtr; sPtr += sBpl)
    {
	imageWord = ULONG_ACCESS(sPtr);
	    /* most of the time, the frame hasn't been soiled and
	     * writes are sort of expensive, so test before writing.
	     */
	if (imageWord & frameMask)
	    ULONG_ACCESS(sPtr) = imageWord & imageMask;
    }

    return 0;
}


/*

 $Log: miframe.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.8  1998/01/13 23:52:17  ddavies
 miframe.c -> /product/ipcore/ipshared/src/RCS/miframe.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.7  1995/12/13 02:49:14  ejaquith
 miframe.c -> /product/ipcore/ipshared/src/RCS/miframe.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/10/18  04:03:13  ddavies
 * miframe.c -> /product/ipcore/ipshared/src/RCS/miframe.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.5  1994/11/07  22:06:08  ddavies
 * miframe.c -> /product/ipcore/ipshared/src/RCS/miframe.c,v
 *
 * Remove irrelevant #ifndef __GNUC__ around inclusion of string.h.
 *
 * Revision 1.4  1994/07/11  23:15:43  ddavies
 * miframe.c -> /product/ipcore/ipshared/src/RCS/miframe.c,v
 *
 * Replace the array rightMask with a variable rightMask that's passed down
 * from i_addFrame to ip_addFrameToLine.  This allows the latter routine to
 * correctly clear exactly the frame bits in the rightmost image word.  We
 * had been assuming that unframed images are padded on the right to byte
 * boundaries with 0's.  This is not true of images that have been rotated
 * in place.
 *
 * Revision 1.3  1994/01/04  03:09:22  ddavies
 * miframe.c -> /product/ipcore/ipshared/src/RCS/miframe.c,v
 *
 * Changed comment to say this is miframe.c, not miborder.c.
 *
 * Revision 1.2  1993/12/22  04:58:11  ddavies
 * miframe.c -> /product/ipcore/ipshared/src/RCS/miframe.c,v
 *
 * Change implementation of ip_removeFrameFromLine so it can deal with
 * frameless images whose bytes/line is not a multiple of 4.
 *
 * Revision 1.1  1993/12/14  21:25:42  ddavies
 * Initial revision
 *

 */


