/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  migendil.c -- generates stub for assembly routines
 *                that would have implemented the dilation operations for a
 *                specified structuring element
 */


#include <stdlib.h>
#include <stddef.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "frames.pub"

    /* prototypes */
#include "shrmorph.pub"
#include "shrmorph.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: migendil.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/************************************************************************
 *  i_getDilateAsmSize():  return number of bytes needed for assembly code
 *			used for dilation with this strel.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getDilateAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
	/* reference all the parms and say the buffer size is 0 */
    PROLOGUE;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    *pNumAsmBytes = 0;

    __error__ = ia_noImpl;
    NORMAL_RETURN;
}


/* PUBLIC */
/************************************************************************
 *  i_getLongDilateAsmSize():  return number of bytes needed for assembly code
 *			used for dilation with this strel using Long (>= 64 bit)
 *			opcodes.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getLongDilateAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
	/* reference all the parms and say the buffer size is 0 */
    PROLOGUE;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    *pNumAsmBytes = 0;

    __error__ = ia_noImpl;
    NORMAL_RETURN;
}



/* This is a public procedure */
/* PUBLIC */
/************************************************************************
 *  i_genDilateAsm():  generate assembly code to do dilate defined by strel
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastDilate as
 *			blockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns ia_noImpl.
 *
 * This code is used in place of C code that creates processor-specific
 * erosion routines.  This code is just a stub.  If no erosion routine is
 * made, erosion is done using rasta_rop's.
 *
 ************************************************************************/
Int32 CDECL 
i_genDilateAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    DilateProc *generated_code,
    Int32      *code_length)
{  
	/* keep the compiler happy by referencing all the unused parms. */
    PROLOGUE;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    data = data;
    blockWidth = blockWidth;
    *code_length = 0;
    *generated_code = NULL;	/* we don't know how to generate code for
				   this processor, so punt */
    __error__ = ia_noImpl;
    NORMAL_RETURN;
}


/* PUBLIC */
/************************************************************************
 *  i_genLongDilateAsm():  generate assembly code to do dilate defined by strel.
 *			This assembly code makes use of long memory accesses
 *			and operations (long >= 64 bits)
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastDilate as
 *			longBlockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *			If the buffer was passed in, we assume it has at
 *			least the number of bytes that i_getLongDilateAsmSize
 *			would calculate.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_SETooBig:		Structure element was too big to
 *				generate accelerated code OR was composed
 *				entirely of DONT_CARE and MISS terms.
 *	ia_noImpl:		The current platform doesn't support long ops.
 *	ia_nomem:		Not enough memory to allocate the code buffer.
 *	ia_bufferExceeded:	The routine tried to generate too much
 *				assembly code.  This is a bug in the routine.
 *				If it happens, please let us know.
 ************************************************************************/
Int32  CDECL 
i_genLongDilateAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    LongDilateProc *generated_code,
    Int32      *code_length)
{
	/* keep the compiler happy by referencing all the unused parms. */
    PROLOGUE;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    data = data;
    blockWidth = blockWidth;
    *code_length = 0;
    *generated_code = NULL;	/* we don't know how to generate code for
				   this processor, so punt */
    __error__ = ia_noImpl;
    NORMAL_RETURN;
}


/* PRIVATE */
/************************************************************************
 *  ip_getLongDilateSEParms():
 *			return dimensions of input cache needed for long
 *			dilate ops on a strel along with an array of the
 *			active columns in the SE.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	pSE:		Pointer to structure element's data.
 *	pCacheWidth:	Pointer to returned width of cache in bytes.
 *			This may be NULL.
 *	pCacheHeight:	Pointer to returned height of input cache.
 *			This may be NULL.
 *	pNumActiveCols:	Pointer to returned number of active columns.
 *			This may be NULL.
 *	pActiveCols:	Pointer to returned array of nx bytes.  This
 *			array should be allocated (and freed) by the caller.
 *			This routine will fill each element in the array
 *			with cHit if the corresponding SE column contains
 *			at least one hit and 0 otherwise.  This may be NULL.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
ip_getLongDilateSEParms (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    const UInt8 *pSE,
    Int32      *pCacheWidth,
    Int32      *pCacheHeight,
    Int32      *pNumActiveCols,
    UInt8      *pActiveCols)
{
    PROLOGUE;
    nx = nx;			/* keep compiler happy. */
    ny = ny;
    pSE = pSE;
    pCacheWidth = pCacheWidth;
    pCacheHeight = pCacheHeight;
    pNumActiveCols = pNumActiveCols;
    pActiveCols = pActiveCols;

	/* This routine should never be called on platforms that have
	 * no long dilate ops, so return ia_noImpl.
	 */
    __error__ = ia_noImpl;
    NORMAL_RETURN;
}


/*

$Log: migendil.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.8  1998/01/13 23:52:19  ddavies
migendil.c -> /product/ipcore/ipshared/src/RCS/migendil.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1996/12/19 22:36:49  ddavies
ddavies on Thu Dec 19 14:34:44 PST 1996

migendil.c -> /product/ipcore/ipshared/src/RCS/migendil.c,v

Have dummy implementations of i_getDilateAsmSize and i_getLongDilateAsmSize
return ia_noImpl instead of ia_successful so upper levels of code know
not to allocate a buffer of size 0.  Purify refuses to free a buffer of
size 0, then complains that it's been leaked.

Revision 1.6  1996/12/18 23:01:01  ddavies
migendil.c -> /product/ipcore/ipshared/src/RCS/migendil.c,v

Remove the mysterious ifdef cLongFrameBits.  Who knows what that was for?
Add dummy implementations of i_getLongDilateAsmSize,
i_genLongDilateAsm and ip_getLongDilateSEParms.  These routines are not
really present in the machine independent implementations, but the linker
wants to see them.

Revision 1.5  1995/12/13 02:49:18  ejaquith
migendil.c -> /product/ipcore/ipshared/src/RCS/migendil.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1995/10/03  21:13:48  ddavies
 * migendil.c -> /product/ipcore/ipshared/src/RCS/migendil.c,v
 *
 * Include shrmorph.pub/prv instead of shrip.pub/prv.  Change arg of i_genDilateAsm
 * from "void **generated_code" to "DilateProc generated_code".
 *
 * Revision 1.3  1994/09/20  02:47:53  ddavies
 * migendil.c -> /product/ipcore/ipshared/src/RCS/migendil.c,v
 *
 * Add dummy implementation of i_getDilateAsmSize.  Update comments and args
 * for dummy implementation of i_genDilateAsm to delete allocator and alloc_shrinker.
 *
 * Revision 1.2  1994/01/13  17:54:36  dferrell
 * migendil.c -> /product/ipcore/ipshared/src/RCS/migendil.c,v
 *
 * Fixed so that "sparc.prv" is included only when SPARC_ENV is defined.
 *
 * Revision 1.1  1994/01/05  03:57:36  ddavies
 * Initial revision
 *
 * Revision 1.14  1993/12/02  19:26:52  ddavies
 * sgendil.c -> /product/ipcore/ipshared/src/RCS/sgendil.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.13  1993/11/05  02:22:05  ddavies
 * sgendil.c -> /product/ipcore/ipshared/src/RCS/sgendil.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


