/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  migenero.c -- generates stub for assembly routines
 *                that would have implemented the erosion operations for a
 *                specified structuring element
 */


#include <stdlib.h>
#include <stddef.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#ifdef SPARC_ENV
#include "sparc.prv"
#endif
#include "frames.pub"

    /* prototypes */
#include "shrmorph.pub"
#include "shrmorph.prv"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: migenero.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* PRIVATE */
/*************************************************************************
 *  ip_longOpIPLibLinked()
 *		There are things we'd really not do if routines capable
 *		of long (64 bit or longer) are linked, like avoid
 *		making superflous tables.  This routine returns a
 *		boolean which is set to cTrue if the
 *		long memory access libraries are linked.
 *
 *	Input:
 *		None
 *	Output:
 *		cTrue if fast libraries are linked, cFalse otherwise.
 *************************************************************************/
Bool  CDECL 
ip_longOpIPLibLinked (
    IPState    *pIPState)
{
    PROLOGUE;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Clearly the fast libs are NOT here since we're here. */
    return cFalse;
}


/* PUBLIC */
/************************************************************************
 *  i_getErodeAsmSize():  return number of bytes needed for assembly code
 *			used for erosion with this strel.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getErodeAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
	/* reference all the parms and return a buffer size of 0 since there
	 * is no assembly code here */
    PROLOGUE;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    *pNumAsmBytes = 0;
    __error__ = ia_noImpl;
    NORMAL_RETURN;
}


/* PUBLIC */
/************************************************************************
 *  i_getLongErodeAsmSize():  return number of bytes needed for assembly code
 *			used for erosion with this strel using Long (>= 64 bit)
 *			opcodes.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getLongErodeAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
	/* reference all the parms and return a buffer size of 0 since there
	 * is no assembly code here */
    PROLOGUE;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    *pNumAsmBytes = 0;
    __error__ = ia_noImpl;
    NORMAL_RETURN;
}


/* PRIVATE */
/************************************************************************
 *  ip_getLongErodeSEParms():
 *			return dimensions of input cache needed for long
 *			erode ops on a strel along with an array of the
 *			active columns in the SE.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	pSE:		Pointer to structure element's data.
 *	pCacheWidth:	Pointer to returned width of cache in bytes.
 *			This may be NULL.
 *	pCacheHeight:	Pointer to returned height of input cache.
 *			This may be NULL.
 *	pNumActiveCols:	Pointer to returned number of active columns.
 *			This may be NULL.
 *	pActiveCols:	Pointer to returned array of nx bytes.  This
 *			array should be allocated (and freed) by the caller.
 *			This routine will fill each element in the array
 *			with 0 if there are no SE elements in the column,
 *			cHit if the SE column contains only HITs and don't
 *			care terms, cMiss if the column contains only MISSes
 *			and don't care terms and (cHit | cMiss) if the
 *			column contains both HITs and MISSes (and possibly
 *			don't care terms).  This may be NULL.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
ip_getLongErodeSEParms (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    const UInt8 *pSE,
    Int32      *pCacheWidth,
    Int32      *pCacheHeight,
    Int32      *pNumActiveCols,
    UInt8      *pActiveCols)
{
    PROLOGUE;
    nx = nx;			/* keep compiler happy. */
    ny = ny;
    pSE = pSE;
    pCacheWidth = pCacheWidth;
    pCacheHeight = pCacheHeight;
    pNumActiveCols = pNumActiveCols;
    pActiveCols = pActiveCols;

	/* This routine should never be called on platforms that have
	 * no long erode ops, so return ia_noImpl.
	 */
    __error__ = ia_noImpl;
    NORMAL_RETURN;
}


/* This is a public procedure */
/* PUBLIC */
/************************************************************************
 *  i_genErodeAsm():  generate assembly code to do erode defined by strel.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastErode as
 *			blockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns ia_noImpl.
 *
 * This code is used in place of C code that creates processor-specific
 * erosion routines.  This code is just a stub.  If no erosion routine is
 * made, erosion is done using rasta_rop's.
 ************************************************************************/
Int32 CDECL 
i_genErodeAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    ErodeProc  *generated_code,
    Int32      *code_length)
{
	/* keep the compiler happy by referencing all the unused parms. */
    PROLOGUE;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    data = data;
    blockWidth = blockWidth;
    *code_length = 0;
    *generated_code = NULL;	/* we don't know how to generate code for
				   this processor, so punt */
    __error__ = ia_noImpl;
    NORMAL_RETURN;
}

/************************************************************************
 *  i_genLongErodeAsm():  generate assembly code to do erode defined by strel.
 *			This assembly code makes use of long memory accesses
 *			and operations (long >= 64 bits)
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastErode as
 *			blockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *			If the buffer was passed in, we assume it has at
 *			least the number of bytes that i_getLongErodeAsmSize
 *			would calculate.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_SETooBig:		Structure element was too big to
 *				generate accelerated code OR was composed
 *				entirely of DONT_CARE terms (which is silly).
 *	ia_nomem:		Not enough memory to allocate the code buffer.
 *	ia_bufferExceeded:	The routine tried to generate too much
 *				assembly code.  This is a bug in the routine.
 *				If it happens, please let us know.
 ************************************************************************/
Int32  CDECL 
i_genLongErodeAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    LongErodeProc *generated_code,
    Int32      *code_length)
{
	/* keep the compiler happy by referencing all the unused parms. */
    PROLOGUE;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    data = data;
    blockWidth = blockWidth;
    *code_length = 0;
    *generated_code = NULL;	/* we don't know how to generate code for
				   this processor, so punt */
    __error__ = ia_noImpl;
    NORMAL_RETURN;
}



/*

$Log: migenero.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.8  1998/01/13 23:52:20  ddavies
migenero.c -> /product/ipcore/ipshared/src/RCS/migenero.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.7  1996/12/19 22:37:22  ddavies
migenero.c -> /product/ipcore/ipshared/src/RCS/migenero.c,v

Have dummy implementations of i_getErodeAsmSize and i_getLongErodeAsmSize
return ia_noImpl instead of ia_successful so upper levels of code know
not to allocate a buffer of size 0.  Purify refuses to free a buffer of
size 0, then complains that it's been leaked.

Revision 1.6  1996/12/18 23:03:11  ddavies
migenero.c -> /product/ipcore/ipshared/src/RCS/migenero.c,v

Add dummy implementations for ip_longOpIPLibLinked (returns cFalse since
code capable of doing long (>= 64 bit) operations is not present),
i_getLongErodeAsmSize, ip_getLongErodeSEParms and i_genLongErodeAsm.
These routines are not really present in the machine independent
implementations, but the linker wants to see them.

Revision 1.5  1995/12/13 02:49:25  ejaquith
migenero.c -> /product/ipcore/ipshared/src/RCS/migenero.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1995/10/03  21:14:34  ddavies
 * migenero.c -> /product/ipcore/ipshared/src/RCS/migenero.c,v
 *
 * Include shrmorph.pub/prv instead of shrip.pub/prv.  Change arg of
 * i_genErodeAsm from "void **generated_code" to "ErodeProc *generated_code".
 *
 * Revision 1.3  1994/09/20  02:48:32  ddavies
 * migenero.c -> /product/ipcore/ipshared/src/RCS/migenero.c,v
 *
 * Add dummy implementation of i_getErodeAsmSize.  Update comments and args
 * for dummy implementation of i_genErodeAsm to delete allocator and alloc_shrinker.
 *
 * Revision 1.2  1994/01/13  17:54:21  dferrell
 * dferrell on Thu Jan 13 09:52:46 PST 1994
 *
 * migenero.c -> /product/ipcore/ipshared/src/RCS/migenero.c,v
 *
 * Fixed so that "sparc.prv" is included only when SPARC_ENV is defined.
 *
 * Revision 1.1  1994/01/05  03:57:12  ddavies
 * Initial revision
 *
 * Revision 1.14  1993/12/02  19:26:56  ddavies
 * sgenero.c -> /product/ipcore/ipshared/src/RCS/sgenero.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.13  1993/11/05  02:22:11  ddavies
 * sgenero.c -> /product/ipcore/ipshared/src/RCS/sgenero.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


