/*************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  miglyph1.c:  contains C implementations of convolution routines.
 *
 *          Contains subroutines
 *
 *                Int32         ip_convolve1x1()
 *                Int32         ip_convolve3x3()
 *
 *	    Stub implementation of
 *		  UInt32	*ip_genConvAsm()
 *
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "imageref.pub"
#include "iaerror.pub"

    /*  prototypes */
#include "shrglyph.prv"		/* If we were using any of the public
				 * pointers to private structs, this would
				 * have to go first since it's where the
				 * private structures are defined.  Today,
				 * we're not using any, but you never know. */
#include "shrglyph.pub"
#include "kernel.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miglyph1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");
/* Most of the comments for the glyph functions are in cglyph.c.  This file
 * contains only C implementations of the convolution routines and a dummy
 * implementation of the routine that generates machine code for convolution.
 *
 * The stub is easy, it just sets the pointer to the C implementation of the
 * convolution routine needed and returns.  It's used when we don't know
 * how to generate machine code for the current platform.
 *
 * ip_convolve1x1 does a simple convolution.  It takes as arguments a pointer
 * to the word containing the upper left corner of the glyphs and a
 * count saying where the upper left corner is within that word.  It also
 * takes the number of 32-bit words/line and a pointer to the kernel
 * data structure.
 *
 * The first thing it does is advance the image pointer down to account
 * for the distance to the kernel's firstActiveRow.  It then moves the
 * x position within the word, and the pointer if necessary, to the right
 * to account for firstActiveColBit.
 *
 * We now loop through the kernel rows.
 *
 * For each kernel row, we process pKernel->numBitsPerRow from right to
 * left.  We don't know the depth of the image, and don't care.  The 
 * table entries are set up to account for the image depth.
 *
 * The bits in each row are processed in multiple groups.  We access the
 * groups starting from the right end of of the bit vector and keep
 * going until we've run out of bits.
 *
 * To process a group of image bits, we need to put the group's bits
 * in the proper position, then use them to access the proper table.
 * This means some shifting and masking.  Any of the groups could cross a
 * word boundary.  Aligning a split group requires a left shift, a
 * right shift and an OR to merge the results.  Bummer.  We could 
 * decide whether each group individually crossed a boundary, but that 
 * would be slow.  Instead, we generate an input word containing all 
 * the bits we'll need in a row.  The first table in the comments of 
 * cglyph.c shows that we'll never need more than 28 bits for a row.  
 * Therefore, we can fit all the image bits required to process one 
 * kernel row into a single register.  If all these bits happen to be 
 * in a single image word, we only need a shift.  If they are in two 
 * image words, we do the two shifts.  All this happens once per row.
 *
 * Furthermore, since we don't want to decide this once per row, we decide
 * it once and have three versions of row loop.  One reads input words and
 * shifts them right to align them (LS Bit of input vector is in bit 2
 * of the register so we can access 32-bit words); one reads input words
 * and shifts them left by 1 or 2 and one reads pairs of input words and
 * does double left shifts.
 *
 * Within each row, we consume the block of input bits from right to
 * left in units of 8 bits.  Each block is used to look up a table entry
 * which is added to the running sum.  The entire group of input bits in
 * the row is masked with pKernel->rowBitsMask before we begin processing.
 * After each step, we see if there are any bits remaining.  If not,
 * we quit.  A nice feature of this scheme is that we don't have to keep
 * an extra loop variable to see if we're done with the row.  Another nice
 * feature is that the last group of bits (probably less than 8 bits) is
 * processed naturally without special conditions or masks.
 *
 * Having added up all the entries from all the rows, we return the sum.
 *
 * The 3x3 convolution is a bit harder.  We want to produce three rows of
 * convolved sums.  These three sums should be centered around the estimated
 * position given to us, so we move up by 1 row and to the left by
 * pKernel->depth bits..  We then go right by firstActiveColBit and down by
 * firstActiveRow.  Since centering kernels often have one inactive row and
 * column around the edges, we often end up the the same place.
 * We then process (pKernel->numActiveRows)+2 rows.  In the first row,
 * the sums affect only the top row of convolved outputs.  In the second
 * row, the convolved sum affect the top two rows of outputs.  The
 * next pKernel->numActiveRows-2 rows affect all three rows of
 * convolved outputs. The second to last row affects on the last two 
 * rows of outputs.  The last row affects only the last row of outputs.
 * Similar things happen as we move across the columns, but the tables 
 * were built with that in mind.
 *
 * The code convolves one row affecting one output row, then one input row
 * affecting two output rows, then does a loop over the middle
 * numActiveRows-2 input rows affecting all three output rows, then
 * convolves an input row affecting the lower two output rows and finally
 * convolves the last input row, putting the result into the last output row.
 *
 * Within each row, we operate pretty much like we did in the 1x1 case.
 * The exception is that we maintain up to three table pointers and use
 * each block of input bits to index 1, 2 or 3 tables - putting the
 * results into 1, 2 or 3 result words.
 *
 * There's another funny thing about the 3x3 case.  When we're done, we don't
 * return all 9 sums.  We must find the biggest one that's less than the
 * upperThreshold.  We return that sum AND modify the image address (we
 * received a pointer to it instead of getting the real thing) AND modify
 * the xoffset within the word to point to the location with the highest
 * sum.  That is, if the pixel in the upper left corner had the highest
 * legal sum, we should move the word pointer up a line and move the x
 * offset to the left by 1 bit.  If that crossed a word boundary, we should
 * adjust the word pointer and xoffset again.
 *
 * The process of searching through the 9 sums should be biased towards the
 * middle sum.  That is, we move the center point only if the sum for
 * another pixel is actually greater than that for the center.
 *
 * IMPROVEMENT: comparing each of the 9 sums against pKernel->upperThreshold
 * is a pain.  There's a faster way.  There are 11 bits allocated for the
 * rightmost two sums in the word and 10 bits for the leftmost sum.  We
 * assume that convolution sums are in the range [511, -512].  Before we
 * begin the convolution, we bias them with +512, so the results lie
 * in [1023, 0].  This only uses 10 bits.
 * After all summing has been done, we add (512 - pKernel->upperThreshold)
 * to each sum.  The result will fit into 11 bits, but we only look at 
 * the LS 10 bits.  If the original sum was greater than or equal to 
 * the upperThreshold, it becomes a small number (MOD 1024).  When we 
 * compare it with the current max, we will see it as a small number, so 
 * will ignore it.  When we're done comparing, we subtract (512 (original bias)
 * + 512 (this last bias) - pKernel->upperThreshold (ditto)) from the 
 * result to un-bias it.
 *
 * SMALL IMPROVEMENT: This has to do with searching through the 3x3 array
 * of output sums for the best one.  Say there are often ties for the
 * best sum.  Say we always search them in the same order, like left-to-right,
 * top-to-bottom.  Furthermore, say we only choose a new sum when it's bigger
 * than the current choice.  In this case, we will tend to drive towards the
 * upper left since that's where we'll see the best sum first.
 *
 * We attempt to fix this problem by alternating search directions.  The
 * last argument to ip_convolve3x3 specifies search order.  If the LS bit of
 * this arg is 0, we search the 3x3 array of output sums from
 * left-to-right, top-to-bottom ((-1,-1), (0, -1), ... , (0, 1), (1, 1)).
 * if the LS bit of the last arg is 1, we search from right-to-left,
 * bottom-to-top ((1, 1), (0, 1), ... , (0, -1), (-1, -1)).  The caller
 * puts something that flips for every glyph in a row, like the number of
 * the glyph, in this parameter.
 */

/* This is a private routine */
/* PRIVATE */
/*
 *  ip_genConvAsm():
 *        This routine generates machine code to process a kernel of a
 *	  particular shape.  The code is specialized for each platform.
 *	  If we don't know how to generate code for this platform, return
 *	  a pointer to the general C implementation.
 *
 *	pKernel:	Pointer to IAKernel data structure containing
 *			information about the kernel's shape.
 *	allocator:	Ptr to function that is used to allocate memory
 *			for the code.  Function's single argument is
 *			the number of BYTES requested.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of malloc
 *			provided by WATCOM on the PC!!
 *	freer:		Ptr to function used to free memory in case of
 *			error
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of free
 *			provided by WATCOM on the PC!!
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		couldn't allocate internal buffers
 *					
 */
Int32 CDECL 
ip_genConvAsm (
    IPState    *pIPState,
    IAKernel   *pKernel,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	),
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *))
{

	/* Keep the compiler happy */
    PROLOGUE;
    allocator = allocator;
    freer = freer;

	/* Since this code doesn't know anything about generating machine
	 * code, always return a pointer to the C implementation.  Do
	 * we want the 1x1 convolution routine or the 3x3 routine?
	 */
    if (pKernel->is3x3)
	pKernel->conv3Code = ip_convolve3x3;
    else
	pKernel->conv1Code = ip_convolve1x1;

    NORMAL_RETURN;
}


/* This is a private routine */
/* PRIVATE */
/*
 *  ip_convolve1x1():
 *        This routine does one convolution operation producing one sum.
 *
 *	pImage:		Pointer to word containing upper left corner of
 *			image block to be convolved.
 *	centerXBit:	holds position of leftmost bit
 *			within word pointed to by pImage
 *	wpl:		Number of words/line in image.
 *	pKernel:	Pointer to IAKernel data structure used to
 *			define the convolution kernel.
 *
 *   Returns:
 *	convResult:	The convolution sum.
 *					
 */
Int32 CDECL 
ip_convolve1x1 (
    UInt32     *pImage,
    Int32	centerXBit,
    UInt32	wpl,
    IAKernel   *pKernel)
{
    UInt32	inputWord;		/* the raw image bits, usually shifted
					 */
    UInt8	*pCompTableRow;		/* a local byte pointer used
					 * both so we won't keep referencing
					 * the pointer in the kernel structure
					 * and so the compiler won't go
					 * around shifting indices.  This
					 * version points to the beginning
					 * of the section of the table
					 * to be used for the current row. */
    UInt8	*pCompTableBlock;	/* a version of pCompTableRow that
					 * tracks across the columns. */
    Int32	shiftCount;		/* amount by which raw image words
					 * should be shifted to be properly
					 * aligned in inputWord */
    Int32	convSum;		/* the result */


	/* add in offset due to vacant kernel columns. */
    centerXBit += pKernel->firstActiveColBit;

	/* if centerX went over 31, increment pImage.  Do this now
	 * because the machine is likely to still have centerX in a
	 * register */
    if (centerXBit > 31)
    {
	pImage++;
	centerXBit &= 0x1F;
    }

	/* add in offset due to vacant kernel rows.  We use the loop
	 * instead of a multiply because the firstActiveRow is most likely
	 * to be 0, then maybe 1 or 2.  Note the SIDE-EFFECT this has
	 * of setting convSum to 0, which is what we wanted anyway. */
    for (convSum = pKernel->firstActiveRow; convSum > 0; convSum--)
        pImage += wpl;

	/* We want to use the input bits as table indices directly
	 * without the having the compiler shift them into place as it
	 * would if we were naturally using a pointer to words.  That
	 * means the LS two bits of inputWord should be blank -
	 * or the LS bit of the string of useful input should reside in
	 * bit 2 of inputWord.  This means there are three possible
	 * actions we could take to align the input bits in the image
	 * properly.  If the LS input bit is to the left of bit 1, we
	 * shift the entire string right to align it.  We know that it
	 * all fits into the current word since centerXBit (the right end)
	 * is by definition in the current word.
	 * If the LS input bit is in bits 0 or 1 of the image word,
	 * we'll need to shift the result left to align it.  We know
	 * the result will fit because the string of input bits is at
	 * most 28 bits long (see first table in cglyph.c).
	 * If the LS bit is below bit 0 (in the next word), we'll need
	 * to do a left shift of the portion of the input in this word,
	 * then a right shift of the portion in the next word, then
	 * OR the results.
	 * We'll clearly need to do one of these actions for every kernel
	 * row.  However, we would like to avoid deciding which one to
	 * do on every row since the decision will be the same on every
	 * row (all images are padded to 32-bit boundaries so the word
	 * boundaries in each row are vertically aligned).  We decide 
	 * once here, and have three versions of row
	 * loop.  One shifts every input word right, the second shifts
	 * every input word left and the last shifts every input word
	 * and the word next to it left.
	 * Initially, we set shiftCount to the distance from the LS bit in
	 * the field to the LS bit in the word (the right shift distance)
	 * and branch on its value.  Each case will turn this
	 * value into a real shift count as appropriate.
	 */
    shiftCount = 32 - (centerXBit + pKernel->numBitsPerRow);

    if (shiftCount > 1)
    {
	    /* in this case, we'll be doing right shifts by shiftCount-2 */
	shiftCount = shiftCount-2;	/* this is the right shift dist */

	    /* loop through the rows of the kernel */
	for (pCompTableRow = (UInt8 *) pKernel->compWeights;
	     (UInt32) pCompTableRow < pKernel->finalTablePtrVal;
	     pImage += wpl, pCompTableRow += pKernel->bytesPerTableRow)
	{
	    pCompTableBlock = pCompTableRow;
	    inputWord = (*pImage >> shiftCount) & pKernel->rowBitsMask;

		/* Here's the loop that will be executed to retire the
		 * bits in the input block.  We just keep going until
		 * there aren't any more bits.  The giant, long, ugly
		 * thinger that finds the table entry to add to convSum
		 * says "get the relevant bits in inputWord, add them
		 * to the pointer to the table as bytes modifying a byte
		 * address, treat the resulting address
		 * as a word address and retrieve the word at that
		 * location."
		 */
	    for (;
		 inputWord != 0;
		 inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	    {
		convSum += ULONG_ACCESS(&pCompTableBlock[inputWord & 0x3FE]);
		pCompTableBlock += 256*sizeof(Int32);
	    }

	}  /* bottom of loop through rows when right shifts are needed */
    }	/* end of right shift case */

    else
    if (shiftCount >= 0)	/* left shift by 1 or 2 needed */

    {
	    /* in this case, we'll be doing left shifts by 2-shiftCount */
	shiftCount = 2-shiftCount;	/* this is the left shift dist */

	    /* loop through the rows of the kernel */
	for (pCompTableRow = (UInt8 *) pKernel->compWeights;
	     (UInt32) pCompTableRow < pKernel->finalTablePtrVal;
	     pImage += wpl, pCompTableRow += pKernel->bytesPerTableRow)
	{
	    pCompTableBlock = pCompTableRow;
	    inputWord = (*pImage << shiftCount) & pKernel->rowBitsMask;

		/* Here's the loop that will be executed to retire the
		 * bits in the input block.  We just keep going until
		 * there aren't any more bits. */
	    for (;
		 inputWord != 0;
		 inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	    {
		convSum += ULONG_ACCESS(&pCompTableBlock[inputWord & 0x3FE]);
		pCompTableBlock += 256*sizeof(Int32);
	    }

	}  /* bottom of loop through rows when right shifts are needed */
    }	/* end of left single shift case */

    else

    {
	    /* in this case, we'll be doing a double left shift.  We
	     * shift the current word left by 2-shiftCount and shift
	     * the next word to the right right by 32-(2-shiftCount),
	     * then combine them.
	     */
	shiftCount = 2-shiftCount;	/* this is the left shift dist */

	    /* loop through the rows of the kernel */
	for (pCompTableRow = (UInt8 *) pKernel->compWeights;
	     (UInt32) pCompTableRow < pKernel->finalTablePtrVal;
	     pImage += wpl, pCompTableRow += pKernel->bytesPerTableRow)
	{
	    pCompTableBlock = pCompTableRow;
	    inputWord = *pImage << shiftCount;
	    inputWord = (inputWord | (pImage[1] >> (32-shiftCount))) &
			  pKernel->rowBitsMask;

		/* Here's the loop that will be executed to retire the
		 * bits in the input block.  We just keep going until
		 * there aren't any more bits.
		 */
	    for (;
		 inputWord != 0;
		 inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	    {
		convSum += ULONG_ACCESS(&pCompTableBlock[inputWord & 0x3FE]);
		pCompTableBlock += 256*sizeof(Int32);
	    }

	}  /* bottom of loop through rows when right shifts are needed */
    }	/* end of left double shift case */

    return convSum;

}


/* This is a private routine */
/* PRIVATE */
/*
 *  ip_convolve3x3():
 *        This routine produces 9 convolution sums in a 3x3 array.
 * 	  It returns the largest of the sums that's less than the upper
 *	  threshold.  It also adjusts the input address pointer and
 *	  input value of centerXBit to point to the position of the largest
 *	  sum relative to the original estimated position.  The original
 *	  position is in the center of the 3x3 array.  If the center's
 *	  sum is as large as the largest sum, it is chosen.
 *
 *	ppImage:	Pointer to pointer to word containing upper
 *			left corner of estimated position of the glyph
 *			to be found.  The image pointer will be adjusted
 *			to point to the glyph center actually found.
 *      pCenterXBit:    Pointer to Int32.  Bottom 16 bits holds estimated
 *                      position of leftmost bit of the glyph
 *                      within word pointed to by *ppImage.
 *                      Next 8 bits (to the left) holds the amount that
 *                      the center moved to the left or right.  Should
 *                      be -1, 0 or 1.  The most significant 8 bits
 *                      holds the amount that the center moved up or
 *                      down.  Should be -1, 0, or 1.
 *	wpl:		Number of words/line in image.
 *	pKernel:	Pointer to IAKernel data structure used to
 *			define the convolution kernel.
 *	numOutBit:	The LS bit of this number is used to determine
 *			the order in which we search the 3x3 output grid
 *			for the biggest score.  If 0, we search left
 *			to right, top to bottom.  If 1, we search right
 *			to left, bottom to top.  This tends to keep
 *			us from wandering off the line of glyphs.
 *
 *   Returns:
 *	convResult:	The largest convolution sum less than
 *			pKernel->upperThreshold.
 *	in *ppImage, *pCenter:	Location of the image word containing
 *			the upper left corner of the image
 *			block corresponding to the best convolution sum.
 *			For example, if the sum in the upper left corner
 *			of the 3x3 array was the biggest legal one,
 *			*ppImage and *pCenter would be adjusted to the
 *			left by 1 and up by 1 from their original values.
 *					
 */
Int32 CDECL 
ip_convolve3x3 (
    UInt32    **ppImage,
    Int32      *pCenterXBit,
    UInt32	wpl,
    IAKernel   *pKernel,
    UInt32	numOutBit)
{
    UInt32	inputWord;		/* the raw image bits */
    UInt8      *pCompTableRow0;		/* local byte pointer used
					 * both so we won't keep referencing
					 * the kernel structure and so the
					 * compiler won't keep shifting 
					 * indices.  This pointer
					 * points to the beginning
					 * of the section of the table
					 * to be used to compute the upper
					 * output row for the current input
					 * row. */
    UInt8      *pCompTableBlock0, *pCompTableBlock1, *pCompTableBlock2;
					/* Versions of pCompTableRow0 that
					 * track across the columns.  We
					 * need three pointers into the
					 * table since we're computing 
					 * sums on three different rows */
    Int32	shiftCount;		/* amount by which raw image words
					 * should be shifted to be properly
					 * aligned in inputWord */
    UInt32	convSum0, convSum1, convSum2;
					/* Each of these holds 3 convolution
					 * sums, for a total of 9 sums */
    Int32	centerXBit;		/* local version of the position
					 * within the pImage word */
    UInt32     *pImage;			/* local version of the image pointer */
    Int32	convResult;		/* the biggest legal sum of the 3x3
					 * sums */
    Int32	depth;			/* we'll use this a lot at the
					 * bottom of the routine */

	/* load the local versions of the image pointer and centerXBit parms */
    pImage = *ppImage;
    centerXBit = *pCenterXBit;	/* the original is always positive */

	/* Move up and to the left to get the proper starting position
	 * for the convolution rectangle */
    pImage -= wpl;
    centerXBit -= 1 * pKernel->depth;

	/* add in offset due to vacant kernel columns */
    centerXBit += pKernel->firstActiveColBit;

	/* if centerXBit went over 31 or below 0, fix pImage.  Do this now
	 * because the machine is likely to still have centerXBit in a
	 * register */
    if (centerXBit & 0x20)		/* catches both "< 0" and "> 31" */
    {
	if (centerXBit > 0)
	    pImage++;
	else
	    pImage--;
	centerXBit &= 0x1F;		/* truncate it back to [0..31] */
    }

	/* add in offset due to vacant kernel rows.  We use the loop
	 * instead of a multiply because the firstActiveRow is most likely
	 * to be 0, then maybe 1 or 2. */
    for (convSum0 = pKernel->firstActiveRow; convSum0 > 0; convSum0--)
        pImage += wpl;

	/* Bias all the original sums so all values will be positive */
    convSum0 = convSum1 = convSum2 = cTripleSumBiasVal;

	/* We want to use the input bits as table indices directly
	 * without the having the compiler shift them into place as it
	 * would if we were naturally using a pointer to words.  That
	 * means the LS two bits of inputWord should be blank -
	 * or the LS bit of the string of useful input should reside in
	 * bit 2 of inputWord.  This means there are three possible
	 * actions we could take to align the input bits in the image
	 * properly.  If the LS input bit is to the left of bit 1, we
	 * shift the entire string right to align it.  We know that it
	 * all fits into the current word since centerXBit (the right end)
	 * is by definition in the current word.
	 * If the LS input bit is in bits 0 or 1 of the image word,
	 * we'll need to shift the result left to align it.  We know
	 * the result will fit because the string of input bits is at
	 * most 28 bits long (see first table in cglyph.c).
	 * If the LS bit is below bit 0 (in the next word), we'll need
	 * to do a left shift of the portion of the input in this word,
	 * then a right shift of the portion in the next word, then
	 * OR the results.
	 * We'll clearly need to do one of these actions for every kernel
	 * row.  However, we would like to avoid deciding which one to
	 * do on every row since the decision will be the same on every
	 * row (all images are padded to 32-bit boundaries so the word
	 * boundaries in each row are vertically aligned).  We decide 
	 * once here, and have three versions of row
	 * loop.  One does right shifts, one does single left shifts
	 * and the last one does double left shifts.
	 * Initially, we set shiftCount to the distance from the LS bit in
	 * the field to the LS bit in the word (the right shift distance)
	 * and branch on its value.  Each case will turn this
	 * value into a real shift count as appropriate.
	 */
    shiftCount = 32 - (centerXBit + pKernel->numBitsPerRow);

	/* ************ 3x3 Convolution, shiftCount > 1 ************ */

    if (shiftCount > 1)
    {
	    /* in this case, we'll be doing right shifts by shiftCount-2 */
	shiftCount = shiftCount-2;	/* this is the right shift dist */

	    /* pt to beginning of first row */
	pCompTableBlock0 = pCompTableRow0 = (UInt8 *) (pKernel->compWeights);

	    /* Process the FIRST kernel row, it only affects convSum0 */
	inputWord = (*pImage >> shiftCount) & pKernel->rowBitsMask;

	    /* Here's the loop that will be executed to retire the
	     * bits in the input block.  We just keep going until
	     * there aren't any more bits.  The giant, long, ugly
	     * thinger that finds the table entry to add to convSum
	     * says "get the relevant bits in inputWord, add them
	     * to the pointer to the table as bytes modifying a byte
	     * address, treat the resulting address
	     * as a word address and retrieve the word at that
	     * location."
	     */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
	    pCompTableBlock0 += 256*sizeof(Int32);
	}

	    /* Process the SECOND kernel row, it only affects convSum0
	     * and convSum1 */
	pImage += wpl;
	pCompTableBlock1 = pCompTableRow0;
	pCompTableRow0 += pKernel->bytesPerTableRow;
	pCompTableBlock0 = pCompTableRow0;

	inputWord = (*pImage >> shiftCount) & pKernel->rowBitsMask;

	    /* Process all the bits in the second row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
	    convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
	    pCompTableBlock0 += 256*sizeof(Int32);
	    pCompTableBlock1 += 256*sizeof(Int32);
	}

	    /* loop through the numActiveRows-2 rows of the kernel in
	     * the MIDDLE */
	pImage += wpl;
	pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;
	pCompTableBlock1 = pCompTableRow0;
	pCompTableRow0 += pKernel->bytesPerTableRow;
	pCompTableBlock0 = pCompTableRow0;

	while ((UInt32) pCompTableRow0 < pKernel->finalTablePtrVal)
	{
	    inputWord = (*pImage >> shiftCount) & pKernel->rowBitsMask;

	    /* Process all the bits in one of the middle row */
	    for (;
		 inputWord != 0;
		 inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	    {
		convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
		convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
		convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
		pCompTableBlock0 += 256*sizeof(Int32);
		pCompTableBlock1 += 256*sizeof(Int32);
		pCompTableBlock2 += 256*sizeof(Int32);
	    }
		/* move all table pointers down one row. */
	    pImage += wpl;
	    pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;
	    pCompTableBlock1 = pCompTableRow0;
	    pCompTableRow0 += pKernel->bytesPerTableRow;
	    pCompTableBlock0 = pCompTableRow0;
	}

	    /* Process the SECOND-TO-LAST kernel row, it only affects convSum1
	     * and convSum2.  At this point, pCompTableRow0 points
	     * one row BELOW the last row of table stuff.  pCompTableBlock1
	     * points to the last table row and pCompTableBlock2 points to
	     * the second-to-last table row. */

	inputWord = (*pImage >> shiftCount) & pKernel->rowBitsMask;

	    /* Process all the bits in the second-to-last row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
	    convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
	    pCompTableBlock1 += 256*sizeof(Int32);
	    pCompTableBlock2 += 256*sizeof(Int32);
	}
		/* move last pointer down to last table row. */
	pImage += wpl;
	pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;

	    /* Process the LAST kernel row, it only affects convSum2 */

	inputWord = (*pImage >> shiftCount) & pKernel->rowBitsMask;

	    /* Process all the bits in the last row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
	    pCompTableBlock2 += 256*sizeof(Int32);
	}

    }	/* end of right shift case */

    else
    if (shiftCount >= 0)	/* left shift by 1 or 2 needed */


	/* ************ 3x3 Convolution, (2 > shiftCount >= 0) ************ */

    {
	    /* in this case, we'll be doing left shifts by 2-shiftCount */
	shiftCount = 2-shiftCount;	/* this is the single left shift dist */

	    /* pt to beginning of first row */
	pCompTableBlock0 = pCompTableRow0 = (UInt8 *) (pKernel->compWeights);

	    /* Process the FIRST kernel row, it only affects convSum0 */
	inputWord = (*pImage << shiftCount) & pKernel->rowBitsMask;

	    /* Here's the loop that will be executed to retire the
	     * bits in the input block.  We just keep going until
	     * there aren't any more bits. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
	    pCompTableBlock0 += 256*sizeof(Int32);
	}

	    /* Process the SECOND kernel row, it only affects convSum0
	     * and convSum1 */
	pImage += wpl;
	pCompTableBlock1 = pCompTableRow0;
	pCompTableRow0 += pKernel->bytesPerTableRow;
	pCompTableBlock0 = pCompTableRow0;

	inputWord = (*pImage << shiftCount) & pKernel->rowBitsMask;

	    /* Process all the bits in the second row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
	    convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
	    pCompTableBlock0 += 256*sizeof(Int32);
	    pCompTableBlock1 += 256*sizeof(Int32);
	}

	    /* loop through the numActiveRows-2 rows of the kernel in
	     * the MIDDLE */
	pImage += wpl;
	pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;
	pCompTableBlock1 = pCompTableRow0;
	pCompTableRow0 += pKernel->bytesPerTableRow;
	pCompTableBlock0 = pCompTableRow0;

	while ((UInt32) pCompTableRow0 < pKernel->finalTablePtrVal)
	{
	    inputWord = (*pImage << shiftCount) & pKernel->rowBitsMask;

	    /* Process all the bits in one of the middle rows */
	    for (;
		 inputWord != 0;
		 inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	    {
		convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
		convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
		convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
		pCompTableBlock0 += 256*sizeof(Int32);
		pCompTableBlock1 += 256*sizeof(Int32);
		pCompTableBlock2 += 256*sizeof(Int32);
	    }
		/* move two extra pointer down one table row. */
	    pImage += wpl;
	    pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;
	    pCompTableBlock1 = pCompTableRow0;
	    pCompTableRow0 += pKernel->bytesPerTableRow;
	    pCompTableBlock0 = pCompTableRow0;
	}

	    /* Process the SECOND-TO-LAST kernel row, it only affects convSum1
	     * and convSum2.  At this point, pCompTableRow0 points
	     * one row BELOW the last row of table stuff.  pCompTableBlock1
	     * points to the last table row and pCompTableBlock2 points to
	     * the second-to-last table row. */

	inputWord = (*pImage << shiftCount) & pKernel->rowBitsMask;

	    /* Process all the bits in the second-to-last row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
	    convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
	    pCompTableBlock1 += 256*sizeof(Int32);
	    pCompTableBlock2 += 256*sizeof(Int32);
	}
		/* move last extra pointer down to last table row. */
	pImage += wpl;
	pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;

	    /* Process the LAST kernel row, it only affects convSum2 */

	inputWord = (*pImage << shiftCount) & pKernel->rowBitsMask;

	    /* Process all the bits in the last row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
	    pCompTableBlock2 += 256*sizeof(Int32);
	}

    }	/* end of single left shift case */


    else

	/* ************ 3x3 Convolution, (shiftCount < 0) ************ */

    {
	    /* in this case, we'll be doing double left shifts by
	     * 2-shiftCount.  A double left shift is implemented as a
	     * leftShift of *pImage and a right shift of pImage[1]
	     * and an OR to combine the results. */
	shiftCount = 2-shiftCount;	/* this is the single left shift dist */

	    /* pt to beginning of first row */
	pCompTableBlock0 = pCompTableRow0 = (UInt8 *) (pKernel->compWeights);

	    /* Process the FIRST kernel row, it only affects convSum0 */
	inputWord = *pImage << shiftCount;
	inputWord = (inputWord | (pImage[1] >> (32-shiftCount))) &
		      pKernel->rowBitsMask;

	    /* Here's the loop that will be executed to retire the
	     * bits in the input block.  We just keep going until
	     * there aren't any more bits. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
	    pCompTableBlock0 += 256*sizeof(Int32);
	}

	    /* Process the SECOND kernel row, it only affects convSum0
	     * and convSum1 */
	pImage += wpl;
	pCompTableBlock1 = pCompTableRow0;
	pCompTableRow0 += pKernel->bytesPerTableRow;
	pCompTableBlock0 = pCompTableRow0;

	inputWord = *pImage << shiftCount;
	inputWord = (inputWord | (pImage[1] >> (32-shiftCount))) &
		      pKernel->rowBitsMask;

	    /* Process all the bits in the second row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
	    convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
	    pCompTableBlock0 += 256*sizeof(Int32);
	    pCompTableBlock1 += 256*sizeof(Int32);
	}

	    /* loop through the numActiveRows-2 rows of the kernel in
	     * the MIDDLE */
	pImage += wpl;
	pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;
	pCompTableBlock1 = pCompTableRow0;
	pCompTableRow0 += pKernel->bytesPerTableRow;
	pCompTableBlock0 = pCompTableRow0;

	while ((UInt32) pCompTableRow0 < pKernel->finalTablePtrVal)
	{
	    inputWord = *pImage << shiftCount;
	    inputWord = (inputWord | (pImage[1] >> (32-shiftCount))) &
			  pKernel->rowBitsMask;

	    /* Process all the bits in one of the middle rows */
	    for (;
		 inputWord != 0;
		 inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	    {
		convSum0 += ULONG_ACCESS(&pCompTableBlock0[inputWord & 0x3FE]);
		convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
		convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
		pCompTableBlock0 += 256*sizeof(Int32);
		pCompTableBlock1 += 256*sizeof(Int32);
		pCompTableBlock2 += 256*sizeof(Int32);
	    }
		/* move all table pointers down one row. */
	    pImage += wpl;
	    pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;
	    pCompTableBlock1 = pCompTableRow0;
	    pCompTableRow0 += pKernel->bytesPerTableRow;
	    pCompTableBlock0 = pCompTableRow0;
	}

	    /* Process the SECOND-TO-LAST kernel row, it only affects convSum1
	     * and convSum2.  At this point, pCompTableRow0 points
	     * one row BELOW the last row of table stuff.  pCompTableBlock1
	     * points to the last table row and pCompTableBlock2 points to
	     * the second-to-last table row. */

	inputWord = *pImage << shiftCount;
	inputWord = (inputWord | (pImage[1] >> (32-shiftCount))) &
		      pKernel->rowBitsMask;

	    /* Process all the bits in the second-to-last row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum1 += ULONG_ACCESS(&pCompTableBlock1[inputWord & 0x3FE]);
	    convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
	    pCompTableBlock1 += 256*sizeof(Int32);
	    pCompTableBlock2 += 256*sizeof(Int32);
	}
		/* move last table pointer down to last table row. */
	pImage += wpl;
	pCompTableBlock2 = pCompTableRow0 - pKernel->bytesPerTableRow;

	    /* Process the LAST kernel row, it only affects convSum2 */

	inputWord = *pImage << shiftCount;
	inputWord = (inputWord | (pImage[1] >> (32-shiftCount))) &
		      pKernel->rowBitsMask;

	    /* Process all the bits in the last row. */
	for (;
	     inputWord != 0;
	     inputWord = (inputWord >> 8) & pKernel->rowBitsMask)
	{
	    convSum2 += ULONG_ACCESS(&pCompTableBlock2[inputWord & 0x3FE]);
	    pCompTableBlock2 += 256*sizeof(Int32);
	}

    }	/* end of double left shift case */


	/* ************* Find the biggest sum ********************* */

	/* At this point, convSum0, convSum1 and convSum2 each have three
	 * sums in them.  Do the trick of adding 512-(pKernel->upperThreshold)
	 * to each sum.  When we look at the lower 10 bits of the result,
	 * the sums that were bigger than the upperThreshold will have
	 * wrapped around to small numbers (MOD 1024), so we'll ignore them.
	 */
    inputWord = cSumBias - (pKernel->upperThreshold + 1);	/* use as scratch reg */
    inputWord = (inputWord << (2*cSumSpacing)) |
		(inputWord << cSumSpacing) |
		 inputWord;
    convSum0 += inputWord;
    convSum1 += inputWord;
    convSum2 += inputWord;

    depth = (Int32) pKernel->depth;	/* we use this to move left or right
					 * by 1 pixel.  Moving right means
					 * moving by -depth, so we make
					 * depth an integer.
					 */

	/* Assume the center has the correct value and
	 * initialize the sum from that.
	 */
    convResult = ((convSum1 >> cSumSpacing) & cSumMask);

	/* To conserve local vars (keeps the compiler happy), we'll use
	 * centerXBit and shiftCount to indicate the x and y positions of the
	 * highest legal sum. */
    centerXBit = 0;

	/* Check the rest of the sums.  Note that convSum0 is for the
	 * row of output pixel towards the top of the page, hence has
	 * more negative row (y) numbers.
	 * If the LS bit of numOutBit is 0, we search from the top left
	 * towards the bottom right.  If numOutBit.0 equals 1, we search
	 * from the bottom right towards the top left.  Alternating search
	 * direction tends to eliminate tracking bias as we move across
	 * a row of glyphs.
	 */

   	/* we ultimately want information about how many lines that we
	 * we up or down and how many pixels that we went right or left,
	 * if any, in the top 16 bits of *pCenterXBit.  I'd put them in
	 * centerXBit now except that centerXBit may need to be modified
	 * before being written into *pCenterXBit (if it crosses a
	 * word boundary) so its easier to put that information somewhere
	 * else.  So, we put that information into shiftCount.
	 */
    shiftCount = cCenterAndCenter;
    if (!(numOutBit & 1))	/* search left-to-right, top-to-bottom */
    {
	inputWord = (convSum0 >> (2*cSumSpacing));
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = -depth;
	    shiftCount = cUpAndLeft;
	}

	inputWord = (convSum0 >> cSumSpacing) & cSumMask;
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = 0;
	    shiftCount = cUpAndCenter;
	}

	inputWord = convSum0 & cSumMask;
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = depth;
	    shiftCount = cUpAndRight;
	}

	inputWord = (convSum1 >> (2*cSumSpacing));
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = -depth;
	    shiftCount = cCenterAndLeft;
	}

	inputWord = convSum1 & cSumMask;
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = depth;
	    shiftCount = cCenterAndRight;
	}

	inputWord = (convSum2 >> (2*cSumSpacing));
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = -depth;
	    shiftCount = cDownAndLeft;
	}

	inputWord = (convSum2 >> cSumSpacing) & cSumMask;
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = 0;
	    shiftCount = cDownAndCenter;
	}

	inputWord = convSum2 & cSumMask;
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = depth;
	    shiftCount = cDownAndRight;
	}
    }

    else		/* search right-to-left, bottom-to-top */

    {
	inputWord = convSum2 & cSumMask;
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = depth;
	    shiftCount = cDownAndRight;
	}
	convSum2 >>= cSumSpacing;		/* look at sum (0, 1) */
    
	inputWord = convSum2 & cSumMask;
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = 0;
	    shiftCount = cDownAndCenter;
	}
	convSum2 >>= cSumSpacing;		/* look at sum (-1, 1) */

	if ((Int32)convSum2 > convResult)
	{
	    convResult = convSum2;
	    centerXBit = -depth;
	    shiftCount = cDownAndLeft;
	}

	inputWord = convSum1 & cSumMask;	/* look at sum (1, 0) */
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = depth;
	    shiftCount = cCenterAndRight;
	}
	convSum1 >>= (2*cSumSpacing);		/* look at sum (-1, 0) */

	if ((Int32)convSum1 > convResult)
	{
	    convResult = convSum1;
	    centerXBit = -depth;
	    shiftCount = cCenterAndLeft;
	}

	inputWord = convSum0 & cSumMask;	/* look at sum (1, -1) */
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = depth;
	    shiftCount = cUpAndRight;
	}
	convSum0 >>= cSumSpacing;		/* look at sum (0, -1) */

	inputWord = convSum0 & cSumMask;
	if ((Int32)inputWord > convResult)
	{
	    convResult = inputWord;
	    centerXBit = 0;
	    shiftCount = cUpAndCenter;
	}
	convSum0 >>= cSumSpacing;		/* look at sum (-1, -1) */
    
	if ((Int32)convSum0 > convResult)
	{
	    convResult = convSum0;
	    centerXBit = -depth;
	    shiftCount = cUpAndLeft;
	}
    }

	/* having decided where the convolution sum is, adjust the
	 * passed-in *ppImage and *pCenter parms if the estimated
	 * position wasn't the best. */
    if (shiftCount != 0)	/* we should move the estimated pt */
    {
	pImage = *ppImage;
	centerXBit += *pCenterXBit;
	if (centerXBit & 0x20)		/* crossed a word boundary */
	{
	    pImage += (shiftRightArithmetic(centerXBit, 5));
	    centerXBit &= 0x1F;
	}
	if ((shiftCount & cVerticalMovementMask) != 0)
	{
	    if (shiftCount < 0)
		pImage -= wpl;
	    else
		pImage += wpl;
	}

	*ppImage = pImage;
	*pCenterXBit = centerXBit | shiftCount;
    }
	/* remove the bias before returning the result */
    convResult = convResult - (Int32)(2*cSumBias) + pKernel->upperThreshold + 1;

	/* if ALL of the results were bigger than pKernel->upperThreshold,
	 * we will have found the biggest one.  We shouldn't return that.
	 * So, if the best sum is still bigger than the upper threshold,
	 * return 0.
	 */
    if (convResult > pKernel->upperThreshold)
	convResult = 0;
    return (convResult);
}

/*

$Log: miglyph1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.5  1998/01/13 23:52:22  ddavies
miglyph1.c -> /product/ipcore/ipshared/src/RCS/miglyph1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1997/07/11 01:49:46  ddavies
miglyph1.c -> /product/ipcore/ipshared/src/RCS/miglyph1.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.3  1995/12/13 02:49:31  ejaquith
miglyph1.c -> /product/ipcore/ipshared/src/RCS/miglyph1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/10/18  04:03:20  ddavies
 * miglyph1.c -> /product/ipcore/ipshared/src/RCS/miglyph1.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.1  1995/03/28  00:01:42  jcraig
 * Initial revision
 *
 * Revision 1.2  1994/03/10  01:27:27  jcraig
 * miglyph.c -> /product/ipcore/ipshared/src/RCS/miglyph.c,v
 *
 * ip_convolve3x3 the most significant bits of *pCenterXBit now
 * contain the amount that the center moved up or down,,the amount that
 * the center moved left or right.
 *
 * Revision 1.1  1994/01/22  03:49:28  ddavies
 * Initial revision
 *
 * Revision 1.10  1993/12/02  19:27:01  ddavies
 * sglyph.c -> /product/ipcore/ipshared/src/RCS/sglyph.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.9  1993/11/05  02:22:21  ddavies
 * sglyph.c -> /product/ipcore/ipshared/src/RCS/sglyph.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


