/*************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  miglyph1.c:  contains C implementation of thresholding routine. 
 *
 *          Contains subroutines
 *
 *		  Int32		ip_glyphThreshLine
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "imageref.pub"
#include "iaerror.pub"

    /*  prototypes */
#include "shrglyph.prv"		/* If we were using any of the public
				 * pointers to private structs, this would
				 * have to go first since it's where the
				 * private structures are defined.  Today,
				 * we're not using any, but you never know. */
#include "shrglyph.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: miglyph2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/*
 * i_glyphthresh()
 *	This procedure is used to threshold one line (of an image that is 
 *	believed to contain glyphs) from 8bpp to 1bpp so that the rectangle 
 * 	finding code can be ran over the resulting image.  This procedure
 *	uses an adaptive threshold and was intended to be used on
 * 	600dpi images which are frequently too light or too dark to
 * 	use a straight 128 threshold.
 *
 * Parameters:
 *	pGray		pointer to the 8bpp gray source line
 *	w		width of the source/dest image in pixels
 *	gBpl		number of bytes per line of source
 *	pSourceEnd	pointer to source word after the last referenced
 *			source word in the line
 *	pDest		pointer to the destination image
 *	mask		mask for last destination word
 *	repThreshTable	table with thresholds repeated in all 4 bytes
 *	thresh		current threshold
 *
 * Returns:
 *	modified threshold
 */
UInt8 CDECL 
ip_glyphThreshLine (
    IPState    *pIPState,
    UInt8      *pGray,
    Int32	w,
    Int32	gBpl,
    UInt8      *pSourceEnd,
    UInt8      *pDest,
    UInt32	mask,
    UInt32     *repThreshTable,
    UInt8	currThresh)
{
    PROLOGUE;
    Int32            sOffset;
    UInt32           oword;
    UInt32		 curr, prev, next, above, below, average;
    UInt32		 msbMask;

    /* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;
    pSourceEnd = pSourceEnd;

    /* we are going to ignore the last 2 bits of the source gray image.
     * why 2?  well, we want to compute the average of 4 source pixels
     * at a time so we going to have to loose some bits or we'll overflow.
     * 4 pixels are going into the average, so we'll need 2 extra bits
     * to hold the sum.
     */
    msbMask = 0xFCFCFCFC;
    oword = 0;

    /* the loop for each source word */
    curr = ULONG_ACCESS(pGray) & msbMask;
    prev = curr;
    next = ULONG_ACCESS(pGray+4) & msbMask;
    for (sOffset = 1; sOffset <= (w+3)/4; sOffset++)
    {
	/* are all the bytes in the current word >= 192 (that
	 * famous magic number)?
	 */
	if (((curr & (curr >> 1)) & 0x40404040) == 0)
	{
	    /* no, we need to adjust the threshold for each pixel.
	     * first, compute the average background for each 
	     * pixel.
	     */
	    above = (ULONG_ACCESS(pGray-2*gBpl) & msbMask) >> 2;
	    below = (ULONG_ACCESS(pGray+2*gBpl) & msbMask) >> 2;
	    average = above + below + (prev << 22) + (next >> 26) + 
		      (curr >> 10) + (curr << 6);

	    /* now adjust the threshold for each pixel.  given the
	     * average for each pixel, move the current threshold one  
	     * quarter of the distance to that new threshold.
	     */
	    oword = oword << 1;
	    currThresh += ((UInt8)(average >> 24) - currThresh) / 4U;
	    if ((curr >> 24) < currThresh)
		oword |= 1;

	    oword = oword << 1;
	    currThresh += ((UInt8)((average >> 16) & 0xFF) - currThresh) / 4U;
	    if (((curr >> 16) & 0xFF) < currThresh)
		oword |= 1;

	    oword = oword << 1;
	    currThresh += ((UInt8)((average >> 8) & 0xFF) - currThresh) / 4U;
	    if (((curr >> 8) & 0xFF) < currThresh)
		oword |= 1;

	    oword = oword << 1;
	    currThresh += ((UInt8)(average & 0xFF) - currThresh) / 4U;
	    if ((curr & 0xFF) < currThresh)
		oword |= 1;
	}
	else
	{
	UInt32	repThresh;

	    /* at least some of the pixels in the current byte are
	     * greater than 192.  we won't adjust the current 
	     * threshold but we will threshold to the current
	     * threshold.  in order to do all the bytes in the source
	     * word at once, we need the threshold repeated in each
	     * of the bytes and to ignore the least
	     * significant bit of the current threshold.  since
	     * the current pixels have had the 2 least significant
	     * bits removed this shouldn't matter.  note that
	     * the entries in the table have a 1 in their most 
	     * significant bits so that the subtract won't bother
	     * into the previous pixel.
	     */
	    repThresh = repThreshTable[currThresh];
	    /* no need to clear out the most significant bit of
	     * each byte since the least 2 significant bits were 
	     * cleared when curr was initialized.
	     */
	    repThresh = (repThresh - (curr >> 1)) & 0x80808080;

	    oword = oword << 4;
	    oword |= (((repThresh >> 28) | (repThresh >> 21) |
		       (repThresh >> 14) | (repThresh >> 7)) & 0xF);
	}

	/* is it time to write a destination word yet? */
	if ((sOffset & 0x7) == 0)
	{
	    ULONG_ACCESS(pDest) = oword;
	    pDest += 4;
	}

	/* get new pixel values */
	prev = curr;
	curr = next;
	pGray += 4;
	next = (ULONG_ACCESS(pGray+4) & msbMask);
    }
    
    /* we may need to write out a partial destination word.  if so,
     * some bits may need to be masked off the end so that we do
     * not write into the frame and then the word needs to be 
     * shifted into position
     */
    if ((w & 0x1F) == 0)
    {
	/* then we finished the loop by writing a destination word.
	 * some of the bits in the word may have actually been in
	 * the frame so mask them out.
	 */
	ULONG_ACCESS(pDest-4) = ULONG_ACCESS(pDest-4) & mask;
    }
    else
    {
	oword = (oword << (32 - ((w + 3) & 0x1F))) & mask;
	ULONG_ACCESS(pDest) = oword;
    }

    return currThresh;
}

/*

$Log: miglyph2.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.5  1998/01/13 23:52:24  ddavies
miglyph2.c -> /product/ipcore/ipshared/src/RCS/miglyph2.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1995/12/13 02:49:37  ejaquith
miglyph2.c -> /product/ipcore/ipshared/src/RCS/miglyph2.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/10/19  00:45:35  jcraig
 * miglyph2.c -> /product/ipcore/ipshared/src/RCS/miglyph2.c,v
 *
 * make constants unsigned to eliminate a compiler whining
 *
 * Revision 1.2  1995/10/18  04:03:25  ddavies
 * miglyph2.c -> /product/ipcore/ipshared/src/RCS/miglyph2.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.1  1995/03/28  00:02:44  jcraig
 * Initial revision
 *
*/


