/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * migreduc.c
 *
 *	Gray-scale min/max image reduction primitives -- "C" versions
 *
 * Exports:
 *
 *	ip_reduce2Min8bpp
 *	ip_reduce2Max8bpp
 *	ip_reduce2Min4bpp
 *	ip_reduce2Max4bpp
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"

	/* prototypes */
#include "shrscal.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: migreduc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/*
 * ip_reduce2Min8bpp
 *
 *	Inner loop of reduce2Min, which is a grayscale reduction by two
 *	which picks the minimum of four points. 
 *
 * Arguments:
 *	pD	-	Destination pointer
 *	wD	-	Number of pixels in destination
 *	pS	-	Source pointer
 *	sBpl	-	Distance between source lines 
 *
 * Returns:
 *	ia_successful, ...
 *
 */

Int32 CDECL 
ip_reduce2Min8bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    UInt32  tL,		/* top left source word */
            tR,		/* top right source word */
	    bL,		/* bottom left source word */
	    bR,		/* bottom right source word */
	    xW,		/* XOR subresult storage */
	    w1,		/* partial result word */
	    w2,		/* partial result word */
	    w3,		/* partial result word */
	    min;	/* destination word */
    Int32   sWpl,	/* source words per line */
            extraPix;	/* number of pixels in last partial word */
    UInt32 *pSW;
    UInt8  *pDEnd;	/* pts to last dest word + 1 */

    /* initialize termination address.  We use an address instead of
     * decrementing w to avoid a memory write */
    pDEnd = pD + ((wD / 4) * 4);
    extraPix = wD % 4;

    pSW = (UInt32 *)pS;
    sWpl = sBpl / 4;


    /* loop through all the destination words in the line, computing each one */
    for (; pD != pDEnd; pSW += 2, pD += 4)
    {
        /* read the four source words which will be used to compute one
	 * destination word */
	tL = ULONG_ACCESS (pSW);
	bL = ULONG_ACCESS (pSW + sWpl);
        tR = ULONG_ACCESS (pSW + 1);
	bR = ULONG_ACCESS (pSW + sWpl + 1);

	/* We'll make heavy use of the neat block of code adapted
	 * from ip_absValDiffLine8 () which simultaneously computes
	 * the difference of 4 8-bit pixels at a time.  See the
	 * comments in i_absdif.c for more info on this hack.  By
	 * skipping the last step, we have a word full of min pixels.
	 */

        /* compute the left word top to bottom min */
	xW = tL ^ bL;
	w1 = (tL | 0x80808080) - (bL & 0x7f7f7f7f);
	w1 = ((~xW) & w1) | (tL & (~bL));
	w1 &= 0x80808080;
	w1 = 0x80808080 - (w1 >> 7);
	w1 ^= 0x80808080;
	w1 &= xW;
	w1 ^= tL;


	/* compute the right word top to bottom min */
	xW = tR ^ bR;
	w2 = (tR | 0x80808080) - (bR & 0x7f7f7f7f);
	w2 = ((~xW) & w2) | (tR & (~bR));
	w2 &= 0x80808080;
	w2 = 0x80808080 - (w2 >> 7);
	w2 ^= 0x80808080;
	w2 &= xW;
	w2 ^= tR;


	/* reorder the pixels so that our two results can be compared
	 * using the same word-wise min operation */
	w3 = ((w1 & 0x00ffff00) << 8) |
	     ((w2 & 0x00ffff00) >> 8);
	w2 = (w1 & 0xff000000) |
	     ((w2 & 0xff000000) >> 16) |
	     ((w1 & 0x000000ff) << 16) |
	     (w2 & 0x000000ff);

	/* compute the final min word */
	xW = w2 ^ w3;
	min = (w2 | 0x80808080) - (w3 & 0x7f7f7f7f);
        min = ((~xW) & min) | (w2 & (~w3));
	min &= 0x80808080;
	min = 0x80808080 - (min >> 7);
	min ^= 0x80808080;
	min &= xW;
	min ^= w2;

	/* write the result word */
	ULONG_ACCESS (pD) = min;
    }

    /* take care of any pixels left in a partially full word */
    if (extraPix)
    {
        /* read the four source words which will be used to compute one
	 * destination word */
	tL = ULONG_ACCESS (pSW);
	bL = ULONG_ACCESS (pSW + sWpl);
        tR = ULONG_ACCESS (pSW + 1);
	bR = ULONG_ACCESS (pSW + sWpl + 1);

	/* We'll make heavy use of the neat block of code adapted
	 * from ip_absValDiffLine8 () which simultaneously computes
	 * the difference of 4 8-bit pixels at a time.  See the
	 * comments in i_absdif.c for more info on this hack.  By
	 * skipping the last step, we have a word full of min pixels.
	 */

        /* compute the left word top to bottom min */
	xW = tL ^ bL;
	w1 = (tL | 0x80808080) - (bL & 0x7f7f7f7f);
	w1 = ((~xW) & w1) | (tL & (~bL));
	w1 &= 0x80808080;
	w1 = 0x80808080 - (w1 >> 7);
	w1 ^= 0x80808080;
	w1 &= xW;
	w1 ^= tL;


	/* compute the right word top to bottom min */
	xW = tR ^ bR;
	w2 = (tR | 0x80808080) - (bR & 0x7f7f7f7f);
	w2 = ((~xW) & w2) | (tR & (~bR));
	w2 &= 0x80808080;
	w2 = 0x80808080 - (w2 >> 7);
	w2 ^= 0x80808080;
	w2 &= xW;
	w2 ^= tR;


	/* reorder the pixels so that our two results can be compared
	 * using the same word-wise min operation */
	w3 = ((w1 & 0x00ffff00) << 8) |
	     ((w2 & 0x00ffff00) >> 8);
	w2 = (w1 & 0xff000000) |
	     ((w2 & 0xff000000) >> 16) |
	     ((w1 & 0x000000ff) << 16) |
	     (w2 & 0x000000ff);

	/* compute the final min word */
	xW = w2 ^ w3;
	min = (w2 | 0x80808080) - (w3 & 0x7f7f7f7f);
        min = ((~xW) & min) | (w2 & (~w3));
	min &= 0x80808080;
	min = 0x80808080 - (min >> 7);
	min ^= 0x80808080;
	min &= xW;
	min ^= w2;

        /* mask out the frame pixels */
	w1 = 0xffffffff << (extraPix >> 3);
	min &= w1;

	/* write the result word */
	ULONG_ACCESS (pD) = min;
    }

    NORMAL_RETURN;

}  /* ip_reduce2Min8bpp () */


/*
 * ip_reduce2Max8bpp
 *
 *	Inner loop of reduce2Max, which is a grayscale reduction by two
 *	which picks the maximum of four points. This first version is
 *	"for research only", meaning that it's not likely to be real
 *	quick.
 *
 * Arguments:
 *	pD	-	Destination pointer
 *	wD	-	Number of pixels in destination
 *	pS	-	Source pointer
 *	sBpl	-	Distance between source lines 
 *
 * Returns:
 *	ia_successful, ...
 *
 */

Int32 CDECL 
ip_reduce2Max8bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    UInt32  tL,		/* top left source word */
            tR,		/* top right source word */
	    bL,		/* bottom left source word */
	    bR,		/* bottom right source word */
	    xW,		/* XOR subresult storage */
	    w1,		/* partial result word */
	    w2,		/* partial result word */
	    w3,		/* partial result word */
	    max;	/* destination word */
    Int32   sWpl,	/* source words per line */
            extraPix;	/* number of pixels in last partial word */
    UInt32 *pSW;
    UInt8  *pDEnd;	/* pts to last dest word + 1 */

    /* initialize termination address.  We use an address instead of
     * decrementing w to avoid a memory write */
    pDEnd = pD + ((wD / 4) * 4);
    extraPix = wD % 4;

    pSW = (UInt32 *)pS;
    sWpl = sBpl / 4;


    /* loop through all the destination words in the line, computing each one */
    for (; pD != pDEnd; pSW += 2, pD += 4)
    {
        /* read the four source words which will be used to compute one
	 * destination word */
	tL = ULONG_ACCESS (pSW);
	bL = ULONG_ACCESS (pSW + sWpl);
        tR = ULONG_ACCESS (pSW + 1);
	bR = ULONG_ACCESS (pSW + sWpl + 1);

	/* We'll make heavy use of the neat block of code adapted
	 * from ip_absValDiffLine8 () which simultaneously computes
	 * the difference of 4 8-bit pixels at a time.  See the
	 * comments in i_absdif.c for more info on this hack.  By
	 * skipping the last step, we have a word full of max pixels.
	 */

        /* compute the left word top to bottom max */
	xW = tL ^ bL;
	w1 = (tL | 0x80808080) - (bL & 0x7f7f7f7f);
	w1 = ((~xW) & w1) | (tL & (~bL));
	w1 &= 0x80808080;
	w1 = 0x80808080 - (w1 >> 7);
	w1 ^= 0x80808080;
	w1 &= xW;
	w1 ^= bL;


	/* compute the right word top to bottom max */
	xW = tR ^ bR;
	w2 = (tR | 0x80808080) - (bR & 0x7f7f7f7f);
	w2 = ((~xW) & w2) | (tR & (~bR));
	w2 &= 0x80808080;
	w2 = 0x80808080 - (w2 >> 7);
	w2 ^= 0x80808080;
	w2 &= xW;
	w2 ^= bR;


	/* reorder the pixels so that our two results can be compared
	 * using the same word-wise max operation */
	w3 = ((w1 & 0x00ffff00) << 8) |
	     ((w2 & 0x00ffff00) >> 8);
	w2 = (w1 & 0xff000000) |
	     ((w2 & 0xff000000) >> 16) |
	     ((w1 & 0x000000ff) << 16) |
	     (w2 & 0x000000ff);

	/* compute the final max word */
	xW = w2 ^ w3;
	max = (w2 | 0x80808080) - (w3 & 0x7f7f7f7f);
        max = ((~xW) & max) | (w2 & (~w3));
	max &= 0x80808080;
	max = 0x80808080 - (max >> 7);
	max ^= 0x80808080;
	max &= xW;
	max ^= w3;

	/* write the result word */
	ULONG_ACCESS (pD) = max;
    }

    /* take care of any pixels left in a partially full word */
    if (extraPix)
    {
        /* read the four source words which will be used to compute one
	 * destination word */
	tL = ULONG_ACCESS (pSW);
	bL = ULONG_ACCESS (pSW + sWpl);
        tR = ULONG_ACCESS (pSW + 1);
	bR = ULONG_ACCESS (pSW + sWpl + 1);

	/* We'll make heavy use of the neat block of code adapted
	 * from ip_absValDiffLine8 () which simultaneously computes
	 * the difference of 4 8-bit pixels at a time.  See the
	 * comments in i_absdif.c for more info on this hack.  By
	 * skipping the last step, we have a word full of min pixels.
	 */

        /* compute the left word top to bottom min */
	xW = tL ^ bL;
	w1 = (tL | 0x80808080) - (bL & 0x7f7f7f7f);
	w1 = ((~xW) & w1) | (tL & (~bL));
	w1 &= 0x80808080;
	w1 = 0x80808080 - (w1 >> 7);
	w1 ^= 0x80808080;
	w1 &= xW;
	w1 ^= tL;


	/* compute the right word top to bottom min */
	xW = tR ^ bR;
	w2 = (tR | 0x80808080) - (bR & 0x7f7f7f7f);
	w2 = ((~xW) & w2) | (tR & (~bR));
	w2 &= 0x80808080;
	w2 = 0x80808080 - (w2 >> 7);
	w2 ^= 0x80808080;
	w2 &= xW;
	w2 ^= tR;


	/* reorder the pixels so that our two results can be compared
	 * using the same word-wise min operation */
	w3 = ((w1 & 0x00ffff00) << 8) |
	     ((w2 & 0x00ffff00) >> 8);
	w2 = (w1 & 0xff000000) |
	     ((w2 & 0xff000000) >> 16) |
	     ((w1 & 0x000000ff) << 16) |
	     (w2 & 0x000000ff);

	/* compute the final max word */
	xW = w2 ^ w3;
	max = (w2 | 0x80808080) - (w3 & 0x7f7f7f7f);
        max = ((~xW) & max) | (w2 & (~w3));
	max &= 0x80808080;
	max = 0x80808080 - (max >> 7);
	max ^= 0x80808080;
	max &= xW;
	max ^= w2;

        /* mask out the frame pixels */
	w1 = 0xffffffff << (extraPix >> 3);
	max &= w1;

	/* write the result word */
	ULONG_ACCESS (pD) = max;
    }

    NORMAL_RETURN;

}  /* ip_reduce2Max8bpp () */


/*
 * ip_reduce2Min4bpp
 *
 *	Inner loop of reduce2Min, which is a grayscale reduction by two
 *	which picks the minimum of four points.
 *
 * Arguments:
 *	pD	-	Destination pointer
 *	wD	-	Number of pixels in destination
 *	pS	-	Source pointer
 *	sBpl	-	Distance between source lines 
 *
 * Returns:
 *	ia_successful, ...
 *
 */

Int32 CDECL 
ip_reduce2Min4bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    UInt32  tL,		/* top left source word */
            tR,		/* top right source word */
	    bL,		/* bottom left source word */
	    bR,		/* bottom right source word */
	    xW,		/* XOR subresult storage */
	    w1,		/* partial result word */
	    w2,		/* partial result word */
	    w3,		/* partial result word */
	    min;	/* destination word */
    Int32   sWpl,	/* source words per line */
            extraPix;	/* number of pixels in last partial word */
    UInt32 *pSW;
    UInt8  *pDEnd;	/* pts to last dest word + 1 */

    /* initialize termination address.  We use an address instead of
     * decrementing w to avoid a memory write */
    pDEnd = pD + ((wD / 8) * 4);
    extraPix = wD % 8;

    pSW = (UInt32 *)pS;
    sWpl = sBpl / 4;


    /* loop through all the destination words in the line, computing each one */
    for (; pD != pDEnd; pSW += 2, pD += 4)
    {
        /* read the four source words which will be used to compute one
	 * destination word */
	tL = ULONG_ACCESS (pSW);
	bL = ULONG_ACCESS (pSW + sWpl);
        tR = ULONG_ACCESS (pSW + 1);
	bR = ULONG_ACCESS (pSW + sWpl + 1);

	/* We'll make heavy use of the neat block of code adapted
	 * from ip_absValDiffLine4 () which simultaneously computes
	 * the difference of 8 4-bit pixels at a time.  See the
	 * comments in i_absdif.c for more info on this hack.  By
	 * skipping the last step, we have a word full of min pixels.
	 */

        /* compute the left word top to bottom min */
	xW = tL ^ bL;
	w1 = (tL | 0x88888888) - (bL & 0x77777777);
	w1 = ((~xW) & w1) | (tL & (~bL));
	w1 &= 0x88888888;
	w1 = 0x88888888 - (w1 >> 3);
	w1 ^= 0x88888888;
	w1 &= xW;
	w1 ^= tL;


	/* compute the right word top to bottom min */
	xW = tR ^ bR;
	w2 = (tR | 0x88888888) - (bR & 0x77777777);
	w2 = ((~xW) & w2) | (tR & (~bR));
	w2 &= 0x88888888;
	w2 = 0x88888888 - (w2 >> 3);
	w2 ^= 0x88888888;
	w2 &= xW;
	w2 ^= tR;


	/* reorder the pixels so that our two results can be compared
	 * using the same word-wise min operation */
	w3 = ((w1 & 0x0ff00000) << 4) |
	     ((w2 & 0x0ff00000) >> 12) |
	     ((w1 & 0x00000ff0) << 12) |
	     ((w2 & 0x00000ff0) >> 4);
	w2 = (w1 & 0xf0000000) |
	     ((w1 & 0x000ff000) << 8) |
	     ((w2 & 0x000ff000) >> 8) |
	     ((w2 & 0xf0000000) >> 16) |
	     ((w1 & 0x0000000f) << 16) |
	     (w2 & 0x0000000f);

	/* compute the final min word */
	xW = w2 ^ w3;
	min = (w2 | 0x88888888) - (w3 & 0x77777777);
        min = ((~xW) & min) | (w2 & (~w3));
	min &= 0x88888888;
	min = 0x88888888 - (min >> 3);
	min ^= 0x88888888;
	min &= xW;
	min ^= w2;

	/* write the result word */
	ULONG_ACCESS (pD) = min;
    }

    /* take care of any pixels left in a partially full word */
    if (extraPix)
    {
        /* read the four source words which will be used to compute one
	 * destination word */
	tL = ULONG_ACCESS (pSW);
	bL = ULONG_ACCESS (pSW + sWpl);
        tR = ULONG_ACCESS (pSW + 1);
	bR = ULONG_ACCESS (pSW + sWpl + 1);

	/* We'll make heavy use of the neat block of code adapted
	 * from ip_absValDiffLine4 () which simultaneously computes
	 * the difference of 8 4-bit pixels at a time.  See the
	 * comments in i_absdif.c for more info on this hack.  By
	 * skipping the last step, we have a word full of min pixels.
	 */

        /* compute the left word top to bottom min */
	xW = tL ^ bL;
	w1 = (tL | 0x88888888) - (bL & 0x77777777);
	w1 = ((~xW) & w1) | (tL & (~bL));
	w1 &= 0x88888888;
	w1 = 0x88888888 - (w1 >> 3);
	w1 ^= 0x88888888;
	w1 &= xW;
	w1 ^= tL;


	/* compute the right word top to bottom min */
	xW = tR ^ bR;
	w2 = (tR | 0x88888888) - (bR & 0x77777777);
	w2 = ((~xW) & w2) | (tR & (~bR));
	w2 &= 0x88888888;
	w2 = 0x88888888 - (w2 >> 3);
	w2 ^= 0x88888888;
	w2 &= xW;
	w2 ^= tR;


	/* reorder the pixels so that our two results can be compared
	 * using the same word-wise min operation */
	w3 = ((w1 & 0x0ff00000) << 4) |
	     ((w2 & 0x0ff00000) >> 12) |
	     ((w1 & 0x00000ff0) << 12) |
	     ((w2 & 0x00000ff0) >> 4);
	w2 = (w1 & 0xf0000000) |
	     ((w1 & 0x000ff000) << 8) |
	     ((w2 & 0x000ff000) >> 8) |
	     ((w2 & 0xf0000000) >> 16) |
	     ((w1 & 0x0000000f) << 16) |
	     (w2 & 0x0000000f);

	/* compute the final min word */
	xW = w2 ^ w3;
	min = (w2 | 0x88888888) - (w3 & 0x77777777);
        min = ((~xW) & min) | (w2 & (~w3));
	min &= 0x88888888;
	min = 0x88888888 - (min >> 3);
	min ^= 0x88888888;
	min &= xW;
	min ^= w2;

        /* mask out the frame pixels */
	w1 = 0xffffffff << (extraPix >> 2);
	min &= w1;

	/* write the result word */
	ULONG_ACCESS (pD) = min;
    }

    NORMAL_RETURN;

}  /* ip_reduce2Min4bpp () */


/*
 * ip_reduce2Max4bpp
 *
 *	Inner loop of reduce2Max, which is a grayscale reduction by two
 *	which picks the maximum of four points. This first version is
 *	"for research only", meaning that it's not likely to be real
 *	quick.
 *
 * Arguments:
 *	pD	-	Destination pointer
 *	wD	-	Number of pixels in destination
 *	pS	-	Source pointer
 *	sBpl	-	Distance between source lines 
 *
 * Returns:
 *	ia_successful, ...
 *
 */

Int32 CDECL 
ip_reduce2Max4bpp (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    UInt32  tL,		/* top left source word */
            tR,		/* top right source word */
	    bL,		/* bottom left source word */
	    bR,		/* bottom right source word */
	    xW,		/* XOR subresult storage */
	    w1,		/* partial result word */
	    w2,		/* partial result word */
	    w3,		/* partial result word */
	    max;	/* destination word */
    Int32   sWpl,	/* source words per line */
            extraPix;	/* number of pixels in last partial word */
    UInt32 *pSW;
    UInt8  *pDEnd;	/* pts to last dest word + 1 */

    /* initialize termaxation address.  We use an address instead of
     * decrementing w to avoid a memory write */
    pDEnd = pD + ((wD / 8) * 4);
    extraPix = wD % 8;

    pSW = (UInt32 *)pS;
    sWpl = sBpl / 4;


    /* loop through all the destination words in the line, computing each one */
    for (; pD != pDEnd; pSW += 2, pD += 4)
    {
        /* read the four source words which will be used to compute one
	 * destination word */
	tL = ULONG_ACCESS (pSW);
	bL = ULONG_ACCESS (pSW + sWpl);
        tR = ULONG_ACCESS (pSW + 1);
	bR = ULONG_ACCESS (pSW + sWpl + 1);

	/* We'll make heavy use of the neat block of code adapted
	 * from ip_absValDiffLine4 () which simultaneously computes
	 * the difference of 8 4-bit pixels at a time.  See the
	 * comments in i_absdif.c for more info on this hack.  By
	 * skipping the last step, we have a word full of min pixels.
	 */

        /* compute the left word top to bottom max */
	xW = tL ^ bL;
	w1 = (tL | 0x88888888) - (bL & 0x77777777);
	w1 = ((~xW) & w1) | (tL & (~bL));
	w1 &= 0x88888888;
	w1 = 0x88888888 - (w1 >> 3);
	w1 ^= 0x88888888;
	w1 &= xW;
	w1 ^= bL;


	/* compute the right word top to bottom max */
	xW = tR ^ bR;
	w2 = (tR | 0x88888888) - (bR & 0x77777777);
	w2 = ((~xW) & w2) | (tR & (~bR));
	w2 &= 0x88888888;
	w2 = 0x88888888 - (w2 >> 3);
	w2 ^= 0x88888888;
	w2 &= xW;
	w2 ^= bR;


	/* reorder the pixels so that our two results can be compared
	 * using the same word-wise max operation */
	w3 = ((w1 & 0x0ff00000) << 4) |
	     ((w2 & 0x0ff00000) >> 12) |
	     ((w1 & 0x00000ff0) << 12) |
	     ((w2 & 0x00000ff0) >> 4);
	w2 = (w1 & 0xf0000000) |
	     ((w2 & 0xf0000000) >> 16) |
	     ((w2 & 0x000ff000) >> 8) |
	     ((w1 & 0x000ff000) << 8) |
	     ((w1 & 0x0000000f) << 16) |
	     (w2 & 0x0000000f);

	/* compute the final max word */
	xW = w2 ^ w3;
	max = (w2 | 0x88888888) - (w3 & 0x77777777);
        max = ((~xW) & max) | (w2 & (~w3));
	max &= 0x88888888;
	max = 0x88888888 - (max >> 3);
	max ^= 0x88888888;
	max &= xW;
	max ^= w3;

	/* write the result word */
	ULONG_ACCESS (pD) = max;
    }

    /* take care of any pixels left in a partially full word */
    if (extraPix)
    {
        /* read the four source words which will be used to compute one
	 * destination word */
	tL = ULONG_ACCESS (pSW);
	bL = ULONG_ACCESS (pSW + sWpl);
        tR = ULONG_ACCESS (pSW + 1);
	bR = ULONG_ACCESS (pSW + sWpl + 1);

	/* We'll make heavy use of the neat block of code adapted
	 * from ip_absValDiffLine4 () which simultaneously computes
	 * the difference of 8 4-bit pixels at a time.  See the
	 * comments in i_absdif.c for more info on this hack.  By
	 * skipping the last step, we have a word full of min pixels.
	 */

        /* compute the left word top to bottom min */
	xW = tL ^ bL;
	w1 = (tL | 0x88888888) - (bL & 0x77777777);
	w1 = ((~xW) & w1) | (tL & (~bL));
	w1 &= 0x88888888;
	w1 = 0x88888888 - (w1 >> 3);
	w1 ^= 0x88888888;
	w1 &= xW;
	w1 ^= tL;


	/* compute the right word top to bottom min */
	xW = tR ^ bR;
	w2 = (tR | 0x88888888) - (bR & 0x77777777);
	w2 = ((~xW) & w2) | (tR & (~bR));
	w2 &= 0x88888888;
	w2 = 0x88888888 - (w2 >> 3);
	w2 ^= 0x88888888;
	w2 &= xW;
	w2 ^= tR;


	/* reorder the pixels so that our two results can be compared
	 * using the same word-wise min operation */
	w3 = ((w1 & 0x0ff00000) << 4) |
	     ((w2 & 0x0ff00000) >> 12) |
	     ((w1 & 0x00000ff0) << 12) |
	     ((w2 & 0x00000ff0) >> 4);
	w2 = (w1 & 0xf0000000) |
	     ((w1 & 0x000ff000) << 8) |
	     ((w2 & 0x000ff000) >> 8) |
	     ((w2 & 0xf0000000) >> 16) |
	     ((w1 & 0x0000000f) << 16) |
	     (w2 & 0x0000000f);

	/* compute the final max word */
	xW = w2 ^ w3;
	max = (w2 | 0x88888888) - (w3 & 0x77777777);
        max = ((~xW) & max) | (w2 & (~w3));
	max &= 0x88888888;
	max = 0x88888888 - (max >> 3);
	max ^= 0x88888888;
	max &= xW;
	max ^= w2;

        /* mask out the frame pixels */
	w1 = 0xffffffff << (extraPix >> 2);
	max &= w1;

	/* write the result word */
	ULONG_ACCESS (pD) = max;
    }

    NORMAL_RETURN;

}  /* ip_reduce2Max4bpp () */

	     
/*
$Log: migreduc.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.9  1998/01/13 23:52:25  ddavies
migreduc.c -> /product/ipcore/ipshared/src/RCS/migreduc.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1996/06/26 22:19:55  dferrell
dferrell on Wed Jun 26 15:17:24 PDT 1996

migreduc.c -> /product/ipcore/ipshared/src/RCS/migreduc.c,v

Optimized reduction by 2 min/max functions and fixed a bug in the 8-bit
cases.  The 8-bit functions are ~30% faster and the 4-bit, 100%.

Revision 1.7  1995/12/13 02:49:44  ejaquith
migreduc.c -> /product/ipcore/ipshared/src/RCS/migreduc.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/02/15  21:12:27  macleod
 * macleod on Wed Feb 15 13:11:16 PST 1995
 *
 * migreduc.c -> /product/ipcore/ipshared/src/RCS/migreduc.c,v
 *
 * if ((unsigned_char & 0xf) < unsigned_char) generates a warning about
 * changing semantics in ANSI C when compiled with "acc". Made the constant
 * unsigned.
 *
 * Revision 1.5  1995/02/14  02:59:00  macleod
 * migreduc.c -> /product/ipcore/ipshared/src/RCS/migreduc.c,v
 *
 * Added 4-bit/pixel versions of min-or-four and max-of-four x2 reductions
 *
*/
