/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  migrybin.c: contains primitives for combining gray and binary images
 *
 *         Combine Gray image with Binary image:
 *                  Int32     	ip_mask2bppLine()
 *                  Int32	ip_paint2bppLine()
 *                  Int32     	ip_mask4bppLine()
 *                  Int32	ip_paint4bppLine()
 *                  Int32     	ip_mask8bppLine()
 *                  Int32	ip_paint8bppLine()
 *
 *                  Int32     	ip_grayBinarySelLine()
 *                  Int32	ip_grayBinarySelLineNoFG()
 *                  Int32	ip_grayBinarySelLineNoBG()
 *                  Int32	ip_grayBinarySelLineNoFGNoBG()
 *                  Int32     	ip_gray4bppSelLine()
 *                  Int32	ip_gray4bppSelLineNoFG()
 *                  Int32	ip_gray4bppSelLineNoBG()
 *                  Int32	ip_gray4bppSelLineNoFGNoBG()
 */


#include <stddef.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /*  prototyes */
#include "shrip.prv"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: migrybn1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* This file contains C implementations of the single line primitives
 * used by i_opGrayBin and i_opGrayBinIP.  Please see i_grybin.c for
 * overall comments.
 */

/* If the depth of the gray image is 2 or 4, each loop iteration writes
 * one 32-bit destination word.  This may write into the frame.  If this
 * is a paint operation it doesn't matter.  If this is a mask operation
 * the frame could contains 1's so we must clear the right frame
 * after we are finished.  If the depth of the gray image is 8, each
 * loop iteration writes two 32-bit destination words.  This garbages
 * more that just the right frame so we have to take care not to write
 * the second word unless absolutely necessary.
 */



/*********************************************************************
 *************************      Procedures     ***********************
 *********************************************************************/
/* PRIVATE */
/*************************************************************************** 
 *  ip_mask2bppLine()
 *  		Combines one 2 bit/pixel gray line with one
 *		line of a binary image, producing a 2 bit/pixel line.
 *		The desired result is to have white in the output gray
 *		image everywhere there is a white pixel in the source image.
 *		A white gray pixel has a value of 3.  A white binary pixel
 *		has a value of 0.  Therefore,
 *		ResultGrayPix(x,y) = SrcGrayPix(x,y) OR ~(BinaryPixel(x,y)).
 *	Input:
 *		pG =   		Pointer to left end of gray source image line
 *				  INSIDE frame
 *      	pB =  		Pointer to left end of binary image line
 *				  INSIDE frame.
 *		pD =   		Pointer to left end of gray dest image line
 *				  INSIDE frame.
 *      	w =		Width of dest image in bytes
 *				  excluding frame
 *		tabExp2 =	Pointer to table used to expand binary image
 *				  to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_mask2bppLine (
    IPState    *pIPState,
    UInt8      *pG,
    UInt8      *pB,
    UInt8      *pD,
    UInt32	w,
    UInt16     *tabExp2)
{
    PROLOGUE;
    UInt32	byte;			/* iteration var */
    UInt32	expBinaryData;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for (byte = 0; byte < w; byte += 4)
    {
	expBinaryData = (tabExp2[(Int32) UCHAR_ACCESS(&pB[byte/2])] << 16) |
			 tabExp2[(Int32) UCHAR_ACCESS(&pB[(byte/2)+1])];
	ULONG_ACCESS(&pD[byte]) = ULONG_ACCESS(&pG[byte]) | ~expBinaryData;
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_paint2bppLine()
 *  		Combines one 2 bit/pixel gray line with one
 *		line of a binary image, producing a 2 bit/pixel line.
 *		The desired result is to have black in the output gray
 *		image everywhere there is a black pixel in the source image.
 *		A black gray pixel has a value of 0.  A black binary pixel
 *		has a value of 1.  Therefore,
 *		ResultGrayPix(x,y) = SrcGrayPix(x,y) AND ~(BinaryPixel(x,y)).
 *	Input:
 *		pG =   		Pointer to left end of gray source image line
 *				  INSIDE frame
 *      	pB =  		Pointer to left end of binary image line
 *				  INSIDE frame.
 *		pD =   		Pointer to left end of gray dest image line
 *				  INSIDE frame.
 *      	w =		Width of dest image in bytes
 *				  excluding frame
 *		tabExp2 =	Pointer to table used to expand binary image
 *				  to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_paint2bppLine (
    IPState    *pIPState,
    UInt8      *pG,
    UInt8      *pB,
    UInt8      *pD,
    UInt32	w,
    UInt16     *tabExp2)
{
    PROLOGUE;
    UInt32	byte;			/* iteration var */
    UInt32	expBinaryData;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for (byte = 0; byte < w; byte += 4)
    {
	expBinaryData = (tabExp2[(Int32) UCHAR_ACCESS(&pB[byte/2])] << 16) |
			 tabExp2[(Int32) UCHAR_ACCESS(&pB[(byte/2)+1])];
	ULONG_ACCESS(&pD[byte]) = ULONG_ACCESS(&pG[byte]) & ~expBinaryData;
    }
}



/* PRIVATE */
/*************************************************************************** 
 *  ip_mask4bppLine()
 *  		Combines one 4 bit/pixel gray line with one
 *		line of a binary image, producing a 4 bit/pixel line.
 *		The desired result is to have white in the output gray
 *		image everywhere there is a white pixel in the source image.
 *		A white gray pixel has a value of 15.  A white binary pixel
 *		has a value of 0.  Therefore,
 *		ResultGrayPix(x,y) = SrcGrayPix(x,y) OR ~(BinaryPixel(x,y)).
 *	Input:
 *		pG =   		Pointer to left end of gray source image line
 *				  INSIDE frame
 *      	pB =  		Pointer to left end of binary image line
 *				  INSIDE frame.
 *		pD =   		Pointer to left end of gray dest image line
 *				  INSIDE frame.
 *      	w =		Width of dest image in bytes
 *				  excluding frame
 *		tabExp4 =	Pointer to table used to expand binary image
 *				  to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_mask4bppLine (
    IPState    *pIPState,
    UInt8      *pG,
    UInt8      *pB,
    UInt8      *pD,
    UInt32	w,
    UInt32     *tabExp4)
{
    PROLOGUE;
    UInt32	byte;			/* iteration var */
    UInt32	expBinaryData;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for (byte = 0; byte < w; byte += 4)
    {
	expBinaryData = tabExp4[(Int32) UCHAR_ACCESS(&pB[byte/4])];
	ULONG_ACCESS(&pD[byte]) = ULONG_ACCESS(&pG[byte]) | ~expBinaryData;
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_paint4bppLine()
 *  		Combines one 4 bit/pixel gray line with one
 *		line of a binary image, producing a 4 bit/pixel line.
 *		The desired result is to have black in the output gray
 *		image everywhere there is a black pixel in the source image.
 *		A black gray pixel has a value of 0.  A black binary pixel
 *		has a value of 1.  Therefore,
 *		ResultGrayPix(x,y) = SrcGrayPix(x,y) AND ~(BinaryPixel(x,y)).
 *	Input:
 *		pG =   		Pointer to left end of gray source image line
 *				  INSIDE frame
 *      	pB =  		Pointer to left end of binary image line
 *				  INSIDE frame.
 *		pD =   		Pointer to left end of gray dest image line
 *				  INSIDE frame.
 *      	w =		Width of dest image in bytes
 *				  excluding frame
 *		tabExp4 =	Pointer to table used to expand binary image
 *				  to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_paint4bppLine (
    IPState    *pIPState,
    UInt8      *pG,
    UInt8      *pB,
    UInt8      *pD,
    UInt32	w,
    UInt32     *tabExp4)
{
    PROLOGUE;
    UInt32	byte;			/* iteration var */
    UInt32	expBinaryData;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for (byte = 0; byte < w; byte += 4)
    {
	expBinaryData = tabExp4[(Int32) UCHAR_ACCESS(&pB[byte/4])];
	ULONG_ACCESS(&pD[byte]) = ULONG_ACCESS(&pG[byte]) & ~expBinaryData;
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_mask8bppLine()
 *  		Combines one 8 bit/pixel gray line with one
 *		line of a binary image, producing a 8 bit/pixel line.
 *		The desired result is to have white in the output gray
 *		image everywhere there is a white pixel in the source image.
 *		A white gray pixel has a value of 255.  A white binary pixel
 *		has a value of 0.  Therefore,
 *		ResultGrayPix(x,y) = SrcGrayPix(x,y) OR ~(BinaryPixel(x,y)).
 *	Input:
 *		pG =   		Pointer to left end of gray source image line
 *				  INSIDE frame
 *      	pB =  		Pointer to left end of binary image line
 *				  INSIDE frame.
 *		pD =   		Pointer to left end of gray dest image line
 *				  INSIDE frame.
 *      	w =		Width of dest image in bytes
 *				  excluding frame
 *		tabExp8 =	Pointer to table used to expand binary image
 *				  to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_mask8bppLine (
    IPState    *pIPState,
    UInt8      *pG,
    UInt8      *pB,
    UInt8      *pD,
    UInt32	w,
    UInt32     *tabExp8)
{
    PROLOGUE;
    UInt32	 byte;			/* iteration var */
    UInt8	 inputByte;
    UInt32	 expBinaryData;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for (byte = 0; byte < w; byte += 4)
    {
	inputByte = UCHAR_ACCESS(&pB[byte/8]);
	expBinaryData = tabExp8[(Int32) (inputByte >> 4)];
	ULONG_ACCESS(&pD[byte]) = ULONG_ACCESS(&pG[byte]) | ~expBinaryData;
	byte += 4;
	if (byte < w)
	{
	expBinaryData = tabExp8[(Int32) (inputByte & 0xF)];
	ULONG_ACCESS(&pD[byte]) = ULONG_ACCESS(&pG[byte]) | ~expBinaryData;
	}
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_paint8bppLine()
 *  		Combines one 8 bit/pixel gray line with one
 *		line of a binary image, producing a 8 bit/pixel line.
 *		The desired result is to have black in the output gray
 *		image everywhere there is a black pixel in the source image.
 *		A black gray pixel has a value of 0.  A black binary pixel
 *		has a value of 1.  Therefore,
 *		ResultGrayPix(x,y) = SrcGrayPix(x,y) AND ~(BinaryPixel(x,y)).
 *	Input:
 *		pG =   		Pointer to left end of gray source image line
 *				  INSIDE frame
 *      	pB =  		Pointer to left end of binary image line
 *				  INSIDE frame.
 *		pD =   		Pointer to left end of gray dest image line
 *				  INSIDE frame.
 *      	w =		Width of dest image in bytes
 *				  excluding frame
 *		tabExp8 =	Pointer to table used to expand binary image
 *				  to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_paint8bppLine (
    IPState    *pIPState,
    UInt8      *pG,
    UInt8      *pB,
    UInt8      *pD,
    UInt32	w,
    UInt32     *tabExp8)
{
    PROLOGUE;
    UInt32	byte;			/* iteration var */
   UInt8	inputByte;
   UInt32	expBinaryData;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    for (byte = 0; byte < w; byte += 4)
    {
	inputByte = UCHAR_ACCESS(&pB[byte/8]);
	expBinaryData = tabExp8[(Int32) (inputByte >> 4)];
	ULONG_ACCESS(&pD[byte]) = ULONG_ACCESS(&pG[byte]) & ~expBinaryData;
	byte += 4;
	if (byte < w)
	{
	expBinaryData = tabExp8[(Int32) (inputByte & 0xF)];
	ULONG_ACCESS(&pD[byte]) = ULONG_ACCESS(&pG[byte]) & ~expBinaryData;
	}
    }
}


/*

$Log: migrybn1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.2  1998/01/13 23:52:27  ddavies
migrybn1.c -> /product/ipcore/ipshared/src/RCS/migrybn1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.1  1997/06/06 04:19:20  ddavies
Initial revision

Revision 1.4  1995/12/13 02:49:50  ejaquith
migrybin.c -> /product/ipcore/ipshared/src/RCS/migrybin.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1994/07/15  23:39:19  ejaquith
 * ejaquith on Fri Jul 15 16:38:47 PDT 1994
 *
 * migrybin.c -> /product/ipcore/ipshared/src/RCS/migrybin.c,v
 *
 * Removed #include "iadef.prv"
 *
 * Revision 1.2  1994/02/04  20:18:15  dferrell
 * migrybin.c -> /product/ipcore/ipshared/src/RCS/migrybin.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/14  04:31:41  ddavies
 * Initial revision
 *
 * Revision 1.15  1993/12/02  19:27:06  ddavies
 * sgrbinop.c -> /product/ipcore/ipshared/src/RCS/sgrbinop.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.14  1993/11/05  02:22:30  ddavies
 * sgrbinop.c -> /product/ipcore/ipshared/src/RCS/sgrbinop.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


