/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  migrybin.c: contains primitives for combining gray and binary images
 *
 *         Combine Gray image with Binary image:
 *                  Int32     	ip_grayBinarySelLine()
 *                  Int32	ip_grayBinarySelLineNoFG()
 *                  Int32	ip_grayBinarySelLineNoBG()
 *                  Int32	ip_grayBinarySelLineNoFGNoBG()
 *                  Int32     	ip_gray4bppSelLine()
 *                  Int32	ip_gray4bppSelLineNoFG()
 *                  Int32	ip_gray4bppSelLineNoBG()
 *                  Int32	ip_gray4bppSelLineNoFGNoBG()
 *
 */


#include <stddef.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"
#include "graysel.prv"

    /*  prototyes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: migrybn2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* This file contains C implementations of the single line primitives
 * used by i_graySelect.  Please see i_grybin.c for overall comments.
 */


/*********************************************************************
 *************************      Procedures     ***********************
 *********************************************************************/


/* PRIVATE */
/*************************************************************************** 
 *  ip_grayBinarySelLine()
 *  		Combines a gray foreground and a gray background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose background pixel.  A 1 in the selector
 *		=> choose foreground pixel.
 *	Input:
 *		pBackLong =   	Pointer to left end of background source
 *				image line INSIDE frame
 *		pForeLong =   	Pointer to left end of foreground source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *		tabExp8 =	Pointer to table used to expand binary image
 *				to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_grayBinarySelLine (
    IPState    *pIPState,
    UInt32     *pBackLong,
    UInt32     *pForeLong,
    UInt8      *pSel,
    UInt32     *pDLong,
    Int32	width,
    UInt32     *tabExp8)
{
    PROLOGUE;
    Int32	 j;		/* iteration var */
    UInt8	 bits;
    UInt32	*tabExp8Hi = tabExp8 + 256;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Take 4 selection bits at a time, and write an output word.
     * Don't get close to the end here so we can eliminate the
     * the test inside the main line code.
     */
    for (j = 0; j < width - 4; j += 8)
    {
	bits = UCHAR_ACCESS (&pSel[j / 8]);
	
	    /* But Wait!  Most pixels in selection bitmaps
	     * are white. if this byte is white, just copy
	     * Background pixels.
	     */
	if (bits == 0)
	{
	    pDLong[0] = pBackLong[0];
	    pDLong[1] = pBackLong[1];
	}
	else	/* need to combine foreground and background */
	{
	    pDLong[0] = (pForeLong[0] & ~tabExp8Hi[bits]) |
			(pBackLong[0] & (tabExp8Hi[bits]));
	    pDLong[1] = (pForeLong[1] & ~tabExp8[bits]) |
			(pBackLong[1] & (tabExp8[bits]));
	}
	pDLong += 2;
	pBackLong += 2;
	pForeLong += 2;
    }

	/* Now see if there's any more to do.  There can
	 * be at most 4 more pixels, so one more store will
	 * do it.
	 */
    if (j < width)
    {
	bits = UCHAR_ACCESS (&pSel[j / 8]);
	
	    /* But Wait!  Most pixels in selection bitmaps
	     * are white.  If this byte is white, just copy
	     * Background pixels.
	     */
	if (bits == 0)
	{
	    pDLong[0] = pBackLong[0];
	}
	else	/* need to combine foreground and background */
	{
	    pDLong[0] = (pForeLong[0] & ~tabExp8Hi[bits]) |
			(pBackLong[0] & (tabExp8Hi[bits]));
	}
    }
}



/* PRIVATE */
/*************************************************************************** 
 *  ip_grayBinarySelLineNoFG()
 *  		Combines a virtual foreground and a gray background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose background pixel.  A 1 in the selector
 *		=> choose foreground pixel.
 *		The virtual foreground image is entire black (pix == 0)
 *	Input:
 *		pBackLong =   	Pointer to left end of background source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *		tabExp8 =	Pointer to table used to expand binary image
 *				to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_grayBinarySelLineNoFG (
    IPState    *pIPState,
    UInt32     *pBackLong,
    UInt8      *pSel,
    UInt32     *pDLong,
    Int32	width,
    UInt32     *tabExp8)
{
    PROLOGUE;
    Int32	 j;		/* iteration var */
    UInt8	 bits;
    UInt32	*tabExp8Hi = tabExp8 + 256;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Take 4 selection bits at a time, and write an output word.
     * Don't get close to the end here so we can eliminate the
     * the test inside the main line code.
     */
    for (j = 0; j < width - 4; j += 8)
    {
	bits = UCHAR_ACCESS (&pSel[j / 8]);
	
	    /* But Wait!  Most pixels in selection bitmaps
	     * are white. if this byte is white, just copy
	     * Background pixels.
	     */
	if (bits == 0)
	{
	    pDLong[0] = pBackLong[0];
	    pDLong[1] = pBackLong[1];
	}
	else	    /* need to combine background with virtual
		     * foreground pixels having values of 0.
		     */
	{
	    pDLong[0] = (pBackLong[0] & tabExp8Hi[bits]);
	    pDLong[1] = (pBackLong[1] & tabExp8[bits]);
	}
	pDLong += 2;
	pBackLong += 2;
    }

	/* Now see if there's any more to do.  There can
	 * be at most 4 more pixels, so one more store will
	 * do it.
	 */
    if (j < width)
    {
	bits = UCHAR_ACCESS (&pSel[j / 8]);
	
	    /* But Wait!  Most pixels in selection bitmaps
	     * are white.  If this byte is white, just copy
	     * Background pixels.
	     */
	if (bits == 0)
	{
	    pDLong[0] = pBackLong[0];
	}
	else	/* need to combine foreground and background */
	{
	    pDLong[0] = (pBackLong[0] & (tabExp8Hi[bits]));
	}
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_grayBinarySelLineNoBG()
 *  		Combines a gray foreground and a virtual background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose background pixel.  A 1 in the selector
 *		=> choose foreground pixel.
 *		The virtual background image is entire white (pix == 0xFF)
 *	Input:
 *		pForeLong =   	Pointer to left end of foreground source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *		tabExp8 =	Pointer to table used to expand binary image
 *				to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_grayBinarySelLineNoBG (
    IPState    *pIPState,
    UInt32     *pForeLong,
    UInt8      *pSel,
    UInt32     *pDLong,
    Int32	width,
    UInt32     *tabExp8)
{
    PROLOGUE;
    Int32	 j;		/* iteration var */
    UInt8	 bits;
    UInt32	*tabExp8Hi = tabExp8 + 256;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Take 4 selection bits at a time, and write an output word.
     * Don't get close to the end here so we can eliminate the
     * the test inside the main line code.
     */
    for (j = 0; j < width - 4; j += 8)
    {
	bits = UCHAR_ACCESS (&pSel[j / 8]);
	
	    /* But Wait!  Most pixels in selection bitmaps
	     * are white. if this byte is white, just copy
	     * Background pixels.
	     */
	if (bits == 0)
	{
	    pDLong[0] = 0xFFFFFFFF;
	    pDLong[1] = 0xFFFFFFFF;
	}
	else	/* need to combine foreground and background */
	{
	    pDLong[0] = pForeLong[0] | tabExp8Hi[bits];
	    pDLong[1] = pForeLong[1] | tabExp8[bits];
	}
	pDLong += 2;
	pForeLong += 2;
    }

	/* Now see if there's any more to do.  There can
	 * be at most 4 more pixels, so one more store will
	 * do it.
	 */
    if (j < width)
    {
	bits = UCHAR_ACCESS (&pSel[j / 8]);
	
	    /* But Wait!  Most pixels in selection bitmaps
	     * are white.  If this byte is white, just copy
	     * Background pixels.
	     */
	if (bits == 0)
	{
	    pDLong[0] = 0xFFFFFFFF;
	}
	else	/* need to combine foreground and background */
	{
	    pDLong[0] = pForeLong[0] | tabExp8Hi[bits];
	}
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_grayBinarySelLineNoFGNoBG()
 *  		Combines a virtual foreground and a virtual background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose background pixel.  A 1 in the selector
 *		=> choose foreground pixel.
 *		The virtual foreground image is entire black (pix == 0)
 *		The virtual background image is entire white (pix == 0xFF)
 *	Input:
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *		tabExp8 =	Pointer to table used to expand binary image
 *				to same width as gray image for logic ops.
 *
 **************************************************************************/
void  CDECL 
ip_grayBinarySelLineNoFGNoBG (
    IPState    *pIPState,
    UInt8      *pSel,
    UInt32     *pDLong,
    Int32	width,
    UInt32     *tabExp8)
{
    PROLOGUE;
    Int32	 j;		/* iteration var */
    UInt8	 bits;
    UInt32	*tabExp8Hi =	 tabExp8 + 256;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* Take 4 selection bits at a time, and write an output word.
     * Don't get close to the end here so we can eliminate the
     * the test inside the main line code.  We don't do the
     * background detection step since using the mask is so
     * trivial in this case.
     */
    for (j = 0; j < width - 4; j += 8)
    {
	bits = UCHAR_ACCESS (&pSel[j / 8]);
	
	pDLong[0] = tabExp8Hi[bits];
	pDLong[1] = tabExp8[bits];

	pDLong += 2;
    }

	/* Now see if there's any more to do.  There can
	 * be at most 4 more pixels, so one more store will
	 * do it.
	 */
    if (j < width)
    {
	bits = UCHAR_ACCESS (&pSel[j / 8]);
	
	pDLong[0] = tabExp8Hi[bits];
    }
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_gray4bppSelLine()
 *  		Combines a gray foreground and a gray background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose foreground pixel.  A 0xF in the selector
 *		=> choose background pixel.  Other values result in weighted
 *		combinations of foreground and background.
 *	Input:
 *		pBack =   	Pointer to left end of background source
 *				image line INSIDE frame
 *		pFore =   	Pointer to left end of foreground source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *
 **************************************************************************/
void  CDECL 
ip_gray4bppSelLine (
    IPState    *pIPState,
    UInt8      *pBack,
    UInt8      *pFore,
    UInt8      *pSel,
    UInt32     *pDLong,
    UInt32	width)
{
    PROLOGUE;
    Int32	 j;		/* iteration var */
    UInt32	 maskBits;
    UInt8	 bits;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* the frames contain 32 gray pixels, so it doesn't
	 * matter if we do some extra so long as we clear
	 * them out.  Note that the frame pixels of both
	 * Back and Fore are zeros, so the combination is also 0.
	 * Therefore, any dest frame pixels will be written
	 * with 0's, which is fine.
	 * We process from right to left to save a little time
	 * testing j.
	 */
    for (j = ((width - 1)/2) & ~0x3,
	   pDLong += ((width - 1) & ~0x7) / 4,
	   pBack += (width - 1) & ~0x7,
	   pFore += (width - 1) & ~0x7;
	 j >= 0;
	 j -= 4, pDLong -= 2, pBack -= 8, pFore -= 8)
    {
	    /* if the next four selector bytes are filled with all
	     * 1's (remember that we invert the sense of gray
	     * selectors), just copy the next eight Back
	     * (background) bytes to the output.
	     */
	if (ULONG_ACCESS(&pSel[j]) == 0xFFFFFFFF)
	{
	    pDLong[0] = ULONG_ACCESS(&pBack[0]);
	    pDLong[1] = ULONG_ACCESS(&pBack[4]);
	}
	else
	{
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
    
	    bits = pSel[j];
	    maskBits = (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[0]] + 
		 foreTab[((UInt32)(bits & 0xF0) << 4) |
			pFore[0]]);

	    maskBits <<= 8;		/* ms bit to bit 15 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[1]] + 
		 foreTab[((UInt32)bits << 8) | pFore[1]]);

	    maskBits <<= 8;		/* ms bit to bit 23 */

	    bits = pSel[j + 1];
	    maskBits |= (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[2]] + 
		 foreTab[((UInt32)(bits & 0xF0) << 4) |
			pFore[2]]);

	    maskBits <<= 8;	/* ms bit to bit 31 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[3]] + 
		 foreTab[((UInt32)bits << 8) | pFore[3]]);

	    pDLong[0] = maskBits;

	    bits = pSel[j + 2];
	    maskBits = (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[4]] + 
		 foreTab[((UInt32)(bits & 0xF0) << 4) |
			pFore[4]]);

	    maskBits <<= 8;		/* ms bit to bit 15 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[5]] + 
		 foreTab[((UInt32)bits << 8) | pFore[5]]);

	    maskBits <<= 8;		/* ms bit to bit 23 */

	    bits = pSel[j + 3];
	    maskBits |= (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[6]] + 
		 foreTab[((UInt32)(bits & 0xF0) << 4) |
			pFore[6]]);

	    maskBits <<= 8;	/* ms bit to bit 31 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[7]] + 
		 foreTab[((UInt32)bits << 8) | pFore[7]]);
    
#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    
	    bits = pSel[j + 3];
	    maskBits = (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[3]] + 
		 foreTab[((UInt32)(bits & 0xF0) << 4) |
			pFore[3]]);

	    maskBits <<= 8;		/* ms bit to bit 15 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[2]] + 
		 foreTab[((UInt32)bits << 8) | pFore[2]]);

	    maskBits <<= 8;		/* ms bit to bit 23 */

	    bits = pSel[j + 2];
	    maskBits |= (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[1]] + 
		 foreTab[((UInt32)(bits & 0xF0) << 4) |
			pFore[1]]);

	    maskBits <<= 8;	/* ms bit to bit 31 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[0]] + 
		 foreTab[((UInt32)bits << 8) | pFore[0]]);

	    pDLong[0] = maskBits;

	    bits = pSel[j + 1];
	    maskBits = (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[7]] + 
		 foreTab[((UInt32)(bits & 0xF0) << 4) |
			pFore[7]]);

	    maskBits <<= 8;		/* ms bit to bit 15 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[6]] + 
		 foreTab[((UInt32)bits << 8) | pFore[6]]);

	    maskBits <<= 8;		/* ms bit to bit 23 */

	    bits = pSel[j];
	    maskBits |= (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[5]] + 
		 foreTab[((UInt32)(bits & 0xF0) << 4) |
			pFore[5]]);

	    maskBits <<= 8;	/* ms bit to bit 31 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[4]] + 
		 foreTab[((UInt32)bits << 8) | pFore[4]]);
    
#endif
	    pDLong[1] = maskBits;
	}
    }   /* do more pixels if necessary */
}



/* PRIVATE */
/*************************************************************************** 
 *  ip_gray4bppSelLineNoFG()
 *  		Combines a virtual foreground and a gray background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose foreground pixel.  A 0xF in the selector
 *		=> choose background pixel.  Other values result in weighted
 *		combinations of foreground and background.
 *		All pixels in the virtual foreground image a black (0x00).
 *
 *	Input:
 *		pBack =   	Pointer to left end of background source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *
 **************************************************************************/
void  CDECL 
ip_gray4bppSelLineNoFG (
    IPState    *pIPState,
    UInt8      *pBack,
    UInt8      *pSel,
    UInt32     *pDLong,
    UInt32	width)
{
    PROLOGUE;
    Int32	 j;		/* iteration var */
    UInt32	 maskBits;
    UInt8	 bits;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* the frames contain 32 gray pixels, so it doesn't
	 * matter if we do some extra so long as we clear
	 * them out.  Note that the frame pixels of both
	 * Back and Fore are zeros, so the combination is also 0.
	 * Therefore, any dest frame pixels will be written
	 * with 0's, which is fine.
	 * We process from right to left to save a little time
	 * testing j.
	 */
    for (j = ((width - 1)/2) & ~0x3,
	   pDLong += ((width - 1) & ~0x7) / 4,
	   pBack += (width - 1) & ~0x7;
	 j >= 0;
	 j -= 4, pDLong -= 2, pBack -= 8)
    {
	    /* if the next four selector bytes are filled with all
	     * 1's (remember that we invert the sense of gray
	     * selectors), just copy the next eight Back
	     * (background) bytes to the output.
	     */
	if (ULONG_ACCESS(&pSel[j]) == 0xFFFFFFFF)
	{
	    pDLong[0] = ULONG_ACCESS(&pBack[0]);
	    pDLong[1] = ULONG_ACCESS(&pBack[4]);
	}
	else
	{
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
    
	    bits = pSel[j];
	    maskBits = (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[0]]);

	    maskBits <<= 8;		/* ms bit to bit 15 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[1]]);

	    maskBits <<= 8;		/* ms bit to bit 23 */

	    bits = pSel[j + 1];
	    maskBits |= (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[2]]);

	    maskBits <<= 8;	/* ms bit to bit 31 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[3]]);

	    pDLong[0] = maskBits;

	    bits = pSel[j + 2];
	    maskBits = (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[4]]);

	    maskBits <<= 8;		/* ms bit to bit 15 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[5]]);

	    maskBits <<= 8;		/* ms bit to bit 23 */

	    bits = pSel[j + 3];
	    maskBits |= (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[6]]);

	    maskBits <<= 8;	/* ms bit to bit 31 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[7]]);
    
#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    
	    bits = pSel[j + 3];
	    maskBits = (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[3]]);

	    maskBits <<= 8;		/* ms bit to bit 15 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[2]]);

	    maskBits <<= 8;		/* ms bit to bit 23 */

	    bits = pSel[j + 2];
	    maskBits |= (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[1]]);

	    maskBits <<= 8;	/* ms bit to bit 31 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[0]]);

	    pDLong[0] = maskBits;

	    bits = pSel[j + 1];
	    maskBits = (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[7]]);

	    maskBits <<= 8;		/* ms bit to bit 15 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[6]]);

	    maskBits <<= 8;		/* ms bit to bit 23 */

	    bits = pSel[j];
	    maskBits |= (UInt32) 
		(backTab[((UInt32)(bits & 0xF0) << 4) |
			pBack[5]]);

	    maskBits <<= 8;	/* ms bit to bit 31 */
	    bits &= 0x0F;

	    maskBits |= (UInt32)
		(backTab[((UInt32)bits << 8) | pBack[4]]);
    
#endif
	    pDLong[1] = maskBits;
	}
    }   /* do more pixels if necessary */
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_gray4bppSelLineNoBG()
 *  		Combines a gray foreground and a virtual background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose foreground pixel.  A 0xF in the selector
 *		=> choose background pixel.  Other values result in weighted
 *		combinations of foreground and background.
 *		All pixels in the virtual background image are white (0xFF).
 *
 *	Input:
 *		pFore =   	Pointer to left end of foreground source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *
 **************************************************************************/
void  CDECL 
ip_gray4bppSelLineNoBG (
    IPState    *pIPState,
    UInt8      *pFore,
    UInt8      *pSel,
    UInt32     *pDLong,
    UInt32	width)
{
    PROLOGUE;
    Int32	 j;		/* iteration var */
    UInt32	 maskBits;
    UInt8	 bits;

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* the frames contain 32 gray pixels, so it doesn't
	 * matter if we do some extra so long as we clear
	 * them out.  Note that the frame pixels of both
	 * Back and Fore are zeros, so the combination is also 0.
	 * Therefore, any dest frame pixels will be written
	 * with 0's, which is fine.
	 * We process from right to left to save a little time
	 * testing j.
	 */
    for (j = ((width - 1)/2) & ~0x3,
	   pDLong += ((width - 1) & ~0x7) / 4,
	   pFore += (width - 1) & ~0x7;
	 j >= 0;
	 j -= 4, pDLong -= 2, pFore -= 8)
    {
	    /* if the next four selector bytes are filled with all
	     * 1's (remember that we invert the sense of gray
	     * selectors), just copy the next eight Back
	     * (background) bytes to the output.
	     */
	if (ULONG_ACCESS(&pSel[j]) == 0xFFFFFFFF)
	{
	    pDLong[0] = 0xFFFFFFFF;
	    pDLong[1] = 0xFFFFFFFF;
	}
	else
	{
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
    
		/* first, insert the background, which is
		 * really just an expanded version of the
		 * selector image.
		 */
	    maskBits = (expNib[pSel[j    ]] & 0xFFFF0000) |
		       (expNib[pSel[j + 1]] & 0x0000FFFF);

		/* Now add in the foreground pix.  We
		 * know that there will be no overflow.
		 */
	    bits = pSel[j + 0];
	    maskBits +=
		    ((UInt32)(foreTab[((UInt32)(bits & 0xF0) << 4) |
				     pFore[0]]) << 24) |
		    ((UInt32)(foreTab[((UInt32)(bits & 0x0F) << 8) |
				     pFore[1]]) << 16);

	    bits = pSel[j + 1];
	    maskBits +=
		    ((UInt32)(foreTab[((UInt32)(bits & 0xF0) << 4) |
				     pFore[2]]) << 8) |
		    ((UInt32)(foreTab[((UInt32)(bits & 0x0F) << 8) |
				     pFore[3]]));

	    pDLong[0] = maskBits;

		/* Insert background for next word */
	    maskBits = (expNib[pSel[j + 2]] & 0xFFFF0000) |
		       (expNib[pSel[j + 3]] & 0x0000FFFF);

		/* Insert 4 weighted foreground bytes */
	    bits = pSel[j + 2];
	    maskBits +=
		    ((UInt32)(foreTab[((UInt32)(bits & 0xF0) << 4) |
				     pFore[4]]) << 24) |
		    ((UInt32)(foreTab[((UInt32)(bits & 0x0F) << 8) |
				     pFore[5]]) << 16);

	    bits = pSel[j + 3];
	    maskBits +=
		    ((UInt32)(foreTab[((UInt32)(bits & 0xF0) << 4) |
				     pFore[6]]) << 8) |
		    ((UInt32)(foreTab[((UInt32)(bits & 0x0F) << 8) |
				     pFore[7]]));


#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    
		/* first, insert the background, which is
		 * really just an expanded version of the
		 * selector image.
		 */
	    maskBits = (expNib[pSel[j + 3]] & 0xFFFF0000) |
		       (expNib[pSel[j + 2]] & 0x0000FFFF);

		/* Now add in the foreground pix.  We
		 * know that there will be no overflow.
		 */
	    bits = pSel[j + 3];
	    maskBits +=
		    ((UInt32)(foreTab[((UInt32)(bits & 0xF0) << 4) |
				     pFore[3]]) << 24) |
		    ((UInt32)(foreTab[((UInt32)(bits & 0x0F) << 8) |
				     pFore[2]]) << 16);

	    bits = pSel[j + 2];
	    maskBits +=
		    ((UInt32)(foreTab[((UInt32)(bits & 0xF0) << 4) |
				     pFore[1]]) << 8) |
		    ((UInt32)(foreTab[((UInt32)(bits & 0x0F) << 8) |
				     pFore[0]]));

	    pDLong[0] = maskBits;

		/* Insert background for next word */
	    maskBits = (expNib[pSel[j + 1]] & 0xFFFF0000) |
		       (expNib[pSel[j + 0]] & 0x0000FFFF);

		/* Insert 4 weighted foreground bytes */
	    bits = pSel[j + 1];
	    maskBits +=
		    ((UInt32)(foreTab[((UInt32)(bits & 0xF0) << 4) |
				     pFore[7]]) << 24) |
		    ((UInt32)(foreTab[((UInt32)(bits & 0x0F) << 8) |
				     pFore[6]]) << 16);

	    bits = pSel[j + 0];
	    maskBits +=
		    ((UInt32)(foreTab[((UInt32)(bits & 0xF0) << 4) |
				     pFore[5]]) << 8) |
		    ((UInt32)(foreTab[((UInt32)(bits & 0x0F) << 8) |
				     pFore[4]]));

#endif
	    pDLong[1] = maskBits;
	}
    }   /* do more pixels if necessary */
}



/* PRIVATE */
/*************************************************************************** 
 *  ip_gray4bppSelLineNoFGNoBG()
 *  		Combines a virtual foreground and a virtual background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose foreground pixel.  A 0xF in the selector
 *		=> choose background pixel.  Other values result in weighted
 *		combinations of foreground and background.
 *		All pixels in the virtual foreground image are black (0x00).
 *		All pixels in the virtual background image are white (0xFF).
 *
 *	Input:
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *
 **************************************************************************/
void  CDECL 
ip_gray4bppSelLineNoFGNoBG (
    IPState    *pIPState,
    UInt8      *pSel,
    UInt32     *pDLong,
    UInt32	width)
{
    PROLOGUE;
    Int32	 j;		/* iteration var */

    /* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* the frames contain 32 gray pixels, so it doesn't
	 * matter if we do some extra so long as we clear
	 * them out.  Note that the frame pixels of both
	 * Back and Fore are zeros, so the combination is also 0.
	 * Therefore, any dest frame pixels will be written
	 * with 0's, which is fine.
	 * We process from right to left to save a little time
	 * testing j.
	 */
    for (j = ((width - 1)/2) & ~0x3,
	   pDLong += ((width - 1) & ~0x7) / 4;
	 j >= 0;
	 j -= 4, pDLong -= 2)
    {
	    /* if the next four selector bytes are filled with all
	     * 1's (remember that we invert the sense of gray
	     * selectors), just copy the next eight Back
	     * (background) bytes to the output.
	     */
	if (ULONG_ACCESS(&pSel[j]) == 0xFFFFFFFF)
	{
	    pDLong[0] = 0xFFFFFFFF;
	    pDLong[1] = 0xFFFFFFFF;
	}
	else
	{
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
    
		/* the only thing to do is insert the background, which is
		 * really just an expanded version of the
		 * selector image.
		 */
	    pDLong[0] = (expNib[pSel[j + 0]] & 0xFFFF0000) |
		        (expNib[pSel[j + 1]] & 0x0000FFFF);

		/* Insert background for next word */
	    pDLong[1] = (expNib[pSel[j + 2]] & 0xFFFF0000) |
		        (expNib[pSel[j + 3]] & 0x0000FFFF);


#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    
		/* first, insert the background, which is
		 * really just an expanded version of the
		 * selector image.
		 */
	    pDLong[0] = (expNib[pSel[j + 3]] & 0xFFFF0000) |
		        (expNib[pSel[j + 2]] & 0x0000FFFF);

		/* Insert background for next word */
	    pDLong[1] = (expNib[pSel[j + 1]] & 0xFFFF0000) |
		        (expNib[pSel[j + 0]] & 0x0000FFFF);

#endif
	}
    }   /* do more pixels if necessary */
}

/*

 $Log: migrybn2.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:52:28  ddavies
 migrybn2.c -> /product/ipcore/ipshared/src/RCS/migrybn2.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1997/06/06 04:19:49  ddavies
 Initial revision


 */



