/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	migsrot.c:
 *
 *          Machine independent Gray Scale rotation primitives,
 *		SMALL ANGLES ONLY! (< 10 degrees)
 *
 *          Low-level ops:
 *				ip_rotateGrayLine();
 *
 */


#include <stdio.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

    /*  prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: migsrot.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");
/*
 * This file contains an a machine independent version of ip_rotateGrayLine
 * as encoded in i_gsrfxp.c.  ip_rotateGrayLine computes one line of a
 * gray scale rotated image by interpolating pixels in an input image.
 *
 * To understand this glop, the reader should be familiar with the
 * voluminous comments in i_gsrfxp.c.
 *
 * The line is calculated as a series of 8 byte chunks.  For each chunk,
 * we take diagonal sweeps of input pixels (the sweeps are horizontal in
 * the input coordinate system - diagonal in the output coordinate system)
 * until we've covered the output, then write it out.  In each sweep, we
 * disperse each input pixel's value among the corresponding output pixels.
 *
 * There one subtle, but important, point to note.  i_rotateGrayLine passes all
 * coordinates down here as through we were doing clockwise rotation.  If it's
 * really counterclockwise, rSrcPitch is negative and we're starting from the
 * lower right corner of the image, but we don't need to know that here.
 * Therefore, when deciding whether a horizontal movement in the source
 * corresponds to a slight upward or downward movement in the destination, we
 * can always assume clockwise rotation (left->right in source => upper left
 * -> lower right in destination).
 *
 ******************* FIRST ASSEMBLY CODE REWORK ***************************
 *
 * This counts as the third rework of the algorithm.  The first two are
 * documented in the comments in i_gsrfxp.c.
 *
 * We're spending a lot of time in overhead.  Instead of doing two or more
 * complete passes through input lines to form an output line, say we do one
 * pass.  In that pass, we follow a jagged path through the input file such
 * that the coordinates of the input pixels in the output's frame of reference
 * stay immediately above the output line being calculated.  At each input
 * pixel position, we read and distribute not only the given input pixel, but
 * also the one immediately below it.
 *
 * Here's another way to look at it.  In the old scheme, we made as many
 * pure horizontal input sweeps as necessary to cover every output pixel with
 * a pair of vertically adjacent input pixels.  In this scheme, we do one
 * double-height sweep that jogs as necessary to keep it over the
 * output pixels.  The only way to make this work reasonably is to assume
 * the same xIOCoord for both upper and lower pixels - handling all the
 * conditions when the upper and lower pixels are in different output columns
 * gives me a headache.  This means we lose a little resolution - say 1/4 of
 * a pixel at 10 degrees, but tough.
 *
 * Without this simplification, there are two levels of loops - the outer loop
 * advances one iteration for each 8-pixel block of output pixels.  The inner
 * loop sweeps through as many input lines as necessary to calculate one block.
 * With the simplification, the inner loop is eliminated, though each step
 * takes longer.  What is the tradeoff here?
 *
 * The code in the old world that distributed the value of a pixel took 15
 * cycles.  We tended to execute a little over two of these 15 cycle blocks
 * of code for each output pixel, for an average of 34 cycles/pixel.  Overhead
 * required for the inner loop took about 20 cycles and was executed at least
 * once, often 2 or 3 times.  Say this adds 40 cycles/8 pixels for 5 cyc/pix.
 * This boosts the total to 39 cycles/pixel.
 *
 * In the new code, it takes about 20 cycles to distribute two pixels at a
 * given position.  We save on overhead by not jumping around so much.  There
 * is some overhead each time we move up or down by one input line, say this
 * happens 2 or 3 times/8 pixels.  The overhead is probably about 8 cycles,
 * or 16 cycles/8 pixels = 2 cyc/pix.  This brings the total to 22 cycles/pixel.
 * This is almost twice as fast as the old way.
 *
 * Since there is no inner loop, there's no switch statement required to
 * restart execution at random x offsets into the block.  Therefore, we don't
 * needs to spend so much time keeping xIOCoord up to date.  In fact, we
 * really only need the fraction bits to index into the table.  Ok, we also
 * need to know when adding sinRotAngle to xIOCoord crosses a pixel boundary
 * in the output world.  Huh?  Well, we normally traverse a horizontal path in
 * the input that's slightly diagonal in the output.  Since we want to stay
 * above the output line (the coordinates of the input pixels in the output's
 * frame of reference should be immediately above the output line), we detect
 * when the slanted path goes too low.  When we does, we move up one row in
 * the input world.  This corresponds to a vertical move by cosRotAngle
 * (approximated by 1 to avoid overlapping pixels) and a horizontal move by
 * sinRotAngle.  If adding sinRotAngle to xIOCoord crosses a pixel boundary in
 * the output world, we need to skip one output pixel position when
 * distributing input pixel values.
 * We'll just keep a -1 in the integer part of xIOCoord.
 * We can easily test to see if adding sinRotAngle forces xIOCoord to cross
 * an integer boundary by testing to see if the result is non-negative.  If we
 * cross an integer boundary, we restore the -1 in the integer, decrement
 * xCount, rSrcAddr and rSrcAddrBL, and skip one block of the inner loop.  What
 * are the extra decrements for?  Instead of incrementing the source address
 * and count at each block of inner loop, we use fixed address offsets.  For
 * example, at the second block of the loop, we address pixel rSrcAddr+2.  When
 * we skip one of these code blocks, we're also effectively skipping an address
 * offset.  In order to maintain our position in the input world, we decrement
 * the address pointers to compensate.
 *
 * The overhead encountered at the end of each block is also reduced because
 * we are already in at the correct input pixel at which to start processing
 * the next block.  The old scheme required pixel movement.
 *
 * There is a change in strategy, though.  Almost all output pixels contain
 * the coordinates of exactly one input pixel.  However, some output pixels
 * contain the coordinates of no input pixels - you can just barely miss at
 * all four corners.  Our scheme generally assumes that we start with the
 * input pixel whose output coords are in to the upper left of the leftmost
 * output pixel in the block.  Where do we start if no input pixel has those
 * coordinates?
 *
 * In the old scheme, we always started in the column to the left of the
 * output block, but could choose either the same row or the upper row (if the
 * same row wasn't reachable).  Here, we always want to start in the row above
 * the output row, either in the column to the left of the leftmost pixel or
 * the column of the first pixel if that isn't reachable.  Since the inner
 * loop code always reads two vertically adjacent pixels, it's important to
 * be in the row above the output line.
 *
 * One other modification - There was a special case to detect whether an
 * output block was entirely outside the image.  This did save some cycles
 * when it was used, totalling about 10 milliseconds for a 3 degree rotation
 * of a 8MB image.  However, there were an extra 6 opcodes to test for this
 * special case.  They were executed an 1,042,800 times for an addition
 * of 156 milliseconds.  Ooops.  The special case went away.
 *
 *
 * BUG FIX - Well, this is the first big post-rework bug.  As mentioned
 * above, a horizontal sweep across the input image is a diagonal sweep
 * (towards the lower right) across the output image.  When the sweep crosses
 * the boundary from the row above the output to the output's row, we need to
 * jump up one input line.  In doing so, we also move to the right by
 * sin(rotAngle) in the output's frame of reference.  So far, so good.  In
 * the new scheme however, we have been using the same xIOCoord for both the
 * upper and the lower input pixels.  This means the lower input pixel also
 * jumped to the right by sin(rotAngle).  This is a problem.  There will be
 * a discontinuity in the way that pixel's value is distributed among the
 * output pixels.  In effect, there will be a gap in the mapping of input
 * pixels to output pixels, leaving a couple of output pixels short changed.
 * The effect is visible as slightly darker, nearly vertical, lines down the
 * rotated image.
 *
 * This is a fairly ugly problem.  One approach is to use different xIOCoords
 * for the upper and lower samples.  This is fine except when they lie in
 * different output pixels.  To handle that, we'd have to shift the results
 * of the table lookups by to different output bytes.  This either requires
 * a messy test and branch in the inner loop (YUCK) or another set of checking
 * and non-checking loops with special shift cases.  Also yuck.
 *
 * Another approach is to try to fill the "gap" when we add do the little
 * horizontal jump.  This is a bit messy since we don't have a lookup table
 * that tells us how much of a pixel fits the gap.  It's also pretty slow to
 * do this gap filling every time we jump up a line in the input space.
 *
 * We`ll try a compromise.  If adding sin(rotAngle) to xIOCoord does not cross
 * into a new output column, we'll effectively use the old (does not reflect
 * the sin(rotAngle) jump) value of xIOCoord for the bottom sample.  We don't
 * actually keep two different xIOCoords, we just transfer the 3 xIOCoord
 * fraction bits from rTopLineTab to rBotLineTab.  This means the index into
 * the lookup table (which is really the only thing that matters here) for the
 * bottom input pixel will reflect the value of the un-shifted xIOCoord.
 *
 * What happens when adding sin(rotAngle) to xIOCoord causes it to cross into a
 * new pixel column?  That's when we do advance the bottom coordinate and fill
 * the gap.  How wide is the gap?  How much to we fill in?  Remember that
 * since we're jumping from one output pixel to the next, we need to add
 * something to *both* of them.  That's because the lookup tables assume that
 * an xIOCoord offset of 7 really means an offset of 15/16.  An offset of 0
 * really means an offset of 1/16.  Therefore, if we jump from 7 to 0, we
 * would need to add 1/16 of the pixel value to both the current output pixel
 * (to bring it up from 15/16 to 16/16) and the next output pixel (to bring
 * it from 0/16 to 1/16 where the next input pixel will start).
 *
 * But, we don't always start from an xIOCoord offset of 7 and, if we wanted
 * to be aligned with the top xIOCoord, we wouldn't always jump to an xIOCoord
 * offset of 0.  At the larger angles, xIOCoord can go from 6 to 0 or 7 to 1.
 * It is reasonably important to do the right thing so as to avoid the very
 * dark streaks we're trying to avoid.
 *
 * To make things deterministic, we'll always jump from wherever we would have
 * been normally (xIOCoord offset of 6 or 7) to offset 0.  Since the bottom
 * xIOCoord often trails the upper one anyway, it's not vital that they're
 * aligned after the jump.  We can get the amount of pixel to be added to
 * fill the gap in the output pixel we're leaving by simply doing a table
 * lookup using the rBotLineTab we would have used anyway with the proper
 * input pixel.  The leftmost byte of the two returned says how much of that
 * pixel lies in the current output pixel.  We get the amount of the input
 * pixel needed to fill the gap in the output pixel we're jumping to by simply
 * dividing the input pixel value by 16.  This is absolutely correct if the
 * yIOCoord offset was 0, which is almost always true.  If the yIOCoord offset
 * was 1, we'll be off by 1/256 of the pixel's value - not a big deal.
 *
 * In summary - each time the input jumps up a line, we transfer the old
 * xIOCoord fraction bits from rTopLineTab to rBotLineTab.  This is cheap and
 * fast.  If, in addition, the jog caused us to cross into a new output pixel
 * column, rBotLineTab is used to fetch the amount of the next pixel that will
 * fill the gap in the current output pixel, the gap in the next output pixel
 * is filled with 1/16th of the input pix's value and we advance the xIOCoord
 * fraction bits in rBotLineTab to 0.  You can think of this as skipping a
 * processing block, but filling in the gap created caused by skipping that
 * block.  This means we read the same pixel that block would have read and
 * apply the results to the output pixels that block would have affected.
 *
 * BUG FIX II - That was all lovely.  It went fast and didn't have ugly black
 * lines through the image.  Unfortunately, it had ugly faint gray bars though
 * the image in the same places.  My hypothesis is that these bars result from
 * small gaps in the mapping of input pixels to output pixels.  Look at this as
 * trying to tile the output image with diagonally-offset input pixels.  When
 * we jump up an input row, we have been jumping over a little in the output.
 * In the fix above, we closed the big gap in this jump by having one of the
 * lookup table pointer stay behind.  However, the upper pointer still jumped,
 * so there was still a small gap.
 *
 * So, what happens if we keep xIOCoord up to date on these vertical jogs, but
 * *DO NOT* update the X Fraction portion of either of the table pointers?
 * If there's a vertical jump, we may be as much as 1/4 of a pixel off?  How
 * often do we get vertical jumps?  Well, sin(10 degrees) = 0.17.  1/sin(10)
 * = 5.76, so we expect to cross to a new row of output bits every 5.76
 * horizontal input moves.  Since 5.76 is safely greater than 4, we'll take no
 * more than 1 vertical jump per 8 byte output block.  Therefore, we'll never
 * be more than 1/4 of a pixel off.  Seems ok to me.
 *
 * To make this work, we do calculate a new pair of table lookup indices at
 * the beginning of each block.  If xIOCoord has gone positive (the last
 * vertical input jump caused us to cross into a new output column), we back
 * up one pixel position.  If the backup causes us to move up too high, we
 * move down one row.
 *
 ******************************* ANISOTROPIC REWORK  ********************
 *
 * The code as is almost works for rotating anistropic images.  However,
 * to account for existing shears (and to maintain the proper new shears)
 * in anisotropic images, we need two values of sinRotAngle.  One is used
 * to adjust the X coord when moving vertically.  The other is used to
 * adjust the y coord when moving horizontally.  We had been only passing
 * in one value of sinRotAngle and using it for both.
 * Therefore, we add vertSinRotAngle and replace sinRotAngle with
 * horizSinRotAngle.
 *
 ****************************** FLAT IMAGE REWORK **********************
 *
 * The code was originally written assuming that the diagonal paths
 * traced could go from in the image to above the image, or from inside
 * the image to inside the image, or from below the image to inside
 * the image.  If you have a very broad, flat image, a path can go
 * from below the image, up through the image and out the top.  That
 * means that we must keep track of yCount.  When it crosses the boundary
 * from 7FFFFFFF down to what would be 7FFFFFFE, we set it to hS-2 instead.
 * Why not do the thing when it crosses from 0x80000000 to 0x7FFFFFFF?
 * Well, we check both yCount and yCount+1 to see if either one is
 * negative (>= 0x80000000).  If we set yCount equal to hS-1 when it
 * just crossed onto the bottom line of the image, we'd miss the fact
 * that yCount+1 wasn't on the image yet, so would fail.
 * Note that sometimes we have to increment yCount again when we've overshot.
 * In this case, we also have to check to see if yCount+1 already equals
 * hS-1.  If so, we set yCount to 0x7FFFFFFF instead of hS-1 so that
 * yCount+1 is properly negative.
 * However, we only need to do this in the loop that happens when we're
 * at or near the bottom.  When in the loop that's executed when we're
 * in the midst of the image (the normal case), we don't need to check
 * to see whether we're crossing the bottom edge of the image, so can just
 * increment and decrement yCount freely.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_rotateGrayLine()
 *		Generates one rotated line.
 *
 *	Input:
 *		pS =	pointer to beginning of next line of unrotated 
 *			gray data
 *		sBpl =	source bytes/line
 *		pD =	pointer to next line of rotated gray data
 *		w =	width of stipe in pixels
 *		vertSinRotAngle = used to adjust the X coord when moving
 *			vertically.
 *		horizSinRotAngle = used to adjust the Y coord when moving
 *			horizontally.
 *		xIOCoord = split integer: x position of current input
 *			pixel inoutput's frame of referecne
 *		yIOCoord = split integer: y position of current input
 *			pixel in output's frame of reference
 *		xCount = increment this count at each move towards the
 *			right.  While xCount < 0, we are off the image.
 *			distance to horizontal edge.
 *		yCount = increment this count at each move towards the
 *			bottom.  While yCount < 0, we are off the image.
 *			distance to vertical edge.
 *              pGrayRotationTable = table for gray rotation
 *		hS =	height of source image.  This is down here because
 *			it's a late addition.
 **************************************************************************/
void  CDECL 
ip_rotateGrayLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	w,
    Int32	vertSinRotAngle,
    Int32	horizSinRotAngle,
    Int32	xIOCoord,
    Int32	yIOCoord,
    UInt32	xCount,
    UInt32	yCount,
    UInt16     *pGrayRotationTable,
    Int32	hS)
{
    PROLOGUE;
    UInt32         gray0, gray1;	/* do 8 pixels at a time */
    UInt8         *pTopLineTab, *pBotLineTab;	/* pointers into
					 * grayRotationTable. */
    Int32	col;		/* iteration variables */
    Int32		xMask;		/* used to select 3 MS fraction
					 * bits from shifted xIOCoord */
    Int32		xInc;		/* used to increment integer
					 * part of xIOCoord */
    Int32		grayTemp;	/* holds value returned from
					 * gray rotate table */
    Int32		xVInc;		/* holds amount by which we
					 * change xIOCoord for vertical
					 * movement. */
	 
#define cXFMask		0x3800		/* selects 3 MS X frac bits */
#define cXIOCoordInc	0x4000		/* used to increment xIOCoord */
#define	cYFMask		0x700		/* selects 3 MS Y frac bits */
#define cYIOCoordInc		0x800		/* used to increment yIOCoord */
#define cMaxYFrac	0x7FF		/* maximum value of Y coord
					 * less than 1.0 */
#define cPixMask	0xFE		/* mask for raw pixel values */


	/* It would be polite to check for all sorts of error conditions
	 * here, but the assembly routine doesn't check so we won't either.
	 * Nonetheless, keep the compiler happy.
	 */
    pIPState = pIPState;
    __error__ = __error__;

	/* Just because it's important in what's coming up, let's set the
	 * format of the table lookup index:
	 * |   13   12   11    |  10     9      8  | 7 6 5 4 3 2 1 | 0 |
	 * | 3b xIOCoord (fff) | 3b yIOCoord (fff) |  7b pixel val | 0 |
	 *
	 * The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
	 * yIOCoord bits are its 3 MS fraction bits.  The 7 pixel bits are
	 * the 7 MS pixel bits.
	 * The 0 at the end makes this an index into a table of short ints.
	 */

	/* We'd like to have the three MS fraction bits in xIOC aligned with
	 * their position in the table lookup.
	 */
    xIOCoord = xIOCoord >> 2;
    xVInc = vertSinRotAngle >> 2;
    xMask = 0x3800;		/* used to select the 3 fraction bits */
    xInc = 0x4000;		/* used to increment integer part */

	/* Ditto for yIOCoord.  This leaves us with 11 bits of fraction.
	 * Knowing the Y position to within 1 part in 2048 is better accuracy
	 * than the skew detection code will return */
    yIOCoord = yIOCoord >> 5;

	/* The horizSinRotAngle quantity is added to yIOCoord in the inner loop.
	 * Align it with yIOCoord to make this easier. */
    horizSinRotAngle = horizSinRotAngle >> 5;

	/* we initialize the regs making up the gray box to 0 since no
	 * input pixels have been inserted yet.
	 */
    gray0 = gray1 = 0;

	/* Looks like we're going to have to start writing real inner loop
	 * code pretty soon.  Hard to put it off much longer.  But wait!
	 * We can still discuss how many flavors of loop we need.
	 * There are two types -  one checks xCount and yCount on
	 * every read from the input image because we're near a boundary, and
	 * one is used when we're totally inside the input image.
	 *
	 */
    for (col = 0; col < w; ) /* loop through dest cols */
    {

	/* which inner loop shall we use? */
	if (((yCount + 2)  >= (UInt32) hS) ||
	    ((yCount - 3)  >= (UInt32) hS) ||
	    ((xCount + 10) >= (UInt32) w) ||
	    ((xCount - 1)  >= (UInt32) w))
	{	/* at least part of the input is outside the input image */

		/* There are two types of inner-inner loop.  One starts in
		 * col 0 and one starts in col -1.  Both types end their
		 * sweeps when they reach col 7.  If start from col -1, we
		 * remember to skip the first block of the loop.
		 */

		/* The Y portion of rTopLineTab should be inverted so
		 * it references the portion of the pixel that lies
		 * in the *next* line.  The ! Y portion of rBotLineTab 
		 * should not be inverted so it get the portion of the 
		 * input pixel that lies in the current line.  We 
		 * implement the inversion by adding the maximum Y coord 
		 * (7) to pTopLineTab and subtracting each Y coord
		 * from this pointer.
		 */

		/* get ptr to portion of table that will be used for
		 * for this xIOCoord.
		 */
	    pBotLineTab = (UInt8 *) pGrayRotationTable +
			  (xIOCoord & xMask);
	    pTopLineTab = pBotLineTab + cYFMask;
	    if (xIOCoord >= 0)
	    {	/* backup one column */
		xCount--;
		pS--;
		xIOCoord -= xInc;
		yIOCoord -= horizSinRotAngle;
		    /* did that take us above row -1 */
		if ((yIOCoord + cYIOCoordInc) < 0)
		{	/* We went up too high, so move down one input line */
		    pS +=sBpl;
		    yCount++;
		    yIOCoord += cYIOCoordInc;
		    xIOCoord -= xVInc;
		}
	    }

		/*
		 * We're calculating pixel -1 relative to the output block.  
		 * Check to see if we're on the input image or not.
		 */
            if ((xCount >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask])) << 24;
	    }
	    else
	    {	/* top pixel is on the image */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[0]) & cPixMask)])) << 24;
	    }
            if ((xCount >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {	/* lower pixel is off the image so use max white value */
		gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask])) << 24;
	    }
	    else
	    {	/* lower pixel is on the image */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl]) & cPixMask)])) << 24;
	    }
	    if ( (yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 0 */
            if (((xCount + 1) >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
		gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask])) << 16;
	    }
	    else
	    {	/* top pixel is on the image */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[1]) & cPixMask)])) << 16;
	    }
            if (((xCount + 1) >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {
		gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask])) << 16;
	    }
	    else
	    {	/* lower pixel is on the image */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl+1]) & cPixMask)])) << 16;
	    }
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 1 */
            if (((xCount + 2) >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
		gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask])) << 8;
	    }
	    else
	    {	/* top pixel is on the image */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[2]) & cPixMask)])) << 8;
	    }
            if (((xCount + 2) >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {
		gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask])) << 8;
	    }
	    else
	    {	/* lower pixel is on the image */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl+2]) & cPixMask)])) << 8;
	    }
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 2 */
            if (((xCount + 3) >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
		gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask]));
	    }
	    else
	    {	/* top pixel is on the image */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[3]) & cPixMask)]));
	    }
            if (((xCount + 3) >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {
		gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask]));
	    }
	    else
	    {	/* lower pixel is on the image */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl+3]) & cPixMask)]));
	    }
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 3 */
            if (((xCount + 4) >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
	        grayTemp = (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask]));
	    }
	    else
	    {	/* top pixel is on the image */
	        grayTemp = (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[4]) & cPixMask)]));
	    }
	    gray0 += grayTemp >> 8;
	    gray1 += grayTemp << 24;
            if (((xCount + 4) >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {	/* lower pixel is off the image so use max white value */
		grayTemp = (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask]));

	    }
	    else
	    {	/* lower pixel is on the image */
	        grayTemp = (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl+4]) & cPixMask)]));
	    }
	    gray0 += grayTemp >> 8;
	    gray1 += grayTemp << 24;
	    if ( (yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 4 */
            if (((xCount + 5) >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
		gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask])) << 16;
	    }
	    else
	    {	/* top pixel is on the image */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[5]) & cPixMask)])) << 16;
	    }
            if (((xCount + 5) >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {
		gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask])) << 16;
	    }
	    else
	    {	/* lower pixel is on the image */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl+5]) & cPixMask)])) << 16;
	    }
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 5 */
            if (((xCount + 6) >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
		gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask])) << 8;
	    }
	    else
	    {	/* top pixel is on the image */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[6]) & cPixMask)])) << 8;
	    }
            if (((xCount + 6) >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {
		gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask])) << 8;
	    }
	    else
	    {	/* lower pixel is on the image */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl+6]) & cPixMask)])) << 8;
	    }
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 6 */
            if (((xCount + 7) >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
		gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask]));
	    }
	    else
	    {	/* top pixel is on the image */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[7]) & cPixMask)]));
	    }
            if (((xCount + 7) >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {
		gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask]));
	    }
	    else
	    {	/* lower pixel is on the image */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl+7]) & cPixMask)]));
	    }
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 7 */
            if (((xCount + 8) >= (UInt32) w) || (yCount >= (UInt32) hS))
	    {	/* top pixel is off the image so use max white value */
		gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
		    cPixMask])) >> 8;
	    }
	    else
	    {	/* top pixel is on the image */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) + 
	            (UCHAR_ACCESS(&pS[8]) & cPixMask)])) >> 8;
	    }
            if (((xCount + 8) >= (UInt32) w) || ((yCount + 1) >= (UInt32) hS))
	    {
		gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
		    cPixMask])) >> 8;
	    }
	    else
	    {	/* lower pixel is on the image */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	            (UCHAR_ACCESS(&pS[sBpl+8]) & cPixMask)])) >> 8;
	    }
		/* Note that we didn't try to advance the pointers.
		 * That's because we're already in the right position
		 * to start the next sweep.
		 */

	}
	else
	{
		/* This is the normal inner loop */
		/* All the comments for the checking loop apply here.
		 * The only difference is that we don't test xCount and
		 * yCount on each input sample.
		 */

		/* set up current table pointer */
	    pBotLineTab = (UInt8 *) pGrayRotationTable +
			  (xIOCoord & xMask);
	    pTopLineTab = pBotLineTab + cYFMask;

		/* Here's where we jump to the correct block of the
		 * checking loop.  the block is based on the integer 
		 * part of the xIOCoord since each block calculated a 
		 * byte.
		 * Note that we can just increment and decrement yCount
		 * freely without checking to see if we're near the bottom
		 * since we wouldn't be here if we were near the bottom.
		 */
	    if (xIOCoord >= 0)
	    {	/* we must be in column 0.  Back up 1 column */
		xCount--;
		pS--;
		xIOCoord -= xInc;
		yIOCoord -= horizSinRotAngle;
		if ((yIOCoord + cYIOCoordInc)  < 0)
		{	/* we are above row -1 */
		    pS += sBpl;
		    yCount++;
		    yIOCoord += cYIOCoordInc;
		    xIOCoord -= xVInc;
		}
	    }

		    /* column -1 */
	    gray0 += (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[0]) & cPixMask)])) << 24;
	    gray0 += (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl]) & cPixMask)])) << 24;
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		/* column 0 */
	    gray0 += (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[1]) & cPixMask)])) << 16;
	    gray0 += (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl+1]) & cPixMask)])) << 16;
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		    /* column 1 */
	    gray0 += (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[2]) & cPixMask)])) << 8;
	    gray0 += (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl+2]) & cPixMask)])) << 8;
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		    /* column 2 */
	    gray0 += (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[3]) & cPixMask)]));
	    gray0 += (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl+3]) & cPixMask)]));
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		    /* column 3 */
	    grayTemp = (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[4]) & cPixMask)]));
	    gray0 += grayTemp >> 8;
	    gray1 += grayTemp << 24;
	    grayTemp = (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl+4]) & cPixMask)]));
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }
	    gray0 += grayTemp >> 8;
	    gray1 += grayTemp << 24;

		    /* column 4 */
	    gray1 += (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[5]) & cPixMask)])) << 16;
	    gray1 += (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl+5]) & cPixMask)])) << 16;
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		    /* column 5 */
	    gray1 += (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[6]) & cPixMask)])) << 8;
	    gray1 += (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl+6]) & cPixMask)])) << 8;
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		    /* column 6 */
	    gray1 += (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[7]) & cPixMask)]));
	    gray1 += (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl+7]) & cPixMask)]));
	    if ((yIOCoord += horizSinRotAngle) >= 0)
	    {	/* entered row 0.  Move up one input row. */
		pS -= sBpl;
		yIOCoord -= cYIOCoordInc;
		xIOCoord += xVInc;
		yCount--;
	    }

		    /* column 7 */
	    gray1 += (UInt32) 
	      ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[8]) & cPixMask)])) >> 8;
	    gray1 += (UInt32) 
	      ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFMask) +
	        (UCHAR_ACCESS(&pS[sBpl+8]) & cPixMask)])) >> 8;

		/* Note that we dind't try to advance the pointers.
		 * They're already in the right position to start the
		 * next sweep.
		 */

	}

	    /* We've completed the formation of this row.  Write it
	     * out.  Note that we always write gray0.  If the line 
	     * ends after that, don't write gray1.
	     */
	ULONG_ACCESS(&(pD[col])) = gray0;
	if ((col += 4) < w)
	{
	    ULONG_ACCESS(&(pD[col])) = gray1;
	}
	col += 4;
	gray0 = gray1 = 0;
	pS += 8;
	xCount += 8;

    }	/* end of loop that processes all columns in the line */
}	/* end of ip_rotateGrayLine */

/*

$Log: migsrot.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.9  1998/01/13 23:52:31  ddavies
migsrot.c -> /product/ipcore/ipshared/src/RCS/migsrot.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1996/10/05 02:22:20  ddavies
migsrot.c -> /product/ipcore/ipshared/src/RCS/migsrot.c,v

Change the semantics of xCount and yCount.  Instead of trying to arrange
for them to be negative outside the image, we just to unsigned compares of
xCount >= width and yCount >= height.  If these are true, we're off the image.
Also get rid of the fixup of xCount at the left boundary.

Revision 1.7  1995/12/13 02:49:57  ejaquith
migsrot.c -> /product/ipcore/ipshared/src/RCS/migsrot.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/12/13  01:52:08  ddavies
 * migsrot.c -> /product/ipcore/ipshared/src/RCS/migsrot.c,v
 *
 * In keeping with the new, resolution-independent, method representing skew
 * angles, use two versions of sinRotAngle - horizSinRotAngle is used to adjust
 * y coords when moving horizontally and vertSinRotAngle is used to adjust
 * x coords when moving vertically.  Also fix a bug in which xCount was
 * being adjusted for use to detect the right edge of the image when it was
 * equal to 0.  In this case, it may still be necessary to read at pixel with
 * x coord = -1, but we wouldn't substitute a white pixel since xCount would
 * have been changed to be relative to the right edge of the image instead of
 * being relative to the left edge.
 *
 * Revision 1.5  1994/07/27  16:01:55  ejaquith
 * migsrot.c -> /product/ipcore/ipshared/src/RCS/migsrot.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 *
 * Revision 1.4  1994/02/25  02:32:35  dferrell
 * dferrell on Thu Feb 24 18:29:07 PST 1994
 *
 * migsrot.c -> /product/ipcore/ipshared/src/RCS/migsrot.c,v
 *
 * Changed parameters for ip_rotateGrayLine.
 * xCount and yCount are now UInt32 rather than Int32.  This allowed me to
 * fix a problem on the Mac with overflow/wrap-around detection of signed
 * integers.
 * Modified checking for being close to the edge (Yes!) of the source image
 * by looking at xCount and yCount.  We now compare them to a constant instead
 * of checking for an overflow in a signed int.
 *
 * Revision 1.3  1993/12/02  19:24:52  ddavies
 * migsrot.c -> /product/ipcore/ipshared/src/RCS/migsrot.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:49:44  ddavies
 * migsrot.c -> /product/ipcore/ipshared/src/RCS/migsrot.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

