/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * mihccorr.c:	    Hausdorff binary image comparator
 *  	    	    portion that actual takes two bitmaps and counts
 *  	    	    the overlap
 *
 *  Public routines:
 *
 *  [none, all routines called from other Hausdorff routines]
 *
 *
 * Extensive comments concerning the Hausdorff comparator method can be found
 * at the top of i_hc32ms.c.
 *
 */


#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"

    /* prototypes */
#include "shrpatt.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mihccorr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

static UInt32 bit32[32] = {
    0x80000000, 0x40000000, 0x20000000, 0x10000000,
    0x08000000, 0x04000000, 0x02000000, 0x01000000,
    0x00800000, 0x00400000, 0x00200000, 0x00100000,
    0x00080000, 0x00040000, 0x00020000, 0x00010000,
    0x00008000, 0x00004000, 0x00002000, 0x00001000,
    0x00000800, 0x00000400, 0x00000200, 0x00000100,
    0x00000080, 0x00000040, 0x00000020, 0x00000010,
    0x00000008, 0x00000004, 0x00000002, 0x00000001};

#ifdef pbmDEBUG
#include "memory.pub"
static UInt8 bit8[8] = {
    0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
#endif /*pbmDEBUG*/


/* PRIVATE */
/******************************************************************************
 * Counts the number of "on" overlapping bits in 2 images
 *
 * Description:
 *  Take a model in its point list representation and an index in this list
 *  pointing to the first overlapping row and take an imagery pointer to the
 *  imagery's first overlapping word, and AND the overlapping bits.
 *  Calculate the number of model bits ON in the region and the number of
 *  model bits that MISS (did not overlap an imagery bit).
 *
 * Returns:
 *  ia_successful
 *
 * Notes:
 *  maxMissCnt parameter:  If, while counting, the number of model bits that
 *  MISS exceeds this number, then terminate the correlation early.
 *  This number should be the number that is deemed least acceptable.
 *  (For example, if you want 90% of the bits in the whole model to overlap,
 *  then pass in 10% * the number of bits in the model).
 *****************************************************************************/
Int32 CDECL 
ip_correlatePtList (
    IPState    *pIPState,
    UInt32     *pI,		/* Ptr to first word of imagery overlap */
    Int32	heightOverlap,	   /* Number of mutual overlapping rows */
    Int32	wplI,		/* Words per line in imagery */
    Int32	x,		/* Relative location model is being placed
				 * over imagery in X-dimension */
    const UInt16 *pX,		/* Ptr to array of x offsets, the model bitmap
				 * in its point list representation */
    UInt32	index,		/* pX Index to first row in array that is in
				 * overlap region */
    UInt32	firstX,		/* x offset of first column in overlap */
    UInt32	lastX,		/* x offset of last column in overlap */
    Int32	maxMissCnt,	/* If the number of model bits that MISS
				 * exceeds this, terminate early */
    Int32      *pMissCnt)   	/* Return the number of total model bits
				 * that MISS imagery bits in the overlap
				 * (if > maxMissCnt, returns maxMissCnt+1) */
{
    PROLOGUE;
    UInt32   tx;
#ifdef pbmDEBUG
    UInt8   *MODELOVER, *pMODELOVER, *MODELAND, *pMODELAND;
    UInt32   MODELbpl;
    UInt32   htOverlap = heightOverlap;
    /* This is for the extra word that may appear with nonzero x
     * offsets that make shifts >= 32 */
    MODELbpl = (lastX - firstX + 1) / 8 +
	((lastX - firstX + 1) & 7 ? 1 : 0);
    pMODELOVER = MODELOVER = (UInt8 *)CALLOC( MODELbpl * htOverlap,
					      sizeof(UInt8));
    pMODELAND  = MODELAND  = (UInt8 *)CALLOC( MODELbpl * htOverlap,
					      sizeof(UInt8));
#endif /*pbmDEBUG*/

    *pMissCnt = maxMissCnt;
    
    for ( ; heightOverlap > 0; pI += wplI, heightOverlap--)
    {
	/* Go to the first overlapping model point in this row */
	while (pX[index] < firstX)
	    index++;

	/* Do the overlapping model points in this row */
	while (pX[index] <= lastX)
	{

	    tx = pX[index] + x;

	    if ((pI[tx >> 5] & bit32[tx & 31]) == 0)
		maxMissCnt--;

#ifdef pbmDEBUG
	    if ((pI[tx >> 5] & bit32[tx & 31]) != 0)
	    {
		tx = pX[index] - firstX /*+ (x & 7)*/;
		*(pMODELAND + (tx >> 3)) |= bit8[tx & 7];
	    }
	    else
		tx = pX[index] - firstX /*+ (x & 7)*/;
	    *(pMODELOVER + (tx >> 3)) |= bit8[tx & 7];
#endif /*pbmDEBUG*/
	    
	    index++;
	}

	/* Check for possilbe EARLY TERMINATION */
	if (maxMissCnt < 0)
	{
	    /* Too little points already, the model with NEVER match the
	     * imagery here */
	    (*pMissCnt)++;
	    NORMAL_RETURN;
	}

	/* Go to the end of this row */
	while (pX[index] != ROW_END)
	    index++;
	
	/* Advance to new model row */
	index++;
	
#ifdef pbmDEBUG
	pMODELOVER += MODELbpl;
	pMODELAND  += MODELbpl;
#endif /*pbmDEBUG*/
	
    } /* Done whole overlap */

#ifdef pbmDEBUG
    DumpPBM("HD_Model.over.pbm", MODELOVER, MODELbpl, lastX-firstX+1,
	    htOverlap);
    DumpPBM("HD_Model.and.pbm",  MODELAND,  MODELbpl, lastX-firstX+1,
	    htOverlap);
    FREE(MODELAND);
    FREE(MODELOVER);
#endif /*pbmDEBUG*/

    *pMissCnt -=maxMissCnt;
    NORMAL_RETURN;
}


/* PRIVATE */
/******************************************************************************
 * Counts the number of "on" overlapping bits in 2 images
 *
 * Description:
 *  Take a model in its point list representation and an index in this list
 *  pointing to the first overlapping row and take an imagery pointer to the
 *  imagery's first overlapping word, and AND the overlapping bits.
 *  Calculate the number of model bits ON in the region and the number of
 *  model bits that MISS (did not overlap an imagery bit).
 *
 * Returns:
 *  ia_successful
 *
 * Notes:
 *  maxMissCnt parameter:  If, while counting, the number of model bits that
 *  MISS exceeds this number, then terminate the correlation early.
 *  This number should be the number that is deemed least acceptable.
 *  (For example, if you want 90% of the bits in the whole model to overlap,
 *  then pass in 10% * the number of bits in the model).
 *****************************************************************************/
Int32 CDECL 
ip_correlateFullPtList (
    IPState    *pIPState,
    UInt32     *pI,		/* Ptr to first word of imagery overlap */
    Int32	heightM,	/* # model rows (excluding padding/frame)
				 * (also # of mutual overlapping rows) */
    Int32	wplI,		/* Words per line in imagery */
    Int32	x,		/* Relative location model is being placed
				 * over imagery in X-dimension */
    const UInt16 *pX,		/* Ptr to array of x offsets, the model bitmap
				 * in its point list representation */
    Int32	maxMissCnt,	/* If the number of model bits that MISS
				 * exceeds this, terminate early */
    Int32      *pMissCnt)   	/* Return the number of total model bits
				 * that MISS imagery bits in the overlap
				 * (if > maxMissCnt, returns maxMissCnt+1) */
{
    PROLOGUE;
    UInt32   tx, index = 0;

    /* This routine is just a simple variant of i_correlatePtList()
     * where we know the model lies entirely inside the imagery */

    *pMissCnt = maxMissCnt;
    
    for (--heightM; heightM >= 0; pI += wplI, heightM--)
    {
	if (pX[index] != ROW_END)
	{
	    /* Do the overlapping model points in this row */
	    do
	    {
		tx = pX[index] + x;
		
		if ((pI[tx >> 5] & bit32[tx & 31]) == 0)
		    maxMissCnt--;
		
		index++;
	    }
	    while (pX[index] != ROW_END);
	    
	    /* Check for possilbe EARLY TERMINATION */
	    if (maxMissCnt < 0)
	    {
		/* Too little points already, the model with NEVER match the
		 * imagery here */
		(*pMissCnt)++;
		NORMAL_RETURN;
	    }
	}

	/* Go to the next model row */
	index++;
    }

    *pMissCnt -=maxMissCnt;
    NORMAL_RETURN;
}


/* PRIVATE */
/******************************************************************************
 * Counts the number of "on" overlapping bits in 2 images
 *
 * Description:
 *  Take a model and an imagery pointer, each to their first overlapping word
 *  in the first overlapping image rows, and AND the overlapping word bits.
 *  Calculate the number of model bits ON in the region and the number of
 *  model bits that MISS (did not overlap an imagery bit).
 *
 * Returns:
 *  ia_successful
 *
 * Notes:
 *  maxMissCnt parameter:  If, while counting, the number of model bits that
 *  MISS exceeds this number, then terminate the correlation early.
 *  This number should be the number that is deemed least acceptable.
 *  (For example, if you want 90% of the bits in the whole model to overlap,
 *  then pass in 10% * the number of bits in the model).
 *****************************************************************************/
Int32 CDECL 
ip_correlateWithShifts (
    IPState    *pIPState,
    Int8	rshift,		  /* The number of bits to right shift the
				   * model before we overlap each model word */
    UInt32     *pI,		  /* Ptr to first word of imagery overlap */
    UInt32     *pM,		  /* Ptr to first word of model overlap */
    Int32	heightOverlap,	  /* Number of mutual overlapping rows */
    Int32	wplOverlap,	  /* Number of mutual overlapping words per
				   * row, minus one */
    UInt32	rowLastWordMask,  /* A left mask for the last overlapping word
				   * of each model row, to chop off excess
				   * bits that don't really overlap the
				   * imagery*/
    UInt32	wplI,		  /* # words per line,including padding/frame*/
    UInt32	wplM,		  /* # words per line,including padding/frame*/
    const UInt8 *pBitCount,	  /* Ptr to bit counting table, containing
				   * number of 1's in a byte */
    Int32	maxMissCnt,	  /* If the number of model bits that MISS
				   * exceeds this, terminate early */
    Int32      *pMissCnt)         /* Return the number of total model bits
				   * that MISS imagery bits in the overlap
				   * (if > maxMissCnt, returns maxMissCnt+1) */
{
    PROLOGUE;
    UInt32    word;
    Int8      lshift;
    Int32     column;
#ifdef pbmDEBUG
    UInt8   *MODELAND, *pMODELAND, *MODELOVER, *pMODELOVER;
    Int32    MODELbpl = (wplOverlap+1) << 2;
    Int32    htOverlap = heightOverlap;
    /* Make an array the size of the overlap */
    pMODELAND  = MODELAND  = (UInt8 *)CALLOC(MODELbpl * htOverlap,
					     sizeof(UInt8));
    pMODELOVER = MODELOVER = (UInt8 *)CALLOC(MODELbpl * htOverlap,
					     sizeof(UInt8));
#endif /*pbmDEBUG*/

    lshift = WORD_SIZE - rshift;
    *pMissCnt = maxMissCnt;

    for ( ; heightOverlap > 0; pM += wplM, pI += wplI, heightOverlap--)
    {
	/* AND the new overlapping row */
	for (column = 0; column < wplOverlap; column++) 
	{
	    /* Count the number of model bits that MISS */
	    word = ((pM[column] >> rshift) | (pM[column-1] << lshift))
		& ~(pI[column]);
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word];

#ifdef pbmDEBUG
	    /* Fill in the array with the MODEL bytes */
	    word = (pM[column] >> rshift) | (pM[column-1] << lshift);
	    *(pMODELOVER++) = (UInt8)(word >> 24);
	    *(pMODELOVER++) = (UInt8)(word >> 16);
	    *(pMODELOVER++) = (UInt8)(word >> 8 );
	    *(pMODELOVER++) = (UInt8) word;

	    /* Fill in the array with the ANDed bytes */
	    word &= pI[column];
	    *(pMODELAND++) = (UInt8)(word >> 24);
	    *(pMODELAND++) = (UInt8)(word >> 16);
	    *(pMODELAND++) = (UInt8)(word >> 8 );
	    *(pMODELAND++) = (UInt8) word;
#endif /* pbmDEBUG */	    
	}

	/* AND the (special) last overlapping word in the row */
	/* Count the number of model bits that MISS */
	word = ((pM[column] >> rshift) | (pM[column-1] << lshift))
	    & rowLastWordMask & ~(pI[column]);
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word];

#ifdef pbmDEBUG
	/* Fill in the array with the MODEL bytes */
	word = ((pM[column] >> rshift) | (pM[column] << lshift))
	    & rowLastWordMask; 
	*(pMODELOVER++) = (UInt8)(word >> 24);
	*(pMODELOVER++) = (UInt8)(word >> 16);
	*(pMODELOVER++) = (UInt8)(word >> 8 );
	*(pMODELOVER++) = (UInt8) word;

	/* Fill in the array with the ANDed bytes */
	word &= pI[column];
	*(pMODELAND++) = (UInt8)(word >> 24);
	*(pMODELAND++) = (UInt8)(word >> 16);
	*(pMODELAND++) = (UInt8)(word >> 8 );
	*(pMODELAND++) = (UInt8) word;
#endif /* pbmDEBUG */
	
	/* Done another overlapping row */
	
	if (maxMissCnt < 0)
	{
	    /* EARLY TERMINATION:  Too little bits overlap, the model will
	     * NEVER match the imagery overlapping here */
	    (*pMissCnt)++;
	    NORMAL_RETURN;
	}
    }

#ifdef pbmDEBUG
    DumpPBM("HD_Model.over",MODELOVER,MODELbpl,MODELbpl*8,htOverlap);
    DumpPBM("HD_Model.and", MODELAND, MODELbpl,MODELbpl*8,htOverlap);
    FREE(MODELOVER); FREE(MODELAND);
#endif /* pbmDEBUG */

    *pMissCnt -= maxMissCnt;
    NORMAL_RETURN;
}


/* PRIVATE */
/******************************************************************************
 * Counts the number of "on" overlapping bits in 2 images
 *
 * Description:
 *  Take a model and an imagery pointer, each to their first overlapping word
 *  in the first overlapping image rows, and AND the overlapping word bits.
 *  Calculate the number of model bits ON in the region and the number of
 *  model bits that MISS (did not overlap an imagery bit).
 *
 * Returns:
 *  ia_successful
 *
 * Notes:
 *  maxMissCnt parameter:  If, while counting, the number of model bits that
 *  MISS exceeds this number, then terminate the correlation early.
 *  This number should be the number that is deemed least acceptable.
 *  (For example, if you want 90% of the bits in the whole model to overlap,
 *  then pass in 10% * the number of bits in the model).
 *****************************************************************************/
Int32 CDECL 
ip_correlateOrderedWithShifts (
    IPState    *pIPState,
    Int8	rshift,		  /* The number of bits to right shift the
				   * model before we overlap each model word */
    UInt32     *pI,		  /* Ptr to first word of imagery overlap */
    UInt32     *pM,		  /* Ptr to first word of model overlap */
    Int32	heightM,	  /* # model rows (excluding padding/frame)
				   * (also # of mutual overlapping rows) */
    Int32	wplOverlap,	  /* Number of mutual overlapping words per
				   * row, minus one */
    UInt32	rowLastWordMask,  /* A left mask for the last overlapping word
				   * of each model row, to chop off excess
				   * bits that don't really overlap the
				   * imagery*/
    Int32      *pOffsetsI,	  /* Ptr to imagery's ordered offsets table */
    Int32      *pOffsetsM,	  /* Ptr to model's ordered offsets table */
    const UInt8 *pBitCount,	  /* Ptr to bit counting table, containing
				   * number of 1's in a byte */
    Int32	maxMissCnt,	  /* If the number of model bits that MISS
				   * exceeds this, terminate early */
    Int32      *pMissCnt)         /* Return the number of total model bits
				   * that MISS imagery bits in the overlap
				   * (if > maxMissCnt, returns maxMissCnt+1) */
{
    PROLOGUE;
    UInt32    word;
    Int8      lshift;
    Int32     column;

    lshift = WORD_SIZE - rshift;
    *pMissCnt = maxMissCnt;

    for (--heightM; heightM >= 0; heightM--)
    {
	/* Move the model and imagery pointers to the new rows to compare */
	column = pOffsetsM[heightM]; /* Temporarily use */
	if (column == 0 /*Zero row flag*/ )
	    /* This means that this row in the model was a zero row (no ON
	     * bits) and thus we can skip over the comparison!
	     */
	    continue;
	
	pM += column;
	pI += pOffsetsI[heightM];
	
	/* AND the new overlapping row */
	for (column = 0; column < wplOverlap; column++) 
	{
	    /* Count the number of model bits that MISS */
	    word = ((pM[column] >> rshift) | (pM[column-1] << lshift))
		& ~(pI[column]);
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word];

	}

	/* AND the (special) last overlapping word in the row */
	/* Count the number of model bits that MISS */
	word = ((pM[column] >> rshift) | (pM[column-1] << lshift))
	    & rowLastWordMask & ~(pI[column]);
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word];

	
	/* Done another overlapping row */
	
	if (maxMissCnt < 0)
	{
	    /* EARLY TERMINATION:  Too little bits overlap, the model will
	     * NEVER match the imagery overlapping here */
	    (*pMissCnt)++;
	    NORMAL_RETURN;
	}
    }

    *pMissCnt -= maxMissCnt;
    NORMAL_RETURN;
}


/* PRIVATE */
/******************************************************************************
 * Counts the number of "on" overlapping bits in 2 images
 *
 * Description:
 *  Take a model and an imagery pointer, each to their first overlapping word
 *  in the first overlapping image rows, and AND the overlapping word bits.
 *  Calculate the number of model bits ON in the region and the number of
 *  model bits that MISS (did not overlap an imagery bit).
 *
 * Returns:
 *  ia_successful
 *
 * Notes:
 *  maxMissCnt parameter:  If, while counting, the number of model bits that
 *  MISS exceeds this number, then terminate the correlation early.
 *  This number should be the number that is deemed least acceptable.
 *  (For example, if you want 90% of the bits in the whole model to overlap,
 *  then pass in 10% * the number of bits in the model).
 *****************************************************************************/
Int32 CDECL 
ip_correlate (
    IPState    *pIPState,
    UInt32     *pI,		  /* Ptr to first word of imagery overlap */
    UInt32     *pM,		  /* Ptr to first word of model overlap */
    Int32	heightOverlap,	  /* Number of mutual overlapping rows */
    Int32	wplOverlap,	  /* Number of mutual overlapping words per
				   * row, minus one */
    UInt32	rowLastWordMask,  /* A left mask for the last overlapping word
				   * of each model row, to chop off excess
				   * bits that don't really overlap the
				   * imagery*/
    UInt32	wplI,		  /* # words per line,including padding/frame*/
    UInt32	wplM,		  /* # words per line,including padding/frame*/
    const UInt8 *pBitCount,	  /* Ptr to bit counting table, containing
				   * number of 1's in a byte */
    Int32	maxMissCnt,	  /* If the number of model bits that MISS
				   * exceeds this, terminate early */
    Int32      *pMissCnt)         /* Return the number of total model bits
				   * that MISS imagery bits in the overlap
				   * (if > maxMissCnt, returns maxMissCnt+1) */
{
    PROLOGUE;
    UInt32    word;
    Int32     column;
#ifdef pbmDEBUG
    UInt8   *MODELAND, *pMODELAND, *MODELOVER, *pMODELOVER;
    Int32    MODELbpl = (wplOverlap+1) << 2;
    Int32    htOverlap = heightOverlap;
    /* Make an array the size of the overlap */
    pMODELAND  = MODELAND  = (UInt8 *)CALLOC(MODELbpl * htOverlap,
					     sizeof(UInt8));
    pMODELOVER = MODELOVER = (UInt8 *)CALLOC(MODELbpl * htOverlap,
					     sizeof(UInt8));
#endif /*pbmDEBUG*/

    *pMissCnt = maxMissCnt;

    for ( ; heightOverlap > 0; pM += wplM, pI += wplI, heightOverlap--)
    {
	/* AND the new overlapping row */
	for (column = 0; column < wplOverlap; column++)
	{
	    /* Count the number of model bits that MISS */
	    word = pM[column] & ~(pI[column]);
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word];

#ifdef pbmDEBUG
	    /* Fill in the array with the ANDed bytes */
	    word = pM[column] & pI[column]; 
	    *(pMODELAND++) = (UInt8)(word >> 24);
	    *(pMODELAND++) = (UInt8)(word >> 16);
	    *(pMODELAND++) = (UInt8)(word >> 8 );
	    *(pMODELAND++) = (UInt8) word;

	    /* Fill in the array with the MODEL bytes */
	    word = pM[column];
	    *(pMODELOVER++) = (UInt8)(word >> 24);
	    *(pMODELOVER++) = (UInt8)(word >> 16);
	    *(pMODELOVER++) = (UInt8)(word >> 8 );
	    *(pMODELOVER++) = (UInt8) word;
#endif /* pbmDEBUG */	    

	}

	/* AND the (special) last overlapping word in the row */
	/* Count the number of model bits that MISS */
	word = pM[column] & ~(pI[column]) & rowLastWordMask;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word];

#ifdef pbmDEBUG
	/* Fill in the array with the ANDed bytes */
	word = pM[column] & pI[column];
	*(pMODELAND++) = (UInt8)(word >> 24);
	*(pMODELAND++) = (UInt8)(word >> 16);
	*(pMODELAND++) = (UInt8)(word >> 8 );
	*(pMODELAND++) = (UInt8) word;

	/* Fill in the array with the MODEL bytes */
	word = pM[column] & rowLastWordMask; 
	*(pMODELOVER++) = (UInt8)(word >> 24);
	*(pMODELOVER++) = (UInt8)(word >> 16);
	*(pMODELOVER++) = (UInt8)(word >> 8 );
	*(pMODELOVER++) = (UInt8) word;
#endif /* pbmDEBUG */
	
	/* Done another overlapping row */
	
	if (maxMissCnt < 0)
	{
	    /* EARLY TERMINATION:  Too little bits overlap, the model will
	     * NEVER match the imagery overlapping here */
	    (*pMissCnt)++;
	    NORMAL_RETURN;
	}
    }

#ifdef pbmDEBUG
    DumpPBM("HD_Model.over",MODELOVER,MODELbpl,MODELbpl*8,htOverlap);
    DumpPBM("HD_Model.and", MODELAND, MODELbpl,MODELbpl*8,htOverlap);
    FREE(MODELOVER); FREE(MODELAND);
#endif /* pbmDEBUG */

    *pMissCnt -= maxMissCnt;
    NORMAL_RETURN;
}


/* PRIVATE */
/******************************************************************************
 * Counts the number of "on" overlapping bits in 2 images
 *
 * Description:
 *  Take a model and an imagery pointer, each to their first overlapping word
 *  in the first overlapping image rows, and AND the overlapping word bits.
 *  Calculate the number of model bits ON in the region and the number of
 *  model bits that MISS (did not overlap an imagery bit).
 *
 * Returns:
 *  ia_successful
 *
 * Notes:
 *  maxMissCnt parameter:  If, while counting, the number of model bits that
 *  MISS exceeds this number, then terminate the correlation early.
 *  This number should be the number that is deemed least acceptable.
 *  (For example, if you want 90% of the bits in the whole model to overlap,
 *  then pass in 10% * the number of bits in the model).
 *****************************************************************************/
Int32 CDECL 
ip_correlateOrdered (
    IPState    *pIPState,
    UInt32     *pI,		  /* Ptr to first word of imagery overlap */
    UInt32     *pM,		  /* Ptr to first word of model overlap */
    Int32	heightM,	  /* # model rows (excluding padding/frame)
				   * (also # of mutual overlapping rows) */
    Int32	wplOverlap,	  /* Number of mutual overlapping words per
				   * row, minus one */
    UInt32	rowLastWordMask,  /* A left mask for the last overlapping word
				   * of each model row, to chop off excess
				   * bits that don't really overlap the
				   * imagery*/
    Int32      *pOffsetsI,	  /* Ptr to imagery's ordered offsets table */
    Int32      *pOffsetsM,	  /* Ptr to model's ordered offsets table */
    const UInt8 *pBitCount,	  /* Ptr to bit counting table, containing
				   * number of 1's in a byte */
    Int32	maxMissCnt,	  /* If the number of model bits that MISS
				   * exceeds this, terminate early */
    Int32      *pMissCnt)         /* Return the number of total model bits
				   * that MISS imagery bits in the overlap
				   * (if > maxMissCnt, returns maxMissCnt+1) */
{
    PROLOGUE;
    UInt32    word;
    Int32     column;
    
    *pMissCnt = maxMissCnt;

    for (--heightM; heightM >= 0; heightM--)
    {
	/* Move the model and imagery pointers to the new rows to compare */
	column = pOffsetsM[heightM]; /* Temporarily use */
	if (column == 0 /*Zero row flag*/ )
	    /* This means that this row in the model was a zero row (no ON
	     * bits) and thus we can skip over the comparison!
	     */
	    continue;
	
	pM += column;
	pI += pOffsetsI[heightM];
	
	/* AND the new overlapping row */
	for (column = 0; column < wplOverlap; column++)
	{
	    /* Count the number of model bits that MISS */
	    word = pM[column] & ~(pI[column]);
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word & 0xFF];
	    word >>= 8;
	    maxMissCnt -= pBitCount[word];
	}

	/* AND the (special) last overlapping word in the row */
	/* Count the number of model bits that MISS */
	word = pM[column] & ~(pI[column]) & rowLastWordMask;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word & 0xFF];
	word >>= 8;
	maxMissCnt -= pBitCount[word];
	
	if (maxMissCnt < 0)
	{
	    /* EARLY TERMINATION:  Too little bits overlap, the model will
	     * NEVER match the imagery overlapping here */
	    (*pMissCnt)++;
	    NORMAL_RETURN;
	}
    }

    *pMissCnt -= maxMissCnt;
    NORMAL_RETURN;
}


/*

$Log: mihccorr.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.6  1998/01/13 23:52:35  ddavies
mihccorr.c -> /product/ipcore/ipshared/src/RCS/mihccorr.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1995/12/13 02:50:12  ejaquith
mihccorr.c -> /product/ipcore/ipshared/src/RCS/mihccorr.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1995/09/14  18:59:57  ejaquith
 * mihccorr.c -> /product/ipcore/ipshared/src/RCS/mihccorr.c,v
 *
 * CommonCheck: Converted all routines that were private in files to be of form
 * ip_*(), which they weren't.  No functionality/noticeable changes.
 *
 * Revision 1.3  1995/03/15  01:25:40  ejaquith
 * mihccorr.c -> /product/ipcore/ipshared/src/RCS/mihccorr.c,v
 *
 * Added routine i_correlateFullPtList() to quickly do correlation when
 * entire model overlays the imagery.
 * Added i_correlateOrderedWithShifts(
 * and i_correlateOrdered() routines to access the model's bitmap
 * non-sequentially in a previously decided faster order.  The order
 * is based on "densest model rows compared first" so that early
 * termination will occur much faster than previously did.
 * Modified all i_correlate**() routines to use array indexed
 * image references (instead of moving a single pointer).
 * This allows a bit of a speed up from:
 *  4 extra adds per invocation and 1 extra mult per invocation
 *  4 adds per row
 *  3 adds per column
 * to:
 *  4 adds per row
 *  2 adds per column
 *
 * Revision 1.2  1995/01/04  18:51:35  ejaquith
 * mihccorr.c -> /product/ipcore/ipshared/src/RCS/mihccorr.c,v
 *
 * Abandoned use of "unit" (variable sized word for storage of models) for
 * se of 32 bit words.  Sped up the three correlation routines by figuring
 * way to get away with only one bit counter throughout matching.  The routines
 * therefore now return the number of MISSes (the number of model points that
 * do not overlap imagery points), and need an acceptable maximum miss count
 * to be passed in.  This also killed a subtraction at the end of every
 * overlapping row.  Made bit counting safer.  Much faster!
 *
 * Revision 1.1  1994/12/06  23:34:08  ejaquith
 * Initial revision
 *

*/
