/*************************************************************
 *  Copyright (c) 1992,1993, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * mihisto.c -- C implementations of lower-level histogram function:
 *           Int32     ip_calcFullHistogram()		
 */


#include <stdio.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /* prototypes */
#include "shrip.pub"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mihisto.c,v $");
/* default: suppress all debugging */
#define  DEBUG_HISTO     0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  DEBUG_HISTO     1
#endif  /* DEBUG_ON */

#define cByteMask	0xFF
#define cNibbleMask	0x0F

/*-----------------------------------------------------------------------*
 *                         Grayscale Histogram                           *
 *-----------------------------------------------------------------------*/



/**************************************************************************
 *  ip_calcHistogramLine()
 *		Takes an 8-bit grayscale image and returns
 *             	a histogram of the gray image in a 256 element array.
 *		If there are n gray pixels of value i, histo[i] will be set
 *		to n.
 *	Input:
 *		pS =   		Pointer to upper left corner of gray image
 *				  INSIDE frame.  This pointer may point
 *				  to bytes that are not a multiple of 4.
 *      	w =		Width of dest image in BYTES excluding frame
 *		histo =		Pointer to table containing 256 Int32's.
 *
 **************************************************************************/
void  CDECL 
ip_calcHistogramLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32      *histo)
{
    PROLOGUE;
    UInt8      *pSLine;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Do a wad of bytes in a row for efficiency.  Start from
	 * the right side of the row so we can work back to
	 * pS.  NOTE THAT pS IS NOT NECESSARILY ON A MULTIPLE OF 4 BYTES!!!
	 *
	 * For the sake of efficiency, we'll do 8 bytes in a block
	 * inside the loop.  To make addressing easy on both
	 * little endian and big endian platforms, we want to
	 * do things in units of words.  That means, we go backwards
	 * from the right end until we get to a word boundary first.
	 */
    pSLine = pS + (w - 1);		/* address the rightmost byte */
    while ((((UInt32)pSLine & 0x3) != 3) && (pSLine >= pS))
    {
	histo[UCHAR_ACCESS(&pSLine[0])]++;
	pSLine--;
    }

	/* Normally, at the top of this loop, pSLine is on a word boundary.
	 * Initially, it points to the rightmost byte (byte 3) of a word.
	 * Since we like to use positive address offsets
	 * (machines tend to like them better), we skip back 8 bytes
	 * )7 bytes initially to get on a word boundary)
	 * before procesing the next 8 bytes in two words.
	 * If skipping back causes us to go below pS, we process the
	 * bytes on the left end of the scan line up to pS.
	 * Note that we don't need to use UCHAR_ACCESS here because
	 * we're going to look at all of the bytes in both words -
	 * the order in which we touch them doesn't matter.  In fact,
	 * That's the whole point of getting on a word boundary.
	 */
    for (pSLine -= 7; pSLine >= pS; pSLine -= 8)
    {
	histo[pSLine[7]]++;
	histo[pSLine[6]]++;
	histo[pSLine[5]]++;
	histo[pSLine[4]]++;
	histo[pSLine[3]]++;
	histo[pSLine[2]]++;
	histo[pSLine[1]]++;
	histo[pSLine[0]]++;
    }

	/* Do the bytes on the left end of the line, if any */
    for (pSLine += 7; pSLine >= pS; pSLine--)
	histo[UCHAR_ACCESS(&pSLine[0])]++;

}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute histogram for one line of image pixels covered by a mask.
 *
 * LongDescription:::
 *   Given an 8bpp gray image and a mask, histogram one line of the 
 *   image which have a corresponding "1" in the mask.  The mask
 *   does not need to start on a 32-bit boundary (it starts at firstBit).
 *
 * Input:::
 *  pS::		Source image, pointing to start of rectangle.
 *  pMLong::		Mask image, pointing to start of rectangle
 *  firstBit::		Bit number of first bit of rectangle (in mask image)
 *  firstWordMask::	32-bit mask used to remove bits (from mask) which 
 *			should not be histogrammed from the first mask
 *			word
 *  numFirstWordBits::  Number of bits to be accessed in first word of
 *			mask
 *  lastWordMask::	32-bit mask used to remove bits (from mask) which 
 *			should not be histogrammed from the last mask
 *			word
 *  w::			Number of pixels to be histogrammed
 *
 * Output:::
 *  pHisto::		The histogram
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  JCraig
 *
 * EndDocumentation:::
 **************************************************************************/

void  CDECL 
ip_calcHistoMaskLine8 (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32     *pMLong,
    Int32	firstBit,
    UInt32	firstWordMask,
    UInt32	numFirstWordBits,
    UInt32	lastWordMask,
    Int32	w,
    Int32      *pHisto)
{
    PROLOGUE;
    Int32	j, k;
    UInt32	bits, bitMask;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* What we want to do is get longwords out of the mask, so that
     * we read it efficiently, and so that if we see all 0 we can skip
     * 32 pixels at a time.
     */

    /* is the first word a full word?  if not, it must be done separately.
     * otherwise just let the next loop handle it since it is more 
     * efficient.
     */
    if (numFirstWordBits != 32)
    {
        bits = pMLong[0] & firstWordMask;
        pMLong++;
        if (bits != 0)
        {
	    bitMask = 1UL << (32 - firstBit - 1);
	    for (k = numFirstWordBits; k > 0; k--, pS++, bitMask >>= 1)
	        if (bits & bitMask)
		    pHisto[UCHAR_ACCESS(pS)]++;
        }

	/* update the width for the number of bits done in the first word */
    	w -= numFirstWordBits;
    }

    /* now do all the full 32 bit mask words */
    for (j = 0; j < (w / 32); j++)
    {
	bits = pMLong[0];
	pMLong++;
	if (bits == 0)
	{
	    /* Nothing here matters; keep going */
	    pS += 32;
	    continue;
	}

	if (bits == 0xffffffff)
	{
	    /* Special case: all ones, so histogram them all */
	    /* note that byteswapping doesn't matter here since
	     * we're looking at all the pixels and order doesn't
	     * matter.
	     */
	    pHisto[*pS]++;
	    pHisto[*(pS+1)]++;
	    pHisto[*(pS+2)]++;
	    pHisto[*(pS+3)]++;
	    pHisto[*(pS+4)]++;
	    pHisto[*(pS+5)]++;
	    pHisto[*(pS+6)]++;
	    pHisto[*(pS+7)]++;
	    pHisto[*(pS+8)]++;
	    pHisto[*(pS+9)]++;
	    pHisto[*(pS+10)]++;
	    pHisto[*(pS+11)]++;
	    pHisto[*(pS+12)]++;
	    pHisto[*(pS+13)]++;
	    pHisto[*(pS+14)]++;
	    pHisto[*(pS+15)]++;
	    pHisto[*(pS+16)]++;
	    pHisto[*(pS+17)]++;
	    pHisto[*(pS+18)]++;
	    pHisto[*(pS+19)]++;
	    pHisto[*(pS+20)]++;
	    pHisto[*(pS+21)]++;
	    pHisto[*(pS+22)]++;
	    pHisto[*(pS+23)]++;
	    pHisto[*(pS+24)]++;
	    pHisto[*(pS+25)]++;
	    pHisto[*(pS+26)]++;
	    pHisto[*(pS+27)]++;
	    pHisto[*(pS+28)]++;
	    pHisto[*(pS+29)]++;
	    pHisto[*(pS+30)]++;
	    pHisto[*(pS+31)]++;
	    pS += 32;
	}
	else
	{
	    /* We look at each bit in the mask to
	     * determine whether to histogram a pixel in the image
	     */
	    bitMask = 1UL << 31;
	    for (k = 0; k < 32; k++, pS++)
	    {
		if (bits & bitMask)
		    pHisto[UCHAR_ACCESS(pS)]++;
		bitMask >>= 1;
	    }
	}
    }

    /* now do the last partial word. 
     * why check for lastWordMask being 0?  well, if the rectangle ends
     * at the right edge of the mask image, then accessing pMLong would
     * access the frame.  not such a big deal but still no need to 
     * reference it if we don't have to.   in addition, the lastWordMask
     * may be 0, if we end on a word boundary.  again, we could ignore
     * this check and just refrence one too many words but just to be a
     * good citizen, check for this case.  note lastWordMask may be 0 
     * either because we ended on a word boundary OR because the we
     * only reference one source word (the rectangle starts and ends
     * in the same word.
     */
    if (lastWordMask != 0)
    {
        bits = pMLong[0] & lastWordMask;
        if (bits != 0)
        {
	    bitMask = 1UL << 31;
	    for (k = 0; k < (w & 0x1F); k++, pS++, bitMask >>= 1)
	        if (bits & bitMask)
		    pHisto[UCHAR_ACCESS(pS)]++;
        }
    }
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute histogram for one line of image pixels covered by a mask.
 *
 * LongDescription:::
 *   Given an 4bpp gray image and a mask, histogram one line of the 
 *   image which have a corresponding "1" in the mask.  The mask
 *   does not need to start on a 32-bit boundary (it starts at firstBit).
 *   As a speedup a histogram of bytes is used if all the bits in a 
 *   mask word are on. The byte histogram must be used later to compute
 *   the nibble histogram.  (See comments in i_histoRect4.).  The
 *   nibble histogram is used when we have to look at individual bits,
 *   either in the first or last mask word or when the mask has holes
 *   in it.
 *
 * Input:::
 *  pS::		Source image, pointing to start of rectangle.
 *  pMLong::		Mask image, pointing to start of rectangle
 *  firstBit::		Bit number of first bit of rectangle (in mask image)
 *  firstWordMask::	32-bit mask used to remove bits (from mask) which 
 *			should not be histogrammed from the first mask
 *			word
 *  numFirstWordBits::  Number of bits to be accessed in first word of
 *			mask
 *  lastWordMask::	32-bit mask used to remove bits (from mask) which 
 *			should not be histogrammed from the last mask
 *			word
 *  w::			Number of pixels to be histogrammed
 *
 * Output:::
 *  pHisto::		The 4bpp histogram
 *  byteHisto::		The 8bpp histogram
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  JCraig
 *
 * EndDocumentation:::
 **************************************************************************/
void  CDECL 
ip_calcHistoMaskLine4 (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32     *pMLong,
    Int32	firstBit,
    UInt32	firstWordMask,
    UInt32	numFirstWordBits,
    UInt32	lastWordMask,
    Int32	w,
    Int32      *pHisto,
    Int32      *byteHisto)
{
    PROLOGUE;
    Int32	j, k;
    UInt32	bits, bitMask;
    UInt8	byte;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    /* What we want to do is get longwords out of the mask, so that
     * we read it efficiently, and so that if we see all 0 we can skip
     * 32 pixels at a time.
     */

    /* is the first word a full word?  if not, it must be done separately.
     * otherwise just let the next loop handle it since it is more 
     * efficient.
     */
    if (numFirstWordBits != 32)
    {
        bits = pMLong[0] & firstWordMask;
        pMLong++;
        if (bits != 0)
        {
	    bitMask = 1UL << (32 - firstBit - 1);
	    if ((firstBit & 0x1) != 0)
	    {
		/* then the first pixel is not on a byte boundary and
		 * we'll treat it separately.
		 */
		byte = UCHAR_ACCESS(pS);
		if (bits & bitMask)
		    pHisto[byte & 0xF]++;
		numFirstWordBits--;
		w--;
		pS++;
		bitMask >>= 1;
	    }

	    /* now do all the full bytes.
	     * note that if only one mask word is touched per line,
	     * numFirstWordBits may be an odd number here.  since the
	     * loop always references 2 pixels, this will cause us to
	     * reference an extra pixel.  however, firstWordMask should
	     * have set the mask bit corresponding to that extra pixel 
	     * to 0, so we will not count a pixel that should not have
	     * been counted. 
	     */
	    for (k = numFirstWordBits; k > 0; k -= 2, pS++)
	    {
		byte = UCHAR_ACCESS(pS);
		if (bits & bitMask)
		    pHisto[byte >> 4]++;
		bitMask >>= 1;
		if (bits & bitMask)
		    pHisto[byte & 0xF]++;
		bitMask >>= 1;
	    }
        }

	/* update the width for the number of bits done in the first word */
    	w -= numFirstWordBits;
    }

    /* now do all the full 32 bit mask words */
    for (j = 0; j < (w / 32); j++)
    {
	bits = pMLong[0];
	pMLong++;
	if (bits == 0)
	{
	    /* Nothing here matters; keep going */
	    pS += 16;
	    continue;
	}

	if (bits == 0xffffffff)
	{
	    /* Special case: all ones, so histogram them all */
	    /* note that byteswapping doesn't matter here since
	     * we're looking at all the pixels and order doesn't
	     * matter.
	     */
	    byteHisto[*pS]++;
	    byteHisto[*(pS+1)]++;
	    byteHisto[*(pS+2)]++;
	    byteHisto[*(pS+3)]++;
	    byteHisto[*(pS+4)]++;
	    byteHisto[*(pS+5)]++;
	    byteHisto[*(pS+6)]++;
	    byteHisto[*(pS+7)]++;
	    byteHisto[*(pS+8)]++;
	    byteHisto[*(pS+9)]++;
	    byteHisto[*(pS+10)]++;
	    byteHisto[*(pS+11)]++;
	    byteHisto[*(pS+12)]++;
	    byteHisto[*(pS+13)]++;
	    byteHisto[*(pS+14)]++;
	    byteHisto[*(pS+15)]++;
	    pS += 16;
	}
	else
	{
	    /* We look at each bit in the mask to
	     * determine whether to histogram a pixel in the image
	     */
	    bitMask = 1UL << 31;
	    for (k = 0; k < 16; k++, pS++)
	    {
		byte = UCHAR_ACCESS(pS);
		if (bits & bitMask)
		    pHisto[byte >> 4]++;
		bitMask >>= 1;
		if (bits & bitMask)
		    pHisto[byte & 0xF]++;
		bitMask >>= 1;
	    }
	}
    }

    /* now do the last partial word. 
     * why check for lastWordMask being 0?  well, if the rectangle ends
     * at the right edge of the mask image, then accessing pMLong would
     * access the frame.  not such a big deal but still no need to 
     * reference it if we don't have to.   in addition, the lastWordMask
     * may be 0, if we end on a word boundary.  again, we could ignore
     * this check and just refrence one too many words but just to be a
     * good citizen, check for this case.  note lastWordMask may be 0 
     * either because we ended on a word boundary OR because the we
     * only reference one source word (the rectangle starts and ends
     * in the same word.
     */
    if (lastWordMask != 0)
    {
        bits = pMLong[0] & lastWordMask;
        if (bits != 0)
        {
	    bitMask = 1UL << 31;
	    /* note that this loop always looks at 2 source pixels even
	     * if we really only needed to be looking at an odd number
	     * of pixels.  however, lastWordMask should have turned 
	     * that bit into a 0 so we haven't counted a pixel that
	     * shouldn't have been counted.
	     */
	    for (k = 0; k < (w & 0x1F); k += 2, pS++)
	    {
		byte = UCHAR_ACCESS(pS);
		if (bits & bitMask)
		    pHisto[byte >> 4]++;
		bitMask >>= 1;
		if (bits & bitMask)
		    pHisto[byte & 0xF]++;
		bitMask >>= 1;
	    }
        }
    }
}


/*
$Log: mihisto.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.5  1998/01/13 23:52:36  ddavies
mihisto.c -> /product/ipcore/ipshared/src/RCS/mihisto.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1997/07/03 21:59:08  jcraig
mihisto.c -> /product/ipcore/ipshared/src/RCS/mihisto.c,v

add ip_calcHistoMaskLine8 and ip_calcHistoMaskLine4 which histogram
one line for i_histoMaskRect

Revision 1.3  1995/12/13 02:50:19  ejaquith
mihisto.c -> /product/ipcore/ipshared/src/RCS/mihisto.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1994/03/25  22:42:57  ddavies
 * mihisto.c -> /product/ipcore/ipshared/src/RCS/mihisto.c,v
 *
 * Changed ip_calcFullHistogram into ip_calcHistogramLine.  The old routine
 * calculated the histogram for an entire image.  The new one calculates the
 * histogram for just one line.  This allows us to use this routine in banded and/or
 * tiled image formats at some point in the future.
 * Also removed the constraint that all the images we look at are word aligned.
 * This is necessary because the histoRect code uses this routine.  This
 * entailed some fussing about byte order.  Generally simplified the code.
 * Since we are retiring the SPARC assembly version of this routine, this
 * C code will be used in all implementations.
 *
 * Revision 1.1  1994/01/19  19:41:21  macleod
 * Initial revision
 *
*/


