/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mildc.c: low level code used in Lossy Dither compression.
 *
 *	Int32	ip_compLDRow()
 *			Compresses one row of data using LD compression.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "mmrcomm.prv"	/* for mmrTableEntry */
#define _NEED_LITERAL_LENGTH_CODES_
#define _NEED_STRING_LENGTH_CODES_
#define _NEED_LITERAL_CODES_
#include "ldc.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mildc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* This file contains the routine used to compress one row of data using
 * Lossy Dither encoding.  This method is novel.
 *
 * There are many methods of compressing binary images.  Virtually all of
 * them do a poor job on dithered (error diffused or halftoned) areas
 * simply because there are a very large number of tiny, noisy, dots
 * there.  It is very hard to predict the value of a pixel given the values
 * if its neighbors.
 *
 * Our approach to this problem is to note that in these areas, the
 * exact placement of each pixel is not nearly so important as the
 * gray level.  Yeah, we should make some attempt to preserve edges,
 * but the gray levels are what's really important.
 *
 * We encode the image in units of bytes.  We map the 256 possible values
 * into a much smaller set of equivalence classes.  There are many
 * mappings possible.  The one we choose (initially?) lets single
 * pixels float anywhere in a byte so long as they remain isolated,
 * pairs of pixels can move by +-2 bit positions and clumps of 3 or
 * 4 pixels can only move +- 1 bit position.  If a byte has more
 * white pixels than black pixels, we decide what's allowed to move
 * by looking at the black pixels.  If it has more black pixels,
 * we look at groups of white pixels.  This means that you can derive the
 * equivalence class for 5 black pixels (in a byte) by inverting all
 * the members of the equivalence class for 3 black pixels.  Similarly,
 * the members of the class of 6 black pixels are the inverses of
 * the members of the class for 2 black pixels and so on.
 *
 * What do we do with the classified bytes?  Like MMR, we have a
 * reference line and a coding line.  Each is the width of the image
 * and is filled with classified bytes (actually with equivalence 
 * class IDs, one per byte).  We start at the left edge of the coding line
 * and encode strings.  There are two types of strings.  One can either encode
 * just a string length, which is the number of bytes for which the
 * reference line is equal to the coding line, or one can encode a
 * literal length followed by the specified number of equivalence class IDs.
 * These two types of encoding always alternate.  A plain string length
 * is always first.  It may specify 0 bytes.
 *
 * Here's an example:
 * bytes	 0  1  2  3  4  5  6  7  8  9 10
 * Line 0:	00 00 01 00 02 00 21 44 46 46 46
 * Line 1:	00 00 01 01 00 00 24 45 46 46 46
 * Line 2: 	01 00 00 00 00 00 24 45 46 46 46
 *
 * We assume a reference line filled with the class used to encode
 * all white (pretend it's 0) above the top of the image.  Line 0
 * is then encoded using a first copy string length of 2 (the two 00's),
 * a literal string length of 1 followed by the value 01; a copy string
 * legnth of 1; a literal string length of 1 followed by the value 02,
 * a copy string length of 1 and a final literal string length of 5
 * followed by values of 21, 44, 46, 46, 46.
 *
 * Encoding line 1 is similar.  We encode a first copy length of
 * 3 (bytes 0, 1 and 2 are the same), a literal length of 2 followed
 * by the codes for classes 01 and 00, a copy length of 1, a literal
 * code of 2 and codes for classes 45 and 46 and a string length of 3.
 * On line 2, we start with a copy length of 0, a literal length of 1
 * followed by the code for equivalence class 1, a copy length of 1,
 * a literal length of 2 followed by the codes for class 0 and class 0
 * again, ending with a copy length of 7.
 *
 * This seems pretty simple, and it is.  Particularly the part about
 * only copying reference line bytes straight down.  Couldn't you find
 * longer matches if you looked around in the reference line instead of
 * only looking straight up?  Well, yes, you can.  However, you must
 * also specify the starting position of the matching string in the
 * reference line.  You encode fewer strings, but the number of bits
 * required to encode each string goes up.  In the tests I did, it
 * always made things a little worse.
 *
 * How do we encode things?  We did trial encodings of lots of images
 * to gather statistics on encoding frequencies of copy string lengths,
 * literal string lengths and equivalence class IDs (literal values).
 * We used the statisitics to build Huffman codes for all the values.
 *
 * The easiest table to understand is the one for equvalence class
 * IDs.  Some classes are much more likely to appear than others.
 * We built Huffman codes for the class numbers based on the frequency
 * of occurence in our training set.
 *
 * The lengths are a little more tricky.  We have two length encoding tables,
 * one for copy string lengths and one for literal string lengths.  The
 * histograms for code lengths are fairly different for these two types
 * of strings as you might expect.  In general, we hope that copy string
 * lengths are substantially longer than literal string lengths, and this
 * turns out to be the case.  In our traning set, the average length of
 * a copy string is 12.5 bytes and the average length of a literal
 * string is 3.6 bytes.
 *
 * The literal string length table contains individual entries for
 * strings of length 1 through 64.  It also contains a special prefix
 * for strings of length 65 through 128.  During encoding, this prefix
 * is followed by a 6 bit code specifying (length - 65).  There is also
 * a prefix for strings of length 129 through 256.  This is followed by
 * a 7 bit code for (length - 129).
 *
 * There's also a prefix code for lengths greater than 256.  It is followed
 * by a sequence of codes.  Each code has 8 bits.  The LS 7 bits
 * are a number.  The MS bit is a flag.  If equal to 0, this is the last
 * code of the sequence.  If equal to 1, another 8-bit code follows.  The
 * bottom 7 bits of this code are a number, which should be concatenated
 * with the 7-bit number from the first code.  The 7 new bits go on the
 * MS end of the new number.  As as before, the MS bit is a flag bit.  
 * If this is the last 8-bit code, it's a 0.  If not, it's a 1 and another
 * 8-bit code follows.  This process continues until we reach the end of
 * the string of 8-bit codes.  The final number is the concatenation of
 * the original 7 bit number (on the LS end) and each successive 7 bit
 * number tacked on top of it.
 *
 * There's one more type of code for literal string lengths.  It signifies
 * that the literal string is long enough to exactly fill out the coding
 * line.  We know the width of the coding line in bytes since that's
 * how we know we've reached the end of a scan line.  This code is useful
 * for eliminating long literal string codes.  It appears fairly often.
 *
 * 
 * The codes for copy strings are similar.  There are individual codes for
 * strings of length 0 through 63 (remember that we need to encode strings
 * of length 0 when the first byte of the coding line differs from the first
 * byte of the reference line).  It also has prefix codes for string lengths
 * of [64..127] and [128..255] and greater than 255.  These follow the same
 * pattern as the codes for the literal string lenghts.  As before, there
 * is also a "to end of coding line" code.
 *
 * The string lengths also have three special codes that turn out to be
 * quite useful.  In Modified Modified Read encoding, one attempts to
 * encode lengths relative to known markers in the reference line.  This
 * is a powerful technique since one can subsitute one code for a large
 * number of individual codes.  In MMR, the markers are calculated as
 * places where the reference line changes from white to black or vice versa.
 * In our coding scheme, markers are placed at the copy string boundaries
 * that were used to create the reference line.  We do this on the assumption
 * that things don't change all that much from line to line.
 *
 * There are four ways to code using markers.  In all four cases, one
 * looks at the next two markers to the right from the current position.
 * There are three codes signifying that the copy string ends one byte
 * to the left of the next marker, exactly at the next marker and one
 * byte to the right of the next marker respectively.  There is a fourth code
 * for a copy string that ends one byte to the left of the marker after
 * the next marker.  I tried lots of other marker-relative positions
 * and stuff, like putting in MMR-like pass codes to skip
 * unused markers, but they all seemed to increase the size of the
 * compressed file.
 *
 *			*	*	*
 *
 * What about decoding?  Reconstructing the strings of equivalence class
 * numbers is relatively straightforward.  The trick comes in deciding how
 * to paint them.  If we simply choose one member from each equivalence
 * class, we'd get a bunch of vertical streaks in the image.  Instead, we
 * use another table to re-dither the members of each class, much like
 * halftoning.  The dither values for a class are chosen from the members
 * of that class.  We also make some attempt to dither clumps of 2 dots
 * differently than single dots on the assumption that each printed dot
 * was scanned as two pixels.
 *
 * Code at this lower level only encodes and decodes strings of equivalence
 * class numbers.  Code at the next higher level turns bytes into
 * equivalence class IDs on compression and redithers these IDs on
 * decompression.
 *
 *			*	*	*
 *
 * The encoding loop is a state machine, much like the decoding
 * loop.  In fact, we'll try to make them as similar as possible
 * to reduce the chance of error.  In the encoder, one trip around
 * the loop produces one copy string length, one literal string length
 * and the codes for a string of literal bytes.  We do not encode
 * a literal string if the copy string filled the coding line.
 *
 * In the decoder, each trip through the loop processes a copy string
 * followed by a literal string.  Again, we stop before the literal
 * string if the copy string filled the coding line.  Once the line
 * is decoded, we use the re-dithering table to paint the scan line.
 *
 */

/* PRIVATE */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress 1 row of a binary image using Lossy Dither compression.
 *
 * LongDescription:::
 *  Compress one row of an image using the Lossy Dither algorithm.
 *  We will stop when either an error is encountered, we finish the row
 *  or we fill the compressed data buffer.
 *
 * Input:::
 *  pLDCompInfo::           Pointer to private LD compression state block
 *  imWidth::               the width of the image not including frame.
 *  pCompDataEnd::          pointer to byte just beyond end of compressed data.
 *
 * Output:::
 *  ppCompData::            Pointer to pointer to beginning of compressed
 *			    data block.  The data block pointer is updated
 *			    at the end of this routine.
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid,
 *  ia_bufferExceeded => ran out of space in compressed data buffer.
 *
 * Notes:::
 *  The compressed data stream is always in natural byte order.
 *  The coding and reference lines are also both in natural byte order
 *  since they were stored that way by higher level code.
 *		
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
ip_compLDRow (
    IPState    *pIPState,
    LDCompressionInfo *pLDCompInfo,
    Int32	imWidth,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    UInt8	*pFirstDataByte, *pFirstRefByte; /* ptrs to first bytes in
					 * coding and reference lines */
    Int32	*markerTable;		/* pointer to first word in marker
					 * table.  The marker table contains
					 * the offsets of the ends of copy
					 * strings in the reference line. */
    Int32	 dataIndex;		/* index of cur byte in coding line */
    Int32	 endIndex;		/* index of byte just beyond end
					 * of coding line. */
    Int32	 stringIndex;		/* records end of matching string */
    Int32	 litIndex;		/* index of next literal byte
					 * to send */
    Int32	 markerIndex;		/* index of next marker location */
    UInt32	 accum;			/* accumulated encoded bits */
    Int32	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */
    Int32	 remainingLength;	/* used when we stop in the middle
					 * of encoding a long thing. */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pLDCompInfo->accumBits;
    vacantBit = pLDCompInfo->vacantBit;
    dataIndex = pLDCompInfo->dataIndex;
    stringIndex = pLDCompInfo->stringIndex;
    litIndex = pLDCompInfo->litIndex;
    remainingLength = pLDCompInfo->remainingLength;
    endIndex = (imWidth + 7) / 8;
    markerIndex = pLDCompInfo->markerIndex;
    markerTable = (Int32 *) ((UInt8 *) pLDCompInfo +
			     pLDCompInfo->markerOffset);

	/* Most of time time, we want to compare to
	 * pCompDataEnd-1, so decrement it here */
    pCompDataEnd--;

	/* point to current byte by first pointing to first real byte,
	 * To do this, we figure out which byte array is the coding line and
	 * which is the reference line.  pLDCompInfo->line1IsCoding
	 * is set when the line1 is the coding line.
	 */
    if (pLDCompInfo->line1IsCoding)
    {
	pFirstDataByte = ((UInt8 *)(pLDCompInfo) + pLDCompInfo->line1Offset);
	pFirstRefByte = ((UInt8 *)(pLDCompInfo) + pLDCompInfo->line0Offset);
    }
    else	/* the first run array holds the coding line */
    {
	pFirstDataByte = ((UInt8 *)(pLDCompInfo) + pLDCompInfo->line0Offset);
	pFirstRefByte = ((UInt8 *)(pLDCompInfo) + pLDCompInfo->line1Offset);
    }

	/* see where we should start off */
    switch (pLDCompInfo->compState)
    {
    case cLDCompNewLine:
	    /* Start at the beginning of the scan line.
	     */
	if (dataIndex >= endIndex)
	    goto compLDGoodCompletion;
	else
	    break;

    case cLDCompEncodeCopyLength:
	goto compLDEncodeCopyLength;

    case cLDCompEncodeLongCopyLength:
	goto compLDEncodeLongCopyLength;

    case cLDCompEncodeLiteralLength:
	goto compLDEncodeLiteralLength;

    case cLDCompEncodeLongLiteralLength:
	goto compLDEncodeLongLiteralLength;

    case cLDCompEncodeLiteralString:
	goto compLDEncodeLiteralString;

    default:	/* Restart state belongs to some other routine */
	NORMAL_RETURN;
    }

	/* ********************* MAIN LOOP ***********************/

	/* Here's the main loop.  In it, we match a string of coding
	 * bytes with the string of reference bytes immediately above
	 * it.  If we find a match, we encode the length of the
	 * reference string.  If not, we encode a copy string length
	 * of 0.  If we're not at the end of a scan line, we
	 * then encode a literal string.
	 *
	 * Note that the reference and coding lines end in different dummy
	 * bytes.  Each is a bigger number than the last equivalence
	 * class ID.  This guarantees that no legal class ID
	 * will match the dummies and the two dummy entries will
	 * not match each other.  To make sure that the check of
	 * the end of a literal string ends, each dummy is
	 * followed by a byte containing 0, which looks like
	 * the end of a string of different bytes.  This enables
	 * us to skip the end of line check most of the time.
	 */
    while (dataIndex < endIndex)
    {
	    /* First, update the index into the marker table.  We
	     * want to stop in front of the first marker that's
	     * to the right of our current position.  There's
	     * always a dummy marker off the right edge of the
	     * image, so we'll hit that if nothing else appears.
	     */
	while (markerTable[markerIndex] <= dataIndex)
	    markerIndex++;

	    /* -------------- FIND STRING LENGTHS ----------------- */

	    /* Here's where we search for a match of coding line
	     * bytes using the reference line.  First, match up
	     * to the next 32-bit word boundary, then try to
	     * match 32-bit words.  When that fails, try to
	     * match bytes within the last word.
	     */
	if ((dataIndex & ~0x3) != dataIndex)	/* go do first partial word */
	{
	    stringIndex = dataIndex;
	    if (pFirstDataByte[stringIndex] !=
			pFirstRefByte[stringIndex])
		goto compLDFindLitLength;
	    if (pFirstDataByte[stringIndex + 1] !=
			pFirstRefByte[stringIndex + 1])
	    {
		stringIndex++;	/* record pos of first miss */
		goto compLDFindLitLength;
	    }
	    if (pFirstDataByte[stringIndex + 2] !=
			pFirstRefByte[stringIndex + 2])
	    {
		stringIndex += 2;	/* record pos of first miss */
		goto compLDFindLitLength;
	    }

		/* all the remaining bytes in the first word, and
		 * perhaps some of the bytes in the second word, matched.
		 * Proceed to match in units of words starting at
		 * the second word.
		 */
	    stringIndex = (dataIndex & ~0x3) + 4;
	}
	else
	    stringIndex = dataIndex;	/* start matching at first word */

	for (; ; stringIndex += 4)
	{
	    if (ULONG_ACCESS(&pFirstDataByte[stringIndex]) !=
			ULONG_ACCESS(&pFirstRefByte[stringIndex]))
		break;
	}

	    /* Find out how many bytes matched in the last word */
	if (pFirstDataByte[stringIndex] !=
		    pFirstRefByte[stringIndex])
	    goto compLDFindLitLength;
	if (pFirstDataByte[stringIndex + 1] !=
		    pFirstRefByte[stringIndex + 1])
	{
	    stringIndex++;	/* record pos of first miss */
	    goto compLDFindLitLength;
	}
	if (pFirstDataByte[stringIndex + 2] !=
		    pFirstRefByte[stringIndex + 2])
	{
	    stringIndex += 2;	/* record pos of first miss */
	    goto compLDFindLitLength;
	}
	stringIndex += 3;	/* record pos of first miss */

compLDFindLitLength:
	    /* Find length of following literal string.  However, if
	     * we're at the end of the line, don't bother.
	     * Note that we can't use the trick of matching entire
	     * words here since we expect them to mismatch.
	     * The average length of a literal string is only
	     * a little over 3 bytes anyway, so there's not a
	     * lot of optimizing to do here.
	     * YET Another bug in Microsoft MSVC 5.0, Service Pack 3.
	     * In the case where the first comparison in the loop
	     * is the one that triggers the loop exit, the dumb 
	     * compiler forgets to store the current index, in ecx,
	     * into memory.  Duh!
	     * In an attempt to work around this, we try storing
	     * the final value of litIndex in remainingLength.  At
	     * the bottom of the loop, we restore it.
	     */
	for (litIndex = stringIndex, remainingLength = endIndex+4;
	     litIndex < endIndex;
	     litIndex += 4)
	{
	    if (pFirstDataByte[litIndex] ==
			pFirstRefByte[litIndex])
	    {
		remainingLength = litIndex;  /* Only here because of
					      * MSVC 5.0 bug! */
		break;
	    }
	    if (pFirstDataByte[litIndex + 1] ==
			pFirstRefByte[litIndex + 1])
	    {
		remainingLength = litIndex + 1;	/* record pos of first match */
		break;
	    }
	    if (pFirstDataByte[litIndex + 2] ==
			pFirstRefByte[litIndex + 2])
	    {
		remainingLength = litIndex + 2;	/* record pos of first match */
		break;
	    }
	    if (pFirstDataByte[litIndex + 3] ==
			pFirstRefByte[litIndex + 3])
	    {
		remainingLength = litIndex + 3;	/* record pos of first match */
		break;
	    }
	}

	    /* There are two conditions under which we arrive here.
	     * 1. the loop exited because litIndex >= endIndex;
	     * 2. We found a match and exited with the offset in
	     *    remainingLength.
	     */
	litIndex = MIN (remainingLength, endIndex);

	    /* -------------- ENCODE COPY STRING LENGTH ----------------- */

	    /* Now we encode the various lengths.  First, encode
	     * the copy string length.  Since all the
	     * copy string codes or prefix codes fit into 16
	     * bits, try making room by clearing out 16 bits
	     * first.
	     */
compLDEncodeCopyLength:
	if (vacantBit >= 16)
	{
	    if (pCurCompData >= pCompDataEnd)
		goto abortEncCopyLength;

		/* write out LS 2 bytes of accumulator and increment
		 * pointers */
	    pCurCompData[0] = reverseByte[(UInt8) accum];
	    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
	    accum >>= 16;		/* get rid of written bytes */
	    pCurCompData += 2;
	    vacantBit -= 16;
	}

	    /* Using this is easier than recomputing it all the time */
	remainingLength = stringIndex - dataIndex;

	    /* Try encoding string length relative to the
	     * next marker or the one after that if the end of
	     * the string is close to either of them.
	     */
	if (ABS((stringIndex - 1) - markerTable[markerIndex]) <= 1)
	{
	    switch ((stringIndex-1) - markerTable[markerIndex])
	    {
	    case -1:
		    /* String ends exactly 1 byte before next marker */
		accum |=
		    ldcStringLengthCodes[cLDCopyLenToMarker0M1].codeBits <<
								     vacantBit;
		vacantBit +=
		    ldcStringLengthCodes[cLDCopyLenToMarker0M1].codeLength;
		break;
	    case 0:
		    /* String ends exactly at next marker */
		accum |=
		    ldcStringLengthCodes[cLDCopyLenToMarker0].codeBits <<
								     vacantBit;
		vacantBit +=
		    ldcStringLengthCodes[cLDCopyLenToMarker0].codeLength;
		break;
	    case 1:
		    /* String ends one byte after at next marker */
		accum |=
		    ldcStringLengthCodes[cLDCopyLenToMarker0P1].codeBits <<
								     vacantBit;
		vacantBit +=
		    ldcStringLengthCodes[cLDCopyLenToMarker0P1].codeLength;
		break;
	    default:
	        return ia_internal;
	    }
	}
	else
	if (markerTable[markerIndex+1] == (stringIndex - 1))
	{
		/* String ends exactly 1 byte before marker after next marker */
	    accum |=
		ldcStringLengthCodes[cLDCopyLenToMarker1].codeBits <<
								 vacantBit;
	    vacantBit +=
		ldcStringLengthCodes[cLDCopyLenToMarker1].codeLength;
	}
	else
	if (stringIndex >= endIndex)
	{
		/* String goes to end of coding line */
	    accum |=
		ldcStringLengthCodes[cLDCopyLenToEOL].codeBits << vacantBit;
	    vacantBit +=
		ldcStringLengthCodes[cLDCopyLenToEOL].codeLength;
	}
	else
	if (remainingLength <= cLDShortCopyLength)
	{
		/* encode exact string length */
	    accum |=
		ldcStringLengthCodes[remainingLength].codeBits << vacantBit;
	    vacantBit +=
		ldcStringLengthCodes[remainingLength].codeLength;
	}
	else
	if (remainingLength <= cLDMediumCopyLength)
	{
		/* encode the length prefix followed by the length
		 * minus cLDMediumCopyLength.
		 * This fits in 16 bits, so we don't have to make
		 * more room.
		 * Note that the length is bit-reversed so it looks
		 * like everything else in the accumulator.  We
		 * only want the LS cLDMediumCopyLengthBits of it,
		 * so shift the reversed bits right by
		 * 8 - cLDMediumCopyLengthBits before shift them left
		 * by vacantBits, or just shift left by
		 * vacantBits - (8 - cLDMediumCopyLengthBits).
		 * We can always shift left because we know that
		 * the length of the prefix is greater than
		 * (8 - cLDMediumCopyLengthBits).
		 */
	    accum |=
		ldcStringLengthCodes[cLDMediumCopyLenPrefix].codeBits <<
								    vacantBit;
	    vacantBit +=
		ldcStringLengthCodes[cLDMediumCopyLenPrefix].codeLength;
	    remainingLength -= cLDShortCopyLength + 1;
	    accum |= reverseByte[remainingLength] <<
				(vacantBit - (8 - cLDMediumCopyLengthBits));
	    vacantBit += cLDMediumCopyLengthBits;
	}
	else
	if (remainingLength <= cLDLongCopyLength)
	{
		/* encode the length prefix followed by the length.
		 * This also fits in 16 bits, so we don't have to make
		 * more room.
		 */
	    accum |=
		ldcStringLengthCodes[cLDLongCopyLenPrefix].codeBits <<
								    vacantBit;
	    vacantBit +=
		ldcStringLengthCodes[cLDLongCopyLenPrefix].codeLength;
	    remainingLength -= cLDMediumCopyLength + 1;
	    accum |= reverseByte[remainingLength] <<
				(vacantBit - (8 - cLDLongCopyLengthBits));
	    vacantBit += cLDLongCopyLengthBits;
	}
	else	/* it was a very long string */
	{
		/* encode the length prefix followed by the length.
		 */
	    accum |=
		ldcStringLengthCodes[cLDVeryLongCopyLenPrefix].codeBits <<
								    vacantBit;
	    vacantBit +=
		ldcStringLengthCodes[cLDVeryLongCopyLenPrefix].codeLength;

	    remainingLength -= cLDLongCopyLength + 1;
		/* Put out the sequence of length bytes.
		 * Always put out at least one.
		 */
	    do
	    {
compLDEncodeLongCopyLength:
		if (vacantBit >= 16)
		{
		    if (pCurCompData >= pCompDataEnd)
			goto abortEncLongCopyLength;
	
			/* write out LS 2 bytes of accumulator and increment
			 * pointers */
		    pCurCompData[0] = reverseByte[(UInt8) accum];
		    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		    accum >>= 16;		/* get rid of written bytes */
		    pCurCompData += 2;
		    vacantBit -= 16;
		}
    
		if (remainingLength < 128)	/* all fits in one byte */
		    accum |= reverseByte[remainingLength] << vacantBit;
		else	/* need more than just that byte */
		    accum |= reverseByte[(remainingLength & 0x7F) | 0x80] <<
								     vacantBit;
		vacantBit += 8;
		remainingLength = (UInt32)remainingLength >> 7;
	    } while (remainingLength > 0);
	}

	    /* --------------- ENCODE LITERAL LENGTH --------------------- */

	    /* Now we encode the literal string length.  If it's a long code,
	     * go do it in another subroutine.  Since all the
	     * literal lengths or prefix codes fit into 16
	     * bits, try making room by clearing out 16 bits
	     * first.
	     * BUT FIRST, if the copy string went to the end of the
	     * scan line, litIndex == stringIndex.  In this case, do nothing.
	     * In particular, don't insist on emptying the accumulator.
	     */
	if (litIndex == stringIndex)
	    break;
	else
	    remainingLength = litIndex - stringIndex;

compLDEncodeLiteralLength:
	if (vacantBit >= 16)
	{
	    if (pCurCompData >= pCompDataEnd)
		goto abortEncLiteralLength;

		/* write out LS 2 bytes of accumulator and increment
		 * pointers */
	    pCurCompData[0] = reverseByte[(UInt8) accum];
	    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
	    accum >>= 16;		/* get rid of written bytes */
	    pCurCompData += 2;
	    vacantBit -= 16;
	}

	    /* Encode the literal length.  This is a lot like encoding
	     * the copy string length.
	     */
	if (litIndex >= endIndex)
	{
		/* String goes to end of coding line */
	    accum |=
		ldcLiteralLengthCodes[cLDLitLenToEOL].codeBits << vacantBit;
	    vacantBit +=
		ldcLiteralLengthCodes[cLDLitLenToEOL].codeLength;
	}
	else
	if (remainingLength <= cLDShortLitLength)
	{
		/* encode exact string length */
	    accum |=
		ldcLiteralLengthCodes[remainingLength].codeBits << vacantBit;
	    vacantBit +=
		ldcLiteralLengthCodes[remainingLength].codeLength;
	}
	else
	if (remainingLength <= cLDMediumLitLength)
	{
		/* encode the length prefix followed by the length.
		 * This fits in 16 bits, so we don't have to make
		 * more room.
		 * Note that the length is bit-reversed so it looks
		 * like everything else in the accumulator.  We
		 * only want the LS cLDMediumLitLengthBits of it,
		 * so shift the reversed bits right by
		 * 8 - cLDMediumLitLengthBits before shift them left
		 * by vacantBits, or just shift left by
		 * vacantBits - (8 - cLDMediumLitLengthBits).
		 * We can always shift left because we know that
		 * the length of the prefix is greater than
		 * (8 - cLDMediumLitLengthBits).
		 */
	    accum |=
		ldcLiteralLengthCodes[cLDMediumLitLenPrefix].codeBits <<
								    vacantBit;
	    vacantBit +=
		ldcLiteralLengthCodes[cLDMediumLitLenPrefix].codeLength;
	    remainingLength -= cLDShortLitLength + 1;
	    accum |= reverseByte[remainingLength] <<
				(vacantBit - (8 - cLDMediumLitLengthBits));
	    vacantBit += cLDMediumLitLengthBits;
	}
	else
	if (remainingLength <= cLDLongLitLength)
	{
		/* encode the length prefix followed by the length.
		 * This also fits in 16 bits, so we don't have to make
		 * more room.
		 */
	    accum |=
		ldcLiteralLengthCodes[cLDLongLitLenPrefix].codeBits <<
								    vacantBit;
	    vacantBit +=
		ldcLiteralLengthCodes[cLDLongLitLenPrefix].codeLength;
	    remainingLength -= cLDMediumLitLength + 1;
	    accum |= reverseByte[remainingLength] <<
				(vacantBit - (8 - cLDLongLitLengthBits));
	    vacantBit += cLDLongLitLengthBits;
	}
	else	/* it was a very long string */
	{
		/* encode the length prefix followed by the length.
		 */
	    accum |=
		ldcLiteralLengthCodes[cLDVeryLongLitLenPrefix].codeBits <<
								    vacantBit;
	    vacantBit +=
		ldcLiteralLengthCodes[cLDVeryLongLitLenPrefix].codeLength;

	    remainingLength -= cLDLongLitLength + 1;
		/* Put out the sequence of length bytes.
		 * Always put out at least one.
		 */
	    do
	    {
compLDEncodeLongLiteralLength:
		if (vacantBit >= 16)
		{
		    if (pCurCompData >= pCompDataEnd)
			goto abortEncLongLiteralLength;
	
			/* write out LS 2 bytes of accumulator and increment
			 * pointers */
		    pCurCompData[0] = reverseByte[(UInt8) accum];
		    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		    accum >>= 16;		/* get rid of written bytes */
		    pCurCompData += 2;
		    vacantBit -= 16;
		}
    
		if (remainingLength < 128)	/* all fits in one byte */
		    accum |= reverseByte[remainingLength] << vacantBit;
		else	/* need more than just that byte */
		    accum |= reverseByte[(remainingLength & 0x7F) | 0x80] <<
								     vacantBit;
		vacantBit += 8;
		remainingLength = (UInt32)remainingLength >> 7;
	    } while (remainingLength > 0);
	}


	    /* --------------- ENCODE LITERAL CODES --------------------- */

	    /* Now we encode a string of literals.  How many?  Well,
	     * litIndex - stringIndex of them.  We may have destroyed
	     * remainingLength by the time we get here.
	     */
	for (remainingLength = stringIndex;
	     remainingLength != litIndex;
	     remainingLength++)
	{
compLDEncodeLiteralString:
	    if (vacantBit >= 16)
	    {
		if (pCurCompData >= pCompDataEnd)
		    goto abortEncLiteralString;
    
		    /* write out LS 2 bytes of accumulator and increment
		     * pointers */
		pCurCompData[0] = reverseByte[(UInt8) accum];
		pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		accum >>= 16;		/* get rid of written bytes */
		pCurCompData += 2;
		vacantBit -= 16;
	    }
		/* encode exact string length */
	    accum |=
		ldcLiteralCodes[pFirstDataByte[remainingLength]].codeBits <<
								    vacantBit;
	    vacantBit +=
		ldcLiteralCodes[pFirstDataByte[remainingLength]].codeLength;
	}

	    /* Advance to the end of the literal area */
	dataIndex = litIndex;
    }	/* done with encoding one line. */


	/* We get here when everything was swell.  We've generated
	 * the codes for an entire line.
	 * We never add padding in LD, so no need to do that.  We also
	 * don't flush the accumulator at the end of a scan line, so
	 * don't do that either.
	 * Just save the state so we know we're done with this scan
	 * line and can concatenate the codes for the next scan line
	 * when it comes.
	 */
compLDGoodCompletion:
    *ppCompData = pCurCompData;
    pLDCompInfo->dataIndex = 0;
    pLDCompInfo->stringIndex = 0;
    pLDCompInfo->litIndex = 0;
    pLDCompInfo->markerIndex = 0;
    pLDCompInfo->remainingLength = 0;
    pLDCompInfo->compState = cLDCompNewLine;
    pLDCompInfo->accumBits = accum;
    pLDCompInfo->vacantBit = vacantBit;

    NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

	/* First, we have all the abort targets.  We do this weird
	 * thing of setting all the states here instead of above
	 * in the hopes that the compiler will generate better
	 * code in the real loops - just a test a jump down here
	 * where the normal case is straight ahead instead of
	 * a test and jump around the error stuff.
	 * We get here when we had at least two bytes to write
	 * to the compressed data stream, but there wasn't room
	 * for 2 bytes.  There might be room for 1 byte.
	 */
abortEncCopyLength:
    pLDCompInfo->compState = cLDCompEncodeCopyLength;
    goto compLDMightWriteOne;
abortEncLongCopyLength:
    pLDCompInfo->compState = cLDCompEncodeLongCopyLength;
    goto compLDMightWriteOne;
abortEncLiteralLength:
    pLDCompInfo->compState = cLDCompEncodeLiteralLength;
    goto compLDMightWriteOne;
abortEncLongLiteralLength:
    pLDCompInfo->compState = cLDCompEncodeLongLiteralLength;
    goto compLDMightWriteOne;
abortEncLiteralString:
    pLDCompInfo->compState = cLDCompEncodeLiteralString;
    goto compLDMightWriteOne;

	/* We get here when we know that two bytes won't fit into the
	 * compressed output buffer.  Try putting one in.
	 * First, restore pCompDataEnd.  Remember that we decremented
	 * it at the top of the routine.
	 */
compLDMightWriteOne:
    pCompDataEnd++;
    if (pCurCompData < pCompDataEnd)
    {
	pCurCompData[0] = reverseByte[(UInt8) accum];
	pCurCompData++;
	accum >>= 8;
	vacantBit -= 8;
    }

	/* This is where we end up when we abort because we ran out
	 * of space in the compression buffer.
	 */
    *ppCompData = pCurCompData;
    pLDCompInfo->dataIndex = dataIndex;
    pLDCompInfo->stringIndex = stringIndex;
    pLDCompInfo->litIndex = litIndex;
    pLDCompInfo->markerIndex = markerIndex;
    pLDCompInfo->remainingLength = remainingLength;
    pLDCompInfo->accumBits = accum;
    pLDCompInfo->vacantBit = vacantBit;

    return ia_bufferExceeded;	/* indicates that we stopped because we
				 * ran out of space.
				 */
}


/*

 $Log: mildc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.3  1998/01/21 21:48:16  ddavies
 mildc.c -> /product/ipcore/ipshared/src/RCS/mildc.c,v

 Work around bug in code generator for Microsoft Visual C/C++ version 5.0,
 Service Pack 3.  When running through the array of equivalence classes
 for Lossy Dither compression trying to find the length of a literal string
 (i.e. a string none of whose members match the reference line), the
 code generator forgot to save the index of the end of the string in one
 of four cases.  We now route the index through another variable, which
 seems to make it happy.

 Revision 1.2  1998/01/13 23:52:38  ddavies
 mildc.c -> /product/ipcore/ipshared/src/RCS/mildc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1997/03/17 19:48:18  ddavies
 Initial revision


 */

