/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mildd.c: low level code used in Lossy Dither decompression.
 *
 *	Int32	ip_decompLDRow()
 *			Decompresses one row of data using LD decompression.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "mmrcomm.prv"	/* for mmrTableEntry */
#define _NEED_LITERAL_LENGTH_DECODING_TABLE_
#define _NEED_COPY_LENGTH_DECODING_TABLE_
#define _NEED_LITERAL_DECODING_TABLE_
#include "ldc.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mildd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* This file contains the routine used to decompress one row of data 
 * compressed using Lossy Dither encoding.  This method is novel.
 *
 * There are many methods of compressing binary images.  Virtually all of
 * them do a poor job on dithered (error diffused or halftoned) areas
 * simply because there are a very large number of tiny, noisy, dots
 * there.  It is very hard to predict the value of a pixel given the values
 * if its neighbors.
 *
 * Our approach to this problem is to note that in these areas, the
 * exact placement of each pixel is not nearly so important as the
 * gray level.  Yeah, we should make some attempt to preserve edges,
 * but the gray levels are what's really important.
 *
 * We encode the image in units of bytes.  We map the 256 possible values
 * into a much smaller set of equivalence classes.  There are many
 * mappings possible.  The one we choose (initially?) lets single
 * pixels float anywhere in a byte so long as they remain isolated,
 * pairs of pixels can move by +-2 bit positions and clumps of 3 or
 * 4 pixels can only move +- 1 bit position.  If a byte has more
 * white pixels than black pixels, we decide what's allowed to move
 * by looking at the black pixels.  If it has more black pixels,
 * we look at groups of white pixels.  This means that you can derive the
 * equivalence class for 5 black pixels (in a byte) by inverting all
 * the members of the equivalence class for 3 black pixels.  Similarly,
 * the members of the class of 6 black pixels are the inverses of
 * the members of the class for 2 black pixels and so on.
 *
 * What do we do with the classified bytes?  Like MMR, we have a
 * reference line and a coding line.  Each is the width of the image
 * and is filled with classified bytes (actually with equivalence 
 * class IDs, one per byte).  We start at the left edge of the coding line
 * and encode strings.  There are two types of strings.  One can either encode
 * just a string length, which is the number of bytes for which the
 * reference line is equal to the coding line, or one can encode a
 * literal length followed by the specified number of equivalence class IDs.
 * These two types of encoding always alternate.  A plain string length
 * is always first.  It may specify 0 bytes.
 *
 * Here's an example:
 * bytes	 0  1  2  3  4  5  6  7  8  9 10
 * Line 0:	00 00 01 00 02 00 21 44 46 46 46
 * Line 1:	00 00 01 01 00 00 24 45 46 46 46
 * Line 2: 	01 00 00 00 00 00 24 45 46 46 46
 *
 * We assume a reference line filled with the class used to encode
 * all white (pretend it's 0) above the top of the image.  Line 0
 * is then encoded using a first copy string length of 2 (the two 00's),
 * a literal string length of 1 followed by the value 01; a copy string
 * legnth of 1; a literal string length of 1 followed by the value 02,
 * a copy string length of 1 and a final literal string length of 5
 * followed by values of 21, 44, 46, 46, 46.
 *
 * Encoding line 1 is similar.  We encode a first copy length of
 * 3 (bytes 0, 1 and 2 are the same), a literal length of 2 followed
 * by the codes for classes 01 and 00, a copy length of 1, a literal
 * code of 2 and codes for classes 45 and 46 and a string length of 3.
 * On line 2, we start with a copy length of 0, a literal length of 1
 * followed by the code for equivalence class 1, a copy length of 1,
 * a literal length of 2 followed by the codes for class 0 and class 0
 * again, ending with a copy length of 7.
 *
 * This seems pretty simple, and it is.  Particularly the part about
 * only copying reference line bytes straight down.  Couldn't you find
 * longer matches if you looked around in the reference line instead of
 * only looking straight up?  Well, yes, you can.  However, you must
 * also specify the starting position of the matching string in the
 * reference line.  You encode fewer strings, but the number of bits
 * required to encode each string goes up.  In the tests I did, it
 * always made things a little worse.
 *
 * How do we encode things?  We did trial encodings of lots of images
 * to gather statistics on encoding frequencies of copy string lengths,
 * literal string lengths and equivalence class IDs (literal values).
 * We used the statisitics to build Huffman codes for all the values.
 *
 * The easiest table to understand is the one for equvalence class
 * IDs.  Some classes are much more likely to appear than others.
 * We built Huffman codes for the class numbers based on the frequency
 * of occurence in our training set.
 *
 * The lengths are a little more tricky.  We have two length encoding tables,
 * one for copy string lengths and one for literal string lengths.  The
 * histograms for code lengths are fairly different for these two types
 * of strings as you might expect.  In general, we hope that copy string
 * lengths are substantially longer than literal string lengths, and this
 * turns out to be the case.  In our traning set, the average length of
 * a copy string is 12.5 bytes and the average length of a literal
 * string is 3.6 bytes.
 *
 * The literal string length table contains individual entries for
 * strings of length 1 through 64.  It also contains a special prefix
 * for strings of length 65 through 128.  During encoding, this prefix
 * is followed by a 6 bit code specifying (length - 65).  There is also
 * a prefix for strings of length 129 through 256.  This is followed by
 * a 7 bit code for (length - 129).
 *
 * There's also a prefix code for lengths greater than 256.  It is followed
 * by a sequence of codes.  Each code has 8 bits.  The LS 7 bits
 * are a number.  The MS bit is a flag.  If equal to 0, this is the last
 * code of the sequence.  If equal to 1, another 8-bit code follows.  The
 * bottom 7 bits of this code are a number, which should be concatenated
 * with the 7-bit number from the first code.  The 7 new bits go on the
 * MS end of the new number.  As as before, the MS bit is a flag bit.  
 * If this is the last 8-bit code, it's a 0.  If not, it's a 1 and another
 * 8-bit code follows.  This process continues until we reach the end of
 * the string of 8-bit codes.  The final number is the concatenation of
 * the original 7 bit number (on the LS end) and each successive 7 bit
 * number tacked on top of it.
 *
 * There's one more type of code for literal string lengths.  It signifies
 * that the literal string is long enough to exactly fill out the coding
 * line.  We know the width of the coding line in bytes since that's
 * how we know we've reached the end of a scan line.  This code is useful
 * for eliminating long literal string codes.  It appears fairly often.
 *
 * 
 * The codes for copy strings are similar.  There are individual codes for
 * strings of length 0 through 63 (remember that we need to encode strings
 * of length 0 when the first byte of the coding line differs from the first
 * byte of the reference line).  It also has prefix codes for string lengths
 * of [64..127] and [128..255] and greater than 255.  These follow the same
 * pattern as the codes for the literal string lenghts.  As before, there
 * is also a "to end of coding line" code.
 *
 * The string lengths also have three special codes that turn out to be
 * quite useful.  In Modified Modified Read encoding, one attempts to
 * encode lengths relative to known markers in the reference line.  This
 * is a powerful technique since one can subsitute one code for a large
 * number of individual codes.  In MMR, the markers are calculated as
 * places where the reference line changes from white to black or vice versa.
 * In our coding scheme, markers are placed at the ends of the
 * copy strings that were used to create the reference line.  We do this
 * on the assumption that things don't change all that much from line to line.
 *
 * There are four ways to code using markers.  In all four cases, one
 * looks at the next two markers to the right from the current position.
 * There are three codes signifying that the copy string ends one byte
 * to the left of the next marker, exactly at the next marker and one
 * byte to the right of the next marker respectively.  There is a fourth code
 * for a copy string that ends one byte to the left of the marker after
 * the next marker.  I tried lots of other marker-relative positions
 * and stuff, like putting in MMR-like pass codes to skip
 * unused markers, but they all seemed to increase the size of the
 * compressed file.
 *
 *			*	*	*
 *
 * What about decoding?  Reconstructing the strings of equivalence class
 * numbers is relatively straightforward.  The trick comes in deciding how
 * to paint them.  If we simply choose one member from each equivalence
 * class, we'd get a bunch of vertical streaks in the image.  Instead, we
 * use another table to re-dither the members of each class, much like
 * halftoning.  The dither values for a class are chosen from the members
 * of that class.  We also make some attempt to dither clumps of 2 dots
 * differently than single dots on the assumption that each printed dot
 * was scanned as two pixels.
 *
 * Code at this lower level only encodes and decodes strings of equivalence
 * class numbers.  Code at the next higher level turns bytes into
 * equivalence class IDs on compression and redithers these IDs on
 * decompression.
 *
 *			*	*	*
 *
 * The decoding loop is a state machine, much like the encoding
 * loop.  In fact, we'll try to make them as similar as possible
 * to reduce the chance of error.  In the encoder, one trip around
 * the loop produces one copy string length, one literal string length
 * and the codes for a string of literal bytes.  We do not encode
 * a literal string if the copy string filled the coding line.
 *
 * In the decoder, each trip through the loop processes a copy string
 * followed by a literal string.  Again, we stop before the literal
 * string if the copy string filled the coding line.  Once the line
 * is decoded, we use the re-dithering table to paint the scan line.
 *
 */

/* PRIVATE */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress 1 row of a binary image using Lossy Dither decompression.
 *
 * LongDescription:::
 *  Decompress one row of an image using the Lossy Dither algorithm.
 *  We will stop when either an error is encountered, we finish the row
 *  or we empty the compressed data buffer.
 *
 * Input:::
 *  pLDDecompInfo::         Pointer to private LD compression state block
 *  widthInBytes::          The width of the image (or tile or strip),
 *			    not including frame, in units of bytes.
 *  flushPipeline::	    If cTrue, this routine should be able to
 *			    get to the end of the scan line using
 *			    only data in the compressed data buffer.
 *			    If set to cFalse, it's ok to stop and
 *			    ask for more compressed data when we're
 *			    starting to run out.  Additional data will
 *			    be seamlessly concatenated to current data.
 *  pCompDataEnd::          pointer to byte just beyond end of compressed data.
 *
 * Output:::
 *  ppCompData::            Pointer to pointer to beginning of compressed
 *			    data block.  The data block pointer is updated
 *			    at the end of this routine.
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid,
 *  ia_bufferExceeded => ran out of bytes in compressed data buffer.
 *
 * Notes:::
 *  The compressed data stream is always in natural byte order.
 *  The coding and reference lines are also both in natural byte order
 *  since they were stored that way by higher level code.
 *		
 *  If the end of the scan line is reached, set
 *  pLDDecompInfo->decompState = cLDDecompNewLine.
 *  If we stop because the compressed data buffer empties,
 *  set pLDDecompInfo->decompState to the state which
 *  describes where we should restart and
 *  put the relevant byte counts and run pointers into
 *  *pLDDecompInfo.
 *
 *  In any case, *ppCompData is set to point to the next
 *  byte of compressed data to be read.  Callers can use this
 *  to determine the number of bytes read so far.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
ip_decompLDRow (
    IPState    *pIPState,
    LDDecompressionInfo *pLDDecompInfo,
    Int32	widthInBytes,
    Bool	flushPipeline,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    UInt8	*pFirstDataByte, *pFirstRefByte; /* ptrs to first bytes in
					 * coding and reference lines */
    Int32	*markerTable;		/* pointer to first word in marker
					 * table.  The marker table contains
					 * the offsets of the ends of copy
					 * strings in the reference line. */
    Int32	 dataIndex;		/* index of cur byte in coding line */
    Int32	 endIndex;		/* index of byte just beyond end
					 * of coding line.  This is just
					 * a renamed version of widthInBytes. */
    Int32	 codeVal;		/* value found in decoding tables */
    UInt32	 codeLen;		/* length of huffman code */
    Int32	 markerIndex;		/* index of next marker location */
    UInt32	 accum;			/* accumulated encoded bits */
    Int32	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */
    Int32	 remainingLength;	/* used when we stop in the middle
					 * of encoding a long thing. */
    Bool	 goAgain;		/* Loop control var */
    mmrTableEntry *pNonTermTable;	/* points to second level tables
					 * of non-terminating codes.
					 */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pLDDecompInfo->accumBits;
    vacantBit = pLDDecompInfo->vacantBit;
    dataIndex = pLDDecompInfo->dataIndex;
    codeLen = pLDDecompInfo->codeLen;
    remainingLength = pLDDecompInfo->remainingLength;
    endIndex = widthInBytes;
    markerIndex = pLDDecompInfo->markerIndex;
    markerTable = (Int32 *) ((UInt8 *) pLDDecompInfo +
			     pLDDecompInfo->markerOffset);

	/* Most of time time, we want to compare to
	 * pCompDataEnd-1, so decrement it here */
    pCompDataEnd--;

	/* point to current byte by first pointing to first real byte,
	 * To do this, we figure out which byte array is the coding line and
	 * which is the reference line.  pLDDecompInfo->line1IsCoding
	 * is set when the line1 is the coding line.
	 */
    if (pLDDecompInfo->line1IsCoding)
    {
	pFirstDataByte = ((UInt8 *)(pLDDecompInfo) +
						pLDDecompInfo->line1Offset);
	pFirstRefByte = ((UInt8 *)(pLDDecompInfo) +
						pLDDecompInfo->line0Offset);
    }
    else	/* the first run array holds the coding line */
    {
	pFirstDataByte = ((UInt8 *)(pLDDecompInfo) +
						pLDDecompInfo->line0Offset);
	pFirstRefByte = ((UInt8 *)(pLDDecompInfo) +
						pLDDecompInfo->line1Offset);
    }


	/* see where we should start off */
    switch (pLDDecompInfo->decompState)
    {
    case cLDDecompNewLine:
	    /* Start at the beginning of the scan line.
	     */
	if (dataIndex >= endIndex)
	    goto decompLDGoodCompletion;
	else
	    break;

    case cLDDecompDecodeCopyLength:
	goto decompLDDecodeCopyLength;

    case cLDDecompDecodeLongCopyLength:
	goAgain = cTrue;
	goto decompLDDecodeLongCopyLength;

    case cLDDecompDecodeLiteralLength:
	goto decompLDDecodeLiteralLength;

    case cLDDecompDecodeLongLiteralLength:
	goAgain = cTrue;
	goto decompLDDecodeLongLiteralLength;

    case cLDDecompDecodeLiteralString:
	goto decompLDDecodeLiteralString;

    default:	/* Restart state belongs to some other routine */
	NORMAL_RETURN;
    }

	/* ********************* MAIN LOOP ***********************/

	/* Here's the main loop.  In it, we decode a copy string
	 * length.  If it's not 0, we copy a string of coding
	 * bytes from the string of reference bytes immediately above
	 * it using this length.  If the first length is 0,
	 * we just skip the copy step.  If, after copying,
	 * we're not at the end of a scan line, we
	 * then decode a literal string length and the specified
	 * number of literal values, placing each decoded
	 * equivalence class ID into the subsequent coding line
	 * byte.
	 */
    while (dataIndex < endIndex)
    {
	    /* First, update the index into the marker table.  We
	     * want to stop in front of the first marker that's
	     * to the right of our current position.  There's
	     * always a dummy marker off the right edge of the
	     * image, so we'll hit that if nothing else appears.
	     */
	while (markerTable[markerIndex] <= dataIndex)
	    markerIndex++;

	   /* ---------------- DECODE COPY STRING LENGTH -------------- */

decompLDDecodeCopyLength:
	    /* Now, see if the accumulator has room for at least more
	     * two bytes of data.  If so, read them.  If not,
	     * read one if possible and quit.  This is also
	     * where we restart if the buffer was empty last time
	     * we got here.
	     */
	if (vacantBit <= 16)
	{
	    if (pCurCompData < pCompDataEnd)
	    {
		accum |= reverseByte[pCurCompData[0]] << vacantBit;
		vacantBit += 8;
		accum |= reverseByte[pCurCompData[1]] << vacantBit;
		vacantBit += 8;
		pCurCompData += 2;
	    }
	    else
	    {
		    /* We're not allowed to fetch 2 new bytes.
		     * If there's one byte, get it.
		     */
		if (pCurCompData == pCompDataEnd)
		{
		    /* There's one more byte in the data stream. Get it. */
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData ++;
		}
		if (!flushPipeline)
		    goto abortDecodeCopyLength;
	    }
	}

	    /* Now decode the first thing in the accumulator.  This
	     * should be a copy string length.  First, get the length
	     * of this code.  If it's less than or equal to 8,
	     * we got it all.
	     */
	codeLen =
	    ldcLevel1CopyLengthDecodes[(accum & cLevel1CopyDecodeMask) +
				       cLevel1CopyLengthOffset];

	    /* See if it could be decoded using that first table lookup.
	     */
	if (codeLen <= cLevel1CopyDecodeMaxLen)
	{
	    codeVal = 
		ldcLevel1CopyLengthDecodes[accum & cLevel1CopyDecodeMask];
	}
	else	/* it was a long copy code.  However, all long
		 * codes fit into 16 bits, so we're still save */
	{
		/* We can safely discard the first 4 code bits, all
		 * codes longer than 8 bits has these set to 0000.
		 */
	    accum >>= 4;
	    vacantBit -= 4;

		/* Get the address of the non-terminating table
		 * pointed to by next 4 bits of code. */
	    pNonTermTable = ldcLevel2CopySelectorTable[accum & 0xF];
	    if (pNonTermTable == NULL)
		goto decompLDDecodeFailure;

		/* We got a valid table pointer using bits (7..4).
		 * Advance past them and decode bits (11..8), which
		 * almost always gets us to the end.
		 */
	    accum >>= 4;
	    vacantBit -= 4;
	    codeLen = pNonTermTable[accum & 0xF].codeLength;

		/* There are two 13-bit codes.  If the length of
		 * this code is 5, we're on one of them.  Read the
		 * code by looking at the 13th bit.  Otherwise,
		 * just use the bits we have now.
		 */
	    if (codeLen > 4)
	    {
		if ((accum & 0x10) == 0)
		    codeVal = 63;
		else
		    codeVal = 66;
	    }
	    else
	    {
		codeVal = pNonTermTable[accum & 0xF].codeBits;
	    }
	}
	    /* Shift out terminating portion of the length code */
	accum >>= codeLen;
	vacantBit -= codeLen;

	    /* You'd think that we could just copy the specified
	     * number of bytes from the reference line to the
	     * coding line, but naaah.
	     * There are a bunch of special values for the copy
	     * string codes that have to sorted out first.  We need
	     * to turn these special codes into lengths before
	     * doing the real copying.
	     */
	if (codeVal >= cLDMediumCopyLenPrefix)
	{
	    switch (codeVal - cLDMediumCopyLenPrefix)
	    {
	    case cLDMediumCopyLenPrefix - cLDMediumCopyLenPrefix:
		    /* this is a prefix code for copy lengths of
		     * 64 through 127.  The (number - 64) is encoded in
		     * the next 6 bits of accum (bit reversed).
		     */
		codeVal = (reverseByte[accum & cLDMediumLengthBitMask] >>
				    (8 - cLDMediumCopyLengthBits)) +
			  cLDShortCopyLength + 1;
		accum >>= cLDMediumCopyLengthBits;
		vacantBit -= cLDMediumCopyLengthBits;
		break;

	    case cLDLongCopyLenPrefix - cLDMediumCopyLenPrefix:
		    /* this is a prefix code for copy lengths of
		     * 128 through 255.  The (number - 128) is encoded in
		     * the next 7 bits of accum (bit reversed).
		     */
		codeVal = (reverseByte[accum & cLDLongLengthBitMask] >>
				    (8 - cLDLongCopyLengthBits)) +
			  cLDMediumCopyLength + 1;
		accum >>= cLDLongCopyLengthBits;
		vacantBit -= cLDLongCopyLengthBits;
		break;

	    case cLDVeryLongCopyLenPrefix - cLDMediumCopyLenPrefix:
		    /* this is a prefix code for copy lengths
		     * greater than 255.  The (number - 255) is encoded in
		     * the concatenation of the low order 7 bits of
		     * the sequence of bytes with their MS bits
		     * on (if any) concatenated with the LS 7 bits
		     * of the last byte of the sequence.  The LS
		     * bits of the count are encoded first.  All
		     * counts should be bit-reversed.
		     */
		remainingLength = 0;
		codeLen = 0;
		goAgain = cTrue;
		while (goAgain)
		{
decompLDDecodeLongCopyLength:
			/* But First, see if the accumulator has room
			 * for at least more two bytes of data.  If so,
			 * read them.  If not, read one if possible and
			 * quit.  This is also where we restart if the
			 * buffer was empty last time we got here.
			 */
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
				reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |= 
				reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data
				 * stream. Get it. */
				accum |= 
				    reverseByte[pCurCompData[0]] <<
								vacantBit;
				vacantBit += 8;
				pCurCompData ++;
			    }
			    if (!flushPipeline)
				goto abortDecodeLongCopyLength;
			}
		    }
			/* if this is the first time through, we reverse
			 * the order of the bits in the LS byte of accum,
			 * then it, discarding the reversed flag big as
			 * the first entry in remainingLength.  If this
			 * after that, we shift the reversed, masked bits
			 * left by codeLen, before inserting the result
			 * into the accumulting remainingLength.
			 * The LS bits of the count are sent
			 * first, the MS bits last.
			 */
		    remainingLength |=
			    (reverseByte[accum & 0xFF] & 0x7F) << codeLen;

		    codeLen += 7;		/* prepare to insert next
						 * bits in sequence. */
			/* The REVERSED flag bit is in the LS bit of accum */
		    if ((accum & 0x1) == 0)
			goAgain = cFalse;	/* check this now before
						 * throwing it away below */

		    accum >>= cLDVeryLongCopyLengthBits;
		    vacantBit -= cLDVeryLongCopyLengthBits;
		}
		    /* add the very long offset in at the end after
		     * all the concatenating.
		     */
		codeVal = remainingLength + cLDLongCopyLength + 1;
		break;

	    case cLDCopyLenToEOL - cLDMediumCopyLenPrefix:
		    /* the copy length goes to the end of the scan line */
		codeVal = endIndex - dataIndex;
		break;

	    case cLDCopyLenToMarker0M1 - cLDMediumCopyLenPrefix:
		    /* the copy length goes through one byte before next
		     * marker.
		     */
		codeVal = (markerTable[markerIndex] - dataIndex - 1) + 1;
		break;

	    case cLDCopyLenToMarker0 - cLDMediumCopyLenPrefix:
		    /* the copy length goes through next marker.
		     */
		codeVal = (markerTable[markerIndex] - dataIndex) + 1;
		break;

	    case cLDCopyLenToMarker0P1 - cLDMediumCopyLenPrefix:
		    /* the copy length goes through one byte after next
		     * marker.
		     */
		codeVal = (markerTable[markerIndex] - dataIndex + 1) + 1;
		break;

	    case cLDCopyLenToMarker1 - cLDMediumCopyLenPrefix:
		    /* the copy length goes through second marker.
		     */
		codeVal = (markerTable[markerIndex+1] - dataIndex) + 1;
		break;

	    default:
		goto decompLDDecodeFailure;
	    }
	}

	    /* Yea!  At this point, codeVal contains the real number
	     * of bytes to be copied from the reference line to
	     * the coding line.  Do it, then update dataIndex.
	     */
	memcpy((void *)(&pFirstDataByte[dataIndex]),
	       (void *)(&pFirstRefByte[dataIndex]),
	       codeVal);

	dataIndex += codeVal;	/* point to spot where literal
				     * string should be written.
				     */
	if (dataIndex >= endIndex)
	{
	    break;			/* we're done with the line */
	}



	   /* ---------------- DECODE LITERAL STRING LENGTH -------------- */

decompLDDecodeLiteralLength:
	    /* Now, see if the accumulator has room for at least more
	     * two bytes of data.  If so, read them.  If not,
	     * read one if possible and quit.  This is also
	     * where we restart if the buffer was empty last time
	     * we got here.
	     */
	if (vacantBit <= 16)
	{
	    if (pCurCompData < pCompDataEnd)
	    {
		accum |= reverseByte[pCurCompData[0]] << vacantBit;
		vacantBit += 8;
		accum |= reverseByte[pCurCompData[1]] << vacantBit;
		vacantBit += 8;
		pCurCompData += 2;
	    }
	    else
	    {
		    /* We're not allowed to fetch 2 new bytes.
		     * If there's one byte, get it.
		     */
		if (pCurCompData == pCompDataEnd)
		{
		    /* There's one more byte in the data stream. Get it. */
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData ++;
		}
		if (!flushPipeline)
		    goto abortDecodeLiteralLength;
	    }
	}

	    /* Now decode the next thing in the accumulator.  This
	     * should be a literal string length.  First, get the length
	     * of this code.  If it's less than or equal to 8,
	     * we got it all.
	     */
	codeLen =
	    ldcLevel1LiteralLengthDecodes[(accum & cLevel1LiteralDecodeMask) +
				          cLevel1LiteralLengthOffset];

	    /* See if it could be decoded using that first table lookup.
	     */
	if (codeLen <= cLevel1LiteralDecodeMaxLen)
	{
	    codeVal = 
		ldcLevel1LiteralLengthDecodes[accum & cLevel1LiteralDecodeMask];
	}
	else	/* it was a long copy code.  However, all long
		 * codes fit into 16 bits, so we're still save */
	{

		/* We can safely discard the first 4 code bits, all
		 * codes longer than 8 bits has these set to 0000.
		 */
	    accum >>= 4;
	    vacantBit -= 4;

		/* Get the address of the non-terminating table
		 * pointed to by next 4 bits of code. */
	    pNonTermTable = ldcLevel2LiteralSelectorTable[accum & 0xF];
	    if (pNonTermTable == NULL)
		goto decompLDDecodeFailure;

		/* We got a valid table pointer using bits (7..4).
		 * Advance past them and look at bits (11..8).
		 * These either are the last bits in the code
		 * or they're not.  We decide by trying to look
		 * up their lengths.
		 */
	    accum >>= 4;
	    vacantBit -= 4;

	    if (pNonTermTable[accum & 0xF].codeLength > 4)
	    {
		    /* ooops, we need one last level of decode.
		     * use bits 11..8 to get a pointer to the last
		     * level of decode table, then discard them.
		     */
		pNonTermTable = ldcLevel3LiteralSelectorTable[accum & 0xF];
		if (pNonTermTable == NULL)
		    goto decompLDDecodeFailure;

		accum >>= 4;
		vacantBit -= 4;
	    }

		/* Get the length and value of the code */
	    codeLen = pNonTermTable[accum & 0xF].codeLength;
	    codeVal = pNonTermTable[accum & 0xF].codeBits;
	}

	    /* Shift out terminating portion of the length code */
	accum >>= codeLen;
	vacantBit -= codeLen;

	    /* You'd think that we could just copy the specified
	     * number of bytes from the compressed data to the
	     * coding line, but naaah.
	     * There are a bunch of special values for the literal
	     * string codes that have to sorted out first.  We need
	     * to turn these special codes into lengths before
	     * decoding the real literal codes.
	     */
	if (codeVal >= cLDMediumLitLenPrefix)
	{
	    switch (codeVal - cLDMediumLitLenPrefix)
	    {
	    case cLDMediumLitLenPrefix - cLDMediumLitLenPrefix:
		    /* this is a prefix code for copy lengths of
		     * 65 through 128.  The (number - 65) is encoded in
		     * the next 6 bits of accum (bit reversed).
		     */
		codeVal = (reverseByte[accum & cLDMediumLengthBitMask] >>
				    (8 - cLDMediumLitLengthBits)) +
			  cLDShortLitLength + 1;
		accum >>= cLDMediumLitLengthBits;
		vacantBit -= cLDMediumLitLengthBits;
		break;

	    case cLDLongLitLenPrefix - cLDMediumLitLenPrefix:
		    /* this is a prefix code for copy lengths of
		     * 129 through 256.  The (number - 129) is encoded in
		     * the next 7 bits of accum (bit reversed).
		     */
		codeVal = (reverseByte[accum & cLDLongLengthBitMask] >>
				    (8 - cLDLongLitLengthBits)) +
			  cLDMediumLitLength + 1;
		accum >>= cLDLongLitLengthBits;
		vacantBit -= cLDLongLitLengthBits;
		break;

	    case cLDVeryLongLitLenPrefix - cLDMediumLitLenPrefix:
		    /* this is a prefix code for copy lengths
		     * greater than 256.  The (number - 256) is encoded in
		     * the concatenation of the low order 7 bits of
		     * the sequence of bytes with their MS bits
		     * on (if any) concatenated with the LS 7 bits
		     * of the last byte of the sequence.  The LS
		     * bits of the count are encoded first.  All
		     * counts should be bit-reversed.
		     */
		remainingLength = 0;
		codeLen = 0;
		goAgain = cTrue;
		while (goAgain)
		{
decompLDDecodeLongLiteralLength:
			/* But First, see if the accumulator has room
			 * for at least more two bytes of data.  If so,
			 * read them.  If not, read one if possible and
			 * quit.  This is also where we restart if the
			 * buffer was empty last time we got here.
			 */
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
				reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |= 
				reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data
				 * stream. Get it. */
				accum |= 
				    reverseByte[pCurCompData[0]] <<
								vacantBit;
				vacantBit += 8;
				pCurCompData ++;
			    }
			    if (!flushPipeline)
				goto abortDecodeLongLiteralLength;
			}
		    }
			/* if this is the first time through, we reverse
			 * the order of the bits in the LS byte of accum,
			 * then it, discarding the reversed flag big as
			 * the first entry in remainingLength.  If this
			 * after that, we shift the reversed, masked bits
			 * left by codeLen, before inserting the result
			 * into the accumulting remainingLength.
			 * The LS bits of the count are sent
			 * first, the MS bits last.
			 */
		    remainingLength |=
			(reverseByte[accum & 0xFF]  & 0x7F) << codeLen;

		    codeLen += 7;		/* prepare to insert next
						 * bits in sequence. */

			/* The REVERSED flag bit is in the LS bit of accum */
		    if ((accum & 0x1) == 0)
			goAgain = cFalse;	/* check this now before
						 * throwing it away below */

		    accum >>= cLDVeryLongLitLengthBits;
		    vacantBit -= cLDVeryLongLitLengthBits;
		}
		    /* add the very long offset in at the end after
		     * all the concatenating.
		     */
		codeVal = remainingLength + cLDLongLitLength + 1;
		break;

	    case cLDLitLenToEOL - cLDMediumLitLenPrefix:
		    /* the copy length goes to the end of the scan line */
		codeVal = endIndex - dataIndex;
		break;

	    default:
		goto decompLDDecodeFailure;
	    }
	}


	   /* ---------------- DECODE LITERAL STRING -------------- */

	    /* Yea!  At this point, codeVal contains the real number
	     * of equivalence class IDs to be decoded from the
	     * compressed data stream.  Put this into remainingLength
	     * so we can recover and start decoding.
	     */
	for (remainingLength = codeVal; remainingLength > 0; remainingLength--)
	{

decompLDDecodeLiteralString:
		/* Before getting (another) literal value,
		 * see if the accumulator has room for at least more
		 * two bytes of data.  If so, read them.  If not,
		 * read one if possible and quit.  This is also
		 * where we restart if the buffer was empty last time
		 * we got here.
		 */
	    if (vacantBit <= 16)
	    {
		if (pCurCompData < pCompDataEnd)
		{
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    accum |= reverseByte[pCurCompData[1]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData += 2;
		}
		else
		{
			/* We're not allowed to fetch 2 new bytes.
			 * If there's one byte, get it.
			 */
		    if (pCurCompData == pCompDataEnd)
		    {
			/* There's one more byte in the data stream. Get it. */
			accum |= reverseByte[pCurCompData[0]] << vacantBit;
			vacantBit += 8;
			pCurCompData ++;
		    }
		    if (!flushPipeline)
			goto abortDecodeLiteralString;
		}
	    }

		/* deocde another byte and store it into the coding line. */
	    pFirstDataByte[dataIndex] =
		ldcLiteralCodeDecodes[accum & cLiteralCodeDecodeMask];
	    dataIndex++;	/* point to next byte */
	    codeLen =
		ldcLiteralCodeDecodes[(accum & cLiteralCodeDecodeMask) +
				      cLiteralCodeOffset];
	    accum >>= codeLen;
	    vacantBit -= codeLen;
	}	/* end of loop that decodes literal strings */

    }	/* end of loop to decode one line */


	/* We get here when everything was swell.  We've decoded
	 * equivalence class IDs for an entire line.
	 * We never add padding in LD, so no need to do that.  We also
	 * don't flush the accumulator at the end of a scan line, so
	 * don't do that either.
	 * Just save the state so we know we're done with this scan
	 * line and can pick up with the codes for the next scan line
	 * when it comes.
	 */
decompLDGoodCompletion:
   *ppCompData = pCurCompData;
    pLDDecompInfo->dataIndex = 0;
    pLDDecompInfo->markerIndex = 0;
    pLDDecompInfo->codeLen = 0;
    pLDDecompInfo->remainingLength = 0;
    pLDDecompInfo->decompState = cLDDecompNewLine;
    pLDDecompInfo->accumBits = accum;
    pLDDecompInfo->vacantBit = vacantBit;

    NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

	/* We get here if we were about to have a decode failure.
	 * If flushPipeline is set and accum is 0, then we
	 * really just ran out of data, so return ia_bufferExceeded
	 * instead.  If the upper level sees an ia_bufferExceeded
	 * with flushPipeline, it knows that the client didn't
	 * supply enough compressed data and we're doomed.
	 */
decompLDDecodeFailure:
    if ((flushPipeline) && (accum == 0))
	return ia_bufferExceeded;
    else
	return ia_decodeFailure;


	/* Next, we have all the abort targets.  We do this weird
	 * thing of setting all the states here instead of above
	 * in the hopes that the compiler will generate better
	 * code in the real loops - just a test a jump down here
	 * where the normal case is straight ahead instead of
	 * a test and jump around the error stuff.
	 * We get here when we had at least two bytes to write
	 * to the compressed data stream, but there wasn't room
	 * for 2 bytes.  There might be room for 1 byte.
	 */

abortDecodeCopyLength:
    pLDDecompInfo->decompState = cLDDecompDecodeCopyLength;
    goto decompLDAbortFinish;
abortDecodeLongCopyLength:
    pLDDecompInfo->decompState = cLDDecompDecodeLongCopyLength;
    goto decompLDAbortFinish;
abortDecodeLiteralLength:
    pLDDecompInfo->decompState = cLDDecompDecodeLiteralLength;
    goto decompLDAbortFinish;
abortDecodeLongLiteralLength:
    pLDDecompInfo->decompState = cLDDecompDecodeLongLiteralLength;
    goto decompLDAbortFinish;
abortDecodeLiteralString:
    pLDDecompInfo->decompState = cLDDecompDecodeLiteralString;
    goto decompLDAbortFinish;

	/* This is where we end up when we abort because we ran out
	 * of bytes in the compression buffer and we weren't supposed
	 * to flush the compression pipeline.
	 */
decompLDAbortFinish:
   *ppCompData = pCurCompData;
    pLDDecompInfo->dataIndex = dataIndex;
    pLDDecompInfo->markerIndex = markerIndex;
    pLDDecompInfo->codeLen = codeLen;
    pLDDecompInfo->remainingLength = remainingLength;
    pLDDecompInfo->accumBits = accum;
    pLDDecompInfo->vacantBit = vacantBit;

    return ia_bufferExceeded;	/* It ran out of compressed input data */
}


/*

 $Log: mildd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.3  1998/01/13 23:52:40  ddavies
 mildd.c -> /product/ipcore/ipshared/src/RCS/mildd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1997/03/18 20:11:36  ddavies
 mildd.c -> /product/ipcore/ipshared/src/RCS/mildd.c,v

 Include string.h to get prototype for memcpy.

 Revision 1.1  1997/03/17 19:49:10  ddavies
 Initial revision


 */


