/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mildut.c: utility routines for Lossy Dithered compression.
 *
 *	Int32	ip_writeLDEOL()
 *			Insert an EndOfLine code into the data stream.
 *	Int32	ip_flushLD()
 *			Flush encoding pipeline at end of large block.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "ldc.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"
#include "shrcmprs.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: mildut.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* This is a set of utility routines used in Lossy Dithered compression.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write out end of line code for Lossy Dither Compression.
 *
 * LongDescription:::
 *  This procedure writes out the end of line code for Lossy Dither
 *  compression.
 *
 * Input:::
 *  pLDCompInfo::           Pointer to private LD compression state block
 *  ppCompData::            Pointer to pointer to beginning of compressed
 *			    data block.  The data block pointer is updated
 *			    at the end of this routine.
 *  pCompDataEnd::          pointer to byte just beyond end of compressed data.
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid,
 *  ia_bufferExceeded => ran out of space in compressed data buffer.
 *
 * Notes:::
 *  If pLDCompInfo->compType == cLDCompNewLine (we're really
 *  at the end of a line), attempt to output an EOL code.
 *
 *  If we stop because the compressed data buffer fills,
 *  set pLDCompInfo->compState to the state which
 *  describes the compressed data byte we're about to write and
 *  put the relevant byte counts and byte pointers into
 *  *pLDCompInfo.
 *
 *  In any case, *ppCompData is set to point to the next
 *  byte of compressed data to be written.  Callers can use this
 *  to determine the number of bytes written so far.
 *		
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
ip_writeLDEOL (
    IPState    *pIPState,
    LDCompressionInfo *pLDCompInfo,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    UInt32	 accum;			/* accumulated encoded bits */
    Int32	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pLDCompInfo->accumBits;
    vacantBit = pLDCompInfo->vacantBit;

	/* Start off in the proper state */
    switch (pLDCompInfo->compState)
    {
    case cLDCompNewLine:
	goto compLDStartEOL;

    case cLDCompFinishEOL:
	goto compLDStartEOL;

    default:	/* restart state belongs to some other routine */
	NORMAL_RETURN;
    }

compLDStartEOL:
    if (vacantBit >= 16)
    {
	if (pCurCompData + 1 >= pCompDataEnd)
	{
	    pLDCompInfo->compState = cLDCompFinishEOL;
	    goto compLDEOLMightWriteOne;
	}

	    /* write out LS 2 bytes of accumulator and increment
	     * pointers */
	pCurCompData[0] = reverseByte[(UInt8) accum];
	pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
	accum >>= 16;			/* get rid of written bytes */
	pCurCompData += 2;
	vacantBit -= 16;
    }

	/* Having inserted the pad if necessary, insert the EOL itself */
    accum |= (cEOLCode) << vacantBit;
    vacantBit += cEOLLength;

	/* Swell, we've inserted the EOL.  Update the saved state do we'll
	 * remember.
	 */
   *ppCompData = pCurCompData;
   pLDCompInfo->compState = cLDCompNewLine;
   pLDCompInfo->accumBits = accum;
   pLDCompInfo->vacantBit = vacantBit;

   NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

	/* We get here when we know that two bytes won't fit into the
	 * compressed output buffer.  Try putting one in.
	 */
compLDEOLMightWriteOne:
    if (pCurCompData < pCompDataEnd)
    {
	pCurCompData[0] = reverseByte[(UInt8) accum];
	pCurCompData++;
	accum >>= 8;
	vacantBit -= 8;
    }

	/* This is where we end up when we abort because we ran out
	 * of space in the compression buffer.
	 */
    *ppCompData = pCurCompData;
    pLDCompInfo->accumBits = accum;
    pLDCompInfo->vacantBit = vacantBit;
    
    return ia_bufferExceeded;	/* This is how we say that we weren't able
				 * to finish.
				 */

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Flush Lossy Dither Compression pipeline after encoding.
 *
 * LongDescription:::
 *  This routine simply attempts to empty the bits in the pipeline to
 *  the compressed data stream.  It is normally called at the end of 
 *  a large section, like a tile or strip.
 *
 * Input:::
 *  pLDCompInfo::           Pointer to private LD compression state block
 *  ppCompData::            Pointer to pointer to beginning of compressed
 *			    data block.  The data block pointer is updated
 *			    at the end of this routine.
 *  pCompDataEnd::          pointer to byte just beyond end of compressed data.
 *
 * Returns:::
 *  ia_successful => it worked; ia_invalidParm => parmeters were invalid,
 *  ia_bufferExceeded => ran out of space in compressed data buffer.
 *
 * Notes:::
 *  If we stop because the compressed data buffer fills,
 *  set pLDCompInfo->compState to the state which
 *  describes the compressed data byte we're about to write and
 *  put the relevant byte counts and byte pointers into
 *  *pLDCompInfo.
 *
 *  In any case, *ppCompData is set to point to the next
 *  byte of compressed data to be written.  Callers can use this
 *  to determine the number of bytes written so far.
 *		
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL 
ip_flushLD (
    IPState    *pIPState,
    LDCompressionInfo *pLDCompInfo,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    UInt32	 accum;			/* accumulated encoded bits */
    Int32	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pLDCompInfo->accumBits;
    vacantBit = pLDCompInfo->vacantBit;

	/* see where we should start off */
    switch (pLDCompInfo->compState)
    {
    case cLDCompNewLine:
	goto compLDSetFlushAmount;

    case cLDCompFinishTotalFlush:
	goto compLDFinishFlush;

    default:	/* not one of our restart codes */
	NORMAL_RETURN;
    }

compLDSetFlushAmount:
	/* Adjust vacantBit so we flush to the next byte boundary. */
    vacantBit = (vacantBit + 7) & ~0x7;

compLDFinishFlush:
	    /* flush the EOL */
    while (vacantBit > 0)
    {
	if (pCurCompData >= pCompDataEnd)
	    goto compLDAbortFlush;

	    /* write out LS byte of accumulator and increment
	     * pointers */
	pCurCompData[0] = reverseByte[(UInt8) accum];
	accum >>= 8;			/* get rid of written bytes */
	pCurCompData++;
	vacantBit -= 8;
    }

	/* Swell, we've flushed the accumulator.  Update the saved
	 * state do we'll remember.
	 */
   *ppCompData = pCurCompData;
   pLDCompInfo->compState = cLDCompNewLine;
   pLDCompInfo->accumBits = accum;
   pLDCompInfo->vacantBit = vacantBit;

   NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

compLDAbortFlush:
    pLDCompInfo->compState = cLDCompFinishTotalFlush;
	/* This is where we end up when we abort because we ran out
	 * of space in the compression buffer.
	 */
    *ppCompData = pCurCompData;
    pLDCompInfo->accumBits = accum;
    pLDCompInfo->vacantBit = vacantBit;
    
    return ia_bufferExceeded;	/* This is how we say that we weren't able
				 * to finish.
				 */

}



/*

 $Log: mildut.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:52:41  ddavies
 mildut.c -> /product/ipcore/ipshared/src/RCS/mildut.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1997/03/17 19:52:20  ddavies
 Initial revision


 */

