/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mimhc.c: low level code used in Modified Huffman and Modified Read
 *	     compression.
 *
 *	Int32	ip_compMHRow()
 *			Compresses one row of data.
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#define _NEED_HORIZONTAL_CODES_
#include "mmrenc.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mimhc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* The restart states for compression are listed in shrcmprs.prv.
 * cCCITTCompNewLine is already state 0.  The remainingLength, accumBits
 * and vacantBit in the state structure hold additional restart info.
 */

/* There's a big comment on our approach to Modified Huffman compression and
 * decompression in i_mhc.c.  You should read that before reading this.
 *
 * To review, MH and MR compression represent a binary image scan line as a
 * sequence of white runs and black runs.  The length of each run is
 * encoded.  One always starts with a white run, possibly of length 0.
 * One always encodes all the bits in the scan line.
 *
 * The difference between MH and MR compression occurs at the ends of
 * scan lines.  In MH, all scan lines are padded to byte or word
 * boundaries.  There are no explicit end of line codes.  In MR, there
 * are explicit end of line codes.  The sequence is:
 * <optional alignment gap><EOL><optional 1D vs 2D bit><CompressedDataLine1>
 * <optional alignment gap><EOL><optional 1D vs 2D bit><CompressedDataLine2>
 * ...
 * <optional alignment gap><EOL><optional 1D vs 2D bit><CompressedDataLineN>
 * <optional alignment gap><EOL><optional 1D vs 2D bit>
 *
 * Since the first data line is preceeded by an End of Line code and
 * the last line is followed by and End Of Line code, one can debate
 * whether these are really BeginningOfLine codes or EndOfLine codes.
 * The optional alignment gap is inserted when the encoder wishes
 * to guarantee that the EOL code ends on a byte boundary.
 *
 * The codes used to represent the run lengths were derived by fax
 * folks.  They seem to work fine.  There's one set of codes for
 * white runs and another set for black runs.  The white run codes
 * work a little better for long runs.  The black run codes are a
 * little better for short runs.
 *
 * Both sets of codes have terminating codes for runs shorter than 64
 * pixels and non-terminating codes for runs longer than that.  There's
 * a discussion of the codes in i_mhmrc.c.
 *
 * One thing that's interesting to note here is the restart mechanism.
 * When we fill the compression buffer (before running out of uncompressed
 * data), we set some state and return.  The calling routine allocates
 * more compression space and simply calls the same routine it called
 * before with new args.  We have to be able to figure out where to
 * restart, especially if a higher level routine calls more than one
 * lower routine that might restart.  So, each routine looks at the
 * restart states.  If the state belongs to that routine, it restarts.
 * If not, it simply returns happily so that the routine that owns
 * the state can be called.
 *
 * We might be called here to start a new line or complete an old one.
 * There's a switch statement at the top of the routine that dispatches
 * on the state.  After some setup, it jumps to the place in the normal
 * code where we should continue.
 *
 * The structure of the main code is a big loop.  It encodes a white
 * run, then encodes a black run, then loops back.  This lets us easily
 * reference the different tables needed to encode the two types of
 * runs.
 *
 * We stop when either we're done or when the compressed buffer is full.
 * If we're done, we return with pMHCompInfo->compState = cMHCompNewLine.
 * If we ran out of room in the compressed data buffer, we return with
 * pMHCompInfo->compState equal to one of the four states listed above.
 * the accumBits and vacantBit entries are set to reflect the current
 * state of the accumulator in any case.  The remainingLength entry
 * is set if the state is cMHCompFinishBlackRun or cMHCompFinishWhiteRun.
 * It holds the number that is still to be encoded.
 *
 * If the flushPipeline arg is set and we're done with the last run,
 * we empty the accumulator to the buffer and pad the last byte with
 * 0's if necessary.  This is the way MH lines end, though not they way
 * MR and MMR lines end.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_compMHRow()
 *
 *	This procedure compresses one line of data using the CCITT
 *	huffman tables.  The compressed data is in natural byte order.
 *	The uncompressed data is in pixr order.
 *
 *	Input:
 *		pCCITTCompInfo:	pointer to compression state.
 *		dummyMarker:	the value in the rightEnd of the
 *				dummy beyond the last valid run.
 *		padType:	If equal to cPadToByte, pad end of
 *				compresssed line to next byte boundary.
 *				If equal to cPadToWord, pad compresssed
 *				line to next 2-byte boundary.
 *				If equal to cNoPad, do not
 *				pad the end of the compressed line.
 *				Any padding is added after all
 *				encoded bits have been flushed.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *
 *	Action:
 *		Attempt to compress a single scan line of data.
 *		The compressed data is assumed to be in natural byte order and
 *		is arbitrarily aligned, so should be written 1 byte at a
 *		time.
 *
 *		If the end of the line is reached, check padType.
 *		If it contains cPadToByte, flush the accumulator and
 *		follow it with padding to the next byte boundary.
 *		If it's set to cPadToWord, flush the accumulator and
 *		pad to the next 2-byte boundary.  Otherwise, just set
 *		pCCITTCompInfo->compState = cCCITTCompNewLine;
 *		If we stop because the compressed data buffer fills,
 *		set pCCITTCompInfo->compState to the state which
 *		describes the compressed data byte we're about to write and
 *		put the relevant byte counts and run pointers into
 *		*pCCITTCompInfo.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be written.  Callers can use this
 *		to determine the number of bytes written so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pProc was NULL
 *		ia_bufferExceeded	stopped because we tried to write
 *					something to the compressed data
 *					buffer but it was full
 *
 **************************************************************************/
Int32  CDECL 
ip_compMHRow (
    IPState    *pIPState,
    CCITTCompressionInfo *pCCITTCompInfo,
    Int32	dummyMarker,
    UInt32	padType,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    Run32	*pCurRun, *pFirstRun;	/* pointer to current Run32 struct */
    UInt32	 accum;			/* accumulated encoded bits */
    Int32	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */
    Int32	 runLength;		/* length of run being encoded */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pCCITTCompInfo->accumBits;
    vacantBit = pCCITTCompInfo->vacantBit;
    runLength = pCCITTCompInfo->remainingLength;
	/* Most of time time, we want to compare to
	 * pCompDataEnd-1, so decrement it here */
    pCompDataEnd--;

	/* point to current run by first pointing to first run,
	 * then offseting to current run */
    pFirstRun = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				   (pCCITTCompInfo->runArrayOffset));
    pCurRun = pFirstRun + pCCITTCompInfo->currentRun;

	/* see where we should start off */
    switch (pCCITTCompInfo->compState)
    {
    case cCCITTCompNewLine:
	goto compMHNewLine;

    case cCCITTCompFinishLongWhite1:
	goto compMHFinishLongWhite1;

    case cCCITTCompFinishLongWhite2:
	goto compMHFinishLongWhite2;

    case cCCITTCompFinishLongWhite3:
	goto compMHFinishLongWhite3;

    case cCCITTCompFinishShortWhite:
	goto compMHFinishShortWhite;

    case cCCITTCompFinishLongBlack1:
	goto compMHFinishLongBlack1;

    case cCCITTCompFinishLongBlack2:
	goto compMHFinishLongBlack2;

    case cCCITTCompFinishLongBlack3:
	goto compMHFinishLongBlack3;

    case cCCITTCompFinishShortBlack:
	goto compMHFinishShortBlack;

    case cCCITTCompFinishFlushLine:
	goto compMHFinishFlushLine;

    default:	/* Restart state belongs to some other routine */
	NORMAL_RETURN;
    }

compMHNewLine:

	/* ********************* MAIN LOOP ***********************/

	/* Here's the main loop.  We encode a white run, then a
	 * black run.  We stop when we reach the end of the scan
	 * line or run out of space in the compressed data buffer.
	 */
    while (cTrue)
    {

	    /* ************* WHITE RUN ************** */
	runLength = pCurRun[1].runLeft - pCurRun[0].runRight - 1;
	if ((runLength == 0) && (pCurRun[1].runRight == dummyMarker))
	    break;	/* quit if there's no white run to encode */

	if (runLength >= 64)
	{
		/************* LONG WHITE RUN ************/

		/* write out the multiples of 2560 */
	    while (runLength >= 2560)
	    {
		    /* write out any full bytes */
compMHFinishLongWhite1:
		if (vacantBit >= 16)
		{
		    if (pCurCompData >= pCompDataEnd)
			goto abortLongWhite1;

			/* write out LS 2 bytes of accumulator and increment
			 * pointers */
		    pCurCompData[0] = reverseByte[(UInt8) accum];
		    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		    accum >>= 16;		/* get rid of written bytes */
		    pCurCompData += 2;
		    vacantBit -= 16;
		}

		accum |= mmrRevWhiteNonTermCodes[2560/64].codeBits << vacantBit;
		vacantBit += mmrRevWhiteNonTermCodes[2560/64].codeLength;
		runLength -= 2560;
	    }
    
		/* Write out codes needed to get us to the next multiple of 64
		 * if necessary */
	    if (runLength >= 64)
	    {
		    /* write out any full bytes */
compMHFinishLongWhite2:
		if (vacantBit >= 16)
		{
		    if (pCurCompData >= pCompDataEnd)
			goto abortLongWhite2;
	
			/* write out LS 2 bytes of accumulator and increment
			 * pointers */
		    pCurCompData[0] = reverseByte[(UInt8) accum];
		    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		    accum >>= 16;		/* get rid of written bytes */
		    pCurCompData += 2;
		    vacantBit -= 16;
		}

		accum |=
		    mmrRevWhiteNonTermCodes[runLength/64].codeBits << vacantBit;
		vacantBit +=
		    mmrRevWhiteNonTermCodes[runLength/64].codeLength;
		runLength &= 0x3F;		/* only len MOD 64 left */
	    }

		/* write out any full bytes */
compMHFinishLongWhite3:
	    if (vacantBit >= 16)
	    {
		if (pCurCompData >= pCompDataEnd)
		    goto abortLongWhite3;

		    /* write out LS 2 bytes of accumulator and increment
		     * pointers */
		pCurCompData[0] = reverseByte[(UInt8) accum];
		pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		accum >>= 16;			/* get rid of written bytes */
		pCurCompData += 2;
		vacantBit -= 16;
	    }

		/* Write the terminating run code */
	    accum |= mmrRevWhiteTermCodes[runLength].codeBits << vacantBit;
	    vacantBit += mmrRevWhiteTermCodes[runLength].codeLength;
	}
	else	/********** SHORT WHITE RUN ************/
	{
		/* write out any full bytes */
compMHFinishShortWhite:
	    if (vacantBit >= 16)
	    {
		if (pCurCompData >= pCompDataEnd)
		    goto abortShortWhite;
    
		    /* write out LS 2 bytes of accumulator and increment
		     * pointers */
		pCurCompData[0] = reverseByte[(UInt8) accum];
		pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		accum >>= 16;			/* get rid of written bytes */
		pCurCompData += 2;
		vacantBit -= 16;
	    }
		/* Write the terminating white run code for the first run */
	    accum |= mmrRevWhiteTermCodes[runLength].codeBits << vacantBit;
	    vacantBit += mmrRevWhiteTermCodes[runLength].codeLength;
	}

	    /****************** BLACK RUN *****************/

	pCurRun += 1;		/* point to the new black run */
	if (pCurRun->runRight == dummyMarker)
	    break;		/* we're done */

	runLength = pCurRun->runRight - pCurRun->runLeft + 1;
	if (runLength >= 64)
	{
		/************ LONG BLACK RUN ***************/

		/* write out the multiples of 2560 */
	    while (runLength >= 2560)
	    {
		    /* write out any full bytes */
compMHFinishLongBlack1:
		if (vacantBit >= 16)
		{
		    if (pCurCompData >= pCompDataEnd)
		        goto abortLongBlack1;
	
			/* write out LS 2 bytes of accumulator and increment
			 * pointers */
		    pCurCompData[0] = reverseByte[(UInt8) accum];
		    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		    accum >>= 16;		/* get rid of written bytes */
		    pCurCompData += 2;
		    vacantBit -= 16;
		}

		accum |= mmrRevBlackNonTermCodes[2560/64].codeBits << vacantBit;
		vacantBit += mmrRevBlackNonTermCodes[2560/64].codeLength;
		runLength -= 2560;
	    }
    
		/* Write out codes needed to get us to the next multiple of 64
		 * if necessary */
	    if (runLength >= 64)
	    {
		    /* write out any full bytes */
compMHFinishLongBlack2:
		if (vacantBit >= 16)
		{
		    if (pCurCompData >= pCompDataEnd)
			goto abortLongBlack2;
	
			/* write out LS 2 bytes of accumulator and increment
			 * pointers */
		    pCurCompData[0] = reverseByte[(UInt8) accum];
		    pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		    accum >>= 16;		/* get rid of written bytes */
		    pCurCompData += 2;
		    vacantBit -= 16;
		}

		accum |=
		    mmrRevBlackNonTermCodes[runLength/64].codeBits << vacantBit;
		vacantBit +=
		    mmrRevBlackNonTermCodes[runLength/64].codeLength;
		runLength &= 0x3F;		/* only len MOD 64 left */
	    }
    
		/* write out any full bytes */
compMHFinishLongBlack3:
	    if (vacantBit >= 16)
	    {
		if (pCurCompData >= pCompDataEnd)
		    goto abortLongBlack3;
    
		    /* write out LS 2 bytes of accumulator and increment
		     * pointers */
		pCurCompData[0] = reverseByte[(UInt8) accum];
		pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		accum >>= 16;			/* get rid of written bytes */
		pCurCompData += 2;
		vacantBit -= 16;
	    }

		/* Write the terminating run code */
	    accum |= mmrRevBlackTermCodes[runLength].codeBits << vacantBit;
	    vacantBit += mmrRevBlackTermCodes[runLength].codeLength;
	}
	else
	{
		/************* SHORT BLACK RUN *****************/

		/* write out any full bytes */
compMHFinishShortBlack:
	    if (vacantBit >= 16)
	    {
		if (pCurCompData >= pCompDataEnd)
		    goto abortShortBlack;

		    /* write out LS 2 bytes of accumulator and increment
		     * pointers */
		pCurCompData[0] = reverseByte[(UInt8) accum];
		pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		accum >>= 16;			/* get rid of written bytes */
		pCurCompData += 2;
		vacantBit -= 16;
	    }

		/* Write the terminating black run code for the first run */
	    accum |= mmrRevBlackTermCodes[runLength].codeBits << vacantBit;
	    vacantBit += mmrRevBlackTermCodes[runLength].codeLength;
	}
    }

	/* Swell!  We've encoded all the runs.  See if we're supposed
	 * to pad the line to a word boundary.  This adjustment to
	 * vacantBit may make it larger than 32, but that's ok.
	 * The bits above the MS bit are all 0's anyway.
	 */
    switch (padType)
    {
    case cPadToByte:
	     /* There several cases here:
	      * 1. vacantBit == 0:
	      *		Do nothing.
	      * 2. 8 >= vacantBit > 0:
	      *		set vacantBit = 8.
	      * 3. 15 >= vacantBit > 8:
	      *		set vacantBit = 16.
	      * The actual writing of accum will be done below.
	      * Note that since we always pad with 0's and
	      * we always shift 0's down from the MS end of
	      * accum, we don't have to supply the pad 0's
	      * explicitly - they're already there.
	      */
	if (vacantBit > 0)
	    vacantBit = ((vacantBit-1) & ~0x7) + 8;
	break;

    case cPadToWord:
	    /* There are several different cases here.  Note that
	     * the Leffler code decides whether it is word aligned
	     * by simply looking at the LS bit of the current
	     * pointer to the compressed data, so we'll do the
	     * same thing.
	     * 1. pCurCompData is even and vacantBit == 0:
	     *		do nothing
	     * 2. pCurCompData is even and 15 >= vacantBit > 0:
	     *		set vacantBit = 16
	     * 3. pCurCompData is odd and 8 >= vacantBit >= 0:
	     *		set vacantBit = 8
	     * 4. pCurCompData is odd and 15 >= vacantBit > 8:
	     *		set vacantBit = 24
	     */
	runLength = ((UInt32)pCurCompData & 0x1);	/* get LS bit of ptr */
	if ((runLength == 0) && (vacantBit > 0))
	    vacantBit = 16;
	else
	if ((runLength == 1) && (vacantBit <= 8))
	    vacantBit = 8;
	else
	if (runLength == 1)	/* vacantBit > 8 */
	    vacantBit = 24;
	break;

    default:	/* no padding required */
        break;
    }

	/* If there are new bits to be flushed because of the padding, do it.
	 * Note that if the padding was set to cPadToByte or cPadToWord,
	 * vacantBit is now a multiple of 8, so we'll flush exactly the
	 * right amount.  If no padding was needed, we'll leave some bits
	 * in the last byte unless the number of bits happened to be an
	 * exact multiple of 8.  In either case, the next encoded bit will be
	 * seamless concatenated with these encoded bits.
	 * First, restore pCompDataEnd.  Remember that
	 * we decremented it at the beginning of the scan line.
	 */
compMHFinishFlushLine:
    pCompDataEnd++;
    while (vacantBit >= 8)
    {
	if (pCurCompData == pCompDataEnd)
	    goto abortFlushLine;

	pCurCompData[0] = reverseByte[(UInt8) accum];
	pCurCompData++;
	accum >>= 8;
	vacantBit -= 8;
    }

	/* Set up the saved state so we will know we're done */
   *ppCompData = pCurCompData;
   pCCITTCompInfo->currentRun = 0;
   pCCITTCompInfo->compState = cCCITTCompNewLine;
   pCCITTCompInfo->accumBits = accum;
   pCCITTCompInfo->vacantBit = vacantBit;

   NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

	/* First, we have all the abort targets.  We do this weird
	 * thing of setting all the states here instead of above
	 * in the hopes that the compiler will generate better
	 * code in the real loops - just a test a jump down here
	 * where the normal case is straight ahead instead of
	 * a test and jump around the error stuff.
	 */
abortLongWhite1:
    pCCITTCompInfo->compState = cCCITTCompFinishLongWhite1;
    goto compMHMightWriteOne;
abortLongWhite2:
    pCCITTCompInfo->compState = cCCITTCompFinishLongWhite2;
    goto compMHMightWriteOne;
abortLongWhite3:
    pCCITTCompInfo->compState = cCCITTCompFinishLongWhite3;
    goto compMHMightWriteOne;
abortShortWhite:
    pCCITTCompInfo->compState = cCCITTCompFinishShortWhite;
    goto compMHMightWriteOne;
abortLongBlack1:
    pCCITTCompInfo->compState = cCCITTCompFinishLongBlack1;
    goto compMHMightWriteOne;
abortLongBlack2:
    pCCITTCompInfo->compState = cCCITTCompFinishLongBlack2;
    goto compMHMightWriteOne;
abortLongBlack3:
    pCCITTCompInfo->compState = cCCITTCompFinishLongBlack3;
    goto compMHMightWriteOne;
abortShortBlack:
    pCCITTCompInfo->compState = cCCITTCompFinishShortBlack;
    goto compMHMightWriteOne;
abortFlushLine:
    pCCITTCompInfo->compState = cCCITTCompFinishFlushLine;
    goto compMHAbort;

	/* We get here when we know that two bytes won't fit into the
	 * compressed output buffer.  Try putting one in.
	 * First, restore pCompDataEnd.  Remember that we decremented
	 * it at the top of the routine.
	 */
compMHMightWriteOne:
    pCompDataEnd++;
    if (pCurCompData < pCompDataEnd)
    {
	pCurCompData[0] = reverseByte[(UInt8) accum];
	pCurCompData++;
	accum >>= 8;
	vacantBit -= 8;
    }

	/* This is where we end up when we abort because we ran out
	 * of space in the compression buffer.  We jump here
	 * directly when we can't flush the accumulator without
	 * trying compMHMightWriteOne.
	 */
compMHAbort:
    *ppCompData = pCurCompData;
    pCCITTCompInfo->currentRun = pCurRun - pFirstRun;
    pCCITTCompInfo->remainingLength = runLength;
    pCCITTCompInfo->accumBits = accum;
    pCCITTCompInfo->vacantBit = vacantBit;
    
    return ia_bufferExceeded;	/* indicates that we stopped because we
				 * ran out of space.
				 */

}

/*
 $Log: mimhc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.6  1998/01/13 23:52:44  ddavies
 mimhc.c -> /product/ipcore/ipshared/src/RCS/mimhc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.5  1996/05/01 23:24:27  ddavies
 ddavies on Wed May 1 16:17:18 PDT 1996

 mimhc.c -> /product/ipcore/ipshared/src/RCS/mimhc.c,v

 Yow, those original checkin messages were what happens when you hit the
 "paste" key by accident.  This is a new file which contains the procedure
 i_compressRowsMMR.  This routine compresses some scan lines of an
 ipshared-style image using MMR (CCITT Group 4).  It's about 3.5x faster
 than the Leffler equivalent (on the SPARC, ~2x-3x faster on PC).
 This uses the same style of interface used by i_compressRowsMH and
 i_compressRowsPackBits (and soon - i_compressRowsMR).  The routine
 stops when either it completes the compression task or it fills the
 output buffer.  It can be restarted in either case.
 NO WAIT - that's the comment for i_mmrc.c.  All I did to this file
 is change ip_compMHRow so that it checks to see if the pipeline has
 room for new codes before inserting them instead of checking to see
 if there's enough to write after inserting new data.  This makes
 ip_compMHRow and ip_compMMRRow consistent in the way they handle the
 pipeline.  They should be consistent since they're both used in MR
 compression.

 Revision 1.4  1996/05/01 23:15:25  ddavies
 mimhc.c -> /product/ipcore/ipshared/src/RCS/mimhc.c,v

 Big stupid mistaken checkin.  See next message.

 Revision 1.3  1996/04/25 18:42:14  ddavies
 mimhc.c -> /product/ipcore/ipshared/src/RCS/mimhc.c,v

 Fix lots of comments.

 Revision 1.2  1996/04/11 01:04:30  ddavies
 ddavies on Wed Apr 10 18:01:39 PDT 1996

 mimhc.c -> /product/ipcore/ipshared/src/RCS/mimhc.c,v

 Add a couple of minor performance tweaks having to do with detecting the
 end of the compression buffer.  The Microsoft Visual C/C++ compiler is
 so WEIRD!!  It cleverly rearranges the code into spaghetti (ignoring all
 the hints about how checks could be cheap) and leaves ALL the variables
 in memory ALL of the time.  Gack!

 Revision 1.1  1996/04/10 22:02:26  ddavies
 Initial revision

 */
