/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mimhd.c: low level code used in Modified Huffman and Modified Read
 *	     decompression.
 *
 *	Int32	ip_decompMHRow()
 *			Decompresses one row of data.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#define _NEED_HORIZONTAL_CODES_
#include "mmrdec.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mimhd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* This routine decompresses one scan line of data.  It stops when
 * it runs out of compressed data.  It can restart in the same spot.
 * This is done by setting, and later dispatching on, restart states.
 * The states for decompression are listed in shrcmprs.prv.
 * cCCITTDecompNewLine is already state 0.  The remaining
 * state variables (number of scan lines left in the current block,
 * the 1-word decoding buffer, number of bits in that buffer, etc) are
 * also saved in the private state structure.
 */

/* To review, MH and MR compression represent a binary image scan line as a
 * sequence of white runs and black runs.  The length of each run is
 * encoded.  One always starts with a white run, possibly of length 0.
 * One always encodes all the bits in the scan line.
 *
 * The difference between MH and MR compression occurs at the ends of
 * scan lines.  In MH, all scan lines are padded to byte or word
 * boundaries.  There are no explicit end of line codes.  In MR, there
 * are explicit end of line codes.  The sequence is:
 * <optional alignment gap><EOL><optional 1D vs 2D bit><CompressedDataLine1>
 * <optional alignment gap><EOL><optional 1D vs 2D bit><CompressedDataLine2>
 * ...
 * <optional alignment gap><EOL><optional 1D vs 2D bit><CompressedDataLineN>
 * <optional alignment gap><EOL><optional 1D vs 2D bit>
 *
 * Since the first data line is preceeded by an End of Line code and
 * the last line is followed by and End Of Line code, one can debate
 * whether these are really BeginningOfLine codes or EndOfLine codes.
 * The optional alignment gap is inserted when the encoder wishes
 * to guarantee that the EOL code ends on a byte boundary.
 *
 * The standard G3/G4 horizontal distance codes are a little weird.
 * There is one code used to specify blocks of white space
 * (used to cover the distance from a right endpoint of one run
 * to the left endpoint of the run to the right) and another
 * code used to specify blocks of black pixels.  Why two codes?
 * It's probably because black runs tend to be shorter than white runs.
 * The black code has fewer bits/distance at the small distances.
 * The white code has fewer bits/distance at the larger distances.
 *
 * Each of the codes has individual entries for all distances from
 * 0 through 63.  They also have individual entries for each
 * multiple of 64 up through 1728.  There is one common code for
 * spanning pixels in blocks of 1792, 1856 (units of 64), ..., 2560.
 * For distances greater than 2560, one concatenates 2560 codes
 * together.  Therefore, a single distances has the form:
 *	(<some number of 2560 codes>)
 *	(<black or white code for 64 through 2496 in blocks of 64))
 *	<terminating black or white code specifying a distance
 *	 from 0 through 63>
 *
 * Only the last code is required.  If the block length is an
 * exact multiple of 64, the terminating code specifies a distance
 * of 0.  If the block is an exact multiple of 2560, the middle
 * code is left out and the last code specifies 0.
 * If n * 2560 < distance < n * 2560 + 64, the middle code is also
 * left out and the terminating code specifies the offset from
 * the last multiple of 2560.
 *
 * As stated above, the black and white codes are different.
 * The white codes are nicely arranged so that using an 8 bit
 * lookup table will return the following info:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code).
 * If the code value is greater than 63, it is a non-terminating
 * code.  If the code length is 9 (the length of the longest
 * white code), it is a 64x code and the final value should be
 * found by adding the ninth bit * 64 to the code value.  If the
 * answer is 1664, Then the answer is really 1728 (weird - is 1664
 * so common that it needs a special code?).  If the code length
 * is > 9, this is a 64x code shared with the black codes.  Use
 * the black lookup method as described next.
 *
 * The black codes have a nice feature.  Either the first four bits
 * are 0000, or they're not.  If they are not 0000, we can use a
 * lookup table indexed by 6 bits that encodes the 9 interesting
 * entries. These represent the distances 1 through 9.  Each entry has
 *	1. code value
 *	2. code length in bits.
 *
 * If the first four bits are 0000, we get rid of them and use an 8-bit
 * lookup table on the remaining bits.  The output of this
 * table is the same as the output of the white table described above:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code after discarding the first 4 bits).
 * There are 13-bit black codes, but they are nicely arranged.  So,
 * if the code length after discarding the first 4 bits is 9, this
 * is a 64x code and the final value should be found by adding the
 * ninth bit * 64 to the code value.
 *
 * For the sake of saving table space, we actually look up the longer codes
 * one nibble at a time.  The early lookups just return a pointer to the
 * next table.  An interesting number of the nibbles have no associated code,
 * so no tables are allocated for them.
 *
 * To encode a distance, we put out one 2560 code for each multiple
 * of 2560 in the distance, then the correct 64x code (if any) for
 * any non-zero multiples of 64 remaining, then the terminating code
 * for the remainder.  Only the terminating code is required.  To
 * decode a number, we read in codes, adding their values, until
 * a terminating code (one whose length is less than 64) is reached.
 *
 * The basic structure of the routine has an outer loop that contains
 * two inner loops.  Each trip through the outer loop fills in the right
 * and left endpoints of one black run.  The first inner
 * loop is used to decode a white run (sets the run's left endpoint).
 * The second inner loop is used to decode the next black run (sets the run's
 * right endpoint).  As seen above, each run length is specified using
 * a sequence of black or white codes.  Each trip through the inner
 * white run loop decodes one token used to specify part of a white
 * run length.  We exit the white run loop when a terminating token is
 * found.  The black run loop works the same way.  This seems like
 * a reasonably efficient structure.
 *
 * The compressed data streams are deserialized using a register called
 * "accum".  Bytes are loaded into the left side of accum just above the
 * current valid bits.  accum is shifted to the right (the LS end) where
 * the bits are used to index decoding tables.  The variable "vacantBit"
 * holds the number of the first vacant bit in accum.  This is the
 * distance by which incoming bytes are shifted left before being put
 * into accum.
 *
 * We're looking at the right end of accum and shifting things to
 * the right.  Unfortunately, all the CCITT tables, as well as the
 * encoded data stream, assume data is encoded and decoded from left
 * to right, not the other way around.  Therefore, we have to bit-reverse
 * each byte before inserting it into accum.  The table reverseByte is
 * used for this.
 *
 * What happens when we are about to run out of compressed data?
 * This condition is detected when there's room in accum for more
 * bytes, but accum isn't empty yet.  Well, the action taken depends on
 * the value of the flushPipeline parameter.  If flushPipeline is not
 * set, the client has told us that there's another compressed data buffer
 * after this one.  The contents of that buffer should be seamlessly
 * concatenated with this buffer.  Therefore, we just stop after loading
 * the very last byte out of the current buffer into accum, set a state
 * that tells us where to restart, save accum and any other relevant state
 * and return.  When we're given the next compressed data buffer, we'll
 * continue.
 *
 * If flushPipeline is set, the client is saying that there isn't any more
 * compressed data.  This often happens in decompression - the amount
 * of data saved is exactly the amount needed.  In this case, we simply
 * continue on with the data left in accum.  If it runs out before we
 * reach the end of the scan line, we return with an ia_decodeFailure
 * return code.
 *
 * This routine does not paint the decoded runs into a bitmap.  That's
 * done elsewhere.  This routine just returns an array contains the
 * run endpoints.  This division of labor was done both because it should
 * be more efficient to have two small, focused tasks and because there
 * are applications (like connected component finding) which will
 * operate on the runs directly.
 *
 * In MH encoding, each line is padded to the next byte or word boundary.
 * We take a padType field that indicates the pad type.  It is checked
 * at the end of each encoded line.  If padding is required, the requisite
 * number of bits is flushed from the accumulator.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_decompMHRow()
 *
 *	This procedure decompresses one line of data using the 1D CCITT
 *	Huffman tables.  The compressed data is in natural byte order.
 *	The uncompressed data is in pixr order.
 *
 *	Input:
 *		pCCITTDecompInfo: pointer to decompression state.
 *		padType:	If equal to cPadToByte, end of
 *				compresssed line is padded to next
 *				byte boundary.
 *				If equal to cPadToWord, compresssed
 *				line is padded to next 2-byte boundary.
 *				If equal to cNoPad, end of the compressed
 *				is not padded.
 *				Any padding is flushed after all
 *				encoded bits have been decoded.
 *		w:		width of image (or tile or stripe) in
 *				pixels.
 *		flushPipeline:	If cTrue, this routine should be able to
 *				get to the end of the scan line using
 *				only data in the compressed data buffer.
 *				If set to cFalse, it's ok to stop and
 *				ask for more compressed data when we're
 *				starting to run out.  Additional data will
 *				be seamlessly concatenated to current data.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *		ppDummy:	pointer to pointer to the Run32 structure
 *				just after the last valid structure.
 *				*ppDummy is updated to point to the
 *				structure beyond the last valid one on return.
 *
 *	Action:
 *		Attempt to decompress a single scan line of data.
 *		The compressed data is assumed to be in natural byte order and
 *		is arbitrarily aligned, so should be read 1 byte at a
 *		time.
 *
 *		If the end of the scan line is reached, check padType.
 *		If it contains cPadToByte, flush any bits required to get
 *		to the next byte boundary from the accumulator.
 *		If it's set to cPadToWord, flush any bits required to get
 *		to the next 2-byte boundary from the accumulator.
 *		In any case, set pCCITTDecompInfo->decompState =
 *		cCCITTDecompNewLine.
 *		If we stop because the compressed data buffer empties,
 *		set pCCITTDecompInfo->decompState to the state which
 *		describes where we should restart and
 *		put the relevant byte counts and run pointers into
 *		*pCCITTDecompInfo.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be read.  Callers can use this
 *		to determine the number of bytes read so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pProc was NULL
 *		ia_bufferExceeded	stopped because we tried to read
 *					something to the compressed data
 *					buffer but it was empty
 *
 **************************************************************************/
Int32  CDECL 
ip_decompMHRow (
    IPState    *pIPState,
    CCITTDecompressionInfo *pCCITTDecompInfo,
    UInt32	padType,
    Int32	w,
    Bool	flushPipeline,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd,
    Run32     **ppDummy)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    Run32	*pCurRun, *pFirstRun;	/* pointer to current  and first
					 * Run32 structs */
    UInt32	 accum;			/* holds bits to be decoded */
    Int8	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */
    UInt8	 codeIndex;		/* the fragment of the input stream
					 * currently being used to index
					 * decoding tables. */
    Int32	 codeVal;		/* value found in decoding tables */
    UInt32	 tokenVal;		/* a version of codeVal that's
					 * used when calculating
					 * non-terminating codes. */
    mmrTableEntry *pNonTermTable;	/* points to second level tables
					 * of non-terminating codes.
					 */


	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pCCITTDecompInfo->accumBits;
    vacantBit = (Int8) pCCITTDecompInfo->vacantBit;
    tokenVal = 0;

	/* Most of time time, we want to compare to
	 * pCompDataEnd-1, so decrement it here */
    pCompDataEnd--;

	/* point to current run by first pointing to second run,
	 * then offseting to current run.  Remember that the
	 * first run is a dummy containing [-1, -1]
	 */
    pFirstRun = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				   (pCCITTDecompInfo->runArrayOffset) +
				   sizeof(Run32));
    pCurRun = pFirstRun + pCCITTDecompInfo->currentRun;

	/* see where we should start off */
    switch (pCCITTDecompInfo->decompState)
    {
    case cCCITTDecompNewLine:
	goto decompMHNewLine;

    case cCCITTDecompMHWhiteRun:
	goto decompMHFinishWhite;

    case cCCITTDecompMHBlackRun:
	goto decompMHFinishBlack;

    default:	/* Restart state belongs to some other routine */
	NORMAL_RETURN;
    }

decompMHNewLine:

	/* ********************* MAIN LOOP ***********************/

	/* Here's the main loop.  We decode a white run, then a
	 * black run.  We stop when we reach the end of the scan
	 * line or run out of bytes in the compressed data buffer.
	 */
    while (cTrue)
    {

	    /* ************* WHITE RUN LOOP ************** */
	    /* Each time through this loop, we decode one more
	     * token.  At the top of the loop, we see if there's
	     * room for more stuff.
	     */
	pCurRun[0].runLeft = 0;	/* initialize accumulated white run length */
	while (cTrue)
	{
decompMHFinishWhite:
	    if (vacantBit <= 16)
	    {
		if (pCurCompData < pCompDataEnd)
		{
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    accum |= reverseByte[pCurCompData[1]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData += 2;
		}
		else
		{
			/* We're not allowed to fetch 2 new bytes.
			 * If there's one byte, get it.
			 */
		    if (pCurCompData == pCompDataEnd)
		    {
			/* There's one more byte in the data stream. Get it. */
			accum |= reverseByte[pCurCompData[0]] << vacantBit;
			vacantBit += 8;
			pCurCompData ++;
		    }
		    if (!flushPipeline)
			goto decompMHAbortWhite;
		}
	    }

	        /* It is sad, but true, the we can find an EOL anywhere.
		 * furthermore, the EOL might be preceeded by an arbitrary
		 * number of 0's.  We can decode the EOL, but ONLY if we've
		 * flushed out the 0's first.
		 */
	    while (((accum & cEOLMask) == 0) && (vacantBit >= cEOLLength))
	    {
		accum >>= 1;
		vacantBit--;
	    }
	        /* If we don't have enough bits, try to get some more. */
	    if ((vacantBit < cEOLLength) && (pCurCompData <= pCompDataEnd))
		continue;

		/* When decoding white runs, we always use 8 bits to
		 * index a table.
		 * First, see if this is a terminating code yet.
		 */
	    codeIndex = (UInt8) accum & cWhiteTermDecodeMask;
	    codeVal = mmrRevWhiteTermDecodes[codeIndex];
	    if (codeVal <= 63)
	    {
		/* *********** Terminating white run code ************** */
		    /* set next run coord */
		pCurRun[0].runLeft += pCurRun[-1].runRight + codeVal + 1;

		    /* get length of code in bits */
		codeVal =
		    mmrRevWhiteTermDecodes[cTermDecodeTableLengthOffset +
					   codeIndex];
		accum >>= codeVal;
		vacantBit -= (Int8) codeVal;

		    /* HERE IS THE EXIT FROM THE WHITE RUN LOOP */
		break;
	    }
	    else	/* look up this token using the
			 * non-terminating tables */
	    {
		/* *********** Non Terminating white run code ************ */
		pNonTermTable =
		    mmrRevWhite1stNibbleNonTermDecodes[codeIndex & 0xF];
		if (pNonTermTable == NULL)
		    goto decompMHDecodeFailure;
	
		    /* We got a valid table pointer using the bottom 4 bits.
		     * advance past them and decode the next 4 bits.
		     */
		accum >>= 4;
		vacantBit -= 4;
		codeIndex = (Int8) accum & 0xF;
		if (pNonTermTable[codeIndex].codeLength > 5)
		{
			/* The code in accum is one of the common codes used
			 * to skip really big blocks of pixels.  To decode
			 * these, we look it up in the black non-terminating
			 * tables.
			 */
		    pNonTermTable =
			mmrRevBlack1stNibbleNonTermDecodes[codeIndex];
		    if (pNonTermTable == NULL)
			goto decompMHDecodeFailure;
	
			/* Now we really have a pointer to the lower level
			 * non-terminating code table, so toss the nibble we
			 * just used.
			 */
		    accum >>= 4;
		    vacantBit -= 4;
		    codeIndex = (Int8) accum & 0xF;
		}
	
		    /* and look up the code length */
		codeVal = pNonTermTable[codeIndex].codeLength;
		if (codeVal == 0)
		    goto decompMHDecodeFailure;
    
		tokenVal = pNonTermTable[codeIndex].codeBits;
		if (codeVal == 5)	/* add 64 x value of 5th bit */
		{
		    tokenVal += (accum * 4) & 64;
		    if (tokenVal == cWeirdWhiteCode)
			tokenVal += 64;
		}
		accum >>= codeVal;
		vacantBit -= (Int8) codeVal;

		    /* now move to the left in the current scan line by
		     * the number of pixels just decoded in tokenVal.
		     * We keep doing this until we find a terminating code.
		     */
		pCurRun[0].runLeft += tokenVal;
		    /* If that was an EOL, treat it like a terminating code.
		     */
		if (tokenVal == 0x8000)
		    break;

	    }	/* end of non-terminating white case */
	}	/* end of loop that decodes white values */

	    /* Was that the white run that went to the edge of the page?
	     * Note that at this point pCurRun points to the
	     * dummy run that's off the edge of the page, which
	     * is the correct exit condition.
	     */
	if (pCurRun->runLeft >= w)
	{
	        /* If we found an EOL when we were looking for data,
	         * pCurRun[0].runLeftt will be much
	         * greater than w.  If this is the case, set it to
	         * w-1 and put the EOL code back into accum and vacantBit.
	         * The next time our caller goes around its loop, it will
	         * expect to find the EOL in accum.
	         */
	    if (tokenVal == 0x8000)
	    {
		pCurRun[0].runLeft = w;
		accum = (accum << cEOLLength) | cEOLCode;
		vacantBit += cEOLLength;
	    }

		/* THIS IS THE FIRST EXIT FROM THE MAIN LOOP */
	    break;
	}

	    /* ************* BLACK RUN LOOP ************** */
	    /* Each time through this loop, we decode one more
	     * token.  At the top of the loop, we see if there's
	     * room for more stuff.
	     */
	pCurRun[0].runRight = 0; /* initialize accumulated black run length */
	while (cTrue)
	{
decompMHFinishBlack:
	    if (vacantBit <= 16)
	    {
		if (pCurCompData < pCompDataEnd)
		{
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    accum |= reverseByte[pCurCompData[1]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData += 2;
		}
		else
		{
			/* We're not allowed to fetch 2 new bytes.
			 * If there's one byte, get it.
			 */
		    if (pCurCompData == pCompDataEnd)
		    {
			/* There's one more byte in the data stream. Get it. */
			accum |= reverseByte[pCurCompData[0]] << vacantBit;
			vacantBit += 8;
			pCurCompData ++;
		    }
		    if (!flushPipeline)
			goto decompMHAbortBlack;
		}
	    }

	        /* It is sad, but true, the we can find an EOL anywhere.
		 * furthermore, the EOL might be preceeded by an arbitrary
		 * number of 0's.  We can decode the EOL, but ONLY if we've
		 * flushed out the 0's first.
		 */
	    while (((accum & cEOLMask) == 0) && (vacantBit >= cEOLLength))
	    {
		accum >>= 1;
		vacantBit--;
	    }
	        /* If we don't have enough bits, try to get some more. */
	    if ((vacantBit < cEOLLength) && (pCurCompData <= pCompDataEnd))
		continue;

		/* In the black world, terminating codes can be short
		 * or long.  First, see if we have a short terminating
		 * code.  If that doesn't work, try a long terminating
		 * code.  If that doesn't work, we'll try a non-terminating
		 * code.
		 */
	    if ((accum & cBlackLSBitsMask) != 0)	/* it's short */
	    {
		    /* get code length and index using the short code
		     * table */
		codeIndex = (Int8) accum & cShortBlackTermDecodeMask;
		    /* get code value */
		codeVal = mmrRevBlackShortTermDecodes[codeIndex];
		    /* get code length */
		codeIndex = (Int8) mmrRevBlackShortTermDecodes[
					cShortTermDecodeTableLengthOffset +
					codeIndex];
	    }
	    else
	    {
		    /* delete the LS four bits since they're all 0, and use
		     * the long terminating code table. */
		accum >>= 4;
		vacantBit -= 4;
		codeIndex = (Int8) accum & cBlackTermDecodeMask;
		codeVal = mmrRevBlackLongTermDecodes[codeIndex];
		codeIndex = (Int8) mmrRevBlackLongTermDecodes[
					cTermDecodeTableLengthOffset +
					codeIndex];
	    }

	    if (codeVal <= 63)
	    {
		/* *********** Terminating black run code ************** */
		    /* set next run coord */
		pCurRun[0].runRight += pCurRun[0].runLeft + codeVal - 1;

		accum >>= codeIndex;
		vacantBit -= codeIndex;

		    /* HERE IS THE EXIT FROM THE BLACK RUN LOOP */
		break;
	    }

	    else	/* look up this token using the
			 * non-terminating tables.  At this point, we've
			 * already deleted the LS four bits, so we have
			 * the right nibble for this table
			 */
	    {
		pNonTermTable = mmrRevBlack1stNibbleNonTermDecodes[accum & 0xF];
		if (pNonTermTable == NULL)
		    goto decompMHDecodeFailure;
	
		    /* We got a valid table pointer using the bottom 4 bits.
		     * advance past them and decode the next 4 bits.
		     */
		accum >>= 4;
		vacantBit -= 4;
		codeIndex = (Int8) accum & 0xF;

		    /* and look up the code length */
		codeVal = pNonTermTable[codeIndex].codeLength;
		if (codeVal == 0)
		    goto decompMHDecodeFailure;
    
		tokenVal = pNonTermTable[codeIndex].codeBits;
		if (codeVal == 5)	/* add 64 x value of 5th bit */
		    tokenVal += (accum * 4) & 64;
		accum >>= codeVal;
		vacantBit -= (Int8) codeVal;

		    /* now move to the left in the current scan line by
		     * the number of pixels just decoded in tokenVal.
		     * We keep doing this until we find a terminating code.
		     */
		pCurRun[0].runRight += tokenVal;

		    /* If that was an EOL, treat it as a terminating code. */
		if (tokenVal == 0x8000)
		    break;

	    }	/* end of non-terminating black code case */
	}	/* end of loop that decodes black values */

	    /* Just so the termination conditions are all the same,
	     * point to the next run now */
	pCurRun++;

	    /* Was that the black run that went to the edge of the page?
	     */
	if (pCurRun[-1].runRight >= w-1)
	{
	        /* If we found an EOL when we were looking for data,
	         * pCurRun[-1].runRight will be much
	         * greater than w-1.  If this is the case, set it to
	         * w-1 and put the EOL code back into accum and vacantBit.
	         * The next time our caller goes around its loop, it will
	         * expect to find the EOL in accum.
	         */
	    if (tokenVal == 0x8000)
	    {
		pCurRun[-1].runRight = w-1;
		accum = (accum << cEOLLength) | cEOLCode;
		vacantBit += cEOLLength;
	    }

		/* THIS IS THE SECOND (and last) EXIT FROM THE MAIN LOOP */
	    break;
	}
    }

	/* If we get here, we must have successfully decoded all the runs
	 * in one scan line.  Whew!
	 * See if the scan line was padded to a byte boundary, a word
	 * boundary or not at all.
	 * How do the contents of vacantBit relate to the padding?
	 * Well, we increase vacantBit by 8 every time we insert a
	 * byte and decrease it by the length of each code token
	 * removed.  We want to flush (shift right) accum until the
	 * number of bits in accum is a muliple of 8.  This is true
	 * when (vacantBit MOD 8) == 0.
	 */
    switch (padType)
    {
    case cPadToByte:
	     /* There two cases here:
	      * 1. vacantBit & 0x7 == 0:
	      *		Do nothing.
	      * 2. vacantBit & 0x7 != 0:
	      *		shift accum right by (vacantBit & 7)
	      *		subtract (vacantBit & 7) from vacantBit
	      */
	codeIndex = vacantBit & 0x7;
	break;
    case cPadToWord:
	     /* There two cases here:
	      * 1. vacantBit & 0xF == 0:
	      *		Do nothing.
	      * 2. vacantBit & 0xF != 0:
	      *		shift accum right by (vacantBit & 0xF)
	      *		subtract (vacantBit & 0xF) from vacantBit
	      */
	codeIndex = vacantBit & 0xF;
	break;
    default:	/* no padding required */
	codeIndex = 0;
	break;
    }

	/* Do the required flushing.  The number of bits to be flushed
	 * is in codeIndex.
	 */
    accum >>= codeIndex;
    vacantBit -= codeIndex;

	/* Set up the saved state so we will know we're done */
    *ppCompData = pCurCompData;
    *ppDummy = pCurRun;
    pCCITTDecompInfo->currentRun = 0;
    pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
    pCCITTDecompInfo->accumBits = accum;
    pCCITTDecompInfo->vacantBit = vacantBit;

    NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

	/* We get here if we were about to have a decode failure.
	 * If flushPipeline is set and accum is 0, then we
	 * really just ran out of data, so return ia_bufferExceeded
	 * instead.  If the upper level sees an ia_bufferExceeded
	 * with flushPipeline, it knows that the client didn't
	 * supply enough compressed data and we're doomed.
	 */
decompMHDecodeFailure:
    if ((flushPipeline) && (accum == 0))
	return ia_bufferExceeded;
    else
	return ia_decodeFailure;

	/* First, we have all the abort targets.  We do this weird
	 * thing of setting all the states here instead of above
	 * in the hopes that the compiler will generate better
	 * code in the real loops - just a test a jump down here
	 * where the normal case is straight ahead instead of
	 * a test and jump around the error stuff.
	 */
decompMHAbortWhite:
    pCCITTDecompInfo->decompState = cCCITTDecompMHWhiteRun;
    goto decompMHFinishAbort;
decompMHAbortBlack:
    pCCITTDecompInfo->decompState = cCCITTDecompMHBlackRun;
    goto decompMHFinishAbort;

	/* This is where we end up when we abort because we ran out
	 * of bytes in the compression buffer and we weren't supposed
	 * to flush the compression pipeline.
	 */
decompMHFinishAbort:
    *ppCompData = pCurCompData;
    pCCITTDecompInfo->currentRun = pCurRun - pFirstRun;
    pCCITTDecompInfo->accumBits = accum;
    pCCITTDecompInfo->vacantBit = vacantBit;
    
    return ia_bufferExceeded;	/* indicates that we stopped because we
				 * ran out of bytes.
				 */

}


/*

 $Log: mimhd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.5  1998/11/12 00:18:52  aber
 aber on Wed Nov 11 16:18:07 PST 1998

 mimhd.c -> /product/ipcore/ipshared/src/RCS/mimhd.c,v

 Synced fix from ddavies:
 Fix problems found when an EOL is unexpectedly found in when looking for
 normal black and white runs.  Some fax machine lie about the number of scan
  lines, then put extra EOL's at the end.  The spec implies we should be able
  to handle this.

 Revision 1.4  1998/01/13 23:52:46  ddavies
 mimhd.c -> /product/ipcore/ipshared/src/RCS/mimhd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1996/05/11 01:04:58  ddavies
 mimhd.c -> /product/ipcore/ipshared/src/RCS/mimhd.c,v

 Fix comments and indentation.

 Revision 1.2  1996/05/06 18:50:05  ddavies
 mimhd.c -> /product/ipcore/ipshared/src/RCS/mimhd.c,v

 Fix comments.  Change name of pCurDecompData to pCurCompData to be
 consistent with other CCITT decompression routines.  Tidy up the
 error handling by having the code that detects that the compression
 buffer is (almost) empty always try to read the last byte immediately
 instead of sometimes reading it immediately and sometimes trying to
 read it later.

 Revision 1.1  1996/04/25 18:45:24  ddavies
 Initial revision


 */

