/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mimmrc.c: low level code used in Modified Modified Read
 *	      compression.
 *
 *	Int32	ip_compMMRRow()
 *			Compresses one row of data using MMR.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#define _NEED_HORIZONTAL_CODES_
#define _NEED_VERTICAL_CODES_
#include "mmrenc.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mimmrc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* This file contains the routine used to compress one row of data using
 * Modified Modified Read encoding.  This method is described in
 * the CCITT Fascicle VII.3, Recommendation T.6.  The Recommendation
 * T.4 for Modified Read encoding has many useful examples of 2D encoding
 * as well.
 *
 * MMR is a pure 2 dimensional (2D) encoding standard.  Each line is
 * coded from left to right with respect to the line above it.  The
 * first line is assumed to be preceeded by a blank line.  The assumed
 * conditions at the left and right ends of the scan line are discussed
 * below.
 *
 * The encoding loop is a state machine, much like the decoding
 * loop.  In fact, we'll try to make them as similar as possible
 * to reduce the chance of error.  We go around the loop once for
 * each complete code put out or taken in.  I'd better explain
 * MMR coding in order for that to make sense.
 *
 * The standard draws several little pictures.  The "Reference Line"
 * is the line above the one being coded.  The "Coding Line" is the
 * image line being coded.  At each stage, we have a starting
 * position, lastX, in the coding line.  That is obvious, but
 * important.  The starting position in the coding line is
 * immediately to the right (since we encode left to right) of
 * the last pixel that was encoded.  An important point is that both
 * the encoder and the decoder can identify the starting position
 * since it reflects only things that have happened in the past.
 * As a hack that has ugly consequences, lastX begins on pixel -1,
 * the leftmost pixel of the line.  We'll discuss this after
 * talking about the normal encoding scheme.
 *
 * Both the encoder and decoder can see the reference line, so
 * let's start with that.  Starting at lastX in the reference
 * line, the encoder looks strictly to the right for the first pixel that a)
 * has a different color than the pixel to its left and b) has the
 * opposite color from the pixel at position lastX in the coding line.
 * Even though we haven't encoded lastX, we can deduce its
 * polarity from the sequence of coding state.
 *
 * For example, say we're about to encode a left endpoint of a black
 * run.  This means that the pixel in the coding line at position
 * lastX is white.  We look to the right of position lastX in the
 * reference line for the first left endpoint of a black run.
 * If we're going to encode a right endpoint, we're starting on a
 * black pixel in the coding line.  We look to the right of lastX
 * in the reference line for the white pixel just beyond the
 * nearest right endpoint.  The CCITT standard calls this first
 * endpoint on the reference line b1.  The location of the next
 * pixel to the left of b1 of opposite color from b1 is called b2.
 *
 * It's in the coding line that encoding and decoding become
 * different.  In encoding, we look to the right for the first
 * endpoint of proper left/right polarity.  Call this
 * location a1.  We then look to the right of a1 for the next pixel
 * of opposite color.  Call the location of this pixel a2.
 * When decoding, we just decode the next code which tells us what
 * the encoding process found.
 * Several things can happen at this point of encoding.
 *   1. There may be no pixels of the desired color (opposite color
 *	of coding line pixel at lastX) on the coding line under the 
 *	reference line run (b1,b2).  This is indicated by b2
 *	(one past the end of the run in the reference line) being 
 *	strictly to the left of a1 (the beginning of the run in the
 *	coding line).  We require a difference of at least one pixel
 *	between b2 and a1 to decide the run has really
 *	ended.  In this case, we issue a PASS code.  This tells
 *	both the encoder and decoder to skip the black or white run
 *	on the reference line by advancing lastX to b2.  Note that
 *	insisting on the 1-pixel buffer between b2 and a1 means
 *	that the color of the pixel under the new lastX will be
 *	the same as the color of the pixel under old lastX.  If
 *	we were looking for a left endpoint originally, we'll
 *	still be looking for one.  Ditto for right endpoints.
 *   2. If we didn't issue a PASS code, the next point on the
 *	reference line, b1, might be within distance 3 of the next
 *	point on the coding line, a1.  That is, | b1 - a1 | <= 3.
 *	If this is true, we use a vertical code to specify the 
 *	relative distance between them and move lastX up to a1's 
 *	position.  Since we're only moving past one endpoint, we swap 
 *	the polarity of the next endpoint to be encoded or decoded.
 *	For example, if we've just encoded a left endpoint, we`ll
 *	next encode a right endpoint.
 *	Group 4 encoding is based on the notion that this is
 *	the common case.
 *   3. If neither of those work, we issue a horizontal code
 *	followed by two distances.  One goes to the next endpoint, 
 *	the other goes to the endpoint after that.  Why two distances?
 *	Because we often want two insert two endpoints to get to
 *	the next vertical match.  A brand new run or a fork inserts
 *	two new endpoints.
 *
 * Here's a point that confused me.  Why do we always look strictly
 * to the right from lastX?  Say we have the following case where
 * each X represents a run of length 1:
 *
 *			X
 *		      X X X
 *
 * It would be nice to encode the second line using 6 vertical
 * codes (2 for each X - one for the left edge and one for the
 * right edge), all relative to the bit on the first
 * line.  However, the standard says that the pixel on the right
 * will be encoded relative to some other pixel on the reference
 * line because the obvious reference pixel isn't to the right of
 * the coding pixel.  Why did they do that?
 *
 * Take the following example:
 *		X	X
 *		     X  X  X
 *
 * Without the rule of looking strictly to the right, the code for
 * the coding line would be -2, -2, 0, 0, +2, +2.
 * Now say we have the following image:
 *		X	X
 *	      X		X X
 *
 * The encoding would be the SAME: -2, -2, 0, 0, +2, +2.  Ooops.
 * That's why they say you always have to look right for the reference run.
 *
 * Ok, but why *strictly* to the right?  Well, you have to decide
 * on some convention to distinguish
 *		X	X
 *	      X		X X
 * from
 *		X	X
 *	      X	X	  X
 *
 * With the rule, the code for the first example is V(-2), V(-2), PASS,
 * V(0), V(0), H(1, 1).  The code for the second case is V(-2), V(-2),
 * V(0), V(0), V(+2), H(1, someDistToTheRightOfTheLastMark).
 *
 * Here's an explanation of the ugly hack at the left edge of the
 * scan line.  What happens if black runs on both the reference and
 * coding lines touch the left end of the image?  If we don't start looking
 * from a position off the edge of the image, we won't see the
 * proper reference line endpoint because it won't be strictly to the
 * right.  If we do start with an imaginary pixel off to the left,
 * we have to remember not to include this imaginary pixel in the
 * first code.  Glug.  Well, that's exactly what the standard says
 * to do, subtract one from the first run length.  We'll finess this
 * by finding the leftmost color change points at the beginning of
 * the scan line, then setting lastX to 0 before calculating the
 * first code.
 *
 * Ok, what is lastX really used for?  It's used to decide where to
 * look for b1 and b2 on the reference line.  It's also used to encode
 * horizontal distances when those are needed.  It's not needed to
 * encode vertical codes or pass codes.
 *
 *			*	*	*
 *
 * There are several layers of encoding.  At the top, is a layer that
 * can find or generate one of the following codes (listed in bit
 * reversed order relative to the standard documents - remember that
 * in our special encoding world, one should read these from right to
 * left, i.e. <--<--<--< ):
 *
 *	Code	Meaning
 *	      1	Next coding line endpoint (left or right) is
 *		immediately below the next reference line endpoint.
 *	    110	Next coding line endpoint is 1 bit to the right of
 *		next reference line endpoint.
 *	    010	Next coding line endpoint is 1 bit to the left of
 *		next reference line the endpoint.
 *	    100	Horizontal distances to next two endpoints in
 *		coding line relative to previous point in *coding*
 *		line follow (see below)
 *	   1000	Pass code - move lastX to just past end of next run in
 *		reference line.
 *	 110000	Next coding line endpoint is 2 bits to the right of
 *		next reference line endpoint.
 *	 010000	Next coding line endpoint is 2 bits to the left of
 *		next reference line endpoint.
 *	1100000	Next coding line endpoint is 3 bits to the right of
 *		next reference line endpoint.
 *	0100000	Next coding line endpoint is 3 bits to the left of
 *		next reference line endpoint.
 *	1000000	G4 extension code.  If we find an extention, we quit.
 *
 * Each pass through the encoding loop generates one of these.
 * Similarly, each pass through the decoding loop decodes one of
 * these.  If the horizontal code is generated, we also generate
 * distances to the next two endpoints before returning to the top
 * of the loop.
 *
 * We can be in either of two states at the top of the loop,
 * preparing to encode/decode a right endpoint or preparing to
 * encode/decode a left endpoint.  Unfortunately, finding the next
 * endpoint on the reference line is a little tricky.  Don't the
 * next right and left endpoint always belong to the same run?
 * No.  Look at this example:
 *
 *      RRRRRRRRRRRRR             SSSSSSSSSSSSSS
 *     AA  BBBBB  CCCCC        DD
 *
 * The reference left endpoint for run A is the left endpoint of
 * R.  The reference left endpoint for runs B, C and D is the left
 * endpoint of S.  The reference right endpoints of A and B is the
 * right endpoint of R.  The reference right endpoint of runs C and D
 * is the right endpoint of S.  So, for runs B and C, the reference
 * right endpoint is to the left of the reference left endpoint.
 * For runs A and D, it's the other way around.  How do we deal with
 * this?  Well, when looking for left endpoints, we know that we
 * always have to look at the run containing the current left or
 * right endpoint and runs to the right of that.  When looking for
 * right endpoints, we might have to back up by one run before
 * starting to look.  We just always back up one run, then start
 * looking since that's easy.
 * 
 * If we get a horizontal code when preparing to generate a
 * left endpoint, the next coordinates are for the left
 * endpoint and the next right endpoint (specifying an entire run).
 * If we're ready to generate a right endpoint, the two coordinates
 * finish off the right end of the current run and start the left
 * end of the next one.
 *
 * If a horizontal code pushes us beyond the right edge of the
 * bitmap, we stop.  If it's the first of the pair that pushes
 * over the edge, the second code is set to represent 0.
 *
 * The standard G3/G4 horizontal distance codes are a little weird.
 * There is one code used to specify blocks of white space
 * (used to cover the distance from a right endpoint of one run
 * to the left endpoint of the run to the right) and another
 * code used to specify blocks of black pixels.  Why two codes?
 * It's probably because black runs tend to be shorter than white runs.
 * The black code has fewer bits/distance than the white codes
 * for the small distances.  The opposite is true for large distances.
 *
 * Each of the codes has individual entries for all distances from
 * 0 through 63.  They also have individual entries for each
 * multiple of 64 up through 1728.  There is one common code for
 * spanning pixels in blocks of 1792, 1856 (units of 64), ..., 2560.
 * For distances greater than 2560, one concatenates 2560 codes
 * together.  Therefore, a single distances has the form:
 *	(<some number of 2560 codes>)
 *	(<black or white code for 64 through 2496 in blocks of 64))
 *	<terminating black or white code specifying a distance
 *	 from 0 through 63>
 *
 * Only the last code is required.  If the block length is an
 * exact multiple of 64, the terminating code specifies a distance
 * of 0.  If the block is an exact multiple of 2560, the middle
 * code is left out and the last code specifies 0.
 * If n * 2560 < distance < n * 2560 + 64, the middle code is also
 * left out and the terminating code specifies the offset from
 * the last multiple of 2560.
 *
 * As stated above, the black and white codes are different.
 * The white codes are nicely arranged so that using an 8 bit
 * lookup table will return the following info:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code).
 * If the code value is greater than 63, it is a non-terminating
 * code.
 * If the code length is 9 (the length of the longest white code),
 * it is a 64x code and the final value should be found by adding
 * the ninth bit * 64 to the code value.  If the answer is 1664,
 * Then the answer is really 1728 (weird - is 1664 so common that
 * it needs a special code?).  If the code length is > 9, this is
 * a 64x code shared with the black codes.  Use the black lookup
 * method as described next.
 *
 * The black codes have a nice feature.  Either the first four bits
 * are 0000, or they're not.  If they are not 0000, we can use a
 * lookup table indexed by 6 bits that encodes the 9 interesting
 * entries. These represent the distances 1 through 9.  Each entry has
 *	1. code value
 *	2. code length in bits.
 *
 * If the first four bits are 0000, we get rid of them and use an 8-bit
 * lookup table on the remaining bits.  The output of this
 * table is the same as the output of the white table described above:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code after discarding the first 4 bits).
 * There are 13-bit black codes, but they are nicely arranged.  So,
 * if the code length after discarding the first 4 bits is 9, this
 * is a 64x code and the final value should be found by adding the
 * ninth bit * 64 to the code value.
 *
 * To save table space, we actually decode long codes 4 bits at a time.
 * The intermediate lookups return pointers to subsequent tables.
 * Since fewer than half of all combinations are actually used, this
 * saves quite a bit of table space.
 *
 * To encode a distance, we put out one 2560 code for each multiple
 * of 2560 in the distance, then the correct 64x code (if any) for
 * any non-zero multiples of 64 remaining, then the terminating code
 * for the remainder.  Only the terminating code is required.  To
 * decode a number, we read in codes, adding their values, until
 * a terminating code (one whose length is less than 64) is reached.
 *
 *			*	*	*
 *
 * So much for reading and writing numbers in horizontal codes.  How
 * does the encoding loop really work?  The encoding loop converts
 * runs on the current and preceeding lines into a sequence of
 * codes.  The decoding loop does the opposite.  The encoding loop
 * tests its current position with respect to runs on the preceeding
 * line to decode what codes to generate.  The decoding loop uses the
 * tokens read to update the current position and record run endpoints.
 *
 * Both loops have two main states: generating/expecting a left
 * endpoint (preparing to encode/decode a white run) and generating/expecting
 * a right endpoint (preparing to encode/decode a black run).  These govern
 * interpretation of input runs during encoding and generation of
 * output runs during decoding.  They also govern the generation/
 * interpretation of horizontal codes - either white then black if
 * looking for a left endpoint or black then white if looking for a
 * right endpoint.
 *
 * All the runs for a scan line are held in an array of Run32 structs.
 * We actually keep two arrays and ping-pong back and forth.  One is
 * used as the coding line and one is the reference line.
 *
 * We start encoding by setting the runs in one array to simulate a
 * blank line and filling the runs in the other array with the runs of
 * the real first line.  We find the reference run endpoints (at the
 * end of the line in this case), move lastX up to position 0 and
 * begin issuing codes.  In the first line, they are all horizontal
 * codes.  Decoding does a similar thing - setting the first set of
 * runs to form a blank line and setting lastX to 0 before decoding
 * the first code.
 *
 * The loop terminates when lastX >= the width of the image, i.e. we've
 * encoded an entire scan line.
 *
 * Let's look a what happens when we want to encode a new left
 * endpoint.
 *   1. Skip left endpoints on the reference line until we
 *	find one that's to the right of the current position.
 *   2. If the right endpoint of the run containing the
 *	reference left endpoint is more than 1 pixel to the left
 *	of the next left endpoint on the coding line, issue a PASS code,
 *	set the current horizontal position equal to the position of
 *	the reference line's right endpoint + 1 (so we're back in white space)
 *	and go back to the top of the loop with the state still 
 *	set to right endpoint.
 *   3. If a PASS code wasn't issued and the next coding line
 *	left endpoint is within 3 pixels of the reference line
 *	left endpoint, issue the corresponding vertical code, 
 *	update the current coding line position to be the position 
 *	of the coding line left endpoint, change the state to 
 *	right endpoint and go back to the top of the loop.
 *   4. If all else fails, issue a horizontal code specifying the
 *	number of white pixels to the next left endpoint (counting the
 *	current white pixel), and the number of pixels in the next
 *	black run.  Set the lastX to the position of the new right
 *	endpoint + 1 (so we're back in white space) and go back to
 *	the top of the loop with the state still set to left endpoint.
 *
 * When generating right endpoints,
 *   1. Skip right endpoints on the reference line until we
 *	find one that's to the right of the current position.
 *	We do this by starting on the run to the left of the run
 *	containing the previous reference left or right endpoint
 *	and looking to the right for the next legal right endpoint.
 *   2. If the left endpoint of the run to the right of the one
 *	containing the reference right endpoint is to the left
 *	of the next right endpoint on the coding line, issue a PASS code.
 *	Also set the current x position to the position of the
 *	new left endpoint (so we're still pointing to a black
 *	pixel) and go back to the top of the loop with
 *	the state still set to right endpoint.
 *   3. If a PASS code wasn't issued and the next reference line
 *	right endpoint is within 3 pixels of the next coding line
 *	right endpoint, issue the corresponding vertical code, 
 *	change the state to left endpoint and go back to the 
 *	top of the loop.
 *   4. If all else fails, issue a horizontal code specifying the
 *	number of pixels to the next endpoint (a run of width 1
 *	generates a 1, not a 0) and the number of white pixels
 *	before the next left endpoint.  Set the current horizontal
 *	position to the position of the new left endpoint and go back to
 *	the top of the loop with the state still set to right endpoint.
 *
 * There's a subtle feature here.  We encode the distance from our
 * current position to the next position.  When we go from the
 * first black pixel of a run to the first white pixel of the next
 * run, we've encoded the length of the length of the black run.
 * The same thing happens when going from the first white pixel to the
 * first black pixel.  However, if a PASS code moves us inside a
 * run, we subsequently encode the remaining length of the run,
 * including the pixel we're starting on.
 *
 * Let's say something about the mechanics of composing and decoding
 * bitstreams.  In both cases, we keep the bitstream in a 32-bit
 * register.  When composing, we keep a pointer to the first vacant
 * bit relative to the LS end of the word.  New codes are shifted
 * left by this much and ORed into the accumulator - to the right
 * of the older codes.  The pointer to the first vacant bit is
 * incremented by the length of the code.  Each time we get ready
 * to insert a new code, we look at the new value of the pointer.  If
 * it's greater than or equal to 16, we write out the bottom two bytes,
 * shift the word right by 16 and subtract 16 from the pointer.
 *
 * Since MMR never requires line padding, we do not flush the
 * accumulator at the bottom of the loop.  There is a special function
 * that can be used to flush the accumulator during encoding.  This
 * function is called only at the end of strips or tiles.
 *
 * When decoding the bitstream, we also keep a pointer to the first
 * vacant bit in the accumulator.  Before doing a decode, we look at
 * the vacant bit pointer.  If it's less than or equal to 16, we
 * shift the next input byte left to align it with the next vacant
 * bit and OR it into the accumulator, then repeat with the next
 * input byte.  To decode words, we mask the lower bits of the
 * accumulator and use the masked bits as an index into the tables
 * described above.
 *
 * Since we're shifting the accumulators right, the head of the bitstream
 * is in the LS bit of the accumulator.  All the tables assume that
 * bitstream should be read from left to right, not right to left.
 * To correct the, we reverse the order of the bits in every byte
 * read or written to the compressed data stream.
 *
 *		*	*	*
 *
 * We should also say something about starting and stopping and general
 * code structure.  There is an outer loop containing two main sections.
 * The outer loop has no explicit exit condition, it runs until
 * one of the internal sections says we're done.
 * 
 * One of the internal sections encodes left endpoints (runs of white
 * pixels to the left of left endpoints), the other encodes right
 * endpoints (runs of black pixels).  Since we don't always encode
 * left and right endpoints alternately (see description above), we
 * can't just cycle between these sections blindly.
 *
 * Each section is coded as a "while (cTrue)" loop.  When we do something
 * that keeps coding the same type of endpoint, we just loop around
 * in one of the inner loops.  When we want to encode the other type
 * of endpoint, we execute a "break" statement.  This drops us out of
 * the inner loop that encodes one type of endpoint and we fall
 * through to the inner loop for the other type of endpoint.
 *
 * At the top of each inner loop is section that detects whether we're
 * done with the line.  How do we detect this?  We simply test to
 * see if lastX >= width.  This means that we've encoded all the pixels
 * in the scan line.
 *
 * The section that looks for the end of the line is followed by
 * a section that finds the next left or right endpoint in the reference
 * line to the right of the current position in the coding line (to the
 * right of lastX).  After that comes a block of code that detects
 * whether the accumulator has at least 2 bytes in it. If so, two
 * bytes are written to the output bytestream.  If there's not
 * enough room in the output bytestream, we record that we should restart
 * in this state and abort.   Finally, we decide whether to issue
 * a PASS code, a vertical code or a horizontal code and a pair of
 * run length codes.
 *
 * Since the run length codes can be large, we also have to check the
 * bitstream accumulator in those states.  The block that issues
 * run length codes is also a pair of loops.  The first loop starts
 * by checking to see if the accumulator has enough data to write
 * and writes 2 bytes if so.  It then issues one horizontal token.
 * As explained above, it may take several tokens to encode a single
 * run length.  Once the run length is fully encoded, the second
 * loop does the same for the next run length.  Since it is possible
 * to fill the output buffer any time data is written to it, each of
 * the loops also has a restart label.  Restarting is done by simply
 * dispatching on the compression state.
 *
 * Since we may use this routine in both MMR and MR encoding, the restart
 * states for this routine must be distinguished from those used in
 * 1D encoding.  These states are defined in shrcmprs.prv as
 * CCITTCompression states.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_compMMRRow()
 *
 *	This procedure compresses one line of data using the 2D CCITT
 *	huffman tables.  The compressed data is in natural byte order.
 *	The uncompressed data is in pixr order.
 *
 *	Input:
 *		pCCITTCompInfo:	pointer to compression state.
 *		inWidth:	the width of the image not including
 *				frame.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *
 *	Action:
 *		Attempt to compress a single scan line of data.
 *		The compressed data is assumed to be in natural byte order and
 *		is arbitrarily aligned, so should be written 1 byte at a
 *		time.
 *
 *		*ppCompData is set to point to the next byte of
 *		compressed data to be written.  Callers can use this
 *		to determine the number of bytes written so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pProc was NULL
 *		ia_bufferExceeded	stopped because we tried to write
 *					something to the compressed data
 *					buffer but it was full
 *
 **************************************************************************/
Int32  CDECL 
ip_compMMRRow (
    IPState    *pIPState,
    CCITTCompressionInfo *pCCITTCompInfo,
    Int32	imWidth,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    Run32	*pFirstRun, *pFirstRefRun; /* ptrs to first runs in
					 * coding and reference lines */
    Run32	*pCurRun;		/* pointer to current run in
					 * coding line */
    Run32	*pRefRun;		/* pointer to run in
					 * the reference line holding the
					 * next right or left endpoint. */
    Int32	 lastX;			/* last x coord in line */
    UInt32	 accum;			/* accumulated encoded bits */
    Int32	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */
    Int32	 runLength;		/* length of run being encoded */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pCCITTCompInfo->accumBits;
    vacantBit = pCCITTCompInfo->vacantBit;
    lastX = pCCITTCompInfo->remainingLength;
	/* Most of time time, we want to compare to
	 * pCompDataEnd-1, so decrement it here */
    pCompDataEnd--;

	/* point to current run by first pointing to first real run,
	 * (skip the [-1,-1] dummy) then offseting to current run.
	 * To do this, we figure out which run array is the coding line and
	 * which is the reference line.  pCCITTCompInfo->run2ArrayIsCoding
	 * is set when the run2Array is the coding line.
	 */
    if (pCCITTCompInfo->run2ArrayIsCoding)
    {
	pFirstRun = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				       (pCCITTCompInfo->run2ArrayOffset) +
				       sizeof(Run32));
	pFirstRefRun = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				          (pCCITTCompInfo->runArrayOffset) +
					  sizeof(Run32));
    }
    else	/* the first run array holds the coding line */
    {
	pFirstRun = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				       (pCCITTCompInfo->runArrayOffset) +
				       sizeof(Run32));
	pFirstRefRun = (Run32 *)((UInt8 *)(pCCITTCompInfo) +
				          (pCCITTCompInfo->run2ArrayOffset) +
					  sizeof(Run32));
    }

    pCurRun = pFirstRun + pCCITTCompInfo->currentRun;
    pRefRun = pFirstRefRun + pCCITTCompInfo->refRun;

	/* see where we should start off */
    switch (pCCITTCompInfo->compState)
    {
    case cCCITTCompNewLine:
	    /* Start at the beginning of the scan line.  Here's where we
	     * do the silly trick of looking for the leftmost left
	     * endpoints in the coding and reference lines and
	     * then set lastX to 0 instead of -1.
	     * Well, hey, since pRefRun already points
	     * to the first run in the reference line, that run must be
	     * the run containing the leftmost left endpoint.
	     * So all we have to do is set lastX to 0 and jump into
	     * the loop that calculates left endpoints at the spot
	     * AFTER it adjusts pRefRun.
	     * lastX was set to 0 from pCCITTCompInfo->remainingLength
	     * above, so the jump is all we need.
	     * Just for the sake of paranoia, we check to see if
	     * we're already done at the beginning.
	     */
	if (lastX >= imWidth)
	    goto compMMRGoodCompletion;
	else
	    goto compMMRLeftEndpoint;

    case cCCITTCompDoLeftEndpoint:
	goto compMMRLeftEndpoint;

    case cCCITTCompDoLeftWhite:
	goto compMMRLeftWhite;

    case cCCITTCompDoLeftBlack:
	goto compMMRLeftBlack;

    case cCCITTCompDoRightEndpoint:
	goto compMMRRightEndpoint;

    case cCCITTCompDoRightBlack:
	goto compMMRRightBlack;

    case cCCITTCompDoRightWhite:
	goto compMMRRightWhite;

    default:	/* Restart state belongs to some other routine */
	NORMAL_RETURN;
    }


	/* ********************* MAIN LOOP ***********************/

	/* Here's the main loop.  It contains two inner loops, one
	 * for encoding left endpoints and one for encoding right
	 * endpoints.  We stop when we reach the end of the scan
	 * line or run out of space in the compressed data buffer.
	 */
    while (cTrue)
    {

	    /* ***************** LEFT ENDPOINTS ******************/
	while (cTrue)
	{
		/* First, are we done encoding this scan line?
		 * This is true if our current X position if off
		 * the right edge of the image.
		 */
	    if (lastX >= imWidth)
		goto compMMRGoodCompletion;

		/* Ok, there's more to do.  Find the next left
		 * endpoint in the reference line */
	    while (pRefRun->runLeft <= lastX)
		pRefRun++;

		/* Here's where we start the scan line.  We wanted to
		 * skip adjusting the right and left reference
		 * endpoints since they're already correct at this time.
		 */
compMMRLeftEndpoint:

		/* Now, see if the accumulator has at least two bytes
		 * of data to write.  If so, write them.  If not,
		 * write one if possible and quit.  This is also
		 * where we restart if the buffer was full last time
		 * we got here.
		 */
	    if (vacantBit >= 16)
	    {
		if (pCurCompData >= pCompDataEnd)
		    goto abortLeftEndpoint;
    
		    /* write out LS 2 bytes of accumulator and increment
		     * pointers */
		pCurCompData[0] = reverseByte[(UInt8) accum];
		pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		accum >>= 16;			/* get rid of written bytes */
		pCurCompData += 2;
		vacantBit -= 16;
	    }

		    /* First see if we should issue a PASS code.  This
		     * happens if the reference run's right endpoint
		     * is more than 1 bit to the left of the next left
		     * endpoint on the current line.  The "more than 1"
		     * weirdness is so that if we skip over the previous run,
		     * we'll still end up in white space.
		     */
	    if (pCurRun->runLeft - pRefRun->runRight > 1)
	    {
		accum |= cMMRPass << vacantBit;
		vacantBit += cMMRPassLen;
		lastX = pRefRun->runRight + 1;	/* advance to passed
						 * position. */
		    /* We don't advance the coding run because it's still
		     * to the right of us by definition.  So, we just
		     * fall through the other cases in the else-if chain
		     * and loop back around to generate another
		     * left endpoint.
		     */
	    }
	    else
		    /* See if we should use a vertical code instead.
		     * This happens when the distance from the
		     * reference run's left endpoint to the coding
		     * run's left endpoint is less than or equal to 3.
		     * The trick below is a way of doing comparison
		     * of absolute values.  We do an unsigned compare
		     * of (a - b + 3) <= 6.  If a > (b+3), then
		     * the sum is clearly >= 6.  If (a-3) < b, then
		     * (a - b + 3) produces a negative number, which
		     * in unsigned arithmetic looks like a big positive
		     * number, so it's still > 6.  In between, we get
		     * a positive number <= 6.
		     */
	    if ((UInt32)(pCurRun->runLeft - pRefRun->runLeft + 3) <= 6)
	    {
		runLength = pCurRun->runLeft - pRefRun->runLeft + 3;
		accum |= mmrRevVerticalCodes[runLength] << vacantBit;
		vacantBit += mmrRevVertCodeLengths[runLength];

		    /* Point past the white space we've just skipped */
		lastX = pCurRun->runLeft;
		    /* Leave pCurRun pointing to the same run since we
		     * now want to generate a right endpoint of the
		     * same run.
		     * Skip down to generate a right endpoint.
		     */

		    /* THIS IS HOW WE GET TO THE RIGHT ENDPOINT LOOP */
		break;
	    }
	    else
		/* Issue a horizontal code (2 edges) that takes us to
		 * the next left endpoint.
		 */
	    {
		    /* First, generate the horizontal code itself */
		accum |= cMMRHoriz << vacantBit;
		vacantBit += cMMRHorizLen;

		    /* Here's the loop used to generate codes for white
		     * space.  It begins with an attempt to make room
		     * in accum, then tries to encode one token in
		     * the white run length.  Each trip through the
		     * loop encodes one token.
		     */
		while (cTrue)
		{
compMMRLeftWhite:
		    if (vacantBit >= 16)
		    {
			if (pCurCompData >= pCompDataEnd)
			    goto abortLeftWhite;

			    /* write out LS 2 bytes of accumulator and increment
			     * pointers */
			pCurCompData[0] = reverseByte[(UInt8) accum];
			pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
			accum >>= 16;		/* get rid of written bytes */
			pCurCompData += 2;
			vacantBit -= 16;
		    }

		    runLength = pCurRun->runLeft - lastX;
		    if (runLength <= 63)
		    {
			    /* Put out terminating code */
			accum |= mmrRevWhiteTermCodes[runLength].codeBits <<
								    vacantBit;
			vacantBit += mmrRevWhiteTermCodes[runLength].codeLength;
			    /* Set lastX so that we get the length of the
			     * next black run by subtracting lastX from
			     * pCurRun->runRight.
			     */
			lastX = pCurRun->runLeft - 1;
			break;
		    }
		    else	/* put out a non-terminating code */
		    if (runLength >= 2560)
		    {
			accum |=
			    mmrRevWhiteNonTermCodes[2560/64].codeBits <<
								   vacantBit;
			vacantBit +=
			    mmrRevWhiteNonTermCodes[2560/64].codeLength;
			lastX += 2560;
		    }
		    else
		    {
			    /* put out 64x code */
			accum |=
			    mmrRevWhiteNonTermCodes[runLength/64].codeBits <<
								   vacantBit;
			vacantBit +=
			    mmrRevWhiteNonTermCodes[runLength/64].codeLength;
			lastX += runLength & ~63;
		    }
		}

		    /* Here's the loop used to generate black runs when
		     * generating horizontal codes for left endpoints.
		     */
		while (cTrue)
		{
compMMRLeftBlack:
		    if (vacantBit >= 16)
		    {
			if (pCurCompData >= pCompDataEnd)
			    goto abortLeftBlack;

			    /* write out LS 2 bytes of accumulator and increment
			     * pointers */
			pCurCompData[0] = reverseByte[(UInt8) accum];
			pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
			accum >>= 16;		/* get rid of written bytes */
			pCurCompData += 2;
			vacantBit -= 16;
		    }

		    runLength = pCurRun->runRight - lastX;
		    if (runLength <= 63)
		    {
			    /* Put out terminating code */
			accum |= mmrRevBlackTermCodes[runLength].codeBits <<
								    vacantBit;
			vacantBit += mmrRevBlackTermCodes[runLength].codeLength;
			    /* We've now put out the horizontal code that
			     * takes us to just beyond the right endpoint
			     * of the current horizontal run.  We want to
			     * mark our current position  We'll loop back
			     * around to generate another left endpoint.
			     * The "break" gets us out of the loop that
			     * generates run length tokens for the black run.
			     */
			lastX = pCurRun->runRight + 1;
			    /* Having skipped past a black run, point to the
			     * next one in the coding line.
			     */
			pCurRun++;
			break;
		    }
		    else	/* put out a non-terminating code */
		    if (runLength >= 2560)
		    {
			accum |=
			    mmrRevBlackNonTermCodes[2560/64].codeBits <<
								   vacantBit;
			vacantBit +=
			    mmrRevBlackNonTermCodes[2560/64].codeLength;
			lastX += 2560;
		    }
		    else
		    {
			    /* put out 64x code */
			accum |=
			    mmrRevBlackNonTermCodes[runLength/64].codeBits <<
								   vacantBit;
			vacantBit +=
			    mmrRevBlackNonTermCodes[runLength/64].codeLength;
			lastX += runLength & ~63;
		    }
		}
	    }	/* End of generating horizontal codes for left endpoint. */
	}	/* End of loop that generates left endpoints */


	    /* ***************** RIGHT ENDPOINTS ******************/
	while (cTrue)
	{
		/* First, are we done encoding this scan line?
		 * This is true if we're off the right edge of the
		 * scan line.
		 */
	    if (lastX >= imWidth)
		goto compMMRGoodCompletion;

		/* Ok, there's more to do.  Find the next
		 * right endpoint in the reference line.
		 * There are two subtle tricks here.  The first
		 * one is that we really want to find the first
		 * WHITE pixel in the reference line that's immediately
		 * to the right a black pixel and to the right of
		 * lastX.  This test is pRefRun->runRight+1 > lastX.
		 * So, we keep going so long as pRefRun->runRight+1 <=
		 * lastX, or pRefRun->runRight < lastX.
		 * The second trick is that we may have to
		 * back up by one run in the reference line
		 * since the right endpoint we find now may be
		 * to the left of the left endpoint found in the
		 * previous case.  So always back up one just
		 * to be sure.  There's a dummy run at the begining
		 * of the scan line, so we won't fall off the
		 * beginning of the scan line.
		 */
	    pRefRun--;
	    while (pRefRun->runRight < lastX)
		pRefRun++;

compMMRRightEndpoint:
		/* Now, see if the accumulator has at least two bytes
		 * of data to write.  If so, write them.  If not,
		 * write one if possible and quit.  This is also
		 * where we restart if the buffer was full last time
		 * we got here.
		 */
	    if (vacantBit >= 16)
	    {
		if (pCurCompData >= pCompDataEnd)
		    goto abortRightEndpoint;
    
		    /* write out LS 2 bytes of accumulator and increment
		     * pointers */
		pCurCompData[0] = reverseByte[(UInt8) accum];
		pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
		accum >>= 16;			/* get rid of written bytes */
		pCurCompData += 2;
		vacantBit -= 16;
	    }

		    /* First see if we should issue a PASS code.  This
		     * happens if we're supposed to skip a little run
		     * of white on the reference line that overlaps a
		     * long black run on the coding line.  This
		     * is indicated by the left endpoint of the run to
		     * the right of the current reference run being
		     * at the same coord or to the left of the current
		     * run's right endpoint.  By limiting our move
		     * to no more than the current run's right endpoint,
		     * we guarantee that we stay in black pixels.  That's
		     * how you tell if issuing a PASS code is ok.
		     */
	    if (pCurRun->runRight >= pRefRun[1].runLeft)
	    {
		accum |= cMMRPass << vacantBit;
		vacantBit += cMMRPassLen;
		lastX = pRefRun[1].runLeft;	/* advance to passed
						 * position. */
		    /* We don't advance the coding run because we still
		     * want to encode its right endpoint.  So, we just
		     * fall through the other cases in the else-if chain
		     * and loop back around to generate another
		     * right endpoint.
		     */
	    }
	    else
		    /* See if we should use a vertical code instead.
		     * This happens when the distance from the
		     * reference run's right endpoint to the coding
		     * run's right endpoint is less than or equal to 3.
		     * The trick below is a way of doing comparison
		     * of absolute values.  We do an unsigned compare
		     * of (a - b + 3) <= 6.  If a > (b+3), then
		     * the sum is clearly >= 6.  If (a-3) < b, then
		     * (a - b + 3) produces a negative number, which
		     * in unsigned arithmetic looks like a big positive
		     * number, so it's still > 6.  In between, we get
		     * a positive number <= 6.
		     */
	    if ((UInt32)(pCurRun->runRight - pRefRun->runRight + 3) <= 6)
	    {
		runLength = pCurRun->runRight - pRefRun->runRight + 3;
		accum |= mmrRevVerticalCodes[runLength] << vacantBit;
		vacantBit += mmrRevVertCodeLengths[runLength];

		    /* Point past the black space we've just skipped */
		lastX = pCurRun->runRight + 1;
		    /* Advance pCurRun so it points to the next run
		     * in the coding line since we want to look
		     * for the next left endpoint now.
		     */
		pCurRun++;

		    /* Loop around to generate a left endpoint. */
		    /* THIS IS HOW WE GET TO THE LEFT ENDPOINT LOOP */
		break;
	    }
	    else
		/* Issue a horizontal code (2 edges) that takes us to
		 * the next right endpoint.
		 */
	    {
		    /* First, generate the horizontal code itself */
		accum |= cMMRHoriz << vacantBit;
		vacantBit += cMMRHorizLen;

		    /* Here's the loop used to generate codes for white
		     * space.  It begins with an attempt to make room
		     * in accum, then tries to encode one token in
		     * the white run length.  Each trip through the
		     * loop encodes one token.
		     */
		    /* Back lastX up by 1 so that we get the right
		     * number of black bits when subtracting lastX
		     * from pCurRun->runRight.
		     */
		lastX--;

		while (cTrue)
		{
compMMRRightBlack:
		    if (vacantBit >= 16)
		    {
			if (pCurCompData >= pCompDataEnd)
			    goto abortRightBlack;

			    /* write out LS 2 bytes of accumulator and increment
			     * pointers */
			pCurCompData[0] = reverseByte[(UInt8) accum];
			pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
			accum >>= 16;		/* get rid of written bytes */
			pCurCompData += 2;
			vacantBit -= 16;
		    }

		    runLength = pCurRun->runRight - lastX;
		    if (runLength <= 63)
		    {
			    /* Put out terminating code */
			accum |= mmrRevBlackTermCodes[runLength].codeBits <<
								    vacantBit;
			vacantBit += mmrRevBlackTermCodes[runLength].codeLength;
			    /* Point to the first white pixel so we get
			     * the correct run length for the white run
			     * by computing pCurRun[1].runLeft - lastX.
			     */
			lastX = pCurRun->runRight + 1;
			pCurRun++;
			break;		/* go write run of white pixels */
		    }
		    else	/* put out a non-terminating code */
		    if (runLength >= 2560)
		    {
			accum |=
			    mmrRevBlackNonTermCodes[2560/64].codeBits <<
								   vacantBit;
			vacantBit +=
			    mmrRevBlackNonTermCodes[2560/64].codeLength;
			lastX += 2560;
		    }
		    else
		    {
			    /* put out 64x code */
			accum |=
			    mmrRevBlackNonTermCodes[runLength/64].codeBits <<
								   vacantBit;
			vacantBit +=
			    mmrRevBlackNonTermCodes[runLength/64].codeLength;
			lastX += runLength & ~63;
		    }
		}

		    /* Here's the loop used to generate white runs when
		     * generating horizontal codes for right endpoints.
		     */
		while (cTrue)
		{
compMMRRightWhite:
		    if (vacantBit >= 16)
		    {
			if (pCurCompData >= pCompDataEnd)
			    goto abortRightWhite;

			    /* write out LS 2 bytes of accumulator and increment
			     * pointers */
			pCurCompData[0] = reverseByte[(UInt8) accum];
			pCurCompData[1] = reverseByte[(UInt8) (accum >> 8)];
			accum >>= 16;		/* get rid of written bytes */
			pCurCompData += 2;
			vacantBit -= 16;
		    }

		    runLength = pCurRun->runLeft - lastX;
		    if (runLength <= 63)
		    {
			    /* Put out terminating code */
			accum |= mmrRevWhiteTermCodes[runLength].codeBits <<
								    vacantBit;
			vacantBit += mmrRevWhiteTermCodes[runLength].codeLength;
			    /* Point to just beyond the end of the run
			     * of white pixels, i.e. the first black pixel.
			     * This is still the run for which we want to
			     * generate a right endpoint, so don't advance
			     * the run pointer.
			     */
			lastX = pCurRun->runLeft;
			break;		/* fall out of loop that generates white
					 * run length code */
		    }
		    else	/* put out a non-terminating code */
		    if (runLength >= 2560)
		    {
			accum |=
			    mmrRevWhiteNonTermCodes[2560/64].codeBits <<
								   vacantBit;
			vacantBit +=
			    mmrRevWhiteNonTermCodes[2560/64].codeLength;
			lastX += 2560;
		    }
		    else
		    {
			    /* put out 64x code */
			accum |=
			    mmrRevWhiteNonTermCodes[runLength/64].codeBits <<
								   vacantBit;
			vacantBit +=
			    mmrRevWhiteNonTermCodes[runLength/64].codeLength;
			lastX += runLength & ~63;
		    }
		}
	    }	/* End of generating horizontal codes for right endpoint. */
	}	/* End of loop that generates right endpoints */
    }		/* End of loop that generates codes for the whole line. */


	/* We get here when everything was just swell.  We've generated
	 * the codes for an entire line.
	 * We never add padding in MMR, so no need to do that.  We also
	 * don't flush the accumulator at the end of a scan line, so
	 * don't do that either.
	 * Just save the state so we know we're done with this scan
	 * line and can concatenate the codes for the next scan line
	 * when it comes.
	 */
compMMRGoodCompletion:
   *ppCompData = pCurCompData;
   pCCITTCompInfo->currentRun = 0;
   pCCITTCompInfo->refRun = 0;
   pCCITTCompInfo->remainingLength = 0;	/* used for lastX initialization */
   pCCITTCompInfo->compState = cCCITTCompNewLine;
   pCCITTCompInfo->accumBits = accum;
   pCCITTCompInfo->vacantBit = vacantBit;

   NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

	/* First, we have all the abort targets.  We do this weird
	 * thing of setting all the states here instead of above
	 * in the hopes that the compiler will generate better
	 * code in the real loops - just a test a jump down here
	 * where the normal case is straight ahead instead of
	 * a test and jump around the error stuff.
	 * We get here when we had at least two bytes to write
	 * to the compressed data stream, but there wasn't room
	 * for 2 bytes.  There might be room for 1 byte.
	 */
abortLeftEndpoint:
    pCCITTCompInfo->compState = cCCITTCompDoLeftEndpoint;
    goto compMMRMightWriteOne;
abortLeftWhite:
    pCCITTCompInfo->compState = cCCITTCompDoLeftWhite;
    goto compMMRMightWriteOne;
abortLeftBlack:
    pCCITTCompInfo->compState = cCCITTCompDoLeftBlack;
    goto compMMRMightWriteOne;
abortRightEndpoint:
    pCCITTCompInfo->compState = cCCITTCompDoRightEndpoint;
    goto compMMRMightWriteOne;
abortRightBlack:
    pCCITTCompInfo->compState = cCCITTCompDoRightBlack;
    goto compMMRMightWriteOne;
abortRightWhite:
    pCCITTCompInfo->compState = cCCITTCompDoRightWhite;
    goto compMMRMightWriteOne;

	/* We get here when we know that two bytes won't fit into the
	 * compressed output buffer.  Try putting one in.
	 * First, restore pCompDataEnd.  Remember that we decremented
	 * it at the top of the routine.
	 */
compMMRMightWriteOne:
    pCompDataEnd++;
    if (pCurCompData < pCompDataEnd)
    {
	pCurCompData[0] = reverseByte[(UInt8) accum];
	pCurCompData++;
	accum >>= 8;
	vacantBit -= 8;
    }

	/* This is where we end up when we abort because we ran out
	 * of space in the compression buffer.
	 */
    *ppCompData = pCurCompData;
    pCCITTCompInfo->currentRun = pCurRun - pFirstRun;
    pCCITTCompInfo->refRun = pRefRun - pFirstRefRun;
    pCCITTCompInfo->remainingLength = lastX;
    pCCITTCompInfo->accumBits = accum;
    pCCITTCompInfo->vacantBit = vacantBit;

    return ia_bufferExceeded;	/* indicates that we stopped because we
				 * ran out of space.
				 */
}


/*

 $Log: mimmrc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.3  1998/01/13 23:52:48  ddavies
 mimmrc.c -> /product/ipcore/ipshared/src/RCS/mimmrc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1996/05/06 18:50:33  ddavies
 mimmrc.c -> /product/ipcore/ipshared/src/RCS/mimmrc.c,v

 Fix comments.

 Revision 1.1  1996/05/01 23:27:29  ddavies
 Initial revision


 */

