/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mimmrd.c: low level code used in Modified Read and Modified Modified Read
 *	      (CCITT Group 3 and Group 4) decompression
 *
 *	Int32	ip_decompMMRRow()
 *			Decompresses one row of data.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#define _NEED_TOPLEVEL_CODES_
#define _NEED_HORIZONTAL_CODES_
#include "mmrdec.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mimmrd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* This routine decompresses one scan line of MMR (CCITT Group 4 or
 * CCITT 2-Dimensionally encoded) data.  It stops when
 * it runs out of compressed data.  It can restart in the same spot.
 * This is done by setting, and later dispatching on, restart states.
 * The states for decompression are listed in shrcmprs.prv.
 * cCCITTDecompNewLine is already state 0.  The remaining
 * state variables (number of scan lines left in the current block,
 * the 1-word decoding buffer, number of bits in that buffer, etc) are
 * also saved in the private state structure.
 */

/* This file contains the routine used to decompress one row of data using
 * Modified Modified Read encoding.  This method is described in
 * the CCITT Fascicle VII.3, Recommendation T.6.  The Recommendation
 * T.4 for Modified Read encoding has many useful examples of 2D encoding
 * as well.
 *
 * MMR is a pure 2 dimensional (2D) encoding standard.  Each line is
 * coded from left to right with respect to the line above it.  The
 * first line is assumed to be preceeded by a blank line.  The assumed
 * conditions at the left and right ends of the scan line are discussed
 * below.
 *
 * The decoding loop is a state machine, much like the encoding
 * loop.  In fact, we'll try to make them as similar as possible
 * to reduce the chance of error.  We go around the loop once for
 * each complete code put out or taken in.  I'd better explain
 * MMR coding in order for that to make sense.
 *
 * The standard draws several little pictures.  The "Reference Line"
 * is the line above the one being coded.  The "Coding Line" is the
 * image line being coded.  At each stage, we have a starting
 * position, lastX, in the coding line.  That is obvious, but
 * important.  The starting position in the coding line is
 * immediately to the right (since we encode left to right) of
 * the last pixel that was encoded.  An important point is that both
 * the encoder and the decoder can identify the starting position
 * since it reflects only things that have happened in the past.
 * As a hack that has ugly consequences, lastX begins on pixel -1,
 * the leftmost pixel of the line.  We'll discuss this after
 * talking about the normal encoding scheme.
 *
 * Both the encoder and decoder can see the reference line, so
 * let's start with that.  Starting at lastX in the reference
 * line, the encoder looks strictly to the right for the first pixel that a)
 * has a different color than the pixel to its left and b) has the
 * opposite color from the pixel at position lastX in the coding line.
 * Even though we haven't encoded lastX, we can deduce its
 * polarity from the sequence of coding state.
 *
 * For example, say we're about to encode a left endpoint of a black
 * run.  This means that the pixel in the coding line at position
 * lastX is white.  We look to the right of position lastX in the
 * reference line for the first left endpoint of a black run.
 * If we're going to encode a right endpoint, we're starting on a
 * black pixel in the coding line.  We look to the right of lastX
 * in the reference line for the white pixel just beyond the
 * nearest right endpoint.  The CCITT standard calls this first
 * endpoint on the reference line b1.  The location of the next
 * pixel to the left of b1 of opposite color from b1 is called b2.
 *
 * It's in the coding line that encoding and decoding become
 * different.  In encoding, we look to the right for the first
 * endpoint of proper left/right polarity.  Call this
 * location a1.  We then look to the right of a1 for the next pixel
 * of opposite color.  Call the location of this pixel a2.
 * When decoding, we just decode the next code which tells us what
 * the encoding process found.
 * Several things can happen at this point of encoding.
 *   1. There may be no pixels of the desired color (opposite color
 *	of coding line pixel at lastX) on the coding line under the 
 *	reference line run (b1,b2).  This is indicated by b2
 *	(one past the end of the run in the reference line) being 
 *	strictly to the left of a1 (the beginning of the run in the
 *	coding line).  We require a difference of at least one pixel
 *	between b2 and a1 to decide the run has really
 *	ended.  In this case, we issue a PASS code.  This tells
 *	both the encoder and decoder to skip the black or white run
 *	on the reference line by advancing lastX to b2.  Note that
 *	insisting on the 1-pixel buffer between b2 and a1 means
 *	that the color of the pixel under the new lastX will be
 *	the same as the color of the pixel under old lastX.  If
 *	we were looking for a left endpoint originally, we'll
 *	still be looking for one.  Ditto for right endpoints.
 *   2. If we didn't issue a PASS code, the next point on the
 *	reference line, b1, might be within distance 3 of the next
 *	point on the coding line, a1.  That is, | b1 - a1 | <= 3.
 *	If this is true, we use a vertical code to specify the 
 *	relative distance between them and move lastX up to a1's 
 *	position.  Since we're only moving past one endpoint, we swap 
 *	the polarity of the next endpoint to be encoded or decoded.
 *	For example, if we've just encoded a left endpoint, we`ll
 *	next encode a right endpoint.
 *	Group 4 encoding is based on the notion that this is
 *	the common case.
 *   3. If neither of those work, we issue a horizontal code
 *	followed by two distances.  One goes to the next endpoint, 
 *	the other goes to the endpoint after that.  Why two distances?
 *	Because we often want two insert two endpoints to get to
 *	the next vertical match.  A brand new run or a fork inserts
 *	two new endpoints.
 *
 * Here's a point that confused me.  Why do we always look strictly
 * to the right from lastX?  Say we have the following case where
 * each X represents a run of length 1:
 *
 *			X
 *		      X X X
 *
 * It would be nice to encode the second line using 6 vertical
 * codes (2 for each X - one for the left edge and one for the
 * right edge), all relative to the bit on the first
 * line.  However, the standard says that the pixel on the right
 * will be encoded relative to some other pixel on the reference
 * line because the obvious reference pixel isn't to the right of
 * the coding pixel.  Why did they do that?
 *
 * Take the following example:
 *		X	X
 *		     X  X  X
 *
 * Without the rule of looking strictly to the right, the code for
 * the coding line would be -2, -2, 0, 0, +2, +2.
 * Now say we have the following image:
 *		X	X
 *	      X		X X
 *
 * The encoding would be the SAME: -2, -2, 0, 0, +2, +2.  Ooops.
 * That's why they say you always have to look right for the reference run.
 *
 * Ok, but why *strictly* to the right?  Well, you have to decide
 * on some convention to distinguish
 *		X	X
 *	      X		X X
 * from
 *		X	X
 *	      X	X	  X
 *
 * With the rule, the code for the first example is V(-2), V(-2), PASS,
 * V(0), V(0), H(1, 1).  The code for the second case is V(-2), V(-2),
 * V(0), V(0), V(+2), H(1, someDistToTheRightOfTheLastMark).
 *
 * Here's an explanation of the ugly hack at the left edge of the
 * scan line.  What happens if black runs on both the reference and
 * coding lines touch the left end of the image?  If we don't start looking
 * from a position off the edge of the image, we won't see the
 * proper reference line endpoint because it won't be strictly to the
 * right.  If we do start with an imaginary pixel off to the left,
 * we have to remember not to include this imaginary pixel in the
 * first code.  Glug.  Well, that's exactly what the standard says
 * to do, subtract one from the first run length.  We'll finess this
 * by finding the leftmost color change points at the beginning of
 * the scan line, then setting lastX to 0 before calculating the
 * first code.
 *
 * Ok, what is lastX really used for?  It's used to decide where to
 * look for b1 and b2 on the reference line.  It's also used to encode
 * horizontal distances when those are needed.  It's not needed to
 * encode vertical codes or pass codes.
 *
 *			*	*	*
 *
 * There are several layers of encoding.  At the top, is a layer that
 * can find or generate one of the following codes (listed in bit
 * reversed order relative to the standard documents - remember that
 * in our special encoding world, one should read these from right to
 * left, i.e. <--<--<--< ):
 *
 *	Code	Meaning
 *	      1	Next coding line endpoint (left or right) is
 *		immediately below the next reference line endpoint.
 *	    110	Next coding line endpoint is 1 bit to the right of
 *		next reference line endpoint.
 *	    010	Next coding line endpoint is 1 bit to the left of
 *		next reference line the endpoint.
 *	    100	Horizontal distances to next two endpoints in
 *		coding line relative to previous point in *coding*
 *		line follow (see below)
 *	   1000	Pass code - move lastX to just past end of next run in
 *		reference line.
 *	 110000	Next coding line endpoint is 2 bits to the right of
 *		next reference line endpoint.
 *	 010000	Next coding line endpoint is 2 bits to the left of
 *		next reference line endpoint.
 *	1100000	Next coding line endpoint is 3 bits to the right of
 *		next reference line endpoint.
 *	0100000	Next coding line endpoint is 3 bits to the left of
 *		next reference line endpoint.
 *	1000000	G4 extension code.  If we find an extention, we quit.
 *
 * Each pass through the encoding loop generates one of these.
 * Similarly, each pass through the decoding loop decodes one of
 * these.  There's a decoding table, mmrRevTopLevelDecodes, used to
 * decide which of these has been seen.
 * If the horizontal code is generated, we also generate
 * distances to the next two endpoints before returning to the top
 * of the loop.
 *
 * We can be in either of two states at the top of the loop,
 * preparing to encode/decode a right endpoint or preparing to
 * encode/decode a left endpoint.  Unfortunately, finding the next
 * endpoint on the reference line is a little tricky.  Don't the
 * next right and left endpoint always belong to the same run?
 * No.  Look at this example:
 *
 *      RRRRRRRRRRRRR             SSSSSSSSSSSSSS
 *     AA  BBBBB  CCCCC        DD
 *
 * The reference left endpoint for run A is the left endpoint of
 * R.  The reference left endpoint for runs B, C and D is the left
 * endpoint of S.  The reference right endpoints of A and B is the
 * right endpoint of R.  The reference right endpoint of runs C and D
 * is the right endpoint of S.  So, for runs B and C, the reference
 * right endpoint is to the left of the reference left endpoint.
 * For runs A and D, it's the other way around.  How do we deal with
 * this?  Well, when looking for left endpoints, we know that we
 * always have to look at the run containing the current left or
 * right endpoint and runs to the right of that.  When looking for
 * right endpoints, we might have to back up by one run before
 * starting to look.  We just always back up one run, then start
 * looking since that's easy.
 * 
 * If we get a horizontal code when preparing to generate a
 * left endpoint, the next coordinates are for the left
 * endpoint and the next right endpoint (specifying an entire run).
 * If we're ready to generate a right endpoint, the two coordinates
 * finish off the right end of the current run and start the left
 * end of the next one.
 *
 * If a horizontal code pushes us beyond the right edge of the
 * bitmap, we stop.  If it's the first of the pair that pushes
 * over the edge, the second code is set to represent 0.
 *
 * The standard G3/G4 horizontal distance codes are a little weird.
 * There is one code used to specify blocks of white space
 * (used to cover the distance from a right endpoint of one run
 * to the left endpoint of the run to the right) and another
 * code used to specify blocks of black pixels.  Why two codes?
 * It's probably because black runs tend to be shorter than white runs.
 * The black code has fewer bits/distance than the white codes
 * for the small distances.  The opposite is true for large distances.
 *
 * Each of the codes has individual entries for all distances from
 * 0 through 63.  They also have individual entries for each
 * multiple of 64 up through 1728.  There is one common code for
 * spanning pixels in blocks of 1792, 1856 (units of 64), ..., 2560.
 * For distances greater than 2560, one concatenates 2560 codes
 * together.  Therefore, a single distances has the form:
 *	(<some number of 2560 codes>)
 *	(<black or white code for 64 through 2496 in blocks of 64))
 *	<terminating black or white code specifying a distance
 *	 from 0 through 63>
 *
 * Only the last code is required.  If the block length is an
 * exact multiple of 64, the terminating code specifies a distance
 * of 0.  If the block is an exact multiple of 2560, the middle
 * code is left out and the last code specifies 0.
 * If n * 2560 < distance < n * 2560 + 64, the middle code is also
 * left out and the terminating code specifies the offset from
 * the last multiple of 2560.
 *
 * As stated above, the black and white codes are different.
 * The white codes are nicely arranged so that using an 8 bit
 * lookup table during decoding will return the following info:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code).
 * If the code value is greater than 63, it is a non-terminating
 * code.
 * If the code length is 9 (the length of the longest white code),
 * it is a 64x code and the final value should be found by adding
 * the ninth bit * 64 to the code value.  If the answer is 1664,
 * Then the answer is really 1728 (weird - is 1664 so common that
 * it needs a special code?).  If the code length is > 9, this is
 * a 64x code shared with the black codes.  Use the black lookup
 * method as described next.
 *
 * The black codes have a nice feature.  Either the first four bits
 * are 0000, or they're not.  If they are not 0000, we can use a
 * lookup table indexed by 6 bits that encodes the 9 interesting
 * entries. These represent the distances 1 through 9.  Each entry has
 *	1. code value
 *	2. code length in bits.
 *
 * If the first four bits are 0000, we get rid of them and use an 8-bit
 * lookup table on the remaining bits.  The output of this
 * table is the same as the output of the white table described above:
 *	1. code value (distance represented)
 *	2. code length in bits (amount by which we should advance
 *	   to get to the next code after discarding the first 4 bits).
 * There are 13-bit black codes, but they are nicely arranged.  So,
 * if the code length after discarding the first 4 bits is 9, this
 * is a 64x code and the final value should be found by adding the
 * ninth bit * 64 to the code value.
 *
 * To save table space, we actually decode long codes 4 bits at a time.
 * The intermediate lookups return pointers to subsequent tables.
 * Since fewer than half of all combinations are actually used, this
 * saves quite a bit of table space.
 *
 * To encode a distance, we put out one 2560 code for each multiple
 * of 2560 in the distance, then the correct 64x code (if any) for
 * any non-zero multiples of 64 remaining, then the terminating code
 * for the remainder.  Only the terminating code is required.  To
 * decode a number, we read in codes, adding their values, until
 * a terminating code (one whose length is less than 64) is reached.
 *
 *			*	*	*
 *
 * So much for reading and writing numbers in horizontal codes.  How
 * does the decoding loop really work?  The encoding loop converts
 * runs on the current and preceeding lines into a sequence of
 * codes.  The decoding loop does the opposite.  The encoding loop
 * tests its current position with respect to runs on the preceeding
 * line to decode what codes to generate.  The decoding loop uses the
 * tokens read to update the current position and record run endpoints.
 *
 * Both the encoding and decoding loops have two main states:
 * generating/expecting a left endpoint (preparing to encode/decode a
 * white run) and generating/expecting a right endpoint (preparing to
 * encode/decode a black run).  These govern interpretation of input
 * runs during encoding and generation of output runs during decoding.
 * They also govern the generation/interpretation of horizontal codes -
 * either white then black if looking for a left endpoint or black then
 * white if looking for a right endpoint.
 *
 * All the runs for a scan line are held in an array of Run32 structs.
 * We actually keep two arrays and ping-pong back and forth.  One is
 * used as the coding line and one is the reference line.
 *
 * We start decoding by setting the runs in one array to simulate a
 * blank line.  We find the reference run endpoints (at the
 * end of the line in this case), move lastX up to position 0 and
 * begin decoding codes codes.  In the first line, they are all horizontal
 * codes.  The result of each code is to move lastX and often, though
 * not always, to set a run endpoint.
 *
 * The loop terminates when lastX >= the width of the image, i.e. we've
 * decoded an entire scan line.
 *
 * Let's look a what happens when we want to decode a new left
 * endpoint.
 *   1. Skip left endpoints on the reference line until we
 *	find one that's to the right of the current position.
 *   2. If we encounter a PASS code, move lastX to just beyond the
 *	right end of the reference run.  This skips the that reference
 *	run.  We then loop up to step 1 again since we're still
 *	looking for a left endpoint.
 *   3. If the next code is a vertical code, update the current
 *	coding line position to be the position of the reference run's
 *	left endpoint plus or minus the encoded vertical offset,
 *	record this position in the next coding line run's left
 *	endpoint and go decode a right endpoints.
 *   4. If the next code was a horizontal code, discard it and decode
 *	the subsequent number of white pixels and black pixels.  Add
 *	the number of white pixels to lastX and set the next coding
 *	run's left endpoint to the sum.  Add the subsequent length of
 *	black pixels to the new lastX to form the next lastX.  Set the
 *	coding line run's right endpoint to be the new lastX-1.  Go back to
 *	the top of the loop with the state still set to left endpoint.
 *
 * When generating right endpoints,
 *   1. Skip right endpoints on the reference line until we
 *	find one that's to the right of the current position.
 *	We do this by starting on the run to the left of the run
 *	containing the previous reference left or right endpoint
 *	and looking to the right for the next legal right endpoint.
 *   2. If the next token is a PASS code, set lastX to the left
 *	endpoint of the reference line run *to the right* of the
 *	current reference run (so we're still pointing to a black
 *	pixel) and go back to the top of the loop with
 *	the state still set to right endpoint.
 *   3. If the next token is a vertical code, set the new X position
 *	equal to the right endpoint found in step 1 plus or minus the
 *	encoded vertical offset, then add 1 to it so we're pointing
 *	at a white pixel.  Set the right endpoint of the next coding
 *	line run to the new lastX-1.  Change the state to left
 *	endpoint and go to the top of the left endpoint loop.
 *   4. If the token was a horizontal code, discard it and decode
 *	the subsequent number of black, then white pixels.  Set
 *	The right endpoint of the next coding run line equal to
 *	the sum of the original position plus the number of black
 *	pixels - 1.  Temporarily set lastX to the old lastX plus
 *	the number of black pixels. Set the left endpoint of the
 *	next coding line run to be the number of white pixels plus
 *	the temporary lastX.  Set the new lastX to this sum also.
 *	Go back to the top of the right endpoint loop.
 *
 * There's a subtle feature here.  We encode the distance from our
 * current position to the next position.  When we go from the
 * first black pixel of a run to the first white pixel of the next
 * run, we've encoded the length of the length of the black run.
 * The same thing happens when going from the first white pixel to the
 * first black pixel.  However, if a PASS code moves us inside a
 * run, we subsequently encode the remaining length of the run,
 * including the pixel we're starting on.
 *
 * Let's say something about the mechanics of composing and decoding
 * bitstreams.  In both cases, we keep the bitstream in a 32-bit
 * register.  When decoding, we keep a pointer to the first vacant
 * bit relative to the LS end of the word.  New bytes from the compressed
 * data stream are shifted left by this much and ORed into the
 * accumulator - to the right of the older codes.  Each time we get ready
 * to decode a code, we look at the value of the pointer.  If
 * it's less than or equal to 16, we insert two more bytes,
 * and add 16 from the pointer.  To decode codes, we mask the lower
 * bits of the accumulator and use the masked bits as an index into the tables
 * described above.
 *
 * Since MMR never requires line padding, we do not skip pad bits in the
 * accumulator at the bottom of the loop.  There is a special function
 * that can be used to flush the accumulator during encoding.  This
 * function is called only at the end of strips or tiles.
 *
 * Since we're shifting the accumulators right, the head of the bitstream
 * is in the LS bit of the accumulator.  All the tables assume that
 * bitstream should be read from left to right, not right to left.
 * To correct the, we reverse the order of the bits in every byte
 * read or written to the compressed data stream.
 *
 *		*	*	*
 *
 * We should also say something about starting and stopping and general
 * code structure.  There is an outer loop containing two main sections.
 * The outer loop has no explicit exit condition, it runs until
 * one of the internal sections says we're done.
 * 
 * One of the internal sections decodes left endpoints (runs of white
 * pixels to the left of left endpoints), the other decodes right
 * endpoints (runs of black pixels).  Since we don't always encode
 * left and right endpoints alternately (see description above), we
 * can't just cycle between these sections blindly.
 *
 * Each section is coded as a "while (cTrue)" loop.  When we do something
 * that keeps coding the same type of endpoint, we just loop around
 * in one of the inner loops.  When we want to encode the other type
 * of endpoint, we execute a "break" statement.  This drops us out of
 * the inner loop that encodes one type of endpoint and we fall
 * through to the inner loop for the other type of endpoint.
 *
 * At the top of each inner loop is section that detects whether we're
 * done with the line.  How do we detect this?  We simply test to
 * see if lastX >= width.  This means that we've decoded all the pixels
 * in the scan line.
 *
 * The section that looks for the end of the line is followed by
 * a section that finds the next left or right endpoint in the reference
 * line to the right of the current position in the coding line (to the
 * right of lastX).  After that comes a block of code that detects
 * whether the accumulator has room for at least 2 bytes. If so, two
 * bytes are read from the input bytestream.  If there are fewer than
 * 2 bytes remaining in the input bytestream, we record that we should restart
 * in this state and abort.   Finally, we decode the top level code
 * resulting in a PASS, a vertical code or a horizontal code.
 *
 * Since the run length codes can be large, we also have to check the
 * bitstream accumulator in those states.  The block that decodes
 * run length codes is also a pair of loops.  The first loop starts
 * by checking to see if the accumulator has space for 2 more bytes
 * and reads 2 bytes if so.  It then decodes one horizontal token.
 * As explained above, it may take several tokens to represent a single
 * run length.  Once the run length is fully decoded, the second
 * loop does the same for the next run length.  Since it is possible
 * to empty the output buffer any time data is read from it, each of
 * the loops also has a restart label.  Restarting is done by simply
 * dispatching on the compression state.
 *
 * Since we may use this routine in both MMR and MR decoding, the restart
 * states for this routine must be distinguished from those used in
 * 1D encoding.  These states are defined in shrcmprs.prv as
 * CCITTDecompression states.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_decompMMRRow()
 *
 *	This procedure decompresses one line of data using the 2D CCITT
 *	Huffman tables.  The compressed data is in natural byte order.
 *	The uncompressed data is in pixr order.
 *
 *	Input:
 *		pCCITTDecompInfo: pointer to decompression state.
 *		imWidth:	width of image (or tile or stripe) in
 *				pixels.
 *		flushPipeline:	If cTrue, this routine should be able to
 *				get to the end of the scan line using
 *				only data in the compressed data buffer.
 *				If set to cFalse, it's ok to stop and
 *				ask for more compressed data when we're
 *				starting to run out.  Additional data will
 *				be seamlessly concatenated to current data.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *		ppDummy:	pointer to pointer to the Run32 structure
 *				just after the last valid structure.
 *				*ppDummy is updated to point to the
 *				structure beyond the last valid one on return.
 *
 *	Action:
 *		Attempt to decompress a single scan line of data.
 *		The compressed data is assumed to be in natural byte order and
 *		is arbitrarily aligned, so should be read 1 byte at a
 *		time.
 *
 *		If the end of the scan line is reached, set
 *		pCCITTDecompInfo->decompState = cCCITTDecompNewLine.
 *		If we stop because the compressed data buffer empties,
 *		set pCCITTDecompInfo->decompState to the state which
 *		describes where we should restart and
 *		put the relevant byte counts and run pointers into
 *		*pCCITTDecompInfo.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be read.  Callers can use this
 *		to determine the number of bytes read so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pProc was NULL
 *		ia_bufferExceeded	stopped because we tried to read
 *					something to the compressed data
 *					buffer but it was empty
 *
 **************************************************************************/
Int32  CDECL 
ip_decompMMRRow (
    IPState    *pIPState,
    CCITTDecompressionInfo *pCCITTDecompInfo,
    Int32	imWidth,
    Bool	flushPipeline,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd,
    Run32     **ppDummy)
{
    PROLOGUE;
    UInt8	*pCurCompData;		/* pointer to current comp data */
    Run32	*pCurRun, *pFirstRun;	/* pointer to current and first
					 * Run32 structs in the coding line */
    Run32	*pRefRun, *pFirstRefRun; /* pointer to current and first
					 * Run32 structs in reference line */
    Int32	 lastX;			/* last x coord in line */
    UInt32	 accum;			/* holds bits to be decoded */
    Int8	 vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */
    UInt8	 codeIndex;		/* the fragment of the input stream
					 * currently being used to index
					 * decoding tables. */
    Int32	 codeVal;		/* value found in decoding tables */
    UInt32	 tokenVal;		/* a version of codeVal that's
					 * used when calculating
					 * non-terminating codes. */
    mmrTableEntry *pNonTermTable;	/* points to second level tables
					 * of non-terminating codes.
					 */

	/* first, initialize things */
    pCurCompData = *ppCompData;
    accum = pCCITTDecompInfo->accumBits;
    vacantBit = (Int8) pCCITTDecompInfo->vacantBit;
    lastX = pCCITTDecompInfo->remainingLength;
	/* Most of time time, we want to compare to
	 * pCompDataEnd-1, so decrement it here */
    pCompDataEnd--;

	/* point to current run by first pointing to first real run,
	 * (skip the [-1,-1] dummy) then offseting to current run.
	 * To do this, we figure out which run array is the coding line and
	 * which is the reference line.  pCCITTDecompInfo->run2ArrayIsCoding
	 * is set when the run2Array is the coding line.
	 */
    if (pCCITTDecompInfo->run2ArrayIsCoding)
    {
	pFirstRun = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				       (pCCITTDecompInfo->run2ArrayOffset) +
				       sizeof(Run32));
	pFirstRefRun = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				          (pCCITTDecompInfo->runArrayOffset) +
					  sizeof(Run32));
    }
    else	/* the first run array holds the coding line */
    {
	pFirstRun = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				       (pCCITTDecompInfo->runArrayOffset) +
				       sizeof(Run32));
	pFirstRefRun = (Run32 *)((UInt8 *)(pCCITTDecompInfo) +
				          (pCCITTDecompInfo->run2ArrayOffset) +
					  sizeof(Run32));
    }

    pCurRun = pFirstRun + pCCITTDecompInfo->currentRun;
    pRefRun = pFirstRefRun + pCCITTDecompInfo->refRun;

	/* see where we should start off */
    switch (pCCITTDecompInfo->decompState)
    {
    case cCCITTDecompNewLine:
	    /* Start at the beginning of the scan line.  Here's where we
	     * do the silly trick of looking for the leftmost left
	     * endpoints in the coding and reference lines and
	     * then set lastX to 0 instead of -1.
	     * Well, hey, since pRefRun already points
	     * to the first run in the reference line, that run must be
	     * the run containing the leftmost left endpoint.
	     * So all we have to do is set lastX to 0 and jump into
	     * the loop that calculates left endpoints at the spot
	     * AFTER it adjusts pRefRun.
	     * lastX was set to 0 from pCCITTDecompInfo->remainingLength
	     * above, so the jump is all we need.
	     * Just for the sake of paranoia, we check to see if
	     * we're already done at the beginning.
	     */
	if (lastX >= imWidth)
	    goto decompMMRGoodCompletion;
	else
	    goto decompMMRLeftEndpoint;

    case cCCITTDecompDoLeftEndpoint:
	goto decompMMRLeftEndpoint;

    case cCCITTDecompDoLeftWhite:
	goto decompMMRLeftWhite;

    case cCCITTDecompDoLeftBlack:
	goto decompMMRLeftBlack;

    case cCCITTDecompDoRightEndpoint:
	goto decompMMRRightEndpoint;

    case cCCITTDecompDoRightBlack:
	goto decompMMRRightBlack;

    case cCCITTDecompDoRightWhite:
	goto decompMMRRightWhite;

    default:	/* Restart state belongs to some other routine */
	NORMAL_RETURN;
    }

	/* ********************* MAIN LOOP ***********************/

	/* Here's the main loop.  It contains two inner loops, one
	 * for encoding left endpoints and one for encoding right
	 * endpoints.  We stop when we reach the end of the scan
	 * line or run out of space in the compressed data buffer.
	 */
    while (cTrue)
    {

	    /* ***************** LEFT ENDPOINTS ******************/
	while (cTrue)
	{
		/* First, are we done encoding this scan line?
		 * This is true if our current X position if off
		 * the right edge of the image.
		 */
	    if (lastX >= imWidth)
		goto decompMMRGoodCompletion;

		/* Ok, there's more to do.  Find the next left
		 * endpoint in the reference line */
	    while (pRefRun->runLeft <= lastX)
		pRefRun++;

		/* Here's where we start the scan line.  We wanted to
		 * skip adjusting the right and left reference
		 * endpoints since they're already correct at this time.
		 */
decompMMRLeftEndpoint:

		/* Now, see if the accumulator has room for at least more
		 * two bytes of data.  If so, read them.  If not,
		 * read one if possible and quit.  This is also
		 * where we restart if the buffer was empty last time
		 * we got here.
		 */
	    if (vacantBit <= 16)
	    {
		if (pCurCompData < pCompDataEnd)
		{
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    accum |= reverseByte[pCurCompData[1]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData += 2;
		}
		else
		{
			/* We're not allowed to fetch 2 new bytes.
			 * If there's one byte, get it.
			 */
		    if (pCurCompData == pCompDataEnd)
		    {
			/* There's one more byte in the data stream. Get it. */
			accum |= reverseByte[pCurCompData[0]] << vacantBit;
			vacantBit += 8;
			pCurCompData ++;
		    }
		    if (!flushPipeline)
			goto abortLeftEndpoint;
		}
	    }

		/* Now decode the first thing in the accumulator.  This
		 * should be a top level code, (PASS, one of the vertical
		 * codes or a horizontal code marker), so use the
		 * top level code table.
		 */
	    codeVal = mmrRevTopLevelDecodes[accum & cTopLevelCodeMask];

		/* Since we know that all the vertical offset constants
		 * are in the range 0..cVertP3Offset, just test to see
		 * if the value is less than or equal to cVertP3Offset.
		 */
	    if (codeVal <= cVertP3Offset)
	    {
		    /* It's a vertical offset.  First, get rid of the code */
		accum >>= mmrRevVertCodeLengths[codeVal];
		vacantBit -= mmrRevVertCodeLengths[codeVal];

		    /* Set the left endpoint of the current run.
		     * Also set current X position to be the endpoint we
		     * just found.  We were on a white pixel and
		     * this is the first black pixel, so that's the
		     * correct position.*/
		lastX = pCurRun->runLeft = pRefRun->runLeft + codeVal -
					   cVert0Offset;

		    /* Leave pCurRun pointing to the same run since we
		     * now want to generate a right endpoint of the
		     * same run.
		     * Skip down to generate a right endpoint.
		     */

		    /* THIS IS HOW WE GET TO THE RIGHT ENDPOINT LOOP. */
		break;
	    }
	    else
	    if (codeVal == cPassCode)
	    {
		    /* In this case, we're just advancing lastX and
		     * emptying the code out of the accumulator.  We
		     * advance it to just past the right endpoint of
		     * pRefRun.
		     */
		accum >>= cMMRPassLen;
		vacantBit -= cMMRPassLen;
		lastX = pRefRun->runRight + 1;
	    }
	    else
	    if (codeVal == cHorizCode)
	    {
		    /* In this case, there's a horizontal code, which
		     * we should delete, followed by an encoded run of
		     * white pixels, followed by an encoded run of
		     * black pixels.
		     */
		accum >>= cMMRHorizLen;
		vacantBit -= cMMRHorizLen;

		    /* Here's the loop that decodes some number of tokens
		     * used to represent a white run length.  At the top
		     * of the loop is a clause that tries to put two more
		     * bytes into the accumulator.  After that, we
		     * decode the code.
		     */
		    /* ************** LEFT WHITE RUN LOOP ******************/
		while (cTrue)
		{
decompMMRLeftWhite:
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
			        reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |=
			        reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data stream.
				 * Get it. */
				accum |=
				    reverseByte[pCurCompData[0]] << vacantBit;
				vacantBit += 8;
				pCurCompData ++;
			    }
			    if (!flushPipeline)
				goto abortLeftWhite;
			}
		    }

			/* When decoding white runs, we always use 8 bits to
			 * index a table.
			 * First, see if this is a terminating code yet.
			 */
		    codeIndex = (UInt8) accum & cWhiteTermDecodeMask;
		    codeVal = mmrRevWhiteTermDecodes[codeIndex];
		    if (codeVal <= 63)
		    {
			/* *********** Terminating white run code ********** */
			    /* set next run coord */
			lastX = pCurRun->runLeft = lastX + codeVal;
	
			    /* get length of code in bits */
			codeVal =
			    mmrRevWhiteTermDecodes[
				cTermDecodeTableLengthOffset + codeIndex];
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* HERE IS THE EXIT FROM THE WHITE RUN LOOP */
			break;
		    }
		    else	/* look up this token using the
				 * non-terminating tables */
		    {
			/* ********* Non Terminating white run code ********* */
			pNonTermTable =
			    mmrRevWhite1stNibbleNonTermDecodes[codeIndex & 0xF];
			if (pNonTermTable == NULL)
			    goto decompMMRDecodeFailure;
		
			    /* We got a valid table pointer using the bottom
			     * 4 bits.
			     * Advance past them and decode the next 4 bits.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & 0xF;
			if (pNonTermTable[codeIndex].codeLength > 5)
			{
				/* The code in accum is one of the common
				 * codes used to skip really big blocks of
				 * pixels.  To decode these, we look it up
				 * in the black non-terminating tables.
				 */
			    pNonTermTable =
				mmrRevBlack1stNibbleNonTermDecodes[codeIndex];
			    if (pNonTermTable == NULL)
				goto decompMMRDecodeFailure;
		
				/* Now we really have a pointer to the
				 * lower level non-terminating code table,
				 * so toss the nibble we just used.
				 */
			    accum >>= 4;
			    vacantBit -= 4;
			    codeIndex = (Int8) accum & 0xF;
			}
		
			    /* and look up the code length */
			codeVal = pNonTermTable[codeIndex].codeLength;
			if (codeVal == 0)
			    goto decompMMRDecodeFailure;
	    
			tokenVal = pNonTermTable[codeIndex].codeBits;
			if (codeVal == 5)	/* add 64 x value of 5th bit */
			{
			    tokenVal += (accum * 4) & 64;
			    if (tokenVal == cWeirdWhiteCode)
				tokenVal += 64;
			} 
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* now move to the left in the current scan line by
			     * the number of pixels just decoded in tokenVal.
			     * We keep doing this until we find a terminating
			     * code.
			     */
			lastX += tokenVal;
		    }	/* end of non-terminating white case */
		}	/* end of loop that decodes white values */

		    /* ************* LEFT BLACK RUN LOOP ************** */
		    /* Each time through this loop, we decode one more
		     * token.  At the top of the loop, we see if there's
		     * room for more stuff.
		     */
		while (cTrue)
		{
decompMMRLeftBlack:
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
				reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |=
				reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data stream.
				 * Get it. */
				accum |=
				    reverseByte[pCurCompData[0]] << vacantBit;
				vacantBit += 8;
				pCurCompData ++;
			    }
			    if (!flushPipeline)
				goto abortLeftBlack;
			}
		    }
	
			/* In the black world, terminating codes can be short
			 * or long.  First, see if we have a short terminating
			 * code.  If that doesn't work, try a long terminating
			 * code.  If that doesn't work, we'll try a
			 * non-terminating code.
			 */
		    if ((accum & cBlackLSBitsMask) != 0)	/* it's short */
		    {
			    /* get code length and index using the short code
			     * table */
			codeIndex = (Int8) accum & cShortBlackTermDecodeMask;
			    /* get code value */
			codeVal = mmrRevBlackShortTermDecodes[codeIndex];
			    /* get code length */
			codeIndex = (Int8) mmrRevBlackShortTermDecodes[
					cShortTermDecodeTableLengthOffset +
					codeIndex];
		    }
		    else
		    {
			    /* delete the LS four bits since they're all 0,
			     * and use the long terminating code table. */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & cBlackTermDecodeMask;
			codeVal = mmrRevBlackLongTermDecodes[codeIndex];
			codeIndex = (Int8) mmrRevBlackLongTermDecodes[
						cTermDecodeTableLengthOffset +
						codeIndex];
		    }
	
		    if (codeVal <= 63)
		    {
			/* ********** Terminating black run code *********** */
			    /* set lastX to point to the first white pixel and
			     * runRight to point to the last black pixel.
			     */
			lastX += codeVal;
			pCurRun->runRight = lastX - 1;
	
			accum >>= codeIndex;
			vacantBit -= codeIndex;

			    /* Having skipped past a black run, point to the
			     * next one in the coding line.
			     */
/* DEBUG */
#if 0
fprintf(stderr, "[%ld, %ld] ", pCurRun->runLeft, pCurRun->runRight);
#endif
			pCurRun++;

			    /* HERE IS THE EXIT FROM THE BLACK RUN LOOP */
			break;
		    }

		    else	/* look up this token using the
				 * non-terminating tables.  At this point, we've
				 * already deleted the LS four bits, so we have
				 * the right nibble for this table
				 */
		    {
			pNonTermTable =
			    mmrRevBlack1stNibbleNonTermDecodes[accum & 0xF];
			if (pNonTermTable == NULL)
			    goto decompMMRDecodeFailure;
		
			    /* We got a valid table pointer using the
			     * bottom 4 bits.  Advance past them and
			     * decode the next 4 bits.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & 0xF;
	
			    /* and look up the code length */
			codeVal = pNonTermTable[codeIndex].codeLength;
			if (codeVal == 0)
			    goto decompMMRDecodeFailure;
	    
			tokenVal = pNonTermTable[codeIndex].codeBits;
			if (codeVal == 5)	/* add 64 x value of 5th bit */
			    tokenVal += (accum * 4) & 64;
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* now move to the left in the current scan line by
			     * the number of pixels just decoded in tokenVal.
			     * We keep doing this until we find a terminating
			     * code.
			     */
			lastX += tokenVal;
	
		    }	/* end of non-terminating black code case */
		}	/* end of loop that decodes black values */
	    }	/* End of horizontal codes for left endpoints */
	    else
		    /* this was an unknown code */
		goto decompMMRDecodeFailure;
	}	/* End of loop that decodes left endpoints */


	    /* ***************** RIGHT ENDPOINTS ******************/
	while (cTrue)
	{
		/* First, are we done encoding this scan line?
		 * This is true if our current X position is off
		 * the right edge of the image.
		 */
	    if (lastX >= imWidth)
		goto decompMMRGoodCompletion;

		/* Ok, there's more to do.  Find the next
		 * right endpoint in the reference line.
		 * There are two subtle tricks here.  The first
		 * one is that we really want to find the first
		 * WHITE pixel in the reference line that's immediately
		 * to the right a black pixel and to the right of
		 * lastX.  This test is pRefRun->runRight+1 > lastX.
		 * So, we keep going so long as pRefRun->runRight+1 <=
		 * lastX, or pRefRun->runRight < lastX.
		 * The second trick is that we may have to
		 * back up by one run in the reference line
		 * since the right endpoint we find now may be
		 * to the left of the left endpoint found in the
		 * previous case.  So always back up one just
		 * to be sure.  There's a dummy run at the begining
		 * of the scan line, so we won't fall off the
		 * beginning of the scan line.
		 */
	    pRefRun--;
	    while (pRefRun->runRight < lastX)
		pRefRun++;

decompMMRRightEndpoint:
		/* Now, see if the accumulator has room for at least more
		 * two bytes of data.  If so, read them.  If not,
		 * read one if possible and quit.  This is also
		 * where we restart if the buffer was empty last time
		 * we got here.
		 */
	    if (vacantBit <= 16)
	    {
		if (pCurCompData < pCompDataEnd)
		{
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    accum |= reverseByte[pCurCompData[1]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData += 2;
		}
		else
		{
			/* We're not allowed to fetch 2 new bytes.
			 * If there's one byte, get it.
			 */
		    if (pCurCompData == pCompDataEnd)
		    {
			/* There's one more byte in the data stream. Get it. */
			accum |= reverseByte[pCurCompData[0]] << vacantBit;
			vacantBit += 8;
			pCurCompData ++;
		    }
		    if (!flushPipeline)
			goto abortRightEndpoint;
		}
	    }

		/* Now decode the first thing in the accumulator.  This
		 * should be a top level code, (PASS, one of the vertical
		 * codes or a horizontal code marker), so use the
		 * top level code table.
		 */
	    codeVal = mmrRevTopLevelDecodes[accum & cTopLevelCodeMask];

		/* Since we know that all the vertical offset constants
		 * are in the range 0..cVertP3Offset, just test to see
		 * if the value is less than or equal to cVertP3Offset.
		 */
	    if (codeVal <= cVertP3Offset)
	    {
		    /* first, get rid of the code */
		accum >>= mmrRevVertCodeLengths[codeVal];
		vacantBit -= mmrRevVertCodeLengths[codeVal];

		    /* Set the right endpoint of the current run.
		     * Also set current X position to be just past
		     * the endpoint we just found.  We were on a black
		     * pixel and we want to go to the next white
		     * pixel.*/
		pCurRun->runRight = pRefRun->runRight + codeVal -
				    cVert0Offset;
		lastX = pCurRun->runRight + 1;

		    /* Advance pCurRun so it points to the next run
		     * in the coding line since we want to look
		     * for the next left endpoint now.
		     */
/* DEBUG */
#if 0
fprintf(stderr, "[%ld, %ld] ", pCurRun->runLeft, pCurRun->runRight);
#endif
		pCurRun++;

		    /* THIS IS HOW WE GET TO THE LEFT ENDPOINT LOOP. */
		break;
	    }
	    else
	    if (codeVal == cPassCode)
	    {
		    /* We found a PASS code.  This happens if we're
		     * supposed to skip a little run of white on the
		     * reference line that overlaps a long black run
		     * on the coding line.  This is indicated by the
		     * left endpoint of the run to the right of the
		     * current reference run being at the same coord
		     * or to the left of the current run's right
		     * endpoint.  By limiting our move to no more
		     * than the current run's right endpoint,
		     * we guarantee that we stay in black pixels.  That's
		     * how you tell if issuing a PASS code is ok.
		     * In this case, we're just advancing lastX and
		     * emptying the code out of the accumulator.  We
		     * advance it to the left endpoint of the next
		     * reference run.
		     */
		accum >>= cMMRPassLen;
		vacantBit -= cMMRPassLen;
		lastX = pRefRun[1].runLeft;
	    }
	    else
	    if (codeVal == cHorizCode)
	    {
		    /* In this case, there's a horizontal code, which
		     * we should delete, followed by an encoded run of
		     * black pixels, followed by an encoded run of
		     * white pixels.
		     */
		accum >>= cMMRHorizLen;
		vacantBit -= cMMRHorizLen;

		    /* ************* RIGHT BLACK RUN LOOP ************** */
		    /* Each time through this loop, we decode one more
		     * token.  At the top of the loop, we see if there's
		     * room for more stuff.
		     */
		while (cTrue)
		{
decompMMRRightBlack:
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
				reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |=
				reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data stream.
				 * Get it. */
				accum |=
				    reverseByte[pCurCompData[0]] << vacantBit;
				vacantBit += 8;
				pCurCompData ++;
			    }
			    if (!flushPipeline)
				goto abortRightBlack;
			}
		    }
	
			/* In the black world, terminating codes can be short
			 * or long.  First, see if we have a short terminating
			 * code.  If that doesn't work, try a long terminating
			 * code.  If that doesn't work, we'll try a
			 * non-terminating code.
			 */
		    if ((accum & cBlackLSBitsMask) != 0)	/* it's short */
		    {
			    /* get code length and index using the short code
			     * table */
			codeIndex = (Int8) accum & cShortBlackTermDecodeMask;
			    /* get code value */
			codeVal = mmrRevBlackShortTermDecodes[codeIndex];
			    /* get code length */
			codeIndex = (Int8) mmrRevBlackShortTermDecodes[
					cShortTermDecodeTableLengthOffset +
					codeIndex];
		    }
		    else
		    {
			    /* delete the LS four bits since they're all 0,
			     * and use the long terminating code table. */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & cBlackTermDecodeMask;
			codeVal = mmrRevBlackLongTermDecodes[codeIndex];
			codeIndex = (Int8) mmrRevBlackLongTermDecodes[
						cTermDecodeTableLengthOffset +
						codeIndex];
		    }
	
		    if (codeVal <= 63)
		    {
			/* ********** Terminating black run code *********** */
			    /* set lastX to point to the first white pixel and
			     * runRight to point to the last black pixel.
			     */
			lastX += codeVal;
			pCurRun->runRight = lastX - 1;
	
			accum >>= codeIndex;
			vacantBit -= codeIndex;
	
			    /* Having skipped past a black run, point to the
			     * next one in the coding line.
			     */
/* DEBUG */
#if 0
fprintf(stderr, "[%ld, %ld] ", pCurRun->runLeft, pCurRun->runRight);
#endif
			pCurRun++;

			    /* HERE IS THE EXIT FROM THE BLACK RUN LOOP */
			break;
		    }

		    else	/* look up this token using the
				 * non-terminating tables.  At this point, we've
				 * already deleted the LS four bits, so we have
				 * the right nibble for this table
				 */
		    {
			pNonTermTable =
			    mmrRevBlack1stNibbleNonTermDecodes[accum & 0xF];
			if (pNonTermTable == NULL)
			    goto decompMMRDecodeFailure;
		
			    /* We got a valid table pointer using the
			     * bottom 4 bits.  Advance past them and
			     * decode the next 4 bits.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & 0xF;
	
			    /* and look up the code length */
			codeVal = pNonTermTable[codeIndex].codeLength;
			if (codeVal == 0)
			    goto decompMMRDecodeFailure;
	    
			tokenVal = pNonTermTable[codeIndex].codeBits;
			if (codeVal == 5)	/* add 64 x value of 5th bit */
			    tokenVal += (accum * 4) & 64;
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* now move to the left in the current scan line by
			     * the number of pixels just decoded in tokenVal.
			     * We keep doing this until we find a terminating
			     * code.
			     */
			lastX += tokenVal;
	
		    }	/* end of non-terminating black code case */
		}	/* end of loop that decodes black values */

		    /* Here's the loop that decodes some number of tokens
		     * used to represent a white run length.  At the top
		     * of the loop is a clause that tries to put two more
		     * bytes into the accumulator.  After that, we
		     * decode the code.
		     */
		    /* ************** RIGHT WHITE RUN LOOP ******************/
		while (cTrue)
		{
decompMMRRightWhite:
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
			        reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |=
			        reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data stream.
				 * Get it. */
				accum |=
				    reverseByte[pCurCompData[0]] << vacantBit;
				vacantBit += 8;
				pCurCompData ++;
			    }
			    if (!flushPipeline)
				goto abortRightWhite;
			}
		    }

			/* When decoding white runs, we always use 8 bits to
			 * index a table.
			 * First, see if this is a terminating code yet.
			 */
		    codeIndex = (UInt8) accum & cWhiteTermDecodeMask;
		    codeVal = mmrRevWhiteTermDecodes[codeIndex];
		    if (codeVal <= 63)
		    {
			/* *********** Terminating white run code ********** */
			    /* set next run coord */
			lastX = pCurRun->runLeft = lastX + codeVal;
	
			    /* get length of code in bits */
			codeVal =
			    mmrRevWhiteTermDecodes[
				cTermDecodeTableLengthOffset + codeIndex];
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;

			    /* Continue to leave pCurRun pointing to
			     * the current run since we've got to go
			     * find it's right endpoint.  We just loop
			     * back around in right endpoint loop after
			     * falling out of this loop and the switch
			     * statement.
			     */
	
			    /* HERE IS THE EXIT FROM THE WHITE RUN LOOP */
			break;
		    }
		    else	/* look up this token using the
				 * non-terminating tables */
		    {
			/* ********* Non Terminating white run code ********* */
			pNonTermTable =
			    mmrRevWhite1stNibbleNonTermDecodes[codeIndex & 0xF];
			if (pNonTermTable == NULL)
			    goto decompMMRDecodeFailure;
		
			    /* We got a valid table pointer using the bottom
			     * 4 bits.
			     * Advance past them and decode the next 4 bits.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & 0xF;
			if (pNonTermTable[codeIndex].codeLength > 5)
			{
				/* The code in accum is one of the common
				 * codes used to skip really big blocks of
				 * pixels.  To decode these, we look it up
				 * in the black non-terminating tables.
				 */
			    pNonTermTable =
				mmrRevBlack1stNibbleNonTermDecodes[codeIndex];
			    if (pNonTermTable == NULL)
				goto decompMMRDecodeFailure;
		
				/* Now we really have a pointer to the
				 * lower level non-terminating code table,
				 * so toss the nibble we just used.
				 */
			    accum >>= 4;
			    vacantBit -= 4;
			    codeIndex = (Int8) accum & 0xF;
			}
		
			    /* and look up the code length */
			codeVal = pNonTermTable[codeIndex].codeLength;
			if (codeVal == 0)
			    goto decompMMRDecodeFailure;
	    
			tokenVal = pNonTermTable[codeIndex].codeBits;
			if (codeVal == 5)	/* add 64 x value of 5th bit */
			{
			    tokenVal += (accum * 4) & 64;
			    if (tokenVal == cWeirdWhiteCode)
				tokenVal += 64;
			} 
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* now move to the left in the current scan line by
			     * the number of pixels just decoded in tokenVal.
			     * We keep doing this until we find a terminating
			     * code.
			     */
			lastX += tokenVal;
		    }	/* end of non-terminating white case */
		}	/* end of loop that decodes white values */
	    }	/* end of horizontal code for right endpoints */
	    else
		    /* this was an unknown code */
		goto decompMMRDecodeFailure;
	}		/* end of loop that decodes right endpoints */
    }	/* End of main loop that decodes whole line */

	/* We get here when everything was just swell.  We've generated
	 * the codes for an entire line.
	 * We never add padding in MMR, so no need to do that.  We also
	 * don't flush the accumulator at the end of a scan line, so
	 * don't do that either.
	 * Just save the state so we know we're done with this scan
	 * line and can concatenate the codes for the next scan line
	 * when it comes.
	 * BUG FIX: Very rarely, the encoder had to encode a pair of
	 * horizontal runs where the second run was black AND that
	 * second run was null because the white run actually touched
	 * the right edge of the image.  If that happens, don't return
	 * that last run, it will just confuse everyone.  This is
	 * a good place to put the check since it doesn't happen very
	 * often.
	 */
decompMMRGoodCompletion:
   *ppCompData = pCurCompData;
   if (pCurRun[-1].runLeft > pCurRun[-1].runRight)
       pCurRun--;
   *ppDummy = pCurRun;
   pCCITTDecompInfo->currentRun = 0;
   pCCITTDecompInfo->refRun = 0;
   pCCITTDecompInfo->remainingLength = 0;   /* used for lastX initialization */
   pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
   pCCITTDecompInfo->accumBits = accum;
   pCCITTDecompInfo->vacantBit = vacantBit;

   NORMAL_RETURN;	/* It really worked */


	/****************** ABORT *******************/

	/* We get here if we were about to have a decode failure.
	 * If flushPipeline is set and accum is 0, then we
	 * really just ran out of data, so return ia_bufferExceeded
	 * instead.  If the upper level sees an ia_bufferExceeded
	 * with flushPipeline, it knows that the client didn't
	 * supply enough compressed data and we're doomed.
	 */
decompMMRDecodeFailure:
    if ((flushPipeline) && (accum == 0))
	return ia_bufferExceeded;
    else
	return ia_decodeFailure;

	/* Next, we have all the abort targets.  We do this weird
	 * thing of setting all the states here instead of above
	 * in the hopes that the compiler will generate better
	 * code in the real loops - just a test a jump down here
	 * where the normal case is straight ahead instead of
	 * a test and jump around the error stuff.
	 * We get here when we had at least two bytes to write
	 * to the compressed data stream, but there wasn't room
	 * for 2 bytes.  There might be room for 1 byte.
	 */
abortLeftEndpoint:
    pCCITTDecompInfo->decompState = cCCITTDecompDoLeftEndpoint;
    goto decompMMRAbortFinish;
abortLeftWhite:
    pCCITTDecompInfo->decompState = cCCITTDecompDoLeftWhite;
    goto decompMMRAbortFinish;
abortLeftBlack:
    pCCITTDecompInfo->decompState = cCCITTDecompDoLeftBlack;
    goto decompMMRAbortFinish;
abortRightEndpoint:
    pCCITTDecompInfo->decompState = cCCITTDecompDoRightEndpoint;
    goto decompMMRAbortFinish;
abortRightBlack:
    pCCITTDecompInfo->decompState = cCCITTDecompDoRightBlack;
    goto decompMMRAbortFinish;
abortRightWhite:
    pCCITTDecompInfo->decompState = cCCITTDecompDoRightWhite;
    goto decompMMRAbortFinish;

	/* This is where we end up when we abort because we ran out
	 * of bytes in the compression buffer and we weren't supposed
	 * to flush the compression pipeline.
	 */
decompMMRAbortFinish:
    *ppCompData = pCurCompData;
    *ppDummy = pCurRun + 1;	/* Don't scribble on the current run */
    pCCITTDecompInfo->currentRun = pCurRun - pFirstRun;
    pCCITTDecompInfo->refRun = pRefRun - pFirstRefRun;
    pCCITTDecompInfo->remainingLength = lastX;
    pCCITTDecompInfo->accumBits = accum;
    pCCITTDecompInfo->vacantBit = vacantBit;
    
    return ia_bufferExceeded;	/* indicates that we stopped because we
				 * ran out of bytes.
				 */
}



/*

 $Log: mimmrd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.3  1998/01/13 23:52:49  ddavies
 mimmrd.c -> /product/ipcore/ipshared/src/RCS/mimmrd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1996/05/21 18:52:31  ddavies
 mimmrd.c -> /product/ipcore/ipshared/src/RCS/mimmrd.c,v

 A minor bug fix: when the encoder is forced to encode a white/black run
 pair and the white run touches the right edge of the image, the length
 of the black run is 0.  Routines that interpret the array of runs get
 confused by this.  We now delete this last null run if it exists.

 Revision 1.1  1996/05/06 18:43:00  ddavies
 Initial revision


 */


