/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mimmrd.c: low level code used in Modified Read and Modified Modified Read
 *	      (CCITT Group 3 and Group 4) decompression
 *
 *	Int32	ip_decompMMRRowList()
 *			Decompresses one row of data to a linked list of runs.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"
#define _NEED_TOPLEVEL_CODES_
#define _NEED_HORIZONTAL_CODES_
#include "mmrdec.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"

    /*  prototyes */
#include "shrcmprs.prv"

/* This routine is a lot like ip_decompMMRRow except the result ends up
 * in a linked list of TCScanLineRuns instead of in an array of Run32's.
 * This routine is used when converting an MMR TIFF file directly into
 * a set of connected components.  The connected component code needs
 * to have the runs in a linked list.
 *
 * Please read the giant comment block in mimmrd.c, this code is almost
 * the same.
 *
 * The main difference is that we are passed pointers to the lists of runs
 * making up the coding and reference lines instead of finding the arrays
 * of runs at the end of the CCITTDecompressionInfo structure.  This
 * is in some ways simpler.  We also advance along by following pointers
 * instead of incrementing a source pointer.
 *
 * When we create a new run, we use i_allocateStruct if there's no new
 * run at the end of the existing list.  Well, ok, we really use the
 * macro ALLOCATE_STRUCT.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_decompMMRRowList()
 *
 *	This procedure decompresses one line of data using the 2D CCITT
 *	Huffman tables.  The compressed data is in natural byte order.
 *	The resulting runs are put into a linked list of TCScanLineRuns.
 *
 *	Input:
 *		pCCITTDecompInfo: pointer to decompression state.
 *		imWidth:	width of image (or tile or stripe) in
 *				pixels.  imWidth does NOT include frame.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *		pRunHeap:	Pointer to heap from which TCScanLineRuns
 *				are allocated.
 *		pCurRun:	pointer to first run in the list
 *				of coding runs.
 *		pRefRun:	pointer to first run in the list of reference
 *				runs.
 *		ppDummy:	pointer to pointer to the TCScanLineRun
 *				structure just after the last valid structure.
 *				*ppDummy is updated to point to the
 *				structure beyond the last valid one on return.
 *
 *	Action:
 *		Attempt to decompress a single scan line of data.
 *		The compressed data is assumed to be in natural byte order and
 *		is arbitrarily aligned, so should be read 1 byte at a
 *		time.
 *
 *		If the end of the scan line is reached, set
 *		pCCITTDecompInfo->decompState = cCCITTDecompNewLine.
 *		If the end of the compressed data is reached before
 *		the end of the scan line, simply return with
 *		ia_bufferExceeded.  This routine is not as general
 *		as ip_decompMMRRow and we cannot restart after running
 *		out of data.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be read.  Callers can use this
 *		to determine the number of bytes read so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pProc was NULL
 *		ia_bufferExceeded	stopped because we tried to read
 *					something to the compressed data
 *					buffer but it was empty
 *
 **************************************************************************/
Int32  CDECL 
ip_decompMMRRowList (
    IPState    *pIPState,
    CCITTDecompressionInfo *pCCITTDecompInfo,
    Int32	imWidth,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd,
    TCBlockHeader *pRunHeap,
    TCScanLineRun *pCurRun,
    TCScanLineRun *pRefRun,
    TCScanLineRun **ppDummy)
{
    PROLOGUE;
    UInt8	  *pCurCompData;	/* pointer to current comp data */
    TCScanLineRun *pPrevRun;		/* often points to run before pCurRun */
    TCScanLineRun *pPrevRefRun;		/* ptr to previous reference run */
    Int32	   lastX;		/* last x coord in line */
    UInt32	   accum;		/* holds bits to be decoded */
    Int8	   vacantBit;		/* offset first vacant bit from
					 * LS end of accum. */
    UInt8	   codeIndex;		/* the fragment of the input stream
					 * currently being used to index
					 * decoding tables. */
    Int32	   codeVal;		/* value found in decoding tables */
    UInt32	   tokenVal;		/* a version of codeVal that's
					 * used when calculating
					 * non-terminating codes. */
    
    mmrTableEntry *pNonTermTable;	/* points to second level tables
					 * of non-terminating codes.
					 */

	/* first, initialize things */
    pPrevRun = NULL;
    pPrevRefRun = pRefRun;		/* When we look for reference runs,
					 * we sometimes back up a run.
					 * That's what pPrevRefRun is for.
					 * The very first time, we don't
					 * want to back up at all.
					 */
    pCurCompData = *ppCompData;
    accum = pCCITTDecompInfo->accumBits;
    vacantBit = (Int8) pCCITTDecompInfo->vacantBit;
    lastX = pCCITTDecompInfo->remainingLength;
	/* Most of time time, we want to compare to
	 * pCompDataEnd-1, so decrement it here */
    pCompDataEnd--;

	/* Skip over the part where we try to find the initial
	 * reference run since if it starts at 0 and lastX is
	 * 0, we do NOT want to skip the first ref run.
	 */
    goto decompMMRStart;


	/* ********************* MAIN LOOP ***********************/

	/* Here's the main loop.  It contains two inner loops, one
	 * for encoding left endpoints and one for encoding right
	 * endpoints.  We stop when we reach the end of the scan
	 * line or run out of space in the compressed data buffer.
	 */
    while (cTrue)
    {

	    /* ***************** LEFT ENDPOINTS ******************/
	while (cTrue)
	{
		/* First, are we done encoding this scan line?
		 * This is true if our current X position if off
		 * the right edge of the image.
		 */
	    if (lastX >= imWidth)
		goto decompMMRGoodCompletion;

		/* Ok, there's more to do.  Find the next left
		 * endpoint in the reference line */
	    while (pRefRun->runLeft <= lastX)
	    {
	        pPrevRefRun = pRefRun;
		pRefRun = pRefRun->pNext;
	    }

		/* Here's where we start the scan line.
		 */
decompMMRStart:
		/* See if the accumulator has room for at least more
		 * two bytes of data.  If so, read them.  If not,
		 * read one if possible.
		 */
	    if (vacantBit <= 16)
	    {
		if (pCurCompData < pCompDataEnd)
		{
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    accum |= reverseByte[pCurCompData[1]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData += 2;
		}
		else
		{
			/* We're not allowed to fetch 2 new bytes.
			 * If there's one byte, get it.
			 */
		    if (pCurCompData == pCompDataEnd)
		    {
			/* There's one more byte in the data stream. Get it. */
			accum |= reverseByte[pCurCompData[0]] << vacantBit;
			vacantBit += 8;
			pCurCompData++;
		    }
		}
	    }

		/* Now decode the first thing in the accumulator.  This
		 * should be a top level code, (PASS, one of the vertical
		 * codes or a horizontal code marker), so use the
		 * top level code table.
		 */
	    codeVal = mmrRevTopLevelDecodes[accum & cTopLevelCodeMask];

		/* Since we know that all the vertical offset constants
		 * are in the range 0..cVertP3Offset, just test to see
		 * if the value is less than or equal to cVertP3Offset.
		 */
	    if (codeVal <= cVertP3Offset)
	    {
		    /* It's a vertical offset.  First, get rid of the code */
		accum >>= mmrRevVertCodeLengths[codeVal];
		vacantBit -= mmrRevVertCodeLengths[codeVal];

		    /* Set the left endpoint of the current run.
		     * Also set current X position to be the endpoint we
		     * just found.  We were on a white pixel and
		     * this is the first black pixel, so that's the
		     * correct position.*/
		pCurRun->runLeft = pRefRun->runLeft + (Int16) codeVal -
				   cVert0Offset;
		lastX = pCurRun->runLeft;
		pCurRun->pClient = NULL;

		    /* Leave pCurRun pointing to the same run since we
		     * now want to generate a right endpoint of the
		     * same run.
		     * Skip down to generate a right endpoint.
		     */

		    /* THIS IS HOW WE GET TO THE RIGHT ENDPOINT LOOP. */
		break;
	    }
	    else
	    if (codeVal == cPassCode)
	    {
		    /* In this case, we're just advancing lastX and
		     * emptying the code out of the accumulator.  We
		     * advance it to just past the right endpoint of
		     * pRefRun.
		     */
		accum >>= cMMRPassLen;
		vacantBit -= cMMRPassLen;
		lastX = pRefRun->runRight + 1;
	    }
	    else
	    if (codeVal == cHorizCode)
	    {
		    /* In this case, there's a horizontal code, which
		     * we should delete, followed by an encoded run of
		     * white pixels, followed by an encoded run of
		     * black pixels.
		     */
		accum >>= cMMRHorizLen;
		vacantBit -= cMMRHorizLen;

		    /* Here's the loop that decodes some number of tokens
		     * used to represent a white run length.  At the top
		     * of the loop is a clause that tries to put two more
		     * bytes into the accumulator.  After that, we
		     * decode the code.
		     */
		    /* ************** LEFT WHITE RUN LOOP ******************/
		while (cTrue)
		{
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
			        reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |=
			        reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data stream.
				 * Get it. */
				accum |=
				    reverseByte[pCurCompData[0]] << vacantBit;
				vacantBit += 8;
				pCurCompData++;
			    }
			}
		    }

			/* When decoding white runs, we always use 8 bits to
			 * index a table.
			 * First, see if this is a terminating code yet.
			 */
		    codeIndex = (UInt8) accum & cWhiteTermDecodeMask;
		    codeVal = mmrRevWhiteTermDecodes[codeIndex];
		    if (codeVal <= 63)
		    {
			/* *********** Terminating white run code ********** */
			    /* set next run coord */
			lastX += codeVal;
			pCurRun->runLeft = (Int16) lastX;
			pCurRun->pClient = NULL;

			    /* get length of code in bits */
			codeVal =
			    mmrRevWhiteTermDecodes[
				cTermDecodeTableLengthOffset + codeIndex];
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;

			    /* HERE IS THE EXIT FROM THE WHITE RUN LOOP */
			break;
		    }
		    else	/* look up this token using the
				 * non-terminating tables */
		    {
			/* ********* Non Terminating white run code ********* */
			pNonTermTable =
			    mmrRevWhite1stNibbleNonTermDecodes[codeIndex & 0xF];
			if (pNonTermTable == NULL)
			    return ia_decodeFailure;
		
			    /* We got a valid table pointer using the bottom
			     * 4 bits.
			     * Advance past them and decode the next 4 bits.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & 0xF;
			if (pNonTermTable[codeIndex].codeLength > 5)
			{
				/* The code in accum is one of the common
				 * codes used to skip really big blocks of
				 * pixels.  To decode these, we look it up
				 * in the black non-terminating tables.
				 */
			    pNonTermTable =
				mmrRevBlack1stNibbleNonTermDecodes[codeIndex];
			    if (pNonTermTable == NULL)
				return ia_decodeFailure;
		
				/* Now we really have a pointer to the
				 * lower level non-terminating code table,
				 * so toss the nibble we just used.
				 */
			    accum >>= 4;
			    vacantBit -= 4;
			    codeIndex = (Int8) accum & 0xF;
			}
		
			    /* and look up the code length */
			codeVal = pNonTermTable[codeIndex].codeLength;
			if (codeVal == 0)
			    return ia_decodeFailure;
	    
			tokenVal = pNonTermTable[codeIndex].codeBits;
			if (codeVal == 5)	/* add 64 x value of 5th bit */
			{
			    tokenVal += (accum * 4) & 64;
			    if (tokenVal == cWeirdWhiteCode)
				tokenVal += 64;
			} 
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* now move to the left in the current scan line by
			     * the number of pixels just decoded in tokenVal.
			     * We keep doing this until we find a terminating
			     * code.
			     */
			lastX += tokenVal;
		    }	/* end of non-terminating white case */
		}	/* end of loop that decodes white values */

		    /* ************* LEFT BLACK RUN LOOP ************** */
		    /* Each time through this loop, we decode one more
		     * token.  At the top of the loop, we see if there's
		     * room for more stuff.
		     */
		while (cTrue)
		{
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
				reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |=
				reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data stream.
				 * Get it. */
				accum |=
				    reverseByte[pCurCompData[0]] << vacantBit;
				vacantBit += 8;
				pCurCompData++;
			    }
			}
		    }
	
			/* In the black world, terminating codes can be short
			 * or long.  First, see if we have a short terminating
			 * code.  If that doesn't work, try a long terminating
			 * code.  If that doesn't work, we'll try a
			 * non-terminating code.
			 */
		    if ((accum & cBlackLSBitsMask) != 0)	/* it's short */
		    {
			    /* get code length and index using the short code
			     * table */
			codeIndex = (Int8) accum & cShortBlackTermDecodeMask;
			    /* get code value */
			codeVal = mmrRevBlackShortTermDecodes[codeIndex];
			    /* get code length */
			codeIndex = (Int8) mmrRevBlackShortTermDecodes[
					cShortTermDecodeTableLengthOffset +
					codeIndex];
		    }
		    else
		    {
			    /* delete the LS four bits since they're all 0,
			     * and use the long terminating code table. */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & cBlackTermDecodeMask;
			codeVal = mmrRevBlackLongTermDecodes[codeIndex];
			codeIndex = (Int8) mmrRevBlackLongTermDecodes[
						cTermDecodeTableLengthOffset +
						codeIndex];
		    }
	
		    if (codeVal <= 63)
		    {
			/* ********** Terminating black run code *********** */
			    /* set lastX to point to the first white pixel and
			     * runRight to point to the last black pixel.
			     */
			lastX += codeVal;
			pCurRun->runRight = (Int16) lastX - 1;
	
			accum >>= codeIndex;
			vacantBit -= codeIndex;

			    /* Having skipped past a black run, point to the
			     * next one in the coding line.
			     * BUT FIRST, this is the only case in which
			     * we might encode a run of length 0, which
			     * confuses thing later.  Remember this run
			     * just in case it's the last run in the
			     * scan line and we have to back up.
			     */
			pPrevRun = pCurRun;
			if (pCurRun->pNext == NULL)
			{
			    ALLOCATE_STRUCT(pRunHeap, TCScanLineRun,
					    pCurRun->pNext);
			    if (__error__ != ia_successful)
				return ia_nomem;
			    pCurRun = pCurRun->pNext;
			    pCurRun->pNext = NULL;
			    pCurRun->pClient = NULL;
			}
			else
			    pCurRun = pCurRun->pNext;

			    /* HERE IS THE EXIT FROM THE BLACK RUN LOOP */
			break;
		    }

		    else	/* look up this token using the
				 * non-terminating tables.  At this point, we've
				 * already deleted the LS four bits, so we have
				 * the right nibble for this table
				 */
		    {
			pNonTermTable =
			    mmrRevBlack1stNibbleNonTermDecodes[accum & 0xF];
			if (pNonTermTable == NULL)
			    return ia_decodeFailure;
		
			    /* We got a valid table pointer using the
			     * bottom 4 bits.  Advance past them and
			     * decode the next 4 bits.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & 0xF;
	
			    /* and look up the code length */
			codeVal = pNonTermTable[codeIndex].codeLength;
			if (codeVal == 0)
			    return ia_decodeFailure;
	    
			tokenVal = pNonTermTable[codeIndex].codeBits;
			if (codeVal == 5)	/* add 64 x value of 5th bit */
			    tokenVal += (accum * 4) & 64;
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* now move to the left in the current scan line by
			     * the number of pixels just decoded in tokenVal.
			     * We keep doing this until we find a terminating
			     * code.
			     */
			lastX += tokenVal;
	
		    }	/* end of non-terminating black code case */
		}	/* end of loop that decodes black values */
	    }	/* End of horizontal codes for left endpoints */
	    else
		    /* this was an unknown code */
		return ia_decodeFailure;
	}	/* End of loop that decodes left endpoints */


	    /* ***************** RIGHT ENDPOINTS ******************/
	while (cTrue)
	{
		/* First, are we done encoding this scan line?
		 * This is true if our current X position is off
		 * the right edge of the image.
		 */
	    if (lastX >= imWidth)
		goto decompMMRGoodCompletion;

		/* Ok, there's more to do.  Find the next
		 * right endpoint in the reference line.
		 * There are two subtle tricks here.  The first
		 * one is that we really want to find the first
		 * WHITE pixel in the reference line that's immediately
		 * to the right a black pixel and to the right of
		 * lastX.  This test is pRefRun->runRight+1 > lastX.
		 * So, we keep going so long as pRefRun->runRight+1 <=
		 * lastX, or pRefRun->runRight < lastX.
		 * The second trick is that we may have to
		 * back up by one run in the reference line
		 * since the right endpoint we find now may be
		 * to the left of the left endpoint found in the
		 * previous case.  So always back up one just
		 * to be sure.  There's a dummy run at the begining
		 * of the scan line, so we won't fall off the
		 * beginning of the scan line.
		 */
	    pRefRun = pPrevRefRun;
	    while (pRefRun->runRight < lastX)
	    {
		pPrevRefRun = pRefRun;
		pRefRun = pRefRun->pNext;
	    }

		/* Now, see if the accumulator has room for at least more
		 * two bytes of data.  If so, read them.  If not,
		 * read one if possible and quit.  This is also
		 * where we restart if the buffer was empty last time
		 * we got here.
		 */
	    if (vacantBit <= 16)
	    {
		if (pCurCompData < pCompDataEnd)
		{
		    accum |= reverseByte[pCurCompData[0]] << vacantBit;
		    vacantBit += 8;
		    accum |= reverseByte[pCurCompData[1]] << vacantBit;
		    vacantBit += 8;
		    pCurCompData += 2;
		}
		else
		{
			/* We're not allowed to fetch 2 new bytes.
			 * If there's one byte, get it.
			 */
		    if (pCurCompData == pCompDataEnd)
		    {
			/* There's one more byte in the data stream. Get it. */
			accum |= reverseByte[pCurCompData[0]] << vacantBit;
			vacantBit += 8;
			pCurCompData++;
		    }
		}
	    }

		/* Now decode the first thing in the accumulator.  This
		 * should be a top level code, (PASS, one of the vertical
		 * codes or a horizontal code marker), so use the
		 * top level code table.
		 */
	    codeVal = mmrRevTopLevelDecodes[accum & cTopLevelCodeMask];

		/* Since we know that all the vertical offset constants
		 * are in the range 0..cVertP3Offset, just test to see
		 * if the value is less than or equal to cVertP3Offset.
		 */
	    if (codeVal <= cVertP3Offset)
	    {
		    /* first, get rid of the code */
		accum >>= mmrRevVertCodeLengths[codeVal];
		vacantBit -= mmrRevVertCodeLengths[codeVal];

		    /* Set the right endpoint of the current run.
		     * Also set current X position to be just past
		     * the endpoint we just found.  We were on a black
		     * pixel and we want to go to the next white
		     * pixel.*/
		pCurRun->runRight = pRefRun->runRight + (Int16) codeVal -
				    cVert0Offset;
		lastX = pCurRun->runRight + 1;

		    /* Advance pCurRun so it points to the next run
		     * in the coding line since we want to look
		     * for the next left endpoint now.
		     */
		if (pCurRun->pNext == NULL)
		{
		    ALLOCATE_STRUCT(pRunHeap, TCScanLineRun,
				    pCurRun->pNext);
		    if (__error__ != ia_successful)
			return ia_nomem;
		    pCurRun = pCurRun->pNext;
		    pCurRun->pNext = NULL;
		    pCurRun->pClient = NULL;
		}
		else
		    pCurRun = pCurRun->pNext;

		    /* THIS IS HOW WE GET TO THE LEFT ENDPOINT LOOP. */
		break;
	    }
	    else
	    if (codeVal == cPassCode)
	    {
		    /* We found a PASS code.  This happens if we're
		     * supposed to skip a little run of white on the
		     * reference line that overlaps a long black run
		     * on the coding line.  This is indicated by the
		     * left endpoint of the run to the right of the
		     * current reference run being at the same coord
		     * or to the left of the current run's right
		     * endpoint.  By limiting our move to no more
		     * than the current run's right endpoint,
		     * we guarantee that we stay in black pixels.  That's
		     * how you tell if issuing a PASS code is ok.
		     * In this case, we're just advancing lastX and
		     * emptying the code out of the accumulator.  We
		     * advance it to the left endpoint of the next
		     * reference run.
		     */
		accum >>= cMMRPassLen;
		vacantBit -= cMMRPassLen;
		lastX = pRefRun->pNext->runLeft;
	    }
	    else
	    if (codeVal == cHorizCode)
	    {
		    /* In this case, there's a horizontal code, which
		     * we should delete, followed by an encoded run of
		     * black pixels, followed by an encoded run of
		     * white pixels.
		     */
		accum >>= cMMRHorizLen;
		vacantBit -= cMMRHorizLen;

		    /* ************* RIGHT BLACK RUN LOOP ************** */
		    /* Each time through this loop, we decode one more
		     * token.  At the top of the loop, we see if there's
		     * room for more stuff.
		     */
		while (cTrue)
		{
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
				reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |=
				reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data stream.
				 * Get it. */
				accum |=
				    reverseByte[pCurCompData[0]] << vacantBit;
				vacantBit += 8;
				pCurCompData++;
			    }
			}
		    }
	
			/* In the black world, terminating codes can be short
			 * or long.  First, see if we have a short terminating
			 * code.  If that doesn't work, try a long terminating
			 * code.  If that doesn't work, we'll try a
			 * non-terminating code.
			 */
		    if ((accum & cBlackLSBitsMask) != 0)	/* it's short */
		    {
			    /* get code length and index using the short code
			     * table */
			codeIndex = (Int8) accum & cShortBlackTermDecodeMask;
			    /* get code value */
			codeVal = mmrRevBlackShortTermDecodes[codeIndex];
			    /* get code length */
			codeIndex = (Int8) mmrRevBlackShortTermDecodes[
					cShortTermDecodeTableLengthOffset +
					codeIndex];
		    }
		    else
		    {
			    /* delete the LS four bits since they're all 0,
			     * and use the long terminating code table. */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & cBlackTermDecodeMask;
			codeVal = mmrRevBlackLongTermDecodes[codeIndex];
			codeIndex = (Int8) mmrRevBlackLongTermDecodes[
						cTermDecodeTableLengthOffset +
						codeIndex];
		    }
	
		    if (codeVal <= 63)
		    {
			/* ********** Terminating black run code *********** */
			    /* set lastX to point to the first white pixel and
			     * runRight to point to the last black pixel.
			     */
			lastX += codeVal;
			pCurRun->runRight = (Int16) lastX - 1;
	
			accum >>= codeIndex;
			vacantBit -= codeIndex;
	
			    /* Having skipped past a black run, point to the
			     * next one in the coding line.
			     */
			if (pCurRun->pNext == NULL)
			{
			    ALLOCATE_STRUCT(pRunHeap, TCScanLineRun,
					    pCurRun->pNext);
			    if (__error__ != ia_successful)
				return ia_nomem;
			    pCurRun = pCurRun->pNext;
			    pCurRun->pNext = NULL;
			    pCurRun->pClient = NULL;
			}
			else
			    pCurRun = pCurRun->pNext;

			    /* HERE IS THE EXIT FROM THE BLACK RUN LOOP */
			break;
		    }

		    else	/* look up this token using the
				 * non-terminating tables.  At this point, we've
				 * already deleted the LS four bits, so we have
				 * the right nibble for this table
				 */
		    {
			pNonTermTable =
			    mmrRevBlack1stNibbleNonTermDecodes[accum & 0xF];
			if (pNonTermTable == NULL)
			    return ia_decodeFailure;
		
			    /* We got a valid table pointer using the
			     * bottom 4 bits.  Advance past them and
			     * decode the next 4 bits.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & 0xF;
	
			    /* and look up the code length */
			codeVal = pNonTermTable[codeIndex].codeLength;
			if (codeVal == 0)
			    return ia_decodeFailure;
	    
			tokenVal = pNonTermTable[codeIndex].codeBits;
			if (codeVal == 5)	/* add 64 x value of 5th bit */
			    tokenVal += (accum * 4) & 64;
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* now move to the left in the current scan line by
			     * the number of pixels just decoded in tokenVal.
			     * We keep doing this until we find a terminating
			     * code.
			     */
			lastX += tokenVal;
	
		    }	/* end of non-terminating black code case */
		}	/* end of loop that decodes black values */

		    /* Here's the loop that decodes some number of tokens
		     * used to represent a white run length.  At the top
		     * of the loop is a clause that tries to put two more
		     * bytes into the accumulator.  After that, we
		     * decode the code.
		     */
		    /* ************** RIGHT WHITE RUN LOOP ******************/
		while (cTrue)
		{
		    if (vacantBit <= 16)
		    {
			if (pCurCompData < pCompDataEnd)
			{
			    accum |=
			        reverseByte[pCurCompData[0]] << vacantBit;
			    vacantBit += 8;
			    accum |=
			        reverseByte[pCurCompData[1]] << vacantBit;
			    vacantBit += 8;
			    pCurCompData += 2;
			}
			else
			{
				/* We're not allowed to fetch 2 new bytes.
				 * If there's one byte, get it.
				 */
			    if (pCurCompData == pCompDataEnd)
			    {
				/* There's one more byte in the data stream.
				 * Get it. */
				accum |=
				    reverseByte[pCurCompData[0]] << vacantBit;
				vacantBit += 8;
				pCurCompData ++;
			    }
			}
		    }

			/* When decoding white runs, we always use 8 bits to
			 * index a table.
			 * First, see if this is a terminating code yet.
			 */
		    codeIndex = (UInt8) accum & cWhiteTermDecodeMask;
		    codeVal = mmrRevWhiteTermDecodes[codeIndex];
		    if (codeVal <= 63)
		    {
			/* *********** Terminating white run code ********** */
			    /* set next run coord */
			lastX += codeVal;
			pCurRun->runLeft = (Int16) lastX;
			pCurRun->pClient = NULL;
	
			    /* get length of code in bits */
			codeVal =
			    mmrRevWhiteTermDecodes[
				cTermDecodeTableLengthOffset + codeIndex];
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;

			    /* Continue to leave pCurRun pointing to
			     * the current run since we've got to go
			     * find it's right endpoint.  We just loop
			     * back around in right endpoint loop after
			     * falling out of this loop and the switch
			     * statement.
			     */
	
			    /* HERE IS THE EXIT FROM THE WHITE RUN LOOP */
			break;
		    }
		    else	/* look up this token using the
				 * non-terminating tables */
		    {
			/* ********* Non Terminating white run code ********* */
			pNonTermTable =
			    mmrRevWhite1stNibbleNonTermDecodes[codeIndex & 0xF];
			if (pNonTermTable == NULL)
			    return ia_decodeFailure;
		
			    /* We got a valid table pointer using the bottom
			     * 4 bits.
			     * Advance past them and decode the next 4 bits.
			     */
			accum >>= 4;
			vacantBit -= 4;
			codeIndex = (Int8) accum & 0xF;
			if (pNonTermTable[codeIndex].codeLength > 5)
			{
				/* The code in accum is one of the common
				 * codes used to skip really big blocks of
				 * pixels.  To decode these, we look it up
				 * in the black non-terminating tables.
				 */
			    pNonTermTable =
				mmrRevBlack1stNibbleNonTermDecodes[codeIndex];
			    if (pNonTermTable == NULL)
				return ia_decodeFailure;
		
				/* Now we really have a pointer to the
				 * lower level non-terminating code table,
				 * so toss the nibble we just used.
				 */
			    accum >>= 4;
			    vacantBit -= 4;
			    codeIndex = (Int8) accum & 0xF;
			}
		
			    /* and look up the code length */
			codeVal = pNonTermTable[codeIndex].codeLength;
			if (codeVal == 0)
			    return ia_decodeFailure;
	    
			tokenVal = pNonTermTable[codeIndex].codeBits;
			if (codeVal == 5)	/* add 64 x value of 5th bit */
			{
			    tokenVal += (accum * 4) & 64;
			    if (tokenVal == cWeirdWhiteCode)
				tokenVal += 64;
			} 
			accum >>= codeVal;
			vacantBit -= (Int8) codeVal;
	
			    /* now move to the left in the current scan line by
			     * the number of pixels just decoded in tokenVal.
			     * We keep doing this until we find a terminating
			     * code.
			     */
			lastX += tokenVal;
		    }	/* end of non-terminating white case */
		}	/* end of loop that decodes white values */
	    }	/* end of horizontal code for right endpoints */
	    else
		    /* this was an unknown code */
		return ia_decodeFailure;
	}		/* end of loop that decodes right endpoints */
    }	/* End of main loop that decodes whole line */

	/* We get here when everything was just swell.  We've generated
	 * the codes for an entire line.
	 * We never add padding in MMR, so no need to do that.  We also
	 * don't flush the accumulator at the end of a scan line, so
	 * don't do that either.
	 * Just save the state so we know we're done with this scan
	 * line and can concatenate the codes for the next scan line
	 * when it comes.
	 * BUG FIX: Very rarely, the encoder had to encode a pair of
	 * horizontal runs where the second run was black AND that
	 * second run was null because the white run actually touched
	 * the right edge of the image.  If that happens, don't return
	 * that last run, it will just confuse everyone.  This is
	 * a good place to put the check since it doesn't happen very
	 * often.
	 */
decompMMRGoodCompletion:
    *ppCompData = pCurCompData;
    if ((pPrevRun != NULL) &&
	(pPrevRun->pNext == pCurRun) &&
	(pPrevRun->runLeft > pPrevRun->runRight))
       pCurRun = pPrevRun;
    *ppDummy = pCurRun;
    pCCITTDecompInfo->currentRun = 0;
    pCCITTDecompInfo->refRun = 0;
    pCCITTDecompInfo->remainingLength = 0;   /* used for lastX initialization */
    pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
    pCCITTDecompInfo->accumBits = accum;
    pCCITTDecompInfo->vacantBit = vacantBit;
    
    NORMAL_RETURN;	/* It really worked */
}

/*

 $Log: mimmrdl.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:52:51  ddavies
 mimmrdl.c -> /product/ipcore/ipshared/src/RCS/mimmrdl.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1996/05/31 20:36:20  ddavies
 Initial revision


 */
