/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	minirt1.c: Machine independent implementation of lower
 *		   level routines for non-interpolated rotation
 *
 *	    C IMPLEMENTATION of
 *          Rotation routines that do NOT use interpolation to
 *		form output pixels, SMALL ANGLES ONLY! (< 10 degrees)
 *
 *          Low-level ops:
 *		Int32		ip_rotateNoInterpLine();
 *		Int32		ip_rotateAssemblyCodeIsPresent();
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

    /*  prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: minirt1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* Please see the EXTENSIVE comments in i_nirfx1.c.  These cover all levels
 * of non-interpolated rotation.
 */


/* PRIVATE */
/**************************************************************************
 *  ip_rotateAssemblyCodeIsPresent()
 *		If rotateNoInterpLine is implemented in C, returns cFalse.
 *		If rotateNoInterpLine is implemented in assembly, returns cTrue
 *************************************************************************/
Int32  CDECL 
ip_rotateAssemblyCodeIsPresent (
    IPState    *pIPState)
{
    PROLOGUE;
    pIPState = pIPState;	/* keep compiler happy */
    __error__ = __error__;

    return cFalse;
}


/* PRIVATE */
/**************************************************************************
 *  ip_rotateNoInterpLine()
 *		Generates one rotated line without using interpolation.
 *
 *	Input:
 *		pS =		pointer to 32 bit word containing first pixel
 *				of unrotated data.
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination line
 *		w =		width of source image in bits
 *		xCount = 	increment this count at each move towards the
 *				right.  While xCount < 0, we are off the image.
 *				Distance to horizontal edge.
 *		yCount =	increment this count at each move towards the
 *				bottom.  While yCount < 0, we are off the image.
 *				Distance to vertical edge.
 *		leftShift =	position of leftmost image bit within
 *				first image word as measured from left
 *				edge of first image word (0 => input and
 *				output words are aligned).  This bit
 *				should be shifted to become bit 0 (leftmost) 
 *				of the first output word, hence it is the
 *				distance by which input words should be
 *				shifted left to be aligned with the output
 *				words.
 *		pMaskTable =	Pointer to sequence of masks used to insert
 *				runs of pixels from input words into output
 *				words.
 *		whiteVal =	A 32 bit word full of white pixels.  This
 *				is 0x00000000 in binary images and
 *				0xFFFFFFFF in gray images.
 *		rightWhiteVal = A 32 bit word used to OR white pixels into
 *				the rightmost partial word of an image.
 *		boundary1 =	Check xCount and yCount when reading input
 *				words for all output words
 *				processed to the left of boundary1.
 *		boundary2 =	Check xCount when reading input
 *				words for all output words
 *				processed to the left of boundary2.
 *		hS =		height of source image in pixels.  This is
 *				down here because it was the solution to
 *				a bug that was found long after everything
 *				else was written.
 *	Output:
 *		Stores 1 rotated line into output image.
 *
 **************************************************************************/
void  CDECL 
ip_rotateNoInterpLine (
    IPState    *pIPState,
    UInt8Ptr	pS,
    Int32	sBpl,
    UInt8Ptr	pD,
    Int32	w,
    Int32	xCount,
    Int32	yCount,
    Int32	leftShift,
    register UInt32Ptr pMaskTable,
    UInt32	whiteVal,
    UInt32	rightWhiteVal,
    Int32	boundary1,
    Int32	boundary2,
    Int32	hS)
{
    PROLOGUE;
    UInt32	leftmostInputBit;	/* a mask containing a single
					 * 1 in the position where
					 * the MS bit of the input
					 * words aligns with the
					 * output words. */
    UInt32	rightmostInputBit;	/* a mask containing a single
					 * 1 in the position where
					 * the LS bit of the input
					 * words aligns with the
					 * output words. */
    UInt32	oWord;			/* the output word under
					 * formation */
    UInt32	iWord;			/* word read from input */
    UInt32	mask;			/* the current mask */
    Int32	rightShift;		/* distance by which input words
					 * should be shifted right.  As
					 * discussed above, this is constant
					 * across the entire line. */
    UInt32Ptr	pSWord;			/* pointer to source words */
    UInt32Ptr	pDWord;			/* pointer to dest words */
    Int32	sWpl;			/* source words/line */


    pIPState = pIPState;		/* keep compiler happy */
    __error__ = __error__;

	/* we'll be reading the source and writing the destination
	 * is units of words, so use word pointers */
    pSWord = (UInt32Ptr) pS;
    pDWord = (UInt32Ptr) pD;
    sWpl = sBpl/4;			/* preserve the sign */
    rightShift = 32 - leftShift;

	/* Do the input and output line up perfectly?  If so, no shifting
	 * is required */
    if (leftShift == 0)
    {
	    /* Loop through the words of this row.  Start
	     * with the first section of code.  This one
	     * checks both xCount and yCount on each fetch of
	     * input word.
	     * ************* SECTION 1, NO SHIFTING ***********
	     */
	for (; w > boundary1; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = whiteVal;
	    if ((xCount | yCount) >= 0)
	    {
		    /* If we're near the right edge of the image,
		     * fill last partial word with white.
		     */
		if (xCount+32 < 0)
		    iWord = *pSWord | rightWhiteVal;
		else
		    iWord = *pSWord;
	    }
	    oWord = iWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */
	        /* if yCount is at the bottom of the image, set it equal
		 * to hS - 1 (the last line of the image.  Otherwise, just
		 * decrement it.  This allows us to process diagonal paths
		 * that go right through the image from below to above.
		 */
	    if (yCount != (Int32)0x80000000)
		yCount--;
	    else
		yCount = hS - 1;

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		iWord = whiteVal;
		if ((xCount | yCount) >= 0)
		{
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if (xCount+32 < 0)
			iWord = *pSWord | rightWhiteVal;
		    else
			iWord = *pSWord;
		}
		oWord |= iWord & mask;	/* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    if (mask == (UInt32)sWpl)
	    {
		if (yCount != hS - 1)
		    yCount++;		/* put yCount back where it belongs */
		else
		    yCount = 0x80000000;
	    }
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks xCount and yCount. */
	if (w <= 0)
	    return;

	    /* Second section of no-shifting code.  This one checks neither
	     * xCount nor yCount, but does update xCount so it's ready for
	     * the third section.
	     * ************* SECTION 2, NO SHIFTING ***********
	     */
	for (; w > boundary2; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    oWord = *pSWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		oWord |= *pSWord & mask; /* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks NEITHER xCount 
		 * nor yCount. */

	    /* Third section of no-shifting code.  This one checks only
	     * xCount.
	     * ************* SECTION 3, NO SHIFTING ***********
	     */
	for (; w > 0; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = whiteVal;
		/* If the entire word is inside the image (rare),
		 * use it.  If only part of the rightmost word is
		 * inside the image, OR in white pixels in the
		 * frame section.  We can do this because the
		 * frame is always 0's.
		 */
	    if (xCount >= 0)
	    {
		if (xCount+32 >= 0)
		    iWord = *pSWord;
		else
		    iWord = *pSWord | rightWhiteVal;
	    }

	    oWord = iWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		iWord = whiteVal;	/* assume the common case */
		    /* Insert white pixels as necessary as explained above */
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
			iWord = *pSWord;
		    else
			iWord = *pSWord | rightWhiteVal;
		}
		oWord |= iWord & mask;	/* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks only xCount. */
	return;

    }	/* end of case where no shifting is required */

    else

    {	/* each input word must be shifted left or right before inserting
	 * its bits */

	    /* rightmostInputBit will have a 1 in the output bit position
	     * corresponding to the least significant bit of the input
	     * words that are shifted left by leftShift.  leftmostInputBit
	     * will have a 1 in the output bit position corresponding
	     * to the most significant bit of input words that are
	     * shifted right by 32-leftShift.
	     */
	rightmostInputBit = 0x80000000;		/* just to put the const
						 * into a register */
	leftmostInputBit = rightmostInputBit >> (32-leftShift);
	rightmostInputBit = leftmostInputBit << 1;

	    /* Loop through the words of this row.  Start
	     * with the first section of code.  This one
	     * checks both xCount and yCount on each fetch of
	     * input word.
	     * ************* SECTION 1, SHIFTING REQUIRED ***********
	     */
	for (; w > boundary1; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = whiteVal;
	    if ((xCount | yCount) >= 0)
	    {
		    /* If we're near the right edge of the image,
		     * fill last partial word with white.
		     */
		if (xCount+32 < 0)
		    iWord = *pSWord | rightWhiteVal;
		else
		    iWord = *pSWord;
	    }
	    oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
	        /* if yCount is at the bottom of the image, set it equal
		 * to hS - 1 (the last line of the image.  Otherwise, just
		 * decrement it.  This allows us to process diagonal paths
		 * that go right through the image from below to above.
		 */
	    if (yCount != (Int32)0x80000000)
		yCount--;
	    else
		yCount = hS - 1;

		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = whiteVal;
		if ((xCount | yCount) >= 0)
		{
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if (xCount+32 < 0)
			iWord = *pSWord | rightWhiteVal;
		    else
			iWord = *pSWord;
		}
		oWord |= (iWord << leftShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word.
		 * Note the funny test for yCount.  We really want to
		 * look at yCount + 1, the line below pSWord.
		 * If yCount+1 is negative, (UInt32)(yCount+1) is bigger
		 * than any possible value for hS.  If we've just set
		 * yCount to correspond to the last image line (hS-1),
		 * then yCount + 1 == hS so the test will fail, as it
		 * should (the line below pSWord is off the image).
		 * If the line below yCount is inside the image,
		 * (UInt32)(yCount+1) will be less than hS.
		 */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		iWord = whiteVal;
		if ((xCount >= 0) && ((UInt32)(yCount + 1) < (UInt32)hS))
		{
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if (xCount+32 < 0)
			iWord = pSWord[sWpl] | rightWhiteVal;
		    else
			iWord = pSWord[sWpl];
		}
		oWord |= (iWord >> rightShift) & mask; /* insert next run */
	    }

	        /* Insert the rest of the runs in this word (if any) into
		 * this output word. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = whiteVal;
		if ((xCount | yCount) >= 0)
		{
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if (xCount+32 < 0)
			iWord = *pSWord | rightWhiteVal;
		    else
			iWord = *pSWord;
		}
		oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    if (mask == (UInt32)sWpl)
	    {
		if (yCount != hS - 1)
		    yCount++;		/* put yCount back where it belongs */
		else
		    yCount = 0x80000000;
	    }
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks both xCount and
		 * yCount when fetching input words. */
	if (w <= 0)
	    return;

	    /* Second section of shifting code.  This one checks neither
	     * xCount nor yCount, but does update xCount so it's ready for
	     * the third section.
	     * WE SHOULD BE SPENDING MOST OF OUR TIME HERE
	     * ************* SECTION 2, SHIFTING REQUIRED ***********
	     */
	for (; w > boundary2; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    oWord = (*pSWord << leftShift) & mask; /* insert 1st run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		oWord |= (*pSWord << leftShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word. */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		oWord |= (pSWord[sWpl] >> rightShift) & mask; /* finish run */
	    }

	        /* Insert the rest of the runs in this word (if any) into
		 * this output word. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		oWord |= (*pSWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks neither xCount nor
		 * yCount when fetching input words. */

	    /* Third section of shifting code.  This one checks only
	     * xCount.
	     * ************* SECTION 3, SHIFTING REQUIRED ***********
	     */
	for (; w > 0; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = whiteVal;
		/* If the entire word is inside the image (rare),
		 * use it.  If only part of the rightmost word is
		 * inside the image, OR in white pixels in the
		 * frame section.  We can do this because the
		 * frame is always 0's.
		 */
	    if (xCount >= 0)
	    {
		if (xCount+32 >= 0)
		    iWord = *pSWord;
		else
		    iWord = *pSWord | rightWhiteVal;
	    }

	    oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = whiteVal;
		    /* Insert white pixels as necessary as explained above.
		     */
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
			iWord = *pSWord;
		    else
			iWord = *pSWord | rightWhiteVal;
		}
		oWord |= (iWord << leftShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word. */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		iWord = whiteVal;
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
			iWord = pSWord[sWpl];
		    else
			iWord = pSWord[sWpl] | rightWhiteVal;
		}
		oWord |= (iWord >> rightShift) & mask; /* insert next run */
	    }

	        /* Insert the rest of the runs in this word (if any) into
		 * this output word. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = whiteVal;
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
			iWord = pSWord[0];
		    else
			iWord = pSWord[0l] | rightWhiteVal;
		}
		oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks only xCount
		 * when fetching input words. */
	return;
    }	/* end of case where shifting IS required */
}

/*

$Log: minirt1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.5  1998/01/13 23:52:53  ddavies
minirt1.c -> /product/ipcore/ipshared/src/RCS/minirt1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1996/10/05 02:25:14  ddavies
minirt1.c -> /product/ipcore/ipshared/src/RCS/minirt1.c,v

Make sort of ugly changes to the semantics of xCount and yCount.  yCount now
is changed when we cross the lower boundary of the image so it's negative
whenever it's outside the image.  This accounts for the fact that it might
cross both the upper and lower boundaries.

Revision 1.3  1995/12/13 02:50:26  ejaquith
minirt1.c -> /product/ipcore/ipshared/src/RCS/minirt1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/09/21  02:59:20  ddavies
 * minirt1.c -> /product/ipcore/ipshared/src/RCS/minirt1.c,v
 *
 * Fixed a bug in ip_rotateNoInterpLine and added the rightWhiteVal arg.
 * The bug appeared in images whose rightmost pixel did not appear on a
 * word boundary.  In this case, we sometimes read the entire word (including
 * frame) containing the right pixel and put it into the output image (causeing
 * frame to appear in the midst of an image).  We also sometimes ignored the
 * rightmost word entirely, causing any pixels in that word to be lost.
 * The calling routines now calculate xCount correctly, fixing the bug that
 * caused us to ignore the rightmost word.  The rightWhiteVal arg contains
 * the value of white in the pixels that would otherwise contain right frame
 * in the rightmost word of each input image line.  It is ORed with the
 * rightmost word whenever it's read, inserting white in place of the frame.
 *
 * Revision 1.1  1995/01/12  03:10:25  ddavies
 * Initial revision
 *
 * Revision 1.4  1994/04/03  02:05:56  ddavies
 * minirot.c -> /product/ipcore/ipshared/src/RCS/minirot.c,v
 *
 * Incorporate pointer types for use in SantaFe project.
 *
 * Revision 1.3  1993/12/02  19:24:58  ddavies
 * minirot.c -> /product/ipcore/ipshared/src/RCS/minirot.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:49:52  ddavies
 * minirot.c -> /product/ipcore/ipshared/src/RCS/minirot.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


