/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	minirt2.c: Machine independent implementation of lower
 *		   level routines for non-interpolated rotation
 *
 *	    C IMPLEMENTATION of
 *          Rotation routines that do NOT use interpolation to
 *		form output pixels, SMALL ANGLES ONLY! (< 10 degrees)
 *
 *          Low-level ops:
 *		Int32		ip_rotateNoInterpLineNiceBinary() C (for now)
 *		Int32		ip_rotateNoInterpLineNiceGray()	C (for now)
 * 
 *
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"

    /*  prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: minirt2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* Please see the EXTENSIVE comments in i_nirfx1.c.  These cover all levels
 * of non-interpolated rotation.
 */

/* This routine is very similar to a binary version ip_rotateNoInterpLine
 * except it tries to avoid shifting input words by amounts that would 
 * cause them to become unaligned with previously produced outputs.
 * It does this by ANDing each input word with the one above it to find 
 * vertical strokes.  This union is shifted by the correct amount and ANDed
 * with the white portion of the destination word above the current dest
 * word.  If the result is zero, things are still aligned so we proceed.
 * if the result is non-zero, things are not aligned.  We blindly shift
 * the input right by 1 bit before inserting it's run.  Why right by 1?
 * As rotateNoInterpolation goes up or down the image, it either holds
 * the left shift distance constant or increases it by 1.  We assume
 * that lack of vertical alignment results from an increase by 1 relative
 * to the line above this one, so we shift back right.
 *
 * ***********************
 *
 * That was promising, but not quite right.  We occasionally got an
 * extraneous right shift.  We now shift right if both a) using the normal
 * shift moved the vertical strokes into white areas, and b) shifting
 * back right one place did not move the vertical strokes in white
 * places.
 *
 ************************
 *
 * That was definitely better on text.  But, it turned out that we got
 * the same results (faster) if we just applied the current run's mask
 * to the shifted vertical strokes when deciding if the vertical strokes
 * strayed into white areas.  This will prevent us from being fooled by
 * areas of the input word that won't appear in the output.  We should
 * have been doing that all along, but had neglected to do so.
 *
 * Unfortunately, in fairly dark dithered areas, this algorithm produced
 * vertical streaks.  In areas where the dithering produced thin
 * vertical stripes, the rotation algorithm would try to keep the stripes
 * aligned.  This is fine except that the vertical stripes might occur in
 * one vertical block of dest words, but not in the block next to it.
 * The algorithm would use differing shift amounts for the two words.
 * Since the two words were no longer contiguous, the algorithm would
 * either replicate a stripe or leave one out to make up the difference.
 * The duplicated or deleted bits would be seen as a streak.
 *
 * Our first attempt to fix this tries to disable alignment in dithered
 * areas.  We've noticed that a) there aren't very many black places on a
 * normal page, and b) there are even fewer places where alignment is needed.
 * In the interests of performance, we should put the extra code inside
 * the clause that detects that alignment is needed - not increase the length
 * of the "if" that detects the need for alignment.  That minimizes the
 * number of times the extra code is executed.
 *
 * To carry this one step further, we break the alignment detection clause
 * into two sections.  The first asks both whether there are any vertical
 * strokes at all (vertBits != 0) and whether shifting vertBits by the 
 * expected amount (and masking it with the current mask) causes the strokes
 * to end up in white areas of the destination.  Avoiding the fancy alignment
 * code when vertBits == 0 gives us a quick out in all white areas and
 * in most parts of dithered images.
 *
 * If the first clause is true, we do a morpholgical open operation on
 * vertBits.  This eliminates vertical strokes that are 1 bit
 * wide while preserving wider strokes (like the edges of characters!).
 * If shifting the opened version of vertBits by the normal amount
 * would put vertical input strokes into white dest areas, we adjust the
 * shift.  If not, we use the normal shift.
 *
 * NOTE: we may need to change this heuristic for other dithering algorithms.
 * It seems to work well for 1 bit/pixel images.
 *
 ****************************************
 *
 * Next BUG FIX.  rightShift will equal 31 at some point.  We will
 * then try to shift some word right by 31+1.  Since the shift amounts
 * are taken MOD 32, this is 0, not 32.  So, before trying to shift by
 * rightShift+1, we check rightShift.  If it's not equal to 31, go ahead.
 * if it's equal to 31, we ignore the shifted term, since it's supposed to
 * turn into 0.
 *
 ****************************************
 *
 * REVISION: We split ip_rotateNoInterpLineNice into two routines,
 * ip_rotateNoInterpLineNiceBinary and ip_rotateNoInterpLineNiceGray.  The
 * Binary version doesn't need to bother with blackVal or depth, so can
 * be a bit faster.  It's also easier for me to figure out what's happening
 * this way.
 *
 ****************************************
 *
 * FEATURE ADDITION for XIS.  They need to save space in the Reading Edge
 * and never use these, so we turn them into noops when XIS_USE is defined.
 * 
 * ***********************************************************************
 * BUG FIX, 9/20/95:  Richard Mlynarik found that rotating gray images
 * sometimes produced artifacts on the right side.  After further
 * investigation, I found that there's a generic problem with the way
 * we handle the rightmost word of an image.  If the right boundary of
 * an input image does not end on a word boundary, we sometimes use the
 * entire word (which doesn't work when the frame is not white) and sometimes
 * ignore the rightmost word entirely.  No good.
 *
 * The problem has the do with confusion about xCount.  In the gray rotate
 * code, we read one pixel at a time and updated xCount each time.  In
 * this code we read 32 bits (some number of pixels) each time.  xCount
 * indicated the distance from some pixel inside the word to the edge
 * of the input image.  This uncertaintly threw us off.
 *
 * In the new scheme, we truncate xCount to the nearest multiple of 32
 * towards -infinity.  This makes it calculate the distance from the leftmost
 * bit in the word being read to the edge of the input image.  We then
 * check the value of xCount in the third section of code that is used when
 * we're approaching the right edge of the image.  If xCount is < 0 (greater
 * than 0x80000000), we know the entire word being read is off the right
 * side of the image and just use whiteval.  This is the common case
 * there.  If xCount is positive, we now have two cases where before we
 * had one.  If xCount + 32 is also postive, we read an entire new word
 * from the source image.  If xCount+32 is negative ( >= 0x80000000 ),
 * we merge the word read from the image with whiteval using a new mask.
 * The mask is a constant computed from the width of the image in bits.
 * We can use the constant since the width of the source image (the one
 * from which we are reading the data) is constant.  In fact, since we
 * know that the stuff off the edge of the source image is frame, therefore
 * filled with 0's, we can just supply a rightWhite mask which is filled
 * with whiteVal bits in the frame section of the input word.  We just
 * OR that with the partial input word read.  The rightWhite mask is
 * computed in the lower level routines since not all of them take a
 * whiteVal anyway.
 *
 */

/* PRIVATE */
/**************************************************************************
 *  ip_rotateNoInterpLineNiceBinary()
 *		Generates one rotated line without using interpolation.
 *		Tries to avoid shearing vertical strokes.
 *
 *	Input:
 *		pS =		pointer to 32 bit word containing first pixel
 *				of unrotated data.
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination line
 *		dBpl =		dest bytes/line
 *		w =		width of source image in bits
 *		xCount = 	increment this count at each move towards the
 *				right.  While xCount < 0, we are off the image.
 *				Distance to horizontal edge.
 *		yCount =	increment this count at each move towards the
 *				bottom.  While yCount < 0, we are off the image.
 *				Distance to vertical edge.
 *		leftShift =	position of leftmost image bit within
 *				first image word as measured from left
 *				edge of first image word (0 => input and
 *				output words are aligned).  This bit
 *				should be shifted to become bit 0 (leftmost) 
 *				of the first output word, hence it is the
 *				distance by which input words should be
 *				shifted left to be aligned with the output
 *				words.
 *		pMaskTable =	Pointer to sequence of masks used to insert
 *				runs of pixels from input words into output
 *				words.
 *		boundary1 =	Check xCount and yCount when reading input
 *				words for all output words
 *				processed to the left of boundary1.
 *		boundary2 =	Check xCount when reading input
 *				words for all output words
 *				processed to the left of boundary2.
 *		hS =		height of source image in pixels.  This is
 *				down here because it was the solution to
 *				a bug that was found long after everything
 *				else was written.
 *	Output:
 *		Stores 1 rotated line into output image.
 *
 **************************************************************************/
void  CDECL 
ip_rotateNoInterpLineNiceBinary (
    IPState    *pIPState,
    UInt8Ptr	pS,
    Int32	sBpl,
    UInt8Ptr	pD,
    Int32	dBpl,
    Int32	w,
    Int32	xCount,
    Int32	yCount,
    Int32	leftShift,
    UInt32Ptr	pMaskTable,
    Int32	boundary1,
    Int32	boundary2,
    Int32	hS)
#ifdef XIS_USE		/* XIS doesn't use this routine, so turn it into
			 * a stub.
			 */
{
    pIPState = pIPState;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dBpl = dBpl;
    w = w;
    xCount = xCount;
    yCount = yCount;
    leftShift = leftShift;
    pMaskTable = pMaskTable;
    boundary1 = boundary1;
    boundary2 = boundary2;
}
#else
{
    UInt32	leftmostInputBit; /* a mask containing a single
				 * 1 in the position where
				 * the MS bit of the input
				 * words aligns with the
				 * output words. */
    UInt32	rightmostInputBit; /* a mask containing a single
				 * 1 in the position where
				 * the LS bit of the input
				 * words aligns with the
				 * output words. */
    UInt32	oWord;		/* the output word under
				 * formation */
    UInt32	iWord;		/* word read from input */
    UInt32	mask;		/* the current mask */
    Int32	rightShift;	/* distance by which input words
				 * should be shifted right.  As
				 * discussed above, this is constant
				 * across the entire line. */
    UInt32Ptr	pSWord;		/* pointer to source words */
    UInt32Ptr	pDWord;		/* pointer to dest words */
    Int32	sWpl;		/* source words/line */
    Int32	dWpl;		/* dest words/line */
    UInt32	vertBits;	/* holds the bits that were
				 * aligned vertically in the input */

	/* Keep the compiler happy */
    pIPState = pIPState;

	/* we'll be reading the source and writing the destination
	 * is units of words, so use word pointers */
    pSWord = (UInt32Ptr) pS;
    pDWord = (UInt32Ptr) pD;
    sWpl = sBpl/4;			/* make sure sign is preserved */
    dWpl = dBpl/4;
    rightShift = 32 - leftShift;

	/* Do the input and output line up perfectly?  If so, no shifting
	 * is required */
    if (leftShift == 0)
    {
	    /* Loop through the words of this row.  Start
	     * with the first section of code.  This one
	     * checks both xCount and yCount on each fetch of
	     * input word.
	     * ************* SECTION 1, NO SHIFTING ***********
	     */
	for (; w > boundary1; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = 0;
	    vertBits = 0;
	    if ((xCount | yCount) >= 0)
	    {
	        iWord = pSWord[0];
		vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
	    }
		/* do the vert input bits line up with vert output? */
	    if (vertBits &&
		(vertBits & ~pDWord[-dWpl] & mask))
	    {
		    /* delete 1 pixel strokes, but leave the wider strokes
		     * unaffected */
		vertBits = vertBits & (vertBits >> 1); 
		vertBits = vertBits | (vertBits << 1);
		if (vertBits & ~pDWord[-dWpl] & mask)
		    iWord = (iWord >> 1) | (pSWord[-1] << 31);
	    }
	    oWord = iWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */
	        /* if yCount is at the bottom of the image, set it equal
		 * to hS - 1 (the last line of the image.  Otherwise, just
		 * decrement it.  This allows us to process diagonal paths
		 * that go right through the image from below to above.
		 */
	    if (yCount != (Int32)0x80000000)
		yCount--;
	    else
		yCount = hS - 1;

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		iWord = 0;
		vertBits = 0;
		if ((xCount | yCount) >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		}
		    /* do the vert input bits line up with vert output? */
		if (vertBits &&
		    (vertBits & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1); 
		    vertBits = vertBits | (vertBits << 1);
		    if (vertBits & ~pDWord[-dWpl] & mask)
			iWord = (iWord >> 1) | (pSWord[-1] << 31);
		}
		oWord |= iWord & mask;	/* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    if (mask == (UInt32)sWpl)
	    {
		if (yCount != hS - 1)
		    yCount++;		/* put yCount back where it belongs */
		else
		    yCount = 0x80000000;
	    }
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks xCount and yCount. */
	if (w <= 0)
	    return;

	    /* Second section of no-shifting code.  This one checks neither
	     * xCount nor yCount, but does update xCount so it's ready for
	     * the third section.
	     * ************* SECTION 2, NO SHIFTING ***********
	     */
	for (; w > boundary2; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = pSWord[0];
	    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
	    if (vertBits &&
		(vertBits & ~pDWord[-dWpl] & mask))
	    {
		vertBits = vertBits & (vertBits >> 1); 
		vertBits = vertBits | (vertBits << 1);
		if (vertBits & ~pDWord[-dWpl] & mask)
		    iWord = (iWord >> 1) | (pSWord[-1] << 31);
	    }
	    oWord = iWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		iWord = pSWord[0];
		vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		if (vertBits &&
		    (vertBits & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1); 
		    vertBits = vertBits | (vertBits << 1);
		    if (vertBits & ~pDWord[-dWpl] & mask)
			iWord = (iWord >> 1) | (pSWord[-1] << 31);
		}
		oWord |= iWord & mask; /* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks NEITHER xCount 
		 * nor yCount. */

	    /* Third section of no-shifting code.  This one checks only
	     * xCount.
	     * ************* SECTION 3, NO SHIFTING ***********
	     */
	for (; w > 0; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = 0;
	    vertBits = 0;
	    if (xCount >= 0)
	    {
	        iWord = pSWord[0];
		vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
	    }
		/* do the vert input bits line up with vert output? */
	    if (vertBits &&
		(vertBits & ~pDWord[-dWpl] & mask))
	    {
		vertBits = vertBits & (vertBits >> 1); 
		vertBits = vertBits | (vertBits << 1);
		if (vertBits & ~pDWord[-dWpl] & mask)
		    iWord = (iWord >> 1) | (pSWord[-1] << 31);
	    }
	    oWord = iWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		iWord = 0;
		vertBits = 0;
		if (xCount >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		}
		if (vertBits &&
		    (vertBits & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1); 
		    vertBits = vertBits | (vertBits << 1);
		    if (vertBits & ~pDWord[-dWpl] & mask)
			iWord = (iWord >> 1) | (pSWord[-1] << 31);
		}
		oWord |= iWord & mask;	/* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks only xCount. */
	return;

    }	/* end of case where no shifting is required */

    else

    {	/* each input word must be shifted left or right before inserting
	 * its bits */

	    /* rightmostInputBit will have a 1 in the output bit position
	     * corresponding to the least significant bit of the input
	     * words that are shifted left by leftShift.  leftmostInputBit
	     * will have a 1 in the output bit position corresponding
	     * to the most significant bit of input words that are
	     * shifted right by 32-leftShift.
	     */
	rightmostInputBit = 0x80000000;		/* just to put the const
						 * into a register */
	leftmostInputBit = rightmostInputBit >> (32-leftShift);
	rightmostInputBit = leftmostInputBit << 1;

	    /* Loop through the words of this row.  Start
	     * with the first section of code.  This one
	     * checks both xCount and yCount on each fetch of
	     * input word.
	     * ************* SECTION 1, SHIFTING REQUIRED ***********
	     */
	for (; w > boundary1; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = 0;
	    vertBits = 0;
	    if ((xCount | yCount) >= 0)
	    {
		iWord = pSWord[0];
		vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
	    }
	    if (vertBits &&
		((vertBits<<leftShift) & ~pDWord[-dWpl] & mask))
	    {
		vertBits = vertBits & (vertBits >> 1); 
		vertBits = vertBits | (vertBits << 1);
		if ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask)
		    oWord = (iWord << (leftShift-1)) & mask;
		else
		    oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    }
	    else
		oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
	        /* if yCount is at the bottom of the image, set it equal
		 * to hS - 1 (the last line of the image.  Otherwise, just
		 * decrement it.  This allows us to process diagonal paths
		 * that go right through the image from below to above.
		 */
	    if (yCount != (Int32)0x80000000)
		yCount--;
	    else
		yCount = hS - 1;

		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0;
		vertBits = 0;
		if ((xCount | yCount) >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		}
		if (vertBits &&
		    ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1); 
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask)
			oWord |= (iWord << (leftShift-1)) & mask;
		    else
			oWord |= (iWord << leftShift) & mask;
		}
		else
		    oWord |= (iWord << leftShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word.
		 * BUG FIX:  In the first case here, we had been ORing the
		 * the shifted versions of iWord and pSWord[sWpl-1] together,
		 * then applying "mask".  This doesn't work in this case
		 * because the mask still covers bits that have been
		 * set above.  Shifting pSWord[sWpl-1] *left* causes it's
		 * bits to intersect the bits we've already set in oWord.
		 * To fix it, we mask pSWord[sWpl-1] with 0x1 before
		 * the left shift.  This is ok, because we only need the
		 * LS bit of that word to account for the extra right shift.
		 * Note that we check rightShift before using rightShift+1.
		 * If rightShift == 31, we just ignore the shifted term
		 * because the SPARC (and 486) only use the LS 5 bits of the
		 * shift distance so a shift by 31+1 bits is no shift at
		 * all instead of a shift by 32 - which should have filled
		 * the word with 0's.
		 * Note the funny test for yCount.  We really want to
		 * look at yCount + 1, the line below pSWord.
		 * If yCount+1 is negative, (UInt32)(yCount+1) is bigger
		 * than any possible value for hS.  If we've just set
		 * yCount to correspond to the last image line (hS-1),
		 * then yCount + 1 == hS so the test will fail, as it
		 * should (the line below pSWord is off the image).
		 * If the line below yCount is inside the image,
		 * (UInt32)(yCount+1) will be less than hS.
		 */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		iWord = 0;
		vertBits = 0;
		if ((xCount >= 0) && ((UInt32)(yCount + 1) < (UInt32)hS))
		{
		    iWord = pSWord[sWpl];
		    vertBits = iWord & pSWord[0]; /* get vert strokes */
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1); 
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
		        ~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= ((iWord >> (rightShift+1)) & mask) |
				     ((pSWord[sWpl-1]&0x1) << (31-rightShift));
			else
			    oWord |= pSWord[sWpl-1] & 0x1;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
	    }
	    else
	    {	    /* The run ended on the input word boundary.  Fill
		     * in first run in the new word knowing that a
		     * right shift will require bits from the
		     * word to the left.  We can still mask pSWord[-1]
		     * with 0x1 since we'll only be using the LS bit in
		     * any case.
		     */
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0;
		vertBits = 0;
		if ((xCount | yCount) >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1); 
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
		        ~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= ((iWord >> (rightShift+1)) & mask) |
				     ((pSWord[sWpl-1] & 0x1) <<(31-rightShift));
			else
			    oWord |= pSWord[sWpl-1] & 0x1;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* Insert the rest of the runs in this word (if any) into
		 * this output word.  At this point, we know that no right
		 * shift can require bits from beyond the left edge of
		 * the current input word.  That case was handled in
		 * the "else" clause immediately above. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0;
		vertBits = 0;
		if ((xCount | yCount) >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1); 
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
		        ~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= (iWord >> (rightShift+1)) & mask;
			/* otherwise, do nothing, which is the same as
			 * ORing with (iWord >> 32). */
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    if (mask == (UInt32)sWpl)
	    {
		if (yCount != hS - 1)
		    yCount++;		/* put yCount back where it belongs */
		else
		    yCount = 0x80000000;
	    }
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks both xCount and
		 * yCount when fetching input words. */
	if (w <= 0)
	    return;

	    /* Second section of shifting code.  This one checks neither
	     * xCount nor yCount, but does update xCount so it's ready for
	     * the third section.
	     * WE SHOULD BE SPENDING MOST OF OUR TIME HERE
	     * ************* SECTION 2, SHIFTING REQUIRED ***********
	     */
	for (; w > boundary2; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = pSWord[0];
	    vertBits = iWord & pSWord[-sWpl];	/* get vert strokes */
	    if (vertBits &&
		((vertBits<<leftShift) & ~pDWord[-dWpl] & mask))
	    {
		/* We might want to do an extra right shift.  This often
		 * results in a duplicated bit on the end of the word, which
		 * introduces vertical streaks in dithered sections of an
		 * image.  Do a morphological open on vertBits so we ignore
		 * dithered section of the image. */
		vertBits = vertBits & (vertBits >> 1);
		vertBits = vertBits | (vertBits << 1);
		if ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask)
		    oWord = (iWord << (leftShift-1)) & mask;
		else
		    oWord = (iWord << leftShift) & mask; /* no extra shift */
	    }
	    else
		oWord = (iWord << leftShift) & mask; /* insert next run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = pSWord[0];
		vertBits = iWord & pSWord[-sWpl];	/* get vert strokes */
		if (vertBits &&
		    ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1);
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask)
			oWord |= (iWord << (leftShift-1)) & mask;
		    else
			oWord |= (iWord << leftShift) & mask;
		}
		else
		    oWord |= (iWord << leftShift) & mask;
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word.
		 * BUG FIX:  In the first case here, we had been ORing the
		 * the shifted versions of iWord and pSWord[sWpl-1] together,
		 * then applying "mask".  This doesn't work in this case
		 * because the mask still covers bits that have been
		 * set above.  Shifting pSWord[sWpl-1] *left* causes it's
		 * bits to intersect the bits we've already set in oWord.
		 * To fix it, we mask pSWord[sWpl-1] with 0x1 before
		 * the left shift.  This is ok, because we only need the
		 * LS bit of that word to account for the extra right shift.
		 */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		iWord = pSWord[sWpl];
		vertBits = iWord & pSWord[0];
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1);
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
			~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= ((iWord >> (rightShift+1)) & mask) |
				     ((pSWord[sWpl-1]&0x1) << (31-rightShift));
			else
			    oWord |= pSWord[sWpl-1] & 0x1;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
	    }
	    else /* the run ended at the input word boundary.  This
		  * is the only time an extra right shift will require
		  * a bit from the word to the left of the input word.  We
		  * can still mask pSWord[-1] with 0x1 because we'll only
		  * use it's LS bit. */
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = pSWord[0];
		vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1);
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
			~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= ((iWord >> (rightShift+1)) & mask) |
				     ((pSWord[sWpl-1] & 0x1) <<(31-rightShift));
			else
			    oWord |= pSWord[sWpl-1] & 0x1;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;
	    }

	        /* Insert the rest of the runs in this word (if any) into
		 * this output word.  None of these runs can required bits
		 * from the word to the left of the input word. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = pSWord[0];
		vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1);
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
			~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= (iWord >> (rightShift+1)) & mask;
			/* Otherwise, do nothing.  This is the same as
			 * ORing oWord with (iWord >> 32) */
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks neither xCount nor
		 * yCount when fetching input words. */

	    /* Third section of shifting code.  This one checks only
	     * xCount.
	     * ************* SECTION 3, SHIFTING REQUIRED ***********
	     */
	for (; w > 0; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = 0;
	    vertBits = 0;
	    if (xCount >= 0)
	    {
		iWord = pSWord[0];
		vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
	    }
	    if (vertBits &&
		((vertBits<<leftShift) & ~pDWord[-dWpl] & mask))
	    {
		vertBits = vertBits & (vertBits >> 1);
		vertBits = vertBits | (vertBits << 1);
		if ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask)
		    oWord = (iWord << (leftShift-1)) & mask;
		else
		    oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    }
	    else
		oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0;
		vertBits = 0;
		if (xCount >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		}
		if (vertBits &&
		    ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1);
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits<<leftShift) & ~pDWord[-dWpl] & mask)
			oWord |= (iWord << (leftShift-1)) & mask;
		    else
			oWord |= (iWord << leftShift) & mask;
		}
		else
		    oWord |= (iWord << leftShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word.
		 * BUG FIX:  In the first case here, we had been ORing the
		 * the shifted versions of iWord and pSWord[sWpl-1] together,
		 * then applying "mask".  This doesn't work in this case
		 * because the mask still covers bits that have been
		 * set above.  Shifting pSWord[sWpl-1] *left* causes it's
		 * bits to intersect the bits we've already set in oWord.
		 * To fix it, we mask pSWord[sWpl-1] with 0x1 before
		 * the left shift.  This is ok, because we only need the
		 * LS bit of that word to account for the extra right shift.
		 */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		iWord = 0;
		vertBits = 0;
		if (xCount >= 0)
		{
		    iWord = pSWord[sWpl];
		    vertBits = iWord & pSWord[0]; /* get vert strokes */
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1);
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
			~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= ((iWord >> (rightShift+1)) & mask) |
				     ((pSWord[sWpl-1]&0x1) << (31-rightShift));
			else
			    oWord |= pSWord[sWpl-1] & 0x1;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
	    }
	    else
	    {	    /* The run ended on the input word boundary.  Fill
		     * in first run in the new word knowing that a
		     * right shift will require bits from the
		     * word to the left.  We
		     * can still mask pSWord[-1] with 0x1 because we'll only
		     * use it's LS bit. 
		     */
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0;
		vertBits = 0;
		if (xCount >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1);
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
			~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= ((iWord >> (rightShift+1)) & mask) |
				     ((pSWord[sWpl-1] & 0x1) <<(31-rightShift));
			else
			    oWord |= pSWord[sWpl-1] & 0x1;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;
	    }

	        /* Insert the rest of the runs in this word (if any) into
		 * this output word.   At this point, we know that no right
		 * shift can require bits from beyond the left edge of
		 * the current input word.  That case was handled in
		 * the "else" clause immediately above. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0;
		vertBits = 0;
		if (xCount >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = iWord & pSWord[-sWpl]; /* get vert strokes */
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & ~pDWord[-dWpl] & mask))
		{
		    vertBits = vertBits & (vertBits >> 1);
		    vertBits = vertBits | (vertBits << 1);
		    if ((vertBits>>rightShift) &
			~pDWord[-dWpl] & mask)
		    {
			if (rightShift != 31)
			    oWord |= (iWord >> (rightShift+1)) & mask;
			/* Otherwise, do nothing.  This is the same as
			 * ORing oWord with (iWord >> 32) */
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks only xCount
		 * when fetching input words. */
	return;
    }	/* end of case where shifting IS required */
}
#endif	/* XIS_USE */

/* PRIVATE */
/**************************************************************************
 *  ip_rotateNoInterpLineNiceGray()
 *		Generates one rotated line without using interpolation.
 *		Tries to avoid shearing vertical strokes.  Operates on gray
 *		pixels, but DOES NOT do averaging.  This routine just
 *		rearranges the pixels with some attempt to avoid jaggies.
 *
 *	Input:
 *		pS =		pointer to 32 bit word containing first pixel
 *				of unrotated data.
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination line
 *		dBpl =		dest bytes/line
 *		w =		width of source image in bits
 *		xCount = 	increment this count at each move towards the
 *				right.  While xCount < 0, we are off the image.
 *				Distance to horizontal edge.
 *		yCount =	increment this count at each move towards the
 *				bottom.  While yCount < 0, we are off the image.
 *				Distance to vertical edge.
 *		leftShift =	position of leftmost image bit within
 *				first image word as measured from left
 *				edge of first image word (0 => input and
 *				output words are aligned).  This bit
 *				should be shifted to become bit 0 (leftmost) 
 *				of the first output word, hence it is the
 *				distance by which input words should be
 *				shifted left to be aligned with the output
 *				words.
 *		pMaskTable =	Pointer to sequence of masks used to insert
 *				runs of pixels from input words into output
 *				words.
 *		depth =		Number of bits/pixel in the image.  Used
 *				to compute masks and shift distances.
 *		boundary1 =	Check xCount and yCount when reading input
 *				words for all output words
 *				processed to the left of boundary1.
 *		boundary2 =	Check xCount when reading input
 *				words for all output words
 *				processed to the left of boundary2.
 *		hS =		height of source image in pixels.  This is
 *				down here because it was the solution to
 *				a bug that was found long after everything
 *				else was written.
 *	Output:
 *		Stores 1 rotated line into output image.
 *
 **************************************************************************/
void  CDECL 
ip_rotateNoInterpLineNiceGray (
    IPState    *pIPState,
    UInt8Ptr	pS,
    Int32	sBpl,
    UInt8Ptr	pD,
    Int32	dBpl,
    Int32	w,
    Int32	xCount,
    Int32	yCount,
    Int32	leftShift,
    UInt32Ptr	pMaskTable,
    Int32	depth,
    UInt32	rightWhiteVal,
    Int32	boundary1,
    Int32	boundary2,
    Int32	hS)
#ifdef XIS_USE		/* XIS doesn't use this routine, so turn it into
			 * a stub.
			 */
{
    pIPState = pIPState;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dBpl = dBpl;
    w = w;
    xCount = xCount;
    yCount = yCount;
    leftShift = leftShift;
    pMaskTable = pMaskTable;
    depth = depth;
    rightWhiteVal = rightWhiteVal;
    boundary1 = boundary1;
    boundary2 = boundary2;
}
#else
{
    UInt32	leftmostInputBit; /* a mask containing a single
				 * 1 in the position where
				 * the MS bit of the input
				 * words aligns with the
				 * output words. */
    UInt32	rightmostInputBit; /* a mask containing a single
				 * 1 in the position where
				 * the LS bit of the input
				 * words aligns with the
				 * output words. */
    UInt32	oWord;		/* the output word under
				 * formation */
    UInt32	iWord;		/* word read from input */
    UInt32	mask;		/* the current mask */
    Int32	rightShift;	/* distance by which input words
				 * should be shifted right.  As
				 * discussed above, this is constant
				 * across the entire line. */
    UInt32Ptr	pSWord;		/* pointer to source words */
    UInt32Ptr	pDWord;		/* pointer to dest words */
    Int32	sWpl;		/* source words/line */
    Int32	dWpl;		/* dest words/line */
    UInt32	vertBits;	/* holds the bits that were
				 * aligned vertically in the input */
    UInt32	msBits;		/* This is ANDed with the shifted
				 * vertical strokes and "mask"
				 * when deciding
				 * if the normal shift keeps vertical
				 * strokes aligned.  We don't want
				 * to be confused by the LS bits of
				 * each pixel. */
    UInt32	lsPixelMask;	/* used to insert 1 pixel during
				 * extra right shifts. */

	/* Keep the compiler happy */
    pIPState = pIPState;

	/* we'll be reading the source and writing the destination
	 * is units of words, so use word pointers */
    pSWord = (UInt32Ptr) pS;
    pDWord = (UInt32Ptr) pD;
    sWpl = sBpl >> 2;
    dWpl = dBpl >> 2;
    rightShift = 32 - leftShift;

    switch (depth)
    {
    case 2:
	msBits = 0xAAAAAAAA;
	lsPixelMask = 0x3;
	break;
    case 4:
	msBits = 0x88888888;
	lsPixelMask = 0xF;
	break;
    case 8:
	msBits = 0x80808080;
	lsPixelMask = 0xFF;
	break;
    default:
	return;			/* should never happen */
    }

	/* Do the input and output line up perfectly?  If so, no shifting
	 * is required */
    if (leftShift == 0)
    {
	    /* Loop through the words of this row.  Start
	     * with the first section of code.  This one
	     * checks both xCount and yCount on each fetch of
	     * input word.
	     * ************* SECTION 1, NO SHIFTING ***********
	     */
	for (; w > boundary1; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = 0xFFFFFFFF;
	    vertBits = 0;
	    if ((xCount | yCount) >= 0)
	    {
		iWord = pSWord[0];
		vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
		    /* If we're near the right edge of the image,
		     * fill last partial word with white.
		     */
		if (xCount+32 < 0)
		    iWord |= rightWhiteVal;
	    }

		/* do the vert input bits line up with vert output? */
	    if (vertBits &&
		(vertBits & pDWord[-dWpl] & msBits & mask))
	    {
		    /* delete 1 pixel strokes, but leave the wider strokes
		     * unaffected */
		vertBits = vertBits & (vertBits >> depth); 
		vertBits = vertBits | (vertBits << depth);
		if (vertBits & pDWord[-dWpl] & msBits & mask)
		    iWord = (iWord >> depth) | (pSWord[-1] << (32 - depth));
	    }
	    oWord = iWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */
	    if (yCount != (Int32)0x80000000)
		yCount--;
	    else
		yCount = hS - 1;

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		if ((xCount | yCount) >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if (xCount+32 < 0)
			iWord |= rightWhiteVal;
		}
		    /* do the vert input bits line up with vert output? */
		if (vertBits &&
		    (vertBits & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth); 
		    vertBits = vertBits | (vertBits << depth);
		    if (vertBits & pDWord[-dWpl] & msBits & mask)
			iWord = (iWord >> depth) | (pSWord[-1] << (32-depth));
		}
		oWord |= iWord & mask;	/* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    if (mask == (UInt32)sWpl)
	    {
		if (yCount != hS - 1)
		    yCount++;		/* put yCount back where it belongs */
		else
		    yCount = 0x80000000;
	    }
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks xCount and yCount. */
	if (w <= 0)
	    return;

	    /* Second section of no-shifting code.  This one checks neither
	     * xCount nor yCount, but does update xCount so it's ready for
	     * the third section.
	     * ************* SECTION 2, NO SHIFTING ***********
	     */
	for (; w > boundary2; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = pSWord[0];
	    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
	    if (vertBits &&
		(vertBits & pDWord[-dWpl] & msBits & mask))
	    {
		vertBits = vertBits & (vertBits >> depth); 
		vertBits = vertBits | (vertBits << depth);
		if (vertBits & pDWord[-dWpl] & msBits & mask)
		    iWord = (iWord >> depth) | (pSWord[-1] << (32-depth));
	    }
	    oWord = iWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		iWord = pSWord[0];
		vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
		if (vertBits &&
		    (vertBits & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth); 
		    vertBits = vertBits | (vertBits << depth);
		    if (vertBits & pDWord[-dWpl] & msBits & mask)
			iWord = (iWord >> depth) | (pSWord[-1] << (32-depth));
		}
		oWord |= iWord & mask; /* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks NEITHER xCount 
		 * nor yCount. */

	    /* Third section of no-shifting code.  This one checks only
	     * xCount.
	     * ************* SECTION 3, NO SHIFTING ***********
	     */
	for (; w > 0; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = 0xFFFFFFFF;
	    vertBits = 0;
		/* If the entire word is inside the image (rare),
		 * use it.  If only part of the rightmost word is
		 * inside the image, OR in white pixels in the
		 * frame section.  We can do this because the
		 * frame is always 0's.
		 */
	    if (xCount >= 0)
	    {
		if (xCount+32 >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */			}
		else
		{
		    iWord = pSWord[0];
		    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
		    iWord |= rightWhiteVal;		/* insert white frame */
		}
	    }

		/* do the vert input bits line up with vert output? */
	    if (vertBits &&
		(vertBits & pDWord[-dWpl] & msBits & mask))
	    {
		vertBits = vertBits & (vertBits >> depth); 
		vertBits = vertBits | (vertBits << depth);
		if (vertBits & pDWord[-dWpl] & msBits & mask)
		    iWord = (iWord >> depth) | (pSWord[-1] << (32-depth));
	    }
	    oWord = iWord & mask;	/* insert first run into word */
	    pSWord -= sWpl;		/* end of run, so move up a line */

	    while (!(mask & 0x1))	/* keep going until we've written
					 * the rightmost bit in this word */
	    {
		mask = *pMaskTable++;	/* get next mask */
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		    /* Insert white bits on right as necessary */
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
		    {
			iWord = pSWord[0];
			vertBits = ~(iWord | pSWord[-sWpl]); /* vert strokes */
		    }
		    else
		    {
			iWord = pSWord[0];
			vertBits = ~(iWord | pSWord[-sWpl]); /* vert strokes */
			iWord |= rightWhiteVal;		/* insert white frame */
		    }
		}

		if (vertBits &&
		    (vertBits & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth); 
		    vertBits = vertBits | (vertBits << depth);
		    if (vertBits & pDWord[-dWpl] & msBits & mask)
			iWord = (iWord >> depth) | (pSWord[-1] << (32-depth));
		}
		oWord |= iWord & mask;	/* insert next run into word */
		pSWord -= sWpl;		/* end of run, so move up a line */
					/* this is corrected below if that run
					 * really extended off the end of
					 * the output word. */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask + 1;
	    xCount += 32;		/* advance xCount */
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks only xCount. */
	return;

    }	/* end of case where no shifting is required */

    else

    {	/* each input word must be shifted left or right before inserting
	 * its bits */

	    /* rightmostInputBit will have a 1 in the output bit position
	     * corresponding to the least significant bit of the input
	     * words that are shifted left by leftShift.  leftmostInputBit
	     * will have a 1 in the output bit position corresponding
	     * to the most significant bit of input words that are
	     * shifted right by 32-leftShift.
	     */
	rightmostInputBit = 0x80000000;		/* just to put the const
						 * into a register */
	leftmostInputBit = rightmostInputBit >> (32-leftShift);
	rightmostInputBit = leftmostInputBit << 1;

	    /* Loop through the words of this row.  Start
	     * with the first section of code.  This one
	     * checks both xCount and yCount on each fetch of
	     * input word.
	     * ************* SECTION 1, SHIFTING REQUIRED ***********
	     */
	for (; w > boundary1; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = 0xFFFFFFFF;
	    vertBits = 0;
	    if ((xCount | yCount) >= 0)
	    {
		iWord = pSWord[0];
		vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
		    /* If we're near the right edge of the image,
		     * fill last partial word with white.
		     */
		if (xCount+32 < 0)
		    iWord |= rightWhiteVal;
	    }
	    if (vertBits &&
		((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask))
	    {
		vertBits = vertBits & (vertBits >> depth); 
		vertBits = vertBits | (vertBits << depth);
		if ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask)
		    oWord = (iWord << (leftShift-depth)) & mask;
		else
		    oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    }
	    else
		oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
	    if (yCount != (Int32)0x80000000)
		yCount--;
	    else
		yCount = hS - 1;

		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		if ((xCount | yCount) >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if (xCount+32 < 0)
			iWord |= rightWhiteVal;
		}
		if (vertBits &&
		    ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth); 
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask)
			oWord |= (iWord << (leftShift-depth)) & mask;
		    else
			oWord |= (iWord << leftShift) & mask;
		}
		else
		    oWord |= (iWord << leftShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
		    /* if yCount is at the bottom of the image, set it equal
		     * to hS - 1 (the last line of the image.  Otherwise, just
		     * decrement it.  This allows us to process diagonal paths
		     * that go right through the image from below to above.
		     */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word.
		 * BUG FIX:  In the first case here, we had been ORing the
		 * the shifted versions of iWord and pSWord[sWpl-1] together,
		 * then applying "mask".  This doesn't work in this case
		 * because the mask still covers bits that have been
		 * set above.  Shifting pSWord[sWpl-1] *left* causes it's
		 * bits to intersect the bits we've already set in oWord.
		 * To fix it, we mask pSWord[sWpl-1] with lsPixelMask before
		 * the left shift.  This is ok, because we only need the
		 * LS pixel of that word to account for the extra right shift.
		 * Note that we check rightShift before using rightShift+depth.
		 * If rightShift == (32-depth), we just ignore the shifted term
		 * because the SPARC (and 486) only use the LS 5 bits of the
		 * shift distance so a shift by (32-depth)+depth bits is
		 * no shift at all instead of a shift by 32 - which should
		 * have filled the word with 0's.
		 * Note the funny test for yCount.  We really want to
		 * look at yCount + 1, the line below pSWord.
		 * If yCount+1 is negative, (UInt32)(yCount+1) is bigger
		 * than any possible value for hS.  If we've just set
		 * yCount to correspond to the last image line (hS-1),
		 * then yCount + 1 == hS so the test will fail, as it
		 * should (the line below pSWord is off the image).
		 * If the line below yCount is inside the image,
		 * (UInt32)(yCount+1) will be less than hS.
		 */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		if ((xCount >= 0) && ((UInt32)(yCount + 1) < (UInt32)hS))
		{
		    iWord = pSWord[sWpl];
		    vertBits = ~(iWord | pSWord[0]); /* get vert strokes */
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if(xCount+32 < 0)
			iWord |= rightWhiteVal;
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth); 
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
		        pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= ((iWord >> (rightShift+depth)) & mask) |
				     ((pSWord[sWpl-1] & lsPixelMask) <<
						(32-(rightShift+depth)));
			else
			    oWord |= pSWord[sWpl-1] & lsPixelMask;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
	    }
	    else
	    {	    /* The run ended on the input word boundary.  Fill
		     * in first run in the new word knowing that a
		     * right shift will require bits from the
		     * word to the left.  We can still mask pSWord[-1]
		     * with lsPixelMask since we'll only be using the LS 
		     * pixel in any case.
		     */
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		if ((xCount | yCount) >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if (xCount+32 < 0)
			iWord |= rightWhiteVal;
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth); 
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
		        pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= ((iWord >> (rightShift+depth)) & mask) |
				     ((pSWord[sWpl-1] & lsPixelMask) <<
						(32 - (rightShift+depth)));
			else
			    oWord |= pSWord[sWpl-1] & lsPixelMask;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* Insert the rest of the runs in this word (if any) into
		 * this output word.  At this point, we know that no right
		 * shift can require bits from beyond the left edge of
		 * the current input word.  That case was handled in
		 * the "else" clause immediately above. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		if ((xCount | yCount) >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
			/* If we're near the right edge of the image,
			 * fill last partial word with white.
			 */
		    if (xCount+32 < 0)
			iWord |= rightWhiteVal;
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth); 
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
		        pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= (iWord >> (rightShift+depth)) & mask;
			/* otherwise, do nothing, which is the same as
			 * ORing with (iWord >> 32). */
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
		if (yCount != (Int32)0x80000000)
		    yCount--;
		else
		    yCount = hS - 1;
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    if (mask == (UInt32)sWpl)
	    {
		if (yCount != hS - 1)
		    yCount++;		/* put yCount back where it belongs */
		else
		    yCount = 0x80000000;
	    }
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks both xCount and
		 * yCount when fetching input words. */
	if (w <= 0)
	    return;

	    /* Second section of shifting code.  This one checks neither
	     * xCount nor yCount, but does update xCount so it's ready for
	     * the third section.
	     * WE SHOULD BE SPENDING MOST OF OUR TIME HERE
	     * ************* SECTION 2, SHIFTING REQUIRED ***********
	     */
	for (; w > boundary2; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = pSWord[0];
	    vertBits = ~(iWord | pSWord[-sWpl]);	/* get vert strokes */
	    if (vertBits &&
		((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask))
	    {
		/* We might want to do an extra right shift.  This often
		 * results in a duplicated bit on the end of the word, which
		 * introduces vertical streaks in dithered sections of an
		 * image.  Do a morphological open on vertBits so we ignore
		 * dithered section of the image. */
		vertBits = vertBits & (vertBits >> depth);
		vertBits = vertBits | (vertBits << depth);
		if ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask)
		    oWord = (iWord << (leftShift-depth)) & mask;
		else
		    oWord = (iWord << leftShift) & mask; /* no extra shift */
	    }
	    else
		oWord = (iWord << leftShift) & mask; /* insert next run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = pSWord[0];
		vertBits = ~(iWord | pSWord[-sWpl]);	/* get vert strokes */
		if (vertBits &&
		    ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth);
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask)
			oWord |= (iWord << (leftShift-depth)) & mask;
		    else
			oWord |= (iWord << leftShift) & mask;
		}
		else
		    oWord |= (iWord << leftShift) & mask;
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word.
		 * BUG FIX:  In the first case here, we had been ORing the
		 * the shifted versions of iWord and pSWord[sWpl-1] together,
		 * then applying "mask".  This doesn't work in this case
		 * because the mask still covers bits that have been
		 * set above.  Shifting pSWord[sWpl-1] *left* causes it's
		 * bits to intersect the bits we've already set in oWord.
		 * To fix it, we mask pSWord[sWpl-1] with lsPixelMask before
		 * the left shift.  This is ok, because we only need the
		 * LS pixel of that word to account for the extra right shift.
		 */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		iWord = pSWord[sWpl];
		vertBits = ~(iWord | pSWord[0]);
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth);
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
			pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= ((iWord >> (rightShift+depth)) & mask) |
				     ((pSWord[sWpl-1] & lsPixelMask) <<
						(32-(rightShift+depth)));
			else
			    oWord |= pSWord[sWpl-1] & lsPixelMask;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
	    }
	    else /* the run ended at the input word boundary.  This
		  * is the only time an extra right shift will require
		  * a bit from the word to the left of the input word.  We
		  * can still mask pSWord[-1] with lsPixelMask because
		  * we'll only use it's LS pixel. */
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = pSWord[0];
		vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth);
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
			pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= ((iWord >> (rightShift+depth)) & mask) |
				     ((pSWord[sWpl-1] & lsPixelMask) <<
						(32-(rightShift+depth)));
			else
			    oWord |= pSWord[sWpl-1] & lsPixelMask;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;
	    }

	        /* Insert the rest of the runs in this word (if any) into
		 * this output word.  None of these runs can required bits
		 * from the word to the left of the input word. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = pSWord[0];
		vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth);
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
			pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= (iWord >> (rightShift+depth)) & mask;
			/* Otherwise, do nothing.  This is the same as
			 * ORing oWord with (iWord >> 32) */
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks neither xCount nor
		 * yCount when fetching input words. */

	    /* Third section of shifting code.  This one checks only
	     * xCount.
	     * ************* SECTION 3, SHIFTING REQUIRED ***********
	     */
	for (; w > 0; w -= 32)
	{
	    mask = *pMaskTable++;	/* get first mask for this word */
	    iWord = 0xFFFFFFFF;
	    vertBits = 0;
		/* If the entire word is inside the image (rare),
		 * use it.  If only part of the rightmost word is
		 * inside the image, OR in white pixels in the
		 * frame section.  We can do this because the
		 * frame is always 0's.
		 */
	    if (xCount >= 0)
	    {
		if (xCount+32 >= 0)
		{
		    iWord = pSWord[0];
		    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */			}
		else
		{
		    iWord = pSWord[0];
		    vertBits = ~(iWord | pSWord[-sWpl]); /* get vert strokes */
		    iWord |= rightWhiteVal;		/* insert white frame */
		}
	    }

	    if (vertBits &&
		((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask))
	    {
		vertBits = vertBits & (vertBits >> depth);
		vertBits = vertBits | (vertBits << depth);
		if ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask)
		    oWord = (iWord << (leftShift-depth)) & mask;
		else
		    oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    }
	    else
		oWord = (iWord << leftShift) & mask; /* insert 1st run */
	    pSWord -= sWpl;		/* end of run, so move up a line */
		/* keep going until we've written the output bit
		 * corresponding to the rightmost input bit */
	    while (!(mask & rightmostInputBit))	
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		    /* Insert white bits on right as necessary */
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
		    {
			iWord = pSWord[0];
			vertBits = ~(iWord | pSWord[-sWpl]); /* vert strokes */
		    }
		    else
		    {
			iWord = pSWord[0];
			vertBits = ~(iWord | pSWord[-sWpl]); /* vert strokes */
			iWord |= rightWhiteVal;		/* insert white frame */
		    }
		}

		if (vertBits &&
		    ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth);
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits<<leftShift) & pDWord[-dWpl] & msBits & mask)
			oWord |= (iWord << (leftShift-depth)) & mask;
		    else
			oWord |= (iWord << leftShift) & mask;
		}
		else
		    oWord |= (iWord << leftShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the left portion of the output
		 * word.  If the most recent mask overlaps the leftmost
		 * word of the next input, we should go back down a
		 * line, move over 1 word and finish filling the mask
		 * Otherwise, just move over 1 word.
		 * BUG FIX:  In the first case here, we had been ORing the
		 * the shifted versions of iWord and pSWord[sWpl-1] together,
		 * then applying "mask".  This doesn't work in this case
		 * because the mask still covers bits that have been
		 * set above.  Shifting pSWord[sWpl-1] *left* causes it's
		 * bits to intersect the bits we've already set in oWord.
		 * To fix it, we mask pSWord[sWpl-1] with lsPixelMask before
		 * the left shift.  This is ok, because we only need the
		 * LS pixel of that word to account for the extra right shift.
		 */
	    pSWord++;
	    xCount += 32;
	    if (mask & leftmostInputBit)
	    {
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
		    {
			iWord = pSWord[sWpl];
			vertBits = ~(iWord | pSWord[0]); /* vert strokes */
		    }
		    else
		    {
			iWord = pSWord[sWpl];
			vertBits = ~(iWord | pSWord[0]); /* vert strokes */
			iWord |= rightWhiteVal;		/* insert white frame */
		    }
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth);
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
			pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= ((iWord >> (rightShift+depth)) & mask) |
				     ((pSWord[sWpl-1] & lsPixelMask) <<
						(32-(rightShift+depth)));
			else
			    oWord |= pSWord[sWpl-1] & lsPixelMask;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
	    }
	    else
	    {	    /* The run ended on the input word boundary.  Fill
		     * in first run in the new word knowing that a
		     * right shift will require bits from the
		     * word to the left.  We
		     * can still mask pSWord[-1] with lsPixelMask
		     * because we'll only use it's LS pixel. 
		     */
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
		    {
			iWord = pSWord[0];
			vertBits = ~(iWord | pSWord[-sWpl]); /* vert strokes */
		    }
		    else
		    {
			iWord = pSWord[0];
			vertBits = ~(iWord | pSWord[-sWpl]); /* vert strokes */
			iWord |= rightWhiteVal;		/* insert white frame */
		    }
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth);
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
			pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= ((iWord >> (rightShift+depth)) & mask) |
				     ((pSWord[sWpl-1] & lsPixelMask) <<
						(32-(rightShift+depth)));
			else
			    oWord |= pSWord[sWpl-1] & lsPixelMask;
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;
	    }

	        /* Insert the rest of the runs in this word (if any) into
		 * this output word.   At this point, we know that no right
		 * shift can require bits from beyond the left edge of
		 * the current input word.  That case was handled in
		 * the "else" clause immediately above. */
	    while (!(mask & 0x1))
	    {
		mask = *pMaskTable++;	/* get next mask for this word */
		iWord = 0xFFFFFFFF;
		vertBits = 0;
		if (xCount >= 0)
		{
		    if (xCount+32 >= 0)
		    {
			iWord = pSWord[0];
			vertBits = ~(iWord | pSWord[-sWpl]); /* vert strokes */
		    }
		    else
		    {
			iWord = pSWord[0];
			vertBits = ~(iWord | pSWord[-sWpl]); /* vert strokes */
			iWord |= rightWhiteVal;		/* insert white frame */
		    }
		}
		if (vertBits &&
		    ((vertBits>>rightShift) & pDWord[-dWpl] & msBits & mask))
		{
		    vertBits = vertBits & (vertBits >> depth);
		    vertBits = vertBits | (vertBits << depth);
		    if ((vertBits>>rightShift) &
			pDWord[-dWpl] & msBits & mask)
		    {
			if (rightShift != (32-depth))
			    oWord |= (iWord >> (rightShift+depth)) & mask;
			/* Otherwise, do nothing.  This is the same as
			 * ORing oWord with (iWord >> 32) */
		    }
		    else
			oWord |= (iWord >> rightShift) & mask;
		}
		else
		    oWord |= (iWord >> rightShift) & mask; /* insert next run */
		pSWord -= sWpl;		/* end of run, so move up a line */
	    }
	        /* We're at the end of the word, correct the source
		 * address with the next word from the mask table.
		 * that word contains sWpl if the run should continue
		 * in the next output word.  It contains 0 if the run
		 * happened to end at the end of the output word. */
	    mask = *pMaskTable++;
	    pSWord += mask;
	    *pDWord++ = oWord;		/* write the rotated word */

	}	/* end of section of code that checks only xCount
		 * when fetching input words. */
	return;
    }	/* end of case where shifting IS required */
}
#endif	/* XIS_USE */


/*

$Log: minirt2.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.6  1998/01/13 23:52:54  ddavies
minirt2.c -> /product/ipcore/ipshared/src/RCS/minirt2.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1996/10/05 02:25:51  ddavies
minirt2.c -> /product/ipcore/ipshared/src/RCS/minirt2.c,v

Make sort of ugly changes to the semantics of xCount and yCount.  yCount now
is changed when we cross the lower boundary of the image so it's negative
whenever it's outside the image.  This accounts for the fact that it might
cross both the upper and lower boundaries.

Revision 1.4  1995/12/13 02:50:32  ejaquith
minirt2.c -> /product/ipcore/ipshared/src/RCS/minirt2.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/09/21  03:00:56  ddavies
 * minirt2.c -> /product/ipcore/ipshared/src/RCS/minirt2.c,v
 *
 * Fixed a bug in ip_rotateNoInterpLineNiceGray.  Involved adding the rightWhiteVal arg.
 * The bug appeared in images whose rightmost pixel did not appear on a
 * word boundary.  In this case, we sometimes read the entire word (including
 * frame) containing the right pixel and put it into the output image (causing
 * frame to appear in the midst of an image).  We also sometimes ignored the
 * rightmost word entirely, causing any pixels in that word to be lost.
 * The calling routines now calculate xCount correctly, fixing the bug that
 * caused us to ignore the rightmost word.  The rightWhiteVal arg contains
 * the value of white in the pixels that would otherwise contain right frame
 * in the rightmost word of each input image line.  It is ORed with the
 * rightmost word whenever it's read, inserting white in place of the frame.
 *
 * Revision 1.2  1995/02/18  05:52:31  ddavies
 * minirt2.c -> /product/ipcore/ipshared/src/RCS/minirt2.c,v
 *
 * Fix really arcane bug in both nice rotation primitives.  When a run ends,
 * we have to decide whether the first word of the next run should be
 * shifted or not.  If we decide to do an extra right shift, we have to get
 * a new pixel on the left from somewhere.  The various sections of code
 * were inconsistent on this.  Some fetched the new pixel from the input
 * line below the current one, some fetched it from the current input line.
 * This caused a *very* small number of pixels in the output image to change
 * by a small amount.  Hard to find!  Anyway, we always fetch from the line
 * below the current one since this is really the extension of the current
 * output run.
 *
 * Revision 1.1  1995/01/12  03:12:48  ddavies
 * Initial revision
 *

 */

