/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miorth48.c--includes
 *
 *		ip_rotate48Block90()	rotates a 4 bit/pixel or
 *					8 bit/pixel row into a column
 *		ip_rotate48Block270()	rotates a 4 bit/pixel or
 *					8 bit/pixel row into a column
 *		ip_make4bppRotTable()	Makes a table used in both
 *					rotations of 4 bpp images.
 */


#include <stddef.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /* prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miorth48.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* A reasonable way to rotate an image by 90 degrees is to divide the
 * image into rectangles, rotate the rectangles, and write them.  
 * The nice feature of this scheme is that we always write full 32-bit 
 * words.  The dimensions of the rectangles in pixels depend on the 
 * number of bits/pixel.
 *
 * How to rotate the tiles?  One way is to test and branch on each bit,
 * conditionally ORing a mask into one of 8 dest registers.  Nice, but
 * this takes about three opcodes per bit.  Can't we do better?
 *
 * Another approach is to use a lookup table to spread the 8 bits of each
 * byte out into a 32-bit word.  When rotating 4 bit/pixel images, we
 * only have two pixels in each byte, so we spread them into two groups.
 * When we've done that 4 times, we have two groups, each with 4 pixels.
 * We'd continue, but this fills a 32 bit word which is all our processors
 * have available.  So, we carefully store these words and repeat the
 * process.  We combine the two sets of 4-pixel chunks found and write
 * out the 2 resulting 32-bit words.
 *
 * So how does this really work?  Here are the steps to rotate a block 
 * by 90 degrees clockwise.  The block is two pixels wide by 4 pixels tall.
 * 1. Read bottommost byte in block.
 * 2. Use lookup table to spread bits [7:0] into bits [19,18,17,16, 3,2,1,0]
 *    in an accumulating register.
 * 3. Shift the accumulated register left by 4.
 * 4. Repeat steps 1-3 for the next three source bytes.  Each new source
 *    byte is closer to the top of the source image.  We also have to
 *    OR the new table lookup bits into the register.
 * 5. Leave the results in inputWord0.
 * 6. Repeat steps 1-4 for the next four source bytes, leaving the result
 *    in inputWord1.  We now have 8 bytes of info, but the most significant
 *    16 bits of each result word is in inputWord0 and the least significant is
 *    in inputWord1.
 * 7. Set DestB0 = (inputWord0 & 0xFFFF0000) | ((inputWord1 >> 16)
 *    This forms the upper output word (stuff on left becomes stuff on top).
 * 9. Set DestB1 = ((inputWord0 << 16) | (inputWord1 & 0x0000FFFF)
 *    This forms the lower output word.
 * 12. If either DestB0 or DestB1 are non-zero, write them both to the
 *    destination, DestB0 on top..
 *
 * This last step bring up an interesting point.  Most of the images we
 * deal with are white.  Shouldn't there be a check for all-0 source bytes?
 * Well, on the SPARC, we'd save a lookup (2 cycles) and an OR (1 cycle).
 * The test costs a tst (1 cycle) and a branch (1 cycle if the compiler
 * can use the interlock).  We'll ignore it for now.  After it works, we'll
 * look at the compiled code and see if it looks worthwhile.
 *
 * In what order do we address the source and destination?  The 90 degree
 * rotation is defined to be clockwise, so the lower left corner becomes
 * the upper left corner.  The first block rotated will be in the lower
 * left corner and will be stored in the upper left corner of the destination
 * image.  To take advantage of the processor's cache, the second block
 * read will be just to the right of the first block read and will be stored
 * just below the first block stored.  Both the SS2 and the 486 have
 * write through caches that don't fetch locations being written into the
 * cache implicitly, so the order in which we write dest words doesn't matter.
 *
 * What about images whose widths are not even multiples of 8 or whose
 * heights are not even mulitples of 32?  Look at the height first.  We
 * want the bottommost line of the source image to become the leftmost bits
 * of the leftmost words of the destination image.  That means we should
 * start the bottom rectangle exactly on the bottommost line of the
 * source image.  The topmost rectangles in the source image may well include
 * some of the upper frame.  Big deal, that upper frame will become the
 * right frame in the destination image.
 *
 * What about the right edge of the source image?  The source image will
 * likely end somewhere inside the rightmost 8 bit source words.  Again,
 * big deal.  The extra 0's to the right of the real image will appear
 * as part of the lower frame in the destination image.
 *
 * In both cases, we can safely rotate entire rectangles without worrying
 * about edge conditions.
 *
 * 270 degree rotation (90 degrees counterclockwise) is almost the same
 * as 90 degree rotation, except
 *	1. We scan across the source image from left->right (fast scan) and
 *	   top->bottom (slow scan) instead of left->right, bottom->top.
 *	2. Each block is rotated counterclockwise instead of clockwise.
 *
 *********************
 *
 * For 8 bit/pixel images, it gets too simple.  We just read four bytes in
 * a row vertically, shifting them into a horizontal word, and write the
 * single result.  We keep the caches fairly happy by reading 4 bytes
 * vertically, moving over a byte, reading four more bytes and so on.
 */

/* dimensions of the rectangles we rotate */

/* DO NOT CHANGE THESE WITHOUT CHANGING THE COPIES IN miorthrt.c AND
 * ssorthst.c!!!
 */

/* 4 bit/pixel images */
#define	cSrcRectWidth4		2	/* in pixels */
#define cLogSrcRectWidth4	1
#define cSrcRectWidthInBytes4 ((cSrcRectWidth4*4)>>3)
#define cSrcRectHeight4		8
#define cLogSrcRectHeight4	3

/* 8 bit/pixel images */
#define	cSrcRectWidth8		1	/* in pixels */
#define cLogSrcRectWidth8	0
#define cSrcRectWidthInBytes8 ((cSrcRectWidth8*8)>>3)
#define cSrcRectHeight8		4
#define cLogSrcRectHeight8	2


/* PRIVATE */
/**************************************************************************
 *  ip_rotate48Block90()
 *		rotates a row of 4bit/pixel or 8 bit/pixel rectangles
 *		by 90 degrees clockwise.
 *
 *	Input:
 *		pS =		Pointer to topmost halfword of source block
 *		sBpl =		Source bytes/line
 *		pD =		Pointer to topmost word of dest block
 *		sBpl =		Dest bytes/line
 *		rotationTable = Ptr to table used to spread source values.
 *				This table lets us do the rotation in two
 *				steps, the first spreads the values out so
 *				we can accumulate contiguous fragments of
 *				dest bytes in a source word.  The second
 *				assembles these fragments into real dest
 *				bytes or words.
 *		blockCount =	number of blocks to be rotated in a
 *				row of blocks across the source image.
 *		srcAddrFixup =	Number which is added to the final source
 *				address for one block to convert it to the
 *				initial source address for the next block
 *				in the row of blocks.
 *		depth =		Number of bits/pixel in src and dest images.
 **************************************************************************/
void  CDECL 
ip_rotate48Block90 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *rotationTable,
    Int32	blockCount,
    Int32	srcAddrFixup,
    Int32	depth)
{
    PROLOGUE;
    UInt32	inputWord0, inputWord1;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Do the correct code for the current image depth */
    switch (depth)
    {

	/* ********** 90 degree Rotate of 4 bit/pixel images ********* */

    case 4:

	    /* If this is a SPARC, we can just address the bytes
	     * without modifying the addresses.  If this is a PC,
	     * we need to XOR each address with 3 before using it.
	     * That's a pain, execution time went up from 0.71
	     * sec to 0.87 sec when we put it in.  So, we note that
	     * we're accessing source bits in a string of vertical
	     * bytes.  We know that all source images are padded
	     * to 4 byte boundaries horizontally, therefore, the
	     * LS two bits of source address do not change in the
	     * 32 source rect accesses.  So, instead of doing
	     * 32 XOR's with 3, we'll do one XOR at the top of
	     * the loop and use that address everywhere.
	     */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* loop through the rectangles in a row of rects */
	for (;
	     blockCount > 0;
	     blockCount--, pS += srcAddrFixup)
#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* loop through the rectangles in a row of rects */
	for (pS = (UInt8 *) (((UInt32) pS) ^ 3);
	     blockCount > 0;
	     blockCount--,
		pS = (UInt8 *) ((((UInt32) pS ^ 3) + srcAddrFixup) ^ 3))
#endif /* cAlpacaPCFmt */

	{

		/* PROCESS BOTTOM 2x4 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord0 <<= 4;

	    inputWord0 |= rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord0 <<= 4;

	    inputWord0 |= rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord0 <<= 4;

	    inputWord0 |= rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;


		/* PROCESS TOP 2x4 SQUARE */

	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord1 <<= 4;

	    inputWord1 |= rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord1 <<= 4;

	    inputWord1 |= rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord1 <<= 4;

	    inputWord1 |= rotationTable[*pS];	/* byte #1 */

		/* We used to test to see if there were any non-blank
		 * registers before writing.  It didn't speed things
		 * up and required the user to clear the image before
		 * calling this routine.  Now we just write the thing.
		 */
	    ULONG_ACCESS(pD) = (inputWord0 & 0xFFFF0000) |
			       (inputWord1 >> 16);
	    pD += dBpl;
	    ULONG_ACCESS(pD) = (inputWord0 << 16) |
			       (inputWord1 & 0x0000FFFF);
	    pD += dBpl;

	}	/* end of rotating all rects in a row of rectangles */
	break;


	/* ************* 90 degree Rotate of 8 bit/pixel images *********** */


    default: /* depth == 8 bits/pixel */


	    /* If this is a SPARC, we can just address the bytes
	     * without modifying the addresses.  If this is a PC,
	     * we need to XOR each address with 3 before using it.
	     * That's a pain.  So, we note that
	     * we're accessing source bits in a string of vertical
	     * bytes.  We know that all source images are padded
	     * to 4 byte boundaries horizontally, therefore, the
	     * LS two bits of source address do not change in the
	     * 32 source rect accesses.  So, instead of doing
	     * 32 XOR's with 3, we'll do one XOR at the top of
	     * the loop and use that address everywhere.
	     */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* loop through the rectangles in a row of rects */
	for (;
	     blockCount > 0;
	     blockCount--, pS += srcAddrFixup)
#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* loop through the rectangles in a row of rects */
	for (pS = (UInt8 *) (((UInt32) pS ) ^ 3);
	     blockCount > 0;
	     blockCount--,
		pS = (UInt8 *) ((((UInt32) pS ^ 3) + srcAddrFixup) ^ 3))
#endif /* cAlpacaPCFmt */
	{
		/* This is one of the cases where we can probably do
		 * a lot better than the WATCOM compiler on the PC.
		 * If there's a big problem, we'll look into it.
		 * On the x86, we can just load byte portions of
		 * registers directly.  This eliminates the ORs
		 * and two of the shifts.
		 */
	    inputWord0 = *pS;
	    pS -= sBpl;
	    inputWord0 <<= 8;

	    inputWord0 |= *pS;
	    pS -= sBpl;
	    inputWord0 <<= 8;

	    inputWord0 |= *pS;
	    pS -= sBpl;
	    inputWord0 <<= 8;

	    inputWord0 |= *pS;

		/* inputWord0 now has a rotated 1 x 4 column of data,
		 * so write it.  It's so simple.
		 */
	    ULONG_ACCESS(pD) = inputWord0;
	    pD += dBpl;
	}
    }
}



/* PRIVATE */
/**************************************************************************
 *  ip_rotate48Block270()
 *		rotates a row of 4bit/pixel or 8 bit/pixel rectangles
 *		by 270 degrees clockwise.
 *
 *	Input:
 *		pS =		Pointer to topmost halfword of source block
 *		sBpl =		Source bytes/line
 *		pD =		Pointer to topmost word of dest block
 *		sBpl =		Dest bytes/line
 *		rotationTable = Ptr to table used to spread source values.
 *				This table lets us do the rotation in two
 *				steps, the first spreads the values out so
 *				we can accumulate contiguous fragments of
 *				dest bytes in a source word.  The second
 *				assembles these fragments into real dest
 *				bytes or words.
 *		blockCount =	number of blocks to be rotated in a
 *				row of blocks across the source image.
 *		srcAddrFixup =	Number which is added to the final source
 *				address for one block to convert it to the
 *				initial source address for the next block
 *				in the row of blocks.
 *		depth =		Number of bits/pixel in src and dest images.
 **************************************************************************/
void  CDECL 
ip_rotate48Block270 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *rotationTable,
    Int32	blockCount,
    Int32	srcAddrFixup,
    Int32	depth)
{
    PROLOGUE;
    UInt32	inputWord0, inputWord1;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Do the correct code for the current image depth */
    switch (depth)
    {

	/* ********** 270 degree Rotate of 4 bit/pixel images ********* */

    case 4:

	    /* If this is a SPARC, we can just address the bytes
	     * without modifying the addresses.  If this is a PC,
	     * we need to XOR each address with 3 before using it.
	     * That's a pain, execution time went up from 0.71
	     * sec to 0.87 sec when we put it in.  So, we note that
	     * we're accessing source bits in a string of vertical
	     * bytes.  We know that all source images are padded
	     * to 4 byte boundaries horizontally, therefore, the
	     * LS two bits of source address do not change in the
	     * 32 source rect accesses.  So, instead of doing
	     * 32 XOR's with 3, we'll do one XOR at the top of
	     * the loop and use that address everywhere.
	     */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* loop through the rectangles in a row of rects */
	for (;
	     blockCount > 0;
	     blockCount--, pS += srcAddrFixup)
#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* loop through the rectangles in a row of rects */
	for (pS = (UInt8 *) (((UInt32) pS) ^ 3);
	     blockCount > 0;
	     blockCount--,
		pS = (UInt8 *) ((((UInt32) pS ^ 3) + srcAddrFixup) ^ 3))
#endif /* cAlpacaPCFmt */

	{

		/* PROCESS TOP 2x4 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord0 <<= 4;

	    inputWord0 |= rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord0 <<= 4;

	    inputWord0 |= rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord0 <<= 4;

	    inputWord0 |= rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;


		/* PROCESS BOTTOM 2x4 SQUARE */

	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord1 <<= 4;

	    inputWord1 |= rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord1 <<= 4;

	    inputWord1 |= rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord1 <<= 4;

	    inputWord1 |= rotationTable[*pS];	/* byte #1 */

		/* We used to test to see if there were any non-blank
		 * registers before writing.  It didn't speed things
		 * up and required the user to clear the image before
		 * calling this routine.  Now we just write the thing.
		 */
	    ULONG_ACCESS(pD) = (inputWord0 & 0xFFFF0000) |
			       (inputWord1 >> 16);
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = (inputWord0 << 16) |
			       (inputWord1 & 0x0000FFFF);
	    pD -= dBpl;

	}	/* end of rotating all rects in a row of rectangles */
	break;


	/* ************* 270 degree Rotate of 8 bit/pixel images *********** */


    default: /* depth == 8 bits/pixel */


	    /* If this is a SPARC, we can just address the bytes
	     * without modifying the addresses.  If this is a PC,
	     * we need to XOR each address with 3 before using it.
	     * That's a pain.  So, we note that
	     * we're accessing source bits in a string of vertical
	     * bytes.  We know that all source images are padded
	     * to 4 byte boundaries horizontally, therefore, the
	     * LS two bits of source address do not change in the
	     * 32 source rect accesses.  So, instead of doing
	     * 32 XOR's with 3, we'll do one XOR at the top of
	     * the loop and use that address everywhere.
	     */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* loop through the rectangles in a row of rects */
	for (;
	     blockCount > 0;
	     blockCount--, pS += srcAddrFixup)
#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* loop through the rectangles in a row of rects */
	for (pS = (UInt8 *) (((UInt32) pS ) ^ 3);
	     blockCount > 0;
	     blockCount--,
		pS = (UInt8 *) ((((UInt32) pS ^ 3) + srcAddrFixup) ^ 3))
#endif /* cAlpacaPCFmt */
	{
		/* This is one of the cases where we can probably do
		 * a lot better than the WATCOM compiler on the PC.
		 * If there's a big problem, we'll look into it.
		 * On the x86, we can just load byte portions of
		 * registers directly.  This eliminates the ORs
		 * and two of the shifts.
		 */
	    inputWord0 = *pS;
	    pS += sBpl;
	    inputWord0 <<= 8;

	    inputWord0 |= *pS;
	    pS += sBpl;
	    inputWord0 <<= 8;

	    inputWord0 |= *pS;
	    pS += sBpl;
	    inputWord0 <<= 8;

	    inputWord0 |= *pS;

		/* inputWord0 now has a rotated 1 x 4 column of data,
		 * so write it.  It's so simple.
		 */
	    ULONG_ACCESS(pD) = inputWord0;
	    pD -= dBpl;
	}
	break;
    }
}


/* This procedure is used by all implementations of 4 bit/pixel 90 degree
 * and 270 degree rotation.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_make4bppRotTable():  Makes the table we used to do 90 degree rotations
 *			    for 4 bit/pixel images.
 *
 *			This is a 256 x 32bit table.  
 *			For 4 bit/pixel images, is spreads bits [7:0]
 *			into 2 clumps - [19,18,17,16] and [3,2,1,0]
 *
 *	rotationTable:	pointer to returned rotation table.  This proc
 *			allocates the table and returns the address in
 *			this pointer.  It's up to the caller to free
 *			the table.
 *
 *	Return code:
 *	 ia_nomem		no memory available for table
 *	 ia_successful		good completion
 *
 *
 **************************************************************************/
Int32  CDECL 
ip_make4bppRotTable (
    IPState    *pIPState,
    UInt32    **pRotationTable)
{
PROLOGUE;
 register UInt32		 byte;
register UInt32 	 expByte;
UInt32			*localTab;

	/* allocate the table */
    *pRotationTable = (UInt32 *) CALLOC (256, sizeof(UInt32));

    if (*pRotationTable == NULL)
	return ia_nomem;

    localTab = (UInt32 *) *pRotationTable;

	/* Set each element of the table with the spread bits of its
	 * index. */
    for (byte = 0; byte < 256; byte++)
    {
	expByte =  ((byte & 0xF0) << (19-7) ) |	/* bits 7,6,5,4 */
		   ((byte & 0x0F));		/* bits 3,2,1,0 */
	localTab[byte] = expByte;
    }
    NORMAL_RETURN;
}


/*

 $Log: miorth48.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.3  1998/01/13 23:52:56  ddavies
 miorth48.c -> /product/ipcore/ipshared/src/RCS/miorth48.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1995/12/13 02:50:38  ejaquith
 miorth48.c -> /product/ipcore/ipshared/src/RCS/miorth48.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1994/02/13  00:59:01  ddavies
 * Initial revision
 *

 */

