/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miorthrt.c--includes
 *
 *		ip_fastRot90Setup()	setup parms for 90 degree rotate
 *		ip_rotateBlock90()	rotates a row into a column
 *		ip_fastRot270Setup()	setup parms for 270 degree rotate
 *		ip_rotateBlock270()	rotates a row into a column
 *		ip_reverseLine()	reverses the order of pixels in
 *					a single scan line.
 */


#include <stddef.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /* prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miorthrt.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* A reasonable way to rotate an image by 90 degrees is to divide the
 * image into rectangles, rotate the rectangles, and write them.  
 * The nice feature of this scheme is that we always write full 32-bit 
 * words.  The dimensions of the rectangles in pixels depend on the 
 * number of bits/pixel.
 *
 * For example, if we want to rotate a 1 bit/pixel image, the destination
 * rectangle are 32 bits wide, so the source rectangles are 32 bits tall.
 * The width of the source rectangles (height of the dest rectangles) is set
 * by the number of registers available for rotation.  Our compilers are
 * stupid enough to have trouble dealing with more than 8 registers, so
 * let's make the rectangles 8 columns wide.  Ok, so this is totally
 * bogus for the PC.  Well, too bad.  We tried a few different schemes and
 * the general plan of distributing the bits of one source thinger
 * (byte, halfword or word) among lots of destinations won every time.
 * In the PC assembly code, we'll probably use the AH, AL, BH, BL, CH, CL,
 * DH and DL registers to get a sufficient number of them.  It will rotate
 * 8 x 8 squares.  Never mind, in the C implemenation, we'll do source
 * rectangles 8 bits wide and 32 bits tall.
 *
 * How to rotate the squares?  One way is to test and branch on each bit,
 * conditionally ORing a mask into one of 8 dest registers.  Nice, but
 * this takes about three opcodes per bit.  Can't we do better?
 *
 * Another approach is to use a lookup table to spread the 8 bits of each
 * byte out into a 32-bit word.  We do this four times for four vertically
 * adjacent bytes.  We then put each of the 4-bit nibbles into its own
 * register.  Since we're moving bits in groups, we win.
 *
 * So how does this really work?  Here are the steps to rotate an 8x8
 * block.
 * 1. Read bottommost byte in block.
 * 2. Use lookup table to spread bits [7:0] into bits [28,24,20,16,12,8,4,0]
 * 3. Shift the accumulated register left by 1 (add to itself).
 * 4. Repeat steps 1-3 for the next three source bytes.  Each new source
 *    byte is closer to the top of the source image.  We also have to
 *    OR the new table lookup bits into the register.
 * 5. Leave the results in inputWord0.
 * 6. Repeat steps 1-4 for the next four source bytes, leaving the result
 *    in inputWord1.  We now have 8 bytes of info, but the most significant
 *    nibble of each byte is in inputWord0 and the least significant is
 *    in inputWord1.
 * 7. Set Dest = (inputWord0 & 0xF0F0F0F0) | ((inputWord1 >> 4) & (0x0F0F0F0F)
 *    This selects all the even numbered pixels (numbering left to right).
 * 8. Store DestB0 = Dest[31:24], DestB2 = Dest[23:16],
 *    DestB4 = Dest[15:8], DestB6 = Dest[7:0];
 * 9. Set Dest = ((inputWord0 << 4) * 0xF0F0F0F0) | (inputWord1 & 0x0F0F0F0F)
 *    This selects all the odd numbered pixels.
 * 10. Store DestB1 = Dest[31:24], DestB3 = Dest[23:16],
 *     DestB5 = Dest[15:8], DestB7 = Dest[7:0];
 * 11. Repeat steps 1-11 three more times, shifting the accumulating bytes
 *     in DestB0 through DestB7 left each time to make room for the
 *     new bytes.
 * 12. If any of them are non-zero, write DestB0 through DestB7 to the
 *     destination.
 *
 * This last step bring up an interesting point.  Most of the images we
 * deal with are white.  Shouldn't there be a check for all-0 source bytes?
 * Well, on the SPARC, we'd save a lookup (2 cycles) and an OR (1 cycle).
 * The test costs a tst (1 cycle) and a branch (1 cycle if the compiler
 * can use the interlock).  We'll ignore it for now.  After it works, we'll
 * look at the compiled code and see if it looks worthwhile.
 *
 * In what order do we address the source and destination?  The 90 degree
 * rotation is defined to be clockwise, so the lower left corner becomes
 * the upper left corner.  The first block rotated will be in the lower
 * left corner and will be stored in the upper left corner of the destination
 * image.  To take advantage of the processor's cache, the second block
 * read will be just to the right of the first block read and will be stored
 * just below the first block stored.  Both the SS2 and the 486 have
 * write through caches that don't fetch locations being written into the
 * cache implicitly, so the order in which we write dest words doesn't matter.
 *
 * What about images whose widths are not even multiples of 8 or whose
 * heights are not even mulitples of 32?  Look at the height first.  We
 * want the bottommost line of the source image to become the leftmost bits
 * of the leftmost words of the destination image.  That means we should
 * start the bottom rectangle exactly on the bottommost line of the
 * source image.  The topmost rectangles in the source image may well include
 * some of the upper frame.  Big deal, that upper frame will become the
 * right frame in the destination image.
 *
 * What about the right edge of the source image?  The source image will
 * likely end somewhere inside the rightmost 8 bit source words.  Again,
 * big deal.  The extra 0's to the right of the real image will appear
 * as part of the lower frame in the destination image.
 *
 * In both cases, we can safely rotate entire rectangles without worrying
 * about edge conditions.
 *
 * 270 degree rotation (90 degrees counterclockwise) is almost the same
 * as 90 degree rotation, except
 *	1. We scan across the source image from left->right (fast scan) and
 *	   top->bottom (slow scan) instead of left->right, bottom->top.
 *	2. Each block is rotated counterclockwise instead of clockwise.
 *
 *********************
 *
 * That was all fine for binary images.  Worked great.  Now, we have to
 * generalize to 2 bit/pixel images.
 *
 * We'll use a variant on the scheme used above.  We're still rotating
 * rectangles.  However, each source byte contains 4 pixels, not 8.  Since
 * our source rectangles will be 4 pixels wide, our dest rectangles will
 * be 4 lines tall.
 *
 * Since there are four dest registers, the intermediate register holds
 * four pixels, each 8 bits wide.  Each source read returns two bits in
 * each pixel, so in four reads, we have 4 dest bytes.  These are
 * distributed among the four destination registers, one byte to each
 * destination register.
 *
 * That process is repeated four times, filling each dest reg with four
 * bytes.  These bytes are written into a column of dest locations.
 */

/* dimensions of the rectangles we rotate */
/* 1 bit/pixel images */
#define	cSrcRectWidth1		8
#define cLogSrcRectWidth1	3
#define cSrcRectWidthInBytes1 (cSrcRectWidth1>>3)
#define cSrcRectHeight1		32
#define cLogSrcRectHeight1	5

/* 2 bit/pixel images */
#define	cSrcRectWidth2		4	/* in pixels */
#define cLogSrcRectWidth2	2
#define cSrcRectWidthInBytes2 ((cSrcRectWidth2*2)>>3)
#define cSrcRectHeight2		16
#define cLogSrcRectHeight2	4

/* DO NOT CHANGE THESE WITHOUT CHANGING THE COPIES IN miorth48.c AND
 * ssorthst.c!!!
 */

/* 4 bit/pixel images */
#define	cSrcRectWidth4		2	/* in pixels */
#define cLogSrcRectWidth4	1
#define cSrcRectWidthInBytes4 ((cSrcRectWidth4*4)>>3)
#define cSrcRectHeight4		8
#define cLogSrcRectHeight4	3

/* 8 bit/pixel images */
#define	cSrcRectWidth8		1	/* in pixels */
#define cLogSrcRectWidth8	0
#define cSrcRectWidthInBytes8 ((cSrcRectWidth8*8)>>3)
#define cSrcRectHeight8		4
#define cLogSrcRectHeight8	2

/* PRIVATE */
/**************************************************************************
 *	ip_fastRot90Setup
 *		Performs the following actions required before doing
 *		90 degree rotation:
 *		1. Allocate, initialize and return address of rotation table
 *		   (if any).
 *		2. Return number of rectangles in the width of the image.
 *		3. Return the number of rectangles that fit into the height
 *		   of the image.
 *		4. Return the value used to go from the end of one
 *		   source rectangle to the beginning of the next when
 *		   processing a row of rectangles.
 *		5. Adjust the initial value of the source address.  This
 *		   will be somewhere on the rectangle in the lower
 *		   left corner of the image depending on the implementation.
 *		6. If necessary, adjust value of destination address.
 *		   This happens when the assembly code uses weird
 *		   tricks with sign bits that leak over into into the
 *		   order in which the dest image blocks are written.
 *		7. Return the displacement that we subtract from the source
 *		   address when moving up one rectangle in the image.  This
 *		   is used to move between rows of rectangles.
 *	   The following steps apply to rotation in place
 *		8. Return the number of source lines in a source rectangle.
 *		9. Return the number of source blocks that fit into the
 *		   destination buffer.
 *
 *	Input:
 *	pSrcAddr:		ptr to initial source address INSIDE frame
 *	pDestAddr:		ptr to initial dest image INSIDE frame
 *	sWidth:			width of source image without frame
 *	sHeight:		height of source image without frame
 *	sDepth:			depth of source image
 *	sBpl:			source image bytes/line
 *	pRotationTable:		pointer to ptr to first element in rotation
 *				table.
 *	pWidthInBlocks:		ptr to width of image in units of rotation
 *				rectangles.
 *	pHeightInBlocks:	ptr to height of image in units of rotation
 *				rectangles.
 *	pInterBlockAddrIncr:	ptr to value we add to the source address
 *				to move from one source block to the next
 *				within one row of source blocks
 *	pInterRowAddrIncr:	ptr to value we subtract from the source addr
 *				to move between rows of rectangles
 *	pSrcBlockHeight:	ptr to height of source blocks in rows;
 *				used only for in-place rotation
 *	Returns:
 *	ia_nomem:	Ran out of memory for rotation table
 *	ia_successful:	Setup was successful.
 *
 **************************************************************************/
Int32 CDECL 
ip_fastRot90Setup (
    IPState    *pIPState,
    UInt8     **pSrcAddr,
    Int32	sWidth,
    Int32	sHeight,
    Int32	sDepth,
    Int32	sBpl,
    UInt8     **pDstAddr,
    Int32	dBpl,
    UInt32    **pRotationTable,
    Int32      *pWidthInBlocks,
    Int32      *pHeightInBlocks,
    Int32      *pInterBlockAddrIncr,
    Int32      *pInterRowAddrIncr,
    Int32      *pSrcBlockHeight)
{
    PROLOGUE;
    

	/* keep the compiler happy */
    pDstAddr = pDstAddr;
    dBpl = dBpl;

	/* Compute the table we'll use to expand bytes into 32-bit things */
    ERR_PASS(ip_makeRotTable (pIPState, pRotationTable, sDepth));

	/* Switch on the depth of the image, this changes the returned
	 * parameters. */
    switch (sDepth)
    {
    case 1:
	    /* set source image pointer to lower left corner of rectangle in
	     * lower left corner of the image.  We need the lower left
	     * corner because the we want to shift left, not right.
	     * This means we need to access the bottom line
	     * first so it ends up on the rightmost (most shifted)
	     * side of the dest registers. */
	*pSrcAddr += (sHeight-1) * sBpl;
    
	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth1-1) >> cLogSrcRectWidth1;
	*pHeightInBlocks = (sHeight+cSrcRectHeight1-1) >> cLogSrcRectHeight1;
	*pSrcBlockHeight = cSrcRectHeight1;

	    /* the loop structure of the assembly code subtracts
	     * rectHeight * srcPitch,
	     * we must add that much back on, then move over 1 rect. */
	*pInterBlockAddrIncr = (sBpl << cLogSrcRectHeight1) - sBpl
			    + cSrcRectWidthInBytes1;

	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight1);
	break;

    case 2:
	    /* set source image pointer to lower left corner of rectangle in
	     * lower left corner of the image.  We need the lower left
	     * corner because the we want to shift left, not right.
	     * This means we need to access the bottom line
	     * first so it ends up on the rightmost (most shifted)
	     * side of the dest registers. */
	*pSrcAddr += (sHeight-1) * sBpl;

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth2-1) >> cLogSrcRectWidth2;
	*pHeightInBlocks = (sHeight+cSrcRectHeight2-1) >> cLogSrcRectHeight2;
	*pSrcBlockHeight = cSrcRectHeight2;

	    /* the loop structure of the assembly code subtracts
	     * rectHeight * srcPitch, we must add that much back on,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = (sBpl << cLogSrcRectHeight2) - sBpl +
				cSrcRectWidthInBytes2;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight2);
	break;

	/* The setup for the 4bpp and 8bpp cases happens on behalf of
	 * primitives contained in miorth48.c
	 */
    case 4:
	    /* set source image pointer to lower left corner of rectangle in
	     * lower left corner of the image.  We need the lower left
	     * corner because the we want to shift left, not right.
	     * This means we need to access the bottom line
	     * first so it ends up on the rightmost (most shifted)
	     * side of the dest registers. */
	*pSrcAddr += (sHeight-1) * sBpl;

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth4-1) >> cLogSrcRectWidth4;
	*pHeightInBlocks = (sHeight+cSrcRectHeight4-1) >> cLogSrcRectHeight4;
	*pSrcBlockHeight = cSrcRectHeight4;

	    /* the loop structure of the assembly code subtracts
	     * rectHeight * srcPitch, we must add that much back on,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = (sBpl << cLogSrcRectHeight4) - sBpl +
				cSrcRectWidthInBytes4;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight4);
	break;

    default:	/* case 8 */
	    /* set source image pointer to lower left corner of rectangle in
	     * lower left corner of the image.  We need the lower left
	     * corner because the we want to shift left, not right.
	     * This means we need to access the bottom line
	     * first so it ends up on the rightmost (most shifted)
	     * side of the dest registers. */
	*pSrcAddr += (sHeight-1) * sBpl;

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth8-1) >> cLogSrcRectWidth8;
	*pHeightInBlocks = (sHeight+cSrcRectHeight8-1) >> cLogSrcRectHeight8;
	*pSrcBlockHeight = cSrcRectHeight8;

	    /* the loop structure of the assembly code subtracts
	     * rectHeight * srcPitch, we must add that much back on,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = (sBpl << cLogSrcRectHeight8) - sBpl +
				cSrcRectWidthInBytes8;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight8);
	break;
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_rotateBlock90()
 *		rotates a row of rectangles by 90 degrees clockwise.
 *
 *	Input:
 *		pS =		Pointer to topmost halfword of source block
 *		sBpl =		Source bytes/line
 *		pD =		Pointer to topmost word of dest block
 *		sBpl =		Dest bytes/line
 *		rotationTable = Ptr to table used to spread source values.
 *				This table lets us do the rotation in two
 *				steps, the first spreads the values out so
 *				we can accumulate contiguous fragments of
 *				dest bytes in a source word.  The second
 *				assembles these fragments into real dest
 *				bytes or words.
 *		blockCount =	number of blocks to be rotated in a
 *				row of blocks across the source image.
 *		srcAddrFixup =	Number which is added to the final source
 *				address for one block to convert it to the
 *				initial source address for the next block
 *				in the row of blocks.
 *		depth =		Number of bits/pixel in src and dest images.
 **************************************************************************/
void  CDECL 
ip_rotateBlock90 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *rotationTable,
    Int32	blockCount,
    Int32	srcAddrFixup,
    Int32	depth)
{
    PROLOGUE;
    UInt32	inputWord0, inputWord1, dest, dest1, destMask;
    UInt32	destB0;		/* topmost destination reg */
    UInt32	destB1, destB2, destB3, destB4, destB5, destB6;
    UInt32	destB7;		/* bottom destination reg */

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Do the correct code for the current image depth */
    switch (depth)
    {

	/* ********** 90 degree Rotate of 1 bit/pixel images ********* */

    case 1:

	    /* If this is a SPARC, we can just address the bytes
	     * without modifying the addresses.  If this is a PC,
	     * we need to XOR each address with 3 before using it.
	     * That's a pain, execution time went up from 0.71
	     * sec to 0.87 sec when we put it in.  So, we note that
	     * we're accessing source bits in a string of vertical
	     * bytes.  We know that all source images are padded
	     * to 4 byte boundaries horizontally, therefore, the
	     * LS two bits of source address do not change in the
	     * 32 source rect accesses.  So, instead of doing
	     * 32 XOR's with 3, we'll do one XOR at the top of
	     * the loop and use that address everywhere.
	     */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* loop through the rectangles in a row of rects */
	for (;
	     blockCount > 0;
	     blockCount--, pS += srcAddrFixup)
#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* loop through the rectangles in a row of rects */
	for (pS = (UInt8 *) (((UInt32) pS) ^ 3);
	     blockCount > 0;
	     blockCount--,
		pS = (UInt8 *) ((((UInt32) pS ^ 3) + srcAddrFixup) ^ 3))
#endif /* cAlpacaPCFmt */

	{
		/* PROCESS BOTTOM 8x8 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* Repeat this for inputWord 1 with next 4 src bytes */
	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* We now have bits [31:24] of each dest word in
		 * inputWord0 and bits [27:24] of each dest word in
		 * inputWord1.  Set dest equal to the 4 even bytes.
		 */
	    destMask = 0xFF000000;
	    dest = (inputWord0 & 0xF0F0F0F0) |
		   ((inputWord1 >> 4) & 0x0F0F0F0F);

		/* Store these bits into the even pixels */
	    destB0 = dest & destMask;
	    dest <<= 8;
	    destB2 = dest & destMask;
	    dest <<= 8;
	    destB4 = dest & destMask;
	    dest <<= 8;
	    destB6 = dest & destMask;

		/* Do the same for the odd pixels */
	    dest = ((inputWord0 << 4) & 0xF0F0F0F0) |
		   (inputWord1 & 0x0F0F0F0F);
	    destB1 = dest & destMask;
	    dest <<= 8;
	    destB3 = dest & destMask;
	    dest <<= 8;
	    destB5 = dest & destMask;
	    dest <<= 8;
	    destB7 = dest & destMask;


		/* PROCESS NEXT-TO-BOTTOM 8x8 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* Repeat this for inputWord 1 with next 4 src bytes */
	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* We now have bits [23:20] of each dest word in
		 * inputWord0 and bits [19:16] of each dest word in
		 * inputWord1.  Set dest equal to the 4 even bytes.
		 */
	    destMask = 0x00FF0000;
	    dest = dest1 = (inputWord0 & 0xF0F0F0F0) |
			    ((inputWord1 >> 4) & 0x0F0F0F0F);

		/* Store these bits into the even pixels */
	    dest >>= 8;	/* rotate into position */
	    destB0 |= dest & destMask;
	    destB2 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB4 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB6 |= dest1 & destMask;

		/* Do the same for the odd pixels */
	    dest = dest1 = ((inputWord0 << 4) & 0xF0F0F0F0) |
			    (inputWord1 & 0x0F0F0F0F);
	    dest >>= 8;	/* rotate into position */
	    destB1 |= dest & destMask;
	    destB3 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB5 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB7 |= dest1 & destMask;

		/* PROCESS NEXT-TO-TOP 8x8 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* Repeat this for inputWord 1 with next 4 src bytes */
	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* We now have bits [15:12] of each dest word in
		 * inputWord0 and bits [11:8] of each dest byte in
		 * inputWord1.  Set dest equal to the 4 even bytes.
		 */
	    destMask = 0x0000FF00;
	    dest = dest1 = (inputWord0 & 0xF0F0F0F0) |
			    ((inputWord1 >> 4) & 0x0F0F0F0F);

		/* Store these bits into the even pixels */
	    dest >>= 16;	/* rotate into position */
	    destB0 |= dest & destMask;
	    dest <<= 8;
	    destB2 |= dest & destMask;
	    destB4 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB6 |= dest1 & destMask;

		/* Do the same for the odd pixels */
	    dest = dest1 = ((inputWord0 << 4) & 0xF0F0F0F0) |
			    (inputWord1 & 0x0F0F0F0F);
	    dest >>= 16;	/* rotate into position */
	    destB1 |= dest & destMask;
	    dest <<= 8;
	    destB3 |= dest & destMask;
	    destB5 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB7 |= dest1 & destMask;


		/* PROCESS TOP 8x8 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* Repeat this for inputWord 1 with next 4 src bytes */
	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #4 */

		/* We now have bits [7:4] of each dest word in
		 * inputWord0 and bits [3:0] of each dest word in
		 * inputWord1.  Set dest equal to the 4 even bytes.
		 */
	    destMask = 0x000000FF;
	    dest = (inputWord0 & 0xF0F0F0F0) |
		    ((inputWord1 >> 4) & 0x0F0F0F0F);

		/* Store these bits into the even pixels */
	    destB6 |= dest & destMask;
	    dest >>= 8;
	    destB4 |= dest & destMask;
	    dest >>= 8;
	    destB2 |= dest & destMask;
	    dest >>= 8;
	    destB0 |= dest & destMask;

		/* Do the same for the odd pixels */
	    dest = ((inputWord0 << 4) & 0xF0F0F0F0) |
		    (inputWord1 & 0x0F0F0F0F);
	    destB7 |= dest & destMask;
	    dest >>= 8;
	    destB5 |= dest & destMask;
	    dest >>= 8;
	    destB3 |= dest & destMask;
	    dest >>= 8;
	    destB1 |= dest & destMask;


		/* We used to test to see if there were any non-blank
		 * registers before writing.  It didn't speed things
		 * up and required the user to clear the image before
		 * calling this routine.  Now we just write the thing.
		 */
	    ULONG_ACCESS(pD) = destB0;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB1;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB2;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB3;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB4;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB5;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB6;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB7;
	    pD += dBpl;

	}	/* end of rotating all rects in a row of rectangles */
	break;


	/* ************* 90 degree Rotate of 2 bit/pixel images *********** */


    default: /* depth == 2 bits/pixel */

	    /* If this is a SPARC, we can just address the bytes
	     * without modifying the addresses.  If this is a PC,
	     * we need to XOR each address with 3 before using it.
	     * That's a pain.  So, we note that
	     * we're accessing source bits in a string of vertical
	     * bytes.  We know that all source images are padded
	     * to 4 byte boundaries horizontally, therefore, the
	     * LS two bits of source address do not change in the
	     * 32 source rect accesses.  So, instead of doing
	     * 32 XOR's with 3, we'll do one XOR at the top of
	     * the loop and use that address everywhere.
	     */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* loop through the rectangles in a row of rects */
	for (;
	     blockCount > 0;
	     blockCount--, pS += srcAddrFixup)
#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* loop through the rectangles in a row of rects */
	for (pS = (UInt8 *) (((UInt32) pS ) ^ 3);
	     blockCount > 0;
	     blockCount--,
		pS = (UInt8 *) ((((UInt32) pS ^ 3) + srcAddrFixup) ^ 3))
#endif /* cAlpacaPCFmt */

	{
		/* PROCESS BOTTOM 4x4 SQUARE */

	    dest = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* We now have bits [31:24] of each dest word in
		 * dest.  Distribute these among the dest regs.
		 */
	    destMask = 0xFF000000;

		/* Store these bits into the MS bytes */
	    destB0 = dest & destMask;
	    dest <<= 8;
	    destB1 = dest & destMask;
	    dest <<= 8;
	    destB2 = dest & destMask;
	    dest <<= 8;
	    destB3 = dest & destMask;


		/* PROCESS NEXT-TO-BOTTOM 4x4 SQUARE */

	    dest = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* We now have bits [23:16] of each dest word in
		 * dest.  Place into the dest regs.
		 */
	    destMask = 0x00FF0000;
	    dest1 = dest;

		/* Store these bits into bits [23:16] of dest regs. */
	    dest >>= 8;	/* rotate into position */
	    destB0 |= dest & destMask;
	    destB1 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB2 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB3 |= dest1 & destMask;


		/* PROCESS NEXT-TO-TOP 4x4 SQUARE */

	    dest = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #4 */
	    pS -= sBpl;

		/* We now have bits [15:8] of each dest byte in
		 * dest.  Place into the dest regs.
		 */
	    destMask = 0x0000FF00;
	    dest1 = dest;

		/* Store these bits into bits [15:8] of dest regs. */
	    dest >>= 16;	/* rotate into position */
	    destB0 |= dest & destMask;
	    dest <<= 8;
	    destB1 |= dest & destMask;
	    destB2 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB3 |= dest1 & destMask;


		/* PROCESS TOP 4x4 SQUARE */

	    dest = rotationTable[*pS];	/* byte #1 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #2 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #3 */
	    pS -= sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #4 */

		/* We now have bits [7:0] of each dest byte in
		 * dest.  Place into the dest regs.
		 */
	    destMask = 0x000000FF;

		/* Store these bits into bits [7:0] of dest regs. */
	    destB3 |= dest & destMask;
	    dest >>= 8;
	    destB2 |= dest & destMask;
	    dest >>= 8;
	    destB1 |= dest & destMask;
	    dest >>= 8;
	    destB0 |= dest & destMask;


		/* We used to test to see if there were any non-blank
		 * registers before writing.  It didn't speed things
		 * up and required the user to clear the image before
		 * calling this routine.  Now we just write the thing.
		 */
	    ULONG_ACCESS(pD) = destB0;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB1;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB2;
	    pD += dBpl;
	    ULONG_ACCESS(pD) = destB3;
	    pD += dBpl;

	}	/* end of rotating all rects in a row of rectangles */
    }
}

/* PRIVATE */
/**************************************************************************
 *	ip_fastRot270Setup
 *		Performs the following actions required before doing
 *		270 degree rotation:
 *		1. Allocate, initialize and return address of rotation table
 *		   (if any).
 *		2. Return number of rectangles in the width of the image.
 *		3. Return the number of rectangles that fit into the height
 *		   of the image.
 *		4. Return the value used to go from the end of one
 *		   source rectangle to the beginning of the next when
 *		   processing a row of rectangles.
 *		5. Adjust the initial value of the source address.  This
 *		   will be somewhere on the rectangle in the lower
 *		   left corner of the image depending on the implementation.
 *		6. Adjust value of destination address.  It will be somewhere
 *		   near the bottom right corner of the image depending on the
 *		   implementation.
 *		7. Return the displacement that we subtract from the source
 *		   address when moving up one rectangle in the image.  This
 *		   is used to move between rows of rectangles.
 *	   The following steps apply to rotation in place
 *		8. Return the number of source lines in a source rectangle.
 *
 *	Input:
 *	pSrcAddr:		ptr to initial source address INSIDE frame
 *	sWidth:			width of source image without frame
 *	sHeight:		height of source image without frame
 *	sDepth:			depth of source image
 *	sBpl:			source image bytes/line
 *	pDestAddr:		ptr to initial dest image INSIDE frame
 *	dWidth:			width of dest image or transfer buffer
 *				in pixels
 *	dBpl:			bytes/line of dest image or transfer buffer
 *	pRotationTable:		pointer to ptr to first element in rotation
 *				table.
 *	pWidthInBlocks:		ptr to width of image in units of rotation
 *				rectangles.
 *	pHeightInBlocks:	ptr to height of image in units of rotation
 *				rectangles.
 *	pInterBlockAddrIncr:	ptr to value we add to the source address
 *				to move from one source block to the next
 *				within one row of source blocks
 *	pInterRowAddrIncr:	ptr to value we subtract from the source addr
 *				to move between rows of rectangles
 *	pSrcBlockHeight:	ptr to height of source blocks in rows;
 *				used only for in-place rotation
 *	Returns:
 *	ia_nomem:	Ran out of memory for rotation table
 *	ia_successful:	Setup was successful.
 *
 **************************************************************************/
Int32 CDECL 
ip_fastRot270Setup (
    IPState    *pIPState,
    UInt8     **pSrcAddr,
    Int32	sWidth,
    Int32	sHeight,
    Int32	sDepth,
    Int32	sBpl,
    UInt8     **pDstAddr,
    Int32	dWidth,
    Int32	dBpl,
    UInt32    **pRotationTable,
    Int32      *pWidthInBlocks,
    Int32      *pHeightInBlocks,
    Int32      *pInterBlockAddrIncr,
    Int32      *pInterRowAddrIncr,
    Int32      *pSrcBlockHeight)
{
    PROLOGUE;

	/* Compute the table we'll use to expand bytes into 32-bit things */
    ERR_PASS(ip_makeRotTable (pIPState, pRotationTable, sDepth));

	/* Switch on the depth of the image, this changes the returned
	 * parameters. */
    switch (sDepth)
    {
    case 1:

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth1-1) >> cLogSrcRectWidth1;
	*pHeightInBlocks = (sHeight+cSrcRectHeight1-1) >> cLogSrcRectHeight1;
	*pSrcBlockHeight = cSrcRectHeight1;

	    /* Set the source so it points to the upper left corner of
	     * the rectangle in the lower left of the image.  The rects
	     * are aligned with the top line of the image.  We do this
	     * weird thing so the in-place rotations can make room for
	     * the dest at the bottom of the image, which allows them
	     * to use regular old i_rasterOp to stretch the dest bits.
	     */
	*pSrcAddr += ((*pHeightInBlocks - 1) * cSrcRectHeight1) * sBpl;

	    /* Set dest image pointer to the last word
	     * in the lower right corner of image */
	*pDstAddr += (sWidth-1) * dBpl + ((dWidth-1)/32) * 4;

	    /* the loop structure of the assembly code adds
	     * rectHeight * srcPitch,
	     * we must subtract that much back off, then move over 1 rect. */
	*pInterBlockAddrIncr = -(sBpl << cLogSrcRectHeight1) + sBpl
			    + cSrcRectWidthInBytes1;

	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight1);
	break;

    case 2:

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth2-1) >> cLogSrcRectWidth2;
	*pHeightInBlocks = (sHeight+cSrcRectHeight2-1) >> cLogSrcRectHeight2;
	*pSrcBlockHeight = cSrcRectHeight2;

	    /* Set the source so it points to the upper left corner of
	     * the rectangle in the lower left of the image.  The rects
	     * are aligned with the top line of the image.  We do this
	     * weird thing so the in-place rotations can make room for
	     * the dest at the bottom of the image, which allows them
	     * to use regular old i_rasterOp to stretch the dest bits.
	     */
	*pSrcAddr += ((*pHeightInBlocks - 1) * cSrcRectHeight2) * sBpl;

	    /* Set dest image pointer to the last word
	     * in the lower right corner of image */
	*pDstAddr += (sWidth-1) * dBpl + ((dWidth-1)/16) * 4;

	    /* the loop structure of the assembly code adds
	     * rectHeight * srcPitch, we must subtract that much back off,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = -(sBpl << cLogSrcRectHeight2) + sBpl +
				cSrcRectWidthInBytes2;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight2);
	break;

	/* The setup for the 4bpp and 8bpp cases happens on behalf of
	 * primitives contained in miorth48.c
	 */
    case 4:

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth4-1) >> cLogSrcRectWidth4;
	*pHeightInBlocks = (sHeight+cSrcRectHeight4-1) >> cLogSrcRectHeight4;
	*pSrcBlockHeight = cSrcRectHeight4;

	    /* Set the source so it points to the upper left corner of
	     * the rectangle in the lower left of the image.  The rects
	     * are aligned with the top line of the image.  We do this
	     * weird thing so the in-place rotations can make room for
	     * the dest at the bottom of the image, which allows them
	     * to use regular old i_rasterOp to stretch the dest bits.
	     */
	*pSrcAddr += ((*pHeightInBlocks - 1) * cSrcRectHeight4) * sBpl;

	    /* Set dest image pointer to the last word
	     * in the lower right corner of image */
	*pDstAddr += (sWidth-1) * dBpl + ((dWidth-1)/8) * 4;

	    /* the loop structure of the assembly code adds
	     * rectHeight * srcPitch, we must subtract that much back off,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = -(sBpl << cLogSrcRectHeight4) + sBpl +
				cSrcRectWidthInBytes4;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight4);
	break;

    default:	/* case 8 */

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth8-1) >> cLogSrcRectWidth8;
	*pHeightInBlocks = (sHeight+cSrcRectHeight8-1) >> cLogSrcRectHeight8;
	*pSrcBlockHeight = cSrcRectHeight8;

	    /* Set the source so it points to the upper left corner of
	     * the rectangle in the lower left of the image.  The rects
	     * are aligned with the top line of the image.  We do this
	     * weird thing so the in-place rotations can make room for
	     * the dest at the bottom of the image, which allows them
	     * to use regular old i_rasterOp to stretch the dest bits.
	     */
	*pSrcAddr += ((*pHeightInBlocks - 1) * cSrcRectHeight8) * sBpl;

	    /* Set dest image pointer to the last word
	     * in the lower right corner of image */
	*pDstAddr += (sWidth-1) * dBpl + ((dWidth-1)/4) * 4;

	    /* the loop structure of the assembly code adds
	     * rectHeight * srcPitch, we must subtract that much back off,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = -(sBpl << cLogSrcRectHeight8) + sBpl +
				cSrcRectWidthInBytes8;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight8);
	break;
    }

    NORMAL_RETURN;
}


/**************************************************************************
 *  ip_rotateBlock270()
 *		rotates a row of rectangles by 270 degrees clockwise.
 *
 *	Input:
 *		pS =		Pointer to bottommost halfword of source block
 *		sBpl =		Source bytes/line
 *		pD =		Pointer to topmost word of dest block
 *		sBpl =		Dest bytes/line
 *		rotationTable = Ptr to table used to spread source values.
 *				This table lets us do the rotation in two
 *				steps, the first spreads the values out so
 *				we can accumulate contiguous fragments of
 *				dest bytes in a source word.  The second
 *				assembles these fragments into real dest
 *				bytes or words.
 *		blockCount =	number of blocks to be rotated in a
 *				row of blocks across the source image.
 *		srcAddrFixup =	Number which is added to the final source
 *				address for one block to convert it to the
 *				initial source address for the next block
 *				in the row of blocks.
 *		depth =		Number of bits/pixel in src and dest images.
 **************************************************************************/
void  CDECL 
ip_rotateBlock270 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *rotationTable,
    Int32	blockCount,
    Int32	srcAddrFixup,
    Int32	depth)
{
    PROLOGUE;
    UInt32	inputWord0, inputWord1, dest, dest1, destMask;
    UInt32	destB0;		/* topmost destination reg */
    UInt32	destB1, destB2, destB3, destB4, destB5, destB6;
    UInt32	destB7;		/* bottom destination reg */

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Do the correct code for the current image depth */
    switch (depth) {

	/* ********** 270 degree Rotate of 1 bit/pixel images ********* */

    case 1:

	    /* If this is a SPARC, we can just address the bytes
	     * without modifying the addresses.  If this is a PC,
	     * we need to XOR each address with 3 before using it.
	     * That's a pain, execution time went up from 0.71
	     * sec to 0.87 sec when we put it in.  So, we note that
	     * we're accessing source bits in a string of vertical
	     * bytes.  We know that all source images are padded
	     * to 4 byte boundaries horizontally, therefore, the
	     * LS two bits of source address do not change in the
	     * 32 source rect accesses.  So, instead of doing
	     * 32 XOR's with 3, we'll do one XOR at the top of
	     * the loop and use that address everywhere.
	     */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* loop through the rectangles in a row of rects */
	for (;
	     blockCount > 0;
	     blockCount--, pS += srcAddrFixup)
#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* loop through the rectangles in a row of rects */
	for (pS = (UInt8 *) (((UInt32) pS ) ^ 3);
	     blockCount > 0;
	     blockCount--,
		pS = (UInt8 *) ((((UInt32) pS ^ 3) +
						srcAddrFixup) ^ 3))
#endif /* cAlpacaPCFmt */

	{
		/* PROCESS TOP 8x8 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* Repeat this for inputWord 1 with next 4 src bytes */
	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* We now have bits [31:28] of each dest word in
		 * inputWord0 and bits [27:24] of each dest word in
		 * inputWord1.  Set dest equal to the 4 even bytes.
		 */
	    destMask = 0xFF000000;
	    dest = (inputWord0 & 0xF0F0F0F0) |
		   ((inputWord1 >> 4) & 0x0F0F0F0F);

		/* Store these bits into the even pixels */
	    destB0 = dest & destMask;
	    dest <<= 8;
	    destB2 = dest & destMask;
	    dest <<= 8;
	    destB4 = dest & destMask;
	    dest <<= 8;
	    destB6 = dest & destMask;

		/* Do the same for the odd pixels */
	    dest = ((inputWord0 << 4) & 0xF0F0F0F0) |
		   (inputWord1 & 0x0F0F0F0F);
	    destB1 = dest & destMask;
	    dest <<= 8;
	    destB3 = dest & destMask;
	    dest <<= 8;
	    destB5 = dest & destMask;
	    dest <<= 8;
	    destB7 = dest & destMask;


		/* PROCESS NEXT-TO-TOP 8x8 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* Repeat this for inputWord 1 with next 4 src bytes */
	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* We now have bits [23:20] of each dest word in
		 * inputWord0 and bits [19:16] of each dest word in
		 * inputWord1.  Set dest equal to the 4 even bytes.
		 */
	    destMask = 0x00FF0000;
	    dest = dest1 = (inputWord0 & 0xF0F0F0F0) |
			    ((inputWord1 >> 4) & 0x0F0F0F0F);

		/* Store these bits into the even pixels */
	    dest >>= 8;	/* rotate into position */
	    destB0 |= dest & destMask;
	    destB2 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB4 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB6 |= dest1 & destMask;

		/* Do the same for the odd pixels */
	    dest = dest1 = ((inputWord0 << 4) & 0xF0F0F0F0) |
			    (inputWord1 & 0x0F0F0F0F);
	    dest >>= 8;	/* rotate into position */
	    destB1 |= dest & destMask;
	    destB3 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB5 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB7 |= dest1 & destMask;

		/* PROCESS NEXT-TO-BOTTOM 8x8 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* Repeat this for inputWord 1 with next 4 src bytes */
	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* We now have bits [15:12] of each dest word in
		 * inputWord0 and bits [11:8] of each dest word in
		 * inputWord1.  Set dest equal to the 4 even bytes.
		 */
	    destMask = 0x0000FF00;
	    dest = dest1 = (inputWord0 & 0xF0F0F0F0) |
			    ((inputWord1 >> 4) & 0x0F0F0F0F);

		/* Store these bits into the even pixels */
	    dest >>= 16;	/* rotate into position */
	    destB0 |= dest & destMask;
	    dest <<= 8;
	    destB2 |= dest & destMask;
	    destB4 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB6 |= dest1 & destMask;

		/* Do the same for the odd pixels */
	    dest = dest1 = ((inputWord0 << 4) & 0xF0F0F0F0) |
			    (inputWord1 & 0x0F0F0F0F);
	    dest >>= 16;	/* rotate into position */
	    destB1 |= dest & destMask;
	    dest <<= 8;
	    destB3 |= dest & destMask;
	    destB5 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB7 |= dest1 & destMask;


		/* PROCESS BOTTOM 8x8 SQUARE */

	    inputWord0 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    inputWord0 += inputWord0;

	    inputWord0 |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* Repeat this for inputWord 1 with next 4 src bytes */
	    inputWord1 = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    inputWord1 += inputWord1;

	    inputWord1 |= rotationTable[*pS];	/* byte #4 */

		/* We now have bits [7:4] of each dest word in
		 * inputWord0 and bits [3:0] of each dest word in
		 * inputWord1.  Set dest equal to the 4 even bytes.
		 */
	    destMask = 0x000000FF;
	    dest = (inputWord0 & 0xF0F0F0F0) |
		    ((inputWord1 >> 4) & 0x0F0F0F0F);

		/* Store these bits into the even pixels */
	    destB6 |= dest & destMask;
	    dest >>= 8;
	    destB4 |= dest & destMask;
	    dest >>= 8;
	    destB2 |= dest & destMask;
	    dest >>= 8;
	    destB0 |= dest & destMask;

		/* Do the same for the odd pixels */
	    dest = ((inputWord0 << 4) & 0xF0F0F0F0) |
		    (inputWord1 & 0x0F0F0F0F);
	    destB7 |= dest & destMask;
	    dest >>= 8;
	    destB5 |= dest & destMask;
	    dest >>= 8;
	    destB3 |= dest & destMask;
	    dest >>= 8;
	    destB1 |= dest & destMask;


		/* We used to test to see if there were any non-blank
		 * registers before writing.  It didn't speed things
		 * up and required the user to clear the image before
		 * calling this routine.  Now we just write the thing.
		 */
	    ULONG_ACCESS(pD) = destB0;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB1;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB2;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB3;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB4;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB5;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB6;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB7;
	    pD -= dBpl;

	}	/* end of rotating all rects in a row of rectangles */
	break;

	/* ************* 270 degree Rotate of 2 bit/pixel images *********** */


    default: /* depth == 2 bits/pixel */

	    /* If this is a SPARC, we can just address the bytes
	     * without modifying the addresses.  If this is a PC,
	     * we need to XOR each address with 3 before using it.
	     * That's a pain, execution time went up from 0.71
	     * sec to 0.87 sec when we put it in.  So, we note that
	     * we're accessing source bits in a string of vertical
	     * bytes.  We know that all source images are padded
	     * to 4 byte boundaries horizontally, therefore, the
	     * LS two bits of source address do not change in the
	     * 32 source rect accesses.  So, instead of doing
	     * 32 XOR's with 3, we'll do one XOR at the top of
	     * the loop and use that address everywhere.
	     */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* loop through the rectangles in a row of rects */
	for (;
	     blockCount > 0;
	     blockCount--, pS += srcAddrFixup)
#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    /* loop through the rectangles in a row of rects */
	for (pS = (UInt8 *) (((UInt32) pS ) ^ 3);
	     blockCount > 0;
	     blockCount--,
		pS = (UInt8 *) ((((UInt32) pS ^ 3) +
						srcAddrFixup) ^ 3))
#endif /* cAlpacaPCFmt */

	{
		/* PROCESS TOP 4x4 SQUARE */

	    dest = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* We now have bits [31:24] of each dest word in
		 * dest.  Distribute these among the dest regs..
		 */
	    destMask = 0xFF000000;

		/* Store these bits into the MS bytes */
	    destB0 = dest & destMask;
	    dest <<= 8;
	    destB1 = dest & destMask;
	    dest <<= 8;
	    destB2 = dest & destMask;
	    dest <<= 8;
	    destB3 = dest & destMask;


		/* PROCESS NEXT-TO-TOP 4x4 SQUARE */

	    dest = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* We now have bits [23:16] of each dest word in
		 * dest.  Place into the dest regs.
		 */
	    destMask = 0x00FF0000;
	    dest1 = dest;

		/* Store these bits into bits [23:16] of dest regs. */
	    dest >>= 8;	/* rotate into position */
	    destB0 |= dest & destMask;
	    destB1 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB2 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB3 |= dest1 & destMask;


		/* PROCESS NEXT-TO-BOTTOM 4x4 SQUARE */

	    dest = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #4 */
	    pS += sBpl;

		/* We now have bits [15:8] of each dest byte in
		 * dest.  Place into the dest regs.
		 */
	    destMask = 0x0000FF00;
	    dest1 = dest;

		/* Store these bits into bits [15:8] of dest regs. */
	    dest >>= 16;	/* rotate into position */
	    destB0 |= dest & destMask;
	    dest <<= 8;
	    destB1 |= dest & destMask;
	    destB2 |= dest1 & destMask;
	    dest1 <<= 8;
	    destB3 |= dest1 & destMask;


		/* PROCESS BOTTOM 4x4 SQUARE */

	    dest = rotationTable[*pS];	/* byte #1 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #2 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #3 */
	    pS += sBpl;
	    dest <<= 2;

	    dest |= rotationTable[*pS];	/* byte #4 */

		/* We now have bits [7:0] of each dest byte in
		 * dest.  Place into the dest regs.
		 */
	    destMask = 0x000000FF;

		/* Store these bits into bits [7:0] of dest regs. */
	    destB3 |= dest & destMask;
	    dest >>= 8;
	    destB2 |= dest & destMask;
	    dest >>= 8;
	    destB1 |= dest & destMask;
	    dest >>= 8;
	    destB0 |= dest & destMask;


		/* We used to test to see if there were any non-blank
		 * registers before writing.  It didn't speed things
		 * up and required the user to clear the image before
		 * calling this routine.  Now we just write the thing.
		 */
	    ULONG_ACCESS(pD) = destB0;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB1;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB2;
	    pD -= dBpl;
	    ULONG_ACCESS(pD) = destB3;
	    pD -= dBpl;

	}	/* end of rotating all rects in a row of rectangles */
    }
}


/* 180 degree rotation is quite different from 90 or 270 degree rotation.
 * All we have to do is reverse the order of the bits in each line and
 * reverse the order of the lines.
 *
 * We'll scan the source image from left to right, top to bottom.  There is
 * some small chance that this will make the cache circuitry happier (there
 * is no more conventional way to scan memory).  We'll scan the destination
 * image from right to left, bottom to top.
 *
 * In the inner loop, we read four bytes of source word, reverse the order
 * of the bits within the bytes using a lookup table, reverse the order
 * of the bytes within the source word, and write the reversed dest word.
 * Because it's easiest to read bytes into the bottom of a register, we'll
 * shift the dest left to make room for newly reversed source bytes.  This
 * means the first byte read will end up on the left end of the dest word.
 * Therefore, we should read the rightmost byte of the source word first.
 *
 * Well, that's all very nice, but it only works if the read image bits
 * exactly fit into 32 bit words.  Not likely.  If the rightmost 32-bit word
 * of each row has n frame bits on the right end, the rotated image will
 * be shifted right by these very same n bits.  This is no good.
 *
 * To fix this, we'll have two cases: the fast one where everything happens
 * to fit in exactly 32 bits and the somewhat slower one where some shifting
 * is required.  If shifting is required, we maintain two accumulator words,
 * newWord and oldWord.  We do the byte reversing as described above to
 * fill newWord with the contents of one reversed word.  Since we're writing
 * the destination line from right to left, oldWord is trailing along to the
 * right of newWord from the destination's point of view.  Therefore, the
 * word written to memory is generated by shifting oldWord right by 32-padBits
 * and shifting newWord left by padBits, then ORing them.  After the write,
 * we stuff newWord into oldWord and continue.
 *
 * How do we start and how many times do we go through the loop?  Initially,
 * we fill oldWord with 0's.  This is used to generate the frame bits on the
 * right end of the destination line.
 * We go through the loop enough times to complete writing an entire dest
 * line.
 *
 * That worked fine for 1 bit/pixel.  How do we extend it to 2 bits/pixel?
 * In the case where the width is an exact multiple of 32, everything's
 * too easy.  We just use a different table to reverse the order of pixels
 * within bytes.  That is, pixels are moved in units of 2 bits instead of
 * moving them in units of 1 bit.  For example, byte 0000 0001 goes to
 * 0100 0000 instead of 1000 0000.
 *
 * For the cases that require shifting, the only other thing we modify is
 * calculation of the position of the rightmost bit in the memory image.
 * Instead of simply taking srcWidth & 0x1F (equals srcWidth REM 32), we
 * calculate (srcWidth * depth) & 0x1F.  This is done because these images
 * are "depth" times as wide as the 1 bit/pixel images.
 *
 * Ok, there's one more tweak.  "White" pixels in binary images are 0.
 * White pixels in gray images are 1's.  When we check to see if we can
 * skip the reversing step, we check the incoming word against a variable
 * that's set as a function of depth (0 for binary, 0xFFFFFFFF) for gray).
 * If we're being picky, words full of all 0's or all 1's don't need to be
 * reversed in any images, but testing for all 0's OR all 1's requires an
 * extra opcode, and there's not too much black, so we won't bother
 * trying to detect black.
 *
 */

/* PRIVATE */
/****************************************************************************
 *  ip_reverseLine():  reverses the pixels in one line of an image.
 *
 *	pS:		Ptr to left end of source line inside frame
 *	pD:		Ptr to left end of dest image inside frame
 *	widthInWords:	width of line in units of 32-bit words.
 *	rightShift:	distance by which reversed lines should be shifted
 *			right in units of bits.
 *	pRevTable:	Table used to reverse the order of bits in a byte.
 *
 *	Returns:
 *		The pixels in the line are reversed.  The number of bits
 *		per pixel is determined by the table used to reverse
 *		bits within each byte.
 *
 ****************************************************************************/
void  CDECL 
ip_reverseLine (
    IPState    *pIPState,
    UInt8      *pS,
    UInt8      *pD,
    Int32	widthInWords,
    Int32	rightShift,
    const UInt8 *pRevTable)
{
    PROLOGUE;
    UInt32	 newWord, oldWord, leftShift;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* adjust the destination pointer so it points to the right
	 * end of its line.  It turns out to be inconvenient to do
	 * this in the calling code when doing rotations in place.
	 */
    pD += (widthInWords-1) * 4;

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

	/* ************ SPARC VERSION *************** */

	/* Can we avoid shifting?  This happens if the width*depth is an
	 * exact multiple of 32. */
    if (rightShift == 0)
    {
	    /* loop through words in this row.  We read the bytes
	     * of each source word from right to left so they
	     * appear in the dest from left to right */
	for (;
	     widthInWords > 0;
	     widthInWords--, pD -= 4, pS += 4)
	{
		/* We only reverse the order of bits in this word if
		 * it's not filled with 0's or 1's.
		 * The cutesy test turns 0000...00 into 0000..01 and
		 * 1111...11 into 0000..00, then ANDs this with
		 * 1111...10.  Note that if the number started out
		 * as 0 or -1, there will be no upper bits set, so
		 * the AND op will produce 0.  Otherwise, it will
		 * produce some non-zero value.
		 */
	    newWord = ULONG_ACCESS(pS);
	    if (((newWord + 1) & (-2)) != 0)
	    {
		newWord = pRevTable[pS[3]];
		newWord <<= 8;
		newWord |= pRevTable[pS[2]];
		newWord <<= 8;
		newWord |= pRevTable[pS[1]];
		newWord <<= 8;
		newWord |= pRevTable[pS[0]];
	    }
	    ULONG_ACCESS(pD) = newWord;
	}
    }	/* end of case where width was exact multiple of 32 */

    else	/* width is not a multiple of 32 */
    {
	    /* compute the left shift distance */
        leftShift = 32 - rightShift;

	    /* loop through words in this row.  We read the bytes
	     * of each source word from right to left so they
	     * appear in the dest from left to right */
	for (oldWord = 0;
	     widthInWords > 0;
	     widthInWords--, pD -= 4, pS += 4)
	{
		/* We only reverse the order of bits in this word if
		 * it's not filled with 0's or 1's.
		 * The cutesy test turns 0000...00 into 0000..01 and
		 * 1111...11 into 0000..00, then ANDs this with
		 * 1111...10.  Note that if the number started out
		 * as 0 or -1, there will be no upper bits set, so
		 * the AND op will produce 0.  Otherwise, it will
		 * produce some non-zero value.
		 */
	    newWord = ULONG_ACCESS(pS);
	    if (((newWord + 1) & (-2)) != 0)
	    {
		newWord = pRevTable[pS[3]];
		newWord <<= 8;
		newWord |= pRevTable[pS[2]];
		newWord <<= 8;
		newWord |= pRevTable[pS[1]];
		newWord <<= 8;
		newWord |= pRevTable[pS[0]];
	    }
	    ULONG_ACCESS(pD) = (newWord << leftShift) | oldWord;
	    oldWord = newWord >> rightShift;
	}
    }	/* end of case where width was exact multiple of 32 */

#endif /* cBigEndianFmt */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

	/* ******** 386 VERSION ************* */

	/* Can we avoid shifting?  This happens if the width is an
	 * exact multiple of 32. */
    if (rightShift == 0)
    {
	    /* loop through words in this row.  We read the bytes
	     * of each source word from right to left so they
	     * appear in the dest from left to right */
	for (;
	     widthInWords > 0;
	     widthInWords--, pD -= 4, pS += 4)
	{
		/* We only reverse the order of bits in this word if
		 * it's not filled with 0's or 1's.
		 * The cutesy test turns 0000...00 into 0000..01 and
		 * 1111...11 into 0000..00, then ANDs this with
		 * 1111...10.  Note that if the number started out
		 * as 0 or -1, there will be no upper bits set, so
		 * the AND op will produce 0.  Otherwise, it will
		 * produce some non-zero value.
		 */
	    newWord = ULONG_ACCESS(pS);
	    if (((newWord + 1) & (-2)) != 0)
	    {
		newWord = pRevTable[pS[0]];
		newWord <<= 8;
		newWord |= pRevTable[pS[1]];
		newWord <<= 8;
		newWord |= pRevTable[pS[2]];
		newWord <<= 8;
		newWord |= pRevTable[pS[3]];
	    }
	    ULONG_ACCESS(pD) = newWord;
	}
    }	/* end of case where width was exact multiple of 32 */

    else	/* width is not a multiple of 32 */
    {
	    /* compute the left shift distance */
        leftShift = 32 - rightShift;

	    /* loop through words in this row.  We read the bytes
	     * of each source word from right to left so they
	     * appear in the dest from left to right */
	for (oldWord = 0;
	     widthInWords > 0;
	     widthInWords--, pD -= 4, pS += 4)
	{
	    newWord = ULONG_ACCESS(pS);
	    if (((newWord + 1) & (-2)) != 0)
	    {
		newWord = pRevTable[pS[0]];
		newWord <<= 8;
		newWord |= pRevTable[pS[1]];
		newWord <<= 8;
		newWord |= pRevTable[pS[2]];
		newWord <<= 8;
		newWord |= pRevTable[pS[3]];
	    }
		ULONG_ACCESS(pD) = (newWord << leftShift) | oldWord;
		oldWord = newWord >> rightShift;
	}
    }	/* end of case where width was not an exact multiple of 32 */

#endif /* cAlpacaPCFmt */

}

/*

$Log: miorthrt.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.6  1998/01/13 23:52:57  ddavies
miorthrt.c -> /product/ipcore/ipshared/src/RCS/miorthrt.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1995/12/13 02:50:44  ejaquith
miorthrt.c -> /product/ipcore/ipshared/src/RCS/miorthrt.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/07/01  04:30:16  ddavies
 * miorthrt.c -> /product/ipcore/ipshared/src/RCS/miorthrt.c,v
 *
 * Removed implementation of ip_makeRotTable.  It's in mirottab.c now.
 *
 * Revision 1.3  1994/02/13  00:57:35  ddavies
 * miorthrt.c -> /product/ipcore/ipshared/src/RCS/miorthrt.c,v
 *
 * Modified ip_fastRot90Setup and ip_fastRot270Setup so they supply an
 * extra parmater needed by the in-place rotation routines and so they can
 * handle 4 bit and 8 bit images.
 *
 * Revision 1.2  1994/02/04  20:18:29  dferrell
 * miorthrt.c -> /product/ipcore/ipshared/src/RCS/miorthrt.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/21  21:16:58  ddavies
 * Initial revision
 *
 * Revision 1.17  1994/01/14  04:34:46  ddavies
 * sorthrt.c -> /product/ipcore/ipshared/src/RCS/sorthrt.c,v
 *
 * Have ip_makeRotTable return a value at the bottom even though it ends
 * in a switch statement, both of whose branches have returns.  This makes teh
 * compilers happy.
 *
 * Revision 1.16  1993/12/02  19:27:19  ddavies
 * sorthrt.c -> /product/ipcore/ipshared/src/RCS/sorthrt.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.15  1993/11/05  02:22:51  ddavies
 * sorthrt.c -> /product/ipcore/ipshared/src/RCS/sorthrt.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


