/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mipackc.c: low level code used in PackBits compression.
 *
 *   	Int32	ip_getPackBitsCompProc()
 *			Returns pointer to the routine used to compress
 *			one line of PackBits data on this platform.
 *
 *	Int32	ip_compPackBitsRow()
 *			Compresses one row of data.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mipackc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* Here are the restart states for compression. cPBCompNewLine is
 * already state 0.  The bytecounts give additional restart info.
 */
#define	cPBCompWriteLiteralByteCount	1
#define cPBCompWriteLiteralData		2
#define	cPBCompWriteReplicaByteCount	3
#define	cPBCompWriteReplicaData		4

/* There's a big comment on our approach to PackBits compression and
 * decompression in i_packc.c.  You should read that before reading this.
 *
 * To review, one compresses an image using PackBits by looking for two
 * types of runs of bytes.  A run of replicas is one in which all the bytes
 * have the same value.  Most of the time, this is used to compress out
 * long runs of white or black.  A run of literals is one in which no
 * byte equals the next byte (almost).  The "almost" part is that a when
 * the compressor finds a run of literals, two bytes with the same value,
 * then another run of literals, it should merge the three runs into a
 * single run of literals.  Therefore, a run of literals may contain pairs
 * of equal bytes occasionally.
 *
 * Much of the compression work has already been done by the time we get
 * to these routines.  We have already created an array of difference bits.
 * There is one difference bit for each byte in the scan line.  Difference bit
 * n is set if byte n does not equal byte n+1.  We have also already filled
 * an array with the addresses of the starting and ending locations of
 * each run of 1's in the difference bits.  A run of replicas is represented
 * by a run of 0's in the difference bits followed by a single 1 where
 * the value changes.  The single 1 is the last byte in the run of
 * replicas (not the first byte of the next run).  A run of literals starts
 * with the second 1 of a run of 1's (the first 1 ended a run of replicas)
 * and continues through the last 1 in the run.
 *
 * The difference bits runs supplied to this routine have not yet been
 * filtered for the [literal run:pair of replicas:literal run] case.
 * Each time we start a new run, we decide whether a) the length in
 * the new run is greater than one (it represents a run of literals
 * and not just a new run of replicas), the next run also has length
 * greater than one and there is only one 0 between them.  If this is
 * true, all three runs should be merged, so we advance the pointer
 * used to find the right end of the run.  We continue to advance this
 * pointer until either the gap between runs is greater than 1 or the
 * length of the next run is equal to 1.
 *
 * Once we have a run, what next?  If the length of the run is greater than
 * one, we put out a run of literals.  If this doesn't reach the end of the
 * scan line, we put out a run of replicas at the end.
 *
 * How do we put out the run of literals?  First, we divide the run into
 * subruns of 128 bytes plus a subrun of the remaining bytes if necessary.
 * Remember that the length of the literal run in bytes is one less than the
 * length of the run of difference bits.  For each run unit, we write
 * out a byte containing the unit's length minus 1.  This is the standard
 * PackBits format for bytecounts of literal runs.  Now we copy the literal
 * bytes from the image to the compressed bytestream.  This is where
 * the weird byte order stuff comes in.  Images are read in units of words.
 * Compressed data is written in units of bytes.  The first word of an
 * image is aligned on a word boundary.  The bytestream is arbitrarily
 * aligned.  Since the order of the bytes within a word unit only matches
 * the order of bytes in a bytestream on big-endian machines (and it would
 * be awfully nice if the C code worked on both big and little endian
 * machines), we cannot just copy words to the bytestream.  Even shifting
 * the data in the words doesn't work on little endian platforms.  So,
 * we'll do the simple and obvious thing.  We read a word and write its
 * four bytes individually.  The LS (rightmost) byte in the word is
 * written to pCurCompData[3], the next to LS byte is written to
 * pCurCompData[2] and so on.  This puts the rightmost byte later in
 * the bytestream as it should be.
 *
 * Is doing all these writes a horrible performance hog?  I hope not.
 * The typical literal run (at least when compressing binary files) is
 * fairly short - say less than 10 bytes.  It's important to spend as
 * little time as possible on overhead of starting and stopping the
 * copying code.  Also, the compressed PackBits files aren't huge, so
 * writing the bytes individually isn't awful.
 *
 * The nice write-word-as-four-bytes code described above works great
 * as long as we're writing whole words.  What happens at the beginnings
 * and ends of runs?  At this point we should say what's actually in the
 * coordinates of the Run32 structures.  Contained in these structures
 * are the actual addresses of the starting and ending bytes of the runs.
 *
 * We first calculate the address of the image word on which the literal
 * run ends.  This is done by simply masking off the LS two bits of the
 * right run coordinate.  Then we check to see if the run crosses a word
 * boundary (masked right run coord > left run coord).  Short runs
 * may be entirely contained in one word.  If a word boundary is not
 * crossed, we simply read and write each byte individually.  This is cheaper
 * than complicating the process of copying longer runs with this case.
 *
 * If a run does cross a word boundary, we have to deal with both the
 * starting and ending conditions.  The basic structure of the code has
 * a switch statement to copy the desired portion of the first image word,
 * a loop that copies entire image words, and a final switch that copies
 * the relevant bytes of the final image word.  In the first switch,
 * we branch on the LS two bits of the left run address.  In the final switch,
 * we branch on the LS two bits of the right run address.
 *
 * The process above is repeated for each subrun of a literal run.
 *
 * That's how we do runs of literals.  How do we handle runs of replicas?
 * That's easy.  We calculate the length of the run as the difference
 * between the rightmost 1 in the current run and the leftmost run in the
 * next run (remember that that leftmost 1 is actually part of the replica
 * run).  As above, this length is broken down in to subruns of maximum
 * length 128.  For each subrun, we follow the PackBits convention by
 * recording -(runLength - 1) as the bytecount and the byte at the left
 * coord of the next run as the value.
 *
 * The run that marks end-of-line has a left coord that's equal to the line
 * width - 1 and a right coord that's equal to dummyMarker (an argument to
 * the line compression routine).  Note that the left coord has the correct
 * value to end any run of replicas that reaches the right edge of the image.
 *
 * A literal run that goes all the way to the end of the scan line ends at
 * lineWidth - 1, which means we don't want to generate a replica run there.
 * When the code that's going to generate replicas notices that there's
 * no space between the right coord of one run and the left coord of the
 * next one, it omits the replica run.
 *
 * What happens of the only the rightmost byte of a line equals 0?  This
 * would look like the frame bytes, so would produce a 0 difference bit,
 * making it look like a run of replicas.  However, we can't have a run of
 * 1 replica.  This case was taken care of in i_compressRowsPackBits.  If
 * the last run of difference bits ended at lineWidth - 2, it is simply
 * extended to the end of the line.
 *
 * Now we come to the touchy area of error handling.  There are two main
 * conditions under which this procedure will stop.  It might actually
 * finish encoding the entire scan line.  When this happens, we update the
 * pointer to the compressed data that was passed in, set the compState
 * to cPBCompNewLine and return happily.
 *
 * The other reason for returning is that we run out of room for
 * compressed data.  We detect this by comparing pCurCompData to the
 * pointer to the end of the buffer before writing any bytecount and
 * before writing the replica data values.  When copying strings of
 * literals, we shorten the literal string if it's going to run off the
 * end and save the difference in a safe place.  At the end of the
 * string copy, we look at the safe place.  If it's non-zero, we know
 * that there was more string.  In any case, we save an indicator of
 * what we should do when we return in compState, recorded the remaining
 * run bytecount and subrun bytecounts, update the compressed data pointer
 * and return.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_getPackBitsCompProc()
 *
 *	This procedure returns a pointer to the procedure used to
 *	compress one line of PackBits data.
 *
 *	Input:
 *		ppProc:		pointer to pointer to returned proc.
 *
 *	Action:
 *		Fills *ppProc with the address of the procedure used
 *		to compress one row of PackBits data.
 *
 *	Returns:
 *		void
 *
 **************************************************************************/
void  CDECL 
ip_getPackBitsCompProc (
    IPState    *pIPState,
    Int32 (CDECL **ppProc) (
		    IPState    *pIPState,
		    PackBitsCompressionInfo *,
		    Int32	,
		    UInt8     **,
		    UInt8      *))
{
	/* Keep compiler happy */
    pIPState = pIPState;

	/* Return pointer to the correct procedure */
    *ppProc = &ip_compPackBitsRow;

}


/* PRIVATE */
/**************************************************************************
 *   ip_compPackBitsRow()
 *
 *	This procedure compresses one line of PackBits data.  The
 *	compressed data is in natural byte order, so byteswapping
 *	is required on little endian platforms when copying literal strings.
 *
 *	Input:
 *		pPBCompInfo:	pointer to compression state.
 *		dummyMarker:	the value in the rightEnd of the
 *				dummy beyond the last valid run.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *
 *	Action:
 *		Attempt to compress a single scan line of data.
 *		The compressed data is assumed to be in natural byte order and
 *		is arbitrarily aligned, so should be written 1 byte at a
 *		time.
 *
 *		If the end of the line is reached, flush remaining
 *		compressed data to the compressed data buffer and set
 *		pPBCompInfo->compState = cPBCompNewLine;
 *		Otherwise, set pPBCompInfo->compState to the state which
 *		describes the compressed data byte we're about to write and
 *		put the relevant byte counts and run pointers into
 *		*pPBCompInfo.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be written.  Callers can use this
 *		to determine the number of bytes written so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pProc was NULL or compDataOrder
 *					had an illegal value
 *
 **************************************************************************/
Int32  CDECL 
ip_compPackBitsRow (
    IPState    *pIPState,
    PackBitsCompressionInfo *pPBCompInfo,
    Int32	dummyMarker,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
    PROLOGUE;
    UInt8      *pCurImage;		/* pointer to current image byte */
    UInt8      *pImageEnd;		/* pointer to byte beyond last byte of
				         * a literal run. */
    UInt8      *pCurCompData;		/* pointer to current comp data */
    Run32      *pRunStart, *pRunEnd;	/* pointers to Run32 structs on which a
					 * a particular literal run starts and
					 * ends after amalgamation with replica
					 * runs of length 2 */
    UInt32	imgWord;		/* current image word */
    UInt32	overrunBytes;		/* number of bytes that wouldn't fit
					 * into the current subrun because we
					 * filled the buffer. */
    Int32	runByteCount;		/* count of bytes remaining in a
					 * run after subRunByteCount bytes
					 * have been processed.
					 */
    Int32	subRunByteCount;	/* number of bytes remaining in current
					 * subrun of a run.  A run may be
					 * composed of only one subrun, 
					 *in which case runByteCount == 0.
					 * No subrun can contain
					 * more than 128 bytes.
					 */
    UInt32	endAddr;		/* used to save LS bits of
					 * pImageEnd when
					 * we're doing transfers that might not
					 * end on pRunEnd->runRight.
					 */

	/* first initialize things */
    pCurCompData = *ppCompData;
    pRunStart = (Run32 *)((UInt8 *)(pPBCompInfo) +
				    pPBCompInfo->runArrayOffset);
    pRunStart = pRunStart + pPBCompInfo->runStart;
    pRunEnd = pRunStart + (pPBCompInfo->runEnd - pPBCompInfo->runStart);
    runByteCount = pPBCompInfo->runByteCount;
    subRunByteCount = pPBCompInfo->subRunByteCount;

	/* switch on the incoming state.  This tells us both what to
	 * initialize and where to start executing.
	 */
    switch (pPBCompInfo->compState)
    {
    case cPBCompNewLine:
	goto compPBNewLine;

    case cPBCompWriteLiteralByteCount:
	    /* next image byte to be written is at end of run minus
	     * number of bytes yet to be written in this run */
	pCurImage = (UInt8 *)(pRunEnd->runRight -
				(runByteCount + subRunByteCount - 1));
	goto compPBWriteLiteralByteCount;

    case cPBCompWriteLiteralData:
	    /* next image byte to be written is at end of run minus
	     * number of bytes yet to be written in this run */
	pCurImage = (UInt8 *)(pRunEnd->runRight -
				(runByteCount + subRunByteCount - 1));
	goto compPBWriteLiteralBytes;

    case cPBCompWriteReplicaByteCount:
	    /* the replica data byte is just to the right of the last byte
	     * in the literal run. */
	pCurImage = (UInt8 *)(pRunEnd->runRight + 1);
	imgWord = UCHAR_ACCESS(pCurImage);
	goto compPBWriteReplicaByteCount;

    case cPBCompWriteReplicaData:
	    /* the replica data byte is just to the right of the last byte
	     * in the literal run. */
	pCurImage = (UInt8 *)(pRunEnd->runRight + 1);
	imgWord = UCHAR_ACCESS(pCurImage);
	goto compPBWriteReplicaData;

    default:	/* flush pipeline */
	return ia_invalidParm;
    }

	/* At the beginning of the scan line, pRunStart and pRunEnd both
	 * point to the dummy node at the beginning of the line.  Both
	 * of its coords are set to (beginning of line - 1), so it
	 * produces no literal run.  We want to start with a replica run only
	 * if there is space between the left edge of the scan line and
	 * the first run of literals (if any).  If the first literal run
	 * abuts the left edge, just skip to the first real Run32 node now.
	 */
compPBNewLine:
    if (pRunEnd[1].runLeft - pRunEnd[0].runRight <= 1)
    {
	pRunStart++;
	pRunEnd++;
	    /* Note that since there really wasn't a replica run, the
	     * very first byte of the first real set of difference bits is
	     * part of that literal run.
	     */
	pCurImage = (UInt8 *)pRunStart->runLeft;
    }
    else	/* pCurImage should still be set up as through we
		 * had just finished a replica run so that the
		 * first real replica run will be processed properly.
		 */
    {
	pCurImage = (UInt8 *)pRunStart->runLeft + 1;
    }


	/* ********************* MAIN LOOP ***********************/

	/* In this loop, we will write a conditional literal run, then
	 * almost always put out a replica run.  The replica run is only
	 * skipped when at the right end of the line when the literals
	 * abut the right edge of the image.
	 *
	 * The first thing to do in the loop is to try to extend the
	 * current run of literals so that it absorbs any replica runs
	 * of length 2.  Note that the pointer to the first byte of
	 * the next run (pCurImage) has already been set up by the
	 * code processing the previous run.  For example, it is
	 * normally set by the code at the bottom of the loop.
	 * However, when starting a new scan line, it was set above.
	 *
	 * When we have the real run length, we put out a literal run
	 * if the length of the run of difference bit 1's is greater
	 * than one (the first 1 in the run belonged to the
	 * end of the previous replica run).
	 * 
	 * Having put out a literal run (or not), we almost always put
	 * out a replica run.  This is avoided only if the last byte
	 * of a literal run is on the right edge of the image.  The
	 * right edge is detected with another dummy Run32 node.  It's
	 * runLeft coord equals the width of the image - 1 (hence equals
	 * the rightmost coord) and its runRight coord equals dummyMarker.
	 * Given the range of uses i_findAllRunsInLineArray could be
	 * put to, allowing the client to mark the last run (and tell us
	 * what the mark is by passing it in dummyMarker) seemed like
	 * the easiest thing to do.
	 */
    while (pRunStart->runRight != dummyMarker)
    {
	    /* extend the run if possible */
	if (pRunStart->runRight >= (Int32)pCurImage)
	{
	    while ((pRunEnd[1].runRight != pRunEnd[1].runLeft) &&
		   (pRunEnd[1].runLeft - pRunEnd[0].runRight == 2))
		pRunEnd++;
	}

	    /* don't go off the end of the line */
	if (pRunEnd->runRight == dummyMarker)
	    pRunEnd--;

	    /* now decide whether to put out a literal run */
	if (pRunEnd->runRight >= (Int32) pCurImage)
	{
	    subRunByteCount = pRunEnd->runRight - (UInt32)pCurImage + 1;

		/* If this is already a short literal run, don't bother with
		 * trying to break it into short runs.  This is
		 * the most common case by far, so we treat it specially.
		 */
	    if (subRunByteCount <= 128)	
	    {
		    /* Initialize runByteCount so we won't store
		     * something bogus in the case where the run
		     * doesn't all fit in the compression buffer.
		     * That check is done after all the cases of
		     * copying literal strings.
		     */
		runByteCount = 0;

		    /* if there's room, put out the bytecount */
		if (pCurCompData < pCompDataEnd)
		{
		    *pCurCompData++ = (subRunByteCount - 1);
		}
		else	/* ran out of room.  Update the saved state
			 * and remember where to restart.  We'll
			 * restart in the code for the general case
			 * (literal run length > 128) to save having
			 * two labels for almost the same function.
			 */
		{
		    pPBCompInfo->runByteCount = 0;
		    pPBCompInfo->subRunByteCount = subRunByteCount;
		    pPBCompInfo->compState = cPBCompWriteLiteralByteCount;
		    goto compPBExhaustedCompData;
		}

		    /* Now see if there's room to put out all the
		     * data.  If not, shorten the bytecount.
		     * Note that pCurImage already points to the
		     * first image literal byte.  If, for example, there
		     * aren't any more vacant bytes (pCompDataEnd
		     * equals pCurCompData), then we want to stop
		     * before writing any data, so we want
		     * pImageEnd to point BEFORE pCurImage.  That's
		     * where the -1 comes from.
		     */
		if (pCompDataEnd - pCurCompData <= subRunByteCount)
		{
		    overrunBytes = subRunByteCount -
				   (pCompDataEnd - pCurCompData);
		    pImageEnd = pCurImage + (pCompDataEnd - pCurCompData) - 1;
		}
		else	/* the run fits */
		{
		    overrunBytes = 0;
		    pImageEnd = (UInt8 *)(pRunEnd->runRight);
		}

		    /* Write out the correct number of bytes, if any.
		     * There are two implementations of this task.
		     * If the run does not cross a word boundary,
		     * we just copy the bytes one at a time.
		     * If it crosses a word boundary, we have a switch
		     * statement that gets us to the word boundary,
		     * a loop that transfers whole words and a switch
		     * at the end that finishes off the last partial
		     * word.
		     * First, point to the word containing the end 
		     * of the run.  If this also the first word, just
		     * move the bytes one at a time.
		     */
		if (((UInt32)pImageEnd & ~0x3) <= (UInt32)pCurImage)
		{
		    while (pCurImage <= pImageEnd)
		    {
			*pCurCompData++ = UCHAR_ACCESS(pCurImage);
			pCurImage++;
		    }
		}
		else	/* calculate the last image word boundary and
			 * transfer the bytes in the first word.
			 * Note that since we may have modified piamgeEnd
			 * so that it correctly points to the endpoint
			 * when the entire run does not fit into the
			 * compressed data buffer, we must save the
			 * LS bits of the modified pImageEnd to dispatch
			 * on in the closing switch statement.  We
			 * can't always use the LS bits of pRunEnd->runRight.
			 */
		{
		    endAddr = (UInt32)pImageEnd;
		    pImageEnd = (UInt8 *)((UInt32)pImageEnd & ~0x3);
		    switch ((UInt32)pCurImage & 0x3)
		    {
		    case 0:
			    /* read the word pointed to by pCurImage
			     * and write each of the bytes individually.
			     * Note that the LS (rightmost) byte is
			     * written to the largest address in the
			     * compressed data bytestream.
			     */
			imgWord = ULONG_ACCESS(&pCurImage[0]);
			pCurImage += 4;
			pCurCompData[3] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[2] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[1] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[0] = (UInt8)imgWord;
			pCurCompData += 4;
			break;
		    case 1:
			imgWord = ULONG_ACCESS(&pCurImage[-1]);
			pCurImage += 3;
			pCurCompData[2] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[1] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[0] = (UInt8)imgWord;
			pCurCompData += 3;
			break;
		    case 2:
			imgWord = ULONG_ACCESS(&pCurImage[-2]);
			pCurImage += 2;
			pCurCompData[1] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[0] = (UInt8)imgWord;
			pCurCompData += 2;
			break;
		    default:	/* case 3 */
			imgWord = ULONG_ACCESS(&pCurImage[-3]);
			pCurImage += 1;
			pCurCompData[0] = (UInt8)imgWord;
			pCurCompData += 1;
			break;
		    }

			/* Here's the loop that transfers whole words */
		    while (pCurImage < pImageEnd)
		    {
			imgWord = ULONG_ACCESS(pCurImage);
			pCurImage += 4;
			pCurCompData[3] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[2] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[1] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[0] = (UInt8)imgWord;
			pCurCompData += 4;
		    }

			/* transfer the bytes in the last word */
		    switch (endAddr & 0x3)
		    {
		    case 0:	/* last byte is the MS byte of next word */
			*pCurCompData++ = UCHAR_ACCESS(pCurImage);
			pCurImage++;
			break;
		    case 1:
			imgWord = ULONG_ACCESS(pCurImage);
			pCurImage += 2;
			imgWord >>= 16;		/* move byte 1 into LS byte */
			pCurCompData[1] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[0] = (UInt8)imgWord;
			pCurCompData += 2;
			break;
		    case 2:
			imgWord = ULONG_ACCESS(pCurImage);
			pCurImage += 3;
			imgWord >>= 8;		/* move byte 2 into LS byte */
			pCurCompData[2] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[1] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[0] = (UInt8)imgWord;
			pCurCompData += 3;
			break;
		    default: /* case 3 */
			imgWord = ULONG_ACCESS(pCurImage);
			pCurImage += 4;
			pCurCompData[3] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[2] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[1] = (UInt8)imgWord;
			imgWord >>= 8;
			pCurCompData[0] = (UInt8)imgWord;
			pCurCompData += 4;
			break;
		    }
		}
	    }
	    else	/* this is the general case of literal transfer.
			 * In it, we break the literal string into subruns
			 * of no more than 128, looping until all of
			 * the are transferred.
			 */
	    {
		overrunBytes = 0;		/* assume no overrun */
		runByteCount = subRunByteCount;	/* initialize length of
						 * the entire run */
		while ((overrunBytes == 0) && (subRunByteCount > 0))
		{
			/* If subRunByteCount > 128, chop it down to
			 * 128 and save the remainder into runByteCount.
			 */
		    if (subRunByteCount > 128)
			subRunByteCount = 128;

			/* In either case, subtract the length of the
			 * run to be written from the total run length.
			 */
		    runByteCount -= subRunByteCount;

compPBWriteLiteralByteCount:
			/* if there's room, put out the bytecount */
		    if (pCurCompData < pCompDataEnd)
		    {
			*pCurCompData++ = (subRunByteCount - 1);
		    }
		    else    /* ran out of room.  Update the saved state
			     * and remember where to restart.  We'll
			     * restart in the code for the general case
			     * (literal run length > 128) to save having
			     * two labels for almost the same function.
			     */
		    {
			pPBCompInfo->runByteCount = runByteCount;
			pPBCompInfo->subRunByteCount = subRunByteCount;
			pPBCompInfo->compState = cPBCompWriteLiteralByteCount;
			goto compPBExhaustedCompData;
		    }

compPBWriteLiteralBytes:
			/* Now see if there's room to put out all the
			 * data.  If not, shorten the bytecount.
			 * Note that pCurImage already points to the
			 * first image literal byte.  If, for example, there
			 * aren't any more vacant bytes (pCompDataEnd
			 * equals pCurCompData), then we want to stop
			 * before writing any data, so we want
			 * pImageEnd to point BEFORE pCurImage.  That's
			 * where the -1 comes from.
			 */
		    if (pCompDataEnd - pCurCompData <= subRunByteCount)
		    {
			overrunBytes = subRunByteCount -
				       (pCompDataEnd - pCurCompData);
			pImageEnd = pCurImage +
				    (pCompDataEnd - pCurCompData) - 1;
		    }
		    else	/* the run fits */
		    {
			overrunBytes = 0;
			pImageEnd = pCurImage + subRunByteCount - 1;
		    }
    
			/* Write out the correct number of bytes, if any.
			 * There are two implementations of this task.
			 * If the run does not cross a word boundary,
			 * we just copy the bytes one at a time.
			 * If it crosses a word boundary, we have a switch
			 * statement that gets us to the word boundary,
			 * a loop that transfers whole words and a switch
			 * at the end that finishes off the last partial
			 * word.
			 * First, point to the word containing the end 
			 * of the run.  If this also the first word, just
			 * move the bytes one at a time.
			 */
		    if (((UInt32)pImageEnd & ~0x3) <= (UInt32)pCurImage)
		    {
			while (pCurImage <= pImageEnd)
			{
			    *pCurCompData++ = UCHAR_ACCESS(pCurImage);
			    pCurImage++;
			}
		    }
		    else    /* calculate the last image word boundary and
			     * transfer the bytes in the first word.
			     * We have to save the LS two bits of pImageEnd
			     * in something so we can dispatch on them
			     * to transfer the last bytes.
			     */
		    {
			endAddr = (UInt32)pImageEnd;
			pImageEnd = (UInt8 *)((UInt32)pImageEnd & ~0x3);
			switch ((UInt32)pCurImage & 0x3)
			{
			case 0:
				/* read the word pointed to by pCurImage
				 * and write each of the bytes individually.
				 * Note that the LS (rightmost) byte is
				 * written to the largest address in the
				 * compressed data bytestream.
				 */
			    imgWord = ULONG_ACCESS(pCurImage);
			    pCurImage += 4;
			    pCurCompData[3] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[2] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[1] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[0] = (UInt8)imgWord;
			    pCurCompData += 4;
			    break;
			case 1:
			    imgWord = ULONG_ACCESS(&pCurImage[-1]);
			    pCurImage += 3;
			    pCurCompData[2] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[1] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[0] = (UInt8)imgWord;
			    pCurCompData += 3;
			    break;
			case 2:
			    imgWord = ULONG_ACCESS(&pCurImage[-2]);
			    pCurImage += 2;
			    pCurCompData[1] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[0] = (UInt8)imgWord;
			    pCurCompData += 2;
			    break;
			default:	/* case 3 */
			    imgWord = ULONG_ACCESS(&pCurImage[-3]);
			    pCurImage += 1;
			    pCurCompData[0] = (UInt8)imgWord;
			    pCurCompData += 1;
			    break;
			}
    
			    /* Here's the loop that transfers whole words */
			while (pCurImage < pImageEnd)
			{
			    imgWord = ULONG_ACCESS(pCurImage);
			    pCurImage += 4;
			    pCurCompData[3] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[2] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[1] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[0] = (UInt8)imgWord;
			    pCurCompData += 4;
			}
    
			    /* transfer the bytes in the last word */
			switch (endAddr & 0x3)
			{
			case 0:	/* last byte is the MS byte of next word */
			    *pCurCompData++ = UCHAR_ACCESS(pCurImage);
			    pCurImage++;
			    break;
			case 1:
			    imgWord = ULONG_ACCESS(pCurImage);
			    pCurImage += 2;
			    imgWord >>= 16;	/* move byte 1 into LS byte */
			    pCurCompData[1] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[0] = (UInt8)imgWord;
			    pCurCompData += 2;
			    break;
			case 2:
			    imgWord = ULONG_ACCESS(pCurImage);
			    pCurImage += 3;
			    imgWord >>= 8;	/* move byte 2 into LS byte */
			    pCurCompData[2] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[1] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[0] = (UInt8)imgWord;
			    pCurCompData += 3;
			    break;
			default: /* case 3 */
			    imgWord = ULONG_ACCESS(pCurImage);
			    pCurImage += 4;
			    pCurCompData[3] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[2] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[1] = (UInt8)imgWord;
			    imgWord >>= 8;
			    pCurCompData[0] = (UInt8)imgWord;
			    pCurCompData += 4;
			    break;
			}	/* end of last switch */
		    }	/* end of case where subrun crosses a word boundary */

			/* get the number of remaining bytes in the run */
		    subRunByteCount = runByteCount;

		}	/* end of loop processing multiple subruns */
	    }	/* end of case where run is composed of multiple
		 * subruns */

		/* There are two ways to get here.  The normal one is
		 * that we've completed a run of literals successfully.
		 * In the abnormal one, the compressed data buffer
		 * wouldn't hold the entire subrun.  If that happens,
		 * we remember that this subrun should contain another
		 * overrunBytes bytes and quit.
		 */
	    if (overrunBytes != 0)
	    {
		pPBCompInfo->subRunByteCount = overrunBytes;
		pPBCompInfo->runByteCount = runByteCount;
		pPBCompInfo->compState = cPBCompWriteLiteralData;
		goto compPBExhaustedCompData;
	    }

	}	/* At this point, we have written a literal run successfully.
		 * pCurImage already points to the first byte of the
		 * next replica run (if any)
		 */

	    /* Is there a replica run to write?  The only case in which
	     * there is no replica is when a literal run extended to the
	     * right end of the scan line.  The dummy coord just beyond
	     * the right edge has a left coord of (lineWidth - 1), so
	     * this case is indicated by pRunEnd[1].runLeft ==
	     * pRunEnd[0].runRight.  Any replica run which ends at the
	     * right edge will have the proper width since it includes
	     * the leftmost 1 of the next run.  Since the leftmost 1 of the
	     * dummy at the right end of the scan line is at the rightmost
	     * coord of the image, so the replica will happily terminate
	     * in the right place.
	     * The caller to this routine already took care of the case
	     * of a single byte of 0 at the right end of the run.  Without
	     * some specific action, this byte would be left out since
	     * it looks like an extension of the frame. Unfortunately,
	     * there's no such thing as a replica run of length 1.
	     * The caller has already looked at the rightmost run of
	     * literals.  If it ended one byte short of the right end
	     * of the scan line, it was extended by one byte.
	     */
	imgWord = UCHAR_ACCESS(pCurImage);
	subRunByteCount = pRunEnd[1].runLeft - pRunEnd[0].runRight;
	if (subRunByteCount > 0)
	{
		/* If this is a short replica run, don't bother with
		 * trying to break it into short runs.  As before,
		 * this is by far the most common case, so we
		 * treat it specially.
		 */
	    if (subRunByteCount <= 128)	
	    {
		    /* Most of the time, there's room in the buffer
		     * for both bytes, so check that first and don't
		     * waste time checking for each byte individually.
		     */
		if (pCompDataEnd - pCurCompData >= 2)
		{
		    *(Int8 *)pCurCompData++ = -(subRunByteCount - 1);
		    *pCurCompData++ = (UInt8) imgWord;
		}
		else
		{	/* check before putting out each byte */
			/* if there's room, put out the bytecount */
		    if (pCurCompData < pCompDataEnd)
		    {
			*(Int8 *)pCurCompData++ = -(subRunByteCount - 1);
		    }
		    else	/* ran out of room.  Update the saved state
			     * and remember where to restart.  We'll
			     * restart in the code for the general case
			     * (literal run length > 128) to save having
			     * two labels for almost the same function.
			     */
		    {
			pPBCompInfo->runByteCount = 0;
			pPBCompInfo->subRunByteCount = subRunByteCount;
			pPBCompInfo->compState = cPBCompWriteReplicaByteCount;
			goto compPBExhaustedCompData;
		    }
    
			/* If there's room, put out the replica data */
		    if (pCurCompData < pCompDataEnd)
		    {
			*pCurCompData++ = (UInt8) imgWord;
		    }
		    else	/* out of room again */
		    {
			pPBCompInfo->runByteCount = 0;
			pPBCompInfo->subRunByteCount = subRunByteCount;
			pPBCompInfo->compState = cPBCompWriteReplicaData;
			goto compPBExhaustedCompData;
		    }
		}	/* end of short replica run where we check before each
			 * write */
	    }	/* end of short replica run */
	    else
	    {
		    /* this is a long replica run.  If it all fits
		     * into the compressed data buffer, write it
		     * quickly.  This is a fairly common case for
		     * blank lines.  Otherwise, write things painfully.
		     *
		     * Well, almost.  If (runLength MOD 128) == 1,
		     * decrease it by 1.  The worst thing that can
		     * happen is to have a literal of length 1, which
		     * is exactly what we'd need anyway.  However, if
		     * there's a literal run following this replica,
		     * we should absorb the last replica byte into it.
		     * If the run following the replica run is another
		     * replica of length 2 which is followed by a
		     * literal, we will happily merge this last replica
		     * byte, the short replica run and the literal.
		     */
		if (((subRunByteCount & 0x7F) == 1) &&
		    (pRunEnd[1].runRight != dummyMarker))
		{
		    pRunEnd[1].runLeft--;
		    subRunByteCount--;
		}

		runByteCount = subRunByteCount;
		    /* if all the run specifiers fit into the buffer,
		     * which is the common case, write them all
		     * in one loop without further checking.
		     */
		if ((2*(runByteCount/128)) + 2 <= (pCompDataEnd - pCurCompData))
		{
		    while (runByteCount > 0)
		    {
			subRunByteCount = MIN(128, runByteCount);
			*(Int8 *)pCurCompData++ = -(subRunByteCount - 1);
			*pCurCompData++ = (UInt8) imgWord;
			runByteCount -= subRunByteCount;
		    }
		}
		else 	/* all the run specifiers don't fit, so check
			 * before writing each one.
			 */
		{
		    while (runByteCount > 0)
		    {
			subRunByteCount = MIN(128, runByteCount);
compPBWriteReplicaByteCount:
			if (pCurCompData < pCompDataEnd)
			    *(Int8 *)pCurCompData++ = -(subRunByteCount - 1);
			else
			{
			    pPBCompInfo->runByteCount = runByteCount;
			    pPBCompInfo->subRunByteCount = subRunByteCount;
			    pPBCompInfo->compState =
						cPBCompWriteReplicaByteCount;
			    goto compPBExhaustedCompData;
			}
compPBWriteReplicaData:
			if (pCurCompData < pCompDataEnd)
			    *pCurCompData++ = (UInt8) imgWord;
			else
			{
			    pPBCompInfo->runByteCount = runByteCount;
			    pPBCompInfo->subRunByteCount = subRunByteCount;
			    pPBCompInfo->compState = cPBCompWriteReplicaData;
			    goto compPBExhaustedCompData;
			}
			runByteCount -= subRunByteCount;
		    }	/* end of loop that writes consecutive subruns
			 * while checking for overrun at each write
			 */
		}	/* end of case that writes consecutive subruns
			 * while checking for overrun at each write
			 */
	    }	/* end of writing long replica runs */
	}	/* end of writing replicas */

	    /* point just beyond end of replica run */
	pRunEnd++;				/* start to process the */
	pRunStart = pRunEnd;			/* next run */
	pCurImage = (UInt8 *)(pRunStart->runLeft) + 1;


    }	/* end of loop that writes out a line of data */

	/* Hooray, we finished successfully!
	 * Update the saved state and return.
	 */
    pPBCompInfo->compState = cPBCompNewLine;
    *ppCompData = pCurCompData;
    NORMAL_RETURN;

	/* We get here when we've filled the compressed data buffer before
	 * completing the correct number of scan lines.  The code that
	 * jumped here has already set pPBCompInfo->runByteCount,
	 * pPBCompInfo->subRunByteCount and pPBCompInfo->compState.
	 */
compPBExhaustedCompData:
    pPBCompInfo->runStart = ((UInt32)pRunStart -
			     ((UInt32)pPBCompInfo +
			      pPBCompInfo->runArrayOffset)) / sizeof(Run32);
    pPBCompInfo->runEnd = pPBCompInfo->runStart + (pRunEnd - pRunStart);
    *ppCompData = pCurCompData;
    NORMAL_RETURN;
}


/*

 $Log: mipackc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.10  1998/01/13 23:52:59  ddavies
 mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.9  1997/07/11 01:49:50  ddavies
 mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.8  1996/04/10 21:59:08  ddavies
 ddavies on Wed Apr 10 14:58:21 PDT 1996

 mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v

 Improve comments, fix errors in them too.

 Revision 1.7  1995/12/13 02:50:51  ejaquith
 mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1995/10/18  04:03:42  ddavies
 * mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.5  1995/10/06  01:53:41  ddavies
 * ddavies on Thu Oct 5 18:49:13 PDT 1995
 *
 * mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v
 *
 * Attempt to fix seg fault that was only detected on Solaris machines (weird).
 * The check in ip_compPackBitsRow that tried to see if a long replica would
 * fit into the buffer looked for (runByteCount/128) + 2 <= <bytes left in buffer>
 * instead of 2*(runByteCount/128) + 2 <= <bytes left in buffer>.  Each
 * literal requires 2 bytes, not one.
 *
 * Revision 1.4  1995/07/11  00:14:23  ddavies
 * mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v
 *
 * Changed CCRun to Run32.
 *
 * Revision 1.3  1995/05/26  01:08:32  ddavies
 * ddavies on Thu May 25 18:07:27 PDT 1995
 *
 * mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v
 *
 * Make ip_getPackBitsCompProc return void.  Fix a lot of comments.
 *
 * Revision 1.2  1995/05/10  02:03:11  ddavies
 * mipackc.c -> /product/ipcore/ipshared/src/RCS/mipackc.c,v
 *
 * Fix off-by-one bug when computing address of last valid image byte when
 * entire run of literals will not fit into the compressed data buffer.  Also
 * save this address so we can dispatch on it's LS bits instead of always
 * dispatching on the LS bits of pRunEnd->runRight.
 *
 * Revision 1.1  1995/05/09  02:58:20  ddavies
 * Initial revision
 *

 */


