/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mipackd.c: high level code used in PackBits decompression.
 *
 *   	Int32	ip_getPackBitsDecompProc()
 *			Returns pointer to the routine used to decompress
 *			one line of PackBits data on this platform.
 *
 *	Int32	ip_decompPackBitsRow()
 *			Decompresses one row of data.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "defines.pub"
#include "iaerror.pub"
#include "imageref.pub"

    /*  prototyes */
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mipackd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* There's a big comment on our approach to PackBits compression and
 * decompression in i_packd.c.  You should read that before reading this.
 *
 * There's not much of a trick to decompressing a PackBits file.  Basically,
 * you read a byte count (n) and decide if it's positive or negative.  If
 * positive (between 0 and 127 inclusive), you copy the next n+1 bytes
 * from the compressed file to the image.  If it's between (-1 and -127)
 * inclusive, you replicate the byte following the byte count -n + 1 times
 * into the image.  If the count is -128, just skip it.  Repeat this process
 * until you reach the end of the line.  Strings of replicas (copies of the
 * same byte) or literals (copies of a string of bytes) must stop at
 * image scan line boundaries.
 *
 * That is too simple, I have to complicate it a little.  We use PackBits
 * for two purposes.  We want to be able to compress and decompress TIFF
 * files that use PackBits format.  The PerfectScan Basic Viewer also uses
 * PackBits as an internal format for cached images.  In both cases, the
 * uncompressed image lines should be accessed with a granularity of 32, not
 * 8, bits.  That is, we read and write the image in 32-bit chunks on
 * both big and little endian machines.  This is a bummer on little endian
 * machines since the compressed data is supposed to be in "natural byte order,"
 * which means there's a byteswap needed between the compressed data and
 * the image data.
 *
 * The byteswap means we can't use memset and memcpy.  Well, for
 * performance reasons, we might not want to use them anyway.
 * Using them implies lots of procedure calls, one for each string being
 * decoded.  Instead, we will have the copying routine embedded in the
 * routine that decompresses entire scan lines.
 *
 * This file contains a C implementation of ip_decompressPackBitsRow.
 * Assembly files might contain more than one version.  For example,
 * the i486 has a byteswap opcode but the i386 doesn't.  The routine
 * ip_getPackBitsDecompProc returns a pointer to the routine being used.
 *
 * Each routine operates by accumulating uncompressed bytes into output
 * words and writing the output words.  Each routine reads the next byte count
 * and decides whether to ignore it (n = 128), copy a run of literals
 * to the image (n >= 0) or replicate the following byte -n + 1 times
 * (n < 0).
 *
 * An important feature to remember is that runs of replicas can be pretty
 * long (for large black or white blocks) while runs of literals tend to
 * be pretty short, say 4 to 12 bytes.  At least this works at 300 dpi
 * binary.  Runs of literals in gray images will be longer, but even for
 * 4 bit images, there are lots of replicas.  Therefore, it doesn't make
 * a lot of sense to write fancy blitting routines for the literal strings
 * that have a fair amount of overhead.  Just get in and get out.
 *
 * Let's look at the two main cases in the decompression routine.
 *
 * 1. Replicate bytes.
 *
 * This is straightforward.  We create a word containing 4 copies of the
 * byte to be replicated.  We AND it with a mask having 1's in the bit
 * positions of the output word that are not yet filled (leftMask) and OR the
 * result into the current output word.  We also calculate the address of
 * the image word containing the end of the replica run.  If that's not the
 * same as the first word, write the output word and increment the output
 * word address.  While we're not pointing to the image word containing
 * the last byte of the run, write the replica word to the image and
 * increment the address.  For efficiency reasons, we may wish to write
 * two words each time through the loop.  This does make the termination
 * condition a little harder since we have to see which word of the final
 * pair should be written.
 *
 * Anyway, when we get to the word containing the last run, we mask the
 * replica word with a mask containing 1's in only the valid image
 * positions (rightMask).  If the word is now full, we write the word,
 * set the offset of the next valid byte to 0 and go to the top of the loop.
 *
 * 2. Copy literal string
 *
 * A reasonably efficient byteswap routine reads each byte from the
 * source individually and inserts them into the output word one at a time.
 * We'll do that.
 *
 * I looked at cutsey C tricks like using bit fields and/or unions to
 * insert bytes into words.  They're all so implementation dependent that
 * you can't depend on code like that from one compiler to the next.
 *
 * The code has an entrance section, a bulk transfer section and an exit
 * section.  In the entrance section, we fill outWord with bytes from
 * the compressed data stream.  If there aren't enough bytes in the literal
 * string to do this, we fill what's available and skip everything else.
 *
 * In the loops of the bulk transfer section, we assume outWord is full
 * on entry and exit.  Each loop writes outWord, then fills it up again.
 * We leave the bulk transfer section when there's not enough data to
 * fill outWord again.  outWord is filled by shifting it left by 8 and
 * filling the vacancy with a new byte.  Nothing fancy.  To make this
 * work, we shift the first output word right by (number vacant bytes
 * - 1) before inserting the first literal byte and left by (new number
 * of vacant bytes) after inserting the last one.
 *
 * In the exit section, we write outWord and put the last bytes of the
 * string into it.  It should now be partially filled.
 *
 *
 *
 * A couple of other notes.  We need to check to see when we're going to
 * run off the end of the scan line AND to see when we're going to run
 * out of compressed data.  We'd like to minimize the number of checks.
 *
 * Well, first, we note that the scan line can only end between runs of
 * compressed data.  So, we only need to check for the end of a scan line
 * then.
 *
 * Next, we adjust the length of any run of compressed data so that it
 * doesn't run off the end of the compressed data block.  The desired
 * run length is the length specified in the file.  The real run length
 * is the minimum of the desired run length and the distance to the end
 * of the compressed data block.  At the start of each run, we see if the
 * the run will extend beyond the end of the compressed data block.  If
 * so, we write the number of extra bytes into tempWord and set the real
 * run length appropriately.  At the end of each run, we look at tempWord.
 * If it's non-zero, we record the decompression state, current output
 * word, number of vacant spaces in the output word, current compressed
 * data pointerl and, the number of remaining bytes in the run.  Calling
 * code can tell there's more to do if a) the offset into the uncompressed
 * image scan line is non-zero, b) the next vacant byte in outWord is
 * non-zero, or c) we're out of compressed data (the compressed data
 * buffer pointer points to the end of the buffer) but we haven't processed
 * the right number of lines yet.  We really need all three of these
 * conditions.  The last one detects running out at any scan line
 * boundary before the last one.  The first two detect running out anywhere
 * else.
 *
 * If we get to the end of the scan line happily, we also set all the
 * entries in pPBDecompInfo as above except we set the decompState to
 * "find new bytecount" and set nextVacantByte = 0 and imageWordOffset = 0.
 * In any case, we set the returned value of *ppCompData to point to
 * the next compressed data byte to be read.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_getPackBitsDecompProc()
 *
 *	This procedure returns a pointer to the procedure used to
 *	decompress one line of PackBits data.
 *
 *	Input:
 *		ppProc:		pointer to pointer to returned proc.
 *
 *	Action:
 *		Fills *ppProc with the address of the procedure used
 *		to decompress one row of PackBits data.
 *	Returns:
 *		void
 *
 **************************************************************************/
void  CDECL 
ip_getPackBitsDecompProc (
    IPState    *pIPState,
    Int32 (CDECL **ppProc) (
		    IPState    *pIPState,
		    PackBitsDecompressionInfo *,
		    UInt8      *,
		    UInt8      *,
		    UInt8     **,
		    UInt8      *))
{
	/* Make compiler happy */
    pIPState = pIPState;

	/* Return pointer to the correct procedure */
    *ppProc = &ip_decompPackBitsRow;

}



/* Here are the restart states for decompression.
 * cPBDecompNewLine is already state 0.  It is used when the next
 * thing to do is read a bytecount.  pPBDecompInfo->nextVacantByte
 * says where the first vacancy is in the output word.
 */
#define	cPBDecompReadData	1

/* PRIVATE */
/**************************************************************************
 *   ip_decompPackBitsRow()
 *
 *	This procedure decompresses one line of PackBits data.  The
 *	compressed data is in natural byte order in a little endian
 *	platform, so byteswapping is required during unpacking
 *	of literal strings.
 *
 *	Input:
 *		pPBDecompInfo:	pointer decompression state.
 *		pImage:		pointer to leftmost byte in image line
 *		pImageEnd:	Pointer to byte just beyond valid byte in
 *				uncompressed image scan line.
 *		ppCompData:	pointer to pointer to the beginning of
 *				compressed data block.  The data block
 *				pointer is updated at the end of this
 *				routine.
 *		pCompDataEnd:	pointer to byte just beyond end of
 *				compressed data
 *
 *	Action:
 *		Attempt to decompress a single scan line of data.
 *		The data is assumed to be in natural byte order and
 *		is arbitrarily aligned, so should be read 1 byte at a
 *		time.
 *
 *		If the end of the line is reached, flush the last remaining
 *		partial word to the image and set
 *		pPBDecompInfo->imageWordOffset = 0 and
 *		pPBDecompInfo->nextVacantByte = 0;
 *		Otherwise, set pPBDecompInfo->imageWordOffset to be the
 *		byte offset of the space where the image word will be
 *		written and pPBDecompInfo->nextVacantByte equal to the
 *		offset in pPBDecompInfo->imageWord.
 *
 *		If the end of the compressed data was reached at the same
 *		time the end of a scanline was reached, set
 *		pPBDecompInfo->decompState = cPBDecompReadByteCount.
 *		Otherwise, set pPBDecompInfo->decompState
 *		to the state to be executed for the next byte and
 *		pPBDecompInfo->runByteCount equal to the number of bytes
 *		remaining in the current run, if any.
 *
 *		In any case, *ppCompData is set to point to the next
 *		byte of compressed data to be read.  Callers can use this
 *		to determine the number of bytes read so far.
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_invalidParm		pProc was NULL or compDataOrder
 *					had an illegal value
 *
 **************************************************************************/
Int32  CDECL 
ip_decompPackBitsRow (
    IPState    *pIPState,
    PackBitsDecompressionInfo *pPBDecompInfo,
    UInt8      *pImage,
    UInt8      *pImageEnd,
    UInt8     **ppCompData,
    UInt8      *pCompDataEnd)
{
PROLOGUE;
 UInt8	*pCurImage;		/* pointer to current image byte */
UInt8	*pCurCompData;		/* pointer to current comp data */
UInt8	*pRunEnd;		/* points to word or byte on which run ends */
UInt32	 outWord;		/* current output word */
UInt32	 mask;			/* mask used at beginnings and ends of runs */
UInt32	 tempWord;		/* filled with replicated bytes or contains
				 * number of literal run bytes that didn't fit
				 * in current compressed buffer. */
Int32	 byteCount;		/* read from compressed data */
Int32	 nextVacantByte;	/* holds a number from 0 through 3 indicating
				 * the next vacant byte in outWord.
				 */
    /* This is the array of masks used to get rid of crud off the end of
     * the right side of each run.  It is indexed with the offset of the
     * last byte in the run relative to the left side of the word.
     * The array is repeated because we put a termination condition bit
     * into bit 2 of the index.
     * Why are entries 3 and 7 set to 0?  Shouldn't they be all 1's?
     * Well, we don't really use the rightmask to mask anything if the
     * run ends on a word boundary.  However, we do initialize the subsequent
     * left mask (bits of the next run that are valid) as the complement
     * of the previous rightmask.  All bits in subsequent image word are
     * new, so we set the table up so that when we complement 0, we get
     * all 1's for the next leftmask.
     */
static	UInt32 rightMask[8] = {0xFF000000, 0xFFFF0000, 0xFFFFFF00, 0x00000000,
			       0xFF000000, 0xFFFF0000, 0xFFFFFF00, 0x00000000};

	/* first initialize things */
    pCurImage = pImage + pPBDecompInfo->imageWordOffset;
    pCurCompData = *ppCompData;
    outWord = pPBDecompInfo->imageWord;
    byteCount = pPBDecompInfo->runByteCount;
    nextVacantByte = pPBDecompInfo->nextVacantByte;
	/* initialize mask with the complement of the first left mask
	 * value so we can use it to initialize the left mask
	 * in the main loop.
	 */
    mask = ~(0xFFFFFFFFUL >> (nextVacantByte * 8));

	/* There are really only two output word states in which we can start
	 * or restart, read a bytecount or read data.
	 * Yeah, the big comment says there are 4 * 2 states, but
	 * the only differences between the four main states were trivial
	 * when there's no byteswappings, so we consolidated them.
	 * We just branch on the decompState to see which state
	 * to start in.
	 */
    if (pPBDecompInfo->decompState == cPBDecompNewLine)
	goto pbdbsReadByteCount;
    else	/* bytecount is already valid, go read the data 
		 * following the bytecount.
		 */
    if (pPBDecompInfo->decompState == cPBDecompReadData)
	goto pbdbsReadData;
    else
	return ia_invalidParm;

	/* *************** MAIN LOOP ******************************* */

	/* The loop goes until we've processed an entire image line.
	 * We read a bytecount from the compressed byte stream and
	 * decide what to do.  If the count is in (0, 127), we copy the
	 * next n+1 bytes literally.  If the count is in (-1, -127), we
	 * replicate the next byte -n + 1 times.  If the count is -128,
	 * we skip it.
	 * Before we read the bytecount and before reading the arguments,
	 * we check to see if that byte exists.  If not, we record the
	 * current image and compression states and return.  If so, we
	 * proceed.
	 */
pbdbsReadByteCount:
    while (pCurImage + nextVacantByte < pImageEnd)
    {
	    /* if we can't read the next byte count, restart here */
	if (pCurCompData >= pCompDataEnd)
	{
	    pPBDecompInfo->decompState = cPBDecompNewLine;
	    goto pbdbsCompExhausted;
	}

	byteCount = *(Int8 *)pCurCompData;	/* get the byte count */
	pCurCompData++;				/* point to next byte */
    
	    /* if the bytecount was the last thing in the compressed
	     * data buffer, stop here.  This is also where we'll
	     * restart if we reached the end of the compressed data
	     * in the midst of a run of literals.  We can't reach the
	     * end of the compressed data in the midst of a run of 
	     * replicas since there's only one replica byte to read.
	     * We either fail here before reading it or fill in the
	     * entire run with it.
	     */
pbdbsReadData:
	if (pCurCompData >= pCompDataEnd)
	{
	    pPBDecompInfo->decompState = cPBDecompReadData;
	    goto pbdbsCompExhausted;
	}
    
	    /* initialize the left mask after reading the
	     * bytecount so that this still happens if we
	     * restart by reading the bytecount.  We initialize it
	     * with the complement of the mask that indicated which
	     * bytes of outWord were correct at the end of the
	     * previous run.  The new value of mask will indicate which
	     * bytes of outWord can be replaced with new run bits.
	     */
	mask = ~mask;

	if (byteCount >= 0)
	{
		/* *********** COPY LITERAL STRINGS ************** */

		/* Here's where we copy the next byteCount + 1
		 * bytes from the compressed data stream to the
		 * image.
		 */
	    byteCount++;	/* get the real run length */

		/* This was basically explained in the comment above the
		 * top of the routine.  There is a bulk copying loop and
		 * a careful finishing loop.  The entry point into the
		 * bulk copying loop is determined by the number of
		 * bytes already in outWord, which happens to equal
		 * nextVacantByte.
		 * Since the loops do byteswapping by shifting each
		 * new byte into the bottom of outWord, we should shift
		 * outWord right by (4 - nextVacantByte) * 8 before
		 * entering the loops to make the current data right
		 * aligned.  When we exit, we shift outWord back left
		 * by (4 - nextVacantByte) * 8.
		 *
		 * In all the other loops in this module, we used
		 * the address of the last image word to determine the
		 * exit condition.  In this loop, since we're not
		 * copying words, we'll use the address of the last
		 * compressed data byte.  First, we clip the run to
		 * fit the compressed data buffer if necessary, then
		 * set pRunEnd to be the address of the byte just
		 * off the end of the string (address of the next
		 * bytecount).
		 *
		 * So, does the compressed run extend beyond the end of
		 * of the compressed data buffer?
		 */
	    if (pCurCompData + byteCount > pCompDataEnd)
	    {
		    /* save the excess and revise byteCount */
		tempWord = byteCount - ((UInt32)pCompDataEnd -
					(UInt32)pCurCompData);
		byteCount -= tempWord;
	    }
	    else	/* make sure we know the run ended */
	        tempWord = 0;

		/* Every loop in the main copying code assumes
		 * that outWord is empty at the top of the loop.
		 * Therefore, we should try to fill outWord before
		 * starting, so we can empty it.  If there aren't
		 * enough bytes in the string to fill outWord,
		 * that won't work.  Deal with that special case
		 * first.  When we leave here, nextVacantByte should
		 * point to the last valid byte, not the next vacant
		 * byte.  That's so the code at the very end of
		 * the main loop (below the replica code below)
		 * works properly.
		 */
		/* right-align outWord.  If outWord is empty, this
		 * either clears outWord or does nothing.  Either
		 * is ok since we're about to fill it.
		 */
	    outWord >>= (4 - nextVacantByte) * 8;

	    if (byteCount + nextVacantByte <= 4)
	    {
		nextVacantByte += byteCount - 1;	/* point to last valid
							 * byte after transfer
							 * is complete */
		while (byteCount > 0)
		{
		    outWord <<= 8;
		    outWord |= (*pCurCompData++);
		    byteCount--;
		}
	    }
	    else	/* fill up the first word and transfer a
			 * bunch of words */
	    {
		    /* Calculate address of end of string in
		     * the compressed data (pRunEnd) and in image data
		     * (byteCount).  Note that the LS two bits of pCurImage
		     * are always 00, so they don't affect the lower two
		     * bits of byteCount.  Those are the only two we use.
		     */
		pRunEnd = pCurCompData + byteCount - 1;
		byteCount += nextVacantByte - 1;
    
		    /* fill in outWord.  We know we can do this
		     * because we know byteCount + nextVacantByte > 4
		     * so the string extends beyond the end of outWord.
		     */
		switch (nextVacantByte)
		{
		case 0:
		    outWord = pCurCompData[0];
		    outWord = (outWord << 8) | pCurCompData[1];
		    outWord = (outWord << 8) | pCurCompData[2];
		    outWord = (outWord << 8) | pCurCompData[3];
		    pCurCompData += 4;
		    break;	/* we're already on a word boundary */
		case 1:
		    outWord = (outWord << 8) | pCurCompData[0];
		    outWord = (outWord << 8) | pCurCompData[1];
		    outWord = (outWord << 8) | pCurCompData[2];
		    pCurCompData += 3;
		    break;
		case 2:
		    outWord = (outWord << 8) | pCurCompData[0];
		    outWord = (outWord << 8) | pCurCompData[1];
		    pCurCompData += 2;
		    break;
		default: /* case 3 */
		    outWord = (outWord << 8) | pCurCompData[0];
		    pCurCompData += 1;
		    break;
		}
		    /* The loops below assume outWord is full
		     * on entry and on exit.
		     */
		pRunEnd -= 7;
		while (pCurCompData <= pRunEnd)
		{
		    ULONG_ACCESS(&pCurImage[0]) = outWord;
		    outWord = pCurCompData[0];
		    outWord = (outWord << 8) | pCurCompData[1];
		    outWord = (outWord << 8) | pCurCompData[2];
		    outWord = (outWord << 8) | pCurCompData[3];
		    ULONG_ACCESS(&pCurImage[4]) = outWord;
		    outWord = pCurCompData[4];
		    outWord = (outWord << 8) | pCurCompData[5];
		    outWord = (outWord << 8) | pCurCompData[6];
		    outWord = (outWord << 8) | pCurCompData[7];
		    pCurCompData += 8;
		    pCurImage += 8;
		}

		    /* We either have one more whole word before
		     * the word that contains the end of the string
		     * or we don't.  Check by seeing if pCurCompData
		     * is less than the last address - 3 (or
		     * pRunEnd - 7 + 4).
		     */
		pRunEnd += 4;
		if (pCurCompData <= pRunEnd)
		{
		    ULONG_ACCESS(&pCurImage[0]) = outWord;
		    outWord = pCurCompData[0];
		    outWord = (outWord << 8) | pCurCompData[1];
		    outWord = (outWord << 8) | pCurCompData[2];
		    outWord = (outWord << 8) | pCurCompData[3];
		    pCurCompData += 4;
		    pCurImage += 4;
		}

		    /* put the remaining portion of the string (if any)
		     * into the last word.  If there is a remaining
		     * portion, remember to write outWord now since
		     * it's full.  Remember that pRunEnd points to
		     * the last byte of the current string, so we
		     * must test for pCurCompData less than OR EQUAL
		     * TO pRunEnd.  In either case, we haven't
		     * processed the last string byte yet.
		     */
		pRunEnd += 3;

		if (pCurCompData <= pRunEnd)
		{
		    ULONG_ACCESS(pCurImage) = outWord;
		    pCurImage += 4;
		    outWord = 0;

		    switch ((UInt32)pRunEnd - (UInt32)pCurCompData)
		    {
		    case 3:
			outWord = pCurCompData[0];
			pCurCompData++;
		    case 2:
			outWord = (outWord << 8) | pCurCompData[0];
			pCurCompData++;
		    case 1:
			outWord = (outWord << 8) | pCurCompData[0];
			pCurCompData++;
		    case 0:
			outWord = (outWord << 8) | pCurCompData[0];
			pCurCompData++;
			break;
		    default:
			return ia_internal;
		    }

		}
		    /* make nextVacantByte point to the last valid
		     * byte in a properly aligned version of outWord.
		     */
		nextVacantByte = byteCount & 0x3;
	    }

		/* align outWord */
	    outWord <<= (3 - nextVacantByte) * 8;

		/* If the run didn't fit into the compressed buffer,
		 * record the amount remaining and the restart state.
		 * Go clean up after that.  Note that we store tempWord
		 * minus 1 since we'll increment it on restart.
		 * The state of outWord and nextVacantByte depends on
		 * where we start and end.  If we happened to end on
		 * a word boundary, write the last known image word
		 * and prepare for another.  If not, just mask off
		 * the crud on the right and prepare to restart.
		 */
	    if (tempWord != 0)
	    {
		byteCount = tempWord - 1;
		if (nextVacantByte == 3)
		{
			/* ending on a word boundary */
		    ULONG_ACCESS(pCurImage) = outWord;
		    pCurImage += 4;
		    nextVacantByte = 0;
		    outWord = 0;     /* mask off everything in the new word */
		}
		else
		{
			/* mask off crud to the right of nextVacantByte,
			 * which is really the last valid byte at this
			 * point.
			 */
		    outWord &= rightMask[nextVacantByte];
		    nextVacantByte++;
		}
		pPBDecompInfo->decompState = cPBDecompReadData;
		goto pbdbsCompExhausted;
	    }
	}
	else
		/* *************** REPLICA BYTES ***************** */
	{
	    if (byteCount == 128)
		continue;		/* skip the noop */
	    byteCount = -byteCount + 1;	/* copy replicas or noop */

		/* Here's the code used to replicate a byte.  We know
		 * that the byte is available in the compressed data
		 * stream because we checked in the main loop.
		 *
		 * The code follows these steps:
		 * 1. Set pRunEnd equal to pCurImage + nextVacantByte +
		 *	  byteCount - 1.  This is the full address of the last
		 *	  byte of the string.
		 * 2. Set nextVacantByte = pRunEnd & 0x3.  This is the
		 *	  offset of the last byte in the run relative to
		 *	  its word.
		 * 3. Set pRunEnd equal to pRunEnd & ~0x3.  This makes
		 *	  it point to the last word to be filled.
		 * 4. Replicate the byte into all four bytes of
		 *	  tempWord.
		 * 5. Insert tempWord & mask into outWord.
		 * 6. If pCurImage equals pRunEnd, skip to step 9.
		 * 7. Write outWord to *pCurImage, load outWord with
		 *	  tempWord and make pCurImage point to the next word.
		 * 8. While pCurImage does not equal pRunEnd, write outWord
		 *	  to *pCurImage, and increment pCurImage.  We will
		 *	  actually do this loop in slightly larger chunks
		 *	  (2 words) for efficiency.
		 */

		/* Calculate address of last byte in run */
	    pRunEnd = (UInt8 *)(pCurImage + nextVacantByte + byteCount - 1);
	
		/* Get byte offset into last word of run */
	    nextVacantByte = (UInt32)pRunEnd & 0x3;
	
		/* make pRunEnd point to a word boundary for easy
		 * comparision later */
	    pRunEnd = (UInt8 *) ((UInt32)pRunEnd & ~0x3);

		/* Create a word containing the replicated byte */
	    tempWord = *pCurCompData;	/* get replica byte */
	    pCurCompData++;		/* point to next bytecount */
	    tempWord |= tempWord << 8;	/* in 2 LS bytes */
	    tempWord |= tempWord << 16;	/* in all 4 bytes */
	
		/* insert replicated byte into outWord.
		 * This frees up mask. */
	    outWord |= tempWord & mask;
	
		/* Here's the loop that fills in the words between the
		 * start and end of the run.  Don't execute it if the
		 * run starts and ends in the same word.
		 */
	    if (pRunEnd > pCurImage)
	    {
		ULONG_ACCESS(pCurImage) = outWord;	/* write first word */
		pCurImage += 4;
		outWord = tempWord;
	
		    /* we really want to test to see if there are
		     * at least two words between pCurImage and pRunEnd.
		     * To optimize this, we temporarily make pRunEnd
		     * point two words earlier than it should.
		     */
		pRunEnd -= 8;
		while (pCurImage <= pRunEnd)
		{
		    ULONG_ACCESS(&(pCurImage[0])) = outWord;
		    ULONG_ACCESS(&(pCurImage[4])) = outWord;
		    pCurImage += 8;
		}
		pRunEnd += 8;
	
		    /* pCurImage either equals pRunEnd-4 or pRunEnd.
		     * The farthest forward we could have written is
		     * pRunEnd-8+4, or the word just before pRunEnd's word.
		     * That only happens if pCurImage now equals pRunEnd.
		     */
		if (pRunEnd != pCurImage)
		{
		    ULONG_ACCESS(pCurImage) = outWord;
		    pCurImage += 4;
		}
	    }
	}

	    /* At this point, either the code for strings of
	     * literals or the code for replicating bytes
	     * has left the last byte of the output string in outWord,
	     * properly aligned.  pCurImage points to the
	     * location where it will be stored.  nextVacantByte
	     * holds the offset of the last valid string byte in
	     * outWord.
	     */

	    /* Get the mask that will be used to discard crud
	     * that's beyond the right end of the run.  Even if
	     * we don't use it below, we'll use it to initialize
	     * leftMask for the next run.
	     */
	mask = rightMask[nextVacantByte];
    
	    /* pCurImage now equals pRunEnd.  If the run ends on a
	     * word boundary, write the word and point to the next
	     * one.  Otherwise, just mask off the crud beyond the
	     * end of the run.
	     */
	if (nextVacantByte == 3)
	{
	    ULONG_ACCESS(pCurImage) = outWord;
	    pCurImage += 4;
	    nextVacantByte = 0;
	    outWord = 0;     /* mask off everything in the new word */
	}
	else
	{
	    outWord &= mask;
	    nextVacantByte++;
	}
    }

	/* **************** END OF MAIN LOOP ************************* */

	/* We get here after happily completing our task.  The last bytes
	 * of the scan line may still be in outWord, so we should write
	 * them.  Only do this if nextVacantByte points to the middle of
	 * a word.  If it doesn't point to the middle of a word, pCurImage
	 * is pointing off the end of the scan line.  This is ok in a
	 * framed image, but some weirdo may want to use this on
	 * unframed images.
	 */
    if (nextVacantByte != 0)
	ULONG_ACCESS(pCurImage) = outWord;

	/* Clear out the saved word and byte offsets so we know we're done
	 * with the line.
	 */
    pPBDecompInfo->nextVacantByte = 0;
    pPBDecompInfo->imageWordOffset = 0;
    pPBDecompInfo->imageWord = 0;

	/* Set the state of the decompression buffer.  The next thing to
	 * to is read a bytecount.  Record the pointer to the next
	 * byte to be read.
	 */
    pPBDecompInfo->decompState = cPBDecompNewLine;
    *ppCompData = pCurCompData;

    NORMAL_RETURN;


	/* We get here when we've exhausted the compressed data buffer.
	 * We are guaranteed that we have NOT run out of scan lines
	 * at this point.  If we're still reading compressed data,
	 * we must not be done with scan line data.  We check to see
	 * if we're at the end of the scan line data at the end of
	 * each run inserted into the image.
	 * at this point, we expect the following to be true:
	 * 	pCurImage points to next image word to be written.
	 *	outWord holds the current partial output word
	 *	nextVacantByte holds the offset of the next vacant byte
	 *		in outWord.
	 *	pPBDecompInfo->decompState holds the startup state.
	 *	byteCount holds the number that would have just been
	 *		read from the image to properly process the
	 *		subsequent data if the state is
	 *		cPBDecompReadData.  I.E., if we're
	 *		supposed to replicate bytes, byteCount is negative.
	 *		If we run out of literal bytes, byteCount says
	 *		how many are left - 1.
	 */
pbdbsCompExhausted:
    pPBDecompInfo->imageWord = outWord;
    pPBDecompInfo->nextVacantByte = nextVacantByte;
    pPBDecompInfo->imageWordOffset = (UInt32)pCurImage - (UInt32)pImage;
    pPBDecompInfo->runByteCount = byteCount;
    *ppCompData = pCompDataEnd;

    NORMAL_RETURN;
}


/*

 $Log: mipackd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.6  1998/01/13 23:53:01  ddavies
 mipackd.c -> /product/ipcore/ipshared/src/RCS/mipackd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.5  1997/07/11 01:49:53  ddavies
 mipackd.c -> /product/ipcore/ipshared/src/RCS/mipackd.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.4  1995/12/13 02:50:57  ejaquith
 mipackd.c -> /product/ipcore/ipshared/src/RCS/mipackd.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1995/05/19  02:53:31  ddavies
 * mipackd.c -> /product/ipcore/ipshared/src/RCS/mipackd.c,v
 *
 * ip_getPackBitsDecompProc no longer returns a meaningless error code.
 * Fix some obsolete comments.
 *
 * Revision 1.2  1995/05/10  02:03:54  ddavies
 * mipackd.c -> /product/ipcore/ipshared/src/RCS/mipackd.c,v
 *
 * Tag constant 0xFFFFFFFF with UL in an attempt to stop acc from complaining
 * about shifting it right.
 *
 * Revision 1.1  1995/05/09  03:00:04  ddavies
 * Initial revision
 *

 */


