/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mipntrun.c
 *
 *    This file provides functions to paint runs into one scan line of
 *	a bitmap.
 *
 *    ip_paintRunsLine() 	Paint an array of runs into a scan line
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "lmask.prv"
#include "imageref.pub"
#include "memory.pub"

/* prototyes */
#include "shrcc.pub"
#include "shrcc.prv"

#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: mipntrun.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

static UInt32   rightMask[32] =
{   0xFFFFFFFF, 0x7FFFFFFF, 0x3FFFFFFF, 0x1FFFFFFF,
    0x0FFFFFFF, 0x07FFFFFF, 0x03FFFFFF, 0x01FFFFFF,
    0x00FFFFFF, 0x007FFFFF, 0x003FFFFF, 0x001FFFFF,
    0x000FFFFF, 0x0007FFFF, 0x0003FFFF, 0x0001FFFF,
    0x0000FFFF, 0x00007FFF, 0x00003FFF, 0x00001FFF,
    0x00000FFF, 0x000007FF, 0x000003FF, 0x000001FF,
    0x000000FF, 0x0000007F, 0x0000003F, 0x0000001F,
    0x0000000F, 0x00000007, 0x00000003, 0x00000001 };

/* PUBLIC */
/***********************************************************************
 * ip_paintRunsLine ()
 *
 *   This function paints the runs specified by an array of Run32 structures
 *   into one scan line of a binary image.
 *   This function used to assume that all runs fit into the scanline.
 *   After some decoding failures, it now checks to see if they do.
 *
 * Arguments:
 *   pDest:		Pointer to left end of the destination bitmap,
 *			INSIDE frame.  First byte is word aligned.
 *   w:			width of scanline in pixels ("bits" in this case)
 *   pRuns:        	Pointer to first run in array of Run32s.
 *   pEndRun:		Pointer to run beyond last valid run in array
 *   combineResults:	Boolean.  If cTrue, runs should be ORed into
 *			the original image.  Otherwise, the original
 *			image is assumed to be already clear.
 *
 * Returns:
 *   ia_successful:    all was well
 *   ia_decodeFailure
 *
 ***********************************************************************/

Int32 CDECL 
ip_paintRunsLine (
    IPState    *pIPState,
    UInt8      *pDest,
    Int32	w,
    Run32      *pRuns,
    Run32      *pEndRun,
    Bool	combineResults)
{
    PROLOGUE;
    Int32	leftPos;		/* Left end of current run */
    Int32	rightPos;		/* right end of current run */
    UInt32	oWord;			/* word in which output values
					 * are being formed */

	/* clear the accumulator for the beginning of the scan line */
    oWord = 0;

	/* this code is mostly copied from i_CCSet, which was in turn
	 * cribbed from ip_symbolToBitmap.  The biggest difference is
	 * the input format.  For minor efficiency reasons, we fill
	 * the line from right to left, not left to right.  So, point
	 * to the last valid run.
	 */
    pEndRun--;

	/* To save a little time in the inner loops, we want pDest to
	 * point directly to the words we want to write, instead of
	 * writing to pDest[rightPos / 8].  Therefore, we want pDest
	 * to start by pointing to the word containing the right
	 * end of the first run.
	 */
    if (pEndRun >= pRuns)
	pDest = &(pDest[(pEndRun->runRight & ~31) / 8]);

	/* For the sake of efficiency, we have two loops that
	 * go through all the runs.  One simply writes runs
	 * into the image, skipping all the image words that
	 * don't have runs in them, and the other ORs the
	 * runs with the original contents of the image.
	 */
    if (!combineResults)
    {
	    /* Just do it the fast way */
	while (pEndRun >= pRuns)
	{
	    leftPos = pEndRun->runLeft;
	    rightPos = pEndRun->runRight;

		/* make sure we're still in the scanline.  This
		 * test catches both numbers that are too large and
		 * negative coords (which appear as big pos coords.
		 */
	    if ( ((UInt32) leftPos >= (UInt32) w) ||
		 ((UInt32) rightPos >= (UInt32) w) )
		return ia_decodeFailure;
    
	    oWord |= leftMask[rightPos & 31];	/* fill in left side of
						     * the word, leaving out
						     * only the bits to the
						     * right of the right end
						     * of the run. */
	    rightPos &= ~31;			/* point to beginning of
						     * current word */
    
		/* See if this run crosses a word boundary.  If so,
		 * we need to write the current contents of oWord
		 * and fill it with all 1's for any subsequent
		 * words.
		 */
	    if (leftPos < rightPos)
	    {
		ULONG_ACCESS(pDest) = oWord;
    
		oWord = 0xFFFFFFFF;
		rightPos -= 32;
		pDest -= 4;
    
		/* Fill in the intermediate words, stopping before
		 * we get to the word containing the left endpoint
		 * The test works because rightPos always points to the
		 * beginning of a word. 
		 */
		while (leftPos < rightPos)
		{
		    ULONG_ACCESS(pDest) = oWord;
		    rightPos -= 32;
		    pDest -= 4;
		}
	    }
    
		/* We are now in the word containing the left end of the
		 * run.  Clear out all bits to the left of the left end.
		 */
	    oWord &= rightMask[leftPos & 31];
    
		/* Should we write oWord to memory now?  This should
		 * happen if there is no next run to the left or if
		 * the right end of that next run is in another word.
		 */
	    pEndRun--;
	    if ((pEndRun < pRuns) || (pEndRun->runRight < rightPos))
	    {
		ULONG_ACCESS(pDest) = oWord;
		oWord = 0;
		pDest -= (rightPos - (pEndRun->runRight & ~31)) / 8;
	    }
	}	/* end of loop that writes filled words */
    }
    else	/* OR the decompressed words into the image */
    {
	    /* Just do it the fast way */
	while (pEndRun >= pRuns)
	{
	    leftPos = pEndRun->runLeft;
	    rightPos = pEndRun->runRight;
    
		/* make sure we're still in the scanline.  This
		 * test catches both numbers that are too large and
		 * negative coords (which appear as big pos coords.
		 */
	    if ( ((UInt32) leftPos >= (UInt32) w) ||
		 ((UInt32) rightPos >= (UInt32) w) )
		return ia_decodeFailure;
    
	    oWord |= leftMask[rightPos & 31];	/* fill in left side of
						     * the word, leaving out
						     * only the bits to the
						     * right of the right end
						     * of the run. */
	    rightPos &= ~31;			/* point to beginning of
						     * current word */
    
		/* See if this run crosses a word boundary.  If so,
		 * we need to write the current contents of oWord
		 * and fill it with all 1's for any subsequent
		 * words.
		 */
	    if (leftPos < rightPos)
	    {
		ULONG_ACCESS(pDest) |= oWord;
    
		oWord = 0xFFFFFFFF;
		rightPos -= 32;
		pDest -= 4;
    
		/* Fill in the intermediate words, stopping before
		 * we get to the word containing the left endpoint
		 * The test works because rightPos always points to the
		 * beginning of a word. 
		 */
		while (leftPos < rightPos)
		{
		    ULONG_ACCESS(pDest) |= oWord;
		    rightPos -= 32;
		    pDest -= 4;
		}
	    }
    
		/* We are now in the word containing the left end of the
		 * run.  Clear out all bits to the left of the left end.
		 */
	    oWord &= rightMask[leftPos & 31];
    
		/* Should we write oWord to memory now?  This should
		 * happen if there is no next run to the left or if
		 * the right end of that next run is in another word.
		 */
	    pEndRun--;
	    if ((pEndRun < pRuns) || (pEndRun->runRight < rightPos))
	    {
		ULONG_ACCESS(pDest) |= oWord;
		oWord = 0;
		pDest -= (rightPos - (pEndRun->runRight & ~31)) / 8;
	    }
	}	/* end of loop that writes filled words */
    }


    NORMAL_RETURN;
}

/*

 $Log: mipntrun.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.5  1998/01/13 23:53:02  ddavies
 mipntrun.c -> /product/ipcore/ipshared/src/RCS/mipntrun.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1997/08/23 01:22:09  ddavies
 ddavies on Fri Aug 22 18:21:29 PDT 1997

 mipntrun.c -> /product/ipcore/ipshared/src/RCS/mipntrun.c,v

 Put in checks to see if we're about to write off the edge of the image.
 This can happen as the result of a decoding error.

 Revision 1.3  1996/07/19 20:23:20  ddavies
 ddavies on Fri Jul 19 13:23:10 PDT 1996

 mipntrun.c -> /product/ipcore/ipshared/src/RCS/mipntrun.c,v

 Fix WATCOM whining about unused parameters.

 Revision 1.2  1996/07/19 00:27:45  ddavies
 mipntrun.c -> /product/ipcore/ipshared/src/RCS/mipntrun.c,v

 Change clearLine arg to combineResults arg.  When combineResults is set,
 the runs are ORed into the original image.  When combineResults is clear,
 we assume that the original image was all white.  This is implemented
 by duplicating the main loop, one version replaces and one ORs.

 Revision 1.1  1996/04/25 18:46:53  ddavies
 Initial revision


 */


