/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miproctp.c
 *
 *  Private routines:
 *	ip_getLongestAccess():	C implementation of routine that
 *				returns the length in bytes of the longest
 *				memory access this processor will
 *				normally make.
 */


#include <stdio.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"

    /* prototypes */
#include "shrpixr.pub"
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miproctp.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");


/* All processors are happiest when accessing n-byte quantites on n-byte
 * boundaries.  This routine returns the length of the longest access
 * this processor will normally make.  The Sun SPARC's can make double
 * work accesses, but hardly ever do, so we just say 4 bytes for them.
 */

/* Define some inline assembly for the MMX opcode EMMS.  This
 * is the opcode that must be executed at the end of every
 * block of MMX code.  The wacky WATCOM compiler wants to define
 * sequences of assembly code as procedures, so we'll do that
 * for everyone.
 */
#ifdef _WATCOMC_		/* WATCOM */
void EMMSOpcode(void);
#pragma aux EMMSOpcode = 0xFF 0x77
#elif defined(_MSC_VER)		/* MS Visual C/C++ */
    /* The definition of STATUS_ILLEGAL_INSTRUCTION was stolen from
     * the winnt.h include file.  Including the entire file (or
     * windef.h which includes it) causes no end of error messages.
     */
#define STATUS_ILLEGAL_INSTRUCTION       ((UInt32)0xC000001DL)
#include <excpt.h>
__inline void EMMSOpcode(void) {__asm _emit 0x0F __asm _emit 0x77}
#else				/* everyone else */
void EMMSOpcode(void);
void EMMSOpcode(void) {} 
#endif

/* This is a private procedure */
/* PRIVATE */
/**************************************************************************
 *  ip_getLongestAccess()
 *	This procedure determines the length of the longest memory access
 *	commonly used on the current processor.
 *
 *	pLongestAccess:	Pointer to the returned number of bytes in the
 *			longest type of memory access this processor
 *			normally makes.
 *
 **************************************************************************/
void CDECL 
ip_getLongestAccess (
    IPState    *pIPState,
    Int32      *pLongestAccess)
{
    PROLOGUE;
    Int32	longestAccess;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* First, see if long accesses are enabled */
    if (pIPState->enableLongAccesses == 0)
	longestAccess = 4;
    else
    {
	    /* If this isn't a PC, then the longest access is 4 bytes. */
#ifndef _386_ARCH_
	longestAccess = 4;
#elif defined(_UNIX_)
	longestAccess = 4;
#elif !defined(_WIN32)
	    /* This is a PC running Windows 3.1, so ask the assembly code
	     * to do the work.
	     * We do this for two reasons: 1) reading Control Register 0
	     * is not a priviledged operation in Windows 3.1, and 2)
	     * there is no error handling mechanism in Win31 that lets
	     * is just try excuting an MMX opcode to see whether it works.
	     * Therefore, the only way to figure out whether MMX opcodes
	     * are allowed is to detect the type of processor, whether
	     * MMX opcodes are available (via the CPUID opcode) and
	     * whether the floating point emulation bit is enabled
	     * (via reading CR0).  That's all assembly language stuff.
	     */
	ip_getLongestWin31Access (pIPState, &longestAccess);
#else
	    /* This is a Win95 or WinNT machine.  We don't know which
	     * one since the same code is compiled for both.  I don't
	     * know how to ask which one we're on.
	     * So what?  Well, if you execute an MMX opcode while the
	     * floating point emulation bit in Control Register 0 (CRO.EM)
	     * is set, you get an undefined/invalid/illegal opcode trap.
	     * You also get this trap if the current processor doesn't
	     * have MMX support.
	     * Reading Control Register 0 to check the EM bit
	     * is fine in Win95 and Window 3.1, but results
	     * in a trap whining about privileged operations in NT.
	     * BUT, we know that all code compiled for Win95 and WinNT
	     * is compiled using the Microsoft Visual C/C++ compiler.
	     * They've implemented a spiffy error handling mechanism.
	     * We use it to try executing an MMX opcode, handling the
	     * error if we're not allowed to do that.
	     */

	    /* See if type of processor we're on allows MMX opcodes.
	     * We do this by setting up an exception handler,
	     * then trying to execute an MMX opcode.  This is the
	     * EMMS opcode, which is the one you should execute
	     * at the end of each segment of MMX opcodes.  Since
	     * this *is* a segment of MMX opcodes, we need an EMMS
	     * anyway.
	     */
	longestAccess = 8;	/* assume MMX opcodes available */
	__try
	{
	    EMMSOpcode();
	}
	__except (GetExceptionCode() == STATUS_ILLEGAL_INSTRUCTION)
	{
	    longestAccess = 4;	/* there are no MMX opcodes available */
	}
#endif
    }
	/* somehow, we've got the longest access, so set it and return */
    *pLongestAccess = longestAccess;
}


/* PRIVATE */
/**************************************************************************
 *  ip_checkImageAccess()
 *
 *    This routine checks the source and destination pointers and source
 *    and destination bytes/line values and the processor type to determine
 *    the longest legal memory access allowed for operations on this
 *    source and destination.
 *
 *	pS:		Pointer to src image inside frame.
 *	sBpl:		Source image bytes/line
 *	pD:		Pointer to dest image inside frame.
 *      dBpl:		Dest image bytes/line
 *	pLongestAccess:	Pointer to the returned number of bytes in the
 *			longest type of memory access that is legal
 *			for the source and destination image and this
 *			processor.
 *
 **************************************************************************/
void CDECL 
ip_checkImageAccess (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    Int32      *pLongestAccess)
{
    PROLOGUE;
    Int32	longestAccess;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* First, see if we can fetch 8-byte memory chunks */
    ip_getLongestAccess (pIPState, &longestAccess);

	/* if so, see if the addresses and bytes/line are all multiples
	 * of 8.
	 */
    if (longestAccess == 8)
    {
	if ((((UInt32)pS | sBpl | (UInt32)pD | dBpl) & 0x7) != 0)
	    longestAccess = 4;		/* nope, some address or bpl was
					 * only a multiple of 4 (or less!)
					 */
    }

    *pLongestAccess = longestAccess;
}


/* PRIVATE */
/**************************************************************************
 *  ip_initLongOps()
 *
 *    This routine does whatever is necessary to cause the current processor
 *    to be able to execute long (64-bit or longer) operations.  This may
 *    be nothing.
 *
 *
 **************************************************************************/
void CDECL 
ip_initLongOps (
    IPState    *pIPState)
{
    PROLOGUE;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* None of our processor require an entry sequence for
	 * long operations.
	 */
}


/* PRIVATE */
/**************************************************************************
 *  ip_finishLongOps()
 *
 *    This routine does whatever is necessary to cause the current processor
 *    to finish executing long (64-bit or longer) operations.  This may
 *    be nothing.
 *
 *
 **************************************************************************/
void CDECL 
ip_finishLongOps (
    IPState    *pIPState)
{
	/* MMX machines require an EMMS opcode.  This routine isn't called
	 * if MMX is not present, so we can always execute it.
	 */
    PROLOGUE;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    EMMSOpcode();
}


/*

$Log: miproctp.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.5  1998/01/21 21:49:06  ddavies
miproctp.c -> /product/ipcore/ipshared/src/RCS/miproctp.c,v

Change reference to ipVar_enableLongAccess to a reference to
pIPState->enableLongAccesses.

Revision 1.4  1998/01/13 23:53:03  ddavies
miproctp.c -> /product/ipcore/ipshared/src/RCS/miproctp.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1997/02/05 02:38:51  jcraig
miproctp.c -> /product/ipcore/ipshared/src/RCS/miproctp.c,v

in another one of those temporary hack things, don't use mmx on
a unix pc.  i'll submit a bug report so that we don't forget that
this needs to check instead of always saying no.

Revision 1.2  1996/10/24 19:52:19  ddavies
miproctp.c -> /product/ipcore/ipshared/src/RCS/miproctp.c,v

Modify ip_getLongestAccess so that it deals with non-PC, Windows 3.1 and
Win95/WinNT separately.  For non-x86 architectures, it always returns 4 bytes
as the longest access.  This will change with VIS if we ever use that.
For Windows 3.1 machines, it uses assembly code to figure out the processor
type, decide if it has MMX hardware and see if the floating point emulation
bit in CR0 is set.  CR0 can be read in protection-free Windows 3.1.
In Win95/WinNT it makes use of the structured exception mechanism.  This
also requires use of the Microsoft Visual C/C++ compiler.  It just executes
an I
ooops, an EMMS opcode and fields the exception if MMX can't be used for
any reason.  EMMS is the MMX opcode you have to execute when you're done
using MMX for awhile and floating point ops may follow.

Revision 1.1  1996/09/24 02:44:33  ddavies
Initial revision


*/


