/*****************************************************************
 *  Copyright (c) 1988, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miredm.c: Dummy C implementations of reduction primitives
 *
 *	     Used to see if MMX stuff is linked
 *		Bool	ip_longOpScaleLibLinked()
 *
 *	     Dummy procedures (should never be called)
 *		void	ip_reduceT1LineLong()
 *		void	ip_reduceT2LineLong()
 *		void	ip_reduceT3LineLong()
 *		void	ip_reduceT4LineLong()
 *
 *		void	ip_reduce4T1LineLong()
 *		void	ip_reduce4T4LineLong()
 *
 *		void	ip_reduce8T1LineLong()
 *		void	ip_reduce8T4LineLong()
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"

/* prototypes */
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miredm.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* Some routines are supplied only so the linker will be happy, but should
 * never be called.  When MMX code is present, the versions of the 
 * long routines implemented in assembly language will be called.  When
 * no assembly code is present, this file is linked in.  The calling
 * routine will see that there is no assembly code and will call the
 * ordinary reduction routines.  The only reason there routines exist
 * is to keep the linker happy.
 */

/* PRIVATE */
/*************************************************************************
 *  ip_longOpScaleLibLinked()
 *		There are things we'd really not do if routines capable
 *		of long (64 bit or longer) are linked, like avoid
 *		making superflous tables.  This routine returns a
 *		boolean which is set to cTrue if the
 *		long memory access libraries are linked.
 *
 *	Input:
 *		None
 *	Output:
 *		cTrue if fast libraries are linked, cFalse otherwise.
 *************************************************************************/
Bool  CDECL 
ip_longOpScaleLibLinked (
    IPState    *pIPState)
{
	/* Clearly the fast libs aren't here since we're here. */
    PROLOGUE;
	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    return cFalse;
}


/* PRIVATE */
/*************************************************************************
 *  ip_reduceT1LineLong()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the four corresponding pixels in the src 
 *		image are ON.  The four pixels are in a 2x2 square.
 *		This differs from ip_reduceT1Line in that is processes
 *		the image in bigger chunks.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab2To1Permute =  table that converts a byte containing
 *			AEBFCGDH -> xxxx xxxx xxxx ABCD xxxx xxxx xxxx EFGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduceT1LineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2To1Permute)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
    tab2To1Permute = tab2To1Permute;
}


/* PRIVATE */
/*************************************************************************
 *  ip_reduceT2LineLong()
 *		The reduced image line has an ON pixel whenever at least
 *              2 of the four corresponding pixels in the src image are ON.
 *		The four pixels are in a 2x2 square.
 *
 *		There are several configuration of 2 bits in a 2x2 square.
 *		We AND the two vertically displaced samples together, then
 *		OR the result with itself shifted by 1 to detect the cases
 *		when the 1 bits are in the same column.
 *		We OR the two samples, then AND the result with a shifted
 *		version of itself to detect all cases where the two samples
 *		were in different columns.
 *		We OR these two results together to merge the cases. 
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab2To1Permute =  table that converts a byte containing
 *			AEBFCGDH -> xxxx xxxx xxxx ABCD xxxx xxxx xxxx EFGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduceT2LineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2To1Permute)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
    tab2To1Permute = tab2To1Permute;
}


/* PRIVATE */
/*************************************************************************
 * ip_reduceT3LineLong()
 *		The reduced image line has an ON pixel whenever at least
 *              3 of the four corresponding pixels in the src image are ON.
 *		The four pixels are in a 2x2 square.
 *
 *		There are several configuration of 3 bits in a 2x2 square.
 *		We AND the two vertically displaced samples together, then
 *		OR the result with itself shifted by 1 to detect the cases
 *		when 2 bits are in the same column.
 *		We OR the two samples, then AND the result with a shifted
 *		version of itself to detect all cases where the two samples
 *		were in different columns.
 *		In order to have at least 3 bits in a 2x2 square, we must have
 *		bit in both the same column and the same row, so we AND the
 *		results of these test to get the result. 
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab2To1Permute =  table that converts a byte containing
 *			AEBFCGDH -> xxxx xxxx xxxx ABCD xxxx xxxx xxxx EFGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduceT3LineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2To1Permute)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
    tab2To1Permute = tab2To1Permute;
}


/* PRIVATE */
/*************************************************************************
 *  ip_reduceT4LineLong()
 *		The reduced image line has an ON pixel whenever all
 *              of the four corresponding pixels in the src image are ON.
 *		The four pixels are in a 2x2 square.
 *
 *		This is basically the same as reduceT1 except we AND the 
 *		pixels nstead of ORing them.  Any 0 will force the result 
 *		to be 0.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab2To1Permute =  table that converts a byte containing
 *			AEBFCGDH -> xxxx xxxx xxxx ABCD xxxx xxxx xxxx EFGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduceT4LineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2To1Permute)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
    tab2To1Permute = tab2To1Permute;
}


/* PRIVATE */
/*************************************************************************
 *  ip_reduce4T1LineLong()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the 16 corresponding pixels in the src 
 *		image are ON.  The 16 pixels are in a 4x4 square.
 *		This differs from ip_reduce4T1Line in that is processes
 *		the image in bigger chunks.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab4To1Permute =  table that converts a byte containing
 *			ACEGBDFH -> xxxx xxAB xxxx xxCD xxxx xxEF xxxx xxGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduce4T1LineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab4To1Permute)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
    tab4To1Permute = tab4To1Permute;
}


/* PRIVATE */
/*************************************************************************
 *  ip_reduce4T4LineLong()
 *		The reduced image line has an ON pixel whenever
 *		all of the 16 corresponding pixels in the src 
 *		image are ON.  The 16 pixels are in a 4x4 square.
 *		This differs from ip_reduce4T4Line in that is processes
 *		the image in bigger chunks.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab4To1Permute =  table that converts a byte containing
 *			ACEGBDFH -> xxxx xxAB xxxx xxCD xxxx xxEF xxxx xxGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduce4T4LineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab4To1Permute)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
    tab4To1Permute = tab4To1Permute;
}


/* PRIVATE */
/*************************************************************************
 *  ip_reduce8T1LineLong()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the 64 corresponding pixels in the src 
 *		image are ON.  The 64 pixels are in a 8x8 square.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab8To1Permute =  table that converts a byte containing
 *			ACEGBDFH -> xxxx xxAB xxxx xxCD xxxx xxEF xxxx xxGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduce8T1LineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab8To1Permute)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
    tab8To1Permute = tab8To1Permute;
}


/* PRIVATE */
/*************************************************************************
 *  ip_reduce8T4LineLong()
 *		The reduced image line has an ON pixel whenever
 *		all of the 64 corresponding pixels in the src 
 *		image are ON.  The 64 pixels are in a 8x8 square.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab8To1Permute =  table that converts a byte containing
 *			ACEGBDFH -> xxxx xxAB xxxx xxCD xxxx xxEF xxxx xxGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduce8T4LineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab8To1Permute)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    sBpl = sBpl;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
    tab8To1Permute = tab8To1Permute;
}




/*

$Log: miredm.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.2  1998/01/13 23:53:06  ddavies
miredm.c -> /product/ipcore/ipshared/src/RCS/miredm.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.1  1996/10/24 19:14:55  ddavies
Initial revision


 */

