/*****************************************************************
 *  Copyright (c) 1988, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  mireduce.c: C implementations of reduction primitives
 *
 *          2x threshold reduction:
 *		void	ip_reduceT1Line()
 *		void	ip_reduceT2Line()
 *		void	ip_reduceT3Line()
 *		void	ip_reduceT4Line()
 *
 *	    Single-pass reductions for Thresholds 1 and 4:
 *		void	ip_reduce4T1Line()
 *		void	ip_reduce4T4Line()
 *		void	ip_reduce8Line()
 *		UInt32	ip_reduce8xT1()
 *		UInt32	ip_reduce8xT4()
 *		void	ip_reduce16Line()
 *		UInt32	ip_reduce16xT1()
 *		UInt32	ip_reduce16xT4()
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"

	/* prototypes */
#include "shrscal.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: mireduce.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* This file contains C implementations of threshold reduction primitives.
 * Each of these routines produces one reduced scan line.  It is our hope
 * that these routines can be used in banded and tiled image routines
 * as well as in normal full-image operations.
 *
 * The first four routines, ip_reduceT1 through ip_reduceT4, do 2x reduction
 * with thresholds of 1, 2, 3 and 4 respectively.  What does this mean?
 * 2x reduction means that each 2x2 square of source pixels maps to a single
 * destination pixel, so the dest image is half as wide and half as tall.
 * Reduction done with threshold n means that a destination pixel is set
 * if at least n of the pixels in the associated 2x2 source square are set.
 *
 * Thresholds 1 and 4 are relatively easy to compute.  To compute threshold
 * 1, we simply OR all the pixels in each square together.  To compute
 * threshold 4, we AND them.  Thresholds 2 and 3 are trickier.  To find
 * a threshold of 2, we need to detect all cases where there were pixels
 * in both columns OR where there were pixels in both rows of the 2x2
 * squares.  To detect a threshold of 2, we need to detect all cases
 * where there were pixels in both columns AND in both rows of the 2x2 squares.
 *
 * Since the cases of threshold = 1 and threshold = 4 are easy to compute,
 * we also provide accelerated cases of 4x, 8x and 16x reduction for those
 * thresholds.
 *
 */

#define cLSBitMask2bpp	0x55555555
#define cMSBitMask2bpp	0xAAAAAAAA
#define cBit3Mask4bpp	0x88888888
#define cBit2Mask4bpp	0x44444444
#define cBit1Mask4bpp	0x22222222
#define cBit0Mask4bpp	0x11111111
#define cMSBitMask8bpp	0x80808080
#define cMSBitMask16bpp 0x80008000
#define c16xBitMask	0xC0000000
#define	cByteMask	0x000000FF

/* PRIVATE */
/*************************************************************************
 *  ip_reduceT1Line()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the four corresponding pixels in the src 
 *		image are ON.  The four pixels are in a 2x2 square.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab2To1Permute =  table that converts a byte containing
 *			AEBFCGDH -> xxxx xxxx xxxx ABCD xxxx xxxx xxxx EFGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduceT1Line (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2To1Permute)
{
    PROLOGUE;
    UInt32	 oword;			/* permuted output word being
					 * assembled */
    UInt32	 data;			/* intermediate data */
    UInt8	*pSLineBelow;		/* ptr to src line below pS */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 2 source words.  
	 */
    pSLineBelow = pS + sBpl;
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;
	data = ULONG_ACCESS(&pS[srcLimit]) |
		ULONG_ACCESS(&pSLineBelow[srcLimit]);	/* combine rows */
	oword |= (data | (data << 1)) & cMSBitMask2bpp;	/* combine cols */
    }

	/* unscramble the bits in oword */
    data = tab2To1Permute[(oword & cByteMask)];		/* [19:16], [3:0] */
    oword >>= 8;
    data |= tab2To1Permute[(oword & cByteMask)] << 4;	/* [23:20],[7:4] */
    oword >>= 8;
    data |= tab2To1Permute[(oword & cByteMask)] << 8;	/* [27:24], [11:8] */
    oword >>= 8;
    data |= tab2To1Permute[oword] << 12;		/* [31:28], [15:12] */
    ULONG_ACCESS(&pD[dstOffset]) = data;

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	data = ULONG_ACCESS(&pS[srcOffset]) |
		ULONG_ACCESS(&pSLineBelow[srcOffset]);	/* combine rows */
	oword = (data | (data >> 1)) & cLSBitMask2bpp;	/* combine cols */
	srcOffset -= 4;
	data = ULONG_ACCESS(&pS[srcOffset]) |
		ULONG_ACCESS(&pSLineBelow[srcOffset]);	/* combine rows */
	oword |= (data | (data << 1)) & cMSBitMask2bpp;	/* combine cols */
	srcOffset -= 4;

	data = tab2To1Permute[(oword & cByteMask)];	/* [19:16], [3:0] */
	oword >>= 8;
	data |= tab2To1Permute[(oword & cByteMask)] << 4; /* [23:20],[7:4] */
	oword >>= 8;
	data |= tab2To1Permute[(oword & cByteMask)] << 8; /* [27:24], [11:8] */
	oword >>= 8;
	data |= tab2To1Permute[oword] << 12;		/* [31:28], [15:12] */
	ULONG_ACCESS(&pD[dstOffset]) = data;
    }
}




/* PRIVATE */
/*************************************************************************
 * ip_reduceT2Line() 
 *		The reduced image line has an ON pixel whenever at least
 *              2 of the four corresponding pixels in the src image are ON.
 *		The four pixels are in a 2x2 square.
 *
 *		There are several configuration of 2 bits in a 2x2 square.
 *		We AND the two vertically displaced samples together, then
 *		OR the result with itself shifted by 1 to detect the cases
 *		when the 1 bits are in the same column.
 *		We OR the two samples, then AND the result with a shifted
 *		version of itself to detect all cases where the two samples
 *		were in different columns.
 *		We OR these two results together to merge the cases. 
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab2To1Permute =  table that converts a byte containing
 *			AEBFCGDH -> xxxx xxxx xxxx ABCD xxxx xxxx xxxx EFGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduceT2Line (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2To1Permute)
{
    PROLOGUE;
    UInt32	 oword;			/* permuted output word being
					 * assembled */
    UInt32	 data, inputU, inputL;	/* scratch vars */
    UInt8	*pSLineBelow;		/* ptr to src line below pS */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 2 source words.  
	 */
    pSLineBelow = pS + sBpl;
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;
	inputU = ULONG_ACCESS(&pS[srcLimit]);
	inputL = ULONG_ACCESS(&pSLineBelow[srcLimit]);
	data = inputU | inputL;			/* OR data vertically */
	data = data & (data << 1);		/* detect bits in diff cols*/
	inputL = inputU & inputL;		/* AND the data vertically */
	inputL = inputL | (inputL << 1);	/* detect bits in same col */
	data = data | inputL;			/* merge the cases */
	oword |= data & cMSBitMask2bpp;		/* combine wiht prev bits */
    }

	/* unscramble the bits in oword */
    data = tab2To1Permute[(oword & cByteMask)];		/* [19:16], [3:0] */
    oword >>= 8;
    data |= tab2To1Permute[(oword & cByteMask)] << 4;	/* [23:20],[7:4] */
    oword >>= 8;
    data |= tab2To1Permute[(oword & cByteMask)] << 8;	/* [27:24], [11:8] */
    oword >>= 8;
    data |= tab2To1Permute[oword] << 12;		/* [31:28], [15:12] */
    ULONG_ACCESS(&pD[dstOffset]) = data;

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	inputU = ULONG_ACCESS(&pS[srcOffset]);
	inputL = ULONG_ACCESS(&pSLineBelow[srcOffset]);
	data = inputU | inputL;			/* OR data vertically */
	data = data & (data >> 1);		/* detect bits in diff cols*/
	inputL = inputU & inputL;		/* AND the data vertically */
	inputL = inputL | (inputL >> 1);	/* detect bits in same col */
	data = data | inputL;			/* merge the cases */
	oword = data & cLSBitMask2bpp;		/* put the cases into
						 * LS bits of each pair */
	srcOffset -= 4;

	inputU = ULONG_ACCESS(&pS[srcOffset]);
	inputL = ULONG_ACCESS(&pSLineBelow[srcOffset]);
	data = inputU | inputL;			/* OR data vertically */
	data = data & (data << 1);		/* detect bits in diff cols*/
	inputL = inputU & inputL;		/* AND the data vertically */
	inputL = inputL | (inputL << 1);	/* detect bits in same col */
	data = data | inputL;			/* merge the cases */
	oword |= data & cMSBitMask2bpp;		/* insert the cases into
						 * MS bits of each pair */
	srcOffset -= 4;

	data = tab2To1Permute[(oword & cByteMask)];	/* [19:16], [3:0] */
	oword >>= 8;
	data |= tab2To1Permute[(oword & cByteMask)] << 4; /* [23:20],[7:4] */
	oword >>= 8;
	data |= tab2To1Permute[(oword & cByteMask)] << 8; /* [27:24], [11:8] */
	oword >>= 8;
	data |= tab2To1Permute[oword] << 12;		/* [31:28], [15:12] */
	ULONG_ACCESS(&pD[dstOffset]) = data;
    }
}


/* PRIVATE */
/*************************************************************************
 * ip_reduceT3Line()
 *		The reduced image line has an ON pixel whenever at least
 *              3 of the four corresponding pixels in the src image are ON.
 *		The four pixels are in a 2x2 square.
 *
 *		There are several configuration of 3 bits in a 2x2 square.
 *		We AND the two vertically displaced samples together, then
 *		OR the result with itself shifted by 1 to detect the cases
 *		when 2 bits are in the same column.
 *		We OR the two samples, then AND the result with a shifted
 *		version of itself to detect all cases where the two samples
 *		were in different columns.
 *		In order to have at least 3 bits in a 2x2 square, we must have
 *		bit in both the same column and the same row, so we AND the
 *		results of these test to get the result. 
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab2To1Permute =  table that converts a byte containing
 *			AEBFCGDH -> xxxx xxxx xxxx ABCD xxxx xxxx xxxx EFGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduceT3Line (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2To1Permute)
{
    PROLOGUE;
    UInt32	 oword;			/* permuted output word being
					 * assembled */
    UInt32	 data, inputU, inputL;	/* scratch vars */
    UInt8	*pSLineBelow;		/* ptr to src line below pS */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 2 source words.  
	 */
    pSLineBelow = pS + sBpl;
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;
	inputU = ULONG_ACCESS(&pS[srcLimit]);
	inputL = ULONG_ACCESS(&pSLineBelow[srcLimit]);
	data = inputU | inputL;			/* OR data vertically */
	data = data & (data << 1);		/* detect bits in diff cols*/
	inputL = inputU & inputL;		/* AND the data vertically */
	inputL = inputL | (inputL << 1);	/* detect bits in same col */
	data = data & inputL;			/* both must be true */
	oword |= data & cMSBitMask2bpp;		/* combine wiht prev bits */
    }

	/* unscramble the bits in oword */
    data = tab2To1Permute[(oword & cByteMask)];		/* [19:16], [3:0] */
    oword >>= 8;
    data |= tab2To1Permute[(oword & cByteMask)] << 4;	/* [23:20],[7:4] */
    oword >>= 8;
    data |= tab2To1Permute[(oword & cByteMask)] << 8;	/* [27:24], [11:8] */
    oword >>= 8;
    data |= tab2To1Permute[oword] << 12;		/* [31:28], [15:12] */
    ULONG_ACCESS(&pD[dstOffset]) = data;

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	inputU = ULONG_ACCESS(&pS[srcOffset]);
	inputL = ULONG_ACCESS(&pSLineBelow[srcOffset]);
	data = inputU | inputL;			/* OR data vertically */
	data = data & (data >> 1);		/* detect bits in diff cols*/
	inputL = inputU & inputL;		/* AND the data vertically */
	inputL = inputL | (inputL >> 1);	/* detect bits in same col */
	data = data & inputL;			/* both must be true */
	oword = data & cLSBitMask2bpp;		/* put the cases into
						 * LS bits of each pair */
	srcOffset -= 4;

	inputU = ULONG_ACCESS(&pS[srcOffset]);
	inputL = ULONG_ACCESS(&pSLineBelow[srcOffset]);
	data = inputU | inputL;			/* OR data vertically */
	data = data & (data << 1);		/* detect bits in diff cols*/
	inputL = inputU & inputL;		/* AND the data vertically */
	inputL = inputL | (inputL << 1);	/* detect bits in same col */
	data = data & inputL;			/* both must be true */
	oword |= data & cMSBitMask2bpp;		/* insert the cases into
						 * MS bits of each pair */
	srcOffset -= 4;

	data = tab2To1Permute[(oword & cByteMask)];	/* [19:16], [3:0] */
	oword >>= 8;
	data |= tab2To1Permute[(oword & cByteMask)] << 4; /* [23:20],[7:4] */
	oword >>= 8;
	data |= tab2To1Permute[(oword & cByteMask)] << 8; /* [27:24], [11:8] */
	oword >>= 8;
	data |= tab2To1Permute[oword] << 12;		/* [31:28], [15:12] */
	ULONG_ACCESS(&pD[dstOffset]) = data;
    }
}


/* PRIVATE */
/*************************************************************************
 *  ip_reduceT4Line()
 *		The reduced image line has an ON pixel whenever all
 *              of the four corresponding pixels in the src image are ON.
 *		The four pixels are in a 2x2 square.
 *
 *		This is basically the same as reduceT1 except we AND the 
 *		pixels nstead of ORing them.  Any 0 will force the result 
 *		to be 0.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab2To1Permute =  table that converts a byte containing
 *			AEBFCGDH -> xxxx xxxx xxxx ABCD xxxx xxxx xxxx EFGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduceT4Line (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab2To1Permute)
{
    PROLOGUE;
    UInt32	 oword;			/* permuted output word being
					 * assembled */
    UInt32	 data;			/* intermediate data */
    UInt8	*pSLineBelow;		/* ptr to src line below pS */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 2 source words.  
	 */
    pSLineBelow = pS + sBpl;
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;
	data = ULONG_ACCESS(&pS[srcLimit]) &
		ULONG_ACCESS(&pSLineBelow[srcLimit]);	/* combine rows */
	oword |= (data & (data << 1)) & cMSBitMask2bpp;	/* combine cols */
    }

	/* unscramble the bits in oword */
    data = tab2To1Permute[(oword & cByteMask)];		/* [19:16], [3:0] */
    oword >>= 8;
    data |= tab2To1Permute[(oword & cByteMask)] << 4;	/* [23:20],[7:4] */
    oword >>= 8;
    data |= tab2To1Permute[(oword & cByteMask)] << 8;	/* [27:24], [11:8] */
    oword >>= 8;
    data |= tab2To1Permute[oword] << 12;		/* [31:28], [15:12] */
    ULONG_ACCESS(&pD[dstOffset]) = data;

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	data = ULONG_ACCESS(&pS[srcOffset]) &
		ULONG_ACCESS(&pSLineBelow[srcOffset]);	/* combine rows */
	oword = (data & (data >> 1)) & cLSBitMask2bpp;	/* combine cols */
	srcOffset -= 4;
	data = ULONG_ACCESS(&pS[srcOffset]) &
		ULONG_ACCESS(&pSLineBelow[srcOffset]);	/* combine rows */
	oword |= (data & (data << 1)) & cMSBitMask2bpp;	/* combine cols */
	srcOffset -= 4;

	data = tab2To1Permute[(oword & cByteMask)];	/* [19:16], [3:0] */
	oword >>= 8;
	data |= tab2To1Permute[(oword & cByteMask)] << 4; /* [23:20],[7:4] */
	oword >>= 8;
	data |= tab2To1Permute[(oword & cByteMask)] << 8; /* [27:24], [11:8] */
	oword >>= 8;
	data |= tab2To1Permute[oword] << 12;		/* [31:28], [15:12] */
	ULONG_ACCESS(&pD[dstOffset]) = data;
    }
}


/*************************************************************************
 *  ip_reduce4T1Line()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the 16 corresponding pixels in the src 
 *		image are ON.  The four pixels are in a 4x4 square.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab4To1Permute =  table that converts a byte containing
 *			ACEGBDFH -> xxxx xxAB xxxx xxCD xxxx xxEF xxxx xxGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduce4T1Line (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab4To1Permute)
{
    PROLOGUE;
    UInt32	 oword, data;			/* permuted output word being
						 * assembled */
    UInt8	*pSDown1, *pSDown2, *pSDown3;	/* ptrs to src lines below pS */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 4 source words.  
	 */
    oword = 0;
    pSDown1 = pS + sBpl;
    pSDown2 = pSDown1 + sBpl;
    pSDown3 = pSDown2 + sBpl;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;	/* previous word wasn't the leftmost one
				 * in the group that makes up the last
				 * dest word, so shift the results to the right
				 * and look to the left in the source.
				 */
	    /* put output bits [31:24] into bits
	     * 31xxx 30xxx 29xxx 28xxx 27xxx 26xxx 25xxx 24xxx
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcLimit]);		/* row 0 */
	data |= ULONG_ACCESS(&pSDown1[srcLimit]);	/* row 1 */
	data |= ULONG_ACCESS(&pSDown2[srcLimit]);	/* row 2 */
	data |= ULONG_ACCESS(&pSDown3[srcLimit]);	/* row 3 */

	    /* combine the columns, adjacent ones first, then column pairs */
	data |= (data << 1);
	data |= (data << 2);
	oword |= data & cBit3Mask4bpp;
    }

	/* unscramble the bits in oword.  We use one table instead of
	 * 4 (one for each shift position) to save space and registers.
	 * First, insert bits [25:24 17:16 9:8 1:0] */
    data = tab4To1Permute[(oword & cByteMask)];
    oword >>= 8;
	/* insert bits [27:26 19:18 11:10 3:2] */
    data |= tab4To1Permute[(oword & cByteMask)] << 2;
    oword >>= 8;
	/* insert bits [29:28 21:20 13:12 5:4] */
    data |= tab4To1Permute[(oword & cByteMask)] << 4;
    oword >>= 8;
	/* insert bits [31:30 23:22 15:14 7:6] */
    data |= tab4To1Permute[oword] << 6;
    ULONG_ACCESS(&pD[dstOffset]) = data;


	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	    /* put output bits [7:0] into bits
	     * xxx7 xxx6 xxx5 xxx4 xxx3 xxx2 xxx1 xxx0
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcOffset]);		/* row 0 */
	data |= ULONG_ACCESS(&pSDown1[srcOffset]);	/* row 1 */
	data |= ULONG_ACCESS(&pSDown2[srcOffset]);	/* row 2 */
	data |= ULONG_ACCESS(&pSDown3[srcOffset]);	/* row 3 */

	    /* combine the columns, adjacent ones first, then column pairs */
	data |= (data >> 1);
	data |= (data >> 2);
	oword = data & cBit0Mask4bpp;
	srcOffset -= 4;

	    /* put output bits [15:8] into bits
	     * xx15x xx14x xx13x xx12x xx11x xx10x xx9x xx8x
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcOffset]);		/* row 0 */
	data |= ULONG_ACCESS(&pSDown1[srcOffset]);	/* row 1 */
	data |= ULONG_ACCESS(&pSDown2[srcOffset]);	/* row 2 */
	data |= ULONG_ACCESS(&pSDown3[srcOffset]);	/* row 3 */
	    /* combine the columns, adjacent ones first, then column pairs */
	data |= (data << 1);
	data |= (data >> 2);
	oword |= data & cBit1Mask4bpp;
	srcOffset -= 4;


	    /* put output bits [23:16] into bits
	     * x23xx x22xx x21xx x20xx x19xx x18xx x17xx x16xx
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcOffset]);		/* row 0 */
	data |= ULONG_ACCESS(&pSDown1[srcOffset]);	/* row 1 */
	data |= ULONG_ACCESS(&pSDown2[srcOffset]);	/* row 2 */
	data |= ULONG_ACCESS(&pSDown3[srcOffset]);	/* row 3 */
	    /* combine the columns, adjacent ones first, then column pairs */
	data |= (data >> 1);
	data |= (data << 2);
	oword |= data & cBit2Mask4bpp;
	srcOffset -= 4;


	    /* put output bits [31:24] into bits
	     * 31xxx 30xxx 29xxx 28xxx 27xxx 26xxx 25xxx 24xxx
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcOffset]);		/* row 0 */
	data |= ULONG_ACCESS(&pSDown1[srcOffset]);	/* row 1 */
	data |= ULONG_ACCESS(&pSDown2[srcOffset]);	/* row 2 */
	data |= ULONG_ACCESS(&pSDown3[srcOffset]);	/* row 3 */
	    /* combine the columns, adjacent ones first, then column pairs */
	data |= (data << 1);
	data |= (data << 2);
	oword |= data & cBit3Mask4bpp;
	srcOffset -= 4;

	    /* unscramble the bits in oword.  We use one table instead of
	     * 4 (one for each shift position) to save space and registers.
	     * First, insert bits [25:24 17:16 9:8 1:0] */
	data = tab4To1Permute[(oword & cByteMask)];
	oword >>= 8;
	    /* insert bits [27:26 19:18 11:10 3:2] */
	data |= tab4To1Permute[(oword & cByteMask)] << 2;
	oword >>= 8;
	    /* insert bits [29:28 21:20 13:12 5:4] */
	data |= tab4To1Permute[(oword & cByteMask)] << 4;
	oword >>= 8;
	    /* insert bits [31:30 23:22 15:14 7:6] */
	data |= tab4To1Permute[oword] << 6;
	ULONG_ACCESS(&pD[dstOffset]) = data;
    }
}


/*************************************************************************
 *  ip_reduce4T4Line()
 *		The reduced image line has an ON pixel whenever
 *		all of the 16 corresponding pixels in the src 
 *		image are ON.  The four pixels are in a 4x4 square.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab4To1Permute =  table that converts a byte containing
 *			ACEGBDFH -> xxxx xxAB xxxx xxCD xxxx xxEF xxxx xxGH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *************************************************************************/
void  CDECL 
ip_reduce4T4Line (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab4To1Permute)
{
    PROLOGUE;
    UInt32	 oword, data;			/* permuted output word being
						 * assembled */
    UInt8	*pSDown1, *pSDown2, *pSDown3;	/* ptrs to src lines below pS */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 4 source words.  
	 */
    oword = 0;
    pSDown1 = pS + sBpl;
    pSDown2 = pSDown1 + sBpl;
    pSDown3 = pSDown2 + sBpl;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;	/* previous word wasn't the leftmost one
				 * in the group that makes up the last
				 * dest word, so shift the results to the right
				 * and look to the left in the source.
				 */
	    /* put output bits [31:24] into bits
	     * 31xxx 30xxx 29xxx 28xxx 27xxx 26xxx 25xxx 24xxx
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcLimit]);		/* row 0 */
	data &= ULONG_ACCESS(&pSDown1[srcLimit]);	/* row 1 */
	data &= ULONG_ACCESS(&pSDown2[srcLimit]);	/* row 2 */
	data &= ULONG_ACCESS(&pSDown3[srcLimit]);	/* row 3 */

	    /* combine the columns, adjacent ones first, then column pairs */
	data &= (data << 1);
	data &= (data << 2);
	oword |= data & cBit3Mask4bpp;
    }

	/* unscramble the bits in oword.  We use one table instead of
	 * 4 (one for each shift position) to save space and registers.
	 * First, insert bits [25:24 17:16 9:8 1:0] */
    data = tab4To1Permute[(oword & cByteMask)];
    oword >>= 8;
	/* insert bits [27:26 19:18 11:10 3:2] */
    data |= tab4To1Permute[(oword & cByteMask)] << 2;
    oword >>= 8;
	/* insert bits [29:28 21:20 13:12 5:4] */
    data |= tab4To1Permute[(oword & cByteMask)] << 4;
    oword >>= 8;
	/* insert bits [31:30 23:22 15:14 7:6] */
    data |= tab4To1Permute[oword] << 6;
    ULONG_ACCESS(&pD[dstOffset]) = data;


	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	    /* put output bits [7:0] into bits
	     * xxx7 xxx6 xxx5 xxx4 xxx3 xxx2 xxx1 xxx0
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcOffset]);		/* row 0 */
	data &= ULONG_ACCESS(&pSDown1[srcOffset]);	/* row 1 */
	data &= ULONG_ACCESS(&pSDown2[srcOffset]);	/* row 2 */
	data &= ULONG_ACCESS(&pSDown3[srcOffset]);	/* row 3 */

	    /* combine the columns, adjacent ones first, then column pairs */
	data &= (data >> 1);
	data &= (data >> 2);
	oword = data & cBit0Mask4bpp;
	srcOffset -= 4;

	    /* put output bits [15:8] into bits
	     * xx15x xx14x xx13x xx12x xx11x xx10x xx9x xx8x
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcOffset]);		/* row 0 */
	data &= ULONG_ACCESS(&pSDown1[srcOffset]);	/* row 1 */
	data &= ULONG_ACCESS(&pSDown2[srcOffset]);	/* row 2 */
	data &= ULONG_ACCESS(&pSDown3[srcOffset]);	/* row 3 */
	    /* combine the columns, adjacent ones first, then column pairs */
	data &= (data << 1);
	data &= (data >> 2);
	oword |= data & cBit1Mask4bpp;
	srcOffset -= 4;


	    /* put output bits [23:16] into bits
	     * x23xx x22xx x21xx x20xx x19xx x18xx x17xx x16xx
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcOffset]);		/* row 0 */
	data &= ULONG_ACCESS(&pSDown1[srcOffset]);	/* row 1 */
	data &= ULONG_ACCESS(&pSDown2[srcOffset]);	/* row 2 */
	data &= ULONG_ACCESS(&pSDown3[srcOffset]);	/* row 3 */
	    /* combine the columns, adjacent ones first, then column pairs */
	data &= (data >> 1);
	data &= (data << 2);
	oword |= data & cBit2Mask4bpp;
	srcOffset -= 4;


	    /* put output bits [31:24] into bits
	     * 31xxx 30xxx 29xxx 28xxx 27xxx 26xxx 25xxx 24xxx
	     * of data.
	     */
	data = ULONG_ACCESS(&pS[srcOffset]);		/* row 0 */
	data &= ULONG_ACCESS(&pSDown1[srcOffset]);	/* row 1 */
	data &= ULONG_ACCESS(&pSDown2[srcOffset]);	/* row 2 */
	data &= ULONG_ACCESS(&pSDown3[srcOffset]);	/* row 3 */
	    /* combine the columns, adjacent ones first, then column pairs */
	data &= (data << 1);
	data &= (data << 2);
	oword |= data & cBit3Mask4bpp;
	srcOffset -= 4;

	    /* unscramble the bits in oword.  We use one table instead of
	     * 4 (one for each shift position) to save space and registers.
	     * First, insert bits [25:24 17:16 9:8 1:0] */
	data = tab4To1Permute[(oword & cByteMask)];
	oword >>= 8;
	    /* insert bits [27:26 19:18 11:10 3:2] */
	data |= tab4To1Permute[(oword & cByteMask)] << 2;
	oword >>= 8;
	    /* insert bits [29:28 21:20 13:12 5:4] */
	data |= tab4To1Permute[(oword & cByteMask)] << 4;
	oword >>= 8;
	    /* insert bits [31:30 23:22 15:14 7:6] */
	data |= tab4To1Permute[oword] << 6;
	ULONG_ACCESS(&pD[dstOffset]) = data;
    }
}


/*************************************************************************
 *  ip_reduce8Line()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the 64 corresponding pixels in the src 
 *		image are ON.  The four pixels are in a 8x8 square.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *              tab8To1Permute =  table that converts a byte containing
 *			ABCDEFGH -> xxxA xxxB xxxC xxxD xxxE xxxF xxxG xxxH.
 *				It is used to unscramble reduced bits in
 *				32 bit words.
 *		pThreshProc =	pointer to routine that turns four 8x8
 *				squares of bits into 4 bits using the
 *				desired threshold (only 1 or 4 allowed).
 *************************************************************************/
void  CDECL 
ip_reduce8Line (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32     *tab8To1Permute,
    UInt32 (CDECL *pThreshProc) (
		    IPState    *pIPState,
		    UInt8      *,
		    Int32	))
{
    PROLOGUE;
    UInt32	 oword, data;			/* permuted output word being
						 * assembled */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 8 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 1;	/* previous word wasn't the leftmost one
				 * in the group that makes up the last
				 * dest word, so shift the results to the right
				 * and look to the left in the source.
				 */
	oword |= pThreshProc(pIPState, &pS[srcLimit], sBpl);
    }
	/* unscramble the bits in oword.  We use one table instead of
	 * 4 (one for each shift position) to save space and registers.
	 * First, insert bits [28,24,20,16,12,8,4,0] */
    data = tab8To1Permute[(oword & cByteMask)];
    oword >>= 8;
	/* insert bits [29,25,21,17,13,9,5,1] */
    data |= tab8To1Permute[(oword & cByteMask)] << 1;
    oword >>= 8;
	/* insert bits [30,26,22,18,14,10,6,2] */
    data |= tab8To1Permute[(oword & cByteMask)] << 2;
    oword >>= 8;
	/* insert bits [31,27,23,19,15,11,7,3] */
    data |= tab8To1Permute[oword] << 3;
    ULONG_ACCESS(&pD[dstOffset]) = data;

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	    /* get output bits [3:0] into bits
	     *  3  x  x  x  x  x  x  x		Byte 3
	     *  2  x  x  x  x  x  x  x		Byte 2
	     *  1  x  x  x  x  x  x  x		Byte 1
	     *  0  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword = pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 1;

	    /* get output bits [7:4] into bits
	     *  7  3  x  x  x  x  x  x 		Byte 3
	     *  6  2  x  x  x  x  x  x 		Byte 2
	     *  5  1  x  x  x  x  x  x 		Byte 1
	     *  4  0  x  x  x  x  x  x 		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 1;

	    /* get output bits [11:8] into bits
	     * 11  7  3  x  x  x  x  x 	 	Byte 3
	     * 10  6  2  x  x  x  x  x 		Byte 2
	     *  9  5  1  x  x  x  x  x 		Byte 1
	     *  8  4  0  x  x  x  x  x 		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 1;

	    /* get output bits [15:12] into bits
	     * 15 11  7  3  x  x  x  x	 	Byte 3
	     * 14 10  6  2  x  x  x  x 		Byte 2
	     * 13  9  5  1  x  x  x  x		Byte 1
	     * 12  8  4  0  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 1;

	    /* get output bits [19:16] into bits
	     * 19 15 11  7  3  x  x  x	 	Byte 3
	     * 18 14 10  6  2  x  x  x		Byte 2
	     * 17 13  9  5  1  x  x  x		Byte 1
	     * 16 12  8  4  0  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 1;

	    /* get output bits [23:20] into bits
	     * 23 19 15 11  7  3  x  x	 	Byte 3
	     * 22 18 14 10  6  2  x  x		Byte 2
	     * 21 17 13  9  5  1  x  x		Byte 1
	     * 20 16 12  8  4  0  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 1;

	    /* get output bits [27:24] into bits
	     * 27 23 19 15 11  7  3  x	 	Byte 3
	     * 26 22 18 14 10  6  2  x		Byte 2
	     * 25 21 17 13  9  5  1  x		Byte 1
	     * 24 20 16 12  8  4  0  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 1;

	    /* get output bits [31:28] into bits
	     * 31 27 23 19 15 11  7  3	 	Byte 3
	     * 30 26 22 18 14 10  6  2		Byte 2
	     * 29 25 21 17 13  9  5  1		Byte 1
	     * 28 24 20 16 12  8  4  0		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;

	    /* unscramble the bits in oword.  We use one table instead of
	     * 4 (one for each shift position) to save space and registers.
	     * First, insert bits [28,24,20,16,12,8,4,0] */
	data = tab8To1Permute[(oword & cByteMask)];
	oword >>= 8;
	    /* insert bits [29,25,21,17,13,9,5,1] */
	data |= tab8To1Permute[(oword & cByteMask)] << 1;
	oword >>= 8;
	    /* insert bits [30,26,22,18,14,10,6,2] */
	data |= tab8To1Permute[(oword & cByteMask)] << 2;
	oword >>= 8;
	    /* insert bits [31,27,23,19,15,11,7,3] */
	data |= tab8To1Permute[oword] << 3;
	ULONG_ACCESS(&pD[dstOffset]) = data;
    }
}


/*************************************************************************
 *  ip_reduce8xT1()
 *		Convert four 8x8 squares to 4 bits in positions 31, 23, 15, 7
 *		of a 32-bit word.  Each bit is set if all the corresponding
 *		bits are set it its square.  The squares are word aligned.
 *
 *	Input:
 *		pS =		pointer to word containing top line of
 *				all squares
 *		sBpl =		source bytes/line
 *
 *	Returns:
 *		A word containing the result bits described above.
 *
 *************************************************************************/
UInt32  CDECL 
ip_reduce8xT1 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    UInt32	 data;			/* word being formed */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* OR together the bits in each column of every square */
    data = ULONG_ACCESS(pS);	/* row 0 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 1 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 2 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 3 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 4 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 5 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 6 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 7 */

	/* combine the columns, leave the result in the MS bits */
    data |= data << 1;
    data |= data << 2;
    data |= data << 4;

	/* return the result bits */
    data &= cMSBitMask8bpp;	/* Macintosh compiler might have trouble
				 * with constant expressions in return
				 * statements
				 */
    return data;
}



/*************************************************************************
 *  ip_reduce8xT4()
 *		Convert four 8x8 squares to 4 bits in positions 31, 23, 15, 7
 *		of a 32-bit word.  Each bit is set if none of the corresponding
 *		bits are set it its square.  The squares are word aligned.
 *
 *	Input:
 *		pS =		pointer to word containing top line of
 *				all squares
 *		sBpl =		source bytes/line
 *
 *	Returns:
 *		A word containing the result bits described above.
 *
 *************************************************************************/
UInt32  CDECL 
ip_reduce8xT4 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    UInt32	data;			/* word being formed */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* OR together the bits in each column of every square */
    data = ULONG_ACCESS(pS);	/* row 0 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 1 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 2 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 3 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 4 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 5 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 6 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 7 */

	/* combine the columns, leave the result in the MS bits */
    data &= data << 1;
    data &= data << 2;
    data &= data << 4;

	/* return the result bits */
    data &= cMSBitMask8bpp;	/* Macintosh compiler might have trouble
				 * with constant expressions in return
				 * statements
				 */
    return data;
}



/*************************************************************************
 *  ip_reduce16Line()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the 256 corresponding pixels in the src 
 *		image are ON.  The pixels are in a 16x16 square.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *		pThreshProc =	pointer to routine that turns two 16x16
 *				squares of bits into 2 bits using the
 *				desired threshold (only 1 or 4 allowed).
 *************************************************************************/
void  CDECL 
ip_reduce16Line (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit,
    UInt32 (CDECL *pThreshProc) (
		    IPState    *pIPState,
		    UInt8      *,
		    Int32	))
{
    PROLOGUE;
    UInt32	 oword;			/* permuted output word being
					 * assembled */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* do the last destination word first because it may require
	 * less than 16 source words.  
	 */
    oword = 0;
    for (; srcLimit >= srcOffset; srcLimit -= 4)
    {
	oword = oword >> 2;	/* previous word wasn't the leftmost one
				 * in the group that makes up the last
				 * dest word, so shift the results to the right
				 * and look to the left in the source.
				 */
	oword |= pThreshProc(pIPState, &pS[srcLimit], sBpl);
    }

	/* The threshold proc returns its bits in the two MS bits of the
	 * 32-bit word, so oword is not scrambled.  Just store it.
	 */
    ULONG_ACCESS(&pD[dstOffset]) = oword;

	/* the loop for each destination word */
    for (srcOffset -= 4, dstOffset -= 4; dstOffset >= 0; dstOffset -= 4)
    {
	    /* get output bits [1:0] into bits
	     *  1  0  x  x  x  x  x  x		Byte 3
	     *  x  x  x  x  x  x  x  x		Byte 2
	     *  x  x  x  x  x  x  x  x		Byte 1
	     *  x  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword = pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 2;

	    /* get output bits [3:2] into bits
	     *  3  2  1  0  x  x  x  x		Byte 3
	     *  x  x  x  x  x  x  x  x		Byte 2
	     *  x  x  x  x  x  x  x  x		Byte 1
	     *  x  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 2;

	    /* get output bits [5:4] into bits
	     *  5  4  3  2  1  0  x  x		Byte 3
	     *  x  x  x  x  x  x  x  x		Byte 2
	     *  x  x  x  x  x  x  x  x		Byte 1
	     *  x  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 2;

	    /* get output bits [7:6] into bits
	     *  7  6  5  4  3  2  1  0		Byte 3
	     *  x  x  x  x  x  x  x  x		Byte 2
	     *  x  x  x  x  x  x  x  x		Byte 1
	     *  x  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 2;

	    /* get output bits [9:8] into bits
	     *  9  8  7  6  5  4  3  2		Byte 3
	     *  1  0  x  x  x  x  x  x		Byte 2
	     *  x  x  x  x  x  x  x  x		Byte 1
	     *  x  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 2;

	    /* get output bits [11:10] into bits
	     * 11 10  9  8  7  6  5  4		Byte 3
	     *  3  2  1  0  x  x  x  x		Byte 2
	     *  x  x  x  x  x  x  x  x		Byte 1
	     *  x  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 2;

	    /* get output bits [13:12] into bits
	     * 13 12 11 10  9  8  7  6		Byte 3
	     *  5  4  3  2  1  0  x  x		Byte 2
	     *  x  x  x  x  x  x  x  x		Byte 1
	     *  x  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 2;

	    /* get output bits [15:14] into bits
	     * 15 14 13 12 11 10  9  8		Byte 3
	     *  7  6  5  4  3  2  1  0		Byte 3
	     *  x  x  x  x  x  x  x  x		Byte 1
	     *  x  x  x  x  x  x  x  x		Byte 0
	     * of oword.
	     */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);
	srcOffset -= 4;
	oword >>= 2;

	/* Well, you get the idea.  Repeat this 8 more times. */
	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);	/* bits 17:16 */
	srcOffset -= 4;
	oword >>= 2;

	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);	/* bits 19:18 */
	srcOffset -= 4;
	oword >>= 2;

	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);	/* bits 21:20 */
	srcOffset -= 4;
	oword >>= 2;

	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);	/* bits 23:22 */
	srcOffset -= 4;
	oword >>= 2;

	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);	/* bits 25:24 */
	srcOffset -= 4;
	oword >>= 2;

	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);	/* bits 27:26 */
	srcOffset -= 4;
	oword >>= 2;

	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);	/* bits 29:28 */
	srcOffset -= 4;
	oword >>= 2;

	oword |= pThreshProc(pIPState, &pS[srcOffset], sBpl);	/* bits 31:30 */
	srcOffset -= 4;

	    /* The threshold proc returns its bits in the two MS bits of the
	     * 32-bit word, so oword is not scrambled.  Just store it.
	     */
	ULONG_ACCESS(&pD[dstOffset]) = oword;
    }
}


/*************************************************************************
 *  ip_reduce16xT1()
 *		Convert four 8x8 squares to 4 bits in positions 31, 23, 15, 7
 *		of a 32-bit word.  Each bit is set if all the corresponding
 *		bits are set it its square.  The squares are word aligned.
 *
 *	Input:
 *		pS =		pointer to word containing top line of
 *				all squares
 *		sBpl =		source bytes/line
 *
 *	Returns:
 *		A word containing the result bits described above.
 *
 *************************************************************************/
UInt32  CDECL 
ip_reduce16xT1 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    UInt32	 data;			/* word being formed */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* OR together the bits in each column of every square */
    data = ULONG_ACCESS(pS);	/* row 0 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 1 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 2 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 3 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 4 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 5 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 6 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 7 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 8 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 9 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 10 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 11 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 12 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 13 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 14 */
    pS += sBpl;
    data |= ULONG_ACCESS(pS);	/* row 15 */

	/* combine the columns, leave the result in the MS bits */
    data |= data << 1;		/* results in odd bits */
    data |= data << 2;		/* results in 31, 27, 23, 19, 15, 11, 7, 3 */
    data |= data << 4;		/* results in 31,     23,     15,     7    */
    data |= data << 8;		/* results in 31,             15          */
    data &= cMSBitMask16bpp;
    data |= data << 15;		/* result in 31:30 */

	/* return the result bits */
    data &= c16xBitMask;	/* Macintosh compiler might have trouble
				 * with constant expressions in return
				 * statements
				 */
    return data;
}



/*************************************************************************
 *  ip_reduce16xT4()
 *		Convert four 8x8 squares to 4 bits in positions 31, 23, 15, 7
 *		of a 32-bit word.  Each bit is set if none of the corresponding
 *		bits are set it its square.  The squares are word aligned.
 *
 *	Input:
 *		pS =		pointer to word containing top line of
 *				all squares
 *		sBpl =		source bytes/line
 *
 *	Returns:
 *		A word containing the result bits described above.
 *
 *************************************************************************/
UInt32  CDECL 
ip_reduce16xT4 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl)
{
    PROLOGUE;
    UInt32	 data;			/* word being formed */

	/* keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* OR together the bits in each column of every square */
    data = ULONG_ACCESS(pS);	/* row 0 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 1 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 2 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 3 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 4 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 5 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 6 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 7 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 8 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 9 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 10 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 11 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 12 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 13 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 14 */
    pS += sBpl;
    data &= ULONG_ACCESS(pS);	/* row 15 */

	/* combine the columns, leave the result in the MS bits */
    data &= data << 1;		/* results in odd bits */
    data &= data << 2;		/* results in 31, 27, 23, 19, 15, 11, 7, 3 */
    data &= data << 4;		/* results in 31,     23,     15,     7    */
    data &= data << 8;		/* results in 31,             15          */
    data &= cMSBitMask16bpp;
    data |= data << 15;		/* result in 31:30 */

	/* return the result bits */
    data &= c16xBitMask;	/* Macintosh compiler might have trouble
				 * with constant expressions in return
				 * statements
				 */
    return data;
}


/*

$Log: mireduce.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.5  1998/01/13 23:53:08  ddavies
mireduce.c -> /product/ipcore/ipshared/src/RCS/mireduce.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1997/07/11 01:49:57  ddavies
mireduce.c -> /product/ipcore/ipshared/src/RCS/mireduce.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.3  1995/12/13 02:51:03  ejaquith
mireduce.c -> /product/ipcore/ipshared/src/RCS/mireduce.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1994/07/14  18:48:36  ejaquith
 * ejaquith on Thu Jul 14 11:47:44 PDT 1994
 *
 * mireduce.c -> /product/ipcore/ipshared/src/RCS/mireduce.c,v
 *
 * removed unnecessary #include "iadef.prv"
 *
 * Revision 1.1  1994/02/03  00:09:44  ddavies
 * Initial revision
 *
 * Revision 1.26  1994/01/27  03:48:52  ddavies
 * ddavies on Wed Jan 26 19:43:58 PST 1994
 *
 * sreduce.c -> /product/ipcore/ipshared/src/RCS/sreduce.c,v
 *
 * Quick fix of the previous fix:  The code that attempted to clear the right
 * and bottom frames of the intermediate images actually cleared some area below
 * the bottom of the image.  I forgot that the frames were already included
 * in wT and hT.  Following Richard Mylnarik's suggestion, the temporary
 * images are now allocated relative to the dest, not the source.  This allows
 * us to skip initializing the bottom and right areas of the temp images
 * even when the DAE folks give us a dest image that's too big.  When this
 * happens, we'll read into the frame of the source image, except in io_reduceX
 * where the source might not have a frame.  The DAE folks assure us they
 * don't use io_reduce and never will.
 * In the next day or so, sreduce.c and freduce.c will become i_reduce.c and
 * mireduce.c.
 *
 * Revision 1.25  1994/01/14  23:28:43  ddavies
 * sreduce.c -> /product/ipcore/ipshared/src/RCS/sreduce.c,v
 *
 * Fix bugs caused by uninitialized frames when making temporary images.
 * Thanks for finding these Richard!!
 *
 * Revision 1.24  1993/12/02  19:29:37  ddavies
 * ddavies on Thu Dec 2 11:29:34 PST 1993
 *
 * sreduce.c -> /product/ipcore/ipshared/src/RCS/sreduce.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.23  1993/11/05  02:23:20  ddavies
 * sreduce.c -> /product/ipcore/ipshared/src/RCS/sreduce.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


