/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * mirescl1.c
 *
 * Private rescaling routines:
 *		void	ip_arbRescale8bppOneLine()
 *		void	ip_arbRescale8bppFourLines()
*/


#include <stdlib.h>
#include <stddef.h>

#ifndef _TYPES_PUB_INCLUDED
#include "types.pub"
#include "ipscore.prv"
#endif

#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "shrscal.pub"
#include "shrrast.pub"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mirescl1.c,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:18:16 $ */

/* PRIVATE */
/**************************************************************************
 *  ip_arbRescale8bppOneLine()
 *		Generates one rescaled line using a linear weighting
 *		interpolation algorithm.
 *
 *	Input:
 *		pS =	pointer to beginning of next line of source 
 *		sBpl =	bytes per line of source image
 *		pD =	pointer to next line of destination image
 *		wD =	width of destination image
 *		dBpl =	byte per line of destination image
 *		xOICoord = split integer: x position of current input
 *			pixel in output's frame of referecne
 *		yOICoord = split integer: y position of current input
 *			pixel in output's frame of reference
 *		xOIIncrement = split integer: amount to increment xOICoord
 *		yOIIncrement = split integer: amount to increment yOICoord
 *		yCount = number of destination lines to produce.
 *			either 1 or 4
 *		pRescaleTable = pointer to rescale table
 *		pStateTable = pointer to state table
 *
 *	Split integers are already in form to be used in index.
 *************************************************************************/
void  CDECL 
ip_arbRescale8bppOneLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	wD,
    UInt32	xOICoord,
    UInt32	yOICoord,
    UInt32	xOIIncrement,
    UInt16     *pRescaleTable,
    UInt8      *pStateTable)
{
    PROLOGUE;
    UInt32	gray0, grayT;
    Int32	dCol;
    UInt8      *pRescaleLeftTop, *pRescaleRightTop;
    UInt8      *pRescaleLeftBot, *pRescaleRightBot;
    UInt32	twiceXOIInc;
    UInt32	xFrac;

#define cResXSmooth	0x4000		/* bit 14 of index.  see below */
#define cResYSmooth	0x8000		/* bit 15 of index.  see below */

#define cResPixMask	0xFE		/* mask for raw pixel values */

	/* states in the state machine.  see comments above */
#define cRes_2_2_State		0x0
#define cRes_2_1_1_State	0x1
#define cRes_1_2_1_State	0x2
#define cRes_1_1_2_State	0x3
#define cRes_1_1_1_1_State	0x4

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Just because it's important in what's coming up, let's set the
	 * format of the table lookup index:
	 * | 15 | 14 |   13   12   11  |  10   9    8  | 7 6 5 4 3 2 1 | 0 |
	 * | yS | xS | 3b xIOCoord     | 3b yIOCoord   |  7b pixel val | 0 |
	 *
	 * The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
	 * yIOCoord bits are its 3 MS fraction bits.  The 7 pixel bits are
	 * the 7 MS pixel bits.
	 * The 0 at the end makes this an index into a table of short ints.
	 * Bit 14 is used to access the table when doing the x smoothing
	 * with the 2nd pixel.
	 * Remember each table entry is 2 pixels wide.  Say that our
	 * 3 MS bits are 010.  Then with the first pixel we will access
	 * using 010 and get entries for 010 and 011 (assuming that our
	 * xOIIncrement is 001).  Now the entires that we want for our
	 * second pixel correspond to 101 and 100.  Fine.  Sometime
	 * later we could make a reference with our first table
	 * corresponding to 101 and 110.  This conflicts with
	 * the entries that we needed above when we indexed with 101.
	 * Clearly then we must have 2 tables - one to be used by the
	 * first pixel and one to be used by the second pixel.  Hence
	 * bit 14.
	 * Bit 15 is used to access the table when doing the y smoothing
	 * with the 2nd line.
	 * This makes some small amount of difference but we could
	 * easily go without this and most people probably wouldn't notice.
	 * Without this bit we must access the table with yOICoord and
	 * yOICoord ^ cResYFMask.  This pairs the table entries so that
	 * we have 4 pairs.  If we use bit 15 we can basically have
	 * 8 pairs and hence a *little* better smoothing.
	 */

    twiceXOIInc = xOIIncrement + xOIIncrement;

	/* point to the table for the 2 lines */
    pRescaleLeftTop = ((UInt8 *)pRescaleTable) + (yOICoord & cResYFMask);
    pRescaleLeftBot = ((UInt8 *)pRescaleTable) + (yOICoord & cResYFMask) + cResYSmooth;
    pRescaleRightTop = ((UInt8 *)pRescaleTable) + (yOICoord & cResYFMask) + cResXSmooth;
    pRescaleRightBot = ((UInt8 *)pRescaleTable) + (yOICoord & cResYFMask) + cResXSmooth + cResYSmooth;

    gray0 = 0;

    for (dCol = 0; dCol < wD; dCol += 4)
    {
	switch (*pStateTable++)
	{
	case cRes_2_2_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = gray0 << 16;
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;
	    break;

	case cRes_2_1_1_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = gray0 << 16;
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
		/* the leftmost pixel is correct and in the correct
		 * position.  we want to recalculate the rightmost
		 * of the 2 pixels, however.
		 */
	    gray0 &= ~0xFF;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 |= grayT >> 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;
	    break;

	case cRes_1_2_1_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = gray0 << 8;
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 |= grayT >> 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;
	    break;

	case cRes_1_1_2_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;
	    break;
	case cRes_1_1_1_1_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF);
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 |= grayT >> 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;
	    break;

	}
	ULONG_ACCESS(pD) = gray0;
	pD += 4;
    }
}


/* PRIVATE */
/**************************************************************************
 *  ip_arbRescale8bppFourLines()
 *		Generates four rescaled lines using a linear weighting
 *		interpolation algorithm.
 *
 *	Input:
 *		pS =	pointer to beginning of next line of source 
 *		sBpl =	bytes per line of source image
 *		pD =	pointer to next line of destination image
 *		wD =	width of destination image
 *		dBpl =	byte per line of destination image
 *		xOICoord = split integer: x position of current input
 *			pixel in output's frame of referecne
 *		yOICoord = split integer: y position of current input
 *			pixel in output's frame of reference
 *		xOIIncrement = split integer: amount to increment xOICoord
 *		yOIIncrement = split integer: amount to increment yOICoord
 *		pRescaleTable = pointer to rescale table
 *		pStateTable = pointer to state table
 *
 *	Split integers are already in form to be used in index.
 *************************************************************************/
void  CDECL 
ip_arbRescale8bppFourLines (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	wD,
    Int32	dBpl,
    UInt32	xOICoord,
    UInt32	yOICoord,
    UInt32	xOIIncrement,
    UInt32	yOIIncrement,
    UInt16     *pRescaleTable,
    UInt8      *pStateTable)
{
    PROLOGUE;
    UInt32	gray0, gray1, gray2, gray3, grayT;
    Int32	dCol;
    UInt8      *pRescaleLeftTop, *pRescaleRightTop;
    UInt8      *pRescaleLeftBot, *pRescaleRightBot;
    UInt32	twiceXOIInc;
    UInt32	xFrac, yFrac0, yFrac1, yFrac2, yFrac3;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    twiceXOIInc = xOIIncrement + xOIIncrement;
    gray0 = gray1 = gray2 = gray3 = 0;

    yFrac0 = yOICoord;
    yFrac1 = yFrac0 + yOIIncrement;
    yFrac2 = yFrac1 + yOIIncrement;
    yFrac3 = yFrac2 + yOIIncrement;
    yFrac0 &= cResYFMask;
    yFrac1 &= cResYFMask;
    yFrac2 &= cResYFMask;
    yFrac3 &= cResYFMask;
    pRescaleLeftTop = ((UInt8 *)pRescaleTable);
    pRescaleLeftBot = ((UInt8 *)pRescaleTable) + cResYSmooth;
    pRescaleRightTop = ((UInt8 *)pRescaleTable) + cResXSmooth;
    pRescaleRightBot = ((UInt8 *)pRescaleTable) + cResXSmooth + cResYSmooth;

    for (dCol = 0; dCol < wD; dCol += 4)
    {
	switch (*pStateTable++)
	{
	case cRes_2_2_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 + 
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = gray0 << 16;
	    gray1 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = gray1 << 16;
	    gray2 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac +  yFrac2 + 
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = gray2 << 16;
	    gray3 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 + 
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = gray3 << 16;
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 + 
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 + 
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 + 
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;
	    break;

	case cRes_2_1_1_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 + 
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac +  yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = gray0 << 16;
	    gray1 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 + 
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 + 
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = gray1 << 16;
	    gray2 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 + 
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = gray2 << 16;
	    gray3 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = gray3 << 16;
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 + 
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac +  yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
		/* the leftmost pixel is correct and in the correct
		 * position.  we want to recalculate the rightmost
		 * of the 2 pixels, however.
		 */
	    gray0 &= ~0xFF;
	    gray1 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 &= ~0xFF;
	    gray2 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 &= ~0xFF;
	    gray3 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 &= ~0xFF;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;

	    xFrac = xOICoord & cResXFMask;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 + 
		    (UCHAR_ACCESS(++pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 + 
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 + 
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 |= grayT >> 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;
	    break;

	case cRes_1_2_1_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac +  yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    gray1 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac +  yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = (gray1 & ~0xFF) << 8;
	    gray2 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac +  yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = (gray2 & ~0xFF) << 8;
	    gray3 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac +  yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = (gray3 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(++pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = gray0 << 8;
	    gray1 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = gray1 << 8;
	    gray2 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = gray2 << 8;
	    gray3 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = gray3 << 8;
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;

	    xFrac = xOICoord & cResXFMask;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 |= grayT >> 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;
	    break;

	case cRes_1_1_2_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    gray1 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = (gray1 & ~0xFF) << 8;
	    gray2 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = (gray2 & ~0xFF) << 8;
	    gray3 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = (gray3 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(++pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    gray1 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = (gray1 & ~0xFF) << 8;
	    gray2 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = (gray2 & ~0xFF) << 8;
	    gray3 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = (gray3 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(++pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    xOICoord = (xOICoord & cResXFAMask) + twiceXOIInc;
	    pS += xOICoord >> cResXIPos;
	    break;
	case cRes_1_1_1_1_State:
	    xFrac = xOICoord & cResXFMask;
	    gray0 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    gray1 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = (gray1 & ~0xFF) << 8;
	    gray2 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = (gray2 & ~0xFF) << 8;
	    gray3 = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = (gray3 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(++pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF) << 8;
	    gray1 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = (gray1 & ~0xFF) << 8;
	    gray2 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = (gray2 & ~0xFF) << 8;
	    gray3 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = (gray3 & ~0xFF) << 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;

	    xFrac = xOICoord & cResXFMask;
	    gray0 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(++pS) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray0 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 = (gray0 & ~0xFF);
	    gray1 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray1 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 = (gray1 & ~0xFF);
	    gray2 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray2 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 = (gray2 & ~0xFF);
	    gray3 += *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(pS) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    gray3 += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 = (gray3 & ~0xFF);
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;

	    xFrac = xOICoord & cResXFMask;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac0 +
		    (UCHAR_ACCESS(++pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac0 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray0 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac1 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac1 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray1 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac2 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac2 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray2 |= grayT >> 8;
	    grayT = *(UInt16 *) &pRescaleLeftTop[xFrac + yFrac3 +
		    (UCHAR_ACCESS(pS) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightTop[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[1]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleLeftBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl]) & cResPixMask)];
	    grayT += *(UInt16 *) &pRescaleRightBot[xFrac + yFrac3 +
		     (UCHAR_ACCESS(&pS[sBpl+1]) & cResPixMask)];
	    gray3 |= grayT >> 8;
	    xOICoord = (xOICoord & cResXFAMask) + xOIIncrement;
	    pS += xOICoord >> cResXIPos;
	    break;

	}
	ULONG_ACCESS(pD) = gray0;
	ULONG_ACCESS(&pD[dBpl]) = gray1;
	ULONG_ACCESS(&pD[dBpl+dBpl]) = gray2;
	ULONG_ACCESS(&pD[dBpl+dBpl+dBpl]) = gray3;
	pD += 4;
    }
}

/*

$Log: mirescl1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.4  1998/01/13 23:53:09  ddavies
mirescl1.c -> /product/ipcore/ipshared/src/RCS/mirescl1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1995/12/13 02:51:12  ejaquith
mirescl1.c -> /product/ipcore/ipshared/src/RCS/mirescl1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/10/18  04:03:50  ddavies
 * mirescl1.c -> /product/ipcore/ipshared/src/RCS/mirescl1.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.1  1994/09/09  00:14:06  lperry
 * Initial revision
 *
 * Revision 1.4  1994/02/04  20:18:36  dferrell
 * mirescal.c -> /product/ipcore/ipshared/src/RCS/mirescal.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:25:03  ddavies
 * mirescal.c -> /product/ipcore/ipshared/src/RCS/mirescal.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:50:03  ddavies
 * mirescal.c -> /product/ipcore/ipshared/src/RCS/mirescal.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


