/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * mirescl2.c
 *
 * Put in separate file to keep code size down
 *
 * Private rescaling routines:
 *		void	ip_areaMap8bppLine()
*/


#include <stdlib.h>
#include <stddef.h>

#ifndef _TYPES_PUB_INCLUDED
#include "types.pub"
#include "ipscore.prv"
#endif

#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "shrscal.pub"
#include "shrrast.pub"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mirescl2.c,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:18:16 $ */

#define cResPixMask	0xFE		/* mask for raw pixel values */

/* PRIVATE */
/**************************************************************************
 *  ip_areaMap8bppLine()
 *		Generates one rescaled line using an area weighting
 *		interpolation algorithm.
 *
 *	Input:
 *		pS =	pointer to beginning of next line of source 
 *		sBpl =	bytes per line of source image
 *		pD =	pointer to next line of destination image
 *		wD =	width of destination image
 *		xIOCoord = split integer: x position of current input
 *			pixel in output's frame of referecne
 *		yIOCoord = split integer: y position of current input
 *			pixel in output's frame of reference
 *		xIOIncrement = split integer: amount to increment xIOCoord
 *		xCount = for every xCount source pixels we produce
 *			xCount+1 destination pixels
 *		pRescaleTable = pointer to rescale table
 *
 *	Split integers are of the form:
 *		bits [31..16] integer portion
 *		bits [15..0] fraction
 *************************************************************************/
void  CDECL 
ip_areaMap8bppLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	wD,
    Int32	xIOCoord,
    Int32	yIOCoord,
    Int32	xIOIncrement,
    Int32	xCount,
    UInt16     *pRescaleTable,
    UInt8      *pStateTable)
{
    PROLOGUE;
    UInt32	gray0;
    Int32	dCol;
    UInt8      *pRescale, *pRescaleTop, *pRescaleBot;
    Int32	extraPixel;
    Int32	xIOCoordNext;

#define cExtraPixel1Start	0
#define cExtraPixel1StartExtra	1
#define cExtraPixel1Finish	2
#define cExtraPixel1FinishXCount2 3
#define cExtraPixel2Start	4
#define cExtraPixel2Finish	5
#define cExtraPixel3Start	6
#define cExtraPixel3Finish	7
#define cExtraPixel4Start	8
#define cExtraPixel4Finish	9
#define cExtraPixel5OrOverStart	10
#define cExtraPixel5OrOverCont	11

#define cExtraPixelDec1		0x4000
#define cExtraPixelDec2		0x8000
#define cExtraPixelDec3		0xC000
#define cExtraPixelDec4		0x10000


	/* Just because it's important in what's coming up, let's set the
	 * format of the table lookup index:
	 * | 14 |   13   12   11    |  10     9      8  | 7 6 5 4 3 2 1 | 0 |
	 * | tab| 3b xIOCoord       | 3b yIOCoord       |  7b pixel val | 0 |
	 *
	 * The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
	 * yIOCoord bits are its 3 MS fraction bits.  The 7 pixel bits are
	 * the 7 MS pixel bits.
	 * The 0 at the end makes this an index into a table of short ints.
	 * The top bit, tab, determines if we are using the special table;
	 * see comments before the procedure.
	 */

	/* We'd like to have the three MS fraction bits in xIOC aligned 
	 * with their position in the table lookup.
	 */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    xIOCoord = xIOCoord >> 2;
    xIOIncrement = xIOIncrement >> 2;
    xIOCoordNext = xIOCoord + xIOIncrement;
    xCount = xCount >> 2;

	/* Same for yIOC */
    yIOCoord = yIOCoord >> 5;


	/* so that the compiler won't automatically do indexing */
    pRescale = (UInt8 *)pRescaleTable;

    gray0 = 0;
    extraPixel = xCount;

    if ((yIOCoord & cResYFMask) == 0)
    {
	for (dCol = 0; dCol < wD; )
	{
	    switch (*pStateTable++)
	    {
	    case cExtraPixel1Start:
		/* for now we will just duplicate the pixel
		 * into two pixels.  In the future this should
		 * possibly go into the magnify by 2-3x case.  
		 * In that case each source pixel could influence
		 * 3 destination pixels unlike magnify by < 2 in
		 * which each source pixel can influence only 2
		 * destination pixels.
		 */
		/* We will just get the values directly from the
		 * source.  As a result we do not need to maintain
		 * xIOCoord and xIOCoordNext.
		 */
		/* The previous version of this code calculated the
		 * next state as we were going through the loop.
		 * In order to get a good jump table, we converted
		 * to an algorithm in which the states are calculated for
		 * the line once and the same table is used for each
		 * line.  The states used to be based on extraPixel
		 * and as a result the numbers were not consecutive - 
		 * and produced a bad jump table.  Since this code
		 * is not recommended for a scale factor > 1.5, the states
		 * of cExtraPixel1Start and cExtraPixel1StartExtra should
		 * not be used.  Hence, they were converted directly to
		 * the new scheme but without extra enhancements.
		 * If this code is really to be used, we should add
		 * several extra states which correspond to the if
		 * statements in these 2 states and get rid of extraPixel.
		 */
		gray0 = (UInt32)(UCHAR_ACCESS(pS) & cResPixMask);
		gray0 = gray0 << 8;
		    /* Occassionally we may not want to drop out an
		     * extra pixel.  This can happen once during a
		     * loop iteration but not more than once.
		     */
		if (extraPixel < cExtraPixelDec2)
		{
			/* put the extra pixel in */
		    gray0 += (UInt32)(UCHAR_ACCESS(pS++) & cResPixMask);
		    gray0 = gray0 << 8;
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
		    gray0 += (UInt32)(UCHAR_ACCESS(pS) & cResPixMask);
		    gray0 = gray0 << 8;
		    if (extraPixel < cExtraPixelDec2)
		    {
			    /* put the extra pixel in */
			gray0 += (UInt32)(UCHAR_ACCESS(pS++) & cResPixMask);
			extraPixel = (extraPixel & cResXFAMask) + xCount;
		    }
		    else
		    {
			    /* don't put the extra pixel in */
		        gray0 += (UInt32)(UCHAR_ACCESS(++pS) & cResPixMask);
			extraPixel -= cExtraPixelDec1;
		    }
		}
		else
		{
		        /* don't duplicate the pixel just do get the next
		         * pixel.
		         */
		    gray0 += (UInt32)(UCHAR_ACCESS(++pS) & cResPixMask);
		    gray0 = gray0 << 8;
		        /* duplicate that pixel */
		    gray0 += (UInt32)(UCHAR_ACCESS(pS++) & cResPixMask);
		    gray0 = gray0 << 8;
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
		        /* and get the next one */
		    gray0 += (UInt32)(UCHAR_ACCESS(pS) & cResPixMask);
		}
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel2Start:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
		  (UCHAR_ACCESS(pS) & cResPixMask);
	        gray0 = gray0 << 8;
		pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel3Start:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
		  (UCHAR_ACCESS(pS) & cResPixMask);
		pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel4Start:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel5OrOverStart:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel1StartExtra:
		    /* put the extra pixel in from the previous state */
		gray0 = (UInt32)(UCHAR_ACCESS(pS++) & cResPixMask);
		gray0 = gray0 << 8;
		extraPixel = (extraPixel & cResXFAMask) + xCount;
		    /* put the new source pixel in */
		gray0 += (UInt32)(UCHAR_ACCESS(pS) & cResPixMask);
		gray0 = gray0 << 8;
		if (extraPixel < cExtraPixelDec2)
		{
		        /* put in the extra pixel */
		    gray0 += (UInt32)(UCHAR_ACCESS(pS++) & cResPixMask);
		    gray0 = gray0 << 8;
			/* put in the next source pixel */
		    gray0 += (UInt32)(UCHAR_ACCESS(pS) & cResPixMask);
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
		    if (extraPixel >= cExtraPixelDec2)
			{
			pS++;
			extraPixel -= cExtraPixelDec1;
			}
		}
		else
		{
		        /* no extra pixel, put in the next source 
			 * pixel 
			 */
		    gray0 += (UInt32)(UCHAR_ACCESS(++pS) & cResPixMask);
		    gray0 = gray0 << 8;
			/* put in the extra pixel */
		    gray0 += (UInt32)(UCHAR_ACCESS(pS++) & cResPixMask);
		    extraPixel -= cExtraPixelDec1;
		}
 	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel1Finish:
		    /* can only get in this state if xCount > 2.
		     * if xCount == 2, we will be in 
		     * cExtraPixel1FinishXCount2 below.  if xCount == 1
		     * we will never leave the cExtraPixel1Start case.
		     */
		gray0 = (UCHAR_ACCESS(pS) & cResPixMask) << 16;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = gray0 << 8;
		pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel2Finish:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
		  (UCHAR_ACCESS(pS) & cResPixMask);
	        gray0 = gray0 << 16;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

		    /* we know that xCount is at least 2 so that we
		     * can safely put in one pixel without worrying about
		     * needing an extra pixel.
		     */
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel3Finish:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
		  (UCHAR_ACCESS(pS) & cResPixMask);
	        gray0 = gray0 << 8;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel4Finish:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
		  (UCHAR_ACCESS(pS) & cResPixMask);
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel5OrOverCont:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel1FinishXCount2:
		gray0 = (UCHAR_ACCESS(pS) & cResPixMask) << 16;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = gray0 << 8;
		pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescale[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
		  (UCHAR_ACCESS(pS) & cResPixMask);
		pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    }
	}
    }
    else 	/* the destination line is composed of 2 source lines */
    {
	/* pRescaleTop = pRescale + ((((yIOCoord & cResYIMask) + cResYIInc) - yIOCoord) & cResYFMask); */
	pRescaleTop = pRescale + (cResYIInc - (yIOCoord & cResYFMask));
	pRescaleBot = pRescale + (yIOCoord & cResYFMask);
	for (dCol = 0; dCol < wD; )
	{
	    switch (*pStateTable++)
	    {
	    case cExtraPixel1Start:
		/* for now we will just duplicate the pixel
		 * into two pixels.  In the future this should
		 * possibly go into the magnify by 2-3x case.  
		 * In that case each source pixel could influence
		 * 3 destination pixels unlike magnify by < 2 in
		 * which each source pixel can influence only 2
		 * destination pixels.
		 */
		/* We will just get the values directly from the
		 * source.  As a result we do not need to maintain
		 * xIOCoord and xIOCoordNext.
		 */
		/* remember that the pixel value we want is actually
		 * in bits [15..8]
		 */
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS++) & cResPixMask)]));
		    /* Occassionally we may not want to drop out an
		     * extra pixel.  This can happen once during a
		     * loop iteration but not more than once.
		     */
		if (extraPixel < cExtraPixelDec2)
		{
			/* put the extra pixel in */
		    gray0 = (gray0 & ~0xFF);
		    gray0 = (gray0 << 16) | (gray0 << 8);
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
			/* next source pixel */
	            gray0 += (UInt32) 
                      ( *(UInt16 *) &(pRescaleTop[
	              (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	            gray0 += (UInt32) 
                      ( *(UInt16 *) &(pRescaleBot[
	              (UCHAR_ACCESS(pS++) & cResPixMask)]));
		    gray0 = (gray0 & ~0xFF);
		    if (extraPixel < cExtraPixelDec2)
		    {
			    /* put the extra pixel in */
			gray0 |= (gray0 & 0xFF00) >> 8;
			extraPixel = (extraPixel & cResXFAMask) + xCount;
		    }
		    else
		    {
			    /* don't put the extra pixel in */
	                gray0 += ((UInt32) 
                          ( *(UInt16 *) &(pRescaleTop[
	                  (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	                gray0 += ((UInt32) 
                          ( *(UInt16 *) &(pRescaleBot[
	                  (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
			extraPixel -= cExtraPixelDec1;
		    }
		}
		else
		{
		        /* don't duplicate the pixel just go get the next
		         * pixel.
		         */
		    gray0 = (gray0 & ~0xFF) << 8;
	            gray0 += (UInt32) 
                      ( *(UInt16 *) &(pRescaleTop[
	              (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	            gray0 += (UInt32) 
                      ( *(UInt16 *) &(pRescaleBot[
	              (UCHAR_ACCESS(pS++) & cResPixMask)]));
		        /* duplicate that pixel */
		    gray0 = (gray0 & 0xFF00) | ((gray0 & ~0xFF) << 8);
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
		        /* and get the next one */
	            gray0 += ((UInt32) 
                      ( *(UInt16 *) &(pRescaleTop[
	              (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	            gray0 += ((UInt32) 
                      ( *(UInt16 *) &(pRescaleBot[
	              (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
		}
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel2Start:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
                  (( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)])) >> 8);
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        gray0 = gray0 << 8;
		pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel3Start:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
                  (( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)])) >> 8);
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
		pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel4Start:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel5OrOverStart:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel1StartExtra:
		    /* put the extra pixel in from the previous state */
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS++) & cResPixMask)]));
		gray0 = (gray0 & ~0xFF) << 8;
		extraPixel = (extraPixel & cResXFAMask) + xCount;
		    /* put the new source pixel in */
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS++) & cResPixMask)]));
		if (extraPixel < cExtraPixelDec2)
		{
		        /* put in the extra pixel */
		    gray0 = ((gray0 & ~0xFF) << 8) | (gray0 & 0xFF00);
			/* put in the next source pixel */
	            gray0 += ((UInt32) 
                      ( *(UInt16 *) &(pRescaleTop[
	              (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	            gray0 += ((UInt32) 
                      ( *(UInt16 *) &(pRescaleBot[
	              (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
		    if (extraPixel >= cExtraPixelDec2)
			{
			pS++;
			extraPixel -= cExtraPixelDec1;
			}
		    /* else state does not change */
		}
		else
		{
		        /* no extra pixel, put in the next source 
			 * pixel 
			 */
		    gray0 = (gray0 & ~0xFF) << 8;
	            gray0 += (UInt32) 
                      ( *(UInt16 *) &(pRescaleTop[
	              (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	            gray0 += (UInt32) 
                      ( *(UInt16 *) &(pRescaleBot[
	              (UCHAR_ACCESS(pS++) & cResPixMask)]));
			/* put in the extra pixel */
		    gray0 = ((gray0 & 0xFF00) >> 8) | (gray0 & ~0xFF);
		    extraPixel -= cExtraPixelDec1;
		}
 	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel1Finish:
		    /* can only get in this state if xCount > 2.
		     * if xCount == 2, we will be in 
		     * cExtraPixel1FinishXCount2 below.  if xCount == 1
		     * we will never leave the cExtraPixel1Start case.
		     */
		gray0 = (UInt32)
                  (*(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  (*(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFF00) << 8;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = gray0 << 8;
		pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel2Finish:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
                  (( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)])) >> 8);
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        gray0 = gray0 << 16;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

		    /* we know that xCount is at least 2 so that we
		     * can safely put in one pixel without worrying about
		     * needing an extra pixel.
		     */
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel3Finish:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
                  ( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]) >> 8);
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        gray0 = gray0 << 8;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel4Finish:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
                  (( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)])) >> 8);
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel5OrOverCont:
	        gray0 = (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        gray0 = gray0 << 8;
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
	        pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]))) >> 8;
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    case cExtraPixel1FinishXCount2:
		gray0 = 
                  *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]);
	        gray0 += ((UInt32) 
                  *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFF00) << 8;
	        pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = gray0 << 8;
		pS++;
	        xIOCoord = xIOCoordNext;
	        xIOCoordNext += xIOIncrement;

	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleTop[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]));
	        gray0 += (UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[(xIOCoord & cResXFMask) +
	          (((xIOCoordNext ^ xIOCoord) & cResXLSMask) << 3) +
	          (UCHAR_ACCESS(pS) & cResPixMask)]));
		gray0 = (gray0 & 0xFFFFFF00) | 
                  ( *(UInt16 *) &(pRescaleTop[
	          (UCHAR_ACCESS(&pS[-sBpl]) & cResPixMask)]) >> 8);
	        gray0 += ((UInt32) 
                  ( *(UInt16 *) &(pRescaleBot[
	          (UCHAR_ACCESS(pS) & cResPixMask)]))) >> 8;
		pS++;
		xIOCoord &= 0xFFFFC000;
		xIOCoordNext = xIOCoord + xIOIncrement;

	        ULONG_ACCESS(pD) = gray0;
	        pD += 4;
	        dCol += 4;
		break;
	    }
	}
    }
}

/*

$Log: mirescl2.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.4  1998/01/13 23:53:11  ddavies
mirescl2.c -> /product/ipcore/ipshared/src/RCS/mirescl2.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1995/12/13 02:51:18  ejaquith
mirescl2.c -> /product/ipcore/ipshared/src/RCS/mirescl2.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/10/18  04:03:55  ddavies
 * mirescl2.c -> /product/ipcore/ipshared/src/RCS/mirescl2.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.1  1994/09/09  00:15:25  lperry
 * Initial revision
 *
 * Revision 1.4  1994/02/04  20:18:36  dferrell
 * mirescal.c -> /product/ipcore/ipshared/src/RCS/mirescal.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:25:03  ddavies
 * mirescal.c -> /product/ipcore/ipshared/src/RCS/mirescal.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:50:03  ddavies
 * mirescal.c -> /product/ipcore/ipshared/src/RCS/mirescal.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


