/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * mirescon.c
 *
 * Private resolution conversion routines:
 *	Primitives that produces a line of output directly from a line
 *	of input:
 *		void	ip_rescon1_5x4bppLine()
 *		void	ip_rescon2x4bppLine()
 *		void	ip_rescon1_5x8bppLine()
 *		void	ip_rescon2x8bppLine()
 *
 *	Line averaging primitives:
 *		void	ip_average4bppLines()
 *		void	ip_average8bppLines()
 *
 *	Down-conversion primitives:
 *		void	ip_rescon2To1Line4bpp()
 *		void	ip_rescon2To1Line8bpp()
 *		void	ip_rescon3To1Line4bpp()
 *		void	ip_rescon3To1Line8bpp()
 */


#include <stdlib.h>
#include <stddef.h>

#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "imageref.pub"
#include "shrrast.pub"
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mirescon.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* This file contains the resolution conversion primitives.  These are
 * called by routines in i_rescon.c.  Each one produces one line of output.
 * The first group listed above expand a line of input horizontally by
 * 1.5x or 2x.  The second group produce a line of output by averaging
 * two lines of input.  The overall scheme is explained in i_rescon.c.
 *
 * The 1.5x procedures turn each pair of input pixels into three output
 * pixels.  They use table lookup to do this.  The 4bpp tables turn a
 * byte into 12 bits.  These bits are shifted into place as needed.
 *
 * The 8 bpp tables produce 24 bits.  This seems perfectly natural until you
 * realize that the tables are indexed by bytes.  Where do we get the
 * information for the second byte of the pair?  We don't.  We just return
 * four bytes containing | 00 | byte/2 | byte | byte/2 |.  At 1.5x
 * expansion, only one of the byte/2 terms will be used.  We choose which
 * one by masking off the unwanted one before doing the require shift and OR
 * op.
 *
 * At 2x, each pixel affects the pixels on *both* sides of it.  Therefore,
 * at 4bpp, each pair of pixels (AB) turns into 20 bits:
 *
 *	|  0  |  0  |  0  | A/2 |  A  | (A+B)/2 |  B  | B/2 |
 *
 * This quantity is shifted as necessary before being added into the
 * word that's forming the output word.
 *
 * Note that the 8bpp expansion table is the same for 1.5x and 2x.  At 2x,
 * we use both "byte/2" terms instead of throwing one of them away.
 *
 * The averaging routines do the same job, but use different methods.  In
 * both cases, we operate on entire 32-bit words of pixels.  The 8 bit case
 * is easier because we cheat.  We shift both source words right by 1,
 * mask off bits 31, 23, 15 and 7 in both, then add them together.  This
 * produces the average of each pair of source pixels, almost.  Since we're
 * tossing the least significant bits, we can be one value too low out of
 * 256.  Big deal.
 *
 * In the 4 bpp case, being off by 1/16 of full value *is* is big deal.
 * We can't just toss the LSB's.  Instead, we mask off the MSB's as before
 * add the resulting words and shift the result right 1.  This gives the
 * correct average for the 3 LS bits of each pixels.  We then select the
 * MSB's out of the original source words, shift them right 1 and add them
 * together.  This gives us the average value of the MSBs.  Finally, we
 * clear bits 31, 27, 23, 19, 15, 11, 7 and 3 of the first result and add
 * in the second result.  This is the sum of the average of the 3 LSBs
 * of each pixel pair and the MSB of each pixel pair, which equals the
 * average of each pixel pair.
 *
 * One of the design goals is to have the procedures be so simple that
 * the compiler can keep all the important variables in registers.  That's
 * the main reason why we have so many small procedures instead of a few
 * larger ones.
 */

/************************ 4 BIT/PIXEL ROUTINES  **************************/

/* PRIVATE */
/**************************************************************************
 * ip_rescon1_5x4bppLine()
 *	Generates a dest line 1.5x as wide as a source line.  The dest
 *	pixels are interpolated.
 *	Used during resolution conversion from 200x200x4 to 300x300x4.
 *
 *	Input:
 *		pS:		pointer to left end of source line
 *		wS:		width of source line in pixels
 *		pD:		pointer to left end of dest line
 *		pResconTable:	pointer to beginning of expansion table
 *
 *	This routine writes into the right frame.
 **************************************************************************/
void  CDECL 
ip_rescon1_5x4bppLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    UInt8      *pD,
    UInt32     *pResconTable)
{
    PROLOGUE;
    UInt32	dataOut;
    UInt8      *endPtr;	/* used to find end of line */	

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Compute the pointer to the last byte.  When we get
	 * there, we'll decide whether one or both nibbles are valid.
	 */
    endPtr = pS + (wS-1) / 2;

	/* use this block to fill in a group of 3 output words.
	 * This is equivalent to 2 input words.  When there are less
	 * than 2 input words left, we go to the loop below that does
	 * things one at a time.  To avoid the silly xors, we have two
	 * copies of the loop.
	 */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

    for ( ; (UInt32) endPtr - (UInt32) pS >= 8; pS += 8, pD += 12)
    {
	    /* first destination word */
	dataOut = pResconTable[pS[0]];		/* x x x x x A AB B */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[1]];		/* x x A AB B C CD D */
	dataOut = (dataOut << 8) |
		  (pResconTable[pS[2]] >> 4);	/* A AB B C CD D E EF */
	ULONG_ACCESS(&pD[0]) = dataOut;

	dataOut = pResconTable[pS[2]] & 0xF;	/* x x x x x x x F */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[3]];		/* x x x x F G GH H */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[4]];		/* x F G GH H I IJ J */
	dataOut = (dataOut << 4) |
		  (pS[5] >> 4);			/* F G GH H I IJ J K */
	ULONG_ACCESS(&pD[4]) = dataOut;

	dataOut = pResconTable[pS[5]] & 0xFF;	/* x x x x x x KL L */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[6]];		/* x x x KL L M MN N */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[7]];		/* KL L M MN N O OP P */
	ULONG_ACCESS(&pD[8]) = dataOut;
    }

	/* At this point, there are 8 or fewer source bytes left,
	 * so we check before turning each byte into 12 bits.  
	 * The byte pointed to by pS always has at least one valid
	 * nibble in it.  If the image ended in the middle of the byte,
	 * It will expand to the middle of a dest byte.  The call
	 * to clearRightFrame in the routine that calls this
	 * one will smash anything beyond that to 0, so we don't
	 * bother to try to calculate it correctly.
	 */

	/* process byte 0 */
    dataOut = pResconTable[pS[0]] << 12;		/* x x A AB B x x x */
    if (pS == endPtr)
    {
	ULONG_ACCESS(&pD[0]) = dataOut << 8;
	return;
    }

	/* process byte 1 */
    dataOut |= pResconTable[pS[1]];
    dataOut <<= 8;					/* A AB B C CD D x x */
    if ((pS + 1) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[0]) = dataOut;
	return;
    }

	/* process byte 2 */
    dataOut |= pResconTable[pS[2]] >> 4;	/* A AB B C CD D E EF */
    ULONG_ACCESS(&pD[0]) = dataOut;
    dataOut = (pS[2] & 0xF) << 12;		/* x x x x F x x x */

    if ((pS + 2) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[4]) = dataOut << 16;	/* F x x x x x x x */
	return;
    }

	/* process byte 3 */
    dataOut |= pResconTable[pS[3]];		/* x x x x F G GH H */
    dataOut <<= 12;				/* x F G GH H x x x */
    if ((pS + 3) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[4]) = dataOut << 4;
	return;
    }

	/* process byte 4 */
    dataOut |= pResconTable[pS[4]];		/* x F G GH H I IJ J */
    dataOut <<= 4;				/* F G GH H I IJ J x */
    if ((pS + 4) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[4]) = dataOut;
	return;
    }

	/* process byte 5 */
	/* we know that at least the upper nibble of this byte is
	 * correct, so put it into dataOut and write that out.
	 */
    dataOut |= (pS[5] >> 4);			/* F G GH H I IJ J K */
    ULONG_ACCESS(&pD[4]) = dataOut;
    dataOut = (pResconTable[pS[5]] & 0xFF) << 12; /* x x x KL L x x x */

    if ((pS + 5) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[8]) = dataOut << 12;	/* KL L x x x x x x */
	return;
    }

	/* process byte 6 */
    dataOut |= pResconTable[pS[6]];		/* x x x KL L M MN N */
    dataOut <<= 12;				/* KL L M MN N x x x */
    if ((pS + 6) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[8]) = dataOut;
	return;
    }

	/* process byte 7.  We must end here since if there were
	 * 8 or more bytes to process, we would still be in the
	 * loop above
	 */
    dataOut |= pResconTable[pS[7]];		/* KL L M MN N O OP P */
    ULONG_ACCESS(&pD[8]) = dataOut;
    return;


#else		/* data in little-endian format */


    for ( ; (UInt32) endPtr - (UInt32) pS >= 8; pS += 8, pD += 12)
    {
	    /* first destination word */
	dataOut = pResconTable[pS[3]];		/* x x x x x A AB B */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[2]];		/* x x A AB B C CD D */
	dataOut = (dataOut << 8) |
		  (pResconTable[pS[1]] >> 4);	/* A AB B C CD D E EF */
	ULONG_ACCESS(&pD[0]) = dataOut;

	dataOut = pResconTable[pS[1]] & 0xF;	/* x x x x x x x F */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[0]];		/* x x x x F G GH H */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[7]];		/* x F G GH H I IJ J */
	dataOut = (dataOut << 4) |
		  (pS[6] >> 4);			/* F G GH H I IJ J K */
	ULONG_ACCESS(&pD[4]) = dataOut;

	dataOut = pResconTable[pS[6]] & 0xFF;	/* x x x x x x KL L */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[5]];		/* x x x KL L M MN N */
	dataOut = (dataOut << 12) |
		   pResconTable[pS[4]];		/* KL L M MN N O OP P */
	ULONG_ACCESS(&pD[8]) = dataOut;
    }

	/* At this point, there are 8 or fewer source bytes left,
	 * so we check before turning each byte into 12 bits.  
	 * The byte "pointed to" by pS always has at least one valid
	 * nibble in it.  If the image ended in the middle of the byte,
	 * It will expand to the middle of a dest byte.  The call
	 * to clearRightFrame in the routine that calls this
	 * one will smash anything beyond that to 0, so we don't
	 * bother to try to calculate it correctly.
	 * Note that we pretend that pS is big endian when
	 * comparing it with endPtr, but actually index from it
	 * in little endian fashion since this is little endian code.
	 */

	/* process byte 0 */
    dataOut = pResconTable[pS[3]] << 12;		/* x x A AB B x x x */
    if (pS == endPtr)
    {
	ULONG_ACCESS(&pD[0]) = dataOut << 8;
	return;
    }

	/* process byte 1 */
    dataOut |= pResconTable[pS[2]];
    dataOut <<= 8;					/* A AB B C CD D x x */
    if ((pS + 1) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[0]) = dataOut;
	return;
    }

	/* process byte 2 */
    dataOut |= pResconTable[pS[1]] >> 4;	/* A AB B C CD D E EF */
    ULONG_ACCESS(&pD[0]) = dataOut;
    dataOut = (pS[1] & 0xF) << 12;		/* x x x x F x x x */

    if ((pS + 2) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[4]) = dataOut << 16;	/* F x x x x x x x */
	return;
    }

	/* process byte 3 */
    dataOut |= pResconTable[pS[0]];		/* x x x x F G GH H */
    dataOut <<= 12;				/* x F G GH H x x x */
    if ((pS + 3) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[4]) = dataOut << 4;
	return;
    }

	/* process byte 4 */
    dataOut |= pResconTable[pS[7]];		/* x F G GH H I IJ J */
    dataOut <<= 4;				/* F G GH H I IJ J x */
    if ((pS + 4) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[4]) = dataOut;
	return;
    }

	/* process byte 5 */
	/* we know that at least the upper nibble of this byte is
	 * correct, so put it into dataOut and write that out.
	 */
    dataOut |= (pS[6] >> 4);			/* F G GH H I IJ J K */
    ULONG_ACCESS(&pD[4]) = dataOut;
    dataOut = (pResconTable[pS[6]] & 0xFF) << 12; /* x x x KL L x x x */

    if ((pS + 5) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[8]) = dataOut << 12;	/* KL L x x x x x x */
	return;
    }

	/* process byte 6 */
    dataOut |= pResconTable[pS[5]];		/* x x x KL L M MN N */
    dataOut <<= 12;				/* KL L M MN N x x x */
    if ((pS + 6) == endPtr)	/* line ends in this byte */
    {
	ULONG_ACCESS(&pD[8]) = dataOut;
	return;
    }

	/* process byte 7.  We must end here since if there were
	 * 8 or more bytes to process, we would still be in the
	 * loop above
	 */
    dataOut |= pResconTable[pS[4]];		/* KL L M MN N O OP P */
    ULONG_ACCESS(&pD[8]) = dataOut;
    return;

#endif	/* _ALPACA_IMAGE_FMT_ == cBigEndianFmt */

}


/* PRIVATE */
/**************************************************************************
 * ip_rescon2x4bppLine()
 *	Generates a dest line 2x as wide as a source line.  The dest
 *	pixels are interpolated.
 *	Used during when doubling resolution of 4bpp images.
 *
 *	Input:
 *		pS:		pointer to left end of source line
 *		wS:		width of source line in pixels
 *		pD:		pointer to left end of dest line
 *		pResconTable:	pointer to beginning of expansion table
 *
 *	This routine writes into the right frame.
 **************************************************************************/
void  CDECL 
ip_rescon2x4bppLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    UInt8      *pD,
    UInt32     *pResconTable)
{
    PROLOGUE;
    UInt32	dataOut;
    UInt8      *endPtr;	/* used to find end of line */	

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Compute the pointer to the last byte.  When we get
	 * there, we'll decide whether one or both nibbles are valid.
	 */
    endPtr = pS + (wS-1) / 2;

	/* use this block to fill in a group of 2 output words.
	 * We could do 1 output word at a time, but this is a little
	 * more efficient.
	 * The weird code to calculate the nibble that links each
	 * byte is there because we can't use tables.  If we put
	 * B/2 in the LS nibble of a table entry and added A/2
	 * from the next table entry, we'd lose the LS bit of each.
	 * With only 16 levels, this is easily seen.  It looks like
	 * faint vertical stripes.  Instead, we compute the average
	 * value of the two.
	 * When there is less than 1 input word left, we go to the loop
	 * below that does things one at a time.
	 * To avoid the silly address xors, we have two copies of the loop.
	 */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

    for ( ; (UInt32) endPtr - (UInt32) pS >= 4; pS += 4, pD += 8)
    {
	    /* first destination word */
	dataOut = pResconTable[pS[0]];		/* x x x 0 A AB B 0 */
	dataOut |= ((dataOut & 0xF0) + (UInt32) pS[1] + 1) >> 5;
						/* x x x 0 A AB B BC */
	dataOut = (dataOut << 16) |
		   (pResconTable[pS[1]]);	/* A AB B BC C CD D 0 */
	dataOut |= ((dataOut & 0xF0) + (UInt32) pS[2] + 1) >> 5;
						/* A AB B BC C CD D DE */
	ULONG_ACCESS(&pD[0]) = dataOut;

	dataOut = pResconTable[pS[2]];		/* x x x 0 E EF F 0 */
	dataOut |= ((dataOut & 0xF0) + (UInt32) pS[3] + 1) >> 5;
						/* x x x 0 E EF F FG */
	dataOut = (dataOut << 16) |
		   pResconTable[pS[3]];		/* E EF F FG G GH H 0 */
	dataOut |= ((dataOut & 0xF0) + (UInt32) pS[4] + 1) >> 5;
						/* E EF F FG G GH H HI */
	ULONG_ACCESS(&pD[4]) = dataOut;
    }

	/* At this point, there are 4 or fewer source bytes left,
	 * so we check before filling each byte pair.  We don't want
	 * to average a frame nibble with a real image nibble, so
	 * we replicate the rightmost image nibble if the image
	 * ends in the middle of a byte.
	 */
	/* process byte 0 */
    if (pS == endPtr)
    {
	if (wS & 0x1)	/* only left nibble is valid  */
	    dataOut = pResconTable[pS[0] | (pS[0] >> 4)];
							/* x x x 0 A AA A 0 */
	else
	{
	    dataOut = pResconTable[pS[0]];		/* x x x 0 A AB B 0 */
		/* Since every nibble should turn into two nibbles, replicate
		 * the LS nibble.
		 */
	    dataOut |= (dataOut & 0xF0) >> 4;		/* x x x 0 A AB B B */
	}
	ULONG_ACCESS(&pD[0]) = dataOut << 16;
	return;
    }

	/* The line doesn't end in this byte and we know that at
	 * least the MS nibble of the next byte is valid, so
	 * insert the BC nibble.
	 */
    dataOut = pResconTable[pS[0]];			/* x x x 0 A AB B 0 */
    dataOut |= ((dataOut & 0xF0) + (UInt32) pS[1] + 1) >> 5;
							/* x x x 0 A AB B BC */
    dataOut <<= 16;					/* A AB B BC x x x x */


	/* process byte 1 */
    if ((pS + 1) == endPtr)
    {
	if (wS & 0x1)	/* only left nibble is valid  */
	    dataOut |= pResconTable[pS[1] | (pS[1] >> 4)];
							/* A AB B BC C CC C 0 */
	else
	{
	    dataOut |= pResconTable[pS[1]];		/* A AB B BC C CD D 0 */
	    dataOut |= (dataOut & 0xF0) >> 4;		/* A AB B BC C CD D D */
	}
	ULONG_ACCESS(&pD[0]) = dataOut;
	return;
    }

	/* The line doesn't end in this byte and we know that at
	 * least the MS nibble of the next byte is valid, so
	 * insert the BC nibble.
	 */
    dataOut |= pResconTable[pS[1]];		/* A AB B BC C CD D 0 */
    dataOut |= ((dataOut & 0xF0) + (UInt32) pS[2] + 1) >> 5;
						/* A AB B BC C CD D DE */
    ULONG_ACCESS(&pD[0]) = dataOut;


	/* process byte 2 */
    if ((pS + 2) == endPtr)
    {
	if (wS & 0x1)	/* only left nibble is valid  */
	    dataOut = pResconTable[pS[2] | (pS[2] >> 4)];
							/* x x x 0 E EE E 0 */
	else
	{
	    dataOut = pResconTable[pS[2]];		/* x x x 0 E EF F 0 */
		/* Since every nibble should turn into two nibbles, replicate
		 * the LS nibble.
		 */
	    dataOut |= (dataOut & 0xF0) >> 4;		/* x x x 0 E EF F F */
	}
	ULONG_ACCESS(&pD[4]) = dataOut << 16;
	return;
    }

	/* The line doesn't end in this byte and we know that at
	 * least the MS nibble of the next byte is valid, so
	 * insert the FG nibble.
	 */
    dataOut = pResconTable[pS[2]];			/* x x x 0 E EF F 0 */
    dataOut |= ((dataOut & 0xF0) + (UInt32) pS[3] + 1) >> 5;
							/* x x x 0 E EF F FG */
    dataOut <<= 16;					/* E EF F FG x x x x */


	/* process byte 3.  This is the last byte; if there were more,
	 * we would still be in the loop above.
	 */
    if (wS & 0x1)	/* only left nibble is valid  */
	dataOut |= pResconTable[pS[3] | (pS[3] >> 4)];
							/* E EF F FG G GG G 0 */
    else
    {
	dataOut |= pResconTable[pS[3]];			/* E EF F FG G GH H 0 */
	dataOut |= (dataOut & 0xF0) >> 4;		/* E EF F FG G GH H H */
    }
    ULONG_ACCESS(&pD[4]) = dataOut;
    return;


#else		/* data in little-endian format */


    for ( ; (UInt32) endPtr - (UInt32) pS >= 4; pS += 4, pD += 8)
    {
	    /* first destination word */
	dataOut = pResconTable[pS[3]];		/* x x x 0 A AB B 0 */
	dataOut |= ((dataOut & 0xF0) + (UInt32) pS[2] + 1) >> 5;
						/* x x x 0 A AB B BC */
	dataOut = (dataOut << 16) |
		   (pResconTable[pS[2]]);	/* A AB B BC C CD D 0 */
	dataOut |= ((dataOut & 0xF0) + (UInt32) pS[1] + 1) >> 5;
						/* A AB B BC C CD D DE */
	ULONG_ACCESS(&pD[0]) = dataOut;

	dataOut = pResconTable[pS[1]];		/* x x x 0 E EF F 0 */
	dataOut |= ((dataOut & 0xF0) + (UInt32) pS[0] + 1) >> 5;
						/* x x x 0 E EF F FG */
	dataOut = (dataOut << 16) |
		   pResconTable[pS[0]];		/* E EF F FG G GH H 0 */
	dataOut |= ((dataOut & 0xF0) + (UInt32) pS[7] + 1) >> 5;
						/* E EF F FG G GH H HI */
	ULONG_ACCESS(&pD[4]) = dataOut;
    }

	/* At this point, there are 4 or fewer source bytes left,
	 * so we check before filling each byte pair.  We don't want
	 * to average a frame nibble with a real image nibble, so
	 * we replicate the rightmost image nibble if the image
	 * ends in the middle of a byte.
	 */
	/* process byte 0 */
    if (pS == endPtr)
    {
	if (wS & 0x1)	/* only left nibble is valid  */
	    dataOut = pResconTable[pS[3] | (pS[3] >> 4)];
							/* x x x 0 A AA A 0 */
	else
	{
	    dataOut = pResconTable[pS[3]];		/* x x x 0 A AB B 0 */
		/* Since every nibble should turn into two nibbles, replicate
		 * the LS nibble.
		 */
	    dataOut |= (dataOut & 0xF0) >> 4;		/* x x x 0 A AB B B */
	}
	ULONG_ACCESS(&pD[0]) = dataOut << 16;
	return;
    }

	/* The line doesn't end in this byte and we know that at
	 * least the MS nibble of the next byte is valid, so
	 * insert the BC nibble.
	 */
    dataOut = pResconTable[pS[3]];			/* x x x 0 A AB B 0 */
    dataOut |= ((dataOut & 0xF0) + (UInt32) pS[2] + 1) >> 5;
							/* x x x 0 A AB B BC */
    dataOut <<= 16;					/* A AB B BC x x x x */


	/* process byte 1 */
    if ((pS + 1) == endPtr)
    {
	if (wS & 0x1)	/* only left nibble is valid  */
	    dataOut |= pResconTable[pS[2] | (pS[2] >> 4)];
							/* A AB B BC C CC C 0 */
	else
	{
	    dataOut |= pResconTable[pS[2]];		/* A AB B BC C CD D 0 */
	    dataOut |= (dataOut & 0xF0) >> 4;		/* A AB B BC C CD D D */
	}
	ULONG_ACCESS(&pD[0]) = dataOut;
	return;
    }

	/* The line doesn't end in this byte and we know that at
	 * least the MS nibble of the next byte is valid, so
	 * insert the BC nibble.
	 */
    dataOut |= pResconTable[pS[2]];		/* A AB B BC C CD D 0 */
    dataOut |= ((dataOut & 0xF0) + (UInt32) pS[1] + 1) >> 5;
						/* A AB B BC C CD D DE */
    ULONG_ACCESS(&pD[0]) = dataOut;


	/* process byte 2 */
    if ((pS + 2) == endPtr)
    {
	if (wS & 0x1)	/* only left nibble is valid  */
	    dataOut = pResconTable[pS[1] | (pS[1] >> 4)];
							/* x x x 0 E EE E 0 */
	else
	{
	    dataOut = pResconTable[pS[1]];		/* x x x 0 E EF F 0 */
		/* Since every nibble should turn into two nibbles, replicate
		 * the LS nibble.
		 */
	    dataOut |= (dataOut & 0xF0) >> 4;		/* x x x 0 E EF F F */
	}
	ULONG_ACCESS(&pD[4]) = dataOut << 16;
	return;
    }

	/* The line doesn't end in this byte and we know that at
	 * least the MS nibble of the next byte is valid, so
	 * insert the FG nibble.
	 */
    dataOut = pResconTable[pS[1]];			/* x x x 0 E EF F 0 */
    dataOut |= ((dataOut & 0xF0) + (UInt32) pS[0] + 1) >> 5;
							/* x x x 0 E EF F FG */
    dataOut <<= 16;					/* E EF F FG x x x x */


	/* process byte 3.  This is the last byte; if there were more,
	 * we would still be in the loop above.
	 */
    if (wS & 0x1)	/* only left nibble is valid  */
	dataOut |= pResconTable[pS[0] | (pS[0] >> 4)];
							/* E EF F FG G GG G 0 */
    else
    {
	dataOut |= pResconTable[pS[0]];			/* E EF F FG G GH H 0 */
	dataOut |= (dataOut & 0xF0) >> 4;		/* E EF F FG G GH H H */
    }
    ULONG_ACCESS(&pD[4]) = dataOut;
    return;


#endif

}


/* PRIVATE */
/**************************************************************************
 * ip_average4bppLines()
 *	Generates an interpolated dest line by averaging two existing
 *	lines of the right width.
 *	Used during during 1.5x and 2x resolution conversion.
 *
 *	This routine averages the two lines.  Since there are only 4
 *	bits/pixel, it can't afford to ignore any of them.  Therefore,
 *	it goes to pains to add a word containing the 3 LS bits of
 *	each pixel, then add the MS bits of each pixel, then add these
 *	together.
 *
 *	Input:
 *		pTop:	pointer to top source line
 *		bpl:	bytes per line in all three lines	
 *		wD:	width of dest line in pixels
 *
 *	This routine writes into the right frame.
 **************************************************************************/
void  CDECL 
ip_average4bppLines (
    IPState    *pIPState,
    UInt8      *pTop,
    UInt32	bpl,
    Int32	wD)
{
    PROLOGUE;
    UInt32	dataOut, upperIn, lowerIn;
    UInt8      *endPtr;	/* used to find end of line */	


	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* since we want to make sure that we get the last pixel and
	 * width may be odd, round up the width.  This will write into
	 * the right frame so it must be cleared later.
	 */
    endPtr = pTop + (wD+1) / 2;

	/* go across the line in units of words.  This allows us to
	 * use the same code for both big and little endian 
	 * platforms.
	 */
    for ( ; (UInt32) endPtr > (UInt32) pTop; pTop += 4)
    {
	    /* Add the 3 LSB's in each pixel first */
	upperIn = ULONG_ACCESS(pTop) & 0x77777777;
	lowerIn = ULONG_ACCESS(&pTop[2*bpl]) & 0x77777777;

	    /* now that the MSB's are gone, we know there won't be
	     * any carries from pixel to pixel.  Add the pixels
	     * and divide by 2 to get the average.
	     */
	dataOut = (upperIn + lowerIn + 0x11111111) / 2;

	    /* go average the MSBs of each pixel.  Note that we
	     * have to do the shift before doing the adds to
	     * avoid having the topmost bit fall off the end.
	     */
	upperIn = (ULONG_ACCESS(pTop) & 0x88888888) / 2;
	lowerIn = (ULONG_ACCESS(&pTop[2*bpl]) & 0x88888888) / 2;

	upperIn += lowerIn;			/* get the average */
	dataOut &= 0x77777777;			/* clear out spaces for the new
						 * MS bits */
	dataOut += upperIn;			/* add them in to get the
						 * real fractions */
	ULONG_ACCESS(&pTop[bpl]) = dataOut;	/* write the word */
    }
}



/************************ 8 BIT/PIXEL ROUTINES  **************************/


/* PRIVATE */
/**************************************************************************
 * ip_rescon1_5x8bppLine()
 *	Generates a dest line 1.5x as wide as a source line.  The dest
 *	pixels are interpolated.
 *	Used during resolution conversion from 200x200x8 to 300x300x8.
 *
 *	Input:
 *		pS:		pointer to left end of source line
 *		wS:		width of source line in pixels
 *		pD:		pointer to left end of dest line
 *
 *	This routine writes into the right frame.
 **************************************************************************/
void  CDECL 
ip_rescon1_5x8bppLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	dataOut, temp;
    UInt8      *endPtr;		/* used to find end of line */	

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Point to the last byte.  Remember, each pixel is
	 * an entire byte. */
    endPtr = pS + wS - 1;

	/* use this block to fill in a group of 3 output words.
	 * This is equivalent to 2 input words.  When there are less
	 * than 2 input words left, we go to the loop below that does
	 * things one at a time.  Note that we do only 32-bit accesses,
	 * so this code runs on all platforms.
	 */

    for ( ; (UInt32) endPtr - (UInt32) pS >= 8; pS += 8, pD += 12)
    {
	    /* first destination word.  We're converting
	     * [A B C D][E F G H] to [A AB B C][CD D E EF][F G GH H] 
	     * where AB means (A + B)/2
	     */
	dataOut = ULONG_ACCESS(&pS[0]) & 0xFEFF0000;	/* A B 0 0 */
	dataOut += ULONG_ACCESS(&pS[0]) >> 8;		/* A 2AB B C */
	dataOut = (dataOut & 0xFE00FFFF) |
		  ((dataOut & 0x01FE0000) / 2); 	/* A AB B C */
	ULONG_ACCESS(&pD[0]) = dataOut;

	dataOut = (ULONG_ACCESS(&pS[0]) << 15) |
		  (ULONG_ACCESS(&pS[4]) >> 17);		/* C/2 D/2 E/2 F/2 */
	dataOut &= 0x7F7F7F7F;				/* get rid of overlap */
	temp = (dataOut & 0x00FF0000) |
		(ULONG_ACCESS(&pS[4]) >> 25);		/* 0 D/2 0 E/2 */
	temp |= temp << 8;				/* D/2 D/2 E/2 E/2 */
	dataOut += temp;				/* CD   D   E  EF  */
	ULONG_ACCESS(&pD[4]) = dataOut;

	dataOut = (ULONG_ACCESS(&pS[4]) & 0x00FFFEFF) << 8;
							/* F G H 0 */
	dataOut += ULONG_ACCESS(&pS[4]) & 0x0000FFFF;
							/* F G 2GH H */
	dataOut = (dataOut & 0xFFFE00FF) | ((dataOut & 0x0001FE00)/2);
							/* F G GH H */
	ULONG_ACCESS(&pD[8]) = dataOut;
    }

	/* At this point, there are 8 or fewer source bytes left.
	 * We again create 1, 2 or 3 output words.  Now, you might
	 * think that we have to fix up all the terms containing
	 * averages (like BC, CD, etc).  It turns out that the
	 * fixups are unnecessary.  Let's look at a byte pair, like [A B],
	 * that turns into a triple like [A AB B].  If only A is
	 * valid, the output image will also only contain A, so the
	 * AB term will be wiped out by clearRightFrame.  If B is
	 * also valid, then we want all three terms.  Therefore,
	 * we can calculate along with impunity.
	 * In the first word, we know that at least part of [A B C D] is
	 * in the image, so the rest of it must still be in the
	 * frame, so we can use the whole thing.
	 */
    dataOut = ULONG_ACCESS(&pS[0]) & 0xFEFF0000;	/* A B 0 0 */
    dataOut += ULONG_ACCESS(&pS[0]) >> 8;		/* A 2AB B C */
    dataOut = (dataOut & 0xFE00FFFF) |
	      ((dataOut & 0x01FE0000) / 2); 		/* A AB B C */
    ULONG_ACCESS(&pD[0]) = dataOut;

	/* If byte 0 or byte 1 was the last byte, we're done.  If byte 2 was
	 * the last byte, we're also done because it was only supposed
	 * to expand to 1 byte (C above).
	 */
    if ((pS + 2) >= endPtr)
	return;

	/* We know that at least byte 3 is ok, so it's ok to make
	 * the CD term.  The next word, [E F G H] is either partly
	 * in the image or entirely in the frame.  Since we have
	 * 32 pixels of frame, we know we won't run out of frame
	 * when using it.
	 */
    dataOut = (ULONG_ACCESS(&pS[0]) << 15) |
	      (ULONG_ACCESS(&pS[4]) >> 17);		/* C/2 D/2 E/2 F/2 */
    dataOut &= 0x7F7F7F7F;				/* get rid of overlap */
    temp = (dataOut & 0x00FF0000) |
	    (ULONG_ACCESS(&pS[4]) >> 25);		/* 0 D/2 0 E/2 */
    temp |= temp << 8;					/* D/2 D/2 E/2 E/2 */
    dataOut += temp;					/* CD   D   E  EF  */
    ULONG_ACCESS(&pD[4]) = dataOut;

	/* if byte E (byte 4) was the last byte, we're done.  If byte
	 * F( or G or H) was the last pixel, we need to make the final
	 * output word.
	 */
    if ((pS + 4) >= endPtr)
	return;

    dataOut = (ULONG_ACCESS(&pS[4]) & 0x00FFFEFF) << 8;	/* F G H 0 */
    dataOut += ULONG_ACCESS(&pS[4]) & 0x0000FFFF;	/* F G 2GH H */
    dataOut = (dataOut & 0xFFFE00FF) | ((dataOut & 0x0001FE00)/2);
							/* F G GH H */
    ULONG_ACCESS(&pD[8]) = dataOut;
    return;

}


/* PRIVATE */
/**************************************************************************
 * ip_rescon2x8bppLine()
 *	Generates a dest line 2x as wide as a source line.  The dest
 *	pixels are interpolated.
 *	Used during resolution conversion from 200x200x8 to 400x400x8.
 *
 *	Input:
 *		pS:		pointer to left end of source line
 *		wS:		width of source line in pixels
 *		pD:		pointer to left end of dest line
 *
 *	This routine writes into the right frame.
 **************************************************************************/
void  CDECL 
ip_rescon2x8bppLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    UInt8      *pD)
{
    PROLOGUE;
    UInt32	dataOut, temp;
    UInt8      *endPtr;	/* used to find end of line */	

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Point to the last byte.  Remember, each pixel is
	 * an entire byte. */
    endPtr = pS + wS - 1;

	/* use this block to fill in a group of 2 output words.
	 * This is equivalent to 1 input word.  When there is less
	 * than 1 input word left, we go to the loop below that does
	 * things one at a time.  Note that we do only 32-bit accesses,
	 * so this code runs on all platforms.
	 */

    for ( ; (UInt32) endPtr - (UInt32) pS >= 4; pS += 4, pD += 8)
    {
	    /* first destination word.  We're converting
	     * [A B C D] to [A AB B BC][C CD D DE] 
	     * where AB means (A + B)/2
	     */
	dataOut = ULONG_ACCESS(&pS[0]) & 0xFEFF0000;	/* A B 0 0 */
	dataOut += ULONG_ACCESS(&pS[0]) >> 8;		/* A 2AB B C */
	dataOut &= 0xFFFFFEFF;				/* take out B.0 */
	dataOut += (dataOut >> 8) & 0xFF;		/* A 2AB B 2BC */
	dataOut = (dataOut &  0xFE00FE00) |
		  ((dataOut & 0x01FE01FE) / 2); 	/* A AB B BC */
	ULONG_ACCESS(&pD[0]) = dataOut;

	temp = ULONG_ACCESS(&pS[0]) & 0xFEFE;
	dataOut = temp << 16;				/* C D 0 0 */
	dataOut += temp << 8;				/* C 2CD D 0 */
	dataOut += temp & 0xFF;				/* C 2CD D D */
	dataOut += UCHAR_ACCESS(&pS[4]);		/* C 2CD D 2DE */
	dataOut = (dataOut &  0xFE00FE00) |
		  ((dataOut & 0x01FE01FE) / 2); 	/* C CD D DE */
	ULONG_ACCESS(&pD[4]) = dataOut;

    }

	/* At this point, there are 4 or fewer bytes left.
	 * Each byte should turn into two bytes.  If a byte
	 * is the last one in the source line, then we
	 * put two copies of it into the dest line.  This avoids
	 * averaging with frame pixels.
	 */
    if (pS == endPtr)	/* A is the last byte */
    {
	dataOut = ULONG_ACCESS(&pS[0]) & 0xFF000000;
	ULONG_ACCESS(&pD[0]) = dataOut | (dataOut >> 8); /* A A 0 0 */
	return;
    }

    if ((pS + 1) == endPtr)	/* B is the last byte */
    {
	dataOut = ULONG_ACCESS(&pS[0]) & 0xFEFF0000;	/* A B 0 0 */
	dataOut += (ULONG_ACCESS(&pS[0]) & 0xFFFF0000) >> 8; /* A 2AB B 0 */
	dataOut |= (dataOut & 0xFF00) >> 8;		/* A 2AB B B */
	dataOut = (dataOut &  0xFE00FEFE) |
		  ((dataOut & 0x01FE0000) / 2); 	/* A AB B B */
	ULONG_ACCESS(&pD[0]) = dataOut;
    }
    else			/* B is not the last byte, so include C */
    {
	dataOut = ULONG_ACCESS(&pS[0]) & 0xFEFF0000;	/* A B 0 0 */
	dataOut += ULONG_ACCESS(&pS[0]) >> 8;		/* A 2AB B C */
	dataOut &= 0xFFFFFEFF;				/* take out B.0 */
	dataOut += (dataOut >> 8) & 0xFF;		/* A 2AB B 2BC */
	dataOut = (dataOut &  0xFE00FE00) |
		  ((dataOut & 0x01FE01FE) / 2); 	/* A AB B BC */
	ULONG_ACCESS(&pD[0]) = dataOut;
    }

    if ((pS + 2) == endPtr)	/* C is the last byte */
    {
	dataOut = (ULONG_ACCESS(&pS[0]) & 0x0000FF00) << 16; /* C 0 0 0 */
	ULONG_ACCESS(&pD[4]) = dataOut | (dataOut >> 8); /* C C 0 0 */
	return;
    }

	/* There can only be 4 source bytes.  If there were more than 4,
	 * we wouldn't be here.  Therefore, D must be the last byte.
	 */
    temp = ULONG_ACCESS(&pS[0]) & 0xFEFE;
    dataOut = temp << 16;				/* C D 0 0 */
    dataOut += temp << 8;				/* C 2CD D 0 */
    dataOut += temp & 0xFF;				/* C 2CD D D */
    dataOut = (dataOut &  0xFE00FEFE) |
	      ((dataOut & 0x01FE0000) / 2); 		/* C CD D D */
    ULONG_ACCESS(&pD[4]) = dataOut;
    return;

}


/* PRIVATE */
/**************************************************************************
 * ip_average8bppLines()
 *	Generates an interpolated dest line by averaging two existing
 *	lines of the right width.
 *	Used during during 1.5x and 2x resolution conversion.
 *
 *	This routine averages the two lines.  Since there are 8
 *	bits/pixel, we can afford to ignore the LS bit of each pixel.
 *	We just shift the two source words right, mask off the new MS bits
 *	of each byte, and add them together to produce the result.
 *
 *	Input:
 *		pTop:	pointer to top source line
 *		bpl:	bytes per line in all three lines	
 *		wD:	width of dest line in pixels
 *
 *	This routine writes into the right frame.
 **************************************************************************/
void  CDECL 
ip_average8bppLines (
    IPState    *pIPState,
    UInt8      *pTop,
    UInt32	bpl,
    Int32	wD)
{
    PROLOGUE;
    UInt32	dataOut, upperIn, lowerIn;
    UInt8      *endPtr;	/* used to find end of line */	

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* since we want to make sure that we get the last pixel and
	 * width may be odd, round up the width.  This will write into
	 * the right frame so it must be cleared later.
	 */
    endPtr = pTop + wD;

	/* go across the line in units of words.  This allows us to
	 * use the same code for both big and little endian 
	 * platforms.
	 */
    for ( ; (UInt32) endPtr > (UInt32) pTop; pTop += 4)
    {
	    /* Add the 3 LSB's in each pixel first */
	upperIn = (ULONG_ACCESS(pTop) / 2) & 0x7F7F7F7F;
	lowerIn = (ULONG_ACCESS(&pTop[2*bpl]) / 2) & 0x7F7F7F7F;

	    /* now that the MSB's are gone, we know there won't be
	     * any carries from pixel to pixel.  Add the pixels
	     * and divide by 2 to get the average.
	     */
	dataOut = (upperIn + lowerIn);

	ULONG_ACCESS(&pTop[bpl]) = dataOut;	/* write the word */
    }
}

/* This array is used by ip_rescon2To1Line4bpp and ip_rescon3To1Line4bpp. */
static UInt32 rightMask[8] =
    { 0xFFFFFFFF, 0xF0000000, 0xFF000000, 0xFFF00000,
      0xFFFF0000, 0xFFFFF000, 0xFFFFFF00, 0xFFFFFFF0 };

/* PRIVATE */
/*****************************************************************************
 * 
 * ip_rescon2To1Line4bpp
 *
 *	Resolution conversion - 2 to 1 in both axes, 4 bit/pixel source
 *	and destination.
 *
 * Arguments:
 *	pS	-	Pointer to pair of source lines
 *	wS	-	Width of a source line
 *	sBpl	-	Byte distance between source lines
 *	pD	-	Pointer to a destination line
 *	wD	-	Width of a destination line
 *  
 * Returns:
 *	nothing
 *
 *****************************************************************************/
void CDECL 
ip_rescon2To1Line4bpp (
    IPState    *pIPState,
    UInt8	*pS,
    UInt32	wS,
    Int32	sBpl,
    UInt8      *pD,
    UInt32	wD)
{
    PROLOGUE;
    UInt32	 sumH, sumL, input, outputWord;
    UInt8	*pEndSrcWord;
    UInt8	*pS2;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* keep address pointers to the two source lines of interest. */
    pS2 = (pS + sBpl);

	/* Point to the last source word to be read.  We stop after
	 * processing this source word with a partial or full result
	 * in outputWord.
	 */
    pEndSrcWord = pS + ((wS - 1) / 8) * 4;

	/* Keep the compiler happy.  It thinks that we might not go
	 * through the loop at all, so outputWord might not be
	 * initialized.  This can only happen if wS is a big negative
	 * number. If that happens, you get what you deserve.
	 */
    outputWord = 0;

	/* Process the data output-word-at-a-time, i.e. 8 output pixels, which
	 * requires 32 (16x2) input pixels.
	 */
    for (; pS <= pEndSrcWord; pD += 4, pS += 8, pS2 += 8)
    {
	    /* We're going to use a box filter for reduction, that is,
	     * we weight all four pixels going into the output pixels
	     * equally (1/4).  However, since we're using 4 bit
	     * pixels, we can't afford to toss any of the bits of
	     * the input pixels until after we've done all the adds.
	     * The loop below has two parts.  One computes the upper
	     * four nibbles of the output and the other computes the
	     * lower four nibbles.  These are basically the same
	     * except for how the nibbles are finally assembled.
	     * In each part, we keep two registers, sumH and sumL.
	     * sumH holds the upper bits of each output pixel, sumL
	     * holds the lower bits.  Initially, we put 2 bits of
	     * each pixel into each register.  Unfortunately, we
	     * need three registers.  The extra one is to hold
	     * the input.
	     */
	input = ULONG_ACCESS(&pS[0]);		/* get input word on top line */
	sumL = input & 0x33333333;
	sumH = (input >> 2) & 0x33333333;

	input = ULONG_ACCESS(&pS2[0]);		/* get word on bottom line */
	sumL += input & 0x33333333;
	sumH += (input >> 2) & 0x33333333;	/* add in pixels below */

	    /* Now we have the sum of each pixel and the one below it
	     * in sumH and sumL.  Add the pixels to the right to the
	     * sum pixels on the left.  This gives us the sum of all
	     * four pixels in each block.  We don't have to do any
	     * masking at this step because the largest sum in any
	     * pixel before this step is 6 (3 + 3), so the largest sum
	     * is 12, which still fits in 4 bits.
	     */
	sumH += (sumH << 4);
	sumL += (sumL << 4);

	    /* sumH now has interesting 4 bit pixels in bits 31:28,
	     * 23:20, 15:12, and 7:4.  sumL has interesting bits in the
	     * same places, but the binary point is in the middle of
	     * each range.  We add 2 to each sumL pixel for rounding,
	     * then shift the results right by 2 (to align them with
	     * sumH) and add them to sumH.  Note that we still don't
	     * need to do any masking before the add since the rounded
	     * values can`t add up to numbers bigger than 15.
	     */
	sumL += 0x20202020;
	sumL &= 0xC0C0C0C0;
	sumH += (sumL >> 2);

	    /* The four output pixels of interest are now complete in
	     * bits 31:28, 23:20, 15:12 and 7:4 of sumH.  First,
	     * Clear out the vacant pixels.
	     */
	sumH &= 0xF0F0F0F0;

	    /* We shift sumH left by 4 and OR it with itself.  This
	     * puts the useful pixels into two clumps of 8 bits each.
	     */
	sumH |= (sumH << 4);

	    /* Mask off the trash, leaving only useful bits */
	sumH &= 0xFF00FF00;

	    /* combine these into a single clump of 4 pixels */
	outputWord = sumH;
	outputWord = (outputWord | (sumH << 8)) & 0xFFFF0000;

	    /* Was that the last pair of source words on the line?
	     * if so, stop now.  We stop by popping right back up
	     * to the top of the loop, where the pointers are
	     * incremented off the end and we quit.  Weird.
	     * This is done so that all methods of exiting the
	     * loop are the same.  We want the meaning of pD
	     * to be consistent at the exit of the loop so we
	     * can reach back and rewrite the last word with
	     * a masked value.
	     */
	if (pS == pEndSrcWord)
	    continue;

	    /* repeat the process, but shift all the stuff to the
	     * right instead of to the left.  This allows us to
	     * compose a single output word easily.
	     */
	input = ULONG_ACCESS(&pS[4]);		/* get input word on top line */
	sumL = input & 0x33333333;
	sumH = (input >> 2) & 0x33333333;

	input = ULONG_ACCESS(&pS2[4]);		/* get word on bottom line */
	sumL += input & 0x33333333;
	sumH += (input >> 2) & 0x33333333;	/* add in pixels below */

	sumH += (sumH >> 4);			/* put sums in lower nibbles */
	sumL += (sumL >> 4);			/* of each byte */

	sumL += 0x02020202;			/* round the LS bits */
	sumL &= 0x0C0C0C0C;
	sumH += (sumL >> 2);			/* and compose final values */

	sumH &= 0x0F0F0F0F;			/* get rid of trash */
	sumH |= (sumH >> 4);			/* good bits in lower bytes
						 * of each halfword */
	sumH &= 0x00FF00FF;			/* get rid of trash */
	sumH |= (sumH >> 8);			/* good bits in bottom
						 * halfword. */
	    /* Write this word to the output.  Make sure the output
	     * is duplicated in outputWord so we can rewrite it
	     * if this is the last word in the row.
	     */
	outputWord |= (sumH & 0x0000FFFF);
	ULONG_ACCESS(pD) = outputWord;
    }

	/* Mask off the unused portion of the last word and write it again.
	 * I think this is faster than testing to see if we're composing
	 * the last word every time through the loop.  Remember than
	 * pD was incremented before we exited the loop, so we have
	 * to point back to it.
	 * Storing outputWord here is required if we broke out of the
	 * loop in the middle.
	 */
    ULONG_ACCESS(&pD[-4]) = outputWord & rightMask[wD & 7];
}


/* PRIVATE */
/*****************************************************************************
 * 
 * ip_rescon3To1Line4bpp
 *
 *	Resolution conversion - 3 to 1 in both axes, 4 bit/pixel source
 *	and destination.
 *
 * Arguments:
 *	pS	-	Pointer to pair of source lines
 *	wS	-	Width of a source line
 *	sBpl	-	Byte distance between source lines
 *	pD	-	Pointer to a destination line
 *	wD	-	Width of a destination line
 *  
 * Returns:
 *	nothing
 *
 *****************************************************************************/
void CDECL 
ip_rescon3To1Line4bpp (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	wS,
    Int32	sBpl,
    UInt8      *pD,
    UInt32	wD)
{
    PROLOGUE;
    UInt32	 sumH, sumL, input, outputWord;
    UInt8	*pEndSrcWord;
    UInt8	*pS2, *pS3;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* keep address pointers to the three source lines of interest. */
    pS2 = (pS + sBpl);
    pS3 = (pS2 + sBpl);

	/* Point to the last source word to be read.  We stop after
	 * processing this source word with a partial or full result
	 * in outputWord.
	 */
    pEndSrcWord = pS + ((wS - 1) / 8) * 4;

	/* Keep the compiler happy.  It thinks that we might not go
	 * through the loop at all, so outputWord might not be
	 * initialized.  This can only happen if wS is a big negative
	 * number. If that happens, you get what you deserve.
	 */
    outputWord = 0;

	/* Process the data output-word-at-a-time, i.e. 8 output pixels, which
	 * requires 72 (24*3) input pixels.
	 */
    for (; pS <= pEndSrcWord; pD += 4, pS += 12, pS2 += 12, pS3 += 12)
    {
	    /* We use the following filter to weight the output bits:
	     *	1/16   1/8   1/16
	     *  1/8    1/4   1/8
	     *  1/16   1/8   1/16
	     * To do this, we first combine words vertically, then
	     * combine them horizontally.  Note that we can combine
	     * the vertical words by weighting the top and bottom words
	     * by 1/4 and the middle word by 1/2; then combine the left
	     * and right columns by weighting them by 1/4 and the
	     * middle column by 1/2.  When multiplied out, this gives
	     * the weights specified above.
	     * How do we do this with 4-bit pixels?  As in the 2:1
	     * reduction case, we divide each word into two registers,
	     * sumH and sumL.  sumH will holds the upper 2 bits of each
	     * pixel and sumL will hold the lower two bits.  Now, we
	     * need to figure out how and when to recombine the upper
	     * and lower bits of each pixel.  In fact, can we guarantee
	     * that there's no overflow out of the 4-bit space allocated
	     * for each original 2-bit half?
	     * When doing the vertical adds, we align the LS bit of the
	     * pixels from the top and bottom rows into the LS bit of
	     * each 4-bit position.  We align the LS bit of the pixels
	     * from the middle rows into the next bit above that.  Assuming
	     * the original pixels has a value of 15 (1111 in binary),
	     * the largest sum in each 4-bit position will be
	     * 0011 (top) + 0110 (middle) + (0011) (bottom) = 1100.
	     * This is true for both sumH and sumL.  Hmmm, we now have
	     * 8 interesting bits for each 4 bit pixel and we haven't
	     * even started the process of combining columns.  What
	     * do we do with all the bits?
	     * My solution is to simply recombine the upper and lower
	     * halves at this point.  We add 0010 to the lower half pixels
	     * for rounding, then add the upper two bits of the lower pixels
	     * to the lower 2 bits of the upper pixels.  After appropriate
	     * masking, we have 4-bit approximations to the correct answers.
	     * I think we can get away with this because the 4-bit
	     * approximations still cover the complete range of 4-bit numbers
	     * so we don't notice the truncation (?).
	     * Now, we combine columns by just doing shifts, masks and adds.
	     * We shift right by 5 and left by 3 so the components from
	     * the sides have 1/2 the weight of the column in the middle
	     * this produces a new number with an extra bit of significance
	     * above the original center pixel's value.  Where does that go?
	     * To answer, we have to look at the position of the center
	     * pixels in each word.
	     * We process the lines in units of three words.  If no shifting
	     * is done, the pixels are arranged as shown below.  The
	     * pixels are labelled l for left, C for center and r for
	     * right:
	     *	lCrlCrlC rlCrlCrl CrlCrlCr
	     * Another way to look at this is to give the center positions
	     * names, such as:
	     *  .a..b..c ..d..e.. f..g..h.
	     * Note there are three center pixels in the words on the 
	     * ends and two in the word in the middle.  Also note that
	     * the word in the middle has a right pixel to contribute to
	     * the word on its left and a left pixel to contribute to the
	     * word on its right.
	     * I thought of lots of variations on processing order.  One
	     * idea was to form the final values for all the center pixels,
	     * leaving them in their native positions, combine all three
	     * result words (turning the 8 desired pixels "abcdefgh" into
	     * "fadgbehc"), then unscramble the pixels.  This is cute, but
	     * has a couple of drawbacks.  The biggest is the tables or
	     * amount of shifting required to do the unscrambling.  In the
	     * end, I decided to do the straightforward thing of
	     * consolidating the result pixels in each word as they are
	     * derived.  This has the extra advantage of allowing us
	     * to stop as soon as the last source word has been processed.
	     * Ok, there's one little addition to this plan.  When forming
	     * the pixels in the first word (abc), the rightmost pixel
	     * is not formed completely.  We leave out the contribution
	     * from the pixel on its right, which is contained in the
	     * second word.  If the first word in the triplet was also
	     * the last word in the source line, that means the rightmost
	     * pixel has a contribution of 0, or black.  So far as I know,
	     * that's ok, but if someone wants to fix it sometime, we
	     * could make it white later.
	     * When the second word is formed, we complete pixel c and
	     * insert pixels d and e into the accumulated output word.
	     * We also put the contribution from the pixel to the left of
	     * pixel f into the bottom nibble of the output word.
	     * This is erased if the second column of the triplet is also
	     * the last source column.
	     * As part of forming the third word of the triplet, we add the
	     * orphaned contribution to pixel f, which was kept on the last
	     * nibble of the output word, to pixel f.
	     */


	    /* ******************** Column 1 *************************/

	    /* Compute the vertical sum of the words on the left end
	     * of the triplet.
	     */
	input = ULONG_ACCESS(&pS[0]);		/* get input word on top line */
	sumL = input & 0x33333333;
	sumH = (input >> 2) & 0x33333333;

	    /* Weight the pixels in the middle by twice as much as
	     * the pixels on the top and bottom.
	     */
	input = ULONG_ACCESS(&pS2[0]);
	sumL += (input << 1) & 0x66666666;
	sumH += (input >> 1) & 0x66666666;

	input = ULONG_ACCESS(&pS3[0]);		/* get word on bottom line */
	sumL += input & 0x33333333;
	sumH += (input >> 2) & 0x33333333;	/* add in pixels below */

	    /* round the lower pixels */
	sumL += 0x22222222;

	    /* form the complete vertical pixel sums */
	sumH += (sumL >> 2) & 0x33333333;

	    /* This is the word with output pixel bits in positions
	     * 0x0F00F00F, so we need to shift the inputs
	     * left before doing the next set of adds to make
	     * room for the to-be-discarded LS bit below the
	     * LS pixel.
	     * Add the pixels on the left to the center pixels
	     * first to make room for the new sum bits above
	     * the center pixels.  First, save the left and
	     * right pixels in sumL so they don't get in the way.
	     */
	sumL = sumH;
	sumH &= 0x0F00F00F;			/* select the center pixels */
	sumH <<= 1;				/* leave room for new LS bit */
	sumH += (sumL >> (5-1)) & 0x0F00F00F;	/* pixels on the left */
	sumH += (sumL << (3+1)) & 0x0F00F00F;	/* pixels on the right */

	    /* At this point, sumH has 3 6-bit pixels with the MS
	     * bits two bits to the left from the natural nibble boundaries.
	     * Strip off the obsolete LS bits.
	     * NOTE, the LS bits below the rightmost pixel is not yet
	     * obsolete because we haven't added in the contribution from
	     * the pixel to its right yet.  Keep those bits around.
	     * Collect the three pixels at the MS end of the word.  We
	     * do one shift to put the pixels into the position
	     * 0xF00F00FC, then two more shifts and ORs to put the pixels
	     * we want at the top of the word.  Note that we only need
	     * one mask step at the end.  That's because shifting the
	     * original word (0xF00F00FC) left by 8 produces 0x0F00FC00,
	     * which sets only the second pixel, and shifting it by 16
	     * produces 0x00FC0000, which sets only the third pixel plus
	     * one bit.
	     */
	sumH <<= 2;				/* produce 0xF00F00F0 */
	sumH &= 0xF00F00FC;			/* delete obsolete bits */
	sumL = sumH;
	sumH |= sumL << 8;
	sumH &= 0xFF000000;			/* clear bit 19, which would
						 * otherwise conflict with
						 * the LS bit below pixel c.
						 */
	outputWord = sumH | (sumL << 16);

	    /* Was that the last pair of source words on the line?
	     * if so, stop now.  We stop by popping right back up
	     * to the top of the loop, where the pointers are
	     * incremented off the end and we quit.  Weird.
	     * This is done so that all methods of exiting the
	     * loop are the same.  We want the meaning of pD
	     * to be consistent at the exit of the loop so we
	     * can reach back and rewrite the last word with
	     * a masked value.
	     */
	if (pS == pEndSrcWord)
	    continue;


	    /* ******************** Column 2 *************************/

	    /* So go on to do the second word.  It's pretty much the same. */

	input = ULONG_ACCESS(&pS[4]);		/* get input word on top line */
	sumL = input & 0x33333333;
	sumH = (input >> 2) & 0x33333333;

	input = ULONG_ACCESS(&pS2[4]);
	sumL += (input << 1) & 0x66666666;
	sumH += (input >> 1) & 0x66666666;

	input = ULONG_ACCESS(&pS3[4]);		/* get word on bottom line */
	sumL += input & 0x33333333;
	sumH += (input >> 2) & 0x33333333;	/* add in pixels below */

	    /* round the lower pixels */
	sumL += 0x22222222;

	    /* form the complete vertical pixel sums */
	sumH += (sumL >> 2) & 0x33333333;

	    /* The rightmost pixel in the output word (in bits 23:20)
	     * needs a contribution from the leftmost pixel in this word.
	     * Apply it now.  We want the MS bit of the term which is in
	     * bit 31 of sumH to appear in bit 21 of the output word.
	     * That's 2 positions down from the MS bit of the result
	     * pixel in bits 23:20.  Why two bits?  Well, the MS bit
	     * of the result is one bit above the MS bit of the center
	     * pixel, which is one bit above the MS bit of the right
	     * pixel.  1+1=2.
	     */
	outputWord += (sumH >> (31-21));
	outputWord &= 0xFFF00000;

	    /* Now copy the left and right pixels into sumL and the
	     * center pixels into sumH as above.  Add the left and
	     * right contributions.
	     */
	sumL = sumH;
	sumH &= 0x00F00F00;			/* select the center pixels */
	sumH += (sumL >> 5) & 0x00780780;	/* pixels on the left */
	sumH += (sumL << 3) & 0x00780780;	/* pixels on the right */

	    /* At this point, sumH has 2 5-bit pixels with the MS
	     * bits one bit to the left of the natural nibble
	     * positions.  Strip off the obsolete LS bits.
	     */
	sumH &= 0x01E01E00;

	    /* Collect the new center pixels into outputWord */
	outputWord |= ((sumH >> 5) | (sumH << 3)) & 0x000FF000;
	    /* insert the remaining rightmost pixel from sumL into
	     * output word so we can apply it to the MS pixel of
	     * the third word in the triplet.
	     */
	outputWord |= sumL & 0xF;

	    /* If that was the last source word, do the continue trick now. */
	if (pS+4 == pEndSrcWord)
	    continue;


	    /* ******************** Column 3 *************************/

	    /* So do the last word as before. */

	input = ULONG_ACCESS(&pS[8]);		/* get input word on top line */
	sumL = input & 0x33333333;
	sumH = (input >> 2) & 0x33333333;

	input = ULONG_ACCESS(&pS2[8]);
	sumL += (input << 1) & 0x66666666;
	sumH += (input >> 1) & 0x66666666;

	input = ULONG_ACCESS(&pS3[8]);		/* get word on bottom line */
	sumL += input & 0x33333333;
	sumH += (input >> 2) & 0x33333333;	/* add in pixels below */

	    /* round the lower pixels */
	sumL += 0x22222222;

	    /* form the complete vertical pixel sums */
	sumH += (sumL >> 2) & 0x33333333;

	    /* This is the word with center pixels at 0xF00F00F0.
	     * Before we can add the remaining pixel hiding in the LS
	     * nibble of output word to the MS pixel of this word,
	     * we need to shift it right to make room for the carry.
	     * Remember this extra shift when shifting the left and
	     * right pixels too.  We shift right by 3 so we can add
	     * in the contribution from pixels to the right without
	     * having to shift those too.
	     * The contribution from outputWord is shifted left by
	     * (31-3) <distance from MS bit of correction value to MS bit of
	     * word> - 3 <MS pixel in word has been shifted right by 3>
	     * - 1 <MS bit of correction value should be 1 bit to the
	     * right of MS bit of original pixel>
	     */
	sumL = sumH;
	sumH = (sumH & 0xF00F00F0) >> 3;
	sumH += (outputWord & 0xF) << (31-3-3-1); /* add contribution from pixel
						 * to the left of MS pixel */
	sumH += (sumL >> (4+1+3)) & 0x0F00F00F;	/* add pixels from above */
	sumH += (sumL & 0x0F00F00F);		/* add pixels from below */

	    /* The final pixels are now in 0x3C03C03C, i.e. the MS bits
	     * are 2 bits to the left of normal nibble positions.
	     * We consolidate these at the bottom of the word, then
	     * insert them into outputWord.
	     */
	sumH >>= 2;
	sumH &= 0x0F00F00F;			/* select the correct bits */
	sumH |= (sumH >> 8) | (sumH >> 16);
	sumH &= 0x00000FFF;			/* those are the only ones we
						 * want. */
	outputWord &= 0xFFFFF000;		/* clear out space for them */
	outputWord |= sumH;

	    /* write this output word.  Keep the word in outputWord
	     * so we can rewrite the last word in the line after it
	     * has been masked.
	     */
	ULONG_ACCESS(pD) = outputWord;
    }

	/* Mask off the unused portion of the last word and write it again.
	 * I think this is faster than testing to see if we're composing
	 * the last word every time through the loop.  Remember than
	 * pD was incremented before we exited the loop, so we have
	 * to point back to it.
	 * Storing outputWord here is required if we broke out of the
	 * loop in the middle.
	 */
    ULONG_ACCESS(&pD[-4]) = outputWord & rightMask[wD & 7];
}


/* PRIVATE */
/*****************************************************************************
 * 
 * ip_rescon2To1Line8bpp
 *
 *	Resolution conversion - 2 to 1 in both axes, 8 bit/pixel source
 *	and destination.
 *
 *	We'll attempt to do four output pixels at once.
 *
 * Arguments:
 *	pS	-	Pointer to pair of source lines
 *	wS	-	Width of a source line
 *	sBpl	-	Byte distance between source lines
 *	pD	-	Pointer to a destination line
 *	wD	-	Width of a destination line
 *  
 * Returns:
 *	nothing
 *
 *****************************************************************************/

void CDECL 
ip_rescon2To1Line8bpp (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	wS,
    Int32	sBpl,
    UInt8      *pD,
    UInt32	wD)
{
    PROLOGUE;
    UInt32	*pSLong, *pSLong2,
		 outputPixels, cSum1, cSum2, outputWords;
    UInt32	 ii;
    UInt8	*pS2;
    UInt32	 i;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    pS2 = pS + sBpl;
    wS = wS;

    outputWords = wD & (~3);
    pSLong = (UInt32 *) pS;
    pSLong2 = (UInt32 *) (pS + sBpl);

    /*
     * Process the data output-word-at-a-time, i.e. 4 output pixels, which
     * is 16 (8x2) input pixels.
     */
    for (i = 0; i < outputWords; i += 4)
    {
	/* 
	 * get two input longs for each line, each of which contains 4x1 pix.
	 * Mask/shift such that we have
	 * 0xxxxxxx0xxxxxxx0xxxxxxx0xxxxxxx
	 * and when we add each pair together, the resulting longword is the
	 * sum of the 4 columns in the original words, divided by 2.
	 */

	cSum1 = ((*pSLong++ & 0xfefefefe) >> 1) + 
	        ((*pSLong2++ & 0xfefefefe) >> 1);
	cSum2 = ((*pSLong++ & 0xfefefefe) >> 1) + 
	        ((*pSLong2++ & 0xfefefefe) >> 1);


 	/* 
	 * Produce the output pixels. This involves adding up the four
	 * pairs of column sums, and dividing by two, and shifting into th
	 * right place. Rather than divide by two, we shift into the right
	 * place and mask out the extra bit
	 */
	outputPixels  =  (((cSum1 >> 24) + 
			  ((cSum1 >> 16) & 0xff)) << 23) & 0xff000000;
	outputPixels |= ((((cSum1 >> 8) & 0xff) + 
			   (cSum1 & 0xff)) << 15) & 0x00ff0000;
	outputPixels |=  (((cSum2 >> 24) + 
			  ((cSum2 >> 16) & 0xff)) << 7)  & 0x0000ff00;
	outputPixels |=  (((cSum2 >> 8) & 0xff) + 
			   (cSum2 & 0xff)) >> 1;

	/* Output the pixels */
	*((UInt32 *) (pD + i)) = outputPixels;
    }

    /* Now do the remaining pixels */
    ii = outputWords << 1;
    for (; i < wD; i++)
    {
	cSum1 = UCHAR_ACCESS (pS + ii) + UCHAR_ACCESS (pS2 + ii++);
	cSum2 = UCHAR_ACCESS (pS + ii) + UCHAR_ACCESS (pS2 + ii++);
	UCHAR_ACCESS (pD + i) =  (UInt8) ((cSum1 + cSum2) >> 2UL);
    }
}



/* PRIVATE */
/*****************************************************************************
 * 
 * ip_rescon3To1Line8bpp
 *
 *	Resolution conversion - 3 to 1 in both axes, 8 bit/pixel source
 *	and destination. It does a center-weighted downaverage. There
 *	are two reasons for doing this: one, improving contrast a bit
 *	(the box filter blurs things a bit too much) and two, it uses
 *	shifts rather than a look-up table, which runs a slight bit faster
 *	on the SPARC, and seems around the same speed on the PC.
 *
 * Arguments:
 *	pS	-	Pointer to pair of source lines
 *	wS	-	Width of a source line
 *	sBpl	-	Byte distance between source lines
 *	pD	-	Pointer to a destination line
 *	wD	-	Width of a destination line
 *  
 * Returns:
 *	nothing
 *
 *****************************************************************************/
void CDECL 
ip_rescon3To1Line8bpp (
    IPState    *pIPState,
    UInt8      *pS,
    UInt32	wS,
    Int32	sBpl,
    UInt8      *pD,
    UInt32	wD)
{
    PROLOGUE;
    UInt32	*pSLong, *pSLong2, *pSLong3;
    UInt32       outputPixels, cSum1, cSum2, cSum3, outputWords;
    UInt32	 ii;
    UInt32	 i;
    UInt8	*pS2, *pS3;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    wS = wS;
    pS2 = pS + sBpl;
    pS3 = pS2 + sBpl;

    /* In this case, we output a longword (4 pix) every 12x3 input bytes. */
    outputWords = wD - (wD % 12);
    pSLong = (UInt32 *) pS;
    pSLong2 = (UInt32 *) (pS + sBpl);
    pSLong3 = (UInt32 *) (pS + 2*sBpl);

    for (i = 0; i < outputWords; i += 4)
    {
	/* 
	 * get 3 longs for each line, each of which contains 4x1 pix.
	 * Mask/shift such that we get
	 *
	 * 00xxxxxx|00xxxxxx|00xxxxxx|00xxxxxx
	 * 0xxxxxxx|0xxxxxxx|0xxxxxxx|0xxxxxxx
	 * 00xxxxxx|00xxxxxx|00xxxxxx|00xxxxxx
	 *
	 * and when we add the three together, the resulting longword is the
	 * sum of the columns in the original words, with the center
	 * row divided by 2, and the outer rows divided by 4, resulting
	 * in a sum still normalized 0-255.
	 *
	 */

	cSum1 = ((*pSLong++ & 0xfcfcfcfc) >> 2) + 
 	        ((*pSLong2++ & 0xfefefefe) >> 1) +
	        ((*pSLong3++ & 0xfcfcfcfc) >> 2);

	cSum2 = ((*pSLong++ & 0xfcfcfcfc) >> 2) + 
	        ((*pSLong2++ & 0xfefefefe) >> 1) +
	        ((*pSLong3++ & 0xfcfcfcfc) >> 2);

	cSum3 = ((*pSLong++ & 0xfcfcfcfc) >> 2) + 
	        ((*pSLong2++ & 0xfefefefe) >> 1) +
	        ((*pSLong3++ & 0xfcfcfcfc) >> 2);


	/* 
	 * The columns are summed in csum[1-3], so sum each
	 * of 4 triples, again weighted, this time so that
	 * the outer two values are weighted by 1/4, and the center
	 * value is weighted by 1/2. The result is that we weight
	 * each resulting output pixel as follows:
	 *
	 * 1/16  1/8  1/16
	 * 1/8   1/4  1/8
	 * 1/16  1/8  1/16
	 *
	 * which sums to one, as one would like.
	 */

	outputPixels =   ((cSum1 & 0xfc000000) >> 2) +
		         ((cSum1 & 0x00fe0000) << 7) +
		         ((cSum1 & 0x0000fc00) << 14);
	outputPixels |=  ((cSum1 & 0x000000fc) << 14) +
			 ((cSum2 & 0xfe000000) >> 9) + 
			 ((cSum2 & 0x00fc0000) >> 2);
	outputPixels |=  ((cSum2 & 0x0000fc00) >> 2) +
			 ((cSum2 & 0x000000fe) << 7) +
			 ((cSum3 & 0xfc000000) >> 18);
	outputPixels |=  ((cSum3 & 0x00fc0000) >> 18) +
			 ((cSum3 & 0x0000fe00) >> 9) +
			 ((cSum3 & 0x000000fc) >> 2);
	/* Output the four pixels */
	*((UInt32 *) (pD + i)) = outputPixels;
    }
    
    /* Now do the remaining pixels which weren't a multiple of 12 */
    ii = outputWords * 3;
    for (; i < wD; i++)
    {
	/* Slow but safe. Just do the convolution sum listed above 
	   directly. We could probably be more efficient here,
	   but it's "only the last few" (up to 11) pixels, so it'll
	   only matter on really narrow images. */
	UCHAR_ACCESS (pD + i) = (UCHAR_ACCESS(pS  + ii) >> 4) +
				(UCHAR_ACCESS(pS2 + ii) >> 3) +
				(UCHAR_ACCESS(pS3 + ii) >> 4) +
				(UCHAR_ACCESS(pS  + ii + 1) >> 3) +
				(UCHAR_ACCESS(pS2 + ii + 1) >> 2) + 
				(UCHAR_ACCESS(pS3 + ii + 1) >> 3) +
				(UCHAR_ACCESS(pS  + ii + 2) >> 4) +
				(UCHAR_ACCESS(pS2 + ii + 2) >> 3) + 
				(UCHAR_ACCESS(pS3 + ii + 2) >> 4);
    }
}


/*
 *	 
$Log: mirescon.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.12  1998/01/13 23:53:13  ddavies
mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.11  1995/12/13 02:51:26  ejaquith
mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.10  1995/10/18  04:04:01  ddavies
 * mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.9  1995/03/08  04:28:16  ddavies
 * mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v
 *
 * Extensive changes in the way the 2x and 1.5x expansion code handles the
 * right sides of images.  Formerly, it just averaged the pixels in the frame
 * into the rightmost output pixel.  Any extra pixels are now duplicates of
 * the rightmost source pixels.  This fixes Macduff Hughes' latest bug in which
 * an all white image is resolution converted and produces a black stripe down
 * the right edge.
 *
 * Revision 1.8  1995/01/12  00:08:28  ddavies
 * ddavies on Wed Jan 11 16:06:09 PST 1995
 *
 * mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v
 *
 * Convert ip_rescon2To1Line4bpp and ip_rescon3To14bpp from dummies to real
 * routines.  These routines produce one output line of 4 bit/pixel images.
 * The first does 2:1 reduction, weighting the four input pixels equally.
 * The second does 3:1 reduction, weighting the middle pixel by 1/4, the
 * four horizontal and vertical neighbors by 1/8 and the four diagonals by
 * 1/15.
 * Ooops, I meant 1/16.
 *
 * Revision 1.7  1994/07/14  20:58:38  macleod
 * macleod on Thu Jul 14 13:58:12 PDT 1994
 *
 * mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v
 *
 * Fixed something which gave a warning in ACC but not GCC.
 *
 * Revision 1.6  1994/05/27  00:43:20  macleod
 * macleod on Thu May 26 17:42:41 PDT 1994
 *
 * mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v
 *
 * Fixed a bug which caused the non-debug Sparc version to mess up
 * the last few pixels on the right edge.
 *
 * Revision 1.5  1994/05/26  00:22:40  macleod
 * mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v
 *
 * Added the inner loops for 2:1 and 3:1 downconversion,
 * ip_rescon2To1Line4bpp, ip_rescon2To1Line8bpp,
 * ip_rescon3To1Line4bpp, ip_rescon3To1Line8bpp
 * *NOTE* that the 4-bit cases don't do anything (literally, except
 * refer to their arguments to shut up the compiler) yet.
 *
 * Revision 1.4  1993/12/02  19:25:10  ddavies
 * mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:50:16  ddavies
 * mirescon.c -> /product/ipcore/ipshared/src/RCS/mirescon.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


