/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  mirnhst1.c: machine independent routines for computing histograms
 *		of horizontal run lengths.
 *
 *	Int32	ip_horizRunOf0sHistoLine()	looks at runs of 0's
 *	Int32	ip_horizRunOf1sHistoLine()	looks at runs of 1's
 *	Int32	ip_horizRunOfCyclesHistoLine()	looks at cycles (adjacent
 *						runs of 0's and 1's)
 *
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "runhist1.prv"
#include "runhist2.prv"

/* prototypes */
#include "shrcc.pub"
#include "shrcc.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: mirnhst1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");


/* It is useful for segmentation to be able to figure out how big things
 * are.  A cheap way to do this is to look at histograms of run lengths.
 * Histograms of lengths of white space can be used to figure out how
 * far apart things are.  This file contains routines that compute histograms
 * of horizontal runs in a single scan line.
 *
 * These routines are called from i_rnhst1.c.  That file has many useful
 * comments at the top.  Since they describe the upper and lower level
 * routines, we'll include them here too.
 *
 *****************************************************************************
 *
 * The code that finds blocks of glyphs on a page uses histograms of run
 * lengths.  XIS uses them extensively for segmentation.  In one sense,
 * they are a cheap alternative to computing fourier transforms.
 *
 * The module i_runhst.c also contains code to compute histograms of
 * horizontal runs.  If this code works better, the code in i_runhst.c
 * will become obsolete.
 *
 * XIS asked for a routine that has a bunch of arguments.  We've added
 * a couple more.  The arguments are:
 *
 *	pointers to image and rectangle within the image;
 *	pHisto: pointer to histogram to be returned;
 *	numHistoBins: equals the number of bins in the histogram
 *	direction: samples are to be taken horizontally or vertically;
 *	sampleFreq: skip sampleFreq-1 lines between each sampled row
 *		or column;
 *	runType: whether to count black runs, white runs or cycles (lengths
 *		of black/white pairs);
 *	useEndRuns: all runs are terminated at the boundary of the
 *		sampling rectangle.  If useEndRuns is true, we should
 *		count the runs that touch the boundary.  If it is
 *		false, we should ignore these runs.  This is a pain,
 *		but is useful.
 *		Well, in fact, NOT ignoring cycles that abut the
 *		borders is not a well defined operation.  Besides,
 *		it's hard to implement.  So we'll say that if we are
 *		measuring cycles, we always ignore cycles that abut
 *		the edges of the rectangle.
 *	countLongRuns:	if false, we ignore runs that have no bins in
 *		the histogram.  If true, all the counts for the large
 *		runs are put into to the last histogram bin.
 *	weightByLength: if false, finding a run of length x causes us to
 *		add 1 to bin x.  If true, finding a run of length x causes
 *		us to add x to bin x.  What is this for anyway?
 *	
 * This is quite a monster - the swiss army knife approach to writing code.
 * Unfortunately, I think we actually use this stuff.
 *
 * We'd really like to avoid making a zillion decisions for every run found.
 * One way to avoid some decisions is to create a temporary histogram,
 * then process it to produce the histogram we really want.  We will create a
 * temporary histogram containing a bin for every possible run length.
 * If we're looking for horizontal runs, the temporary histogram will have
 * bins for 1 through rectangleWidth.  If we're tracing vertical runs,
 * we need bins for 1 through rectangleHeight.  When we've filled the
 * temporary histogram with run info from the image, we start post processing.
 *
 * If weightByLength is true, we multiply the number in each bin by the
 * bin number.
 *
 * We then add the contents of the first numHistoBins bins from the
 * temporary histogram to the contents of the output histogram.
 *
 * If countLargeRuns is true, we add the contents of the remaining bins
 * in the temporary histogram to the pHisto[numHistoBins-1].
 *
 * That takes care of several of the parameters.  There are several
 * significant problems remaining.  The first one is the horizontal/vertical
 * direction.  Calculating histograms of horizontal runs is certainly
 * easier since you get to see several bits in a run in each sample.
 * Most of the overhead is encountered in detecting that a run has ended
 * and taking appropriate action.  In the vertical direction, you have
 * a choice of maintaining several runs concurrently or processing each
 * pixel in the image individually - yuck.  Let's describe the horizontal
 * run code first, getting back to the vertical code later.
 *
 * Even with just horizontal histograms, the choice of counting runs of
 * 1's, runs of 0's or cycles is not one we'd like to make on every
 * image sample.  We'll have different code for each of these cases.
 * What are cycles?  When finding areas containing periodic data, like
 * glyphs or halftones, it's best to count not just the width of each
 * mark or the width of the space between them, but the period of
 * white/black pairs.  This is the thing that's most likely to be constant.
 * Another way to think of this is that we count the distance between
 * 0-to-1 transitions.
 *
 * The really messy remaining parameter is useEndRuns.  It is much more
 * convenient to count all the runs.  When useEndRuns is false, we must
 * efficiently trace the leftmost and rightmost runs on each scan line,
 * then ignore them.  If we were tracing each run serially, this would
 * be trivial.  Since we're trying to process as much data as we can
 * on each sample, this will be messy.  We will have separate verions of
 * run counting code for these too.
 *
 * The routine at this level, i_runHistogram, will call lower routines
 * to process each line.  These routines will have the real inner loops.
 * What do the inner loops of these routines look like?  The inner loops
 * are similar for all the routines.  The cycle routine looks for somewhat
 * different things than the routines that look for 1's or 0's, but
 * the principle is the same.  The SkipEnds and UseEnds routines differ
 * only in how they start and stop.
 *
 * We clearly want to look at more than one pixel at a time.  The inner
 * loops should be able to skip over runs of 0's and runs of 1's in large
 * jumps.  This is especially true since these are fairly common.  On the
 * other hand, a large chunk of bits might contain many internal runs,
 * so we can't just deal with large chunks all the time.
 *
 * One cute observation is that the inner loops have two basic jobs.  One
 * is to track runs that bridge chunks.  The other is to track runs that
 * start and stop inside individual chunks.  Tracking runs that start and
 * stop purely within chunks need not be done as each chunk is encountered.
 * We can create yet another histogram, this one of chunk values seen.
 * When we get done with the image, we calculate the number of runs that
 * started and stopped inside each chunk and add that into the temporary
 * histogram.  For example, say chunks are bytes and we find an 0xAA.  We
 * increment the 0xAA bin.  At the end of the image, say we've found
 * 70,000 0xAA's.  Say we're tracking runs of 1's.  This byte (1010 1010)
 * has three internal runs (plus one on the left that bridges to the next
 * byte, each of length 1.  We would add 3 * 70,000 to the
 * temporary histogram bin for runs of length 1.
 *
 * Using this trick, we only have to track runs between chunks when
 * we're actually going through the image - the runs inside chunks are
 * done later.  This greatly simplifies the code done for each byte.
 *
 * We use tables to compute things about chunks.  To keep table sizes
 * small, we'd like the chunk size to be fairly small.
 * To be able to skip over runs of black or white efficiently, we'd like
 * the chunk size to be large.  We'll use a combination.  We process the
 * image in units of 32-bit words.  If a word is entirely black or entirely
 * white, we skip over it easily.  If it contains both black and white,
 * we process its four bytes individually.  At the end of the four bytes,
 * we look at the next word.
 *
 * There are two ways we can do the calculations on each byte: we can use
 * a state machine that uses each byte as input or we can derive both
 * the state and output from the region around each byte.  Huh?  The
 * state machine approach is straightforward.  Say we're looking for runs
 * of 1's.  The current state is set by the previous byte - we're either
 * in a run of 0's or a run of 1's:
 *	State	byte starts with	Action
 *	0's	0			keep looking for 1's
 *	0's	1			start new run
 *	1's	0			end current run, start looking for
 *					another
 *	1's	1			continue current run.
 *
 * This is lovely, but state machines are sort of complicated what with
 * all the testing a branching that goes on.
 *
 * Note that the state from the previous byte is equal to the last bit of
 * that byte together with the length of the run we're now in (if any).
 * We could eliminate the state machine stuff by just concatenating that
 * bit with the others in the byte when doing our table lookups.
 *
 * What table lookups?  We use three tables for each byte:
 *   1.	Incremental run length
 *	This table performs two functions.  It indicates when a run from
 *	the previous byte ends and gives the distance by which that run
 *	extends into this byte.  The run ends when the value in this
 *	table is negative.  The trick is that we need two values of 0.
 *	When a run ends at the beginning of this byte, we should store
 *	negative 0.  When there is no run coming into this byte, we store
 *	positive 0.  Since we need both positive and negative values for
 *	0, we use one's complement to store negative values.
 *   2.	Byte Value
 *	This is a histogram of byte values.  When we come to the end of
 *	a run, we increment the bin corresponding to this byte.  We use
 *	these counts later to add in the runs that were purely internal
 *	to the bytes.
 *   3.	New Run Length
 *	This is the number of bits in the run that includes the last
 *	bit of the byte.  We use it to set the run Length counter before
 *	encountering the next byte.
 * 
 * One nice feature of organizing things this way is that the inner loops
 * of the code that finds cycles and the code that finds runs of 1's (or 0's)
 * is all the same.  The only difference is the contents of the tables.
 * But, we use different routines anyway.  If we use different routines, we
 * can address the tables directly in each routine.  If we used one
 * routine, we'd have to pass a pointer to the table, which would use
 * a register.  There aren't a lot of registers in a x86 processor, so
 * we'll use different routines.
 *
 * Do we scan left to right or right to left within a line?  When we
 * encounter a new word, we would like to perform a quick check to see
 * if its the entension to an existing run or contains the end of that
 * run.  We can do this by looking at the last bit of the previous word
 * and the value of the new word:
 *	last bit of prev word	new word	Action
 *		0		00...00		Add 32 to runLength unless
 *						we're looking for runs of 1's
 *		1		11...11		Add 32 to runLength unless
 *						we're looking for runs of 0's
 *		all other combinations		Find end of run in word
 * It's easy to branch on the last bit of the previous word if that bit
 * is in the sign bit position (MSBit).  Therefore, we will go from right
 * to left within a scan line.
 *
 * Do we read each byte individually or do we read entire words into a
 * register and do shifting and masking to form the table lookup values?
 * Reading bytes individually is generally a good idea, but maybe not in
 * this case.  First, there is the annoyance of dealing with different
 * byte orders.  If we read by words, this goes away.  That's not a
 * big deal.  Second, at least the byte value histogram entries are not
 * bytes.  Therefore, the machine will do some shifting anyway.  The
 * current run length will not be a byte, so the value read from the
 * Incremental run length table will either be expanded into a word
 * (an operation that our compilers do very stupidly on PCs) or the
 * incremental run length table entries will be words already.  Ditto for
 * the new run length table.  We can save some shifting of address
 * offsets by making all the tables have the same entry size, so pick
 * 32-bit words.  This also eliminates the operations used to expand
 * smaller entries to words when they're read.
 *
 * Next, we're not addressing at least the incremental run length table
 * with bytes.  We use 9 bits - 8 from the current byte and 1 from the
 * previous byte.  We'd have to compose these entries anyway, so we might
 * as well do this from a register that hold all the bytes at once.  So
 * long as one of the tables is addressed using 9 bits, we might as well
 * address all of them that way - again to avoid extra address computation.
 * Reading the image in units of words and processing it in units of runs
 * of bits from these words has another advantage when processing the
 * end runs as seen below.
 *
 * What do we do with the end runs anyway?  When we are keeping the end
 * runs and looking for runs of 1's (or 0's), we just smash the portions of the
 * works on the far left and right that are outside our rectangle with
 * 0's (or 1's).  When looking for cycles, we smash the portion of the
 * word on the right with 0's (since that cannot start a new run) and
 * the portion of the leftmost word outside the rectangle with 1's
 * (since that cannot end a run).  When ignoring the end runs, we'll
 * duplicate the inner loop.  The first copy is executed only for the
 * very first, rightmost, run.  It does the normal things, except that
 * it doesn't update the temporary histogram for the first run length.
 * It then falls through to the ordinary loop.
 *
 * To ignore the leftmost run, we simply make the inner loop stop before
 * processing the word containing the left end of the rectangle.  The
 * portion of the this word that's outside the rectangle is filled with 1's
 * (0's if we're looking for runs of 0's).  When we get to the end of this
 * word, we simply neglect to update the temporary histogram for the last run.
 *
 * We can combine the routine that ignores ends with the routine that
 * pays attention to ends by choosing the appropriate starting and ending
 * code.  Since this branch only happens twice per line, the performance
 * penalty for this branch should be negligible.
 *
 * How do we actually compute the internal runs inside each byte?  The
 * first thing to notice is that there are only 512 different values.
 * No matter what we do, it won't take very long.  We just take each byte
 * and look up the length of the run at the right end.  We shift right
 * by that amount plus 1, since we were supposed to ignore the run's bit
 * in the word plus the extra bit that belonged to the previous byte.
 * We also calculate the number of bits in the byte that are left over
 * when we subtract off the bits belonging to the runs on the left and
 * right ends.  In a loop, we then look for the length of the run on
 * the right end of the byte using the regular tables.  If the length
 * is negative, the run terminates in the byte.  Increment the temp
 * histogram bin corresponding to (runLength + 1) (to account for the
 * bit in the LS bit of the index - that bit is part of our byte but is
 * not counted by the table) by the number in the byte's entry in the
 * byte histogram.  Let me try that again.  The number in the byte's bin
 * in the byte histogram corresponds to the number of times this byte
 * was seen in the image.  Therefore, each run we find inside this byte
 * should be weighted by this occurance count before we put it into
 * the final histogram of run lengths.
 * Anyway, we shift the remaining byte bits right by runLength+1 and
 * subtract runLength+1 from the number of bits remaining in the byte
 * and go to the top of the loop.  The loop terminates when the number
 * of bits remaining in the byte is 0.
 *
 * The tables are in the include file runhist.prv.  To utilize processor
 * caches efficiently, we have folded the four tables described above into
 * two.  Almost every time we get the distance a run extends into the byte
 * from the right, we also want the distance the next run extends into
 * the same byte from the left.  So, we interleave the two values so that
 * when one appears in the caches, the other is probably also brought in.
 * The table onesOnEnds holds the number of 1's on the left and right ends
 * of each byte.  The table cyclesOnEnds holds the lengths of the cycles
 * on each end of a byte.
 *
 * Yeah, but what about vertical runs?  Well, the routine that looks at
 * vertical runs will be entirely different than the one that looks at
 * horizontal runs, so we'll just make it a different routine altogether.
 * Here's the horizontal routine.
 *
 ****************************************************************************
 *
 * This file contains implementations of the low level routines that find
 * runs of 0's, runs of 1's and cycles in a single scan line.  They are
 * pretty well described by the comments above.
 *
 * We work our way across the scan line from right to left.  If we're supposed
 * to be discarding the runs at the ends, we carefully follow the first
 * run with another copy of the inner loop, then jump into the main loop.
 * When we get to the left end of the scan line, we either process the
 * last word normally (useEndRuns is true) or process all but the last
 * run in the end word normally (useEndRuns is false).  If useEndRuns is
 * false, we discard the length of the leftmost run.
 *
 * Declaring the following three pointers is our way of telling the compiler
 * not to shift the indexes used to address the tables.
 *
 ****************************************************************************
 *
 * SPACE SAVING, 1/4/95.
 * These routines worked fine, but were too big for XIS's Reading Edge
 * product.  In an effort to save space, we have taken the following
 * actions:
 *   1.	We have combined the routines that detect runs of 1's and runs of
 *	0's into a single routine.  This routine, ip_horizRunHistoLine,
 *	takes an argument that's used to conditionally invert every image
 *	word read.  We do the invert when looking for runs of 0's, so we're
 *	always looking for runs of 1's.  This makes a little more
 *	intuitive sense when debugging (I think).
 *   2.	Since we really only look for runs of 1's now, the table used to
 *	find runs of 0's is eliminated.
 *   3.	The run length table entries have been changed from 32-bit words
 *	to 8-bit bytes.  This saves quite a bit of space, so should improve
 *	cache performance.  However, we do insert an extra sign-extension
 *	when reading the run-length table (takes longer on x86 machines)
 *	and an extra shift when incrementing the histogram of byte values.
 *
 ***************************************************************************
 *
 * BUG FIX: 3/10/95:
 *	XIS wants to use these routines with frameless images.  This should
 *	work except ip_horizRunHistogram processes the image from right to
 *	left and reads the word in the left frame before quitting.  Put
 *	in a check so this word is not read.
 *
 * AESTHETICS FIX: 3/31/95:
 * 	Remove an unnecessary check for byteRunLength equal to 0 in 
 *	ip_horizRunOfCyclesHistoLine.  After retrieving byteRunLength from
 *	the cyclesOnEnds table, the code checks to see if the current
 *	run ends in this byte by checking byteRunLength < 0.  If the 
 *	run doesn't end then the entire byte must be in the run; hence
 *	we should increment the run length by 8.  However, the code 
 *	used to check byteRunLength == 0 before assuming that the run
 *	extends through this byte, but byteRunLength will never be 0
 *	(indicating that we are not currently in a run) since cycles are
 *	always in a run.  Of course, if we are counting runs of 0's or 1's
 *	we may not be in a run and this check is necessary.
 */
#if 0
static Int8 * pZerosOnEndsBytePtr = (Int8 *)(&zerosOnEnds[0]);
#endif /* 0 */
static Int8 * pOnesOnEndsBytePtr = (Int8 *)(&onesOnEnds[0]);
static Int8 * pCyclesOnEndsBytePtr = (Int8 *)(&cyclesOnEnds[0]);

/* The following routine is here only to avoid compiling in two copies
 * of the tables.  This routine just returns a pointer to the table
 * being used for the operation described.
 *
 * 
 */

/* PRIVATE */
/**************************************************************************
 *  ip_getHorizRunTablePtr()
 *		Returns pointer to table used to compute run lengths
 *		for give type of run.
 *
 *	Input:
 *		runType:	Says whether we're looking for runs of 1's,
 *				runs of 0's or cycles.
 *		ppRunTable:	pointer to pointer to first element of
 *				run table.
 *	Action:
 *		*ppRunTable is set to the address of the proper table
 *
 *	Returns:
 *		ia_successful	it worked
 *		ia_invalidParm	runType was illegal
 *
 **************************************************************************/
Int32 CDECL 
ip_getHorizRunTablePtr (
    IPState    *pIPState,
    Int32	runType,
    Int8      **ppRunTable)
{
    PROLOGUE;
    switch (runType)
    {
    case cRunOfZeros:
	/* since the bytes have already been inverted before being stored
	 * in the byte histogram, we really want the table used to find
	 * runs of 1's,  So, fall through to that case.
	 */
#if 0
	*ppRunTable = &zerosOnEnds[0];
	NORMAL_RETURN;
#endif
    case cRunOfOnes:
	*ppRunTable = &onesOnEnds[0];
	NORMAL_RETURN;

    case cRunOfCycles:
	*ppRunTable = &cyclesOnEnds[0];
	NORMAL_RETURN;

    default:
	return ia_invalidParm;
    }
}




/* PRIVATE */
/**************************************************************************
 *  ip_horizRunHistoLine()  
 *		Computes histogram of lengths of runs of 0's or 1's in a
 *		scan line.
 *
 *	Input:
 *		pImage:		Pointer to word containing left end of
 *				rectangle.
 *		rightOffset:	Byte offset of word containing left end
 *				of scan line.
 *		leftWordMask:	Used to mark left end of rectangle inside
 *				leftmost word.
 *		rightWordMask:	Used to mark right end of rectangle inside
 *				rightmost word.
 *		useEndRuns:	When true, record the lengths of the runs
 *				that abut the rectangle boundaries.  Otherwise,
 *				discard them.
 *		pHisto:		Pointer to histogram that will hold lengths
 *				of runs that bridge bytes
 *		pByteHisto:	Pointer to histogram of byte values.  These
 *				are used to reconstruct lengths of runs
 *				contained purely inside bytes.
 *		runOfZeros:	Filled with 0xFFFFFFFF when we're looking
 *				for runs of 0's and 0x00000000 when we're
 *				looking for runs of 1's.
 *
 *	Return code:
 *		ia_successful	it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_horizRunHistoLine (
    IPState    *pIPState,
    UInt8      *pImage,
    UInt32	rightOffset,
    UInt32	leftWordMask,
    UInt32	rightWordMask,
    Int32	useEndRuns,
    UInt32     *pHisto,
    UInt32     *pByteHisto,
    UInt32	runOfZeros)
{
    PROLOGUE;
    UInt32	 iWord;		/* the binary input word */
    UInt8	*pCurImage;	/* pointer to word being processed */
    UInt32	 trialByte;	/* used to hold each byte being examined */
    Int32	 curRunLength;	/* length of the run we're currently
				 * tracking between bytes or words */
    Int32	 byteRunLength;	/* amount by which a run extends into a byte */

	/* point to the right end of the scan line */
    pCurImage = pImage + rightOffset;


	/* Get this word and either get rid of extra bits on the right
	 * end (when we're supposed to use the rightmost run) or
	 * extend the rightmost run to the end (when we're supposed to
	 * ignore it).
	 * If we're supposed to ignore the end run, go trace it.
	 */
	/* NOTE!  Every word read from the image is XORed with
	 * runOfZeros.  The word contains 0x0 if we're really looking
	 * for runs of 1's and 0xFFFFFFFF if we're supposed to be
	 * looking for runs of 0's.  After the XOR, we're always looking for
	 * runs of ones.
	 */
    if (useEndRuns)
    {
	iWord = (ULONG_ACCESS(pCurImage) ^ runOfZeros) & rightWordMask;
	curRunLength = 0;
	trialByte = 0;			/* contents of fake previous word */
    }
    else
    {
	iWord = (ULONG_ACCESS(pCurImage) ^ runOfZeros) | rightWordMask;

	while ((iWord == 0xFFFFFFFF) && (pCurImage > pImage))
	{
	    pCurImage -= 4;		/* point to the next word */
		/* and fetch it */
	    iWord = (ULONG_ACCESS(pCurImage) ^ runOfZeros);
	}

	    /* we've now either found a word with the end of the rightmost
	     * run in it, or have found the leftmost word in the scan line,
	     * or both.  If this is the leftmost word, apply the leftWordMask.
	     * If, after conditionally applying the mask, the word has some
	     * zeros, go find all the runs in it.  If we are supposed
	     * to ignore the leftmost run, applying leftWordMask will make
	     * it extend off the end of the word, so we won't record its
	     * length here anyway.  If we are supposed to record the
	     * length of the leftmost run, applying the leftWordMask may
	     * or may not put 0's at the left end of the word (since the
	     * end of the rectangle may or may not be on a word boundary).
	     * If the rectangle extends to a word boundary and there's a
	     * run there, curRunLength will be nonzero when we're done with
	     * the line.  Otherwise it will be zero.  If curRunLength
	     * is nonzero and useEndRuns is true, (and we're at the
	     * left end of the rectangle), we update
	     * the bin corresponding to curRunLength.
	     *
	     * Note that when we look at the first byte, we know we can
	     * ignore any value in curRunLength.  But when we look at
	     * later bytes, we might be setting curRunLength.  Is this ok?
	     * What if byteRunLength is 0 on the second byte - doesn't that
	     * set curRunLength to some non-negative value?  Won't we
	     * forget we were tracking the first run?  It's ok.  The
	     * only way byteRunLength is set to 0 is when there's a 0
	     * is the LSBit of the byte.  This can only happen if we
	     * have already seen that 0, so the first run has already been
	     * terminated.
	     */
	if (pCurImage == pImage)
	    iWord |= leftWordMask;

	curRunLength = -1000000;	/* big negative number */
	if (iWord != 0xFFFFFFFF) /* there's something in here, so go find the
				  * runs, ignoring the first one.  We find
				  * the first one by initializing
				  * curRunLength to a big negative number,
				  * and testing it before each update to
				  * see if it's still negative */
	{
		/* do byte 3, assume the last bit from the previous word
		 * was a 1.
		 */
	    trialByte = ((iWord << (cRunAddrPos+1)) |
			 (1 << cRunAddrPos)) & cRunAddrMask;
	    byteRunLength = pOnesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of the first
					 * run */
	    {
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run extending from this byte */
		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }

		/* do byte 2 */
	    trialByte = (iWord >> (8-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pOnesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		if (curRunLength > 0)	/* this is not the first run */
		    pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else
	    if (byteRunLength == 0) /* there was no run, but one might start */
	    {
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;

		/* do byte 1 */
	    trialByte = (iWord >> (16-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pOnesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		if (curRunLength > 0)	/* this is not the first run */
		    pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else
	    if (byteRunLength == 0) /* there was no run, but one might start */
	    {
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;

		/* do byte 0 */
	    trialByte = (iWord >> (24-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pOnesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		if (curRunLength > 0)	/* this is not the first run */
		    pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else
	    if (byteRunLength == 0) /* there was no run, but one might start */
	    {
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;
	}	/* finished processing word containing end of rightmost run */
	    /* Since we've just processed the data in iWord, get another word */
	pCurImage -= 4;
	trialByte = iWord;			/* remember contents of
						 * prev word */
	    /* if there's another word in this scan line, get it */
	if (pCurImage >= pImage)
	    iWord = (ULONG_ACCESS(pCurImage) ^ runOfZeros);
    }

	/* At this point, we can be in one of three places.
	 *  1.	iWord contains the word at the right end of the line,
	 *		This happens when we are supposed to use the
	 *		end runs in the line.
	 *  2.	iWord contains some word to the left of the rightmost
	 *	word, possibly the leftmost word.
	 *  3.	iWord contains the word to the left of the leftmost word.
	 *
	 * In all cases, pCurImage points to the memory word that
	 * corresponds to iWord and trialByte contains the previous
	 * word in the line.
	 *
	 * We first work our way to the left until we get to the first
	 * word.
	 * If we get there (weren't already to the left of it), we
	 * apply leftWordMask to it and process it.
	 * When that's done, we look at useEndRuns.  If true (and
	 * curRunLength > 0), we increment pHisto bin corresponding to
	 * this last run.
	 */
    while (pCurImage >= pImage)
    {
	    /* Does iWord contain the leftmost word of the scan line?
	     * If so, apply leftWordMask before processing it.  I could
	     * put this test outside the loop and replicate all this
	     * code, but that's too weird.  Besides, we've just tested
	     * pCurImage agains pImage, so this should just a another
	     * jump.
	     */
	if (pCurImage == pImage)
	{
	    if (useEndRuns)
		iWord &= leftWordMask;
	    else
		iWord |= leftWordMask;
	}

	    /* is the current word full of 1's?  If so, a run does not
	     * end in this word, so just make the previous run (if any)
	     * longer.
	     */
	if (iWord == 0xFFFFFFFF)
	{
	    curRunLength += 32;		/* extend the run */
	}
	else
	    /* is the current word full of 0's and is there a run
	     * in progress?  If so, terminate that run.  If the
	     * word was full of 0's and there was no current run,
	     * just skip the word.
	     */
	if (iWord == 0)
	{
	    if (curRunLength > 0)
	    {
		pHisto[curRunLength]++;	/* end of that run */
		curRunLength = 0;	/* no bits of this word in new run */
	    }
	}
	else	/* runs start and/or stop in here, so look at all
		 * four bytes.
		 */
	{
		/* do byte 3 */
	    trialByte = ((iWord << (cRunAddrPos + 1)) |
			 (trialByte >> (31 - cRunAddrPos))) & cRunAddrMask;
	    byteRunLength = pOnesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of the first
					 * run */
	    {
	        pHisto[curRunLength + (~byteRunLength)]++;
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else
	    if (byteRunLength == 0) /* there was no run, but one might start */
	    {
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else		/* the run extends through this byte */
		curRunLength += 8;

		/* do byte 2 */
	    trialByte = (iWord >> (8-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pOnesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else
	    if (byteRunLength == 0) /* there was no run, but one might start */
	    {
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;

		/* do byte 1 */
	    trialByte = (iWord >> (16-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pOnesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else
	    if (byteRunLength == 0) /* there was no run, but one might start */
	    {
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;

		/* do byte 0 */
	    trialByte = (iWord >> (24-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pOnesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else
	    if (byteRunLength == 0) /* there was no run, but one might start */
	    {
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		curRunLength =
			pOnesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;
	}	/* finished processing all four bytes of a word */

	    /* Since we've just processed the data in iWord, get another word */
	pCurImage -= 4;
	trialByte = iWord;			/* remember contents of
						 * prev word */
	    /* if there's another word in this scan line, get it */
	if (pCurImage >= pImage)
	    iWord = (ULONG_ACCESS(pCurImage) ^ runOfZeros);
    }

	/* At this point, we have just processed the leftmost word.
	 * curRunLength holds the length of the leftmost run, if any.
	 * If useEndRuns is true, update the corresonding histogram
	 * entry.
	 */
    if ((useEndRuns) && (curRunLength > 0))
	pHisto[curRunLength]++;


    NORMAL_RETURN;
}


/* XIS doesn't every look for cycles.  They need the space.  So, if XIS_USE
 * is defined, we turn this routine into a noop.  Otherwise, it's the
 * real routine.
 */


/* PRIVATE */
/**************************************************************************
 *  ip_horizRunOfCyclesHistoLine()  
 *		Computes histogram of lengths of pairs of adjacent runs
 *		of 0's and runs of 1's in a scan line.
 *
 *	Input:
 *		pImage:		Pointer to word containing left end of
 *				rectangle.
 *		rightOffset:	Byte offset of word containing left end
 *				of scan line.
 *		leftWordMask:	Used to mark left end of rectangle inside
 *				leftmost word.
 *		rightWordMask:	Used to mark right end of rectangle inside
 *				rightmost word.
 *		pHisto:		Pointer to histogram that will hold lengths
 *				of runs that bridge bytes
 *		pByteHisto:	Pointer to histogram of byte values.  These
 *				are used to reconstruct lengths of runs
 *				contained purely inside bytes.
 *
 *	Return code:
 *		ia_successful	it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_horizRunOfCyclesHistoLine (
    IPState    *pIPState,
    UInt8      *pImage,
    UInt32	rightOffset,
    UInt32	leftWordMask,
    UInt32	rightWordMask,
    UInt32     *pHisto,
    UInt32     *pByteHisto)
{
#ifdef XIS_USE	/* this routine isn't used by XIS.  They really need the
		 * space on the Reading Edge, so make it go away there.
		 */
    pImage = pImage;
    rightOffset = rightOffset;
    leftWordMask = leftWordMask;
    rightWordMask = rightWordMask;
    pHisto = pHisto;
    pByteHisto = pByteHisto;
    return ia_noImpl;

#else

    PROLOGUE;
    UInt32	 iWord;		/* the binary input word */
    UInt8	*pCurImage;	/* pointer to word being processed */
    UInt32	 trialByte;	/* used to hold each byte being examined */
    Int32	 curRunLength;	/* length of the run we're currently
				 * tracking between bytes or words */
    Int32	 byteRunLength;	/* amount by which a run extends into a byte */

	/* point to the right end of the scan line */
    pCurImage = pImage + rightOffset;

	/* Get this word and extend the rightmost run to the end
	 * since we're always supposed to ignore it.
	 */
    iWord = ULONG_ACCESS(pCurImage) & rightWordMask;
    trialByte = 0;		/* the word off the right end of the image */

	/* We want to test for "end of run"  A run ends where there's
	 * a 0 to the left of a 1.  So we see if (iWord << 1) &
	 * ~iWord is zero.  If this is true, there are NO ends of
	 * runs in this word.
	 */
    while ((((trialByte | (iWord << 1)) & (~iWord)) == 0) &&
	   (pCurImage > pImage))
    {
	pCurImage -= 4;				/* point to the next word */
	trialByte = iWord >> 31;		/* remember MS bit of
						 * prev word */
	iWord = ULONG_ACCESS(pCurImage);	/* and fetch next word */
    }

	/* we've now either found a word with the end of the rightmost
	 * run in it, or have found the leftmost word in the scan line,
	 * or both.  If this is the leftmost word, apply the leftWordMask.
	 * If, after conditionally applying the mask, the word has some
	 * cycles, go find all the runs in it.  Applying leftWordMask
	 * will make the leftmost run extend off the end of the word,
	 * so we won't record its length here anyway.  Since we always
	 * ignore the last cycle in each line, we'll always ignore curRunLength
	 * at the bottom of the loop.  If we find the left word at
	 * this point, we'll skip over the loop, then ignore curRunLength,
	 * so nothing will be recorded when the line contains only one run.
	 *
	 * Note that when we look at the first byte, we know we can
	 * ignore any value in curRunLength.  But when we look at
	 * later bytes, we might be setting curRunLength.  Is this ok?
	 * What if byteRunLength is 0 on the second byte - doesn't that
	 * set curRunLength to some non-negative value?  Won't we
	 * forget we were tracking the first run?  It's ok.  The
	 * only way byteRunLength is set to 0 is when there's a 0
	 * is the LSBit of the byte.  This can only happen if we
	 * have already seen that 0, so the first run has already been
	 * terminated.
	 */
    if (pCurImage == pImage)
	iWord |= leftWordMask;

    curRunLength = -1000000;	/* big negative number */

    if ((trialByte | (iWord << 1)) & (~iWord))
	/* there's something in here, so go find the runs, ignoring
	 * the first one.  We find the first one by initializing
	 * curRunLength to a big negative number, and testing it
	 * before each update to see if it's still negative */
    {
	    /* do byte 3, assume the last bit from the previous word
	     * was a 1.
	     */
	trialByte = ((iWord << (cRunAddrPos+1)) |
		     (trialByte << cRunAddrPos)) & cRunAddrMask;
	byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	if (byteRunLength < 0)	/* we've found the end of the first
				     * run */
	{
		/* there might be other runs inside this byte */
	    pByteHisto[trialByte]++;

		/* get length of run extending from this byte */
	    curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	}

	    /* do byte 2 */
	trialByte = (iWord >> (8-cRunAddrPos-1)) & cRunAddrMask;
	byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	if (byteRunLength < 0)	/* we've found the end of a run */
	{
	    if (curRunLength > 0)	/* this is not the first run */
		pHisto[curRunLength + (~byteRunLength)]++;

		/* there might be other runs inside this byte */
	    pByteHisto[trialByte]++;

		/* get length of run on left extending from this byte */
	    curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	}
	else	/* the run just extends through this byte */
	    curRunLength += 8;

	    /* do byte 1 */
	trialByte = (iWord >> (16-cRunAddrPos-1)) & cRunAddrMask;
	byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	if (byteRunLength < 0)	/* we've found the end of a run */
	{
	    if (curRunLength > 0)	/* this is not the first run */
		pHisto[curRunLength + (~byteRunLength)]++;

		/* there might be other runs inside this byte */
	    pByteHisto[trialByte]++;

		/* get length of run on left extending from this byte */
	    curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	}
	else	/* the run just extends through this byte */
	    curRunLength += 8;

	    /* do byte 0 */
	trialByte = (iWord >> (24-cRunAddrPos-1)) & cRunAddrMask;
	byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	if (byteRunLength < 0)	/* we've found the end of a run */
	{
	    if (curRunLength > 0)	/* this is not the first run */
		pHisto[curRunLength + (~byteRunLength)]++;

		/* there might be other runs inside this byte */
	    pByteHisto[trialByte]++;

		/* get length of run on left extending from this byte */
	    curRunLength =
		    pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	}
	else	/* the run just extends through this byte */
	    curRunLength += 8;
    }	/* finished processing word containing end of rightmost run */
	/* Since we've just processed the data in iWord, get another word */
    pCurImage -= 4;
    trialByte = iWord >> 31;		/* remember contents of
					 * prev word */
	/* if there's another word in this scan line, get it */
    if (pCurImage >= pImage)
	iWord = ULONG_ACCESS(pCurImage);

	/* At this point, we can be in one of three places.
	 *  1.	iWord contains the word at the right end of the line,
	 *		This happens when we are supposed to use the
	 *		end runs in the line.
	 *  2.	iWord contains some word to the left of the rightmost
	 *	word, possibly the leftmost word.
	 *  3.	iWord contains the word to the left of the leftmost word.
	 *
	 * In all cases, pCurImage points to the memory word that
	 * corresponds to iWord and trialByte contains the previous
	 * word in the line.
	 *
	 * We first work our way to the left until we get to the first
	 * word.
	 * If we get there (weren't already to the left of it), we
	 * apply leftWordMask to it and process it.
	 * When that's done, we look at useEndRuns.  If true (and
	 * curRunLength > 0), we increment pHisto bin corresponding to
	 * this last run.
	 */
    while (pCurImage >= pImage)
    {
	    /* Does iWord contain the leftmost word of the scan line?
	     * If so, apply leftWordMask before processing it.  I could
	     * put this test outside the loop and replicate all this
	     * code, but that's too weird.  Besides, we've just tested
	     * pCurImage agains pImage, so this should just a another
	     * jump.
	     */
	if (pCurImage == pImage)
	{
	    iWord |= leftWordMask;
	}

	    /* Does a cycle end in the current word? If not, just make the
	     * previous run (if any) longer.
	     */
	if (((trialByte | (iWord << 1)) & (~iWord)) == 0)
	{
	    curRunLength += 32;		/* extend the run */
	}
	else	/* runs start and/or stop in here, so look at all
		 * four bytes.
		 */
	{
		/* do byte 3 */
	    trialByte = ((iWord << (cRunAddrPos + 1)) |
			 (trialByte << cRunAddrPos)) & cRunAddrMask;
	    byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of the first
					 * run */
	    {
	        pHisto[curRunLength + (~byteRunLength)]++;
		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else		/* the run extends through this byte */
		curRunLength += 8;

		/* do byte 2 */
	    trialByte = (iWord >> (8-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;

		/* do byte 1 */
	    trialByte = (iWord >> (16-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;

		/* do byte 0 */
	    trialByte = (iWord >> (24-cRunAddrPos-1)) & cRunAddrMask;
	    byteRunLength = pCyclesOnEndsBytePtr[trialByte];
	    if (byteRunLength < 0)	/* we've found the end of a run */
	    {
		pHisto[curRunLength + (~byteRunLength)]++;

		    /* there might be other runs inside this byte */
		pByteHisto[trialByte]++;

		    /* get length of run on left extending from this byte */
		curRunLength =
			pCyclesOnEndsBytePtr[trialByte+cRunOnLeftAddrOffset];
	    }
	    else	/* the run just extends through this byte */
		curRunLength += 8;
	}	/* finished processing all four bytes of a word */

	    /* Since we've just processed the data in iWord, get another word */
	pCurImage -= 4;
	trialByte = iWord >> 31;		/* remember contents of
						 * prev word */
	    /* if there's another word in this scan line, get it */
	if (pCurImage >= pImage)
	    iWord = ULONG_ACCESS(pCurImage);
    }

	/* At this point, we have just processed the leftmost word.
	 * curRunLength holds the length of the leftmost run, if any.
	 * Since we never count the last run when counting cycles,
	 * ignore curRunLength.
	 */


    NORMAL_RETURN;

#endif	/* XIS_USE */
}



/*

$Log: mirnhst1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.9  1998/01/13 23:53:15  ddavies
mirnhst1.c -> /product/ipcore/ipshared/src/RCS/mirnhst1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1995/12/13 02:51:32  ejaquith
mirnhst1.c -> /product/ipcore/ipshared/src/RCS/mirnhst1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.7  1995/04/14  01:19:17  jcraig
 * jcraig on Thu Apr 13 18:16:15 PDT 1995
 *
 * mirnhst1.c -> /product/ipcore/ipshared/src/RCS/mirnhst1.c,v
 *
 * remove a useless check for byteRunLength == 0  in the cycle
 * histogramming code.  byteRunLength == 0, indicates that there is
 * no cycle to keep track of when going between bytes.  this cannot
 * happen in the cycle code - we are always in a cycle (although it
 * may happen if looking for runs of ones or zeros).
 *
 * Revision 1.6  1995/03/10  20:35:30  ddavies
 * ddavies on Fri Mar 10 12:34:05 PST 1995
 *
 * mirnhst1.c -> /product/ipcore/ipshared/src/RCS/mirnhst1.c,v
 *
 * ip_horizRunHistoLine and ip_horizRunOfCyclesHistoLine were needlessly
 * reading from the left frame.  This caused problems for XIS who had frameless
 * images.  Put in an extra check before each image read to fix this.
 *
 * Revision 1.5  1995/01/11  22:32:58  dferrell
 * mirnhst1.c -> /product/ipcore/ipshared/src/RCS/mirnhst1.c,v
 *
 * Changed include of shrip.pub and shrip.prv to shrcc.pub and shrcc.prv,
 * respectively.
 *
 * Revision 1.4  1995/01/05  04:04:47  ddavies
 * mirnhst1.c -> /product/ipcore/ipshared/src/RCS/mirnhst1.c,v
 *
 * The routines ip_horizRunOf0sHistoLine and ip_horizRunOf1sHistoLine have
 * been consolidated into ip_horizRunHistoLine.  The routine
 * ip_horizRunOfCyclesHistoLine is now a dummy when XIS_USE is defined.
 * The XIS folks never call it and need the space on the Reading Edge.
 * The run length tables now contain bytes instead of words to save space.
 * All these measures saved about 11.5KB in this file.  The macros
 * CONST_LONG_ACCESS have been removed when accessing the run length tables
 * since we get bytes now (and expect the processor to sign-extend them).
 * The ULONG_ACCESS macros have been removed from updates to the byte
 * histograms since we now expect the compiler to shift the table index
 * instead of having it preshifted (we used to use the same index for
 * the run length tables and the byte histograms).  The net effect of
 * reducing the table size and adding the shifts seems to be negligible
 * in performance.
 *
 * Revision 1.3  1994/08/01  21:08:20  ddavies
 * ddavies on Mon Aug 1 14:08:16 PDT 1994
 *
 * mirnhst1.c -> /product/ipcore/ipshared/src/RCS/mirnhst1.c,v
 *
 * Change "static const" to "static" to work around buggy gcc/Purify interaction.
 *
 * Revision 1.2  1994/06/28  22:08:18  ddavies
 * mirnhst1.c -> /product/ipcore/ipshared/src/RCS/mirnhst1.c,v
 *
 * Add real implementations of ip_horizRunOf1sHistoLine and
 * ip_horizRunOfCyclesHistoLine.  These routines produce histograms of
 * lengths of runs of 1's and lengths from one 01 transition to the next one
 * respectively.  This completes (first pass of) the horizontal run length
 * histogramming code.
 *
 * Revision 1.1  1994/06/27  22:54:16  ddavies
 * Initial revision
 *

*/


