/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  mirnhst2.c: machine independent routines for computing histograms
 *		of horizontal run lengths.
 *
 *	Int32	ip_getVertRunTablePtr()	returns pointer to table used to
 *					find vertical run lengths
 *	Int32	ip_vertRunHistoLine()	looks at runs of 0's, 1's or cycles
 *					in a row of 8x8 blocks
 *	void	ip_vertRunHistoBlock()	looks at runs of 0's 1's or cycles
 *					in a single 8x8 block
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "vrunhist.prv"

	/* prototypes */
#include "shrcc.pub"
#include "shrcc.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: mirnhst2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");


/* It is useful for segmentation to be able to figure out how big things
 * are.  A cheap way to do this is to look at histograms of run lengths.
 * Histograms of lengths of white space can be used to figure out how
 * far apart things are.  This file contains routines that compute histograms
 * of vertical runs in a single scan line.
 *
 * These routines are called from i_rnhst2.c.  That file has many useful
 * comments at the top.  Since they describe the upper and lower level
 * routines, we'll include them here too.
 *
 *****************************************************************************
 * Because the horizontal and vertical implementations are similar, we'll
 * include the bulk of the comments from i_rnhst1.c.
 *
 ***************************************************************************
 *
 * XIS asked for a routine that has a bunch of arguments.  We've added
 * a couple more.  The arguments are:
 *
 *	pointers to image and rectangle within the image;
 *	pHisto: pointer to histogram to be returned;
 *	numHistoBins: equals the number of bins in the histogram
 *	sampleFreq: skip sampleFreq-1 lines between each sampled row
 *		or column (our case, skip in units of 8 columns);
 *	runType: whether to count black runs, white runs or cycles (lengths
 *		of black/white pairs);
 *	useEndRuns: all runs are terminated at the boundary of the
 *		sampling rectangle.  If useEndRuns is true, we should
 *		count the runs that touch the boundary.  If it is
 *		false, we should ignore these runs.  This is a pain,
 *		but is useful.
 *		Well, in fact, NOT ignoring cycles that abut the
 *		borders is not a well defined operation.  Besides,
 *		it's hard to implement.  So we'll say that if we are
 *		measuring cycles, we always ignore cycles that abut
 *		the edges of the rectangle.
 *	countLongRuns:	if false, we ignore runs that have no bins in
 *		the histogram.  If true, all the counts for the large
 *		runs are put into to the last histogram bin.
 *	weightByLength: if false, finding a run of length x causes us to
 *		add 1 to bin x.  If true, finding a run of length x causes
 *		us to add x to bin x.  What is this for anyway?
 *	
 * This is quite a monster - the swiss army knife approach to writing code.
 * Unfortunately, I think we actually use this stuff.
 *
 * We'd really like to avoid making a zillion decisions for every run found.
 * One way to avoid some decisions is to create a temporary histogram,
 * then process it to produce the histogram we really want.  We will create a
 * temporary histogram containing a bin for every possible run length.
 * If we're looking for horizontal runs, the temporary histogram will have
 * bins for 1 through rectangleWidth.  If we're tracing vertical runs,
 * we need bins for 1 through rectangleHeight.  When we've filled the
 * temporary histogram with run info from the image, we start post processing.
 *
 * If weightByLength is true, we multiply the number in each bin by the
 * bin number.
 *
 * We then add the contents of the first numHistoBins bins from the
 * temporary histogram to the contents of the output histogram.
 *
 * If countLargeRuns is true, we add the contents of the remaining bins
 * in the temporary histogram to the pHisto[numHistoBins-1].
 *
 * Let's review how horizontal histograms are counted.  We will rotate each
 * byte in the image, then use pretend we're counting horizontally
 * on the rotated bytes.
 *
 * Even with just horizontal histograms, the choice of counting runs of
 * 1's, runs of 0's or cycles is not one we'd like to make on every
 * image sample.  We'll have different code for each of these cases.
 * What are cycles?  When finding areas containing periodic data, like
 * glyphs or halftones, it's best to count not just the width of each
 * mark or the width of the space between them, but the period of
 * white/black pairs.  This is the thing that's most likely to be constant.
 * Another way to think of this is that we count the distance between
 * 0-to-1 transitions.
 *
 * The really messy remaining parameter is useEndRuns.  It is much more
 * convenient to count all the runs.  When useEndRuns is false, we must
 * efficiently trace the leftmost and rightmost runs on each scan line,
 * then ignore them.  If we were tracing each run serially, this would
 * be trivial.  Since we're trying to process as much data as we can
 * on each sample, this will be messy.  We will have separate verions of
 * run counting code for these too.
 *
 * The routine at this level, i_horizRunHistogram, will call lower routines
 * to process each line.  These routines will have the real inner loops.
 * What do the inner loops of these routines look like?  The inner loops
 * are similar for all the routines.  The cycle routine looks for somewhat
 * different things than the routines that look for 1's or 0's, but
 * the principle is the same.  The SkipEnds and UseEnds routines differ
 * only in how they start and stop.
 *
 * We clearly want to look at more than one pixel at a time.  The inner
 * loops should be able to skip over runs of 0's and runs of 1's in large
 * jumps.  This is especially true since these are fairly common.  On the
 * other hand, a large chunk of bits might contain many internal runs,
 * so we can't just deal with large chunks all the time.
 *
 * One cute observation is that the inner loops have two basic jobs.  One
 * is to track runs that bridge chunks.  The other is to track runs that
 * start and stop inside individual chunks.  Tracking runs that start and
 * stop purely within chunks need not be done as each chunk is encountered.
 * We can create yet another histogram, this one of chunk values seen.
 * When we get done with the image, we calculate the number of runs that
 * started and stopped inside each chunk and add that into the temporary
 * histogram.  For example, say chunks are bytes and we find an 0xAA.  We
 * increment the 0xAA bin.  At the end of the image, say we've found
 * 70,000 0xAA's.  Say we're tracking runs of 1's.  This byte (1010 1010)
 * has three internal runs (plus one on the left that bridges to the next
 * byte, each of length 1.  We would add 3 * 70,000 to the
 * temporary histogram bin for runs of length 1.
 *
 * Using this trick, we only have to track runs between chunks when
 * we're actually going through the image - the runs inside chunks are
 * done later.  This greatly simplifies the code done for each byte.
 *
 * We use tables to compute things about chunks.  To keep table sizes
 * small, we'd like the chunk size to be fairly small.
 * To be able to skip over runs of black or white efficiently, we'd like
 * the chunk size to be large.  We'll use a combination.  We process the
 * image in units of 32-bit words.  If a word is entirely black or entirely
 * white, we skip over it easily.  If it contains both black and white,
 * we process its four bytes individually.  At the end of the four bytes,
 * we look at the next word.
 *
 * There are two ways we can do the calculations on each byte: we can use
 * a state machine that uses each byte as input or we can derive both
 * the state and output from the region around each byte.  Huh?  The
 * state machine approach is straightforward.  Say we're looking for runs
 * of 1's.  The current state is set by the previous byte - we're either
 * in a run of 0's or a run of 1's:
 *	State	byte starts with	Action
 *	0's	0			keep looking for 1's
 *	0's	1			start new run
 *	1's	0			end current run, start looking for
 *					another
 *	1's	1			continue current run.
 *
 * This is lovely, but state machines are sort of complicated what with
 * all the testing a branching that goes on.
 *
 * Note that the state from the previous byte is equal to the last bit of
 * that byte together with the length of the run we're now in (if any).
 * We could eliminate the state machine stuff by just concatenating that
 * bit with the others in the byte when doing our table lookups.
 *
 * What table lookups?  We use three tables for each byte:
 *   1.	Incremental run length
 *	This table performs two functions.  It indicates when a run from
 *	the previous byte ends and gives the distance by which that run
 *	extends into this byte.  The run ends when the value in this
 *	table is negative.  The trick is that we need two values of 0.
 *	When a run ends at the beginning of this byte, we should store
 *	negative 0.  When there is no run coming into this byte, we store
 *	positive 0.  Since we need both positive and negative values for
 *	0, we use one's complement to store negative values.
 *   2.	Byte Value
 *	This is a histogram of byte values.  When we come to the end of
 *	a run, we increment the bin corresponding to this byte.  We use
 *	these counts later to add in the runs that were purely internal
 *	to the bytes.
 *   3.	New Run Length
 *	This is the number of bits in the run that includes the last
 *	bit of the byte.  We use it to set the run Length counter before
 *	encountering the next byte.
 * 
 * One nice feature of organizing things this way is that the inner loops
 * of the code that finds cycles and the code that finds runs of 1's (or 0's)
 * is all the same.  The only difference is the contents of the tables.
 * But, we use different routines anyway.  If we use different routines, we
 * can address the tables directly in each routine.  If we used one
 * routine, we'd have to pass a pointer to the table, which would use
 * a register.  There aren't a lot of registers in a x86 processor, so
 * we'll use different routines.
 *
 * Do we scan left to right or right to left within a line?  When we
 * encounter a new word, we would like to perform a quick check to see
 * if its the entension to an existing run or contains the end of that
 * run.  We can do this by looking at the last bit of the previous word
 * and the value of the new word:
 *	last bit of prev word	new word	Action
 *		0		00...00		Add 32 to runLength unless
 *						we're looking for runs of 1's
 *		1		11...11		Add 32 to runLength unless
 *						we're looking for runs of 0's
 *		all other combinations		Find end of run in word
 * It's easy to branch on the last bit of the previous word if that bit
 * is in the sign bit position (MSBit).  Therefore, we will go from right
 * to left within a scan line.
 *
 * Do we read each byte individually or do we read entire words into a
 * register and do shifting and masking to form the table lookup values?
 * Reading bytes individually is generally a good idea, but maybe not in
 * this case.  First, there is the annoyance of dealing with different
 * byte orders.  If we read by words, this goes away.  That's not a
 * big deal.  Second, at least the byte value histogram entries are not
 * bytes.  Therefore, the machine will do some shifting anyway.  The
 * current run length will not be a byte, so the value read from the
 * Incremental run length table will either be expanded into a word
 * (an operation that our compilers do very stupidly on PCs) or the
 * incremental run length table entries will be words already.  Ditto for
 * the new run length table.  We can save some shifting of address
 * offsets by making all the tables have the same entry size, so pick
 * 32-bit words.  This also eliminates the operations used to expand
 * smaller entries to words when they're read.
 *
 * Next, we're not addressing at least the incremental run length table
 * with bytes.  We use 9 bits - 8 from the current byte and 1 from the
 * previous byte.  We'd have to compose these entries anyway, so we might
 * as well do this from a register that hold all the bytes at once.  So
 * long as one of the tables is addressed using 9 bits, we might as well
 * address all of them that way - again to avoid extra address computation.
 * Reading the image in units of words and processing it in units of runs
 * of bits from these words has another advantage when processing the
 * end runs as seen below.
 *
 * What do we do with the end runs anyway?  When we are keeping the end
 * runs and looking for runs of 1's (or 0's), we just smash the portions of the
 * works on the far left and right that are outside our rectangle with
 * 0's (or 1's).  When looking for cycles, we smash the portion of the
 * word on the right with 0's (since that cannot start a new run) and
 * the portion of the leftmost word outside the rectangle with 1's
 * (since that cannot end a run).  When ignoring the end runs, we'll
 * duplicate the inner loop.  The first copy is executed only for the
 * very first, rightmost, run.  It does the normal things, except that
 * it doesn't update the temporary histogram for the first run length.
 * It then falls through to the ordinary loop.
 *
 * To ignore the leftmost run, we simply make the inner loop stop before
 * processing the word containing the left end of the rectangle.  The
 * portion of the this word that's outside the rectangle is filled with 1's
 * (0's if we're looking for runs of 0's).  When we get to the end of this
 * word, we simply neglect to update the temporary histogram for the last run.
 *
 * We can combine the routine that ignores ends with the routine that
 * pays attention to ends by choosing the appropriate starting and ending
 * code.  Since this branch only happens twice per line, the performance
 * penalty for this branch should be negligible.
 *
 * How do we actually compute the internal runs inside each byte?  The
 * first thing to notice is that there are only 512 different values.
 * No matter what we do, it won't take very long.  We just take each byte
 * and look up the length of the run at the right end.  We shift right
 * by that amount plus 1, since we were supposed to ignore the run's bit
 * in the word plus the extra bit that belonged to the previous byte.
 * We also calculate the number of bits in the byte that are left over
 * when we subtract off the bits belonging to the runs on the left and
 * right ends.  In a loop, we then look for the length of the run on
 * the right end of the byte using the regular tables.  If the length
 * is negative, the run terminates in the byte.  Increment the temp
 * histogram bin corresponding to (runLength + 1) (to account for the
 * bit in the LS bit of the index - that bit is part of our byte but is
 * not counted by the table) by the number in the byte's entry in the
 * byte histogram.  Let me try that again.  The number in the byte's bin
 * in the byte histogram corresponds to the number of times this byte
 * was seen in the image.  Therefore, each run we find inside this byte
 * should be weighted by this occurance count before we put it into
 * the final histogram of run lengths.
 * Anyway, we shift the remaining byte bits right by runLength+1 and
 * subtract runLength+1 from the number of bits remaining in the byte
 * and go to the top of the loop.  The loop terminates when the number
 * of bits remaining in the byte is 0.
 *
 * The tables are in the include file runhist.prv.  To utilize processor
 * caches efficiently, we have folded the four tables described above into
 * two.  Almost every time we get the distance a run extends into the byte
 * from the right, we also want the distance the next run extends into
 * the same byte from the left.  So, we interleave the two values so that
 * when one appears in the caches, the other is probably also brought in.
 * The table onesOnEnds holds the number of 1's on the left and right ends
 * of each byte.  The table cyclesOnEnds holds the lengths of the cycles
 * on each end of a byte.
 *
 * Yeah, but what about vertical runs?  Well, the routine that looks at
 * vertical runs will be entirely different than the one that looks at
 * horizontal runs, so we'll just make it a different routine altogether.
 * Here's the horizontal routine.
 *
 ************************************************************************
 *
 * So much for histograms of horizontal runs.  How does this apply to
 * vertical runs?
 *
 * Let's say more about this rotation stuff.  Clearly, it would be easy
 * to rotate the entire image, then do horizontal histograms.  Clients
 * could do that easily.  The only reason this routine exists is because
 * the clients ran out of memory - they couldn't afford the space required
 * for the rotated copy.  Therefore, we have to do something clever.
 *
 * We could just read down entire columns of the image.  Unfortunately,
 * the image is stored in memory by horizontal scan lines.  If we move
 * vertically through the image, we access a slice of every scan line.
 * This means we fault every level of the memory heirarchy (processor primary
 * cache, secondary cache and main memory) on every major access.
 * Any possible performance gains we could see by accessing the image
 * vertically are overwhelmed by all this faulting.  We really want
 * to read the image (roughly) horizontally.
 *
 * However, we don't want to read individual scan lines either.  Since we're
 * really computing vertical run lengths, we'd have to look at each
 * pixel individually - incrementing the corresponding run length count
 * if the pixel has the right polarity.  We could try keeping the run
 * length counts in some weird vertical order (one byte holds bit 0 of
 * 8 different counts, the next one holds bit 1 of the eight counts and
 * so on).  Since we need 16 bit counters, we need 32 operations to increment
 * the 8 counts AND we have to decide if we've reached the end of a run
 * in any of the columns anyway.  So, we really don't want to do something
 * with histogram data every time we see a scan line.
 *
 * We'll compromise.  We read the image in units of 8 scan lines - 7 new
 * ones and 1 old one.  In the horizontal code described above, we used
 * 8 new ones and 1 old one.  Why change?  Well, we really read columns
 * of bytes, rotating them into the 8 nibbles of a word as we go.  Each
 * nibble represents the bits in one column of data.  We do this twice,
 * then combine the nibbles to produce 8 bytes.  The 8 bytes are used in
 * a table lookup (as described above) to update the current run length
 * for each column as well as decide when a run is complete.  Yes, this
 * means using different tables.  Bummer.
 *
 * What do we do about the edge conditions?  We would like to base this code
 * on the orthogonal rotation code that blithely ignore the edges - it rotates
 * the frame as needed.  We have to stop exactly on the edges.
 *
 * First, we decide which type of rotation we want.  We choose 270 degree
 * (as opposed to 90 degree) rotation.  This has two advantages.  The
 * existing code works from the bottom of the source image towards the
 * top.  This corresponds to working from right to left in the rotated
 * image, which is what we want.
 *
 * Both 90 and 270 rotation divide the image into blocks of width 8 and
 * height 32.  The rotation operation puts together 8 32-bit words and
 * writes them to memory.  We could also produce 8 32-bit words, then
 * compute histograms from each.  This allows us to closely follow the
 * logic of the horizontal histogram code.  On the other hand, it forces
 * us to spend a bunch of cycles putting the words together, then take
 * them apart to compute the histograms.  We'll just compute the histograms
 * on each block of 8 bytes, then throw them away.
 *
 * The simplest thing to do is to process an 8x8 block, then move to the
 * next block on to the right.  Do we gain anything interesting from
 * processing several 8x8 blocks vertically before moving horizontally?
 * We could keep using the same run length counters for a longer time
 * (there's one run length counter for each column of bits).  Well, the
 * run length counters are going to be kept in an array anyway.  There
 * are at least 8 of them per 8x8 block.  That's too many to keep in
 * x86 registers - probably too many for the SPARC compiler to handle
 * as well.  So the compiler will almost certainly keep the things in
 * memory anyway.  Since we don't eliminate any memory reads or writes,
 * we might as well keep the loop simple and just go from right to left.
 *
 * Back to edge conditions.  In 270 degree rotation, the upper right corner
 * of the source image becomes the upper left corner of the dest.
 * Therefore, the grid of blocks is aligned over the source image so
 * that the upper right corner of the grid corresponds to the upper
 * right corner of the image.  The top row of blocks is exactly aligned
 * with the top row of the image and the right row of blocks lines up
 * with the right row of the image.  Therefore, the left and bottom sides
 * either happen to also line up or the blocks hang off the edges of the
 * image.
 *
 * In our case, we will align the top row of the source rectangle with
 * the top row of blocks.  The right side of the column of bytes
 * containing the right edge of the rectangle will be aligned with the
 * right edge of the blocks.  In our case, there can be extra bits on
 * the left, right and bottom.  In fact, even if there are no extra
 * bits on the bottom, we have to make one up for the very bottom row
 * since we need to make up a bit from the "previous" row.
 *
 * So, the things we can do when we read a blocks are:
 *	1. apply the left mask.
 *	2. apply the right mask.
 *	3. apply the left and right masks.
 *	4. apply the bottom mask.
 *	5. apply the left and bottom masks.
 *	6. apply the right and bottom masks.
 *	7. apply the left, right and bottom masks.
 *	8. apply no mask.
 *
 * The last condition happens almost all the time, except in very small
 * images where performance isn't a problem.  We'll have two
 * versions of the code that processes 8x8 blocks.  One applies all the
 * masks.  The other applies none of them.  We have to set up all the masks
 * before calling the first block of code.  We call it all along the bottom
 * of the rectangle and at both left and right edges of each line.  Just
 * to keep things tidy, we'll make it a subroutine.  We only call the
 * no-mask processing code from one place, so we just write it out.
 *
 * The inner loops are implemented in mirnhst2.c
 *
 ************************************************************************
 *
 * So much for the comments from i_rnhst2.c.
 *
 * One difference from the horizontal run length code is that we have
 * one primitive routine instead of three.  We used three in the
 * horizontal code so we wouldn't have to use up a register with a
 * pointer to the table used to find run lengths.  Well, there are already
 * so many registers here, that another one won't matter.  Besides, this
 * routine is already pretty big, so I don't want to replicate it.
 *
 */


/* The following routine is here only to avoid compiling in two copies
 * of the tables.  This routine just returns a pointer to the table
 * being used for the operation described.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_getVertRunTablePtr()
 *		Returns pointer to table used to compute run lengths
 *		for give type of run.
 *
 *	Input:
 *		runType:	Says whether we're looking for runs of 1's,
 *				runs of 0's or cycles.
 *		ppRunTable:	pointer to pointer to first element of
 *				run table.
 *	Action:
 *		*ppRunTable is set to the address of the proper table
 *
 *	Returns:
 *		ia_successful	it worked
 *		ia_invalidParm	runType was illegal
 *
 **************************************************************************/
Int32 CDECL 
ip_getVertRunTablePtr (
    IPState    *pIPState,
    Int32	runType,
    Int8      **ppRunTable)
{
    PROLOGUE;
    switch (runType)
    {
    case cRunOfZeros:
	*ppRunTable = &zerosOnVertEnds[0];
	NORMAL_RETURN;

    case cRunOfOnes:
	*ppRunTable = &onesOnVertEnds[0];
	NORMAL_RETURN;

    case cRunOfCycles:
	*ppRunTable = &cyclesOnVertEnds[0];
	NORMAL_RETURN;

    default:
	return ia_invalidParm;
    }
}


/* PRIVATE */
/**************************************************************************
 *  ip_vertRunHistoLine()  
 *		Computes histogram of lengths of runs of 0's, 1's or
 *		cycles in a row of 8x8 blocks across an image.
 *
 *	Input:
 *		pImage:			Pointer to word containing left end of
 *					rectangle.
 *		bpl:			image bytes/line
 *		leftByteAndMask, leftByteOrMask: Used to mark left end of
 *					rectangle inside leftmost byte.
 *		rightByteAndMask, rightByteOrMask: Used to mark right end of
 *					rectangle inside rightmost word.
 *		pHisto:			Pointer to histogram that will hold
 *					lengths of runs that bridge bytes
 *		pByteHisto:		Pointer to histogram of byte values.
 *					These are used to reconstruct lengths 
 *					of runs contained purely inside bytes.
 *		pRotationTable:		Pointer to table used to turn
 *					bytes into rotated nibbles.
 *		pRunTable:		Pointer to table used to find
 *					run lengths
 *		pRunLengths:		Pointer to array of current run lengths.
 *					There's one entry for every image column
 *					we're considering
 *		blockCount:		number of rotation blocks we see in one
 *					horizontal pass through the image.
 *		sampleFreq:		Skip sampleFreq - 1 blocks between
 *					blocks that are sampled.
 *		blockAndMask:		Specifies the bits to be preserved
 *					in the block.  If this is not all 1's,
 *					we are at the bottom of the rect.
 *		blockOrMask:		Contains 1's in bits that should be
 *					set in rotated bytes from bottom
 *					row.
 *
 *	Return code:
 *		ia_successful	it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_vertRunHistoLine (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	bpl,
    UInt8	leftByteAndMask,
    UInt8	leftByteOrMask,
    UInt8	rightByteAndMask,
    UInt8	rightByteOrMask,
    UInt32     *pHisto,
    UInt32     *pByteHisto,
    UInt32     *pRotationTable,
    Int8       *pRunTable,
    Int32      *pRunLengths,
    UInt32	blockCount,
    Int32	sampleFreq,
    UInt32	blockAndMask,
    UInt32	blockOrMask)
{
    PROLOGUE;
    UInt32	 iWord0, iWord1;    /* rotated input words; each nibble holds
				     * rotated data from one of the 8
				     * columns being read. */
    UInt32	 rotBytes;	    /* holds rotated bytes, these are composed
				     * from rotated nibbles in iWord0
				     * and iWord1.
				     */
    UInt32	 trialByte;	    /* used to hold each byte being examined */
    Int32	 byteRunLength;	    /* amount by which a run extends
				     * into a byte */
    Int32	 interBlockAddrOffset; /* the number you add to get from the end
				     * of one block (after processing it) to
				     * the beginning of the next one */

	/* We're going to process a row of blocks.  If the row is
	 * the bottom row, we need to process all its blocks using
	 * ip_vertRunHistoBlock.  Otherwise, we only process the
	 * far left and right blocks using ip_vertRunHistoBlock.
	 * If blockAndMask is not 0xFFFFFFFF, we're on the bottom
	 * row, otherwise, we're not.
	 */
    if (blockAndMask != 0xFFFFFFFF)
    {
	    /* Do the block on the left.  If there's only one block,
	     * apply right masks too.  Note that the right masks were
	     * computed so that they only do something if the rightmost
	     * block intersected the right edge of the rectangle, so
	     * we can just look at blockCount to decide whether to
	     * apply the right mask here.
	     */
	ip_vertRunHistoBlock (pIPState,
			      pImage,
			      bpl,
			      (UInt8)(blockCount == 1 ?
				rightByteAndMask & leftByteAndMask :
				leftByteAndMask),
			      (UInt8)(blockCount == 1 ?
				rightByteOrMask | leftByteOrMask :
				leftByteOrMask),
			      blockAndMask,
			      blockOrMask,
			      pHisto,
			      pByteHisto,
			      pRotationTable,
			      pRunTable,
			      pRunLengths);

	    /* Point to the next block.  While there are blocks before
	     * the rightmost one, process them.
	     */
	pImage += sampleFreq;
	blockCount--;
	pRunLengths += 8;

	for (;
	     blockCount > 1;
	     pImage += sampleFreq, blockCount--, pRunLengths += 8)
	{
	    ip_vertRunHistoBlock (pIPState,
				  pImage,
				  bpl,
				  0xFF,		/* AndMask */
				  0,		/* orMask */
				  blockAndMask,
				  blockOrMask,
				  pHisto,
				  pByteHisto,
				  pRotationTable,
				  pRunTable,
				  pRunLengths);
	}

	    /* if we haven't already done it, do the rightmost block */
	if (blockCount == 1)
	{
	    ip_vertRunHistoBlock (pIPState,
				  pImage,
				  bpl,
				  rightByteAndMask,
				  rightByteOrMask,
				  blockAndMask,
				  blockOrMask,
				  pHisto,
				  pByteHisto,
				  pRotationTable,
				  pRunTable,
				  pRunLengths);
	}
    }	/* end of processing the bottom row */

    else	/* we're not on the bottom row, so do all the blocks
		 * between the left and right ones right here
		 */
    {
	    /* Do the block on the left.  If there's only one block,
	     * apply right masks too.  Note that the right masks were
	     * computed so that they only do something if the rightmost
	     * block intersected the right edge of the rectangle, so
	     * we can just look at blockCount to decide whether to
	     * apply the right mask here.
	     */
	ip_vertRunHistoBlock (pIPState,
			      pImage,
			      bpl,
			      (UInt8)(blockCount == 1 ?
				rightByteAndMask & leftByteAndMask :
				leftByteAndMask),
			      (UInt8)(blockCount == 1 ?
				rightByteOrMask | leftByteOrMask :
				leftByteOrMask),
			      0xFFFFFFFF,	/* blockAndMask */
			      0,		/* blockOrMask */
			      pHisto,
			      pByteHisto,
			      pRotationTable,
			      pRunTable,
			      pRunLengths);

	    /* Point to the next block.  While there are blocks before
	     * the rightmost one, process them.
	     */
	pImage += sampleFreq;
	blockCount--;
	pRunLengths += 8;

	    /* We process each block from the top to the bottom.  When
	     * we're done, we add interBlockAddrOffset to the address
	     * to make it point to the top of the next block.  This
	     * is the distance back up to the top plus the distance
	     * we move sideways.
	     */
	interBlockAddrOffset = -(bpl * 8) + bpl + sampleFreq;

	    /* if we're on a little endian machine, we have to fix
	     * up the bottom bits of pImage.  Since we know that bpl
	     * is a multiple of four, we only have to do this at the
	     * top of the loop.  Note that we convert it back at
	     * the bottom of the loop so we're consistent entering
	     * and leaving.
	     */
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	for (;
	     blockCount > 1;
	     pImage = (UInt8 *)(((UInt32)pImage ^ 3) + interBlockAddrOffset),
		blockCount--, pRunLengths += 8)
	{
	    pImage = (UInt8 *)((UInt32)pImage ^ 3); /* convert to little endian
						 * address just for the loop */
#else
	for (;
	     blockCount > 1;
	     pImage += interBlockAddrOffset, blockCount--, pRunLengths += 8)
	{
#endif
		/* rotate the top four bytes from the block */
	    iWord0 = pRotationTable[*pImage];	/* byte 1 */
	    pImage += bpl;
	    iWord0 += iWord0;
	    
	    iWord0 |= pRotationTable[*pImage];	/* byte 2 */
	    pImage += bpl;
	    iWord0 += iWord0;
	    
	    iWord0 |= pRotationTable[*pImage];	/* byte 3 */
	    pImage += bpl;
	    iWord0 += iWord0;
	    
	    iWord0 |= pRotationTable[*pImage];	/* byte 4 */
	    pImage += bpl;

		/* iWord0 is now filled with 8 nibbles.
		 * Repeat the process for iWord1.
		 */
	    iWord1 = pRotationTable[*pImage];	/* byte 5 */
	    pImage += bpl;
	    iWord1 += iWord1;
	    
	    iWord1 |= pRotationTable[*pImage];	/* byte 6 */
	    pImage += bpl;
	    iWord1 += iWord1;
	    
	    iWord1 |= pRotationTable[*pImage];	/* byte 7 */
	    pImage += bpl;
	    iWord1 += iWord1;
	    
	    iWord1 |= pRotationTable[*pImage];	/* byte 8 */

		/* Let's try to save some time.  There's lots of
		 * white space in images.  If all the bits are
		 * 0's, then increment the run length pointers
		 * appropriately and just move on.  Ditto if they're
		 * both all 1's.
		 */
	    if (((iWord0 | iWord1) == 0) ||
		((iWord0 & iWord1) == 0xFFFFFFFF))
	    {
		    /* fetch either 7 or 0 and increment the counts by that
		     * amount
		     */
		byteRunLength = pRunTable[iWord0 & cVRunAddrMask];

		if (byteRunLength != 0)
		{
		    pRunLengths[0] += byteRunLength;
		    pRunLengths[1] += byteRunLength;
		    pRunLengths[2] += byteRunLength;
		    pRunLengths[3] += byteRunLength;
		    pRunLengths[4] += byteRunLength;
		    pRunLengths[5] += byteRunLength;
		    pRunLengths[6] += byteRunLength;
		    pRunLengths[7] += byteRunLength;
		}
	    }
	    else	/* runs start and end somewhere in here */
	    {
		    /* We now have bits 7:4 of each rotated byte in
		     * iWord0 and bits 3:0 of each rotated byte in
		     * iWord1.  Compose 4 bytes and look them up
		     * in the run length tables as we do in
		     * horizontal rotation.
		     */
		rotBytes = (iWord0 & 0xF0F0F0F0) |
			    ((iWord1 >> 4) & 0x0F0F0F0F);
    
		    /* look at first byte, first set of nibbles */
		trialByte = (rotBytes << cVRunAddrPos) & cVRunAddrMask;
		byteRunLength = pRunTable[trialByte];
    
		if (byteRunLength < 0)	/* we've found the end of a run */
		{
		    if (pRunLengths[0] > 0)	/* this is not the first run */
			pHisto[pRunLengths[0] + (~byteRunLength)]++;
    
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
			/* get length of run on left extending from this byte */
		    pRunLengths[0] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else
		if (byteRunLength == 0) /* there was no run, does one start? */
		{
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
		    pRunLengths[0] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else	/* the run just extends through this byte */
		    pRunLengths[0] += 7;
    
    
    
		    /* look at second byte, first set of nibbles */
		trialByte = (rotBytes >> (8 - cVRunAddrPos)) & cVRunAddrMask;
		byteRunLength = pRunTable[trialByte];
    
		if (byteRunLength < 0)	/* we've found the end of a run */
		{
		    if (pRunLengths[1] > 0)	/* this is not the first run */
			pHisto[pRunLengths[1] + (~byteRunLength)]++;
    
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
			/* get length of run on left extending from this byte */
		    pRunLengths[1] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else
		if (byteRunLength == 0) /* there was no run, does one start? */
		{
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
		    pRunLengths[1] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else	/* the run just extends through this byte */
		    pRunLengths[1] += 7;
    
    
		    /* look at third byte, first set of nibbles */
		trialByte = (rotBytes >> (16 - cVRunAddrPos)) & cVRunAddrMask;
		byteRunLength = pRunTable[trialByte];
    
		if (byteRunLength < 0)	/* we've found the end of a run */
		{
		    if (pRunLengths[2] > 0)	/* this is not the first run */
			pHisto[pRunLengths[2] + (~byteRunLength)]++;
    
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
			/* get length of run on left extending from this byte */
		    pRunLengths[2] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else
		if (byteRunLength == 0) /* there was no run, does one start? */
		{
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
		    pRunLengths[2] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else	/* the run just extends through this byte */
		    pRunLengths[2] += 7;
    
    
		    /* look at fourth byte, first set of nibbles */
		trialByte = (rotBytes >> (24 - cVRunAddrPos)) & cVRunAddrMask;
		byteRunLength = pRunTable[trialByte];
    
		if (byteRunLength < 0)	/* we've found the end of a run */
		{
		    if (pRunLengths[3] > 0)	/* this is not the first run */
			pHisto[pRunLengths[3] + (~byteRunLength)]++;
    
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
			/* get length of run on left extending from this byte */
		    pRunLengths[3] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else
		if (byteRunLength == 0) /* there was no run, does one start? */
		{
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
		    pRunLengths[3] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else	/* the run just extends through this byte */
		    pRunLengths[3] += 7;
    
    
    
		    /* Look at the other set of nibbles */
		rotBytes = ((iWord0 << 4) & 0xF0F0F0F0) |
			    (iWord1 & 0x0F0F0F0F);
		
		    /* look at first byte, second set of nibbles */
		trialByte = (rotBytes << cVRunAddrPos) & cVRunAddrMask;
		byteRunLength = pRunTable[trialByte];
    
		if (byteRunLength < 0)	/* we've found the end of a run */
		{
		    if (pRunLengths[4] > 0)	/* this is not the first run */
			pHisto[pRunLengths[4] + (~byteRunLength)]++;
    
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
			/* get length of run on left extending from this byte */
		    pRunLengths[4] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else
		if (byteRunLength == 0) /* there was no run, does one start? */
		{
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
		    pRunLengths[4] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else	/* the run just extends through this byte */
		    pRunLengths[4] += 7;
    
    
		    /* look at second byte, second set of nibbles */
		trialByte = (rotBytes >> (8 - cVRunAddrPos)) & cVRunAddrMask;
		byteRunLength = pRunTable[trialByte];
    
		if (byteRunLength < 0)	/* we've found the end of a run */
		{
		    if (pRunLengths[5] > 0)	/* this is not the first run */
			pHisto[pRunLengths[5] + (~byteRunLength)]++;
    
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
			/* get length of run on left extending from this byte */
		    pRunLengths[5] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else
		if (byteRunLength == 0) /* there was no run, does one start? */
		{
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
		    pRunLengths[5] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else	/* the run just extends through this byte */
		    pRunLengths[5] += 7;
    
    
		    /* look at third byte, second set of nibbles */
		trialByte = (rotBytes >> (16 - cVRunAddrPos)) & cVRunAddrMask;
		byteRunLength = pRunTable[trialByte];
    
		if (byteRunLength < 0)	/* we've found the end of a run */
		{
		    if (pRunLengths[6] > 0)	/* this is not the first run */
			pHisto[pRunLengths[6] + (~byteRunLength)]++;
    
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
			/* get length of run on left extending from this byte */
		    pRunLengths[6] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else
		if (byteRunLength == 0) /* there was no run, does one start? */
		{
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
		    pRunLengths[6] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else	/* the run just extends through this byte */
		    pRunLengths[6] += 7;
    
    
		    /* look at fourth byte, second set of nibbles */
		trialByte = (rotBytes >> (24 - cVRunAddrPos)) & cVRunAddrMask;
		byteRunLength = pRunTable[trialByte];
    
		if (byteRunLength < 0)	/* we've found the end of a run */
		{
		    if (pRunLengths[7] > 0)	/* this is not the first run */
			pHisto[pRunLengths[7] + (~byteRunLength)]++;
    
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
			/* get length of run on left extending from this byte */
		    pRunLengths[7] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else
		if (byteRunLength == 0) /* there was no run, does one start? */
		{
			/* there might be other runs inside this byte */
		    pByteHisto[trialByte]++;
    
		    pRunLengths[7] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
		}
		else	/* the run just extends through this byte */
		    pRunLengths[7] += 7;
	    }
	}	/* end of processing one 8x8 block of bits */


	    /* if we haven't already done it, do the rightmost block */
	if (blockCount == 1)
	{
	    ip_vertRunHistoBlock (pIPState,
				  pImage,
				  bpl,
				  rightByteAndMask,
				  rightByteOrMask,
				  0xFFFFFFFF,		/* blockAndMask */
				  0,			/* blockOrMask */
				  pHisto,
				  pByteHisto,
				  pRotationTable,
				  pRunTable,
				  pRunLengths);
	}

    }	/* end of processing some row other than the bottom one */

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_vertRunHistoBlock()  
 *		Computes histogram of lengths of vertical runs of 0's, 1's
 *		or cycles in a single 8x8 block.
 *
 *	Input:
 *		pImage:			Pointer to word containing left end of
 *					rectangle.
 *		bpl:			image bytes/line
 *		andMask, orMask:	Used to set bits outside the rect
 *					to all 1's or all 0's.
 *		bottomBlockAndMask:	used to get rid of extra bits on the
 *					bottom of the rect.
 *		bottomBlockOrMask:	used to set extra bits on bottom of
 *					rect to 1's..
 *		pHisto:			Pointer to histogram that will hold
 *					lengths of runs that bridge bytes
 *		pByteHisto:		Pointer to histogram of byte values.
 *					These are used to reconstruct lengths 
 *					of runs contained purely inside bytes.
 *		pRotationTable:		Pointer to table used to turn
 *					bytes into rotated nibbles.
 *		ppRunLengths:		Pointer to pointer to array of
 *					current run lengths.
 *					There's one entry for every image column
 *					we're considering
 *
 *	Return code:
 *		ia_successful	it worked
 *
 **************************************************************************/
void  CDECL 
ip_vertRunHistoBlock (
    IPState    *pIPState,
    UInt8      *pImage,
    Int32	bpl,
    UInt8	andMask,
    UInt8	orMask,
    UInt32	blockAndMask,
    UInt32	blockOrMask,
    UInt32     *pHisto,
    UInt32     *pByteHisto,
    UInt32     *pRotationTable,
    Int8       *pRunTable,
    Int32      *pRunLengths)
{
    PROLOGUE;
    UInt32	 iWord0, iWord1;    /* rotated input words; each nibble holds
				     * rotated data from one of the 8
				     * columns being read. */
    UInt32	 rotBytes;	    /* holds rotated bytes, these are composed
				     * from rotated nibbles in iWord0
				     * and iWord1. */
    UInt32	 trialByte;	    /* used to hold each byte being examined */
    Int32	 byteRunLength;	    /* amount by which a run extends
				     * into a byte */

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* If this is a little endian machine, convert the address */
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    pImage = (UInt8 *)((UInt32) pImage ^ 3);
#endif
	/* rotate the top four bytes from the block */
    rotBytes = (*pImage & andMask) | orMask;
    iWord0 = pRotationTable[rotBytes];	/* byte 1 */
    pImage += bpl;
    iWord0 += iWord0;

    rotBytes = (*pImage & andMask) | orMask;
    iWord0 |= pRotationTable[rotBytes];	/* byte 2 */
    pImage += bpl;
    iWord0 += iWord0;

    rotBytes = (*pImage & andMask) | orMask;
    iWord0 |= pRotationTable[rotBytes];	/* byte 3 */
    pImage += bpl;
    iWord0 += iWord0;

    rotBytes = (*pImage & andMask) | orMask;
    iWord0 |= pRotationTable[rotBytes];	/* byte 4 */
    pImage += bpl;

	/* iWord0 is now filled.  Repeat the process for
	 * iWord1
	 */

    rotBytes = (*pImage & andMask) | orMask;
    iWord1 = pRotationTable[rotBytes];	/* byte 5 */
    pImage += bpl;
    iWord1 += iWord1;

    rotBytes = (*pImage & andMask) | orMask;
    iWord1 |= pRotationTable[rotBytes];	/* byte 6 */
    pImage += bpl;
    iWord1 += iWord1;

    rotBytes = (*pImage & andMask) | orMask;
    iWord1 |= pRotationTable[rotBytes];	/* byte 7 */
    pImage += bpl;
    iWord1 += iWord1;

    rotBytes = (*pImage & andMask) | orMask;
    iWord1 |= pRotationTable[rotBytes];	/* byte 8 */


	/* We now have bits 7:4 of each rotated byte in
	 * iWord0 and bits 3:0 of each rotated byte in
	 * iWord1.  Compose 4 bytes and look them up
	 * in the run length tables as we do in
	 * horizontal rotation.
	 */
    rotBytes = (iWord0 & 0xF0F0F0F0) |
		((iWord1 >> 4) & 0x0F0F0F0F);

    rotBytes &= blockAndMask;	/* delete extra bits */
    rotBytes |= blockOrMask;	/* add in the padding bits */

	/* go update the run lengths just like normal */

	/* look at first byte, first set of nibbles */
    trialByte = (rotBytes << cVRunAddrPos) & cVRunAddrMask;
    byteRunLength = pRunTable[trialByte];

    if (byteRunLength < 0)	/* we've found the end of a run */
    {
	if (pRunLengths[0] > 0)	/* this is not the first run */
	    pHisto[pRunLengths[0] + (~byteRunLength)]++;

	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	    /* get length of run on left extending from this byte */
	pRunLengths[0] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else
    if (byteRunLength == 0) /* there was no run, does one start? */
    {
	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	pRunLengths[0] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else	/* the run just extends through this byte */
	pRunLengths[0] += 7;



	/* look at second byte, first set of nibbles */
    trialByte = (rotBytes >> (8 - cVRunAddrPos)) & cVRunAddrMask;
    byteRunLength = pRunTable[trialByte];

    if (byteRunLength < 0)	/* we've found the end of a run */
    {
	if (pRunLengths[1] > 0)	/* this is not the first run */
	    pHisto[pRunLengths[1] + (~byteRunLength)]++;

	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	    /* get length of run on left extending from this byte */
	pRunLengths[1] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else
    if (byteRunLength == 0) /* there was no run, does one start? */
    {
	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	pRunLengths[1] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else	/* the run just extends through this byte */
	pRunLengths[1] += 7;


	/* look at third byte, first set of nibbles */
    trialByte = (rotBytes >> (16 - cVRunAddrPos)) & cVRunAddrMask;
    byteRunLength = pRunTable[trialByte];

    if (byteRunLength < 0)	/* we've found the end of a run */
    {
	if (pRunLengths[2] > 0)	/* this is not the first run */
	    pHisto[pRunLengths[2] + (~byteRunLength)]++;

	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	    /* get length of run on left extending from this byte */
	pRunLengths[2] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else
    if (byteRunLength == 0) /* there was no run, does one start? */
    {
	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	pRunLengths[2] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else	/* the run just extends through this byte */
	pRunLengths[2] += 7;


	/* look at fourth byte, first set of nibbles */
    trialByte = (rotBytes >> (24 - cVRunAddrPos)) & cVRunAddrMask;
    byteRunLength = pRunTable[trialByte];

    if (byteRunLength < 0)	/* we've found the end of a run */
    {
	if (pRunLengths[3] > 0)	/* this is not the first run */
	    pHisto[pRunLengths[3] + (~byteRunLength)]++;

	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	    /* get length of run on left extending from this byte */
	pRunLengths[3] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else
    if (byteRunLength == 0) /* there was no run, does one start? */
    {
	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	pRunLengths[3] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else	/* the run just extends through this byte */
	pRunLengths[3] += 7;



	/* Look at the other set of nibbles */
    rotBytes = ((iWord0 << 4) & 0xF0F0F0F0) |
		(iWord1 & 0x0F0F0F0F);

    rotBytes &= blockAndMask;	/* delete extra bits */
    rotBytes |= blockOrMask;	/* add in the padding bits */

	/* look at first byte, second set of nibbles */
    trialByte = (rotBytes << cVRunAddrPos) & cVRunAddrMask;
    byteRunLength = pRunTable[trialByte];

    if (byteRunLength < 0)	/* we've found the end of a run */
    {
	if (pRunLengths[4] > 0)	/* this is not the first run */
	    pHisto[pRunLengths[4] + (~byteRunLength)]++;

	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	    /* get length of run on left extending from this byte */
	pRunLengths[4] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else
    if (byteRunLength == 0) /* there was no run, does one start? */
    {
	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	pRunLengths[4] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else	/* the run just extends through this byte */
	pRunLengths[4] += 7;


	/* look at second byte, second set of nibbles */
    trialByte = (rotBytes >> (8 - cVRunAddrPos)) & cVRunAddrMask;
    byteRunLength = pRunTable[trialByte];

    if (byteRunLength < 0)	/* we've found the end of a run */
    {
	if (pRunLengths[5] > 0)	/* this is not the first run */
	    pHisto[pRunLengths[5] + (~byteRunLength)]++;

	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	    /* get length of run on left extending from this byte */
	pRunLengths[5] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else
    if (byteRunLength == 0) /* there was no run, does one start? */
    {
	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	pRunLengths[5] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else	/* the run just extends through this byte */
	pRunLengths[5] += 7;


	/* look at third byte, second set of nibbles */
    trialByte = (rotBytes >> (16 - cVRunAddrPos)) & cVRunAddrMask;
    byteRunLength = pRunTable[trialByte];

    if (byteRunLength < 0)	/* we've found the end of a run */
    {
	if (pRunLengths[6] > 0)	/* this is not the first run */
	    pHisto[pRunLengths[6] + (~byteRunLength)]++;

	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	    /* get length of run on left extending from this byte */
	pRunLengths[6] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else
    if (byteRunLength == 0) /* there was no run, does one start? */
    {
	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	pRunLengths[6] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else	/* the run just extends through this byte */
	pRunLengths[6] += 7;


	/* look at fourth byte, second set of nibbles */
    trialByte = (rotBytes >> (24 - cVRunAddrPos)) & cVRunAddrMask;
    byteRunLength = pRunTable[trialByte];

    if (byteRunLength < 0)	/* we've found the end of a run */
    {
	if (pRunLengths[7] > 0)	/* this is not the first run */
	    pHisto[pRunLengths[7] + (~byteRunLength)]++;

	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	    /* get length of run on left extending from this byte */
	pRunLengths[7] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else
    if (byteRunLength == 0) /* there was no run, does one start? */
    {
	    /* there might be other runs inside this byte */
	pByteHisto[trialByte]++;

	pRunLengths[7] = pRunTable[trialByte+cVRunOnLeftAddrOffset];
    }
    else	/* the run just extends through this byte */
	pRunLengths[7] += 7;

	/* that's it for this block */
    return;
}


/*

$Log: mirnhst2.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.7  1998/01/13 23:53:18  ddavies
mirnhst2.c -> /product/ipcore/ipshared/src/RCS/mirnhst2.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1995/12/13 02:51:37  ejaquith
mirnhst2.c -> /product/ipcore/ipshared/src/RCS/mirnhst2.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/10/13  23:53:08  dferrell
 * mirnhst2.c -> /product/ipcore/ipshared/src/RCS/mirnhst2.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.4  1995/01/11  22:33:22  dferrell
 * mirnhst2.c -> /product/ipcore/ipshared/src/RCS/mirnhst2.c,v
 *
 * Changed include of shrip.pub and shrip.prv to shrcc.pub and shrcc.prv,
 * respectively.
 *
 * Revision 1.3  1995/01/05  04:06:35  ddavies
 * mirnhst2.c -> /product/ipcore/ipshared/src/RCS/mirnhst2.c,v
 *
 * The run length tables now contain bytes instead of words to save space.
 * This saved about 4.5KB in this file.  The macro
 * CONST_LONG_ACCESS has been removed when accessing the run length tables
 * since we get bytes now (and expect the processor to sign-extend them).
 * The ULONG_ACCESS macro has been removed from updates to the byte
 * histograms since we now expect the compiler to shift the table index
 * instead of having it preshifted (we used to use the same index for
 * the run length tables and the byte histograms).  The net effect of
 * reducing the table size and adding the shifts seems to be negligible
 * in performance.
 *
 * Revision 1.2  1994/07/01  05:09:05  ddavies
 * ddavies on Thu Jun 30 22:08:41 PDT 1994
 *
 * mirnhst2.c -> /product/ipcore/ipshared/src/RCS/mirnhst2.c,v
 *
 * Fixed bug in little-endian addressing.  Had forgotten to XOR 3 with the
 * byte addresses.
 *

*/


