/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  mirottab.c--includes
 *
 *		ip_makeRotTable()	makes a table used by C
 *					implementations of 90 and 270
 *					degree rotation.
 */


#include <stddef.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

    /* prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mirottab.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* PRIVATE */
/**************************************************************************
 *  ip_makeRotTable():  Makes the table we used to do 90 degree rotations.
 *
 *			This is a 256 x 32bit table.  For 1 bit/pixel
 *			images, it simply spreads the
 *			bits [7:0] in each byte into bit positions
 *			[28, 24, 20, 16, 12, 8, 4, 0].
 *			For 2 bit/pixel images, is spreads bits [7:0]
 *			into 4 clumps - [25,24, 17,16, 9,8, 1,0]
 *
 *	rotationTable:	pointer to returned rotation table.  This proc
 *			allocates the table and returns the address in
 *			this pointer.  It's up to the caller to free
 *			the table.
 *
 *	Return code:
 *	 ia_nomem		no memory available for table
 *	 ia_depthNotSupported	cannot make tables for rotation of
 *				images with this depth
 *	 ia_successful		good completion
 *
 *
 **************************************************************************/
Int32  CDECL 
ip_makeRotTable (
    IPState    *pIPState,
    UInt32    **pRotationTable,
    Int32	depth)
{
    PROLOGUE;
    UInt32	byte;
    UInt32 	expByte;
    UInt32     *localTab;

	/* if depth == 4, use the table making routine that's used
	 * in all implementations
	 */
    if (depth == 4)
	return ip_make4bppRotTable (pIPState, pRotationTable);
    else
    if (depth == 8)
    {
	*pRotationTable = NULL;
	NORMAL_RETURN;
    }
    else
	/* can we make tables for images of this depth? */
    if ((depth != 1) && (depth != 2))
        return ia_depthNotSupported;

	/* At this point, we know the image depth is 1 or 2. */

	/* allocate the table */
    *pRotationTable = (UInt32 *) CALLOC (256, sizeof(UInt32));

    if (*pRotationTable == NULL)
	return ia_nomem;

    localTab = (UInt32 *) *pRotationTable;

	/* Set each element of the table with the spread bits of its
	 * index. */
    switch (depth) {
    case 1: /* depth = 1 bit/pixel */
	for (byte = 0; byte < 256; byte++)
	{
	    expByte =  ((byte & 0x80) << (28-7) ) |	/* bit 7 */
		       ((byte & 0x40) << (24-6) ) |	/* bit 6 */
		       ((byte & 0x20) << (20-5) ) |	/* bit 5 */
		       ((byte & 0x10) << (16-4) ) |	/* bit 4 */
		       ((byte & 0x08) << (12-3) ) |	/* bit 3 */
		       ((byte & 0x04) << ( 8-2) ) |	/* bit 2 */
		       ((byte & 0x02) << ( 4-1) ) |	/* bit 1 */
		       ((byte & 0x01) << ( 0-0) );	/* bit 0 */
	    localTab[byte] = expByte;
	}
	break;

    default: /* depth = 2 bits/pixel */
	for (byte = 0; byte < 256; byte++)
	{
	    expByte =  ((byte & 0xC0) << (25-7) ) |	/* bits 7,6 */
		       ((byte & 0x30) << (17-5) ) |	/* bits 5,4 */
		       ((byte & 0x0C) << ( 9-3) ) |	/* bits 3,2 */
		       ((byte & 0x03));			/* bits 1,0 */
	    localTab[byte] = expByte;
	}
	break;
    }
    NORMAL_RETURN;
}

/*

 $Log: mirottab.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:17  jlin
 Imported sources

 Revision 1.3  1998/01/13 23:53:19  ddavies
 mirottab.c -> /product/ipcore/ipshared/src/RCS/mirottab.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.2  1995/12/13 02:51:44  ejaquith
 mirottab.c -> /product/ipcore/ipshared/src/RCS/mirottab.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1994/07/01  04:29:54  ddavies
 * Initial revision
 *

 */

