/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miscgr41.c: machine independent routine for scale to 4bpp gray,
 *		1.5x reduction
 *
 *          1.5x scale to gray reduction:
 *                  void	ip_scaleToGray4bpp1_5xLine()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: miscgr41.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* This is code that does 1.5x reduction from binary to gray image, producing
 * a single output line.  We repeat the comments found in i_scgr41.c.
 ***************************************************************************.
 * The basic method is to use each 6-bit block of input bits as an index
 * into one of two tables.  The table entries for two vertically adjacent
 * rows are summed, producing the gray output pixels.
 *
 * Unfortunately, the details are a little harder.  We want to repeatedly
 * map 9 binary pixels (3x3) in the input to 4 gray pixels (2x2) in the output.
 * This is equivalent to mulitplying the width and height by 2/3, or 1/1.5.
 * Here's a picture of 9 binary input pixels:
 *
 * 	0  1  2		0    1
 *	3  4  5     =>
 *	6  7  8		2    3
 *
 * The four gray output pixels would ideally be calculated as follows:
 * 
 *	Out0 = (In0 + In1/2 + In3/2 + In4/4) * 15/2.25
 *	Out1 = (In2 + In1/2 + In5/2 + In4/4) * 15/2.25
 *	Out2 = (In6 + In7/2 + In3/2 + In4/4) * 15/2.25
 *	Out3 = (In8 + In7/2 + In5/2 + In4/4) * 15/2.25
 *
 * This mapping is perfectly linear and covers the entire range of outputs.
 * How do we approximate it efficiently?
 *
 * We use two tables, one for the first and third input rows, another for the
 * second input row.  Each table is addressed with 6 bits, so we will map
 * two groups of 3 pixels into 2 gray pixels.  Let's just look at the
 * weights used to translate one set of 3 pixels into 2 gray pixels.  In
 * fact, lets just look at the weights used to calculate pixel Out0 above;
 * the other weights in the table are calculated symmetrically.
 *
 * If we can meet these goals, we'll be happy:
 *	1. When none of the pixels are active, the result is 0.
 *	2. When all of the pixels are active, the result is 15.
 *	3. There is no need to check for rollover.
 *	4. The mapping from number of input pixels to output value is
 *	   roughly linear (within roundoff error) and monotonic.
 *	5. The mapping is only a function of the number of pixels active,
 *	   not their position.  For example, pixels In1 and In3 should
 *	   count equally to the sum.
 *	6. Several sums can be computed in parallel.
 *	7. Any fixup effort required at the end of the summing process,
 *	   like to add in carries from fractional parts, is minimized.
 *
 * One way to look at the problem is that we'd like to weight each pixel
 * by (4/9) * 15.  That's hard to calculate, but 4/8 is trivial.  However, then
 * the sum of the weights is too big.  To compensate, we drop the LS bit
 * of the 15, making it (4/8) * 14.  This happens to work as seen below
 * The weights we'll use:
 *  Pixel	weighted sum	In binary
 *   In0		7.0	0111.00
 *   In1		3.5	0011.10
 *   In3		3.5	0011.10
 *   In4	        1.75	0001.11
 *
 * Another set of mappings that doesn't use any fraction bits is:
 *  Pixel	weighted sum	In binary
 *   In0		7	0111
 *   In1		3	0011
 *   In3		3	0011
 *   In4	        2	0010
 *
 *
 * In the table below, we calculate the perfect sum of pixel values, the
 * sum found using the first set of weights then truncating and the sum
 * found using the second set of weights (no truncation required):
 *
 * In0 In1 In3 In4	Ideal	Sum1	Sum2
 *   0   0   0   0	0.00	0	0
 *   0   0   0   1	1.67	1	2
 *   0   0   1   0	3.33	3	3
 *   0   0   1   1	5.00	5	5
 *   0   1   0   0	3.33	3	3
 *   0   1   0   1	5.00	5	5
 *   0   1   1   0	6.67	7	6
 *   0   1   1   1	8.34	8	8
 *   1   0   0   0	6.67	7	7
 *   1   0   0   1	8.34	8	9
 *   1   0   1   0	10.00	10	10
 *   1   0   1   1	11.67	12	12
 *   1   1   0   0	10.00	10	10
 *   1   1   0   1	11.67	12	12
 *   1   1   1   0	13.33	14	13
 *   1   1   1   1	15.00	15	15
 *
 * The second mapping seems fine to me and doesn't involve any fraction
 * bits, so is faster to compute.  We'll use it.
 *
 * How does the algorithm work?
 *
 *	1. Read an input image 6-bit block.  The byte is divided into 2
 *	   triples of input pixels.
 *	2. Use it to index an array of 32-bit words.  The lower half of
 *	   each word contains four 4-bit numbers which correspond to the
 *	   integer parts of summed weights associated with each pair of
 *	   input pixels.
 *	3. Read the input byte directly below the one read first.
 *	4. Look it up in the the other half of the table since it
 *	   corresponds to the pixels that are weighted by 1/2 and 1/4
 *	   instead of 1.0 and 1/2.
 *	5. Add the two table-lookup entries.  This produces the summed
 *	   weights for 1 pair of output pixels.
 *
 * Now things start getting funny.  We produce 4 output pixels each time
 * we do steps 1 through 5.  We want to write 8 output pixels at a time.
 * On the other hand, we need groups of 6 input pixels.  32 is not a
 * multiple of 6.  So we perform steps 1..5 in the following sequence:
 *	Read bits [7:2] from byte 0, produce 4 pixels
 *	Combine byte0.[1:0] and byte1.[7:4], produce 4 more, write 1 word
 *	Combine byte1.[3:0] and byte2.[7:6}, produce 4
 *	Use byte2.[5:0] to produce 4 more, write 1 word.
 *
 * We use this sequence to produce a line of outputs.  We then advance
 * down *1* line in both the input and output images; and negate the
 * offset used to get to the next input line when performing step 3
 * above.  This allows us to use the second line of input image
 * twice.  After producing another line of output, we skip a line of input
 * entirely.
 *
 * As a matter of software engineering, this file contains the piece of code
 * that accumulates some image and calls the lower level routine that
 * produces one output line.  When we switch between banded images and
 * whole images, we can just use different routines here while using the
 * same real routine below.
 *
 * The first (only?) user of this function is EasyScan.  They will want to
 * apply the function to small images (extracted from large images) many
 * times a second.  In this world, it's a bad idea to create and destroy
 * the table every time we're called.  A separate procedure,
 * i_createScaleToGrayTab4bpp2x is called by the user to create the table
 * beforehand.  A pointer to this table is passed down.  If the pointer is
 * NULL, the table is created, used and destroyed before returning.
 */

/*********************************************************************
 ************       Scale To 4 bpp Gray, 1.5x Reduction    *************
 *********************************************************************/

/* PRIVATE */
/**************************************************************************
 *  ip_scaleToGray4bpp1_5xLine()  
 *		Each 3x3 square of the source binary image is mapped
 *		to a 2x2 square of 4-bit output pixels.  The values
 *		of the output pixels are a (roughly) linear function of
 *		the number of bits turned on in the input square.
 *
 *	Input:
 *		pD =		pointer to upper left corner of dest image
 *				(includes frame)
 *		wD =		width in pixels of dest (including frame)
 *		pS =		pointer to upper left corner of source image
 *				(includes frame)
 *		sBpl =		source image bytes/line
 *		pScaleTable =	ptr to table used to do scaling
 *		rightMask =	used to set unused pixel on the right
 *				end of the line to zero.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_scaleToGray4bpp1_5xLine (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    UInt8      *pS,
    Int32	sBpl,
    UInt32     *pScaleTable,
    UInt32	rightMask)
{
    PROLOGUE;
    UInt32	oWord;			/* the output word being produced */
	/* input bytes, two on this line and two on next line */
    UInt8	iByte00, iByte01, iByteN0, iByteN1;	

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* prepare to loop over the words in the output image.
	 * The images is always left aligned.  The last word on the 
	 * right is masked with rightMask.  We start by reading the
	 * first 6 pixels of the row.
	 * Each output pixel is composed of inputs from the major and
	 * minor lines as described above.  Each 6-bit block of inputs
	 * provides 4 output pixels, so we need two blocks of input
	 * pixels for each output word.
	 */
    iByte00 = UCHAR_ACCESS(&pS[0]);
    iByteN0 = UCHAR_ACCESS(&pS[sBpl]);
    oWord = pScaleTable[iByte00 >> 2];
    oWord += pScaleTable[(iByteN0 >> 2) + c1_5xTableOffset];
    oWord <<= 16;	/* shift these pixels into position */

	/* repeat the process for the next pair of bytes to the right */
    pS++;
    iByte01 = UCHAR_ACCESS(&pS[0]);
    iByteN1 = UCHAR_ACCESS(&pS[sBpl]);
    oWord += pScaleTable[((iByte00 << 4) & 0x30) + (iByte01 >> 4)];
    oWord += pScaleTable[((iByteN0 << 4) & 0x30) + (iByteN1 >> 4) +
			 c1_5xTableOffset];

	/* Now do the real loop.  We start by writing the word we
	 * already have so the last time through the loop we can
	 * fall out the bottom with the last word.  We'll mask it
	 * before writing it.  The loop has enough sections so that
	 * when we get back up to the top, iByte00 and iByteN0 are
	 * empty, iByte01 and iByteN1 each have 4 useful bits and
	 * oWord is full of pixels.
	 */
    for (; wD > 8; wD -= 16)
    {
	ULONG_ACCESS(pD) = oWord;
	pD += 4;

	    /* get the next bytes so we can compute another pair of
	     * 6-bit blocks */
	pS++;
	iByte00 = UCHAR_ACCESS(&pS[0]);
	iByteN0 = UCHAR_ACCESS(&pS[sBpl]);

	oWord  = pScaleTable[((iByte01 << 2) & 0x3C) + (iByte00 >> 6)];
	oWord += pScaleTable[((iByteN1 << 2) & 0x3C) + (iByteN0 >> 6) +
			      c1_5xTableOffset];
	oWord <<= 16;	/* shift these pixels into position */

	    /* Take the last 6 input bits out of iByte00 and iByteN0.
	     * After this step, all the iByteXX vars will be empty.
	     */
	oWord += pScaleTable[iByte00 & 0x3F];
	oWord += pScaleTable[(iByteN0 & 0x3F) + c1_5xTableOffset];

	    /* if that's the last word in the line, quit now.  Otherwise,
	     * write it to the output and continue.
	     */
	if (wD <= 16)
	    break;

	ULONG_ACCESS(pD) = oWord;
	pD += 4;

	pS++;
	iByte00 = UCHAR_ACCESS(&pS[0]);
	iByteN0 = UCHAR_ACCESS(&pS[sBpl]);
	oWord  = pScaleTable[iByte00 >> 2];
	oWord += pScaleTable[(iByteN0 >> 2) + c1_5xTableOffset];
	oWord <<= 16;	/* shift these pixels into position */
    
	    /* repeat the process for the next pair of bytes to the right */
	pS++;
	iByte01 = UCHAR_ACCESS(&pS[0]);
	iByteN1 = UCHAR_ACCESS(&pS[sBpl]);
	oWord += pScaleTable[((iByte00 << 4) & 0x30) + (iByte01 >> 4)];
	oWord += pScaleTable[((iByteN0 << 4) & 0x30) + (iByteN1 >> 4) +
			     c1_5xTableOffset];
    }

	/* mask the last word and write it */
    ULONG_ACCESS(pD) = oWord & rightMask;
}

/*

$Log: miscgr41.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.6  1998/01/13 23:53:22  ddavies
miscgr41.c -> /product/ipcore/ipshared/src/RCS/miscgr41.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1995/12/13 02:51:55  ejaquith
miscgr41.c -> /product/ipcore/ipshared/src/RCS/miscgr41.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/02/04  20:18:45  dferrell
 * miscgr41.c -> /product/ipcore/ipshared/src/RCS/miscgr41.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:25:14  ddavies
 * miscgr41.c -> /product/ipcore/ipshared/src/RCS/miscgr41.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:50:25  ddavies
 * miscgr41.c -> /product/ipcore/ipshared/src/RCS/miscgr41.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


