/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miscgr48.c: machine independent routine for scale to 4bpp gray, 8x reduction
 *
 *          8x scale to gray reduction:
 *                  void	ip_scaleToGray4bpp8xLine()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: miscgr48.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* This is code that does 8x reduction from binary to gray image, producing
 * a single output line.  We repeat the comments found in i_scgr48.c.
 ***************************************************************************.
 * The basic method is to use each byte of input as an index into a table.
 * The table entries for four vertically adjacent rows are summed, producing
 * the gray output pixels.
 *
 * Unfortunately, the details are a little harder.  We want to repeatedly
 * map 64 binary pixels in the input to 1 gray pixel in the output.
 * The group of 64 binary pixels can take one of *65* values: 0, 1, 2, .. 
 * 64 pixels turned on.  The output pixel has a range from 0 to 15.  To the
 * extent possible, we want the input range to linearly to the output range.
 * The mapping should depend only on the number of input 1's, not on their
 * position within the 8x8 square.  A nice mapping is:
 *
 * Number of input pixels	Output value
 *		0		0
 *		1		15/64
 *		2		2 * 15/64
 *		...		...
 *		63		63 * 15/64
 *		64		64 * 15/64
 *
 * This mapping is linear and covers the entire range of output.
 *
 * For the sake of efficiency, we would like to process as many pixels at
 * once as possible.  Table lookup is a good way to space out the pixels and
 * do some computing.  On the other hand, we don't like to make huge tables,
 * so we like to restrict the number of pixels used to index the tables.
 * We often use 8 as a nice table index size.
 *
 * After thinking about several schemes, we settled on one that weights each
 * 1 in the input by 15/64.  We use a table to compute the summed weights.
 * It would be conceptually simple to keep each pixel as a fixed point, 16 bit
 * binary number.  However, after computing the fixed point sum, we would need
 * to discard the 8 fraction bits.  This is pretty easy, so we'll just do it.
 *
 * How does the algorithm compute a single 32-bit output word?
 *
 *	1. Read an input image byte.
 *	2. Use it to index an array of 32-bit words.  The lower halfword of
 *	   each word contains a fixed point number in the form
 *	   xxxx.xxxxxxxx.  Each number equals 15/64 times the
 *	   number of active pixels (0's in our case) in the corresponding
 *	   nibble.  The result is adjusted so the integer byte is byte
 *	   aligned.
 *	3. Read the input byte directly below the one read first.
 *	4. Look it up in the same table.
 *	5. Add the two table-lookup entries.
 *	6. Repeat steps 4 and 5 six more times.  This produces the weights
 *	   associated with 8x8 squares of pixels.
 *	7. Mask out the fractional bits. This puts the sum in byte 2 of
 *	   a 32 bit word.
 *	8. Repeat steps 1 through 7 for the next 7 vertical stacks of 7 bytes
 *	   on the right.
 *	   Shift the existing sum of inputs left by 4 and OR in the new
 *	   set at each stage before the last two stages.  In those two,
 *	   shift the sum right into position before combining it..  Write
 *	   the output word to the image.
 *
 * As a matter of software engineering, this file contains the piece of code
 * that accumulates some image and calls the lower level routine that
 * produces one output line.  When we switch between banded images and
 * whole images, we can just use different routines here while using the
 * same real routine below.
 *
 * The first (only?) user of this function is EasyScan.  They will want to
 * apply the function to small images (extracted from large images) many
 * times a second.  In this world, it's a bad idea to create and destroy
 * the table every time we're called.  A separate procedure,
 * i_createScaleToGrayTab4bpp8x is called by the user to create the table
 * beforehand.  A pointer to this table is passed down.  If the pointer is
 * NULL, the table is created, used and destroyed before returning.
 */

/*********************************************************************
 ************       Scale To 4 bpp Gray, 8x Reduction    *************
 *********************************************************************/

/* PRIVATE */
/**************************************************************************
 *  ip_scaleToGray4bpp8xLine()  
 *		Each 8x8 square of the source binary image is mapped
 *		to a 4-bit output pixel.  The value of the output
 *		pixel is a (roughly) linear function of the number of
 *		bits turned on in the 8x8 input square.
 *
 *	Input:
 *		pD =		pointer to upper left corner of dest image
 *				(includes frame)
 *		wD =		width in pixels of dest (including frame)
 *		pS =		pointer to upper left corner of source image
 *				(includes frame)
 *		sBpl =		source image bytes/line
 *		pScaleTable =	ptr to table used to do scaling
 *		rightMask =	used to set unused pixel on the right
 *				end of the line to zero.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_scaleToGray4bpp8xLine (
    IPState    *pIPState,
    UInt8      *pD,
    Int32	wD,
    UInt8      *pS,
    Int32	sBpl,
    UInt32     *pScaleTable,
    UInt32	rightMask)
{
    PROLOGUE;
    UInt32	oWord1, oWord2;	/* the output words being produced */
    Int32	tmpsBpl;		/* used to get at source image */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* prepare to loop over the words in the output image.
	 * The images is always left aligned.  The last word on the 
	 * right is masked with rightMask.  We add the two table 
	 * entries corresponding to the two bytes read (one above
	 * the other), then add in the carries from the sums of the
	 * fractional parts of the weights.
	 */
    oWord1 = pScaleTable[UCHAR_ACCESS(&pS[0])];
    for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
    {
	oWord1 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
    }
	/* delete the fraction */
    oWord1 &= 0xF00;

	/* get nibble number 1 */
    pS++;
    oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
    for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
    {
	oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
    }
	/* delete the fraction */
    oWord2 &= 0xF00;
	/* and insert the nibble into the output word */
    oWord1 = (oWord1 << 4) | oWord2;

	/* get nibble number 2 */
    pS++;
    oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
    for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
    {
	oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
    }
	/* delete the fraction */
    oWord2 &= 0xF00;
	/* and insert the nibble into the output word */
    oWord1 = (oWord1 << 4) | oWord2;

	/* get nibble number 3 */
    pS++;
    oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
    for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
    {
	oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
    }
	/* delete the fraction */
    oWord2 &= 0xF00;
	/* and insert the nibble into the output word */
    oWord1 = (oWord1 << 4) | oWord2;

	/* get nibble number 4 */
    pS++;
    oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
    for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
    {
	oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
    }
	/* delete the fraction */
    oWord2 &= 0xF00;
	/* and insert the nibble into the output word */
    oWord1 = (oWord1 << 4) | oWord2;

	/* get nibble number 5 */
    pS++;
    oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
    for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
    {
	oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
    }
	/* delete the fraction */
    oWord2 &= 0xF00;
	/* and insert the nibble into the output word */
    oWord1 = (oWord1 << 4) | oWord2;

	/* get nibble number 6 */
    pS++;
    oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
    for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
    {
	oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
    }
	/* delete the fraction */
    oWord2 &= 0xF00;
	/* and insert the nibble into the output word */
    oWord1 = oWord1 | (oWord2 >> 4);

	/* get nibble number 7 */
    pS++;
    oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
    for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
    {
	oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
    }
	/* and prepare the last nibble */
    oWord2 = oWord2 >> 8;



	/* Now do the real loop.  We start by writing the word we
	 * already have so the last time through the loop we can
	 * fall out the bottom with the last word.  We'll mask it
	 * before writing it.
	 */
    for (pS++; wD > 8; wD -= 8, pS++, pD += 4)
    {
	ULONG_ACCESS(pD) = oWord1 | oWord2;

	oWord1 = pScaleTable[UCHAR_ACCESS(&pS[0])];
	for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
	{
	    oWord1 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
	}
	    /* delete the fraction */
	oWord1 &= 0xF00;
    
	    /* get nibble number 1 */
	pS++;
	oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
	for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
	{
	    oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
	}
	    /* delete the fraction */
	oWord2 &= 0xF00;
	    /* and insert the nibble into the output word */
	oWord1 = (oWord1 << 4) | oWord2;
    
	    /* get nibble number 2 */
	pS++;
	oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
	for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
	{
	    oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
	}
	    /* delete the fraction */
	oWord2 &= 0xF00;
	    /* and insert the nibble into the output word */
	oWord1 = (oWord1 << 4) | oWord2;
    
	    /* get nibble number 3 */
	pS++;
	oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
	for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
	{
	    oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
	}
	    /* delete the fraction */
	oWord2 &= 0xF00;
	    /* and insert the nibble into the output word */
	oWord1 = (oWord1 << 4) | oWord2;
    
	    /* get nibble number 4 */
	pS++;
	oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
	for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
	{
	    oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
	}
	    /* delete the fraction */
	oWord2 &= 0xF00;
	    /* and insert the nibble into the output word */
	oWord1 = (oWord1 << 4) | oWord2;
    
	    /* get nibble number 5 */
	pS++;
	oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
	for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
	{
	    oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
	}
	    /* delete the fraction */
	oWord2 &= 0xF00;
	    /* and insert the nibble into the output word */
	oWord1 = (oWord1 << 4) | oWord2;
    
	    /* get nibble number 6 */
	pS++;
	oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
	for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
	{
	    oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
	}
	    /* delete the fraction */
	oWord2 &= 0xF00;
	    /* and insert the nibble into the output word */
	oWord1 = oWord1 | (oWord2 >> 4);
    
	    /* get nibble number 7 */
	pS++;
	oWord2 = pScaleTable[UCHAR_ACCESS(&pS[0])];
	for (tmpsBpl = 7*sBpl; tmpsBpl > 0; tmpsBpl -= sBpl)
	{
	    oWord2 += pScaleTable[UCHAR_ACCESS(&pS[tmpsBpl])];
	}
	    /* and prepare the last nibble */
	oWord2 = oWord2 >> 8;
    }

	/* mask the last word and write it */
    ULONG_ACCESS(pD) = (oWord1 | oWord2) & rightMask;
}

/*

$Log: miscgr48.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.6  1998/01/13 23:53:26  ddavies
miscgr48.c -> /product/ipcore/ipshared/src/RCS/miscgr48.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1995/12/13 02:52:14  ejaquith
miscgr48.c -> /product/ipcore/ipshared/src/RCS/miscgr48.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/02/04  20:19:25  dferrell
 * miscgr48.c -> /product/ipcore/ipshared/src/RCS/miscgr48.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:25:26  ddavies
 * miscgr48.c -> /product/ipcore/ipshared/src/RCS/miscgr48.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:50:51  ddavies
 * miscgr48.c -> /product/ipcore/ipshared/src/RCS/miscgr48.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


