/*************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  miskew.c:  contains machine independent implementations of lower
 *		level skew detection routines.
 *
 *	    Assembler-Level routines
 *		  Int32	      	ip_findDiffPixelsSum()
 *		  void	        ip_doEdgeSample()
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrskew.pub"
#include "shrskew.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miskew.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* The overall comments for skew detection are in i_skwfxp.c.  We repeat
 * them here because almost all of them pertain to the lower level code
 * anyway.  That's where the interesting work is done.
 */

/* This version of skew code is a variant of the original skew.c.  The
 * job of skew.c is to detect and correct image skew.  It depends on
 * images having a lot of horizontal stuff (lines, text).
 *
 * The original skew.c accepted binary images as input and produced them
 * as output.  We would like to improve the speed and quality of skew
 * detection.  We would also like to explore grey scale rotation for skew
 * correction.  We'll work on skew detection first.
 *
 * Skew.c`s detection algorithm had three major steps:
 *	1. Reduce image to a tractable size.
 *	2. Coarsely rotate the reduced image over a range of angles (say
 *	   -10 degrees to +10 degrees in 1 degree increments) to try to
 *	   find the position in which the skew is minimized.
 *	3. Use a bifurcation algorithm around the best angle found in the
 *	   sweep to fine tune the results.  Stop looking when the amount by
 *	   which the image will be rotated for the next step reaches a
 *	   value passed in as a parameter.
 *
 * The method used to obtain a "skew score" for a particular angle is to
 * vertically shear the reduced image by the angle and count the number of
 * pixels in each horizontal line.  The score equals the sum of the squares
 * of the differences in pixel counts across all adjacent lines.  In effect,
 * we're looking for skew angles at which we find the steepest horizontal
 * edges.  The squaring operation sharpens our sensitivity to steep edges.
 *
 * We'd like to explore improving this algorithm on two fronts.  First,
 * information is lost when the image is reduced.  This degrades our ability
 * to get good answers even with an ideal detection algorithm.  We'd like
 * to avoid reducing the image.
 *
 * Second, the process of continually shearing the image is relatively slow,
 * especially for a full scale image.  We'd like to avoid shearing.
 *
 * Why is the image sheared?  Because it's easier to count pixels
 * horizontally than diagonally.  However, if we could count diagonally,
 * we could dispense with the shearing, hence with the reduction.  We will
 * do so.
 *
 * We also note that it's not necessary to count every pixel in the image for
 * every potential shear angle.  Skew detection is a matter of statistics.
 * The more image samples we make, the better the statistics.  However, there
 * is probably some point of diminishing returns, but we don't know where
 * that is now.  We'll start by sampling roughly every other long word
 * horizontally and a pair of lines out of every 2*reductionFactor lines
 * (see below) vertically.  For a full scale 2300 x 3300 image, that's
 * about 32 horizontal samples and (reduction factor of 8), 400 samples
 * vertically, for a total of 12,800 samples.  One of the interesting jobs
 * will be to see how much this can be reduced while retaining sufficient
 * accuracy.
 *
 * Not only is it not necessary to count every pixel in a line, it's really
 * not necessary to count every pixel in a word.  If we think of looking for
 * horizontal rivers of white space, we can get by with saying the value of
 * a particular 32 bit sample is "1" if there are any bits on and 0 otherwise.
 * This improves counting speed.
 *
 * The original skew.c had three levels of code hierarchy.  The top level,
 * deskew(), used detectSkew to find the skew angle and rotateArb to
 * correct the image.  detectSkew used detectSkewSweep and detectSkewBif to
 * do the sweep and bifurcation steps respectively.  Each of those reduced
 * the image (!!) and used shearC and findDiffPowerSum to compute the scores.
 *
 * The sweep stage is used because the bifurcate step can otherwise choose a
 * local, but not global, maximum in skew score.  The number of false peaks
 * and amount of general noise of a hypothethical skew score vs angle curve
 * appears to be strong function of reduction factor.  This is because when
 * the program effectively looks at every 4th or 8th line, it's resolution
 * for edge discrimination is reduced.  Naturally, you say.  Edges can only
 * be determined to within 4 or 8 lines.  However, in addition to this
 * effect, the algorithm starts to lose its ability to distinguish between
 * one long edge and two separate horizontal edges which are slightly
 * displaced relative to each other.  For example, take a page with three
 * columns of text.  Say the inter-line spacing in all columns is identical,
 * but the columns are verticaly shifted relative to each other.  This is
 * reasonably common.
 *
 * Say the skew detection algorithm samples along a slanted line that goes
 * through all the black pixels of every column even though it is misaligned
 * with each column individually.  If the next sample line is far enough down
 * (due to reduction), it can go through all the white pixels of every column
 * even though it is also misaligned with each column.  The difference in
 * pixel counts between these two lines will be high, giving a misleading
 * high score.  I have an image where this happens - the "deskewed" image is
 * consistently skewed by about 0.5 degrees.
 *
 * One way to fix this is to sample every line of the image.  That's too
 * expensive.  A cheaper way is to sample adjacent pairs of lines of the
 * original image.  That is, instead of sampling every 4th or 8th line, as
 * is effectively done by reduction, sample 2 lines, then skip 6 or 14.
 * The skew score is calculated only between these pairs.  This is what we'll
 * do.
 *
 * Sampling adjacent pairs of lines has the effect of giving the skew score
 * curve a VERY thin, spikey peak.  This gives wonderful final resolution if
 * the algorithm can find it.  If the curve has no slope except in a very
 * thin region around the peak, the peak becomes a needle in the haystack.
 *
 * The spike will be widened by the fact that text lines, a prominent
 * horizontal feature do not present perfect horizontal surfaces.  They're
 * a bit noisy, so will contribute non-zero skew scores when they are close
 * to aligned.
 *
 * Unfortunately, even with this wonderful technique, we cannot eliminate
 * false peaks in the curve.  Even if the skew evaluation were perfect, pages
 * with slanted areas (almost any page with a picture on it) will have
 * blips in their skew curves.  These blips can potentially fool a pure
 * bifurcation algorithm.
 * 
 * The brute force way to identify the real peak is to sweep across the entire
 * band of angles.  The sweep can be a bit coarse, using bifurcation to home
 * in on the exact value.  But it's boring to be brutes.  Slow too.
 *
 * Another way to use two methods of generating skew scores.  One isn't very
 * exact, but produces a wide curve with enough amplitude to swamp out most
 * of the noise.  The other produces a thin, spikey curve for good resolution.
 * The second method is the one described above.  What function would produce
 * a big, wide curve?  Simply performing the previous algorithm with fewer
 * samples flattens the peak of the curve, but also decreases the signal to
 * noise ratio.
 *
 * A property of text lines is that they are very long, thin, horizontal bars.
 * We could construct a set of slanted paths across the image, counting
 * ON pixels encountered by these paths.  Higher counts would tend to
 * indicate that the paths are stay inside text lines.  However, this measure
 * is sensitve to both local orientation of text lines (which is good) and
 * global placement of lines and graphics across the page (which is bad).
 * A line's score can be high simply because it goes through many
 * independent text lines or through two or three text lines in different
 * columns that are displaced relative to each other.
 *
 * Instead, we sample pixels along linear paths in the image, remembering for
 * each path the *maximum number of consecutive pixels* encountered.  A path
 * that stays inside one text line will get a high score.  A path that goes
 * through a large number of text lines will get a low score since it will not
 * see a large section of any one of the lines.
 *
 * This scoring scheme doesn't give very good resolution since it can't
 * distinguish any two paths that follow a text line across the page.  Its
 * minimum resolution is the height of a text line.  However, it does produce
 * a nice broad curve that's reasonably insensitive to noise.  As the angle
 * of the paths relative to the text lines increases, the skew scores fall off
 * almost linearly.
 *
 * We could use bifurcation with the MaxPixels curve to find the skew angle
 * to within a degree, then use the HorizontalEdge method to fine tune. 
 * This is probably what we'll end up doing.  However, just to be adventurous,
 * we'll try another scheme first.  We assume the broad curve is pretty
 * linear.  We find a couple of points on each slope (four samples in all),
 * compute the angle at which the slopes intersect and use the bifurcation
 * method starting from there.  If that doesn't give a sufficiently accurate
 * first guess, we'll go back to bifurcation with the wide curve.
 *
 * How do we do bifurcation?  The original skew.c used 5 points, choosing one
 * of the interior three as the new center, then halving the angular distance
 * between them.  The number of samples required for this method is
 * 5 + 2 * number of steps.  Initially, we'll try a method that relies on the
 * the curves being fairly smooth.  We pick three inital points separated by
 * the maximum initial separation.  While the middle sample is not the highest,
 * take one more sample at the maximum separation on the other side of the
 * highest endpoint.  Once the middle is the highest, take a sample half way
 * between the middle and the next highest endpoint.  If everything were
 * smooth and symmetric, we'd know that the highest point was somewhere
 * in the range between the two highest samples, so we could just keep taking
 * a new sample between the two highest values and find it.  This gives a
 * total number of samples of 3 + number of samples.  If the curves for real
 * images are too noisy, we'll just repeat the first part of the algorithm -
 * taking samples on the far side of the highest endpoint until the middle
 * sample is the highest, then halving the separation and taking a sample
 * between the two highest samples.
 *
 * Efficiently tracing diagonal paths is more complex than I originally
 * thought.  Moving horizontally is easy, figuring out when to move
 * vertically and by how much is tricky.  To get sufficient accuracy,
 * we need to move vertically by fractional pixels.  Well, at least we need
 * to approximate fractional pixel movement.  This can be done pretty well
 * by moving up or down in units of whole lines at selected intervals.
 *
 * The job of plotting a diagonal path through the image is reduced then to
 * determining the size and frequency of vertical movement as a function of
 * angle.  One way to approach the problem is to pick an angle,
 * compute its tangent, multiply that by the width of the image (giving the
 * total rise at the end of the path) and divide the result by the number of
 * horizontal samples to be taken.  Yikes!  A trig function, a multiply and
 * a divide for every sample line!  Simple for scientists, but scary
 * for engineers.
 *
 * Instead of using angles as the space through which we search or sweep,
 * say we use vertical path displacement.  For small angles, the vertical
 * displacement is very nearly linear with angle anyway.  We can compute the
 * maximum positive and negative vertical displacments of the initial
 * search space as well as the vertical displacement associated with the
 * initial deltaAngle.  We then search or sweep through these displacements
 * skipping the number of lines associated with deltaAngle at each step.
 * To do bifurcation, we can just divide the incremental number of displacment
 * lines in half.  All this eliminates the trig function and the multiply
 * in the original algorithm.
 *
 * This is still a little messy because the width of the image still figures
 * into the equations.  We note that the angle of a path is (obviously)
 * independent of the width of the image.  In fact, we could use a hypothetical
 * image of any convenient width to compute vertical displacements so long as
 * the code that's traversing the real image checks to see when it's reached
 * an edge.  Convenient image widths are powers of 2 so divides and multiplies
 * turn into shifts.  We will follow this approach - a hypothetical image
 * used to compute vertical displacements that are used in turn to traverse
 * the real image.
 *
 * We discussed simulating fractional vertical line movement by occasional
 * full line movement.  For example, if the path rises 10 lines in
 * 20 samples, we move up one line every other sample.  Rather than have the
 * sampling code compute some marvelous error function, we'll just define
 * when vertical movements should be taken in a set of bit vectors.  We will
 * not produce a set of bit vectors for every possible image width.  Instead,
 * we will produce sets of fairly short bit vectors (16 bits) that can be
 * reused if the code takes more than 16 horizontal samples in a line.
 *
 * Nice theory, but what really happens?  Here are the steps in calculating
 * the necessary vertical displacement and choosing a bit vector corresponding
 * to the angle of the path.
 * 	1. Decide on the spacing between horizontal samples.  One algorithm
 *	   might want a 4 byte sample every 4 bytes.  Another might take a
 *	   4 byte sample every 12 or 16 bytes.
 *	2. Since our traversal algorithm will be periodic every 16 samples,
 *	   multiply the sample distance chosen in step 1 by 16 (shift left 4).
 *	   This gives the width of the hypothetical image.
 *	3. To determine the vertical offset in the hypothetical image for
 *	   maxAngle, multiply tan(maxAngle) by the width of the hypothetical
 *	   image computed in step 2.  Repeat for minAngle and delAngle.
 *	4. Sweep or search through the space of vertical offsets between
 *	   maxAngle's offset and minAngle's offset using an (initial)
 *	   increment of delAngle's offset.
 *	5. The intra-sample vertical offset for a particular path comes in
 * 	   two parts, integer and fraction.  The integer part is added to
 *	   the address on every sample.  This is simply equal to the
 *	   hypothetical vertical offset (the things we're searching through)
 *	   divided by 16 (shifted right 4).  The fractional part equals the
 *	   hypothetical vertical offset MOD 16 (AND with 0xF).  This is used
 *	   to choose a bit vector.  The bit vector is shifted left one place
 *	   after taking each sample.  If the bit that fell off the end was a
 *	   1, the bytes/line is added to (subtracted from) the next sample
 *	   address.
 *
 ***************************************************************************
 *
 * That worked, but we can do better in the coarse search.  Cycling through
 * (and reloading) the bit vectors is harder than we'd like.  We still end
 * up computing too many trig functions, doing too many multiplies and
 * divides, etc.  There are also too many cache refills.
 *
 * It turns out that a rise of 0.5 bits across 32 bits is 0.9 degrees.  That's
 * pretty close to 1 degree, which is the traditional value of deltaAngle
 * for coarse skew detection.
 * Anyway, it's good enough for a coarse search.  We'll divide the range from
 * minAngle to maxAngle into vertical displacements that change in units of
 * 8 bits.
 *
 * The assembly code that actually computes scores does so in a slightly
 * unrolled loop.  On every other read, it adds rSrcPitch to the source
 * address.  The code passes 0, +sourcePitch or -sourcePitch in this argument
 * depending on whether the current vertical displacment/32 horiz bits
 * is an integer (0), a positive non-integer (+sourcePitch) or a
 * negative non-integer (-sourcePitch).  Things become simpler, hence faster.
 *
 * We can improve the cache hit ratio on sweep detection algorithms by
 * dividing the image into blocks and doing the entire sweep on each block,
 * then moving on to the next block.  After a few passes, most of the bits
 * in a block will already be in the cache, so the remaining passes will go
 * faster.  However, if the blocks aren't wide enough, we end up adding lots
 * of little maximums to the skew score instead of discarding them in favor
 * of fewer, bigger maximums.  This flattens the score curves.
 *
 * Since the slant calculation for the coarse algorithm is independent of
 * the width of the block or image, we might as well make blocks cover the
 * entire width of the image.  This will cause more cache misses than we'd
 * like, but will improve the accuracy of the result.
 *
 ***************************************************************************
 *
 * That was successful and still pretty fast.  Now, we change from a sweep
 * to a binary search.  In the search algorithm, we sample both extreme grid
 * points and the middle.  We then continue to sample half way between the
 * two highest grid points until we reach a sample interval of 1 degree.
 *
 * Well, almost.  Say the image is almost centered, the skew angle is just
 * a bit positive (for example).  Noise in the skew function can cause
 * minAngle's score to be a little bigger than maxAngle's score.  This has us
 * eliminating all the positive angles in the first step, a mistake.  Instead
 * we don't adjust the trial angle quite so far in the first couple of steps.
 * For example, instead of curMaxAngle = (curMaxAngle+curMinAngle)/2, we say
 * curMaxAngle = (curMaxAngle+curMinAngle)/2 + (curMaxAngle-curMinAngle)/8.
 * The second term shortens the first couple of steps.
 *
 ***************************************************************************
 *
 * The fine skew detection algorithm will also use a binary search.  Instead
 * of using angles, we deal in terms of vertical displacment of slanted lines
 * over a hypothetical 4k bit horizontal span.  Yeah, but real images aren't
 * 4K bits wide, right?  This just means that if the slanted line would have
 * risen n bits over a 4Kb stride, it rises n*(imageWidth/4K) bits over a
 * imageWidth stride.
 *
 * Since everything is proportional, why not choose a smaller hypothetical
 * stride?  If we choose a stride smaller than the width of a real image, we
 * get insufficient angular resolution.  In an extreme case, say the
 * hypothetical image were 32 bits wide and the smallest angle we could
 * handle had 1 bit rise in 32 bit stride.  This corresponds to a rise of
 * 80 lines in a 2560 bit real image.  We'd like a resolution of +- 1 line
 * in 2560 bits, so need to choose a hypothetical image wider than the real
 * one.  The next larger power of 2 is 4K.
 *
 * The fine skew detection algorithms is based on adding up differences
 * vertically adjacant samples.  It's not necessary to sample every point
 * along each grid line, so we'll spread the samples out a bit.  Initially,
 * we'll try sampling 32 bits (8 bits?) out of every 128.  That would give
 * 4K/128 = 32 samples in the hypothetical image.  Real images are less wide,
 * so we have to stop when we reach the edge.
 *
 * Since we'll need good accuracy, we can't get away with the 1 degree trick
 * used by the coarse routine.  In fact, we need to be able to rise or fall
 * 0 to 31 lines smoothly over the course of 32 samples.  There is a
 * cute way to do this using one counter setting the interval at which
 * the pitch is added to the address and another counter setting the interval
 * for subtracting the pitch (error correction).  However, since we know
 * there will never be more than 32 samples, we'll just use a bit vector
 * to decide when to add the (positive or negative) pitch to the current
 * address when moving "horizontally".
 *
 *********************
 * The original version of this just got noise, no peak to the curve.  More
 * samples are probably needed.  But that brings up the question, "Why are
 * so many samples needed?"  Because very few of them actually find edges
 * to detect.  We need to take lots of samples so that some interesting
 * number actually find edges to measure.
 *
 * That sounds awfully inefficient.  What if we looked for edges instead of
 * simply hoping that they'd show up sometimes?  This is the idea behind
 * ip_detectSkewRobot.
 *
 * Say we view an image as a maze of text and data also having white and
 * black deserts.  The text and data (especially text) have nice horizontal
 * edges we'd like to find and whose angle we'd like to record.  The deserts
 * are big vacant areas.
 *
 * We'll pick a bunch of points (not at random, that takes too long).  An easy
 * set of points forms a big X across the image.  At each image point, we put
 * down a little "robot."  If the little fellow lands in a white space
 * (all 32 bits are white), it looks vertically down for some black.  It
 * samples every 16th line.  If no black is found in a few samples, the
 * robot concludes it's in a desert and gives up (returns).
 *
 * If the robot finds some black, or lands in black to begin with, it
 * looks below for white.  It samples every 8th line (white space being
 * generally thin between text lines).  If no white is found in a few samples
 * the robot concludes it's in a black desert and gives up.
 *
 * If white space is found, we have a black/white edge.  These tend to be
 * the best kind since text is usually darker than the background (not
 * always, but tough) and the underside of a text line isn't as jagged as
 * the top side.  We do a short binary search until we have found the edge
 * exactly.
 *
 * We now attempt to follow the edge to the right.  Starting on the white
 * line under the black line, we sample the next word to the right.  If it's
 * also white, we look up to see if the black stuff is still there.  If
 * so, we record that one horizontal step was taken and no vertical steps
 * were taken.  If not, we keep looking up for 5 lines.  If the black stuff
 * is found, we record one horizontal step and as many vertical steps
 * (a negative number) as were required to find the black stuff.  We also
 * move up to the line immediately below the black stuff.  If no black
 * stuff was found, we conclude that the edge has ended (or was too steep
 * to consider) and quit.
 *
 * If the step to the right found black stuff, we look down for up to 5 lines.
 * If white stuff is found again, we record one horizontal step and as many
 * the number of lines (this time a positive number) as required to find white.
 * We also move down.  If no white is found, we just quit.
 *
 * The previous two steps are repeated so long as the edge is seen.
 *
 * The purpose of all this is to record the slope of the edges.  We pick some
 * number of total horizontal steps to record and keep putting down robots
 * until that number is found.  We then get the slope by dividing the total
 * number of vertical steps by the total number of horizontal steps.
 *
 * Clearly, we need to traverse the image several times to get rid of noise.
 * The number of horizontal steps required will be determined by
 * experimentation.
 *
 * A more complex version of this code could maintain several vertical
 * counters for different slopes.  We'd then pick the biggest one as the
 * dominant slope.  As it is now, slanted lines in the image contribute to
 * the average slope calculation.
 *
 **************************************************************
 *
 * We had some problems.  First, there weren`t enough samples.  We now
 * take sample points in columns.  The sample points within a column are
 * close enough to catch the majority of text lines.
 *
 * Second, the samples were so wide that we could bridge across slanted
 * text rows.  The sample width is now 16 bits instead of 32.
 *
 *************************************************************
 *
 * All this worked well on most documents.  Z Smith noticed that it got
 * confused with normal subsampling factors in line drawings.  We think
 * this is because the lines are so thin.  To get some indication about
 * when the answer is confused, we now return a ratio of the maximum
 * power found to the minimum power found.  Applications can use this
 * heuristically to decide when to change the subsample factor.  A pointer
 * to the ratio, called "confidence" is passed to i_detectSkew and then
 * to ip_detectSkewFine.  ip_detectSkewFine fills in the value pointed to when
 * it returns.
 *
 *************************************************************
 *
 * That was nice, but wasn't a very nice measure of confidence.  We'd like
 * something that depends not only on the relative height of the peak
 * (the previous measure depended only on the difference between the
 * highest and lowest points so could be confused by noise) but also on
 * the peak's width.
 *
 * Say the total power over all samples is sumOfPowers.  The sum of the
 * absolute values of angular distances from the final skew angle to the
 * sample points is sumOfAngles.  The new confidence measure returned by
 * ip_detectSkewFine is
 *	confidence = ( Sum( ABS(finalAngle - angle) * power(angle) ) /
 *			  (sumOfPowers) )
 * That is, we weight each other angle searched by the fraction of the total
 * power assigned to that angle.  This weighted sum is the average angular
 * distance from the result.
 *
 * i_detectSkew computes the fraction of this number compared to 1 degree.
 * If the number is greater than 1 degree, the returned confidence is 0.
 * If the average offset is 0, the returned number is 100.  From a few
 * trials on images, confidences of more than 65 seem to indicate
 * reasonable certainty.  The caller may also wish to call this
 * routine with a smaller subsampling factor and compare answers.  If
 * the confidence stays sort of low (50's, lower 60's) and the answers
 * vary by more than 0.2 degrees from one subsampling interval to the
 * next, we're probably confused.
 *
 *************************************************************
 *
 * That was better, but as Dan Bloomberg noted, rotating an image in which
 * the confidence was ok (but not great) by 90 degrees and computing
 * the skew angle again gave about the same confidence.  Basically, it
 * does fine on images that have lots of long horizontal edges and not
 * much else.
 *
 * We'll try looking out +- 2 degrees and +- 3 degrees from the angle found
 * The powers we get from these will be used in the confidence study.  The
 * idea is that a lowish, noise peak should be emphasized if we look farther
 * away if it's good and be shown as noise if it's bad.
 *
 * Post change: adding this change and fixing a bug further increased the
 * utility of the confidence measure.  However, if you run the skew detect
 * code on normal magazine images that have been rotated 90 degrees, you
 * still get a pretty high confidence.  Why?  Well, there are lots of
 * vertical strokes in text.  These become horizontal strokes.  Also, the
 * page often contains vertical rules that become horizontal rules.  Finally,
 * text is usually in columns with justified edges.  These edges are also
 * found.  However, I did find that the presence of a slightly skewed line
 * or two can result in a skew measure that's almost a degree off, yet still
 * reports high confidence.  I recommend that users use the landscape/portrait
 * heuristic to orient the image, then run this code to find the really
 * correct angle.
 *
 * In this iteration, we also started having ip_detectSkewFine look +- 2
 * degrees around the offset returned by ip_detectSkewRobot.  This helps
 * for fairly large skews (6-8 degrees).
 *
 */ 

/* PRIVATE */
/*****************************************************************************
 *  ip_findDiffPixelsSum():
 *
 *        Calculates differences in pixel sums across adjacent diagonal lines.
 *	  All differences are (sloppily) raised to a power and these results
 *	  are added together.  The final sum, a 63-bit number (32 bits in
 *	  upper word, 31 bits in lower 31 bits of lower word) of
 *	  sums-to-a-power is returned.
 *	pImage: 	pointer to start of first image line to be sampled
 *	numVSamp:	number of vertical lines to be sampled
 *	numHSamp:	number of horizontal samples/line
 *	vAddrDisp:	difference between addresses of vertical samples
 *	hAddrDisp:	difference between addresses of horizontal samples
 *	imBpl:		image's bytes/line (pos if sample lines slope down,
 *			neg if sample lines slope up)
 *	fracBitVector:	a bit vector that indicates when the sample path
 *			should move up/down an extra line.  This is how we
 *			trace exact diagonal paths.
 *	pResultHi:	Hi word of result
 *	pResultLo:	Low word of result
 * NOTE: real result = *pResultHi[31:0] : *pResultLo[30:0], i.e. the sign
 * bit of *pResultLo is always clear.  We use it as a carry bit to detect
 * when to increment *pResultHi.
 *****************************************************************************/
Int32 CDECL 
ip_findDiffPixelsSum (
    IPState    *pIPState,
    UInt8Ptr	pImage,
    Int32	numVSamp,
    Int32	numHSamp,
    Int32	vAddrDisp,
    Int32	hAddrDisp,
    Int32	imBpl,
    UInt32	fracBitVector,
    Int32Ptr	pResultLo,
    Int32Ptr	pResultHi)
{
PROLOGUE;
Int32		hSamp, vSamp;		/* iteration vars */
UInt8Ptr	pImageLo;		/* points to other sample line
				         */
Int32		lineOffset;		/* distance (in memory bytes)
					 * from first sample of
					 * current line */
UInt32		lineScore;		/* holds differnces for current sample
					 * line.  We assume no single sample
					 * line can go beyond 4Gig. */
UInt32		totalScore[2];		/* holds differences for
					 * total for all sample lines.
					 * totalScore[0] has LS word,
					 * totalScore[1] has MS word.
					 */
UInt32		shiftedScore;		/* used to compute totalScore */
UInt32		curFracBitVec;		/* current fraction bit vec */
UInt32		curInputDiff;		/* bit-difference in current
					 * inputs */
UInt32		input, otherInput;	/* the 2 input words to
					 * be compared */


#define	cScoreShiftDist	4	/* Used to compute the non-linear function
				 * that emphasizes long edges.
				 * lineScore =
				 * lineScore << (lineScore >> cScoreShiftDist)
				 */
#define	cByteSamples	0x7F7F7F7F
#define	cFracWidth	32		/* number of samples before reloading
					   fracBitVector */
#define cFracCountMask	0x1F		/* detect hSamp being a multiple of 32 */

/* This code trundles across a line, computing the difference in vertically
 * adjacent pixels at each step.  Our difference is recorded if one sample has
 * any pixels and the other has none.
 *
 * The constant "scoreShiftDist" helps implement our shoddy non-linear
 * score adjustment.  What's that for?  Say we do a bunch of samples along
 * trajectories that cross lots of text lines.  The scores for the sample
 * lines will all have some medium value since they'll all find about the
 * same number of edges.  Say we sample along trajectories that are perfectly
 * aligned with the text lines in the image.  Lots of the scores will be 0
 * since they fall right in the middle of inter-line gaps or go right along
 * the middle of a text line.  A few scores will be really big since they'll
 * go right along an edge.  However, the sum of all the scores will be the
 * same as the sum of the scores along any other set of trajectories.  We need
 * some way to make the few really big scores stand out.
 *
 * We do this by shifting each score left by cScoreShiftDist, then using this
 * as a shift count for shifting the score left.  Little differences in the
 * shift count make big differences in shifted scores.  This seems to work
 * quite well and is a LOT faster than raising the score to a power.
 *
 * Moving diagonally is messy.  Why not move vertically instead?  Well, if we
 * moved through an entire column, we'd run out of cache so would miss all the
 * time.  We'd have to break the image into horizontal blocks.  In order to
 * make sure the cache didn't fill, the blocks couldn't be very tall.  We
 * wouldn't take very many samples per block, so there would be a fair amount
 * of loop/call/return overhead.  We'll just move horizontally instead.
 *
 * Diagonal movement has two parts.  After each sample, we add an address
 * displacment (hAddrDisp).  We also shift fracBitVector left by 1.  If a 1
 * fell out, we also add imBpl to the address.  This is how we approximate
 * moving up or down by fractional lines - we move up or down by whole lines
 * at selected intervals.
 *
 * MODIFICATIONS:
 *
 * We were recording an edge difference if a full 32 bit word was zero/non-zero
 * and the word underneath was non-zero/zero.  The results were not
 * sufficiently accurate.  We will now record a difference *for each byte pair*
 * in the 4 byte words.  The byte pair has a difference if one byte is zero
 * and the other is not.
 *
 * BUG FIX, 2/21/95: The total scores can overflow for some images.  To
 * fix this, we keep the total score in two words, with the carry bit
 * in big 31 of the LS word.  These two words are totalScore[0] (LS word),
 * and totalScore[1].
 */

	/* Initialize the score */
    totalScore[0] = totalScore[1] = 0;

	/* Point to the line below the current one so we can take differences */
    pImageLo = pImage + 2*imBpl;

	/* Loop through all the sample rows */
    for (vSamp = 0;
	 vSamp < numVSamp;
	 pImage += vAddrDisp, pImageLo += vAddrDisp, vSamp++)
    {
	lineScore = 0;			/* initialize score for this line */
	lineOffset = 0;			/* address offset for the line */
	curFracBitVec = fracBitVector;

	    /* Loop through all samples in a row.  Start at 1 so we can
	     * detect every 32nd iteration by anding with 0x1F and getting
	     * 0. */
	for (hSamp = 1; hSamp <= numHSamp; hSamp++)
	{
	    input = *(UInt32Ptr) &pImage[lineOffset];
	    otherInput = *(UInt32Ptr) &pImageLo[lineOffset];

	    lineOffset += hAddrDisp;		/* move across the line */
	    if (curFracBitVec & 0x80000000)	/* move fraction line now? */
		lineOffset += imBpl;		/* yes */
	    curFracBitVec = curFracBitVec << 1;	/* and shift fraction */

		/* Did we run out of bits in curFracBitVec? */
	    if (!(hSamp & cFracCountMask))
		curFracBitVec = fracBitVector;

		/* Was there a difference? Compute individually for each byte.
		 * Well, ok for speed, we really only see if there were
		 * differences in the least significant 7 bits of each byte.
		 * We do this by anding with 0x7F7F7F7F, (clears top bit of
		 * each byte) and adding the same constant.  This puts a 1
		 * into the top bit of each byte if there were any bits on.
		 * The two input words are then xor'd together to determine
		 * if there were any differences.
		 * These bits are shifted right by 7 and added to lineScore.
		 * lineScore is really four individual counts.  They get
		 * combined at the end of the line.
		 * Since difference should be noted if one sample has
		 * any pixels and the other has none, the input words
		 * are not xor'd together until after determining if
		 * there were any bits on.  If the input words were
		 * xor'd together first, differences are found if any
		 * pixels are different.
		 */
	    input &= cByteSamples;
	    input += cByteSamples;
	    otherInput &= cByteSamples;
	    otherInput += cByteSamples;
	    curInputDiff = input ^ otherInput;
	    curInputDiff &= ~cByteSamples;	/* get just indicator bits */
	    curInputDiff = curInputDiff >> 7;
	    lineScore += curInputDiff;

	}

	    /* We're at the end of the row.  There are 4 edge counts in
	     * rRunScr.  Since there are at most 127 samples in a line,
	     * we know that the samples are at most 127.  Add them together,
	     * then perform a nice non-linear function on the score and
	     * add the result to rScoreHi:rScoreLo.  The
	     * non-linear functions is here so that big scores (as you'd get
	     * if you were sampling along a real edge) really stand out while
	     * average scores (as you'd get on any diagonal sample) don't.
	     * Current non-linear function is lineScore * 2^(lineScore/16) 
	     */
	shiftedScore = lineScore >> 16;		/* get 2 scores from 4 */
	lineScore += shiftedScore;
	shiftedScore = lineScore >> 8;		/* get 1 score from 2 */
	shiftedScore &= 0xFF;			/* get rid of old trash */
	lineScore &= 0xFF;
	lineScore += shiftedScore;		/* compute line's score */

	    /* do the crude non-linear operation (lineScore  =
	     * lineScore << (lineScore/16) )
	     * We do the shifting is two pieces.  The part that
	     * fits only in the upper score word is shifted and
	     * added separately.  This part is everything that
	     * would have been shifted into bits [63:31] of the
	     * output word.  We include bit 31 here because we
	     * want to leave bit 31 of the lower word empty.  We
	     * use it as a sign bit when adding to the low score word.
	     */
	shiftedScore = lineScore >> cScoreShiftDist; /* lineScore/16 */
	totalScore[1] += lineScore >> (31 - shiftedScore);
	lineScore = lineScore << shiftedScore;	/* big values of lineScore get
						   really big here */
	lineScore &= 0x7FFFFFFF;		/* get rid of sign bit */
	totalScore[0] += lineScore;		/* add into total */
	if (totalScore[0] & 0x80000000)
	{
	    totalScore[1]++;
	    totalScore[0] &= 0x7FFFFFFF;
	}

    } /* end of all sample rows */
    *pResultLo = totalScore[0];
    *pResultHi = totalScore[1];
    NORMAL_RETURN;
}



/*
 *  ip_doEdgeSample():
 */
/* This routine is quite unlike the others.  Instead of taking statistics
 * from fixed points in the image, this routine looks for an edge, then
 * plots its slope.

 * This routine implements the "robot" referred to in ip_detectSkewRobot.
 * After getting plopped down at some random point in the
 * image, we take the following actions:
 *	1. If the sample at the initial location has no black pixels (is
 *	   white), sample every cSampleSpacing'th line of the image
 *	   straight down.
 *	   If no black pixels are found in 3 samples, we just give up and 
 *	   return.
 *	2. If the initial location had black pixels or we have found black
 *	   pixels in step 1, we look down every cSampleSpacing'th line of
 *	   the image straight down.  If no white is found in 3 samples,
 *	   just give up and return.
 *	3. If white was found, do a linear search to find the exact boundary
 *	   between black and white.
 *	4. Starting from the lowest black word, move one halfword to the right.
 *	5. If the new halfword is black, look down to see if the halfword below
 *	   is white.  If so, increment hCount by 1 and go to step 4.
 *	6. If the new halfword was black, but the word below it was also black,
 *	   keep looking down for a total of 5 additional lines.  If white is
 *	   found, add the number of lines above from vCount
 *	   and increment hCount.  Adjust the vertical position to the
 *	   uppermost white word and go to step 4 again.  If no white was
 *	   found, just quit.
 *	7. If the new word was white, look upwards for black.  If black is
 *	   found in no more than 5 lines, subtract the offset from vCount,
 *	   increment hCount, adjust the current position to the uppermost
 *	   black word and continue with step 4.  If no black was found, quit.
 *	8. When we quit, add the final values of vCount and hCount in the
 *	   memory structure pointed to by rImSlopeAddr.

 * Alterations:

 *	This whole thing is both more complex and simpler than I had
 *	thought at first.  What we're really doing is measuring the slope
 *	of individual *lines* of text, then averaging these.  We can think
 *	of this as setting a start point, tracking along the slope, setting
 *	the end point and recording the slope between the start and end
 *	points.  We happen to record the slope all the way along because we
 *	never know when the end will show up.  The important insight here
 *	is that the placements of the start and endpoints are critical.
 *
 *	Unfortunately, these are the hardest to place.  When starting a line,
 *	we can easily start too high on a sloped character.  When finishing
 *	a line, we can be fooled by a descender (go too low) or a rounded
 *	character (move up too high).  We'd like to measure slopes on "good
 *	samples."  But how?
 *
 *	Just tossing the first and last samples is not good, the second and
 *	second to last samples are just as likely to be funny.  What we'll
 *	do instead is look for runs of samples that seem to be following a
 *	nice edge.  We'll use the leftmost sample of the first run as the
 *	beginning of the line.  We'll use the rightmost sample of the last
 *	run as the end of the line and record that slope.
 *
 *	We use a simple state machine to implement this.  Initially, the
 *	state machine looks for a run.  Each time the step is "big" vertically,
 *	it clears the horizontal and vertical counters.  When the first run
 *	is found, the horizontal and vertical counters automatically have the
 *	slope since the last big step.
 *
 *	We then follow runs.  At each little step inside a run, we record the
 *	latest slope in hRunEnd and vRunEnd.  If a big step is seen, we
 *	look for another run, but do not clear hCount or vCount.  We also do
 *	NOT change hRunEnd and vRunEnd.
 *
 *	The routine ends when no step can be taken because we've run out of
 *	edge.  If the edge has been long enough, we record the last slope in
 *	hRunEnd and vRunEnd.
 *
 * **************************************
 *
 *	That sort of worked ok, except it made the lines over which slopes
 *	are calculated too short.  A difference of 0.1 degrees equals +- 1
 *	line of vertical displacement in 573 bits horizontal displacement!
 *	To get good accuracy, we really need to measure over long lines.
 *
 *	However, we can't just jump over white space willy-nilly.  An inter-
 *	word gap (one we'd like to cross) in one document can be bigger
 *	than an inter-column gap (should NOT cross) in another.  How far can
 *	we jump?
 *
 *	Initially, we thought of using the character height as a measure of
 *	allowable white space gap.  Unfortunately, justified text has much
 *	more space between words than unjustified text, yet the character
 *	height does not change.  Instead, we'll let the higher level routine
 *	try different white space widths.  This routine will return the
 *	number of spaces actually skipped.
 *
 *	The caller will start out with a large value for allowable white
 *	white space.  If the number of spaces actually skipped is small
 *	compared to the number of samples taken (but non-zero), we assume
 *	columns have been skipped and reduce the white space value until
 *	the white space skipped value goes to zero.
 *
 * ************************************************************************
 * 
 * That got pretty close to reality, but still failed (off by 0.5 degrees)
 * for pages that had lots of short lines - phone books or data sheets.
 * At that point, we decided to punt.  This routine will only be used to
 * do a very quick estimate of coarse skew - within a degree.  The fine skew
 * detection routine above will be used to narrow it down from there.
 *
 * There was yet another bug.  One of the CCITT test charts has a diagonal
 * line running from upper left to lower right.  This robot routine followed
 * that line.  Oooops.  The fix was to put a check just after the portion
 * of the loop that tries to track descenders.  If the total horizontal
 * distance covered in that loop is much greater than the amount of white
 * space we're allowed to skip plus the minimum run length required to
 * exit the loop, we decide that we've followed some big alien object.
 * 
 ***************************************************************
 *
 * 9/24/93 BUG FIX:.  There's a possibility of an infinite loop if a
 * horizontal black line extends completely across the image AND the caller
 * doesn't have a frame around the image.  We would keep following the
 * horizontal line forever.  XIS may have seen this.  Anyway, we'll say that
 * if we're following an edge for more than the width of the image - stop
 * following it.
 */

/* PRIVATE */
/************************************************************************
 *  ip_doEdgeSample():
 *
 *        Tries to find an follow one diagonal edge
 *	  Input:  Initial Address, source pitch, trackLimit, min and max legal
 *		  addresses, ptr to result structure.
 *	  Output: slope of edge followed in result structure (vert dist,
 *		  horiz. dist)
 *	  This routine implements a single "probe" described above.
 *	pImage: 	pointer to initial probe address
 *	imBpl:		image's bytes/line
 *	imSlope:	pointer to resulting slope structure defined below
 *	maxVertSearch:	Give up if we look this many lines down without
 *			finding a horizontal edge.
 *	minLegalAddr:	Addresses less than this are above the top of the image
 *	maxLegalAddr:	Ditto for off the bottom of the image.
 ************************************************************************/
void CDECL 
ip_doEdgeSample (
    IPState    *pIPState,
    UInt8Ptr	pImage,
    Int32	imBpl,
    SlopePtr	imSlope,
    Int32	maxVertSearch,
    Int32	minLegalAddr,
    Int32	maxLegalAddr)
{
#define	cSampleSpacing	8	/* sample every 8th vert line when
				 * looking for first edge */
#define	cLogSampleSpacing 3

#define	cMaxSpaceCount	2	/* maximum amount of white space that
				 * we can skip over */
#define	cBytesPerSample	2	/* there are 2 bytes/space.  We
				 * need this to compute maxRunLength
				 * from imBpl */

#define	cMinValidRunLen 2	/* ignore runs shorter than this */
#define	cMinSampleLength cMinValidRunLen+1 /* toss edges shorter than this */
#define cBigStepValue	3	/* if the edge changes by more than
				 * this in one sample, it's a
				 * "big step."  Track it, but remember
				 * it might not be well behaved. */
#define	cLogHeightToSkip 4

Int32	hCount, vCount;		/* current values of horiz and vert
				 * counts */
Int32	srcOffset, lineOffset;	/* address offset from beginning of
				 * source line; between src lines */
Int32	hRunEnd, vRunEnd;	/* horiz and very components of latest
			         * run */
Int32	runLength;		/* length of current run, used to
				 * decide whether its long enough to
				 * count */
Int32	runLengthBound;		/* Start with this equal to the number
				 * of samples in the page.  Decrement it
				 * for each sample.  If the sample doesn't
				 * stop before this gets to 0, stop
				 * it then. */
Int32	currentSpaces;		/* number of white spaces to skip over
				 * before we give up */
	

	/* Stop following edges longer than the width of the page. */
    runLengthBound = imBpl / cBytesPerSample;

	/* Assume we're in a white place, look down every
	   2**(cLogSampleSpacing) lines for some black. */
    srcOffset = 0;
    lineOffset = imBpl << cLogSampleSpacing;
    if (! *((UInt32Ptr) &pImage[srcOffset]))
    {
	srcOffset += lineOffset;	/* pt to next line */
	if (! *((UInt32Ptr) &pImage[srcOffset]))
	{
	    srcOffset += lineOffset;
	    if (! *((UInt32Ptr) &pImage[srcOffset]))
		return;			/* lost in a white desert so return */
	}
    }

	/* Have found some black.  Look down for white so we can get a
	 * black/white horizontal edge to track */
    srcOffset += lineOffset;
    if (*((UInt32Ptr) &pImage[srcOffset]))
    {
	srcOffset += lineOffset;	/* pt to next line */
	if (*((UInt32Ptr) &pImage[srcOffset]))
	{
	    srcOffset += lineOffset;
	    if (*((UInt32Ptr) &pImage[srcOffset]))
		return;			/* lost in a black desert so return */
	}
    }

	/* We've found a black-on-top-of-white edge, sort of.  Refine the
	 * location using a linear search.  This guarantees we really find
	 * the first black edge here.
	 */
    while (!*((UInt32Ptr) &pImage[srcOffset]))
	srcOffset -= imBpl;

	/* At this point, pImage[scrOffset] points to the lowest black line,
	 * which is exactly what we want.  However, we'd like pImage to be
	 * pointing there by itself so we can use srcOffset for offsets relative
	 * to this point */
    pImage += srcOffset;

/************ TRACKING STATE MACHINE STARTS HERE ********************/

	/* We're ready to start the tracking state machine. In the initial
	 * state, we clear rHCount and rVCount.  We return here so long as
	 * one of the first cMinValidRunLen steps is big. */

	/* First, initialize the vars used to record the slope we've seen
	 * in this run.
	 * Why do we maintain both hRunEnd and hCount?  hCount says how far
	 * we've come since the start of the first bona fide run.  It includes
	 * any white space or runt-runs we're currently working on.  We don't
	 * want to return that trash.  hRunEnd is updated with hCount only when
	 * hCount is guaranteed to be in a run.  Therefore hRunEnd filters
	 * out possible crud on the far left and right ends of the line. */
    hRunEnd = vRunEnd = 0;
    hCount = vCount = 0;
    currentSpaces = 0;			/* not in white space now */
    runLength = 0;			/* haven't seen a run yet */

    while ((runLength < cMinValidRunLen) && (runLengthBound > 0) &&
	   (currentSpaces < cMaxSpaceCount))
    {
	if (cBigStepValue <
		ip_doFindEdgeTakeStep (pIPState, &pImage, imBpl, minLegalAddr,
				       maxLegalAddr, maxVertSearch, &hCount,
				       &vCount, &currentSpaces))
	{
	    runLength = 0;		/* still haven't found a run */
	    hCount = vCount = 0;
	}
	else
	{
	    runLength++;		/* we might be in the start of a run */
	}
	runLengthBound--;		/* we've moved over 1 */
    }

	/* At this point, we've either found the very first run or have run
	 * into too much white space and given up.  Enter the loop that
	 * follows runs and tries to bridge between runs.
	 */
    while ((currentSpaces < cMaxSpaceCount) && (runLengthBound > 0))
    {
	hRunEnd = hCount;		/* update latest legal displacements */
	vRunEnd = vCount;
	if (cBigStepValue <
		ip_doFindEdgeTakeStep (pIPState, &pImage, imBpl, minLegalAddr,
				       maxLegalAddr, maxVertSearch, &hCount,
				       &vCount, &currentSpaces))
	{
		/* a big step was just taken, look for a new run */
	    runLength = 0;
	    while ((runLength < cMinValidRunLen) && (runLengthBound > 0) &&
		   (currentSpaces < cMaxSpaceCount))
	    {
		if (cBigStepValue <
			ip_doFindEdgeTakeStep (pIPState, &pImage,
					       imBpl, minLegalAddr,
					       maxLegalAddr, maxVertSearch,
					       &hCount, &vCount,
					       &currentSpaces))
		{
		    runLength = 0;
		}
		else
		{
		    runLength++;
		}
		runLengthBound--;	/* have moved over by 1 sample */
	    }  /* have found a new run or end of all runs */

		/* if the amount of horizontal stuff we've skipped over
		 * is much greater than cMinValidRunLen + cMaxSpaceCount,
		 * We've been tracking some weird thing for a long time.
		 * We don't want to see it, so arbitrarily set the
		 * currentSpaces to a big number we we'll ignore it.
		 */
	    if (hCount - hRunEnd > cMaxSpaceCount + cMinValidRunLen + 1)
		currentSpaces = cMaxSpaceCount+1;

	}	/* end of case that ends a run */
	runLengthBound--;		/* have moved over 1 step */
    }

	/* We've now ended the last run.  Was it big enough to record? */
    if (hRunEnd > cMinSampleLength)
    {
	imSlope->hCount += hRunEnd;
	imSlope->vCount += vRunEnd;
    }
}


/* PRIVATE */
/************************************************************************
 *  ip_doFindEdgeTakeStep():
 *
 *        Tries to take one step along a diagonal edge
 *	  Input:  ptr to Initial Address, source pitch, trackLimit, min and max
 *		  legal addresses, ptr to horizontal and vertical distances and
 *		  ptr to number of white space chars we're allowed to skip
 *		  before giving up.
 *	  Output: updated initial address, horizontal and vert distances and
 *		  updated number of white space blocks we have crossed in
 *		  a row.  Return value is size of vertical step taken.
 *	ppImage: 	pointer to pointer to current probe address.  This is
 *			updated after the step is taken
 *	imBpl:		image's bytes/line
 *	minLegalAddr:	Addresses less than this are above the top of the image
 *	maxLegalAddr:	Ditto for off the bottom of the image.
 *	maxVertSearch:	Give up if we look this many lines down without
 *			finding a horizontal edge.
 *	pHCount:	pointer to count of horizontal steps taken by probe
 *	pVCount:	pointer to count of vertical steps taken by probe
 *	pCurrentSpaces:	pointer to count of horizontal white space steps
 *			taken in current block of white space.  If this gets
 *			too big, quit.  This allows us to cross inter-word
 *			gaps.
 ************************************************************************/
Int32 CDECL 
ip_doFindEdgeTakeStep (
    IPState    *pIPState,
    UInt8Ptr   *ppImage,
    Int32	imBpl,
    Int32	minLegalAddr,
    Int32	maxLegalAddr,
    Int32	maxVertSearch,
    Int32Ptr	pHCount,
    Int32Ptr	pVCount,
    Int32Ptr	pCurrentSpaces)
{
    PROLOGUE;
    UInt8Ptr	pImage;		/* a more useful form */
    Int32		srcOffset;	/* addr offset from pImage */
    Int32		lineOffset;	/* number of lines we move */
    UInt16		sample;		/* memory sample */
/* START DEBUG ONLY*/
/*
    UInt16        	*debug;
*/ 
/* END DEBUG */

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* Start by loading the image pointer and testing to see if it's
	 * out of range */
    pImage = *ppImage;
    if (((Int32) pImage <= minLegalAddr) || ((Int32) pImage >= maxLegalAddr))
    {
	*pCurrentSpaces = cMaxSpaceCount + 1;	/* we're done */
	 return(cBigStepValue+1);
    }

	/* Move 2 bytes to the right and take a sample.  If it's black, look
	 * down one line.  If it's white, start looking up for the black edge.
	 */
    pImage += 2;
    sample = USHORT_ACCESS(pImage);

    if (sample)				/* sample was black */
    {

	sample = USHORT_ACCESS(pImage+imBpl);	/* look down */

	if (!sample)	/* we're still on the edge */
	{
	    (*pHCount)++;		/* moved over by 1 sample */
	    *pCurrentSpaces = 0;	/* not in the middle of white space */
	    *ppImage = (UInt8Ptr) (pImage);
	    return (0);			/* went up/down by 0 steps */
	}

	    /* The sample below was black too.  That means we need to
	     * look down until we find some white.  Apparently some very
	     * smart person decided that we should try even harder to find
	     * runs that encounter descenders in text with positive
	     * angles.  That means we really look down a long way.
	     * Anyway, that's what the assembly version does.  Also,
	     * we really want the loop to execute 2*maxVertSearch
	     * times.  Since lineOffset starts at 2, we need to bump
	     * the limit up by 2.  However, the assembly code tests
	     * does this loop one less time than you would think, so
	     * we use 2*maxVertSearch+1 instead of 2*maxVertSearch+2 as
	     * the ending condition.
	     */
	for (lineOffset = 2, srcOffset = imBpl<<1;
	     ((lineOffset < 2*maxVertSearch+1) && (sample));
	     lineOffset++, srcOffset += imBpl)
	{
	    sample = USHORT_ACCESS(pImage+srcOffset);
	}

	    /* we are now either 1 line below the black/white boundary or
	     * have found a black desert. */
	if (sample)	/* in a black desert */
	{
	    *pCurrentSpaces = cMaxSpaceCount + 1;	/* we're done */
	     return(cBigStepValue+1);
	}
	else
	{
		/* We found a black/white edge.  Because of the loop structure,
		 * lineOffset and srcOffset point one line too low.
		 * Update *ppImage, *pHCount, *pVCount and return the number
		 * of lines we skipped.
		 * The loop above stopped when it found a WHITE sample,
		 * then it incremented srcOffset one line beyond that.
		 * We want to point to the lowest black line, so
		 * need to back up 2 lines.
		 */
	    *ppImage = (UInt8Ptr) &pImage[srcOffset - 2*imBpl];
	    (*pHCount)++;
	    *pVCount += lineOffset-2;
	    *pCurrentSpaces = 0;
	    return(lineOffset-2);
	}

    }	/* end of edge-moved-down case */
    else
    {
	    /* moving immediately to the right found white.  Look up for
	     * black.  This end conditions looks silly, but it's
	     * what the assembly code does and we're trying to be
	     * consistent.
	     */
	for (lineOffset = 1, srcOffset = -imBpl;
	     ((lineOffset < maxVertSearch) && (!sample));
	     lineOffset++, srcOffset -= imBpl)
	{
	    sample = USHORT_ACCESS(pImage+srcOffset);
	}

	    /* we are now either 1 line above the black/white boundary or
	     * have found white space. */
	if (!sample)	/* in white space */
	{
	    *ppImage = (UInt8Ptr) (pImage);
	    (*pHCount)++;
	    (*pCurrentSpaces)++;		/* found one more white space */	    
	     return(cBigStepValue+1);		/* end of run */
	}
	else
	{
		/* We found a black/white edge.  Because of the loop structure,
		 * lineOffset and srcOffset point one line too high.
		 * Update *ppImage, *pHCount, *pVCount and return the number
		 * of lines we skipped. */
	    *ppImage = (UInt8Ptr) &pImage[srcOffset + imBpl];
	    (*pHCount)++;
	    *pVCount -= lineOffset-1;
	    *pCurrentSpaces = 0;
	    return(lineOffset-1);
	}
    }	/* end of edge-moved-up case */

#ifdef __GNUC__
    NORMAL_RETURN; /* WE NEVER GET HERE, this dummy is needed so gcc compiler
		 * won't complain that there's no return.  Unfortunately,
		 * acc *will* complain if this is here, so only put it in
		 * for gcc.  Blah. */
#endif

}

/*

$Log: miskew.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.9  1998/01/13 23:53:28  ddavies
miskew.c -> /product/ipcore/ipshared/src/RCS/miskew.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1995/12/13 02:52:20  ejaquith
miskew.c -> /product/ipcore/ipshared/src/RCS/miskew.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.7  1995/02/24  02:33:29  ddavies
 * miskew.c -> /product/ipcore/ipshared/src/RCS/miskew.c,v
 *
 * Change ip_findDiffPixelsSum so that it computes the total score into a
 * double word instead of into a single 32-bit word.  This is used to
 * prevent overflows.  This routine assumes that there are at most 127
 * horizontal samples per line, which is guaranteed by ip_detectSkewFine.
 *
 * Revision 1.6  1994/04/03  02:06:08  ddavies
 * miskew.c -> /product/ipcore/ipshared/src/RCS/miskew.c,v
 *
 * Incorporate pointer types for use in SantaFe project.
 *
 * Revision 1.5  1994/02/04  20:19:31  dferrell
 * miskew.c -> /product/ipcore/ipshared/src/RCS/miskew.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:25:30  ddavies
 * miskew.c -> /product/ipcore/ipshared/src/RCS/miskew.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:50:58  ddavies
 * miskew.c -> /product/ipcore/ipshared/src/RCS/miskew.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

 */


