/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  mitaghrn.c: machine independent routines for tagging each pixel
 *		with the length of the horizontal run containing that
 *		pixel.
 *
 *          gray comparison:
 *                  void	ip_tagHorizontalRunsLine()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#define _USE_FIRST_BIT_TABLE_		/* only want firstBitTable from
					 * iopbyte.prv */
#include "iopbyte.prv"

/* prototypes */
#include "shrcc.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: mitaghrn.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* During page segmentation it is useful to identify the white gutters
 * separating areas of text.  Every pixel in a gutter is in the middle of
 * both a horizontal run and a vertical run.  If the gutter is mainly
 * vertical, say separating two columns of text, the vertical run is
 * significantly longer than the horizontal run.  We would like to be
 * able to tag each pixel with the lengths of the horizontal and vertical
 * runs containing it so we can decide whether the gutters containing them
 * are horizontal or vertical.
 *
 * ip_tagHorizontalRunsLine does the low level work of tagging each pixel
 * with the length of the horizontal run containing it.  It is called once
 * per line by i_tagHorizontalRuns in i_tagrun.c.
 *
 * How do we decide whether we're talking about white runs
 * or black runs?  I don't know, probably invert the incoming binary image.
 * This routine performs the same function as the inner loop of
 * HorizGranulometry (the version that's not commented out) in XIS's
 * giputils.c.  That routine doesn't conditionally invert the input image,
 * so we won't either.
 *
 * The inputs to this routine are pointers to a line of a binary image,
 * a pointer to a line of a gray image and the width of both images in
 * pixels.  The binary image contains the pixels of interest.  When we
 * return, the gray image line holds the output tag values.  This is a bit
 * different than XIS's HorizGranulometry.  That routine takes a single
 * gray image as both input and output.  Before HorizGranulometry is called,
 * the routine BitmapToImage creates the gray image, setting each pixel 
 * to 0x00 or 0x01, depending on the value of the corresponding binary pixel
 * (0 => 0x00, 1 => 0x01).  We will create the final gray image directly
 * from the binary image.
 *
 * The gray image always has 8 bits/pixel.  What do we do with runs longer
 * than 255 pixels?  The XIS routine simply records the run length as
 * 255.  We'll follow the same convention.
 *
 * The XIS routine did two passes through each line of their gray image.
 * In the first one, each pixel was tagged with the number of contiguous
 * non-zero pixels to its left + 1.  In the second pass, the sum at the
 * right end of each run was copied back to all the pixels in the run.
 *
 * This is a very nice, clear algorithm.  We will try to eliminate
 * some steps.  First, we don't translate the binary image into a gray image
 * containing one interesting bit in each pixel before starting.  This would
 * cause us to write the image one extra time.  Second, we do each line in
 * one pass, not two.  The combination saves us two passes through the
 * gray image.
 *
 * Assume we're starting in a region of no run (pixels are 0).  We look ahead
 * until we find the left end of the next run.  We fill all the corresponding
 * gray pixels from the current position to that position with 0.  We then
 * look ahead to find the right end of the run.  We fill all gray pixels
 * between the current position and the new position with Min(runLength, 255).
 * This process continues to the end of the line.
 *
 * This is harder to make fast than I thought when I started.  The tricks
 * revolve around reading 32 binary pixels at a time and writing 4 gray
 * pixels at a time.
 *
 * Why read 32 pixels at once?  It's faster than reading some smaller number..
 * So long as the they are all 0 or all 1 (depending on
 * whether we're looking for the start or the end of a run), we can continue.
 * When we find a break, we isolate it down to a byte within the word,
 * then use table lookup to find the end of the run.
 *
 * We'd like to write the output image in units of 4 bytes, not 1. This is
 * easy in the middle of big runs, what do we do at the ends?  We keep the
 * word to next be written in a variable.  Usually, it will contain the
 * end of the last run.  It will contain nothing if the last run ended on
 * a word boundary.  If the next run fits entirely into the word and
 * does not fill it, we just OR it in and continue.  If the next run crosses
 * into the next word of the gray image, we write the accumulated output word
 * before filling it with new data for the next output word.
 *
 * That's the easy part.  Things get more complex when we deal with
 * addresses and run lengths.  There are several considerations:
 *  1.	The number of variables should be kept at a minimum.  The more
 *	variables, the more confused compilers get.
 *  2.	The number of operations done in the inner loops should be
 *	minimized.  In particular, we would like to have the end
 *	condition of loops depend on addresses that are being
 *	incremented anyway, not on extraneous loop counters.
 *  3.	We would like to read the source and write the destination
 *	1 word at a time.  There are 32 pixels in a binary wod, but
 *	only 4 pixels in a gray image word.
 *  4.	A result of statements 2 and 3 is that we would like to
 *	check addresses only at 32-bit boundaries.
 *  5.	We assume both source and destination image lines are padded to
 *	32-bit boundaries.
 *  6.	As a result of statements 4 and 5, we would like to write the
 *	pad bits of the destination image.
 *  7.	One may add as many as 31 pixels to pad the
 *	source binary image, but only a maximum of 3 pixels to pad
 *	the destination gray image.  Therefore, the number of pixels
 *	to be written in the padded gray image does not necessarily
 *	equal the number of pixels in the padded binary source image.
 *	Therefore, we must keep the widths of the two padded images
 *	in separate variables.  Note that we need only check to see if
 *	we're going to run off the end of the gray line when we're writing 0's.
 *  8.	There are at least two ways to compute run lengths.  We can
 *	have a run length variable that is incremented as we go along.
 *	We could also compute it from the addresses and bit offsets.
 *	The latter scheme is nice in that we don't have to increment
 *	a run length variable in the inner search loop.  After a false
 *	start or two, I finally figured out that we need to calculate two
 *	interrelated, but distinct, quanties - the address of the gray
 *	image word where the run ends and the number we put into each pixel.
 *	The final address must account for the number of pixels computed so
 *	far (especially those in the output word that haven't been written
 *	yet) as well as the number of pixels in the current run.  The number
 *	we write into the pixels is only a function of the length of the
 *	current run.  Say, for example, the current output word already has
 *	3 valid pixels and we need to write a run of 10 pixels.  We should
 *	put 1 in the current word, 4 in each of the next two words and have
 *	1 left over for the next output word to be computed.  This advances
 *	the gray output address by 3 words.  Originally, I was just adding
 *	10 to the gray output address, (which would only advance by 2)
 *	because I had forgotten about the valid pixels that are already in the
 *	output word before we start writing the current run.
 *
 *	The latest algorithm computes the gray address from the binary address
 *	and the run length from both addresses AND the number of pixels
 *	already in the output word when we start.
 *	Say the address of the leftmost byte in the binary line is binStartAddr
 *	and the address of the corresonding byte in the gray line is
 *	grayStartAddr.  When we've found the end of a binary run, the address of
 *	the word containing the end of the run is pB.  The offset of the end
 *	of the run in the binary word is bitOffset.  The address
 *	of the next gray *byte* to be written is pG.  While writing gray
 *	pixels, pG contains word addresses, but while searching for binary
 *	pixels, we keep the pixel address itself in pG to save variables.
 *
 *	Here's a picture in which +'s represent word boundaries.  Note there
 *	are more pixels in each binary word than in each gray word.  The
 *	picture is meant to show the state at the end of a search for a
 *	binary run, before the corresponding gray run has been filled in.
 *	
 *	binStartAddr			     pB 
 *		|----+----+----+----+----+----|---|-+----+----+----+
 *					      |<->| bitOffset
 *		|-+-+-+-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *	grayStartAddr		     pG
 *
 *	The scheme depends on the fact that we will eventually write one
 *	gray pixel for every binary pixel.  Therefore, the *byte* address of
 *	the end of the gray run equals
 *	   finGrayRun = (8 * (pB - binStartAddr) + bitOffset) + grayStartAddr
 *	The word address is
 *		= ((8 * (pB - binStartAddr) + bitOffset) + grayStartAddr) & ~3
 *	The run length is equal to
 *	   run length = (8 * (pB - binStartAddr) + bitOffset) - (pG -
 *								 grayStartAddr)
 *	Note that the first portion of this calculation is the same as the
 *	first portion of finGrayRun.  There are obvious economies.
 *  9.	We need to correctly position the pixels at the start and end 
 *	of a run in the output word.  Their position is a function of their
 *	address MOD 4.  This is equal to the binary bit offset MOD 4.  That's
 *	true because the binary bit offset is calculated with respect to
 *	a binary pixels whose address equals 0 MOD 4 with respect to the
 *	beginning of both the binary and gray scan lines.
 *
 *	The algorithm proceeds as follows:
 *
 *	A. Find the word containing the end of the current run.  pB
 *	   contains the address of this word.  The variable oMask
 *	   contains a mask used to select the invalid pixels.  pG points to
 *	   the first invalid byte in oWord.
 *	B. Set
 *	   finGrayRun = (8 * (pB - binStartAddr) + bitOffset) + grayStartAddr.
 *	C. If this is a run of 0's, set
 *	   finGrayRun = MIN(finGrayRun, finGrayLine) and skip steps D, E,
 *	   F and G.
 *	D. Set newOWord = finGrayAddr - pG.  This is the raw runLength.
 *	E. Set newOWord = MIN(255, newOWord).  This is the corrected 
 *	   run length.
 *	F. Move this number into all 4 pixels of newOWord.
 *	G. Mask newOWord with the mask computed in step B above and OR the
 *	   result into oWord.  This inserts the pixel at the left end of
 *	   the run into oWord.
 *	H. Convert pG and finGrayRun to word addresses by ANDin both with ~3.
 *	I. If the run extends beyond the output word represented by oWord,
 *	   write oWord to memory, increment pG by 4 and set oWord to newOWord.
 *	J. While there are more complete words to fill in the run, write them
 *	   with oWord.
 *	K. If the run ends in the middle of a gray word, calculate a mask
 *	   from the LS two bits of bitOffset of the end of the run
 *	   (see below).  Use this mask to set the rightmost bits of oWord to 0.
 *	L. In keeping with our convention, set the LS two bits of pG to the
 *	   two LS bits of the bitOffset.  Complement the mask in oMask so
 *	   we're ready for the next run.
 *
 * What about the mask at the right end of the run?  The LS two bits of
 * the bit offset gives the distance from the left end of the word to the
 * first pixel NOT in the run.  The mapping is:
 *
 *	BitOffset.[1:0]		right gray pixel mask
 *		00		0x00000000
 *		01		0xFF000000
 *		10		0xFFFF0000
 *		11		0xFFFFFF00
 *
 * By no accident, the mask used to insert new pixels is the complement of
 * this mask.
 *
 * How do we calculate the mappings?  A table lookup involves a mask, a shift
 * and computing the address of the base of the table and a read.  We
 * can also use 0xFFFFFFFE << (31 - BitOffset.[1:0] * 8).  The
 * arithmetic for this may be marginally easier than calculating
 * the address of the base of the table and putting more stuff in the cache,
 * so we'll use it.
 *
 * One more note - instead of keeping two constant values, startGAddr and
 * startBAddr, why don't we just keep startGAddr - 8*startBAddr?  That's the
 * thing we always use anyway, right?  Unfortunately, we don't know if any
 * interested address bits disappeared when we did the "8*".  If they did,
 * the number calculated will be trash.  Instead, we get the difference
 * between pB and startBAddr first, doing the 8* second.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_tagHorizontalRunsLine()  
 *		Takes pointers to a scan line of a binary image and a
 *		scan line of a gray image.  Records
 *		MIN(255, horizontalRunLength) in each pixel of the gray
 *		image.
 *
 *	Input:
 *		pB =		pointer binary source
 *		pG =		pointer to gray dest image
 *		w =		width of both images in pixels.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_tagHorizontalRunsLine (
    IPState    *pIPState,
    UInt8      *pB,
    UInt8      *pG,
    UInt32	w)
{
    PROLOGUE;
    UInt32	 oWord;			/* the gray output word */
    UInt32	 iWord;			/* the binary input word */
    UInt32	 newOWord;		/* contains 4 copies of the pixels
					 * used to fill in the next run. */
    UInt8	 bitOffset;		/* byte containing bit offset of
					 * the start or stop of a run relative
					 * to the MS bit of a 32-bit word. */
    UInt32	 startGAddr, startBAddr; /* starting addresses, used to compute
					 * run addreses and lengths. */
    UInt32       finGAddrRun;		/* the rightmost gray address for
					 * the current run + 4.
					 * This is a word address.  We stop
					 * before writing this address.
					 */
    UInt32	 finGAddrLine;		/* the rightmost gray address for the
					 * line + 4.  We stop before writing
					 * this address.
					 */
    UInt32	 finBAddrLine;		/* the address of the rightmost word
					 * in the binary image line + 4.
					 * Filling stops when this address
					 * is reached.
					 */

	/* keep the compiler happy */
    __error__ = __error__;
    pIPState = pIPState;

	/* set the address of the first word beyond the right end
	 * of the binary and gray lines.  We will stop just before processing
	 * this word.
	 */
    startGAddr = (UInt32) pG;		/* where we start */
    startBAddr = (UInt32) pB;
    finBAddrLine = (UInt32) pB + ((w+31)/32) * 4;
    finGAddrLine = (UInt32) pG + ((w+3)/4) * 4;

	/* start by assuming the first run is 0's.  In the loop, we'll
	 * alternatively process runs of 0's and 1's until we reach
	 * finBAddr while processing 0's.  Note that pG.[1:0] already has the
	 * number of valid pixels in oWord (0).
	 */
    oWord = 0;
    iWord = ULONG_ACCESS(pB);		/* get first binary word */

	/* Enter the main loop.  We keep reading binary stuff until we
	 * reach the end of the line.
	 * In the loop, we alternately look for the end of a run of 0's
	 * (the start of a run of 1's) and look for the end of a run
	 * of 1's.  When the run of 0's is found, we fill the corresponding
	 * pixels in the gray image with 0's.  When the end of the run
	 * of 1's is found, we fill the gray pixels with MIN(255, runLength)
	 */
    while ((UInt32) pB < finBAddrLine)
    {

	    /* look for end of a run of 0's */
	while ((iWord == 0) && ((UInt32) pB < finBAddrLine))
	{
	    pB += 4;
	    iWord = ULONG_ACCESS(pB);
	}

	    /* We have either found the start of a non-zero run or the end
	     * of the line.  Either way, we want to fill everything from the
	     * current value of pG to the place we've stopped with 0's.
	     * Calculate the address of the word containing the
	     * right end of the run.  This word will not be
	     * written until oWord has the right stuff in it.
	     * To do this, first find the end of the run within the
	     * binary word.
	     */

	if (iWord & 0xFFFF0000)
	{	/* 1's in top half */
	    if (iWord & 0xFF000000)
	    {
		    /* There are 1's in the top byte */
		bitOffset = firstBitTable[iWord >> 24];
	    }
	    else
	    {
		    /* 1's start in byte 1 */
		bitOffset = firstBitTable[iWord >> 16] + 8;
	    }
	}
	else	/* 1's in bottom half of word */
	{
	    if (iWord & 0x0000FF00)
	    {
		    /* 1's start in byte 2 */
		bitOffset = firstBitTable[iWord >> 8] + 16;
	    }
	    else
	    {
		    /* 1's start in byte 3 */
		bitOffset = firstBitTable[iWord] + 24;
	    }
	}

	    /* calculate the final *byte* address of the run */
	finGAddrRun = 8 * ((UInt32) pB - startBAddr) + bitOffset + startGAddr;

	    /* get the word addresses and make sure we don't go off the
	     * end of the line */
	finGAddrRun = MIN((finGAddrRun & ~3), finGAddrLine);
	pG = (UInt8 *)((UInt32) pG & ~3);

	    /* Fill the run with 0's */
	if ((UInt32) pG < finGAddrRun)	/* finish this word */
	{
	    ULONG_ACCESS(pG) = oWord;
	    pG += 4;
	    oWord = 0;
	}
	    /* fill the rest of the run */
	while ((UInt32) pG < finGAddrRun)
	{
	    ULONG_ACCESS(pG) = oWord;
	    pG += 4;
	}

	    /* If pB >= finBAddrLine, we finished because we hit the end of the
	     * line.  We've filled the line with 0's, so quit.
	     */
	if ((UInt32)pB >= finBAddrLine)
	    break;

	    /* Now we find the end of the run of 1's.  We set the
	     * the top bitOffset bits of iWord with 1's so we don't
	     * mistake those 0's for the end of the run.
	     * The shift count is a little funny.  Ideally, we want
	     * to shift a word filled with 1's left by (32 - bitOffset).
	     * However, if bitOffset = 0, this turns into all 1's
	     * instead of all 0's on 386, 486 and SPARC platforms because
	     * the shifter only looks at the least significant 5 bits
	     * of the shift count.  So, we use (31 - bitOffset) instead
	     * and take a 1 off the right end of the shift constant.
	     */
	iWord |= 0xFFFFFFFE << (31 - bitOffset);


	    /* Put the number of valid pixels remaining in oWord into
	     * the LS two bits of pG.
	     */
	pG = (UInt8 *) ((UInt32) pG | (bitOffset & 3));

	    /* look for end of a run of 1's.  Note that we don't have
	     * to check for the end of the line since frames are filled
	     * with 0's, so that will stop us.
	     */
	while (iWord == 0xFFFFFFFF)
	{
	    pB += 4;
	    iWord = ULONG_ACCESS(pB);
	}
    
	    /* Calculate the real run length by finding the byte
	     * within iWord containing the first 0's.  This is
	     * easier to do if we invert iWord and look for 1's.  This
	     * Means we can use the same code and the same tables
	     * as were used above.
	     */
	iWord = ~iWord;

	if (iWord & 0xFFFF0000)
	{	/* 1's in top half */
	    if (iWord & 0xFF000000)
	    {
		    /* There are 1's in the top byte */
		bitOffset = firstBitTable[iWord >> 24];
	    }
	    else
	    {
		    /* 1's start in byte 1 */
		bitOffset = firstBitTable[iWord >> 16] + 8;
	    }
	}
	else	/* 1's in bottom half of word */
	{
	    if (iWord & 0x0000FF00)
	    {
		    /* 1's start in byte 2 */
		bitOffset = firstBitTable[iWord >> 8] + 16;
	    }
	    else
	    {
		    /* 1's start in byte 3 */
		bitOffset = firstBitTable[iWord] + 24;
	    }
	}

	    /* calculate the final *byte* address of the run */
	finGAddrRun = 8 * ((UInt32) pB - startBAddr) + bitOffset + startGAddr;

	    /* Calculate the run length */
	newOWord = MIN(finGAddrRun - (UInt32) pG, 255);

	    /* At this point, newOWord contain the number we are going to
	     * put into all pixels of the gray image corresponding to the
	     * run of binary pixels we've found.  finGAddrRun points to
	     * the first gray word that should not be written.
	     * Replicate the value in the bottom byte of newOWord into
	     * all it's bytes, then insert the relevant bits
	     * into oWord.  This is the value to be written at the left
	     * end of the run.
	     */
	newOWord |= newOWord << 8;
	newOWord |= newOWord << 16;

	oWord |= newOWord & ((UInt32) 0xFFFFFFFF >> (((UInt32) pG & 0x3) * 8));

	    /* get the word addresses.  We can't run off the end of
	     * the line when following a run of 1's, so don't bother
	     * checking.
	     */
	finGAddrRun &= ~3;
	pG = (UInt8 *) ((UInt32) pG & ~3);

	    /* If the run extends beyond the word currently pointed to
	     * by pG (i.e. all the pixels in oWord are valid now), write
	     * oWord and reload it with newOWord.  The new value will
	     * be written into all further words in the run.
	     */
	if ((UInt32) pG < finGAddrRun)
	{
	    ULONG_ACCESS(pG) = oWord;
	    pG += 4;
	    oWord = newOWord;
	}

	    /* Write the rest of the full words in this run */
	while ((UInt32) pG < finGAddrRun)
	{
	    ULONG_ACCESS(pG) = oWord;
	    pG += 4;
	}

	    /* Set all the 0's we've found at the left end of iWord and
	     * re-invert the result.  That's to make sure we don't find
	     * them again when looking for the leftmost string of 1's
	     * that hasn't been seen yet (i.e. the right end of the next
	     * run of 0's).
	     */
	iWord = ~(iWord | (0xFFFFFFFE << (31 - bitOffset)));

	    /* Calculate the mask used for the right end of the run.
	     * We use the bit offset of the end of the run since it's
	     * computed relative to a location that is 0 MOD 4
	     * (the start of a binary byte) relative to the left end
	     * of the line.  Also put the number of valid pixels remaining in
	     * oWord into the LS two bits of pG.
	     */
	bitOffset &= 3;
	pG = (UInt8 *) ((UInt32) pG | bitOffset);
	oWord &= (0xFFFFFFFE << (31 - (8 * bitOffset)));

    }	/* the end of the loop that finds a run of 0's followed by a
	 * run of 1's */
}

/*

$Log: mitaghrn.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.9  1998/01/13 23:53:29  ddavies
mitaghrn.c -> /product/ipcore/ipshared/src/RCS/mitaghrn.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1997/04/17 02:37:22  ddavies
mitaghrn.c -> /product/ipcore/ipshared/src/RCS/mitaghrn.c,v

Fix bug in check for end of line after a run of 0's has been found.
We now see if pB >= finBAddrLine instead of seeing if iWord==0.  The
second case depends on finding an entire word of 0's to the right of
the image, which may not happen (pad can be non-zero).

Revision 1.7  1997/04/15 19:39:18  ddavies
mitaghrn.c -> /product/ipcore/ipshared/src/RCS/mitaghrn.c,v

Fix bug in ip_tagHorizontalRunsLine.  The address of the location beyond
the end of the line is pB + ((w+31)/32) * 4, not pB + ((w/32) + 1) * 4.
Ditto for the end of the gray line.

Revision 1.6  1995/12/13 02:52:25  ejaquith
mitaghrn.c -> /product/ipcore/ipshared/src/RCS/mitaghrn.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/01/11  22:33:53  dferrell
 * mitaghrn.c -> /product/ipcore/ipshared/src/RCS/mitaghrn.c,v
 *
 * Changed shrip.prv to shrcc.prv.
 *
 * Revision 1.4  1994/02/04  20:19:37  dferrell
 * mitaghrn.c -> /product/ipcore/ipshared/src/RCS/mitaghrn.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:25:35  ddavies
 * mitaghrn.c -> /product/ipcore/ipshared/src/RCS/mitaghrn.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:51:06  ddavies
 * mitaghrn.c -> /product/ipcore/ipshared/src/RCS/mitaghrn.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


