/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  mitagvrn.c: machine independent routines for tagging each pixel
 *		with the length of the vertical run containing that
 *		pixel.
 *
 *          gray comparison:
 *                  void	ip_calculateVerticalRunLengths()
 *		    void	ip_tagVerticalPixels()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#include "nibmask.prv"

/* prototypes */
#include "shrcc.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: mitagvrn.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */


/* During page segmentation it is useful to identify the white gutters
 * separating areas of text.  Every pixel in a gutter is in the middle of
 * both a horizontal run and a vertical run.  If the gutter is mainly
 * vertical, say separating two columns of text, the vertical run is
 * significantly longer than the horizontal run.  We would like to be
 * able to tag each pixel with the lengths of the horizontal and vertical
 * runs containing it so we can decide whether the gutters containing them
 * are horizontal or vertical.
 *
 * Finding vertical runs is quite different than finding horizontal runs.
 * When finding horizontal runs, we can look ahead down a horizontal line,
 * some contiguous portion of which is in each word we read.  Consecutive
 * words tend to fall in the same cache line so the cache is used
 * efficiently.  Image bands tend to be horizontal, so it's fairly
 * straightforward to process the image by bands.  Compressed images
 * tend to compress scan lines (possibly with respect to previous scan lines),
 * so we can walk down the scan line in a straightforward manner.  Life
 * is easy.
 *
 * One might say that the grain of images is horizontal.  Moving vertically
 * goes against the grain.  It can certainly be done, but it's harder.
 * The caches are used inefficiently, bands fall apart and it's a real
 * battle to go through compressed images.
 *
 * For these reasons, we will break one of our cardinal rules: we will write
 * the output gray image twice.  Even if we ignore bands and compression,
 * taking a cache fault on every word read is a lot more expensive than doing
 * an extra write for every word read.  If there is some chance that the
 * image doesn't fit in memory, so is paged, then reading vertically will
 * thrash the poor disk out of its box.
 *
 * How do we process the image horizontally if we're looking for vertical
 * runs?  We use a modified version of the Luc Vincent's original scheme
 * implemented in the routine VertGranulometry in giputils.c.  We work
 * from left to right along a scan line.  We assign each gray pixel the
 * value of the pixel above it + 1 unless a) the corresponding binary
 * pixel is 0, or b) the gray pixel already has a value of 255.  If the
 * corresponding binary pixel is 0, we assign the gray pixel a value of 0.
 * If the gray pixel above this one is already 255 (and the corresponding 
 * binary pixel is 1), we set this pixel to 255 also.  This job is done by
 * ip_calculateVerticalRunLengths.  This routine is called once per scan
 * line by i_tagVerticalRuns.  It processes the image from top to bottom.
 *
 * When the first top-to-bottom pass through the image is complete,
 * i_tagVerticalRuns calls ip_tagVerticalPixels from bottom to top.  This
 * routine copies the lowest (largest) count in each vertical run on top
 * of the other counts of that run, tagging each pixel in the run with
 * MIN(runLength, 255).
 *
 * So far, except for the fact that we process the image a scan line at
 * a time instead of a vertical column at a time, this sounds just like
 * Luc's original scheme.  The trick is making it run fast.  This
 * means dealing with the image a word at a time instead of a pixel
 * at a time.  Each gray word holds pixels from 4 independent columns.  Each
 * binary word holds pixels from 32 independent columns.  In most cases,
 * we can't use the processor's normal tests and branches to decide what to
 * do since the results of any given operation may be different for
 * different columns.
 *
 * Let's start with ip_calculateVerticalRunsLengths.  I thought of several
 * exotic schemes.  For example, wouldn't it be cool to just write the final
 * run length into the image without bothering to write all the intervening
 * junk?  Ummm.  There are two things we need to know after the first pass
 * through the image.  The first is the number we will write into each pixel
 * (MIN(runLength, 255)).  The other is the starting and ending locations
 * of each run.  Say we wrote the pixel value at the bottom of each run and
 * nothing else.  Since reals runs can be longer than 255, we wouldn't know
 * how far to back up.  Well, ok, you could write the pixel value, and if it
 * was 255, write the real run length in the two bytes above it.  Decoding
 * a word full of staggered runs is messy, and as we shall see below,
 * writing it is no faster than just writing all the counts into the image.
 * Say we write the pixel value at the top and bottom of each
 * run.  Cool - to fill in the runs all we would have to do is XOR the value
 * in the image with the current value for the column and write it.  Ooops,
 * what's the current value for the column?  It's another line that's the
 * same width as the image that's separately maintained?  Doesn't that double
 * the number of writes required?  Ouch.  Besides, this scheme implicitly
 * assumes that the space in the middle of the runs is filled with 0's.  How
 * did that happen?  If it happened in the first pass, then we did all that
 * writing anyway PLUS keeping the run lengths somewhere else - again doubling
 * the amount of writing.  Oh, and we'd have to remember the start of each
 * run in each column so we could write the run length there when we found
 * the end of the run.  And this doesn't work for runs of length 1.  No good.
 *
 * Well, what about filling the gray image with some other constant, like 0xFF
 * wherever there's a run and only writing the run length in at the end?
 * This allows runs to be of length 1 and doesn't require remembering where
 * the run starts, right?  Yes, but we still require two writes for each
 * gray pixel (the constant and the accumulating run length) on the way down.
 * When we're copying run lengths back over all the constants, does the fact
 * that they're constants make things significantly easier?  I can't find a way
 * to take advantage of it, especially enough advantage to offset the extra
 * write.
 *
 * Ok, what about just writing the count at the end of the run and leaving
 * the rest of the gray image alone?  In this scheme, we'd use the binary
 * image to define the runs both on the way down and on the way back up.
 * Cool, we get to avoid writing the most of the gray image pixels on the
 * way down, right?  Wrong.  While it's true that we don't have to write
 * the pixels in the gray image itself, we still have to have a private scan
 * line buffer that accumulates the run lengths.  We still have to write
 * something to into that buffer for every pixel in the image.  So, instead
 * of doing a write for every word, we do a write for every word PLUS a
 * write at the end of every run.  Still a bummer.  This is the same
 * problem found with writing both the pixel value and the real run length into
 * the gray image on the way down.
 *
 * Yeah, but wouldn't this go faster if we could somehow get rid of the
 * check for 255?  Since we're keeping the pixel values in bytes, we can't
 * let them wrap around, so they can be allowed to get no bigger than
 * 255.  Seems like a pain to check every pixel count to see if it's too
 * big.  One way to fix it is to have a separate buffer that contains the
 * current maximum pixel value for some group of pixels.  We would increment
 * this pixel value on every scan line.  When it got to 255, we'd check
 * every pixel in the scan line.  If any of them stayed at 255, we'd keep the
 * count at 255.  Otherwise, we'd set it to the new maximum.  Doing this would
 * save us the trouble of checking each pixel at the cost of an extra write
 * for every n pixels plus the overhead of finding the maximum of the group
 * of pixels.  Hmmm.  Since I wrote this comment after writing the first
 * draft of the code, I scurried down to see what I could get out of it.
 * On the SPARC it adds significant complexity and the performance is almost
 * breakeven (26 avg cycles/gray word vs 27 without it).  On the PC, we might
 * get a 25% performance improvement.  Ummm.  Maybe I'll wait on this.
 *
 * Another way to avoid checking each pixel to see if it's already 255 is
 * to have a private scan line has a 16 bit number for every
 * column.  Well, then we essentially write twice as much stuff on the first
 * pass through the image.  This is not fun.  The scan line would be about
 * 5KB on a full size 300 dpi image, larger than the cache on a 486, so we
 * cache fault on every 32-byte block.  We have to check each pixel to see
 * if it's the end of the run (and if so, do the MIN function and write the
 * appropriate byte).  All this adds up.  We'll stick with just writing the
 * write set of pixels to the image a word at a time.
 *
 * Ok, we just write the accumulating counts into the gray image on the
 * way down.  How?  One way is to look at each gray pixel to see if it's below
 * 255, look at each binary pixel to see if it's a 1 and increment the
 * gray pixels that pass both tests.  This is a little slow.  We can do better
 * by incrementing all pixels, decrementing the pixels that wrapped around to
 * 0, then blindly clearing the gray pixels that correspond to 0's in the
 * binary image.  This avoids making decisions.
 *
 * What's that trick with the incrementing and conditionally decrementing
 * again?  The basic point is that it's easier to figure out whether we
 * wrapped around than to decide if a pixel equals 0xFF.  We add
 * 0x01010101 to the word containing the gray pixels.  We AND the complement
 * of the result with the original gray word.  Every pixel that went from
 * 1 to 0 is now set.  We will say that a pixel was initially 0xFF if BOTH
 * the MS bit and the LS bit went from 1 to 0.  The MS bit is obvious.  However
 * if the LS bit was originally 0 (pixel was 0xFE), the entire pixel could
 * still wrap around to 0 because of the carry from a pixel to the right.
 * We detect this by checking that both the MS and LS bits flipped.  Anyway,
 * we shift the result of the AND right by 7 and AND it with the unshifted
 * version.  This puts a 1 in the LS bit of each pixel that really
 * wrapped around, and junk elsewhere.  We AND this with 0x01010101 to get
 * rid of the junk and subtract it from (iWord + 0x01010101), which should
 * still be around somewhere if the compiler isn't bonkers.  It turns out
 * this is an opcode or two shorter than identifying all the pixels that
 * are not equal to 0xFF and incrementing them.  At least I haven't found
 * a better sequence of opcodes that does that.
 *
 * After all that stuff, we have a word full of pixels that have been
 * incremented.  If all pixels are in the middle of runs, this is the right
 * word.  We choose the correct nibble from the binary image, expand it into
 * a 32-bit mask that has 0xFF for each 1 in the nibble and 0x00 for each 0.
 * We AND the mask with the word of pixels.  All pixels that are not in
 * runs are cleared.  The result is written to the output image.  If there were
 * some guarantee that all incoming gray pixels were already equal to 0x00, we
 * could skip the write if the output word were 0.  Guaranteeing this
 * involves writing the whole image.  We can skip initializing the image
 * if we do the write, so we will.
 *
 * That's how we go down the image.  What do we do on the way up in
 * ip_tagVerticalPixels?  We start on the next-to-bottom scan line.  There
 * are two main approaches.  We could look at the binary image to decide where
 * the runs are or look at the gray image.  Looking at the binary image has
 * the advantage that the information is very compact.  Looking at the gray
 * image has the advantage of minimizing the number of variables and pointers
 * we have to deal with as well as minimizing cache contention (dealing
 * with one image, not two).  We need to decide exactly what we need to
 * do before we can decide how to do it.  This is obvious, but I spent a lot
 * of time trying to optimize buggy algorithms before seeing the light.
 *
 * We can be in one of four states in each column: 1) in the middle of a
 * run of 0's => put out 0's or don't bother writing; 2) at the bottom of
 * a run of 1's => don't destroy the run length; 3) in the middle of run =>
 * do copy the run length in the line below to the current line; 4) at
 * the top of a run => stop copying the run length.
 *
 * This can be simplified a bit.  At the ends of runs, we need to avoid
 * replacing the current pixel with the one below it.  This
 * both preserves the run length at the bottom of each column and stops
 * the run length from being copied at the top of the run.  In the middle of
 * runs, we must do the replacement if we're in the middle of a run of 1's
 * and may or may not copy if in the middle of a run of 0's.
 *
 * Say we detect the case when either the pixel in the current line or the
 * pixel in the previous line equals 0.  If this condition is true, we
 * are either at the top or bottom of a run or in the middle of a run
 * of 0's.  In all three cases, we can avoid copying pixels.  If this
 * condition is not true, we are in the middle of a run of 1's, so we must
 * replace the current pixel with the one below it.  Not too bad.
 *
 * At this point, I spent a lot of time trying to decide whether to look at
 * the binary pixels or the gray pixels to find the runs.  It's really
 * tempting to use the gray image to minimize the amount of stuff.
 * However, as hard as I tried, I couldn't find an algorithm that beats
 * looking at the binary image to determine run boundaries.  The binary
 * data is 8 times as compact, so every binary word can be used to process
 * 8 gray words.  Oh well.
 *
 * We read a binary word and the word below it and AND them.  The resulting
 * word has 0's where the corresponding gray pixels should be left alone
 * and 1's where gray pixels should be replaced by the pixels below them.
 * We enter the inner loop, which is unrolled, that repeatedly shifts a
 * nibble into position from the binary word, expands it into a mask.
 * We AND this mask with the gray word below, AND the inverse of the
 * mask with the current gray word, OR the results and write to the
 * current gray word.
 *
 * Whew.  When you actually see the code, it's not as awful as it sounds.
 * So far as I can tell, it is definitely faster than dealing with each
 * pixel individually - not 4 times as fast, but about twice as fast.
 */

/* PRIVATE */
/**************************************************************************
 *  ip_calculateVerticalRunLengths()  
 *		Takes pointers to a scan line of a binary image and a
 *		gray image.  If corresponding binary pixel == 1, writes
 *		MIN(255, gray value above current pixel + 1) to each gray
 *		pixel.  If corresponding binary pixel == 0, write 0 to
 *		gray pixel.
 *		This function is used by i_tagVerticalRuns to tag
 *		each pixel with the length of the vertical run containing
 *		it.
 *
 *	Input:
 *		pB =	pointer binary source
 *		pG =	pointer to gray dest image
 *		gBpl =	gray dest's bytes/line
 *		w =	width of both images in pixels.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_calculateVerticalRunLengths (
    IPState    *pIPState,
    UInt8      *pB,
    UInt8      *pG,
    Int32	gBpl,
    Int32	w)
{
    PROLOGUE;
    UInt32	 oWord;			/* the gray output word */
    UInt32	 iNibble;		/* binary nibble corresponding to
				         * the current gray word. */
    UInt32	 iBWord;		/* the binary input word */
    UInt32	 iGWord;		/* the gray input word */
    UInt32	 finGAddrLine;		/* the rightmost gray address for the
				         * line + 4.  We stop before writing
				         * this address.
				         */
    const UInt8	*nibMaskBytePtr = (const UInt8 *)nibMask; /* try to discourage
					 * the compiler from doing
					 * unnecessary shifts when computing
					 * addresses.  This is a pointer to
					 * an array of 16 masks.  Each mask
					 * expands every bit in the addressing
					 * nibble by 8.
					 */
	/* Keep the compiler happy. */
    pIPState = pIPState;
    __error__ = __error__;

	/* We want to address the line above this one, but the
	 * normal addressing modes only allow ADDING a displacement.
	 * The compiler could figure out that we only use -gBpl, but
	 * let's not tax its abilities.
	 */
    gBpl = -gBpl;

	/* set the address of the first word beyond the right end
	 * of the gray line.  We will stop just before processing
	 * this word.  We use the rightmost gray word since it has fewer
	 * pad bits than the rightmost binary word.
	 */
    finGAddrLine = (UInt32) pG + ((w/4) + 1) * 4;

	/* Enter the main loop.  We keep reading words until we
	 * reach the end of the line.
	 * In the loop, we get a new binary word, then enter an inner
	 * unrolled loop in which we process the corresponding 8 gray
	 * words.  To process each gray word, we conditionally increment
	 * each pixel (increment all pixels and decrement the pixels
	 * that wrapped around), then use the mask drived from the
	 * proper binary nibble to smash the pixel to 0 if necessary.
	 */
    while (cTrue)
    {
	iBWord = ULONG_ACCESS(pB);
	pB += 4;

	    /********* process GRAY WORD 0 ***********/
	iNibble = (iBWord >> 26) & 0x3C;

	if (iNibble == 0)			/* all gray pix should be 0 */
	{
	    ULONG_ACCESS(pG) = 0;
	}
	else
	{
	    iGWord = ULONG_ACCESS(&pG[gBpl]);	/* get word above this one */
	    oWord = iGWord + 0x01010101;	/* increment all pixels */
	    iGWord = ~oWord & iGWord;		/* find all bits that went from
						 * 1 to 0 */
	    iGWord = (iGWord >> 7);		/* put 1 into LS bit of each
						 * pixel whose MS bit wrapped */
	    iGWord = iGWord & 0x01010101;	/* get rid of extra dreck */
	    oWord -= iGWord;			/* decrement pixels that
						 * wrapped around */
	    oWord &= *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
						/* clear pixels not in runs */
	    ULONG_ACCESS(pG) = oWord;
	}

	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 1 ***********/
	iNibble = (iBWord >> 22) & 0x3C;

	if (iNibble == 0)
	{
	    ULONG_ACCESS(pG) = 0;
	}
	else
	{
	    iGWord = ULONG_ACCESS(&pG[gBpl]);	/* get word above this one */
	    oWord = iGWord + 0x01010101;	/* increment all pixels */
	    iGWord = ~oWord & iGWord;		/* find all bits that went from
						 * 1 to 0 */
	    iGWord = (iGWord >> 7);		/* put 1 into LS bit of each
						 * pixel whose MS bit wrapped */
	    iGWord = iGWord & 0x01010101;	/* get rid of extra dreck */
	    oWord -= iGWord;			/* decrement pixels that
						 * wrapped around */
	    oWord &= *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
						/* clear pixels not in runs */
	    ULONG_ACCESS(pG) = oWord;
	}

	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 2 ***********/
	iNibble = (iBWord >> 18) & 0x3C;

	if (iNibble == 0)
	{
	    ULONG_ACCESS(pG) = 0;
	}
	else
	{
	    iGWord = ULONG_ACCESS(&pG[gBpl]);	/* get word above this one */
	    oWord = iGWord + 0x01010101;	/* increment all pixels */
	    iGWord = ~oWord & iGWord;		/* find all bits that went from
						 * 1 to 0 */
	    iGWord = (iGWord >> 7);		/* put 1 into LS bit of each
						 * pixel whose MS bit wrapped */
	    iGWord = iGWord & 0x01010101;	/* get rid of extra dreck */
	    oWord -= iGWord;			/* decrement pixels that
						 * wrapped around */
	    oWord &= *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
						/* clear pixels not in runs */
	    ULONG_ACCESS(pG) = oWord;
	}

	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 3 ***********/
	iNibble = (iBWord >> 14) & 0x3C;

	if (iNibble == 0)
	{
	    ULONG_ACCESS(pG) = 0;
	}
	else
	{
	    iGWord = ULONG_ACCESS(&pG[gBpl]);	/* get word above this one */
	    oWord = iGWord + 0x01010101;	/* increment all pixels */
	    iGWord = ~oWord & iGWord;		/* find all bits that went from
						 * 1 to 0 */
	    iGWord = (iGWord >> 7);		/* put 1 into LS bit of each
						 * pixel whose MS bit wrapped */
	    iGWord = iGWord & 0x01010101;	/* get rid of extra dreck */
	    oWord -= iGWord;			/* decrement pixels that
						 * wrapped around */
	    oWord &= *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
						/* clear pixels not in runs */
	    ULONG_ACCESS(pG) = oWord;
	}

	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 4 ***********/
	iNibble = (iBWord >> 10) & 0x3C;

	if (iNibble == 0)
	{
	    ULONG_ACCESS(pG) = 0;
	}
	else
	{
	    iGWord = ULONG_ACCESS(&pG[gBpl]);	/* get word above this one */
	    oWord = iGWord + 0x01010101;	/* increment all pixels */
	    iGWord = ~oWord & iGWord;		/* find all bits that went from
						 * 1 to 0 */
	    iGWord = (iGWord >> 7);		/* put 1 into LS bit of each
						 * pixel whose MS bit wrapped */
	    iGWord = iGWord & 0x01010101;	/* get rid of extra dreck */
	    oWord -= iGWord;			/* decrement pixels that
						 * wrapped around */
	    oWord &= *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
						/* clear pixels not in runs */
	    ULONG_ACCESS(pG) = oWord;
	}

	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 5 ***********/
	iNibble = (iBWord >> 6) & 0x3C;

	if (iNibble == 0)
	{
	    ULONG_ACCESS(pG) = 0;
	}
	else
	{
	    iGWord = ULONG_ACCESS(&pG[gBpl]);	/* get word above this one */
	    oWord = iGWord + 0x01010101;	/* increment all pixels */
	    iGWord = ~oWord & iGWord;		/* find all bits that went from
						 * 1 to 0 */
	    iGWord = (iGWord >> 7);		/* put 1 into LS bit of each
						 * pixel whose MS bit wrapped */
	    iGWord = iGWord & 0x01010101;	/* get rid of extra dreck */
	    oWord -= iGWord;			/* decrement pixels that
						 * wrapped around */
	    oWord &= *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
						/* clear pixels not in runs */
	    ULONG_ACCESS(pG) = oWord;
	}

	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 6 ***********/
	iNibble = (iBWord >> 2) & 0x3C;

	if (iNibble == 0)
	{
	    ULONG_ACCESS(pG) = 0;
	}
	else
	{
	    iGWord = ULONG_ACCESS(&pG[gBpl]);	/* get word above this one */
	    oWord = iGWord + 0x01010101;	/* increment all pixels */
	    iGWord = ~oWord & iGWord;		/* find all bits that went from
						 * 1 to 0 */
	    iGWord = (iGWord >> 7);		/* put 1 into LS bit of each
						 * pixel whose MS bit wrapped */
	    iGWord = iGWord & 0x01010101;	/* get rid of extra dreck */
	    oWord -= iGWord;			/* decrement pixels that
						 * wrapped around */
	    oWord &= *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
						/* clear pixels not in runs */
	    ULONG_ACCESS(pG) = oWord;
	}

	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 7 ***********/
	iNibble = (iBWord << 2) & 0x3C;

	if (iNibble == 0)
	{
	    ULONG_ACCESS(pG) = 0;
	}
	else
	{
	    iGWord = ULONG_ACCESS(&pG[gBpl]);	/* get word above this one */
	    oWord = iGWord + 0x01010101;	/* increment all pixels */
	    iGWord = ~oWord & iGWord;		/* find all bits that went from
						 * 1 to 0 */
	    iGWord = (iGWord >> 7);		/* put 1 into LS bit of each
						 * pixel whose MS bit wrapped */
	    iGWord = iGWord & 0x01010101;	/* get rid of extra dreck */
	    oWord -= iGWord;			/* decrement pixels that
						 * wrapped around */
	    oWord &= *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
						/* clear pixels not in runs */
	    ULONG_ACCESS(pG) = oWord;
	}

	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;

    }	/* the end of the loop that processes 8 gray words */
}



/* PRIVATE */
/**************************************************************************
 *  ip_tagVerticalPixels()  
 *		Takes pointers to a scan line of a binary image and a
 *		gray image.  If the binary pixel is in the midst of a
 *		run (above the bottom pixel of the run) it is replaced
 *		with the pixel below it.  Otherwise, it is left alone.
 *		This function is used by i_tagVerticalRuns to tag
 *		each pixel with the length of the vertical run containing
 *		it.
 *
 *	Input:
 *		pB =	pointer binary source
 *		bBpl =	binary source's bytes/line
 *		pG =	pointer to gray dest image
 *		gBpl =	gray dest's bytes/line
 *		w =	width of both images in pixels.
 *
 *	Return code:
 *		void
 **************************************************************************/
void  CDECL 
ip_tagVerticalPixels (
    IPState    *pIPState,
    UInt8      *pB,
    UInt32	bBpl,
    UInt8      *pG,
    UInt32	gBpl,
    UInt32	w)
{
    PROLOGUE;
    UInt32	 oMask;			/* used to combine pixels */
    UInt32	 iNibble;		/* binary nibble corresponding to
				         * the current gray word. */
    UInt32	 iBWord;		/* the binary input word */
    UInt32	 finGAddrLine;		/* the rightmost gray address for the
				         * line + 4.  We stop before writing
				         * this address.
				         */
   const UInt8	*nibMaskBytePtr = (const UInt8 *)nibMask; /* try to discourage
					 * the compiler from doing
					 * unnecessary shifts when computing
					 * addresses.  This is a pointer to
					 * an array of 16 masks.  Each mask
					 * expands every bit in the addressing
					 * nibble by 8.
					 */

	/* Keep the compller happy */
    pIPState = pIPState;
    __error__ = __error__;

	/* set the address of the first word beyond the right end
	 * of the gray line.  We will stop just before processing
	 * this word.  We use the rightmost gray word since it has fewer
	 * pad bits than the rightmost binary word.
	 */
    finGAddrLine = (UInt32) pG + ((w/4) + 1) * 4;

	/* Enter the main loop.  We keep reading words until we
	 * reach the end of the line.
	 * In the loop, we AND the binary word for this line with the
	 * one below it.  In places where the result is 0, we don't want
	 * to change the output image since we are either in the middle
	 * of a run of 0's (which were written in the first pass through
	 * the gray image by ip_calculateVerticalRunLengths) or are at
	 * the end of a run.  In places where the result is 1, we need
	 * to copy the gray pixels from the line below this one to the
	 * current line.  Since a group of 4 gray pixels (1 word) will
	 * often have some pixels in runs and some not, we must be
	 * able to merge runs.  We do this using masks derived from
	 * the nibbles of the binary word - one nibble of binary data
	 * corresponds to four pixels or one gray word.
	 */
    while (cTrue)
    {
	iBWord = ULONG_ACCESS(pB) & ULONG_ACCESS(&pB[bBpl]);
	pB += 4;
	if (iBWord == 0)	/* none of these 32 pixels change */
	{
	    pG += 32;
	    if ((UInt32) pG >= finGAddrLine)
		break;		/* we're done */
	    else
		continue;	/* get the next binary word */
	}

	    /********* process GRAY WORD 0 ***********/
	iNibble = (iBWord >> 26) & 0x3C;

	if (iNibble != 0)			/* some pixels change */
	{
	    oMask = *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
	    ULONG_ACCESS(pG) = (ULONG_ACCESS(&pG[gBpl]) & oMask) |
				(ULONG_ACCESS(&pG[0]) & ~oMask);
	}
	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 1 ***********/
	iNibble = (iBWord >> 22) & 0x3C;

	if (iNibble != 0)			/* some pixels change */
	{
	    oMask = *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
	    ULONG_ACCESS(pG) = (ULONG_ACCESS(&pG[gBpl]) & oMask) |
				(ULONG_ACCESS(&pG[0]) & ~oMask);
	}
	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 2 ***********/
	iNibble = (iBWord >> 18) & 0x3C;

	if (iNibble != 0)			/* some pixels change */
	{
	    oMask = *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
	    ULONG_ACCESS(pG) = (ULONG_ACCESS(&pG[gBpl]) & oMask) |
				(ULONG_ACCESS(&pG[0]) & ~oMask);
	}
	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 3 ***********/
	iNibble = (iBWord >> 14) & 0x3C;

	if (iNibble != 0)			/* some pixels change */
	{
	    oMask = *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
	    ULONG_ACCESS(pG) = (ULONG_ACCESS(&pG[gBpl]) & oMask) |
				(ULONG_ACCESS(&pG[0]) & ~oMask);
	}
	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 4 ***********/
	iNibble = (iBWord >> 10) & 0x3C;

	if (iNibble != 0)			/* some pixels change */
	{
	    oMask = *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
	    ULONG_ACCESS(pG) = (ULONG_ACCESS(&pG[gBpl]) & oMask) |
				(ULONG_ACCESS(&pG[0]) & ~oMask);
	}
	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 5 ***********/
	iNibble = (iBWord >> 6) & 0x3C;

	if (iNibble != 0)			/* some pixels change */
	{
	    oMask = *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
	    ULONG_ACCESS(pG) = (ULONG_ACCESS(&pG[gBpl]) & oMask) |
				(ULONG_ACCESS(&pG[0]) & ~oMask);
	}
	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 6 ***********/
	iNibble = (iBWord >> 2) & 0x3C;

	if (iNibble != 0)			/* some pixels change */
	{
	    oMask = *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
	    ULONG_ACCESS(pG) = (ULONG_ACCESS(&pG[gBpl]) & oMask) |
				(ULONG_ACCESS(&pG[0]) & ~oMask);
	}
	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;


	    /********* process GRAY WORD 7 ***********/
	iNibble = (iBWord << 2) & 0x3C;

	if (iNibble != 0)			/* some pixels change */
	{
	    oMask = *((const UInt32 *)(&nibMaskBytePtr[iNibble]));
	    ULONG_ACCESS(pG) = (ULONG_ACCESS(&pG[gBpl]) & oMask) |
				(ULONG_ACCESS(&pG[0]) & ~oMask);
	}
	    /* point to next gray word.  Quit if we're done. */
	pG += 4;
	if ((UInt32) pG >= finGAddrLine)
	    break;
    }
}

/*

$Log: mitagvrn.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.9  1998/01/13 23:53:31  ddavies
mitagvrn.c -> /product/ipcore/ipshared/src/RCS/mitagvrn.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1995/12/13 02:52:29  ejaquith
mitagvrn.c -> /product/ipcore/ipshared/src/RCS/mitagvrn.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.7  1995/10/18  04:04:14  ddavies
 * mitagvrn.c -> /product/ipcore/ipshared/src/RCS/mitagvrn.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.6  1995/01/11  22:34:11  dferrell
 * mitagvrn.c -> /product/ipcore/ipshared/src/RCS/mitagvrn.c,v
 *
 * Changed shrip.prv to shrcc.prv.
 *
 * Revision 1.5  1994/02/04  20:19:42  dferrell
 * mitagvrn.c -> /product/ipcore/ipshared/src/RCS/mitagvrn.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.4  1993/12/02  19:25:40  ddavies
 * mitagvrn.c -> /product/ipcore/ipshared/src/RCS/mitagvrn.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.3  1993/11/05  01:51:14  ddavies
 * mitagvrn.c -> /product/ipcore/ipshared/src/RCS/mitagvrn.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


