/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mitcenc.c: Contains the utility procedures used to turn an image into a
 *		set of symbols and positions of symbols.  These are called
 *		by procedures in i_tcenc.c.
 *
 *		Int32	ip_makeVerticalRuns()
 *			Composes a set of vertical runs from a set of
 *			horizontal runs.
 *
 *		Int32	ip_calculateSlant()
 *			Calculate the slant of a symbol from the horizontal
 *			and vertical runs.  Used to discriminate between
 *			italic and non-italic characters
 *
 *		Int32	ip_makeGroupsFromRuns()
 *			Composes the groups of merged lines from the
 *			original lines of the symbol.  These are used
 *			to classify the symbol.  This routine is used
 *			to make groups from both horizontal and vertical
 *			runs.
 *
 *		Int32	ip_symbolIsComplete()
 *			Does final classification of symbols, assigns
 *			symbolIDs and tentative positions.
 *
 *		Int32	ip_createUncommittedTCPosition()
 *			Create uncommitted position for new symbol
 *			relative to old symbol.
 *
 *		Int32	ip_commitTCPositions()
 *			Writes final position info and symbol ID
 *			to internal position buffer.
 *
 *		Int32	ip_foundEquivalentSymbol()
 *			Called when two equivalent symbols have been
 *			found in the list of symbols.  This procedure
 *			inserts them into an equivalence class.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#include "bitcount.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /*  prototyes */
#include "shrcmprs.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mitcenc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* The process we use to compress images by finding repeated symbols is
 * well documented (at least there's lots of documentation) in the
 * file i_tcompr.c.  You'll have to read that to understand this.
 *
 * This file contains utility routines called by the main encoding functions
 * in i_tcenc.c.  These functions create the shape and position information.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_makeVerticalRuns()
 *
 *	This procedure is called for to create either horizontal or
 *	vertical groups from a set of horizontal or vertical runs.
 *	The groups are used to classify the symbol.
 *
 *	Input:
 *		pTextCompInfo:		Ptr to all text compression info
 *		symWidth:		width of symbol in pixels
 *		symHeight:		height of symbol in pixels
 *		pHorizRuns:		Ptr to list of horiz runs from
 *					which vert runs will be made.
 *		ppVertRuns:		Ptr to ptr to first resulting
 *					vertical run.
 *
 *	Action:
 *		Creates a set of vertical runs from the set of horizontal
 *		runs.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_makeVerticalRuns (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    Int32	symWidth,
    Int32	symHeight,
    TCScanLineRun *pHorizRuns,
    TCScanLineRun **ppVertRuns)
{
PROLOGUE;
 TCScanLineRun	*pCurHorizRun;			/* ptr to current horiz run */
TCScanLineRun	*pBelowHorizRun;		/* ptr to horiz run under
						 * pCurHorizRun */
TCScanLineRun	*pCurVertRun;			/* ptr to current vert run */
TCScanLineRun	**pHeadArray, **pTailArray;	/* ptrs to arrays of pointers
						 * to runs.  There's one
						 * second level ptr for each
						 * column of the symbol.
						 */
Int16		 curYCoord;			/* the current Y offset from
						 * the center of the symbol.
						 * Like all coords, this
						 * has an extra bit of
						 * significance so that
						 * symbols with an even number
						 * of rows can have symmetrical
						 * distances to the center.
						 */
Int32		 index, maxIndex;		/* indices into array of
						 * ptrs to runs. */


    /* We use a state machine to work our way across each horizontal
     * line, initiating and terminating vertical runs as we go.
     * These are 4 active state bits, but only 8 of the combinations
     * are meaningful.  The LS two bits refer to the colors of the
     * pixels in the current and previous line at each horizontal
     * position.  The next two bits say whether we should pay attention
     * to the current-line and below-line run pointers.
     */
#define	cAllWhite	0
#define	cBlackBelow	1
#define cBlackAbove	2
#define cAllBlack	3
#define	cCurAtEOL	0x4	/* when set, this bit indicates that the
				 * current-line run pointer is at the
				 * end of its line so the run it points
				 * to should be ignored. */
#define cBelowAtEOL	0x8	/* when set, this bit indicates that the
				 * below-line run pointer is at the
				 * end of its line so the run it points
				 * to should be ignored. */

	/* In the original version of this algorithm, we created
	 * or extended every vertical run touched by every horizontal
	 * run.  This did something interesting for *every* black pixel
	 * in *every* instance of *every* symbol.  We were spending too
	 * much time in here.
	 *
	 * The new algorithm tries to get rid of one of the every's.
	 * We will compare each horizontal run to the run under it, if
	 * any.  In places where the run under the current one has
	 * nothing above it, we complete a vertical run.  In places
	 * where the run on top has nothing under it, we start a
	 * vertical run.  Otherwise, we do nothing.  This allows us
	 * to only do things around the edges of strokes instead of
	 * touching every black pixel.
	 *
	 * There is one chain of vertical runs for each column in the
	 * symbol.  We use the arrays pVertRunHeadArray and pVertRunTailArray
	 * in pTextCompInfo to point to the head and tail of each chain.
	 * First, clear the pointers in these arrays.
	 */

    pHeadArray = pTextCompInfo->pVertRunHeadArray->ptrArray;
    pTailArray = pTextCompInfo->pVertRunTailArray->ptrArray;
    for (index = symWidth-1; index >= 0; index--)
    {
	pHeadArray[index] = NULL;
	pTailArray[index] = NULL;
    }

	/* HET, WATCH OUT!! Since we only use symWidth-1, we just
	 * decrement it here.
	 */
    symWidth--;

	/* Get ready to start marking the vertical runs.  The first y coord
	 * is height-1.  curYCoord gets more negative on each line since the
	 * matching and group-making code expects that runs go from large
	 * coordinates to smaller (more negative) coordinates as we go along
	 * the chain.  curYCoord gets decrements by 2 because we have an
	 * extra bit of significance.
	 */
    curYCoord = (Int16) (symHeight - 1);

	/* First, initialize a set of vertical runs for every horizontal
	 * run in the first (bottom) horizontal scan line.
	 */
    for (pCurHorizRun = pHorizRuns;
	 pCurHorizRun != NULL &&
		pCurHorizRun->pClient == (void *)cEvenLineTag;
	 pCurHorizRun = pCurHorizRun->pNext)
    {
	    /* touch all of the vertical runs intersected by this
	     * horizontal run.
	     */
	index = (pCurHorizRun->runRight + symWidth)/2;
	maxIndex = (pCurHorizRun->runLeft + symWidth)/2;
	for (; index >= maxIndex; index--)
	{
	    __error__ =
		i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
				  (void **)&pCurVertRun);
	    if (__error__ != ia_successful)
		goto tcMakeVertRunCleanup;

	    pCurVertRun->runRight = curYCoord;
	    if ((index & 0x1) == 0)
		pCurVertRun->pClient = (void *)cEvenLineTag;
	    else
		pCurVertRun->pClient = (void *)cOddLineTag;
	    pHeadArray[index] = pCurVertRun;
	    pTailArray[index] = pCurVertRun;
	}
    }

	/* Since we're done with the bottom line, decrement curYCoord */
    curYCoord -= 2;


	/* Now go through each run of each subsequent line, initiating
	 * and terminating vertical runs as we go.  This is a more
	 * subtle process than I originally thought.  We clearly
	 * can't assume a 1-to-1 relationship between runs on the
	 * the current line and runs on the previous line, though
	 * that's the most common case.
	 *
	 * We'll use a state machine to work our way across each 
	 * horizontal line.  There are four main states, cAllWhite,
	 * cAllBlack, cBlackAbove and cBlackBelow.  cAllWhite
	 * means the pixels in the current and below lines at
	 * some horizontal position are both white.  cAllBlack means
	 * they're both black.  cBlackAbove means the pixel in the
	 * current line is black and the pixel below is white.
	 * cBlackBelow is the opposite.
	 * There are two more state bits.  They are used to indicate
	 * that one or the other run pointer is at the end of its
	 * line so should be ignored.  These are both never set
	 * since when that happens, we advance to the next line.
	 * We'll use the variable "index" to mark our current horizontal
	 * position and the variable "state" to record the state.
	 * Even though the horizontal coords in the TCScanLineRun
	 * structs range from -(symWidth-1) to (symWidth-1) in units
	 * of 2, index ranges from symWidth-1 down to 0 in units of 1.
	 * This makes it a useful index for the arrays of pointers to
	 * vertical runs pHeadArray and pTailArray.
	 * The initial value of the state is cAllWhite.
	 *
	 * When can we advance the run pointers?  At the left end of each
	 * run.  Which pointers do we advance?
	 *
	 *	CurrentState	NextState	Advance
	 *	------------	---------	-------
	 *	cAllBlack	cAllWhite	pCurHorizRun, pBelowHorizRun
	 *	cAllBlack	cBlackBelow	pCurHorizRun
	 *	cAllBlack	cBlackAbove	pBelowHorizRun
	 *	cBlackBelow	cAllWhite	pBelowHorizRun
	 *	cBlackBelow	cBlackAbove	pBelowHorizRun
	 *	cBlackAbove	cAllWhite	pCurHorizRun
	 *	cBlackAbove	cBlackBelow	pCurHorizRun
	 *
	 * If advancing a pointer would cause it to go to the next line,
	 * set that pointer's EOL bit in the state and do not advance it.
	 * When the state becomes (cAllWhite | cCurAtEOL | cBelowAtEOL),
	 * we advance both run pointers to their respective next lines.
	 *
	 * Performance Improvement: The speedup wasn't as big as I would
	 * have hoped.  So, I'm going to replace the standard state
	 * machine loop-containing-switch-statement with a bunch of
	 * gotos that jump from state to state.  The execution path
	 * is the same, we just eliminate a bunch of jumps and switches.
	 * We'll see if this helps much.
	 */
    pBelowHorizRun = pHorizRuns;	/* start second line at the
					 * beginning since pCurHorizRun
					 * is already at the second line
					 */
mvrAllWhite:
	/* We know that neither pointer is off the end of its
	 * line, so we just move up to the rightmost point
	 * of the two.
	 * This is the state in which we always start a line.  If
	 * pCurHorizRun is NULL at this point, go terminate
	 * all the remaining vertical runs.
	 */
    if (pCurHorizRun == NULL)
	goto mvrTerminateAllRuns;

    if (pCurHorizRun->runRight > pBelowHorizRun->runRight)
    {
	index = (pCurHorizRun->runRight + symWidth)/2;
	goto mvrBlackAbove;
    }
    else
    {
	index =  (pBelowHorizRun->runRight + symWidth)/2;
	if (pCurHorizRun->runRight == pBelowHorizRun->runRight)
	    goto mvrAllBlack;		/* go from allWhite to allBlack */

	    /* if the weren't equal, pCurHorizRun->runRight must be
	     * to the left of pBelowHorizRun->runRight, so we fall
	     * through to the mvrBlackBelow state.
	     */
    }


mvrBlackBelow:
	/* terminate all the vert runs touched by just the
	 * run below.  We end when either we run off the
	 * end of pBelowHorizRun or we run into pCurHorizRun,
	 * or both.  First, figure out which vertical runs
	 * to terminate.
	 * Remember that the runRight and runLeft values
	 * change in units of 2, so use this convention here.
	 */
    if (pCurHorizRun->runRight + 2 > pBelowHorizRun->runLeft)
	maxIndex = (pCurHorizRun->runRight + 2 + symWidth)/2;
    else
	maxIndex = (pBelowHorizRun->runLeft + symWidth)/2;

	/* terminate all the runs that end on the line below.
	 * Note that line`s y coord is curYCoord + 2.
	 */
    for (; index >= maxIndex; index--)
    {
	if (pHeadArray[index] == NULL)
	    goto tcMakeVertRunInternalError; /* should never happen */
	pHeadArray[index]->runLeft = curYCoord + 2;
    }

	/* now decide where to go next */
    if (pCurHorizRun->runRight + 2 > pBelowHorizRun->runLeft)
    {
	goto mvrAllBlack;
    }
    else
    {
	    /* since we've run off the end of pBelowHorizRun,
	     * advance it unless doing so would run it off the
	     * end of its line.
	     */
	if (pBelowHorizRun->pNext->pClient == pBelowHorizRun->pClient)
	{
	    if (pCurHorizRun->runRight + 2 < pBelowHorizRun->runLeft)
	    {
		pBelowHorizRun = pBelowHorizRun->pNext;
		goto mvrAllWhite;
	    }
	    else
	    {
		pBelowHorizRun = pBelowHorizRun->pNext;
		goto mvrBlackAbove;
	    }
	}
	else	/* go to the BelowAtEOL versions of these states.
		 * This is in lieu of actually moving pBelowHorizRun.
		 */
	{
	    if (pCurHorizRun->runRight + 2 < pBelowHorizRun->runLeft)
	    {
		goto mvrAllWhiteBelowAtEOL;
	    }
	    else
	    {
		goto mvrBlackAboveBelowAtEOL;
	    }
	}
    }


mvrBlackAbove:
	/* Initiate new vert runs for each pixel touched by just the
	 * run above.  We end when either we run off the
	 * end of pCurHorizRun or we run into pBelowHorizRun,
	 * or both.
	 */
    if (pBelowHorizRun->runRight + 2 > pCurHorizRun->runLeft)
	maxIndex = (pBelowHorizRun->runRight + 2 + symWidth)/2;
    else
	maxIndex = (pCurHorizRun->runLeft + symWidth)/2;

    for (; index >= maxIndex; index--)
    {
	__error__ =
	    i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
			      (void **)&pCurVertRun);
	if (__error__ != ia_successful)
	    goto tcMakeVertRunCleanup;

	if (pHeadArray[index] == NULL)
	{	/* this is the very first vert run in this column */
	    pTailArray[index] = pCurVertRun;
	}
	else	/* tack this run onto the end of an existing run */
	{
	    pHeadArray[index]->pNext = pCurVertRun;
	}
	pHeadArray[index] = pCurVertRun;

	pCurVertRun->runRight = curYCoord;
	if ((index & 0x1) == 0)
	    pCurVertRun->pClient = (void *)cEvenLineTag;
	else
	    pCurVertRun->pClient = (void *)cOddLineTag;
    }

	/* decide where to go next */
    if (pBelowHorizRun->runRight + 2 > pCurHorizRun->runLeft)
    {
	goto mvrAllBlack;
    }
    else
    {
	    /* since we've reached the end of pCurHorizRun, try
	     * to advance.  If we can't go to the CurAtEOL states
	     * instead.
	     */
	if ((pCurHorizRun->pNext != NULL) &&
	    (pCurHorizRun->pNext->pClient == pCurHorizRun->pClient))
	{
	    if (pBelowHorizRun->runRight + 2 < pCurHorizRun->runLeft)
	    {
		pCurHorizRun = pCurHorizRun->pNext;
		goto mvrAllWhite;
	    }
	    else	/* pBelowHorizRun starts where pCurHorizRun
			 * stops */
	    {
		pCurHorizRun = pCurHorizRun->pNext;
		goto mvrBlackBelow;
	    }
	}
	else	/* pCurHorizRun is at the end of its line, so
		 * jump to the CurAtEOL states in lieu of advancing it */
	{
	    if (pBelowHorizRun->runRight + 2 < pCurHorizRun->runLeft)
	    {
		goto mvrAllWhiteCurAtEOL;
	    }
	    else	/* pBelowHorizRun starts where pCurHorizRun
			 * stops */
	    {
		goto mvrBlackBelowCurAtEOL;
	    }
	}
    }


mvrAllBlack:
	/* Advance until the first left endpoint is reached.
	 * Advance that run.  If both endpoints are
	 * reached concurrently, advance both of them.
	 */
    if (pCurHorizRun->runLeft > pBelowHorizRun->runLeft)
    {
	index = (pCurHorizRun->runLeft + symWidth)/2 - 1;

	    /* if we can advance pCurHorizRun, do so.  Otherwise,
	     * go to the state that ignores it.
	     */
	if ((pCurHorizRun->pNext != NULL) &&
	    (pCurHorizRun->pNext->pClient == pCurHorizRun->pClient))
	{
	    pCurHorizRun = pCurHorizRun->pNext;
	    goto mvrBlackBelow;
	}
	else
	{
	    goto mvrBlackBelowCurAtEOL;
	}
    }
    else
    {
	index =  (pBelowHorizRun->runLeft + symWidth)/2 - 1;
	if (pBelowHorizRun->runLeft > pCurHorizRun->runLeft)
	{
		/* if we can advance pBelowHorizRun, do so.  Otherwise,
		 * go to the state that ignores it.
		 */
	    if (pBelowHorizRun->pNext->pClient == pBelowHorizRun->pClient)
	    {
		pBelowHorizRun = pBelowHorizRun->pNext;
		goto mvrBlackAbove;
	    }
	    else
	    {
		goto mvrBlackAboveBelowAtEOL;
	    }
	}
	else	/* they're equal.  There are four cases here.  A) Neither
		 * pCurHorizRun nor pBelowHorizRun is at the end of
		 * a line; B) pCurHorizRun is at the end of a line bu
		 * pBelowHorizRun is not; C) pBelowHorizRun is at the
		 * end of a line and pCurHorizRun is not; d) Both are
		 * at the ends of lines.  We branch on the state of
		 * pCurHorizRun in an outer IF statement and pBelowHorizRun
		 * in an inner IF.
		 */
	{
		/* See whether we can advance pCurHorizRun */
	    if ((pCurHorizRun->pNext != NULL) &&
		(pCurHorizRun->pNext->pClient == pCurHorizRun->pClient))
	    {
		if (pBelowHorizRun->pNext->pClient ==
			    pBelowHorizRun->pClient)
		{
			/* We can advance both pointers */
		    pCurHorizRun = pCurHorizRun->pNext;
		    pBelowHorizRun = pBelowHorizRun->pNext;
		    goto mvrAllWhite;
		}
		else	/* can only advance pCurHorizRun */
		{
		    pCurHorizRun = pCurHorizRun->pNext;
		    goto mvrAllWhiteBelowAtEOL;
		}
	    }
	    else	/* pCurHorizRun is already at EOL */
	    {
		if (pBelowHorizRun->pNext->pClient ==
			    pBelowHorizRun->pClient)
		{	/* can only advance pBelowHorizRun */
		    pBelowHorizRun = pBelowHorizRun->pNext;
		    goto mvrAllWhiteCurAtEOL;
		}
		else	/* both at EOL, so go to a new line */
		{
		    curYCoord -= 2;
		    pCurHorizRun = pCurHorizRun->pNext;
		    pBelowHorizRun = pBelowHorizRun->pNext;
		    goto mvrAllWhite;
		}
	    }
	}
    }


mvrAllWhiteCurAtEOL:
	/* We know that the current pointer is off the end
	 * of its row, so just move up to the end of the
	 * below-line pointer.
	 */
    index =  (pBelowHorizRun->runRight + symWidth)/2;
	/* We'll just fall through to the blackBelow state.
	 */

mvrBlackBelowCurAtEOL:
	/* Just terminate the runs specified by the run below */
    maxIndex = (pBelowHorizRun->runLeft + symWidth)/2;
    for (; index >= maxIndex; index--)
    {
	if (pHeadArray[index] == NULL)
	    goto tcMakeVertRunInternalError; /* should never happen */
	pHeadArray[index]->runLeft = curYCoord + 2;
    }

	/* since we've run off the end of pBelowHorizRun,
	 * advance it unless doing so would run it off the
	 * end of its line.
	 */
    if (pBelowHorizRun->pNext->pClient == pBelowHorizRun->pClient)
    {
	pBelowHorizRun = pBelowHorizRun->pNext;
	goto mvrAllWhiteCurAtEOL;
    }
    else
    {	/* we're off the end of both, so advance both pointers. */
        curYCoord -= 2;
	pBelowHorizRun = pBelowHorizRun->pNext;
	pCurHorizRun = pCurHorizRun->pNext;
	goto mvrAllWhite;
    }



    /* can't have black above or all black if there's no line above,
     * so skip (cBlackAbove | cCurAtEOL) and (cAllBlack | cCurAtEOL).
     */


mvrAllWhiteBelowAtEOL:
	/* We know that the below pointer is off the end
	 * of its row, so just move up to the end of the
	 * current pointer.
	 */
    index =  (pCurHorizRun->runRight + symWidth)/2;
	/* We'll just fall through to the blackAbove state */

mvrBlackAboveBelowAtEOL:
	/* Just initiate the runs specified by the run above */
    maxIndex = (pCurHorizRun->runLeft + symWidth)/2;
    for (; index >= maxIndex; index--)
    {
	__error__ =
	    i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
			      (void **)&pCurVertRun);
	if (__error__ != ia_successful)
	    goto tcMakeVertRunCleanup;

	if (pHeadArray[index] == NULL)
	{	/* this is the very first vert run in this column */
	    pTailArray[index] = pCurVertRun;
	}
	else	/* tack this run onto the end of an existing run */
	{
	    pHeadArray[index]->pNext = pCurVertRun;
	}
	pHeadArray[index] = pCurVertRun;

	pCurVertRun->runRight = curYCoord;
	if ((index & 0x1) == 0)
	    pCurVertRun->pClient = (void *)cEvenLineTag;
	else
	    pCurVertRun->pClient = (void *)cOddLineTag;
    }

	/* since we've run off the end of pCurHorizRun,
	 * advance it unless doing so would run it off the
	 * end of its line.
	 */
    if ((pCurHorizRun->pNext != NULL) &&
	(pCurHorizRun->pNext->pClient == pCurHorizRun->pClient))
    {
	pCurHorizRun = pCurHorizRun->pNext;
	goto mvrAllWhiteBelowAtEOL;
    }
    else
    {	/* we're off the end of both, so advance both pointers.
	 */
	curYCoord -= 2;
	pCurHorizRun = pCurHorizRun->pNext;
	pBelowHorizRun = pBelowHorizRun->pNext;
	goto mvrAllWhite;
    }


    /* can't have black below or all black if there's no line below,
     * so skip (cBlackBelow | cBelowAtEOL) and
     * (cAllBlack | cBelowAtEOL).
     */



mvrTerminateAllRuns:
	/* At this point, pCurHorizRun points above the symbol and
	 * pBelowHorizRun points to the top line of the symbol.
	 * Terminate all the runs pointed to by pBelowHorizRun.
	 */
    for (;
	 pBelowHorizRun != NULL;
	 pBelowHorizRun = pBelowHorizRun->pNext)
    {
	    /* terminate all of the vertical runs intersected by this
	     * horizontal run.
	     */
	index = (pBelowHorizRun->runRight + symWidth)/2;
	maxIndex = (pBelowHorizRun->runLeft + symWidth)/2;
	for (; index >= maxIndex; index--)
	{
	    if (pHeadArray[index] == NULL)
		goto tcMakeVertRunInternalError; /* should never happen */
	    pHeadArray[index]->runLeft = curYCoord + 2;
	}
    }


	/* At this point, we have a bunch of chains of runs.  String
	 * them together into a single big chain by connecting the
	 * head of each list to the tail of the next list.
	 */
    for (index = 0; index < symWidth; index++)
    {
	pHeadArray[index]->pNext = pTailArray[index+1];
    }
    pHeadArray[symWidth]->pNext = NULL;
    *ppVertRuns = pTailArray[0];

	/* if we make it to here, things went fine, so return happily */
    __error__ = ia_successful;

tcMakeVertRunCleanup:
    NORMAL_RETURN;

/* get here if something that should never happen happened */
tcMakeVertRunInternalError:
    return ia_internal;

}


/* PRIVATE */
/**************************************************************************
 *   ip_calculateSlant()
 *
 *	This procedure calculates the slant, volume and several flavors
 *	of signed asymmetry for a symbol.
 *	Slant is mostly used to differentiate italic from non-italic
 *	characters.  Volume is used to distinguish round things,
 *	like 8 vs B and e vs a.
 *
 *	Input:
 *		pInstance:	Ptr to structure containing symbol data.
 *
 *	Action:
 *		From the runs, calculates a measure of the tilt of a
 *		symbol.  This is used to tell the difference between
 *		italic and non-italic characters.  Also calculate a
 *		measure of the amount of a symbol's bounding box that
 *		it takes up (volume) and the crookedness of its
 *		horizontal and vertical centerlines (asymmetry).
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_calculateSlant (
    IPState    *pIPState,
    TCInstance *pInstance)
{
PROLOGUE;
 TCScanLineRun	*pPrevRun, *pCurRun;
Int16		 prevLeft, prevRight;	/* x coordinates of left and right
					 * edges of symbol in line below
					 * current one. */
Int32		 lineFlags;		/* used to keep track of line
					 * boundaries. */
Int32		 curSlant;		/* the current value of slant */
Int32		 curVolume;		/* current volume. */
Int32		 curSAsymmetry;		/* current signed asymmetry */
UInt32		 curProfileBit;		/* used to insert bits into profiles */
Int32		 curLeftProfile;	/* binary approximation to left side
					 * of instance symbol. */
Int32		 curRightProfile;	/* binary approximation to right side
					 * of instance symbol. */
Int32		 firstSAsymmetry;	/* signed asymmetry calculated for
					 * the first scan line or column. */
Int32		 lastSAsymmetry;	/* signed asymmetry calculated for
					 * the last scan line or column. */
Int32		 profileWidth;		/* width used to calculate profiles */
Int32		 minEnvWidth;		/* used to decide whether runs are
					 * wide enough to count towards the
					 * total volume */
Int32		 area;			/* the fraction of the total area
					 * of the symbol that the volume
					 * must occupy to become a volume
					 * that we'd like to check. */
Int32		 index;			/* used to index histograms */
Int32		 runLenHist[128] = {0};	/* histogram of run lengths in symbol */
Int32		 runMassHist[128] = {0}; /* table indexed by run
					 * lengths, each box holds
					 * total number of pixels
					 * in runs of that length. */

	/* Slant is a funny measure.  It's main purpose is to
	 * distinguish italic round things from non-italic round things
	 * (like o's and e's) because the other tests don't do a very
	 * good job at that task.  The slant attempts to represent
	 * the extent of outside edges that slope to the right
	 * versus outside edges that slope to the left.  Italic characters
	 * tend to have lines sloping to the upper right.
	 * A few things make the slant measure tricky.  There aren't many
	 * pixels in the tiny symbols in which this measure is supposed
	 * to work, so we don't have a lot of leeway.  We don't want to
	 * have to figure out if we're dealing with little italic round
	 * things when we apply the measure, we'd like to apply it on
	 * every character.  Therefore, it can't be affected (much) by
	 * noise that appears on any character.  Measures like the number
	 * of pixels in the four quadrants of the image don't work.
	 * Raw distances from the corners of the bounding box of the
	 * symbol to the nearest pixel aren't very discrminating either.
	 * The measure we've chosen keeps track of sloping outside edges.
	 * Each time the outside edge of the symbol changes (is not vertical),
	 * we add MIN( amount by which edge changed, 2) to the slant in
	 * the direction of the change.  Remember that all coordinates are
	 * multiplied by 2 (have an extra bit of precision), so we really
	 * add MIN(coord diff, 4) and keep slant in units of double coords.
	 *
	 * Trial Bug Fix: try skipping the first and last scan lines on
	 * both the horizontal and vertical runs.  That's where most of
	 * the noise is.  The horizontal calculation can tolerate noise
	 * on the east and west sides, so we ignore the north and south.
	 * Similarly, the vertical calculation can ignore noise on the north
	 * and south, so we ignore noise on the east and west.
	 *
	 * Add Volume:  Distinguishing tiny, bold a's from tiny bold e's
	 * is really a pain.  They have the same weight and basically look
	 * the same.  We're going to try to use a measure derived from the
	 * slant.  The a's differ from the e's in that they take up a little
	 * more room all around the edge.  We don't just use mass because
	 * that counts the holes in the middles of the characters and
	 * those often close up on little bold blobs.  The "volume" of
	 * a character is the number of pixels inside the left and right
	 * edges of the character plus the number of pixels inside the
	 * top and bottom edges of a character.  Most pixels will be
	 * counted twice, but that's ok.
	 *
	 * Add Asymmetry: That almost worked, but after
	 * fooling around with the same image, another a/e confusion popped
	 * up.  The volume measure for one was 600.  The other's volume
	 * was 614.  That's too close.  So, I peered at these again.  One
	 * thing that stands out after awhile is that even though all the
	 * pixels of the a are within distance 1 of all the pixels of the
	 * e, the a is generally less symmetric around both the horizontal
	 * and vertical midlines.  So, we calculate asymmetry as the
	 * sum of the absolute values of the sums of envelope endpoints
	 * on each scan line.  Remember that endpoints on the left have
	 * negative coordinates relative to the midpoint and endpoints
	 * on the right have positive coordinates.  The sum of the endpoints
	 * in a symmetric symbol is 0 on every scan line.
	 * We calculate the symmetry for both horizontal and vertical runs
	 * to catch things that are symmetric around only one of the midlines.
	 * Like most of thes funny little measures, this one just gets in
	 * the way for large symbols.  So we multiply the range of allowable
	 * asymmetries by a factor that grows slowly with size.
	 *
	 * I tried replacing volume with asymmetry, but got stuck on
	 * 8 vs. B, so had to put volume back in again.  And there was
	 * still a case where it didn't work, so I changed the way in
	 * which volume is calculated.  We only add the space occupied by
	 * the symbol in a given scan line to the total volume if this
	 * space is larger than a given fraction of the total symbol width.
	 * I'm trying to stress square things over round things here.
	 * The minimum width for a qualifying slice of envelope is
	 * held in minEnvWidth.
	 *
	 * To distinguish I's that are slightly shifted to the right from
	 * 1's, I had to add horizontal and vertical signed asymmetry.
	 * Their purpose is to try to detect when an instance is shifted
	 * relative to the exemplar.  These are calculated by summing up
	 * the signed coordinates of the points on the outer envelope (sum
	 * is 0 for symmetric objects), then dividing by the number of
	 * scan lines used to compute the sum.  This gives the center of
	 * gravity.  We then compute the distance from this point to the
	 * scan line containing the first orthogonal run.  Huh?  When
	 * computing asymmetry using horizontal runs, we compute the distance
	 * from the central vertical axis of the character to the left
	 * edge.  That's where the first vertical runs is.  When computing
	 * asymmetry using vertical runs, we compute the distance from
	 * the central horizontal axis to the bottom of the symbol.  That`s
	 * where the first horizontal run is.  When preparing to compare
	 * exemplar and instance horizontal runs, we look at the difference
	 * in the respective distances from horizontal axes to bottom lines.
	 * If the difference is greater than 3.5 (remember we're using
	 * 2x coords), we quit.  If it's less than 2, the symbols are
	 * already aligned.  If it's around 2, we shift the instance or
	 * exemplar to line them up.
	 *
	 * Add strokeWidth calculation: Just comparing the total number
	 * of pixels in two symbols doesn't do a very good job of
	 * discriminating bold from non-bold characters.  We'd really
	 * like to measure the widths of the strokes making up the symbol.
	 * Well, the strokes may be diagonal and we only have horizontal
	 * and vertical runs, but we'll try to make do.
	 *
	 * You can't just divide the mass by the number of runs.  That
	 * measure only works if you have only vertical strokes and use
	 * only horizontal runs or vice versa.  Most symbols have strokes
	 * going in all directions.  To attempt to capture this, we
	 * only want to count the relatively short horizontal and vertical
	 * runs.
	 *
	 * We keep two histograms.  One is indexed by run length and is
	 * incremented each time we find a run of that length.  The other
	 * is also indexed by run length, but the length of the run is
	 * added each time a run is found.  This holds the total number
	 * of pixels in runs of that length.
	 *
	 * After creating the histograms, we inspect them starting from
	 * the shortest runs.  At each run length, we add the total number
	 * of pixels having that run length to the total number of pixels
	 * seen so far.  If this is a sufficiently large fraction of the
	 * number of pixels in the entire symbols (maybe 50%?), we stop.
	 * This total is divided by the number of runs required to make
	 * that total to get the average stroke width.
	 *
	 * For historical purposes, note that we tried out an edge
	 * pixel measure which was supposed to keep tiny compact things
	 * from turning into not-quite-tiny diffuse things.  It worked
	 * ok, but just clamping the size of tiny things worked better.
	 * OCR results from both methods were almost identical, except
	 * when the clamp was a tiny bit more effective.
	 *
	 * Ok, but we need yet another measure.  This one will be used
	 * to help us align characters vertically, i.e. maintain a common
	 * baseline.  We tried using counts of pixels in the bottom row
	 * or two, but wasn't very effective.
	 *
	 * Our new measure is a binary profile of each side of the
	 * symbol.  The way a human aligns symbols in the presence of
	 * noise is to align the salient featuers.  We'll try to do the
	 * same thing.  We'll maintain two profiles, left and right.
	 * Each profile has one bit position for each of the bottom
	 * 32 rows of the symbol.  When a piece of the symbol gets
	 * close to the edge of the symbol's bounding box (7/8 of the
	 * way from the center in our current implementation), we put
	 * a 1 in the profile bit corresponding to the current row.
	 * The result is a word that contains bits where the symbol
	 * really sticks out.  When we're replacing one symbol with
	 * another on the page, we adjust the position of the replacement
	 * symbol relative to the original symbol to get the best
	 * match between the profiles.  We do this by moving the
	 * the profiles up or down by 1 relative to each other, XORing
	 * them and counting the difference bits.
	 * 
	 *
	 * There are a few more details.  First, we don't bother to do this
	 * for symbols whose max dimension is greater than 128.  That's
	 * because we'd like to use fixed size arrays and telling bold from
	 * non-bold isn't a subtle thing in symbols that big.  Second,
	 * we'd like to have some extra digits on the stroke width, so we
	 * shift it left before dividing by number of runs.
	 */
    curSlant = 0;			/* initialization */
	/* We calculate this every time we compare two symbols, so
	 * why not do it once now?
	 */
    pInstance->maxDimension = MAX(pInstance->w, pInstance->h);

	/* calculate the minimum width of an envelope slice that we count
	 * towards the total volume.  Remember that the coordinates for
	 * runs are multiplied by 2 already, so we have to multiply the
	 * width by 2 also to make minEnvWidth a valid thing to compare.
	 */
    curLeftProfile = curRightProfile = 0;
    curProfileBit = 0x80000000;		/* line 0 corresponds to MS bit */
    minEnvWidth = (((cMinEnvWidth * pInstance->w) - pInstance->w) * 2)
					/ cMinEnvWidth;
    profileWidth = ((cProfileFrac * pInstance->w) - pInstance->w)
					/ cProfileFrac;
    pCurRun = pInstance->pHorizRuns;
    prevRight = pCurRun->runRight;
	/* Skip the first scan line */
    pPrevRun = pCurRun;
    while ((pCurRun != NULL) && ((UInt32)pCurRun->pClient == cEvenLineTag))
    {
	pPrevRun = pCurRun;
	pCurRun = pCurRun->pNext;
    }

    prevLeft = pPrevRun->runLeft;
    lineFlags = cEvenLineTag;
    curVolume = 0;
    if (-prevLeft >= profileWidth)
	curLeftProfile |= curProfileBit;
    if (prevRight >= profileWidth)
	curRightProfile |= curProfileBit;
    curProfileBit >>= 1;
    if ((prevRight - prevLeft + 2) >= minEnvWidth)
	curVolume = prevRight - prevLeft + 2;
    firstSAsymmetry = MAX(0, prevRight) + MIN(0, prevLeft);
    curSAsymmetry = firstSAsymmetry;
    lastSAsymmetry = 0;

	/* process the remaining lines */
    while (pCurRun != NULL)
    {
	lineFlags ^= cLineTagMask;		/* move to the new line */
	if (pCurRun->runRight != prevRight)
	{
	    if (pCurRun->runRight < prevRight)
	    {
	        curSlant -= MIN(prevRight - pCurRun->runRight, 4);
	    }
	    else
	    {
		curSlant += MIN(pCurRun->runRight - prevRight, 4);
	    }
	    prevRight = pCurRun->runRight;
	}

	    /* go find the left end of this scan line */
	while ((pCurRun != NULL) && ((Int32)pCurRun->pClient == lineFlags))
	{
	    pPrevRun = pCurRun;
	    pCurRun = pCurRun->pNext;
	}

	    /* check the left end of the scan line */
	if (pPrevRun->runLeft != prevLeft)
	{
	    if (pPrevRun->runLeft < prevLeft)
	    {
		curSlant -= MIN(prevLeft - pPrevRun->runLeft, 4);
	    }
	    else
	    {
		curSlant += MIN(pPrevRun->runLeft - prevLeft, 4);
	    }
	    prevLeft = pPrevRun->runLeft;
	}

	if (-prevLeft >= profileWidth)
	    curLeftProfile |= curProfileBit;
	if (prevRight >= profileWidth)
	    curRightProfile |= curProfileBit;
	curProfileBit >>= 1;
	if ((prevRight - prevLeft + 2) >= minEnvWidth)
	    curVolume += prevRight - prevLeft + 2;
	lastSAsymmetry = MAX(0, prevRight) + MIN(0, prevLeft);
	curSAsymmetry += lastSAsymmetry;
    }
    pInstance->slant = curSlant;
    pInstance->hVolume = curVolume;
    pInstance->leftProfile = curLeftProfile;
    pInstance->rightProfile = curRightProfile;

	/* Should we check the horizontal volume?  This is only done if
	 * is is a sufficiently large fraction of the total area.
	 */
    area = pInstance->w * pInstance->h;
    area = ((cMinVolFraction * area) - area)/cMinVolFraction;
    if (area <= curVolume)
	pInstance->checkHVolume = cTrue;
    else
	pInstance->checkHVolume = cFalse;

	/* The horizontal asymmetry should disregard the noisy top and
	 * bottom lines, then find the location of the central 
	 * vertical axis, then subtract the distance to the left line.
	 * The coordinate of the left line is -(w-1).
	 * We disregard the -1 since it will be in all signed
	 * asymmetries.  Therefore the distance to the bottom
	 * is curSAsymmetry - (-w) = curSAsymmetry + w.
	 * To avoid rounding errors, let's keep an extra bit
	 * of precision, so everything is multiplied by 2.
	 */
    curSAsymmetry -= firstSAsymmetry + lastSAsymmetry;
    if (pInstance->h > 2)
	curSAsymmetry = 2*curSAsymmetry / (pInstance->h - 2);
    else
	curSAsymmetry = 0;

    pInstance->hSAsymmetry = curSAsymmetry + 2 * pInstance->w;

	/* Calculate stroke width from horizontal runs (which are still
	 * in the cache, right?) if the symbol is small enough.
	 */
    if (pInstance->maxDimension < 128)
    {
	for (pCurRun = pInstance->pHorizRuns;
	     pCurRun != NULL;
	     pCurRun = pCurRun->pNext)
	{
		/* to use the histogram efficiently, we want 1x coords */
	    index = ((pCurRun->runRight - pCurRun->runLeft) / 2) + 1;
	    runLenHist[index]++;
	    runMassHist[index] += index;
	}
    }

	/* Do the same thing for the vertical runs.  Since the image
	 * is rotated by 90 degrees, we have to reverse the polarity of
	 * the slant calcualted.
	 */
    curSlant = 0;			/* initialization */

    pCurRun = pInstance->pVertRuns;
    prevRight = pCurRun->runRight;
	/* Skip the first scan line */
    pPrevRun = pCurRun;
    while ((pCurRun != NULL) && ((UInt32)pCurRun->pClient == cEvenLineTag))
    {
	pPrevRun = pCurRun;
	pCurRun = pCurRun->pNext;
    }


    prevLeft = pPrevRun->runLeft;
    lineFlags = cEvenLineTag;
    curVolume = 0;
    minEnvWidth = ((cMinEnvWidth - 1) * 2 * pInstance->h) / cMinEnvWidth;
    if ((prevRight - prevLeft + 2) >= minEnvWidth)
	curVolume = prevRight - prevLeft + 2;
    firstSAsymmetry = MAX(0, prevRight) + MIN(0, prevLeft);
    curSAsymmetry = firstSAsymmetry;
    lastSAsymmetry = 0;

    while (pCurRun != NULL)
    {
	lineFlags ^= cLineTagMask;		/* move to the new line */
	if (pCurRun->runRight != prevRight)
	{
	    if (pCurRun->runRight < prevRight)
	    {
	        curSlant -= MIN(prevRight - pCurRun->runRight, 4);
	    }
	    else
	    {
		curSlant += MIN(pCurRun->runRight - prevRight, 4);
	    }
	    prevRight = pCurRun->runRight;
	}

	    /* go find the left end of this scan line */
	while ((pCurRun != NULL) && ((Int32)pCurRun->pClient == lineFlags))
	{
	    pPrevRun = pCurRun;
	    pCurRun = pCurRun->pNext;
	}

	    /* check the left end of the scan line */
	if (pPrevRun->runLeft != prevLeft)
	{
	    if (pPrevRun->runLeft < prevLeft)
		curSlant -= MIN(prevLeft - pPrevRun->runLeft, 4);
	    else
		curSlant += MIN(pPrevRun->runLeft - prevLeft, 4);
	    prevLeft = pPrevRun->runLeft;
	}
	if ((prevRight - prevLeft + 2) >= minEnvWidth)
	    curVolume += prevRight - prevLeft + 2;
	lastSAsymmetry = MAX(0, prevRight) + MIN(0, prevLeft);
	curSAsymmetry += lastSAsymmetry;
    }
	/* Since we're rotated by 90 degrees, the sign of the slant changed,
	 * so subtract it instead of adding it.  Normalize the asymmetry
	 * for shifting of the character by subtracting off the signed
	 * horizontal and vertical asymmetries.
	 */
    pInstance->slant -= curSlant;
    pInstance->vVolume = curVolume;
	/* Should we check the horizontal volume?  This is only done if
	 * is is a sufficiently large fraction of the total area.
	 */
    if (area <= curVolume)
	pInstance->checkVVolume = cTrue;
    else
	pInstance->checkVVolume = cFalse;

	/* We calculate the vertical asymmetry the same way as the
	 * horizontal asymmetry, except the positions of w and h are
	 * reversed.  Well, not quite.  Remember that we want the
	 * distance to the BOTTOM of the symbol and that's where the
	 * positive coordinates are.  Therefore, the larger the
	 * curSAsymmetry, the smaller the distance to the bottom.
	 * Therefore, we subtract curSAsymmetry from the height to
	 * get the distance.
	 */
    curSAsymmetry -= firstSAsymmetry + lastSAsymmetry;
    if (pInstance->w > 2)
	curSAsymmetry = 2*curSAsymmetry / (pInstance->w - 2);
    else
	curSAsymmetry = 0;

    pInstance->vSAsymmetry = -curSAsymmetry + 2 * pInstance->h;

	/* Calculate stroke width from vertical runs (which are still
	 * in the cache, right?) if the symbol is small enough.
	 */
    if (pInstance->maxDimension < 128)
    {
	for (pCurRun = pInstance->pVertRuns;
	     pCurRun != NULL;
	     pCurRun = pCurRun->pNext)
	{
		/* to use the histogram efficiently, we want 1x coords */
	    index = ((pCurRun->runRight - pCurRun->runLeft) / 2) + 1;
	    runLenHist[index]++;
	    runMassHist[index] += index;
	}

	    /* See which runs are needed to calculate the stroke width. */
	for (index = 0, area = 0, curVolume = 0;
	     area <
		(cStrokeWidthSlopNum * pInstance->mass) / cStrokeWidthSlopDen;
	     index++)
	{
	    area += runMassHist[index];		/* increase area coverage */
	    curVolume += runLenHist[index];	/* and number of runs */
	}

	    /* Calculate the average stroke width */
	pInstance->strokeWidth = (area << cStrokeWidthFracDigits) /
				  curVolume;
    }
    else	/* initialize the stroke Width to 0.  That's as good
		 * a guess as any.
		 */
    {
	pInstance->strokeWidth = 0;
    }


	/* We get here after looking at all the runs.  Store the result
	 * and return */
    NORMAL_RETURN;
}



/* PRIVATE */
/**************************************************************************
 *   ip_makeGroupsFromRuns()
 *
 *	This procedure is called for to create either horizontal or
 *	vertical groups from a set of horizontal or vertical runs.
 *	The groups are used to classify the symbol.
 *
 *	Input:
 *		pTextCompInfo:		Ptr to all text compression info
 *		pInstance:		Ptr to info about symbol.
 *		pp1stGroup:		Ptr to ptr to first resulting group.
 *					This is the vehicle for returning the
 *					list of groups.
 *		ppDupGroup:		Ptr to ptr to group holding duplicate
 *					of groups in first scan line.  Used
 *					to shift exemplar up or left by 1 line.
 *		makeHorizRuns:		cTrue => make groups from horiz runs.
 *					cFalse => make groups from vert runs
 *
 *	Action:
 *		Compose a set of run groups for each line of the symbol.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_makeGroupsFromRuns (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCInstance *pInstance,
    TCGroupRun **pp1stGroup,
    TCGroupRun **ppDupGroup,
    Int32	makeHorizRuns)
{
PROLOGUE;
 TCScanLineRun	*pRuns;				/* ptr to raw runs */
TCScanLineRun	*pAbove, *pMid, *pBelow;	/* ptrs to runs in the
						 * three current scan lines
						 */
TCScanLineRun	*pAboveRight, *pAboveLeft;	/* the rightmost and leftmost
						 * runs in the top line of
						 * the group. */
TCScanLineRun	*pMidRight, *pMidLeft;		/* the rightmost and leftmost
						 * runs in the middle line of
						 * the group. */
TCScanLineRun	*pBelowRight, *pBelowLeft;	/* the rightmost and leftmost
						 * runs in the bottom line of
						 * the group. */
TCScanLineRun	*pEndRun;			/* points to run at the
						 * end of the list */
Int16		 leftPt;			/* the leftmost point of
						 * the group with noise
						 * tolerance built in. */
Int16		 rightPt;			/* the rightmost point of
						 * the group with noise
						 * tolerance built in. */
Int16		 leftTight, rightTight;		/* the tight ranges for the
						 * left and right endpoints */
TCScanLineRun	*pPrevRun;			/* used to follow runs */
TCGroupRun	*pGroup;			/* ptr to the group being
						 * composed. */
TCGroupRun	*pDupGroup;			/* Used to compose groups
						 * that duplicate those in
						 * the first scan line. */
UInt32		 flags;				/* used to tag runs */
Int16		 smallNoise, bigNoise, endNoise; /* how much noise to allow */


	/* A group describes the horizontal extent of a set of overlapping
	 * runs on three contiguous lines.  For example, if
	 * one run goes from [-1, 3] on one line and the run above
	 * it goes from [0, 4], the group covers [-1, 4].
	 *
	 * Each set of overlapping runs on three scan lines produces
	 * 1 or more groups.  Most cases produce 1 group.  Only when
	 * a run on the top or bottom of a stroke actually bridges two
	 * strokes do we produce more than one group - a group for
	 * each portion above/below a bridged stroke and a group for
	 * each gap between strokes.
	 *
	 * BUG FIX: the concept of "overlap" was slightly broken, as
	 * illustrated with synthetic images containing diagonal strokes
	 * that were one pixel wide.  We were saying that two strokes
	 * overlapped if the left endpoint of one stroke minus either
	 * bigNoise (for a run in the middle of a group) or smallNoise
	 * (for the top or bottom runs of a group) overlapped the
	 * right endpoint of the other run.  For most smallish things,
	 * smallNoise = 1.  The distance between strokes that are only
	 * connected diagonally is 2.  Therefore, things that were
	 * only connected diagonally weren't overlapping.  This led to
	 * the production of lots of inappropriate noise-checking groups
	 * that discarded runs they should have kept and generally got
	 * confused.  Besides, deciding whether things connect using
	 * a measure that changes with the width of the symbol is kind
	 * of weird anyway.  The fix is to decree that overlap really
	 * means 2.  To do this, we change the meaning of leftPt to
	 * be the actual leftmost endpoint found so far - 2.  Before
	 * using leftPt to set the group's tolerances, we adjust it
	 * by bigNoise-2 if leftPt+2 was equal to pMidLeft->runLeft
	 * or by smallNoise-2 if it was a top or bottom run endpoint.
	 *
	 * Each group has a type, part of which is whether the middle line
	 * represented by the group is on an even or odd scan line of
	 * the symbol.  The type is used to select the code used to match
	 * the group to one or more runs from the instance symbol.
	 * Groups also contain endpoint ranges which specify where endpoints
	 * of instance runs must lie and a discardBdry which specifies
	 * (in some groups) which runs should be discarded.
	 *
	 * Note that we don't have to worry about scan line boundaries
	 * in the instance or the exemplar.  Each run and group is
	 * tagged with line tags.  These will match in all valid
	 * comparisons.  Ideally, we cruise through both lists,
	 * matching them as we go along.
	 * How does the matching really work?  There are several
	 * types of groups.  Each one does a particular type of
	 * matching against runs and consumes one or more runs.
	 * The types are:
	 *
	 *    * cCheckNoise
	 *	At the edges of symbols, we allow noise.  These
	 *	groups are used to make sure that if noise does
	 *	exist there, it is confined to legal areas.  To
	 *	allow us to use this type of run for two purposes,
	 *	only consume runs that are entirely to the right
	 *	of the group's discardBdry.
	 *
	 *    * cCheck1Run
	 *	Most groups represent isolated vertical strokes.
	 *	This type of group expects to find exactly one run
	 *	in this range.  This group checks the endpoint of
	 *	the next run.  If they match, we consume the run
	 *	and continue.  If they fail, we fail.
	 *
	 *    *	cCheckEndpoints
	 *	Near forks and joins, we allow the scan lines at
	 *	which the fork or join happens to move up or down.
	 *	This type of group expects to find a run matching
	 *	the right endpoint range, some number of runs in
	 *	the middle and a run (possibly the same run as the
	 *	first one) matching the left endpoint range.  All
	 *	runs are consumed.
	 *
	 *    * cCheck1RunNoAdvance
	 *	Sometimes, the run on the top or bottom of a symbol
	 *	is the only one connecting two strokes.  We use
	 *	this type of group to check to make sure the gap is
	 *	bridged.  The gap is usually made up of some space
	 *	under the stroke on the right, the gap, then some
	 *	space under the stroke on the left.  We use a
	 *	cCheckNoise group to consume any runs contained
	 *	entirely in the space to the right.  A 
	 *	cCheck1RunNoAdvance group checks the gap and another
	 *	cCheckNoise checks for noise on the left.  This
	 *	continues for all gaps.
	 *
	 * There are actually two groups for each type mentioned
	 * above.  One consumes scan lines on even runs and one
	 * consumes scan lines on odd runs.  Yeah, we could just
	 * use a variable to check the scan line polarity, but
	 * I think this may be faster.
	 *
	 * How does the matching process for each type of group
	 * really work?
	 *
	 * cCheckNoise:
	 *   1.	If the current run pointer is NULL or the run it
	 *	points to is on another scan line or is on the
	 *	same scan line but doesn't overlap this group,
	 *	Pass.  We just go on to the next group.
	 *   2.	If the runs overlaps this group, make sure it
	 *	is covered by the group.
	 *   3.	Consume all runs to the right of discardBdry.
	 *
	 * cCheck1Run:
	 *   1.	If the current run pointer is NULL or the run it
	 *	points to is on another scan line or is on the
	 *	same scan line but has endpoints not inside the
	 *	range of legal endpoints, Fail.  That means
	 *	we try another symbol.
	 *   2. If those checks succeeded, go on to the next group
	 *	and the next run.
	 *
	 * cCheckEndpoints:
	 *   1.	If the current run pointer is NULL or the run it
	 *	points to is on another scan line or is on the
	 *	same scan line but has a right endpoint not inside the
	 *	range of legal right endpoints, Fail.  That means
	 *	we try another symbol.
	 *   2.	Consume all runs that are on the same scan line
	 *	and are to the right of the maximum legal left
	 *	endpoint.
	 *   3. If the next run's left endpoint falls outside the
	 *	range of legal left endpoints, Fail.
	 *   4.	If that check succeeded, go on to the next group
	 *	and run.
	 *
	 * cCheck1RunNoAdvance:
	 *   1.	If the current run pointer is NULL or the run it
	 *	points to is on another scan line or is on the
	 *	same scan line but has endpoints not inside the
	 *	range of legal endpoints, Fail.  That means
	 *	we try another symbol.
	 *   2. If those checks succeeded, go on to the next group
	 *	BUT NOT the next run.
	 *
	 * What happens if we run out of groups before running out of
	 * instance runs?  There are enough groups to cover noise on
	 * the top of the exemplar.  If we run out of them before
	 * running out of instance runs, the instance is too tall,
	 * so we look for another symbol.
	 *
	 * What happens if we run out of instance runs before running
	 * out of groups?  Since there are extra groups, this is likely
	 * to happen.  All of the groups on the top of the exemplar
	 * are cCheckNoise groups.  If there are no instance runs
	 * there, we happily skip them.
	 *
	 * A dummy node with both even and odd flags set is appended
	 * to the end of the list.  The entire process stops when
	 * the top and middle pointers are equal after processing
	 * a scan line.
	 */

    *pp1stGroup = *ppDupGroup = NULL;	/* initialize so we can remember
					 * what to clean up. */

	/* point to the horizontal or vertical runs as required */
    if (makeHorizRuns)
	pRuns = pInstance->pHorizRuns;
    else
	pRuns = pInstance->pVertRuns;

	/* We assume the line flags are already set in the runs.
	 * However, we need to find the last run so we can put the
	 * dummy run on the end.  While we're doing this, find the
	 * largest and smallest coordinates.  These determine the
	 * amount of noise we'll allow.
	 */
    leftPt = pRuns->runLeft;
    rightPt = pRuns->runRight;
    for (pMid = pRuns; pMid->pNext != NULL; pMid = pMid->pNext)
    {
	if (pMid->runLeft < leftPt)
	    leftPt = pMid->runLeft;
	if (pMid->runRight > rightPt)
	    rightPt = pMid->runRight;
    }

	/* Set the noise margin.  We allow smallNoise slop when matching a
	 * run against runs above or below it.  We allow bigNoise slop
	 * when matching a run against an exemplar run on the same line.
	 * We allow endNoise slop on the right and left edges of the top
	 * and bottom runs of a symbol.
	 */
    smallNoise = MIN(cMaxSmallNoise, (rightPt-leftPt)/cSmallNoiseDivisor) +
		 (rightPt - leftPt)/cBigSymbolDivisor;
    bigNoise = MIN(cMaxBigNoise, (rightPt-leftPt)/cBigNoiseDivisor) +
		 (rightPt - leftPt)/cBigSymbolDivisor;
    endNoise = MIN(cMaxEndNoise, (rightPt-leftPt)/cEndNoiseDivisor) +
		 (rightPt - leftPt)/cBigSymbolDivisor;

	/* add the dummy node onto the end of the list */
    pEndRun = pMid;		/* remember where it's attached */
    __error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
				(void **) &(pMid->pNext));
    if (__error__ != ia_successful)
	goto tcMakeGroupsCleanup;

    pMid = pMid->pNext;
    pMid->pClient = NULL;
    pMid->pNext = NULL;
    pMid->runRight = pMid->runLeft = -cVeryLargeNum;	/* useful to avoid
							 * setting group
							 * limits */

	/* Make the group that's centered on the bottom scan line.  In this
	 * loop, we ignore pBelow since it doesn't point to anything
	 * by definition.  First, make pAbove point to the first run of the
	 * second scan line.
	 */
    for (pAbove = pRuns->pNext;
	 (UInt32) pAbove->pClient == cEvenLineTag;
	 pAbove = pAbove->pNext)
    {}
    pMid = pRuns;

	/* allocate a dummy group since the loops below always
	 * assume we're adding a group, not putting in the first
	 * group.  We'll toss this one later.
	 */
    __error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stGroupBlock,
			        (void **) &pGroup);
    if (__error__ != ia_successful)
	goto tcMakeGroupsCleanup;

    pGroup->leftEndMin = cVeryLargeNum;	/* remember that this is
					 * is the dummy */
    *pp1stGroup = pGroup;
    *ppDupGroup = pGroup;

	/* loop through the first two scan lines.  This loop increments
	 * once per new group.
	 */
    while (((UInt32) pAbove->pClient == cOddLineTag) ||
	   ((UInt32) pMid->pClient == cEvenLineTag))
    {
	pAboveRight = pAboveLeft = NULL;	/* initialize run ptrs */
	pMidRight = pMidLeft = NULL;

	    /* Initialize the group from the rightmost run */
	if ( ((UInt32) pMid->pClient == cEvenLineTag) &&
	     (((UInt32) pAbove->pClient != cOddLineTag) ||
	      (pMid->runRight >= pAbove->runRight)) )
	{
		/* allow extra 1.5 pixel slop for the middle line */
	    rightPt = pMid->runRight + bigNoise;
	    leftPt = pMid->runLeft - 2;
	    pMidRight = pMidLeft = pMid;
	    pMid = pMid->pNext;
	    flags = cGroupRunMiddle | cEvenLineTag;
	}
	else
	{
	    rightPt = pAbove->runRight + smallNoise;
	    leftPt = pAbove->runLeft - 2;
	    pAboveRight = pAboveLeft = pAbove;
	    pAbove = pAbove->pNext;
	    flags = cGroupRunAbove | cEvenLineTag;
	}

	    /* while there are more overlapping runs in the group, track
	     * them.  We keep going until the group size doesn't change
	     * in a pass through the loop.
	     */
	__error__ = cTrue;		/* we have just made the group bigger */
	while (__error__)
	{
	    __error__ = cFalse;		/* assume no changes */
	    if (((UInt32) pAbove->pClient == cOddLineTag) &&
		(pAbove->runRight >= leftPt))
	    {
		pPrevRun = pAbove;
		if (pAboveRight == NULL)
		    pAboveRight = pAbove;
		pAbove = pAbove->pNext;
		while (((UInt32) pAbove->pClient == cOddLineTag) &&
		       (pAbove->runRight >= leftPt))
		{
		    pPrevRun = pAbove;
		    pAbove = pAbove->pNext;
		}
		if (pPrevRun->runLeft - 2 < leftPt)
		{
		    leftPt = pPrevRun->runLeft - 2;
		    __error__ = cTrue;	/* the group has been changed */
		}
		pAboveLeft = pPrevRun;
		flags |= cGroupRunAbove;
	    }

	    if (((UInt32) pMid->pClient == cEvenLineTag) &&
		(pMid->runRight >= leftPt))
	    {
		pPrevRun = pMid;
		if (pMidRight == NULL)
		    pMidRight = pMid;
		pMid = pMid->pNext;
		while (((UInt32) pMid->pClient == cEvenLineTag) &&
		       (pMid->runRight >= leftPt))
		{
		    pPrevRun = pMid;
		    pMid = pMid->pNext;
		}
		if (pPrevRun->runLeft - 2 < leftPt)
		{
		    leftPt = pPrevRun->runLeft - 2;
		    __error__ = cTrue;	/* the group has been changed */
		}
		pMidLeft = pPrevRun;
		flags |= cGroupRunMiddle;
	    }
	}	/* end of adding extra runs to a group */

	    /* Remove the 2 from leftPt and adjust by proper
	     * amount of noise.
	     */
	leftPt += 2;
	if ((pMidLeft != NULL) && (leftPt == pMidLeft->runLeft))
	    leftPt -= bigNoise;
	else
	    leftPt -= smallNoise;

	    /* At this point, we have pointers to the leftmost and
	     * rightmost runs on each scan line.  See if we are going
	     * to make 1 or 3 groups.  If the group contains only a
	     * single run (as will happen if it contains runs on
	     * only one scan line), adjust the line pointers to
	     * avoid further checking for NULL.
	     */
	if ((pMidLeft == pMidRight) && (pAboveLeft == pAboveRight))
	{	/* make one group */
	    __error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stGroupBlock,
				        (void **) &(pGroup->pNext));

	    if (__error__ != ia_successful)
		goto tcMakeGroupsCleanup;

	    pGroup = pGroup->pNext;
	    pGroup->leftEndMin = leftPt;
	    pGroup->rightEndMax = rightPt;

	    if ((pAboveLeft != NULL) && (pMidLeft != NULL))
	    {
		pGroup->leftEnd = pMidLeft->runLeft;
		pGroup->rightEnd = pMidLeft->runRight;
		    /* if the bottom line sticks out the farthest, keep
		     * a tight bound on it.  Otherwise, use a loose bound
		     * for noise.
		     */
		if ((pAboveLeft->runRight < pMidLeft->runRight) ||
		    (pAboveLeft->runLeft > pMidLeft->runLeft))
		{
		    pGroup->groupType = cCheckEndpoints;
		    if (pAboveLeft->runRight < pMidLeft->runRight)
			pGroup->rightEndMin = pMidLeft->runRight - smallNoise;
		    else
			pGroup->rightEndMin = leftPt;

		    if (pAboveLeft->runLeft > pMidLeft->runLeft)
			pGroup->leftEndMax = pMidLeft->runLeft + smallNoise;
		    else
			pGroup->leftEndMax = rightPt;
		}
		else
		{
		    pGroup->discardBdry = leftPt;
		    pGroup->groupType = cCheckNoise;
		}
	    }
	    else
	    if (pMidLeft == NULL) /* any noise below here is ok */
	    {
		pGroup->leftEnd = pAboveLeft->runLeft;
		pGroup->rightEnd = pAboveLeft->runRight;
		pGroup->discardBdry = leftPt;
		pGroup->groupType = cCheckNoise;
	    }
	    else		/* happens in symbols that are 1 line tall */
	    {
		pGroup->leftEnd = pMidLeft->runLeft;
		pGroup->rightEnd = pMidLeft->runRight;
		pGroup->leftEndMax = pMidLeft->runLeft + bigNoise;
		    /* in this step, we can use pMidLeft because it's
		     * equal to pMidRight and we don't want the compiler
		     * accessing yet another variable.
		     */
		pGroup->rightEndMin = pMidLeft->runRight - bigNoise;
		pGroup->groupType = cCheck1Run;
	    }
	}
	else	/* make several groups */
	{
		/* For each run in the Above line, put out a group saying
		 * that there's only runs on the line above and giving
		 * endpoint ranges that cover the entire group.  For
		 * each gap, put out a group saying there are runs
		 * in the Above and Middle line with endpoint ranges
		 * starting just inside the gap and going out to the ends.
		 */
	    for (;
		 pAboveRight != pAboveLeft;
		 pAboveRight = pAboveRight->pNext)
	    {
		    /* make a group to catch noise below pAboveRight */
		__error__ = i_allocateStruct (pIPState,
					    pTextCompInfo->p1stGroupBlock,
					    (void **) &(pGroup->pNext));

		if (__error__ != ia_successful)
		    goto tcMakeGroupsCleanup;

		pGroup = pGroup->pNext;
		pGroup->leftEndMin = leftPt;
		pGroup->rightEndMax = rightPt;
		pGroup->discardBdry = pAboveRight->runLeft;
		pGroup->groupType = cCheckNoise;
		pGroup->leftEnd = pMidLeft->runLeft;
		pGroup->rightEnd = pMidRight->runRight;

		    /* make a group to bridge to the next run above */
		__error__ = i_allocateStruct (pIPState,
					    pTextCompInfo->p1stGroupBlock,
					    (void **) &(pGroup->pNext));
    
		if (__error__ != ia_successful)
		    goto tcMakeGroupsCleanup;
    
		pGroup = pGroup->pNext;
		pGroup->leftEndMin = leftPt;
		pGroup->leftEndMax = pAboveRight->pNext->runRight + bigNoise ;
		pGroup->rightEndMin = pAboveRight->runLeft - bigNoise;
		pGroup->rightEndMax = rightPt;
		pGroup->groupType = cCheck1RunNoAdvance;
		pGroup->leftEnd = pMidLeft->runLeft;
		pGroup->rightEnd = pMidRight->runRight;
	    }


		/* make the group to detect noise below pAboveLeft */
	    __error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stGroupBlock,
				        (void **) &(pGroup->pNext));
    
	    if (__error__ != ia_successful)
		goto tcMakeGroupsCleanup;
    
	    pGroup = pGroup->pNext;
	    pGroup->leftEndMin = leftPt;
	    pGroup->rightEndMax = rightPt;
	    pGroup->discardBdry = leftPt;
	    pGroup->groupType = cCheckNoise;
	    pGroup->leftEnd = pMidLeft->runLeft;
	    pGroup->rightEnd = pMidRight->runRight;
	}
    }	/* end of loop that makes groups for the first line */

	/* *************** MAIN LOOP *********************** */

	/* there are no more runs in this group.  Set the pointer to
	 * the line below and start the main group-making loop.
	 */
    pBelow = pRuns;
    while (cTrue)
    {
	while (((UInt32) pAbove->pClient == cEvenLineTag) ||
	       ((UInt32) pMid->pClient == cOddLineTag) ||
	       ((UInt32) pBelow->pClient == cEvenLineTag))
	{
	    pAboveRight = pAboveLeft = NULL;	/* initialize run ptrs */
	    pMidRight = pMidLeft = NULL;
	    pBelowRight = pBelowLeft = NULL;

		/* Initialize the group from the rightmost run.  We
		 * have to check each pointer to see if it's still
		 * pointing to the correct line before using the
		 * run it points to.
		 */
	    if ( ((UInt32) pMid->pClient == cOddLineTag) &&
		 (((UInt32) pAbove->pClient != cEvenLineTag) ||
		  (pMid->runRight >= pAbove->runRight)) &&
		 (((UInt32) pBelow->pClient != cEvenLineTag) ||
		  (pMid->runRight >= pBelow->runRight)) )
	    {	/* pMid points to the rightmost legal run in the group */
		    /* allow extra 1.5 pixel slop for the middle line */
		rightPt = pMid->runRight + bigNoise;
		leftPt = pMid->runLeft - 2;
		pMidRight = pMidLeft = pMid;
		pMid = pMid->pNext;
		flags = cGroupRunMiddle;
	    }
	    else
	    if ( ((UInt32) pAbove->pClient == cEvenLineTag) &&
		 (((UInt32) pBelow->pClient != cEvenLineTag) ||
		  (pAbove->runRight >= pBelow->runRight)) )
	    {	/* pAbove points to the rightmost legal run in the group */
		rightPt = pAbove->runRight + smallNoise;
		leftPt = pAbove->runLeft - 2;
		pAboveRight = pAboveLeft = pAbove;
		pAbove = pAbove->pNext;
		flags = cGroupRunAbove;
	    }
	    else
	    {	/* pBelow points to the rightmost legal run in the group */
		rightPt = pBelow->runRight + smallNoise;
		leftPt = pBelow->runLeft - 2;
		pBelowRight = pBelowLeft = pBelow;
		pBelow = pBelow->pNext;
		flags = cGroupRunBelow;
	    }

		/* while there are more overlapping runs in the group, track
		 * them.  We keep going until the group size doesn't change
		 * in a pass through the loop.
		 */
	    __error__ = cTrue;		/* we have just made the group bigger */
	    while (__error__)
	    {
		__error__ = cFalse;		/* assume no changes */
		if (((UInt32) pAbove->pClient == cEvenLineTag) &&
		    (pAbove->runRight >= leftPt))
		{
		    pPrevRun = pAbove;
		    if (pAboveRight == NULL)
			pAboveRight = pAbove;
		    pAbove = pAbove->pNext;
		    while (((UInt32) pAbove->pClient == cEvenLineTag) &&
			   (pAbove->runRight >= leftPt))
		    {
			pPrevRun = pAbove;
			pAbove = pAbove->pNext;
		    }
		    if (pPrevRun->runLeft - 2 < leftPt)
		    {
			leftPt = pPrevRun->runLeft - 2;
			__error__ = cTrue;	/* the group has been changed */
		    }
		    pAboveLeft = pPrevRun;
		    flags |= cGroupRunAbove;
		}
    
		if (((UInt32) pMid->pClient == cOddLineTag) &&
		    (pMid->runRight >= leftPt))
		{
		    pPrevRun = pMid;
		    if (pMidRight == NULL)
			pMidRight = pMid;
		    pMid = pMid->pNext;
		    while (((UInt32) pMid->pClient == cOddLineTag) &&
			   (pMid->runRight >= leftPt))
		    {
			pPrevRun = pMid;
			pMid = pMid->pNext;
		    }
		    if (pPrevRun->runLeft - 2 < leftPt)
		    {
			leftPt = pPrevRun->runLeft - 2;
			__error__ = cTrue;	/* the group has been changed */
		    }
		    pMidLeft = pPrevRun;
		    flags |= cGroupRunMiddle;
		}

		if (((UInt32) pBelow->pClient == cEvenLineTag) &&
		    (pBelow->runRight >= leftPt))
		{
		    pPrevRun = pBelow;
		    if (pBelowRight == NULL)
			pBelowRight = pBelow;
		    pBelow = pBelow->pNext;
		    while (((UInt32) pBelow->pClient == cEvenLineTag) &&
			   (pBelow->runRight >= leftPt))
		    {
			pPrevRun = pBelow;
			pBelow = pBelow->pNext;
		    }
		    if (pPrevRun->runLeft - 2 < leftPt)
		    {
			leftPt = pPrevRun->runLeft - 2;
			__error__ = cTrue;	/* the group has been changed */
		    }
		    pBelowLeft = pPrevRun;
		    flags |= cGroupRunBelow;
		}
	    }	/* end of adding extra runs to a group */

		/* Remove the 2 from leftPt and adjust by proper
		 * amount of noise.
		 */
	    leftPt += 2;
	    if ((pMidLeft != NULL) && (leftPt == pMidLeft->runLeft))
		leftPt -= bigNoise;
	    else
		leftPt -= smallNoise;

		/* At this point, we have pointers to the leftmost and
		 * rightmost runs on each scan line.  If there is no
		 * more than one run on any of the scan lines,
		 * we make a single simple group.
		 */
	    if ((pAboveLeft == pAboveRight) &&
		(pMidLeft == pMidRight) &&
		(pBelowLeft == pBelowRight))
	    {	/* make one group */
		__error__ = i_allocateStruct (pIPState,
					    pTextCompInfo->p1stGroupBlock,
					    (void **) &(pGroup->pNext));
	
		if (__error__ != ia_successful)
		    goto tcMakeGroupsCleanup;

		pGroup = pGroup->pNext;

		switch (flags)
		{
		case 1:		/* only runs in line below, allow lots of
				 * slop on this line */
		    pGroup->leftEndMin = leftPt - endNoise;
		    pGroup->rightEndMax = rightPt + endNoise;
		    pGroup->discardBdry = leftPt - endNoise;
		    pGroup->groupType = cOddLineTag | cCheckNoise;
		    pGroup->leftEnd = pBelowLeft->runLeft;
		    pGroup->rightEnd = pBelowRight->runRight;
		    break;

		case 2:		/* only runs in the middle line.  Allow
				 * very little slop */
		    pGroup->leftEndMin = leftPt;
		    pGroup->leftEndMax = pMidLeft->runLeft + bigNoise;
		    pGroup->rightEndMin = pMidLeft->runRight - bigNoise;
		    pGroup->rightEndMax = rightPt;
		    pGroup->groupType = cOddLineTag | cCheck1Run;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;

		    break;

		case 3:		/* runs on the bottom and middle lines only */
				/* if the middle line sticks out farther than
				 * the bottom line on either end, require
				 * it to be here, otherwise, pretend it's just
				 * noise.
				 */
		    pGroup->leftEndMin = leftPt;
		    pGroup->rightEndMax = rightPt;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;
		    if ((pBelowLeft->runRight < pMidLeft->runRight) ||
			(pBelowLeft->runLeft > pMidLeft->runLeft))
		    {
			pGroup->groupType = cOddLineTag | cCheckEndpoints;
			if (pBelowLeft->runRight < pMidLeft->runRight)
			    pGroup->rightEndMin = pMidLeft->runRight -
								smallNoise;
			else
			    pGroup->rightEndMin = leftPt;

			if (pBelowLeft->runLeft > pMidLeft->runLeft)
			    pGroup->leftEndMax = pMidLeft->runLeft + smallNoise;
			else
			    pGroup->leftEndMax = rightPt;
		    }
		    else
		    {
			pGroup->discardBdry = leftPt;
			pGroup->groupType = cOddLineTag | cCheckNoise;
		    }

		    break;

		case 4:		/* runs on the top line only, allow lots of
				 * slop on this line */
		    pGroup->leftEndMin = leftPt - endNoise;
		    pGroup->rightEndMax = rightPt + endNoise;
		    pGroup->discardBdry = leftPt - endNoise;
		    pGroup->groupType = cOddLineTag | cCheckNoise;
		    pGroup->leftEnd = pAboveLeft->runLeft;
		    pGroup->rightEnd = pAboveRight->runRight;
		    break;

		case 5:		/* no middle line.  There should be no
				 * instance run here.  If there is one,
				 * the next match will fail.
				 */
		    break;

		case 6:		/* runs on the top and middle lines only.
				 * As before, be strict if the middle line
				 * sticks out more than the top line.
				 */
		    pGroup->leftEndMin = leftPt;
		    pGroup->rightEndMax = rightPt;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;
		    if ((pAboveLeft->runRight < pMidLeft->runRight) ||
			(pAboveLeft->runLeft > pMidLeft->runLeft))
		    {
			pGroup->groupType = cOddLineTag | cCheckEndpoints;
			if (pAboveLeft->runRight < pMidLeft->runRight)
			    pGroup->rightEndMin = pMidLeft->runRight -
								smallNoise;
			else
			    pGroup->rightEndMin = leftPt;

			if (pAboveLeft->runLeft > pMidLeft->runLeft)
			    pGroup->leftEndMax = pMidLeft->runLeft + smallNoise;
			else
			    pGroup->leftEndMax = rightPt;
		    }
		    else
		    {
			pGroup->discardBdry = leftPt;
			pGroup->groupType = cOddLineTag | cCheckNoise;
		    }

		    break;

		case 7:		/* runs on all three lines, the normal case. */
		    if ((pMidLeft->runLeft < pAboveLeft->runLeft) &&
			(pMidLeft->runLeft < pBelowLeft->runLeft))
			leftTight = pMidLeft->runLeft + bigNoise;
		    else
			leftTight = MAX(pAboveLeft->runLeft + smallNoise,
					MAX(pMidLeft->runLeft + bigNoise,
					    pBelowLeft->runLeft + smallNoise));

		    if ((pMidLeft->runRight > pAboveLeft->runRight) &&
			(pMidLeft->runRight > pBelowLeft->runRight))
			rightTight = pMidLeft->runRight - bigNoise;
		    else
			rightTight = MIN(pAboveLeft->runRight - smallNoise,
					MIN(pMidLeft->runRight - bigNoise,
					    pBelowLeft->runRight - smallNoise));

			/* assign the main group */
		    pGroup->leftEndMin = leftPt;
		    pGroup->leftEndMax = leftTight;
		    pGroup->rightEndMin = rightTight;
		    pGroup->rightEndMax = rightPt;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;
			/* if either of the ranges is big, use cCheckEndpoints
			 * instead of cCheck1Run.  cCheckEndpoints allows
			 * for the presence of noise blobs in the ranges.
			 */
		    if ((leftTight - leftPt > cMinNoiseEdgeLength) ||
			(rightPt - rightTight > cMinNoiseEdgeLength))
			pGroup->groupType = cOddLineTag | cCheckEndpoints;
		    else
			pGroup->groupType = cOddLineTag | cCheck1Run;

		    break;

		default:	/* should never get here */
		    break;
		}
	    }
	    else	/* make three groups, odd line tags */
	    {
		    /* first, figure out what values to use for the
		     * endpoint ranges.
		     */
		switch (flags)
		{
		case 3:		/* runs on the bottom and middle lines only */

			/* Since the code for cases 3 and 6 is virtually
			 * identical, just set pAboveRight to pBelowRight
			 * and pAboveLeft to pBelowLeft and drop through.
			 * These pointers aren't used after the groups are
			 * generated anyway.
			 */
		    pAboveRight = pBelowRight;
		    pAboveLeft = pBelowLeft;

		case 6:		/* runs on the top and middle lines only */
			/* For each run in the top line, put out a
			 * group saying that there's only runs on the top
			 * line and giving endpoint ranges that cover the
			 * entire group.  For each gap between top runs,
			 * put out a group saying there are runs
			 * in the Bottom and Middle line with endpoint ranges
			 * starting just inside the gap and going out to the
			 * ends.
			 */
		    for (;
			 pAboveRight != pAboveLeft;
			 pAboveRight = pAboveRight->pNext)
		    {
			    /* make a group to catch noise below pAboveRight */
			__error__ =
			    i_allocateStruct (pIPState,
					      pTextCompInfo->p1stGroupBlock,
					      (void **) &(pGroup->pNext));
	    
			if (__error__ != ia_successful)
			    goto tcMakeGroupsCleanup;

			pGroup = pGroup->pNext;
			pGroup->leftEndMin = leftPt;
			pGroup->rightEndMax = rightPt;
			pGroup->discardBdry = pAboveRight->runLeft;
			pGroup->groupType = cOddLineTag | cCheckNoise;
			pGroup->leftEnd = pMidLeft->runLeft;
			pGroup->rightEnd = pMidRight->runRight;

			    /* make a group to bridge to the next run below */
			__error__ =
			    i_allocateStruct (pIPState,
					      pTextCompInfo->p1stGroupBlock,
					      (void **) &(pGroup->pNext));
	    
			if (__error__ != ia_successful)
			    goto tcMakeGroupsCleanup;
	    
			pGroup = pGroup->pNext;
			pGroup->leftEndMin = leftPt;
			pGroup->leftEndMax = pAboveRight->pNext->runRight
								+ bigNoise ;
			pGroup->rightEndMin = pAboveRight->runLeft
								- bigNoise;
			pGroup->rightEndMax = rightPt;
			pGroup->groupType = cOddLineTag | cCheck1RunNoAdvance;
			pGroup->leftEnd = pMidLeft->runLeft;
			pGroup->rightEnd = pMidRight->runRight;
		    }
	
	
			/* make the group to detect noise below pAboveLeft */
		    __error__ =
		        i_allocateStruct (pIPState,
					  pTextCompInfo->p1stGroupBlock,
					  (void **) &(pGroup->pNext));
	    
		    if (__error__ != ia_successful)
			goto tcMakeGroupsCleanup;

		    pGroup = pGroup->pNext;
		    pGroup->leftEndMin = leftPt;
		    pGroup->rightEndMax = rightPt;
		    pGroup->discardBdry = leftPt;
		    pGroup->groupType = cOddLineTag | cCheckNoise;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;

		    break;

		case 7:		/* runs on all three lines, the normal case. */
		    if ((pMidLeft->runLeft < pAboveLeft->runLeft) &&
			(pMidLeft->runLeft < pBelowLeft->runLeft))
			leftTight = pMidLeft->runLeft + bigNoise;
		    else
			leftTight = MAX(pAboveLeft->runLeft + smallNoise,
					MAX(pMidLeft->runLeft + bigNoise,
					    pBelowLeft->runLeft + smallNoise));

		    if ((pMidRight->runRight > pAboveRight->runRight) &&
			(pMidRight->runRight > pBelowRight->runRight))
			rightTight = pMidRight->runRight - bigNoise;
		    else
			rightTight = MIN(pAboveRight->runRight - smallNoise,
				      MIN(pMidRight->runRight - bigNoise,
					  pBelowRight->runRight - smallNoise));


			/* make the group */
		    __error__ =
			i_allocateStruct (pIPState,
					  pTextCompInfo->p1stGroupBlock,
					  (void **) &(pGroup->pNext));
    
		    if (__error__ != ia_successful)
			goto tcMakeGroupsCleanup;

		    pGroup = pGroup->pNext;
		    pGroup->leftEndMin = leftPt;
		    pGroup->leftEndMax = leftTight;
		    pGroup->rightEndMin = rightTight;
		    pGroup->rightEndMax = rightPt;
		    pGroup->groupType = cOddLineTag | cCheckEndpoints;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;

		    break;

		default:	/* Case 0, 1, 2 and 4 cannot happen.  Case 5
				 * causes the matching process to ignore
				 * all the endpoint settings
				 * except leftEndMin anyway.
				 */
		    break;
		}

	    }	/* end of case that makes three groups, odd line tags */
	}	/* end of loop that makes groups for lines ending in a
		 * flag of cOddLineTag
		 */

	    /* if pBelow, pMid and pAbove are now equal, we're done since we've
	     * since we've processed the very top line of the image in
	     * all three positions.
	     */
	if ((pMid == pAbove) && (pMid == pBelow))
	    break;


	    /**************** Even Line Copy Of MAIN LOOP *******************/

	while (((UInt32) pAbove->pClient == cOddLineTag) ||
	       ((UInt32) pMid->pClient == cEvenLineTag) ||
	       ((UInt32) pBelow->pClient == cOddLineTag))
	{
	    pAboveRight = pAboveLeft = NULL;	/* initialize run ptrs */
	    pMidRight = pMidLeft = NULL;
	    pBelowRight = pBelowLeft = NULL;

		/* Initialize the group from the rightmost run.  We
		 * have to check each pointer to see if it's still
		 * pointing to the correct line before using the
		 * run it points to.
		 */
	    if ( ((UInt32) pMid->pClient == cEvenLineTag) &&
		 (((UInt32) pAbove->pClient != cOddLineTag) ||
		  (pMid->runRight >= pAbove->runRight)) &&
		 (((UInt32) pBelow->pClient != cOddLineTag) ||
		  (pMid->runRight >= pBelow->runRight)) )
	    {	/* pMid points to the rightmost legal run in the group */
		    /* allow extra 1.5 pixel slop for the middle line */
		rightPt = pMid->runRight + bigNoise;
		leftPt = pMid->runLeft - 2;
		pMidRight = pMidLeft = pMid;
		pMid = pMid->pNext;
		flags = cGroupRunMiddle;
	    }
	    else
	    if ( ((UInt32) pAbove->pClient == cOddLineTag) &&
		 (((UInt32) pBelow->pClient != cOddLineTag) ||
		  (pAbove->runRight >= pBelow->runRight)) )
	    {	/* pAbove points to the rightmost legal run in the group */
		rightPt = pAbove->runRight + smallNoise;
		leftPt = pAbove->runLeft - 2;
		pAboveRight = pAboveLeft = pAbove;
		pAbove = pAbove->pNext;
		flags = cGroupRunAbove;
	    }
	    else
	    {	/* pBelow points to the rightmost legal run in the group */
		rightPt = pBelow->runRight + smallNoise;
		leftPt = pBelow->runLeft - 2;
		pBelowRight = pBelowLeft = pBelow;
		pBelow = pBelow->pNext;
		flags = cGroupRunBelow;
	    }

		/* while there are more overlapping runs in the group, track
		 * them.  We keep going until the group size doesn't change
		 * in a pass through the loop.
		 */
	    __error__ = cTrue;		/* we have just made the group bigger */
	    while (__error__)
	    {
		__error__ = cFalse;		/* assume no changes */
		if (((UInt32) pAbove->pClient == cOddLineTag) &&
		    (pAbove->runRight >= leftPt))
		{
		    pPrevRun = pAbove;
		    if (pAboveRight == NULL)
			pAboveRight = pAbove;
		    pAbove = pAbove->pNext;
		    while (((UInt32) pAbove->pClient == cOddLineTag) &&
			   (pAbove->runRight >= leftPt))
		    {
			pPrevRun = pAbove;
			pAbove = pAbove->pNext;
		    }
		    if (pPrevRun->runLeft - 2 < leftPt)
		    {
			leftPt = pPrevRun->runLeft - 2;
			__error__ = cTrue;	/* the group has been changed */
		    }
		    pAboveLeft = pPrevRun;
		    flags |= cGroupRunAbove;
		}
    
		if (((UInt32) pMid->pClient == cEvenLineTag) &&
		    (pMid->runRight >= leftPt))
		{
		    pPrevRun = pMid;
		    if (pMidRight == NULL)
			pMidRight = pMid;
		    pMid = pMid->pNext;
		    while (((UInt32) pMid->pClient == cEvenLineTag) &&
			   (pMid->runRight >= leftPt))
		    {
			pPrevRun = pMid;
			pMid = pMid->pNext;
		    }
		    if (pPrevRun->runLeft - 2 < leftPt)
		    {
			leftPt = pPrevRun->runLeft - 2;
			__error__ = cTrue;	/* the group has been changed */
		    }
		    pMidLeft = pPrevRun;
		    flags |= cGroupRunMiddle;
		}

		if (((UInt32) pBelow->pClient == cOddLineTag) &&
		    (pBelow->runRight >= leftPt))
		{
		    pPrevRun = pBelow;
		    if (pBelowRight == NULL)
			pBelowRight = pBelow;
		    pBelow = pBelow->pNext;
		    while (((UInt32) pBelow->pClient == cOddLineTag) &&
			   (pBelow->runRight >= leftPt))
		    {
			pPrevRun = pBelow;
			pBelow = pBelow->pNext;
		    }
		    if (pPrevRun->runLeft - 2 < leftPt)
		    {
			leftPt = pPrevRun->runLeft - 2;
			__error__ = cTrue;	/* the group has been changed */
		    }
		    pBelowLeft = pPrevRun;
		    flags |= cGroupRunBelow;
		}
	    }	/* end of adding extra runs to a group */

		/* Remove the 2 from leftPt and adjust by proper
		 * amount of noise.
		 */
	    leftPt += 2;
	    if ((pMidLeft != NULL) && (leftPt == pMidLeft->runLeft))
		leftPt -= bigNoise;
	    else
		leftPt -= smallNoise;

		/* At this point, we have pointers to the leftmost and
		 * rightmost runs on each scan line.  See if we are going
		 * to make 1 or 3 groups.  If the group contains only a
		 * single run (as will happen if it contains runs on
		 * only one scan line), adjust the line pointers to
		 * avoid further checking for NULL.
		 */
	    if ((pAboveLeft == pAboveRight) &&
		(pMidLeft == pMidRight) &&
		(pBelowLeft == pBelowRight))
	    {	/* make one group */
		__error__ = i_allocateStruct (pIPState,
					    pTextCompInfo->p1stGroupBlock,
					    (void **) &(pGroup->pNext));
	
		if (__error__ != ia_successful)
		    goto tcMakeGroupsCleanup;

		pGroup = pGroup->pNext;

		switch (flags)
		{
		case 1:		/* only runs in line below, allow lots of
				 * slop on this line */
		    pGroup->leftEndMin = leftPt - endNoise;
		    pGroup->rightEndMax = rightPt + endNoise;
		    pGroup->discardBdry = leftPt - endNoise;
		    pGroup->groupType = cCheckNoise;
		    pGroup->leftEnd = pBelowLeft->runLeft;
		    pGroup->rightEnd = pBelowRight->runRight;
		    break;

		case 2:		/* only runs in the middle line.  Allow
				 * very little slop */
		    pGroup->leftEndMin = leftPt;
		    pGroup->leftEndMax = pMidLeft->runLeft + bigNoise;
		    pGroup->rightEndMin = pMidLeft->runRight - bigNoise;
		    pGroup->rightEndMax = rightPt;
		    pGroup->groupType = cCheck1Run;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;

		    break;

		case 3:		/* runs on the bottom and middle lines only */
				/* if the middle line sticks out farther than
				 * the bottom line on either end, require
				 * it to be here, otherwise, pretend it's just
				 * noise.
				 */
		    pGroup->leftEndMin = leftPt;
		    pGroup->rightEndMax = rightPt;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;
		    if ((pBelowLeft->runRight < pMidLeft->runRight) ||
			(pBelowLeft->runLeft > pMidLeft->runLeft))
		    {
			pGroup->groupType = cCheckEndpoints;
			if (pBelowLeft->runRight < pMidLeft->runRight)
			    pGroup->rightEndMin = pMidLeft->runRight -
								smallNoise;
			else
			    pGroup->rightEndMin = leftPt;

			if (pBelowLeft->runLeft > pMidLeft->runLeft)
			    pGroup->leftEndMax = pMidLeft->runLeft + smallNoise;
			else
			    pGroup->leftEndMax = rightPt;
		    }
		    else
		    {
			pGroup->discardBdry = leftPt;
			pGroup->groupType = cCheckNoise;
		    }

		    break;

		case 4:		/* runs on the top line only, allow lots of
				 * slop on this line */
		    pGroup->leftEndMin = leftPt - endNoise;
		    pGroup->rightEndMax = rightPt + endNoise;
		    pGroup->discardBdry = leftPt - endNoise;
		    pGroup->groupType = cCheckNoise;
		    pGroup->leftEnd = pAboveLeft->runLeft;
		    pGroup->rightEnd = pAboveRight->runRight;
		    break;

		case 5:		/* no middle line.  There should be no
				 * instance run here.  If there is one,
				 * the next match will fail.
				 */
		    break;

		case 6:		/* runs on the top and middle lines only.
				 * As before, be strict if the middle line
				 * sticks out more than the top line.
				 */
		    pGroup->leftEndMin = leftPt;
		    pGroup->rightEndMax = rightPt;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;
		    if ((pAboveLeft->runRight < pMidLeft->runRight) ||
			(pAboveLeft->runLeft > pMidLeft->runLeft))
		    {
			pGroup->groupType = cCheckEndpoints;
			if (pAboveLeft->runRight < pMidLeft->runRight)
			    pGroup->rightEndMin = pMidLeft->runRight -
								smallNoise;
			else
			    pGroup->rightEndMin = leftPt;

			if (pAboveLeft->runLeft > pMidLeft->runLeft)
			    pGroup->leftEndMax = pMidLeft->runLeft + smallNoise;
			else
			    pGroup->leftEndMax = rightPt;
		    }
		    else
		    {
			pGroup->discardBdry = leftPt;
			pGroup->groupType = cCheckNoise;
		    }

		    break;

		case 7:		/* runs on all three lines, the normal case. */
		    if ((pMidLeft->runLeft < pAboveLeft->runLeft) &&
			(pMidLeft->runLeft < pBelowLeft->runLeft))
			leftTight = pMidLeft->runLeft + bigNoise;
		    else
			leftTight = MAX(pAboveLeft->runLeft + smallNoise,
					MAX(pMidLeft->runLeft + bigNoise,
					    pBelowLeft->runLeft + smallNoise));

		    if ((pMidLeft->runRight > pAboveLeft->runRight) &&
			(pMidLeft->runRight > pBelowLeft->runRight))
			rightTight = pMidLeft->runRight - bigNoise;
		    else
			rightTight = MIN(pAboveLeft->runRight - smallNoise,
					MIN(pMidLeft->runRight - bigNoise,
					    pBelowLeft->runRight - smallNoise));

			/* assign the main group */
		    pGroup->leftEndMin = leftPt;
		    pGroup->leftEndMax = leftTight;
		    pGroup->rightEndMin = rightTight;
		    pGroup->rightEndMax = rightPt;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;
			/* if either of the ranges is big, use cCheckEndpoints
			 * instead of cCheck1Run.  cCheckEndpoints allows
			 * for the presence of noise blobs in the ranges.
			 */
		    if ((leftTight - leftPt > cMinNoiseEdgeLength) ||
			(rightPt - rightTight > cMinNoiseEdgeLength))
			pGroup->groupType = cCheckEndpoints;
		    else
			pGroup->groupType = cCheck1Run;

		    break;

		default:	/* should never get here */
		    break;
		}
	    }
	    else	/* make three groups, odd line tags */
	    {
		    /* first, figure out what values to use for the
		     * endpoint ranges.
		     */
		switch (flags)
		{
		case 3:		/* runs on the bottom and middle lines only */

			/* Since the code for cases 3 and 6 is virtually
			 * identical, just set pAboveRight to pBelowRight
			 * and pAboveLeft to pBelowLeft and drop through.
			 * These pointers aren't used after the groups are
			 * generated anyway.
			 */
		    pAboveRight = pBelowRight;
		    pAboveLeft = pBelowLeft;

		case 6:		/* runs on the top and middle lines only */
			/* For each run in the top line, put out a
			 * group saying that there's only runs on the top
			 * line and giving endpoint ranges that cover the
			 * entire group.  For each gap between top runs,
			 * put out a group saying there are runs
			 * in the Bottom and Middle line with endpoint ranges
			 * starting just inside the gap and going out to the
			 * ends.
			 */
		    for (;
			 pAboveRight != pAboveLeft;
			 pAboveRight = pAboveRight->pNext)
		    {
			    /* make a group to catch noise below pAboveRight */
			__error__ =
			    i_allocateStruct (pIPState,
					      pTextCompInfo->p1stGroupBlock,
					      (void **) &(pGroup->pNext));
	    
			if (__error__ != ia_successful)
			    goto tcMakeGroupsCleanup;

			pGroup = pGroup->pNext;
			pGroup->leftEndMin = leftPt;
			pGroup->rightEndMax = rightPt;
			pGroup->discardBdry = pAboveRight->runLeft;
			pGroup->groupType = cCheckNoise;
			pGroup->leftEnd = pMidLeft->runLeft;
			pGroup->rightEnd = pMidRight->runRight;

			    /* make a group to bridge to the next run below */
			__error__ =
			    i_allocateStruct (pIPState,
					      pTextCompInfo->p1stGroupBlock,
					      (void **) &(pGroup->pNext));
	    
			if (__error__ != ia_successful)
			    goto tcMakeGroupsCleanup;
	    
			pGroup = pGroup->pNext;
			pGroup->leftEndMin = leftPt;
			pGroup->leftEndMax = pAboveRight->pNext->runRight
								+ bigNoise ;
			pGroup->rightEndMin = pAboveRight->runLeft
								- bigNoise;
			pGroup->rightEndMax = rightPt;
			pGroup->groupType = cCheck1RunNoAdvance;
			pGroup->leftEnd = pMidLeft->runLeft;
			pGroup->rightEnd = pMidRight->runRight;
		    }
	
	
			/* make the group to detect noise below pAboveLeft */
		    __error__ =
		        i_allocateStruct (pIPState,
					  pTextCompInfo->p1stGroupBlock,
					  (void **) &(pGroup->pNext));
	    
		    if (__error__ != ia_successful)
			goto tcMakeGroupsCleanup;

		    pGroup = pGroup->pNext;
		    pGroup->leftEndMin = leftPt;
		    pGroup->rightEndMax = rightPt;
		    pGroup->discardBdry = leftPt;
		    pGroup->groupType = cCheckNoise;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;

		    break;

		case 7:		/* runs on all three lines, the normal case. */
		    if ((pMidLeft->runLeft < pAboveLeft->runLeft) &&
			(pMidLeft->runLeft < pBelowLeft->runLeft))
			leftTight = pMidLeft->runLeft + bigNoise;
		    else
			leftTight = MAX(pAboveLeft->runLeft + smallNoise,
					MAX(pMidLeft->runLeft + bigNoise,
					    pBelowLeft->runLeft + smallNoise));

		    if ((pMidRight->runRight > pAboveRight->runRight) &&
			(pMidRight->runRight > pBelowRight->runRight))
			rightTight = pMidRight->runRight - bigNoise;
		    else
			rightTight = MIN(pAboveRight->runRight - smallNoise,
				      MIN(pMidRight->runRight - bigNoise,
					  pBelowRight->runRight - smallNoise));


			/* make the group */
		    __error__ =
			i_allocateStruct (pIPState,
					  pTextCompInfo->p1stGroupBlock,
					  (void **) &(pGroup->pNext));
    
		    if (__error__ != ia_successful)
			goto tcMakeGroupsCleanup;
	    
		    pGroup = pGroup->pNext;
		    pGroup->leftEndMin = leftPt;
		    pGroup->leftEndMax = leftTight;
		    pGroup->rightEndMin = rightTight;
		    pGroup->rightEndMax = rightPt;
		    pGroup->groupType = cCheckEndpoints;
		    pGroup->leftEnd = pMidLeft->runLeft;
		    pGroup->rightEnd = pMidRight->runRight;

		    break;

		default:	/* Case 0, 1, 2 and 4 cannot happen.  Case 5
				 * causes the matching process to ignore
				 * all the endpoint settings
				 * except leftEndMin anyway.
				 */
		    break;
		}

	    }	/* end of case that makes three groups, odd line tags */
	}	/* end of loop that makes groups for lines ending in a
		 * flag of cEvenLineTag
		 */

	    /* if pMid and pAbove are now equal, we're done since we've
	     * since we've processed the very top line of the image in
	     * all three positions.
	     */
	if ((pMid == pAbove) && (pMid == pBelow))
	    break;

    }	/* end of loop that makes groups for all lines. */

	/* If we get here, everything was ducky.
	 * End the chain of groups.
	 */
   pGroup->pNext = NULL;

	/* Turn on the flags that indicate which groups are at the
	 * right and left ends of scan lines.  These are used when
	 * we track the variance of the endpoints of an instance
	 * relative to the endpoints of an exemplar.
	 * Since these flags control whether the endspoints are
	 * checked, we only turn them on for symbols on which we
	 * want to check the endpoints.  In particular, small, round,
	 * dense symbols.  We use smallNoise as the minimum of
	 * the symbol's width and height; bigNoise equals the maximum
	 * of the symbol's width and height and endNoise equals the
	 * symbol's area.  We only want to do the endpoint check
	 * if bigNoise * (3/4) <= smallNoise AND area * (3/4) <= mass
	 * AND (bigNoise < cLittleSymbolSize.
	 */
    if (pInstance->w < pInstance->h)
    {
	smallNoise = (Int16) pInstance->w;
	bigNoise = (Int16) pInstance->h;
    }
    else
    {
	smallNoise = (Int16) pInstance->h;
	bigNoise = (Int16) pInstance->w;
    }
    endNoise = smallNoise * bigNoise * 2;	/* calculate the sym's area
						 * in 2x coords */
    if ((bigNoise < cLittleSymbolSize) &&
	(((bigNoise * cMinVolFraction - bigNoise)/cMinVolFraction)
							 <= smallNoise) &&
	(((endNoise * cMinVolFraction - endNoise)/cMinVolFraction)
							 <= pInstance->mass))
    {
	pGroup = *pp1stGroup;
	pGroup->groupType |= cAtRightEnd;		/* first run is at right
						     * end of scan line */
	for (; pGroup->pNext != NULL; pGroup = pGroup->pNext)
	{
	    if ((pGroup->pNext->groupType ^ pGroup->groupType) & cOddLineTag)
	    {
		    /* We're at the boundary between two scan lines, so
		     * mark both groups.
		     */
		pGroup->groupType |= cAtLeftEnd;
		pGroup->pNext->groupType |= cAtRightEnd;
	    }
	}
	    /* mark the last run as being at the left end of the last line */
	pGroup->groupType |= cAtLeftEnd;
    }

	/* Duplicate the groups contained
	 * in the first scan line.  These are used to shift the exemplar
	 * up or left by one row or column.  We actually use the 
	 * extra first group at the head of the list as the first
	 * duplicate group.
	 * Note that since even and odd line groups are distinguished
	 * by the presence or absence of cOddLineTag, (instead of the
	 * presence of cOddLineTag or the presence of cEvenLineTag),
	 * we XOR with cOddLineTag to switch the even/odd tag instead
	 * of XORing with cLineTagMask.
	 */
 
    pDupGroup = *pp1stGroup;
    pGroup = pDupGroup->pNext;
    *pp1stGroup = pGroup;
    *ppDupGroup = pDupGroup;
    pDupGroup->leftEndMin = pGroup->leftEndMin;
    pDupGroup->leftEndMax = pGroup->leftEndMax;
    pDupGroup->rightEndMin = pGroup->rightEndMin;
    pDupGroup->rightEndMax = pGroup->rightEndMax;
    pDupGroup->leftEnd = pGroup->leftEnd;
    pDupGroup->rightEnd = pGroup->rightEnd;
    pDupGroup->discardBdry = pGroup->discardBdry;
    pDupGroup->groupType = pGroup->groupType ^ cOddLineTag;
	/* Make sure the chain is always complete so we can free
	 * everything in one step if one of the later allocations
	 * fails.
	 */
    pDupGroup->pNext = *pp1stGroup;

    for (pGroup = pGroup->pNext;
	 (pGroup->groupType & cOddLineTag) == 0;
	 pGroup = pGroup->pNext)
    {
	__error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stGroupBlock,
				    (void **) &(pDupGroup->pNext));
	if (__error__ != ia_successful)
	    goto tcMakeGroupsCleanup;

	pDupGroup = pDupGroup->pNext;
	pDupGroup->leftEndMin = pGroup->leftEndMin;
	pDupGroup->leftEndMax = pGroup->leftEndMax;
	pDupGroup->rightEndMin = pGroup->rightEndMin;
	pDupGroup->rightEndMax = pGroup->rightEndMax;
	pDupGroup->leftEnd = pGroup->leftEnd;
	pDupGroup->rightEnd = pGroup->rightEnd;
	pDupGroup->discardBdry = pGroup->discardBdry;
	pDupGroup->groupType = pGroup->groupType ^ cOddLineTag;
	pDupGroup->pNext = *pp1stGroup;
    }

	/* Make the last duplicate group point to the first real group */
    pDupGroup->pNext = *pp1stGroup;

	/* That all worked.  Set __error__ equal to ia_successful and
	 * drop through to the cleanup code.
	 */
    __error__ = ia_successful;

tcMakeGroupsCleanup:
	/* If we failed, free all the groups starting from
	 * *ppDupGroup.
	 */
    if (__error__ != ia_successful)
    {
	for (pDupGroup = *ppDupGroup;
	     pDupGroup != NULL;
	     )
	{
	    pGroup = pDupGroup->pNext;
	    i_freeStruct (pIPState, pTextCompInfo->p1stGroupBlock,
			  (void **)&pDupGroup);
	    pDupGroup = pGroup;
	}
    }

	/* toss the dummy run on the end of the list.  This
	 * also sets the pointer to the dummy to NULL
	 */
    if (pEndRun->pNext != NULL)
	i_freeStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
		      (void **)&(pEndRun->pNext));

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_symbolIsComplete()
 *
 *	This procedure is called for each completed symbol.  It determines
 *	the final identity of the symbol and outputs position information.
 *
 *	Input:
 *		pTextCompInfo:		Ptr to all text compression info
 *		pSymbol:		Ptr to symbol matching the new
 *					instance.  This may be NULL if
 *					no match was found.
 *		pInstance:		Ptr to information about instance
 *		insertOnly:		When cTrue, just insert this symbol
 *					into the list of symbols, don't put
 *					out any position info.  It's also ok
 *					for the slant vertical runs to be
 *					NULL if this is true.  ALSO, when
 *					this is set, we just attach the incoming
 *					set of horizontal runs to the symbol
 *					instead of copying them to a new set.
 *
 *	Action:
 *		If no match was found, make up a new symbol and fill in
 *		its entries, insert it into the chain of symbols and
 *		copy the horizontal runs to new, permanent structures.  In
 *		any case, put position information into the output.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		ran out of memory
 *		ia_bufferExceeded	ran off the end of the position buffer.
 *
 **************************************************************************/
Int32 CDECL 
ip_symbolIsComplete (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCSymbol   *pSymbol,
    TCInstance *pInstance,
    Bool	insertOnly)
{
    PROLOGUE;
    TCScanLineRun *pNewScanLineRun;	/* used to allocate new scan line
				         * runs when we make a new symbol.
				         */
    TCScanLineRun *pOldScanLineRun;	/* used to copy old scan line runs */
    TCSymbol   *pTempSymbol;		/* scratch ptr */
    Int32	delta;			/* amount by which various measures
				         * of symbol are allowed to vary */
    

    __error__ = ia_successful;		/* hope it stays this way */

	/* if we're only supposed to classify stuff, go no further.
	 * If a new symbol was found, just return a pointer to it.
	 * If no new symbol was found, return NULL.
	 */
    if (pTextCompInfo->classifyOnly)
    {
	if (pSymbol != NULL)
	    pTextCompInfo->pClassifiedSymbol = pSymbol;
	goto symIsCompCleanup;
    }

	/* We are still building the symbol list.  Did we find a symbol?
	 * If not, fill in the rest of they symbol's entries and
	 * insert it into the shape tree.  This
	 * involves allocating new scan line runs and copying all their
	 * data.
	 */
    pTextCompInfo->totalSymbols++;	/* for debugging */

    if (pSymbol == NULL)
    {
	__error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stSymbolBlock,
			            (void **) &pSymbol);
	if (__error__ != ia_successful)
	    goto symIsCompCleanup;

	pSymbol->width = pInstance->w;
	pSymbol->height = pInstance->h;
	pSymbol->minX = -(pInstance->w - 1);
	pSymbol->maxX = (pInstance->w - 1);
	pSymbol->minY = -(pInstance->h - 1);
	pSymbol->maxY = (pInstance->h - 1);
	delta = MAX(pInstance->w, pInstance->h);
	delta = MIN((delta - cDimensionOffset)/cDimensionDivisor,
		    cMaxDimensionSlop);
	delta = MAX(delta, 0);
	pSymbol->minWidth = pInstance->w - delta;
	pSymbol->maxWidth = pInstance->w + delta;
	pSymbol->minHeight = pInstance->h - delta;
	pSymbol->maxHeight = pInstance->h + delta;
	pSymbol->mass = pInstance->mass;

	    /* Set the symbol's funny measures from those recorded for
	     * this instance.
	     */
	pSymbol->leftProfile = pInstance->leftProfile;
	pSymbol->rightProfile = pInstance->rightProfile;
	pSymbol->strokeWidth = pInstance->strokeWidth;
	pSymbol->maxDimension = pInstance->maxDimension;
	pSymbol->slant = pInstance->slant;
	pSymbol->checkHVolume = pInstance->checkHVolume;
	pSymbol->checkVVolume = pInstance->checkVVolume;
	pSymbol->hVolume = pInstance->hVolume;
	pSymbol->vVolume = pInstance->vVolume;
	pSymbol->hSAsymmetry = pInstance->hSAsymmetry;
	pSymbol->vSAsymmetry = pInstance->vSAsymmetry;
	pSymbol->numActiveRuns = 0;
	pSymbol->instanceCount = 1;
	pSymbol->pRunHead = NULL;
	pSymbol->pClassSymbol = NULL;
	pSymbol->pNewSymbol = NULL;
	pSymbol->pNext = NULL;		/* temporarily */
	pSymbol->pPrev = NULL;		/* temporarily */
	pSymbol->bytesPerLine = 0;
	pSymbol->pBitmap = NULL;
	pSymbol->xOffset = 0;
	pSymbol->pNextDictSym = NULL;
	pSymbol->dictOrder = -1;
	pSymbol->dictNum = pTextCompInfo->currentDictionaryNum;
	pSymbol->pageID = pTextCompInfo->curPageID;
	pSymbol->huffmanCodeLen = 0;
	pSymbol->huffmanCode = 0;

	    /* Make classification info and stuff only we're really
	     * going to do classification.  If we're just inserting this
	     * symbol into a list of symbols, as happens when the
	     * symbols are being deserialized, don't bother.
	     */
	if (!insertOnly)
	{
		/* make the horizontal groups */
	    __error__ = ip_makeGroupsFromRuns (pIPState, pTextCompInfo,
					     pInstance,
					     &(pSymbol->p1stHorizGroup),
					     &(pSymbol->pDupHorizGroup),
					     cTrue);
	    if (__error__ != ia_successful)
		goto symIsCompCleanup;
    
		/* make the vertical groups */
	    __error__ = ip_makeGroupsFromRuns (pIPState, pTextCompInfo,
					     pInstance,
					     &(pSymbol->p1stVertGroup),
					     &(pSymbol->pDupVertGroup),
					     cFalse);
	    if (__error__ != ia_successful)
		goto symIsCompCleanup;
    
		/* copy the incoming set of horizontal runs to a new set
		 * especially for this symbol.  The incoming set will
		 * be deallocated when we return to
		 * i_findConnectedComponentRuns.
		 */
	    __error__ = i_allocateStruct (pIPState,
					pTextCompInfo->p1stScanLineBlock,
				        (void **) &pNewScanLineRun);
	    if (__error__ != ia_successful)
		goto symIsCompCleanup;

	    pSymbol->pRunTail = pNewScanLineRun;
	    pOldScanLineRun = pInstance->pHorizRuns;
	    pNewScanLineRun->runLeft = pOldScanLineRun->runLeft;
	    pNewScanLineRun->runRight = pOldScanLineRun->runRight;
		/* keep line tags for encoding */
	    pNewScanLineRun->pClient = pOldScanLineRun->pClient;

	    for (pOldScanLineRun = pOldScanLineRun->pNext;
		 pOldScanLineRun != NULL;
		 pOldScanLineRun = pOldScanLineRun->pNext)
	    {
		__error__ = i_allocateStruct (pIPState,
					    pTextCompInfo->p1stScanLineBlock,
					    (void **)
						&(pNewScanLineRun->pNext));
		if (__error__ != ia_successful)
		    goto symIsCompCleanup;
    
		pNewScanLineRun = pNewScanLineRun->pNext;
		pNewScanLineRun->runLeft = pOldScanLineRun->runLeft;
		pNewScanLineRun->runRight = pOldScanLineRun->runRight;
		pNewScanLineRun->pClient = pOldScanLineRun->pClient;
	    }
	    pNewScanLineRun->pNext = NULL;
	}
	else	/* just use the runs that were supplied as the horizontal
		 * runs and don't worry about the rest */
	{
	    pSymbol->p1stHorizGroup = NULL;
	    pSymbol->p1stVertGroup = NULL;
	    pSymbol->pRunTail = pInstance->pHorizRuns;
	}
	pSymbol->runsPtToTail = cFalse;	/* the runs' pNext ptrs point to
					 * pRunHead. */

	    /* insert this symbol into the array of symbols.  For want of
	     * a more beautiful policy, just put it at the head of
	     * the list of symbols with the same width and height.
	     * If either the width or height is too big, put it
	     * in the special list of big symbols.
	     */
	pSymbol->symbolID = pTextCompInfo->nextSymbolNum;
	pTextCompInfo->nextSymbolNum++;
	pSymbol->pNextNum = NULL;
	pSymbol->pPrevNum = pTextCompInfo->pLastSymbol;
	if (pSymbol->pPrevNum != NULL)	/* This is NOT the first symbol */
	    pSymbol->pPrevNum->pNextNum = pSymbol;
	else	/* this IS the first symbol */
	    pTextCompInfo->p1stSymbol = pSymbol;
	pTextCompInfo->pLastSymbol = pSymbol;

	pSymbol->pPrev = NULL;
	if ((pInstance->w > cMaxSymIndexDim) ||
	    (pInstance->h > cMaxSymIndexDim))
	{
		/* It will be convenient later to have this list
		 * sorted by height, then width.  This makes
		 * it easier to put symbols out in a Tundra
		 * dictionary, in height/width order.
		 */
	    pTempSymbol = pTextCompInfo->pBigSymbols;
	    if ((pTempSymbol == NULL) ||
		(pSymbol->maxY < pTempSymbol->maxY) ||
		((pSymbol->maxY == pTempSymbol->maxY) &&
		 (pSymbol->maxX <= pTempSymbol->maxX)))
	    {
		    /* the list of big symbols is empty or we
		     * should insert the new symbol before the
		     * first existing symbol
		     */
		pSymbol->pNext = pTextCompInfo->pBigSymbols;
		pSymbol->pPrev = NULL;
		if (pSymbol->pNext != NULL)
		    pSymbol->pNext->pPrev = pSymbol;
		pTextCompInfo->pBigSymbols = pSymbol;
	    }
	    else
	    {
		    /* Insert the new symbol somewhere after the
		     * first symbol.  First find the right height,
		     * then the right place in symbols of the same
		     * height.
		     */
		while ((pTempSymbol->pNext != NULL) &&
			(pSymbol->maxY > pTempSymbol->pNext->maxY))
		    pTempSymbol = pTempSymbol->pNext;
		while ((pTempSymbol->pNext != NULL) &&
			(pSymbol->maxY == pTempSymbol->pNext->maxY) &&
			(pSymbol->maxX > pTempSymbol->pNext->maxX))
		    pTempSymbol = pTempSymbol->pNext;

		    /* Insert pSymbol before pTempSymbol->pNext */
		pSymbol->pNext = pTempSymbol->pNext;
		if (pSymbol->pNext != NULL)
		    pSymbol->pNext->pPrev = pSymbol;
		pSymbol->pPrev = pTempSymbol;
		pTempSymbol->pNext = pSymbol;
	    }
	}
	else
	{
	    pSymbol->pNext =
		(pTextCompInfo->pSymIndex->array)[pInstance->w][pInstance->h];
	    (pTextCompInfo->pSymIndex->array)[pInstance->w][pInstance->h] =
		pSymbol;
	}
	if (pSymbol->pNext != NULL)
	    pSymbol->pNext->pPrev = pSymbol;

    }
    else	/* we found a match and pSymbol points to it */
	pSymbol->instanceCount++;		/* count the instances */

	/* if we're only supposed to insert the symbol into the data
	 * structures, as happens when we're decoding shape info from
	 * a serial bytestream, just stop now.
	 */
    if (insertOnly)
    {
	__error__ = ia_successful;
	goto symIsCompCleanup;
    }

	/* At this point, pSymbol points to the symbol we want to encode
	 * into the output position file.  The Y position of
	 * the symbol might be the same as the Y position of the
	 * instance, one line above it or one line below it.
	 * This makes it difficult to just put out the position
	 * info since we'd potentially have to insert it into the
	 * middle of the line above or the line below (that doesn't
	 * exist yet).  Rather than doing this, we will put
	 * the symbol into the list of uncommitted symbols.
	 */
    __error__ = ip_createUncommittedTCPosition (pIPState,
					      pTextCompInfo,
					      pInstance, pSymbol);

	/* if we arrive here with __error__ still equal to ia_successful, great.
	 * Even if something went wrong, all we do is return __error__.
	 */
symIsCompCleanup:
    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_createUncommittedTCPosition()
 *
 *	When substituting one symbol for another, we need to adjust the
 *	the position of the replacement symbol to properly align it with
 *	the original.  This routine allows us to create a list of
 *	replacement symbols that's indexed by x position.  Each replacement
 *	symbol's list element holds the replacement symbol's y coord.
 *	When all the symbols for a given line have finally been determined,
 *	that line is flushed using ip_commitTCPositions.
 *
 *	Input:
 *		pTextCompInfo:	Ptr to all text compression info
 *		pInstance:	Ptr to original symbol being replaced.
 *		pSymbol:	Ptr to replacement symbol.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_bufferExceeded	Couldn't get enough space for
 *					the committed position bytes.
 *		ia_nomem		Ran out of memory
 *
 **************************************************************************/
Int32  CDECL 
ip_createUncommittedTCPosition (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCInstance *pInstance,
    TCSymbol   *pSymbol)
{
    PROLOGUE;
    Int32		 x;		/* x position of symbol */
    Int32		 currentLine;	/* the line on which the output should
					 * be written. */
    UInt32		 profileDiff;	/* XOR of two profiles */
    Int32		 levelLeft;	/* number of profile error bits
					 * found when left profile is
					 * unshifted. */
    Int32		 downLeft;	/* number of profile error bits
					 * found when replacment left profile
					 * is shifted down 1. */
    Int32		 upLeft;	/* number of profile error bits
					 * found when replacment left profile
					 * is shifted up 1. */
    Int32		 levelRight;	/* number of profile error bits
					 * found when right profile is
					 * unshifted. */
    Int32		 downRight;	/* number of profile error bits
					 * found when replacment right profile
					 * is shifted down 1. */
    Int32		 upRight;	/* number of profile error bits
					 * found when replacment right profile
					 * is shifted up 1. */
    TCScanLineRun	*pOldRun, *pNewRun;
    

	/* At this point, pSymbol points to the symbol we want to encode
	 * into the output position file.  We encode the position of the
	 * LOWER, RIGHT corner of the symbol.  The Y position of
	 * the symbol might be the same as the Y position of the
	 * instance, one line above it or one line below it.
	 * This makes it difficult to just put out the position
	 * info since we'd potentially have to insert it into the
	 * middle of the line above or the line below (that doesn't
	 * exist yet).  Rather than doing this, we will put
	 * the symbol into the list of uncommitted symbols.
	 *
	 * But when do we write out a row of positions from the
	 * uncommitted symbol list?  Well, when the current row
	 * can't affect it anymore.  We know that the current row
	 * is the bottom row of the instance.  We put out all
	 * rows of uncommitted symbols that are more than 1 row
	 * above the bottom of the instance.
	 * SMALL IMPROVEMENT:  Try positioning the *right* side of the
	 * bounding box at x + w/2 (gets to middle of bounding box) +
	 * (exemplar width/2).  This should place the right side of
	 * the exemplar bounding box so the middle of the exemplar
	 * is lined up with the middle of the instance.
	 * Since pSymbol->width is a 1x coord,
	 * we just add the (instance Width + symbol Width) + 1
	 * <to make the rounding work out right>) / 2.
	 * We save the divide to the end to preserve significance as
	 * long as possible.
	 * Note that x cannot be less than 0.  We don't care if
	 * is off the right side of the image, the blits done by
	 * the decoder will be clipped to the image.
	 */
    x = pInstance->x + (pInstance->w + pSymbol->width + 1)/2 - 1;
    currentLine = pInstance->y + pInstance->h - 1;

	/* Commit any rows of symbols that won't be affected by symbols
	 * on this line */
    while (pTextCompInfo->lowUncomY < currentLine - 1)
    {
	    /* since we're on a new line, just reset the last
	     * uncommited position ptr.  Otherwise, it might
	     * be left pointing at some deleted position.
	     */
	pTextCompInfo->pLastUncom = &(pTextCompInfo->dum1stLoc);

	    /* Go convert uncommited positions to bytes in the
	     * position buffer.
	     */
	__error__ = ip_commitTCPositions (pIPState, pTextCompInfo,
				        pTextCompInfo->lowUncomY);
	if (__error__ != ia_successful)
	    goto finCreateUncomTCPos;

	    /* It worked, so increment pTextCompInfo->lowUncomY.  If
	     * it's greater than pTextCompInfo->hiUncomY, stop since
	     * the list is empty.
	     */
	pTextCompInfo->lowUncomY++;
	if (pTextCompInfo->lowUncomY > pTextCompInfo->hiUncomY)
	    break;
    }

	/* Set the list of rows that might contain uncommitted
	 * symbols.
	 */
    pTextCompInfo->lowUncomY = currentLine - 1;
    pTextCompInfo->hiUncomY = currentLine + 1;

	/* Now insert this symbol in the list of uncommitted symbols.
	 * First, figure out where it goes.  pTextCompInfo->pLastUncom
	 * points to the last uncommited symbol inserted.  If we
	 * need to move backwards, go to the beginning of the list.
	 * This is usually the right thing to do since we usually
	 * move back at the beginning of a scan line.
	 */
    pOldRun = pTextCompInfo->pLastUncom;
    if (x < pOldRun->runLeft)
	pOldRun = &(pTextCompInfo->dum1stLoc);

    while (x > pOldRun->pNext->runLeft)
	pOldRun = pOldRun->pNext;

	/* pOldRun is now pointing to the uncommitted indicator
	 * just before the one we want to insert.
	 */
    __error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stScanLineBlock,
				(void **) &pNewRun);
    if (__error__ != ia_successful)
	goto finCreateUncomTCPos;

    pNewRun->pClient = (void *) pSymbol->symbolID;
    pNewRun->runLeft = (Int16) x;

	/* This is the tricky part.  We want to set the new Y coord
	 * so that the symbol and instance are aligned vertially.
	 * Originally, I tried to line up horizontal centerlines, but
	 * that broke about as many placements as it fixed.  Then I
	 * tried counting the number of pixels on the bottom line
	 * the instance and symbol - if the difference was big enough,
	 * I decided that one had noise on the bottom.  That worked
	 * better (surprisingly enough), but not well enough.
	 * Then I tried various comparisons between the lengths of
	 * the spans on the bottom two lines of the symbols.  That
	 * was sort of ok, but still suffered from the fact that
	 * you just can't tell enough by only looking at the bottom
	 * lines.
	 *
	 * In the current scheme, we use a cheap way of looking at
	 * the bottom 32 rows of the symbol.  We have two profile
	 * masks, left and right.  Each has one bit per row, MS bit
	 * corresponds to row 0.  A bit is set in a profile if the
	 * bits in the symbol are close to the edge of the bounding
	 * box on that row.  For example, the profiles for an "F"
	 * might have lots of bits on on the left while the right
	 * profile would have bits only at the top and in the middle.
	 *
	 * We match the profiles of the original and replacement symbols
	 * togeher without shifting first.  Then, we try shifting them
	 * by 1 up or down relative to each other and see how well
	 * those match.  Generally, it's a good idea to put the
	 * bottom line of the replacement on the bottom line of the
	 * original.  However, if both the left and right profiles
	 * line up better when shifted up (down), we move the
	 * replacement symbol up (down).
	 *
	 * To be symmetric, we either shift the original (pInstance)
	 * up or shift the replacment (pSymbol) up.  This means we
	 * never lose the bottom line of the symbol.  Also, since
	 * we are shifting integers and doing right shifts and the
	 * bottom row bit is in the MS end of the word, the bit for
	 * the bottom line is duplicated in our shifts.  This makes
	 * it easier to match a symbol with a fat bottom to one
	 * with a thin bottom.
	 */
    pNewRun->runRight = (Int16) currentLine;

    profileDiff = pInstance->leftProfile ^ pSymbol->leftProfile;
    levelLeft = bitCount[profileDiff & 0xFF] +
		bitCount[(profileDiff >> 8) & 0xFF] +
		bitCount[(profileDiff >> 16) & 0xFF] +
		bitCount[(profileDiff >> 24)];
    profileDiff = pInstance->rightProfile ^ pSymbol->rightProfile;
    levelRight = bitCount[profileDiff & 0xFF] +
		 bitCount[(profileDiff >> 8) & 0xFF] +
		 bitCount[(profileDiff >> 16) & 0xFF] +
		 bitCount[(profileDiff >> 24)];

    profileDiff = ((UInt32)(pInstance->leftProfile) >> 1) ^
						pSymbol->leftProfile;
    downLeft =  bitCount[profileDiff & 0xFF] +
		bitCount[(profileDiff >> 8) & 0xFF] +
		bitCount[(profileDiff >> 16) & 0xFF] +
		bitCount[(profileDiff >> 24)];
    profileDiff = ((UInt32)(pInstance->rightProfile) >> 1) ^
						pSymbol->rightProfile;
    downRight =  bitCount[profileDiff & 0xFF] +
		 bitCount[(profileDiff >> 8) & 0xFF] +
		 bitCount[(profileDiff >> 16) & 0xFF] +
		 bitCount[(profileDiff >> 24)];

    profileDiff = pInstance->leftProfile ^
				((UInt32)(pSymbol->leftProfile) >> 1);
    upLeft =    bitCount[profileDiff & 0xFF] +
		bitCount[(profileDiff >> 8) & 0xFF] +
		bitCount[(profileDiff >> 16) & 0xFF] +
		bitCount[(profileDiff >> 24)];
    profileDiff = pInstance->rightProfile ^
				((UInt32)(pSymbol->rightProfile) >> 1);
    upRight =    bitCount[profileDiff & 0xFF] +
		 bitCount[(profileDiff >> 8) & 0xFF] +
		 bitCount[(profileDiff >> 16) & 0xFF] +
		 bitCount[(profileDiff >> 24)];

	/* Symbols are the same height */
    if (pInstance->h == pSymbol->height)
    {
	    /* Now, if the left and right profiles say to move
	     * up ((upLeft < levelLeft) and (upRight < levelRight))
	     * and the left and right profiles do not say to move
	     * down ((downLeft >= levelLeft) and (downRight >= levelRight)),
	     * move up.  Otherwise, if the corresponding test for
	     * moving down is true, move down.
	     * Note that we aren't allowed to move above the top
	     * or below the bottom of the image.
	     */
	if ( ( ((upLeft < levelLeft) && (upRight <= levelRight)) ||
	       ((upLeft <= levelLeft) && (upRight < levelRight))
	     ) &&
	     ( (downLeft >= levelLeft) && (downRight >= levelRight)
	     ) &&
	     (currentLine > 0)
	   )
	    pNewRun->runRight--;	/* move up */
	else
	if ( ( ((downLeft < levelLeft) && (downRight <= levelRight)) ||
	       ((downLeft <= levelLeft) && (downRight < levelRight))
	     ) &&
	     ( (upLeft >= levelLeft) && (upRight >= levelRight)
	     ) &&
	     (currentLine + 1 < pTextCompInfo->imageHeight)
	   )
	    pNewRun->runRight++;	/* move down */
    }
    else
    if (pInstance->h > pSymbol->height)
    {
	    /* Here, we know that the replacement is shorter, so we
	     * only have to worry about the possibility of moving it
	     * up.
	     */
	if ( ( ((upLeft < levelLeft) && (upRight <= levelRight)) ||
	       ((upLeft <= levelLeft) && (upRight < levelRight))
	     ) &&
	     ( (downLeft >= levelLeft) && (downRight >= levelRight)
	     ) &&
	     (currentLine > 0)
	   )
	    pNewRun->runRight--;	/* move up */
    }
    else	/* replacement is taller */
    {
	    /* Here, we know that the replacement is taller, so we
	     * only have to worry about the possibility of moving it
	     * down.
	     */
	if ( ( ((downLeft < levelLeft) && (downRight <= levelRight)) ||
	       ((downLeft <= levelLeft) && (downRight < levelRight))
	     ) &&
	     ( (upLeft >= levelLeft) && (upRight >= levelRight)
	     ) &&
	     (currentLine + 1 < pTextCompInfo->imageHeight)
	   )
	    pNewRun->runRight++;	/* move down */
    }


	/* Insert the new uncommitted position into the list and
	 * remember where it is and increment the count of things
	 * in the list.
	 */
    pNewRun->pNext = pOldRun->pNext;
    pOldRun->pNext = pNewRun;
    pTextCompInfo->pLastUncom = pNewRun;
    pTextCompInfo->numUncomPos++;

    __error__ = ia_successful;

finCreateUncomTCPos:
    NORMAL_RETURN;
}



/* PRIVATE */
/**************************************************************************
 *   ip_commitTCPositions()
 *
 *	This procedure looks through the list of uncommitted positions
 *	for positions that are in the specified line.  It then creates
 *	position codes for these positions and removes the associated
 *	list structures.
 *
 *	Input:
 *		pTextCompInfo:	Ptr to all text compression info
 *		lineNum:	the Y coord of the positions to be
 *				committed.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_bufferExceeded	Couldn't get enough space for
 *					the committed position bytes.
 *
 **************************************************************************/
Int32  CDECL 
ip_commitTCPositions (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    Int32	lineNum)
{
    PROLOGUE;
    TCScanLineRun	*pPrevPos, *pCurPos;
    Int32		 last, curXPos;
    UInt8		*pPosBuf = pTextCompInfo->pPositionBuf;

	/* Make a worst-case guess about whether this stuff will
	 * fit into the position buffer.  If this is a bit convervative,
	 * big deal.  I don't think it's vital that we use up
	 * every last byte in the pos buffer before noticing
	 * that we can't fit any more bytes.
	 * We assume 5 bytes per position (all symbols have the same
	 * x offset on same line, so dX = 0, dYLo = -1, dYHi = -1,
	 * xLo, xHi) plus 7 bytes for Y displacement
	 * (dX = 0 + 2 bytes of dY + (2 * ( dX = 0, dY = +1))) as the
	 * max.
	 */
    if ((pTextCompInfo->numUncomPos * 5) + 7 + pTextCompInfo->resultSize >
	 pTextCompInfo->positionBufSize)
	return ia_bufferExceeded;

	/* There's no guarantee that there are any symbols in the
	 * list that are on this line.  Rather than check for whether
	 * every symbol found is the first symbol on the line,
	 * just go look through the list to see if we can find the
	 * first symbol in this separate loop.
	 */
    for (pPrevPos = &(pTextCompInfo->dum1stLoc), pCurPos = pPrevPos->pNext;
	 pCurPos->pNext != NULL;
	 pPrevPos = pCurPos, pCurPos = pCurPos->pNext)
    {
	if (pCurPos->runRight == (Int16) lineNum)
	    break;
    }

	/* did we find a first symbol? */
    if (pCurPos->pNext == NULL)
    {
	/* there was no symbol on this line in the list, so quit now.
	 * We haven't changed any state, so there's nothing to update.
	 */
	NORMAL_RETURN;
    }

	/* Yea, we found a first symbol!
	 * Put out the delta Y to get to this line.  If
	 * we happened to find a symbol that ended on line 0,
	 * skip that.
	 * In the old version of the internal format, a position
	 * buffer usually began with a first X of 0 and a deltaY
	 * relative to 0 to get to the first line.  This didn't
	 * work if the first symbol in the image was a single
	 * dot in the upper left corner (has a real firstX of 0).
	 * In the new format, the first thing in the position
	 * buffer is a deltaY relative to the previous line.
	 * The initial value for lastY is -1 so that we can
	 * distinguish a buffer that begins with a symbol
	 * on line 0 (deltaY of 1) from an empty position
	 * buffer (first deltaY of 0).
	 * We now put a deltaX of 0 at the END of the line
	 * where it belongs.
	 */
    if (lineNum != pTextCompInfo->lastY)
    {
	last = lineNum - pTextCompInfo->lastY;
	if (last < 128)
	    *pPosBuf++ = (UInt8) last;
	else
	{
	    *pPosBuf++ = (UInt8) (last | 0x80);
	    *pPosBuf++ = (UInt8) (last >> 7);
	}
	pTextCompInfo->lastY = lineNum;
    }

	/* Put out first X for this line */
    curXPos = pCurPos->runLeft;
    if (curXPos < 128)
	*pPosBuf++ = (UInt8) curXPos;
    else
    {
	*pPosBuf++ = (UInt8) (curXPos | 0x80);
	*pPosBuf++ = (UInt8) (curXPos >> 7);
    }

	/* Put out the symbol ID for this first symbol */
    last = (Int32) (pCurPos->pClient);
    if (last < 128)
	*pPosBuf++ = (UInt8) last;
    else
    {
	*pPosBuf++ = (UInt8) (last | 0x80);
	*pPosBuf++ = (UInt8) (last >> 7);
    }

	/* and remove this sym from the list */
    pPrevPos->pNext = pCurPos->pNext;
    i_freeStruct (pIPState,
		  pTextCompInfo->p1stScanLineBlock, (void **)&pCurPos);
    pTextCompInfo->numUncomPos--;

	/* Go through the list looking for any more symbols on
	 * this page.
	 */
    for (pCurPos = pPrevPos->pNext;
	 pCurPos->pNext != NULL;
	 pPrevPos = pCurPos, pCurPos = pCurPos->pNext)
    {
	if (pCurPos->runRight != (Int16) lineNum)
	    continue;		/* ignore irrevelent syms */

	    /* Found another symbol in this line.  Put out X
	     * coord and symbol.
	     */
	last = pCurPos->runLeft - curXPos;
	if (last != 0)		/* the normal case */
	{
	    if ((UInt32) last < 128)		/* don't put out neg coords
						 * as 1 byte */
		*pPosBuf++ = (UInt8) last;
	    else
	    {
		*pPosBuf++ = (UInt8) (last | 0x80);
		*pPosBuf++ = (UInt8) (last >> 7);
	    }
	}
	else		/* put out the special sequence for symbols with
			 * no X displacement */
	{
	    *pPosBuf++ = 0;		/* dX = 0 */
	    *pPosBuf++ = 0xFF;		/* dYLo = -1 */
	    *pPosBuf++ = 0xFF;		/* dYHi = -1 */
	    if (curXPos < 128)
		*pPosBuf++ = (UInt8) curXPos;
	    else
	    {
		*pPosBuf++ = (UInt8) (curXPos | 0x80);
		*pPosBuf++ = (UInt8) (curXPos >> 7);
	    }
	}
	curXPos += last;

	    /* Put out symbol ID */
	last = (Int32) (pCurPos->pClient);
	if (last < 128)
	    *pPosBuf++ = (UInt8) last;
	else
	{
	    *pPosBuf++ = (UInt8) (last | 0x80);
	    *pPosBuf++ = (UInt8) (last >> 7);
	}

	    /* and remove this sym from the list */
	pPrevPos->pNext = pCurPos->pNext;
	i_freeStruct (pIPState,
		      pTextCompInfo->p1stScanLineBlock, (void **)&pCurPos);
	pTextCompInfo->numUncomPos--;
	    /* to make the loop advancement condition work right,
	     * set pCurPos to pPrevPos.  This is like running
	     * in place since we just ate the thing in front of
	     * pPrevPos.
	     */
	pCurPos = pPrevPos;
    }

	/* That's it.  All the symbols on this line have been put out
	 * and their list entries removed.  Record the number of
	 * bytes used, update the position buf pointer and return.
	 * BUT FIRST, put out the deltaX of 0 signifying end of line.
	 */
    *pPosBuf++ = 0;
    pTextCompInfo->resultSize += (pPosBuf - pTextCompInfo->pPositionBuf);
    pTextCompInfo->pPositionBuf = pPosBuf;
    NORMAL_RETURN;
}




/* PRIVATE */
/**************************************************************************
 *   ip_foundEquivalentSymbol()
 *
 *	This procedure is called by i_consolidateShapeInfo when an
 *	equivalent symbol is found.  It inserts symbols into the linked
 *	lists of equivalence classes, or joins existing linked lists.
 *
 *	Input:
 *		pTextCompInfo:	Ptr to all text compression info
 *		pSymbol1:	Pointer to one of the equivalent symbols.
 *		pSymbol2:	Pointer to other equivalent symbol
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		Couldn't get enough memory for
 *					internal data structures.
 *
 **************************************************************************/
Int32  CDECL 
ip_foundEquivalentSymbol (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCSymbol   *pSymbol1,
    TCSymbol   *pSymbol2)
{
    PROLOGUE;
    TCSymbol	*pTempSymbol, *pTempSymbol2;
    Int32	 sumOfInstances;
    Int32	 delta;			/* used to compute ranges */
    

	/* We found two equivalent symbols.
	 * If neither symbol is equivalent to anything, make them
	 * equivalent to each other.
	 * If one of the symbols is already equivalent to something,
	 * insert the other one into the existing list.
	 * If both symbols are already equivalent to things,
	 * connect the lists to each other by swapping the pNewSymbol
	 * pointer of the symbols (makes a figure 8).
	 * Also increment the number of instances associated with
	 * both symbols.
	 * We use the instance numbers to choose which
	 * symbol will survive.
	 * BUG FIX: We were getting long chains of equivalent symbols
	 * like a == b == c == d == e, but a != e.  The first workaround
	 * for this was to check dimesions for a new symbol against
	 * all the elements of an equivalence class before adding it
	 * to the class.  This got messy (and slow), especially when
	 * trying to decide whether to merge two equivalence classes.
	 * The new method is half done here and half by ip_classifySymbol.
	 * When a symbol is created, we record its minimum and maximum
	 * matching widths and heights.  ip_classifySymbol won't match
	 * two symbols with widths that differ by too much.  It also
	 * won't match two equivalence classes if the range of dimensions
	 * in one doesn't include all the elements of the other.  Our job
	 * here is to maintain the dimension tolerances of all of the
	 * elements of each equivalence class so that a symbol matching
	 * any of them has legal dimensions to match all of them.
	 * Instead of modifying all the dimensions of all elements
	 * in a class, we create a new symbol that represents the
	 * dimensions of the class as a whole.  This symbol is created
	 * When two elements are joined.  When two classes are joined,
	 * their joint info is put into one of their common symbols
	 * and the other one is deleted.
	 *
	 * The common symbols really contain two sets of data.  One
	 * is the current range of widths and heights of symbols in
	 * the class.  The other is the range of minimum and maximum
	 * matching widths of symbols that could be included into
	 * this class.  The real widths and heights are kept in the
	 * minX, maxX, minY and maxY entries.  The ranges are kept
	 * in the minWidth, maxWidth, minHeight and maxHeight entries.
	 *
	 * For example, say we join two symbols.  Entries in the common
	 * class symbol are set as follows:
	 *	minX:	When joining two symbols: minimum of widths of syms;
	 *		When joining a symbol to a class: min of width of
	 *			symbol and width of class;
	 *		When joining two classes: minimum of widths of
	 *			two classes.
	 *	maxX:	Ditto for maximum widths.
	 *	minY:	Ditto for minimum heights.
	 *	maxY:	Ditto for maximum heights.
	 *	delta:	computed based on maximum dimension of the symbol
	 *	minWidth: When joining two symbols: maximum of (first sym's
	 *			width - delta and second sym's width
	 *			- delta).  If this is greater than
	 *			minX (minimum width of an element in the
	 *			class), then set minWidth to minX.
	 *		 When joining a symbol to a class: maximum of
	 *			(new sym's width - delta and class's maxX
	 *			- delta).  If this is greater than
	 *			minX (minimum width of an element in the
	 *			class), then set minWidth to minX.
	 *		When joining two classes: maximum of
	 *			(1st class's maxX - first class's delta
	 *			 and 2nd class's maxX - delta).  If this is
	 *			 greater than minX (minimum width of an
	 *			 element in the class), then set minWidth
	 *			 to minX.
	 *	maxWidth: Ditto for maximum widths.  Use minimum instead
	 *			of maximum in first step and compare for
	 *			less maxX than instead of greater than minX
	 *			in the second.
	 *	minHeight: Like minWidth, but use heights instead.
	 *	maxHeight: Like maxWidth, but use heights instead.
	 *
	 */
    if ((pSymbol1->pNewSymbol == NULL) && (pSymbol2->pNewSymbol == NULL))
    {
	    /* create a new symbol to represent the common class */
	__error__ =
	    i_allocateStruct (pIPState, pTextCompInfo->p1stSymbolBlock,
			      (void **)&pTempSymbol);
	if (__error__ != ia_successful)
	    goto tcFoundEquivCleanup;

	    /* make both symbols point to the common symbol too */
	pSymbol1->pClassSymbol = pTempSymbol;
	pSymbol2->pClassSymbol = pTempSymbol;

	pSymbol1->pNewSymbol = pSymbol2;
	pSymbol2->pNewSymbol = pSymbol1;

	    /* Set the tolerances in the class symbol. */
	pTempSymbol->maxX = MAX(pSymbol1->maxX, pSymbol2->maxX) + 1;
	pTempSymbol->minX = MIN(pSymbol1->maxX, pSymbol2->maxX) + 1;
	pTempSymbol->maxY = MAX(pSymbol1->maxY, pSymbol2->maxY) + 1;
	pTempSymbol->minY = MIN(pSymbol1->maxY, pSymbol2->maxY) + 1;
	delta = (MAX(pTempSymbol->minX, pTempSymbol->maxY) -
			cDimensionOffset)/cDimensionDivisor;
	delta = MIN(delta, cMaxDimensionSlop);
	delta = MAX(delta, 0);	/* I'm too confused by negative deltas */
	    /* Save the class's deltaWidth in slant.  It will
	     * be used when joining symbols to this class or this
	     * class to another class.
	     */
	pTempSymbol->slant = delta;
	pTempSymbol->minWidth = MIN(pTempSymbol->maxX - delta,
				    pTempSymbol->minX);
	pTempSymbol->maxWidth = MAX(pTempSymbol->minX + delta,
				    pTempSymbol->maxX);

	pTempSymbol->minHeight = MIN(pTempSymbol->maxY - delta,
				     pTempSymbol->minY);
	pTempSymbol->maxHeight = MAX(pTempSymbol->minY + delta,
				     pTempSymbol->maxY);
	pTempSymbol->strokeWidth = (pSymbol1->strokeWidth +
				       pSymbol2->strokeWidth) / 2;
	    /* We want to be conservative about turning little blobs
	     * into anything larger.  For this reason, we decide that
	     * the class symbol is small (when in doubt, do the test
	     * for small stuff) and has a small number of edge pixels
	     * (produces the tightest check).
	     */
	pTempSymbol->maxDimension = MIN(pSymbol1->maxDimension,
				        pSymbol2->maxDimension);
    }
    else	/* At least one of these is equivalent to something */
    if (pSymbol2->pNewSymbol == NULL)	/* and it's not pSymbol2 */
    {
	pSymbol2->pNewSymbol = pSymbol1->pNewSymbol;
	pSymbol1->pNewSymbol = pSymbol2;
	    /* We adjust the dimensions of the class symbol in the
	     * same way as when joining two symbols,
	     * except we substitute pSymbol1->pClassSymbol for
	     * pSymbol1 when doing the calculations.
	     * First, make both symbols point to the common class symbol.
	     * Remember that slant holds the class's delta.  The new
	     * class element can only make the deltas smaller.
	     */
	pTempSymbol = pSymbol1->pClassSymbol;
	pSymbol2->pClassSymbol = pTempSymbol;

	    /* Set the tolerances in the class symbol.  Remember that
	     * pTempSymbol->minX already holds the minimum width of
	     * any element of the class.  pSymbol2->maxX + 1 holds the
	     * width of symbol 2.
	     */
	pTempSymbol->minX = MIN(pTempSymbol->minX, pSymbol2->maxX + 1);
	pTempSymbol->maxX = MAX(pTempSymbol->maxX, pSymbol2->maxX + 1);
	pTempSymbol->minY = MIN(pTempSymbol->minY, pSymbol2->maxY + 1);
	pTempSymbol->maxY = MAX(pTempSymbol->maxY, pSymbol2->maxY + 1);
	delta = MIN(pTempSymbol->slant,
		    (MAX(pTempSymbol->maxX, pTempSymbol->maxY) -
		     cDimensionOffset)/cDimensionDivisor);
	delta = MAX(delta, 0);	/* I'm too confused by negative deltas */
	if (pTempSymbol->slant != delta)
	    pTempSymbol->slant = delta;
	pTempSymbol->minWidth = MIN(pTempSymbol->maxX - delta,
				    pTempSymbol->minX);
	pTempSymbol->maxWidth = MAX(pTempSymbol->minX + delta,
				    pTempSymbol->maxX);

	pTempSymbol->minHeight = MIN(pTempSymbol->maxY - delta,
				     pTempSymbol->minY);
	pTempSymbol->maxHeight = MAX(pTempSymbol->minY + delta,
				     pTempSymbol->maxY);

    }
    else	/* Maybe pSymbol1 isn't equivalent to anything */
    if (pSymbol1->pNewSymbol == NULL)
    {
	pSymbol1->pNewSymbol = pSymbol2->pNewSymbol;
	pSymbol2->pNewSymbol = pSymbol1;
	    /* Do the same thing we did above with pSymbol1 and
	     * pSymbol2 reversed.
	     */
	pTempSymbol = pSymbol2->pClassSymbol;
	pSymbol1->pClassSymbol = pTempSymbol;

	    /* Set the tolerances in the class symbol.  Remember that
	     * pTempSymbol->minX already holds the minimum width of
	     * any element of the class.  pSymbol1->maxX + 1 holds the
	     * width of symbol 1.
	     */
	pTempSymbol->minX = MIN(pTempSymbol->minX, pSymbol1->maxX + 1);
	pTempSymbol->maxX = MAX(pTempSymbol->maxX, pSymbol1->maxX + 1);
	pTempSymbol->minY = MIN(pTempSymbol->minY, pSymbol1->maxY + 1);
	pTempSymbol->maxY = MAX(pTempSymbol->maxY, pSymbol1->maxY + 1);
	delta = MIN(pTempSymbol->slant,
		    (MAX(pTempSymbol->maxX, pTempSymbol->maxY) -
		     cDimensionOffset)/cDimensionDivisor);
	delta = MAX(delta, 0);	/* I'm too confused by negative deltas */
	if (pTempSymbol->slant != delta)
	    pTempSymbol->slant = delta;
	pTempSymbol->minWidth = MIN(pTempSymbol->maxX - delta,
				    pTempSymbol->minX);
	pTempSymbol->maxWidth = MAX(pTempSymbol->minX + delta,
				    pTempSymbol->maxX);

	pTempSymbol->minHeight = MIN(pTempSymbol->maxY - delta,
				     pTempSymbol->minY);
	pTempSymbol->maxHeight = MAX(pTempSymbol->minY + delta,
				     pTempSymbol->maxY);

    }
    else	/* they both already have lists.  We know they're
		 * not in the same list, that was checked in
		 * ip_classifySymbol.  Connect the two different
		 * lists by swapping the list pointers.  This makes
		 * a single list in the shapeof a figure 8.
		 */
    {
	pTempSymbol = pSymbol1->pNewSymbol;
	pSymbol1->pNewSymbol = pSymbol2->pNewSymbol;
	pSymbol2->pNewSymbol = pTempSymbol;

		/* Consolidate all the information in the two
		 * class symbols into the class symbol associated with
		 * pSymbol1.  This looks like joining pSymbol2 to a
		 * the class of pSymbol1, except we use pTempSymbol2
		 * instead of pSymbol2.  Also remember that we already
		 * have all the min and max quantities calculated, so
		 * don't need to do the maxX + 1 deal.
		 */
	pTempSymbol = pSymbol1->pClassSymbol;
	pTempSymbol2 = pSymbol2->pClassSymbol;

	    /* Set the tolerances in the class symbol.  Remember that
	     * pTempSymbol->minX already holds the minimum width of
	     * any element of the first class and pTempSymbol2 holds
	     * the min width of the second.
	     */
	pTempSymbol->minX = MIN(pTempSymbol->minX, pTempSymbol2->minX);
	pTempSymbol->maxX = MAX(pTempSymbol->maxX, pTempSymbol2->maxX);
	pTempSymbol->minY = MIN(pTempSymbol->minY, pTempSymbol2->minY);
	pTempSymbol->maxY = MAX(pTempSymbol->maxY, pTempSymbol2->maxY);
	delta = MIN(pTempSymbol->slant, pTempSymbol2->slant);
	if (pTempSymbol->slant != delta)
	    pTempSymbol->slant = delta;

	pTempSymbol->minWidth = MIN(pTempSymbol->maxX - delta,
				    pTempSymbol->minX);
	pTempSymbol->maxWidth = MAX(pTempSymbol->minX + delta,
				    pTempSymbol->maxX);
	pTempSymbol->minHeight = MIN(pTempSymbol->maxY - delta,
				     pTempSymbol->minY);
	pTempSymbol->maxHeight = MAX(pTempSymbol->minY + delta,
				     pTempSymbol->maxY);
	pTempSymbol->strokeWidth = (pTempSymbol->strokeWidth +
				       pTempSymbol2->strokeWidth) / 2;
	    /* We want to be conservative about turning little blobs
	     * into anything larger.  For this reason, we decide that
	     * the class symbol is small (when in doubt, do the test
	     * for small stuff) and has a small number of edge pixels
	     * (produces the tightest check).
	     */
	pTempSymbol->maxDimension = MIN(pTempSymbol->maxDimension,
				        pTempSymbol2->maxDimension);

	    /* free the class symbol from pSymbol2's class since we've
	     * consolidated its info into the other class symbol.
	     */
	i_freeStruct (pIPState, pTextCompInfo->p1stSymbolBlock,
		      (void **)&pTempSymbol2);

	    /* go around the list of pSymbol2's symbols and change
	     * the pClassSymbol pointers to point to the new
	     * class symbol for the merged class.
	     */
	for (pTempSymbol2 = pSymbol1->pNewSymbol;
	     pTempSymbol2 != pSymbol2->pNewSymbol;
	     pTempSymbol2 = pTempSymbol2->pNewSymbol)
	{
	    pTempSymbol2->pClassSymbol = pTempSymbol;
	}
    }

	/* having done all the joining of symbols and classes, record
	 * the popularity of both symbols.
	 */
    sumOfInstances = pSymbol1->instanceCount + pSymbol2->instanceCount;
    pSymbol2->numActiveRuns += sumOfInstances;
    pSymbol1->numActiveRuns += sumOfInstances;
    __error__ = ia_successful;

tcFoundEquivCleanup:
    NORMAL_RETURN;
}



/*

 $Log: mitcenc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:17  jlin
 Imported sources

 Revision 1.31  1998/08/04 00:17:05  aber
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 CommonCheck: Retain the page ID such that the dictionary can be split into a
 shared dictionary, and a single-use dictionary for each page.
 Support multiple dictionaries per page.

 Revision 1.30  1998/01/13 23:53:33  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.29  1997/12/05 20:01:09  ddavies
 ddavies on Fri Dec 5 12:00:44 PST 1997

 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Move ip_compareSymCounts to i_tctnut.c, where is should probably have been
 all along.

 Revision 1.28  1997/06/25 03:48:51  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Remove the routine ip_sortTCSymbolsInTundraDictOrder.  It's been moved
 to mitcence.c since it's needed by both lossless and lossy compression.

 Revision 1.27  1997/05/08 02:36:24  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Change type of ip_symbolIsComplete's insertOnly arg from Int32 to Bool.

 Revision 1.26  1997/04/29 02:49:31  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Fix up comments generally.
 Remove temporary code used before Tundra format was finished.

 Revision 1.25  1997/02/12 03:11:59  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Many changes having to do with using profiles to better align replacement
 symbols with the originals.  Also, create a new routine to do this so we
 can use it both when creating the position buffer initially and when
 recreating a position buffer after symbol consolidation.
 In ip_calculateSlant (that trash can), create left and right profiles of
 a new symbol.  These are words that have a bit for each of the bottom
 32 rows of a symbol.  A bit is set if the symbol sticks out far enough
 to the left or right in that row.  Also modify the calculation of
 minEnvWidth so that it's 7/8's of the width instead of 8/8's of the width.
 Was that on purpose?
 When a new symbol is made, copy the profiles and maxDimension from the
 instance.
 Also use ip_createUncommittedTCPosition to create the position buffer
 initially instead of just doing it in ip_symbolIsComplete.
 This new routine, ip_createUncommittedTCPosition, puts replacement
 symbols into the list of uncommitted symbols after aligning the
 replacement symbols with the originals.  It uses the new profiles
 to do the vertical alignment - moving the replacement up or down until
 the profiles match the best.
 In ip_foundEquivalentSymbol, set the joined symbols's maxDimension.

 Revision 1.24  1997/02/07 03:09:17  ddavies
 ddavies on Thu Feb 6 19:08:15 PST 1997

 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Fix bug in ip_symbolIsComplete in which __error__ was not being set by
 calls to i_allocateStruct, but was being checked.  This resulted in a
 segmentation fault when we ran out of memory instead of a polite
 ia_nomem return.

 Revision 1.23  1997/01/30 00:50:13  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Fix bug found when testing effects of symbol-based compression on OCR
 accuracy.  During symbol consolidation, we now require the stroke widths
 of the target symbol and the class's symbol to be very close.  This keeps
 urk, ensures that all all symbols in a class have the same stroke width.
 This, in turn, keeps us from doing violence to very small symbols when
 consolidating them.

 Revision 1.22  1997/01/04 00:33:43  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Replace the mass and interior measures with the strokeWidth measurement.
 This does a much better job of distinguishing bold from non-bold characters.
 It also does ok for distinguishing commas from periods.
 ip_calculateSlant now calculates a symbol's stroke width using horizontal
 and vertical runs.
 Remove calculation of massRange.  Have ip_foundEquivalentSymbol use a symbol's
 slant entry as scratch space instead of using massRange (since massRange
 is gone).

 Revision 1.21  1996/09/04 01:29:00  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Use a symbol's width and height entries instead of calculating them from
 maxX, minX, maxY, minY.  When placing symbols, don't place any either above
 the top of the image or below the bottom since these can't be decoded.

 Revision 1.20  1996/08/29 01:09:23  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Start using the new width and height members of a TCSymbol so we don't
 have to worry about whether the run coords have been converted.
 Activate and debug ip_sortTCSymbolsInTundraDictOrder.  It now goes through
 the symbols in order of width and height, but also reversed the order
 within symbols of the same dimension so that when they're read back in,
 the order is preserved.

 Revision 1.19  1996/08/23 03:02:52  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Fix bug in which residual symbols were not being assigned dictOrder values.
 This resulted in sorts which differed on the SPARC/gcc and PC/MSVC/C++
 platforms when consolidating shape info.
 Modify parsing of internal position info so the block begins with a deltaY
 relative to -1 instead of a deltaX.
 This allows us to represent a symbol consisting of a single dot in the
 upper left corner of a page, which would have caused confusion in the
 old format.

 Revision 1.18  1996/08/22 01:49:53  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Fussed with ip_compareSymCounts and ip_sortTCSymbolsInTundraDictOrder.
 The first routine is used by qsort.  It now puts symbols having both
 larger instanceCounts (used more often) and larger dictOrder values
 at one end of the list.  We had been putting large instanceCounts and
 small dictOrder at one end.  The new way is consistent with the way
 the Tundra spec says Huffman codes should be assigned (I hope).
 ip_sortTCSymbolsInTundraDictOrder is changed to a temporary dummy
 routine.  It used to do the right thing for the final Tundra dictionaries,
 which don't exist yet.  It now assigns the dictOrder value to symbols
 corresponding to reverse symbol ID, since symbols are now encoded
 into the old serialized dictionaries using symbolID.  We use reverse
 symbolID order because of a complicated interaction between
 the code that consolidates symbols (and produces a list that increases
 in symbol order and usage count) and the code used to generate
 Huffman codes (which wants a list that increases in usage count and
 dictOrder).  Ooops, above, I meant to say that the symbol consolidation
 code produces a list that increases in symbol ID and decreases in
 usage count (highest counts at the beginning).  We want to be able
 to use a slightly modified (reversed) version of this same list as the
 input to the Huffman code generator.

 Revision 1.17  1996/07/27 02:08:49  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Change TCCompressionInfo boolean symsHaveBeenEncoded to general encodeState
 variable with bit fields.
 Initialize the new pNextDictSym, dictOrder, huffmanCodeLen and huffmanCode
 entries for new dictionary symbols.
 Sort new big symbols by height, then width as they are inserted into the
 pTextCompInfo->pBigSymbols list.  This is the order in which they'll be
 written to the Tundra dictionary.
 Move ip_compareSymCounts here from i_tcenc.c.  Change it so that the secondar
 key is dictOrder instead of symbolID.  This is in keeping with the Tundra
 way of assigning huffman codes.
 Add new routine ip_sortTCSymbolsInTundraDictOrder.  This sorts the symbols
 by height, then width.  It creates a list of symbols and tags each one
 with its offset in the list.  This is used both when putting out symbols
 into the dictionary and when generating huffman codes.

 Revision 1.16  1996/07/22 18:43:51  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Added support for mild improvements in the vertical positioning of symbols.
 In the old code, the bottom line of each symbol was placed on the same
 line as the bottom line of the instance it replaced.  If there was noise
 on one, but not the other, symbols would jitter up or down a line.
 In the new code, we maintain a list of uncommitted symbol positions sorted
 by X coord.  This list can contain symbols on at most three lines at any
 one time, since instances on line n can affect at most symbols on lines
 n-1, n and n+1.  ip_symbolIsComplete now just puts symbols into the
 sorted list instead of putting out the position info directly.
 ip_symbolIsComplete calls a new routine, ip_commitTCPositions when the
 current line number is more than 1 greater than the lowest line number
 of any symbol in the uncommitted list.  ip_commitTCPositions removes
 any symbols in a given line from the uncommitted list, creating a new
 block of committed position info in the position buffer.
 How does ip_symbolIsComplete decide where to put a symbol?  It counts
 the number of pixels on the bottom rows of both symbol and instance.
 If the difference between these counts is too large, the one with the
 small count is deemed to have a row of noise which should be ignored.

 Revision 1.15  1996/07/19 00:28:43  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Initialize the new TCSymbol entry runsPtToTail to cFalse when creating a
 new dictionary symbol.

 Revision 1.14  1996/07/04 03:02:42  ddavies
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 Change order of symbol linked list, so p1stSymbol points to symbol with
 lowest symbol ID instead of symbol with the highest symbol ID.
 This also means that pSymbol->pNextNum points to the symbol with the next
 higher symbolID instead of the next lower one.

 Revision 1.13  1995/12/13 02:52:33  ejaquith
 mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.12  1995/10/18  04:04:20  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.11  1995/07/11  00:17:59  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Change references to arrays of pointers to TCScanLineRuns used to generate
 * vertical runs to reference to structure containing the array of pointers.
 * Change all casts in calls to i_allocateStruct from (TCBlockHeader **) to
 * (void **).  This is cleaner since we shouldn't pretend that each structure
 * in every heap is a TCBlockHeader.
 * Also change 2D array of symbol pointers to structure containing 2D array of
 * symbol pointers.  The beginning of the structure contains a PointerFixup
 * struct so pointers can be fixed if things move.
 *
 * Revision 1.10  1995/04/01  03:23:45  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Change references to pTextCompInfo->symIndex to (*(pTextCompInfo->pSymIndex)).
 * That's because the TCCompressionInfo structure now has a pointer to a 2D
 * array of symbol pointers instead of the array itself.
 *
 * Revision 1.9  1995/03/15  03:12:47  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Lotsa changes. ip_foundEquivalentSymbol now returns a return code.  The
 * routines ip_calculateSlant and ip_symbolIsComplete now take a TCInstance
 * structure instead of a bunch of parameters.
 * In ip_calculateSlant: delete calculate of global asymmetry.  Add calculate
 * of signed horizontal and vertical asymmetries, which are really the distance
 * from the central vertical and horizontal axes to the left and bottom lines
 * respectively.  Change the calculation of volume to horizontal volume and
 * vertical volume.  Further modify these so that we only count portions of
 * the symbol's envelope that are wide or tall enough.  This discriminates
 * against round things.  Add checkHVolume and checkVVolume which are only
 * set when the horizontal or vertical volume is a sufficiently large fraction
 * of the symbol's area (i.e. means something).
 * In ip_makeGroupsFromRuns: Fix the overlap calculation so that symbols
 * ooops, back up - so that diagonal lines that are 1 pixel wide are classified
 * as single lines, not as a bunch of disconnected dots.  These are found in
 * synthetic images.  Add assignment of leftEnd and rightEnd in each group.
 * These are the left and right endpoints of the middle scan line of the group.
 * They're used to track small differences between symbols and instances.
 * Also set the cAtLeftEnd and cAtRightEnd flags in each group that's at the
 * left and/or right end of the symbol.  Duplicate the groups in the first
 * scan line.  These are used to adjust position vertically or horizontally
 * by changing how the instance and exemplar are initially aligned.
 * In ip_symbolIsComplete: pass in a TCInstance instead of a bunch of parmaeters.
 * Delete the volumeDelta and asymmetry.  Add hVolume, vVolume, hSAsymmetry,
 * vSAsymmetry, checkHVolume and checkVVolume.  Insert a new symbol into the
 * list pointed to by the entry in pTextCompInfo's 2D array having the same
 * width and height as the symbol.  Symbols that are too big all go into the
 * list reserved for big symbols.
 * In ip_foundEquivalentSymbols: maintain the minimum and maximum allowed widths
 * and heights for equivalence classes when two symbols join or a symbol joins
 * a class or two classes join.  This prevents a long chain of equivalences
 * which leads to symbols of widely different dimensions being called equivalent.
 * Also set up a dummy symbol which holds the common width and height
 * parameters of the class as a whole.
 *
 * Revision 1.8  1994/11/20  22:58:47  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Minor improvement in horizontal placement of symbols on a page.  Attempted
 * to align the midpoint of the encoded symbol with the midpoint of the
 * symbol found on the page.
 * The old scheme simply put the right side of the exemplar at the right side
 * of the instance symbol.
 *
 * Revision 1.7  1994/11/16  02:46:27  ddavies
 * ddavies on Tue Nov 15 18:44:12 PST 1994
 *
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Completely redo ip_makeVerticalRuns for roughly a 50% performance
 * improvement.  This routine was taking about 1.6 sec when compressing
 * cseditor, now it takes 0.9 sec.  The new version uses a state machine
 * to only touch the vertical runs that are initiated or terminated.
 * The old version touched every bit of every vertical run.
 *
 * Revision 1.6  1994/11/10  21:53:07  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Changes to increase robustness and symbol discrimination.  Volume is a
 * new measure used to distinguish little, round, bold blobs.  It's calculated
 * by ip_calculateSlant since it uses most of the same input data.  Simplify
 * ip_calculateSlant, removing rightDelta since it wasn't used.
 * Also increase the amount of noise we tolerate on big symbols when
 * creating the groups that specify run endpoint tolerances.
 *
 * Revision 1.5  1994/10/14  01:31:05  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Fix bug in initializing vertical run arrays, was going beyond the end of
 * the arrays.  Add insertOnly arg to ip_classifySymbol.  When set, ip_classifySymbol
 * makes a new symbol from a set of runs and a bounding box and puts it into
 * the lists of symbols.  It does not put out position info.  This is used
 * when decoding symbols from a bytestream.  Also when decoding the symbols,
 * just attach the runs passed in to the new symbol instead of copying them
 * to a new set of structures as is done during classification.
 *
 * Revision 1.4  1994/09/08  00:35:21  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Zillions of changes.  We now make group runs for both horizontal and
 * vertical groups.  Each group run has a type that specifies the section
 * of code used to match the group to a run from the instance symbol.
 * All of the magic constants have been labelled and moved to shrcmprs.prv.
 *
 * Revision 1.2  1994/08/03  19:17:47  ddavies
 * mitcenc.c -> /product/ipcore/ipshared/src/RCS/mitcenc.c,v
 *
 * Snapshot of version of symbol-based compression code that does a good job
 * of distinguishing symbols.
 *
 * Revision 1.1  1994/07/14  21:04:46  ddavies
 * Initial revision
 *

 */


