/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mitcence.c: Contains the utility procedures used to turn an image into a
 *		set of symbols and positions of symbols using lossless
 *		symbol comparison.  These are called by procedures in
 *		i_tcence.c.
 *
 *		Int32	ip_exactSymbolIsComplete()
 *			Does final classification of symbols, assigns
 *			symbolIDs and positions.
 *
 *		Int32	ip_sortTCSymbolsInTundraDictOrder()
 *			Go through the symbols in the order in which they'll
 *			be written out in the dictionary, stringing them
 *			together in a list and marking each symbol with
 *			its number in the list.
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

    /*  prototyes */
#include "shrcmprs.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mitcence.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* The process we use to compress images by finding repeated symbols is
 * well documented (at least there's lots of documentation) in the
 * file i_tcompr.c.  However, much of that doesn't apply here since we're
 * doing a simple exact match.
 *
 * This file contains utility routines called by the main encoding functions
 * in i_tcence.c.
 */

/* PRIVATE */
/**************************************************************************
 *   ip_exactSymbolIsComplete()
 *
 *	This procedure is called for each completed symbol.  It determines
 *	the final identity of the symbol and outputs position information.
 *
 *	Input:
 *		pTextCompInfo:		Ptr to all text compression info
 *		pSymbol:		Ptr to symbol matching the new
 *					instance.  This may be NULL if
 *					no match was found.
 *		pInstance:		Ptr to information about instance
 *		insertOnly:		When cTrue, just insert this symbol
 *					into the list of symbols, don't put
 *					out any position info.  ALSO, when
 *					this is set, we just attach the incoming
 *					set of horizontal runs to the symbol
 *					instead of copying them to a new set.
 *
 *	Action:
 *		If no match was found, make up a new symbol and fill in
 *		its entries, insert it into the chain of symbols and
 *		copy the horizontal runs to new, permanent structures.  In
 *		any case, put position information into the output.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_nomem		ran out of memory
 *		ia_bufferExceeded	ran off the end of the position buffer.
 *
 **************************************************************************/
Int32 CDECL 
ip_exactSymbolIsComplete (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    TCSymbol   *pSymbol,
    TCInstance *pInstance,
    Bool	insertOnly)
{
    PROLOGUE;
    TCScanLineRun	*pNewScanLineRun;  /* used to allocate new scan line
					    * runs when we make a new symbol.
					    */
    TCScanLineRun	*pOldScanLineRun;  /* used to copy old scan line runs */
    TCSymbol		*pTempSymbol;	   /* scratch ptr */
    Int32		 x, y;		   /* coords of sym's Lower right
					    * corner */
    Int32		 delta;		   /* dist to move */
    UInt8		*pPosBuf;	   /* local position buf ptr */
    

    __error__ = ia_successful;		   /* hope it stays this way */

	/* Did we find a symbol?
	 * If not, fill in the rest of they symbol's entries and
	 * insert it into the shape data structure.  This
	 * involves allocating new scan line runs and copying all their
	 * data.
	 */
    pTextCompInfo->totalSymbols++;	/* for debugging */

    if (pSymbol == NULL)
    {
	__error__ = i_allocateStruct (pIPState, pTextCompInfo->p1stSymbolBlock,
			            (void **) &pSymbol);
	if (__error__ != ia_successful)
	    goto symIsCompCleanup;

	pSymbol->width = pInstance->w;
	pSymbol->height = pInstance->h;
	pSymbol->minX = -(pInstance->w - 1);
	pSymbol->maxX = (pInstance->w - 1);
	pSymbol->minY = -(pInstance->h - 1);
	pSymbol->maxY = (pInstance->h - 1);
	pSymbol->maxDimension = pInstance->maxDimension;
	pSymbol->minWidth = 0;
	pSymbol->maxWidth = 0;
	pSymbol->minHeight = 0;
	pSymbol->maxHeight = 0;
	pSymbol->mass = pInstance->mass;
	pSymbol->numRuns = pInstance->numRuns;

	    /* Reset all the funny stuff just to be tidy.
	     */
	pSymbol->leftProfile = 0;
	pSymbol->rightProfile = 0;
	pSymbol->strokeWidth = 0;
	pSymbol->slant = 0;
	pSymbol->checkHVolume = 0;
	pSymbol->checkVVolume = 0;
	pSymbol->hVolume = 0;
	pSymbol->vVolume = 0;
	pSymbol->hSAsymmetry = 0;
	pSymbol->vSAsymmetry = 0;
	pSymbol->numActiveRuns = 0;
	pSymbol->instanceCount = 1;
	pSymbol->pRunHead = NULL;
	pSymbol->pClassSymbol = NULL;
	pSymbol->pNewSymbol = NULL;
	pSymbol->pNext = NULL;		/* temporarily */
	pSymbol->pPrev = NULL;		/* temporarily */
	pSymbol->bytesPerLine = 0;
	pSymbol->pBitmap = NULL;
	pSymbol->xOffset = 0;
	pSymbol->pNextDictSym = NULL;
	pSymbol->dictOrder = -1;
	pSymbol->dictNum = pTextCompInfo->currentDictionaryNum;
	pSymbol->huffmanCodeLen = 0;
	pSymbol->huffmanCode = 0;
	pSymbol->p1stHorizGroup = NULL;
	pSymbol->p1stVertGroup = NULL;

	    /* make up new runs only if this is really a new symbol */
	if (!insertOnly)
	{
		/* copy the incoming set of horizontal runs to a new set
		 * especially for this symbol.  The incoming set will
		 * be deallocated when we return to
		 * i_findConnectedComponentRuns.
		 */
	    __error__ = i_allocateStruct (pIPState,
					pTextCompInfo->p1stScanLineBlock,
				        (void **) &pNewScanLineRun);
	    if (__error__ != ia_successful)
		goto symIsCompCleanup;
    
	    pSymbol->pRunTail = pNewScanLineRun;
	    pOldScanLineRun = pInstance->pHorizRuns;
	    pNewScanLineRun->runLeft = pOldScanLineRun->runLeft;
	    pNewScanLineRun->runRight = pOldScanLineRun->runRight;
		/* keep line tags for encoding */
	    pNewScanLineRun->pClient = pOldScanLineRun->pClient;
    
	    for (pOldScanLineRun = pOldScanLineRun->pNext;
		 pOldScanLineRun != NULL;
		 pOldScanLineRun = pOldScanLineRun->pNext)
	    {
		__error__ = i_allocateStruct (pIPState,
					    pTextCompInfo->p1stScanLineBlock,
					    (void **)
						&(pNewScanLineRun->pNext));
		if (__error__ != ia_successful)
		    goto symIsCompCleanup;
    
		pNewScanLineRun = pNewScanLineRun->pNext;
		pNewScanLineRun->runLeft = pOldScanLineRun->runLeft;
		pNewScanLineRun->runRight = pOldScanLineRun->runRight;
		pNewScanLineRun->pClient = pOldScanLineRun->pClient;
	    }
	    pNewScanLineRun->pNext = NULL;
	}
	else	/* just use the runs that were supplied as the horizontal
		 * runs and don't worry about the rest */
	{
	    pSymbol->pRunTail = pInstance->pHorizRuns;
	}
	pSymbol->runsPtToTail = cFalse;	/* the runs' pNext ptrs point to
					     * pRunHead. */

	    /* insert this symbol into the array of symbols.  For want of
	     * a more beautiful policy, just put it at the head of
	     * the list of symbols with the same width and height.
	     * If either the width or height is too big, put it
	     * in the special list of big symbols.
	     */
	pSymbol->symbolID = pTextCompInfo->nextSymbolNum;
	pTextCompInfo->nextSymbolNum++;
	pSymbol->pNextNum = NULL;
	pSymbol->pPrevNum = pTextCompInfo->pLastSymbol;
	if (pSymbol->pPrevNum != NULL)	/* This is NOT the first symbol */
	    pSymbol->pPrevNum->pNextNum = pSymbol;
	else	/* this IS the first symbol */
	    pTextCompInfo->p1stSymbol = pSymbol;
	pTextCompInfo->pLastSymbol = pSymbol;

	pSymbol->pPrev = NULL;
	if ((pInstance->w > cMaxSymIndexDim) ||
	    (pInstance->h > cMaxSymIndexDim))
	{
		/* It will be convenient later to have this list
		 * sorted by height, then width.  This makes
		 * it easier to put symbols out in a Tundra
		 * dictionary, in height/width order.
		 */
	    pTempSymbol = pTextCompInfo->pBigSymbols;
	    if ((pTempSymbol == NULL) ||
		(pSymbol->maxY < pTempSymbol->maxY) ||
		((pSymbol->maxY == pTempSymbol->maxY) &&
		 (pSymbol->maxX <= pTempSymbol->maxX)))
	    {
		    /* the list of big symbols is empty or we
		     * should insert the new symbol before the
		     * first existing symbol
		     */
		pSymbol->pNext = pTextCompInfo->pBigSymbols;
		pSymbol->pPrev = NULL;
		if (pSymbol->pNext != NULL)
		    pSymbol->pNext->pPrev = pSymbol;
		pTextCompInfo->pBigSymbols = pSymbol;
	    }
	    else
	    {
		    /* Insert the new symbol somewhere after the
		     * first symbol.  First find the right height,
		     * then the right place in symbols of the same
		     * height.
		     */
		while ((pTempSymbol->pNext != NULL) &&
			(pSymbol->maxY > pTempSymbol->pNext->maxY))
		    pTempSymbol = pTempSymbol->pNext;
		while ((pTempSymbol->pNext != NULL) &&
			(pSymbol->maxY == pTempSymbol->pNext->maxY) &&
			(pSymbol->maxX > pTempSymbol->pNext->maxX))
		    pTempSymbol = pTempSymbol->pNext;

		    /* Insert pSymbol before pTempSymbol->pNext */
		pSymbol->pNext = pTempSymbol->pNext;
		if (pSymbol->pNext != NULL)
		    pSymbol->pNext->pPrev = pSymbol;
		pSymbol->pPrev = pTempSymbol;
		pTempSymbol->pNext = pSymbol;
	    }
	}
	else
	{
	    pSymbol->pNext =
		(pTextCompInfo->pSymIndex->array)[pInstance->w][pInstance->h];
	    (pTextCompInfo->pSymIndex->array)[pInstance->w][pInstance->h] =
		pSymbol;
	}
	if (pSymbol->pNext != NULL)
	    pSymbol->pNext->pPrev = pSymbol;

    }
    else	/* we found a match and pSymbol points to it */
	pSymbol->instanceCount++;		/* count the instances */

	/* if we're only supposed to insert the symbol into the data
	 * structures, as happens when we're decoding shape info from
	 * a serial bytestream, just stop now.
	 */
    if (insertOnly)
    {
	__error__ = ia_successful;
	goto symIsCompCleanup;
    }

	/* At this point, we're ready to insert the new symbol's position
	 * into the position file.
	 * But first, make a worst-case guess about whether this stuff will
	 * fit into the position buffer.  If this is a bit convervative,
	 * big deal.  I don't think it's vital that we use up
	 * every last byte in the pos buffer before noticing
	 * that we can't fit any more bytes.
	 * We assume 5 bytes per position (all symbols have the same
	 * x offset on same line, so dX = 0, dYLo = -1, dYHi = -1,
	 * xLo, xHi) plus 7 bytes for Y displacement
	 * (dX = 0 + 2 bytes of dY + (2 * ( dX = 0, dY = +1))) as the
	 * max.
	 */
    if (5 + 7 + pTextCompInfo->resultSize > pTextCompInfo->positionBufSize)
	return ia_bufferExceeded;

	/* point to right location in position buffer */
     pPosBuf = pTextCompInfo->pPositionBuf;

	/* calculate the correct X and Y positions of the lower right
	 * corner of this symbol.
	 */
     x = pInstance->x + pInstance->w - 1;
     y = pInstance->y + pInstance->h - 1;

	/* See if we need to issue a new y position.  This happens if
	 * the current Y position is different from the previous
	 * Y position.
	 */
    if (y != pTextCompInfo->lastY)
    {
	    /* If this isn't the first symbol, end the previous line */
	if (pTextCompInfo->lastY >= 0)
	    *pPosBuf++ = 0;

	    /* Put out the delta Y to this line */
	delta = y - pTextCompInfo->lastY;
	if (delta < 128)
	    *pPosBuf++ = (UInt8) delta;
	else
	{
	    *pPosBuf++ = (UInt8) (delta | 0x80);
	    *pPosBuf++ = (UInt8) (delta >> 7);
	}
	pTextCompInfo->lastY = y;

	    /* Put out first X for this line */
	if (x < 128)
	    *pPosBuf++ = (UInt8) x;
	else
	{
	    *pPosBuf++ = (UInt8) (x | 0x80);
	    *pPosBuf++ = (UInt8) (x >> 7);
	}
    }
    else
    {
	    /* put out the deltaX to next symbol.  If it's 0,
	     * put out the special code.
	     */
	delta = x - pTextCompInfo->lastX;
	if (delta != 0)
	{
		/* This is the normal case */
	    if ((delta > 0) && (delta < 128))
		*pPosBuf++ = (UInt8) delta;
	    else
	    {
		*pPosBuf++ = (UInt8) (delta | 0x80);
		*pPosBuf++ = (UInt8) (delta >> 7);
	    }
	}
	else	/* put out the special sequence for symbols with
		 * no X displacement */
	{
	    *pPosBuf++ = 0;		/* dX = 0 */
	    *pPosBuf++ = 0xFF;		/* dYLo = -1 */
	    *pPosBuf++ = 0xFF;		/* dYHi = -1 */
	    if (x < 128)
		*pPosBuf++ = (UInt8) x;
	    else
	    {
		*pPosBuf++ = (UInt8) (x | 0x80);
		*pPosBuf++ = (UInt8) (x >> 7);
	    }
	}
    }

	/* Put out the symbol ID for this first symbol */
    if (pSymbol->symbolID < 128)
	*pPosBuf++ = (UInt8) pSymbol->symbolID;
    else
    {
	*pPosBuf++ = (UInt8) (pSymbol->symbolID | 0x80);
	*pPosBuf++ = (UInt8) (pSymbol->symbolID >> 7);
    }

	/* update the current position buf info */
    pTextCompInfo->lastX = x;
    pTextCompInfo->resultSize += (pPosBuf - pTextCompInfo->pPositionBuf);
    pTextCompInfo->pPositionBuf = pPosBuf;

	/* if we arrive here with __error__ still equal to ia_successful, great.
	 * Even if something went wrong, all we do is return __error__.
	 */
symIsCompCleanup:
    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_sortTCSymbolsInTundraDictOrder()
 *
 *	This procedure runs through the list of symbols in the
 *	TCCompressionInfo structure in the order in which they'll
 *	appear in an output dictionary.  The symbols are strung
 *	together in a list and each symbol is marked with the symbol's
 *	dictionary number (distance from the beginning of the dict).
 *
 *	Input:
 *		pTextCompInfo:		Ptr to all text compression info
 *		preserveReadOrder:	Preserve the order the symbols were
 *					read from the dictionaries
 *
 *	Returns:
 *		ia_successful		things went well
 *
 **************************************************************************/
Int32  CDECL 
ip_sortTCSymbolsInTundraDictOrder (
    IPState    *pIPState,
    TCCompressionInfoPtr pTextCompInfo,
    UInt32 preserveReadOrder)
{
    PROLOGUE;
    TCSymbol	*pSymbol;	/* the destination symbols */
    TCSymbol	*pBigSymbol, *pRegSymbol;	/* source symbols */
    TCSymbol	*pEndBigSymGroup;		/* end of current big sym
						 * group. */
    TCSymbol	*pEndPrevBigSymGroup;		/* end of previous big sym
						 * group.
						 */
    Int32	 dictOrder;	/* number of current symbol */
    Int32	 width, height;	/* used to iterate through all regular
				 * symbols */

	/* If there aren't any symbols or they're already sorted,
	 * return happily */
    if ((pTextCompInfo->nextSymbolNum == 0) ||
	(pTextCompInfo->encodeState & cTCSymsAreInDictOrder))
	NORMAL_RETURN;

	/* Point to the first normal symbol and the first big symbol.
	 * We put symbols out sorted by height, then width.  If a
	 * big symbol belongs in a height class, we'll put it out
	 * in the proper class.  That means we have to inspect
	 * the list of big symbols at the end of each height class.
	 *
	 * There is one funny feature of this sorting.  When a symbol
	 * of width w and height h is read back in from an incoming
	 * dictionary, it's put at the head of the list of symbols
	 * of dimension [w, h].  This is true whether the symbol
	 * is a regular symbol or a big symbol.  If we put symbols into
	 * the dictionary by starting from the head of the list, the
	 * order of the symbols will be reversed inside each dimension
	 * class when they're read back in.  So?  Well,
	 * the order in which the symbols appear should be consistent
	 * because this order is used to generate the symbol's
	 * Huffman codes.  Yeah, when reading the symbols from
	 * the dictionary, we could just assign them the numbers
	 * from the dictionaries, and be careful never to call
	 * this routine, but that looks like a giant bug opportunity
	 * to me.
	 * To solve this problem, we have to reverse things somewhere.
	 * The somewhere is here.  We will declare that dictionary order
	 * starts at the *end* of each group of symbols having the
	 * same dimensions and proceeds towards the head.
	 *
	 * That's ok for symbols pointed to by the big 2D symbol
	 * array.  What about big symbols?  All dimension classes
	 * for big symbols are in the same list.  As before, new big
	 * symbols are inserted before older symbols having the
	 * same dimensions, so when computing dictionary order,
	 * we must go to the end of each block of symbols of the
	 * same dimension and work backwards.
	 * Instead of doing this weird sorting everywhere we
	 * want to follow the next big symbol, we just supply
	 * a new set of pointers to big symbols right now.
	 * Within a group of big symbols having the same
	 * dimensions, we want to follow the pPrev pointers,
	 * the first symbol of a big symbol group should point
	 * to the last symbol of the next *larger* group.
	 */
    pBigSymbol = pTextCompInfo->pBigSymbols;
    if (pBigSymbol != NULL)
    {
	    /* first, find the end of the list of big symbols */
	while (pBigSymbol->pNext != NULL)
	    pBigSymbol = pBigSymbol->pNext;

	pEndPrevBigSymGroup = NULL;	/* there was no prev dimension class */
	pEndBigSymGroup = pBigSymbol;	/* end of current dimension class */

	height = pBigSymbol->height;
	width = pBigSymbol->width;

	    /* Here's the loop where we move from the
	     * end of the list of big symbols towards
	     * the beginning, stringing them together.
	     * Each time we find the first symbol in a
	     * dimension group, we point it at the last
	     * symbol in the previous dimension group
	     * and recompute the dimensions.  If the symbol
	     * being investigated is not the first symbol
	     * of a dimension group, we just point it
	     * at the previous symbol so the order within
	     * the dimension groups is properly reversed.
	     */
	while (pBigSymbol->pPrev != NULL)
	{
	    if ((pBigSymbol->pPrev->height != height) ||
		(pBigSymbol->pPrev->width != width))
	    {
		pBigSymbol->pClassSymbol = pEndPrevBigSymGroup;
		pEndPrevBigSymGroup = pEndBigSymGroup;
		pEndBigSymGroup = pBigSymbol = pBigSymbol->pPrev;
		height = pBigSymbol->height;
		width = pBigSymbol->width;
	    }
	    else
	    {
		pBigSymbol->pClassSymbol = pBigSymbol->pPrev;
		pBigSymbol = pBigSymbol->pPrev;
	    }
	}
	    /* The first big symbol is by definition the first one
	     * in its group, so string it together that way.
	     */
	pBigSymbol->pClassSymbol = pEndPrevBigSymGroup;
	pBigSymbol = pEndBigSymGroup;	/* this is the first big symbol
					 * we want to process.
					 */
    }

    pTextCompInfo->p1stDictSymbol = NULL;	/* start over */
    dictOrder = 0;
    pSymbol = NULL;
    width = 1;					/* to keep the compiler happy */
    pRegSymbol = NULL;				/* ditto */

	/* Go find the first symbol.  Initializing this one
	 * separately makes the main loop a lot easier.
	 */
    if (pBigSymbol != NULL)
    {
	for (height = 1;
	     height <= MIN(pBigSymbol->height, cMaxSymIndexDim);
	     height++)
	{
	    for (width = 1; width <= cMaxSymIndexDim; width++)
	    {
		if (pTextCompInfo->pSymIndex->array[width][height] != NULL)
		{
		    for (pSymbol =
			    pTextCompInfo->pSymIndex->array[width][height];
			 pSymbol->pNext != NULL;
			 pSymbol = pSymbol->pNext)
			;
		    pRegSymbol = pSymbol->pPrev;
		    break;
		}
	    }

	    if (pSymbol != NULL)
		break;
	}

	    /* Did we find a regular symbol, or do we start with
	     * the first big symbol? */
	if (pSymbol == NULL)
	{
	    pSymbol = pBigSymbol;   /* Start with first big symbol */
	    pBigSymbol = pBigSymbol->pClassSymbol;

	    /* Account for the fact that we incremented height one
	     * more than we wanted in the for loop, fix 6/17/98 DA */
	    height--;
	    width = 1;
	}
    }
    else	/* no big symbols */
    {
	for (height = 1; height <= cMaxSymIndexDim; height++)
	{
	    for (width = 1; width <= cMaxSymIndexDim; width++)
	    {
		if (pTextCompInfo->pSymIndex->array[width][height] != NULL)
		{
		    /* If this dimension class is not empty,
		     * find the last symbol in it so the order within
		     * the class will be reversed.
		     */
		    for (pSymbol =
			    pTextCompInfo->pSymIndex->array[width][height];
			 pSymbol->pNext != NULL;
			 pSymbol = pSymbol->pNext)
			;
		    pRegSymbol = pSymbol->pPrev;
		    break;
		}
	    }

	    if (pSymbol != NULL)
		break;
	}
    }

	/* We now encounter the major cases:
	 * 1. There are both big and little symbols;
	 * 2. There are only little symbols;
	 * 3. There are only big symbols;
	 * 4. There are no symbols.
	 * The main line code handles the first case.
	 * The main line code minus a little clause at the bottom
	 * of the routine handles the second case.
	 * If there are no little symbols, we skip the loop that
	 * continues to look through little symbols.
	 * The last case can happen if we've already called this
	 * routine and ALL the original symbols were residual.
	 * In that case, they will all have been removed from the
	 * size array and from the big symbol list.  If that
	 * happens, we just skip everything.
	 */
    if (pSymbol != NULL)
    {
	pTextCompInfo->p1stDictSymbol = pSymbol;
    
	    /* Now find the rest of the symbols in dictionary order.
	     * Note that the inner loop is weird so that we can continue
	     * with the symbol after the first one found above.  Just
	     * starting over (either all the symbols or just the
	     * current row) inserts the first symbol twice.
	     */
	for (; height <= cMaxSymIndexDim; height++)
	{
	    while (cTrue)
	    {
		while (pRegSymbol != NULL)
		{
		    pSymbol->pNextDictSym = pRegSymbol;
		    pSymbol->dictOrder = dictOrder++;
		    pSymbol = pRegSymbol;
		    pRegSymbol = pRegSymbol->pPrev;
		}
		width++;
		if (width <= cMaxSymIndexDim)
		{
		    pRegSymbol = pTextCompInfo->pSymIndex->array[width][height];
			/* If this dimension class is not empty,
			 * find the last symbol in it so the order within
			 * the class will be reversed.
			 */
		    if (pRegSymbol != NULL)
		    {
			while (pRegSymbol->pNext != NULL)
			     pRegSymbol = pRegSymbol->pNext;
		    }
		}
		else
		    break;
	    }
    
	    width = 1;	/* reset for next time through the loop */
		/* Read pRegSymbol for the next time through the loop.
		 * If we don't do it here, we'll skip over symbols with
		 * width = 1.
		 */
	    if (height < cMaxSymIndexDim)
		pRegSymbol = pTextCompInfo->pSymIndex->array[width][height+1];
    
	       /* Link in any big symbols that are this tall */
	    while ((pBigSymbol != NULL) &&
		    (pBigSymbol->height == height))
	    {
		pSymbol->pNextDictSym = pBigSymbol;
		pSymbol->dictOrder = dictOrder++;
		pSymbol = pBigSymbol;
		pBigSymbol = pBigSymbol->pClassSymbol;
	    }
	}
    }	/* end of case where there are some little symbols */
    else	/* there are no little symbols.  We should start
		 * with the first big symbol if it exists.
		 */
    if (pBigSymbol != NULL)
    {
	pTextCompInfo->p1stDictSymbol = pBigSymbol;
	pSymbol = pBigSymbol;
	pBigSymbol = pBigSymbol->pClassSymbol;
    }

	/* Link in any remaining big symbols */
    while (pBigSymbol != NULL)
    {
	pSymbol->pNextDictSym = pBigSymbol;
	pSymbol->dictOrder = dictOrder++;
	pSymbol = pBigSymbol;
	pBigSymbol = pBigSymbol->pClassSymbol;
    }

	/* Mark the next pointer of the last symbol, if any */
    if (pSymbol != NULL)
    {
	pSymbol->pNextDictSym = NULL;
	pSymbol->dictOrder = dictOrder++;
    }

    /* The "combined" dictionary must really be split into a concatenation
     * of multiple dictionaries, whether reading or writing. At this point,
     * we have the dictionaries ranked by sizes (as written in Tundra
     * format), but they are interleaved based on dictionary separation.
     * If appropriate, the dictionary list is reordered here to account
     * for the fact that when split, they will be concatenated for decoding.
     */
    if (pTextCompInfo->maxDictionaryNum > 0)
    {
	Int32 dictNum;

	/* Loop through each dictionary set, and separate the symbols from
	 * that dictionary and the ones before, from the symbols in a
	 * later dictionary */
	for (dictNum = 0; dictNum < pTextCompInfo->maxDictionaryNum; dictNum++)
	{
	    TCSymbol *pFirstChain;	/* Chain of symbols in size class with
					 * count > 1 */
	    TCSymbol *pSecondChain;	/* Chain of symbols in size class with
					 * count == 1 */
	    TCSymbol *pEndOfFirstChain; /* Last in first chain */
	    TCSymbol *pEndOfSecondChain;/* Last in second chain */

	    /* Loop through all of the symbols */
	    pSymbol = pTextCompInfo->p1stDictSymbol;

	    /* Start with empty chains */
	    pFirstChain = pEndOfFirstChain = NULL;
	    pSecondChain = pEndOfSecondChain = NULL;

	    /* Loop through the remaining symbols */
	    while (pSymbol)
	    {
		/* Decide which chain to place it on */
		if (pSymbol->dictNum <= dictNum)
		{
		    /* Add to the end of the first chain */
		    if (pFirstChain == NULL)
		    {
			pFirstChain = pEndOfFirstChain = pSymbol;
		    }
		    else
		    {
			pEndOfFirstChain->pNextDictSym = pSymbol;
		    }
		    pEndOfFirstChain = pSymbol;
		}
		else
		{
		    /* Add to the end of the first chain */
		    if (pSecondChain == NULL)
		    {
			pSecondChain = pEndOfSecondChain = pSymbol;
		    }
		    else
		    {
			pEndOfSecondChain->pNextDictSym = pSymbol;
		    }
		    pEndOfSecondChain = pSymbol;
		}
		pSymbol = pSymbol->pNextDictSym;
	    }
		    
	    /* If there are two distinct chains, relink them back in */
	    if (pFirstChain != NULL && pSecondChain != NULL)
	    {
		UInt32 nextID = 0;	/* Next ID during renumbering */
		
		pTextCompInfo->p1stDictSymbol = pFirstChain;

		pEndOfFirstChain->pNextDictSym = pSecondChain;
		pEndOfSecondChain->pNextDictSym = NULL;

		/* Renumber the two chains */
		for (pSymbol = pFirstChain; pSymbol != NULL;
		pSymbol = pSymbol->pNextDictSym)
		{
		    pSymbol->dictOrder = nextID++;
		}
	    }
	}
    }

    /* Check to see if the sort is really needed */
    if (preserveReadOrder)
    {
	Int32 nextID = 0;

	for (pSymbol = pTextCompInfo->p1stDictSymbol;
	    pSymbol != NULL; pSymbol = pSymbol->pNextDictSym)
	{
	    if (pSymbol->dictOrder != nextID++)
		break;
	}
	if (pSymbol == NULL)
	    preserveReadOrder = cFalse;	/* No need to do the sort that
					 * comes next */
    }

    /* If maintaining read order, we need to sort them in the linked-list,
     * according to symbolID. */
    if (preserveReadOrder)
    {
	/* Ideally, we won't ever get here. In case we do, this sort assures we
	 * don't run into trouble. In the past, I've tests hundreds of files
	 * with:
	 * return ia_internal;
	 * here instead, and this code was never run */
	TCSymbol *pNewStart = NULL, *pNextNew = NULL;
	TCSymbol *pPrevSymbol;
	Int32 nextID = 0;

	pPrevSymbol = NULL;
	pSymbol = pTextCompInfo->p1stDictSymbol;

	while (pSymbol)
	{
	    /* Search for nextID */
	    while (pSymbol != NULL && pSymbol->symbolID != nextID)
	    {
		pPrevSymbol = pSymbol;
		pSymbol = pSymbol->pNextDictSym;
	    }
	    if (pSymbol == NULL || pSymbol->symbolID != nextID)
		return ia_internal;

	    /* Add it to the new list */
	    if (pNextNew == NULL)
	    {
		pNextNew = pNewStart = pSymbol;
	    }
	    else
	    {
		pNextNew->pNextDictSym = pSymbol;
		pNextNew = pSymbol;
	    }

	    /* Remove it from the old list */
	    if (pPrevSymbol == NULL)	/* The first symbol in the list */
	    {
		pTextCompInfo->p1stDictSymbol = pSymbol->pNextDictSym;
	    }
	    else
	    {
		pPrevSymbol->pNextDictSym = pSymbol->pNextDictSym;
	    }

	    /* Prepare for the next ID */
	    pSymbol = pTextCompInfo->p1stDictSymbol;
	    pPrevSymbol = NULL;
	    nextID++;
	}

	/* Use new list */
	pTextCompInfo->p1stDictSymbol = pNewStart;
	pNextNew->pNextDictSym = NULL;
    }

    /* If maintaining read order, simply assign the order read as the
     * dictionary order, as it will be used as the secondary key to sort
     * before assigning Huffman codes */
    if (preserveReadOrder)
    {
	pSymbol = pTextCompInfo->p1stDictSymbol;
	while (pSymbol)
	{
	    pSymbol->dictOrder = pSymbol->symbolID;
	    pSymbol = pSymbol->pNextDictSym;
	}
    }
 
    NORMAL_RETURN;
}



/*

 $Log: mitcence.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:17  jlin
 Imported sources

 Revision 1.6  1998/06/23 21:45:48  aber
 mitcence.c -> /product/ipcore/ipshared/src/RCS/mitcence.c,v

 CommonCheck: Fixes to force Tundra dictionary reading order, esp. for multiple
 dictionaries per page. Some cosmetic comment enhancements. A couple of
 extra error conditions caught.

 Revision 1.5  1998/01/13 23:53:36  ddavies
 mitcence.c -> /product/ipcore/ipshared/src/RCS/mitcence.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1997/06/25 03:51:30  ddavies
 mitcence.c -> /product/ipcore/ipshared/src/RCS/mitcence.c,v

 Add revised version of ip_sortTCSymbolsInTundraDictOrder.  This routine
 was moved from mitcenc.c since it's needed by both lossy and lossless
 compression.  The revisions allow it to handle the case where it's already
 been called for an image that contains only residual symbols, then is
 called again when there are no dictionary symbols.  Also fixed a bug
 in which it could overrun the pTextCompInfo->pSymIndex->array if the
 image contained only very tall symbols.


 */

