/*****************************************************************
 *  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mitcptr.c:	Utility routines used to fix up pointers in memory blocks
 *		after DAE garbage collection.
 *
 *   	Int32	ip_fixupTCCompressionInfoPtrs()
 *			Fix pointers inside the TCCompressionInfo block.
 *
 *   	Int32	ip_fixupSymbolArrayPtrs()
 *			Fix pointers inside 2D array of ptrs to TCSymbols.
 *
 *   	Int32	ip_fixupSymbolTablePtrs()
 *			Fix pointers inside 1D array of ptrs to TCSymbols.
 *
 *   	Int32	ip_fixupTCScanLineArrayPtrs()
 *			Fix pointers inside 1D array of ptrs to TCScanLineRuns.
 *
 *	Int32	ip_fixupTCGroupRunPtrs()
 *			Fix pointers inside memory block holding
 *			TCGroupRuns
 *
 *	Int32	ip_fixupTCSymbolPtrs()
 *			Fix pointers inside memory block holding
 *			TCSymbols
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"

    /*  prototyes */
#include "shrcmprs.pub"
#include "shrcmprs.prv"
#include "ptrfixup.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: mitcptr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* PRIVATE */
/**************************************************************************
 *   ip_fixupTCCompressionInfoPtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a TCCompressionInfo block.
 *
 *	Input:
 *		pThisBlock:	address of the TCCompressionInfoBlock.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in a TCCompressionInfo block to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupTCCompressionInfoPtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
    PROLOGUE;
    TCCompressionInfoPtr	pTextCompInfo;
    UInt32	oldAddr;
    UInt32	addrDiff;

	/* convert the local pointer to the right type */
    pTextCompInfo = (TCCompressionInfoPtr) pThisBlock;

	/* prepare for the comparisions */
    oldAddr = (UInt32) pOldAddr;

	/* Just go through all the pointers in a TCCompressionInfo, testing
	 * each one to see if it pointed to something that moved.
	 * We do the comparison by subtracting the old address from
	 * the pointer, treating the result as an unsigned number.  If
	 * the pointer was less than the old address (the old address
	 * always points to the beginning of the moved block), we
	 * don't want to touch the pointer.  In this case, the difference
	 * will be a very large unsigned number.  If the pointer was
	 * past the end of the block, the difference will be larger than
	 * blkLength.  In either case, the difference is larger than
	 * blkLength, so that's the test.
	 */
    addrDiff = ((UInt32)(pTextCompInfo->pSymIndex) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->pSymIndex = (TCSymbolArray *)
				   (addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pTextCompInfo->pSymTable) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->pSymTable = (TCSymbolTable *)
				   (addrDiff + (UInt32)pNewAddr);

	/* look at ptr to first big symbol */
    addrDiff = ((UInt32)(pTextCompInfo->pBigSymbols) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->pBigSymbols = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	/* look at ptr to first and last symbols */
    addrDiff = ((UInt32)(pTextCompInfo->p1stSymbol) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->p1stSymbol = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pTextCompInfo->pLastSymbol) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->pLastSymbol = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pTextCompInfo->p1stDictSymbol) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->p1stDictSymbol =
				(TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	/* We don't need to bother with the pointer to the position buffer,
	 * it's reloaded on every call anyway.
	 */

	/* See if either of the arrays of pointers to vertical runs has moved */
    addrDiff = ((UInt32)(pTextCompInfo->pVertRunHeadArray) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->pVertRunHeadArray = (TCScanLineArray *)
						(addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pTextCompInfo->pVertRunTailArray) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->pVertRunTailArray = (TCScanLineArray *)
						(addrDiff + (UInt32)pNewAddr);

	/* The allocator and freer don't move, so we don't have to check
	 * them.  The pClassifiedSymbol is strictly a temp that's reloaded
	 * immediately before it's used, so we don't have to check it
	 * either.
	 * So, the next set of things are the pointers to the big heaps. */
    addrDiff = ((UInt32)(pTextCompInfo->p1stScanLineBlock) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->p1stScanLineBlock = (TCBlockHeader *)
						(addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pTextCompInfo->p1stGroupBlock) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->p1stGroupBlock = (TCBlockHeader *)
						(addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pTextCompInfo->p1stSymbolBlock) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->p1stSymbolBlock = (TCBlockHeader *)
						(addrDiff + (UInt32)pNewAddr);

	/* Look at the bitmap too. */
    addrDiff = ((UInt32)(pTextCompInfo->pBitmapBlock) - oldAddr);
    if (addrDiff < blkLength)
	pTextCompInfo->pBitmapBlock = (UInt8 *)(addrDiff + (UInt32)pNewAddr);

	/* That's it, we win! */
    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_fixupSymbolArrayPtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a TCSymbolArray block.
 *
 *	Input:
 *		pThisBlock:	address of the TCCompressionInfoBlock.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in a TCSymbolArray block to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupSymbolArrayPtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
    PROLOGUE;
    TCSymbolArray	*pSymArray;
    UInt32	row, col;
    UInt32	oldAddr;
    UInt32	addrDiff;

	/* correct the type of the block pointer */
    pSymArray = (TCSymbolArray *) pThisBlock;

	/* prepare for the comparisions */
    oldAddr = (UInt32) pOldAddr;

	/* loop through the array, inspecting each pointer */
    for (row = 0; row <= cMaxSymIndexDim; row++)
    {
	for (col = 0; col <= cMaxSymIndexDim; col++)
	{
	    addrDiff = ((UInt32)((pSymArray->array)[row][col]) - oldAddr);
	    if (addrDiff < blkLength)
		(pSymArray->array)[row][col] = (TCSymbol *)
						(addrDiff + (UInt32)pNewAddr);
	}
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_fixupTCSymbolTablePtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a TCSymbolTable block.
 *
 *	Input:
 *		pThisBlock:	address of the TCCompressionInfoBlock.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in a TCSymbolTable block to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupTCSymbolTablePtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
    PROLOGUE;
    TCSymbolTable	*pSymTable;
    Int32	i;
    UInt32	oldAddr;
    UInt32	addrDiff;

	/* correct the type of the block pointer */
    pSymTable = (TCSymbolTable *) pThisBlock;

	/* prepare for the comparisions */
    oldAddr = (UInt32) pOldAddr;

	/* loop through the table, inspecting each pointer */
    for (i = 0; i < pSymTable->numPtrs; i++)
    {
	addrDiff = ((UInt32)((pSymTable->table)[i]) - oldAddr);
	if (addrDiff < blkLength)
	    (pSymTable->table)[i] = (TCSymbol *)
					(addrDiff + (UInt32)pNewAddr);
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_fixupTCScanLineArrayPtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a TCScanLineArray block.
 *
 *	Input:
 *		pThisBlock:	address of the TCCompressionInfoBlock.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in a TCSymbolArray block to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupTCScanLineArrayPtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
    PROLOGUE;
    TCScanLineArray	*pScanLineArray;
    Int32	col;
    UInt32	oldAddr;
    UInt32	addrDiff;

	/* correct the type of the block pointer */
    pScanLineArray = (TCScanLineArray *) pThisBlock;

	/* prepare for the comparisions */
    oldAddr = (UInt32) pOldAddr;

	/* loop through the array, inspecting each pointer */
    for (col = 0; col < pScanLineArray->numPtrs; col++)
    {
	addrDiff = ((UInt32)((pScanLineArray->ptrArray)[col]) - oldAddr);
	if (addrDiff < blkLength)
	    (pScanLineArray->ptrArray)[col] = (TCScanLineRun *)
					    (addrDiff + (UInt32)pNewAddr);
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_fixupTCGroupRunPtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a heap memory block
 *	containing TCGroupRuns.
 *
 *	Input:
 *		pThisBlock:	address of the TCCompressionInfoBlock.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in every TCGroupRun struct. to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupTCGroupRunPtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
    PROLOGUE;
    TCGroupRun	*pGroup;
    TCBlockHeader	*pBlock;
    UInt32	numOfStructs;
    UInt32	oldAddr;
    UInt32	addrDiff;

	/* The TCGroupRun structures are kept in (string of) memory block(s).
	 * There are two ways to look at each structure, either it's on
	 * the free list or it's in use.  In either case, the first four
	 * bytes are a pointer so can be freely corrected.
	 * The number of structures in the memory block is in the
	 * numOfStructs entry of the TCBlockHeader.
	 */

	/* correct the type of the block pointer */
    pBlock = (TCBlockHeader *) pThisBlock;
    oldAddr = (UInt32) pOldAddr;

	/* Fixup pointers in the TCBlockHeader itself */
    addrDiff = ((UInt32)(pBlock->pNext) - oldAddr);
    if (addrDiff < blkLength)
	pBlock->pNext = (TCBlockHeader *) (addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pBlock->pTop) - oldAddr);
    if (addrDiff < blkLength)
	pBlock->pTop = (void *) (addrDiff + (UInt32)pNewAddr);

	/* get the number of structures in this block */
    numOfStructs = pBlock->numOfStructs;
    pGroup = (TCGroupRun *) ((UInt32)pBlock + sizeof(TCBlockHeader));

	/* loop through the array of structures, inspecting each pointer */
    for (; numOfStructs > 0; numOfStructs--, pGroup++)
    {
	addrDiff = ((UInt32)(pGroup->pNext) - oldAddr);
	if (addrDiff < blkLength)
	    pGroup->pNext = (TCGroupRun *) (addrDiff + (UInt32)pNewAddr);
    }

    NORMAL_RETURN;
}

/* PRIVATE */
/**************************************************************************
 *   ip_fixupTCSymbolPtrs()
 *
 *	This is a standard procedure used to fixup pointers after
 *	garbage collection moves a memory block.
 *	This one works fixes up pointers in a heap memory block
 *	containing TCSymbols.
 *
 *	Input:
 *		pThisBlock:	address of the TCCompressionInfoBlock.
 *		pOldAddr:	former address of block that moved.
 *		pNewAddr:	new address of block that moved.
 *		blkLength:	length of moved block.
 *
 *	Action:
 *		Look at every pointer in every TCSymbols struct. to
 *		see if any of them point to the block that moved.
 *	Returns:
 *		ia_successful		it worked
 *
 **************************************************************************/
Int32  CDECL 
ip_fixupTCSymbolPtrs (
    IPState    *pIPState,
    void       *pThisBlock,
    void       *pOldAddr,
    void       *pNewAddr,
    UInt32	blkLength)
{
    PROLOGUE;
    TCSymbol	*pSymbol;
    TCBlockHeader	*pBlock;
    UInt32	numOfStructs;
    UInt32	oldAddr;
    UInt32	addrDiff;

	/* The TCSymbols structures are kept in (string of) memory block(s).
	 * There are two ways to look at each structure, either it's on
	 * the free list or it's in use.  In either case, the first four
	 * bytes are a pointer so can be freely corrected.
	 * The number of structures in the memory block is in the
	 * numOfStructs entry of the TCBlockHeader.
	 */
	/* correct the type of the block pointer */
    pBlock = (TCBlockHeader *) pThisBlock;
    oldAddr = (UInt32) pOldAddr;

	/* Fixup pointers in the TCBlockHeader itself */
    addrDiff = ((UInt32)(pBlock->pNext) - oldAddr);
    if (addrDiff < blkLength)
	pBlock->pNext = (TCBlockHeader *) (addrDiff + (UInt32)pNewAddr);

    addrDiff = ((UInt32)(pBlock->pTop) - oldAddr);
    if (addrDiff < blkLength)
	pBlock->pTop = (void *) (addrDiff + (UInt32)pNewAddr);

	/* get the number of structures in this block */
    numOfStructs = pBlock->numOfStructs;
    pSymbol = (TCSymbol *) ((UInt32)pBlock + sizeof(TCBlockHeader));

	/* loop through the array of structures, inspecting each pointer */
    for (; numOfStructs > 0; numOfStructs--, pSymbol++)
    {
	addrDiff = ((UInt32)(pSymbol->pNext) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pNext = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pPrev) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pPrev = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pRunHead) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pRunHead = (TCScanLineRun *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pRunTail) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pRunTail = (TCScanLineRun *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pClassSymbol) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pClassSymbol = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pNewSymbol) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pNewSymbol = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pNextNum) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pNextNum = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pPrevNum) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pPrevNum = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pNextDictSym) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pNextDictSym = (TCSymbol *) (addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->p1stHorizGroup) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->p1stHorizGroup = (TCGroupRun *)
						(addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->p1stVertGroup) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->p1stVertGroup = (TCGroupRun *)
						(addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pDupHorizGroup) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pDupHorizGroup = (TCGroupRun *)
						(addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pDupVertGroup) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pDupVertGroup = (TCGroupRun *)
						(addrDiff + (UInt32)pNewAddr);

	addrDiff = ((UInt32)(pSymbol->pBitmap) - oldAddr);
	if (addrDiff < blkLength)
	    pSymbol->pBitmap = (UInt8 *) (addrDiff + (UInt32)pNewAddr);

    }

    NORMAL_RETURN;
}





/*

 $Log: mitcptr.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:17  jlin
 Imported sources

 Revision 1.8  1998/01/13 23:53:38  ddavies
 mitcptr.c -> /product/ipcore/ipshared/src/RCS/mitcptr.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.7  1997/04/24 23:34:15  ddavies
 mitcptr.c -> /product/ipcore/ipshared/src/RCS/mitcptr.c,v

 Add new routine ip_fixupTCSymbolPtrs.  This fixes the pointers inside
 a TCSymbol during garbage collection.

 Revision 1.6  1996/07/27 02:03:46  ddavies
 mitcptr.c -> /product/ipcore/ipshared/src/RCS/mitcptr.c,v

 Add fixup for new p1stDictSymbol pointer in TCCompressionInfo.

 Revision 1.5  1996/07/19 00:29:34  ddavies
 mitcptr.c -> /product/ipcore/ipshared/src/RCS/mitcptr.c,v

 Add ip_fixupTCSymbolTablePtrs.  This routine fixes the pointers in a
 new table that maps symbol IDs to symbols.

 Revision 1.4  1996/07/04 03:03:45  ddavies
 mitcptr.c -> /product/ipcore/ipshared/src/RCS/mitcptr.c,v

 Check new TCCompressionInfo pointer, pLastSymbol, to see if it should
 be adjusted when the thing it points to moves.

 Revision 1.3  1995/12/13 02:52:38  ejaquith
 mitcptr.c -> /product/ipcore/ipshared/src/RCS/mitcptr.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/07/14  02:31:25  ddavies
 * mitcptr.c -> /product/ipcore/ipshared/src/RCS/mitcptr.c,v
 *
 * Fixed bugs in pointer fixup routines.  They were supposed to fix all pointers
 * in heap memory blocks, but they didn't fix the heap pointers themselves.  They
 * were only fixing the pointers of the structures in the heaps.
 * This applies to ip_fixupTCGroupRunPtrs.
 *
 * Revision 1.1  1995/07/10  23:08:23  ddavies
 * Initial revision
 *

 */
