/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mitctnd.c:	Contains tables used to perform Tundra-format decoding
 *		of symbol-based compressed images.  These tables
 *		are specified in William Rucklidge's "Xerox Proposal
 *		for JBIG-2 Coding"
 *
 *		ip_decodeTundraInteger()
 *			Decode 32-bit integer using Table 3 from spec
 *
 *		ip_decodeTundraBigVal()
 *			Decode using one of the unbounded classes.
 *
 *		ip_putValIntoTundraDecodeTable()
 *			Initialize a portion of a Huffman decoding table.
 *
 *		firstXDecTable0, firstXDecTable1
 *			First X decoding tables 0 and 1.
 *
 *		deltaXDecTable0, deltaXDecTable1, deltaXDecTable2
 *			Delta X decoding tables 0, 1 and 2
 *
 *		deltaYDecTable0, deltaYDecTable1, deltaYDecTable2
 *			Delta Y decoding tables 0, 1 and 2
 *
 *		widthHeightDecTable0
 *			Symbol Width/Height decoding table 0
 *
 *		useDecTable0, useDecTable1
 *			Symbol Use Count decoding tables 0 and 1
 *
 * These tables are global variables, but they're read only.  This
 * makes it possible for reentrant routines to use them.
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "shrcmprs.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mitctnd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* This file contains global tables and procedures used to perform
 * Tundra decoding of symbol-based compressed image information.  There
 * are also some tables used for both encoding and decoding.  They are
 * in mitctnt.c.
 *
 * The encoding information is comprised of a symbol dictionary
 * (pictures of the symbols), position information for each page
 * (where the symbols go on the page) and a residual image for each
 * page (stuff that only appeared once).
 *
 * This file contains global tables that are used to encode these files.
 * Yes, these are global variables.  However, they are read only, so
 * reentrant routines can use them.  Making them constant global tables means
 * we can use pointers with impunity.  They become part of the code
 * and code doesn't get garbage collected!
 */


/* PRIVATE */
/**************************************************************************
 *   ip_decodeTundraInteger()
 *
 *	This procedure is called to decode a number that was encoded
 *	using Table 3 as specified in William Rucklidge's "Xerox Proposal for
 *	JBIG-2 Coding."  This type of number doesn't appear very
 *	often, so calling a procedure is ok.
 *	This type of encoding is strange in that it can take as
 *	input very long numbers and put out very long numbers.
 *	This means we must be careful about emptying out the
 *	accumulator.
 *
 *	Input:
 *		pAccum:		ptr to the current set of encoded bits
 *		pVacantBits:	ptr to the number of vacant bits on the right
 *				end of accum.
 *		ppInputBuf:	Ptr to ptr to input byte buffer.
 *		pInputEnd:	Ptr to last byte of data buffer.
 *		pVal:		The number that was decoded.
 *
 *	Action:
 *		Decodes the value given.  If we run out of bytes in
 *		the input buffer, returns ia_bufferExceeded.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_bufferExceeded	ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_decodeTundraInteger (
    IPState    *pIPState,
    UInt32     *pAccum,
    Int32      *pVacantBits,
    UInt8     **ppInputBuf,
    UInt8      *pInputEnd,
    UInt32     *pVal)
{
    PROLOGUE;
    UInt32	 accum = *pAccum;
    Int32	 vacantBits = *pVacantBits;
    UInt8	*pInputBuf = *ppInputBuf;

	/* See which of the four cases applies.  But first, we have
	 * to have at least 8 bits to decode the header + min value.
	 * Make sure we have at least 16 bits and try for 24.
	 */
    if (vacantBits >= 16)
    {
	    /* We know there are at least 16 vacant bits,  If
	     * there are 24 and there's enough data, get 3 bytes.
	     * BUG FIX: If you can't fetch anything, go with
	     * what you have instead of dying (removed
	     * "else return ia_bufferExceeded at end of "if" chain).
	     */
	if ((vacantBits >= 24) && (pInputBuf < pInputEnd - 1))
	{
	    vacantBits -= 8;
	    accum |= (UInt32)pInputBuf[0] << vacantBits;
	    vacantBits -= 8;
	    accum |= (UInt32)pInputBuf[1] << vacantBits;
	    vacantBits -= 8;
	    accum |= (UInt32)pInputBuf[2] << vacantBits;
	    pInputBuf += 3;
	}
	else
	if (pInputBuf < pInputEnd)	/* get 2 bytes */
	{
	    vacantBits -= 8;
	    accum |= (UInt32)pInputBuf[0] << vacantBits;
	    vacantBits -= 8;
	    accum |= (UInt32)pInputBuf[1] << vacantBits;
	    pInputBuf += 2;
	}
	else
	if (pInputBuf == pInputEnd)	/* get 1 byte */
	{
	    vacantBits -= 8;
	    accum |= (UInt32)pInputBuf[0] << vacantBits;
	    pInputBuf += 1;
	}
    }

    if ((accum & 0x80000000) == 0)
    {
	    /* value is in next 7 bits */
	*pVal = accum >> (32 - 8);
	accum <<= 8;
	vacantBits += 8;
    }
    else
    if ((accum & 0xC0000000) == 0x80000000)	/* val - 128 in next 10 bits */
    {
	    /* need at least 12 bits to decode this.  If we don't
	     * have 12 bits now, there must not be 12 bits left
	     * in the buffer since the code above tried to get 16
	     * bits if they were there.  So if we don't have
	     * 12 bits, we're dead.
	     */
	if (vacantBits > (32 - 12))
	    return ia_bufferExceeded;
	*pVal = ((accum >> (32 - 12)) & 0x3FF) + 128;
	accum <<= 12;
	vacantBits += 12;
    }
    else
    if ((accum & 0xFE000000) == 0xC0000000)	/* 32-bit val follows
						 * 7-bit hdr */
    {
	    /* We need at least 39 bits to decode this.  if the
	     * sum of (32-vacantBits) and (pInputEnd-pInputBuf+1)*8
	     * is less than 32, we're done.
	     */
	if (((32-vacantBits) + (pInputEnd - pInputBuf + 1) * 8) < 39)
	    return ia_bufferExceeded;

	    /* Shift out the header and fill the rest of the input buffer. */
	accum <<= 7;
	vacantBits += 7;
	while (vacantBits >= 8)
	{
	    vacantBits -= 8;
	    accum |= (*pInputBuf++) << vacantBits;
	}
	    /* Insert as much of the next byte as fits */
	if (vacantBits > 0)
	    accum |= (*pInputBuf) >> (8-vacantBits);

	    /* The 32 bits in accum is the value */
	*pVal = accum;
	vacantBits += 24;
	if (pInputBuf <= pInputEnd)
	    accum = (*pInputBuf++) << vacantBits;
	else
		/* the position buffer can't end on a byte boundary
		 * after encoding a really big value, so we must
		 * need stuff after this, so complain.
		 */
	    return ia_bufferExceeded;
    }
    else	/* 15 big value follows 2 bit header */
    {
	    /* We need at least 17 bits to decode this.  If we don't
	     * have at least 17 bits now, there aren't 17 bits left,
	     * so punt.
	     */
	if (vacantBits > (32 - 17))
	    return ia_bufferExceeded;

	*pVal = (accum >> (32 - 17)) & 0x7FFF;
	accum <<= 17;
	vacantBits += 17;
    }

	/* update the passed in values */
    *pAccum = accum;
    *pVacantBits = vacantBits;
    *ppInputBuf = pInputBuf;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_decodeTundraBigVal()
 *
 *	The Tundra decoding tables normally specify a class that includes
 *	positive or negative infinity.  This procedure encodes these
 *	very large values.  It isn't called very often, so it's ok
 *	that it's a procedure.
 *
 *	Input:
 *		pDecTable:	ptr to decoding table
 *		pAccum:		ptr to the current set of unwritten bits
 *		pVacantBits:	ptr to the number of vacant bits on the right
 *				end of accum.
 *		ppInputBuf:	Ptr to ptr to input byte buffer.
 *		pInputEnd:	Ptr to last byte of data buffer.
 *		pVal:		Ptr to the number to be decoded.
 *
 *	Action:
 *		Decodes the value given.  If we run out of bytes in
 *		the output buffer, returns ia_bufferExceeded.
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_bufferExceeded	ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_decodeTundraBigVal (
    IPState    *pIPState,
    TCTundraDecTable *pDecTable,
    UInt32     *pAccum,
    Int32      *pVacantBits,
    UInt8     **ppInputBuf,
    UInt8      *pInputEnd,
    UInt32     *pVal)
{
    PROLOGUE;
    UInt32		 accum = *pAccum;
    Int32		 vacantBits = *pVacantBits;
    UInt8		*pInputBuf = *ppInputBuf;
    TCTundraEncClass	*pEncClass;
    Int32		 headerLen = pDecTable->headerLen;
    Int32		 val;
    

	/* first, get enough bits to decode the header */
    while ((32 - vacantBits) < headerLen)
    {
	if (pInputBuf > pInputEnd)
	    return ia_bufferExceeded;	/* couldn't get the next byte */

	vacantBits -= 8;
	accum |= (*pInputBuf++) << vacantBits;
    }

	/* See if we're supposed to use the min class or the max class */
    if (accum & (1 << (32 - headerLen)))
	pEncClass = &(pDecTable->pClassArray[pDecTable->bigPosNumClass]);
    else	/* use the bit negative class */
	pEncClass = &(pDecTable->pClassArray[0]);

	/* Get rid of the header */
    accum <<= pDecTable->headerLen;
    vacantBits += pDecTable->headerLen;

	/* Go get the big positive integer that follows */
    __error__ = ip_decodeTundraInteger (pIPState, &accum, &vacantBits, &pInputBuf,
				     pInputEnd, (UInt32 *)(&val));
    if (__error__ != ia_successful)
	NORMAL_RETURN;

	/* Convert it into a signed integer.  Remember that the encoded
	 * values in the big negative classes are equal to (offset - val),
	 * so val = offset - encodedVal.
	 * In the positive classes encodedVal = (val - offset), so
	 * val = encodedVal + offset;
	 */
    if (pEncClass == &(pDecTable->pClassArray[0]))
    {
	    /* this is an unbounded negative class */
	*pVal = pEncClass->offset - val;
    }
    else
    {
	    /* This is an unbounded positive class */
	*pVal = pEncClass->offset + val;
    }

	/* It worked, update the vars passed in and return */
    *pAccum = accum;
    *pVacantBits = vacantBits;
    *ppInputBuf = pInputBuf;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_calcSizeOfTundraDecodeTable()
 *
 *	We use an array of linked tables to decode one of the Huffman-
 *	encoded Tundra values.  We use the same style of table for
 *	every type of decoding process.  This routine calculates the
 *	total size of the array of tables in units of TCTundraDecSelectors.
 *
 *	Input:
 *		numItemsArray:	ptr to 32-element array containing number of
 *				decoded items as a function of code length.
 *				For example, numItemsArray[3] contains the
 * 				number of codes of length 3.
 *		pLength:	ptr to the calculated length.
 *
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_bufferExceeded	ran out of memory
 *
 **************************************************************************/
Int32 CDECL 
ip_calcSizeOfTundraDecodeTable (
    IPState    *pIPState,
    Int32     *numItemsArray,
    Int32     *pLength)
{
    PROLOGUE;
    Int32	slotsUsed, slotsAvailable;
    Int32	lowerTables;
    Int32	index;

	/* How does this glorious routine work?  Huffman codes
	 * will be decoded using several levels of tables.
	 * You can think of this structure as a tree.  At the root of
	 * the tree is one 256 entry table.  It is used to
	 * parse the first 8 bits of every code.  Codes whose
	 * lengths are less than or equal to 8 bits take up
	 * 2**(8-codeLength) entries in this table.
	 * For codes that are longer than 8 bits, we have a
	 * pointer in the top level table (at the location specified
	 * by the first 8 bits of the code) to a lower 16-entry table.
	 * These tables are allocated as needed.  Codes that are
	 * longer than 8+4 bits have entries in the second level
	 * tables that point to third level tables and so on.
	 *
	 * This routine calculates the number of 16-entry tables
	 * needed in addition to the top level 256-entry table.
	 *
	 * One generates Huffman codes (theoretically) by dividing
	 * some 2**N space among the possible code
	 * values such that the number of entries corresponding to
	 * a code roughly corresponds to the frequency of code
	 * occurence.  For example, a code that occurs a lot
	 * corresponds to lots of entries in the 2**N space.
	 * For our purposes now, the interesting notion is that
	 * a complete 2**N binary space is divided up.  This is
	 * different than, say, generating a sequence of numbers
	 * until we're done.  The nice property of dividing up
	 * the space is that every entry in a space of 2**(N-m)
	 * entries corresponds to at least 1 code value.
	 *
	 * Therefore, we can compute the number of descendent tables
	 * at every level of the tree by subtracting the number
	 * of table entries taken by code values that terminate
	 * at that level from the total number of table entries
	 * at that level.  You can't just count the number of
	 * code values that don't terminate since they will correspond
	 * to some unknown number of lower level tables.
	 */

	/* Figure out how many slots are left over after accounting
	 * for all the terminating codes in the top level table.
	 */
    slotsUsed = numItemsArray[1] * 128 +
		numItemsArray[2] * 64 +
		numItemsArray[3] * 32 +
		numItemsArray[4] * 16 +
		numItemsArray[5] * 8 +
		numItemsArray[6] * 4 +
		numItemsArray[7] * 2 +
		numItemsArray[8] * 1;

    slotsAvailable = 256 - slotsUsed;
    lowerTables = slotsAvailable;

	/* figure it out for most of the lower levels */
    for (index = 9, slotsAvailable = lowerTables;
	 (index < 29) && (slotsAvailable != 0);
	 index += 4)
    {
	slotsAvailable = slotsAvailable * 16;	/* on the new level */
	slotsUsed = numItemsArray[index] * 8 +
		    numItemsArray[index+1] * 4 +
		    numItemsArray[index+2] * 2 +
		    numItemsArray[index+3] * 1;

       slotsAvailable = slotsAvailable - slotsUsed;
       lowerTables += slotsAvailable;
    }

	/* do the last level if necessary.  We can't do it in
	 * the regular loop since we're not allowed to read
	 * numItemsArray[32].
	 */
    if (slotsAvailable != 0)
    {
	slotsAvailable = slotsAvailable * 16;
	slotsUsed = numItemsArray[29] * 8 +
		    numItemsArray[30] * 4 +
		    numItemsArray[31] * 2;

       slotsAvailable = slotsAvailable - slotsUsed;
       lowerTables += slotsAvailable;
    }

	/* We must be happy.  Add the number of entries in the
	 * lower tables to the number in the root table and
	 * return that.
	 * BUG FIX: if numItemsArray was empty (there are no symbols),
	 * then lowerTables will be huge (it thinks all the codes are
	 * very long since none of them were short) and slotsUsed will
	 * be 0.  In this case, return a length of 4 since we don't
	 * need the tables.  We don't return a length of 0 since some
	 * wise allocators return NULL when asked for a buffer of size 0.
	 * Since higher level routines always check for a NULL
	 * pointer after calling an allocator, that would make things
	 * look like there was no memory.  The code that calls this
	 * code should happily allocate and deallocate a 4-byte
	 * thing.
	 */
    if (slotsUsed > 0)
	*pLength = 256 + (lowerTables * 16);
    else
	*pLength = 4;

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *   ip_putValIntoTundraDecodeTable()
 *
 *	We use an array of linked tables to decode one of the Huffman-
 *	encoded Tundra values.  We use the same style of table for
 *	every type of decoding process.  This routine sets the
 *	values and pointers in a table corresponding to one Huffman code.
 *
 *	Input:
 *		pSelArray:	ptr to structure containing the array
 *				of decode selectors.
 *		newCode:	the new code to be inserted
 *		newCodeLen:	length of the new code in bits.
 *		pEntry:		ptr to be inserted in all leaf nodes
 *				corresponding to newCode.
 *
 *
 *	Returns:
 *		ia_successful		things went well
 *		ia_bufferExceeded	pSelArray was too small
 *
 **************************************************************************/
Int32 CDECL 
ip_putValIntoTundraDecodeTable (
    IPState    *pIPState,
    TCTundraDecSelectorArray *pSelArray,
    UInt32	newCode,
    UInt32	newCodeLen,
    void       *pEntry)
{
    PROLOGUE;
    TCTundraDecSelector		*pTable;	/* ptr to current table */
    Int32			 minIndex;	/* index of 1st val
						 * corresponding to newCode
						 * in table */
    Int32			 maxIndex;	/* index of last val
						 * corresponding to newCode
						 * in table */
    void			*newEntry;	/* what we mark each table
						 * entry with */
    Int32			 bitsAtThisLevel; /* number of bits that this
						 * level of table parses */
    Int32			 curCodeLen;	/* number of code bits to
						 * be put into tables */

	/* align the new code at the MS end of its word so that
	 * we can chop pieces off the left end easily.
	 */
    curCodeLen = newCodeLen;
    newCode <<= (32 - curCodeLen);

	/* Loop through the levels of tables, assigning ptrs as we go */
    for (pTable = &(pSelArray->array[0]), bitsAtThisLevel = 8;
	 curCodeLen > 0;
	 pTable = (TCTundraDecSelector *) newEntry,
		newCode <<= bitsAtThisLevel, curCodeLen -= bitsAtThisLevel,
		bitsAtThisLevel = 4)
    {
	minIndex = newCode >> (32 - bitsAtThisLevel);
	if (curCodeLen <= bitsAtThisLevel)
	{
		/* We're at the terminating level for this code */
	    maxIndex = minIndex + (1 << (bitsAtThisLevel - curCodeLen)) - 1;
	    newEntry = pEntry;
	}
	else
	{
		/* We need to point to a lower level table, get the
		 * index we need */
	    maxIndex = minIndex;
	    if (pTable[minIndex].pEntry == NULL)
	    {
		    /* We need to allocate a new lower level table */
		if (pSelArray->nextTable + 16 <= pSelArray->arraySize)
		{
		    newEntry =
		        (void *) &(pSelArray->array[pSelArray->nextTable]);
		    pSelArray->nextTable += 16;
		}
		else	/* ooops */
		    return ia_bufferExceeded;
	    }
	    else	/* use the entry that was there */
		newEntry = pTable[minIndex].pEntry;
	}
    
	    /* mark all the top level table values corresponding to newCode */
	for (; minIndex <= maxIndex; minIndex++)
	{
		/* check to make sure we're not overwriting some other code.
		 * We hope this will become an obsolete check soon (code
		 * gets debugged, right?)
		 */
	    if ((pTable[minIndex].pEntry != NULL) &&
		(pTable[minIndex].pEntry != newEntry))
		return ia_internal;
    
	    pTable[minIndex].pEntry = newEntry;
	    pTable[minIndex].bitLength = newCodeLen;
	}
    }

	/* We're done */
    NORMAL_RETURN;
}



/*********************** First X Decoding Tables *************************/

/* firstXClass0 is in mitctnt.c */

TCTundraDecTable firstXDecTable0 =
{
    cx111011111 << (32-9),	/* headerRoot */
    cx111111111 << (32-9),	/* headerMask */
    10,				/* headerLen */
    13,				/* bigPosNumClass */
    &firstXClass0[0]		/* first encoding class */
};


/* firstXClass1 is in mitctnt.c */

TCTundraDecTable firstXDecTable1 =
{
    cx101111111 << (32-9),	/* headerRoot */
    cx111111111 << (32-9),	/* headerMask */
    10,				/* headerLen */
    14,				/* bigPosNumClass */
    &firstXClass1[0]		/* first encoding class */
};


TCTundraDecTable *firstXDecTableArray[4] =
{
    &firstXDecTable0,
    &firstXDecTable1,
    NULL,
    NULL
};

/* deltaXClass0 is in mitctnt.c */

TCTundraDecTable deltaXDecTable0 =
{
    cx1111100111 << (32-10),	/* headerRoot */
    cx1111111111 << (32-10),	/* headerMask */
    11,				/* headerLen */
    19,				/* bigPosNumClass */
    &deltaXClass0[0]		/* first encoding class */
};


/* deltaXClass1 is in mitctnt.c */

TCTundraDecTable deltaXDecTable1 =
{
    cx1111100111 << (32-10),	/* headerRoot */
    cx1111111111 << (32-10),	/* headerMask */
    11,				/* headerLen */
    20,				/* bigPosNumClass */
    &deltaXClass1[0]		/* first encoding class */
};


/* deltaXClass2 is in mitctnt.c */

TCTundraDecTable deltaXDecTable2 =
{
    cx110110111 << (32-9),	/* headerRoot */
    cx111111111 << (32-9),	/* headerMask */
    10,				/* headerLen */
    19,				/* bigPosNumClass */
    &deltaXClass2[0]		/* first encoding class */
};

TCTundraDecTable *deltaXDecTableArray[4] =
{
    &deltaXDecTable0,
    &deltaXDecTable1,
    &deltaXDecTable2,
    NULL
};


/* Here are the deltaY classes.  Note that we specify the entire header
 * in the headerRoot and headerMask sections.  That's because we still
 * decide to process an unbounded class if (accum & headerMask) == headerRoot.
 * In the classes with both positive and negative unbounded classes,
 * the header root held all but the LS bit of the two headers since all
 * those bits distinguished these headers from all others.  In the
 * deltaY classes, you need to look at all the header bits.
 */

/* deltaYClass0 is in mitctnt.c */

TCTundraDecTable deltaYDecTable0 =
{
    cx1111111 << (32-7),	/* headerRoot */
    cx1111111 << (32-7),	/* headerMask */
    7,				/* headerLen */
    12,				/* bigPosNumClass */
    &deltaYClass0[0]		/* first encoding class */
};


/* deltaYClass1 is in mitctnt.c */

TCTundraDecTable deltaYDecTable1 =
{
    cx11111111 << (32-8),	/* headerRoot */
    cx11111111 << (32-8),	/* headerMask */
    8,				/* headerLen */
    12,				/* bigPosNumClass */
    &deltaYClass1[0]		/* first encoding class */
};


/* deltaYClass2 is in mitctnt.c */

TCTundraDecTable deltaYDecTable2 =
{
    cx101111 << (32-6),		/* headerRoot */
    cx111111 << (32-6),		/* headerMask */
    6,				/* headerLen */
    12,				/* bigPosNumClass */
    &deltaYClass2[0]		/* first encoding class */
};


TCTundraDecTable *deltaYDecTableArray[4] =
{
    &deltaYDecTable0,
    &deltaYDecTable1,
    &deltaYDecTable2,
    NULL
};



/* Here are the width/height classes.  Note that we specify the entire header
 * in the headerRoot and headerMask sections.  That's because we
 * decide to process an unbounded class if (accum & headerMask) == headerRoot..
 */

/* widthHeightClass0 is in mitctnt.c */

TCTundraDecTable widthHeightDecTable0 =
{
    cx1111 << (32-4),		/* headerRoot */
    cx1111 << (32-4),		/* headerMask */
    4,				/* headerLen */
    4,				/* bigPosNumClass */
    &widthHeightClass0[0]	/* first encoding class */
};


/* Here are the symbol use count classes.  Note that we specify
 * the entire header in the headerRoot and headerMask sections.
 * That's because we decide to process an unbounded class
 * if (accum & headerMask) == headerRoot.
 */

/* useClass0 is in mitctnt.c */

TCTundraDecTable useDecTable0 =
{
    cx111 << (32-3),		/* headerRoot */
    cx111 << (32-3),		/* headerMask */
    3,				/* headerLen */
    4,				/* bigPosNumClass */
    &useClass0[0]		/* first encoding class */
};

/* useClass1 is in mitctnt.c */

TCTundraDecTable useDecTable1 =
{
    cx11111 << (32-5),		/* headerRoot */
    cx11111 << (32-5),		/* headerMask */
    5,				/* headerLen */
    8,				/* bigPosNumClass */
    &useClass1[0]		/* first encoding class */
};

TCTundraDecTable *useDecTableArray[2] =
{
    &useDecTable0,
    &useDecTable1
};


/*

 $Log: mitctnd.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:17  jlin
 Imported sources

 Revision 1.5  1998/01/13 23:53:39  ddavies
 mitctnd.c -> /product/ipcore/ipshared/src/RCS/mitctnd.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1997/08/27 02:43:57  ddavies
 mitctnd.c -> /product/ipcore/ipshared/src/RCS/mitctnd.c,v

 Have ip_decodeTundraInteger carry on to the end of the compressed data
 stream instead of giving up if it has room for data and can't load any.

 Revision 1.3  1997/06/25 03:53:54  ddavies
 mitctnd.c -> /product/ipcore/ipshared/src/RCS/mitctnd.c,v

 Fixed a bug in the routine that generates the size of the tables used to
 decode symbol IDs in position blocks.  If there are no symbols at all,
 it says the table should be 4 bytes (a very small non-zero number) instead
 of a huge number.  The number 4 was chosen so that when higher level code
 allocates a table using the size returned, it won't get a NULL pointer.
 The NULL pointer would be checked and would be used to return an
 out-of-memory error.

 Revision 1.2  1996/08/29 01:09:56  ddavies
 mitctnd.c -> /product/ipcore/ipshared/src/RCS/mitctnd.c,v

 Add tables used to decode Tundra dictionaries.

 Revision 1.1  1996/08/22 01:28:08  ddavies
 Initial revision


 */


