/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	
 *  mitctnt.c:	Contains tables and procedures used for both
 *		Tundra-format encoding and decoding of symbol-based
 *		compressed images.  The tables are specified in
 *		William Rucklidge's "Xerox Proposal for JBIG-2 Coding"
 *
 *		Int32	ip_genSymbolHuffmanCodes();
 *			Assign huffman codes to symbols in dictionary.
 *
 *		firstXClass0, firstXClass1
 *			First X encoding classes 0 and 1.
 *
 *		deltaXClass0, deltaXClass1, deltaXClass2
 *			Delta X encoding classes 0, 1 and 2
 *
 *		deltaYClass0, deltaYClass1, deltaYClass2
 *			Delta Y encoding classes 0, 1 and 2
 *
 *		widthHeightClass0
 *			Symbol Width/Height encoding class
 *
 *		useClass0, useClass1
 *			Symbol use count encoding classes
 *
 * These tables are global variables, but they're read only.  This
 * makes it possible for reentrant routines to use them.
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "shrcmprs.prv"
#include "memory.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: mitctnt.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* PRIVATE */
/**************************************************************************
 *   ip_genSymbolHuffmanCodes()
 *
 *	This procedure generates a set of Huffman codes for all the
 *	symbols in the dictionary.
 *
 *	NOTE: for the time being, there is no provision to encode the
 *	page using only a subset of the dictionaries in the internal
 *	compression info.  We assume symbols from all dictionaries can
 *	be used.
 *
 *	Input:
 *		pTextCompInfo:		pointer to text compression info
 *		numItemsArray:		ptr to array of 32 Int32s.  If
 *					non-NULL, the array will be filled
 *					in with the number of symbols 
 *					having each possible code length.
 *					For example, numItemsArray[3] 
 *					will contain the number of symbols 
 *					having Huffman codes of length 3. 
 *		pageRequest:		Page ID of single-count items to add,
 *					or 0
 *		preserveReadOrder:	preserve the read order of symbols
 *
 *	Returns:
 *		ia_successful		it worked
 *		ia_nomem		no memory for internal tables.
 *
 **************************************************************************/
Int32  CDECL 
ip_genSymbolHuffmanCodes (
    IPState    *pIPState,
    TCCompressionInfo *pTextCompInfo,
    Int32      *numItemsArray,
    UInt32	pageRequest,
    Bool	preserveReadOrder)
{
    PROLOGUE;
    TCSymbol		*pSymbol;		/* scratch symbol ptr */
    Int32		 numSymbols;		/* the number of non-residual
						 * symbols. */
    Int32		 temp;
    HuffTreeNode	*huffTreeArray = NULL;	/* array of HuffTreeNodes used
						 * to assign codes to symbols.
						 */
    TCSymbol	       **symPtrArray = NULL;	/* ptr to array of symbol
						 * ptrs. used by qsort */
    UInt32		 firstCodeArray[32];	/* array of values of first
						 * Huffman codes indexed by
						 * code length */
    

	/* We need to generate Huffman codes for all the symbols.
	 * We do this each time a page is encoded since in the
	 * future, we'll be encoding each page relative to a new
	 * set of dictionaries.  The Huffman code generating procedure
	 * takes as input a bunch of symbols in order of instanceCount
	 * and dictionary order (primary and secondary keys).
	 * So, first, establish the dictionary order.  It will be
	 * needed in any case.
	 */
    if (!(pTextCompInfo->encodeState & cTCSymsAreInDictOrder))
    {
	__error__ = ip_sortTCSymbolsInTundraDictOrder (pIPState,
						       pTextCompInfo,
						       preserveReadOrder);
	if (__error__ != ia_successful)
	    goto cleanupGenSymHuffCodes;
    }

#ifdef DDDEBUG
    printf("Dict order:\n");
    for (pSymbol = pTextCompInfo->p1stDictSymbol; pSymbol != NULL;
    pSymbol = pSymbol->pNextDictSym)
    {
	printf("%3d (%3d): dict %d, %dx%d, count %d, code %d-%d\n",
	    pSymbol->dictOrder, pSymbol->symbolID, pSymbol->dictNum, pSymbol->width, pSymbol->height,
	    pSymbol->instanceCount, pSymbol->huffmanCode, pSymbol->huffmanCodeLen);
    }
#endif
	/* Allocate the array of symbols for which we're going to
	 * assign Huffman codes.  We first have to count the
	 * number of symbols with instanceCount > 1 if we're using
	 * a residual image or just all the symbols if there is
	 * no residual.  Note that we're supposed to supply an array
	 * of 2N nodes.  The remainder are used to build the Huffman
	 * tree that's used to assign code lengths.
	 */
    if (pTextCompInfo->useResidual)
    {
	if (pageRequest == 0)
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol, numSymbols = 0;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextNum)
	    {   
		if (pSymbol->instanceCount > 1)
		    numSymbols++;
	    }
	}
	else
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol, numSymbols = 0;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextNum)
	    {   
		if (pSymbol->pageID == 0 || pSymbol->pageID == pageRequest)
		    numSymbols++;
	    }
	}
    }
    else
    {
	if (pageRequest == 0)
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol, numSymbols = 0;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextNum)
	    {   
		if (pSymbol->pageID == 0)
		    numSymbols++;
	    }
	}
	else
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol, numSymbols = 0;
		pSymbol != NULL;
		pSymbol = pSymbol->pNextNum)
	    {   
		if (pSymbol->pageID == 0 || pSymbol->pageID == pageRequest)
		    numSymbols++;
	    }
	}
    }

	/* If there are no non-residual symbols, i.e. nothing to put
	 * in the dictionary, i.e. no symbols for which Huffman codes
	 * should be generated, quit after initializing numItemsArray.
	 */
    if (numSymbols == 0)
    {
	if (numItemsArray != NULL)
	    memset( numItemsArray, 0, 32 * sizeof(UInt32));
	NORMAL_RETURN;
    }

	/* Otherwise, allocate the array of HuffTreeNodes */
    huffTreeArray = (HuffTreeNode *)MALLOC((2*numSymbols + 1) *
					   sizeof(HuffTreeNode));
    if (huffTreeArray == NULL)
    {
	__error__ = ia_nomem;
	goto cleanupGenSymHuffCodes;
    }

	/* Now, if the symbols are already sorted in order of
	 * both count and dictionary order (all symbols are in
	 * several lists, this is one of them), just fill in the primary
	 * and secondary keys with the instanceCount and dictOrder
	 * entries of the symbols.  Note that we follow the
	 * order backwards since the symbols with the smallest
	 * instance counts (and smallest dictionary numbers)
	 * are supposed to be at the beginning of the array.
	 */
    if (pTextCompInfo->encodeState & cTCSymsAreInCountAndDictOrder)
    {
	if (pTextCompInfo->useResidual)
	{
	    for (pSymbol = pTextCompInfo->pLastSymbol, temp = 0;
		 pSymbol != NULL;
		 pSymbol = pSymbol->pPrevNum)
	    {
		if (pSymbol->instanceCount > 1)
		{
		    huffTreeArray[temp].primaryKey = pSymbol->instanceCount;
		    huffTreeArray[temp].secondaryKey = pSymbol->dictOrder;
		    temp++;
		}
	    }
	}
	else	/* there's no residual image, so use all the symbols */
	{
	    for (pSymbol = pTextCompInfo->pLastSymbol, temp = 0;
		 pSymbol != NULL;
		 pSymbol = pSymbol->pPrevNum)
	    {
		if (pSymbol->pageID == 0 || pSymbol->pageID == pageRequest)
		{
		    huffTreeArray[temp].primaryKey = pSymbol->instanceCount;
		    huffTreeArray[temp].secondaryKey = pSymbol->dictOrder;
#ifdef DDDEBUG
		    printf("ID %d has count %d\n", pSymbol->dictOrder, pSymbol->instanceCount);
#endif
		    temp++;
		 }
	    }
	}
    }
    else	/* the symbol need to be sorted.  Allocate an array
		 * of symbol ptrs and sort them */
    {
	symPtrArray = (TCSymbol **) MALLOC((numSymbols) * sizeof(TCSymbol **));
	if (symPtrArray == NULL)
	{
	    __error__ = ia_nomem;
	    goto cleanupGenSymHuffCodes;
	}

	if (pTextCompInfo->useResidual)
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol, temp = 0;
		 pSymbol != NULL;
		 pSymbol = pSymbol->pNextNum)
	    {
		if (pSymbol->instanceCount > 1)
		{
		    if (temp >= numSymbols)
		    {
			__error__ = ia_internal;
			goto cleanupGenSymHuffCodes;
		    }
		    symPtrArray[temp] = pSymbol;
		    temp++;
		}
	    }
	}
	else	/* there's no residual image, so use all the symbols */
	{
	    for (pSymbol = pTextCompInfo->p1stSymbol, temp = 0;
		 pSymbol != NULL;
		 pSymbol = pSymbol->pNextNum)
	    {
		if (pSymbol->pageID == 0 || pSymbol->pageID == pageRequest)
		{
		    if (temp >= numSymbols)
		    {
			__error__ = ia_internal;
			goto cleanupGenSymHuffCodes;
		    }
		    symPtrArray[temp] = pSymbol;
		    temp++;
		}
	    }
	}

	if (temp != numSymbols)
	{
	    __error__ = ia_internal;
	    goto cleanupGenSymHuffCodes;
	}

	    /* Sort the pointers so we can present the symbols in the
	     * right order.  We use the same comparison routine that
	     * i_consolidateShapeInfo does, so we get the same sort.
	     */
	qsort((void *)(&symPtrArray[0]), numSymbols, sizeof(TCSymbol *),
	      ip_compareSymCounts);

	    /* Now copy the primary and secondary keys of the sorted
	     * symbols into the Huffman tree node array.  Note that
	     * the symbols are sorted so that the symbols with big
	     * instanceCounts and big dictionary numbers appear at
	     * the front, so we need to reverse the order when
	     * copying.
	     */
#ifdef DDDEBUG
	printf("Sorted order:\n");
#endif
	for (temp = numSymbols-1; temp >= 0; temp--)
	{
	    huffTreeArray[numSymbols - temp - 1].primaryKey =
		symPtrArray[temp]->instanceCount;
	    huffTreeArray[numSymbols - temp - 1].secondaryKey =
		symPtrArray[temp]->dictOrder;
#ifdef DDDEBUG
	    {
		TCSymbol *pSym = symPtrArray[temp];

		printf("%3d (%3d): dict %d, %dx%d, count %d, code %d-%d\n",
		    pSym->dictOrder, pSym->symbolID, pSym->dictNum, pSym->width,
		    pSym->height, pSym->instanceCount, pSym->huffmanCode,
		    pSym->huffmanCodeLen);
	    }
#endif
	}
    }

	/* The input to the Huffman code generation routine is set up.
	 * Go generate the Huffman codes for the symbols.
	 */
    __error__ = i_genHuffmanCodeLengths (pIPState, huffTreeArray,
				         (2*numSymbols + 1),
				         &firstCodeArray[0],
				         numItemsArray);
    if (__error__ != ia_successful)
	goto cleanupGenSymHuffCodes;

	/* Copy the code lengths back to the symbols.  Note that we
	 * use exactly the same type of loop used to assign primary
	 * and secondary keys.  This maintains synchronization.
	 */
    if (pTextCompInfo->encodeState & cTCSymsAreInCountAndDictOrder)
    {
	if (pTextCompInfo->useResidual)
	{
	    for (pSymbol = pTextCompInfo->pLastSymbol, temp = 0;
		 pSymbol != NULL;
		 pSymbol = pSymbol->pPrevNum)
	    {
		if (pSymbol->instanceCount > 1)
		{
		    pSymbol->huffmanCodeLen = huffTreeArray[temp].codeLength;
		    temp++;
		}
	    }
	}
	else	/* there is no residual image so use all the symbols */
	{
	    for (pSymbol = pTextCompInfo->pLastSymbol, temp = 0;
		 pSymbol != NULL;
		 pSymbol = pSymbol->pPrevNum)
	    {
		pSymbol->huffmanCodeLen = huffTreeArray[temp].codeLength;
		temp++;
	    }
	}
    }
    else	/* the symbol needed to be sorted.  Use the pointer array
		 * to access them. */
    {
	for (temp = 0; temp < numSymbols; temp++)
	{
	    symPtrArray[temp]->huffmanCodeLen =
			huffTreeArray[numSymbols - temp - 1].codeLength;
	}
    }

	/* Now go assign the codes to the symbols.  Note that we do
	 * this in dictionary order, as specified in William Rucklidge's
	 * "Xerox Proposal for JBIG-2 Coding"
	 */
    if (pTextCompInfo->useResidual)
    {
	for (pSymbol = pTextCompInfo->p1stDictSymbol;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextDictSym)
	{
	    if (pSymbol->instanceCount > 1)
	    {
		if (pSymbol->huffmanCodeLen >= 32)
		{
		    __error__ = ia_decodeFailure;
		    goto cleanupGenSymHuffCodes;
		}
		pSymbol->huffmanCode = firstCodeArray[pSymbol->huffmanCodeLen];
		firstCodeArray[pSymbol->huffmanCodeLen]++;
	    }
	}
    }
    else	/* no residual, use all the symbols. */
    {
	for (pSymbol = pTextCompInfo->p1stDictSymbol;
	     pSymbol != NULL;
	     pSymbol = pSymbol->pNextDictSym)
	{
	    if (pSymbol->instanceCount > 1 || pSymbol->pageID == pageRequest)
	    {
		pSymbol->huffmanCode = firstCodeArray[pSymbol->huffmanCodeLen];
		firstCodeArray[pSymbol->huffmanCodeLen]++;
	    }
	}
    }

#ifdef DDEBUG
    {
	TCSymbol *pSym = symPtrArray[temp];

	printf("Sorted order:\n");

	for (pSym = pTextCompInfo->p1stDictSymbol;
	     pSym != NULL;
	     pSym = pSym->pNextDictSym)
	{
	    printf("%3d (%3d): dict %d, %dx%d, count %d, code %d-%d\n",
		pSym->dictOrder, pSym->symbolID, pSym->dictNum, pSym->width,
		pSym->height, pSym->instanceCount, pSym->huffmanCode,
		pSym->huffmanCodeLen);
	}
    }
#endif
    
    __error__ = ia_successful;

cleanupGenSymHuffCodes:
    if (huffTreeArray != NULL)
	FREE(huffTreeArray);
    if (symPtrArray != NULL)
	FREE(symPtrArray);

    ERR_RETURN;
}



/*********************** First X Encoding Class 0 *************************/

TCTundraEncClass firstXClass0[14] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx1110111110,	10,	-2048,	39	},	/* 0 */
    {	cx1010,		4,	-2047,	10	},	/* 1 */
    {	cx1011,		4,	-1023,	9	},	/* 2 */
    {	cx1100,		4,	-511,	8	},	/* 3 */
    {	cx1101,		4,	-255,	7	},	/* 4 */
    {	cx11110,	5,	-127,	6	},	/* 5 */
    {	cx11111,	5,	-63,	5	},	/* 6 */
    {	cx1110,		4,	-31,	5	},	/* 7 */
    {	cx00,		2,	 0,	7	},	/* 8 */
    {	cx010,		3,	 128,	7	},	/* 9 */
    {	cx011,		3,	 256,	8	},	/* 10 */
    {	cx1000,		4,	 512,	9	},	/* 11 */
    {	cx1001,		4,	 1024,	10	},	/* 12 */
    {	cx1110111111,	10,	 2048,	39	}	/* 13 */
};


/*********************** First X Encoding Class 1 **********************/

TCTundraEncClass firstXClass1[15] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx1011111110,	10,	-1024,	39	},	/* 0 */
    {	cx000,		3,	-1023,	9	},	/* 1 */
    {	cx001,		3,	-511,	8	},	/* 2 */
    {	cx1010,		4,	-255,	7	},	/* 3 */
    {	cx11100,	5,	-127,	6	},	/* 4 */
    {	cx11101,	5,	-63,	5	},	/* 5 */
    {	cx1011,		4,	-31,	5	},	/* 6 */
    {	cx1100,		4,	 0,	5	},	/* 7 */
    {	cx11110,	5,	 32,	5	},	/* 8 */
    {	cx11111,	5,	 64,	6	},	/* 9 */
    {	cx1101,		4,	 128,	7	},	/* 10 */
    {	cx010,		3,	 256,	8	},	/* 11 */
    {	cx011,		3,	 512,	9	},	/* 12 */
    {	cx100,		3,	 1024,	10	},	/* 13 */
    {	cx1011111111,	10,	 2048,	39	}	/* 14 */
};


/*********************** Delta X Encoding Class 0 **********************/

TCTundraEncClass deltaXClass0[20] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx11111001110,	11,	-15,	39	},	/* 0 */
    {	cx1111100,	7,	-14,	3	},	/* 1 */
    {	cx111111110,	9,	-7,	1	},	/* 2 */
    {	cx11111110,	8,	-5,	1	},	/* 3 */
    {	cx111111111,	9,	-3,	0	},	/* 4 */
    {	cx1111101,	7,	-2,	0	},	/* 5 */
    {	cx1010,		4,	-1,	0	},	/* 6 */
    {	cx01,		2,	 0,	1	},	/* 7 */
    {	cx11010,	5,	 2,	0	},	/* 8 */
    {	cx111010,	6,	 3,	0	},	/* 9 */
    {	cx100,		3,	 4,	4	},	/* 10 */
    {	cx111011,	6,	 20,	1	},	/* 11 */
    {	cx1011,		4,	 22,	4	},	/* 12 */
    {	cx1100,		4,	 38,	5	},	/* 13 */
    {	cx11011,	5,	 70,	6	},	/* 14 */
    {	cx11100,	5,	 134,	7	},	/* 15 */
    {	cx111100,	6,	 262,	7	},	/* 16 */
    {	cx1111110,	7,	 390,	8	},	/* 17 */
    {	cx111101,	6,	 646,	10	},	/* 18 */
    {	cx11111001111,	11,	 1670,	39	}	/* 19 */
};


/*********************** Delta X Encoding Class 1 **********************/

TCTundraEncClass deltaXClass1[21] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx11111001110,	11,	-30,	39	},	/* 0 */
    {	cx1111100,	7,	-29,	4	},	/* 1 */
    {	cx111111110,	9,	-15,	2	},	/* 2 */
    {	cx11111110,	8,	-11,	2	},	/* 3 */
    {	cx111111111,	9,	-7,	1	},	/* 4 */
    {	cx1111101,	7,	-5,	1	},	/* 5 */
    {	cx1010,		4,	-3,	1	},	/* 6 */
    {	cx010,		3,	-1,	1	},	/* 7 */
    {	cx011,		3,	 1,	1	},	/* 8 */
    {	cx11010,	5,	 3,	1	},	/* 9 */
    {	cx111010,	6,	 5,	1	},	/* 10 */
    {	cx100,		3,	 7,	5	},	/* 11 */
    {	cx111011,	6,	 39,	2	},	/* 12 */
    {	cx1011,		4,	 43,	5	},	/* 13 */
    {	cx1100,		4,	 75,	6	},	/* 14 */
    {	cx11011,	5,	 139,	7	},	/* 15 */
    {	cx11100,	5,	 267,	8	},	/* 16 */
    {	cx111100,	6,	 523,	8	},	/* 17 */
    {	cx1111110,	7,	 779,	9	},	/* 18 */
    {	cx111101,	6,	 1291,	11	},	/* 19 */
    {	cx11111001111,	11,	 3339,	39	}	/* 20 */
};


/*********************** Delta X Encoding Class 2 **********************/

TCTundraEncClass deltaXClass2[20] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx1101101110,	10,	-20,	39	},	/* 0 */
    {	cx110110,	6,	-19,	4	},	/* 1 */
    {	cx11111110,	8,	-5,	0	},	/* 2 */
    {	cx1111100,	7,	-4,	0	},	/* 3 */
    {	cx11000,	5,	-3,	0	},	/* 4 */
    {	cx01,		2,	-2,	2	},	/* 5 */
    {	cx11001,	5,	 2,	0	},	/* 6 */
    {	cx110111,	6,	 3,	0	},	/* 7 */
    {	cx1111101,	7,	 4,	0	},	/* 8 */
    {	cx11111111,	8,	 5,	0	},	/* 9 */
    {	cx10,		2,	 6,	6	},	/* 10 */
    {	cx11010,	5,	 70,	5	},	/* 11 */
    {	cx111000,	6,	 102,	5	},	/* 12 */
    {	cx111001,	6,	 134,	6	},	/* 13 */
    {	cx111010,	6,	 198,	7	},	/* 14 */
    {	cx111011,	6,	 326,	8	},	/* 15 */
    {	cx111100,	6,	 582,	9	},	/* 16 */
    {	cx111101,	6,	 1094,	10	},	/* 17 */
    {	cx1111110,	7,	 2118,	11	},	/* 18 */
    {	cx1101101111,	10,	 4166,	39	}	/* 19 */
};


/*********************** Delta Y Encoding Class 0 **********************/

TCTundraEncClass deltaYClass0[13] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx0,		1,	 1,	0	},	/* 0 */
    {	cx10,		2,	 2,	1	},	/* 1 */
    {	cx1100,		4,	 4,	0	},	/* 2 */
    {	cx1101,		4,	 5,	1	},	/* 3 */
    {	cx11100,	5,	 7,	1	},	/* 4 */
    {	cx11101,	5,	 9,	2	},	/* 5 */
    {	cx111100,	6,	 13,	2	},	/* 6 */
    {	cx1111010,	7,	 17,	2	},	/* 7 */
    {	cx1111011,	7,	 21,	3	},	/* 8 */
    {	cx1111100,	7,	 29,	4	},	/* 9 */
    {	cx1111101,	7,	 45,	5	},	/* 10 */
    {	cx1111110,	7,	 77,	6	},	/* 11 */
    {	cx1111111,	7,	 141,	39	}	/* 12 */
};


/*********************** Delta Y Encoding Class 1 **********************/

TCTundraEncClass deltaYClass1[13] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx0,		1,	 1,	0	},	/* 0 */
    {	cx10,		2,	 2,	0	},	/* 1 */
    {	cx110,		3,	 3,	1	},	/* 2 */
    {	cx11100,	5,	 5,	0	},	/* 3 */
    {	cx11101,	5,	 6,	1	},	/* 4 */
    {	cx111100,	6,	 8,	1	},	/* 5 */
    {	cx1111010,	7,	 10,	0	},	/* 6 */
    {	cx1111011,	7,	 11,	1	},	/* 7 */
    {	cx1111100,	7,	 13,	2	},	/* 8 */
    {	cx1111101,	7,	 17,	3	},	/* 9 */
    {	cx1111110,	7,	 25,	4	},	/* 10 */
    {	cx11111110,	8,	 41,	5	},	/* 11 */
    {	cx11111111,	8,	 73,	39	}	/* 12 */
};


/*********************** Delta Y Encoding Class 2 **********************/

TCTundraEncClass deltaYClass2[13] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx0,		1,	 1,	0	},	/* 0 */
    {	cx100,		3,	 2,	0	},	/* 1 */
    {	cx1100,		4,	 3,	0	},	/* 2 */
    {	cx11100,	5,	 4,	0	},	/* 3 */
    {	cx1101,		4,	 5,	1	},	/* 4 */
    {	cx101,		3,	 7,	3	},	/* 5 */
    {	cx111010,	6,	 14,	1	},	/* 6 */
    {	cx111011,	6,	 16,	2	},	/* 7 */
    {	cx111100,	6,	 20,	3	},	/* 8 */
    {	cx111101,	6,	 28,	4	},	/* 9 */
    {	cx111110,	6,	 44,	5	},	/* 10 */
    {	cx111111,	6,	 76,	6	},	/* 11 */
    {	cx101111,	6,	 140,	39	}	/* 12 */
};


/****************** Symbol Width/Height Encoding Class 0 *****************/

TCTundraEncClass widthHeightClass0[5] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx0,		1,	 0,	0	},	/* 0 */
    {	cx10,		2,	 1,	0	},	/* 1 */
    {	cx110,		3,	 2,	0	},	/* 2 */
    {	cx1110,		4,	 3,	3	},	/* 3 */
    {	cx1111,		4,	 10,	39	}	/* 4 */
};
/* The end code for the width height class is 1110111, which would
 * otherwise be the encoding for 10 in class 3.  However, 10 is
 * encoded as part of class 4. */


/* The following table is used when residual images are not
 * included in the file.
 */

/****************** Symbol Use Count Encoding Class 0 *****************/

TCTundraEncClass useClass0[5] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx100,		3,	 0,	0	},	/* 0 */
    {	cx0,		1,	 1,	0	},	/* 1 */
    {	cx101,		3,	 2,	0	},	/* 2 */
    {	cx110,		3,	 3,	5	},	/* 3 */
    {	cx111,		3,	 35,	39	}	/* 4 */
};

/* This table is used when residual images are use, so counts of
 * 0 and 1 are not encoded.
 */

/****************** Symbol Use Count Encoding Class 1 *****************/

TCTundraEncClass useClass1[9] =
{									
     /* bitHeader    bitHdrLen	offset eValLen */
    {	cx0,		1,	 2,	0	},	/* 0 */
    {	cx100,		3,	 3,	0	},	/* 1 */
    {	cx1010,		4,	 4,	0	},	/* 2 */
    {	cx1011,		4,	 5,	1	},	/* 3 */
    {	cx1100,		4,	 7,	2	},	/* 4 */
    {	cx11110,	5,	 11,	2	},	/* 5 */
    {	cx1101,		4,	 15,	4	},	/* 6 */
    {	cx1110,		4,	 31,	6	},	/* 7 */
    {	cx11111,	5,	 95,	39	}	/* 8 */
};


/*
 $Log: mitctnt.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:17  jlin
 Imported sources

 Revision 1.7  1998/08/04 00:17:07  aber
 mitctnt.c -> /product/ipcore/ipshared/src/RCS/mitctnt.c,v

 CommonCheck: Retain the page ID such that the dictionary can be split into a
 shared dictionary, and a single-use dictionary for each page.
 Support multiple dictionaries per page.

 Revision 1.6  1998/06/23 21:45:49  aber
 mitctnt.c -> /product/ipcore/ipshared/src/RCS/mitctnt.c,v

 CommonCheck: Fixes to force Tundra dictionary reading order, esp. for multiple
 dictionaries per page. Some cosmetic comment enhancements. A couple of
 extra error conditions caught.

 Revision 1.5  1998/01/13 23:53:44  ddavies
 mitctnt.c -> /product/ipcore/ipshared/src/RCS/mitctnt.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.4  1997/08/27 02:46:02  ddavies
 mitctnt.c -> /product/ipcore/ipshared/src/RCS/mitctnt.c,v

 Generate Huffman codes with instanceCounts of 1 if pTextCompInfo->useResidual
 is not set.  Ok, try that in better English.  Generate Huffman codes for
 symbols with instanceCounts of 1 if pTextCompInfo->useResidual is not
 set.

 Revision 1.3  1997/06/25 03:55:03  ddavies
 mitctnt.c -> /product/ipcore/ipshared/src/RCS/mitctnt.c,v

 Fix a bug in the routine that generates Huffman codes for symbols.  If
 there are no symbols, it simply clears the contents of numItemsArray and
 returns.  It used to get all confused.

 Revision 1.2  1996/08/29 01:11:04  ddavies
 mitctnt.c -> /product/ipcore/ipshared/src/RCS/mitctnt.c,v

 Add tables used to encode/decode Tundra dictionaries.

 Revision 1.1  1996/08/22 01:39:11  ddavies
 Initial revision

 */

