/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	wchuff.c
 *
 *  Private routine:
 *     ip_derived_tbl()
 *     ip_waveHuffInit()
 *     ip_emit_huffcode()
 *     ip_dc_huffCoding()
 *     ip_ac_huffCoding()
 *     ip_set_huffTable()
 *     ip_clearZerolen()
 *
 */

#include <stdio.h>
#include <math.h>

#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"

#include "wavelib.pub"
#include "wavelib.prv"

IP_RCSINFO(miwchuff_c_RCSInfo, "$RCSfile: miwchuff.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Derive huffman coding table.
 *
 * LongDescription:::
 *  Derive uffman coding table ( copy from jpeg ). 
 *
 * Input:::
 *     bits::      Ptr to input code bit string
 *  huffval::      Ptr to input code value, it will be huffman value
 * 
 * Output:::
 *  huffco::       Ptr to huff code
 *  huffsi::       Ptr to huff code len
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_derived_tbl (
    IPState    *pIPState,
    const UInt8 *bits,
    const UInt8 *huffval,
    UInt16     *huffco,
    Int8       *huffsi)
{
  PROLOGUE;

  Int32 p, i, l, lastp, si;
  Int8 huffsize[257];
  UInt16 huffcode[257];
  UInt16 code;

  /* Figure C.1: make table of Huffman code length for each symbol */
  /* Note that this is in code-length order. */

  p = 0;
  for (l = 1; l <= 16; l++) {
    for (i = 1; i <= (Int32)bits[l]; i++)
      huffsize[p++] = (Int8) l;
  }
  huffsize[p] = 0;
  lastp = p;
  
  /* Figure C.2: generate the codes themselves */
  /* Note that this is in code-length order. */
  
  code = 0;
  si = huffsize[0];
  p = 0;
  while (huffsize[p]) {
    while (((Int32) huffsize[p]) == si) {
      huffcode[p++] = code;
      code++;
    }
    code <<= 1;
    si++;
  }
  
  /* Figure C.3: generate encoding tables */
  /* These are code and size indexed by symbol value */

  /* Set any codeless symbols to have code length 0;
   * this allows emit_bits to detect any attempt to emit such symbols.
   */

  for (p = 0; p < lastp; p++) {
    huffco[huffval[p]] = huffcode[p];
    huffsi[huffval[p]] = huffsize[p];
  }

  NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman coding init.
 *
 * LongDescription:::
 *  Create huffman coding table for DC and AC ( copy from jpeg ) and init.
 * coding coef. 
 *
 * Input:::
 *      pdata::      Ptr to data used in IO callback func
 * outputFunc::      Ptr to callback function
 *  phuffInfo::      Ptr to huffman info structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveHuffInit (
    IPState    *pIPState,
    void       *pdata,
    Int32 (CDECL *outputFunc) (
		    IPState    *pIPState,
		    void       *userData,
		    UInt8      *buf,
		    UInt32	count),
    WHUFFINFO  *phuffInfo)
{
  PROLOGUE;

  static const UInt8 bits_dc_luminance[17] =
    { /* 0-base */ 0, 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 };
  static const UInt8 val_dc_luminance[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
  static const UInt8 bits_dc_chrominance[17] =
    { /* 0-base */ 0, 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 };
  static const UInt8 val_dc_chrominance[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
  static const UInt8 bits_ac_luminance[17] =
    { /* 0-base */ 0, 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d };
  static const UInt8 val_ac_luminance[] =
    { 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
      0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
      0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
      0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
      0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
      0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
      0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
      0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
      0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
      0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
      0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
      0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
      0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
      0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
      0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
      0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
      0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
      0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
      0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
      0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };
  
  static const UInt8 bits_ac_chrominance[17] =
    { /* 0-base */ 0, 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77 };
  static const UInt8 val_ac_chrominance[] =
    { 0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
      0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
      0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
      0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
      0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
      0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
      0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
      0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
      0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
      0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
      0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
      0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
      0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
      0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
      0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
      0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
      0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
      0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
      0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
      0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };

    phuffInfo->pdata = pdata;
    phuffInfo->IOPath = outputFunc;
    phuffInfo->bits_to_go =8;
    phuffInfo->last_dc_val =0;
    phuffInfo->zero_l =0;
    phuffInfo->buffer =0;

    ERR_PASS(ip_derived_tbl (pIPState, bits_dc_luminance, val_dc_luminance,
			 phuffInfo->dc_lum_hufco, phuffInfo->dc_lum_hufsi));
    ERR_PASS(ip_derived_tbl (pIPState, bits_dc_chrominance, val_dc_chrominance,
			 phuffInfo->dc_chr_hufco, phuffInfo->dc_chr_hufsi));
    ERR_PASS(ip_derived_tbl (pIPState, bits_ac_luminance, val_ac_luminance,
			 phuffInfo->ac_lum_hufco, phuffInfo->ac_lum_hufsi));
    ERR_PASS(ip_derived_tbl (pIPState, bits_ac_chrominance, val_ac_chrominance,
			 phuffInfo->ac_chr_hufco, phuffInfo->ac_chr_hufsi));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Emit huffman code.
 *
 * LongDescription:::
 *  After huffman coing, emit the huffman code to output buf.
 *
 * Input:::
 *          k::      Bit length of code 
 *       code::      Output huffman code
 *  phuffInfo::      Ptr to huffman info structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_emit_huffcode (
    IPState    *pIPState,
    Int16	k,
    Int32	code,
    WHUFFINFO  *phuffInfo)
{
  PROLOGUE;

  if( k <= phuffInfo->bits_to_go)
  {
      phuffInfo->bits_to_go -= k;
      phuffInfo->buffer |= (UInt8)(code << phuffInfo->bits_to_go);
  }
  else
  {
      k -= phuffInfo->bits_to_go;
      phuffInfo->buffer |= (UInt8)(code >> k);
      phuffInfo->IOPath(pIPState, phuffInfo->pdata, &phuffInfo->buffer, 1);
      ++ phuffInfo->codeLen;
      if( phuffInfo->buffer == 0xff )
      {
	  phuffInfo->buffer =0;
	  phuffInfo->IOPath(pIPState, phuffInfo->pdata, &phuffInfo->buffer, 1);
	  ++ phuffInfo->codeLen;
      }
      while( k >= 8)
      {
	  k -= 8;
	  phuffInfo->buffer = (UInt8)((code >> k)&0xff);
	  phuffInfo->IOPath(pIPState, phuffInfo->pdata, &phuffInfo->buffer, 1);
	  ++ phuffInfo->codeLen;
	  if( phuffInfo->buffer == 0xff )
	  {
	      phuffInfo->buffer =0;
	      phuffInfo->IOPath(pIPState, phuffInfo->pdata,
				&phuffInfo->buffer, 1);
	      ++ phuffInfo->codeLen;
	  }
      }
      phuffInfo->bits_to_go = 8-k;
      phuffInfo->buffer = (UInt8)((code << phuffInfo->bits_to_go)&0xff);
  }
  NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman coding for DC.
 *
 * LongDescription:::
 *  Create huffman coding on DC components.
 *
 * Input:::
 *     symbol::      Data to be coded 
 *  phuffInfo::      Ptr to huffman info structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_dc_huffCoding (
    IPState    *pIPState,
    Int16	symbol,
    WHUFFINFO  *phuffInfo)		  
{
    PROLOGUE;

    Int32 temp, temp2;
    Int32 nbits; 
    Int32 l;
    UInt16 *phuffco;
    Int8 *phuffsi;

    phuffco = phuffInfo->phuffco;
    phuffsi = phuffInfo->phuffsi;

  /* Encode the DC coefficient difference per section F.1.2.1 */
  
    temp = symbol - phuffInfo->last_dc_val;
    phuffInfo->last_dc_val = symbol;
    
    temp2 = abs(temp);
    nbits =1;
    if(temp2 == 0) nbits =0;
    while((temp2 >>= 1) >0) ++ nbits;
    if(nbits > 11) 
	nbits = 11;
    temp2 = temp;
    
    if (temp < 0) {
	temp = -temp;             /* temp is abs value of input */
    /* For a negative input, want temp2 = bitwise complement of abs(input) */
    /* This code assumes we are on a two's complement machine */
	temp2--;
  /* Mask off any extra bits in code so emit bits doesn't have to */
	temp2 &= (((Int32) 1)<<nbits) - 1; 
    }
    else temp2 &= 0x7ff;

  /* Emit that number of bits of the value, if positive, */
  /* or the complement of its magnitude, if negative. */
    if(nbits<=0)
    {
	temp  = phuffco[nbits];
	l  = (Int32)phuffsi[nbits];
    }
    else
    {
	temp = (phuffco[nbits]<<nbits) | temp2;
	l = nbits + (Int32) phuffsi[nbits];
    }
    ERR_PASS(ip_emit_huffcode (pIPState, (Int16) l, temp, phuffInfo));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman coding for AC.
 *
 * LongDescription:::
 *  Create huffman coding on AC components.
 * coding coef. 
 *
 * Input:::
 *     symbol::      Data to be coded 
 *  phuffInfo::      Ptr to huffman info structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_ac_huffCoding (
    IPState    *pIPState,
    Int16	symbol,
    WHUFFINFO  *phuffInfo)
{
    PROLOGUE;

    Int32 temp, temp2;
    Int32 nbits; 
    Int32 i;
    Int16 l;
    UInt16 *phuffco;
    Int8 *phuffsi;

    phuffco = phuffInfo->phuffco;
    phuffsi = phuffInfo->phuffsi;
  /* Encode the AC coefficients per section F.1.2.2 */
  
    temp = (Int32)symbol;
    if ( temp == 0) {
	phuffInfo->zero_l ++;
    } else {
      /* if run length > 15, must emit special run-length-16 codes (0xF0) */
      while (phuffInfo->zero_l > 15) {
	  ip_emit_huffcode (pIPState, (Int16) phuffsi[0xf0], 
			    (Int32) phuffco[0xf0], phuffInfo);
	  phuffInfo->zero_l -= 16;
      }

      temp2 = abs(temp);
      nbits =1;
      if(temp2 == 0) nbits =0;
      while((temp2 >>= 1) >0) ++ nbits;
      temp2 = temp;
      if( nbits > 10)
	  nbits =10;

      if (temp < 0) {
	temp = (-temp);		/* temp is abs value of input */
    /* For a negative input, want temp2 = bitwise complement of abs(input) */
    /* This code assumes we are on a two's complement machine */
	temp2--;
    /* Mask off any extra bits in code so emit bits doesn't have to */
	temp2 &= (((Int32) 1)<<nbits) - 1; 
      }
      else temp2 &= 0x3ff;

      /* Emit Huffman symbol for run length / number of bits */
      i = (phuffInfo->zero_l << 4) + nbits;
      temp2 |= (phuffco[i] << nbits);
      nbits += (Int32)phuffsi[i];
      l = (Int16) nbits;
      ip_emit_huffcode (pIPState, l, temp2, phuffInfo);
      phuffInfo->zero_l = 0;
  }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set huffman table for current coding.
 *
 * LongDescription:::
 *  Huffman table for DC, AC , luminent and color are different. Before
 * coding, we have to set the right table for current coding.
 *
 * Input:::
 *       chan::      Channel number 
 *     dcFlag::      DC flag ( 1: dc huffman table )
 *  phuffInfo::      Ptr to huffman info structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_set_huffTable (
    IPState    *pIPState,
    Int32	chan,
    Int32	dcFlag,
    WHUFFINFO  *phuffInfo)
{
    PROLOGUE;

    if(dcFlag)
    {
	phuffInfo->phuffco = phuffInfo->dc_lum_hufco;
	phuffInfo->phuffsi = phuffInfo->dc_lum_hufsi;
	if(chan >0)
	{
	    phuffInfo->phuffco = phuffInfo->dc_chr_hufco;
	    phuffInfo->phuffsi = phuffInfo->dc_chr_hufsi;
	}
	phuffInfo->last_dc_val =0;
    }
    else 
    {
	phuffInfo->phuffco = phuffInfo->ac_lum_hufco;
	phuffInfo->phuffsi = phuffInfo->ac_lum_hufsi;
	if(chan >0)
	{
	    phuffInfo->phuffco = phuffInfo->ac_chr_hufco;
	    phuffInfo->phuffsi = phuffInfo->ac_chr_hufsi;
	}
	phuffInfo->zero_l =0;
    }
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Emit remain zero len or emit remain bits
 *
 * LongDescription:::
 *   Emit remain zero len or emit remain bits
 *
 * Input:::
 *        flag::      1: emit zero len, 0: emit remain bits 
 *   phuffInfo::      Ptr to huffman info structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_clearZerolen (
    IPState    *pIPState,
    Int32	flag,
    WHUFFINFO  *phuffInfo)
{
  PROLOGUE;
  Int32  idx;
  Int16  pixel;

  if(flag)
  {
      if (phuffInfo->zero_l > 0)
      {
	  idx = (Int32)phuffInfo->phuffco[0];
	  pixel = (Int16)phuffInfo->phuffsi[0];
	  ERR_PASS(ip_emit_huffcode (pIPState, pixel, idx, phuffInfo));
      } 
      phuffInfo->zero_l =0;
  }
  else
  {
      if(phuffInfo->bits_to_go < 8)
      {
	  phuffInfo->IOPath(pIPState, phuffInfo->pdata, &phuffInfo->buffer, 1);
	  ++ phuffInfo->codeLen;
      }
  }

  NORMAL_RETURN;
}


/*
$Log: miwchuff.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.4  1998/01/13 23:53:47  ddavies
miwchuff.c -> /product/ipcore/ipshared/src/RCS/miwchuff.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1997/09/17 23:30:07  xzhu
miwchuff.c -> /product/ipcore/ipshared/src/RCS/miwchuff.c,v

CommonCheck: Huff-info will be a passed in variable and not a global variable. So we can get huffman coding infomation directly from Huff-info.

Revision 1.2  1997/07/11 01:51:23  ddavies
miwchuff.c -> /product/ipcore/ipshared/src/RCS/miwchuff.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.1  1996/10/16 00:02:44  xzhu
Initial revision

*/
