/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	wdhuff.c
 *
 *  Private routine:
 *      ip_d_derived_tbl ()
 *      ip_waveDehuffInit()
 *           ip_fill_bit ()
 *           ip_h_decode ()
 *       ip_huffDecoding ()
 *         ip_ac_decoding()
 *         ip_dc_decoding()
 *      ip_set_dehuffTable ()
 *
 */

#include <stdio.h>
#include <math.h>

#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"

#include "wavelib.pub"
#include "wavelib.prv"

IP_RCSINFO(miwdhuff_c_RCSInfo, "$RCSfile: miwdhuff.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

static DEHUFFINFO dehuffInfo;

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Derive huffman decoding table.
 *
 * LongDescription:::
 *    Derive huffman decoding table (copy from JPEG).
 *
 * Input:::
 *       bits::      Ptr to input code bit string
 *    huffval::      Ptr to input code value, it will be huffman value
 *     valLen::      String length of code value
 *     valptr::      ptr to index of first symbol with code length x
 *    mincode::      ptr to min code with colength x
 *    maxcode::      ptr to max code with code length x
 * look_nbits::      ptr to look ahead table bits string
 *   look_sym::      ptr to look ahead table symbol string
 *
 * Output:::
 *    hbits::      Ptr to output code bit string
 *      val::      Ptr to output code value, it will be huffman value
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_d_derived_tbl (
    IPState    *pIPState,
    const UInt8 *bits,
    const UInt8 *huffval,
    Int32	valLen,
    Int32      *valptr,
    Int32      *mincode,
    Int32      *maxcode,
    Int32      *look_nbits,
    UInt8      *look_sym,
    UInt8      *val,
    UInt8      *hbits)
{
  PROLOGUE;

  Int32 p, i, l, lookbits, ctr, si;
  Int8 huffsize[257];
  UInt16 huffcode[257];
  UInt16 code;

  memcpy(val, (const void*)huffval, valLen);
  memcpy(hbits, (const void*)bits, 17);

  /* Figure C.1: make table of Huffman code length for each symbol */
  /* Note that this is in code-length order. */

  p = 0;
  for (l = 1; l <= 16; l++) {
    for (i = 1; i <= (Int32) bits[l]; i++)
      huffsize[p++] = (Int8) l;
  }
  huffsize[p] = 0;
  
  /* Figure C.2: generate the codes themselves */
  /* Note that this is in code-length order. */
  
  code = 0;
  si = huffsize[0];
  p = 0;
  while (huffsize[p]) {
    while (((Int32) huffsize[p]) == si) {
      huffcode[p++] = code;
      code++;
    }
    code <<= 1;
    si++;
  }

  /* Figure F.15: generate decoding tables for bit-sequential decoding */

  p = 0;
  for (l = 1; l <= 16; l++) {
    if (bits[l]) {
      valptr[l] = p; /* huffval[] index of 1st symbol of code length l */
      mincode[l] = huffcode[p]; /* minimum code of length l */
      p += bits[l];
      maxcode[l] = huffcode[p-1]; /* maximum code of length l */
    } else {
      maxcode[l] = -1;	/* -1 if no codes of this length */
    }
  }
  maxcode[17] = 0xFFFFFL; /* ensures jpeg_huff_decode terminates */

  /* Compute lookahead tables to speed up decoding.
   * First we set all the table entries to 0, indicating "too long";
   * then we iterate through the Huffman codes that are short enough and
   * fill in all the entries that correspond to bit sequences starting
   * with that code.
   */

  p = 0;
  for (l = 1; l <= 8; l++) {
    for (i = 1; i <= (Int32) bits[l]; i++, p++) {
      /* l = current code's length, p = its index in huffcode[] & huffval[]. */
      /* Generate left-justified code followed by all possible bit sequences */
      lookbits = huffcode[p] << (8-l);
      for (ctr = 1 << (8-l); ctr > 0; ctr--) {
	look_nbits[lookbits] = l;
	look_sym[lookbits] = huffval[p];
	lookbits++;
      }
    }
  }

  NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman decoding init.
 *
 * LongDescription:::
 *  Create huffman decoding table for DC and AC ( copy from jpeg ) and init.
 * coding coef. 
 *
 * Input:::
 *      pdata::      Ptr to data used in IO callback func
 *  inputFunc::      Ptr to callback function
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveDehuffInit (
    IPState    *pIPState,
    void       *pdata,
    Int32 (CDECL *inputFunc) (
		    IPState    *pIPState,
		    void       *userData,
		    UInt8      *buf,
		    UInt32	count))
{
    PROLOGUE;
    Int32    valLen;

  static const UInt8 bits_dc_luminance[17] =
    { /* 0-base */ 0, 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 };
  static const UInt8 val_dc_luminance[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
  static const UInt8 bits_dc_chrominance[17] =
    { /* 0-base */ 0, 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 };
  static const UInt8 val_dc_chrominance[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
  static const UInt8 bits_ac_luminance[17] =
    { /* 0-base */ 0, 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d };
  static const UInt8 val_ac_luminance[] =
    { 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
      0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
      0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
      0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
      0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
      0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
      0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
      0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
      0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
      0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
      0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
      0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
      0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
      0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
      0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
      0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
      0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
      0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
      0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
      0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };
  
  static const UInt8 bits_ac_chrominance[17] =
    { /* 0-base */ 0, 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77 };
  static const UInt8 val_ac_chrominance[] =
    { 0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
      0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
      0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
      0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
      0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
      0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
      0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
      0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
      0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
      0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
      0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
      0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
      0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
      0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
      0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
      0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
      0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
      0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
      0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
      0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };

    DEHUFFINFO *phuffInfo;

    phuffInfo = &dehuffInfo;
    phuffInfo->pdata = pdata;
    phuffInfo->IOPath = inputFunc;
    phuffInfo->bits_left = 0;
    phuffInfo->last_dc_val =0;
    phuffInfo->zero_l =0;
    phuffInfo->buffer =0;

    valLen = (Int32)sizeof(val_dc_luminance);
    ERR_PASS(ip_d_derived_tbl (pIPState, bits_dc_luminance,
			 val_dc_luminance, valLen, 
			 phuffInfo->dc_lum_valptr, phuffInfo->dc_lum_mincode,
			 phuffInfo->dc_lum_maxcode,phuffInfo->dc_lum_nbits,
			 phuffInfo->dc_lum_sym,phuffInfo->dc_lum_huffval,
			 phuffInfo->dc_lum_bits ));
    valLen = (Int32)sizeof(val_dc_chrominance);
    ERR_PASS(ip_d_derived_tbl (pIPState, bits_dc_chrominance,
			 val_dc_chrominance, valLen,
			 phuffInfo->dc_chr_valptr, phuffInfo->dc_chr_mincode,
			 phuffInfo->dc_chr_maxcode,phuffInfo->dc_chr_nbits,
			 phuffInfo->dc_chr_sym, phuffInfo->dc_chr_huffval,
			 phuffInfo->dc_chr_bits));
    valLen = (Int32)sizeof(val_ac_luminance);
    ERR_PASS(ip_d_derived_tbl (pIPState, bits_ac_luminance,
			 val_ac_luminance, valLen,
			 phuffInfo->ac_lum_valptr, phuffInfo->ac_lum_mincode,
			 phuffInfo->ac_lum_maxcode,phuffInfo->ac_lum_nbits,
			 phuffInfo->ac_lum_sym, phuffInfo->ac_lum_huffval,
			 phuffInfo->ac_lum_bits));
    valLen = (Int32)sizeof(val_ac_chrominance);
    ERR_PASS(ip_d_derived_tbl (pIPState, bits_ac_chrominance,
			 val_ac_chrominance, valLen,
			 phuffInfo->ac_chr_valptr, phuffInfo->ac_chr_mincode,
			 phuffInfo->ac_chr_maxcode,phuffInfo->ac_chr_nbits,
			 phuffInfo->ac_chr_sym, phuffInfo->ac_chr_huffval,
			 phuffInfo->ac_chr_bits));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Fill in bit.
 *
 * LongDescription:::
 *  After huffman decoing, fill bits in output buf.
 *
 * Input:::
 *       nbits::      Bits to be filled
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_fill_bit (
    IPState    *pIPState,
    Int32	nbits)
{
    PROLOGUE;
    UInt8  c;
    DEHUFFINFO *pInfo;

/* Load up the bit buffer to a depth of at least nbits */
    pInfo = &dehuffInfo;
    while (pInfo->bits_left < MIN_GET_BITS) {
    /* Attempt to read a byte */
	pInfo->IOPath(pIPState, pInfo->pdata, &c, sizeof(UInt8));

    /* If it's 0xFF, check and discard stuffed zero byte */
	if (c == 0xFF) {
	    do {
		pInfo->IOPath(pIPState, pInfo->pdata, &c, sizeof(UInt8));
	    } while (c == 0xFF);

	    if (c == 0) {
		/* Found FF/00, which represents an FF data byte */
		c = 0xFF;
	    } else {
	/* Oops, it's actually a marker indicating end of compressed data. */
	/* Better put it back for use later */

/*no_more_data*/
		pInfo->buffer <<= 8;
		pInfo->bits_left += 8;
		while(pInfo->bits_left < nbits) {
		    pInfo->buffer <<= 8;
		    pInfo->bits_left += 8;
		}
		break;
	    }
	}

    /* OK, load c into get_buffer */
	pInfo->buffer = (pInfo->buffer << 8) | c;
	pInfo->bits_left += 8;
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Decode huffman code.
 *
 * LongDescription:::
 *    Decode one huffman code. It will be called by ip_huffDecoding().
 *
 * Input:::
 *   min_bits::      Huffman code min_bits
 *
 * Output:::
 *    symbol::       Ptr to decoded symbol 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_h_decode (
    IPState    *pIPState,
    Int32	min_bits,
    Int16      *symbol)
{
    PROLOGUE;

    Int32   l;
    Int32   code;
    Int32   *maxcode; 
     DEHUFFINFO *pInfo;

  /* HUFF_DECODE has determined that the code is at least min_bits */
  /* bits long, so fetch that many bits in one swoop. */

    pInfo = &dehuffInfo;
    maxcode = pInfo->pmaxcode;
    l = min_bits;
    if(pInfo->bits_left < 16)
    {
	if(pInfo->bits_left < min_bits)
	    ip_fill_bit (pIPState, min_bits);
	code = (pInfo->buffer >> (pInfo->bits_left -= (Int16)min_bits)) & 
	     ((1<<min_bits)-1);

  /* Collect the rest of the Huffman code one bit at a time. */
  /* This is per Figure F.16 in the JPEG spec. */

	while (code > maxcode[l]) {
	    code <<= 1;
	    if(pInfo->bits_left < 1)
		ip_fill_bit (pIPState, 1);
	    code |=  (Int32)((pInfo->buffer >> (-- pInfo->bits_left)) & 1);
	    l++;
	}
    }
    else
    {
	code = (Int32)(pInfo->buffer >> (pInfo->bits_left -= (Int16) l)) & 
	     ((1<<l)-1);

  /* Collect the rest of the Huffman code one bit at a time. */
  /* This is per Figure F.16 in the JPEG spec. */

	while (code > maxcode[l]) {
	    code <<= 1;
	    code |=  ((pInfo->buffer >> (-- pInfo->bits_left)) & 1);
	    l++;
	}
    }

  /* With garbage input we may reach the sentinel value l = 17. */

    if (l > 16) {
	ERR_RETURN;
    }

    *symbol = pInfo->phuffval[pInfo->pvalptr[l] + (code - pInfo->pmincode[l])];

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Huffman decoding.
 *
 * LongDescription:::
 *    Huffman decoding and will call ip_h_decoding() to decode 
 *    huffman code.
 *
 * Input:::
 *
 * Output:::
 *    symbol::       Ptr to decoded symbol 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_huffDecoding (
    IPState    *pIPState,
    Int16      *symbol)
{
    PROLOGUE;
    
    Int32  nb, look;
    DEHUFFINFO *phuffInfo;

    phuffInfo = &dehuffInfo;
    if (phuffInfo->bits_left < 8) 
    { 
	ip_fill_bit (pIPState, 0);
	if (phuffInfo->bits_left < 8) {
	    nb = 1; 
	    goto slowlabel;
	} 
    } 
    look = (Int32) (phuffInfo->buffer >> (phuffInfo->bits_left - 8)) & 0xff; 
    if ((nb = phuffInfo->plook_nbits[look]) != 0) { 
	 phuffInfo->bits_left -= (Int16)nb; 
	 *symbol = phuffInfo->plook_sym[look]; 
    } else { 
	nb = 8+1; 
slowlabel: 
	ERR_PASS(ip_h_decode (pIPState, nb, symbol));
    } 
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    AC Huffman decoding.
 *
 * LongDescription:::
 *    AC Huffman decoding and will call ip_huffDecoding() to decode 
 *    huffman code.
 *
 * Input:::
 *
 * Output:::
 *    symbol::       Ptr to decoded symbol 
 *       len::       Ptr to bits length
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_ac_decoding (
    IPState    *pIPState,
    Int16      *symbol,
    Int16      *len)
{
    PROLOGUE;
    
    Int16 index;
    Int32 pixel;
    DEHUFFINFO *phuffInfo;

    phuffInfo = &dehuffInfo;
    *len =0;
    ERR_PASS(ip_huffDecoding (pIPState, &index));

    *len = index >> 4;
    index &= 15;
    *symbol =0;  

    if (index) 
    {
	if(phuffInfo->bits_left < index)
	    ip_fill_bit (pIPState, index);
	pixel =(phuffInfo->buffer >> (phuffInfo->bits_left -= index))& 
	    ((1<<index)-1);
	*symbol = (Int16)( (pixel < (1<<(index-1))) ? 
		 pixel + ((-1<<index) + 1) : pixel );	
    } 
    else {
	if (*len != 15)
	    *len = MAX_LEN;
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    DC Huffman decoding.
 *
 * LongDescription:::
 *    DC Huffman decoding and will call ip_huffDecoding() to decode 
 *    huffman code.
 *
 * Input:::
 *
 * Output:::
 *    symbol::       Ptr to decoded symbol 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_dc_decoding (
    IPState    *pIPState,
    Int16      *symbol)
{
    PROLOGUE;
    Int16     index;
    Int32     r;

    ip_huffDecoding (pIPState, &index);
    if ( index ) 
    {
	if(dehuffInfo.bits_left < index)
	    ip_fill_bit (pIPState, index);
	r = (dehuffInfo.buffer >> (dehuffInfo.bits_left -= index)) & 
			((1<<index)-1);
	index = (Int16)(r < (1<<(index-1)) ? r + ((-1<<index) + 1) : r);
    }

    index += dehuffInfo.last_dc_val;
    dehuffInfo.last_dc_val = index;
    *symbol = index;

    NORMAL_RETURN;

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set huffman table for current decoding.
 *
 * LongDescription:::
 *  Huffman table for DC, AC , luminent and color are different. Before
 * decoding, we have to set the right table for current decoding.
 *
 * Input:::
 *       chan::      Channel number 
 *     dcFlag::      DC flag ( 1: dc huffman table )
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_set_dehuffTable (
    IPState    *pIPState,
    Int32	chan,
    Int32	dcFlag)
{
    PROLOGUE;

    if(dcFlag)
    {
	dehuffInfo.pvalptr = dehuffInfo.dc_lum_valptr;
	dehuffInfo.pmincode = dehuffInfo.dc_lum_mincode;
	dehuffInfo.pmaxcode = dehuffInfo.dc_lum_maxcode;
	dehuffInfo.plook_nbits = dehuffInfo.dc_lum_nbits;
	dehuffInfo.plook_sym = dehuffInfo.dc_lum_sym;
	dehuffInfo.phuffval = dehuffInfo.dc_lum_huffval;
	if(chan >0)
	{
	    dehuffInfo.pvalptr = dehuffInfo.dc_chr_valptr;
	    dehuffInfo.pmincode = dehuffInfo.dc_chr_mincode;
	    dehuffInfo.pmaxcode = dehuffInfo.dc_chr_maxcode;
	    dehuffInfo.plook_nbits = dehuffInfo.dc_chr_nbits;
	    dehuffInfo.plook_sym = dehuffInfo.dc_chr_sym;
	    dehuffInfo.phuffval = dehuffInfo.dc_chr_huffval;
	}
	dehuffInfo.last_dc_val =0;
    }
    else 
    {
	dehuffInfo.pvalptr = dehuffInfo.ac_lum_valptr;
	dehuffInfo.pmincode = dehuffInfo.ac_lum_mincode;
	dehuffInfo.pmaxcode = dehuffInfo.ac_lum_maxcode;
	dehuffInfo.plook_nbits = dehuffInfo.ac_lum_nbits;
	dehuffInfo.plook_sym = dehuffInfo.ac_lum_sym;
	dehuffInfo.phuffval = dehuffInfo.ac_lum_huffval;
	if(chan >0)
	{
	    dehuffInfo.pvalptr = dehuffInfo.ac_chr_valptr;
	    dehuffInfo.pmincode = dehuffInfo.ac_chr_mincode;
	    dehuffInfo.pmaxcode = dehuffInfo.ac_chr_maxcode;
	    dehuffInfo.plook_nbits = dehuffInfo.ac_chr_nbits;
	    dehuffInfo.plook_sym = dehuffInfo.ac_chr_sym;
	    dehuffInfo.phuffval = dehuffInfo.ac_chr_huffval;
	}
    }
    
    NORMAL_RETURN;
}

 
/*
$Log: miwdhuff.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.5  1998/01/13 23:53:53  ddavies
miwdhuff.c -> /product/ipcore/ipshared/src/RCS/miwdhuff.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1997/09/17 22:48:05  xzhu
miwdhuff.c -> /product/ipcore/ipshared/src/RCS/miwdhuff.c,v

CommonCheck: Added #include wavelib.pub, since some changes made in huffman coding.

Revision 1.3  1997/07/11 01:51:30  ddavies
miwdhuff.c -> /product/ipcore/ipshared/src/RCS/miwdhuff.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.2  1997/05/19 23:10:38  xzhu
xzhu on Mon May 19 16:08:25 PDT 1997

miwdhuff.c -> /product/ipcore/ipshared/src/RCS/miwdhuff.c,v

Input string length into ip_d_derived_tbl() to avoid solaris random crash.

Revision 1.1  1996/10/16 00:02:47  xzhu
Initial revision


*/
