 /*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*	wdtrans.c:
 *
 *  Private routine:
 *      ip_waveInvertTrans()
 *      ip_init_invertTbl()
 *      ip_waveInvertTrans_LUT()
 *      ip_waveInvertLift()
 *
 */

#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"

#include "wavelib.pub"
#include "wavelib.prv"

IP_RCSINFO(miwdtra_c_RCSInfo, "$RCSfile: miwdtra.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

#define TABLE_LEN    256

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set up inverse filter look up table to speed up.
 *
 * LongDescription:::
 *  filter parameters have been scaled during forward transformation.
 *  Since we do uantization on wavelets, after inverse transformation,
 *  the parameter range is difficulty to limit in 255*level*2. So we
 *  set up two 256 table, one for first byte, one for high byte.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *  r_lowpass::      Ptr to inverse low pass filter
 * r_highpass::      Ptr to inverse high pass filter
 *     lowlen::      inverse low pass filter length
 *    highlen::      inverse high pass filter length
 *     tbllen::      lookup table length  (now we use fixed length)
 *
 * Output:::
 *   lowTab::         low pass look up table
 *  highTab::        high pass look up table
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_init_invertTbl (
    IPState    *pIPState,
    Int32      *r_lowpass,
    Int32      *r_highpass,
    Int32	lowlen,
    Int32	highlen,
    Int32	tbllen,
    Int32     **lowTab,
    Int32     **highTab)
{
    PROLOGUE;
    Int32  idx, col;
    Int32  min, max;
    Int32  *ptbl;
 
    tbllen = tbllen;
    min = - TABLE_LEN;
    max = TABLE_LEN;
    for(idx=0; idx<lowlen; ++idx)
    {
	ptbl = lowTab[idx]+ TABLE_LEN;
	for(col=min; col<=max; ++col)
	    ptbl[col] = r_lowpass[idx]*col;
    }

    for(idx =0; idx <highlen; ++idx)
    {
	ptbl = highTab[idx]+ TABLE_LEN;
	for(col=min; col<=max; ++col)
	    ptbl[col] = r_highpass[idx]*col;
    }    

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  wavelet inverse transform.
 *
 * LongDescription:::
 *    Do wavelet inverse transform on assigned region.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *  r_lowpass::     Ptr to invert lowpass filter
 * r_highpass::     Ptr to invert highpass filter
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveInvertTrans (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32      *r_lowpass,
    Int32      *r_highpass,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh, temp;
    Int16       *coef, *buffer, *linebuf, *pSrc, *pDes;
    Int16       t0,t1,t2,t3,t4;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer +4;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+=bpl)
    {
	pDes = linebuf;
	
	for(col=0, idx=halfw; idx <coefw; ++col, ++idx)
	{
	    *pDes++ = *(pSrc+col);
	    *pDes++ = *(pSrc+idx);
	}
	if( col<halfw ) *pDes++ = *(pSrc+col);
	for(idx=0, col= 8; idx<4; ++idx, --col)
	    *(buffer+idx) = *(buffer+col);
	for(idx=1, col=coefw-1; idx<5; ++idx)
	    *(linebuf+idx+col) = *(linebuf+col-idx);

	pDes = pSrc;
	for(idx=0; idx<coefw; ++idx)
	{
	    t0 = *(buffer+idx+4);
	    t1 = *(buffer+idx+5) + *(buffer+idx+3);
	    t2 = *(buffer+idx+6) + *(buffer+idx+2);
	    t3 = *(buffer+idx+7) + *(buffer+idx+1);
	    temp =r_highpass[0] * t0 +
		r_highpass[1] * t1 +
		r_highpass[2] * t2 +
		r_highpass[3] * t3;
	    *pDes++ =(Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefw) break;
	    t0 = *(buffer+idx+4);
	    t1 = *(buffer+idx+5) + *(buffer+idx+3);
	    t2 = *(buffer+idx+6) + *(buffer+idx+2);
	    t3 = *(buffer+idx+7) + *(buffer+idx+1);
	    t4 = *(buffer+idx+8) + *(buffer+idx);
	    temp = r_lowpass[0] * t0 +
		r_lowpass[1] * t1 +
		r_lowpass[2] * t2 +
		r_lowpass[3] * t3 +
		r_lowpass[4] * t4;
	    *pDes++ =(Int16)((temp+ONEHALF)>>SCALEBITS);
	}
    }

/* collunm */
    for(col=0; col<coefw; ++col)
    {
/* copy image */
	pDes = linebuf;
	pSrc = coef + halfh * bpl;
	for(idx=0, row=col; idx<coefh; ++idx, row += bpl )
	{
	    *pDes ++ = *(coef + row);
	    if( ++idx >= coefh) break;
	    *pDes ++ = *(pSrc + row);
	}
	for(idx=0, row = 8; idx<4; ++idx, --row)
	    *(buffer+idx)= *(buffer+row);
	for(idx=1, row=coefh-1; idx<5; ++idx)
	    *(linebuf+idx+row) = *(linebuf+row-idx);

/* convelution */
	for(idx=0, pDes = coef+col; idx<coefh; ++idx, pDes += bpl)
	{
	    t0 = *(buffer+idx+4);
	    t1 = *(buffer+idx+5) + *(buffer+idx+3);
	    t2 = *(buffer+idx+6) + *(buffer+idx+2);
	    t3 = *(buffer+idx+7) + *(buffer+idx+1);
	    temp =r_highpass[0] * t0 +
		r_highpass[1] * t1 +
		r_highpass[2] * t2 +
		r_highpass[3] * t3;
	    *pDes=(Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefh) break;
	    pDes += bpl;
	    t0 = *(buffer+idx+4);
	    t1 = *(buffer+idx+5) + *(buffer+idx+3);
	    t2 = *(buffer+idx+6) + *(buffer+idx+2);
	    t3 = *(buffer+idx+7) + *(buffer+idx+1);
	    t4 = *(buffer+idx+8) + *(buffer+idx);
	    temp = r_lowpass[0] * t0 +
		r_lowpass[1] * t1 +
	        r_lowpass[2] * t2 +
		r_lowpass[3] * t3 +
		r_lowpass[4] * t4;
	    *pDes=(Int16)((temp+ONEHALF)>>SCALEBITS);
	}	
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  wavelet inverse transform.
 *
 * LongDescription:::
 *    Do wavelet inverse transform on assigned region.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveInvertLift (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh, temp, temp1;
    Int16       *coef, *buffer, *linebuf, *pSrc, *pDes, para;
    Int32       sp1, s0, sn1, sn2, dp1, d0, dn1, prev;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer +4;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+=bpl)
    {
	pDes = linebuf;
	
	for(col=0, idx=halfw; idx <coefw; ++col, ++idx)
	{
	    *pDes++ =(Int16) (((Int32)(*(pSrc+col))<< SCALEBITS)/ALPH4);
	    *pDes++ =(Int16)(((Int32)(*(pSrc+idx))*ALPH4 +ONEHALF)>>SCALEBITS);
	}
	if( col<halfw ) 
	    *pDes++ = (Int16) (((Int32)(*(pSrc+col))<< SCALEBITS)/ALPH4);
	for(idx=0, col= 8; idx<4; ++idx, --col)
	    *(buffer+idx) = *(buffer+col);
	for(idx=1, col=coefw-1; idx<6; ++idx)
	    *(linebuf+idx+col) = *(linebuf+col-idx);

	pDes = pSrc;

	sp1 = ((Int32)(*(buffer + 2))<< SCALEBITS) - 
	    ALPH3 *(*(buffer + 3) + *(buffer +1));
	s0 = ((Int32)(*(buffer + 4))<< SCALEBITS) - 
	    ALPH3 *(*(buffer + 5) + *(buffer +3));
	para = (Int16)((sp1 + s0 +ONEHALF)>>SCALEBITS);
	dp1 = ((Int32)(*(buffer +3))<< SCALEBITS) - ALPH2 * para;
	sn1 = ((Int32)(*(buffer +6))<< SCALEBITS) - 
	    ALPH3 *(*(buffer +7) + *(buffer +5));
	para = (Int16)((s0+sn1+ONEHALF)>>SCALEBITS);
	d0 = ((Int32)(*(buffer +5))<< SCALEBITS) - ALPH2 * para;
	prev = s0 - ALPH1 * ((dp1+d0 +ONEHALF)/ONEHALF1);
	for(idx=0; idx<coefw; idx += 2)
	{
	    *pDes++ =(Int16)((prev +ONEHALF)>>SCALEBITS);
	    if((idx+1) >= coefw) break;
	    sn2 = ((Int32)(*(buffer+idx+8))<< SCALEBITS) - 
		ALPH3 *(*(buffer+idx+9) + *(buffer+idx+7));
	    para = (Int16)((sn1+ sn2 +ONEHALF)>>SCALEBITS);
	    dn1 = ((Int32)(*(buffer+idx+7))<< SCALEBITS) -ALPH2 * para;
	    temp1 = prev;
	    para = (Int16)((d0 + dn1+ONEHALF)>>SCALEBITS);
	    prev = sn1 - ALPH1 * para;
	    temp = d0 - ALPH0 *((temp1 + prev+ONEHALF)/ONEHALF1); 
	    *pDes++ =(Int16)((temp+ONEHALF)>>SCALEBITS);
	    sp1 = s0;
	    s0 = sn1;
	    sn1 = sn2;
	    dp1 = d0;
	    d0 = dn1;
	}
    }

/* collunm */
    for(col=0; col<coefw; ++col)
    {
/* copy image */
	pDes = linebuf;
	pSrc = coef + halfh * bpl;
	for(idx=0, row=col; idx<coefh; ++idx, row += bpl )
	{
	    *pDes ++ = (*(coef + row)<< SCALEBITS)/ALPH4;
	    if( ++idx >= coefh) break;
	    *pDes ++ = ((Int32)(*(pSrc + row))*ALPH4 +ONEHALF)>>SCALEBITS;
	}
	for(idx=0, row = 8; idx<4; ++idx, --row)
	    *(buffer+idx)= *(buffer+row);
	for(idx=1, row=coefh-1; idx<6; ++idx)
	    *(linebuf+idx+row) = *(linebuf+row-idx);

/* convelution */
	sp1 = ((Int32)(*(buffer + 2))<< SCALEBITS) - 
	    ALPH3 *(*(buffer + 3) + *(buffer +1));
	s0 = ((Int32)(*(buffer + 4))<< SCALEBITS) - 
	    ALPH3 *(*(buffer + 5) + *(buffer +3));
	para = (Int16)((sp1 + s0 +ONEHALF)>>SCALEBITS);
	dp1 = ((Int32)(*(buffer +3))<< SCALEBITS) - ALPH2 * para;
	sn1 = ((Int32)(*(buffer +6))<< SCALEBITS) - 
	    ALPH3 *(*(buffer +7) + *(buffer +5));
	para = (Int16)((s0+sn1+ONEHALF)>>SCALEBITS);
	d0 = ((Int32)(*(buffer +5))<< SCALEBITS) - ALPH2 * para;
	prev = s0 - ALPH1 * ((dp1+d0 +ONEHALF)/ONEHALF1);
	for(idx=0, pDes = coef+col; idx<coefh; idx+=2, pDes += bpl)
	{
	    *pDes =(Int16)((prev +ONEHALF)>>SCALEBITS);
	    if( (idx+1) >= coefh) break;
	    pDes += bpl;
	    sn2 = ((Int32)(*(buffer+idx+8))<< SCALEBITS) - 
		ALPH3 *(*(buffer+idx+9) + *(buffer+idx+7));
	    para = (Int16)((sn1+ sn2 +ONEHALF)>>SCALEBITS);
	    dn1 = ((Int32)(*(buffer+idx+7))<< SCALEBITS) -ALPH2 * para;
	    temp1 = prev;
	    para = (Int16)((d0 + dn1+ONEHALF)>>SCALEBITS);
	    prev = sn1 - ALPH1 * para;
	    temp = d0 - ALPH0 *((temp1 + prev+ONEHALF)/ONEHALF1); 
	    *pDes =(Int16)((temp+ONEHALF)>>SCALEBITS);
	    sp1 = s0;
	    s0 = sn1;
	    sn1 = sn2;
	    dp1 = d0;
	    d0 = dn1;
	}	
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  wavelet inverse transform.
 *
 * LongDescription:::
 *    Do wavelet inverse transform on assigned region.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_wave44InvertLift (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh;
    Int16       *coef, *buffer, *linebuf, *pSrc, *pDes, *pbuf;
    Int32        p1, p2, xp1, x0, xn1, xn2, x1;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer +6;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+=bpl)
    {
	pDes = linebuf;
	
	for(col=0, idx=halfw; idx <coefw; ++col, ++idx)
	{
	    *pDes++ =(Int16) (*(pSrc+col));
	    *pDes++ =(Int16) (*(pSrc+idx));
	}
	if( col<halfw ) 
	    *pDes++ = (Int16) (*(pSrc+col));
	for(idx=0, col= 12; idx<6; ++idx, --col)
	    *(buffer+idx) = *(buffer+col);
	for(idx=1, col=coefw-1; idx<7; ++idx)
	    *(linebuf+idx+col) = *(linebuf+col-idx);

	pDes = pSrc;
	pbuf = linebuf;
	p1 = 9 * (*(pbuf -3) + *(pbuf -1));
	p2 = *(pbuf -5) + *(pbuf +1);
	xp1 = (Int32)(*(pbuf - 2)) - ((p1 - p2)/32);
	p1 = 9 * (*(pbuf -1) + *(pbuf +1));
	p2 = *(pbuf -3) + *(pbuf +3);
	x0 = (Int32)(*(pbuf)) - ((p1 - p2)/32);
	p1 = 9 * (*(pbuf + 1) + *(pbuf + 3));
	p2 = *(pbuf -1) + *(pbuf +5);
	xn1 = (Int32)(*(pbuf + 2)) - ((p1 - p2)/32);
	for(idx=0; idx<coefw; idx += 2, pbuf += 2)
	{
	    *pDes++ =(Int16)( x0 );
	    if((idx+1) >= coefw) break;
	    p1 = 9 * (*(pbuf + 3) + *(pbuf + 5));
	    p2 = *(pbuf + 1) + *(pbuf +7);
	    xn2 = (Int32)(*(pbuf + 4)) - ((p1 - p2)/32);
	    x1 = (Int32)(*(pbuf+1)) +(( 9* (x0+xn1) - (xp1+xn2))/16);
	    *pDes++ =(Int16)( x1 );
	    xp1 = x0;
	    x0 = xn1;
	    xn1 = xn2;
	}
    }

/* collunm */
    for(col=0; col<coefw; ++col)
    {
/* copy image */
	pDes = linebuf;
	pSrc = coef + halfh * bpl;
	for(idx=0, row=col; idx<coefh; ++idx, row += bpl )
	{
	    *pDes ++ = (*(coef + row));
	    if( ++idx >= coefh) break;
	    *pDes ++ = (*(pSrc + row));
	}
	for(idx=0, row = 12; idx<6; ++idx, --row)
	    *(buffer+idx)= *(buffer+row);
	for(idx=1, row=coefh-1; idx<7; ++idx)
	    *(linebuf+idx+row) = *(linebuf+row-idx);

/* convelution */
	pbuf = linebuf;
	p1 = 9 * (*(pbuf -3) + *(pbuf -1));
	p2 = *(pbuf -5) + *(pbuf +1);
	xp1 = (Int32)(*(pbuf - 2)) - ((p1 - p2)/32);
	p1 = 9 * (*(pbuf -1) + *(pbuf +1));
	p2 = *(pbuf -3) + *(pbuf +3);
	x0 = (Int32)(*(pbuf)) - ((p1 - p2)/32);
	p1 = 9 * (*(pbuf + 1) + *(pbuf + 3));
	p2 = *(pbuf -1) + *(pbuf +5);
	xn1 = (Int32)(*(pbuf + 2)) - ((p1 - p2)/32);

	for(idx=0, pDes = coef+col; idx<coefh; idx+=2, pDes += bpl, pbuf+=2)
	{
	    *pDes =(Int16)( x0 );
	    if( (idx+1) >= coefh) break;
	    pDes += bpl;
	    p1 = 9 * (*(pbuf + 3) + *(pbuf + 5));
	    p2 = *(pbuf + 1) + *(pbuf +7);
	    xn2 = (Int32)(*(pbuf + 4)) - ((p1 - p2)/32);
	    x1 = (Int32)(*(pbuf+1)) +(( 9* (x0+xn1) - (xp1+xn2))/16);
	    *pDes =(Int16)( x1 );
	    xp1 = x0;
	    x0 = xn1;
	    xn1 = xn2;
	}	
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  wavelet inverse transform.
 *
 * LongDescription:::
 *    Do wavelet inverse transform on assigned region.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveD4InvertLift (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh, temp;
    Int16       *coef, *buffer, *linebuf, *pSrc, *pDes, *pbuf;
    Float32     s0, d0, dn1, p1, p0, pp;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer +4;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+=bpl)
    {
	pDes = linebuf;
	
	for(col=0, idx=halfw; idx <coefw; ++col, ++idx)
	{
	    *pDes++ =(*(pSrc+col));
	    *pDes++ =(*(pSrc+idx));
	}
	if( col<halfw ) 
	    *pDes++ =(*(pSrc+col));
	for(idx=0, col= 8; idx<4; ++idx, --col)
	    *(buffer+idx) = *(buffer+col);
	for(idx=1, col=coefw-1; idx<6; ++idx)
	    *(linebuf+idx+col) = *(linebuf+col-idx);

	pDes = pSrc;
	pbuf = linebuf;

	p0 = D4P3 * (Float32)(*(pbuf +1 ));
	d0 = p0 - D4P4 * (Float32)(*(pbuf - 2));
	for(idx=0; idx<coefw; idx += 2, pbuf += 2)
	{
	    p1 = D4P3 * (Float32)(*(pbuf +3 ));
	    pp = D4P4 * (Float32)(*(pbuf ));
	    dn1 = p1 - pp;
	    s0 = pp - (D4P1 * d0) - (D4P2 * dn1); 
	    *pDes++ =(Int16)( s0 );
	    if((idx+1) >= coefw) break;
	    temp = (Int32)( d0 + D4P0*s0 ); 
	    *pDes++ =(Int16)(temp);
	    d0 = dn1;
	    p0 = p1;
	}
    }

/* collunm */
    for(col=0; col<coefw; ++col)
    {
/* copy image */
	pDes = linebuf;
	pSrc = coef + halfh * bpl;
	for(idx=0, row=col; idx<coefh; ++idx, row += bpl )
	{
	    *pDes ++ = (*(coef+row));
	    if( ++idx >= coefh) break;
	    *pDes ++ = (*(pSrc+row));
	}
	for(idx=0, row = 8; idx<4; ++idx, --row)
	    *(buffer+idx)= *(buffer+row);
	for(idx=1, row=coefh-1; idx<6; ++idx)
	    *(linebuf+idx+row) = *(linebuf+row-idx);

	pbuf = linebuf;
/* convelution */
	p0 = D4P3 * (Float32)(*(pbuf +1 ));
	d0 = p0 - D4P4 * (Float32)(*(pbuf - 2));
	for(idx=0, pDes = coef+col; idx<coefh; idx+=2, pDes += bpl, pbuf +=2)
	{
	    p1 = D4P3 * (Float32)(*(pbuf +3 ));
	    pp = D4P4 * (Float32)(*(pbuf ));
	    dn1 = p1 - pp;
	    s0 = pp - (D4P1 * d0) - (D4P2 * dn1); 
	    *pDes =(Int16)(s0);
	    if( (idx+1) >= coefh) break;
	    pDes += bpl;
	    temp = (Int32)( d0 + D4P0*s0 ); 
	    *pDes =(Int16)(temp);
	    d0 = dn1;
	}	
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  wavelet inverse transform.
 *
 * LongDescription:::
 *    Do wavelet inverse transform on assigned region.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveTTInvert (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh;
    Int16       *coef, *buffer, *linebuf, *pSrc, *pDes, *pbuf;
    Int32        p0, s0, d0;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer +6;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+=bpl)
    {
	pDes = linebuf;
	
	for(col=0, idx=halfw; idx <coefw; ++col, ++idx)
	{
	    *pDes++ =(Int16) (*(pSrc+col));
	    *pDes++ =(Int16) (*(pSrc+idx));
	}
	if( col<halfw ) 
	    *pDes++ = (Int16) (*(pSrc+col));
	for(idx=0, col= 12; idx<6; ++idx, --col)
	    *(buffer+idx) = *(buffer+col);
	for(idx=1, col=coefw-1; idx<7; ++idx)
	    *(linebuf+idx+col) = *(linebuf+col-idx);

	pDes = pSrc;
	pbuf = linebuf;
	for(idx=0; idx<coefw; idx += 2, pbuf += 2)
	{
	    p0 =(3*(*(pbuf -4)- *(pbuf + 4))-22*(*(pbuf -2)- *(pbuf+2))+32)/64;
	    d0 = ( *(pbuf +1) - p0)/2;
	    s0 = (Int32)*pbuf + d0;
	    *pDes++ =(Int16)( s0 );
	    if((idx+1) >= coefw) break;
	    s0 = (Int32)*pbuf - d0;
	    *pDes++ =(Int16)( s0 );
	}
    }

/* collunm */
    for(col=0; col<coefw; ++col)
    {
/* copy image */
	pDes = linebuf;
	pSrc = coef + halfh * bpl;
	for(idx=0, row=col; idx<coefh; ++idx, row += bpl )
	{
	    *pDes ++ = (*(coef + row));
	    if( ++idx >= coefh) break;
	    *pDes ++ = (*(pSrc + row));
	}
	for(idx=0, row = 12; idx<6; ++idx, --row)
	    *(buffer+idx)= *(buffer+row);
	for(idx=1, row=coefh-1; idx<7; ++idx)
	    *(linebuf+idx+row) = *(linebuf+row-idx);

/* convelution */
	pbuf = linebuf;
	for(idx=0, pDes = coef+col; idx<coefh; idx+=2, pDes += bpl, pbuf+=2)
	{
	    p0 =(3*(*(pbuf -4)- *(pbuf + 4))-22*(*(pbuf -2)- *(pbuf+2))+32)/64;
	    d0 = (*(pbuf +1) - p0)/2;
	    s0 = (Int32)*pbuf + d0;
	    *pDes =(Int16)( s0 );
	    if( (idx+1) >= coefh) break;
	    pDes += bpl;
	    s0 = (Int32)*pbuf - d0;
	    *pDes =(Int16)( s0 );
	}	
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  wavelet inverse transform.
 *
 * LongDescription:::
 *    Do wavelet inverse transform on assigned region with LUT.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *     lowTab::     Ptr to invert lowpass filter LUT
 *    highTab::     Ptr to invert highpass filter LUT
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveInvertTrans_LUT (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32     **lowTab,
    Int32     **highTab,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh, temp;
    Int16       *coef, *buffer, *linebuf, *pSrc, *pDes, *ptmp;
    Int16       t0,t1,t2,t3,t4, tup, tlow;
    Int32       *low1, *low2, *low3, *low4, *low5;
    Int32       *high1, *high2, *high3, *high4;

    coef = *coefbuf;
    buffer = *tempbuf;
    low1 = lowTab[0]+TABLE_LEN;
    low2 = lowTab[1]+TABLE_LEN;
    low3 = lowTab[2]+TABLE_LEN;
    low4 = lowTab[3]+TABLE_LEN;
    low5 = lowTab[4]+TABLE_LEN;
    high1 = highTab[0]+TABLE_LEN;
    high2 = highTab[1]+TABLE_LEN;
    high3 = highTab[2]+TABLE_LEN;
    high4 = highTab[3]+TABLE_LEN;

    linebuf = buffer +4;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+=bpl)
    {
	pDes = linebuf;
	
	for(col=0, idx=halfw; idx <coefw; ++col, ++idx)
	{
	    *pDes++ = *(pSrc+col);
	    *pDes++ = *(pSrc+idx);
	}
	if(col<halfw ) *pDes++ = *(pSrc+col); 
	for(idx=0, col= 8; idx<4; ++idx, --col)
	    *(buffer+idx) = *(buffer+col);
	for(idx=1, col=coefw-1; idx<5; ++idx)
	    *(linebuf+idx+col) = *(linebuf+col-idx);

	pDes = pSrc;
	ptmp = buffer;
	for(idx=0; idx<coefw; ++idx, ++ptmp)
	{
	    t0 = *(ptmp+4);
	    t1 = *(ptmp+5) + *(ptmp+3);
	    t2 = *(ptmp+6) + *(ptmp+2);
	    t3 = *(ptmp+7) + *(ptmp+1);
	    tup = (t0>>8);
	    tlow = t0 - (tup<<8);
	    temp = high1[tlow] + (high1[tup]<<8);
	    if( t1 )
	    {
		tup = (t1>>8);
		tlow = t1 - (tup<<8);
		temp += high2[tlow] + (high2[tup]<<8);
	    }
	    tup = (t2>>8);
	    tlow = t2 - (tup<<8);
	    temp += high3[tlow] + (high3[tup]<<8);
	    if( t3 )
	    {
		tup = (t3>>8);
		tlow = t3 - (tup<<8);
		temp += high4[tlow] + (high4[tup]<<8);
	    }
	    *pDes++ =(Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefw) break;
	    ++ptmp;
	    t0 = *(ptmp+4);
	    t1 = *(ptmp+5) + *(ptmp+3);
	    t2 = *(ptmp+6) + *(ptmp+2);
	    t3 = *(ptmp+7) + *(ptmp+1);
	    t4 = *(ptmp+8) + *(ptmp);
	    tup = (t0>>8);
	    tlow = t0 - (tup<<8);
	    temp = low1[tlow] + (low1[tup]<<8);
	    tup = (t1>>8);
	    tlow = t1 - (tup<<8);
	    temp += low2[tlow] + (low2[tup]<<8);
	    if( t2 )
	    {
		tup = (t2>>8);
		tlow = t2 - (tup<<8);
		temp += low3[tlow] + (low3[tup]<<8);
	    }
	    tup = (t3>>8);
	    tlow = t3 - (tup<<8);
	    temp += low4[tlow] + (low4[tup]<<8);
	    if( t4 )
	    {
		tup = (t4>>8);
		tlow = t4 - (tup<<8);
		temp += low5[tlow] + (low5[tup]<<8);
	    }
	    *pDes++ =(Int16)((temp+ONEHALF)>>SCALEBITS);
	}
    }

/* collunm */
    for(col=0; col<coefw; ++col)
    {
/* copy image */
	pDes = linebuf;
	pSrc = coef + halfh * bpl;
	for(idx=0, row=col; idx<coefh; ++idx, row += bpl )
	{
	    *pDes ++ = *(coef + row);
	    if( ++idx >= coefh) break;
	    *pDes ++ = *(pSrc + row);
	}
	for(idx=0, row = 8; idx<4; ++idx, --row)
	    *(buffer+idx)= *(buffer+row);
	for(idx=1, row=coefh-1; idx<5; ++idx)
	    *(linebuf+idx+row) = *(linebuf+row-idx);

/* convelution */
	ptmp = buffer;
	for(idx=0, pDes = coef+col; idx<coefh; ++idx, pDes += bpl, ++ptmp)
	{
	    t0 = *(ptmp+4);
	    t1 = *(ptmp+5) + *(ptmp+3);
	    t2 = *(ptmp+6) + *(ptmp+2);
	    t3 = *(ptmp+7) + *(ptmp+1);
	    tup = (t0>>8);
	    tlow = t0 - (tup<<8);
	    temp = high1[tlow] + (high1[tup]<<8);
	    if( t1 )
	    {
		tup = (t1>>8);
		tlow = t1 - (tup<<8);
		temp += high2[tlow] + (high2[tup]<<8);
	    }
	    tup = (t2>>8);
	    tlow = t2 - (tup<<8);
	    temp += high3[tlow] + (high3[tup]<<8);
	    if( t3 )
	    {
		tup = (t3>>8);
		tlow = t3 - (tup<<8);
		temp += high4[tlow] + (high4[tup]<<8);
	    }
	    *pDes=(Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefh) break;
	    pDes += bpl;
	    ++ptmp;
	    t0 = *(ptmp+4);
	    t1 = *(ptmp+5) + *(ptmp+3);
	    t2 = *(ptmp+6) + *(ptmp+2);
	    t3 = *(ptmp+7) + *(ptmp+1);
	    t4 = *(ptmp+8) + *(ptmp);
	    tup = (t0>>8);
	    tlow = t0 - (tup<<8);
	    temp = low1[tlow] + (low1[tup]<<8);
	    tup = (t1>>8);
	    tlow = t1 - (tup<<8);
	    temp += low2[tlow] + (low2[tup]<<8);
	    if( t2 )
	    {
		tup = (t2>>8);
		tlow = t2 - (tup<<8);
		temp += low3[tlow] + (low3[tup]<<8);
	    }
	    tup = (t3>>8);
	    tlow = t3 - (tup<<8);
	    temp += low4[tlow] + (low4[tup]<<8);
	    if( t4 )
	    {
		tup = (t4>>8);
		tlow = t4 - (tup<<8);
		temp += low5[tlow] + (low5[tup]<<8);
	    }
	    *pDes=(Int16)((temp+ONEHALF)>>SCALEBITS);
	}	
    }

    NORMAL_RETURN;
}


/*
$Log: miwdtra.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.5  1998/11/30 19:40:36  xzhu
xzhu on Mon Nov 30 11:40:28 PST 1998

miwdtra.c -> /product/ipcore/ipshared/src/RCS/miwdtra.c,v

Added new functions.

Revision 1.4  1998/10/20 17:51:26  xzhu
xzhu on Tue Oct 20 10:50:52 PDT 1998

miwdtra.c -> /product/ipcore/ipshared/src/RCS/miwdtra.c,v

Create new algorithm (lift) invert transform.

Revision 1.3  1998/01/13 23:53:54  ddavies
miwdtra.c -> /product/ipcore/ipshared/src/RCS/miwdtra.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1997/09/17 22:48:07  xzhu
miwdtra.c -> /product/ipcore/ipshared/src/RCS/miwdtra.c,v

CommonCheck: Added #include wavelib.pub, since some changes made in huffman coding.

Revision 1.1  1996/10/16 00:02:52  xzhu
Initial revision


*/
