;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  mortemp2.asm:  template for morphology assembly code.

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C

; This is primarily a template for fast erosion code.  We'll add an OR
; operation so it can be used for dilation as well.  Just to make sure
; this stuff really works, this code is actually executable.  It does an
; erosion using the SE s_4_8_l as defined in thin4.se.  The operations
; that weren't used in s_4_8_l are defined after the last active code.

; Note that we use no local or global data references, all variables are
; referenced as offsets from the ebp register.  The C code that generates
; real assembly need not worry about relocable variables.

; The algorithm used here is quite different from the one used in the
; SPARC assembly code.  Instead of maintaining as many partial
; results as possible, we compute each complete result and store it.
; In the first version of this code, we shifted each source word once
; for each SE element.  That's a lot of relatively slow shift operations.
; This is especially true of the double shifts we need.  We thought
; we had to recompute every input value because the 386 doesn't have
; enough registers to hold either lots of inputs or lots of outputs.

; The SPARC algorithms broke the entire image into blocks, processing
; each block vertically.  This was a good compromise between efficient
; processing (processing the image vertically minimizes memory writes of
; partial results) and cache utilization (minimized by horizontal image
; reads).  In the first version of 386 algorithm, all inputs read for
; one dest word were read again as necessary for the next dest word.  It
; didn't matter whether we went horizontally or vertically, so we went
; horizontally because it was easier.

; The second version of 386 morphology saves the shifted versions of each
; source word in a buffer.  While we still calculate one destination
; word at a time, we can usually use preshifted input words, cutting out the
; 4 cycles used to do double shifts out of the inner loop.

; Given that we want to use pre-shifted inputs, do we move through the
; image horizontally or vertically?  If we moved horizontally (and computed
; one output word at a time), each horizontal move would require us to read
; a new input word for *each row* of the SE.  That's because each word is
; only 1 bit tall.

; If we move through the image vertically, we need to read only 3 new words
; for each move down.  That's because the words are wide enough to cover the
; maximum legal shift done in our accelerated code (32 bits).  So, we need
; to read the word directly below the previous ones and the words on either
; side of that.  This seems preferable to the extra reads horizontal
; movement would require.

; How do we address the buffer?  It will be at least as large as the
; SE, except each SE point will correspond to a (shifted) 32-bit input word.
; We'll process the image in vertical stripes one word wide.  That means
; we'll make a vertical pass through the rows of the array of shifted inputs
; for each output word.  We could move each shifted input word up one row
; in the array when it is used.  That generates a lot of writes.  It would
; be nicer to leave the words in place and use a moving pointer.

; We can't use one pre-allocated global array for the shifted inputs
; needed in all erosion operations because the code wouldn't be reentrant.
; Therefore, we must use a register to point to an array allocated by the
; calling procedure.  We know the width of the array, so can use constant
; displacements to address shifted inputs within a row.  It would be very
; nice to use just register+displacment addressing, it adds no cycles.
; However, we need some way to detect when the buffer wraps around.

; Or do we?  We'll process a limited number of words vertically before
; starting the next stripe.  This is done to make reasonable use of the
; processor's cache.  Say the calling routine allocates a buffer that's
; so large that we'll never overflow it.  If we process n lines vertically and 
; have an H by V structure element, we'll need a buffer that can hold
; H*(V+n)*4 bytes.  Typical values of H and V are under 10.  Typical
; processing block height is 64.  That's a 10*74*4 = 3KB block.  Not too bad,
; especially when compared to the 1MB required for the image.  Say
; the SE were a maximum 64 x 64.  That requires a 64*128*4 = 32KB block.
; Seems ok, especially because it saves a compare and branch on every row
; being processed.  This is a ridiculously large strel.

; In fact, since the rows don't wrap around, we can address
; the entire array of shifted inputs with displacements relative to a
; single pointer.  The pointer points to the word in the upper left corner
; of the array.  Zowee.  We can do AND or OR ops in 2 cycles.  ANDN ops
; take 3 cycles since we have to invert the input before the AND.  Very nice.
; There's a somewhat more subtle benefit from eliminating extra address
; arithmetic.  The 486 and P5 have pipeline stalls when addresses are modified
; immediately after being used.  By addressing the entire input array
; relative to a single unchanging pointer, we eliminate these stalls.

; What do we do in the degenerate cases of purely horizontal and purely
; vertical strels?  The purely horizontal strels use shifted inputs to
; compute each output word, but don't use those shifted inputs to compute
; any other output word.  We'll just do the shifting, but won't bother
; storing the results in an array.

; Purely vertical strels don't use shifted inputs.  However, they do reuse
; input values for multiple output words.  We could continue to address the
; input words in place, skipping source bytes/line to get to each new one.
; This is a lot of address arithmetic in the inner loop.  For purely vertical
; strels, we'll continue to use the array of inputs so we can minimize
; address arithmetic.

; In the SPARC algorithms, we went to a great deal of trouble to use double
; word (64 bit) write operations.  Writes on an SS2 are slow, so it was
; beneficial.  The 386/486 has relatively fast 32 bit writes and no 64 bit
; writes, so we will discard that complication.

; Like the SPARC algorithms, we will have restrictions on the size of
; structure elements we can handle.  The distance from the SE's center to
; the edge of the SE (dimensions cx, cy, (nx-cx) and (ny-cy)) must all be
; less than or equal to 32.  This is the width and height of the pad all
; around the image.

; To specify the algorithm, define the ShiftRight(L, R, d) operator.  It
; returns the least significant 32 bits of the result of shifting the 64 bit
; quantity L,,R right by d bits.  Similarly, ShiftLeft(L, R, d) returns the
; most significant 32 bits of the result of shifting the 64 bit quantity
; L,,R left by d bits.

; Define the array of shifted inputs as InputCache.

; The basic algorithm is ("word" = 32 bits):
;    1.	Start in the upper left corner of the current block in the
;	output image.
;    2. Given that output word, read the word of input image cy rows
;	above it into Input.
;    3. If cx > 0, read the word to the left of the input word into
;	InputLeft.
;    4. For each non-empty column i of the strel to the left of center,
;	write ShiftRight(InputLeft, Input, cx-i) into location (0,i) of
;	InputCache.  This is most easily done by working from the center's
;	column (no shift) towards the left.
;    5. If the center column of the strel is not empty, write Input
;	into location (0, cx) of the array of shifted inputs.
;    6. If cx < nx-1, read the word to the right of input into InputRight.
;    7. For each non-empty column i of the strel to the right of center,
;	write ShiftLeft(Input, InputRight, i-cx) into location (0,i) of
;	InputCache.
;    8. Repeat steps 2 through 8 for each row of the strel except for
;	the last, moving down one image row and one row in InputCache
;	on each iteration.
;
;	We have now preloaded the InputCache.  We can start processing.
;
;    9. Process one row of inputs from the InputCache from left to right.
;	Start with the top row.  Each InputCache element (i,j) corresponds
;	to SE location (i,j). Combine each input word with the current
;	output word using the corresponding SE element as a guide.  There
;	are five possibilities.  They are:
;	   a. Don't Care: do nothing
;	   b. partial result currently empty, current SE element is a HIT:
;		move current (possibly shifted) input to partial result.
;	   c. partial result currently empty, current SE element is a MISS:
;		move inverse of (shifted) input to partial result.
;	   d. partial result not empty, current SE element is a HIT:
;		AND (shifted) input word with partial result.
;	   e. partial result not empty, current SE element is a MISS:
;		AND inverse of (shifted) input word with partial result.
;   10. We have now processed all but the bottom row of the SE.  Read the
;	word from the input image that's ny-1-cy rows below the output
;	word.
;   11. Proceed as we did in steps 3 through 7, with two exceptions.  First,
;	combine each shifted input with the output before storing the input
;	to the InputCache.  Second, if this is a purely horizontal strel,
;	don't bother with the store.
;   12. We have how processed one output word.  Write it to the output image.
;	Adjust the pointer to the InputCache to point down one row.
;   11. Repeat steps 9 through 12 for each remaining word in the current
;	vertical output image stripe.

; The real algorithm will have some optimizations.  We will simply skip the
; Don't Care states.  We'll also put the loop controls at the top of the
; loops instead of at the bottom.  This makes address calculation easier
; when generating the code automatically.

; After immense sillyness, I found out that the WATCOM Linker's PHARLAP format
; causes the linker to produce code in the FLAT memory model.  All that stuff
; in the Microsoft Assembler manual (and in the WATCOM
; compiler manual and in the flags we give to the WATCOM compiler) about
; small models is inoperative.  So, that makes things easy.  We write code as
; though we had the small model and everything's peachy.  In the final system,
; we'll be able to generate code relative to the data segment and execute it
; with respect to the code segment.  This will work because both segments have
; the same base (gee, I hope the code segment doesn't have a smaller limit!).

; It didn't.  All the calling and returning worked fine.

_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: mortemp2.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0

; To get the Microsoft assembler to believe the following labels are PUBLIC,
; we have to declare them with double colons, e.g. "erT2Invert::".
	PUBLIC	erT2Start, erT2PreLoadDec, erT2PreLoadInpCac
	PUBLIC	erT2LoadInputCenter, erT2LoadInputLeft, erT2LoadInputRight
	PUBLIC	erT2ShiftInputRight, erT2ShiftInputLeft, erT2IncInputAddr
	PUBLIC	erT2CacheMiss, erT2NonCacheMiss, erT2NonCacheMissRestore
	PUBLIC	erT2NonCacheHitAnd, erT2NonCacheHitOr, erT2NonCacheInit
	PUBLIC	erT2LoadDestAddr, erT2BottomLoop, erT2LoopJmp, erT2Return
	PUBLIC	erT2CacheStore0, erT2CacheStoreShort, erT2CacheStoreLong
	PUBLIC	erT2CacheLoadEAX0, erT2CacheLoadEAXShort, erT2CacheLoadEAXLong
	PUBLIC	erT2NotEAX
	PUBLIC	erT2CacheLoadEBX0, erT2CacheLoadEBXShort, erT2CacheLoadEBXLong
	PUBLIC	erT2CacheAnd0, erT2CacheAndShort, erT2CacheAndLong
	PUBLIC	erT2CacheOr0, erT2CacheOrShort, erT2CacheOrLong
	PUBLIC	erT2IncCachePtrShort, erT2IncCachePtrLong
	PUBLIC	erT2SingleRightShift, erT2SingleLeftShift, erT2DoubleLeftShift1
	PUBLIC	erT2Nop

	PUBLIC	startBlkLen2, preLoadDecBlkLen2, preLoadInpCacBlkLen2
	PUBLIC	loadInputCenterBlkLen2, loadInputLeftBlkLen2
	PUBLIC	loadInputRightBlkLen2
	PUBLIC	shiftInputRightBlkLen2, shiftInputLeftBlkLen2
	PUBLIC	incInputAddrBlkLen2, cacheMissBlkLen2
	PUBLIC	nonCacheMissBlkLen2, nonCacheMissRestoreBlkLen2
	PUBLIC	nonCacheHitOrBlkLen2, nonCacheHitAndBlkLen2
	PUBLIC	nonCacheInitBlkLen2, loadDestAddrBlkLen2
	PUBLIC	bottomLoopBlkLen2, jumpBlkLen2, returnBlkLen2
	PUBLIC	cacheStore0BlkLen2, cacheStoreShortBlkLen2
	PUBLIC	cacheStoreLongBlkLen2, cacheLoadEAX0BlkLen2
	PUBLIC	cacheLoadEAXShortBlkLen2, cacheLoadEAXLongBlkLen2
	PUBLIC	cacheNotEAXBlkLen2, cacheLoadEBX0BlkLen2
	PUBLIC	cacheLoadEBXShortBlkLen2, cacheLoadEBXLongBlkLen2
	PUBLIC	cacheAnd0BlkLen2 , cacheAndShortBlkLen2
	PUBLIC	cacheAndLongBlkLen2 , cacheOr0BlkLen2
	PUBLIC	cacheOrShortBlkLen2 , cacheOrLongBlkLen2
	PUBLIC	incCachePtrShortBlkLen2 , incCachePtrLongBlkLen2
	PUBLIC	singleLeftShiftBlkLen2, singleRightShiftBlkLen2
	PUBLIC	doubleLeftShift1BlkLen2, nopBlkLen2

_DATA	ENDS


; Register Assignment:
; EAX:	scratch, output word
; EBX:	input word, possibly shifted
; ECX:	input word's shift input for double shifts, address of output words
; EDX:	holds dest addr
; EBP:	points to vars on stack
; ESI:	address of input words
; EDI:	address of InputCache
; ESP:	stack pointer

; Local Variables:
; srcAddr:	address of top source word
; destAddr:	address of top dest word
; inputCacheAddr: address of beginning of input cache.
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image
; destPitch:	number of bytes to be added to address to move down one line
;		in destination (output) image

InputCachePitch	EQU	12	; there are 12 bytes/row of input cache

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT	BYTE PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name erS248l.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by ferode.c

erT2Start::
	PUBLIC	erS248l
erS248l	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI,\
	srcAddr:DWORD, destAddr:DWORD, srcPitch:DWORD, destPitch:DWORD,\
	endDstAddr:DWORD, inputCacheAddr:DWORD

; create local variables on the stack automatically if we need them (we don't)
;	LOCAL	curDstAddr:DWORD

; All registers used are also pushed on the stack automatically.

; Preload the input cache.  Each location in the SE corresponds to one
; preshifted input word.
	mov	esi, srcAddr		; load input address
erT2StartEnd::

; move srcAddr up by cy lines.
erT2PreLoadDec::
	sub	esi, srcPitch		; do address processing early
erT2PreLoadDecEnd::

erT2PreLoadInpCac::
	mov	edi, inputCacheAddr	; load address of input cache
erT2PreLoadInpCacEnd::

erT2LoadDestAddr::
	mov	edx, destAddr
erT2LoadDestAddrEnd::

; get word directly above center
erT2LoadInputCenter::
	mov	ebx, [esi]
erT2LoadInputCenterEnd::
; get word to the left of that in preparation for the right shift
erT2LoadInputLeft::
	mov	ecx, [esi - 4]
erT2LoadInputLeftEnd::
; Even though this strel doesn't have any active elements in the center's
; column, we'll store the unshifted word for practice.
	mov	[edi + 4], ebx

; store right shifted input word in input cache.
erT2ShiftInputRight::
	shrd	ebx, ecx, 1
erT2ShiftInputRightEnd::
	mov	[edi], ebx

; get the unshifted word and the word to the right in preparation for a
; left shift.
	mov	ebx, [esi]
erT2LoadInputRight::
	mov	ecx, [esi + 4]
erT2LoadInputRightEnd::

; address the next row early
erT2IncInputAddr::
	add	esi, srcPitch
erT2IncInputAddrEnd::

; Store the left shifted word in the last entry of the first row of the
; input cache.
erT2ShiftInputLeft::
	shld	ebx, ecx, 1
erT2ShiftInputLeftEnd::
	mov	[edi + 2*4], ebx

; Repeat the performance for the next input row.
	mov	ebx, [esi]
	mov	ecx, [esi - 4]
	mov	[edi + InputCachePitch + 4], ebx
	shrd	ebx, ecx, 1
	mov	[edi + InputCachePitch], ebx
	mov	ebx, [esi]
	mov	ecx, [esi + 4]
	add	esi, srcPitch
	shld	ebx, ecx, 1
	mov	[edi + InputCachePitch + 2*4], ebx

; We've now preloaded the first two rows of the input cache.  We enter
; the main loop that processes the output using stored data from the
; first two rows, then creates the third row and applies it to the
; output at the same time.

erT2Loop::

; The strel looks like
; 0 0 H
; M 0 H
; M 0 H
; where 0 = Don't care, M = Miss and H = Hit.  There's only one operation
; to do in the first row.  Since this is the first operation we've seen,
; do a copy.  Since it's a hit instead of a miss, do a non-inverted copy.
	mov	eax, [edi + 2*4]

; There are two operations in the second row.  The first is Miss, so invert
; the input before ANDing.
	mov	ebx, [edi + InputCachePitch]
erT2CacheMiss::
	not	ebx
	and	eax, ebx
erT2CacheMissEnd::
; the second is a hit, so just do the AND
	and	eax, [edi + InputCachePitch + 2*4]

; So much for the first two rows.  Create the next row for the input cache
; and apply the words as needed.  Note that this particular strel doesn't
; use the center's column, so we don't bother storing the un-shifted input.
; Note that esi still points to the next input row.
	mov	ebx, [esi]
	mov	ecx, [esi - 4]
	shrd	ebx, ecx, 1
	mov	[edi + 2*InputCachePitch], ebx
; Process the miss.  Note that we don't need to move the data before
; inverting it because we're not going to use ebx again without reloading.
; A version of this code that inverts ebx, does the AND and re-inverts ebx
; appears later individual templates.
erT2NonCacheMiss::
	not	ebx
	and	eax, ebx
erT2NonCacheMissEnd::

	mov	ebx, [esi]
	mov	ecx, [esi + 4]
	add	esi, srcPitch			; do addr mods early
	shld	ebx, ecx, 1

; Complete storing the last input cache entry.
	mov	[edi + 2*InputCachePitch + 2*4], ebx
; Process the hit in the lower left corner of the strel.  This is trivial.
erT2NonCacheHitAnd::
	and	eax, ebx
erT2NonCacheHitAndEnd::

; Point down to the next line of cached input data
	add	edi, InputCachePitch

; The output word has been fully generated, store it.
erT2BottomLoop::
	mov	[edx], eax
	add	edx, destPitch		; point to next output word
; See if we've reached the bottom of the loop.
	cmp	edx, endDstAddr
erT2BottomLoopEnd::
	jmp	erT2LoopJmp		;  jump around big gap.  We want the
					; assembler to generate a near jump,
					; since our generated code might
					; require a jump of more than -128 bytes

	DB	256 DUP (?)

erT2LoopJmp::
	jnz	erT2Loop		; if more words to do in this row,
					; do them.
erT2LoopJmpEnd::

erT2Return::
; We're done, so return.  The assembler turns this into a PILE of code!
	ret
erT2ReturnEnd::


; Other interesting pieces of template
erT2CacheStore0::
	mov	[edi], ebx			; requires no address disp
erT2CacheStore0End::

erT2CacheStoreShort::
	mov	[edi + 4], ebx			; requires byte addr disp
erT2CacheStoreShortEnd::

erT2CacheStoreLong::
	mov	[edi + 4000], ebx		; requires dword addr disp
erT2CacheStoreLongEnd::


erT2CacheLoadEAX0::
	mov	eax, [edi]
erT2CacheLoadEAX0End::

erT2CacheLoadEAXShort::
	mov	eax, [edi + 4]
erT2CacheLoadEAXShortEnd::

erT2CacheLoadEAXLong::
	mov	eax, [edi + 4000]
erT2CacheLoadEAXLongEnd::

erT2NotEAX::
	not	eax
erT2NotEAXEnd::

erT2CacheLoadEBX0::
	mov	ebx, [edi]
erT2CacheLoadEBX0End::

erT2CacheLoadEBXShort::
	mov	ebx, [edi + 4]
erT2CacheLoadEBXShortEnd::

erT2CacheLoadEBXLong::
	mov	ebx, [edi + 4000]
erT2CacheLoadEBXLongEnd::

erT2CacheAnd0::
	and	eax, [edi]
erT2CacheAnd0End::

erT2CacheAndShort::
	and	eax, [edi + 4]
erT2CacheAndShortEnd::

erT2CacheAndLong::
	and	eax, [edi + 4000]
erT2CacheAndLongEnd::

erT2CacheOr0::
	or	eax, [edi]
erT2CacheOr0End::

erT2CacheOrShort::
	or	eax, [edi + 4]
erT2CacheOrShortEnd::

erT2CacheOrLong::
	or	eax, [edi + 4000]
erT2CacheOrLongEnd::

erT2NonCacheHitOr::
	or	eax, ebx
erT2NonCacheHitOrEnd::

erT2NonCacheInit::
	mov	eax, ebx
erT2NonCacheInitEnd::

erT2IncCachePtrShort::
	add	edi, 4
erT2IncCachePtrShortEnd::

erT2IncCachePtrLong::
	add	edi, 4000
erT2IncCachePtrLongEnd::


erT2NonCacheMissRestore::
	not	ebx
	and	eax, ebx
	not	ebx
erT2NonCacheMissRestoreEnd::

erT2SingleRightShift::
	shr	ecx, 10			; any shift amount requiring an
					; immediate byte for the shift dist.
erT2SingleRightShiftEnd::

erT2SingleLeftShift::
	shl	ecx, 10			; any shift amount requiring an
					; immediate byte for the shift dist.
erT2SingleLeftShiftEnd::

erT2DoubleLeftShift1::
	add	ecx, ecx		; use two adds (2 cycles) instead of
	adc	ebx, ebx		; two shifts (4 cycles).
erT2DoubleLeftShift1End::

erT2Nop::
	nop
erT2NopEnd::


erS248l ENDP

_TEXT	ENDS				; End the segment


_DATA	SEGMENT PUBLIC 'DATA'

; Lengths of the code blocks.  Put these down here so the disassembler
; output is easier to read (starts with code at offset 0).
startBlkLen2		DWORD	erT2StartEnd - erT2Start
preLoadDecBlkLen2	DWORD	erT2PreLoadDecEnd - erT2PreLoadDec
preLoadInpCacBlkLen2	DWORD	erT2PreLoadInpCacEnd - erT2PreLoadInpCac
loadInputCenterBlkLen2	DWORD	erT2LoadInputCenterEnd - erT2LoadInputCenter
loadInputLeftBlkLen2	DWORD	erT2LoadInputLeftEnd - erT2LoadInputLeft
loadInputRightBlkLen2	DWORD	erT2LoadInputRightEnd - erT2LoadInputRight
shiftInputRightBlkLen2	DWORD	erT2ShiftInputRightEnd - erT2ShiftInputRight
shiftInputLeftBlkLen2	DWORD	erT2ShiftInputLeftEnd - erT2ShiftInputLeft
incInputAddrBlkLen2	DWORD	erT2IncInputAddrEnd - erT2IncInputAddr
cacheMissBlkLen2	DWORD	erT2CacheMissEnd - erT2CacheMiss
nonCacheMissBlkLen2	DWORD	erT2NonCacheMissEnd - erT2NonCacheMiss
nonCacheMissRestoreBlkLen2 DWORD erT2NonCacheMissRestoreEnd - erT2NonCacheMissRestore
nonCacheHitAndBlkLen2	DWORD	erT2NonCacheHitAndEnd - erT2NonCacheHitAnd
nonCacheHitOrBlkLen2	DWORD	erT2NonCacheHitOrEnd - erT2NonCacheHitOr
nonCacheInitBlkLen2	DWORD	erT2NonCacheInitEnd - erT2NonCacheInit
loadDestAddrBlkLen2	DWORD	erT2LoadDestAddrEnd - erT2LoadDestAddr
bottomLoopBlkLen2	DWORD	erT2BottomLoopEnd - erT2BottomLoop
jumpBlkLen2		DWORD	erT2LoopJmpEnd - erT2LoopJmp
returnBlkLen2		DWORD	erT2ReturnEnd - erT2Return
cacheStore0BlkLen2	DWORD	erT2CacheStore0End - erT2CacheStore0
cacheStoreShortBlkLen2	DWORD	erT2CacheStoreShortEnd - erT2CacheStoreShort
cacheStoreLongBlkLen2	DWORD	erT2CacheStoreLongEnd - erT2CacheStoreLong
cacheLoadEAX0BlkLen2	DWORD	erT2CacheLoadEAX0End - erT2CacheLoadEAX0
cacheLoadEAXShortBlkLen2 DWORD	erT2CacheLoadEAXShortEnd - erT2CacheLoadEAXShort
cacheLoadEAXLongBlkLen2	DWORD	erT2CacheLoadEAXLongEnd - erT2CacheLoadEAXLong
cacheNotEAXBlkLen2	DWORD	erT2NotEAXEnd - erT2NotEAX
cacheLoadEBX0BlkLen2	DWORD	erT2CacheLoadEBX0End - erT2CacheLoadEBX0
cacheLoadEBXShortBlkLen2 DWORD	erT2CacheLoadEBXShortEnd - erT2CacheLoadEBXShort
cacheLoadEBXLongBlkLen2	DWORD	erT2CacheLoadEBXLongEnd - erT2CacheLoadEBXLong
cacheAnd0BlkLen2	DWORD	erT2CacheAnd0End - erT2CacheAnd0
cacheAndShortBlkLen2	DWORD	erT2CacheAndShortEnd - erT2CacheAndShort
cacheAndLongBlkLen2	DWORD	erT2CacheAndLongEnd - erT2CacheAndLong
cacheOr0BlkLen2		DWORD	erT2CacheOr0End - erT2CacheOr0
cacheOrShortBlkLen2	DWORD	erT2CacheOrShortEnd - erT2CacheOrShort
cacheOrLongBlkLen2	DWORD	erT2CacheOrLongEnd - erT2CacheOrLong
incCachePtrShortBlkLen2	DWORD	erT2IncCachePtrShortEnd - erT2IncCachePtrShort
incCachePtrLongBlkLen2	DWORD	erT2IncCachePtrLongEnd - erT2IncCachePtrLong
singleRightShiftBlkLen2	DWORD	erT2SingleRightShiftEnd - erT2SingleRightShift
singleLeftShiftBlkLen2	DWORD	erT2SingleLeftShiftEnd - erT2SingleLeftShift
doubleLeftShift1BlkLen2	DWORD	erT2DoubleLeftShift1End - erT2DoubleLeftShift1
nopBlkLen2		DWORD	erT2NopEnd - erT2Nop

; all locals are on the stack
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END

IF 0				; These are all comments

COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: mortemp2.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.7  1998/01/13 23:57:56  ddavies
mortemp2.asm -> /product/ipcore/ipshared/src/RCS/mortemp2.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.6  1996/09/17 00:45:14  ddavies
ddavies on Mon Sep 16 17:45:14 PDT 1996

mortemp2.asm -> /product/ipcore/ipshared/src/RCS/mortemp2.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.5  1996/09/16 00:48:48  ddavies
ddavies on Sun Sep 15 17:47:09 PDT 1996

mortemp2.asm -> /product/ipcore/ipshared/src/RCS/mortemp2.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA segment; 2) don't declare DGROUP when assembling for NT; 3) ASSUME
DS is FLAT when assembling for WinNT.

Revision 1.4  1995/09/15 21:47:20  ddavies
ddavies on Fri Sep 15 14:47:19 PDT 1995

mortemp2.asm -> /product/ipcore/ipshared/src/RCS/mortemp2.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

# Revision 1.3  1993/11/05  01:51:22  ddavies
# mortemp2.asm -> /product/ipcore/ipshared/src/RCS/mortemp2.asm,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

^

ENDIF

