;*****************************************************************
;*  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  mortemp3.asm:  template for MMX morphology assembly code.

	.586		; compile code for a Pentium processor

;	OPTION	LANGUAGE:C

INCLUDE     iammx.inc

_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: mortemp3.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; This is primarily a template for fast MMX erosion code.  We'll add an OR
; operation so it can be used for dilation as well.  This code isn't really
; executable, but it has lots of nice fragments that C routines compose into
; real erosion and dilation routines.
;
; In subsequent comments, I'll refer to 8 bit things as bytes, 16 bit
; things as halfwords, 32 bit things as words and 64 bit things as
; doublewords, or dwords.  Intel thinks in terms of bytes, words, doublewords
; and quadwords, but tough.
;
; Note that we use no local or global data references, all variables are
; referenced as offsets from the ebp register.  The C code that generates
; real assembly need not worry about relocable variables.
;
; The method we use for MMX code is very different from the SPARC
; method and somewhat different from the method used in regular 386 code.
;
; Before I start talking about details, let me give an overall view of
; erosion and dilation.
;
; In binary morphology, we have an input image, an output image and
; a structure element, or SE.  The SE is itself an array which contains
; a HIT, a MISS or a don't-care term at each location.  Each location
; in an SE is called an "SE element".  An active SE element is one with
; a HIT or MISS, as opposed to a don't-care term.  An active SE column is
; one that contain at least one active SE element.
;
; One of the SE elements is also chosen as the "center".  It need not be
; the element that's actually at the midpoint of the SE.
;
; In the simplest (and least efficient) implementation of erosion, one
; moves the SE around on the input image.  At every spot where the pattern
; is matched on the input image, a 1 is written on the output image at the
; location corresponding to the current location of the SE's center.
; You can think of the output as a record of all the places that the pattern
; specified by the SE was matched.  The pattern may be quite specific and
; used to detect particular blobs in the input.  It may contain a lot of
; don't-care terms, so only matches loosely defined shapes or textures.
; It may also be quite small.  In this case, it matches portions
; of all shapes.  For example, one might detect all left edges of
; things this way.
;
; Dilation is almost the opposite.  At all places where the center is over
; a 1 in the input image, the HITs of the SE are ORed into the output image.
; This generally extends the boundaries of things (often things that were
; eroded with the same SE) in a structured way.
;
; Moving the SE around to every position in the input is not very efficient.
; However, notice that results do not depend on each other so they can be
; calculated in parallel.  The cononical way to implement this is to combine
; shifted versions of the entire original input image to produce the output.
; For example, say the strel contained only two elements, a HIT to the left
; of a MISS, with the center over the MISS.  We could AND the image shifted
; right (corresponds to the HIT one position to the left of center) with
; the inverse of the original image (corresponds to the MISS at the center)
; to produce the output.  An output pixel will be set if and only if it
; was a 0 in the input image and there was a 1 immediately to the left
; of it.  In this view, the SE is really a little program that dictates
; a series of shifts and logic operations.  We have C code that compiles
; these little programs into machine code.  This file contains code
; fragments that act as building blocks when on an MMX machine.
;
;				*	*	*
;
; The SPARC algorithms broke the entire image into short blocks, processing
; each block vertically.  We moved horizontally from one block to the next.
; This was a good compromise between efficient processing (processing
; the image vertically minimizes memory writes of partial results) and
; cache utilization (minimized by horizontal image reads).  In the first
; version of regular 386 algorithm, all inputs read for one dest word
; were read again as necessary for the next dest word.  It didn't
; matter whether we went horizontally or vertically, so we went
; horizontally because it was easier.
;
; In that first version, we also shifted each source word once
; for each active SE element.  These are really double shifts because
; each shifted word really contains bits from two source words.
; That's a lot of slow shift operations.  Double shifts are still
; slow in MMX.  We thought we had to recompute every input value
; because the 386 doesn't have enough registers to hold lots
; of partial outputs.  MMX adds 8 registers, but that's still
; not a lot of values.
;
; The second version of 386 morphology saved the shifted versions of each
; source word in a temporary buffer.  We saved a shifted source word
; for each SE column that had a HIT or MISS in it.  While we still
; calculated one destination word at a time, we could usually use
; preshifted input words, cutting the 5 (Pentium) cycles used to
; do double shifts out of the inner loop.
;
; Huh?  Don't you have to do the shifts anyway?  Well, yes, but we only
; have to shift a word once for all the SE elements in a given column
; instead of repeating the shift for each SE element in that column.
; This optimization clearly works best if there are lots of active
; elements in an SE column.
;
; The MMX code follows this idea, with a twist.  Since we have more
; MMX registers, we compute 7 destination words at a time, again using
; preshifted inputs.  The eighth MMX register holds one of the inputs.
; Having lots of partial results available at once allows us to be
; more efficient.  It usually allows us to do two AND ops per cycle
; in a Pentium instead of only one.  The total number of loads and
; stores and ANDs is the same whether we compute 1 result at a time or
; multiple results.  Computing multiple results lets us do a better job
; of filling clock cycles with 2 operations instead of 1.
;
; Given that we want to use pre-shifted inputs, do we move through the
; image horizontally or vertically?  If we moved horizontally (and computed
; one output word at a time), each horizontal move would require us to read
; a new input word for *each row* of the SE.  That's because each word is
; only 1 bit tall.
;
; If we move through the image vertically, we always read one dword and
; two words for each move down.  That's because the words are wide enough
; to cover the maximum legal shift done in our accelerated code (32 bits).
; So, we need to read the dword directly below the previous one and the
; words on either side of that.  This seems preferable to the extra
; reads horizontal movement would require.
;
; All this means we'll go back to the SPARC method of breaking the
; image into relatively short blocks.  Each block will be processed
; vertically.  To avoid trashing the cache, we'll move horizontally from
; block to block.
;
; How large is the buffer of preshifted inputs?  It will be at
; least as wide as the SE, except each active SE column will correspond
; to a (wordswapped, shifted) 64-bit input dword.  Because inverting
; words to process MISS elements is a pain when doing the ANDs, we'll
; actually store both positive and inverted versions of the shifted
; inputs as needed.
;
; Editorial Comment: Can you believe the idiocy of the MMX designers in
; designing the pandn opcode?  Instead of inverting the source, then
; ANDing with the dest, it inverts the dest, then ANDs with the source.
; Gag me.  This makes it useless for combining a dest with a bunch of
; inverted and non-inverted masks.  End of Editorial Comment
;
; We'll process the image in blocks that are one dword wide.  That means
; we'll make a vertical pass through the rows of the shifted input buffer
; for each output dword.  After processing one output dword, there's
; a row of shifted inputs we'll never use again.  We'll also need a
; new row.  We could move each shifted input word up one row in the buffer
; and insert one new row of shifted dwords at the bottom.  That
; generates a lot of writes.  It would be nicer to leave the buffer
; dwords in place and use a moving pointer.
;
; We can't use one pre-allocated global array for the shifted inputs
; needed in all morphology operations because the code wouldn't be reentrant.
; Therefore, we must use a register to point to an array allocated by the
; calling procedure.  We know the width of the array, so can use constant
; displacements to address shifted inputs within a row.  It would be very
; nice to use just register+displacment addressing, it adds no cycles.
; However, we need some way to detect when the buffer wraps around.
;
; Or do we?  We'll process a limited number of output dwords vertically before
; starting the next block.  This is done to make reasonable use of the
; processor's cache.  The buffer of preshifted inputs is reused for each
; block.  Say the calling routine allocates a buffer that's
; so large that we'll never overflow it.  If we process n lines vertically and 
; have an H by V structure element, we'll need a buffer that can hold
; H * (V+n) * 8 bytes.  Typical values of H and V are under 10.  Typical
; processing block height is 64.  That's a 10 * 74 * 8 = 6 KB block.
; Not too bad, especially when compared to the 1MB required for the image.  Say
; the SE were a maximum 64 x 64.  That requires a 64 * 128 * 8 = 64 KB block.
; Seems ok on an MMX machine.  This is a ridiculously large strel.
;
; In fact, we can address the entire array of shifted inputs with
; displacements relative to a single pointer.  The pointer points to
; the shifted dword in the upper left corner of the buffer.
; Well, if the generated code processes the output words in a loop,
; it can adjust the register to point to first preshifted input needed.
; Zowee.  We can do AND or OR ops in 1 cycle.  Well, actually,
; if we process two output lines at a time, we can do two ANDs or two ORs in
; a cycle.  Since there's no AND-with-inverse-of-source op, we'll store
; the inverse of the shifted words in the buffer as needed.  Very nice.
;
; There's a somewhat more subtle benefit from eliminating extra buffer address
; arithmetic.  The 486 and Pentium have pipeline stalls when addresses
; are used immediately after being modified.  By addressing the entire
; input array relative to a single unchanging pointer, we eliminate
; these stalls.
;
; In the plain 386 code, we combined the code that did the shifting with
; the code that did the ANDing.  We would start creating a new output
; word by ANDing a bunch of preshifted words together.  These words
; represented all the SE elements except for those on the last SE row.
; We'd then create one new preshifted input row, put it into the buffer
; and combine any of these new elements with the current output word
; as necessary.  The result would be a new output word and one new row
; of preshifted input words.
;
; In MMX code, instruction pairing is important.  It's really a good idea
; to do two things at once so you can keep both execution units occupied
; on every clock cycle.  It's easier to design code that does this
; if you're basically doing the same thing on two data streams.  That's
; what we'll do here, we'll compute two rows of preshifted inputs in
; parallel.
;
; We're really doing double shifts.  Each MMX double shift requires two input
; registers plus a scratch register.  If you figure out a fast scheme that
; uses fewer registers, let me know.  Anyway, doing two of these at
; a time uses up 2 * 3 = 6 MMX registers.  There are only 8 MMX registers.
;
; Well, we could still integrate the shifting with the output word production.
; We'd produce two output dwords at a time in the remaining two MMX registers.
; We don't want to do that.  The reason is subtle, as we'll try to explain next.
;
; We want to do two AND or OR operations per clock cycle.  Since you can't do
; two MMX operations with memory in the same cycle (!), at least one of
; the source operands has to be in a register.  That means we have to
; do an MMX read to fetch that operand.  If there are only two output
; words being produced at a time, the word fetched can be used at most
; twice, often only once.  If we produce more output dwords in parallel,
; we can almost always amortize the cost of that extra fetch over more
; AND ops.  Since there are fewer reads per useful op, the code runs faster.
;
; Therefore, in the MMX implementation, we'll separate the code that
; produces shifted inputs from the code that does the ANDing and ORing.
; This lets us use lots of registers for each task.  For the sake of
; simplicity, all of the shifted inputs needed by the block are
; produced first, then all of the outputs are produced.  Note
; that we still have two spare registers during the shifting stage.
; We'll fill one of them with all 1's, using that to invert the shifted
; inputs as needed.  As stated above, both positive and inverted inputs
; are stored in the shifted input buffer.  We can produce the inverted
; input in one operation and write it with one more.  If we wait until
; we need the inverted inputs to invert them, we need to reserve two MMX
; registers for inputs (if we're going to do two ANDs in parallel
; eventually) instead of one and need to read both of them instead of
; just 1 and need to XOR both with a word of all 1's.  All told, a bummer.
;
;				*	*	*
;
; We need to say something about wordswapping.  When a 64-bit dword is
; read from an image, the 32 bits on the left ends up in the *LS* end
; of the word since it has the lower address.  If we're going to shift
; the inputs, we'd really like to swap the two 32-bit halves of the
; input first.  Otherwise, a double shift requires us to insert bits
; into the ends of the two 32 bit halves independently (requiring at least two
; operations) instead of into the end of one 64-bit thing.  So, we
; wordswap the input words before shifting.
;
; Sometime before the output word is written out, its bits must be wordswapped
; back.  We could do this by re-wordswapping all the preshifted input dwords.
; That's a lot of shifting.  We'll just wordswap each the output word before
; writing it.
;
; What do we do in the degenerate cases of purely horizontal and purely
; vertical strels?  For purely vertical strels, no shifting is needed.  Since
; no shifting is done, we can skip the wordswapping, so we do.  Since
; there's no shifting, do we need the buffer of preshifted inputs?
; Well, yeah, we do.  We buffer them for two reasons.  First, it provides
; a place to store both inverted and non-inverted versions of the input
; words if needed.  Second, the preshifted input buffer is more easily
; addressed than the original image during the AND/OR phase.  Its geometry
; is known by the code that generates the strel, so we can use displacement
; addressing instead of having to repeatedly add the input images's 
; bytes/line to an address register.  Also, using the buffer in this
; case just makes all the cases more uniform.
;
; Purely horizontal strels really are a special case.  Since there's no
; vertical component, each shifted input dword is used exactly once.  It
; would be nicely uniform to store each shifted dword into the buffer,
; then read it out before using it that one time, but it's too ugly.
; Remember that we calculate two rows of shifted inputs in parallel.
; We have two MMX registers left over.  In the purely horizontal case, we
; use the two MMX registers to hold the outputs.  As each shifted input
; is created, it's either ANDed or inverted and ANDed with the corresponding
; output word.  Since we can't use one of the registers to hold the all
; 1's mask used to do inverting, we kicked the mask out to memory.  There's
; a word of 1's kept in the data segment.  Since this piece of template
; code gets loaded, the loader will fix up the pointer of the template
; opcode that points to that mask word.  Therefore, we can just copy that
; opcode, address and all, to each piece of new machine code.
;
; Since the purely horizontal case doesn't use the preshifted input buffer,
; it could process the image in vertical blocks or could just process
; entire scan lines at a time.  We choose to process the image in vertical
; blocks purely for the sake of uniformity.  If nothing else, it makes
; writing the upper level code easier.
;
;				*	*	*
;
; We've seen that processing a block is done in two major steps - the
; preshifted input buffer is filled, then the outputs are created.  Let's
; say a little more about each step.
;
; To efficiently use two ALUs in each clock cycle, there are really two
; independent trains of execution taking place in parallel.  Since
; they're basically the same, we'll just describe one of them.
;
; When calculating the preshifted inputs, we dedicate one MMX register to
; holding the input dword that's centered on the column of outputs we're
; going to calculate.  A second MMX register holds 32 bits (read using
; the movd opcode so we don't have to worry about the other 32 bits) to the
; right or left of the column.  It holds the bits to the left when we're
; shifting right and vice versa.
;
; This is a good place to distinguish between what the C code that generates
; machine code does and what the machine code does.  The C code parses the
; SE and puts out machine code.  The reason we go to all of this trouble
; is to eliminate virtually all of the conditional branches that would
; otherwise soak up time in a general morphology routine.  The C code generates
; the code that would be executed once all the branches have been
; decided.  Therefore, the C code does all the tests.  The steps listed
; below show what the C code does, and by inference, what the machine
; code does.
;  1.	Generate code to load and wordswap the central register.
;  2.	If there are no active columns to the left of the center's column,
;	generate code to byteswap in the central dword and jump to step 12.
;	This is not a jumps that are executed at run time, they're
;	jumps that are executed by the C code that generates machine
;	code.
;  3.	Generate code to load the 32 bits to the left of the central dword.
;  4.   If the central dword is not already wordswapped, generate code to
;	wordswap it.  It will already be wordswapped if we are executing this
;	step as a result of step 14.
;  5.	Generate code to copy the central dword to the scratch register.
;  6.	Generate code to shift the central dword as far right as
;	necessary to account for the distance from the leftmost active
;	column of the strel to the center of the strel.  This is
;   	Cx bit positions, where the center of the strel is a position
;   	(Cx,Cy) inside the strel (starting from (0,0) in the upper left corner).
;  7.	Generate code to shift the outboard register left so that Cx
;	bits from the word on the left are visible in the MS end of 
;	the register.  Initially, this is a shift of 32 + (32 - Cx) bits
;	since we must move the bits read in step 3 to the upper word of
;	the dword register in addition to moving them into position in that
;	upper word.
;  8.	Generate code to combine the outboard register with the scratch
;	register, forming the shifted input.
;  9.	If the leftmost active SE column has any HITs, generate code to
;	write this value to the buffer of shifted inputs.
; 10.	If the leftmost active SE column has any MISSes, generate code to
;	invert this value and write it to the next dword location in
;   	the buffer of shifted inputs.  The dword is inverted by
;	pxor-ing it with one of the two remaining MMX registers which
;	is filled with all 1's.  The 1's register was filled in the
;	prologue, which is not shown here.
; 11.	If there are any more active columns (contain HITs and/or MISSes) of
;   	the SE to the left of the center column, repeat steps 5
;	through 10 for each active column, working from the leftmost column
;	of the strel towards the center.  Modify the shift distances so
;	that the central bits in the scratch register are always shifted right
;	by the distance from the column in question to the central
;	strel column.  The outboard register is shifted left by the
;   	distance betweeen the last column and the current column.
;	It has progresively fewer bits in its top end as we compute
;   	dwords with smaller and smaller shifts.
; 12.	If the central column of the strel has HITs, generate code to 
;	write the central dword to the shifted input buffer.
; 13.   If the central column has MISSes, generate code to copy the
;	central dword to the scratch register, invert it, and write it.
; 14.	If any of the columns of the strel to the right of the central
;	column have HITs and/or MISSes, repeat steps 3 through 11
;	where the outboard register holds bits from the 32 bits to the
;	right of the central dword and the direction of all shifts is reversed
;	(left <-> right).  This time, work from the rightmost column of the
;	strel back towards the center.
;
; To increase efficiency, we actually do two rows at once, interleaving their
; opcodes so that the shifts and memory references don't collide.  This
; code is hand written out below.  All the C code does is assembly little
; modular units of the template code.
;
;
; On to the section of code that does ANDs (erodes) and ORs (dilation).  Ok,
; dilation is really more complicated than just replacing ANDs with ORs.
; In dilation we need to project the center pixel out to all the HIT
; locations in the strel instead of carrying pixels at all the HIT/MISS
; locations back to the center.  This introduces an interesting set
; of reversals, but tough.
;
; To maximize efficiency, we want to concurrently calculate as many
; partial results as possible.  We also want to do two ANDs per cycle
; when possible.  Since one can't do two MMX memory operations in the
; same cycle, one of the AND source operands must be in an MMX register.
; Therefore, we must reserve an MMX register for a source operand.
; Since there are 8 MMX registers and one must be used for a source
; operand, we can only hold 7 partial results at once.  Therefore,
; we break the block into subblocks of height 7.  To make this
; efficient, we'll tell the upper level code that the height of
; the standard block is some multiple of 7.
;
; I fretted that things would be most efficient if *both* of the source
; operands were in registers.  It's not true.  In MMX land, operating with
; a memory operand costs nothing in time, though it does entail awkward
; pairing restrictions.  However, loading a memory operand into a
; register takes an extra cycle.  So, in every clock cycle, we'll do
; two ANDs, one from memory and one from a register that was preloaded.
; As noted above, the real reason we hold as many output results as
; possible is to minimize the number of times we need to do the preloading.
;
; Yeah, but what about strels that have 7 or fewer rows?  Couldn't we
; process them more efficiently if we constructed the loop so that it
; retired two fully formed outputs and started two new outputs on each pass?
; That's how the SPARC code works, though it does one at a time since it
; doesn't have these pesky pairing restrictions.  Well, maybe.
;
; The total number of ANDs done in both cases is the same.  The number
; of times a preshifted source word is preloaded goes up in our scheme
; since source dwords that lie in rows above and below the current
; output block are read twice.  The amount of redundant reading depends on
; the height and width of the strel.  The strel having the most redundant
; reads has a height of 7.  Remember that we're only concerned about
; strels whose height is less than or equal to 7 now.  Say we have
; a 7x7 strel that's fully populated with active elements.  This
; means we need to do a total of (3 + 7 + 3)x7 reads per subblock
; of which (3+3)x7 will be repeated in the subblocks above and below.
; Well, we really only do an extra (6x7)/2 reads since only every row
; is preloaded.
;
; So say we're operating on a 2560 x 3300 image.  That image can be
; divided into 40 columns by 471.43 rows of 64x7 subblocks.  We need
; an extra (6x7)/2 * 40 * 471.43 = 396,002 operations to process
; the entire image.  What percentage is that of the total operations?
; Well, we need to produce 7 shifted dwords for every source dword and
; it takes 5 cycles to produce 2 dwords.  That's ((7 * 5)/2) * 40 * 3300
; or 2,310,000 cycles.  Then we need roughly 7*7 operations on each dword,
; but can do 2 per cycle, so add another (7*7)/2 * 40 * 3300 = 3,234,000
; cycles.  Thats a total of 5,544,000 cycles.  If we eliminate those
; extra loads, we increase performance by
; (5,544,000 + 396,002) / 5,544,000 = 1.07x.  Whoopee.  This is about the
; largest performance improvement we could expect to see.  Strels with
; fewer rows will have less overlap outside the 7 row block.  Strels with
; more rows will be processed 7 rows at a time anyway.
;
; Yeah, but what if we wanted to do this anyway?  There are at least
; two ugly problems you'll have to overcome.  First, you'll have to
; figure out how to keep track of the moving window of outputs.  Second,
; you'll have to efficiently wordswap the two output registers and
; write them out before you need them again.
;
; What's the deal with the revolving window of outputs?  You have
; an ugly choice of either moving the valid registers up as the
; output registers become vacant (an extra 5 ops per pair of input rows - or
; 5 * (3300/2) * 40 columns = 330,000 ops) or changing the "names" of the
; registers each time an output pair is written.  In the second option, the
; contents of the registers don't move, but we simply declare that if mm0
; held the top row of outputs before the 2 writes, mm2 holds them after the
; writes.  Not only is this confusing when trying to write and debug
; the code, it means that any loop must be unrolled so that it only
; repeats when the register names come around again, yet we must check
; to see if we're done after *every* register pair (2 ops * (3300/2) *
; 40 cols = 132,000 ops). 
;
; Ok, but how hard can it be to wordswap and write registers?  You need
; two unpack ops to do the wordswap, a dest address modification
; (not paired with an MMX mem ref, since you can't do that) and an
; MMX write to vacate a register.  We need to vacate two registers,
; so we need to do all this twice.  This is particularly challenging when
; strels have a small number of elements since you may need the
; registers before they're written, which means you need some way of
; detecting this case and completing the write even if its inefficient.
; To make the writing process efficient in the normal case,
; you need a fancy code generator that can combine the write
; preparation and write ops with subsequent processing ops for the
; rest of the strel, whatever they may be.  This is all breaking my head.
;
; So we'll do it the simpler way.  We process the columns in sub-blocks of
; height 7, then write all 7 out in one fell swoop.  This can be coded
; by hand and just copied by the template code.  It can be a tiny bit
; slower than the complicated way, but too bad.
;
; Like the SPARC algorithms, we will have restrictions on the size of
; structure elements we can handle.  The distance from the SE's center to
; the edge of the SE (dimensions cx, cy, (nx-cx) and (ny-cy)) must all be
; less than 32.  This is the width and height of the pad all around
; the image.
;
;				*	*	*
;
; After immense sillyness, I found out that the WATCOM Linker's PHARLAP format
; causes the linker to produce code in the FLAT memory model.  All that stuff
; in the Microsoft Assembler manual (and in the WATCOM compiler manual
; and in the flags we give to the WATCOM compiler) about small models
; is inoperative.  So, that makes things easy.  We write code as though
; we had the small model and everything's peachy.  In the final system,
; we'll be able to generate code relative to the data segment and execute it
; with respect to the code segment.  This will work because both segments have
; the same base (gee, I hope the code segment doesn't have a smaller limit!).

; It didn't.  All the calling and returning worked fine.


; Register Assignment:
; EAX:	number of remaining rows
; EBX:	output bytes/line, scratch
; ECX:	input bytes/line, scratch
; EDX:	holds dest addr, scratch address reg
; EBP:	points to vars on stack
; ESI:	holds source addr, scratch address reg
; EDI:	address of InputCache
; ESP:	stack pointer

; During shift stage in which input cache is filled with pre-shifted inputs:
; mm0:	central dword from upper row
; mm1:  32 bits to the left or right of central dword in upper row
; mm2:	upper row scratch shift register, also result of shift
; mm3:	central dword from lower row
; mm4:	32 bits to the left or right of central dword in lower row
; mm5:	lower row scratch shift register, also result of shift
; mm6:	holds all 1's, used to invert shift outputs.
; mm7:

; During shift stage in which outputs for purely horizontal strel are computed:
; mm0:	central dword from upper row
; mm1:  32 bits to the left or right of central dword in upper row
; mm2:	upper row scratch shift register, also result of shift
; mm3:	central dword from lower row
; mm4:	32 bits to the left or right of central dword in lower row
; mm5:	lower row scratch shift register, also result of shift
; mm6:	output for upper row.
; mm7:	output for lower row.

; During AND/OR stage in which outputs are computed:
; mm0:	output dword from row 0 MOD 7 (closest to top of image)
; mm1:  output dword from row 1 MOD 7
; mm2:	output dword from row 2 MOD 7
; mm3:	output dword from row 3 MOD 7
; mm4:	output dword from row 4 MOD 7
; mm5:	output dword from row 5 MOD 7
; mm6:	output dword from row 6 MOD 7
; mm7:  holds preshifted input

rRows		TEXTEQU	<eax>
rDstPitch	TEXTEQU	<ebx>
rSrcPitch	TEXTEQU	<ecx>
rDstAddr	TEXTEQU	<edx>
rSrcAddr	TEXTEQU	<esi>
rCache		TEXTEQU	<edi>

; Local Variables:
; srcAddr:	address of top source dword.  This is the dword that is
;		cy rows above the word that is aligned with destAddr.
;		Doing the offset by cy rows is trivial in the calling code
;		and saves us work on every block here.
; destAddr:	address of top dest word
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image
; destPitch:	number of bytes to be added to address to move down one line
;		in destination (output) image
; inputCacheRows: the number of rows in the cache of preshifted inputs
; blockHeight:	number of rows in block.
; inputCacheAddr: address of beginning of input cache.

InputCachePitch	EQU	48	; there are 48 bytes/row of input cache in this
				; example

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32-bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT	BYTE PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; To get the Microsoft assembler to believe the following labels are PUBLIC,
; we have to declare them with double colons, e.g. "mphT3Start::".
	PUBLIC	mphT3Start, mphT3PreLoadSrcAddrPitch
	PUBLIC	mphT3PreLoadCacheHeight
	PUBLIC	mphT3PreLoadCacheAddr, mphT3LoadUpperSrcDWord
	PUBLIC	mphT3PreLoad1sMask, mphT3PreLoad1sMaskToMM7
	PUBLIC	mphT3PreLoadUpperLeftSideWord
	PUBLIC	mphT3PreLoadUpperRightSideWord, mphT3WordSwapUpperStep1
	PUBLIC	mphT3LoadLowerSrcDWord, mphT3WordSwapUpperStep2
	PUBLIC	mphT3WordSwapUpperMM2Step1, mphT3WordSwapUpperMM2Step2
	PUBLIC	mphT3WordSwapLowerMM5Step1, mphT3WordSwapLowerMM5Step2
	PUBLIC	mphT3PreLoadLowerLeftSideWord, mphT3PreLoadLowerRightSideWord
	PUBLIC	mphT3WordSwapLowerStep1, mphT3LeftShiftUpperSideWord
	PUBLIC	mphT3RightShiftUpperSideWord, mphT3WordSwapLowerStep2
	PUBLIC	mphT3CopyUpperDWord, mphT3LeftShiftLowerSideWord
	PUBLIC	mphT3RightShiftLowerSideWord, mphT3LeftShiftUpperScratch
	PUBLIC	mphT3RightShiftUpperScratch, mphT3CopyLowerDWord
	PUBLIC	mphT3CombineShiftedUpperDWord, mphT3LeftShiftLowerScratch
	PUBLIC	mphT3RightShiftLowerScratch, mphT3CombineShiftedLowerDWord
	PUBLIC	mphT3WriteUpperScratchToCacheDisp8
	PUBLIC	mphT3WriteUpperScratchToCacheDisp32
	PUBLIC	mphT3WriteLowerScratchToCacheDisp8
	PUBLIC	mphT3WriteLowerScratchToCacheDisp32
	PUBLIC	mphT3WriteUpperScratchToDest
	PUBLIC	mphT3WriteLowerScratchToDest
	PUBLIC	mphT3InvertUpperShiftedDWord, mphT3InvertLowerShiftedDWord
	PUBLIC	mphT3InvertAndUpperShiftedDWord, mphT3InvertAndLowerShiftedDWord
	PUBLIC	mphT3InvertUpperOrigDWord, mphT3InvertLowerOrigDWord
	PUBLIC	mphT3InvertUpperOrigDWordToMM6, mphT3InvertLowerOrigDWordToMM7
	PUBLIC	mphT3WriteUpperOrigToCacheDisp8
	PUBLIC	mphT3WriteUpperOrigToCacheDisp32
	PUBLIC	mphT3WriteLowerOrigToCacheDisp8
	PUBLIC	mphT3WriteLowerOrigToCacheDisp32
	PUBLIC	mphT3AdjustSrcAddrBy2Lines
	PUBLIC	mphT3AdjustDstAddrBy2Lines
	PUBLIC	mphT3AdjustCachePtr
	PUBLIC	mphT3DecrRowsBy2AndBranch
	PUBLIC	mphT3DecrRowsBy2
	PUBLIC	mphT3JABranch
	PUBLIC	mphT3LoadUpperShiftedDWord, mphT3LoadLowerShiftedDWord
	PUBLIC	mphT3LoadUpperOrigDWord, mphT3LoadLowerOrigDWord
	PUBLIC	mphT3AndUpperShiftedDWord, mphT3AndLowerShiftedDWord
	PUBLIC	mphT3AndUpperShiftedDWordToMM2, mphT3AndLowerShiftedDWordToMM5
	PUBLIC	mphT3OrUpperShiftedDWordToMM2, mphT3OrLowerShiftedDWordToMM5
	PUBLIC	mphT3OrUpperShiftedDWord, mphT3OrLowerShiftedDWord
	PUBLIC	mphT3LoadBlockHeight, mphT3LoadDstAddr, mphT3LoadDstPitch
	PUBLIC	mphT3TestBlockHeight, mphT3LoadMM0Cache8
	PUBLIC	mphT3LoadMM1Cache8, mphT3LoadMM2Cache8
	PUBLIC	mphT3LoadMM3Cache8, mphT3LoadMM4Cache8
	PUBLIC	mphT3LoadMM5Cache8, mphT3LoadMM6Cache8
	PUBLIC	mphT3LoadMM7Cache8, mphT3LoadMM0Cache32
	PUBLIC	mphT3LoadMM1Cache32, mphT3LoadMM2Cache32
	PUBLIC	mphT3LoadMM3Cache32, mphT3LoadMM4Cache32
	PUBLIC	mphT3LoadMM5Cache32, mphT3LoadMM6Cache32
	PUBLIC	mphT3LoadMM7Cache32, mphT3LoadMM0Scratch
	PUBLIC	mphT3LoadMM1Scratch, mphT3LoadMM2Scratch
	PUBLIC	mphT3LoadMM3Scratch, mphT3LoadMM4Scratch
	PUBLIC	mphT3LoadMM5Scratch, mphT3LoadMM6Scratch
	PUBLIC	mphT3AndMM0Cache8
	PUBLIC	mphT3AndMM1Cache8, mphT3AndMM2Cache8
	PUBLIC	mphT3AndMM3Cache8, mphT3AndMM4Cache8
	PUBLIC	mphT3AndMM5Cache8, mphT3AndMM6Cache8
	PUBLIC	mphT3AndMM0Cache32, mphT3AndMM1Cache32
	PUBLIC	mphT3AndMM2Cache32, mphT3AndMM3Cache32
	PUBLIC	mphT3AndMM4Cache32, mphT3AndMM5Cache32
	PUBLIC	mphT3AndMM6Cache32, mphT3AndMM0Scratch
	PUBLIC	mphT3AndMM1Scratch, mphT3AndMM2Scratch
	PUBLIC	mphT3AndMM3Scratch, mphT3AndMM4Scratch
	PUBLIC	mphT3AndMM5Scratch, mphT3AndMM6Scratch
	PUBLIC	mphT3OrMM0Cache8, mphT3OrMM1Cache8, mphT3OrMM2Cache8
	PUBLIC	mphT3OrMM3Cache8, mphT3OrMM4Cache8, mphT3OrMM5Cache8
	PUBLIC	mphT3OrMM6Cache8, mphT3OrMM0Cache32, mphT3OrMM1Cache32
	PUBLIC	mphT3OrMM2Cache32, mphT3OrMM3Cache32, mphT3OrMM4Cache32
	PUBLIC	mphT3OrMM5Cache32, mphT3OrMM6Cache32, mphT3OrMM0Scratch
	PUBLIC	mphT3OrMM1Scratch, mphT3OrMM2Scratch, mphT3OrMM3Scratch
	PUBLIC	mphT3OrMM4Scratch, mphT3OrMM5Scratch, mphT3OrMM6Scratch
	PUBLIC	mphT3WriteSubBlockStart, mphT3WriteSubBlockIncCachePtr
	PUBLIC	mphT3WriteSubBlockRemainder, mphT3WriteVertSubBlockStart
	PUBLIC	mphT3WriteVertSubBlockRemainder, mphT3WriteTinySubBlockStart
	PUBLIC	mphT3WriteTinySubBlockRemainder
	PUBLIC	mphT3WriteTinyVertSubBlockStart
	PUBLIC	mphT3WriteTinyVertSubBlockRemainder, mphT3Return

; block lengths
	PUBLIC	mphT3StartBlkLen, mphT3PreLoadSrcAddrPitchBlkLen
	PUBLIC	mphT3PreLoadCacheHeightBlkLen
	PUBLIC	mphT3PreLoadCacheAddrBlkLen, mphT3LoadUpperSrcDWordBlkLen
	PUBLIC	mphT3PreLoad1sMaskBlkLen, mphT3PreLoad1sMaskToMM7BlkLen
	PUBLIC	mphT3PreLoadUpperLeftSideWordBlkLen
	PUBLIC	mphT3PreLoadUpperRightSideWordBlkLen
	PUBLIC	mphT3WordSwapUpperStep1BlkLen
	PUBLIC	mphT3LoadLowerSrcDWordBlkLen, mphT3WordSwapUpperStep2BlkLen
	PUBLIC	mphT3WordSwapUpperMM2Step1BlkLen
	PUBLIC	mphT3WordSwapUpperMM2Step2BlkLen
	PUBLIC	mphT3WordSwapLowerMM5Step1BlkLen
	PUBLIC	mphT3WordSwapLowerMM5Step2BlkLen
	PUBLIC	mphT3PreLoadLowerLeftSideWordBlkLen
	PUBLIC	mphT3PreLoadLowerRightSideWordBlkLen
	PUBLIC	mphT3WordSwapLowerStep1BlkLen
	PUBLIC	mphT3LeftShiftUpperSideWordBlkLen
	PUBLIC	mphT3RightShiftUpperSideWordBlkLen
	PUBLIC	mphT3WordSwapLowerStep2BlkLen
	PUBLIC	mphT3CopyUpperDWordBlkLen, mphT3LeftShiftLowerSideWordBlkLen
	PUBLIC	mphT3RightShiftLowerSideWordBlkLen
	PUBLIC	mphT3LeftShiftUpperScratchBlkLen
	PUBLIC	mphT3RightShiftUpperScratchBlkLen, mphT3CopyLowerDWordBlkLen
	PUBLIC	mphT3CombineShiftedUpperDWordBlkLen
	PUBLIC	mphT3LeftShiftLowerScratchBlkLen
	PUBLIC	mphT3RightShiftLowerScratchBlkLen
	PUBLIC	mphT3CombineShiftedLowerDWordBlkLen
	PUBLIC	mphT3WriteUpperScratchToCacheDisp8BlkLen
	PUBLIC	mphT3WriteUpperScratchToCacheDisp32BlkLen
	PUBLIC	mphT3WriteLowerScratchToCacheDisp8BlkLen
	PUBLIC	mphT3WriteLowerScratchToCacheDisp32BlkLen
	PUBLIC	mphT3WriteUpperScratchToDestBlkLen
	PUBLIC	mphT3WriteLowerScratchToDestBlkLen
	PUBLIC	mphT3InvertUpperShiftedDWordBlkLen
	PUBLIC	mphT3InvertLowerShiftedDWordBlkLen
	PUBLIC	mphT3InvertAndUpperShiftedDWordBlkLen
	PUBLIC	mphT3InvertAndLowerShiftedDWordBlkLen
	PUBLIC	mphT3InvertUpperOrigDWordBlkLen
	PUBLIC	mphT3InvertLowerOrigDWordBlkLen
	PUBLIC	mphT3InvertUpperOrigDWordToMM6BlkLen
	PUBLIC	mphT3InvertLowerOrigDWordToMM7BlkLen
	PUBLIC	mphT3WriteUpperOrigToCacheDisp8BlkLen
	PUBLIC	mphT3WriteUpperOrigToCacheDisp32BlkLen
	PUBLIC	mphT3WriteLowerOrigToCacheDisp8BlkLen
	PUBLIC	mphT3WriteLowerOrigToCacheDisp32BlkLen
	PUBLIC	mphT3AdjustSrcAddrBy2LinesBlkLen
	PUBLIC	mphT3AdjustDstAddrBy2LinesBlkLen
	PUBLIC	mphT3AdjustCachePtrBlkLen
	PUBLIC	mphT3DecrRowsBy2AndBranchBlkLen
	PUBLIC	mphT3DecrRowsBy2BlkLen
	PUBLIC	mphT3JABranchBlkLen
	PUBLIC	mphT3LoadUpperShiftedDWordBlkLen
	PUBLIC	mphT3LoadLowerShiftedDWordBlkLen
	PUBLIC	mphT3LoadUpperOrigDWordBlkLen
	PUBLIC	mphT3LoadLowerOrigDWordBlkLen
	PUBLIC	mphT3AndUpperShiftedDWordBlkLen, mphT3AndLowerShiftedDWordBlkLen
	PUBLIC	mphT3AndUpperShiftedDWordToMM2BlkLen
	PUBLIC	mphT3AndLowerShiftedDWordToMM5BlkLen
	PUBLIC	mphT3OrUpperShiftedDWordToMM2BlkLen
	PUBLIC	mphT3OrLowerShiftedDWordToMM5BlkLen
	PUBLIC	mphT3OrUpperShiftedDWordBlkLen, mphT3OrLowerShiftedDWordBlkLen
	PUBLIC	mphT3LoadBlockHeightBlkLen, mphT3LoadDstAddrBlkLen
	PUBLIC	mphT3LoadDstPitchBlkLen
	PUBLIC	mphT3TestBlockHeightBlkLen, mphT3LoadMM0Cache8BlkLen
	PUBLIC	mphT3LoadMM1Cache8BlkLen, mphT3LoadMM2Cache8BlkLen
	PUBLIC	mphT3LoadMM3Cache8BlkLen, mphT3LoadMM4Cache8BlkLen
	PUBLIC	mphT3LoadMM5Cache8BlkLen, mphT3LoadMM6Cache8BlkLen
	PUBLIC	mphT3LoadMM7Cache8BlkLen, mphT3LoadMM0Cache32BlkLen
	PUBLIC	mphT3LoadMM1Cache32BlkLen, mphT3LoadMM2Cache32BlkLen
	PUBLIC	mphT3LoadMM3Cache32BlkLen, mphT3LoadMM4Cache32BlkLen
	PUBLIC	mphT3LoadMM5Cache32BlkLen, mphT3LoadMM6Cache32BlkLen
	PUBLIC	mphT3LoadMM7Cache32BlkLen, mphT3LoadMM0ScratchBlkLen
	PUBLIC	mphT3LoadMM1ScratchBlkLen, mphT3LoadMM2ScratchBlkLen
	PUBLIC	mphT3LoadMM3ScratchBlkLen, mphT3LoadMM4ScratchBlkLen
	PUBLIC	mphT3LoadMM5ScratchBlkLen, mphT3LoadMM6ScratchBlkLen
	PUBLIC	mphT3AndMM0Cache8BlkLen
	PUBLIC	mphT3AndMM1Cache8BlkLen, mphT3AndMM2Cache8BlkLen
	PUBLIC	mphT3AndMM3Cache8BlkLen, mphT3AndMM4Cache8BlkLen
	PUBLIC	mphT3AndMM5Cache8BlkLen, mphT3AndMM6Cache8BlkLen
	PUBLIC	mphT3AndMM0Cache32BlkLen, mphT3AndMM1Cache32BlkLen
	PUBLIC	mphT3AndMM2Cache32BlkLen, mphT3AndMM3Cache32BlkLen
	PUBLIC	mphT3AndMM4Cache32BlkLen, mphT3AndMM5Cache32BlkLen
	PUBLIC	mphT3AndMM6Cache32BlkLen, mphT3AndMM0ScratchBlkLen
	PUBLIC	mphT3AndMM1ScratchBlkLen, mphT3AndMM2ScratchBlkLen
	PUBLIC	mphT3AndMM3ScratchBlkLen, mphT3AndMM4ScratchBlkLen
	PUBLIC	mphT3AndMM5ScratchBlkLen, mphT3AndMM6ScratchBlkLen
	PUBLIC	mphT3OrMM0Cache8BlkLen, mphT3OrMM1Cache8BlkLen
	PUBLIC	mphT3OrMM2Cache8BlkLen
	PUBLIC	mphT3OrMM3Cache8BlkLen, mphT3OrMM4Cache8BlkLen
	PUBLIC	mphT3OrMM5Cache8BlkLen
	PUBLIC	mphT3OrMM6Cache8BlkLen, mphT3OrMM0Cache32BlkLen
	PUBLIC	mphT3OrMM1Cache32BlkLen
	PUBLIC	mphT3OrMM2Cache32BlkLen, mphT3OrMM3Cache32BlkLen
	PUBLIC	mphT3OrMM4Cache32BlkLen
	PUBLIC	mphT3OrMM5Cache32BlkLen, mphT3OrMM6Cache32BlkLen
	PUBLIC	mphT3OrMM0ScratchBlkLen
	PUBLIC	mphT3OrMM1ScratchBlkLen, mphT3OrMM2ScratchBlkLen
	PUBLIC	mphT3OrMM3ScratchBlkLen
	PUBLIC	mphT3OrMM4ScratchBlkLen, mphT3OrMM5ScratchBlkLen
	PUBLIC	mphT3OrMM6ScratchBlkLen
	PUBLIC	mphT3WriteSubBlockStartBlkLen
	PUBLIC	mphT3WriteSubBlockIncCachePtrBlkLen
	PUBLIC	mphT3WriteSubBlockRemainderBlkLen
	PUBLIC	mphT3WriteVertSubBlockStartBlkLen
	PUBLIC	mphT3WriteVertSubBlockRemainderBlkLen
	PUBLIC	mphT3WriteTinySubBlockStartBlkLen
	PUBLIC	mphT3WriteTinySubBlockRemainderBlkLen
	PUBLIC	mphT3WriteTinyVertSubBlockStartBlkLen
	PUBLIC	mphT3WriteTinyVertSubBlockRemainderBlkLen, mphT3ReturnBlkLen


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name morphT3.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by i_erode.c

mphT3Start::
	PUBLIC	morphT3
morphT3	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	srcAddr:DWORD, destAddr:DWORD, srcPitch:DWORD, destPitch:DWORD,\
	inputCacheRows:DWORD, blockHeight:DWORD, inputCacheAddr:DWORD
mphT3StartEnd::

; All registers used are also pushed on the stack automatically.

; Load source address and source bytes/line.
mphT3PreLoadSrcAddrPitch::
	mov	rSrcPitch, srcPitch
	mov	rSrcAddr, srcAddr
mphT3PreLoadSrcAddrPitchEnd::

; load number of rows in input cache
mphT3PreLoadCacheHeight::
	mov	rRows, inputCacheRows
mphT3PreLoadCacheHeightEnd::

; load address of input cache
mphT3PreLoadCacheAddr::
	mov	rCache, inputCacheAddr
mphT3PreLoadCacheAddrEnd::

; This is a good place to map out the generic sequence for filling the
; input cache.  The stuff shown is executed once for each row of the
; input cache.  We process two lines at a time, that's what "upper" and
; "lower" mean.  The upper address is addressed using rSrcAddr.  The
; lower address is addressed using rSrcAddr + rSrcPitch.
;
; First, examine the columns of the strel in the following order:
; center column, leftmost col, next-to-leftmost col, ... col left of
; center, rightmost col, next-to-rightmost col, col to right of center.
; This peculiar order allows us to keep the pipeline flowing reasonably
; well.  How?  Well, processing from the outside in towards the middle
; allows us to keep throwing away sideword bits.  This means we can use
; a single register for the sideword instead of keeping the sideword in
; a register and copying it to a scratch reg.  This, in turn, means we don't
; have to reload the central dword (also) when moving from the left cols
; to the right cols since we never change it.
;
; How else does our method keep things flowing?  There's obviously no
; preparation for the initial load.  This is generated when we find the
; first active strel column when investigating them in the order
; specified above.  The only preparation we need when writing the center
; column's data is done by the initial load, so we might as well get that
; out of the way immediately.  The left and right shifting code can always
; interleave some writes with preparation for the next writes.  This
; preparation is normally shifting.   However, when coming to the end
; of the row, there's no more shifting to do, but there is some general
; updating of pointers and counts.  Sadly, one cannot pair the opcodes
; that do this with MMX writes, so there are bunch of empty cycles on the
; last column.
;
; Here is the order in which we decide which type of code to generate.
; The Preload code may come as a prelude to any other block.  Below that,
; only one type of block will be generated at a time:
;
; Preload (loads registers and starts the shifts)
; See if col is center col, if so generate code for it.
;    HITs only
;    MISSes only
;    Both HITs and MISSes
; See if col is last column to be processed (will be to left or right of
; center).  If so, generate code for it.
;    HITs only
;    MISSes only
;    Both HITs and MISSes
; See if EITHER the col is to the left of center and is followed by
; another column that's also to the left of center OR the col is to the
; right of center and is followed by another col that's also to the right
; of center.  If so, generate code for it.  We think this is the common
; case.
;    HITs only
;    MISSes only
;    Both HITs and MISSes
; If none of the cases above is true, this must be the last left col to
; be processed.  It must be followed by a right col.  Generate code for
; this case.
;    HITs only
;    MISSes only
;    Both HITs and MISSes
;
;
; There's only one type of preload code.  It is:
;START:	load upper dword
;	load 1's mask
;
;	load left/right upper side word
;	step 1 of upper wordswap
;
;	load lower dword
;	step 2 of upper wordswap
;
;	load left/right lower side word
;	step 1 of lower wordswap
;
;	shift upper sideword left/right
;	step 2 of lower wordswap
;
;	copy upper dword to upper scratch
;	shift lower sideword left/right
;
;	shift upper dword in scratch reg right/left
;	copy lower dword to lower scratch
;
;
; Here are the three types of code used to process the center block.
; This always happens immediately after a preload if it happens at all.
;
; Here's what happens if we want to write only the positive dword for
; the central columns:
;LOOP:
;	write unshifted upper dword to input cache
;	; empty
;
;	write unshifted lower dword to input cache + bytes/row
;	; empty
;
; Go back to LOOP above.  This doesn't mean that the MMX code loops.
; It means that the C code that generates MMX code loops around to generate
; the code to process another column.  "LOOP" signifies the place where the
; code for each column starts/stops.
;
; If this is a purely horizontal strel, these become:
;LOOP:
;	copy unshifted upper dword to mm6 temp
;	copy unshifted lower dword to mm7 temp
;
;
; Here's what happens if we want to write only the negative dword for
; the central columns:
;LOOP:
;	invert upper unshifted dword
;	invert lower unshifted dword
;
;	write inverted, unshifted upper dword to input cache
;	re-invert upper unshifted dword
;
;	write inverted, unshifted lower dword to input cache + bytes/row
;	re-invert lower unshifted dword
;
;LOOP: ...
;
; If this is a purely horizontal strel, this becomes:
;LOOP:
;	Invert upper unshifted dword, leaving result in mm6.
;	load mm7 with all 1's.
;
;	invert lower unshifted dword, leaving result in mm7.
;	; empty
;
; Here's what happens if we want to write both the positive and the
; negative dword for the central columns:
;LOOP:
;	write positive unshifted upper dword to input cache
;	invert upper unshifted dword
;
;	write positive unshifted lower dword to input cache + bytes/row
;	invert lower unshifted dword
;
;	write inverted, unshifted upper dword to input cache + 8
;	re-invert upper unshifted dword
;
;	write inverted, unshifted lower dword to input cache + 8 + bytes/row
;	re-invert lower unshifted dword
;
;LOOP: ...
;
; Note that we can't have a column with both hits and misses in a
; purely horizontal strel (there's only one thing in a column in this case)
; so we don't have to make a hit-and-miss horizontal strel case.
;
; Here's how we process the last column to the left or right of center.
; The basic difference between this case and the normal case is that we
; don't get ready for a new write, but we do do the end-of-loop code.
; Here's the code for HITs only (shifted words are coming in):
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	increment inputCache address by 2 * number of input cache bytes/row 
;	combine (OR) lower dword and sideword
;
;	add 2 * source bytes/line to source address	; wait for AGI
;	; empty
;
;	write last shifted upper dword to inputCache - 2 * bytes/row. 
;	; empty
;
;	write last shifted lower dword to inputCache - bytes/row
;	; empty
;
;	decrement number of rows left by 2 in inputCache
;	branch back to START if number of rows left > 0.
;
;
; If this is a purely horizontal strel, the code becomes:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	AND upper shifted dword with mm6 temp, result to mm2
;	combine (OR) lower dword and sideword
;
;	wordswap step 1 for mm2, mm6 as scratch
;	AND lower shifted dword with mm7 temp, result to mm5
;
;	add 2 * source bytes/line to source address
;	wordswap step 2 for mm2, mm6 as scratch
;
;	wordswap step 1 for mm5, mm7 as scratch
;	decrement number of rows left by 2
;
;	write mm2 to dstAddress			; note 1 cycle delay
;	wordswap step 2 for mm5, mm7 as scratch
;
;	; empty
;	; empty
;
;	write mm5 to dstAddress + dest bytes/row
;	; empty
;
;	add 2 * dest bytes/line to dest address using lea (no cc's affected)
;	branch back to START if number of rows left > 0.
;
;
; Here's the last col code for MISSes only:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	invert (pxor) upper dword in scratch reg
;	combine (OR) lower dword and sideword
;
;	increment inputCache address by 2 * number of input cache bytes/row
;	invert (pxor) lower dword in scratch reg
;
;	add 2 * source bytes/line to source address	; wait for AGI
;	; empty
;
;	write last inverted, shifted upper dword to inputCache - 2 * bytes/row. 
;	; empty
;
;	write last inverted, shifted lower dword to inputCache - bytes/row
;	; empty
;
;	decrement number of rows left by 2 in inputCache
;	branch back to START if number of rows left > 0.
;
; If this is a purely horizontal strel, the code becomes:
; Here's the last col code for MISSes only:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	invert (pandn) upper dword in scratch reg, result to scratch reg (mm2)
;	combine (OR) lower dword and sideword
;
;	wordswap step 1 for mm2, mm6 as scratch
;	invert (pandn) lower dword in scratch reg, result to scratch reg (mm5)
;
;	add 2 * source bytes/line to source address
;	wordswap step 2 for mm2, mm6 as scratch
;
;	wordswap step 1 for mm5, mm7 as scratch
;	decrement number of rows left by 2
;
;	write mm2 to dstAddress		; note 1 cycle delay
;	wordswap step 2 for mm5, mm7 as scratch
;
;	; empty
;	; empty
;
;	write mm5 to dstAddress + dest bytes/row
;	; empty
;
;	decrement number of rows left by 2 in inputCache
;	branch back to START if number of rows left > 0.
;
;
; Here's the last col code for both HITs and MISSes:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	increment inputCache address by 2 * number of input cache bytes/row
;	combine (OR) lower dword and sideword
;
;	add 2 * source bytes/line to source address	; wait for AGI
;	; empty
;
;	write last shifted upper dword to inputCache - 2 * bytes/row. 
;	invert (pxor) upper shifted dword
;
;	write last shifted lower dword to inputCache - bytes/row
;	invert (pxor) lower shifted dword
;
;	write last inverted, shifted upper dword to inputCache+8-2*(bytes/row). 
;	; empty
;
;	write last inverted, shifted lower dword to inputCache+8-(bytes/row)
;	; empty
;
;	decrement number of rows left by 2 in inputCache
;	branch back to START if number of rows left > 0.
;
;
;
; Here's the code for processing a left or right col that's followed by
; a col on the same side of center. In other words, this isn't the last
; col in the row and it isn't a left col followed by a right col.
; As usual, we start with the code for HITs only:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	shift upper sideword left/right
;	combine (OR) lower dword and sideword
;
;	write upper dword to input cache	; note 1 cycle delay from OR
;	shift lower sideword left/right
;
;	write lower dword to input cache + bytes/row
;	copy upper dword to upper scratch
;
;	shift upper dword in scratch reg right/left
;	copy lower dword to lower scratch
;
; If this is a purely horizontal strel, the code becomes:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	shift upper sideword left/right
;	combine (OR) lower dword and sideword
;
;	pand upper dword to mm6
;	shift lower sideword left/right
;
;	pand lower dword to mm7
;	copy upper dword to upper scratch
;
;	shift upper dword in scratch reg right/left
;	copy lower dword to lower scratch
;
;
; Here's the code for MISSes only:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	invert (pxor) upper dword in scratch reg
;	combine (OR) lower dword and sideword
;
;	shift upper sideword left/right
;	invert (pxor) lower dword in scratch reg
;
;	write upper dword to input cache	; note 1 cycle delay from XOR
;	shift lower sideword left/right
;
;	write lower dword to input cache + bytes/row
;	copy upper dword to upper scratch
;
;	shift upper dword in scratch reg right/left
;	copy lower dword to lower scratch
;	
; Go back to LOOP above.
;
; If this is a purely horizontal strel, the code becomes:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	invert (pandn) upper dword in scratch reg, result in mm2
;	combine (OR) lower dword and sideword
;
;	shift upper sideword left/right
;	invert (pandn) lower dword in scratch reg, result in mm5
;
;	copy upper dword in mm2 back to mm6
;	shift lower sideword left/right
;
;	copy lower dword in mm5 back to mm7
;	copy upper dword to upper scratch
;
;	shift upper dword in scratch reg right/left
;	copy lower dword to lower scratch
;	
;
; If we need both positive and inverted versions of a dword, we get:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	shift upper sideword left/right
;	combine (OR) lower dword and sideword
;
;	write upper dword to input cache	; note 1 cycle delay from OR
;	invert upper dword in scratch reg
;
;	write lower dword to input cache + bytes/row
;	invert lower dword in scratch reg
;
;	write inverted upper dword to input cache + 8
;	shift lower sideword left/right
;
;	write inverted lower dword to input cache + 8 + bytes/row
;	copy upper dword to upper scratch
;
;	shift upper dword in scratch reg right/left
;	copy lower dword to lower scratch
;
; Go back to LOOP above.
;
;
;
; Finally, we need to show the code used for a left col that's followed
; by a right col.  If there was a center col, it's already been processed,
; so we don't need to worry about it here.  First, the code for HITs only:.
;
; Here's how things go when we write no dwords for the central column
; (it's not active).
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	read right upper sideword (32 bits).
;	combine (OR) lower dword and sideword
;
;	write last shifted upper dword to input cache
;	copy upper dword to upper scratch
;
;	write last shifted lower dword to input cache + bytes/row
;	shift upper dword in scratch reg left
;
;	load right lower sideword (32 bits).
;	shift upper sideword right
;
;	copy lower dword to lower scratch
;	shift lower sideword right
;
;LOOP: ...
;
; If this is a purely horizontal strel, the code becomes:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	read right upper sideword (32 bits).
;	combine (OR) lower dword and sideword
;
;	pand upper dword, result to mm6
;	copy upper dword to upper scratch
;
;	pand lower dword, result to mm7
;	shift upper dword in scratch reg left
;
;	load right lower sideword (32 bits).
;	shift upper sideword right
;
;	copy lower dword to lower scratch
;	shift lower sideword right
;
;
; Here's the code that processes MISSes only:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	invert (pxor) upper dword in scratch reg
;	combine (OR) lower dword and sideword
;
;	read right upper sideword (32 bits).
;	invert (pxor) lower dword in scratch reg
;
;	write last shifted upper dword to input cache
;	copy upper dword to upper scratch
;
;	write last shifted lower dword to input cache + bytes/row
;	shift upper dword in scratch reg left
;
;	load right lower sideword (32 bits).
;	shift upper sideword right
;
;	copy lower dword to lower scratch
;	shift lower sideword right
;
;LOOP: ...
;
; If this is a purely horizontal strel, the code becomes:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	invert (pandn) upper dword, result in mm2
;	combine (OR) lower dword and sideword
;
;	read right upper sideword (32 bits).
;	invert (pandn) lower dword, result in mm5
;
;	copy upper dword in mm2 back to mm6
;	copy upper dword to upper scratch
;
;	copy lower dword in mm5 back to mm7
;	shift upper dword in scratch reg left
;
;	load right lower sideword (32 bits).
;	shift upper sideword right
;
;	copy lower dword to lower scratch
;	shift lower sideword right
;
;
; Here's the version that processes both HITs and MISSes:
;LOOP:
;	combine (OR) upper dword and sideword
;	shift lower dword in scratch reg right/left
;
;	read right upper sideword (32 bits).
;	combine (OR) lower dword and sideword
;
;	write last shifted upper dword to input cache
;	invert (pxor) upper dword in scratch reg
;
;	write last shifted lower dword to input cache + bytes/row
;	invert (pxor) lower dword in scratch reg
;
;	write last inverted, shifted upper dword to input cache + 8
;	copy upper dword to upper scratch
;
;	write last inverted, shifted lower dword to input cache + 8 + bytes/row
;	shift upper dword in scratch reg left
;
;	load right lower sideword (32 bits).
;	shift upper sideword right
;
;	copy lower dword to lower scratch
;	shift lower sideword right
;
;LOOP: ...
;
;
;
; We should also say how the input cache is filled for purely vertical
; strels.  For purely vertical strels containg only hits.  There is no
; LOOP section, only a START:
;
;START:	read upper dword
;	; empty
;
;	read lower dword
;	; empty
;
;	add 2 * source bytes/line to source address
;	increment input cache address by 2 * number of bytes/row of input cache
;
;	write upper dword to input cache - 2 * bytes/row
;	; empty
;
;	write lower dword to input cache - bytes/row
;	; empty
;
;	decrement number of rows left by 2 in input cache
;	branch back to START if number of rows left > 0.
;
; Don't all those empties make you sick?  The dumb thing just can't pair
; integer operations with MMX memory operations.
;
; If we have a vertical strel containing only misses:
;
;START:	read upper dword
;	(re)load mm6 with all 1's
;
;	read lower dword
;	invert upper dword using mm6
;
;	invert lower dword using mm6
;	; empty
;
;	add 2 * source bytes/line to source address
;	increment input cache address by 2 * number of bytes/row of input cache
;
;	write inverted upper dword to input cache - 2 * bytes/row
;	; empty
;
;	write inverted lower dword to input cache - bytes/row
;	; empty
;
;	decrement number of rows left by 2 in input cache
;	branch back to START if number of rows left > 0.
;
; If we have a vertical strel containing both hits and misses:
;
;START:	read upper dword
;	(re)load mm6 with all 1's
;
;	read lower dword
;	; empty	
;
;	write upper dword to input cache
;	invert upper dword using mm6
;
;	write lower dword to input cache + cache bytes/row
;	invert lower dword using mm6
;
;	add 2 * source bytes/line to source address
;	increment input cache address by 2 * number of bytes/row of input cache
;
;	write inverted upper dword to input cache - 2 * bytes/row
;	; empty
;
;	write inverted lower dword to input cache - bytes/row
;	; empty
;
;	decrement number of rows left by 2 in input cache
;	branch back to START if number of rows left > 0.
;


; the first opcode of filling the shift buffer is loading the central dword
; from the upper line
mphT3LoadUpperSrcDWord::
	movq	mm0, [rSrcAddr]		; get upper src dword
mphT3LoadUpperSrcDWordEnd::

; If we're really going to fill the shift buffer (not purely horizontal
; strel), fill a word with 1's. This goes in the delay slot of the word above.
mphT3PreLoad1sMask::
	pcmpeqb	mm6, mm6		; a byte is always equal to itself,
					; so all masks are 1's.
mphT3PreLoad1sMaskEnd::

mphT3PreLoad1sMaskToMM7::
	pcmpeqb	mm7, mm7		; a byte is always equal to itself,
					; so all masks are 1's.
mphT3PreLoad1sMaskToMM7End::

; Load the side word to the left and byteswap the source dword.
; Where do these strange offsets come from?  Here's a map of
; word addresses.  Remember that we organize the image into
; chunks of 32-bit words.  In the x86 world, byte 0 is really
; the LS byte of each word, but that's ok.  These are the addresses
; you see if you read the image from left to right.
; ...  ||   -8   |   -4   ||   +0   |   +4   ||   +8   |   +12  | ...
; When we read in the central dword (address 0), we see [  +4  |  +0  ],
; so it must be wordswapped.  However, when reading the 32 bits to the left
; of the dword, we want the 32 bits at address -4.  To get the 32 bits to
; the right of the dword, we want address +8.  Note that these offsets
; work because we do 32-bit reads.  If we did 64-bit reads, we'd need
; -8 and +8.

mphT3PreLoadUpperLeftSideWord::
	movd	mm1, [rSrcAddr - 4]
mphT3PreLoadUpperLeftSideWordEnd::

mphT3PreLoadUpperRightSideWord::
	movd	mm1, [rSrcAddr + 8]
mphT3PreLoadUpperRightSideWordEnd::

mphT3WordSwapUpperStep1::
	punpckldq mm7, mm0			; put mm0(31:0) into mm7(63:32)
mphT3WordSwapUpperStep1End::

; Load the lower row's dword and finish wordswap
mphT3LoadLowerSrcDWord::
	movq	mm3, [rSrcAddr + rSrcPitch]	; get lower src dword
mphT3LoadLowerSrcDWordEnd::

mphT3WordSwapUpperStep2::
	punpckhdq mm0, mm7			; put mm0(63:32) into mm0(31:0),
						; put mm7(63:32) into mm0(63:32)
mphT3WordSwapUpperStep2End::

; load the lower word's side word from the left or right and start wordswapping
; the lower word
mphT3PreLoadLowerLeftSideWord::
	movd	mm4, [rSrcAddr + rSrcPitch - 4]
mphT3PreLoadLowerLeftSideWordEnd::

mphT3PreLoadLowerRightSideWord::
	movd	mm4, [rSrcAddr + rSrcPitch + 8]
mphT3PreLoadLowerRightSideWordEnd::

mphT3WordSwapLowerStep1::
	punpckldq mm7, mm3			; put mm3(31:0) into mm7(63:32)
mphT3WordSwapLowerStep1End::

; Here's normal pair, a shift of the side word in the upper opcode and
; a copy of the lower src dword to its scratch register in the lower.
; again, we list them separately so the shift distance can be adjusted.
; For the sake of generality, we list both left and right shifts.
mphT3LeftShiftUpperSideWord::
	psllq	mm1, 32+(32-1)			; shift upper side word
mphT3LeftShiftUpperSideWordEnd::

mphT3RightShiftUpperSideWord::
	psrlq	mm1, 32+(32-1)			; shift upper side word
mphT3RightShiftUpperSideWordEnd::

mphT3WordSwapLowerStep2::
	punpckhdq mm3, mm7			; put mm3(63:32) into mm3(31:0),
						; put mm7(63:32) into mm3(63:32)
mphT3WordSwapLowerStep2End::

mphT3CopyUpperDWord::
	movq	mm2, mm0			; copy upper dword
mphT3CopyUpperDWordEnd::

mphT3LeftShiftLowerSideWord::
	psllq	mm4, 32+(32-1)			; shift lower side word
mphT3LeftShiftLowerSideWordEnd::

mphT3RightShiftLowerSideWord::
	psrlq	mm4, 32+(32-1)			; shift lower side word
mphT3RightShiftLowerSideWordEnd::

; For the sake of generality, we list both left and right shifts.
mphT3LeftShiftUpperScratch::
	psllq	mm2, 1				; shift upper dword
mphT3LeftShiftUpperScratchEnd::

mphT3RightShiftUpperScratch::
	psrlq	mm2, 1				; shift upper dword
mphT3RightShiftUpperScratchEnd::

mphT3CopyLowerDWord::
	movq	mm5, mm3			; copy lower dword
mphT3CopyLowerDWordEnd::

; This is the location for LOOP in the main comments.

; Here's the next normal pair.  We OR the shifted upper words and start
; shifting the lower dword.
mphT3CombineShiftedUpperDWord::
	por	mm2, mm1			; compose shifted upper dword
mphT3CombineShiftedUpperDWordEnd::

mphT3LeftShiftLowerScratch::
	psllq	mm5, 1				; shift lower dword
mphT3LeftShiftLowerScratchEnd::

mphT3RightShiftLowerScratch::
	psrlq	mm5, 1				; shift lower dword
mphT3RightShiftLowerScratchEnd::


; The first opcode in this pair is a shift upper sideword, which is already
; listed above.  The second is a combine lower DWord.
mphT3CombineShiftedLowerDWord::
	por	mm5, mm4			; compose shifted lower dword
mphT3CombineShiftedLowerDWordEnd::

; The next pair begins with a write upper.  There are two flavors of this,
; one with 8 bit displacements and one with a 32-bit displacement.
mphT3WriteUpperScratchToCacheDisp8::
	movq	[rCache + 4], mm2
mphT3WriteUpperScratchToCacheDisp8End::

mphT3WriteUpperScratchToCacheDisp32::
	movq	[rCache + 2560], mm2
mphT3WriteUpperScratchToCacheDisp32End::

; This opcode is used for the purely horizontal case.
mphT3WriteUpperScratchToDest::
	movq	[rDstAddr], mm2
mphT3WriteUpperScratchToDestEnd::

; The second opcode is this pair is a shift of the lower sideword.
; That's listed above.


; The next pair begins with a write lower scratch to memory.  Again,
; we have two flavors.
mphT3WriteLowerScratchToCacheDisp8::
	movq	[rCache + 4], mm5
mphT3WriteLowerScratchToCacheDisp8End::

mphT3WriteLowerScratchToCacheDisp32::
	movq	[rCache + 2560], mm5
mphT3WriteLowerScratchToCacheDisp32End::

mphT3WriteLowerScratchToDest::
	movq	[rDstAddr + rDstPitch], mm5
mphT3WriteLowerScratchToDestEnd::

; The second opcode of the pair is a copy from the upper dword to the
; scratch reg, but that's listed above.

; The last pair in the loop contains a shift of the upper scratch
; reg (already seen) paired with a copy of the lower dword to scratch,
; also already seen.
; That's the bottom of the main loop.



; If we need to store inverted versions of stuff, we need inverting
; opcodes, one for the upper dword and one for the lower dword.
mphT3InvertUpperShiftedDWord::
	pxor	mm2, mm6			; xor with all 1's in mm6
mphT3InvertUpperShiftedDWordEnd::

mphT3InvertLowerShiftedDWord::
	pxor	mm5, mm6			; xor with all 1's in mm6
mphT3InvertLowerShiftedDWordEnd::



; When we get to the central column, we need, in addition to the stuff
; above, some way to write the unshifted upper and lower dwords.
mphT3WriteUpperOrigToCacheDisp8::
	movq	[rCache + 4], mm0
mphT3WriteUpperOrigToCacheDisp8End::

mphT3WriteUpperOrigToCacheDisp32::
	movq	[rCache + 2560], mm0
mphT3WriteUpperOrigToCacheDisp32End::

mphT3WriteLowerOrigToCacheDisp8::
	movq	[rCache + 4], mm3
mphT3WriteLowerOrigToCacheDisp8End::

mphT3WriteLowerOrigToCacheDisp32::
	movq	[rCache + 2560], mm3
mphT3WriteLowerOrigToCacheDisp32End::

mphT3InvertUpperOrigDWord::
	pxor	mm0, mm6			; xor with all 1's in mm6
mphT3InvertUpperOrigDWordEnd::

mphT3InvertLowerOrigDWord::
	pxor	mm3, mm6			; xor with all 1's in mm6
mphT3InvertLowerOrigDWordEnd::

mphT3InvertUpperOrigDWordToMM6::
	pxor	mm6, mm0			; xor with all 1's, leave
						; result in mm6
mphT3InvertUpperOrigDWordToMM6End::

mphT3InvertLowerOrigDWordToMM7::
	pxor	mm7, mm3			; xor with all 1's, leave
						; result in mm7
mphT3InvertLowerOrigDWordToMM7End::

; After crossing the center column, we'll need to load the right side words,
; but we already coded those opcodes.


; When we finish a row, we need to write out the last (shifted) dwords,
; which we already know how to do.  We also need to increment the source
; address by 2 lines and add the number of bytes/input cache row to
; rCache.
mphT3AdjustSrcAddrBy2Lines::
	lea	rSrcAddr, [rSrcAddr + 2*rSrcPitch]	; move down 2 lines
mphT3AdjustSrcAddrBy2LinesEnd::

mphT3AdjustDstAddrBy2Lines::
	lea	rDstAddr, [rDstAddr + 2*rDstPitch]	; move down 2 lines
mphT3AdjustDstAddrBy2LinesEnd::

mphT3AdjustCachePtr::
	add	rCache, 256			; move down one row in cache
						; "256" is a placeholder for
						; the number of cache
						; bytes/line
mphT3AdjustCachePtrEnd::

; Finally, we have to decrement the number of rows in the input cache
; and branch back to the top of the loop.  These are often a pair.
mphT3DecrRowsBy2AndBranch::
	sub	rRows, 2
	ja	mphT3LoadUpperSrcDWord		; keep going if there's at
						; least one line left.
						; The input cache is padded
						; in rows to the next multiple
						; of 2, so there will always
						; be room for 2 more rows.
mphT3DecrRowsBy2AndBranchEnd::

mphT3DecrRowsBy2::
	sub	rRows, 2
mphT3DecrRowsBy2End::

mphT3JABranch::
	ja	mphT3LoadUpperSrcDWord
mphT3JABranchEnd::


; For purely vertical strels, we also need to be able to write the upper
; and lower dwords to the input cache ptr minus cache bytes/row.  That
; only uses the write opcodes with displacement that we've already got.


; We're getting close.  What about purely horizontal strels?  When
; calculating these, we keep temporary outputs in mm6 and mm7 instead
; of all 1's in mm6 and nothing in mm7.  Each shifted input is moved into,
; or ANDed with, one of the temporary registers.  The input might be
; inverted before being ANDed.  In this case, we use the pandn opcode
; to invert the shifted data and AND it with the temporary output.
; Unfortunately, the idiotic pandn opcode leaves the result in the
; shifted data register.  So, we have to follow it with a movq opcode.
;
; The way horizontal strels will work is that in the normal case, we
; substitute pand opcodes for writes.

mphT3LoadUpperShiftedDWord::
	movq	mm6, mm2			; load upper row
mphT3LoadUpperShiftedDWordEnd::

mphT3LoadLowerShiftedDWord::
	movq	mm7, mm5			; load lower row
mphT3LoadLowerShiftedDWordEnd::

mphT3LoadUpperOrigDWord::
	movq	mm6, mm0			; load upper row
mphT3LoadUpperOrigDWordEnd::

mphT3LoadLowerOrigDWord::
	movq	mm7, mm3			; load lower row
mphT3LoadLowerOrigDWordEnd::

mphT3AndUpperShiftedDWord::
	pand	mm6, mm2			; ordinary AND, upper row
mphT3AndUpperShiftedDWordEnd::

mphT3AndLowerShiftedDWord::
	pand	mm7, mm5			; ordinary AND, lower row
mphT3AndLowerShiftedDWordEnd::

mphT3AndUpperShiftedDWordToMM2::
	pand	mm2, mm6			; reverse AND, upper row
mphT3AndUpperShiftedDWordToMM2End::

mphT3AndLowerShiftedDWordToMM5::
	pand	mm5, mm7			; reverse AND, lower row
mphT3AndLowerShiftedDWordToMM5End::

mphT3OrUpperShiftedDWordToMM2::
	por	mm2, mm6			; reverse OR, upper row
mphT3OrUpperShiftedDWordToMM2End::

mphT3OrLowerShiftedDWordToMM5::
	por	mm5, mm7			; reverse OR, lower row
mphT3OrLowerShiftedDWordToMM5End::

mphT3WordSwapUpperMM2Step1::
	punpckldq mm6, mm2			; put mm2(31:0) into mm6(63:32)
mphT3WordSwapUpperMM2Step1End::

mphT3WordSwapUpperMM2Step2::
	punpckhdq mm2, mm6			; put mm2(63:32) into mm2(31:0),
						; put mm6(63:32) into mm2(63:32)
mphT3WordSwapUpperMM2Step2End::

mphT3WordSwapLowerMM5Step1::
	punpckldq mm7, mm5			; put mm5(31:0) into mm7(63:32)
mphT3WordSwapLowerMM5Step1End::

mphT3WordSwapLowerMM5Step2::
	punpckhdq mm5, mm7			; put mm5(63:32) into mm5(31:0),
						; put mm7(63:32) into mm5(63:32)
mphT3WordSwapLowerMM5Step2End::

mphT3OrUpperShiftedDWord::
	por	mm6, mm2			; ordinary OR, upper row
mphT3OrUpperShiftedDWordEnd::

mphT3OrLowerShiftedDWord::
	por	mm7, mm5			; ordinary OR, lower row
mphT3OrLowerShiftedDWordEnd::

; The inverting operations invert the shifted word in mm2 and
; AND it with the accumulated word.  This is fine, except that the
; result is left in mm2.  We'll use the movq opcodes above to
; move it back.  We use this method instead pxor-ing with a mask
; in memory since it makes the opcodes shorter and reduces memory
; accesses.
mphT3InvertAndUpperShiftedDWord::
	pandn	mm2, mm6 
mphT3InvertAndUpperShiftedDWordEnd::

mphT3InvertAndLowerShiftedDWord::
	pandn	mm5, mm7
mphT3InvertAndLowerShiftedDWordEnd::

; Insert some space here so that we can always use a big branch displacment.

	DB	256 DUP (?)


; We could assemble the code for purely vertical strels from the stuff
; shown above, but it wouldn't be as efficient as we could make it.
; for one thing, 


; That's it for the shift loop.  Well, almost.  What happens if the block
; height is odd?  We fill two rows at a time.  Doesn't that mean that we
; might try to fill one row too far and extend off the bottom of the image?
; It would, if we allowed the distance from the strel's center to the bottom
; of the strel to be 32.  We only accelerate strels where the distance from
; the center to the top, bottom or sides is less than 32.  Therefore, even
; though we may be trying to calculate the data for the row below the last
; row in the block, we'll read at most 31 rows below that, so will still be
; within the 32-row frame.




; So swell, now for the ANDs and ORs.  First, load up the block height
; and dest address and bytes/line.

mphT3LoadBlockHeight::
	mov	rRows, blockHeight
mphT3LoadBlockHeightEnd::

mphT3LoadDstAddr::
	mov	rDstAddr, destAddr
mphT3LoadDstAddrEnd::

mphT3LoadDstPitch::
	mov	rDstPitch, destPitch
mphT3LoadDstPitchEnd::

; If the block height is less than 7, jump down to the code that
; processes the block one line at a time.  This happens only at the
; very bottom of the image.  We jump back here after each subblock
; to see if the next subblock is full height.  If not, we finish up
; one line at a time.
mphT3TestBlockHeight::
	cmp	rRows, 7
	jb	mphT3WriteSubBlockStart
mphT3TestBlockHeightEnd::


; We process the block in subblocks of height 7.  The C code will create
; a loop to do the processing of each subblock.
;
; The basic loop structure has us working across the input cache,
; reading data for two rows at a time.  The two shifted, wordswapped,
; dwords are applied to two output dwords.  This is where things get
; confusing (as if they weren't already).
;
; Each cached dword is associated with one column of the strel.  Each
; cached dword also has a polarity - it corresponds to a HIT or a MISS.
; Let's number the columns of the strel from 0 to nx-1 with the center
; of the strel at column cx.  The C code will maintain a table that
; produces a strel column and polarity given a horizontal position in the
; input cache.
;
; We also need to worry about input cache rows.  Say we want to produce
; a block of height blockHeight which starts on row blockStart.  Say
; the strel has ny rows with a center on row cy.  Then the first input
; cache row will correspond to row blockStart - cy of the input (and output).
; The last cache row will correspond to row blockStart + (blockHeight-1) +
; ((ny-1) - cy).
;
; A block is produced in units of subblocks.  Each subblock has 7 rows
; and it 1 dword in width.  The first row of the subblock
; is on row subBlockStart.  We process the subblock by looping through
; the rows of the input cache starting from row subBlockStart-cy and
; finishing with row subBlockStart + 7 - 1 + (ny - 1 - cy).
;
; For each entry in the input cache on row R, we look up its
; corresponding strel column, C, and polarity P.  We then inspect all
; output dwords in the subblock.  Each output dword is in a row, say Ro.
; Ro = subBlockStart for the top dword, Ro = subBlockStart+1 for the
; next output word and so on.  For each output word and cache entry,
; we see if there's a strel element matching the input cache dword's
; polarity P at strel row (R - Ro + cy) and strel column C.  If so,
; we generate an AND opcode.  If not, we don't.
;
; For a particular pair of cache words (one from row R and one from row
; R+1, both from col C and having polarity P), we cycle through all the
; output dwords asking the question above.  This means we generally get
; to use a given input word more than once before getting the next input
; pair.  We move across the input cache two rows at a time, then move down
; two rows so each input cache word is reused as many times as possible
; before moving on.
;
; It's almost that easy.  First, there are two types of AND opcodes and
; a load opcode.  The load just loads an input cache entry into the
; MMX scratch register.  It's used the first time an output dword is
; touched, since we can't AND with the trash that was there previously.
; One type of AND opcode combines a register directly with an input
; cache dword.  The other type combines with the scratch reg.
; To do two ANDs in a single cycle, we use both types at once.
;
; Can we always use both types in a cycle?  Usually, but not always.
; Usually, we can compare words from input cache rows R and R+1 with
; output dwords from rows Ro and Ro+1.  However, at the very top and bottom
; of the range of input cache rows, Ro or Ro+1 might not be in the range
; of output dwords.  In those cases, we can only do one AND per cycle.
;
; Let's try an example.  Say the strel looks like this:
;
;	H  _  M
;	_  c  _		<- "c" is the center
;	H  _  M
;	H  _  H
;
; For this strel, nx = 3, ny = 4, cx = 1, cy = 1.  The input cache will
; have 3 dwords per row.  The first dword in a row will correspond to
; the input shifted right by 1.  That's for the HITs in the leftmost strel
; column (col 0).  That dword will not be inverted.  There is no dword
; corresponding to the center column (column 1) since there are no
; HITs or MISSes there.  The second input cache column will correspond
; to the input shifted left since there is a HIT in column 2.  The
; third dword in an input cache row is the inverse of the input
; shifted left by 1 since there is also a MISS in column 2.
;
; The number of rows in a block will be some multiple of 7.  If there
; are 49 (7 * 7) rows in a block, there will be 49 + (4-1) = 52 rows
; in the input cache.  Each row will have 3 dwords * 8 bytes/dword =
; 24 bytes.
;
; Say we want to process a block starting at row 49 of the input image.
; The input cache will have data from row number (49 - 1) to row number
; (49 - 1) + 52 - 1 = 99.
;
; When processing the first subblock in the block, we start on the 
; input cache row corresponding to row 49 - 1 = 48 of the image.  The
; output dwords of this subblock have row numbers (Ro) from 49 through
; 55.  We read the first pair of input cache dwords.  Their rows, R, are 48
; and 49.  They both have column, C, equal to 0.  Both have a polarity of HIT.
;
; Start looking at pairs of output dword rows (48, 49), (49, 50), (50, 51),...
; (54, 55), (55, 56).  Hey, but there are no dwords with output rows 48 or
; 56!  Yeah, but we have to deal with these special cases somehow, and this
; seems as rational as any.  The first input cache pair is (R, R+1) = (48, 49),
; C = 0, P = HIT.  Well, golly, at strel row (48 - 48 + 1), there's nothing.
; In particular, there's no HIT, so we do issue no opcodes.  Had there been
; a HIT there, we would have generated one AND since there's really no
; output dword for row 48, only for row 49.  Go on to the next pair of
; outputs.  At strel row (49 - 48 + 1) = 1, col 0, there's a HIT.  Hooray!
; We generate two AND opcodes.  Well, not quite.  First, since we notice that
; the upper input cache word hasn't been loaded yet, we really issue a load
; opcode first.  This puts the cached input word into mm7, the scratch
; register.  Second, since we notice that the output dwords in question
; haven't been touched yet, we convert the ANDs into simple moves to
; initialize them.  This moves the remaining cached dword and the dword
; that we just loaded into mm7 into the output's registers.
; You shouldn't AND with trash.
;
; We continue in this vein through output pair rows with the following
; results:
; output pair	input pair		strel (Rs,Cs)	Action
; (48, 49)	((48, 49), 0, HIT)	(1, 0) = _	nothing
; (49, 50)	((48, 49), 0, HIT)	(0, 0) = HIT	Two loads to init
;							first two dwords
; (51, 52)	((48, 49), 0, HIT)	(-1, 0) = ?	nothing
;
; and in fact, we can stop looking at output pairs at this point since
; (R - Ro + cy) corresponds to negative strel rows (above top of strel)
; for Ro > 49.
;
; So we move onto the next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (48, 49)	((48, 49), 2, HIT)	(1, 2) = _	nothing
; (49, 50)	((48, 49), 2, HIT)	(0, 2) = MISS	nothing 
;
; and onto the next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (48, 49)	((48, 49), 2, MISS)	(1, 2) = _	nothing
; (49, 50)	((48, 49), 2, MISS)	(0, 2) = MISS	two ANDS
;
; So much for the first two input cache rows.  Let's do the next two for
; practice.  Note that we can go farther down into the outputs this
; time.
; output pair	input pair		strel (Rs,Cs)	Action
; (48, 49)	((50, 51), 0, HIT)	(3, 0) = HIT	one AND (no row 48)
; (49, 50)	((50, 51), 0, HIT)	(2, 0) = HIT	two ANDS
; (50, 51)	((50, 51), 0, HIT)	(1, 0) = _	nothing
; (51, 52)	((50, 51), 0, HIT)	(0, 0) = HIT	two loads
;
; next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (48, 49)	((50, 51), 2, HIT)	(3, 2) = HIT	two ANDs
; (49, 50)	((50, 51), 2, HIT)	(2, 2) = MISS	nothing 
; (50, 51)	((50, 51), 2, HIT)	(1, 2) = _	nothing
; (51, 52)	((50, 51), 2, HIT)	(0, 2) = MISS	nothing 
;
; next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (48, 49)	((50, 51), 2, MISS)	(3, 2) = HIT	nothing
; (49, 50)	((50, 51), 2, MISS)	(2, 2) = MISS	two ANDs 
; (50, 51)	((50, 51), 2, MISS)	(1, 2) = _	nothing
; (51, 52)	((50, 51), 2, MISS)	(0, 2) = MISS	two ANDs 
;
; Let's do one more row for fun.  Note that the set of interesting
; output pairs shifts down by 2.
; output pair	input pair		strel (Rs,Cs)	Action
; (50, 51)	((52, 53), 0, HIT)	(3, 0) = HIT	two ANDs
; (51, 52)	((52, 53), 0, HIT)	(2, 0) = HIT	two ANDS
; (52, 53)	((52, 53), 0, HIT)	(1, 0) = _	nothing
; (53, 54)	((52, 53), 0, HIT)	(0, 0) = HIT	two loads
;
; next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (50, 51)	((52, 53), 2, HIT)	(3, 2) = HIT	two ANDs
; (51, 52)	((52, 53), 2, HIT)	(2, 2) = MISS	nothing 
; (52, 53)	((52, 53), 2, HIT)	(1, 2) = _	nothing
; (53, 54)	((52, 53), 2, HIT)	(0, 2) = MISS	nothing 
;
; next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (50, 51)	((52, 53), 2, MISS)	(3, 2) = HIT	nothing
; (51, 52)	((52, 53), 2, MISS)	(2, 2) = MISS	two ANDs 
; (52, 53)	((52, 53), 2, MISS)	(1, 2) = _	nothing
; (53, 54)	((52, 53), 2, MISS)	(0, 2) = MISS	two ANDs 
;
; What the heck. let's finish it. 
; output pair	input pair		strel (Rs,Cs)	Action
; (52, 53)	((54, 55), 0, HIT)	(3, 0) = HIT	two ANDs
; (53, 54)	((54, 55), 0, HIT)	(2, 0) = HIT	two ANDS
; (54, 55)	((54, 55), 0, HIT)	(1, 0) = _	nothing
; (55, 56)	((54, 55), 0, HIT)	(0, 0) = HIT	one load (no row 56)
;
; next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (52, 53)	((54, 55), 2, HIT)	(3, 2) = HIT	two ANDs
; (53, 54)	((54, 55), 2, HIT)	(2, 2) = MISS	nothing 
; (54, 55)	((54, 55), 2, HIT)	(1, 2) = _	nothing
; (55, 56)	((54, 55), 2, HIT)	(0, 2) = MISS	nothing 
;
; next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (52, 53)	((54, 55), 2, MISS)	(3, 2) = HIT	nothing
; (53, 54)	((54, 55), 2, MISS)	(2, 2) = MISS	two ANDs 
; (54, 55)	((54, 55), 2, MISS)	(1, 2) = _	nothing
; (55, 56)	((54, 55), 2, MISS)	(0, 2) = MISS	one AND (no row 56)
;
; and finally:
; output pair	input pair		strel (Rs,Cs)	Action
; (54, 55)	((56, 57), 0, HIT)	(3, 0) = HIT	two ANDs
; (55, 56)	((56, 57), 0, HIT)	(2, 0) = HIT	one AND (no row 56)
;
; next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (54, 55)	((56, 57), 2, HIT)	(3, 2) = HIT	two ANDs
; (55, 56)	((56, 57), 2, HIT)	(2, 2) = MISS	nothing 
;
; next input cache dword:
; output pair	input pair		strel (Rs,Cs)	Action
; (54, 55)	((56, 57), 2, MISS)	(3, 2) = HIT	nothing
; (55, 56)	((56, 57), 2, MISS)	(2, 2) = MISS	one AND (no row 56) 
;
; That's it.
;
; So the opcodes we need are loads for each register, a load for the
; scratch register and two types of ANDs for each register.
; We encode two types of loads and two types of ANDs with memory.
; One type of each has 8 bit displacements and one has 32-bit
; displacements.

; Loads with 8 bit address displacements
mphT3LoadMM0Cache8::
	movq	mm0, [rCache + 4]
mphT3LoadMM0Cache8End::

mphT3LoadMM1Cache8::
	movq	mm1, [rCache + 4]
mphT3LoadMM1Cache8End::

mphT3LoadMM2Cache8::
	movq	mm2, [rCache + 4]
mphT3LoadMM2Cache8End::

mphT3LoadMM3Cache8::
	movq	mm3, [rCache + 4]
mphT3LoadMM3Cache8End::

mphT3LoadMM4Cache8::
	movq	mm4, [rCache + 4]
mphT3LoadMM4Cache8End::

mphT3LoadMM5Cache8::
	movq	mm5, [rCache + 4]
mphT3LoadMM5Cache8End::

mphT3LoadMM6Cache8::
	movq	mm6, [rCache + 4]
mphT3LoadMM6Cache8End::

mphT3LoadMM7Cache8::
	movq	mm7, [rCache + 4]
mphT3LoadMM7Cache8End::


; Loads with 32 bit address displacements
mphT3LoadMM0Cache32::
	movq	mm0, [rCache + 2560]
mphT3LoadMM0Cache32End::

mphT3LoadMM1Cache32::
	movq	mm1, [rCache + 2560]
mphT3LoadMM1Cache32End::

mphT3LoadMM2Cache32::
	movq	mm2, [rCache + 2560]
mphT3LoadMM2Cache32End::

mphT3LoadMM3Cache32::
	movq	mm3, [rCache + 2560]
mphT3LoadMM3Cache32End::

mphT3LoadMM4Cache32::
	movq	mm4, [rCache + 2560]
mphT3LoadMM4Cache32End::

mphT3LoadMM5Cache32::
	movq	mm5, [rCache + 2560]
mphT3LoadMM5Cache32End::

mphT3LoadMM6Cache32::
	movq	mm6, [rCache + 2560]
mphT3LoadMM6Cache32End::

mphT3LoadMM7Cache32::
	movq	mm7, [rCache + 2560]
mphT3LoadMM7Cache32End::


; Loads from scratch register
mphT3LoadMM0Scratch::
	movq	mm0, mm7
mphT3LoadMM0ScratchEnd::

mphT3LoadMM1Scratch::
	movq	mm1, mm7
mphT3LoadMM1ScratchEnd::

mphT3LoadMM2Scratch::
	movq	mm2, mm7
mphT3LoadMM2ScratchEnd::

mphT3LoadMM3Scratch::
	movq	mm3, mm7
mphT3LoadMM3ScratchEnd::

mphT3LoadMM4Scratch::
	movq	mm4, mm7
mphT3LoadMM4ScratchEnd::

mphT3LoadMM5Scratch::
	movq	mm5, mm7
mphT3LoadMM5ScratchEnd::

mphT3LoadMM6Scratch::
	movq	mm6, mm7
mphT3LoadMM6ScratchEnd::

mphT3LoadMM7Scratch::
	movq	mm7, mm7
mphT3LoadMM7ScratchEnd::


; ANDs with 8 bit address displacements
mphT3AndMM0Cache8::
	pand	mm0, [rCache + 4]
mphT3AndMM0Cache8End::

mphT3AndMM1Cache8::
	pand	mm1, [rCache + 4]
mphT3AndMM1Cache8End::

mphT3AndMM2Cache8::
	pand	mm2, [rCache + 4]
mphT3AndMM2Cache8End::

mphT3AndMM3Cache8::
	pand	mm3, [rCache + 4]
mphT3AndMM3Cache8End::

mphT3AndMM4Cache8::
	pand	mm4, [rCache + 4]
mphT3AndMM4Cache8End::

mphT3AndMM5Cache8::
	pand	mm5, [rCache + 4]
mphT3AndMM5Cache8End::

mphT3AndMM6Cache8::
	pand	mm6, [rCache + 4]
mphT3AndMM6Cache8End::


; ANDs with 32 bit address displacements
mphT3AndMM0Cache32::
	pand	mm0, [rCache + 2560]
mphT3AndMM0Cache32End::

mphT3AndMM1Cache32::
	pand	mm1, [rCache + 2560]
mphT3AndMM1Cache32End::

mphT3AndMM2Cache32::
	pand	mm2, [rCache + 2560]
mphT3AndMM2Cache32End::

mphT3AndMM3Cache32::
	pand	mm3, [rCache + 2560]
mphT3AndMM3Cache32End::

mphT3AndMM4Cache32::
	pand	mm4, [rCache + 2560]
mphT3AndMM4Cache32End::

mphT3AndMM5Cache32::
	pand	mm5, [rCache + 2560]
mphT3AndMM5Cache32End::

mphT3AndMM6Cache32::
	pand	mm6, [rCache + 2560]
mphT3AndMM6Cache32End::


; ANDs with the input register.
mphT3AndMM0Scratch::
	pand	mm0, mm7
mphT3AndMM0ScratchEnd::

mphT3AndMM1Scratch::
	pand	mm1, mm7
mphT3AndMM1ScratchEnd::

mphT3AndMM2Scratch::
	pand	mm2, mm7
mphT3AndMM2ScratchEnd::

mphT3AndMM3Scratch::
	pand	mm3, mm7
mphT3AndMM3ScratchEnd::

mphT3AndMM4Scratch::
	pand	mm4, mm7
mphT3AndMM4ScratchEnd::

mphT3AndMM5Scratch::
	pand	mm5, mm7
mphT3AndMM5ScratchEnd::

mphT3AndMM6Scratch::
	pand	mm6, mm7
mphT3AndMM6ScratchEnd::


; So much for ANDing.  Ok, we need another set for ORing for dilation.
; ORs with 8 bit address displacements
mphT3OrMM0Cache8::
	por	mm0, [rCache + 4]
mphT3OrMM0Cache8End::

mphT3OrMM1Cache8::
	por	mm1, [rCache + 4]
mphT3OrMM1Cache8End::

mphT3OrMM2Cache8::
	por	mm2, [rCache + 4]
mphT3OrMM2Cache8End::

mphT3OrMM3Cache8::
	por	mm3, [rCache + 4]
mphT3OrMM3Cache8End::

mphT3OrMM4Cache8::
	por	mm4, [rCache + 4]
mphT3OrMM4Cache8End::

mphT3OrMM5Cache8::
	por	mm5, [rCache + 4]
mphT3OrMM5Cache8End::

mphT3OrMM6Cache8::
	por	mm6, [rCache + 4]
mphT3OrMM6Cache8End::


; ORs with 32 bit address displacements
mphT3OrMM0Cache32::
	por	mm0, [rCache + 2560]
mphT3OrMM0Cache32End::

mphT3OrMM1Cache32::
	por	mm1, [rCache + 2560]
mphT3OrMM1Cache32End::

mphT3OrMM2Cache32::
	por	mm2, [rCache + 2560]
mphT3OrMM2Cache32End::

mphT3OrMM3Cache32::
	por	mm3, [rCache + 2560]
mphT3OrMM3Cache32End::

mphT3OrMM4Cache32::
	por	mm4, [rCache + 2560]
mphT3OrMM4Cache32End::

mphT3OrMM5Cache32::
	por	mm5, [rCache + 2560]
mphT3OrMM5Cache32End::

mphT3OrMM6Cache32::
	por	mm6, [rCache + 2560]
mphT3OrMM6Cache32End::


; ORs with the input register
mphT3OrMM0Scratch::
	por	mm0, mm7
mphT3OrMM0ScratchEnd::

mphT3OrMM1Scratch::
	por	mm1, mm7
mphT3OrMM1ScratchEnd::

mphT3OrMM2Scratch::
	por	mm2, mm7
mphT3OrMM2ScratchEnd::

mphT3OrMM3Scratch::
	por	mm3, mm7
mphT3OrMM3ScratchEnd::

mphT3OrMM4Scratch::
	por	mm4, mm7
mphT3OrMM4ScratchEnd::

mphT3OrMM5Scratch::
	por	mm5, mm7
mphT3OrMM5ScratchEnd::

mphT3OrMM6Scratch::
	por	mm6, mm7
mphT3OrMM6ScratchEnd::


; Having computed all that stuff, we need to write out the result registers.
; This code is fairly simple.  Note that we also need to wordswap each
; dword before writing it out.
; We'll write this out as one hunk.  We deal with partial subblocks in
; completely different code below.
; Note the use of two address registers to avoid the Address Generation
; Interlock (AGI) delays.

; Insert some space here so that we can always use a big branch displacment.

	DB	256 DUP (?)

mphT3WriteSubBlockStart::
	mov	rSrcAddr, rDstAddr	; point to line 0
	punpckldq mm7, mm0
mphT3WriteSubBlockStartEnd::

mphT3WriteSubBlockIncCachePtr::
	add	rCache, 2560		; advance cache pointer
mphT3WriteSubBlockIncCachePtrEnd::

mphT3WriteSubBlockRemainder::
	punpckhdq mm0, mm7

	add	rDstAddr, rDstPitch	; point to line 1
	punpckldq mm7, mm1		; wait for mm0 to percolate

	movq	[rSrcAddr], mm0		; store line 0
	punpckhdq mm1, mm7

	lea	rSrcAddr, [rSrcAddr + 2*rDstPitch] ; point to line 2
	punpckldq mm7, mm2		; wait for mm1 to percolate

	movq	[rDstAddr], mm1		; store line 1
	punpckhdq mm2, mm7

	lea	rDstAddr, [rDstAddr + 2*rDstPitch] ; point to line 3
	punpckldq mm7, mm3		; wait for mm2 to percolate

	movq	[rSrcAddr], mm2		; store line 2
	punpckhdq mm3, mm7

	lea	rSrcAddr, [rSrcAddr + 2*rDstPitch] ; point to line 4
	punpckldq mm7, mm4		; wait for mm3 to percolate

	movq	[rDstAddr], mm3		; store line 3
	punpckhdq mm4, mm7

	lea	rDstAddr, [rDstAddr + 2*rDstPitch] ; point to line 5
	punpckldq mm7, mm5		; wait for mm4 to percolate

	movq	[rSrcAddr], mm4		; store line 4
	punpckhdq mm5, mm7

	lea	rSrcAddr, [rSrcAddr + 2*rDstPitch] ; point to line 6
	punpckldq mm7, mm6		; wait for mm5 to percolate

	movq	[rDstAddr], mm5		; store line 5
	punpckhdq mm6, mm7

	lea	rDstAddr, [rDstAddr + 2*rDstPitch] ; point to line 7, which
					; is line 0 for next loop
	; empty	

	movq	[rSrcAddr], mm6		; store line 6
	; empty

	sub	rRows, 7		; did 7 more rows
	jg	mphT3LoadMM0Cache8	; jump back to top of loop
mphT3WriteSubBlockRemainderEnd::


; Here's the loop we use to write results from vertical strels.  These
; strels have no byteswaps.
mphT3WriteVertSubBlockStart::
	lea	rSrcAddr, [4*rDstPitch] ; offset to line 4
mphT3WriteVertSubBlockStartEnd::

;	add	rCache, blockHeight	; advance cache pointer

mphT3WriteVertSubBlockRemainder::
	sub	rSrcAddr, rDstPitch	; offset to line 3
	lea	rSrcPitch, [rDstAddr + 4*rDstPitch] ; point to line 4

	movq	[rDstAddr], mm0		; store line 0

	movq	[rDstAddr + rDstPitch], mm1 ; store line 1

	movq	[rDstAddr + 2*rDstPitch], mm2 ; store line 2

	movq	[rDstAddr + rSrcAddr], mm3 ; store line 3

	movq	[rSrcPitch], mm4		; store line 4

	movq	[rSrcPitch + rDstPitch], mm5 ; store line 5

	movq	[rSrcPitch + 2*rDstPitch], mm6 ; store line 6

	lea	rDstAddr, [rSrcPitch + rSrcAddr] ; point to line 7,
					; which is line 0
					; for next subblock.

	sub	rRows, 7		; did 7 more rows
	jg	mphT3LoadMM0Cache8	; jump back to top of loop
mphT3WriteVertSubBlockRemainderEnd::


; We're getting closer all the time.  We've dealt with purely vertical
; strels, purely horizontal strels and all other strels.  The thing we
; haven't dealt with is subblocks that are less than 7 rows tall.  This
; happens at the bottom of the image.  This routine will usually be called
; with a block height that's a multiple of the subblock height.  The
; only time this rule is violated is at the bottom of the image.
; There, we may be called with some number of subblocks followed by
; one partial subblock.  The normal code loops back up to the check
; that looks at the remaining block height.  If it's less than 7, it
; jumps down here where we process one line at a time.  The looping
; jump at the bottom of the one-line-at-a-time code jumps back up to
; the top of the one-line-at-a-time code.  That's ok because once we've
; started on one-line-at-a-time processing for a partial block, that's
; how we'll finish the block.
;
; This routine looks like the normal subblock processing except that
; it only does one output row before looping back.  This is less efficient,
; but we only execute it in a tiny strip at the bottom of the image, so
; no big deal.
;
; The only extra template code we need for this thing is the code that
; writes out results.  Here it is.
mphT3WriteTinySubBlockStart::
	mov	rSrcAddr, rDstAddr
	punpckldq mm7, mm0
mphT3WriteTinySubBlockStartEnd::

;	add	rCache, blockHeight		; advance cache pointer

mphT3WriteTinySubBlockRemainder::
	punpckhdq mm0, mm7

	add	rDstAddr, rDstPitch	; wait for mm0 to percolate

	movq	[rSrcAddr], mm0		; store line 0

	sub	rRows, 1
	jnz	mphT3LoadMM0Cache8
mphT3WriteTinySubBlockRemainderEnd::

; We also need a version of this code used when processing vertical strels.
; No wordswapping is needed there.
mphT3WriteTinyVertSubBlockStart::
	mov	rSrcAddr, rDstAddr
mphT3WriteTinyVertSubBlockStartEnd::

;	add	rCache, blockHeight		; advance cache pointer

mphT3WriteTinyVertSubBlockRemainder::

	add	rDstAddr, rDstPitch	; this needs to happen sometime

	movq	[rSrcAddr], mm0		; store line 0

	sub	rRows, 1
	jnz	mphT3LoadMM0Cache8
mphT3WriteTinyVertSubBlockRemainderEnd::
;
; Hey, we're done!  Return here.
mphT3Return::
	ret
mphT3ReturnEnd::

morphT3 ENDP


; Lengths of the code blocks.  Put these down here so the disassembler
; output is easier to read (starts with code at offset 0).
mphT3StartBlkLen \
	DWORD	mphT3StartEnd - \
		mphT3Start
mphT3PreLoadSrcAddrPitchBlkLen \
	DWORD	mphT3PreLoadSrcAddrPitchEnd - \
		mphT3PreLoadSrcAddrPitch
mphT3PreLoadCacheHeightBlkLen \
	DWORD	mphT3PreLoadCacheHeightEnd - \
		mphT3PreLoadCacheHeight
mphT3PreLoadCacheAddrBlkLen \
	DWORD	mphT3PreLoadCacheAddrEnd - \
		mphT3PreLoadCacheAddr
mphT3LoadUpperSrcDWordBlkLen \
	DWORD	mphT3LoadUpperSrcDWordEnd - \
		mphT3LoadUpperSrcDWord
mphT3PreLoad1sMaskBlkLen \
	DWORD	mphT3PreLoad1sMaskEnd - \
		mphT3PreLoad1sMask
mphT3PreLoad1sMaskToMM7BlkLen \
	DWORD	mphT3PreLoad1sMaskToMM7End - \
		mphT3PreLoad1sMaskToMM7
mphT3PreLoadUpperLeftSideWordBlkLen \
	DWORD	mphT3PreLoadUpperLeftSideWordEnd - \
		mphT3PreLoadUpperLeftSideWord
mphT3PreLoadUpperRightSideWordBlkLen \
	DWORD	mphT3PreLoadUpperRightSideWordEnd - \
		mphT3PreLoadUpperRightSideWord
mphT3WordSwapUpperStep1BlkLen \
	DWORD	mphT3WordSwapUpperStep1End - \
		mphT3WordSwapUpperStep1
mphT3WordSwapUpperMM2Step1BlkLen \
	DWORD	mphT3WordSwapUpperMM2Step1End - \
		mphT3WordSwapUpperMM2Step1
mphT3LoadLowerSrcDWordBlkLen \
	DWORD	mphT3LoadLowerSrcDWordEnd - \
		mphT3LoadLowerSrcDWord
mphT3WordSwapUpperStep2BlkLen \
	DWORD	mphT3WordSwapUpperStep2End - \
		mphT3WordSwapUpperStep2
mphT3WordSwapUpperMM2Step2BlkLen \
	DWORD	mphT3WordSwapUpperMM2Step2End - \
		mphT3WordSwapUpperMM2Step2
mphT3PreLoadLowerLeftSideWordBlkLen \
	DWORD	mphT3PreLoadLowerLeftSideWordEnd - \
		mphT3PreLoadLowerLeftSideWord
mphT3PreLoadLowerRightSideWordBlkLen \
	DWORD	mphT3PreLoadLowerRightSideWordEnd - \
		mphT3PreLoadLowerRightSideWord
mphT3WordSwapLowerStep1BlkLen \
	DWORD	mphT3WordSwapLowerStep1End - \
		mphT3WordSwapLowerStep1
mphT3WordSwapLowerMM5Step1BlkLen \
	DWORD	mphT3WordSwapLowerMM5Step1End - \
		mphT3WordSwapLowerMM5Step1
mphT3LeftShiftUpperSideWordBlkLen \
	DWORD	mphT3LeftShiftUpperSideWordEnd - \
		mphT3LeftShiftUpperSideWord
mphT3RightShiftUpperSideWordBlkLen \
	DWORD	mphT3RightShiftUpperSideWordEnd - \
		mphT3RightShiftUpperSideWord
mphT3WordSwapLowerStep2BlkLen \
	DWORD	mphT3WordSwapLowerStep2End - \
		mphT3WordSwapLowerStep2
mphT3WordSwapLowerMM5Step2BlkLen \
	DWORD	mphT3WordSwapLowerMM5Step2End - \
		mphT3WordSwapLowerMM5Step2
mphT3CopyUpperDWordBlkLen \
	DWORD	mphT3CopyUpperDWordEnd - \
		mphT3CopyUpperDWord
mphT3LeftShiftLowerSideWordBlkLen \
	DWORD	mphT3LeftShiftLowerSideWordEnd - \
		mphT3LeftShiftLowerSideWord
mphT3RightShiftLowerSideWordBlkLen \
	DWORD	mphT3RightShiftLowerSideWordEnd - \
		mphT3RightShiftLowerSideWord
mphT3LeftShiftUpperScratchBlkLen \
	DWORD	mphT3LeftShiftUpperScratchEnd - \
		mphT3LeftShiftUpperScratch
mphT3RightShiftUpperScratchBlkLen \
	DWORD	mphT3RightShiftUpperScratchEnd - \
		mphT3RightShiftUpperScratch
mphT3CopyLowerDWordBlkLen \
	DWORD	mphT3CopyLowerDWordEnd - \
		mphT3CopyLowerDWord
mphT3CombineShiftedUpperDWordBlkLen \
	DWORD	mphT3CombineShiftedUpperDWordEnd - \
		mphT3CombineShiftedUpperDWord
mphT3LeftShiftLowerScratchBlkLen \
	DWORD	mphT3LeftShiftLowerScratchEnd - \
		mphT3LeftShiftLowerScratch
mphT3RightShiftLowerScratchBlkLen \
	DWORD	mphT3RightShiftLowerScratchEnd - \
		mphT3RightShiftLowerScratch
mphT3CombineShiftedLowerDWordBlkLen \
	DWORD	mphT3CombineShiftedLowerDWordEnd - \
		mphT3CombineShiftedLowerDWord
mphT3WriteUpperScratchToCacheDisp8BlkLen \  
	DWORD	mphT3WriteUpperScratchToCacheDisp8End - \
		mphT3WriteUpperScratchToCacheDisp8
mphT3WriteUpperScratchToCacheDisp32BlkLen \ 
	DWORD	mphT3WriteUpperScratchToCacheDisp32End - \
		mphT3WriteUpperScratchToCacheDisp32
mphT3WriteLowerScratchToCacheDisp8BlkLen \  
	DWORD	mphT3WriteLowerScratchToCacheDisp8End - \
		mphT3WriteLowerScratchToCacheDisp8
mphT3WriteLowerScratchToCacheDisp32BlkLen \ 
	DWORD	mphT3WriteLowerScratchToCacheDisp32End - \
		mphT3WriteLowerScratchToCacheDisp32
mphT3WriteUpperScratchToDestBlkLen \  
	DWORD	mphT3WriteUpperScratchToDestEnd - \
		mphT3WriteUpperScratchToDest
mphT3WriteLowerScratchToDestBlkLen \  
	DWORD	mphT3WriteLowerScratchToDestEnd - \
		mphT3WriteLowerScratchToDest
mphT3InvertUpperShiftedDWordBlkLen \
	DWORD	mphT3InvertUpperShiftedDWordEnd - \
		mphT3InvertUpperShiftedDWord
mphT3InvertLowerShiftedDWordBlkLen \
	DWORD	mphT3InvertLowerShiftedDWordEnd - \
		mphT3InvertLowerShiftedDWord
mphT3InvertUpperOrigDWordBlkLen \
	DWORD	mphT3InvertUpperOrigDWordEnd - \
		mphT3InvertUpperOrigDWord
mphT3InvertLowerOrigDWordBlkLen \
	DWORD	mphT3InvertLowerOrigDWordEnd - \
		mphT3InvertLowerOrigDWord
mphT3InvertUpperOrigDWordToMM6BlkLen \
	DWORD	mphT3InvertUpperOrigDWordToMM6End - \
		mphT3InvertUpperOrigDWordToMM6
mphT3InvertLowerOrigDWordToMM7BlkLen \
	DWORD	mphT3InvertLowerOrigDWordToMM7End - \
		mphT3InvertLowerOrigDWordToMM7
mphT3WriteUpperOrigToCacheDisp8BlkLen \
	DWORD	mphT3WriteUpperOrigToCacheDisp8End - \
		mphT3WriteUpperOrigToCacheDisp8
mphT3WriteUpperOrigToCacheDisp32BlkLen \
	DWORD	mphT3WriteUpperOrigToCacheDisp32End - \
		mphT3WriteUpperOrigToCacheDisp32
mphT3WriteLowerOrigToCacheDisp8BlkLen \
	DWORD	mphT3WriteLowerOrigToCacheDisp8End - \
		mphT3WriteLowerOrigToCacheDisp8
mphT3WriteLowerOrigToCacheDisp32BlkLen \
	DWORD	mphT3WriteLowerOrigToCacheDisp32End - \
		mphT3WriteLowerOrigToCacheDisp32
mphT3AdjustSrcAddrBy2LinesBlkLen \
	DWORD	mphT3AdjustSrcAddrBy2LinesEnd - \
		mphT3AdjustSrcAddrBy2Lines
mphT3AdjustDstAddrBy2LinesBlkLen \
	DWORD	mphT3AdjustDstAddrBy2LinesEnd - \
		mphT3AdjustDstAddrBy2Lines
mphT3AdjustCachePtrBlkLen \
	DWORD	mphT3AdjustCachePtrEnd - \
		mphT3AdjustCachePtr
mphT3DecrRowsBy2AndBranchBlkLen \
	DWORD	mphT3DecrRowsBy2AndBranchEnd - \
		mphT3DecrRowsBy2AndBranch
mphT3DecrRowsBy2BlkLen \
	DWORD	mphT3DecrRowsBy2End - \
		mphT3DecrRowsBy2
mphT3JABranchBlkLen \
	DWORD	mphT3JABranchEnd - \
		mphT3JABranch
mphT3LoadUpperShiftedDWordBlkLen \
	DWORD	mphT3LoadUpperShiftedDWordEnd - \
		mphT3LoadUpperShiftedDWord
mphT3LoadLowerShiftedDWordBlkLen \
	DWORD	mphT3LoadLowerShiftedDWordEnd - \
		mphT3LoadLowerShiftedDWord
mphT3LoadUpperOrigDWordBlkLen \
	DWORD	mphT3LoadUpperOrigDWordEnd - \
		mphT3LoadUpperOrigDWord
mphT3LoadLowerOrigDWordBlkLen \
	DWORD	mphT3LoadLowerOrigDWordEnd - \
		mphT3LoadLowerOrigDWord
mphT3AndUpperShiftedDWordBlkLen \
	DWORD	mphT3AndUpperShiftedDWordEnd - \
		mphT3AndUpperShiftedDWord
mphT3AndLowerShiftedDWordBlkLen \
	DWORD	mphT3AndLowerShiftedDWordEnd - \
		mphT3AndLowerShiftedDWord
mphT3AndUpperShiftedDWordToMM2BlkLen \
	DWORD	mphT3AndUpperShiftedDWordToMM2End - \
		mphT3AndUpperShiftedDWordToMM2
mphT3AndLowerShiftedDWordToMM5BlkLen \
	DWORD	mphT3AndLowerShiftedDWordToMM5End - \
		mphT3AndLowerShiftedDWordToMM5
mphT3OrUpperShiftedDWordToMM2BlkLen \
	DWORD	mphT3OrUpperShiftedDWordToMM2End - \
		mphT3OrUpperShiftedDWordToMM2
mphT3OrLowerShiftedDWordToMM5BlkLen \
	DWORD	mphT3OrLowerShiftedDWordToMM5End - \
		mphT3OrLowerShiftedDWordToMM5
mphT3OrUpperShiftedDWordBlkLen \
	DWORD	mphT3OrUpperShiftedDWordEnd - \
		mphT3OrUpperShiftedDWord
mphT3OrLowerShiftedDWordBlkLen \
	DWORD	mphT3OrLowerShiftedDWordEnd - \
		mphT3OrLowerShiftedDWord
mphT3InvertAndUpperShiftedDWordBlkLen \
	DWORD	mphT3InvertAndUpperShiftedDWordEnd - \
		mphT3InvertAndUpperShiftedDWord
mphT3InvertAndLowerShiftedDWordBlkLen \
	DWORD	mphT3InvertAndLowerShiftedDWordEnd - \
		mphT3InvertAndLowerShiftedDWord
mphT3LoadBlockHeightBlkLen \
	DWORD	mphT3LoadBlockHeightEnd - \
		mphT3LoadBlockHeight
mphT3LoadDstAddrBlkLen \
	DWORD	mphT3LoadDstAddrEnd - \
		mphT3LoadDstAddr
mphT3LoadDstPitchBlkLen \
	DWORD	mphT3LoadDstPitchEnd - \
		mphT3LoadDstPitch
mphT3TestBlockHeightBlkLen \
	DWORD	mphT3TestBlockHeightEnd - \
		mphT3TestBlockHeight
mphT3LoadMM0Cache8BlkLen \
	DWORD	mphT3LoadMM0Cache8End - \
		mphT3LoadMM0Cache8
mphT3LoadMM1Cache8BlkLen \
	DWORD	mphT3LoadMM1Cache8End - \
		mphT3LoadMM1Cache8
mphT3LoadMM2Cache8BlkLen \
	DWORD	mphT3LoadMM2Cache8End - \
		mphT3LoadMM2Cache8
mphT3LoadMM3Cache8BlkLen \
	DWORD	mphT3LoadMM3Cache8End - \
		mphT3LoadMM3Cache8
mphT3LoadMM4Cache8BlkLen \
	DWORD	mphT3LoadMM4Cache8End - \
		mphT3LoadMM4Cache8
mphT3LoadMM5Cache8BlkLen \
	DWORD	mphT3LoadMM5Cache8End - \
		mphT3LoadMM5Cache8
mphT3LoadMM6Cache8BlkLen \
	DWORD	mphT3LoadMM6Cache8End - \
		mphT3LoadMM6Cache8
mphT3LoadMM7Cache8BlkLen \
	DWORD	mphT3LoadMM7Cache8End - \
		mphT3LoadMM7Cache8
mphT3LoadMM0Cache32BlkLen \
	DWORD	mphT3LoadMM0Cache32End - \
		mphT3LoadMM0Cache32
mphT3LoadMM1Cache32BlkLen \
	DWORD	mphT3LoadMM1Cache32End - \
		mphT3LoadMM1Cache32
mphT3LoadMM2Cache32BlkLen \
	DWORD	mphT3LoadMM2Cache32End - \
		mphT3LoadMM2Cache32
mphT3LoadMM3Cache32BlkLen \
	DWORD	mphT3LoadMM3Cache32End - \
		mphT3LoadMM3Cache32
mphT3LoadMM4Cache32BlkLen \
	DWORD	mphT3LoadMM4Cache32End - \
		mphT3LoadMM4Cache32
mphT3LoadMM5Cache32BlkLen \
	DWORD	mphT3LoadMM5Cache32End - \
		mphT3LoadMM5Cache32
mphT3LoadMM6Cache32BlkLen \
	DWORD	mphT3LoadMM6Cache32End - \
		mphT3LoadMM6Cache32
mphT3LoadMM7Cache32BlkLen \
	DWORD	mphT3LoadMM7Cache32End - \
		mphT3LoadMM7Cache32
mphT3LoadMM0ScratchBlkLen \
	DWORD	mphT3LoadMM0ScratchEnd - \
		mphT3LoadMM0Scratch
mphT3LoadMM1ScratchBlkLen \
	DWORD	mphT3LoadMM1ScratchEnd - \
		mphT3LoadMM1Scratch
mphT3LoadMM2ScratchBlkLen \
	DWORD	mphT3LoadMM2ScratchEnd - \
		mphT3LoadMM2Scratch
mphT3LoadMM3ScratchBlkLen \
	DWORD	mphT3LoadMM3ScratchEnd - \
		mphT3LoadMM3Scratch
mphT3LoadMM4ScratchBlkLen \
	DWORD	mphT3LoadMM4ScratchEnd - \
		mphT3LoadMM4Scratch
mphT3LoadMM5ScratchBlkLen \
	DWORD	mphT3LoadMM5ScratchEnd - \
		mphT3LoadMM5Scratch
mphT3LoadMM6ScratchBlkLen \
	DWORD	mphT3LoadMM6ScratchEnd - \
		mphT3LoadMM6Scratch
mphT3AndMM0Cache8BlkLen \
	DWORD	mphT3AndMM0Cache8End - \
		mphT3AndMM0Cache8
mphT3AndMM1Cache8BlkLen \
	DWORD	mphT3AndMM1Cache8End - \
		mphT3AndMM1Cache8
mphT3AndMM2Cache8BlkLen \
	DWORD	mphT3AndMM2Cache8End - \
		mphT3AndMM2Cache8
mphT3AndMM3Cache8BlkLen \
	DWORD	mphT3AndMM3Cache8End - \
		mphT3AndMM3Cache8
mphT3AndMM4Cache8BlkLen \
	DWORD	mphT3AndMM4Cache8End - \
		mphT3AndMM4Cache8
mphT3AndMM5Cache8BlkLen \
	DWORD	mphT3AndMM5Cache8End - \
		mphT3AndMM5Cache8
mphT3AndMM6Cache8BlkLen \
	DWORD	mphT3AndMM6Cache8End - \
		mphT3AndMM6Cache8
mphT3AndMM0Cache32BlkLen \
	DWORD	mphT3AndMM0Cache32End - \
		mphT3AndMM0Cache32
mphT3AndMM1Cache32BlkLen \
	DWORD	mphT3AndMM1Cache32End - \
		mphT3AndMM1Cache32
mphT3AndMM2Cache32BlkLen \
	DWORD	mphT3AndMM2Cache32End - \
		mphT3AndMM2Cache32
mphT3AndMM3Cache32BlkLen \
	DWORD	mphT3AndMM3Cache32End - \
		mphT3AndMM3Cache32
mphT3AndMM4Cache32BlkLen \
	DWORD	mphT3AndMM4Cache32End - \
		mphT3AndMM4Cache32
mphT3AndMM5Cache32BlkLen \
	DWORD	mphT3AndMM5Cache32End - \
		mphT3AndMM5Cache32
mphT3AndMM6Cache32BlkLen \
	DWORD	mphT3AndMM6Cache32End - \
		mphT3AndMM6Cache32
mphT3AndMM0ScratchBlkLen \
	DWORD	mphT3AndMM0ScratchEnd - \
		mphT3AndMM0Scratch
mphT3AndMM1ScratchBlkLen \
	DWORD	mphT3AndMM1ScratchEnd - \
		mphT3AndMM1Scratch
mphT3AndMM2ScratchBlkLen \
	DWORD	mphT3AndMM2ScratchEnd - \
		mphT3AndMM2Scratch
mphT3AndMM3ScratchBlkLen \
	DWORD	mphT3AndMM3ScratchEnd - \
		mphT3AndMM3Scratch
mphT3AndMM4ScratchBlkLen \
	DWORD	mphT3AndMM4ScratchEnd - \
		mphT3AndMM4Scratch
mphT3AndMM5ScratchBlkLen \
	DWORD	mphT3AndMM5ScratchEnd - \
		mphT3AndMM5Scratch
mphT3AndMM6ScratchBlkLen \
	DWORD	mphT3AndMM6ScratchEnd - \
		mphT3AndMM6Scratch
mphT3OrMM0Cache8BlkLen \
	DWORD	mphT3OrMM0Cache8End - \
		mphT3OrMM0Cache8
mphT3OrMM1Cache8BlkLen \
	DWORD	mphT3OrMM1Cache8End - \
		mphT3OrMM1Cache8
mphT3OrMM2Cache8BlkLen \
	DWORD	mphT3OrMM2Cache8End - \
		mphT3OrMM2Cache8
mphT3OrMM3Cache8BlkLen \
	DWORD	mphT3OrMM3Cache8End - \
		mphT3OrMM3Cache8
mphT3OrMM4Cache8BlkLen \
	DWORD	mphT3OrMM4Cache8End - \
		mphT3OrMM4Cache8
mphT3OrMM5Cache8BlkLen \
	DWORD	mphT3OrMM5Cache8End - \
		mphT3OrMM5Cache8
mphT3OrMM6Cache8BlkLen \
	DWORD	mphT3OrMM6Cache8End - \
		mphT3OrMM6Cache8
mphT3OrMM0Cache32BlkLen \
	DWORD	mphT3OrMM0Cache32End - \
		mphT3OrMM0Cache32
mphT3OrMM1Cache32BlkLen \
	DWORD	mphT3OrMM1Cache32End - \
		mphT3OrMM1Cache32
mphT3OrMM2Cache32BlkLen \
	DWORD	mphT3OrMM2Cache32End - \
		mphT3OrMM2Cache32
mphT3OrMM3Cache32BlkLen \
	DWORD	mphT3OrMM3Cache32End - \
		mphT3OrMM3Cache32
mphT3OrMM4Cache32BlkLen \
	DWORD	mphT3OrMM4Cache32End - \
		mphT3OrMM4Cache32
mphT3OrMM5Cache32BlkLen \
	DWORD	mphT3OrMM5Cache32End - \
		mphT3OrMM5Cache32
mphT3OrMM6Cache32BlkLen \
	DWORD	mphT3OrMM6Cache32End - \
		mphT3OrMM6Cache32
mphT3OrMM0ScratchBlkLen \
	DWORD	mphT3OrMM0ScratchEnd - \
		mphT3OrMM0Scratch
mphT3OrMM1ScratchBlkLen \
	DWORD	mphT3OrMM1ScratchEnd - \
		mphT3OrMM1Scratch
mphT3OrMM2ScratchBlkLen \
	DWORD	mphT3OrMM2ScratchEnd - \
		mphT3OrMM2Scratch
mphT3OrMM3ScratchBlkLen \
	DWORD	mphT3OrMM3ScratchEnd - \
		mphT3OrMM3Scratch
mphT3OrMM4ScratchBlkLen \
	DWORD	mphT3OrMM4ScratchEnd - \
		mphT3OrMM4Scratch
mphT3OrMM5ScratchBlkLen \
	DWORD	mphT3OrMM5ScratchEnd - \
		mphT3OrMM5Scratch
mphT3OrMM6ScratchBlkLen \
	DWORD	mphT3OrMM6ScratchEnd - \
		mphT3OrMM6Scratch
mphT3WriteSubBlockStartBlkLen \
	DWORD	mphT3WriteSubBlockStartEnd - \
		mphT3WriteSubBlockStart
mphT3WriteSubBlockIncCachePtrBlkLen \
	DWORD	mphT3WriteSubBlockIncCachePtrEnd - \
		mphT3WriteSubBlockIncCachePtr
mphT3WriteSubBlockRemainderBlkLen \
	DWORD	mphT3WriteSubBlockRemainderEnd - \
		mphT3WriteSubBlockRemainder
mphT3WriteVertSubBlockStartBlkLen \
	DWORD	mphT3WriteVertSubBlockStartEnd - \
		mphT3WriteVertSubBlockStart
mphT3WriteVertSubBlockRemainderBlkLen \
	DWORD	mphT3WriteVertSubBlockRemainderEnd - \
		mphT3WriteVertSubBlockRemainder
mphT3WriteTinySubBlockStartBlkLen \
	DWORD	mphT3WriteTinySubBlockStartEnd - \
		mphT3WriteTinySubBlockStart
mphT3WriteTinySubBlockRemainderBlkLen \
	DWORD	mphT3WriteTinySubBlockRemainderEnd - \
		mphT3WriteTinySubBlockRemainder
mphT3WriteTinyVertSubBlockStartBlkLen \
	DWORD	mphT3WriteTinyVertSubBlockStartEnd - \
		mphT3WriteTinyVertSubBlockStart
mphT3WriteTinyVertSubBlockRemainderBlkLen \
	DWORD	mphT3WriteTinyVertSubBlockRemainderEnd - \
		mphT3WriteTinyVertSubBlockRemainder
mphT3ReturnBlkLen \
	DWORD	mphT3ReturnEnd - \
		mphT3Return

_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
; not global initialized data used
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END

IF 0				; These are all comments

COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: mortemp3.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.1  1996/12/18 22:54:51  ddavies
Initial revision


^

ENDIF



