/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  pcgendil.c -- generates assembly routines implementing dilate ops
 */


#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#ifdef __WATCOMC__
#include <string.h>
#endif
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "frames.pub"
#include "memory.pub"

    /* prototypes */
#include "shrmorph.pub"
#include "shrmorph.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: pcgendil.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* This code is very closely related to that in pcgenero.c, except it
 * generates dilation instead of erosion code.  pcgenero.c has extensive
 * comments.  Here, we'll discuss the difference between these
 * implementations of erosion and dilation.
 *
 * In erosion, we match a pattern in the input image (using hits for 1's,
 * misses for 0's and don't care terms), writing a dot in the output image
 * whereever the pattern matched the input image.  In dilation, we write
 * the pattern (really only the hits from the Structure Element) in the
 * output image at each position where a 1 was found in the input image.
 *
 * One way to implement dilation is to search for all the 1's in the input
 * image.  When a 1 is found at location (x,y), the SE is written into the
 * output image with the SE's defined center location (cx, cy) placed at
 * the location corresponding to the 1 in the input image.  While correct,
 * this is a very slow approach.
 *
 * This approach treats the structure element as a child's stamp and the
 * input image as a pattern that governs where the stamp is applied to the
 * output image.  What happens if we reverse the interpretations - calling
 * the input image the stamp and the SE the pattern in which the stamp is
 * applied?  With some reflection, you can convince yourself that the
 * output is still correct.
 *
 * A nice feature of the second interpretation is that the algorithm can
 * be much more efficient. Instead of constantly inserting a small pattern
 * into a larger one (with the attendent shifting, masking and merging), we
 * simply shift and OR the entire image in a very uniform manner.  If we had
 * a parallel processor, we'd be in really good shape, but even with a serial
 * processor new approach is simpler and faster.
 *
 * There's a fly in the ointment though.  The straightforward implementation
 * of this approach has us grabbing an input bit (or word) and doing a set
 * of shifts and ORs with the surrounding output bits (or words).
 * That means a lot of memory writes.  Memory operations are slow and writes,
 * especially lots of them so the write buffer overflows, are very slow.
 * We'd like to recast the algorithm so that it does lots of reads and
 * a few writes.  Each output word should be completely computed in a
 * register using the surrounding image words, then written once.
 *
 * This isn't too hard, but you have to do another conceptual reversal (these
 * usually take me weeks to get used to).  Normally, we think of the SE
 * as being centered (in the sense of (cx,cy), not the physical center) on
 * the word being computed.  The SE element at ((cx+i),(cy+j)) tells us
 * how to combine the image bit at offset (i,j) from the output bit in
 * question.  This is how we did the erosion code where we naturally moved
 * all the surrounding image bits to the center.
 *
 * In dilation, we natually want to move the center's image bit towards the
 * periphery.  So quit thinking of one SE centered at the current output.
 * Think instead about sliding the SE around into all positions where it still
 * intersects the current output bit.  For example, if the SE were 2 wide and
 * 3 tall, there would be 6 positions in which the SE would intersect
 * a given output bit.  The union of the centers of these SE positions defines
 * the set of bits that can affect the current output bit in a dilation
 * operation.
 *
 * Fine theory, but how do we use it?  For example, say we want to compute
 * the output image bit at location (x,y).  What input image bits should we
 * consider?  To see how far up to look, we place (x,y) in the *bottom* row
 * of the SE.  The distance to the center (the highest bit that could
 * affect this output bit) is (ny-cy-1) rows, so the highest row we need
 * to inspect is y - (ny-cy-1).  Similarly, the distance from the top row
 * of the SE to the center is cy, so we need to look at image bits that are
 * as much as cy rows below y.  In all, we need to inspect all bits at offsets
 * (dx, dy) from (x,y) where {-(nx-cx-1) <= dx <= cx} and
 * {-(ny-cy-1) <= dy <= cy}.  When considering an input image bit at offset
 * (dx, dy) from our output, we use SE element (cx-dx, cy-dy) to determine
 * the operation (cHit => OR; cDontCare or cMiss => nil) to be performed.
 * 
 * How does this interact with our cache of shifted inputs?  In the erosion
 * code, things were very nice since the offset into the SE and the offset
 * into the input cache were the same.  Can we do that here?
 *
 * No.  An important feature of the cache of inputs is that when we go down
 * one line in the source and dest images, we go down one line in the
 * input cache.  Well, ok, we could probably go through the input cache
 * from bottom to top while going through the images from top to bottom,
 * but that's too weird.  The easiest thing to do is to maintain two sets
 * of coordinates: one for the input cache and one for the SE.  Since we
 * still want to go through the input cache top to bottom, left to right;
 * we'll go through the SE bottom to top, right to left.
 *
 * One non-obvious question is "what do we do with MISSes?"  We ignore them.
 * Dilate operations that do something with misses have been defined, but
 * Dan Bloomberg says they're seldom used.
 *
 * Originally, we used global variables to pass data between the
 * three procs in this module.  That made the code non-reentrant.  All vars
 * are now passed as parameters.  The converted vars are pCodeBuf,
 * bufIndex, bufSize and firstActiveFound.  To make things tidy, we use
 * macros to call the subroutines that have lots of parameters, most of
 * which do not change from call to call.
 *
 * OPTIMIZATION:  The 486 processor inserts 1 pipeline delay
 * if you load a register in one instruction and use it to generate an
 * address in the next.  We would like to avoid doing this in the inner loops.
 * To do so, we detect whether each SE row has active elements to the left,
 * in the center and to the right before generating any opcodes for that row.
 * If there is something in the center, but nothing to left or right, we
 * advance the source address ptr down one line immediately after the load,
 * but before processing the center's element.
 *
 * Similarly, if there is something on the left, but nothing to the right, 
 * we advance the source ptr immediately after loading the word to the left.
 * Finally, we always advance the source pointer immediately after loading
 * the word to the right (which is only done if there is something to the
 * right).  If none of these were true, we advance the source pointer
 * after generating no code at all.
 *
 * INTERFACE CHANGE, 9/17/94.  Instead of having the routine that generates
 * assembly code always allocate the assembly buffer, we'll let the client
 * do it optionally.  This means we have to provide a routine that tells the
 * client the size of the buffer - i_getDilateAsmSize.  If the *generated_code
 * is NULL on entry to i_getDilateAsm, we allocate a buffer using MALLOC.
 * Otherwise, we assume the client has already allocated one.  We do not
 * shrink the buffer to fit after generating the code, that's the client's
 * job.
 *
 * BUG FIX 12/7/94:  Leigh Klotz found out that if we give this code a
 * strel that's entirely DONT_CARE terms, it generates random code.
 * Check for this case.  If it occurs, don't allocate any assembly code
 * at all.  i_fastDilate will fix things up.
 *
 */

    /* dimensions of the largest structure element we can handle */
#define cMaxShiftDist	31

    /* Labels in the template assembly code */
extern	UInt8	erT2Start, erT2PreLoadDec, erT2PreLoadInpCac;
extern	UInt8	erT2LoadInputCenter, erT2LoadInputLeft, erT2LoadInputRight;
extern	UInt8	erT2ShiftInputRight, erT2ShiftInputLeft;
extern	UInt8	erT2IncInputAddr;
extern	UInt8	erT2CacheMiss, erT2NonCacheInit, erT2NonCacheMiss;
extern	UInt8	erT2NonCacheMissRestore;
extern	UInt8	erT2NonCacheHitAnd, erT2NonCacheHitOr;
extern	UInt8	erT2LoadDestAddr;
extern	UInt8	erT2BottomLoop, erT2LoopJmp, erT2Return;
extern	UInt8	erT2CacheStore0, erT2CacheStoreShort, erT2CacheStoreLong;
extern	UInt8	erT2CacheLoadEAX0, erT2CacheLoadEAXShort;
extern	UInt8	erT2CacheLoadEAXLong, erT2NotEAX;
extern	UInt8	erT2CacheAnd0, erT2CacheAndShort, erT2CacheAndLong;
extern	UInt8	erT2CacheOr0, erT2CacheOrShort, erT2CacheOrLong;
extern	UInt8	erT2IncCachePtrShort, erT2IncCachePtrLong;
extern	UInt8	erT2SingleRightShift, erT2SingleLeftShift;
extern	UInt8	erT2DoubleLeftShift1, erT2Nop;

extern	UInt32	startBlkLen2, preLoadDecBlkLen2, preLoadInpCacBlkLen2;
extern	UInt32	loadInputCenterBlkLen2, loadInputLeftBlkLen2;
extern	UInt32	loadInputRightBlkLen2;
extern	UInt32	shiftInputRightBlkLen2, shiftInputLeftBlkLen2;
extern	UInt32	incInputAddrBlkLen2, cacheMissBlkLen2;
extern	UInt32	nonCacheMissBlkLen2, nonCacheMissRestoreBlkLen2;
extern	UInt32	nonCacheHitOrBlkLen2, nonCacheHitAndBlkLen2;
extern	UInt32	nonCacheInitBlkLen2, loadDestAddrBlkLen2;
extern	UInt32	bottomLoopBlkLen2, jumpBlkLen2, returnBlkLen2;
extern	UInt32	cacheStore0BlkLen2, cacheStoreShortBlkLen2;
extern	UInt32	cacheStoreLongBlkLen2, cacheLoadEAX0BlkLen2;
extern	UInt32	cacheLoadEAXShortBlkLen2, cacheLoadEAXLongBlkLen2;
extern	UInt32	cacheNotEAXBlkLen2;
extern	UInt32	cacheAnd0BlkLen2 , cacheAndShortBlkLen2;
extern	UInt32	cacheAndLongBlkLen2 , cacheOr0BlkLen2;
extern	UInt32	cacheOrShortBlkLen2 , cacheOrLongBlkLen2;
extern	UInt32	incCachePtrShortBlkLen2 , incCachePtrLongBlkLen2;
extern	UInt32	singleLeftShiftBlkLen2, singleRightShiftBlkLen2;
extern	UInt32	doubleLeftShift1BlkLen2, nopBlkLen2;

/* Define macros for common subroutine calls with fancy abort sequences.
 * This should clean up the code a lot since most of the information in
 * the calls does not change.
 */
#define	M_COPY_CODE(blkAddr, blkLen)	if ((__error__ = \
			ip_copyCode (pIPState, &blkAddr, blkLen, \
				      pCodeBuf, &bufIndex, bufSize)) \
				      != ia_successful) \
						goto genCleanup

#define M_GEN_DILATE_SE_ELEMENT(cacheRowOffset, cacheCol, lastRow) \
		if ((__error__ = \
			ip_genDilateSEElement (pIPState, \
					       cacheRowOffset+cacheCol, \
					       pCodeBuf, &bufIndex,\
					       bufSize, &firstActiveFound,\
					       lastRow)) != ia_successful) \
		  goto genCleanup
#define	M_STORE_CACHE(cacheRowOffset, cacheCol) \
	    if ((__error__ = ip_storeCache (pIPState, pCodeBuf, \
					    &bufIndex, bufSize, \
				            cacheRowOffset+cacheCol)) \
			!= ia_successful) \
		goto genCleanup

/* define better names for some of the variables. */
#define	numSECols	nx
#define	numSERows	ny
#define	centerX		cx
#define	centerY		cy
#define	cacheCenterX	numSECols-centerX-1
#define cacheCenterY	numSERows-centerY-1

/* PUBLIC */
/************************************************************************
 *  i_getDilateAsmSize():  return number of bytes needed for assembly code
 *			used for dilation with this strel.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getDilateAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
    PROLOGUE;
    Int32       maxBotSEElementBlkLen; /* max # of code bytes required
					* to process any SE element that's
					*  above the last row. */
    Int32       maxLastSEElementBlkLen; /* max # of code bytes required
					 * to process any SE element that's
					 * in the last row. */

	/* Reference cx to keep the compilers happy. */
    cx = cx;

	/* We'd like to compute the maximum number of opcode bytes required
	 * to retire any of the SE elements below the top row (remember we
	 * go through the SE from bottom to top).  The ones
	 * in the top row are different because we're doing shifting
	 * at the same time.  There are two potential operations for any
	 * of the SE elements, initialize because of cHit and process a
	 * later cHit.
	 * Iniitalizing because of a cHit merely requires a load.
	 * Processing a cHit is an OR, which is the same length.
	 */
    maxBotSEElementBlkLen = cacheLoadEAXLongBlkLen2;

	/* This is the number of opcode bytes required to process a
	 * word on the bottom row in addition to the normal loads and
	 * shifts.  The shifted data will already be in EBX.  There are
	 * two cases: 1) Found a cHit and first time we've touched
	 * dest word; 2) Found a cHit and not first time we've touched
	 * dest word.  The first case turns into a reg-reg move, 1 byte.
	 * The second turns into a reg-reg OR, still 1 byte.
	 */
    maxLastSEElementBlkLen = nonCacheHitOrBlkLen2;

	/* figure out how big a buffer to allocate for the code */
    *pNumAsmBytes =
		    /* initial boilerplate */
	       (startBlkLen2 +
		    /* adjust src address */
		((numSERows - centerY - 1) * preLoadDecBlkLen2) +
		preLoadInpCacBlkLen2 +		/* load input cache addr */
		    /* load center's word twice and words to left and
		     * right once each */
		(numSERows * (2*loadInputCenterBlkLen2 +
				loadInputLeftBlkLen2 +
				loadInputRightBlkLen2)) +
		    /* do one shift for each non-center column in
		     * every row.  The shift left and shift right opcodes 
		     * have the same length so we use the length from
		     * right shift. */
		(numSERows * (numSECols - 1) *
		    (shiftInputRightBlkLen2 + singleRightShiftBlkLen2)) +
		    /* store one word for each position in the input cache */
		(numSERows * numSECols * cacheStoreLongBlkLen2) +
		    /* process each input not corresponding to last (top)
		     * SE row */
		((numSERows-1) * numSECols * maxBotSEElementBlkLen) +
		    /* process each input corresponding to last (top)
		    * SE row */
		(numSECols * maxLastSEElementBlkLen) +
		loadDestAddrBlkLen2 + incCachePtrLongBlkLen2 +
		bottomLoopBlkLen2 +		/* bottom of inner loop */
		jumpBlkLen2 +		 	/* inner loop jump */
		returnBlkLen2);		/* subroutine return */

    NORMAL_RETURN;
}



/* PUBLIC */
/************************************************************************
 *  i_genDilateAsm():  generate assembly code to do dilate defined by strel
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastDilate as
 *			blockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_SETooBig:		Structure element was too big to
 *				generate accelerated code OR was composed
 *				entirely of DONT_CARE and MISS terms
 *				(which is silly).
 *	ia_bufferExceeded:	The routine tried to generate too much
 *				assembly code.  This is a bug in the routine.
 *				If it happens, please let us know.
 ************************************************************************/
Int32  CDECL 
i_genDilateAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    DilateProc *generated_code,
    Int32      *code_length)
{
    PROLOGUE;
    UInt8      *pCodeBuf;		/* output code buffer */
    Bool	bufAllocatedHere;	/* set if this proc allocated the
					 * asm code buffer. */
    Int32	bufIndex;		/* index of code buffer */
    Int32	bufSize;		/* size of mem allocated for code
					 * buffer */
    Bool	firstActiveFound;	/* cTrue if first active SE
					 * element has been
					 * encountered.  The first SE
					 * element requires different code
					 * than the rest. */
    Int32	cacheRow, cacheCol;	/* index through the SE
					 * and cache */
    Int32	innerLoopOffset;	/* top of inner loop */
    UInt8	*cacheColActive;	/* array of booleans telling whether
					 * each cache col has any HITs or
					 * MISSes below the top row. */
    Int32	lastPosition;		/* used to compute shift distance
					 * for double shift opcodes */
    UInt8	lastShift, shiftDist;	/* used to remember previous shift
					 * distance.  Is the amount by which
					 * we need to shift ECX before the
					 * next shift */
    Bool	activeFound;		/* used to explore SE rows */
    Int32      *pJumpDisplacement;	/* point to jump offset */
    Int32      *pWordDisplacement;	/* point to word address offset */
    Int8	activeOnLeft, activeOnRight, activeInCenter;
    Int32	minRowUsed, maxRowUsed, rowUsed, oneRowUsed;
					/* sometimes all the hits are in
					 * a single row.  If so, we can
					 * dispense with the input cache.
					 * Sometimes an outer row or rows
					 * contains only misses so can be
					 * ignored. */
    Int32	vertSrcShiftDist;	/* distance by which we shift the
					 * source address up or down before
					 * beginning processing. Normally,
					 * we shift the source address up
					 * and fill the input cache.  However,
					 * if the SE has only one row with
					 * hits in it and that row is
					 * above the center, we have to
					 * shift the source address down.
					 */
    Int32	leftmostCol, rightmostCol;
					/* booleans used to decide when to
					 * generate code */
    
    Int32       seRowOffset, cacheRowOffset; /* holds (row * numSECols)
					      * for SE and cache respectively.
					      */


	/* First, decide if the SE is too big. */
    if ((centerX > cMaxShiftDist) ||
	(centerY > cMaxShiftDist) ||
	((numSECols - centerX - 1) > cMaxShiftDist) ||
	((numSERows - centerY - 1) > cMaxShiftDist) )
    {
        return ia_SETooBig;
    }

	/* see if the strel is composed entirely of cDontCare and
	 * cMiss terms. If so, fail as above.
	 */
    __error__ = ia_SETooBig;
    for (seRowOffset = 0;
	 (seRowOffset < ny*nx) && (__error__ == ia_SETooBig);
	 seRowOffset++)
    {
	if (data[seRowOffset] == cHit)
	{
	    __error__ = ia_successful;
	    break;
	}
    }
	/* if we didn't find any hits, return now */
    if (__error__ == ia_SETooBig)
	return ia_SETooBig;

	/* the strel's blockWidth is 4 bytes in all cases, so we might as well
	   set it now. */
    *blockWidth = 4;

	/* if we don't already have an assembly buffer, get one. */
    i_getDilateAsmSize (pIPState, nx, ny, cx, cy, &bufSize);

	/* In Visual C++, turn off the warning that says we're about to
	 * cast a pointer to a function to a pointer to data, since
	 * that's exactly what we intend to do.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    pCodeBuf = (UInt8 *) *generated_code;
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

    if (pCodeBuf == NULL)
    {
	pCodeBuf = (UInt8 *) MALLOC(bufSize);
	bufAllocatedHere = cTrue;
    }
    else
	bufAllocatedHere = cFalse;

	/* if the pointer is null, punt.  Otherwise, return the address
	 * of the new buffer.
	 */
    if (pCodeBuf == NULL)
        return ia_nomem;
    else
	/* In Visual C++, turn off the warning that says we're casting
	 * a pointer to data to a pointer to a function.  We're going
	 * to do exactly that.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
	*generated_code = (DilateProc) pCodeBuf;
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif

	/* Initialize the index into the code buffer.  We use an index
	 * because we usually want to use the offset into the buffer
	 * to compute things, not the absolute buffer pointer.
	 */
    bufIndex = 0;

	/* Trim the SE down so only the rows with hits in them count.
	 * This will help us avoid extraneous uses of the input cache.
	 */
    oneRowUsed = cTrue;
    minRowUsed = numSERows;
    maxRowUsed = -1;
    rowUsed = -1;
    for (cacheRow = 0, seRowOffset = (numSERows-1) * numSECols;
	 cacheRow < numSERows;
	 seRowOffset -= numSECols, cacheRow++)
    {
	for (cacheCol = 0; cacheCol < numSECols; cacheCol++)
	{
	    if (data[seRowOffset + (numSECols-cacheCol-1)] == cHit)
	    {
		    /* find the extent of the HITs in the SE */
		if (cacheRow < minRowUsed) minRowUsed = cacheRow;
		if (cacheRow > maxRowUsed) maxRowUsed = cacheRow;
	    }
	}
    }

	/* Set the amount by which we'll adjust the source address
	 * assuming nothing is trimmed. */
    vertSrcShiftDist = numSERows - centerY - 1;

	/* Can we trim rows off the top of the SE?  Note that since
	 * we address the SE from bottom to top, we look for maxRowUsed
	 * here. */
    if (maxRowUsed < numSERows-1)
    {
	    /* We need to advance the pointer "data" beyond these rows
	     * and subtract them from numSERows.  Since vertSrcShiftDist
	     * is measured from the bottom of the SE, it need not
	     * change.
	     */
	data += (numSERows-maxRowUsed-1)*numSECols;
	numSERows = maxRowUsed+1;
	centerY = MIN(centerY, numSERows-1);
    }

	/* Can we trim rows off the bottom of the SE? */
    if (minRowUsed > 0)
    {
	    /* No need to advance the pointer "data", we're just pulling
	     * in the end of the SE.  However, we need to adjust the
	     * center and vertSrcShiftDist since they're relative to
	     * the bottom. */
	vertSrcShiftDist -= minRowUsed;
	centerY = MAX(centerY-minRowUsed, 0);
	numSERows -= minRowUsed;
    }


	/* Record the presence of active elements for each column
	 * in the remaining SE rows.
	 * We don't need to store a shifted input for a column if
	 * it has no active elements below the top row.  Note that
	 * here the terms "right" and "left" refer to data to the input cache
	 * as they did in pcgenero, not to the SE.  We are indexing
	 * the SE in reverse order so we can use the same series
	 * of shifts and loops as in pcgenero.c.
	 */
    cacheColActive = (UInt8 *)CALLOC(numSECols, sizeof(UInt8));

    activeOnLeft = activeOnRight = activeInCenter = cFalse;
    leftmostCol = numSECols-1;
    rightmostCol = 0;
    for (cacheCol = 0; cacheCol < numSECols; cacheCol++)
    {
	for (cacheRow = 0, seRowOffset = (numSERows-1) * numSECols;
	     cacheRow < numSERows;
	     seRowOffset -= numSECols, cacheRow++)
	{
	    if (data[seRowOffset + (numSECols-cacheCol-1)] == cHit)
	    {
		if (cacheRow < numSERows-1) cacheColActive[cacheCol] = cTrue;
		rightmostCol = cacheCol;
		if (cacheCol < leftmostCol) leftmostCol = cacheCol;
		if (cacheCol < cacheCenterX) activeOnLeft = cTrue;
		else
		if (cacheCol == cacheCenterX) activeInCenter = cTrue;
		else activeOnRight = cTrue;
	    }
	}
    }

/* *************** START OF CODE GENERATION ********************** */


	/* Start with the initialization code */
    M_COPY_CODE(erT2Start, startBlkLen2);

	/* We now move the source address up by vertSrcShiftDist
	 * image lines.  This is how the strel->cy parameter affects the output.
	 * Note that nothing happens if the SE only has one row (since
	 * centerY must be 0 in that case).
	 * If vertSrcShiftDist is negative, we move the source addr
	 * down instead of up.  This can only happen if there was only
	 * one active row in the SE and it was above the center.
	 */
    if (vertSrcShiftDist > 0)
    {
	for (cacheRow = 0; cacheRow < vertSrcShiftDist; cacheRow++)
	    M_COPY_CODE(erT2PreLoadDec, preLoadDecBlkLen2);
    }
    else
    {
	vertSrcShiftDist = -vertSrcShiftDist;
	for (cacheRow = 0; cacheRow < vertSrcShiftDist; cacheRow++)
	    M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
    }

	/* load the address of the input cache into EDI register.  We
	 * do this now to avoid the interlock between the last adjustment
	 * of ESI above and the first use of ESI. */
    if (numSERows > 1)
	M_COPY_CODE(erT2PreLoadInpCac, preLoadInpCacBlkLen2);

	/* load the dest address into the EDX register. */
    M_COPY_CODE(erT2LoadDestAddr, loadDestAddrBlkLen2);


/* *********************** PRELOAD INPUT CACHE *********************** */


	/* Preload the input cache.  This is only done if there is more
	 * than one row in the SE. */
    for (cacheRow = 0, seRowOffset = (numSERows-1)*numSECols,
		cacheRowOffset = 0;
	 cacheRow < numSERows-1 ;
	 seRowOffset -= numSECols, cacheRowOffset += numSECols, cacheRow++)
    {
	    /* There's bound to be something active somewhere in the SE
	     * so load the center's word.  Note that this word may well be
	     * used to create shifted inputs that will be used by many
	     * different rows of SE processing (NOT just this row!!, that's
	     * the whole point of keeping a cache of shifted inputs). */
	M_COPY_CODE(erT2LoadInputCenter, loadInputCenterBlkLen2);

	    /* If there are no active cols to the left or right, increment the
	     * address pointer as soon as possible to avoid interlocks. */
	if (!activeOnLeft && !activeOnRight)
	{
	    M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
	}

	    /* if the center's column is active, store this word in the
	     * cache of shifted inputs. */
	if (cacheColActive[cacheCenterX])
	{
	    M_STORE_CACHE(cacheRowOffset, cacheCenterX);
	}

	/* ********************* PRELOAD LEFT SIDE OF CACHE *************** */


	    /* If there are any active cache columns to the left of center,
	     * load the word to the left. */
	if (activeOnLeft)
	{
	    M_COPY_CODE(erT2LoadInputLeft, loadInputLeftBlkLen2);

		/* If there are no active elements to the right, increment the
		 * address pointer as soon as possible to
		 * avoid interlocks. */
	    if (!activeOnRight)
	    {
		M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
	    }

		/* set activeFound to cFalse here.  We'll set it to true
		 * when the first double shift has been done.  Each subsequent
		 * double shift will be immediately preceeded with a
		 * single shift of ECX so the incoming bits are correct
	 	 * (what a weird chip!)
		 */
	    lastPosition = cacheCenterX;
	    activeFound = cFalse;
	    lastShift = 0;		/* keep the compiler happy */

		/* Shift and store each version of input if the column is
		 * active. */
	    for (cacheCol = cacheCenterX - 1; cacheCol >= 0; cacheCol--)
	    {
		if (cacheColActive[cacheCol])
		{
		    if (activeFound) /* have already put out an shrd, so
					need to preceed the next one with
					an shr of ecx */
		    {
			M_COPY_CODE(erT2SingleRightShift,
				        singleRightShiftBlkLen2);
			pCodeBuf[bufIndex-1] = lastShift;
		    }
		    activeFound = cTrue;
		    M_COPY_CODE(erT2ShiftInputRight,
				   shiftInputRightBlkLen2);
			/* fix up the shift distance */
		    lastShift = (UInt8) (lastPosition - cacheCol);
		    pCodeBuf[bufIndex-1] = lastShift;
		    lastPosition = cacheCol;
			/* Store the shifted input into the cache. */
		    M_STORE_CACHE(cacheRowOffset, cacheCol);
		}
	    }  /* end of preloading input cache with left side of row */

		/* If there are any cols on the right, reload the center's
		 * value.  That way, the code below can be assured that
		 * it is entered with the unshifted data in EBX. */
	    if (activeOnRight)
	    {
		M_COPY_CODE(erT2LoadInputCenter, loadInputCenterBlkLen2);
	    }

	}	/* end of case where there are cols to the left of center
		 * when preloading input cache. */



	/* ********************* PRELOAD RIGHT SIDE ********************** */


	    /* If there are columns to the right of center, preload them too. */
	if (activeOnRight)
	{
		/* load the word to the right */
	    M_COPY_CODE(erT2LoadInputRight, loadInputRightBlkLen2);

		/* Increment the address pointer as soon as possible to
		 * avoid interlocks if we haven't already done so. */
	    M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);

		/* remember where we started shifting.  Each left shift
		 * by 1 will be done with two adds (2 cycles).  Left
		 * shifts by more than 1 will use two shifts (4 cycles).
		 */
	    lastPosition = cacheCenterX;

		/* Shift and store each version of input if the column is
		 * active. */
	    for (cacheCol = cacheCenterX + 1; cacheCol < numSECols; cacheCol++)
	    {
		if (cacheColActive[cacheCol])
		{
		    shiftDist = (UInt8) (cacheCol - lastPosition);
		    if (shiftDist > 1) /* shift by more than 1 */
		    {
			M_COPY_CODE(erT2ShiftInputLeft, shiftInputLeftBlkLen2);
			pCodeBuf[bufIndex-1] = shiftDist;
			if (cacheCol < rightmostCol) /* shift ECX to catch up */
			{
			    M_COPY_CODE(erT2SingleLeftShift,
					   singleLeftShiftBlkLen2);
			    pCodeBuf[bufIndex-1] = shiftDist;
			}
		    }
		    else	/* shift by 1 */
		    {
			M_COPY_CODE(erT2DoubleLeftShift1,
				  doubleLeftShift1BlkLen2);
		    }
			/* fix up the shift distance */
		    lastPosition = cacheCol;
			/* Store the shifted input into the cache. */
		    M_STORE_CACHE(cacheRowOffset, cacheCol);
		}
	    }  /* end of preloading input cache with right side of row */
	}	/* end of activeOnRight case for preloading cache rows */
    }	/* end of loop that generates code to preload the input cache */



/* *********************** INNER LOOP *********************** */


	 /* Remember this address so we can branch back to it.
 	 */
    innerLoopOffset =  bufIndex;

	/* Go through the cache top to bottom, left to right (through the
	 * SE (bottom to top, right to left).  At each
	 * location, if there's a cHit, generate the appropriate
	 * code to combine the cached input with the current output.
	 * Stop before the last (top) SE row.  Note that nothing will
	 * happen here if the SE has only one row. */
    firstActiveFound = cFalse;
    for (cacheRow = 0, seRowOffset = (numSERows-1)*numSECols,
		cacheRowOffset = 0;
	 cacheRow < numSERows-1;
	 seRowOffset -= numSECols, cacheRowOffset += numSECols, cacheRow++)
    {
	for (cacheCol = 0; cacheCol < numSECols; cacheCol++)
	{
	    if (data[seRowOffset+(numSECols-cacheCol-1)] == cHit)
	    {
		M_GEN_DILATE_SE_ELEMENT(cacheRowOffset, cacheCol, cFalse);
	    }
	}
    }


/* ********** INNER LOOP: BOTTOM LINE OF INPUT (TOP LINE OF SE) ********** */

	/* This looks a lot like prefilling the input cache, except we
	 * only do one row.  It has two other differences.  First, if
	 * necessary, we combine each shifted input with the current 
	 * output before storing it.  Second, we only store the input 
	 * if the SE is not horizontal (will need the input again sometime).
	 */

	/* There's no need to fixup "cacheRow", "seRowOffset" or
	 * "cacheRowOffset".  They are already equal to (numSERows-1), 0 and
	 * (numSERows-1)*numSECols respectively.
	 */

	/* However, we would like to load the destination address into
	 * ECX as soon as possible to avoid interlock.  "As soon as
	 * possible" is immediately after the last shift - or immediately
	 * before the last store to the input cache.  We figure out
	 * when that is by using rightmostCol and leftmostCol.  They were
	 * set above.  We need leftmostCol when we're working from right
	 * to left filling the left side of the cache.  We use rightmost
	 * col when working from left to right filling the right side
	 * of the cache.
	 */

	/* There's bound to be something active somewhere in the SE
	 * so load the center's word.  Note that this word may well be
	 * used to create shifted inputs that will be used by many
	 * different rows of SE processing (NOT just this row!!, that's
	 * the whole point of keeping a cache of shifted inputs).
	 */
    M_COPY_CODE(erT2LoadInputCenter, loadInputCenterBlkLen2);

	/* If there are no active cols to the left or right, increment the
	 * address pointer as soon as possible to avoid interlocks.  Note
	 * that "left" and "right" refer to direction in the input cache,
	 * not in the SE.
	 */
    if (!activeOnLeft && !activeOnRight)
    {
	M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
    }

	/* if the center's column is active below the top row,
	 * store this word in the cache of shifted inputs. */
    if (cacheColActive[cacheCenterX])
    {
	M_STORE_CACHE(cacheRowOffset, cacheCenterX);
    }

	/* If this element of the SE is active, combine the unshifted
	 * input with the current dest value.  Note the lastRow parameter
	 * is passed as cTrue here. */
    if (data[seRowOffset+centerX] == cHit)
    {
	M_GEN_DILATE_SE_ELEMENT(cacheRowOffset, cacheCenterX, cTrue);
    }


    /* **************** PROCESS LEFT SIDE OF BOTTOM LINE ************* */


	/* If there are any active cache columns to the left of center, load
	 * the word to the left. */
    if (activeOnLeft)
    {
	M_COPY_CODE(erT2LoadInputLeft, loadInputLeftBlkLen2);

	    /* If there are no active elements to the right, increment the
	     * address pointer as soon as possible to avoid interlocks. */
	if (!activeOnRight)
	{
	    M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
	}

	    /* set activeFound to cFalse here.  We'll set it to true
	     * when the first double shift has been done.  Each subsequent
	     * double shift will be immediately preceeded with a
	     * single shift of ECX so the incoming bits are correct
	     * (what a weird chip!)
	     */
	lastPosition = cacheCenterX;
	activeFound = cFalse;
	lastShift = 0;		/* keep the compiler happy */

	    /* Shift and store each version of input if the column is
	     * active. */
	for (cacheCol = cacheCenterX - 1; cacheCol >= 0; cacheCol--)
	{
	    if (cacheColActive[cacheCol] ||
		(data[seRowOffset+(numSECols-cacheCol-1)] == cHit))
	    {
		if (activeFound) /* have already put out an shrd, so
				    need to preceed the next one with
				    an shr of ECX */
		{
		    M_COPY_CODE(erT2SingleRightShift,
				   singleRightShiftBlkLen2);
		    pCodeBuf[bufIndex-1] = lastShift;
		}
		activeFound = cTrue;
		M_COPY_CODE(erT2ShiftInputRight, shiftInputRightBlkLen2);
		    /* fix up the shift distance */
		lastShift = (UInt8) (lastPosition - cacheCol);
		pCodeBuf[bufIndex-1] = lastShift;
		lastPosition = cacheCol;

		    /* Store the shifted input into the cache if we'll
		     * ever need this input again. */
		if (cacheColActive[cacheCol])
		{
		    M_STORE_CACHE(cacheRowOffset, cacheCol);
		}

		    /* If there's an active element at this SE location,
		     * combine the shifted input with the current dest.
		     * There's no way to miss this statement if it's needed
		     * because if there's something at this SE location,
		     * this SE column must be active. */
		if (data[seRowOffset+(numSECols-cacheCol-1)] == cHit)
		{
		    M_GEN_DILATE_SE_ELEMENT(cacheRowOffset, cacheCol, cTrue);
		}
	    }
	}  /* end of processing input cache with left side of row */

	    /* If there are any cols on the right, reload the center's
	     * value.  That way, the code below can be assured that
	     * it is entered with the unshifted data in EBX. */
	if (activeOnRight)
	{
	    M_COPY_CODE(erT2LoadInputCenter, loadInputCenterBlkLen2);
	}

    }	/* end of case where there are cols to the left of center
	 * when processing bottom line of SE. */


    /* **************** PROCESS RIGHT SIDE OF BOTTOM LINE ************* */


	/* If there are columns to the right of center, process them too. */
    if (activeOnRight)
    {
	    /* load the word to the right */
	M_COPY_CODE(erT2LoadInputRight, loadInputRightBlkLen2);

	    /* Increment the address pointer as soon as possible to
	     * avoid interlocks if we haven't already done so. */
	M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);

	    /* remember where we started shifting.  Each left shift
	     * by 1 will be done with two adds (2 cycles).  Left
	     * shifts by more than 1 will use two shifts (4 cycles).
	     */
	lastPosition = cacheCenterX;

	    /* Shift and store each version of input if the column is
	     * active. */
	for (cacheCol = cacheCenterX + 1; cacheCol < numSECols; cacheCol++)
	{
	    if (cacheColActive[cacheCol] ||
		(data[seRowOffset+(numSECols-cacheCol-1)] == cHit))
	    {
		shiftDist = (UInt8) (cacheCol - lastPosition);
		if (shiftDist > 1) /* shift by more than 1 */
		{
		    M_COPY_CODE(erT2ShiftInputLeft, shiftInputLeftBlkLen2);
		    pCodeBuf[bufIndex-1] = shiftDist;
		    if (cacheCol < rightmostCol) /* shift ECX to catch up */
		    {
			M_COPY_CODE(erT2SingleLeftShift,
				   singleLeftShiftBlkLen2);
			pCodeBuf[bufIndex-1] = shiftDist;
		    }
		}
		else	/* shift by 1 */
		{
		    M_COPY_CODE(erT2DoubleLeftShift1,
			      doubleLeftShift1BlkLen2);
		}
		    /* fix up the shift distance */
		lastPosition = cacheCol;

		    /* Store the shifted input into the cache if we'll
		     * ever need this input again. */
		if (cacheColActive[cacheCol])
		{
		    M_STORE_CACHE(cacheRowOffset, cacheCol);
		}

		    /* If there's an active element at this SE location,
		     * combine the shifted input with the current dest.
		     * There's no way to miss this statement if it's needed
		     * because if there's something at this SE location,
		     * this SE column must be active. */
		if (data[seRowOffset+(numSECols-cacheCol-1)] == cHit)
		{
		    M_GEN_DILATE_SE_ELEMENT(cacheRowOffset, cacheCol, cTrue);
		}
	    }
	}  /* end of processing input cache with right side of row */
    }	/* end of activeOnRight case for processing cache rows */


	/* ENTIRE SE HAS BEEN PROCESSED FOR ONE OUTPUT WORD */

	/* If we're using the input cache, increment the input cache
	 * pointer.  There are two possible types of increment.  One
	 * uses a byte constant and the other uses a 32-bit immediate.  
	 * We use the byte immediate if the number of bytes/line in 
	 * the *input cache* is less than 128.
	 */
    if (numSERows > 1)
    {
	if (numSECols < 128/4)
	{
	    M_COPY_CODE(erT2IncCachePtrShort, incCachePtrShortBlkLen2);
	    pCodeBuf[bufIndex-1] = (UInt8) (4*numSECols);
	}
	else
	{
	    M_COPY_CODE(erT2IncCachePtrLong, incCachePtrLongBlkLen2);
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = (Int32) (4*numSECols);
	}
    }

	/* put out the code for the bottom of the loop.  This writes
	 * the value to memory, increments the dest address down one
	 * word and decrements the height count. */
    M_COPY_CODE(erT2BottomLoop, bottomLoopBlkLen2);

	/* and the jump at the bottom of the inner loop */
    M_COPY_CODE(erT2LoopJmp,  jumpBlkLen2);

	/* fix up the row loop's offset */
    pJumpDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
    *pJumpDisplacement = (Int32) (innerLoopOffset - bufIndex);

	/* and finally, the return code */
    M_COPY_CODE(erT2Return, returnBlkLen2);

	/* Now for a little error checking.  If we've generated more code
	   than fits in the buffer, complain.  This should never happen. */
    if (bufIndex >= bufSize)
    {
	/* fprintf(stderr, " Buffer overflow, memory has been *smashed* in %s\n %d bytes written into buffer of size %d\n",
		procName, bufIndex, bufSize); */
	__error__ = ia_bufferExceeded;
	goto genCleanup;
    }

    *code_length = bufIndex;
    __error__ = ia_successful;

genCleanup:
    FREE(cacheColActive);		/* free the array of booleans */
    if ((__error__ != ia_successful) && (bufAllocatedHere))
	FREE(pCodeBuf);
    NORMAL_RETURN;
}




/* Define the macro used to call ip_copyCode from ip_genDilateSEElement. */
#define	G_COPY_CODE(blkAddr, blkLen) \
		__error__ = ip_copyCode (pIPState, &blkAddr, blkLen, \
				         pCodeBuf, pBufIndex, \
				         bufSize)


/* This is a private procedure */
/* PRIVATE */
/************************************************************************
 *  ip_genDilateSEElement()
 *		ONLY USED IN TO GENERATE x86 ASSEMBLY LANGUAGE.
 *		Decides what type of assembly code to generate for the
 *		specified sub-element.  Normally, a cHit generates an OR
 *		operation.
 *		However, if the output word is being touched for the first
 *		time, a cHit generates a COPY.  Once the choice has been
 *		made, the assembly code is actually put into the output
 *		buffer by ip_copyCode.
 *
 *	currentElement:	Index into input cache; used to generate proper
 *			address displacement for OR op.
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			by ip_copyCode.  This routine will also
 *			update the index.
 *	bufSize:	Max size of the code buffer.  ip_copyCode
 *			will check for buffer overflow before doing
 *			the write.
 *	pFirstActiveFound: Pointer to boolean which is cFalse until the
 *			first operation affecting the output word
 *			has been generated.  This is used to
 *			distinguish between the OR case and the COPY case.
 *			Once any code has been generated which touches
 *			the output, word, this routine will set the
 *			boolean to cTrue.
 *	lastRow:	Boolean.  Set to cTrue if this is the last row of
 *			the SE.  Otherwise, set to cFalse.  The opcodes
 *			generated are different for the two cases.
 *
 *  This routine returns one of two status codes directly from ip_copyCode.
 * They are:
 *	ia_successful:		Code was put into the output buffer.
 *	ia_bufferExceeded:	Tried to put too much code into the output
 *				buffer.  This should never happen.
 *				The calling routine, i_genDilateAsm, will return
 *				ia_bufferExceeded, someone will complain and
 *				we'll debug the code.
 ************************************************************************/
Int32  CDECL 
ip_genDilateSEElement (
    IPState    *pIPState,
    Int32	currentElement,
    UInt8      *pCodeBuf,
    Int32      *pBufIndex,
    Int32	bufSize,
    UInt32     *pFirstActiveFound,
    Int32	lastRow)
{
    PROLOGUE;
    
    Int32	addrDisplacement;	/* number of bytes needed for
					 * address displacement of opcodes */
    Int32      *pWordDisplacement;	/* pts to address displacements */

	/* Is this the last row?  If not, we use the array of shifted inputs
	 * as the source of the operation. */
    if (!lastRow)
    {
	    /* calculate the length of the address displacement needed in
	     * the OR opcode below that accesses the cache of shifted
	     * inputs.
	     * If the offset (4 * currentElement) from the beginning of
	     * the input cache is 0, we use no address displacement.  If the
	     * offset is less than 128 BYTES, we use a single byte
	     * displacement.  If the displacement is bigger than 128 bytes,
	     * we use a load op with a 4 byte displacement.
	     */
	if (currentElement == 0)
	    addrDisplacement = 0;
	else
	if (currentElement < 128/4)
	    addrDisplacement = 1;
	else
	    addrDisplacement = 4;

	    /* Have we put out any processing code yet? */
	if ((*pFirstActiveFound) == cFalse)
	{
		/* nope, this is the first time.  We always load the input into
		 * the partial result.  Choose the load op based on the
		 * value of addrDisplacement.
		 */
	    switch (addrDisplacement) {
	    case 0:
		G_COPY_CODE(erT2CacheLoadEAX0, cacheLoadEAX0BlkLen2);
		break;
	    case 1:
		G_COPY_CODE(erT2CacheLoadEAXShort,
			       cacheLoadEAXShortBlkLen2);
		if (__error__ == ia_successful)
		    pCodeBuf[(*pBufIndex)-1] = (UInt8) (4*currentElement);
		break;
	    default:
		G_COPY_CODE(erT2CacheLoadEAXLong, 
			       cacheLoadEAXLongBlkLen2);
		if (__error__ == ia_successful)
		{
		    pWordDisplacement = (Int32 *) &(pCodeBuf[(*pBufIndex)-4]);
		    *pWordDisplacement = (Int32) (4*currentElement);
		}
		break;
	    }
	    *pFirstActiveFound = cTrue;	/* don't do this again */
	    NORMAL_RETURN;
	}
	else
	{
		/* this isn't the first time, so we'll do an OR with
		 * memory. */
	    switch (addrDisplacement) {
	    case 0:
		G_COPY_CODE(erT2CacheOr0, cacheOr0BlkLen2);
		NORMAL_RETURN;
	    case 1:
		G_COPY_CODE(erT2CacheOrShort,  cacheOrShortBlkLen2);
		if (__error__ == ia_successful)
		    pCodeBuf[(*pBufIndex)-1] = (UInt8) (4*currentElement);
		NORMAL_RETURN;
	    default:
		G_COPY_CODE(erT2CacheOrLong, cacheOrLongBlkLen2);
		if (__error__ == ia_successful)
		{
		    pWordDisplacement =
			    (Int32 *) &(pCodeBuf[(*pBufIndex)-4]);
		    *pWordDisplacement = (Int32) (4*currentElement);
		}
		NORMAL_RETURN;
	    }
	}	/* done processing hits that weren't touching
		 * the dest for the first time and weren't on the last
		 * row of the SE.
		 */
    }	/* done processing hits and misses that weren't on the last line
	 * of the SE. */

    else

    {

    /* ********************* PROCESS LAST ROW ************************* */

	/* This is the last row.  We have the same cases again, except
         * the code generated is different because the source is in EBX.
	 */
	    /* Have we put out any processing code yet?  If the SE has
	     * only one active row, we might not have touched the
	     * destination yet.
	     */
	if ((*pFirstActiveFound) == cFalse)
	{
		/* This is the first time the dest has been touched.
		 * We always move the data from EBX into EAX.
		 */
	    G_COPY_CODE(erT2NonCacheInit, nonCacheInitBlkLen2);

	    *pFirstActiveFound = cTrue;	/* don't do this again */
	    NORMAL_RETURN;
	}
	else
	{
		/* this isn't the first time, so we'll do an OR. */
	    G_COPY_CODE(erT2NonCacheHitOr, nonCacheHitOrBlkLen2);
	    NORMAL_RETURN;
	}
    }	/* done processing hits and misses that *were* on the last line
	 * of the SE. */
}

/*

$Log: pcgendil.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.23  1998/01/13 23:53:57  ddavies
pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.22  1997/02/06 01:36:14  ddavies
pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v

Fix gcc uninitialized var warnings on freebsd.  Fix one apparent real
problem that would occur when loading the input cache for dilation when
there was stuff on the left side of the strel that required shifting by
more than one bit position.  Fixing this didn't change the outputs, but
I feel better about it.

Revision 1.21  1995/12/13 02:52:42  ejaquith
pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.20  1995/10/18  04:04:37  ddavies
 * pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.19  1995/10/03  21:16:11  ddavies
 * pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v
 *
 * Include shrmorph.pub/prv instead of shrip.pub/prv.  Change arg of
 * i_genDilateAsm from "void **generated_code" to "DilateProc *generated_code".
 * Change casts when using generated_code.
 *
 * Revision 1.18  1994/12/08  04:36:51  ddavies
 * pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v
 *
 * Check to see if structure element contains any hits in i_genDilateAsm.
 * If not, return ia_SETooBig.  This really means "the structure element is
 * too weird," in this case.
 *
 * Revision 1.17  1994/09/20  02:51:31  ddavies
 * pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v
 *
 * Giant changes to get rid of allocator, alloc_shrinker and generally allow
 * clients to pass in buffers for generated code.  Add a version of i_getDilateAsmSize
 * for the PC.  Make i_genDilateAsm call it.  Get rid of the generated_code and
 * alloc_shrinker args in all of the private subroutines and macros since we
 * no longer free the buffer on error.
 *
 * Revision 1.16  1994/02/04  20:20:19  dferrell
 * dferrell on Fri Feb 4 12:20:04 PST 1994
 *
 * pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v
 *
 * Changed memory.prv to memory.pub
 *
 * Revision 1.15  1993/12/02  19:25:50  ddavies
 * pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.14  1993/11/05  01:52:24  ddavies
 * pcgendil.c -> /product/ipcore/ipshared/src/RCS/pcgendil.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


