/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  pcgendim.c -- generates MMX assembly routines implementing dilate ops
 */


#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#ifdef __WATCOMC__
#include <string.h>
#endif
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "frames.pub"
#include "memory.pub"

    /* prototypes */
#include "shrmorph.pub"
#include "shrmorph.prv"
#include "shrpixr.pub"
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: pcgendim.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* This code is very closely related to that in pcgenerm.c, except it
 * generates dilation instead of erosion code.  pcgenerm.c has extensive
 * comments.  Here, we'll discuss the difference between these
 * implementations of erosion and dilation.
 *
 * In erosion, we match a pattern in the input image (using hits for 1's,
 * misses for 0's and don't care terms), writing a dot in the output image
 * whereever the pattern matched the input image.  In dilation, we write
 * the pattern (really only the hits from the Structure Element) in the
 * output image at each position where a 1 was found in the input image.
 *
 * One way to implement dilation is to search for all the 1's in the input
 * image.  When a 1 is found at location (x,y), the SE is written into the
 * output image with the SE's defined center location (cx, cy) placed at
 * the location corresponding to the 1 in the input image.  While correct,
 * that is a very slow approach.
 *
 * That approach treats the structure element as a child's stamp and the
 * input image as a pattern that governs where the stamp is applied to the
 * output image.  What happens if we reverse the interpretations - calling
 * the input image the stamp and the SE the pattern in which the stamp is
 * applied?  With some reflection, you can convince yourself that the
 * output is still correct.
 *
 * A nice feature of the second interpretation is that the algorithm can
 * be much more efficient. Instead of constantly inserting a small pattern
 * into a larger one (with the attendent shifting, masking and merging), we
 * simply shift and OR the entire image in a very uniform manner.  If we had
 * a parallel processor, we'd be in really good shape, but even with a serial
 * processor new approach is simpler and faster.
 *
 * There's a fly in the ointment though.  The straightforward implementation
 * of this approach has us grabbing an input bit (or word) and doing a set
 * of shifts and ORs with the surrounding output bits (or words).
 * That means a lot of memory writes.  Memory operations are slow and writes,
 * especially lots of them so the write buffer overflows, are very slow.
 * We'd like to recast the algorithm so that it does lots of reads and
 * a few writes.  Each output word should be completely computed in a
 * register using the surrounding image words, then written once.
 *
 * This isn't too hard, but you have to do another conceptual reversal (these
 * usually take me weeks to get used to).  Normally, we think of the SE
 * as being centered (in the sense of (cx,cy), not the physical center) on
 * the word being computed.  The SE element at ((cx+i),(cy+j)) tells us
 * how to combine the image bit at offset (i,j) from the output bit in
 * question.  This is how we did the erosion code where we naturally moved
 * all the surrounding image bits to the center.
 *
 * In dilation, we natually want to move the center's image bit towards the
 * periphery.  So quit thinking of one SE centered at the current output.
 * Think instead about sliding the SE around into all positions where it still
 * intersects the current output bit.  For example, if the SE were 2 wide and
 * 3 tall, there would be 6 positions in which the SE would intersect
 * a given output bit.  The union of the centers of these SE positions defines
 * the set of bits that can affect the current output bit in a dilation
 * operation.
 *
 * Fine theory, but how do we use it?  For example, say we want to compute
 * the output image bit at location (x,y).  What input image bits should we
 * consider?  To see how far up to look, we place (x,y) in the *bottom* row
 * of the SE.  The distance to the center (the highest bit that could
 * affect this output bit) is (ny-cy-1) rows, so the highest row we need
 * to inspect is y - (ny-cy-1).  Similarly, the distance from the top row
 * of the SE to the center is cy, so we need to look at image bits that are
 * as much as cy rows below y.  In all, we need to inspect all bits at offsets
 * (dx, dy) from (x,y) where {-(nx-cx-1) <= dx <= cx} and
 * {-(ny-cy-1) <= dy <= cy}.  When considering an input image bit at offset
 * (dx, dy) from our output, we use SE element (cx-dx, cy-dy) to determine
 * the operation (cHit => OR; cDontCare or cMiss => nil) to be performed.
 * 
 * How does this interact with our cache of shifted inputs?  In the erosion
 * code, things were very nice since the offset into the SE and the offset
 * into the input cache were the same.  Can we do that here?
 *
 * No.  An important feature of the cache of inputs is that when we go down
 * one line in the source and dest images, we go down one line in the
 * input cache.  Well, ok, we could probably go through the input cache
 * from bottom to top while going through the images from top to bottom,
 * but that's too weird.  The easiest thing to do is to maintain two sets
 * of coordinates: one for the input cache and one for the SE.  Since we
 * still want to go through the input cache top to bottom, left to right;
 * we'll go through the SE bottom to top, right to left.
 *
 * One non-obvious question is "what do we do with MISSes?"  We ignore them.
 * Dilate operations that do something with misses have been defined, but
 * Dan Bloomberg says they're seldom used.
 *
 * As a result of these two points, the code here looks like the code in
 * pcgenerm.c with the following exceptions:
 *   1.	No MISS or HIT&MISS cases.
 *   2. Use ORs instead of ANDs.
 *   3. When loading the input cache, we look at left side of the strel
 *	when deciding how much to shift left and the right side of the
 *	strel when deciding how much to shift right.
 *   4. When deciding what OR operations to generate for a preshifted input
 *	on row R, strel col C and an output word on row Ro, we look at
 *	strel row (cy + (Ro - R)) and strel col C to see if there's a HIT.
 *
 */


    /* dimensions of the largest structure element we can handle */
#define cMaxShiftDist	31

    /* Labels in the template assembly code */
extern	UInt8	mphT3Start, mphT3PreLoadSrcAddrPitch;
extern	UInt8	mphT3PreLoadCacheHeight;
extern	UInt8	mphT3PreLoadCacheAddr, mphT3LoadUpperSrcDWord;
extern	UInt8	mphT3PreLoad1sMask, mphT3PreLoad1sMaskToMM7;
extern	UInt8	mphT3PreLoadUpperLeftSideWord;
extern	UInt8	mphT3PreLoadUpperRightSideWord, mphT3WordSwapUpperStep1;
extern	UInt8	mphT3LoadLowerSrcDWord, mphT3WordSwapUpperStep2;
extern	UInt8	mphT3WordSwapUpperMM2Step1, mphT3WordSwapUpperMM2Step2;
extern	UInt8	mphT3WordSwapLowerMM5Step1, mphT3WordSwapLowerMM5Step2;
extern	UInt8	mphT3PreLoadLowerLeftSideWord, mphT3PreLoadLowerRightSideWord;
extern	UInt8	mphT3WordSwapLowerStep1, mphT3LeftShiftUpperSideWord;
extern	UInt8	mphT3RightShiftUpperSideWord, mphT3WordSwapLowerStep2;
extern	UInt8	mphT3CopyUpperDWord, mphT3LeftShiftLowerSideWord;
extern	UInt8	mphT3RightShiftLowerSideWord, mphT3LeftShiftUpperScratch;
extern	UInt8	mphT3RightShiftUpperScratch, mphT3CopyLowerDWord;
extern	UInt8	mphT3CombineShiftedUpperDWord, mphT3LeftShiftLowerScratch;
extern	UInt8	mphT3RightShiftLowerScratch, mphT3CombineShiftedLowerDWord;
extern	UInt8	mphT3WriteUpperScratchToCacheDisp8;
extern	UInt8	mphT3WriteUpperScratchToCacheDisp32;
extern	UInt8	mphT3WriteLowerScratchToCacheDisp8;
extern	UInt8	mphT3WriteLowerScratchToCacheDisp32;
extern	UInt8	mphT3WriteUpperScratchToDest;
extern	UInt8	mphT3WriteLowerScratchToDest;
extern	UInt8	mphT3InvertUpperShiftedDWord, mphT3InvertLowerShiftedDWord;
extern	UInt8	mphT3InvertAndUpperShiftedDWord;
extern	UInt8	mphT3InvertAndLowerShiftedDWord;
extern	UInt8	mphT3InvertUpperOrigDWord, mphT3InvertLowerOrigDWord;
extern	UInt8	mphT3InvertUpperOrigDWordToMM6, mphT3InvertLowerOrigDWordToMM7;
extern	UInt8	mphT3WriteUpperOrigToCacheDisp8;
extern	UInt8	mphT3WriteUpperOrigToCacheDisp32;
extern	UInt8	mphT3WriteLowerOrigToCacheDisp8;
extern	UInt8	mphT3WriteLowerOrigToCacheDisp32;
extern	UInt8	mphT3AdjustSrcAddrBy2Lines;
extern	UInt8	mphT3AdjustDstAddrBy2Lines;
extern	UInt8	mphT3AdjustCachePtr;
extern	UInt8	mphT3DecrRowsBy2AndBranch;
extern	UInt8	mphT3DecrRowsBy2;
extern	UInt8	mphT3JABranch;
extern	UInt8	mphT3LoadUpperShiftedDWord, mphT3LoadLowerShiftedDWord;
extern	UInt8	mphT3LoadUpperOrigDWord, mphT3LoadLowerOrigDWord;
extern	UInt8	mphT3AndUpperShiftedDWord, mphT3AndLowerShiftedDWord;
extern	UInt8	mphT3AndUpperShiftedDWordToMM2, mphT3AndLowerShiftedDWordToMM5;
extern	UInt8	mphT3OrUpperShiftedDWordToMM2, mphT3OrLowerShiftedDWordToMM5;
extern	UInt8	mphT3OrUpperShiftedDWord, mphT3OrLowerShiftedDWord;
extern	UInt8	mphT3LoadBlockHeight, mphT3LoadDstAddr, mphT3LoadDstPitch;
extern	UInt8	mphT3TestBlockHeight, mphT3LoadMM0Cache8;
extern	UInt8	mphT3LoadMM1Cache8, mphT3LoadMM2Cache8;
extern	UInt8	mphT3LoadMM3Cache8, mphT3LoadMM4Cache8;
extern	UInt8	mphT3LoadMM5Cache8, mphT3LoadMM6Cache8;
extern	UInt8	mphT3LoadMM7Cache8, mphT3LoadMM0Cache32;
extern	UInt8	mphT3LoadMM1Cache32, mphT3LoadMM2Cache32;
extern	UInt8	mphT3LoadMM3Cache32, mphT3LoadMM4Cache32;
extern	UInt8	mphT3LoadMM5Cache32, mphT3LoadMM6Cache32;
extern	UInt8	mphT3LoadMM7Cache32, mphT3LoadMM0Scratch;
extern	UInt8	mphT3LoadMM1Scratch, mphT3LoadMM2Scratch;
extern	UInt8	mphT3LoadMM3Scratch, mphT3LoadMM4Scratch;
extern	UInt8	mphT3LoadMM5Scratch, mphT3LoadMM6Scratch;
extern	UInt8	mphT3AndMM0Cache8;
extern	UInt8	mphT3AndMM1Cache8, mphT3AndMM2Cache8;
extern	UInt8	mphT3AndMM3Cache8, mphT3AndMM4Cache8;
extern	UInt8	mphT3AndMM5Cache8, mphT3AndMM6Cache8;
extern	UInt8	mphT3AndMM0Cache32, mphT3AndMM1Cache32;
extern	UInt8	mphT3AndMM2Cache32, mphT3AndMM3Cache32;
extern	UInt8	mphT3AndMM4Cache32, mphT3AndMM5Cache32;
extern	UInt8	mphT3AndMM6Cache32, mphT3AndMM0Scratch;
extern	UInt8	mphT3AndMM1Scratch, mphT3AndMM2Scratch;
extern	UInt8	mphT3AndMM3Scratch, mphT3AndMM4Scratch;
extern	UInt8	mphT3AndMM5Scratch, mphT3AndMM6Scratch;
extern	UInt8	mphT3OrMM0Cache8, mphT3OrMM1Cache8, mphT3OrMM2Cache8;
extern	UInt8	mphT3OrMM3Cache8, mphT3OrMM4Cache8, mphT3OrMM5Cache8;
extern	UInt8	mphT3OrMM6Cache8, mphT3OrMM0Cache32, mphT3OrMM1Cache32;
extern	UInt8	mphT3OrMM2Cache32, mphT3OrMM3Cache32, mphT3OrMM4Cache32;
extern	UInt8	mphT3OrMM5Cache32, mphT3OrMM6Cache32, mphT3OrMM0Scratch;
extern	UInt8	mphT3OrMM1Scratch, mphT3OrMM2Scratch, mphT3OrMM3Scratch;
extern	UInt8	mphT3OrMM4Scratch, mphT3OrMM5Scratch, mphT3OrMM6Scratch;
extern	UInt8	mphT3WriteSubBlockStart, mphT3WriteSubBlockIncCachePtr;
extern	UInt8	mphT3WriteSubBlockRemainder, mphT3WriteVertSubBlockStart;
extern	UInt8	mphT3WriteVertSubBlockRemainder, mphT3WriteTinySubBlockStart;
extern	UInt8	mphT3WriteTinySubBlockRemainder;
extern	UInt8	mphT3WriteTinyVertSubBlockStart;
extern	UInt8	mphT3WriteTinyVertSubBlockRemainder, mphT3Return;

/* block lengths */
extern	UInt32	mphT3StartBlkLen, mphT3PreLoadSrcAddrPitchBlkLen;
extern	UInt32	mphT3PreLoadCacheHeightBlkLen;
extern	UInt32	mphT3PreLoadCacheAddrBlkLen, mphT3LoadUpperSrcDWordBlkLen;
extern	UInt32	mphT3PreLoad1sMaskBlkLen, mphT3PreLoad1sMaskToMM7BlkLen;
extern	UInt32	mphT3PreLoadUpperLeftSideWordBlkLen;
extern	UInt32	mphT3PreLoadUpperRightSideWordBlkLen;
extern	UInt32	mphT3WordSwapUpperStep1BlkLen;
extern	UInt32	mphT3LoadLowerSrcDWordBlkLen, mphT3WordSwapUpperStep2BlkLen;
extern	UInt32	mphT3WordSwapUpperMM2Step1BlkLen;
extern	UInt32	mphT3WordSwapUpperMM2Step2BlkLen;
extern	UInt32	mphT3WordSwapLowerMM5Step1BlkLen;
extern	UInt32	mphT3WordSwapLowerMM5Step2BlkLen;
extern	UInt32	mphT3PreLoadLowerLeftSideWordBlkLen;
extern	UInt32	mphT3PreLoadLowerRightSideWordBlkLen;
extern	UInt32	mphT3WordSwapLowerStep1BlkLen;
extern	UInt32	mphT3LeftShiftUpperSideWordBlkLen;
extern	UInt32	mphT3RightShiftUpperSideWordBlkLen;
extern	UInt32	mphT3WordSwapLowerStep2BlkLen;
extern	UInt32	mphT3CopyUpperDWordBlkLen, mphT3LeftShiftLowerSideWordBlkLen;
extern	UInt32	mphT3RightShiftLowerSideWordBlkLen;
extern	UInt32	mphT3LeftShiftUpperScratchBlkLen;
extern	UInt32	mphT3RightShiftUpperScratchBlkLen, mphT3CopyLowerDWordBlkLen;
extern	UInt32	mphT3CombineShiftedUpperDWordBlkLen;
extern	UInt32	mphT3LeftShiftLowerScratchBlkLen;
extern	UInt32	mphT3RightShiftLowerScratchBlkLen;
extern	UInt32	mphT3CombineShiftedLowerDWordBlkLen;
extern	UInt32	mphT3WriteUpperScratchToCacheDisp8BlkLen;
extern	UInt32	mphT3WriteUpperScratchToCacheDisp32BlkLen;
extern	UInt32	mphT3WriteLowerScratchToCacheDisp8BlkLen;
extern	UInt32	mphT3WriteLowerScratchToCacheDisp32BlkLen;
extern	UInt32	mphT3WriteUpperScratchToDestBlkLen;
extern	UInt32	mphT3WriteLowerScratchToDestBlkLen;
extern	UInt32	mphT3InvertUpperShiftedDWordBlkLen;
extern	UInt32	mphT3InvertLowerShiftedDWordBlkLen;
extern	UInt32	mphT3InvertAndUpperShiftedDWordBlkLen;
extern	UInt32	mphT3InvertAndLowerShiftedDWordBlkLen;
extern	UInt32	mphT3InvertUpperOrigDWordBlkLen;
extern	UInt32	mphT3InvertLowerOrigDWordBlkLen;
extern	UInt32	mphT3InvertUpperOrigDWordToMM6BlkLen;
extern	UInt32	mphT3InvertLowerOrigDWordToMM7BlkLen;
extern	UInt32	mphT3WriteUpperOrigToCacheDisp8BlkLen;
extern	UInt32	mphT3WriteUpperOrigToCacheDisp32BlkLen;
extern	UInt32	mphT3WriteLowerOrigToCacheDisp8BlkLen;
extern	UInt32	mphT3WriteLowerOrigToCacheDisp32BlkLen;
extern	UInt32	mphT3AdjustSrcAddrBy2LinesBlkLen;
extern	UInt32	mphT3AdjustDstAddrBy2LinesBlkLen;
extern	UInt32	mphT3AdjustCachePtrBlkLen;
extern	UInt32	mphT3DecrRowsBy2AndBranchBlkLen;
extern	UInt32	mphT3DecrRowsBy2BlkLen;
extern	UInt32	mphT3JABranchBlkLen;
extern	UInt32	mphT3LoadUpperShiftedDWordBlkLen;
extern	UInt32	mphT3LoadLowerShiftedDWordBlkLen;
extern	UInt32	mphT3LoadUpperOrigDWordBlkLen;
extern	UInt32	mphT3LoadLowerOrigDWordBlkLen;
extern	UInt32	mphT3AndUpperShiftedDWordBlkLen;
extern	UInt32	mphT3AndLowerShiftedDWordBlkLen;
extern	UInt32	mphT3AndUpperShiftedDWordToMM2BlkLen;
extern	UInt32	mphT3AndLowerShiftedDWordToMM5BlkLen;
extern	UInt32	mphT3OrUpperShiftedDWordToMM2BlkLen;
extern	UInt32	mphT3OrLowerShiftedDWordToMM5BlkLen;
extern	UInt32	mphT3OrUpperShiftedDWordBlkLen, mphT3OrLowerShiftedDWordBlkLen;
extern	UInt32	mphT3LoadBlockHeightBlkLen, mphT3LoadDstAddrBlkLen;
extern	UInt32	mphT3LoadDstPitchBlkLen;
extern	UInt32	mphT3TestBlockHeightBlkLen, mphT3LoadMM0Cache8BlkLen;
extern	UInt32	mphT3LoadMM1Cache8BlkLen, mphT3LoadMM2Cache8BlkLen;
extern	UInt32	mphT3LoadMM3Cache8BlkLen, mphT3LoadMM4Cache8BlkLen;
extern	UInt32	mphT3LoadMM5Cache8BlkLen, mphT3LoadMM6Cache8BlkLen;
extern	UInt32	mphT3LoadMM7Cache8BlkLen, mphT3LoadMM0Cache32BlkLen;
extern	UInt32	mphT3LoadMM1Cache32BlkLen, mphT3LoadMM2Cache32BlkLen;
extern	UInt32	mphT3LoadMM3Cache32BlkLen, mphT3LoadMM4Cache32BlkLen;
extern	UInt32	mphT3LoadMM5Cache32BlkLen, mphT3LoadMM6Cache32BlkLen;
extern	UInt32	mphT3LoadMM7Cache32BlkLen, mphT3LoadMM0ScratchBlkLen;
extern	UInt32	mphT3LoadMM1ScratchBlkLen, mphT3LoadMM2ScratchBlkLen;
extern	UInt32	mphT3LoadMM3ScratchBlkLen, mphT3LoadMM4ScratchBlkLen;
extern	UInt32	mphT3LoadMM5ScratchBlkLen, mphT3LoadMM6ScratchBlkLen;
extern	UInt32	mphT3AndMM0Cache8BlkLen;
extern	UInt32	mphT3AndMM1Cache8BlkLen, mphT3AndMM2Cache8BlkLen;
extern	UInt32	mphT3AndMM3Cache8BlkLen, mphT3AndMM4Cache8BlkLen;
extern	UInt32	mphT3AndMM5Cache8BlkLen, mphT3AndMM6Cache8BlkLen;
extern	UInt32	mphT3AndMM0Cache32BlkLen, mphT3AndMM1Cache32BlkLen;
extern	UInt32	mphT3AndMM2Cache32BlkLen, mphT3AndMM3Cache32BlkLen;
extern	UInt32	mphT3AndMM4Cache32BlkLen, mphT3AndMM5Cache32BlkLen;
extern	UInt32	mphT3AndMM6Cache32BlkLen, mphT3AndMM0ScratchBlkLen;
extern	UInt32	mphT3AndMM1ScratchBlkLen, mphT3AndMM2ScratchBlkLen;
extern	UInt32	mphT3AndMM3ScratchBlkLen, mphT3AndMM4ScratchBlkLen;
extern	UInt32	mphT3AndMM5ScratchBlkLen, mphT3AndMM6ScratchBlkLen;
extern	UInt32	mphT3OrMM0Cache8BlkLen, mphT3OrMM1Cache8BlkLen;
extern	UInt32	mphT3OrMM2Cache8BlkLen;
extern	UInt32	mphT3OrMM3Cache8BlkLen, mphT3OrMM4Cache8BlkLen;
extern	UInt32	mphT3OrMM5Cache8BlkLen;
extern	UInt32	mphT3OrMM6Cache8BlkLen, mphT3OrMM0Cache32BlkLen;
extern	UInt32	mphT3OrMM1Cache32BlkLen;
extern	UInt32	mphT3OrMM2Cache32BlkLen, mphT3OrMM3Cache32BlkLen;
extern	UInt32	mphT3OrMM4Cache32BlkLen;
extern	UInt32	mphT3OrMM5Cache32BlkLen, mphT3OrMM6Cache32BlkLen;
extern	UInt32	mphT3OrMM0ScratchBlkLen;
extern	UInt32	mphT3OrMM1ScratchBlkLen, mphT3OrMM2ScratchBlkLen;
extern	UInt32	mphT3OrMM3ScratchBlkLen;
extern	UInt32	mphT3OrMM4ScratchBlkLen, mphT3OrMM5ScratchBlkLen;
extern	UInt32	mphT3OrMM6ScratchBlkLen;
extern	UInt32	mphT3WriteSubBlockStartBlkLen;
extern	UInt32	mphT3WriteSubBlockIncCachePtrBlkLen;
extern	UInt32	mphT3WriteSubBlockRemainderBlkLen;
extern	UInt32	mphT3WriteVertSubBlockStartBlkLen;
extern	UInt32	mphT3WriteVertSubBlockRemainderBlkLen;
extern	UInt32	mphT3WriteTinySubBlockStartBlkLen;
extern	UInt32	mphT3WriteTinySubBlockRemainderBlkLen;
extern	UInt32	mphT3WriteTinyVertSubBlockStartBlkLen;
extern	UInt32	mphT3WriteTinyVertSubBlockRemainderBlkLen, mphT3ReturnBlkLen;

/* the Opcode arrays loadFromScratch, loadFromCache, writeLowerOrig,
 * writeUpperOrig, writeLowerScratch and writeUpperScratch are defined
 * in pcgenerm.c.
 */
extern OpcodeBlock loadFromScratch;
extern OpcodeBlock loadFromCache;
extern OpcodeBlock writeLowerOrig;
extern OpcodeBlock writeUpperOrig;
extern OpcodeBlock writeLowerScratch;
extern OpcodeBlock writeUpperScratch;

OpcodeBlock orFromCache = {
    {&mphT3OrMM0Cache8, &mphT3OrMM1Cache8,
     &mphT3OrMM2Cache8, &mphT3OrMM3Cache8,
     &mphT3OrMM4Cache8, &mphT3OrMM5Cache8,
     &mphT3OrMM6Cache8, NULL},
    {&mphT3OrMM0Cache8BlkLen, &mphT3OrMM1Cache8BlkLen,
     &mphT3OrMM2Cache8BlkLen, &mphT3OrMM3Cache8BlkLen,
     &mphT3OrMM4Cache8BlkLen, &mphT3OrMM5Cache8BlkLen,
     &mphT3OrMM6Cache8BlkLen, NULL},
    {&mphT3OrMM0Cache32, &mphT3OrMM1Cache32,
     &mphT3OrMM2Cache32, &mphT3OrMM3Cache32,
     &mphT3OrMM4Cache32, &mphT3OrMM5Cache32,
     &mphT3OrMM6Cache32, NULL},
    {&mphT3OrMM0Cache32BlkLen, &mphT3OrMM1Cache32BlkLen,
     &mphT3OrMM2Cache32BlkLen, &mphT3OrMM3Cache32BlkLen,
     &mphT3OrMM4Cache32BlkLen, &mphT3OrMM5Cache32BlkLen,
     &mphT3OrMM6Cache32BlkLen, NULL}};
OpcodeBlock orWithScratch = {
    {&mphT3OrMM0Scratch, &mphT3OrMM1Scratch,
     &mphT3OrMM2Scratch, &mphT3OrMM3Scratch,
     &mphT3OrMM4Scratch, &mphT3OrMM5Scratch,
     &mphT3OrMM6Scratch, NULL},
    {&mphT3OrMM0ScratchBlkLen, &mphT3OrMM1ScratchBlkLen,
     &mphT3OrMM2ScratchBlkLen, &mphT3OrMM3ScratchBlkLen,
     &mphT3OrMM4ScratchBlkLen, &mphT3OrMM5ScratchBlkLen,
     &mphT3OrMM6ScratchBlkLen, NULL},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}};

/* Define macros for common subroutine calls with fancy abort sequences.
 * This should clean up the code a lot since most of the information in
 * the calls does not change.
 */
#define	M_COPY_CODE(blkAddr) \
	    if ((__error__ = ip_copyCode (pIPState, &(blkAddr), \
				        (blkAddr##BlkLen), \
				        pCodeBuf, &bufIndex, bufSize)) \
		!= ia_successful) \
		goto genCleanup

#define	M_COPY_DISP(opcode, reg, disp) \
	    if ((__error__ = ip_longCopyDisp (pIPState, &(opcode), \
					    (reg), (disp), \
					    pCodeBuf, &bufIndex, bufSize)) \
		!= ia_successful) \
		goto genCleanup


/* Define scratch register used when doing ORing */
#define	cScratchReg 7


/* PRIVATE */
/************************************************************************
 *  i_getLongDilateAsmSize():  return number of bytes needed for assembly code
 *			used for dilation with this strel using Long (>= 64 bit)
 *			opcodes.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getLongDilateAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
    PROLOGUE;
    UInt32	blockLen;
    UInt32	numActiveCols;
    UInt32	numActiveElements;
    Int32	longestAccess;	/* longest data type supported on
				 * current platform in bytes.  Must
				 * be 8 for this routine to work */

	/* Reference cx to keep the compiler happy */
    cx = cx;
    cy = cy;

	/* Since we're making worst case assumptions, and don't have
	 * the strel to inspect anyway.  Assume every strel column is
	 * active.  Assume every possible strel element is active.
	 */
    numActiveCols = nx;
    numActiveElements = nx * ny;

	/* See if long accesses are allowed on this platform.  If
	 * not, return ia_noImpl.
	 */
    ip_getLongestAccess (pIPState, &longestAccess);
    if (longestAccess != 8)
	return ia_noImpl;

	/* let's do this a little at a time.  Start with the length
	 * of the prologue code.
	 */
    blockLen = mphT3StartBlkLen + mphT3PreLoadSrcAddrPitchBlkLen +
	       mphT3PreLoadCacheHeightBlkLen +
	       mphT3PreLoadCacheAddrBlkLen;

	/* Now we need to generate code to fill the input cache.  If
	 * This is a purely horizontal strel, we don't actually do
	 * that, but the code isn't all that much smaller, so we
	 * pretend we do.  There's some start up code, then code to
	 * process all the elements to the left of center, then code
	 * to process the center, then code to process all the elements
	 * to the right of center.
	 * We start with the cononical preload.  Preloading for vertical
	 * strels is shorter, but that's ok.
	 */
    blockLen += mphT3LoadUpperSrcDWordBlkLen +
		mphT3PreLoad1sMaskBlkLen +
		mphT3PreLoadUpperLeftSideWordBlkLen +
		mphT3WordSwapUpperStep1BlkLen +
		mphT3LoadLowerSrcDWordBlkLen +
		mphT3WordSwapUpperStep2BlkLen +
		mphT3PreLoadLowerLeftSideWordBlkLen +
		mphT3WordSwapLowerStep1BlkLen +
		mphT3LeftShiftUpperSideWordBlkLen +
		mphT3WordSwapLowerStep2BlkLen +
		mphT3CopyUpperDWordBlkLen +
		mphT3LeftShiftLowerSideWordBlkLen +
		mphT3LeftShiftUpperScratchBlkLen +
		mphT3CopyLowerDWordBlkLen;


	/* For all the SE columns to the left and right of center (yeah
	 * the code for the last col is a little shorter, but so what),
	 * we add the standard loop processing that stores positive
	 * shifted words.
	 */
    blockLen += (numActiveCols - 1) *
		 (mphT3CombineShiftedUpperDWordBlkLen +
		  mphT3LeftShiftLowerScratchBlkLen +
		  mphT3LeftShiftUpperSideWordBlkLen +
		  mphT3CombineShiftedLowerDWordBlkLen +
		  mphT3WriteUpperScratchToCacheDisp32BlkLen +
		  mphT3LeftShiftLowerSideWordBlkLen +
		  mphT3WriteLowerScratchToCacheDisp32BlkLen +
		  mphT3CopyUpperDWordBlkLen +
		  mphT3LeftShiftUpperScratchBlkLen +
		  mphT3CopyLowerDWordBlkLen);

	/* Add in cycles to deal with HITS in the central column */
    blockLen += mphT3WriteUpperOrigToCacheDisp32BlkLen +
		mphT3WriteLowerOrigToCacheDisp32BlkLen;

	/* At the end of the loop, we test and branch back */
    blockLen += mphT3DecrRowsBy2AndBranchBlkLen;

	/* Now, we reinitialize things for the ANDs part of the loop. */
    blockLen += mphT3LoadBlockHeightBlkLen +
		mphT3LoadDstAddrBlkLen +
		mphT3LoadDstPitchBlkLen +
		mphT3PreLoadCacheAddrBlkLen +
		mphT3TestBlockHeightBlkLen;

	/* Now for the ORs and stuff for the subblock.  Since this
	 * is a worst-case analysis, we assume that every OR is with
	 * memory.
	 */
    blockLen += numActiveElements * cLongSubBlockHeight *
		mphT3OrMM0Cache32BlkLen;

	/* Now we write all this stuff out to memory */
    blockLen += mphT3WriteSubBlockStartBlkLen +
		mphT3WriteSubBlockIncCachePtrBlkLen +
		mphT3WriteSubBlockRemainderBlkLen;

	/* Here's the first copy of the return code.  We return instead
	 * of jumping to the return below the 1-line-at-a-time code.
	 */
    blockLen += mphT3ReturnBlkLen;

	/* Here's the one-line-at-a-time code.  It's a lot like the regular
	 * code except that we don't multiply by cLongSubBlockHeight since it's
	 * equal to 1.
	 */
    blockLen += numActiveElements *
		mphT3OrMM0Cache32BlkLen;

	/* Now we write all this stuff out to memory */
    blockLen += mphT3WriteTinySubBlockStartBlkLen +
		mphT3WriteSubBlockIncCachePtrBlkLen +
		mphT3WriteTinySubBlockRemainderBlkLen;

	/* Here's the last copy of the return code.  This is how we
	 * return if we execute the 1-line-at-a-time code.
	 */
    blockLen += mphT3ReturnBlkLen;

	/* return the number we calculated */
    *pNumAsmBytes = blockLen;

    NORMAL_RETURN;
}


/* PUBLIC */
/************************************************************************
 *  i_genLongDilateAsm():  generate assembly code to do dilate defined by strel.
 *			This assembly code makes use of long memory accesses
 *			and operations (long >= 64 bits)
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastDilate as
 *			longBlockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *			If the buffer was passed in, we assume it has at
 *			least the number of bytes that i_getLongDilateAsmSize
 *			would calculate.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_SETooBig:		Structure element was too big to
 *				generate accelerated code OR was composed
 *				entirely of DONT_CARE and MISS terms
 *				OR was a 1x1 strel.
 *	ia_noImpl:		The current platform doesn't support long ops.
 *	ia_nomem:		Not enough memory to allocate the code buffer.
 *	ia_bufferExceeded:	The routine tried to generate too much
 *				assembly code.  This is a bug in the routine.
 *				If it happens, please let us know.
 ************************************************************************/
Int32  CDECL 
i_genLongDilateAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    LongDilateProc *generated_code,
    Int32      *code_length)
{
    PROLOGUE;
    UInt8      *pCodeBuf;		/* output code buffer */
    Bool	bufAllocatedHere;	/* set if this proc allocated the
					 * asm code buffer. */
    Int32      	bufIndex;		/* index of code buffer */
    Int32      	bufSize;		/* size of mem allocated for code
					 * buffer */
    Int32	row, col;		/* index through the SE */
    Int32	cacheWidthInBytes;	/* width of cache in units of
					 * bytes. */
    Int32	cacheHeight;		/* total height of input cache */
    Int32	numActiveCols;		/* number of active SE columns */
    UInt8      *seColActive = NULL;	/* array of booleans telling whether
					 * each SE col has any HITs or
					 * MISSes above the bottom row. */
    Int32	prevCol;		/* the previous active column, used
					 * when filling input cache. */
    Int32	curCol;			/* the current active column, used
					 * when filling input cache. */
    Int32	nextCol;		/* next active column, used when
					 * filling input cache. */
    Int32      *pWordDisplacement;	/* point to word address offset */
    UInt8      *colIDArray = NULL;	/* array of column numbers.  used to
					 * get the SE column of a dword
					 * in an input cache row. */
    Int32	colIDIndex;		/* index into colIDArray */
    Int32	topOfLoopIndex;		/* Offset of top of shift loop from
					 * beginning of buffer. */
    Int32	subBlockJumpIndex;	/* remembers location of jump that
					 * decides where to process an entire
					 * subblock or one line at a time.
					 */
    Int32	topOfLowerLoopIndex;	/* location of top of loop that does
					 * ANDS. */
    Bool	outInitArray[cLongSubBlockHeight];  /* used to decide whether
					 * an output reg has been touched */
    Int32	longestAccess;		/* longest data type supported on
					 * current platform in bytes.  Must
					 * be 8 for this routine to work */
     


	/* First, decide if the SE is too big. */
    if ((cx > cMaxShiftDist) ||
	(cy > cMaxShiftDist) ||
	((nx - cx - 1) > cMaxShiftDist) ||
	((ny - cy - 1) > cMaxShiftDist) ||
	((nx == 1) && (ny == 1)) )
    {
        return ia_SETooBig;
    }

	/* See if long accesses are allowed on this platform.  If
	 * not, return ia_noImpl.
	 */
    ip_getLongestAccess (pIPState, &longestAccess);
    if (longestAccess != 8)
	return ia_noImpl;

	/* Go create an array of active columns.  First, allocate
	 * it.
	 */
    seColActive = (UInt8 *)CALLOC(nx, sizeof(UInt8));
    if (seColActive == NULL)
	return ia_nomem;

    __error__ = ip_getLongDilateSEParms (pIPState, nx, ny,
				       data, &cacheWidthInBytes,
				       &cacheHeight, &numActiveCols,
				       &(seColActive[0]));
    if (__error__ != ia_successful)
	goto genCleanup;

	/* If the SE has no active components, complain */
    if (numActiveCols == 0)
    {
	__error__ = ia_SETooBig;
	goto genCleanup;
    }

	/* our long blockWidth is 8, so just set that now. */
    *blockWidth = 8;

	/* If we don't already have an assembly buffer, figure out how
	 * big one should be and allocate it.
	 * Even if we were given a buffer, assume it's at least this big.
	 * If the real code doesn't fit into a buffer of this
	 * size, i_getLongDilateAsmSize should be fixed.
	 */
    i_getLongDilateAsmSize (pIPState, nx, ny, cx, cy, &bufSize);

	/* In Visual C++, turn off the warning that says we're about to
	 * cast a pointer to a function to a pointer to data, since
	 * that's exactly what we intend to do.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    pCodeBuf = (UInt8 *) *generated_code;
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

    if (pCodeBuf == NULL)
    {
	pCodeBuf = (UInt8 *) MALLOC(bufSize);
	bufAllocatedHere = cTrue;
    }
    else
	bufAllocatedHere = cFalse;

	/* if the pointer is null, punt.  Otherwise, return the address
	 * of the new buffer.
	 */
    if (pCodeBuf == NULL)
        return ia_nomem;
    else
	/* In Visual C++, turn off the warning that says we're casting
	 * a pointer to data to a pointer to a function.  We're going
	 * to do exactly that.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
	*generated_code = (LongDilateProc) pCodeBuf;
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif

	/* allocate the arrays used to convert a displacement into
	 * an input cache row into SE column for the dword recorded there.
	 */
    colIDArray = (UInt8 *) CALLOC((cacheWidthInBytes / 8), sizeof(UInt8));
    if (colIDArray == NULL)
    {
	__error__ = ia_nomem;
	goto genCleanup;
    }
    colIDIndex = 0;		/* start with nothing in here */

	/* Initialize the index into the code buffer.  We use an index
	 * because we usually need to compute things using the *offset*
	 * into the buffer, as opposed to using the absolute buffer
	 * pointer itself.
	 */
    bufIndex = 0;


/* *************** START OF CODE GENERATION ********************** */


	/* Start with the initialization code */
    M_COPY_CODE(mphT3Start);

	/* preload the source address and source address pitch */
    M_COPY_CODE(mphT3PreLoadSrcAddrPitch);

	/* If this is not a purely horizontal strel, we need code
	 * that will load the input cache.  If this is a horizontal
	 * strel (height == 1), we don't load the input cache, but
	 * do a bunch of shifting and writing to the destination.
	 * So, if this is NOT a purely horizontal strel, generate
	 * code to load the height of the input cache and a pointer
	 * to the input cache.  If this is a purely horizontal strel,
	 * generate code to load the dest address and dest image
	 * pitch and the height of the entire block.  Purely horizontal
	 * strels are computed an entire block at a time.
	 */
    if (ny > 1)
    {
	M_COPY_CODE(mphT3PreLoadCacheAddr);
	M_COPY_CODE(mphT3PreLoadCacheHeight);
    }
    else	/* purely horizontal strel */
    {
	M_COPY_CODE(mphT3LoadDstAddr);
	M_COPY_CODE(mphT3LoadDstPitch);
	M_COPY_CODE(mphT3LoadBlockHeight);
    }
	/* Here's where we either fill the input cache (not a purely
	 * horizontal strel) or process the entire block (purely
	 * horizontal strel).  In both cases, we go through the columns
	 * of the strel starting at the center column, then working
	 * from the left edge towards column center - 1, then working from
	 * the right edge towards column center + 1.  ip_mmxCreateShiftCode
	 * puts out the right code for each column.
	 */
    prevCol = curCol = nextCol = -1;	/* no cols seen yet */
    topOfLoopIndex = 0xFFFFFFFF;	/* not initialized yet */

	/* If the center's column is active, it will be the first
	 * column.  Otherwise, go look for the first column.
	 */
    if (seColActive[cx] == cHit)
    {
	nextCol = cx;
    }
    else	/* go look for first col.  Start at left end of strel.
		 * Since we know that there's at least one active
		 * column (we would have given up if there were
		 * no active columns) we can just assume this
		 * search is going to work.
		 */
    {
	for (nextCol = 0; nextCol < cx; nextCol++)
	{
	    if (seColActive[nextCol] == cHit)
		break;
	}
	if (nextCol >= cx)	/* haven't found the first active col yet,
				 * look on the right */
	{
	    for (nextCol = nx-1; nextCol > cx; nextCol--)
	    {
		if (seColActive[nextCol] == cHit)
		    break;
	    }
	}
    }

	/* Here's the loop where we generate columns.  First, update
	 * the current values of curCol and prevCol.  Then go find
	 * the next column to be active.
	 */
    while (nextCol != -1)
    {
	prevCol = curCol;	/* pass the columns down the chain */
	curCol = nextCol;

	    /* if we enter with nextCol == cx, start looking on
	     * the left for the next column.  If we were already
	     * on the left, move right by 1.  If we were already
	     * on the right, move left by 1.
	     */
	if (nextCol == cx)
	    nextCol = 0;
	else
	if (nextCol < cx)
	{
	    nextCol++;
	    if (nextCol == cx)
		nextCol = nx-1;
	}
	else
	if (nextCol > cx)
	{
	    nextCol--;
	    if (nextCol == cx)
		nextCol = -1;	/* there is no next col, so just finish
				 * the current column */
	}

	if (nextCol != -1)
	{
	    if (nextCol < cx)	/* look for next col on left */
	    {
		for (; nextCol < cx; nextCol++)
		{
		    if (seColActive[nextCol] == cHit)
			break;
		}
	    }
    
	    if (nextCol == cx)	/* haven't found one yet */
		nextCol = nx-1;	/* so start on right edge */
    
	    if (nextCol > cx)	/* keep looking on the right */
	    {
		for (; nextCol > cx; nextCol--)
		{
		    if (seColActive[nextCol] == cHit)
			    break;
		}
	    }
    
	    if (nextCol == cx)	/* there are no more columns */
		nextCol = -1;	/* do the last one */
	}

	__error__ =
	    ip_mmxCreateDilateShiftCode (pIPState, prevCol, curCol, nextCol,
					 cx,			/* centerCol */
					 ny,			/* numRows */
					 nx,			/* numCols */
					 cacheWidthInBytes,	/* cacheBpl */
					 colIDArray,
					 &colIDIndex,
					 &topOfLoopIndex,
					 pCodeBuf,
					 &bufIndex,
					 bufSize);
	if (__error__ != ia_successful)
	    goto genCleanup;
    }

	/* Swell, we've generated the code to fill the input cache
	 * (not purely horizontal strel) or have generated the code
	 * to process the entire subblock (purely horizontal strel).
	 * If this isn't a purely horizontal strel, generate the
	 * code to do all the ORs.
	 */
    if (ny > 1)
    {
	    /* First, load the stuff we'll need for the lower code. */
	M_COPY_CODE(mphT3LoadBlockHeight);
	M_COPY_CODE(mphT3LoadDstAddr);
	M_COPY_CODE(mphT3LoadDstPitch);
	M_COPY_CODE(mphT3PreLoadCacheAddr);

	    /* We'll loop back to here after processing one
	     * subblock.  This way, we can test to see if the
	     * next subblock is full height.  If not, we'll
	     * process it one line at a time.
	     */
	topOfLowerLoopIndex = bufIndex;

	    /* Insert the test/jump that used to decide if we're
	     * processing a full subblock.  If not, we process
	     * one row at a time.
	     */
	M_COPY_CODE(mphT3TestBlockHeight);

	    /* Remember the jump`s location so we can fix up the
	     * target address later.
	     */
	subBlockJumpIndex = bufIndex;
    
	    /* Cycle through the rows of the input cache.  We really
	     * only cycle through the rows needed for one subblock.
	     * The code we generate here processes the block in
	     * units of subblocks.  At the bottom, there's a loop
	     * that does additional subblocks.  Note that we
	     * process the rows two at a time, so increment the
	     * row number by 2.  We won't generate any opcodes
	     * for portions of the input cache that aren't used
	     * by a particular subblock.  First, initialize
	     * the array used to decide whether an output
	     * element has been touched yet.
	     */
	for (row = 0; row < cLongSubBlockHeight; row++)
	    outInitArray[row] = cFalse;

	for (row = -(ny - 1 - cy); row < cLongSubBlockHeight + cy; row += 2)
	{
		/* Loop through all the dwords in this input cache
		 * row.  Generate the right sequence of loads and
		 * ORs for each dword.
		 */
	    for (col = 0; col < (cacheWidthInBytes/8); col++)
	    {
		__error__ = ip_mmxCreateORCode (pIPState, nx, ny, cy, data, row,
					     (row+(ny-1-cy))*cacheWidthInBytes +
							(col * 8),
					     colIDArray[col],
					     cacheWidthInBytes,
					     cLongSubBlockHeight,
					     &(outInitArray[0]),
					     pCodeBuf,
					     &bufIndex,
					     bufSize);
		if (__error__ != ia_successful)
		    goto genCleanup;
	    }
	}

	    /* We've generated all the ORs for one subblock.  Generate
	     * the code that writes the subblock to memory.  There are
	     * two such blocks, one for vertical strels (no byteswapping)
	     * and one for strels that need some shifting.
	     */
	if ((numActiveCols > 1) || (seColActive[cx] != cHit))
	{
		/* Strel needs some shifting */
	    M_COPY_CODE(mphT3WriteSubBlockStart);
	    M_COPY_CODE(mphT3WriteSubBlockIncCachePtr);
		/* We increment the rCache register by the difference
		 * between the starting position of this subblock and
		 * the starting position of the next subblock.
		 * That's cLongSubBlockHeight rows in the input cache.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = cLongSubBlockHeight * cacheWidthInBytes;

	    M_COPY_CODE(mphT3WriteSubBlockRemainder);
		/* Set the negative displacement back to the top of the
		 * lower loop.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = topOfLowerLoopIndex - bufIndex;
	}
	else
	{
		/* Strel is purely vertical */
	    M_COPY_CODE(mphT3WriteVertSubBlockStart);
	    M_COPY_CODE(mphT3WriteSubBlockIncCachePtr);
		/* We increment the rCache register by the difference
		 * between the starting position of this subblock and
		 * the starting position of the next subblock.
		 * That's cLongSubBlockHeight rows in the input cache.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = cLongSubBlockHeight * cacheWidthInBytes;

	    M_COPY_CODE(mphT3WriteVertSubBlockRemainder);
		/* Set the negative displacement back to the top of the
		 * lower loop.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = topOfLowerLoopIndex - bufIndex;
	}

	    /* If we exit that code, just return now instead of
	     * jumping around the one-row-at-at-time code. */
	M_COPY_CODE(mphT3Return);

	    /* That was fine.  Now we need to generate the code used to
	     * create one output row at a time.  This looks a lot like
	     * the code above, but we use 1 instead of cLongSubBlockHeight.
	     */
	pWordDisplacement = (Int32 *) &(pCodeBuf[subBlockJumpIndex-4]);
	*pWordDisplacement = bufIndex - subBlockJumpIndex;

	    /* This will also be the top of the lowest loop.  We jump
	     * back here after processing one line.
	     */
	topOfLowerLoopIndex = bufIndex;

	    /* initialize the single output register to be untouched */
	outInitArray[0] = cFalse;

	    /* go through the rows and columns of the input cache that
	     * apply to one output row.
	     */
	for (row = -(ny - 1 - cy); row < 1 + cy; row += 2)
	{
		/* Loop through all the dwords in this input cache
		 * row.  Generate the right sequence of loads and
		 * ANDs for each dword.
		 */
	    for (col = 0; col < (cacheWidthInBytes/8); col++)
	    {
		__error__ = ip_mmxCreateORCode (pIPState, nx, ny, cy, data, row,
					     (row+(ny-1-cy))*cacheWidthInBytes +
							(col * 8),
					      colIDArray[col],
					      cacheWidthInBytes,
					      1,
					      &(outInitArray[0]),
					      pCodeBuf,
					      &bufIndex,
					      bufSize);
		if (__error__ != ia_successful)
		    goto genCleanup;
	    }
	}

	    /* We've generated all the ORs for one subblock of height 1.
	     * Generate the code that writes the single dword reg to memory.
	     * There are two such code blocks, one for vertical strels
	     * (no byteswapping) and one for strels that need some shifting.
	     */
	if ((numActiveCols > 1) || (seColActive[cx] != cHit))
	{
		/* Strel needs some shifting */
	    M_COPY_CODE(mphT3WriteTinySubBlockStart);
	    M_COPY_CODE(mphT3WriteSubBlockIncCachePtr);
		/* We increment the rCache register by the difference
		 * between the starting position of this subblock and
		 * the starting position of the next subblock.
		 * That's 1 row in the input cache.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = 1 * cacheWidthInBytes;

	    M_COPY_CODE(mphT3WriteTinySubBlockRemainder);
		/* Set the negative displacement back to the top of the
		 * lowest loop.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = topOfLowerLoopIndex - bufIndex;
	}
	else
	{
		/* Strel is purely vertical */
	    M_COPY_CODE(mphT3WriteTinyVertSubBlockStart);
	    M_COPY_CODE(mphT3WriteSubBlockIncCachePtr);
		/* We increment the rCache register by the difference
		 * between the starting position of this subblock and
		 * the starting position of the next subblock.
		 * That's 1 row in the input cache.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = 1 * cacheWidthInBytes;

	    M_COPY_CODE(mphT3WriteTinyVertSubBlockRemainder);
		/* Set the negative displacement back to the top of the
		 * lowest loop.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = topOfLowerLoopIndex - bufIndex;
	}
    }	/* end of generating code to do ANDs in non-horizontal strel */

	/* We're almost done.  Generate the code to return. */
    M_COPY_CODE(mphT3Return);

	/* Now for a little error checking.  If we've generated more code
	   than fits in the buffer, complain */
    if (bufIndex >= bufSize)
    {
#if 0
	fprintf(stderr, " Buffer overflow, memory has been *smashed* in %s\n %d bytes written into buffer of size %d\n",
		i_genLongDilateAsm, bufIndex, bufSize);
#endif
	__error__ = ia_bufferExceeded;
	goto genCleanup;
    }

    *code_length = bufIndex;
    __error__ = ia_successful;

genCleanup:
    if (seColActive != NULL)
	FREE(seColActive);
    if (colIDArray != NULL)
	FREE(colIDArray);

    NORMAL_RETURN;
}


/* PRIVATE */
/************************************************************************
 *  ip_getLongDilateSEParms():
 *			return dimensions of input cache needed for long
 *			dilate ops on a strel along with an array of the
 *			active columns in the SE.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	pSE:		Pointer to structure element's data.
 *	pCacheWidth:	Pointer to returned width of cache in bytes.
 *			This may be NULL.
 *	pCacheHeight:	Pointer to returned height of input cache.
 *			This may be NULL.
 *	pNumActiveCols:	Pointer to returned number of active columns.
 *			This may be NULL.
 *	pActiveCols:	Pointer to returned array of nx bytes.  This
 *			array should be allocated (and freed) by the caller.
 *			This routine will fill each element in the array
 *			with cHit if the corresponding SE column contains
 *			at least one hit and 0 otherwise.  This may be NULL.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
ip_getLongDilateSEParms (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    const UInt8 *pSE,
    Int32      *pCacheWidth,
    Int32      *pCacheHeight,
    Int32      *pNumActiveCols,
    UInt8      *pActiveCols)
{
    PROLOGUE;
    Int32	 numActiveCols;
    Int32	 cacheWidth;
    Int32	 currentRow, currentCol;
    Bool	 allocatedHere;

	/* See if the parameters are ok. */
    if (pSE == NULL)
	return ia_invalidParm;

	/* If we weren't passed a pActiveCols, create one.
	 * We'll need it to calculate stuff.
	 */
    if (pActiveCols == NULL)
    {
	pActiveCols = (UInt8 *)CALLOC(nx, sizeof(UInt8));
	if (pActiveCols == NULL)
	    return ia_nomem;
	allocatedHere = cTrue;
    }
    else
    {
	    /* Clear all the entries in pActiveCols */
	for (currentCol = nx-1; currentCol >= 0; currentCol--)
	    pActiveCols[currentCol] = 0;
	allocatedHere = cFalse;
    }

	/* Go through the SE looking for active elements. */
    for (currentRow = ny; currentRow > 0; currentRow--)
    {
	for (currentCol = 0; currentCol < nx; currentCol++, pSE++)
	{
		/* If the current element is a HIT, add indicator
		 * to current column.  We're avoiding a superflous
		 * branch on the theory that long pipeline machines
		 * don't like them.
		 */
	    pActiveCols[currentCol] |= (*pSE) & cHit;
	}
    }

	/* Count the number of active columns in the SE.  Also reserve
	 * one spot for each dword in the input cache rows.
	 */
    for (numActiveCols = 0, cacheWidth = 0, currentCol = nx-1;
	 currentCol >= 0;
	 currentCol--)
    {
	if (pActiveCols[currentCol] != 0)
	{
	    numActiveCols++;
	    cacheWidth++;
	}
    }

	/* Return the calculated values.  Note that the height of the
	 * input cache is equal to the height of the strel plus the
	 * height of the blocks we'll be processing.  Since we always
	 * process two rows at a time, we round this up to the next
	 * multiple of 2.
	 */
    if (pCacheWidth != NULL)
	*pCacheWidth = cacheWidth * 8;		/* 8 bytes/dword */
    if (pCacheHeight != NULL)
	    /* should really be ((cMaxLongBlockHeight + ny-1)+1) & ~0x1 */
	*pCacheHeight = (cMaxLongBlockHeight + ny) & ~0x1;
    if (pNumActiveCols != NULL)
	*pNumActiveCols = numActiveCols;
    if (allocatedHere)
	FREE(pActiveCols);	/* otherwise, return it */

    NORMAL_RETURN;
}


/* PRIVATE */
/************************************************************************
 *  ip_mmxCreateDilateShiftCode():
 *			Create a block of MMX code that puts out the shifted
 *			64-bit dword(s) for one Structure Element column.
 *
 *	prevCol:	Number of the previous SE column for which shifted
 *			dwords were put out.  This is equal to -1 if curCol is
 *			the first column.
 *	curCol:		Number of the current SE column for which shifted
 *			dwords will be put out.
 *	nextCol;	Number of the next SE column for which shifted dwords
 *			will be put out.
 *	centerCol:	The number of the SE's center column.
 *	numRows:	The total number of rows in the strel.
 *	numCols:	The total number of columns in the strel.
 *	cacheBpl:	The bytes/line in the input cache.
 *	colIDArray	Pointer to array which maps horizontal displacements
 *			in the input cache to SE column numbers.  One or
 *			two entries will be added to this array here.
 *	pCurDisp:	Pointer to current displacement in colIDArray.
 *			The displacement will be updated upon
 *			returning from this routine.
 *	pTopOfLoopIndex: Pointer to the offset into pCodeBuf of the first
 *			byte of the shift loop.  The offset is set by
 *			this routine when the loop starts.  It is referenced
 *			by a later invocation of this routine when the
 *			code for the bottom of the loop is generated.
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			If no buffer overflow, *pBufIndex will be incremented
 *			to point to next available byte in code buffer.
 *	bufSize:	Max size of the code buffer.  This routine
 *			will check for buffer overflow before doing
 *			the write.
 * Return codes:
 *	ia_successful:		The operation worked.
 *	ia_bufferExceeded:	opcode didn't fit in buffer.
 ************************************************************************/
Int32  CDECL 
ip_mmxCreateDilateShiftCode (
    IPState    *pIPState,
    Int32	prevCol,
    Int32	curCol,
    Int32	nextCol,
    Int32	centerCol,
    Int32	numRows,
    Int32	numCols,
    Int32	cacheBpl,
    UInt8      *colIDArray,
    Int32      *pCurDisp,
    Int32      *pTopOfLoopIndex,
    UInt8      *pCodeBuf,
    Int32      *pBufIndex,
    Int32	bufSize)
{
    PROLOGUE;
    Int32	 bufIndex = *pBufIndex;	/* needed for macros */
    Int32	 curDisp = *pCurDisp;	/* current displacement in
					 * colIDArray. */
    Int32	*pWordDisplacement;	/* pts to displacements */
    

	/* Keep the compilers happy. */
    numCols = numCols;

	/* First, see if this is a vertical strel.
	 * We have to use this weird test since some weirdo
	 * might pass in a strel with lots of columns in which only
	 * the center column has hits.
	 */
    if ((curCol == centerCol) && (prevCol == -1) && (nextCol == -1))
    {	
	    /* VERTICAL STREL */
	    /* This is a vertical strel.  There is special code for
	     * this case.  It puts out the entire cache-filling loop
	     * in one step.
	     *
	     * The sequence is:
	     **START: read upper dword
	     **	; empty
	     **
	     **	read lower dword
	     **	; empty
	     **
	     **	add 2 * source bytes/line to source address
	     **	increment input cache address by
	     **			2 * number of bytes/row of input cache
	     **
	     **	write upper dword to input cache - 2 * bytes/row
	     **	; empty
	     **
	     **	write lower dword to input cache - bytes/row
	     **	; empty
	     **
	     **	decrement number of rows left by 2 in input cache
	     **	branch back to START if number of rows left > 0.
	     **
	     */
	*pTopOfLoopIndex = bufIndex;	/* remember where we start */
	M_COPY_CODE(mphT3LoadUpperSrcDWord);
	/* ; empty */

	M_COPY_CODE(mphT3LoadLowerSrcDWord);
	/* ; empty */

	M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
	M_COPY_CODE(mphT3AdjustCachePtr);
	    /* Set the existing 32-bit displacment to be 2 * cache Bpl. */
	pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
	*pWordDisplacement = 2 * cacheBpl;

	    /* Put out first write opcode */
	M_COPY_DISP(writeUpperOrig, 0, (-2 * cacheBpl));
	    /* ; empty */

	M_COPY_DISP(writeLowerOrig, 0, (-cacheBpl));
	    /* ; empty */

	    /* Remember what was written here */
	colIDArray[curDisp] = (UInt8) curCol;
	curDisp++;

	M_COPY_CODE(mphT3DecrRowsBy2AndBranch);
	    /* Set the jump's relative offset correctly */
	pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
	*pWordDisplacement = (*pTopOfLoopIndex - bufIndex);

    }	/* end of vertical strel case */
    else	/* this is not a vertical strel */
    {
	    /* There are several factors that determine what kind of code
	     * will be generated:
	     *  Column Position (left of center, in center col, right of center)
	     *  Is First Column?
	     *  Is Last Column?
	     *
	     * Since we dealt with purely vertical strels above, we
	     * know that no column we see here will be both the first
	     * and last column.
	     *
	     * If this is the first column, we need to generate some
	     * preamble code.  Then there's always some middle code to
	     * process the column.  If this is the last column, there's
	     * some code at the end of finish writing the last words
	     * and loop.  We'll deal with the three cases in this order.
	     *
	     * Note that the preamble and finish-up code are basically
	     * independent of the column position (left/center/right).
	     * The preamble only varies with whether the right or
	     * left sideword is loaded.  This makes things fairly
	     * straightforward.
	     *
	     * First, test to see if we need to insert preamble.
	     * The preamble code is:
	     *
	     **START: load upper dword
	     **	load 1's mask
	     **
	     **	load left/right upper side word
	     **	step 1 of upper wordswap
	     **
	     **	load lower dword
	     **	step 2 of upper wordswap
	     **
	     **	load left/right lower side word
	     **	step 1 of lower wordswap
	     **
	     **	shift upper sideword left/right
	     **	step 2 of lower wordswap
	     **
	     **	copy upper dword to upper scratch
	     **	shift lower sideword left/right
	     **
	     **	shift upper dword in scratch reg right/left
	     **	copy lower dword to lower scratch
	     **
	     */
	if (prevCol == -1)	/* preamble needed */
	{
	    *pTopOfLoopIndex = bufIndex;	/* remember where we start */
	    M_COPY_CODE(mphT3LoadUpperSrcDWord);
	    M_COPY_CODE(mphT3PreLoad1sMask);

		/* If the first off-center column is to the left of center, load
		 * the upper right sideword.  Otherwise, load the upper left
		 * sideword.
		 */
	    if ((curCol < centerCol) ||
		((curCol == centerCol) && (nextCol < centerCol)))
	    {
		M_COPY_CODE(mphT3PreLoadUpperRightSideWord);
	    }
	    else
	    {
		M_COPY_CODE(mphT3PreLoadUpperLeftSideWord);
	    }
	    M_COPY_CODE(mphT3WordSwapUpperStep1);

	    M_COPY_CODE(mphT3LoadLowerSrcDWord);
	    M_COPY_CODE(mphT3WordSwapUpperStep2);


		/* Same deal for loading lower sideword */
	    if ((curCol < centerCol) ||
		((curCol == centerCol) && (nextCol < centerCol)))
	    {
		M_COPY_CODE(mphT3PreLoadLowerRightSideWord);
	    }
	    else
	    {
		M_COPY_CODE(mphT3PreLoadLowerLeftSideWord);
	    }
	    M_COPY_CODE(mphT3WordSwapLowerStep1);

		/* The amount and direction we shift the sideword depends on
		 * whether we're to the left or right of center.
		 * If we're to the left of center, we want to shift
		 * the right sideword right by 32 - (centerCol - curCol).
		 * That's because we've just loaded the sideword
		 * into the lower 32 bits of a reg and we want
		 * to start with (centerCol - curCol) bits
		 * at the bottom (right) end of the sideword.
		 * If we're in the center column or on the right side,
		 * then we just read 32 bits from the word on the
		 * left into the lower 32 bits of a reg.  We want
		 * (curCol - centerCol) bits to be left in the top
		 * of that register, so we shift it left by
		 * 32 + 32 - (curCol - centerCol).
		 */
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3RightShiftUpperSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (centerCol - curCol));
	    }
	    else	/* see if we start with left or right shift
			 * for nextCol */
	    if ((curCol == centerCol) && (nextCol < centerCol))
	    {
		M_COPY_CODE(mphT3RightShiftUpperSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (centerCol - nextCol));
	    }
	    else	/* start with left shift */
	    if ((curCol == centerCol) && (nextCol > centerCol))
	    {
		M_COPY_CODE(mphT3LeftShiftUpperSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 + 32 -
						  (nextCol - centerCol));
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3LeftShiftUpperSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 + 32 - 
						  (curCol - centerCol));
	    }
	    M_COPY_CODE(mphT3WordSwapLowerStep2);


	    M_COPY_CODE(mphT3CopyUpperDWord);
		/* Insert the left/right shift for the lower word */
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3RightShiftLowerSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (centerCol - curCol));
	    }
	    else	/* see if we start with left or right shift
			 * for nextCol */
	    if ((curCol == centerCol) && (nextCol < centerCol))
	    {
		M_COPY_CODE(mphT3RightShiftLowerSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (centerCol - nextCol));
	    }
	    else	/* start with right shift */
	    if ((curCol == centerCol) && (nextCol > centerCol))
	    {
		M_COPY_CODE(mphT3LeftShiftLowerSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 + 32 -
						  (nextCol - centerCol));
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3LeftShiftLowerSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 + 32 - 
						  (curCol - centerCol));
	    }


		/* Now we get ready to shift the dwords.  If we
		 * are to the left of center, we want to shift the
		 * dwords left so that the "unshifted" input (which
		 * is where the center is in dilation) bits will
		 * move to the output word at the position of
		 * the active SE element.  If we're to the right
		 * of center (or will be soon), we want to shift
		 * the dwords right.  The trick is to remember that
		 * the center is where the input is and the active
		 * element is where the output is.  This is the
		 * opposite of erosion.
		 */
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3LeftShiftUpperScratch);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
	    }
	    else	/* see if we start with left or right shift
			 * for nextCol */
	    if ((curCol == centerCol) && (nextCol < centerCol))
	    {
		M_COPY_CODE(mphT3LeftShiftUpperScratch);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - nextCol);
	    }
	    else	/* start with right shift */
	    if ((curCol == centerCol) && (nextCol > centerCol))
	    {
		M_COPY_CODE(mphT3RightShiftUpperScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3RightShiftUpperScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
	    }
	    M_COPY_CODE(mphT3CopyLowerDWord);

	}	/* end of case where preamble is needed */

	    /* Now we decide which kind of column block to generate.  Here's
	     * the basic search order.  Note that purely horizontal strels
	     * can't have both hits and misses in the same column since
	     * there's only one entry per column (only 1 row in the strel).
	     ** Preload (loads registers and starts the shifts)
	     ** See if col is center col, if so generate code for it.
	     **	   non-horizontal strels
	     **	   pure horizontal strels
	     ** See if col is last column to be processed (will be to left or
	     ** right of center).  If so, generate code for it.
	     **	   non-horizontal strels
	     **	   pure horizontal strels
	     ** See if EITHER the col is to the left of center and is
	     ** followed by another column that's also to the left of
	     ** center OR the colis to the right of center and is
	     **  followed by another col that's also to the right of
	     ** center.  If so, generate code for it.  We think this
	     ** is the common case.
	     **	   non-horizontal strels
	     **	   pure horizontal strels
	     ** If none of the cases above is true, this must be the last left
	     ** col to be processed.  It must be followed by a right col.
	     ** Generate code for this case.
	     **	   non-horizontal strels
	     **	   pure horizontal strels
	     */
	if (curCol == centerCol)
	{
		/* CENTER COLUMN */
	    if (numRows > 1)	/* not a purely horizontal strel */
	    {
		    /*
		     **LOOP:
		     **	write unshifted upper dword to input cache
		     **	; empty
		     **
		     **	write unshifted lower dword to
		     **			input cache + bytes/row
		     **	; empty
		     **
		     */
		    /* Put out code to write inverted inputs */
		M_COPY_DISP(writeUpperOrig, 0, 0);
		    /* ; empty */
    
		M_COPY_DISP(writeLowerOrig, 0, +cacheBpl);
		    /* ; empty */

		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		curDisp++;
	    }
	    else	/* is purely horizontal strel */
	    {
		    /*
		     **LOOP:
		     **	copy unshifted upper dword to mm6 temp
		     **	copy unshifted lower dword to mm7 temp
		     **
		     */
		M_COPY_CODE(mphT3LoadUpperOrigDWord);
		M_COPY_CODE(mphT3LoadLowerOrigDWord);
	    }
	}	/* end of center column case */

	else
	    /* see if this is the last column we'll process in the row */
	if (nextCol == -1)
	{
		/* LAST COLUMN */
		/*
		 **LOOP:
		 **	combine (OR) upper dword and sideword
		 **	shift lower dword in scratch reg right/left
		 **
		 **	increment inputCache address by
		 **			2 * number of input cache bytes/row 
		 **	combine (OR) lower dword and sideword
		 **
		 ** ; wait for AGI
		 **	add 2 * source bytes/line to source address
		 **	; empty
		 **
		 **	write last shifted upper dword to
		 **			inputCache - 2 * bytes/row. 
		 **	; empty
		 **
		 **	write last shifted lower dword to
		 **			inputCache - bytes/row
		 **	; empty
		 **
		 **	decrement number of rows left by 2 in inputCache
		 **	branch back to START if number of rows left > 0.
		 **
		 ** If this is a purely horizontal strel, the code becomes:
		 **LOOP:
		 **	combine (OR) upper dword and sideword
		 **	shift lower dword in scratch reg right/left
		 **
		 **	AND upper shifted dword with mm6 temp, result to mm2
		 **	combine (OR) lower dword and sideword
		 **
		 **	wordswap step 1 for mm2, mm6 as scratch
		 **	AND lower shifted dword with mm7 temp, result to mm5
		 **
		 **	add 2 * source bytes/line to source address
		 **	wordswap step 2 for mm2, mm6 as scratch
		 **
		 **	wordswap step 1 for mm5, mm7 as scratch
		 **	decrement number of rows left by 2
		 **
		 **	write mm2 to dstAddress		; note 1 cycle delay
		 **	wordswap step 2 for mm5, mm7 as scratch
		 **
		 **	; empty
		 **	; empty
		 **
		 **	write mm5 to dstAddress + dest bytes/row
		 **	; empty
		 **
		 **	add 2 * dest bytes/line to dest address using
		 **					lea (no cc's affected)
		 **	branch back to START if number of rows left > 0.
		 **
		 **
		 */
	    M_COPY_CODE(mphT3CombineShiftedUpperDWord);
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3LeftShiftLowerScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3RightShiftLowerScratch);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
	    }

	    if (numRows > 1 )
	    {
		M_COPY_CODE(mphT3AdjustCachePtr);
		    /* Set the existing 32-bit displacment
		     * to be 2*cacheBpl. */
		pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		*pWordDisplacement = 2 * cacheBpl;
	    }
	    else	/* purely horizontal strel */
	    {
		M_COPY_CODE(mphT3OrUpperShiftedDWordToMM2);
	    }
	    M_COPY_CODE(mphT3CombineShiftedLowerDWord);


	    if (numRows > 1 )
	    {
		M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
		/* ; empty */

		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp - 2*cacheBpl);
		/* ; empty */

		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp - cacheBpl);
		/* ; empty */
		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		curDisp++;

		M_COPY_CODE(mphT3DecrRowsBy2AndBranch);
		    /* Set the jump's relative offset correctly */
		pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		*pWordDisplacement = (*pTopOfLoopIndex - bufIndex);
	    }
	    else	/* purely horizontal strel */
	    {
		M_COPY_CODE(mphT3WordSwapUpperMM2Step1);
		M_COPY_CODE(mphT3OrLowerShiftedDWordToMM5);

		M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
		M_COPY_CODE(mphT3WordSwapUpperMM2Step2);

		M_COPY_CODE(mphT3WordSwapLowerMM5Step1);
		M_COPY_CODE(mphT3DecrRowsBy2);

		M_COPY_CODE(mphT3WriteUpperScratchToDest);
		M_COPY_CODE(mphT3WordSwapLowerMM5Step2);

		/* ; empty */
		/* ; empty */

		M_COPY_CODE(mphT3WriteLowerScratchToDest);
		/* ; empty */

		M_COPY_CODE(mphT3AdjustDstAddrBy2Lines);
		M_COPY_CODE(mphT3JABranch);
		    /* Set the jump's relative offset correctly */
		pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		*pWordDisplacement = (*pTopOfLoopIndex - bufIndex);
	    }
	}	/* end of last column case */

	else
	    /* see if this is a left col followed by a left col or a
	     * right col followed by a right col.
	     */
	if ( ((curCol < centerCol) && (nextCol < centerCol)) ||
	     ((curCol > centerCol) && (nextCol > centerCol)) )
	{

		/*  CONTINUATION COLUMN 
		 *
		 **LOOP:
		 **	combine (OR) upper dword and sideword
		 **	shift lower dword in scratch reg right/left
		 **
		 **	shift upper sideword left/right
		 **	combine (OR) lower dword and sideword
		 **
		 ** ; note 1 cycle delay from OR
		 **	write upper dword to input cache
		 **	shift lower sideword left/right
		 **
		 **	write lower dword to input cache + bytes/row
		 **	copy upper dword to upper scratch
		 **
		 **	shift upper dword in scratch reg right/left
		 **	copy lower dword to lower scratch
		 **
		 ** If this is a purely horizontal strel, the code becomes:
		 **LOOP:
		 **	combine (OR) upper dword and sideword
		 **	shift lower dword in scratch reg right/left
		 **
		 **	shift upper sideword left/right
		 **	combine (OR) lower dword and sideword
		 **
		 **	pand upper dword to mm6
		 **	shift lower sideword left/right
		 **
		 **	pand lower dword to mm7
		 **	copy upper dword to upper scratch
		 **
		 **	shift upper dword in scratch reg right/left
		 **	copy lower dword to lower scratch
		 **
		 */

	    M_COPY_CODE(mphT3CombineShiftedUpperDWord);
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3LeftShiftLowerScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3RightShiftLowerScratch);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
	    }

		/* Note that this shift is for the next time
		 * through the loop, so we shift by the difference
		 * between nextCol and curCol.
		 */
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3RightShiftUpperSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - curCol);
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3LeftShiftUpperSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (curCol - nextCol);
	    }
	    M_COPY_CODE(mphT3CombineShiftedLowerDWord);

	    if (numRows > 1)
	    {
		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
	    }
	    else
	    {
		if (prevCol != -1)
		{
		    M_COPY_CODE(mphT3OrUpperShiftedDWord);
		}
		else
		{
		    M_COPY_CODE(mphT3LoadUpperShiftedDWord);
		}
	    }
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3RightShiftLowerSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - curCol);
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3LeftShiftLowerSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (curCol - nextCol);
	    }

	    if (numRows > 1)
	    {
		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		curDisp++;
	    }
	    else	/* purely horizontal strel */
	    {
		if (prevCol != -1)
		{
		    M_COPY_CODE(mphT3OrLowerShiftedDWord);
		}
		else
		{
		    M_COPY_CODE(mphT3LoadLowerShiftedDWord);
		}
	    }
	    M_COPY_CODE(mphT3CopyUpperDWord);


	    if (nextCol < centerCol)
	    {
		M_COPY_CODE(mphT3LeftShiftUpperScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - nextCol);
	    }
	    else	/* nextCol > centerCol */
	    {
		M_COPY_CODE(mphT3RightShiftUpperScratch);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);
	    }
	    M_COPY_CODE(mphT3CopyLowerDWord);

	}	/* End of continuation column case */

	else	/* Well, it's not the center column, it's not the last
		 * column, it's not a left followed by left or right
		 * followed by right; so it must be a left followed by
		 * right.
		 */
	{
		/* LAST LEFT COLUMN */
		/*
		 **LOOP:
		 **	combine (OR) upper dword and sideword
		 **	shift lower dword in scratch reg right/left
		 **
		 **	read left upper sideword (32 bits).
		 **	combine (OR) lower dword and sideword
		 **
		 **	write last shifted upper dword to input cache
		 **	copy upper dword to upper scratch
		 **
		 **	write last shifted lower dword to
		 **				input cache + bytes/row
		 **	shift upper dword in scratch reg right
		 **
		 **	load left lower sideword (32 bits).
		 **	shift upper sideword left
		 **
		 **	copy lower dword to lower scratch
		 **	shift lower sideword left
		 **
		 **
		 ** If this is a purely horizontal strel, the code becomes:
		 **LOOP:
		 **	combine (OR) upper dword and sideword
		 **	shift lower dword in scratch reg right/left
		 **
		 **	read left upper sideword (32 bits).
		 **	combine (OR) lower dword and sideword
		 **
		 **	pand upper dword, result to mm6
		 **	copy upper dword to upper scratch
		 **
		 **	pand lower dword, result to mm7
		 **	shift upper dword in scratch reg right
		 **
		 **	load left lower sideword (32 bits).
		 **	shift upper sideword left
		 **
		 **	copy lower dword to lower scratch
		 **	shift lower sideword left
		 **
		 */
	    M_COPY_CODE(mphT3CombineShiftedUpperDWord);
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3LeftShiftLowerScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3RightShiftLowerScratch);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
	    }

	    M_COPY_CODE(mphT3PreLoadUpperLeftSideWord);
	    M_COPY_CODE(mphT3CombineShiftedLowerDWord);

	    if (numRows > 1)
	    {
		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
	    }
	    else	/* purely horizontal strel */
	    {
		if (prevCol != -1)
		{
		    M_COPY_CODE(mphT3OrUpperShiftedDWord);
		}
		else
		{
		    M_COPY_CODE(mphT3LoadUpperShiftedDWord);
		}
	    }
	    M_COPY_CODE(mphT3CopyUpperDWord);

	    if (numRows > 1)
	    {
		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		curDisp++;
	    }
	    else
	    {
		if (prevCol != -1)
		{
		    M_COPY_CODE(mphT3OrLowerShiftedDWord);
		}
		else
		{
		    M_COPY_CODE(mphT3LoadLowerShiftedDWord);
		}
	    }
	    M_COPY_CODE(mphT3RightShiftUpperScratch);
		/* Set the right shift distance */
	    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);

	    M_COPY_CODE(mphT3PreLoadLowerLeftSideWord);
	    M_COPY_CODE(mphT3LeftShiftUpperSideWord);
		/* Set the left shift distance */
	    pCodeBuf[bufIndex - 1] =
				(UInt8) (32 + (32 - (nextCol - centerCol)));

	    M_COPY_CODE(mphT3CopyLowerDWord);
	    M_COPY_CODE(mphT3LeftShiftLowerSideWord);
		/* Set the left shift distance */
	    pCodeBuf[bufIndex - 1] =
				(UInt8) (32 + (32 - (nextCol - centerCol)));

	}	/* end of last left case */
    }	/* end of non-vertical strel case */


genCleanup:
    if (__error__ == ia_successful)
    {
	*pCurDisp = curDisp;
	*pBufIndex = bufIndex;
    }
    NORMAL_RETURN;
}


/* PRIVATE */
/************************************************************************
 *  ip_mmxCreateORCode():
 *			Create a block of MMX code that combines two shifted
 *			dwords from the input cache with the numOutputRows
 *			output registers.
 *			The two shifted dwords are vertically adjacent in the
 *			input cache.  Therefore, they share the same
 *			SE column and polarity.
 *			
 *			When this routine is called, the code used to
 *			update the input cache pointer so that rCache plus
 *			topInputDisp points to the upper cached dword.
 *			This routine does not generate code to update
 *			rCache.
 *
 *	nx:		Number of columns in Structure Element (SE)
 *	ny:		Number of rows in Structure Element.
 *	cy:		Number of the SE's "center" row.
 *	pSE:		Pointer to byte array containing Structure Elements.
 *	topInputRow:	Each cache dword is formed from a dword (and usually
 *			a sideword) from one row of the input image. This
 *			parameter is equal to that input image row number.
 *			It is relative to the number of the top output row.
 *	topInputDisp:	The displacement from the "current" (in the sense
 *			of "the value that's the result of the code
 *			that's been generated so far") value of rCache to
 *			the upper cached input word.
 *	inputSECol:	Number of SE column for both input cache dwords.
 *	cacheBpl:	Input cache bytes/line.
 *	numOutputRows:	The number of output rows.
 *	pOutInitArray:	Pointer to an array of Booleans that records whether
 *			each of the output words has been initialized.  Used
 *			to convert the first operation touching an output
 *			word from an OR to a load.  This routine updates
 *			this array as each output word is initialized.  The
 *			calling routine should set all components of this array,
 *			there are nunOutputRows of them, to cFalse before
 *			calling this routine for the first time for a given
 *			output subblock.
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			If no buffer overflow, *pBufIndex will be incremented
 *			to point to next available byte in code buffer.
 *	bufSize:	Max size of the code buffer.  This routine
 *			will check for buffer overflow before doing
 *			the write.
 * Return codes:
 *	ia_successful:		The operation worked.
 *	ia_bufferExceeded:	opcode didn't fit in buffer.
 ************************************************************************/
Int32  CDECL 
ip_mmxCreateORCode (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cy,
    const UInt8 *pSE,
    Int32	topInputRow,
    Int32	topInputDisp,
    Int32	inputSECol,
    Int32	cacheBpl,
    Int32	numOutputRows,
    Bool       *pOutInitArray,
    UInt8      *pCodeBuf,
    Int32      *pBufIndex,
    Int32	bufSize)
{
    PROLOGUE;
    Int32	 bufIndex = *pBufIndex;	/* needed for macros */
    Int32	 upperUses;		/* number of times upper dword
					 * is used */
    Bool	 upperLoaded;		/* if cTrue, upper input dword
					 * has been loaded into the scratch
					 * reg. */
    Int32	 lowerUses;		/* number of times lower dword
					 * is used */
    Bool	 lowerLoaded;		/* if cTrue, lower input dword
					 * has been loaded into the scratch
					 * reg. */
    Int32	 outRow;		/* temp number of output row */
    Int32	 seIndex, seEnd;	/* indices for pSE. */
    

	/* Initialize local vars */
    seEnd = nx * ny;

	/* First, we count the number of times the upper and lower
	 * dwords are used.  Whichever one is used the most will
	 * be loaded into a register before being used.
	 *
	 * The counting process takes a little thought.  Since
	 * we'll use basically the same method to generate the
	 * OR code, we'd better explain it.
	 *
	 * Since we're doing dilation, we want to place each strel
	 * HIT on the output dword, then see what we have to do to
	 * move the dword at the resulting center's location to
	 * the output dword.  This is the opposite of erosion where
	 * we fix the center of the strel on the output and inspect
	 * all of the dwords at the active element locations.
	 *
	 * Therefore, we imagine putting each strel HIT on top of
	 * each output dword.  To see if a particular input dword
	 * (one of the two we're given) is used, we see if the
	 * strel center falls on top that input dword.
	 * If so, we need to generate an OR opcode.
	 * We perform this process for the upper input dword, incrementing
	 * upperUses if we get a match.  We repeat it for the lower
	 * input dword, incrementing lowerUses if that gives a match.
	 * Note that since the the offsets from the output dword to the
	 * two cached input dwords are different, we need to look at
	 * two different SE elements.
	 *
	 * That's the obvious way to do it.  However, there's a little
	 * problem organizing things cleanly.  Another way to do the
	 * same thing is to start with the input cache dwords and
	 * iterate through the rows of the SE.  We put the SE's center
	 * on top of each input dword, then for each SE row, we
	 * see if there's a HIT on the location corresponding to an
	 * output dword.  If the HIT ends up on top of one of the
	 * output dwords, we increment the corresponding use counter
	 * (upperUses if we're looking at the upper input dword,
	 * lowerUses otherwise).  This method has a couple of nice
	 * features.  First, it's a little easier to iterate through
	 * the rows of the SE, reading each one once, and checking
	 * output row numbers than to interate through output row
	 * numbers and check for array bounds problems on the SE,
	 * then read the proper SE element.  That's a minor thing.
	 * A slightly nicer point is that we will want to generate
	 * ORs in pairs based on two input dwords being combined in
	 * parallel with two output dwords based on the same SE
	 * element.  Iterating through the SE elements one at
	 * a time seems to make this flow more nicely.
	 * There's also the problem of what we do at the top of and
	 * bottom of the subblock.  We only generate one OR of the
	 * pair there.  Iterating through the SE elements makes this
	 * special case a bit more natural.
	 *
	 * Ok, so we iterate through the SE rows from top to bottom.
	 * This means we initially place the SE center on the input
	 * dword.  Each time we move down one row in the
	 * SE, we move the output row down one.  This is a consequence of
	 * keeping the input location fixed.  Since we're keeping the
	 * position of the SE fixed, moving down one row in the SE
	 * means moving down one row in the output.
	 */
    upperUses = lowerUses = 0;
    for (seIndex = inputSECol, outRow = topInputRow - cy;
	 seIndex < seEnd;
	 seIndex += nx, outRow++)
    {
	if (pSE[seIndex] == cHit)
	{
	    if ((outRow >= 0) && (outRow < numOutputRows))
		upperUses++;		/* upper input is used again */
	    if ((outRow + 1 >= 0) && (outRow + 1 < numOutputRows))
		lowerUses++;		/* lower input is used again */
	}
    }

	/* If the strel is pretty tall and/or there aren't many
	 * output rows, we might not use the inputs at all.
	 * If this is the case, we're done.
	 */
    if ((upperUses + lowerUses) == 0)
	NORMAL_RETURN;

	/* See which we should load into a register.  We load the
	 * one with more uses.  If there's a tie, and both counts
	 * are greater than 1, we load the lower word.
	 */
    upperLoaded = lowerLoaded = cFalse;
    if ((upperUses + lowerUses) > 2)
    {
	if (upperUses > lowerUses)
	{
	    M_COPY_DISP(loadFromCache, cScratchReg, topInputDisp);
	    upperLoaded = cTrue;
	}
	else
	{
	    M_COPY_DISP(loadFromCache, cScratchReg, topInputDisp + cacheBpl);
	    lowerLoaded = cTrue;
	}
    }

	/* Now go through the strel rows again, generating OR ops as
	 * we find proper places where the input dwords in question
	 * should be combined with the current output dwords.  Note
	 * that if we want to combine both the upper and lower dwords
	 * with output dwords, one of the combinations will be done
	 * using a OR-with-memory and the other will be OR-with-
	 * register.  The pairing rules dictate that the OR-with-memory
	 * be first.  This means we must look to see which input
	 * has been loaded before generating the opcodes.
	 *
	 * Also, if one of the output registers hasn't been touched
	 * yet, we must generate a load (from memory or a reg) instead
	 * of an OR opcode and must mark the corresponding Boolean
	 * in the initialization array.
	 *
	 * Normally, we operate on the upper word, then on the lower
	 * word.  The only case where we reverse the order is when
	 * a) we will operate on both, and b) the upper input word has
	 * been loaded into the scratch register.  In this case, we
	 * do the memory operation with the lower word in the first
	 * opcode and the scratch op with the upper dword in the second.
	 * We do this because memory ops must be in the first opcode of
	 * a pair.
	 */
    for (seIndex = inputSECol, outRow = topInputRow - cy;
	 seIndex < seEnd;
	 seIndex += nx, outRow++)
    {
	if (pSE[seIndex] == cHit)	/* there's something to do */
	{
	    if ((upperLoaded) &&
		(outRow >= 0) && (outRow < numOutputRows) &&
		(outRow + 1 >= 0) && (outRow + 1 < numOutputRows))
	    {
		/* This is the weird case in which we process the
		 * lower input word first.
		 */
		    /* lower must not be loaded, so operate from
		     * memory.  Do we want a load or an OR?
		     */
		if (pOutInitArray[outRow+1])
		{
		    M_COPY_DISP(orFromCache, outRow + 1,
				topInputDisp + cacheBpl);
		}
		else
		{
		    M_COPY_DISP(loadFromCache, outRow + 1,
				topInputDisp + cacheBpl);
		    pOutInitArray[outRow+1] = cTrue;
		}

		    /* In second op, operate on upper word */
		if (pOutInitArray[outRow])
		{
			/* put out an OR opcode with the loaded
			 * scratch reg.  We don't use M_COPY_DISP
			 * because there's no displacement.
			 */
		    __error__ =
			ip_copyCode (pIPState, orWithScratch.disp8[outRow],
				     *(orWithScratch.disp8Len[outRow]),
				     pCodeBuf, &bufIndex, bufSize);
		    if (__error__ != ia_successful)
			goto genCleanup;
		}
		else	/* this is the first time we've touched
			     * this output register */
		{
			/* put out a LOAD opcode from the loaded
			 * scratch reg.  We don't use M_COPY_DISP
			 * because there's no displacement.
			 */
		    __error__ =
			ip_copyCode (pIPState, loadFromScratch.disp8[outRow],
				     *(loadFromScratch.disp8Len[outRow]),
				     pCodeBuf, &bufIndex, bufSize);
		    if (__error__ != ia_successful)
			goto genCleanup;
		    pOutInitArray[outRow] = cTrue;
		}
	    }
	    else	/* process in the normal order.  First,
			 * look at the upper word.
			 */
	    {
		if ((outRow >= 0) && (outRow < numOutputRows))
		{
			/* We need to process the upper word */
		    if (upperLoaded)
		    {
			    /* upper input dword is in scratch register */
			if (pOutInitArray[outRow])
			{
				/* put out an OR opcode with the loaded
				 * scratch reg.  We don't use M_COPY_DISP
				 * because there's no displacement.
				 */
			    __error__ =
				ip_copyCode (pIPState,
					     orWithScratch.disp8[outRow],
					     *(orWithScratch.disp8Len[outRow]),
					     pCodeBuf, &bufIndex, bufSize);
			    if (__error__ != ia_successful)
				goto genCleanup;
			}
			else	/* this is the first time we've touched
				     * this output register */
			{
				/* put out a LOAD opcode from the loaded
				 * scratch reg.  We don't use M_COPY_DISP
				 * because there's no displacement.
				 */
			    __error__ =
			      ip_copyCode (pIPState,
					   loadFromScratch.disp8[outRow],
					   *(loadFromScratch.disp8Len[outRow]),
					   pCodeBuf, &bufIndex, bufSize);
			    if (__error__ != ia_successful)
				goto genCleanup;
			    pOutInitArray[outRow] = cTrue;
			}
		    }
		    else	/* upper input dword is in memory */
		    {
			if (pOutInitArray[outRow])
			{
			    M_COPY_DISP(orFromCache, outRow,
					topInputDisp);
			}
			else
			{
			    M_COPY_DISP(loadFromCache, outRow,
					topInputDisp);
			    pOutInitArray[outRow] = cTrue;
			}
		    }
		}	/* end of processing the upper input dword */

		    /* See if there's anything to do for the lower dword. */
		if ((outRow + 1 >= 0) && (outRow + 1 < numOutputRows))
		{
			/* Now process the lower word.  See if it's
			 * already in the scratch register.
			 */
		    if (lowerLoaded)
		    {
			    /* lower dword is in scratch reg */
			if (pOutInitArray[outRow + 1])
			{
				/* put out an OR opcode with the loaded
				 * scratch reg.  We don't use M_COPY_DISP
				 * because there's no displacement.
				 */
			    __error__ =
			       ip_copyCode (pIPState,
					    orWithScratch.disp8[outRow+1],
					    *(orWithScratch.disp8Len[outRow+1]),
					    pCodeBuf, &bufIndex, bufSize);
			    if (__error__ != ia_successful)
				goto genCleanup;
			}
			else	/* this is the first time we've touched
				     * this output register */
			{
				/* put out a LOAD opcode from the loaded
				 * scratch reg.  We don't use M_COPY_DISP
				 * because there's no displacement.
				 */
			    __error__ =
			     ip_copyCode (pIPState,
					  loadFromScratch.disp8[outRow+1],
					  *(loadFromScratch.disp8Len[outRow+1]),
					  pCodeBuf, &bufIndex, bufSize);
			    if (__error__ != ia_successful)
				goto genCleanup;
			    pOutInitArray[outRow+1] = cTrue;
			}
		    }
		    else	/* lower dword is in memory */
		    {
			if (pOutInitArray[outRow+1])
			{
			    M_COPY_DISP(orFromCache, outRow + 1,
					topInputDisp + cacheBpl);
			}
			else
			{
			    M_COPY_DISP(loadFromCache, outRow + 1,
					topInputDisp + cacheBpl);
			    pOutInitArray[outRow+1] = cTrue;
			}
		    }
		}	/* end of processing the lower input dword */
	    }	/* end of case where upper dword is processed before
		 * lower dword.
		 */
	}	/* end of case where SE polarity is a HIT */
    }	/* end of loop that goes through SE rows, dropping out code
	 * to process each SE element in a column.
	 */

genCleanup:
    if (__error__ == ia_successful)
    {
	*pBufIndex = bufIndex;
    }
    NORMAL_RETURN;
}


/*

$Log: pcgendim.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.4  1998/01/13 23:54:00  ddavies
pcgendim.c -> /product/ipcore/ipshared/src/RCS/pcgendim.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1997/08/22 01:29:16  ddavies
ddavies on Thu Aug 21 18:28:09 PDT 1997

pcgendim.c -> /product/ipcore/ipshared/src/RCS/pcgendim.c,v

Make the MMX code generators generate ia_SETooBig errors for 1x1 strels.
The strels weren't going to be used anyway, but the bogus strels confused
Richard.

Revision 1.2  1996/12/19 23:05:01  ddavies
pcgendim.c -> /product/ipcore/ipshared/src/RCS/pcgendim.c,v

Have i_getLongDilateAsmSize both use the cy parameter and check to see if
its on an MMX machine.  If not, return ia_noImpl.  Also have
ip_mmxCreateDilateShiftCode use the numCols parameter.

Revision 1.1  1996/12/18 22:53:41  ddavies
Initial revision


 */



