/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  pcgenero.c -- generates assembly routines implementing erode ops
 */


#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#ifdef __WATCOMC__
#include <string.h>
#endif
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "frames.pub"
#include "memory.pub"

    /* prototypes */
#include "shrmorph.pub"
#include "shrmorph.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: pcgenero.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* This code has the formidible job of generating (nearly) optimal 386
 * assembly code for a wide range of structure elements to be used for
 * erode (as opposed to dilates).  It's not so bad as it sounds, however.
 * Code for all the SE's corresponds to a template.  We use the file
 * mortemp2.asm (morphological template) as a souce of assembly code.
 *
 * There are extensive comments in mortemp2.asm.  They explain the algorithm
 * used to do the actual processing.  Seems a waste to duplicate those
 * comments here, but we'll do it anyway.
 *
 *******************************************************************
 *
 * Note that we use no local or global data references, all variables are
 * referenced as offsets from the EBP register.  The C code that generates
 * real assembly need not worry about relocable variables.

 * The algorithm used here is quite different from the one used in the
 * SPARC assembly code.  Instead of maintaining as many partial
 * results as possible, we compute each complete result and store it.
 * In the first version of this code, we shifted each source word once
 * for each SE element.  That's a lot of relatively slow shift operations.
 * This is especially true of the double shifts we need.  We thought
 * we had to recompute every input value because the 386 doesn't have
 * enough registers to hold either lots of inputs or lots of outputs.

 * The SPARC algorithms broke the entire image into blocks, processing
 * each block vertically.  This was a good compromise between efficient
 * processing (processing the image vertically minimizes memory writes of
 * partial results) and cache utilization (minimized by horizontal image
 * reads).  In the first version of 386 algorithm, all inputs read for
 * one dest word were read again as necessary for the next dest word.  It
 * didn't matter whether we went horizontally or vertically, so we went
 * horizontally because it was easier.

 * The second version of 386 morphology saves the shifted versions of each
 * source word in a buffer.  While we still calculate one destination
 * word at a time, we can usually use preshifted input words, cutting out the
 * 4 cycles used to do double shifts out of the inner loop.

 * Given that we want to use pre-shifted inputs, do we move through the
 * image horizontally or vertically?  If we moved horizontally (and computed
 * one output word at a time), each horizontal move would require us to read
 * a new input word for *each row* of the SE.  That's because each word is
 * only 1 bit tall.

 * If we move through the image vertically, we need to read only 3 new words
 * for each move down.  That's because the words are wide enough to cover the
 * maximum legal shift done in our accelerated code (32 bits).  So, we need
 * to read the word directly below the previous ones and the words on either
 * side of that.  This seems preferable to the extra reads horizontal
 * movement would require.

 * How do we address the buffer?  It will be at least as large as the
 * SE, except each SE point will correspond to a (shifted) 32-bit input word.
 * We'll process the image in vertical stripes one word wide.  That means
 * we'll make a vertical pass through the rows of the array of shifted inputs
 * for each output word.  We could move each shifted input word up one row
 * in the array when it is used.  That generates a lot of writes.  It would
 * be nicer to leave the words in place and use a moving pointer.

 * We can't use one pre-allocated global array for the shifted inputs
 * needed in all erosion operations because the code wouldn't be reentrant.
 * Therefore, we must use a register to point to an array allocated by the
 * calling procedure.  We know the width of the array, so can use constant
 * displacements to address shifted inputs within a row.  It would be very
 * nice to use just register+displacment addressing, it adds no cycles.
 * However, we need some way to detect when the buffer wraps around.

 * Or do we?  We'll process a limited number of words vertically before
 * starting the next stripe.  This is done to make reasonable use of the
 * processor's cache.  Say the calling routine allocates a buffer that's
 * so large that we'll never overflow it.  If we process n lines vertically and 
 * have an H by V structure element, we'll need a buffer that can hold
 * H*(V+n)*4 bytes.  Typical values of H and V are under 10.  Typical
 * processing block height is 64.  That's a 10*74*4 = 3KB block.  Not too bad,
 * especially when compared to the 1MB required for the image.  Say
 * the SE were a maximum 64 x 64.  That requires a 64*128*4 = 32KB block.
 * Seems ok, especially because it saves a compare and branch on every row
 * being processed.  This is a ridiculously large strel.

 * In fact, since the rows don't wrap around, we can address
 * the entire array of shifted inputs with displacements relative to a
 * single pointer.  The pointer points to the word in the upper left corner
 * of the array.  Zowee.  We can do AND or OR ops in 2 cycles.  ANDN ops
 * take 3 cycles since we have to invert the input before the AND.  Very nice.
 * There's a somewhat more subtle benefit from eliminating extra address
 * arithmetic.  The 486 and Pentium have pipeline stalls when addresses
 * are modified immediately after being used.  By addressing the entire 
 * input array relative to a single unchanging pointer, we eliminate 
 * these stalls.

 * What do we do in the degenerate cases of purely horizontal and purely
 * vertical strels?  The purely horizontal strels use shifted inputs to
 * compute each output word, but don't use those shifted inputs to compute
 * any other output word.  We'll just do the shifting, but won't bother
 * storing the results in an array.

 * Purely vertical strels don't use shifted inputs.  However, many output
 * words to share the same inputs..  We could continue to address the
 * input words in place, skipping source bytes/line to get to each new one.
 * This is a lot of address arithmetic in the inner loop.  For purely vertical
 * strels, we'll continue to use the array of inputs so we can minimize
 * address arithmetic.

 * In the SPARC algorithms, we went to a great deal of trouble to use double
 * word (64 bit) write operations.  Writes on an SS2 are slow, so it was
 * beneficial.  The 386/486 has relatively fast 32 bit writes and no 64 bit
 * writes, so we will discard that complication.

 * Like the SPARC algorithms, we will have restrictions on the size of
 * structure elements we can handle.  The distance from the SE's center to
 * the edge of the SE (dimensions cx, cy, (nx-cx) and (ny-cy)) must all be
 * less than or equal to 31.  This fits within the frame added around the
 * image.

 * To specify the algorithm, define the ShiftRight(L, R, d) operator.  It
 * returns the least significant 32 bits of the result of shifting the 64 bit
 * quantity L,,R right by d bits.  Similarly, ShiftLeft(L, R, d) returns the
 * most significant 32 bits of the result of shifting the 64 bit quantity
 * L,,R left by d bits.

 * Define the current array of shifted inputs as InputCache.  Each time
 * we move down one image line, the current position of InputCache moves
 * down one line in the buffer allocated by the caller.  Thus, if we
 * process a total of v vertical source/dest words, the buffer holds
 * v overlapped versions of InputCache, each separated by one line of
 * inputs.  The width of a line of inputs in the input cache is (the
 * width of the SE) * (4 bytes/entry), not the width of the image.

 * The basic algorithm is ("word" = 32 bits):
 *    1. Start in the upper left corner of the current block in the
 *	output image.
 *    2. Given that output word, read the word of input image cy rows
 *	above it into Input.
 *    3. If cx > 0, read the word to the left of the input word into
 *	InputLeft.
 *    4. For each non-empty column i of the strel to the left of center,
 *	write ShiftRight(InputLeft, Input, cx-i) into location (0,i) of
 *	InputCache.  This is most easily done by working from the center's
 *	column (no shift) towards the left.
 *    5. If the center column of the strel is not empty, write Input
 *	into location (0, cx) of the array of shifted inputs.
 *    6. If cx < nx-1, read the word to the right of input into InputRight.
 *    7. For each non-empty column i of the strel to the right of center,
 *	write ShiftLeft(Input, InputRight, i-cx) into location (0,i) of
 *	InputCache.
 *    8. Repeat steps 2 through 8 for each row of the strel except for
 *	the last, moving down one image row and one row in InputCache
 *	on each iteration.
 *
 *	We have now preloaded the InputCache.  We can start processing.
 *
 *    9. Process one row of inputs from the InputCache from left to right.
 *	Start with the top row.  Each InputCache element (i,j) corresponds
 *	to SE location (i,j). Combine each input word with the current
 *	output word using the corresponding SE element as a guide.  There
 *	are five possibilities.  They are:
 *	   a. Don't Care: do nothing
 *	   b. partial result currently empty, current SE element is a HIT:
 *		move current (possibly shifted) input to partial result.
 *	   c. partial result currently empty, current SE element is a MISS:
 *		move inverse of (shifted) input to partial result.
 *	   d. partial result not empty, current SE element is a HIT:
 *		AND (shifted) input word with partial result.
 *	   e. partial result not empty, current SE element is a MISS:
 *		AND inverse of (shifted) input word with partial result.
 *   10. We have now processed all but the bottom row of the SE.  Read the
 *	word from the input image that's ny-1-cy rows below the output
 *	word.  WE may also need to read the words immediately to the left
 *	and right of that word to do double shifts.
 *   11. Proceed as we did in steps 3 through 7, with two exceptions.  First,
 *	combine each shifted input with the output after storing the input
 *	to the InputCache.  Second, if this is a purely horizontal strel,
 *	don't bother with the store.
 *   12. We have now processed one output word.  Write it to the output image.
 *	Adjust the InputCache pointer to the buffer of shifted inputs
 *	to point down one row.
 *   11. Repeat steps 9 through 12 for each remaining word in the current
 *	vertical output image stripe.

 * The real algorithm will have some optimizations.  We will simply skip the
 * Don't Care states.  The code generated to process the bottom loop can
 * be optimized if we know whether or not the current value in the shift
 * register must be restored after processing a cMiss.  It need not be
 * restored if there are no more shifts to be done in the bottom row..

 * After immense sillyness, I found out that the WATCOM Linker's PHARLAP format
 * causes the linker to produce code in the FLAT memory model.  All that stuff
 * in the Microsoft Assembler manual (and in the WATCOM
 * compiler manual and in the flags we give to the WATCOM compiler) about
 * small models is inoperative.  So, that makes things easy.  We write code as
 * though we had the small model and everything's peachy.  In the final system,
 * we'll be able to generate code relative to the data segment and execute it
 * with respect to the code segment.  This will work because both segments have
 * the same base (gee, I hope the code segment doesn't have a smaller limit!).

 * It didn't.  All the calling and returning worked fine.

 *******************************************************************

 * So how does the C code generate this stuff?  First, it decides whether
 * to do anything at all.  If the SE is too big, it just returns without
 * generating any code.
 *
 * We then figure out how big the code will be.  The size is needed for
 * buffer allocation.  It's ok to ask for a little more space than required,
 * so we'll just do a worst-case computation.  This equals the size of the
 * overhead stuff plus the code required to do loads for left and right
 * shifts for the last SE row plus enough code to do an AND-WITH-INVERSE (the
 * longest operation) for each element in every row.  We then ask for that much
 * memory.
 *
 * In the SPARC code, we replicated the basic processing loop many times with
 * different registers.  It made sense to figure out what the basic
 * opcodes were once, then plug them in as needed.  Here, there is only one
 * basic loop.  We'll just figure out what x86 opcodes are required as
 * needed.
 *
 * One word on the source of assembly opcodes: if I were very industrious,
 * I'd write a complete compiler producing 386 opcodes.  I'm not.  Instead,
 * the file mortemp2.asm contains a sample of a working assembly routine.
 * It's the successor to mortempl.asm, the template for the first version
 * of assembly code.
 *
 * Each of the blocks in mortemp2 are labelled.  We simply copy blocks of
 * code from mortemp2 as required.  Note that the particular routines being
 * generated only reference variables kept in the local frame (on the stack),
 * or passed in as parameters (also on the stack).  In particular, there
 * are no references to globals or other remote variables or labels.  This
 * allows us to use most addresses (which are really offsets from EBP)
 * from mortemp2 directly.  We must supply address offsets into the cache
 * of addressed inputs (since they depend on the SE) and jump displacements
 * ourselves.
 *
 * So, we just start in, writing the block of overhead code.  Next comes
 * the code that preloads the input cache with shifted inputs.  How is
 * this cache structured?  It contains one 32-bit word for each SE element.
 * That's because each SE element specifies one combination of (shifted)
 * input with the current output.  The whole idea of the input cache is
 * to keep one copy of the appropriately shifted input for each desired
 * combination.  The current input cache pointer points to the upper
 * left corner of the array of shifted inputs.  We still need to know how
 * to address the cache elements.
 *
 * When we do erosion, we naturally want to combine all the words surrounding
 * the destination word with the destination word.  For example, if the
 * SE element above and to the left of center is a cHit, we want to AND each
 * dest bit with the bit above and to the left in the source image.
 * The word holding all the bits that are above and to the left of the bits
 * in the dest word is in input cache word [cx-1, cy-1].  Therefore the
 * input cache words will correspond directly to SE elements.  So what,
 * you say?  In dilation, input cache word [x, y] will correspond to
 * SE element [cx-x, cy-y].  My continual confusion on this issue forces me
 * to spell it out each time.
 *
 * This code starts by reading the source image word cy rows above the
 * first source word.  If cx > 0, it also reads the source word to the
 * left in preparation for right shifts.  If there are any active SE
 * elements in the center's column, we store the unshifted input in the
 * center's column in the input cache, element (cx, 0).  We then proceed
 * to shift the input word right with incoming bits coming from the word
 * to the left.  After each shift i, we store the shifted word into
 * input cache element (cx-i, 0), working from the center's column towards
 * the left.  Well, we really only store the word if there are any active
 * SE elements in SE column cx-i.
 *
 * After reaching column 0, we re-read the word cy rows above the first
 * dest word and read the word to its right and shift the input word left
 * with incoming bits coming from that word on the left.  After each shift
 * i, we store the result into input cache element (cx+i, 0) if there are
 * and active elements in SE column cx+i.
 *
 * This process is repeated for cy-2 *more* input rows.  Each time we move
 * down one row in the image, one row in the SE and one row in the input
 * cache.  We stop just before the last row in the SE.  The input cache
 * now has enough data to enter the inner loop.  The inner loop expects
 * to read one more row of data at the end.
 *
 * For the sake of cleanliness, we'll have subroutines to generate the
 * opcodes that store data into the input cache.  The address displacements
 * for the opcodes may be null, one byte or 4 bytes.  The subroutine,
 * getStoreCache, will take all the current code buf pointers and the
 * desired input cache displacement.
 *
 * Following that is the start of the inner loop.  We're finally ready to
 * do something interesting - start putting out code that does real image 
 * processing.  The basic plan is to scan the structure element left-to-right,
 * top-to-bottom, generating assembly code as we go.  There are 5 possible
 * code blocks (well, 4 non-null blocks plus "don't care") each SE element.
 * If the dest is touched for the first time because of a cHit, we copy
 * the (shifted) input to the dest.  If the dest is touched the first time
 * because of a cMiss, we follow the copy with a NOT opcode..  After the
 * first time, each cHit results in a single AND with the (shifted) input 
 * and each cMiss loads the (shifted) input, inverts it, then does the AND.
 * Furthermore, the opcodes used may have either no address displacement,
 * a single byte address displacement or a 32-bit address displacement.
 *
 * We don't want to put in code that makes this choice at every spot where  
 * code is required, so we put this specialized type of code generation 
 * into a subroutine, called ip_genErodeSEElement.  This routine takes the  
 * current SE element (don't care, hit or miss), a boolean saying  
 * whether this is the first time the destination word has been touched
 * and the input cache offset used to access the proper shifted input.
 * It adds the proper x86 code to the current buffer.  None of the 
 * information used by ip_genErodeSEElement can be global as that would make 
 * the code non-reentrant.  All the info is passed in parameters.
 *
 * Code is generated in this fashion until we reach the bottom row of
 * the SE.  We revert to the behavior shown to fill the input cache;
 * reading the word ny-cy-1 image rows below the dest and the words to
 * the left and right as needed.  The difference between the prefill code
 * and this code is that we combine each shifted word with the dest as
 * needed as well as storing the input word in the input cache.  The code
 * to do this combining is also generated by ip_genErodeSEElement, which is
 * called with a special flag to indicate that we're on the last SE row
 * so the input data is in EBX, not in the input cache.
 *
 * That's about it.  At the bottom of the loop, we advance the input cache
 * pointer down one input line, write out the dest word, advance the
 * source and dest pointers down one line each and check to see if we're
 * at the end of the input block.  All this code is generated from the
 * prototype in mortemp2.asm.  After the conditional jump at the bottom of
 * the loop, we insert the return sequence.
 *
 * How does this all work with purely vertical structure elements?  Just fine.
 * We never read the words to the left and right of the center's column
 * since left/right shifting is not required.  We still use the input
 * cache since it saves us the time required to modify addresses to repeatedly
 * sweep through the source image.  We can use constant displacements to
 * address the input cache since we know it's pitch (set by size of SE).
 * We would have to use indexed arithmetic to access the source image, 
 * which adds a cycle on every access plus a possible interlock cycle 
 * because we increment the address immediately before each access.
 *
 * How does it work for purely horizontal structure elements?  Pretty good.
 * Since there is only one row to the structure element, we don't preload
 * anything into the input cache.  Since there's only one SE row, we don't
 * use any of the inputs after they've been shifted.  So, we totally ignore
 * the input cache.  It should go the same speed as the cache-less 
 * implementation of the algorithm.
 *
 *
 * INTERFACE CHANGE, 9/17/94.  Instead of having the routine that generates
 * assembly code always allocate the assembly buffer, we'll let the client
 * do it optionally.  This means we have to provide a routine that tells the
 * client the size of the buffer - i_getErodeAsmSize.  If the *generated_code
 * is NULL on entry to i_getErodeAsm, we allocate a buffer using MALLOC.
 * Otherwise, we assume the client has already allocated one.  We do not
 * shrink the buffer to fit after generating the code, that's the client's
 * job.
 *
 * BUG FIX 12/7/94:  Leigh Klotz found out that if we give this code a
 * strel that's entirely DONT_CARE terms, it generates random code.
 * Check for this case.  If it occurs, don't allocate any assembly code
 * at all.  i_fastDilate will fix things up.
 *
 */


    /* dimensions of the largest structure element we can handle */
#define cMaxShiftDist	31

    /* Labels in the template assembly code */
extern	UInt8	erT2Start, erT2PreLoadDec, erT2PreLoadInpCac;
extern	UInt8	erT2LoadInputCenter, erT2LoadInputLeft, erT2LoadInputRight;
extern	UInt8	erT2ShiftInputRight, erT2ShiftInputLeft;
extern	UInt8	erT2IncInputAddr;
extern	UInt8	erT2CacheMiss, erT2NonCacheInit, erT2NonCacheMiss;
extern	UInt8	erT2NonCacheMissRestore;
extern	UInt8	erT2NonCacheHitAnd, erT2NonCacheHitOr;
extern	UInt8	erT2LoadDestAddr;
extern	UInt8	erT2BottomLoop, erT2LoopJmp, erT2Return;
extern	UInt8	erT2CacheStore0, erT2CacheStoreShort, erT2CacheStoreLong;
extern	UInt8	erT2CacheLoadEAX0, erT2CacheLoadEAXShort;
extern	UInt8	erT2CacheLoadEAXLong, erT2NotEAX;
extern	UInt8	erT2CacheLoadEBX0, erT2CacheLoadEBXShort;
extern	UInt8	erT2CacheLoadEBXLong;
extern	UInt8	erT2CacheAnd0, erT2CacheAndShort, erT2CacheAndLong;
extern	UInt8	erT2CacheOr0, erT2CacheOrShort, erT2CacheOrLong;
extern	UInt8	erT2IncCachePtrShort, erT2IncCachePtrLong;
extern	UInt8	erT2SingleRightShift, erT2SingleLeftShift;
extern	UInt8	erT2DoubleLeftShift1;


extern	UInt32	startBlkLen2, preLoadDecBlkLen2, preLoadInpCacBlkLen2;
extern	UInt32	loadInputCenterBlkLen2, loadInputLeftBlkLen2;
extern	UInt32	loadInputRightBlkLen2;
extern	UInt32	shiftInputRightBlkLen2, shiftInputLeftBlkLen2;
extern	UInt32	incInputAddrBlkLen2, cacheMissBlkLen2;
extern	UInt32	nonCacheMissBlkLen2, nonCacheMissRestoreBlkLen2;
extern	UInt32	nonCacheHitOrBlkLen2, nonCacheHitAndBlkLen2;
extern	UInt32	nonCacheInitBlkLen2, loadDestAddrBlkLen2;
extern	UInt32	bottomLoopBlkLen2, jumpBlkLen2, returnBlkLen2;
extern	UInt32	cacheStore0BlkLen2, cacheStoreShortBlkLen2;
extern	UInt32	cacheStoreLongBlkLen2, cacheLoadEAX0BlkLen2;
extern	UInt32	cacheLoadEAXShortBlkLen2, cacheLoadEAXLongBlkLen2;
extern	UInt32	cacheNotEAXBlkLen2, cacheLoadEBX0BlkLen2;
extern	UInt32	cacheLoadEBXShortBlkLen2, cacheLoadEBXLongBlkLen2;
extern	UInt32	cacheAnd0BlkLen2 , cacheAndShortBlkLen2;
extern	UInt32	cacheAndLongBlkLen2 , cacheOr0BlkLen2;
extern	UInt32	cacheOrShortBlkLen2 , cacheOrLongBlkLen2;
extern	UInt32	incCachePtrShortBlkLen2 , incCachePtrLongBlkLen2;
extern	UInt32	singleLeftShiftBlkLen2, singleRightShiftBlkLen2;
extern	UInt32	doubleLeftShift1BlkLen2;

/* Define macros for common subroutine calls with fancy abort sequences.
 * This should clean up the code a lot since most of the information in
 * the calls does not change.
 */
#define	M_COPY_CODE(blkAddr, blkLen)	if ((__error__ = \
			ip_copyCode (pIPState, &blkAddr, blkLen, \
				     pCodeBuf, &bufIndex, bufSize)) \
					!= ia_successful) \
						goto genCleanup

#define M_GEN_ERODE_SE_ELEMENT(currentElement, lastRow, isLastCol)	if ((__error__ = \
			ip_genErodeSEElement (pIPState, data, currentElement, \
					      pCodeBuf, &bufIndex,\
					      bufSize, &firstActiveFound,\
					      lastRow, isLastCol)) \
							!= ia_successful) \
			  goto genCleanup
#define	M_STORE_CACHE(offset) \
	    if ((__error__ = ip_storeCache (pIPState, pCodeBuf, \
					  &bufIndex, bufSize, \
					  offset)) != ia_successful) \
		goto genCleanup

/* define better names for some of the variables. */
#define	numSECols	nx
#define	numSERows	ny
#define	centerX		cx
#define	centerY		cy


/* PUBLIC */
/************************************************************************
 *  i_getErodeAsmSize():  return number of bytes needed for assembly code
 *			used for erosion with this strel.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getErodeAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
    PROLOGUE;
    UInt32      maxTopSEElementBlkLen; /* max # of code bytes required
					   to process any SE element that's
					   above the last row. */
    UInt32      maxLastSEElementBlkLen; /* max # of code bytes required
					   to process any SE element that's
					   in the last row. */

	/* Reference cx to keep the compilers happy. */
    cx = cx;

	/* We'd like to compute the maximum number of opcode bytes required
	 * to retire any of the SE elements above the bottom row.  The ones
	 * in the bottom row are different because we're doing shifting
	 * at the same time.  There are four potential operations for any
	 * of the SE elements, initialize because of cHit, initialize
	 * because of cMiss, process a later cHit and process a later MISS.
	 * Iniitalizing because of a cHit merely requires a load.
	 * Processing a cHit is an AND, which is the same length.
	 * Initializing a cMiss has a load and a NOT opcode, one byte
	 * longer than initializing a cHit.
	 * Processing a later cMiss has a load, a NOT and an AND.  That's
	 * the longest sequence.  Luckily, it's also reasonable since
	 * an SE might have lots of MISSES.
	 */
    maxTopSEElementBlkLen = cacheLoadEBXLongBlkLen2 + cacheMissBlkLen2;

	/* This is the number of opcode bytes required to process a
	 * word on the bottom row in addition to the normal loads and
	 * shifts.  The shifted data will already be in EBX.  There are
	 * four cases: 1) Found a cHit and first time we've touched
	 * dest word; 2) Found a cHit and not first time we've touched
	 * dest word; 3) Found a cMiss and first time we've touched
	 * dest word; 4) Found a cMiss and not first time we've touched
	 * dest word.  The first case turns into a reg-reg move, 1 byte.
	 * The second turns into a reg-reg AND, still 1 byte.
	 * The third requires a reg-reg move and a NOT, 2 bytes.  The
	 * last case requires a reg-reg move, a NOT and an AND, 3 bytes.
	 * That's the winner.
	 */
    maxLastSEElementBlkLen = nonCacheMissBlkLen2;

	/* figure out how big a buffer to allocate for the code */
    *pNumAsmBytes =
		    /* initial boilerplate */
	       (startBlkLen2 +
		(centerY * preLoadDecBlkLen2) + /* adjust src address */
		preLoadInpCacBlkLen2 +		/* load input cache addr */
		    /* load center's word twice and words to left and
		     * right once each */
		(numSERows * (2*loadInputCenterBlkLen2 +
				loadInputLeftBlkLen2 +
				loadInputRightBlkLen2)) +
		    /* do one shift for each non-center column in
		     * every row.  The shift left and shift right opcodes 
		     * have the same length so we use the length from
		     * right shift. */
		(numSERows * (numSECols - 1) *
		    (shiftInputRightBlkLen2 + singleRightShiftBlkLen2)) +
		    /* store one word for each position in the input cache */
		(numSERows * numSECols * cacheStoreLongBlkLen2) +
		    /* process each input not corresponding to last SE row */
		((numSERows-1) * numSECols * maxTopSEElementBlkLen) +
		    /* process each input corresponding to last SE row */
		(numSECols * maxLastSEElementBlkLen) +
		loadDestAddrBlkLen2 + incCachePtrLongBlkLen2 +
		bottomLoopBlkLen2 +		/* bottom of inner loop */
		jumpBlkLen2 +		 	/* inner loop jump */
		returnBlkLen2);		/* subroutine return */

    NORMAL_RETURN;
}



/* PUBLIC */
/************************************************************************
 *  i_genErodeAsm():  generate assembly code to do erode defined by strel
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastErode as
 *			blockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_SETooBig:		Structure element was too big to
 *				generate accelerated code OR was composed
 *				entirely of DONT_CARE terms (which is silly).
 *	ia_bufferExceeded:	The routine tried to generate too much
 *				assembly code.  This is a bug in the routine.
 *				If it happens, please let us know.
 ************************************************************************/
Int32  CDECL 
i_genErodeAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    ErodeProc  *generated_code,
    Int32      *code_length)
{
    PROLOGUE;
    UInt8      *pCodeBuf;		/* output code buffer */
    Bool	bufAllocatedHere;	/* set if this proc allocated the
					 * asm code buffer. */
    Int32       bufIndex;		/* index of code buffer */
    Int32       bufSize;		/* size of mem allocated for code
					 *  buffer */
    Bool        firstActiveFound;	/* cTrue if first active SE
					 * element has been
					 * encountered.  The first SE
					 * element requires different code
					 * than the rest. */
    Int32	row, col;		/* index through the SE */
    Int32       innerLoopOffset;	/* top of inner loop */
    UInt8      *seColActive;		/* array of booleans telling whether
					 * each SE col has any HITs or
					 * MISSes above the bottom row. */
    Int32       lastPosition;		/* used to compute shift distance
					 * for double shift opcodes */
    Int8        lastShift, shiftDist;	/* used to remember previous shift
					 * distance.  Is the amount by which
					 * we need to shift ECX before the
					 * next shift */
    Bool        activeFound;		/* used to explore SE rows */
    Int32      *pJumpDisplacement;	/* point to jump offset */
    Int32      *pWordDisplacement;	/* point to word address offset */
    Int8	activeOnLeft, activeOnRight, activeInCenter;
    Int32	leftmostCol, rightmostCol;
					/* booleans used to decide when to
					 * generate code */
    
    UInt32      rowOffset;		/* holds (row * numSECols) */


	/* First, decide if the SE is too big. */
    if ((centerX > cMaxShiftDist) ||
	(centerY > cMaxShiftDist) ||
	((numSECols - centerX - 1) > cMaxShiftDist) ||
	((numSERows - centerY - 1) > cMaxShiftDist) )
    {
        return ia_SETooBig;
    }

	/* see if the strel is composed entirely of cDontCare terms.
	 * If so, fail as above.
	 */
    __error__ = ia_SETooBig;
    for (rowOffset = 0;
	 (rowOffset < ny*nx) && (__error__ == ia_SETooBig);
	 rowOffset++)
    {
	if ((data[rowOffset] == cHit) || (data[rowOffset] == cMiss))
	{
	    __error__ = ia_successful;
	    break;
	}
    }
	/* if we didn't find any hits, return now */
    if (__error__ == ia_SETooBig)
	return ia_SETooBig;

	/* the strel's blockWidth is 4 bytes in all cases, so we might as well
	   set it now. */
    *blockWidth = 4;

	/* if we don't already have an assembly buffer, get one. */
    i_getErodeAsmSize (pIPState, nx, ny, cx, cy, &bufSize);
	/* In Visual C++, turn off the warning that says we're about to
	 * cast a pointer to a function to a pointer to data, since
	 * that's exactly what we intend to do.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    pCodeBuf = (UInt8 *) *generated_code;
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

    if (pCodeBuf == NULL)
    {
	pCodeBuf = (UInt8 *) MALLOC(bufSize);
	bufAllocatedHere = cTrue;
    }
    else
	bufAllocatedHere = cFalse;

	/* if the pointer is null, punt.  Otherwise, return the address
	 * of the new buffer.
	 */
    if (pCodeBuf == NULL)
        return ia_nomem;
    else
	/* In Visual C++, turn off the warning that says we're casting
	 * a pointer to data to a pointer to a function.  We're going
	 * to do exactly that.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
	*generated_code = (ErodeProc) pCodeBuf;
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif

	/* Initialize the index into the code buffer.  We use an index
	 * because we usually want to use the offset into the buffer
	 * to compute things, not the absolute buffer pointer.
	 */
    bufIndex = 0;

	/* Record the presence of active elements for each SE column
	 * *if the row is above the bottom row*.  seColActive will
	 * be used to decide whether to store data into the input
	 * cache.
	 * We don't need to store a shifted input for a column if
	 * it has no active elements. */
    seColActive = (UInt8 *)CALLOC(numSECols, sizeof(UInt8));

    activeOnLeft = activeOnRight = activeInCenter = cFalse;
    leftmostCol = numSECols;
    rightmostCol = 0;
    for (col = 0; col < numSECols; col++)
    {
	for (row = 0, rowOffset = 0;
	     row < numSERows;
	     rowOffset += numSECols, row++)
	{
	    if (data[rowOffset + col] != cDontCare)
	    {
		if (row < numSERows-1) seColActive[col] = cTrue;
		rightmostCol = col;
		if (col < leftmostCol) leftmostCol = col;
		if (col < centerX) activeOnLeft = cTrue;
		else
		if (col == centerX) activeInCenter = cTrue;
		else activeOnRight = cTrue;
		row = numSERows;		/* no need to look further */
	    }
	}
    }


/* *************** START OF CODE GENERATION ********************** */


	/* Start with the initialization code */
    M_COPY_CODE(erT2Start, startBlkLen2);

	/* We now move the source address up by centerY image lines.  This
	 * is how the strel->cy parameter affects the output.
	 * Note that nothing happens if the SE only has one row (since
	 * centerY must be 0 in that case). */
    for (row = 0; row < centerY; row++)
	M_COPY_CODE(erT2PreLoadDec, preLoadDecBlkLen2);

	/* If we're using the input cache, load its address into EDI
	 * register.  We do this now to avoid the interlock between the
	 * last adjustment of ESI above and the first use of ESI. */
    if (numSERows > 1)
	M_COPY_CODE(erT2PreLoadInpCac, preLoadInpCacBlkLen2);

	/* Load dest address into EDX. */
    M_COPY_CODE(erT2LoadDestAddr, loadDestAddrBlkLen2);


/* *********************** PRELOAD INPUT CACHE *********************** */


	/* Preload the input cache.  This is only done if there is more
	 * than one row in the SE. */
    for (row = 0, rowOffset = 0;
	 row < numSERows-1;
	 rowOffset += numSECols, row++)
    {
	    /* There's bound to be something active somewhere in the SE
	     * so load the center's word.  Note that this word may well be
	     * used to create shifted inputs that will be used by many
	     * different rows of SE processing (NOT just this row!!, that's
	     * the whole point of keeping a cache of shifted inputs). */
	M_COPY_CODE(erT2LoadInputCenter, loadInputCenterBlkLen2);

	    /* If there are no active cols to the left or right, increment the
	     * address pointer as soon as possible to avoid interlocks. */
	if (!activeOnLeft && !activeOnRight)
	{
	    M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
	}

	    /* if the center's column is active above bottom row, store
	     * this word in the cache of shifted inputs. */
	if (seColActive[centerX])
	{
	    M_STORE_CACHE((rowOffset + centerX));
	}

	/* ********************* PRELOAD LEFT SIDE ********************** */


	    /* If there are any active SE columns to the left of center, load
	     * the word to the left. */
	if (activeOnLeft)
	{
	    M_COPY_CODE(erT2LoadInputLeft, loadInputLeftBlkLen2);

		/* If there are no active elements to the right, increment the
		 * address pointer as soon as possible to
		 * avoid interlocks. */
	    if (!activeOnRight)
	    {
		M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
	    }

		/* set activeFound to cFalse here.  We'll set it to true
		 * when the first double shift has been done.  Each subsequent
		 * double shift will be immediately preceeded with a
		 * single shift of ECX so the incoming bits are correct
	 	 * (what a weird chip!)
		 */
	    lastPosition = centerX;
	    activeFound = cFalse;
	    lastShift = 0;	/* keep the compiler happy */

		/* Shift and store each version of input if the column is
		 * active above bottom row. */
	    for (col = centerX - 1; col >= 0; col--)
	    {
		if (seColActive[col])
		{
		    if (activeFound) /* have already put out an shrd, so
					need to preceed the next one with
					an shr of ecx */
		    {
			M_COPY_CODE(erT2SingleRightShift,
				        singleRightShiftBlkLen2);
			pCodeBuf[bufIndex-1] = lastShift;
		    }
		    activeFound = cTrue;
		    M_COPY_CODE(erT2ShiftInputRight,
				   shiftInputRightBlkLen2);
			/* fix up the shift distance */
		    lastShift = (UInt8) (lastPosition - col);
		    pCodeBuf[bufIndex-1] = lastShift;
		    lastPosition = col;
			/* Store the shifted input into the cache. */
		    M_STORE_CACHE((rowOffset + col));
		}
	    }  /* end of preloading input cache with left side of row */

		/* If there are any cols on the right, reload the center's
		 * value.  That way, the code below can be assured that
		 * it is entered with the unshifted data in EBX. */
	    if (activeOnRight)
	    {
		M_COPY_CODE(erT2LoadInputCenter, loadInputCenterBlkLen2);
	    }

	}	/* end of case where there are cols to the left of center
		 * when preloading input cache. */



	/* ********************* PRELOAD RIGHT SIDE ********************** */


	    /* If there are columns to the right of center, preload them too. */
	if (activeOnRight)
	{
		/* load the word to the right */
	    M_COPY_CODE(erT2LoadInputRight, loadInputRightBlkLen2);

		/* Increment the address pointer as soon as possible to
		 * avoid interlocks if we haven't already done so. */
	    M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);

		/* remember where we started shifting.  Each left shift
		 * by 1 will be done with two adds (2 cycles).  Left
		 * shifts by more than 1 will use two shifts (4 cycles).
		 */
	    lastPosition = centerX;

		/* Shift and store each version of input if the column is
		 * active. */
	    for (col = centerX + 1; col < numSECols; col++)
	    {
		if (seColActive[col])
		{
		    shiftDist = (UInt8) (col - lastPosition);
		    if (shiftDist > 1) /* shift by more than 1 */
		    {
			M_COPY_CODE(erT2ShiftInputLeft, shiftInputLeftBlkLen2);
			pCodeBuf[bufIndex-1] = shiftDist;
			if (col < rightmostCol)	/* shift ECX to catch up */
			{
			    M_COPY_CODE(erT2SingleLeftShift,
				       singleLeftShiftBlkLen2);
			    pCodeBuf[bufIndex-1] = shiftDist;
			}
		    }
		    else	/* shift by 1 */
		    {
			M_COPY_CODE(erT2DoubleLeftShift1,
				  doubleLeftShift1BlkLen2);
		    }
			/* fix up the shift distance */
		    lastPosition = col;
			/* Store the shifted input into the cache. */
		    M_STORE_CACHE((rowOffset + col));
		}
	    }  /* end of preloading input cache with right side of row */
	}	/* end of activeOnRight case for preloading cache rows */
    }	/* end of loop that generates code to preload the input cache */



/* *********************** INNER LOOP *********************** */


	/* We're at the beginning of the inner loop.  Remember this
	 * address so we can branch back to it.
 	 */
    innerLoopOffset =  bufIndex;

	/* Go through the SE top to bottom, left to right.  At each
	 * location, if there's a cHit or cMiss, generate the appropriate
	 * code to combine the cached input with the current output.
	 * Stop before the last SE row.  Note that nothing will happen here
	 * if the SE has only one row. */
    firstActiveFound = cFalse;
    for (row = 0, rowOffset = 0;
	 row < numSERows - 1;
	 rowOffset += numSECols, row++)
    {
	for (col = 0; col < numSECols; col++)
	{
	    if (data[rowOffset+col] != cDontCare)
	    {
		M_GEN_ERODE_SE_ELEMENT((rowOffset + col), cFalse, cFalse);
	    }
	}
    }


/* ******************* BOTTOM LINE OF INNER LOOP *************** */

	/* This looks a lot like prefilling the input cache, except we
	 * only do one row.  It has two other differences.  First, if
	 * necessary, we combine each shifted input with the current 
	 * output before storing it.  Second, we only store the input 
	 * if the SE is not horizontal (will need the input again sometime).
	 */

	/* There's no need to fixup "row" or "rowOffset", they are
	 * already equal to numSERows-1 and (numSERows-1)*numSECols
	 * respectively.
	 */

	/* However, we would like to load the destination address into
	 * ECX as soon as possible to avoid interlock.  "As soon as
	 * possible" is immediately after the last shift - or immediately
	 * before the last store to the input cache.  We figure out
	 * when that is by using rightmostCol.  It was set above with
	 * the number of the rightmost active column in the SE.
	 */

	/* There's bound to be something active somewhere in the SE
	 * so load the center's word.  Note that this word may well be
	 * used to create shifted inputs that will be used by many
	 * different rows of SE processing (NOT just this row!!, that's
	 * the whole point of keeping a cache of shifted inputs).
	 */
    M_COPY_CODE(erT2LoadInputCenter, loadInputCenterBlkLen2);

	/* If there are no active cols to the left or right, increment the
	 * address pointer as soon as possible to avoid interlocks. */
    if (!activeOnLeft && !activeOnRight)
    {
	M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
    }

	/* if the center's column is active above the bottom row,
	 * store this word in the cache of shifted inputs. */
    if (seColActive[centerX])
    {
	M_STORE_CACHE((rowOffset + centerX));
    }

	/* If this element of the SE is active, combine the unshifted
	 * input with the current dest value.  Note the lastRow parameter
	 * is passed as cTrue here.
	 * BUG FIX: we were passing "rightmostCol == centerX" as the
	 * islastCol parameter.  If the center were on the right side
	 * of the SE, and the center's element was a cMiss, this would
	 * cause ip_genErodeSEElement to neglect to restore ebx after
	 * processing the miss.  We added the leftmostCol term.
	 */
    if (data[rowOffset+centerX] != cDontCare)
    {
	M_GEN_ERODE_SE_ELEMENT((rowOffset + centerX), cTrue,
			   ((rightmostCol == centerX) &&
			    (leftmostCol == centerX)));
    }


    /* **************** PROCESS LEFT SIDE OF BOTTOM LINE ************* */


	/* If there are any active SE columns to the left of center, load
	 * the word to the left. */
    if (activeOnLeft)
    {
	M_COPY_CODE(erT2LoadInputLeft, loadInputLeftBlkLen2);

	    /* If there are no active elements to the right, increment the
	     * address pointer as soon as possible to avoid interlocks. */
	if (!activeOnRight)
	{
	    M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);
	}

	    /* set activeFound to cFalse here.  We'll set it to true
	     * when the first double shift has been done.  Each subsequent
	     * double shift will be immediately preceeded with a
	     * single shift of ECX so the incoming bits are correct
	     * (what a weird chip!)
	     */
	lastPosition = centerX;
	activeFound = cFalse;
	lastShift = 0;		/* keep the compiler happy */

	    /* Shift and store each version of input if the column is
	     * active. */
	for (col = centerX - 1; col >= 0; col--)
	{
	    if (seColActive[col] || (data[rowOffset+col] != cDontCare))
	    {
		if (activeFound) /* have already put out an shrd, so
				    need to preceed the next one with
				    an shr of ECX */
		{
		    M_COPY_CODE(erT2SingleRightShift,
				   singleRightShiftBlkLen2);
		    pCodeBuf[bufIndex-1] = lastShift;
		}
		activeFound = cTrue;
		M_COPY_CODE(erT2ShiftInputRight, shiftInputRightBlkLen2);
		    /* fix up the shift distance */
		lastShift = (UInt8) (lastPosition - col);
		pCodeBuf[bufIndex-1] = lastShift;
		lastPosition = col;

		    /* Store the shifted input into the cache if we'll
		     * ever need this input again. */
		if (seColActive[col])
		{
		    M_STORE_CACHE((rowOffset + col));
		}

		    /* If there's an active element at this SE location,
		     * combine the shifted input with the current dest.
		     * There's no way to miss this statement if it's needed
		     * because if there's something at this SE location,
		     * this SE column must be active. */
		if (data[rowOffset+col] != cDontCare)
		{
		    M_GEN_ERODE_SE_ELEMENT((rowOffset + col), cTrue,
				       (leftmostCol == col));
		}
	    }
	}  /* end of processing input cache with left side of row */

	    /* If there are any cols on the right, reload the center's
	     * value.  That way, the code below can be assured that
	     * it is entered with the unshifted data in EBX. */
	if (activeOnRight)
	{
	    M_COPY_CODE(erT2LoadInputCenter, loadInputCenterBlkLen2);
	}

    }	/* end of case where there are cols to the left of center
	 * when processing bottom line of SE. */


    /* **************** PROCESS RIGHT SIDE OF BOTTOM LINE ************* */


	/* If there are columns to the right of center, process them too. */
    if (activeOnRight)
    {
	    /* load the word to the right */
	M_COPY_CODE(erT2LoadInputRight, loadInputRightBlkLen2);

	    /* Increment the address pointer as soon as possible to
	     * avoid interlocks if we haven't already done so. */
	M_COPY_CODE(erT2IncInputAddr, incInputAddrBlkLen2);

	    /* remember where we started shifting.  Each left shift
	     * by 1 will be done with two adds (2 cycles).  Left
	     * shifts by more than 1 will use two shifts (4 cycles).
	     */
	lastPosition = centerX;

	    /* Shift and store each version of input if the column is
	     * active. */
	for (col = centerX + 1; col < numSECols; col++)
	{
	    if (seColActive[col] || (data[rowOffset+col] != cDontCare))
	    {
		shiftDist = (UInt8) (col - lastPosition);
		if (shiftDist > 1) /* shift by more than 1 */
		{
		    M_COPY_CODE(erT2ShiftInputLeft, shiftInputLeftBlkLen2);
		    pCodeBuf[bufIndex-1] = shiftDist;
		    if (col < rightmostCol)	/* shift ECX to catch up */
		    {
			M_COPY_CODE(erT2SingleLeftShift,
				   singleLeftShiftBlkLen2);
			pCodeBuf[bufIndex-1] = shiftDist;
		    }
		}
		else	/* shift by 1 */
		{
		    M_COPY_CODE(erT2DoubleLeftShift1,
			      doubleLeftShift1BlkLen2);
		}
		    /* fix up the shift distance */
		lastPosition = col;

		    /* Store the shifted input into the cache if we'll
		     * ever need this input again. */
		if (seColActive[col])
		{
		    M_STORE_CACHE((rowOffset + col));
		}

		    /* If there's an active element at this SE location,
		     * combine the shifted input with the current dest.
		     * There's no way to miss this statement if it's needed
		     * because if there's something at this SE location,
		     * this SE column must be active. */
		if (data[rowOffset+col] != cDontCare)
		{
		    M_GEN_ERODE_SE_ELEMENT((rowOffset + col), cTrue,
				       (rightmostCol == col));
		}
	    }
	}  /* end of processing input cache with right side of row */
    }	/* end of activeOnRight case for processing cache rows */


	/* ENTIRE SE HAS BEEN PROCESSED FOR ONE OUTPUT WORD */

	/* If we're using the input cache, increment the input cache
	 * pointer.  There are two possible types of increment.  One
	 * uses a byte constant and the other uses a 32-bit immediate.  
	 * We use the byte immediate if the number of bytes/line in 
	 * the *input cache* is less than 128.
	 */
    if (numSERows > 1)
    {
	if (numSECols < 128/4)
	{
	    M_COPY_CODE(erT2IncCachePtrShort, incCachePtrShortBlkLen2);
	    pCodeBuf[bufIndex-1] = (UInt8) (4*numSECols);
	}
	else
	{
	    M_COPY_CODE(erT2IncCachePtrLong, incCachePtrLongBlkLen2);
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = (Int32) (4*numSECols);
	}
    }

	/* put out the code for the bottom of the loop.	 */
    M_COPY_CODE(erT2BottomLoop, bottomLoopBlkLen2);

	/* and the jump at the bottom of the inner loop */
    M_COPY_CODE(erT2LoopJmp,  jumpBlkLen2);

	/* fix up the row loop's offset */
    pJumpDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
    *pJumpDisplacement = (Int32) (innerLoopOffset - bufIndex);

	/* and finally, the return code */
    M_COPY_CODE(erT2Return, returnBlkLen2);

	/* Now for a little error checking.  If we've generated more code
	   than fits in the buffer, complain */
    if (bufIndex >= bufSize)
    {
	/* fprintf(stderr, " Buffer overflow, memory has been *smashed* in %s\n %d bytes written into buffer of size %d\n",
		procName, bufIndex, bufSize); */
	__error__ = ia_bufferExceeded;
	goto genCleanup;
    }

    *code_length = bufIndex;
    __error__ = ia_successful;

genCleanup:
    FREE(seColActive);		/* free the array of booleans */
    if ((__error__ != ia_successful) && (bufAllocatedHere))
	FREE(pCodeBuf);
    NORMAL_RETURN;

}




/* Define the macro used to call ip_copyCode from ip_genErodeSEElement and
 * ip_storeCache. */
#define	G_COPY_CODE(blkAddr, blkLen) \
		__error__ = ip_copyCode (pIPState, &blkAddr, blkLen, \
				         pCodeBuf, pBufIndex, \
				         bufSize)


/* This is a private procedure */
/* PRIVATE */
/**************************************************************************
 *  ip_genErodeSEElement()
 *		Decides what type of assembly code to generate for the
 *		specified sub-element (cHit, cMiss or cDontCare) of the
 *		structure element.  Normally, a cHit generates an AND
 *		operation and a cMiss generates an AND_WITH_INVERSE op.
 *		However, if the output word is being touched for the first
 *		time, a cHit generates a COPY and a cMiss turns into a
 *		COPY_INVERSE.  Once the choice has been made, the assembly
 *		code is actually put into the output buffer by ip_copyCode.
 *
 *	data:		Pointer to array of SE elements as defined
 *			in i_genErodeAsm.  Used to identify type of
 *			erosion op required (cHit or cMiss).
 *	currentElement:	Index into data; used to identify type of
 *			erosion op required (cHit or cMiss).
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			by ip_copyCode.  This routine will also
 *			update the index.
 *	bufSize:	Max size of the code buffer.  ip_copyCode
 *			will check for buffer overflow before doing
 *			the write.
 *	pFirstActiveFound: Pointer to boolean which is cFalse until the
 *			first operation affecting the output word
 *			has been generated.  This is used to
 *			distinguish between the AND/AND_WITH_INVERSE
 *			case and the COPY/COPY_INVERSE case.
 *			Once any code has been generated which touches
 *			the output, word, this routine will set the
 *			boolean to cTrue.
 *	lastRow:	Boolean.  Set to cTrue if this is the last row of
 *			the SE.  Otherwise, set to cFalse.  The opcodes
 *			generated are different for the two cases.
 *	isLastCol:	Boolean.  Set to cTrue if this is the last active
 *			column of the SE.  This is used to decide whether
 *			EBX should be restored if we are processing the
 *			last row of the SE.  If this is the last (rightmost
 *			active) column, there is no need to restore EBX.
 *
 *  This routine returns one of two status codes directly from ip_copyCode.
 * They are:
 *	ia_successful:		Code was put into the output buffer.
 *	ia_bufferExceeded:	Tried to put too much code into the output
 *				buffer.  This should never happen.
 *				The calling routine, i_genErodeAsm, will return
 *				ia_bufferExceeded, someone will complain and
 *				we'll debug the code.
 ****************************************************************************/
Int32  CDECL 
ip_genErodeSEElement (
    IPState    *pIPState,
    const UInt8 *data,
    Int32	currentElement,
    UInt8      *pCodeBuf,
    Int32     *pBufIndex,
    Int32	bufSize,
    UInt32     *pFirstActiveFound,
    Int32	lastRow,
    Int32	isLastCol)
{
    PROLOGUE;
    
    Int32	addrDisplacement;	/* number of bytes needed for
					 * address displacement of opcodes */
    Int32      *pWordDisplacement;	/* pts to address displacements */

	/* Is this the last row?  If not, we use the array of shifted inputs
	 * as the source of the operation. */
    if (!lastRow)
    {
	    /* calculate the length of the address displacement needed in
	     * the various opcodes below that access the cache of shifted
	     * inputs.
	     * If the offset (4 * currentElement) from the beginning of
	     * the input cache is 0, we use no address displacement.  If the
	     * offset is less than 128 BYTES, we use a single byte
	     * displacement.  If the displacement is bigger than 128 bytes,
	     * we use a load op with a 4 byte displacement.
	     */
	if (currentElement == 0)
	    addrDisplacement = 0;
	else
	if (currentElement < 128/4)
	    addrDisplacement = 1;
	else
	    addrDisplacement = 4;

	    /* Have we put out any processing code yet? */
	if ((*pFirstActiveFound) == cFalse)
	{
		/* nope, this is the first time.  We always load the input into
		 * the partial result.  Choose the load op based on the
		 * value of addrDisplacement.
		 */
	    switch (addrDisplacement) {
	    case 0:
		G_COPY_CODE(erT2CacheLoadEAX0, cacheLoadEAX0BlkLen2);
		break;
	    case 1:
		G_COPY_CODE(erT2CacheLoadEAXShort,
			       cacheLoadEAXShortBlkLen2);
		if (__error__ == ia_successful)
		    pCodeBuf[(*pBufIndex)-1] = (UInt8) (4*currentElement);
		break;
	    default:
		G_COPY_CODE(erT2CacheLoadEAXLong, 
			       cacheLoadEAXLongBlkLen2);
		if (__error__ == ia_successful)
		{
		    pWordDisplacement = (Int32 *) &(pCodeBuf[(*pBufIndex)-4]);
		    *pWordDisplacement = (Int32) (4*currentElement);
		}
		break;
	    }

		/* and if it was a cMiss, we invert the data */
	    if ((data[currentElement] == cMiss) && 
		      (__error__ == ia_successful))
		G_COPY_CODE(erT2NotEAX, cacheNotEAXBlkLen2);
	    *pFirstActiveFound = cTrue;	/* don't do this again */
	    NORMAL_RETURN;
	}
	else
	{
		/* this isn't the first time, so we'll do an AND (cHit) or 
		    AND-WITH-INVERSE (cMiss). */
	    if (data[currentElement] == cHit)
	    {
		switch (addrDisplacement) {
		case 0:
		    G_COPY_CODE(erT2CacheAnd0, cacheAnd0BlkLen2);
		    NORMAL_RETURN;
		case 1:
		    G_COPY_CODE(erT2CacheAndShort,  cacheAndShortBlkLen2);
		    if (__error__ == ia_successful)
			pCodeBuf[(*pBufIndex)-1] = (UInt8) (4*currentElement);
		    NORMAL_RETURN;
		default:
		    G_COPY_CODE(erT2CacheAndLong, cacheAndLongBlkLen2);
		    if (__error__ == ia_successful)
		    {
			pWordDisplacement =
				(Int32 *) &(pCodeBuf[(*pBufIndex)-4]);
			*pWordDisplacement = (Int32) (4*currentElement);
		    }
		    NORMAL_RETURN;
		}
	    }
	    else
	    {
		    /* processing a miss takes two steps.  We load
		     * the data into EBX first, then invert and and it.
		     * The load requires one of three displacment sizes.
		     */
		switch (addrDisplacement) {
		case 0:
		    G_COPY_CODE(erT2CacheLoadEBX0, cacheLoadEBX0BlkLen2);
		    break;
		case 1:
		    G_COPY_CODE(erT2CacheLoadEBXShort,
				   cacheLoadEBXShortBlkLen2);
		    if (__error__ == ia_successful)
			pCodeBuf[(*pBufIndex)-1] = (UInt8) (4*currentElement);
		    break;
		default:
		    G_COPY_CODE(erT2CacheLoadEBXLong, 
				   cacheLoadEBXLongBlkLen2);
		    if (__error__ == ia_successful)
		    {
			pWordDisplacement =
				(Int32 *) &(pCodeBuf[(*pBufIndex)-4]);
			*pWordDisplacement = (Int32) (4*currentElement);
		    }
		    break;
		}
		if (__error__ == ia_successful)
		{
		    G_COPY_CODE(erT2CacheMiss, cacheMissBlkLen2);
		}
		NORMAL_RETURN;
	    }	/* done Processing cMiss using cache as source of inputs. */
	}	/* done processing hits and misses that weren't touching
		 * the dest for the first time and weren't on the last
		 * row of the SE.
		 */
    }	/* done processing hits and misses that weren't on the last line
	 * of the SE. */

    else

    {

    /* ********************* PROCESS LAST ROW ************************* */

	/* This is the last row.  We have the same cases again, except
         * the code generated is different because the source is in EBX.
	 * The code is also different because we can't afford to
	 * destroy the data in EBX if we'll need it for later shifts.
	 * How do we know if we'll need it again?  The parameter isLastCol
	 * is set to cTrue if we are generating code for the rightmost
	 * active column.
	 */
	    /* Have we put out any processing code yet?  If the SE has
	     * only one active row, we might not have touched the
	     * destination yet.
	     */
	if ((*pFirstActiveFound) == cFalse)
	{
		/* This is the first time the dest has been touched.
		 * We always move the data from EBX into EAX.
		 */
	    G_COPY_CODE(erT2NonCacheInit, nonCacheInitBlkLen2);

		/* and if it was a cMiss, we invert the data */
	    if ((data[currentElement] == cMiss) && 
		      (__error__ == ia_successful))
		G_COPY_CODE(erT2NotEAX, cacheNotEAXBlkLen2);
	    *pFirstActiveFound = cTrue;	/* don't do this again */
	    NORMAL_RETURN;
	}
	else
	{
		/* this isn't the first time, so we'll do an AND (cHit) or 
		    AND-WITH-INVERSE (cMiss). */
	    if (data[currentElement] == cHit)
	    {
		G_COPY_CODE(erT2NonCacheHitAnd, nonCacheHitAndBlkLen2);
		NORMAL_RETURN;
	    }
	    else
	    {
		    /* If this is not the rightmost column, invert EBX,
		     * do the AND and restore EBX. */
		if (!isLastCol)
		{
		    G_COPY_CODE(erT2NonCacheMissRestore, 
				   nonCacheMissRestoreBlkLen2);
		    NORMAL_RETURN;
		}
		else
		    /* Invert EBX, do the AND, but don't bother restoring
		     * EBX since we won't need it again. */
		{
		    G_COPY_CODE(erT2NonCacheMiss, nonCacheMissBlkLen2);
		    NORMAL_RETURN;
		}
	    }	/* done Processing cMiss using EBX as source of inputs. */
	}	/* done processing hits and misses that weren't touching
		 * the dest for the first time and *were* on the last
		 * row of the SE.
		 */
    }	/* done processing hits and misses that *were* on the last line
	 * of the SE. */
}


/* This is a private routine */
/* PRIVATE */
/****************************************************************************
 *  ip_storeCache()
 *		Uses ip_copyCode to generate the appropriate
 *		store opcode.  The 3 versions of store opcode
 *		differ only in the length of their address displacements,
 *		0, 1 or 4 bytes.
 *		If things went well, *pBufIndex is incremented
 *		to point to next available byte and routine
 *		returns ia_successful.
 *
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			If no buffer overflow, *pBufIndex will be incremented
 *			to point to next available byte in code buffer.
 *	bufSize:	Max size of the code buffer.  This routine
 *			will check for buffer overflow before doing
 *			the write.
 *	offset:		offset to current input cache element in units of
 *			32-bit words.  For example, if we're on row 2, column
 *			3, this is 2*numSECols + 3.
 ***************************************************************************/
Int32  CDECL 
ip_storeCache (
    IPState    *pIPState,
    UInt8      *pCodeBuf,
    Int32     *pBufIndex,
    Int32	bufSize,
    Int32	offset)
{
    PROLOGUE;
    
    UInt32	addrDisplacement;	/* number of bytes needed for
					 * address displacement of opcodes */
    Int32      *pWordDisplacement;	/* pts to address displacements */

	/* calculate the length of the address displacement needed in
	 * the store to the input cache.
	 * If the offset (4 * currentElement) from the beginning of
	 * the input cache is 0, we use no address displacement.  If the
	 * offset is less than 128 BYTES, we use a single byte
	 * displacement.  If the displacement is bigger than 128 bytes,
	 * we use a load op with a 4 byte displacement.
	 */
    if (offset == 0)
	addrDisplacement = 0;
    else
    if (offset < 128/4)
	addrDisplacement = 1;
    else
	addrDisplacement = 4;

	/* Choose the load op based on the value of addrDisplacement. */
    switch (addrDisplacement) {
    case 0:
	G_COPY_CODE(erT2CacheStore0, cacheStore0BlkLen2);
	NORMAL_RETURN;
    case 1:
	G_COPY_CODE(erT2CacheStoreShort, cacheStoreShortBlkLen2);
	if (__error__ == ia_successful)
	    pCodeBuf[(*pBufIndex)-1] = (UInt8) (4*offset);
	NORMAL_RETURN;
    default:
	G_COPY_CODE(erT2CacheStoreLong, cacheStoreLongBlkLen2);
	if (__error__ == ia_successful)
	{
	    pWordDisplacement = (Int32 *) &(pCodeBuf[(*pBufIndex)-4]);
	    *pWordDisplacement = (Int32) (4*offset);
	}
	NORMAL_RETURN;
    }
}



/* This is a private routine */
/* PRIVATE */
/***************************************************************************
 *  ip_copyCode()
 *		Check bufIndex against bufSize and
 *		copies code from template to code buffer.
 *              Code stored into pCodeBuf[*pBufIndex].
 *		If *pBufIndex is too big, routine returns ia_bufferExceeded.
 *		If things went well, *pBufIndex is incremented
 *		to point to next available byte and routine
 *		returns ia_successful.
 *
 *	source:		Pointer to beginning of code template in mortempl.asm.
 *	length:		Length of code template in mortempl.asm.
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			If no buffer overflow, *pBufIndex will be incremented
 *			to point to next available byte in code buffer.
 *	bufSize:	Max size of the code buffer.  This routine
 *			will check for buffer overflow before doing
 *			the write.
 ***************************************************************************/
Int32  CDECL 
ip_copyCode (
    IPState    *pIPState,
    UInt8      *source,
    Int32	length,
    UInt8      *pCodeBuf,
    Int32     *pBufIndex,
    Int32	bufSize)
{
    PROLOGUE;
    if (*pBufIndex + length  > bufSize)
    {
	    /* DEBUGGING PRINT STATEMENT, remove in production code */
	/* fprintf(stderr, "Erode's assembly code buffer size exceeded!!"); */
	return ia_bufferExceeded;
    }
    else
    {
	memcpy(&(pCodeBuf[(*pBufIndex)]), source, length);
	*pBufIndex += length;
	NORMAL_RETURN;
    }
}

/*

$Log: pcgenero.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.26  1998/01/13 23:54:08  ddavies
pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.25  1997/02/06 01:36:41  ddavies
pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v

Fix gcc uninitialized var warnings on freebsd.  So far as I know, these
didn't do anyting.

Revision 1.24  1995/12/13 02:52:46  ejaquith
pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.23  1995/10/18  04:04:43  ddavies
 * pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.22  1995/10/03  21:16:48  ddavies
 * pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v
 *
 * Include shrmorph.pub/prv instead of shrip.pub/prv.  Change arg of
 * i_genErodeAsm from "void **generated_code" to "ErodeProc *generated_code".
 * Change casts when using generated_code.
 *
 * Revision 1.21  1994/12/08  04:37:36  ddavies
 * pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v
 *
 * Check to see if structure element contains any hits or any misses in
 * i_genErodeAsm.  If not, return ia_SETooBig.  This really means "the
 * structure element is too weird," in this case.
 *
 * Revision 1.20  1994/09/20  02:52:15  ddavies
 * pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v
 *
 * Giant changes to get rid of allocator, alloc_shrinker and generally allow
 * clients to pass in buffers for generated code.  Add a version of i_getErodeAsmSize
 *  for the PC.  Make i_genErodeAsm call it.  Get rid of the generated_code and
 * alloc_shrinker args in all of the private subroutines and macros since we
 * no longer free the buffer on error.
 *
 * Revision 1.19  1994/02/04  20:20:30  dferrell
 * pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v
 *
 * Changed memory.prv to memory.pub
 *
 * Revision 1.18  1993/12/02  19:25:56  ddavies
 * pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.17  1993/11/05  01:52:33  ddavies
 * pcgenero.c -> /product/ipcore/ipshared/src/RCS/pcgenero.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


