/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: sscmpgry.s
 *
 * This file contains assembly code to compare scaled versions of
 * two 8 bit gray lines, producing a binary result.  A bit in the binary
 * result is set if image1Pixel * mult1 > image2Pixel * mult2.
 *
 * This routine is called from i_compareGrayImages in i_cmpgry.c.  It
 * performs the same function as ip_compareGrayLineScaleP1 and
 * ip_compareGrayLineScaleP2 in micmpgry.c, but has a very different
 * structure.
 *
 * The basic idea here is to read the two source bytes, run one of them
 * (the byte from image 1 in the ip_compareGrayLineScaleP1, the byte from
 * image 2 in ip_compareGrayLineScaleP2) through a table that computes
 * either mult2/mult1 * pixel (...ScaleP2) or mult1/mult2 * pixel (...ScaleP1)
 * and compare the results.  The carry bit set by the comparison is
 * left shifted into a result word by adding the result word to itself
 * with carry.
 *
 * Since the result word is being shifted left, we process the line left to
 * right.  We explicitly write out 8 copies of this sequence to avoid
 * incrementing pointers, decrementing counters and testing results too
 * often.  We execute the 8x loop as four times in a row, writing out
 * a result word at the end.
 *
 * I started writing out 32 copies of the code that produces a bit (instead
 * of 8, but that seemed excessive.
 * When there are less than 32 pixels left in a line, we fall back
 * into a normal loop that does one pixel at a time.  When it completes,
 * it shifts the result word left by (32 - (number of output bits computed))
 * before writing it to the output.
 *
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: sscmpgry.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	.align	4

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPState	%i0		/* current addr in source image 1 */
#define	rSrcAddr1	%i1		/* current addr in source image 1 */
#define rSrcAddr2	%i2		/* current addr in source image 2 */
#define rDstAddr	%i3		/* current addr in dest image */
#define rScaleTable	%i4		/* base address of scaling table */
#define rWidth		%i5		/* width of dest block in 32-bit wds */

#define	rInput1		%l1		/* input byte from image 1 */
#define	rInput2		%l2		/* input byte from image 2 */
#define rOut		%l3		/* output word, contains binary data */
#define	rInputAux	%l4		/* used to avoid interlock */

#define rCount		%l5		/* loop counter for code that fills
					 * a complete output word
					 */

#define	rRetAddr	%o7		/* defined by SPARC's call inst */

/* Boilerplate constants */
#define IPStateOffset	0x44
#define	srcAddr1Offset	0x48
#define	srcAddr2Offset	0x4c
#define	dstAddrOffset	0x50
#define	scaleTabOffset	0x54
#define	widthOffset	0x58

/* magical C Frame constant (no local variables) */
#define	LF26		-64

LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! _ip_compareGrayLineScaleP1 Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_compareGrayLineScaleP1)
MK_GLOBAL_NAME(ip_compareGrayLineScaleP1):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState,    [%fp+IPStateOffset]
	st	rSrcAddr1,   [%fp+srcAddr1Offset]
	st	rSrcAddr2,   [%fp+srcAddr2Offset]
	st	rDstAddr,    [%fp+dstAddrOffset]
	st	rScaleTable, [%fp+scaleTabOffset]
	st	rWidth,	     [%fp+widthOffset]

! If there are more than 32 pixels left in the line, process them in the
! 32-pixel-at-a-time section.  Otherwise, go to the 1-pixel-at-a-time loop.
	cmp	rWidth, 32
	blu	cmpGLScaleP1BitLoopEntry
	ldub	[rSrcAddr1], rInput1			! fill delay slot

! START OF LOOP OVER THE ROW

cmpGLScaleP1RowLoop:
	mov	32, rCount				! number of pix/word

! This block of code fills an output byte.
cmpGLScaleP1WordLoop:
	ldub	[rSrcAddr2], rInput2
	ldub	[rScaleTable + rInput1], rInputAux	! scale image1 byte
	ldub	[rSrcAddr1 + 1], rInput1		! fill interlock
	cmp	rInput2, rInputAux			! compute bit 0

! compute bit 1
	ldub	[rSrcAddr2 + 1], rInput2
	ldub	[rScaleTable + rInput1], rInput1	! scale image1 byte
	addx	rOut, rOut, rOut			! insert bit 0
	cmp	rInput2, rInput1			! compute bit 1

! compute bit 2
	ldub	[rSrcAddr1 + 2], rInput1
	ldub	[rSrcAddr2 + 2], rInput2
	ldub	[rScaleTable + rInput1], rInput1	! scale image1 byte
	addx	rOut, rOut, rOut			! insert bit 1
	cmp	rInput2, rInput1			! compute bit 2

! compute bit 3
	ldub	[rSrcAddr1 + 3], rInput1
	ldub	[rSrcAddr2 + 3], rInput2
	ldub	[rScaleTable + rInput1], rInput1	! scale image1 byte
	addx	rOut, rOut, rOut			! insert bit 2
	cmp	rInput2, rInput1			! compute bit 3

! compute bit 4
	ldub	[rSrcAddr1 + 4], rInput1
	ldub	[rSrcAddr2 + 4], rInput2
	ldub	[rScaleTable + rInput1], rInput1	! scale image1 byte
	addx	rOut, rOut, rOut			! insert bit 3
	cmp	rInput2, rInput1			! compute bit 4

! compute bit 5
	ldub	[rSrcAddr1 + 5], rInput1
	ldub	[rSrcAddr2 + 5], rInput2
	ldub	[rScaleTable + rInput1], rInput1	! scale image1 byte
	addx	rOut, rOut, rOut			! insert bit 4
	cmp	rInput2, rInput1			! compute bit 5

! compute bit 6
	ldub	[rSrcAddr1 + 6], rInput1
	ldub	[rSrcAddr2 + 6], rInput2
	ldub	[rScaleTable + rInput1], rInput1	! scale image1 byte
	addx	rOut, rOut, rOut			! insert bit 5
	cmp	rInput2, rInput1			! compute bit 6

! compute bit 7
	ldub	[rSrcAddr1 + 7], rInput1
	ldub	[rSrcAddr2 + 7], rInput2
	ldub	[rScaleTable + rInput1], rInput1	! scale image1 byte
	addx	rOut, rOut, rOut			! insert bit 6
	cmp	rInput2, rInput1			! compute bit 7
	addx	rOut, rOut, rOut			! insert bit 7

	add	rSrcAddr1, 8, rSrcAddr1			! increment src addr
	add	rSrcAddr2, 8, rSrcAddr2
	subcc	rCount, 8, rCount			! done with word yet?
	bnz	cmpGLScaleP1WordLoop
	ldub	[rSrcAddr1], rInput1			! fill delay slot

! At this point, rOut is full, write it to memory and return to the top
! of the loop.
	st	rOut, [rDstAddr]
	sub	rWidth, 32, rWidth			! we've done 32 more pix
	cmp	rWidth, 32				! are there 32 more?
	bgeu	cmpGLScaleP1RowLoop			! yes, process them
	add	rDstAddr, 4, rDstAddr			! point to next word

! This loop is used when we have less than 32 pixels to compute.  We remember
! the number of pixels remaining so we know how much to shift the result to
! align it.
cmpGLScaleP1BitLoopEntry:
	orcc	rWidth, 0, rCount			! are we done now?
	bz	cmpGLScaleP1Fin
	ldub	[rSrcAddr2], rInput2

cmpGLScaleP1BitLoop:
	ldub	[rScaleTable + rInput1], rInput1	! scale image1 byte
	add	rSrcAddr1, 1, rSrcAddr1			! increment src addr
	cmp	rInput2, rInput1			! compute next bit
	addx	rOut, rOut, rOut			! insert it
	add	rSrcAddr2, 1, rSrcAddr2
	subcc	rCount, 1, rCount			! done with word yet?
	ldub	[rSrcAddr1], rInput1
	bnz	cmpGLScaleP1BitLoop
	ldub	[rSrcAddr2], rInput2			! fill delay slot

! At this point, rOut contains 32-rWidth invalid bits, so shift the
! valid bits 32-rWidth places to the right
	sub	rWidth, 32, rWidth
	sub	%g0, rWidth, rWidth			! make the result pos
	sll	rOut, rWidth, rOut
	st	rOut, [rDstAddr]

! We're done
cmpGLScaleP1Fin:
	mov	0,%i0
	ret
	restore


! -----------------------------------------------------------------------------
! _ip_compareGrayLineScaleP2 Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_compareGrayLineScaleP2)
MK_GLOBAL_NAME(ip_compareGrayLineScaleP2):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState,    [%fp+IPStateOffset]
	st	rSrcAddr1,   [%fp+srcAddr1Offset]
	st	rSrcAddr2,   [%fp+srcAddr2Offset]
	st	rDstAddr,    [%fp+dstAddrOffset]
	st	rScaleTable, [%fp+scaleTabOffset]
	st	rWidth,	     [%fp+widthOffset]

! If there are more than 32 pixels left in the line, process them in the
! 32-pixel-at-a-time section.  Otherwise, go to the 1-pixel-at-a-time loop.
	cmp	rWidth, 32
	blu	cmpGLScaleP2BitLoopEntry
	ldub	[rSrcAddr2], rInput2			! fill delay slot

! START OF LOOP OVER THE ROWS

cmpGLScaleP2RowLoop:
	mov	32, rCount				! number of pix/word

! This block of code fills an output byte.
cmpGLScaleP2WordLoop:
	ldub	[rSrcAddr1], rInput1
	ldub	[rScaleTable + rInput2], rInputAux	! scale image2 byte
	ldub	[rSrcAddr2 + 1], rInput2		! fill interlock
	cmp	rInputAux, rInput1			! compute bit 0

! compute bit 1
	ldub	[rSrcAddr1 + 1], rInput1
	ldub	[rScaleTable + rInput2], rInput2	! scale image2 byte
	addx	rOut, rOut, rOut			! insert bit 0
	cmp	rInput2, rInput1			! compute bit 1

! compute bit 2
	ldub	[rSrcAddr2 + 2], rInput2
	ldub	[rSrcAddr1 + 2], rInput1
	ldub	[rScaleTable + rInput2], rInput2	! scale image2 byte
	addx	rOut, rOut, rOut			! insert bit 1
	cmp	rInput2, rInput1			! compute bit 2

! compute bit 3
	ldub	[rSrcAddr2 + 3], rInput2
	ldub	[rSrcAddr1 + 3], rInput1
	ldub	[rScaleTable + rInput2], rInput2	! scale image2 byte
	addx	rOut, rOut, rOut			! insert bit 2
	cmp	rInput2, rInput1			! compute bit 3

! compute bit 4
	ldub	[rSrcAddr2 + 4], rInput2
	ldub	[rSrcAddr1 + 4], rInput1
	ldub	[rScaleTable + rInput2], rInput2	! scale image2 byte
	addx	rOut, rOut, rOut			! insert bit 3
	cmp	rInput2, rInput1			! compute bit 4

! compute bit 5
	ldub	[rSrcAddr2 + 5], rInput2
	ldub	[rSrcAddr1 + 5], rInput1
	ldub	[rScaleTable + rInput2], rInput2	! scale image2 byte
	addx	rOut, rOut, rOut			! insert bit 4
	cmp	rInput2, rInput1			! compute bit 5

! compute bit 6
	ldub	[rSrcAddr2 + 6], rInput2
	ldub	[rSrcAddr1 + 6], rInput1
	ldub	[rScaleTable + rInput2], rInput2	! scale image2 byte
	addx	rOut, rOut, rOut			! insert bit 5
	cmp	rInput2, rInput1			! compute bit 6

! compute bit 7
	ldub	[rSrcAddr2 + 7], rInput2
	ldub	[rSrcAddr1 + 7], rInput1
	ldub	[rScaleTable + rInput2], rInput2	! scale image2 byte
	addx	rOut, rOut, rOut			! insert bit 6
	cmp	rInput2, rInput1			! compute bit 7
	addx	rOut, rOut, rOut			! insert bit 7

	add	rSrcAddr1, 8, rSrcAddr1			! increment src addr
	subcc	rCount, 8, rCount			! done with word yet?
	add	rSrcAddr2, 8, rSrcAddr2
	bnz	cmpGLScaleP2WordLoop
	ldub	[rSrcAddr2], rInput2			! fill delay slot

! At this point, rOut is full, write it to memory and return to the top
! of the loop.
	st	rOut, [rDstAddr]
	sub	rWidth, 32, rWidth			! we've done 32 more pix
	cmp	rWidth, 32				! are there 32 more?
	bgeu	cmpGLScaleP2RowLoop			! yes, process them
	add	rDstAddr, 4, rDstAddr			! point to next word

! This loop is used when we have less than 32 pixels to compute.  We remember
! the number of pixels remaining so we know how much to shift the result to
! align it.
cmpGLScaleP2BitLoopEntry:
	orcc	rWidth, 0, rCount			! are we done now?
	bz	cmpGLScaleP2Fin
	ldub	[rSrcAddr1], rInput1

cmpGLScaleP2BitLoop:
	ldub	[rScaleTable + rInput2], rInput2	! scale image2 byte
	add	rSrcAddr1, 1, rSrcAddr1			! increment src addr
	cmp	rInput2, rInput1			! compute next bit
	addx	rOut, rOut, rOut			! insert it
	add	rSrcAddr2, 1, rSrcAddr2
	subcc	rCount, 1, rCount			! done with word yet?
	ldub	[rSrcAddr2], rInput2
	bnz	cmpGLScaleP2BitLoop
	ldub	[rSrcAddr1], rInput1			! fill delay slot

! At this point, rOut contains 32-rWidth invalid bits, so shift the
! valid bits 32-rWidth places to the right
	sub	rWidth, 32, rWidth
	neg	rWidth					! make the result pos
	sll	rOut, rWidth, rOut
	st	rOut, [rDstAddr]

! We're done
cmpGLScaleP2Fin:
	mov	0,%i0
	ret
	restore


/*

$Log: sscmpgry.s,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.4  1998/01/13 23:57:31  ddavies
ddavies on Tue Jan 13 15:57:30 PST 1998

sscmpgry.s -> /product/ipcore/ipshared/src/RCS/sscmpgry.s,v

CommonCheck: Add IPState parameter to almost all private routines.

# Revision 1.3  1993/12/04  04:59:29  ddavies
# sscmpgry.s -> /product/ipcore/ipshared/src/RCS/sscmpgry.s,v
#
# Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
#
# Revision 1.2  1993/11/05  02:23:41  ddavies
# sscmpgry.s -> /product/ipcore/ipshared/src/RCS/sscmpgry.s,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

*/

