/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: ssdepth.s
 *
 * This file contains assembly code to convert 8-bit gray scale pixrects
 * to 1 bit binary pixrects.  It is loosely based on the 8:1 reduction code in
 * ssredn.s
 *
 * As of August 24, 1992, there is also code to convert 2-bit pixrects to
 * binary pixrects.
 *
 * As of October 1992, there is also code to convert 4-bit pixrects to
 * binary pixrects.
 *
 * As of January 1994, the modules have been rearranged so each assembly
 * routine processes only 1 image line.  The routines are called
 * repeatedly to process the entire image.
 *
 * The inner loop of each reduction routine reads a source word,
 * filters it according to the rules for the current threshold, selects
 * the relevant bits of output and shifts these bits into the current
 * output word.
 *
 * In a change from the first version of the code, we
 * don't immediately shift the 4 source bits into position.  Instead, we
 * collect 8 sets of 4 bits into a 32 bit word, but leave the bits from
 * each group separated (by 8) instead of consolidating them at each
 * step.  Using the bit numbers from the final (desired) positions, the
 * accumulated word contains bits ordered as:
 * Byte 0 (MS byte): 31, 27, 23, 19, 15, 11, 7, 3
 * Byte 1	   : 30, 26, 22, 18, 14, 10, 6, 2
 * Byte 2	   : 29, 25, 21, 17, 13,  9, 5, 1
 * Byte 3 (LS byte): 28, 24, 20, 16, 12,  8, 4, 0
 *
 * There are four permutation tables, each containing 256 32-bit words.
 * TabBinarizeByte0 is indexed with the upper 8 bits of the source
 * word shown above and rearranges the bits into their proper positions.
 * TabBinarizeByte1, TabBinarizeByte2 and TabBinarizeByte3 do the same for
 * the other bytes.  These tables are created by C code.
 *
 * The code that produces 2-bit output or starts with 4-bit input does
 * roughly the same thing.  The tables used to rearrange the output bits
 * are different and we produce a different number of output bits from each
 * input word.
 *
 * The inner loops fill a register pair with outputs.  The register pair is
 * then stored using a double write.  Words are processed from right to
 * left across a line so that the destination address displacement can also
 * act as a line counter.
 * However, within the computation of a single destination word, source
 * words are read from left to right.  This allows a few small code
 * optimizations. The loop structure is:
 *
 * OddAddrEntryPoint:
 *	process source words for rightmost dest word at odd address
 *	if source address is off right end of image, fix up result and quit
 *	jump to EvenAddrInMainLoop
 *	PIXELS IN THIS LOOP ARE PROCESSED FROM LEFT TO RIGHT (so we can find
 *	the right edge).
 * EvenAddrEntryPoint:
 *	process source words for rightmost dest word at even address
 *	if source address is off right end of image, fix up result and quit
 *	write dest word at even address
 *	PIXELS IN THIS LOOP ARE PROCESSED FROM LEFT TO RIGHT (so we can find
 *	the right edge).
 * MainLoop:
 *	compute dest word at odd address in 4-bit chunks
 *	THIS LOOP USES UNROLLED CODE.  It is executed more often and
 *	doesn't have to check for the end conditions as is necessary
 *	in the entry point loops.
 *	if now at beginning of dest line, write out odd dest word and quit
 *	PIXELS IN THIS LOOP ARE PROCESSED FROM RIGHT TO LEFT (to minimize
 *	shifting).
 * EvenAddrInMainLoop:
 *	compute dest word at even address in 4-bit chunks
 *	THIS LOOP ALSO USES UNROLLED CODE.
 *	write out even/odd register pair using double write
 *	if not yet at beginning of dest line, go back to MainLoop
 *	PIXELS IN THIS LOOP ARE PROCESSED FROM RIGHT TO LEFT (to minimize
 *	shifting).
 *
 * The OddAddrEntryPoint and EvenAddrEntryPoint blocks are used to produce the
 * rightmost destination word.  This word may require data from beyond the
 * right edge of the source image due to the way the image widths have been
 * rounded.  There is special code in these blocks to detect this case.  The
 * special code is unneeded in the main loop so is left out for speed.
 * Note that the EvenAddrEntryPoint also writes one word at an even dest
 * word address.  That word would otherwise have been skipped by the main loop.
 *
 * ADDITION: The same strategy was used in toTwoBitsPerPixel.  This routine
 * converts an 8 bit/pixel image to a 2 bit/pixel image by simply selecting
 * the top two bits of each pixel.  The main changes are in the formats
 * of the accumulated data, hence in the permutation tables.  The data
 * formats are:
 *
 *	Byte0: 31,30, 23,22, 15,14, 7,6
 *	Byte1: 29,28, 21,20, 13,12, 5,4
 *	Byte2: 27,26, 19,18, 11,10, 3,2
 *	Byte3: 25,24, 17,16,  9, 8, 1,0
 *
 * BUG FIX (11/15/93): When taking the EvenAddrEntryPoint, we normally go
 * to code that unscrambles the even word in rOut0, then does a *double*
 * write.  Whatever trash is in rOut1 is written into the frame and pad.
 * Rather than have a completely separate copy of the unscrambling code
 * or testing to see if we're at the right edge of the image in this
 * inner loop, we'll just fill rOut1 with 0's at the beginning of each
 * row.  If we take the even branch, we'll be writing 0's to the frame.
 * Yeah, this doesn't work if there's no frame, but lots of things don't
 * work if there's no frame.
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssdepth.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	.align	4

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPStateIn	%i0		/* ptr to IPState */
#define	rSrcAddrIn	%i1		/* current addr in source image */
#define rDstAddrIn	%i2		/* current addr in dest image */
#define rDstOffsetIn	%i3		/* current horizontal dest offset */
#define rSrcOffsetIn	%i4		/* current horizontal source offset */
#define rSrcLimitIn	%i5		/* rightmost legal source offset */

#define	rSrcAddr	%i0		/* current addr in source image */
#define rDstAddr	%i1		/* current addr in dest image */
#define rDstOffset	%i2		/* current horizontal dest offset */
#define rSrcOffset	%i3		/* current horizontal source offset */
#define rSrcLimit	%i4		/* rightmost legal source offset */
#define rTabByte0	%i5		/* points to tabBinarizeByte0 */

#define rTabByte1	%l0		/* points to tabBinarizeByte1 */
#define rTabByte2	%l1		/* points to tabBinarizeByte2 */
#define rTabByte3	%l2		/* points to tabBinarizeByte3 */

#define	rThresh		%l3		/* HighPass's only threshold */
#define	rLower		%l3		/* BandPass's lower threshold */
#define	rUpper		%l4		/* BandPass's Upper-Lower threshold */

#define rScr0		%l5		/* scratch registers */
#define rScr1		%l6

#define	rInput		%l7		/* input word */

#define	rOut0		%o0		/* output register pair */
#define rOut1		%o1

#define	rCount		%o2		/* for loop iteration */
#define	rMSBitMask	%o3		/* mask selecting MS bit of each byte */
#define	rLSBitMask	%o4		/* mask selecting LS bit of bytes 1&2 */

#define	rRetAddr	%o7		/* defined by SPARC's call inst */

/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	srcAddrOffset	0x48
#define	dstAddrOffset	0x4c
#define	dstOffsetOffset	0x50
#define	srcOffsetOffset	0x54
#define	srcLimitOffset	0x58
#define	tabByte0Offset	0x5c
#define tabByte1Offset	0x60
#define	tabByte2Offset	0x64
#define	tabByte3Offset	0x68

#define	thresh1Offset	0x6c		/* thresh for toBinaryThr, lower thresh
					   for toBinaryBP */
#define	thresh2Offset	0x70		/* upper thresh for toBinaryBP */

#define	cMSBitMask	0x80808080	/* selects 4 sign bits of a word */
#define	cLSBitMask	0x01010101	/* used to clear LS bits to break
					 * the carry chain.
					 */
#define	cMS2BitsMask	0xC0C0C0C0	/* selects upper 2 bits of each byte
					 * for toTwoBitsPerPixel */
#define	cMS16BitsMask	0xAAAAAAAA	/* selects upper bit of each 2-bit
					 * pixel in a word */
#define cMS8BitsMask	0x88888888	/* selects upper bit of each 4-bit
					 * pixel in a word */

/* magical C Frame constant (no local variables) */
#define	LF26		-64

LL0:					! does this really do anything?


! -----------------------------------------------------------------------------
! _ip_toBinary128Line Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_toBinary128Line)
MK_GLOBAL_NAME(ip_toBinary128Line):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPStateIn,   [%fp+ipStateOffset]
	st	rSrcAddrIn,   [%fp+srcAddrOffset]
	st	rDstAddrIn,   [%fp+dstAddrOffset]
	st	rDstOffsetIn, [%fp+dstOffsetOffset]
	st	rSrcOffsetIn, [%fp+srcOffsetOffset]
	st	rSrcLimitIn,  [%fp+srcLimitOffset]
	ld	[%fp+srcAddrOffset], rSrcAddr
	ld	[%fp+dstAddrOffset], rDstAddr
	ld	[%fp+dstOffsetOffset], rDstOffset
	ld	[%fp+srcOffsetOffset], rSrcOffset
	ld	[%fp+srcLimitOffset], rSrcLimit

	ld	[%fp+tabByte0Offset], rTabByte0
	ld	[%fp+tabByte1Offset], rTabByte1
	ld	[%fp+tabByte2Offset], rTabByte2
	ld	[%fp+tabByte3Offset], rTabByte3

! The MS bit mask register is set up in delay slots below.
! Increment rDstOffset by 4 so we decrement it before it is used.
! This makes the loop nicer.
	add	rDstOffset, 4, rDstOffset

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  Both cases
! have particular code for starting the row.  Note that rDstOffset is
! decremented before use, changing the sense of the test.

	add	rDstAddr, rDstOffset, rScr0	! calculate dest address
	andcc	rScr0, 0x4, %g0			! is first addr even?
	bnz	toBinary128EvenEntry		! yes, process rightmost even wd
	sethi	%hi(cMSBitMask), rMSBitMask

! calculate the rightmost dest word, which has an odd word address.
toBinary128OddEntry:
	call	toBinary128Comp			! reduce first group of 4 bytes
	or	rMSBitMask, %lo(cMSBitMask), rMSBitMask
! return with 1st 4 bits
	mov	rScr0, rOut1			! get first 4 bits
	mov	7, rCount			! set count register

toBinary128OddEntryLoop:
	subcc	rSrcOffset, rSrcLimit, %g0	! at end of src line now?
	bpos	toBinary128OddEntryExit		! yes, quit now and
	or	rOut1, rScr0, rOut1		! combine next 4 bits.
	call	toBinary128Comp			! get nth 4 bits
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next word to the right
! return with nth 4 bits, rCount decremented in toBinary128Comp
	bnz	toBinary128OddEntryLoop
	sll	rOut1, 1, rOut1			! align previous bits
	or	rOut1, rScr0, rOut1		! combine last 4 bits.

toBinary128OddEntryExit:
	sll	rOut1, rCount, rOut1		! yes, align output bits
	sll	rCount, 2, rCount		! compute remaining byte offset
	add	rSrcOffset, rCount, rSrcOffset	! and correct addr offset.
						! rSrcOffset should point to
						! the last word of a block
						! so we can subtract blockLen
						! from it to get to the right
						! end of the block to the left.
	ba	toBinary128UnscrambleOddWd	! go unpermute this word
	add	rSrcAddr, rSrcOffset, rSrcOffset ! and set up rScrOffset

!------------------------------------------------------------------
! ROW ENTRY POINT USED WHEN RIGHTMOST DEST WORD HAS AN EVEN ADDRESS

! calculate the rightmost dest word, which has an even word address.
toBinary128EvenEntry:
	call	toBinary128Comp			! reduce first group of 4 bytes
	or	rMSBitMask, %lo(cMSBitMask), rMSBitMask
! return with 1st 4 bits
	mov	rScr0, rOut0			! get first 4 bits
	mov	7, rCount			! set count register

toBinary128EvenEntryLoop:
	subcc	rSrcOffset, rSrcLimit, %g0	! at end of src line now?
	bpos	toBinary128CheckEvenEntryWrite	! yes, quit now and
	or	rOut0, rScr0, rOut0		! combine next 4 bits.
	call	toBinary128Comp			! get nth 4 bits
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next word to the right
! return with nth 4 bits, rCount decremented in toBinary128Comp
	bnz	toBinary128EvenEntryLoop
	sll	rOut0, 1, rOut0			! align previous bits
	or	rOut0, rScr0, rOut0		! combine last 4 bits.

toBinary128CheckEvenEntryWrite:
	mov	0, rOut1			! prepare for double write
						! after rOut0 is unscrambled.
						! This 0 is written over the
						! already 0 right frame.
	sll	rOut0, rCount, rOut0		! output bits aligned
	sll	rCount, 2, rCount		! compute remaining byte offset
	add	rSrcOffset, rCount, rSrcOffset	! gen correct addr offset
	ba	toBinary128UnscrambleEvenWd
	add	rSrcAddr, rSrcOffset, rSrcOffset ! set up rScrOffset



! START OF LOOP OVER REST OF 1 ROW

toBinary128RowLoop:
	ld	[rSrcOffset - 0x20], rInput	! get 1st 4 bits
	sub	rSrcOffset, 0x20, rSrcOffset	! pt to next wd group to left
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 4 bits
	srl	rScr0, 1, rOut1
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x8], rInput	! get 3rd 4 bits
	or	rOut1, rScr0, rOut1
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0xC], rInput	! get 4th 4 bits
	or	rOut1, rScr0, rOut1
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x10], rInput	! get 5th 4 bits
	or	rOut1, rScr0, rOut1
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x14], rInput	! get 6th 4 bits
	or	rOut1, rScr0, rOut1
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x18], rInput	! get 7th 4 bits
	or	rOut1, rScr0, rOut1
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x1c], rInput	! get 8th 4 bits
	or	rOut1, rScr0, rOut1
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1
	or	rOut1, rScr0, rOut1

! At this point, rOut1 contains the scrambled bits.  We un-scramble them
! using
! rOut1 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
toBinary128UnscrambleOddWd:
	sll	rOut1, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut1, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut1, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut1, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut1		! have unscrambled bits.
	bz,a	toBinary128Fin			! yes, done with this row
	st	rOut1, [rDstAddr+rDstOffset]	! store the odd word if this is
						! left end of row

! **** CALCULATE EVEN DESTINATION WORD ****

toBinary128EvenDest:
	ld	[rSrcOffset - 0x20], rInput	! get 1st 4 bits
	sub	rSrcOffset, 0x20, rSrcOffset	! pt to next 8 wd group to left
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 4 bits
	srl	rScr0, 1, rOut0
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x8], rInput	! get 3rd 4 bits
	or	rOut0, rScr0, rOut0
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0xC], rInput	! get 4th 4 bits
	or	rOut0, rScr0, rOut0
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x10], rInput	! get 5th 4 bits
	or	rOut0, rScr0, rOut0
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x14], rInput	! get 6th 4 bits
	or	rOut0, rScr0, rOut0
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x18], rInput	! get 7th 4 bits
	or	rOut0, rScr0, rOut0
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x1c], rInput	! get 8th 4 bits
	or	rOut0, rScr0, rOut0
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0
	or	rOut0, rScr0, rOut0

! At this point, rOut0 contains the scrambled bits.  We un-scramble them
! using
! rOut0 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
toBinary128UnscrambleEvenWd:
	sll	rOut0, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut0, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut0, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut0, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut0		! have unscrambled bits.
	bnz	toBinary128RowLoop		! yes, prepare for next row
	std	rOut0, [rDstAddr+rDstOffset]	! store even/odd pair

! END OF LOOP ACROSS 1 ROW

! We've put out the right number of words, so quit
toBinary128Fin:
	ret
	restore


!------------------------------------------------------------
! toBinary128Comp
!------------------------------------------------------------

! A subroutine used by toBinary128 to reduce 4 8-bit bytes into 4 bits.
toBinary128Comp:
	ld	[rSrcAddr+rSrcOffset], rInput	! get 1st word
	subcc	rCount, 1, rCount		! dec loop count (interlock)
	and	rInput, rMSBitMask, rInput
	retl
	srl	rInput, 7, rScr0		! align at LS end of each byte

!------------------------------------------------------------
 


! -----------------------------------------------------------------------------
! _ip_toBinaryFrom2bppLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_toBinaryFrom2bppLine)
MK_GLOBAL_NAME(ip_toBinaryFrom2bppLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPStateIn,   [%fp+ipStateOffset]
	st	rSrcAddrIn,   [%fp+srcAddrOffset]
	st	rDstAddrIn,   [%fp+dstAddrOffset]
	st	rDstOffsetIn, [%fp+dstOffsetOffset]
	st	rSrcOffsetIn, [%fp+srcOffsetOffset]
	st	rSrcLimitIn,  [%fp+srcLimitOffset]
	ld	[%fp+srcAddrOffset], rSrcAddr
	ld	[%fp+dstAddrOffset], rDstAddr
	ld	[%fp+dstOffsetOffset], rDstOffset
	ld	[%fp+srcOffsetOffset], rSrcOffset
	ld	[%fp+srcLimitOffset], rSrcLimit

	ld	[%fp+tabByte0Offset], rTabByte0
	ld	[%fp+tabByte1Offset], rTabByte1
	ld	[%fp+tabByte2Offset], rTabByte2
	ld	[%fp+tabByte3Offset], rTabByte3

! The MS bit mask register is set up in delay slots below.
! Increment rDstOffset by 4 so we decrement it before it is used.
! This makes the loop nicer.
	add	rDstOffset, 4, rDstOffset

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  If it's
! odd, we start right in.  If it's even, we need to do a special cycle that
! does a single (as opposed to double) write for the even word calculated.
! Note the sense of the test.  Note rDstOffset will be decremented 
! before its used.  Therefore, the sense of the test on (rDstAddr+rDstOffset)
! is reversed.

	add	rDstAddr, rDstOffset, rScr0	! calculate dest address
	andcc	rScr0, 0x4, %g0			! is (first addr+1) odd?
	bnz	tBF2EvenEntry			! yes, first addr is even
	sethi	%hi(cMS16BitsMask), rMSBitMask

! ODD WORD ENTRY

! calculate the rightmost dest word, which has an odd word address.
tBF2OddEntry:
	or	rMSBitMask, %lo(cMS16BitsMask), rMSBitMask
	ld	[rSrcAddr+rSrcOffset], rOut1	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?
	and	rOut1, rMSBitMask, rOut1	! select MS bits

! have 1st 16 bits
	bneg	tBF2OddEntryWd1			! no, keep going
	add	rSrcOffset, 0x4, rSrcOffset	! fix up offset

	ba	tBF2UnscrambleOddWord		! yes, finish up
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop

tBF2OddEntryWd1:
	ld	[rSrcAddr+rSrcOffset], rInput	! get the word
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop
	and	rInput, rMSBitMask, rInput	! select MS bits
	srl	rInput, 1, rInput		! align at LS end of each pixel
! have last 16 bits
	ba	tBF2UnscrambleOddWord		! go unscramble the bits
	or	rOut1, rInput, rOut1		! combine with other 16 bits.

! EVEN WORD ENTRY

! calculate the rightmost dest word, which has an even word address.
tBF2EvenEntry:
	mov	0, rOut1			! prepare for double write
						! after rOut0 is unscrambled.
						! This 0 is written over the
						! already 0 right frame.
	or	rMSBitMask, %lo(cMS16BitsMask), rMSBitMask
	ld	[rSrcAddr+rSrcOffset], rOut0	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?
	and	rOut0, rMSBitMask, rOut0	! select MS bits

! now have 1st 16 bits
	bneg	tBF2EvenEntryWd1		! no, keep going
	add	rSrcOffset, 0x4, rSrcOffset	! with proper offset

	ba	tBF2UnscrambleEvenWd
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop

tBF2EvenEntryWd1:
	ld	[rSrcAddr+rSrcOffset], rInput	! get the word
	add	rSrcAddr, rSrcOffset, rSrcOffset
	and	rInput, rMSBitMask, rInput	! select MS bits
	srl	rInput, 1, rInput		! align at LS end of each pixel
! now have last 16 bits
	ba	tBF2UnscrambleEvenWd		! yes, prepare for next row
	or	rOut0, rInput, rOut0		! get 32 bits.


! START OF LOOP OVER 1 ROW
! FORM ODD DEST WORD

! The first word computed has an odd address at this point
tBF2RowLoop:
	ld	[rSrcOffset - 0x8], rInput	! get rightmost 32 bits of
						! next 64 bits to the left
	sub	rSrcOffset, 0x8, rSrcOffset	! pt to right end of next 2
						! wd group to left
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 4 bits
	srl	rScr0, 1, rOut1
	and	rInput, rMSBitMask, rScr0
	or	rOut1, rScr0, rOut1

! At this point, rOut1 contains the scrambled bits.  We un-scramble them
! using
! rOut1 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
tBF2UnscrambleOddWord:
	sll	rOut1, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut1, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut1, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut1, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut1		! have unscrambled bits.
	bz,a	tBF2Fin				! yes, done with this row
	st	rOut1, [rDstAddr+rDstOffset]	! store the odd word if this is
						! left end of row

! **** CALCULATE EVEN DESTINATION WORD ****

tBF2EvenDest:
	ld	[rSrcOffset - 0x8], rInput	! get 1st 16 bits
	sub	rSrcOffset, 0x8, rSrcOffset	! pt to rightmost word of
						! next 2 wd group to left
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get other 16 bits
	srl	rScr0, 1, rOut0
	and	rInput, rMSBitMask, rScr0
	or	rOut0, rScr0, rOut0

! At this point, rOut0 contains the scrambled bits.  We un-scramble them
! using
! rOut0 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
tBF2UnscrambleEvenWd:
	sll	rOut0, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut0, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut0, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut0, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut0		! have unscrambled bits.
	bnz	tBF2RowLoop			! yes, prepare for next row
	std	rOut0, [rDstAddr+rDstOffset]	! store even/odd pair

! We've put out the right number of words, so quit
tBF2Fin:
	ret
	restore

! -----------------------------------------------------------------------------
! _ip_toBinaryFrom4bppLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_toBinaryFrom4bppLine)
MK_GLOBAL_NAME(ip_toBinaryFrom4bppLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPStateIn,   [%fp+ipStateOffset]
	st	rSrcAddrIn,   [%fp+srcAddrOffset]
	st	rDstAddrIn,   [%fp+dstAddrOffset]
	st	rDstOffsetIn, [%fp+dstOffsetOffset]
	st	rSrcOffsetIn, [%fp+srcOffsetOffset]
	st	rSrcLimitIn,  [%fp+srcLimitOffset]
	ld	[%fp+srcAddrOffset], rSrcAddr
	ld	[%fp+dstAddrOffset], rDstAddr
	ld	[%fp+dstOffsetOffset], rDstOffset
	ld	[%fp+srcOffsetOffset], rSrcOffset
	ld	[%fp+srcLimitOffset], rSrcLimit

	ld	[%fp+tabByte0Offset], rTabByte0
	ld	[%fp+tabByte1Offset], rTabByte1
	ld	[%fp+tabByte2Offset], rTabByte2
	ld	[%fp+tabByte3Offset], rTabByte3

! The MS bit mask register is set up in delay slots below.
! Increment rDstOffset by 4 so we decrement it before it is used.
! This makes the loop nicer.
	add	rDstOffset, 4, rDstOffset

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  If it's
! odd, we start right in.  If it's even, we need to do a special cycle that
! does a single (as opposed to double) write for the even word calculated.
! Note the sense of the test.  Note rDstOffset will be decremented 
! before its used.  Therefore, the sense of the test on (rDstAddr+rDstOffset)
! is reversed.

	add	rDstAddr, rDstOffset, rScr0	! calculate dest address
	andcc	rScr0, 0x4, %g0			! is (first addr+1) odd?
	bnz	tBF4EvenEntry			! yes, first addr is even
	sethi	%hi(cMS8BitsMask), rMSBitMask

! ODD WORD ENTRY

! calculate the rightmost dest word, which has an odd word address.
tBF4OddEntry:
	or	rMSBitMask, %lo(cMS8BitsMask), rMSBitMask
	ld	[rSrcAddr+rSrcOffset], rOut1	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?

! have 1st 8 bits
	bneg	tBF4OddEntryWd1			! no, get next 8 bits
	and	rOut1, rMSBitMask, rOut1	! select MS bits

	ba	tBF4UnscrambleOddWord		! yes, finish up
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop

tBF4OddEntryWd1:
	srl	rOut1, 2, rOut1			! no, align at LS end of each pixel
	add	rSrcOffset, 0x4, rSrcOffset	! fix up offset
	ld	[rSrcAddr+rSrcOffset], rInput	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?
	and	rInput, rMSBitMask, rInput	! select MS bits
	srl	rInput, 3, rInput		! align at LS end of each pixel
! have next 8 bits
	bneg	tBF4OddEntryWd2			! no get next 8 bits
	or	rOut1, rInput, rOut1		! combine with other 8
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop
	sub	rSrcOffset, 4, rSrcOffset
	ba	tBF4UnscrambleOddWord		! yes, finish up
	sll	rOut1, 2, rOut1			! put into correct place for
						! unscrambling

tBF4OddEntryWd2:
	add	rSrcOffset, 0x4, rSrcOffset	! no, fix up offset
	ld	[rSrcAddr+rSrcOffset], rInput	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?
	and	rInput, rMSBitMask, rInput	! select MS bits
	sll	rOut1, 1, rOut1
	srl	rInput, 3, rInput		! align at LS end of each pixel
! have next 8 bits
	bneg	tBF4OddEntryWd3			! no get next 8 bits
	or	rOut1, rInput, rOut1		! combine with other 8
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop
	sub	rSrcOffset, 8, rSrcOffset
	ba	tBF4UnscrambleOddWord		! yes, finish up
	sll	rOut1, 1, rOut1			! put into correct place for
						! unscrambling

tBF4OddEntryWd3:
	add	rSrcOffset, 0x4, rSrcOffset	! no, fix up offset
	ld	[rSrcAddr+rSrcOffset], rInput	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?
	and	rInput, rMSBitMask, rInput	! select MS bits
	sll	rOut1, 1, rOut1
	srl	rInput, 3, rInput		! align at LS end of each pixel
! have last 8 bits
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop
	sub	rSrcOffset, 12, rSrcOffset
	ba	tBF4UnscrambleOddWord		! go unscramble the bits
	or	rOut1, rInput, rOut1		! combine with other 16 bits.

! EVEN WORD ENTRY

! calculate the rightmost dest word, which has an even word address.
tBF4EvenEntry:
	mov	0, rOut1			! prepare for double write
						! after rOut0 is unscrambled.
						! This 0 is written over the
						! already 0 right frame.
	or	rMSBitMask, %lo(cMS8BitsMask), rMSBitMask
	ld	[rSrcAddr+rSrcOffset], rOut0	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?

! now have 1st 8 bits
	bneg	tBF4EvenEntryWd1		! no, keep going
	and	rOut0, rMSBitMask, rOut0	! select MS bits

	ba	tBF4UnscrambleEvenWord		! yes, finish up
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop

tBF4EvenEntryWd1:
	srl	rOut0, 2, rOut0			! no, align at LS end of each pixel
	add	rSrcOffset, 0x4, rSrcOffset	! fix up offset
	ld	[rSrcAddr+rSrcOffset], rInput	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?
	and	rInput, rMSBitMask, rInput	! select MS bits
	srl	rInput, 3, rInput		! align at LS end of each pixel
! have next 8 bits
	bneg	tBF4EvenEntryWd2		! no get next 8 bits
	or	rOut0, rInput, rOut0		! combine with other 8
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop
	sub	rSrcOffset, 4, rSrcOffset
	ba	tBF4UnscrambleEvenWord		! yes, finish up
	sll	rOut0, 2, rOut0			! put into correct place for
						! unscrambling

tBF4EvenEntryWd2:
	add	rSrcOffset, 0x4, rSrcOffset	! no, fix up offset
	ld	[rSrcAddr+rSrcOffset], rInput	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?
	and	rInput, rMSBitMask, rInput	! select MS bits
	sll	rOut0, 1, rOut0
	srl	rInput, 3, rInput		! align at LS end of each pixel
! have next 8 bits
	bneg	tBF4EvenEntryWd3			! no get next 8 bits
	or	rOut0, rInput, rOut0		! combine with other 8
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop
	sub	rSrcOffset, 8, rSrcOffset
	ba	tBF4UnscrambleEvenWord		! yes, finish up
	sll	rOut0, 1, rOut0			! put into correct place for
						! unscrambling

tBF4EvenEntryWd3:
	add	rSrcOffset, 0x4, rSrcOffset	! no, fix up offset
	ld	[rSrcAddr+rSrcOffset], rInput	! get the word
	subcc	rSrcOffset, rSrcLimit, %g0	! was that end of src line?
	and	rInput, rMSBitMask, rInput	! select MS bits
	sll	rOut0, 1, rOut0
	srl	rInput, 3, rInput		! align at LS end of each pixel
! have last 8 bits
	add	rSrcAddr, rSrcOffset, rSrcOffset ! fixup rSrcOffset for loop
	sub	rSrcOffset, 12, rSrcOffset
	ba	tBF4UnscrambleEvenWord		! go unscramble the bits
	or	rOut0, rInput, rOut0		! combine with other 16 bits.


! START OF LOOP OVER 1 ROW
! FORM ODD DEST WORD

! The first word computed has an odd address at this point
tBF4RowLoop:
	ld	[rSrcOffset - 0x4], rInput	! get rightmost 32 bits of
						! next 128 bits to the left
	sub	rSrcOffset, 0x10, rSrcOffset	! pt to leftmost end of the
						! group
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset + 0x8], rInput	! get other 8 bits
	srl	rScr0, 1, rOut1			! prev 8 bits in out word
	and	rInput, rMSBitMask, rScr0
	ld	[rSrcOffset + 0x4], rInput	! get other 8 bits
	or	rOut1, rScr0, rOut1		! prev 8 bits in out word
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1
	ld	[rSrcOffset + 0], rInput	! get final 8 bits
	or	rOut1, rScr0, rOut1		! prev 8 bits in out word
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1
	or	rOut1, rScr0, rOut1		! prev 8 bits in out word

! At this point, rOut1 contains the scrambled bits.  We un-scramble them
! using
! rOut1 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
tBF4UnscrambleOddWord:
	sll	rOut1, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut1, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut1, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut1, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut1		! have unscrambled bits.
	bz,a	tBF4Fin				! yes, donw with this row
	st	rOut1, [rDstAddr+rDstOffset]	! store the odd word if this is
						! left end of row

! **** CALCULATE EVEN DESTINATION WORD ****

tBF4EvenDest:
	ld	[rSrcOffset - 0x4], rInput	! get 1st 8 bits
	sub	rSrcOffset, 0x10, rSrcOffset	! pt to leftmost word of
						! next 4 wd group to left
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset + 0x8], rInput	! get other 8 bits
	srl	rScr0, 1, rOut0			! prev 8 bits in out word
	and	rInput, rMSBitMask, rScr0
	ld	[rSrcOffset + 0x4], rInput	! get other 8 bits
	or	rOut0, rScr0, rOut0		! prev 8 bits in out word
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0
	ld	[rSrcOffset + 0], rInput	! get final 8 bits
	or	rOut0, rScr0, rOut0		! prev 8 bits in out word
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0
	or	rOut0, rScr0, rOut0		! prev 8 bits in out word


! At this point, rOut0 contains the scrambled bits.  We un-scramble them
! using
! rOut0 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
tBF4UnscrambleEvenWord:
	sll	rOut0, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut0, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut0, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut0, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut0		! have unscrambled bits.
	bnz	tBF4RowLoop			! yes, prepare for next row
	std	rOut0, [rDstAddr+rDstOffset]	! store even/odd pair


! We've put out the right number of words, so quit
tBF4Fin:
	ret
	restore


!----------------------------------------------------------------------------
! toBinaryThresh
!----------------------------------------------------------------------------

! This routine is very much like toBinary128 except it compares each pixel
! against a threshold instead of simply taking the MS bit of each pixel.

! The order in which source and destination are read and written respectively
! is the same as in toBinary128.  The only difference is in generation of
! the outputs.

! We'd very much like to compute 4 thresholds in parallel given a 32 bit
! word.  To compare pixels with thresholds, we need to do four independent
! subtract operations in parallel.  How is a subtract implemented?  In
! 2's complement logic, the subtrahend is inverted and added to the minuend
! with the carry input set.  Thus, we must guarantee that there's a carry
! into the LS bit of each pixel during the subtract.

! To do this, we assume we may disregard the LS bit of both the threshold
! and the gray values.  Scanners aren't really linear to 256 levels, right?
! Here's the algorithm:

! 1. Take a new 32 bit sample.
! 2. Shift it right by 1 so the bits recording the carry out from each
!    pixel value become visible.
! 2. Set bits 31, 23, 15 and 7 (the "carry out" bits above each pixel).
! 3. Subtract a 32 bit register containing a shifted copy of the threshold in
!    each pixel position from the modified sample.  The MS bits of the
!    bytes must be 0.
! 4. Use the "carry out" bits of the results (bits 31, 23, 15, 7) as the four
!    threshold values.  These bits will be set if the original pixel
!    values were greater than the thresholds (sample-threshold did not
!    generate a borrow, hence maintained the original 1's in these positions).

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_toBinaryThreshLine)
MK_GLOBAL_NAME(ip_toBinaryThreshLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  Also load addresses of byte permutation tables.
	st	rIPStateIn,   [%fp+ipStateOffset]
	st	rSrcAddrIn,   [%fp+srcAddrOffset]
	st	rDstAddrIn,   [%fp+dstAddrOffset]
	st	rDstOffsetIn, [%fp+dstOffsetOffset]
	st	rSrcOffsetIn, [%fp+srcOffsetOffset]
	st	rSrcLimitIn,  [%fp+srcLimitOffset]
	ld	[%fp+srcAddrOffset], rSrcAddr
	ld	[%fp+dstAddrOffset], rDstAddr
	ld	[%fp+dstOffsetOffset], rDstOffset
	ld	[%fp+srcOffsetOffset], rSrcOffset
	ld	[%fp+srcLimitOffset], rSrcLimit

	ld	[%fp+tabByte0Offset], rTabByte0
	ld	[%fp+tabByte1Offset], rTabByte1
	ld	[%fp+tabByte2Offset], rTabByte2
	ld	[%fp+tabByte3Offset], rTabByte3
	ld	[%fp+thresh1Offset], rThresh

! The MS bit mask register is set up in delay slots below.
! Increment rDstOffset by 4 so we decrement it before it is used.
! This makes the loop nicer.
	add	rDstOffset, 4, rDstOffset

! Comments about the two loop entry points are the same (since the code
! is the same).  One is used if the rightmost dest address is
! even, the other if it's odd.  Note that rDstOffset is
! decremented before use, changing the sense of the test.

	add	rDstAddr, rDstOffset, rScr0	! calculate dest address
	andcc	rScr0, 0x4, %g0			! is first addr even?
	bnz	toBinaryThrEvenEntry		! yes, process rightmost even wd
	sethi	%hi(cMSBitMask), rMSBitMask

! calculate the rightmost dest word, which has an odd word address.
toBinaryThrOddEntry:
	call	toBinaryThrComp			! reduce first group of 4 8x8's
	or	rMSBitMask, %lo(cMSBitMask), rMSBitMask
! return with 1st 4 bits
	mov	rScr0, rOut1			! get first 4 bits
	or	%g0, 7, rCount			! set count register

toBinaryThrOddEntryLoop:
	subcc	rSrcOffset, rSrcLimit, %g0	! at end of src line now?
	bpos	toBinaryThrOddEntryExit		! yes, quit now and
	or	rOut1, rScr0, rOut1		! combine next 4 bits.
	call	toBinaryThrComp			! get nth 4 bits
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next word to the right
! return with nth 4 bits, rCount decremented in toBinary128Comp
	bnz	toBinaryThrOddEntryLoop
	sll	rOut1, 1, rOut1			! align previous bits
	or	rOut1, rScr0, rOut1		! combine last 4 bits.

toBinaryThrOddEntryExit:
	sll	rOut1, rCount, rOut1		! yes, align output bits
	sll	rCount, 2, rCount		! compute remaining byte offset
	add	rSrcOffset, rCount, rSrcOffset	! and correct addr offset.
						! rSrcOffset should point to
						! the last word of a block
						! so we can subtract blockLen
						! from it to get to the right
						! end of the block to the left.
	ba	toBinaryThrUnscrambleOddWd
	add	rSrcAddr, rSrcOffset, rSrcOffset ! set up rScrOffset

!------------------------------------------------------------------
! ROW ENTRY POINT USED WHEN RIGHTMOST DEST WORD HAS AN EVEN ADDRESS

! calculate the rightmost dest word, which has an even word address.
toBinaryThrEvenEntry:
	call	toBinaryThrComp			! reduce first group of 4 8x8's
	or	rMSBitMask, %lo(cMSBitMask), rMSBitMask
! return with 1st 4 bits
	mov	rScr0, rOut0			! get first 4 bits
	mov	7, rCount			! set count register

toBinaryThrEvenEntryLoop:
	subcc	rSrcOffset, rSrcLimit, %g0	! at end of src line now?
	bpos	toBinaryThrCheckEvenEntryWrite	! yes, quit now and
	or	rOut0, rScr0, rOut0		! combine next 4 bits.
	call	toBinaryThrComp			! get nth 4 bits
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next word to the right
! return with nth 4 bits, rCount decremented in toBinary128Comp
	bnz	toBinaryThrEvenEntryLoop
	sll	rOut0, 1, rOut0			! align previous bits
	or	rOut0, rScr0, rOut0		! combine last 4 bits.

toBinaryThrCheckEvenEntryWrite:
	mov	0, rOut1			! prepare for double write
						! after rOut0 is unscrambled.
						! This 0 is written over the
						! already 0 right frame.
	sll	rOut0, rCount, rOut0		! output bits aligned
	sll	rCount, 2, rCount		! compute remaining byte offset
	add	rSrcOffset, rCount, rSrcOffset	! gen correct addr offset
	ba	toBinaryThrUnscrambleEvenWd
	add	rSrcAddr, rSrcOffset, rSrcOffset ! set up rScrOffset

!------------------------------------------------------------------
! START OF LOOP OVER REST OF 1 ROW

toBinaryThrRowLoop:
	ld	[rSrcOffset - 0x20], rInput	! get 1st 4 bits
	sub	rSrcOffset, 0x20, rSrcOffset	! pt to next wd group to left
	srl	rInput, 1, rInput		! align input with thresholds
	or	rInput, rMSBitMask, rInput	! set MS bits for subtract
	sub	rInput, rThresh, rInput		! compute thresholds
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 4 bits
	srl	rScr0, 1, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x8], rInput	! get 3rd 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0xC], rInput	! get 4th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x10], rInput	! get 5th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x14], rInput	! get 6th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x18], rInput	! get 7th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x1c], rInput	! get 8th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1
	or	rOut1, rScr0, rOut1

! At this point, rOut1 contains the scrambled bits.  We un-scramble them
! using
! rOut1 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
toBinaryThrUnscrambleOddWd:
	sll	rOut1, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut1, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut1, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut1, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut1		! have unscrambled bits.
	bz,a	toBinaryThrFin			! yes, done with this row
	st	rOut1, [rDstAddr+rDstOffset]	! store the odd word if this is
						! left end of row

! **** CALCULATE EVEN DESTINATION WORD ****

toBinaryThrEvenDest:
	ld	[rSrcOffset - 0x20], rInput	! get 1st 4 bits
	sub	rSrcOffset, 0x20, rSrcOffset	! pt to next 8 wd group to left
	srl	rInput, 1, rInput		! align input with thresholds
	or	rInput, rMSBitMask, rInput	! set MS bits for subtract
	sub	rInput, rThresh, rInput		! compute thresholds
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 4 bits
	srl	rScr0, 1, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x8], rInput	! get 3rd 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0xC], rInput	! get 4th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x10], rInput	! get 5th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x14], rInput	! get 6th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x18], rInput	! get 7th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x1c], rInput	! get 8th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rThresh, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0
	or	rOut0, rScr0, rOut0

! At this point, rOut0 contains the scrambled bits.  We un-scramble them
! using
! rOut0 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
toBinaryThrUnscrambleEvenWd:
	sll	rOut0, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut0, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut0, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut0, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut0		! have unscrambled bits.
	bnz	toBinaryThrRowLoop		! yes, prepare for next row
	std	rOut0, [rDstAddr+rDstOffset]	! store even/odd pair

! We've put out the right number of words, so quit
toBinaryThrFin:
	ret
	restore

!------------------------------------------------------------
! toBinaryThrComp
!------------------------------------------------------------

! A subroutine used by toBinaryThr to reduce 4 8-bit bytes into 4 bits.
toBinaryThrComp:
	ld	[rSrcAddr+rSrcOffset], rInput	! get 1st word
	subcc	rCount, 1, rCount		! dec loop count (interlock)
	srl	rInput, 1, rInput		! align input with thresholds
	or	rInput, rMSBitMask, rInput	! set MS bits for subtract
	sub	rInput, rThresh, rInput		! compute thresholds
	and	rInput, rMSBitMask, rInput
	retl
	srl	rInput, 7, rScr0		! align at LS end of each byte

!------------------------------------------------------------
 


!----------------------------------------------------------------------------
! toBinaryBandPass
!----------------------------------------------------------------------------

! This routine is very much like toBinaryThresh except it compares each pixel
! against two thresholds instead of one.  The pixel must be greater than the
! lower threshold and less than the upper.

! In toBinaryThresh, we did four unsigned compares on each 32-bit word in
! parallel.  We did this by shifting the input right one place (we will
! ignore the least significant bits of the pixels in these comparisons) to
! make the "carry-out" bits visible, pre-setting these carry-out bits so that
! borrows don't propagate between the bytes, then subtracting a word
! containing 4 copies of the threshold from this data word.  The "carry-out"
! bits (31, 23, 15, 7) were 1's if the data was greater than the threshold.

! Now we need to add another comparison.  Rather than doing two actual
! comparisons and ANDing the results, we just do another compare with a
! modified threshold value as follows:
! 1. Read data word (32 bits)
! 2. Shift data word right by 1 to make "carry-out" bits from each pixel
!    available.
! 3. Preset the carry out bits (bits 31, 23, 15, 7) to 1.  This prevents
!    borrows from propagating to between bytes (pixels).
! 4. Subtract a word containing four copies of the lower threshold (right
!    shifted by 1 to be aligned) from the data.  The carry-out bit positions
!    in the threshold word are set to 0.

! So far, this is identical to toBinaryThresh.  Values in the range
! [LowerThreshold, UpperThreshold] have been mapped into the range
! [0, UpperThreshold-LowerThreshold].

! 5. Set the carry-out bits in the resulting difference to 0's.
! 6. Subtract the modified difference from a word containing four copies of
!    (UpperThreshold - LowerThreshold) whose carry-out bits are preset.
!    The resulting carry-out bit for each pixel will be set if the original
!    data was greater than or equal to the lower threshold and less than
!    or equal to the upper one.
!    
	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_toBinaryBandPassLine)
MK_GLOBAL_NAME(ip_toBinaryBandPassLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  Also load addresses of byte permutation tables.
	st	rIPStateIn,   [%fp+ipStateOffset]
	st	rSrcAddrIn,   [%fp+srcAddrOffset]
	st	rDstAddrIn,   [%fp+dstAddrOffset]
	st	rDstOffsetIn, [%fp+dstOffsetOffset]
	st	rSrcOffsetIn, [%fp+srcOffsetOffset]
	st	rSrcLimitIn,  [%fp+srcLimitOffset]
	ld	[%fp+srcAddrOffset], rSrcAddr
	ld	[%fp+dstAddrOffset], rDstAddr
	ld	[%fp+dstOffsetOffset], rDstOffset
	ld	[%fp+srcOffsetOffset], rSrcOffset
	ld	[%fp+srcLimitOffset], rSrcLimit

	ld	[%fp+tabByte0Offset], rTabByte0
	ld	[%fp+tabByte1Offset], rTabByte1
	ld	[%fp+tabByte2Offset], rTabByte2
	ld	[%fp+tabByte3Offset], rTabByte3
	ld	[%fp+thresh1Offset], rLower
	ld	[%fp+thresh2Offset], rUpper


! The MS bit mask register is set up in delay slots below.
! Increment rDstOffset by 4 so we decrement it before it is used.
! This makes the loop nicer.
	add	rDstOffset, 4, rDstOffset

! Comments about the two loop entry points are the same (since the code
! is the same).  One is used if the rightmost dest address is
! even, the other if it's odd.  Note that rDstOffset is
! decremented before use, changing the sense of the test.

	add	rDstAddr, rDstOffset, rScr0	! calculate dest address
	andcc	rScr0, 0x4, %g0			! is first addr even?
	bnz	toBinaryBPEvenEntry		! yes, process rightmost even wd
	sethi	%hi(cMSBitMask), rMSBitMask

! calculate the rightmost dest word, which has an odd word address.
toBinaryBPOddEntry:
	call	toBinaryBPComp			! reduce first group of 4 8x8's
	or	rMSBitMask, %lo(cMSBitMask), rMSBitMask
! return with 1st 4 bits
	mov	rScr0, rOut1			! get first 4 bits
	or	%g0, 7, rCount			! set count register

toBinaryBPOddEntryLoop:
	subcc	rSrcOffset, rSrcLimit, %g0	! at end of src line now?
	bpos	toBinaryBPOddEntryExit		! yes, quit now and
	or	rOut1, rScr0, rOut1		! combine next 4 bits.
	call	toBinaryBPComp			! get nth 4 bits
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next word to the right
! return with nth 4 bits, rCount decremented in toBinary128Comp
	bnz	toBinaryBPOddEntryLoop
	sll	rOut1, 1, rOut1			! align previous bits
	or	rOut1, rScr0, rOut1		! combine last 4 bits.

toBinaryBPOddEntryExit:
	sll	rOut1, rCount, rOut1		! yes, align output bits
	sll	rCount, 2, rCount		! compute remaining byte offset
	add	rSrcOffset, rCount, rSrcOffset	! and correct addr offset.
						! rSrcOffset should point to
						! the last word of a block
						! so we can subtract blockLen
						! from it to get to the right
						! end of the block to the left.
	ba	toBinaryBPUnscrambleOddWd
	add	rSrcAddr, rSrcOffset, rSrcOffset ! set up rScrOffset

!------------------------------------------------------------------
! ROW ENTRY POINT USED WHEN RIGHTMOST DEST WORD HAS AN EVEN ADDRESS

! calculate the rightmost dest word, which has an even word address.
toBinaryBPEvenEntry:
	call	toBinaryBPComp			! reduce first group of 4 8x8's
	or	rMSBitMask, %lo(cMSBitMask), rMSBitMask
! return with 1st 4 bits
	mov	rScr0, rOut0			! get first 4 bits
	mov	7, rCount			! set count register

toBinaryBPEvenEntryLoop:
	subcc	rSrcOffset, rSrcLimit, %g0	! at end of src line now?
	bpos	toBinaryBPCheckEvenEntryWrite	! yes, quit now and
	or	rOut0, rScr0, rOut0		! combine next 4 bits.
	call	toBinaryBPComp			! get nth 4 bits
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next word to the right
! return with nth 4 bits, rCount decremented in toBinary128Comp
	bnz	toBinaryBPEvenEntryLoop
	sll	rOut0, 1, rOut0			! align previous bits
	or	rOut0, rScr0, rOut0		! combine last 4 bits.

toBinaryBPCheckEvenEntryWrite:
	mov	0, rOut1			! prepare for double write
						! after rOut0 is unscrambled.
						! This 0 is written over the
						! already 0 right frame.
	sll	rOut0, rCount, rOut0		! output bits aligned
	sll	rCount, 2, rCount		! compute remaining byte offset
	add	rSrcOffset, rCount, rSrcOffset	! gen correct addr offset
	ba	toBinaryBPUnscrambleEvenWd
	add	rSrcAddr, rSrcOffset, rSrcOffset ! set up rScrOffset

!------------------------------------------------------------------
! START OF LOOP OVER REST OF 1 ROW

toBinaryBPRowLoop:
	ld	[rSrcOffset - 0x20], rInput	! get 1st 4 bits
	sub	rSrcOffset, 0x20, rSrcOffset	! pt to next wd group to left
	srl	rInput, 1, rInput		! align input with thresholds
	or	rInput, rMSBitMask, rInput	! set MS bits for subtract
	sub	rInput, rLower, rInput		! compute thresholds
	andn	rInput, rMSBitMask, rInput	! prepare as subtrahend
	sub	rUpper, rInput, rInput		! MS bits set if < Upper
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 4 bits
	srl	rScr0, 1, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x8], rInput	! get 3rd 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0xC], rInput	! get 4th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x10], rInput	! get 5th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x14], rInput	! get 6th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x18], rInput	! get 7th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1

	ld	[rSrcOffset - 0x1c], rInput	! get 8th 4 bits
	or	rOut1, rScr0, rOut1
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 1, rOut1
	or	rOut1, rScr0, rOut1

! At this point, rOut1 contains the scrambled bits.  We un-scramble them
! using
! rOut1 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
toBinaryBPUnscrambleOddWd:
	sll	rOut1, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut1, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut1, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut1, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut1		! have unscrambled bits.
	bz,a	toBinaryBPFin			! yes, done with this row
	st	rOut1, [rDstAddr+rDstOffset]	! store the odd word if this is
						! left end of row

! **** CALCULATE EVEN DESTINATION WORD ****

toBinaryBPEvenDest:
	ld	[rSrcOffset - 0x20], rInput	! get 1st 4 bits
	sub	rSrcOffset, 0x20, rSrcOffset	! pt to next 8 wd group to left
	srl	rInput, 1, rInput		! align input with thresholds
	or	rInput, rMSBitMask, rInput	! set MS bits for subtract
	sub	rInput, rLower, rInput		! compute thresholds
	andn	rInput, rMSBitMask, rInput	! prepare as subtrahend
	sub	rUpper, rInput, rInput		! MS bits set if < Upper
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 4 bits
	srl	rScr0, 1, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x8], rInput	! get 3rd 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0xC], rInput	! get 4th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x10], rInput	! get 5th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x14], rInput	! get 6th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x18], rInput	! get 7th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0

	ld	[rSrcOffset - 0x1c], rInput	! get 8th 4 bits
	or	rOut0, rScr0, rOut0
	srl	rInput, 1, rInput
	or	rInput, rMSBitMask, rInput
	sub	rInput, rLower, rInput
	andn	rInput, rMSBitMask, rInput
	sub	rUpper, rInput, rInput
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 1, rOut0
	or	rOut0, rScr0, rOut0

! At this point, rOut0 contains the scrambled bits.  We un-scramble them
! using
! rOut0 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
toBinaryBPUnscrambleEvenWd:
	sll	rOut0, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut0, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut0, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut0, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut0		! have unscrambled bits.
	bnz	toBinaryBPRowLoop		! yes, prepare for next row
	std	rOut0, [rDstAddr+rDstOffset]	! store even/odd pair

! We've put out the right number of words, so quit
toBinaryBPFin:
	ret
	restore

!------------------------------------------------------------
! toBinaryBPComp
!------------------------------------------------------------

! A subroutine used by toBinaryBP to reduce 4 8-bit bytes into 4 bits.
toBinaryBPComp:
	ld	[rSrcAddr+rSrcOffset], rInput	! get 1st word
	subcc	rCount, 1, rCount		! dec loop count (interlock)
	srl	rInput, 1, rInput		! align input with thresholds
	or	rInput, rMSBitMask, rInput	! set MS bits for subtract
	sub	rInput, rLower, rInput		! compute thresholds
	andn	rInput, rMSBitMask, rInput	! prepare as subtrahend
	sub	rUpper, rInput, rInput		! MS bits set if < Upper
	and	rInput, rMSBitMask, rInput
	retl
	srl	rInput, 7, rScr0		! align at LS end of each byte

!------------------------------------------------------------
 


! -----------------------------------------------------------------------------
! _ip_toTwoBitsPerPixelLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_toTwoBitsPerPixelLine)
MK_GLOBAL_NAME(ip_toTwoBitsPerPixelLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPStateIn,   [%fp+ipStateOffset]
	st	rSrcAddrIn,   [%fp+srcAddrOffset]
	st	rDstAddrIn,   [%fp+dstAddrOffset]
	st	rDstOffsetIn, [%fp+dstOffsetOffset]
	st	rSrcOffsetIn, [%fp+srcOffsetOffset]
	st	rSrcLimitIn,  [%fp+srcLimitOffset]
	ld	[%fp+srcAddrOffset], rSrcAddr
	ld	[%fp+dstAddrOffset], rDstAddr
	ld	[%fp+dstOffsetOffset], rDstOffset
	ld	[%fp+srcOffsetOffset], rSrcOffset
	ld	[%fp+srcLimitOffset], rSrcLimit

	ld	[%fp+tabByte0Offset], rTabByte0
	ld	[%fp+tabByte1Offset], rTabByte1
	ld	[%fp+tabByte2Offset], rTabByte2
	ld	[%fp+tabByte3Offset], rTabByte3

! The MS bit mask register is set up in delay slots below.
! Increment rDstOffset by 4 so we decrement it before it is used.
! This makes the loop nicer.
	add	rDstOffset, 4, rDstOffset

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  Both cases
! have particular code for starting the row.  Note that rDstOffset is
! decremented before use, changing the sense of the test.

	add	rDstAddr, rDstOffset, rScr0	! calculate dest address
	andcc	rScr0, 0x4, %g0			! is first addr even?
	bnz	to2bppEvenEntry			! yes, process rightmost even wd
	sethi	%hi(cMS2BitsMask), rMSBitMask

! calculate the rightmost dest word, which has an odd word address.
to2bppOddEntry:
	call	to2bppComp			! reduce first group of 4 bytes
	or	rMSBitMask, %lo(cMS2BitsMask), rMSBitMask
! return with 1st 8 bits
	mov	rScr0, rOut1			! get first 8 bits
	mov	6, rCount			! set count register

to2bppOddEntryLoop:
	subcc	rSrcOffset, rSrcLimit, %g0	! at end of src line now?
	bpos	to2bppOddEntryExit		! yes, quit now and
	or	rOut1, rScr0, rOut1		! combine next 8 bits.
	call	to2bppComp			! get nth 8 bits
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next word to the right
! return with nth 8 bits, rCount decremented in to2bppComp
	bnz	to2bppOddEntryLoop
	sll	rOut1, 2, rOut1			! align previous bits
	or	rOut1, rScr0, rOut1		! combine last 8 bits.

to2bppOddEntryExit:
	sll	rOut1, rCount, rOut1		! yes, align output bits
	sll	rCount, 1, rCount		! compute remaining byte offset
	add	rSrcOffset, rCount, rSrcOffset	! and correct addr offset.
						! rSrcOffset should point to
						! the last word of a block
						! so we can subtract blockLen
						! from it to get to the right
						! end of the block to the left.
	ba	to2bppUnscrambleOddWd	! go unpermute this word
	add	rSrcAddr, rSrcOffset, rSrcOffset ! and set up rScrOffset

!------------------------------------------------------------------
! ROW ENTRY POINT USED WHEN RIGHTMOST DEST WORD HAS AN EVEN ADDRESS

! calculate the rightmost dest word, which has an even word address.
to2bppEvenEntry:
	call	to2bppComp			! reduce first group of 4 bytes
	or	rMSBitMask, %lo(cMS2BitsMask), rMSBitMask
! return with 1st 8 bits
	mov	rScr0, rOut0			! get first 8 bits
	mov	6, rCount			! set count register

to2bppEvenEntryLoop:
	subcc	rSrcOffset, rSrcLimit, %g0	! at end of src line now?
	bpos	to2bppCheckEvenEntryWrite	! yes, quit now and
	or	rOut0, rScr0, rOut0		! combine next 8 bits.
	call	to2bppComp			! get nth 8 bits
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next word to the right
! return with nth 4 bits, rCount decremented in to2bppComp
	bnz	to2bppEvenEntryLoop
	sll	rOut0, 2, rOut0			! align previous bits
	or	rOut0, rScr0, rOut0		! combine last 8 bits.

to2bppCheckEvenEntryWrite:
	mov	0, rOut1			! prepare for double write
						! after rOut0 is unscrambled.
						! This 0 is written over the
						! already 0 right frame.
	sll	rOut0, rCount, rOut0		! output bits aligned
	sll	rCount, 1, rCount		! compute remaining byte offset
	add	rSrcOffset, rCount, rSrcOffset	! gen correct addr offset
	ba	to2bppUnscrambleEvenWd
	add	rSrcAddr, rSrcOffset, rSrcOffset ! set up rScrOffset



! START OF LOOP OVER REST OF 1 ROW
! processing here is done right to left

to2bppRowLoop:
	ld	[rSrcOffset - 0x10], rInput	! get 1st 8 bits
	sub	rSrcOffset, 0x10, rSrcOffset	! pt to next 4 wd group to left
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 8 bits
	srl	rScr0, 2, rOut1
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x8], rInput	! get 3rd 8 bits
	or	rOut1, rScr0, rOut1
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 2, rOut1

	ld	[rSrcOffset - 0xC], rInput	! get 4th 8 bits
	or	rOut1, rScr0, rOut1
	and	rInput, rMSBitMask, rScr0
	srl	rOut1, 2, rOut1
	or	rOut1, rScr0, rOut1

! At this point, rOut1 contains the scrambled bits.  We un-scramble them
! using
! rOut1 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
to2bppUnscrambleOddWd:
	sll	rOut1, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut1, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut1, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut1, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut1		! have unscrambled bits.
	bz,a	to2bppFin			! yes, done with this row
	st	rOut1, [rDstAddr+rDstOffset]	! store the odd word if this is
						! left end of row

! **** CALCULATE EVEN DESTINATION WORD ****
to2bppEvenDest:
	ld	[rSrcOffset - 0x10], rInput	! get 1st 8 bits
	sub	rSrcOffset, 0x10, rSrcOffset	! pt to next 4 wd group to left
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x4], rInput	! get 2nd 8 bits
	srl	rScr0, 2, rOut0
	and	rInput, rMSBitMask, rScr0

	ld	[rSrcOffset - 0x8], rInput	! get 3rd 8 bits
	or	rOut0, rScr0, rOut0
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 2, rOut0

	ld	[rSrcOffset - 0xC], rInput	! get 4th 8 bits
	or	rOut0, rScr0, rOut0
	and	rInput, rMSBitMask, rScr0
	srl	rOut0, 2, rOut0
	or	rOut0, rScr0, rOut0

! At this point, rOut0 contains the scrambled bits.  We un-scramble them
! using
! rOut0 = remaining scrambled bits
! rScr0 = 8 bits currently being unscrambled
! rInput = 32 unscrambled bits
! rScr1	= accumulated unscrambled bits
to2bppUnscrambleEvenWd:
	sll	rOut0, 2, rScr0			! convert byte 3 to word address
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte3 + rScr0], rScr1
	srl	rOut0, 6, rScr0			! get byte 2
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte2 + rScr0], rInput
	srl	rOut0, 14, rScr0		! get byte 1
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte1 + rScr0], rInput
	srl	rOut0, 22, rScr0		! get byte 0
	or	rScr1, rInput, rScr1
	and	rScr0, 0x3FC, rScr0
	ld	[rTabByte0 + rScr0], rInput

	subcc	rDstOffset, 0x4, rDstOffset	! done with row yet?
	or	rScr1, rInput, rOut0		! have unscrambled bits.
	bnz	to2bppRowLoop			! yes, prepare for next row
	std	rOut0, [rDstAddr+rDstOffset]	! store even/odd pair

! We've put out the right number of words, so quit
to2bppFin:
	ret
	restore


!------------------------------------------------------------
! to2bppComp
!------------------------------------------------------------

! A subroutine used by to2bpp to reduce 4 8-bit bytes into 4 bits.
to2bppComp:
	ld	[rSrcAddr+rSrcOffset], rInput	! get 1st word
	subcc	rCount, 2, rCount		! dec loop count (interlock)
	and	rInput, rMSBitMask, rInput
	retl
	srl	rInput, 6, rScr0		! align at LS end of each byte

!------------------------------------------------------------

/*

$Log: ssdepth.s,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.2  1998/01/13 23:57:33  ddavies
ssdepth.s -> /product/ipcore/ipshared/src/RCS/ssdepth.s,v

CommonCheck: Add IPState parameter to almost all private routines.

# Revision 1.1  1994/01/28  04:19:30  ddavies
# Initial revision
#
! Revision 1.7  1993/12/04  04:58:47  ddavies
! ddavies on Fri Dec 3 20:57:47 PST 1993
!
! ssbinarz.s -> /product/ipcore/ipshared/src/RCS/ssbinarz.s,v
!
! Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
!
! Revision 1.6  1993/11/16  03:04:19  ddavies
! ssbinarz.s -> /product/ipcore/ipshared/src/RCS/ssbinarz.s,v
!
! All of these functions had been writing into the right frame.  All had been
! using the source bytes/line to find the rightmost source word and the
! dest bytes/line to find the rightmost dest word.  This doesn't work because
! bytes/line - 2*frame = image + pad, not just image.  All have been changed
! to use the image width in pixels to calculate source and destination
! word offsets.
!
! Revision 1.5  1993/11/05  02:23:31  ddavies
! ssbinarz.s -> /product/ipcore/ipshared/src/RCS/ssbinarz.s,v
!
! Add RCS magic to have string specifying this file's name, rev number and
! last edit date into the object file.  Also add change log at the bottom.
!

*/


