/*****************************************************************
 *  Copyright (c) 1991, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: ssexpand.s
 *
 * This file contains assembly code for 2:1 expansion by replication.
 * In 2:1 expansion, each pixel in the source image becomes a 2x2 square
 * of pixels in the expanded image.  All expanded pixels take on the
 * value of the single original pixel.
 *
 * All five routines are given the following arguments:
 * i0 = pointer to upper left corner of active source bits
 * i1 = source's bytes/line (is multiple of 4)
 * i2 = pointer to upper left corner of active destination bits
 * i3 = dest's bytes/line (is multiple of 4)
 * i4 = width of active area of src pixrect in units of bits
 * i5 = height of active area of src pixrect in units of lines
 *
 * Expansion by replication should be trivial, but we've managed to make it
 * hard.  There are two main complications.  First, we must remember that
 * there are probably fewer valid bits in the source pixrect than would
 * otherwise fit.  When Pixrects are allocated, the width of memory given is
 * rounded up to the next 32 bit boundary.  Luckily, the number of valid bits
 * is recorded in the pixrect's width parameter.  Say a pixrect had a width
 * of 33 bits and we were supposed to expand it by 32x.  We should get an
 * image 32 words wide (with some pads on the side, but ignore that).  If we
 * blindly expanded every bit in the pixrect image, we'd get an image that was
 * 64 words wide, a big mistake.  This means we have to be careful when
 * deciding how much to expand.
 *
 * However, we would like to do writes in units of words.  Say our 33-bit wide
 * pixrect is to be expanded by 2.  This gives 66 bits.  We'll want to write
 * 3 full words in each line.  Where do we get the other 30 bits in the
 * rightmost word?  Dan Bloomberg says it's ok to just expand whatever dreck
 * is in the 15 bits to the right of the last valid bit to fill in
 * the rightmost word.
 *
 * The other big complication has to do with the write operations themselves.
 * To do expansion, one reads a source word, does a little bit of
 * processing and writes many copies of the output.  Writes are slow on an
 * SS2 and likely to be relatively slow on future machines.  The tricky part
 * of making expansion go fast is optimizing the writes.
 *
 * First, it's a bad idea to write partial results.  Not only does this add
 * to the total number of writes, it causes extra cache misses when the
 * partial results must be read back in.
 *
 * Second, one should use double writes whenever possible.  A double write
 * takes almost exactly the same time as a single write, but we do half as
 * many.
 *
 * Unfortunately, the SPARC puts two restrictions on the use of double
 * writes.  First, the source must be a consecutive pair of registers where
 * the first register has an even number.  Second, the destination
 * address must be a multiple of 8 bytes (even word address).  The memory
 * alignment restrictions usually result in special code to "get in phase"
 * before we start writing every other word.  The register restrictions
 * also force us to generate extra code to align register usage with
 * dest line offset.  For example, one set of code might store results into
 * rOut0, then rOut1, then rOut0, etc while another set would store into
 * rOut1, then rOut0, then rOut1, etc.  The first set of code would be used
 * when the first destination address was even, the second set when the
 * destination address is odd.
 *
 * That was all bad enough in the morphology and reduction code.  In expansion,
 * we must also be able to write multiple lines of image with the same data.
 * This is particularly tricky if the dest image's words/line is an odd number.
 * A double write that works on one line would cause a bus error on the next.
 *
 * The variables are:
 *	A. Whether current dest address is even or odd;
 *	B. Whether words/line of destination image is even/odd;
 *	C. Whether we're at the left edge, right edge or middle of an image;
 *	D. Whether we're doing 2x, 4x, 8x, 16x or 32x expansion.
 *
 * In all, that's 2*2*3*5 = 60 choices!
 *
 * To handle this complexity, we'll use one state machine for each expansion
 * routine.  The state machine's job is to handle output.  It's called once for
 * every 32-bit word to be written.  The state machines for the various
 * expansion factors differ only in the number of writes they do (more
 * expansion -> more writes).  In any given state, the state machine will do
 * some combination of writing a column of output data and saving data to be
 * written next time.  For example, the state machine for 4x expansion might
 * concatenate the word it was given with a previous word and write the pair to
 * lines 0 and 2.  It might also save the word it was given so it can be
 * written to lines 1 and 3 next time.
 *
 * The state machines maintain the rDstOffset register and compute destination
 * addresses.  The also maintain four data registers, rOutE0, rOutE1, rOutO0
 * and rOutO1.  rOoutE0 and rOutE1 hold data destined for even and odd word
 * addresses on the even numbered lines (counting from the top of the
 * destination block to be written).  rOutO0 and rOutO1 perform a similar
 * function for odd numbered lines.
 *
 * We could process lines from left to right or right to left.  Processing
 * from right to left makes generating double word addresses a little easier.
 * Processing from left to right makes it easier to expand arbitrary numbers
 * of bits in a line since bits are left justified.  We'll go from left to
 * right.
 *
 * Each state machine has 4 initial (Left) states:
 *	L1. Dest words/line is even; addr of leftmost word is even.
 *	L2. Dest words/line is even; addr of leftmost word is odd.
 *	L3. Dest words/line is odd; addr of leftmost word is even.
 *	L4. Dest words/line is odd; addr of leftmost word is odd.
 *
 * These are followed by a choice of 4 Middle states:
 *	M1. Dest words/line is even; addr of current word is even.
 *	M2. Dest words/line is even; addr of current word is odd.
 *	M3. Dest words/line is odd; addr of current word is even.
 *	M4. Dest words/line is odd; addr of current word is odd.
 *
 * And 4 end (Right) states:
 *	R1. Dest words/line is even; addr of rightmost word is even.
 *	R2. Dest words/line is even; addr of rightmost word is odd.
 *	R3. Dest words/line is odd; addr of rightmost word is even.
 *	R4. Dest words/line is odd; addr of rightmost word is odd.
 *
 * For example, if the number of words/line is even and the address of
 * the left most word is even, we'd start in state L1, oscillate between
 * M2 and M1, then finish with R2.
 *
 * The left states always write the incoming word to words along the
 * left edge of the output block that have odd addresses (since there
 * won't be subsequent double writes for these words).  These states will
 * advance rDstOffset only if there will be no further writes in this
 * column (i.e. only if all addresses in this column are odd).
 *
 * The middle states do double writes to any words in the *previous* column
 * that have even addresses using previous data plus the new data.  If any of
 * the words in the current column are left unwritten, the current data is
 * saved.  rDstOffset is always advanced.
 *
 * We can reach the end states from any of the middle states.  Calling code
 * will decrement the number of source bits remaining AFTER the write has
 * taken place just before calling the output routines.
 * Each of thoses routines begins with a Branch Less Than or Equal to the
 * appropriate end state.  The end states write all words in the current
 * column with the incoming data.  We know there will be no subsequent double
 * writes since there will be no subsequent writes at all in this block.
 *
 * The end states are likely to do at least some double writes.  If the
 * destination pixrect were only 1 word wide, this wouldn't work.  We have
 * restricted destination pixrects to be at least 2 words wide.  This
 * is checked in the calling code.
 *
 * How do we get from state to state?  A register, rOState, holds the
 * address of the next state routine.  Each state routine adds the difference
 * between the current state's address and the next state's address to
 * rOState.  At the beginning of each line, rOState is initialized
 * to one of the four initial states by branching on the (dest words/line is
 * even) and (addr of first dest word is even).
 *
 * The general expansion scheme is to read an input word, expand portions of
 * it into words laterally, then call the state machine with each of these
 * words.  How do we do the lateral expansion?  For 2x and 4x expansion, we
 * use table lookup.  It's too slow to expand each bit by hand.  For 8x, 16x
 * and 32x expansion, we do the expansion in code.  Each bit is expanded by
 * adding the input to itself, then subtracting 0 from itself with borrow.
 * This extends the single input bit into 32 bits.  We then mask the result
 * (leaving 8, 16 or 32 bits at the right edge of the word),
 * shift the output left to make room and insert the masked bits.
 *
 * Spreading is a lot like expansion except the pixel is only stored into
 * the upper left bit of each 2x2 square.  We do this by using a new table
 * to do the horizontal spreading, then always writing 0's to the second line.
 *
 * Spreading by 3x is a tiny bit more tricky.  We divide the loop into 4
 * parts corresponding to the four possible positions of the expanded 3 bytes.
 * After inserting the expanded 24 bits into its location (and writing out
 * a dest word if necessary), we check the bytecount to see if we're done.
 * If not, the loop continues.  If so, we write out the last expanded bits
 * if necessary and quit.
 *
 * Note also that we must clear two lines below the current one in 3x
 * spreading.
 *
 * ********************* Expand and Mask *****************************
 *
 * The easiest way to do this is to produce another set of output routines.
 * These read the destination locations first, AND them with the source,
 * then write them.  The new entry routines, expandAndMask2Asm,
 * expandAndMask4Asm and expandAndMask8Asm just load pointers to the new
 * output routines, then jump to ip_expand2Asm, ip_expand4Asm and ip_expand8Asm
 * respectively.
 *
 * BUG FIX 10/24/94: We`ve had trouble forever when running Purify with
 * this code.  The problem was that the output state machine set its new
 * state by using
 *	add	rOState, (newState - currentState), rOState.
 *
 * When Purify added code to all the routines, these constants were wrong,
 * but Purify didn't know that so it didn't change them.  We jumped into
 * the middles of output routines.  We now use full sethi/or pairs to set
 * the new output addresses.  Purify takes care of these.
 * 
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssexpand.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	.align	4

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPStateIn	%i0		/* current addr in source image */
#define	rSrcAddrIn	%i1		/* current addr in source image */
#define rSrcPitchIn	%i2		/* source image's bytes/line */
#define rDstAddrIn	%i3		/* current addr in dest image */
#define rDstPitchIn	%i4		/* dest image's bytes/line */
#define rWidthIn	%i5		/* width of source block in bits */

#define	rSrcAddr	%i0		/* current addr in source image */
#define rSrcPitch	%i1		/* source image's bytes/line */
#define rDstAddr	%i2		/* current addr in dest image */
#define rDstPitch	%i3		/* dest image's bytes/line */
#define rWidth		%i4		/* width of source block in bits */
#define rHeight		%i5		/* # of lines to emit */

#define rtabExpAddr	%g1		/* ptr to expansion table */
#define rtabExpAddrHi	%g2		/* ptr to hi half of expansion table,
					 * used only in 8x expansion */


#define rScr0		%g4		/* scratch register PAIR (also used */
#define rScr1		%g5		/* for double reads and writes) */
#define	rScr2		%l0		/* another scratch register PAIR */
#define	rScr3		%l1		/* only used in 2x, 4x and 8x
					 * expansion routines. */

#define	rInput		%g3		/* input word */

#define	rOutE0		%o0		/* output reg pair for even lines */
#define rOutE1		%o1
#define	rOutO0		%o2		/* output reg pair for odd lines */
#define rOutO1		%o3

#define	rSrcOffset	%o4		/* current horizontal source offset */
#define rDstOffset	%o5		/* current horizontal dest offset */
#define	rBlockPitch	%l2		/* skips n lines in a block */

#define	rOState		%l3		/* selects output routine */
#define	r2xDstPitch	%l4		/* used to skip two lines */
#define rDstAddrX	%l5		/* used to form dest addresses */
#define	rDstAddrX1	%l6		/* SAME AS rCount!!.  Only used in
					 * 4x, 8x expansion. */
#define	rInitOState	%l7		/* Initial value of OState for
					 * every line. */

#define	rRetAddr	%o7		/* general return addr */

/* Constants for 2x expansion */
#define	exp2TabAddrMask	0x000001FE	/* need bytes addressing halfwords */

/* Constants for 4x expansion */
#define	exp4TabAddrMask	0x000003FC	/* need bytes addressing words */

/* Constants for 8x expansion */
#define	exp8TabAddrMask	0x000003FC	/* need bytes addressing words */

/* Constants for 16x expansion */
#define	exp16TabAddrMask 0x000003FC	/* need bytes addressing words */

/* Boilerplate parameter offsets */
#define cIPStateOffset	0x44
#define cSrcAddrOffset	0x48
#define cSrcPitchOffset 0x4c
#define cDstAddrOffset	0x50
#define cDstPitchOffset	0x54
#define cWidthOffset	0x58
#define	cHeightOffset	0x5c
#define cExpTableOffset	0x60

/* magical C Frame constant (no local variables) */
#define	LF26		-64

LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! _ip_expandAndMask2Asm Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_expandAndMask2Asm)
MK_GLOBAL_NAME(ip_expandAndMask2Asm):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  We'll need to reload
! rWidth from here.
	st	rIPStateIn,  [%fp+cIPStateOffset]
	st	rSrcAddrIn,  [%fp+cSrcAddrOffset]
	st	rSrcPitchIn, [%fp+cSrcPitchOffset]
	st	rDstAddrIn,  [%fp+cDstAddrOffset]
	st	rDstPitchIn, [%fp+cDstPitchOffset]
	st	rWidthIn,    [%fp+cWidthOffset]

	ld	[%fp+cSrcAddrOffset], rSrcAddr
	ld	[%fp+cSrcPitchOffset], rSrcPitch
	ld	[%fp+cDstAddrOffset], rDstAddr
	ld	[%fp+cDstPitchOffset], rDstPitch
	ld	[%fp+cWidthOffset], rWidth
	ld	[%fp+cHeightOffset], rHeight
	
! load the pointer to expansion table
        ld      [%fp+cExpTableOffset],	rtabExpAddr

! compute distance by which destination addresses will be advanced at the
! end of each line.
	sll	rDstPitch, 1, rBlockPitch	! loop will skip 2 dst lines

! The choice of Left output routine depends on whether the dest pitch and
! the original dest address are even or odd (4 choices in all).  The dest
! pitch obviously does not change.  Since rBlockPitch is an even number
! (in words), rDstAddr is remains even for all lines or odd for all lines.
! Therefore, we can choose the Left output routine outside the loop.

	andcc	rDstPitch, 0x4, %g0		! is rDstPitch even?
	bnz	expandAM2xOddPitch		! no, do the odd cases
	andcc	rDstAddr, 0x4, %g0		! and test rDstAddr on the way
	
! The destination pitch is even.
	bnz	expandAM2xEvenPitchOddAddr	! but first addr is odd
	sethi	%hi(expAM2x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is even
	ba	expand2xImageLoop
	or	rInitOState, %lo(expAM2x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is odd
expandAM2xEvenPitchOddAddr:
	sethi	%hi(expAM2x_L_EP_OA), rInitOState
	ba	expand2xImageLoop
	or	rInitOState, %lo(expAM2x_L_EP_OA), rInitOState

! The destination pitch is odd.
expandAM2xOddPitch:
	bnz	expandAM2xOddPitchOddAddr	! and first addr is odd
	sethi	%hi(expAM2x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is even
	ba	expand2xImageLoop
	or	rInitOState, %lo(expAM2x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is odd
expandAM2xOddPitchOddAddr:
	sethi	%hi(expAM2x_L_OP_OA), rInitOState
	ba	expand2xImageLoop
	or	rInitOState, %lo(expAM2x_L_OP_OA), rInitOState



! -----------------------------------------------------------------------------
! _ip_expand2Asm Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_expand2Asm)
MK_GLOBAL_NAME(ip_expand2Asm):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  We'll need to reload
! rWidth from here.
	st	rIPStateIn,  [%fp+cIPStateOffset]
	st	rSrcAddrIn,  [%fp+cSrcAddrOffset]
	st	rSrcPitchIn, [%fp+cSrcPitchOffset]
	st	rDstAddrIn,  [%fp+cDstAddrOffset]
	st	rDstPitchIn, [%fp+cDstPitchOffset]
	st	rWidthIn,    [%fp+cWidthOffset]

	ld	[%fp+cSrcAddrOffset], rSrcAddr
	ld	[%fp+cSrcPitchOffset], rSrcPitch
	ld	[%fp+cDstAddrOffset], rDstAddr
	ld	[%fp+cDstPitchOffset], rDstPitch
	ld	[%fp+cWidthOffset], rWidth
	ld	[%fp+cHeightOffset], rHeight

! load the pointer to expansion table
	ld	[%fp+cExpTableOffset],	rtabExpAddr

! compute distance by which destination addresses will be advanced at the
! end of each line.
	sll	rDstPitch, 1, rBlockPitch	! loop will skip 2 dst lines

! The choice of Left output routine depends on whether the dest pitch and
! the original dest address are even or odd (4 choices in all).  The dest
! pitch obviously does not change.  Since rBlockPitch is an even number
! (in words), rDstAddr is remains even for all lines or odd for all lines.
! Therefore, we can choose the Left output routine outside the loop.

	andcc	rDstPitch, 0x4, %g0		! is rDstPitch even?
	bnz	expand2xOddPitch		! no, do the odd cases
	andcc	rDstAddr, 0x4, %g0		! and test rDstAddr on the way
	
! The destination pitch is even.
	bnz	expand2xEvenPitchOddAddr	! but first addr is odd
	sethi	%hi(exp2x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is even
	ba	expand2xImageLoop
	or	rInitOState, %lo(exp2x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is odd
expand2xEvenPitchOddAddr:
	sethi	%hi(exp2x_L_EP_OA), rInitOState
	ba	expand2xImageLoop
	or	rInitOState, %lo(exp2x_L_EP_OA), rInitOState

! The destination pitch is odd.
expand2xOddPitch:
	bnz	expand2xOddPitchOddAddr	! and first addr is odd
	sethi	%hi(exp2x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is even
	ba	expand2xImageLoop
	or	rInitOState, %lo(exp2x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is odd
expand2xOddPitchOddAddr:
	sethi	%hi(exp2x_L_OP_OA), rInitOState
	or	rInitOState, %lo(exp2x_L_OP_OA), rInitOState

! START OF LOOP OVER ALL ROWS

! Load the source and destination offsets.  Since we're starting at the
! left edge of the line, they're 0.  Also reload the width and initial
! output state.

expand2xImageLoop:
	mov	0, rSrcOffset
	mov	0, rDstOffset
	ld	[%fp+cWidthOffset], rWidth
	mov	rInitOState, rOState
! Since we write two lines, point to second one.
	add	rDstAddr, rDstPitch, rDstAddrX


! PROCESS ONE ROW

! We read each word, expand the upper, then lower, halves into words and
! have the output state machine store them.

expand2xRowLoop:
	ld	[rSrcAddr+rSrcOffset], rInput	! get current source word
expand2xRowLoop1:
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next one

	srl	rInput, 23, rScr0		! select [31:24]
	and	rScr0, exp2TabAddrMask, rScr0
	lduh	[rtabExpAddr+rScr0], rScr0
	srl	rInput, 15, rScr1		! select [23:16]
	and	rScr1, exp2TabAddrMask, rScr1
	lduh	[rtabExpAddr+rScr1], rScr1
	sll	rScr0, 16, rScr0		! align upper expanded halfword
	or	rScr0, rScr1, rScr0		! compute expanded word
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 16, rWidth		! decrement width (interlock)

! have retired 1st 16 source bits
	ble	expand2xFinRow			! quit if that was end of row

	srl	rInput, 7, rScr0		! select [15:8]
	and	rScr0, exp2TabAddrMask, rScr0
	lduh	[rtabExpAddr+rScr0], rScr0
	sll	rInput, 1, rScr1		! select [7:0]
	and	rScr1, exp2TabAddrMask, rScr1
	lduh	[rtabExpAddr+rScr1], rScr1
	sll	rScr0, 16, rScr0		! align upper expanded halfword
	or	rScr0, rScr1, rScr0		! compute expanded word
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 16, rWidth		! decrement width

! have retired other 16 source bits from current word.
	bg,a	expand2xRowLoop1		! keep going if there's more
	ld	[rSrcAddr+rSrcOffset], rInput	! get next source word

! have finished with that row, are there more rows?
expand2xFinRow:
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point to next src row
	subcc	rHeight, 1, rHeight		! decrement height count
	bnz	expand2xImageLoop		! if more lines, keep going
	add	rDstAddr, rBlockPitch, rDstAddr	! pt to new Dst row

! END OF LOOP ACROSS ALL ROWS

! We've put out the right number of words, so quit
expand2Fin:
	mov	0,%i0
	ret
	restore


! *********** Ordinary 2x expansion output state machine **************

! This is a fake routine that we use to tell Purify to
! pay attention to the code that is otherwise only reached by
! indirect jumps.  Purify seems to need a branch to decide that
! a piece of code should be Purified.  Without these, the jump
! offsets in these routines are not changed and things blow up.
! Purify decides that things are really code if there are at least
! 4 normal opcodes before the first branch.  So put in some
! noops.  We don't really have to touch all the states since
! many of them have 4 ops at the beginning without a branch,
! but let's be safe (in case a branch appears later).
	.global	MK_GLOBAL_NAME(ip_fake2xDispTable)
MK_GLOBAL_NAME(ip_fake2xDispTable):
	nop
	nop
	nop
	nop
	deccc	%i0
	bz	exp2x_L_EP_EA
	deccc	%i0
	bz	exp2x_L_EP_OA
	deccc	%i0
	bz	exp2x_L_OP_EA
	deccc	%i0
	bz	exp2x_L_OP_OA
	deccc	%i0
	bz	exp2x_M_EP_EA
	deccc	%i0
	bz	exp2x_M_EP_OA
	deccc	%i0
	bz	exp2x_M_OP_EA
	deccc	%i0
	bz	exp2x_M_OP_OA
	deccc	%i0
	bz	exp2x_R_EP_EA
	deccc	%i0
	bz	exp2x_R_EP_OA
	deccc	%i0
	bz	exp2x_R_OP_EA
	deccc	%i0
	bz	exp2x_R_OP_OA
	deccc	%i0
	nop
	ret
	restore

!------------------------------------------
! exp2x_L_EP_EA

! Left state, Even Pitch, Even Starting address
! Just save the output, it will be written as part of a double
! write next time.  If output is 1 word wide, do both writes now.
exp2x_L_EP_EA:
	ble	exp2x_L_EP_OA
	mov	rScr0, rOutE0
	sethi	%hi(exp2x_M_EP_OA), rOState
	retl
	or	rOState, %lo(exp2x_M_EP_OA), rOState


!------------------------------------------
! exp2x_L_EP_OA

! Left state, Even Pitch, Odd Starting address
! Write output to both lines.
! DO NOT advance rDstOffset here. It is supposed to be 1 "behind"
! so that the next double write (in two words) will work.

exp2x_L_EP_OA:
	st	rScr0, [rDstAddr+rDstOffset]
	st	rScr0, [rDstAddrX+rDstOffset]
	sethi	%hi(exp2x_M_EP_EA), rOState
	retl
	or	rOState, %lo(exp2x_M_EP_EA), rOState


!------------------------------------------
! exp2x_L_OP_EA

! Left state, Odd Pitch, Even Starting address
! Write output to lower line, save for even lines next

exp2x_L_OP_EA:
	ble	exp2x_L_EP_OA
	mov	rScr0, rOutE0
	st	rScr0, [rDstAddrX+rDstOffset]
	sethi	%hi(exp2x_M_OP_OA), rOState
	retl
	or	rOState, %lo(exp2x_M_OP_OA), rOState


!------------------------------------------
! exp2x_L_OP_OA

! Left state, Odd Pitch, Odd Starting address
! Write output to upper line, save for lower line next, do NOT inc rDstOffset

exp2x_L_OP_OA:
	ble	exp2x_L_EP_OA
	mov	rScr0, rOutO0
	st	rScr0, [rDstAddr+rDstOffset]
	sethi	%hi(exp2x_M_OP_EA), rOState
	retl
	or	rOState, %lo(exp2x_M_OP_EA), rOState


!------------------------------------------
! exp2x_M_EP_EA

! Middle state, Even Pitch, Even address
! save data for double writes on both lines later, inc rDstOffset

exp2x_M_EP_EA:
	ble	exp2x_R_EP_EA
	mov	rScr0, rOutE0
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp2x_M_EP_OA), rOState
	retl
	or	rOState, %lo(exp2x_M_EP_OA), rOState


!------------------------------------------
! exp2x_M_EP_OA

! Middle state, Even Pitch, Odd address
! do double writes on both lines, inc rDstOffset

exp2x_M_EP_OA:
	ble	exp2x_R_EP_OA
	mov	rScr0, rOutE1
	std	rOutE0, [rDstAddr+rDstOffset]
	std	rOutE0, [rDstAddrX+rDstOffset]
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp2x_M_EP_EA), rOState
	retl
	or	rOState, %lo(exp2x_M_EP_EA), rOState


!------------------------------------------
! exp2x_M_OP_EA

! Middle state, Odd Pitch, Even address
! save data for double write on upper line next, do double write on lower line,
! inc rDstOffset

exp2x_M_OP_EA:
	ble	exp2x_R_OP_EA
	mov	rScr0, rOutO1
	mov	rScr0, rOutE0
	std	rOutO0, [rDstAddrX+rDstOffset]
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp2x_M_OP_OA), rOState
	retl
	or	rOState, %lo(exp2x_M_OP_OA), rOState


!------------------------------------------
! exp2x_M_OP_OA

! Middle state, Odd Pitch, Odd address
! save data for double write on lower line next, do double write on upper line,
! inc rDstOffset

exp2x_M_OP_OA:
	ble	exp2x_R_OP_OA
	mov	rScr0, rOutE1
	mov	rScr0, rOutO0
	std	rOutE0, [rDstAddr+rDstOffset]
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp2x_M_OP_EA), rOState
	retl
	or	rOState, %lo(exp2x_M_OP_EA), rOState


!------------------------------------------
! exp2x_R_EP_EA

! Right state, Even Pitch, Even address
! do single writes on both lines now, don't bother with rDstOffset or next
! state

exp2x_R_EP_EA:
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	st	rScr0, [rDstAddr+rDstOffset]
	retl
	st	rScr0, [rDstAddrX+rDstOffset]


!------------------------------------------
! exp2x_R_EP_OA

! Right state, Even Pitch, Odd address
! do double writes on both lines now, don't bother with rDstOffset or next
! state. Output data is already in rOutE1.

exp2x_R_EP_OA:
	std	rOutE0, [rDstAddr+rDstOffset]
	retl
	std	rOutE0, [rDstAddrX+rDstOffset]


!------------------------------------------
! exp2x_R_OP_EA

! Right state, Odd Pitch, Even address
! do single writes on upper line, double write on lower line,
! don't bother with rDstOffset or next state.  Data is already in rOutO1.

exp2x_R_OP_EA:
	std	rOutO0, [rDstAddrX+rDstOffset]
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	retl
	st	rScr0, [rDstAddr+rDstOffset]

!------------------------------------------
! exp2x_R_OP_OA

! Right state, Odd Pitch, Odd address
! do single writes on lower line, double write on upper line,
! don't bother with rDstOffset or next state.  Data is already in rOutE1.

exp2x_R_OP_OA:
	std	rOutE0, [rDstAddr+rDstOffset]
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	retl
	st	rScr0, [rDstAddrX+rDstOffset]



! *********** 2x Expand and Mask output state machine **************

! This is a fake routine that we use to tell Purify to
! pay attention to the code that is otherwise only reached by
! indirect jumps.  Purify seems to need a branch to decide that
! a piece of code should be Purified.  Without these, the jump
! offsets in these routines are not changed and things blow up.
! Purify decides that things are really code if there are at least
! 4 normal opcodes before the first branch.  So put in some
! noops.  We don't really have to touch all the states since
! many of them have 4 ops at the beginning without a branch,
! but let's be safe (in case a branch appears later).
	.global	MK_GLOBAL_NAME(ip_fakeAM2xDispTable)
MK_GLOBAL_NAME(ip_fakeAM2xDispTable):
	nop
	nop
	nop
	nop
	deccc	%i0
	bz	expAM2x_L_EP_EA
	deccc	%i0
	bz	expAM2x_L_EP_OA
	deccc	%i0
	bz	expAM2x_L_OP_EA
	deccc	%i0
	bz	expAM2x_L_OP_OA
	deccc	%i0
	bz	expAM2x_M_EP_EA
	deccc	%i0
	bz	expAM2x_M_EP_OA
	deccc	%i0
	bz	expAM2x_M_OP_EA
	deccc	%i0
	bz	expAM2x_M_OP_OA
	deccc	%i0
	bz	expAM2x_R_EP_EA
	deccc	%i0
	bz	expAM2x_R_EP_OA
	deccc	%i0
	bz	expAM2x_R_OP_EA
	deccc	%i0
	bz	expAM2x_R_OP_OA
	nop
	ret
	restore

!------------------------------------------
! expAM2x_L_EP_EA

! Left state, Even Pitch, Even Starting address
! Just save the output, it will be written as part of a double
! write next time.  If output is 1 word wide, do both writes now.
expAM2x_L_EP_EA:
	ble	expAM2x_L_EP_OA
	mov	rScr0, rOutE0
	sethi	%hi(expAM2x_M_EP_OA), rOState
	retl
	or	rOState, %lo(expAM2x_M_EP_OA), rOState


!------------------------------------------
! expAM2x_L_EP_OA

! Left state, Even Pitch, Odd Starting address
! Write output to both lines.
! DO NOT advance rDstOffset here. It is supposed to be 1 "behind"
! so that the next double write (in two words) will work.

expAM2x_L_EP_OA:
	ld	[rDstAddr+rDstOffset], rScr2
	ld	[rDstAddrX+rDstOffset], rScr3
	and	rScr2, rScr0, rScr1
	st	rScr1, [rDstAddr+rDstOffset]
	and	rScr3, rScr0, rScr1
	st	rScr1, [rDstAddrX+rDstOffset]
	sethi	%hi(expAM2x_M_EP_EA), rOState
	retl
	or	rOState, %lo(expAM2x_M_EP_EA), rOState


!------------------------------------------
! expAM2x_L_OP_EA

! Left state, Odd Pitch, Even Starting address
! Write output to lower line, save for even lines next

expAM2x_L_OP_EA:
	ble	expAM2x_L_EP_OA
	ld	[rDstAddrX+rDstOffset], rScr2
	mov	rScr0, rOutE0
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]
	sethi	%hi(expAM2x_M_OP_OA), rOState
	retl
	or	rOState, %lo(expAM2x_M_OP_OA), rOState


!------------------------------------------
! expAM2x_L_OP_OA

! Left state, Odd Pitch, Odd Starting address
! Write output to upper line, save for lower line next, do NOT inc rDstOffset

expAM2x_L_OP_OA:
	ble	expAM2x_L_EP_OA
	ld	[rDstAddr+rDstOffset], rScr2
	mov	rScr0, rOutO0
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rDstOffset]
	sethi	%hi(expAM2x_M_OP_EA), rOState
	retl
	or	rOState, %lo(expAM2x_M_OP_EA), rOState


!------------------------------------------
! expAM2x_M_EP_EA

! Middle state, Even Pitch, Even address
! save data for double writes on both lines later, inc rDstOffset

expAM2x_M_EP_EA:
	ble	expAM2x_R_EP_EA
	mov	rScr0, rOutE0
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(expAM2x_M_EP_OA), rOState
	retl
	or	rOState, %lo(expAM2x_M_EP_OA), rOState


!------------------------------------------
! expAM2x_M_EP_OA

! Middle state, Even Pitch, Odd address
! do double writes on both lines, inc rDstOffset

expAM2x_M_EP_OA:
	ble	expAM2x_R_EP_OA
	mov	rScr0, rOutE1
	ldd	[rDstAddr+rDstOffset], rScr0	! load rScr0,,rScr1
	ldd	[rDstAddrX+rDstOffset], rScr2	! load rScr2,,rScr3
	and	rScr0, rOutE0, rScr0
	and	rScr1, rOutE1, rScr1
	std	rScr0, [rDstAddr+rDstOffset]
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(expAM2x_M_EP_EA), rOState
	retl
	or	rOState, %lo(expAM2x_M_EP_EA), rOState


!------------------------------------------
! expAM2x_M_OP_EA

! Middle state, Odd Pitch, Even address
! save data for double write on upper line next, do double write on lower line,
! inc rDstOffset

expAM2x_M_OP_EA:
	ble	expAM2x_R_OP_EA
	mov	rScr0, rOutO1
	ldd	[rDstAddrX+rDstOffset], rScr2
	mov	rScr0, rOutE0
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(expAM2x_M_OP_OA), rOState
	retl
	or	rOState, %lo(expAM2x_M_OP_OA), rOState


!------------------------------------------
! expAM2x_M_OP_OA

! Middle state, Odd Pitch, Odd address
! save data for double write on lower line next, do double write on upper line,
! inc rDstOffset

expAM2x_M_OP_OA:
	ble	expAM2x_R_OP_OA
	mov	rScr0, rOutE1
	ldd	[rDstAddr+rDstOffset], rScr2
	mov	rScr0, rOutO0
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(expAM2x_M_OP_EA), rOState
	retl
	or	rOState, %lo(expAM2x_M_OP_EA), rOState


!------------------------------------------
! expAM2x_R_EP_EA

! Right state, Even Pitch, Even address
! do single writes on both lines now, don't bother with rDstOffset or next
! state

expAM2x_R_EP_EA:
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	ld	[rDstAddr+rDstOffset], rScr2
	ld	[rDstAddrX+rDstOffset], rScr3
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rDstOffset]
	and	rScr3, rScr0, rScr3
	retl
	st	rScr3, [rDstAddrX+rDstOffset]


!------------------------------------------
! expAM2x_R_EP_OA

! Right state, Even Pitch, Odd address
! do double writes on both lines now, don't bother with rDstOffset or next
! state. Output data is already in rOutE1.

expAM2x_R_EP_OA:
	ldd	[rDstAddr+rDstOffset], rScr0
	ldd	[rDstAddrX+rDstOffset], rScr2
	and	rScr0, rOutE0, rScr0
	and	rScr1, rOutE1, rScr1
	std	rScr0, [rDstAddr+rDstOffset]
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	retl
	std	rScr2, [rDstAddrX+rDstOffset]


!------------------------------------------
! expAM2x_R_OP_EA

! Right state, Odd Pitch, Even address
! do single writes on upper line, double write on lower line,
! don't bother with rDstOffset or next state.  Data is already in rOutO1.

expAM2x_R_OP_EA:
	ldd	[rDstAddrX+rDstOffset], rScr0
	add	rDstOffset, 0x4, rScr3		/* was pting to prev word */
	ld	[rDstAddr+rScr3	], rScr2
	and	rScr0, rOutO0, rScr0
	and	rScr1, rOutO1, rScr1
	std	rScr0, [rDstAddrX+rDstOffset]
	and	rScr2, rOutO1, rScr2
	retl
	st	rScr2, [rDstAddr+rScr3]

!------------------------------------------
! expAM2x_R_OP_OA

! Right state, Odd Pitch, Odd address
! do single writes on lower line, double write on upper line,
! don't bother with rDstOffset or next state.  Data is already in rOutE1.

expAM2x_R_OP_OA:
	ldd	[rDstAddr+rDstOffset], rScr0
	add	rDstOffset, 0x4, rScr3		/* was pting to prev word */
	ld	[rDstAddrX+rScr3], rScr2
	and	rScr0, rOutE0, rScr0
	and	rScr1, rOutE1, rScr1
	std	rScr0, [rDstAddr+rDstOffset]
	and	rScr2, rOutE1, rScr2
	retl
	st	rScr2, [rDstAddrX+rScr3]





! -----------------------------------------------------------------------------
! _ip_expandAndMask4Asm Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_expandAndMask4Asm)
MK_GLOBAL_NAME(ip_expandAndMask4Asm):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  We'll need to reload
! rWidth from here.
	st	rIPStateIn,  [%fp+cIPStateOffset]
	st	rSrcAddrIn,  [%fp+cSrcAddrOffset]
	st	rSrcPitchIn, [%fp+cSrcPitchOffset]
	st	rDstAddrIn,  [%fp+cDstAddrOffset]
	st	rDstPitchIn, [%fp+cDstPitchOffset]
	st	rWidthIn,    [%fp+cWidthOffset]

	ld	[%fp+cSrcAddrOffset], rSrcAddr
	ld	[%fp+cSrcPitchOffset], rSrcPitch
	ld	[%fp+cDstAddrOffset], rDstAddr
	ld	[%fp+cDstPitchOffset], rDstPitch
	ld	[%fp+cWidthOffset], rWidth
	ld	[%fp+cHeightOffset], rHeight

! load the pointer to expansion table
	ld	[%fp+cExpTableOffset], rtabExpAddr

! compute distance by which destination addresses will be advanced at the
! end of each line.
	sll	rDstPitch, 1, r2xDstPitch	! used to skip two lines
	sll	rDstPitch, 2, rBlockPitch	! loop will skip 4 dst lines

! The choice of Left output routine depends on whether the dest pitch and
! the original dest address are even or odd (4 choices in all).  The dest
! pitch obviously does not change.  Since rBlockPitch is an even number
! (in words), rDstAddr is remains even for all lines or odd for all lines.
! Therefore, we can choose the Left output routine outside the loop.

	andcc	rDstPitch, 0x4, %g0		! is rDstPitch even?
	bnz	expandAM4xOddPitch		! no, do the odd cases
	andcc	rDstAddr, 0x4, %g0		! and test rDstAddr on the way
	
! The destination pitch is even.
	bnz	expandAM4xEvenPitchOddAddr	! but first addr is odd
	sethi	%hi(expAM4x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is even
	ba	expand4xImageLoop
	or	rInitOState, %lo(expAM4x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is odd
expandAM4xEvenPitchOddAddr:
	sethi	%hi(expAM4x_L_EP_OA), rInitOState
	ba	expand4xImageLoop
	or	rInitOState, %lo(expAM4x_L_EP_OA), rInitOState

! The destination pitch is odd.
expandAM4xOddPitch:
	bnz	expandAM4xOddPitchOddAddr	! and first addr is odd
	sethi	%hi(expAM4x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is even
	ba	expand4xImageLoop
	or	rInitOState, %lo(expAM4x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is odd
expandAM4xOddPitchOddAddr:
	sethi	%hi(expAM4x_L_OP_OA), rInitOState
	ba	expand4xImageLoop
	or	rInitOState, %lo(expAM4x_L_OP_OA), rInitOState



! -----------------------------------------------------------------------------
! _ip_expand4Asm Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_expand4Asm)
MK_GLOBAL_NAME(ip_expand4Asm):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  We'll need to reload
! rWidth from here.
	st	rIPStateIn,  [%fp+cIPStateOffset]
	st	rSrcAddrIn,  [%fp+cSrcAddrOffset]
	st	rSrcPitchIn, [%fp+cSrcPitchOffset]
	st	rDstAddrIn,  [%fp+cDstAddrOffset]
	st	rDstPitchIn, [%fp+cDstPitchOffset]
	st	rWidthIn,    [%fp+cWidthOffset]

	ld	[%fp+cSrcAddrOffset], rSrcAddr
	ld	[%fp+cSrcPitchOffset], rSrcPitch
	ld	[%fp+cDstAddrOffset], rDstAddr
	ld	[%fp+cDstPitchOffset], rDstPitch
	ld	[%fp+cWidthOffset], rWidth
	ld	[%fp+cHeightOffset], rHeight

! load the pointer to expansion table
	ld      [%fp+cExpTableOffset],	rtabExpAddr

! compute distance by which destination addresses will be advanced at the
! end of each line.
	sll	rDstPitch, 1, r2xDstPitch	! used to skip two lines
	sll	rDstPitch, 2, rBlockPitch	! loop will skip 4 dst lines

! The choice of Left output routine depends on whether the dest pitch and
! the original dest address are even or odd (4 choices in all).  The dest
! pitch obviously does not change.  Since rBlockPitch is an even number
! (in words), rDstAddr is remains even for all lines or odd for all lines.
! Therefore, we can choose the Left output routine outside the loop.

	andcc	rDstPitch, 0x4, %g0		! is rDstPitch even?
	bnz	expand4xOddPitch		! no, do the odd cases
	andcc	rDstAddr, 0x4, %g0		! and test rDstAddr on the way
	
! The destination pitch is even.
	bnz	expand4xEvenPitchOddAddr	! but first addr is odd
	sethi	%hi(exp4x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is even
	ba	expand4xImageLoop
	or	rInitOState, %lo(exp4x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is odd
expand4xEvenPitchOddAddr:
	sethi	%hi(exp4x_L_EP_OA), rInitOState
	ba	expand4xImageLoop
	or	rInitOState, %lo(exp4x_L_EP_OA), rInitOState

! The destination pitch is odd.
expand4xOddPitch:
	bnz	expand4xOddPitchOddAddr	! and first addr is odd
	sethi	%hi(exp4x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is even
	ba	expand4xImageLoop
	or	rInitOState, %lo(exp4x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is odd
expand4xOddPitchOddAddr:
	sethi	%hi(exp4x_L_OP_OA), rInitOState
	or	rInitOState, %lo(exp4x_L_OP_OA), rInitOState

! START OF LOOP OVER ALL ROWS

! Load the source and destination offsets.  Since we're starting at the
! left edge of the line, they're 0.  Also reload the width and initial
! output routine.

expand4xImageLoop:
	mov	0, rSrcOffset
	mov	0, rDstOffset
	ld	[%fp+cWidthOffset], rWidth
	mov	rInitOState, rOState

! PROCESS ONE ROW

! We read each word, expand the bytes (upper to lower) into words and
! have the output state machine store them.

expand4xRowLoop:
	ld	[rSrcAddr+rSrcOffset], rInput	! get current source word
expand4xRowLoop1:
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next one

	srl	rInput, 22, rScr0		! select [31:24]
	and	rScr0, exp4TabAddrMask, rScr0
	ld	[rtabExpAddr+rScr0], rScr0
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 8, rWidth		! decrement width

! retire bits [23:16]
	ble	expand4xFinRow			! quit if that was end of row

	srl	rInput, 14, rScr0		! select [23:16]
	and	rScr0, exp4TabAddrMask, rScr0
	ld	[rtabExpAddr+rScr0], rScr0
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 8, rWidth		! decrement width

! retire bits [15:8]
	ble	expand4xFinRow			! quit if that was end of row

	srl	rInput, 6, rScr0		! select [15:8]
	and	rScr0, exp4TabAddrMask, rScr0
	ld	[rtabExpAddr+rScr0], rScr0
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 8, rWidth		! decrement width

! retire bits [7:0]
	ble	expand4xFinRow			! quit if that was end of row

	sll	rInput, 2, rScr0		! select [7:0]
	and	rScr0, exp4TabAddrMask, rScr0
	ld	[rtabExpAddr+rScr0], rScr0
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 8, rWidth		! decrement width

! have retired other 16 source bits from current word.
	bg,a	expand4xRowLoop1		! keep going if there's more
	ld	[rSrcAddr+rSrcOffset], rInput	! get next source word

! have finished with that row, are there more rows?
expand4xFinRow:
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point to next src row
	subcc	rHeight, 1, rHeight		! decrement height count
	bnz	expand4xImageLoop		! if more lines, keep going
	add	rDstAddr, rBlockPitch, rDstAddr	! with new DstAddr

! END OF LOOP ACROSS ALL ROWS

! We've put out the right number of words, so quit
expand4Fin:
	mov	0,%i0
	ret
	restore


! ***************** 4x Expansion Output State Machine  *****************

! This is a fake routine that we use to tell Purify to
! pay attention to the code that is otherwise only reached by
! indirect jumps.  Purify seems to need a branch to decide that
! a piece of code should be Purified.  Without these, the jump
! offsets in these routines are not changed and things blow up.
! Purify decides that things are really code if there are at least
! 4 normal opcodes before the first branch.  So put in some
! noops.  We don't really have to touch all the states since
! many of them have 4 ops at the beginning without a branch,
! but let's be safe (in case a branch appears later).
	.global	MK_GLOBAL_NAME(ip_fake4xDispTable)
MK_GLOBAL_NAME(ip_fake4xDispTable):
	nop
	nop
	nop
	nop
	deccc	%i0
	bz	exp4x_L_EP_EA
	deccc	%i0
	bz	exp4x_L_EP_OA
	deccc	%i0
	bz	exp4x_L_OP_EA
	deccc	%i0
	bz	exp4x_L_OP_OA
	deccc	%i0
	bz	exp4x_M_EP_EA
	deccc	%i0
	bz	exp4x_M_EP_OA
	deccc	%i0
	bz	exp4x_M_OP_EA
	deccc	%i0
	bz	exp4x_M_OP_OA
	deccc	%i0
	bz	exp4x_R_EP_EA
	deccc	%i0
	bz	exp4x_R_EP_OA
	deccc	%i0
	bz	exp4x_R_OP_EA
	deccc	%i0
	bz	exp4x_R_OP_OA
	nop
	ret
	restore

!------------------------------------------
! exp4x_L_EP_EA

! Left state, Even Pitch, Even Starting address
! Just save the output, it will be written as part of a double
! write next time.  If output is 1 word wide, do all writes now.
exp4x_L_EP_EA:
	ble	exp4x_L_EP_OA
	mov	rScr0, rOutE0
	sethi	%hi(exp4x_M_EP_OA), rOState
	retl
	or	rOState, %lo(exp4x_M_EP_OA), rOState

!------------------------------------------
! exp4x_L_EP_OA

! Left state, Even Pitch, Odd Starting address
! Write output to 4 lines
! DO NOT advance rDstOffset here. It is supposed to be 1 "behind"
! so that the next double write (in two words) will work.

exp4x_L_EP_OA:
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	sethi	%hi(exp4x_M_EP_EA), rOState
	retl
	or	rOState, %lo(exp4x_M_EP_EA), rOState


!------------------------------------------
! exp4x_L_OP_EA

! Left state, Odd Pitch, Even Starting address
! Write output to odd lines, save for even lines next

exp4x_L_OP_EA:
	ble	exp4x_L_EP_OA
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	mov	rScr0, rOutE0
	sethi	%hi(exp4x_M_OP_OA), rOState
	retl
	or	rOState, %lo(exp4x_M_OP_OA), rOState


!------------------------------------------
! exp4x_L_OP_OA

! Left state, Odd Pitch, Odd Starting address
! Write output to even lines, save for odd lines next

exp4x_L_OP_OA:
	ble	exp4x_L_EP_OA
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	mov	rScr0, rOutO0
	sethi	%hi(exp4x_M_OP_EA), rOState
	retl
	or	rOState, %lo(exp4x_M_OP_EA), rOState


!------------------------------------------
! exp4x_M_EP_EA

! Middle state, Even Pitch, Even address
! save data for double writes on all lines later, inc rDstOffset

exp4x_M_EP_EA:
	ble	exp4x_R_EP_EA
	mov	rScr0, rOutE0
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp4x_M_EP_OA), rOState
	retl
	or	rOState, %lo(exp4x_M_EP_OA), rOState


!------------------------------------------
! exp4x_M_EP_OA

! Middle state, Even Pitch, Odd address
! do double writes on all lines, inc rDstOffset

exp4x_M_EP_OA:
	ble	exp4x_R_EP_OA
	mov	rScr0, rOutE1
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp4x_M_EP_EA), rOState
	retl
	or	rOState, %lo(exp4x_M_EP_EA), rOState


!------------------------------------------
! exp4x_M_OP_EA

! Middle state, Odd Pitch, Even address
! save data for double write on even lines next, do double write on odd lines,
! inc rDstOffset

exp4x_M_OP_EA:
	ble	exp4x_R_OP_EA
	mov	rScr0, rOutO1
	mov	rScr0, rOutE0
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp4x_M_OP_OA), rOState
	retl
	or	rOState, %lo(exp4x_M_OP_OA), rOState


!------------------------------------------
! exp4x_M_OP_OA

! Middle state, Odd Pitch, Odd address
! save data for double write on odd lines next, do double write on even lines,
! inc rDstOffset

exp4x_M_OP_OA:
	ble	exp4x_R_OP_OA
	mov	rScr0, rOutE1
	mov	rScr0, rOutO0
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp4x_M_OP_EA), rOState
	retl
	or	rOState, %lo(exp4x_M_OP_EA), rOState


!------------------------------------------
! exp4x_R_EP_EA

! Right state, Even Pitch, Even address
! do single writes on all lines now, don't bother with rDstOffset or next
! state

exp4x_R_EP_EA:
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */


!------------------------------------------
! exp4x_R_EP_OA

! Right state, Even Pitch, Odd address
! do double writes on all lines now, don't bother with rDstOffset or next
! state. Output data is already in rOutE1.

exp4x_R_EP_OA:
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	retl
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 3 */


!------------------------------------------
! exp4x_R_OP_EA

! Right state, Odd Pitch, Even address
! do single writes on even lines, double write on odd lines,
! don't bother with rDstOffset or next state.  Data is already in rOutO1.

exp4x_R_OP_EA:
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */

!------------------------------------------
! exp4x_R_OP_OA

! Right state, Odd Pitch, Odd address
! do single writes on odd lines, double write on even lines,
! don't bother with rDstOffset or next state.  Data is already in rOutE1.

exp4x_R_OP_OA:
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */



! *************** 4x Expand And Mask Output State Machine  ***************

! This is a fake routine that we use to tell Purify to
! pay attention to the code that is otherwise only reached by
! indirect jumps.  Purify seems to need a branch to decide that
! a piece of code should be Purified.  Without these, the jump
! offsets in these routines are not changed and things blow up.
! Purify decides that things are really code if there are at least
! 4 normal opcodes before the first branch.  So put in some
! noops.  We don't really have to touch all the states since
! many of them have 4 ops at the beginning without a branch,
! but let's be safe (in case a branch appears later).
	.global	MK_GLOBAL_NAME(ip_fakeAM4xDispTable)
MK_GLOBAL_NAME(ip_fakeAM4xDispTable):
	nop
	nop
	nop
	nop
	deccc	%i0
	bz	expAM4x_L_EP_EA
	deccc	%i0
	bz	expAM4x_L_EP_OA
	deccc	%i0
	bz	expAM4x_L_OP_EA
	deccc	%i0
	bz	expAM4x_L_OP_OA
	deccc	%i0
	bz	expAM4x_M_EP_EA
	deccc	%i0
	bz	expAM4x_M_EP_OA
	deccc	%i0
	bz	expAM4x_M_OP_EA
	deccc	%i0
	bz	expAM4x_M_OP_OA
	deccc	%i0
	bz	expAM4x_R_EP_EA
	deccc	%i0
	bz	expAM4x_R_EP_OA
	deccc	%i0
	bz	expAM4x_R_OP_EA
	deccc	%i0
	bz	expAM4x_R_OP_OA
	nop
	ret
	restore

!------------------------------------------
! expAM4x_L_EP_EA

! Left state, Even Pitch, Even Starting address
! Just save the output, it will be written as part of a double
! write next time.  If output is 1 word wide, do all writes now.
expAM4x_L_EP_EA:
	ble	expAM4x_L_EP_OA
	mov	rScr0, rOutE0
	sethi	%hi(expAM4x_M_EP_OA), rOState
	retl
	or	rOState, %lo(expAM4x_M_EP_OA), rOState

!------------------------------------------
! expAM4x_L_EP_OA

! Left state, Even Pitch, Odd Starting address
! Write output to 4 lines
! DO NOT advance rDstOffset here. It is supposed to be 1 "behind"
! so that the next double write (in two words) will work.

expAM4x_L_EP_OA:
	ld	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 2 */

	ld	[rDstAddrX+rDstOffset], rScr2
	sethi	%hi(expAM4x_M_EP_EA), rOState
	or	rOState, %lo(expAM4x_M_EP_EA), rOState
	and	rScr2, rScr0, rScr2
	retl
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 3 */


!------------------------------------------
! expAM4x_L_OP_EA

! Left state, Odd Pitch, Even Starting address
! Write output to odd lines, save for even lines next

expAM4x_L_OP_EA:
	ble	expAM4x_L_EP_OA
	add	rDstAddr, rDstPitch, rDstAddrX
	ld	[rDstAddrX+rDstOffset], rScr2	
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	mov	rScr0, rOutE0
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 3 */
	sethi	%hi(expAM4x_M_OP_OA), rOState
	retl
	or	rOState, %lo(expAM4x_M_OP_OA), rOState


!------------------------------------------
! expAM4x_L_OP_OA

! Left state, Odd Pitch, Odd Starting address
! Write output to even lines, save for odd lines next

expAM4x_L_OP_OA:
	ble	expAM4x_L_EP_OA
	ld	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ld	[rDstAddrX+rDstOffset], rScr2
	mov	rScr0, rOutO0
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 2 */
	sethi	%hi(expAM4x_M_OP_EA), rOState
	retl
	or	rOState, %lo(expAM4x_M_OP_EA), rOState


!------------------------------------------
! expAM4x_M_EP_EA

! Middle state, Even Pitch, Even address
! save data for double writes on all lines later, inc rDstOffset

expAM4x_M_EP_EA:
	ble	expAM4x_R_EP_EA
	mov	rScr0, rOutE0
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(expAM4x_M_EP_OA), rOState
	retl
	or	rOState, %lo(expAM4x_M_EP_OA), rOState


!------------------------------------------
! expAM4x_M_EP_OA

! Middle state, Even Pitch, Odd address
! do double writes on all lines, inc rDstOffset

expAM4x_M_EP_OA:
	ble	expAM4x_R_EP_OA
	mov	rScr0, rOutE1
	ldd	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 2 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	sethi	%hi(expAM4x_M_EP_EA), rOState
	or	rOState, %lo(expAM4x_M_EP_EA), rOState
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 3 */
	retl
	add	rDstOffset, 0x4, rDstOffset


!------------------------------------------
! expAM4x_M_OP_EA

! Middle state, Odd Pitch, Even address
! save data for double write on even lines next, do double write on odd lines,
! inc rDstOffset

expAM4x_M_OP_EA:
	ble	expAM4x_R_OP_EA
	mov	rScr0, rOutO1
	mov	rScr0, rOutE0
	add	rDstAddr, rDstPitch, rDstAddrX
	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	sethi	%hi(expAM4x_M_OP_OA), rOState
	or	rOState, %lo(expAM4x_M_OP_OA), rOState
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 3 */
	retl
	add	rDstOffset, 0x4, rDstOffset


!------------------------------------------
! expAM4x_M_OP_OA

! Middle state, Odd Pitch, Odd address
! save data for double write on odd lines next, do double write on even lines,
! inc rDstOffset

expAM4x_M_OP_OA:
	ble	expAM4x_R_OP_OA
	mov	rScr0, rOutE1
	mov	rScr0, rOutO0
	ldd	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, r2xDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	sethi	%hi(expAM4x_M_OP_EA), rOState
	or	rOState, %lo(expAM4x_M_OP_EA), rOState
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 2 */
	retl
	add	rDstOffset, 0x4, rDstOffset


!------------------------------------------
! expAM4x_R_EP_EA

! Right state, Even Pitch, Even address
! do single writes on all lines now, don't bother with rDstOffset or next
! state

expAM4x_R_EP_EA:
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	ld	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 2 */

	ld	[rDstAddrX+rDstOffset], rScr2
							/* INTERLOCK */
	and	rScr2, rScr0, rScr2
	retl
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 3 */


!------------------------------------------
! expAM4x_R_EP_OA

! Right state, Even Pitch, Odd address
! do double writes on all lines now, don't bother with rDstOffset or next
! state. Output data is already in rOutE1.

expAM4x_R_EP_OA:
	ldd	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 2 */

	ldd	[rDstAddrX+rDstOffset], rScr2
							/* INTERLOCK */
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	retl
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 3 */


!------------------------------------------
! expAM4x_R_OP_EA

! Right state, Odd Pitch, Even address
! do single writes on even lines, double write on odd lines,
! don't bother with rDstOffset or next state.  Data is already in rOutO1.

expAM4x_R_OP_EA:
	add	rDstAddr, rDstPitch, rDstAddrX
	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstOffset, 0x4, rScr1		/* was pting to prev word */
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 3 */

	ld	[rDstAddr+rScr1], rScr2
	add	rDstAddr, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rScr1]			/* line 0 */

	ld	[rDstAddrX+rScr1], rScr2
							/* INTERLOCK */
	and	rScr2, rScr0, rScr2
	retl
	st	rScr2, [rDstAddrX+rScr1]		/* line 2 */

!------------------------------------------
! expAM4x_R_OP_OA

! Right state, Odd Pitch, Odd address
! do single writes on odd lines, double write on even lines,
! don't bother with rDstOffset or next state.  Data is already in rOutE1.

expAM4x_R_OP_OA:
	ldd	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, r2xDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstOffset, 0x4, rScr1		/* was pting to prev word */
	add	rDstAddr, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 2 */

	ld	[rDstAddrX1+rScr1], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rScr1]		/* line 1 */

	ld	[rDstAddrX+rScr1], rScr2
							/* INTERLOCK */
	and	rScr2, rScr0, rScr2
	retl
	st	rScr2, [rDstAddrX+rScr1]		/* line 3 */



! -----------------------------------------------------------------------------
! _ip_expandAndMask8Asm Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_expandAndMask8Asm)
MK_GLOBAL_NAME(ip_expandAndMask8Asm):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  We'll need to reload
! rWidth from here.
	st	rIPStateIn,  [%fp+cIPStateOffset]
	st	rSrcAddrIn,  [%fp+cSrcAddrOffset]
	st	rSrcPitchIn, [%fp+cSrcPitchOffset]
	st	rDstAddrIn,  [%fp+cDstAddrOffset]
	st	rDstPitchIn, [%fp+cDstPitchOffset]
	st	rWidthIn,    [%fp+cWidthOffset]

	ld	[%fp+cSrcAddrOffset], rSrcAddr
	ld	[%fp+cSrcPitchOffset], rSrcPitch
	ld	[%fp+cDstAddrOffset], rDstAddr
	ld	[%fp+cDstPitchOffset], rDstPitch
	ld	[%fp+cWidthOffset], rWidth
	ld	[%fp+cHeightOffset], rHeight

! load the pointer to expansion table
	ld	[%fp+cExpTableOffset], rtabExpAddr
	add	rtabExpAddr, 256*4, rtabExpAddrHi	/* point to upper
							 * half of table */

! compute distance by which destination addresses will be advanced at the
! end of each line.
	sll	rDstPitch, 1, r2xDstPitch	! used to skip two lines
	sll	rDstPitch, 3, rBlockPitch	! loop will skip 8 dest lines

! The choice of Left output routine depends on whether the dest pitch and
! the original dest address are even or odd (4 choices in all).  The dest
! pitch obviously does not change.  Since rBlockPitch is an even number
! (in words), rDstAddr is remains even for all lines or odd for all lines.
! Therefore, we can choose the Left output routine outside the loop.

	andcc	rDstPitch, 0x4, %g0		! is rDstPitch even?
	bnz	expandAM8xOddPitch		! no, do the odd cases
	andcc	rDstAddr, 0x4, %g0		! and test rDstAddr on the way

! The destination pitch is even.
	bnz	expandAM8xEvenPitchOddAddr	! but first addr is odd
	sethi	%hi(expAM8x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is even
	ba	expand8xImageLoop
	or	rInitOState, %lo(expAM8x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is odd
expandAM8xEvenPitchOddAddr:
	sethi	%hi(expAM8x_L_EP_OA), rInitOState
	ba	expand8xImageLoop
	or	rInitOState, %lo(expAM8x_L_EP_OA), rInitOState

! The destination pitch is odd.
expandAM8xOddPitch:
	bnz	expandAM8xOddPitchOddAddr	! and first addr is odd
	sethi	%hi(expAM8x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is even
	ba	expand8xImageLoop
	or	rInitOState, %lo(expAM8x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is odd
expandAM8xOddPitchOddAddr:
	sethi	%hi(expAM8x_L_OP_OA), rInitOState
	ba	expand8xImageLoop
	or	rInitOState, %lo(expAM8x_L_OP_OA), rInitOState



! -----------------------------------------------------------------------------
! _ip_expand8Asm Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_expand8Asm)
MK_GLOBAL_NAME(ip_expand8Asm):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  We'll need to reload
! rWidth from here.
	st	rIPStateIn,  [%fp+cIPStateOffset]
	st	rSrcAddrIn,  [%fp+cSrcAddrOffset]
	st	rSrcPitchIn, [%fp+cSrcPitchOffset]
	st	rDstAddrIn,  [%fp+cDstAddrOffset]
	st	rDstPitchIn, [%fp+cDstPitchOffset]
	st	rWidthIn,    [%fp+cWidthOffset]

	ld	[%fp+cSrcAddrOffset], rSrcAddr
	ld	[%fp+cSrcPitchOffset], rSrcPitch
	ld	[%fp+cDstAddrOffset], rDstAddr
	ld	[%fp+cDstPitchOffset], rDstPitch
	ld	[%fp+cWidthOffset], rWidth
	ld	[%fp+cHeightOffset], rHeight

! load the pointer to expansion table
	ld      [%fp+cExpTableOffset],	rtabExpAddr
	add	rtabExpAddr, 256*4, rtabExpAddrHi	/* point to upper
							 * half of table */

! compute distance by which destination addresses will be advanced at the
! end of each line.
	sll	rDstPitch, 1, r2xDstPitch	! used to skip two lines
	sll	rDstPitch, 3, rBlockPitch	! loop will skip 8 dest lines

! The choice of Left output routine depends on whether the dest pitch and
! the original dest address are even or odd (4 choices in all).  The dest
! pitch obviously does not change.  Since rBlockPitch is an even number
! (in words), rDstAddr is remains even for all lines or odd for all lines.
! Therefore, we can choose the Left output routine outside the loop.

	andcc	rDstPitch, 0x4, %g0		! is rDstPitch even?
	bnz	expand8xOddPitch		! no, do the odd cases
	andcc	rDstAddr, 0x4, %g0		! and test rDstAddr on the way

! The destination pitch is even.
	bnz	expand8xEvenPitchOddAddr	! but first addr is odd
	sethi	%hi(exp8x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is even
	ba	expand8xImageLoop
	or	rInitOState, %lo(exp8x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is odd
expand8xEvenPitchOddAddr:
	sethi	%hi(exp8x_L_EP_OA), rInitOState
	ba	expand8xImageLoop
	or	rInitOState, %lo(exp8x_L_EP_OA), rInitOState

! The destination pitch is odd.
expand8xOddPitch:
	bnz	expand8xOddPitchOddAddr	! and first addr is odd
	sethi	%hi(exp8x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is even
	ba	expand8xImageLoop
	or	rInitOState, %lo(exp8x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is odd
expand8xOddPitchOddAddr:
	sethi	%hi(exp8x_L_OP_OA), rInitOState
	or	rInitOState, %lo(exp8x_L_OP_OA), rInitOState

! START OF LOOP OVER ALL ROWS

! Load the source and destination offsets.  Since we're starting at the
! left edge of the line, they're 0.  Also reload the width and initial
! output routine.

expand8xImageLoop:
	mov	0, rSrcOffset
	mov	0, rDstOffset
	ld	[%fp+cWidthOffset], rWidth
	mov	rInitOState, rOState

! PROCESS ONE ROW

! We read each word, expand the nibbles (upper to lower) into words and
! have the output state machine store them.

expand8xRowLoop:
	ld	[rSrcAddr+rSrcOffset], rInput	! get current source word
expand8xRowLoop1:
	add	rSrcOffset, 0x4, rSrcOffset	! pt to next one

	srl	rInput, 22, rScr0		! select [31:24]
	and	rScr0, exp8TabAddrMask, rScr0
	ld	[rtabExpAddrHi+rScr0], rScr0	! use upper nibble
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 4, rWidth		! decrement width

! retire bits [27:24]
	ble	expand8xFinRow			! quit if that was end of row

	srl	rInput, 22, rScr0		! select [31:24]
	and	rScr0, exp8TabAddrMask, rScr0
	ld	[rtabExpAddr+rScr0], rScr0	! use lower nibble
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 4, rWidth		! decrement width

! retire bits [23:20]
	ble	expand8xFinRow			! quit if that was end of row

	srl	rInput, 14, rScr0		! select [23:16]
	and	rScr0, exp8TabAddrMask, rScr0
	ld	[rtabExpAddrHi+rScr0], rScr0	! use upper nibble
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 4, rWidth		! decrement width

! retire bits [19:16]
	ble	expand8xFinRow			! quit if that was end of row

	srl	rInput, 14, rScr0		! select [23:16]
	and	rScr0, exp8TabAddrMask, rScr0
	ld	[rtabExpAddr+rScr0], rScr0	! use lower nibble
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 4, rWidth		! decrement width

! retire bits [15:12]
	ble	expand8xFinRow			! quit if that was end of row

	srl	rInput, 6, rScr0		! select [15:8]
	and	rScr0, exp8TabAddrMask, rScr0
	ld	[rtabExpAddrHi+rScr0], rScr0	! use upper nibble
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 4, rWidth		! decrement width

! retire bits [11:8]
	ble	expand8xFinRow			! quit if that was end of row

	srl	rInput, 6, rScr0		! select [15:8]
	and	rScr0, exp8TabAddrMask, rScr0
	ld	[rtabExpAddr+rScr0], rScr0	! use lower nibble
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 4, rWidth		! decrement width

! retire bits [7:4]
	ble	expand8xFinRow			! quit if that was end of row

	sll	rInput, 2, rScr0		! select [7:0]
	and	rScr0, exp8TabAddrMask, rScr0
	ld	[rtabExpAddrHi+rScr0], rScr0	! use upper nibble
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 4, rWidth		! decrement width

! retire bits [3:0]
	ble	expand8xFinRow			! quit if that was end of row

	sll	rInput, 2, rScr0		! select [7:0]
	and	rScr0, exp8TabAddrMask, rScr0
	ld	[rtabExpAddr+rScr0], rScr0	! use lower nibble
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 4, rWidth		! decrement width

! have retired all bits from current input word.  Is there another one on
! this row?
	bg,a	expand8xRowLoop1		! keep going if there's more
	ld	[rSrcAddr+rSrcOffset], rInput	! get next source word

! END OP LOOP ACROSS ONE ROW

! have finished with that row, are there more rows?
expand8xFinRow:
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point to next src row
	subcc	rHeight, 1, rHeight		! decrement height count
	bnz	expand8xImageLoop		! if more lines, keep going
	add	rDstAddr, rBlockPitch, rDstAddr	! with new DstAddr

! END OF LOOP ACROSS ALL ROWS

! We've put out the right number of words, so quit
expand8Fin:
	mov	0,%i0
	ret
	restore


! ***************** 8x Expansion Output State Machine  *****************

! This is a fake routine that we use to tell Purify to
! pay attention to the code that is otherwise only reached by
! indirect jumps.  Purify seems to need a branch to decide that
! a piece of code should be Purified.  Without these, the jump
! offsets in these routines are not changed and things blow up.
! Purify decides that things are really code if there are at least
! 4 normal opcodes before the first branch.  So put in some
! noops.  We don't really have to touch all the states since
! many of them have 4 ops at the beginning without a branch,
! but let's be safe (in case a branch appears later).
	.global	MK_GLOBAL_NAME(ip_fake8xDispTable)
MK_GLOBAL_NAME(ip_fake8xDispTable):
	nop
	nop
	nop
	nop
	deccc	%i0
	bz	exp8x_L_EP_EA
	deccc	%i0
	bz	exp8x_L_EP_OA
	deccc	%i0
	bz	exp8x_L_OP_EA
	deccc	%i0
	bz	exp8x_L_OP_OA
	deccc	%i0
	bz	exp8x_M_EP_EA
	deccc	%i0
	bz	exp8x_M_EP_OA
	deccc	%i0
	bz	exp8x_M_OP_EA
	deccc	%i0
	bz	exp8x_M_OP_OA
	deccc	%i0
	bz	exp8x_R_EP_EA
	deccc	%i0
	bz	exp8x_R_EP_OA
	deccc	%i0
	bz	exp8x_R_OP_EA
	deccc	%i0
	bz	exp8x_R_OP_OA
	nop
	ret
	restore

!------------------------------------------
! exp8x_L_EP_EA

! Left state, Even Pitch, Even Starting address
! Just save the output, it will be written as part of a double
! write next time.  If dest is 1 word wide, do all writes now.
exp8x_L_EP_EA:
	ble	exp8x_L_EP_OA
	mov	rScr0, rOutE0
	sethi	%hi(exp8x_M_EP_OA), rOState
	retl
	or	rOState, %lo(exp8x_M_EP_OA), rOState

!------------------------------------------
! exp8x_L_EP_OA

! Left state, Even Pitch, Odd Starting address
! Write output to 8 lines
! DO NOT advance rDstOffset here. It is supposed to be 1 "behind"
! so that the next double write (in two words) will work.

exp8x_L_EP_OA:
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 7 */
	sethi	%hi(exp8x_M_EP_EA), rOState
	retl
	or	rOState, %lo(exp8x_M_EP_EA), rOState


!------------------------------------------
! exp8x_L_OP_EA

! Left state, Odd Pitch, Even Starting address
! Write output to odd lines, save for even lines next

exp8x_L_OP_EA:
	ble	exp8x_L_EP_OA
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 7 */
	mov	rScr0, rOutE0
	sethi	%hi(exp8x_M_OP_OA), rOState
	retl
	or	rOState, %lo(exp8x_M_OP_OA), rOState


!------------------------------------------
! exp8x_L_OP_OA

! Left state, Odd Pitch, Odd Starting address
! Write output to even lines, save for odd lines next

exp8x_L_OP_OA:
	ble	exp8x_L_EP_OA
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 6 */
	mov	rScr0, rOutO0
	sethi	%hi(exp8x_M_OP_EA), rOState
	retl
	or	rOState, %lo(exp8x_M_OP_EA), rOState


!------------------------------------------
! exp8x_M_EP_EA

! Middle state, Even Pitch, Even address
! save data for double writes on all lines later, inc rDstOffset

exp8x_M_EP_EA:
	ble	exp8x_R_EP_EA
	mov	rScr0, rOutE0
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp8x_M_EP_OA), rOState
	retl
	or	rOState, %lo(exp8x_M_EP_OA), rOState


!------------------------------------------
! exp8x_M_EP_OA

! Middle state, Even Pitch, Odd address
! do double writes on all lines, inc rDstOffset

exp8x_M_EP_OA:
	ble	exp8x_R_EP_OA
	mov	rScr0, rOutE1
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp8x_M_EP_EA), rOState
	retl
	or	rOState, %lo(exp8x_M_EP_EA), rOState


!------------------------------------------
! exp8x_M_OP_EA

! Middle state, Odd Pitch, Even address
! save data for double write on even lines next, do double write on odd lines,
! inc rDstOffset

exp8x_M_OP_EA:
	ble	exp8x_R_OP_EA
	mov	rScr0, rOutO1
	mov	rScr0, rOutE0
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp8x_M_OP_OA), rOState
	retl
	or	rOState, %lo(exp8x_M_OP_OA), rOState


!------------------------------------------
! exp8x_M_OP_OA

! Middle state, Odd Pitch, Odd address
! save data for double write on odd lines next, do double write on even lines,
! inc rDstOffset

exp8x_M_OP_OA:
	ble	exp8x_R_OP_OA
	mov	rScr0, rOutE1
	mov	rScr0, rOutO0
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp8x_M_OP_EA), rOState
	retl
	or	rOState, %lo(exp8x_M_OP_EA), rOState


!------------------------------------------
! exp8x_R_EP_EA

! Right state, Even Pitch, Even address
! do single writes on all lines now, don't bother with rDstOffset or next
! state

exp8x_R_EP_EA:
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 7 */


!------------------------------------------
! exp8x_R_EP_OA

! Right state, Even Pitch, Odd address
! do double writes on all lines now, don't bother with rDstOffset or next
! state. Output data is already in rOutE1.

exp8x_R_EP_OA:
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	retl
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 7 */


!------------------------------------------
! exp8x_R_OP_EA

! Right state, Odd Pitch, Even address
! do single writes on even lines, double write on odd lines,
! don't bother with rDstOffset or next state.  Data is already in rOutO1.

exp8x_R_OP_EA:
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 6 */


!------------------------------------------
! exp8x_R_OP_OA

! Right state, Odd Pitch, Odd address
! do single writes on odd lines, double write on even lines,
! don't bother with rDstOffset or next state.  Data is already in rOutE1.

exp8x_R_OP_OA:
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 7 */




! *************** 8x Expand and Mask Output State Machine  ***************


! This is a fake routine that we use to tell Purify to
! pay attention to the code that is otherwise only reached by
! indirect jumps.  Purify seems to need a branch to decide that
! a piece of code should be Purified.  Without these, the jump
! offsets in these routines are not changed and things blow up.
! Purify decides that things are really code if there are at least
! 4 normal opcodes before the first branch.  So put in some
! noops.  We don't really have to touch all the states since
! many of them have 4 ops at the beginning without a branch,
! but let's be safe (in case a branch appears later).
	.global	MK_GLOBAL_NAME(ip_fakeAM8xDispTable)
MK_GLOBAL_NAME(ip_fakeAM8xDispTable):
	nop
	nop
	nop
	nop
	deccc	%i0
	bz	expAM8x_L_EP_EA
	deccc	%i0
	bz	expAM8x_L_EP_OA
	deccc	%i0
	bz	expAM8x_L_OP_EA
	deccc	%i0
	bz	expAM8x_L_OP_OA
	deccc	%i0
	bz	expAM8x_M_EP_EA
	deccc	%i0
	bz	expAM8x_M_EP_OA
	deccc	%i0
	bz	expAM8x_M_OP_EA
	deccc	%i0
	bz	expAM8x_M_OP_OA
	deccc	%i0
	bz	expAM8x_R_EP_EA
	deccc	%i0
	bz	expAM8x_R_EP_OA
	deccc	%i0
	bz	expAM8x_R_OP_EA
	deccc	%i0
	bz	expAM8x_R_OP_OA
	nop
	ret
	restore

!------------------------------------------
! expAM8x_L_EP_EA

! Left state, Even Pitch, Even Starting address
! Just save the output, it will be written as part of a double
! write next time.  If dest is 1 word wide, do all writes now.
expAM8x_L_EP_EA:
	ble	expAM8x_L_EP_OA
	mov	rScr0, rOutE0
	sethi	%hi(expAM8x_M_EP_OA), rOState
	retl
	or	rOState, %lo(expAM8x_M_EP_OA), rOState

!------------------------------------------
! expAM8x_L_EP_OA

! Left state, Even Pitch, Odd Starting address
! Write output to 8 lines
! DO NOT advance rDstOffset here. It is supposed to be 1 "behind"
! so that the next double write (in two words) will work.

expAM8x_L_EP_OA:
	ld	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 2 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 3 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 4 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 5 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 6 */

	ld	[rDstAddrX+rDstOffset], rScr2
	sethi	%hi(expAM8x_M_EP_EA), rOState
	or	rOState, %lo(expAM8x_M_EP_EA), rOState
	and	rScr2, rScr0, rScr2
	retl
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 7 */


!------------------------------------------
! expAM8x_L_OP_EA

! Left state, Odd Pitch, Even Starting address
! Write output to odd lines, save for even lines next

expAM8x_L_OP_EA:
	ble	expAM8x_L_EP_OA
	add	rDstAddr, rDstPitch, rDstAddrX

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 3 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 5 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	mov	rScr0, rOutE0
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 7 */
	sethi	%hi(expAM8x_M_OP_OA), rOState
	retl
	or	rOState, %lo(expAM8x_M_OP_OA), rOState


!------------------------------------------
! expAM8x_L_OP_OA

! Left state, Odd Pitch, Odd Starting address
! Write output to even lines, save for odd lines next

expAM8x_L_OP_OA:
	ble	expAM8x_L_EP_OA
	ld	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 2 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 4 */

	ld	[rDstAddrX+rDstOffset], rScr2
	mov	rScr0, rOutO0
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 6 */
	sethi	%hi(expAM8x_M_OP_EA), rOState
	retl
	or	rOState, %lo(expAM8x_M_OP_EA), rOState


!------------------------------------------
! expAM8x_M_EP_EA

! Middle state, Even Pitch, Even address
! save data for double writes on all lines later, inc rDstOffset

expAM8x_M_EP_EA:
	ble	expAM8x_R_EP_EA
	mov	rScr0, rOutE0
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(expAM8x_M_EP_OA), rOState
	retl
	or	rOState, %lo(expAM8x_M_EP_OA), rOState


!------------------------------------------
! expAM8x_M_EP_OA

! Middle state, Even Pitch, Odd address
! do double writes on all lines, inc rDstOffset

expAM8x_M_EP_OA:
	ble	expAM8x_R_EP_OA
	mov	rScr0, rOutE1

	ldd	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 2 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 3 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 4 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 5 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 6 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	sethi	%hi(expAM8x_M_EP_EA), rOState
	or	rOState, %lo(expAM8x_M_EP_EA), rOState
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 7 */

	retl
	add	rDstOffset, 0x4, rDstOffset


!------------------------------------------
! expAM8x_M_OP_EA

! Middle state, Odd Pitch, Even address
! save data for double write on even lines next, do double write on odd lines,
! inc rDstOffset

expAM8x_M_OP_EA:
	ble	expAM8x_R_OP_EA
	mov	rScr0, rOutO1
	mov	rScr0, rOutE0
	add	rDstAddr, rDstPitch, rDstAddrX

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 3 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 5 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	sethi	%hi(expAM8x_M_OP_OA), rOState
	or	rOState, %lo(expAM8x_M_OP_OA), rOState
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 7 */

	retl
	add	rDstOffset, 0x4, rDstOffset


!------------------------------------------
! expAM8x_M_OP_OA

! Middle state, Odd Pitch, Odd address
! save data for double write on odd lines next, do double write on even lines,
! inc rDstOffset

expAM8x_M_OP_OA:
	ble	expAM8x_R_OP_OA
	mov	rScr0, rOutE1
	mov	rScr0, rOutO0

	ldd	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, r2xDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 2 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 4 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	sethi	%hi(expAM8x_M_OP_EA), rOState
	or	rOState, %lo(expAM8x_M_OP_EA), rOState
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 6 */
	retl
	add	rDstOffset, 0x4, rDstOffset


!------------------------------------------
! expAM8x_R_EP_EA

! Right state, Even Pitch, Even address
! do single writes on all lines now, don't bother with rDstOffset or next
! state

expAM8x_R_EP_EA:
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */

	ld	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 2 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 3 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 4 */

	ld	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 5 */

	ld	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rDstOffset]		/* line 6 */

	ld	[rDstAddrX+rDstOffset], rScr2
							/* INTERLOCK */
	and	rScr2, rScr0, rScr2
	retl
	st	rScr2, [rDstAddrX+rDstOffset]		/* line 7 */


!------------------------------------------
! expAM8x_R_EP_OA

! Right state, Even Pitch, Odd address
! do double writes on all lines now, don't bother with rDstOffset or next
! state. Output data is already in rOutE1.

expAM8x_R_EP_OA:
	ldd	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 2 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 3 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 4 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 5 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, rDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 6 */

	ldd	[rDstAddrX+rDstOffset], rScr2
							/* INTERLOCK */
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	retl
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 7 */


!------------------------------------------
! expAM8x_R_OP_EA

! Right state, Odd Pitch, Even address
! do single writes on even lines, double write on odd lines,
! don't bother with rDstOffset or next state.  Data is already in rOutO1.

expAM8x_R_OP_EA:
	add	rDstAddr, rDstPitch, rDstAddrX

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 1 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 3 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 5 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstOffset, 0x4, rScr1		/* was pting to prev word */
	and	rScr2, rOutO0, rScr2
	and	rScr3, rOutO1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 7 */

	ld	[rDstAddr+rScr1], rScr2
	add	rDstAddr, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddr+rScr1]			/* line 0 */

	ld	[rDstAddrX+rScr1], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rScr1]		/* line 2 */

	ld	[rDstAddrX1+rScr1], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rScr1]		/* line 4 */

	ld	[rDstAddrX+rScr1], rScr2
							/* INTERLOCK */
	and	rScr2, rScr0, rScr2
	retl
	st	rScr2, [rDstAddrX+rScr1]		/* line 6 */


!------------------------------------------
! expAM8x_R_OP_OA

! Right state, Odd Pitch, Odd address
! do single writes on odd lines, double write on even lines,
! don't bother with rDstOffset or next state.  Data is already in rOutE1.

expAM8x_R_OP_OA:
	ldd	[rDstAddr+rDstOffset], rScr2
	add	rDstAddr, r2xDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddr+rDstOffset]		/* line 0 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 2 */

	ldd	[rDstAddrX1+rDstOffset], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX1+rDstOffset]		/* line 4 */

	ldd	[rDstAddrX+rDstOffset], rScr2
	add	rDstOffset, 0x4, rScr1		/* was pting to prev word */
	add	rDstAddr, rDstPitch, rDstAddrX1
	and	rScr2, rOutE0, rScr2
	and	rScr3, rOutE1, rScr3
	std	rScr2, [rDstAddrX+rDstOffset]		/* line 6 */

	ld	[rDstAddrX1+rScr1], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rScr1]		/* line 1 */

	ld	[rDstAddrX+rScr1], rScr2
	add	rDstAddrX, r2xDstPitch, rDstAddrX1
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX+rScr1]		/* line 3 */

	ld	[rDstAddrX1+rScr1], rScr2
	add	rDstAddrX1, r2xDstPitch, rDstAddrX
	and	rScr2, rScr0, rScr2
	st	rScr2, [rDstAddrX1+rScr1]		/* line 5 */

	ld	[rDstAddrX+rScr1], rScr2
							/* INTERLOCK */
	and	rScr2, rScr0, rScr2
	retl
	st	rScr2, [rDstAddrX+rScr1]		/* line 7 */



! -----------------------------------------------------------------------------
! _ip_expand16Asm Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_expand16Asm)
MK_GLOBAL_NAME(ip_expand16Asm):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  We'll need to reload
! rWidth from here.
	st	rIPStateIn,  [%fp+cIPStateOffset]
	st	rSrcAddrIn,  [%fp+cSrcAddrOffset]
	st	rSrcPitchIn, [%fp+cSrcPitchOffset]
	st	rDstAddrIn,  [%fp+cDstAddrOffset]
	st	rDstPitchIn, [%fp+cDstPitchOffset]
	st	rWidthIn,    [%fp+cWidthOffset]

	ld	[%fp+cSrcAddrOffset], rSrcAddr
	ld	[%fp+cSrcPitchOffset], rSrcPitch
	ld	[%fp+cDstAddrOffset], rDstAddr
	ld	[%fp+cDstPitchOffset], rDstPitch
	ld	[%fp+cWidthOffset], rWidth
	ld	[%fp+cHeightOffset], rHeight

! load the pointer to expansion table
	ld	[%fp+cExpTableOffset], rtabExpAddr

! compute distance by which destination addresses will be advanced at the
! end of each line.
	sll	rDstPitch, 1, r2xDstPitch	! used to skip two lines
	sll	rDstPitch, 4, rBlockPitch	! loop will skip 16 dest lines

! The choice of Left output routine depends on whether the dest pitch and
! the original dest address are even or odd (4 choices in all).  The dest
! pitch obviously does not change.  Since rBlockPitch is an even number
! (in words), rDstAddr is remains even for all lines or odd for all lines.
! Therefore, we can choose the Left output routine outside the loop.

	andcc	rDstPitch, 0x4, %g0		! is rDstPitch even?
	bnz	expand16xOddPitch		! no, do the odd cases
	andcc	rDstAddr, 0x4, %g0		! and test rDstAddr on the way

! The destination pitch is even.
	bnz	expand16xEvenPitchOddAddr	! but first addr is odd
	sethi	%hi(exp16x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is even
	ba	expand16xImageLoop
	or	rInitOState, %lo(exp16x_L_EP_EA), rInitOState
! Dest Pitch is even, first addr is odd
expand16xEvenPitchOddAddr:
	sethi	%hi(exp16x_L_EP_OA), rInitOState
	ba	expand16xImageLoop
	or	rInitOState, %lo(exp16x_L_EP_OA), rInitOState

! The destination pitch is odd.
expand16xOddPitch:
	bnz	expand16xOddPitchOddAddr	! and first addr is odd
	sethi	%hi(exp16x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is even
	ba	expand16xImageLoop
	or	rInitOState, %lo(exp16x_L_OP_EA), rInitOState
! Dest Pitch is odd, first addr is odd
expand16xOddPitchOddAddr:
	sethi	%hi(exp16x_L_OP_OA), rInitOState
	or	rInitOState, %lo(exp16x_L_OP_OA), rInitOState

! START OF LOOP OVER ALL ROWS

! Load the source and destination offsets.  Since we're starting at the
! left edge of the line, they're 0.  Also reload the width and initial
! output routine.

expand16xImageLoop:
	mov	0, rSrcOffset
	mov	0, rDstOffset
	ld	[%fp+cWidthOffset], rWidth
	mov	rInitOState, rOState

! PROCESS ONE ROW

! We read each word, expand bytes (upper to lower) into words and
! have the output state machine store them.
expand16xRowLoop:
	ldub	[rSrcAddr+rSrcOffset], rInput	! get current source byte
expand16xRowLoop1:
	add	rSrcOffset, 0x1, rSrcOffset	! pt to next one
	sll	rInput, 2, rInput		! make it point to words

! put out rightmost expanded words
	add	rInput, 3*256*4, rScr0		! index the byte3 table
	ld	[rtabExpAddr+rScr0], rScr0
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 2, rWidth		! decrement width
! return from output state machine with cc's intact.  If we put
! the ld in the jmpl's delay slot, we'd get an interlock cycle.
	ble	expand16xFinRow			! quit if that was end of row
	add	rInput, 2*256*4, rScr0		! index the byte2 table

! put out next-to-rightmost expanded word
	ld	[rtabExpAddr+rScr0], rScr0
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 2, rWidth		! decrement width
! return from output state machine with cc's intact.  If we put
! the ld in the jmpl's delay slot, we'd get an interlock cycle.
	ble	expand16xFinRow			! quit if that was end of row
	add	rInput, 256*4, rScr0		! index the byte1 table

! put out next-to-leftmost expanded word
	ld	[rtabExpAddr+rScr0], rScr0
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 2, rWidth		! decrement width
! return from output state machine with cc's intact.  If we put
! the ld in the jmpl's delay slot, we'd get an interlock cycle.
	ble	expand16xFinRow			! quit if that was end of row

! put out leftmost expanded word
	ld	[rtabExpAddr+rInput], rScr0
	jmpl	rOState, rRetAddr		! and call output routine
	subcc	rWidth, 2, rWidth		! decrement width
! return from output state machine with cc's intact.  If we put
! the ld in the jmpl's delay slot, we'd get an interlock cycle.
	bg	expand16xRowLoop1		! keep looping if there's more
	ldub	[rSrcAddr+rSrcOffset], rInput	! get current source byte


! END OP LOOP ACROSS ONE ROW

! have finished with that row, are there more rows?
expand16xFinRow:
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point to next src row
	subcc	rHeight, 1, rHeight		! decrement height count
	bnz	expand16xImageLoop		! if more lines, keep going
	add	rDstAddr, rBlockPitch, rDstAddr	! with new DstAddr

! END OF LOOP ACROSS ALL ROWS

! We've put out the right number of words, so quit
expand16Fin:
	mov	0,%i0
	ret
	restore



! *************** 16x Expand Output State Machine  ***************


! This is a fake routine that we use to tell Purify to
! pay attention to the code that is otherwise only reached by
! indirect jumps.  Purify seems to need a branch to decide that
! a piece of code should be Purified.  Without these, the jump
! offsets in these routines are not changed and things blow up.
! Purify decides that things are really code if there are at least
! 4 normal opcodes before the first branch.  So put in some
! noops.  We don't really have to touch all the states since
! many of them have 4 ops at the beginning without a branch,
! but let's be safe (in case a branch appears later).
	.global	MK_GLOBAL_NAME(ip_fake16xDispTable)
MK_GLOBAL_NAME(ip_fake16xDispTable):
	nop
	nop
	nop
	nop
	deccc	%i0
	bz	exp16x_L_EP_EA
	deccc	%i0
	bz	exp16x_L_EP_OA
	deccc	%i0
	bz	exp16x_L_OP_EA
	deccc	%i0
	bz	exp16x_L_OP_OA
	deccc	%i0
	bz	exp16x_M_EP_EA
	deccc	%i0
	bz	exp16x_M_EP_OA
	deccc	%i0
	bz	exp16x_M_OP_EA
	deccc	%i0
	bz	exp16x_M_OP_OA
	deccc	%i0
	bz	exp16x_R_EP_EA
	deccc	%i0
	bz	exp16x_R_EP_OA
	deccc	%i0
	bz	exp16x_R_OP_EA
	deccc	%i0
	bz	exp16x_R_OP_OA
	nop
	ret
	restore


!------------------------------------------
! exp16x_L_EP_EA

! Left state, Even Pitch, Even Starting address
! Just save the output, it will be written as part of a double
! write next time.  If dest is 1 word wide, do all writes now.
exp16x_L_EP_EA:
	ble	exp16x_L_EP_OA
	mov	rScr0, rOutE0
	sethi	%hi(exp16x_M_EP_OA), rOState
	retl
	or	rOState, %lo(exp16x_M_EP_OA), rOState


!------------------------------------------
! exp16x_L_EP_OA

! Left state, Even Pitch, Odd Starting address
! Write output to 8 lines
! DO NOT advance rDstOffset here. It is supposed to be 1 "behind"
! so that the next double write (in two words) will work.

exp16x_L_EP_OA:
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 8 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 9 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 10 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 11 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 12 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 13 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 14 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 15 */
	sethi	%hi(exp16x_M_EP_EA), rOState
	retl
	or	rOState, %lo(exp16x_M_EP_EA), rOState


!------------------------------------------
! exp16x_L_OP_EA

! Left state, Odd Pitch, Even Starting address
! Write output to odd lines, save for even lines next

exp16x_L_OP_EA:
	ble	exp16x_L_EP_OA
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 9 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 11 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 13 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 15 */
	mov	rScr0, rOutE0
	sethi	%hi(exp16x_M_OP_OA), rOState
	retl
	or	rOState, %lo(exp16x_M_OP_OA), rOState


!------------------------------------------
! exp16x_L_OP_OA

! Left state, Odd Pitch, Odd Starting address
! Write output to even lines, save for odd lines next

exp16x_L_OP_OA:
	ble	exp16x_L_EP_OA
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 8 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 10 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 12 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 14 */
	mov	rScr0, rOutO0
	sethi	%hi(exp16x_M_OP_EA), rOState
	retl
	or	rOState, %lo(exp16x_M_OP_EA), rOState


!------------------------------------------
! exp16x_M_EP_EA

! Middle state, Even Pitch, Even address
! save data for double writes on all lines later, inc rDstOffset

exp16x_M_EP_EA:
	ble	exp16x_R_EP_EA
	mov	rScr0, rOutE0
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp16x_M_EP_OA), rOState
	retl
	or	rOState, %lo(exp16x_M_EP_OA), rOState


!------------------------------------------
! exp16x_M_EP_OA

! Middle state, Even Pitch, Odd address
! do double writes on all lines, inc rDstOffset

exp16x_M_EP_OA:
	ble	exp16x_R_EP_OA
	mov	rScr0, rOutE1
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 8 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 9 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 10 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 11 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 12 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 13 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 14 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 15 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp16x_M_EP_EA), rOState
	retl
	or	rOState, %lo(exp16x_M_EP_EA), rOState


!------------------------------------------
! exp16x_M_OP_EA

! Middle state, Odd Pitch, Even address
! save data for double write on even lines next, do double write on odd lines,
! inc rDstOffset

exp16x_M_OP_EA:
	ble	exp16x_R_OP_EA
	mov	rScr0, rOutO1
	mov	rScr0, rOutE0
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 9 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 11 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 13 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 15 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp16x_M_OP_OA), rOState
	retl
	or	rOState, %lo(exp16x_M_OP_OA), rOState


!------------------------------------------
! exp16x_M_OP_OA

! Middle state, Odd Pitch, Odd address
! save data for double write on odd lines next, do double write on even lines,
! inc rDstOffset

exp16x_M_OP_OA:
	ble	exp16x_R_OP_OA
	mov	rScr0, rOutE1
	mov	rScr0, rOutO0
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 8 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 10 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 12 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 14 */
	add	rDstOffset, 0x4, rDstOffset
	sethi	%hi(exp16x_M_OP_EA), rOState
	retl
	or	rOState, %lo(exp16x_M_OP_EA), rOState


!------------------------------------------
! exp16x_R_EP_EA

! Right state, Even Pitch, Even address
! do single writes on all lines now, don't bother with rDstOffset or next
! state

exp16x_R_EP_EA:
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 8 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 9 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 10 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 11 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 12 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 13 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 14 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 15 */


!------------------------------------------
! exp16x_R_EP_OA

! Right state, Even Pitch, Odd address
! do double writes on all lines now, don't bother with rDstOffset or next
! state. Output data is already in rOutE1.

exp16x_R_EP_OA:
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 8 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 9 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 10 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 11 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 12 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 13 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 14 */
	add	rDstAddrX, rDstPitch, rDstAddrX
	retl
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 15 */


!------------------------------------------
! exp16x_R_OP_EA

! Right state, Odd Pitch, Even address
! do single writes on even lines, double write on odd lines,
! don't bother with rDstOffset or next state.  Data is already in rOutO1.

exp16x_R_OP_EA:
	add	rDstAddr, rDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 9 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 11 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 13 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutO0, [rDstAddrX+rDstOffset]		/* line 15 */
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	st	rScr0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 8 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 10 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 12 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 14 */


!------------------------------------------
! exp16x_R_OP_OA

! Right state, Odd Pitch, Odd address
! do single writes on odd lines, double write on even lines,
! don't bother with rDstOffset or next state.  Data is already in rOutE1.

exp16x_R_OP_OA:
	std	rOutE0, [rDstAddr+rDstOffset]		/* line 0 */
	add	rDstAddr, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 2 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 4 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 6 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 8 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 10 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 12 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	std	rOutE0, [rDstAddrX+rDstOffset]		/* line 14 */
	add	rDstOffset, 0x4, rDstOffset	/* was pting to prev word */
	add	rDstAddr, rDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 1 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 3 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 5 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 7 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 9 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 11 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 13 */
	add	rDstAddrX, r2xDstPitch, rDstAddrX
	retl
	st	rScr0, [rDstAddrX+rDstOffset]		/* line 15 */




/*

$Log: ssexpand.s,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.15  1998/01/13 23:57:35  ddavies
ssexpand.s -> /product/ipcore/ipshared/src/RCS/ssexpand.s,v

CommonCheck: Add IPState parameter to almost all private routines.

! Revision 1.14  1994/10/27  02:42:37  ddavies
! ddavies on Wed Oct 26 19:40:34 PDT 1994
!
! ssexpand.s -> /product/ipcore/ipshared/src/RCS/ssexpand.s,v
!
! Given new insight into how Purify decides blocks of memory are code,
! add 4 nops to the beginning of the fake procedures (they like to see four
! real opcodes before deciding something is code).  Also reference EVERY
! state entry point since some others of them had branches at the beginning
! (hence didn't look like code!).  This seems to have fixed the problems
! that gave us incorrect data in expandAndMask ops.
!
! Revision 1.13  1994/10/25  22:34:33  ddavies
! ddavies on Tue Oct 25 15:33:54 PDT 1994
!
! ssexpand.s -> /product/ipcore/ipshared/src/RCS/ssexpand.s,v
!
! Made names of Purify workaround routines begin with ip_fake... instead of
! Fake... to conform with naming convention.
!
! Revision 1.12  1994/10/25  03:14:30  ddavies
! ddavies on Mon Oct 24 20:11:29 PDT 1994
!
! ssexpand.s -> /product/ipcore/ipshared/src/RCS/ssexpand.s,v
!
! Fixed two interaction problems with Purify.  The output states calculated
! the address of the next output state by subtracting the next state's address
! from the current state`s address.  Purify had no way of knowing this was
! happening, so didn't modify this constant.  We now use the normal
! sethi/or sequence and it is recognized.
! Purify also only purifies code that it thinks is reachable.  Since the
! output states were only reached via indirect jumps, it decided they weren't
! reachable and didn't touch them.  That meant that not only were they not
! checked, but their branches were not modified.  They branched into the
! middles of other routines.  Installed a fake routine for each state machine
! that pretends to branch to all the state machine entry points.
!
! Revision 1.11  1994/07/27  15:53:27  ejaquith
! ssexpand.s -> /product/ipcore/ipshared/src/RCS/ssexpand.s,v
!
! Reentrancy!
! Removed globals and made functions be called with pointers to expansion
! tables instead.
!
! Revision 1.10  1993/12/04  04:59:42  ddavies
! ssexpand.s -> /product/ipcore/ipshared/src/RCS/ssexpand.s,v
!
! Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
!
! Revision 1.9  1993/11/05  02:23:49  ddavies
! ssexpand.s -> /product/ipcore/ipshared/src/RCS/ssexpand.s,v
!
! Add RCS magic to have string specifying this file's name, rev number and
! last edit date into the object file.  Also add change log at the bottom.
!

*/


