/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: ssfloyd.s
 *
 * This file contains assembly code to dither one horizontal block of gray
 * pixels. It uses the Floyd-Steinberg error diffusion method.  It implements
 * much the same function as the routine ip_ditherFloydLine in mifloyd.c.
 * i_floyd.c (which calls this routine) contains massive comments on basic
 * methodology.
 *
 * An addition to the original module is ip_ditherFloydLine2.  This routine,
 * like the implementation in mifloyd.c, generates 2 bit/pixel output.
 *
 * The biggest difference between the C and assembler versions of
 * ip_ditherFloydLine is the number of lines of dithered code being produced.
 * The C routine produces 2 lines of dithered code on each pass, this
 * routine produces 7.  We define the number of lines of dithered data
 * produced by a pass as DBLOCK_HEIGHT (see asfloyd.c/mifloyd.c). Increasing
 * DBLOCK_HEIGHT reduces the overhead of reading and writing lines
 * of modified gray pixels.  One such line must be written for each pass
 * and handed to the next pass.
 *
 * To avoid repetition, I'll try to keep these comments centered on the aspects
 * of the implementation peculiar to assembly language.  From that point of
 * view, how does this thing work?
 *
 * When called, this routine loads up the register pointers in the normal
 * fashion.  The DRCTable, HDitherErrorTable, VDitherErrorTable, and
 * addresses are passed in.  (HDitherErrorTable2 and VDitherErrorTable2 get
 * passed to ip_ditherFloydLine2).
 * Because we ran out of registers, we don't keep the pointer to the
 * truncate table or the destination pitch in dedicated registers.  They're
 * loaded as needed.  These are the least referenced values and they don't
 * change.
 *
 * Next, the routines load the registers holding gray data from the first 7
 * lines.  We load the topmost line's (line0) register with the three pixels
 * in the upper left corner of the rectangle being loaded.  Each subsequent
 * register is loaded with the set of 3 pixels positioned down one line and
 * to the *left* by 1 pixel.  For example, the second gray register is loaded
 * with the leftmost 2 pixels from the second line and the pixel to the left
 * of that.  The third register is loaded with two pixels to the left of the
 * rectangle and one in the rectangle, and so on.  We use pixels from outside
 * the rectangle as our best approximation of surrounding gray scale data.
 * The errors from these pixels can diffuse into the rectangle, so we shouldn't
 * use random data.
 *
 * Like the C implementations, we use marker bits in the digital
 * registers to determine when they should be written.  When this marker bit
 * falls off the right end after a shift, we know the output register is full
 * so it's time to do a write.  How are these flags initialized?  The topmost
 * output register should be written after accumulating 32 or 16 pixel, so
 * we set either bit 0 (ip_ditherFloydLine) or bit 1 (ip_ditherFloydLine2).
 * However, the first pixel accumulated by the next output register is trash,
 * so this register should be written after accumulating 33 pixels for
 * ip_ditherFloydLine or 17 pixels for ip_ditherFloydLine2.  We can't set bit
 * -1 of this register.  Similarly, the register for line 2 should be written
 * after accumulating 34 or 18 pixels and so on.
 *
 * We intialize the other output registers to 0.  When the topmost register's
 * flag bit falls off the top end, we know the data in the other output
 * registers is in position to have the flag bits inserted.  We mask off
 * the bogus data in the upper portions of those registers and insert
 * flag bits into the correct (one pixel to the right for each line we move
 * down) positions.  Thereafter, just after the contents of a digital output
 * register is written to memory, that register's flag bit (bit 0 or bit 1)
 * is set.  Doing all this silliness with flag bits allows us to avoid having
 * the code for each output line test the bottom bits of loopCt for a
 * particular pattern. 
 *
 * After initialization, we enter the main loop.  The memory location that would
 * normally hold rWidth is modified to hold {w + (DBLOCK_HEIGHT-1) +
 * ((32 - (w % 32)) & 0x1F)}.  In ip_ditherFloydLine2, we use {w +
 * (DBLOCK_HEIGHT-1) + ((16 - (w % 16)) & 0xF)}.  These bizarre numbers are
 * the number of times we'll go through the loop that computes
 * DBLOCK_HEIGHT dithered lines of data plus one modified gray line.  Where do
 * they come from?
 *
 * The "w" term is the width of the image in pixels.
 *
 * The "DBLOCK_HEIGHT-1" term specifies the number of extra cycles required
 * to guarantee that the lowest line producing digital data has processed it's
 * rightmost pixel.
 *
 * The "((32 - (w % 32)) & 0x1F)" or "((16 - (w % 16)) & 0xF)" term has us
 * continue until the lowest line gets to a word boundary (so the last valid
 * pixels are written).  This is equivalent to rounding (w + DBLOCK_HEIGHT-1)
 * up to the next multiple of 32 or 16.
 *
 * The rest of the code is pretty straightforward (assuming you understand
 * the basic scheme described in i_floyd.c).  The usual assembly code
 * register dance provides some excitement.
 *
 * One of the goals of the assembly code is to process many lines at once.
 * Each line requires 2 registers, one holding gray data and
 * one holding processed digital data.  In addition, we need pointers to
 * the GrayBuf, the source image, the destination image and the four tables.
 * All of those addresses are use fairly often, so we'd like to keep them in
 * registers.  Ideally, we'd like to have a source and address pointer for
 * every line of processed data, but that's too many registers.
 *
 * Let's talk about the pointers to the GrayBuf first.  pGrayBuf points to
 * the beginning of the buffer, assumed to be on a 4 byte boundary.  The
 * GrayBuf holds both partially modified pixels from the previous block and
 * pixels going to the next block.  GrayBuf pixels are consumed by the top
 * line (line0) and produced by the bottom line (line7).  Line0 reads GrayBuf
 * ahead of Line7`s writes.  The naive implementation would have both
 * source and destination pointers, but that uses too many registers (or
 * too many reads and writes of local variables).  Instead, we compute the
 * distance between the read and write pointers.  In fact, we arrange things
 * so this distance is a constant.
 *
 * Line7 writes a word (4 bytes) into GrayBuf starting at location 0 every
 * time it has a word ready.  Since Line7 starts DBLOCK_HEIGHT bytes to the
 * left of the rectangle, the first DBLOCK_HEIGHT bytes in GrayBuf are
 * trash.  This trash also provides some room between the writer and reader.
 * So, we initially increment rGrayBuf by DBLOCK_HEIGHT.  We then read 3 bytes
 * and put them into rGLine0.  We must process 4 pixels before preparing to
 * write.  Therefore, the destination pointer = (modified rGrayBuf) -
 * DBLOCK_HEIGHT - 3 - 4.
 *
 * The source image pointer is also interesting.  As we move down from line
 * to line in the code, we want to add rSrcPitch, then subtract 1 to get the
 * address of the next line's source byte.  We actually just add rSrcPitch
 * and use increasingly negative address offsets.  At the bottom of the loop,
 * we subtract (DBLOCK_HEIGHT-2)*rSrcPitch, and add 1.  Remember that
 * rSrcAddr points to source data for line *1*, not line 0.  Line 0's data
 * comes from GrayBuf.  Therefore. the distance from the bottom line to
 * line 1 is DLBLOCK_HEIGHT-2.
 *
 * The destination pointer is a bit different.  All destination writes are
 * done on word boundaries.  Data from line N must be written directly
 * under that for line N-1 even though the pixels in rGLineN are to the left
 * of those in rGLine(N-1).  We do this by only writing dithered data for a
 * line when that line's register is full.  For example, the code will wait for
 * 32 dithered bits to accumulate in rDLine0, then write them.  Say this
 * happens in pass i through the loop.  In pass i+1 through the loop,
 * rDLine1 is written to memory.  In pass i+2 through the loop, rDLine2 is
 * written, and so on.  Therefore, the destination pointer goes straight
 * down (+rDstPitch) for DBLOCK_HEIGHT passes through the loop, then goes
 * back to the top (- (DBLOCK_HEIGHT-1)*rDstPitch) and over 4 bytes.
 *
 * The method used by ip_ditherFloydLine to compute the dithered data has
 * changed from the C implementation.  We grab the pixel to be thresholded
 * first, get the horizontal and vertical dither values and add them to the
 * appropriate registers.  However, current Gray data register is not shifted
 * before the add as was done in the C implementation.  The horizontal table
 * is modified to contain bits at the upper end that cause the 7/16 error add
 * to set the carry flag if and only if the binary data should be white.  It
 * would have been better to set the carry flag if the binary data were
 * black, but I can't figure out how to do that reliably.  For example, if the
 * thresholded pixel were 0, there's nothing we can add to it to get a carry.
 * Anyway, after the 7/16 error add, we add the digital reg to itself with
 * carry in.  This shifts in the inverse of the required data.  This also
 * shifts the marker bit left by 1.  If the marker bit goes into the carry,
 * we invert the dithered data reg and write it to memory.  The combination
 * of the modified horizontal table and the marker bit saves us 5 opcodes
 * (about 25%) in each line's code.
 *
 * In ip_ditherFloydLine2, the same tables are used by the C and assembly
 * implementations.  There are a couple of reasons for this.  First, shifting
 * two bits through the condition codes takes more cycles anyway.  Second,
 * there's nothing we can add to the upper bits of a pixel that always
 * generates the correct carry and MSB.  For example, say the pixel is
 * negative.  The correct digital output value should be 00.  Therefore, we
 * need something that can be added to the pixel that flips the top bit to
 * 0 and *does not* set the carry.  But if 1 in the MSB flips, there will be
 * a carry, so that doesn't work.  Say we're willing to invert all the digital
 * data afterwords.  Then a pixel of value 0 doesn't work.  Under this scheme,
 * it should produce a value of 11, so we'd need to add something that would
 * set the carry, 0 + <anything> leaves the carry unset.
 *
 * Instead, we do a couple of shifts and an or just like the C implementation.
 * The 2 bit shift of the digital output register is done with 2 adds so we
 * can detect the full flag falling into the carry.
 *
 * Even with all that convoluted cleverness, we still needed to use a couple
 * of registers twice.  As stated above, we compute a modified width and
 * store it into rWidth's memory location.  rWidth is then reused as rScr.
 * Since we don't modify rDstAddr too often, we reuse rDstPitch as
 * the pointer to the table used to truncate 10 bit pixels to 8 bit pixels.
 *
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssfloyd.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	.align	4

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define rIPStateIn	%i0
#define	rSrcAddrIn	%i1
#define rSrcPitchIn	%i2
#define rDstAddrIn	%i3
#define rDstPitchIn	%i4
#define rGrayBufIn	%i5

#define	rSrcAddr	%i0		/* current addr in source image */
#define rSrcPitch	%i1		/* source image's bytes/line */
#define rDstAddr	%i2		/* current addr in dest image */
#define rDstPitch	%i3		/* dest image's bytes/line */
#define rGrayBuf	%i4		/* ptr to line of modified pixels */
#define rWidth		%i5		/* width of rectangle in pixels */

#define	rDRCTable	%g1		/* ptr to Dynamic Range Correction
					 * Table */
#define	rHDErrorTab	%g2		/* ptr to error table holding 7/16
					 * of diffusion errors */
#define	rVDErrorTab	%g3		/* ptr to error table holding
					 * {3/16 : 5/16 : 1/16} of error */
	
#define rScr		%i3		/* scratch reg. SAME AS rDstPitch! */
#define rScr1		%i5		/* scratch reg. SAME AS rWidth! */
#define rCount		%o7		/* records current position relative
					 * to left edge of line */

#define rGLine0		%g4		/* Gray pixels, top line */
#define	rDLine0		%g5		/* binary pixels, top line */
#define	rGLine1		%o0
#define rDLine1		%o1
#define	rGLine2		%o2
#define rDLine2		%o3
#define	rGLine3		%o4
#define rDLine3		%o5
#define	rGLine4		%l0
#define rDLine4		%l1
#define	rGLine5		%l2
#define rDLine5		%l3
#define	rGLine6		%l4
#define rDLine6		%l5
#define	rGLine7		%l6		/* Gray pixels, bottom line */
#define rGLine7Out	%l7		/* output pixels, bottom line */



/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	srcAddrOffset	0x48
#define	srcPitchOffset	0x4c
#define	dstAddrOffset	0x50
#define	dstPitchOffset	0x54
#define	grayBufOffset	0x58
#define	widthOffset	0x5c
#define hdeTableOffset	0x64            /* addr of horizontal error table
					 * This offset is +8 from the previous
					 * because the parameter before it
					 * passed in gets ignored */
#define vdeTableOffset	0x68            /* addr of vertical error table */
#define dtTableOffset	0x6c            /* addr of pixel truncation table */
#define drcTableOffset	0x70		/* addr of DRCTable */
#define pixelOffset	0x78            /* This offset is +8 from the previous
	                                 * because one parameter passed in 
	                                 * gets loaded indirectly */
	

/* Addresses of local variables relative to frame pointer (fp) */
#define realWidthOffset -0x4
#define srcFixupOffset	-0x8
#define	dstFixupOffset	-0xC
#define srcFixupOffsetOver1	-0x10

/* magical C Frame constant (a few local variables) */
#define	LF26		-80

/* Number of lines of dithered data produced in a pass.
 * DO NOT CHANGE without changing arithmetic used to advance rSrcAddr
 * and rDstAddr at the bottom of the loop!! */
#define	DBLOCK_HEIGHT		7
#define GRAY_DEST_OFFSET 	(DBLOCK_HEIGHT+7)

#define lDRCOffset		0x100

LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! _ip_ditherFloydLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_ditherFloydLine)
MK_GLOBAL_NAME(ip_ditherFloydLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  As stated above
! we also use some of this stuff.
	st	rIPStateIn,  [%fp+ipStateOffset]
	st	rSrcAddrIn,  [%fp+srcAddrOffset]
	st	rSrcPitchIn, [%fp+srcPitchOffset]
	st	rDstAddrIn,  [%fp+dstAddrOffset]
	st	rDstPitchIn, [%fp+dstPitchOffset]
	st	rGrayBufIn,  [%fp+grayBufOffset]

	ld	[%fp+srcAddrOffset],  rSrcAddr
	ld	[%fp+srcPitchOffset], rSrcPitch
	ld	[%fp+dstAddrOffset],  rDstAddr
	ld	[%fp+dstPitchOffset], rDstPitch
	ld	[%fp+grayBufOffset],  rGrayBuf
	ld	[%fp+widthOffset],    rWidth
	ld	[%fp+drcTableOffset], rDRCTable
	
! compute the modified width (number of times we go through the loop).  We
! add DBLOCK_HEIGHT-1 because we actually test rCount against this count
! before incrementing rCount, but start rCount at 1 instead of 0.
	add	rWidth, 0x1F, rWidth		! round up to next mult of 32
	andn	rWidth, 0x1F, rWidth
	add	rWidth, DBLOCK_HEIGHT-1, rWidth	! = additional cycles
	st	rWidth, [%fp+realWidthOffset]

! construct constants used to modify rSrcAddr and rDstAddr at the bottom
! of the loop.  REMEMBER THAT rDstPitch IS THE SAME AS rScr!!!
	sll	rSrcPitch, 2, rScr1		! 4 lines
	add	rSrcPitch, rScr1, rScr1		! 5 lines
	add	rSrcPitch, rScr1, rScr1		! 6 lines
	sub	rScr1, 1, rScr1			! and over 1 pixel
	st	rScr1, [%fp+srcFixupOffset]

	sll	rDstPitch, 2, rScr1		! 4 lines
	add	rScr1, rDstPitch, rScr1		! 5 lines
	sub	rScr1, 4, rScr1			! and over 4 bytes
	st	rScr1, [%fp+dstFixupOffset]

! dereference pointers to the two most referenced table parameters
	ld	[%fp+hdeTableOffset],	rHDErrorTab
	ld	[%fp+vdeTableOffset],	rVDErrorTab

! Load data into the gray registers.  Start with the top line from GrayBuf.
	ldub	[rGrayBuf + DBLOCK_HEIGHT+0], rScr
	ldub	[rGrayBuf + DBLOCK_HEIGHT+1], rScr1
	ldub	[rGrayBuf + DBLOCK_HEIGHT+2], rGLine0
	add	rGrayBuf, DBLOCK_HEIGHT+3, rGrayBuf
	sll	rScr, 22, rScr			! align first byte
	or	rGLine0, rScr, rGLine0
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine0, rScr1, rGLine0

! Get raw data from the image.  Each byte must be corrected individually.

! Load rGLine1
	ldub	[rSrcAddr - 1], rScr
	ldub	[rSrcAddr + 0], rScr1
	ldub	[rSrcAddr + 1], rGLine1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine1], rGLine1
	sll	rScr, 22, rScr			! align first byte
	or	rGLine1, rScr, rGLine1
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine1, rScr1, rGLine1

! Load rGLine2
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 2], rScr
	ldub	[rSrcAddr - 1], rScr1
	ldub	[rSrcAddr + 0], rGLine2
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine2], rGLine2
	sll	rScr, 22, rScr			! align first byte
	or	rGLine2, rScr, rGLine2
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine2, rScr1, rGLine2

! Load rGLine3
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 3], rScr
	ldub	[rSrcAddr - 2], rScr1
	ldub	[rSrcAddr - 1], rGLine3
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine3], rGLine3
	sll	rScr, 22, rScr			! align first byte
	or	rGLine3, rScr, rGLine3
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine3, rScr1, rGLine3

! Load rGLine4
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 4], rScr
	ldub	[rSrcAddr - 3], rScr1
	ldub	[rSrcAddr - 2], rGLine4
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine4], rGLine4
	sll	rScr, 22, rScr			! align first byte
	or	rGLine4, rScr, rGLine4
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine4, rScr1, rGLine4

! Load rGLine5
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 5], rScr
	ldub	[rSrcAddr - 4], rScr1
	ldub	[rSrcAddr - 3], rGLine5
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine5], rGLine5
	sll	rScr, 22, rScr			! align first byte
	or	rGLine5, rScr, rGLine5
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine5, rScr1, rGLine5

! Load rGLine6
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 6], rScr
	ldub	[rSrcAddr - 5], rScr1
	ldub	[rSrcAddr - 4], rGLine6
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine6], rGLine6
	sll	rScr, 22, rScr			! align first byte
	or	rGLine6, rScr, rGLine6
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine6, rScr1, rGLine6

! There's no need to initialize rGLine7.  It is never used to compute
! thresholds, hence is not a source for error diffusion.
! You'd think we would put the marker bits into the rDLinex registers here,
! but NO.  rDLine0 could use a 1, but the other registers should have their
! bits placed farther off their right ends.  Those registers need to get
! through the trash at the beginning of the rectangle before starting to
! accumulate real data.  So, we'll initialize rDLine0 here.  When rDLine0's
! marker bit shifts into the carry, we'll also test rCount.  If it's equal
! to 32, we'll set the marker bits in the other rDLineX registers.
! However, DO initialize these registers so initial trash doesn't cause
! spurious writes.
	mov	1, rDLine0
	mov	0, rDLine1
	mov	0, rDLine2
	mov	0, rDLine3
	mov	0, rDLine4
	mov	0, rDLine5
	mov	0, rDLine6

! restore rSrcAddr
	sll	rSrcPitch, 2, rScr		! compute 4*rSrcPitch
	sub	rSrcAddr, rScr, rSrcAddr	! now down by 1 lines
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! points to line1 again
	add	rSrcAddr, 2, rSrcAddr		! point to first new pixel


! Initialize rCount for use in the main loop
	mov	1, rCount

! ****************** MAIN LOOP **************************

dFloydRowLoop:

! ********** LINE 0

! Get threshold pixel for line 0
	srl	rGLine0, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine0, rScr1, rGLine0		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine0, rDLine0, rDLine0	! did marker 1 shift out?
	bcc	dFloydThreshLine0		! no, skip the write
	add	rGLine1, rScr, rGLine1		! apply {3/16 5/16 1/16} error

! time to do a write
	not	rDLine0				! invert dithered data
	st	rDLine0, [rDstAddr]
	xorcc	rCount, 32, %g0			! was that the first time?
	bnz	dFloydThreshLine0		! no, just keep going.
	mov	1, rDLine0			! turn on marker bit again

! We have just written the first 32 bits of the top line.  That means the
! other data bits will be ready soon.  Turn on their marker bits.  We couldn't
! do this before because the correct positions for the marker bits weren't
! in the registers.  For example, rDLine1 should be written after the first
! 33 times through the loop, but there are only 32 bits in a register.
! BUG FIX:  Just because we cleared all the rDLineX registers doesn't mean
! they don't have trash in them now.  We've done 32 cycles.  The MS bit in
! rDLine1 is trash because it was derived from data to the left of the
! rectangle.  The MS 2 bits in rDLine2 are trash for the same reason and
! so on.  We not only need to set the marker bits in all these registers,
! we need to clear out the trash bits.
	sethi	%hi(0x80000000), rScr		! prepare trash mask
	sethi	%hi(0x40000000), rScr1		! prepare rDLine1's marker bit
	andn	rDLine1, rScr, rDLine1		! clear trash bit
	or	rDLine1, rScr1, rDLine1		! set marker bit

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine2, rScr, rDLine2
	or	rDLine2, rScr1, rDLine2

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine3, rScr, rDLine3
	or	rDLine3, rScr1, rDLine3

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine4, rScr, rDLine4
	or	rDLine4, rScr1, rDLine4

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine5, rScr, rDLine5
	or	rDLine5, rScr1, rDLine5

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine6, rScr, rDLine6
	or	rDLine6, rScr1, rDLine6


! shift new pixel into rGLine0
dFloydThreshLine0:
	ldub	[rGrayBuf + 0], rScr1		! get next source pixel
	add	rGrayBuf, 1, rGrayBuf		! pt to next pixel
! DO NOT CONVERT this pixel, the pixels in GrayBuf were converted before
! being written there.
	sll	rGLine0, 11, rGLine0		! make room
	or	rGLine0, rScr1, rGLine0		! insert new byte

! ********** LINE 1

! Get threshold pixel for line 1
	srl	rGLine1, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine1, rScr1, rGLine1		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine1, rDLine1, rDLine1	! did marker 1 shift out?
	bcc	dFloydThreshLine1		! no, skip the write
	add	rGLine2, rScr, rGLine2		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine1				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 1
	st	rDLine1, [rDstAddr]
	mov	1, rDLine1			! turn on marker bit again

! shift new pixel into rGLine1
dFloydThreshLine1:
	ldub	[rSrcAddr - 0], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine1, 11, rGLine1		! make room
	or	rGLine1, rScr1, rGLine1		! insert new byte

! ********** LINE 2

! Get threshold pixel for line 2
	srl	rGLine2, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine2, rScr1, rGLine2		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine2, rDLine2, rDLine2	! did marker 1 shift out?
	bcc	dFloydThreshLine2		! no, skip the write
	add	rGLine3, rScr, rGLine3		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine2				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine2, [rDstAddr]
	mov	1, rDLine2			! turn on marker bit again

! shift new pixel into rGLine2
dFloydThreshLine2:
	ldub	[rSrcAddr - 1], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine2, 11, rGLine2		! make room
	or	rGLine2, rScr1, rGLine2		! insert new byte

! ********** LINE 3

! Get threshold pixel for line 3
	srl	rGLine3, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine3, rScr1, rGLine3		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine3, rDLine3, rDLine3	! did marker 1 shift out?
	bcc	dFloydThreshLine3		! no, skip the write
	add	rGLine4, rScr, rGLine4		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine3				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 3
	st	rDLine3, [rDstAddr]
	mov	1, rDLine3			! turn on marker bit again

! shift new pixel into rGLine3
dFloydThreshLine3:
	ldub	[rSrcAddr - 2], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine3, 11, rGLine3		! make room
	or	rGLine3, rScr1, rGLine3		! insert new byte

! ********** LINE 4

! Get threshold pixel for line 4
	srl	rGLine4, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine4, rScr1, rGLine4		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine4, rDLine4, rDLine4	! did marker 1 shift out?
	bcc	dFloydThreshLine4		! no, skip the write
	add	rGLine5, rScr, rGLine5		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine4				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 4
	st	rDLine4, [rDstAddr]
	mov	1, rDLine4			! turn on marker bit again

! shift new pixel into rGLine4
dFloydThreshLine4:
	ldub	[rSrcAddr - 3], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine4, 11, rGLine4		! make room
	or	rGLine4, rScr1, rGLine4		! insert new byte

! ********** LINE 5

! Get threshold pixel for line 5
	srl	rGLine5, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine5, rScr1, rGLine5		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine5, rDLine5, rDLine5	! did marker 1 shift out?
	bcc	dFloydThreshLine5		! no, skip the write
	add	rGLine6, rScr, rGLine6		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine5				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 5
	st	rDLine5, [rDstAddr]
	mov	1, rDLine5			! turn on marker bit again

! shift new pixel into rGLine5
dFloydThreshLine5:
	ldub	[rSrcAddr - 4], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine5, 11, rGLine5		! make room
	or	rGLine5, rScr1, rGLine5		! insert new byte

! ********** LINE 6

! Get threshold pixel for line 6
	srl	rGLine6, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine6, rScr1, rGLine6		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine6, rDLine6, rDLine6	! did marker 1 shift out?
	bcc	dFloydThreshLine6		! no, skip the write
	add	rGLine7, rScr, rGLine7		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine6				! invert dithered data
	st	rDLine6, [rDstAddr + rScr1]	! store to line 6
	ld	[%fp+dstFixupOffset], rScr1	! get amount by which rDstAddr
						! should be changed at bottom
						! of the loop
	mov	1, rDLine6			! turn on marker bit again
	sub	rDstAddr, rScr1, rDstAddr	! point to line 0 again

! shift new pixel into rGLine6
dFloydThreshLine6:
	ldub	[rSrcAddr - 5], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine6, 11, rGLine6		! make room
	or	rGLine6, rScr1, rGLine6		! insert new byte

! ********** LINE 7

! calculate the output gray data for line 7
	ld	[%fp+dtTableOffset], rScr1      ! ptr to truncate table
	srl	rGLine7, 22, rScr		! grab next output pixel
	ldub	[rScr1+rScr], rScr		! truncate it
	sll	rGLine7Out, 8, rGLine7Out	! make room for it
	or	rGLine7Out, rScr, rGLine7Out	! insert it

! is rGLine7Out full of gray data?  This will happen every 4th time through
! the loop ((rCount & 0x3) == 0).
	andcc	rCount, 0x3, rScr		! (rCount) % 3 == 0?
	bnz	dFloydShiftLine7		! nope, skip the write
	ld	[%fp+srcFixupOffset], rScr	! prepare to adjust rSrcAddr

! time to do a write
	st	rGLine7Out, [rGrayBuf - GRAY_DEST_OFFSET]

! shift rGLine7
dFloydShiftLine7:
	ldub	[rSrcAddr - 6], rScr1		! get next source pixel
	sub	rSrcAddr, rScr, rSrcAddr	! pt to line 1 and over 1 byte
	ldub	[rDRCTable + rScr1], rScr1	! convert next source pixel
	sll	rGLine7, 11, rGLine7		! make room
	or	rGLine7, rScr1, rGLine7		! insert new byte

! We're finished with one skewed column of data.  Test to see
! if we should keep going (rCount < stored value of rWidth)

	ld	[%fp+realWidthOffset], rScr1	! get end count
	cmp	rScr1, rCount			! is rCount bigger yet?
	bge	dFloydRowLoop
	inc	rCount				! increment rCount

! We shouldn't put out any more dithered data, but we should finish filling
! GrayBuf.  The top line is DBLOCK_HEIGHT bytes ahead of us now.  We need
! to convert at least DBLOCK_HEIGHT more bytes, then keep going until
! we reach the next multiple of 4.
	add	rCount, DBLOCK_HEIGHT+3, rGLine0 ! round up to next mult of 4
	andn	rGLine0, 3, rGLine0
! repair rSrcAddr
	add	rSrcAddr, rScr, rSrcAddr	! pt to line 7 again
	add	rSrcAddr, 1, rSrcAddr		! and advance 1 byte

	ld	[%fp+dtTableOffset], rDLine0    ! ptr to truncate table

! read and translate more bytes
dFloydFillGrayLoop:
	add	rGrayBuf, 1, rGrayBuf
	srl	rGLine7, 22, rScr		! grab next output pixel
	ldub	[rDLine0+rScr], rScr		! truncate it
	sll	rGLine7Out, 8, rGLine7Out	! make room for it
	or	rGLine7Out, rScr, rGLine7Out	! insert it

! is rGLine7Out full of gray data?  This will happen every 4th time through
! the loop ((rCount & 0x3) == 3).
	andcc	rCount, 0x3, %g0		! (rCount) % 3 == 0?
	bz,a	dFloydFillGrayShift		! yes, do the write
	st	rGLine7Out, [rGrayBuf-GRAY_DEST_OFFSET]

! shift rGLine7
dFloydFillGrayShift:
	ldub	[rSrcAddr - 6], rScr1		! get next source pixel
	add	rSrcAddr, 1, rSrcAddr		! pt to next byte to the right
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine7, 11, rGLine7		! make room
	or	rGLine7, rScr1, rGLine7		! insert new byte

! done yet?
	cmp	rGLine0, rCount
	bge	dFloydFillGrayLoop		! nope, keep going
	add	rCount, 1, rCount


! END OF LOOP ACROSS ROW

! We've put out the right number of words, so quit
dFloydFin:
	mov	0,%i0
	ret
	restore




! -----------------------------------------------------------------------------
! _ip_ditherFloydLine2 Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_ditherFloydLine2)
MK_GLOBAL_NAME(ip_ditherFloydLine2):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  As stated above
! we also use some of this stuff.
	st	rIPStateIn,  [%fp+ipStateOffset]
	st	rSrcAddrIn,  [%fp+srcAddrOffset]
	st	rSrcPitchIn, [%fp+srcPitchOffset]
	st	rDstAddrIn,  [%fp+dstAddrOffset]
	st	rDstPitchIn, [%fp+dstPitchOffset]
	st	rGrayBufIn,  [%fp+grayBufOffset]

	ld	[%fp+srcAddrOffset],  rSrcAddr
	ld	[%fp+srcPitchOffset], rSrcPitch
	ld	[%fp+dstAddrOffset],  rDstAddr
	ld	[%fp+dstPitchOffset], rDstPitch
	ld	[%fp+grayBufOffset],  rGrayBuf
	ld	[%fp+widthOffset],    rWidth
	ld	[%fp+drcTableOffset], rDRCTable
	
! compute the modified width (number of times we go through the loop).  We
! add DBLOCK_HEIGHT-1 because we actually test rCount against this count
! before incrementing rCount, but start rCount at 1 instead of 0.
	add	rWidth, 0xF, rWidth		! round up to next mult of 16
	andn	rWidth, 0xF, rWidth
	add	rWidth, DBLOCK_HEIGHT-1, rWidth	! = additional cycles
	st	rWidth, [%fp+realWidthOffset]

! construct constants used to modify rSrcAddr and rDstAddr at the bottom
! of the loop.  REMEMBER THAT rDstPitch IS THE SAME AS rScr!!!
	sll	rSrcPitch, 2, rScr1		! 4 lines
	add	rSrcPitch, rScr1, rScr1		! 5 lines
	add	rSrcPitch, rScr1, rScr1		! 6 lines
	sub	rScr1, 1, rScr1			! and over 1 pixel
	st	rScr1, [%fp+srcFixupOffset]

	sll	rDstPitch, 2, rScr1		! 4 lines
	add	rScr1, rDstPitch, rScr1		! 5 lines
	add	rScr1, rDstPitch, rScr1		! 6 lines
	sub	rScr1, 4, rScr1			! and over 4 bytes
	st	rScr1, [%fp+dstFixupOffset]

! dereference pointers to the two most referenced table parameters
	ld	[%fp+hdeTableOffset],	rHDErrorTab
	ld	[%fp+vdeTableOffset],	rVDErrorTab

! Load data into the gray registers.  Start with the top line from GrayBuf.
	ldub	[rGrayBuf + DBLOCK_HEIGHT+0], rScr
	ldub	[rGrayBuf + DBLOCK_HEIGHT+1], rScr1
	ldub	[rGrayBuf + DBLOCK_HEIGHT+2], rGLine0
	add	rGrayBuf, DBLOCK_HEIGHT+3, rGrayBuf
	sll	rScr, 22, rScr			! align first byte
	or	rGLine0, rScr, rGLine0
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine0, rScr1, rGLine0

! Get raw data from the image.  Each byte must be corrected individually.

! Load rGLine1
	ldub	[rSrcAddr - 1], rScr
	ldub	[rSrcAddr + 0], rScr1
	ldub	[rSrcAddr + 1], rGLine1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine1], rGLine1
	sll	rScr, 22, rScr			! align first byte
	or	rGLine1, rScr, rGLine1
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine1, rScr1, rGLine1

! Load rGLine2
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 2], rScr
	ldub	[rSrcAddr - 1], rScr1
	ldub	[rSrcAddr + 0], rGLine2
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine2], rGLine2
	sll	rScr, 22, rScr			! align first byte
	or	rGLine2, rScr, rGLine2
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine2, rScr1, rGLine2

! Load rGLine3
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 3], rScr
	ldub	[rSrcAddr - 2], rScr1
	ldub	[rSrcAddr - 1], rGLine3
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine3], rGLine3
	sll	rScr, 22, rScr			! align first byte
	or	rGLine3, rScr, rGLine3
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine3, rScr1, rGLine3

! Load rGLine4
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 4], rScr
	ldub	[rSrcAddr - 3], rScr1
	ldub	[rSrcAddr - 2], rGLine4
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine4], rGLine4
	sll	rScr, 22, rScr			! align first byte
	or	rGLine4, rScr, rGLine4
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine4, rScr1, rGLine4

! Load rGLine5
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 5], rScr
	ldub	[rSrcAddr - 4], rScr1
	ldub	[rSrcAddr - 3], rGLine5
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine5], rGLine5
	sll	rScr, 22, rScr			! align first byte
	or	rGLine5, rScr, rGLine5
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine5, rScr1, rGLine5

! Load rGLine6
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 6], rScr
	ldub	[rSrcAddr - 5], rScr1
	ldub	[rSrcAddr - 4], rGLine6
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine6], rGLine6
	sll	rScr, 22, rScr			! align first byte
	or	rGLine6, rScr, rGLine6
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine6, rScr1, rGLine6

! There's no need to initialize rGLine7.  It is never used to compute
! thresholds, hence is not a source for error diffusion.
! You'd think we would put the marker bits into the rDLinex registers here,
! but NO.  rDLine0 could use a 2, but the other registers should have their
! bits placed farther off their right ends.  Those registers need to get
! through the trash at the beginning of the rectangle before starting to
! accumulate real data.  So, we'll initialize rDLine0 here.  When rDLine0's
! marker bit shifts into the carry, we'll also test rCount.  If it's equal
! to 32, we'll set the marker bits in the other rDLineX registers.
! However, DO initialize these registers so initial trash doesn't cause
! spurious writes.
	mov	2, rDLine0
	mov	0, rDLine1
	mov	0, rDLine2
	mov	0, rDLine3
	mov	0, rDLine4
	mov	0, rDLine5
	mov	0, rDLine6

! restore rSrcAddr
	sll	rSrcPitch, 2, rScr		! compute 4*rSrcPitch
	sub	rSrcAddr, rScr, rSrcAddr	! now down by 1 lines
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! points to line1 again
	add	rSrcAddr, 2, rSrcAddr		! point to first new pixel


! Initialize rCount for use in the main loop
	mov	1, rCount

! ****************** MAIN LOOP **************************

dFloyd2RowLoop:

! ********** LINE 0

! Get threshold pixel for line 0
	srl	rGLine0, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	add	rGLine0, rScr1, rGLine0		! apply 7/16 error

! shift dithered data into binary reg
	addcc	rDLine0, rDLine0, rDLine0	! did marker 1 shift out?
	add	rDLine0, rDLine0, rDLine0
	srl	rGLine0, 30, rScr1		! prepare digital input
	or	rDLine0, rScr1, rDLine0		! insert it
	bcc	dFloyd2ThreshLine0		! no, skip the write
	add	rGLine1, rScr, rGLine1		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	st	rDLine0, [rDstAddr]
	add	rDstAddr, rScr1, rDstAddr	! point to line 1
	xorcc	rCount, 16, %g0			! was that the first time?
	bnz	dFloyd2ThreshLine0		! no, just keep going.
	mov	2, rDLine0			! turn on marker bit again

! We have just written the first 32 bits of the top line.  That means the
! other data bits will be ready soon.  Turn on their marker bits.  We couldn't
! do this before because the correct positions for the marker bits weren't
! in the registers.  For example, rDLine1 should be written after the first
! 17 times through the loop (accumulating 17*2 = 34 bits), but there are only
! 32 bits in a register.
! BUG FIX:  Just because we cleared all the rDLineX registers doesn't mean
! they don't have trash in them now.  We've done 16 cycles.  The 2 MS bits in
! rDLine1 are trash because they were derived from data to the left of the
! rectangle.  The MS 4 bits in rDLine2 are trash for the same reason and
! so on.  We not only need to set the marker bits in all these registers,
! we need to clear out the trash bits.
	sethi	%hi(0xC0000000), rScr		! prepare trash mask
	sethi	%hi(0x20000000), rScr1		! prepare rDLine1's marker bit
	andn	rDLine1, rScr, rDLine1		! clear trash bit
	or	rDLine1, rScr1, rDLine1		! set marker bit

	sra	rScr, 2, rScr
	srl	rScr1, 2, rScr1
	andn	rDLine2, rScr, rDLine2		! clear top 4 bits
	or	rDLine2, rScr1, rDLine2

	sra	rScr, 2, rScr
	srl	rScr1, 2, rScr1
	andn	rDLine3, rScr, rDLine3		! clear top 6 bits
	or	rDLine3, rScr1, rDLine3

	sra	rScr, 2, rScr
	srl	rScr1, 2, rScr1
	andn	rDLine4, rScr, rDLine4		! clear top 8 bits
	or	rDLine4, rScr1, rDLine4

	sra	rScr, 2, rScr
	srl	rScr1, 2, rScr1
	andn	rDLine5, rScr, rDLine5		! clear top 10 bits
	or	rDLine5, rScr1, rDLine5

	sra	rScr, 2, rScr
	srl	rScr1, 2, rScr1
	andn	rDLine6, rScr, rDLine6		! clear top 12 bits
	or	rDLine6, rScr1, rDLine6


! shift new pixel into rGLine0
dFloyd2ThreshLine0:
	ldub	[rGrayBuf + 0], rScr1		! get next source pixel
	add	rGrayBuf, 1, rGrayBuf		! pt to next pixel
! DO NOT CONVERT this pixel, the pixels in GrayBuf were converted before
! being written there.
	sll	rGLine0, 11, rGLine0		! make room
	or	rGLine0, rScr1, rGLine0		! insert new byte

! ********** LINE 1

! Get threshold pixel for line 1
	srl	rGLine1, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	add	rGLine1, rScr1, rGLine1		! apply 7/16 error

! shift dithered data into binary reg
	addcc	rDLine1, rDLine1, rDLine1	! did marker 1 shift out?
	add	rDLine1, rDLine1, rDLine1
	srl	rGLine1, 30, rScr1		! prepare digital input
	or	rDLine1, rScr1, rDLine1		! insert it
	bcc	dFloyd2ThreshLine1		! no, skip the write
	add	rGLine2, rScr, rGLine2		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	st	rDLine1, [rDstAddr]
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	mov	2, rDLine1			! turn on marker bit again

! shift new pixel into rGLine1
dFloyd2ThreshLine1:
	ldub	[rSrcAddr - 0], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine1, 11, rGLine1		! make room
	or	rGLine1, rScr1, rGLine1		! insert new byte

! ********** LINE 2

! Get threshold pixel for line 2
	srl	rGLine2, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	add	rGLine2, rScr1, rGLine2		! apply 7/16 error

! shift dithered data into binary reg
	addcc	rDLine2, rDLine2, rDLine2	! did marker 1 shift out?
	add	rDLine2, rDLine2, rDLine2
	srl	rGLine2, 30, rScr1		! prepare digital input
	or	rDLine2, rScr1, rDLine2		! insert it
	bcc	dFloyd2ThreshLine2		! no, skip the write
	add	rGLine3, rScr, rGLine3		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	st	rDLine2, [rDstAddr]
	add	rDstAddr, rScr1, rDstAddr	! pt to line 3
	mov	2, rDLine2			! turn on marker bit again

! shift new pixel into rGLine2
dFloyd2ThreshLine2:
	ldub	[rSrcAddr - 1], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine2, 11, rGLine2		! make room
	or	rGLine2, rScr1, rGLine2		! insert new byte

! ********** LINE 3

! Get threshold pixel for line 3
	srl	rGLine3, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	add	rGLine3, rScr1, rGLine3		! apply 7/16 error

! shift dithered data into binary reg
	addcc	rDLine3, rDLine3, rDLine3	! did marker 1 shift out?
	add	rDLine3, rDLine3, rDLine3
	srl	rGLine3, 30, rScr1		! prepare digital input
	or	rDLine3, rScr1, rDLine3		! insert it
	bcc	dFloyd2ThreshLine3		! no, skip the write
	add	rGLine4, rScr, rGLine4		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	st	rDLine3, [rDstAddr]
	add	rDstAddr, rScr1, rDstAddr	! pt to line 4
	mov	2, rDLine3			! turn on marker bit again

! shift new pixel into rGLine3
dFloyd2ThreshLine3:
	ldub	[rSrcAddr - 2], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine3, 11, rGLine3		! make room
	or	rGLine3, rScr1, rGLine3		! insert new byte

! ********** LINE 4

! Get threshold pixel for line 4
	srl	rGLine4, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	add	rGLine4, rScr1, rGLine4		! apply 7/16 error

! shift dithered data into binary reg
	addcc	rDLine4, rDLine4, rDLine4	! did marker 1 shift out?
	add	rDLine4, rDLine4, rDLine4
	srl	rGLine4, 30, rScr1		! prepare digital input
	or	rDLine4, rScr1, rDLine4		! insert it
	bcc	dFloyd2ThreshLine4		! no, skip the write
	add	rGLine5, rScr, rGLine5		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	st	rDLine4, [rDstAddr]
	add	rDstAddr, rScr1, rDstAddr	! pt to line 5
	mov	2, rDLine4			! turn on marker bit again

! shift new pixel into rGLine4
dFloyd2ThreshLine4:
	ldub	[rSrcAddr - 3], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine4, 11, rGLine4		! make room
	or	rGLine4, rScr1, rGLine4		! insert new byte

! ********** LINE 5

! Get threshold pixel for line 5
	srl	rGLine5, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	add	rGLine5, rScr1, rGLine5		! apply 7/16 error

! shift dithered data into binary reg
	addcc	rDLine5, rDLine5, rDLine5	! did marker 1 shift out?
	add	rDLine5, rDLine5, rDLine5
	srl	rGLine5, 30, rScr1		! prepare digital input
	or	rDLine5, rScr1, rDLine5		! insert it
	bcc	dFloyd2ThreshLine5		! no, skip the write
	add	rGLine6, rScr, rGLine6		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	st	rDLine5, [rDstAddr]
	add	rDstAddr, rScr1, rDstAddr	! pt to line 6
	mov	2, rDLine5			! turn on marker bit again

! shift new pixel into rGLine5
dFloyd2ThreshLine5:
	ldub	[rSrcAddr - 4], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine5, 11, rGLine5		! make room
	or	rGLine5, rScr1, rGLine5		! insert new byte

! ********** LINE 6

! Get threshold pixel for line 6
	srl	rGLine6, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word index
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	add	rGLine6, rScr1, rGLine6		! apply 7/16 error

! shift dithered data into binary reg
	addcc	rDLine6, rDLine6, rDLine6	! did marker 1 shift out?
	add	rDLine6, rDLine6, rDLine6
	srl	rGLine6, 30, rScr1		! prepare digital input
	or	rDLine6, rScr1, rDLine6		! insert it
	bcc	dFloyd2ThreshLine6		! no, skip the write
	add	rGLine7, rScr, rGLine7		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstFixupOffset], rScr1	! get amount by which rDstAddr
						! should be changed at bottom
						! of the loop
	st	rDLine6, [rDstAddr]		! store to line 6
	sub	rDstAddr, rScr1, rDstAddr	! point to line 0 again
	mov	2, rDLine6			! turn on marker bit again

! shift new pixel into rGLine6
dFloyd2ThreshLine6:
	ldub	[rSrcAddr - 5], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine6, 11, rGLine6		! make room
	or	rGLine6, rScr1, rGLine6		! insert new byte

! ********** LINE 7

! calculate the output gray data for line 7
	ld	[%fp+dtTableOffset], rScr1      ! ptr to truncate table
	srl	rGLine7, 22, rScr		! grab next output pixel
	ldub	[rScr1+rScr], rScr		! truncate it
	sll	rGLine7Out, 8, rGLine7Out	! make room for it
	or	rGLine7Out, rScr, rGLine7Out	! insert it

! is rGLine7Out full of gray data?  This will happen every 4th time through
! the loop ((rCount & 0x3) == 0).
	andcc	rCount, 0x3, rScr		! (rCount) % 3 == 0?
	bnz	dFloyd2ShiftLine7		! nope, skip the write
	ld	[%fp+srcFixupOffset], rScr	! prepare to adjust rSrcAddr

! time to do a write
	st	rGLine7Out, [rGrayBuf - GRAY_DEST_OFFSET]

! shift rGLine7
dFloyd2ShiftLine7:
	ldub	[rSrcAddr - 6], rScr1		! get next source pixel
	sub	rSrcAddr, rScr, rSrcAddr	! pt to line 1 and over 1 byte
	ldub	[rDRCTable + rScr1], rScr1	! convert next source pixel
	sll	rGLine7, 11, rGLine7		! make room
	or	rGLine7, rScr1, rGLine7		! insert new byte

! We're finished with one skewed column of data.  Test to see
! if we should keep going (rCount < stored value of rWidth)

	ld	[%fp+realWidthOffset], rScr1	! get end count
	cmp	rScr1, rCount			! is rCount bigger yet?
	bge	dFloyd2RowLoop
	inc	rCount				! increment rCount

! We shouldn't put out any more dithered data, but we should finish filling
! GrayBuf.  The top line is DBLOCK_HEIGHT bytes ahead of us now.  We need
! to convert at least DBLOCK_HEIGHT more bytes, then keep going until
! we reach the next multiple of 4.
	add	rCount, DBLOCK_HEIGHT+3, rGLine0 ! round up to next mult of 4
	andn	rGLine0, 3, rGLine0
! repair rSrcAddr
	add	rSrcAddr, rScr, rSrcAddr	! pt to line 7 again
	add	rSrcAddr, 1, rSrcAddr		! and advance 1 byte

	ld	[%fp+dtTableOffset], rDLine0    ! ptr to truncate table

! read and translate more bytes
dFloyd2FillGrayLoop:
	add	rGrayBuf, 1, rGrayBuf
	srl	rGLine7, 22, rScr		! grab next output pixel
	ldub	[rDLine0+rScr], rScr		! truncate it
	sll	rGLine7Out, 8, rGLine7Out	! make room for it
	or	rGLine7Out, rScr, rGLine7Out	! insert it

! is rGLine7Out full of gray data?  This will happen every 4th time through
! the loop ((rCount & 0x3) == 3).
	andcc	rCount, 0x3, %g0		! (rCount) % 3 == 0?
	bz,a	dFloyd2FillGrayShift		! yes, do the write
	st	rGLine7Out, [rGrayBuf-GRAY_DEST_OFFSET]

! shift rGLine7
dFloyd2FillGrayShift:
	ldub	[rSrcAddr - 6], rScr1		! get next source pixel
	add	rSrcAddr, 1, rSrcAddr		! pt to next byte to the right
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine7, 11, rGLine7		! make room
	or	rGLine7, rScr1, rGLine7		! insert new byte

! done yet?
	cmp	rGLine0, rCount
	bge	dFloyd2FillGrayLoop		! nope, keep going
	add	rCount, 1, rCount


! END OF LOOP ACROSS ROW

! We've put out the right number of words, so quit
dFloyd2Fin:
	mov	0,%i0
	ret
	restore


!---------------------------------------------------------------------------
! _ip__ditherFloydLine4to1 Routine
!---------------------------------------------------------------------------

! see general algorithm comments in i_floyd.c

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_ditherFloydLine4to1)
MK_GLOBAL_NAME(ip_ditherFloydLine4to1):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  As stated above
! we also use some of this stuff.
	st	rIPStateIn,  [%fp+ipStateOffset]
	st	rSrcAddrIn,  [%fp+srcAddrOffset]
	st	rSrcPitchIn, [%fp+srcPitchOffset]
	st	rDstAddrIn,  [%fp+dstAddrOffset]
	st	rDstPitchIn, [%fp+dstPitchOffset]
	st	rGrayBufIn,  [%fp+grayBufOffset]

	ld	[%fp+srcAddrOffset],  rSrcAddr
	ld	[%fp+srcPitchOffset], rSrcPitch
	ld	[%fp+dstAddrOffset],  rDstAddr
	ld	[%fp+dstPitchOffset], rDstPitch
	ld	[%fp+grayBufOffset],  rGrayBuf
	ld	[%fp+widthOffset],    rWidth
	ld	[%fp+drcTableOffset], rDRCTable

	ld	[%fp+pixelOffset], %o0		! get high or low pixel

! compute the modified width (number of times we go through the loop).  We
! would like to always exit the loop at the bottom.  In addition, we would
! like the next source pixel for line 7 to be the high nibble of the byte.
! To do this for the high pixel starting case, we need to add 1 to the 
! count.  As a result we add DBLOCK_HEIGHT instead of DBLOCK_HEIGHT-1
! (as you might expect since we actually test rCount against this count
! before incrementing rCount, but start rCount at 1 instead of 0).
	add	rWidth, 0x1F, rWidth		! round up to next mult of 32
	tst	%o0				! do we start with an
						! high or low pixel?
	bnz	dFloyd4to1OddPixelSetup	
	andn	rWidth, 0x1F, rWidth
	add	rWidth, DBLOCK_HEIGHT, rWidth	! = additional cycles
	st	rWidth, [%fp+realWidthOffset]

! construct constants used to modify rSrcAddr and rDstAddr at the bottom
! of the loop.  REMEMBER THAT rDstPitch IS THE SAME AS rScr!!!
	sll	rSrcPitch, 2, rScr1		! 4 lines
	add	rSrcPitch, rScr1, rScr1		! 5 lines
	add	rSrcPitch, rScr1, rScr1		! 6 lines
	st	rScr1, [%fp+srcFixupOffset]
	sub	rScr1, 1, rScr1			! and over 1 pixel
	st	rScr1, [%fp+srcFixupOffsetOver1]

	sll	rDstPitch, 2, rScr1		! 4 lines
	add	rScr1, rDstPitch, rScr1		! 5 lines
	sub	rScr1, 4, rScr1			! and over 4 bytes
	st	rScr1, [%fp+dstFixupOffset]

! dereference pointers to the two most referenced table parameters
	ld	[%fp+hdeTableOffset],	rHDErrorTab
	ld	[%fp+vdeTableOffset],	rVDErrorTab

! Load data into the gray registers.  Start with the top line from GrayBuf.
	ldub	[rGrayBuf + DBLOCK_HEIGHT+0], rScr
	ldub	[rGrayBuf + DBLOCK_HEIGHT+1], rScr1
	ldub	[rGrayBuf + DBLOCK_HEIGHT+2], rGLine0
	add	rGrayBuf, DBLOCK_HEIGHT+3, rGrayBuf
	sll	rScr, 22, rScr			! align first byte
	or	rGLine0, rScr, rGLine0
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine0, rScr1, rGLine0

! Get raw data from the image.  Each byte must be corrected individually.

! Load rGLine1
	ldub	[rSrcAddr - 1], rScr
	ldub	[rSrcAddr + 0], rScr1
	ldub	[rSrcAddr + 0], rGLine1
	add	rScr, lDRCOffset, rScr
	add	rGLine1, lDRCOffset, rGLine1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine1], rGLine1
	sll	rScr, 22, rScr			! align first byte
	or	rGLine1, rScr, rGLine1
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine1, rScr1, rGLine1

! Load rGLine2
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 1], rScr
	ldub	[rSrcAddr - 1], rScr1
	ldub	[rSrcAddr + 0], rGLine2
	add	rScr1, lDRCOffset, rScr1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine2], rGLine2
	sll	rScr, 22, rScr			! align first byte
	or	rGLine2, rScr, rGLine2
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine2, rScr1, rGLine2

! Load rGLine3
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 2], rScr
	ldub	[rSrcAddr - 1], rScr1
	ldub	[rSrcAddr - 1], rGLine3
	add	rScr, lDRCOffset, rScr
	add	rGLine3, lDRCOffset, rGLine3
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine3], rGLine3
	sll	rScr, 22, rScr			! align first byte
	or	rGLine3, rScr, rGLine3
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine3, rScr1, rGLine3

! Load rGLine4
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 2], rScr
	ldub	[rSrcAddr - 2], rScr1
	ldub	[rSrcAddr - 1], rGLine4
	add	rScr1, lDRCOffset, rScr1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine4], rGLine4
	sll	rScr, 22, rScr			! align first byte
	or	rGLine4, rScr, rGLine4
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine4, rScr1, rGLine4

! Load rGLine5
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 3], rScr
	ldub	[rSrcAddr - 2], rScr1
	ldub	[rSrcAddr - 2], rGLine5
	add	rScr, lDRCOffset, rScr
	add	rGLine5, lDRCOffset, rGLine5
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine5], rGLine5
	sll	rScr, 22, rScr			! align first byte
	or	rGLine5, rScr, rGLine5
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine5, rScr1, rGLine5

! Load rGLine6
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 3], rScr
	ldub	[rSrcAddr - 3], rScr1
	ldub	[rSrcAddr - 2], rGLine6
	add	rScr1, lDRCOffset, rScr1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine6], rGLine6
	sll	rScr, 22, rScr			! align first byte
	or	rGLine6, rScr, rGLine6
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine6, rScr1, rGLine6

! There's no need to initialize rGLine7.  It is never used to compute
! thresholds, hence is not a source for error diffusion.
! You'd think we would put the marker bits into the rDLinex registers here,
! but NO.  rDLine0 could use a 1, but the other registers should have their
! bits placed farther off their right ends.  Those registers need to get
! through the trash at the beginning of the rectangle before starting to
! accumulate real data.  So, we'll initialize rDLine0 here.  When rDLine0's
! marker bit shifts into the carry, we'll also test rCount.  If it's equal
! to 32, we'll set the marker bits in the other rDLineX registers.
! However, DO initialize these registers so initial trash doesn't cause
! spurious writes.
	mov	1, rDLine0
	mov	0, rDLine1
	mov	0, rDLine2
	mov	0, rDLine3
	mov	0, rDLine4
	mov	0, rDLine5
	mov	0, rDLine6

! restore rSrcAddr
	sll	rSrcPitch, 2, rScr		! compute 4*rSrcPitch
	sub	rSrcAddr, rScr, rSrcAddr	! now down by 1 lines
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! points to line1 again
	add	rSrcAddr, 1, rSrcAddr		! point to first new pixel


! Initialize rCount for use in the main loop
	mov	1, rCount

! ****************** MAIN LOOP **************************

dFloyd4to1RowLoop:

! ********** LINE 0

! Get threshold pixel for line 0
	srl	rGLine0, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine0, rScr1, rGLine0		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine0, rDLine0, rDLine0	! did marker 1 shift out?
	bcc	dFloyd4to1ThreshLine0		! no, skip the write
	add	rGLine1, rScr, rGLine1		! apply {3/16 5/16 1/16} error

! time to do a write
	not	rDLine0				! invert dithered data
	st	rDLine0, [rDstAddr]
	xorcc	rCount, 32, %g0			! was that the first time?
	bnz	dFloyd4to1ThreshLine0		! no, just keep going.
	mov	1, rDLine0			! turn on marker bit again

! We have just written the first 32 bits of the top line.  That means the
! other data bits will be ready soon.  Turn on their marker bits.  We couldn't
! do this before because the correct positions for the marker bits weren't
! in the registers.  For example, rDLine1 should be written after the first
! 33 times through the loop, but there are only 32 bits in a register.
! BUG FIX:  Just because we cleared all the rDLineX registers doesn't mean
! they don't have trash in them now.  We've done 32 cycles.  The MS bit in
! rDLine1 is trash because it was derived from data to the left of the
! rectangle.  The MS 2 bits in rDLine2 are trash for the same reason and
! so on.  We not only need to set the marker bits in all these registers,
! we need to clear out the trash bits.
	sethi	%hi(0x80000000), rScr		! prepare trash mask
	sethi	%hi(0x40000000), rScr1		! prepare rDLine1's marker bit
	andn	rDLine1, rScr, rDLine1		! clear trash bit
	or	rDLine1, rScr1, rDLine1		! set marker bit

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine2, rScr, rDLine2
	or	rDLine2, rScr1, rDLine2

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine3, rScr, rDLine3
	or	rDLine3, rScr1, rDLine3

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine4, rScr, rDLine4
	or	rDLine4, rScr1, rDLine4

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine5, rScr, rDLine5
	or	rDLine5, rScr1, rDLine5

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine6, rScr, rDLine6
	or	rDLine6, rScr1, rDLine6


! shift new pixel into rGLine0
dFloyd4to1ThreshLine0:
	ldub	[rGrayBuf + 0], rScr1		! get next source pixel
	add	rGrayBuf, 1, rGrayBuf		! pt to next pixel
! DO NOT CONVERT this pixel, the pixels in GrayBuf were converted before
! being written there.
	sll	rGLine0, 11, rGLine0		! make room
	or	rGLine0, rScr1, rGLine0		! insert new byte

! ********** LINE 1 First Pixel
	srl	rGLine1, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine1, rScr1, rGLine1		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine1, rDLine1, rDLine1	! did marker 1 shift out?
	bcc	dFloydThreshLine1High		! no, skip the write
	add	rGLine2, rScr, rGLine2		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine1				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine1, [rDstAddr]
	mov	1, rDLine1			! turn on marker bit again

! shift new pixel into rGLine1
dFloydThreshLine1High:
	ldub	[rSrcAddr - 0], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine1, 11, rGLine1		! make room
	or	rGLine1, rScr1, rGLine1		! insert new byte


! ********** LINE 2 First Pixel
	srl	rGLine2, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine2, rScr1, rGLine2		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine2, rDLine2, rDLine2	! did marker 1 shift out?
	bcc	dFloydThreshLine2Low		! no, skip the write
	add	rGLine3, rScr, rGLine3		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine2				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine2, [rDstAddr]
	mov	1, rDLine2			! turn on marker bit again

! shift new pixel into rGLine2
dFloydThreshLine2Low:
	ldub	[rSrcAddr - 1], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	add	rScr1, lDRCOffset, rScr1	! point to drc table for
						! low byte
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine2, 11, rGLine2		! make room
	or	rGLine2, rScr1, rGLine2		! insert new byte


! ********** LINE 3 First Pixel
	srl	rGLine3, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine3, rScr1, rGLine3		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine3, rDLine3, rDLine3	! did marker 1 shift out?
	bcc	dFloydThreshLine3High		! no, skip the write
	add	rGLine4, rScr, rGLine4	! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine3				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine3, [rDstAddr]
	mov	1, rDLine3			! turn on marker bit again

! shift new pixel into rGLine3
dFloydThreshLine3High:
	ldub	[rSrcAddr - 1], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine3, 11, rGLine3		! make room
	or	rGLine3, rScr1, rGLine3		! insert new byte


! ********** LINE 4 First Pixel
	srl	rGLine4, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine4, rScr1, rGLine4		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine4, rDLine4, rDLine4	! did marker 1 shift out?
	bcc	dFloydThreshLine4Low		! no, skip the write
	add	rGLine5, rScr, rGLine5		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine4				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine4, [rDstAddr]
	mov	1, rDLine4			! turn on marker bit again

! shift new pixel into rGLine4
dFloydThreshLine4Low:
	ldub	[rSrcAddr - 2], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	add	rScr1, lDRCOffset, rScr1	! point to drc table for
						! low byte
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine4, 11, rGLine4		! make room
	or	rGLine4, rScr1, rGLine4		! insert new byte


! ********** LINE 5 First Pixel
	srl	rGLine5, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine5, rScr1, rGLine5		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine5, rDLine5, rDLine5	! did marker 1 shift out?
	bcc	dFloydThreshLine5High		! no, skip the write
	add	rGLine6, rScr, rGLine6		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine5				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine5, [rDstAddr]
	mov	1, rDLine5			! turn on marker bit again

! shift new pixel into rGLine5
dFloydThreshLine5High:
	ldub	[rSrcAddr - 2], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine5, 11, rGLine5		! make room
	or	rGLine5, rScr1, rGLine5		! insert new byte


! ********** LINE 6 First Pixel
	srl	rGLine6, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine6, rScr1, rGLine6		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine6, rDLine6, rDLine6	! did marker 1 shift out?
	bcc	dFloydThreshLine6Low		! no, skip the write
	add	rGLine7, rScr, rGLine7		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine6				! invert dithered data
!	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
!	st	rDLine6, [rDstAddr]
	st	rDLine6, [rDstAddr + rScr1]	! store to line 6
	ld	[%fp+dstFixupOffset], rScr1	! get amount by which rDstAddr
	mov	1, rDLine6			! turn on marker bit again
						! should be changed at bottom
						! of the loop
	sub	rDstAddr, rScr1, rDstAddr	! point to line 0 again

! shift new pixel into rGLine6
dFloydThreshLine6Low:
	ldub	[rSrcAddr - 3], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	add	rScr1, lDRCOffset, rScr1	! point to drc table for
						! low byte
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine6, 11, rGLine6		! make room
	or	rGLine6, rScr1, rGLine6		! insert new byte

! ********** LINE 7 First Pixel

! calculate the output gray data for line 7
	ld	[%fp+dtTableOffset], rScr1      ! ptr to truncate table
	srl	rGLine7, 22, rScr		! grab next output pixel
	ldub	[rScr1+rScr], rScr		! truncate it
	sll	rGLine7Out, 8, rGLine7Out	! make room for it
	or	rGLine7Out, rScr, rGLine7Out	! insert it

! is rGLine7Out full of gray data?  This will happen every 4th time through
! the loop ((rCount & 0x3) == 0).
	andcc	rCount, 0x3, rScr		! (rCount) % 3 == 0?
	bnz	dFloyd4to1ShiftLine7High	! nope, skip the write
	ld	[%fp+srcFixupOffset], rScr	! prepare to adjust rSrcAddr

! time to do a write
	st	rGLine7Out, [rGrayBuf - GRAY_DEST_OFFSET]

! shift rGLine7
dFloyd4to1ShiftLine7High:
	ldub	[rSrcAddr - 3], rScr1		! get next source pixel
	sub	rSrcAddr, rScr, rSrcAddr	! pt to line 1 
	ldub	[rDRCTable + rScr1], rScr1	! convert next source pixel
	sll	rGLine7, 11, rGLine7		! make room
	or	rGLine7, rScr1, rGLine7		! insert new byte

! We're finished with one skewed column of data.  Test to see
! if we should keep going (rCount < stored value of rWidth)
	ld	[%fp+realWidthOffset], rScr1	! get end count
	cmp	rScr1, rCount			! is rCount bigger yet?
	bl	dFloyd4to1RowDone
	inc	rCount				! increment rCount


! ********** LINE 0 Second Pixel
dFloydRowLoopSecondPixel:

! Get threshold pixel for line 0
	srl	rGLine0, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine0, rScr1, rGLine0		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine0, rDLine0, rDLine0	! did marker 1 shift out?
	bcc	dFloyd4to1ThreshLine0Second	! no, skip the write
	add	rGLine1, rScr, rGLine1		! apply {3/16 5/16 1/16} error

! time to do a write
	not	rDLine0				! invert dithered data
	st	rDLine0, [rDstAddr]
	xorcc	rCount, 32, %g0			! was that the first time?
	bnz	dFloyd4to1ThreshLine0Second	! no, just keep going.
	mov	1, rDLine0			! turn on marker bit again

! We have just written the first 32 bits of the top line.  That means the
! other data bits will be ready soon.  Turn on their marker bits.  We couldn't
! do this before because the correct positions for the marker bits weren't
! in the registers.  For example, rDLine1 should be written after the first
! 33 times through the loop, but there are only 32 bits in a register.
! BUG FIX:  Just because we cleared all the rDLineX registers doesn't mean
! they don't have trash in them now.  We've done 32 cycles.  The MS bit in
! rDLine1 is trash because it was derived from data to the left of the
! rectangle.  The MS 2 bits in rDLine2 are trash for the same reason and
! so on.  We not only need to set the marker bits in all these registers,
! we need to clear out the trash bits.
	sethi	%hi(0x80000000), rScr		! prepare trash mask
	sethi	%hi(0x40000000), rScr1		! prepare rDLine1's marker bit
	andn	rDLine1, rScr, rDLine1		! clear trash bit
	or	rDLine1, rScr1, rDLine1		! set marker bit

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine2, rScr, rDLine2
	or	rDLine2, rScr1, rDLine2

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine3, rScr, rDLine3
	or	rDLine3, rScr1, rDLine3

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine4, rScr, rDLine4
	or	rDLine4, rScr1, rDLine4

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine5, rScr, rDLine5
	or	rDLine5, rScr1, rDLine5

	sra	rScr, 1, rScr
	srl	rScr1, 1, rScr1
	andn	rDLine6, rScr, rDLine6
	or	rDLine6, rScr1, rDLine6


! shift new pixel into rGLine0
dFloyd4to1ThreshLine0Second:
	ldub	[rGrayBuf + 0], rScr1		! get next source pixel
	add	rGrayBuf, 1, rGrayBuf		! pt to next pixel
! DO NOT CONVERT this pixel, the pixels in GrayBuf were converted before
! being written there.
	sll	rGLine0, 11, rGLine0		! make room
	or	rGLine0, rScr1, rGLine0		! insert new byte

! ********** LINE 1 Second Pixel
	srl	rGLine1, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine1, rScr1, rGLine1	! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine1, rDLine1, rDLine1	! did marker 1 shift out?
	bcc	dFloydThreshLine1Low		! no, skip the write
	add	rGLine2, rScr, rGLine2		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine1				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine1, [rDstAddr]
	mov	1, rDLine1			! turn on marker bit again

! shift new pixel into rGLine1
dFloydThreshLine1Low:
	ldub	[rSrcAddr - 0], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	add	rScr1, lDRCOffset, rScr1	! point to drc table for
						! low byte
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine1, 11, rGLine1		! make room
	or	rGLine1, rScr1, rGLine1	! insert new byte


! ********** LINE 2 Second Pixel
	srl	rGLine2, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine2, rScr1, rGLine2		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine2, rDLine2, rDLine2	! did marker 1 shift out?
	bcc	dFloydThreshLine2High		! no, skip the write
	add	rGLine3, rScr, rGLine3	! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine2			! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine2, [rDstAddr]
	mov	1, rDLine2			! turn on marker bit again

! shift new pixel into rGLine2
dFloydThreshLine2High:
	ldub	[rSrcAddr - 0], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine2, 11, rGLine2		! make room
	or	rGLine2, rScr1, rGLine2	! insert new byte


! ********** LINE 3 Second Pixel
	srl	rGLine3, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine3, rScr1, rGLine3	! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine3, rDLine3, rDLine3	! did marker 1 shift out?
	bcc	dFloydThreshLine3Low		! no, skip the write
	add	rGLine4, rScr, rGLine4		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine3				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine3, [rDstAddr]
	mov	1, rDLine3			! turn on marker bit again

! shift new pixel into rGLine3
dFloydThreshLine3Low:
	ldub	[rSrcAddr - 1], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	add	rScr1, lDRCOffset, rScr1	! point to drc table for
						! low byte
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine3, 11, rGLine3		! make room
	or	rGLine3, rScr1, rGLine3		! insert new byte


! ********** LINE 4 Second Pixel
	srl	rGLine4, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine4, rScr1, rGLine4		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine4, rDLine4, rDLine4	! did marker 1 shift out?
	bcc	dFloydThreshLine4High		! no, skip the write
	add	rGLine5, rScr, rGLine5		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine4				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine4, [rDstAddr]
	mov	1, rDLine4			! turn on marker bit again

! shift new pixel into rGLine4
dFloydThreshLine4High:
	ldub	[rSrcAddr - 1], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine4, 11, rGLine4		! make room
	or	rGLine4, rScr1, rGLine4		! insert new byte


! ********** LINE 5 Second Pixel
	srl	rGLine5, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine5, rScr1, rGLine5	! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine5, rDLine5, rDLine5	! did marker 1 shift out?
	bcc	dFloydThreshLine5Low		! no, skip the write
	add	rGLine6, rScr, rGLine6		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine5				! invert dithered data
	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
	st	rDLine5, [rDstAddr]
	mov	1, rDLine5			! turn on marker bit again

! shift new pixel into rGLine5
dFloydThreshLine5Low:
	ldub	[rSrcAddr - 2], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	add	rScr1, lDRCOffset, rScr1	! point to drc table for
						! low byte
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine5, 11, rGLine5		! make room
	or	rGLine5, rScr1, rGLine5	! insert new byte


! ********** LINE 6 Second Pixel
	srl	rGLine6, 22, rScr		! get the threshold value
	sll	rScr, 2, rScr			! convert to word 
	ld	[rHDErrorTab + rScr], rScr1	! get 7/16 error
	ld	[rVDErrorTab + rScr], rScr	! get {3/16 5/16 1/16} error
	addcc	rGLine6, rScr1, rGLine6		! apply 7/16 error

! shift dithered data into binary reg
	addxcc	rDLine6, rDLine6, rDLine6	! did marker 1 shift out?
	bcc	dFloydThreshLine6High		! no, skip the write
	add	rGLine7, rScr, rGLine7		! apply {3/16 5/16 1/16} error

! time to do a write
	ld	[%fp+dstPitchOffset], rScr1	! get the dest pitch
	not	rDLine6				! invert dithered data
!	add	rDstAddr, rScr1, rDstAddr	! pt to line 2
!	st	rDLine6, [rDstAddr]
	st	rDLine6, [rDstAddr + rScr1]	! store to line 6
	ld	[%fp+dstFixupOffset], rScr1	! get amount by which rDstAddr
	mov	1, rDLine6			! turn on marker bit again
						! should be changed at bottom
						! of the loop
	sub	rDstAddr, rScr1, rDstAddr	! point to line 0 again


! shift new pixel into rGLine6
dFloydThreshLine6High:
	ldub	[rSrcAddr - 2], rScr1		! get next source pixel
	add	rSrcAddr, rSrcPitch, rSrcAddr	! pt down 1 line
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine6, 11, rGLine6		! make room
	or	rGLine6, rScr1, rGLine6		! insert new byte


! ********** LINE 7 Second Pixel

! calculate the output gray data for line 7
	ld	[%fp+dtTableOffset], rScr1      ! ptr to truncate table
	srl	rGLine7, 22, rScr		! grab next output pixel
	ldub	[rScr1+rScr], rScr		! truncate it
	sll	rGLine7Out, 8, rGLine7Out	! make room for it
	or	rGLine7Out, rScr, rGLine7Out	! insert it

! is rGLine7Out full of gray data?  This will happen every 4th time through
! the loop ((rCount & 0x3) == 0).
	andcc	rCount, 0x3, rScr		! (rCount) % 3 == 0?
	bnz	dFloyd4to1ShiftLine7Low		! nope, skip the write
	ld	[%fp+srcFixupOffsetOver1], rScr	! prepare to adjust rSrcAddr

! time to do a write
	st	rGLine7Out, [rGrayBuf - GRAY_DEST_OFFSET]

! shift rGLine7
dFloyd4to1ShiftLine7Low:
	ldub	[rSrcAddr - 3], rScr1		! get next source pixel
	sub	rSrcAddr, rScr, rSrcAddr	! pt to line 1 and 1 byte over
	add	rScr1, lDRCOffset, rScr1	! point to drc table for
						! low nibble
	ldub	[rDRCTable + rScr1], rScr1	! convert next source pixel
	sll	rGLine7, 11, rGLine7		! make room
	or	rGLine7, rScr1, rGLine7		! insert new byte


! We're finished with one skewed column of data.  Test to see
! if we should keep going (rCount < stored value of rWidth)

	ld	[%fp+realWidthOffset], rScr1	! get end count
	cmp	rScr1, rCount			! is rCount bigger yet?
	bge	dFloyd4to1RowLoop
	inc	rCount				! increment rCount

! We shouldn't put out any more dithered data, but we should finish filling
! GrayBuf.  The top line is DBLOCK_HEIGHT bytes ahead of us now.  We need
! to convert at least DBLOCK_HEIGHT more bytes, then keep going until
! we reach the next multiple of 4.
dFloyd4to1RowDone:
	add	rCount, DBLOCK_HEIGHT+3, rGLine0 ! round up to next mult of 4
	andn	rGLine0, 3, rGLine0
! repair rSrcAddr
	add	rSrcAddr, rScr, rSrcAddr	! pt to line 7 again
	sub	rSrcAddr, 3, rSrcAddr		! and the next source byte

	ld	[%fp+dtTableOffset], rDLine0    ! ptr to truncate table

! read and translate more pixels
dFloyd4to1FillGrayLoop:
	add	rGrayBuf, 1, rGrayBuf
	srl	rGLine7, 22, rScr		! grab next output pixel
	ldub	[rDLine0+rScr], rScr		! truncate it
	sll	rGLine7Out, 8, rGLine7Out	! make room for it
	or	rGLine7Out, rScr, rGLine7Out	! insert it

! is rGLine7Out full of gray data?  This will happen every 4th time through
! the loop ((rCount & 0x3) == 3).
	andcc	rCount, 0x3, %g0		! (rCount) % 3 == 0?
	bz,a	dFloyd4to1FillGrayShift		! yes, do the write
	st	rGLine7Out, [rGrayBuf-GRAY_DEST_OFFSET]

! shift rGLine7
dFloyd4to1FillGrayShift:
	ldub	[rSrcAddr], rScr1		! get next source pixel
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine7, 11, rGLine7		! make room
	or	rGLine7, rScr1, rGLine7		! insert new byte
! done yet?
	cmp	rGLine0, rCount
	bl	dFloyd4to1Fin			! nope, keep going
	add	rCount, 1, rCount

	add	rGrayBuf, 1, rGrayBuf
	srl	rGLine7, 22, rScr		! grab next output pixel
	ldub	[rDLine0+rScr], rScr		! truncate it
	sll	rGLine7Out, 8, rGLine7Out	! make room for it
	or	rGLine7Out, rScr, rGLine7Out	! insert it

! is rGLine7Out full of gray data?  This will happen every 4th time through
! the loop ((rCount & 0x3) == 3).
	andcc	rCount, 0x3, %g0		! (rCount) % 3 == 0?
	bz,a	dFloyd4to1FillGrayShiftLow	! yes, do the write
	st	rGLine7Out, [rGrayBuf-GRAY_DEST_OFFSET]

! shift rGLine7
dFloyd4to1FillGrayShiftLow:
	ldub	[rSrcAddr], rScr1		! get next source pixel
	add	rSrcAddr, 1, rSrcAddr		! pt to next byte to the right
	add	rScr1, lDRCOffset, rScr1
	ldub	[rDRCTable + rScr1], rScr1	! convert it
	sll	rGLine7, 11, rGLine7		! make room
	or	rGLine7, rScr1, rGLine7		! insert new byte

! done yet?
	cmp	rGLine0, rCount
	bge	dFloyd4to1FillGrayLoop		! nope, keep going
	add	rCount, 1, rCount


! END OF LOOP ACROSS ROW

! We've put out the right number of words, so quit
dFloyd4to1Fin:
	mov	0,%i0
	ret
	restore

! compute the modified width (number of times we go through the loop).  We
! would like to always exit the loop at the bottom.  In addition, we would
! like the next source pixel for line 7 to be the high nibble of the byte.
! To do this for the low pixel starting case, we do not need to add 1 to the
! count as we did in the high pixel starting case.  As a result we add
! DBLOCK_HEIGHT-1 (as you might expect since we actually test rCount 
! against this count before incrementing rCount, but start rCount at 1
! instead of 0).
dFloyd4to1OddPixelSetup:
	add	rWidth, DBLOCK_HEIGHT, rWidth	! = additional cycles
	st	rWidth, [%fp+realWidthOffset]

! construct constants used to modify rSrcAddr and rDstAddr at the bottom
! of the loop.  REMEMBER THAT rDstPitch IS THE SAME AS rScr!!!
	sll	rSrcPitch, 2, rScr1		! 4 lines
	add	rSrcPitch, rScr1, rScr1		! 5 lines
	add	rSrcPitch, rScr1, rScr1		! 6 lines
	st	rScr1, [%fp+srcFixupOffset]
	sub	rScr1, 1, rScr1			! and over 1 pixel
	st	rScr1, [%fp+srcFixupOffsetOver1]

	sll	rDstPitch, 2, rScr1		! 4 lines
	add	rScr1, rDstPitch, rScr1		! 5 lines
	sub	rScr1, 4, rScr1			! and over 4 bytes
	st	rScr1, [%fp+dstFixupOffset]

! dereference pointers to the two most referenced table parameters
	ld	[%fp+hdeTableOffset],	rHDErrorTab
	ld	[%fp+vdeTableOffset],	rVDErrorTab

! Load data into the gray registers.  Start with the top line from GrayBuf.
	ldub	[rGrayBuf + DBLOCK_HEIGHT+0], rScr
	ldub	[rGrayBuf + DBLOCK_HEIGHT+1], rScr1
	ldub	[rGrayBuf + DBLOCK_HEIGHT+2], rGLine0
	add	rGrayBuf, DBLOCK_HEIGHT+3, rGrayBuf
	sll	rScr, 22, rScr			! align first byte
	or	rGLine0, rScr, rGLine0
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine0, rScr1, rGLine0

! Get raw data from the image.  Each byte must be corrected individually.

! Load rGLine1
	ldub	[rSrcAddr + 0], rScr
	ldub	[rSrcAddr + 0], rScr1
	ldub	[rSrcAddr + 1], rGLine1
	add	rScr1, lDRCOffset, rScr1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine1], rGLine1
	sll	rScr, 22, rScr			! align first byte
	or	rGLine1, rScr, rGLine1
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine1, rScr1, rGLine1

! Load rGLine2
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 1], rScr
	ldub	[rSrcAddr + 0], rScr1
	ldub	[rSrcAddr + 0], rGLine2
	add	rScr, lDRCOffset, rScr
	add	rGLine2, lDRCOffset, rGLine2
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine2], rGLine2
	sll	rScr, 22, rScr			! align first byte
	or	rGLine2, rScr, rGLine2
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine2, rScr1, rGLine2

! Load rGLine3
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 1], rScr
	ldub	[rSrcAddr - 1], rScr1
	ldub	[rSrcAddr + 0], rGLine3
	add	rScr1, lDRCOffset, rScr1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine3], rGLine3
	sll	rScr, 22, rScr			! align first byte
	or	rGLine3, rScr, rGLine3
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine3, rScr1, rGLine3

! Load rGLine4
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 2], rScr
	ldub	[rSrcAddr - 1], rScr1
	ldub	[rSrcAddr - 1], rGLine4
	add	rScr, lDRCOffset, rScr
	add	rGLine4, lDRCOffset, rGLine4
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine4], rGLine4
	sll	rScr, 22, rScr			! align first byte
	or	rGLine4, rScr, rGLine4
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine4, rScr1, rGLine4

! Load rGLine5
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 2], rScr
	ldub	[rSrcAddr - 2], rScr1
	ldub	[rSrcAddr - 1], rGLine5
	add	rScr1, lDRCOffset, rScr1
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine5], rGLine5
	sll	rScr, 22, rScr			! align first byte
	or	rGLine5, rScr, rGLine5
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine5, rScr1, rGLine5

! Load rGLine6
	add	rSrcAddr, rSrcPitch, rSrcAddr	! point down 1 line
	ldub	[rSrcAddr - 3], rScr
	ldub	[rSrcAddr - 2], rScr1
	ldub	[rSrcAddr - 2], rGLine6
	add	rScr, lDRCOffset, rScr
	add	rGLine6, lDRCOffset, rGLine6
	ldub	[rDRCTable + rScr], rScr
	ldub	[rDRCTable + rScr1], rScr1
	ldub	[rDRCTable + rGLine6], rGLine6
	sll	rScr, 22, rScr			! align first byte
	or	rGLine6, rScr, rGLine6
	sll	rScr1, 11, rScr1		! align second byte
	or	rGLine6, rScr1, rGLine6

! There's no need to initialize rGLine7.  It is never used to compute
! thresholds, hence is not a source for error diffusion.
! You'd think we would put the marker bits into the rDLinex registers here,
! but NO.  rDLine0 could use a 1, but the other registers should have their
! bits placed farther off their right ends.  Those registers need to get
! through the trash at the beginning of the rectangle before starting to
! accumulate real data.  So, we'll initialize rDLine0 here.  When rDLine0's
! marker bit shifts into the carry, we'll also test rCount.  If it's equal
! to 32, we'll set the marker bits in the other rDLineX registers.
! However, DO initialize these registers so initial trash doesn't cause
! spurious writes.
	mov	1, rDLine0
	mov	0, rDLine1
	mov	0, rDLine2
	mov	0, rDLine3
	mov	0, rDLine4
	mov	0, rDLine5
	mov	0, rDLine6

! restore rSrcAddr
	sll	rSrcPitch, 2, rScr		! compute 4*rSrcPitch
	sub	rSrcAddr, rScr, rSrcAddr	! now down by 1 lines
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! points to line1 again
	add	rSrcAddr, 1, rSrcAddr		! point to first new pixel


! Initialize rCount for use in the main loop
	ba	dFloydRowLoopSecondPixel
	mov	1, rCount


/*

$Log: ssfloyd.s,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.8  1998/01/13 23:57:39  ddavies
ssfloyd.s -> /product/ipcore/ipshared/src/RCS/ssfloyd.s,v

CommonCheck: Add IPState parameter to almost all private routines.

! Revision 1.7  1994/08/25  17:43:26  ejaquith
! ssfloyd.s -> /product/ipcore/ipshared/src/RCS/ssfloyd.s,v
!
! Updated comments.  Added extra redundant parameter to the ditherFloydLine
! routines which is needed in the C/slow code versions.
! Bug Fix. _ip__ditherFloydLine4to1 was writing the destination image 1 pixel
! over to the right more than it should have in the case when pixel==1.
! It was also, in turn, writing 1 pixel over to the right to the GrayBuffer.
! Fixed it by calling the second half of the main loop when pixel==1 and
! by adding a few changes to make it work.
!
! Revision 1.6  1994/07/27  15:42:01  ejaquith
! ssfloyd.s -> /product/ipcore/ipshared/src/RCS/ssfloyd.s,v
!
! Reentrancy!
! Removed globals and made functions be called with pointers to error
! diffusion tables instead.
!
! Revision 1.5  1993/12/04  04:59:51  ddavies
! ssfloyd.s -> /product/ipcore/ipshared/src/RCS/ssfloyd.s,v
!
! Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
!
! Revision 1.4  1993/11/05  02:23:59  ddavies
! ssfloyd.s -> /product/ipcore/ipshared/src/RCS/ssfloyd.s,v
!
! Add RCS magic to have string specifying this file's name, rev number and
! last edit date into the object file.  Also add change log at the bottom.
!

*/

