/*****************************************************************
 *  Copyright (c) 1991, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  ssgenero.c -- generates assembly routines for ss (sparcstation)
 *                that implement the erosion operation for a
 *                specified structuring element
 */


#include <stdlib.h>
#include <stddef.h>
#include "iaerror.pub"
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "sparc.prv"
#include "frames.pub"
#include "memory.pub"

    /* prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrmorph.pub"
#include "shrmorph.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: ssgenero.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* PRIVATE */
/*************************************************************************
 *  ip_longOpIPLibLinked()
 *		There are things we'd really not do if routines capable
 *		of long (64 bit or longer) are linked, like avoid
 *		making superflous tables.  This routine returns a
 *		boolean which is set to cTrue if the
 *		long memory access libraries are linked.
 *
 *	Input:
 *		None
 *	Output:
 *		cTrue if fast libraries are linked, cFalse otherwise.
 *************************************************************************/
Bool  CDECL 
ip_longOpIPLibLinked (
    IPState    *pIPState)
{
	/* There are no long operations on the SPARC (yet). */
    PROLOGUE;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    return cFalse;
}


/* PUBLIC */
/************************************************************************
 *  i_getLongErodeAsmSize():  return number of bytes needed for assembly code
 *			used for erosion with this strel using Long (>= 64 bit)
 *			opcodes.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getLongErodeAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
	/* reference all the parms and return a buffer size of 0 since there
	 * is no assembly code for *LONG* operations here */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    *pNumAsmBytes = 0;
    return ia_noImpl;
}

/************************************************************************
 *  i_genLongErodeAsm():  generate assembly code to do erode defined by strel.
 *			This assembly code makes use of long memory accesses
 *			and operations (long >= 64 bits)
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastErode as
 *			blockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *			If the buffer was passed in, we assume it has at
 *			least the number of bytes that i_getLongErodeAsmSize
 *			would calculate.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_SETooBig:		Structure element was too big to
 *				generate accelerated code OR was composed
 *				entirely of DONT_CARE terms (which is silly).
 *	ia_nomem:		Not enough memory to allocate the code buffer.
 *	ia_bufferExceeded:	The routine tried to generate too much
 *				assembly code.  This is a bug in the routine.
 *				If it happens, please let us know.
 ************************************************************************/
Int32  CDECL 
i_genLongErodeAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    LongErodeProc *generated_code,
    Int32      *code_length)
{
	/* keep the compiler happy by referencing all the unused parms. */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    nx = nx;
    ny = ny;
    cx = cx;
    cy = cy;
    data = data;
    blockWidth = blockWidth;
    *code_length = 0;
    *generated_code = NULL;	/* we don't know how to generate code 
				 * using *LONG* operations for
				 * this processor, so punt */
    return ia_noImpl;
}


/* PRIVATE */
/************************************************************************
 *  ip_getLongErodeSEParms():
 *			return dimensions of input cache needed for long
 *			erode ops on a strel along with an array of the
 *			active columns in the SE.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	pSE:		Pointer to structure element's data.
 *	pCacheWidth:	Pointer to returned width of cache in bytes.
 *			This may be NULL.
 *	pCacheHeight:	Pointer to returned height of input cache.
 *			This may be NULL.
 *	pNumActiveCols:	Pointer to returned number of active columns.
 *			This may be NULL.
 *	pActiveCols:	Pointer to returned array of nx bytes.  This
 *			array should be allocated (and freed) by the caller.
 *			This routine will fill each element in the array
 *			with 0 if there are no SE elements in the column,
 *			cHit if the SE column contains only HITs and don't
 *			care terms, cMiss if the column contains only MISSes
 *			and don't care terms and (cHit | cMiss) if the
 *			column contains both HITs and MISSes (and possibly
 *			don't care terms).  This may be NULL.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
ip_getLongErodeSEParms (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    const UInt8 *pSE,
    Int32      *pCacheWidth,
    Int32      *pCacheHeight,
    Int32      *pNumActiveCols,
    UInt8       *pActiveCols)
{
    PROLOGUE;
	/* keep the compiler happy by referencing all the unused parms. */
    pIPState = pIPState;
    __error__ = __error__;
    nx = nx;
    ny = ny;
    pSE = pSE;
    pCacheWidth = pCacheWidth;
    pCacheHeight = pCacheHeight;
    pNumActiveCols = pNumActiveCols;
    pActiveCols = pActiveCols;

	/* This routine should never be called on platforms that have
	 * no long erode ops, so return ia_noImpl.
	 */
    return ia_noImpl;
}


/* PUBLIC */
/*------------------------------------------------------------------------*
 *
 *  i_genErodeAsm():  generate assembly code for erosion defined by strel
 *                  Puts address of the code into the SE.
 *                  Returns size in 4-byte words of code segment
 *                      generated, or non-zero on error.
 *
 * This code generates assembly code that erodes an image according to a
 * structure element.  The assembly code is meant to be called from a
 * special version of erode() that breaks images up into vertical slices.
 * The assembly code processes one slice.
 *
 * The basic algorithm is:
 *	1. Determine if we can generate code for this SE.  If not, simply
 *	   return with the SE's erodeOp pointer set to NULL.
 *	2. Determine the number of words required for the assembly code.
 *	3. Allocate a suitable buffer.
 *	4. Fill the Buffer with assembly code.
 *	5. Return a pointer to the code in the SE->erodeOp.
 *
 * The fundamental idea of the assembly code is to perform all necessary
 * operations on a destination word before writing it.
 * To do this, we need to maintain partial results in processor registers.
 * The number of partial results equals (the number of rows in the SE
 * minus 1) * (the number of horizontal words we process in parallel).
 *
 * We will fail to generate assembly code if the number of partial results
 * required is greater than the number of registers available.  The SPARC has
 * 32 registers.  One of them holds the constant 0.  Five are used to pass
 * arguments, though we steal one of those so only decrease the number
 * available by 4.  Global registers 6 and 7 are reserved for the system.
 * Out register 6 holds the stack pointer.  In registers 6 and 7 hold the
 * frame pointer and return address respectively.  That leaves us
 * 32 - 1 - 4 - 2 - 1 - 2 = 22 registers.  Of these, two are used for to
 * hold input data and two hold output data, so we're down to 18.  This is
 * the number of partial results we can hold for purely vertical structure
 * elements.  That restricts us to purely vertical structure elements of
 * height 18+1 = 19.
 *
 * If the structure element has any width, we need two more registers for
 * horizontal shifting.  This reduces the number of registers available for
 * partial results to 16.
 *
 * Therefore, if the SE is purely vertical (nx=1 and cx=0), we will reject it
 * only if it has more than 19 rows (ny > 19).  If the SE has width (nx>1 or
 * cx != 0), we will reject it if it has more than 17 rows (ny > 17).
 *
 * SE width is also a factor.  While complex code could handle arbitrarily
 * wide SE's, we will reject any which require more than 32 left shifts
 * (nx-cx > 33) or 32 right shifts (cx > 32).
 *
 * The assembly code processes vertical stripes of source image data.
 * We'll try not to use the word "columns" to mean more than 3 or 4 different
 * things.  The height of the stripes is set by the calling code.  The width
 * is set by the number of registers available for partial results.  Widths
 * increase in units of 32 bits.  Each increase in width requires ny-1
 * registers.  We arbitrarily set the maximum stripe width to 8*32 bits.
 * The benefit of wide stripes is reduction of redundant source reads
 * (each word affects outputs to the left and right) and an increase
 * in the number of double writes possible.  Beyond a stripe width of 8*32
 * bits, we hit diminishing returns.  In general, the stripe widths are
 * equal to MIN(16/(ny-1), 8)*32.  Possible stripe widths are:
 *
 *	SE height (ny)	Column Width in units of 32 bits
 *			 nx>1	nx=1
 *		1	  8	 8
 *		2	  8	 8
 *		3	  8	 8
 *		4	  5	 6
 *		5	  4	 4
 *		6	  3	 3
 *		7	  2	 3
 *		8	  2	 2
 *		9	  2	 2
 *	       10	  1	 2
 *	       11-17	  1	 1
 *	       18-19	 N.A.	 1
 *	       > 19	 N.A.	N.A.
 *		
 * Having decided to generate assembly code, we need to decide how much
 * space to ask for.  This is a good place to lay out the template of the
 * assembly code.
 *
 *   Call Overhead Ops
 *   {
 *	new register window (3 ops)
 *	save incoming parms in local vars (6 ops)
 *	save global registers for Purify (5 ops)
 *   }
 *   Pipeline Loading
 *   {
 *	adjust source address cy (cy+1 subtractions)
 *	load partials (ny * col Width ops + 1)
 *	adjust col height for added writes (1 op)
 *   }
 *   Erode ops
 *   {
 *	reload width and both pitches (3 ops)
 *	decrement height, set cc's (1 op)
 *	adjust current dest and src pointers down one line (2 ops)
 *	if there are no more words in the image stripe 
 *	    Return Overhead
 *		restore global registers for Purify (5 ops)
 *		{load return value, return, restore (3 ops)
 *	test starting dest addr and branch (2 ops)
 *	for either even or odd dest addresses
 *	{
 *	    read 1 input word (1 op)
 *	    if there will be right shifts (cx > 0)
 *		read input word to the left (1 op)
 *	    for each word in the stripe of image words
 *	    {
 *		if there are any SE columns to the left of center
 *		{
 *		    for leftmost non-vacant SE column
 *		    {
 *		        shift image into position (right) (3 ops)
 *		        ny-1 ops to migrate partial results up one row
 *		        if SE[0][thisCol] has hit or miss, add 1 op
 *		    }
 *		    for remaining SE cols to left of center
 *		    {
 *		        if col is not vacant
 *			{
 *			    shift image into position (right) (3 ops)
 *			    for each hit or miss
 *			        add one op
 *			}
 *		    }
 *		} -- all SE cols to left of center
 *	        if there are any SE columns to the right of center
 *	            load next input word to the left (1 op)
 *	        if center's SE column is not vacant
 *		{
 *		    if cx = 0 (this is the leftmost SE column)
 *		    {
 *		        ny-1 ops to migrate partial results up one row
 *		        if SE[0][cx] is not vacant, add 1 op
 *		    }
 *		    else
 *		    {
 *		        1 op for each hit or miss
 *		    }
 *		}
 *	        for each SE column starting from cx through nx-1
 *		{
 *		    shift input left by 1 bit (2 ops)
 *		    if this is the first SE column to be touched
 *		    {
 *		        ny-1 ops to migrate partial results up one row
 *		        if SE[0][thisCol] is not vacant, add 1 op
 *		    }
 *		    else
 *		    {
 *			1 op for each hit or miss
 *		    }
 *		}
 *		decrement width and branch if done (2 ops)
 *		if (we have just computed a destination word whose address
 *		    is odd) OR (have just computed the rightmost word in
 *		    this stripe)  -- do a write
 *		{
 *		    if (there's an even word ready, but not yet written)
 *			do a double write (1 op)
 *		    else
 *			do a single write (1 op)
 *		}
 *	    } -- end of loop for each word in image stripe
 *	} -- end of sections for even and odd dest start addresses
 *   } -- end of processing loop
 *
 *
 * Pretty messy, right?  Well, not so bad once you go through it a few times.
 * The inner loop reads 1 image word and applies it to all partial results
 * the word could affect.  This is done by going through the SE columns
 * from left to right.  This loop is repeated for each word in the image
 * stripe.  Rows of image words are processed from the top of the stripe
 * to the bottom.
 *
 * To increase performance, we do double stores whenever possible.  Data in
 * a SPARC double store must come from an even-odd numbered register pair (like
 * registers 2 and 3) and must be written to an even memory word address
 * (a multiple of 8 bytes).  We use a single even-odd register pair for all
 * outputs.  The are called rOut0 (the even one) and rOut1 (the odd one).
 *
 * Some destination rows start with words having even addresses, some with
 * words having odd addresses.  Therefore, we can't simply do double writes
 * on every other word.  The outputs of even dest words (row started in an even
 * address and this is an even word relative to the start OR row started
 * in an odd address and this is an odd word relative to the start) are put
 * into rOut0.  The outputs of odd image words (the other two cases) are
 * put into rOut1.  We do writes whenever an odd image word has just been
 * computed or when the last word of a row has been computed.  These writes
 * are double writes if an even word is ready.
 * The timing of writes relative to the beginning of the current row of
 * output word depends on whether that row began with an even or odd address.
 * Rather than test and branch at each possible write location, there are
 * two complete blocks of code (the even and odd blocks).
 * The even block assumes the first dest address was even, the odd block
 * assumes the first dest address was odd.  At the top of the loop that
 * processes a row of words, there's a test to see which block should be
 * executed.
 * 
 * The loops are constructed a bit strangely with the return and loop address
 * adjustments at the top.  That's so that most of the branches are backwards.
 * We can fill in the displacements of backward branches as soon as they are
 * generated.  The jump to the beginning of the odd block is the remaining
 * forward jump.  We generate this branch opcode when we figure out where
 * the beginning of the odd block is.
 *
 * There are a few more subtleties.  First, the very first time a destination
 * word is touched (the first time any hit or miss is encountered in the
 * topmost non-vacant row of the SE), we need to put either the appropriate
 * (possibly shifted) source or the inverse of that source into the
 * destination.  The source is used if the SE element is a hit, the inverted
 * source is used if the SE element is a miss.
 * This can be uniformly implemented by xoring the source with either 0 or -1.
 *
 * While integrating this code with the rest of the system, we discovered
 * a sizeable number of ops between pixrects of different sizes.  The code
 * generator has been modified to allow the source and destination pixrects
 * to have different values of pitch (number of bytes between vertically
 * adjacent pixels).
 *
 ************************ PURIFY fixup **********************************
 *
 * The Purify program from Pure, Inc. tracks memory usage.  It's a swell tool.
 * However, it gloms global registers for it's own use.  They normally get
 * away with this because they go through ALL code that's linked in and
 * make sure that everyone else leaves their register alone.  Since this code
 * is generated at runtime, they can't adjust it.  Bummer.  We'll go to the
 * trouble of saving the global registers on entry and restoring them on
 * exit.   We use 5 "local variables" for this purpose.  It appears that
 * this means we claim 120 bytes on the stack (instead of 64 - why does
 * adding 20 bytes of local variable mean we increase the local frame by
 * 56 bytes?).  The first opcodes store the globals, the opcodes just
 * before the return restore them.
 *
 ************************************************************************
 *
 * BUG FIX 12/7/94:  Leigh Klotz found out that if we give this code a
 * strel that's entirely DONT_CARE terms, it generates random code.
 * Check for this case.  If it occurs, don't allocate any assembly code
 * at all.  i_fastErode will fix things up.
 *
 * PRECAUTION 7/26/95: On one occasion, the DAE folks seemed to fall off
 * the end of the generated code.  The generated code does end with a
 * conditional branch, but the condition should always be true.  Just to
 * aid in debugging if this ever happens again, put a word of 0's, which
 * is an unimplemented opcode after the end of the code.  This should
 * cause a timely trap.
 *
 *------------------------------------------------------------------------*/


/* Input and constant register definitions */
/* The C Preprocessor can't handle the standard notation for SPARC
 * registers, "%i0".  We use an underline instead of percent sign.  */
#define	rSrcAddr	_i0
#define	rDstAddr	_i1
#define	rSrcPitch	_i2	/* reused as rInput0 */
#define	rDstPitch	_i3	/* reused as rInput1 */
#define	rHeight		_i4
#define	rWidth		_i5

#define	rScr0		_g1
#define	rScr1		_g2

#define	rInput0		_i2	/* originally rSrcPitch */
#define	rInput1		_i3	/* originally rDstPitch */

#define rOut0		_o0
#define rOut1		_o1

/* Names of global registers.  These are saved and restored so Purify will
 * work.
 */
#define	rG1		_g1
#define	rG2		_g2
#define	rG3		_g3
#define	rG4		_g4
#define	rG5		_g5

#define numSERows	ny		/* these are better names */
#define	numSECols	nx
#define numLeftSECols	cx
#define	numRightSECols	(numSECols - cx - 1)

#define cMaxPartials	18
#define cMaxStripeWidth	8
#define cMaxShiftDist	32

/* dimensions of the largest structure element we can handle */
#define	cMaxSERows	cMaxPartials+1
#define cMaxSECols	2*cMaxShiftDist+1

/* Sizes of various blocks of code */
#define	cCallOverheadSize	   14	/* and save global regs */
#define cLoopInitOverheadSize	   2
#define cEvenOddBlockOverheadSize  2
#define	cRightShiftOverheadSize	   3
#define cLeftShiftOverheadSize	   2
#define cWidthTestOverheadSize	   2
#define cLoopOverheadSize	   7
#define cRetOverheadSize	   8	/* and restore global regs */

/* Magic C Routine Boilerplate */
#define	LF26		-120		/* room for 5 local variables */
#define cSrcOffset	0x44
#define	cDstOffset	0x48
#define	cSrcPitchOffset	0x4c
#define	cDstPitchOffset	0x50
#define	cHeightOffset	0x54
#define	cWidthOffset	0x58

/* Local variable offsets */
#define cSavedg1Offset	-4
#define cSavedg2Offset	-8
#define cSavedg3Offset	-12
#define cSavedg4Offset	-16
#define cSavedg5Offset	-20


/* PUBLIC */
/************************************************************************
 *  i_getErodeAsmSize():  return number of bytes needed for assembly code
 *			used for erosion with this strel.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getErodeAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
    PROLOGUE;
    Int32         leftSEColCodeSize, rightSEColCodeSize;
    Int32         strelIsVert;	/* set if strel is purely vertical */
    Int32         stripeWidth;	/* width of stripe in 32-bit words */

	/* Initialize useful vars */
    strelIsVert = ((numSECols == 1) && (cx == 0));

	/*  At this point, we know we're going to generate code.
	 *  Figure out how much space we'll need. */
    if (ny == 1)
	stripeWidth = cMaxStripeWidth;
    else if (strelIsVert)
	stripeWidth = MIN(cMaxPartials/(ny-1), cMaxStripeWidth);
    else
	stripeWidth = MIN((cMaxPartials-2)/(ny-1), cMaxStripeWidth);
    
  leftSEColCodeSize = numLeftSECols * (cRightShiftOverheadSize + numSERows);
  rightSEColCodeSize = numRightSECols * (cLeftShiftOverheadSize + numSERows);
  
  *pNumAsmBytes =     4*(cCallOverheadSize +		/* call overhead */
		      cy + 1 +				/* source pipe adj */
		      (numSERows-1)*(stripeWidth+1) +	/* dest pipe adj */
		      1 +				/* dest addr adj */
		      1 +				/* height pipe adj */
		      cLoopOverheadSize +
		      cRetOverheadSize +
		      cLoopInitOverheadSize +		/* test even/odd addr */
		      2*
		      (cEvenOddBlockOverheadSize +
		       stripeWidth *
		       (leftSEColCodeSize +
			1 +		/* load word to the right */
			numSERows + 	/* to process center column */
			rightSEColCodeSize +
			cWidthTestOverheadSize +
			1)		/* store result word */
		       ) +
		      1 +			/* for Unimp trap op at end */
		      1);			/* just for good measure */


    NORMAL_RETURN;
}


/* This is a public procedure */
/* PUBLIC */
/************************************************************************
 *  i_genErodeAsm():  generate assembly code to do erode defined by strel.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastErode as
 *			blockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_SETooBig:		Structure element was too big to
 *				generate accelerated code OR was composed
 *				entirely of DONT_CARE terms (which is silly).
 *	ia_bufferExceeded:	The routine tried to generate too much
 *				assembly code.  This is a bug in the routine.
 *				If it happens, please let us know.
 *	ia_noImpl		not implemented
 *
 ************************************************************************/
Int32 CDECL
i_genErodeAsm(
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    PtrToConstUInt8 data,
    Int32Ptr	blockWidth,
    ErodeProc  *generated_code,
    Int32Ptr	code_length)
{
  PROLOGUE;

  /* id's of regs available for partial results */
  static Int32	partialRegs[cMaxPartials] =
      { _g3, _g4, _g5,
	_o2, _o3, _o4, _o5, _o7,
	_l0, _l1, _l2, _l3, _l4, _l5, _l6, _l7,
	rScr0, rScr1		/* only used for vert SE's */
	};
  /* the partial regs that are used for current SE.
   * BUG FIX, 8/5/93: We address this with row values (first coord) of
   * 1 through cMaxSERows-1, NOT 0 through cMaxPartials-1.  Therefore
   * we need an extra entry.
   */
  Int32         curPartials[cMaxSERows][cMaxStripeWidth];

  /* set with 0, xor, and or andn for each point in the SE.  Also
     indexed by current stripe pos and current output reg */
  UInt32    SEOps[cMaxStripeWidth][2][cMaxSERows][cMaxSECols];
  Int32     seColActive[cMaxSECols];  /* set if SE colunm contains
				      * hit or miss */
  UInt32    seOp;		/* an individual operation */
  Int32	    rS10, rS11, rS2, rD0, rD1; /* register field in an operation */
  Int32	    curInst;		/* instruction about to be stored */
  Int32	    haveSeenActiveSECol;   /* records whether non-empty col
				      has been seen yet */
  Int32     haveSeenHitOrMiss;	/* used to detect non-vacant SE columns */
  Int32	    r, c, i, j;		/* general iterative */
  Int32	    evenOdd;		/* iterates between even and odd blocks */
  Int32	    stripeOffset;	/* iterates across the stripe */
  Int32	    curInputReg;	/* input register holding current input */
  Int32	    otherInputReg;	/* input register holding data to the left or
				   right of the current input */
  Int32     outputRegIndex;	/* selects rOut0 vs rOut1 for current word */
  Int32	    evenWordReady;	/* set when an even dest word is sitting in
				 * rOut0 ready to be written */
  
  Int32     strelIsVert;	/* set if strel is purely vertical */
  Int32     stripeWidth;	/* width of stripe in 32-bit words */

  Int32    *pCodeBuf;		/* pointer to the code block */
  Bool      bufAllocatedHere;	/* set if this proc allocated the
				 * asm code buffer. */
  Int32     bufSizeInBytes;	/* size of code buffer */
  Int32     mainLoopOffset;	/* address of start of main loop */
  Int32     oddBlockBrOffset;	/* address of branch to odd block */
  Int32     curBufOffset;	/* offset to currently generated inst word */
  Int32     brDisp;		/* calculated branch displacement */
  Int32	    lastActiveCol;	/* used to calculate shift distances */
  

  /* Initialize useful vars */
  strelIsVert = ((numSECols == 1) && (cx == 0));

  /* Figure out if we can generate code at all */
  if ( ( strelIsVert && (numSERows-1 > cMaxPartials)) ||
      (!strelIsVert && (numSERows-1 > cMaxPartials-2)) ||
      (numLeftSECols > cMaxShiftDist) ||
      (numRightSECols > cMaxShiftDist) )
    {
      return ia_SETooBig;
    }
  
  /* see if the strel is composed entirely of cDontCare terms. If so,
   * fail as above.
   */
  __error__ = ia_SETooBig;
  for (c = 0; (c < ny*nx) && (__error__ == ia_SETooBig); c++)
  {
    if ((data[c] == cHit) || (data[c] == cMiss))
    {
	__error__ = ia_successful;
	break;
    }
  }
  /* if we didn't find any hits or misses, return now */
  if (__error__ == ia_SETooBig)
    ERR_RETURN;

  /* Set the stripeWidth, using it to set the block width */
  if (ny == 1)
    stripeWidth = cMaxStripeWidth;
  else if (strelIsVert)
    stripeWidth = MIN(cMaxPartials/(ny-1), cMaxStripeWidth);
  else
    stripeWidth = MIN((cMaxPartials-2)/(ny-1), cMaxStripeWidth);

  /* the blockWidth is 4 times the stripe width, so set it now. */
  *blockWidth = 4*stripeWidth;	/* set width of stripe in bytes */

  /*  At this point, we know we're going to generate code.
   *  If we don't already have an assembly buffer, get one.
   */
  i_getErodeAsmSize (pIPState, nx, ny, cx, cy, &bufSizeInBytes);
  pCodeBuf = (Int32 *) *generated_code;

  if (pCodeBuf == NULL)
  {
    pCodeBuf = (Int32 *) MALLOC(bufSizeInBytes);
    bufAllocatedHere = cTrue;
  }
  else
    bufAllocatedHere = cFalse;

  /* if the pointer is null, punt.  Otherwise, return the address
   * of the new buffer.
   */
  if (pCodeBuf == NULL)
    return ia_nomem;
  else
    *generated_code = (ErodeProc) pCodeBuf;

  /* Before starting to fill the block, fill in the identities of
   * the partials used for this SE.  The assignment of SPARC registers
   * to partial registers is a function of the SE. */
  for (r=1, i=0; r<numSERows; r++)
    for (c=0; c<stripeWidth; c++, i++)
    {
      curPartials[r][c] = partialRegs[i];
    }

  /*-------------------------------------------------------------------*
   * Fill in the SEOps operations array.  We fill the array in four steps.
   * First, clear the array.  Second, we figure out where to insert the
   * only xor op.  This is used the first time a (possibly shifted) input
   * word is inserted into the computation pipeline.  We process the
   * structure element from left to right, bottom to top.  In each column,
   * a shifted version of the input is applied to the same partial
   * registers.  The first time we find a non-empty SE
   * column, the pipeline is advanced by moving the results down the chain
   * of partial result registers. Since we process from bottom to top in
   * an SE column, the pipeline migrates from top to bottom.  Therefore,
   * the SE operations in which (possible shifted) inputs enter the pipeline
   * are those in the top row of the SE.
   *
   * Well, not exactly.  The top row of the SE might be vacant.  This would
   * happen if the SE's center were above all the active points in the SE.
   * New inputs really enter the pipeline in the SE operation that's
   * furthest to the left in the topmost *active* SE row.
   *
   * In the second step, we fill in entries for the leftmost active column.
   * This is the one in which the pipeline will be advanced.  An "active"
   * column is one in which there are any hits or misses.  Entries for
   * this column are different than entries for the other columns in two
   * ways.  First, each operation takes data from one partial result reg
   * and deposits the result in the next partial result register.  All
   * other operations apply the input to a partial and leave the result
   * in the same partial.
   * Second, "don't cares" in the SE become mov opcodes just to advance
   * the pipeline.  This does not apply to a "don't care" which would have
   * used the input word as the source - that's handled by the xor above.
   *
   * In the final step, we fill in all the remaining columns of SEOps.
   *
   * There are a couple of other things to watch out for.  If the column
   * being filled is the center's column, the input word is not shifted.
   * It comes directly from rInput0 or rInput1.  Otherwise, the input comes
   * from rScr0.  The choice of rInput0 or rInput1 depends on whether we
   * are currently processing an even or odd word relative to the left
   * edge of the stripe.  We actually make two copies of the SEOps array,
   * one used when we want rInput0/rScr0 and one used when we want
   * rInput1/rScr0.  These two copies are selected by the FIRST SEOps index.
   *
   * There's also a difference in output registers.  We want to write
   * completed results (operations in bottom SE row) to rOut0 if we are
   * computing a result word that will go to an even destination address.
   * Similarly, rOut1 is used when going to an odd destination address.
   * We make two copies of the two versions of SEOps to distinguish these
   * cases.  The output choice is selected by the SECOND SEOps index.
   *-------------------------------------------------------------------*/
  
  /* First, clear the SEOps and seColActive arrays */
  for (i = 0; i < cMaxStripeWidth; i++)
    for (j = 0; j < 2; j++)
      for (r = 0; r < cMaxSERows; r++)
	for (c = 0; c < cMaxSECols; c++)
	  SEOps[i][j][r][c] = 0;
  
  for (c = 0; c < cMaxSECols; c++)
    seColActive[c] = cFalse;
  
  haveSeenHitOrMiss = cFalse;
  for (r = 0; (r < numSERows) && !haveSeenHitOrMiss; r++)
    for (c = 0; (c < numSECols) && !haveSeenHitOrMiss; c++)
      /* if (data[r][c] != 0) */
      if (data[ (r * numSECols) + c ] != 0)
	haveSeenHitOrMiss = cTrue;
  
  /* We've found the operation that needs an xor op.  Fill in all
   * versions of SEOps.  BUT FIRST, decrement r and c to account
   * for extra increments above. */
  r--; c--;
  seColActive[c] = cTrue;
  for (i=0; i<stripeWidth; i++)
    {
      if (c == cx) 			/* in center's column */
	{
	  if ((i & 0x1) == 0) rS10 = rS11 = rInput0;
	  else rS10 = rS11 = rInput1;
	}
      else rS10 = rS11 = rScr0;			/* shifting needed */
      
      rS2 = (1 - data[ (r * numSECols) + c ]/*data[r][c]*/);	 	/* 1 (hit) => 0, */
      /* 2 (miss) => -1 */
      
      if (r == numSERows-1)			/* gen final output */
	{
	  rD0 = rOut0;
	  rD1 = rOut1;
	}
      else					/* update partials */
	rD0 = rD1 = curPartials[numSERows-r-1][i];
      
      SEOps[i][0][r][c] =
	GEN_INST(cXorCOp, rS10, rS2, rD0); /* new partial reg */
      SEOps[i][1][r][c] =
	GEN_INST(cXorCOp, rS11, rS2, rD1); /* new partial reg */
    }
  
  /* Fill in the rest of the SEOps, avoiding the one's we've just
   * filled in.   Note that to find the active columns, we index
   * down the columns instead of across the rows. */
  haveSeenActiveSECol = cFalse;
  for (c=0; c < numSECols; c++)
    {
      haveSeenHitOrMiss = cFalse;
      for (r=0; r < numSERows; r++)	/* see if anything's in this col */
	{
	  if (data[ (r * numSECols) + c ] /* data[r][c] */ != 0)
	    {
	      haveSeenHitOrMiss = cTrue;
	      seColActive[c] = cTrue;
	      break;
	    }
	}
      if (haveSeenHitOrMiss)
	{
	  if (!haveSeenActiveSECol)	/* is this the first one? */
	    {
	      /* if so, go through the SE col from bottom to top.  Quit
		 if we find the xor.  Also quit before doing the top
		 row.  It will either have the xor in it, or we shouldn't
		 be doing anything.  All the pipeline advancing happens
		 below the top row.  */
	      for (r = numSERows-1;
		   (r > 0) && (SEOps[0][0][r][c] == 0);
		   r--)
		{
		  /* choose between don't care => move; hit => and,
		     miss => andn */
		  switch (data[ (r * numSECols) + c ] /* data[r][c] */) {
		  case 0:			/* don't care */
		    seOp = cOrROp;
		    break;
		  case 1:			/* hit */
		    seOp = cAndROp;
		    break;
		  default:			/* miss */
		    seOp = cAndnROp;
		    break;
		  }	
		  
		  for (i=0; i < stripeWidth; i++)
		    {
		      /* rS1 = partial result from previous pipe stage,
		       * rS2 = _g0 if just doing a move,
		       *	 input if this is center's column,
		       *	 shifted input if this is not center's col
		       * rD = rOutx if this is bottom SE row,
		       *	current partial for other rows.
		       * NOTE: if rS1 and rS2 are swapped, andn won't work.
		       */
		      rS10 = rS11 = curPartials[numSERows-r][i];
		      
		      if (seOp == cOrROp)
			rS2 = _g0;
		      else if (c == cx) /* in center's column */
			{
			  if ((i & 0x1) == 0) rS2 = rInput0;
			  else rS2 = rInput1;
			}
		      else rS2 = rScr0;	/* shifting needed */			
		      if (r == numSERows-1)	/* gen final output */
			{
			  rD0 = rOut0;
			  rD1 = rOut1;
			}
		      else			/* update partials */
			rD0 = rD1 =
			  curPartials[numSERows-r-1][i];
		      
		      SEOps[i][0][r][c] = GEN_INST(seOp, rS10, rS2, rD0);
		      SEOps[i][1][r][c] = GEN_INST(seOp, rS11, rS2, rD1);
		      
		    } /* end of loop across stripe */
		} /* end of loop up first active column */
	      haveSeenActiveSECol = cTrue;
	    } /* end of first active column case */
	  
	  else	/* this is some other active column */
	    
	    {
	      /* pretty much the same as the first active column, except
	       * we don't put out anything in rows without hits or misses.
	       * We still skip the xor.
	       */
	      for (r = numSERows-1;
		   (r >= 0) && (SEOps[0][0][r][c] == 0);
		   r--)
		{
		  if (data[ (r * numSECols) + c ] /* data[r][c]*/ != 0)	/* skip "don't cares" */
		    {
		      /* choose between; hit => and, miss => andn */
		      switch (data[ (r * numSECols) + c ] /* data[r][c] */) {
		      case 1:			/* hit */
			seOp = cAndROp;
			break;
		      case 2:			/* miss */
			seOp = cAndnROp;
			break;
		      default:			/* we cannot get here since we checked
						 * for don't cares in the if statement
						 */
			seOp = cAndROp;
			break;
		      }	
		      for (i=0; i < stripeWidth; i++)
			{
			  if (c == cx) /* in center's column */
			    {
			      if ((i & 0x1) == 0) rS2 = rInput0;
			      else rS2 = rInput1;
			    }
			  else rS2 = rScr0; /* shifting needed */
			  if (r == numSERows-1)     /* gen final output */
			    {
			      rS10 = rD0 = rOut0;
			      rS11 = rD1 = rOut1;
			    }
			  else		      /* update partials */
			    rS10 = rS11 = rD0 = rD1 =
			      curPartials[numSERows-r-1][i];
			  
			  SEOps[i][0][r][c] = GEN_INST(seOp, rS10, rS2, rD0);
			  SEOps[i][1][r][c] = GEN_INST(seOp, rS11, rS2, rD1);
			  
			} /* end of loop across stripe */
		    } /* end of if that selects active SE elements */
		} /* end of loop through rows in this col */
	    } /* end of filling some column other than the first */
	} /* end of if that selects active SE cols */
    } /* end of loop through all the SE cols */
  
  
  /*-------------------------------------------------------------------*
   * SEOps is now filled with the operations to be performed.
   * It is indexed by
   *     [stripe offset][even/odd dest word address][SE row][SE col]
   * These entries can be used directly by the code that puts instructions
   * into memory.
   *-------------------------------------------------------------------*/
  
  /* START PUTTING CODE INTO THE CODE BUFFER */
  
  /* Start by filling in the call overhead */
  curBufOffset = 0;
  pCodeBuf[curBufOffset++] = GEN_SETHI_INST(LF26, _g1);
  pCodeBuf[curBufOffset++] = GEN_INST(cAddCOp, _g1, _lo(LF26), _g1);
  pCodeBuf[curBufOffset++] = GEN_INST(cSaveROp, _sp, _g1, _sp);
  
  /* Store the incoming parameters.  More C Boilerplate */
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rSrcAddr, _fp, cSrcOffset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rDstAddr, _fp, cDstOffset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rSrcPitch, _fp, cSrcPitchOffset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rDstPitch, _fp, cDstPitchOffset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rHeight, _fp, cHeightOffset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rWidth, _fp, cWidthOffset);
  
  /* Save the global registers for the sake of Purify */
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rG1, _fp, cSavedg1Offset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rG2, _fp, cSavedg2Offset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rG3, _fp, cSavedg3Offset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rG4, _fp, cSavedg4Offset);
  pCodeBuf[curBufOffset++] = GEN_ST_INST(cStCOp, rG5, _fp, cSavedg5Offset);

  /*-------------------------------------------------------------------*
   * The inner loop reads a line of words, applies them to a bunch of
   * partial results, and writes out one line of completed results.
   * This assumes that the partials are already in place.  When we begin
   * processing, we have to load the partials with something.  We load
   * them with the data from the block immediately above.  The processing
   * on the reloaded data will be repeated, but this is still a cheap
   * way to restart the pipeline.
   * First, adjust the source pointer.  It should point cy lines *above*
   * the beginning of the block so the first dest line will be affected
   * by the source cy lines above it (distance from top to center of SE).
   * Toss in one extra decrement to account for the extra increment at the
   * top of the loop.
   *-------------------------------------------------------------------*/
  for (i=0; i<(cy+1); i++)
    pCodeBuf[curBufOffset++] = GEN_INST(cSubROp, rSrcAddr, rSrcPitch, rSrcAddr);
  
  
  /*-------------------------------------------------------------------*
   * Load up the pipeline.  This is a bit confusing, at least to me.  We
   * start loading the partial registers which correspond to image lines
   * immediately above to top of the block.  These are the lines below
   * the first source line.  They will be combined with the first source
   * line using the second row of SE elements.  We work our way back to
   * lines above the source line.  These are the "almost complete" partial
   * results.
   *-------------------------------------------------------------------*/
  for (i=numSERows-1; i>0; i--)
    {
      pCodeBuf[curBufOffset++] = GEN_INST(cSubROp, rDstAddr, rDstPitch, rDstAddr);
      for (j=0; j<stripeWidth; j++)
	{
	  pCodeBuf[curBufOffset++] =
	    GEN_INST(cLdCOp, rDstAddr, j*4, curPartials[i][j]);
	}
    }
  
  /*-------------------------------------------------------------------*
   * adjust the number of words to be written to account for the extra
   * lines we've put into the pipeline plus the extra decrement that we'll
   * get at the top of the loop.  Put in an extra decrement of rDstAddr
   * to account for the extra increment at the top of the loop.
   *-------------------------------------------------------------------*/
  pCodeBuf[curBufOffset++] = GEN_INST(cAddCOp, rHeight, numSERows, rHeight);
  pCodeBuf[curBufOffset++] = GEN_INST(cSubROp, rDstAddr, rDstPitch, rDstAddr);
  
  /* We're now at the top of the main loop.
   * Remember this location since we'll need to compute
   * relative offsets to branch back to it. */
  mainLoopOffset = curBufOffset;
  
  /* Insert the looping code.  It reloads the stripe width,
   * since we'll be decrementing that as we go across the stripe.
   * It also reloads the line pitch since that register is also
   * used as one of the input regs.  We then decrement the height,
   * setting the conditition codes, add the line pitch to the
   * destination, branch if the height was non-zero to the start
   * of the loop and add the line pitch to the source in the branch's
   * delay slot.
   */
  pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, _fp, cWidthOffset, rWidth);
  pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, _fp, cSrcPitchOffset, rSrcPitch);
  pCodeBuf[curBufOffset++] = GEN_INST(cSubccCOp, rHeight, 1, rHeight);
  pCodeBuf[curBufOffset++] = GEN_INST(cAddROp, rSrcAddr, rSrcPitch, rSrcAddr);
  pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, _fp, cDstPitchOffset, rDstPitch);
  pCodeBuf[curBufOffset++] = GEN_BR_INST(cBNZOp, (cRetOverheadSize+2)<<2);
  pCodeBuf[curBufOffset++] = GEN_INST(cAddROp, rDstAddr, rDstPitch, rDstAddr);
  
  /* Here's the return code.  This is skipped in the normal case,
   * executed when we're done.  Before returning, restore the global registers
   * so Purify will work.
   */
  pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, _fp, cSavedg1Offset, rG1);
  pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, _fp, cSavedg2Offset, rG2);
  pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, _fp, cSavedg3Offset, rG3);
  pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, _fp, cSavedg4Offset, rG4);
  pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, _fp, cSavedg5Offset, rG5);

  pCodeBuf[curBufOffset++] = GEN_INST(cOrROp, _g0, _g0, _i0);
  pCodeBuf[curBufOffset++] = cRetOp;
  pCodeBuf[curBufOffset++] = cRestoreOp;
  
  /* CODE TO DO EVEN OR ODD BLOCK */
  
  /* test to see if we should do the even or odd block. */
  pCodeBuf[curBufOffset++] = GEN_INST(cAndccCOp, rDstAddr, 0x4, _g0);
  
  /* This is the address of the branch to the odd block.
   * Remember it so we can generate the branch when we figure
   * out where the odd block starts.  Skip the location for now,
   * we'll fill in the branch later.  */
  oddBlockBrOffset = curBufOffset++;
  
  /* Starting the even and odd blocks. */
  for (evenOdd = 0; evenOdd < 2; evenOdd++)
    {
      /* do the initial register loads */
      if (evenOdd == 0)	/* the "even" version of this op will
			 * be executed in the delay slot of the
			 * branch above for both even and odd cases,
			 * so don't put it into the odd case.
			 */ 
	pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, rSrcAddr, 0x0, rInput0);
      /* if there will be any right shifts, load the word to the left */
      if (numLeftSECols > 0)
	pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp, rSrcAddr, -4, rInput1);
      
      /* Loop across the stripe.  Remember that rOut0 is now empty */
      evenWordReady = cFalse;
      for (stripeOffset = 0; stripeOffset < stripeWidth; stripeOffset++)
	{
	  /* set the current and adjoining input registers
	   * for the block */
	  if ((stripeOffset & 0x1) == 0)
	    {curInputReg = rInput0; otherInputReg = rInput1;}
	  else
	    {curInputReg = rInput1; otherInputReg = rInput0;}
	  outputRegIndex = (evenOdd + stripeOffset) & 0x1;
	  
	  /* loop through all the right shifts */
	  for (c=0; c<numLeftSECols; c++)
	    {
	      if (seColActive[c])	/* make sure this column has some
					   hits or misses it in */
		{
		  /* shift data into position */
		  if (cx - c != 32)
		    {
		      pCodeBuf[curBufOffset++] =
			GEN_INST(cSrlCOp, curInputReg, (cx - c), rScr0);
		      pCodeBuf[curBufOffset++] =
			GEN_INST(cSllCOp, otherInputReg,
				(32 - (cx - c)), rScr1);
		      pCodeBuf[curBufOffset++] =
			GEN_INST(cOrROp, rScr0, rScr1, rScr0);
		    }
		  else /* shift == 32, just use entire word to the right */
		    pCodeBuf[curBufOffset++] =
		      GEN_INST(cOrROp, _g0, otherInputReg, rScr0);
		  
		  for (r = (numSERows-1); r >= 0; r--)
		    {
		      curInst = SEOps[stripeOffset][outputRegIndex][r][c];
		      if (curInst != 0)
			pCodeBuf[curBufOffset++] = curInst;
		    } /* of loop through rows of this column */
		} /* of if selecting active columns */
	    } /* of loop going through SE cols left of center */
	  
	  /* if we'll be doing any left shifts OR we'll need the word to
	   * to the right for the next stripe, load it.  */
	  if ((numRightSECols > 0) || (stripeOffset < stripeWidth-1))
	    pCodeBuf[curBufOffset++] = GEN_INST(cLdCOp,
					       rSrcAddr, 4*(stripeOffset+1), otherInputReg);
	  
	  /* if the center's column is active, do the ops */
	  if (seColActive[c])
	    {
	      for (r = (numSERows-1); r >= 0; r--)
		{
		  curInst = SEOps[stripeOffset][outputRegIndex][r][c];
		  if (curInst != 0)
		    pCodeBuf[curBufOffset++] = curInst;
		} /* of loop through rows of center's column */
	    } /* of if selecting active center column */
	  
	  /* loop through all the left shifts */
	  for (c = cx + 1, lastActiveCol = cx; c < numSECols; c++)
	    {
	      
	      /* The normal way to do a double left shift is with two shifts
	       * and an OR.  There's a trickier way.  You can do an add
	       * and addWithCarry to shift one bit.  This is two opcodes
	       * instead of three.  That makes an interesting difference in
	       * very wide SE's.
	       * This is an incremental shift - shifting my n bits takes n
	       * times as long.  The barrel shift takes three opcodes for
	       * any length (up to 32) shift.  For the sake of simplicity,
	       * we used to do the fast left shift whether or not the column
	       * was active.  However, SE's containing lots of cDontCare terms
	       * got a lot of useless shifts.  Now, we remember the location
	       * of the last shift and use the add/addx trick for single
	       * shifts and the shift/shift/OR for multiple shifts.
	       * Note that we must maintain both rScr1 and rScr0 in either
	       * case.  Note also that if this is the first time we're
	       * generating opcodes, we use curInputReg,,otherInputReg as the
	       * source.  On subsequent ops, we use rScr0,,rScr1.  We can't
	       * destroy the proper contents of curInputReg and otherInputReg
	       * because they'll be needed to process the next word in the
	       * stripe.
	       */
	      if (seColActive[c])	/* make sure this column has some
					   hits or misses it in */
		{
		  if (c - lastActiveCol == 1)	/* use add/addx trick */
		  {
		    pCodeBuf[curBufOffset++] =
			GEN_INST(cAddccROp,
				(lastActiveCol == cx ? otherInputReg : rScr1),
				(lastActiveCol == cx ? otherInputReg : rScr1),
				rScr1);
		    pCodeBuf[curBufOffset++] =
			GEN_INST(cAddxROp,
				(lastActiveCol == cx ? curInputReg : rScr0),
				(lastActiveCol == cx ? curInputReg : rScr0),
				rScr0);
		  }
		  else
		  if (c - lastActiveCol == 32)	/* weird, but could happen.
						 * need to use otherInputReg
						 * as the arg.  This must be
						 * the last shift, so there's
						 * no need to maintain rScr1.
						 */
		  {
		    pCodeBuf[curBufOffset++] =
			GEN_INST(cOrROp, _g0, otherInputReg, rScr0);
		  }
		  else				/* use shift/shift/or/shift */
		  {
		    pCodeBuf[curBufOffset++] =
			GEN_INST(cSllCOp,
				(lastActiveCol == cx ? curInputReg : rScr0),
				(c-lastActiveCol),
				rScr0);
		    pCodeBuf[curBufOffset++] =
			GEN_INST(cSrlCOp,
				(lastActiveCol == cx ? otherInputReg : rScr1),
				(32 - (c-lastActiveCol)),
				rScr1);
		    pCodeBuf[curBufOffset++] =
			GEN_INST(cOrROp, rScr0, rScr1, rScr0);
		    pCodeBuf[curBufOffset++] =
			GEN_INST(cSllCOp, otherInputReg, (c-cx), rScr1);
		  }
		  lastActiveCol = c;

		  for (r = (numSERows-1); r >= 0; r--)
		    {
		      curInst = SEOps[stripeOffset][outputRegIndex][r][c];
		      if (curInst != 0)
			pCodeBuf[curBufOffset++] = curInst;
		    } /* of loop through rows of this column */
		} /* of if selecting active columns */
	    } /* of loop going through SE cols right of center */
	  
	  /* We've put out all the morph ops for a word in the stripe.
	   * Put out the code that tests to see if it should quit
	   * processing the stripe.  BUT, don't put out the
	   * branch yet (see below).    */
	  pCodeBuf[curBufOffset++] = GEN_INST(cSubccCOp, rWidth, 0x1, rWidth);
	  brDisp = (mainLoopOffset-curBufOffset)<<2;
	  
	  /* Now, figure out whether to write something.
	   * And, if so, is it a single or a double write?
	   * We write something if we have just computed an
	   * odd word (outputRegIndex == 1) or we have just
	   * computed the last word in the stripe.  We write
	   * a double word if there's an even word ready to go
	   * (evenWordReady == cTrue).   */
	  if ((outputRegIndex == 1) || (stripeOffset == stripeWidth-1))
	    {
	      /* we need to do a write.  Put out a normal branch */
	      pCodeBuf[curBufOffset++] = GEN_BR_INST(cBZOp, brDisp);
	      
	      /* Single or double write? */
	      if (evenWordReady)
		/* do a double store */
		pCodeBuf[curBufOffset++] =
		  GEN_ST_INST(cStdCOp, rOut0, rDstAddr, 4*(stripeOffset-1));
	      
	      else	/* do a single store of the word we just computed */
		{
		  switch (outputRegIndex) {
		  case 0:
		    pCodeBuf[curBufOffset++] =
		      GEN_ST_INST(cStCOp, rOut0, rDstAddr,
				4*(stripeOffset));
		    break;
		  case 1:
		    pCodeBuf[curBufOffset++] =
		      GEN_ST_INST(cStCOp, rOut1, rDstAddr,
				4*(stripeOffset));
		    break;
		  }
		} /* of single store case */
	      evenWordReady = cFalse;	/* just wrote it out */
	    } /* of statements that do a write */
	  
	  /*------------------------------------------------------------*
	   * set evenWordReady if we just computed an even word but didn't
	   * write it out.  This happens if and only if we computed something
	   * and didn't write it out.  BUT, what if this is the last word of
	   * this particular stripe?  This could happen if the stripe width
	   * passed in was less than the maxiumum allowed for this code.
	   * If that happened, we should a write after all.  To have the
	   * write happen if and only if this happened to be the last
	   * word in the stripe, we put in an annulled branch and a
	   * store op in the delay slot.  If the branch is not taken (there
	   * are more words in the stripe), the store is ignored.  If the
	   * stripe ends here, the store is executed on the way back to
	   * the beginning of the loop.
	   *------------------------------------------------------------*/
	  else
	    {
	      evenWordReady = cTrue;
	      pCodeBuf[curBufOffset++] = GEN_BR_INST(cBZaOp, brDisp);
	      pCodeBuf[curBufOffset++] = 
		GEN_ST_INST(cStCOp, rOut0, rDstAddr, 4*(stripeOffset));
	    }
	  
	} /* of loop that goes through a stripe */
      
      /* we're now at the end of either the even or the odd block.  If we're
       * at the end of the even block, we know the odd block's address,
       * so can insert the branch instruction now.
       */
      if (evenOdd == 0)	/* just finished the even block */
	{
	  brDisp = (curBufOffset-oddBlockBrOffset) << 2;
	  pCodeBuf[oddBlockBrOffset] = GEN_BR_INST(cBNZOp, brDisp);
	}
    } /* of loop that goes through even and odd blocks */

	/* Add a word of 0's just past the last word of code.  We should
	 * never get here, but if we do, it will cause a immediate
	 * unimplemented instruction trap and the registers should
	 * be preserved.
	 */
    pCodeBuf[curBufOffset++] = GEN_INST(cUnimpOp, _g0, _g0, _g0);

    if (4*(curBufOffset) >= bufSizeInBytes)
    {
	__error__ = ia_bufferExceeded;
	goto genCleanup;
    }

    /* return the number of bytes actually put into the buffer */
  *code_length = 4*curBufOffset;

    /* UCKYNESS!!
     * On the SPARC 5 machines, we get mysterious illegal instructure traps.
     * Those machines have separate instruction and data caches.  We're
     * going to guess that the instructions we've just written don't get
     * to the instruction cache on time.  First, try writing extra stuff
     * to flush the write buffers.
     * P.S. putting this in fixed the problem, so leave it in.
     */
  for (r = 0; r < cMaxSERows; r++)
    SEOps[0][0][r][0] = 0;

    /* Now try flushing the I cache at these locations,  the buffer length
     * is given in bytes
     */
  i_flushICache (pIPState, (void *)(*generated_code), *code_length);
  __error__ = ia_successful;

genCleanup:
  if ((__error__ != ia_successful) && (bufAllocatedHere))
    FREE(pCodeBuf);
  NORMAL_RETURN;
}

/*

$Log: ssgenero.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.31  1998/01/13 23:54:18  ddavies
ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.30  1996/12/19 23:07:47  ddavies
ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v

Have dummy implementation of i_getLongErodeAsmSize return ia_noImpl
instead of ia_successful so upper levels of code know
not to allocate a buffer of size 0.  Purify refuses to free a buffer of
size 0, then complains that it's been leaked.

Revision 1.29  1996/12/18 23:06:06  ddavies
ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v

Add dummy implementations of i_getLongErodeAsmSize,
i_genLongErodeAsm and ip_getLongErodeSEParms.  Also add implementation
of ip_longOpIPLibLinked (returns cFalse since code capable of doing
long operations (>= 64 bit) is not present in the SPARC world (yet).

Revision 1.28  1995/12/13 02:52:54  ejaquith
ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.27  1995/10/03  21:17:26  ddavies
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Include shrmorph.pub/prv instead of shrip.pub/prv.  Change arg of
 * i_genErodeAsm from "void **generated_code" to "ErodeProc *generated_code".
 * Change casts when using generated_code.
 *
 * Revision 1.26  1995/07/26  21:20:58  ddavies
 * ddavies on Wed Jul 26 14:19:57 PDT 1995
 *
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Add an unimplemented opcode (0x00000000) to the end of the generated
 * SPARC assembly code for erosion.  This will cause an immediate trap,
 * hopefully with all state intact, if someone falls off the end.
 *
 * Revision 1.25  1994/12/08  04:38:20  ddavies
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Check to see if structure element contains any hits or any misses in
 * i_genErodeAsm.  If not, return ia_SETooBig.  This really means "the
 * structure element is too weird," in this case.
 *
 * Revision 1.24  1994/09/20  21:31:35  ddavies
 * ddavies on Tue Sep 20 14:30:35 PDT 1994
 *
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Change bufSizeInBytes to be a UInt32 from an Int32 so it can be used
 * as an arg to i_getErodeAsmSize without complaint.
 *
 * Revision 1.23  1994/09/20  02:55:39  ddavies
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Changes to get rid of allocator, alloc_shrinker and generally allow
 * clients to pass in buffers for generated code.  Add a version of i_getErodeAsmSize
 * for the SPARC.  Make i_genErodeAsm call it.  Get rid of the generated_code
 * and alloc_shrinker args since we no longer free the buffer on error.
 * Return the code_length as bytes, not 32-bit words.
 *
 * Revision 1.22  1994/08/11  17:31:52  ddavies
 * ddavies on Thu Aug 11 10:31:17 PDT 1994
 *
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Fix bug in call to ip_flushICache.  We were passing the length of the
 * buffer to be flushed in words, not bytes.
 *
 * Revision 1.21  1994/08/11  01:35:13  ddavies
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Add call to ip_flushICache immediately after generating new assembly code.
 * Yeah, it would have been tidier to do this after deallocating the strel,
 * but that's done in ipwrap (which isn't used by all our clients), so we'll
 * do it here.
 *
 * Revision 1.20  1994/08/01  21:08:33  ddavies
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Change "static const" to "static" to work around buggy gcc/Purify interaction.
 *
 * Revision 1.19  1994/07/27  23:19:38  ejaquith
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Made partialRegs a static const declaration for reentrancy.
 *
 * Revision 1.18  1994/02/04  20:20:37  dferrell
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Changed memory.prv to memory.pub
 *
 * Revision 1.17  1993/12/02  19:29:53  ddavies
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.16  1993/11/05  02:24:24  ddavies
 * ssgenero.c -> /product/ipcore/ipshared/src/RCS/ssgenero.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


