/*************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  ssgengly.c:  contains routines that generate machine code to do
 *		 convolutions needed to decode glyphs.
 *
 *
 *	    Contains SPARC implementations of
 *		  UInt32	ip_gen1x1ConvAsm()
 *		  UInt32	ip_gen3x3ConvAsm()
 *
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"
#include "memory.pub"
#include "sparc.prv"
#include "kernel.prv"

    /*  prototypes */
#include "shrglyph.prv"		/* If we were using any of the public
				 * pointers to private structs, this would
				 * have to go first since it's where the
				 * private structures are defined.  Today,
				 * we're not using any, but you never know. */
#include "shrglyph.pub"
#include "shrip.pub"		/* for i_flushICache */

IP_RCSINFO(RCSInfo, "$RCSfile: ssgengly.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");
/* SET THIS cTrue TO GET ERROR MESSAGES */
#ifdef DEBUG
#undef DEBUG
#endif /* DEBUG */
/* #define	DEBUG	0 */

/* Most of the comments for the glyph functions are in cglyph.c.  This file
 * contains the SPARC-specific parts of ip_genConvAsm, which generates
 * machine code to perform glyph convolutions.
 *
 * The code generated looks a lot like convolve1x1 and ip_convolve3x3 in
 * sglyph.c.  The reader should use that code as a guide.  The
 * only real difference is that we remove the loops, replacing them with
 * appropriate amounts of straight-line code.
 *
 * To keep me sane, we'll actually have different procedures to generate
 * the two types of convolution code.  One type of convolution code produces
 * a single convolution sum (like sglyph.c's version of convolve1x1).
 * The other type generates a 3x3 array of output sums, picks the biggest
 * one that's less than pKernel->upperThreshold and returns it while also 
 * modifying the image pointers to point to the position corresponding 
 * to the biggest sum (like sglyph.c's version of ip_convolve3x3).
 * ip_genConvAsm simply figures out which type of code to generate and 
 * calls the appropriate procedure.
 *
 * gen1x1ConvAsm's basic algorithm is:
 *	1. Determine the number of words required for the assembly code.
 *	2. Allocate a suitable buffer.
 *	3. Fill the Buffer with assembly code.
 *	4. Return a pointer to the buffer in pKernel->conv1Code or
 *	   pKernel->conv3Code.
 *
 * Both the 1x1 and 3x3 convolution algorithms operate on a range 
 * of kernel sizes and image depths.  A feature of all convolution 
 * algorithms is that the input bits required to retire one row of 
 * the convolution kernel fit into a single 32-bit word (possibly after 
 * some shifting to put them there).
 *
 * Doing things efficiently in the SPARC means keeping everything you need
 * in registers.  How many registers are available?  The SPARC has
 * 32 registers.  One of them holds the constant 0.  Four are used to pass
 * arguments, but we use all four of these later, so we'll count them
 * as available.  Global registers 6 and 7 are reserved for the system.
 * Out register 6 holds the stack pointer.  In registers 6 and 7 hold the
 * frame pointer and return address respectively.  That leaves us
 * 32 - 1 - 2 - 1 - 2 = 26 registers.
 *
 * The basic method of doing convolution is to grab input bits in units
 * 8 bits or less, use these bits as an index into a series of lookup tables,
 * and add the numbers from the lookup tables to the running convolution
 * sum or sums.  The tables contain pre-computed inner products of image
 * pixels times convolution weights.
 *
 * If the product of the number of active columns in the
 * convolution kernel times the image depth is less than or equal to 8,
 * we do a single table lookup per kernel row.  If the total is greater
 * than 8, we do one lookup for each group of 8 plus a lookup for any bits
 * left over.
 *
 * The major difference between the routines producing a single convolution
 * sum and the ones producing 9 sums is obviously the amount of adding that
 * goes on.  In the 1x1 case, there is one table pointer and one add per
 * group of inputs.  In the 3x3 case, there are 3 table pointers and 3
 * adds per sum.  Why not 9?  Because each table entry holds 3 sums and
 * the 3 registers each hold 3 sums.
 *
 * A lesser difference is the way parameters are passed and the style of the
 * results.  In 1x1 convolution, we pass in information used to find the
 * upper left corner of the glyph, image words/line and a pointer to the
 * kernel data structure.  The convolution sum is returned.  This is
 * straightforward.  For efficiency reasons, the 3x3 case is more complex.
 * The 3x3 routine is used to find the center of each glyph.  We pass 
 * in *pointers to* image information and and a pointer to the kernel.
 * After the 9 sums are computed, this routine finds the largest sum less
 * than the upper threshold.  In addition to returning this sum, it
 * modifies the image address and offset information so further routines can
 * point directly at our best guess of the glyph's position without
 * further computation.
 *
 * The exact parameters to convolve1x1 are:
 *	pImage:		Pointer to word containing upper left corner of
 *			image block to be convolved.
 *	centerXBit:	Holds position of leftmost bit
 *			within word pointed to by pImage
 *	wpl:		Number of words/line in image.
 *	pKernel:	Pointer to IAKernel data structure used to
 *			define the convolution kernel.
 *
 * This is a good place to lay out the template of the assembly code  Here's
 * the template for convolve1x1.  When the code below says "Loop," it does
 * NOT mean that there are loops in the generated machine code.  It means
 * that the generating C code should loop through code that puts out
 * machine code.  The difference still confuses me.  The machine code has
 * several cases, but always proceeds forwards.  There are no backwards
 * jumps in it.
 *
 ************************ PURIFY fixup **********************************
 *
 * The Purify program from Pure, Inc. tracks memory usage.  It's a swell tool.
 * However, it gloms global registers for it's own use.  They normally get
 * away with this because they go through ALL code that's linked in and
 * make sure that everyone else leaves their register alone.  Since this code
 * is generated at runtime, they can't adjust it.  Bummer.  We'll go to the
 * trouble of saving the global registers on entry and restoring them on
 * exit.   We use 5 "local variables" for this purpose.  It appears that
 * this means we claim 120 bytes on the stack (instead of 64 - why does
 * adding 20 bytes of local variable mean we increase the local frame by
 * 56 bytes?).  The first opcodes store the globals, the opcodes just
 * before the return restore them.
 *
 ************************************************************************
 *
 * *********************** convolve1x1 ***************************
 *
 *   Call Overhead Ops
 *   {
 *	new register window (3 ops)
 *	save incoming parms in local vars (4 ops)
 *	save global registers for Purify (5 ops)
 *   }
 *   Initial Parameter Adjustment Ops
 *   {
 *	Set rBpl = 4*rBpl (had contained words/line, 1 op)
 *	Add offset due to vacant kernel bits to rCenterXBit (2 ops)
 *	Test to see if rCenterXBit went over 31.  If so, increment rImage
 *		and truncate rCenterXBit (4 ops)
 *	Add in address offset due to vacant kernel rows
 *		(pKernel->firstActiveRow ops)
 *	Load rCompTableRow (1 op)
 *	Load rBytesPerTableRow (1 op)
 *	Set rShiftCount to 32 - (rCenterXBit + pKernel->numBitsPerRow) (4 ops)
 *   }
 *   Branch based on value of rShiftCount (if it's > 1, we'll always shift
 *	image data right.  If it's 1 or 0, we do single left shifts.
 *	if it's < 0, we need to do double left shifts.  There will be
 *	a complete section of convolution code for each case.
 *   {
 *	Test rShiftCount (1 op)
 *	If rShiftCount <= 1,
 *	{
 *	    Branch to Case where image should be shifted left (1 op)
 *	    Something useful in delay slot (1 op)
 *	}
 *   }
 *
 *   **** Case where image bits should be shifted right ****
 *   Parameter setup
 *   {
 *	Set rShiftCount = rShiftCount-2	(equals right shift dist, 1 op)
 *   }
 *   Loop Through Rows of Kernel (do this pKernel->numActiveRows times)
 *   {
 *	Set pCompTableBlock = pCompTableRow (1 op)
 *	Set rInput = ([rImage] >> rShiftCount) (2 ops)
 *	If there are any row after this one,
 *	    rImage += rBpl (1 op)
 *	Loop Through column blocks of Kernel (do this
 *	(pKernel->numBitsPerRow+7)/8 times)
 *	{
 *	    If there are at least 8 bits remaining in the input row,
 *		Set rBlockBits = rInput & 0x3FC (1 op)
 *	    Else
 *		Set rBlockBits = rInput & lastMask (1 op)
 *	    If this is the first time we've touched rConvSum and this row
 *	    of the kernel is non-zero,
 *		set rConvSum = [rCompTableBlock+rBlockBits] (1 op)
 *	    Else if this row of the kernel is non-zero.
 *		set rScratch = [rCompTableBlock+rBlockBits] (1 op)
 *	    If there are more blocks in this row,
 *	    {
 *		Shift rInput right by 8 (1 op).
 *		pCompTableBlock += 256 * sizeof(Int32) (1 op)
 *	    }
 *	    Else if there are any more rows
 *	    {
 *		rCompTableRow += rBytesPerTableRow (1 op)
 *	    }
 *	    If that was NOT the first time we've touched rConvSum and
 *	    this row of the kernel is non-zero,  We wait to do this
 *	    to fill the interlock between loading rScratch and using it.
 *		rConvSum += rScratch (1 op)
 *	}
 *   }
 *   restore global registers for Purify (5 ops)
 *   return rConvSum
 *
 *
 *   **** Case where image bits should be single shifted left ****
 *   Compare rShiftCount to 0 (1 op)
 *   if rShiftCount < 0,
 *   {
 *	Branch to double left shift case (1 op)
 *	Something useful in delay slot (1 op)
 *   }
 *   Parameter setup
 *   {
 *	Set rShiftCount = 2-rShiftCount	(equals left shift dist, 2 ops)
 *   }
 *   Loop Through Rows of Kernel (do this pKernel->numActiveRows times)
 *   {
 *	Set pCompTableBlock = pCompTableRow (1 op)
 *	Set rInput = ([rImage] >> rShiftCount) (2 ops)
 *	If there are any row after this one,
 *	    rImage += rBpl (1 op)
 *	Loop Through column blocks of Kernel (do this
 *	(pKernel->numBitsPerRow+7)/8 times)
 *	{
 *	    ... same as above ...
 *	}
 *   }
 *   restore global registers for Purify (5 ops)
 *   return rConvSum
 *
 *
 *   **** Case where image bits should be double shifted left ****
 *   **** rShiftCount holds a negative number ****
 *   Parameter setup
 *   {
 *	Set rRightShiftCount = 30 + rShiftCount (equals right shift dist
 *				for word to the right, 1 op)
 *	Set rShiftCount = 32 - rRightShiftCount (equals left shift dist, 2 ops)
 *   }
 *   Loop Through Rows of Kernel (do this pKernel->numActiveRows times)
 *   {
 *	Set pCompTableBlock = pCompTableRow (1 op)
 *	Set rInput = ([rImage] << rShiftCount) (2 ops)
 *	Set rInputRight = ([rImage+4] >> rRightShiftCount) (2 ops)
 *	Set rInput = rInput | rInputRight (1 op)
 *	Loop Through column blocks of Kernel (do this
 *	(pKernel->numBitsPerRow+7)/8 times)
 *	{
 *	    ... same as above ...
 *	}
 *   }
 *   restore global registers for Purify (5 ops)
 *   return rConvSum
 *
 **********************************************************************
 *
 * A few words on code construction.  How do we do branches?  We remember
 * That address of every branch target (if it's a backwards branch) or
 * the address of the branch itself (if it's a forwards branch).  When we
 * get to the branch or branch target respectively, we compute the branch
 * distance accordingly.
 *
 * How do we deal with the code buffer?  The easiest way to to have a
 * pointer to the buffer and an offset from the pointer.  The offset
 * provides relative addresses for computing branch addresses.  It also
 * becomes the size of the code when we're done.
 *
 */

/* Input and constant register definitions */
/* The C Preprocessor can't handle the standard notation for SPARC
 * registers, "%i0".  We use an underline instead of percent sign.  */
#define	rImage			_i0
#define rCenterXBit		_i1
#define	rBpl			_i2
#define	rKernel			_i3
#define	rNumOutBit		_i4

#define	rCompTableRow		_g1

#define	rBytesPerTableRow	_g2
#define	rCompTableBlock		_g3	/* used in convolve1x1 */
#define	rCompTableBlock0	_g3	/* used in ip_convolve3x3 */
#define rCompTableBlock1	_g4
#define	rCompTableBlock2	_g5

#define	rConvSum		_o0
#define	rConvSum0		_o1
#define	rConvSum1		_o2
#define	rConvSum2		_o3

#define	rShiftCount		_o4
#define	rRightShiftCount	_o5

#define	rInput			_l0
#define	rInputRight		_l1

#define	rScratch		_l2
#define	rScratch0		_l3
#define	rScratch1		_l4
#define	rScratch2		_l5
#define	rBlockBits		_l6

/* Names of global registers.  These are saved and restored so Purify will
 * work.
 */
#define	rG1			_g1
#define	rG2			_g2
#define	rG3			_g3
#define	rG4			_g4
#define	rG5			_g5

/* Sizes of various blocks of code */
#define	cCall1x1OverheadSize	12	/* includes saving global regs */
#define	cParm1x1SetupSize	13
#define	cShiftRightTestSize	3
#define	cRowLoop1x1SetupSize1	3
#define	cRowLoop1x1SetupSize2	3
#define	cRowLoop1x1SetupSize3	7
#define	cColBlock1x1Size	5
#define	cRowLoop1x1BottomSize	2
#define	cRetOverheadSizeSize	8	/* includes restoring global regs */

/* Magic C Routine Boilerplate */
#define	LF26			-120	/* room for 5 local variables */
#define	cImageOffset		0x44
#define	cCenterXBitOffset	0x48
#define	cBplOffset		0x4c
#define	cKernelOffset		0x50
#define cNumOutBitOffset	0x54

/* Local variable offsets */
#define cSavedg1Offset		-4
#define cSavedg2Offset		-8
#define cSavedg3Offset		-12
#define cSavedg4Offset		-16
#define cSavedg5Offset		-20

/* This is a private routine */
/* PRIVATE */
/*
 *  ip_gen1x1ConvAsm():
 *        This routine generates machine code to process a kernel of a
 *	  particular shape.  The code is specialized for each platform.
 *	  The code generated here produces exactly 1 convolution result.
 *
 *	pKernel:	Pointer to IAKernel data structure containing
 *			information about the kernel's shape.
 *	allocator:	Ptr to function that is used to allocate memory
 *			for the code.  Function's single argument is
 *			the number of BYTES requested.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of malloc
 *			provided by WATCOM on the PC!!
 *	freer:		Ptr to function used to free memory in case of
 *			error
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of free
 *			provided by WATCOM on the PC!!
 *	codeLength	Pointer to length of code generated in units of
 *			UInt32's.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		couldn't allocate internal buffers
 *		ia_bufferExceeded	code generated didn't fit into buffer
 *					
 */
Int32 CDECL 
ip_gen1x1ConvAsm (
    IPState    *pIPState,
    IAKernel   *pKernel,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	),
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *),
    UInt32     *codeLength)
{
    PROLOGUE;
    Int32	 r, c, i;		/* general iterative */
    UInt32	*pCodeBuf;		/* pointer to the code block */
    Int32	 bufSizeInBytes;	/* size of code buffer */
    Int32	 addr;			/* offset to currently generated
					 * inst word */
    Int32	 testShiftCnt1, testShiftCnt2;	/* offsets of test ops */
    Int32	 sumTouched;		/* BOOLEAN, true if convSum has been
					 * touched. */
    Int32	 need2TabPtrs;		/* BOOLEAN, true if we'll need both
					 * a pointer to each table row and
					 * a pointer to each block in a row.
					 * cFalse if there's only one block per
					 * row. */
    UInt32	 lastMask;		/* mask used to select the last
					 * set of bits in each input row. */
    UInt32	 activeRow[cMaxGlyphRows]; /* array of BOOLEANS.
					 * activeRow[r] is cTrue if there are
					 * non-zero weights in kernel row r */

    need2TabPtrs = (pKernel->numBitsPerRow > 8);
    lastMask = (0xFF >> (8 - (pKernel->numBitsPerRow % 8)) << 2);

	/* Find holes in the kernel inside the region of active rows.
	 * We won't put out opcodes to do convolution in those rows. */
    for (r = 0, i = pKernel->firstActiveRow * pKernel->width;
	 r < pKernel->numActiveRows;
	 r++)
    {
	activeRow[r] = cFalse;
	for (c = 0; c < pKernel->width; c++, i++)
	{
	    if (pKernel->weights[i] != 0)
	    {
		activeRow[r] = cTrue;
		i += pKernel->width - c - 1;	/* look at next row AFTER
						 * increment at loop bottom */
		c = pKernel->width - 1;		/* ditto */
	    }
	}
    }

	/* Figure out how much space we'll need. */
    bufSizeInBytes = 4*(cCall1x1OverheadSize +	/* call overhead */
			cParm1x1SetupSize + pKernel->firstActiveRow +
			cShiftRightTestSize +
				/* for each row, we have setup plus column
				 * processing ops.  There is one set of
				 * column processing ops for each block of
				 * 8 bits plus one for the last set of
				 * leftover bits.  However, the first
				 * column processing op in the first row
				 * is 1 shorter, we subtract that op
				 * off at the end of the row loops.
				 * Also, the last col block in
				 * every row is 2 ops shorter, so we
				 * subtract 2 from the col block total.  Also,
				 * we don't do one op of the bottom loop
				 * overhead the last time through the row
				 * loop, so subtract it back off. */
			pKernel->numActiveRows *
			   (cRowLoop1x1SetupSize1 +
			    ((((pKernel->numBitsPerRow+7)/8) * cColBlock1x1Size)
			     - 2) +
			    cRowLoop1x1BottomSize) +
			-(cRowLoop1x1BottomSize) +
			cRetOverheadSizeSize +

				/* for block that shifts left */
			cShiftRightTestSize +
			pKernel->numActiveRows *
			   (cRowLoop1x1SetupSize2 +
			    ((((pKernel->numBitsPerRow+7)/8) * cColBlock1x1Size)
			     - 2) +
			    cRowLoop1x1BottomSize) +
			-(cRowLoop1x1BottomSize) +
			cRetOverheadSizeSize +

				/* for block that double shifts left */
			pKernel->numActiveRows *
			   (cRowLoop1x1SetupSize3 +
			    ((((pKernel->numBitsPerRow+7)/8) * cColBlock1x1Size)
			     - 2) +
			    cRowLoop1x1BottomSize) +
			-(cRowLoop1x1BottomSize) +
			cRetOverheadSizeSize);

	/* get the memory to be used for the code */
    pCodeBuf = (UInt32 *) ((allocator) (pIPState, bufSizeInBytes));

	/* quit if we didn't get any memory */
    if (pCodeBuf == NULL)
	return ia_nomem;

	/* Start filing the buffer */
    addr = 0;		/* initial offset into buffer */
    pCodeBuf[addr++] = GEN_SETHI_INST(LF26, _g1);
    pCodeBuf[addr++] = GEN_INST(cAddCOp, _g1, _lo(LF26), _g1);
    pCodeBuf[addr++] = GEN_INST(cSaveROp, _sp, _g1, _sp);
  
	/* Store the incoming parameters.  More C Boilerplate */
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rImage, _fp, cImageOffset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rCenterXBit, _fp, cCenterXBitOffset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rBpl, _fp, cBplOffset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rKernel, _fp, cKernelOffset);

	/* Save the global registers for the sake of Purify */
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG1, _fp, cSavedg1Offset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG2, _fp, cSavedg2Offset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG3, _fp, cSavedg3Offset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG4, _fp, cSavedg4Offset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG5, _fp, cSavedg5Offset);

	/* Set up the incoming parameters */
	/* Add offsets due to vacant kernel rows to rCenterXBit */
    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCenterXBit, \
				       pKernel->firstActiveColBit, \
				       rCenterXBit);
	/* Test to see if rCenterXBit is now over 31. */
    pCodeBuf[addr++] = GEN_INST(cAndccCOp, rCenterXBit, 0x20, _g0);
    pCodeBuf[addr++] = GEN_BR_INST(cBZOp, 4*3);	/* skip 2 opcodes */
    pCodeBuf[addr++] = GEN_INST(cAndCOp, rCenterXBit, 0x1F, rCenterXBit);
	/* this is the only one that's really skipped, truncating rCenterXBit
	 * is always safe anyway */
    pCodeBuf[addr++] = GEN_INST(cAddCOp, rImage, 0x4, rImage);

	/* Convert words/line to bytes/line */
    pCodeBuf[addr++] = GEN_INST(cSllCOp, rBpl, 2, rBpl);

	/* add in the address offset due to vacant kernel rows.  There are
	 * usually 0 of these, sometimes 1, so repeated adding is a lot
	 * faster than multiplying. */
    for (r = pKernel->firstActiveRow; r > 0; r--)
    {
	pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
    }

	/* load rCompTableRow with pointer to tables of compiled weights */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rKernel, cIAKCompWeights, rCompTableRow);
	/* we'll load rBytesPerTableRow in the delay slot of the branch below */

	/* load rScratch with pKernel->numBitsPerRow */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rKernel, cIAKNumBitsPerRow, rScratch);
	/* load rShiftCount with 32 in interlock slot */
    pCodeBuf[addr++] = GEN_INST(cOrCOp, _g0, 32, rShiftCount);
	/* put 32 - pKernel->numBitsPerRow into rShiftCount */
    pCodeBuf[addr++] = GEN_INST(cSubROp, rShiftCount, rScratch, rShiftCount);
	/* put 32 - (pKernel->numBitsPerRow + rCenterXBit) into rShiftCount */
    pCodeBuf[addr++] = GEN_INST(cSubROp, rShiftCount, rCenterXBit, rShiftCount);

	/* if rShiftCount is less than 1 (using signed comparisons)
	 * we'll do left shifts.  Those are the target of the branch. */
    pCodeBuf[addr++] = GEN_INST(cSubccCOp, rShiftCount, 0x1, _g0);
	/* This is the location of the first branch on the value of
	 * rShiftCount.  Remember this offset since we'll need to set
	 * the branch distance later */
    testShiftCnt1 = addr;
    pCodeBuf[addr++] = GEN_BR_INST(cBLEOp, 0);	/* fill in offset later */
	/* load rBytesPerTableRow in the delay slot */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rKernel, cIAKBytesPerTabRow, 
				rBytesPerTableRow);


	/* **************** 1x1 Right Shift Case *********************** */


	/* In this set of code, we'll be doing right shifts of image words.
	 * That means we load up each image word and shift it right to
	 * get it into position.
	 * We do this for every non-zero kernel row.  First, set rShiftCount to
	 * rShiftCount-2 to get the proper right shift distance for
	 * the image data read. */
    pCodeBuf[addr++] = GEN_INST(cSubCOp, rShiftCount, 0x2, rShiftCount);
    sumTouched = cFalse;

	/* Loop through the kernel rows, emitting opcodes for each one. */
    for (r = pKernel->numActiveRows - 1; r >= 0; r--)
    {
	    /* Get the image word for this row */
	if (activeRow[r])
	{
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
		/* load table pointer in interlock */
	    if (need2TabPtrs)
		pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rCompTableRow,
						       rCompTableBlock);
	}
	    /* Sometimes we only need one table pointer, so that didn't
	     * happen.  We still need something to fill the interlock.
	     * if there are more rows in the kernel table, increment
	     * rImage to point to the next row of corresponding image bits. */
	if (r > 0)
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);

	    /* Shift the input word right to align it */
	if (activeRow[r])
	    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInput, rShiftCount, rInput);

	    /* Loop through the blocks in this row. */
	for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
	{
		/* Put the resulting table entry into convSum.  If this is
		 * the first time we've touched rConvSum, just load it.
		 * otherwise, Add it to the existing value.
		 * BUT FIRST, select hte portion of the inputs we
		 * want for this column block.  If c >= 8, we use
		 * the next 8 bits.  If not, we use the bits convered
		 * by lastMask. */
	    if ((c < 8) && activeRow[r])
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask,
						   rBlockBits);
	    if ((c >= 8) && activeRow[r])
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    if (!sumTouched) /* just load rConvSum */
	    {
		if (need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock,
						      rBlockBits,
						      rConvSum);
		if (!need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableRow,
						      rBlockBits,
						      rConvSum);
		if ((c > 8) && activeRow[r])	/* there's more to do
						 * in this row */
		{
		    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 0x8, rInput);
		    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock, 1024,
						       rCompTableBlock);
		}
		    /* if there's no more to do in this row, and there are
		     * more rows, point to them.  NOTE that we do NOT
		     * look at activeRow[r] here because this should
		     * happen whether or not the current row is active.
		     * So why do we put it here?  So it will look like
		     * the sumTouched == cTrue case below where we need
		     * the opcode to fill the interlock between loading
		     * rScratch and adding it to convSum. */
		if ((c <= 8) && (r > 0))
		{
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						       rBytesPerTableRow,
						       rCompTableRow);
		}
		if (activeRow[r])
		    sumTouched = cTrue;
	    }	/* at the end of this, rConvSum is updated and we've put
	         * out the opcodes needed to either get us to the next
		 * block in this column or the next row. */
	    else
		/* Need to add the thing fetched to rConvSum */
	    {
		if (need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock,
						      rBlockBits,
						      rScratch);
		if (!need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableRow,
						      rBlockBits,
						      rScratch);

		if ((c > 8) && activeRow[r])	/* there's more to do
						 * in this row */
		{
		    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 0x8, rInput);
		    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock, 1024,
						       rCompTableBlock);
		}
		    /* if there's no more to do in this row, and there are
		     * more rows, point to them.  This goes here to fill
		     * the interlock between fetching rScratch and adding it
		     * to rConvSum.  Note that we do NOT look at activeRow[r]
		     * here since this action is independent of whether
		     * this row had anything in it. */
		if ((c <= 8) && (r > 0))
		{
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						       rBytesPerTableRow,
						       rCompTableRow);
		}

		    /* The interlock should be as filled as it's going to get,
		     * add the thing fetched to rConvSum. */
		if (activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum, rScratch,
						       rConvSum);
	    }	/* at the end of this, rConvSum is updated and we've put
	         * out the opcodes needed to either get us to the next
		 * block in this column or the next row. */

	}	/* end of loop generating code to cover columns */
    }	/* end of loop generating code to cover rows in right shift case */

	/* we're done, so just return the result.  But first, restore the
	 * global registers saved above.
	 */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg1Offset, rG1);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg2Offset, rG2);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg3Offset, rG3);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg4Offset, rG4);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg5Offset, rG5);

    pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rConvSum, _i0);
    pCodeBuf[addr++] = cRetOp;
    pCodeBuf[addr++] = cRestoreOp;

	/* This is where we test to see if rShiftCount is negative.  If so,
	 * we jump down to do double left shifts.
	 * But first, since this is the target of the branch above, we
	 * should modify its branch displacement.
	 */
    pCodeBuf[testShiftCnt1] |= (addr - testShiftCnt1);

	/* if rShiftCount is less than 0 (using signed comparisons)
	 * we'll do double left shifts.  Those are the target of the branch. */
    pCodeBuf[addr++] = GEN_INST(cSubccCOp, rShiftCount, 0x0, _g0);
	/* This is the location of the second branch on the value of
	 * rShiftCount.  Remember this offset since we'll need to set
	 * the branch distance later */
    testShiftCnt2 = addr;
    pCodeBuf[addr++] = GEN_BR_INST(cBLOp, 0);	/* fill in offset later */
	/* Compute rShiftCount-2 in the delay slot.  In the single shift
	 * case, we'll just negate this to get the shift count.  In
	 * the double shift case, we add this to 32 to get the right
	 * shift count, then negate it to get the left shift count.
	 */
    pCodeBuf[addr++] = GEN_INST(cSubCOp, rShiftCount, 0x2, rShiftCount);


	/* **************** 1x1 Single Left Shift Case ******************* */


	/* In this set of code, we'll be doing single left shifts of
	 * image words.
	 * That means we load up each image word and shift it left to
	 * get it into position.
	 * We do this for every non-zero kernel row.  First, set rShiftCount to
	 * -rShiftCount to get the proper left shift distance for
	 * the image data read. */
    pCodeBuf[addr++] = GEN_INST(cSubROp, _g0, rShiftCount, rShiftCount);
    sumTouched = cFalse;

	/* Loop through the kernel rows, emitting opcodes for each one. */
    for (r = pKernel->numActiveRows - 1; r >= 0; r--)
    {
	    /* Get the image word for this row */
	if (activeRow[r])
	{
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
		/* load table pointer in interlock */
	    if (need2TabPtrs)
		pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rCompTableRow,
						       rCompTableBlock);
	}
	    /* Sometimes we only need one table pointer, so that didn't
	     * happen.  We still need something to fill the interlock.
	     * if there are more rows in the kernel table, increment
	     * rImage to point to the next row of corresponding image bits. */
	if (r > 0)
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);

	    /* Shift the input word left to align it */
	if (activeRow[r])
	    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);

	    /* Loop through the blocks in this row. */
	for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
	{
		/* Put the resulting table entry into convSum.  If this is
		 * the first time we've touched rConvSum, just load it.
		 * otherwise, Add it to the existing value.
		 * BUT FIRST, select hte portion of the inputs we
		 * want for this column block.  If c >= 8, we use
		 * the next 8 bits.  If not, we use the bits convered
		 * by lastMask. */
	    if ((c < 8) && activeRow[r])
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask,
						   rBlockBits);
	    if ((c >= 8) && activeRow[r])
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC,
						   rBlockBits);

	    if (!sumTouched) /* just load rConvSum */
	    {
		if (need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock,
						      rBlockBits,
						      rConvSum);
		if (!need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableRow,
						      rBlockBits,
						      rConvSum);
		if ((c > 8) && activeRow[r])	/* there's more to do
						 * in this row */
		{
		    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 0x8, rInput);
		    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock, 1024,
						       rCompTableBlock);
		}
		    /* if there's no more to do in this row, and there are
		     * more rows, point to them.  NOTE that we do NOT
		     * look at activeRow[r] here because this should
		     * happen whether or not the current row is active.
		     * So why do we put it here?  So it will look like
		     * the sumTouched == cTrue case below where we need
		     * the opcode to fill the interlock between loading
		     * rScratch and adding it to convSum. */
		if ((c <= 8) && (r > 0))
		{
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						       rBytesPerTableRow,
						       rCompTableRow);
		}
		if (activeRow[r])
		    sumTouched = cTrue;
	    }	/* at the end of this, rConvSum is updated and we've put
	         * out the opcodes needed to either get us to the next
		 * block in this column or the next row. */
	    else
		/* Need to add the thing fetched to rConvSum */
	    {
		if (need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock,
						      rBlockBits,
						      rScratch);
		if (!need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableRow,
						      rBlockBits,
						      rScratch);

		if ((c > 8) && activeRow[r])	/* there's more to do
						 * in this row */
		{
		    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 0x8, rInput);
		    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock, 1024,
						       rCompTableBlock);
		}
		    /* if there's no more to do in this row, and there are
		     * more rows, point to them.  This goes here to fill
		     * the interlock between fetching rScratch and adding it
		     * to rConvSum.  Note that we do NOT look at activeRow[r]
		     * here since this action is independent of whether
		     * this row had anything in it. */
		if ((c <= 8) && (r > 0))
		{
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						       rBytesPerTableRow,
						       rCompTableRow);
		}

		    /* The interlock should be as filled as it's going to get,
		     * add the thing fetched to rConvSum. */
		if (activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum, rScratch,
						       rConvSum);
	    }	/* at the end of this, rConvSum is updated and we've put
	         * out the opcodes needed to either get us to the next
		 * block in this column or the next row. */

	}	/* end of loop generating code to cover columns */
    }	/* end of loop generating code to cover rows in
	 * single left shift case */

	/* we're done, so just return the result.  But first, restore the
	 * global registers saved above.
	 */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg1Offset, rG1);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg2Offset, rG2);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg3Offset, rG3);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg4Offset, rG4);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg5Offset, rG5);

    pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rConvSum, _i0);
    pCodeBuf[addr++] = cRetOp;
    pCodeBuf[addr++] = cRestoreOp;

	/* We're going to do double left shifts of image data.  
	 * But first, since this is the target of the second branch above, we
	 * should modify its branch displacement.
	 */
    pCodeBuf[testShiftCnt2] |= (addr - testShiftCnt2);


	/* **************** 1x1 Double Left Shift Case ****************** */


	/* In this set of code, we'll be doing double left shifts of
	 * image words.
	 * That means we load up each image word and shift it left to
	 * get it into position.  We then load the word to the right
	 * and shift it right, then OR the two together.
	 * We do this for every non-zero kernel row.  First, set
	 * rRightShiftCount to 32 + rShiftCount (which is negative).  This
	 * gives the right shift distance.  Then set rShiftCount to
	 * -rShiftCount to get the proper left shift distance for
	 * the image data read. */
    pCodeBuf[addr++] = GEN_INST(cAddCOp, rShiftCount, 32, rRightShiftCount);
    pCodeBuf[addr++] = GEN_INST(cSubROp, _g0, rShiftCount, rShiftCount);
    sumTouched = cFalse;

	/* Loop through the kernel rows, emitting opcodes for each one. */
    for (r = pKernel->numActiveRows - 1; r >= 0; r--)
    {
	    /* Get the image word for this row */
	if (activeRow[r])
	{
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, rImage, 4, rInputRight);
		/* align both words */
	    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);
	    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInputRight, rRightShiftCount,
					       rInputRight);
		/* merge them */
	    pCodeBuf[addr++] = GEN_INST(cOrROp, rInput, rInputRight, rInput);

		/* load table pointer */
	    if (need2TabPtrs)
		pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rCompTableRow,
						       rCompTableBlock);
	}
	    /* The interlock's all taken care of, but we'll leave this
	     * code here for the sake of convention.
	     * If there are more rows in the kernel table, increment
	     * rImage to point to the next row of corresponding image bits. */
	if (r > 0)
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);

	    /* Loop through the blocks in this row. */
	for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
	{
		/* Put the resulting table entry into convSum.  If this is
		 * the first time we've touched rConvSum, just load it.
		 * otherwise, Add it to the existing value.
		 * BUT FIRST, select hte portion of the inputs we
		 * want for this column block.  If c >= 8, we use
		 * the next 8 bits.  If not, we use the bits convered
		 * by lastMask. */
	    if ((c < 8) && activeRow[r])
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask,
						   rBlockBits);
	    if ((c >= 8) && activeRow[r])
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC,
						   rBlockBits);

	    if (!sumTouched) /* just load rConvSum */
	    {
		if (need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock,
						      rBlockBits,
						      rConvSum);
		if (!need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableRow,
						      rBlockBits,
						      rConvSum);
		if ((c > 8) && activeRow[r])	/* there's more to do
						 * in this row */
		{
		    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 0x8, rInput);
		    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock, 1024,
						       rCompTableBlock);
		}
		    /* if there's no more to do in this row, and there are
		     * more rows, point to them.  NOTE that we do NOT
		     * look at activeRow[r] here because this should
		     * happen whether or not the current row is active.
		     * So why do we put it here?  So it will look like
		     * the sumTouched == cTrue case below where we need
		     * the opcode to fill the interlock between loading
		     * rScratch and adding it to convSum. */
		if ((c <= 8) && (r > 0))
		{
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						       rBytesPerTableRow,
						       rCompTableRow);
		}
		if (activeRow[r])
		    sumTouched = cTrue;
	    }	/* at the end of this, rConvSum is updated and we've put
	         * out the opcodes needed to either get us to the next
		 * block in this column or the next row. */
	    else
		/* Need to add the thing fetched to rConvSum */
	    {
		if (need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock,
						      rBlockBits,
						      rScratch);
		if (!need2TabPtrs && activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableRow,
						      rBlockBits,
						      rScratch);

		if ((c > 8) && activeRow[r])	/* there's more to do
						 * in this row */
		{
		    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 0x8, rInput);
		    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock, 1024,
						       rCompTableBlock);
		}
		    /* if there's no more to do in this row, and there are
		     * more rows, point to them.  This goes here to fill
		     * the interlock between fetching rScratch and adding it
		     * to rConvSum.  Note that we do NOT look at activeRow[r]
		     * here since this action is independent of whether
		     * this row had anything in it. */
		if ((c <= 8) && (r > 0))
		{
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						       rBytesPerTableRow,
						       rCompTableRow);
		}

		    /* The interlock should be as filled as it's going to get,
		     * add the thing fetched to rConvSum. */
		if (activeRow[r])
		    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum, rScratch,
						       rConvSum);
	    }	/* at the end of this, rConvSum is updated and we've put
	         * out the opcodes needed to either get us to the next
		 * block in this column or the next row. */

	}	/* end of loop generating code to cover columns */
    }	/* end of loop generating code to cover rows in
	 * single left shift case */

	/* we're done, so just return the result.  But first, restore the
	 * global registers saved above.
	 */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg1Offset, rG1);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg2Offset, rG2);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg3Offset, rG3);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg4Offset, rG4);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg5Offset, rG5);

    pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rConvSum, _i0);
    pCodeBuf[addr++] = cRetOp;
    pCodeBuf[addr++] = cRestoreOp;

	/* We're done generating code. */
    if (4*addr >= bufSizeInBytes)
    {
#ifdef DEBUG
	fprintf(stderr, " ERROR in gen1x1ConvAsm: storing %d bytes into buffer of size %d\n",
	    4*addr, bufSizeInBytes);
#endif /* DEBUG */
	freer(pIPState, pCodeBuf);
	*codeLength = 0;
	pKernel->conv1Code = NULL;
	return ia_bufferExceeded;
    }

    *codeLength = addr;		/* number of 32-bit words generated */
    pKernel->conv1Code = (Convolve1Proc) pCodeBuf;

    /* Now try flushing the I cache at these locations,  the buffer length
     * is given in bytes
     */
  i_flushICache (pIPState, pCodeBuf, 4 * (*codeLength));

    NORMAL_RETURN;
}




/*
 * So much for the 1x1 case.  How does the 3x3 case differ?  Basically,
 * we process pKernel->numActiveRows + 2 rows in units of 1 row producing
 * 1 sum, 1 row producing 2 sums, pKernel->numActiveRows - 2 rows producing
 * 3 sums, 1 row producing 2 sums and 1 row producing 1 sum.  We also
 * move the initial image pointers up and to the left by 1 pixel before
 * we start.  This points to the upper left corner of the 3x3 set of kernel
 * positions relative to the image.  Here's a map of what the generated
 * code should look like:
 *
 * There's an added twist in Purify fixup that doesn't occur in the
 * 1x1 code.  All the routines generated here
 * end up in a piece of common assembly code that was written by hand and
 * linked in.  Given the chance, Purify will almost certainly mess with
 * any globals that appear in the linked code, so that won't work.
 * HOWEVER, by pure luck, we happen to only use the global registers
 * in the section of code that's generated.  So, we can just restore
 * the globals before we jump to the common code.
 *
 * *********************** ip_convolve3x3 ***************************
 *
 *   Call Overhead Ops
 *   {
 *	new register window (3 ops)
 *	save incoming parms in local vars (5 ops)
 *	save global registers for Purify (5 ops)
 *   }
 *   Initial Parameter Adjustment Ops
 *   {
 *	Dereference pointers in rImage, rCenterXBit (2 ops)
 *	Set rBpl = 4*rBpl (had contained words/line, 1 op)
 *	If pKernel->firstActiveCol == 0,
 *	{
 *	    Subtract pKernel->depth from rCenterXBit (1 op)
 *	    Test to see if the result was negative (1 op)
 *	    Clip rCenterXBit to the range [0..31] in delay slot (1 op)
 *	    If result was negative, subtract 4 from rImage (1 op)
 *	}
 *	Else if pKernel->firstActiveCol > 1,
 *	{
 *	    Add offset due to vacant kernel cols - pKernel->depth
 *		to rCenterXBit (1 op)
 *	    Test to see if rCenterXBit went over 31, (2 ops)
 *	    Clip rCenterXBit to the range [0..31] in delay slot (1 op)
 *	    If rCenterXBit was > 31, add 4 from rImage (1 op)
 *	}
 *	Add in address offset due to vacant kernel rows - 1
 *		(MAX(0, pKernel->firstActiveRow-1) ops)
 *	Load rCompTableRow (1 op)
 *	Load rBytesPerTableRow (1 op)
 *	Set rConvSum0 = rConvSum1 = rConvSum2 = cTripleSumBiasVal (4 ops)
 *	Set rShiftCount to 32 - (rCenterXBit + pKernel->numBitsPerRow) (4 ops)
 *   }
 *   Branch based on value of rShiftCount (if it's > 1, we'll always shift
 *	image data right.  If it's 1 or 0, we do single left shifts.
 *	if it's < 0, we need to do double left shifts.  There will be
 *	a complete section of convolution code for each case.
 *   {
 *	Test rShiftCount (1 op)
 *	If rShiftCount <= 1,
 *	{
 *	    Branch to Case where image should be shifted left (1 op)
 *	    Something useful in delay slot (1 op)
 *	}
 *   }
 *
 *   **** Case where image bits should be shifted right ****
 *   Parameter setup
 *   {
 *	Set rShiftCount = rShiftCount-2	(equals right shift dist, 1 op)
 *   }
 *   Set pCompTableBlock0 = pCompTableRow (1 op)
 *
 *   ** Process first row **
 *   Set rInput = ([rImage] >> rShiftCount) (2 ops)
 *   rImage += rBpl (1 op)
 *   Loop Through column blocks of Kernel (do this
 *   (pKernel->numBitsPerRow+7)/8 times)
 *   {
 *	If there are at least 8 bits remaining in the input row,
 *	    Set rBlockBits = rInput & 0x3FC (1 op)
 *	Else
 *	    Set rBlockBits = rInput & lastMask (1 op)
 *	Set rScratch0 = [rCompTableBlock+rBlockBits] (1 op)
 *	If there are more blocks in this row,
 *	{
 *	    Shift rInput right by 8 (1 op).
 *	    pCompTableBlock0 += 256 * sizeof(Int32) (1 op)
 *	}
 *	Else
 *	{
 *	    rCompTableBlock1 = pCompTableRow (1 op)
 *	    rCompTableRow += rBytesPerTableRow (1 op)
 *	    rCompTableBlock0 = pCompTableRow (1 op)
 *	}
 *	Set rConvSum0 += rScratch0 (1 op)
 *   }
 *
 *   ** Process second row **
 *   Set rInput = ([rImage] >> rShiftCount) (2 ops)
 *   rImage += rBpl (1 op)
 *   Loop Through column blocks of Kernel (do this
 *   (pKernel->numBitsPerRow+7)/8 times)
 *   {
 *	If there are at least 8 bits remaining in the input row,
 *	    Set rBlockBits = rInput & 0x3FC (1 op)
 *	Else
 *	    Set rBlockBits = rInput & lastMask (1 op)
 *	Set rScratch0 = [rCompTableBlock0+rBlockBits] (1 op)
 *	Set rScratch1 = [rCompTableBlock1+rBlockBits] (1 op)
 *	Set rConvSum0 += rScratch0 (1 op)
 *	Set rConvSum1 += rScratch1 (1 op)
 *	If there are more blocks in this row,
 *	{
 *	    Shift rInput right by 8 (1 op)
 *	    pCompTableBlock0 += 256 * sizeof(Int32) (1 op)
 *	    pCompTableBlock1 += 256 * sizeof(Int32) (1 op)
 *	}
 *	Else
 *	{
 *	    rCompTableBlock2 = pCompTableRow - rBytesPerTableRow (1 op)
 *	    rCompTableBlock1 = pCompTableRow (1 op)
 *	    rCompTableRow += rBytesPerTableRow (1 op)
 *	    rCompTableBlock0 = pCompTableRow (1 op)
 *	}
 *   }
 *
 *   ** Process middle rows **
 *   Loop through the next code pKernel->numActiveRows - 2 times
 *   {
 *	Set rInput = ([rImage] >> rShiftCount) (2 ops)
 *	rImage += rBpl (1 op)
 *	Loop Through column blocks of Kernel (do this
 *	(pKernel->numBitsPerRow+7)/8 times)
 *	{
 *	    If there are at least 8 bits remaining in the input row,
 *		Set rBlockBits = rInput & 0x3FC (1 op)
 *	    Else
 *		Set rBlockBits = rInput & lastMask (1 op)
 *	    Set rScratch0 = [rCompTableBlock0+rBlockBits] (1 op)
 *	    Set rScratch1 = [rCompTableBlock1+rBlockBits] (1 op)
 *	    Set rScratch2 = [rCompTableBlock2+rBlockBits] (1 op)
 *	    Set rConvSum0 += rScratch0 (1 op)
 *	    Set rConvSum1 += rScratch1 (1 op)
 *	    Set rConvSum2 += rScratch2 (1 op)
 *	    If there are more blocks in this row,
 *	    {
 *		Shift rInput right by 8 (1 op).
 *		pCompTableBlock0 += 256 * sizeof(Int32) (1 op)
 *		pCompTableBlock1 += 256 * sizeof(Int32) (1 op)
 *		pCompTableBlock2 += 256 * sizeof(Int32) (1 op)
 *	    }
 *	    Else
 *	    If there are more rows in the middle pKernel->numActiveRows - 2
 *		set of rows,
 *	    {
 *		rCompTableBlock2 = pCompTableRow - rBytesPerTableRow (1 op)
 *		rCompTableBlock1 = pCompTableRow (1 op)
 *		rCompTableRow += rBytesPerTableRow (1 op)
 *		rCompTableBlock0 = pCompTableRow (1 op)
 *	    }
 *	    Else this is the last block of the last of the middle rows,
 *	    {
 *	        rCompTableBlock2 = pCompTableRow - rBytesPerTableRow (1 op)
 *	        rCompTableBlock1 = pCompTableRow (1 op)
 *	    }
 *	}
 *   }
 *
 *   ** Process second-to-last row **
 *   At this point, rCompTableRow points to the last row of
 *   table stuff.  pCompTableBlock1 points to the last table row and
 *   pCompTableBlock2 points to the second-to-last table row.
 *
 *   Set rInput = ([rImage] >> rShiftCount) (2 ops)
 *   rImage += rBpl (1 op)
 *   Loop Through column blocks of Kernel (do this
 *   (pKernel->numBitsPerRow+7)/8 times)
 *   {
 *	If there are at least 8 bits remaining in the input row,
 *	    Set rBlockBits = rInput & 0x3FC (1 op)
 *	Else
 *	    Set rBlockBits = rInput & lastMask (1 op)
 *	Set rScratch1 = [rCompTableBlock1+rBlockBits] (1 op)
 *	Set rScratch2 = [rCompTableBlock2+rBlockBits] (1 op)
 *	Set rConvSum1 += rScratch1 (1 op)
 *	Set rConvSum2 += rScratch2 (1 op)
 *	If there are more blocks in this row,
 *	{
 *	    Shift rInput right by 8 (1 op).
 *	    pCompTableBlock1 += 256 * sizeof(Int32) (1 op)
 *	    pCompTableBlock2 += 256 * sizeof(Int32) (1 op)
 *	}
 *	Else
 *	{
 *	    rCompTableBlock2 = pCompTableRow (1 op)
 *	}
 *   }
 *
 *   ** Process last row **
 *
 *   Set rInput = ([rImage] >> rShiftCount) (2 ops)
 *   Loop Through column blocks of Kernel (do this
 *   (pKernel->numBitsPerRow+7)/8 times)
 *   {
 *	If there are at least 8 bits remaining in the input row,
 *	    Set rBlockBits = rInput & 0x3FC (1 op)
 *	Else
 *	    Set rBlockBits = rInput & lastMask (1 op)
 *	Set rScratch2 = [rCompTableBlock2+rBlockBits] (1 op)
 *	If there are more blocks in this row,
 *	{
 *	    Shift rInput right by 8 (1 op).
 *	    pCompTableBlock2 += 256 * sizeof(Int32) (1 op)
 *	}
 *	Else
 *	{
 *	    Set rScratch to address of findMaxOf3x3 (2 ops)
 *	    jump to [rScratch]	(1 op)
 *	}
 *	Set rConvSum2 += rScratch2 (1 op) (after interlock or in delay
 *					   slot of jump!)
 *   }
 *   restore global registers for Purify (5 ops)
 *   jump to common code in ssglyph.s
 *
 *   **** Case where image bits should be single shifted left ****
 *   Compare rShiftCount to 0 (1 op)
 *   if rShiftCount < 0,
 *   {
 *	Branch to double left shift case (1 op)
 *	Set rShiftCount = rShiftCount-2 in delay slot (1 op)
 *   }
 *   Parameter setup
 *   {
 *	Set rShiftCount = -rShiftCount	(equals left shift dist, 1 ops)
 *   }
 *   ... Same as above with right shifts (1 op) replaced with
 *   ... left shifts (1 op) ...
 *
 *
 *
 *   **** Case where image bits should be double shifted left ****
 *   **** rShiftCount holds a negative number ****
 *   Parameter setup
 *   {
 *	Set rRightShiftCount = 32 + rShiftCount (equals right shift dist
 *				for word to the right, 1 op)
 *	Set rShiftCount = -rShiftCount (equals left shift dist, 1 op)
 *   }
 *   ... Same as above with single loads and left shifts (2 ops) replaced with
 *   ... 2 loads and double left shifts (5 ops) ...
 *
 *
 * All of these cases jump to a common piece of assembly code.  In fact,
 * The assembly code is common to all convolutions producing 3x3 results,
 * so it's implemented in ssglyph.s.  They all jump there.  That piece of
 * code does a return to get back to the code that called the code
 * generated here.
 */

/* Sizes of various blocks of code */
#define	cCall3x3OverheadSize	13	/* includes storing global regs */
#define	cParm3x3SetupSize	18
#define	cShiftTestSize		3

#define cShift1SetupSize	2
#define	cShiftLeft2SetupSize	3

#define	cRowLoop3x3SetupSize1	3	/* first row */
#define	cColBlock3x3Size1	5
#define	cRowLoop3x3BottomSize1	1	/* this is really the number of
					 * extra ops executed when we reach
					 * the end of the row loop = 3 - 2 */

#define	cRowLoop3x3SetupSize2	3	/* second row */
#define	cColBlock3x3Size2	8
#define	cRowLoop3x3BottomSize2	1	/* equals 4 - 3 */

#define	cRowLoop3x3SetupSizeM	3	/* middle rows */
#define	cColBlock3x3SizeM	11
#define	cRowLoop3x3BottomSizeM	0	/* equals 4 - 4.  Yeah, this is -2
					 * for the very last row, so we'll
					 * allocate a couple of extra words.
					 * Big Deal. */

#define	cRowLoop3x3SetupSizeL1	3	/* second to last row */
#define	cColBlock3x3SizeL1	8
#define	cRowLoop3x3BottomSizeL1	-2	/* equals 1 - 3 */

#define	cRowLoop3x3SetupSizeL	2	/* last row */
#define	cColBlock3x3SizeL	5
#define	cRowLoop3x3BottomSizeL	6	/* equals 8 - 2, includes restoring
					 * all the global regs for Purify */

#define cRowLoop2xLeftShiftAdd	3	/* number of additional opcodes
					 * needed at the top of code that
					 * processes one input row using
					 * a double shift, as opposed to
					 * a single shift, for alignment. */

extern	UInt32	findMaxOf3x3;		/* this is the entry point of the
					 * assembly code that computes the
					 * maximum of the 9 sums in rConvSum0,
					 * rConvSum1 and rConvSum2.  This 
					 * code is in ssglyph.s.
					 */
/* This is a private routine */
/* PRIVATE */
/*
 *  ip_gen3x3ConvAsm():
 *        This routine generates machine code to process a kernel of a
 *	  particular shape.  The code is specialized for each platform.
 *	  The code generated here produces a 3x3 array of convolution
 *	  results and jumps to a body of code that's common to all such
 *	  routines that selects the best sum and returns it.  This
 *	  common code block also modifies the pointers passed in to point
 *	  to the image location giving the best sum.
 *
 *	pKernel:	Pointer to IAKernel data structure containing
 *			information about the kernel's shape.
 *	allocator:	Ptr to function that is used to allocate memory
 *			for the code.  Function's single argument is
 *			the number of BYTES requested.
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of malloc
 *			provided by WATCOM on the PC!!
 *	freer:		Ptr to function used to free memory in case of
 *			error
 *			NOTE!! This function must have C-style parameter
 *			passing, UNLIKE the standard version of free
 *			provided by WATCOM on the PC!!
 *	codeLength	Pointer to length of code generated in units of
 *			UInt32's.
 *
 *	Return code:
 *		ia_successful		good completion
 *		ia_nomem		couldn't allocate internal buffers
 *		ia_bufferExceeded	code generated didn't fit into buffer
 *					
 */
Int32 CDECL 
ip_gen3x3ConvAsm (
    IPState    *pIPState,
    IAKernel   *pKernel,
    void *(CDECL *allocator) (
		    IPState    *pIPState,
		    UInt32	),
    void (CDECL *freer) (
		    IPState    *pIPState,
		    void       *),
    UInt32     *codeLength)
{
    PROLOGUE;
    Int32	 r, c;			/* general iterative */
    UInt32	*pCodeBuf;		/* pointer to the code block */
    Int32	 bufSizeInBytes;	/* size of code buffer */
    Int32	 addr;			/* offset to currently generated
					 * inst word */
    Int32	 testShiftCnt1, testShiftCnt2;	/* offsets of test ops */
    UInt32	 lastMask;		/* mask used to select the last
					 * set of bits in each input row. */
    UInt32	 numBlocks;		/* equals number of 8-bit blocks
					 * in a row of input bits. */

	/* set constants.  lastMask selects the relevant bits in the
	 * last (leftmost) group of input bits in each row.  numBlocks
	 * equals the number of blocks in an input row (including the last
	 * partial one covered by lastMask).  If pKernel->numBitsPerRow is
	 * a multiple of 8, lastMask is not used.
	 */
    lastMask = (0xFF >> (8 - (pKernel->numBitsPerRow % 8)) << 2);
    numBlocks = (pKernel->numBitsPerRow + 7) / 8;

	/* Figure out how much space we'll need.  At the end of each
	 * case, we jump to findMaxOf3x3, so there's no return here.
	 */
    bufSizeInBytes = 4*(cCall3x3OverheadSize +	/* call overhead */
			cParm3x3SetupSize + pKernel->firstActiveRow +
				/* For each case (shift right, single left
				 * shift, double left shift) we process the
				 * first row, the second row, the middle
				 * pKernel->numActiveRows - 2 rows, the second
				 * to last row and the last row.  We do
				 * the double left shift case separately
				 * because the number of opcodes required
				 * to do the double left shift is larger
				 * that that required to do a single right
				 * or left shift. */
			2 * (cShiftTestSize +
			     cShift1SetupSize +
			     cRowLoop3x3SetupSize1 +
			     numBlocks * cColBlock3x3Size1 +
			     cRowLoop3x3BottomSize1 +
			     cRowLoop3x3SetupSize2 +
			     numBlocks * cColBlock3x3Size2 +
			     cRowLoop3x3BottomSize2 +
			     (pKernel->numActiveRows - 2) *
				 (cRowLoop3x3SetupSizeM +
				  numBlocks * cColBlock3x3SizeM +
				  cRowLoop3x3BottomSizeM) +
			     cRowLoop3x3SetupSizeL1 +
			     numBlocks * cColBlock3x3SizeL1 +
			     cRowLoop3x3BottomSizeL1 +
			     cRowLoop3x3SetupSizeL +
			     numBlocks * cColBlock3x3SizeL +
			     cRowLoop3x3BottomSizeL) +

			cShiftLeft2SetupSize +
			1 * (cRowLoop3x3SetupSize1 + cRowLoop2xLeftShiftAdd +
			     numBlocks * cColBlock3x3Size1 +
			     cRowLoop3x3BottomSize1 +
			     cRowLoop3x3SetupSize2 + cRowLoop2xLeftShiftAdd +
			     numBlocks * cColBlock3x3Size2 +
			     cRowLoop3x3BottomSize2 +
			     (pKernel->numActiveRows - 2) *
				 (cRowLoop3x3SetupSizeM + cRowLoop2xLeftShiftAdd +
				  numBlocks * cColBlock3x3SizeM +
				  cRowLoop3x3BottomSizeM) +
			     cRowLoop3x3SetupSizeL1 + cRowLoop2xLeftShiftAdd +
			     numBlocks * cColBlock3x3SizeL1 +
			     cRowLoop3x3BottomSizeL1 +
			     cRowLoop3x3SetupSizeL + cRowLoop2xLeftShiftAdd +
			     numBlocks * cColBlock3x3SizeL +
			     cRowLoop3x3BottomSizeL));

	/* get the memory to be used for the code */
    pCodeBuf = (UInt32 *) ((allocator) (pIPState, bufSizeInBytes));

	/* quit if we didn't get any memory */
    if (pCodeBuf == NULL)
	return ia_nomem;

	/* Start filing the buffer with initial boilerplate */
    addr = 0;		/* initial offset into buffer */
    pCodeBuf[addr++] = GEN_SETHI_INST(LF26, _g1);
    pCodeBuf[addr++] = GEN_INST(cAddCOp, _g1, _lo(LF26), _g1);
    pCodeBuf[addr++] = GEN_INST(cSaveROp, _sp, _g1, _sp);
  
	/* Store the incoming parameters.  More C Boilerplate */
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rImage, _fp, cImageOffset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rCenterXBit, _fp, cCenterXBitOffset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rBpl, _fp, cBplOffset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rKernel, _fp, cKernelOffset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rNumOutBit, _fp, cNumOutBitOffset);

	/* Save the global registers for the sake of Purify */
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG1, _fp, cSavedg1Offset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG2, _fp, cSavedg2Offset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG3, _fp, cSavedg3Offset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG4, _fp, cSavedg4Offset);
    pCodeBuf[addr++] = GEN_ST_INST(cStCOp, rG5, _fp, cSavedg5Offset);

	/* Set up the incoming parameters.  rCenterXBit and rImage point to
	 * their args, so dereference them. */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rCenterXBit, 0, rCenterXBit);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rImage, 0, rImage);
	/* Convert words/line to bytes/line */
    pCodeBuf[addr++] = GEN_INST(cSllCOp, rBpl, 2, rBpl);

	/* Add in offset due to vacant kernel columns minus pKernel->depth.
	 * The -pKernel->depth term accounts for the fact that we want
	 * to move up and to the left by 1 pixel to get to the upper
	 * left corner of the 3x3 set of locations where we'll put the kernel.
	 */
    if (pKernel->firstActiveCol == 0)
    {
	pCodeBuf[addr++] = GEN_INST(cSubccCOp, rCenterXBit,
					     pKernel->depth, rCenterXBit);
	pCodeBuf[addr++] = GEN_BR_INST(cBPOSOp, 4*3);	/* skip 2 opcodes */
	    /* It's always safe to clip rCenterXBit back to the range [0..1F]
	     * in the branch's delay slot */
	pCodeBuf[addr++] = GEN_INST(cAndCOp, rCenterXBit, 0x1F, rCenterXBit);
	pCodeBuf[addr++] = GEN_INST(cSubCOp, rImage, 4, rImage);
    }
    else
    if (pKernel->firstActiveCol > 1)
    {
	pCodeBuf[addr++] = GEN_INST(cAddCOp, rCenterXBit,
					   pKernel->firstActiveColBit -
						pKernel->depth, rCenterXBit);
	    /* Test to see if rCenterXBit is now over 31 or below 0. */
	pCodeBuf[addr++] = GEN_INST(cAndccCOp, rCenterXBit, 0x20, _g0);
	pCodeBuf[addr++] = GEN_BR_INST(cBZOp, 4*3);	/* skip 2 opcodes */
	    /* It's always safe to clip rCenterXBit back to the range [0..1F]
	     * in the branch's delay slot */
	pCodeBuf[addr++] = GEN_INST(cAndCOp, rCenterXBit, 0x1F, rCenterXBit);
	pCodeBuf[addr++] = GEN_INST(cAddCOp, rImage, 4, rImage);
    }

	/* this is the target of the jumps above.  Account for the
	 * two vertical offsets.  A positive value in pKernel->firstActiveRow
	 * implies we should move down in the image (because the kernel
	 * is actually smaller than the caller thinks it is).  We should
	 * also move up 1 row to get to the upper left corner of the 3x3
	 * set of kernel positions.
	 */
    if (pKernel->firstActiveRow == 0)
	pCodeBuf[addr++] = GEN_INST(cSubROp, rImage, rBpl, rImage);
    else
	for (r = pKernel->firstActiveRow-1; r > 0; r--)
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);

	/* load rCompTableRow with pointer to tables of compiled weights */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rKernel, cIAKCompWeights, rCompTableRow);
	/* we'll load rBytesPerTableRow in the delay slot of the branch below */

	/* Load rConvSum0, rConvSum1 and rConvSum2 with cTripleSumBiasVal */
    pCodeBuf[addr++] = GEN_SETHI_INST(cTripleSumBiasVal, rConvSum0);
    pCodeBuf[addr++] = GEN_INST(cOrCOp, rConvSum0, _lo(cTripleSumBiasVal),
				      rConvSum0);
    pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rConvSum0, rConvSum1);
    pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rConvSum0, rConvSum2);

	/* load rScratch with pKernel->numBitsPerRow */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rKernel, cIAKNumBitsPerRow, rScratch);
	/* load rShiftCount with 32 in interlock slot */
    pCodeBuf[addr++] = GEN_INST(cOrCOp, _g0, 32, rShiftCount);
	/* put 32 - pKernel->numBitsPerRow into rShiftCount */
    pCodeBuf[addr++] = GEN_INST(cSubROp, rShiftCount, rScratch, rShiftCount);
	/* put 32 - (pKernel->numBitsPerRow + rCenterXBit) into rShiftCount */
    pCodeBuf[addr++] = GEN_INST(cSubROp, rShiftCount, rCenterXBit, rShiftCount);

	/* if rShiftCount is less than 1 (using signed comparisons)
	 * we'll do left shifts.  Those are the target of the branch. */
    pCodeBuf[addr++] = GEN_INST(cSubccCOp, rShiftCount, 0x1, _g0);
	/* This is the location of the first branch on the value of
	 * rShiftCount.  Remember this offset since we'll need to set
	 * the branch distance later */
    testShiftCnt1 = addr;
    pCodeBuf[addr++] = GEN_BR_INST(cBLEOp, 0);	/* fill in offset later */
	/* load rBytesPerTableRow in the delay slot */
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rKernel, cIAKBytesPerTabRow, 
				rBytesPerTableRow);


	/* **************** 3x3 Right Shift Case *********************** */


	/* In this set of code, we'll be doing right shifts of image words.
	 * That means we load up each image word and shift it right to
	 * get it into position.
	 * We do this for every non-zero kernel row.  First, set rShiftCount to
	 * rShiftCount-2 to get the proper right shift distance for
	 * the image data read. */
    pCodeBuf[addr++] = GEN_INST(cSubCOp, rShiftCount, 0x2, rShiftCount);
    pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rCompTableRow, rCompTableBlock0);

	/* ************* PROCESS FIRST ROW, right shift ***************** */

	/* Read the input word for row 0 and align it.  Point to next row
	 * in the interlock slot. */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInput, rShiftCount, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entry selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					  rScratch0);

	    /* We fill the interlock with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock0.  If this is the last block
	     * in the row, we advance rCompTableRow and load rCompTableBlock0
	     * and rCompTableBlock1.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
					       256 * sizeof(UInt32),
					       rCompTableBlock0);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableRow);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock0);
	}

	    /* Finally, add the stuff we fetched from the lookup table to
	     * rConvSum0. */
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);

    }	/* end of loop through blocks of first row in right shift case */


	/* ************* PROCESS SECOND ROW, right shift ***************** */

	/* Read the input word for second row and align it.  Point to next row
	 * in the interlock slot. */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInput, rShiftCount, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since the later load fill the interlock steps, do the
	     * adds now.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					  rScratch0);
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					  rScratch1);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);

	    /* We finish the block with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock0 and rCompTableBlock1.
	     * If this is the last block in the row, we advance rCompTableRow
	     * and load rCompTableBlock0, rCompTableBlock1 and rCompTableBlock2.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
					       256 * sizeof(UInt32),
					       rCompTableBlock0);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
					       256 * sizeof(UInt32),
					       rCompTableBlock1);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableBlock2);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableRow);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock0);
	}

    }	/* end of loop through blocks of second row in right shift case */


	/* ************* PROCESS MIDDLE ROWS, right shift ***************** */

    for (r = pKernel->numActiveRows - 2; r > 0; r--)
    {
	    /* Read the input word for this row and align it.  Point to
	     * next row in the interlock slot. */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
	pCodeBuf[addr++] = GEN_INST(cSrlROp, rInput, rShiftCount, rInput);
    
	    /* Loop through the blocks in this row. */
	for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
	{
		/* if there are at least 8 bits in this block, set rBlockBits
		 * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
		 * lastMask.
		 */
	    if (c < 8)
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask,
						   rBlockBits);
	    else
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);
    
		/* load the table entries selected by these input bits.
		 * This is the main step of our style of convolution.
		 * It takes the place of all the multiplies normally
		 * involved in a convolution.
		 * Since the later load fill the interlock steps, do the
		 * adds now.
		 */
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					      rScratch0);
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					      rScratch1);
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					      rScratch2);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);

		/* We finish the block with one of two actions.
		 * If there are more blocks to process in this row, we
		 * shift rInput right by 8 and advance rCompTableBlock0,
		 * rCompTableBlock1 and rCompTableBlock2.
		 * If this is the last block in a row that isn't the
		 * last row, we advance rCompTableRow, rCompTableBlock0,
		 * rCompTableBlock1 and rCompTableBlock2.  If this is the
		 * last block of the last row, we only advance
		 * rCompTableBlock1 and rCompTableBlock2.  We leave
		 * rCompTableBlockRow pointing to the last table row when
		 * we leave.
		 */
	    if (c > 8)
	    {
		pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
						   256 * sizeof(UInt32),
						   rCompTableBlock0);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
						   256 * sizeof(UInt32),
						   rCompTableBlock1);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
						   256 * sizeof(UInt32),
						   rCompTableBlock2);
	    }
	    else
	    if (r > 1) /* not the last row */
	    {
		pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableBlock2);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock1);
		pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableRow);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock0);
	    }
	    else	/* this is the last of the middle rows */
	    {
		pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableBlock2);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock1);
	    }
    
	}	/* end of loop through blocks of each middle row in
		 * right shift case */
    }	/* end of loop through middle rows in right shift case */


	/* ************* PROCESS SECOND-TO-LAST ROW, right shift ********** */

	/* Read the input word for this row and align it.  Point to next row
	 * in the interlock slot. */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInput, rShiftCount, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since the later load fill the interlock steps, do the
	     * adds now.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					  rScratch1);
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					  rScratch2);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);

	    /* We finish the block with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock1 and rCompTableBlock2.
	     * If this is the last block in the row, we advance
	     * rCompTableBlock2.  Remember that rCompTableRow still points
	     * to the last row of the tables.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
					       256 * sizeof(UInt32),
					       rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
					       256 * sizeof(UInt32),
					       rCompTableBlock2);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock2);
	}

    }	/* end of loop through blocks of second-to-last row
	 * in right shift case */


	/* ************* PROCESS LAST ROW, right shift ***************** */

	/* Read the input word for this row and align it.  Don't bother
	 * filling interlock slot. */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInput, rShiftCount, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since there's only 1 load, try to fill the interlock with
	     * code and do the add below.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					  rScratch2);

	    /* If there are more blocks to process in this row, we shift
	     * rInput right by 8 and advance rCompTableBlock2.
	     * If this is the last block in the row, we jump to the code
	     * that will find the largest legal sum of the 9 sums in
	     * rConvSum0, rConvSum1 and rConvSum2.  This code is in
	     * ssglyph.s.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
					       256 * sizeof(UInt32),
					       rCompTableBlock2);
	}
	else
	{
		/* Restore the global registers for the sake of Purify */
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg1Offset, rG1);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg2Offset, rG2);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg3Offset, rG3);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg4Offset, rG4);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg5Offset, rG5);

	    pCodeBuf[addr++] = GEN_SETHI_INST((UInt32)&findMaxOf3x3, rScratch);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rScratch,
					     _lo((UInt32)&findMaxOf3x3),
					      rScratch);
	    pCodeBuf[addr++] = GEN_INST(cJmplCOp, rScratch, 0, _g0);
	}

	    /* do the add after the interlock stuff or in the jump's
	     * delay slot */
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);

    }	/* end of loop through blocks of last row
	 * in right shift case */


	/* This is where we test to see if rShiftCount is negative.  If so,
	 * we jump down to do double left shifts.
	 * But first, since this is the target of the branch above, we
	 * should modify its branch displacement.
	 */
    pCodeBuf[testShiftCnt1] |= (addr - testShiftCnt1);

	/* if rShiftCount is less than 0 (using signed comparisons)
	 * we'll do double left shifts.  Those are the target of the branch. */
    pCodeBuf[addr++] = GEN_INST(cSubccCOp, rShiftCount, 0x0, _g0);
	/* This is the location of the second branch on the value of
	 * rShiftCount.  Remember this offset since we'll need to set
	 * the branch distance later */
    testShiftCnt2 = addr;
    pCodeBuf[addr++] = GEN_BR_INST(cBLOp, 0);	/* fill in offset later */
	/* Compute rShiftCount-2 in the delay slot.  In the single shift
	 * case, we'll just negate this to get the shift count.  In
	 * the double shift case, we add this to 32 to get the right
	 * shift count, then negate it to get the left shift count.
	 */
    pCodeBuf[addr++] = GEN_INST(cSubCOp, rShiftCount, 0x2, rShiftCount);


	/* **************** 3x3 Single Left Case ******************* */


	/* In this set of code, we'll be doing single left shifts of
	 * image words.
	 * That means we load up each image word and shift it left to
	 * get it into position.
	 * We do this for every non-zero kernel row.  First, negate
	 * rShiftCount to get the proper left shift distance for
	 * the image data read. */
    pCodeBuf[addr++] = GEN_INST(cSubROp, _g0, rShiftCount, rShiftCount);
    pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rCompTableRow, rCompTableBlock0);

	/* ********** PROCESS FIRST ROW, single left shift ************** */

	/* Read the input word for row 0 and align it.  Point to next row
	 * in the interlock slot. */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entry selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					  rScratch0);

	    /* We fill the interlock with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock0.  If this is the last block
	     * in the row, we advance rCompTableRow and load rCompTableBlock0
	     * and rCompTableBlock1.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
					       256 * sizeof(UInt32),
					       rCompTableBlock0);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableRow);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock0);
	}

	    /* Finally, add the stuff we fetched from the lookup table to
	     * rConvSum0. */
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);

    }	/* end of loop through blocks of first row in single left shift case */


	/* ********** PROCESS SECOND ROW, single left shift ************** */

	/* Read the input word for second row and align it.  Point to next row
	 * in the interlock slot. */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since the later load fill the interlock steps, do the
	     * adds now.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					  rScratch0);
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					  rScratch1);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);

	    /* We finish the block with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock0 and rCompTableBlock1.
	     * If this is the last block in the row, we advance rCompTableRow
	     * and load rCompTableBlock0, rCompTableBlock1 and rCompTableBlock2.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
					       256 * sizeof(UInt32),
					       rCompTableBlock0);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
					       256 * sizeof(UInt32),
					       rCompTableBlock1);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableBlock2);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableRow);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock0);
	}

    }	/* end of loop through blocks of second row in single left shift case */


	/* ********** PROCESS MIDDLE ROWS, single left shift ************** */

    for (r = pKernel->numActiveRows - 2; r > 0; r--)
    {
	    /* Read the input word for this row and align it.  Point to
	     * next row in the interlock slot. */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
	pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);
    
	    /* Loop through the blocks in this row. */
	for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
	{
		/* if there are at least 8 bits in this block, set rBlockBits
		 * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
		 * lastMask.
		 */
	    if (c < 8)
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	    else
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);
    
		/* load the table entries selected by these input bits.
		 * This is the main step of our style of convolution.
		 * It takes the place of all the multiplies normally
		 * involved in a convolution.
		 * Since the later load fill the interlock steps, do the
		 * adds now.
		 */
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					      rScratch0);
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					      rScratch1);
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					      rScratch2);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);
    
		/* We finish the block with one of two actions.
		 * If there are more blocks to process in this row, we
		 * shift rInput right by 8 and advance rCompTableBlock0,
		 * rCompTableBlock1 and rCompTableBlock2.
		 * If this is the last block in a row that isn't the
		 * last row, we advance rCompTableRow, rCompTableBlock0,
		 * rCompTableBlock1 and rCompTableBlock2.  If this is the
		 * last block of the last row, we only advance
		 * rCompTableBlock1 and rCompTableBlock2.  We leave
		 * rCompTableBlockRow pointing to the last table row when
		 * we leave.
		 */
	    if (c > 8)
	    {
		pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
						   256 * sizeof(UInt32),
						   rCompTableBlock0);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
						   256 * sizeof(UInt32),
						   rCompTableBlock1);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
						   256 * sizeof(UInt32),
						   rCompTableBlock2);
	    }
	    else
	    if (r > 1) /* not the last row */
	    {
		pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableBlock2);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock1);
		pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableRow);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock0);
	    }
	    else	/* this is the last of the middle rows */
	    {
		pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableBlock2);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock1);
	    }
    
	}	/* end of loop through blocks of each middle row in
		 * single left shift case */
    }	/* end of loop through middle rows in single left shift case */


	/* ********** PROCESS SECOND-TO-LAST ROW, single left shift ******* */

	/* Read the input word for this row and align it.  Point to next row
	 * in the interlock slot. */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);
    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since the later load fill the interlock steps, do the
	     * adds now.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					  rScratch1);
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					  rScratch2);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);

	    /* We finish the block with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock1 and rCompTableBlock2.
	     * If this is the last block in the row, we advance
	     * rCompTableBlock2.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
					       256 * sizeof(UInt32),
					       rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
					       256 * sizeof(UInt32),
					       rCompTableBlock2);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock2);
	}

    }	/* end of loop through blocks of second-to-last row
	 * in single left shift case */


	/* ********** PROCESS LAST ROW, single left shift ************** */

	/* Read the input word for this row and align it.  Don't bother
	 * filling interlock slot. */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since there's only 1 load, try to fill the interlock with
	     * code and do the add below.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					  rScratch2);

	    /* If there are more blocks to process in this row, we shift
	     * rInput right by 8 and advance rCompTableBlock2.
	     * If this is the last block in the row, we jump to the code
	     * that will find the largest legal sum of the 9 sums in
	     * rConvSum0, rConvSum1 and rConvSum2.  This code is in
	     * ssglyph.s.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
					       256 * sizeof(UInt32),
					       rCompTableBlock2);
	}
	else
	{
		/* Restore the global registers for the sake of Purify */
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg1Offset, rG1);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg2Offset, rG2);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg3Offset, rG3);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg4Offset, rG4);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg5Offset, rG5);

	    pCodeBuf[addr++] = GEN_SETHI_INST((UInt32)&findMaxOf3x3, rScratch);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rScratch,
					     _lo((UInt32)&findMaxOf3x3),
					      rScratch);
	    pCodeBuf[addr++] = GEN_INST(cJmplCOp, rScratch, 0, _g0);
	}

	    /* do the add after the interlock stuff or in the jump's
	     * delay slot */
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);

    }	/* end of loop through blocks of last row
	 * in single left shift case */



	/* We're going to do double left shifts of image data.  
	 * But first, since this is the target of the second branch above, we
	 * should modify its branch displacement.
	 */
    pCodeBuf[testShiftCnt2] |= (addr - testShiftCnt2);



	/* **************** 3x3 Double Left Case ******************* */


	/* In this set of code, we'll be doing double left shifts of
	 * image words.
	 * That means we load up each image word and shift it left to
	 * get it into position.  We then load the word to the right
	 * and shift it right, then OR the two together.
	 * We do this for every non-zero kernel row.  First, set
	 * rRightShiftCount to 32 + rShiftCount (which is negative).  This
	 * gives the right shift distance.  Then set rShiftCount to
	 * -rShiftCount to get the proper left shift distance for
	 * the image data read. */
    pCodeBuf[addr++] = GEN_INST(cAddCOp, rShiftCount, 32, rRightShiftCount);
    pCodeBuf[addr++] = GEN_INST(cSubROp, _g0, rShiftCount, rShiftCount);
    pCodeBuf[addr++] = GEN_INST(cOrROp, _g0, rCompTableRow, rCompTableBlock0);

	/* ********** PROCESS FIRST ROW, double left shift ************** */

	/* Read the input words */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rImage, 4, rInputRight);
	/* align both words */
    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);
    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInputRight, rRightShiftCount,
				       rInputRight);
	/* merge them and point to next line down */
    pCodeBuf[addr++] = GEN_INST(cOrROp, rInput, rInputRight, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entry selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					  rScratch0);

	    /* We fill the interlock with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock0.  If this is the last block
	     * in the row, we advance rCompTableRow and load rCompTableBlock0
	     * and rCompTableBlock1.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
					       256 * sizeof(UInt32),
					       rCompTableBlock0);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableRow);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock0);
	}

	    /* Finally, add the stuff we fetched from the lookup table to
	     * rConvSum0. */
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);

    }	/* end of loop through blocks of first row in double left shift case */


	/* ********** PROCESS SECOND ROW, double left shift ************** */

	/* Read the input words */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rImage, 4, rInputRight);
	/* align both words */
    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);
    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInputRight, rRightShiftCount,
				       rInputRight);
	/* merge them and point to next line down */
    pCodeBuf[addr++] = GEN_INST(cOrROp, rInput, rInputRight, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since the later load fill the interlock steps, do the
	     * adds now.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					  rScratch0);
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					  rScratch1);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);

	    /* We finish the block with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock0 and rCompTableBlock1.
	     * If this is the last block in the row, we advance rCompTableRow
	     * and load rCompTableBlock0, rCompTableBlock1 and rCompTableBlock2.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
					       256 * sizeof(UInt32),
					       rCompTableBlock0);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
					       256 * sizeof(UInt32),
					       rCompTableBlock1);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableBlock2);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow, rBytesPerTableRow,
					       rCompTableRow);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock0);
	}

    }	/* end of loop through blocks of second row in double left shift case */


	/* ********** PROCESS MIDDLE ROWS, double left shift ************** */

    for (r = pKernel->numActiveRows - 2; r > 0; r--)
    {
	    /* Read the input words */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
	pCodeBuf[addr++] = GEN_INST(cLdCOp, rImage, 4, rInputRight);
	    /* align both words */
	pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);
	pCodeBuf[addr++] = GEN_INST(cSrlROp, rInputRight, rRightShiftCount,
					   rInputRight);
	    /* merge them and point to next line down */
	pCodeBuf[addr++] = GEN_INST(cOrROp, rInput, rInputRight, rInput);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);

	    /* Loop through the blocks in this row. */
	for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
	{
		/* if there are at least 8 bits in this block, set rBlockBits
		 * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
		 * lastMask.
		 */
	    if (c < 8)
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask,
						   rBlockBits);
	    else
		pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);
    
		/* load the table entries selected by these input bits.
		 * This is the main step of our style of convolution.
		 * It takes the place of all the multiplies normally
		 * involved in a convolution.
		 * Since the later load fill the interlock steps, do the
		 * adds now.
		 */
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock0, rBlockBits, 
					      rScratch0);
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					      rScratch1);
	    pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					      rScratch2);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum0, rScratch0, rConvSum0);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);
	    pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);
    
		/* We finish the block with one of two actions.
		 * If there are more blocks to process in this row, we
		 * shift rInput right by 8 and advance rCompTableBlock0,
		 * rCompTableBlock1 and rCompTableBlock2.
		 * If this is the last block in a row that isn't the
		 * last row, we advance rCompTableRow, rCompTableBlock0,
		 * rCompTableBlock1 and rCompTableBlock2.  If this is the
		 * last block of the last row, we only advance
		 * rCompTableBlock1 and rCompTableBlock2.  We leave
		 * rCompTableBlockRow pointing to the last table row when
		 * we leave.
		 */
	    if (c > 8)
	    {
		pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock0,
						   256 * sizeof(UInt32),
						   rCompTableBlock0);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
						   256 * sizeof(UInt32),
						   rCompTableBlock1);
		pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
						   256 * sizeof(UInt32),
						   rCompTableBlock2);
	    }
	    else
	    if (r > 1) /* not the last row */
	    {
		pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableBlock2);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock1);
		pCodeBuf[addr++] = GEN_INST(cAddROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableRow);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock0);
	    }
	    else	/* this is the last of the middle rows */
	    {
		pCodeBuf[addr++] = GEN_INST(cSubROp, rCompTableRow,
						   rBytesPerTableRow,
						   rCompTableBlock2);
		pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
						  rCompTableBlock1);
	    }
    
	}	/* end of loop through blocks of each middle row in
		 * single left shift case */
    }	/* end of loop through middle rows in double left shift case */


	/* ********** PROCESS SECOND-TO-LAST ROW, double left shift ******* */

	/* Read the input words */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rImage, 4, rInputRight);
	/* align both words */
    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);
    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInputRight, rRightShiftCount,
				       rInputRight);
	/* merge them and point to next line down */
    pCodeBuf[addr++] = GEN_INST(cOrROp, rInput, rInputRight, rInput);
    pCodeBuf[addr++] = GEN_INST(cAddROp, rImage, rBpl, rImage);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since the later load fill the interlock steps, do the
	     * adds now.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock1, rBlockBits, 
					  rScratch1);
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					  rScratch2);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum1, rScratch1, rConvSum1);
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);

	    /* We finish the block with one of two actions.  If there are
	     * more blocks to process in this row, we shift rInput right by
	     * 8 and advance rCompTableBlock1 and rCompTableBlock2.
	     * If this is the last block in the row, we advance
	     * rCompTableBlock2.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock1,
					       256 * sizeof(UInt32),
					       rCompTableBlock1);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
					       256 * sizeof(UInt32),
					       rCompTableBlock2);
	}
	else
	{
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rCompTableRow, 0,
					      rCompTableBlock2);
	}

    }	/* end of loop through blocks of second-to-last row
	 * in double left shift case */


	/* ********** PROCESS LAST ROW, double left shift ************** */

	/* Read the input words */
    pCodeBuf[addr++] = GEN_INST(cLdROp, rImage, _g0, rInput);
    pCodeBuf[addr++] = GEN_INST(cLdCOp, rImage, 4, rInputRight);
	/* align both words */
    pCodeBuf[addr++] = GEN_INST(cSllROp, rInput, rShiftCount, rInput);
    pCodeBuf[addr++] = GEN_INST(cSrlROp, rInputRight, rRightShiftCount,
				       rInputRight);
	/* merge them */
    pCodeBuf[addr++] = GEN_INST(cOrROp, rInput, rInputRight, rInput);

	/* Loop through the blocks in this row. */
    for (c = pKernel->numBitsPerRow; c > 0; c -= 8)
    {
	    /* if there are at least 8 bits in this block, set rBlockBits
	     * to rInput & 0x3FC.  Otherwise, set rBlockBits to rInput &
	     * lastMask.
	     */
	if (c < 8)
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, lastMask, rBlockBits);
	else
	    pCodeBuf[addr++] = GEN_INST(cAndCOp, rInput, 0x3FC, rBlockBits);

	    /* load the table entries selected by these input bits.  This is
	     * the main step of our style of convolution.  It takes the place
	     * of all the multiplies normally involved in a convolution.
	     * Since there's only 1 load, try to fill the interlock with
	     * code and do the add below.
	     */
	pCodeBuf[addr++] = GEN_INST(cLdROp, rCompTableBlock2, rBlockBits, 
					  rScratch2);

	    /* If there are more blocks to process in this row, we shift
	     * rInput right by 8 and advance rCompTableBlock2.
	     * If this is the last block in the row, we jump to the code
	     * that will find the largest legal sum of the 9 sums in
	     * rConvSum0, rConvSum1 and rConvSum2.  This code is in
	     * ssglyph.s.
	     */
	if (c > 8)
	{
	    pCodeBuf[addr++] = GEN_INST(cSrlCOp, rInput, 8, rInput);
	    pCodeBuf[addr++] = GEN_INST(cAddCOp, rCompTableBlock2,
					       256 * sizeof(UInt32),
					       rCompTableBlock2);
	}
	else
	{
		/* Restore the global registers for the sake of Purify */
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg1Offset, rG1);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg2Offset, rG2);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg3Offset, rG3);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg4Offset, rG4);
	    pCodeBuf[addr++] = GEN_INST(cLdCOp, _fp, cSavedg5Offset, rG5);

	    pCodeBuf[addr++] = GEN_SETHI_INST((UInt32)&findMaxOf3x3, rScratch);
	    pCodeBuf[addr++] = GEN_INST(cOrCOp, rScratch,
					     _lo((UInt32)&findMaxOf3x3),
					      rScratch);
	    pCodeBuf[addr++] = GEN_INST(cJmplCOp, rScratch, 0, _g0);
	}

	    /* do the add after the interlock stuff or in the jump's
	     * delay slot */
	pCodeBuf[addr++] = GEN_INST(cAddROp, rConvSum2, rScratch2, rConvSum2);

    }	/* end of loop through blocks of last row
	 * in double left shift case */



	/* We're done generating code. */
    if (4*addr >= bufSizeInBytes)
    {
#ifdef DEBUG
	fprintf(stderr, " ERROR in gen1x1ConvAsm: storing %d bytes into buffer of size %d\n",
	    4*addr, bufSizeInBytes);
#endif /* DEBUG */
	freer(pIPState, pCodeBuf);
	*codeLength = 0;
	pKernel->conv3Code = NULL;
	return ia_bufferExceeded;
    }

    *codeLength = addr;		/* number of 32-bit words generated */
    pKernel->conv3Code = (Convolve3Proc) pCodeBuf;

    /* Now try flushing the I cache at these locations,  the buffer length
     * is given in bytes
     */
  i_flushICache (pIPState, pCodeBuf, 4 * (*codeLength));

    NORMAL_RETURN;
}

/*

$Log: ssgengly.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.19  1998/01/13 23:54:20  ddavies
ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.18  1997/07/11 01:51:39  ddavies
ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.17  1995/12/13 02:52:57  ejaquith
ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.16  1995/10/18  04:05:09  ddavies
 * ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.15  1994/09/20  02:56:29  ddavies
 * ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v
 *
 * Call i_flushICache instead of ip_flushICache.  Include shrip.pub instead
 * of shrip.prv for this purpose.
 *
 * Revision 1.14  1994/09/07  20:34:37  ddavies
 * ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v
 *
 * Add a call to the instruction cache flushing routine after generating
 * assembly code for 1x1 and 3x3 convolutions.
 *
 * Revision 1.13  1994/02/04  20:20:43  dferrell
 * ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v
 *
 * Changed memory.prv to memory.pub
 *
 * Revision 1.12  1994/01/07  22:44:28  jcraig
 * jcraig on Fri Jan 7 14:44:01 PST 1994
 *
 * ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v
 *
 * Dan's fix to ip_gen3x3ConvAsm to fix a bus error: when generating
 * the instruction to more the image down some rows because the
 * the first active row is not 0, the bpl is in a register not a constant.
 *
 * Revision 1.11  1993/12/02  19:29:59  ddavies
 * ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.10  1993/11/05  02:24:36  ddavies
 * ssgengly.c -> /product/ipcore/ipshared/src/RCS/ssgengly.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


