/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#include "kernel.prv"	/* to get ia_kernel offsets and information about
			 * the layout of sums in convolution registers */

! File: ssglyph.s
! 
! This file contains assembly code to find the largest sum out of a group
! of 9 sums contained in 3 registers.  This code is not an entire
! subroutine - it is executed as the tail end of all 3x3 convolution
! operations.
!
! We generate custom machine code to do the actual convolution operations.
! This code varies by the size and shape of the convolution kernel.
! However, at the end of every 3x3 convolution operation, we still need
! to find the largest sum that's less than pKernel->upperThreshold.
! This code does not change with kernel, so will be written here.
!
! We enter this code via a jump.  The registers are already set up as
! shown below.  We do a return to leave from here.
!
! There are three registers containing sums, rConvSum0, rConvSum1 and
! rConvSum2,  The final result should end up in rConvSum.  Each register
! contains 3 sums laid out as
!		
!	Bit number:	31	    22 | 21	     11 | 10		0
!	Value:		Sum 0 (10 bits)	 Sum 1 (11 bits)  Sum 2 (11 bits)
!
! The sums have already had a bias value (cSumBias = 512) added into
! them before the convolution.  This guarantees they are all positive
! 10 bit numbers.
!
! To avoid comparing each of the sums to pKernel->upperThreshold individually,
! we add cSumBias - pKernel->upperThreshold to each sum.  This makes the
! sums that were bigger than the upper threshold wrap around to small
! numbers when we look at only the bottom 10 bits.  We can then do
! regular unsigned compares.  Before returning the final sum, we subtract
! 2*cSumBias and add pKernel-upperThreshold back in to get the proper
! value.
!
! The other strange feature of this code is the search order.  Our job
! is to find the biggest legal sum.  What do we do in case of ties?
! We stick with the sum found first.  Therefore, search order matters.
! We initialize the output sum with the value of the sum in the center
! of the 3x3 array, call it sum (0,0).  If no sum is bigger, this means
! the glyph is right where it was supposed to be.
!
! The somewhat funny feature is that we use alternating search orders.
! Say several of the sums around the outside of the glyph are the same
! value, but larger than the center sum.  Say this happened repeatedly
! on a series of glyphs.  If we always used the same search order, 
! we'd tend to drift off course.  Instead, we go to the trouble of 
! having two search orders.  One searches the sums from left-to-right, 
! top-to-bottom.  The other searches from right-to-left, bottom-to-top.
! The choice of search order is based on the least significant bit of 
! rNumOutBit.  If it's 0, we use the first search order.  If it's 1, 
! we use the second one.  Higher level routines should increment this
! quantity as we go along.  Normally, this is the parameter that indicates 
! which glpyh in a row is now being inspected.
!
! On entry, we expect the following of the register contents:
!
!	%fp:		A pointer to the saved parameters (see offsets below)
!	rKernel:	A pointer to the kernel data structure
!	rConvSum0:	The top row of convolution sums
!	rConvSum1:	The middle row of convolution sums
!	rConvSum2:	The bottom row of convolution sums
!	rNumOutBit:	LS bit indicates search order (see above)

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssglyph.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	.align	4

! DO NOT REDEFINE REGISTERS WITHOUT CHANGING THE REGISTER DEFINITIONS IN
! fglyph.c!!!!!

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rImage			%i0
#define rCenterXBit		%i1
#define	rBpl			%i2
#define	rKernel			%i3
#define	rNumOutBit		%i4

#define	rCompTableRow		%g1

#define	rBytesPerTableRow	%g2
#define	rCompTableBlock		%g3	/* used in convolve1x1 */
#define	rCompTableBlock0	%g3	/* used in convolve3x3 */
#define rCompTableBlock1	%g4
#define	rCompTableBlock2	%g5

#define	rConvSum		%o0
#define	rConvSum0		%o1
#define	rConvSum1		%o2
#define	rConvSum2		%o3

#define	rShiftCount		%o4
#define	rRightShiftCount	%o5

#define	rInput			%l0
#define	rInputRight		%l1

#define	rScratch		%l2
#define	rScratch0		%l3
#define	rScratch1		%l4
#define	rScratch2		%l5
#define	rBlockBits		%l6
#define rDepth			%l7

/* Magic C Routine Boilerplate.  These are positions of the subroutine's
 * parameters relative to %fp */
#define	imageOffset		0x44
#define	centerXBitOffset	0x48
#define	bplOffset		0x4c
#define	kernelOffset		0x50
#define numOutBit		0x54

LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! Common _findMaxOf3x3 ending to all center-finding convolution routines.
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(findMaxOf3x3)
MK_GLOBAL_NAME(findMaxOf3x3):

! First, set the three bias sums into rInput.  These are used to make all
! sums bigger than pKernel-(upperThreshold+1) wrap around.  The bias sum
! equals cSumBias - pKernel->(upperThreshold+1).  Use upperThreshold+1 
! since we want a result <= upperThreshold
	ld	[rKernel+cIAKUpperThr], rInput
	mov	cSumBias, rScratch
	sub	rScratch, rInput, rScratch	! contains 1 copy of bias val
	sub	rScratch, 1, rScratch		! really wanted 
						! upperThreshold + 1
	sll	rScratch, cSumSpacing, rInput
	or	rScratch, rInput, rScratch	! contains 2 copies of bias val
	sll	rInput, cSumSpacing, rInput
	or	rScratch, rInput, rScratch	! contains 3 copies of bias val

! Add the bias val to all the sums.  This makes all sums that were above
! the upper threshold appear to be smaller than all sums below it.  Thus
! legal sums will naturally win in a direct comparison.
	add	rConvSum0, rScratch, rConvSum0
	add	rConvSum1, rScratch, rConvSum1
	add	rConvSum2, rScratch, rConvSum2

! Put the image depth into rDepth.  We'll use it a lot below.
	ld	[rKernel+cIAKDepth], rDepth

! Put the mask 0xFFC00000 into rScratch0.  This will be used to extract
! the relevant 10 bits of the next sum.  Well, ok, it's really used to
! smash the LS bits to 1's using an orn (or not) opcode.
	sethi	%hi(0xFFC00000), rScratch0

! So far as I can tell, the easiest way to do the individual compares
! is to put the numbers we want to compare at the MS end of the
! registers and compare them.  This tends to ignore LS bits.  To
! guarantee that we don't choose a new sum if it's only equal to the
! old one, we force all the LS bits of the current best sum to be 1's.
! This makes this sum bigger than any other one having equal bits
! in the relevant portion and other stuff below.  Initialize the result
! from the middle sum using this alignment scheme.
	sll	rConvSum1, cSumSpacing, rConvSum
	orn	rConvSum, rScratch0, rConvSum

! To conserve registers, reuse rCenterXBit and rShiftCount.  rCenterXBit
! Will indicate how far we move to the right and rShiftCount will
! indicate who far we move down.  Both are relative to 0,0 at the
! middle of the 3x3 array.  Initialize rCenterXBit here and rShiftCount in
! the delay slot of the branch on rNumOutBit below.
	mov	0, rCenterXBit

! Finally, branch on the value of numOutBit.  If this number is even
! we search the sums from top left to bottom right (MS end of rConvSum0
! to LS end of rConvSum2)  IF this number is odd, we search in the
! other direction.  This tends to eliminate tracking bias as we move
! across a row of glyphs.
	btst	0x1, rNumOutBit
	bnz	fmSearchBackwards
	mov	cCenterAndCenter, rShiftCount	! use delay slot

! -----------------------------------------------------------------------

! This is the case where we search from left to right, top to bottom.
! Note that the current sum is kept left aligned.  We can compare it
! to each sum in each convolution reg in turn by merely shifting the
! convolution reg into position and doing a compare.  Since we have
! filled the lower bits of the current sum with 1's, the compare will
! naturally ignore any less significant bits.
! Coordinates shown below are in (x, y) notation.
fmForwardsM1M1:
	cmp	rConvSum, rConvSum0
	bgeu	fmForwards0M1
	sll	rConvSum0, cSumSpacing, rScratch
! The sum at (-1, -1) was bigger
	orn	rConvSum0, rScratch0, rConvSum	! get new biggest sum
	sub	%g0, rDepth, rCenterXBit	! new x position (-1)
	sethi	%hi(cUpAndLeft), rShiftCount	! y and x offsets

fmForwards0M1:
	cmp	rConvSum, rScratch
	bgeu	fmForwards1M1
	sll	rScratch, cSumSpacing, rConvSum0
! The sum at (0, -1) was bigger
	orn	rScratch, rScratch0, rConvSum	! get new biggest sum
	mov	0, rCenterXBit			! new x position
	sethi	%hi(cUpAndCenter), rShiftCount	! y and x offsets

fmForwards1M1:
	cmp	rConvSum, rConvSum0
	bgeu	fmForwardsM10
	cmp	rConvSum, rConvSum1
! The sum at (1, -1) was bigger
	orn	rConvSum0, rScratch0, rConvSum	! get new biggest sum
	cmp	rConvSum, rConvSum1		! redo the comparison
	mov	rDepth, rCenterXBit		! new x position
	sethi	%hi(cUpAndRight), rShiftCount	! y and x offsets

fmForwardsM10:
	bgeu	fmForwards10
	sll	rConvSum1, 2*cSumSpacing, rScratch
! The sum at (-1, 0) was bigger
	orn	rConvSum1, rScratch0, rConvSum	! get new biggest sum
	sub	%g0, rDepth, rCenterXBit	! new x position (-1)
	sethi	%hi(cCenterAndLeft), rShiftCount ! y and x offsets

fmForwards10:
	cmp	rConvSum, rScratch
	bgeu	fmForwardsM11
	cmp	rConvSum, rConvSum2
! The sum at (1, 0) was bigger
	orn	rScratch, rScratch0, rConvSum	! get new biggest sum
	cmp	rConvSum, rConvSum2		! redo the comparison
	mov	rDepth, rCenterXBit		! new x position
	sethi	%hi(cCenterAndRight), rShiftCount ! y and x offsets

fmForwardsM11:
	bgeu	fmForwards01
	sll	rConvSum2, cSumSpacing, rScratch
! The sum at (-1, 1) was bigger
	orn	rConvSum2, rScratch0, rConvSum	! get new biggest sum
	sub	%g0, rDepth, rCenterXBit	! new x position (-1)
	sethi	%hi(cDownAndLeft), rShiftCount 	! y and x offsets

fmForwards01:
	cmp	rConvSum, rScratch
	bgeu	fmForwards11
	sll	rScratch, cSumSpacing, rConvSum2
! The sum at (0, 1) was bigger
	orn	rScratch, rScratch0, rConvSum	! get new biggest sum
	mov	0, rCenterXBit			! new x position
	sethi	%hi(cDownAndCenter), rShiftCount ! y and x offsets

fmForwards11:
	cmp	rConvSum, rConvSum2
	bgeu	fmSetPointers
	ld	[%fp+imageOffset], rInput	! get ptr to ptr to image
! The sum at (1, 1) was bigger
	orn	rConvSum2, rScratch0, rConvSum	! get new biggest sum
	mov	rDepth, rCenterXBit		! new x position
	ba	fmSetPointers
	sethi	%hi(cDownAndRight), rShiftCount ! y and x offsets


! -----------------------------------------------------------------------

! This is the case where we search the sums from the lower right
! towards the upper left.  It's a lot like the version above
! except the search order is backwards (obviously) and we have to
! use rScratch1 to left align individual pixels.  We can't just shift
! the convolution sums themselves as was done above.

fmSearchBackwards:
fmBackwards11:
	sll	rConvSum2, 2*cSumSpacing, rScratch1

	cmp	rConvSum, rScratch1
	bgeu	fmBackwards01
	sll	rConvSum2, cSumSpacing, rScratch2
! The sum at (1, 1) was bigger
	orn	rScratch1, rScratch0, rConvSum	! get new biggest sum
	mov	rDepth, rCenterXBit		! new x position
	sethi	%hi(cDownAndRight), rShiftCount ! y and x offsets

fmBackwards01:
	cmp	rConvSum, rScratch2
	bgeu	fmBackwardsM11
	cmp	rConvSum, rConvSum2
! The sum at (0, 1) was bigger
	orn	rScratch2, rScratch0, rConvSum	! get new biggest sum
	cmp	rConvSum, rConvSum2		! redo the comparision
	mov	0, rCenterXBit			! new x position
	sethi	%hi(cDownAndCenter), rShiftCount ! y and x offsets

fmBackwardsM11:
	bgeu	fmBackwards10
	sll	rConvSum1, 2*cSumSpacing, rScratch1
! The sum at (-1, 1) was bigger
	orn	rConvSum2, rScratch0, rConvSum	! get new biggest sum
	sub	%g0, rDepth, rCenterXBit	! new x position (-1)
	sethi	%hi(cDownAndLeft), rShiftCount ! y and x offsets

fmBackwards10:
	cmp	rConvSum, rScratch1
	bgeu	fmBackwardsM10
	cmp	rConvSum, rConvSum1
! The sum at (1, 0) was bigger
	orn	rScratch1, rScratch0, rConvSum	! get new biggest sum
	cmp	rConvSum, rConvSum1		! redo the comparision
	mov	rDepth, rCenterXBit		! new x position
	sethi	%hi(cCenterAndRight), rShiftCount ! y and x offsets

fmBackwardsM10:
	bgeu	fmBackwards1M1
	sll	rConvSum0, 2*cSumSpacing, rScratch1
! The sum at (-1, 0) was bigger
	orn	rConvSum1, rScratch0, rConvSum	! get new biggest sum
	sub	%g0, rDepth, rCenterXBit	! new x position (-1)
	sethi	%hi(cCenterAndLeft), rShiftCount ! y and x offsets

fmBackwards1M1:
	cmp	rConvSum, rScratch1
	bgeu	fmBackwards0M1
	sll	rConvSum0, cSumSpacing, rScratch2
! The sum at (1, -1) was bigger
	orn	rScratch1, rScratch0, rConvSum	! get new biggest sum
	mov	rDepth, rCenterXBit		! new x position
	sethi	%hi(cUpAndRight), rShiftCount 	! y and x offsets

fmBackwards0M1:
	cmp	rConvSum, rScratch2
	bgeu	fmBackwardsM1M1
	cmp	rConvSum, rConvSum0
! The sum at (0, -1) was bigger
	orn	rScratch2, rScratch0, rConvSum	! get new biggest sum
	cmp	rConvSum, rConvSum0		! redo the comparision
	mov	0, rCenterXBit			! new x position
	sethi	%hi(cUpAndCenter), rShiftCount 	! y and x offsets

fmBackwardsM1M1:
	bgeu	fmSetPointers
	ld	[%fp+imageOffset], rInput	! get ptr to ptr to image
! The sum at (-1, -1) was bigger
	orn	rConvSum0, rScratch0, rConvSum	! get new biggest sum
	sub	%g0, rDepth, rCenterXBit	! new x position (-1)
	sethi	%hi(cUpAndLeft), rShiftCount 	! y and x offsets

! ---------------------------------------------------------------------

! Here we are all ready to set the return pointers.  rConvSum contains
! the best threshold (biased).  If the y coordinate in rShiftCount has
! changed from 0, adjust rImage.  If the x coordinate in rCenterXBit has
! changed, adjust the saved x offset.  If that moves us across a word
! boundary (should be rare), adjust rImage again.  Originally, we either
! updated neither coordinate or both of them.  This way, we should be
! able to do only the required updates - which should speed things up?
fmSetPointers:
	cmp	rShiftCount, 0			! did the x or y coord 
						! change?
	bz	fmReturnResult			! no, get the conv. sum
	ld	[rKernel + cIAKUpperThr], rScratch ! get ready to de-bias
						! the convolution sum

	sra	rShiftCount, cVerticalRightShift, rScratch0
						! get the y coord change
	cmp	rScratch0, 0			! did the y coord change
	bz	fmTestXCoord			! no, go see if the xCoord
						! changed
	ld	[%fp+centerXBitOffset], rInputRight ! get x offset in delay slot

! The y coordinate is +1 or -1.  Fetch the original address pointer, modify
! it appropriately and store it.  Note that rInput already has a pointer
! to this pointer.
	ld	[rInput], rImage		! get image pointer
	bneg,a	fmFUpdateImagePointer
	sub	rImage, rBpl, rImage		! done if y = -1
	add	rImage, rBpl, rImage		! done if y = +1
! Store the updated image pointer.
fmFUpdateImagePointer:
	st	rImage, [rInput]

! Did the X coordinate change?
fmTestXCoord:
	ld	[rInputRight], rScratch0	! get x offset
	cmp	rCenterXBit, 0
	bz	fmUpdateCenterXBit		! nope, but there was 
						! change in the y so
						! update centerXBit
! At this point, we need to adjust the centerXBit value.
! rInputRight will have the address of the centerXBit offset.  We'll save
! these addresses so we can update the args at the end of this routine.
	add	rCenterXBit, rScratch0, rCenterXBit ! calculate new centerXBit

! if rCenterXBit has crossed a word boundary, fix up rImage.  
! We must adjust rImage if rCenterXBit has either gone above 31 or below 0.
! Both can be detected by ANDing rCenterXBit with 0x20.
	btst	0x20, rCenterXBit		! was there a cross to -1
						! or +32?
	bz	fmUpdateCenterXBit
	! This is empty because it's ok to execute the next op here

! The xoffset crossed a word boundary.  Bummer.  Add the integer part of
! rCenterXBit that's above the 32-bit part of rCenterXBit
! to rImage.  That gives us an automatic increment or decrement.  Note that
! we need to multiply this increment by 4 since rImage should point to
! word boundaries.
	sra	rCenterXBit, 5, rScratch1
! must reload rImage.  If we didn't have to adjust y above, rImage
! will point to the last line that we used to compute the convolution.
! Instead of slowing down all cases to make sure that rImage is
! valid here, we reload now and only slow down this case.
	ld	[rInput], rImage		! get image pointer
	sll	rScratch1, 2, rScratch1		! make into word offset
	add	rImage, rScratch1, rImage	! add to address offset
	st	rImage, [rInput]		! store updated pointer
	and	rCenterXBit, 0x1F, rCenterXBit	! convert to a legal offset

! go update the image pointer.
fmUpdateCenterXBit:
	or	rShiftCount, rCenterXBit, rCenterXBit 
						! put the delta y and x
						! information in the top
						! 16 bits of centerXBit
	st	rCenterXBit, [rInputRight]	! update the offset

! At this point, the address pointers have been updated.  De-bias the
! sum in rConvSum by subtracting 2*cSumBias and adding pKernel->upperThreshold+1
! back in.  The upper threshold is in rScratch now.
fmReturnResult:
	srl	rConvSum, 2*cSumSpacing, rConvSum ! re-align the result
	sub	rConvSum, 2*cSumBias, rConvSum
	add	rConvSum, rScratch, rConvSum
	add	rConvSum, 1, rConvSum

! If the result is less than the upper threshold, return it.  Otherwise
! return 0.
	cmp	rConvSum, rScratch
	ble,a	fmRealReturn
	mov	rConvSum, %i0			! return the result
	mov	0, %i0				! return 0
fmRealReturn:
	ret
	restore


/*

$Log: ssglyph.s,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.7  1994/03/10 01:28:23  jcraig
ssglyph.s -> /product/ipcore/ipshared/src/RCS/ssglyph.s,v

in ip_convolve3x3 the most significant bits of *pCenterXBit now
contain the amount that the center moved up or down,,the amount that
the center moved left or right.

# Revision 1.6  1993/12/04  05:00:09  ddavies
# ssglyph.s -> /product/ipcore/ipshared/src/RCS/ssglyph.s,v
#
# Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
#
# Revision 1.5  1993/11/05  02:24:46  ddavies
# ssglyph.s -> /product/ipcore/ipshared/src/RCS/ssglyph.s,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

*/

