/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: ssgrybin.s
 *
 * This file contains assembly code to paint and mask gray images with
 * data in binary images.
 *
 * There are several small files, each one paints or masks one line from
 * one type of gray image.  The gray depths supported are 2, 4 and 8 bits/pixel.
 *
 * Only two binary-gray operations are currently supported: mask and paint.
 * The mask operation causes the output gray image to be white whereever the
 * binary image had a 0.  In other pixels, the source gray image is copied to
 * the destination gray image.
 *
 * The paint operation is symmetric.  It causes the output gray image to be
 * black everywhere the binary image had a 1.  In other locations, the input
 * gray image is copied to the output.
 *
 * For our purposes (Towada not withstanding), a 0 gray pixel is defined as
 * black.  The mask operation therefore forces pixels to be all 1's where
 * there's a 0 in the binary image.  It does this by expanding the binary
 * image to the depth of the gray image, then ORing the inverse of the
 * binary bits with the source gray bits.  Similarly, the paint operation
 * AND's the inverse of the expanded binary image with the source gray image.
 *
 * The inner loop of each routine reads a word of binary image and divides
 * it into bytes.  Each byte is expanded by the depth of the gray image and
 * ANDN'd or ORN'd with the gray image.  In the case of 2 bit/pixel gray
 * files, we do this twice to get a gray word.  In the case of 4 bit/pixel
 * gray files, each expansion covers an entire word.  For 8 bit/pixel gray
 * files, each nibble in the binary image is expanded to produce a gray
 * word.
 *
 * To get a little extra speed, we'll use double writes whenever possible.
 * To do this, we'll have two versions of each inner loop, one in which
 * the rightmost destination word has an even word address and one in which
 * it has an odd word address.  Even and odd destination words are calculated
 * separately and written as a pair.  Since double writes can only be done
 * to even word addresses, each version of the inner loop deals with the
 * end conditions appropriately.
 *
 * There's one thing to watch out for here.  The Binary pixrects are padded to
 * 32 bit boundaries.  The gray pixrects are also.  The binary and gray
 * pixrects have the same dimensions in units of pixels.  Does this mean that
 * the gray pixrects end up being padded to 32*depth bit boundaries?  No.
 * If, for example, the interesting part of the binary pixrect were 33 bits
 * wide, the interesting part of a 2 bit/pixel pixrect would be 66 bits wide
 * The binary pixrect would occupy 2 words, the gray pixrect 3 words.  An
 * 8 bit gray pixrect would be 264 bits wide, or 9 words wide (not 16).
 * So, we can't find the rightmost dest address word by simply multiplying
 * the offset of the rightmost binary word by depth.  Instead, we have to take
 * the width of the gray pixrect and divide by depth to find the width of the
 * binary pixrect.
 *
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssgrybin.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:19 $"
	.align	4

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPState	%i0		/* current addr in Dest gray image */
#define	rGrayAddr	%i1		/* current addr in Dest gray image */
#define rBinAddr	%i2		/* current addr in binary image */
#define rDstAddr	%i3		/* current addr in dest image */
#define rWidth		%i4		/* dest image's bytes/line */
#define rExpTable	%i5		/* address of byte expansion table */

#define	rBinData	%g1		/* holds binary data */
#define	rGrayData	%g2		/* holds source gray data */
#define	rExpBinData	%g3		/* holds expanded binary data */

#define rScr0		%g4		/* scratch registers */
#define rScr1		%g5

#define	rOut0		%o0		/* output register pair */
#define rOut1		%o1

#define rBinAddrP1	%o3		/* base used to access odd byte of
					 * each byte pair in 2bpp */
#define rBinOffset	%o4		/* current horizontal Binary offset */

#define	rRetAddr	%o7		/* defined by SPARC's call inst */

/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	grayAddrOffset	0x48
#define	binAddrOffset	0x4c
#define	dstAddrOffset	0x50
#define	widthOffset	0x54
#define	expTableOffset	0x58

/* magical C Frame constant (no local variables) */
#define	LF26		-64

LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! _ip_paint2bppLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_paint2bppLine)
MK_GLOBAL_NAME(ip_paint2bppLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState,  [%fp+ipStateOffset]
	st	rGrayAddr, [%fp+grayAddrOffset]
	st	rBinAddr,  [%fp+binAddrOffset]
	st	rDstAddr,  [%fp+dstAddrOffset]
	st	rWidth,    [%fp+widthOffset]
	st	rExpTable, [%fp+expTableOffset]

! To save on an index, we'll process each line from right to left.  That way,
! we can use the address offset as the loop index.  rWidth contains the
! number of bytes to be processed in the gray world.
! The first byte to be read in the binary world therefore has an offset of
! (rWidth-1)/2.  To save a little time, we want to read binary bytes in pairs,
! left one first.  For example, in a line with bytes 0, 1, 2, 3; we'd read them
! in the order 2, 3, 0, 1.  rBinOffset should therefore start at
! ((rWidth-1)/2)-1.  Except rBinOffset is decremented by 2 before it's
! used (that's the way the loop came out), so the initial value should be
! ((rWidth-1)/2)+1.
! Ummm.  How do we guarantee that the byte pair we read actually lines up with
! the first dest word computed?  Since we'll always compute integral numbers
! of dest words, we want to guarantee that the leftmost byte of the first byte
! pair is at an EVEN offset relative to the beginning of the line.  Since we're
! already pointing at the last byte of the line, we want to round DOWN to get
! that first byte.  So, starting from first principles, the first binary byte
! offset ought to be ((rWidth-1)/2 AND ~1) to get the even offset, then +2
! to account for the initial subtraction.
	sub	rWidth, 1, rWidth		! rWidth-1
	srl	rWidth, 1, rBinOffset		! (rWidth-1)/2
	andn	rBinOffset, 1, rBinOffset	! round down to next even offset
	add	rBinOffset, 2, rBinOffset	! ((rWidth-1)/2 AND ~1) + 2
	add	rBinAddr, 1, rBinAddrP1		! base used to access odd byte

! The rightmost byte of the gray line is at offset (rWidth-1).  However, we
! need a word address, not a byte address.  To get it, we round up to the
! next multiple of 4 by calculating ((rWidth-1)+3) & ~3.  Yeah, but isn't that
! the address of the word to the right of the one we want?  Yes, but we'll
! decrement rWidth before use, so it's ok.  Yeah, but we can't use this
! offset to do the first read, right?  Correct, so we decrement rGrayAddr
! by 4 to account for this.
! NOT QUITE!  Say rWidth = 17.  That means we want to start writing at byte
! offset 16.  rWidth-1 = 16.  (rWidth-1)+3 & ~3 = 16.  When we do the first
! decrement BEFORE the write, we'll start with byte offset 12, missing the
! last byte.  What we really want is (rWidth+3) & ~3.
	add	rWidth, 4, rWidth		! (rWidth-1) + 4
	andn	rWidth, 3, rWidth		! offset of word above first
						! dest word.

! START OF LOOP OVER ROW

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  Both cases
! have particular code for starting the row.  Note that rWidth is decremented
! before use, so the sense of the test is reversed.

	mov	0, rOut1			! if the rightmost addr is
						! even, this will be written
						! over the already 0'd frame
	add	rDstAddr, rWidth, rScr0		! calculate dest address
	andcc	rScr0, 0x4, %g0			! is first addr even?
	bnz	paint2CalcEvenWord		! yes, jump into the middle of
						! the loop
	sub	rGrayAddr, 0x4, rGrayAddr	! base used to read gray data
						! since rWidth is 4 too big.

paint2RowLoop:
	sub	rBinOffset, 2, rBinOffset	! point to next byte pair
	ldub	[rBinAddr+rBinOffset], rBinData ! get even binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! next gray word
	sll	rBinData, 1, rScr0		! convert to halfword index
	ldub	[rBinAddrP1+rBinOffset], rBinData ! get odd binary byte
	lduh	[rExpTable+rScr0],rExpBinData	! get upper half of expanded wd
	sll	rBinData, 1, rScr0
	lduh	[rExpTable+rScr0], rScr0	! get lower half of expanded wd
	sll	rExpBinData, 16, rExpBinData	! shift upper half into position
	or	rExpBinData, rScr0, rExpBinData	! combine the halves
	andn	rGrayData, rExpBinData, rOut1	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bz,a	paint2Fin			! yes, quit
	st	rOut1, [rDstAddr+rWidth]	! and write leftmost word
	
! **** CALCULATE EVEN DESTINATION WORD ****

paint2CalcEvenWord:
	sub	rBinOffset, 2, rBinOffset	! point to next byte pair
	ldub	[rBinAddr+rBinOffset], rBinData ! get even binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! next gray word
	sll	rBinData, 1, rScr0		! convert to halfword index
	ldub	[rBinAddrP1+rBinOffset], rBinData ! get odd binary byte
	lduh	[rExpTable+rScr0],rExpBinData	! get upper half of expanded wd
	sll	rBinData, 1, rScr0		! convert to halfword index
	lduh	[rExpTable+rScr0], rScr0	! get lower half of expanded wd
	sll	rExpBinData, 16, rExpBinData	! shift upper half into position
	or	rExpBinData, rScr0, rExpBinData	! combine the halves
	andn	rGrayData, rExpBinData, rOut0	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bnz	paint2RowLoop			! no, keep going
	std	rOut0, [rDstAddr+rWidth]	! and write current double wd

! END OF LOOP ACROSS 1 ROW

! We've put out the right number of words, so quit
paint2Fin:
	mov	0,%i0
	ret
	restore


! -----------------------------------------------------------------------------
! _ip_mask2bppLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_mask2bppLine)
MK_GLOBAL_NAME(ip_mask2bppLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState,  [%fp+ipStateOffset]
	st	rGrayAddr, [%fp+grayAddrOffset]
	st	rBinAddr,  [%fp+binAddrOffset]
	st	rDstAddr,  [%fp+dstAddrOffset]
	st	rWidth,    [%fp+widthOffset]
	st	rExpTable, [%fp+expTableOffset]

! To save on an index, we'll process each line from right to left.  That way,
! we can use the address offset as the loop index.  rWidth contains the
! number of bytes to be processed in the gray world.
! The first word to be read in the binary world therefore has an offset of
! (rWidth-1)/2.  To save a little time, we want to read binary bytes in pairs,
! left one first.  For example, in a line with bytes 0, 1, 2, 3; we'd read them
! in the order 2, 3, 0, 1.  rBinOffset should therefore start at
! ((rWidth-1)/2)-1.  Well, except rBinOffset is decremented by 2 before it's
! used (that's the way the loop came out), so the initial value should be
! ((rWidth-1)/2)+1.
! Ummm.  How do we guarantee that the byte pair we read actually lines up with
! the first dest word computed?  Since we'll always compute integral numbers
! of dest words, we want to guarantee that the leftmost byte of the first byte
! pair is at an EVEN offset relative to the beginning of the line.  Since we're
! already pointing at the last byte of the line, we want to round DOWN to get
! that first byte.  So, starting from first principles, the first binary byte
! offset ought to be ((rWidth-1)/2 AND ~1) to get the even offset, then +2
! to account for the initial subtraction.
	sub	rWidth, 1, rWidth		! rWidth-1
	srl	rWidth, 1, rBinOffset		! (rWidth-1)/2
	andn	rBinOffset, 1, rBinOffset	! round down to next even offset
	add	rBinOffset, 2, rBinOffset	! ((rWidth-1)/2 AND ~1) + 2
	add	rBinAddr, 1, rBinAddrP1		! base used to access odd byte

! The rightmost byte of the gray line is at offset (rWidth-1).  However, we
! need a word address, not a byte address.  To get it, we round up to the
! next multiple of 4 by calculating ((rWidth-1)+3) & ~3.  Yeah, but isn't that
! the address of the word to the right of the one we want?  Yes, but we'll
! decrement rWidth before use, so it's ok.  Yeah, but we can't use this
! offset to do the first read, right?  Correct, so we decrement rGrayAddr
! by 4 to account for this.
! NOT QUITE!  Say rWidth = 17.  That means we want to start writing at byte
! offset 16.  rWidth-1 = 16.  (rWidth-1)+3 & ~3 = 16.  When we do the first
! decrement BEFORE the write, we'll start with byte offset 12, missing the
! last byte.  What we really want is (rWidth+3) & ~3.
	add	rWidth, 4, rWidth		! (rWidth-1) + 4
	andn	rWidth, 3, rWidth		! offset of word above first
						! dest word.

! START OF LOOP OVER ROW

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  Both cases
! have particular code for starting the row.  Note that rWidth is decremented
! before use, so the sense of the test is reversed.

	mov	0, rOut1			! if the rightmost addr is
						! even, this will be written
						! over the already 0'd frame
	add	rDstAddr, rWidth, rScr0		! calculate dest address
	andcc	rScr0, 0x4, %g0			! is first addr even?
	bnz	mask2CalcEvenWord		! yes, jump into the middle of
						! the loop
	sub	rGrayAddr, 0x4, rGrayAddr	! base used to read gray data
						! since rWidth is 4 too big.

mask2RowLoop:
	sub	rBinOffset, 2, rBinOffset	! point to next byte pair
	ldub	[rBinAddr+rBinOffset], rBinData ! get even binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! next gray word
	sll	rBinData, 1, rScr0		! convert to halfword index
	ldub	[rBinAddrP1+rBinOffset], rBinData ! get odd binary byte
	lduh	[rExpTable+rScr0],rExpBinData	! get upper half of expanded wd
	sll	rBinData, 1, rScr0
	lduh	[rExpTable+rScr0], rScr0	! get lower half of expanded wd
	sll	rExpBinData, 16, rExpBinData	! shift upper half into position
	or	rExpBinData, rScr0, rExpBinData	! combine the halves
	orn	rGrayData, rExpBinData, rOut1	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bz,a	mask2Fin			! yes, quit
	st	rOut1, [rDstAddr+rWidth]	! and write leftmost word
	
! **** CALCULATE EVEN DESTINATION WORD ****

mask2CalcEvenWord:
	sub	rBinOffset, 2, rBinOffset	! point to next byte pair
	ldub	[rBinAddr+rBinOffset], rBinData ! get even binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! next gray word
	sll	rBinData, 1, rScr0		! convert to halfword index
	ldub	[rBinAddrP1+rBinOffset], rBinData ! get odd binary byte
	lduh	[rExpTable+rScr0],rExpBinData	! get upper half of expanded wd
	sll	rBinData, 1, rScr0		! convert to halfword index
	lduh	[rExpTable+rScr0], rScr0	! get lower half of expanded wd
	sll	rExpBinData, 16, rExpBinData	! shift upper half into position
	or	rExpBinData, rScr0, rExpBinData	! combine the halves
	orn	rGrayData, rExpBinData, rOut0	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bnz	mask2RowLoop			! no, keep going
	std	rOut0, [rDstAddr+rWidth]	! and write current double wd

! END OF LOOP ACROSS 1 ROW

! We've put out the right number of words, so quit
mask2Fin:
	mov	0,%i0
	ret
	restore



! -----------------------------------------------------------------------------
! _ip_paint4bppLine Routine
! -----------------------------------------------------------------------------

! This routine is a lot like paint2bppLine except each byte is expanded into
! a full word immediately.  We don't have to bother with reading byte pairs.

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_paint4bppLine)
MK_GLOBAL_NAME(ip_paint4bppLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState,  [%fp+ipStateOffset]
	st	rGrayAddr, [%fp+grayAddrOffset]
	st	rBinAddr,  [%fp+binAddrOffset]
	st	rDstAddr,  [%fp+dstAddrOffset]
	st	rWidth,    [%fp+widthOffset]
	st	rExpTable, [%fp+expTableOffset]

! To save on an index, we'll process each line from right to left.  That way,
! we can use the address offset as the loop index.  rWidth contains the
! number of bytes to be processed in the gray world.
! The first byte to be read in the binary world therefore has an offset of
! (rWidth-1)/4.
	sub	rWidth, 1, rWidth		! rWidth-1
	srl	rWidth, 2, rBinOffset		! (rWidth-1)/4

! The rightmost byte of the gray line is at offset (rWidth-1).  However, we
! need a word address, not a byte address.  To get it, we round up to the
! next multiple of 4 by calculating ((rWidth-1)+3) & ~3.  Yeah, but isn't that
! the address of the word to the right of the one we want?  Yes, but we'll
! decrement rWidth before use, so it's ok.  Yeah, but we can't use this
! offset to do the first read, right?  Correct, so we decrement rGrayAddr
! by 4 to account for this.
! NOT QUITE!  Say rWidth = 17.  That means we want to start writing at byte
! offset 16.  rWidth-1 = 16.  (rWidth-1)+3 & ~3 = 16.  When we do the first
! decrement BEFORE the write, we'll start with byte offset 12, missing the
! last byte.  What we really want is (rWidth+3) & ~3.
	add	rWidth, 4, rWidth		! (rWidth-1) + 4
	andn	rWidth, 3, rWidth		! offset of word above first
						! dest word.

! START OF LOOP OVER ROW

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  Both cases
! have particular code for starting the row.  Note that rWidth is decremented
! before use, so the sense of the test is reversed.

	mov	0, rOut1			! if the rightmost addr is
						! even, this will be written
						! over the already 0'd frame
	add	rDstAddr, rWidth, rScr0		! calculate dest address
	andcc	rScr0, 0x4, %g0			! is first addr even?
	bnz	paint4CalcEvenWord		! yes, jump into the middle of
						! the loop
	sub	rGrayAddr, 0x4, rGrayAddr	! base used to read gray data
						! since rWidth is 4 too big.

paint4RowLoop:
	ldub	[rBinAddr+rBinOffset], rBinData ! get next binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	sll	rBinData, 2, rScr0		! convert to word index
	ld	[rExpTable+rScr0],rExpBinData	! expand byte to cover word
	sub	rBinOffset, 1, rBinOffset	! point to next byte
	andn	rGrayData, rExpBinData, rOut1	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bz,a	paint4Fin			! yes, quit
	st	rOut1, [rDstAddr+rWidth]	! and write leftmost word
	
! **** CALCULATE EVEN DESTINATION WORD ****

paint4CalcEvenWord:
	ldub	[rBinAddr+rBinOffset], rBinData ! get next binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	sll	rBinData, 2, rScr0		! convert to word index
	ld	[rExpTable+rScr0],rExpBinData	! expand byte to cover word
	sub	rBinOffset, 1, rBinOffset	! point to next byte
	andn	rGrayData, rExpBinData, rOut0	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bnz	paint4RowLoop			! no, keep going
	std	rOut0, [rDstAddr+rWidth]	! and write current double word


! END OF LOOP ACROSS 1 ROW

! We've put out the right number of words, so quit
paint4Fin:
	mov	0,%i0
	ret
	restore


! -----------------------------------------------------------------------------
! _ip_mask4bppLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_mask4bppLine)
MK_GLOBAL_NAME(ip_mask4bppLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState,  [%fp+ipStateOffset]
	st	rGrayAddr, [%fp+grayAddrOffset]
	st	rBinAddr,  [%fp+binAddrOffset]
	st	rDstAddr,  [%fp+dstAddrOffset]
	st	rWidth,    [%fp+widthOffset]
	st	rExpTable, [%fp+expTableOffset]

! To save on an index, we'll process each line from right to left.  That way,
! we can use the address offset as the loop index.  rWidth contains the
! number of bytes to be processed in the gray world.
! The first byte to be read in the binary world therefore has an offset of
! (rWidth-1)/4.
	sub	rWidth, 1, rWidth		! rWidth-1
	srl	rWidth, 2, rBinOffset		! (rWidth-1)/4

! The rightmost byte of the gray line is at offset (rWidth-1).  However, we
! need a word address, not a byte address.  To get it, we round up to the
! next multiple of 4 by calculating ((rWidth-1)+3) & ~3.  Yeah, but isn't that
! the address of the word to the right of the one we want?  Yes, but we'll
! decrement rWidth before use, so it's ok.  Yeah, but we can't use this
! offset to do the first read, right?  Correct, so we decrement rGrayAddr
! by 4 to account for this.
! NOT QUITE!  Say rWidth = 17.  That means we want to start writing at byte
! offset 16.  rWidth-1 = 16.  (rWidth-1)+3 & ~3 = 16.  When we do the first
! decrement BEFORE the write, we'll start with byte offset 12, missing the
! last byte.  What we really want is (rWidth+3) & ~3.
	add	rWidth, 4, rWidth		! (rWidth-1) + 4
	andn	rWidth, 3, rWidth		! offset of word above first
						! dest word.

! START OF LOOP OVER ROW

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  Both cases
! have particular code for starting the row.  Note that rWidth is decremented
! before use, so the sense of the test is reversed.

	mov	0, rOut1			! if the rightmost addr is
						! even, this will be written
						! over the already 0'd frame
	add	rDstAddr, rWidth, rScr0		! calculate dest address
	andcc	rScr0, 0x4, %g0			! is first addr even?
	bnz	mask4CalcEvenWord		! yes, jump into the middle of
						! the loop
	sub	rGrayAddr, 0x4, rGrayAddr	! base used to read gray data
						! since rWidth is 4 too big.

mask4RowLoop:
	ldub	[rBinAddr+rBinOffset], rBinData ! get next binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	sll	rBinData, 2, rScr0		! convert to word index
	ld	[rExpTable+rScr0],rExpBinData	! expand byte to cover word
	sub	rBinOffset, 1, rBinOffset	! point to next byte
	orn	rGrayData, rExpBinData, rOut1	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bz,a	mask4Fin			! yes, quit
	st	rOut1, [rDstAddr+rWidth]	! and write leftmost word
	
! **** CALCULATE EVEN DESTINATION WORD ****

mask4CalcEvenWord:
	ldub	[rBinAddr+rBinOffset], rBinData ! get next binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	sll	rBinData, 2, rScr0		! convert to word index
	ld	[rExpTable+rScr0],rExpBinData	! expand byte to cover word
	sub	rBinOffset, 1, rBinOffset	! point to next byte
	orn	rGrayData, rExpBinData, rOut0	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bnz	mask4RowLoop			! no, keep going
	std	rOut0, [rDstAddr+rWidth]	! and write current double word


! END OF LOOP ACROSS 1 ROW

! We've put out the right number of words, so quit
mask4Fin:
	mov	0,%i0
	ret
	restore




! -----------------------------------------------------------------------------
! _ip_paint8bppLine Routine
! -----------------------------------------------------------------------------

! This is similar to paint4bppLine except we expand nibbles instead of bytes.

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_paint8bppLine)
MK_GLOBAL_NAME(ip_paint8bppLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState,  [%fp+ipStateOffset]
	st	rGrayAddr, [%fp+grayAddrOffset]
	st	rBinAddr,  [%fp+binAddrOffset]
	st	rDstAddr,  [%fp+dstAddrOffset]
	st	rWidth,    [%fp+widthOffset]
	st	rExpTable, [%fp+expTableOffset]

! The rightmost byte of the gray line is at offset (rWidth-1).  However, we
! need a word address, not a byte address.  To get it, we round up to the
! next multiple of 4 by calculating ((rWidth-1)+3) & ~3.  Yeah, but isn't that
! the address of the word to the right of the one we want?  Yes, but we'll
! decrement rWidth before use, so it's ok.  Yeah, but we can't use this
! offset to do the first read, right?  Correct, so we decrement rGrayAddr
! by 4 to account for this.
! NOT QUITE!  Say rWidth = 17.  That means we want to start writing at byte
! offset 16.  rWidth-1 = 16.  (rWidth-1)+3 & ~3 = 16.  When we do the first
! decrement BEFORE the write, we'll start with byte offset 12, missing the
! last byte.  What we really want is (rWidth+3) & ~3.
	sub	rWidth, 1, rScr1		! rWidth-1
	add	rWidth, 3, rWidth		! rWidth + 3
	andn	rWidth, 3, rWidth		! offset of word above first
						! dest word.

! To save on an index, we'll process each line from right to left.  That way,
! we can use the address offset as the loop index.  rWidth contains the
! number of bytes to be processed in the gray world.
! The first byte to be read in the binary world therefore has an offset of
! (rWidth-1)/8.
! Yeah, but which nibble from that byte do we use?  There are really two
! types of inner loop: one in which even (MS) nibbles are used to calculate
! destination words with odd addresses and one in which odd (LS) nibbles are
! used to calculate dest words with odd addresses.  We decide which loop
! to enter based on bit 2 of (rWidth-1).  Since we're going to divide
! (rWidth-1) by 8 to get the byte address, bit 2 selects the nibble.
	srl	rScr1, 3, rBinOffset		! (rWidth-1)/8

! START OF LOOP OVER ROW

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  Both cases
! have particular code for starting the row.  Note that rWidth is decremented
! before use, so the sense of the test is reversed.

	mov	0, rOut1			! if the rightmost addr is
						! even, this will be written
						! over the already 0'd frame
	andcc	rScr1, 0x4, %g0			! look at nibble select bit
	bnz	paint8UseOddNibbleFirst		! if it's set, use lower nib
	add	rDstAddr, rWidth, rScr0		! calculate dest address

	ldub	[rBinAddr+rBinOffset], rBinData	! get first binary byte
	andcc	rScr0, 0x4, %g0			! is first addr even?
	srl	rBinData, 2, rScr0		! convert MS nibble to wd index
	bnz	paint8CalcEvenWordEvenNib	! yes, jump into the middle of
						! the loop way below
	sub	rGrayAddr, 0x4, rGrayAddr	! base used to read gray data
						! since rWidth is 4 too big.

! If we get here, we know the even nibble will be used first and the first
! word to be calculated is at an odd address.  Therefore,
! this pair of loops is used when the even (MS) nibble is used to
! calculate the destination word at the odd address.
! REMEMBER we're processing from right to left so must read a new byte
! just before the odd nibble is used.
paint8RowLoopEvenNib:
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	and	rScr0, 0x3C, rScr0		! get rid of low nibble
	ld	[rExpTable+rScr0], rExpBinData	! expand byte to cover word
	sub	rBinOffset, 1, rBinOffset	! point to next byte
	andn	rGrayData, rExpBinData, rOut1	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bz,a	paint8Fin			! yes, quit
	st	rOut1, [rDstAddr+rWidth]	! and write leftmost word
	
! **** CALCULATE EVEN DESTINATION WORD ****

paint8CalcEvenWordOddNib:
	ldub	[rBinAddr+rBinOffset], rBinData ! get next binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	sll	rBinData, 2, rScr0		! align odd nibble
	and	rScr0, 0x3C, rScr0		! get rid of high nibble
	ld	[rExpTable+rScr0], rExpBinData	! expand byte to cover word
	srl	rBinData, 2, rScr0		! convert MS nibble to wd index
	andn	rGrayData, rExpBinData, rOut0	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bnz	paint8RowLoopEvenNib		! no, keep going
	std	rOut0, [rDstAddr+rWidth]	! and write current double word
	ba	paint8Fin			! we're done
	nop


! Here, we know that the odd nibble is to be used first.  The question is
! whether the first dest word is at an even or odd address.
paint8UseOddNibbleFirst:
	andcc	rScr0, 0x4, %g0			! is first dest addr even?
	bnz	paint8CalcEvenWordOddNib	! yes, jump into the middle of
						! the loop above
	sub	rGrayAddr, 0x4, rGrayAddr	! base used to read gray data
						! since rWidth is 4 too big.

! If we get here, we know the first nibble to be used is odd and the first
! dest address is odd.  Therefore, this is the set of loops that uses the
! odd nibble to calculate words at odd addresses.
paint8RowLoopOddNib:
	ldub	[rBinAddr+rBinOffset], rBinData ! get next binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	sll	rBinData, 2, rScr0		! align odd nibble
	and	rScr0, 0x3C, rScr0		! get rid of even nibble
	ld	[rExpTable+rScr0], rExpBinData	! expand byte to cover word
	srl	rBinData, 2, rScr0		! convert MS nibble to wd index
	andn	rGrayData, rExpBinData, rOut1	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bz,a	paint8Fin			! yes, quit
	st	rOut1, [rDstAddr+rWidth]	! and write leftmost word
	
! **** CALCULATE EVEN DESTINATION WORD ****

paint8CalcEvenWordEvenNib:
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	and	rScr0, 0x3C, rScr0		! get rid of low nibble
	ld	[rExpTable+rScr0], rExpBinData	! expand byte to cover word
	sub	rBinOffset, 1, rBinOffset	! point to next byte
	andn	rGrayData, rExpBinData, rOut0	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bnz	paint8RowLoopOddNib		! no, keep going
	std	rOut0, [rDstAddr+rWidth]	! and write current double word


! END OF LOOP ACROSS 1 ROW

! We've put out the right number of words, so quit
paint8Fin:
	mov	0,%i0
	ret
	restore


! -----------------------------------------------------------------------------
! _ip_mask8bppLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_mask8bppLine)
MK_GLOBAL_NAME(ip_mask8bppLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState,  [%fp+ipStateOffset]
	st	rGrayAddr, [%fp+grayAddrOffset]
	st	rBinAddr,  [%fp+binAddrOffset]
	st	rDstAddr,  [%fp+dstAddrOffset]
	st	rWidth,    [%fp+widthOffset]
	st	rExpTable, [%fp+expTableOffset]


! The rightmost byte of the gray line is at offset (rWidth-1).  However, we
! need a word address, not a byte address.  To get it, we round up to the
! next multiple of 4 by calculating ((rWidth-1)+3) & ~3.  Yeah, but isn't that
! the address of the word to the right of the one we want?  Yes, but we'll
! decrement rWidth before use, so it's ok.  Yeah, but we can't use this
! offset to do the first read, right?  Correct, so we decrement rGrayAddr
! by 4 to account for this.
! NOT QUITE!  Say rWidth = 17.  That means we want to start writing at byte
! offset 16.  rWidth-1 = 16.  (rWidth-1)+3 & ~3 = 16.  When we do the first
! decrement BEFORE the write, we'll start with byte offset 12, missing the
! last byte.  What we really want is (rWidth+3) & ~3.
	sub	rWidth, 1, rScr1		! rWidth-1
	add	rWidth, 3, rWidth		! rWidth + 3
	andn	rWidth, 3, rWidth		! offset of word above first
						! dest word.

! To save on an index, we'll process each line from right to left.  That way,
! we can use the address offset as the loop index.  rWidth contains the
! number of bytes to be processed in the gray world.
! The first byte to be read in the binary world therefore has an offset of
! (rWidth-1)/8.
! Yeah, but which nibble from that byte do we use?  There are really two
! types of inner loop: one in which even (MS) nibbles are used to calculate
! destination words with odd addresses and one in which odd (LS) nibbles are
! used to calculate dest words with odd addresses.  We decide which loop
! to enter based on bit 2 of (rWidth-1).  Since we're going to divide
! (rWidth-1) by 8 to get the byte address, bit 2 selects the nibble.
	srl	rScr1, 3, rBinOffset		! (rWidth-1)/8

! START OF LOOP OVER ROW

! We're ready to start.  To decide how to enter the row loop, we figure out
! whether the address of the rightmost dest word is even or odd.  Both cases
! have particular code for starting the row.  Note that rWidth is decremented
! before use, so the sense of the test is reversed.

	mov	0, rOut1			! if the rightmost addr is
						! even, this will be written
						! over the already 0'd frame
	andcc	rScr1, 0x4, %g0			! look at nibble select bit
	bnz	mask8UseOddNibbleFirst		! if it's set, use lower nib
	add	rDstAddr, rWidth, rScr0		! calculate dest address

	ldub	[rBinAddr+rBinOffset], rBinData	! get first binary byte
	andcc	rScr0, 0x4, %g0			! is first addr even?
	srl	rBinData, 2, rScr0		! convert MS nibble to wd index
	bnz	mask8CalcEvenWordEvenNib	! yes, jump into the middle of
						! the loop way below
	sub	rGrayAddr, 0x4, rGrayAddr	! base used to read gray data
						! since rWidth is 4 too big.

! If we get here, we know the even nibble will be used first and the first
! word to be calculated is at an odd address.  Therefore,
! this pair of loops is used when the even (MS) nibble is used to
! calculate the destination word at the odd address.
! REMEMBER we're processing from right to left so must read a new byte
! just before the odd nibble is used.
mask8RowLoopEvenNib:
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	and	rScr0, 0x3C, rScr0		! get rid of low nibble
	ld	[rExpTable+rScr0], rExpBinData	! expand byte to cover word
	sub	rBinOffset, 1, rBinOffset	! point to next byte
	orn	rGrayData, rExpBinData, rOut1	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bz,a	mask8Fin			! yes, quit
	st	rOut1, [rDstAddr+rWidth]	! and write leftmost word
	
! **** CALCULATE EVEN DESTINATION WORD ****

mask8CalcEvenWordOddNib:
	ldub	[rBinAddr+rBinOffset], rBinData ! get next binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	sll	rBinData, 2, rScr0		! align odd nibble
	and	rScr0, 0x3C, rScr0		! get rid of high nibble
	ld	[rExpTable+rScr0], rExpBinData	! expand byte to cover word
	srl	rBinData, 2, rScr0		! convert MS nibble to wd index
	orn	rGrayData, rExpBinData, rOut0	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bnz	mask8RowLoopEvenNib		! no, keep going
	std	rOut0, [rDstAddr+rWidth]	! and write current double word
	ba	mask8Fin			! we're done
	nop


! Here, we know that the odd nibble is to be used first.  The question is
! whether the first dest word is at an even or odd address.
mask8UseOddNibbleFirst:
	andcc	rScr0, 0x4, %g0			! is first dest addr even?
	bnz	mask8CalcEvenWordOddNib	! yes, jump into the middle of
						! the loop above
	sub	rGrayAddr, 0x4, rGrayAddr	! base used to read gray data
						! since rWidth is 4 too big.

! If we get here, we know the first nibble to be used is odd and the first
! dest address is odd.  Therefore, this is the set of loops that uses the
! odd nibble to calculate words at odd addresses.
mask8RowLoopOddNib:
	ldub	[rBinAddr+rBinOffset], rBinData ! get next binary byte
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	sll	rBinData, 2, rScr0		! align odd nibble
	and	rScr0, 0x3C, rScr0		! get rid of even nibble
	ld	[rExpTable+rScr0], rExpBinData	! expand byte to cover word
	srl	rBinData, 2, rScr0		! convert MS nibble to wd index
	orn	rGrayData, rExpBinData, rOut1	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bz,a	mask8Fin			! yes, quit
	st	rOut1, [rDstAddr+rWidth]	! and write leftmost word
	
! **** CALCULATE EVEN DESTINATION WORD ****

mask8CalcEvenWordEvenNib:
	ld	[rGrayAddr+rWidth], rGrayData	! get next gray word
	and	rScr0, 0x3C, rScr0		! get rid of low nibble
	ld	[rExpTable+rScr0], rExpBinData	! expand byte to cover word
	sub	rBinOffset, 1, rBinOffset	! point to next byte
	orn	rGrayData, rExpBinData, rOut0	! calculate output word
	subcc	rWidth, 0x4, rWidth		! done with row now?
	bnz	mask8RowLoopOddNib		! no, keep going
	std	rOut0, [rDstAddr+rWidth]	! and write current double word


! END OF LOOP ACROSS 1 ROW

! We've put out the right number of words, so quit
mask8Fin:
	mov	0,%i0
	ret
	restore


/*

$Log: ssgrybin.s,v $
Revision 1.1.1.1  1999/09/20 18:18:19  jlin
Imported sources

Revision 1.2  1998/01/13 23:57:46  ddavies
ssgrybin.s -> /product/ipcore/ipshared/src/RCS/ssgrybin.s,v

CommonCheck: Add IPState parameter to almost all private routines.

# Revision 1.1  1994/01/14  04:32:12  ddavies
# Initial revision
#
! Revision 1.5  1993/12/04  05:00:16  ddavies
! ssgrbin.s -> /product/ipcore/ipshared/src/RCS/ssgrbin.s,v
!
! Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
!
! Revision 1.4  1993/11/05  02:24:58  ddavies
! ssgrbin.s -> /product/ipcore/ipshared/src/RCS/ssgrbin.s,v
!
! Add RCS magic to have string specifying this file's name, rev number and
! last edit date into the object file.  Also add change log at the bottom.
!

*/

