/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: ssgsrot.s
 *
 * This file contains an assembly version of ip_rotateGrayLine as encoded in
 * gsrot.c.  ip_rotateGrayLine computes one line of a gray scale rotated image
 * by interpolating pixels in an input image.
 *
 * To understand this glop, the reader should be familiar with both the
 * voluminous comments in gsrot.c and the ip_rotateGrayLine code, also in
 * gsrot.c.
 *
 * The line is calculated as a series of 8 byte chunks.  For each chunk,
 * we take diagonal sweeps of input pixels (the sweeps are horizontal in
 * the input coordinate system - diagonal in the output coordinate system)
 * until we've covered the output, then write it out.  In each sweep, we
 * disperse each input pixel's value among the corresponding output pixels.
 *
 * We thought about writing the two output words using a double write, but
 * when you include the four flavors of inner loops required (see ssframe.s),
 * the cost doesn't seem to justify the tiny extra performance.  It takes about
 * 0.6 seconds to write 15 MB using double writes and 1.2 seconds using single
 * writes.  A rotation of a 15 MB image should take about 15-20 seconds total.
 * The 0.6 sec saved doesn't make a lot of difference here.
 *
 * There one subtle, but important, point to note.  rotateGrayPr passes all
 * coordinates down here as through we were doing clockwise rotation.  If it's
 * really counterclockwise, rSrcPitch is negative and we're starting from the
 * lower right corner of the image, but we don't need to know that here.
 * Therefore, when deciding whether a horizontal movement in the source
 * corresponds to a slight upward or downward movement in the destination, we
 * can always assume clockwise rotation (left->right in source => upper left
 * -> lower right in destination).
 *
 ******************* FIRST ASSEMBLY CODE REWORK ***************************
 *
 * This counts as the third rework of the algorithm.  The first two are
 * documented in the comments in fgsrot.c and gsrot.c.
 *
 * We're spending a lot of time in overhead.  Instead of doing two or more
 * complete passes through input lines to form an output line, say we do one
 * pass.  In that pass, we follow a jagged path through the input file such
 * that the coordinates of the input pixels in the output's frame of reference
 * stay immediately above the output line being calculated.  At each input
 * pixel position, we read and distribute not only the given input pixel, but
 * also the one immediately below it.
 *
 * Here's another way to look at it.  In the old scheme, we made as many
 * pure horizontal input sweeps as necessary to cover every output pixel with
 * a pair of vertically adjacent input pixels.  In this scheme, we do one
 * double-height sweep that jogs as necessary to keep it over the
 * output pixels.  The only way to make this work reasonably is to assume
 * the same xIOCoord for both upper and lower pixels - handling all the
 * conditions when the upper and lower pixels are in different output columns
 * gives me a headache.  This means we lose a little resolution - say 1/4 of
 * a pixel at 10 degrees, but tough.
 *
 * Without this simplification, there are two levels of loops - the outer loop
 * advances one iteration for each 8-pixel block of output pixels.  The inner
 * loop sweeps through as many input lines as necessary to calculate one block.
 * With the simplification, the inner loop is eliminated, though each step
 * takes longer.  What is the tradeoff here?
 *
 * The code in the old world that distributed the value of a pixel took 15
 * cycles.  We tended to execute a little over two of these 15 cycle blocks
 * of code for each output pixel, for an average of 34 cycles/pixel.  Overhead
 * required for the inner loop took about 20 cycles and was executed at least
 * once, often 2 or 3 times.  Say this adds 40 cycles/8 pixels for 5 cyc/pix.
 * This boosts the total to 39 cycles/pixel.
 *
 * In the new code, it takes about 20 cycles to distribute two pixels at a
 * given position.  We save on overhead by not jumping around so much.  There
 * is some overhead each time we move up or down by one input line, say this
 * happens 2 or 3 times/8 pixels.  The overhead is probably about 8 cycles,
 * or 16 cycles/8 pixels = 2 cyc/pix.  This brings the total to 22 cycles/pixel.
 * This is almost twice as fast as the old way.
 *
 * Since there is no inner loop, there's no switch statement required to
 * restart execution at random x offsets into the block.  Therefore, we don't
 * needs to spend so much time keeping xIOCoord up to date.  In fact, we
 * really only need the fraction bits to index into the table.  Ok, we also
 * need to know when adding sinRotAngle to xIOCoord crosses a pixel boundary
 * in the output world.  Huh?  Well, we normally traverse a horizontal path in
 * the input that's slightly diagonal in the output.  Since we want to stay
 * above the output line (the coordinates of the input pixels in the output's
 * frame of reference should be immediately above the output line), we detect
 * when the slanted path goes too low.  When we does, we move up one row in
 * the input world.  This corresponds to a vertical move by cosRotAngle
 * (approximated by 1 to avoid overlapping pixels) and a horizontal move by
 * sinRotAngle.  If adding sinRotAngle to xIOCoord crosses a pixel boundary in
 * the output world, we need to skip one output pixel position when
 * distributing input pixel values.
 * We'll just keep a -1 in the integer part of xIOCoord.
 * We can easily test to see if adding sinRotAngle forces xIOCoord to cross
 * an integer boundary by testing to see if the result is non-negative.  If we
 * cross an integer boundary, we restore the -1 in the integer, decrement
 * xCount, rSrcAddr and rSrcAddrBL, and skip one block of the inner loop.  What
 * are the extra decrements for?  Instead of incrementing the source address
 * and count at each block of inner loop, we use fixed address offsets.  For
 * example, at the second block of the loop, we address pixel rSrcAddr+2.  When
 * we skip one of these code blocks, we're also effectively skipping an address
 * offset.  In order to maintain our position in the input world, we decrement
 * the address pointers to compensate.
 *
 * The overhead encountered at the end of each block is also reduced because
 * we are already in at the correct input pixel at which to start processing
 * the next block.  The old scheme required pixel movement.
 *
 * There is a change in strategy, though.  Almost all output pixels contain
 * the coordinates of exactly one input pixel.  However, some output pixels
 * contain the coordinates of no input pixels - you can just barely miss at
 * all four corners.  Our scheme generally assumes that we start with the
 * input pixel whose output coords are in to the upper left of the leftmost
 * output pixel in the block.  Where do we start if no input pixel has those
 * coordinates?
 *
 * In the old scheme, we always started in the column to the left of the
 * output block, but could choose either the same row or the upper row (if the
 * same row wasn't reachable).  Here, we always want to start in the row above
 * the output row, either in the column to the left of the leftmost pixel or
 * the column of the first pixel if that isn't reachable.  Since the inner
 * loop code always reads two vertically adjacent pixels, it's important to
 * be in the row above the output line.
 *
 * One other modification - There was a special case to detect whether an
 * output block was entirely outside the image.  This did save some cycles
 * when it was used, totalling about 10 milliseconds for a 3 degree rotation
 * of a 8MB image.  However, there were an extra 6 opcodes to test for this
 * special case.  They were executed an 1,042,800 times for an addition
 * of 156 milliseconds.  Ooops.  The special case went away.
 *
 *
 * BUG FIX - Well, this is the first big post-rework bug.  As mentioned
 * above, a horizontal sweep across the input image is a diagonal sweep
 * (towards the lower right) across the output image.  When the sweep crosses
 * the boundary from the row above the output to the output's row, we need to
 * jump up one input line.  In doing so, we also move to the right by
 * sin(rotAngle) in the output's frame of reference.  So far, so good.  In
 * the new scheme however, we have been using the same xIOCoord for both the
 * upper and the lower input pixels.  This means the lower input pixel also
 * jumped to the right by sin(rotAngle).  This is a problem.  There will be
 * a discontinuity in the way that pixel's value is distributed among the
 * output pixels.  In effect, there will be a gap in the mapping of input
 * pixels to output pixels, leaving a couple of output pixels short changed.
 * The effect is visible as slightly darker, nearly vertical, lines down the
 * rotated image.
 *
 * This is a fairly ugly problem.  One approach is to use different xIOCoords
 * for the upper and lower samples.  This is fine except when they lie in
 * different output pixels.  To handle that, we'd have to shift the results
 * of the table lookups by to different output bytes.  This either requires
 * a messy test and branch in the inner loop (YUCK) or another set of checking
 * and non-checking loops with special shift cases.  Also yuck.
 *
 * Another approach is to try to fill the "gap" when we add do the little
 * horizontal jump.  This is a bit messy since we don't have a lookup table
 * that tells us how much of a pixel fits the gap.  It's also pretty slow to
 * do this gap filling every time we jump up a line in the input space.
 *
 * We`ll try a compromise.  If adding sin(rotAngle) to xIOCoord does not cross
 * into a new output column, we'll effectively use the old (does not reflect
 * the sin(rotAngle) jump) value of xIOCoord for the bottom sample.  We don't
 * actually keep two different xIOCoords, we just transfer the 3 xIOCoord
 * fraction bits from rTopLineTab to rBotLineTab.  This means the index into
 * the lookup table (which is really the only thing that matters here) for the
 * bottom input pixel will reflect the value of the un-shifted xIOCoord.
 *
 * What happens when adding sin(rotAngle) to xIOCoord causes it to cross into a
 * new pixel column?  That's when we do advance the bottom coordinate and fill
 * the gap.  How wide is the gap?  How much to we fill in?  Remember that
 * since we're jumping from one output pixel to the next, we need to add
 * something to *both* of them.  That's because the lookup tables assume that
 * an xIOCoord offset of 7 really means an offset of 15/16.  An offset of 0
 * really means an offset of 1/16.  Therefore, if we jump from 7 to 0, we
 * would need to add 1/16 of the pixel value to both the current output pixel
 * (to bring it up from 15/16 to 16/16) and the next output pixel (to bring
 * it from 0/16 to 1/16 where the next input pixel will start).
 *
 * But, we don't always start from an xIOCoord offset of 7 and, if we wanted
 * to be aligned with the top xIOCoord, we wouldn't always jump to an xIOCoord
 * offset of 0.  At the larger angles, xIOCoord can go from 6 to 0 or 7 to 1.
 * It is reasonably important to do the right thing so as to avoid the very
 * dark streaks we're trying to avoid.
 *
 * To make things deterministic, we'll always jump from wherever we would have
 * been normally (xIOCoord offset of 6 or 7) to offset 0.  Since the bottom
 * xIOCoord often trails the upper one anyway, it's not vital that they're
 * aligned after the jump.  We can get the amount of pixel to be added to
 * fill the gap in the output pixel we're leaving by simply doing a table
 * lookup using the rBotLineTab we would have used anyway with the proper
 * input pixel.  The leftmost byte of the two returned says how much of that
 * pixel lies in the current output pixel.  We get the amount of the input
 * pixel needed to fill the gap in the output pixel we're jumping to by simply
 * dividing the input pixel value by 16.  This is absolutely correct if the
 * yIOCoord offset was 0, which is almost always true.  If the yIOCoord offset
 * was 1, we'll be off by 1/256 of the pixel's value - not a big deal.
 *
 * In summary - each time the input jumps up a line, we transfer the old
 * xIOCoord fraction bits from rTopLineTab to rBotLineTab.  This is cheap and
 * fast.  If, in addition, the jog caused us to cross into a new output pixel
 * column, rBotLineTab is used to fetch the amount of the next pixel that will
 * fill the gap in the current output pixel, the gap in the next output pixel
 * is filled with 1/16th of the input pix's value and we advance the xIOCoord
 * fraction bits in rBotLineTab to 0.  You can think of this as skipping a
 * processing block, but filling in the gap created caused by skipping that
 * block.  This means we read the same pixel that block would have read and
 * apply the results to the output pixels that block would have affected.
 *
 * BUG FIX II - That was all lovely.  It went fast and didn't have ugly black
 * lines through the image.  Unfortunately, it had ugly faint gray bars though
 * the image in the same places.  My hypothesis is that these bars result from
 * small gaps in the mapping of input pixels to output pixels.  Look at this as
 * trying to tile the output image with diagonally-offset input pixels.  When
 * we jump up an input row, we have been jumping over a little in the output.
 * In the fix above, we closed the big gap in this jump by having one of the
 * lookup table pointer stay behind.  However, the upper pointer still jumped,
 * so there was still a small gap.
 *
 * So, what happens if we keep xIOCoord up to date on these vertical jogs, but
 * *DO NOT* update the X Fraction portion of either of the table pointers?
 * If there's a vertical jump, we may be as much as 1/4 of a pixel off?  How
 * often do we get vertical jumps?  Well, sin(10 degrees) = 0.17.  1/sin(10)
 * = 5.76, so we expect to cross to a new row of output bits every 5.76
 * horizontal input moves.  Since 5.76 is safely greater than 4, we'll take no
 * more than 1 vertical jump per 8 byte output block.  Therefore, we'll never
 * be more than 1/4 of a pixel off.  Seems ok to me.
 *
 * To make this work, we do calculate a new pair of table lookup indices at
 * the beginning of each block.  If xIOCoord has gone positive (the last
 * vertical input jump caused us to cross into a new output column), we back
 * up one pixel position.  If the backup causes us to move up too high, we
 * move down one row.
 *
 ***************************** ANISOTROPIC REWORK ***********************
 *
 * The glorious new way to make gray rotation of anisotropic images work
 * is to have two versions of sinRotAngle.  vertSinRotAngle is used to
 * move adjust the X coord when moving vertically and horizSinRotAngle is
 * used to adjust the Y coord when moving horizontally.  Basically,
 * vertSinRotAngle = (yRes/xRes) * sinRotAngle and horizSinRotAngle =
 * (xRes/yRes) * sinRotAngle.
 *
 * To make this work, we need another parameter.  rSinRotAngle will now
 * contain horizSinRotAngle (since that's used most often) and rXVInc
 * will contain vertSinRotAngle (since that's what rXVInc is really used
 * for anyway).
 *
 ************************ HIGH ASPECT-RATIO IMAGE REWORK *******************
 *
 * The code was originally written assuming that the diagonal paths
 * traced could go from in the image to above the image, or from inside
 * the image to inside the image, or from below the image to inside
 * the image.  It also assumed that you could compute at least 8 output
 * pixels horizontally without going through both the left and right source
 * image boundaries.  These assumptions are clearly not true.
 * The code used what I thought was a cute trick of setting xCount and
 * yCount to negative numbers outside the image and positive numbers
 * inside.  xCount is always incremented and yCount is always decremented
 * as you move across the image.  Making them negative on the left, right
 * top and bottom of the image required fooling around with the overflow
 * and stuff.  You also had to change the definition of the xCount and
 * yCount vars at the left and right edges (from a simple pos/neg number
 * at the left edge to some number subtracted from 0x80000000 at the
 * right edge).  Well, there's no time to change the definition in very
 * slim images.
 *
 * Our new scheme has us just setting xCount to the x coord in the input
 * image and yCount to the y coord in the input image.  We do an unsigned
 * version of rXCount >= rWidth to see if we're off the image.  If rXCount
 * is negative, the comparison is true since all negative numbers are bigger
 * than all positive numbers when doing unsigned compares.  Similarly, if
 * xCount is off the image to the right, it's bigger than rWidth.  The
 * same type of thing happens with yCount.  The effect of this is that
 * we have to do two comparisons instead of just ORing xCount and yCount
 * together, but tough.  This only happens in the version of the code that's
 * executed around the edges of the image anyway.
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssgsrot.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:19 $"
	.align	4


! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPStateIn	%i0
#define	rSrcAddrIn	%i1
#define rSrcPitchIn	%i2
#define rDstAddrIn	%i3
#define rWidthIn	%i4
#define rXVIncIn	%i5

#define	rSrcAddr	%i0		/* pointer to source image */
#define rSrcPitch	%i1		/* source image bytes/line */
#define rDstAddr	%i2		/* pointer to dest image */
#define rWidth		%i3		/* width of frame in bytes */
#define rXVInc		%i4		/* amount by which xIOCoord is changed
					 * for vertical movement */
#define rSinRotAngle	%i5		/* split integer, sin(rotAngle) */
#define rXIOCoord	%l1		/* Input's x coordinate in output's
					 * frame of reference */
#define rYIOCoord	%g1		/* Input's y coordinate in output's
					 * frame of reference */
#define rXCount		%g2		/* pseudo x coordinate in input
					 * frame of reference.  Used to detect
					 * edge of input pixrect */
#define rYCount		%g3		/* ditto for Y coord. */

#define	rGray0		%g4		/* holds Left 4 output bytes */
#define	rGray1		%g5		/* holds Right 4 output bytes */

#define rGrayRotTab	%o0		/* ptr to gray rotate table */
#define rTopLineTab	%o1		/* gray rotate tab ptr for top line */
#define rBotLineTab	%o2		/* ditto for bottom line */
#define rSrcAddrBL	%o3		/* source address of bottom line */
#define rCol		%o4		/* current col in dest. */

#define rXMask		%o5		/* selects 3 xIOCoord fraction bits */
#define rXInc		%l0		/* Increments int part of xIOCoord */
#define rYFrac		%l3		/* holds fractional y bits */
#define rXCntTemp	%l4		/* holds xCount+n in checking loop */

#define rScr0		%l5		/* scratch reg */
#define rScr1		%l6

#define rSrcHeight	%l2		/* holds height of source image */

#define	rRetAddr	%o7		/* defined by SPARC's call inst */

/* Boilerplate constants */
#define	ipStateOffset		0x44
#define	srcAddrOffset		0x48
#define	srcPitchOffset		0x4c
#define	dstAddrOffset		0x50
#define	widthOffset		0x54
#define	vertSinRotAngOffset	0x58
#define horizSinRotAngOffset	0x5c
#define	xIOCoordOffset		0x60
#define	yIOCoordOffset		0x64
#define	xCountOffset		0x68
#define	yCountOffset		0x6c
#define	pGrayRotTabOffset	0x70
#define srcHeightOffset		0x74

! Just because it's important in what's coming up, let's set the
! format of the table lookup index:
! |   13   12   11    |  10     9      8  | 7 6 5 4 3 2 1 | 0 |
! | 3b xIOCoord (fff) | 3b yIOCoord (fff) |  7b pixel val | 0 |
!
! The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
! yIOCoord bits are its 3 MS fraction bits.  The 7 pixel bits are
! the 7 MS pixel bits.
! The 0 at the end makes this an index into a table of short ints.

#define XF_MASK		0x3800		/* selects 3 MS xIOCoord frac bits */
#define XI_INC		0x4000		/* used to increment xIOCoord */
#define XI_POS		14		/* position of LS int bit in xIOCoord */
#define YF_MASK		0x700		/* selects 7 fraction bits */
#define YI_INC		0x800		/* used to increment yIOCoord */
#define YI_POS		11		/* position of LS int bit in yIOCoord */
#define MAX_Y_FRAC	0x7FF		/* maximum value of yIOCoord fraction */
#define PIX_MASK	0xFE		/* mask for raw pixel values */

#define	WHITE_VAL	0xFAFAFAFA	/* a word of "white" pixels.  These are
					 * chosen to be consistent with
					 * computed values of "white" */

/* magical C Frame constant (no local variables) */
#define	LF26		-64

LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! _ip_rotateGrayLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_rotateGrayLine)
MK_GLOBAL_NAME(ip_rotateGrayLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the args that didn't fit in registers here.
	st	rIPStateIn,     [%fp+ipStateOffset]
	st	rSrcAddrIn,     [%fp+srcAddrOffset]
	st	rSrcPitchIn,    [%fp+srcPitchOffset]
	st	rDstAddrIn,     [%fp+dstAddrOffset]
	st	rWidthIn,	[%fp+widthOffset]
	st	rXVIncIn,	[%fp+vertSinRotAngOffset]

	ld	[%fp+srcAddrOffset],     rSrcAddr
	ld	[%fp+srcPitchOffset],    rSrcPitch
	ld	[%fp+dstAddrOffset],     rDstAddr
	ld	[%fp+widthOffset],	 rWidth
	ld	[%fp+vertSinRotAngOffset], rXVInc
	ld	[%fp+horizSinRotAngOffset], rSinRotAngle
	ld	[%fp+xIOCoordOffset],    rXIOCoord   
	ld	[%fp+yIOCoordOffset],	 rYIOCoord
	ld	[%fp+xCountOffset],	 rXCount
	ld	[%fp+yCountOffset],	 rYCount
	ld	[%fp+pGrayRotTabOffset], rGrayRotTab
	ld	[%fp+srcHeightOffset],	 rSrcHeight

! We'd like to have the three MS fraction bits in xIOC aligned with
! their position in the table lookup.
	sra	rXIOCoord, 2, rXIOCoord
	sra	rXVInc, 2, rXVInc		! when we move vertically in
						! the input coord system, we
						! move horizontally by this much
						! in the output coord system
	sethi	%hi(XF_MASK), rXMask		! used to select xIOCoord
	or	rXMask, %lo(XF_MASK), rXMask	! fraction bits
	sethi	%hi(XI_INC), rXInc		! used to increment integer part
	or	rXInc, %lo(XI_INC), rXInc	! of xIOCoord

! Ditto for yIOCoord.  This leaves us with 11 bits of fraction.
! Knowing the Y position to within 1 part in 2048 is better accuracy
! than the skew detection code will return
	sra	rYIOCoord, 5, rYIOCoord

! The sinRotAngle quantity is added to yIOCoord in the inner loop.
! Align it with yIOCoord to make this easier.
	sra	rSinRotAngle, 5, rSinRotAngle

! we initialize the regs making up the gray box to 0 since no
! input pixels have been inserted yet.
	mov	0, rGray0
	mov	0, rGray1

! Initialize the pointer to the next input line.
	add	rSrcAddr, rSrcPitch, rSrcAddrBL

! Looks like we're going to have to start writing real inner loop
! code pretty soon.  Hard to put it off much longer.  But wait!
! We can still discuss how many flavors of loop we need.
! There are two types - one checks xCount and yCount on
! every read from the input image because we're near a boundary, and
! one is used when we're totally inside the input image.

	mov	0, rCol			! start at dest column 0

rGRowLoop:

! Here's where we check to see if only part of the block lies outside the
! input image.  If so, we'll have to look at each input pixel individually.
! We test to see if we're begining to the left/right of the image, beginning
! above/below the image or ending above/below the image.
! The conditions for entering the checking loop are:
! 1. yCount - 3 is negative (might rise up above image in this block)
! 2. yCount + 2 is greater than or equal hS (normally we go down 1,
!		if we backup 1, we might go down 2)
! 3. xCount - 1 is negative (starting to the left of image when potential
!		backup is taken into account)
! 4. xCount + 10 is greater than wS (sample could end to right of image)
rGRowTestPartOutside:
	and	rXIOCoord, rXMask, rScr1 ! get ptr to grayRotation table
	add	rYCount, 2, rScr0
	cmp	rScr0, rSrcHeight	! is rYCount+2 >= hS?
	bgeu	rGRowPartial		! yes, need to check as we go across

	add	rYCount, -3, rScr0
	cmp	rScr0, rSrcHeight	! is rYCount-3 >= hS?
	bgeu	rGRowPartial		! yes, need to check as we go across

	add	rXCount, -1, rScr0
	cmp	rScr0, rWidth		! is xCount-1 >= wS?
	bgeu	rGRowPartial		! yes, need to check as we go across

	add	rXCount, 10, rScr0
	cmp	rScr0, rWidth		! is xCount+10 >= wS?
	blu	rGRowNoCheckBlock	! no, do the loop that doesn't look
					! for input image boundaries
	nop

! At least part of the input is outside the input image

! There are two types of inner-inner loop.  One starts in
! column 0 and one starts in column -1.  Both types end their
! sweeps when they reach column 7.  If start from column 0,
! remember to skip the first block of the loop.
!
! The Y portion of rTopLineTab should be inverted so it references
! the portion of the pixel that lies in the *next* line.  The
! Y portion of rBotLineTab should not be inverted so it get the
! portion of the input pixel that lies in the current line.
! We implement the inversion by adding the maximum Y coord (7)
! to rTopLineTab and subtracting each Y coord from this pointer.

rGRowPartial:
	add	rGrayRotTab, rScr1, rBotLineTab ! get ptr to portion of gray
						! rotation table that will be
						! used for this xIOCoord.
	add	rBotLineTab, YF_MASK, rTopLineTab

! Here's where we jump to the correct block of the checking loop.  The
! block is based on the integer part of xIOCoord since each block calculates
! a byte.
	tst	rXIOCoord
	bneg	rGRowPartM1			! rXIOCoord is neg (the normal
						! case) so start at a negative
						! output offset
	cmp	rXCount, rWidth			! off image on left/right?

! rXIOCoord is positive, so we must be column 0.  Backup one column using our
! usual approximations.
	dec	rXCount
	dec	rSrcAddr
	dec	rSrcAddrBL
	sub	rXIOCoord, rXInc, rXIOCoord
	sub	rYIOCoord, rSinRotAngle, rYIOCoord
! Did that take us above row -1?
	addcc	rYIOCoord, YI_INC, %g0		! does adding 1 to integer make
						! result positive?
	bpos	rGRowPartM1			! yes, just start now
	cmp	rXCount, rWidth			! off image on left/right?

! We went up too high, so move down one input line.
	add	rSrcAddr, rSrcPitch, rSrcAddr
	add	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	inc	rYCount				! increment rYCount
	add	rYIOCoord, YI_INC, rYIOCoord
	sub	rXIOCoord, rXVInc, rXIOCoord
	cmp	rXCount, rWidth			! off image on left/right?

!--------------------------------
! We're calculating pixel -1 relative to the output block.  Check to see if
! we're on the input image or not.  Note we've compared rXCount and rWidth.
rGRowPartM1:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	bgeu	rGRowPartM1TopIsOff
	cmp	rYCount, rSrcHeight		! off top/bottom of image?
	blu	rGRowPartM1FetTopPix		! nope.
! This is here because Purify is whines if it sees a branch to the opcode
! in the delay slot of a branch.  AAAARGH!
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPartM1TopIsOff:
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPartM1FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPartM1FetTopPix:
	ldub	[rSrcAddr+0], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPartM1FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	cmp	rXCount, rWidth			! prepare to test lower pixel
	sll	rScr0, 24, rScr0		! align gray value
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot
	bgeu	rGRowPartM1BotIsOff		! off image on left/right
	add	rYCount, 1, rScr1		! bottom pix off top/bottom?
	cmp	rScr1, rSrcHeight
	blu	rGRowPartM1FetBotPix		! we're on image
	nop					! Pure Idiots!


! bottom pixel is off the image
rGRowPartM1BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPartM1FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPartM1FetBotPix:
	ldub	[rSrcAddrBL+0], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPartM1FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 24, rScr0		! align gray value
	bneg	rGRowPart0			! entered row 0?
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount

!--------------------------------
rGRowPart0:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	add	rXCount, 1, rXCntTemp
	cmp	rXCntTemp, rWidth		! off image to left/right?
	bgeu	rGRowPart0TopIsOff		! yes
	cmp	rYCount, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart0FetTopPix		! nope
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPart0TopIsOff:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPart0FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPart0FetTopPix:
	ldub	[rSrcAddr+1], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart0FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	add	rYCount, 1, rScr1		! prepare to test lower pixel
	sll	rScr0, 16, rScr0		! align gray value
	cmp	rXCntTemp, rWidth		! off image on left/right?
	add	rGray0, rScr0, rGray0		! apply gray val
	bgeu	rGRowPart0BotIsOff		! yes
	cmp	rScr1, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart0FetBotPix		! no
	nop					! Pure Idiots!


! bottom pixel is off the image
rGRowPart0BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPart0FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPart0FetBotPix:
	ldub	[rSrcAddrBL+1], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart0FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 16, rScr0		! align gray value
	bneg	rGRowPart1			! entered row 0?
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowPart1:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	add	rXCount, 2, rXCntTemp
	cmp	rXCntTemp, rWidth		! off image to left/right?
	bgeu	rGRowPart1TopIsOff		! yes
	cmp	rYCount, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart1FetTopPix		! nope
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPart1TopIsOff:
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPart1FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPart1FetTopPix:
	ldub	[rSrcAddr+2], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart1FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	add	rYCount, 1, rScr1		! prepare to test lower pixel
	sll	rScr0, 8, rScr0			! align gray value
	cmp	rXCntTemp, rWidth		! off image on left/right?
	add	rGray0, rScr0, rGray0		! apply gray val
	bgeu	rGRowPart1BotIsOff		! yes
	cmp	rScr1, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart1FetBotPix		! no
	nop					! Pure Idiots!

! bottom pixel is off the image
rGRowPart1BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPart1FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPart1FetBotPix:
	ldub	[rSrcAddrBL+2], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart1FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 8, rScr0			! align gray value
	bneg	rGRowPart2			! entered row 0?
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowPart2:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	add	rXCount, 3, rXCntTemp
	cmp	rXCntTemp, rWidth		! off image to left/right?
	bgeu	rGRowPart2TopIsOff		! yes
	cmp	rYCount, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart2FetTopPix		! nope
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPart2TopIsOff:
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPart2FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPart2FetTopPix:
	ldub	[rSrcAddr+3], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart2FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	add	rYCount, 1, rScr1		! prepare to test lower pixel
!	sll	rScr0, 0, rScr0			! align gray value
	cmp	rXCntTemp, rWidth		! off image on left/right?
	add	rGray0, rScr0, rGray0		! apply gray val
	bgeu	rGRowPart2BotIsOff		! yes
	cmp	rScr1, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart2FetBotPix		! no
	nop					! Pure Idiots!

! bottom pixel is off the image
rGRowPart2BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPart2FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPart2FetBotPix:
	ldub	[rSrcAddrBL+3], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart2FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
!	sll	rScr0, 0, rScr0			! align gray value
	bneg	rGRowPart3			! entered row 0?
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowPart3:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	add	rXCount, 4, rXCntTemp
	cmp	rXCntTemp, rWidth		! off image to left/right?
	bgeu	rGRowPart3TopIsOff		! yes
	cmp	rYCount, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart3FetTopPix		! nope
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPart3TopIsOff:
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPart3FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPart3FetTopPix:
	ldub	[rSrcAddr+4], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart3FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	add	rYCount, 1, rScr1		! prepare to test lower pixel
	cmp	rScr1, rSrcHeight		! off image on top/bottom?
	srl	rScr0, 8, rScr1			! align gray value
	add	rGray0, rScr1, rGray0		! apply gray val in delay slot
	sll	rScr0, 24, rScr0		! align gray value
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot
	bgeu	rGRowPart3BotIsOff		! yes, we're off img on top/bot
	cmp	rXCntTemp, rWidth		! off image on left/right?
	blu	rGRowPart3FetBotPix		! no
	nop					! Pure Idiots!

! bottom pixel is off the image
rGRowPart3BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPart3FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPart3FetBotPix:
	ldub	[rSrcAddrBL+4], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart3FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	srl	rScr0, 8, rScr1			! align gray value
	add	rGray0, rScr1, rGray0		! apply gray val in delay slot
	sll	rScr0, 24, rScr0		! align gray value
	bneg	rGRowPart4			! entered row 0?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowPart4:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	add	rXCount, 5, rXCntTemp
	cmp	rXCntTemp, rWidth		! off image to left/right?
	bgeu	rGRowPart4TopIsOff		! yes
	cmp	rYCount, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart4FetTopPix		! nope
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPart4TopIsOff:
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPart4FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPart4FetTopPix:
	ldub	[rSrcAddr+5], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart4FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	add	rYCount, 1, rScr1		! prepare to test lower pixel
	sll	rScr0, 16, rScr0		! align gray value
	cmp	rXCntTemp, rWidth		! off image on left/right?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot
	bgeu	rGRowPart4BotIsOff		! yes
	cmp	rScr1, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart4FetBotPix		! no
	nop					! Pure Idiots!

! bottom pixel is off the image
rGRowPart4BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPart4FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPart4FetBotPix:
	ldub	[rSrcAddrBL+5], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart4FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 16, rScr0		! align gray value
	bneg	rGRowPart5			! entered row 0?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowPart5:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	add	rXCount, 6, rXCntTemp
	cmp	rXCntTemp, rWidth		! off image to left/right?
	bgeu	rGRowPart5TopIsOff		! yes
	cmp	rYCount, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart5FetTopPix		! nope
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPart5TopIsOff:
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPart5FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPart5FetTopPix:
	ldub	[rSrcAddr+6], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart5FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	add	rYCount, 1, rScr1		! prepare to test lower pixel
	sll	rScr0, 8, rScr0			! align gray value
	cmp	rXCntTemp, rWidth		! off image on left/right?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot
	bgeu	rGRowPart5BotIsOff		! yes
	cmp	rScr1, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart5FetBotPix		! no
	nop					! Pure Idiots!

! bottom pixel is off the image
rGRowPart5BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPart5FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPart5FetBotPix:
	ldub	[rSrcAddrBL+6], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart5FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 8, rScr0			! align gray value
	bneg	rGRowPart6			! entered row 0?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowPart6:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	add	rXCount, 7, rXCntTemp
	cmp	rXCntTemp, rWidth		! off image to left/right?
	bgeu	rGRowPart6TopIsOff		! yes
	cmp	rYCount, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart6FetTopPix		! nope
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPart6TopIsOff:
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPart6FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPart6FetTopPix:
	ldub	[rSrcAddr+7], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart6FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	add	rYCount, 1, rScr1		! prepare to test lower pixel
!	sll	rScr0, 0, rScr0			! align gray value
	cmp	rXCntTemp, rWidth		! off image on left/right?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot
	bgeu	rGRowPart6BotIsOff		! yes
	cmp	rScr1, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart6FetBotPix		! no
	nop					! Pure Idiots!

! bottom pixel is off the image
rGRowPart6BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPart6FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPart6FetBotPix:
	ldub	[rSrcAddrBL+7], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart6FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
!	sll	rScr0, 0, rScr0			! align gray value
	bneg	rGRowPart7			! entered row 0?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowPart7:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	add	rXCount, 8, rXCntTemp
	cmp	rXCntTemp, rWidth		! off image to left/right?
	bgeu	rGRowPart7TopIsOff		! yes
	cmp	rYCount, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart7FetTopPix		! nope
	nop					! Pure Idiots!

! top pixel is off the image
rGRowPart7TopIsOff:
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	b	rGRowPart7FetTopTab
	mov	PIX_MASK, rScr1			! use max white value
! top pixel is on the image
rGRowPart7FetTopPix:
	ldub	[rSrcAddr+8], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart7FetTopTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
	add	rYCount, 1, rScr1		! prepare to test lower pixel
	srl	rScr0, 8, rScr0			! align gray value
	cmp	rXCntTemp, rWidth		! off image on left/right?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot
	bgeu	rGRowPart7BotIsOff		! yes
	cmp	rScr1, rSrcHeight		! off image on top/bottom?
	blu	rGRowPart7FetBotPix		! no
	nop					! Pure Idiots!

! bottom pixel is off the image
rGRowPart7BotIsOff:
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	b	rGRowPart7FetBotTab
	mov	PIX_MASK, rScr1			! use max white value
! bottom pixel is on the image
rGRowPart7FetBotPix:
	ldub	[rSrcAddrBL+8], rScr1		! get bottom source pixel
	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
rGRowPart7FetBotTab:
	lduh	[rScr0+rScr1], rScr0		! get gray value
						! INTERLOCK
	srl	rScr0, 8, rScr0			! align gray value
	b	rGRowWrite
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! Note that we didn't try to advance the pointers.  That's because we're
! already in the right position to start the next sweep.



!-------------------- End of Checking Case -----------------------



! This inner loop is very much like the checking loops except that we
! don't check to see if we're off the image.  We just read the image byte
! and use it.  This is the set of code that should be executed almost all of
! the time.

rGRowNoCheckBlock:
	add	rGrayRotTab, rScr1, rBotLineTab ! get ptr to portion of gray
						! rotation table that will be
						! used for this xIOCoord.
	add	rBotLineTab, YF_MASK, rTopLineTab

! Here's where we jump to the correct block of the checking loop.  The
! block is based on the integer part of xIOCoord since each block calculates
! a byte.
	tst	rXIOCoord
	bneg	rGRowAllM1			! rXIOCoord is neg (the normal
						! case) so start at a negative
						! output offset
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits

! rXIOCoord is positive, so we must be column 0.  Backup one column using our
! usual approximations.
	dec	rXCount
	dec	rSrcAddr
	dec	rSrcAddrBL
	sub	rXIOCoord, rXInc, rXIOCoord
	sub	rYIOCoord, rSinRotAngle, rYIOCoord
! Did that take us above row -1?
	addcc	rYIOCoord, YI_INC, %g0		! does adding 1 to integer make
						! result positive?
	bpos	rGRowPartM1			! yes, just start now
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits

! We went up too high, so move down one input line.
	add	rSrcAddr, rSrcPitch, rSrcAddr
	add	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	inc	rYCount
	add	rYIOCoord, YI_INC, rYIOCoord
	sub	rXIOCoord, rXVInc, rXIOCoord
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits

!--------------------------------
! We're calculating pixel -1 relative to the output block.  Check to see if
! we're on the input image or not.  Note we've already ORed rXCount and rYCount.
rGRowAllM1:
	ldub	[rSrcAddr+0], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+0], rScr1		! get bottom source pixel
	sll	rScr0, 24, rScr0		! align gray value
	add	rGray0, rScr0, rGray0		! apply gray val

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 24, rScr0		! align gray value
	bneg	rGRowAll0			! entered row 0?
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount

!--------------------------------
rGRowAll0:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	ldub	[rSrcAddr+1], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+1], rScr1		! get bottom source pixel
	sll	rScr0, 16, rScr0		! align gray value
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 16, rScr0		! align gray value
	bneg	rGRowAll1			! entered row 0?
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowAll1:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	ldub	[rSrcAddr+2], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+2], rScr1		! get bottom source pixel
	sll	rScr0, 8, rScr0			! align gray value
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 8, rScr0			! align gray value
	bneg	rGRowAll2			! entered row 0?
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount

!--------------------------------
rGRowAll2:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	ldub	[rSrcAddr+3], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+3], rScr1		! get bottom source pixel
!	sll	rScr0, 0, rScr0			! align gray value
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
!	sll	rScr0, 0, rScr0			! align gray value
	bneg	rGRowAll3			! entered row 0?
	add	rGray0, rScr0, rGray0		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowAll3:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	ldub	[rSrcAddr+4], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+4], rScr1		! get bottom source pixel
	srl	rScr0, 8, rXCntTemp		! align gray value in temp reg
	add	rGray0, rXCntTemp, rGray0	! apply gray val in delay slot
	sll	rScr0, 24, rScr0		! align gray value
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	srl	rScr0, 8, rScr1			! align gray value
	add	rGray0, rScr1, rGray0		! apply gray val in delay slot
	sll	rScr0, 24, rScr0		! align gray value
	bneg	rGRowAll4			! entered row 0?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowAll4:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	ldub	[rSrcAddr+5], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+5], rScr1		! get bottom source pixel
	sll	rScr0, 16, rScr0		! align gray value
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 16, rScr0		! align gray value
	bneg	rGRowAll5			! entered row 0?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowAll5:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	ldub	[rSrcAddr+6], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+6], rScr1		! get bottom source pixel
	sll	rScr0, 8, rScr0			! align gray value
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
	sll	rScr0, 8, rScr0			! align gray value
	bneg	rGRowAll6			! entered row 0?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowAll6:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	ldub	[rSrcAddr+7], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+7], rScr1		! get bottom source pixel
!	sll	rScr0, 0, rScr0			! align gray value
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	addcc	rYIOCoord, rSinRotAngle, rYIOCoord ! fill interlock
!	sll	rScr0, 0, rScr0			! align gray value
	bneg	rGRowAll7			! entered row 0?
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! entered row 0.  Move up one input row.
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sub	rSrcAddrBL, rSrcPitch, rSrcAddrBL
	sub	rYIOCoord, YI_INC, rYIOCoord	! adjust yIOCoord for move up
	add	rXIOCoord, rXVInc, rXIOCoord	! adjust xIOCoord
	dec	rYCount


!--------------------------------
rGRowAll7:
	and	rYIOCoord, YF_MASK, rYFrac	! select yIOCoord frac bits
	ldub	[rSrcAddr+8], rScr1		! get top source pixel
	sub	rTopLineTab, rYFrac, rScr0	! get top line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
	ldub	[rSrcAddrBL+8], rScr1		! get bottom source pixel
	srl	rScr0, 8, rScr0			! align gray value
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

	add	rBotLineTab, rYFrac, rScr0	! get bottom line's table ptr
	and	rScr1, PIX_MASK, rScr1		! mask off LS bit(s?)
	lduh	[rScr0+rScr1], rScr0		! get gray value
						! INTERLOCK
	srl	rScr0, 8, rScr0			! align gray value
!	b	rGRowWrite
	add	rGray1, rScr0, rGray1		! apply gray val in delay slot

! Note that we didn't try to advance the pointers.  That's because we're
! already in the right position to start the next sweep.



!-------------------- End of No Checking Case -----------------------



! We've completed formation of this row.  Write it out.  Note that we always
! write the rGray0.  If the line ends after that, don't write rGray1.
rGRowWrite:
	st	rGray0, [rDstAddr+rCol]
	add	rCol, 4, rCol			! at the end of the line?
	cmp	rCol, rWidth
	bl,a	rGRowNextBlock
	st	rGray1, [rDstAddr+rCol]		! only if still in line
rGRowNextBlock:
	add	rCol, 4, rCol

! ARE WE ALL DONE WITH THE ROW??
	cmp	rCol, rWidth
	bge	rGLineFin			! yes, just go quit
! we hardly ever quit and when we do, it doesn't matter what's in rGray0, so
! we might as well clear it whether we quit or not.
	mov	0, rGray0			! clear out gray regs for
	mov	0, rGray1			! new block

! rSrcAddr, rScrAddrBL and xCount are all pointing to the beginning of the
! line.  Update them here all at once.
	add	rSrcAddr, 8, rSrcAddr
	add	rSrcAddrBL, 8, rSrcAddrBL
	b	rGRowLoop
	add	rXCount, 8, rXCount

! END OF THIS ROW

! We get here from a branch taken immediately after the last word of the row
! is written.

! We've put out the right number of words, so quit
rGLineFin:
	mov	0,%i0
	ret
	restore


/*

$Log: ssgsrot.s,v $
Revision 1.1.1.1  1999/09/20 18:18:19  jlin
Imported sources

Revision 1.8  1998/01/13 23:57:48  ddavies
ssgsrot.s -> /product/ipcore/ipshared/src/RCS/ssgsrot.s,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.7  1996/10/05 02:27:43  ddavies
ssgsrot.s -> /product/ipcore/ipshared/src/RCS/ssgsrot.s,v

Change the semantics of xCount and yCount.  Instead of trying to arrange
for them to be negative outside the image, we just to unsigned compares of
xCount >= width and yCount >= height.  If these are true, we're off the image.
Also get rid of the fixup of xCount at the left boundary.

Revision 1.6  1995/12/13 01:55:19  ddavies
ssgsrot.s -> /product/ipcore/ipshared/src/RCS/ssgsrot.s,v

In keeping with the new, resolution-independent, method representing skew
angles, use two versions of sinRotAngle - horizSinRotAngle is used to adjust
y coords when moving horizontally and vertSinRotAngle is used to adjust
x coords when moving vertically.  Also fix a bug in which xCount was
being adjusted for use to detect the right edge of the image when it was
equal to 0.  In this case, it may still be necessary to read at pixel with
x coord = -1, but we wouldn't substitute a white pixel since xCount would
have been changed to be relative to the right edge of the image instead of
being relative to the left edge.

! Revision 1.5  1994/07/27  16:02:41  ejaquith
! ssgsrot.s -> /product/ipcore/ipshared/src/RCS/ssgsrot.s,v
!
! Reentrancy!
! Removed globals and passed in gray rotation table pointers instead.
!
! Revision 1.4  1993/12/04  05:00:23  ddavies
! ssgsrot.s -> /product/ipcore/ipshared/src/RCS/ssgsrot.s,v
!
! Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
!
! Revision 1.3  1993/11/05  02:25:06  ddavies
! ssgsrot.s -> /product/ipcore/ipshared/src/RCS/ssgsrot.s,v
!
! Add RCS magic to have string specifying this file's name, rev number and
! last edit date into the object file.  Also add change log at the bottom.
!

*/

