/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: ssnirot.s
 *
 * This file contains assembly code to do small angle, non-interpolated
 * rotations.
 *
 * 		ip_rotateNoInterpLine()			produces one line of
 *							rotated image.  Works
 *							for any depth image
 *		ip_rotateAssemblyCodeIsPresent()	Since this is assembly
 *							code, returns 1.
 *
 *
 * Rotation  by small angles is tricky.  There's a huge comment in fnirot.c
 * that explains what we're doing and how we decide to do it.  It seems
 * like a waste of effort, but we'll copy that comment here.
 *
 ****************************************************************************
 *
 * People don't always put documents on copiers straight.  The ability
 * make these skewed images look nice by rotating them a few degrees is
 * thought to be a valuable feature.
 *
 * Rotation is an ugly operation.  We have to move all the pixels in the
 * image by distances that appear straightforward to image procesing
 * grad students, but arbitary to microcoders.  WE DON'T LIKE TRANCENDENTAL
 * FUNCTIONS!!  Heck, we don't even like mutiplication.
 *
 * Image processing experts are to assembly programmers as Newtonian
 * mechanics is to quantum mechanics.  There are no continuous functions
 * down here.
 *
 * The "rotateGray" family of routines attack this problem by doing
 * interpolated rotation of 8-bit gray images.  Each pixel is formed from
 * the weighted sum of the corresponding pixels in the input image.
 * Interpolation works well because there are lots of gray levels in both
 * input and output images.  That is, the gray image does a very good job of
 * representing a diagonal line as a group of pixels whose values vary
 * smoothly.  There are no input "jaggies."  Similarly, it makes perfect
 * sense to compose an output pixel from a weighted sum of continuous values,
 * since output pixels also take on gray values.
 *
 * While gray rotation gives very nice results (after working on it for a
 * month!), it's slow.  There's too much data and the operations are too
 * complex.  For performance reasons, we'd like to rotate images after they've
 * been converted to binary.
 *
 * Can we make these images look nice?  It's really hard.  First, consider the
 * problem of jaggies.  All binary images have them since diagonal lines must
 * be represented by a "staircase" of binary pixels.  One's first inclination
 * is to simply smooth the jaggies back into gray.  Well, if you round off
 * the corners of a staircase, you end up with a staircase with rounded corners;
 * not at all the same as a straight line.  What if you smooth things even
 * more?  The problem is that there are edges in the binary image that you
 * want to keep nice and sharp.  Very thin lines and characters should be
 * retained.  If the image was converted to binary using error diffusion, there
 * are lots of isolated dots and precise grids of dots that you can't afford
 * to lose or blur into nothing.  How do you tell the difference betweeen
 * nice clean sharp edges and ugly jaggies?
 *
 * Take this one step further.  Say a line which should be horizontal is
 * represented as a staircase because the image was skewed.  We're trying to
 * deskew the image, so will be following a similar staircase across the
 * input image to produce the output image.  The problem is that the staircase
 * we follow will inevitably be out of phase with the staircase in the drawing.
 * This is because staircases in the drawing have arbitrary phase with respect
 * to each other.  Any fixed phase we choose will be wrong.  Big deal, so
 * we're out of phase, so what?  This means that at the edge of the line,
 * we'll dip down into black pixels for awhile, that line of black pixels will
 * end so we'll go through some white pixels then we dip down into black
 * again and so on.  Any stateless method that computes output pixels 
 * based solely on input pixels in the neighborhood of the output 
 * (and is capable of producing both 0's and 1's) will see neighborhoods 
 * dominated by alternately by 0's and 1's, so will produce an output 
 * line with bumps along the edges.  This is the biggest, though not the 
 * only, problem we're trying to avoid.
 *
 * At this point, we could start messing around with algorithms that have
 * state, perhaps using some sort of hysteresis to smooth the bumps.  Neat.
 * We'll go into that in a later set of routines in different files.
 * For now, lets see if we can just do a plain rotation very quickly.
 * This type of rotation will just move each pixel to some other location 
 * without trying to smooth the bumps.  It will add a new set of jaggies.  
 * However, at sufficiently high resolution, this looks sort of ok.  
 * Anyway, it's the basis for the more complicated methods.
 *
 * This job is currently done in the Alpaca world by the procedure "rotateArb".
 * The procedure use two or three shears of the image to get the results.
 * Each shear writes the entire image using a bunch of separate bitblts.  We'd
 * like to get improved performance by avoiding shears and only writing the
 * output image once.
 *
 * How?  We'll form each output line by tracing a diagonal path across the
 * input image.  For example, if we're supposed to be rotating clockwise,
 * the input is a staircase that rises as it goes across the image from
 * left to right.
 *
 * The tricky part is assigning an exact position to each input bit.  The
 * most straightforward thing to do is to concatenate strings of bits read
 * from the input image.  For example, we might read n bits from line x,
 * another n bits from line x-1 and so on.  The problem with this scheme is
 * that we're assuming that a horizontal move of n bits in the input image
 * exactly equals a horizontal move of n bits in the output image.  This is
 * equivalent to assuming the the cosine of our small angle equals 1.
 *
 * Well, how bad is this assumption?  Say we're rotating by 5 degrees.
 * The cosine of 5 degrees is 0.996195.  Assume we're rotating an 11 x 17
 * image across the 17 inch dimension.  Tracing a diagonal path, we 
 * ought to see 17/0.996195 = 17.06 inches of image.  Instead, we see 
 *  exactly 17 inches since every horizontal bit in the source image 
 * equals one horizontal bit in the dest image.  So we're 0.06 inches, 
 * or 0.4% off.  Big Deal.  This seems like a sufficiently good approximation.
 * Putting the thing through a copier adds around 2% distortion anyway..
 *
 * For comparison, say we're scanning across 8.5 inches.  We'll be off by
 * 8.5/0.996195 - 8.5 = 0.03 inches.  Seems ok to me.  This buys several
 * very nice simplifications in the rotation algorithm as we'll see below..
 *
 * Ok, each output bit will be composed of the data from exactly one input
 * bit.  The input bits will be read in stairstep fashion from the input image
 * (read a run of bits, move up/down one line, read another run of bits, etc).
 * How many input bits should we read from any given scan line?
 *
 * Think of the input and output images as x,y grids - each grid square
 * containing one pixel.  Put the two grids exactly on top of each other.
 * Every output pixel will correspond to exactly one input pixel.  Put a pin
 * through the center of both grids.  Rotate the input by a few degrees
 * with respect to the output around the pin.  The grid square of any output
 * pixel will now intersect four input pixels.  Ideally, the value of the
 * output pixel is equal to the weighted sum of these four pixels.  The
 * weight for each input pixel equals the percentage of output pixel covered
 * by that input pixel (which equals the percentage of input pixel
 * covered by the output pixel).
 *
 * However, in the interests of speed, we will not be computing weighted sums.
 * The next most logical thing to do is to assign the output pixel the value
 * of the input pixel with the greatest overlap.  We'll do that.
 *
 * To do this exactly, we'd need to account for the fact that the individual
 * input pixels are tilted with respect to the output pixels.  The overlapped
 * areas have between 3 and 8 sides.  Yuck.  In our first approximation, we'll
 * pretend that the input grid is really a sheared version of the original
 * input.  That is, the individual input squares have the same orientation as
 * the output squares.  This means the input squares no longer form a
 * rectangular grid, but that's easier to handle, trust me. Think of an
 * arrangement where the upper left corner of each input pixel is in exactly
 * the right place, then we rotate all the input pixels individually around
 * their upper left corners so all the individual input pixels are aligned
 * with all the individual output pixels.
 *
 * In simplest version of the algorithm, we cruise across the output image
 * considering each output pixel.  For each one, we look at the four input
 * pixels, choosing the one with the greatest overlap.  If we're doing
 * clockwise rotation, moving over 1 pixel in output space means we're moving
 * over a little less than one pixel and up some fraction of a pixel in input
 * space.  Our cos(rotAngle) = 1 approximation says that we'll assume
 * moving over by 1 in output space corresponds to moving over by exactly 1
 * pixel in input space, though we still move up by sin(rotAngle).
 *
 * To make a faster algorithm, we don't really want to carefully inspect every
 * output pixel.  We really want to assign runs of input pixels to runs of
 * output pixels.  How are the run lengths computed?  For each x,y position
 * *within a pixel* at a given angle, we get back a run length.  Say the
 * input run slants from lower left to upper right relative to the output.
 * The top half of the input pixels at the left edge of the run overlap the
 * bottom half of the leftmost output pixels.  At the right edge of the run,
 * the bottom half of the input pixels overlap the top half of the output
 * pixels.

 * Given this view of things, run lengths could be found by indexing a
 * table with the fractional portions of the current (x,y) position.  At
 * this point in the narrative, we should have a grasp of the basic problem
 * and assumptions.  From here, we'll start exploring the general idea
 * of using position to determine run lengths.  We can make substantial
 * improvements in the algorithm.
 *
 * It would be nice if the table lookup also retured the next y coordinate.
 * This equals (initial y coord) + (runLength*sin(rotAngle)).  To what accuracy
 * do we keep the overall y coordinate?  Well, 10-12 fractional bits ought
 * to do it.  That will let us compute our trajectory across thousands of
 * pixels with good accuracy.  But, indexing a table with 10-12 y bits plus
 * a few x bits and retrieving 4 bytes each time quickly adds up to a lot
 * of table bits.  A big table produces a lot of costly cache misses.  Is
 * there some way we can cheat?
 *
 * Yes.  We'll take a trick from the gray scale rotation code.  The angles
 * for which changes in y coord result in the largest changes in run length
 * are very small angles (run length is proportional to
 * (y coord)/sin(rotAngle)).  Say we decide that rotating images by less
 * than 1 degree is silly because the jaggies introduced are more
 * objectional than the original skew.  A one degree angle gives a rise of
 * 1 bit in a span of 58 bits.  If we want to determine run lengths to
 * within 1 bit, we need to determine y position to within 1/58 bits.  We
 * will put 6 bits of fractional y coordinate for an accuracy of 1/64 bit
 * into the table.  (This turns out not to be a wonderful idea, but it lets
 * us make progress towards our ultimate goal.)
 *
 * Run length is not nearly so sensitive to fractional x coordinate.  In fact,
 * given our assumption that cos(rotAngle) = 1, run length is independent of
 * fractional X coordinate.    This is not true in normal geometry.  If we
 * can prove run length is independent of fractional X coordinate, we can
 * ignore it in the inner loop.
 *
 * In the proof, we'll concentrate on the path we claim
 * input pixels take across the output array with this assumption.
 *
 * The coordinates of a pixel will be measured relative to the upper left
 * corner of the pixel.  Fractional x and y coordinates range between 0 and 1.
 * An output pixel, O, is assigned the value of an input pixel, I, if the
 * intersection of O with I is greater than the intersection of O with any
 * other pixel.  Define a run of pixels to be {Xn}.  This contains n pixels
 * numbered 0 through n-1 from left to right.  The ith pixel in the run is
 * called Xi.  A run of output pixels {On} is assigned the values of a run
 * of input pixels {In} if and only if for every value of i, 0 <= i < n, Oi
 * is assigned the value of Ii.
 *
 * Say O0 is assigned the value of I0.  This can only happen if both I0's
 * fractional x coordinate, xf, satisfies -1/2 < xf < +1/2 relative to O0's
 * x coordinate.  If xf were less than -1/2, then the input to the right of
 * I0 would have a greater overlap with O0 than I0.  Similarly, if xf were
 * greater than 1/2, the input pixel to the left of I0 would have the greater
 * overlap. This is true for any value of y coordinate.
 *
 * Similarly, yf (the fractional y coordinate) must be in -1/2 < yf < 1/2.
 *
 * Without loss of generality, say we're rotating clockwise.  That means
 * the y coordinate of the input pixels in the output's frame of reference
 * is gradually decreasing (moving towards the top of the image).  Since we
 * assume that cos(rotAngle) = 1, the fractional x coordinate does not change
 * as we move across the input and output images.
 *
 * The run starts when xf is in its legal range and yf is just barely less
 * than 1/2.  As we move along the run, xf does not change and yf decreases
 * towards -1/2.  The run ends at I(n-1) whose fractional x coordinate is still
 * xf and whose fractional y coordinate is almost -1/2 (relative to the
 * output line).
 * The length of this run is
 *  - ([final yf] - [original yf])/sin(rotAngle) 
 * = ([original yf] - [final yf])/sin(rotAngle).
 * The original value of xf and the final value of xf are the same.  That
 * value does not appear here. Therefore, the length of the run is
 * independent of xf. Q.E.D.
 *
 * What do we do with this information?  Well, first it tells us that we
 * can ignore the fractional x coord in the inner loop.  That's a nice
 * simplification.  What else happens in the inner loop?
 *
 * Several conditions can cause us to break a contiguous run of bits:
 *  1. Reach the end of an input word
 *  2. Reach the end of an output word
 *  3. Reach the end of the current run
 *  4. Reach the left edge of the input image.
 *  5. Reach the right edge of the input image.
 *  6. Reach the right edge of the output image.
 *  7. Reach the top edge of the input image.
 *  8. Reach the bottom edge of the input image.
 *
 * Luckily, the last two conditions can only occur when we move up or down
 * in the input image, which only happens on a run boundary, thus are
 * subcases of case 3.
 *
 * The remaining 6 cases are bad enough.
 *
 * Furthermore, the input staircase may be shifted arbitrarily in a
 * horizontal direction with respect to the output line.  Since
 * cos(rotAngle) = 1, the horizontal distance between an input bit and
 * its output position will be constant for all input bits, but will
 * vary from one output line to the next.  Most of the shift distance will
 * be handled by the difference in input and output image pointers.
 * However, composition of output words will usually require shifting.
 *
 * Finally, runs may begin and end at arbitrary bit boundaries.  That means
 * the inner loop must keep track of shift amounts and use masks to
 * merge adjacent input runs into contiguous output words.  It must also
 * decide whether the next set of input words comes from the word to the
 * left, the word above, or the word above and to the left of the previous
 * input word.  Having decided this, it must decide whether the new input bits
 * are to be shifted right or left and how many of them should be combined
 * with the output word.
 *
 * How do we make sense of this mess?  Clearly, our code should spend most
 * of its time moving pixels and not so much time figuring out what to
 * do next.  We need some way of organizing things so that deciding what
 * to do next is relatively simple.
 *
 * Using constants instead of variables is usually a good idea.  Say that
 * instead of organizing the code so that each loop iteration transfers a
 * run of bits, we organize it so that each iteration computes a 32 bit
 * output word.  The x coordinate will always change by 32 in both worlds
 * (because of our assumption that cos(rotAngle) = 1).  The input's y
 * coordinate in the output's frame of reference always changes by
 * (32 * sin(rotAngle)).  That's a constant for any given rotation.  Things
 * are already looking up, we don't need to find the new y coordinate
 * in a table.
 * 
 * Any given output word is formed from contiguous sets of bits.  Each
 * set comes from one input word.  What do we know about the input words
 * and the sets?  Say the input words are stacked vertically so that in every
 * input word, the set of bits that went to the output word is aligned
 * with its final position in the output word.  This is a graphic depiction
 * of the staircase of inputs being mapped to an output.  One interesting
 * feature of this stack of input words is that all the bit 0 positions
 * line up. That's because every horizontal move of n bits in output space
 * corresponds to a move of exactly n bits in input space.  In fact, for
 * any given scan line, the position of bit 0 of the input words relative
 * to the output words will be constant across *all* the output words.
 *
 * What can we tell about the runs?  A really interesting feature is that
 * for any given rotation, there are only two possible run lengths.  For
 * example, say we wish to rotate by 3 degrees.  This means we should move
 * up one row in input space for every 19.1 bits we move horizontally.
 * The two possible run lengths are therefore 19 and 20.  Why is this
 * interesting?  It means that we don't (usually) need to determine the
 * the length of the current run from a fancy table, we can have the table
 * tell us whether to use run length A or run length B.
 *
 * Why do we say "usually"?  The very first run of input bits put into an
 * output word is often the remains of the last run put into the previous
 * output word.  That means it may range in length from 0 to MIN(max run
 * length, 32).  So, for each word, we need to know how long the first run
 * is exactly, then know whether each of the succeeding runs is of length
 * A or length B.  How many runs can fit into a 32 bit word?  We'll
 * assume that this code is only called for rotation angles of less than
 * 10 degrees.  That corresponds to runs of length 5.75 degrees.  5.55
 * such runs can fit into a 32 bit word, so we'll say 6.  That's not the
 * end of it.  Say the very first run only has 1 bit in it.  This is
 * followed by 5 runs, each of length 5.75 bits, for a total of 28.75 bits.
 * We've produced 1 + 28.75 = 29.75 bits.  There's clearly room for one
 * more run, for a total of 7 runs.  Therefore, run information for any 
 * given 32 bit word should contain at least 5 bits to specify the length 
 * of the first run plus 6 more bits to specify the lengths of the 
 * succeeding runs, for a total of 11 bits.  (Unless we find another way
 * to encode runs, which we will.  But this keep us moving down the path.)
 * 
 * Say bit 0 in the input words corresponds to bit n in the output words.
 * Let's call all the bits to the left of bit n in the output words the "left"
 * side of the output word.  Similarly, all the bits from bit n to the
 * right end of the output word are called the "right" side of the
 * output word.
 * 
 * Filling the output word can be done in four stages:
 *   1. Fill in first run.  This goes into the left side of the output
 *	word, though it may overlap into the right side.
 *   2. Fill in any remaining runs that start in the left side.
 *   3. If it exists, fill in the remaining part of the last left side run
 *	into the left portion of the right side of the output word.
 *   4. Fill in any remaining portions of the right side of the output
 *	word.
 * 
 * One pleasant feature of this arrangement is that execution proceed fairly
 * smoothly from one state to the next.  States 1 and 3 may be skipped if
 * the preceeding runs ended exactly on word boundaries.  This is pretty
 * easy to determine.  In any case, we don't need a fancy lookup table with
 * multi-way dispatch to decide what to do next.
 * 
 * That's all fine for the inner loop.  What is the basic structure of
 * the code?  We'll have an outer loop that calls this inner loop once
 * per output line.  The outer loop will be responsible for finding the
 * initial position, in both x and y, of the first input bit.  What does
 * this entail?
 *
 * By our assumption that cos(rotAngle) = 1, moving down one line in the
 * output image requires us to move down exactly one line in the input
 * image.  That's easy.  We add sin(rotAngle) to the current x coordinate
 * (relative to the input image).  That tells us when to move over by a
 * bit in the input.  Do we use exactly the same fractional y coordinate
 * for the start of all input lines?  Our assumption that cos(rotAngle)
 * equals 1 says we should.  We'll try it and see what the results look
 * like.  Using exactly the same fractional y coordinate at the start of
 * all input lines means we'd get exactly the same sequence of run lengths
 * in every line.  Is this is good idea?  Well, if the lines started with
 * the same integral x coordinate, it seems swell.  The stairsteps of the
 * two paths through the input will dovetail perfectly.  Every input bit
 * will appear in exactly one output pixel which seems nice.
 *
 * However, say the starting point of one input line is both down 1
 * and over 1 from the starting point of the previous input line.  Say
 * the rotation direction is clockwise, so the stairsteps rise gradually
 * as we move to the right (this diagonal path is reproduced as a
 * horizontal line in the output image, rotating it in a clockwise direction).
 * The vertical path taken by the left ends of the staircase moves gradually
 * to the right as we go down the image.  Relative to the staircase that
 * nests into the ones above it, our new staircase will be displaced 1 bit
 * to the right.  This means it will ignore 1 input bit at every transition
 * between runs.  How bad is this?  Well, I really don't know, but here's
 * a rationalization for deciding it's ok.  Say the pixel skipped is a
 * black pixel in the midst of black pixels or a white pixel in the midst
 * of white pixels.  Clearly, the effect will not be noticeable.  Say the
 * pixel skipped was part of a big edge.  The edge will have a jaggie at
 * the point where the pixel was skipped.  This is what you expect when
 * you rotate an image, so I can't get too excited about it.  Say the pixel
 * was black, but surrounded by white pixels or vice versa.  It seems to me
 * that there are two subcases.  If the pixel was noise, removing it is
 * a fine thing.  If the pixel is part of a carefully arranged set of dots
 * intended to portray a gray level, we are essentially sampling a random
 * set of data.  Say there are b black dots and w white dots in the region
 * of interest.  Say we ignore some fraction i of the dots.  The final
 * ratio of black to white dots will be b*i / w*i = b/w just as before.
 *
 * Clearly, we'll get in trouble if there is an input pattern that matches
 * the pattern of dots being ignored.  I guess we'll hope this hardly
 * ever happens.
 *
 * All this brings up another nice simplification.  If we start *all* lines
 * with the same fractional y position so we get exactly the same sequence
 * of run lengths for all output lines. we can create a lookup table that is
 * indexed simply by horizontal position in the output image.  This is both
 * much smaller and easier to index than a fully general lookup table.
 * Furthermore, since the very first fractional y coordinate is probably
 * arbitrary, we can choose any one we like; 0 for example.  This means we
 * can dispense with keeping the fractional y coordinate around at all!
 *
 * Well, gee, if the table is just a list of run lengths, why not have 1
 * byte per run length?  Because some runs are longer than 255 bits.  Ok
 * why not have one 16-bit thing per run length?  Because we'd still need
 * to generate a mask for each run length.  That will take time.  If we
 * branch on 1 bit specifying short/long run, we can go to code that has
 * both the run length and the appropriate mask in registers all ready to go.
 *
 * Ok, then why not just have a single long bit array that contains
 * nothing but run length specifiers?  Because the code would have to
 * compute how much run was left over from the previous word on every
 * new word.  That takes longer than just looking up the answer (right?).
 * The real trick is to simulate processing a line and record all the
 * information that would be useful at each step.
 *
 * From that point of view, say the preprocessing step outputs a sequence
 * of properly shifted masks.  Each mask is correctly positioned for the
 * output word on which it will be used.  There's one entry in the sequence
 * for each run plus on entry for each output word crossing (the masks
 * necessarily end at word boundaries).  The code tests each mask to
 * see if it should change states.  It can do this by ANDing the
 * boundary bits for both input a output words with the masks to see if
 * the boundary bits have just been filled.  The boundary bit for
 * output words is always the LSbit.  If a mask has that bit turned on,
 * we're done with this output word.  Similarly, if bit 0 (LS Bit) of the input
 * word corresponds to bit n of the output words and bit n is a 1 in the
 * latest mask, we're done with the left side of the output word.
 * Further inputs will be shifted right instead of left to align them.
 * If bit n-1 is also turned on, we should use the input word to the 
 * right to complete the current run.  Otherwise, move up/down a row in 
 * the input image.
 *
 * There's still a problem with this.  How do we tell the difference
 * between a run that *ends on* an output word boundary and a run that
 * *crosses* an output word boundary just by looking at the masks?  After
 * all, masks must end on output word boundaries.  Since masks may be
 * 32 bits wide, we can't hide secret bits inside the masks.  It seems like
 * the easiest thing to do is to follow each output-word-ending mask
 * with the number we should add to the current source address to
 * get the next source word.  Normally, runs cross output word boundaries
 * so this number is 0.  If the run ends on an output word boundary,
 * this number is +- sourcePitch.  This seems simpler than having yet
 * another table (with its own base and index).
 *
 * Yes, we'll need a duplicate set of code that's executed when the input
 * words are exactly aligned with the output words.  This set of code will
 * just ignore the "right side" processing.
 *
 * What happens at the edges of the input image?  The rotated image will
 * have triangles of new stuff at its edges corresponding to areas that
 * were off the edge of the input image.  By fiat, this stuff should be
 * white (like the platen).  In binary images, it should therefore be
 * 0's (1's in gray images).  Frames are filled with 0's, so we're done,
 * right?  No.  We will often be much further off the edge of the image
 * than can be handled by the frames.  A nice way to decide whether the
 * next input word is off the edge is to maintain two counters, xCount 
 * and yCount.  The xCount is incremented as we move left, yCount is 
 * incremented as we move down.  The counters are initialized
 * at the beginning of each line with the distance to the next image edge,
 * sort of.  We would like it to be true that when either counter is
 * negative, we're off the input image.  Therefore, when we're to the left
 * of the input image, xCount contains the x coordinate, a negative number.
 * The same is true of yCount when we're above the image.  However, when
 * we're in the middle of the image, we'd still like to use xCount and yCount
 * to detect the right and bottom edges.  How?  We use a dirty trick.  We
 * set them to 0x80000000 - (distance to edge).  We increment these as we
 * move.  When we reach the edge, they overflow from the largest positive
 * number, 0x7FFFFFFF, to the largest negative number, 0x80000000.  Yeah, it's
 * overflow, but it's still a negative number.  This way, we can always
 * blindly increment the counters and test for negative values.
 *
 * As we learned in the gray rotation case, it is just barely possible to
 * encounter both the left and right edges of the input image in a single
 * sweep.  That's because we'll read a few more input bits than output bits.
 * So, when xCount goes from negative to positive on the left side of the
 * image (xCount < 32), we change it to 0x80000000 - image width - xCount.
 *
 * Let's get back to the structure of the code.  There will be one routine
 * that produces one output line.  It will be called in a loop that locates
 * the leftmost input bit.  Having found the leftmost input bit, we set
 * up xCount, yCount, the original source address and a mask containing
 * a bit in the output bit position corresponding to input bit 0.  Together
 * with address of the sequence of masks, these are passed to the routine
 * that produces one output line.
 *
 * Here's a really annoying detail.  It's of interest only because I got it
 * wrong the first time.  When we move horizontally in the input image, 
 * *exactly* how do we adjust the coordinates of the input pixel in the output's
 * frame of reference?  Our convention is that positive rotation angles imply
 * clockwise rotation.  The cosine function is symmetric and positive around
 * 0.  As agreed, we will approximate cos(rotAngle) with 1. The sine
 * function crosses through (0, 0) with a positive slope.  To
 * rotate an image clockwise, we sweep across it in paths that go from lower
 * left to upper right.  These paths put out horizontal pixels in the rotated
 * image.  Therefore, relative the *rotated*, or output, image, a move to the
 * left in the input image is a move towards the *lower* right in the output
 * image.  I always get confused by inversions.  Here's a table:
 *
 *	Direction in		Coord. mods in
 *	Input Image		Output Image
 *	  +X (to the right)	X = X + 1; Y = Y + sin(rotAngle)
 *	  -X (to the left)	X = X - 1; Y = Y - sin(rotAngle)
 *	  +Y (down)		X = X - sin(rotAngle); Y = Y + 1
 *	  -Y (up)		X = X + sin(rotAngle); Y = Y - 1
 *
 * Note that these formulas work for both positive and negative rotation
 * angles.  In these formulas, the cosine term is the major change of
 * direction and the sine just specifies the tilt.
 *
 * How do we find the very first input point?  This is a function of both
 * the rotation angle and the point about which the image should be rotated.
 * for example, if we rotate around the upper left corner, the first input
 * point might always be the upper left corner!
 *
 * In the gray rotate code, we always rotated about the center of the image.
 * That was usually ok, but there have been grumblings.  For example, if 
 * the scanner returned an image that contained a smaller, tilted, piece
 * of paper, it would be nice to rotate just the piece of paper and move
 * the rotated image to the upper left corner of the result.  The only
 * problem with this scenario is that we don't know how to find the piece
 * of paper inside the image.  For now, we'll just rotate around the center
 * as before.  If someone wants some other sort of rotation, they can
 * ask.
 *
 * Yeah, but how do you find the first input point?  There's a big comment
 * from the gray rotation code that we'll copy here.
 *
	 * Get ready to draw triangles in your notebook.  We want to start
	 * generating output bits at the output's upper left corner.  In the
	 * output's coordinate system, the first output position is (0,0).
	 * Which input image pixel will affect the output pixel at location
	 * (0,0)?  To find that pixel, we need to calculate the location of
	 * output point (0,0) in the input image's frame of reference.  This
	 * is the only place where we need to find an output point using the
	 * input frame of reference.
	 *
 	 * We're rotating about the center, so the width and height of the
	 * image figure into this.  Let's start with an easy calculation.
 	 * Draw a right triangle connecting the upper left corner of the
	 * image, the center and a point on the left edge level with the
	 * center.  The X coordinate of the upper left corner of the image
	 * with respect to the center is the length of the horizontal leg
	 * of this triangle.  Similarly, the y coordinate is the length
	 * of the vertical leg.  These can be found by multiplying the
	 * hypotenuse of the triangle by the cosine and sine respectively
	 * of the angle at the center.  We just happen to know that the
	 * x coordinate is w/2 and the y coordinate is h/2, so we can
	 * construct the following (hopefully tautological) formulas:
	 *
	 * X = SQRT( (h/2)**2 + (w/2)**2 ) * cos(arctan(h/w))
	 * Y = SQRT( (h/2)**2 + (w/2)**2 ) * sin(arctan(h/w))
	 *
	 * JAC Note: hypotenuse = SQRT( (h/2)**2 + (w/2)**2 )
	 * Great. However, the input is tilted at an rotAngle with respect
	 * to the output.  So, the X and Y coords of the input relative
	 * to the center (we rotate around the center so it is the same
	 * for both input and output images) are:
	 * Xi = SQRT((h/2)**2 + (w/2)**2) * cos(arctan(h/w) + rotAngle)
	 * Yi = SQRT((h/2)**2 + (w/2)**2) * sin(arctan(h/w) + rotAngle)
	 *
	 * (JAC Note: The length of the hypotenuse remains constant so we
	 * can continue to use SQRT( (h/2)**2 + (w/2)**2 ) is the formula.)
	 * However, we measure coordinates from the upper left corner, not
	 * the middle.  The coordinates we want are the differences between
	 * these two (positive rotAngle => clockwise rotation):
	 * Xi' = SQRT((h/2)**2 + (w/2)**2) * (cos(arctan(h/w)) -
	 *				      cos(arctan(h/w) + rotAngle)
	 * Yi' = SQRT((h/2)**2 + (w/2)**2) * (sin(arctan(h/w)) -
	 *				      sin(arctan(h/w) + rotAngle)
	 *
	 * Neat, except it's backwards.  This tells us how to rotate the upper
	 * left corner of the input in the input's frame of reference.  We
	 * really want to know the position of the upper left corner of the
	 * output in the input's frame of reference.  To find that, we
	 * just rotate the other way.
	 * Xi' = SQRT((h/2)**2 + (w/2)**2) * (cos(arctan(h/w) + rotAngle) -
	 *				      cos(arctan(h/w))
	 * Yi' = SQRT((h/2)**2 + (w/2)**2) * (sin(arctan(h/w) + rotAngle) -
	 *				      sin(arctan(h/w))
	 *
 * Too easy, huh?
 *
 * That's about it.  The basic steps of this routine are:
 *   1. Boilerplate (strip off frame bits, etc).
 *   2. Make the table containing sequences of masks and source address
 *	offsets.
 *   3. Find the first (upper left) input pixel using the calculations
 *	described above.
 *   4. Loop through all the output rows, calling the proc that produces one
 *	row at a time for each row.  After the proc returns, calculate
 *	the position of the next input pixel.
 *
 * *********************** FIRST REWORK ***************************
 *
 * That was very nice.  It produced nice images and was pretty fast.  But
 * not fast enough.  One (minor) way to speed things up is to rework the
 * routine that creates the masks.  In the first version, it found each
 * run length by dividing the distance from the current vertical position
 * to the top of the line of pixels by sinRotAngle.  We don't need to
 * do all those divides (and corresponding multiplies to advance the
 * current position).
 *
 * There are really only two possible run lengths.  Say the exact run
 * length is 29.3.  There are only two possible run lengths, 29 and 30.
 * each time we use 29, we're 0.3 under the perfect amount.  Every time we
 * use 30, we're 0.7 over the perfect amount.  Say we keep a running sum
 * of these differences (-0.3 when we use the short run, +0.7 when we use
 * the long run).  Whenever the sum is negative, use a long run, otherwise
 * use the short run.  Too easy, right?  All this happens in
 * makeNoInterpRotTable.
 *
 * Implementing that idea that made me feel better, but didn't have any
 * perceptible effect on performance.
 *
 * The next point to note is that our code checks every input word to see
 * if it's off the input image.  Almost all words are not off the input
 * image.  Isn't there some way to skip the test most of the time?
 *
 * Our current solution is to divide each code for procesing lines into
 * three sections.  The first section checks (and maintains) both xCount
 * and yCount.  It is used near the top and bottom of the image and at
 * the left edge of lines that start off the left side of the input image.
 *
 * The second section of code maintains xCount, but fetches input words
 * without checking xCount.  This section of code is executed
 * either after crossing the left edge of the image (so we know we won't be
 * crossing the right edge) or when we start the input line in the input
 * image.  We should be spending by far the most time here.
 *
 * The third section of code maintains xCount and checks it when fetching
 * input words.  This section of code is used when we're getting ready to
 * cross the right side of the image, or have already crossed it.  This
 * section of code ignores yCount.
 *
 * How do we get from section to section?  We pass two boundaries down to
 * ip_rotateNoInterpLine.  These boundaries refer to the number of output
 * words put out.  ip_rotateNoInterpLine executes the first section of code
 * until is has put out enough words to pass the first boundary, then (if
 * any words remain on the line) executes the second section until is 
 * passes the second boundary, then, if any output words remain, it 
 * finishes the line with the third section.
 *
 * In practice, the boundaries will be set as follows:
 * 1. Near top or bottom of image: execute first section of code only.
 * 2. Starting to left of input image: execute first and second sections only.
 * 3. Starting inside image: execute second and third sections only.
 *
 * I toyed with the idea of generating lines either from left to right (if
 * line goes off the left side of the input image) or right to left.  This
 * was more complicated in that we had to remember both the left and right
 * endpoints of the lines.
 *
 * ***********************************************************************
 * BUG FIX, 9/20/95:  Richard Mlynarik found that rotating gray images
 * sometimes produced artifacts on the right side.  After further
 * investigation, I found that there's a generic problem with the way
 * we handle the rightmost word of an image.  If the right boundary of
 * an input image does not end on a word boundary, we sometimes use the
 * entire word (which doesn't work when the frame is not white) and sometimes
 * ignore the rightmost word entirely.  No good.
 *
 * The problem has the do with confusion about xCount.  In the gray rotate
 * code, we read one pixel at a time and updated xCount each time.  In
 * this code we read 32 bits (some number of pixels) each time.  xCount
 * indicated the distance from some pixel inside the word to the edge
 * of the input image.  This uncertaintly threw us off.
 *
 * In the new scheme, we truncate xCount to the nearest multiple of 32
 * towards -infinity.  This makes it calculate the distance from the leftmost
 * bit in the word being read to the edge of the input image.  We then
 * check the value of xCount in the third section of code that is used when
 * we're approaching the right edge of the image.  If xCount is < 0 (greater
 * than 0x80000000), we know the entire word being read is off the right
 * side of the image and just use whiteval.  This is the common case
 * there.  If xCount is positive, we now have two cases where before we
 * had one.  If xCount + 32 is also postive, we read an entire new word
 * from the source image.  If xCount+32 is negative ( >= 0x80000000 ),
 * we merge the word read from the image with whiteval using a new mask.
 * The mask is a constant computed from the width of the image in bits.
 * We can use the constant since the width of the source image (the one
 * from which we are reading the data) is constant.  In fact, since we
 * know that the stuff off the edge of the source image is frame, therefore
 * filled with 0's, we can just supply a rightWhite mask which is filled
 * with whiteVal bits in the frame section of the input word.  We just
 * OR that with the partial input word read.  The rightWhite mask is
 * computed in the lower level routines since not all of them take a
 * whiteVal anyway.
 *
 * 
 *
 **************************************************************************
 *
 * So much for the comments from the C implementation.  We'll basically just
 * try to beat the compiler here.  No real changes in algorithm.  A nice
 * source of documentation might be the complete, functional C Implementation
 * in minirt1.c
 */
/**************************************************************************
 *  ip_rotateNoInterpLine()
 *		Generates one rotated line without using interpolation.
 *
 *	Input:
 *		pS =		pointer to 32 bit word containing first pixel
 *				of unrotated data.
 *		sBpl =		source bytes/line
 *		pD =		pointer to destination line
 *		w =		width of source image in bits
 *		xCount = 	increment this count at each move towards the
 *				right.  While xCount < 0, we are off the image.
 *				Distance to horizontal edge.
 *		yCount =	increment this count at each move towards the
 *				bottom.  While yCount < 0, we are off the image.
 *				Distance to vertical edge.
 *		leftShift =	position of leftmost image bit within
 *				first image word as measured from left
 *				edge of first image word (0 => input and
 *				output words are aligned).  This bit
 *				should be shifted to become bit 0 (leftmost) 
 *				of the first output word, hence it is the
 *				distance by which input words should be
 *				shifted left to be aligned with the output
 *				words.
 *		pMaskTable =	Pointer to sequence of masks used to insert
 *				runs of pixels from input words into output
 *				words.
 *		whiteVal =	A 32 bit word full of white pixels.  This
 *				is 0x00000000 in binary images and
 *				0xFFFFFFFF in gray images.
 *		boundary1 =	Check xCount and yCount when reading input
 *				words for all output words
 *				processed to the left of boundary1.
 *		boundary2 =	Check xCount when reading input
 *				words for all output words
 *				processed to the left of boundary2.
 *	Output:
 *		Stores 1 rotated line into output image.
 *
 **************************************************************************/

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssnirot.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:19 $"
	.align	4


! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPStateIn	%i0
#define	rSrcAddrIn	%i1
#define rSrcPitchIn	%i2
#define rDstAddrIn	%i3
#define rWidthIn	%i4
#define rXCountIn	%i5

#define	rSrcAddr	%i0		/* current addr in source image */
#define rSrcPitch	%i1		/* source image's bytes/line */
#define rDstAddr	%i2		/* current addr in dest image */
#define rWidth		%i3		/* dest image's BITS */
#define rXCount		%i4		/* used to determine when we're
					 * off the right or left edges of
					 * the input image. */
#define rYCount		%i5		/* used to define when we're off the
					 * top or bottom edges of the input
					 * image. */
#define	rLeftShift	%g1		/* distance by which inputs should
					 * be shifted left */
#define rMaskTable	%g2		/* ptr to table containing sequence
					 * of masks. */
#define	rWhiteVal	%g3		/* value of "white".  If this is a
					 * binary image, whiteVal = 0.  If
					 * this is a gray image, whiteVal
					 * is 0xFFFFFFFF. */
#define rBoundary1	%g4		/* check xCount and yCount while
					 * processing output words when
					 * rWidth is > rBoundary1. */
#define rBoundary2	%g5		/* check neither xCount nor yCount
					 * while processing output words when
					 * rBoundary1 < rWidth < rBoundary2. */

#define	rLeftmostBit	%o0		/* a mask containing a single
					 * 1 in the position where
					 * the MS bit of the input
					 * words aligns with the
					 * output words. */
#define rRightmostBit	%o1		/* a mask containing a single
					 * 1 in the position where
					 * the LS bit of the input
					 * words aligns with the
					 * output words. */
#define rOut0		%o2		/* the output word under
					 * formation */
#define rInput		%o3		/* word read from input */
#define rMask		%o4		/* the current mask */
#define rRightShift	%o5		/* distance by which input words
					 * should be shifted right.  As
					 * discussed above, this is constant
					 * across the entire line. */
#define rRightWhiteVal	%l0		/* holds filler for right edge
					 * of line.
					 */
#define rSrcHeight	%l1		/* height of source image - 1.  Used
					 * to reinitialize rYCount */
#define r80000000	%l2		/* holds the constant 0x80000000.
					 * used to check rYCount */

/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	srcAddrOffset	0x48
#define	srcPitchOffset	0x4c
#define	dstAddrOffset	0x50
#define	dstWidthOffset	0x54
#define xCountOffset	0x58
#define	yCountOffset	0x5c
#define leftShiftOffset 0x60
#define	maskTableOffset	0x64
#define	whiteValOffset	0x68
#define rightWhiteOffset 0x6c
#define	boundary1Offset	0x70
#define	boundary2Offset	0x74
#define srcHeightOffset	0x78

/* magical C Frame constant (no local variables) */
#define	LF26		-64

LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! _ip_rotateNoInterpLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_rotateNoInterpLine)
MK_GLOBAL_NAME(ip_rotateNoInterpLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPStateIn,  [%fp+ipStateOffset]
	st	rSrcAddrIn,  [%fp+srcAddrOffset]
	st	rSrcPitchIn, [%fp+srcPitchOffset]
	st	rDstAddrIn,  [%fp+dstAddrOffset]
	st	rWidthIn,    [%fp+dstWidthOffset]
	st	rXCountIn,   [%fp+xCountOffset]

	ld	[%fp+srcAddrOffset],  rSrcAddr
	ld	[%fp+srcPitchOffset], rSrcPitch
	ld	[%fp+dstAddrOffset],  rDstAddr
	ld	[%fp+dstWidthOffset], rWidth
	ld	[%fp+xCountOffset],   rXCount
! load vars that were passed in memory
	ld	[%fp+yCountOffset],   rYCount
	ld	[%fp+leftShiftOffset], rLeftShift
	ld	[%fp+maskTableOffset], rMaskTable
	ld	[%fp+whiteValOffset], rWhiteVal
	ld	[%fp+rightWhiteOffset], rRightWhiteVal
	ld	[%fp+boundary1Offset], rBoundary1
	ld	[%fp+boundary2Offset], rBoundary2
	ld	[%fp+srcHeightOffset], rSrcHeight
	dec	rSrcHeight			! really want height - 1.
	sethi	%hi(0x80000000), r80000000

! If the source and dest are lined up perfectly, execute a special body
! of code, otherwise go to the regular body.
	tst	rLeftShift			! shift distance == 0?
	bnz	rNIShiftingRequired		! no, do regular code

! Loop through the words of this row.  Start
! with the first section of code.  This one
! checks both xCount and yCount on each fetch of
! input word.
	cmp	rWidth, rBoundary1		! is rWidth > rBoundary1?
	ble	rNINoShiftSection2Entry		! no, go do section 2 now.
	cmp	rWidth, rBoundary2		! see if we should do sec 2

	ld	[rMaskTable], rMask		! get first mask

! ************* SECTION 1, NO SHIFTING ***********

rNINoShiftSection1:
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, rYCount, %g0		! are either xCount or yCount
						! negative?
	bneg	rNINoShiftSec1aOffInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNINoShiftSec1aOffInput		! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNINoShiftSec1aOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	and	rInput, rMask, rOut0		! initialize output word
	cmp	rYCount, r80000000		! just below bottom of image?
	bne	rNINoShiftSection1aa		! no, just decrement rYCount
	dec	rYCount

! we're at the bottom of the image, so load rYCount with height - 1.
	mov	rSrcHeight, rYCount

! Keep inserting runs until we've filled this word.
rNINoShiftSection1aa:
	andcc	rMask, 0x1, %g0			! have we touched the LSbit?
	bnz	rNINoShiftSec1Fin		! yes, finish up
	ld	[rMaskTable], rMask		! and get next mask
rNINoShiftSection1b:
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, rYCount, %g0		! are either xCount or yCount
						! negative?
	bneg	rNINoShiftSec1bOffInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNINoShiftSec1bOffInput		! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNINoShiftSec1bOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it

	cmp	rYCount, r80000000		! just below bottom of image?
	bne	rNINoShiftSection1ba		! no, just decrement rYCount
	dec	rYCount

	mov	rSrcHeight, rYCount		! yes, load with height-1

rNINoShiftSection1ba:
	andcc	rMask, 0x1, %g0			! have we touched the LSbit?
	bz	rNINoShiftSection1b		! no, keep going
	ld	[rMaskTable], rMask		! and get next mask

! We've finished one word in the no-shifting case.  If the next mask
! entry equals sBpl, increment yCount because we weren't supposed to have
! moved up at the end of the last run (because the end of the word wasn't
! really the end of the run).  Also update rSrcAddr accordingly,
! update xCount because we moved over a word and write the output word.
rNINoShiftSec1Fin:
	add	rMaskTable, 4, rMaskTable	! pt to next mask
	add	rSrcAddr, 4, rSrcAddr		! move over 1 word
	cmp	rMask, 0			! should we have moved
						! up a line?
	be	rNINoShiftSec1SrcAddrMod	! yes, so skip fixup
	add	rSrcAddr, rMask, rSrcAddr
	cmp	rYCount, rSrcHeight		! if we moved up into the
						! image, move back down,
	bne	rNINoShiftSec1SrcAddrMod
	inc	rYCount				! fixup rYCount
	mov	r80000000, rYCount		! otherwise, remember we're
						! at the bottom of the image.
rNINoShiftSec1SrcAddrMod:
	add	rXCount, 32, rXCount		! moved over 1 word
	st	rOut0, [rDstAddr]		! store output word
	add	rDstAddr, 4, rDstAddr		! and inc ptr

! are there more words to process in this line?
	sub	rWidth, 32, rWidth		! have done 32 bits
	subcc	rWidth, rBoundary1, %g0		! is rWidth > rBoundary1?
	bg	rNINoShiftSection1		! stay on section 1.
	ld	[rMaskTable], rMask		! and get first mask

! Having completed section 1, are we done?
	tst	rWidth
	bg	rNINoShiftSection2Entry		! nope, do section 2.
	cmp	rWidth, rBoundary2		! is rWidth > rBoundary2?

! We were only supposed to do section 1, so return now
	ret
	restore

! Loop through the words of this row.  Start
! with the first section of code.  This one
! checks neither xCount nor yCount on each fetch of
! input word.
! ************* SECTION 2, NO SHIFTING ***********
rNINoShiftSection2Entry:
	ble	rNINoShiftSection3Entry		! no, go do section 3 now.
	tst	rWidth				! see if we`re done

	ld	[rMaskTable], rMask		! and get first mask

rNINoShiftSection2:
	add	rMaskTable, 4, rMaskTable
	ld	[rSrcAddr], rInput		! always use real input word
rNINoShiftSec2aOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	and	rInput, rMask, rOut0		! initialize output word

! Keep inserting runs until we've filled this word.
	andcc	rMask, 0x1, %g0			! have we touched the LSbit?
	bnz	rNINoShiftSec2Fin		! yes, finish up
	ld	[rMaskTable], rMask		! and get next mask
rNINoShiftSection2b:
	add	rMaskTable, 4, rMaskTable
	ld	[rSrcAddr], rInput		! no, use real input word
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it
	andcc	rMask, 0x1, %g0			! have we touched the LSbit?
	bz	rNINoShiftSection2b		! no, keep going
	ld	[rMaskTable], rMask		! and get next mask

! We've finished one word in the no-shifting case.  Update rSrcAddr
! accordingly,
rNINoShiftSec2Fin:
	add	rMaskTable, 4, rMaskTable	! pt to next mask
	add	rSrcAddr, 4, rSrcAddr		! move over 1 word
	add	rSrcAddr, rMask, rSrcAddr
	add	rXCount, 32, rXCount		! moved over 1 word
	st	rOut0, [rDstAddr]		! store output word
	add	rDstAddr, 4, rDstAddr		! and inc ptr

! are there more words to process in this line?
	sub	rWidth, 32, rWidth		! have done 32 bits
	subcc	rWidth, rBoundary2, %g0		! is rWidth > rBoundary2?
	bg	rNINoShiftSection2		! stay on section 2
	ld	[rMaskTable], rMask		! and get first mask

! Having completed section 2, are we done?
	tst	rWidth
	bg	rNINoShiftSection3Entry		! nope, do section 3.
	tst	rWidth				! see if we're all done

! We were only supposed to do section 2, so return now
	ret
	restore

! Loop through the words of this row.  Start
! with the first section of code.  This one
! checks only xCount on each fetch of
! input word.
! ************* SECTION 3, NO SHIFTING ***********

rNINoShiftSection3Entry:
	ble	rNINoShiftSection3End		! yep, quit
	ld	[rMaskTable], rMask		! and get first mask

rNINoShiftSection3:
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, %g0, %g0		! is xCount negative?
	bneg	rNINoShiftSec3aOffInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNINoShiftSec3aOffInput		! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNINoShiftSec3aOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	and	rInput, rMask, rOut0		! initialize output word

! Keep inserting runs until we've filled this word.
	andcc	rMask, 0x1, %g0			! have we touched the LSbit?
	bnz	rNINoShiftSec3Fin		! yes, finish up
	ld	[rMaskTable], rMask		! and get next mask
rNINoShiftSection3b:
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, %g0, %g0		! is xCount negative?
	bneg	rNINoShiftSec3bOffInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNINoShiftSec3bOffInput		! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNINoShiftSec3bOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it
	andcc	rMask, 0x1, %g0			! have we touched the LSbit?
	bz	rNINoShiftSection3b		! no, keep going
	ld	[rMaskTable], rMask		! and get next mask

! We've finished one word in the no-shifting case.  Update rSrcAddr
! accordingly, update xCount because we moved over a word and write
! the output word.
rNINoShiftSec3Fin:
	add	rMaskTable, 4, rMaskTable	! pt to next mask
	add	rSrcAddr, 4, rSrcAddr		! move over 1 word
	add	rSrcAddr, rMask, rSrcAddr
	add	rXCount, 32, rXCount		! moved over 1 word
	st	rOut0, [rDstAddr]		! store output word
	add	rDstAddr, 4, rDstAddr		! and inc ptr

! are there more words to process in this line?
	subcc	rWidth, 32, rWidth		! have done 32 bits
	bg	rNINoShiftSection3		! stay on section 3.
	ld	[rMaskTable], rMask		! and get first mask

! We've finished the last section of this line, so return now
rNINoShiftSection3End:
	ret
	restore


! **************************************************************************
!
! *****************  CODE THAT SHIFTS INPUT WORDS  *************************
!
! **************************************************************************

! This is very similar to the code above, except that it shifts the input
! words as necessary.  It also detects when the left portion of an output
! word has been loaded so it can change from left shifting to right shifting.

rNIShiftingRequired:

! Setup registers that weren't needed in the code above.
	sub	rLeftShift, 32, rRightShift	! will negate this below
	neg	rRightShift			! equals 32 - rLeftShift
	sethi	%hi(0x80000000), rRightmostBit	! just a place to put this
	srl	rRightmostBit, rRightShift, rLeftmostBit
	sll	rLeftmostBit, 1, rRightmostBit


! Loop through the words of this row.  Start
! with the first section of code.  This one
! checks both xCount and yCount on each fetch of
! input word.
	cmp	rWidth, rBoundary1		! is rWidth > rBoundary1?
	ble	rNIShiftSection2Entry		! no, go do section 2 now.
	cmp	rWidth, rBoundary2		! is rWidth > rBoundary2?

	ld	[rMaskTable], rMask		! and get first mask

! ************* SECTION 1, SHIFTING REQUIRED ***********

rNIShiftSection1:
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, rYCount, %g0		! are either xCount or yCount
						! negative?
	bneg	rNIShiftSec1aOffInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNIShiftSec1aOffInput		! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNIShiftSec1aOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	sll	rInput, rLeftShift, rInput	! shift into position
	and	rInput, rMask, rOut0		! initialize output word

! Keep inserting runs until we've filled the left portion of this word.
	andcc	rMask, rRightmostBit, %g0	! have we touched the
						! LS bit of left portion?
	bnz	rNIShiftSec1ToRightHalf		! yes, finish up
	cmp	rYCount, r80000000		! and see if we're at the
						! bottom of the image
! Here's where we decrement rYCount or set it to hS - 2
rNIShiftSection1b:
	bne	rNIShiftSection1c		! just decrement rYCount
	dec	rYCount
! we're at the bottom of the image, so reinitialize rYCount
	mov	rSrcHeight, rYCount

rNIShiftSection1c:
	ld	[rMaskTable], rMask		! and get next mask
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, rYCount, %g0		! are either xCount or yCount
						! negative?
	bneg	rNIShiftSec1bOffInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNIShiftSec1bOffInput		! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNIShiftSec1bOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	sll	rInput, rLeftShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it
	andcc	rMask, rRightmostBit, %g0	! have we touched the
						! LS bit of left portion?
	bz	rNIShiftSection1b		! no, keep going
	cmp	rYCount, r80000000		! and see if we're at the
						! bottom of the image

! We're at the end of the left portion of the output
! word.  If the most recent mask overlaps the leftmost
! word of the next input, we should go back down a
! line, move over 1 word and finish filling the mask
! Otherwise, just move over 1 word.
! But first, see if we should decrement rYCount because we're
! not at the bottom of the image, or should set it to srcHeight - 1
! because that's where we are.
rNIShiftSec1ToRightHalf:
	bne	rNIShiftSec1ToRightHalf1	! just decrement rYCount
	dec	rYCount
! we're at the bottom of the image, so reinitialize rYCount
	mov	rSrcHeight, rYCount
rNIShiftSec1ToRightHalf1:
	add	rSrcAddr, 4, rSrcAddr		! move input over
	andcc	rMask, rLeftmostBit, %g0	! does the run continue?
	bz	rNIShiftSec1RightHalfEntry	! no, just do right half
	add	rXCount, 32, rXCount

! The previous run continued here, so continue it with the word that's
! below us.  Note that if the run went beyond the end of the output word,
! we would not insert any more runs.  If we just moved rYCount up to
! rSrcHeight, we can skip this read.  Otherwise, look at rYCount+1 and
! rXCount.
	cmp	rYCount, rSrcHeight		! are we really off the
	be	rNIShiftSec1ContRunInput	! bottom of the image?
	mov	rWhiteVal, rInput		! yes, use whiteVal
	add	rYCount, 1, rInput		! test with rYCount+1
	orcc	rXCount, rInput, %g0
	bneg	rNIShiftSec1ContRunInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNIShiftSec1ContRunInput	! not near, use entire src wd
	ld	[rSrcAddr+rSrcPitch], rInput	! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNIShiftSec1ContRunInput:
	srl	rInput, rRightShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it

! while the mask doesn't touch the LS bit of rOut0, keep putting runs into
! the right half of rOut0.
rNIShiftSec1RightHalfEntry:
	andcc	rMask, 0x1, %g0			! touched LS bit?
	bnz	rNIShiftSec1Fin			! yes, we're done
	ld	[rMaskTable], rMask		! get next mask in either case

rNIShiftSec1RightHalf:
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, rYCount, %g0		! are either xCount or yCount
						! negative?
	bneg	rNIShiftSec1RightHalfInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNIShiftSec1RightHalfInput	! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNIShiftSec1RightHalfInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	srl	rInput, rRightShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it
	cmp	rYCount, r80000000		! if we're not at the bottom of
	bne	rNIShiftSec1RightHalfInput1	! the image, just decrem
	dec	rYCount				! just decrement rYCount
	mov	rSrcHeight, rYCount		! otherwise, load with coord
						! of bottom line.
rNIShiftSec1RightHalfInput1:
	andcc	rMask, 0x1, %g0			! have we touched the
						! LS bit of the word?
	bz	rNIShiftSec1RightHalf		! no, keep going
	ld	[rMaskTable], rMask		! and get next mask

! We've finished one word in the shifting case.  If the next mask
! entry equals sBpl, increment yCount because we weren't supposed to have
! moved up at the end of the last run (because the end of the word wasn't
! really the end of the run).  Also update rSrcAddr accordingly,
rNIShiftSec1Fin:
	add	rMaskTable, 4, rMaskTable	! pt to next mask
	cmp	rMask, 0			! should we have moved
						! up a line?
	be	rNIShiftSec1SrcAddrMod		! yes, so skip fixup
	add	rSrcAddr, rMask, rSrcAddr
! If we just moved up to the bottom line of the image, move back
! down to line 80000000 so we remember to move up next time.
	cmp	rYCount, rSrcHeight
	bne	rNIShiftSec1SrcAddrMod
	inc	rYCount				! fixup rYCount the ordinary way
	mov	r80000000, rYCount		! or, move back off the image.
rNIShiftSec1SrcAddrMod:
	st	rOut0, [rDstAddr]		! store output word
	add	rDstAddr, 4, rDstAddr		! and inc ptr

! are there more words to process in this line?
	sub	rWidth, 32, rWidth		! have done 32 bits
	subcc	rWidth, rBoundary1, %g0		! is rWidth > rBoundary1?
	bg	rNIShiftSection1		! stay on section 1.
	ld	[rMaskTable], rMask		! and get first mask

! Having completed section 1, are we done?
	tst	rWidth
	bg	rNIShiftSection2Entry		! nope, do section 2.
	cmp	rWidth, rBoundary2		! is rWidth > rBoundary2?

! We were only supposed to do section 1, so return now
	ret
	restore

! ************* SECTION 2, SHIFTING REQUIRED ***********

! WE SHOULD SPEND MOST OF OUR TIME IN THIS SECTION
rNIShiftSection2Entry:
	ble	rNIShiftSection3Entry		! no, go to section 3
	tst	rWidth				! and see if we want to

	ld	[rMaskTable], rMask		! and get first mask

rNIShiftSection2:
	add	rMaskTable, 4, rMaskTable
	ld	[rSrcAddr], rInput		! always use real input word
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	sll	rInput, rLeftShift, rInput	! shift into position

! Keep inserting runs until we've filled the left portion of this word.
	andcc	rMask, rRightmostBit, %g0	! have we touched the
						! LS bit of left portion?
	bnz	rNIShiftSec2ToRightHalf		! yes, finish up
	and	rInput, rMask, rOut0		! initialize output word
rNIShiftSection2b:
	ld	[rMaskTable], rMask		! get next mask
	add	rMaskTable, 4, rMaskTable
	ld	[rSrcAddr], rInput		! always use real input word
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	sll	rInput, rLeftShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	andcc	rMask, rRightmostBit, %g0	! have we touched the
						! LS bit of left portion?
	bz	rNIShiftSection2b		! no, keep going
	or	rOut0, rInput, rOut0		! and insert this run

! We're at the end of the left portion of the output
! word.  If the most recent mask overlaps the leftmost
! word of the next input, we should go back down a
! line, move over 1 word and finish filling the mask
! Otherwise, just move over 1 word.
rNIShiftSec2ToRightHalf:
	add	rSrcAddr, 4, rSrcAddr		! move input over
	andcc	rMask, rLeftmostBit, %g0	! does the run continue?
	bz	rNIShiftSec2RightHalfEntry	! no, just do right half
	add	rXCount, 32, rXCount

! The previous run continued here, so continue it with the word that's
! below us.  Note that if the run went beyond the end of the output word,
! we would not insert any more runs.
	ld	[rSrcAddr+rSrcPitch], rInput	! always use real input word
rNIShiftSec2ContRunInput:
	srl	rInput, rRightShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it

! while the mask doesn't touch the LS bit of rOut0, keep putting runs into
! the right half of rOut0.
rNIShiftSec2RightHalfEntry:
	andcc	rMask, 0x1, %g0			! touched LS bit?
	bnz	rNIShiftSec2Fin			! yes, we're done
	ld	[rMaskTable], rMask		! get next mask in either case

rNIShiftSec2RightHalf:
	add	rMaskTable, 4, rMaskTable
	ld	[rSrcAddr], rInput		! always use real input word
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	srl	rInput, rRightShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it
	andcc	rMask, 0x1, %g0			! have we touched the
						! LS bit of the word?
	bz	rNIShiftSec2RightHalf		! no, keep going
	ld	[rMaskTable], rMask		! and get next mask

! We've finished one word in the shifting case.  If the next mask
! entry equals sBpl, update rSrcAddr accordingly,
rNIShiftSec2Fin:
	add	rMaskTable, 4, rMaskTable	! fill interlock
	add	rSrcAddr, rMask, rSrcAddr
	st	rOut0, [rDstAddr]		! store output word
	add	rDstAddr, 4, rDstAddr		! and inc ptr

! are there more words to process in this line?
	sub	rWidth, 32, rWidth		! have done 32 bits
	subcc	rWidth, rBoundary2, %g0		! is rWidth > rBoundary2?
	bg	rNIShiftSection2		! stay in section 2
	ld	[rMaskTable], rMask		! and get first mask

! Having completed section 2, are we done?
	tst	rWidth
	bg	rNIShiftSection3Entry		! nope, do section 3.
	tst	rWidth				! done now?

! We were supposed to stop after section 2, so return now
	ret
	restore


! ************* SECTION 3, SHIFTING REQUIRED ***********
rNIShiftSection3Entry:
	ble	rNIShiftSection3End		! yes, skip section 3.
	ld	[rMaskTable], rMask		! and get first mask

rNIShiftSection3:
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, %g0, %g0		! is xCount negative?
	bneg	rNIShiftSec3aOffInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNIShiftSec3aOffInput		! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNIShiftSec3aOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	sll	rInput, rLeftShift, rInput	! shift into position

! Keep inserting runs until we've filled the left portion of this word.
	andcc	rMask, rRightmostBit, %g0	! have we touched the
						! LS bit of left portion?
	bnz	rNIShiftSec3ToRightHalf		! yes, finish up
	and	rInput, rMask, rOut0		! initialize output word
rNIShiftSection3b:
	ld	[rMaskTable], rMask		! get next mask
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, %g0, %g0		! is xCount negative?
	bneg	rNIShiftSec3bOffInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNIShiftSec3bOffInput		! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNIShiftSec3bOffInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	sll	rInput, rLeftShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	andcc	rMask, rRightmostBit, %g0	! have we touched the
						! LS bit of left portion?
	bz	rNIShiftSection3b		! no, keep going
	or	rOut0, rInput, rOut0		! and insert this run

! We're at the end of the left portion of the output
! word.  If the most recent mask overlaps the leftmost
! word of the next input, we should go back down a
! line, move over 1 word and finish filling the mask
! Otherwise, just move over 1 word.
rNIShiftSec3ToRightHalf:
	add	rSrcAddr, 4, rSrcAddr		! move input over
	andcc	rMask, rLeftmostBit, %g0	! does the run continue?
	bz	rNIShiftSec3RightHalfEntry	! no, just do right half
	add	rXCount, 32, rXCount

! The previous run continued here, so continue it with the word that's
! below us.  Note that if the run went beyond the end of the output word,
! we would not insert any more runs.
	orcc	rXCount, %g0, %g0		! is xCount negative?
	bneg	rNIShiftSec3ContRunInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNIShiftSec3ContRunInput	! not near, use entire src wd
	ld	[rSrcAddr+rSrcPitch], rInput	! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNIShiftSec3ContRunInput:
	srl	rInput, rRightShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it

! while the mask doesn't touch the LS bit of rOut0, keep putting runs into
! the right half of rOut0.
rNIShiftSec3RightHalfEntry:
	andcc	rMask, 0x1, %g0			! touched LS bit?
	bnz	rNIShiftSec3Fin			! yes, we're done
	ld	[rMaskTable], rMask		! get next mask in either case

rNIShiftSec3RightHalf:
	add	rMaskTable, 4, rMaskTable
	orcc	rXCount, %g0, %g0		! is xCount negative?
	bneg	rNIShiftSec3RightHalfInput
	mov	rWhiteVal, rInput		! yes, use whiteVal
! no, see if we're close to the right edge of the image.
	addcc	rXCount, 32, %g0
	bpos	rNIShiftSec3RightHalfInput	! not near, use entire src wd
	ld	[rSrcAddr], rInput		! no, use real input word
	or	rInput, rRightWhiteVal, rInput	! yes, fill in white stuff

rNIShiftSec3RightHalfInput:
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! move up a line
	srl	rInput, rRightShift, rInput	! shift into position
	and	rInput, rMask, rInput		! select input section
	or	rOut0, rInput, rOut0		! and insert it
	andcc	rMask, 0x1, %g0			! have we touched the
						! LS bit of the word?
	bz	rNIShiftSec3RightHalf		! no, keep going
	ld	[rMaskTable], rMask		! and get next mask

! We've finished one word in the shifting case.  If the next mask
! entry equals sBpl, increment yCount because we weren't supposed to have
! moved up at the end of the last run (because the end of the word wasn't
! really the end of the run).  Also update rSrcAddr accordingly,
rNIShiftSec3Fin:
	add	rMaskTable, 4, rMaskTable	! pt to next mask
	add	rSrcAddr, rMask, rSrcAddr
	st	rOut0, [rDstAddr]		! store output word
	add	rDstAddr, 4, rDstAddr		! and inc ptr

! are there more words to process in this line?
	subcc	rWidth, 32, rWidth		! have done 32 bits
	bg	rNIShiftSection3		! yes, stay on section 3.
	ld	[rMaskTable], rMask		! and get first mask

! Having completed section 3, we are done, so return now
rNIShiftSection3End:
	ret
	restore


! -----------------------------------------------------------------------------
! _ip_rotateAssemblyCodeIsPresent Routine
! -----------------------------------------------------------------------------

! This routine dutifully saves and restores its one dummy
! parameter.  It just puts a 1 into %o0 and returns.

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_rotateAssemblyCodeIsPresent)
MK_GLOBAL_NAME(ip_rotateAssemblyCodeIsPresent):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPStateIn,  [%fp+ipStateOffset]
	mov	1, %i0

	ret
	restore


/*

$Log: ssnirot.s,v $
Revision 1.1.1.1  1999/09/20 18:18:19  jlin
Imported sources

Revision 1.8  1998/01/13 23:57:51  ddavies
ssnirot.s -> /product/ipcore/ipshared/src/RCS/ssnirot.s,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.7  1996/10/05 02:28:31  ddavies
ssnirot.s -> /product/ipcore/ipshared/src/RCS/ssnirot.s,v

Make sort of ugly changes to the semantics of xCount and yCount.  yCount now
is changed when we cross the lower boundary of the image so it's negative
whenever it's outside the image.  This accounts for the fact that it might
cross both the upper and lower boundaries.

Revision 1.6  1995/09/21 03:02:02  ddavies
ssnirot.s -> /product/ipcore/ipshared/src/RCS/ssnirot.s,v

Fixed a bug in ip_rotateNoInterpLine and added the rightWhiteVal arg.
The bug appeared in images whose rightmost pixel did not appear on a
word boundary.  In this case, we sometimes read the entire word (including
frame) containing the right pixel and put it into the output image (causing
frame to appear in the midst of an image).  We also sometimes ignored the
rightmost word entirely, causing any pixels in that word to be lost.
The calling routines now calculate xCount correctly, fixing the bug that
caused us to ignore the rightmost word.  The rightWhiteVal arg contains
the value of white in the pixels that would otherwise contain right frame
in the rightmost word of each input image line.  It is ORed with the
rightmost word whenever it's read, inserting white in place of the frame.

# Revision 1.5  1993/12/04  05:00:35  ddavies
# ssnirot.s -> /product/ipcore/ipshared/src/RCS/ssnirot.s,v
#
# Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
#
# Revision 1.4  1993/11/05  02:25:23  ddavies
# ssnirot.s -> /product/ipcore/ipshared/src/RCS/ssnirot.s,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

*/

