/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: ssorthrt.s
 *
 * This file contains assembly code to do orthogonal (90, 180, 270 degree)
 * rotations.
 *
 * 		ip_rotateBlock90	rotates a row of 8x32 source
 *					binary rectangles
 *					to a column of 32x8 dest rectangles
 * 		ip_fastRotate180	rotates an entire source image
 *					180 degrees using double writes.
 * 		ip_rotateBlock270	rotates a row of 8x32 source
 *					binary rectangles
 *					to a column of 32x8 dest rectangles
 *
 *
 * Rotation on the SPARC is cute.  The C implementation in sorthrt.c uses
 * a lookup table to spread each input byte out into a 32 bit word.  After
 * doing this for 4 vertical input bytes, we have the leftmost bit of each
 * byte in the leftmost 4 bits of the accumulated word, and so on.  That's
 * done most of a rotate.  That code does this process a couple of times,
 * then forms full bytes from the nibbles stored in two accumulated words.
 *
 * This is very clever, but just a bit too clever for the SPARC.  While it
 * moves things around in groups of 8 or 16 bits at a time, it has a few
 * too many moves.  Can we simplify things?
 *
 * The obvious approach to simplifying life is to read a source reg and
 * shift each bit into a different output reg.  Say we read something
 * 16 bits wide.  We'd shift into 16 output registers.  However, we need
 * a double shift, or equivalent, to do this.  When done, we'd write the
 * output regs.
 *
 * The SPARC doesn't have a double shift, but it does have a mutiply step
 * opcode.  The opcode has some warts.  First, it requires one of the
 * operands to be in the Y register, which must be explicilty loaded from
 * one of the other registers.  Second, it's wonderfully clever about
 * copying the sign of the previous result into the current partial product
 * as the product is right shifted.  This is swell for multiplication, but
 * a pain for a straight double shift.
 *
 * What does the multiply-step opcode really do?  See page 6-81 of the
 * Cypress SPARC RISC User's Guide, or equivalent.  The operations are:
 *
 *	op1 = (n XOR v)..r[rs1]<31:1>	; shift partial product right
 *					; and fill hi bit with sign bit
 *					; from previous result
 *	if (Y<0> == 0)			; if next mulitplier bit = 0,
 *	    op2 = 0;			; op2 = 0
 *	else
 *	    op2 = r[rs2] or sign extend[simm13] ; op2 = multiplicand
 *	r[rd] = op1 + op2		; form partial product
 *	Y = r[rs1]<0>..Y<31:1>		; shift Y right, Old multiplier
 *					; bit falls off the right and
 *					; retired partial product bit
 *					; arrives from the left.
 *	    ; set the flags for next time
 *	n = r[rd]<31>			; set sign bit by sign of last
 *					; partial product
 *	if (r[rd] == 0)			; set zero flag
 *	    z = 1;
 *	else
 *	    z = 0;
 *
 *	    ; SET the overflow bit.  This will be used
 *	    ; in the next iteration
 *	v = ( op1<31> &  op2<31> & ~r[rd]<31>) | ; (pos+pos = neg) => overflow
 *	    (~op1<31> & ~op2<31> &  r[rd]<31>) ;  (neg+neg = pos) => overflow
 *
 *	c = (op1<31> & op2<31>) |	; carry flags is normal
 *	    ((op1<31> | op2<31>) & r[rd]<31>
 *
 * The most obvious way to use this opcode is to read a column of source
 * words or bytes into a bunch of registers.  We would then shift one bit
 * from each register into the Y register by specifying each register in turn
 * as r[rs1].  Keen, except we don't have enough registers.  Since writes
 * on the SS2 are so slow, it's really important to write at least 32 bits
 * at a time.  Say we want to read a column 32 bits wide from the source.
 * after one set of shifts, we'd have 32 partially digested inputs and
 * 1 completely formed destinations - a total of 33 things in registers.
 * We only have 26 (or so) registers to play with.  So, say we read only
 * 16 bits from the source at a time - this requires at least 16 + 16
 * registers since we have to save the results of the first 16x16 square
 * must be saved while we rotate the second one.  Still too many.  We're
 * reduced to reading bytes from the source. This isn't horrible, but
 * could certainly be better.
 *
 * The obvious problem is that we have too many partial source and
 * dest registers.  Say we read each input into the Y register and produce
 * a bunch of output registers from that.  We could completely retire
 * each input as it was encountered, so would need only 16 (say) dest
 * registers plus a few address and counter regs.  Great!  Except for
 * the little problem about moving bits from the Y register to the
 * destinations.
 *
 * To do this, we'd like the bit falling off the right end of the Y
 * register to land in the MS bit of the destination register.  All we
 * have to do is us a constant 0x80000000 in r[rs2] and use the same
 * register for rs1 and rd, right?  Almost.  Notice that the very first
 * thing we do is shift rs1 right by 1 to make room for the new bit
 * (which is fine), than fill the empty bit with the sign of the previous
 * addition.  If we use 0x80000000 as the multiplicand, a lot of the
 * operations will produce negative results, which means a lot of other
 * dest regs will have 1's shifted into their MS bits.  Yuck.
 *
 * Our solution to this problem is to use a multiplicand of 0x40000000 for
 * the first 31 times through the loop.  When we're done with that,
 * each of the dest registers will have 31 valid bits in <30:0>.  In
 * the last iteration, we add each dest register to itself (setting cc's)
 * just before doing the multiply step.  This both shifts the dest left
 * by 1 so the multcc's right shift will do the right thing AND it sets
 * both the n and v flags to 0.  When mulscc shifts the dest right, it will
 * insert n XOR v or 0.  Yea!
 *
 * We can do even better.  When a source halfword containing only 0's is
 * read, we only need to shift all the dest regs right by 1.  That's no
 * gain, shifting right takes the same time as a mulscc.  But, if there
 * are two all-zero words in a row, we can do one shift instead of
 * two mulscc's, saving time.  In fact, any run of of all-zero
 * source words turns into a single shift.
 *
 * There are several states we could be in.  rCount is initialized to 31.
 * 1. rCount > 0 and source word is not blank.
 *	A. Distribute bits of source word to dest words.
 *	B. Decrement rCount.  If rCount went to 0, go to state 3.
 *	C. If next source word != 0, stay in state 1, else set rSkipCount
 *	   to 1 and go to state 2.
 * 2. Have just seen a blank source word and rCount > 0.
 *	A. Decrement rCount.  If rCount went to 0, go to state 4.
 *	B. If next source word != 0, shift all dest words by rSkipCount
 *	   and go to state 1 where the source bits will be distributed
 *	   among the dest words.
 *	C. If next source word == 0, increment rSkipCount and stay in state 2.
 * 3. rCount has just gone to 0 and dest words do not need shifting.
 *	A. If next source word is blank and image depth = 1, we're done
 *	   (MS bit(s) already 0) so go to state 6.  If image depth is > 1,
 *	   shift right by depth-1 and go to state 6.
 *	B. If next source word is not blank, to go state 5.
 * 4. rCount has just gone to 0 and dest words need shifting.
 *	Originally, I had code here to detect the case where the entire
 *	block was blank.  It didn't improve performance appreciably, so
 *	it's gone.
 *	A. If next source word is blank, shift all
 *	   dest words by shiftCount and go to state 6.
 *	C. If next source word is not blank, shift all dest words by
 *	   shiftCount and go to state 5.
 * 5. Insert last source word into MS bits of all dest words.
 *	A. Do the trick of adding each dest to itself before it's mulscc.
 * 6. Write all dest words to memory.  Originally, I had code that would
 *	only write non-zero words to memory.  That didn't increase
 *	performance noticably but did require someone to clear the image
 *	before calling this routine, so forget it.  Writing every word
 *	also makes this code a comparable alternative to hardware assist.
 *
 *
 * ******************* 2 bit/pixel code ***********************
 *
 * Originally, I was just going to do everything twice.  We'd read 32 bits
 * from the source rectangles, distribute them into 16 destination registers
 * using 2 mulscc ops per register.  That's fine, except when we start
 * thinking about skipping white space.  The bits we shift in should
 * be 1's, not 0's.  We could do arithmetic right shifts, except the MS
 * bit's are all 0's, not 1's, right?
 *
 * Well, maybe we can make this work.  Say we guarantee that the MS
 * bit of each partial result is always 1.  When the mulscc shifts the
 * partial result right by 1 before inserting the new bit, the 1 will be
 * duplicated.  Say the constant we add is 0xC0000000.  The constant is
 * not added, bit 30 of the partial result will remain a 1.  If the constant
 * is added, bit 30 will go to 0, but the sign bit will remain set.
 * This will result in a stable situation.  Therefore, we must put
 * the inverse of the partial product into the Y register so 1's in the
 * input become 0's in the Y register, which turn into 1's (because there
 * was NO add!) in the destination registers.  The real advantage is that
 * we can use arithmetic right shifts to fill in the bit positions we've
 * skipped.
 *
 * We'll still insert new bits into bit 30 of the destination register.
 * Why not bit 29?  Because that would require extra opcodes to insert the
 * final pixels.  If we inserting new bits into bit 29, the dest registers
 * would be full (except for bits [31:30]) after 15 pixels.  To insert the
 * last pixel into each register, we would have to shift that register
 * left by 1, set the insertMask to 0x40000000, insert the LS bit of the
 * new pixel, then shift the register left by 1 again, change the
 * insertMask to 0x80000000 and insert the MS bit.  If we insert all the
 * bits into bit position 30, we just insert the LS bit, then do one left
 * shift, change the insertMask register to 0x80000000 and insert the MS bit.
 *
 *
 * BUG FIX:  One must clear the n and v flags just before performing the
 * multiply step opcodes.  I had done this by adding source address plus
 * source pitch, knowing that source addresses are never negative.  Ooops.
 * When using shared memory (at least in Solaris), memory addresses are
 * negative.  Use and explicit clear of the n and v flags now.
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssorthrt.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:19 $"
	.align	4

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPStateIn	%i0
#define	rSrcAddrIn	%i1
#define rSrcPitchIn	%i2
#define rDstAddrIn	%i3
#define rDstPitchIn	%i4
#define rRotTableIn	%i5		/* unused in this implementation */

#define	rSrcAddr	%i0		/* current addr in source image */
#define rSrcPitch	%i1		/* source image's bytes/line */
#define rDstAddr	%i2		/* current addr in dest image */
#define rDstPitch	%i3		/* dest image's bytes/line */
#define rInput		%i4		/* current input word */
#define rBlkCount	%i5		/* number of src blocks to go. */

#define	rDest0		%g1		/* top dest word in rot 90, bottom
					 * dest word in rot 270 */
#define	rDest1		%g2
#define rDest2		%g3
#define rDest3		%g4
#define	rDest4		%g5

#define	rDest5		%o0
#define rDest6		%o1
#define	rDest7		%o2
#define rDest8		%o3
#define	rDest9		%o4
#define	rDest10		%o5

#define	rDest11		%l0
#define	rDest12		%l1
#define rDest13		%l2
#define rDest14		%l3
#define rDest15		%l4		/* bottom dest word in rot 90, top
					 * dest word in 270 */

#define	rInsertBit	%l5		/* contains 0x40000000 or
					 * 0x80000000 */
#define rCount		%l6		/* number of src rows within
					 * a block */
#define rSkipCount	%l7		/* number of source words full of
					 * 0's found = number of columns
					 * to skip in dest. */

/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	srcAddrOffset	0x48
#define	srcPitchOffset	0x4c
#define	dstAddrOffset	0x50
#define	dstPitchOffset	0x54
#define rotTableOffset	0x58		/* unused in this implementation */
#define	blkCountOffset	0x5c
#define srcAddrFixupOffset 0x60
#define	depthOffset	0x64

/* height of source rectangle */
#define	cSrcRectHeight1	32		/* for 1 bit/pixel */
#define cSrcRectHeight2 16		/* for 2 bits/pixel */

/* magical C Frame constant (no local variables) */
#define	LF26		-64

LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! _ip_rotateBlock90 Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_rotateBlock90)
MK_GLOBAL_NAME(ip_rotateBlock90):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPStateIn,  [%fp+ipStateOffset]
	st	rSrcAddrIn,  [%fp+srcAddrOffset]
	st	rSrcPitchIn, [%fp+srcPitchOffset]
	st	rDstAddrIn,  [%fp+dstAddrOffset]
	st	rDstPitchIn, [%fp+dstPitchOffset]

	ld	[%fp+srcAddrOffset],  rSrcAddr
	ld	[%fp+srcPitchOffset], rSrcPitch
	ld	[%fp+dstAddrOffset],  rDstAddr
	ld	[%fp+dstPitchOffset], rDstPitch
	ld	[%fp+blkCountOffset], rBlkCount
! leave the srcAddrFixup in memory, we'll fetch it at the bottom of the
! loop.

! choose execution code based on depth of image
	ld	[%fp+depthOffset], rInput
	mov	0, rDest14
	cmp	rInput, 1
	bne	rB902Entry
	mov	0, rDest15


! **************** ROTATE 1 bit/pixel IMAGE 90 DEGREES *****************

! clear the accumulator regs.  We really only need to clear the MS bit
! of each reg, but it's just as easy to clear the entire things.
	mov	0, rDest0
	mov	0, rDest1
	mov	0, rDest2
	mov	0, rDest3
	mov	0, rDest4
	mov	0, rDest5
	mov	0, rDest6
	mov	0, rDest7
	mov	0, rDest8
	mov	0, rDest9
	mov	0, rDest10
	mov	0, rDest11
	mov	0, rDest12
	mov	0, rDest13

! START OF LOOP THROUGH THE BLOCKS (rectangles) ON THIS ROW

rB901BlockLoop:
	mov	cSrcRectHeight1-1, rCount	! we do loop 31 times.
	sethi	%hi(0x40000000), rInsertBit

! Read the first word ahead of time so we can have it ready for
! loading into the Y reg.  We have to wait 3 instructions after loading
! the Y register before doing the first mulscc.  The Cypress
! SPARC is weird.
	lduh	[rSrcAddr], rInput
	add	rSrcAddr, rSrcPitch, rSrcAddr	! fills interlock
	tst	rInput				! is it blank?
	bz	rB901State2			! yes, process blank words
	mov	0, rSkipCount			! and remember to skip
						! this word

! This section of code distributes the bits of a non-blank source word
! to all the dest words.  It also fetches the next source word.
	mov	rInput, %y
	lduh	[rSrcAddr], rInput
rB901State1:
	add	rSrcAddr, rSrcPitch, rSrcAddr
	xorcc	%g0, %g0, %g0			! clear the n and v flags
						! for mulscc opcodes.

! rotate one src row into one dest col
	mulscc	rDest15, rInsertBit, rDest15	! bit in upper right goes
	mulscc	rDest14, rInsertBit, rDest14	! to lower right.
	mulscc	rDest13, rInsertBit, rDest13
	mulscc	rDest12, rInsertBit, rDest12
	mulscc	rDest11, rInsertBit, rDest11
	mulscc	rDest10, rInsertBit, rDest10
	mulscc	rDest9, rInsertBit, rDest9
	mulscc	rDest8, rInsertBit, rDest8
	mulscc	rDest7, rInsertBit, rDest7
	mulscc	rDest6, rInsertBit, rDest6
	mulscc	rDest5, rInsertBit, rDest5
	mulscc	rDest4, rInsertBit, rDest4
	mulscc	rDest3, rInsertBit, rDest3
	mulscc	rDest2, rInsertBit, rDest2
	mulscc	rDest1, rInsertBit, rDest1
	mulscc	rDest0, rInsertBit, rDest0
	mov	rInput, %y			! do this early!
	deccc	rCount				! are we done with first
	bz	rB901State3			! 31 cols?
	tst	rInput				! is next word blank?
	bnz	rB901State1			! no, distribute its bits
	lduh	[rSrcAddr], rInput

! In this state, we just accumulate blank words and increment rSkipCount
! for each one.  If we find a non-blank word, we shift all the dest
! words by rSkipCount before going back to rB901State1.  If we run out of
! source words before finding a blank one, we go to state 4.
	mov	0, rSkipCount
rB901State2:
	lduh	[rSrcAddr], rInput		! get a new source word
	add	rSrcAddr, rSrcPitch, rSrcAddr
	deccc	rCount				! are we done with first
	bz	rB901State4			! 31 cols?
	tst	rInput				! is next word blank?
	bz	rB901State2
	inc	rSkipCount			! here's the increment that's
						! done even if only one
						! blank word was seen.
! shift all the dest words by rSkipCount before going back to state 1.
	srl	rDest15, rSkipCount, rDest15	! bit in upper right goes
	srl	rDest14, rSkipCount, rDest14	! to lower right.
	srl	rDest13, rSkipCount, rDest13
	srl	rDest12, rSkipCount, rDest12
	srl	rDest11, rSkipCount, rDest11
	srl	rDest10, rSkipCount, rDest10
	srl	rDest9, rSkipCount, rDest9
	srl	rDest8, rSkipCount, rDest8
	srl	rDest7, rSkipCount, rDest7
	srl	rDest6, rSkipCount, rDest6
	srl	rDest5, rSkipCount, rDest5
	srl	rDest4, rSkipCount, rDest4
	srl	rDest3, rSkipCount, rDest3
	srl	rDest2, rSkipCount, rDest2
	srl	rDest1, rSkipCount, rDest1
	srl	rDest0, rSkipCount, rDest0
	mov	rInput, %y
	ba	rB901State1
	lduh	[rSrcAddr], rInput


! In this state, we've just finished distributing the bits of source row
! 31 (dest col 31).  Is the final source word blank?
rB901State3:
	bz	rB901State6			! yes, just write out
						! dest words now because
						! the MS bits of all the
						! dest words are already
						! blank.
	sll	rInsertBit, 1, rInsertBit	! shift to 0x80000000
	ba	rB901State5a			! fill the MS bit
	addcc	rDest15, rDest15, rDest15


! In this state, we've just finished a run of blank words ending at
! source row 31.  Do the final shift before inserting last source word.
rB901State4:
	inc	rSkipCount			! got here with rSkipCount
						! 1 too small
	srl	rDest15, rSkipCount, rDest15	! bit in upper right goes
	srl	rDest14, rSkipCount, rDest14	! to lower right.
	srl	rDest13, rSkipCount, rDest13
	srl	rDest12, rSkipCount, rDest12
	srl	rDest11, rSkipCount, rDest11
	srl	rDest10, rSkipCount, rDest10
	srl	rDest9, rSkipCount, rDest9
	srl	rDest8, rSkipCount, rDest8
	srl	rDest7, rSkipCount, rDest7
	srl	rDest6, rSkipCount, rDest6
	srl	rDest5, rSkipCount, rDest5
	srl	rDest4, rSkipCount, rDest4
	srl	rDest3, rSkipCount, rDest3
	srl	rDest2, rSkipCount, rDest2
	srl	rDest1, rSkipCount, rDest1
	srl	rDest0, rSkipCount, rDest0
	mov	rInput, %y			! set up for final word
	tst	rInput				! is there anything in the
						! last word?
	bz	rB901State6			! no, just write what we've
						! got
	sll	rInsertBit, 1, rInsertBit	! shift to 0x80000000


! Rotate the final source word into the leftmost column.
rB901State5:
	addcc	rDest15, rDest15, rDest15
rB901State5a:
	mulscc	rDest15, rInsertBit, rDest15
	addcc	rDest14, rDest14, rDest14
	mulscc	rDest14, rInsertBit, rDest14
	addcc	rDest13, rDest13, rDest13
	mulscc	rDest13, rInsertBit, rDest13
	addcc	rDest12, rDest12, rDest12
	mulscc	rDest12, rInsertBit, rDest12
	addcc	rDest11, rDest11, rDest11
	mulscc	rDest11, rInsertBit, rDest11
	addcc	rDest10, rDest10, rDest10
	mulscc	rDest10, rInsertBit, rDest10
	addcc	rDest9, rDest9, rDest9
	mulscc	rDest9, rInsertBit, rDest9
	addcc	rDest8, rDest8, rDest8
	mulscc	rDest8, rInsertBit, rDest8
	addcc	rDest7, rDest7, rDest7
	mulscc	rDest7, rInsertBit, rDest7
	addcc	rDest6, rDest6, rDest6
	mulscc	rDest6, rInsertBit, rDest6
	addcc	rDest5, rDest5, rDest5
	mulscc	rDest5, rInsertBit, rDest5
	addcc	rDest4, rDest4, rDest4
	mulscc	rDest4, rInsertBit, rDest4
	addcc	rDest3, rDest3, rDest3
	mulscc	rDest3, rInsertBit, rDest3
	addcc	rDest2, rDest2, rDest2
	mulscc	rDest2, rInsertBit, rDest2
	addcc	rDest1, rDest1, rDest1
	mulscc	rDest1, rInsertBit, rDest1
	addcc	rDest0, rDest0, rDest0
	mulscc	rDest0, rInsertBit, rDest0

! Write each dest word.  We use the already-written rDestx regs
! to address lower locations.  Why?  Well, the next time through the
! loop, we would have to initialize their MS bits anyway so subsequent
! mulscc ops would have clear sign bits.  This kills two birds with
! one stone.
! Originally, there was code here that would avoid writing words that
! were blank.  It didn't increase performance appreciably, but did require
! that all incoming images be blank.  We removed it.
rB901State6:
	st	rDest0, [rDstAddr]
	mov	rDstPitch, rDest0		! will use in later writes
						! and initializes MSbit
	st	rDest1, [rDstAddr + rDest0]
	add	rDest0, rDstPitch, rDest1
	st	rDest2, [rDstAddr + rDest1]
	add	rDest1, rDstPitch, rDest2
	st	rDest3, [rDstAddr + rDest2]
	add	rDest2, rDstPitch, rDest3
	st	rDest4, [rDstAddr + rDest3]
	add	rDest3, rDstPitch, rDest4
	st	rDest5, [rDstAddr + rDest4]
	add	rDest4, rDstPitch, rDest5
	st	rDest6, [rDstAddr + rDest5]
	add	rDest5, rDstPitch, rDest6
	st	rDest7, [rDstAddr + rDest6]
	add	rDest6, rDstPitch, rDest7
	st	rDest8, [rDstAddr + rDest7]
	add	rDest7, rDstPitch, rDest8
	st	rDest9, [rDstAddr + rDest8]
	add	rDest8, rDstPitch, rDest9
	st	rDest10, [rDstAddr + rDest9]
	add	rDest9, rDstPitch, rDest10
	st	rDest11, [rDstAddr + rDest10]
	add	rDest10, rDstPitch, rDest11
	st	rDest12, [rDstAddr + rDest11]
	add	rDest11, rDstPitch, rDest12
	st	rDest13, [rDstAddr + rDest12]
	add	rDest12, rDstPitch, rDest13
	st	rDest14, [rDstAddr + rDest13]
	add	rDest13, rDstPitch, rDest14
	st	rDest15, [rDstAddr + rDest14]

rB901FinBlkLoop:
	add	rDest14, rDstPitch, rDest15
	ld	[%fp+srcAddrFixupOffset], rInput	! fixup source addr
	add	rDstAddr, rDest15, rDstAddr		! advance rDstAddr up
							! by height of
							! dest block
	deccc	rBlkCount				! done with last block?
	bnz	rB901BlockLoop
	add	rSrcAddr, rInput, rSrcAddr


! END OF LOOP ACROSS ALL ROWS

! We've put out the right number of words, so quit
rB901Fin:
	ret
	restore


! **************** ROTATE 2 bits/pixel IMAGE 90 DEGREES *****************

! START OF LOOP THROUGH THE BLOCKS (rectangles) ON THIS ROW

! Remember that in gray scale, we want to shift in 1's as white space,
! not 0's.  We do this by always having a 1 in the sign bit.
! So set the accumulator regs.  We really only need to set the MS bit
! of each reg, but it's just as easy to clear the entire things.
rB902Entry:
	mov	-1, rDest0
	mov	-1, rDest1
	mov	-1, rDest2
	mov	-1, rDest3
	mov	-1, rDest4
	mov	-1, rDest5
	mov	-1, rDest6
	mov	-1, rDest7
	mov	-1, rDest8
	mov	-1, rDest9
	mov	-1, rDest10
	mov	-1, rDest11
	mov	-1, rDest12
	mov	-1, rDest13
	mov	-1, rDest14
	mov	-1, rDest15

	sethi	%hi(0xC0000000), rInsertBit

rB902BlockLoop:
	mov	cSrcRectHeight2-1, rCount	! we do loop 15 times.

! Read the first word ahead of time so we can have it ready for
! loading into the Y reg.  We have to wait 3 instructions after loading
! the Y register before doing the first mulscc.  The Cypress
! SPARC is weird.
	ld	[rSrcAddr], rInput
	add	rSrcAddr, rSrcPitch, rSrcAddr	! fills interlock
	cmp	rInput, -1			! is it white?
	bz	rB902State2			! yes, process blank words
	mov	0, rSkipCount			! and remember to skip
						! this word

! This section of code distributes the bits of a non-blank source word
! to all the dest words.  It also fetches the next source word.
	wr	rInput, -1, %y			! load inverse of rInput
	ld	[rSrcAddr], rInput
rB902State1:
	add	rSrcAddr, rSrcPitch, rSrcAddr
	orcc	rInsertBit, %g0, %g0		! set n flag without v flag
! rotate one src row (16 2 bit pixels) into one dest col
	mulscc	rDest15, rInsertBit, rDest15	! bit in upper right goes
	mulscc	rDest15, rInsertBit, rDest15	! to lower right.
	mulscc	rDest14, rInsertBit, rDest14
	mulscc	rDest14, rInsertBit, rDest14
	mulscc	rDest13, rInsertBit, rDest13
	mulscc	rDest13, rInsertBit, rDest13
	mulscc	rDest12, rInsertBit, rDest12
	mulscc	rDest12, rInsertBit, rDest12
	mulscc	rDest11, rInsertBit, rDest11
	mulscc	rDest11, rInsertBit, rDest11
	mulscc	rDest10, rInsertBit, rDest10
	mulscc	rDest10, rInsertBit, rDest10
	mulscc	rDest9, rInsertBit, rDest9
	mulscc	rDest9, rInsertBit, rDest9
	mulscc	rDest8, rInsertBit, rDest8
	mulscc	rDest8, rInsertBit, rDest8
	mulscc	rDest7, rInsertBit, rDest7
	mulscc	rDest7, rInsertBit, rDest7
	mulscc	rDest6, rInsertBit, rDest6
	mulscc	rDest6, rInsertBit, rDest6
	mulscc	rDest5, rInsertBit, rDest5
	mulscc	rDest5, rInsertBit, rDest5
	mulscc	rDest4, rInsertBit, rDest4
	mulscc	rDest4, rInsertBit, rDest4
	mulscc	rDest3, rInsertBit, rDest3
	mulscc	rDest3, rInsertBit, rDest3
	mulscc	rDest2, rInsertBit, rDest2
	mulscc	rDest2, rInsertBit, rDest2
	mulscc	rDest1, rInsertBit, rDest1
	mulscc	rDest1, rInsertBit, rDest1
	mulscc	rDest0, rInsertBit, rDest0
	mulscc	rDest0, rInsertBit, rDest0
	wr	rInput, -1, %y			! do this early!
	deccc	rCount				! are we done with first
	bz	rB902State3			! 15 cols?
	cmp	rInput, -1			! is it white?
	bne	rB902State1			! no, distribute its bits
	ld	[rSrcAddr], rInput

! In this state, we just accumulate blank words and increment rSkipCount
! for each one.  If we find a non-blank word, we shift all the dest
! words by rSkipCount before going back to rB902State1.  If we run out of
! source words before finding a blank one, we go to state 4. Note that
! each blank word counts for 2 bit positions in each dest word since
! pixel from that word has 2 bits.
! Note also that the MS bit of each rDestX register is set.  Therefore,
! when we do arithmetic right shifts, we fill the blank pixels with 1's
! as expected.

	mov	0, rSkipCount			! we'll want to use
						! rSkipCount-1 to generate
						! the 1's to be inserted
						! into the output words.
rB902State2:
	ld	[rSrcAddr], rInput		! get a new source word
	add	rSrcAddr, rSrcPitch, rSrcAddr
	deccc	rCount				! are we done with first
	bz	rB902State4			! 15 cols?
	cmp	rInput, -1			! is it white?
	bz	rB902State2
	add	rSkipCount, 2, rSkipCount	! here's the increment that's
						! done even if only one
						! blank word was seen.
! shift all the dest words by rSkipCount before going back to state 1.
	sra	rDest15, rSkipCount, rDest15
	sra	rDest14, rSkipCount, rDest14
	sra	rDest13, rSkipCount, rDest13
	sra	rDest12, rSkipCount, rDest12
	sra	rDest11, rSkipCount, rDest11
	sra	rDest10, rSkipCount, rDest10
	sra	rDest9, rSkipCount, rDest9
	sra	rDest8, rSkipCount, rDest8
	sra	rDest7, rSkipCount, rDest7
	sra	rDest6, rSkipCount, rDest6
	sra	rDest5, rSkipCount, rDest5
	sra	rDest4, rSkipCount, rDest4
	sra	rDest3, rSkipCount, rDest3
	sra	rDest2, rSkipCount, rDest2
	sra	rDest1, rSkipCount, rDest1
	sra	rDest0, rSkipCount, rDest0
	wr	rInput, -1, %y
	ba	rB902State1
	ld	[rSrcAddr], rInput


! In this state, we've just finished distributing the bits of source row
! 31 (dest col 31).  Is the final source word blank?
rB902State3:
	bne	rB902State5			! No, fill the MS bits
	addcc	rInsertBit, rInsertBit, rInput	! the hard way
! Yes, shift all the outputs right by 1.
	sra	rDest15, 1, rDest15		! insert 1 more bit to go
	sra	rDest14, 1, rDest14		! with the 1 that's already
	sra	rDest13, 1, rDest13		! in bit 31.
	sra	rDest12, 1, rDest12
	sra	rDest11, 1, rDest11
	sra	rDest10, 1, rDest10
	sra	rDest9, 1, rDest9
	sra	rDest8, 1, rDest8
	sra	rDest7, 1, rDest7
	sra	rDest6, 1, rDest6
	sra	rDest5, 1, rDest5
	sra	rDest4, 1, rDest4
	sra	rDest3, 1, rDest3
	sra	rDest2, 1, rDest2
	sra	rDest1, 1, rDest1
	ba	rB902State6			! and print this rectangle
	sra	rDest0, 1, rDest0

! In this state, we've just finished a run of blank words ending at
! source row 15.  Do the final shift before inserting last source word.
! If the last word is blank, shift by an extra bit to fill the last word.
! Otherwise, shift by the correct amount, we'll fill in the last pixel
! in state 5.
rB902State4:
	be,a	rB90State4Done			! do an extra shift if the
	add	rSkipCount, 1, rSkipCount	! last word is blank
rB90State4Done:
	add	rSkipCount, 2, rSkipCount	! got here with rSkipCount
						! 2 too small
	sra	rDest15, rSkipCount, rDest15
	sra	rDest14, rSkipCount, rDest14
	sra	rDest13, rSkipCount, rDest13
	sra	rDest12, rSkipCount, rDest12
	sra	rDest11, rSkipCount, rDest11
	sra	rDest10, rSkipCount, rDest10
	sra	rDest9, rSkipCount, rDest9
	sra	rDest8, rSkipCount, rDest8
	sra	rDest7, rSkipCount, rDest7
	sra	rDest6, rSkipCount, rDest6
	sra	rDest5, rSkipCount, rDest5
	sra	rDest4, rSkipCount, rDest4
	sra	rDest3, rSkipCount, rDest3
	sra	rDest2, rSkipCount, rDest2
	sra	rDest1, rSkipCount, rDest1
	sra	rDest0, rSkipCount, rDest0
	wr	rInput, -1, %y			! set up for final word
	cmp	rInput, -1			! is there anything in the
						! last word?
	bz	rB902State6			! no, just write what we've
						! got
	addcc	rInsertBit, rInsertBit, rInput	! shift to 0x80000000 and set
						! n flag for mulscc below.


! Rotate the final source word into the leftmost column.  The n flag
! should already be set from the addcc ops immediately preceeding all
! jumps to here.
rB902State5:
	mulscc	rDest15, rInsertBit, rDest15
	addcc	rDest15, rDest15, rDest15
	mulscc	rDest15, rInput, rDest15
	mulscc	rDest14, rInsertBit, rDest14
	addcc	rDest14, rDest14, rDest14
	mulscc	rDest14, rInput, rDest14
	mulscc	rDest13, rInsertBit, rDest13
	addcc	rDest13, rDest13, rDest13
	mulscc	rDest13, rInput, rDest13
	mulscc	rDest12, rInsertBit, rDest12
	addcc	rDest12, rDest12, rDest12
	mulscc	rDest12, rInput, rDest12
	mulscc	rDest11, rInsertBit, rDest11
	addcc	rDest11, rDest11, rDest11
	mulscc	rDest11, rInput, rDest11
	mulscc	rDest10, rInsertBit, rDest10
	addcc	rDest10, rDest10, rDest10
	mulscc	rDest10, rInput, rDest10
	mulscc	rDest9, rInsertBit, rDest9
	addcc	rDest9, rDest9, rDest9
	mulscc	rDest9, rInput, rDest9
	mulscc	rDest8, rInsertBit, rDest8
	addcc	rDest8, rDest8, rDest8
	mulscc	rDest8, rInput, rDest8
	mulscc	rDest7, rInsertBit, rDest7
	addcc	rDest7, rDest7, rDest7
	mulscc	rDest7, rInput, rDest7
	mulscc	rDest6, rInsertBit, rDest6
	addcc	rDest6, rDest6, rDest6
	mulscc	rDest6, rInput, rDest6
	mulscc	rDest5, rInsertBit, rDest5
	addcc	rDest5, rDest5, rDest5
	mulscc	rDest5, rInput, rDest5
	mulscc	rDest4, rInsertBit, rDest4
	addcc	rDest4, rDest4, rDest4
	mulscc	rDest4, rInput, rDest4
	mulscc	rDest3, rInsertBit, rDest3
	addcc	rDest3, rDest3, rDest3
	mulscc	rDest3, rInput, rDest3
	mulscc	rDest2, rInsertBit, rDest2
	addcc	rDest2, rDest2, rDest2
	mulscc	rDest2, rInput, rDest2
	mulscc	rDest1, rInsertBit, rDest1
	addcc	rDest1, rDest1, rDest1
	mulscc	rDest1, rInput, rDest1
	mulscc	rDest0, rInsertBit, rDest0
	addcc	rDest0, rDest0, rDest0
	mulscc	rDest0, rInput, rDest0

! Write each dest word starting from the bottom of the block.  We
! use the already-written rDestX regs to address higher locations.  Why?
! Well, the next time through the loop, we would have to initialize
! their MS bits anyway so subsequent mulscc ops would have their sign
! bits SET.  This kills two birds with one stone.
! Originally, there was code here that would avoid writing words that
! were blank.  It didn't increase performance appreciably, but did require
! that all incoming images be blank.  We removed it.
rB902State6:
	st	rDest15, [rDstAddr]
	sub	%g0, rDstPitch, rDest15		! will use in later writes
						! and initializes MSbit to 1
	st	rDest14, [rDstAddr + rDest15]
	sub	rDest15, rDstPitch, rDest14
	st	rDest13, [rDstAddr + rDest14]
	sub	rDest14, rDstPitch, rDest13
	st	rDest12, [rDstAddr + rDest13]
	sub	rDest13, rDstPitch, rDest12
	st	rDest11, [rDstAddr + rDest12]
	sub	rDest12, rDstPitch, rDest11
	st	rDest10, [rDstAddr + rDest11]
	sub	rDest11, rDstPitch, rDest10
	st	rDest9, [rDstAddr + rDest10]
	sub	rDest10, rDstPitch, rDest9
	st	rDest8, [rDstAddr + rDest9]
	sub	rDest9, rDstPitch, rDest8
	st	rDest7, [rDstAddr + rDest8]
	sub	rDest8, rDstPitch, rDest7
	st	rDest6, [rDstAddr + rDest7]
	sub	rDest7, rDstPitch, rDest6
	st	rDest5, [rDstAddr + rDest6]
	sub	rDest6, rDstPitch, rDest5
	st	rDest4, [rDstAddr + rDest5]
	sub	rDest5, rDstPitch, rDest4
	st	rDest3, [rDstAddr + rDest4]
	sub	rDest4, rDstPitch, rDest3
	st	rDest2, [rDstAddr + rDest3]
	sub	rDest3, rDstPitch, rDest2
	st	rDest1, [rDstAddr + rDest2]
	sub	rDest2, rDstPitch, rDest1
	st	rDest0, [rDstAddr + rDest1]

rB902FinBlkLoop:
	sub	rDest1, rDstPitch, rDest0
	ld	[%fp+srcAddrFixupOffset], rInput	! fixup source addr
	sub	rDstAddr, rDest0, rDstAddr		! advance rDstAddr
	deccc	rBlkCount				! done will last block?
	bnz	rB902BlockLoop
	add	rSrcAddr, rInput, rSrcAddr


! END OF LOOP ACROSS ALL ROWS

! We've put out the right number of words, so quit
rB902Fin:
	ret
	restore


! -----------------------------------------------------------------------------
! _ip_rotateBlock270 Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_rotateBlock270)
MK_GLOBAL_NAME(ip_rotateBlock270):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPStateIn,  [%fp+ipStateOffset]
	st	rSrcAddrIn,  [%fp+srcAddrOffset]
	st	rSrcPitchIn, [%fp+srcPitchOffset]
	st	rDstAddrIn,  [%fp+dstAddrOffset]
	st	rDstPitchIn, [%fp+dstPitchOffset]

	ld	[%fp+srcAddrOffset],  rSrcAddr
	ld	[%fp+srcPitchOffset], rSrcPitch
	ld	[%fp+dstAddrOffset],  rDstAddr
	ld	[%fp+dstPitchOffset], rDstPitch
	ld	[%fp+blkCountOffset], rBlkCount
! leave the srcAddrFixup in memory, we'll fetch it at the bottom of the
! loop.

! choose execution code based on depth of image
	ld	[%fp+depthOffset], rInput
	mov	0, rDest14
	cmp	rInput, 1
	bne	rB2702Entry
	mov	0, rDest15


! **************** ROTATE 1 bit/pixel IMAGE 270 DEGREES *****************

! clear the accumulator regs.  We really only need to clear the MS bit
! of each reg, but it's just as easy to clear the entire things.
	mov	0, rDest0
	mov	0, rDest1
	mov	0, rDest2
	mov	0, rDest3
	mov	0, rDest4
	mov	0, rDest5
	mov	0, rDest6
	mov	0, rDest7
	mov	0, rDest8
	mov	0, rDest9
	mov	0, rDest10
	mov	0, rDest11
	mov	0, rDest12
	mov	0, rDest13

! START OF LOOP THROUGH THE BLOCKS (rectangles) ON THIS ROW

rB2701BlockLoop:
	mov	cSrcRectHeight1-1, rCount	! we do loop 31 times.
	sethi	%hi(0x40000000), rInsertBit

! Read the first word ahead of time so we can have it ready for
! loading into the Y reg.  We have to wait 3 instructions after loading
! the Y register before doing the first mulscc.  The Cypress
! SPARC is weird.
	lduh	[rSrcAddr], rInput
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! fills interlock
	tst	rInput				! is it blank?
	bz	rB2701State2			! yes, process blank words
	mov	0, rSkipCount			! and remember to skip
						! this word

! This section of code distributes the bits of a non-blank source word
! to all the dest words.  It also fetches the next source word.
	mov	rInput, %y
	lduh	[rSrcAddr], rInput
rB2701State1:
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	xorcc	%g0, %g0, %g0			! and clear n and v for
						! mulscc below.

! rotate one src row into one dest col
	mulscc	rDest15, rInsertBit, rDest15	! bit in lower right goes
	mulscc	rDest14, rInsertBit, rDest14	! to upper right.
	mulscc	rDest13, rInsertBit, rDest13
	mulscc	rDest12, rInsertBit, rDest12
	mulscc	rDest11, rInsertBit, rDest11
	mulscc	rDest10, rInsertBit, rDest10
	mulscc	rDest9, rInsertBit, rDest9
	mulscc	rDest8, rInsertBit, rDest8
	mulscc	rDest7, rInsertBit, rDest7
	mulscc	rDest6, rInsertBit, rDest6
	mulscc	rDest5, rInsertBit, rDest5
	mulscc	rDest4, rInsertBit, rDest4
	mulscc	rDest3, rInsertBit, rDest3
	mulscc	rDest2, rInsertBit, rDest2
	mulscc	rDest1, rInsertBit, rDest1
	mulscc	rDest0, rInsertBit, rDest0
	mov	rInput, %y			! do this early!
	deccc	rCount				! are we done with first
	bz	rB2701State3			! 31 cols?
	tst	rInput				! is next word blank?
	bnz	rB2701State1			! no, distribute its bits
	lduh	[rSrcAddr], rInput

! In this state, we just accumulate blank words and increment rSkipCount
! for each one.  If we find a non-blank word, we shift all the dest
! words by rSkipCount before going back to rB2701State1.  If we run out of
! source words before finding a blank one, we go to state 4.
	mov	0, rSkipCount
rB2701State2:
	lduh	[rSrcAddr], rInput		! get a new source word
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	deccc	rCount				! are we done with first
	bz	rB2701State4			! 31 cols?
	tst	rInput				! is next word blank?
	bz	rB2701State2
	inc	rSkipCount			! here's the increment that's
						! done even if only one
						! blank word was seen.
! shift all the dest words by rSkipCount before going back to state 1.
	srl	rDest15, rSkipCount, rDest15	! bit in lower right goes
	srl	rDest14, rSkipCount, rDest14	! to upper right.
	srl	rDest13, rSkipCount, rDest13
	srl	rDest12, rSkipCount, rDest12
	srl	rDest11, rSkipCount, rDest11
	srl	rDest10, rSkipCount, rDest10
	srl	rDest9, rSkipCount, rDest9
	srl	rDest8, rSkipCount, rDest8
	srl	rDest7, rSkipCount, rDest7
	srl	rDest6, rSkipCount, rDest6
	srl	rDest5, rSkipCount, rDest5
	srl	rDest4, rSkipCount, rDest4
	srl	rDest3, rSkipCount, rDest3
	srl	rDest2, rSkipCount, rDest2
	srl	rDest1, rSkipCount, rDest1
	srl	rDest0, rSkipCount, rDest0
	mov	rInput, %y
	ba	rB2701State1
	lduh	[rSrcAddr], rInput


! In this state, we've just finished distributing the bits of source row
! 31 (dest col 31).  Is the final source word blank?
rB2701State3:
	bz	rB2701State6			! yes, just write out
						! dest words now because
						! the MS bits of all the
						! dest words are already
						! blank.
	sll	rInsertBit, 1, rInsertBit	! shift to 0x80000000
	ba	rB2701State5a			! fill the MS bit
	addcc	rDest15, rDest15, rDest15


! In this state, we've just finished a run of blank words ending at
! source row 31.  Do the final shift before inserting last source word.
rB2701State4:
	inc	rSkipCount			! got here with rSkipCount
						! 1 too small
	srl	rDest15, rSkipCount, rDest15	! bit in lower right goes
	srl	rDest14, rSkipCount, rDest14	! to upper right.
	srl	rDest13, rSkipCount, rDest13
	srl	rDest12, rSkipCount, rDest12
	srl	rDest11, rSkipCount, rDest11
	srl	rDest10, rSkipCount, rDest10
	srl	rDest9, rSkipCount, rDest9
	srl	rDest8, rSkipCount, rDest8
	srl	rDest7, rSkipCount, rDest7
	srl	rDest6, rSkipCount, rDest6
	srl	rDest5, rSkipCount, rDest5
	srl	rDest4, rSkipCount, rDest4
	srl	rDest3, rSkipCount, rDest3
	srl	rDest2, rSkipCount, rDest2
	srl	rDest1, rSkipCount, rDest1
	srl	rDest0, rSkipCount, rDest0
	mov	rInput, %y			! set up for final word
	tst	rInput				! is there anything in the
						! last word?
	bz	rB2701State6			! no, just write what we've
						! got
	sll	rInsertBit, 1, rInsertBit	! shift to 0x80000000


! Rotate the final source word into the leftmost column.
rB2701State5:
	addcc	rDest15, rDest15, rDest15
rB2701State5a:
	mulscc	rDest15, rInsertBit, rDest15
	addcc	rDest14, rDest14, rDest14
	mulscc	rDest14, rInsertBit, rDest14
	addcc	rDest13, rDest13, rDest13
	mulscc	rDest13, rInsertBit, rDest13
	addcc	rDest12, rDest12, rDest12
	mulscc	rDest12, rInsertBit, rDest12
	addcc	rDest11, rDest11, rDest11
	mulscc	rDest11, rInsertBit, rDest11
	addcc	rDest10, rDest10, rDest10
	mulscc	rDest10, rInsertBit, rDest10
	addcc	rDest9, rDest9, rDest9
	mulscc	rDest9, rInsertBit, rDest9
	addcc	rDest8, rDest8, rDest8
	mulscc	rDest8, rInsertBit, rDest8
	addcc	rDest7, rDest7, rDest7
	mulscc	rDest7, rInsertBit, rDest7
	addcc	rDest6, rDest6, rDest6
	mulscc	rDest6, rInsertBit, rDest6
	addcc	rDest5, rDest5, rDest5
	mulscc	rDest5, rInsertBit, rDest5
	addcc	rDest4, rDest4, rDest4
	mulscc	rDest4, rInsertBit, rDest4
	addcc	rDest3, rDest3, rDest3
	mulscc	rDest3, rInsertBit, rDest3
	addcc	rDest2, rDest2, rDest2
	mulscc	rDest2, rInsertBit, rDest2
	addcc	rDest1, rDest1, rDest1
	mulscc	rDest1, rInsertBit, rDest1
	addcc	rDest0, rDest0, rDest0
	mulscc	rDest0, rInsertBit, rDest0

! Write each dest word.  We use the already-written rDestx regs
! to address lower locations.  Why?  Well, the next time through the
! loop, we would have to initialize their MS bits anyway so subsequent
! mulscc ops would have clear sign bits.  This kills two birds with
! one stone.
! Originally, there was code here that would avoid writing words that
! were blank.  It didn't increase performance appreciably, but did require
! that all incoming images be blank.  We removed it.
rB2701State6:
	st	rDest15, [rDstAddr]
	mov	rDstPitch, rDest15		! will use in later writes
						! and initializes MSbit
	st	rDest14, [rDstAddr + rDest15]
	add	rDest15, rDstPitch, rDest14
	st	rDest13, [rDstAddr + rDest14]
	add	rDest14, rDstPitch, rDest13
	st	rDest12, [rDstAddr + rDest13]
	add	rDest13, rDstPitch, rDest12
	st	rDest11, [rDstAddr + rDest12]
	add	rDest12, rDstPitch, rDest11
	st	rDest10, [rDstAddr + rDest11]
	add	rDest11, rDstPitch, rDest10
	st	rDest9, [rDstAddr + rDest10]
	add	rDest10, rDstPitch, rDest9
	st	rDest8, [rDstAddr + rDest9]
	add	rDest9, rDstPitch, rDest8
	st	rDest7, [rDstAddr + rDest8]
	add	rDest8, rDstPitch, rDest7
	st	rDest6, [rDstAddr + rDest7]
	add	rDest7, rDstPitch, rDest6
	st	rDest5, [rDstAddr + rDest6]
	add	rDest6, rDstPitch, rDest5
	st	rDest4, [rDstAddr + rDest5]
	add	rDest5, rDstPitch, rDest4
	st	rDest3, [rDstAddr + rDest4]
	add	rDest4, rDstPitch, rDest3
	st	rDest2, [rDstAddr + rDest3]
	add	rDest3, rDstPitch, rDest2
	st	rDest1, [rDstAddr + rDest2]
	add	rDest2, rDstPitch, rDest1
	st	rDest0, [rDstAddr + rDest1]

rB2701FinBlkLoop:
	add	rDest1, rDstPitch, rDest0
	ld	[%fp+srcAddrFixupOffset], rInput	! fixup source addr
	sub	rDstAddr, rDest0, rDstAddr		! advance rDstAddr up
							! by height of
							! dest block
	deccc	rBlkCount				! done with last block?
	bnz	rB2701BlockLoop
	add	rSrcAddr, rInput, rSrcAddr


! END OF LOOP ACROSS ALL ROWS

! We've put out the right number of words, so quit
rB2701Fin:
	ret
	restore



! **************** ROTATE 2 bits/pixel IMAGE 270 DEGREES *****************

! START OF LOOP THROUGH THE BLOCKS (rectangles) ON THIS ROW

! Remember that in gray scale, we want to shift in 1's as white space,
! not 0's.  We do this by always having a 1 in the sign bit.
! So set the accumulator regs.  We really only need to set the MS bit
! of each reg, but it's just as easy to clear the entire things.
rB2702Entry:
	mov	-1, rDest0
	mov	-1, rDest1
	mov	-1, rDest2
	mov	-1, rDest3
	mov	-1, rDest4
	mov	-1, rDest5
	mov	-1, rDest6
	mov	-1, rDest7
	mov	-1, rDest8
	mov	-1, rDest9
	mov	-1, rDest10
	mov	-1, rDest11
	mov	-1, rDest12
	mov	-1, rDest13
	mov	-1, rDest14
	mov	-1, rDest15

	sethi	%hi(0xC0000000), rInsertBit

rB2702BlockLoop:
	mov	cSrcRectHeight2-1, rCount	! we do loop 15 times.

! Read the first word ahead of time so we can have it ready for
! loading into the Y reg.  We have to wait 3 instructions after loading
! the Y register before doing the first mulscc.  The Cypress
! SPARC is weird.
	ld	[rSrcAddr], rInput
	sub	rSrcAddr, rSrcPitch, rSrcAddr	! fills interlock
	cmp	rInput, -1			! is it white?
	bz	rB2702State2			! yes, process blank words
	mov	0, rSkipCount			! and remember to skip
						! this word

! This section of code distributes the bits of a non-blank source word
! to all the dest words.  It also fetches the next source word.
	wr	rInput, -1, %y			! load inverse of rInput
	ld	[rSrcAddr], rInput
rB2702State1:
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	orcc	rInsertBit, %g0, %g0		! set n flag without v flag
! rotate one src row (16 2 bit pixels) into one dest col
	mulscc	rDest15, rInsertBit, rDest15	! bit in lower right goes
	mulscc	rDest15, rInsertBit, rDest15	! to upper right.
	mulscc	rDest14, rInsertBit, rDest14
	mulscc	rDest14, rInsertBit, rDest14
	mulscc	rDest13, rInsertBit, rDest13
	mulscc	rDest13, rInsertBit, rDest13
	mulscc	rDest12, rInsertBit, rDest12
	mulscc	rDest12, rInsertBit, rDest12
	mulscc	rDest11, rInsertBit, rDest11
	mulscc	rDest11, rInsertBit, rDest11
	mulscc	rDest10, rInsertBit, rDest10
	mulscc	rDest10, rInsertBit, rDest10
	mulscc	rDest9, rInsertBit, rDest9
	mulscc	rDest9, rInsertBit, rDest9
	mulscc	rDest8, rInsertBit, rDest8
	mulscc	rDest8, rInsertBit, rDest8
	mulscc	rDest7, rInsertBit, rDest7
	mulscc	rDest7, rInsertBit, rDest7
	mulscc	rDest6, rInsertBit, rDest6
	mulscc	rDest6, rInsertBit, rDest6
	mulscc	rDest5, rInsertBit, rDest5
	mulscc	rDest5, rInsertBit, rDest5
	mulscc	rDest4, rInsertBit, rDest4
	mulscc	rDest4, rInsertBit, rDest4
	mulscc	rDest3, rInsertBit, rDest3
	mulscc	rDest3, rInsertBit, rDest3
	mulscc	rDest2, rInsertBit, rDest2
	mulscc	rDest2, rInsertBit, rDest2
	mulscc	rDest1, rInsertBit, rDest1
	mulscc	rDest1, rInsertBit, rDest1
	mulscc	rDest0, rInsertBit, rDest0
	mulscc	rDest0, rInsertBit, rDest0
	wr	rInput, -1, %y			! do this early!
	deccc	rCount				! are we done with first
	bz	rB2702State3			! 15 cols?
	cmp	rInput, -1			! is it white?
	bne	rB2702State1			! no, distribute its bits
	ld	[rSrcAddr], rInput

! In this state, we just accumulate blank words and increment rSkipCount
! for each one.  If we find a non-blank word, we shift all the dest
! words by rSkipCount before going back to rB2702State1.  If we run out of
! source words before finding a blank one, we go to state 4. Note that
! each blank word counts for 2 bit positions in each dest word since
! pixel from that word has 2 bits.
! Note also that the MS bit of each rDestX register is set.  Therefore,
! when we do arithmetic right shifts, we fill the blank pixels with 1's
! as expected.

	mov	0, rSkipCount			! we'll want to use
						! rSkipCount-1 to generate
						! the 1's to be inserted
						! into the output words.
rB2702State2:
	ld	[rSrcAddr], rInput		! get a new source word
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	deccc	rCount				! are we done with first
	bz	rB2702State4			! 15 cols?
	cmp	rInput, -1			! is it white?
	bz	rB2702State2
	add	rSkipCount, 2, rSkipCount	! here's the increment that's
						! done even if only one
						! blank word was seen.
! shift all the dest words by rSkipCount before going back to state 1.
	sra	rDest15, rSkipCount, rDest15
	sra	rDest14, rSkipCount, rDest14
	sra	rDest13, rSkipCount, rDest13
	sra	rDest12, rSkipCount, rDest12
	sra	rDest11, rSkipCount, rDest11
	sra	rDest10, rSkipCount, rDest10
	sra	rDest9, rSkipCount, rDest9
	sra	rDest8, rSkipCount, rDest8
	sra	rDest7, rSkipCount, rDest7
	sra	rDest6, rSkipCount, rDest6
	sra	rDest5, rSkipCount, rDest5
	sra	rDest4, rSkipCount, rDest4
	sra	rDest3, rSkipCount, rDest3
	sra	rDest2, rSkipCount, rDest2
	sra	rDest1, rSkipCount, rDest1
	sra	rDest0, rSkipCount, rDest0
	wr	rInput, -1, %y
	ba	rB2702State1
	ld	[rSrcAddr], rInput


! In this state, we've just finished distributing the bits of source row
! 31 (dest col 31).  Is the final source word blank?
rB2702State3:
	bne	rB2702State5			! No, fill the MS bits
	addcc	rInsertBit, rInsertBit, rInput	! the hard way
! Yes, shift all the outputs right by 1.
	sra	rDest15, 1, rDest15		! insert 1 more bit to go
	sra	rDest14, 1, rDest14		! with the 1 that's already
	sra	rDest13, 1, rDest13		! in bit 31.
	sra	rDest12, 1, rDest12
	sra	rDest11, 1, rDest11
	sra	rDest10, 1, rDest10
	sra	rDest9, 1, rDest9
	sra	rDest8, 1, rDest8
	sra	rDest7, 1, rDest7
	sra	rDest6, 1, rDest6
	sra	rDest5, 1, rDest5
	sra	rDest4, 1, rDest4
	sra	rDest3, 1, rDest3
	sra	rDest2, 1, rDest2
	sra	rDest1, 1, rDest1
	ba	rB2702State6			! and print this rectangle
	sra	rDest0, 1, rDest0

! In this state, we've just finished a run of blank words ending at
! source row 15.  Do the final shift before inserting last source word.
! If the last word is blank, shift by an extra bit to fill the last word.
! Otherwise, shift by the correct amount, we'll fill in the last pixel
! in state 5.
rB2702State4:
	be,a	rB270State4Done			! do an extra shift if the
	add	rSkipCount, 1, rSkipCount	! last word is blank
rB270State4Done:
	add	rSkipCount, 2, rSkipCount	! got here with rSkipCount
						! 2 too small
	sra	rDest15, rSkipCount, rDest15
	sra	rDest14, rSkipCount, rDest14
	sra	rDest13, rSkipCount, rDest13
	sra	rDest12, rSkipCount, rDest12
	sra	rDest11, rSkipCount, rDest11
	sra	rDest10, rSkipCount, rDest10
	sra	rDest9, rSkipCount, rDest9
	sra	rDest8, rSkipCount, rDest8
	sra	rDest7, rSkipCount, rDest7
	sra	rDest6, rSkipCount, rDest6
	sra	rDest5, rSkipCount, rDest5
	sra	rDest4, rSkipCount, rDest4
	sra	rDest3, rSkipCount, rDest3
	sra	rDest2, rSkipCount, rDest2
	sra	rDest1, rSkipCount, rDest1
	sra	rDest0, rSkipCount, rDest0
	wr	rInput, -1, %y			! set up for final word
	cmp	rInput, -1			! is there anything in the
						! last word?
	bz	rB2702State6			! no, just write what we've
						! got
	addcc	rInsertBit, rInsertBit, rInput	! shift to 0x80000000 and set
						! n flag for mulscc below.


! Rotate the final source word into the leftmost column.  The n flag
! should already be set from the addcc ops immediately preceeding all
! jumps to here.
rB2702State5:
	mulscc	rDest15, rInsertBit, rDest15
	addcc	rDest15, rDest15, rDest15
	mulscc	rDest15, rInput, rDest15
	mulscc	rDest14, rInsertBit, rDest14
	addcc	rDest14, rDest14, rDest14
	mulscc	rDest14, rInput, rDest14
	mulscc	rDest13, rInsertBit, rDest13
	addcc	rDest13, rDest13, rDest13
	mulscc	rDest13, rInput, rDest13
	mulscc	rDest12, rInsertBit, rDest12
	addcc	rDest12, rDest12, rDest12
	mulscc	rDest12, rInput, rDest12
	mulscc	rDest11, rInsertBit, rDest11
	addcc	rDest11, rDest11, rDest11
	mulscc	rDest11, rInput, rDest11
	mulscc	rDest10, rInsertBit, rDest10
	addcc	rDest10, rDest10, rDest10
	mulscc	rDest10, rInput, rDest10
	mulscc	rDest9, rInsertBit, rDest9
	addcc	rDest9, rDest9, rDest9
	mulscc	rDest9, rInput, rDest9
	mulscc	rDest8, rInsertBit, rDest8
	addcc	rDest8, rDest8, rDest8
	mulscc	rDest8, rInput, rDest8
	mulscc	rDest7, rInsertBit, rDest7
	addcc	rDest7, rDest7, rDest7
	mulscc	rDest7, rInput, rDest7
	mulscc	rDest6, rInsertBit, rDest6
	addcc	rDest6, rDest6, rDest6
	mulscc	rDest6, rInput, rDest6
	mulscc	rDest5, rInsertBit, rDest5
	addcc	rDest5, rDest5, rDest5
	mulscc	rDest5, rInput, rDest5
	mulscc	rDest4, rInsertBit, rDest4
	addcc	rDest4, rDest4, rDest4
	mulscc	rDest4, rInput, rDest4
	mulscc	rDest3, rInsertBit, rDest3
	addcc	rDest3, rDest3, rDest3
	mulscc	rDest3, rInput, rDest3
	mulscc	rDest2, rInsertBit, rDest2
	addcc	rDest2, rDest2, rDest2
	mulscc	rDest2, rInput, rDest2
	mulscc	rDest1, rInsertBit, rDest1
	addcc	rDest1, rDest1, rDest1
	mulscc	rDest1, rInput, rDest1
	mulscc	rDest0, rInsertBit, rDest0
	addcc	rDest0, rDest0, rDest0
	mulscc	rDest0, rInput, rDest0

! Write each dest word starting from the bottom of the block.  We
! use the already-written rDestX regs to address higher locations.  Why?
! Well, the next time through the loop, we would have to initialize
! their MS bits anyway so subsequent mulscc ops would have their sign
! bits SET.  This kills two birds with one stone.
! Originally, there was code here that would avoid writing words that
! were blank.  It didn't increase performance appreciably, but did require
! that all incoming images be blank.  We removed it.
rB2702State6:
	st	rDest0, [rDstAddr]
	sub	%g0, rDstPitch, rDest0		! will use in later writes
						! and initializes MSbit to 1
	st	rDest1, [rDstAddr + rDest0]
	sub	rDest0, rDstPitch, rDest1
	st	rDest2, [rDstAddr + rDest1]
	sub	rDest1, rDstPitch, rDest2
	st	rDest3, [rDstAddr + rDest2]
	sub	rDest2, rDstPitch, rDest3
	st	rDest4, [rDstAddr + rDest3]
	sub	rDest3, rDstPitch, rDest4
	st	rDest5, [rDstAddr + rDest4]
	sub	rDest4, rDstPitch, rDest5
	st	rDest6, [rDstAddr + rDest5]
	sub	rDest5, rDstPitch, rDest6
	st	rDest7, [rDstAddr + rDest6]
	sub	rDest6, rDstPitch, rDest7
	st	rDest8, [rDstAddr + rDest7]
	sub	rDest7, rDstPitch, rDest8
	st	rDest9, [rDstAddr + rDest8]
	sub	rDest8, rDstPitch, rDest9
	st	rDest10, [rDstAddr + rDest9]
	sub	rDest9, rDstPitch, rDest10
	st	rDest11, [rDstAddr + rDest10]
	sub	rDest10, rDstPitch, rDest11
	st	rDest12, [rDstAddr + rDest11]
	sub	rDest11, rDstPitch, rDest12
	st	rDest13, [rDstAddr + rDest12]
	sub	rDest12, rDstPitch, rDest13
	st	rDest14, [rDstAddr + rDest13]
	sub	rDest13, rDstPitch, rDest14
	st	rDest15, [rDstAddr + rDest14]

rB2702FinBlkLoop:
	sub	rDest14, rDstPitch, rDest15
	ld	[%fp+srcAddrFixupOffset], rInput	! fixup source addr
	add	rDstAddr, rDest15, rDstAddr		! advance rDstAddr
	deccc	rBlkCount				! done will last block?
	bnz	rB2702BlockLoop
	add	rSrcAddr, rInput, rSrcAddr


! END OF LOOP ACROSS ALL ROWS

! We've put out the right number of words, so quit
rB2702Fin:
	ret
	restore



! -----------------------------------------------------------------------------
! _ip_fastRotate180 Routine
! -----------------------------------------------------------------------------

! This routine is quite different from 90 and 270 degree rotation.  Most of
! the game is to reverse the order of bits within bytes, assemble the
! reversed bytes into double words and write the double words.

! There is a small trick, however.  We scan the source image from left to
! right, top to bottom.  Therefore, we scan the dest image right to left,
! bottom to top.  If the width of the image in bits is an exact multiple of
! 32, we just reverse each byte, stuff it into a double word and write the
! double word (modulo the obvious problems with starting and finishing each
! line on double word boundaries).

! However, if the image width is not a multiple of 32, there are some pad
! bits on the right end of the rightmost word in the source and dest images.
! if we just rotate the source image on 32 bit boundaries, the dest ends up
! being shifted right by the width of the pad.  No good.  So, we shift each
! reversed source word left by the width of the pad before writing it to
! the destination.  Naturally, we need to do a double shift to get a dest
! word full of bits.  When finished with the double shift, rNewWord is
! stored into rOldWord.

! Well, since we're doing double shifts, we'll be creating two output
! words.  It doesn't make sense to put rNewWord into rOldWord twice, so
! we'll have two rNewWords, rNewWord0 and rNewWord1.

! How do we start the loop and how many times do we execute it?  We'll use
! two registers to accumulate input data, rNewWord and rOldWord.  We start
! with rOldWord set to 0, representing the frame bits on the right side of
! the image.  We also use rInput to hold each byte is it's read from memory
! and rInputRev to hold each reversed byte.  The basic scheme is to read
! four bytes from right to left, reverse them, and pack the reversed versions
! into rNewWord.  The first byte read ends up on the left side of rNewWord.
! We then form an output word by shifting rNewWord left by the width of the
! pad on the right and ORing it with rOldWord shifted right by 32-padWidth.

! We do this twice, putting the results into rOut0 and rOut1, then do a
! double word write to memory.

! There's one final tweak.  Before starting the task of reversing the bytes
! in a word, we test to see if the word is empty.  If so, we just form
! the output word from rOldWord and fill rOldWord with 0's.

! This routine is the same for all image depths.  The caller is responsible
! for passing in the width of the image in BITS (width in pixel * depth)
! and for passing in the appropriate table for reversing the bits within
! a byte.

! BUG FIX: 11/16/93.  Say the destination line has an odd number of words
! in it (2n - 1) and the rightmost word is at an odd address.  We will
! write only word pairs, so will write 2n words.  That means we write one
! word off the left edge of the image.  If the image width wasn't a
! multiple of 32, we will probably shift some 1's into this area.  If we
! were cheating to make the image width a multiple of 32 by saying that
! a wider image was less wide, we'll write some of the non-zero pseudo-frame
! bits on the right side of the source to the dest.
! Rather than check before each double write to see if it's legal, we'll
! just write a word of 0's into the left frame if we ended the row by
! writing it.  This applies to both the 1 bpp and 2bpp loops.

! Revised 1/20/94: The upper level code now calls the primitive operation
! once per line instead of having the primitive do the entire image.
! This makes it easier to do banding and to use the same primitive for
! rotates in place.  We just strip the outer loop off this code and
! rearrange the arguments.

#undef	rSrcAddr
#undef	rSrcPitch
#undef	rDstAddr
#undef	rDstPitch
#undef	rInput
#undef	rBlkCount
#undef	rDest0
#undef	rDest1
#undef	rDest2
#undef	rDest3
#undef	rDest4
#undef	rDest5
#undef	rDest6
#undef	rDest7
#undef	rDest8
#undef	rDest9
#undef	rDest10
#undef	rDest11
#undef	rDest12
#undef	rDest13
#undef	rDest14
#undef	rDest15
#undef	rInsertBit
#undef	rCount
#undef	rSkipCount
#undef	ipStateOffset
#undef	srcAddrOffset
#undef	srcPitchOffset
#undef	dstAddrOffset
#undef	dstPitchOffset
#undef	rotTableOffset
#undef	blkCountOffset
#undef	srcAddrFixupOffset


! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPState	%i0		/* ptr to unused IPState */
#define	rSrcAddr	%i1		/* address of left end of src line */
#define rDstAddr	%i2		/* address of left end of dest line */
#define rWordCount	%i3		/* width of line in 32-bit words */
#define rRightShift	%i4		/* amount by which reversed image
					 * words should be shifted right */
#define rRevTable	%i5		/* ptr to table used to reverse
					 * pixels within bytes */

#define	rInput0		%o0		/* byte fetched from src image */
#define rInput1		%o1		/* use two of these for pipelining */
#define	rInputRev	%o2		/* byte with bits reversed */

#define rNewWord0	%o3		/* latest word seen */
#define rNewWord1	%o4
#define	rOldWord	%o5		/* from previous new word */

#define	rOut0		%l0		/* pair of output registers */
#define rOut1		%l1

#define rLeftShift	%l2		/* amount to shift dest bits left */

#define	rTest		%l3		/* used to see if word all 0's or
					 * all 1's (if so, it doesn't need
					 * to be reversed) */

/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	srcAddrOffset	0x48
#define	dstAddrOffset	0x4c
#define	wordCountOffset	0x50
#define	shiftOffset	0x54
#define revTableOffset	0x58

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_reverseLine)
MK_GLOBAL_NAME(ip_reverseLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState, [%fp+ipStateOffset]
	st	rSrcAddr, [%fp+srcAddrOffset]
	st	rDstAddr,  [%fp+dstAddrOffset]
	st	rWordCount, [%fp+wordCountOffset]
	st	rRightShift, [%fp+shiftOffset]
	st	rRevTable, [%fp+revTableOffset]

! Adjust the destination address so it points to the right end of the line
	sub	rWordCount, 1, rInput0	! get offset of last word
	sll	rInput0, 2, rInput0	! convert to byte count

! We can process the image faster if its width is an exact multiple of 32.
! Test for this case.
	tst	rRightShift
	bnz	rvLineHavePad			! need to do shifting
	add	rDstAddr, rInput0, rDstAddr	! finish adjusting dest addr

! *********** WIDTH IS MULTIPLE OF 32 *************************

! Here's the fast version of the loop through a row.  Is the first (rightmost)
! dest word address odd or even?  If it's even, we need to put out it's
! word first before beginning the main loop that writes doubles.
	andcc	rDstAddr, 4, %g0
	bnz	rvLineFastRowLoop
	ld	[rSrcAddr], rOut0		! half the time we'll repeat
						! this opcode at
						! rvLineFastRowLoop, but this
						! seems like the least of
						! the evils.

! The dest word was even.  Is it zero?
	ldub	[rSrcAddr + 3], rInput0
	add	rOut0, 1, rTest			! is entire word 0's or 1's?
	andcc	rTest, -2, %g0			! make sure is stays in rOut0
	bz	rvLineFastRowEntryFin		! yes, just store it
	ldub	[rSrcAddr + 2], rInput1
	ldub	[rRevTable + rInput0], rOut0	! rev 1st byte
	ldub	[rSrcAddr + 1], rInput0
	ldub	[rRevTable + rInput1], rInputRev ! rev 2nd byte
	sll	rOut0, 8, rOut0
	or	rOut0, rInputRev, rOut0
	ldub	[rSrcAddr + 0], rInput1
	ldub	[rRevTable + rInput0], rInputRev ! rev 3rd byte
	sll	rOut0, 8, rOut0
	or	rOut0, rInputRev, rOut0
	ldub	[rRevTable + rInput1], rInputRev ! rev 4th byte
	sll	rOut0, 8, rOut0
	or	rOut0, rInputRev, rOut0

! store the word into the even dest location
rvLineFastRowEntryFin:
	st	rOut0, [rDstAddr]
	add	rSrcAddr, 4, rSrcAddr		! pt to next src word
	deccc	rWordCount
	bz	rvLineFastRowLoopFin		! quit if that was
						! the last word
	sub	rDstAddr, 4, rDstAddr

! Here's the main loop that stores word using double writes.  Initially,
! we're on an odd dest address.  Fill the dest word.
! Note that this loop may fill one word too many on the left side.  It
! will just write frame bits into frame bits.  No biggee. (right?)
rvLineFastRowLoop:
	ld	[rSrcAddr], rOut1		! prepare for next test
	ldub	[rSrcAddr + 3], rInput0
	add	rOut1, 1, rTest			! is entire word 0's or 1's?
	andcc	rTest, -2, %g0			! if so, keep it in rOut1
	bz	rvLineFastRowLoopEvenWord
	ldub	[rSrcAddr + 2], rInput1
	ldub	[rRevTable + rInput0], rOut1	! rev 1st byte
	ldub	[rSrcAddr + 1], rInput0
	ldub	[rRevTable + rInput1], rInputRev ! rev 2nd byte
	sll	rOut1, 8, rOut1
	or	rOut1, rInputRev, rOut1
	ldub	[rSrcAddr + 0], rInput1
	ldub	[rRevTable + rInput0], rInputRev ! rev 3rd byte
	sll	rOut1, 8, rOut1
	or	rOut1, rInputRev, rOut1
	ldub	[rRevTable + rInput1], rInputRev ! rev 4th byte
	sll	rOut1, 8, rOut1
	or	rOut1, rInputRev, rOut1

! Get the even dest word
rvLineFastRowLoopEvenWord:
	ld	[rSrcAddr + 4], rOut0
	ldub	[rSrcAddr + 7], rInput0
	add	rOut0, 1, rTest			! is entire word 0's or 1's?
	andcc	rTest, -2, %g0			! (all but MS bit are 0)
	bz	rvLineFastRowLoopFin
	ldub	[rSrcAddr + 6], rInput1
	ldub	[rRevTable + rInput0], rOut0	! rev 1st byte
	ldub	[rSrcAddr + 5], rInput0
	ldub	[rRevTable + rInput1], rInputRev ! rev 2nd byte
	sll	rOut0, 8, rOut0
	or	rOut0, rInputRev, rOut0
	ldub	[rSrcAddr + 4], rInput1
	ldub	[rRevTable + rInput0], rInputRev ! rev 3rd byte
	sll	rOut0, 8, rOut0
	or	rOut0, rInputRev, rOut0
	ldub	[rRevTable + rInput1], rInputRev ! rev 4th byte
	sll	rOut0, 8, rOut0
	or	rOut0, rInputRev, rOut0

! store the double word
rvLineFastRowLoopFin:
	std	rOut0, [rDstAddr - 4]
	add	rSrcAddr, 8, rSrcAddr		! pt to next src word
	subcc	rWordCount, 2, rWordCount
	bg	rvLineFastRowLoop		! if more to do, keep going
	sub	rDstAddr, 8, rDstAddr
! END OF FAST LOOP ACROSS ONE ROW

! At this point, we move rScrAddr down one row, rDstAddr up one row and
! decrement rHeight.  If it's still positive, do another row.
! The reason we don't have to check to see if we've just written into
! the left frame is that it doesn't matter if we did.  Since we're
! reversing in units of words on word boundaries, the thing we would
! have just written into the left frame is the right frame, so it's
! ok.

! We've put out the right number of words, so quit
rvLineFastFin:
	ret
	restore


! *********** WIDTH IS NOT A MULTIPLE OF 32 *************************

! Since we'll be doing shifting, derive the shift distances.
rvLineHavePad:
	sub	rRightShift, 32, rLeftShift
	neg	rLeftShift			! equals number of pad bits
						! in rightmost word.  This
						! is the number of bits we
						! must avoid when shifting
						! the current output word
						! left before storing it.

! Here's the fast version of the loop through rows.  Is the first (rightmost)
! dest word address odd or even?  If it's even, we need to put out it's
! word first before beginning the main loop that writes doubles.
	andcc	rDstAddr, 4, %g0
	bnz	rvLineSlowRowLoop
	mov	0, rOldWord			! simulate frame bits to the
						! right of the first word
						! to be stored.

! The dest word was even.  Is it zero?
	ld	[rSrcAddr], rNewWord0
	ldub	[rSrcAddr + 3], rInput0
	add	rNewWord0, 1, rTest		! is word all 0's or 1's?
	andcc	rTest, -2, %g0			! (all but MS bit are 0)
	bz	rvLineSlowRowEntryFin
	ldub	[rSrcAddr + 2], rInput1
	ldub	[rRevTable + rInput0], rNewWord0 ! rev 1st byte
	ldub	[rSrcAddr + 1], rInput0
	ldub	[rRevTable + rInput1], rInputRev ! rev 2nd byte
	sll	rNewWord0, 8, rNewWord0
	or	rNewWord0, rInputRev, rNewWord0
	ldub	[rSrcAddr + 0], rInput1
	ldub	[rRevTable + rInput0], rInputRev ! rev 3rd byte
	sll	rNewWord0, 8, rNewWord0
	or	rNewWord0, rInputRev, rNewWord0
	ldub	[rRevTable + rInput1], rInputRev ! rev 4th byte
	sll	rNewWord0, 8, rNewWord0
	or	rNewWord0, rInputRev, rNewWord0

! store the word into the even dest location
rvLineSlowRowEntryFin:
	sll	rNewWord0, rLeftShift, rOut0	! we know rOldWord is 0, so
						! there's no reason to or it
						! in
	st	rOut0, [rDstAddr]
	mov	rNewWord0, rOldWord		! save previous NewWord
	add	rSrcAddr, 4, rSrcAddr		! pt to next src word
	deccc	rWordCount
	bz	rvLineSlowRowLoopFin		! quit if that was
						! the last word
	sub	rDstAddr, 4, rDstAddr

! Here's the main loop that stores word using double writes.  Initially,
! we're on an odd dest address.  Fill the dest word.
! Note that this loop may fill one word too many on the left side.  It
! will just write frame bits into frame bits.  No biggee. (right?)
rvLineSlowRowLoop:
	ld	[rSrcAddr], rNewWord1
	ldub	[rSrcAddr + 3], rInput0
	add	rNewWord1, 1, rTest		! is word all 0's or 1's?
	andcc	rTest, -2, %g0			! (all but MS bit are 0)
	bz	rvLineSlowRowLoopEvenWord
	ldub	[rSrcAddr + 2], rInput1
	ldub	[rRevTable + rInput0], rNewWord1 ! rev 1st byte
	ldub	[rSrcAddr + 1], rInput0
	ldub	[rRevTable + rInput1], rInputRev ! rev 2nd byte
	sll	rNewWord1, 8, rNewWord1
	or	rNewWord1, rInputRev, rNewWord1
	ldub	[rSrcAddr + 0], rInput1
	ldub	[rRevTable + rInput0], rInputRev ! rev 3rd byte
	sll	rNewWord1, 8, rNewWord1
	or	rNewWord1, rInputRev, rNewWord1
	ldub	[rRevTable + rInput1], rInputRev ! rev 4th byte
	sll	rNewWord1, 8, rNewWord1
	or	rNewWord1, rInputRev, rNewWord1

! Get the even dest word
rvLineSlowRowLoopEvenWord:
	ld	[rSrcAddr + 4], rNewWord0
	ldub	[rSrcAddr + 7], rInput0
	add	rNewWord0, 1, rTest		! is word all 0's or 1's?
	andcc	rTest, -2, %g0			! (all but MS bit are 0)
	bz	rvLineSlowRowLoopFin
	ldub	[rSrcAddr + 6], rInput1
	ldub	[rRevTable + rInput0], rNewWord0 ! rev 1st byte
	ldub	[rSrcAddr + 5], rInput0
	ldub	[rRevTable + rInput1], rInputRev ! rev 2nd byte
	sll	rNewWord0, 8, rNewWord0
	or	rNewWord0, rInputRev, rNewWord0
	ldub	[rSrcAddr + 4], rInput1
	ldub	[rRevTable + rInput0], rInputRev ! rev 3rd byte
	sll	rNewWord0, 8, rNewWord0
	or	rNewWord0, rInputRev, rNewWord0
	ldub	[rRevTable + rInput1], rInputRev ! rev 4th byte
	sll	rNewWord0, 8, rNewWord0
	or	rNewWord0, rInputRev, rNewWord0

! store the double word
rvLineSlowRowLoopFin:
	srl	rOldWord, rRightShift, rOut1	! LS bits of rOut1
	sll	rNewWord1, rLeftShift, rOldWord	! MS bits of rOut1
	or	rOut1, rOldWord, rOut1
	srl	rNewWord1, rRightShift, rOut0	! LS bits of rOut0
	sll	rNewWord0, rLeftShift, rOldWord	! MS bits of rOut0
	or	rOut0, rOldWord, rOut0
	std	rOut0, [rDstAddr - 4]
	add	rSrcAddr, 8, rSrcAddr		! pt to next src word
	mov	rNewWord0, rOldWord
	subcc	rWordCount, 2, rWordCount
	bg	rvLineSlowRowLoop		! if more to do, keep going
	sub	rDstAddr, 8, rDstAddr

! END OF SLOW LOOP ACROSS ONE ROW

! At this point, we move rScrAddr down one row, rDstAddr up one row and
! decrement rHeight.  If it's still positive, do another row.
! BUT FIRST, check to see if we've written off the left edge of the
! image, as we would do if there were an odd number of words in the
! dest line and the rightmost address were odd.  If we have written
! off the left end, see if rOut0 already contains 0.  If so, things
! are fine - we wrote a 0 to the frame.  If not, cover the trash up
! with 0.  It's faster to check once here than to check every double
! write above.
	bz	rvLineSlowFin
	cmp	rOut0, 0			! did we write a 0 there?
	bnz,a	rvLineSlowFin
	st	%g0, [rDstAddr + 4]		! only write if rOut0 != 0

! We've put out the right number of words, so quit
rvLineSlowFin:
	ret
	restore


/*

$Log: ssorthrt.s,v $
Revision 1.1.1.1  1999/09/20 18:18:19  jlin
Imported sources

Revision 1.10  1998/01/13 23:57:53  ddavies
ssorthrt.s -> /product/ipcore/ipshared/src/RCS/ssorthrt.s,v

CommonCheck: Add IPState parameter to almost all private routines.

# Revision 1.9  1995/02/16  00:40:19  ddavies
# ddavies on Wed Feb 15 16:38:18 PST 1995
#
# ssorthrt.s -> /product/ipcore/ipshared/src/RCS/ssorthrt.s,v
#
# The 90 degree and 270 degree code uses the mulscc opcode.  This opcodes
# assumes the n and v condition code flags are clear (or at least our use
# of it does).  I was clearing these flags by adding rSrcPitch to rSrcAddress,
# assuming that addresses are always positive.  In (at least) Solaris,
# addresses of shared memory are negative, so this didn't work.  Added a
# new opcode to explicitly clear the n and v flags in State1 of both the
# binary 90 degree rotation and binary 270 degree rotate.
#
# Revision 1.8  1994/01/21  21:23:14  ddavies
# ssorthrt.s -> /product/ipcore/ipshared/src/RCS/ssorthrt.s,v
#
# Modified SPARC assembly implementations of ip_rotateBlock90 and
# ip_rotateBlock270.  Well, no that's a lie.  I didn't do anything to those
# routines, but if I abort now, I lose the entire checkin.  Deleted
# ip_fastRotate180.  Replaced it with ip_reverseLine.  This routine reversed
# the order of the pixels in a single scan line instead of doing the entire
# rotate.  The new primitive can be used both when the source and dest
# are different and when the are the same (by using a temporary buffer).
#
# Revision 1.7  1993/12/04  05:00:49  ddavies
# ssorthrt.s -> /product/ipcore/ipshared/src/RCS/ssorthrt.s,v
#
# Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
#
# Revision 1.6  1993/11/17  00:44:32  ddavies
# ddavies on Tue Nov 16 16:42:52 PST 1993
#
# ssorthrt.s -> /product/ipcore/ipshared/src/RCS/ssorthrt.s,v
#
# Fix bug in which the left frame of the destination was being written with
# non-zero data during 180 degree rotates.  This happened when the the
# destination image was an odd number of words in width and the rightmost
# word address was odd.  It happened because we use double writes.  We now
# check for this case and write 0's into the frame (there had better be a
# frame!) if it happens.
#
# Revision 1.5  1993/11/05  02:25:31  ddavies
# ssorthrt.s -> /product/ipcore/ipshared/src/RCS/ssorthrt.s,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

*/

