/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  ssorthst.c--includes
 *
 *           ip_fastRot90Setup(): SPARC version of setup code needed for
 *				  90 degree clockwise rotation
 *           ip_fastRot270Setup(): SPARC version of setup code needed for
 *				   270 degree clockwise rotation
 */


#include <stddef.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"

    /* prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: ssorthst.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* This file contains the setup procedures used by the SPARC assembly
 * implementation of 90, 180 and 270 degree rotation.
 */

/* dimensions of the rectangles we rotate */
/* 1 bit/pixel images */
#define	cSrcRectWidth1			16
#define cLogSrcRectWidth1		4
#define cSrcRectWidthInBytes1		(cSrcRectWidth1>>3)
#define cSrcRectHeight1			32
#define cLogSrcRectHeight1		5

/* 2 bit/pixel images */
#define	cSrcRectWidth2			16	/* in pixels */
#define cLogSrcRectWidth2		4
#define cSrcRectWidthInBytes2		((cSrcRectWidth2*2)>>3)
#define cSrcRectHeight2			16
#define cLogSrcRectHeight2		4

/* These dimensions are for routines that are implemented in C
 * code even in the fast libraries.
 */
/* DO NOT CHANGE THESE WITHOUT CHANGING THE COPIES IN miorthrt.c AND
 * miorth48.c!!!
 */

/* 4 bit/pixel images */
#define	cSrcRectWidth4			2	/* in pixels */
#define cLogSrcRectWidth4		1
#define cSrcRectWidthInBytes4		((cSrcRectWidth4*4)>>3)
#define cSrcRectHeight4			8
#define cLogSrcRectHeight4		3

/* 8 bit/pixel images */
#define	cSrcRectWidth8			1	/* in pixels */
#define cLogSrcRectWidth8		0
#define cSrcRectWidthInBytes8		((cSrcRectWidth8*8)>>3)
#define cSrcRectHeight8			4
#define cLogSrcRectHeight8		2

/* PRIVATE */
/**************************************************************************
 *	ip_fastRot90Setup
 *		Performs the following actions required before doing
 *		90 degree rotation:
 *		1. Allocate, initialize and return address of rotation table
 *		   (if any).
 *		2. Return number of rectangles in the width of the image.
 *		3. Return the number of rectangles that fit into the height
 *		   of the image.
 *		4. Return the value used to go from the end of one
 *		   source rectangle to the beginning of the next when
 *		   processing a row of rectangles.
 *		5. Adjust the initial value of the source address.  This
 *		   will be somewhere on the rectangle in the lower
 *		   left corner of the image depending on the implementation.
 *		6. If necessary, adjust value of destination address.
 *		   This happens when the assembly code uses weird
 *		   tricks with sign bits that leak over into into the
 *		   order in which the dest image blocks are written.
 *		7. Return the displacement that we subtract from the source
 *		   address when moving up one rectangle in the image.  This
 *		   is used to move between rows of rectangles.
 *	   The following steps apply to rotation in place
 *		8. Return the number of source lines in a source rectangle.
 *
 *	Input:
 *	pSrcAddr:		ptr to initial source address INSIDE frame
 *	pDestAddr:		ptr to initial dest image INSIDE frame
 *	sWidth:			width of source image without frame
 *	sHeight:		height of source image without frame
 *	sDepth:			depth of source image
 *	sBpl:			source image bytes/line
 *	pRotationTable:		pointer to ptr to first element in rotation
 *				table.
 *	pWidthInBlocks:		ptr to width of image in units of rotation
 *				rectangles.
 *	pHeightInBlocks:	ptr to height of image in units of rotation
 *				rectangles.
 *	pInterBlockAddrIncr:	ptr to value we add to the source address
 *				to move from one source block to the next
 *				within one row of source blocks
 *	pInterRowAddrIncr:	ptr to value we subtract from the source addr
 *				to move between rows of rectangles
 *	pSrcBlockHeight:	height of source blocks in rows; used only
 *				for in-place rotation
 *	Returns:
 *	ia_nomem:	Ran out of memory for rotation table
 *	ia_successful:	Setup was successful.
 *
 **************************************************************************/
Int32 CDECL 
ip_fastRot90Setup (
    IPState    *pIPState,
    UInt8     **pSrcAddr,
    Int32	sWidth,
    Int32	sHeight,
    Int32	sDepth,
    Int32	sBpl,
    UInt8     **pDstAddr,
    Int32	dBpl,
    UInt32    **pRotationTable,
    Int32      *pWidthInBlocks,
    Int32      *pHeightInBlocks,
    Int32      *pInterBlockAddrIncr,
    Int32      *pInterRowAddrIncr,
    Int32      *pSrcBlockHeight)
{
    PROLOGUE;
    

	/* Switch on the depth of the image, this changes the returned
	 * parameters. */
    switch (sDepth)
    {
    case 1:
	    /* The SPARC assembly code doesn't use a table */
	*pRotationTable = NULL;

	    /* set source image pointer to upper left corner of rectangle in
	     * lower left corner of the image.  We need the upper left
	     * corner because the silly multiply step op shifts right,
	     * not left.  This means we need to access the top line
	     * first so it ends up on the rightmost (most shifted)
	     * side of the dest registers. */
	*pSrcAddr += (sHeight-1-(cSrcRectHeight1-1)) * sBpl;
    
	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth1-1) >> cLogSrcRectWidth1;
	*pHeightInBlocks = (sHeight+cSrcRectHeight1-1) >> cLogSrcRectHeight1;
	*pSrcBlockHeight = cSrcRectHeight1;

	    /* the loop structure of the assembly code adds 32 * srcPitch,
	     * we must subtract that much back off, then move over 1 rect. */
	*pInterBlockAddrIncr = -(sBpl << cLogSrcRectHeight1)
			    + cSrcRectWidthInBytes1;

	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight1);
	break;

    case 2:
	    /* The SPARC assembly code doesn't use a table */
	*pRotationTable = NULL;

	    /* set source image pointer to upper left corner of rectangle in
	     * lower left corner of the image.  We need the upper left
	     * corner because the silly multiply step op shifts right,
	     * not left.  This means we need to access the top line
	     * first so it ends up on the rightmost (most shifted)
	     * side of the dest registers. */
	*pSrcAddr += (sHeight-1-(cSrcRectHeight2-1)) * sBpl;

	    /* set dest pointer to point to bottom of upper block.  This is
	     * so we can use negative offsets to address the block.
	     * Negative offsets are used so we can initialize the dest
	     * registers to negative numbers.  Negative numbers are used
	     * because white = 11 in gray images and we want to
	     * initialize the data registers to white.  Note that we
	     * don't do this in the binary case.  That's because white
	     * is 0 in binary so we want to initialize the data registers
	     * to 0 there.
	     */
	*pDstAddr += (dBpl << cLogSrcRectWidth2) - dBpl;
    
	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth2-1) >> cLogSrcRectWidth2;
	*pHeightInBlocks = (sHeight+cSrcRectHeight2-1) >> cLogSrcRectHeight2;
	*pSrcBlockHeight = cSrcRectHeight2;

	    /* the loop structure of the assembly code adds
	     * rectHeight * srcPitch, we must subtract that much back off,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = -(sBpl << cLogSrcRectHeight2) +
				cSrcRectWidthInBytes2;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight2);
	break;

	/* The setup for the 4bpp and 8bpp cases happens on behalf of
	 * primitives contained in miorth48.c
	 */
    case 4:
	    /* The C implementation needs a table */
	ERR_PASS(ip_make4bppRotTable (pIPState, pRotationTable));

	    /* set source image pointer to lower left corner of rectangle in
	     * lower left corner of the image.  We need the lower left
	     * corner because the we want to shift left, not right.
	     * This means we need to access the bottom line
	     * first so it ends up on the rightmost (most shifted)
	     * side of the dest registers. */
	*pSrcAddr += (sHeight-1) * sBpl;

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth4-1) >> cLogSrcRectWidth4;
	*pHeightInBlocks = (sHeight+cSrcRectHeight4-1) >> cLogSrcRectHeight4;
	*pSrcBlockHeight = cSrcRectHeight4;

	    /* the loop structure of the assembly code subtracts
	     * rectHeight * srcPitch, we must add that much back on,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = (sBpl << cLogSrcRectHeight4) - sBpl +
				cSrcRectWidthInBytes4;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight4);
	break;

    default:	/* case 8 */
	    /* The C implementation doesn't use a table */
	*pRotationTable = NULL;

	    /* set source image pointer to lower left corner of rectangle in
	     * lower left corner of the image.  We need the lower left
	     * corner because the we want to shift left, not right.
	     * This means we need to access the bottom line
	     * first so it ends up on the rightmost (most shifted)
	     * side of the dest registers. */
	*pSrcAddr += (sHeight-1) * sBpl;

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth8-1) >> cLogSrcRectWidth8;
	*pHeightInBlocks = (sHeight+cSrcRectHeight8-1) >> cLogSrcRectHeight8;
	*pSrcBlockHeight = cSrcRectHeight8;

	    /* the loop structure of the assembly code subtracts
	     * rectHeight * srcPitch, we must add that much back on,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = (sBpl << cLogSrcRectHeight8) - sBpl +
				cSrcRectWidthInBytes8;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight8);
	break;
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *	ip_fastRot270Setup
 *		Performs the following actions required before doing
 *		270 degree rotation:
 *		1. Allocate, initialize and return address of rotation table
 *		   (if any).
 *		2. Return number of rectangles in the width of the image.
 *		3. Return the number of rectangles that fit into the height
 *		   of the image.
 *		4. Return the value used to go from the end of one
 *		   source rectangle to the beginning of the next when
 *		   processing a row of rectangles.
 *		5. Adjust the initial value of the source address.  This
 *		   will be somewhere on the rectangle in the lower
 *		   left corner of the image depending on the implementation.
 *		6. If necessary, adjust value of destination address.
 *		   This happens when the assembly code uses weird
 *		   tricks with sign bits that leak over into into the
 *		   order in which the dest image blocks are written.
 *		7. Return the displacement that we subtract from the source
 *		   address when moving up one rectangle in the image.  This
 *		   is used to move between rows of rectangles.
 *	   The following steps apply to rotation in place
 *		8. Return the number of source lines in a source rectangle.
 *
 *	Input:
 *	pSrcAddr:		ptr to initial source address INSIDE frame
 *	sWidth:			width of source image without frame
 *	sHeight:		height of source image without frame
 *	sDepth:			depth of source image
 *	sBpl:			source image bytes/line
 *	pDestAddr:		ptr to initial dest image INSIDE frame
 *	dWidth:			width of dest image or transfer buffer
 *				in pixels
 *	dBpl:			bytes/line of dest image or transfer buffer
 *	pRotationTable:		pointer to ptr to first element in rotation
 *				table.
 *	pWidthInBlocks:		ptr to width of image in units of rotation
 *				rectangles.
 *	pHeightInBlocks:	ptr to height of image in units of rotation
 *				rectangles.
 *	pInterBlockAddrIncr:	ptr to value we add to the source address
 *				to move from one source block to the next
 *				within one row of source blocks
 *	pInterRowAddrIncr:	ptr to value we subtract from the source addr
 *				to move between rows of rectangles
 *	pSrcBlockHeight:	height of source blocks in rows; used only
 *				for in-place rotation
 *	Returns:
 *	ia_nomem:	Ran out of memory for rotation table
 *	ia_successful:	Setup was successful.
 *
 **************************************************************************/
Int32 CDECL 
ip_fastRot270Setup (
    IPState    *pIPState,
    UInt8     **pSrcAddr,
    Int32	sWidth,
    Int32	sHeight,
    Int32	sDepth,
    Int32	sBpl,
    UInt8     **pDstAddr,
    Int32	dWidth,
    Int32	dBpl,
    UInt32    **pRotationTable,
    Int32      *pWidthInBlocks,
    Int32      *pHeightInBlocks,
    Int32      *pInterBlockAddrIncr,
    Int32      *pInterRowAddrIncr,
    Int32      *pSrcBlockHeight)
{
    PROLOGUE;
    

	/* Switch on the depth of the image, this changes the returned
	 * parameters. */
    switch (sDepth)
    {
    case 1:
	    /* The SPARC assembly code doesn't use a table */
	*pRotationTable = NULL;

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth1-1) >> cLogSrcRectWidth1;
	*pHeightInBlocks = (sHeight+cSrcRectHeight1-1) >> cLogSrcRectHeight1;
	*pSrcBlockHeight = cSrcRectHeight1;

	    /* set source image pointer to lower left corner of block
	     * in lower left corner of image.  Remember that the rectangle
	     * must be aligned with the top line, not the bottom one.
	     * We will read the source words
	     * within each block from bottom to top.  This is becaue the
	     * mulscc op does right shifts.  We want the first words read
	     * to end up on the right side of the destination words.
	     */
	*pSrcAddr += ((*pHeightInBlocks * cSrcRectHeight1) - 1) * sBpl;
    
	    /* Set the original dest pointer to the upper left corner of the
	     * lower right dest rectangle.  We use the upper corner so we can
	     * use positive address displacements.  We're initializing the MS
	     * bit of each destination register by using it as a positive
	     * address displacement in the assembly code.
	     */
	*pDstAddr += ((sWidth-1) - (cSrcRectWidth1-1)) * dBpl +
			((dWidth-1)/32) * 4;
    
	    /* the loop structure of the assembly code subtracts
	     * rectHeight * srcPitch,
	     * we must add that much back on, then move over 1 rect. */
	*pInterBlockAddrIncr = (sBpl << cLogSrcRectHeight1)
			    + cSrcRectWidthInBytes1;

	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight1);
	break;

    case 2:
	    /* The SPARC assembly code doesn't use a table */
	*pRotationTable = NULL;

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth2-1) >> cLogSrcRectWidth2;
	*pHeightInBlocks = (sHeight+cSrcRectHeight2-1) >> cLogSrcRectHeight2;
	*pSrcBlockHeight = cSrcRectHeight2;

	    /* set source image pointer to lower left corner of block
	     * in lower left corner of image.   Remember that the rectangle
	     * must be aligned with the top source line, not the bottom one.
	     * We will read the source words
	     * within each block from bottom to top.  This is becaue the
	     * mulscc op does right shifts.  We want the first words read
	     * to end up on the right side of the destination words.
	     */
	*pSrcAddr += ((*pHeightInBlocks * cSrcRectHeight2) - 1) * sBpl;

	    /* Set the original dest pointer to the lower left corner of the
	     * lower right dest rectangle.  We use the lower corner so we can
	     * use negative address displacements.  We're initializing the MS
	     * bit of each destination register by using it as a negative
	     * address displacement in the assembly code.
	     */
	*pDstAddr += (sWidth - 1) * dBpl + ((dWidth-1)/16) * 4;
    
	    /* the loop structure of the assembly code subtracts
	     * rectHeight * srcPitch, we must add that much back on,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = (sBpl << cLogSrcRectHeight2) +
				cSrcRectWidthInBytes2;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight2);
	break;


	/* The setup for the 4bpp and 8bpp cases happens on behalf of
	 * primitives contained in miorth48.c
	 */
    case 4:
	    /* The C implementation needs a table */
	ERR_PASS(ip_make4bppRotTable(pIPState, pRotationTable));

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth4-1) >> cLogSrcRectWidth4;
	*pHeightInBlocks = (sHeight+cSrcRectHeight4-1) >> cLogSrcRectHeight4;
	*pSrcBlockHeight = cSrcRectHeight4;

	    /* Set the source so it points to the upper left corner of
	     * the rectangle in the lower left of the image.  The rects
	     * are aligned with the top line of the image.  We do this
	     * weird thing so the in-place rotations can make room for
	     * the dest at the bottom of the image, which allows them
	     * to use regular old i_rasterOp to stretch the dest bits.
	     */
	*pSrcAddr += ((*pHeightInBlocks - 1) * cSrcRectHeight4) * sBpl;

	    /* Set dest image pointer to the last word
	     * in the lower right corner of image */
	*pDstAddr += (sWidth-1) * dBpl + ((dWidth-1)/8) * 4;

	    /* the loop structure of the assembly code adds
	     * rectHeight * srcPitch, we must subtract that much back off,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = -(sBpl << cLogSrcRectHeight4) + sBpl +
				cSrcRectWidthInBytes4;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight4);
	break;

    default:	/* case 8 */
	    /* The C implementation doesn't use a table */
	*pRotationTable = NULL;

	    /* Set parms to be passed to assembly routine. */
	*pWidthInBlocks = (sWidth+cSrcRectWidth8-1) >> cLogSrcRectWidth8;
	*pHeightInBlocks = (sHeight+cSrcRectHeight8-1) >> cLogSrcRectHeight8;
	*pSrcBlockHeight = cSrcRectHeight8;

	    /* Set the source so it points to the upper left corner of
	     * the rectangle in the lower left of the image.  The rects
	     * are aligned with the top line of the image.  We do this
	     * weird thing so the in-place rotations can make room for
	     * the dest at the bottom of the image, which allows them
	     * to use regular old i_rasterOp to stretch the dest bits.
	     */
	*pSrcAddr += ((*pHeightInBlocks - 1) * cSrcRectHeight8) * sBpl;

	    /* Set dest image pointer to the last word
	     * in the lower right corner of image */
	*pDstAddr += (sWidth-1) * dBpl + ((dWidth-1)/4) * 4;

	    /* the loop structure of the assembly code adds
	     * rectHeight * srcPitch, we must subtract that much back off,
	     * then move over 1 rect. */
	*pInterBlockAddrIncr = -(sBpl << cLogSrcRectHeight8) + sBpl +
				cSrcRectWidthInBytes8;
	*pInterRowAddrIncr = (sBpl << cLogSrcRectHeight8);
	break;
    }

    NORMAL_RETURN;
}


/*

$Log: ssorthst.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.4  1998/01/13 23:54:23  ddavies
ssorthst.c -> /product/ipcore/ipshared/src/RCS/ssorthst.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1995/12/13 02:53:01  ejaquith
ssorthst.c -> /product/ipcore/ipshared/src/RCS/ssorthst.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1994/02/13  01:00:05  ddavies
 * ssorthst.c -> /product/ipcore/ipshared/src/RCS/ssorthst.c,v
 *
 * Modified ip_fastRot90Setup and ip_fastRot270Setup so they supply an
 * extra parmater needed by the in-place rotation routines and so they can
 * handle 4 bit and 8 bit images.
 * Actually, this routine is doing setup for 4-bit and 8-bit images on
 * behalf of the C implementation in miorth48.c.  There is no SPARC
 * assembly code for rotation of 4 and 8 bit images.
 *
 * Revision 1.1  1994/01/21  21:19:15  ddavies
 * Initial revision
 *

 */
