/*****************************************************************
 *  Copyright (c) 1991, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*  File: ssskew.s
 *
 * This file contains assembly code for "Robot" (AI edge following) and Fine
 * skew detection, on a sparcstation (ss).  This is the first viable
 * demonstration of successful AI robotics techniques in the history
 * of the galaxy.
 *
 * The "robot" skew detection finds horizontal edges, recording offsets
 * required to follow edges across the page.  It averages the offsets
 * for several edges.
 *
 * The fine skew detection algorithm traces a set of slanted paths,
 * and samples each line and the line directly above or below it.  The
 * absolute value of the differences in the number of pixels in these pairs
 * of samples is summed to produce the score.
 *
 * ip_detectSkewFine() is passed 32 bit words containing:
 * 	* Address of the left end of the topmost sample line
 *	* Number of vertical sample lines
 *	* Number of samples per horizontal line
 *	* Bytes to be added to get from one sample to the next horizontally.
 *	  This often includes some vertical movement as well.
 *	* Bit vector used to determine when to add +/- image pitch during
 *	  horizontal movement
 *	* +/- Image pitch
 *
*/

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssskew.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:19 $"
	.align	4

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPStateIn	%i0
#define	rSrcAddrIn	%i1
#define rNumVSampIn	%i2
#define rNumHSampIn	%i3
#define	rSrcVAddrIncIn	%i4
#define rSrcHAddrIncIn	%i5

#define	rSrcAddr	%i0		/* current addr in source image */
#define rNumVSamp	%i1		/* number of grod lines in sample */
#define rNumHSamp	%i2		/* number of samples/line */
#define	rSrcVAddrInc	%i3		/* vert distance between grid lines */
#define rSrcHAddrInc	%i4		/* distance between samples */
#define rSrcPitch	%i5		/* number of bytes/image line */

#define	rFracVector	%g1		/* contains bit vector used to decide
					   when to move up or down 1 line */

#define	rShiftedScore	%g2		/* used only in difference counting */
#define rScoreLo	%g3		/* lo word of score for all lines */
#define rScoreHi	%l1		/* hi word of score from all lines */

#define	rInput		%g4		/* input word */

#define rSrcOffset	%o0		/* addr offset from beginning of line */

#define	rRunScr		%o1		/* scratch register */
#define rOtherInput	%o2		/* for difference counting */
#define	rSrcAddr1	%o3		/* to address next image line */
#define rSampleMask	%o4		/* selects bits to compare */
#define	rCurNumHSamp	%l0		/* can only do this many samples
					   before reloading rFracVector */


/* magical C Frame constant (no local variables) */
#define	LF26		-64

#define	scoreShiftDist	4	/* lineScore =
				   lineScore << (lineScore >> scoreShiftDist) */
#define	cByteSamples	0x7F7F7F7F
#define	cFracWidth	32		/* number of samples before reloading
					   rFracVector */

/* define the offsets used for arguments */
#define	cIPStateOffset	0x44
#define	cSrcAddrOffset	0x48
#define	numVSampOffset	0x4c
#define numHSampOffset	0x50
#define vAddrDispOffset	0x54
#define hAddrDispOffset	0x58
#define	srcPitchOffset	0x5c
#define fracVecOffset	0x60
#define resultLoPtrOffset 0x64
#define resultHiPtrOffset 0x68


LL0:					! does this really do anything?

! -----------------------------------------------------------------------------
! _ip_findDiffPixelsSum Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_findDiffPixelsSum)
MK_GLOBAL_NAME(ip_findDiffPixelsSum):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  We do need to reload
! rFracVector every now and then.
	st	rIPStateIn,     [%fp+cIPStateOffset]
	st	rSrcAddrIn,     [%fp+cSrcAddrOffset]
	st	rNumVSampIn,    [%fp+numVSampOffset]
	st	rNumHSampIn,    [%fp+numHSampOffset]
	st	rSrcVAddrIncIn, [%fp+vAddrDispOffset]
	st	rSrcHAddrIncIn, [%fp+hAddrDispOffset]

	ld	[%fp+cSrcAddrOffset],  rSrcAddr
	ld	[%fp+numVSampOffset],  rNumVSamp
	ld	[%fp+numHSampOffset],  rNumHSamp
	ld	[%fp+vAddrDispOffset], rSrcVAddrInc
	ld	[%fp+hAddrDispOffset], rSrcHAddrInc
	ld	[%fp+srcPitchOffset],  rSrcPitch
	ld	[%fp+fracVecOffset],   rFracVector

! This code trundles across a line, computing the difference in vertically
! adjacent pixels at each step.  Our difference is recorded if one sample has
! any pixels and the other has none.
!
! Moving diagonally is messy.  Why not move vertically instead?  Well, if we
! moved through an entire column, we'd run out of cache so would miss all the
! time.  We'd have to break the image into horizontal blocks.  In order to
! make sure the cache didn't fill, the blocks couldn't be very tall.  We
! wouldn't take very many samples per block, so there would be a fair amount
! of loop/call/return overhead.  We'll just move horizontally instead.
!
! Diagonal movement has two parts.  After each sample, we add an address
! displacment (rSrcHAddrInc).  We also shift rFracVector left by 1.  If a 1
! fell out, we also add rSrcPitch to the address.  This is how we approximate
! moving up or down by fractional lines - we move up or down by whole lines
! at selected intervals.
!
! MODIFICATIONS:
!
! We were recording an edge difference if a full 32 bit word was zero/non-zero
! and the word underneath was non-zero/zero.  The results were not
! sufficiently accurate.  We will now record a difference *for each byte pair*
! in the 4 byte words.  The byte pair has a difference if one byte is zero
! and the other is not.
!

! first initialize the accumulated score and sample mask
	mov	0, rScoreLo
	mov	0, rScoreHi
	sethi	%hi(cByteSamples), rSampleMask
	or	rSampleMask, %lo(cByteSamples), rSampleMask
	
! and point to line 2 below current one
	add	rSrcAddr, rSrcPitch, rSrcAddr1
	add	rSrcAddr1, rSrcPitch, rSrcAddr1

! START OF LOOP ACROSS ALL GRID ROWS

! Initialize all of the stuff that isn't already initialized.  Before looping
! back up here, we'll also reload rNumHSamp and rFracVector.
findDiffPixelsImageLoop:
	mov	0, rSrcOffset
	mov	0, rRunScr

! Continue on same row after just reloading rFracVector
findDiffPixelsContinueRow:
	subcc	rNumHSamp, cFracWidth, rNumHSamp
	bge	findDiffPixelsRowPreload
	mov	cFracWidth, rCurNumHSamp

! the remaining row width is less than cFracWidth, so adjust number of
! samples
	add	rNumHSamp, cFracWidth, rCurNumHSamp

findDiffPixelsRowPreload:
	ld	[rSrcAddr+rSrcOffset], rInput	! get next input word

findDiffPixelsRowLoop:
	ld	[rSrcAddr1+rSrcOffset], rOtherInput ! and get other sample
! see if we should jump up/down by the occasional line
	addcc	rFracVector, rFracVector, rFracVector
	bcc	findDiffPixelsComputeDiff
	add	rSrcOffset, rSrcHAddrInc, rSrcOffset
! it's time to bump
	add	rSrcOffset, rSrcPitch, rSrcOffset
findDiffPixelsComputeDiff:
	xorcc	rInput, rOtherInput, %g0	! any edges at all?
	bz	findDiffPixelsCheckRow		! nope, see if we're at the
						! end of a row
	and	rInput, rSampleMask, rInput	! get rid of MS bit of each byte
	add	rInput, rSampleMask, rInput	! MS bit of each byte now set
						! if any of the lower 7 were set
	and	rOtherInput, rSampleMask, rOtherInput
	add	rOtherInput, rSampleMask, rOtherInput
	xor	rInput, rOtherInput, rInput	! edge bits in MS bits
	andn	rInput, rSampleMask, rInput	! select just the right bits
	srl	rInput, 7, rInput		! prepare to add into counts
	add	rRunScr, rInput, rRunScr	! add into 4 counts (1/byte)
findDiffPixelsCheckRow:
	deccc	rCurNumHSamp			! see if that was end of row
	bnz	findDiffPixelsRowLoop		! no, continue down row
	ld	[rSrcAddr+rSrcOffset], rInput	! and get first input word

! We've either exhausted rFracVector (so it needs to be reloaded) or are
! really at the end of the row.  If rNumHSamp is <= 0 (from subtraction at
! findDiffPixelsContinueRow), we're really done.  Otherwise, just reload
! rFracVector and branch back to findDiffPixelsContinueRow where
! rCurNumHSamp will be reloaded and rNumHSamp decremented.
	addcc	rNumHSamp, %g0, %g0		! are we really done?
	bg	findDiffPixelsContinueRow	! no keep going
	ld	[%fp+fracVecOffset], rFracVector ! with a new rFracVector

! We're at the end of the row.  There are 4 edge counts in rRunScr.  Add them
! together, then perform a nice non-linear function on the score
! and add the result to rScoreHi:rScoreLo.  The non-linear functions
! is here so that big scores (as you'd get if you were sampling along
! a real edge) really stand out while average scores (as you'd get on
! any diagonal sample) don't.
! Current non-linear function is lineScore * 2^(lineScore/16)
! BUG FIX: the score was overflowing 32 bits for some images.  So we
! put it into two registers.  We know that since the number of horizontal
! samples is less than 128, we can just add two of the counts without
! getting a carry.  We have to watch out for carry when doing the final
! add.
! The final sum can be as large as 4*127.  The final shift count can be
! as large as (4*127)/16 = 31.  Therefore, we shift part of the count
! right by (32-shiftCount) and add it directly to the upper score word.
! The score shifted left by shiftCount is added to the lower word.  Any
! carries from the lower word are added to the upper word.
findDiffPixelsEndRow:
	srl	rRunScr, 16, rShiftedScore
	add	rRunScr, rShiftedScore, rRunScr	! get 2 counts from 4
	srl	rRunScr, 8, rShiftedScore
	and	rShiftedScore, 0xFF, rShiftedScore ! delete old trash
	and	rRunScr, 0xFF, rRunScr		! delete old trash
	add	rRunScr, rShiftedScore, rRunScr	! get 1 count from 2
	
findDiffPixelsModifyScore:
	srl	rRunScr, scoreShiftDist, rShiftedScore
! add (rRunScr >> (32 - rShiftedScore)) to the upper score word.  Note
! that for our purposes, (32-rShiftedScore) = -rShiftedScoresince
! the shifter only looks at the 5 LSBs of the shift count. HOWEVER, if
! rShiftedScore == 0, -rShiftedScore also equals 0, not 32.  We want to
! add 0 to rScoreHi in that case, so just do nothing.
	subcc	%g0, rShiftedScore, rInput
	bz	findDiffPixelsSumUpdateLowScore
	srl	rRunScr, rInput, rInput
	add	rScoreHi, rInput, rScoreHi

! now add the lower part of the sum to the lower score, taking care of
! carries as well.
findDiffPixelsSumUpdateLowScore:
	sll	rRunScr, rShiftedScore, rRunScr
	addcc	rScoreLo, rRunScr, rScoreLo
	addx	rScoreHi, 0, rScoreHi		! add carry to high word
! update the addresses for the new row
	add	rSrcAddr, rSrcVAddrInc, rSrcAddr
	add	rSrcAddr1, rSrcVAddrInc, rSrcAddr1

	ld	[%fp+numHSampOffset], rNumHSamp
	subcc	rNumVSamp, 1, rNumVSamp
	bnz	findDiffPixelsImageLoop
	ld	[%fp+fracVecOffset], rFracVector

! END OF LOOP ACROSS ALL ROWS

! We've looked through all the rows, so return the final score
! Note that we want to shift rScoreHi left by 1 and put the MS
! bit of rScoreLo into rScoreHi.  The MS bit of rScoreLo should
! then be cleared.  This makes it consistent with the scores returned
! by the C code.
findDiffPixelsSumFin:
	addcc	rScoreLo, rScoreLo, rScoreLo
	addx	rScoreHi, rScoreHi, rScoreHi
	srl	rScoreLo, 1, rScoreLo
	ld	[%fp+resultLoPtrOffset], rSrcAddr
	ld	[%fp+resultHiPtrOffset], rSrcAddr1
	st	rScoreLo, [rSrcAddr]
	st	rScoreHi, [rSrcAddr1]
	mov	0, %i0
	ret
	restore


! UNDEFINE REGISTER USAGE.
! NOTE: cpp requires c-style comments
#undef	rIPStateIn	%i0
#undef	rSrcAddrIn	%i1
#undef	rNumVSampIn	%i2
#undef	rNumHSampIn	%i3
#undef	rSrcVAddrIncIn	%i4
#undef	rSrcHAddrIncIn	%i5
#undef	rSrcAddr	%i0		/* current addr in source image */
#undef	rNumVSamp	%i1		/* number of grod lines in sample */
#undef	rNumHSamp	%i2		/* number of samples/line */
#undef	rSrcVAddrInc	%i3		/* vert distance between grid lines */
#undef	rSrcHAddrInc	%i4		/* distance between samples */
#undef	rSrcPitch	%i5		/* number of bytes/image line */

#undef	rFracVector	%g1		/* contains bit vector used to decide
					   when to move up or down 1 line */

#undef	rShiftedScore	%g2		/* used only in difference counting */
#undef	rScoreLo	%g3		/* low word of score for all lines */
#undef	rScoreHi	%l1		/* hi word of score from all lines */

#undef	rInput		%g4		/* input word */

#undef	rSrcOffset	%o0		/* addr offset from beginning of line */

#undef	rRunScr		%o1		/* scratch register */
#undef	rOtherInput	%o2		/* for difference counting */
#undef	rSrcAddr1	%o3		/* to address next image line */
#undef	rSampleMask	%o4		/* selects bits to compare */
#undef	rCurNumHSamp	%l0		/* can only do this many samples
					   before reloading rFracVector */

! -----------------------------------------------------------------------------
! _ip_doEdgeSample Routine
! -----------------------------------------------------------------------------

! This routine is quite unlike the others.  Instead of taking statistics
! from fixed points in the image, this routine looks for an edge, then
! plots its slope.

! This routine implements the "robot" referred to in fastSkew.c's
! ip_detectSkewRobot.  After getting plopped down at some random point in the
! image, we take the following actions:
!	1. If the sample at the initial location has no black pixels (is
!	   white), sample every cSampleSpacing'th line of the image
!	   straight down.
!	   If no black pixels are found in 3 samples, we just give up and 
!	   return.
!	2. If the initial location had black pixels or we have found black
!	   pixels in step 1, we look down every cSampleSpacing'th line of
!	   the image straight down.  If no white is found in 3 samples,
!	   just give up and return.
!	3. If white was found, do a linear search to find the exact boundary
!	   between black and white.
!	4. Starting from the lowest black word, move one halfword to the right.
!	5. If the new halfword is black, look down to see if the halfword below
!	   is white.  If so, increment hCount by 1 and go to step 4.
!	6. If the new halfword was black, but the word below it was also black,
!	   keep looking down for a total of 5 additional lines.  If black is
!	   found, add the number of lines above to vCount
!	   and increment hCount.  Adjust the vertical position to the
!	   uppermost black word and go to step 4 again.  If no white was
!	   found, just quit.
!	7. If the new word was white, look upwards for black.  If white is
!	   found in no more than 5 lines, subtrackt the offset from vCount,
!	   increment hCount, adjust the current position to the uppermost
!	   black word and continue with step 4.  If no black was found, quit.
!	8. When we quit, add the final values of vCount and hCount in the
!	   memory structure pointed to by rImSlopeAddr.

! Alterations:

!	This whole thing is both more complex and simpler than I had
!	thought at first.  What we're really doing is measuring the slope
!	of individual *lines* of text, then averaging these.  We can think
!	of this as setting a start point, tracking along the slope, setting
!	the end point and recording the slope between the start and end
!	points.  We happen to record the slope all the way along because we
!	never know when the end will show up.  The important insight here
!	is that the placements of the start and endpoints are critical.
!
!	Unfortunately, these are the hardest to place.  When starting a line,
!	we can easily start too high on a sloped character.  When finishing
!	a line, we can be fooled by a descender (go too low) or a rounded
!	character (move up too high).  We'd like to measure slopes on "good
!	samples."  But how?
!
!	Just tossing the first and last samples is not good, the second and
!	second to last samples are just as likely to be funny.  What we'll
!	do instead is look for runs of samples that seem to be following a
!	nice edge.  We'll use the leftmost sample of the first run as the
!	beginning of the line.  We'll use the rightmost sample of the last
!	run as the end of the line and record that slope.
!
!	We use a simple state machine to implement this.  Initially, the
!	state machine looks for a run.  Each time the step is "big" vertically,
!	it clears the horizontal and vertical counters.  When the first run
!	is found, the horizontal and vertical counters automatically have the
!	slope since the last big step.
!
!	We then follow runs.  At each little step inside a run, we record the
!	latest slope in rHRunEnd and rVRunEnd.  If a big step is seen, we
!	look for another run, but do not clear rHCount or rVCount.  We also do
!	NOT change rHRunEnd and rVRunEnd.
!
!	The routine ends when no step can be taken because we've run out of
!	edge.  If the edge has been long enough, we record the last slope in
!	rHRunEnd and rVRunEnd.
!
! **************************************
!
!	That sort of worked ok, except it made the lines over which slopes
!	are calculated too short.  A difference of 0.1 degrees equals +- 1
!	line of vertical displacement in 573 bits horizontal displacement!
!	To get good accuracy, we really need to measure over long lines.
!
!	However, we can't just jump over white space willy-nilly.  An inter-
!	word gap (one we'd like to cross) in one document can be bigger
!	than an inter-column gap (should NOT cross) in another.  How far can
!	we jump?
!
!	Initially, we thought of using the character height as a measure of
!	allowable white space gap.  Unfortunately, justified text has much
!	more space between words than unjustified text, yet the character
!	height does not change.  Instead, we'll let the higher level routine
!	try different white space widths.  This routine will return the
!	number of spaces actually skipped.
!
!	The caller will start out with a large value for allowable white
!	white space.  If the number of spaces actually skipped is small
!	compared to the number of samples taken (but non-zero), we assume
!	columns have been skipped and reduce the white space value until
!	the white space skipped value goes to zero.
!	the white space skipped value goes to zero.
!
! ************************************************************************
! 
!	That got pretty close to reality, but still failed (off by 0.5 degrees)
!	for pages that had lots of short lines - phone books or data sheets.
!	At that point, we decided to punt.  This routine will only be used to
!	do a very quick estimate of coarse skew - within a degree.  The fine
!	skew detection routine above will be used to narrow it down from there.
! 
!	 There was yet another bug.  One of the CCITT test charts has a
!	 diagonal line running from upper left to lower right.  This robot
!	 routine followed that line.  Oooops.  The fix was to put a check
!	 just after the portion of the loop that tries to track descenders.
!	 If the total horizontal distance covered in that loop is much 
!	 greater than the amount of white space we're allowed to skip plus
!	 the minimum run length required to exit the loop, we decide that
!	 we've followed some big alien object.
! 
! ***************************************************************
! 
!  9/24/93 BUG FIX:.  There's a possibility of an infinite loop if a
!  horizontal black line extends completely across the image AND the caller
!  doesn't have a frame around the image.  We would keep following the
!  horizontal line forever.  XIS may have seen this.  Anyway, we'll say that
!  if we're following an edge for more than the width of the image - stop
!  following it.  Like the path of return when we find too much white space
!  in a row, we return directly from the ip_doFindEdgeTakeStep routine.
!  How?  Well, subroutine calling at this level leaves the return address
!  in a SPARC register.  If we just ignore this address, we can return
!  to the address stored in the previous register frame`s return register.
!  Hence, the ip_doFindEdgeTakeStep routine can return to the caller of
!  ip_doEdgeSample.


! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments

#define	rIPStateIn	%i0
#define	rSrcAddrIn	%i1
#define	rSrcPitchIn	%i2
#define	rImSlopeAddrIn	%i3
#define	rMaxSlopeTrackIn %i4
#define	rMinLegalAddrIn	%i5

#define	rSrcAddr	%i0		/* address of current point */
#define	rSrcPitch	%i1		/* bytes/line of source image */
#define	rImSlopeAddr	%i2		/* ptr to dest arg */
#define	rMaxSlopeTrack	%i3		/* Don't track slopes that have bumps
					   bigger than this */
#define	rMinLegalAddr	%i4		/* Legal addrs are greater than this */
#define	rMaxLegalAddr	%i5		/* Legal addrs are less than this */

#define	rInput		%g1		/* input sample value */
#define	rHCount		%g2		/* current value of hCount */
#define	rVCount		%g3		/* current value of vCount */

#define	rSrcOffset	%g4		/* offset used to sample image
					   relative to current location */
#define	rLineOffset	%g5		/* holds cSampleSpacing*rSrcPitch */
#define	rSearchCount	%l0		/* number of lines yet to search */
#define rMarkReg	%l1		/* used to scribble on image for
					   debugging */
#define	rHRunEnd	%l2		/* H component of end of latest run */
#define rVRunEnd	%l3		/* V component of end of latest run */
#define	rScratch	%l4		/* scratch reg */
#define	rRunLength	%l5		/* records length of current run of
					   "little steps." */
#define	rCurSpaces	%l6		/* number of white space samples left
					   before we give up */
#define	rTotalSpaces	%l7		/* number of white space samples we
					   actually skipped */
#define	rRetAddr	%o7		/* Used for subroutine return addrs */
#define	rRunLengthBound %o0		/* Used to stop tracking infinite
					   edges */

#define	cSampleSpacing	8		/* sample every 8th vert line when
					   looking for first edge */
#define	cLogSampleSpacing 3

#define	hCountOffset	0
#define	vCountOffset	4

#define	cMaxSpaceCount	2		/* maximum amount of white space that
					   we can skip over */

#define	cMinValidRunLen 2		/* ignore runs shorter than this */
#define	minSampleLength	cMinValidRunLen+1 /* toss edges shorter than this */
#define bigStepValue	3		/* if the edge changes by more than
					   this in one sample, it's a
					   "big step."  Track it, but remember
					   it might not be well behaved. */
#define	logHeightToSkip	4


	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_doEdgeSample)
MK_GLOBAL_NAME(ip_doEdgeSample):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.
	st	rIPStateIn,	[%fp+0x44]
	st	rSrcAddrIn,	[%fp+0x48]
	st	rSrcPitchIn,	[%fp+0x4c]
	st	rImSlopeAddrIn,	[%fp+0x50]
	st	rMaxSlopeTrackIn, [%fp+0x54]
	st	rMinLegalAddrIn, [%fp+0x58]

	ld	[%fp+0x48], rSrcAddr
	ld	[%fp+0x4c], rSrcPitch
	ld	[%fp+0x50], rImSlopeAddr
	ld	[%fp+0x54], rMaxSlopeTrack
	ld	[%fp+0x58], rMinLegalAddr
	ld	[%fp+0x5c], rMaxLegalAddr

! initialize the register used to scribble on the image and the one
! recording how many spaces we've skipped so far.
	mov	-1, rMarkReg
	mov	0, rTotalSpaces

! Initialize the number of steps we're allowed to take.
	srl	rSrcPitch, 1, rRunLengthBound

! Assume we're in a white place, look down for black.
	mov	0, rSrcOffset
	sll	rSrcPitch, cLogSampleSpacing, rLineOffset

doEdgeSampleFindBlack:
	ld	[rSrcAddr+rSrcOffset], rInput
! **************** TEMPORARY *******************
!	st	rMarkReg, [rSrcAddr]
! **********************************************
	add	rSrcOffset, rLineOffset, rSrcOffset ! add offset
	orcc	rInput, %g0, %g0		! have we found black yet?
	bnz	doEdgeSampleFindWhite		! yes, look for black edge
	ld	[rSrcAddr+rSrcOffset], rInput
	add	rSrcOffset, rLineOffset, rSrcOffset ! add offset
	orcc	rInput, %g0, %g0		! have we found black yet?
	bnz	doEdgeSampleFindWhite		! yes, look for black edge
	ld	[rSrcAddr+rSrcOffset], rInput
	add	rSrcOffset, rLineOffset, rSrcOffset ! add offset
	orcc	rInput, %g0, %g0		! have we found black yet?
	bz	finDoEdgeSample			! nope, just quit
	nop					! Use nop instead of .empty
						! to keep Purify happy
		
! Having found some black, look downwards for white
doEdgeSampleFindWhite:
	ld	[rSrcAddr+rSrcOffset], rInput
	add	rSrcOffset, rLineOffset, rSrcOffset ! add offset
	orcc	rInput, %g0, %g0		! have we found white yet?
	bz	doEdgeSampleFoundEdge		! yes, narrow it down
	ld	[rSrcAddr+rSrcOffset], rInput
	add	rSrcOffset, rLineOffset, rSrcOffset ! add offset
	orcc	rInput, %g0, %g0		! have we found white yet?
	bz	doEdgeSampleFoundEdge		! yes, narrow it down
	ld	[rSrcAddr+rSrcOffset], rInput
	add	rSrcOffset, rLineOffset, rSrcOffset ! add offset
	orcc	rInput, %g0, %g0		! have we found white yet?
	bnz	finDoEdgeSample			! nope, just quit
	nop					! Use nop instead of .empty
						! to keep Purify happy

! We've found a black-on-top-of-white edge, sort of.  Refine the location
! using a linear search.  This guarantees we really find the first
! black edge here.
doEdgeSampleFoundEdge:
	sub	rSrcOffset, rLineOffset, rSrcOffset ! back up to prev sample
	sub	rSrcOffset, rSrcPitch, rSrcOffset ! prepare for 1st sample
	ld	[rSrcAddr+rSrcOffset], rInput
doEdgeSampleFind1stEdge:
	sub	rSrcOffset, rSrcPitch, rSrcOffset
	orcc	rInput, %g0, %g0		! have we found black yet?
	bz	doEdgeSampleFind1stEdge
	ld	[rSrcAddr+rSrcOffset], rInput

! At this point, rSrcOffset points one line above the first black line
! found,  We want to point to the lowest black line, so add
! rSrcPitch.
	add	rSrcOffset, rSrcPitch, rSrcOffset
	add	rSrcAddr, rSrcOffset, rSrcAddr

! ********** TRACKING STATE MACHINE STARTS HERE ********************

! We're ready to start the tracking state machine. In the initial state,
! we clear rHCount and rVCount.  We return here so long as one of the first
! cMinValidRunLen steps is big.

	mov	cMaxSpaceCount, rCurSpaces	! say we're not in white space
	mov	0, rHRunEnd
	mov	0, rVRunEnd

doEdgeSampleTrackInit:
	mov	0, rHCount
	mov	0, rVCount
	mov	cMinValidRunLen, rRunLength
doEdgeSampleFind1stRun:
	call	ip_doFindEdgeTakeStep
	lduh	[rSrcAddr + 2], rInput

! We've taken a step trying to find the first run.  Was it big or little?
	bg	doEdgeSampleTrackInit		! was big, start over
	deccc	rRunLength

! It was a little step.  Have we found a valid run?
	bnz	doEdgeSampleFind1stRun
	mov	rHCount, rHRunEnd

! We've found a(nother) valid run, record its location and take a step.
doEdgeSampleFoundRun:
! ******************** TEMPORARY **********************
!	sth	rMarkReg, [rSrcAddr]
! *****************************************************
	mov	rVCount, rVRunEnd
	call	ip_doFindEdgeTakeStep
	lduh	[rSrcAddr + 2], rInput

! Was that a big or little step?
	ble,a	doEdgeSampleFoundRun		! if little step, keep going
	mov	rHCount, rHRunEnd

! It was a big step, so that run is over.  See if we can find another.
doEdgeSampleFindLaterRunInit:
	mov	cMinValidRunLen, rRunLength
doEdgeSampleFindLaterRun:
	call	ip_doFindEdgeTakeStep
	lduh	[rSrcAddr + 2], rInput

! We've taken a step trying to find a later run.  Was it big or little?
	bg	doEdgeSampleFindLaterRunInit	! was big, start over
	deccc	rRunLength

! It was a little step.  Have we found a valid run?
	bnz	doEdgeSampleFindLaterRun
	nop

! Found a later run.  If it was short, record it's length and loop back
! to continue tracking the run.
	sub	rHCount, rHRunEnd, rScratch	! how long have we been
						! following that thing?
	cmp	rScratch, cMaxSpaceCount+cMinValidRunLen+1
	ble,a	doEdgeSampleFoundRun
	mov	rHCount, rHRunEnd

! We normally exit this machine when ip_doFindEdgeTakeStep is unable to
! find an edge.  If we have just tracked a large alien object, we also
! quit now.
	ba	doEdgeSamplePostResults
	nop

! ************** ip_doFindEdgeTakeStep SUBROUTINE ********************

! This code takes one step to the right and tries to follow the edge.
! It's used by the state machine above.
ip_doFindEdgeTakeStep:
! ******************** TEMPORARY **********************
!	sth	rMarkReg, [rSrcAddr + 2]
! *****************************************************
	cmp	rSrcAddr, rMinLegalAddr		! have we gone off the top?
	ble	doEdgeSamplePostResults
	cmp	rSrcAddr, rMaxLegalAddr		! have we gone off the bottom?
	bge	doEdgeSamplePostResults
	add	rSrcAddr, 2, rSrcAddr		! move to the right
	orcc	rInput, %g0, %g0		! is this sample still black?
	bz	doEdgeSampleWasWhite		! no, look up one line
	mov	rMaxSlopeTrack, rSearchCount	! prepare line count

! Pure horizontal edges are the most common, so we'll have a special
! case for them.
	lduh	[rSrcAddr+rSrcPitch], rInput
	sll	rSrcPitch, 1, rSrcOffset	! prepare to look down
	orcc	rInput, %g0, %g0		! is this sample white?
	bz,a	doFindEdgeTakeStepHorizRet
	mov	0, rSearchCount

! the sample to the right was black, sample below it is black
! so we're really supposed to slant down.
	sll	rSearchCount, 1, rSearchCount	! We should be willing to look
						! a long way down to find the
						! the bottom of this line.
doFindEdgeTestPosEdgeLoop:
	lduh	[rSrcAddr+rSrcOffset], rInput
	deccc	rSearchCount			! should we keep looking?
	bz	doEdgeSamplePostResults		! no, really quit
	orcc	rInput, %g0, %g0		! test sample below
	bnz,a	doFindEdgeTestPosEdgeLoop	! if it's black, keep going
	add	rSrcOffset, rSrcPitch, rSrcOffset

! We found the white edge.
! Number of lines edge went down (pos dir) = 2*rMaxSlopeTrack-rSearchCount;
! Number of halfwords we went over = 1;
! Addr of lowest black line = rSrcAddr+rSrcOffset-rSrcPitch
	add	rSrcAddr, rSrcOffset, rSrcAddr
	sub	rSrcAddr, rSrcPitch, rSrcAddr
	sll	rMaxSlopeTrack, 1, rScratch
	sub	rScratch, rSearchCount, rSearchCount
	add	rVCount, rSearchCount, rVCount
doFindEdgeTakeStepHorizRet:
	add	rHCount, 1, rHCount
	deccc	rRunLengthBound			! have we gone too far?
	bz	doEdgeSamplePostResults		! yes, quit now
	mov	cMaxSpaceCount, rCurSpaces	! white space has ended
	retl
	subcc	rSearchCount, bigStepValue, %g0	! was it a big step?


! The sample to the right was white, so the edge has moved up.  Set up
! rSrcOffset correctly, then go up looking for it.
doEdgeSampleWasWhite:
	sub	%g0, rSrcPitch, rSrcOffset
doFindEdgeTestNegEdgeLoop:
	lduh	[rSrcAddr+rSrcOffset], rInput
	deccc	rSearchCount			! should we keep looking?
	bz	doEdgeSampleTestWhiteSpace	! no, see if we should skip
						! some white space
	orcc	rInput, %g0, %g0		! test sample above
	bz,a	doFindEdgeTestNegEdgeLoop	! if it's white, keep going
	sub	rSrcOffset, rSrcPitch, rSrcOffset

! We found the black edge.
! Number of lines edge went up (neg direction) = rMaxSlopeTrack-rSearchCount;
! Number of halfwords we went over = rCurSkipCount+1;
! Addr of lowest black line = rSrcAddr+rSrcOffset
	add	rSrcAddr, rSrcOffset, rSrcAddr
	sub	rMaxSlopeTrack, rSearchCount, rSearchCount
	sub	rVCount, rSearchCount, rVCount
	add	rHCount, 1, rHCount
	deccc	rRunLengthBound			! have we gone too far?
	bz	doEdgeSamplePostResults		! yes, quit now
	mov	cMaxSpaceCount, rCurSpaces	! white space has ended
	retl
	subcc	rSearchCount, bigStepValue, %g0	! was it a big step?


! We get here when an edge vanished.  First, see if this is the start of
! a new white block.  If so, increment rTotalSpaces.  We want to know the
! number of white blocks seen, not the total amount of white space.
! Then, check to see if we're allowed to skip some white space in search
! of an edge.
! If so, inc rHCount, set the flags so this will look like a big step (not
! part of a run) and return.
! If we're at the end of allowable white space, record the current values
! of rHRunEnd and rVRunEnd.
doEdgeSampleTestWhiteSpace:
! ******************** TEMPORARY **********************
!	sth	rMarkReg, [rSrcAddr]
! *****************************************************
	deccc	rCurSpaces
	bz	doEdgeSamplePostResults
	deccc	rRunLengthBound			! have we gone too far?
	bz	doEdgeSamplePostResults		! yes, quit now
	inc	rHCount
	retl
	addcc	%g0, bigStepValue, %g0		! set cc's to "big step"

! Have reached max white space, so quit
doEdgeSamplePostResults:
! ******************** TEMPORARY **********************
!	sth	rMarkReg, [rSrcAddr]
! *****************************************************
	cmp	rHRunEnd, minSampleLength	! was that edge big enough?
	ble	finDoEdgeSample			! no, just quit
	ld	[rImSlopeAddr+vCountOffset], rVCount
	add	rVCount, rVRunEnd, rVCount
	st	rVCount, [rImSlopeAddr+vCountOffset]
	ld	[rImSlopeAddr+hCountOffset], rHCount
	add	rHCount, rHRunEnd, rHCount
	st	rHCount, [rImSlopeAddr+hCountOffset]

! And just return
finDoEdgeSample:
	mov	0, %i0
	ret
	restore


/*

$Log: ssskew.s,v $
Revision 1.1.1.1  1999/09/20 18:18:19  jlin
Imported sources

Revision 1.9  1998/01/13 23:57:55  ddavies
ssskew.s -> /product/ipcore/ipshared/src/RCS/ssskew.s,v

CommonCheck: Add IPState parameter to almost all private routines.

! Revision 1.8  1995/02/24  02:34:12  ddavies
! ssskew.s -> /product/ipcore/ipshared/src/RCS/ssskew.s,v
!
! Change ip_findDiffPixelsSum so that it computes the total score into a
! double word instead of into a single 32-bit word.  This is used to
! prevent overflows.  This routine assumes that there are at most 127
! horizontal samples per line, which is guaranteed by ip_detectSkewFine.
!
! Revision 1.7  1993/12/04  05:01:12  ddavies
! ssskew.s -> /product/ipcore/ipshared/src/RCS/ssskew.s,v
!
! Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
!
! Revision 1.6  1993/11/05  02:25:56  ddavies
! ssskew.s -> /product/ipcore/ipshared/src/RCS/ssskew.s,v
!
! Add RCS magic to have string specifying this file's name, rev number and
! last edit date into the object file.  Also add change log at the bottom.
!

*/

