/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  t_scgr41.c: machine independent routine to make the table used to
 *		scale to 4bpp gray, 1.5x reduction
 *
 *          1.5x scale to gray reduction:
 *                  Int32	i_getSizeOfScaleToGray4bpp1_5xTab()
 *                  Int32	i_fillScaleToGray4bpp1_5xTab()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: t_scgr41.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* This procedure is used to compute the table used to reduce binary images
 * by a factor of 1.5, producing a 2x2 square of 4 bit gray pixels for
 * each 3x3 square of binary pixels.  The procedure is PUBLIC because it
 * can be called by upper level routines directly when they wish to provide
 * the lower routines with a ready-made table instead of requring them to
 * remake the table on each call.
 *
 * The algorithms is described in detail in the comments in i_scgr41.c.
 * 
 * The table is indexed with a 7 bit block containing a mode select bit and
 * two pixel triples.  We could have use 16 bit, but use 32 for the sake
 * of consistency with other scale to gray operations.
 *
 * The goal is to turn each 3x3 square of binary pixels into a 2x2 square
 * of 4-bit pixels that takes values from 0 through 15.
 * Here's a picture of 9 binary input pixels mapped to 4 output pixels:
 *
 * 	0  1  2		0    1
 *	3  4  5     =>
 *	6  7  8		2    3
 *
 * The four gray output pixels would ideally be calculated as follows:
 * 
 *	Out0 = (In0 + In1/2 + In3/2 + In4/4) * 15/2.25
 *	Out1 = (In2 + In1/2 + In5/2 + In4/4) * 15/2.25
 *	Out2 = (In6 + In7/2 + In3/2 + In4/4) * 15/2.25
 *	Out3 = (In8 + In7/2 + In5/2 + In4/4) * 15/2.25
 *
 * This mapping is perfectly linear and covers the entire range of outputs.
 * How do we approximate it efficiently?
 *
 * We divide the table into two sections, one for the first and third, or
 * major, input rows, another for the second, or minor, input row.
 * Each half of the table is addressed with 6 bits, so each half will map
 * two groups of 3 pixels into 2 gray pixels.  Let's just look at the
 * weights used to translate one set of 3 pixels into 2 gray pixels.  In
 * fact, lets just look at the weights used to calculate pixel Out0 above;
 * the other weights in the table are calculated symmetrically.
 *
 * If we can meet these goals, we'll be happy:
 *	1. When none of the pixels are active, the result is 0.
 *	2. When all of the pixels are active, the result is 15.
 *	3. There is no need to check for rollover.
 *	4. The mapping from number of input pixels to output value is
 *	   roughly linear (within roundoff error) and monotonic.
 *	5. The mapping is only a function of the number of pixels active,
 *	   not their position.  For example, pixels In1 and In3 should
 *	   count equally to the sum.
 *	6. Several sums can be computed in parallel.
 *	7. Any fixup effort required at the end of the summing process,
 *	   like to add in carries from fractional parts, is minimized.
 *
 * One way to look at the problem is that we'd like to weight each pixel
 * by (4/9) * 15.  That's hard to calculate, but 4/8 is trivial.  However, then
 * the sum of the weights is too big.  To compensate, we drop the LS bit
 * of the 15, making it (4/8) * 14.  This happens to work as seen below
 * The weights we'll use:
 *  Pixel	weighted sum	In binary
 *   In0		7.0	0111.00
 *   In1		3.5	0011.10
 *   In3		3.5	0011.10
 *   In4	        1.75	0001.11
 *
 * Another set of mappings that doesn't use any fraction bits is:
 *  Pixel	weighted sum	In binary
 *   In0		7	0111
 *   In1		3	0011
 *   In3		3	0011
 *   In4	        2	0010
 *
 *
 * In the table below, we calculate the perfect sum of pixel values, the
 * sum found using the first set of weights then truncating and the sum
 * found using the second set of weights (no truncation required):
 *
 * In0 In1 In3 In4	Ideal	Sum1	Sum2
 *   0   0   0   0	0.00	0	0
 *   0   0   0   1	1.67	1	2
 *   0   0   1   0	3.33	3	3
 *   0   0   1   1	5.00	5	5
 *   0   1   0   0	3.33	3	3
 *   0   1   0   1	5.00	5	5
 *   0   1   1   0	6.67	7	6
 *   0   1   1   1	8.34	8	8
 *   1   0   0   0	6.67	7	7
 *   1   0   0   1	8.34	8	9
 *   1   0   1   0	10.00	10	10
 *   1   0   1   1	11.67	12	12
 *   1   1   0   0	10.00	10	10
 *   1   1   0   1	11.67	12	12
 *   1   1   1   0	13.33	14	13
 *   1   1   1   1	15.00	15	15
 *
 * The second mapping seems fine to me and doesn't involve any fraction
 * bits, so is faster to compute.  We'll use it.
 *
 */

/* PUBLIC */
/**************************************************************************
 *  i_getSizeOfScaleToGray4bpp1_5xTab()  
 *		Return the size of a table used to scale binary
 *		images to 4bpp by 1.5x.
 *
 *	Input:
 *		pSize =		ptr to returned size of table
 *
 *	Return code:
 *		ia_successful	good completion
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfScaleToGray4bpp1_5xTab (
    IPState    *pIPState,
    UInt32     *pSize)
{
    PROLOGUE;
    *pSize = 2*64 * sizeof(UInt32);
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_fillScaleToGray4bpp1_5xTab()  
 *		Each 2x2 square of the source binary image is mapped
 *		to a 4-bit output pixel.  The value of the output
 *		pixel is a (roughly) linear function of the number of
 *		bits turned on in the 2x2 input square.
 *
 *	Input:
 *		pTable =	ptr to allocated table.
 *		tabSize =	size of allocated table in bytes.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_invalidParm	table passed in is too small or not there
 *
 **************************************************************************/
Int32  CDECL 
i_fillScaleToGray4bpp1_5xTab (
    IPState    *pIPState,
    UInt32     *pTable,
    UInt32	tabSize)
{
    /* This is the table of weights used to create the output image.
     * NOTE: Since 0 = white in binary images and 0 = black in gray
     * images, we do an inversion at this point.
     */
    PROLOGUE;
    UInt32	 majorWeightsLeft[4] = {10,	/* 00 => In0 and In1 => 10 */
					7,	/* 01 => In0 active => 7 */
					3,	/* 10 => In1 active => 3 */
					0};	/* 11 => no pix => 0 */
   UInt32	 majorWeightsRight[4] ={10,	/* 00 => In2 and In1 => 10 */
					3,	/* 01 => In1 active => 3 */
					7,	/* 10 => In2 active => 7 */
					0};	/* 11 => no pix => 0 */
   UInt32	 minorWeightsLeft[4] = {5,	/* 00 => In3 and In4 => 5 */
					3,	/* 01 => In3 active => 3 */
					2,	/* 10 => In4 active => 2 */
					0};	/* 11 => no pix => 0 */
   UInt32	 minorWeightsRight[4] ={5,	/* 00 => In5 and In4 => 5 */
					2,	/* 01 => In4 active => 2 */
					3,	/* 10 => In5 active => 3 */
					0};	/* 11 => no pix => 0 */

    UInt32	 index;				/* index into table */
    UInt32	 realSize;			/* used to check pTable */

	/* Figure out how big a table is needed */
    i_getSizeOfScaleToGray4bpp1_5xTab (pIPState, &realSize);

    if ((pTable == NULL) || (realSize > tabSize))
	return ia_invalidParm;

	/* Fill both halves of the table.  Each 6-bit block contains
	 * two triples.
	 */
    for (index = 0; index < 64; index++)
    {
	pTable[index] = (majorWeightsLeft[  (index >> 4)      ] << 12) |
			(majorWeightsRight[((index >> 3) & 0x3)] << 8) |
			(majorWeightsLeft[ ((index >> 1) & 0x3)] << 4) |
			(majorWeightsRight[ (index       & 0x3)]);
	pTable[index + c1_5xTableOffset] =
			(minorWeightsLeft[  (index >> 4)      ] << 12) |
			(minorWeightsRight[((index >> 3) & 0x3)] << 8) |
			(minorWeightsLeft[ ((index >> 1) & 0x3)] << 4) |
			(minorWeightsRight[ (index       & 0x3)]);
    }

    NORMAL_RETURN;	/* That's it */

}

/*

$Log: t_scgr41.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.7  1998/01/13 23:54:25  ddavies
t_scgr41.c -> /product/ipcore/ipshared/src/RCS/t_scgr41.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1996/03/26 23:19:13  ddavies
t_scgr41.c -> /product/ipcore/ipshared/src/RCS/t_scgr41.c,v

Add i_getSizeOfScaleToGray4bpp1_5xTab, which returns the size of the table
used by i_ScaleToGray4bpp1_5x.  Also change i_makeScaleToGray4bpp1_5xTab
into i_fillScaleToGray4bpp1_5xTab, which only fills the table.  The new
routine does not allocate the table.

Revision 1.5  1995/12/13 02:53:03  ejaquith
t_scgr41.c -> /product/ipcore/ipshared/src/RCS/t_scgr41.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/02/04  20:21:51  dferrell
 * dferrell on Fri Feb 4 12:21:31 PST 1994
 *
 * t_scgr41.c -> /product/ipcore/ipshared/src/RCS/t_scgr41.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:30:21  ddavies
 * t_scgr41.c -> /product/ipcore/ipshared/src/RCS/t_scgr41.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  02:26:13  ddavies
 * t_scgr41.c -> /product/ipcore/ipshared/src/RCS/t_scgr41.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


