/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  t_scgr42.c: machine independent routine to make the table used to
 *		scale to 4bpp gray, 2x reduction
 *
 *          2x scale to gray reduction:
 *                  Int32	i_getSizeOfScaleToGray4bpp2xTab()
 *                  Int32	i_fillScaleToGray4bpp2xTab()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: t_scgr42.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* This procedure is used to compute the table used to reduce binary images
 * by a factor of 2, producing one 4 bit gray pixel for each 2x2 square of
 * binary pixels.  The procedure is PUBLIC because it can be called by
 * upper level routines directly when they wish to provide the lower routines
 * with a ready-made table instead of requring them to remake the table
 * on each call.
 *
 * The table is indexed with a byte containing four pairs
 * of pixels.  It returns a 32-bit word containing four 4-bit integers and
 * four numbers composed of 2 bits of fraction and 2 bits of carry.  These
 * are used as described in the comments to i_scgr42.c and miscgr42.c.
 * The goal is to turn each 2x2 square of binary pixels into a single 4-bit
 * pixel that takes values from 0 through 15.  The values as a function of
 * number of active pixels is:
 *
 * Number of input pixels	Output value
 *		0		0
 *		1		3
 *		2		7
 *		3		11
 *		4		15
 *
 * For the sake of efficiency, we would like to process as many pixels at
 * once as possible.  Table lookup is a good way to space out the pixels and
 * do some computing.  On the other hand, we don't like to make huge tables,
 * so we like to restrict the number of pixels used to index the tables.
 * We often use 8 as a nice table index size.
 *
 * After thinking about several schemes, we settled on one that weights each
 * 1 in the input by 15/4.  The weights are:
 *  # of pix		weighted sum	In binary
 *	0		0		0000.00
 *	1		3.75		0011.11
 *	2		7.5		0111.10
 *	3		11.25		1011.01
 *	4		15.0		1111.00
 *
 * This routine computes a table containing these weights..
 */

/* PUBLIC */
/**************************************************************************
 *  i_getSizeOfScaleToGray4bpp2xTab()  
 *		Return the size of a table used to scale binary
 *		images to 4bpp by 2x.
 *
 *	Input:
 *		pSize =		ptr to returned size of table
 *
 *	Return code:
 *		ia_successful	good completion
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfScaleToGray4bpp2xTab (
    IPState    *pIPState,
    UInt32     *pSize)
{
    PROLOGUE;
    *pSize = 256 * sizeof(UInt32);
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_fillScaleToGray4bpp2xTab()  
 *		Each 2x2 square of the source binary image is mapped
 *		to a 4-bit output pixel.  The value of the output
 *		pixel is a (roughly) linear function of the number of
 *		bits turned on in the 2x2 input square.
 *
 *	Input:
 *		pTable =	ptr to allocated table.
 *		tabSize =	size of allocated table in bytes.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_invalidParm	table passed in is too small or not there
 *
 **************************************************************************/
Int32  CDECL 
i_fillScaleToGray4bpp2xTab (
    IPState    *pIPState,
    UInt32     *pTable,
    UInt32	tabSize)
{
    /* This is the table of weights used to create the output image.
     * NOTE: Since 0 = white in binary images and 0 = black in gray
     * images, we do an inversion at this point.
     */
    PROLOGUE;
    UInt32	 weightVals[4] = {0x00020007,	/* 00 => 0 pix => 0000.00 */
				  0x00030003,	/* 01 => 1 pix => 0011.11 */
				  0x00030003,	/* 10 => 1 pix => 0011.11 */
				  0x00000000};	/* 11 => 2 pix => 0111.10 */
   UInt32	 index;				/* index into table */
   UInt32	 realSize;			/* used to check pTable */


	/* Figure out how big a table is needed */
    i_getSizeOfScaleToGray4bpp2xTab (pIPState, &realSize);

    if ((pTable == NULL) || (realSize > tabSize))
	return ia_invalidParm;

	/* fill the table.  Each byte contains four bit-pairs.  We compute
	 * the index for each bit pair, then combine them.
	 */
    for (index = 0; index < 256; index++)
    {
	pTable[index] = (weightVals[index & 0x3]) |
			(weightVals[(index >> 2) & 0x3] << 4) |
			(weightVals[(index >> 4) & 0x3] << 8) |
			(weightVals[(index >> 6) & 0x3] << 12);
    }

    NORMAL_RETURN;	/* That's it */

}

/*

$Log: t_scgr42.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.7  1998/01/13 23:54:26  ddavies
t_scgr42.c -> /product/ipcore/ipshared/src/RCS/t_scgr42.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1996/03/26 23:20:05  ddavies
t_scgr42.c -> /product/ipcore/ipshared/src/RCS/t_scgr42.c,v

Add i_getSizeOfScaleToGray4bpp2xTab, which returns the size of the table
used by i_ScaleToGray4bpp2x.  Also change i_makeScaleToGray4bpp2xTab
into i_fillScaleToGray4bpp2xTab, which only fills the table.  The new
routine does not allocate the table.

Revision 1.5  1995/12/13 02:53:06  ejaquith
t_scgr42.c -> /product/ipcore/ipshared/src/RCS/t_scgr42.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/02/04  20:22:01  dferrell
 * t_scgr42.c -> /product/ipcore/ipshared/src/RCS/t_scgr42.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:30:26  ddavies
 * t_scgr42.c -> /product/ipcore/ipshared/src/RCS/t_scgr42.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  02:26:21  ddavies
 * t_scgr42.c -> /product/ipcore/ipshared/src/RCS/t_scgr42.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


