/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  t_scgr44.c: machine independent routine to make the table used to
 *		scale to 4bpp gray, 4x reduction
 *
 *          4x scale to gray reduction:
 *                  Int32	i_getSizeOfScaleToGray4bpp4xTab()
 *                  Int32	i_fillScaleToGray4bpp4xTab()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: t_scgr44.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* This procedure is used to compute the table used to reduce binary images
 * by a factor of 4, producing one 4 bit gray pixel for each 4x4 square of
 * binary pixels.  The procedure is PUBLIC because it can be called by
 * upper level routines directly when they wish to provide the lower routines
 * with a ready-made table instead of requring them to remake the table
 * on each call.
 *
 * The table is indexed with a byte.
 * It returns a 32-bit word containing two 8-bit numbers each composed
 * of 4 bits of integer and 4 bits of fraction.  These
 * are used as described in the comments to i_scgr44.c and miscgr44.c.
 * The goal is to turn each 4x4 square of binary pixels into a single 4-bit
 * pixel that takes values from 0 through 15.  The values as a function of
 * number of active pixels is:
 *
 * Number of active pixels	Output value
 *		0		0
 *		1		15/16
 *		2		2*15/16
 *		3		3*15/16
 *		...		...
 *		15		15*15/16
 *		16		16*15/16
 *
 * Since there are 4 fraction bits, we can ignore the "/16" in the calculation
 * above when making the table.  We'll toss the fraction bits when the final
 * image is made.
 *
 * For the sake of efficiency, we would like to process as many pixels at
 * once as possible.  Table lookup is a good way to space out the pixels and
 * do some computing.  On the other hand, we don't like to make huge tables,
 * so we like to restrict the number of pixels used to index the tables.
 * We often use 8 as a nice table index size.
 *
 * This routine computes a table containing these weights.
 */

/* PUBLIC */
/**************************************************************************
 *  i_getSizeOfScaleToGray4bpp4xTab()  
 *		Return the size of a table used to scale binary
 *		images to 4bpp by 4x.
 *
 *	Input:
 *		pSize =		ptr to returned size of table
 *
 *	Return code:
 *		ia_successful	good completion
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfScaleToGray4bpp4xTab (
    IPState    *pIPState,
    UInt32     *pSize)
{
    PROLOGUE;
    *pSize = 256 * sizeof(UInt32);
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_fillScaleToGray4bpp4xTab()  
 *		Each 4x4 square of the source binary image is mapped
 *		to a 4-bit output pixel.  The value of the output
 *		pixel is a (roughly) linear function of the number of
 *		bits turned on in the 4x4 input square.
 *
 *	Input:
 *		pTable =	ptr to allocated table.
 *		tabSize =	size of allocated table in bytes.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_invalidParm	table passed in is too small or not there
 *
 **************************************************************************/
Int32  CDECL 
i_fillScaleToGray4bpp4xTab (
    IPState    *pIPState,
    UInt32     *pTable,
    UInt32	tabSize)
{
    /* This is the table of weights used to create the output image.
     * NOTE: Since 0 = white in binary images and 0 = black in gray
     * images, we do an inversion by counting the number of binary 0's
     * instead of the number of binary 1's.
     */
    PROLOGUE;
    UInt32	 index, bitsHi, bitsLo;		/* index into table and the
						 * number of 0's (white bits)
						 * in the constituent
						 * nibbles of the index */
	/* this table holds the number of white bits in each index */
    UInt8	 numBits[16] = {4, 3, 3, 2,	/* 0, 1, 2, 3 */
				3, 2, 2, 1,	/* 4, 5, 6, 7 */
				3, 2, 2, 1,	/* 8, 9, A, B */
				2, 1, 1, 0};	/* C, D, E, F */
    UInt32	 realSize;			/* used to check pTable */

	/* Figure out how big a table is needed */
    i_getSizeOfScaleToGray4bpp4xTab (pIPState, &realSize);

    if ((pTable == NULL) || (realSize > tabSize))
	return ia_invalidParm;

	/* fill the table.  Each byte contains two nibbles.  We compute
	 * the index for each nibble, then combine them.
	 */
    for (index = 0; index < 256; index++)
    {
        bitsLo = numBits[index & 0xF];
	bitsHi = numBits[index >> 4];
	pTable[index] = (((bitsHi * 16) - bitsHi) << 8) | /* 15 * bitsHi */
			 ((bitsLo * 16) - bitsLo);	  /* 15 * bitsLo */
    }

    NORMAL_RETURN;	/* That's it */

}

/*

$Log: t_scgr44.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.7  1998/01/13 23:54:27  ddavies
t_scgr44.c -> /product/ipcore/ipshared/src/RCS/t_scgr44.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1996/03/26 23:20:44  ddavies
t_scgr44.c -> /product/ipcore/ipshared/src/RCS/t_scgr44.c,v

Add i_getSizeOfScaleToGray4bpp4xTab, which returns the size of the table
used by i_ScaleToGray4bpp4x.  Also change i_makeScaleToGray4bpp4xTab
into i_fillScaleToGray4bpp4xTab, which only fills the table.  The new
routine does not allocate the table.

Revision 1.5  1995/12/13 02:53:08  ejaquith
t_scgr44.c -> /product/ipcore/ipshared/src/RCS/t_scgr44.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/02/04  20:22:06  dferrell
 * t_scgr44.c -> /product/ipcore/ipshared/src/RCS/t_scgr44.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:30:30  ddavies
 * t_scgr44.c -> /product/ipcore/ipshared/src/RCS/t_scgr44.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  02:26:28  ddavies
 * t_scgr44.c -> /product/ipcore/ipshared/src/RCS/t_scgr44.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


