/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  t_scgr48.c: machine independent routine to make the table used to
 *		scale to 4bpp gray, 8x reduction
 *
 *          8x scale to gray reduction:
 *                  Int32	i_getSizeOfScaleToGray4bpp8xTab()
 *                  Int32	i_fillScaleToGray4bpp8xTab()
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscal.pub"
#include "shrscal.prv"
#include "imageref.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: t_scgr48.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $");

/* default: suppress all debugging */
#define  TIMING        0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING        1
#endif  /* DEBUG_ON */



/* This procedure is used to compute the table used to reduce binary images
 * by a factor of 8, producing one 4 bit gray pixel for each 8x8 square of
 * binary pixels.  The procedure is PUBLIC because it can be called by
 * upper level routines directly when they wish to provide the lower routines
 * with a ready-made table instead of requring them to remake the table
 * on each call.
 *
 * The table is indexed with a byte.
 * It returns a 32-bit word containing one 16-bit number composed
 * of 8 bits of integer and 8 bits of fraction.  These numbers
 * are used as described in the comments to i_scgr48.c and miscgr48.c.
 * The goal is to turn each 8x8 square of binary pixels into a single 4-bit
 * pixel that takes values from 0 through 15.  The values as a function of
 * number of active pixels is:
 *
 * Number of active pixels	Output value
 *		0		0
 *		1		15/64
 *		2		2*15/64
 *		3		3*15/64
 *		...		...
 *		63		63*15/64
 *		64		64*15/64
 *
 * Since there are 8 fraction bits, we generate the "/64" with a *left* shift
 * by 2.  We use 8 fraction bits to keep the byte sums byte aligned within
 * the sum word..
 *
 * For the sake of efficiency, we would like to process as many pixels at
 * once as possible.  Table lookup is a good way to space out the pixels and
 * do some computing.  On the other hand, we don't like to make huge tables,
 * so we like to restrict the number of pixels used to index the tables.
 * We often use 8 as a nice table index size.
 *
 * This routine computes a table containing these weights.
 */

/* PUBLIC */
/**************************************************************************
 *  i_getSizeOfScaleToGray4bpp8xTab()  
 *		Return the size of a table used to scale binary
 *		images to 4bpp by 8x.
 *
 *	Input:
 *		pSize =		ptr to returned size of table
 *
 *	Return code:
 *		ia_successful	good completion
 *
 **************************************************************************/
Int32  CDECL 
i_getSizeOfScaleToGray4bpp8xTab (
    IPState    *pIPState,
    UInt32     *pSize)
{
    PROLOGUE;
    *pSize = 256 * sizeof(UInt32);
    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_fillScaleToGray4bpp8xTab()  
 *		Each 8x8 square of the source binary image is mapped
 *		to a 4-bit output pixel.  The value of the output
 *		pixel is a (roughly) linear function of the number of
 *		bits turned on in the 8x8 input square.
 *
 *	Input:
 *		pTable =	ptr to allocated table.
 *		tabSize =	size of allocated table in bytes.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_invalidParm	table passed in is too small or not there
 *
 **************************************************************************/
Int32  CDECL 
i_fillScaleToGray4bpp8xTab (
    IPState    *pIPState,
    UInt32     *pTable,
    UInt32	tabSize)
{
    /* This is the table of weights used to create the output image.
     * NOTE: Since 0 = white in binary images and 0 = black in gray
     * images, we do an inversion by counting the number of binary 0's
     * instead of the number of binary 1's.
     */
    PROLOGUE;
    UInt32	 index, bits;			/* index into table and the
						 * number of 0's (white bits)
						 * in the index */
    UInt8 numBits[256] = {8, 7, 7, 6, 7, 6, 6, 5,	/* 00 .. 07 */
			  7, 6, 6, 5, 6, 5, 5, 4,	/* 08 .. 0F */
			  7, 6, 6, 5, 6, 5, 5, 4,	/* 10 .. 17 */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* 18 .. 1F */
			  7, 6, 6, 5, 6, 5, 5, 4,	/* 20 .. 27 */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* 28 .. 2F */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* 30 .. 37 */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* 38 .. 3F */
			  7, 6, 6, 5, 6, 5, 5, 4,	/* 40 .. 47 */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* 48 .. 4F */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* 50 .. 57 */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* 58 .. 5F */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* 60 .. 67 */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* 68 .. 6F */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* 70 .. 7F */
			  4, 3, 3, 2, 3, 2, 2, 1,	/* 78 .. 7F */
			  7, 6, 6, 5, 6, 5, 5, 4,	/* 80 .. 87 */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* 88 .. 8F */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* 90 .. 97 */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* 98 .. 9F */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* A0 .. A7 */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* A8 .. AF */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* B0 .. BF */
			  4, 3, 3, 2, 3, 2, 2, 1,	/* B8 .. BF */
			  6, 5, 5, 4, 5, 4, 4, 3,	/* C0 .. C7 */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* C8 .. CF */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* D0 .. DF */
			  4, 3, 3, 2, 3, 2, 2, 1,	/* D8 .. DF */
			  5, 4, 4, 3, 4, 3, 3, 2,	/* E0 .. EF */
			  4, 3, 3, 2, 3, 2, 2, 1,	/* E8 .. EF */
			  4, 3, 3, 2, 3, 2, 2, 1,	/* F0 .. F7 */
			  3, 2, 2, 1, 2, 1, 1, 0};	/* F8 .. FF */
    UInt32	 realSize;			/* used to check pTable */


	/* Figure out how big a table is needed */
    i_getSizeOfScaleToGray4bpp8xTab (pIPState, &realSize);

    if ((pTable == NULL) || (realSize > tabSize))
	return ia_invalidParm;

	/* fill the table.  Each word contains a two-byte number.  We compute
	 * the index for each nibble, then combine them.
	 */
    for (index = 0; index < 256; index++)
    {
	bits = numBits[index];
	pTable[index] = (((bits * 16) - bits) << 2);	/* a byte aligned
							 * version of 15*bits */
    }

    NORMAL_RETURN;	/* That's it */

}

/*

$Log: t_scgr48.c,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.7  1998/01/13 23:54:29  ddavies
t_scgr48.c -> /product/ipcore/ipshared/src/RCS/t_scgr48.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1996/03/26 23:21:23  ddavies
t_scgr48.c -> /product/ipcore/ipshared/src/RCS/t_scgr48.c,v

Add i_getSizeOfScaleToGray4bpp8xTab, which returns the size of the table
used by i_ScaleToGray4bpp8x.  Also change i_makeScaleToGray4bpp8xTab
into i_fillScaleToGray4bpp8xTab, which only fills the table.  The new
routine does not allocate the table.

Revision 1.5  1995/12/13 02:53:11  ejaquith
t_scgr48.c -> /product/ipcore/ipshared/src/RCS/t_scgr48.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/02/04  20:22:11  dferrell
 * t_scgr48.c -> /product/ipcore/ipshared/src/RCS/t_scgr48.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:30:35  ddavies
 * t_scgr48.c -> /product/ipcore/ipshared/src/RCS/t_scgr48.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  02:26:35  ddavies
 * t_scgr48.c -> /product/ipcore/ipshared/src/RCS/t_scgr48.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


