/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1996, Xerox Corporation.                       *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/



/*
 *  arithops.c--basic arithmetic operations. 
 *
 *           w_arithOpConst () - Constant operation on an image
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "geomadt.h"
#include "pixr.h"
#include "arithops.h"

/* PROTOTYPES */
#include "props.pub"
#include "prxform.pub"
#include "arithops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: arithops.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:13 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform an arithmetic operation between an image and a constant
 *
 * LongDescription:::
 *  Performs some constant arithmetic operation on a gray image. The
 *  operations supported are cAdd, cSubtract, and cScale. All operations
 *  will clip at 0.0 and 1.0 (where 1.0 is 15 or 255, depending on the
 *  depth of the image, 4 or 8 bit). All operations are conceptually
 *  done by converting the image pixel to floating-point and doing
 *  the computation with the value given, with no scaling of the value
 *  (i.e. a number added or subtracted is in the scale of the image
 *  pixels to which it is applied).
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  op::             cAddOp, cSubtractOp, or cScaleOp
 *  value::          The constant value used in the operation.
 *
 * Output:::
 *  prD::            The image on which to operate
 *
 * Notes:::
 *  Uses a look-up table at the moment, which is OK for scaling, but 
 *  rather inefficient for addition/subtraction. These should be 
 *  made into primitives if speed is an issue.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. MacLeod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_arithOpConst (
    IPWState           *pIPWState,
    Int32               op,
    Float64             value,
    PIXR               *prD)
{
    PROLOGUE;
    Int32         dd, chD, levels, i, roundedValue;
    Float64       resultValue;
    UInt8        *pMap = NULL;

    ERR_GEN(prD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prD, &dd));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prD, &chD));

    levels = (1 << dd);
    
    pMap = IPW_CALLOC (levels, sizeof(UInt8), cAllocatorData);
    ERR_GEN (pMap == NULL, cErrNomem);

    /* Fill the table. This is not very efficient, but it's only
       a wee table */
    for (i = 0; i < levels; i++)
    {
	if (op == cAdd)
	{
	    resultValue = ((Float64) i) + value;
	}
	else if (op == cSubtract)
	{
	    resultValue = ((Float64) i) - value;
	}
	else if (op == cScale)
	{
	    resultValue = ((Float64) i) * value;
	}
	else
	{
	    ERR_GEN_CLEANUP (-1, cErrInvalidParm);
	}
	
	/* Clip, round, etc. */
	roundedValue = (Int32) (resultValue + 0.5);
	if (roundedValue < 0)
	    roundedValue = 0;
	if (roundedValue >= levels)
	    roundedValue = levels - 1;

	pMap[i] = (UInt8) roundedValue;
	
    }

    /* Do the lookup */
    ERR_PASS_CLEANUP (w_map(pIPWState, prD, pMap, &prD));

    /* And give back the table */
    IPW_FREE (pMap, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (pMap)
	IPW_FREE (pMap, cAllocatorData);
    ERR_RETURN;
}

/*

 $Log: arithops.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:13  jlin
 Imported sources

 Revision 1.3  1997/01/02 15:48:49  macleod
 macleod on Thu Jan 2 07:48:41 PST 1997

 arithops.c -> /product/ipcore/ipwrap/src/RCS/arithops.c,v

 Do it right this time...

 Revision 1.2  1997/01/02 15:47:20  macleod
 macleod on Thu Jan 2 07:47:07 PST 1997

 arithops.c -> /product/ipcore/ipwrap/src/RCS/arithops.c,v

 Get rid of NT warning

 Revision 1.1  1996/12/30 15:15:36  macleod
 Initial revision


 */

