/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * bctable.c - brightness/contrast/
 *
 *	makeBrightnessContrastTable ()	-   Make a gamma/contrast table
 *	makeLinearTable () - Make a linear re-ranging table 
 *
 */

#include <math.h>

#include "ipwcore.prv"
#include "spline.pub"
#include "bctable.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: bctable.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:14 $");

#define cSplinePoints	5

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a brightness/contrast look-up table
 *
 * LongDescription:::
 *  Given brightness and contrast values, make a look-up table which
 *  maps 8-bit numbers. The brightness value is just what some call
 *  "gamma", i.e. out = pow(in, 1/gamma) where gamma=brightness value.
 *  Brightness = 1.0 does nothing, brightness > 1.0 "brightens" the
 *  picure, brightness < 1.0 "darkens" the picture.
 *  Contrast does nothing at 0.0, increases contrast > 0.0, decreases
 *  contrast < 0.0. Its useful range is around -1 to 1, but is
 *  defined everywhere.

 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  brightness::     Brightness value, 0.0 to n, useful around 0-4 
 *  contrast::       Contrast value, -1.0 to 1.0. 
 *  tableLength::    Table length, up to 256
 *  range::          Maximum value for the table's range (0...range)
 *  invert::         Invert the curve?  (TRUE for dithering op)
 *
 * Output:::
 *  ppTable::        Table created
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  tone correction
 *
 * Keywords:::
 *  tone correction
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeBrightnessContrastTable (
    IPWState           *pIPWState,
    Float64             brightness,
    Float64             contrast,
    Int32               tableLength,
    Int32               range,
    Int32               invert,
    UInt8             **ppTable)
{
    PROLOGUE;
    Float64	   contrastX[cSplinePoints + 1], contrastY[cSplinePoints + 1], 
		   power, value, in, rise, run;
    Float64	   contrastCurve[256], *gradients = NULL, ctmp[257];
    Int32	   i, tldiv2, tlm1;
    UInt8         *table;

    if (*ppTable == NULL)
    {
	*ppTable = IPW_CALLOC (tableLength, sizeof(Int32), cAllocatorData);
	ERR_GEN (!*ppTable, cErrNomem);
    }
    
    table = *ppTable;
    tldiv2 = tableLength / 2;
    tlm1 = tableLength - 1;

    /* Fill in the contrast points for the interpolating spline */
    /* The idea is that 1.0 gives a large slope, and -1.0 gives
       a small slope. */
    contrastX[1] =   0.0; contrastY[1] =   0.0;

    /* Square the parameter, to give it a better-feeling low end scale */
    rise = contrast * contrast * 10 + 2.0;
    run = 2.0;
    
    contrastX[2] = (Float64) tldiv2 - run / 2;
    contrastY[2] = (Float64) tldiv2 - rise / 2;
    contrastX[3] = (Float64) tldiv2;
    contrastY[3] = (Float64) tldiv2;
    contrastX[4] = (Float64) tldiv2 + run / 2;
    contrastY[4] = (Float64) tldiv2 + rise / 2;
    contrastX[5] = (Float64) tlm1; 
    contrastY[5] = (Float64) tlm1;
    contrastX[0] = contrastY[0] = 0;
    /* Make the gradients arary */
#if 0   
    ERR_PASS( w_spline (pIPWState, &contrastX[0], &contrastY[0], cSplinePoints, 
		     1.0e30, 1.0e30, y2));
#else
    ERR_PASS (w_cubicSplineGradients (pIPWState, &contrastX[0], &contrastY[0],
				      1, cSplinePoints, &gradients));
#endif

    /* Make a look-up table */
    for (i = 0; i < tableLength; i++)
    {
	in = i;
#if 0
	ERR_PASS( w_splint (pIPWState, &contrastX[0], &contrastY[0], y2, 
			 cSplinePoints, in, &value));
#else
	ERR_PASS( w_cubicSplineInterpolate (pIPWState, 
					    &contrastX[0], 
					    &contrastY[0], gradients,
					    1, 5, in, &value));
#endif

	if (contrast < 0)
	{
	    if ((i < tldiv2 && value > tldiv2) || 
		(i > tldiv2 && value < tldiv2))
		value = (Float64) tldiv2;
	}
	if (value < 0)
	    value = 0;
	if (value > tlm1)
	    value = tlm1;
	contrastCurve[i] = value;
    }

    /* Invert and fill in missing codes, if needed */
    if (contrast < 0)
    {
	for (i = 0; i < tableLength; i++)
	    ctmp[i] = -1;

	for (i = 0; i < tableLength; i++)
	    ctmp[(Int32) contrastCurve[i]] = i;

	for (i = 0; i < tableLength; i++)
	{
	    if (ctmp[i] < 0)
	    {
		if (i)
		    ctmp[i] = ctmp[i - 1];
		else
		    ctmp[i] = 0;
	    }
	    contrastCurve[i] = ctmp[i];
	}
    }

    /* Brightness is inverse of "gamma" */
    power = 1.0 / brightness;

    /* Fill the table */
    for (i = 0; i < tableLength; i++)
    {
	/* First, do brightness */
	value = i / (Float64) tlm1;
	value = pow(value, power) * (Float64) tlm1;

	/* Now, do contrast */
	value = contrastCurve[(Int32) value];

	if (value < 0)
	    value = 0;
	if (value > (Float64) tlm1)
	    value = (Float64) tlm1;

        if (invert)
	    table[tableLength - i] = (UInt8)((range * value / tableLength) +
					      0.5);
	else
	    table[i] = (UInt8)((range * value / tableLength) + 0.5);
    }

    /* Free the gradients array */
    IPW_FREE (gradients, cAllocatorData);
    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a linear look-up table for re-ranging input values
 *
 * LongDescription:::
 *  Make a linear look-up table for re-ranging input values
 *  The typical use of this is to expand an image for a given
 *  black/white point.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  tableLength::    # of entries in table
 *  blackPoint::     All this and below will map to zero
 *  whitePoint::     All this an above will map to tableLength-1
 *
 * Output:::
 *  ppTable::        Table created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  tone correction
 *
 * Keywords:::
 *  tone correction
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeLinearTable (
    IPWState           *pIPWState,
    Int32               tableLength,
    Int32               blackPoint,
    Int32               whitePoint,
    UInt8             **ppTable)
{
    PROLOGUE;
    Int32   i, delta;
    UInt8   *table;

    if (*ppTable == NULL)
    {
	*ppTable = IPW_CALLOC (tableLength, sizeof(Int32), cAllocatorData);
	ERR_GEN (!ppTable, cErrNomem);
    }

    table = *ppTable;
	
    delta = (whitePoint - blackPoint);
 
    for (i = 0; i < tableLength; i++)
    {
	if (i <= blackPoint)
	    table[i] = 0;
	else if (i >= whitePoint)
	    table[i] = (UInt8) (tableLength - 1);
	else
	    table[i] = (UInt8) (((i - blackPoint) * tableLength) / delta);
    }

    NORMAL_RETURN;
}

/*
 * $Log: bctable.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:14  jlin
 * Imported sources
 *
 * Revision 1.26  1997/05/24 01:25:23  jcraig
 * jcraig on Fri May 23 18:23:54 PDT 1997
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * fix bug in w_makeLinearTable.  previously if memory for table was not
 * allocated a memory fault occurred.
 *
 * Revision 1.25  1996/01/04 20:20:57  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.24  1995/12/13  04:04:38  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.23  1995/11/13  20:09:15  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.22  1995/10/13  21:58:22  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.21  1995/10/12  04:23:33  ddavies
 * ddavies on Wed Oct 11 21:23:04 PDT 1995
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.20  1995/10/05  21:49:59  macleod
 * macleod on Thu Oct 5 14:49:36 PDT 1995
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Always clear the entire array for spline interpolation.
 *
 * Revision 1.19  1995/10/04  18:30:21  macleod
 * macleod on Wed Oct 4 11:30:02 PDT 1995
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Args to w_cubicSplineGradients changed, so changed call
 *
 * Revision 1.18  1995/10/03  21:09:02  macleod
 * macleod on Tue Oct 3 14:08:46 PDT 1995
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Changed calls to cubic spline interpolation
 * Also updated comment blocks
 *
 * Revision 1.17  1995/09/21  22:51:12  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/11  21:56:12  macleod
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.15  1995/09/07  01:29:22  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.14  1995/09/06  20:16:58  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.13  1995/09/01  20:44:30  macleod
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Added back #include files
 *
 * Revision 1.12  1995/08/31  21:10:48  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.11  1995/08/25  01:34:52  ejaquith
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Fixed warnings
 *
 * Revision 1.10  1995/08/17  21:53:27  macleod
 * macleod on Thu Aug 17 14:53:23 PDT 1995
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.9  1995/01/13  01:54:12  macleod
 * macleod on Thu Jan 12 17:53:50 PST 1995
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * makeLinearTable scales the table output to the table size, rather than
 * arbitrarily setting it at 256.
 *
 * Revision 1.8  1994/10/10  21:16:06  macleod
 * macleod on Mon Oct 10 14:14:51 PDT 1994
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Added makeLinearTable, which, given a black point, white point, and
 * look-up table, fills the look-up table such that everything <= the
 * black point maps to zero, everything >= the white point maps to the
 * maximum value, and values in between are linearly interpolated.
 *
 * Revision 1.7  1994/10/06  20:53:51  lperry
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.6  1994/06/08  04:06:36  ddavies
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.5  1994/05/20  22:40:59  dferrell
 * dferrell on Fri May 20 15:36:13 PDT 1994
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Added range and invert parameters to makeBrightnessContrastTable so
 * that it can be used to generate DRC tables for dithering.  To make
 * it function as it originally did, pass in the the table size as the
 * range and do not invert the table.
 *
 * Revision 1.4  1994/02/10  05:34:54  macleod
 * macleod on Wed Feb 9 21:34:15 PST 1994
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * A header file wasn't being included, so one of the functions
 * which was declared CDECL was getting an undefined reference on the
 * PC.
 *
 * Revision 1.3  1994/01/22  01:35:43  ddavies
 * ddavies on Fri Jan 21 17:35:08 PST 1994
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Remove unused variables offset and ii that generated warnings in WATCOM.
 *
 * Revision 1.2  1994/01/06  22:49:41  macleod
 * macleod on Thu Jan 6 14:49:17 PST 1994
 *
 * bctable.c -> /product/ipcore/ipwrap/src/RCS/bctable.c,v
 *
 * Added a table-size parameter, so one can make tables up to 256
 * entries
 *
 * Revision 1.1  1993/12/21  23:22:30  macleod
 * Initial revision
 *
 */
