/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*************************************************************************/
/*                                                                       */
/*                         Modified Berlekamp-Massey                     */
/*                         Error Correcting Decoder                      */
/*                                    from                               */
/*                     George C. Clark, Jr and J. Bibb Cain              */
/*            "Error-Correction Coding For Digital Communications"       */
/*                               Pages 205 - 219.                        */
/*              Basic Decoder Flow Outlined in Table 5-11 pg. 218        */
/*                                                                       */
/*  Programmer   : Henry Minsky                                          */
/*                                                                       */
/*  Version Number       Version Date              Version Comments      */
/*     1.00.00           Jun 01, 1991                Original Code       */
/*     1.00.01           Feb 17, 1992         GWP : Inline Comments      */
/*                       August  1992         REW : Parameterized,       */
/*                                                  separated error      */
/*                                                  detection from       */
/*                                                  correction, ANSIfied */
/*                                                                       */
/*  Notation :                                                           */
/*             x_y => "x sub y" or yth element of x                      */
/*                                                                       */
/*************************************************************************/

#include <stdlib.h>
#include <string.h>


#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"


#define _BERLEKAM_C_DEFINED
#include "shrglyph.pub"
#include "dgcom.prv"
#include "dgecc.prv"
#include "glydrend.prv"
#include "glydebug.prv"
#include "shrutils.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: berlekam.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:14 $");


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Detect and correct, if possible, errors in the given data
 *
 * LongDescription:::
 * Main controller for ECC decoding and error correction.
 * Combined erasure and error magnitude computation.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::			pointer to IPCore state
 *  block_nr::       		index of codeword currently being corrected
 *  csize::          		size in bytes of codeword
 *  nerasures::      		number of erasures found in this codeword
 *  erasures::       		erasure array
 *  nr_parity::      		number of ecc bytes in this codeword
 *  syn_bytes::      		syndrome bytes
 *  omega::          		polynomial used in correction
 *  lambda::         		polynomial used in correction
 *  glog::			log table
 *  gexp::			table of powers of alpha
 *  error_locs::     		arry of bytes with errors
 *  errorProc::    		procedure used to print error info
 *  errorInfo::			private data needed by errorProc
 *
 * Output:::
 *  codeword::       		pointer to data

 * Returns:::
 *  0				no errors found
 *  -1				cannot correct
 *  n				number of errors found and corrected
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * Citations:::
 * Clark and Cain "Error-Correction Coding For Digital Communications"
 * pages 205-219.
 *
 * EndDocumentation:::
 **************************************************************************/
Int32 CDECL
wp_correctErrorsErasures (
    IPWState	       *pIPWState, 
    Int32               block_nr,
    UInt8              *codeword,
    Int32               csize,
    Int32               nerasures,
    Int32              *erasures,
    Int32               nr_parity,
    UInt8              *syn_bytes,
    UInt8              *omega,
    UInt8              *lambda,
    UInt32	       *glog,
    UInt8	       *gexp,
    Int32              *error_locs,
    ClientErrorProc     errorProc,
    void	       *errorInfo)
{
    Int32 idx;        /* Loop index                                */
    Int32 status;     /* Return code from function */
    Int32 nerrors;    /* Number of error locations found (Chien's search) */

	/* Keep compiler happy */
    pIPState = pIPState;

    block_nr = block_nr;
    memset((void *)lambda, 0, sizeof(*lambda)*2*(nr_parity+1));

    /* How big a codeword does this symbol size allow? */

    if ( csize > MAX_CODEWORD_SIZE ) 
    {
        /* Can't encode so much data with this symbol size */
	return -1;
    }

    /* Is the number of erasures within the processing capability? */
    if (nerasures > nr_parity)
	return -1;


    /* Reverse the indexing of the erasures */
    if ( nerasures > 0 )
    {
        for (idx = 0; idx < nerasures; idx++)
            erasures[idx] = csize - 1 - erasures[idx];
    }

   /* Perform Basic Decoder Operation- Returns Error-Location polynomial*/
    if (ERR_IGNORE(wp_calSynBytes (pIPWState, codeword, csize, glog, gexp, 
				   syn_bytes, nr_parity)) == 0)
    	return 0;
    ERR_IGNORE(wp_modifiedBerlekampMassey (pIPWState, nerasures, erasures,
					   nr_parity,
					   glog, gexp, syn_bytes, omega, 
					   lambda));
#if 0
    {
    Int32	 nonZero = 0;

    for (idx = 1; idx < (2*(nr_parity+1)); idx++)
	if (lambda[idx] != 0) nonZero++;
    if ((nonZero-nerasures) > ((nr_parity-nerasures)/2))
    {
	printf("Returning with incorrect lambda %d. %d erasures. %d errors max\n", nonZero, nerasures, (nr_parity-nerasures)/2);
	return -1;
    }
    }
#endif

    nerrors = ERR_IGNORE(wp_findRoots (pIPWState, nr_parity,
				       error_locs, lambda, glog,
				       gexp));

    /*  Were any errors detected? */
    if (((nerrors-nerasures) <= ((nr_parity-nerasures)/2))  && 
        (nerrors > 0))
    { 			/*  Yes, and they can be corrected */

	/* first check for illegal error locs */
	for (idx = 0; idx < nerrors; idx++)
	    if (error_locs[idx] >= csize)
		return -1;

#ifdef GLY_INFO
	if (getenv(ENV_CORE_STATS))
	{
#if 0
        char string[80];
#endif
        Int32 errors[256];
            /* Restore the order of the erasures */
            for (idx = 0; idx < nerasures; idx++)
                erasures[idx] = csize - 1 - erasures[idx];
#if 0
            if ( nerasures > 0 )
            {
                sprintf(string, "Block %ld, erasures: ", block_nr);
                wp_printDecimal (pIPWState, errorProc, errorInfo,
				 string, erasures, nerasures);
            }
#endif
            /* Reverse-index the error locations */
            for (idx = 0; idx < nerrors; idx++)
                errors[idx] = csize - 1 - error_locs[idx];
#if 0
            sprintf(string, "Block %ld, error locations: ", block_nr);
            wp_printDecimal (pIPWState, errorProc, errorInfo,
			     string, errors, nerrors);
#endif
	}
#endif /* GLY_INFO */

        ERR_IGNORE(wp_correctErrors (pIPWState, codeword, csize,
				     error_locs, nerrors, 
				     nr_parity, omega, lambda, glog, gexp,
				     errorProc, errorInfo));
        status = nerrors;
    } 

    else {
        /* Too many errors */

#ifdef GLY_INFO
	if (getenv(ENV_CORE_STATS))
            i_printf (pIPWState, errorProc, errorInfo,
		     "Too many errors in block %ld: %ld", block_nr, nerrors);
#endif /* GLY_INFO */
        status = -1;
    }
    return status;
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Calculate syndrome bytes for this codeword and determine if all zero
 *
 * PrivateDescription:::
 * This procedure calculates all the syndrome bytes for the given 
 * codeword.  It also determines if all the syndrome bytes are 0.
 * (If all the syndrome bytes were 0 then the codeword was received 
 * without errors.)
 * There are as many syndrome bytes as there are ecc bytes.  The
 * ith syndrome byte is computed by dividing the received codeword
 * by the ith factor of the generator polynomial.
 * Once again we can do this with shift registers.  Fortunately, this
 * is a very simple shift register since the ith factor of the 
 * generator polynomial is (x + alpha^i).  As a result, this shift
 * register has only one element.
 * To calculate the ith syndrome byte:  multiply the current shift 
 * register contents by alpha^(i+1).  exclusive or this with the next 
 * codeword byte.  this becomes the new shift register value.  do these
 * steps for each codeword byte.
 *
 * Input:::
 *  pIPWState::			pointer to IPCore state
 *  data::           		ptr to the codeword
 *  nbytes::         		codeword size
 *  glog::			log table
 *  gexp::			antilog table
 *  nr_parity::      		number of ecc bytes per codeword
 *
 * Output:::
 *  syn_bytes::      		syndrome bytes
 *
 * Returns:::
 *  0				if all zeroes in syndrome bytes
 *  n				if syndrome bytes not all zero
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * EndDocumentation:::
 **************************************************************************/
static UInt8 CDECL
wp_calSynBytes (
    IPWState	       *pIPWState,
    UInt8              *data,
    Int32               nbytes,
    UInt32	       *glog,
    UInt8	       *gexp,
    UInt8              *syn_bytes,
    Int32               nr_parity)
{
    Int32 i;        /* loop index                               */
    Int32 syn;      /* loop index                               */
    UInt8 sum;      /* Intermediate sum for cal of syndrome     */
    UInt8 total;    /* Sum of syndrome bytes                    */

    for (syn = 0, total = 0; syn < nr_parity;  syn++) 
    {
        for (i = 0, sum = 0; i < nbytes; i++) 
	    /* multiply the current shift register contents with
	     * alpha^(syn+1).  note that we must add 1 to syn since
	     * the generator polynomial was built with roots
	     * alpha^1, alpha^2, ... alpha^nr_parity.
	     * then exclusive or this with the current codeword byte.
	     * this result is the new shift register contents.
	     */
            sum = data[i] ^ ERR_IGNORE(wp_gMultVal (pIPWState,
						    gexp[syn+1], sum,
						    glog, gexp));
        syn_bytes[syn]  = sum;
	/* all we care about is whether all the syndrome bytes are zero
	 * so just or them all together.  if any of the syndrome bytes
 	 * are nonzero the result will be nonzero.  if all the syndrome
	 * bytes are zero, the result will be zero.
	 */
        total |= sum;
    }
    return total;
}


#ifdef GLY_INFO
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Calculate number of on bits in the given word
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  inWord::       	data word  
 *
 * Returns:::
 *  n			number of on bits
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static Int32 CDECL
wp_onBits (
    IPWState	       *pIPWState,
    UInt32              inWord)
{
    Int32	count;	/* number of on bits */
    Int32	bit;	/* counter */

	/* Keep compiler happy */
    pIPWState = pIPWState;

    count = 0;
    for (bit = 0; bit < 32; bit++, inWord = inWord >> 1)
	if (inWord & 1)
	    count++;
    return count;
} 
#endif

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Correct errors in specified locations in codeword
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  csize::          		number of bytes in each codeword
 *  error_locs::     		array of error locations
 *  nerrors::        		number of bytes in error
 *  nr_parity::      		number of ecc bytes per codeword
 *  omega::          		error correction polynomial
 *  lambda::         		error correction polynomial
 *  glog::			log table
 *  gexp::			antilog table
 *  errorProc::    		procedure used to print error info
 *  errorInfo::			private data needed by errorProc
 *
 * Output:::
 *  codeword::       		codeword in which to correct errors
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL
wp_correctErrors (
    IPWState	       *pIPWState,
    UInt8              *codeword,
    Int32               csize,
    Int32              *error_locs,
    Int32               nerrors,
    Int32               nr_parity,
    UInt8              *omega,
    UInt8              *lambda,
    UInt32	       *glog,
    UInt8	       *gexp,
    ClientErrorProc     errorProc,
    void	       *errorInfo)
{
    PROLOGUE;
    Int32 idx;        /* Loop index                                */
    Int32 rdx;        /* Loop index                                */
    Int32 jdx;        /* Loop index                                */
    Int32 err;        /* Error flag                                */
    UInt8 num;        /* */
    UInt8 denom;      /* */
    Int32 i;
#ifdef GLY_INFO
    char  *printCorrections = getenv(ENV_CORE_STATS);
#endif /* GLY_INFO */

	/* Keep compiler happy */
    __error__ = __error__;

    for (rdx = 0; rdx < nerrors; rdx++) 
    {
        idx = error_locs[rdx];

        /* Evaluate Omega/Lambda' at the roots alpha^(-idx) */

	/* first find Omega(alpha^(-idx) */
        num = 0;
        for (jdx = 0; jdx < nr_parity; jdx++) 
            num ^= ERR_IGNORE(wp_gMultVal (
		     pIPWState, omega[jdx],
		     gexp[((MAX_CODEWORD_SIZE-idx)*jdx) % MAX_CODEWORD_SIZE],
		     glog, gexp));

        /* evaluate Lambda' (derivative) at alpha^(-idx).
	 * When taking the derivative of lambda, the coefficients of
 	 * the even powers become 0 (these are the odd powers after
	 * differentiating).  
	 * Why?  Say lambda = alpha^5 * z^3 + alpha^3 * z^2
	 * Then lambda' = 3 * alpha^5 * z^2 + 2 * alpha^3 * z
	 * = (alpha^5 + alpha^5 + alpha^5) * z^2 + (alpha^3 + alpha^3) * z
	 * = alpha^5 * z^2 + 0 * z = alpha^5 * z^2
	 */
        denom = 0;
        for (jdx = 1; jdx < nr_parity; jdx += 2) 
	{
            denom ^= ERR_IGNORE(wp_gMultVal (
			pIPWState, lambda[jdx],
			gexp[((MAX_CODEWORD_SIZE-idx)*(jdx-1)) %
							MAX_CODEWORD_SIZE],
			glog, gexp));
        }
      
        err = ERR_IGNORE(wp_gMultVal (pIPWState, num, 
				      ERR_IGNORE(wp_gInvVal(pIPWState, denom,
							    glog, gexp)), 
				      glog, gexp));
        i = csize-idx-1;
#ifdef GLY_INFO
	if (printCorrections)
            i_printf (pIPWState, errorProc, errorInfo,
		     "Correcting index %ld from %lx to %lx. %ld bits\n", 
                     i, (Int32)codeword[i], (Int32)(codeword[i]^err),
		     ERR_IGNORE(wp_onBits (pIPWState, err)));
#endif /* GLY_INFO */

        codeword[i] ^= err;

    }
    NORMAL_RETURN;
}


/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * ECC Decoder by Modified Berlekamp-Massey Method
 *
 * LongDescription:::
 * 1.0) Set up initial condition
 *      1.a) Compute Erasure Locator Polynomial from known number of erasure
 *           using equation 5-58 ( gamma(z) = product (1-z*a^j_k) ; k = 1, e)
 *      1.b) Compute initial Connection polynomial ( D(z) = z*gamma(z)
 *      1.c) Initialize shift register oldest symbol (K_n) and length (L_n)
 * 2.0) Calculate Connection polynomial
 *      2.a) Calculate discrepancy between existing connection polynomial
 *           and the received syndrome polynomial
 *      2.b) If no discrepancy then existing shift register (i.e. connection
 *           polynomial) is satisfactory. 
 *           If not, then modify shift register (connection polynomial). 
 *      2.c) Test length of feedback shift register compared to iteration
 *           and oldest symbol in register
 *           Condition 1 : Length < iteration - oldest symbol
 *           2.c.1) Compute length of new connection polynomia
 *           2.c.2) Update oldest symbol location
 *           2.c.3) Compute new connection poly Eq : sixth Page : 205 
 *           2.c.4) Update current length to new compute length
 *           End Condition 1 : 
 *      2.d) Update connection polynomial for next iteration 
 *      2.e) Update correction polynomial for next iteration [D_(n+1)=z*D_n]
 *      2.f)  Perform next iteration until interation equals 2t 
 * 3.0) Compute the Error-evaluator polynominal Eq : 5-59 Page : 215
 * 4.0) End basic decoder calculations
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  nerasures::      	number of erasure bytes
 *  erasures::       	array of erasure bytes
 *  nr_parity::      	number of ecc bytes
 *  glog::		log table
 *  gexp::		antilog table
 *
 * Output:::
 *  syn_bytes::      	syndrome bytes
 *  omega::          	error evaluator polynomial
 *  psi::         	error locator polynomial
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * Citations:::
 * Clark and Cain "Error-Correction Coding for Digital Communications"
 * Basic Decoder Flow Outlined in Table  5-11 pg. 218 
 * Flow Chart    in Figure 5-8  pg. 206     
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL
wp_modifiedBerlekampMassey (
    IPWState	       *pIPWState,
    Int32               nerasures,
    Int32              *erasures,
    Int32               nr_parity,
    UInt32	       *glog,
    UInt8	       *gexp,
    UInt8              *syn_bytes,
    UInt8              *omega,
    UInt8              *psi)
{
    PROLOGUE;   
    Int32    n;             /* Current process iteration : 0 <= n <= 2t */
    Int32    L;             /* Length of nth feedback shift register */
    Int32    L2;            /* Length of nth + 1 feedback shift register */
    Int32    k;             /* Oldest symbol in feedback shift register */
    UInt8    d;             /* Discrepancy between con. poly/rcv'ed data */
    Int32    idx;           /* Loop index */
    UInt8    D[MAX_CODEWORD_SIZE+1]; /* Error and Erasure Correction 
			     * Polynominal  */
    UInt8    psi2[MAX_CODEWORD_SIZE+1]; /* temporary error and erasure
			     * locator polynomial */
    Int32    degree;        /* Polynomial degree */

	/* Keep compiler happy */
    __error__ = __error__;

    degree = nr_parity + 1;

    /* Initialize Erasure Locator polynomial:
     * gamma(z) = product (1-z*a^j_k) 
     * where j_k is the location of the kth erasure.
     * Note that even though we are computing gamma, we are putting
     * it into a location called psi.  We do not need gamma and this
     * keeps us from copying it into psi later.
     */
    ERR_IGNORE(wp_init_gamma (pIPWState, psi, nerasures, erasures, degree, glog,
			      gexp));

    /* Initialize Error and Erasure Correction Polynomial 
     * D(z) = z*gamma(z)
     */
    ERR_IGNORE(wp_copyPoly (pIPWState, D, psi, degree));
    ERR_IGNORE(wp_shiftPoly (pIPWState, D, degree));
    
    /* Initialize location of oldest symbol in feedback shift register (k) 
     * and Length (L) of the associated feedback shift register
     */
    k = -1;
    L = nerasures;

    /* Main decoder processing loop.
     * Since the location and number of the erasures is known, the locator
     * and the correction polynomials initial values are preset to 
     * computed values.  In addition, the main decoder loop begins at the
     * number of erasures instead of at 0.   
     * Note that if there were no erasures, main processing loop would
     * begin at 0 and only process unknown errors (and the above code
     * would set the locator and correction polynomials as if only 
     * unknown errors were being processed).
     * This code will only solve any combination of unknown errors
     * (v) and known erasures (e), such that,  2*v + e is less than or equal
     * to the number of parity bytes (2t) of the code: 2*v + e <= 2t.
     * The main loop tires to find the minimum length shift register
     * which generates the first 2t terms of S(z), the syndrome
     * polynomial.  See diagram on page 206 of Clark and Cain.
     * 
     * (Comment below from previous version)
     * WARNING  WARNING  WARNING  WARNING  WARNING  WARNING WARNING
     * ???What happens when nerasures is = to nr_parity  and additional
     * ***error exist?  This is fixed by re-checking the corrected solution 
     * with NErasures set to zero and seeing if nerrors is equal to zero.  
     *  If not, then the wrong solution was found.
     */

    for (n = nerasures; n < nr_parity; n++) 
    {

	/* Compute discrepancy - the difference between the connection
	 * polynomial and the received data.  See equation 5-45 on page
	 * page 205 of Cain and Clark.  Note that since we are
	 * allowing erasures, we are substituting psi for lambda.
	 */
        d = ERR_IGNORE(wp_computeDiscrepancy (pIPWState, psi, syn_bytes,
					      L, n, glog,
					      gexp));

	/* If no discrepancy then existing shift register (i.e. connection
	 * polynomial) is satifactory.  If so, then we need to modify
	 * the shift register.
	 */
        if (d != 0) 
	{
	    /* Compare the length of the current shift register with
	     * the length of the new register.  If the new length is
	     * longer, then we also need to update the oldest symbol, 
	     * the shift register length and the correction polynomial.
	     */
            if (L < (n-k)) 
	    {
	        /* modify the connection polynomial. Eqn: 5-46  Page: 205
	         * psi_n+1 = psi_n - d_n * D_n. 
	         */
                for (idx = 0; idx < degree; idx++)
                    psi2[idx] = psi[idx] ^
                                ERR_IGNORE(wp_gMultVal(pIPWState, d,
						       D[idx], glog,
						       gexp));

		/* the new shift register length */
                L2 = n-k;
		/* location of the oldest symbol in the shift register */
                k = n-L;
                L = L2;
		/* the new correction polynomial. */
                for (idx = 0; idx < degree; idx++) 
                    D[idx] = 
			ERR_IGNORE(wp_gMultVal(pIPWState, psi[idx], 
					       ERR_IGNORE(wp_gInvVal(pIPWState,
								     d,
								     glog,
								     gexp)),
					        glog, gexp));
                ERR_IGNORE(wp_copyPoly (pIPWState, psi, psi2, degree));
            }
	    else
	    {
	        /* modify the connection polynomial. Eqn: 5-46  Page: 205
	         * psi_n+1 = psi_n - d_n * D_n. 
	         */
                for (idx = 0; idx < degree; idx++)
                    psi[idx] = psi[idx] ^
                               ERR_IGNORE(wp_gMultVal(pIPWState, d,
						      D[idx], glog,
						      gexp));
	    }
            
        }  /* If discrepancy */

        /* Update the connection polynomial for the next iteration:
	 * D_(m+1) = z * D_n
	 */
        ERR_IGNORE(wp_shiftPoly (pIPWState, D, degree));

    }  /* for each parity byte */

    /* Compute the error evaluator polynomial. eqn 5-59 page 215:
     * syn_bytes * lambda * gamma = syn_bytes * psi = omega mod z^2t
     */
    ERR_IGNORE(wp_computeModifiedOmega (pIPWState, nr_parity,
					degree, syn_bytes, glog, 
				        gexp, omega, psi));

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Compute the Error-evaluator polynominal
 *
 * LongDescription:::
 * Compute the error evaluator polynomial which is given by:
 * syn_bytes(z) * Lambda(z) * Gamma(z) =
 * syn_bytes(z) * psi(z)               = Omega(z) mod z^2t
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  nr_parity::      	number of ecc bytes
 *  degree::         	degree of polynomials
 *  syn_bytes::      	syndrome bytes
 *  glog::		log table
 *  gexp::		antilog table
 *  lambda::         	error locator polynomial
 *
 * Output:::
 *  omega::          	error evaluator polynomial
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * Citations:::
 * Clark and Cain "Error-Correction Coding for Digital Communications"
 *  eq. : 5-59 / pg. : 215
 *
 * EndDocumentation:::
 **************************************************************************/
static Int32 CDECL
wp_computeModifiedOmega (
    IPWState	       *pIPWState,
    Int32               nr_parity,
    Int32               degree,
    UInt8              *syn_bytes,
    UInt32	       *glog,
    UInt8	       *gexp,
    UInt8              *omega,
    UInt8              *lambda)
{
    Int32   status = cSuccessful;
    Int32   i;

    ERR_IGNORE(wp_multPolys (pIPWState, omega, lambda, syn_bytes, nr_parity,
			     glog, gexp));
    for (i = nr_parity; i < degree; i++)
        omega[i] = 0;
    return status;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Initialize the Erasure Locator polynomial (gamma)
 *
 * PrivateDescription:::
 * gamma[0] = 1 => initial condition
 * gamma[z] = product (1-z*alpha^j_k) where k = 1..e and j_k is the 
 *				      location of the kth erasure
 * 
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  nerasures::      	number of erasure bytes
 *  erasures::       	erasure array
 *  degree::         	degree of polynomial
 *  glog::		log table
 *  gexp::		antilog table
 *
 * Output:::
 *  gamma::          	erasure locator polynomial
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * Citations:::
 * Clark and Cain "Error-Correction Coding for Digital Communications"
 * eq. : 5-58 / pg. : 215 
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL
wp_init_gamma (
    IPWState	       *pIPWState,
    UInt8              *gamma,
    Int32               nerasures,
    Int32              *erasures,
    Int32               degree,
    UInt32	       *glog,
    UInt8	       *gexp)
{
    PROLOGUE;
    Int32    e;      /* Loop index for number of erasures         */
    UInt8    tmp[(MAX_CODEWORD_SIZE+1)*2];   /* Copy of (N-1) product result */

	/* Keep compiler happy */
    __error__ = __error__;

    /* Initialize (clear) gamma array and set first value */
    ERR_IGNORE(wp_zeroPoly (pIPWState, gamma, degree));
    gamma[0] = 1;

    /* Note that this code could execute faster if it didn't loop
     * through the entire polynomial each time through the loop.
     * degree should be set to the degree currently being computed,
     * not the maximum.
     */

    /* Perform product calculation for remaining gamma terms */
    for (e = 0; e < nerasures; e++) {
        ERR_IGNORE(wp_copyPoly (pIPWState, tmp, gamma, degree));
        ERR_IGNORE(wp_scalePoly (pIPWState, gexp[erasures[e]],
				 tmp, degree, glog, gexp));
        ERR_IGNORE(wp_shiftPoly (pIPWState, tmp, degree));
        ERR_IGNORE(wp_addPolys (pIPWState, gamma, tmp, degree));
    }
    NORMAL_RETURN;
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Compute Discrepancy between connection polynomial and data polynomial
 *
 * LongDescription:::
 * Compute Discrepancy between existing connection polynomial and
 * received data polynomial
 * d_n = sum( lamda(i) * s(n-i) ) ; i = 1, L
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  lambda::         	connection polynomial
 *  syn_bytes::      	syndrome bytes
 *  L::              	length of the associated feedback shift register
 *  n::              	iteration index
 *  glog::		log table
 *  gexp::		table of powers of alpha
 *
 * Returns:::
 *  discrepancy
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * Citations:::
 * Clark and Cain "Error-Correction Coding for Digital Communications"
 * eq. : 5-45 / pg. : 205
 *
 * EndDocumentation:::
 **************************************************************************/
static UInt8 CDECL
wp_computeDiscrepancy (
    IPWState	       *pIPWState,
    UInt8              *lambda,
    UInt8              *syn_bytes,
    Int32               L,
    Int32               n,
    UInt32	       *glog,
    UInt8	       *gexp)
{
    Int32 idx;            /* Loop index                                */
    UInt8 sum = 0;        /* Discrepancy summation and final result    */

    /* Perform polynomial addition */

    for (idx = 0; idx <= L; idx++) 
        sum ^= ERR_IGNORE(wp_gMultVal (pIPWState, lambda[idx],
				       syn_bytes[n-idx], glog, gexp));

    return (sum);
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Find Roots to Error-Location Polynomial
 *
 * LongDescription:::
 * Find Roots to Error-Location Polynomial by "Chien Search".  
 * An error occurs in position i if and only if lambda(alpha^(-i)) = 0.
 * As a result we want to find the inverse roots of the error location
 * polynomial.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  nr_parity::      	number of ecc bytes per codeword
 *  lambda::         	error-locator polynomial
 *  glog::		log table
 *  gexp::		antilog table
 *
 * Output:::
 *  error_locs::     	array of error locations
 *
 * Returns:::
 *  n			number of errors
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *  Julie Craig
 *
 * Citations:::
 * Clark and Cain "Error-Correction Coding for Digital Communications"
 * pages 192-194
 *
 * EndDocumentation:::
 **************************************************************************/
static Int32 CDECL
wp_findRoots (
    IPWState	       *pIPWState,
    Int32               nr_parity,
    Int32              *error_locs,
    UInt8              *lambda,
    UInt32	       *glog,
    UInt8	       *gexp)
{
    Int32 rdx;        /* Loop (root) index */
    Int32 kdx;        /* Loop index */
    Int32 sum;        /* running sum */
    Int32 nerrors;
    UInt8 tmp[MAX_CODEWORD_SIZE+1];

    /* Clear the number of errors in current code word data */
    nerrors = 0;

    /* We need to find the roots of lambda in order to find the
     * actual error locations.  This is done by using a Chien search.
     * To do this we need a series of temporary variables, one for
     * each coefficeint in lambda.  Note that although the array is
     * declared to be of maximum size, no positions beyond the number
     * of coefficients of lambda are referenced and so they are not
     * initialized.
     * 1) Initialize the temporary variables to be the coeffients of 
     * lambda.
     * 2) for each possible root
     *    2.1) for each coefficient of lambda, the new value for the
     *         corresponding temporary vraible is found by multiplying 
     *	       the current contents with alpha^(coefficient number).
     *         This amount is exclusive or'd with the running total.
     *    2.2) If the running total is 0, then a root (and hence an
     *	       error location) has been found, record it.
     * Why does this work?  Well, the straightforward way to 
     * decide if a value is a root is to substitute each possible
     * finite field value into lambda and see if the result is 0.
     * This works but it does require lots of multiplication.
     * Another method is the one used here, that builds on the values
     * computed in the last iteration.  Lets say that 
     * lambda = x^4 + x^2 + x + 1.  When checking for alpha^1 the
     * brute force method says to substitute alpha^1 into the equation
     * yielding alpha^4 + alpha^2 + alpha + 1.  Note that this is
     * is exactly what we get if we multiply each coeffient of lambda
     * by alpha^(coefficient number).  Now we keep each of these
     * individual results in a temporary variable.  For the next
     * iteration, simply mutiply each of the temporary variables
     * with alpha^(coefficient number) again.  The brute force method
     * would yield (alpha^2)^4 + (alpha^2)^2 + alpha^2 + 1.
     * This is exactly what we get when we multiply the previous
     * iteration (alpha^4 + alpha^2 + alpha + 1) by alpha^coefficient.
     * (alpha^4 * alpha^4) + (alpha^2 * alpha^2) + (alpha * alpha) + 1.
     *
     * The previous version of the code is still below but "if 0'd" out.
     * This version did not keep the results of the previous iteration
     * and so had to recompute it each time; hence the multiply and
     * divide in the inner loop.
     */
    memcpy((char *)tmp, (char *)lambda, sizeof(*tmp)*(nr_parity+1));
    /* for each possible root */
    for (rdx = 1; rdx < MAX_CODEWORD_SIZE+1; rdx++) 
    {
  	sum = 0;
	/* for each coefficient of lambda */
	for (kdx = 0; kdx <= nr_parity; kdx++)
	{
	    tmp[kdx] = ERR_IGNORE(wp_gMultVal (pIPWState, gexp[kdx],
					       tmp[kdx], glog, gexp));
	    sum ^= tmp[kdx];
	}

	/* did we find a root? */
	if (sum == 0)
	{
	    /* we found the inverse of the error location so get the 
	     * actual error location.
	     */
            error_locs[nerrors] = (MAX_CODEWORD_SIZE-rdx);
            nerrors++; 
	}
    }

#if 0
    /* loop through all possible roots */
    for (rdx = 1; rdx < MAX_CODEWORD_SIZE+1; rdx++) 
    {
        sum = 0;

        /* evaluate Lambda at rdx */
        for (kdx = 0; kdx < nr_parity+1; kdx++)
            sum ^= ERR_IGNORE(wp_gMultVal (pIPWState,
					   gexp[(kdx*rdx)%MAX_CODEWORD_SIZE],
					   lambda[kdx], glog, gexp));

        /* Did an error occur? */
        if (sum == 0) 
	{ 
	    /* we found the inverse of the error location so get the 
	     * actual error location.
	     */
            error_locs[nerrors] = (MAX_CODEWORD_SIZE-rdx);
            nerrors++; 
        }
    }
#endif

    return nerrors;
}


/*

 $Log: berlekam.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:14  jlin
 Imported sources

 Revision 1.35  1998/01/14 00:10:41  ddavies
 berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.34  1996/03/21 21:26:35  ddavies
 ddavies on Thu Mar 21 13:25:13 PST 1996

 berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v

 Part of a general change in which the glog table becomes a UInt32 from UInt8,
 and all the polynomial coefficients (lambda, omega, psi, syn_bytes) become
 UInt8's.

 Revision 1.33  1996/03/15 19:15:20  jcraig
 jcraig on Fri Mar 15 11:08:56 PST 1996

 berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v

 add comments in wp_correctErrors.
 after finding the number (and location) of errors, take erasures into
 account when checking against the number of ecc bytes.

 Revision 1.32  1996/03/14 21:45:20  jcraig
 jcraig on Thu Mar 14 13:42:20 PST 1996

 berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v

 add comments.  bug fix to allow codeword to correct when number of
 erasures is equal to the number of ecc bytes.
 eliminate wp_ckSynBytes (actually check the syndrome bytes for zero
 while calculating them.  there's no need to look at them again.)
 minor performance improvement to wp_findErrors.

 Revision 1.31  1996/01/04 20:21:00  ejaquith
 berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.30  1995/12/13  04:04:42  ejaquith
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.29  1995/11/13  20:09:19  ejaquith
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.28  1995/11/09  01:18:51  jcraig
 * jcraig on Wed Nov 8 17:18:50 PST 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: wrap ERR_IGNORE around all private procs which do not return an
 * ErrCode
 *
 * Revision 1.27  1995/11/08  22:30:12  jcraig
 * jcraig on Wed Nov 8 14:29:38 PST 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * remove unused status from wp_computeModifiedOmega
 *
 * Revision 1.26  1995/11/07  21:29:10  jcraig
 * jcraig on Tue Nov 7 13:28:32 PST 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * ecc tables are now passed as parameters instead of being global vars
 *
 * Revision 1.25  1995/10/13  21:58:24  ejaquith
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.24  1995/10/07  02:57:44  ddavies
 * ddavies on Fri Oct 6 19:53:34 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * Change all occurances of errorStream to uses of errorProc and errorInfo
 * since the way stuff is printed has changed.  Also change all fprintf's to
 * i_printf's.  errorProc prints a single string.  errorInfo is private data
 * for errorProc.
 *
 * Revision 1.23  1995/10/05  01:28:50  jcraig
 * jcraig on Wed Oct 4 18:28:35 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * remove compiler warning about unused variable
 *
 * Revision 1.22  1995/10/04  22:13:28  jcraig
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * ifdef'd out a couple of print statements
 *
 * Revision 1.21  1995/09/30  00:18:23  ejaquith
 * ejaquith on Fri Sep 29 17:18:23 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: Converted private-ish glyph routines to be of "wp_" form and 100% private.
 *
 * Revision 1.20  1995/09/27  23:41:00  jcraig
 * jcraig on Wed Sep 27 16:40:09 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * new procedure comments
 *
 * Revision 1.19  1995/09/22  00:38:16  ejaquith
 * ejaquith on Thu Sep 21 17:38:15 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: Fixed up wp_ issues.
 *
 * Revision 1.18  1995/09/21  22:51:16  ejaquith
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.17  1995/09/07  00:50:33  jcraig
 * jcraig on Wed Sep 6 17:47:40 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * dgi prefix becomes w_
 *
 * Revision 1.16  1995/09/06  20:17:02  ejaquith
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.15  1995/09/01  02:11:14  jcraig
 * jcraig on Thu Aug 31 19:10:55 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.14  1995/08/31  21:10:51  ejaquith
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.13  1995/08/17  22:28:44  jcraig
 * jcraig on Thu Aug 17 15:28:37 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.12  1995/07/21  21:40:44  jcraig
 * jcraig on Fri Jul 21 14:38:55 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * don't print information about bytes corrected unless the enviroment
 * variable EDB_CORE_STATS is set (when the libraries were compiled
 * with GLY_INFO defined).
 *
 * Revision 1.11  1995/06/23  23:11:18  jcraig
 * jcraig on Fri Jun 23 16:10:43 PDT 1995
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * fixup file header comments
 *
 * Revision 1.10  1994/09/28  21:10:16  jcraig
 * jcraig on Wed Sep 28 14:09:28 PDT 1994
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * change edd prefix to dg.
 *
 * Revision 1.9  1994/09/23  00:07:02  jcraig
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * use GLY_INFO instead of PRINT_ERRORS around printing of error messages
 *
 * Revision 1.8  1994/05/02  20:25:14  jcraig
 * jcraig on Mon May 2 13:24:23 PDT 1994
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * get rid of warnings generated when compiling with gcc
 *
 * Revision 1.7  1994/03/10  00:35:28  jcraig
 * jcraig on Wed Mar 9 16:22:59 PST 1994
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * include glydebug.prv
 *
 * Revision 1.6  1994/01/27  00:48:25  jcraig
 * jcraig on Wed Jan 26 16:47:22 PST 1994
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * include eddcom.h and eddecc.h instead of libglyph.prv.
 * functions internal to the library but that are called from other
 * modules have eddi not edd prepended.
 *
 * Revision 1.5  1994/01/15  03:56:52  jcraig
 * jcraig on Fri Jan 14 19:56:31 PST 1994
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * removed include of libglyph.h - now only libglyph.prv is included.
 *
 * Revision 1.4  1993/12/02  23:19:22  ddavies
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.3  1993/11/10  02:22:23  jcraig
 * jcraig on Tue Nov 9 18:20:42 PST 1993
 *
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * remove prototypes from beginning of file and put into libglyph.prv.
 * procedures that are not private to this module have edd prepended.
 *
 * Revision 1.2  1993/11/06  03:09:44  ddavies
 * berlekam.c -> /product/ipcore/ipwrap/src/RCS/berlekam.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

