/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* 
 * bkfit.c
 *
 *	Subdivide a "page background" image into a bunch of rectangles
 *
 * Functions Defined:
 *	fitPageBackground()
 *	destroyRegionTree()
 *	renderRegionTree()
 */

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "morphg.h"
#include "bkfit.pub"
#include "bkfit.prv"
#include "shrcc.pub"
#include "imageref.pub"
#include "alplib.h"
#include "defines.pub"
#include "thrshold.h"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "curves.pub"
#include "strel.pub"
#include "prxform.pub"
#include "morphg.pub"
#include "morph.pub"
#include "rectops.pub"
#include "fill.pub"
#include "intops.pub"
#include "thrshold.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: bkfit.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Find a geometric description of a page background image
 *
 * LongDescription:::
 *   A page for which a background image has been computed by reduction
 *   and grayscale closing by makeBackgroundImage will be used to fit
 *   a model of the background which attempts to preserve tints and washes
 *   but reconstruct smoothed versions. I'm intending to model background
 *   regions as rectangles, with each rectangle containing either 1, 4,
 *   or 4 points corresponding to either a constant tint or a wash.
 *   It may be necessary to use 9 points rather than 4, in order to
 *   model non-linear washes, since most washes are non-linear. It would
 *   be better, though, if we could just use 4 points.
 *
 * PrivateDescription:::
 * Getting the rectangles is not so simple, due to the fact that a page
 * can have tints and washes which are only rectangular with respect to
 * the edge of the page. In order to accomodate these and other odd cases,
 * we're going to assume that we cannot always find the "right" 4 points
 * for a given tint region, but are going to use a subdivision algorithm
 * to generate the regions.
 * <BR>
 * The scheme is as follows: take the reduced, grayscale-closed version
 * of the image (probably the green channel, but it could be the luminance
 * channel), and run an edge detecting algorithm, such as the difference
 * between a grayscale dilate and a grayscale erode. Threshold these
 * edges, and separate them into horizontals and verticals using
 * binary openings, perhaps doing some additional closing to get
 * adjacent edges to merge into one. Then, find the connected components
 * for both horizontals and verticals, and sort by length (process the
 * largest first). Then, when going through either cc list, an edge
 * subdivides all of the rectangles it touches. So in the initial state,
 * there is one big rectangle, and an edge would divide it in two.
 * The next edge found would divide it further. It might make sense to
 * alternate horizontal and vertical edges, or it might make sense to
 * combine the connected component list and sort for size--the idea is
 * to get the smallest subdivision in terms of number of rectangles.
 * <BR>
 * Once the rectangles are computed, it might make sense to go through and
 * merge adjacent rectangles which have the same color on their edges.
 * We'll just have to see.
 * <BR>
 * Once that's done, go through and find the color on the corner of each
 * rectangle. Because the edges of the rectangles represent non-continuous
 * edges in closed grayscale, the colors should be sampled a bit in from
 * the edges, to avoid getting the wrong side of the edge. (And of course,
 * the other color planes need to be reduced and closed, though it might
 * make sense to do a closing just in the region of the corners to avoid
 * doing a close on the entire planes, though at least initially, I probably
 * won't do that.
 * <BR>
 * The result, then, is a map of the colors which should be fairly small.
 * The result should be a smoothly-reconstructable map of the tints and
 * washes in the background.
 * <BR>
 * Another note: the image should have already been segmented, and we
 * should wipe out the pictures in the reduced map to avoid subdividing
 * them needlessly.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to analyze
 *
 * Output:::
 *  ppRTree::        Region tree computed for image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_fitPageBackground (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RTREE             **ppRTree)
{
    PROLOGUE;
    RTREE	*rTree = NULL;
    RECTANGL	*hBoxes = NULL, *vBoxes = NULL, *rectP = NULL;
    EDGE	*edgeArray = NULL;
    Int32	 depth, /*tThresh, */i, rCount, channels;
/*    Bool	 ignore; */
    PIXR	*tmp1 = NULL, *tmp2 = NULL, *tmp3 = NULL, *horiz = NULL,
	        *vert = NULL;
    STREL       *pStrel3 = NULL, *pStrel10h = NULL, *pStrel10v = NULL,
	        *pStrel20h = NULL, *pStrel20v = NULL;

    /* Validate the PIXR */
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    
    /* Init the required strels - 3, 10h, 10v, 20h, 20v */
    ERR_PASS (w_initStrelByName (pIPWState, strel3));
    ERR_PASS (w_initStrelByName (pIPWState, strel10h));
    ERR_PASS (w_initStrelByName (pIPWState, strel10v));
    ERR_PASS (w_initStrelByName (pIPWState, strel20h));
    ERR_PASS (w_initStrelByName (pIPWState, strel20v));

    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));

    ERR_PASS (w_getStrel (pIPWState, strel3, &pStrel3));
    ERR_PASS (w_getStrel (pIPWState, strel10h, &pStrel10h));
    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel10v));
    ERR_PASS (w_getStrel (pIPWState, strel20h, &pStrel20h));
    ERR_PASS (w_getStrel (pIPWState, strel20v, &pStrel20v));

    /* Create an initial box for the image */
    rTree = (RTREE *) IPW_MALLOC (sizeof(RTREE), cAllocatorData);
    ERR_GEN(rTree == NULL, cErrNomem);
    *ppRTree = rTree;

    /* Fill it */
    rTree->extent.x = 0;
    rTree->extent.y = 0;
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prSrc, &(rTree->extent.w)));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prSrc, &(rTree->extent.h)));
    rTree->child[0] = NULL;
    rTree->child[1] = NULL;
    rTree->type = cRTREE_UNKNOWN;

    /* Find edges */
    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, prSrc, pStrel3, &tmp1));
    ERR_PASS_CLEANUP (w_erodeGray (pIPWState, prSrc, pStrel3, &tmp2));
    ERR_PASS_CLEANUP (w_subtractGray (pIPWState, tmp1, tmp2, &tmp3));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmp1));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmp2));
    ERR_PASS_CLEANUP (w_invert (pIPWState, tmp3, &tmp3));

    /* Threshold that */
/*    findTextThreshold (tmp3, &tThresh, &ignore); */
/*    tmp1 = lowpassPr (tmp3, tThresh); */
    /* Try just using a constant out of a hat */
    ERR_PASS_CLEANUP (w_threshold (pIPWState, tmp3, 0, 220, THRESH_NORM,
				   &tmp1));

    /* Get the horizontals and verticals */
    ERR_PASS_CLEANUP (w_open (pIPWState, tmp1, pStrel20h, &horiz));
    ERR_PASS_CLEANUP (w_close (pIPWState, horiz, pStrel10h, &horiz));

    ERR_PASS_CLEANUP (w_open (pIPWState, tmp1, pStrel20v, &vert));
    ERR_PASS_CLEANUP (w_close (pIPWState, vert, pStrel10v, &vert));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmp1));

    /* Find the connected components in both */
    ERR_PASS_CLEANUP (w_rabb (pIPWState, horiz, 20, 1, &hBoxes));
    ERR_PASS_CLEANUP (w_rabb (pIPWState, vert, 1, 20, &vBoxes));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &horiz));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &vert));

    /* Count them all up and make an array for sorting */
    rCount = 0;
    for (rectP = hBoxes; rectP; rectP = rectP->next)
	rCount++;
    for (rectP = vBoxes; rectP; rectP = rectP->next)
	rCount++;

    edgeArray = (EDGE *) IPW_CALLOC(rCount, sizeof (EDGE), cAllocatorData);
    ERR_GEN_CLEANUP (!edgeArray, cErrNomem);

    i = 0;
    for (rectP = hBoxes; rectP; rectP = rectP->next)
    {
	edgeArray[i].x = rectP->x;
	edgeArray[i].y = rectP->y + (rectP->h / 2);
	edgeArray[i].length = rectP->w;
	edgeArray[i++].direction = cEDGE_HORIZ;
    }

    for (rectP = vBoxes; rectP; rectP = rectP->next)
    {
	edgeArray[i].x = rectP->x + (rectP->w / 2);
	edgeArray[i].y = rectP->y;
	edgeArray[i].length = rectP->h;
	edgeArray[i++].direction = cEDGE_VERT;
    }

    /* Sort by lengths (descending) */
    qsort ((char *) edgeArray, rCount, sizeof (EDGE), compareEdgeLengths); 
    
    /* Do the subdivision */
    for (i = 0; i < rCount; i++)
    {
	ERR_PASS_CLEANUP (wp_subdivideRegions (pIPWState, &(edgeArray[i]), rTree));
    }

    /* Descend the tree and label the leaves with their colors */
    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prSrc, &channels));
    ERR_PASS_CLEANUP (wp_colorRegionTree (pIPWState, rTree, prSrc, channels));    

    /* Free up stuff */
    IPW_FREE (edgeArray, cAllocatorData);
    edgeArray = NULL;
    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &hBoxes));
    hBoxes = NULL;
    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &vBoxes));
    vBoxes = NULL;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (tmp1) w_destroyPixr (pIPWState, &tmp1);
    if (tmp2) w_destroyPixr (pIPWState, &tmp2);
    if (tmp3) w_destroyPixr (pIPWState, &tmp3);
    if (horiz) w_destroyPixr (pIPWState, &horiz);
    if (vert) w_destroyPixr (pIPWState, &vert);
    if (hBoxes) w_destroyRectList (pIPWState, &hBoxes);
    if (vBoxes) w_destroyRectList (pIPWState, &vBoxes);
    if (edgeArray) IPW_FREE (edgeArray, cAllocatorData);
     
    ERR_RETURN;
}

/*
 * compareEdgeLengths
 *
 *	Compare the lengths of EDGE structures -- qsort(3) callback
 *	Does it "backwards" so that we get a descending sort order
 */

static int compareEdgeLengths (const void *s1, const void *s2)
{
    const EDGE *e1 = (const EDGE *) s1;
    const EDGE *e2 = (const EDGE *) s2;

    /* Backwards, so that we sort descending */
    return e2->length - e1->length;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Subdivide the region tree along an edge
 *
 * LongDescription:::
 *   Given an edge, go down the region tree and subdivide any leaf
 *   rectangle which is touched by the edge. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pEdge::          The edge in question
 *  pTree::          The current region tree
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_subdivideRegions (
    IPWState           *pIPWState,
    EDGE               *pEdge,
    RTREE              *pTree)
{
    PROLOGUE;
    /* Does the edge intersect the extent of this box? */
    if (pEdge->direction == cEDGE_HORIZ)
    {
	if ((EDGEY0(*pEdge)   < Y0(pTree->extent)) || 
	    (EDGEY0(*pEdge)+1 > Y1(pTree->extent)) ||
	    (EDGEX0(*pEdge)   > X1(pTree->extent)) ||
	    (EDGEX1(*pEdge)   < X0(pTree->extent)))
	    return cSuccessful;
    }
    else
    {
	if ((EDGEX0(*pEdge)   < X0(pTree->extent)) ||
	    (EDGEX0(*pEdge)+1 > X1(pTree->extent)) ||
	    (EDGEY0(*pEdge)   > Y1(pTree->extent)) ||
	    (EDGEY1(*pEdge)   < Y0(pTree->extent)))
	    return cSuccessful;
    }

    /* Is it a non-leaf node? */
    if (pTree->child[0])
    {
	/* If so, call recursively */
	ERR_PASS_CLEANUP (wp_subdivideRegions (pIPWState, pEdge, pTree->child[0]));
	ERR_PASS_CLEANUP (wp_subdivideRegions (pIPWState, pEdge, pTree->child[1]));
    }
    else
    {
	/* This node is a leaf; we must split it */
	if (pEdge->direction == cEDGE_HORIZ)
	{
	    RTREE   *top, *bottom;

	    if (Y0(pTree->extent) == EDGEY0(*pEdge) ||
		Y1(pTree->extent) == EDGEY0(*pEdge))
		return cSuccessful;

	    top = (RTREE *) IPW_CALLOC (1, sizeof (RTREE), cAllocatorData);
	    ERR_GEN(!top, cErrNomem);
	    bottom = (RTREE *) IPW_CALLOC (1, sizeof (RTREE), cAllocatorData);
	    if (!bottom)
	    {
		IPW_FREE(top, cAllocatorData);
		ERR_GEN (1, cErrNomem);
	    }

	    top->extent.x = pTree->extent.x;
	    top->extent.w = pTree->extent.w;
	    top->extent.y = pTree->extent.y;	    
	    top->extent.h = pEdge->y - top->extent.y;
	    bottom->extent.x = pTree->extent.x;
	    bottom->extent.w = pTree->extent.w;
	    bottom->extent.y = pEdge->y;
	    bottom->extent.h = Y1(pTree->extent) - pEdge->y + 1;
	    pTree->child[0] = top;
	    pTree->child[1] = bottom;
	}
	else
	{
	    RTREE   *left, *right;

	    if (X0(pTree->extent) == EDGEX0(*pEdge) ||
		X1(pTree->extent) == EDGEX0(*pEdge))
		return cSuccessful;

	    left = (RTREE *) IPW_CALLOC (1, sizeof (RTREE), cAllocatorData);
	    ERR_GEN (!left, cErrNomem);
	    right = (RTREE *) IPW_CALLOC (1, sizeof (RTREE), cAllocatorData);
	    if (!right)
	    {
		IPW_FREE(left, cAllocatorData);
		ERR_GEN (1, cErrNomem);
	    }

	    left->extent.y = pTree->extent.y;
	    left->extent.h = pTree->extent.h;
	    left->extent.x = pTree->extent.x;	    
	    left->extent.w = pEdge->x - left->extent.x;
	    right->extent.y = pTree->extent.y;
	    right->extent.h = pTree->extent.h;
	    right->extent.x = pEdge->x;
	    right->extent.w = X1(pTree->extent) - pEdge->x + 1;
	    pTree->child[0] = left;
	    pTree->child[1] = right;
	}
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Descend the tree and record the color of the corner points of each region
 *
 * LongDescription:::
 *  Descend the tree and record the color of the corner points of each region
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pTree::          Region tree computed
 *  prSrc::          Source image
 *  channels::       Number of channels in image. This is a parameter so
 *                   that we don't compute it every time, since this
 *                   function is recursive.
 *
 * Output:::
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation
 * 
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_colorRegionTree (
    IPWState           *pIPWState,
    RTREE              *pTree,
    PIXR               *prSrc,
    Int32               channels)
{
    PROLOGUE;
    Int32	 i, nColors;
    Bool	 vWash, hWash;
    PIXR	*pC;
    Int32	 xSlop, ySlop, pixVal;
    ColorRecord	 corners[4];

    if (pTree->child[0])
    {
	/* Color the children */
	ERR_PASS (wp_colorRegionTree (pIPWState, pTree->child[0], 
				   prSrc, channels));
	ERR_PASS (wp_colorRegionTree (pIPWState, pTree->child[1], 
				   prSrc, channels));
    }
    else
    {
	vWash = cFalse;
	hWash = cFalse;
	xSlop = (pTree->extent.w * cInteriorSlopPercent) / 100;
	if (!xSlop)
	    xSlop = 1;
	ySlop = (pTree->extent.h * cInteriorSlopPercent) / 100;
	if (!ySlop)
	    ySlop = 1;


	/* This is a leaf; get the corner colors */
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS (w_pixrGetChannel (pIPWState, prSrc, i, &pC));
	
	    /* UL */
	    ERR_PASS (w_pixrGetPixel (pIPWState, pC,
				    X0(pTree->extent) + xSlop,
				    Y0(pTree->extent) + ySlop,
				    &pixVal));
	    corners[cUpperLeft][i] = (UInt8) pixVal;
	    
	    /* UR */
	    ERR_PASS (w_pixrGetPixel (pIPWState, pC,
				    X1(pTree->extent) - xSlop,
				    Y0(pTree->extent) + ySlop,
				    &pixVal));
	    corners[cUpperRight][i] = (UInt8) pixVal;
		
	    /* LL */
	    ERR_PASS (w_pixrGetPixel (pIPWState, pC,
				    X0(pTree->extent) + xSlop,
				    Y1(pTree->extent) - ySlop,
				    &pixVal));
	    corners[cLowerLeft][i] = (UInt8) pixVal;

	    /* LR */
	    ERR_PASS (w_pixrGetPixel (pIPWState, pC,
				    X1(pTree->extent) - xSlop,
				    Y1(pTree->extent) - ySlop,
				    &pixVal));
	    corners[cLowerRight][i] = (UInt8) pixVal;

	    if (ABS((Int32) corners[cLowerLeft][i] - (Int32) corners[cUpperLeft][i]) > cSameColor)
		vWash = cTrue;
	    if (ABS((Int32) corners[cUpperRight][i] - (Int32) corners[cUpperLeft][i]) > cSameColor)
		hWash = cTrue;
		
	}

	if (hWash && vWash)
	{
	    pTree->type = cRTREE_WASH_4;
	    nColors = 4;
	}
	else if (hWash)
	{
	    pTree->type = cRTREE_WASH_H;
	    nColors = 2;
	}
	else if (vWash)
	{
	    pTree->type = cRTREE_WASH_V;
	    nColors = 2;
	}
	else
	{
	    pTree->type = cRTREE_TINT;
	    nColors = 1;
	}

	pTree->colors = (ColorRecord *) 
	    IPW_CALLOC (nColors, sizeof(ColorRecord), cAllocatorData);
	ERR_GEN (pTree->colors == NULL, cErrNomem);

	for (i = 0; i < channels; i++)
	{
	    switch (pTree->type)
	    {
	    case cRTREE_WASH_4:
		pTree->colors[cUpperLeft][i] = corners[cUpperLeft][i];
		pTree->colors[cUpperRight][i] = corners[cUpperRight][i];
		pTree->colors[cLowerLeft][i] = corners[cLowerLeft][i];
		pTree->colors[cLowerRight][i] = corners[cLowerRight][i];
		break;
	    case cRTREE_WASH_H:
		pTree->colors[0][i] = corners[cUpperLeft][i];
		pTree->colors[1][i] = corners[cUpperRight][i];
		break;
	    case cRTREE_WASH_V:
		pTree->colors[0][i] = corners[cUpperLeft][i];
		pTree->colors[1][i] = corners[cLowerLeft][i];
		break;
	    case cRTREE_TINT:
		pTree->colors[0][i] = corners[cUpperLeft][i];
		break;
	    default:
		ERR_GEN (1, cErrTypeNotSupported);
	    }
	}
#if 0
	{
	    Int32 x, y, w, h;
	    x = pTree->extent.x;
	    y = pTree->extent.y;
	    w = pTree->extent.w;
	    h = pTree->extent.h;
fprintf(stderr, "(%ld,%ld), w=%ld, h=%ld, ulc=%d\n", x, y, w, h, corners[0][0]);

#if 0
	    pixr_rop (pIPWState, testPixr, x, y, w, 1, PIX_SET, NULL, 0, 0);
	    pixr_rop (pIPWState, testPixr, x, y, 1, h, PIX_SET, NULL, 0, 0);
	    pixr_rop (pIPWState, testPixr, x + w - 1, y, 1, h, PIX_SET, NULL, 0, 0);
	    pixr_rop (pIPWState, testPixr, x, y + h - 1 , w, 1, PIX_SET, NULL, 0, 0);
#endif
	}
    
#endif
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Render a region tree (background description) into a PIXR
 *
 * LongDescription:::
 *	Render a region tree into a PIXR, scaling appropriately, i.e.
 *	attempt to reconstruct the page background. This is going to 
 *	use a lot of floating point for now; it could be made a lot
 *	more efficient..
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  rTree::          Region tree to render
 *
 * Output:::
 *  prDest::         Image into which to render. NOT created by this
 *                   routine, since the scale factor is determined by
 *                   the size of the image.
 * 
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page rendering
 *
 * Keywords:::
 *  page rendering
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_renderRegionTree (
    IPWState           *pIPWState,
    RTREE              *rTree,
    PIXR               *prDest)
{
    PROLOGUE;
    Float64	scale;
    Int32       width;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);
    ERR_GEN (!rTree, cErrInvalidParm);

    /* Compute scale factor (we don't do anamorphic scaling) */
    ERR_PASS (w_pixrGetWidth (pIPWState, prDest, &width));
    scale = (Float64) width / rTree->extent.w;

    /* Descend the tree */
    ERR_PASS (wp_doRenderTree (pIPWState, rTree, scale, prDest));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Internal routine of w_renderRegionTree. 
 *
 * LongDescription:::
 *	Internal routine of renderRegionTree. The actual recursive thing,
 *	without error checks or computation of total scale factors
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pTree::          Tree to render
 *  scale::          Scale factor to get from tree to image space
 *
 * Output:::
 *  prDest::          Image into which to render
 *  
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page rendering
 *
 * Keywords:::
 *  page rendering
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_doRenderTree (
    IPWState           *pIPWState,
    RTREE              *pTree,
    Float64             scale,
    PIXR               *prDest)
{
    PROLOGUE;

    if (pTree->child[0])
    {
	ERR_PASS (wp_doRenderTree (pIPWState, pTree->child[0], scale, prDest));
	ERR_PASS (wp_doRenderTree (pIPWState, pTree->child[1], scale, prDest));
    }
    else
    {
	/* This is a leaf; render it */
	switch (pTree->type)
	{
	    case cRTREE_TINT:
		ERR_PASS (wp_renderTint (
		    pIPWState, 
		    (Int32) (pTree->extent.x * scale),
		    (Int32) (pTree->extent.y * scale),
		    (Int32) ((pTree->extent.w + 1) * scale - 1),
		    (Int32) ((pTree->extent.h + 1) * scale - 1),
		    pTree->colors[0], 
		    prDest));
		break;
		
	    case cRTREE_WASH_H:
		ERR_PASS (wp_renderWashH (
		    pIPWState,
		    (Int32) (pTree->extent.x * scale),
		    (Int32) (pTree->extent.y * scale),
		    (Int32) ((pTree->extent.w + 1) * scale - 1),
		    (Int32) ((pTree->extent.h + 1) * scale - 1),
		    pTree->colors[0],
		    pTree->colors[1], 
		    prDest));

		break;

	    case cRTREE_WASH_V:
		ERR_PASS (wp_renderWashV (
		    pIPWState,
		    (Int32) (pTree->extent.x * scale),
		    (Int32) (pTree->extent.y * scale),
		    (Int32) ((pTree->extent.w + 1) * scale - 1),
		    (Int32) ((pTree->extent.h + 1) * scale - 1),
		    pTree->colors[0],
		    pTree->colors[1], 
		    prDest));
		break;

	    case cRTREE_WASH_4:
		 ERR_PASS (wp_renderWash4 (
		     pIPWState,
		     (Int32) (pTree->extent.x * scale),
		     (Int32) (pTree->extent.y * scale),
		     (Int32) ((pTree->extent.w + 1) * scale - 1),
		     (Int32) ((pTree->extent.h + 1) * scale - 1),
		     pTree->colors[0],
		     pTree->colors[1],
		     pTree->colors[2],
		     pTree->colors[3],
		     prDest));
		break;

	    default:
		/* Ugh */
		break;

	}
    }

    return cSuccessful;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Render a flat tint
 *
 * LongDescription:::
 *  Render a flat tint
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  x::              Bounding box x0
 *  y::              Bounding box y0
 *  w::              Bounding box width
 *  h::              Bounding box height
 *  color::          Color to render
 *
 * Output:::
 *  prDest::          Image into which to render
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page rendering
 *
 * Keywords:::
 *  page rendering
 *
 * Level:::
 *  deterministic
 *  heuristic
 *  directive
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_renderTint (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    ColorRecord         color,
    PIXR               *prDest)
{
    PROLOGUE;
    RECTANGL	r;

    r.x = x;
    r.y = y; 
    r.w = w;
    r.h = h;

    ERR_PASS (w_paint (pIPWState, prDest, &r, color));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Render a horizontal wash
 *
 * LongDescription:::
 *  Render a horizontal wash. 
 *
 * PrivateDescription:::
 *   What we'll do is fill the first line and then copy it. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  x::              Bounding box x0
 *  y::              Bounding box y0
 *  w::              Bounding box width
 *  h::              Bounding box height
 *  leftColor::      Color on left edge
 *  rightColor::     Color on right edge
 *
 * Output:::
 *  prDest::         Image into which to render wash
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page rendering
 *
 * Keywords:::
 *  page rendering
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_renderWashH (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    ColorRecord         leftColor,
    ColorRecord         rightColor,
    PIXR               *prDest)
{
    PROLOGUE;
    Int32    i, channels, channel, color, depth, bpl;
    UInt8 *p;
    PIXR  *prChannel;

    ERR_PASS (w_pixrGetChannelCount (pIPWState, prDest, &channels));
    ERR_PASS (w_pixrGetDepth (pIPWState, prDest, &depth));
    ERR_PASS (w_pixrGetBpl (pIPWState, prDest, &bpl));

    for (channel = 0; channel < channels; channel++)
    {
	prChannel = NULL;
	ERR_PASS (w_pixrGetChannel (pIPWState, prDest, channel, &prChannel));
	ERR_PASS (w_pixrGetImage (pIPWState, prChannel, &p));
	p += y * bpl;

	for (i = 0; i < w; i++)
	{
	    /* A floating point interpolation; good thing it's only
	       one line.. */
	    color = (Int32) 
		(((Float64) leftColor[channel] * (Float64) (w - (i + 1)) +
		  (Float64) rightColor[channel] * (Float64) i) / 
		 (Float64) (w - 1));
	    if (depth == 8)
		UCHAR_ACCESS(p + x + i) = (UInt8) color;
	    else
		SET_IMAGE_BITS (p, x + i, color, depth);
	}
    }

    /* Copy that line to the rest of the lines */
    for (i = 1; i < h; i++)
    {
	ERR_PASS (w_rasterop (pIPWState, x, y, prDest, PIX_SRC,
			      x, y + i, w, 1, prDest));
    }

    return cSuccessful;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Render a vertical wash.
 *
 * LongDescription:::
 *   Render a vertical wash.
 *
 * PrivateDescription:::
 *   What we do is fill the first column and then copy it. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  x::              Bounding box x0
 *  y::              Bounding box y0
 *  w::              Bounding box width
 *  h::              Bounding box height
 *  topColor::       Color on top edge
 *  bottomColor::    Color on bottom edge
 *
 * Output:::
 *  prDest::         Image into which to render wash
 *  
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page rendering
 *
 * Keywords:::
 *  page rendering
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_renderWashV (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    ColorRecord         topColor,
    ColorRecord         bottomColor,
    PIXR               *prDest)
{
    PROLOGUE;
    Int32    i, channels, channel, color, depth, bpl;
    UInt8  *p;
    PIXR   *prChannel;

    ERR_PASS (w_pixrGetChannelCount (pIPWState, prDest, &channels));
    ERR_PASS (w_pixrGetDepth (pIPWState, prDest, &depth)); 
    ERR_PASS (w_pixrGetBpl (pIPWState, prDest, &bpl));

    for (channel = 0; channel < channels; channel++)
    {
	prChannel = NULL;
	ERR_PASS (w_pixrGetChannel (pIPWState, prDest, channel, &prChannel));
	ERR_PASS (w_pixrGetImage (pIPWState, prChannel, &p));
	p += y * bpl;

	for (i = 0; i < h; i++)
	{
	    /* A floating point interpolation; good thing it's only
	       one column.. */
	    color = (Int32) (((Float64) topColor[channel] * 
			      (Float64)(h - (i + 1)) +
			      ((Float64) bottomColor[channel] * 
			       (Float64) i)) / (Float64) (h - 1));
	    if (depth == 8)
		UCHAR_ACCESS(p + x) = (Int8) color;
	    else
		SET_IMAGE_BITS (p, x, color, depth);

	    p += bpl;
	}
    }

    /* Copy that column over */
    for (i = 1; i < w; i++)
    {
	ERR_PASS (w_rasterop (pIPWState, x, y, prDest, PIX_SRC,
			      x + i, y, 1, h, prDest));
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Render a 4-point wash, verrry slowly.
 *
 * LongDescription:::
 *   Render a 4-point wash, verrry slowly.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  x::              Bounding box x0
 *  y::              Bounding box y0
 *  w::              Bounding box width
 *  h::              Bounding box height
 *  ul::             Upper left color
 *  ur::             Upper right color
 *  ll::             Lower left color
 *  lr::             Lower right color
 *
 * Output:::
 *  prDest::          
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page rendering
 *
 * Keywords:::
 *  page rendering
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_renderWash4 (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    ColorRecord         ul,
    ColorRecord         ur,
    ColorRecord         ll,
    ColorRecord         lr,
    PIXR               *prDest)
{
    PROLOGUE;
    Int32    i = 0, j, channels, channel, color, l, r, depth, bpl;
    UInt8   *p;
    PIXR    *prChannel;

    ERR_PASS (w_pixrGetChannelCount (pIPWState, prDest, &channels));
    ERR_PASS (w_pixrGetDepth (pIPWState, prDest, &depth));
    ERR_PASS (w_pixrGetBpl (pIPWState, prDest, &bpl));

    for (j = 0; j < h; j++)
    {
	for (channel = 0; channel < channels; channel++)
	{
	    l = (Int32) (((Float64) ul[channel] * (h - (j + 1)) +
		 (Float64) ll[channel] * j) / (Float64) (h - 1));
	    r = (Int32) (((Float64) ur[channel] * (h - (j + 1)) +
		 (Float64) lr[channel] * j) / (Float64) (h - 1));

	    prChannel = NULL;
	    ERR_PASS (w_pixrGetChannel (pIPWState, prDest, channel, &prChannel));
	    ERR_PASS (w_pixrGetImage (pIPWState, prChannel, &p));
	    p += (y + j) * bpl;

	    for (i = 0; i < w; i++)
	    {
		color = (Int32) (((Float64) l * (Float64) (w - (i + 1)) +
			 (Float64) r * (Float64) i) / (Float64) (w - 1));
		if (depth == 8)
		    UCHAR_ACCESS(p + x + i) = (UInt8) color;
		else
		    SET_IMAGE_BITS (p, x + i, color, depth);
	    }
	}
    }

    NORMAL_RETURN;
}
	    
/*
 $Log: bkfit.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.22  1997/06/26 20:00:25  ddavies
 bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.21  1996/01/04 20:21:04  ejaquith
 bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.20  1995/12/13  04:04:51  ejaquith
 * bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.19  1995/11/13  20:09:27  ejaquith
 * bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.18  1995/10/13  21:58:32  ejaquith
 * bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/10/12  17:42:37  macleod
 * macleod on Thu Oct 12 10:41:48 PDT 1995
 *
 * bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v
 *
 * added explicit casts (float->int, long->short, long->char, etc) and
 * made floating constants the right size to shut up the NT compiler
 *
 * Revision 1.16  1995/10/09  19:04:19  macleod
 * macleod on Mon Oct 9 12:03:46 PDT 1995
 *
 * bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v
 *
 * Attempt to eliminate some NT compiler whinings
 *
 * Revision 1.15  1995/10/03  01:59:29  macleod
 * bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.14  1995/09/21  22:51:23  ejaquith
 * bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/11  21:54:32  macleod
 * bkfit.c -> /product/ipcore/ipwrap/src/RCS/bkfit.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 * 
*/

