/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  boxes1.c:  includes subroutines
 *
 *      Creation/annihilation
 *          BOXES    *createBoxes():  creates struct and Rect ptr array
 *          BOXES    *destroyBoxes():  frees Boxes, rectlist and Rect array
 *          BOXES    *copyBoxes():  copies a Boxes, with copies of the Rects
 *          Int32     addRectToBoxes(): adds rect & updates count
 *
 *      Extractors
 *          Int32     numBoxes():  returns num of RECTANGLs in the BOXES struct.
 *
 */



#include "ipwcore.prv"
#include "lists.h"
#include "geomadt.h"
#include "pixr.h"

/* Prototypes */
#include "rectops.pub"
#include "boxes.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: boxes1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

/*----------------------------------------------------------------------*
 
           Comments on the rectlist (RL) and boxeslist (BL)
	   data structures, and the transformations between
	   them and Rects and Boxes.

   The rectlist is a convenient data structure, typically used
   within a procedure for computation, where the number of Rects is not known or
   is changing.  It is simply a list of Rects.  A Boxes can
   contain both an array of Rects and a Rectlist, although in practice
   only one is typically used for any Boxes.  Procedures that
   convert between Boxes of the two types are provided; these
   create new Boxes with new Rects, leaving the input version unchanged.

   A boxeslist is simply a list of Boxes.  Like the rectlist, it
   is a convenient data structure for use where the number of Boxes,
   or of Rects within a rectlist within a Boxes, is changing or unknown.

   The boxeslist, like the Boxar, is useful for computing two-dimensional
   arrangements of Rects.  For example, each Boxes in a boxeslist
   can represent a line of text, being composed of rectangles
   that represent the characters or words within the line.

   Procedures are provided to convert a RL into a Boxes, and to convert
   a BL into a single (flattened) Boxes.
   Because the RL is considered to be a temporary structure used
   within a procedure, we provide versions of these conversion
   routines that destroy the input RL.  Likewise, a version of the
   BLtoBoxes() is provided that destroys the input BL.

   Additionally, for transformations in the other direction (from
   Boxes to RL), a version is provided that destroys the input Boxes.
   This is really dangerous!

   In boxar.c, conversions are given between BL and Boxars.
   For converting from BL to Boxar, as from BL to Boxes, two
   versions are given, one that preserves the BL and the other
   that destroys it.  For the other direction, from Boxar to BL,
   two versions are again provided, but the version that destroys
   a Boxar is very dangerous.



	    Comments on the the arrays of Rects in Boxes and
	    the arrays of Boxes in Boxars.

   Boxes and Boxars typically have an array of Rects and Boxes, respectively.
   These are constructed, one Rect or Boxes at a time, and the function
   addRectToBoxes() should always be used, because it checks
   if there is room in the array and also increments the n field, which
   contains the actual number of structures contained in the array.


 *----------------------------------------------------------------------*/



/**************************************************************
 ***********        Creation/Annihilation           ***********
 *************************************************************/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a BOXES structure.
 *
 * LongDescription:::
 *  Create a BOXES structure, which is an array of pointers to rectangles.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  numrects::       Number of rectangles in this box array
 *
 * Output:::
 *  ppBoxes::        Box list created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createBoxes (
    IPWState           *pIPWState,
    Int32               numrects,
    BOXES             **ppBoxes)
{
    PROLOGUE;
BOXES        *boxes = NULL;

    ERR_GEN (numrects < 0, cErrInvalidParm);

	/* make the boxes structure */
    boxes = (BOXES *) IPW_CALLOC(1, sizeof(BOXES), cAllocatorData);
    ERR_GEN_CLEANUP (boxes == NULL, cErrNomem);

    boxes->n = 0;
    boxes->nalloc = numrects;
    INIT_LINKS(boxes)

	/* make the array of pointers to rects */
    if (numrects) 
    {
	boxes->rect = (RECTANGL **) IPW_CALLOC(numrects, 
					       sizeof(RECTANGL *),
					       cAllocatorData);
	ERR_GEN_CLEANUP (boxes->rect  == NULL, cErrNomem);
    }

    *ppBoxes = boxes;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (boxes) IPW_FREE (boxes->rect, cAllocatorData);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroy a BOXES structure and associated RECTANGLs
 *
 * LongDescription:::
 *  Frees BOXES and array of associated RECTANGLS;
 *  always returns a NULL pointer.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  ppBoxes::        BOXES object, which is set to NULL on successful exit
 *
 * Output:::
 *  
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyBoxes (
    IPWState           *pIPWState,
    BOXES             **ppBoxes)
{
    PROLOGUE;
Int32         i;
RECTANGL     *rect, *nextRect;
BOXES        *boxes;

    ERR_GEN (!ppBoxes, cErrNullPtr);

    /* It is not considered an error to free a null box list
     * This is an historical artifact, which if exploited is IMHO a bit
     * sleazy. 
     */
    boxes = *ppBoxes;
    if (boxes)
    {
	/* free rect array */

	for (i = 0; i < boxes->n; i++)
	{
	    if (boxes->rect[i])
	    {
		IPW_FREE(boxes->rect[i], cAllocatorData);
	    }
	}
	
	IPW_FREE(boxes->rect, cAllocatorData);

	/* free rect list */
	if (boxes->rectlist)
	{
	    for (rect = boxes->rectlist; rect != NULL; rect = nextRect)
	    {
		nextRect = rect->next;
		IPW_FREE (rect, cAllocatorData);
	    }
	}

	IPW_FREE(boxes, cAllocatorData);
    }

    *ppBoxes = NULL;
    NORMAL_RETURN;
}
 
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Duplicate a BOXES structure, including RECTANGLs
 *
 * LongDescription:::
 *  Copies all RECTANGLs in the original BOXES, in either or both of the
 *  array and rectlist
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  boxes::          BOXES object
 *
 * Output:::
 *  ppNewBoxes::     Duplicate BOXES object
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copyBoxes (
    IPWState           *pIPWState,
    BOXES              *boxes,
    BOXES             **ppNewBoxes)
{
    PROLOGUE;
Int32         i;
RECTANGL     *first, *rect, *copyrect = NULL;
BOXES        *copyboxes = NULL;

    ERR_GEN (!boxes, cErrNullPtr);

    copyboxes = NULL;
    copyrect = NULL;

	/* allocate */
    ERR_PASS_CLEANUP( w_createBoxes(pIPWState, boxes->n, &copyboxes));

	/* copy each rect and insert */
    for (i = 0; i < boxes->n; i++) 
    {
	ERR_PASS_CLEANUP( w_copyRect(pIPWState, boxes->rect[i], &copyrect));
	ERR_PASS_CLEANUP( w_addRectToBoxes(pIPWState, copyboxes, copyrect));
    }

	/* copy the rectlist */
    if ((first = boxes->rectlist) != NULL)
    {
	/* copy the first rect */
	ERR_PASS_CLEANUP( w_copyRect(pIPWState, first, &copyrect));
	copyboxes->rectlist = copyrect;
	    /* now that copyboxes->rectlist is initialized,
	     * copy the rest of the boxes. */
	for (rect = first->next; rect != NULL; rect = rect->next)
	{
	    ERR_PASS_CLEANUP( w_copyRect(pIPWState, rect, &copyrect));
	    copyrect->next = copyboxes->rectlist;
	    copyrect->next->prev = copyrect;
	    copyboxes->rectlist = copyrect;
	}
    }

    *ppNewBoxes = copyboxes;

NORMAL_RETURN;

ERR_CLEANUP:
    if (copyboxes)
	IPW_FREE (copyboxes, cAllocatorData);

    ERR_RETURN;
	
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a RECTANGL to a BOXES array
 *
 * LongDescription:::
 *  if there is room in the array, the Rect is inserted at the next 
 *  available position and the counter is incremented. If there is
 *  no room, an error code is returned.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  boxes::          BOXES object in which to insert RECTANGL
 *  rect::           RECTANGL to insert
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addRectToBoxes (
    IPWState           *pIPWState,
    BOXES              *boxes,
    RECTANGL           *rect)
{
    PROLOGUE;

    ERR_GEN (!boxes, cErrInvalidParm);
    ERR_GEN (!rect, cErrInvalidParm);

    ERR_GEN (boxes->nalloc == 0, cErrInvalidParm);
    ERR_GEN (boxes->n >= boxes->nalloc, cErrArrayFull);

    boxes->rect[boxes->n] = rect;
    boxes->n += 1;

    NORMAL_RETURN;
}

/**************************************************************
 ***********               Extractors               ***********
 *************************************************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Count the number of RECTANGLs in a BOXES object.
 *
 * LongDescription:::
 *  Takes a ptr to BOXES and returns the number of Rects.
 *  All valid boxes structs have at least one rect.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  boxes::          BOXES object
 *
 * Output:::
 *  pBoxCount::      # of RECTANGLs (not BOXES) in the BOXES object
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_numBoxes (
    IPWState           *pIPWState,
    BOXES              *boxes,
    Int32              *pBoxCount)
{
    PROLOGUE;

    ERR_GEN (!boxes, cErrInvalidParm);

    *pBoxCount = boxes->n;
    NORMAL_RETURN;
}

/*

 $Log: boxes1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.18  1996/01/04 20:21:09  ejaquith
 boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.17  1995/12/13  04:04:58  ejaquith
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/13  20:09:32  ejaquith
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.15  1995/10/13  21:58:38  ejaquith
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.14  1995/10/03  21:12:48  macleod
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.13  1995/09/21  22:51:29  ejaquith
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/11  21:55:18  macleod
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.11  1995/09/06  20:17:16  ejaquith
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  20:43:30  macleod
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * Added back #include files
 *
 * Revision 1.9  1995/08/31  21:11:05  ejaquith
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/29  21:53:17  ddavies
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.7  1995/08/25  01:35:03  ejaquith
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.6  1995/08/17  21:53:44  macleod
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.5  1994/10/06  20:53:17  lperry
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.4  1993/12/17  22:34:53  ddavies
 * ddavies on Fri Dec 17 14:33:29 PST 1993
 *
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * Replace the list processing macros with plain old code when processing
 * linked lists of RECTANGLs.  These lists are no longer circular, they have
 * NULL pointers on the ends.  This is generally easier (and faster!) to deal with.
 *
 * Revision 1.3  1993/12/03  01:13:17  ddavies
 * ddavies on Thu Dec 2 17:13:16 PST 1993
 *
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:09:58  ddavies
 * boxes1.c -> /product/ipcore/ipwrap/src/RCS/boxes1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

