/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  boxes2.c:  includes subroutines
 *
 *
 *      Spatial transforms
 *          BOXES  *xformBoxes()       translation and scaling
 *
 */



#include "ipwcore.prv"
#include "lists.h"
#include "geomadt.h"
#include "pixr.h"

/* Prototypes */
#include "rectops.pub"
#include "boxes.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: boxes2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");
#define  NULL_RECT_ERROR_MESSAGE      1
#define  ZERO_RECT_ERROR_MESSAGE      0



/*-------------------------------------------------------------------*
 *                         Spatial Transforms                        *
 *-------------------------------------------------------------------*/
/*
 *  xformBoxes(): This first shifts the origin and then scales each
 *                rectangle in the input Boxes.  This procedure is
 *                used when computing the corresponding regions
 *                on a shifted and/or scaled image.  There is
 *                no clipping of the shifted and scaled rects
 *
 *                Returns NULL on error; an xformed Boxes if OK.
 *
 *                N.B.  The shifts are integers; the scale factor is a float.
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Shift and scale the rectangles in a BOXES object
 *
 * LongDescription:::
 *   This first shifts the origin and then scales each  rectangle in the 
 *   input Boxes.  This procedure is used when computing the corresponding
 *   regions on a shifted and/or scaled image.  There is no clipping of the 
 *   shifted and scaled rects
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::          State structure
 *  boxes::              BOXES object to transform
 *  shiftX::             X-shift
 *  shiftY::             Y-shift
 *  scale::              Scale factor
 *
 * Output:::
 *  pTransformedBoxes::  A transformed BOXES array
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_xformBoxes (
    IPWState           *pIPWState,
    BOXES              *boxes,
    Int32               shiftX,
    Int32               shiftY,
    Float32             scale,
    BOXES             **pTransformedBoxes)
{
    PROLOGUE;
Int32        i;
BOXES       *xfboxes;
RECTANGL    *xfrect;


    ERR_GEN (!boxes, cErrInvalidParm);
    *pTransformedBoxes = xfboxes = NULL;

	/* make xformed boxes and ptrs to rects */
    ERR_PASS( w_createBoxes(pIPWState, boxes->n, &xfboxes));

	/* shift the origin and then scale */
    for (i = 0; i < boxes->n; i++)
    {
	xfrect = NULL;
	ERR_PASS_CLEANUP( w_xformRect(pIPWState, boxes->rect[i], 
				      shiftX, shiftY,
				      scale, scale,
				      &xfrect));
	ERR_PASS_CLEANUP( w_addRectToBoxes(pIPWState, xfboxes, xfrect));
    }

    *pTransformedBoxes = xfboxes;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (xfboxes)
	w_destroyBoxes (pIPWState, &xfboxes);

    ERR_RETURN;
}

/*

 $Log: boxes2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.19  1996/01/27 22:37:19  ejaquith
 ejaquith on Sat Jan 27 14:37:06 PST 1996

 boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v

 Updated call to w_xformRect()

 * Revision 1.18  1996/01/04  20:21:11  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.17  1995/12/13  04:05:03  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/13  20:09:36  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.15  1995/11/04  01:35:48  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.14  1995/10/13  21:58:41  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.13  1995/10/03  21:12:50  macleod
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.12  1995/09/21  22:51:33  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.11  1995/09/11  21:55:26  macleod
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.10  1995/09/06  20:17:19  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  20:43:37  macleod
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * Added back #include files
 *
 * Revision 1.8  1995/08/31  21:11:08  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/29  21:53:22  ddavies
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.6  1995/08/25  01:35:06  ejaquith
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * Fixed warnings
 *
 * Revision 1.5  1995/08/17  21:53:48  macleod
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.4  1994/10/06  20:53:22  lperry
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.3  1993/12/03  01:13:22  ddavies
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:10:05  ddavies
 * boxes2.c -> /product/ipcore/ipwrap/src/RCS/boxes2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

