/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
** File:        cchisto.c
**
** Description: Histograms of number of connected components as a function
**              of the threshold... Cute algorithms
**
** Author: Luc Vincent
**	   Hacked into IPWRAP by P. MacLeod
*/


#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "histo.h"
#include "defines.pub"
#include "imageref.pub"

/* prototypes */
#include "props.pub"
#include "hutils.pub"
#include "cchisto.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: cchisto.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


/*
**  Exported Functions:
**************************
*
*  horizRunsHistogram()  Histogram of connected components per line
*                               or per column of image, as a function of the
*                               threshold. Threshold can be either
*                               0 = below, 1 = above
*
*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute histogram of number of horiz. black runs as function of threshold
 * LongDescription:::
 *  Computes the histogram of how many black runs
 *  there are if we were to take a threshold of the image at
 *  a given level (above this level or below this level).
 
 * PrivateDescription:::
 *  Suppose for example that threshold_above == cFalse. This means
 *  that every threshold considered is taken between 0 (inclusive) and
 *  the threshold value v. If you think about it a few minutes, you soon
 *  realize that there are only two cases that can make the number of
 *  black runs vary from one threshold to the next: the minima and the
 *  maxima in each line. By minimum, I mean a run of pixels at the same
 *  value such that the pixel before that and the pixel after that are
 *  strictly higher.
 *  <BR>
 *     - suppose that in the considered line, we have a minimum at
 *     value v. If we threshold between 0 and v, we obtain a run
 *     for this particular minimum. However, if we threshold between
 *     0 and w<v, the pixels in this minimum have a too high value (v),
 *     so we get nothing. In clear, this means that when we lower the
 *     threshold level from v to v-1, the number of runs in the resulting
 *     binary image decreases by 1.
 *  <BR>
 *     - suppose now that in the considered line, we have a maximum
 *     at value v. Thresholding between 0 and v, we obtain a run for
 *     this maximum as well as the neighboring pixels. But if we lower the
 *     threshold by 1, the pixels in the maximum are non longer in the
 *     resulting threshold. However, the neighboring pixels on the left and
 *     the right are still on in the resulting threshold. In clear, this
 *     means that when we lower the threshold level from v to v-1, the
 *     number of runs in the resulting binary image increases by 1.
 *  <BR>
 *  The principle of the algorithm is therefore to directly determine the
 *  *derivative* of the black runs histogram desired. Each line of the
 *  image is scanned, and everytime a minimum is discovered, the bin of
 *  the derivative of the histogram corresponding to the value of this
 *  mimimum is decremented by 1. Similarly, everytime a maximum is
 *  discovered, the bin of the derivative of the histogram corresponding
 *  to the value of this mimimum is incremented by 1. At the end of the
 *  routine, the histogram can be integrated if desired.
 *  <BR>
 *  The routine is extremely fast, because only the minima and maxima
 *  cause it to do something, and the operation done in each case is very
 *  very simple (++ or -- !). Undoubtedly, the implementation below can
 *  still be improved a lot, or perhaps assembly-coded.
 *
 * Input:::
 *  pIPWState::       State structure
 *  prSrc::           Source image
 *  threshold_above:: if !0, consider pixels between threshold and highest
 *                    pixel value, otherwise consider pixels between 0
 *                    and threshold
 *  differentiate::   if cTrue, don't integrate the resulting histogram
 *
 * Output:::
 *  ppHisto::         Resulting histogram
 *
 * Returns:::
 *
 * Notes:::
 *  Idea of computing such histograms obtained from a paper
 *  by O'Gorman's. Implementation is completely different and
 *  many times more efficient here!
 *  <BR>
 *  Note that when differentiation is requested, the value
 *  of histogram at MINPIXU for threshold_above == cTrue
 *  (or MAXPIXU for threshold_above == cFalse) is not
 *  a differenciation value, but in fact the number of c.c.'s
 *  at this level. Make sure to change this to 0 if using the
 *  differentiated histogram!
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, run length, threshold finding
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_horizRunsHistogram (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Bool                threshold_above,
    Bool                differentiate,
    Histo             **ppHisto)
{
    PROLOGUE;
    register Int32  *hdata;
    Int32            width, height, depth, vdepth;
    UInt32	     this;
    UInt8           *pixelsS, byte = 0;
    UInt8           *lineS;
    register UInt8  *endS;
    register UInt8  *pixS;
    Int32            pplS;
    register Bool    potential_max;
    register UInt8   val;
    Int32            i, maxVal;

        /*----- Initializations and parameter checking -----*/

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1|cChannels3,cFrame32);

    ERR_PASS( w_pixrGetDepth (pIPWState, prSrc, &depth));
    vdepth = 8;

    maxVal = (1 << vdepth) - 1;
    if (*ppHisto != NULL)
    {
	ERR_GEN (HISTO_GET_MAXVAL(*ppHisto) != maxVal ||
	    HISTO_GET_MINVAL(*ppHisto) != 0, cErrInvalidParm);
    }
    ERR_PASS( w_pixrGetWidth(pIPWState, prSrc, &width));
    ERR_PASS( w_pixrGetHeight(pIPWState, prSrc, &height));
    ERR_PASS( w_pixrGetBpl(pIPWState, prSrc, &pplS));
    ERR_PASS( w_pixrGetImage(pIPWState, prSrc, &pixelsS));

    if (*ppHisto == NULL) 
    {
	ERR_PASS( w_newBlankHisto(pIPWState, 0, maxVal, 1, ppHisto));
    }
    else
	ERR_PASS( w_clearHisto(pIPWState, *ppHisto));

    hdata = HISTO_GET_DATA(*ppHisto);

       /*----- Loops on lines -----*/

    endS = pixelsS + pplS * height;

    if (threshold_above) 
    {

	hdata[0] = height;

	for (lineS = pixelsS; lineS < endS; lineS += pplS)
	{

	    potential_max = cTrue;
	    if (depth == 4)
	    {
		val = UCHAR_ACCESS (lineS) & 0xf0;
	    }
	    else
	    {
		val = UCHAR_ACCESS (lineS);
	    }
	    pixS = lineS;

	    for (i = 0; i < width; i++)
	    {
		if (depth == 4)
		{
		    if (i & 1)
			this = (byte & 0xf) << 4;
		    else
		    {
			byte = UCHAR_ACCESS (pixS + (i >> 1));
			this = byte & 0xf0;
		    }
		}
		else
		{
		    this = UCHAR_ACCESS(pixS + i);
		}

		    /* If same value as previous pixel,
		       nothing new happened. Keep going */
		if (this == val)
		    continue;

		    /* If current pixel pixS is higher than val,
		       then it all depends. If previous pixel
		       was a potential maximum, it no longer is,
		       but pixS is now one. If however, previous
		       pixel was instead a potential minimum,
		       we are now sure that it is a minimum. The
		       derivative of the histogram that we are looking
		       for will therefore be increased by 1
		       where appropriate */

		else if (this > val) 
		{
		    if (potential_max)
			val = (UInt8) this;
		    else
		    {
			hdata[val+1]++;
			potential_max = cTrue;
		    }
		}

		    /* Otherwise, pixS is strictly smaller than val.
		       Do something similar as what was done above */

		else 
		{
		    if (!potential_max)
			val = (UInt8) this;
		    else 
		    {
			hdata[val+1]--;
			potential_max = cFalse;
		    }
		}
	    }

	        /* Don't forget last pixel of line */

	    if (potential_max)
		hdata[val+1]--;

	}
    }

    else 
    {
	hdata[maxVal] = height;

	for (lineS = pixelsS; lineS < endS; lineS += pplS) 
	{
	    potential_max = cFalse;
	    if (depth == 4)
	    {
		/* It's already shifted up */
		val = UCHAR_ACCESS (lineS) & 0xf0;
	    }
	    else
	    {
		val = UCHAR_ACCESS (lineS);
	    }
	    pixS = lineS;

	    for (i = 0; i < width; i++)
	    {
		if (depth == 4)
		{
		    /* Shift up */
		    if (i & 1)
			this = (byte & 0xf) << 4;
		    else
		    {
			byte = UCHAR_ACCESS (pixS + (i >> 1));
			this = byte & 0xf0;
		    }
		}
		else
		    this = UCHAR_ACCESS(pixS + i);

		if (this == val)
		    continue;

		else if (this < val) 
		{
		    if (!potential_max)
			val = (UInt8) this;
		    else 
		    {
			hdata[val-1]++;
			potential_max = cFalse;
		    }
		}

		else 
		{
		    if (potential_max)
			val = (UInt8) this;
		    else 
		    {
			hdata[val-1]--;
			potential_max = cTrue;
		    }
		}
	    }

	    if (!potential_max)
		hdata[val-1]--;

	}
    }
	
        /*----- Now integrates the histogram if requested -----*/

    if (!differentiate) 
    {
	if (threshold_above) 
	{
	    for (i = 0 + 1; i <= maxVal; i++)
		hdata[i] += hdata[i-1];
	}
	else 
	{
	    for (i = maxVal - 1; i >= 0; i--)
		hdata[i] += hdata[i+1];
	}
    }

    NORMAL_RETURN;
} /* HorizRunsHistogram */


/* 
 * $Log: cchisto.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 * Imported sources
 *
 * Revision 1.19  1996/01/04 20:21:15  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.18  1995/12/13  04:05:08  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.17  1995/11/13  20:09:39  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/10/13  21:58:44  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/10/12  17:42:43  macleod
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * added explicit casts (float->int, long->short, long->char, etc) and
 * made floating constants the right size to shut up the NT compiler
 *
 * Revision 1.14  1995/10/03  01:59:21  macleod
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.13  1995/09/21  22:51:36  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/07  01:29:41  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:17:22  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  21:02:37  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.9  1995/08/31  21:11:11  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/25  01:35:09  ejaquith
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * Fixed warnings
 *
 * Revision 1.7  1995/08/17  21:53:53  macleod
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.6  1994/10/24  19:38:41  macleod
 * macleod on Mon Oct 24 12:36:31 PDT 1994
 *
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * changed horizRunsHistogram to upscale 4-bit images to 8-bit rather
 * rather than make a 4-bit histogram. This alleviates a problem which
 * caused 4-bit histograms to sometimes be pathological. However,
 * it doesn't really address the root cause, which could probably
 * occur in 8-bit images, so it should be examined more closely.
 *
 * Revision 1.5  1994/10/07  19:06:33  lperry
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.4  1994/09/29  00:42:22  lperry
 * lperry on Wed Sep 28 17:41:49 PDT 1994
 *
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * Eliminate use of utilpros. Use hutils.pub instead.
 *
 * Revision 1.3  1994/08/01  20:17:22  macleod
 * macleod on Mon Aug 1 13:16:50 PDT 1994
 *
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * Fixed a couple of comparisons which generated a (stupid) warning with
 * Solaris "cc".
 *
 * Revision 1.2  1994/06/16  22:34:44  macleod
 * macleod on Thu Jun 16 15:34:12 PDT 1994
 *
 * cchisto.c -> /product/ipcore/ipwrap/src/RCS/cchisto.c,v
 *
 * Modified horizRunsHistogram so that it works on 4 bit/pixel images
 *
 * Revision 1.1  1994/06/08  20:32:06  macleod
 * Initial revision
 *
 */
