/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   ccittutl.c:
 *
 *	Utility routines for CCITT compression.
 *	
 *		
 *	w_getSizeOfCCITTCompressionInfo():	returns size of memory block
 *						needed for opaque state block
 *						used by CCITT
 *						compression.
 *
 *	w_getSizeOfCCITTDecompressionInfo():	returns size of memory block
 *						needed for opaque state block
 *						used by CCITT
 *						decompression.
 *
 *	w_initializeCCITTCompressionInfo():	initializes opaque CCITT
 *						state block
 *
 *	w_initializeCCITTDecompressionInfo():	initializes opaque CCITT
 *						state block
 *
 *	w_writeCCITTEOL():			write EndOfLine code to
 *						pipeline
 *
 *	w_flushCCITT():				flushes CCITT compression
 *						pipeline
 *
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "ccitt.pub"
#include "shrcmprs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: ccittutl.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


/* This file contains wrappers for routines used to compress images
 * into Modified Huffman bytestreams.
 *
 * The compressed data stream is always in the natural byte order of the
 * current platform.
 *
 * To avoid interactions with client memory allocators, we allow the
 * client to allocate and free all memory structures.  The routine
 * getSizeOfCCITTCompressionInfo is used by the client to get the
 * size of the opaque state block used in CCITT MH, MR and MMR compression.
 * Similarly, getSizeOfCCITTDecompressionInfo returns the size of the
 * opaque state block used in decompression.
 *
 * The routines initializeCCITTCompressionInfo and
 * initializeCCITTDecompressionInfo are used to initialize these state blocks.
 * They sets up the things that you wouldn't want to calculate on every
 * compression call.
 *
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Returns size of private CCITT compression state block.
 *
 * LongDescription:::
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform CCITT
 *	compression.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block
 *  w::              Width of image not including frame.
 *
 * Output:::
 *  pSize::          Pointer to returned size in bytes.
 *
 * Notes:::
 *  CCITT compression uses a private state block that is allocated either
 *  locally or by the client.  This procedure returns the size of this
 *  block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MH, MR, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how Modified Huffman
 *  compression really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSizeOfCCITTCompressionInfo (
    IPWState           *pIPWState,
    UInt32              w,
    UInt32             *pSize)
{
    PROLOGUE;

    ERR_GEN( pSize == NULL, cErrNullPtr);

    ERR_PASS( i_getSizeOfCCITTCompressionInfo (pIPWState,
					       w + 2*cFrameBits,
					       (Int32 *)pSize));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Returns size of private CCITT decompression state block.
 *
 * LongDescription:::
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform CCITT
 *	decompression.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block
 *  w::              Width of image not including frame.
 *
 * Output:::
 *  pSize::          Pointer to returned size in bytes.
 *
 * Notes:::
 *  CCITT decompression uses a private state block that is allocated either
 *  locally or by the client.  This procedure returns the size of this
 *  block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MH, MR, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how Modified Huffman
 *  compression really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSizeOfCCITTDecompressionInfo (
    IPWState           *pIPWState,
    UInt32              w,
    UInt32             *pSize)
{
    PROLOGUE;

    ERR_GEN( pSize == NULL, cErrNullPtr);

    ERR_PASS( i_getSizeOfCCITTDecompressionInfo (pIPWState,
						 w + 2*cFrameBits,
						 (Int32 *)pSize));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize CCITT compression state block.
 *
 * LongDescription:::
 *  This procedure fills in a previously allocated state block with
 *  the information required for CCITT compression.
 *
 * Input:::
 *  pIPWState::         Pointer to standard IPWrap state block.
 *  pCCITTCompInfoIn::  Pointer to private CCITT state block.
 *  w::                 Width of image to be compressed not including frame.
 *  useMR2D::		If set, use MR (CCITT Group 3) compression with
 *			the hybrid 1D/2D option.  If not set, either use
 *			some other type of compression or don't use the
 *			1D/2D option for MR.
 *  maxK::		if useMR2D is set, this should be either 2 or 4.
 *			If equal to 2, each 1D encoded line is followed by
 *			a 2D encoded line.  If equal to 4, each 1D line
 *			is followed by 3 2D lines.
 *  padType::		Type of scan line padding.
 *			If compression is MH:
 *			    cPadToByte => pad each scan
 *				line to a byte boundary.
 *			    cPadToWord => pad each scan
 *				line to a word boundary.
 *			 if compression is MR:
 *			    cNoPad => insert no pad before EOL.
 *			    cPadToByte => insert pad before
 *				EOL so EOL ends on byte bdry.
 *			 if compression = MMR:
 *				ignored.
 *  compInfoBufSize::   Size of the private state block in bytes.
 *
 * Returns:::
 *  In addition to the normal codes, this routine returns cErrInvalidParm if
 *  the private structure's size (compInfoBufSize) is too small or useMR2D
 *  is defined and maxK is not equal to 2 or 4.
 *
 * Notes:::
 *  CCITT (MH, MR, MMR) compression uses a private state block that is
 *  allocated either locally or by the client.  This procedure fills
 *  in the contents of this block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MH, MR, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how Modified Huffman
 *  really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initializeCCITTCompressionInfo (
    IPWState           *pIPWState,
    void               *pCCITTCompInfoIn,
    UInt32              w,
    Bool		useMR2D,
    Int32	 	maxK,
    UInt32		padType,
    UInt32              compInfoBufSize)
{
    PROLOGUE;

    ERR_GEN( pCCITTCompInfoIn == NULL, cErrNullPtr);

    ERR_GEN( useMR2D && (maxK != 2) && (maxK != 4), cErrInvalidParm);

    ERR_GEN( (padType != cNoPad) && (padType != cPadToByte) &&
	     (padType != cPadToWord), cErrInvalidParm);

    ERR_PASS( i_initializeCCITTCompressionInfo (pIPWState, pCCITTCompInfoIn,
					        w + 2*cFrameBits,
					        useMR2D, maxK, padType,
					        compInfoBufSize));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize CCITT decompression state block.
 *
 * LongDescription:::
 *  This procedure fills in a previously allocated state block with
 *  the information required for CCITT decompression.
 *
 * Input:::
 *  pIPWState::           Pointer to standard IPWrap state block.
 *  pCCITTDecompInfoIn::  Pointer to private CCITT state block.
 *  w::                   Width of image to be compressed not including frame.
 *  useMR2D::		If set, use MR (CCITT Group 3) compression with
 *			the hybrid 1D/2D option.  If not set, either use
 *			some other type of compression or don't use the
 *			1D/2D option for MR.
 *  padType::		Type of scan line padding.
 *			If compression is MH:
 *			    cPadToByte => pad each scan
 *				line to a byte boundary.
 *			    cPadToWord => pad each scan
 *				line to a word boundary.
 *			 if compression is MR:
 *			    cNoPad => insert no pad before EOL.
 *			    cPadToByte => insert pad before
 *				EOL so EOL ends on byte bdry.
 *			 if compression = MMR:
 *				ignored.
 *  decompInfoBufSize::   Size of the private state block in bytes.
 *
 * Returns:::
 *  In addition to the normal codes, this routine returns cErrInvalidParm if
 *  the private structure's size (decompInfoBufSize) is too small or useMR2D
 *  is defined and maxK is not equal to 2 or 4.
 *
 * Notes:::
 *  CCITT (MH, MR, MMR) decompression uses a private state block that is
 *  allocated either locally or by the client.  This procedure fills
 *  in the contents of this block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MH, MR, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how Modified Huffman
 *  really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initializeCCITTDecompressionInfo (
    IPWState           *pIPWState,
    void               *pCCITTDecompInfoIn,
    UInt32              w,
    Bool		useMR2D,
    UInt32		padType,
    UInt32              decompInfoBufSize)
{
    PROLOGUE;

    ERR_GEN( pCCITTDecompInfoIn == NULL, cErrNullPtr);

    ERR_GEN( (padType != cNoPad) && (padType != cPadToByte) &&
	     (padType != cPadToWord), cErrInvalidParm);

    ERR_PASS( i_initializeCCITTDecompressionInfo (pIPWState, pCCITTDecompInfoIn,
					          w + 2*cFrameBits,
					          useMR2D, padType,
					          decompInfoBufSize));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Flushes the CCITT compression pipeline to the compressed data stream.
 *
 * LongDescription:::
 *	The CCITT compression routines maintain a pipeline used to
 *	accumulate codes into a bytestream.  This routine is used to
 *	flush the bits remaining in the pipeline to the compressed data
 *	stream.  It is used during MR and MMR compression when the
 *	pipeline is not flushed at the end of individual scan lines.
 *
 * Input:::
 *  pIPWState::		    Pointer to standard IPWrap state block
 *  pCCITTCompInfoIn::	    Pointer to private CCITT state block.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    written.  If several calls
 *			    to this routine are needed to create
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *
 * Output:::
 *  pBytesWritten::         Number of compressed bytes actually
 *			    written to the compressed data
 *			    block.  If flushPipeline is false,
 *			    some bytes required to complete
 *			    a scan line may still be in the
 *			    pipeline and are not included in
 *			    this count.
 *
 * Notes:::
 *  CCITT compression uses a private state block that is allocated either
 *  locally or by the client.  The pipeline is contained in that state
 *  block.  The pipeline is normally flushed at the end of a strip or tile,
 *  or at the end of the entire image if that was written in one block..
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MR, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  "Facsimile Coding Schemes and Coding Control Functions for Group 4
 *  Facsimile Apparatus," Recommendation T.6, Volume VII, Fascicle VII.3,
 *  Terminal Equipment and Protocols for Telematic Services.  The
 *  International Telegraph and Telephone Consultative Committee (CCITT),
 *  Geneva, 1985, pages 40 through 48.  You can write to ANSI,
 *  Attention: Sales, 1430 Broadway, New York, N.Y., 10018.  Ask for the
 *  publication listed above.
 *  Or, if this isn't out of date by the time you read this, try
 *  http://icib.igd.fhg.de/icib/telecom/ccitt/rec_t.6/gen.html.  The
 *  version I saw was lame - it didn't have any of the VERY useful figures.
 *  Section 11 of the TIFF 6.0 Specification is also helpful since it
 *  talks about using compression in files instead of in fax machines.
 *  
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_flushCCITT (
    IPWState           *pIPWState,
    void               *pCCITTCompInfoIn,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32             *pBytesWritten)
{
    PROLOGUE;

    ERR_GEN( pCCITTCompInfoIn == NULL || pCompData == NULL ||
	     pBytesWritten == NULL, cErrNullPtr);

    ERR_PASS( i_flushCCITT (pIPWState, pCCITTCompInfoIn,
			    pCompData, compDataBufSize,
			    (Int32 *)pBytesWritten));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Writes an EndOfLine code to a CCITT bytestream.
 *
 * LongDescription:::
 *	The CCITT compression routines maintain a pipeline used to
 *	accumulate codes into a bytestream.  This routine is used to
 *	insert an EndOfLine code into this bytestream.  The EOL may
 *	optionally be preceeded by padding so that the EOL ends on
 *	a byte boundary.  It may also optionally be followed
 *	by a bit that indicates whether the next line is encoded
 *	using a 1D MH method or a 2D MMR method.
 *
 * Input:::
 *  pIPWState::		    Pointer to standard IPWrap state block
 *  pCCITTCompInfoIn::	    Pointer to private CCITT state block.
 *  padType::		    If equal to cPadToByte, insert pad
 *			    before EOL so that EOL ends on a
 *			    byte boundary.  If equal to cNoPad, 
 *			    do not insert padding before EOL.
 *  EOLType::		    If equal to cEOL1D, follow the EOL
 *			    with a bit indicating that the next 
 *			    line will be encoded with 1D encoding.
 *			    If equal to cEOL2D, append a bit 
 *			    indicating that the next line will 
 *			    be encoded using 2D codes.  If 
 *			    cPlainEOL, do not append a bit.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    written.  If several calls
 *			    to this routine are needed to create
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *
 * Output:::
 *  pBytesWritten::         Number of compressed bytes actually
 *			    written to the compressed data
 *			    block.  If flushPipeline is false,
 *			    some bytes required to complete
 *			    a scan line may still be in the
 *			    pipeline and are not included in
 *			    this count.
 *
 * Notes:::
 *  CCITT compression uses a private state block that is allocated either
 *  locally or by the client.  The pipeline is contained in that state
 *  block.  This routine may be used to insert an EOL at the end of each
 *  scan line (MR compression) and/or at the end of the entire strip or
 *  tile (MMR compression)..
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MR, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  "Facsimile Coding Schemes and Coding Control Functions for Group 4
 *  Facsimile Apparatus," Recommendation T.6, Volume VII, Fascicle VII.3,
 *  Terminal Equipment and Protocols for Telematic Services.  The
 *  International Telegraph and Telephone Consultative Committee (CCITT),
 *  Geneva, 1985, pages 40 through 48.  You can write to ANSI,
 *  Attention: Sales, 1430 Broadway, New York, N.Y., 10018.  Ask for the
 *  publication listed above.
 *  Or, if this isn't out of date by the time you read this, try
 *  http://icib.igd.fhg.de/icib/telecom/ccitt/rec_t.6/gen.html.  The
 *  version I saw was lame - it didn't have any of the VERY useful figures.
 *  Section 11 of the TIFF 6.0 Specification is also helpful since it
 *  talks about using compression in files instead of in fax machines.
 *  
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeCCITTEOL (
    IPWState           *pIPWState,
    void               *pCCITTCompInfoIn,
    UInt32		padType,
    UInt32		EOLType,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32             *pBytesWritten)
{
    PROLOGUE;

    ERR_GEN( pCCITTCompInfoIn == NULL || pCompData == NULL ||
	     pBytesWritten == NULL, cErrNullPtr);

	/* Make sure the pad and EOL types are legal */
    ERR_GEN( (padType != cPadToByte) && (padType != cNoPad),
	     cErrInvalidParm);
    ERR_GEN( (EOLType != cEOL1D) && (EOLType != cEOL2D) &&
	     (EOLType != cPlainEOL), cErrInvalidParm);

    ERR_PASS( i_writeCCITTEOL (pIPWState, pCCITTCompInfoIn, 
			       padType, EOLType,
			       pCompData, compDataBufSize,
			       (Int32 *)pBytesWritten));
    
    NORMAL_RETURN;
}



/*

 $Log: ccittutl.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.5  1998/01/14 00:10:49  ddavies
 ccittutl.c -> /product/ipcore/ipwrap/src/RCS/ccittutl.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.4  1996/05/11 01:28:42  ddavies
 ccittutl.c -> /product/ipcore/ipwrap/src/RCS/ccittutl.c,v

 Add useMR2D, maxK and padType to w_initializeCCITTCompressionInfo.The first two are used when compressing using the 2D option of CCITT group 3.
 The last used to be a parameter for the MH compression functions, but has
 been moved to the private structures.
 Add useMR2D and padType to w_initializeCCITTDecompressionInfo.  We don't need
 maxK since we determine the type of the next scan line's encoding by
 looking at the type of the EndOfLine code that preceeds (!) it.

 Revision 1.3  1996/05/01 23:40:43  ddavies
 ddavies on Wed May 1 16:39:17 PDT 1996

 ccittutl.c -> /product/ipcore/ipwrap/src/RCS/ccittutl.c,v

 Add two new routines, w_flushCCITT and w_writeCCITTEOL.  The first routine
 flushes the pipeline used in MH, MR and MMR compression to the output
 bytestream.  The second writes an EndOfLine code to the output.

 Revision 1.2  1996/04/25 19:08:04  ddavies
 ccittutl.c -> /product/ipcore/ipwrap/src/RCS/ccittutl.c,v

 Added implementations of wrapper routines w_getSizeOfCCITTDecompInfo and
 whoops! that's w_getSizeOfCCITTDecompressionInfo and
 w_initializeCCITTDecompressionInfo.  These routines are used by clients
 to get the size of the opaque state block used by CCITT decompression
 routines and to initialize that state block.  The client is responsible
 for freeing the block.


 */

