/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  ccs.c - Functions that deal with the ConnectedComponent data structure
 *
 * Exports:
 *    w_CCSet ()
 *    w_CCClear ()
 *    w_CCInterpolateEdges ()
 *    w_CCDestroy ()
 *    w_CCPlugHorizontalGaps ()
 *    w_runListToCC ()
 *    w_findConnectedComponents ()
 *    w_paintConnectedComponents ()
 *    w_interpolateEdges ()
 *    w_destroyConnectedComponents ()
 *    w_plugHorizontalGaps ()
 */

#include "ipwcore.prv"

#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrcc.pub"

#include "ccs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: ccs.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

/* local def of callback "packet" */
typedef struct _ccPacket
{
    ConnectedComponent    *ccList;
} ccPacket;


/* Forward refs for local functions */
static Int32 CDECL
wp_findCCCallback(
    IPWState	       *pIPWState,
    void 	       *packet,
    Int32		x,
    Int32		y,
    Int32		w,
    Int32		h, 
    Int32		mass,
    Int32		nRuns,
    TCScanLineRun      *pRuns);


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Paint the runs of a connected component into a binary PIXR, writing black
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  component::      Connected component to paint
 *  destX::          X coordinate of upper left corner of component
 *  destY::          Y coordinate of upper left corner of component
 *
 * Output:::
 *  prDest::         PIXR in which to paint the component
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis 
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_CCSet (
    IPWState              *pIPWState,
    ConnectedComponent    *component,
    Int32                  destX,
    Int32                  destY,
    PIXR                  *prDest)
{
    Int32   bplD;
    UInt8  *iD;
    PROLOGUE;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetImageAll (pIPWState, prDest, &iD));
    ERR_PASS (w_pixrGetBpl (pIPWState, prDest, &bplD));

    ERR_PASS (i_CCSet (pIPWState, iD,
		       bplD,
		       destX,
		       destY,
		       component));

    NORMAL_RETURN;

}  /* w_CCSet () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Paint a connected component into a gray PIXR, interpolating the gray value.
 *
 * LongDescription:::
 *  Paint what's under a connected component with an interpolated value which 
 *  runs smoothly from the value at the left edge to the value at the right
 *  edge.  The destination must be 8 bits/pixel.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  component::      Component to paint
 *  destX::          X coordinate of upper left corner of component
 *  destY::          Y coordinate of upper left corner of component
 *
 * Output:::
 *  prDest::         PIXR in which to paint the component
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_CCInterpolateEdges (
    IPWState              *pIPWState,
    ConnectedComponent    *component,
    Int32                  destX,
    Int32                  destY,
    PIXR                  *prDest)
{
    Int32   bplD;
    UInt8  *iD;
    PROLOGUE;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetImageAll (pIPWState, prDest, &iD));
    ERR_PASS (w_pixrGetBpl (pIPWState, prDest, &bplD));

    ERR_PASS (i_CCInterpolateEdges (pIPWState, iD,
				    bplD,
				    destX,
				    destY,
				    component));

    NORMAL_RETURN;

}  /* w_CCInterpolateEdges () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Paint the runs of a connected component into a PIXR, writing white
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  component::      Connected component to paint
 *  destX::          X coordinate of upper left corner of component
 *  destY::          Y coordinate of upper left corner of component
 *
 * Output:::
 *  prDest::         PIXR in which to paint the component
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_CCClear (
    IPWState              *pIPWState,
    ConnectedComponent    *component,
    Int32                  destX,
    Int32                  destY,
    PIXR                  *prDest)
{
    Int32   bplD;
    UInt8  *iD;
    PROLOGUE;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetImageAll (pIPWState, prDest, &iD));
    ERR_PASS (w_pixrGetBpl (pIPWState, prDest, &bplD));


    ERR_PASS (i_CCClear (pIPWState, iD,
		         bplD,
		         destX,
		         destY,
		         component));

    NORMAL_RETURN;

}  /* w_CCClear () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free a ConnectedComponent object
 *
 * LongDescription:::
 *  Sets the component pointer to NULL
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  component::      Connected component to be freed
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_CCDestroy (
    IPWState              *pIPWState,
    ConnectedComponent   **component)
{
    PROLOGUE;

    *component = i_CCDestroy (pIPWState, *component);

    ERR_GEN ((*component != NULL), cErrInternal);

    NORMAL_RETURN;

}  /* w_CCDestroy () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Transforms a connected component into a new one with no interior white runs
 *
 * LongDescription:::
 *  Transforms a connected component into one where there is only
 *  one run per scanline, starting at the left edge of the leftmost
 *  run of the scanline, and ending at the right edge of the rightmost
 *  run of the scanline. This guarantees that we plug all interior holes,
 *  which allows one to use fairly stupid schemes for "peeling off" the
 *  cc and filling the background with some color (like the mean or
 *  median around the edges of the cc, for example).
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  src::            Source component
 *
 * Output:::
 *  dest::           Destination component
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 * connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_CCPlugHorizontalGaps (
    IPWState              *pIPWState,
    ConnectedComponent    *src,
    ConnectedComponent   **dest)
{
    Int32 size;
    PROLOGUE;

    if (*dest == NULL)
    {
	ERR_PASS( i_CCPlugSize (pIPWState, src, &size));
	ERR_GEN ((size <= 0), cErrInternal);

	*dest = (ConnectedComponent *) IPW_MALLOC (size, cAllocatorData);
	ERR_GEN (!(*dest), cErrNomem);
    }

    ERR_PASS (i_CCPlugHorizontalGaps (pIPWState, *dest, src));

    NORMAL_RETURN;

}  /* w_CCPlugHorizontalGaps () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a list of TCScanLineRun to a ConnectedComponent structure
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  scanLineRuns::   The TCScanLineRun to convert
 *  x::              Minimum X of bounding box (including frame)
 *  y::              Minimum Y of bounding box (including frame)
 *  w::              Width of bounding box (including frame)
 *  h::              Height of bounding box (including frame)
 *  nRuns::          Number of runs in the list
 *
 * Output:::
 *  component::      Resulting connected component
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: Bounding box of the runs. Note that the bounding box is relative 
 *        to the OUTSIDE of the frame, since this is usually called as a 
 *        result of rabbRuns, which gives its bounding box that way.

 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_runListToCC (
    IPWState              *pIPWState,
    TCScanLineRun         *scanLineRuns,
    Int32                  x,
    Int32                  y,
    Int32                  w,
    Int32                  h,
    Int32                  nRuns,
    ConnectedComponent   **component)
{
    ConnectedComponent *cc;
    PROLOGUE;

    ERR_PASS( i_runListToCC (pIPWState, scanLineRuns, x, y, w, h, nRuns, &cc));

    ERR_GEN ((cc == NULL), cErrNomem);

    *component = cc;
    NORMAL_RETURN;

}  /* w_runListToCC () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Finds all connected components in a binary PIXR
 *
 * LongDescription:::
 *  Makes a linked list of all components which are at least minWidth
 *  wide and minHeight tall.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  pImage::         Source PIXR
 *  minWidth::       Minimum Connected component width to consider
 *  minHeight::      Minimum Connected component height to consider
 *  connectType::    0 = 4-connected, 1 = 8-connected
 *
 * Output:::
 *  ccList::         List of ConnectedComponent structures
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findConnectedComponents (
	IPWState    *pIPWState,
	PIXR       *pImage,
	UInt32      minWidth,
	UInt32      minHeight,
	UInt32      connectType,
	ConnectedComponent **ccList)
{
    Int32   bpl, w, h, retCode;
    UInt8  *iS;
    ccPacket packet;
    PROLOGUE;

    ERR_GEN_PIXR (pImage, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, pImage, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, pImage, &h));
    ERR_PASS (w_pixrGetBpl (pIPWState, pImage, &bpl));
    ERR_PASS (w_pixrGetImageAll (pIPWState, pImage, &iS));

    /* Do the real work */
    packet.ccList = NULL;
    ERR_PASS (i_findConnectedComponentRuns (pIPWState, iS,
                                            bpl,
					    cFrameBits,
					    cFrameBits,
					    w, h,
					    minWidth, minHeight,
					    connectType,
					    (void *)(&packet),
					    wp_findCCCallback,
					    &retCode));
    *ccList = packet.ccList;
    NORMAL_RETURN;

}  /* w_findConnectedComponents () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Private callback used by w_findConnectedComponents ()
 *
 * LongDescription:::
 *  This is passed into the IPShared level CC finder.  The packet
 *  contains a list pointer where CCs structures are collected by
 *  insertion at the head.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Pointer to IPCore state
 *  packetV::        Pointer to info packet
 *  x::              Minimum x value of current CC
 *  y::              Minimum y value of current CC
 *  w::              Width of current CC
 *  h::              Height of current CC
 *  mass::           Mass (number of 1 pixels) of current CC
 *  nRuns::          Number of horizontal runs in current CC
 *  pRuns::          Pointer to list of runs for current CC
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static Int32 CDECL
wp_findCCCallback (
    IPWState	       *pIPWState,
    void               *packetV,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               mass,
    Int32               nRuns,
    TCScanLineRun      *pRuns)
{
    ccPacket           *packet;
    ConnectedComponent *cc;
    ErrCode             __error__;

    packet = (ccPacket *)packetV;
    mass = mass;

    /* The, uh, error macro expansion here was done because our return type
     * is Int32, but the "official" return type for error macros is
     * "ErrCode".  So be explicit here to avoid warning whining by
     * error wrapping checking.
     *
     * This is intended as a stop-gap until IPShared gets updated
     * (and depending on your disposition, append ", if ever" to above).
     * --ejaquith--
     */

    /* Make a cc */
    ERR_CHECK(w_runListToCC (pIPWState, pRuns, 
			     x, y, w, h, nRuns, &cc)) {
	return (Int32) __error__;
    }

    ERR_GEN_CHECK(cc == NULL, cErrNomem) {
	return (Int32) __error__;
    }
	    

    /* Add it to the head */
    cc->next = packet->ccList;
    packet->ccList = cc;
    
    return cSuccessful;

}  /* wp_findCCCallback () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Given a list of CCs, paint them into a binary PIXR
 *
 * LongDescription:::
 *  The CCs are painted with the value in color and with offset (destX,
 *  destY) in the PIXR.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  ccList::         List of CCs to paint
 *  destX::          X offset in destination PIXR
 *  destY::          Y offset in destination PIXR
 *  color::          Value to paint runs of each CC
 *
 * Output:::
 *  prDest::        Destination PIXR 
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_paintConnectedComponents (
    IPWState              *pIPWState,
    ConnectedComponent    *ccList,
    Int32                  destX,
    Int32                  destY,
    Int32                  color,
    PIXR                  *prDest)
{
    PROLOGUE;
    UInt8 *pImg;
    Int32  bpl;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);


    ERR_PASS (w_pixrGetImageAll (pIPWState, prDest, &pImg));
    ERR_PASS (w_pixrGetBpl (pIPWState, prDest, &bpl));

    while (ccList)
    {
	if (color == 1)
	    ERR_PASS (i_CCSet (pIPWState, pImg, bpl, destX, destY, ccList));
	else
	    ERR_PASS (i_CCClear (pIPWState, pImg, bpl, destX, destY, ccList));

	ccList = ccList->next;
    }

    NORMAL_RETURN;

}  /* w_paintConnectedComponentsPr () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Uses w_CCInterpolateEdges () to paint a list of CCs into an 8-bit PIXR
 *
 * LongDescription:::
 *  
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  ccList::         List of components to paint
 *  destX::          X offset into destination PIXR
 *  destY::          Y offset into destination PIXR
 *
 * Output:::
 *  prDest::         Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_interpolateEdges (
    IPWState              *pIPWState,
    ConnectedComponent    *ccList,
    Int32                  destX,
    Int32                  destY,
    PIXR                  *prDest)
{
    PROLOGUE;
    UInt8 *pImg;
    Int32  bpl;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetImageAll (pIPWState, prDest, &pImg));
    ERR_PASS (w_pixrGetBpl (pIPWState, prDest, &bpl));

    while (ccList)
    {
	ERR_PASS (i_CCInterpolateEdges (pIPWState, pImg,
					bpl, destX, destY, ccList));

	ccList = ccList->next;
    }

    NORMAL_RETURN;

}  /* w_interpolateEdgesPr () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroys a linked list of connected components
 *
 * LongDescription:::
 *  Also sets the list pointer to NULL.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  ccList::         List to be freed
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyConnectedComponents (
    IPWState              *pIPWState,
    ConnectedComponent   **ccList)
{
    PROLOGUE;
    ConnectedComponent *ccNext;

    while (*ccList)
    {
	ccNext = (*ccList)->next;
	*ccList = i_CCDestroy (pIPWState, *ccList);
	ERR_GEN ((*ccList != NULL), cErrInternal);
	*ccList = ccNext;
    }

    NORMAL_RETURN;

}  /* w_destroyConnectedComponents () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Given a list of CCs, calls w_CCPlugHorizontalGaps () on each component
 *
 * LongDescription:::
 *  Creates a new list of CCs, where each new CC is the result of calling
 *  w_CCPlugHorizontalGaps () on the corresponding CC in the input list.
 *  The original list is not destroyed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  ccList::         Input list of CCs
 *
 * Output:::
 *  plugList::       Output list of CCs (with holes plugged)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_plugHorizontalGaps (
    IPWState              *pIPWState,
    ConnectedComponent    *ccList,
    ConnectedComponent   **plugList)
{
    ConnectedComponent *ccHead, *ccNew, *ccPrev;
    PROLOGUE;

    ccHead = ccPrev = NULL;

    while (ccList != NULL)
    {
	ccNew = NULL;
	ERR_PASS_CLEANUP (w_CCPlugHorizontalGaps (pIPWState, ccList, &ccNew));

	ccNew->next = NULL;
	if (ccHead == NULL)
	{
	    ccHead = ccNew;
	    ccPrev = ccNew;
	}
	else
	{
	    ccPrev->next = ccNew;
	    ccPrev = ccNew;
	}

	ccList = ccList->next;
    }
    
    *plugList = ccHead;

    NORMAL_RETURN;

ERR_CLEANUP:
    /* After some discussion it was decided that since we only
     * are able to handle one level of error, error checking
     * should not occur during cleanup from a previous error
     * in order that the original error be returned.
     */
    if (ccHead != NULL)
	w_destroyConnectedComponents (pIPWState, &ccHead);

    ERR_RETURN;

}  /* w_plugHorizontalGaps () */


/*
 * $Log: ccs.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 * Imported sources
 *
 * Revision 1.19  1998/01/14 00:10:51  ddavies
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.18  1996/01/04 20:21:18  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.17  1995/12/13  04:05:12  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/20  19:27:24  ejaquith
 * ejaquith on Mon Nov 20 11:26:19 PST 1995
 *
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * Added some workaround code to wp_findCCCallback () to gag warnings
 * generated from nightly C code checking.
 *
 * Revision 1.15  1995/11/13  20:09:43  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.14  1995/11/08  19:52:13  dferrell
 * dferrell on Wed Nov 8 11:51:21 PST 1995
 *
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * Changed logic to replace ERR_CHECKs with ERR_PASS_CLEANUPs even in the
 * most brain-dead cases (I hate gotos even in macros).
 *
 * Revision 1.13  1995/10/13  21:58:47  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.12  1995/09/29  01:41:50  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * Fixed keywords typo
 *
 * Revision 1.11  1995/09/28  21:40:09  dferrell
 * dferrell on Thu Sep 28 14:39:38 PDT 1995
 *
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.10  1995/09/21  22:51:39  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.9  1995/09/11  21:20:53  dferrell
 * dferrell on Mon Sep 11 14:20:34 PDT 1995
 *
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.8  1995/09/07  01:29:44  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.7  1995/09/06  20:17:25  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.6  1995/09/01  17:07:49  dferrell
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.5  1995/08/31  21:11:13  ejaquith
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.4  1995/08/23  22:33:05  jcraig
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * in findConnectedComponents correctly return the ccList
 *
 * Revision 1.3  1995/08/17  21:52:31  dferrell
 * dferrell on Thu Aug 17 14:51:15 PDT 1995
 *
 * ccs.c -> /product/ipcore/ipwrap/src/RCS/ccs.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 *
 *
 */

