/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
** File:        ccthresh.c
**
** Description: MAIN for connected components based thresholding
**
** Author: Luc Vincent
**	   Edited into IPWRAP by P. MacLeod
**	   P. MacLeod added a bunch of background-thresholding stuff
*/

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "alplib.h"
#include "histo.h"
#include "defines.pub"
#include "memory.pub"

/* prototypes */
#include "props.pub"
#include "ccthresh.prv"
#include "ccthresh.pub"
#include "curves.pub"
#include "hutils.pub"
#include "cchisto.pub"
#include "shrip.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: ccthresh.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


    /* #define this to ouput a bunch of histograms for debugging */

/*#define DEBUG_CCTHRESH */

/*********************************************************************
 ****************** PARAMETERS USED BY PROGRAM ***********************
 *********************************************************************/

    /*
     ** Parameter controlling size for smoothing of original histogram.
     ** The window size actually used for the smoothing (moving
     ** average) is (WINDOW_SIZE * dynamic_range), where
     ** dynamic_range = (max(image) - min(image)) / 255.
     ** This way, we do not overly smooth histograms corresponding
     ** to images having very few levels (ex: binary images!)
     ** The size of this parameter is not really crucial and has not
     ** really been fined tuned yet.
     */

#define WINDOW_SIZE 12


    /*
    ** The parameter below characterizes how "connected" the resulting
    ** binary image should be: the higher this parameter, the more the
    ** characters will tend to be non-broken, nice, but other types
    ** of aberrations tend to occur also (missed text that is lighter
    ** than the rest, wrong dealing with shaded background areas, etc).
    ** The smaller this parameter, the more emphasis is put on how
    ** many connected components are extracted by each threshold value.
    ** In other words, characters will be more broken up (this gives
    ** more connected components!)... An optimal value has to be found
    ** for the series of images on which the thresholding algorithm will
    ** run, as well as on the OCR package to be used afterwards. Some
    ** package deal better with broken characters, some deal better
    ** with touching characters... In the series of experiments
    ** done with the xx.g images from AODS, as well as some of the
    ** grayscale images of the University of Washington CD ROM, the
    ** best default value seemed to be 10.
    **    But definitely, given the relatively small amount of
    ** experimentation performed, this parameter should be adjusted,
    ** and even totally different methods for extracting the optimal
    ** threshold from the histogram(s) should be investigated.
    **    Another information that I did not take into account is the
    ** size of the connected components extracted at each threshold level.
    ** It may be interesting at some point to look into this. I was
    ** thinking of estimating character width first, then doing some
    ** sort of linear closing on the page before applying the
    ** connected component histogramming technique.
    */

#define DEFAULT_CHARACTER_CONNECTIVITY 10

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Find text threshold using connected components
 *
 * LongDescription:::
 *      Computes the histogram of the number of connected components of
 *      each binary image obtained by thresholding the input image
 *      below each threshold between 0 and 255. 
 *   
 * PrivateDescription:::
 *      Computes the histogram of the number of connected components of
 *      each binary image obtained by thresholding the input image
 *      below each threshold between 0 and 255. In fact, for speed
 *      and memory considerations, we use the number of "black runs"
 *      at each threshold instead of the number of connected components.
 *      The slowly-varying areas of the extracted histogram correspond
 *      to candidate correct threshold points. In addition, if the
 *      number of connected components is high, the potential threshold
 *      is even better. Indeed, suppose that on the connected components
 *      histogram, we find two "flat areas": The one corresponding to the
 *      larger number of connected components (c.c.'s) is probably where
 *      the text is!
 *      <BR>
 *      To extract the best threshold, we first smooth the original
 *      histogram, then computes the absolute value of its derivative.
 *      The potential thresholds are in the "widest areas of minimal
 *      absolute derivative". A cute algorithm is therefore run, that
 *      associates with each "pixel" of the histogram a value corresponding
 *      to how wide an area of low value there is around this pixel
 *      in the absolute derivative of the original histogram. Call this
 *      the grayscale granulometry function (ggran). In fact, we need
 *      to invert the absolute derivative histogram before running
 *      the 'ggran' algorithm.
 *      <BR>
 *      The threshold corresponding to the highest value of the resulting
 *      ggran histogram is a good threshold to pick. To refine the technique
 *      even more, observe that there are often wide areas where the
 *      ggran histogram is pretty much constant. What to do in such cases?
 *      The best is usually to look at the actual number of connected
 *      components, the assumption being that in these relatively constant
 *      areas, the best threshold is probably the one giving the maximal
 *      number of connected components. In order to avoid any hacky code
 *      with a bunch of thresholds and tricks that would probably fail
 *      often, we compute a normalized sum of the original smoothed
 *      histogram, and the smoothed ggran histogram. The latter weighs
 *      much more (DEFAULT_CHARACTER_CONNECTIVITY times) than the former.
 *      Normalization is done according to the integral of each of these
 *      histogram.
 *
 *    - The index corresponding to the maximal value of the resulting
 *      weighted sum of histograms is the threshold extracted.
 *
 * Input:::
 *  pIPWState::         State structure
 *  prSrc::             Image on which to compute threshold
 *  connectivity_index::"Connectivity index", about 10 seems good
 *  IsNew::             Switch between two methods, cFlase->old, cTrue->new
 *
 * Output:::
 *  threshComputed::    The threshold computed
 *
 * Returns:::
 *  Error status
 * Notes:::
 *
 * FunctionCategory:::
 *  threshold finding
 *
 * Keywords:::
 *  threshold finding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ccTextThreshold (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32               connectivity_index,
    Bool                IsNew,
    Int32              *threshComputed)
{
    PROLOGUE;
    Histo *histo = NULL;
    Int32  depth;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1, cFrame32);
    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));

        /*----- Computes differentiated components histogram -----*/
    histo = NULL;
    ERR_PASS (w_horizRunsHistogram(pIPWState, prSrc, cFalse, cFalse, &histo));

#ifdef DEBUG_CCTHRESH
	/* Plot 'em */
    fprintf(stderr, "Here's the cc histo...\n");
    plotPoints (pIPWState, HISTO_GET_DATA(histo), NULL, cPlotInt32, 
		HISTO_GET_SIZE (histo), cPlotLine);

#endif

        /*----- Computes optimal threshold -----*/
    ERR_PASS_CLEANUP (wp_thresholdFromHisto(pIPWState, histo, 
					   connectivity_index, 
					   0, 0, IsNew, threshComputed));

    if (depth == 4)
	*threshComputed = (*threshComputed + 8) / 16;

	/*----- Free the histogram -----*/
    ERR_PASS_CLEANUP (w_destroyHisto (pIPWState, &histo));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (histo)
	w_destroyHisto (pIPWState, &histo);

    ERR_RETURN;

} /* ccTextThreshold */


/* ===================================================================
**  ccBackgroundThreshold -- Find "background" threshold using connected
**			    components
**
**  ARGUMENTS:	prSrc	- Image to process
**              connectivity_index - (roughly) index of character
**                  connectivity in resulting thresholded image. Is
**                  generally equal to default value
**                  DEFAULT_CHARACTER_CONNECTIVITY.
**		farthestValue - the "white point", used to detect failure
**
**  OPERATION:
**		Same as ccTextThreshold, except that after the peak
**		is found, we walk along the curve until it drops off.
**		This usually works because the region on which we find
**		the peak is a plateau, and the rightmost region is where
**		the connected-component-ness changes suddenly, i.e. where
**		background starts coming through. We might want to back
**		off a bit from this value.
**
**  RETURNS:	error status
**
=================================================================== */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Find "background" threshold using connected components.
 *
 * LongDescription:::
 *   Find "background" threshold using connected components. Like
 *   "w_ccTextThreshold", except that we're trying to find not the
 *   best looking text, but the point before the background would
 *   start to come up.
 *
 * PrivateDescription:::
 *   Same as ccTextThreshold, except that after the peak
 *   is found, we walk along the curve until it drops off.
 *   This usually works because the region on which we find
 *   the peak is a plateau, and the rightmost region is where
 *   the connected-component-ness changes suddenly, i.e. where
 *   background starts coming through. We might want to back
 *   off a bit from this value.
 *
 * Input:::
 *  pIPWState::          State structure
 *  prSrc::              Image to analyze
 *  connectivity_index:: Connectivity index, usually around 10
 *  whitePoint::         The white point of the image, used to detect failure
 *
 * Output:::
 *  threshComputed::     Threshold computed
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  threshold finding
 *
 * Keywords:::
 *  threshold finding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod:: hacked out of w_ccTextThreshold
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ccBackgroundThreshold (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32               connectivity_index,
    Int32               whitePoint,
    Int32              *threshComputed)
{
    PROLOGUE;
    Histo *histo = NULL;
    Int32  depth;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4|cDepth8, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));

        /*----- Computes differentiated components histogram -----*/
    ERR_PASS (w_horizRunsHistogram(pIPWState, prSrc, cFalse, cFalse, &histo));

#ifdef DEBUG_CCTHRESH 
	/* Plot 'em */
    fprintf(stderr, "Here's the cc histo...\n");
    plotPoints (pIPWState, HISTO_GET_DATA(histo), NULL, cPlotInt32, HISTO_GET_SIZE (histo), cPlotLine);
#endif

        /*----- Computes optimal threshold -----*/
    ERR_PASS_CLEANUP (wp_thresholdFromHisto(pIPWState, histo, 
					   connectivity_index, 1, 
					   whitePoint, cFalse,threshComputed));

	/*----- Free the histogram -----*/
    ERR_PASS (w_destroyHisto (pIPWState, &histo));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (histo)
	w_destroyHisto (pIPWState, &histo);

    ERR_RETURN;
} /* ccBackgroundThreshold */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Pick the right threshold from a histogram of connected components
 *
 * LongDescription:::
 *  This routine takes a connected-component histogram, and from
 *  it computes either a "best for text" or "best for background"
 *  threshold.
 *
 * PrivateDescription:::
 * psm 11/94: added findBackground stuff. A connected-component
 * histogram gives us a fairly good hint as to what the background
 * threshold should be. Whereas the text threshold is at a fairly flat peak
 * of connected-component count, i.e. where the text forms characters but
 * before it starts connecting together into words, the background
 * threshold is where the connected component count falls as text characters
 * coalesce, before the cc count starts to climb again due to the background
 * starting to come through. This is easier in pages where there is
 * a simple white background--there is often a connected component
 * peak that is higher than the text peak. In this case, we find that
 * peak, then look for the valley in between. If that is not the case,
 * then we look for upward-sloping "bumps" on the downward-sloping 
 * connected-component curve. This is usually either the page coming up,
 * or in the case of thin originals, bleed-through from the back of the
 * page starting to show through. In either case, we want to threshold
 * before any significant bleed-through comes through. In the third case,
 * we don't actually have any upward-sloping stuff between the text point
 * and the white point (which is passed in). In that case, we look for
 * any little wiggles along the downward-sloping curves which have
 * a higher (more upward) slope than the rest of the curve. 
 *
 * Input:::
 *  pIPWState::         State structure
 *  histo::             Connected-component histogram
 *  connectivity_index::Connectivity index; 10 is a decent value. The 
 *                      connectivity index characterizes how "connected"
 *                      the resulting binary image should be: the
 *                      higher this parameter, the more the characters
 *                      will be unbroken, the lower this parameter,
 *                      the more emphasis is put on number of connected
 *                      components.
 *  findBackground::    If true, run an algorithm to
 *		        find a "background threshold" for the page. This
 *                      is a bit of a misnomer; the threshold attempts to be
 *		        the highest threshold below where significant amounts
 *                      of bleedthrough and background start to appear.If 
 *                      cTrue, search for a background threshold
 *  whitePoint::        If findBackground == cTrue, this should be
 *                      the alleged "paper" white, used by the background
 *                      finding code to know which method to use
 *  IsNew::             Switch between two methods, cFlase->old, cTrue->new
 *
 * Output:::
 *  result::            The resulting threshold
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   threshold finding
 *
 * Keywords:::
 *   threshold finding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv::
 *  macleod:: added background-finding code
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_thresholdFromHisto (
    IPWState           *pIPWState,
    Histo              *histo,
    Int32               connectivity_index,
    Int32               findBackground,
    Int32               whitePoint,
    Bool                IsNew,
    Int32              *result)
{
    PROLOGUE;
    register Int32 *hdata_smooth = NULL, *hdata_deriv = NULL, 
	*hdata_ggran = NULL;
    Int32           minval;
    register Int32  maxval;
    register Int32  i;
    Int32           index = 0;
    register Int32  maxdata;
    double          integ_smooth, integ_ggran;
    double          maxsum, newsum /*, delta */;
    Histo          *histo_smooth = NULL, *histo_deriv = NULL, 
		   *histo_ggran = NULL;
    Float64	   *weightedCurve = NULL;
    Int32           temp;
    Int32           average;

    Bool            potential_minimum = cFalse;
    register Int32 *hdata_dyn         = NULL;
    Histo          *histo_dyn         = NULL;
    register Int32  mindata, maxdata2;
    register Int32  j;
    register Int32  previous_val;

#ifdef DEBUG_CCTHRESH
    char            temp_string[128];
#endif

        /*----- Some initializations -----*/

    ERR_PASS (w_copyHisto(pIPWState, histo, &histo_smooth));

    minval = HISTO_GET_MINVAL(histo_smooth);
    maxval = HISTO_GET_MAXVAL(histo_smooth);

    hdata_smooth = HISTO_GET_DATA(histo_smooth);

       /*----- Recomputes max and min of histogram. This is so that if
	       the dynamic range of the image is smaller than
	       [MINPIX8U;MAXPIX8U], we do not get screwed up in
	       thinking that there are good thresholds to pick
	       that white-out or black-out the whole image -----*/

    while (hdata_smooth[minval] == 0 && minval < maxval)
	minval++;

    temp = hdata_smooth[maxval];
    while (hdata_smooth[maxval-1] == temp && maxval > minval)
	maxval--;

	/* BUG FIX: DDavies 9/15/94.  In PerfectScan testing, maxval was
	 * found to be 0 when given an all-black image.  In this
	 * degenerate case, just return a threshold of 0.
	 */
    if (maxval == 0)
    {
	*result = 0;
	goto done;
    }


    HISTO_SET_MINVAL(histo_smooth, minval);
    HISTO_SET_MAXVAL(histo_smooth, maxval);


        /*----- Needs to smooth this histogram a tiny bit. I would prefer
	        to avoid this, since it introduces a smoothing window
		size, but it does not seem to be possible easily. For
		right now, we simply do a moving average -----*/
    ERR_PASS_CLEANUP(
	w_smoothHisto(pIPWState, histo_smooth,
		      MAX(((WINDOW_SIZE * (maxval - minval + 1)) / maxval), 1),
		      &histo_smooth));

    /* We take this improved part. For gray image, it doesn't
       change too much, but it is good for some color image */
        /********** BEGIN NEW *********/
        /* NEW NEW NEW */
	/* 0. Recomputes minval and maxval based on the following:
	   in hdata_smooth, we need to make sure that we only consider
	   the parts of this histogram that are above 30% of the
	   average value of the histogram. */

    /* First compute average value of this histogram: */
    average = 0;
    for (i = minval; i <= maxval; i++)
	average += hdata_smooth[i];

    average /= (maxval - minval + 1);
    average = (average * 3) / 10;

    /* Now recomputes minval and maxval */
    for (i = minval; i <= maxval; i++) 
    {
	if (hdata_smooth[i] >= average)
	    break;
    }
    minval = i;
    for (i = maxval; i >= minval; i--) 
    {
	if (hdata_smooth[i] >= average)
	    break;
    }
    maxval = i;	

    HISTO_SET_MINVAL(histo_smooth, minval);
    HISTO_SET_MAXVAL(histo_smooth, maxval);

    /********** END NEW *********/

    /*----- Now computes the derivative of this histogram.
      Note that this computation is done in such a way that
      even binary images are correctly "thresholded"
      by routine (left-derivative)... -----*/

    ERR_PASS_CLEANUP (w_newHisto(pIPWState, minval, maxval, 1, &histo_deriv));

    hdata_deriv = HISTO_GET_DATA(histo_deriv);

    for (i = maxval; i > minval; i--) 
    {
	hdata_deriv[i] = hdata_smooth[i] - hdata_smooth[i-1];
    }

        /* This is added in order to correctly process
	   binary images. Logically, it should be hdata_smooth[0] */
    hdata_deriv[minval] = 0;


        /*----- Takes the absolute value of histo_deriv
	        supposed to be the derivative of the c.c. histogram -----*/

    for (i = minval; i <= maxval; i++)
	hdata_deriv[i] = ABS(hdata_deriv[i]);


        /*----- Computes the maximal value of this derivative
	        histogram and use it to invert it -----*/

    maxdata = 0;
    for (i = minval; i <= maxval; i++) 
    {
	if (hdata_deriv[i] > maxdata)
	    maxdata = hdata_deriv[i];
    }

    for (i = minval; i <= maxval; i++)
	hdata_deriv[i] = maxdata - hdata_deriv[i];

        /*----- Grayscale linear granulometry function (ggran)... -----*/

    ERR_PASS_CLEANUP (w_newHisto(pIPWState, minval, maxval, 1, &histo_ggran));

    hdata_ggran = HISTO_GET_DATA(histo_ggran);

    ERR_PASS_CLEANUP( i_grayLinearGranulFunctionLine32 (
        pIPWState, 
	hdata_deriv + minval,
	(Int16) (maxval - minval + 1),
	(Int16) cEuclidean, 
	hdata_ggran + minval));

#ifdef DEBUG_CCTHRESH
	/* Plot 'em */
    fprintf(stderr, "Here's the granulometry function...\n");
    plotPoints (pIPWState, hdata_ggran + minval, NULL, cPlotInt32, maxval - minval + 1,
		cPlotLine);

#endif

    if(IsNew == cTrue)
    {
        /**************************************************
	** New technique tried: simply extract the
	** maximum of the granulometry function *before* its
	** minimum of highest dynamics, computed in
	** geodesic border mode
	***************************************************/
	     /* Allocate new histogram */

	ERR_PASS_CLEANUP (w_newHisto(pIPWState, minval, maxval, 
					 1, &histo_dyn));
	hdata_dyn = HISTO_GET_DATA(histo_dyn);
	MEMSET((void *) (hdata_dyn + minval), 0,
	       (maxval - minval + 1) * sizeof(Int32));

	/* 1. Extract minimum value in hggran */

	mindata = hdata_ggran[minval];
	for (i = minval + 1; i <= maxval; i++) 
	{
	    if (hdata_ggran[i] < mindata)
		mindata = hdata_ggran[i];
	}

	/* 2. Now extracts the dynamics of each maximum.
	   Strictly speaking, this is not a real dynamics because
	   the case of neighboring minima with same depth is
	   treated slightly differently than for regular dynamics */

	potential_minimum = cFalse;
	previous_val      = mindata;

	for (i = minval; i <= maxval; i++) 
	{
	    if (hdata_ggran[i] > previous_val) 
	    {
		if (potential_minimum == cTrue) 
		{
		    /* Previous value was indeed a minimum. We need
		       to compute its dynamics */

		    for (maxdata = hdata_ggran[i], j = i + 1;
			 j <= maxval && hdata_ggran[j] >= previous_val; j++) 
		    {
			if (hdata_ggran[j] > maxdata)
			    maxdata = hdata_ggran[j];
		    }
		    for (maxdata2 = previous_val, j = i - 2;
			 j >= minval && hdata_ggran[j] >= previous_val; j--) 
		    {
			if (hdata_ggran[j] > maxdata2)
			    maxdata2 = hdata_ggran[j];
		    }
		    hdata_dyn[i-1] = (MIN(maxdata, maxdata2)) - previous_val;
		    potential_minimum = cFalse;
		}
	    }
	    else if (hdata_ggran[i] < previous_val)
		potential_minimum = cTrue;

	    previous_val = hdata_ggran[i];
	}

	    /* Takes care of the last index in histogram!!
	       Here, we assume that the value on the right is
	       +infinity, so the dynamics value is solely computed
	       by going left */

	if (potential_minimum == cTrue) 
	{
	    for (maxdata2 = previous_val, j = i - 2;
		 j >= minval && hdata_ggran[j] >= previous_val; j--) 
	    {
		if (hdata_ggran[j] > maxdata2)
		    maxdata2 = hdata_ggran[j];
	    }
	    hdata_dyn[i-1] = maxdata2 - previous_val;
	}
	   
	/* 3. Extract index of first minimum of highest dynamics */

	maxdata = -1;
	for (i = minval; i <= maxval; i++) 
	{
	    if (hdata_dyn[i] > maxdata) 
	    {
		maxdata = hdata_dyn[i];
		index = i;
	    }
	}

	    /* change maxval to be the new index found (minus 1)
	       since we will never look to the right of it */

/*	maxval = index - 1; */

	ERR_PASS_CLEANUP( w_destroyHisto(pIPWState, &histo_dyn));

	/* 4. Extract maximum of hdata_ggran before last
	      possible index (which is now maxval) */

	maxdata = 0;
	index = 128;
	for (i = minval; i <= maxval; i++) 
	{
	    if (hdata_ggran[i] > maxdata) 
	    {
		maxdata = hdata_ggran[i];
		index = i;
	    }
	}

	/* Now goes left before index until we get to maxdata - 1%,
	   goes right after index similarly, take the middle of these
	   two indices. */

	maxdata = (maxdata * 997) / 1000;
	j = index - 1;
	while (j >= minval && hdata_ggran[j] > maxdata)
	    j--;

	if (j >= minval) j++;

	i = index + 1;
	while (i <= maxval && hdata_ggran[i] > maxdata)
	    i++;

	if (i <= minval) i--;;

/*	*result = (j + j + i + i + i) / 5; */
	*result = (j + i) / 2; 

#ifdef DEBUG
	fprintf(stderr, "Threshold level: %d\n", *result);
#endif

#ifdef DEBUG
	{
	    char filename[40];
	    FILE *stream;

	    sprintf(filename, "ccsmooth.%d.histo", *result);
	    stream = fopen(filename, "w");
	    for (i = minval; i <= maxval; i++)
		fprintf(stream, "%ld\t%ld\n", i, hdata_smooth[i]);
	    fclose(stream);

	    sprintf(filename, "ggran.%d.histo", *result);
	    stream = fopen(filename, "w");
	    for (i = minval; i <= maxval; i++)
		fprintf(stream, "%ld\t%ld\n", i, hdata_ggran[i]);
	    fclose(stream);
	}
#endif
    }
    else
    {
        /*----- Compute integrals of histo_smooth and histo_ggran.
	        They are used as normalizing factors in the sum
		of histograms computed below. The maximal value
		of these histograms was also used as normalizing factor,
		but found to be less accurate, because more sensitive
		to noise in original image -----*/

	integ_smooth = 0.0;
	for (i = minval; i <= maxval; i++)
	    integ_smooth += ((double) hdata_smooth[i]);

	integ_ggran = 0.0;
	for (i = minval; i <= maxval; i++)
	    integ_ggran += ((double) hdata_ggran[i]);


        /*----- Now extract the index corresponding to the maximal
	        value of the weighed sum of these histograms -----*/
	weightedCurve = (Float64 *)IPW_CALLOC((maxval - minval + 1),  
					  sizeof (Float64), cAllocatorData);
	ERR_GEN_CLEANUP (!weightedCurve, cErrNomem);

	maxsum = -1.0; /* Why does the NT compiler generate a warning here? */

	for (i = minval; i <= maxval; i++) 
	{
	    newsum = ((double) hdata_ggran[i])  * integ_smooth * 
		    connectivity_index +
		    ((double) hdata_smooth[i]) * integ_ggran;
	    if (newsum > maxsum) 
	    {
		index = i;
		maxsum = newsum;
	    }
	    weightedCurve[i - minval] = newsum;
	}

#ifdef DEBUG_CCTHRESH 
	/* Plot 'em */
	fprintf(stderr, "max Index %ld\n", index);
	fprintf(stderr, "Here's the weighted curve...\n");
	plotPoints (pIPWState, weightedCurve, NULL, cPlotFloat64, 
		    maxval - minval + 1, cPlotLine);

#endif
	if (findBackground)
	{
	/* Try something different for background thresholding.
	 * The theory is that there is a "peak", found above in 
	 * "index", which is the peak of "high, flat regions", i.e.
	 * the text peak. Search over to find the absolute connected-
	 * component count peak, and then find the lowest point between
	 * the two. That, we theorize, would be a good background point,
	 * because it's after characters have started merging together but
	 * before the background starts coming through. 
	 * ---
	 * In retrospect, that worked some of the time, but there are
	 * cases where bleed-through is so low that it starts to contribute
	 * to the cc count as the text is merging together. One could
	 * argue (quite rightly) that there IS no correct threshold for
	 * such a page, but we must continue to support the fantasy
	 * until a more reasonable method is developed. See the "two cases"
	 * below.
         * ---
	 * There are now four cases...
	 */
	    Int32	absolutePeak, furthestValue, lowestIndex = 0, *hdata;

	/* Look at the real thing */
	    hdata = HISTO_GET_DATA(histo);

	/* Find the peak */
	    furthestValue = 0;
	    absolutePeak = 0;
	    for (i = index; i < maxval; i++)
	    {
		if (hdata[i] > furthestValue)
		{
		    absolutePeak = i;
		    furthestValue = hdata[i];
		}
	    }

	/*
	 * Case #1:
	 * there is a peak "somewhere out there" which is higher than
	 * the weighted peak--then just find the valley in between. If
	 * that's NOT the case, we have to do a more error-prone search.
	 */
	    if (absolutePeak - index >= MAX((maxval - minval) / 16, 1))
	    {
	    /* Find the lowest value in between the peaks */
		furthestValue = 0x7ffffff0;
		lowestIndex = -1;
		for (i = index; i < absolutePeak; i++)
		{
		    if (hdata[i] < furthestValue)
		    {
			furthestValue = hdata[i];
			lowestIndex = i;
		    }
		}
	    }
	    else
	    {
		Int32 *dx = NULL, *ddx = NULL;
		dx = (Int32 *) IPW_MALLOC (maxval * sizeof(Int32), 
					   cAllocatorData);
		ERR_GEN_CLEANUP (!dx, cErrNomem);
		ddx = (Int32 *) IPW_MALLOC (maxval * sizeof(Int32), 
					    cAllocatorData);
		ERR_CHECK (!ddx)
		{
		    IPW_FREE (dx, cAllocatorData);
		    ERR_GEN_CLEANUP (1, cErrNomem);
		}

	    /* Make smoothed 1st and 2nd derivatives */
		ERR_PASS_CLEANUP (w_numericalDiff (pIPWState, hdata, 
						   maxval, 1, 1, &dx));
		ERR_PASS_CLEANUP (w_numericalDiff (pIPWState, hdata, 
						   maxval, 2, 1, &ddx));

	    /*
	     * Case #2:
	     * Look along the 1st derivative until it's positive and upward
	     * sloping (2nd derivative positive). The idea is that we're
	     * looking for the point at which "stuff" starts coming up from
	     * either the background or the back of the paper. The 
	     * connected component count goes up because little dots start
	     * to appear before everything coalesces.
	     */

		for (i = index; i < maxval; i++)
		{
		    if (dx[i] > 0 && ddx[i] > 0)
		    {
			lowestIndex = (i == index) ? i : i - 1;
			break;
		    }
		}

	    /* Did case #2 fail? */
		if (lowestIndex >= whitePoint)
		{
		/* 
		 * Case #3: 
		 * Sort of the same as case #2, but since there was
		 * no actual upward slope between the text point and the
		 * white point, look for something more murky--
		 * look along the 2nd derivative until it goes upward.
		 * We're really clutching at straws here, basically looking
		 * for "kinks" along the curve, i.e. places where the derivative
		 * is relatively higher (hence the 2nd derivative is upward)
		 */

		    lowestIndex = -1;
		    for (i = index; i < maxval; i++)
		    {
			if (ddx[i] > ddx[i-1])
			{
			    lowestIndex = (i == index) ? i : i - 1;
			    break;
			}
		    }
		}
		
		IPW_FREE (dx, cAllocatorData);
		IPW_FREE (ddx, cAllocatorData);
	    }

	/* Well, did that fail? */
	    if (lowestIndex >= whitePoint)
	    {
	    /*
	     * Case #4:
	     * If all else fails, just go halfway between the white
	     * point and the text point. This doesn't happen very much
	     * (at least in the stuff I've been trying, which is 'documents').
	     * I'm just keeping this in as a "fallback" so that
	     * we don't produce black for the whole page.
	     */
	     
		lowestIndex = (index + whitePoint) / 2;
	    }
	    
	    if (lowestIndex != -1)
		index = lowestIndex;	
	}

	*result = index;
    }

done:
        /*----- Clean up -----*/
    if (weightedCurve) IPW_FREE (weightedCurve, cAllocatorData);
    if (histo_ggran) ERR_PASS_CLEANUP (w_destroyHisto(pIPWState, 
						      &histo_ggran));
    if (histo_deriv) ERR_PASS_CLEANUP (w_destroyHisto(pIPWState, 
						      &histo_deriv));
    if (histo_smooth) ERR_PASS_CLEANUP (w_destroyHisto(pIPWState, 
						       &histo_smooth));
    if (hdata_dyn) ERR_PASS_CLEANUP (w_destroyHisto(pIPWState, 
						       &histo_dyn));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (weightedCurve) IPW_FREE (weightedCurve, cAllocatorData);
    if (histo_ggran) w_destroyHisto(pIPWState, &histo_ggran);
    if (histo_deriv) w_destroyHisto(pIPWState, &histo_deriv);
    if (histo_smooth) w_destroyHisto(pIPWState, &histo_smooth);

    ERR_RETURN;
} /* threshold_from_histo */

/*
 * $Log: ccthresh.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 * Imported sources
 *
 * Revision 1.33  1998/01/14 00:10:53  ddavies
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.32  1997/05/10 00:22:16  xzhu
 * xzhu on Fri May 9 17:20:14 PDT 1997
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Put improved algorithm for macval and minval of histogram in wp_thresholdFromHisto(). It is almost no change for gray, but it is better for color images.
 *
 * Revision 1.31  1997/04/22 18:22:36  xzhu
 * xzhu on Tue Apr 22 11:20:19 PDT 1997
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Merged algorithm used in UNLV testing into wp_thresholdFromHisto() and added IsNew to select old algorithm or new one.
 *
 * Revision 1.30  1996/01/04 20:21:21  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.29  1995/12/13  04:05:17  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.28  1995/11/13  20:09:46  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.27  1995/11/03  22:48:09  ddavies
 * ddavies on Fri Nov 3 14:48:08 PST 1995
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.26  1995/10/13  21:58:50  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.25  1995/10/12  17:43:38  macleod
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * added explicit casts (float->int, long->short, long->char, etc) and
 * made floating constants the right size to shut up the NT compiler.
 * There's still a mysterious problem, where "foo = -1.0" generates
 * a warning about constant arithmetic overflow (foo is a double, and
 * so I'm confused)
 *
 * Revision 1.24  1995/10/09  19:04:30  macleod
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Attempt to eliminate some NT compiler whinings
 *
 * Revision 1.23  1995/10/05  19:13:49  ejaquith
 * ejaquith on Thu Oct 5 12:13:30 PDT 1995
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Added missing FunctionCategory.
 *
 * Revision 1.22  1995/10/03  01:59:19  macleod
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.21  1995/09/21  22:51:42  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.20  1995/09/11  21:54:14  macleod
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.19  1995/09/07  01:29:47  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.18  1995/09/06  20:17:28  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.17  1995/09/01  21:02:35  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.16  1995/08/31  21:11:16  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.15  1995/08/25  01:35:13  ejaquith
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Fixed warnings
 *
 * Revision 1.14  1995/08/17  21:53:57  macleod
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.13  1994/12/06  17:52:50  dferrell
 * dferrell on Tue Dec 6 09:51:52 PST 1994
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Added typecastes to MALLOC return values so MPW C compiler could
 * compile the file.
 *
 * Revision 1.12  1994/11/22  18:52:23  macleod
 * macleod on Tue Nov 22 10:45:48 PST 1994
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Added code for "background" thresholding. The idea is that if one
 * has a histogram of connected components by threshold, which is used to
 * compute the text threshold, one can also compute the background threshold
 * by looking down the curve from the text threshold towards the white
 * point. As the threshold increases, characters tend to merge together
 * into words, and the connected-component count falls, but as
 * background starts to come up, the connected-component count rises.
 * In some cases, the right point is obvious (an upward slope in the curve)
 * but if such a point isn't found, we look for "upward wiggles" not
 * necessary upward sloping but of a higher relative slope.
 * In initial testing (about 50 images) this code gives "better"
 * thresholds than the previous histogram-based method, although
 * the segmentation results may be worse, due to the fact that the segmenter
 * was able to deal with a bit of bleed-through in some cases, because
 * of gutter and rule finding. This will require more work; it may make
 * more sense to somehow detect the gutters and fill in the box they surround,
 * thereby making images with rectangular gutters work as well as they
 * did before, but not screwing up as many images with irregularly-shaped
 * masks.
 *
 * Revision 1.11  1994/10/24  19:39:23  macleod
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Scale down threshold when doing 4-bit images.
 *
 * Revision 1.10  1994/10/07  19:06:37  lperry
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.9  1994/09/29  00:45:22  lperry
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Eliminate use of utilpros. Use hutils.pub instead.
 * Also use utils.pub.
 *
 * Revision 1.8  1994/09/20  22:19:59  macleod
 * macleod on Tue Sep 20 15:18:39 PDT 1994
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * thresholdFromHisto should now free the histograms it allocates
 * even when maxval == 0 or errors occur.
 *
 * Revision 1.7  1994/09/15  17:26:21  ddavies
 * ddavies on Thu Sep 15 10:25:26 PDT 1994
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Ouch.  thresholdFromHisto now returns with a threshold of 0 if maxval == 0,
 * instead of when maxval != 0.  Sorry, Macduff.  whimper.
 *
 * Revision 1.6  1994/09/12  20:34:43  ddavies
 * ddavies on Mon Sep 12 13:32:25 PDT 1994
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Change thresholdFromHisto so that it returns a threshold of 0 if the
 * histogram is empty or the maximum bin in the histogram containing non-zero
 * counts is bin 0.  This fixes a divide-by-zero error found by Macduff.
 * It occurs when trying to smooth the histogram and we calculate the
 * window width by taking (maxval - minval)/maxval where maxval = 0.
 *
 * Revision 1.5  1994/08/01  21:24:30  dferrell
 * dferrell on Mon Aug 1 14:23:16 PDT 1994
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * Added typecaste to return value from CALLOC so that MPW C can compile
 * this file.
 *
 * Revision 1.4  1994/07/28  01:41:52  macleod
 * macleod on Wed Jul 27 18:39:43 PDT 1994
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * added ccBackgroundThreshold, which is like ccTextThreshold only it
 * walks in an increasing direction along the plateau on which the
 * text peak is found until the plateau drops off. The theory is that
 * this is where the background starts coming through. Modified
 * thresholdFromHisto to keep its weighted curve, and to take an
 * option for walking over to the edge of the plateau
 *
 * Revision 1.3  1994/06/16  22:35:21  macleod
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * ccTextThreshold and thresholdFromHisto can now handle 4 bit/pixel images
 *
 * Revision 1.2  1994/06/08  23:59:22  macleod
 * macleod on Wed Jun 8 16:58:51 PDT 1994
 *
 * ccthresh.c -> /product/ipcore/ipwrap/src/RCS/ccthresh.c,v
 *
 * There was an "unused variable" warning.
 *
 * Revision 1.1  1994/06/08  20:30:48  macleod
 * Initial revision
 * 
 */
