/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  clean.c--includes
 *
 *		Drycleaning:
 *			ErrCode	w_cleanPepper3()
 *			ErrCode	w_cleanStipple6 ()
 *			ErrCode	w_cleanStipple ()
 *			ErrCode	w_findStipple ()
 *			ErrCode	w_createStippleMask ()
 *			ErrCode	w_removeRules ()
 *			ErrCode	w_removeHVRules ()
 *
 *		Private:
 *			Int32	wp_rr_filterRuleCCs ()
 */



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "logicops.h"
#include "morphg.h"
#include "shrscgr.pub"
#include "frames.pub"
#include "defines.pub"
#include "alplib.h"
#include "shrcc.pub"

/* prototypes */
#include "clean.pub"
#include "props.pub"
#include "rast.pub"
#include "logicops.pub"
#include "halftone.pub"
#include "rectops.pub"
#include "morph.pub"
#include "expand.pub"
#include "reduce.pub"
#include "strel.pub"
#include "intops.pub"
#include "clean.prv"
#include "subsmpl.pub"
#include "runfilt.pub"
#include "fill.pub"
#include "shrcc.pub"

#define DEBUG_STIPPLE 0


#if DEBUG_STIPPLE
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: clean.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");
#define  DEFAULT_PEPPER_LEVEL      2


/****************************************************************
 *************             Drycleaning             **************
 ***************************************************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Pepper removal
 *
 * LongDescription:::
 *  Removes pepper noise from an image
 *
 * PrivateDescription:::
 * action:: removes pepper noise from image at 4x reduction.
 * (1) reduce with threshold(1,1) and close to
 *     consolidate graphics and text regions;
 * (2) pepper removal for small isolated pepper;
 * (3) large dilate to fully consolidate remaining regions;
 * (4) union of large openings removes large isolated pepper;
 * (5) expand and mask with original;
 * (6) final pepper removal at full scale.
 *
 * Input:::
 *  pIPWState::      State structure.
 *  prSrc::          The original image.
 *  option::         An option flag, which appears to be useless.
 *
 * Output:::
 *  pPrDest::        Image with pepper removed.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 * 
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cleanPepper3 (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32               option,
    PIXR              **pPrDest)
{
    PROLOGUE;
    STREL        *strel_close, *strel_dilate;
    STREL        *strel_open_h, *strel_open_v;
    STREL        *pStrel4 = NULL, *pStrel8h = NULL, *pStrel8v = NULL; 
    PIXR         *prt1 = NULL, *prt2 = NULL, *prt3 = NULL, *prDest;
    Int32         rightColsToClear, bottomRowsToClear;
    Bool	  allocatedHere = cFalse;
    option = option;	/* keep the compiler happy */

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_PASS (wp_clonePixrConditionalNoInit (pIPWState, prSrc, pPrDest,
					    &rightColsToClear, 
					    &bottomRowsToClear,
					    &allocatedHere));
    prDest = *pPrDest;
    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_initStrelByName(pIPWState, strel4));
    ERR_PASS (w_initStrelByName(pIPWState, strel8h));
    ERR_PASS (w_initStrelByName(pIPWState, strel8v));
    
    prt1 = prt2 = prt3 = NULL;

    ERR_PASS (w_getStrel (pIPWState, strel4, &pStrel4));
    ERR_PASS (w_getStrel (pIPWState, strel8h, &pStrel8h));
    ERR_PASS (w_getStrel (pIPWState, strel8v, &pStrel8v));

    strel_close = pStrel4;
    strel_dilate = pStrel4;
    strel_open_h = pStrel8h;
    strel_open_v = pStrel8v;

	/* copy to prDest if necessary; further work is on prDest */
    if (prSrc != prDest)
	ERR_PASS (w_logicOp (pIPWState, prSrc, COPY, prDest));

    ERR_PASS_CLEANUP (w_reduce4 (pIPWState, prDest, 1, 1, &prt1));
    ERR_PASS_CLEANUP (w_close (pIPWState, prt1, strel_close, &prt1));
    ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prt1, &prt2));

    ERR_PASS_CLEANUP (w_removePepper (pIPWState, prt1, 
				      DEFAULT_PEPPER_LEVEL, &prt2));

	/* dilate mask */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prt2, strel_dilate, &prt1));

	/* remove large pepper with union of openings */
    ERR_PASS_CLEANUP (w_open (pIPWState, prt1, strel_open_h, &prt2));
    ERR_PASS_CLEANUP (w_open (pIPWState, prt1, strel_open_v, &prt3));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prt2, OR, prt3));

	/* Expand back to full-size and mask in dest. image */
    ERR_PASS_CLEANUP (w_expandAndMask(pIPWState, prDest, prt3, &prDest));

	/* Take out pepper at full-scale */
    ERR_PASS_CLEANUP (w_removePepper (pIPWState, prDest, DEFAULT_PEPPER_LEVEL,
				      &prDest));


	/* If we need to clear any parts of the frame, do so */
    if ((rightColsToClear | bottomRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prDest,
						     rightColsToClear,
						     bottomRowsToClear));

    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prt1));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prt2));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prt3));
    
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prt1) w_destroyPixr(pIPWState, &prt1);
    if (prt2) w_destroyPixr(pIPWState, &prt2);
    if (prt3) w_destroyPixr(pIPWState, &prt3);
    if (allocatedHere) w_destroyPixr (pIPWState, pPrDest);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Pepper removal within a mask
 *
 * LongDescription:::
 *  Like cleanPepper3, but also takes a mask, which protects regions in
 *  the original in which pepper should not be removed
 *
 * PrivateDescription:::
 * action:: removes pepper noise from image at 4x reduction.
 * (1) reduce with threshold(1,1) and close to
 *     consolidate graphics and text regions
 * (2) pepper removal for small isolated pepper
 * (3) large dilate to fully consolidate remaining regions
 * (4) union of large openings removes large isolated pepper
 * (5) expand and mask with original
 * (6) final pepper removal at full scale
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Original image
 *  prMask::         Mask image, where 1's allow pepper removal
 *  option::         Useless parameter which should be removed
 *
 * Output:::
 *  pPrDest::        Image with pepper removed in relevant areas
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cleanPepper3Masked (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    PIXR               *prMask,
    Int32               option,
    PIXR              **pPrDest)
{
    PROLOGUE;
    STREL        *strel_close, *strel_dilate;
    STREL        *strel_open_h, *strel_open_v;
    STREL        *pStrel4 = NULL, *pStrel8h = NULL, *pStrel8v = NULL; 
    PIXR         *prt1 = NULL, *prt2 = NULL, *prt3 = NULL, *prt4 = NULL, 
	         *prt5 = NULL, *prDest;
    Int32         rightColsToClear, bottomRowsToClear;
    Bool	  allocatedHere = cFalse;
    
    option = option;	/* keep the compiler happy */
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_PASS (wp_clonePixrConditionalNoInit (pIPWState, prSrc, pPrDest,
					    &rightColsToClear, 
					    &bottomRowsToClear,
					    &allocatedHere));
    prDest = *pPrDest;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR (prMask, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_initStrelByName(pIPWState, strel4));
    ERR_PASS (w_initStrelByName(pIPWState, strel8h));
    ERR_PASS (w_initStrelByName(pIPWState, strel8v));

    ERR_PASS (w_getStrel (pIPWState, strel4, &pStrel4));
    ERR_PASS (w_getStrel (pIPWState, strel8h, &pStrel8h));
    ERR_PASS (w_getStrel (pIPWState, strel8v, &pStrel8v));

    strel_close = pStrel4;
    strel_dilate = pStrel4;
    strel_open_h = pStrel8h;
    strel_open_v = pStrel8v;

	/* copy to prDest if necessary; further work is on prDest */
    if (prSrc != prDest)
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prDest));

	/* Cut out the protected stuff from the original */
    ERR_PASS_CLEANUP (w_copy(pIPWState, prSrc, &prt5));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prMask, SUBTRACT, prt5));

    ERR_PASS_CLEANUP (w_reduce4 (pIPWState, prDest, 1, 1, &prt1));
    ERR_PASS_CLEANUP (w_close (pIPWState, prt1, strel_close, &prt1));

    ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prt1, &prt2));

    ERR_PASS_CLEANUP (w_removePepper (pIPWState, prt1, 
				      DEFAULT_PEPPER_LEVEL, &prt2));

	/* dilate mask */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prt2, strel_dilate, &prt1));

	/* remove large pepper with union of openings */
    ERR_PASS_CLEANUP (w_open (pIPWState, prt1, strel_open_h, &prt2));
    ERR_PASS_CLEANUP (w_open (pIPWState, prt1, strel_open_v, &prt3));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prt2, OR, prt3));

	/* expand to f.s. and intersect with source */
    ERR_PASS_CLEANUP (w_copy(pIPWState, prDest, &prt4));
    ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, prt4, prt3, &prt4));

    ERR_PASS_CLEANUP (w_removePepper (pIPWState, prt4, 
				   DEFAULT_PEPPER_LEVEL, &prDest));

    /* Mask out the protected stuff */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prMask, INTERSECTION, prDest));

    /* Put back in the original protected stuff */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prt5, UNION, prDest));
    
	/* If we need to clear any parts of the frame, do so */
    if ((rightColsToClear | bottomRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prDest,
						     rightColsToClear,
						     bottomRowsToClear));

    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prt1)); 
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prt2)); 
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prt3));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prt4));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prt5));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prt1) w_destroyPixr(pIPWState, &prt1);
    if (prt2) w_destroyPixr(pIPWState, &prt2);
    if (prt3) w_destroyPixr(pIPWState, &prt3);
    if (prt4) w_destroyPixr(pIPWState, &prt4);
    if (prt5) w_destroyPixr(pIPWState, &prt5);
    if (allocatedHere) w_destroyPixr (pIPWState, pPrDest);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Remove stipple (large, regular stuff like halftones) in an image
 *
 * LongDescription:::
 *  This routine attempts to remove stipple, which is large regular
 *  stuff (larger than "pepper" anyway) usually behind text. It will
 *  indiscriminately remove stipple out of halftoned pictures, so
 *  one should probably have segmented the page first.
 *
 * PrivateDescription:::
 * One of Dan Bloomberg's scripts just found stippled regions,
 * made some boxes, and then did an open(..., strel4) on the boxes 
 * to get rid of the stipple. This worked better than the fancy
 * attempt to match a stipple STREL. Anyway, this routine does
 * approximately that, but uses masks rather than boxes.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Original image
 *
 * Output:::
 *  pPrDest::        Image with stipple removed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 * 
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod:: method from a script of Bloomberg's
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cleanStipple6 (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    PIXR              **pPrDest)
{
    PROLOGUE;
    Int32 w, h;
    PIXR    *prReduced = NULL, *stippleMask = NULL, *prStippledRegion = NULL, 
	    *prDest = NULL;
    STREL   *pStrel2 = NULL, *pStrel4 = NULL, *pStrel8h = NULL, 
	    *pStrel8v = NULL, *pStrel10h = NULL, *pStrel10v = NULL;
    Int32         rightColsToClear, bottomRowsToClear;
    Bool	  allocatedHere = cFalse;
    ScaledInt1616 xDPI, yDPI;


    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_initStrelByName(pIPWState, strel2));
    ERR_PASS (w_initStrelByName(pIPWState, strel4));
    ERR_PASS (w_initStrelByName(pIPWState, strel8h));
    ERR_PASS (w_initStrelByName(pIPWState, strel8v));
    ERR_PASS (w_initStrelByName(pIPWState, strel10h));
    ERR_PASS (w_initStrelByName(pIPWState, strel10v));

    ERR_PASS (w_getStrel (pIPWState, strel2, &pStrel2));
    ERR_PASS (w_getStrel (pIPWState, strel4, &pStrel4));
    ERR_PASS (w_getStrel (pIPWState, strel8h, &pStrel8h));
    ERR_PASS (w_getStrel (pIPWState, strel8v, &pStrel8v));
    ERR_PASS (w_getStrel (pIPWState, strel10h, &pStrel10h));
    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel10v));

    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &h));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    
    ERR_PASS (wp_clonePixrConditionalNoInit (pIPWState, prSrc, pPrDest,
					    &rightColsToClear, 
					    &bottomRowsToClear,
					    &allocatedHere));
    ERR_GEN_PIXR (*pPrDest, cByteOrderBig, cDepth1, cChannels1, 
		  cFrame32);
    prDest = *pPrDest;

    /* Reduce it by the "right" amount */

    /* Reduce it by 4 */
    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 1, 0, 0, &prReduced));

    /* Turn the stippled regions into masks */
    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel2, &prReduced));
    ERR_PASS_CLEANUP (w_open (pIPWState, prReduced, pStrel10v, &prReduced));
    ERR_PASS_CLEANUP (w_open (pIPWState, prReduced, pStrel10v, &prReduced));

    /* Dilate the mask a bit to get some extra slop around the edge */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prReduced, pStrel8h, &prReduced));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prReduced, pStrel8v, &prReduced));


    /* Blow the mask back up */
    ERR_PASS_CLEANUP (w_expand (pIPWState, prReduced, 4, &stippleMask));

    /* Make a pixr for the stippled parts */
    ERR_PASS_CLEANUP (w_createPixrNoInit(pIPWState, w, h, 1, xDPI, yDPI,
					 cFrameBits, &prStippledRegion));
    /* Cut out the stuff that we want to open from the mask */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prStippledRegion));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, stippleMask, INTERSECTION, 
				 prStippledRegion));

    /* Mask out the stippled region in the original */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prDest));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, stippleMask, SUBTRACT, prDest));

    /* Get rid of the stipples */
    ERR_PASS_CLEANUP (w_open (pIPWState, prStippledRegion, 
			      pStrel4, &prStippledRegion));

    /* Stick the de-stippled stuff back in to the result */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prStippledRegion, UNION, 
				 prDest));

	/* If we need to clear any parts of the frame, do so */
    if ((rightColsToClear | bottomRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prDest,
						     rightColsToClear,
						     bottomRowsToClear));

    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prReduced));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prStippledRegion));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &stippleMask));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prReduced)
	w_destroyPixr(pIPWState, &prReduced);
    if (prStippledRegion) 
	w_destroyPixr(pIPWState, &prStippledRegion);
    if (stippleMask) 
	w_destroyPixr(pIPWState, &stippleMask);
    if (allocatedHere)
	w_destroyPixr (pIPWState, pPrDest);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Remove stipple (large, regular stuff like halftones) in an image
 *
 * LongDescription:::
 *  This routine attempts to remove stipple, which is large regular
 *  stuff (larger than "pepper" anyway) usually behind text. It will
 *  indiscriminately remove stipple out of halftoned pictures, so
 *  one should probably have segmented the page first.
 *
 * PrivateDescription:::
 *  One of Dan Bloomberg's scripts just found stippled regions,
 *  made some boxes, and then did an open(..., strel4) on the boxes 
 *  to get rid of the stipple. This worked better than the fancy
 *  attempt to match a stipple STREL. Anyway, this routine does
 *  approximately that, but uses masks rather than boxes.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Original image
 *  reduction::      Reduction factor. At 300dpi or greater, reduce by 4. 
 *                   At 200 dpi, reduce by 2. At < 200 dpi, reduce by 1.
 *
 * Output:::
 *  pPrDest::        Image with stipple removed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  The reduction factor parameter should probably be removed, and
 *  "cleanStipple6" should be removed in favor of this routine.
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod:: method from a script of Bloomberg's
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cleanStipple (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32               reduction,
    PIXR              **pPrDest)
{
    PROLOGUE;
    Int32 w, h;
    PIXR    *prReduced = NULL, *prStippledRegion = NULL, *prDest = NULL;
    STREL   *stippleStrel;
    STREL   *pStrel2 = NULL, *pStrel4 = NULL, *pStrel8h = NULL, 
	    *pStrel8v = NULL, *pStrel10h = NULL, *pStrel10v = NULL,

	    *pStrel3 = NULL;
    Int32    rightColsToClear, bottomRowsToClear;
    Bool     allocatedHere = cFalse;
    ScaledInt1616 xDPI, yDPI;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_initStrelByName(pIPWState, strel2));
    ERR_PASS (w_initStrelByName(pIPWState, strel3));
    ERR_PASS (w_initStrelByName(pIPWState, strel4));
    ERR_PASS (w_initStrelByName(pIPWState, strel8h));
    ERR_PASS (w_initStrelByName(pIPWState, strel8v));
    ERR_PASS (w_initStrelByName(pIPWState, strel10h));
    ERR_PASS (w_initStrelByName(pIPWState, strel10v));

    ERR_PASS (w_getStrel (pIPWState, strel2, &pStrel2));
    ERR_PASS (w_getStrel (pIPWState, strel3, &pStrel3));
    ERR_PASS (w_getStrel (pIPWState, strel4, &pStrel4));
    ERR_PASS (w_getStrel (pIPWState, strel8h, &pStrel8h));
    ERR_PASS (w_getStrel (pIPWState, strel8v, &pStrel8v));
    ERR_PASS (w_getStrel (pIPWState, strel10h, &pStrel10h));
    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel10v));

    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &h));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    ERR_PASS (wp_clonePixrConditionalNoInit (pIPWState, prSrc, pPrDest,
					    &rightColsToClear, 
					    &bottomRowsToClear,
					    &allocatedHere));
    ERR_GEN_PIXR (*pPrDest, cByteOrderBig, cDepth1, cChannels1, 
		  cFrame32);
    prDest = *pPrDest;

    switch (reduction)
    {
	case 1:
	    ERR_PASS_CLEANUP (w_copy (pIPWState, prSrc, &prReduced));
	    stippleStrel = pStrel2;
	    break;
	case 2:
	    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 0, 0, 0, 
					  &prReduced));
	    stippleStrel = pStrel3;
	    break;
	case 4:
	    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 1, 0, 0, 
					  &prReduced));
	    stippleStrel = pStrel4;
	    break;
	default:
	    ERR_GEN_CLEANUP (1, cErrApps_badReduction);
	    break;

    }

    /* Turn the stippled regions into masks */
    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel2, &prReduced));
    ERR_PASS_CLEANUP (w_open (pIPWState, prReduced, pStrel10v, &prReduced));
    ERR_PASS_CLEANUP (w_open (pIPWState, prReduced, pStrel10h, &prReduced));

    /* Dilate the mask a bit to get some extra slop around the edge */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prReduced, pStrel8h, &prReduced));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prReduced, pStrel8v, &prReduced));

    /* Copy the image to the result */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prDest));

    /* Make a pixr for the stippled parts */
    ERR_PASS_CLEANUP (w_createPixrNoInit(pIPWState, w, h, 1, xDPI, yDPI,
					 cFrameBits, &prStippledRegion));

    /* Get rid of stippled stuff, copying into "prStippledRegion" */
    ERR_PASS_CLEANUP (w_open (pIPWState, prSrc, 
			      stippleStrel, &prStippledRegion));

    /* Mask out the non-stippled regions */
    if (reduction == 1)
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReduced, AND, 
				     prStippledRegion));
    else
	ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, 
					   prStippledRegion, 
					   prReduced,
					   &prStippledRegion));

    /* Mask only the non-stippled regions in the result */
    ERR_PASS_CLEANUP (w_invert (pIPWState, prReduced, &prReduced));
    if (reduction == 1)
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReduced, AND, prDest));
    else
	ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, prDest,
					   prReduced, &prDest));

    /* Combine them */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prStippledRegion, UNION, 
				 prDest));

	/* If we need to clear any parts of the frame, do so */
    if ((rightColsToClear | bottomRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prDest,
						     rightColsToClear,
						     bottomRowsToClear));

    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prReduced));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prStippledRegion));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prReduced)
	w_destroyPixr (pIPWState, &prReduced);
    if (prStippledRegion)
	w_destroyPixr (pIPWState, &prStippledRegion);
    if (allocatedHere)
	w_destroyPixr (pIPWState, &prDest);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return stipple (large, regular stuff like halftones) in an image
 *
 * LongDescription:::
 *  This routine attempts to identify stipple and other high frequency
 *  stuff.  This is done primarily because you want to compress this
 *  stuff using a different method than would be used for text, line art
 *  or big blobs.  It will indiscriminately remove dots from halftoned
 *  pictures, so one should probably have segmented the page first.
 *  This routine does try to leave text basically alone.
 *
 * PrivateDescription:::
 *  This is a new routine.  The steps are:
 * 1. Use halftoneThresh to identify regions that contain
 *    stipple and large blobs and possibly text on stipple.
 * 2. If there are no such blobs, return with dest ptr NULL.
 * 3. AND the original image with this mask to get an image
 *    that contains only stipple/blobs/text-on-stipple.
 * 4. Use w_cleanStipple to find all the non-stipple in these
 *    regions.  This has almost certainly eaten away at the text
 *    if there was text on stipple.
 * 5. Seed fill this back into the image/stipple mask. This
 *    will completely reconstruct the characters, but add
 *    little globs of stipple onto their edges.
 * 6. Open that image to remove the hair.
 * 7. We now have an image that has all the text and big
 *    blobs that were in the original image.  Subtract this
 *    from the image that holds the image/stipple/text-on-stipple
 *    stuff to get just the stipple.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Original image
 *
 * Output:::
 *  pFoundStipple::  Ptr to boolean that's set to true if some stipple
 *		     was found.  If stipple was found *pPrDest has it.
 *		     If no stipple was found and *pPrDest was initially
 *		     NULL, *pPrDest is returned NULL.  If no stipple
 *		     is found and *pPrDest is initially not NULL, the
 *		     contents of *pPrDest are changed in an undefined way.
 *  pPrDest::        Image containing only stipple.
 *
 * Returns:::
 *  error status
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  DDavies
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findStipple (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Bool	       *pFoundStipple,
    PIXR              **pPrDest)
{
    PROLOGUE;
    PIXR    *prMask = NULL, *prDest = NULL;
    STREL   *pStrel2 = NULL, *pStrel3 = NULL;
    Int32    dpi;
    RECTANGL *rectList = NULL;
    Int32    rightColsToClear, bottomRowsToClear;
    Bool     allocatedHere = cFalse;
    ScaledInt1616 xDPI, yDPI;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_initStrelByName(pIPWState, strel2));
    ERR_PASS (w_initStrelByName(pIPWState, strel3));

    ERR_PASS (w_getStrel (pIPWState, strel2, &pStrel2));
    ERR_PASS (w_getStrel (pIPWState, strel3, &pStrel3));

	/* make a copy of the source in the dest.  If there was no
	 * dest, create one and remember it.
	 */
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prSrc, pPrDest,
					   &rightColsToClear, 
					   &bottomRowsToClear,
					   &allocatedHere));
    ERR_GEN_PIXR (*pPrDest, cByteOrderBig, cDepth1, cChannels1, 
		  cFrame32);
    prDest = *pPrDest;
    ERR_PASS_CLEANUP( w_copy(pIPWState, prSrc, &prDest));

	/* Assume we find no stipple */
    *pFoundStipple = cFalse;

	/* halftoneThresh likes to know lots of stuff based on
	 * the incoming picture's resolution.
	 */
    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32 (xDPI),
	       ScaledInt1616_TO_Int32 (yDPI));
    
    if (dpi >= 300)
    {
	ERR_PASS_CLEANUP (w_halftoneThresh(pIPWState, prSrc,
					   4, 8, 1, 99, 15,
					   &rectList, &prMask));
    }
    else if (dpi >= 200)
    {
	ERR_PASS_CLEANUP (w_halftoneThresh(pIPWState, prSrc,
					   2, 4, 1, 66, 10,
					   &rectList, &prMask));
    }
    else
    {
	ERR_PASS_CLEANUP (w_halftoneThresh(pIPWState, prSrc,
					   1, 2, 1, 33, 5,
					   &rectList, &prMask));
    }

	/* We don't need the rect list (maybe later?) */
    ERR_PASS_CLEANUP( w_destroyRectList(pIPWState, &rectList));

	/* if there's no image, quit now */
    if (prMask == NULL)
	goto ERR_CLEANUP;	/* where things are tidied up */

	/* make an image that contains only the image sections. */
    ERR_PASS_CLEANUP( w_logicOp(pIPWState, prMask, INTERSECTION, prDest));

	/* get rid of prMask before the next step */
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prMask));

	/* create a mask of all the non-stipple stuff */
    if (dpi >= 300)
    {
	ERR_PASS_CLEANUP (w_cleanStipple(pIPWState, prDest, 4, &prMask));
    }
    else if (dpi >= 200)
    {
	ERR_PASS_CLEANUP (w_cleanStipple(pIPWState, prDest, 2, &prMask));
    }
    else
    {
	ERR_PASS_CLEANUP (w_cleanStipple(pIPWState, prDest, 1, &prMask));
    }

	/* if there's no stipple, quit now */
    if (prMask == NULL)
	goto ERR_CLEANUP;	/* where things are tidied up */

	/* fill out the characters again. */ 
    ERR_PASS_CLEANUP( w_fillClipSweep8(pIPWState, prMask, prDest, 3,
				       &prMask));

	/* gnaw off the little hairs where stipple grew on the
	 * edge of the characters.
	 */
    if (dpi >= 400)
    {
	ERR_PASS_CLEANUP (w_open(pIPWState, prMask, pStrel3, &prMask));
    }
    else
    {
	ERR_PASS_CLEANUP (w_open(pIPWState, prMask, pStrel2, &prMask));
    }

	/* This has all the text and big blobs.  Subtract it from
	 * the areas that have possible stipple to get only the
	 * stipple.
	 */
    ERR_PASS_CLEANUP( w_logicOp(pIPWState, prMask, SUBTRACT, prDest));

	/* We're done with prMask */
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prMask));

	/* If we need to clear any parts of the frame, do so */
    if ((rightColsToClear | bottomRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prDest,
						     rightColsToClear,
						     bottomRowsToClear));

    *pFoundStipple = cTrue;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prMask)
	w_destroyPixr (pIPWState, &prMask);
    if (allocatedHere)
	w_destroyPixr (pIPWState, pPrDest);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Makes a reduced mask for removing stipple from a PIXR.
 *
 * LongDescription:::
 *  Since cleanStipplePixr tends to damage certain characters
 *  which have thin strokes (most notably, Kanji and Chinese
 *  characters with thin horizontal strokes) a new algorithm
 *  which preserves these characters has been implemented.
 *  The algorithm also attempts to preserve "text on tint"	
 *  which has the unfortunate side effect of also preserving
 *  dense halftoned/dithered regions.  Higher resolution (300+
 *  dpi) scans work best and resolutions less than 200 dpi in
 *  either dimension are rejected as invalid.
 *<BR>
 *  Makes a reduced mask for removing stipple from a PIXR.
 *  The mask has 1's where there is no stipple and is suitable
 *  for use with expandAndMask to remove stipple from the
 *  input image.
 * <BR>
 *  On exit, *pPrOutMask contains the final stipple mask which is
 *  suitable for use with expandAndMask to remove stipple
 *  from the original.  Stippled regions are 0 in the mask
 *  and non-stipple regions are 1.
 *  <BR>
 *  If no stipple is detected, *pPrOutMask will be NULL
 *
 * PrivateDescription:::
 *  The basis of the algorithm is the use of local region
 *  densities to detect halftone regions.  A mask of regions
 *  which may contain halftone is made using thresholded
 *  reduction to ~100 dpi.  This mask is processed 
 *  morphologically to remove small regions and rules leaving 
 *  only "large" stippled regions.  A second thresholded reduction 
 *  is performed at a higher threshold to produce an image
 *  of dense objects which should not be considered stipple.
 *  Small pepper is removed from the "unmask" and then the
 *  "unmask" is subtracted from the mask.  This result is 
 *  inverted and returned.
 *<BR>
 *  To improve performance on images which contain no stipple
 *  the algorithm checks for an empty result at two places
 *  and immediately returns with a NULL outMaskPr.
 *
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image in which to find stipple
 *
 * Output:::
 *  pPrOutMask::     Mask of where we find stipple
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 * Known Bugs::
 *	- The algorithm does not remove dense regions of halftone/
 *	dither/stipple.
 *	- The algorithm does not always remove small, isolated
 *	regions of halftone/dither/stipple.
 *	- The algorithm does not completely preserve text on tint.
 *	At 300 dpi and up the results are slightly worse.
 * Null Images::
 *  The no stipple detected result is NOT considered an
 *  error condition, and the function will return cSuccessful
 *  for this case.
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createStippleMask (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    PIXR              **pPrOutMask)
{
    PROLOGUE;
    PIXR    *prMask = NULL, *prTmp = NULL,
            *prUnmask = NULL;
    Bool     isZero;
    STREL   *pStrel2v = NULL, *pStrel5h = NULL, *pStrel19v = NULL,
	    *pS_pepper_1 = NULL;
    ScaledInt1616 xDPI, yDPI;
    Int32    xres, yres;

    ERR_GEN (*pPrOutMask != NULL, cErrPtrNotNull);
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    xres = ScaledInt1616_TO_Int32 (xDPI);
    yres = ScaledInt1616_TO_Int32 (yDPI);

    ERR_PASS (w_initStrelByName(pIPWState, strel2v));
    ERR_PASS (w_initStrelByName(pIPWState, strel5h));
    ERR_PASS (w_initStrelByName(pIPWState, strel19v));
    ERR_PASS (w_initStrelByName(pIPWState, s_pepper_1));

    ERR_PASS (w_getStrel (pIPWState, strel2v, &pStrel2v));
    ERR_PASS (w_getStrel (pIPWState, strel5h, &pStrel5h));
    ERR_PASS (w_getStrel (pIPWState, strel19v, &pStrel19v));
    ERR_PASS (w_getStrel (pIPWState, s_pepper_1, &pS_pepper_1));

    ERR_GEN ((xres < 200) || (yres < 200) || (xres > 400) || (yres > 400),
	     cErrApps_invalidResolution);


    /* Be sure to handle anamorphic resolutions */

    /* h res smaller than v res? */
    if ((xres + 10) < yres)
    {
        ERR_PASS (w_reduceRows (pIPWState, prSrc, 2, OR, &prTmp));

        if (yres >= 300)
	{
	    ERR_PASS_CLEANUP (w_reduce (pIPWState, prTmp, 1, 0, 0, 0, 
					&prMask));
            ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
        }
	else
	{
	    prMask = prTmp;
	    prTmp = NULL;
	}
    }
    else
    {
        /* v res smaller than h res */
        if ((xres - 10) > yres)
	{
            ERR_PASS_CLEANUP (w_reduceCols (pIPWState, prSrc, 2, OR, &prTmp));

	    if (xres >= 300)
	    {
		ERR_PASS_CLEANUP (w_reduce (pIPWState, prTmp, 1, 0, 0, 0, 
					      &prMask));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	    }
	    else
	    {
		prMask = prTmp;
		prTmp = NULL;
	    }
	}
	else
	{
	    if (xres >= 300)
		ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 1, 0, 0,
					    &prMask));
	    else
		ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 0, 0, 0, 
					      &prMask));
	}
    }

#if DEBUG_STIPPLE
    ERR_PASS_CLEANUP (w_write (pIPWState, "mask1", prMask, TIFF_G3));
#endif

    /* close up the stipple */
    ERR_PASS_CLEANUP (w_close (pIPWState, prMask, pStrel5h, &prMask));
    ERR_PASS_CLEANUP (w_close (pIPWState, prMask, pStrel2v, &prMask));

#if DEBUG_STIPPLE
    ERR_PASS_CLEANUP (w_write ("mask2", prMask, TIFF_G3));
#endif

    /* remove rules & small stuff from the mask */
    ERR_PASS_CLEANUP (w_open (pIPWState, prMask, pStrel19v, &prMask));
    ERR_PASS_CLEANUP (w_open (pIPWState, prMask, pStrel5h, &prMask));


#if DEBUG_STIPPLE
    ERR_PASS_CLEANUP (w_write ("mask3", prMask, TIFF_G3));
#endif

    ERR_PASS_CLEANUP (w_isZero (pIPWState, prMask, &isZero));
    if (isZero)
    {
        /* nothing in the stipple mask to start with */
        ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMask));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prUnmask));
    }
    else  /* something was in the mask */
    {
        /* make the "unmask" */

	/* Be sure to handle anamorphic resolutions */
    
	/* h res smaller than v res? */
	if ((xres + 10) < yres)
	{
	    ERR_PASS_CLEANUP (w_reduceRows (pIPWState, prSrc, 2, 
					    AND, &prTmp));
    
	    if (yres >= 300)
	    {
		ERR_PASS_CLEANUP (w_reduce (pIPWState, prTmp, 2, 0, 0, 0,
					    &prUnmask));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	    }
	    else
	    {
		prUnmask = prTmp;
		prTmp = NULL;
	    }
	}
	else
	{
	    /* v res smaller than h res */
	    if ((xres - 10) > yres)
	    {
		ERR_PASS_CLEANUP (w_reduceCols (pIPWState, prSrc, 2, AND, 
						&prTmp));
    
		if (xres >= 300)
		{
		    ERR_PASS_CLEANUP (w_reduce (pIPWState, prTmp, 
						2, 0, 0, 0, &prUnmask));
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		}
		else
		{
		    prUnmask = prTmp;
		    prTmp = NULL;
		}
	    }
	    else
	    {
		if (xres >= 300)
		    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 
						  4, 2, 0, 0, &prUnmask));
		else
		    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 
						  3, 0, 0, 0, &prUnmask));
	    }
	}

#if DEBUG_STIPPLE
    ERR_PASS_CLEANUP (w_write ("unmask1", prUnmask, TIFF_G3));
#endif

	/* remove 1x1 pepper from the "unmask" */
	ERR_PASS_CLEANUP (w_erode (pIPWState, prUnmask, pS_pepper_1,
				   &prTmp));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, prUnmask));

#if DEBUG_STIPPLE
	ERR_PASS_CLEANUP (w_write (pIPWState, "pep1", prTmp, TIFF_G3));
	ERR_PASS_CLEANUP (w_write (pIPWState, "mask3", prMask, TIFF_G3));
	ERR_PASS_CLEANUP (w_write (pIPWState, "unmask2", prUnmask, TIFF_G3));
#endif

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prUnmask, SUBTRACT, prMask));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prUnmask));


#if DEBUG_STIPPLE
	ERR_PASS_CLEANUP (w_write (pIPWState, "mask4", prMask, TIFF_G3));
#endif

	ERR_PASS_CLEANUP (w_isZero (pIPWState, prMask, &isZero));
        if (!isZero)
	{
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prMask, INVERT, prMask));
            *pPrOutMask = prMask;
        }
	else
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMask));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prUnmask)
	w_destroyPixr (pIPWState, &prUnmask);
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if (prMask)
	w_destroyPixr (pIPWState, &prMask);

    ERR_RETURN;
}  /* createStippleMaskPr () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Takes vertical and horizontal rules out of a binary image.
 *
 * LongDescription:::
 *  Pages often contain "rules" which are used to separate sections,
 *  surround tables, etc. This routine attempst to find and remove
 *  such elements while trying to preserve stuff that the rules touch
 *  or come close to.
 *
 * PrivateDescription:::
 *  What it does is: open with a vertical and horizontal STREL to get
 *  the lines, ORs them together, dilates the result a bit, and subtracts
 *  that from the original. The result is then closed to get rid of
 *  holes that might have been made through text, dilated a bit, and then
 *  ANDED back with the original, so that we hopefully preserve stuff
 *  that was chopped off.
 * <BR>
 * ------
 * <BR>
 * The original version wasn't really good enough, so I've written a new
 * version. What this one does is use 1D AND reductions to isolate long
 * horizontal or vertical features. Each axis is done separately.
 * Before the reduction is done, we attempt to remove "tall, thick" 
 * features by doing an open to get the tall stuff, doing an open in the
 * other direction to get rid of thin stuff (like orthogonal rules), 
 * and then subtracting that from the original to keep only thin stuff. 
 * This scheme works OK, though there is the occasional rule protruding
 * out of something like a stylized logo (the BOSE logo, for example)
 * that gets by, and the rule sticking out of the top of the "E" in BOSE 
 * gets blasted.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Original image
 *
 * Output:::
 *  pPrDest::        Image with rules suppressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 * document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_removeRules (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    PIXR              **pPrDest)
{
    PROLOGUE;
    PIXR    *h1 = NULL, *h2 = NULL, *v1 = NULL, *v2 = NULL, *exp = NULL,
	    *reduced = NULL;
    Int32    r, dpi;
    STREL   *pStrel3 = NULL, *pStrel5h = NULL, *pStrel5v = NULL,
	    *pStrel10h = NULL, *pStrel10v = NULL, *pStrel15h = NULL,
	    *pStrel15v = NULL;
    Int32    rightColsToClear, bottomRowsToClear;
    Bool     allocatedHere = cFalse;
    ScaledInt1616    xDPI, yDPI;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    ERR_PASS (wp_clonePixrConditionalNoInit (pIPWState, prSrc, pPrDest,
					    &rightColsToClear, 
					    &bottomRowsToClear,
					    &allocatedHere));
    ERR_GEN_PIXR (*pPrDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);
	
	
    ERR_PASS (w_initStrelByName(pIPWState, strel3));
    ERR_PASS (w_initStrelByName(pIPWState, strel5h));
    ERR_PASS (w_initStrelByName(pIPWState, strel5v));
    ERR_PASS (w_initStrelByName(pIPWState, strel10h));
    ERR_PASS (w_initStrelByName(pIPWState, strel10v));
    ERR_PASS (w_initStrelByName(pIPWState, strel15h));
    ERR_PASS (w_initStrelByName(pIPWState, strel15v));

    ERR_PASS (w_getStrel (pIPWState, strel3, &pStrel3));
    ERR_PASS (w_getStrel (pIPWState, strel5h, &pStrel5h));
    ERR_PASS (w_getStrel (pIPWState, strel5v, &pStrel5v));
    ERR_PASS (w_getStrel (pIPWState, strel10h, &pStrel10h));
    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel10v));
    ERR_PASS (w_getStrel (pIPWState, strel15h, &pStrel15h));
    ERR_PASS (w_getStrel (pIPWState, strel15v, &pStrel15v));

    if (dpi <= 100)
	reduced = prSrc;
    else
	ERR_PASS (w_reduce2 (pIPWState, prSrc, 2, &reduced));

    /* Do a 4x rows/cols reduction */
    r = 4;

    /* Find "thick" stuff */
    ERR_PASS_CLEANUP (w_open (pIPWState, reduced, pStrel10h, &v1));

    /* But not "thick" stuff that's thin in the other direction (like rules) */
    ERR_PASS_CLEANUP (w_open (pIPWState, v1, pStrel5v, &v1));

    /* And subtract the "thick" stuff out */
    ERR_PASS_CLEANUP (w_copy (pIPWState, reduced, &h1));

    ERR_PASS_CLEANUP (w_logicOp (pIPWState, v1, SUBTRACT, h1));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &v1));

    /* Do an AND-reduction, to get long stuff */
    ERR_PASS_CLEANUP (w_reduceRows (pIPWState, h1, r, AND, &h2));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &h1));

    /* Get rid of non-long stuff */
    ERR_PASS_CLEANUP (w_open (pIPWState, h2, pStrel15v, &h2));

    /* Do an OR reduction the other way */
    ERR_PASS_CLEANUP (w_reduceCols (pIPWState, h2, r, OR, &h1));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &h2));

    /* Do the same thing in the other direction */
    ERR_PASS_CLEANUP (w_open (pIPWState, reduced, pStrel10v, &h2));

    ERR_PASS_CLEANUP (w_open (pIPWState, h2, pStrel5h, &h2));
    ERR_PASS_CLEANUP (w_copy (pIPWState, reduced, &v1));

    /* If we made a reduced thing, we can get rid of it here */
    if (dpi > 100)
    {
	ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &reduced));
    }

    ERR_PASS_CLEANUP (w_logicOp (pIPWState, h2, SUBTRACT, v1));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &h2));
    ERR_PASS_CLEANUP (w_reduceCols (pIPWState, v1, r, AND, &v2));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &v1));
    ERR_PASS_CLEANUP (w_open (pIPWState, v2, pStrel15h, &v2));
    ERR_PASS_CLEANUP (w_reduceRows (pIPWState, v2, r, OR, &v1));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &v2));

    /* Or them together, expand, dilate a bit, and subtract out those
       pesky rules. */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, v1, OR, h1));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &v1));

    if (dpi <= 100)
	ERR_PASS_CLEANUP (w_expand (pIPWState, h1, r, &exp));
    else
	ERR_PASS_CLEANUP (w_expand (pIPWState, h1, r*2, &exp));

    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &h1));

    ERR_PASS_CLEANUP (w_dilate (pIPWState, exp, pStrel3, &exp));
    if (*pPrDest != prSrc)
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, *pPrDest));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, exp, SUBTRACT, *pPrDest));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &exp));

	/* If we need to clear any parts of the frame, do so */
    if ((rightColsToClear | bottomRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, *pPrDest,
						     rightColsToClear,
						     bottomRowsToClear));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (h1) w_destroyPixr(pIPWState, &h1);
    if (h2) w_destroyPixr(pIPWState, &h2);
    if (v1) w_destroyPixr(pIPWState, &v1);
    if (v2) w_destroyPixr(pIPWState, &v2);
    if (exp) w_destroyPixr(pIPWState, &exp);
    if (dpi > 100 && reduced) w_destroyPixr(pIPWState, &reduced);
    if (allocatedHere)
	w_destroyPixr(pIPWState, pPrDest);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Seed only "rule-like" connected components.
 *
 * LongDescription:::
 *  Seed only "rule-like" connected components. This is a callback to
 *  i_findConnectedComponentBoundingBoxes,
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Pointer to IPCore state
 *  packet::         Generic callback struct
 *  x::              x location of this CC.
 *  y::              y location of this CC.
 *  w::              Width of this CC.
 *  h::              Height of this CC.
 *  m::              Mass of this CC.
 *  nRuns::          Number of runs in this CC.
 *  px::             Seed pixel x location
 *  py::             Seed pixel y location
 *
 * Output:::
 *
 * Returns:::
 *  cSuccessful if successfully seeded CC 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_rr_filterRuleCCs (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    clean_CC_packet *pack;

    /* keep compilers quiet */
    x = x;
    y = y;
    m = m;
    nRuns = nRuns;

    pack = (clean_CC_packet *)packet;

    if (((w >= pack->minLinearDimension) && (w <= pack->maxLinearDimension)) ||
        ((h >= pack->minLinearDimension) && (h <= pack->maxLinearDimension)))
        ERR_PASS (w_pixrPutPixel (pIPWState, (PIXR *)(pack->data),
		  px - cFrameBits, py - cFrameBits, 1));

    NORMAL_RETURN;

}  /* rr_filterRuleCCs () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Removes horizontal/vertical rules from binary images using run filtering.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Input PIXR
 *  minLinearSize::  Minimum "length" (V/H) in pixels of rules to be removed
 *  maxLinearSize::  Maximum "length" (V/H) in pixels of rules to be removed
 *  minThickness::   Minimum line thickness (V/H) in pixels of rules to be
 removed
 *  maxThickness::   Maximum line thickness (V/H) in pixels of rules to be
 removed
 *
 * Output:::
 *  pPrDest::        Output PIXR, with rules removed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 * limits::
 *	prSrc and prDest must be the same size and depth.
 *
 *	maxThickness should be limited to less than 32 for
 *	best performance.
 * 
 *      This one works most of the time, but loses on text
 *	intersected by rules (e.g. strikeouts).
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 * dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_removeHVRules (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32               minLinearSize,
    Int32               maxLinearSize,
    Int32               minThickness,
    Int32               maxThickness,
    PIXR              **pPrDest)
{
    PROLOGUE;
    Int32     ret, wSrc, hSrc, wDest, hDest, bplH, w, h, frame;
    STREL    *thicken = NULL,
             *connect = NULL;
    PIXR     *hPr = NULL,
             *vPr = NULL,
             *tPr = NULL,
	     *prSrcReduced = NULL,
	     *seedPr = NULL,
	     *prDest;
    clean_CC_packet packet;
    UInt8    *pH = NULL;
    void     *runFilterState;
    Int32     rightColsToClear, bottomRowsToClear;
    Bool      allocatedHere = cFalse;


    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_PASS (wp_clonePixrConditionalNoInit (pIPWState, prSrc, pPrDest,
					    &rightColsToClear, 
					    &bottomRowsToClear,
					    &allocatedHere));
    prDest = *pPrDest;

    /* Sanity checks */
    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &wSrc));
    ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &hSrc));
    ERR_PASS (w_pixrGetWidth (pIPWState, prDest, &wDest));
    ERR_PASS (w_pixrGetHeight (pIPWState, prDest, &hDest));
    ERR_GEN (hSrc != hDest || wSrc != wDest, cErrInvalidParm);


    /* Create a reduced version of the input for speedup and
     * lower memory footprint.
     */
    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 0, 0, 0, &prSrcReduced));

    /* Create intermediate result pixrs -- this is the high water mark */
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prSrcReduced, &hPr));
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, prSrcReduced, &vPr));
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, prSrcReduced, &tPr));

    /* Maybe later this should be related to resolution */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 3, 3, 1, 1, &thicken));

    /* Scale down the rule sizes by 2 (being careful about rounding) */
    minLinearSize = (minLinearSize + 1) / 2;
    maxLinearSize = (maxLinearSize + 1) / 2;
    minThickness = (minThickness + 1) / 2;
    maxThickness = (maxThickness + 1) / 2;

    /** Find the horizontal rule-like things **/

    /* get ready for run filtering */
    ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, prSrcReduced, cBothRuns,
                                       &runFilterState));

    /* Get only things which are the right thickness */
    ERR_PASS_CLEANUP (w_filterVerticalRuns (pIPWState, runFilterState,
                                            prSrcReduced, 1, 0,
					    minThickness, maxThickness, 
					    cPassRuns, &hPr));

    /* Get the long runs of thin things */
    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, runFilterState,
                                              prSrcReduced, 1, 0,
					      minLinearSize/2, maxLinearSize, 
					      cPassRuns, &tPr));
    /* Keep long AND thin things */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, tPr, AND, hPr));
    
    /* Try to close small gaps between runs */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, MIN (19, maxThickness), 1, 
					MIN (19, maxThickness) / 2, 0, 
					&connect));
    /* Do a close and keep the footprint down by using tPr! */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, hPr, connect, &tPr));
    ERR_PASS_CLEANUP (w_erode (pIPWState, tPr, connect, &hPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &connect));

    /** Find the vertical rule-like things **/

    /* Find the long runs of thin things */
    ERR_PASS_CLEANUP (w_filterVerticalRuns (pIPWState, runFilterState,
                                            prSrcReduced, 1, 0,
					    minLinearSize/2, maxLinearSize,
					    cPassRuns, &tPr));

    /* Find only things which are the right thickness */
    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, runFilterState,
                                              prSrcReduced, 1, 0,
					      minThickness, maxThickness,
					      cPassRuns, &vPr));

    /* we're done with the run filtering */
    ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    /* Keep tall AND thin things */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, tPr, AND, vPr));

    /* Try to close small gaps between runs */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1, MIN (19, maxThickness),  
					0, MIN (19, maxThickness) / 2, 
					&connect));

    /* Do a close and keep the footprint down by using tPr! */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, vPr, connect, &tPr));
    ERR_PASS_CLEANUP (w_erode (pIPWState, tPr, connect, &vPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &connect));

    /* combine the tall thin and wide thin things */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, hPr, OR, vPr));

    /* thicken them and clip to the reduced original */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, vPr, thicken, &hPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &vPr));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrcReduced, AND, hPr));

    /* init the CC callback packet */
    /* High water mark! */
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, prSrcReduced, &seedPr));
    packet.data = (void *)seedPr;
    packet.minLinearDimension = minLinearSize;
    packet.maxLinearDimension = maxLinearSize;
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, hPr, &pH));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, hPr, &bplH));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, hPr, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, hPr, &h));
    ERR_PASS_CLEANUP (w_pixrGetFrame (pIPWState, hPr, &frame));

    /* Run CC-BB and eliminate non-rules */
    ERR_PASS_CLEANUP( i_findConnectedComponentBoundingBoxes (pIPWState,
					       pH, bplH,
					       frame, frame, w, h,
					       minThickness, minThickness,
					       1, (void *)(&packet),
					       wp_rr_filterRuleCCs, &ret));
    
    /* grow the seeds of the rules */
    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, seedPr, hPr, 3, &seedPr));

    /* Now for the "funny stuff" which tries to preserve things which
     * were intersected by or adjacent to rules which were removed.
     * this only works for largish things (i.e. not text).
     */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, seedPr, thicken, &tPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &thicken));

    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrcReduced, AND, tPr));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, seedPr, XOR, tPr));
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, maxThickness, 1,
				      maxThickness / 2, 0, &thicken));
    ERR_PASS_CLEANUP (w_close (pIPWState, tPr, thicken, &tPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &thicken));

    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1, maxThickness, 
				      0, maxThickness / 2, &thicken));
    ERR_PASS_CLEANUP (w_close (pIPWState, tPr, thicken, &tPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &thicken));

    ERR_PASS_CLEANUP (w_logicOp (pIPWState, tPr, SUBTRACT, seedPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tPr));

    /* invert for expand and mask */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, seedPr, INVERT, seedPr));

    /* if not in-place, copy the input to the output before masking */
    if (prDest != prSrc)
    {
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prDest));
    }

    /* mask out the rules */
    ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, prDest, seedPr, &prDest));

	/* If we need to clear any parts of the frame, do so */
    if ((rightColsToClear | bottomRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prDest,
						     rightColsToClear,
						     bottomRowsToClear));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &seedPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &vPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &hPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSrcReduced));
    NORMAL_RETURN;
    
ERR_CLEANUP:

    if (tPr) w_destroyPixr (pIPWState, &tPr);
    if (seedPr) w_destroyPixr (pIPWState, &seedPr);
    if (vPr) w_destroyPixr (pIPWState, &vPr);
    if (hPr) w_destroyPixr (pIPWState, &hPr);
    if (prSrcReduced) w_destroyPixr (pIPWState, &prSrcReduced);
    if (thicken) w_destroyStrel (pIPWState, &thicken);
    if (connect) w_destroyStrel (pIPWState, &connect);

    ERR_RETURN;
}  /* removeHVRulesPr () */


/*

 $Log: clean.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.50  1998/01/14 00:10:55  ddavies
 clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.49  1997/06/26 20:00:28  ddavies
 clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.48  1997/05/15 18:46:57  ddavies
 ddavies on Thu May 15 11:45:43 PDT 1997

 clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v

 Fix bug found by Ron Danisewicz.  i_findConnectedComponentBoundingBoxes
 was being called with the x,y coords of the rectangle being [0,0] instead
 of [frame, frame].  This meant it was missing stuff on the right and
 bottom edges of the image.

 Revision 1.47  1997/04/15 20:13:11  ddavies
 clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v

 Substitute NoInit (uninitialized) version of pixr creation routines where
 possible.

 Revision 1.46  1997/01/14 02:45:17  ddavies
 ddavies on Mon Jan 13 18:44:01 PST 1997

 clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v

 Add new routine, w_findStipple.  This routine returns an image that contains
 only the high frequency little dots in the original image.  It does a pretty
 good job of stripping stipple off text.  It also can fairly quickly detect
 whether there is any stipple at all.

 Revision 1.45  1996/01/04 20:21:24  ejaquith
 clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.44  1995/12/13  04:05:21  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.43  1995/12/12  20:23:09  dferrell
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Updated calls to run filters in w_removeHVRules ().
 *
 * Revision 1.42  1995/11/13  20:09:50  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.41  1995/11/04  01:35:51  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing
 ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.40  1995/10/13  21:58:53  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.39  1995/10/03  20:53:27  ejaquith
 * ejaquith on Tue Oct 3 13:53:12 PDT 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * FF_TIFF_G3 ==> TIFF_G3
 *
 * Revision 1.38  1995/10/03  01:58:56  macleod
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in
 interfaces.
 *
 * Revision 1.37  1995/09/21  22:51:45  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.36  1995/09/19  23:29:16  ddavies
 * ddavies on Tue Sep 19 16:28:55 PDT 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Change type of allocatedHere var from Int32 to Bool.
 *
 * Revision 1.35  1995/09/19  19:19:06  macleod
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi
 parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.34  1995/09/11  21:53:45  macleod
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.33  1995/09/07  01:29:50  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's
 declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.32  1995/09/06  20:17:31  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.31  1995/09/06  01:03:18  ddavies
 * ddavies on Tue Sep 5 18:02:33 PDT 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Use the option parameter in cleanPepper3 and cleanPepper3Masked so that
 * WATCOM doesn't whine about it being unused.
 *
 * Revision 1.30  1995/09/05  19:00:39  macleod
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Removed calls to GET_STREL macro, instead calling getStrel()
 function.
 * This required using local variables rather than invoking GET_STREL
 * as an argument to another function, since getStrel can return an error
 * code which should be checked.
 *
 * Revision 1.29  1995/09/01  21:02:14  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.28  1995/08/31  21:11:19  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.27  1995/08/30  17:20:01  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * removeHVRulesPr() had a cleanup logic bug, and was calling createPixrT()
 * with a nonNull pixr pointer.  Fixed these.
 *
 * Revision 1.26  1995/08/28  23:57:06  ejaquith
 * ejaquith on Mon Aug 28 16:55:17 PDT 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Fixed bug in clean ipwtest target.
 * Was destroying 2 STRELs that had already been destroyed.
 * Additionally, added ERR_PASS_CLEANUP wrappings to w_destroyPixr() calls
 * (which we REALLY should do!)
 *
 * Revision 1.25  1995/08/25  01:35:17  ejaquith
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Fixed warnings
 *
 * Revision 1.24  1995/08/17  23:43:02  dferrell
 * dferrell on Thu Aug 17 16:41:48 PDT 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Removed include of "appserr.pub".
 *
 * Revision 1.23  1995/08/17  21:54:02  macleod
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.22  1995/07/26  23:55:40  dferrell
 * dferrell on Wed Jul 26 16:55:20 PDT 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Fixed procName string in createStippleMaskPr ().
 *
 * Revision 1.21  1995/03/15  00:48:13  dferrell
 * dferrell on Tue Mar 14 16:46:50 PST 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Fixed bugs in and generally improved removeHVRulesPr ().
 *
 * Revision 1.20  1995/03/09  00:13:01  dferrell
 * dferrell on Wed Mar 8 16:10:09 PST 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Updated createStippleMaskPr () to handle anamorphic resolutions.  Note
 * that the reduced mask for these cases may not work with expandAndMask.
 * Added new removeHVRulesPr () which attempts to remove horizontal
 * and vertical rules from images.  It uses run filtering, morphology,
 * and connected component analysis and can operate "in-place."
 *
 * Revision 1.19  1995/01/27  01:51:01  macleod
 * macleod on Thu Jan 26 17:50:27 PST 1995
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * The "reduced" image, when created, was not being freed, leaking hundreds
 * of K at a time.
 *
 * Revision 1.18  1995/01/24  21:49:06  lperry
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Incorporate changes to use new strel manager interface.
 * API's to all functions changed to use passed in StrelInfo arg.
 *
 * Revision 1.17  1994/12/13  19:41:36  macleod
 * macleod on Tue Dec 13 11:41:35 PST 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * New version of removeRulesPr.  What this one does is use 1D AND reductions
 * to isolate long horizontal or vertical features. Each axis is done
 * separately.  Before the reduction is done, we attempt to remove "tall,
 * thick" features by doing an open to get the tall stuff, doing an open in the
 * other direction to get rid of thin stuff (like orthogonal rules), and then
 * subtracting that from the original to keep only thin stuff.  This scheme
 * works OK, though there is the occasional rule protruding out of something
 * like a stylized logo (the BOSE logo, for example) that gets by, and the rule
 * sticking out of the top of the "E" in BOSE gets blasted.  It does work
 * better than the previous version in the limited testing I've done.  It's
 * also faster. (No performace or quality requirements were ever provided, so I
 * cannot judge how well it meets them.)
 *
 * Revision 1.16  1994/12/09  23:41:08  dferrell
 * dferrell on Fri Dec 9 15:39:24 PST 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Added new function createStippleMaskPr () which creates a reduced (by
 * 2 or 4) mask for removing stipple.  The algorithm attempts to preserve
 * text on stipple.
 *
 * Revision 1.15  1994/10/21  18:17:03  macleod
 * macleod on Fri Oct 21 11:16:07 PDT 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * cleanStipplePixr was calling expandAndMask even if the reduction
 * factor was 1, causing expandAndMask to barf. Changed it so that it
 * calls opPrPr if the reduction factor is 1.
 *
 * Revision 1.14  1994/10/11  21:24:41  macleod
 * macleod on Tue Oct 11 14:23:19 PDT 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Changed stipple removal so that it doesn't create a full-sized mask,
 * thus saving some memory.
 *
 * Revision 1.13  1994/10/11  18:45:28  macleod
 * macleod on Tue Oct 11 11:43:31 PDT 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * cleanPepper3 made an unnecessary copy of the full-sized image. Changed so
 * that it works directly on the destination image.
 *
 * Revision 1.12  1994/10/07  19:05:58  lperry
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.11  1994/08/10  21:56:46  lperry
 * lperry on Wed Aug 10 14:55:41 PDT 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Changes made to use new strel manager.
 *
 * Revision 1.10  1994/05/26  19:19:28  ddavies
 * ddavies on Thu May 26 12:18:24 PDT 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Change the abortI's whose return values are not used to abortV's.
 *
 * Revision 1.9  1994/05/19  20:55:18  macleod
 * macleod on Thu May 19 13:54:40 PDT 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * cleanStipplePixr now uses a smaller strel at lower reductions, so that
 * 100dpi images don't get some of their text blown away.
 *
 * Revision 1.8  1994/05/19  01:02:44  macleod
 * macleod on Wed May 18 18:02:43 PDT 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Did some work on removeRulesPr. It does a better job at 100dpi, and doesn't
 * get rid of stuff out of the middle of big, solid things. It still probably
 * needs more work--it doesn't know the difference between a light stem of a
 * big letter and a rule, especially if the letter has nothing around the stem,
 * like "L" or "I".
 *
 * Revision 1.7  1994/01/05  22:06:56  macleod
 * macleod on Wed Jan 5 14:06:31 PST 1994
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * added removeRulesPr, which removes horizontal and vertical rules from
 * a binary image
 *
 * Revision 1.6  1993/12/15  06:34:07  macleod
 * macleod on Tue Dec 14 22:33:51 PST 1993
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Added a version of cleanStipple which takes a reduction factor
 *
 * Revision 1.5  1993/12/03  01:13:31  ddavies
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.4  1993/11/16  23:08:15  macleod
 * macleod on Tue Nov 16 15:07:14 PST 1993
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Added a version of cleanPepper3 called cleanPepper3Masked, which takes
 * a mask used to protect regions that shouldn't be cleaned. A zero in
 * in the mask means that a pixel is protected, a one means that a
 * pixel can be de-peppered.
 *
 * Revision 1.3  1993/11/08  22:38:36  macleod
 * macleod on Mon Nov 8 14:37:51 PST 1993
 *
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Added cleanStipple6Pixr, which attempts to take the stipple out of
 * a binary image.
 *
 * Revision 1.2  1993/11/06  03:10:18  ddavies
 * clean.c -> /product/ipcore/ipwrap/src/RCS/clean.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

