/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  clipzone.c--Forms processing clip zone algorithm.
 *
 *	Public:
 *		extractClipZone
 *
 *	Private:
 *		wp_czFilterEdgeCCs
 *		wp_czFreeEdgeList
 *		wp_czReduceResolution
 *
 */


#define DEBUG_CZ 0
#define DEBUG_CZ_P 0


#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "alplib.h"
#include "logicops.h"
#include "morphg.h"
#include "shrscgr.pub"
#include "shrcc.pub"
#include "frames.pub"
#include "defines.pub"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "clipzone.prv"
#include "clipzone.pub"
#include "morph.pub"
#include "runmorph.pub"
#include "reduce.pub"
#include "strel.pub"
#include "fill.pub"
#include "expand.pub"
#include "subsmpl.pub"
#include "runfilt.pub"
#include "logicops.pub"
#include "rectops.pub"
#include "cut.pub"

#if DEBUG_CZ | DEBUG_CZ_P 
#include <stdio.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif
#endif

#if DEBUG_CZ
#include "imagerep.h"
#include "io.pub"
#endif


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Callback procedure passed into i_findConnectedComponentBoundingBoxes
 *
 * LongDescription:::
 *   Callback procedure passed into i_findConnectedComponentBoundingBoxes
 *   which creates a list of CCs which are potential clip box edges.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Pointer to IPCore state
 *  packet::         Generic CC analysis structure.
 *  x::              x location of current CC.
 *  y::              y location of current CC.
 *  w::              Width of current CC.
 *  h::              Height of current CC.
 *  m::              Mass (number of ON bits) in
 *		     current CC.
 *  nRuns::          Number of runs in current CC.
 *  px::             x location of an ON pixel in the
 *		     current CC (for use as a seed).
 *  py::             y location of an ON pixel in the
 *		     current CC (for use as a seed).
 *
 * Output:::
 * 
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  clipzones
 *
 * Keywords:::
 *  clipzones, connected component analysis
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_czFilterEdgeCCs (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    clip_zone_packet *czPacket;
    clip_box_edge *current;

    nRuns = nRuns;
    m = m;

    czPacket = (clip_zone_packet *)packet;

    /* Check for correct dimensions */
    if ((w >= czPacket->minWidth) &&
        (w <= czPacket->maxWidth) &&
	(h >= czPacket->minHeight) &&
	(h <= czPacket->maxHeight))
    {
        /* allocate a list node and insert it at the head */
        current = (clip_box_edge *)IPW_MALLOC (sizeof (clip_box_edge),
					       cAllocatorData);
	ERR_GEN(current == NULL, cErrNomem);

	current->x = x - cFrameBits;
	current->y = y - cFrameBits;
	current->w = w;
	current->h = h;
	current->px = px - cFrameBits;
	current->py = py - cFrameBits;
	current->next = (clip_box_edge *)(czPacket->data);
	czPacket->data = (void *)current;
	czPacket->count += 1;
    }

    NORMAL_RETURN;

}  /* wp_czFilterEdgeCCs () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free a list of clip_box-edge data structures
 *
 * LongDescription:::
 *  Given a pointer to the head of a list of clip_box_edge
 *  data structs, frees them.  If the  list is empty,
 *  does nothing.  Sets the head pointer to NULL on return.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  head::           Head of list, set to NULL on exit
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  clipzones
 *
 * Keywords:::
 *  clipzones
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_czFreeEdgeList (
    IPWState           *pIPWState,
    clip_box_edge     **head)
{
    PROLOGUE;
    clip_box_edge *thisOne;

    /* proceed through the list freeing each element along the way */
    thisOne = *head;
    while (thisOne != NULL)
    {
	clip_box_edge *tmpOne;

        /* keep a pointer to this element */
	tmpOne = thisOne;

        /* go to the next element */
	thisOne = thisOne->next;

        /* free the previously visited element */
	IPW_FREE (tmpOne, cAllocatorData);
    }

    *head = NULL;
    NORMAL_RETURN;

}  /* wp_czFreeEdgeList () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reduce resolution of input PIXR to between 100 and 150 dpi square
 *
 * LongDescription:::
 *  Attempts to reduce the input PIXR to between 100 and 150 dpi.
 *  Takes into account anamorphism and unity reduction.  If no reduction
 *  is needed, *outPr is set to inPr.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           Input PIXR to be reduced
 *
 * Output:::
 *  hRed::           Horizontal reduction factor
 *  vRed::           Vertical reduction factor
 *  outPr::          Reduced output PIXR
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  clipzones
 *
 * Keywords:::
 *  clipzones
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_czReduceResolution (
    IPWState           *pIPWState,
    PIXR	       *inPr,
    Int32	       *hRed,
    Int32	       *vRed,
    PIXR	      **outPr)

{
    PROLOGUE;
    Int32           xRes, yRes;
    ScaledInt1616   xDPI, yDPI;


    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, inPr, &xDPI, &yDPI));

    xRes = ScaledInt1616_TO_Int32(xDPI);
    yRes = ScaledInt1616_TO_Int32(yDPI);

    ERR_GEN (((xRes < 1) || (yRes < 1)), cErrPixr_resolution);

    /* Approximately symmetrical resolution */
    if ((MAX (xRes, yRes) - MIN (xRes, yRes)) < 16)
    {
        if (xRes > 390)
	{
	    ERR_PASS_CLEANUP (w_reduce4 (pIPWState, inPr, 1, 2,
					 outPr));
	    *hRed = 4;
	}
	else
	{
	    if (xRes > 195)
	    {
	        ERR_PASS_CLEANUP (w_reduce2 (pIPWState, inPr, 1, 
					     outPr));
		*hRed = 2;
	    }
	    else
	    {
	        /* no reduction */
		*hRed = 1;
		*outPr = inPr;
	    }
	}
	*vRed = *hRed;
    }
    else
    {
        /* This section kind of assumes 200 x 100 or 100 x 200
	 * since it only reduces the higher resolution of the 
	 * dimensions by 2.
	 */
	if (xRes > yRes)
	{
	    ERR_PASS_CLEANUP (w_reduceCols (pIPWState, inPr, 2, OR,
					    outPr));
	    *hRed = 2;
	    *vRed = 1;
	}
	else
	{
	    ERR_PASS_CLEANUP (w_reduceRows (pIPWState, inPr, 2, OR, 
					    outPr));
	    *hRed = 1;
	    *vRed = 2;
	}
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    /* Destroy reduced pixr if we made it in this function */
    if ((*hRed != 1) || (*vRed != 1))
        w_destroyPixr (pIPWState, outPr);

    ERR_RETURN;

}  /* wp_czReduceResolution () */


#if 0
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  From candidate clip zone box edges, find the true box corners
 *
 * LongDescription:::
 *  Given lists of candidate vertical and horizontal box edges, attempts
 *  to match them with one another to produce the true corner locations
 *  of the clip zone bounding box.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  clipzones
 *
 * Keywords:::
 *  clipzones
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_czMatchBoxSides (
    IPWState           *pIPWState,
    Int32		xOffset,
    Int32		yOffset,
    clip_box_edge      *horizontalEdges,
    clip_box_edge      *verticalEdges,
    PIXR              **boxPr)
{
    clip_box_edge *pv, *ph;


    /* attempt to match parallel box sides first */

    /* left and right */

    /* top and bottom */

    /* if we didn't match at least one parallel pair then try to find
     * an edge pair which forms one corner of the box and deduce the
     * shape and size of the box from their location and orientation.
     */
    ph = horizontalEdges;
    pv = verticalEdges;

    while (ph != NULL)
    {
        while (pv != NULL)
	{
            if (ph.x
	}  /* for each vertical edge */

    }  /* for each horizontal edge */

}  /* wp_czMatchBoxSides () */
#endif


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clip the contents of a rectangular, outlined region.
 *
 * LongDescription:::
 *  Clip the contents of a rectangular, outlined region.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Input image with one clip region in it
 *  roi::            Region of interest rectangle.  If this is
 *		     NULL, all of prSrc will be searched.
 *  xLineThickness:: Expected thickness of vertical lines in 
 *		     horizontal pixels at xres.
 *  yLineThickness:: Expected thickness of horizontal lines in 
 *		     vertical pixels at yres.
 *  boxW::           Expected width of the outlining box in
 *		     horizontal pixels at xres.
 *  boxH::           Expected height of the outlining box in
 *		     vertical pixels at yres.
 *  xRange::         Range of variation in clip box width in
 *		     horizontal pixels:
 *		     (boxW - xRange) < w < (boxW + xRange)
 *  yRange::         Range of variation in clip box height in
 *		     vertical pixels:
 *		     (boxH - yRange) < h < (boxH + yRange)
 *  clipToSize::     If cTrue, clip the final result to the size of
 *                   the smallest rectangle which contains the clip
 *                   region.  If cFalse, the resulting image will
 *                   be the size of the roi (if non-NULL) or the
 *                   source image.
 *  clipLoc::        The x, y, w, and h fields of this rectangle
 *                   will be filled with the coordinates of the
 *                   minimum rectangle which contains the clip box.
 *                   Coordinates are in source image pixels even if
 *                   a valid roi rectangle is passed.
 *
 * Output:::
 *  pPrDest::        Result image with clip region contents.
 *		     If an roi rectangle is specified, this
 *		     pixr must be the same size as the roi,
 *		     otherwise it must be the same size as
 *		     prSrc.  If a pointer to NULL is passed in,
 *		     the output pixr will be created by the
 *		     algorithm.
 *
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  clipzones
 *
 * Keywords:::
 *  clipzones
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_extractClipZone (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *roi,
    Int32               xLineThickness,
    Int32               yLineThickness,
    Int32               boxW,
    Int32               boxH,
    Int32               xRange,
    Int32               yRange,
    Bool		clipToSize,
    RECTANGL	       *clipLoc,
    PIXR              **pPrDest)
{
    PROLOGUE;
    Int32           ret, i, c,
                    hRed, vRed,         /* reduction factors */
                    xResRed, yResRed,   /* resolution after reduction */
		    xThick, yThick,     /* expected line thickness of box */
                    minBoxW, maxBoxW,   /* range of box width after reduction*/
		    minBoxH, maxBoxH,   /* range of box height "           " */
		    xTol, yTol,         /* location tolerances at reduction */
		    x0, y0, x1, y1,     /* first box edge endpoints */
		    x2, y2, x3, y3,     /* second box edge endpoints */
		    hFound = cFalse,     /* found horizontal edges flag */
		    vFound = cFalse,     /* found vertical edges flag */
                    wF, hF, bpl,        /* framed width and height, pitch */
	            w, h, xRes = 0, yRes = 0;
    ScaledInt1616   xDPI, yDPI;
    UInt8          *pImgAll;            /* .r to frame of image */
    PIXR           *prRect = NULL,	/* pixr containing roi rect */
                   *prSrcReduced = NULL,       /* reduced original pixr */
		   *prTmp1 = NULL,
		   *prTmp2 = NULL,
		   *prTmp3 = NULL,
		   *prSeed = NULL;
    STREL          *hStrel = NULL,
                   *vStrel = NULL;
    clip_zone_packet CCpacket;          /* CCBB callback packet */
    clip_box_edge  *edge1, *edge2, *edget;/* list pointers for candidate edges*/

    ERR_GEN (prSrc == NULL, cErrNullPtr);
    ERR_GEN (pPrDest == NULL, cErrNullPtr);

    /* Make sure this is OK in case of abort */
    CCpacket.data = NULL;
    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));

    xRes = ScaledInt1616_TO_Int32(xDPI);
    yRes = ScaledInt1616_TO_Int32(yDPI);

    ERR_GEN (((xRes < 1) || (yRes < 1)), cErrPixr_resolution);

    /* Deal with the rectangle information, if any */
    if (roi == NULL)
    {
	if (*pPrDest == NULL)
	{
	    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, prSrc, pPrDest));
        }

        ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prSrc, &prRect));

	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prRect));
    }
    else
    {
	Int8 valid;

        /* check for valid rect */
        ERR_PASS_CLEANUP (w_isRectValid (pIPWState, prSrc, roi, &valid));
	ERR_GEN (!valid, cErrInvalidParm);

	if (*pPrDest == NULL)
	{
	    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, roi->w, roi->h,
						  1, xDPI, yDPI,
						  cFrameBits, pPrDest));
        }
	else
	{
	    /* check to make sure prDest is the same size as roi */
	    Int32 w, h;

	    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, *pPrDest, &w));
	    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, *pPrDest, &h));

	    ERR_GEN_CLEANUP (w != roi->w, cErrInvalidParm);
	    ERR_GEN_CLEANUP (h != roi->h, cErrInvalidParm);
	}

        /* create the rectPixr using the specified roi */
	ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, roi->w, roi->h, 1,
					      xDPI, yDPI,
					      cFrameBits, &prRect));

        ERR_PASS_CLEANUP (w_rasterop (pIPWState,
				      roi->x, roi->y, prSrc, PIX_SRC,
				      0, 0, roi->w, roi->h, prRect));

    }

#if DEBUG_CZ
    ERR_PASS_CLEANUP (w_write (pIPWState, "rect", prRect, TIFF_G3));
#endif

    /* Keep the compiler quiet */
    x0 = 0;
    x1 = 0;
    x2 = 0;
    x3 = 0;
    y0 = 0;
    y1 = 0;
    y2 = 0;
    y3 = 0;

#if DEBUG_CZ_P
    fprintf (stdout, "boxW = %ld  boxH = %ld\n", boxW, boxH);
    fprintf (stdout, "xThick = %ld  yThick = %ld\n", 
             xLineThickness, yLineThickness);
    fprintf (stdout, "xRange = %ld  yRange = %ld\n", xRange, yRange);
    fprintf (stdout, "xRes = %ld  yRes = %ld\n\n", xRes, yRes);
#endif

    /* Reduce to ~100dpi square (or 150dpi if 300dpi originally) */
    ERR_PASS_CLEANUP (wp_czReduceResolution (pIPWState, prRect, 
                                             &hRed, &vRed, &prSrcReduced));

#if DEBUG_CZ
    ERR_PASS_CLEANUP (w_write (pIPWState, "red", prSrcReduced, TIFF_G3));
#endif

    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prSrcReduced, &prTmp1));
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, prSrcReduced, &prTmp2));
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prSrcReduced, &prTmp3));
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, prSrcReduced, &prSeed));

    xResRed = xRes / hRed;
    yResRed = yRes / vRed;
    boxW /= hRed;
    boxH /= vRed;

    /* It's not a heuristic without fudge...this allows some variation
     * in box size due to image degradation during the document's
     * lifetime.  Note the use of _integer_ arithmetic.
     */
    minBoxW = boxW - ((xRange * xResRed) / xRes);
    maxBoxW = boxW + ((xRange * xResRed) / xRes);
    minBoxH = boxH - ((yRange * yResRed) / yRes);
    maxBoxH = boxH + ((yRange * yResRed) / yRes);

    xThick = MAX (1, (xLineThickness * xResRed) / xRes);
    yThick = MAX (1, (yLineThickness * yResRed) / yRes);

    /* More fudge based on max acceptable skew--note the use of 
     * _integer_ arithmetic. 
     */
    xTol = MAX (((xRange * xResRed) / xRes),
                ((Int32)(maxBoxH / MIN_SKEWED_RUN)));
    yTol = MAX (((yRange * yResRed) / yRes),
                ((Int32)(maxBoxW / MIN_SKEWED_RUN)));

#if DEBUG_CZ_P
    fprintf (stdout, "boxW = %ld  boxH = %ld\n", boxW, boxH);
    fprintf (stdout, "hRed = %ld  vRed = %ld\n", hRed, vRed);
    fprintf (stdout, "minW = %ld  maxW = %ld  minH = %ld  maxH = %ld\n",
             minBoxW, maxBoxW, minBoxH, maxBoxH);
    fprintf (stdout, "xThick = %ld  yThick = %ld\n", xThick, yThick);
    fprintf (stdout, "xTol = %ld  yTol = %ld\n", xTol, yTol);
#endif

    /* Choose the correct "close" strels depending on resolution */
    if (xResRed > 145)
	ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 5, 1, 2, 0, &hStrel));
    else
	ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 3, 1, 1, 0, &hStrel));

    if (yResRed > 145)
	ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1, 5, 0, 2, &vStrel));
    else
	ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1, 3, 0, 1, &vStrel));

    /*********************************
     * Look for horizontal box sides *
     *********************************/

    /* Close up any small horizontal gaps in the box outline */
    ERR_PASS_CLEANUP (w_close (pIPWState, prSrcReduced, hStrel, &prTmp1));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &hStrel));

    /* Find "long" horizontal segments */
    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, NULL, prTmp1, 1, 0,
	                           (Int32) MIN (minBoxW / 2, MIN_SKEWED_RUN), 
	                           maxBoxW,
	                           cPassRuns, &prTmp2));

#if DEBUG_CZ
    ERR_PASS_CLEANUP (w_write (pIPWState, "hedges", prTmp2, TIFF_G3));
#endif

    /* Set up the callback packet */
    CCpacket.minWidth = minBoxW;
    CCpacket.maxWidth = maxBoxW;
    CCpacket.minHeight = 1;
    CCpacket.maxHeight = yThick + MAX (1, (Int32)(maxBoxW / MIN_SKEWED_RUN));
    CCpacket.count = 0;
    CCpacket.data = NULL;

    /* Collect CCs which are the correct length/width */
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prTmp2, &pImgAll));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prTmp2, &bpl));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prTmp2, &wF));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prTmp2, &hF));

    ERR_PASS_CLEANUP (i_findConnectedComponentBoundingBoxes (pIPWState,
					       pImgAll, bpl, 
                                               0, 0, wF, hF,
					       0, 0, 1, (void *)(&CCpacket),
					       wp_czFilterEdgeCCs, &ret));

    /* Look for horizontal edges which are the correct distance apart,
     * vertically aligned with each other, and parallel.
     */
    if (CCpacket.count > 1)
    {
        edge1 = (clip_box_edge *)CCpacket.data;
        for (i = CCpacket.count - 2; i >= 0; i--)
        {
            /* Starting at the head of the list of horizontal lines,
	     * compare the current line's position relative to the
	     * ones after it in the list.
	     */
	    edge2 = edge1;
	    for (c = 0; c <= i; c++)
	    {
	        /* advance edge2 until the end of the list */
	        edge2 = edge2->next;

		/* compare positions of edge1 and edge2 */
		if ((ABS (edge1->x - edge2->x) < xTol) &&
		    (ABS ((edge1->x + edge1->w) - (edge2->x + edge2->w)) 
		                       < xTol) &&
		   ((ABS (edge1->y - edge2->y) >= minBoxH) &&
		    (ABS (edge1->y - edge2->y) <= maxBoxH)))
		{
		    /* These appear to form the top and bottom of the box
		     * so seed fill them into a PIXR from which we construct
		     * the extraction mask.
		     */
                    ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, prSeed, 
						    edge1->px, edge1->py, 1));
                    ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, prSeed, 
						    edge2->px, edge2->py, 1));

                    /* grow the two interesting edges */
		    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, 
							prSeed, prTmp2, 2, 
							&prSeed));

                    /* dilate the edges slightly to make sure there are
		     * no corner gaps in the case where all 4 side are found.
		     */
		    ERR_PASS_CLEANUP (w_dilateHorizontalRuns (pIPWState,
		                                              NULL,
							      prSeed,
							      5,
							      2,
							      &prSeed));

                    /* store the corners so we can draw in the vertical
		     * edges if we don't find them later.
		     */
		    if (edge1->y > edge2->y)
		    {
		        edget = edge1;
			edge1 = edge2;
			edge2 = edget;
		    }

		    if (edge1->x < edge2->x)
		    {
			x0 = edge1->x;
			y0 = edge1->y + edge1->h - 1;
			x1 = edge2->x;
			y1 = edge2->y + edge2->h - 1;
			x2 = edge1->x + edge1->w - 1;
			y2 = edge1->y;
			x3 = edge2->x + edge2->w - 1;
			y3 = edge2->y;
		    }
		    else
		    {
			x0 = edge1->x;
			y0 = edge1->y;
			x1 = edge2->x;
			y1 = edge2->y;
			x2 = edge1->x + edge1->w - 1;
			y2 = edge1->y + edge1->h - 1;
			x3 = edge2->x + edge2->w - 1;
			y3 = edge2->y + edge2->h - 1;
		    }

		    hFound = cTrue;
#if DEBUG_CZ_P
                    fprintf (stdout, "Found horizontal box sides.\n");
#endif
		    break;
		}
	    }

	    if (hFound)
	        break;

            edge1 = edge1->next;
	}
    }

    /* we're done with the horizontal edge CCs */
    ERR_PASS_CLEANUP (wp_czFreeEdgeList (pIPWState, (clip_box_edge **)&(CCpacket.data)));

    /* prSeed now contains the horizontal box sides if they were present */

    /*******************************
     * Look for vertical box sides *
     *******************************/

    /* Close up any small vertical gaps in the box outline */
    ERR_PASS_CLEANUP (w_close (pIPWState, prSrcReduced, vStrel, &prTmp1));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &vStrel));

    /* Find box-like parts */
    ERR_PASS_CLEANUP (w_filterVerticalRuns (pIPWState, NULL, prTmp1, 1, 0,
			    (Int32)MIN (minBoxH / 2, MIN_SKEWED_RUN), maxBoxH,
			    cPassRuns, &prTmp3));

#if DEBUG_CZ
    ERR_PASS_CLEANUP (w_write (pIPWState, "vedges", prTmp3, TIFF_G3));
#endif

    /* Set up the callback packet */
    CCpacket.minWidth = 1;
    CCpacket.maxWidth = xThick + MAX (1, (Int32)(maxBoxH / MIN_SKEWED_RUN));
    CCpacket.minHeight = minBoxH;
    CCpacket.maxHeight = maxBoxH;
    CCpacket.count = 0;
    CCpacket.data = NULL;

    /* Collect CCs which are the correct length/width */
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prTmp3, &pImgAll));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prTmp3, &bpl));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prTmp3, &wF));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prTmp3, &hF));

    ERR_PASS_CLEANUP( i_findConnectedComponentBoundingBoxes (pIPWState,
					       pImgAll, bpl, 0, 0, wF, hF,
					       0, 0, 1, (void *)(&CCpacket),
					       wp_czFilterEdgeCCs, &ret));

    /* clear the pixr we'll use for seeding */
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prTmp2, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prTmp2, &h));
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL,  PIX_CLR,
				  0, 0, w, h, prTmp2));

    /* look for vertical edges which are the correct distance apart,
     * horizontally aligned with each other, and parallel.
     */
    if (CCpacket.count > 1)
    {
        edge1 = (clip_box_edge *)CCpacket.data;
        for (i = CCpacket.count - 2; i >= 0; i--)
        {
            /* Starting at the head of the list of horizontal lines,
	     * compare the current line's position relative to the
	     * ones after it in the list.
	     */
	    edge2 = edge1;
	    for (c = 0; c <= i; c++)
	    {
	        edge2 = edge2->next;

		/* compare positions of edge1 and edge2 */
		if ((ABS (edge1->y - edge2->y) < yTol) &&
		    (ABS ((edge1->y + edge1->h) - (edge2->y + edge2->h)) 
		                       < yTol) &&
		   ((ABS (edge1->x - edge2->x) >= minBoxW) &&
		    (ABS (edge1->x - edge2->x) <= maxBoxW)))
		{
		    /* These appear to form the top and bottom of the box
		     * so seed fill them into a PIXR from which we construct
		     * the extraction mask.
		     */
                    ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, prTmp2, 
						    edge1->px, edge1->py, 1));
                    ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, prTmp2, 
						    edge2->px, edge2->py, 1));

                    /* grow the two interesting edges */
		    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState,
							prTmp2, prTmp3, 2, 
							&prTmp2));

                    /* dilate the edges slightly to make sure there are
		     * no corner gaps in the case where all 4 side are found.
		     */
		    ERR_PASS_CLEANUP (w_dilateVerticalRuns (pIPWState,
		                                            NULL,
							    prTmp2,
							    5,
							    2,
							    &prTmp2));

                    /* store the corners so we can draw in the horizontal
		     * edges if we didn't find them.
		     */
		    if (edge1->x > edge2->x)
		    {
		        edget = edge1;
			edge1 = edge2;
			edge2 = edget;
		    }

		    if (edge1->y < edge2->y)
		    {
			x0 = edge1->x;
			y0 = edge1->y;
			x1 = edge2->x;
			y1 = edge2->y;
			x2 = edge1->x + edge1->w - 1;
			y2 = edge1->y + edge1->h - 1;
			x3 = edge2->x + edge2->w - 1;
			y3 = edge2->y + edge2->h - 1;
		    }
		    else
		    {
			x0 = edge1->x + edge1->w - 1;
			y0 = edge1->y;
			x1 = edge2->x + edge2->w - 1;
			y1 = edge2->y;
			x2 = edge1->x;
			y2 = edge1->y + edge1->h - 1;
			x3 = edge2->x;
			y3 = edge2->y + edge2->h - 1;
		    }

		    vFound = cTrue;
#if DEBUG_CZ_P
                    fprintf (stdout, "Found vertical box sides.\n");
#endif
		    break;
		}
	    }

	    if (vFound)
	        break;

            edge1 = edge1->next;
        }
    }

    /* we're done with the vertical edge CCs */
    ERR_PASS_CLEANUP (wp_czFreeEdgeList (pIPWState, (clip_box_edge **)&(CCpacket.data)));

    /* prTmp2 now contains the vertical box sides if they were present */

    if (hFound || vFound)
    {
	/* Make the complete mask */
	if (hFound && vFound)
	{
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp2, OR, prSeed));
	}
	else
	{

	    /* Draw in the missing edges */
	    ERR_PASS_CLEANUP (w_pixrPutLine (pIPWState, prSeed, x0, y0, x1, y1, 1));

	    ERR_PASS_CLEANUP (w_pixrPutLine (pIPWState, prSeed, x2, y2, x3, y3, 1));

	    if (vFound)
	    {
		ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp2, OR, prSeed));
	    }
	}

#if DEBUG_CZ
        ERR_PASS_CLEANUP (w_write (pIPWState, "box", prSeed, TIFF_G3));
#endif

	/* Calculate the expected line thickness--note the use of
	 * _integer_ arithmetic.
	 */
	xThick = MAX (3, ((2 * ((xLineThickness * xResRed) / xRes)) - 1));
	yThick = MAX (3, ((2 * ((yLineThickness * yResRed) / yRes)) - 1));

	/* Dilate by the expected line thickness--note the use of 
	 *_integer_ arithmetic. 
	 */
	ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, xThick, yThick, 
					  xThick / 2, yThick / 2, &hStrel));
	ERR_PASS_CLEANUP (w_dilate (pIPWState, prSeed, hStrel, &prTmp2));
	ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &hStrel));

        /* make the reduced mask */
	ERR_PASS_CLEANUP (w_edgeFillSweep4 (pIPWState, prTmp2, 1, &prTmp1));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp1, INVERT, prTmp1));

        /* Copy the input to the output before masking if not in-place */
	if (prSrc != *pPrDest)
	{
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prRect, COPY, *pPrDest));
	}

	/* Apply the mask */
	ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, *pPrDest, prTmp1,
					   pPrDest));
#if DEBUG_CZ
        ERR_PASS_CLEANUP (w_write (pIPWState, "dest", *pPrDest, TIFF_G3));
#endif

    }
    else  /* we didn't find any two parallel box edges */
        ERR_GEN_CLEANUP(1, cErrNoLines);

    /* Destroy reduced pixr if we made it in this function */
    if ((xRes != 1) || (yRes != 1))
        ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSrcReduced));

    /* clean up the intermediate pixrs */
    if (prRect) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRect));
    if (prSeed) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSeed));
    if (prTmp1) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
    if (prTmp2) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp2));
    if (prTmp3) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp3));

    clipLoc->x = hRed * MIN (MIN (x0, x1), MIN (x2, x3));
    clipLoc->y = vRed * MIN (MIN (y0, y1), MIN (y2, y3));
    clipLoc->w = (hRed * MAX (MAX (x0, x1), MAX (x2, x3))) - clipLoc->x;
    clipLoc->h = (vRed * MAX (MAX (y0, y1), MAX (y2, y3))) - clipLoc->y;

#if DEBUG_CZ_P
    fprintf (stdout, "clipLoc - [%ld %ld %ld %ld]\n", clipLoc->x,
             clipLoc->y, clipLoc->w, clipLoc->h);
#endif

    if (clipToSize)
        ERR_PASS_CLEANUP (w_cutRectIP (pIPWState, clipLoc->x,
	                                          clipLoc->y,
	                                          clipLoc->w,
	                                          clipLoc->h,
						  *pPrDest));

    /* if subrect originally specified, add this offset as well */
    if (roi != NULL)
    {
        clipLoc->x += roi->x;
        clipLoc->y += roi->y;
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    /* Destroy reduced pixr if we made it in this function */
    if ((hRed != 1) || (vRed != 1))
        w_destroyPixr (pIPWState, &prSrcReduced);

    /* clean up the intermediate pixrs */
    if (prRect) w_destroyPixr (pIPWState, &prRect);
    if (prSeed) w_destroyPixr (pIPWState, &prSeed);
    if (prTmp1) w_destroyPixr (pIPWState, &prTmp1);
    if (prTmp2) w_destroyPixr (pIPWState, &prTmp2);
    if (prTmp3) w_destroyPixr (pIPWState, &prTmp3);

    /* clean up the edge list if we aborted */
    wp_czFreeEdgeList (pIPWState, (clip_box_edge **)&(CCpacket.data));

    ERR_RETURN;

}  /* extractClipZone () */

/*
 * $Log: clipzone.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 * Imported sources
 *
 * Revision 1.28  1998/01/14 00:10:57  ddavies
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.27  1997/06/26 20:00:32  ddavies
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Add inclusion of new file rast.pub.
 *
 * Revision 1.26  1997/04/15 21:42:52  ddavies
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Substitute NoInit (uninitialized) version of pixr creation routines where
 * possible.
 *
 * Revision 1.25  1996/05/10 22:10:19  dferrell
 * dferrell on Fri May 10 15:08:45 PDT 1996
 *
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Fixed problem with box edge size checks (was using < & > rather than
 * <= and >=).
 *
 * Revision 1.24  1996/05/10 18:09:50  dferrell
 * dferrell on Fri May 10 11:07:54 PDT 1996
 *
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Added new function wp_czReduceResolution () which had been a chunk
 * of inline code.  Improved reliability in cases where there is line
 * jitter or other distortion near the corner of a clip box.  Renamed
 * private functions with more conforming and pleasant style.
 *
 * Revision 1.23  1996/04/22 17:52:26  dferrell
 * dferrell on Mon Apr 22 10:52:14 PDT 1996
 *
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Expanded w_extractClipZone () interface for use by SmartPaper.
 *
 * Revision 1.22  1996/01/04  20:21:27  ejaquith
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.21  1995/12/13  18:44:55  dferrell
 * dferrell on Wed Dec 13 10:43:47 PST 1995
 *
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Wrapped debug-only code in error macros as a sacrifice to the nightly
 * nag.
 *
 * Revision 1.20  1995/12/12  20:24:10  dferrell
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Updated calls to run filters in w_extractClipZone ().
 *
 * Revision 1.19  1995/11/13  20:09:54  ejaquith
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.18  1995/10/16  22:54:27  dferrell
 * dferrell on Mon Oct 16 15:54:09 PDT 1995
 *
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.17  1995/10/13  21:58:57  ejaquith
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.16  1995/10/09  19:04:25  macleod
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Attempt to eliminate some NT compiler whinings
 *
 * Revision 1.15  1995/10/03  01:59:37  macleod
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.14  1995/10/02  00:23:46  ejaquith
 * ejaquith on Sun Oct 1 17:23:27 PDT 1995
 *
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * FF_TIFF_G3 ==> TIFF_G3
 *
 * Revision 1.13  1995/09/29  01:57:28  ejaquith
 * ejaquith on Thu Sep 28 18:57:15 PDT 1995
 *
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * TRUE => cTrue; FALSE => cFalse
 *
 * Revision 1.12  1995/09/21  22:51:49  ejaquith
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.11  1995/09/20  00:10:57  macleod
 * macleod on Tue Sep 19 17:10:33 PDT 1995
 *
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * Took out xRes and yRes paramters; compute xRes and yRes from the
 * scaled versions gotten out of the PIXR
 *
 * Revision 1.10  1995/09/19  19:19:44  macleod
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.9  1995/09/11  21:54:44  macleod
 * clipzone.c -> /product/ipcore/ipwrap/src/RCS/clipzone.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 * 
 */


