/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * cntr.c
 *
 * Region contouring functions:
 *		w_contourRegionBinary ()
 *		w_thinPolygon ()
 *
 *
 */

#include <math.h>


#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "bitacces.h"
#include "logicops.h"
#include "imageref.pub"

/* Prototypes */
#include "props.pub"
#include "rast.pub"
#include "shrscgr.pub"
#include "reduce.pub"
#include "cut.pub"
#include "logicops.pub"
#include "rectops.pub"
#include "geoutil.pub"
#include "cntr.pub"
#include "cntr.prv"
#include "fxpmath.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: cntr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

#define TESTING 0


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Contours an object (8-connected components) in an image
 *
 * LongDescription:::
 *    Contours an object (8-connected components) in an image.  Will
 *    optionally reduce the image first for faster, but less accurate
 *    operation.  Currently works only on binary images.
 *
 * PrivateDescription:::
 *  Notes on the algorithm::
 *  The heart of the contour tracing algorithm is a simple binary edge
 *  tracer.  In the imaginary binary bitmap representing the image which is
 *  being contoured we define direction to be a byte which contains
 *  exactly one bit which is a "1."  The direction is considered to be from
 *  the current pixel towards the ON bit.  The diagram below shows the
 *  direction encoding scheme.  Each number is the bit-place in the
 *  direction byte and X is the current pixel in the image:
 *
 *              bit-place               decimal value
 *            +---+---+---+             +---+---+---+
 *	      | 0 | 1 | 2 |             | 1 | 2 | 4 |
 *	      +---+---+---+             +---+---+---+
 *	      | 7 | X | 3 |             |128| X | 8 |
 *	      +---+---+---+             +---+---+---+
 *	      | 6 | 5 | 4 |             | 64| 32| 16|
 *	      +---+---+---+             +---+---+---+
 *
 *  So, for example, a direction towards the north-east would have a
 *  decimal value of 4 (the 2^2 bit being ON).
 *
 *  The first implementation of this algorithm may be a bit slower 
 *  than necessary.  The inner direction loop could be removed and
 *  replaced with successive if-elses until a new direction is found.
 *  Instead of tracking the x,y location in the virtual binary image, we
 *  could use a bit offset which would reduce the bit addressing arithmetic
 *  somewhat.  These optimizations will be made in the subsequent versions
 *  of the algorithm.
 *
 *  If gray region tracing is needed, it may be possible to use
 *  this algorithm after thresholding or gradient-finding then
 *  thresholding (or some other gray to binary edge preserving
 *  conversion) the original image.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          A binary pixr containing the region to be contoured.
 *  roi::            An optional rectangle specifying a region of interest.
 *  reduction::      Reduction factor (1, 2, 4, 8).
 *  filter::         Toggle for straight line point filtering :
 *		     0 = include all points, !0 = remove points on
 *                   straight lines in cardinal directions.
 *
 * Output:::
 *  polysize::       address of variable to hold size of polygon
 *		     created (size = # of points).
 *  polygon::        Address of polygon which is returned.  This must
 *		     be freed by the caller.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  The reduction sequences are not necessarily the best.
 *  They should be tailored to remove pepper and close smallish gaps
 *  on the region outline.
 *
 * FunctionCategory:::
 *  contour following
 *
 * Keywords:::
 *  contour following, polygon
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_contourRegionBinary (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *roi,
    Int32               reduction,
    Int32               filter,
    Int32              *polysize,
    ipPoint           **polygon)
{
    PROLOGUE;
    Int32     x_off,                /* x offset of rectangle */
	      y_off,                /* y offset of rectangle */
              w, h;
    Int8      valid;
    RECTANGL *roi_reduced = NULL;   /* reduced rect */
    PIXR     *prReduced = NULL;     /* prSrc after performing reduction */

    /* Validate the PIXR */
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* check for valid rect */
    if (roi != NULL)
    {
	ERR_PASS (w_isRectValid (pIPWState, prSrc, roi, &valid));
	ERR_GEN (!valid, cErrInvalidParm);
    }


    /* reduce image if necessary */
    switch (reduction)
    {
        case (1) :   /* no reduction */
	    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prSrc, &prReduced));
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prReduced));
	    break;

	case (2) :   /* 2-1 reduction */
	    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 0, 0, 0, 
					  &prReduced));
	    break;

	case (4) :  /* 4-1 reduction */
	    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 1, 0, 0, 
					  &prReduced));
	    break;

	case (8) :  /* 8-1 reduction */
	    ERR_PASS_CLEANUP (w_reduce (pIPWState, prSrc, 1, 1, 1, 0, 
					  &prReduced));
	    break;

	default :    /* invalid reduction specified */
	    ERR_GEN (1, cErrInvalidParm);

    }  /* switch on reduction */

#if TESTING
    writePr ("contour.reduced", prReduced, SUNRAS_COMP);
#endif

    /* cut out rectangle if specified */
    x_off = 0;
    y_off = 0;
    if (roi != NULL)
    {
	x_off = roi->x;
	y_off = roi->y;

        /* scale the rect and clip it to the image size */
	ERR_PASS_CLEANUP (w_xformRect (pIPWState, roi, 0, 0, 
				       (Float32)1.0 / (Float32)reduction, 
				       (Float32)1.0 / (Float32)reduction, 
				       &roi_reduced));
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prSrc, &w));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prSrc, &h));
        roi_reduced->w = MIN (roi_reduced->w, w - roi_reduced->x);
        roi_reduced->h = MIN (roi_reduced->h, h - roi_reduced->y);

        ERR_PASS_CLEANUP (w_cutRectIP (pIPWState, 
				       roi_reduced->x, 
				       roi_reduced->y, 
				       roi_reduced->w, 
				       roi_reduced->h,
				       prReduced));

	IPW_FREE (roi_reduced, cAllocatorData);

#if TESTING
        writePr (pIPWState, "contour.reduced.cut", prReduced, SUNRAS_COMP);
#endif
    }

    /* connected component analysis? */

    /* contour the object */
    ERR_PASS_CLEANUP (wp_cntrTraceObject (pIPWState, prReduced, reduction, 
				       x_off, y_off, filter, polygon, 
				       polysize));

    /* clean up */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prReduced) w_destroyPixr (pIPWState, &prReduced);
    if (roi_reduced) IPW_FREE (roi_reduced, cAllocatorData);
    ERR_RETURN;

}  /* w_contourRegionBinary () */


/* constants & defs */
#define ROTL_BYTE(b,s)      (((b) << s) | ((b) >> (8 - s)))
#define ROTR_BYTE(b,s)      (((b) >> s) | ((b) << (8 - s)))
#define CNTR_MORE_POINTS    0x80

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Given an edge pixel, trace the object's edge and return a polygon.
 *
 * LongDescription:::
 *     Given a starting location on a binary image which is an edge pixel
 *  (i.e. the pixel is ON and has at least one neighbor which is OFF), trace
 *  the object's edge and return the polygon (scaled) which describes the
 *  outline.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to trace
 *  reduction::      Reduction factor
 *  x_off::          X co-ordinate of edge point
 *  y_off::          Y co-ordinate of edge point
 *  filter::         Filter points
 *
 * Output:::
 *  polygon::        Polygon created
 *  polysize::       Size of polygon created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  contour following
 *
 * Keywords:::
 *  contour following, polygon
 *
 * Level:::
 *  deterministic
 *  heuristic
 *  directive
 *
 * Maturity:::
 *  work-in-progress
 *  stable
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_cntrTraceObject (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32               reduction,
    Int32               x_off,
    Int32               y_off,
    Int32               filter,
    ipPoint           **polygon,
    Int32              *polysize)
{
    PROLOGUE;
    UInt8   dir,         /* direction */
            ldir;        /* last direction */
    Int32   done,        /* closed polygon flag */
            found;       /* found a new direction flag */
    UInt32 *data;        /* pointer to start of "real" image data */
    ipPoint *cur_point;  /* index into polygon */
    Int32   bpl,         /* bits/line in the pixr */
            x,           /* x location in virtual binary image */
	    y,           /* y "                              " */
	    sx,          /* starting x loc in virtual binary image */
	    sy,          /* starting y "                         " */
	    w,           /* width of image in pixels */
	    h,           /* height of image in pixels */
	    i,           /* direction count index */
            vertices,    /* number of polygon vertices found */
	    max_points;  /* max points alloced for polygon */

/* this macro is used to lookup binary data in the packed binary pixr given
   x, y in virtual binary image coordinates
 */
#define PIX_IS_ON(x,y) GET_IMAGE_BIT (data, (Int32)(((Int32)(x) + ((Int32)(y) * (Int32)bpl))))

    x_off += reduction >> 1;  /* center polygon points in reduced pixels */
    y_off += reduction >> 1;
    *polygon = NULL;
    *polysize = 0;
    vertices = 0;
    ERR_PASS (w_pixrGetBpl (pIPWState, prSrc, &bpl));
    bpl <<= 3;
    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &h));
    ERR_PASS (w_pixrGetImage (pIPWState, prSrc, (UInt8 **) &data));

    /* find the edge of the blob */
    dir = 8;
    found = cFalse;
    for (y = 0, x = 0; y < h; y++)	/* init x here to keep gcc happy */
    {
        for (x = 0; x < w; x++)
	{
	    if (PIX_IS_ON (x, y))
	    {
	        found = cTrue;
		break; 
            }
	}
	if (found)
	    break; 
    }

    ERR_GEN (!found, cErrInternal);

    /* allocate polygon space */
    max_points = 4 * (w + h);
    *polygon = (ipPoint *)IPW_MALLOC (max_points * sizeof (ipPoint),
				      cAllocatorData);

    ERR_GEN (*polygon == NULL, cErrNomem);

    /* trace the blob */
    done = cFalse;
    cur_point = *polygon;
    sx = x;
    sy = y;

    /* add the start point to the polygon */
    cur_point->x = (x * reduction) + x_off;
    cur_point->y = (y * reduction) + y_off;
#if TESTING
    fprintf (stderr, "%d->(%d, %d)\n", dir, cur_point->x, cur_point->y);
#endif
    cur_point++;
    vertices++;

    while (!done)
    {
        /* allocate more polygon space if necessary */
        if (vertices == max_points)
	{
	    *polygon = (ipPoint *)
		IPW_REALLOC ((void *)(*polygon), 
	     (size_t)((max_points + CNTR_MORE_POINTS) * sizeof (ipPoint)),
			     cAllocatorData);
	    ERR_GEN_CLEANUP (*polygon == NULL, cErrNomem);
	    max_points += CNTR_MORE_POINTS;
	    cur_point = *polygon + vertices;
	}

        /* store the current direction for filtering */
        ldir = dir;

	/* start looking 135 degrees clockwise from the direction
	 * which we last traveled.
	 */
        dir = ROTL_BYTE (dir, 3);
	found = cFalse;
	/* look in each of the 8-adjacent bits sequentially until
	 * another on bit is found.  Move to the frst on bit found.
	 */
        for (i = 0; i < 8; i++)
	{
#if TESTING
	    fprintf (stderr, "%d->", dir);
#endif
            switch (dir)
	    {
	        case 1:
		    if (PIX_IS_ON (x-1, y-1))
		    {
		        x--;
			y--;
                        found = cTrue;
		    }
		    break;

	        case 2:
		    if (PIX_IS_ON (x, y-1))
		    {
		        y--;
                        found = cTrue;
		    }
		    break;

	        case 4:
		    if (PIX_IS_ON (x+1, y-1))
		    {
		        x++;
			y--;
                        found = cTrue;
		    }
		    break;

	        case 8:
		    if (PIX_IS_ON (x+1, y))
		    {
		        x++;
                        found = cTrue;
		    }
		    break;

	        case 16:
		    if (PIX_IS_ON (x+1, y+1))
		    {
		        x++;
			y++;
                        found = cTrue;
		    }
		    break;

	        case 32:
		    if (PIX_IS_ON (x, y+1))
		    {
		        y++;
                        found = cTrue;
		    }
		    break;

	        case 64:
		    if (PIX_IS_ON (x-1, y+1))
		    {
		        x--;
			y++;
                        found = cTrue;
		    }
		    break;

	        case 128:
		    if (PIX_IS_ON (x-1, y))
		    {
		        x--;
                        found = cTrue;
		    }
		    break;

                default:
		    /* this shouldn't happen */
                    break;

	    }
	    if (found)
	        break;   /* exit the direction loop */
	    else
	        dir = ROTR_BYTE (dir, 1);  /* "increment" the direction */

	}   /* for each of the 8 directions */

	if (found)
	{
	    /* check for end of polygon */
	    if ((x == sx) && (y == sy))
	       done = cTrue;

            /* Only add points when necessary:
	     * We always add the last point to close the polygon.
	     * If we're filtering points, then only add points
	     * when the direction changes.
	     */
	    if (done || !(filter && (ldir == dir)))
            {
                /* add the point to the polygon */
                cur_point->x = (x * reduction) + x_off;
	        cur_point->y = (y * reduction) + y_off;
#if TESTING
                fprintf (stderr, "(%d, %d)\n", dir, cur_point->x, cur_point->y);
#endif
	        cur_point++;
	        vertices++;
	    }

	}
	else    /* this is bad and should happen only when we hit pepper */
	    done = cTrue;

    }   /* while we haven't closed the polygon */

    /* make the polygon memory chunk only as large as it needs to be */
    if (vertices < max_points)
    {
        *polygon = (ipPoint *) 
	    IPW_REALLOC (*polygon, vertices * sizeof (ipPoint),
			 cAllocatorData);

	ERR_GEN_CLEANUP (*polygon == NULL, cErrNomem);
    }

    *polysize = vertices;

    NORMAL_RETURN;

ERR_CLEANUP:
    /* Actually, the only error that happens is that REALLOC fails,
       in which case the polygon goes away (?) anyway */
    if (*polygon)
	IPW_FREE (*polygon, cAllocatorData);

    ERR_RETURN;

}  /* wp_cntrTraceObject () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Thin an array of points which define a closed polygon
 *
 * LongDescription:::
 *  Given an array of points which define a closed polygon, 
 *  thin the polygon by removing vertices which are within
 *  maxMisAlignment of being colinear between their two adjacent
 *  vertices. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::        State structure
 *  poly::             Array of points for input polygon
 *  polySize::         Length of input polygon array
 *  maxMisAlignment::  Maximum allowable deviation (in pixels) from
 *		       colinearity for removal of points.
 *
 * Output:::
 *  thinPoly::         Thinned polygon
 *  thinPolySize::     Length of thinned polygon array
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  polygon
 *
 * Keywords:::
 *  polygon
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *  stable
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_thinPolygon (
    IPWState           *pIPWState,
    ipPoint            *poly,
    Int32               polySize,
    Int32               maxMisAlignment,
    ipPoint           **thinPoly,
    Int32              *thinPolySize)
{
    PROLOGUE;
    UInt8   *inflectionFlag = NULL;
    Int32    retCode, found,
             x, y, dx, dy,
             p1, p2, pC;

    retCode = cSuccessful;
    found = cFalse;
    *thinPoly = NULL;

    if (polySize < 4)
    {
        (*thinPoly) = (ipPoint *)IPW_CALLOC (sizeof (ipPoint), 
					     polySize, cAllocatorData);
	ERR_GEN (*thinPoly == NULL, cErrNomem);
	MEMCPY ((*thinPoly), (poly), polySize * sizeof (ipPoint));
	*thinPolySize = polySize;
    }
    else
    {

	inflectionFlag = (UInt8 *) IPW_CALLOC (sizeof (UInt8), 
					       polySize, cAllocatorData);

	ERR_GEN (inflectionFlag == NULL, cErrNomem);

	p1 = 0;
	p2 = 2;
	pC = 1;

	inflectionFlag[p1] = cTrue;
	*thinPolySize = 1;
        maxMisAlignment *= maxMisAlignment;


	/* Loop thru the entire polygon checking for non-colinearities */
	while (p2 < polySize)
	{
	    /* Check all the points between p1 & p2 for approximate
	     * colinearity.
	     */
	    for (pC = p1 + 1; pC < p2; pC++)
	    {
		ERR_PASS_CLEANUP (w_geo_pointOnLineSeg (
		    pIPWState, poly[p1].x, poly[p1].y,
		    poly[p2].x, poly[p2].y,
		    poly[pC].x, poly[pC].y, &x, &y));

		/* Use squared distance for computational efficiency */
		dx = ABS (poly[pC].x - x);
		dy = ABS (poly[pC].y - y);
		if (((dx * dx) + (dy * dy)) > maxMisAlignment)
		{
		    inflectionFlag[p2 - 1] = cTrue;
		    p1 = p2 - 1;
		    pC = p1 + 1;
		    p2 = pC + 1;
		    found = cTrue;
		    (*thinPolySize)++;
		    break;
		}
	    }

	    if (!found)
		p2++;
	    else
		found = cFalse;
	}

	(*thinPolySize)++;
	inflectionFlag[pC] = cTrue;

	/* Construct the thinned polygon point array */
	(*thinPoly) = (ipPoint *)IPW_CALLOC (sizeof (ipPoint),
					     *thinPolySize, cAllocatorData);
	ERR_GEN_CLEANUP (*thinPoly == NULL, cErrNomem);

	/* insert each inflection point into the thinned polygon array */
	for (p1 = 0, p2 = 0; p1 < polySize; p1++)
	{
	    if (inflectionFlag[p1])
	    {
		((*thinPoly)[p2]).x = poly[p1].x;
		((*thinPoly)[p2]).y = poly[p1].y;
		p2++;
	    }
	}

	IPW_FREE (inflectionFlag, cAllocatorData);
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (inflectionFlag)
	IPW_FREE (inflectionFlag, cAllocatorData);

    ERR_RETURN;
}  /* thinPolygon () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reduce a polygon to a bounding rectangle
 *
 * LongDescription:::
 *  Given an array of points which define a closed polygon, 
 *  thin the polygon by attempting to locate the best rectanglular
 *  region that encompasses the entire polygon.
 *
 * PrivateDescription:::
 *  1. Precompute the lengths of each segment, and the directions.
 *     Directions will be: 0 = horizontal, 16384 = straight up (90 degrees),
 *     and 32767 = almost 180 degrees.
 *  2. Find the total of the segment lengths.
 *  3. Categorize each of the segments into 5 degree increments. The two
 *     adjacent bins with matching 90 degree adjacent bins totalling the
 *     largest combined segment length, is used to compute the initial
 *     perpendiculars. Perpendiculars are computed as the weighted averages
 *     of both angles, folding the 90 degree off angles back to the the
 *     more horizontal, before computing the average.
 *  4. Categorize each of the segments for a match to one of the two
 *     perpendiculars (angle deviation < threshold).
 *  5. Recompute the perpendiculars as the weighted averages of the
 *     segments assigned to the two perpendiculars.
 *  6. Repeat #4 & #5 until skew angle deviates < threshold.
 *  7. If segment lengths categorized to the two perpendiculars is less
 *     than some fraction of the combined length, loosen threshold in #4
 *     and repeat.
 *  8. Use result as the estimated perpendiculars.
 *  9. Traverse the polygon, reaccumulate segments in the direction of
 *     perpendiculars when the segments approach the corners (computed
 *     from estimated perpendiculars and min/max X & Y). When segments
 *     inbetween those tending towards corners aren't within threshold,
 *     remove them.
 * 10. Recompute new perpendiculars from thinned polygon.
 * 11. If new segment contributions < threshold, loosen angle restriction
 *     in #9 and repeat, until > 50%.
 *
 * Input:::
 *  pIPWState::		State structure
 *  poly::		Array of points for input polygon
 *  polySize::		Length of input polygon array
 *  flags::		Rectangle placement (none currently defined)
 *
 * Output:::
 *  pRadianAngle::	Resulting angle in radians (-pi/2 to pi/2)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  polygon
 *
 * Keywords:::
 *  polygon
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *  somewhate stable
 *  partially released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findRectFromPolygon (
    IPWState           *pIPWState,
    ipPoint            *poly,
    Int32               polySize,
    Int32               flags,
    Float64		*pRadianAngle
)
{
    ipPoint *pt1, *pt2;		/* Points in segment loop */
    ipPoint *upperPt, *lowerPt;	/* Upper and lower points */
    Int32 segment, segments;	/* Number of polygon segments */
    UInt32 *segLengths, *len;	/* Segment lengths */
    UInt16 *segAngles, *ang;	/* Segment angles (0-180 normalized) */
    UInt32 distance;		/* Distance between points */
    ScaledAngle angle;		/* Angle between points */
    UInt32 minX, maxX, minY, maxY;  /* Bounding rectangle */
    UInt32 max_length;		/* Maximum length of all segments */
    UInt32 tot_length;		/* Total length of all segments */
    UInt32 bin, bins;		/* Bin number and count */
    UInt32 bina, binb, binc, bind;  /* Bin numbers for adjacency */
    UInt32 *angleBins;		/* Bin array */
    float divisor;		/* For bin assignment calculation */
    UInt32 maxBinNum;		/* Bin with the peak */
    UInt32 maxBinCount;		/* Peak bin count */
    UInt32 matchCount;		/* Matching perpendiculars length total */
    UInt32 prevEstAngle, newEstAngle;	/* Previous and new estimates */
    Int32 prevScaledAngle, newScaledAngle;	/* Previous and new scaled */
    Int32 diff, diffThresh;	/* Scaled angle difference and threshold */
    UInt32 primLength, perpLength; /* Within tolerance segment lengths */
    UInt32 matchLength, prevMatchLength; /* Length of within tolerance totals */
    Int32 primX, primY;		/* Accumulated primary X/Y directions */
    Int32 perpX, perpY;		/* Accumulated perpendicular X/Y directions */
    Bool repeat;
    UInt32 iterations;		/* Number of loop iterations */
    Int32 rotateAngle;		/* Final rotation angle (scaled) */
    PROLOGUE;

#define UPPER_N_LOWER(pp1, pp2)				\
    if (pp1->y < pp2->y) {		/* p1 higher*/	\
	upperPt = pp1, lowerPt = pp2;			\
    } else if (pp1->y > pp2->y) {	/* p2 higher */	\
	upperPt = pp2, lowerPt = pp1;			\
    } else if (pp1->x < pp2->x) {	/* p1 to left */\
	upperPt = pp2, lowerPt = pp1;			\
    } else {				/* p2 to left */\
	upperPt = pp1, lowerPt = pp2;			\
    }
#define SQUARE(a)	((a)*(a))
#define SEG_LENGTH(upperPt, lowerPt, distPtr)		\
    isqrt(pIPWState, SQUARE(upperPt->x - lowerPt->x) + SQUARE(upperPt->y - lowerPt->y), \
	&distance);					\
    *distPtr = distance;
#define SEG_ANGLE(upperPt, lowerPt, anglePtr)		\
    iatan2(pIPWState, upperPt->x - lowerPt->x, lowerPt->y - upperPt->y, &angle); \
    *anglePtr = (UInt16)angle;
#define SCALED_ANGLE(a)	    ((a) * 16384 / 180)
#define UNSCALED_ANGLE(a)   ((a) * 180 / 16384)
#define SCALED_TO_RADIANS(a)	(Float32)((Float64)(a) * acos(-1.0) / 16384.0)

#define BIN_ANGLE	5	/* 5 degrees per bin */

    /* Return nothing initially */
    *pRadianAngle = (Float32)0.0;

    /* Initialize */
    segLengths = NULL; segAngles = NULL;

    /* Check input */
    ERR_GEN_CLEANUP(flags != 0, cErrInvalidParm);

    /* 1. Precompute the lengths of each segment, and the directions.
     * Directions will be: 0 = horizontal, 16384 = straight up (90 degrees),
     * and 32767 = almost 180 degrees.
     */
    
    /* Prepare the arrays for segment lengths and angles */
    segments = polySize;
    segLengths = (UInt32 *)IPW_CALLOC(sizeof(*segLengths), 
	segments, cAllocatorData);
    ERR_GEN_CLEANUP(segLengths == NULL, cErrNomem);
    segAngles = (UInt16 *)IPW_CALLOC(sizeof(*segAngles), 
	segments, cAllocatorData);
    ERR_GEN_CLEANUP(segAngles == NULL, cErrNomem);

    /* Loop through each point pair, and compute the Euclidean distance
     * between points, and the upward angle (0-180) */
    pt1 = poly, pt2 = pt1 + 1;
    len = segLengths;
    ang = segAngles;
    minX = maxX = pt1->x;
    minY = maxY = pt1->y;
    for (segment = 1; segment < segments; segment++) {
      UPPER_N_LOWER(pt1, pt2);
      SEG_LENGTH(upperPt, lowerPt, len);
      SEG_ANGLE(upperPt, lowerPt, ang);
      if (upperPt->y < minY) {
	  minY = upperPt->y;
      }
      if (lowerPt->y > maxY) {
	  maxY = upperPt->y;
      }
      if (upperPt->x < minX) {
	  minX = upperPt->x;
      }
      if (upperPt->x > maxX) {
	  maxX = upperPt->x;
      }
      if (lowerPt->x < minX) {
	  minX = lowerPt->x;
      }
      if (lowerPt->x > maxX) {
	  maxX = lowerPt->x;
      }
      pt1++, pt2++, len++, ang++;
    }

    /* Handle last segment specially for wrap around */
    pt2 = poly;
    UPPER_N_LOWER(pt1, pt2);
    SEG_LENGTH(upperPt, lowerPt, len);
    SEG_ANGLE(upperPt, lowerPt, ang);

    
    /*  2. Find the total of the segment lengths.
     */
    len = segLengths;
    ang = segAngles;
    tot_length = max_length = 0;
    for (segment = 0; segment < segments; segment++) {
      if (*len > max_length) {
	max_length = *len;
      }
      tot_length += *len;
      len++;
    }


    /* 3. Categorize each of the segments into 5 degree increments. The two
     * adjacent bins with matching 90 degree adjacent bins totalling the
     * largest combined segment length, is used to compute the initial
     * perpendiculars. Perpendiculars are computed as the weighted averages
     * of both angles, folding the 90 degree off angles back to the the
     * more horizontal, before computing the average.
     */

    /* Place the segments lengths into the appropriate bins */
    bins = (180 / BIN_ANGLE) & ~1;	/* Make it even too */
    angleBins = (UInt32 *)IPW_CALLOC(sizeof(*angleBins), 
	bins, cAllocatorData);
    ERR_GEN_CLEANUP(angleBins == NULL, cErrNomem);
    len = segLengths;
    ang = segAngles;
    divisor = SCALED_ANGLE(180) / (float)bins;
    for (segment = 0; segment < segments; segment++) {
      bin = (UInt32)((float)*ang++ / divisor);
      if (bin >= bins) {
	bin = bins - 1;
      }
      angleBins[bin] += *len++;
    }

    /* Search for the best matching perpendiculars */
    maxBinCount = 0;
    maxBinNum = 0;
    for (bin = 0; bin < (bins >> 1); bin++) {
      /* Use the previous bin also */
      bina = bin, binb = bin + (bins >> 1);
      if (bin == 0) {
	  binc = (bins >> 1) - 1, bind = bins - 1;
      } else {
	  binc = bin - 1, bind = bin + (bins >> 1) - 1;
      }
      matchCount = angleBins[bina] + angleBins[binb] +
	  angleBins[binc] + angleBins[bind];
      if (matchCount > maxBinCount) {
	maxBinCount = matchCount;
	maxBinNum = bin;
      }
    }

    /* Estimate the perpendiculars */
    newEstAngle = maxBinNum * 180 / bins;
    newScaledAngle = SCALED_ANGLE(newEstAngle);

    /* Start with 5 degrees of allowable deviation */
    diffThresh = SCALED_ANGLE(5);
    matchLength = 0;
    iterations = 0;	/* Keep track of iteration */
    do {
	prevMatchLength = matchLength;
	prevEstAngle = newEstAngle;
	prevScaledAngle = newScaledAngle;

        /* 4. Categorize each of the segments for a match to one of the two
	 * perpendiculars (angle deviation < threshold).
	 */
	len = segLengths;
	ang = segAngles;
	primLength = perpLength = 0;
	primX = primY = perpX = perpY = 0;
	for (segment = 0; segment < segments; segment++) {
	    
	    /* Check primary perpendicular angle */
	    diff = *ang - newScaledAngle;
	    if (ABS(diff) < diffThresh) {
		primLength += *len;
		pt1 = poly + segment;
		if (segment == segments - 1) {
		    pt2 = poly;
		} else {
		    pt2 = pt1 + 1;
		}
		UPPER_N_LOWER(pt1, pt2);
		primX += upperPt->x - lowerPt->x;
		primY += lowerPt->y - upperPt->y;   /* Flip from raster to
						     * quadrant */
	    } else {

		/* Check secondary perpendicular angle */
		diff = *ang - (newScaledAngle + SCALED_ANGLE(90));
		if (ABS(diff) < diffThresh) {
		    perpLength += *len;
		    pt1 = poly + segment;
		    if (segment == segments - 1) {
			pt2 = poly;
		    } else {
			pt2 = pt1 + 1;
		    }
		    UPPER_N_LOWER(pt1, pt2);
		    perpX += upperPt->x - lowerPt->x;
		    perpY += lowerPt->y - upperPt->y;   /* Flip from raster to
							 * quadrant */
		}
	    }
	    ang++, len++;
	}


	/* 5. Recompute the perpendiculars as the weighted averages of the
	 * segments assigned to the two perpendiculars.
	 */

	/* Fold the perpendicular 90 degress onto the primary direction,
	 * then compute the combined contribution in this direction */
	iatan2(pIPWState, primX + perpY, primY - perpX, &newScaledAngle);
	newEstAngle = UNSCALED_ANGLE(newScaledAngle);

	
	/* 6. Repeat #4 & #5 until skew angle deviates < threshold.
	 */

	/* Compute the contributing lengths */
	matchLength = primLength + perpLength;

	/* If the skew angle sufficiently resolved, or we've iterated to much
	 * (possibly due to an unstable algorithm, who knows) */
	if (++iterations > 20 || diffThresh < SCALED_ANGLE(0.5) /* 45 */) {
	    repeat = cFalse;

	/* We're still looking for the dominant angles */
	} else if (matchLength < (tot_length >> 1)) {

	    /* If the angle has changed, leave the threshold alone and try again
	     */
	    if (newScaledAngle != prevScaledAngle) {
		repeat = cTrue;

    /* 7. If segment lengths categorized to the two perpendiculars is less
     * than some fraction of the combined length, loosen threshold in #4
     * and repeat.
     */

	    /* If we are making progress, or we haven't found any yet,
	     * loosen the threshold */
	    } else if (matchLength > prevMatchLength || prevMatchLength == 0) {
		diffThresh = diffThresh * 110 / 100;
		repeat = cTrue;
	    } else {
		repeat = cFalse;
	    }
	} else {

	    /* If seemingly stable, tighten threshold quite a bit */
	    if (newScaledAngle == prevScaledAngle) {
		diffThresh = diffThresh * 50 / 100;
	    } else {

		/* Tighten threshold a bit and repeat */
		diffThresh = diffThresh * 90 / 100;
	    }
	    repeat = cTrue;
	}
    } while (repeat);


    /* 8. Use result as the estimated perpendiculars.
     */

    /* Determine the smallest angle needed to make the rectangle upright */
    if (newScaledAngle > SCALED_ANGLE(45)) {
	rotateAngle = newScaledAngle - SCALED_ANGLE(90); /* Use -pi/2 to pi/2
							  * range */
    } else {
	rotateAngle = newScaledAngle;		    /* Use -pi/2 to pi/2 range */
    }
    *pRadianAngle = SCALED_TO_RADIANS(rotateAngle);   /* Convert to radians */


     
    /* 9. Traverse the polygon, reaccumulate segments in the direction of
     * perpendiculars when the segments approach the corners (computed
     * from estimated perpendiculars and min/max X & Y). When segments
     * inbetween those tending towards corners aren't within threshold,
     * remove them.
     */

    /* ADD CODE HERE */ 
     

    /*10. Recompute new perpendiculars from thinned polygon.
     */

    /* ADD CODE HERE */ 
     
     
    /*11. If new segment contributions < threshold, loosen angle restriction
     * in #9 and repeat, until > 50%.
     */

     

    /* Free the temporary data */
    IPW_FREE (segLengths, cAllocatorData);
    IPW_FREE(segAngles, cAllocatorData);
    
    NORMAL_RETURN;

ERR_CLEANUP:
    if (segLengths) {
      IPW_FREE (segLengths, cAllocatorData);
    }
    if (segAngles) {
      IPW_FREE(segAngles, cAllocatorData);
    }

    ERR_RETURN;
}  /* rectPolygon () */
   

    
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate a polygon about a point
 *
 * LongDescription:::
 *  Given an array of points which define a closed polygon, compute the
 *  radial distance and angle from a fixed point to each polygon point,
 *  add an angle delta to each, then convert back to points of a new
 *  polygon (rotated).
 *
 * Input:::
 *  pIPWState::		State structure
 *  pivotX::		X coordinate to pivot about
 *  pivotY::		Y coordinate to pivot about
 *  poly::		Array of points for input polygon
 *  polySize::		Length of input polygon array
 *  deltaAngle::	Angle to rotate the polygon points about (in radians,
 *			clockwise direction)
 *
 * Output:::
 *  pOutPoly::		Rotated polygon (same number of points as input, user
 *			is responsible for freeing this data)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  polygon
 *
 * Keywords:::
 *  polygon, rotation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *  stable
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotatePolygon (
    IPWState           *pIPWState,
    Int32		pivotX,
    Int32		pivotY,
    ipPoint            *poly,
    Int32               polySize,
    Float64		deltaAngle,
    ipPoint           **pOutPoly)
{
    ipPoint *outPoly = NULL;	/* Output polygon list */
    ipPoint *ptIn, *ptOut;	/* Polygon list pointers */
    Int32 point;		/* Point offset in loop */
    Int32 dx, dy;		/* Delta X/Y */
#ifdef USE_FIXED
    UInt32 square;		/* Distance between points */
    ScaledAngle angle;		/* Angle between points */
    ScaledInt1715 scaledDelta;	/* Delta angle scaled up */
    ScaledInt1715 sq_mant, dist_mant;	/* Mantissas */
    ScaledInt1715 scaledCos, scaledSin;	/* Scaled cos()/sin() */
    UInt32 sq_exp, dist_exp;	/* Exponents */
#else
    Float64 square;
    Float64 dist;
    Float64 angle;
#endif

    PROLOGUE;

    /* Return nothing initially */
    *pOutPoly = NULL;

    /* Allocate the polygon points to be returned */
    outPoly = (ipPoint *)IPW_CALLOC(sizeof(*outPoly), 
	polySize, cAllocatorData);
    ERR_GEN_CLEANUP(outPoly == NULL, cErrNomem);

    /* Adjust the angle for "unit circle direction" (counter-clockwise) */
    deltaAngle = -deltaAngle;

    /* Loop through each point, compute the square of the distance
     * to the referecence, and the angle from (1,0). Then add the
     * angle delta, and compute the new point using the new angle
     * and square of the radial distance. */
    ptIn = poly, ptOut = outPoly;
#ifdef USE_FIXED
    scaledDelta = (ScaledInt1715)(deltaAngle * 32768.0);
#endif
    for (point = 0; point < polySize; point++) {

      /* Compute the distance between pivot and point */
      dx = pivotX - ptIn->x;
      dy = pivotY - ptIn->y;
#ifdef USE_FIXED
      sq_mant = square = dx * dx + dy * dy;
      sq_exp = 15;
      while (sq_mant & 0xffff8000) {
	  sq_mant >>= 1;
	  sq_exp++;
      }
      ERR_CHECK(fract_sqrt(sq_mant, sq_exp, &dist_mant, &dist_exp));
#else
      square = dx * dx + dy * dy;
      dist = sqrt(square);
#endif

      /* Compute the angle from the pivot to the point */
      if (ptIn->y == pivotY) {
	if (ptIn->x > pivotX) {
#ifdef USE_FIXED
	  angle = 0;
#else
	  angle = 0.0;
#endif
	} else if (ptIn->x < pivotX) {
#ifdef USE_FIXED
	  angle = SCALED_ANGLE(180);
#else
	  angle = acos(-1.0);	/* pi */
#endif
	} else {	/* ptIn->x == pivotX */
#ifdef USE_FIXED
	  angle = 0;	/* Point will rotate in place */
#else
	  angle = 0.0;	/* Point will rotate in place */
#endif
	}
      } else {
#ifdef USE_FIXED
	ERR_CHECK(iatan2(ptIn->x - pivotX, pivotY - ptIn->y, &angle));
#else
	angle = atan2((Float64)(pivotY - ptIn->y), (Float64)(ptIn->x - pivotX));
#endif
      }

#ifdef USE_FIXED
      /* Increase the angle by the delta */
      angle += scaledDelta;

      /* Compute the point from the new angle and the distance */
      ERR_CHECK(fract_cos(angle, &scaledCos));
      ERR_CHECK(fract_sin(angle, &scaledSin));
      ptOut->x = pivotX + ((dist_mant * scaledCos / 32768) >> dist_exp);
      ptOut->y = pivotY + ((dist_mant * scaledSin / 32768) >> dist_exp);
#else
      /* Increase the angle by the delta */
      angle += deltaAngle;

      /* Compute the point from the new angle and the distance */
      ptOut->x = pivotX + (Int32)(dist * cos(angle) + 0.5);
      ptOut->y = pivotY - (Int32)(dist * sin(angle) + 0.5);
#endif

      ptIn++, ptOut++;
    }

    /* Return the polygon */
    *pOutPoly = outPoly, outPoly = NULL;
    
    NORMAL_RETURN;

ERR_CLEANUP:
    if (outPoly) {
      IPW_FREE (outPoly, cAllocatorData);
    }
    ERR_RETURN;
}  /* w_rotatePolygon () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Given an array of points which define a polygon, draw it.
 *
 * LongDescription:::
 *	Given an array of points which define a polygon, draw it.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      
 *  poly::           Array of points for input polygon
 *  polySize::       Length of input polygon array
 *  close::          cTrue = draw segment from p0 to pn,
 *		     cFalse = draw only explicit segments
 *  markVertices::   cTrue = put small rect at each vertex,
 *		     cFalse = no marks at vertices.
 *
 * Output:::
 *  outPr::          PIXR to draw in
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  If the polygon extends beyond the edge of the image, the
 *  boundary is computed, and drawn to correctly.
 *
 * FunctionCategory:::
 *  polygon
 *
 * Keywords:::
 *  polygon
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_drawPolygon (
    IPWState           *pIPWState,
    ipPoint            *poly,
    Int32               polySize,
    Int32               close,
    Int32               markVertices,
    PIXR               *outPr)
{
    PROLOGUE;
    Int32	p0, p1;
    Int32	channels;
    RECTANGL	vert;

    ERR_PASS (w_pixrGetChannelCount(pIPWState, outPr, &channels));

    if (channels > 1) {
	PIXR *subPr;
	Int32 channel;

	for (channel = 0; channel < channels; channel++) {
	    subPr = NULL;
	    ERR_PASS (w_pixrGetChannel(pIPWState, outPr, channel, &subPr));
	    ERR_PASS (w_drawPolygon(pIPWState, poly, polySize, close,
		markVertices, subPr));
	}

	NORMAL_RETURN;
    }

    /* The size of the vertex marker */
    vert.w = 2;
    vert.h = 2;

    for (p0 = 0, p1 = 1; p1 < polySize; p0++, p1++)
    {
        if (markVertices)
	{
	    vert.x = poly[p0].x;
	    vert.y = poly[p0].y;
	    ERR_PASS (w_pixrPutRect (pIPWState, outPr, &vert, 1, 1));
	}

        ERR_PASS (w_pixrPutLine (pIPWState, outPr, poly[p0].x, poly[p0].y, 
			       poly[p1].x, poly[p1].y, 1));
    }

    if (close)
        ERR_PASS (w_pixrPutLine (pIPWState, outPr, poly[0].x, poly[0].y, 
			       poly[p0].x, poly[p0].y, 1));

    NORMAL_RETURN;

}  /* drawPolygon () */


/*

$Log: cntr.c,v $
Revision 1.1.1.1  1999/09/20 18:19:15  jlin
Imported sources

Revision 1.32  1998/01/31 01:01:18  aber
cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

Use clock-wise direction as positive angle, for polygon routines. Matches
rotate routines.
Use Float64 vs. Float32, to match angle type propogated up through w_crop5(),
which will match the Float64 type returned from w_crop4() (just to be
consistent).
Extend polygon drawing routine to handle multiple-channel images.

Revision 1.31  1998/01/20 18:06:29  aber
cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

Fix sign check, causes rotation to be 90 degrees off.

Revision 1.30  1998/01/14 00:10:59  ddavies
cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.29  1997/08/26 16:30:37  aber
aber on Tue Aug 26 09:30:25 PDT 1997

cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

Okay, now fix sun warning, geez!

Revision 1.28  1997/08/26 01:13:07  aber
aber on Mon Aug 25 18:13:01 PDT 1997

cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

Now fix solaris warnings.

Revision 1.27  1997/08/26 01:11:41  aber
aber on Mon Aug 25 18:11:33 PDT 1997

cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

Fix watcom warnings.

Revision 1.26  1997/08/26 01:03:08  aber
aber on Mon Aug 25 18:01:15 PDT 1997

cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

Two new routines: w_findRectFromPolygon() which locates the best
orthogonal polygon segments and returns the angle; and w_rotatePolygon()
which rotate a polygon an arbitrary angle.
I originally tried to use the fixed point routines in the polygon rotation,
but didn't get it to work, yet to be fixed.

Revision 1.25  1997/06/26 20:00:36  ddavies
cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

CommonCheck: Add inclusion of new file rast.pub.

Revision 1.24  1997/04/15 21:43:14  ddavies
cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

Substitute NoInit (uninitialized) version of pixr creation routines where
possible.

Revision 1.23  1996/09/26 21:17:01  dferrell
cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

Made w_thinPolygon more accurate by using squared pythagorean distance
rather than manhattan distance when checking for near-colinearity.

Revision 1.22  1996/01/27 22:37:30  ejaquith
cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v

updated call to w_xformRect()

 * Revision 1.21  1996/01/04  20:21:30  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.20  1995/12/13  04:05:31  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.19  1995/11/13  20:09:57  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.18  1995/10/13  21:59:00  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/10/03  21:12:41  macleod
 * macleod on Tue Oct 3 14:12:40 PDT 1995
 *
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.16  1995/09/21  22:51:52  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/11  21:54:56  macleod
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.14  1995/09/07  01:29:56  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:17:38  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  20:43:09  macleod
 * macleod on Fri Sep 1 13:43:00 PDT 1995
 *
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * Added back #include files
 *
 * Revision 1.11  1995/08/31  21:11:25  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/25  01:35:39  ejaquith
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * Fixed warnings
 *
 * Revision 1.9  1995/08/17  21:54:13  macleod
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.8  1995/07/26  23:56:56  dferrell
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * Fixed memory leak of one (1, uno, I, sizeof (char)) RECTANGL struct.
 *
 * Revision 1.7  1995/06/23  23:16:08  dferrell
 * dferrell on Fri Jun 23 16:14:07 PDT 1995
 *
 * cntr.c -> /product/ipcore/ipwrap/src/RCS/cntr.c,v
 *
 * Added two new functions:
 * thinPolygon () which uses manhattan distance and some trig to remove
 * nearly colinear points from a given polygon, and drawPolygon () which
 * blits a polygon into a pixr.
 *

*/
