/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * color.c
 *
 *	Color space transformations, format conversions, color maps, etc.
 *
 * Exported:
 *
 *	w_RGBToY ()	      - Create a luminance (Y) channel from R,G,B
 *	w_RGBToY2 ()	      - Create a DITC luminance (Y) channel from R,G,B
 *	w_RGBToYIP ()         - Create a luminance (Y) channel from R,G,B
 *			        but don't allocate any new space
 *	w_RGBToYFast ()	      - Create a luminance (Y) channel from R,G,B
 *	w_colorMaskPr ()      - Create a color mask from an RGB image
 *	w_colorMaskInit ()    - Create a LUT for colorMaskPr
 *	w_colorMaskFinish ()  - Clean up from colorMaskInit
 *  	w_indexedToRGBPr ()   - Convert a pseudo-color image to separate planes
 *	w_rgbToLuvInit ()     - Make some tables for RGB to L* u* v* conversion
 *	w_rgbToLuv ()	      - Convert RGB to L* u* v*. SLOW!
 *	w_rgbToLuvFinish ()   - IPW_FREE tables for L* u* v* conversion
 *	w_createColorMap ()   - Create a color map
 *	w_destroyColorMap ()  - Destroy a color map
 *	w_fillLinearColorMap () - Fill a color map with a linear palette
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "imageref.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "rectops.pub"
#include "color.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: color.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute the luminance channel of an RGB color image.
 *
 * LongDescription:::
 *  The actual color in a color image is often not needed for analysis
 *  or printing.  Simply choosing one of the RGB channels sort of works
 *  sometimes, but not always.  This routine computes the luminance
 *  at every pixel of a color image.  It's output can often be used
 *  as though the image were scanned in gray.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to RGB image to be converted.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to gray dest image.  If none
 *		     is provided, one will be created.  If one is provided,
 *		     its dimensions, depth, etc will be checked.
 *
 * Notes:::
 *  This is a fairly expensive (slow) operation.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  See Sandpiper's "cspace" utility.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGBToY (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subR, *subG, *subB;
    UInt8		*pR, *pG, *pB, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels3, cFrame32);

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, dS,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* Now that we've got legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));

    ERR_PASS_CLEANUP( i_RGBToY (pIPWState, pR, pG, pB, bplS, pD, bplD, wS, hS));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute the luminance channel of an RGB color image, using DITC's formula.
 *
 * LongDescription:::
 *  This does the same thing as w_RGBToY(), but uses the formula for computing
 *  the transform that DITC uses for doing AIE.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to RGB image to be converted.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to gray dest image.  If none
 *		     is provided, one will be created.  If one is provided,
 *		     its dimensions, depth, etc will be checked.
 *
 * Notes:::
 *  This is a fairly expensive (slow) operation.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *  See Sandpiper's "cspace" utility.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGBToY2 (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subR, *subG, *subB;
    UInt8		*pR, *pG, *pB, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels3, cFrame32);

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, dS,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* Now that we've got legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));

    ERR_PASS_CLEANUP( i_RGBToY2 (pIPWState, pR, pG, pB,
				 bplS, pD, bplD, wS, hS));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert RGB image into gray image using computed luminance channel.
 *
 * LongDescription:::
 *  w_RGBToY converts an RGB color image to a gray image by computing the
 *  the luminance channel.  This routine does the same thing, but it
 *  replaces one of the source channels with this new channel and discards
 *  the other two channels.  This allows the client to avoid the memory
 *  required to hold both an RGB image and a gray image concurrently.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  prS::            Pointer to source and dest image.  On entry, this
 *		     must point to an RGB (3 channel) image. On exit,
 *		     the image will be converted into a gray (1 channel)
 *		     image.  The other two channels will be freed.
 *
 * Notes:::
 *  This is still a relatively slow operation.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  See Sandpiper's "cspace" utility.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGBToYIP (
    IPWState           *pIPWState,
    PIXR               *prS)
{
    PROLOGUE;
    Int32	 w, h, depth, bpl, channels;
    PIXR	*subR, *subG, *subB;
    UInt8	*pR, *pG, *pB;

    ERR_GEN(prS == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels3, cFrame32);

    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &depth));
    ERR_PASS( w_pixrGetChannelCount (pIPWState, prS, &channels));

    ERR_GEN(depth != 8, cErrDepthNotSupported);
    ERR_GEN(channels != 3, cErrInvalidParm);

    ERR_PASS( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS( w_pixrGetImageAll (pIPWState, subR, &pR));
    ERR_PASS( w_pixrGetImageAll (pIPWState, subG, &pG));
    ERR_PASS( w_pixrGetImageAll (pIPWState, subB, &pB));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrFramedWidth (pIPWState, prS, &w));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &h));

    ERR_PASS( i_RGBToY (pIPWState, pR, pG, pB, bpl, pG, bpl, w, h));

	/* Convert the original PIXR structure into a clone
	 * of the former green channel.  That's where the
	 * luminence data is now.
	 */
    ERR_PASS (w_clonePixrStructOnly(pIPWState, subG, &prS));

    /* Now prS is gray image not RGB color image. We need to set color type
       as cColorGray. */
    ERR_PASS (w_pixrSetType (pIPWState, prS, cColorGray));

    /* Get rid of the original sub-channels */
    ERR_PASS( w_destroyPixr(pIPWState, &subR));
    ERR_PASS( w_destroyPixr(pIPWState, &subG));
    ERR_PASS( w_destroyPixr(pIPWState, &subB));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute the approximate luminance channel of an RGB color image.
 *
 * LongDescription:::
 *  The actual color in a color image is often not needed for analysis
 *  or printing.  Simply choosing one of the RGB channels sort of works
 *  sometimes, but not always.  This routine computes the luminance
 *  at every pixel of a color image.  It's output can often be used
 *  as though the image were scanned in gray.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to RGB image to be converted.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to gray dest image.  If none
 *		     is provided, one will be created.  If one is provided,
 *		     its dimensions, depth, etc will be checked.
 *
 * Notes:::
 *	The correct formula is:  Y = 0.299R + 0.587G + 0.114B and the
 *	approximation is: Y = 0.247R + 0.620G + 0.122B which saturates
 *	at 252 rather than 255.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGBToYFast (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subR, *subG, *subB;
    UInt8		*pR, *pG, *pB, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels3, cFrame32);

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

    /* if we already have a dest pixr, make sure it's the right size.
     * otherwise, create one.
     */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, dS,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

    /* Now that we've got legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));

    ERR_PASS_CLEANUP( i_RGBToYFast (pIPWState, pR, pG, pB,
				    bplS, bplD, wS, hS, pD));

    /* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

    /* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;

}  /* w_RGBToYFast () */


#define cColorMaskChannelDelta	16
#define cColorMaskEntries	4096
#define cLongShift		5
#define cLongMask		0x1f

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Prepare to compute a mask that identifies all color pixels in an image.
 *
 * LongDescription:::
 *  It is often useful to be able to identify all color pixels in an image.
 *  This is a cheap way of determining which compression method might be
 *  used, for example.  This routine allocates a private state block and
 *  fills in tables attached to the state block.
 *
 * PrivateDescription:::
 *  In this version, we're going to 
 *  use a slightly different method. In the first version, we just
 *  summed RGB deltas, and used a cutoff. It wasn't very good. In
 *  this version, we're going to convert to L*u*v*, and use the u*v*
 *  axis to compute a color distance from (0.0, 0.0), and use the
 *  L* axis to weight the color distance so that we tend to suppress
 *  color in the dark regions.
 *  
 *  RGB->L*u*v* conversion is slow in the best of algorithms,
 *  unless one is willing to make a massive table. Since we're
 *  not that concerned with absolute accuracy, we'll just
 *  make a bit vector indexed by the high-order 4 bits of each
 *  RGB value. This function creates that table.
 *
 * Input:::
 *  pIPWState::        Pointer to standard IPWrap state block.
 *  minChromaSquared:: Square of minimum chroma distance that we use
 *		       to call a pixel "colored" as opposed to "gray".
 *  lumaWeightFactor:: Inverse-luminance weight on chrominance.
 *		       If this is 0.0, no weighting happens.
 *		       If this is 1.0, chrominance distance is
 *		       multiplied by (100-luminance)/100, so that
 *		       darks get their chrominance suppressed.
 *		       Between 0 and 1, the weight is blended with
 *		       a non-weighted version.
 *
 * Output:::
 *  ppCMaskHandle::   Pointer to pointer to private state block that will
 *		      be used during the calculation.  *ppCMaskHandle must
 *		      be NULL on entry.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrPtrNotNull
 *  if *ppCMaskHandle is not NULL on entry.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison, color, colored text, segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_colorMaskInit (
    IPWState           *pIPWState,
    Float64             minChromaSquared,
    Float64             lumaWeightFactor,
    void              **ppCMaskHandle)
{
    PROLOGUE;
    UInt32  *pCMaskTable = NULL, word, index;
    UInt32   r, g, b;
    Float64  L, u, v, chromaDSquared, lumaWeight;
    void    *pState = NULL;

    ERR_GEN(ppCMaskHandle == NULL, cErrNullPtr);
    ERR_GEN(*ppCMaskHandle != NULL, cErrPtrNotNull);

    /* Allocate the cmask table */
    *ppCMaskHandle = NULL;
    pCMaskTable = (UInt32 *) IPW_MALLOC (cColorMaskEntries / 8, cAllocatorData);
    ERR_GEN(pCMaskTable == NULL, cErrNomem);

    /* Fill the table  */
    ERR_PASS_CLEANUP( w_RGBToLuvInit (pIPWState, 0.0, 0.0, 0.0, &pState));
    word = 0;

    for (r = 0; r < 256; r += cColorMaskChannelDelta)
    {
	for (g = 0; g < 256; g += cColorMaskChannelDelta)
	{
	    for (b = 0; b < 256; b += cColorMaskChannelDelta)
	    {
		/* Make the index bits. */
		index = (r << 4) | g | (b >> 4);

		/* Convert this one */
		ERR_PASS_CLEANUP( w_RGBToLuvProcess (pIPWState, r, g, b, pState,
					   &L, &u, &v));

		/* Compute square of chroma distance */
		chromaDSquared = u * u + v * v;

		/* Weight it by luminance */
		lumaWeight = (L / 100.0);
		lumaWeight *= lumaWeight;		    
		chromaDSquared = chromaDSquared * lumaWeight * lumaWeightFactor +
				 chromaDSquared * (1.0 - lumaWeightFactor);

		/* Set the bit if needed */
		if (chromaDSquared > minChromaSquared)
		    word |= (1 << (index & cLongMask));

		/* Store the word if we need to */
		if ((index & cLongMask) == cLongMask)
		{
		    pCMaskTable[index >> cLongShift] = word;
		    word = 0;
		}
	    }
	}
    }

    ERR_PASS_CLEANUP( w_RGBToLuvFinish(pIPWState, &pState));
    
    *ppCMaskHandle = pCMaskTable;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    IPW_FREE(pCMaskTable, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a binary mask that identifies all pixels in an image that really
 *  are colored, as opposed to gray.
 *
 * LongDescription:::
 *  It is often useful to be able to identify all color pixels in an image.
 *  This is a cheap way of determining which compression method might be
 *  used, for example.  This routine uses a private state block allocated
 *  by w_colorMaskInit to generate a mask identifying all such pixels.
 *  The state block is allocated separately so that clients may call this
 *  routine many times with minimum overhead.
 *
 * PrivateDescription:::
 *  In former versions, it tried some stupid heuristic to suppress
 *  color from alignment errors.
 *  It was lame, however, and doing it using morphology on the
 *  resulting bitmap is probably more useful, so I'm taking it out.
 *<BR><BR>
 *  The other difference between this and previous versions is that
 *  one must call w_colorMaskInit() to set the parameters, and then
 *  pass in a handle from that.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pCMaskHandle::   Pointer to private color-finding state block.  This
 *		     block was created by w_colorMaskInit.
 *  prS::            Pointer to RGB image to be analyzed.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to binary mask image produced. If none
 *		     is provided, one will be created.  If one is provided,
 *		     its dimensions, depth, etc will be checked.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison, color, colored text, segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_colorMask (
    IPWState           *pIPWState,
    void               *pCMaskHandle,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    UInt8		*pMask;
    Int32		 wS, hS, sBpl, dBpl, i, j;
    UInt32		 bits, *outP, *cMaskTable, index;
    UInt8		*saveMask;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subR, *subG, *subB;
    UInt8		*pR, *pG, *pB;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL || pCMaskHandle == NULL,
	    cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels3,
		 cFrame32);

    cMaskTable = (UInt32 *) pCMaskHandle;
    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, 1,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* Now that we have legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, prD, &pMask));
    ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prD, &dBpl));

    saveMask = pMask;

    for (i = 0; i < hS; i++)
    {
	/* Do a line */
	outP = (UInt32 *) pMask;
	bits = 0;

	for (j = 0; j < wS; j++)
	{
	    /* Make the index bits */
	    index = ((UCHAR_ACCESS (pR + j) & 0xf0) << 4) | 
		     (UCHAR_ACCESS (pG + j) & 0xf0) | 
		     (UCHAR_ACCESS (pB + j) >> 4);

	    /* Set the bit if it's set in the table */
	    if (cMaskTable[index >> cLongShift] & (1 << (index  & cLongMask)))
		bits = (bits << 1) | 1;
	    else
		bits <<= 1;

	    /* If we're on the last pixel of a 32-pixel boundary,
	     * write the bits we've accumulated
	     */
	    if ((j & cLongMask) == cLongMask)
	    {
		/* We have some bits to write */
		*outP++ = bits;
		bits = 0;
	    }
	}

	/* Output the last one, if needed, but not if we already did */
	if (((j - 1) & cLongMask) && (((j - 1) & cLongMask) != cLongMask))
	    *outP++ = bits;

	pR += sBpl;
	pG += sBpl;
	pB += sBpl;
	pMask += dBpl;
    }

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
/*    i_swapImageBytes (pIPWState, saveMask + (dBpl << cLogFrameBits),
			dBpl, h); */
#endif

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Deallocates private color-finding state block.
 *
 * LongDescription:::
 *  w_colorMask uses a private state block allocated by w_colorMaskInit.
 *  This routine frees that state block and all associated memory blocks.
 *
 * Input:::
 *  pIPWState::      Pointer to private IPWrap state block.
 *
 * Output:::
 *  ppCMaskHandle::  Pointer to pointer to private state block that will
 *		     be used during the calculation.  **ppCMaskHandle
 *		     (and associated memory blocks) will be freed and
 *		     *ppCMaskHandle will be set to NULL on return.
 *
 * Notes:::
 *  Clients should not simply free **ppCMaskHandle since that may lose any
 *  associated memory blocks.
 *<BR><BR>
 *  Like all freeing routines, this routine does nothing if
 *  *ppCMaskHandle is already NULL on entry.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison, color, colored text, segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_colorMaskFinish (
    IPWState           *pIPWState,
    void              **ppCMaskHandle)
{
    PROLOGUE;

    ERR_GEN(ppCMaskHandle == NULL, cErrNullPtr);

    /* Since on some systems (i.e. system V) IPW_FREE doesn't return anything,
     * we'll just trust it.
     */
    if (*ppCMaskHandle != NULL)
    {
	IPW_FREE (*ppCMaskHandle, cAllocatorData);
	*ppCMaskHandle = NULL;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines whether a single [R,G,B] triplet is "colored" or "gray."
 *
 * LongDescription:::
 *  Who knows why this exists?  It decides whether a single pixel is
 *  gray or colored.  It is not called by w_colorMask, though it uses
 *  the same method.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pCMaskHandle::   Pointer to private color-finding state block.  This
 *		     block was created by w_colorMaskInit.
 *  r::              Red intensity value, between 0 and 255.
 *  g::              Green intensity value, between 0 and 255.
 *  b::              Blue intensity value, between 0 and 255.
 *
 * Output:::
 *  pIsColor::       Pointer to output boolean.  If this routine returns
 *		     the Bool equal to cTrue, the pixel is colorful,
 *		     otherwise it's gray..
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison, color, colored text, segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_isColor (
    IPWState           *pIPWState,
    void               *pCMaskHandle,
    Int32               r,
    Int32               g,
    Int32               b,
    Bool               *pIsColor)
{
    PROLOGUE;
    Int32    index;
    UInt32  *pCMaskTable;

    ERR_GEN(pIsColor == NULL || pCMaskHandle == NULL, cErrNullPtr);

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    pCMaskTable = (UInt32 *) pCMaskHandle;

    index = ((r & 0xf0) << 4) |
	    (g & 0xf0) |
	    (b >> 4);

    if (pCMaskTable[index >> cLongShift] & (1 << (index & cLongMask)))
	*pIsColor = cTrue;
    else
	*pIsColor = cFalse;

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert an indexed color image with color map to full RGB image.
 *
 * LongDescription:::
 *  Indexed color images are much smaller than full RGB images, but much
 *  harder to analyze and compress.  Some I/O devices also require
 *  full RGB images.  This routine converts an indexed color image to
 *  a full RGB image by simply replacing each pixel with the RGB triplet
 *  indexed by the pixel in the color map.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to indexed color image to be converted.
 *		     This may have 4 bit/pixel or 8 bits/pixel.
 *  pRect::          Pointer to RECTANGL that determines the portion of
 *		     the indexed color image to be converted.  If
 *		     pRect is NULL on entry, the entire image is converted.
 *  cMap::           Pointer to the ColorMap for the image.  As we really
 *		     begin using the color maps in the Pixr structures,
 *		     this should go away.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to RGB mask image produced. If none
 *		     is provided, one will be created.  If one is provided,
 *		     its dimensions, depth, etc will be checked.  Note that
 *		     this image always has 8 bits/channel irrespective of
 *		     the depth of the indexed color source image.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, pixel mapping, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_indexedToRGB (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    ColorMap           *cMap,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS;
    Int32		 x, y, w, h, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subR, *subG, *subB;
    UInt8		*pR, *pG, *pB, *pS;
    Int8		 valid;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL || cMap == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* On entry, the rect (x, y, w, h) specifies a rectangle with respect
	 * to the upper left corner of the "proper" image (inside the frame).
	 * The output binary image is always justified to
	 * the upper left corner of its framed pixrect (i.e. inside the
	 * frame).   If the rect is NULL, the entire image is used.  */
    if (pRect == NULL)
    {
	x = 0;
	y = 0;
	w = wS;
	h = hS;
    }
    else
    {
	ERR_PASS( w_isRectValid(pIPWState, prS, pRect, &valid));
	ERR_GEN( !valid, cErrInvalidParm);
	x = pRect->x;
	y = pRect->y;
	w = pRect->w;
	h = pRect->h;
    }

	/* check the width & height of the frameless image.  The
	 * frame is in units of PIXELS, not bits.
	 */
    ERR_GEN((x < 0) || (y < 0), cErrOutOfBounds);
    ERR_GEN((w <= 0) || (h <= 0), cErrOutOfBounds);
    ERR_GEN((x+w) > wS || (y+h) > hS, cErrOutOfBounds);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrExtendedConditionalNoInit (pIPWState, w, h, 8,
						      xSdpi, ySdpi, cFrameBits,
						      3, cColorRGB, NULL, &prD,
						      &rightFrameColsToClear,
						      &bottomFrameRowsToClear,
						      &allocatedHere));

	/* Now that we have legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));

    if (dS == 4)
	ERR_PASS_CLEANUP( i_indexed4ToRGB (pIPWState, pS,
					   x + cFrameBits, y + cFrameBits,
					   w, h,
					   bplS, pR, pG, pB,
					   bplD, (UInt8 *) &cMap->map[0]));
    else
	ERR_PASS_CLEANUP( i_indexed8ToRGB (pIPWState, pS,
					   x + cFrameBits, y + cFrameBits,
					   w, h,
					   bplS, pR, pG, pB,
					   bplD, (UInt8 *) &cMap->map[0]));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Prepare to convert an image from RGB to L*u*v*.
 *
 * LongDescription:::
 *  There are lots of color spaces.  The RGB space is popular and the
 *  one most heavily supported by this library.  However, the L*u*v*
 *  space has advantages.  This routine allocates a private state block
 *  and tables used to convert from RGB to L*u*v*.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  x::              X coord of white point.  If 0, uses D50.
 *  y::              Y coord of white point,  If 0, uses D50
 *  z::              Z coord of white point,  If 0, uses D50
 *
 * Output:::
 *  ppState::        Pointer to output state block.  *ppState must be NULL
 *		     on entry.  The state block is always allocated here.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrPtrNotNull
 *  if *ppState is not NULL on entry.
 *
 * Notes:::
 *  The conversion process if Very Slow.  Don't do this for fun.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGBToLuvInit (
    IPWState           *pIPWState,
    Float64             x,
    Float64             y,
    Float64             z,
    void              **ppState)
{
    PROLOGUE;

    ERR_GEN(ppState == NULL, cErrNullPtr);
    ERR_GEN(*ppState != NULL, cErrPtrNotNull);

    /* moved to ipshared */
    ERR_PASS( i_RGBToLuvInit (pIPWState, x, y, z, ppState));
    
    NORMAL_RETURN;
}
    

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a single [R,G,B] triplet to a single [L*u*v*] triplet.
 *
 * LongDescription:::
 *  This routine converts a single [R,G,B] triplet to a single
 *  [L*u*v*] triplet.  This is Very Slow. It's only
 *  intended use is to initialize a table. If you pass it 
 *  image data, you get what you deserve.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  r::              Red intensity, between 0 and 255.
 *  g::              Green intensity, between 0 and 255.
 *  b::              Blue intensity, between 0 and 255.
 *  pState::         Pointer to private state block allocated by
 *		     w_RGBToLuvInit.
 *
 * Output:::
 *  L::              L* portion of output pixel.
 *  u::              u* portion of output pixel.
 *  v::              v* portion of output pixel.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGBToLuvProcess (
    IPWState           *pIPWState,
    Int32               r,
    Int32               g,
    Int32               b,
    void               *pState,
    Float64            *L,
    Float64            *u,
    Float64            *v)
{
    PROLOGUE;

    ERR_GEN(pState == NULL || L == NULL || u == NULL || v == NULL,
	    cErrNullPtr);

    /* moved to ipshared */
    ERR_PASS( i_RGBToLuv (pIPWState, r, g, b, pState, L, u, v));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Deallocated the internal structures used to convert RGB to L*u*v*
 *
 * LongDescription:::
 *  w_RGBToLuv uses a private state block to convert one RGB pixel to
 *  one L*u*v* pixel.  That routine is likely to be called a lot.
 *  This routine is called by clients to free the private state block
 *  when finished.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  ppState::        Pointer to pointer to private state block that will
 *		     be used during the calculation.  **ppState
 *		     (and associated memory blocks) will be freed and
 *		     *ppState will be set to NULL on return.
 *
 * Notes:::
 *  Clients should not simply free **ppState since that may lose any
 *  associated memory block.
 *<BR><BR>
 *  Like all freeing routines, this routine does nothing if
 *  *ppCMaskHandle is already NULL on entry.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGBToLuvFinish (
    IPWState           *pIPWState,
    void              **ppState)
{
    PROLOGUE;

    ERR_GEN(ppState == NULL, cErrNullPtr);

    /* move to ipshared */
    ERR_PASS( i_RGBToLuvFinish (pIPWState, *ppState));
    *ppState = NULL;

    NORMAL_RETURN;
}
    

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Prepare to convert an image from RGB to L*a*b*.
 *
 * LongDescription:::
 *  There are lots of color spaces.  The RGB space is popular and the
 *  one most heavily supported by this library.  However, the L*a*b*
 *  space has advantages.  This routine allocates a private state block
 *  and tables used to convert from RGB to L*a*b*.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  x::              X coord of white point.  If 0, uses D50.
 *  y::              Y coord of white point,  If 0, uses D50
 *  z::              Z coord of white point,  If 0, uses D50
 *
 * Output:::
 *  ppState::        Pointer to output state block.  *ppState must be NULL
 *		     on entry.  The state block is always allocated here.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrPtrNotNull
 *  if *ppState is not NULL on entry.
 *
 * Notes:::
 *  The conversion process if Very Slow.  Don't do this for fun.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, depth conversion, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  JCraig
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_RGBToLabInit (
    IPWState           *pIPWState,
    Float64             x,
    Float64             y,
    Float64             z,
    void              **ppState)
{
    PROLOGUE;

    ERR_GEN(ppState == NULL, cErrNullPtr);
    ERR_GEN(*ppState != NULL, cErrPtrNotNull);

    /* moved to ipshared */
    ERR_PASS(i_RGBToLabInit (pIPWState, x, y, z, ppState));
    
    NORMAL_RETURN;
}
    

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a single [R,G,B] triplet to a single [L*a*b*] triplet.
 *
 * LongDescription:::
 *  This routine converts a single [R,G,B] triplet to a single
 *  [L*a*b*] triplet.  This is Very Slow. It's only
 *  intended use is to initialize a table. If you pass it 
 *  image data, you get what you deserve.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  r::              Red intensity, between 0 and 255.
 *  g::              Green intensity, between 0 and 255.
 *  b::              Blue intensity, between 0 and 255.
 *  pState::         Pointer to private state block allocated by
 *		     w_RGBToLabInit.
 *
 * Output:::
 *  L::              L* portion of output pixel.
 *  aStar::          a* portion of output pixel.
 *  bStar::          b* portion of output pixel.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  JCraig
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_RGBToLabProcess (
    IPWState           *pIPWState,
    Int32               r,
    Int32               g,
    Int32               b,
    void               *pState,
    Float64            *L,
    Float64            *aStar,
    Float64            *bStar)
{
    PROLOGUE;

    ERR_GEN(pState == NULL || L == NULL || aStar == NULL || bStar == NULL,
	    cErrNullPtr);

    /* moved to ipshared */
    ERR_PASS(i_RGBToLab (pIPWState, r, g, b, pState, L, aStar, bStar));

    NORMAL_RETURN;
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Deallocated the internal structures used to convert RGB to L*a*b*
 *
 * LongDescription:::
 *  w_RGBToLab uses a private state block to convert one RGB pixel to
 *  one L*a*b* pixel.  That routine is likely to be called a lot.
 *  This routine is called by clients to free the private state block
 *  when finished.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  ppState::        Pointer to pointer to private state block that will
 *		     be used during the calculation.  **ppState
 *		     (and associated memory blocks) will be freed and
 *		     *ppState will be set to NULL on return.
 *
 * Notes:::
 *  Clients should not simply free **ppState since that may lose any
 *  associated memory block.
 *<BR><BR>
 *  Like all freeing routines, this routine does nothing if
 *  *ppCMaskHandle is already NULL on entry.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  JCraig
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_RGBToLabFinish (
    IPWState           *pIPWState,
    void              **ppState)
{
    PROLOGUE;

    ERR_GEN(ppState == NULL, cErrNullPtr);

    /* move to ipshared */
    ERR_PASS(i_RGBToLabFinish (pIPWState, *ppState));
    *ppState = NULL;

    NORMAL_RETURN;
}
    
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Allocate a ColorMap structure.
 *
 * LongDescription:::
 *  Indexed color images need an associated color map.  This routine
 *  allocates an official ColorMap structure and fills in the overhead
 *  sections.  The map itself is left filled with 0's.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  colors::         Number of colors in the color map.  This should
 *		     probably be 16 or 256.
 *
 * Output:::
 *  ppMap::          Pointer to pointer to returned ColorMap structure.
 *		     On entry *ppMap must always be NULL.  The color map
 *		     is always allocated here.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrPtrNotNull
 *  if *ppMap is not NULL on entry.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createColorMap (
    IPWState           *pIPWState,
    Int32               colors,
    ColorMap          **ppMap)
{
    PROLOGUE;
    ColorMap    *pMap = NULL;

    ERR_GEN(ppMap == NULL, cErrNullPtr);
    ERR_GEN(*ppMap != NULL, cErrPtrNotNull);

    pMap = (ColorMap *) IPW_MALLOC (sizeof(ColorMap), cAllocatorData);
    ERR_GEN(pMap == NULL, cErrNomem);

    pMap->count = (Int16) colors;
    pMap->type = cColorRGB;	/* Default, since we don't use it yet */
    pMap->map = (ColorRecord *) IPW_CALLOC(colors, sizeof(ColorRecord),
					   cAllocatorData);
    ERR_GEN_CLEANUP(pMap->map == NULL, cErrNomem);

    *ppMap = pMap;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pMap->map, cAllocatorData);
    IPW_FREE(pMap, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Deallocate a ColorMap allocated by w_createColorMap.
 *
 * LongDescription:::
 *  w_createColorMap allocates and partially initializes a ColorMap
 *  structure and associated memory.  This routine destroys this stuff.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  ppMap::          Pointer to pointer to ColorMap.  **ppMap
 *		     (and associated memory blocks) will be freed and
 *		     *ppMap will be set to NULL on return.
 *
 * Notes:::
 *  Clients should not simply free **ppMap since that may lose any
 *  associated memory blocks.
 *<BR><BR>
 *  Like all freeing routines, this routine does nothing if
 *  *ppMap is already NULL on entry.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyColorMap (
    IPWState           *pIPWState,
    ColorMap          **ppMap)
{
    PROLOGUE;

    ERR_GEN(ppMap == NULL, cErrNullPtr);

    if (*ppMap != NULL)
    {
	IPW_FREE((*ppMap)->map, cAllocatorData);
	IPW_FREE (*ppMap, cAllocatorData);
	*ppMap = NULL;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Fill in an existing color map with linear values.
 *
 * LongDescription:::
 *  Having created a ColorMap, clients might like to put something interesting
 *  into it.  This routine fills a color map wiht linear values. For example,
 *  if one requested 5 red levels, 4 green levels, and 4 blue levels,
 *  one would get 5*5*4=100 colors.
 *<BR><BR>
 *  If a trc (Tone Reproduction Curve) is supplied, the levels on ALL THREE
 *  channels of the color map are mapped using the single TRC before
 *  being stored in the ColorMap.
 *
 * Input:::
 *  pIPWState::      Pointer to private IPWrap state block.
 *  redColors::      Number of levels of Red to appear in color map.
 *  greenColors::    Number of levels of Green to appear in color map.
 *  blueColors::     Number of levels of Blue to appear in color map.
 *  colorOffset::    New entries start at this offset inside the color map.
 *  pColorMap::      Pointer to an existing ColorMap.  redColors*
 *		     greenColors*blueColors entries starting at colorOffset
 *		     will be filled in by this routine.
 *  trc::            Pointer to optional Tone Reproduction Curve.  If not NULL,
 *		     this table will be used to modify the linear values
 *		     generated before being stored in the ColorMap.
 *
 * Notes:::
 *  "Linear" colors are ones in which we simply count through all possible
 *  combinattions of red, blue and green (given the number of levels of
 *  each specified).
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, color, pixel mapping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_linearColorMap (
    IPWState           *pIPWState,
    Int32               redColors,
    Int32               greenColors,
    Int32               blueColors,
    Int32               colorOffset,
    ColorMap           *pColorMap,
    UInt8              *trc)
{
    PROLOGUE;
    Int32     i, use_trc;
    Int32     redIndex, greenIndex, blueIndex;
    Float32   redFrac, greenFrac, blueFrac;
    Float32   redLevel, greenLevel, blueLevel;

    ERR_GEN(pColorMap == NULL, cErrNullPtr);
    ERR_GEN((redColors * greenColors * blueColors + colorOffset >
	     pColorMap->count), cErrInvalidParm);

    /* get trc correction */
    use_trc = (trc != NULL);

    redFrac = 255.0F / (redColors - 1);
    greenFrac = 255.0F / (greenColors - 1);
    blueFrac = 255.0F / (blueColors - 1);

    redLevel = 0.5F;
    i = colorOffset;
    for (redIndex = 0; redIndex < redColors; ++redIndex)
    {
	greenLevel = 0.5F;
	for (greenIndex = 0; greenIndex < greenColors; ++greenIndex)
	{
	    blueLevel = 0.5F;
	    for (blueIndex = 0; blueIndex < blueColors; ++blueIndex)
	    {
		if (use_trc)
		{
		    pColorMap->map[i][cRedPlane] = trc[(UInt8) redLevel];
		    pColorMap->map[i][cGreenPlane] = trc[(UInt8) greenLevel];
		    pColorMap->map[i][cBluePlane] = trc[(UInt8) blueLevel];
		}
		else
		{
		    pColorMap->map[i][cRedPlane] = (UInt8) redLevel;
		    pColorMap->map[i][cGreenPlane] = (UInt8) greenLevel;
		    pColorMap->map[i][cBluePlane] = (UInt8) blueLevel;
		}
		++i;
		blueLevel += blueFrac;
	    }
	    greenLevel += greenFrac;
	}
	redLevel += redFrac;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert an RGB color image to Lab color image.
 *
 * LongDescription:::
 *  Lots color image analysis need working in Lab color space. This
 *  function convert RGB color image to Lab color image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to RGB image to be converted.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to Lab dest image.  If none
 *		     is provided, one will be created.  If one is provided,
 *		     its dimensions, depth, etc will be checked.
 *
 * Notes:::
 *  This function will spend time to setup tables. If image is not
 *  large enough ( > 2048 ), use this function will be very slow.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * Citations:::
 *  See Sandpiper's "cspace" utility.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGB2Lab (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subR, *subG, *subB;
    UInt8		*pRed, *pGreen, *pBlue, *pL, *pA, *pB;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;


    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels3, cFrame32);

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrExtendedConditional (pIPWState, wS, hS, dS,
						xSdpi, ySdpi, cFrameBits,
						3, cColorRGB, NULL, &prD,
						&rightFrameColsToClear,
						&bottomFrameRowsToClear,
						&allocatedHere));
    ERR_PASS( w_pixrSetType( pIPWState, prD, cColorCIELAB) );

	/* Now that we've got legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subR, &pRed));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subG, &pGreen));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pBlue));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subR, &pL));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subG, &pA));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pB));

    ERR_PASS_CLEANUP( i_RGB2Lab(pIPWState, pRed, pGreen, pBlue,
				bplS, bplD, wS, hS,
				pL, pA, pB));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert an strip of  RGB color image to strip of  Lab color image.
 *
 * LongDescription:::
 *  This function convert RGB color image to Lab color image. It works
 *  strip by strip after creating the table. It also can work under mask.
 *  If there is mask image, the convertion will be happen only for masked
 *  pixels.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pTable::         Pointer to RGB2LAB table.
 *  prS::            Pointer to RGB image to be converted.
 *  prMask::         Pointer to mask image.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to Lab dest image.  If none
 *		     is provided, one will be created.  If one is provided,
 *		     its dimensions, depth, etc will be checked.
 *
 * Notes:::
 *  This function will spend time to setup tables. If image is not
 *  large enough ( > 2048 ), use this function will be very slow.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * Citations:::
 *  See Sandpiper's "cspace" utility.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGB2LabUseTable (
    IPWState           *pIPWState,
    void               *pTable,
    PIXR               *prS,
    PIXR               *prMask,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, bplS, bplD, bplM;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subR, *subG, *subB;
    UInt8		*pRed, *pGreen, *pBlue, *pL, *pA, *pB, *pMask;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;


    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels3, cFrame32);

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrExtendedConditional (pIPWState, wS, hS, dS,
						xSdpi, ySdpi, cFrameBits,
						3, cColorRGB, NULL, &prD,
						&rightFrameColsToClear,
						&bottomFrameRowsToClear,
						&allocatedHere));
    ERR_PASS( w_pixrSetType( pIPWState, prD, cColorCIELAB) );

	/* Now that we've got legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subR, &pRed));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subG, &pGreen));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pBlue));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, prMask, &pMask));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prMask, &bplM));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subR, &pL));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subG, &pA));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pB));

    ERR_PASS_CLEANUP( i_RGB2LabUseTable(pIPWState, pTable,
				     pRed, pGreen, pBlue, pMask,
				     bplS, bplD, bplM, wS, hS,
				     pL, pA, pB));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create RGB2Lab table for converting  an strip of  RGB color image 
 *  to strip of  Lab color image.
 *
 * LongDescription:::
 *  This function create the table for converting strip of  RGB color 
 *  image to the strip of  Lab color image. 
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  ppTable::        Point to pointer of RGB2LAB table.
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * Citations:::
 *  See Sandpiper's "cspace" utility.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createRGB2LabTable (
    IPWState           *pIPWState,
    void              **ppTable)
{
    PROLOGUE;

    ERR_GEN(ppTable == NULL || *ppTable != NULL, cErrNullPtr);
    
    ERR_PASS( i_createRGB2LabTable(pIPWState, ppTable));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free RGB2Lab table.
 *
 * LongDescription:::
 *  This function free the table for converting strip of  RGB color 
 *  image to the strip of  Lab color image. 
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  ppTable::        Point to pointer of RGB2LAB table.
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * Citations:::
 *  See Sandpiper's "cspace" utility.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_freeRGB2LabTable (
    IPWState           *pIPWState,
    void              **ppTable)
{
    PROLOGUE;

    ERR_GEN(*ppTable == NULL, cErrNullPtr);
    
    ERR_PASS( i_freeRGB2LabTable(pIPWState, ppTable));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert an Lab color image to RGB color image.
 *
 * LongDescription:::
 *  The function convert Lab color image to RGB color image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to Lab image to be converted.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to RGB dest image.  If none
 *		     is provided, one will be created.  If one is provided,
 *		     its dimensions, depth, etc will be checked.
 *
 * Notes:::
 *  This function will spend time to setup tables. If image is not
 *  large enough ( > 2048 ), use this function will be very slow.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * Citations:::
 *  See Sandpiper's "cspace" utility.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_Lab2RGB (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subR, *subG, *subB;
    UInt8		*pRed, *pGreen, *pBlue, *pL, *pA, *pB;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels3, cFrame32);

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrExtendedConditional (pIPWState, wS, hS, dS,
						xSdpi, ySdpi, cFrameBits,
						3, cColorCIELAB, NULL, &prD,
						&rightFrameColsToClear,
						&bottomFrameRowsToClear,
						&allocatedHere));
    ERR_PASS( w_pixrSetType( pIPWState, prD, cColorRGB ) );

	/* Now that we've got legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subR, &pL));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subG, &pA));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subR, &pRed));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subG, &pGreen));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pBlue));

    ERR_PASS_CLEANUP( i_Lab2RGB(pIPWState, pL, pA, pB,
				bplS, bplD, wS, hS,
				pRed, pGreen, pBlue));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/*

 $Log: color.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.49  1998/08/27 23:46:23  xzhu
 xzhu on Thu Aug 27 16:45:58 PDT 1998

 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 Wrap function for RGB to Lab strip by strip.

 Revision 1.48  1998/01/26 19:03:09  xzhu
 xzhu on Mon Jan 26 11:02:14 PST 1998

 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 Added color type in dest pixr when doing RGB->Lab and Lab->RGB transform.

 Revision 1.47  1998/01/23 23:17:24  xzhu
 xzhu on Fri Jan 23 15:16:34 PST 1998

 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 Created new wrap function for converting RGB to Lab and Lab to RGB.

 Revision 1.46  1998/01/14 00:11:01  ddavies
 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.45  1997/11/14 22:06:57  xzhu
 xzhu on Fri Nov 14 14:04:35 PST 1997

 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 Fixed bug not change color type after transform RGB color image into luminance gray image. (Reported by Andrew.)

 Revision 1.44  1997/11/13 22:51:35  aber
 aber on Thu Nov 13 14:50:49 PST 1997

 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 Add new wrapper for i_RGBToY2().

 Revision 1.43  1997/05/21 01:56:19  jcraig
 jcraig on Tue May 20 18:54:44 PDT 1997

 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 add new procedures w_RGBToLabInit, w_RGBToLabProcess and w_RGBToLabFinish
 which convert between RGB and L*a*b*

 Revision 1.42  1997/04/15 21:43:34  ddavies
 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 Substitute NoInit (uninitialized) version of pixr creation routines where
 possible.

 Revision 1.41  1997/04/01 21:53:38  dferrell
 dferrell on Tue Apr 1 13:52:03 PST 1997

 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 Added new wrapper function w_RGBToYFast () which computes a close
 approximation to the RGB to Y transform in about half the time it
 takes to do the exact transform.

 Revision 1.40  1996/02/13 17:28:54  dferrell
 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 Added parens around & operation for gcc 2.7.2

 Revision 1.39  1996/01/04 20:21:32  ejaquith
 color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.38  1995/12/13  04:05:35  ejaquith
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.37  1995/11/14  22:18:33  ddavies
 * ddavies on Tue Nov 14 14:18:32 PST 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Fix documentation bugs found by Eric's scripts.
 *
 * Revision 1.36  1995/11/13  20:10:02  ejaquith
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.35  1995/10/28  01:30:57  ddavies
 * ddavies on Fri Oct 27 18:30:31 PDT 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Change w_rgbToYIP so that it creates a clone of the green plane in the
 * original source image, then destroys the red, blue and green planes.  This
 * is the safe way to do this sort of slight-of-hand now since PIXR planes
 * have pointers.  If a pointer is intended to point to its own block and
 * we copy the pointer to a new block (as was done before), the pointer is
 * now pointing to some other (old, freed) block so things don't work.
 *
 * Revision 1.34  1995/10/13  21:59:04  ejaquith
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.33  1995/10/12  04:23:58  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.32  1995/10/05  19:37:46  ddavies
 * ddavies on Thu Oct 5 12:37:00 PDT 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Initialize pState in w_colorma
 * Ooops, in w_colorMaskInit so that w_RGBToLuvInit doesn't complain that it's
 * not NULL.
 *
 * Revision 1.31  1995/10/05  04:30:38  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 *
 * Revision 1.30  1995/09/21  22:51:55  ejaquith
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.29  1995/09/20  21:34:43  macleod
 * macleod on Wed Sep 20 14:33:07 PDT 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * w_colorMaskInit was passing "pState" instead of "&pState" to
 * w_RGBToLuvFinish().
 *
 * Revision 1.28  1995/09/15  18:01:59  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.27  1995/09/13  18:28:42  ddavies
 * ddavies on Wed Sep 13 11:26:47 PDT 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Fix bug in which pPrD was used instead of &prD in w_createPixr.
 *
 * Revision 1.26  1995/09/11  22:12:31  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.25  1995/09/10  00:35:45  ddavies
 * ddavies on Sat Sep 9 17:35:45 PDT 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.24  1995/09/07  01:29:59  ejaquith
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.23  1995/09/06  20:17:41  ejaquith
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.22  1995/09/01  01:15:13  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.21  1995/08/31  21:11:28  ejaquith
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.20  1995/08/29  23:58:54  ejaquith
 * ejaquith on Tue Aug 29 16:58:19 PDT 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Bug fix for "cmask" target in ipwtest.
 * Was passing in pixrGetImageAll() instead of pixrGetImage() to ipshared
 * routine.  (sigh, so much for stdization).
 *
 * Revision 1.19  1995/08/22  02:20:34  ejaquith
 * ejaquith on Mon Aug 21 19:20:14 PDT 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Nailed some warnings (PC).
 *
 * Revision 1.18  1995/08/17  22:03:49  ejaquith
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.17  1995/07/29  04:35:29  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.16  1995/02/28  18:51:54  macleod
 * macleod on Tue Feb 28 10:51:36 PST 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * abortI was being passed NULL instead of non-zero, which was both
 * incorrect and caused a warning.
 *
 * Revision 1.15  1995/02/25  00:25:47  macleod
 * macleod on Fri Feb 24 16:24:03 PST 1995
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Added RGBToYInPlace, which does a conversion from RGB to Y into the
 * green channel, and then promotes the green channel to the parent PIXR.
 * Destroys the red and blue channels, as well as the PIXR structure for
 * the green channel itself, which brings up a point: beware of aliased
 * pointers to sub-PIXRs.
 *
 * Revision 1.14  1994/10/07  19:19:19  lperry
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.13  1994/07/27  22:31:29  ejaquith
 * ejaquith on Wed Jul 27 15:28:51 PDT 1994
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Changed calls to RGBToLuv and RGBToLuvInit to contain new state information
 * pointer, as well as i_RGBToLuv and i_RGBToLuvInit calls.  Created function
 * RGBToLuvFinish to clean up state information.  Done to make i_color.c
 * in ipshared reentrant.
 *
 * Revision 1.12  1994/06/08  04:06:48  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.11  1994/05/20  21:31:28  macleod
 * macleod on Fri May 20 14:30:22 PDT 1994
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * RGBToY creates a grayscale PIXR rather than creating a 3-channel
 * color PIXR with luminance in the green channel, which was a
 * bit of a waste of space..
 *
 * Revision 1.10  1994/05/13  02:30:40  macleod
 * macleod on Thu May 12 19:30:28 PDT 1994
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * indexedToRGB takes 4-bit images
 *
 * Revision 1.9  1994/04/12  01:12:44  macleod
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.8  1994/04/05  00:39:38  macleod
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Changed all routines to take ColorPixr rather than three PIXRs
 *
 * Revision 1.7  1994/03/21  20:08:49  macleod
 * macleod on Mon Mar 21 12:08:30 PST 1994
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Was missing an #include file which caused PC version to not compile.
 *
 * Revision 1.6  1994/03/17  02:58:55  macleod
 * macleod on Wed Mar 16 18:55:32 PST 1994
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Changed colorMaskPr to use a different scheme. One first calls colorMaskInit
 * to create a lookup table indexed by the high 4 bits of each color channel.
 * The lookup table is created by computing the L*u*v* values for each color,
 * and computing a color distance (^2), which is compared against a
 * value supplied by the caller. There is a second parameter which
 * can range from 0 to 1, which controls the amount of color suppression
 * that happens in the dark regions. 0 does no suppression, 1 suppresses a lot.
 * Also added an RGBToLuv function. This function is extremely slow--only
 * useful for making tables.
 *
 * Revision 1.5  1993/12/15  20:24:36  macleod
 * macleod on Wed Dec 15 12:24:29 PST 1993
 *
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Added indexedToRGBPr()
 *
 * Revision 1.4  1993/12/03  01:13:35  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.3  1993/11/06  03:10:26  ddavies
 * color.c -> /product/ipcore/ipwrap/src/RCS/color.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

