/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  coords2.c:  includes subroutines
 *
 *     Transforms to/from PIXR:
 *
 *         COORDS  *locateCornerCoords():    finds the four cornermost pixels
 *
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"

/* Prototypes */
#include "shrcc.pub"
#include "props.pub"
#include "coords.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: coords2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

    /*  constants for placing marks on pixrects */
#define  WID_MARK        1
#define  LEN_MARK        7


/*************************************************************************
*************************************************************************/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find the corner coordinates of a black object in a binary image
 * 
 * LongDescription:::
 *  This routine searches in from the corners of a binary image
 *  to find a bounding rectangle for the first black image data found
 *  searching from each corner.
 *  First, this routine finds the first nonzero line starting at the
 *  top and bottom of the image and then finds the first nonzero word
 *  on the left and right of the image.  (Note this is nonzero word and
 *  so is not quite precise.)  These offset provide a bounding rectangle
 *  for the black in the image.  Finally, we find the four corners by
 *  scanning the image from each corner at a 45 degree angle and 
 *  looking for the first black pixel.  If useQuadrant is true, the bounding
 *  box is divided into quadrants and a corner will be found (if any 
 *  on pixels exists) in each quadrant.  If useQuadrant is false, we
 *  do not use quadrants - this is especially helpful if the image is
 *  relatively thin and skewed so that 2 corners may be in a single
 *  quadrant.
 *  Previously this routine divided the input pixr into 4 quadrants and
 *  a black pixel had to be found in each quadrant.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to search
 *  useQuadrant::    If true, look for a corner in each of the 4 quadrants
 *
 * Output:::
 *  ppCoords::       Coordinates of the four "corner" pixels
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_locateCornerCoords (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Bool                useQuadrant,
    COORDS            **ppCoords)
{
    PROLOGUE;
    UInt32	*linePtr, *pData;
    Int32	 width, height, Wpl, foundCorners;
    Int32	 boundX, boundY, minValue;
    Int32	 x, y, midX, midY, minDim;
    Int32	 upperY, lowerY, leftX, rightX;
    UInt32	*upperLinePtr, *lowerLinePtr;
    COORDS	*coords = NULL;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetHeight(pIPWState, prSrc, &height));
    ERR_PASS( w_pixrGetWidth(pIPWState, prSrc, &width));
    ERR_PASS( w_pixrGetBpl(pIPWState, prSrc, &Wpl));
    Wpl >>= 2;

	/* allocate COORDS structure  */
    ERR_PASS( w_createCoords(pIPWState, 4, &coords));

    foundCorners = 0;


	/* Skip blank lines at the top of the image */
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prSrc, (UInt8 **) &upperLinePtr));
    pData = upperLinePtr + (width-1)/32;
    upperY = 0;
    while ((upperY < height) && (*pData == 0))
    {
	pData = upperLinePtr + (width-1)/32;
	for (; (pData >= upperLinePtr) && (*pData == 0); pData--)
	    ;
	if (*pData == 0)	/* only increment if we haven't found
				 * the last line yet */
	{
	    upperY++;
	    upperLinePtr += Wpl;
	}
    }

    ERR_GEN_CLEANUP (upperY == height, cErrNoImageData);

	/* Skip blank lines at the bottom of the image */
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prSrc, (UInt8 **) &lowerLinePtr));
    lowerLinePtr += (height-1) * Wpl;
    pData = lowerLinePtr + (width-1)/32;
    lowerY = height-1;
    while ((lowerY > upperY) && (*pData == 0))
    {
	pData = lowerLinePtr + (width-1)/32;
	for (; (pData >= lowerLinePtr) && (*pData == 0); pData--)
	    ;
	if (*pData == 0)	/* only increment if we arent't stopping */
	{
	    lowerY--;
	    lowerLinePtr -= Wpl;
	}
    }

	/* Skip blank words on the left side of the image */
    pData = linePtr = upperLinePtr;
    leftX = 0;
    while ((leftX < width) && (*pData == 0))
    {
	for (y = upperY; (y <= lowerY) && (*pData == 0); pData += Wpl, y++)
	    ;
	if (*pData == 0)
	{
	    leftX += 32;
	    pData = ++linePtr;
	} 
    }

	/* Skip blank words on the right side of the image */
    rightX = (width - 1) & 0xFFFFFFE0;
    pData = linePtr = upperLinePtr + (rightX / 32);
    while ((rightX > leftX) && (*pData == 0))
    {
	for (y = upperY; (y <= lowerY) && (*pData == 0); pData += Wpl, y++)
	    ;
	if (*pData == 0)
	{
	    rightX -= 32;
	    pData = --linePtr;
	} 
    }
    /* at the end of the while loop, rightX it the bit offset to the
     * word which contains some non-zero bits.  we need to include those
     * bits so skip past them.  the easy way to do this is to add 32 and
     * point to the next word.  we do not want to do this if that would
     * point us past the end of the image, so check for this and set 
     * rightX to the width in that case.
     */
    if ((rightX + 32) > width)
	rightX = width;
    else rightX += 32;
    /* note that lowerY and upperY point exactly to the lines that
     * are non-zero and so to get the number of non-zero lines we must
     * add 1 after subtracting.  however, rightX points past the non-zero
     * bit and hence we do not want to subtract 1.  in fact, we cannot
     * subtract one because numbers may end up negative later.  for 
     * example, assume that we have a rectangle less than one word wide
     * (which is the minimum dimension).  then rightX - leftX + 1 is 33.
     * let's say that you try to find the upper right corner not using
     * quadrants.  midX is calculated as rightX - minDim which will 
     * be negative if rightX is only 32.
     */
    minDim = MIN(rightX - leftX, lowerY - upperY + 1);


	/* find upper-left corner */
    coords->x[0] = leftX;
    coords->y[0] = upperY;
    if (useQuadrant == cTrue)
    {
	midX = (rightX + leftX) / 2;
	midY = (lowerY + upperY) / 2;
    }
    else
    {
	midX = leftX + minDim;
	midY = upperY + minDim;
    }
    boundX       = midX;
    boundY       = midY;
    minValue     = boundX + boundY;
    linePtr	 = upperLinePtr;
    for (y = upperY; y < boundY; y++, linePtr += Wpl)
    {
	if (i_nextBitToTheRight (pIPWState, (UInt8 *)linePtr,
				 boundX, leftX, &x)
		== cSuccessful)
	{
	    if (minValue > (x + y))
	    {
		minValue     = x + y;
		coords->x[0] = x;
		coords->y[0] = y;
		boundX       = x;	/* since we started search at top, 
				       we won't need to look any farther
				       than x in following rows */
		boundY       = MIN(midY, x + y);
	    }
        }
    }
    if (minValue < (midX + midY))
        foundCorners++;


	/* find upper-right corner */
    coords->x[1] = coords->x[0];
    coords->y[1] = coords->y[0];
    if (useQuadrant == cTrue)
    {
	midX = (rightX + leftX) / 2;
	midY = (lowerY + upperY) / 2;
    }
    else
    {
	midX = rightX - minDim;
	midY = upperY + minDim;
    }
    boundX       = midX;
    boundY       = midY;
    minValue     = width - boundX + boundY;
    linePtr	 = upperLinePtr;
    for (y = upperY; y < boundY; y++, linePtr += Wpl)
    {
	if (i_nextBitToTheLeft (pIPWState, (UInt8 *)linePtr,
				boundX, width-1, &x)
		== cSuccessful)
	{
	    if (minValue > (y + width - x))
	    {
		minValue = y + width - x;
		coords->x[1] = x;
		coords->y[1] = y;
		boundX = x;
		boundY = MIN(midY, width - x + y);
	    }
        }
    }
	/* We test minValue + width) because to be perfectly correct,
	 * we should have been setting minValue to y + width - x all
	 * along.
	 */
    if ((minValue) < (width - midX + midY))
        foundCorners++;


	/* find lower-left corner */
    coords->x[2] = coords->x[0];
    coords->y[2] = coords->y[0];
    if (useQuadrant == cTrue)
    {
	midX = (rightX + leftX) / 2;
	midY = (lowerY + upperY) / 2;
    }
    else
    {
	midX = leftX + minDim;
	midY = lowerY - minDim;
    }
    boundX       = midX;
    boundY       = midY;
    minValue     = boundX + height - boundY;
    linePtr	 = lowerLinePtr;
    for (y = lowerY; y >= boundY; y--, linePtr -= Wpl)
    {
	if (i_nextBitToTheRight (pIPWState, (UInt8 *)linePtr,
				 boundX, leftX, &x)
		== cSuccessful)
	{
	    if (minValue > (x + height - y))
	    {
		minValue = x + height - y;
		coords->x[2] = x;
		coords->y[2] = y;
		boundX       = x;
		boundY       = MAX(midY, y - x);
	    }
        }
    }
    if (minValue < (midX + height - midY))
        foundCorners++;


	/* find lower-right corner */
    coords->x[3] = coords->x[2];
    coords->y[3] = coords->y[2];
    if (useQuadrant == cTrue)
    {
	midX = (rightX + leftX) / 2;
	midY = (lowerY + upperY) / 2;
    }
    else
    {
	midX = rightX - minDim;
	midY = lowerY - minDim;
    }
    boundX       = midX;
    boundY       = midY;
    minValue     = boundX + boundY;
    linePtr	 = lowerLinePtr;
    for (y = lowerY; y >= boundY; y--, linePtr -= Wpl)
    {
	if (i_nextBitToTheLeft (pIPWState, (UInt8 *)linePtr,
				boundX, width-1, &x)
		== cSuccessful)
	{
	    if (minValue < (x + y))
	    {
		minValue = x + y;
		coords->x[3] = x;
		coords->y[3] = y;
#if 0
		boundX       = width - x;
		boundY       = MIN(midY, width - x + y);
#endif
		boundX	     = x;
		boundY	     = MAX(midY, y - (width - x));
	    }
        }
    }
    if (minValue > (midX + midY))
        foundCorners++;
    
    /* Must be at least 3 corners */
    ERR_GEN_CLEANUP (foundCorners < 3, cErrHeuristic_foundNothing);
    *ppCoords = coords;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (coords)
	w_destroyCoords(pIPWState, &coords);

    ERR_RETURN;
}


/*

 $Log: coords2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.32  1998/01/14 00:11:05  ddavies
 coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.31  1996/01/04 20:21:37  ejaquith
 coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.30  1995/12/13  04:05:42  ejaquith
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.29  1995/11/13  20:10:09  ejaquith
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.28  1995/10/13  21:59:10  ejaquith
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.27  1995/10/03  21:12:46  macleod
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.26  1995/09/21  22:52:01  ejaquith
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.25  1995/09/11  21:55:12  macleod
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.24  1995/09/06  20:17:48  ejaquith
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.23  1995/09/01  20:43:24  macleod
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Added back #include files
 *
 * Revision 1.22  1995/08/31  21:11:33  ejaquith
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.21  1995/08/29  21:53:38  ddavies
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.20  1995/08/25  01:35:47  ejaquith
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Fixed warnings
 *
 * Revision 1.19  1995/08/17  21:54:22  macleod
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.18  1995/05/12  17:55:11  jcraig
 * jcraig on Fri May 12 10:54:09 PDT 1995
 *
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * fix locateCornerCoords so that it will not fail if the rectangle is less than
 * one word wide and quadrants are not being used.
 *
 * Revision 1.17  1995/04/13  01:34:18  jcraig
 * jcraig on Wed Apr 12 18:31:34 PDT 1995
 *
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * add new parameter to locateCornerCoords: useQuadrant.  If
 * useQuadrant is true, the bounding box of on pixels is divided into
 * quadrants and a corner will be found (if any on pixels exists) in
 * each quadrant.  This is the old behavior.  If useQuadrant is false,
 * do not use quadrants - this is especially helpful if the rectangle
 * is relatively thin and skewed so that 2 corners may be in a single
 * quadrant.
 *
 * Revision 1.16  1995/01/11  22:39:09  dferrell
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Changed include of shrcmprs.pub to shrcc.pub.
 *
 * Revision 1.15  1994/10/06  20:53:13  lperry
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.14  1994/09/16  17:34:36  jcraig
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * remove include of bitacces.h which was not needed.
 *
 * Revision 1.13  1994/09/15  00:58:35  jcraig
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * when checking on the right for the first nonzero column, the width
 * is in bytes and must be divided by 32 (not 4) when added to the
 * word pointer.
 *
 * Revision 1.12  1994/09/01  18:58:52  jcraig
 * jcraig on Thu Sep 1 11:57:23 PDT 1994
 *
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * in locateCornerCoords when skipping columns on the left and right
 * side of the input image, don't reference off the end of the image -
 * actually increment the loop counter and look at the correct bits....
 *
 * Revision 1.11  1994/08/31  21:54:12  jcraig
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * change locateCornerCoords so that it looks for the first and last row
 * and column that contain black pixels and uses those to set the
 * quadrants to search for corners (instead of dividing the entire input
 * image into quadrants).  this should only change cases in which
 * the black pixels were localized instead of spread over the entire
 * input image - should find "corners" instead of returning with an error
 * because not enough corners could be found.
 *
 * Revision 1.10  1993/12/17  22:37:41  ddavies
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Fix locateCornerc
 * Ok, start over.  Fix locateCornerCoords.  It was skipping blank lines at the
 * bottom of an image correctly, but returning the y coords of the lower left and
 * lower right points as though the line at the top of the bottom white space was
 * the bottom line of the image.  Also generally simplified the use of variables
 * and the calculations done (I think).
 *
 * Revision 1.9  1993/12/03  01:13:44  ddavies
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.8  1993/11/23  00:41:10  ddavies
 * ddavies on Mon Nov 22 16:40:27 PST 1993
 *
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Fix bug in locateCornerCoords.  I had removed a for loop without removing
 * the corresponding "break" in the middle of it.
 *
 * Revision 1.7  1993/11/20  03:06:35  ddavies
 * ddavies on Fri Nov 19 19:05:10 PST 1993
 *
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Remove getFirstONBinInWord and getLastONBitInWord.  Change locateCornerCoords
 * so that it uses i_nextBitToTheRight and i_nextBitToTheLeft instead.
 *
 * Revision 1.6  1993/11/06  03:10:40  ddavies
 * coords2.c -> /product/ipcore/ipwrap/src/RCS/coords2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

