/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * crop.c
 *
 *	"Auto-crop" public functions
 *
 * Functions defined:
 *
 *	crop () - Find a crop box for an image
 *	crop4 () - Find a 4-point "box" (not necessarily a rectangle,
		   can be skewed) for an image
 */




#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"

/* prototypes */
#include "props.pub"
#include "crop.pub"
#include "crop.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: crop.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   find a crop box from a grayscale image
 *
 * LongDescription:::
 *	Takes a grayscale image and finds a crop box. If you want to
 *	crop an RGB image, convert it to luminance first (use RGBToY), 
 *	and then call this.
 *      <BR>
 *      "crop" means find the thing on the platen, i.e. a photograph.
 *      This was initially intended to find pieces of paper as well,
 *      but that was problematic due to the fact that platens are
 *      white, and the difference between the paper and platen was sometimes
 *      smaller than the scanner noise level.
 *      <BR>
 *	The crop routine we call finds a four-point crop; we just compute
 *	the bounding box from that.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to analyze
 *  rect::           Region of interest
 *
 * Output:::
 *  error::          Error in fitting crop box, 0=perfect
 *  cropRect::       Bounding rectangle of crop box found
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_crop (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *rect,
    Float64            *error,
    RECTANGL           *cropRect)
{
    PROLOGUE;
    Int32   w, h, x, y, points[8], xMin, yMin, xMax, yMax, i, depth,
	    frame, bpl, dpi;
    UInt8  *pImgAll;
    Float64 theta;
    ScaledInt1616    xDPI, yDPI;


    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
		  cChannels1 | cChannels3, cFrame32);
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    if (!rect)
    {
	x = 0;
	y = 0;
	ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &h));
    }
    else
    {
	x = rect->x;
	y = rect->y;
	w = rect->w;
	h = rect->h;
    }

    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));
    ERR_PASS (w_pixrGetImageAll (pIPWState, prSrc, &pImgAll));
    ERR_PASS (w_pixrGetBpl (pIPWState, prSrc, &bpl));
    ERR_PASS (w_pixrGetFrame (pIPWState, prSrc, &frame));
    ERR_PASS (wp_do_crop (pIPWState, pImgAll, bpl, depth, x + frame, y + frame,
		       w, h, dpi, error, &theta, 
		       &points[0], &points[1], &points[2], &points[3], 
		       &points[4], &points[5], &points[6], &points[7]));

    /* for the purposes of cropping, we find the extent */
    xMin = 0x10000000;
    yMin = 0x10000000;
    xMax = -0x10000000;
    yMax = -0x10000000;
    for (i = 0; i < 4; i++)
    {
	if (points[i*2] > xMax)	xMax = points[i*2];
	if (points[i*2] < xMin) xMin = points[i*2];
	if (points[i*2+1] > yMax) yMax = points[i*2+1];
	if (points[i*2+1] < yMin) yMin = points[i*2+1];
    }

    cropRect->x = xMin - frame;
    cropRect->y = yMin - frame;
    cropRect->w = xMax - xMin + 1;
    cropRect->h = yMax - yMin + 1;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   find a crop box from a grayscale image
 *
 * LongDescription:::
 *	Takes a grayscale image and finds a crop box. If you want to
 *	crop an RGB image, convert it to luminance first (use RGBToY), 
 *	and then call this.
 *      <BR>
 *      "crop" means find the thing on the platen, i.e. a photograph.
 *      This was initially intended to find pieces of paper as well,
 *      but that was problematic due to the fact that platens are
 *      white, and the difference between the paper and platen was sometimes
 *      smaller than the scanner noise level.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to analyze, 4- or 8-bits/pixel
 *  rect::           Region of interest
 *
 * Output:::
 *  error::          Error in fitting crop box
 *  angle::          Angle required to correct the skew, positive angle
 *		     means rotate clockwise
 *  cropPoints::     Corner points of crop box found
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_crop4 (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *rect,
    Float64            *error,
    Float64            *angle,
    COORDS             *cropPoints)
{
    PROLOGUE;
    Int32   w, h, x, y, points[8],  i, depth,
	    frame, bpl, dpi;
    UInt8  *pImgAll;
    ScaledInt1616    xDPI, yDPI;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
		  cChannels1 | cChannels3, cFrame32);

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    if (!rect)
    {
	x = 0;
	y = 0;
	ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &h));
    }
    else
    {
	x = rect->x;
	y = rect->y;
	w = rect->w;
	h = rect->h;
    }

    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));
    ERR_PASS (w_pixrGetImageAll (pIPWState, prSrc, &pImgAll));
    ERR_PASS (w_pixrGetBpl (pIPWState, prSrc, &bpl));
    ERR_PASS (w_pixrGetFrame (pIPWState, prSrc, &frame));
    ERR_PASS (wp_do_crop (pIPWState, pImgAll, bpl, depth, x + frame, y + frame,
		       w, h, dpi, error, angle,
		       &points[0], &points[1], &points[2], &points[3], 
		       &points[4], &points[5], &points[6], &points[7]));

    /* Since wp_do_crop returns the angle of the crop box, negate it, to
     * return the clockwise rotation necessary to correct. */
    *angle *= -1.0;

    for (i = 0 ; i < 4; i++)
    {
	cropPoints->x[i] = points[i * 2] - frame;
	cropPoints->y[i] = points[i * 2 + 1] - frame;
    }

    NORMAL_RETURN;
}
/*

 $Log: crop.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.32  1998/02/25 19:25:04  aber
 aber on Wed Feb 25 11:24:27 PST 1998

 crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v

 I should have compiled first. Pointer value needs to be dereferenced.

 Revision 1.31  1998/02/25 19:20:14  aber
 aber on Wed Feb 25 11:20:14 PST 1998

 crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v

 CommonCheck: To be consistent, the skew detection functions return the angle required
 to correct skew. This is the same angle passed on to the rotation algorithm
 itself, and is positive is clockwise rotation is necessary. w_findSkewSample
 does this also, but w_crop4 was returning the angle of skew, not the angle
 to repair the skew. Both w_crop4 and w_crop5 now return the angle to
 correct for skew.

 Revision 1.30  1996/01/04 20:21:42  ejaquith
 crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.29  1995/12/13  04:05:49  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.28  1995/11/16  19:45:51  macleod
 * macleod on Thu Nov 16 11:45:42 PST 1995
 *
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Included crop.prv
 *
 * Revision 1.27  1995/11/13  20:10:16  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.26  1995/10/13  21:59:18  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.25  1995/10/03  01:59:06  macleod
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.24  1995/09/21  22:52:08  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.23  1995/09/19  19:19:15  macleod
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.22  1995/09/11  21:54:00  macleod
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.21  1995/09/07  01:30:10  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.20  1995/09/06  20:17:54  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.19  1995/09/01  21:02:24  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.18  1995/08/31  21:11:39  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.17  1995/08/25  01:35:50  ejaquith
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Fixed warnings
 *
 * Revision 1.16  1995/08/17  21:54:27  macleod
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.15  1994/10/28  02:03:55  macleod
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * crop4 is passed a pointer to a COORDS structure, which should have
 * been allocated, along with some individual points, by createCoords().
 * crop4 was, for some reason, allocating the points again, stomping
 * on what was there. So it doesn't anymore. What this means is that
 * the COORDS thing passed to crop4 must have its components allocated,
 * which createCoords() will do.
 *
 * Revision 1.14  1994/10/07  19:06:14  lperry
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.13  1994/09/29  00:45:59  lperry
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Eliminate use of utilpros. Use utils.pub instead.
 *
 * Revision 1.12  1994/09/21  21:39:09  macleod
 * macleod on Wed Sep 21 14:38:00 PDT 1994
 *
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * crop4 computes a skew angle. This is an additional parameter to crop4,
 * so those calling crop4 will need to modify their code
 *
 * Revision 1.11  1994/07/27  21:52:00  ejaquith
 * ejaquith on Wed Jul 27 14:51:38 PDT 1994
 *
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Removed #include "iadef.prv"
 *
 * Revision 1.10  1994/06/08  04:06:59  ddavies
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.9  1994/03/23  02:57:31  macleod
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Fixed a couple of places where it returned NULL for an error code
 * when it should have returned non-zero
 *
 * Revision 1.8  1994/02/09  00:04:08  dferrell
 * dferrell on Tue Feb 8 16:03:51 PST 1994
 *
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Moved private functions to cropprv.c
 *
 * Revision 1.7  1994/02/04  19:29:57  macleod
 * macleod on Fri Feb 4 11:29:18 PST 1994
 *
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Moved all of the internals from ipshared into here
 *
 * Revision 1.6  1994/01/13  02:19:24  macleod
 * macleod on Wed Jan 12 18:19:16 PST 1994
 *
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Added a dpi parameter to crop
 *
 * Revision 1.5  1994/01/02  22:45:02  macleod
 * macleod on Sun Jan 2 14:44:49 PST 1994
 *
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * auto-crop can take a 4- or 8-bit pixr
 *
 * Revision 1.4  1993/12/03  01:13:57  ddavies
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.3  1993/11/12  01:22:00  macleod
 * macleod on Thu Nov 11 17:21:50 PST 1993
 *
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Added crop4, which does a "four-point" crop
 *
 * Revision 1.2  1993/11/06  03:10:54  ddavies
 * crop.c -> /product/ipcore/ipwrap/src/RCS/crop.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

