/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * crop2.c
 *
 *	Another "Auto-crop"
 *
 * Functions defined:
 *
 *	crop2 () - Find a crop box for an image
 *	crop2_4 () - Find a 4-point "box" (not necessarily a rectangle,
 *		   can be skewed) for an image
 *	documentBoundingBox () - Find the bounding box (no rotation) for 
 *				 a binary image
 *	photoBoundingBox () - Find the bounding box (no rotatoin) for
 *			      a photo
 *
 * Notes:
 * 
 * The first crop function, which looked for low- and high-contrast
 * edge points and fit edges by linear regression, worked OK, but had 
 * its limitations--it didn't deal well when there was a page with a lot
 * of elements and a very light background: the page was missed, and the
 * elements were missed, so it defaulted to the whole platen.
 *
 * Since we now have a decent background threshold finder which works
 * reasonably well even on an un-cropped image, I'm going to use that
 * for this next method. Basically, there are two cases I'm going to handle:
 * The case where there is text, and the case where there is no text.
 * In the first case, since we can threshold the image, we can then:
 *	- detect the skew angle
 *	- de-skew the binary image
 *	- do textured reductions to get a reduced image, probably
 *	  (1, ..., 4) to get things to clump together and then to
 *	  get rid of single-pixel things (the "..." are as many ones
 *	  as needed to get it to a ~100 pixel wide image.
 *	- Get rid of all blobs that touch the edge, by dropping seeds
 *	  around the edge. This is needed to eliminate funny stuff on 
 *	  the edges of platens.
 *	- Close stuff up a bit
 *	- march down using nextXXXToTheRight, to find the first "stuff"
 *	  along each of the four edges.
 *	  Note that since we deskewed, we need to be careful not to include
 *	  the deskewed part, which isn't really part of the image. 
 *	  Perhaps we should do the rotation after the morphology and
 *	  edge crap elimination.
 *	- Rotate and scale the resulting box back to get crop points
 *	  in the coordinate system of the original.
 *
 * In the second case, we might want to use the technique of the original
 * crop, only using high-contrast edges, and without doing any region
 * averaging. It would be nice if we could just threshold and find the
 * corners, but there are two problems with this. First, we can't really
 * threshold things without text very well, and second, we can't reliably
 * eliminate stuff on the edges of the platen. For example, I was given
 * a sample which had a photo in the upper left, and extending out of the
 * photo to the right was a big black band which was where the auto-document
 * -feeder belt curved away from the glass. The original scheme worked on
 * this and similar images, just not very well on documents.
 ************************************************************************/

#include <math.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "thrshold.h"
#include "morphg.h"
#include "shrscgr.pub"
#include "defines.pub"
#include "alplib.h"
#include "shrcc.pub"
#include "logicops.h"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "crop.pub"
#include "crop.prv"
#include "thrshold.pub"
#include "reduce.pub"
#include "prxform.pub"
#include "morph.pub"
#include "expand.pub"
#include "strel.pub"
#include "skew.pub"
#include "fill.pub"
#include "intops.pub"
#include "shrcc.pub"
#include "coords.pub"
#include "rotatop.pub"
#include "logicops.pub"
#include "fthresh.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: crop2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find a crop box from a grayscale image
 *
 * LongDescription:::
 *	Takes a grayscale image and finds a crop box. If you want to
 *	crop an RGB image, convert it to luminance first (use RGBToY), 
 *	and then call this.
 *      <BR>
 *	The crop routine we call finds a four-point crop; we just compute
 *	the bounding box from that.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          The image to search
 *  rect::           Area out of which to crop, NULL = look at whole image
 *  thresh::         If <0, find a threshold; otherwise, use this as threshold
 *
 * Output:::
 *  error::          Error in fitting crop box
 *  cropRect::       Crop rectangle found
 *
 * Returns:::
 *  Error status
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications::: 
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_crop2 (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *rect,
    Int32               thresh,
    Float64            *error,
    RECTANGL           *cropRect)
{
    PROLOGUE;
    Int32   xMin, yMin, xMax, yMax, i;
    COORDS  *points = NULL;
    Float64  theta;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
		  cChannelsAny, cFrame32);


    ERR_PASS (w_createCoords(pIPWState, 4, &points));
    ERR_PASS_CLEANUP (wp_do_crop2 (pIPWState, prSrc, rect, thresh, error, 
				  &theta, points));

    /* for the purposes of cropping, we find the extent */
    xMin = 0x10000000;
    yMin = 0x10000000;
    xMax = -0x10000000;
    yMax = -0x10000000;
    for (i = 0; i < 4; i++)
    {
	if (points->x[i] > xMax)    xMax = points->x[i];
	if (points->x[i] < xMin)    xMin = points->x[i];
	if (points->y[i] > yMax)    yMax = points->y[i];
	if (points->y[i] < yMin)    yMin = points->y[i];
    }

    cropRect->x = xMin /*- pixrGetFrame (image) */ ;
    cropRect->y = yMin /* - pixrGetFrame (image) */;
    cropRect->w = xMax - xMin + 1;
    cropRect->h = yMax - yMin + 1;

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyCoords(pIPWState, &points);

    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Find a 4-point crop box
 *
 * LongDescription:::
 *	Takes a grayscale image and finds a four-point crop box.
 *	If you want to crop an RGB image, convert it to luminance 
 *	first (use RGBToY), and then call this.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to crop
 *  rect::           Look inside this rect, if NULL, look at whole image
 *  thresh::         If <0, find a threshold; otherwise, use this as threshold
 *
 * Output:::
 *  error::          Error in fitting crop box, 0 = perfect fit
 *  angle::          Angle at which crop box lies, in radians
 *  cropPoints::     four points (ul, ur, ll, lr) of bounding box.
 *
 * Returns:::
 *  error status
 * 
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_crop2_4 (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *rect,
    Int32               thresh,
    Float64            *error,
    Float64            *angle,
    COORDS             *cropPoints)
{
    PROLOGUE;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
		  cChannelsAny, cFrame32);
    
    ERR_PASS (wp_do_crop2 (pIPWState,  prSrc, rect, thresh, error, 
			angle, cropPoints));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Given a grayscale image, determine a crop box.
 *
 * LongDescription:::
 *	Given an image, determine a crop box. 
 *
 *	Note: this can find a "rotated" crop box. Figuring out the extent box
 *	is obviously straightforward, but I'll probably write another function
 *	to do it just for convenience, which will have this one do the work.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to search
 *  cropRect::       Rectangle in which to look; if NULL examine entire img.
 *  thresh::         If <0, find a threshold, otherwise use this as threshold
 *
 * Output:::
 *  err::            Error term, i.e. goodness-of-fit, 0.0=it fit exactly
 *  angle::          Angle (radians) at which the crop box lies
 *  cropPoints::     Points of crop box found, ul, ur, ll, lr
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_do_crop2 (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *cropRect,
    Int32               thresh,
    Float64            *err,
    Float64            *angle,
    COORDS             *cropPoints)
{
    PROLOGUE;
    PIXR    *thresholded = NULL;
    Int32    backThresh, depth, confidence, redFactor, scaleConf,
	     black, white, dpi;
    Bool     invert, probablyText;
    Float32  skew;
    PIXR    *irrelevant = NULL;
    ScaledInt1616    xDPI, yDPI;

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));
    if (depth == 1)
    {
	/* Need not threshold; it's already done, obviously */
	thresholded = prSrc;
	backThresh = -1;
    }
    else
    {
	if (thresh < 0)
	{
	    ERR_PASS (w_findBackgroundThreshold (pIPWState, prSrc, cropRect, 
						 cFalse, &backThresh, &black,
						 &white, &invert,
						 &probablyText, 
						 &irrelevant));
	}
	else
	{
	    backThresh = thresh;
	    probablyText = cTrue;
	}

	/* If we go outside a reasonable range, make it arbitrary */
	if (depth == 4 && (backThresh <= 4 || backThresh >= 13))
	{
		backThresh = 8;
	}
	else if (depth == 8 && (backThresh < 64 || backThresh > 220))
	{
	    backThresh = 128;
	}
	else
	    /* Hack: back off a little */
	    backThresh = (backThresh * 3) / 4;

	/* Threshold it */
	ERR_PASS (w_threshold (pIPWState, prSrc, 0, backThresh,
				THRESH_INVERSE, &thresholded));

    }

    /* Try to find a skew angle */
    if (dpi <= 100)
    {
	redFactor = 2;
	scaleConf = 4;
    }
    else if (dpi <= 200)
    {
	redFactor = 4;
	scaleConf = 3;
    }
    else if (dpi <= 300)
    {
	redFactor = 8;
	scaleConf = 2;
    }
    else if (dpi <= 600)
    {
	redFactor = 16;
	scaleConf = 1;
    }
    else
    {
	redFactor = 32;
	scaleConf = 1;
    }
   
    if (probablyText == cFalse)
    {
	/* Not a document.. */
	ERR_PASS (w_cropByPoints (pIPWState, prSrc, cropRect, err, angle, 
				  cropPoints));
    }
    else
    {
	ERR_PASS (w_findSkewSample (pIPWState, thresholded, redFactor, 
				  &confidence, &skew));
	*angle = skew;

	/* It's a document! */
	ERR_PASS (w_cropReduced (pIPWState, thresholded, cropRect,
			         skew, err, cropPoints));
	if (backThresh >= 0)
	    ERR_PASS (w_destroyPixr (pIPWState, &thresholded));
    }

    if (backThresh >= 0 && thresholded != NULL)
    {
	ERR_PASS (w_destroyPixr (pIPWState, &thresholded));
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Reduced-resolution cropping algorithm.
 *
 * LongDescription:::
 *   This is a cropping algorithm more suited to cropping documents 
 *   than photographs.
 *
 * PrivateDescription:::
 *	We have a thresholded image and a 
 *	skew angle. Make a reduced version, get rid of stuff around the
 *	edges, maybe clean it up a little, deskew it, and then march
 *	in from the edges and find the first stuff we run into.
 *	
 *	The edges are then moved out 1/2", to give a nice border around
 *	stuff, and then transformed back to the input coordinate space.
 *	One nice thing to do might be to count the # of connected components,
 *	and not put the border around if there is only one. Another nice
 *	thing to do might be to have a set of default paper sizes, and
 *	if we find that our rectangle is near one of them, set the 
 *	rectangle to that paper size.
 *
 * Input:::
 *  pIPWState::      State structure
 *  imageP::         Image to crop
 *  cropRect::       Rectangle in which to look
 *  skew::           Skew angle of document
 *
 * Output:::
 *  err::            Error of fit, which is always 0
 *  cropPoints::     Crop box, ul, ur, ll, lr
 *
 * Returns:::
 *  error status
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *  stable
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cropReduced (
    IPWState           *pIPWState,
    PIXR               *imageP,
    RECTANGL           *cropRect,
    Float32             skew,
    Float64            *err,
    COORDS             *cropPoints)
{
    PROLOGUE;
    PIXR	*reduced = NULL, *rotated = NULL, *cropMask = NULL, *tmp = NULL;
    Int32	 reduction, i, rW, rH,
		 top, left, bottom, right, bpl, w, h,
		 topx, bottomx, lefty, righty, wOrig, hOrig, edgeSlop, dpi;
    Int32	 bit;
    RECTANGL	 localRect;
    Float64	 tx, ty, theta, cx, cy, margin, delta;
    UInt8	*ip;
    ScaledInt1616    xDPI, yDPI;

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, imageP, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    *err = 0.0;

    if (cropRect == NULL)
    {
	cropRect = &localRect;
	cropRect->x = 0;
	cropRect->y = 0;
	ERR_PASS (w_pixrGetWidth (pIPWState, imageP, &(cropRect->w)));
	ERR_PASS (w_pixrGetHeight (pIPWState, imageP, (&cropRect->h)));
    }

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, imageP, &wOrig));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, imageP, &hOrig));
    cx = (Float64) wOrig / 2.0;
    cy = (Float64) hOrig / 2.0;
    theta = -skew;

    /* Do the reduction, which is resolution-dependent. Do
     * all but the last reductions at threshold 1, which makes dark things get
     * darker, and then do the last one at threshold 4, which gets rid
     * of single-pixel things. 
     */
    if (dpi <= 100)
    {
	ERR_PASS (w_reduce4 (pIPWState, imageP, 1, 4, &reduced));
	reduction = 4;
    }
    else if (dpi <= 200)
    {
	/* 8x reduction */
	ERR_PASS (w_reduce8 (pIPWState, imageP, 1, 1, 4, &reduced));
	reduction = 8;
    }
    else
    {
	/* 16x reduction */
	ERR_PASS (w_reduce16 (pIPWState, imageP, 1, 1, 1, 4, &reduced));
	reduction = 16;
    }

    /* Fill in the edges. NOTE that we don't use the crop box here, since
     * we really are concerned with the scanner edges. Maybe this
     * step should be an option, just in case one knows that the
     * image is already cropped. 
     */
    /* Invert it first, so we don't make it all  black */
    ERR_PASS_CLEANUP (w_invert (pIPWState, reduced, &reduced));
    ERR_PASS_CLEANUP (w_copy (pIPWState, reduced, &tmp));

    /* Fill in everything along the edges that's not black */
    ERR_PASS_CLEANUP (w_edgeFillSweep4 (pIPWState, tmp, 10, &tmp));

    /* Get just the blown-away parts */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, reduced, SUBTRACT, tmp));

    /* 3/22/93: Just mask out the middle, so that we only get rid of stuff 
       along the edges. For now, take 3%  */
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, reduced, &rW));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, reduced, &rH));
    edgeSlop = MAX (rW, rH);
    edgeSlop = (edgeSlop * 3) / 100;
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
				  edgeSlop, edgeSlop, 
				  rW - edgeSlop*2, rH - edgeSlop*2, tmp));


    /* Now, we have just a mask (black stuff) to blow away in "reduced"*/
    /* Invert 'em back */
    ERR_PASS_CLEANUP (w_invert (pIPWState, tmp, &tmp));
    ERR_PASS_CLEANUP (w_invert (pIPWState, reduced, &reduced));

    /* Blow away the stuff around the edges that we don't want */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, tmp, AND, reduced));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmp));

    /* Blow away stuff that's not in our crop rectangle */
    /* Make a mask for the crop rectangle. */
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, reduced, &cropMask));

    /* Set crop mask region. */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_SET,
				  cropRect->x / reduction, 
				  cropRect->y / reduction,
				  cropRect->w / reduction,
				  cropRect->h / reduction,
				  cropMask));

    /* AND it to get rid of non-cropped stuff*/
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, cropMask, INTERSECTION, reduced));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &cropMask));

    /* Rotate the thing. Not "nice". */
    ERR_PASS_CLEANUP (w_rotateArb (pIPWState, reduced, skew, 0, 1, &rotated));

    /* Search for the first "on" pixel in each direction */
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, rotated, &bpl));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, rotated, &ip));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, rotated, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, rotated, &h));

    /* Top: search down until we find a line with a pixel on it */
    for (i = 0; i < h; i++)
    {
	if (i_nextBitToTheRight (pIPWState, ip, w, 0, &bit) == cSuccessful)
	    break;
	ip += bpl;
    }

    ERR_GEN (i == h, cErrHeuristic_foundNothing);

    if(i >0) top = i - 1;
    else top =0;
    topx = bit;

    /* Bottom: search up until we find a line with a pixel on it */
    ERR_PASS (w_pixrGetImage (pIPWState, rotated, &ip));
    ip += bpl * (h - 1);

    for (i = h - 1; i >= 0; i--)
    {
	if (i_nextBitToTheRight (pIPWState, ip, w, 0, &bit) == cSuccessful)
	    break;
	ip -= bpl;
    }

    ERR_GEN_CLEANUP (i < 0, cErrHeuristic_foundNothing);

    if(i < (h-1)) bottom = i + 1;
    else bottom = i;
    bottomx = bit;

    /* For left and right, look on each line and find the min and max.
     * Two things could be faster: 1, just looking for the first column
     * with a "1" in it, and 2, only looking in the crop region 
     */
    ERR_PASS_CLEANUP (w_pixrGetWidth(pIPWState, rotated, &left));
    right = 0;
    lefty = righty = 0;
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, rotated, &ip));

    for (i = 0; i < h; i++)
    {
	if (i_nextBitToTheRight (pIPWState, ip, w, 0, &bit) == cSuccessful)
	{
	    if (bit < left)
	    {
		left = bit;
		lefty = i;
	    }
	}

	if (i_nextBitToTheLeft (pIPWState, ip, 0, w - 1, &bit) == cSuccessful)
	{
	    if (bit > right)
	    {
		right = bit;
		righty = i;
	    }
	}

	ip += bpl;
    }

    if(left >0) left--; 
    if(right < (w-1)) right++;

/* fprintf(stderr, "top %d bottom %d left %d right %d\n", top, bottom, left, right); */
    /* we now have top, left, bottom, and right. Make some points and 
     * transform them back to the original coordinate space.
     * We must not forget to compensate for the offset to the left and
     * top edges when we rotated the points.
     */

    left *= reduction;
    right *= reduction;
    top *= reduction;
    bottom *= reduction;
    ERR_PASS_CLEANUP (w_rotatedBoxOffset (pIPWState, wOrig, hOrig, skew, 
					lefty * reduction, &delta));
    left -= (Int32) delta;
    ERR_PASS_CLEANUP (w_rotatedBoxOffset (pIPWState, wOrig, hOrig, skew, 
					righty * reduction, &delta));
    right -= (Int32) delta;
    ERR_PASS_CLEANUP (w_rotatedBoxOffset (pIPWState, hOrig, wOrig, skew,
					topx * reduction, &delta));
    top -= (Int32) delta;
    ERR_PASS_CLEANUP (w_rotatedBoxOffset (pIPWState, hOrig, wOrig, skew, 
					bottomx * reduction, &delta));
    bottom -= (Int32) delta;

    /* Transform them back. Not so efficient, but it only happens 4 times. */
    margin = (Float64) dpi / 2.0;

#define UNREDUCE(xr, yr, m, mx, my, p) \
tx = (Int32) (xr - cx + (mx * margin)); \
ty = (Int32) (yr - cy + (my * margin)); \
cropPoints->x[p] = (Int32) (tx * cos(theta) - ty * sin(theta) + cx); \
cropPoints->y[p] = (Int32) (tx * sin(theta) + ty * cos(theta) + cy);

    /* Upper left */
    UNREDUCE(left, top, margin, -1, -1, 0);
    UNREDUCE(right, top, margin, 1, -1, 1);
    UNREDUCE(left, bottom, margin, -1, 1, 2);
    UNREDUCE(right, bottom, margin, 1, 1, 3);

    if (reduced)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &reduced));
    if (rotated)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &rotated));
    if (cropMask)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &cropMask));

    NORMAL_RETURN;

ERR_CLEANUP:

    if (reduced)
	w_destroyPixr (pIPWState, &reduced);
    if (rotated)
	w_destroyPixr (pIPWState, &rotated);
    if (cropMask)
	w_destroyPixr (pIPWState, &cropMask);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Crop using the "original" cropping algorithm.
 *
 * LongDescription:::
 *   Calls w_crop4. What this routine was for, I do not know.
 *   It appears to be a wrapper to deal with oddness in crop4
 *   (crop4 doesn't add in the frame)
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to crop
 *  cropRect::       Region of interest; if NULL, use the whole image
 *
 * Output:::
 *  error::          Goodness-of-fit, 0=perfect
 *  angle::          Skew angle, in radians, of crop box
 *  cropPoints::     Crop box, possibly rotated, 4 corners
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cropByPoints (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *cropRect,
    Float64            *error,
    Float64            *angle,
    COORDS             *cropPoints)
{
    PROLOGUE;

    ERR_PASS (w_crop4 (pIPWState, prSrc, cropRect,
		       error, angle, cropPoints));

#if 0
    /* Add back in the frame. Rather absurd, but crop4 subtracts it off. */
    frame = pixrGetFrame (pIPWState, prSrc);

    for (i = 0; i < 4; i++)
    {
	cropPoints->x[i] += frame;
	cropPoints->y[i] += frame;
    }
#endif
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Compute distance between left bound of rotated box and original
 *
 * LongDescription:::
 *	When one rotates an image, one gets a somewhat larger image.
 *	This routine computes, for a given y-offset in the larger image,
 *	the distance between the left edge of the larger image and
 *	the original image. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  w::              Width of box
 *  h::              Height of box
 *  theta::          Angle, in radians, that box will be rotated
 *  yOffset::        Offset, in y, from 0,0
 *
 * Output::: 
 *  pXOffset::       Offset, in x, given y offset, of rotated box       
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotatedBoxOffset (
    IPWState           *pIPWState,
    Int32               w,
    Int32               h,
    Float64             theta,
    Int32               yOffset,
    Float64            *pXOffset)
{
    PROLOGUE;
    Float64 x[4], y[4], xtop, xbottom, yleft, ylefti, tmp, 
	    w2, h2;
    Int32 i;

    pIPWState = pIPWState;
    /* Not rotated--it touches */
    if (theta == 0.0)
    {
	*pXOffset = 0.0;
    }
    else
    {
	/* Rotate the points about the center */
	w2 = w/2;
	h2 = h/2;
	x[0] = -w2;
	y[0] = -h2; 
	x[1] = w2;
	y[1] = -h2;
	x[2] = -w2; 
	y[2] = h2;
	x[3] = w2;
	y[3] = h2;

	for (i = 0; i < 4; i++)
	{
	    tmp = x[i] * cos(-theta) - y[i] * sin(-theta) + w2;
	    y[i] = x[i] * sin(-theta) + y[i] * cos(-theta) + h2;
	    x[i] = tmp;
/*	fprintf(stderr, "xy[%d] %f,%f\n", i, x[i], y[i]); */
	}

	/* Compute xtop and xbottom */
	if (theta > 0.0)
	{
	    /* xtop is ur point, xbottom is ll point, ul sticks out left,
	       so x[0] should be negative, so subtract */
	    xtop = x[1] - x[0];
	    xbottom = x[2] - x[0];
	    yleft = y[0] - y[1];
	    ylefti = (h + 2 * y[1]) - yleft;
	}
	else
	{
	    /* xtop is ul point, xbottom is lr point, ll sticks out left */
	    xtop = x[0] - x[2];
	    xbottom = x[3] - x[2];
	    yleft = y[2] - y[0];
	    ylefti = (h + 2 * y[0]) - yleft;
	}

	if ((Float64) yOffset <= yleft)
	{
	    /* use upper triangle */   
	    *pXOffset = (xtop * (yleft - (Float64) yOffset)) / yleft;
	}
	else
	{
	    /* use lower triangle */
	    *pXOffset = (xbottom * (ylefti - (Float64)(h - yOffset))) / ylefti;
	}
/*fprintf(stderr, "xtop %f xbottom %f w,h (%d, %d), theta %f, y %d = %f\n",
	xtop, xbottom, w, h, theta, yOffset, *); */
    }

    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find the bounding box of a document. Doesn't deal with skew.
 *
 * LongDescription:::
 *	Same as cropReduced, only less. I took out all of the skew
 *	stuff, and it'll just generate a RECTANGL. Also, takes a margin
 *	as a parameter, so that we don't hard-code it.
 *      <BR>
 *      This algorithm looks for dark image features rather than
 *      potentially low-contrast edges like the "crop" functions.
 *      It is used to find the actual content of a document.
 *    
 * PrivateDescription:::
 *	We are passed a thresholded image. What we do is
 *	make a reduced version, get rid of stuff around the
 *	edges, maybe clean it up a little, and then march
 *	in from the edges and find the first stuff we run into.
 *	<BR>
 *	The edges are then moved out by the margin to give a nice border around
 *	stuff, and then transformed back to the input coordinate space.
 *	One nice thing to do might be to count the # of connected components,
 *	and not put the border around if there is only one. Another nice
 *	thing to do might be to have a set of default paper sizes, and
 *	if we find that our rectangle is near one of them, set the 
 *	rectangle to that paper size.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image in which to search for bounding box (binary)
 *  lookInside::     Region of interest; if NULL, look in entire image
 *  margin::         Margin to add to crop box (i.e. expand it)
 *  fillEdges::      If cTrue, get rid of stuff connected to edges of the image
 *
 * Output:::
 *  boundingBox::    Bounding box computed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_documentBoundingBox (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *lookInside,
    Int32               margin,
    Bool                fillEdges,
    RECTANGL           *boundingBox)
{
    PROLOGUE;
    PIXR	*reduced = NULL, *lookInsideMask = NULL;
    Int32	 reduction, i,
		 top, left, bottom, right, bpl, w, h, dpi;
    Int32	 bit;
    RECTANGL	 localRect;
    UInt8	*ip;
    ScaledInt1616    xDPI, yDPI;


    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));


    if (lookInside == NULL)
    {
	lookInside = &localRect;
	lookInside->x = 0;
	lookInside->y = 0;
	ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &(lookInside->w)));
	ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &(lookInside->h)));
    }


    /* Do the reduction, which is resolution-dependent. Do
     * all but the last reductions at threshold 1, which makes dark things get
     * darker, and then do the last one at threshold 4, which gets rid
     * of single-pixel things. 
     */
    if (dpi <= 100)
    {
	ERR_PASS (w_reduce4 (pIPWState, prSrc, 1, 4, &reduced));
	reduction = 4;
    }
    else if (dpi <= 200)
    {
	/* 8x reduction */
	ERR_PASS (w_reduce8 (pIPWState, prSrc, 1, 1, 4, &reduced));
	reduction = 8;
    }
    else
    {
	/* 16x reduction */
	ERR_PASS (w_reduce16 (pIPWState, prSrc, 1, 1, 1, 4, &reduced));
	reduction = 16;
    }

    /* Fill in the edges. NOTE that we don't use the crop box here, since
     * we really are concerned with the scanner edges. Maybe this
     * step should be an option, just in case one knows that the
     * image is already cropped. 
     */
    if (fillEdges)
    {
	PIXR *prTmp = NULL;
	/* Invert it first, so we don't make it all  black */
	ERR_PASS_CLEANUP (w_invert (pIPWState, reduced, &reduced));

	/* Fill in everything along the edges that's not black */
	prTmp = NULL;
	ERR_PASS_CLEANUP (w_edgeFillSweep4 (pIPWState, reduced, 10, &prTmp));

	/* Figure out what was different */
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, reduced, XOR, prTmp));

	/* Invert and subtract it out */
	ERR_PASS_CLEANUP (w_invert (pIPWState, reduced, &reduced));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, reduced));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    }

    /* Blow away stuff that's not in our "lookInside" rectangle */
    /* Make a mask for the crop rectangle. */
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, reduced, &lookInsideMask));

    /* Set "look inside" mask region. */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_SET,
				  lookInside->x / reduction, 
				  lookInside->y / reduction,
				  lookInside->w / reduction,
				  lookInside->h / reduction,
				  lookInsideMask));

    /* AND it to get rid of non-cropped stuff*/
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, lookInsideMask,INTERSECTION, 
				 reduced));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &lookInsideMask));

    /* Search for the first "on" pixel in each direction */
    ERR_PASS (w_pixrGetBpl (pIPWState, reduced, &bpl));
    ERR_PASS (w_pixrGetImage (pIPWState, reduced, &ip));
    ERR_PASS (w_pixrGetWidth (pIPWState, reduced, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, reduced, &h));

    /* Top: search down until we find a line with a pixel on it */
    for (i = 0; i < h; i++)
    {
	if (i_nextBitToTheRight (pIPWState, ip, w, 0, &bit) == cSuccessful)
	    break;
	ip += bpl;
    }

    ERR_GEN ((i == h), cErrHeuristic_foundNothing);

    if(i >0) top = i - 1;
    else top =0;

    /* Bottom: search up until we find a line with a pixel on it */
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, reduced, &ip));
    ip += bpl * (h - 1);

    for (i = h - 1; i >= 0; i--)
    {
	if (i_nextBitToTheRight (pIPWState, ip, w, 0, &bit) == cSuccessful)
	    break;
	ip -= bpl;
    }

    ERR_GEN_CLEANUP ((i < 0), cErrHeuristic_foundNothing);

    if(i < (h-1)) bottom = i + 1;
    else bottom = i;

    /* For left and right, look on each line and find the min and max.
     * Two things could be faster: 1, just looking for the first column
     * with a "1" in it, and 2, only looking in the crop region 
     */
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, reduced, &left));
    right = 0;
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, reduced, &ip));
    for (i = 0; i < h; i++)
    {
	if (i_nextBitToTheRight (pIPWState, ip, w, 0, &bit) == cSuccessful)
	{
	    if (bit < left)
		left = bit;
	}

	if (i_nextBitToTheLeft (pIPWState, ip, 0, w - 1, &bit) == cSuccessful)
	{
	    if (bit > right)
		right = bit;
	}

	ip += bpl;
    }

    if(left >0 ) left--; 
    if(right < (w-1)) right++;

    /* Scale them back up */
    left = (left * reduction) - margin;
    right = (right * reduction) + margin;
    top = (top * reduction) - margin;
    bottom = (bottom * reduction) + margin;

    /* Stuff the bounding box */
    boundingBox->x = left;
    boundingBox->y = top;
    boundingBox->w = right - left + 1;
    boundingBox->h = bottom - top + 1;

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &reduced));
    if (lookInsideMask)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &lookInsideMask));

    NORMAL_RETURN;

ERR_CLEANUP:

    if (reduced)
	w_destroyPixr (pIPWState, &reduced);
    if (lookInsideMask)
	w_destroyPixr (pIPWState, &lookInsideMask);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Another method of finding the bounding box for a photograph.
 *
 * LongDescription:::
 *	Assuming that an image of a photo (or whatever contone thing)
 *	has already been deskewed, we want to find the bounding box
 *	for the photo.
 *
 * PrivateDescription:::
 *	The method is as follows: find the "best-for-text" threshold
 *	for the image, threshold it and call documentBoundingBox on that. 
 *	One should crop in a bit to get rid of edge yutz from the scanner.
 *	<BR>---<BR>
 *	The above method worked on all of the "official" test images, but
 *	didn't really work on pictures with "sky" that had no detail.
 *	My original thought for a method for this function was to find
 *	edges using a morphological filter, threshold them, restrict them 
 *	to horizontal and vertical, and then pass that to documentBounding
 *	box. The problem was that if an image had undetectable edges,
 *	it would be nice to find non-edge stuff. The next attempt is
 *	to do both: find the edges, and OR that with the best-for-text-
 *	thresholded image. This may end up taking too much time.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to crop
 *  pLookInside::    Look inside this rectangle, or whole image if NULL
 *
 * Output:::
 *  pBoundingBox::   Bounding box of photo
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *  I don't think anyone is using this function, and I will probably
 *  take it out.
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_photoBoundingBox (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *pLookInside,
    RECTANGL           *pBoundingBox)
{
    PROLOGUE;   
    Int32    threshold, slop, margin, depth, dpi;
    PIXR    *prThresh = NULL, *prRedHi = NULL, *prRedLo = NULL, *prTmp = NULL, 
	    *prRedThreshed = NULL, *prRedH = NULL, *prRedV = NULL, 
	    *prExp = NULL;
    RECTANGL rect;
    STREL   *hStrel = NULL, *vStrel = NULL;
    Bool     inv;
    ScaledInt1616    xDPI, yDPI;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4 | cDepth8, cChannelsAny,
		  cFrame32);

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    if (dpi >= 300)
    {
	ERR_PASS (w_initStrelByName (pIPWState, strel20h));
	ERR_PASS (w_getStrel (pIPWState, strel20h, &hStrel));

	ERR_PASS (w_initStrelByName (pIPWState, strel20v));
	ERR_PASS (w_getStrel (pIPWState, strel20v, &vStrel));
    }
    else 
    {
	ERR_PASS (w_initStrelByName (pIPWState, strel10h));
	ERR_PASS (w_getStrel (pIPWState, strel10h, &hStrel));

	ERR_PASS (w_initStrelByName (pIPWState, strel10v));
	ERR_PASS (w_getStrel (pIPWState, strel10v, &vStrel));
    }

    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &slop));
    slop /= 100;
    margin = dpi / 10;
    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));

    if (pLookInside == NULL)
    {
	/* make our own */
	pLookInside = &rect;
	rect.x = slop;
	rect.y = slop;
	ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &(rect.w)));
	rect.w -= slop;
	ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &(rect.h)));
	rect.h -= slop;
    }

    ERR_PASS (w_findTextThreshold (pIPWState, prSrc, &threshold, &inv));

    /* Threshold that */
    ERR_PASS (w_threshold (pIPWState, prSrc, 0, threshold, THRESH_NORM,
			   &prThresh));

    /* Do some max and min reductions */
    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, prSrc, 1, &prTmp));
    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, prTmp, 1, &prRedHi));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, prSrc, 0, &prTmp));
    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, prTmp, 0, &prRedLo));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

    /* The difference will be a gradient */
    ERR_PASS_CLEANUP (w_subtractGray (pIPWState, prRedHi, prRedLo,
				      &prRedHi));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedLo));

    /* Threshold the gradient */
    ERR_PASS_CLEANUP (w_threshold (pIPWState, prRedHi, 
				   (depth == 8 ? 40 : 3),
				   (depth == 8 ? 255 : 15),
				   THRESH_NORM,
				   &prRedThreshed));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedHi));

    ERR_PASS_CLEANUP (w_open (pIPWState, prRedThreshed, hStrel, &prRedH));
    ERR_PASS_CLEANUP (w_open (pIPWState, prRedThreshed, vStrel, &prRedV));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedThreshed));

    /* Combine them */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prRedV, OR, prRedH));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedV));

    /* Expand and OR with the thresholded thing*/
    ERR_PASS_CLEANUP (w_expand (pIPWState, prRedH, 4, &prExp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedH));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prExp, OR, prThresh));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prExp));

    /* Call documentBoundingBox on it */
    ERR_PASS_CLEANUP (w_documentBoundingBox (pIPWState, prThresh, pLookInside,
					     margin, cFalse, pBoundingBox));


    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prThresh));
    

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prThresh) w_destroyPixr (pIPWState, &prThresh);
    if (prRedHi) w_destroyPixr (pIPWState, &prRedHi);
    if (prRedLo) w_destroyPixr (pIPWState, &prRedLo);
    if (prRedThreshed) w_destroyPixr (pIPWState, &prRedThreshed);
    if (prRedH) w_destroyPixr (pIPWState, &prRedH);
    if (prRedV) w_destroyPixr (pIPWState, &prRedV);
    if (prTmp) w_destroyPixr (pIPWState, &prTmp);
    ERR_RETURN;
}


/*
 $Log: crop2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.40  1998/01/14 00:11:09  ddavies
 crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.39  1997/06/26 20:00:40  ddavies
 crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.38  1997/04/15 21:43:52  ddavies
 crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v

 Substitute NoInit (uninitialized) version of pixr creation routines where
 possible.

 Revision 1.37  1996/11/20 01:31:15  xzhu
 xzhu on Tue Nov 19 17:22:25 PST 1996

 crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v

 When the detected image bounding box is equal to orignal image,  w_documentBoundingBox and w_cropReduced still move bounding box out image one pixel and cause bounding box left and top negtive and width and height larger than orignal image width and height. Now we added condition for moving bounding box out the image ragion one pixel. Fixed tht problem of bounding box out of orignal image.

 Revision 1.36  1996/01/04 20:21:45  ejaquith
 crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.35  1995/12/13  04:05:53  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.34  1995/11/17  01:06:09  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * The code in w_documentBoundingBox to get rid of stuff around the edges
 * of images had a bug
 *
 * Revision 1.33  1995/11/16  19:46:58  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * The nag script noticed that there were "wp_" routines in crop.pub,
 * or more specifically, not in a ".prv" file, since there wasn't one.
 * Created one, and #included it here.
 *
 * Revision 1.32  1995/11/13  20:10:20  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.31  1995/10/13  21:59:21  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.30  1995/10/12  17:43:51  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * added explicit casts (float->int, long->short, long->char, etc) and
 * made floating constants the right size to shut up the NT compiler.
 *
 * Revision 1.29  1995/10/03  01:59:08  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.28  1995/09/21  22:52:11  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.27  1995/09/19  19:19:18  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.26  1995/09/11  21:54:03  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.25  1995/09/07  01:30:13  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.24  1995/09/06  20:17:58  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.23  1995/09/05  19:00:45  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Removed calls to GET_STREL macro, instead calling getStrel() function.
 * This required using local variables rather than invoking GET_STREL
 * as an argument to another function, since getStrel can return an error
 * code which should be checked.
 *
 * Revision 1.22  1995/09/01  21:02:27  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.21  1995/08/31  21:11:41  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.20  1995/08/25  01:35:54  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Fixed warnings
 *
 * Revision 1.19  1995/08/25  01:08:20  jcraig
 * jcraig on Thu Aug 24 18:08:10 PDT 1995
 *
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * get rid of pc compiler warnings
 *
 * Revision 1.18  1995/08/17  21:54:32  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.17  1995/02/14  02:57:42  macleod
 * macleod on Mon Feb 13 18:55:40 PST 1995
 *
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Added photoBoundingBox, which finds the bounding box of a deskewed photo.
 * The method is to make two masks: one of a thresholded version of the image,
 * which finds "stuff" in the photo, and another of an edge-filtered and
 * then thresholded version of the image, to find the edges of photos.
 * The reason for two versions is to be able to find the photo edges, but
 * to not fail too badly if the photo edge is nearly the intensity of the
 * background.
 *
 * Revision 1.16  1995/01/24  21:49:36  lperry
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Incorporate changes to use new strel manager interface.
 *
 * Revision 1.15  1995/01/11  22:39:35  dferrell
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Added include of shrcc.pub.
 *
 * Revision 1.14  1994/10/18  22:43:20  macleod
 * macleod on Tue Oct 18 15:43:13 PDT 1994
 *
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Changed call to findBackgroundThreshold
 *
 * Revision 1.13  1994/10/07  19:06:18  lperry
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.12  1994/09/29  00:46:13  lperry
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Eliminate use of utilpros. Use utils.pub instead.
 *
 * Revision 1.11  1994/09/21  21:39:26  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * crop2_4 computes a skew angle
 *
 * Revision 1.10  1994/07/27  21:52:06  ejaquith
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Removed #include "iadef.prv"
 *
 * Revision 1.9  1994/07/21  18:48:08  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Added an error check to documentBoundingBox to enforce bit depth
 *
 * Revision 1.8  1994/06/16  22:36:33  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Added a "threshold" parameter to crop. This allows one to pass in a
 * threshold rather than have it computed inside crop, thus saving time.
 * If you pass -1 as the threshold, a threshold will be computed.
 *
 * Revision 1.7  1994/06/08  04:07:08  ddavies
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.6  1994/05/26  19:20:43  ddavies
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Change the abortI's whose return values are not used to abortV's.
 *
 * Revision 1.5  1994/03/24  23:05:15  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * There was a scheme to restrict blowing away stuff touching the edge of the
 * platen to a few percent of the edge, so that pictures touching the edge
 * wouldn't get completely blown away and cropped over, but it didn't work.
 * This should fix it.
 *
 * Revision 1.4  1994/03/23  02:58:43  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Fixed a couple of places where it returned NULL for an error code;
 * restricted the amount of crud that can be eaten off of the edges
 * (for platen-edge-crud removal)
 *
 * Revision 1.3  1994/03/17  03:00:37  macleod
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Changed call to findBackgroundThreshold
 *
 * Revision 1.2  1994/03/10  02:02:23  macleod
 * macleod on Wed Mar 9 18:01:53 PST 1994
 *
 * crop2.c -> /product/ipcore/ipwrap/src/RCS/crop2.c,v
 *
 * Added documentBoundingBox; fixed a bug where the frame was getting
 * subtracted twice
 *
 * Revision 1.1  1994/03/02  02:48:37  macleod
 * Initial revision
 *
 */

