/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1997-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * crop3.c
 *
 *	Another "Auto-crop"
 *
 * Functions defined:
 *
 *	w_cropPhoto () - Find a crop box for an image, rotate and crop
 *	w_crop5 () - Find a crop box for an image, return the box
 *	w_cropAndRotate () - Given the rectangle, precrop, rotate, and
 *		final crop the matching image region
 *	wp_countBlackCorners () - count the number of black corners in
 *		a binary image
 *	w_subtractHisto () - subtract one histogram from another
 *
 * Notes:
 * 
 * The first crop function, which looked for low- and high-contrast
 * edge points and fit edges by linear regression, worked OK, but had 
 * its limitations--it didn't deal well when there was a page with a lot
 * of elements and a very light background: the page was missed, and the
 * elements were missed, so it defaulted to the whole platen.
 *
 * Since we now have a decent background threshold finder which works
 * reasonably well even on an un-cropped image, I'm going to use that
 * for this next method. Basically, there are two cases I'm going to handle:
 * The case where there is text, and the case where there is no text.
 * In the first case, since we can threshold the image, we can then:
 *	- detect the skew angle
 *	- de-skew the binary image
 *	- do textured reductions to get a reduced image, probably
 *	  (1, ..., 4) to get things to clump together and then to
 *	  get rid of single-pixel things (the "..." are as many ones
 *	  as needed to get it to a ~100 pixel wide image.
 *	- Get rid of all blobs that touch the edge, by dropping seeds
 *	  around the edge. This is needed to eliminate funny stuff on 
 *	  the edges of platens.
 *	- Close stuff up a bit
 *	- march down using nextXXXToTheRight, to find the first "stuff"
 *	  along each of the four edges.
 *	  Note that since we deskewed, we need to be careful not to include
 *	  the deskewed part, which isn't really part of the image. 
 *	  Perhaps we should do the rotation after the morphology and
 *	  edge crap elimination.
 *	- Rotate and scale the resulting box back to get crop points
 *	  in the coordinate system of the original.
 *
 * In the second case, we might want to use the technique of the original
 * crop, only using high-contrast edges, and without doing any region
 * averaging. It would be nice if we could just threshold and find the
 * corners, but there are two problems with this. First, we can't really
 * threshold things without text very well, and second, we can't reliably
 * eliminate stuff on the edges of the platen. For example, I was given
 * a sample which had a photo in the upper left, and extending out of the
 * photo to the right was a big black band which was where the auto-document
 * -feeder belt curved away from the glass. The original scheme worked on
 * this and similar images, just not very well on documents.
 ************************************************************************/

#include <math.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "thrshold.h"
#include "morphg.h"
#include "shrscgr.pub"
#include "defines.pub"
#include "alplib.h"
#include "shrcc.pub"
#include "logicops.h"
#include "histo.h"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "crop.pub"
#include "crop.prv"
#include "thrshold.pub"
#include "reduce.pub"
#include "prxform.pub"
#include "morph.pub"
#include "expand.pub"
#include "strel.pub"
#include "skew.pub"
#include "fill.pub"
#include "intops.pub"
#include "coords.pub"
#include "rotatop.pub"
#include "logicops.pub"
#include "fthresh.pub"
#include "frames.pub"    /* for cFrameBits */
#include "shrip.pub"
#include "hutils.pub"
#include "runfilt.pub"
#include "runproj.pub"
#include "imginfo.pub"
#include "ccthresh.pub"
#include "color.pub"
#include "gradient.pub"
#include "wshed.pub"
#include "curves.pub"
#include "subsmpl.pub"
#include "cntr.pub"
#include "shrpixr.pub"	 /* for StartTime () & StopTime () */
#include "imageref.pub"
#include "warp.pub"
#define _IPWSTATE_DEFINED_
#include "librastw.pub"
#include "cut.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif


IP_RCSINFO(RCSInfo, "$RCSfile: crop3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>


#define DEBUG_CROP 0
#define DEBUG_PRINT 0

#if DEBUG_CROP || DEBUG_PRINT
#include "imagerep.h"    /* for FH */
#include "io.pub"
#include "rotatop.pub"
#include "grayproj.pub"
char project[] = "phtocr";
#endif


/* local protots */

static ErrCode CDECL
wp_countBlackCorners (
                      IPWState *pIPWState, 
		      PIXR *pr, 
		      Int32 *count);

static ErrCode CDECL
w_subtractHisto (
                 IPWState *pIPWState,
	         Int32 *h1,
	         Int32 *h2,
	         Int32  length);


#define cConIndex	1
#define cWhitePoint	250
#define FRAME_PCT	5	/* Percentage of size to include as frame */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Detect the rectangular photo in the image, deskew it, and crop it
 *
 * LongDescription:::
 *   w_cropPhoto() is the front-end to the photo detection functions
 *   w_crop5() and w_crop4(), and the crop and rotate function
 *   w_cropAndRotate().
 *   <BR>
 *   In the case of failure during w_crop5(), w_crop4() is called.
 *
 * Input:::
 *  pIPWState::	    State structure
 *  inPr::	    Input image
 *  platen::	    Platen mode (0 = white platen, 1 = black, 2 = unknown)
 *  deskew::	    Image is to be deskewed (boolean)
 *  crop::	    Image is to be cropped (boolean)
 *  
 * Output:::
 *  outPr::	    Resulting image (deskewed and/or cropped)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  dferrell, aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_cropPhoto (
              IPWState *pIPWState,
	      PIXR *inPr,
	      Int32 platen,
	      Bool deskew,
	      Bool crop,
	      PIXR **outPr)
{
    COORDS *pt;		/* Points to form the bounding rectangle */
    Float64 angle, error;
    Bool detectSuccess = cTrue;
    Int32 depth;
    ColorMap *colormap = NULL;
    PROLOGUE;

    /* Initialize the coordinate list */
    ERR_PASS(w_createCoords(pIPWState, 4, &pt));

    /* Run w_crop5 if support for current mode */
    ERR_PASS(w_pixrGetDepth(pIPWState, inPr, &depth));
    ERR_PASS(w_pixrGetColorMap(pIPWState, inPr, &colormap));

    if (depth == 8 && colormap == NULL) {

	/* Get the recommendation for photo region */
	ERR_CHECK_NO_MSG(w_crop5(pIPWState,
				 inPr,
				 platen,
				 &angle,
				 pt)) {
	    detectSuccess = cFalse;
	}
    } else {
	detectSuccess = cFalse;
    }

    /* If w_crop5() can't find anything, use the "old" algorithm */
    if (detectSuccess != cTrue) {
	ERR_PASS_CLEANUP(w_crop4(pIPWState,
				 inPr,
				 NULL,
				 &error,
				 &angle,
				 pt));
    }

    /* Crop to that region */
    ERR_PASS_CLEANUP(w_cropAndRotate(pIPWState,
				     inPr,
				     pt->x[0], pt->y[0],
				     pt->x[1], pt->y[1],
				     pt->x[2], pt->y[2],
				     pt->x[3], pt->y[3],
				     angle,
				     deskew,
				     crop,
				     outPr));

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Detect the rectangular photo in the image, return it's location and angle
 *
 * LongDescription:::
 *   w_crop5() determines where the photo is on the platen, finds the
 *   boundary between the photo and the platen, determines the best
 *   orthogonal edges, computes the rotation angle, computes the coordinates
 *   of the bounding rectangle of the photo, and returns those results.
 *
 * PrivateDescription:::
 *   The task of automatically detecting a scanned photo placed on a platen
 *   results in lots of ideas, and always problem cases. This algorithm is
 *   the result of Doug Ferrell's last efforts (before bailing... I wonder if
 *   it had anything to do with it) at photo crop detect.
 *   <BR>
 *   The algorithm goes something like this, but there's probably more to it:
 *   1. Histogram the edge (frame) and look for a peak at what should be
 *	the platen (I think there's a lot more than this, but I think this
 *	is the basic idea).
 *   2. Locate a spot on the platen, and on the photo.
 *   3. Use watershedding to bring the two spots together, hopefully at
 *	the photo boundary.
 *   4. Convert the boundary to a polygon
 *   5. Determine the polygon segments that have a peak in the orgthogonal
 *	directions.
 *   6. Compute the rotation angle between -45 and +45 degrees.
 *   7. Rotate the polygon according to the rotation angle.
 *   8. Compute the bounding rectangle of the rotated polygon.
 *   9. Rotate the rectangle back.
 *   10. Return the angle, and rectangle (in original image coordinates).
 *
 * Input:::
 *  pIPWState::	    State structure
 *  inPr::	    Input image
 *  platen::	    Platen mode (0 = white platen, 1 = black, 2 = unknown)
 *  
 * Output:::
 *  pRotAngle::	    Rotation angle needed to correct the skew of the photo,
 *		    in radians in the clockwise direction. A positive value
 *		    means rotate the image clockwise, and that the photo lies
 *		    at a negative of this angle.
 *  pCropRect::	    Points of crop rectangle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  dferrell, aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_crop5 (
	 IPWState *pIPWState,
	 PIXR *inPr,
	 Int32 platen,
	 Float64 *pRotAngle,
	 COORDS *pCropRect)
{
    PROLOGUE;
    Int32            w, h, d, c, sw, sh, 
                     px, py, maxM, hi, hi2,
		     polySize, 
		     thinPolySize,
		     count,
		     pivotX, pivotY;
    Int32           *hf = NULL,
		    *hc = NULL;
#if DEBUG_CROP
    Histo            histo;
    char             fileN[128];
#endif
    RECTANGL         roi, 
                    *thisRect,
                    *rList = NULL,
		    *rp;
    ScaledInt1616    xdpi, ydpi;
    ipPoint         *poly = NULL, *thinPoly = NULL,
		    *rotPoly = NULL;
    STREL           *s = NULL;
    PIXR            *tmpPr = NULL,
                    *phMaskPr = NULL,
		    *plMaskPr = NULL,
		    *markerPr = NULL,
		    *gradPr = NULL,
		    *redPr = NULL;
    Float64	    rotationAngle;	/* In radians */
    Int32	    fact;		/* X & Y reduction factor */
    Int32 minX, minY, maxX, maxY;
    Int32 point;
    ColorType	    colortype;
    ColorMap	    *colorMap = NULL;
#ifdef DEBUG_PRINT
    Bool allWhite;
#endif

    /* Default to no skew angle */
    *pRotAngle = 0.0;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth4 | cDepth8, 
		  cChannels1 | cChannels3, cFrame32);
    ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &d));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, inPr, &c));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xdpi, &ydpi));
    ERR_PASS (w_pixrGetColorMap (pIPWState, inPr, &colorMap));
    ERR_CHECK_NO_MSG(colorMap != NULL) {
	__error__ = cErrPixr_colormap;
	goto ERR_CLEANUP;
    }
    if (c == 3 ) {
	colortype = cColorRGB;
    } else {
	colortype = cColorGray;
    }

    /* Initialize the coordinates */
    ERR_GEN (pCropRect->n != 4, cErrInvalidParm);

    px = 0;
    py = 0;
    polySize = thinPolySize;

    StartTime();

    /* make the strel for processing the markers */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 5, 5, 2, 2, &s));

    /* find the gradient */
    /* We do this at full resolution because any reduction prior to
     * finding the gradient causes loss of sharpness and/or dropouts
     * in the gradient, either of which can ruin watershed results.
     */
    ERR_PASS_CLEANUP (w_findMagnitudeOfGradient (pIPWState, inPr,
						 cGradOpMaxLInfFromCenter,
						 cGradNeighborhoodDownRight,
						 &tmpPr));

#if DEBUG_CROP
        sprintf (fileN, "%s-gradient.tif", project);
        ERR_PASS_CLEANUP (w_write (pIPWState, fileN, tmpPr, TIFF_DEFAULT));
#endif

    if (c == 1) /* single plane */
    {
        if (ScaledInt1616_TO_Int32 (xdpi) > 160)
	{
	    /* reduce the gradient, preserving high gradient information */
	    ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, tmpPr, 2,
					      cTrue, cTrue, cTrue, cTrue, 
					      &gradPr));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));

	    /* make a reduction of the original input */
            ERR_PASS_CLEANUP (w_reduceS (pIPWState, inPr, 4, &redPr));
	    fact = 4;
	}
	else
	{
	    /* reduce the gradient, preserving high gradient information */
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, tmpPr, cTrue, 
					      &gradPr));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));

	    /* make a subsampled reduction of the original input */
            ERR_PASS_CLEANUP (w_reduceS (pIPWState, inPr, 2, &redPr));
	    fact = 2;
	}

    }
    else /* RGB */
    {

        if (ScaledInt1616_TO_Int32 (xdpi) > 160)
	{
	    /* reduce the gradient, preserving high gradient information */
	    ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, tmpPr, 2,
					      cTrue, cTrue, cTrue, cTrue, 
					      &gradPr));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));

            /* convert RGB to intensity */
	    ERR_PASS_CLEANUP (w_RGBToYFast (pIPWState, inPr, &tmpPr));

	    /* make a subsampled reduction of the original input */
            ERR_PASS_CLEANUP (w_reduceS (pIPWState, tmpPr, 4, &redPr));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));
	    fact = 4;
        }
	else
	{
	    /* reduce the gradient, preserving high gradient information */
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, tmpPr, cTrue,  
					      &gradPr));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));

            /* convert RGB to intensity */
	    ERR_PASS_CLEANUP (w_RGBToYFast (pIPWState, inPr, &tmpPr));

	    /* make a subsampled reduction of the original input */
            ERR_PASS_CLEANUP (w_reduceS (pIPWState, tmpPr, 2, &redPr));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));
	    fact = 2;
	}
    }

#if DEBUG_CROP
        sprintf (fileN, "%s-reducedGradient.tif", project);
        ERR_PASS_CLEANUP (w_write (pIPWState, fileN, gradPr, TIFF_DEFAULT));
#endif

    ERR_PASS (w_pixrGetWidth (pIPWState, redPr, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, redPr, &h));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, redPr, &xdpi, &ydpi));

    /* sample the whole image */
    ERR_PASS_CLEANUP (w_histo (pIPWState, redPr, NULL, &hf));

    /* compute width/height of center sample region */
    sw = ((100 - 2 * FRAME_PCT) * w) / 100;
    sh = ((100 - 2 * FRAME_PCT) * h) / 100;

    roi.x = (w - sw) / 2;
    roi.y = (h - sh) / 2;
    roi.w = sw;
    roi.h = sh;

    /* get the histogram of the center part of the image */
    ERR_PASS_CLEANUP (w_histo (pIPWState, redPr, &roi, &hc));

    /* subtract to get the histo of the outer part--this is faster than
     * running two histoMask operations. */
    ERR_PASS_CLEANUP (w_subtractHisto (pIPWState, hc, hf, 
                                      (d == 4 ? 16 : 256)));

    switch (platen)
    {
        case 0:  /* white platen */
	    ERR_PASS_CLEANUP (w_findGlobalMaximum (pIPWState, hf, 
	                                           180, 255, &hi));
	    break;

	case 1:  /* black platen */
	    ERR_PASS_CLEANUP (w_findGlobalMaximum (pIPWState, hf, 0, 75, &hi));
	    break;

	default:  /* don't know, so make an intelligent guess */
	    /* this is a simple scheme and will be fooled by already 
	     * cropped photo scans or scans with very little platen present.
	     */

	    /* find the local peak at the bright end */
	    ERR_PASS_CLEANUP (w_findGlobalMaximum (pIPWState, hf, 
	                                           180, 255, &hi));
	    /* find the local peak at the dark end */
	    ERR_PASS_CLEANUP (w_findGlobalMaximum (pIPWState, hf, 0, 75, &hi2));
            if (hi2 > hi)
	    {
	        hi = hi2;
	        platen = 1;
	    }
	    else
	        platen = 0;
	    break;
    }

#if DEBUG_PRINT
    fprintf (stdout, "Found platen mode in %f seconds.\n", StopTime ());
    StartTime();
#endif

#if DEBUG_CROP
    fprintf (stdout, "platen mode : %ld\n", hi);

    HISTO_SET_MINVAL (&histo, 0);
    HISTO_SET_MAXVAL (&histo, (d == 4 ? 15 : 255));
    HISTO_SET_SIZE (&histo, HISTO_GET_MAXVAL (&histo) + 1);
    HISTO_SET_ARRAY (&histo, hf);
    HISTO_SET_DATA (&histo, hf);

    sprintf (fileN, "%s-hf.histo", project);
    ERR_PASS_CLEANUP (w_writeHisto (pIPWState, &histo, fileN));

    HISTO_SET_ARRAY (&histo, hc);
    HISTO_SET_DATA (&histo, hc);

    sprintf (fileN, "%s-hc.histo", project);
    ERR_PASS_CLEANUP (w_writeHisto (pIPWState, &histo, fileN));

#endif

    /* make a photo mask pixr */
    ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, h, 1, xdpi, ydpi,
                                    cFrameBits, &phMaskPr));

    /* paint the center rect black */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_SET,
                                  roi.x, roi.y, roi.w, roi.h, phMaskPr));

    /* create a frame mask */
    ERR_PASS_CLEANUP (w_invert (pIPWState, phMaskPr, &plMaskPr));

    /* make a marker pixr */
    ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, h, 1, xdpi, ydpi,
                                    cFrameBits, &markerPr));

    if (platen == 0)
    {
        /* threshold for photo marker */
	ERR_PASS_CLEANUP (w_threshold (pIPWState, redPr, 0, 165, 
	                               THRESH_NORM, &markerPr));

	/* apply the "center rect" mask */
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, phMaskPr, AND, markerPr));

#if DEBUG_PRINT
	fprintf (stdout, "Found photo marker in %f seconds.\n", StopTime ());
	StartTime();
#endif

#if DEBUG_CROP
        sprintf (fileN, "%s-photoMarker.tif", project);
        ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
	                           markerPr, TIFF_PACKBITS));
#endif

        /* shrink the markers so they don't overlap relevant gradients */
        ERR_PASS_CLEANUP (w_erode (pIPWState, markerPr, s, &phMaskPr));

#if DEBUG_CROP
        sprintf (fileN, "%s-erodedPhotoMarker.tif", project);
        ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
	                           phMaskPr, TIFF_PACKBITS));
#endif

	/* now find and blit the largest photo marker */
        ERR_PASS_CLEANUP (w_rabb4Connected (pIPWState, phMaskPr, 1, 1, &rList));

	/* search the rList for the most massive photo marker */
	maxM = 0;
	rp = rList;
	while (rp != NULL)
	{
            if (rp->mass > maxM)
	    {
	        maxM = rp->mass;
		px = rp->px;
		py = rp->py;
            }
            thisRect = rp;
	    rp = rp->next;
	    IPW_FREE (thisRect, cAllocatorData);
	}
	rList = NULL;

        /* clear the marker pixr */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
	                              0, 0, w, h, markerPr));

        /* seed and fill the photo marker component */
	ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, markerPr, px, py, 1));
	ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, markerPr, phMaskPr, 6,
	                                    &markerPr));

        /* put this result in the photo mask pixr for later use */
	ERR_PASS_CLEANUP (w_copy (pIPWState, markerPr, &phMaskPr));

#if DEBUG_PRINT
	fprintf (stdout, "Filled photo region from marker in %f seconds.\n",
		 StopTime ());
	StartTime();
#endif

#if DEBUG_CROP
        sprintf (fileN, "%s-seededPhoto.tif", project);
        ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
	                           phMaskPr, TIFF_PACKBITS));
#endif

	/* threshold for platen marker */
	ERR_PASS_CLEANUP (w_threshold (pIPWState, redPr, 
	                               MIN (255, hi + 15) - 30, 
				       MIN (255, hi + 15), 
	                               THRESH_NORM, &tmpPr));

	/* apply the frame mask */
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, plMaskPr, AND, tmpPr));

#if DEBUG_CROP
        sprintf (fileN, "%s-frameMask.tif", project);
        ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
	                           tmpPr, TIFF_PACKBITS));
#endif

	/* shrink the markers so they don't overlap relevant gradients */
        ERR_PASS_CLEANUP (w_erode (pIPWState, tmpPr, s, &plMaskPr));

#if DEBUG_CROP
        sprintf (fileN, "%s-erodedFrameMarker.tif", project);
        ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
	                           plMaskPr, TIFF_PACKBITS));
#endif

	/* now find and blit the largest platen marker */
        ERR_PASS_CLEANUP (w_rabb4Connected (pIPWState, plMaskPr, 1, 1, &rList));

	/* search the rList for the most massive photo marker */
	maxM = 0;
	rp = rList;
	while (rp != NULL)
	{
            if (rp->mass > maxM)
	    {
	        maxM = rp->mass;
		px = rp->px;
		py = rp->py;
            }
            thisRect = rp;
	    rp = rp->next;
	    IPW_FREE (thisRect, cAllocatorData);
	}
	rList = NULL;

        /* clear the tmp pixr */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
	                              0, 0, w, h, tmpPr));

        /* seed and fill the platen marker component */
	ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, tmpPr, px, py, 1));
	ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, tmpPr, plMaskPr, 6,
	                                    &tmpPr));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, tmpPr, OR, markerPr));

    }
    else  /* dark platen */
    {
        /* threshold for photo marker */
	ERR_PASS_CLEANUP (w_threshold (pIPWState, redPr, 75, 190, 
	                               THRESH_NORM, &tmpPr));

	/* apply the "center rect" mask */
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, phMaskPr, AND, markerPr));

        /* shrink the markers so they don't overlap relevant gradients */
        ERR_PASS_CLEANUP (w_erode (pIPWState, markerPr, s, &tmpPr));

	/* now find and blit the largest photo marker */
        ERR_PASS_CLEANUP (w_rabb4Connected (pIPWState, tmpPr, 1, 1, &rList));

	/* search the rList for the most massive photo marker */
	maxM = 0;
	while (rList != NULL)
	{
            if (rList->mass > maxM)
	    {
	        maxM = rList->mass;
		px = rList->px;
		py = rList->py;
            }
            thisRect = rList;
	    rList = rList->next;
	    IPW_FREE (thisRect, cAllocatorData);
	}

        /* clear the marker pixr */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
	                              0, 0, w, h, markerPr));

        /* seed and fill the photo marker component */
	ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, markerPr, px, py, 1));
	ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, markerPr, tmpPr, 6,
	                                    &markerPr));


	/* threshold for platen marker */
	ERR_PASS_CLEANUP (w_threshold (pIPWState, redPr, 
	                               MAX (0, hi - 10), 
				       MAX (0, hi - 10) + 20, 
	                               THRESH_NORM, &tmpPr));

	/* apply the frame mask */
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, plMaskPr, AND, tmpPr));

        /* shrink the markers so they don't overlap relevant gradients */
        ERR_PASS_CLEANUP (w_erode (pIPWState, tmpPr, s, &plMaskPr));

	/* now find and blit the largest platen marker */
        ERR_PASS_CLEANUP (w_rabb4Connected (pIPWState, plMaskPr, 1, 1, &rList));

	/* search the rList for the most massive platen marker */
	maxM = 0;
	while (rList != NULL)
	{
            if (rList->mass > maxM)
	    {
	        maxM = rList->mass;
		px = rList->px;
		py = rList->py;
            }
            thisRect = rList;
	    rList = rList->next;
	    IPW_FREE (thisRect, cAllocatorData);
	}

        /* clear the tmp pixr */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
	                              0, 0, w, h, tmpPr));

        /* seed and fill the platen marker component */
	ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, tmpPr, px, py, 1));
	ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, tmpPr, plMaskPr, 6,
	                                    &tmpPr));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, tmpPr, OR, markerPr));
    }

#if DEBUG_CROP
    sprintf (fileN, "%s-platenMarker.tif", project);
    ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
			       markerPr, TIFF_PACKBITS));
#endif

#if DEBUG_PRINT
    fprintf (stdout, "Found platen marker in %f seconds.\n", StopTime ());
    StartTime();
#endif

    /* clear the dest image for watersheds */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
                                  0, 0, w, h, plMaskPr));

    /* run watershed boundary finder */
    ERR_PASS_CLEANUP (w_watershedSegment (pIPWState, gradPr, markerPr, 0,
                                          &plMaskPr));

    /* invert to get watershed regions */
    ERR_PASS_CLEANUP (w_invert (pIPWState, plMaskPr, &plMaskPr));

    /* clear the dest image for seed fill */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
                                  0, 0, w, h, tmpPr));

    /* seed fill the region with the photo marker in it */
    ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, phMaskPr, plMaskPr, 
                                        2, &tmpPr));

#if DEBUG_PRINT
    fprintf (stdout, "Found photo/platen edge in %f seconds.\n", StopTime ());
    StartTime();
#endif

#if DEBUG_CROP
        sprintf (fileN, "%s-photo.tif", project);
        ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
	                           tmpPr, TIFF_PACKBITS));
#endif

    /* Check for at least three corners being black at this point.
     * If they are, then say the photo is already cropped and deskewed.
     */
    ERR_PASS_CLEANUP (wp_countBlackCorners (pIPWState, tmpPr, &count));
    if (count >= 3)
    {
	/* Output the whole image */
	pCropRect->x[0] = pCropRect->y[0] = 0;
	pCropRect->x[1] = pCropRect->y[3] = 0;

	ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &h));

	pCropRect->y[1] = pCropRect->y[2] = h - 1;
	pCropRect->x[2] = pCropRect->x[3] = w - 1;

    } else {

	/* Check to see if it's all white */
	ERR_PASS_CLEANUP (w_isAllWhite(pIPWState, tmpPr, &allWhite));

	/* If it is, the watershedding failed. Gracefully exit. */
	ERR_CHECK_NO_MSG(allWhite) {
	    __error__ = cErrHeuristic_notEnoughPoints;
	    goto ERR_CLEANUP;
	}

	/* construct a polygon for the region's outline */
	ERR_PASS_CLEANUP (w_contourRegionBinary (pIPWState, tmpPr, NULL, 
	                                         1, cTrue,
						 &polySize, &poly));
    
	/* thin the polygon geometrically */
	ERR_PASS_CLEANUP (w_thinPolygon (pIPWState, poly, polySize, 1, 
					 &thinPoly, &thinPolySize));
    
#if DEBUG_CROP
	/* clear the dest image before drawing the polygon */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
				      0, 0, w, h, plMaskPr));
    
	ERR_PASS_CLEANUP (w_drawPolygon (pIPWState, thinPoly, thinPolySize,
					 cTrue, cTrue, plMaskPr));
	    sprintf (fileN, "%s-photoOutline.tif", project);
	    ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
				       plMaskPr, TIFF_PACKBITS));
#endif
#if DEBUG_PRINT
    fprintf (stdout, "Converted edge to a polygon in %f seconds.\n",
	     StopTime ());
	StartTime();
#endif
    
#if DEBUG_CROP
	{
	    PIXR *tmpPr = NULL;

	    ERR_PASS_CLEANUP (w_rotatePolygon (pIPWState, 0, 0, thinPoly,
		thinPolySize, 0.0, &rotPoly));

	    for (point = 0; point < thinPolySize; point++) {
		rotPoly[point].x *= fact;
		rotPoly[point].y *= fact;
	    }
	    ERR_PASS_CLEANUP (w_copy (pIPWState, inPr, &tmpPr));

	    ERR_PASS_CLEANUP (w_drawPolygon (pIPWState, rotPoly, thinPolySize,
		cTrue, cTrue, tmpPr));
	    sprintf (fileN, "%s-origWithCropPoly.tif", project);
	    ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
		tmpPr, TIFF_PACKBITS));
	}
#endif

	/* analyze the polygon */

	/* Find the angle of the best rectangle to represent the polygon.
	 * Compute the rotation angle between -45 and +45 degrees. */
	ERR_PASS_CLEANUP (w_findRectFromPolygon (pIPWState, thinPoly, thinPolySize,
	    0, &rotationAngle));

    
#if DEBUG_PRINT
	/* clear the dest image before drawing the polygon */
	fprintf(stdout, "Rotation angle to correct image is %.2f degrees\n",
	    rotationAngle * 180.0 / 3.1415926);
#endif
	/* Rotate the polygon according to the rotation angle. */
	ERR_PASS (w_pixrGetWidth (pIPWState, tmpPr, &pivotX));
	ERR_PASS (w_pixrGetHeight (pIPWState, tmpPr, &pivotY));
	pivotX /= 2, pivotY /= 2;
	ERR_PASS_CLEANUP (w_rotatePolygon (pIPWState, pivotX, pivotY, thinPoly,
	    thinPolySize, rotationAngle, &rotPoly));

#if DEBUG_CROP
	/* clear the dest image before drawing the polygon */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,

	    0, 0, w, h, plMaskPr));
    
	ERR_PASS_CLEANUP (w_drawPolygon (pIPWState, rotPoly, thinPolySize,
	    cTrue, cTrue, plMaskPr));
	sprintf (fileN, "%s-rotatedPhotoOutline.tif", project);
	ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
	    plMaskPr, TIFF_PACKBITS));
#endif


	/* Compute the bounding rectangle of the rotated polygon. */
	minX = maxX = rotPoly[0].x;
	minY = maxY = rotPoly[0].y;
	for (point = 0; point < thinPolySize; point++) {
	    if (rotPoly[point].x < minX) minX = rotPoly[point].x;
	    if (rotPoly[point].x > maxX) maxX = rotPoly[point].x;
	    if (rotPoly[point].y < minY) minY = rotPoly[point].y;
	    if (rotPoly[point].y > maxY) maxY = rotPoly[point].y;
	}

	/* Replace the polygon with the rectangle */
	IPW_FREE(thinPoly, cAllocatorData);
	thinPoly = NULL;
        thinPoly = (ipPoint *)IPW_CALLOC(sizeof (ipPoint), 
					 thinPolySize = 4, cAllocatorData);
	thinPoly[0].x = thinPoly[1].x = minX - 1;
	thinPoly[2].x = thinPoly[3].x = maxX + 1;
	thinPoly[0].y = thinPoly[3].y = minY - 1;
	thinPoly[1].y = thinPoly[2].y = maxY + 1;

	/* Rotate the rectangle back. */
	ERR_PASS_CLEANUP (w_rotatePolygon (pIPWState, pivotX, pivotY, thinPoly,
	    thinPolySize, -rotationAngle, &rotPoly));

#if DEBUG_CROP
	/* clear the dest image before drawing the polygon */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
	    0, 0, w, h, plMaskPr));
    
	ERR_PASS_CLEANUP (w_drawPolygon (pIPWState, rotPoly, thinPolySize,
	    cTrue, cTrue, plMaskPr));
	sprintf (fileN, "%s-finalCropRect.tif", project);
	ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
	    plMaskPr, TIFF_PACKBITS));
#endif


	/* Return the angle, and rectangle (in original image coordinates).
	 * The angle returned is the angle to correct the skew, and is
	 * positive if clockwise rotation is necessary to fix it.
	 */
	*pRotAngle = rotationAngle;
	for (point = 0; point < thinPolySize; point++) {
	    rotPoly[point].x *= fact;
	    rotPoly[point].y *= fact;
	    pCropRect->x[point] = rotPoly[point].x;
	    pCropRect->y[point] = rotPoly[point].y;
	}

#if DEBUG_CROP
	{
	    PIXR *tmpPr = NULL;

	    ERR_PASS_CLEANUP (w_copy (pIPWState, inPr, &tmpPr));

	    ERR_PASS_CLEANUP (w_drawPolygon (pIPWState, rotPoly, thinPolySize,
		cTrue, cTrue, tmpPr));
	    sprintf (fileN, "%s-origWithCropRect.tif", project);
	    ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
		tmpPr, TIFF_PACKBITS));
	}
#endif

#if DEBUG_PRINT
	fprintf (stdout, "Analyzed the polygon in %f seconds.\n", StopTime ());
	StartTime();
#endif
    }  /* three black corners in photo marker */

    /* Clean up */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &redPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &phMaskPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &plMaskPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &gradPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &markerPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &s));
    IPW_FREE (hf, cAllocatorData);
    hf = NULL;
    IPW_FREE (hc, cAllocatorData);
    hc = NULL;
    if (poly) {
	IPW_FREE(poly, cAllocatorData);
	poly = NULL;
    }
    if (thinPoly) {
	IPW_FREE(thinPoly, cAllocatorData);
	thinPoly = NULL;
    }
    if (rotPoly) {
	IPW_FREE(rotPoly, cAllocatorData);
	rotPoly = NULL;
    }
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &redPr);
    w_destroyPixr (pIPWState, &tmpPr);
    w_destroyPixr (pIPWState, &phMaskPr);
    w_destroyPixr (pIPWState, &plMaskPr);
    w_destroyPixr (pIPWState, &gradPr);
    w_destroyPixr (pIPWState, &markerPr);
    w_destroyStrel (pIPWState, &s);
    IPW_FREE (hf, cAllocatorData);
    IPW_FREE (hc, cAllocatorData);
    if (poly) {
	IPW_FREE(poly, cAllocatorData);
	poly = NULL;
    }
    if (thinPoly) {
	IPW_FREE(thinPoly, cAllocatorData);
	thinPoly = NULL;
    }
    if (rotPoly) {
	IPW_FREE(rotPoly, cAllocatorData);
	rotPoly = NULL;
    }
    ERR_RETURN;

}  /* w_crop5 () */



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Given a rectangular region, crop and/or rotate the region.
 *
 * LongDescription:::
 *   w_cropAndRotate() crops and/or rotates a region-of-interest, as given by
 *   a rectangular region (that would normally be at the rotation angle with
 *   respect to the input image). Three basic operations are performed: 1)
 *   the image is cropped to the boundaries of the rectangle, 2) the cropped
 *   image is rotated; and 3) the rotated image is cropped to match the rotated
 *   rectangle.
 *   <BR>
 *   This routine is a natural follow-on to the w_crop4() and w_crop5()
 *   routines.
 *
 * Input:::
 *  pIPWState::		State structure
 *  inPr::		Input image
 *  rectX1::		Rectangle X coordinate for first point
 *  rectY1::		Rectangle Y coordinate for first point
 *  rectX2::		Rectangle X coordinate for second point
 *  rectY2::		Rectangle Y coordinate for second point
 *  rectX3::		Rectangle X coordinate for third point
 *  rectY3::		Rectangle Y coordinate for third point
 *  rectX4::		Rectangle X coordinate for fourth point
 *  rectY4::		Rectangle Y coordinate for fourth point
 *  rotationAngle::	Angle to rotate the image in the clockwise direction,
 *			given in radians
 *  deskew::		Image is to be deskewed (boolean)
 *  crop::		Image is to be cropped (boolean)
 *  
 * Output:::
 *  outPr::		Resulting image (deskewed and/or cropped)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping, rotation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_cropAndRotate (
	IPWState *pIPWState,
	PIXR *inPr,
	Int32 rectX1,
	Int32 rectY1,
	Int32 rectX2,
	Int32 rectY2,
	Int32 rectX3,
	Int32 rectY3,
	Int32 rectX4,
	Int32 rectY4,
	Float64 rotationAngle,
	Bool deskew,
	Bool crop,
	PIXR **outPr)
{
    PROLOGUE;
    Int32            w, h, d, c,
		     pivotX, pivotY;
    ScaledInt1616    xdpi, ydpi;
    ipPoint         *poly = NULL,
		    *rotPoly = NULL;
    PIXR            *preCropPr = NULL,
		    *rotPr = NULL;
    Int32 minX, minY, maxX, maxY;
    Int32 dx, dy;
    Int32 point;
    Int32 preCropW, preCropH;
    ColorType	    colortype;


    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
		  cChannels1 | cChannels3, cFrame32);
    ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &d));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, inPr, &c));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xdpi, &ydpi));
    if (c == 3 ) {
	colortype = cColorRGB;
    } else {
	colortype = cColorGray;
    }

    /* Check to see if something should actually be done */
    if (deskew == cFalse && crop == cFalse)
    {
	ERR_PASS_CLEANUP(w_copy(pIPWState, inPr, outPr));
	NORMAL_RETURN;
    }

    /* Unless precropping is necessary, don't precrop */
    preCropPr = inPr;

    if (crop) {

	/* Prepare to perform the precrop
	 */
	ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xdpi, &ydpi));
	ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &h));
	minX = maxX = rectX1;
	minY = maxY = rectY1;
	if (rectX2 < minX) minX = rectX2;
	if (rectX3 < minX) minX = rectX3;
	if (rectX4 < minX) minX = rectX4;
	if (rectY2 < minY) minY = rectY2;
	if (rectY3 < minY) minY = rectY3;
	if (rectY4 < minY) minY = rectY4;
	if (rectX2 > maxX) maxX = rectX2;
	if (rectX3 > maxX) maxX = rectX3;
	if (rectX4 > maxX) maxX = rectX4;
	if (rectY2 > maxY) maxY = rectY2;
	if (rectY3 > maxY) maxY = rectY3;
	if (rectY4 > maxY) maxY = rectY4;

	/* Don't really go beyond the image boundaries */
	if (minX < 0) minX = 0;
	if (minY < 0) minY = 0;
	if (maxX >= w) maxX = w - 1;
	if (maxY >= h) maxY = h - 1;

	if (minX != 0 || minY != 0 || maxX != w || maxY != h) {

	    /* Precrop
	     */

	    dx = maxX - minX;
	    dy = maxY - minY;

	    /* Create an image of the precropped size */
	    preCropPr = NULL;
	    ERR_PASS_CLEANUP(w_createPixrExtended(pIPWState,
						  dx,
						  dy,
						  d,
						  xdpi, ydpi,
						  cFrameBits,
						  c,
						  colortype,
						  NULL,
						  &preCropPr));

	    /* What does this do???? */
#if 0
	    ERR_PASS_CLEANUP(w_invert(pIPWState, preCropPr, &preCropPr));
#endif

	    /* Extract the relevant portion (precrop) */
	    ERR_PASS_CLEANUP(w_rasterop(pIPWState,
					minX,
					minY,
					inPr,
					PIX_SRC,
					0,
					0,
					dx,
					dy,
					preCropPr));

	} /* If precropping */

#if DEBUG_CROP
	{
	    char fileN[100];
	    sprintf (fileN, "%s-preCrop.tif", project);
	    ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
				       preCropPr, TIFF_DEFAULT));
	}
#endif

	/* Adjust the cropping recangle by the precrop position */
	rectX1 -= minX;
	rectX2 -= minX;
	rectX3 -= minX;
	rectX4 -= minX;
	rectY1 -= minY;
	rectY2 -= minY;
	rectY3 -= minY;
	rectY4 -= minY;

    } /* If crop requested */

    /* Whether rotating or not, default to current image */
    rotPr = preCropPr;

    /* Calculate the pivot point, the center of the relevant rectangle */
    pivotX = (rectX1 + rectX2 + rectX3 + rectX4 + 2) / 4;
    pivotY = (rectY1 + rectY2 + rectY3 + rectY4 + 2) / 4;

    if (deskew && rotationAngle != 0.0) {
	char rotSuccess = cFalse;

	/* First try small-angle interpolated rotate */
	if (ABS(rotationAngle) < 10.0F * 3.1415926F / 180.0F) {
	    rotPr = NULL;
	    ERR_CHECK_NO_MSG (w_rotateGray(pIPWState,
					   preCropPr,
					   (Float32)rotationAngle,
					   &rotPr)) {
		rotSuccess = cFalse;
	    } else {
		rotSuccess = cTrue;
	    }
	}

	/* If that failed (might be that angle is too large), then try
	 * arbitrary interpolated rotate with magnification 1.0 */
	if (rotSuccess != cTrue) {
	    rotPr = NULL;
	    ERR_CHECK_NO_MSG (w_rotateMag(pIPWState,
					  preCropPr,
					  rotationAngle,
					  1.0, 1.0,
					  (Float64)pivotX,
					  (Float64)pivotY,
					  &rotPr)) {
		rotSuccess = cFalse;
	    } else {
		rotSuccess = cTrue;
	    }
	}

	/* If that failed (who knows why), then try arbitrary
	 * interpolated rotate (takes twice as long as the magnified
	 * one, but is straight-forward) */
	if (rotSuccess != cTrue) {
	    rotPr = NULL;
	    ERR_CHECK_NO_MSG (w_rotateInterpolated(pIPWState,
						   preCropPr,
						   (Float32)rotationAngle,
						   0L,
						   &rotPr)) {
		rotSuccess = cFalse;
	    } else {
		rotSuccess = cTrue;
	    }
	}

	/* If that failed, then try non-interpolated (shearing) rotate. */
	if (rotSuccess != cTrue) {
	    rotPr = NULL;
	    ERR_PASS_CLEANUP(w_rotateArb(pIPWState,
					 preCropPr,
					 (Float32)rotationAngle,
					 cTrue,
					 cFalse,
					 &rotPr));
	}

#if DEBUG_CROP
	    {
		char fileN[100];
		sprintf (fileN, "%s-preCropRotated.tif", project);
		ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
		    rotPr, TIFF_PACKBITS));
	    }
#endif
	/* could free the original image at this point */

	/* Deskew complete. Since deskew required, then continue on with
	 * the final cropping, if cropping. */

	if (crop) {

	    /* Rotate the rectangle by the same amount */
	    poly = (ipPoint *)IPW_CALLOC(sizeof(*poly), 4, cAllocatorData);
	    ERR_GEN_CLEANUP(poly == NULL, cErrNomem);
	    poly[0].x = rectX1;
	    poly[0].y = rectY1;
	    poly[1].x = rectX2;
	    poly[1].y = rectY2;
	    poly[2].x = rectX3;
	    poly[2].y = rectY3;
	    poly[3].x = rectX4;
	    poly[3].y = rectY4;
	    ERR_PASS_CLEANUP (w_rotatePolygon (pIPWState,
					   pivotX, pivotY,
					   poly,
					   4,
					   rotationAngle,
					   &rotPoly));

	    /* Compute the final cropping points from the rotated polygon
	     */
	    minX = maxX = rotPoly[0].x;
	    minY = maxY = rotPoly[0].y;
	    for (point = 1; point < 4; point++) {
		if (rotPoly[point].x < minX) minX = rotPoly[point].x;
		if (rotPoly[point].x > maxX) maxX = rotPoly[point].x;
		if (rotPoly[point].y < minY) minY = rotPoly[point].y;
		if (rotPoly[point].y > maxY) maxY = rotPoly[point].y;
	    }

	    /* Don't really go beyond the image boundaries */
	    ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, preCropPr,
		&preCropW));
	    ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, preCropPr,
		&preCropH));
	    if (minX < 0) minX = 0;
	    if (minY < 0) minY = 0;
	    if (maxX >= preCropW) maxX = preCropW - 1;
	    if (maxY >= preCropH) maxY = preCropH - 1;

	    if (minX > 0 || minY > 0 || maxX < preCropW || maxY < preCropH) {

		ERR_PASS_CLEANUP(w_cutRectIP(pIPWState,
					 minX,
					 minY,
					 maxX - minX,
					 maxY - minY,
					 rotPr));
	    }

#if DEBUG_CROP
	    {
		char fileN[100];
		sprintf (fileN, "%s-finalCrop.tif", project);
		ERR_PASS_CLEANUP (w_write (pIPWState, fileN, 
		    rotPr, TIFF_PACKBITS));
	    }
#endif
	} /* if doing final crop */

    }	/* Deskewing */

#if DEBUG_PRINT
    fprintf (stdout, "Rotated and cropped in %f seconds.\n",
	     StopTime ());
    StartTime();
#endif

    /* If nothing was done, copy the input */
    if (rotPr == inPr) {
	ERR_PASS_CLEANUP(w_copy(pIPWState, inPr, outPr));
    } else {

	/* Save the result */
	*outPr = rotPr, rotPr = NULL;
    }

ERR_CLEANUP:

    /* Clean up */
    if (preCropPr != inPr && preCropPr != *outPr) {
	w_destroyPixr (pIPWState, &preCropPr);
    }
    if (rotPr != inPr) {
	w_destroyPixr (pIPWState, &rotPr);
    }
    if (poly) {
	IPW_FREE(poly, cAllocatorData);
	poly = NULL;
    }
    if (rotPoly) {
	IPW_FREE(rotPoly, cAllocatorData);
	rotPoly = NULL;
    }
    NORMAL_RETURN;

}  /* w_cropAndRotate () */



static ErrCode CDECL
w_subtractHisto (
                 IPWState *pIPWState,
	         Int32 *h1,
	         Int32 *h2,
	         Int32  length)
{
    PROLOGUE;
    Int32 i;

    ERR_GEN(pIPWState == NULL || h1 == NULL || h2 == NULL, cErrNullPtr);

    for (i = 0; i < length; i++)
        h2[i] -= h1[i];

    NORMAL_RETURN;

}  /* w_subtractHisto () */


/***********************************************
 ***********************************************/
static ErrCode CDECL
wp_countBlackCorners (
                      IPWState *pIPWState, 
		      PIXR *pr, 
		      Int32 *count)
{
    PROLOGUE;
    UInt8  *image;
    Int32   pCount, w, h, bpl;

    *count = 0;

    ERR_PASS (w_pixrGetImage (pIPWState, pr, &image));
    ERR_PASS (w_pixrGetWidth (pIPWState, pr, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, pr, &h));
    ERR_PASS (w_pixrGetBpl (pIPWState, pr, &bpl));

    /* top left */
    pCount = GET_IMAGE_BIT (image, 0);
    pCount += GET_IMAGE_BIT (image, 1);
    pCount += GET_IMAGE_BIT (image + bpl, 0);
    pCount += GET_IMAGE_BIT (image + bpl, 1);

    if (pCount >= 3)
        (*count)++;

    /* top right */
    pCount = GET_IMAGE_BIT (image, w - 2);
    pCount += GET_IMAGE_BIT (image, w - 1);
    pCount += GET_IMAGE_BIT (image + bpl, w - 2);
    pCount += GET_IMAGE_BIT (image + bpl, w - 1);

    if (pCount >= 3)
        (*count)++;

    image += (h - 1) * bpl;

    /* bottom left */
    pCount = GET_IMAGE_BIT (image, 0);
    pCount += GET_IMAGE_BIT (image, 1);
    pCount += GET_IMAGE_BIT (image - bpl, 0);
    pCount += GET_IMAGE_BIT (image - bpl, 1);

    if (pCount >= 3)
        (*count)++;

    /* bottom right */
    pCount = GET_IMAGE_BIT (image, w - 2);
    pCount += GET_IMAGE_BIT (image, w - 1);
    pCount += GET_IMAGE_BIT (image - bpl, w - 2);
    pCount += GET_IMAGE_BIT (image - bpl, w - 1);

    if (pCount >= 3)
        (*count)++;

    NORMAL_RETURN;

}  /* wp_countBlackCorners () */

/*
$Log: crop3.c,v $
Revision 1.1.1.1  1999/09/20 18:19:15  jlin
Imported sources

Revision 1.12  1998/02/25 19:20:16  aber
crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

CommonCheck: To be consistent, the skew detection functions return the angle required
to correct skew. This is the same angle passed on to the rotation algorithm
itself, and is positive is clockwise rotation is necessary. w_findSkewSample
does this also, but w_crop4 was returning the angle of skew, not the angle
to repair the skew. Both w_crop4 and w_crop5 now return the angle to
correct for skew.

Revision 1.11  1998/02/19 21:27:14  aber
aber on Thu Feb 19 13:25:58 PST 1998

crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Don't run crop5 if it doesn't handle the mode (binary or paletted).
Allow rotate and crop routine to work if image is binary.

Revision 1.10  1998/02/11 23:28:34  aber
aber on Wed Feb 11 15:27:22 PST 1998

crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Correct the sign on the rotation angle, to match that of w_crop4() --
namely, the angle to correct the skew, not the angle of skew.

Revision 1.9  1998/01/31 01:05:17  aber
crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Move new photo crop detection algorithm into a separate function w_crop5().
Added a new function to handle the crop and deskew, and correctly handle
one or the other, as well as both.
Keep functionality of w_cropPhoto() basically the same, but it now calls
w_crop5() and w_cropAndRotate(). It also now does what was in autoscan.c,
namely if w_crop5() fails to detect anything, w_crop4() is called.

Revision 1.8  1998/01/20 23:58:02  aber
aber on Tue Jan 20 15:56:48 PST 1998

crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Bug fix, for large images, when reducing by a factor of 4, use 4
in the calculations, not 2.
Put debug code back in, ifdef'd out.

Revision 1.7  1998/01/20 03:18:11  aber
aber on Mon Jan 19 19:17:19 PST 1998

crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Rework of rotation routine checking: use ERR_CHECK_NO_MSG() macro to
avoid annoying assertion during print.

Revision 1.6  1998/01/16 23:18:20  aber
crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Use w_rotateMag to do interpolated rotation, 2X faster.

Revision 1.5  1998/01/16 00:51:16  aber
aber on Thu Jan 15 16:50:38 PST 1998

crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Use arbitrary interpolated rotation before using the non-interpolated
method.

Revision 1.4  1998/01/07 02:18:17  aber
aber on Tue Jan 6 18:18:03 PST 1998

crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Fixed memory leaks.

Revision 1.3  1997/09/30 22:04:50  aber
aber on Tue Sep 30 15:03:39 PDT 1997

crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Check to see if the watershedding found a separation between photo
and platen. If not, return an error, rather than having contour routine
complain loudly.

Revision 1.2  1997/09/19 16:51:05  aber
aber on Fri Sep 19 09:50:53 PDT 1997

crop3.c -> /product/ipcore/ipwrap/src/RCS/crop3.c,v

Remove debug code.

Revision 1.1  1997/08/27 00:52:01  aber
Initial revision
*/
