/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * cropprv.c
 *
 *	"Auto-crop" private functions
 *
 * Functions defined:
 *
 */

#include <stdio.h>
#include <math.h>
#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "alplib.h"
#include "imageref.pub"
#include "defines.pub"
/* prototypes */
#include "crop.pub"
#include "crop.prv"
#include "shrcc.pub"
#include "geoutil.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: cropprv.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


/* Constants (tunable) */
#define	    cDiffSignificant     /*100.00*/ 30.0	/* High-contrast significant diff^2 */
#define	    cReasonableErr	 50.0	/* TBD */

#define	    cMinData		6	/* Min. points to fit (LC only) */
#define	    cMinDataPercent	60	/* Min % of points to fit (HC only) */
#define	    cSlopPercentage	1	/* % of rect that we should ignore, due to 
					   the common black stripe on platen edges */
#define	    cColinearCutoff	5.0	/* RMS pixels for lines to be nearly colinear */

/* Note! The code assumes that cAvgWidth is a power of two, so we can mask
   to get a modulo. */
#define	    cAvgWidth		8	    /* average region is this ^ 2 */
#define	    cAvgMask		7	    /* Mask for average width */
#define	    cAvgHalf		4	    /* Half of above */
#define	    cLog2AvgPixels	6	    /* log2 of (avg width)^2 */
#define	    cInvAvgPixels	0.015625    /* 1/64 */

#define cTop	0
#define cLeft	1
#define cBottom 2
#define cRight	3
#define cEdges	4

#define cVertical   0
#define cHorizontal 1

/*#define DEBUG_MSGS */

#ifdef DEBUG_MSGS
static FILE    *debug_file;
static void writeEdge (FILE *file, const char *str, ipPoint *points, Int32 count);
#endif

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Given an image, fit a crop box to high-contrast edges
 *
 * LongDescription:::
 *  This function searches in from each image edge for transitions
 *  that would indicate the edge of an object on the platen. It then
 *  attempts to fit a crop box to the points by fitting lines on each
 *  of four sides individually.
 *  <BR>
 *  The interface is kind of strange--it used to live in ipshared,
 *  and so it takes "pointer to image data" and "bytes/line" stuff
 *  rather than PIXR pointers.
 *
 * PrivateDescription:::
 *	The general scheme is:
 *	1.  take M vertical and N horizontal passes slices through the
 *	    image, and find "likely" points using two methods, a low-contrast
 *	    "absolute" method, and a high-contrast "relative" method.
 *	2.  Fit a crop box to the points using some method. First, I'm
 *	    going to try linear least squares.
 *
 *
 * Input:::
 *  pIPWState::      State structure
 *  pImage::         Pointer to image data, including frame
 *  Bpl::            Bytes/line of image data
 *  depth::          Depth (bits/pixel) of image data
 *  x::              Region of interest
 *  y::              
 *  w::              
 *  h::              
 *  dpi::            Image's DPI
 *
 * Output:::
 *  err::            Error term
 *  angle::          Angle of crop box, in radians
 *  x0::             Upper left x
 *  y0::             Upper left y
 *  x1::             Upper right x
 *  y1::             Upper right y
 *  x2::             Lower left x
 *  y2::             Lower left y
 *  x3::             Lower right x
 *  y3::             Lower right y
 *
 * Returns:::
 *  Errror status
 * Notes:::
 *	This can find a "rotated" crop box. Figuring out the extent box
 *	is obviously straightforward, but I'll probably write another function
 *	to do it just for convenience, which will have this one do the work.
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_do_crop (
    IPWState           *pIPWState,
    UInt8              *pImage,
    Int32               Bpl,
    Int32               depth,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               dpi,
    Float64            *err,
    Float64            *angle,
    Int32              *x0,
    Int32              *y0,
    Int32              *x1,
    Int32              *y1,
    Int32              *x2,
    Int32              *y2,
    Int32              *x3,
    Int32              *y3)
{
    PROLOGUE;
    ipPoint  edgePoints[cEdges][cMaxSamples];
    ipPoint  cropBox[4];

#ifdef DEBUG_MSGS
    int i;
#if defined(__WINDOWS_386__) || defined(_WIN32)
    debug_file = fopen("c:\CROPDB.TXT", "w");
#else
    debug_file = fopen("cropdb.txt", "w");
#endif
    if (!debug_file)
	return -1;
    fprintf(debug_file, "Crop: xy %ld, %ld, w=%ld, h=%ld\n", x, y, w, h);
    fprintf(debug_file, "DPI %ld\n", dpi);
#endif

    /* Find the points */
    ERR_PASS( wp_crop_findEdgePoints (pIPWState, pImage, Bpl, depth, 
				   x, y, w, h, edgePoints));

#ifdef DEBUG_MSGS
    writeEdge (debug_file, "edgePoints[cTop]", &edgePoints[cTop][0], cColumnSamples);
    writeEdge (debug_file, "edgePoints[cLeft]", &edgePoints[cLeft][0], cRowSamples);
    writeEdge (debug_file, "edgePoints[cBottom]", &edgePoints[cBottom][0], cColumnSamples);
    writeEdge (debug_file, "edgePoints[cRight]", &edgePoints[cRight][0], cRowSamples);
#endif

    /* Do the fits */
    ERR_PASS( wp_crop_fitCropBox (pIPWState, pImage, Bpl, depth, x, y, w, h, 
			       edgePoints, dpi, &cropBox[0], err, angle));
#ifdef DEBUG_MSGS
    fprintf(debug_file, "box (%f): ", *err);
    for (i = 0; i < 4; i++)
	fprintf(debug_file, "(%ld, %ld) ", cropBox[i].x, cropBox[i].y);
    fprintf(debug_file, "\n");
#endif

    /* Stuff whichever box it was into the return variables */
    *x0 = cropBox[0].x;
    *y0 = cropBox[0].y;
    *x1 = cropBox[1].x;
    *y1 = cropBox[1].y;
    *x2 = cropBox[2].x;
    *y2 = cropBox[2].y;
    *x3 = cropBox[3].x;
    *y3 = cropBox[3].y;

#ifdef DEBUG_MSGS
    fclose (debug_file);
#endif	    
    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *      Look for the edge intersections in an image
 *
 * LongDescription:::
 *      For each of the four boundaries of the region of interest,
 *      look inwards to find the first edge point for some samples
 *      along each boundary.
 *
 * PrivateDescription:::
 *	Look for the edge points in the image. We're using 8x8 averaging.
 *	Rather than being stupid and re-adding all 8x8 pixels on every line,
 *	we can just keep a rolling list of the sum of 1x8 segments, and a 
 *	running total. Then, when we get to  new line, we take the bottom sum, 
 *	subtract it from the running total, add the new segment's sum to the 
 *	total, and stick the new segment's sum into the rolling list. 
 *	Then, just divide to get the average for that point. Of course, 
 *	that will be the center of the average region,
 *	so we'll have to keep track of all of that. Life is much easier if
 *	we just ignore the 4 outermost pixels of the image. We can do this 
 *	since we have a frame, so we can always enlarge the area of interest by
 *	four pixels in each direction, but it would be better to just
 *	move in by four, so the zeros don't throw us off.
 *
 * Input:::
 *  pIPWState::    State structure
 *  pImage::       Pointer to image data, including frame
 *  Bpl::          Bytes/line
 *  depth::        Image depth, 1, 4 or 8        
 *  x::            Region of interest
 *  y::                              
 *  w::                              
 *  h::                              
 *
 * Output:::
 *  edgePoints[4][cMaxSamples]::   For each of the four edges (left,
 *                                      right, top, bottom) the edge
 *                                      intersections for each sample
 *                                      in the edge (cRowSamples or
 *                                      cColumnSamples samples)
 *                                      
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_findEdgePoints (
    IPWState           *pIPWState,
    UInt8              *pImage,
    Int32               Bpl,
    Int32               depth,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    ipPoint             edgePoints[4][cMaxSamples])
{
    PROLOGUE;

    Float64  **colDiffs=0;
    Float64  **rowDiffs=0;
    Int32   xOffset, yOffset, xDelta, yDelta, i, result;


    if (depth > 1)
    {
	/* Allocate memory for the arrays of row/column vectors */
	colDiffs = (Float64 **) IPW_CALLOC(cColumnSamples, sizeof(Float64 *), 
					   cAllocatorData);
	ERR_GEN_CLEANUP (!colDiffs, cErrNomem);

	rowDiffs = (Float64 **) IPW_CALLOC(cRowSamples, sizeof(Float64 *),
					   cAllocatorData);
	ERR_GEN_CLEANUP (!rowDiffs, cErrNomem);

	/* Allocate the row and column average vectors */
	for (i = 0; i < cColumnSamples; i++)
	{
	    colDiffs[i] = (Float64 *)  IPW_CALLOC (sizeof (Float64), h,
						   cAllocatorData);
	    ERR_GEN_CLEANUP (!colDiffs[i], cErrNomem);
	}
	    
	for (i = 0; i < cRowSamples; i++)
	{

	    rowDiffs[i] = (Float64 *) IPW_CALLOC (sizeof (Float64), w,
						 cAllocatorData);
	    ERR_GEN_CLEANUP (!rowDiffs[i], cErrNomem);
	}
    }

    /* Move the image pointer down to the right place */
    switch (depth)
    {
	case 1: pImage += (x >> 3) + (y * Bpl); break;
	case 4: pImage += (x >> 1) + (y * Bpl); break;
	case 8: pImage += x + y * Bpl; break;
    	default: ERR_GEN_CLEANUP (1, cErrTypeNotSupported);
    }

    xDelta = w / (cColumnSamples + 1);
    yDelta = h / (cRowSamples + 1);

    /* Do the columns */
    for (i = 0; i < cColumnSamples; i++)
    {
	xOffset = (i + 1) * xDelta;

	if (depth == 1)
	{
	    /* We do depth 1 differently... */
	    ERR_PASS_CLEANUP( wp_crop_computeBinaryColumnEdge (pIPWState, pImage, Bpl, xOffset, 
					      h, 1, &result));
	    edgePoints[cTop][i].x = xOffset + x;
	    edgePoints[cTop][i].y = (result < 0) ? result : result + y; 

	    ERR_PASS_CLEANUP( wp_crop_computeBinaryColumnEdge (pIPWState, pImage, Bpl, xOffset,
					      h, -1, &result));
	    edgePoints[cBottom][i].x = xOffset + x;
	    edgePoints[cBottom][i].y = (result < 0) ? result : 
		                                      h - result + y;
	}
	else
	{
	    /* Get the averages for that slice */
	    ERR_PASS_CLEANUP( wp_crop_sliceDiffs (pIPWState, pImage, Bpl, depth, 
					       xOffset, h, cVertical, 
					       colDiffs[i]));
	    
	    /* Compute points from both ends using both methods */
	    edgePoints[cTop][i].x = xOffset + x;
	    ERR_PASS_CLEANUP( wp_crop_computeEdgePoint (pIPWState, colDiffs[i], 1,
						     h, &result));
	    edgePoints[cTop][i].y = ((result < 0) ? result : result + y);

	    edgePoints[cBottom][i].x = xOffset + x;
	    ERR_PASS_CLEANUP( wp_crop_computeEdgePoint (pIPWState, 
						     colDiffs[i] + h - 1, -1,
						     h, &result));
	    edgePoints[cBottom][i].y = ((result < 0) ? result : 
					               h - result + y);
	}
    }

    /* Do the rows */
    for (i = 0; i < cRowSamples; i++)
    {
	yOffset = (i + 1) * yDelta;

	if (depth == 1)
	{
	    /* We do depth 1 differently... */
	    ERR_PASS_CLEANUP( wp_crop_computeBinaryRowEdge (pIPWState, pImage, Bpl, yOffset, w, 
					   1, &result));
	    edgePoints[cLeft][i].y = yOffset + y;
	    edgePoints[cLeft][i].x = (result < 0) ? result : result + x; 

	    ERR_PASS_CLEANUP( wp_crop_computeBinaryRowEdge (pIPWState, pImage, Bpl, yOffset, w, 
					   -1, &result));
	    edgePoints[cRight][i].y = yOffset + y;
	    edgePoints[cRight][i].x = (result < 0) ? result : result + x;
	}
	else
	{
	    ERR_PASS_CLEANUP( wp_crop_sliceDiffs (pIPWState, pImage, Bpl, depth, 
					       yOffset, w, cHorizontal, 
					       rowDiffs[i]));

	    /* Compute points from both ends */
	    edgePoints[cLeft][i].y = yOffset + y;
	    ERR_PASS_CLEANUP( wp_crop_computeEdgePoint (pIPWState, rowDiffs[i], 1,
						     w, &result));
	    edgePoints[cLeft][i].x = ((result < 0) ? result : result + x);

	    edgePoints[cRight][i].y = yOffset + y;
	    ERR_PASS_CLEANUP( wp_crop_computeEdgePoint (pIPWState, 
						     rowDiffs[i] + w - 1, -1, 
						     w, &result));
	    edgePoints[cRight][i].x = ((result < 0) ? result : w - result + x);
	}
    }


done:

    if (depth > 1)
    {
	if (colDiffs)
	{
	    /* Free the row and column average vectors */
	    for (i = 0; i < cColumnSamples; i++)
	    {
		if (colDiffs[i])
		    IPW_FREE (colDiffs[i], cAllocatorData);
	    }
	    
	    IPW_FREE(colDiffs, cAllocatorData);
	}

	if (rowDiffs)
	{
	    for (i = 0; i < cRowSamples; i++)
	    {
		if (rowDiffs[i])
		    IPW_FREE (rowDiffs[i], cAllocatorData);
	    }

	    IPW_FREE(rowDiffs, cAllocatorData);
	}
    }

    if (__error__)
	ERR_RETURN;

    NORMAL_RETURN;

ERR_CLEANUP:
    goto done;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   slice through an image, compute rolling average of pixels and diffs^2
 *
 * LongDescription:::
 *  Take a slice through an image, and average cAvgWidth^2 blocks
 *  of pixels to get an average vector. Also compute a square-of-differences
 *  vector. Note that the cAvgWidth/2 pixels around the edge won't have
 *  a valid average, and the cAvgWidth/2+1 pixels around the edge won't have
 *  a valid difference^2.
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pImg::           Pointer to frame of image
 *  Bpl::            Bytes between lines of image
 *  depth::          Depth--1, 4, or 8
 *  offset::         Offset (x or y) at which to take slice
 *  samples::        Size of the slice
 *  direction::      cVertical or cHorizontal
 *
 * Output:::
 *  diffs::          Vector in which to stick differences of averages
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_sliceDiffs (
    IPWState           *pIPWState,
    UInt8              *pImg,
    Int32               Bpl,
    Int32               depth,
    Int32               offset,
    Int32               samples,
    Int32               direction,
    Float64            *diffs)
{
    PROLOGUE;
    Int32   rollingSums[cAvgWidth], totalSum, i, thisSum;
    Float64 diff, thisAverage, lastAverage;
    totalSum = 0;
    lastAverage = 0;

    for (i = 0; i < cAvgWidth; i++)
	rollingSums[i] = 0;

    if (offset < cAvgHalf)
	offset = cAvgHalf;

    /* Run through this slice. */
    for (i = 0; i < samples - cAvgHalf; i++)
    {
	/* get the sum for this segment */
	ERR_PASS( wp_crop_sumPixels (pIPWState, pImg, Bpl, depth, i, 
				  offset - cAvgHalf,
				  cAvgWidth, direction, &thisSum));

	/* Sort of a kludge: scale 'em up, so that the same constants work */
	if (depth == 4)
	    thisSum <<= 4;

	/* Add in the sum we just got */
	totalSum += thisSum;
	if (i >= cAvgWidth)
	{
	    /*  
	    * If we've gone far enough to roll, take out the lowest one.
	    * Note that  ((i-avgWidth) mod avgWidth) is the same as i mod avgWidth,
	    * meaning that the one we're taking out is the same one we're
	    * putting back in, which only makes sense. 
	    */
	    totalSum -= rollingSums[i & cAvgMask];

	    /* Stick the average in the average vector */
	    thisAverage = totalSum * cInvAvgPixels;

	    /* Compute the difference^2 */
	    if (i >= cAvgWidth)
	    {
		diff = thisAverage - lastAverage;
		diffs[i - cAvgHalf] = diff * diff;
	    }
	    lastAverage = thisAverage;
	}
	    
	/* Put the sum at the tail of the rolling list */
	rollingSums[i & cAvgMask] = thisSum;
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  sum some pixels along the minor axis.
 *
 * LongDescription:::
 *  Add up some pixels along the minor axis. "direction" is the direction 
 *  of the major axis, that is, the axis along which a slice is being taken.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pImg::           Pointer to frame of image
 *  Bpl::            Bytes between lines in image
 *  depth::          Depth, 1, 4, or 8
 *  major::          Major axis value, depends on direction
 *  minor::          Minor axis starting value, depends on direction
 *  pixels::         Number of pixels to sum
 *  direction::      Direction, cHorizontal or cVertical
 *
 * Output:::
 *  pPixelSum::      Sum of pixels 
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_sumPixels (
    IPWState           *pIPWState,
    UInt8              *pImg,
    Int32               Bpl,
    Int32               depth,
    Int32               major,
    Int32               minor,
    Int32               pixels,
    Int32               direction,
    Int32              *pPixelSum)
{
    PROLOGUE;
    register UInt8 *img;
    register Int32 i, sum, bpl, byte, offset;

    pIPWState = pIPWState;
    sum = 0;
    img = pImg;
    bpl = Bpl;

    if (direction == cVertical)
    {
	/* Sum in the x-direction */

	if (depth == 8)
	{
	    img += bpl * major + minor;
	    for (i = 0; i < pixels; i++)
	    {
		sum += UCHAR_ACCESS(img + i);
	    }
	}
	else
	{
	    img += bpl * major;
	    offset = minor >> 1;
	    pixels >>= 1;
	    if (minor & 1)
	    {

		/* Odd starting offset; take the low nibble */
		sum += UCHAR_ACCESS (img + offset) & 0xf;

		for (i = 1; i < pixels; i++)
		{
		    byte = UCHAR_ACCESS (img + offset + i);
		    sum += (byte & 0xf) + (byte >> 4);
		}

		/* We missed the last one; take the high nibble */
		sum += UCHAR_ACCESS (img + offset + i) >> 4; 
	    }
	    else
	    {
		for (i = 0; i < pixels; i++)
		{
		    byte = UCHAR_ACCESS (img + offset + i);
		    sum += (byte & 0xf) + (byte >> 4);
		}
	    }
	}
    }
    else
    {
	/* Sum in the y-direction */
	if (depth == 4)
	{
	    img += (major >> 1) + minor * bpl;
	    /* Sum the even or odd nibble, depending on which line we're on */
	    if (major & 1)
	    {
		for (i = 0; i < pixels; i++)
		{
		    sum += UCHAR_ACCESS(img) >> 4;
		    img += bpl;
		}
	    }
	    else
	    {
		for (i = 0; i < pixels; i++)
		{
		    sum += UCHAR_ACCESS(img) & 0xf;
		    img += bpl;
		}
	    }
	}
	else
	{
	    img += major + minor * bpl;
	    for (i = 0; i < pixels; i++)
	    {
		sum += UCHAR_ACCESS(img);
		img += bpl;
	    }
	}
    }

    *pPixelSum = sum;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute an edge point by maching along a square-of-differences array
 *
 * LongDescription:::
 *	Compute an edge point by marching along a square-of-differences array.
 *	This is the "high contrast" case, where we try to walk over a peak of
 *	differences.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pDiff::          Pointer to difference vetor
 *  direction::      Distance between elements, 1 or -1
 *  max::            Max. # of points to examine before giving up
 *
 * Output:::
 *  pEdgePoint::     The index into the vector where the point was found
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_computeEdgePoint (
    IPWState           *pIPWState,
    Float64            *pDiff,
    Int32               direction,
    Int32               max,
    Int32              *pEdgePoint)
{
    PROLOGUE;
    Int32 i, lim, initial;
    Float64 /*lastDiff = 0, */ diff;

    pIPWState = pIPWState;
    /* Well, instead of doing that (11/29/93), ignore cSlopPercentage pixels */
    initial = (max * cSlopPercentage) / 100;
    if (initial < 8)
	initial = 8;
    /* Limit is 4* initial, which is a bit of a hack to prevent us from
       running into scanner platen edge artifacts on the other end */
    lim = max - initial * 4 ;
    *pEdgePoint = -1;

    for (i = initial; i < lim; i++)
    {
	diff = pDiff[(direction > 0) ? i : -i];

	if (diff > cDiffSignificant)
	{
	    /*lastDiff = diff; */
	    *pEdgePoint = i + cAvgHalf;
	    break;
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Look for the first transition down or up a column of a binary image
 *
 * LongDescription:::
 *	Look for the first transition down or up a column of a binary image
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::        State structure
 *  pImage::           Pointer to image, including frame
 *  bpl::              Bytes/line in image
 *  xOffset::          X-offset of column
 *  h::                Height of image, in pixels
 *  dir::              >0 = top down, <0 = bottom up
 *
 * Output:::
 *  pFirstTransition:: line at which a transition was found, 
 *                     or -1 if none found (which is not an error)
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_computeBinaryColumnEdge (
    IPWState           *pIPWState,
    register UInt8     *pImage,
    Int32               bpl,
    Int32               xOffset,
    Int32               h,
    Int32               dir,
    Int32              *pFirstTransition)
{
    PROLOGUE;
    register UInt8 byteMask;
    register Int32 i;

    pIPWState = pIPWState;

    /* Make a mask for that bit */
    byteMask = 1 << (xOffset & 7);

    /* Offset the image pointer to the right byte-aligned column */
    pImage += (xOffset >> 3);

    /* Get the endian-adjusted offset. The idea is that since lines are
       always longword-aligned, we can add bpl to it without screwing up */
    pImage = &(UCHAR_ACCESS(pImage));
    *pFirstTransition = -1;

    if (dir > 0)
    {
	/* The first one is the starting state */
	if (*pImage & byteMask)
	{
	    pImage += bpl;
	    
	    /* Look for a zero */
	    for (i = 1; i < h; i++)
	    {
		if ((*pImage & byteMask) == 0)
		{
		    *pFirstTransition = i;
		    break;
		}
		pImage += bpl;
	    }
	}
	else
	{
	    pImage += bpl;

	    /* Look for a one */
	    for (i = 1; i < h; i++)
	    {
		if (*pImage & byteMask)
		{
		    *pFirstTransition = i;
		    break;
		}
		pImage += bpl;
	    }
	}
    }
    else
    {
	/* Get to the last line */
	pImage += bpl * (h - 1);
	if (*pImage & byteMask)
	{
	    pImage -= bpl;
	    
	    /* Look for a zero */
	    for (i = 1; i < h; i++)
	    {
		if ((*pImage & byteMask) == 0)
		{
		    *pFirstTransition = i;
		    break;
		}
		pImage -= bpl;
	    }
	}
	else
	{
	    pImage -= bpl;

	    /* Look for a one */
	    for (i = 1; i < h; i++)
	    {
		if (*pImage & byteMask)
		{
		    *pFirstTransition = i;
		    break;
		}
		pImage -= bpl;
	    }
	}
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Look for the first transition forward or backward along a row
 *
 * LongDescription:::
 *	Look for the first transition forward or backward along a row
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::        State structure
 *  pImage::           Pointer to image, including frame
 *  bpl::              Bytes/line in image
 *  yOffset::          Y-offset, in pixels
 *  w::                Width of image, in pixels
 *  dir::              >0 = left-right, <0 = right-left
 *
 * Output:::
 *  pFirstTransition:: Column at which a transition was found,
 *                     or -1 if none found, which is not an error
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_computeBinaryRowEdge (
    IPWState           *pIPWState,
    register UInt8     *pImage,
    Int32               bpl,
    Int32               yOffset,
    Int32               w,
    Int32               dir,
    Int32              *pFirstTransition)
{
    PROLOGUE;
    Int32   result;
    Bool     failFlag;
    UInt8   *byte;

    pIPWState = pIPWState;

    /* Move the pointer to the right row */
    pImage += yOffset * bpl;
    failFlag = cFalse;

    if (dir > 0)
    {
	if (UCHAR_ACCESS(pImage) & 7)
	{
	    /* First bit is 1, find the first zero */
	    if (i_next0ToTheRight (pIPWState, pImage, w, 0, &result))
		failFlag = cTrue;
	    if (result >= w)
		failFlag = cTrue;
	}
	else
	{
	    /* First bit is 0, find the first one */
	    if (i_nextBitToTheRight (pIPWState, pImage, w, 0, &result))
		failFlag = cTrue;
	    if (result >= w)
		failFlag = cTrue;
	}
    }
    else
    {
	byte = pImage + ((w - 1) >> 3);

	if (UCHAR_ACCESS(byte) & (1 << (8 - (w & 7))))
	{
	    /* First bit is 1, find the first zero */
	    if (i_next0ToTheLeft (pIPWState, pImage, 0, w - 1, &result))
		failFlag = cTrue;
	    if (result == 0)
		failFlag = cTrue;
	}
	else
	{
	    /* First bit is 0, find the first one */
	    if (i_nextBitToTheLeft (pIPWState, pImage, 0, w - 1, &result))
		failFlag = cTrue;
	    if (result == 0)
		failFlag = cTrue;
	}
    }

    if (failFlag)
	*pFirstTransition = -1;
    else
	*pFirstTransition = (Int32) result;

    NORMAL_RETURN;

}
	
/* MACROS used in the following routine: */
#define REPLACE_EDGE(edge, p0, p1, p2) \
{\
edges[edge][0] = p0;\
edges[edge][1] = p1;\
edges[edge][2] = p2;\
absDev[edge] = 1e+06;\
}

#define EDGE_COEFS(n) edges[n][0], edges[n][1], edges[n][2]
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *     Fit a crop box given some edge points
 *
 * LongDescription:::
 *	Given a set of points, find a rectangle (can be rotated) that
 *	best encloses the points. If the points are actually points on the
 *	crop rectangle, we should be guaranteed to find that rectangle.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::   State structure
 *  pImage::      Pointer to frame of image                   
 *  Bpl::         Distance between lines, in bytes
 *  depth::       Depth of image
 *  x::           Region of interest     
 *  y::                              
 *  w::                              
 *  h::                              
 *  edgePoints[4][cMaxSamples]:: The edge intersection points, as 
 *                                    computed by wp_crop_findEdgePoints
 *  dpi::         DPI of image
 *
 * Output:::
 *  theBox[4]::   The corner points computed for crop box
 *  error::       Goodness-of-fit measure, where 0.0=perfect fit
 *  angle::       Skew angle, in radians
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_fitCropBox (
    IPWState           *pIPWState,
    UInt8              *pImage,
    Int32               Bpl,
    Int32               depth,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    ipPoint             edgePoints[4][cMaxSamples],
    Int32               dpi,
    ipPoint             theBox[4],
    Float64            *error,
    Float64            *angle)
{
    PROLOGUE;
    Bool linesParallel;
    /*
     * Another attempt: fit each edge individually. The last scheme 
     * (doing a multi-dimensional fit) worked OK, but had two problems:
     * first, it was a bit too sensitive to outlying points, and
     * second, it had no clue about dealing with the case where there
     * is no edge, i.e. one of the sides is outside the platen.
     * The other problem was that it is deathly slow (~30 sec on a SPARC)
     * What I'm going to try is doing a robust fit on each edge
     * individually, and then finding the intersecting corners.
     * In that case, if we detect (by a poor fit) that we don't have
     * an edge, we'll just return the edge of the image for that
     * side of the crop box.
     */
    /*
     * 10/28/93: well, that was pretty good, but it was problematic
     * for two reasons. First, there are sometimes things which have
     * a real nice high-contrast edge on two sides, and a real nice 
     * low contrast edge on the other two sides. So, we're going to
     * fit all of the edges at once, and then pick them individually.
     * Also, for now, we're going to set the threshold for rejecting
     * an edge to some ridiculously high value so that we almost never
     * do it, because we don't seem to figure them out very well.
     */
    Float64 edges[cEdges][3], absDev[cEdges],
	    xI, yI, *diffVec = NULL, sum0, sum1, minError;
    Int32 i, beenHere = 0, midPoint, xDelta, yDelta, minIndex;
    Bool isCoLinear;

    /* Shut the compiler up */
    dpi = dpi;

    xDelta = w / (cColumnSamples + 1);
    yDelta = h / (cRowSamples + 1);

    /* Move the image pointer down to the right place */
    pImage += (depth == 4 ? x >> 1 : x) + (y * Bpl);

    /* Allocate difference vector for colinear edge handling */
    diffVec = (Float64 *) IPW_CALLOC (sizeof (Float64), MAX(h,w), 
				      cAllocatorData);
    ERR_GEN (!diffVec, cErrNomem);

    /* Fit all of the edges to lines */
    ERR_PASS_CLEANUP( wp_crop_fitHorizontal (pIPWState, edgePoints[cTop], dpi,
					    &edges[cTop][0], &absDev[cTop]));
    ERR_PASS_CLEANUP( wp_crop_fitHorizontal (pIPWState, edgePoints[cBottom],
					    dpi, &edges[cBottom][0],
					    &absDev[cBottom]));
    ERR_PASS_CLEANUP( wp_crop_fitVertical (pIPWState, edgePoints[cLeft], dpi,
					  &edges[cLeft][0],
					  &absDev[cLeft]));
    ERR_PASS_CLEANUP( wp_crop_fitVertical (pIPWState, edgePoints[cRight], dpi,
					  &edges[cRight][0], 
					  &absDev[cRight]));

    *error = 0.0;
    minError = 1.0e+6;
    minIndex = 0;
    for (i = 0; i < cEdges; i++)
    {
	if (absDev[i] < minError)
	{
	    minError = absDev[i];
	    minIndex = i;
	}
	*error += absDev[i];
    }

    /* Compute the angle before we replace any edges. slope is in 
       either -edge[0] or edge[1] depending on orientation.
       Note that we negate the angle, because that's what findSkewSample
       does.. This means that it's not the angle of the line, it's the
       rotation necessary to straighten it.*/
    if (minError == 1.0e+6)
	*angle = 0.0;
    else if (minIndex == cTop || minIndex == cBottom)
    {
	*angle = atan (-edges[minIndex][0]);
    }
    else
    {
	*angle = atan (edges[minIndex][1]);
    }

    /*
     * Check edges for rejection.
     */

    if (absDev[cLeft] > cReasonableErr)
    {
	REPLACE_EDGE(cLeft, 1.0, 0.0, x);
    } 

    if (absDev[cRight] > cReasonableErr)
    {
	REPLACE_EDGE(cRight, 1.0, 0.0, x + w - 1);
    }

    if (absDev[cBottom] > cReasonableErr)
    {
	REPLACE_EDGE(cBottom, 0.0, 1.0, y + h - 1);
    }

    if (absDev[cTop] > cReasonableErr)
    {
	REPLACE_EDGE(cTop, 0.0, 1.0, y);
    }

    /* Create the box. Keep doing it until there are no more colinear sides */
    for (beenHere = 0; beenHere < 3; beenHere++)
    {
	ERR_PASS_CLEANUP( w_geo_intersect (pIPWState, EDGE_COEFS(cTop), 
					 EDGE_COEFS(cLeft), &xI, &yI,
					 &linesParallel));
	theBox[0].x = (Int32) xI;
	theBox[0].y = (Int32) yI;

	ERR_PASS_CLEANUP( w_geo_intersect (pIPWState, EDGE_COEFS(cTop),
					 EDGE_COEFS(cRight), &xI, &yI, 
					 &linesParallel));
	theBox[1].x = (Int32) xI;
	theBox[1].y = (Int32) yI;

	ERR_PASS_CLEANUP( w_geo_intersect (pIPWState, EDGE_COEFS(cBottom),
					 EDGE_COEFS(cLeft), &xI, &yI,
					 &linesParallel));
	theBox[2].x = (Int32) xI;
	theBox[2].y = (Int32) yI;

	ERR_PASS_CLEANUP( w_geo_intersect (pIPWState, EDGE_COEFS(cBottom),
					 EDGE_COEFS(cRight), &xI, &yI,
					 &linesParallel));
	theBox[3].x = (Int32) xI;
	theBox[3].y = (Int32) yI;

	/* Any co-linear edges? */

	ERR_PASS_CLEANUP( w_geo_colinear (pIPWState, EDGE_COEFS(cTop),
					EDGE_COEFS(cBottom), 
					x, x + w, 0, cColinearCutoff, 
					&isCoLinear));

	if (isCoLinear)
	{
	    /*
	     * Compute a vertical slice through the picture, in order try
	     * to determine which side to use.
	     */
	    midPoint = (theBox[1].x - theBox[0].x) / 2;
	    ERR_PASS_CLEANUP( wp_crop_sliceDiffs (pIPWState, pImage, Bpl, depth, midPoint, 
					       h, cVertical, diffVec));

	    /* Take the sum-of-differences on each side of the line */
	    sum0 = 0; sum1 = 0;
	    for (i = 0; i < h; i++)
	    {
		if (i < theBox[0].y)
		    sum0 += diffVec[i];
		else
		    sum1 += diffVec[i];
	    }

	    /* Replace whichever edge has a greater diff per pixel */
	    if (sum0 / theBox[0].y > sum1 / (h - theBox[0].y))
	    {
		REPLACE_EDGE(cTop, 0.0, 1.0, y);
	    }
	    else
	    {
		REPLACE_EDGE(cBottom, 0.0, 1.0, y + h - 1);
	    }

	}
	else
	{
	    ERR_PASS_CLEANUP( w_geo_colinear (pIPWState, EDGE_COEFS(cLeft), 
					    EDGE_COEFS(cRight), 
					    y, y+h, 1, cColinearCutoff, 
					    &isCoLinear));

	    if (isCoLinear)
	    {
		/*
		 * Compute a horizontal slice through the picture, in order 
		 * to attempt to determine which side to use.
		 */
		midPoint = (theBox[1].y - theBox[0].y);
		ERR_PASS_CLEANUP( wp_crop_sliceDiffs (pIPWState, pImage, Bpl, 
						   depth, 
						   midPoint, w, cHorizontal,
						   diffVec));

		/* Take the sum-of-differences on each side of the line */
		sum0 = 0; sum1 = 0;
		for (i = 0; i < w; i++)
		{
		    if (i < theBox[0].x)
			sum0 += diffVec[i];
		    else
			sum1 += diffVec[i];
		}

		/* Replace whichever edge has a greater diff per pixel */
		if (sum0 / theBox[0].x  > sum1 / (w - theBox[0].x))
		{
		    REPLACE_EDGE(cLeft, 1.0, 0.0, x);
		}
		else
		{
		    REPLACE_EDGE(cRight, 1.0, 0.0, x + w - 1);
		}
	    }
	    else 
	    {
		/* Neither direction was colinear -- so we're done */
		break;
	    }
	}
    }

    IPW_FREE (diffVec, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:
    if (diffVec)
	IPW_FREE (diffVec, cAllocatorData);
    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Extract the good edge points from the raw edge information
 *  
 * LongDescription:::
 *	Extract the good edge points from the raw edge information
 *	Also counts the number of rejected points inside the edge
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  edge::           Raw points
 *  len::            Number of points in raw edge
 *
 * Output:::
 *  points::         Number of points in good edge
 *  fitX[]::         The points themselves
 *  fitY[]::         
 *  pRejected::      Number of points rejected
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_getEdge (
    IPWState           *pIPWState,
    ipPoint            *edge,
    Int32               len,
    Int32              *points,
    Int32               fitX[],
    Int32               fitY[],
    Int32              *pRejected)
{
    PROLOGUE;
    register Int32 i;
    Int32 firstFound, lastFound, rejected;

    pIPWState = pIPWState;
    *points = 0;
    firstFound = -1;
    lastFound = -1;

    for (i = 0; i < len; i++)
    {
	if (edge[i].x >= 0 && edge[i].y >= 0)
	{
	    fitX[*points] = edge[i].x; 
	    fitY[*points] = edge[i].y; 
	    (*points)++;
	    if (firstFound == -1)
		firstFound = i;
	    lastFound = i;
	}
    }

    rejected = 0;

    if (firstFound != -1)
    {
	/* Now, count up the rejected points */
	for (i = firstFound; i <= lastFound; i++)
	{
	    if (edge[i].x < 0 || edge[i].y < 0)
		rejected++;
	}
    }

    *pRejected = rejected;
    NORMAL_RETURN;
}

#define GET_EDGE(edge, len, points) \
    points = 0; \
    for (i = 0; i < len; i++) \
	if (edge[i].x >= 0 && edge[i].y >= 0) \
	{ fitX[points] = edge[i].x; fitY[points] = edge[i].y; points++; }


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Fit some points on an allegedly horizontal edge
 *
 * LongDescription:::
 *	Fit some points on an allegedly horizontal edge. Points with
 *	ordinates of < 0 are rejected, as that is the flag used to indicate
 *	an invalid point. Original scheme was fairly fancy but flawed.
 *	What we do now is march along the edge, and fit lines based on
 *	some percentage of valid points, and find the best fit.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  edge::           Edge points
 *  dpi::            DPI of image 
 *
 * Output:::
 *  coefs::          resulting a, b, c for line ax + by = c
 *  dev::            Error term.
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_fitHorizontal (
    IPWState           *pIPWState,
    ipPoint            *edge,
    Int32               dpi,
    Float64            *coefs,
    Float64            *dev)
{
    PROLOGUE;
    Int32 fitX[cRowSamples], fitY[cRowSamples];	
    Float64 fitResults[2], minErr, fitError;
    Int32 nPoints;
    Int32 i, minPoints, minIndex;
    Int32 rejected;

    ERR_PASS( wp_crop_getEdge (pIPWState, edge, cColumnSamples, 
			      (Int32 *) &nPoints, fitX, fitY, &rejected));
#ifdef DEBUG_MSGS 
    fprintf(debug_file, "rejected %ld points!\n", rejected);
#endif
    dpi = dpi;
    minPoints = (nPoints * cMinDataPercent) / 100;
    minPoints = MAX(minPoints, cMinData);

    *dev = 1.0e+09;

    minErr = 1e+12;
    minIndex = -1;

    for (i = 0; i < nPoints - minPoints; i++)
    {

	ERR_PASS( w_linearFit (pIPWState, &fitX[i], &fitY[i], minPoints, 
			     fitResults, &fitError));
	if (fitError < minErr)
	{   
	    minErr = fitError;
	    minIndex = i;
	}
    }

    if (minIndex == -1)
    {
	/* No fit. */
	*dev = 1e+32;
	coefs[0] = 1.0;
	coefs[1] = 1.0;
	coefs[2] = 1.0;
    }
    else
    {
	ERR_PASS( w_linearFit (pIPWState, &fitX[minIndex], &fitY[minIndex], 
			       minPoints, fitResults, &fitError));

	*dev = fitError;
	
	/* convert it to ax + by = c from y = b + mx */
	coefs[0] = -fitResults[0];
	coefs[1] = 1.0;
	coefs[2] = fitResults[1];
    }

#ifdef DEBUG_MSGS
    fprintf(debug_file, "hEdge\t%fx + %fy = %f; err=%f\n", coefs[0], coefs[1], coefs[2], *dev);
#endif
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Fit some points on an allegedly vertical edge
 *
 * LongDescription:::
 *	Fit some points on an allegedly vertical edge. Points with
 *	ordinates of < 0 are rejected, as that is the flag used to indicate
 *	an invalid point. Original scheme was fairly fancy but flawed.
 *	What we do now is march along the edge, and fit lines based on
 *	some percentage of valid points, and find the best fit.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  edge::           Edge points
 *  dpi::            DPI of image 
 *
 * Output:::
 *  coefs::          resulting a, b, c for line ax + by = c
 *  dev::            Error term.
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_crop_fitVertical (
    IPWState           *pIPWState,
    ipPoint            *edge,
    Int32               dpi,
    Float64            *coefs,
    Float64            *dev)
{
    PROLOGUE;
    Int32 fitX[cRowSamples], fitY[cRowSamples];	
    Float64 fitResults[2], minErr, fitError;
    Int32 nPoints;
    Int32 i, minPoints, minIndex;
    Int32 rejected;

    ERR_PASS( wp_crop_getEdge (pIPWState, edge, cColumnSamples, 
			    (Int32 *) &nPoints, 
			    fitX, fitY, &rejected));
#ifdef DEBUG_MSGS 
    fprintf(debug_file, "rejected %ld points!\n", rejected);
#endif

    minPoints = (nPoints * cMinDataPercent) / 100;
    minPoints = MAX(minPoints, cMinData);

    *dev = 1.0e+09;
    minErr = 1e+12;
    minIndex = -1;
    dpi = dpi;

    for (i = 0; i < nPoints - minPoints; i++)
    {
	ERR_PASS( w_linearFit (pIPWState, &fitY[i], &fitX[i], minPoints, 
			     fitResults, &fitError));
	if (fitError < minErr)
	{   
	    minErr = fitError;
	    minIndex = i;
	}
    }


    if (minIndex == -1)
    {
	/* No fit. */
	*dev = 1e+32;
	coefs[0] = 1.0;
	coefs[1] = 1.0;
	coefs[2] = 1.0;
    }
    else
    {
	ERR_PASS( w_linearFit (pIPWState, &fitY[minIndex], &fitX[minIndex], 
			 minPoints, fitResults, &fitError));

	*dev = fitError;

	/* convert it to ax + by = c from y = b + mx */
	coefs[0] = 1.0;
	coefs[1] = -fitResults[0];
	coefs[2] = fitResults[1];
    }
    
#ifdef DEBUG_MSGS
    fprintf(debug_file, "vEdge\t%fx + %fy = %f; err=%f\n", coefs[0], coefs[1], coefs[2], *dev);
#endif
    NORMAL_RETURN;
}

#ifdef DEBUG_MSGS
static void writeEdge (FILE *file, const char *str, ipPoint *points, Int32 count)
{
    Int32 i;
    fprintf(file, "edge %s: [", str);
    for (i = 0; i < count; i++)
    {
	fprintf(file, "%ld,%ld  ", points[i].x, points[i].y);
    }
    fprintf(file, "]\n");
}
#endif

/*

 $Log: cropprv.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.41  1998/01/14 00:11:12  ddavies
 cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.40  1996/07/22 18:39:33  macleod
 cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v

 If there was nothing found on the platen (i.e. a completely white image)
 then an index variable was never checked, causing a read from the
 -1th element of an array and a randomly bad result in the fit.

 Revision 1.39  1996/01/04 20:21:48  ejaquith
 cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.38  1995/12/13  04:05:57  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.37  1995/11/16  19:47:09  macleod
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * The nag script noticed that there were "wp_" routines in crop.pub,
 * or more specifically, not in a ".prv" file, since there wasn't one.
 * Created one, and #included it here.
 *
 * Revision 1.36  1995/11/13  20:10:23  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.35  1995/10/21  02:03:30  ddavies
 * ddavies on Fri Oct 20 19:02:51 PDT 1995
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Change __WATCOMC__ flag that selects file names to __WINDOWS_386__ or
 * _WIN32 so the same conventions apply on all PCs.
 *
 * Revision 1.34  1995/10/16  23:24:22  ejaquith
 * ejaquith on Mon Oct 16 16:23:28 PDT 1995
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * A couple routines used MACRO constants for array indices where it makes
 * more sense to actually insert the value, 4, instead.  (crop boxes).
 *
 * Revision 1.33  1995/10/13  21:59:24  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.32  1995/10/12  17:43:57  macleod
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * added explicit casts (float->int, long->short, long->char, etc) and
 * made floating constants the right size to shut up the NT compiler.
 *
 * Revision 1.31  1995/10/07  01:21:44  ddavies
 * ddavies on Fri Oct 6 18:21:44 PDT 1995
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Include stdio.h since ipwcore.prv no longer includes it (will be true soon).
 *
 * Revision 1.30  1995/10/03  01:59:12  macleod
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.29  1995/09/27  21:30:23  ejaquith
 * ejaquith on Wed Sep 27 14:29:49 PDT 1995
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Added back file as it should have been after last checkin (it somehow
 * all dropped out after a shaky checkin).
 *
 * Revision 1.27  1995/09/11  21:54:07  macleod
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.26  1995/09/07  01:30:16  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.25  1995/09/06  20:18:02  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.24  1995/09/01  21:02:30  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.23  1995/08/31  21:11:44  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.22  1995/08/25  01:35:58  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Fixed warnings
 *
 * Revision 1.21  1995/08/25  01:08:31  jcraig
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * get rid of pc compiler warnings
 *
 * Revision 1.20  1995/08/17  21:54:37  macleod
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.19  1995/02/15  21:12:58  macleod
 * macleod on Wed Feb 15 13:12:44 PST 1995
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Referenced an unused parameter to shut up a compiler
 *
 * Revision 1.18  1995/02/10  20:35:10  macleod
 * macleod on Fri Feb 10 12:33:22 PST 1995
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Changed two things.
 * First, the method used to fit edges was changed quite a bit. It used
 * to attempt to iteratively fit and throw out points, which could get
 * a bad result on a bad initial fit. The new method is to just march
 * along the edge and do fits, and pick the fit with the minimum error.
 * Also changed the algorithm used to do the fitting, because the
 * original algorithm was copyrighted code.
 *
 * Revision 1.17  1995/01/11  22:40:37  dferrell
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Removed includes of shrcmprs.pub and shrip.pub and replaced with shrcc.pub.
 *
 * Revision 1.16  1995/01/11  21:39:48  macleod
 * macleod on Wed Jan 11 13:38:16 PST 1995
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Moving to starting xy in crop_findEdgePoints was incorrect in
 * the 4-bit case (actually, it wasn't done at all, because the
 * case label was 2 instead of 4)
 *
 * Revision 1.15  1994/10/20  22:10:51  macleod
 * macleod on Thu Oct 20 15:10:07 PDT 1994
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * If there were NO points found on an edge, the error wasn't being set.
 * This could cause bad things to happen.
 * (in crop_fitCropBox)
 *
 * Revision 1.14  1994/10/19  00:46:02  macleod
 * macleod on Tue Oct 18 17:45:00 PDT 1994
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Computed angle was different direction depending on whether the "best"
 * edge was horizontal or vertical.
 *
 * Revision 1.13  1994/10/07  19:06:23  lperry
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.12  1994/09/29  00:46:18  lperry
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Eliminate use of utilpros. Use utils.pub instead.
 *
 * Revision 1.11  1994/09/26  21:42:05  macleod
 * macleod on Mon Sep 26 14:40:48 PDT 1994
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * The skew angle computed was backwards from what findSkewSample
 * produces--it was giving the angle of the edge, in radians, rather
 * than the amount to rotate the image to get it straight (which is
 * just -(the angle)). Changed it to negate the angle.
 *
 * Revision 1.10  1994/09/21  21:41:32  macleod
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * crop_fitCropBox computes a skew angle on the edge that has the
 * best fit (lowest error). The angle is computed in radians.
 * Also reduced the threshold above which a difference is considered an edge,
 * and reduced the "throw-away" region from 5% to 0% (it always throws
 * away the first 8 pixels anyway).
 *
 * Revision 1.9  1994/07/27  21:52:10  ejaquith
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Removed #include "iadef.prv"
 *
 * Revision 1.8  1994/07/01  01:44:38  lperry
 * lperry on Thu Jun 30 18:42:22 PDT 1994
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Changed the static variables colDiffs and rowDiffs to
 * auto and allocated their space with malloc.
 *
 * Revision 1.7  1994/06/08  04:07:17  ddavies
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.6  1994/05/05  01:09:27  ddavies
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Change a bunch of %d to %ld's.  Make writeEdge take "const char *"
 * instead of "char *" to make gcc happy.
 *
 * Revision 1.5  1994/04/05  15:16:56  lperry
 * lperry on Tue Apr 5 08:14:28 PDT 1994
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Declared debugging function writeEdge to be static
 * Eliminates nightly messages about diffs in header files
 *
 * Revision 1.4  1994/03/28  22:10:49  macleod
 * macleod on Mon Mar 28 14:09:57 PST 1994
 *
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Fixed a bug in crop where if both pairs of edges were colinear,
 * one of them was not replaced, so the result was a very
 * narrow crop box.
 *
 * Revision 1.3  1994/03/10  02:02:55  macleod
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Took out all of the "low-contrast" edge detection stuff, which never
 * really worked on enough images to call useful.
 *
 * Revision 1.2  1994/03/02  02:48:45  macleod
 * cropprv.c -> /product/ipcore/ipwrap/src/RCS/cropprv.c,v
 *
 * Debugging
 *
 * Revision 1.1  1994/02/09  00:05:38  dferrell
 * Initial revision
 *

 */
