/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  curves.c -- Heuristics on curves (i.e. 1-d functions, e.g. histograms)
 *
 *	w_twinPeaks
 *	w_numericalDiff
 *	w_interpolateMissingCodes
 *	w_partitionCurve
 *	w_findGlobalMaximum
 *	w_findLocalMinimum
 *	w_findPeakExtent
 *
 */




#include "ipwcore.prv"
#include "defines.pub"

/* Prototypes */
#include "curves.pub"


/*#define DEBUG_MSGS */

IP_RCSINFO(RCSInfo, "$RCSfile: curves.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find two peaks of a bimodal distribution
 *
 * LongDescription:::
 * Given a curve of some sort, usually a histogram, find the largest
 * point. Then, look for the local minima left and right of that
 * point, i.e. inflection points in a (perhaps low-pass filtered)
 * search left or right along the curve. Then, search the regions
 * between (0, min1) and (min2, 1) for "global" (within the
 * regions) maxima, and pick the higher one. 
 * <BR><BR>
 *
 * The purpose of this is to take some bimodal distribution, like
 * a histogram of an image with text on a colored background,
 * and find the peak at each point. It might more sense to find
 * the mean rather than the peak. It also might make sense to
 * compute the variance about the two means, in order to
 * have a criterion for deciding whether something really
 * is text-on-tint. However, for this go-round, that's all I"m
 * going to do.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  curve::          Curve to search
 *  points::         # of points in curve
 *  smoothing::      # of points to use when smoothing
 *
 * Output:::
 *  peak1::          First peak found
 *  peak2::          Second peak found
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram analysis
 *
 * Keywords:::
 *  histogram analysis, histogram
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_twinPeaks (
    IPWState           *pIPWState,
    Int32              *curve,
    Int32               points,
    Int32               smoothing,
    Int32              *peak1,
    Int32              *peak2)
{
    PROLOGUE;
    Int32 maxIndex, lMin1, lMin2, gMax1, gMax2;



    /* Find the maximum across the curve */
    ERR_PASS( w_findGlobalMaximum (pIPWState, curve, 0, points - 1, &maxIndex));
    *peak1 = maxIndex;

    /* Find some local minima going in either direction */
    ERR_PASS( w_findLocalMinimum (pIPWState, curve, points, maxIndex, 0, -1, 
				smoothing, &lMin1));
    ERR_PASS( w_findLocalMinimum (pIPWState, curve, points, maxIndex, points - 1,
				1, smoothing, &lMin2));

    /* Find the global maxima in those regions */
    ERR_PASS( w_findGlobalMaximum (pIPWState, curve, 0, lMin1, &gMax1));
    ERR_PASS( w_findGlobalMaximum (pIPWState, curve, lMin2, points - 1, &gMax2));

/*fprintf(stderr, "global max %ld, lmin1 %ld, lmin2 %ld, gmax1 %ld, gmax2 %ld\n", 
	maxIndex, lMin1, lMin2, gMax1, gMax2); */

    ERR_GEN (lMin1 < 0 && lMin2 < 0, cErrHeuristic_foundNothing);

    /* Don't forget about failing searches */
    if (lMin1 < 0 || lMin1 == maxIndex)
    {
	*peak2 = gMax2;
    }
    else if (lMin2 < 0 || lMin2 == maxIndex)
    {
	*peak2 = gMax1;
    }
    else if (curve[gMax1] > curve[gMax2])
    {
	/* The peak to the left is the higher; use it */
	*peak2 = gMax1;
    }
    else
    {
	/* The peak to the right is the higher one; use it */
	*peak2 = gMax2;
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Search a region of a curve for a global maximum
 *
 * LongDescription:::
 *	Search a region of a curve for a global maximum
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  curve::          Curve to search
 *  lower::          Lower index bound
 *  upper::          Upper index bound
 *
 * Output:::
 *  pIndex::         Global maximum inside bounds
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram analysis
 *
 * Keywords:::
 *  histogram analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findGlobalMaximum (
    IPWState           *pIPWState,
    Int32              *curve,
    Int32               lower,
    Int32               upper,
    Int32              *pIndex)
{
    PROLOGUE;
    Int32 maxIndex, i;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    maxIndex = lower;

    for (i = lower; i <= upper; i++)
    {
	if (curve[i] > curve[maxIndex])
	{
	    maxIndex = i;
	}
    }

    *pIndex = maxIndex;
    NORMAL_RETURN;
}
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Directional search down a slope for a local minimum with smoothing.
 *
 * LongDescription:::
 *  Directional search down a slope for a local minimum.
 *  Does smoothing on the numbers as it searches, so that
 *  one can avoid finding very local minima under some noise floor
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  curve::          Pointer to curve
 *  size::           Size of curve
 *  start::          Starting index bound
 *  end::            Ending index bound 
 *  direction::      -1 or 1, direction of search
 *  smoothing::      Number of adjacent numbers by which to smooth
 *
 * Output:::
 *  pIndex::         index into curve of local minimum
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram analysis
 *
 * Keywords:::
 *  histogram analysis
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findLocalMinimum (
    IPWState           *pIPWState,
    Int32              *curve,
    Int32               size,
    Int32               start,
    Int32               end,
    Int32               direction,
    Int32               smoothing,
    Int32              *pIndex)
{
    PROLOGUE;
    Int32 i = 0, j, sum,  smoothOffset, smoothIndex, value, lastValue, 
	  gradient;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    smoothOffset = smoothing / 2;

    if (start == end)
    {
	*pIndex = -1;
    }
    else
    {
	/* Compute the direction of the gradient, + or - */
	gradient = (curve[start] > curve[start + direction]) ? -1 : 1;

	/* n.b.: the check is !=, not >= or <=, so if you pass stupid
	   values, it will go off forever. */
	/* This is not very efficient, however, the curves are usually small */
	lastValue = -1;

	for (i = start; i != end; i += direction)
	{
	    if (smoothing == 1)
	    {
		value = curve[i];
	    }
	    else
	    {
		sum = 0;
		for (j = 0; j < smoothing; j++)
		{
		    smoothIndex = i - smoothOffset + j;
		    if (smoothIndex < 0) smoothIndex = 0;
		    if (smoothIndex >= size) smoothIndex = size - 1;
		    sum += curve[smoothIndex];
		}
		value = sum / smoothing;
	    }
	    if (lastValue != -1)
	    {
		if ((value != lastValue) &&
		    (((value - lastValue) > 0 ? 1 : -1) != gradient))
		    /* The tide has turned */
		    break;
	    }

	    lastValue = value;
	}
    }

    *pIndex = i;
    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Directional search down a slope for a local minimum ignoring some noise.
 *
 * LongDescription:::
 *  Directional search down a slope for a local minimum.
 *  Instead of smoothing, this procedure takes a parameter that indicates
 *  how long a gradient change can be and still be considered noise.  
 *  Gradient changes for less than this extent are ignored.
 *
 *  This procedure assumes that start is the index of a global maximum -
 *  probably found with w_findGlobalMaximum().  The curve is searched
 *  from this maximum in the given direction for the minimum.  Do NOT
 *  call this procedure if this is not the case!
 *
 *  What is noiseFraction exactly?  Well, we are attempting to ignore
 *  noise by ignoring gradient changes that last less than the given
 *  extent.  But, what is noise?  Change by a small amount is probably
 *  noise.  However, change by a large amount is probably NOT noise and
 *  should be considered a legitmate gradient change even if the range
 *  of this change is less than the given extent.  noiseFraction
 *  attempts to give the user some control over this percentage.
 *
 * PrivateDescription:::
 *  See the description of noiseFraction in the LongDescription.  When
 *  does this cause problems in a curve?  Well, say that the gradient
 *  indicates that the curve is decreasing.  Now we get a large spike
 *  that causes the gradient to increase.  If the previous decreases
 *  had been progressing in a nice, reasonably smooth manner, this
 *  spike should probably be considered a legitimate gradient change
 *  and the peak should be said to extend to the previous index even
 *  if it does not fall below the correct threshold. 
 *  There are some other cases that we do NOT check for because they
 *  seem to legitmately happen.  Say that the curve is decreasing and
 *  then it suddently decreases by a large amount and then continues
 *  smoothly decreasing.  In the case the noise threshold is NOT
 *  considered.  Or, say that the curve is decreasing and then 
 *  it suddenly decreases by a large amount.  The amount at the next
 *  index is much larger than the current amount but more in line with
 *  the previous "smooth" decrease of the cureve.  Then, we would like
 *  to ignore this index and continue looking for the peak extent.
 *  This case is NOT considered either although at some point if may
 *  be necessary.
 *
 * Input:::
 *  pIPWState::      	IPWrap state structure
 *  curve::          	Pointer to curve
 *  start::          	Starting index bound.  See note in LongDescription
 *			above.
 *  end::            	Ending index bound 
 *  direction::      	-1 or 1, direction of search
 *  fraction::	     	Must be less than this fraction of the maximum in
 *		     	order to be considered a minimum
 *  range::		Number of positions to ignore as noise if
 *			gradient changes
 *  noiseFraction::	If the current curve amount is less than
 *			this fraction times the previous amount, this is
 *			acceptable noise.  if greater than this fraction
 *			then it should be ignored.  See LongDescription.
 *
 * Output:::
 *  pIndex::         	index into curve of local minimum
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram analysis
 *
 * Keywords:::
 *  histogram analysis
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig::
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_findPeakExtent(
    IPWState           *pIPWState,
    Int32              *curve,
    Int32               start,
    Int32               end,
    Int32               direction,
    Float64             fraction,
    Int32		range,
    Float64		noiseFraction,
    Int32              *pIndex)
{
    PROLOGUE;
    Int32 	 i = -1, lastValue, minIndex;
    Int32	 gradientChange, newGradient, lastGradient, origGradient;
    Int32	 curveThreshold;

    /* NOTE.  may need to put in a change where the lowest value is
     * picked if the gradient is changing frequently.
     * may need to limit the height of the noise from the lowest value
     * that we have seen.
     */

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    if (start != end)
    {
	curveThreshold = (Int32)(curve[start] * fraction);

	/* Compute the direction of the gradient, + or - */
	origGradient = lastGradient = newGradient =
	    (curve[start] >= curve[start + direction]) ? -1 : 1;
	ERR_GEN(origGradient == 1, cErrInvalidParm);
	gradientChange = start;

	/* n.b.: the check is !=, not >= or <=, so if you pass stupid
	   values, it will go off forever. */
	/* This is not very efficient, however, the curves are usually 
	 * small.
	 * assumes that histogram values are never negative.
	 */
	lastValue = curve[start];
	minIndex = start;

	for (i = start + direction; i != end; i += direction)
	{
	    if (curve[i] == lastValue)
	        /* then we are on a plateau.  the gradient has not 
		 * changed so continue on.
		 */
		continue;

	    newGradient = ((curve[i] - lastValue) > 0 ? 1 : -1);
	    minIndex = ((curve[i] <= curve[minIndex]) ? i: minIndex);
	    /* are we decreasing? */
	    if (newGradient == origGradient)
	    {
		/* was this a trend or had we been ignoring noise? */
		if (newGradient == lastGradient)
		{
		    /* have we reached the minimum? */
		    if (curve[i] >= curveThreshold)
		    {
		        lastValue = curve[i];
		        continue;
		    }
		    else
		        break;
		}
		else
		{
		    /* we are at a transition, determine if this was
		     * just noise or a significant hump which we
		     * should have stopped at.
		     */
		    if (abs(i - gradientChange) <= range)
		    {
		        lastValue = curve[i];
		        lastGradient = newGradient;
		        continue;
		    }
		    else
		    {
			i = minIndex;
		        break;
		    }
		}
	    }

	    /* the current gradient is not the same as the original,
	     * is this a new change or are we in the middle of a slight
	     * change in the curve?  
	     * note that we assume that even if we are on a slight reverse
	     * trend the max value of the curve will not exceed the 
	     * value of the curve at the start.
	     */
	    if (newGradient == lastGradient)
	    {
		lastValue = curve[i];
		continue;
	    }
	    else
	    {
	        /* then this is a change.  keep with it for a while. */
		if (curve[i] >= curveThreshold)
		{
		    /* check to see if this is a small enough change to
		     * just be considered noise or whether we should 
		     * stop.
		     */
		    if ((newGradient == 1) && 
			(curve[i] > (lastValue * noiseFraction)))
		    {
			i -= direction;
			break;
		    }
		    else
		    {
		    	gradientChange = i;
		    	lastValue = curve[i];
		    	lastGradient = newGradient;
		    	continue;
		    }
		}
		else
		    break;
	    }
	}

	/* sometimes at the edge of the curve we have problems.  one of
	 * two things may happen.  1) we are in a downward trend but
	 * but have not reached our threshold.  there may or may not
	 * have been an index with a smaller value than our current one.
	 * if there was a smaller one, we would like to use it. (this
	 * could have if we had a slight "noise" upward trend which 
	 * reversed itself but has not yet gotten down to our previous
	 * minimum.  2) we are in an upward trend which was being
	 * ignored as noise.  if so, we want to return the place where
	 * the gradient changed.
	 */
	if (i == end)
	    i = minIndex;
    }

    *pIndex = i;
    NORMAL_RETURN;
}


/* 
 * The following coefficients were generated by savgen.c, which is in 
 * ipcore/tools. They are Savitsky-Golay filter coefficients, which
 * when applied to equally-spaced sampled data, smooth the data, or the
 * derivatives. This is used by numericalDiff to get derivatives.
 * See Numerical Recipes in C, vol. 2, sec 14.8 for more info.
 */

/*
 * I'm going to define three filters, a "low", "mid" and "high"--the "low"
 * won't do much smoothing, the "mid" will do some, and the "high" will do
 * a lot.
 */
#define cSidePointsLow 2
/* 2-point-per-side fit of a 4th order polynomial. For curves with few points */

Float32 SGCoefsLow[3][5][5] =
{
    /* The 0th derivative (the function itself) */
    {
	{ 0.000000F,-0.000001F,1.000002F,-0.000001F,0.000000F },
	{ 1.000000F,-0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.150000F,0.550000F,0.450000F,-0.150000F,0.000000F },
	{ 0.000000F,-0.000000F,1.000000F,0.000000F,0.000000F },
	{ -0.150000F,0.450000F,0.550000F,0.150000F,0.000000F }

    },
    /* Derivative 1 */
    {
	{ -0.200000F,-0.100000F,0.000000F,0.100000F,0.200000F },
	{ -1.500001F,2.000002F,-0.500001F,0.000000F,0.000000F },
	{ -0.550000F,0.150000F,0.350000F,0.050000F,0.000000F },
	{ 0.500001F,-2.000002F,1.500001F,0.000000F,0.000000F },
	{ -0.050000F,-0.350000F,-0.150000F,0.550000F,0.000000F }

    },
    /* Derivative 2 */
    {
	{ 0.142857F,-0.071429F,-0.142857F,-0.071429F,0.142857F },
	{ 0.500000F,-1.000001F,0.500000F,0.000000F,0.000000F },
	{ 0.250000F,-0.250000F,-0.250000F,0.250000F,0.000000F },
	{ 0.500000F,-1.000001F,0.500000F,0.000000F,0.000000F },
	{ 0.250000F,-0.250000F,-0.250000F,0.250000F,0.000000F }

    }
};

#if 0
/* This is a 3-point-per-side fit of a 4th order polynomial. It preserves
   shape quite well, but is less than useful for finding global inflection
   points in higher derivatives */
Float32 SGCoefsLow[3][7][7] =
{
    /* The 0th derivative (the function itself) */
    {
	{ 0.021645F,-0.129870F,0.324675F,0.567099F,0.324675F,-0.129870F,
	  0.021645F },
	{ 0.999997F,0.000005F,-0.000003F,0.000000F,0.000000F,0.000000F,
	  0.000000F },
	{ 0.057143F,0.771429F,0.342857F,-0.228572F,0.057143F,0.000000F,
	  0.000000F },
	{ -0.111111F,0.412698F,0.460317F,0.253968F,0.015873F,-0.031746F,
	  0.000000F },
	{ 0.000000F,-0.000003F,0.000005F,0.999997F,0.000000F,0.000000F,
	  0.000000F },
	{ 0.057143F,-0.228572F,0.342857F,0.771429F,0.057143F,0.000000F,
	  0.000000F },
	{ -0.031746F,0.015873F,0.253968F,0.460317F,0.412698F,-0.111111F,
	  0.000000F }

    },
    /* Derivative 1 */
    {
	{ 0.087302F,-0.265873F,-0.230159F,0.000000F,0.230159F,0.265873F,
	  -0.087302F },
	{ -1.833296F,2.999927F,-1.499956F,0.333313F,0.000000F,0.000000F,
	  0.000000F },
	{ -0.452381F,-0.023810F,0.285714F,0.309524F,-0.119048F,0.000000F,
	  0.000000F },
	{ -0.033069F,-0.346561F,-0.116402F,0.259259F,0.382275F,-0.145503F,
	  0.000000F },
	{ -0.333313F,1.499956F,-2.999927F,1.833296F,0.000000F,0.000000F,
	  0.000000F },
	{ 0.119048F,-0.309524F,-0.285714F,0.023810F,0.452381F,0.000000F,
	  0.000000F },
	{ 0.145503F,-0.382275F,-0.259259F,0.116402F,0.346561F,0.033069F,
	  0.000000F }

    },
    /* Derivative 2 */
    {
	{ 0.059524F,0.000000F,-0.035714F,-0.047619F,-0.035714F,0.000000F,
	  0.059524F },
	{ 0.999962F,-2.499931F,1.999960F,-0.499985F,0.000000F,0.000000F,
	  0.000000F },
	{ 0.392857F,-0.571429F,-0.142857F,0.428571F,-0.107143F,0.000000F,
	  0.000000F },
	{ 0.158730F,-0.115079F,-0.126984F,-0.015873F,0.079365F,0.019841F,
	  0.000000F },
	{ -0.499985F,1.999960F,-2.499931F,0.999962F,0.000000F,0.000000F,
	  0.000000F },
	{ -0.107143F,0.428571F,-0.142857F,-0.571429F,0.392857F,0.000000F,
	  0.000000 },
	{ 0.019841F,0.079365F,-0.015873F,-0.126984F,-0.115079F,0.158730F,
	  0.000000F }

    }
};
#endif

#define cSidePointsMid	6

/* This is a 6-point-per-side fit of a 2nd order polynomial */
Float32 SGCoefsMid[3][13][13] =
{
    /* The 0th derivative (the function itself) */
    {
	{ -0.076923F,-0.000000F,0.062937F,0.111888F,0.146853F,0.167832F,0.174825F,0.167832F,0.146853F,0.111888F,0.062937F,-0.000000F,-0.076923F },
	{ 0.761904F,0.357143F,0.071429F,-0.095238F,-0.142857F,-0.071428F,0.119047F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.375000F,0.279762F,0.196429F,0.125000F,0.065476F,0.017857F,-0.017857F,-0.041667F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.163636F,0.190909F,0.200866F,0.193507F,0.168831F,0.126840F,0.067532F,-0.009091F,-0.103030F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.045455F,0.118182F,0.168182F,0.195455F,0.200000F,0.181818F,0.140909F,0.077273F,-0.009091F,-0.118182F,0.000000F,0.000000F,0.000000F },
	{ -0.020979F,0.064336F,0.128671F,0.172028F,0.194406F,0.195804F,0.176224F,0.135664F,0.074126F,-0.008392F,-0.111888F,0.000000F,0.000000F },
	{ -0.057692F,0.026224F,0.092657F,0.141608F,0.173077F,0.187063F,0.183566F,0.162587F,0.124126F,0.068182F,-0.005245F,-0.096154F,0.000000F },
	{ 0.119047F,-0.071428F,-0.142857F,-0.095238F,0.071429F,0.357143F,0.761904F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.041667F,-0.017857F,0.017857F,0.065476F,0.125000F,0.196429F,0.279762F,0.375000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.103030F,-0.009091F,0.067532F,0.126840F,0.168831F,0.193507F,0.200866F,0.190909F,0.163636F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.118182F,-0.009091F,0.077273F,0.140909F,0.181818F,0.200000F,0.195455F,0.168182F,0.118182F,0.045455F,0.000000F,0.000000F,0.000000F },
	{ -0.111888F,-0.008392F,0.074126F,0.135664F,0.176224F,0.195804F,0.194406F,0.172028F,0.128671F,0.064336F,-0.020979F,0.000000F,0.000000F },
	{ -0.096154F,-0.005245F,0.068182F,0.124126F,0.162587F,0.183566F,0.187063F,0.173077F,0.141608F,0.092657F,0.026224F,-0.057692F,0.000000F }

    },
    /* Derivative 1 */
    {
	{ -0.032967F,-0.027473F,-0.021978F,-0.016484F,-0.010989F,-0.005495F,0.000000F,0.005495F,0.010989F,0.016484F,0.021978F,0.027473F,0.032967F },
	{ -0.464285F,-0.071429F,0.178571F,0.285714F,0.250000F,0.071429F,-0.250000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.291667F,-0.089286F,0.053571F,0.136905F,0.160714F,0.125000F,0.029762F,-0.125000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.187879F,-0.080303F,0.001299F,0.056926F,0.086580F,0.090260F,0.067965F,0.019697F,-0.054545F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.122727F,-0.065152F,-0.018939F,0.015909F,0.039394F,0.051515F,0.052273F,0.041667F,0.019697F,-0.013636F,0.000000F,0.000000F,0.000000F },
	{ -0.080420F,-0.050350F,-0.024942F,-0.004196F,0.011888F,0.023310F,0.030070F,0.032168F,0.029604F,0.022378F,0.010490F,0.000000F,0.000000F },
	{ -0.052198F,-0.037712F,-0.024725F,-0.013237F,-0.003247F,0.005245F,0.012238F,0.017732F,0.021728F,0.024226F,0.025225F,0.024725F,0.000000F },
	{ 0.250000F,-0.071429F,-0.250000F,-0.285714F,-0.178571F,0.071429F,0.464285F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.125000F,-0.029762F,-0.125000F,-0.160714F,-0.136905F,-0.053571F,0.089286F,0.291667F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.054545F,-0.019697F,-0.067965F,-0.090260F,-0.086580F,-0.056926F,-0.001299F,0.080303F,0.187879F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.013636F,-0.019697F,-0.041667F,-0.052273F,-0.051515F,-0.039394F,-0.015909F,0.018939F,0.065152F,0.122727F,0.000000F,0.000000F,0.000000F },
	{ -0.010490F,-0.022378F,-0.029604F,-0.032168F,-0.030070F,-0.023310F,-0.011888F,0.004196F,0.024942F,0.050350F,0.080420F,0.000000F,0.000000F },
	{ -0.024725F,-0.025225F,-0.024226F,-0.021728F,-0.017732F,-0.012238F,-0.005245F,0.003247F,0.013237F,0.024725F,0.037712F,0.052198F,0.000000F }

    },
    /* Derivative 2 */
    {
	{ 0.010989F,0.005495F,0.000999F,-0.002498F,-0.004995F,-0.006494F,-0.006993F,-0.006494F,-0.004995F,-0.002498F,0.000999F,0.005495F,0.010989F },
	{ 0.059524F,-0.000000F,-0.035714F,-0.047619F,-0.035714F,0.000000F,0.059524F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.041667F,0.005952F,-0.017857F,-0.029762F,-0.029762F,-0.017857F,0.005952F,0.041667F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.030303F,0.007576F,-0.008658F,-0.018398F,-0.021645F,-0.018398F,-0.008658F,0.007576F,0.030303F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.022727F,0.007576F,-0.003788F,-0.011364F,-0.015152F,-0.015152F,-0.011364F,-0.003788F,0.007576F,0.022727F,0.000000F,0.000000F,0.000000F },
	{ 0.017483F,0.006993F,-0.001166F,-0.006993F,-0.010490F,-0.011655F,-0.010490F,-0.006993F,-0.001166F,0.006993F,0.017483F,0.000000F,0.000000F },
	{ 0.013736F,0.006244F,0.000250F,-0.004246F,-0.007243F,-0.008741F,-0.008741F,-0.007243F,-0.004246F,0.000250F,0.006244F,0.013736F,0.000000F },
	{ 0.059524F,0.000000F,-0.035714F,-0.047619F,-0.035714F,-0.000000F,0.059524F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.041667F,0.005952F,-0.017857F,-0.029762F,-0.029762F,-0.017857F,0.005952F,0.041667F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.030303F,0.007576F,-0.008658F,-0.018398F,-0.021645F,-0.018398F,-0.008658F,0.007576F,0.030303F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.022727F,0.007576F,-0.003788F,-0.011364F,-0.015152F,-0.015152F,-0.011364F,-0.003788F,0.007576F,0.022727F,0.000000F,0.000000F,0.000000F },
	{ 0.017483F,0.006993F,-0.001166F,-0.006993F,-0.010490F,-0.011655F,-0.010490F,-0.006993F,-0.001166F,0.006993F,0.017483F,0.000000F,0.000000F },
	{ 0.013736F,0.006244F,0.000250F,-0.004246F,-0.007243F,-0.008741F,-0.008741F,-0.007243F,-0.004246F,0.000250F,0.006244F,0.013736F,0.000000F }

    }
};

#if 0
#define cSidePointsHigh 8
/* This is a 8-point-per-side fit of a 3rd order polynomial. */
Float32 SGCoefsHigh[3][17][17] =
{
    /* The 0th derivative (the function itself) */
    {
	{ -0.065015F,-0.018576F,0.021672F,0.055728F,0.083591F,0.105263F,0.120743F,0.130031F,0.133127F,0.130031F,0.120743F,0.105263F,0.083591F,0.055728F,0.021672F,-0.018576F,-0.065015F },
	{ 0.858584F,0.282830F,-0.020200F,-0.121211F,-0.090909F,-0.000002F,0.080805F,0.080806F,-0.070705F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.313287F,0.301632F,0.248018F,0.168764F,0.080186F,-0.001399F,-0.059673F,-0.078321F,-0.041026F,0.068531F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.055945F,0.223776F,0.286713F,0.270396F,0.200466F,0.102564F,0.002331F,-0.074592F,-0.102564F,-0.055944F,0.090909F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.058608F,0.141192F,0.245088F,0.272505F,0.242868F,0.175602F,0.090132F,0.005883F,-0.057720F,-0.081252F,-0.045288F,0.069597F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.101648F,0.076923F,0.184316F,0.232767F,0.234516F,0.201798F,0.146853F,0.081918F,0.019231F,-0.028971F,-0.050450F,-0.032967F,0.035714F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.109244F,0.032805F,0.127343F,0.181260F,0.201446F,0.194791F,0.168185F,0.128519F,0.082682F,0.037565F,0.000059F,-0.022948F,-0.024564F,0.002101F,0.000000F,0.000000F,0.000000F },
	{ -0.100000F,0.005042F,0.081157F,0.131739F,0.160181F,0.169877F,0.164221F,0.146606F,0.120427F,0.089076F,0.055947F,0.024434F,-0.002069F,-0.020168F,-0.026471F,0.000000F,0.000000F },
	{ -0.083591F,-0.010836F,0.046440F,0.089486F,0.119552F,0.137890F,0.145749F,0.144380F,0.135032F,0.118957F,0.097404F,0.071624F,0.042867F,0.012384F,-0.018576F,-0.048762F,0.000000F },
	{ -0.070705F,0.080806F,0.080805F,-0.000002F,-0.090909F,-0.121211F,-0.020200F,0.282830F,0.858584F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.068531F,-0.041026F,-0.078321F,-0.059673F,-0.001399F,0.080186F,0.168764F,0.248018F,0.301632F,0.313287F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.090909F,-0.055944F,-0.102564F,-0.074592F,0.002331F,0.102564F,0.200466F,0.270396F,0.286713F,0.223776F,0.055945F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.069597F,-0.045288F,-0.081252F,-0.057720F,0.005883F,0.090132F,0.175602F,0.242868F,0.272505F,0.245088F,0.141192F,-0.058608F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.035714F,-0.032967F,-0.050450F,-0.028971F,0.019231F,0.081918F,0.146853F,0.201798F,0.234516F,0.232767F,0.184316F,0.076923F,-0.101648F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.002101F,-0.024564F,-0.022948F,0.000059F,0.037565F,0.082682F,0.128519F,0.168185F,0.194791F,0.201446F,0.181260F,0.127343F,0.032805F,-0.109244F,0.000000F,0.000000F,0.000000F },
	{ -0.026471F,-0.020168F,-0.002069F,0.024434F,0.055947F,0.089076F,0.120427F,0.146606F,0.164221F,0.169877F,0.160181F,0.131739F,0.081157F,0.005042F,-0.100000F,0.000000F,0.000000F },
	{ -0.048762F,-0.018576F,0.012384F,0.042867F,0.071624F,0.097404F,0.118957F,0.135032F,0.144380F,0.145749F,0.137890F,0.119552F,0.089486F,0.046440F,-0.010836F,-0.083591F,0.000000F }

    },
    /* Derivative 1 */
    {
	{ 0.032164F,-0.004214F,-0.027649F,-0.039990F,-0.043086F,-0.038786F,-0.028939F,-0.015394F,0.000000F,0.015394F,0.028939F,0.038786F,0.043086F,0.039990F,0.027649F,0.004214F,-0.032164F },
	{ -0.735685F,0.102690F,0.432053F,0.404758F,0.173161F,-0.110385F,-0.293524F,-0.223902F,0.250837F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.393938F,-0.035353F,0.146464F,0.194444F,0.151515F,0.060606F,-0.035353F,-0.093434F,-0.070707F,0.075757F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.203963F,-0.067599F,0.019036F,0.064879F,0.078866F,0.069930F,0.047009F,0.019037F,-0.005050F,-0.016317F,-0.005828F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.096459F,-0.063659F,-0.032523F,-0.004348F,0.019573F,0.037944F,0.049469F,0.052855F,0.046805F,0.030026F,0.001221F,-0.040904F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.035256F,-0.049451F,-0.048701F,-0.036797F,-0.017524F,0.005328F,0.027972F,0.046620F,0.057484F,0.056777F,0.040709F,0.005494F,-0.052656F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.000700F,-0.034395F,-0.049127F,-0.048721F,-0.036997F,-0.017779F,0.005113F,0.027855F,0.046625F,0.057602F,0.056963F,0.040886F,0.005548F,-0.052871F,0.000000F,0.000000F,0.000000F },
	{ 0.018246F,-0.021553F,-0.043217F,-0.049929F,-0.044875F,-0.031237F,-0.012201F,0.009050F,0.029331F,0.045458F,0.054248F,0.052515F,0.037076F,0.004746F,-0.047658F,0.000000F,0.000000F },
	{ 0.027950F,-0.011567F,-0.035395F,-0.046002F,-0.045855F,-0.037422F,-0.023170F,-0.005567F,0.012920F,0.029822F,0.042674F,0.049006F,0.046352F,0.032244F,0.004214F,-0.040205F,0.000000F },
	{ -0.250837F,0.223902F,0.293524F,0.110385F,-0.173161F,-0.404758F,-0.432053F,-0.102690F,0.735685F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.075757F,0.070707F,0.093434F,0.035353F,-0.060606F,-0.151515F,-0.194444F,-0.146464F,0.035353F,0.393938F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.005828F,0.016317F,0.005050F,-0.019037F,-0.047009F,-0.069930F,-0.078866F,-0.064879F,-0.019036F,0.067599F,0.203963F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.040904F,-0.001221F,-0.030026F,-0.046805F,-0.052855F,-0.049469F,-0.037944F,-0.019573F,0.004348F,0.032523F,0.063659F,0.096459F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.052656F,-0.005494F,-0.040709F,-0.056777F,-0.057484F,-0.046620F,-0.027972F,-0.005328F,0.017524F,0.036797F,0.048701F,0.049451F,0.035256F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.052871F,-0.005548F,-0.040886F,-0.056963F,-0.057602F,-0.046625F,-0.027855F,-0.005113F,0.017779F,0.036997F,0.048721F,0.049127F,0.034395F,0.000700F,0.000000F,0.000000F,0.000000F },
	{ 0.047658F,-0.004746F,-0.037076F,-0.052515F,-0.054248F,-0.045458F,-0.029331F,-0.009050F,0.012201F,0.031237F,0.044875F,0.049929F,0.043217F,0.021553F,-0.018246F,0.000000F,0.000000F },
	{ 0.040205F,-0.004214F,-0.032244F,-0.046352F,-0.049006F,-0.042674F,-0.029822F,-0.012920F,0.005567F,0.023170F,0.037422F,0.045855F,0.046002F,0.035395F,0.011567F,-0.027950F,0.000000F }

    },
    /* Derivative 2 */
    {
	{ 0.005160F,0.003225F,0.001548F,0.000129F,-0.001032F,-0.001935F,-0.002580F,-0.002967F,-0.003096F,-0.002967F,-0.002580F,-0.001935F,-0.001032F,0.000129F,0.001548F,0.003225F,0.005160F },
	{ 0.171715F,-0.063130F,-0.139969F,-0.109306F,-0.021645F,0.072509F,0.122653F,0.078282F,-0.111109F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.108391F,-0.020979F,-0.075175F,-0.074592F,-0.039627F,0.009324F,0.051865F,0.067599F,0.036131F,-0.062937F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.069930F,-0.003496F,-0.039627F,-0.047203F,-0.034965F,-0.011655F,0.013986F,0.033217F,0.037296F,0.017482F,-0.034965F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.045788F,0.003330F,-0.020147F,-0.028527F,-0.025697F,-0.015540F,-0.001943F,0.011211F,0.020036F,0.020646F,0.009158F,-0.018315F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.030220F,0.005495F,-0.009491F,-0.016484F,-0.017233F,-0.013487F,-0.006993F,0.000500F,0.007243F,0.011489F,0.011489F,0.005494F,-0.008242F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.019958F,0.005656F,-0.003717F,-0.008932F,-0.010761F,-0.009975F,-0.007346F,-0.003643F,0.000360F,0.003893F,0.006185F,0.006464F,0.003959F,-0.002101F,0.000000F,0.000000F,0.000000F },
	{ 0.013072F,0.005019F,-0.000664F,-0.004291F,-0.006177F,-0.006635F,-0.005979F,-0.004525F,-0.002586F,-0.000476F,0.001490F,0.002999F,0.003735F,0.003385F,0.001634F,0.000000F,0.000000F },
	{ 0.008385F,0.004128F,0.000866F,-0.001500F,-0.003069F,-0.003941F,-0.004214F,-0.003989F,-0.003364F,-0.002438F,-0.001311F,-0.000082F,0.001150F,0.002285F,0.003225F,0.003870F,0.000000F },
	{ -0.111109F,0.078282F,0.122653F,0.072509F,-0.021645F,-0.109306F,-0.139969F,-0.063130F,0.171715F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.062937F,0.036131F,0.067599F,0.051865F,0.009324F,-0.039627F,-0.074592F,-0.075175F,-0.020979F,0.108391F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.034965F,0.017482F,0.037296F,0.033217F,0.013986F,-0.011655F,-0.034965F,-0.047203F,-0.039627F,-0.003496F,0.069930F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.018315F,0.009158F,0.020646F,0.020036F,0.011211F,-0.001943F,-0.015540F,-0.025697F,-0.028527F,-0.020147F,0.003330F,0.045788F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.008242F,0.005494F,0.011489F,0.011489F,0.007243F,0.000500F,-0.006993F,-0.013487F,-0.017233F,-0.016484F,-0.009491F,0.005495F,0.030220F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.002101F,0.003959F,0.006464F,0.006185F,0.003893F,0.000360F,-0.003643F,-0.007346F,-0.009975F,-0.010761F,-0.008932F,-0.003717F,0.005656F,0.019958F,0.000000F,0.000000F,0.000000F },
	{ 0.001634F,0.003385F,0.003735F,0.002999F,0.001490F,-0.000476F,-0.002586F,-0.004525F,-0.005979F,-0.006635F,-0.006177F,-0.004291F,-0.000664F,0.005019F,0.013072F,0.000000F,0.000000F },
	{ 0.003870F,0.003225F,0.002285F,0.001150F,-0.000082F,-0.001311F,-0.002438F,-0.003364F,-0.003989F,-0.004214F,-0.003941F,-0.003069F,-0.001500F,0.000866F,0.004128F,0.008385F,0.000000F }

    }
};
#else
#define cSidePointsHigh 16
/* This is a 16-point-per-side fit of a 2nd order polynomial. It really
   smooths quite a bit */

Float32 SGCoefsHigh[3][33][33] =
{
    /* The 0th derivative (the function itself) */
    {
	{ -0.038961F,-0.025974F,-0.013825F,-0.002514F,0.007960F,0.017595F,0.026393F,0.034353F,0.041475F,0.047759F,0.053205F,0.057813F,0.061584F,0.064516F,0.066611F,0.067868F,0.068287F,0.067868F,0.066611F,0.064516F,0.061584F,0.057813F,0.053205F,0.047759F,0.041475F,0.034353F,0.026393F,0.017595F,0.007960F,-0.002514F,-0.013825F,-0.025974F,-0.038961F },
	{ 0.422085F,0.325077F,0.238390F,0.162023F,0.095975F,0.040248F,-0.005160F,-0.040248F,-0.065016F,-0.079463F,-0.083591F,-0.077399F,-0.060888F,-0.034056F,0.003096F,0.050567F,0.108359F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.315790F,0.254902F,0.199690F,0.150155F,0.106295F,0.068111F,0.035604F,0.008772F,-0.012384F,-0.027864F,-0.037668F,-0.041796F,-0.040248F,-0.033024F,-0.020124F,-0.001548F,0.022704F,0.052632F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.234587F,0.198496F,0.165207F,0.134719F,0.107032F,0.082147F,0.060062F,0.040778F,0.024296F,0.010615F,-0.000265F,-0.008344F,-0.013622F,-0.016099F,-0.015775F,-0.012649F,-0.006723F,0.002005F,0.013534F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.172078F,0.153076F,0.135099F,0.118148F,0.102221F,0.087321F,0.073445F,0.060595F,0.048770F,0.037970F,0.028195F,0.019446F,0.011722F,0.005024F,-0.000649F,-0.005297F,-0.008920F,-0.011517F,-0.013090F,-0.013636F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.123659F,0.116431F,0.109144F,0.101798F,0.094392F,0.086927F,0.079402F,0.071818F,0.064174F,0.056471F,0.048709F,0.040887F,0.033005F,0.025065F,0.017064F,0.009005F,0.000886F,-0.007293F,-0.015531F,-0.023828F,-0.032185F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.085968F,0.086815F,0.086957F,0.086392F,0.085121F,0.083145F,0.080463F,0.077075F,0.072981F,0.068182F,0.062676F,0.056465F,0.049548F,0.041925F,0.033597F,0.024562F,0.014822F,0.004376F,-0.006776F,-0.018634F,-0.031197F,-0.044466F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.056522F,0.062846F,0.068097F,0.072276F,0.075381F,0.077414F,0.078374F,0.078261F,0.077075F,0.074816F,0.071485F,0.067081F,0.061604F,0.055054F,0.047431F,0.038735F,0.028967F,0.018125F,0.006211F,-0.006776F,-0.020836F,-0.035968F,-0.052174F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.033462F,0.043428F,0.052133F,0.059576F,0.065757F,0.070676F,0.074334F,0.076730F,0.077864F,0.077736F,0.076347F,0.073696F,0.069783F,0.064608F,0.058171F,0.050473F,0.041513F,0.031291F,0.019807F,0.007061F,-0.006946F,-0.022215F,-0.038746F,-0.056538F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.015385F,0.027692F,0.038662F,0.048294F,0.056589F,0.063545F,0.069164F,0.073445F,0.076388F,0.077993F,0.078261F,0.077191F,0.074783F,0.071037F,0.065953F,0.059532F,0.051773F,0.042676F,0.032241F,0.020468F,0.007358F,-0.007090F,-0.022876F,-0.040000F,-0.058462F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001221F,0.014945F,0.027326F,0.038364F,0.048059F,0.056410F,0.063419F,0.069084F,0.073407F,0.076386F,0.078022F,0.078315F,0.077265F,0.074872F,0.071136F,0.066056F,0.059634F,0.051868F,0.042759F,0.032308F,0.020513F,0.007375F,-0.007106F,-0.022930F,-0.040098F,-0.058608F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.009852F,0.004631F,0.017810F,0.029683F,0.040251F,0.049514F,0.057471F,0.064124F,0.069471F,0.073513F,0.076249F,0.077681F,0.077807F,0.076628F,0.074144F,0.070355F,0.065260F,0.058861F,0.051156F,0.042146F,0.031830F,0.020210F,0.007284F,-0.006947F,-0.022483F,-0.039325F,-0.057471F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.018473F,-0.003695F,0.009842F,0.022136F,0.033188F,0.042998F,0.051566F,0.058892F,0.064976F,0.069818F,0.073418F,0.075776F,0.076891F,0.076765F,0.075397F,0.072786F,0.068934F,0.063839F,0.057503F,0.049924F,0.041104F,0.031041F,0.019736F,0.007189F,-0.006600F,-0.021631F,-0.037904F,-0.055419F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.025139F,-0.010392F,0.003189F,0.015605F,0.026855F,0.036941F,0.045860F,0.053615F,0.060204F,0.065628F,0.069887F,0.072981F,0.074909F,0.075671F,0.075269F,0.073701F,0.070968F,0.067069F,0.062005F,0.055776F,0.048382F,0.039822F,0.030097F,0.019207F,0.007151F,-0.006070F,-0.020457F,-0.036008F,-0.052725F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.030242F,-0.015754F,-0.002348F,0.009975F,0.021216F,0.031374F,0.040450F,0.048443F,0.055353F,0.061181F,0.065926F,0.069589F,0.072169F,0.073667F,0.074082F,0.073415F,0.071665F,0.068832F,0.064917F,0.059920F,0.053840F,0.046677F,0.038432F,0.029104F,0.018693F,0.007200F,-0.005375F,-0.019033F,-0.033774F,-0.049597F,0.000000F,0.000000F,0.000000F },
	{ -0.034091F,-0.020015F,-0.006937F,0.005142F,0.016223F,0.026304F,0.035388F,0.043473F,0.050559F,0.056646F,0.061735F,0.065826F,0.068917F,0.071011F,0.072105F,0.072201F,0.071299F,0.069398F,0.066498F,0.062600F,0.057703F,0.051808F,0.044914F,0.037021F,0.028130F,0.018240F,0.007352F,-0.004535F,-0.017421F,-0.031305F,-0.046188F,0.000000F,0.000000F },
	{ -0.036932F,-0.023369F,-0.010722F,0.001008F,0.011822F,0.021719F,0.030700F,0.038765F,0.045913F,0.052144F,0.057460F,0.061859F,0.065341F,0.067907F,0.069556F,0.070290F,0.070106F,0.069007F,0.066990F,0.064058F,0.060209F,0.055444F,0.049762F,0.043163F,0.035649F,0.027218F,0.017870F,0.007606F,-0.003574F,-0.015671F,-0.028684F,-0.042614F,0.000000F },
	{ 0.108359F,0.050567F,0.003096F,-0.034056F,-0.060888F,-0.077399F,-0.083591F,-0.079463F,-0.065016F,-0.040248F,-0.005160F,0.040248F,0.095975F,0.162023F,0.238390F,0.325077F,0.422085F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.052632F,0.022704F,-0.001548F,-0.020124F,-0.033024F,-0.040248F,-0.041796F,-0.037668F,-0.027864F,-0.012384F,0.008772F,0.035604F,0.068111F,0.106295F,0.150155F,0.199690F,0.254902F,0.315790F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.013534F,0.002005F,-0.006723F,-0.012649F,-0.015775F,-0.016099F,-0.013622F,-0.008344F,-0.000265F,0.010615F,0.024296F,0.040778F,0.060062F,0.082147F,0.107032F,0.134719F,0.165207F,0.198496F,0.234587F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.013636F,-0.013090F,-0.011517F,-0.008920F,-0.005297F,-0.000649F,0.005024F,0.011722F,0.019446F,0.028195F,0.037970F,0.048770F,0.060595F,0.073445F,0.087321F,0.102221F,0.118148F,0.135099F,0.153076F,0.172078F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.032185F,-0.023828F,-0.015531F,-0.007293F,0.000886F,0.009005F,0.017064F,0.025065F,0.033005F,0.040887F,0.048709F,0.056471F,0.064174F,0.071818F,0.079402F,0.086927F,0.094392F,0.101798F,0.109144F,0.116431F,0.123659F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.044466F,-0.031197F,-0.018634F,-0.006776F,0.004376F,0.014822F,0.024562F,0.033597F,0.041925F,0.049548F,0.056465F,0.062676F,0.068182F,0.072981F,0.077075F,0.080463F,0.083145F,0.085121F,0.086392F,0.086957F,0.086815F,0.085968F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.052174F,-0.035968F,-0.020836F,-0.006776F,0.006211F,0.018125F,0.028967F,0.038735F,0.047431F,0.055054F,0.061604F,0.067081F,0.071485F,0.074816F,0.077075F,0.078261F,0.078374F,0.077414F,0.075381F,0.072276F,0.068097F,0.062846F,0.056522F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.056538F,-0.038746F,-0.022215F,-0.006946F,0.007061F,0.019807F,0.031291F,0.041513F,0.050473F,0.058171F,0.064608F,0.069783F,0.073696F,0.076347F,0.077736F,0.077864F,0.076730F,0.074334F,0.070676F,0.065757F,0.059576F,0.052133F,0.043428F,0.033462F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.058462F,-0.040000F,-0.022876F,-0.007090F,0.007358F,0.020468F,0.032241F,0.042676F,0.051773F,0.059532F,0.065953F,0.071037F,0.074783F,0.077191F,0.078261F,0.077993F,0.076388F,0.073445F,0.069164F,0.063545F,0.056589F,0.048294F,0.038662F,0.027692F,0.015385F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.058608F,-0.040098F,-0.022930F,-0.007106F,0.007375F,0.020513F,0.032308F,0.042759F,0.051868F,0.059634F,0.066056F,0.071136F,0.074872F,0.077265F,0.078315F,0.078022F,0.076386F,0.073407F,0.069084F,0.063419F,0.056410F,0.048059F,0.038364F,0.027326F,0.014945F,0.001221F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.057471F,-0.039325F,-0.022483F,-0.006947F,0.007284F,0.020210F,0.031830F,0.042146F,0.051156F,0.058861F,0.065260F,0.070355F,0.074144F,0.076628F,0.077807F,0.077681F,0.076249F,0.073513F,0.069471F,0.064124F,0.057471F,0.049514F,0.040251F,0.029683F,0.017810F,0.004631F,-0.009852F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.055419F,-0.037904F,-0.021631F,-0.006600F,0.007189F,0.019736F,0.031041F,0.041104F,0.049924F,0.057503F,0.063839F,0.068934F,0.072786F,0.075397F,0.076765F,0.076891F,0.075776F,0.073418F,0.069818F,0.064976F,0.058892F,0.051566F,0.042998F,0.033188F,0.022136F,0.009842F,-0.003695F,-0.018473F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.052725F,-0.036008F,-0.020457F,-0.006070F,0.007151F,0.019207F,0.030097F,0.039822F,0.048382F,0.055776F,0.062005F,0.067069F,0.070968F,0.073701F,0.075269F,0.075671F,0.074909F,0.072981F,0.069887F,0.065628F,0.060204F,0.053615F,0.045860F,0.036941F,0.026855F,0.015605F,0.003189F,-0.010392F,-0.025139F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.049597F,-0.033774F,-0.019033F,-0.005375F,0.007200F,0.018693F,0.029104F,0.038432F,0.046677F,0.053840F,0.059920F,0.064917F,0.068832F,0.071665F,0.073415F,0.074082F,0.073667F,0.072169F,0.069589F,0.065926F,0.061181F,0.055353F,0.048443F,0.040450F,0.031374F,0.021216F,0.009975F,-0.002348F,-0.015754F,-0.030242F,0.000000F,0.000000F,0.000000F },
	{ -0.046188F,-0.031305F,-0.017421F,-0.004535F,0.007352F,0.018240F,0.028130F,0.037021F,0.044914F,0.051808F,0.057703F,0.062600F,0.066498F,0.069398F,0.071299F,0.072201F,0.072105F,0.071011F,0.068917F,0.065826F,0.061735F,0.056646F,0.050559F,0.043473F,0.035388F,0.026304F,0.016223F,0.005142F,-0.006937F,-0.020015F,-0.034091F,0.000000F,0.000000F },
	{ -0.042614F,-0.028684F,-0.015671F,-0.003574F,0.007606F,0.017870F,0.027218F,0.035649F,0.043163F,0.049762F,0.055444F,0.060209F,0.064058F,0.066990F,0.069007F,0.070106F,0.070290F,0.069556F,0.067907F,0.065341F,0.061859F,0.057460F,0.052144F,0.045913F,0.038765F,0.030700F,0.021719F,0.011822F,0.001008F,-0.010722F,-0.023369F,-0.036932F,0.000000F }

    },
    /* Derivative 1 */
    {
	{ -0.005348F,-0.005013F,-0.004679F,-0.004345F,-0.004011F,-0.003676F,-0.003342F,-0.003008F,-0.002674F,-0.002340F,-0.002005F,-0.001671F,-0.001337F,-0.001003F,-0.000668F,-0.000334F,0.000000F,0.000334F,0.000668F,0.001003F,0.001337F,0.001671F,0.002005F,0.002340F,0.002674F,0.003008F,0.003342F,0.003676F,0.004011F,0.004345F,0.004679F,0.005013F,0.005348F },
	{ -0.102167F,-0.068756F,-0.039474F,-0.014319F,0.006708F,0.023607F,0.036378F,0.045021F,0.049536F,0.049923F,0.046182F,0.038313F,0.026316F,0.010191F,-0.010062F,-0.034443F,-0.062952F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.083333F,-0.058050F,-0.035668F,-0.016189F,0.000387F,0.014061F,0.024832F,0.032701F,0.037668F,0.039732F,0.038893F,0.035152F,0.028509F,0.018963F,0.006514F,-0.008836F,-0.027090F,-0.048246F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.068421F,-0.049123F,-0.031889F,-0.016718F,-0.003612F,0.007430F,0.016409F,0.023323F,0.028173F,0.030960F,0.031682F,0.030341F,0.026935F,0.021465F,0.013932F,0.004334F,-0.007327F,-0.021053F,-0.036842F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.056494F,-0.041661F,-0.028309F,-0.016439F,-0.006049F,0.002859F,0.010287F,0.016234F,0.020699F,0.023684F,0.025188F,0.025211F,0.023753F,0.020813F,0.016393F,0.010492F,0.003110F,-0.005753F,-0.016097F,-0.027922F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.046866F,-0.035404F,-0.025011F,-0.015688F,-0.007436F,-0.000253F,0.005860F,0.010904F,0.014877F,0.017781F,0.019614F,0.020378F,0.020072F,0.018696F,0.016250F,0.012734F,0.008149F,0.002493F,-0.004232F,-0.012027F,-0.020892F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.039032F,-0.030138F,-0.022021F,-0.014681F,-0.008117F,-0.002329F,0.002682F,0.006917F,0.010375F,0.013058F,0.014963F,0.016093F,0.016446F,0.016022F,0.014822F,0.012846F,0.010093F,0.006564F,0.002259F,-0.002823F,-0.008682F,-0.015316F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.032609F,-0.025692F,-0.019339F,-0.013552F,-0.008329F,-0.003670F,0.000423F,0.003953F,0.006917F,0.009317F,0.011152F,0.012422F,0.013128F,0.013269F,0.012846F,0.011858F,0.010305F,0.008187F,0.005505F,0.002259F,-0.001553F,-0.005929F,-0.010870F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.027308F,-0.021923F,-0.016949F,-0.012385F,-0.008232F,-0.004489F,-0.001157F,0.001765F,0.004276F,0.006377F,0.008068F,0.009348F,0.010217F,0.010676F,0.010725F,0.010363F,0.009591F,0.008408F,0.006815F,0.004811F,0.002397F,-0.000427F,-0.003662F,-0.007308F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.022906F,-0.018718F,-0.014827F,-0.011234F,-0.007938F,-0.004939F,-0.002237F,0.000167F,0.002274F,0.004084F,0.005596F,0.006812F,0.007729F,0.008350F,0.008673F,0.008699F,0.008428F,0.007860F,0.006994F,0.005831F,0.004370F,0.002612F,0.000557F,-0.001795F,-0.004444F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.019231F,-0.015983F,-0.012949F,-0.010128F,-0.007521F,-0.005128F,-0.002949F,-0.000983F,0.000769F,0.002308F,0.003632F,0.004744F,0.005641F,0.006325F,0.006795F,0.007051F,0.007094F,0.006923F,0.006538F,0.005940F,0.005128F,0.004103F,0.002863F,0.001410F,-0.000256F,-0.002137F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.016147F,-0.013642F,-0.011288F,-0.009086F,-0.007035F,-0.005137F,-0.003389F,-0.001794F,-0.000349F,0.000943F,0.002084F,0.003074F,0.003911F,0.004598F,0.005132F,0.005516F,0.005747F,0.005827F,0.005756F,0.005532F,0.005158F,0.004631F,0.003954F,0.003124F,0.002143F,0.001010F,-0.000274F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.013547F,-0.011631F,-0.009821F,-0.008115F,-0.006516F,-0.005021F,-0.003631F,-0.002347F,-0.001168F,-0.000095F,0.000874F,0.001737F,0.002495F,0.003147F,0.003695F,0.004137F,0.004473F,0.004705F,0.004831F,0.004852F,0.004768F,0.004579F,0.004284F,0.003884F,0.003379F,0.002768F,0.002053F,0.001232F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.011346F,-0.009900F,-0.008524F,-0.007220F,-0.005985F,-0.004822F,-0.003729F,-0.002707F,-0.001755F,-0.000874F,-0.000064F,0.000676F,0.001345F,0.001944F,0.002472F,0.002929F,0.003316F,0.003632F,0.003877F,0.004052F,0.004156F,0.004190F,0.004153F,0.004045F,0.003867F,0.003618F,0.003298F,0.002908F,0.002447F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.009476F,-0.008405F,-0.007379F,-0.006398F,-0.005461F,-0.004570F,-0.003722F,-0.002920F,-0.002162F,-0.001449F,-0.000781F,-0.000157F,0.000422F,0.000956F,0.001446F,0.001891F,0.002291F,0.002647F,0.002958F,0.003224F,0.003445F,0.003622F,0.003754F,0.003842F,0.003884F,0.003882F,0.003836F,0.003744F,0.003608F,0.003427F,0.000000F,0.000000F,0.000000F },
	{ -0.007881F,-0.007111F,-0.006367F,-0.005648F,-0.004954F,-0.004285F,-0.003642F,-0.003024F,-0.002431F,-0.001863F,-0.001321F,-0.000804F,-0.000312F,0.000154F,0.000595F,0.001011F,0.001402F,0.001767F,0.002107F,0.002422F,0.002711F,0.002976F,0.003214F,0.003428F,0.003616F,0.003779F,0.003917F,0.004030F,0.004117F,0.004179F,0.004216F,0.000000F,0.000000F },
	{ -0.006517F,-0.005989F,-0.005472F,-0.004965F,-0.004469F,-0.003984F,-0.003509F,-0.003046F,-0.002593F,-0.002151F,-0.001720F,-0.001299F,-0.000889F,-0.000491F,-0.000102F,0.000275F,0.000641F,0.000997F,0.001342F,0.001677F,0.002000F,0.002313F,0.002614F,0.002906F,0.003186F,0.003455F,0.003714F,0.003962F,0.004199F,0.004426F,0.004641F,0.004846F,0.000000F },
	{ 0.062952F,0.034443F,0.010062F,-0.010191F,-0.026316F,-0.038313F,-0.046182F,-0.049923F,-0.049536F,-0.045021F,-0.036378F,-0.023607F,-0.006708F,0.014319F,0.039474F,0.068756F,0.102167F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.048246F,0.027090F,0.008836F,-0.006514F,-0.018963F,-0.028509F,-0.035152F,-0.038893F,-0.039732F,-0.037668F,-0.032701F,-0.024832F,-0.014061F,-0.000387F,0.016189F,0.035668F,0.058050F,0.083333F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.036842F,0.021053F,0.007327F,-0.004334F,-0.013932F,-0.021465F,-0.026935F,-0.030341F,-0.031682F,-0.030960F,-0.028173F,-0.023323F,-0.016409F,-0.007430F,0.003612F,0.016718F,0.031889F,0.049123F,0.068421F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.027922F,0.016097F,0.005753F,-0.003110F,-0.010492F,-0.016393F,-0.020813F,-0.023753F,-0.025211F,-0.025188F,-0.023684F,-0.020699F,-0.016234F,-0.010287F,-0.002859F,0.006049F,0.016439F,0.028309F,0.041661F,0.056494F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.020892F,0.012027F,0.004232F,-0.002493F,-0.008149F,-0.012734F,-0.016250F,-0.018696F,-0.020072F,-0.020378F,-0.019614F,-0.017781F,-0.014877F,-0.010904F,-0.005860F,0.000253F,0.007436F,0.015688F,0.025011F,0.035404F,0.046866F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.015316F,0.008682F,0.002823F,-0.002259F,-0.006564F,-0.010093F,-0.012846F,-0.014822F,-0.016022F,-0.016446F,-0.016093F,-0.014963F,-0.013058F,-0.010375F,-0.006917F,-0.002682F,0.002329F,0.008117F,0.014681F,0.022021F,0.030138F,0.039032F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.010870F,0.005929F,0.001553F,-0.002259F,-0.005505F,-0.008187F,-0.010305F,-0.011858F,-0.012846F,-0.013269F,-0.013128F,-0.012422F,-0.011152F,-0.009317F,-0.006917F,-0.003953F,-0.000423F,0.003670F,0.008329F,0.013552F,0.019339F,0.025692F,0.032609F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.007308F,0.003662F,0.000427F,-0.002397F,-0.004811F,-0.006815F,-0.008408F,-0.009591F,-0.010363F,-0.010725F,-0.010676F,-0.010217F,-0.009348F,-0.008068F,-0.006377F,-0.004276F,-0.001765F,0.001157F,0.004489F,0.008232F,0.012385F,0.016949F,0.021923F,0.027308F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.004444F,0.001795F,-0.000557F,-0.002612F,-0.004370F,-0.005831F,-0.006994F,-0.007860F,-0.008428F,-0.008699F,-0.008673F,-0.008350F,-0.007729F,-0.006812F,-0.005596F,-0.004084F,-0.002274F,-0.000167F,0.002237F,0.004939F,0.007938F,0.011234F,0.014827F,0.018718F,0.022906F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.002137F,0.000256F,-0.001410F,-0.002863F,-0.004103F,-0.005128F,-0.005940F,-0.006538F,-0.006923F,-0.007094F,-0.007051F,-0.006795F,-0.006325F,-0.005641F,-0.004744F,-0.003632F,-0.002308F,-0.000769F,0.000983F,0.002949F,0.005128F,0.007521F,0.010128F,0.012949F,0.015983F,0.019231F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.000274F,-0.001010F,-0.002143F,-0.003124F,-0.003954F,-0.004631F,-0.005158F,-0.005532F,-0.005756F,-0.005827F,-0.005747F,-0.005516F,-0.005132F,-0.004598F,-0.003911F,-0.003074F,-0.002084F,-0.000943F,0.000349F,0.001794F,0.003389F,0.005137F,0.007035F,0.009086F,0.011288F,0.013642F,0.016147F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.001232F,-0.002053F,-0.002768F,-0.003379F,-0.003884F,-0.004284F,-0.004579F,-0.004768F,-0.004852F,-0.004831F,-0.004705F,-0.004473F,-0.004137F,-0.003695F,-0.003147F,-0.002495F,-0.001737F,-0.000874F,0.000095F,0.001168F,0.002347F,0.003631F,0.005021F,0.006516F,0.008115F,0.009821F,0.011631F,0.013547F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.002447F,-0.002908F,-0.003298F,-0.003618F,-0.003867F,-0.004045F,-0.004153F,-0.004190F,-0.004156F,-0.004052F,-0.003877F,-0.003632F,-0.003316F,-0.002929F,-0.002472F,-0.001944F,-0.001345F,-0.000676F,0.000064F,0.000874F,0.001755F,0.002707F,0.003729F,0.004822F,0.005985F,0.007220F,0.008524F,0.009900F,0.011346F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ -0.003427F,-0.003608F,-0.003744F,-0.003836F,-0.003882F,-0.003884F,-0.003842F,-0.003754F,-0.003622F,-0.003445F,-0.003224F,-0.002958F,-0.002647F,-0.002291F,-0.001891F,-0.001446F,-0.000956F,-0.000422F,0.000157F,0.000781F,0.001449F,0.002162F,0.002920F,0.003722F,0.004570F,0.005461F,0.006398F,0.007379F,0.008405F,0.009476F,0.000000F,0.000000F,0.000000F },
	{ -0.004216F,-0.004179F,-0.004117F,-0.004030F,-0.003917F,-0.003779F,-0.003616F,-0.003428F,-0.003214F,-0.002976F,-0.002711F,-0.002422F,-0.002107F,-0.001767F,-0.001402F,-0.001011F,-0.000595F,-0.000154F,0.000312F,0.000804F,0.001321F,0.001863F,0.002431F,0.003024F,0.003642F,0.004285F,0.004954F,0.005648F,0.006367F,0.007111F,0.007881F,0.000000F,0.000000F },
	{ -0.004846F,-0.004641F,-0.004426F,-0.004199F,-0.003962F,-0.003714F,-0.003455F,-0.003186F,-0.002906F,-0.002614F,-0.002313F,-0.002000F,-0.001677F,-0.001342F,-0.000997F,-0.000641F,-0.000275F,0.000102F,0.000491F,0.000889F,0.001299F,0.001720F,0.002151F,0.002593F,0.003046F,0.003509F,0.003984F,0.004469F,0.004965F,0.005472F,0.005989F,0.006517F,0.000000F }

    },
    /* Derivative 2 */
    {
	{ 0.000764F,0.000621F,0.000487F,0.000362F,0.000246F,0.000140F,0.000043F,-0.000045F,-0.000123F,-0.000193F,-0.000253F,-0.000303F,-0.000345F,-0.000377F,-0.000400F,-0.000414F,-0.000419F,-0.000414F,-0.000400F,-0.000377F,-0.000345F,-0.000303F,-0.000253F,-0.000193F,-0.000123F,-0.000045F,0.000043F,0.000140F,0.000246F,0.000362F,0.000487F,0.000621F,0.000764F },
	{ 0.005160F,0.003225F,0.001548F,0.000129F,-0.001032F,-0.001935F,-0.002580F,-0.002967F,-0.003096F,-0.002967F,-0.002580F,-0.001935F,-0.001032F,0.000129F,0.001548F,0.003225F,0.005160F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.004386F,0.002838F,0.001483F,0.000322F,-0.000645F,-0.001419F,-0.001999F,-0.002386F,-0.002580F,-0.002580F,-0.002386F,-0.001999F,-0.001419F,-0.000645F,0.000322F,0.001483F,0.002838F,0.004386F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.003759F,0.002506F,0.001401F,0.000442F,-0.000369F,-0.001032F,-0.001548F,-0.001917F,-0.002138F,-0.002211F,-0.002138F,-0.001917F,-0.001548F,-0.001032F,-0.000369F,0.000442F,0.001401F,0.002506F,0.003759F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.003247F,0.002221F,0.001310F,0.000513F,-0.000171F,-0.000740F,-0.001196F,-0.001538F,-0.001766F,-0.001880F,-0.001880F,-0.001766F,-0.001538F,-0.001196F,-0.000740F,-0.000171F,0.000513F,0.001310F,0.002221F,0.003247F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.002823F,0.001976F,0.001218F,0.000550F,-0.000030F,-0.000520F,-0.000921F,-0.001233F,-0.001456F,-0.001590F,-0.001635F,-0.001590F,-0.001456F,-0.001233F,-0.000921F,-0.000520F,-0.000030F,0.000550F,0.001218F,0.001976F,0.002823F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.002470F,0.001765F,0.001129F,0.000565F,0.000071F,-0.000353F,-0.000706F,-0.000988F,-0.001200F,-0.001341F,-0.001412F,-0.001412F,-0.001341F,-0.001200F,-0.000988F,-0.000706F,-0.000353F,0.000071F,0.000565F,0.001129F,0.001765F,0.002470F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.002174F,0.001581F,0.001045F,0.000565F,0.000141F,-0.000226F,-0.000536F,-0.000791F,-0.000988F,-0.001129F,-0.001214F,-0.001242F,-0.001214F,-0.001129F,-0.000988F,-0.000791F,-0.000536F,-0.000226F,0.000141F,0.000565F,0.001045F,0.001581F,0.002174F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001923F,0.001421F,0.000965F,0.000555F,0.000190F,-0.000129F,-0.000403F,-0.000631F,-0.000813F,-0.000950F,-0.001041F,-0.001087F,-0.001087F,-0.001041F,-0.000950F,-0.000813F,-0.000631F,-0.000403F,-0.000129F,0.000190F,0.000555F,0.000965F,0.001421F,0.001923F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001709F,0.001282F,0.000892F,0.000539F,0.000223F,-0.000056F,-0.000297F,-0.000502F,-0.000669F,-0.000799F,-0.000892F,-0.000948F,-0.000966F,-0.000948F,-0.000892F,-0.000799F,-0.000669F,-0.000502F,-0.000297F,-0.000056F,0.000223F,0.000539F,0.000892F,0.001282F,0.001709F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001526F,0.001160F,0.000824F,0.000519F,0.000244F,-0.000000F,-0.000214F,-0.000397F,-0.000549F,-0.000672F,-0.000763F,-0.000824F,-0.000855F,-0.000855F,-0.000824F,-0.000763F,-0.000672F,-0.000549F,-0.000397F,-0.000214F,0.000000F,0.000244F,0.000519F,0.000824F,0.001160F,0.001526F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001368F,0.001053F,0.000762F,0.000497F,0.000257F,0.000042F,-0.000147F,-0.000312F,-0.000451F,-0.000564F,-0.000653F,-0.000716F,-0.000754F,-0.000766F,-0.000754F,-0.000716F,-0.000653F,-0.000564F,-0.000451F,-0.000312F,-0.000147F,0.000042F,0.000257F,0.000497F,0.000762F,0.001053F,0.001368F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001232F,0.000958F,0.000705F,0.000474F,0.000263F,0.000074F,-0.000095F,-0.000242F,-0.000368F,-0.000474F,-0.000558F,-0.000621F,-0.000663F,-0.000684F,-0.000684F,-0.000663F,-0.000621F,-0.000558F,-0.000474F,-0.000368F,-0.000242F,-0.000095F,0.000074F,0.000263F,0.000474F,0.000705F,0.000958F,0.001232F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001112F,0.000874F,0.000653F,0.000450F,0.000265F,0.000097F,-0.000053F,-0.000185F,-0.000300F,-0.000397F,-0.000477F,-0.000539F,-0.000583F,-0.000609F,-0.000618F,-0.000609F,-0.000583F,-0.000539F,-0.000477F,-0.000397F,-0.000300F,-0.000185F,-0.000053F,0.000097F,0.000265F,0.000450F,0.000653F,0.000874F,0.001112F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001008F,0.000799F,0.000606F,0.000427F,0.000263F,0.000114F,-0.000020F,-0.000139F,-0.000243F,-0.000333F,-0.000407F,-0.000467F,-0.000511F,-0.000541F,-0.000556F,-0.000556F,-0.000541F,-0.000511F,-0.000467F,-0.000407F,-0.000333F,-0.000243F,-0.000139F,-0.000020F,0.000114F,0.000263F,0.000427F,0.000606F,0.000799F,0.001008F,0.000000F,0.000000F,0.000000F },
	{ 0.000916F,0.000733F,0.000562F,0.000404F,0.000259F,0.000126F,0.000006F,-0.000101F,-0.000196F,-0.000278F,-0.000348F,-0.000404F,-0.000449F,-0.000480F,-0.000499F,-0.000506F,-0.000499F,-0.000480F,-0.000449F,-0.000404F,-0.000348F,-0.000278F,-0.000196F,-0.000101F,0.000006F,0.000126F,0.000259F,0.000404F,0.000562F,0.000733F,0.000916F,0.000000F,0.000000F },
	{ 0.000836F,0.000674F,0.000523F,0.000383F,0.000253F,0.000135F,0.000027F,-0.000070F,-0.000156F,-0.000232F,-0.000296F,-0.000350F,-0.000394F,-0.000426F,-0.000447F,-0.000458F,-0.000458F,-0.000447F,-0.000426F,-0.000394F,-0.000350F,-0.000296F,-0.000232F,-0.000156F,-0.000070F,0.000027F,0.000135F,0.000253F,0.000383F,0.000523F,0.000674F,0.000836F,0.000000F },
	{ 0.005160F,0.003225F,0.001548F,0.000129F,-0.001032F,-0.001935F,-0.002580F,-0.002967F,-0.003096F,-0.002967F,-0.002580F,-0.001935F,-0.001032F,0.000129F,0.001548F,0.003225F,0.005160F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.004386F,0.002838F,0.001483F,0.000322F,-0.000645F,-0.001419F,-0.001999F,-0.002386F,-0.002580F,-0.002580F,-0.002386F,-0.001999F,-0.001419F,-0.000645F,0.000322F,0.001483F,0.002838F,0.004386F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.003759F,0.002506F,0.001401F,0.000442F,-0.000369F,-0.001032F,-0.001548F,-0.001917F,-0.002138F,-0.002211F,-0.002138F,-0.001917F,-0.001548F,-0.001032F,-0.000369F,0.000442F,0.001401F,0.002506F,0.003759F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.003247F,0.002221F,0.001310F,0.000513F,-0.000171F,-0.000740F,-0.001196F,-0.001538F,-0.001766F,-0.001880F,-0.001880F,-0.001766F,-0.001538F,-0.001196F,-0.000740F,-0.000171F,0.000513F,0.001310F,0.002221F,0.003247F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.002823F,0.001976F,0.001218F,0.000550F,-0.000030F,-0.000520F,-0.000921F,-0.001233F,-0.001456F,-0.001590F,-0.001635F,-0.001590F,-0.001456F,-0.001233F,-0.000921F,-0.000520F,-0.000030F,0.000550F,0.001218F,0.001976F,0.002823F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.002470F,0.001765F,0.001129F,0.000565F,0.000071F,-0.000353F,-0.000706F,-0.000988F,-0.001200F,-0.001341F,-0.001412F,-0.001412F,-0.001341F,-0.001200F,-0.000988F,-0.000706F,-0.000353F,0.000071F,0.000565F,0.001129F,0.001765F,0.002470F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.002174F,0.001581F,0.001045F,0.000565F,0.000141F,-0.000226F,-0.000536F,-0.000791F,-0.000988F,-0.001129F,-0.001214F,-0.001242F,-0.001214F,-0.001129F,-0.000988F,-0.000791F,-0.000536F,-0.000226F,0.000141F,0.000565F,0.001045F,0.001581F,0.002174F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001923F,0.001421F,0.000965F,0.000555F,0.000190F,-0.000129F,-0.000403F,-0.000631F,-0.000813F,-0.000950F,-0.001041F,-0.001087F,-0.001087F,-0.001041F,-0.000950F,-0.000813F,-0.000631F,-0.000403F,-0.000129F,0.000190F,0.000555F,0.000965F,0.001421F,0.001923F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001709F,0.001282F,0.000892F,0.000539F,0.000223F,-0.000056F,-0.000297F,-0.000502F,-0.000669F,-0.000799F,-0.000892F,-0.000948F,-0.000966F,-0.000948F,-0.000892F,-0.000799F,-0.000669F,-0.000502F,-0.000297F,-0.000056F,0.000223F,0.000539F,0.000892F,0.001282F,0.001709F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001526F,0.001160F,0.000824F,0.000519F,0.000244F,0.000000F,-0.000214F,-0.000397F,-0.000549F,-0.000672F,-0.000763F,-0.000824F,-0.000855F,-0.000855F,-0.000824F,-0.000763F,-0.000672F,-0.000549F,-0.000397F,-0.000214F,-0.000000F,0.000244F,0.000519F,0.000824F,0.001160F,0.001526F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001368F,0.001053F,0.000762F,0.000497F,0.000257F,0.000042F,-0.000147F,-0.000312F,-0.000451F,-0.000564F,-0.000653F,-0.000716F,-0.000754F,-0.000766F,-0.000754F,-0.000716F,-0.000653F,-0.000564F,-0.000451F,-0.000312F,-0.000147F,0.000042F,0.000257F,0.000497F,0.000762F,0.001053F,0.001368F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001232F,0.000958F,0.000705F,0.000474F,0.000263F,0.000074F,-0.000095F,-0.000242F,-0.000368F,-0.000474F,-0.000558F,-0.000621F,-0.000663F,-0.000684F,-0.000684F,-0.000663F,-0.000621F,-0.000558F,-0.000474F,-0.000368F,-0.000242F,-0.000095F,0.000074F,0.000263F,0.000474F,0.000705F,0.000958F,0.001232F,0.000000F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001112F,0.000874F,0.000653F,0.000450F,0.000265F,0.000097F,-0.000053F,-0.000185F,-0.000300F,-0.000397F,-0.000477F,-0.000539F,-0.000583F,-0.000609F,-0.000618F,-0.000609F,-0.000583F,-0.000539F,-0.000477F,-0.000397F,-0.000300F,-0.000185F,-0.000053F,0.000097F,0.000265F,0.000450F,0.000653F,0.000874F,0.001112F,0.000000F,0.000000F,0.000000F,0.000000F },
	{ 0.001008F,0.000799F,0.000606F,0.000427F,0.000263F,0.000114F,-0.000020F,-0.000139F,-0.000243F,-0.000333F,-0.000407F,-0.000467F,-0.000511F,-0.000541F,-0.000556F,-0.000556F,-0.000541F,-0.000511F,-0.000467F,-0.000407F,-0.000333F,-0.000243F,-0.000139F,-0.000020F,0.000114F,0.000263F,0.000427F,0.000606F,0.000799F,0.001008F,0.000000F,0.000000F,0.000000F },
	{ 0.000916F,0.000733F,0.000562F,0.000404F,0.000259F,0.000126F,0.000006F,-0.000101F,-0.000196F,-0.000278F,-0.000348F,-0.000404F,-0.000449F,-0.000480F,-0.000499F,-0.000506F,-0.000499F,-0.000480F,-0.000449F,-0.000404F,-0.000348F,-0.000278F,-0.000196F,-0.000101F,0.000006F,0.000126F,0.000259F,0.000404F,0.000562F,0.000733F,0.000916F,0.000000F,0.000000F },
	{ 0.000836F,0.000674F,0.000523F,0.000383F,0.000253F,0.000135F,0.000027F,-0.000070F,-0.000156F,-0.000232F,-0.000296F,-0.000350F,-0.000394F,-0.000426F,-0.000447F,-0.000458F,-0.000458F,-0.000447F,-0.000426F,-0.000394F,-0.000350F,-0.000296F,-0.000232F,-0.000156F,-0.000070F,0.000027F,0.000135F,0.000253F,0.000383F,0.000523F,0.000674F,0.000836F,0.000000F }

    }
};
#endif

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Numerical differentiation
 *
 * LongDescription:::
 *  Numerical differentiation -- computes a smoothed  1st or 2nd derivative 
 *  of a curve. Incidentally, if you ask for the 0th derivative, you get 
 *  a smoothed version of the curve itself.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  curve::          The curve in question 
 *  points::         # of points along the curve
 *  derivative::     Which derivative. If you pass 0, you just get
 *		     a smoothed curve.
 *  smoothing::      0 = low smoothing, 
 *		     1 = medium smooth, 
 *		     2 = smooth a lot
 *
 * Output:::
 *  ppDCurve::       The differentiated curve
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_numericalDiff (
    IPWState           *pIPWState,
    Int32              *curve,
    Int32               points,
    Int32               derivative,
    Int32               smoothing,
    Int32             **ppDCurve)
{
    PROLOGUE;
    register Float32 *fCurve, sum, *filter;
    register Int32    j, i, convPoints, sidePoints, offset, *dcurve = NULL;
    Int32	      filterIndex;
    Int32 spLUT[] = { cSidePointsLow, cSidePointsMid, cSidePointsHigh };

    ERR_GEN (smoothing < 0 || smoothing > 2, cErrInvalidParm);
    ERR_GEN (derivative < 0 || derivative > 2, cErrInvalidParm);

    if (*ppDCurve == NULL)
    {
	dcurve = IPW_CALLOC (points, sizeof (Int32), cAllocatorData);
	ERR_GEN (!dcurve, cErrNomem);
	*ppDCurve = dcurve;
    }
    else
    {
	dcurve = *ppDCurve;
    }

    sidePoints = spLUT[smoothing];
    ERR_GEN (points < sidePoints, cErrHeuristic_notEnoughPoints);

    /* Allocate a floating-point version */
    fCurve = (Float32 *) IPW_CALLOC (points, sizeof(Float32), cAllocatorData);
    ERR_GEN (!fCurve, cErrNomem);

    for (i = 0; i < points; i++)
    {
	fCurve[i] = (Float32) curve[i];
    }
 
    /* Convolve each point */
    for (i = 0; i < points; i++)
    {
	if (i < sidePoints + 1)
	{
	    /* Not enough left points, so use a left-skewed one */
	    convPoints = i + sidePoints + 1;
	    offset = i;
	    filterIndex = i + 1;
	}
	else if (points - i <= sidePoints)
	{
	    /* Not enough right-points */
	    convPoints = (points - i - 1) + sidePoints + 1;
	    offset = sidePoints;
	    filterIndex = (points - i - 1) + sidePoints + 1;
	}
	else
	{
	    convPoints = 2 * sidePoints + 1;
	    filterIndex = 0;
	    offset = sidePoints;
	}

	switch (smoothing)	/* the default case is 2.  We use default
				 * to keep gcc from whining that filter
				 * might not be set. */
	{
	    case 0: filter = &(SGCoefsLow[derivative][filterIndex][0]); break;
	    case 1: filter = &(SGCoefsMid[derivative][filterIndex][0]); break;
	    default: filter = &(SGCoefsHigh[derivative][filterIndex][0]); break;
	}

	/* Do the convolve */
	sum = 0.0F;
	for (j = 0; j < convPoints; j++)
	{
	    sum += fCurve[i - offset + j] * filter[j];
	}

	/* Stuff the result */
	dcurve[i] = (Int32) sum;
    }

    IPW_FREE (fCurve, cAllocatorData);

    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Interpolate missing codes (zeroes) in a curve, such as a histogram
 *
 * LongDescription:::
 *  Scanned image histograms have a tendency to have a lot of
 *  missing codes, i.e. a bunch of zeroes spaced along an otherwise
 *  fairly smooth curve.
 *  This routine goes through a histogram and interpolates
 *  between the spikes.
 *  
 *  What results is no longer a "valid" histogram, i.e.
 *  it doesn't count the codes in an image. If you look at
 *  some histograms from e.g. the HP scanner, you will notice
 *  that the points around the missing codes are raised a bit.
 *  So to make it valid, we would have to subtract some surrounding
 *  energy from the neighborhood. However, it appears to be more
 *  than just one or two points, so we'd have to figure out
 *  how wide to make it, which might be a good thing to do
 *  sometime, since the histograms are significantly wiggly around
 *  the holes.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  curve::          Curve to interpolate, such as a histogram
 *  points::         Number of points in the curve
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram analysis
 *
 * Keywords:::
 *  histogram, histogram analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::u
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_interpolateMissingCodes (
    IPWState           *pIPWState,
    Int32              *curve,
    Int32               points)
{
    PROLOGUE;
    Int32 i, j, k, previous, next, interval;
    Bool seenData;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    previous = 0;
    seenData = cFalse;

    for (i = 0; i < points; i++)
    {
	if (curve[i] == 0)
	{
	    /* We don't start interpolating until we have seen non-zero data */
	    if (seenData == cFalse)
		continue;

	    /* Search forwards for the next non-zero value */
	    for (j = i; j < points; j++)
	    {
		if (curve[j])
		    break;
	    }

	    /* If that was the last non-zero point, don't interpolate */
	    if (j == points)
		continue;

	    /* Linearly interpolate those values. */
	    interval = j - i;
	    next = curve[j];
	    for (k = 0; k < interval; k++)
		curve[k + i] = (next * k + previous * (interval - k) ) / interval;

	}
	else
	{
	    previous = curve[i];
	    seenData = cTrue;
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Partition a curve into regions
 *
 * LongDescription:::
 *  The purpose of this routine is to divide a curve up into
 *  regions of bumps. The most common use is to divide a 
 *  histogram into regions for thresholding or other analysis.
 *  Our method is to take a smoothed second derivative of
 *  the curve, and divide the curve by finding each "valley" edge
 *  in the curve, i.e. peaks in the second derivative. 
 *  <BR><BR>
 *
 *  A very smoothed second derivative will have positive and
 *  negative lumps. The right place to pick for subdivision
 *  is probably the peaks of the positive lumps.
 *  Well, not really, because the peak of the positive (accelerating)
 *  portion is not necessarily the bottom of the valley. 
 *  A hack for now will be to take the midpoint between the
 *  zero-crossing between decel/accel and the peak accel point.
 *  A better solution might be to search for the valley between
 *  these points, but that cause some pathologies.
 *  <BR><BR>
 *
 *  NOTE that the histogram of most scanners will reveal a shocking
 *  number of missing codes. Therefore, it is advisable to 
 *  pass such curves to w_interpolateMissingCodes before passing it
 *  here.
 *  <BR><BR>
 *
 *  Once the curve is divided, a pair of arrays of integers is 
 *  computed, which contain the starting and ending points for
 *  each region (rendundant, perhaps, but hopefully convenient).
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  curve::          The curve to analyze
 *  nPoints::        Number of points in the curve
 *
 * Output:::
 *  nRegions::       Number of regions found
 *  startIndex::     Starting points for each region
 *  endIndex::       Ending points for each region
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This routine is somewhat squirrely. There are some heuristic constants
 *  that should really be arguments.
 *
 * FunctionCategory:::
 *  histogram analysis
 *
 * Keywords:::
 *  histogram analysis, math
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_partitionCurve (
    IPWState           *pIPWState,
    Int32              *curve,
    Int32               nPoints,
    Int32              *nRegions,
    Int32             **startIndex,
    Int32             **endIndex)
{
    PROLOGUE;
    Int32    i, state, peaks1, s, e,  peaks, highestValue,
	     highestIndex, first, last, j,  histoSum, partialSum,
	     firstIndex = 0,		/* keep gcc happy */
	    *ddx = NULL, 
	    *tmpPeaks = NULL, 
	    *regionPeaks = NULL, 
	    *filteredPeaks = NULL;
/* Float64 meanLowest; */


    *startIndex = NULL;
    *endIndex = NULL;

    /* Allocate a curve for the second derivative */
    ddx = (Int32 *) IPW_CALLOC(nPoints, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (!ddx, cErrNomem);

    /* Allocate a curve for the peak values in each peak region */
    regionPeaks = (Int32 *) IPW_CALLOC(nPoints, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (!regionPeaks, cErrNomem);

    /* Allocate temp curve for the found peaks */
    tmpPeaks = (Int32 *) IPW_CALLOC(nPoints, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (!tmpPeaks, cErrNomem);

    /* Compute the second derivative */
    ERR_PASS_CLEANUP( w_numericalDiff (pIPWState, curve, nPoints, 2,
				     (nPoints <= 32) ? 0 : 2, &ddx));

    /* Find the peaks */
    state = 0;
    peaks = 0;
    highestIndex = 0;
    highestValue = -1; 

    for (i = 0; i < nPoints; i++)
    {
	switch (state)
	{
	    case 0:
		/* Initial state: pick the first direction we see */
		if (ddx[i])
		{
		    state = (ddx[i] > 0) ? 1 : -1;
		    if (state == 1)
		    {
			firstIndex = i;
			highestIndex = i;
			highestValue = ddx[i];
		    }
		}
		break;
	
	    case 1:
		if (ddx[i] < 0)
		{
		    /* +- transition: there was a peak back there */
		    state = -1;
#if 1	/* just use the highest. */
		    if (nPoints > 32)
		    {
#if 0
			/* find lowest point in between this and the next */
			foundIndex = 0;
			foundValue = 0x7ffffffe;
			for (j = firstIndex; j < /*nPoints */ i && ddx[j] >= 0; j++)
			{
			    if (curve[j] < foundValue)
			    {
				foundIndex = j;
				foundValue = curve[j];
			    }
			}
			tmpPeaks[peaks++] = foundIndex;
#else
			/* Another attempt: sum from firstIndex to highestIndex along
			 * ddx, and then move along to a percentage (30%?) of that,
			 * the hope being we find the place where it kind of starts
			 * to curve up.
			 */
			histoSum = 0;
			for (j = firstIndex; j <= highestIndex; j++)
			    histoSum += ddx[j];
			/* Make it 20% just for a first try */
			histoSum = histoSum / 5;
			partialSum = 0;
			for (j = firstIndex; j <= highestIndex; j++)
			{
			    partialSum += ddx[j];
			    if (partialSum > histoSum)
				break;
			}
			tmpPeaks[peaks++] = j;
#endif
		    }
		    else
#endif
			tmpPeaks[peaks++] = highestIndex;
		    highestValue = -1;
		    highestIndex = i;
		}
		else
		{
		    if (ddx[i] > highestValue)
		    {
			highestValue = ddx[i];
			highestIndex = i;
		    }
		}
		break;

	    case -1:
		if (ddx[i] > 0)
		{
		    state = 1;
		    firstIndex = i;
		    highestValue = ddx[i];
		    highestIndex = i;
		}
		break;
	}
    }

    /* If we were left in a positive state, record peak */
    if (state == 1)
	tmpPeaks[peaks++] = highestIndex;

    /* Don't need the 2nd derivative anymore */
    IPW_FREE (ddx, cAllocatorData);
    ddx = NULL;

    /* Allocate a curve for the filtered peaks */
    filteredPeaks = (Int32 *) IPW_CALLOC(peaks, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (!filteredPeaks, cErrNomem);

    for (i = tmpPeaks[0] ? -1 : 0; i < peaks; i++)
    {
	Int32 regionPeak = 0;

	for (j = (i == -1 ? 0 : tmpPeaks[i]); 
	     j < ((i == peaks - 1) ? nPoints - 1: tmpPeaks[i+1]);
	     j++)
	{
	    if (curve[j] > regionPeak)
		regionPeak = curve[j];
	}

	if (i >= 0)
	    regionPeaks[i] = regionPeak;
    }

    /* Filter out peaks which are too far above the mean */
    /*
     * Well, no, let's try a better filtering strategy, since
     * it is now the largest cause of bad thresholds.
     * What we want to detect is regions where one end is
     * significantly higher than the other, and the region
     * next to the higher end has a peak higher than the first
     * region's end, i.e. that the first region is part of a
     * "lump" with the second region.
     */
    for (i = 0, peaks1 = 0; i < peaks; i++)
    {
#ifdef DEBUG_MSGS
fprintf(stderr, "region %d: peak i %d curve[peak i] %d + %d,%d\n", i, tmpPeaks[i], curve[tmpPeaks[i]], tmpPeaks[i+1], curve[tmpPeaks[i+1]]);
#endif
	if (i == 0 || i == peaks - 1 || (Float32) (tmpPeaks[i] - tmpPeaks[i-1]) / nPoints > 0.3)
	{
	    filteredPeaks[peaks1++] = tmpPeaks[i];
	}
	else 
	{
	    Float32 p1, p2, p3;

	    p1 = (Float32) curve[tmpPeaks[i]] / 
		 (Float32) curve[tmpPeaks[i-1]];
	    p2 = (Float32) regionPeaks[i] / 
		 (Float32) curve[tmpPeaks[i]];
	    p3 = (Float32) regionPeaks[i] /
		 (Float32) regionPeaks[i-1];
#ifdef DEBUG_MSGS
fprintf(stderr, "^ p1 %f, p2 %f p3 %f\n", p1, p2, p3);
#endif
	    if (p1 > 1.5 && p2 > 1.0 && p3 >  0.8 /* 1.2 */)
	    {
#ifdef DEBUG_MSGS
fprintf(stderr, "region %d rejected [upwards]\n", i);
#endif
	    }
	    else
	    {
		p1 = (Float32) curve[tmpPeaks[i]] / 
		    (Float32) curve[tmpPeaks[i + 1]];
		p2 = (Float32) regionPeaks[i - 1] / 
		    (Float32) curve[tmpPeaks[i]];
	    p3 = (Float32) regionPeaks[i-1] /
		 (Float32) regionPeaks[i];
#ifdef DEBUG_MSGS
fprintf(stderr, "v p1 %f, p2 %f p3 %f\n", p1, p2, p3);
#endif

		if (!(p1 > 1.5 && p2 > 1.0 && p3 > 0.8 /* 1.2 */))
		{
		    filteredPeaks[peaks1++] = tmpPeaks[i];
		}
		else
		{
#ifdef DEBUG_MSGS
fprintf(stderr, "region %d rejected [downwards]\n", i);
#endif
		}
	    }
	}
    }

    IPW_FREE (tmpPeaks, cAllocatorData);
    IPW_FREE (regionPeaks, cAllocatorData);
    peaks = peaks1;

    /* Make the regions. Be careful about peaks at start or end */
    first = (filteredPeaks[0] == 0) ? 1 : 0;
    last = peaks - ((filteredPeaks[peaks - 1] == nPoints - 1) ? 2 : 1);

    *nRegions = (last - first + 1) + 1 ;
    *startIndex = (Int32 *) IPW_CALLOC(*nRegions, sizeof(Int32), 
				       cAllocatorData);
    ERR_GEN_CLEANUP (*startIndex == NULL, cErrNomem);

    *endIndex = (Int32 *) IPW_CALLOC(*nRegions, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (*endIndex == NULL, cErrNomem);

    /* Get the points into regions */
    for (i = first; i <= last + 1; i++)
    {
	s = (*startIndex)[i - first] = (i == first) ? 0 : filteredPeaks[i - 1];
	e = (*endIndex)[i - first] = (i == last + 1) ? nPoints - 1 : 
	                                               filteredPeaks[i] - 1;
    }

    IPW_FREE (filteredPeaks, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (filteredPeaks) IPW_FREE (filteredPeaks, cAllocatorData);
    if (tmpPeaks) IPW_FREE (tmpPeaks, cAllocatorData);
    if (regionPeaks) IPW_FREE (regionPeaks, cAllocatorData);
    if (ddx) IPW_FREE (ddx, cAllocatorData);
    if (*startIndex) IPW_FREE (*startIndex, cAllocatorData);
    if (*endIndex) IPW_FREE (*endIndex, cAllocatorData);
    ERR_RETURN;
}
   

/*
 * $Log: curves.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 * Imported sources
 *
 * Revision 1.37  1998/07/28 21:50:55  aber
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Warning fixes for gcc v2.8x.
 *
 * Revision 1.36  1997/06/17 20:16:47  jcraig
 * jcraig on Tue Jun 17 13:13:34 PDT 1997
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * fix compiler warning from pc - (used = instead of == )
 *
 * Revision 1.35  1997/06/17 01:42:17  jcraig
 * jcraig on Mon Jun 16 18:40:23 PDT 1997
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * w_findPeakExtent now takes a noiseFraction parameter to describe how
 * much change can be considered noise.  more than this fraction is
 * not noise.
 *
 * Revision 1.34  1997/05/21 16:41:55  jcraig
 * jcraig on Wed May 21 09:41:32 PDT 1997
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * remove unused parameter size from w_findPeakExtent
 *
 * Revision 1.33  1997/05/21 01:59:15  jcraig
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * add new procedure w_findPeakExtent which given the index of the
 * global maximum attempts to find the extent of the peak.  instead
 * of smoothing, the procedure takes a parameter that indicates how
 * long the gradient can change and still be considered noise.  it
 * declares a minimum when a index with the given percent of the maximum
 * is found.
 *
 * Revision 1.32  1996/01/04 20:21:52  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.31  1995/12/13  04:06:01  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.30  1995/11/13  20:10:27  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.29  1995/10/16  22:54:50  dferrell
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.28  1995/10/13  21:59:28  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.27  1995/10/12  01:20:41  ddavies
 * ddavies on Wed Oct 11 18:19:54 PDT 1995
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Append an F to all floating constants so Microsoft's Visual C++ won't whine
 * about conversions from double to float.
 *
 * Revision 1.26  1995/10/03  21:13:04  macleod
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.25  1995/09/21  22:52:19  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.24  1995/09/11  21:56:08  macleod
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.23  1995/09/07  01:30:20  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.22  1995/09/06  20:18:06  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.21  1995/09/01  20:44:24  macleod
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Added back #include files
 *
 * Revision 1.20  1995/08/31  21:11:47  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.19  1995/08/29  21:53:48  ddavies
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.18  1995/08/25  01:36:03  ejaquith
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Fixed warnings
 *
 * Revision 1.17  1995/08/17  21:54:42  macleod
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.16  1994/10/06  20:53:44  lperry
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Revision 1.15  1994/09/29  00:46:44  lperry
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Eliminate use of utilpros. Use utils.pub instead.
 *
 * Revision 1.14  1994/09/27  00:21:41  macleod
 * macleod on Mon Sep 26 17:20:56 PDT 1994
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Changed static coefficient table from 64 bit to 32 bit floating point,
 * hopefully saving some memory
 *
 * Revision 1.13  1994/09/16  17:34:53  jcraig
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * remove include of bitacces.h which was not needed.
 *
 * Revision 1.12  1994/08/16  06:40:37  macleod
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Simplified twinPeaks, which I ended up not using anyway.
 * (Does anyone call this? It doesn't work very well, so I'll probably
 * take it out)
 *
 * Revision 1.11  1994/05/05  01:10:55  ddavies
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Give twinPeaks a procName and use it in the abortI inside twinPeaks.
 * Generally add default cases to switches and initialize vars to make
 * gcc happy.
 *
 * Revision 1.10  1994/03/17  03:01:19  macleod
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Fixed a memory leak
 *
 * Revision 1.9  1994/03/02  02:49:03  macleod
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * I think I just tuned a couple of parameters; other than that, I don't
 * remember what I changed
 *
 * Revision 1.8  1994/02/18  05:20:58  macleod
 * macleod on Thu Feb 17 21:20:25 PST 1994
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Fixes to partitionCurve. Nothing spectacular, but the threshold finder
 * was changed to be a bit more robust, so it isn't quite as critical.
 *
 * Revision 1.7  1994/02/08  23:35:19  macleod
 * macleod on Tue Feb 8 15:34:47 PST 1994
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Better scheme for filtering regions which are really part of larger
 * lumps in partitionCurve
 *
 * Revision 1.6  1994/02/03  19:53:01  macleod
 * macleod on Thu Feb 3 11:52:36 PST 1994
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Made some fixes to the way partitionCurve combines regions.
 * It's not perfect, and needs more work.
 *
 * Revision 1.5  1994/01/31  20:56:42  macleod
 * macleod on Mon Jan 31 12:55:27 PST 1994
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Added partitionCurve, which partitions a curve such as a histogram
 * into regions, each region being a fairly large "bump" on the curve.
 * Added numericalDiff, which does smoothed numerical differentation
 * (1st and 2nd derivatives)
 * Also added interpolateMissingCodes, used to fill in missing codes in
 * scanner histograms
 *
 * Revision 1.4  1994/01/14  22:23:13  macleod
 * macleod on Fri Jan 14 14:21:41 PST 1994
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Twin peaks now takes a peak histo and a power histo, and two additional
 * parameters which indicate which histo to search for the first (highest) and
 * second peaks. This is done so that one can search for the histo
 * region with the most "energy" for background, but with the highest peak
 * for text.
 *
 * Revision 1.3  1993/12/03  01:14:01  ddavies
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/13  00:50:47  macleod
 * macleod on Fri Nov 12 16:50:40 PST 1993
 *
 * curves.c -> /product/ipcore/ipwrap/src/RCS/curves.c,v
 *
 * Fixed a couple of warnings
 *
 * Revision 1.1  1993/11/10  01:23:38  macleod
 * Initial revision
 *
 */

