/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  cut.c -- procedures that extract a rectangle from an image,
 *	     and resize the original image to fit the rectangle.
 *
 *              w_cutRectIP()     	resizes image to given rectangle
 *
 */



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "frames.pub"

/* prototypes */
#include "props.pub"
#include "shrrast.pub"
#include "cut.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: cut.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


/* When scanning an image, you usually get the entire platen.  The interesting
 * portion of the image (the page) is somewhere in that page.  After
 * figuring out where the image is, you'd like to discard the platen portion.
 * The obvious way to do this is to blit the rectangle of interest into
 * a brand new bitmap.  However, when scanning 24-bit color images on a
 * PC, the amount of memory required to hold 4 color planes (three original
 * plus one temp to be a blit target) is too high.  The EasyScan folks
 * asked for a way to do the cut "in place."
 *
 * This is actually pretty straightforward.  Remember that the bitmap
 * is really a 1-D array.  Each scan line of the original image occupies
 * a sequence of sourceBytesPerLine bytes.  After doing the cut, the
 * number of bytes/line in the destination will be less than or equal to
 * source bytes/line and the rectangle will start no higher than the top
 * of the image.  Therefore, we can simply blit the relevant portion of 
 * each scan line from the source rectangle to the upper left corner
 * of the former source image.  So long as we blit from left to right,
 * top to bottom, the source bits we cover up are guaranteed to be either
 * a) not needed, or b) already moved into position.
 *
 * After the rectangle is moved into position, we clear the four frames
 * of the destination image and return new bytes/line and the number of
 * bytes consumed by the image data plus frame.  The last argument
 * can be used by the client to resize the image buffer.
 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clip/cut a PIXR in-place down to some rectangle in the PIXR.
 *
 * LongDescription:::
 *      This procedure takes an image and a rectangle inside the image.
 *      It moves the pixels in the rectangle to the upper left corner
 *      of the image and resizes the image to fit the rectangle.  The memory
 *      not used by the rectangle is returned to the system.  Thus, the
 *      PIXR will have been replaced by its sub-PIXR the size of the input
 *      rectangle.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  x::              rectangle to be cut out relative to INSIDE of frame.
 *  y::              
 *  w::              
 *  h::              
 *
 * Output:::
 *  pixr::           Ptr to source PIXR to cut in-place.
 *
 * Returns:::
 *
 * Notes:::
 *  As of 8/1/97, the rectangle is cropped to the original image.  A rect
 *  which has a negative width or height after clipping results in an error.
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cutRectIP (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    PIXR               *pixr)
{
    PROLOGUE;
    Int32   sBpl, dBpl;                     /* original and resized images'
					     * bytes/line */
    Int32    depth;                         /* image's bits/pixel */
    UInt8   *pS;                            /* pointer to the source image */
    Int32    i, channels;
    Int32    origW, origH;
    PIXR    *subPixr;

    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetChannelCount (pIPWState, pixr, &channels));

	/* Bug Fix: 8/1/97: Check to see if the rectangle should be cropped.
	 * It must fit inside the original image.
	 * First, make sure the rect and the original image
	 * intersect.  If they don't return an error.
	 */
    ERR_PASS (w_pixrGetWidth (pIPWState, pixr, &origW));
    ERR_PASS (w_pixrGetHeight (pIPWState, pixr, &origH));
    if (((x + w) < 0) || ((y + h) < 0) || (x >= origW) || (y >= origH) ||
	(w < 0) || (h < 0))
	return cErrInvalidParm;

	/* knowing that this is a well formed rect that intersects the
	 * original image, clip it to fit if necessary.  First, pull
	 * in the left and top edges.
	 */
    if (x < 0)
    {
	w += x;		/* reduce width by part that falls off left edge */
	x = 0;
    }
    if (y < 0)
    {
	h += y;
	y = 0;
    }
	/* Now pull in the right and bottom edges.
	 */
    if (x + w > origW)		/* right side of rect is too far right */
	w = origW - x;
    if (y + h > origH)
	h = origH - y;

    for (i = 0; i < channels; i++)
    {
	/* calculate the starting address and depth */
	ERR_PASS( w_pixrGetChannel (pIPWState, pixr, i, &subPixr));
	ERR_PASS( w_pixrGetImageAll(pIPWState, subPixr, &pS));
	ERR_PASS( w_pixrGetDepth(pIPWState, subPixr, &depth));
	ERR_PASS( w_pixrGetBpl(pIPWState, subPixr, &sBpl));

	/* set the new pixr width so we can get the dest bytes/line */
	ERR_PASS( w_pixrSetWidth(pIPWState, subPixr, w));
	ERR_PASS( w_pixrBpl(pIPWState, subPixr, &dBpl));

	IP_YIELD;
	/* cut out the rectangle */

	ERR_PASS( i_rasterOp (pIPWState, pS,	/* dest pointer */
			      cFrameBits,	/* dest x offset */
			      cFrameBits,	/* dest y offset */
			      w,		/* width of rect */
			      h,		/* height of rect */
			      depth,		/* dest depth */
			      dBpl,		/* dest bytes/line */
			      PIX_SRC,		/* operation */
			      pS,		/* source pointer */
			      x + cFrameBits,	/* source x offset */
			      y + cFrameBits,	/* source y offset */
			      depth,		/* source depth */
			      sBpl));		/* source bytes/line */

	/* clear out the top frame ( redundant, right?) */
		 
	ERR_PASS( i_rasterOp (pIPWState,
			      pS,		/* dest pointer */
			      0,		/* dest x offset */
			      0,		/* dest y offset */
			      w + 2*cFrameBits,	/* width of rect */
			      cFrameBits,	/* height of rect */
			      depth,		/* dest depth */
			      dBpl,		/* dest bytes/line */
			      PIX_CLR,		/* operation */
			      NULL, 0, 0, 0, 0));

	/* clear out the left frame */
	ERR_PASS( i_rasterOp (pIPWState, pS,	/* dest pointer */
			      0,		/* dest x offset */
			      cFrameBits,	/* dest y offset */
			      cFrameBits,	/* width of rect */
			      h,		/* height of rect */
			      depth,		/* dest depth */
			      dBpl,		/* dest bytes/line */
			      PIX_CLR,		/* operation */
			      NULL, 0, 0, 0, 0));

	IP_YIELD;

	/* clear out the right frame */
	ERR_PASS( i_rasterOp (pIPWState, pS,	/* dest pointer */
			      w + cFrameBits,	/* dest x offset */
			      cFrameBits,	/* dest y offset */
			      cFrameBits,	/* width of rect */
			      h,		/* height of rect */
			      depth,		/* dest depth */
			      dBpl,		/* dest bytes/line */
			      PIX_CLR,		/* operation */
			      NULL, 0, 0, 0, 0));

	/* clear out the bottom frame */
	ERR_PASS( i_rasterOp (pIPWState, pS,	/* dest pointer */
			      0,		/* dest x offset */
			      h + cFrameBits,	/* dest y offset */
			      w + 2*cFrameBits,	/* width of rect */
			      cFrameBits,	/* height of rect */
			      depth,		/* dest depth */
			      dBpl,		/* dest bytes/line */
			      PIX_CLR,		/* operation */
			      NULL, 0, 0, 0, 0));

	/* update the PIXR's width, height, bytes/line and reallocate
	 * the image buffer */
	ERR_PASS( w_pixrSetHeight(pIPWState, subPixr, h));
	ERR_PASS( w_pixrSetBpl(pIPWState, subPixr, dBpl));
	ERR_PASS( w_reallocPixrData(pIPWState, subPixr));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


/*

 $Log: cut.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.28  1998/01/14 00:11:15  ddavies
 cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.27  1997/08/02 03:12:30  ddavies
 ddavies on Fri Aug 1 20:11:41 PDT 1997

 cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v

 Improve the rectangle checking in w_cutRectIP.  It now checks to see if
 the rect is entirely outside the image.  This would have confused the
 original fix.

 Revision 1.26  1997/08/02 02:17:58  ddavies
 ddavies on Fri Aug 1 19:16:16 PDT 1997

 cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v

 Make w_cutRectIP check it's incoming rectangle.  The input rect is now
 cropped at the original image boundaries.  The photocrop code occasionally
 calls this routine with bad rectangles.  These cause rasterOp to seg fault.

 Revision 1.25  1996/09/24 02:47:32  ddavies
 ddavies on Mon Sep 23 19:46:30 PDT 1996

 cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v

 Use w_pixrBpl to calculate the bytes/line of the destination image in
 w_cutRectInPlace instead of trying to calculate it here.  The bytes/line
 can depend on the processor, so it's a subtle calculation.

 Revision 1.24  1996/01/04 20:21:55  ejaquith
 cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.23  1995/12/13  04:06:07  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/13  20:10:31  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/13  21:59:32  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/09/29  00:36:45  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.19  1995/09/21  22:52:22  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/12  16:55:59  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.17  1995/09/08  00:15:52  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.16  1995/09/07  01:30:23  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.15  1995/09/06  20:18:11  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.14  1995/09/01  02:21:27  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.13  1995/08/31  21:11:51  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.12  1995/08/22  02:20:45  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.11  1995/08/17  22:02:24  ejaquith
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.10  1994/09/30  15:52:14  lperry
 * lperry on Fri Sep 30 08:51:49 PDT 1994
 *
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Add include of utils.pub
 *
 * Revision 1.9  1994/09/29  23:51:19  lperry
 * lperry on Thu Sep 29 16:50:22 PDT 1994
 *
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Use cut.pub and props.pub instead of pixrpros.h
 *
 * Revision 1.8  1994/09/20  04:27:11  ddavies
 * ddavies on Mon Sep 19 21:26:22 PDT 1994
 *
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Use the new reallocPixrData to resize the amount of memory associated with
 * a PIXR instead of duplicating the calculation here.
 *
 * Revision 1.7  1994/08/26  16:06:49  ddavies
 * ddavies on Fri Aug 26 09:03:47 PDT 1994
 *
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Fix bug in cutRectInPlacePr that caused a heap smash in PerfectScan.
 * The memory allocated for PIXRs should have enough extra bits at the end
 * so that after a rotate in place, there's room for the extra pad bits that
 * were on the right of the image.  The old version of cutRectInPlace did
 * not ask for this room when realloc'ing space for the smaller PIXR created.
 * When this was followed by a rotate in place, memory below the PIXR was
 * written, smashing the heap.
 *
 * Revision 1.6  1994/06/08  04:07:27  ddavies
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.5  1994/06/07  02:39:13  macleod
 * macleod on Mon Jun 6 19:39:11 PDT 1994
 *
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.4  1994/04/12  01:12:53  macleod
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.3  1994/03/02  03:55:18  ddavies
 * ddavies on Tue Mar 1 19:54:25 PST 1994
 *
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Add IP_YIELD to give up control during cutRectInPlacePr.
 *
 * Revision 1.2  1993/12/14  21:31:02  ddavies
 * ddavies on Tue Dec 14 13:30:08 PST 1993
 *
 * cut.c -> /product/ipcore/ipwrap/src/RCS/cut.c,v
 *
 * Change implementation of cutRectInPlace so that it does all the work
 * at this level instead of calling i_cutRectInPlace.  This involves
 * doing 5 rasterOps and fixing the dimensions of the PIXR.
 *
 * Revision 1.1  1993/12/10  05:39:53  ddavies
 * Initial revision
 *

 */
