/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * cyclseed.c
 *
 * Functions defined:
 *
 *	w_horizCycleSeed()
 */




#include "ipwcore.prv"

#include "pixr.h"
#include "frames.pub"
#include "defines.pub"     /* for cTrue and cFalse */
#include "geomadt.h"
#include "props.pub"
#include "shrcc.pub"

#include "cyclseed.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: cyclseed.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Seed all runs which are part of a cycle with length [minCycleThresh..maxCycleThresh]
 *
 * LongDescription:::
 *  The seed pixel is in the leftmost position of the black run in the
 *  cycle.  Source and destination PIXRs must be binary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  sPixr::          Source PIXR to be analyzed
 *  minCycleThresh:: Minimum cycle length to be seeded
 *  maxCycleThresh:: Maximum cycle length to be seeded
 *
 * Output:::
 *  seedPixr::       The destination PIXR where the seeds are written
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cycle seed
 *
 * Keywords:::
 *  cycle seed, run length
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_horizCycleSeed (
    IPWState           *pIPWState,
    PIXR               *sPixr,
    Int32               minCycleThresh,
    Int32               maxCycleThresh,
    PIXR              **seedPixr)
{
    Int32   wS, hS, bplS, bplD;      /* width, height, bpl of images */
    Bool    destCreated = cFalse;
    ScaledInt1616   xRes, yRes;
    Int32           rfc, bfc;
    UInt8  *iS, *iD;
    PROLOGUE;

    ERR_GEN_PIXR (sPixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*seedPixr != NULL)
        ERR_GEN_PIXR (*seedPixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetHeight (pIPWState, sPixr, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, sPixr, &wS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, sPixr, &xRes, &yRes));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS, hS, 1, xRes, yRes,
					      cFrameBits, seedPixr, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, sPixr, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *seedPixr, &hS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, sPixr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *seedPixr, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, sPixr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *seedPixr, &iD));

    /* Call the ipshared function */
    ERR_PASS_CLEANUP (i_horizCycleSeed (pIPWState, iS, bplS,
					wS, hS,
					minCycleThresh, maxCycleThresh,
					iD, bplD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *seedPixr, 
                                                 rfc, bfc));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, seedPixr);
    ERR_RETURN;

}  /* w_horizCycleSeed () */


/*
 $Log: cyclseed.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.18  1998/01/14 00:11:17  ddavies
 cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.17  1997/04/15 21:44:11  ddavies
 cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v

 Substitute NoInit (uninitialized) version of pixr creation routines where
 possible.

 Revision 1.16  1996/01/04 20:21:58  ejaquith
 cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.15  1995/12/13  04:06:11  ejaquith
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.14  1995/11/13  20:10:33  ejaquith
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.13  1995/10/13  21:59:35  ejaquith
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.12  1995/09/28  21:40:18  dferrell
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.11  1995/09/21  22:52:25  ejaquith
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.10  1995/09/19  22:27:40  dferrell
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * Updated to use new conditional pixr creation and right/
 * bottom frame clearing functionality.
 *
 * Revision 1.9  1995/09/15  22:37:56  ddavies
 * ddavies on Fri Sep 15 15:37:55 PDT 1995
 *
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.8  1995/09/11  21:21:01  dferrell
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.7  1995/09/07  01:30:25  ejaquith
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.6  1995/09/06  20:18:14  ejaquith
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.5  1995/09/01  17:07:44  dferrell
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.4  1995/08/31  21:11:53  ejaquith
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.3  1995/08/23  22:32:10  jcraig
 * jcraig on Wed Aug 23 15:30:43 PDT 1995
 *
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * pass the framed height to i_horizCycleSeed not the unframed height
 *
 * Revision 1.2  1995/08/17  21:52:39  dferrell
 * cyclseed.c -> /product/ipcore/ipwrap/src/RCS/cyclseed.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.1  1995/04/14  01:29:40  jcraig
 * Initial revision
 *
 */
