/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  	d_serial.c:   Wrappers for (de)serialization of a
 *                   "description" of an image
 *
 *	(De)serializing page descriptions to (from) an array of bytes:
 *		w_serializeDescriptionSize()
 *		w_serializeDescription()
 *		w_deserializeDescription()
 *
 *	Getting header info of page description array of bytes:
 *		w_getSerializedDescriptionHeader()
 *
 *	Combining two page descriptions:
 *		w_mergeSerializedDescriptions()
 */

#include <stdio.h>

#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "alplib.h"
#include "hauscomp.h"
#include "descrip.h"

/* prototypes */
#include "rectops.pub"
#include "descrip.pub"
#include "hauscomp.pub"
#include "descrip.prv"

#ifdef __GNUC__
#include "ansiprot.h"
#endif


/* This is ALWAYS 0 on checkin/release, and 1 only when checked out 
 * if some printed debugging information is desired.
 */
#define DEBUG_PROGRESS   0


IP_RCSINFO(RCSInfo, "$RCSfile: d_serial.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");



#define SWAP32_IF_NEEDED(bs, n)  (n) = ((bs) ? BYTE_SWAP_INT(n)   : (n))
#define SWAP16_IF_NEEDED(bs, n)  (n) = ((bs) ? BYTE_SWAP_SHORT(n) : (n))



typedef struct descripStream 
{
    unsigned char   *base;
    unsigned char   *ptr;
    Int32            size;
} descripStream;



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Search the stream for some character, without modifying stream
 *
 * LongDescription:::
 *  Finds the number of characters until the search character.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  stream::            Ptr to stream to inspect.  It will not be modified.
 *  searchChar::        Character to find in stream, beginning at the present
 *                      location of the stream's buffer pointer.
 *
 * Output:::
 *  pLen::              Ptr to number of characters until the search character
 *                      is found in the stream.  If the character is not
 *                      present, it is considered an error.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  serialization
 *
 * Keywords:::
 *  serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_streamUntil (
    IPWState        *pIPWState,
    descripStream   *stream,
    unsigned char    searchChar,
    Int32           *pLen)
{
    PROLOGUE;
    unsigned char   *ptr2;
    
    ptr2 = stream->ptr;

    /* Check if at or past end of stream already! */
    ERR_GEN(ptr2 >= stream->base + stream->size, cErrIO_ReadError);

    while (*ptr2 != searchChar)
    {
	ptr2++;

	/* Check if at or past end of stream already! */
	ERR_GEN(ptr2 >= stream->base + stream->size, cErrIO_ReadError);
    }

    *pLen = ptr2 - stream->ptr;
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform a memcpy from a stream to a buffer
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  stream::            Ptr to stream to write from
 *  elsize::            Size of elements to write
 *  nelem::             Number of elements to write
 *
 * Output:::
 *  buf::               Ptr to buffer where we'll write to
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  serialization
 *
 * Keywords:::
 *  serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_streamRead (
    IPWState      *pIPWState,
    descripStream *stream,
    size_t         elsize,
    size_t         nelem,
    void          *buf)
{
    PROLOGUE;

    /* See if enough room to read from stream */
    ERR_GEN(stream->ptr + (nelem*elsize) > stream->base + stream->size,
	    cErrIO_ReadError);

    memcpy(buf, stream->ptr, elsize*nelem);
    
    stream->ptr += elsize*nelem;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform a memcpy from a buffer to a stream
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  buf::               Ptr to buffer to write from
 *  elsize::            Size of elements to write
 *  nelem::             Number of elements to write
 *
 * Output:::
 *  stream::            Ptr to stream where we'll write to
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  serialization
 *
 * Keywords:::
 *  serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_streamWrite (
    IPWState        *pIPWState,
    const void      *buf,
    size_t           elsize,
    size_t           nelem,
    descripStream   *stream)
{
    PROLOGUE;

    /* See if enough room to write to stream */
    ERR_GEN(stream->ptr + (nelem*elsize) > stream->base + stream->size,
	    cErrIO_WriteError);

    memcpy(stream->ptr, buf, elsize*nelem);
    
    stream->ptr += elsize*nelem;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read "land marks" in a page's description from a serial bytestream.
 *
 * LongDescription:::
 *  On error, resets stream to state as when passed in.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  stream::            Ptr to stream where we'll read the data
 *  byteswap::          0 to indicate contents in stream are not byteswapped;
 *                      O.w. stream contents are byteswapped.
 *
 * Output:::
 *  pageData::          Ptr to (single) page description data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_streamReadLandMarks (
    IPWState         *pIPWState,
    descripStream    *stream,
    UInt8             byteswap,
    pageDataType     *pageData)
{
    PROLOGUE;
    Int32           i;
    UInt32          tmp;
    SLOTTYPEIO      type;
    unsigned char  *savePtr;

    /* In the event of error -- for cleanup */
    savePtr = stream->ptr;

    /* Read in the number of different kinds of landmarks there are */
    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
				   stream, sizeof(pageData->nSlots), 1,
				   &(pageData->nSlots)));
    SWAP32_IF_NEEDED(byteswap, pageData->nSlots);

    if (pageData->nSlots > 0)
    {
	
	pageData->lmSlots = (landmarksSlot *)IPW_CALLOC(
	    pageData->nSlots, 
	    sizeof(landmarksSlot), cAllocatorData);
	ERR_GEN_CLEANUP(pageData->lmSlots == NULL, cErrNomem);
    
	for (i = 0; i < pageData->nSlots; i++)
	{
	    /* Read in the type of this slot */
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream, sizeof(SLOTTYPEIO), 1,
					   &type));
	    SWAP32_IF_NEEDED(byteswap, type);

	    pageData->lmSlots[i].type = (landmarktype)type;

	    switch (pageData->lmSlots[i].type)
	    {
	    case TEXTSEGMENTS:
	    {
		textSegmentsType  *textSegs;
		RECTANGL          *rect, *prev, iorect;

		textSegs = (textSegmentsType *)
		    IPW_CALLOC(1, sizeof(textSegmentsType), cAllocatorData);
		ERR_GEN_CLEANUP(textSegs == NULL, cErrNomem);

		pageData->lmSlots[i].lms = (void *)textSegs;
	    
		/* Read in the number of bytes (total!) this entire slot will
		 * take up */
		ERR_PASS_CLEANUP(wp_streamRead(pIPWState, 
					       stream, sizeof(tmp), 1, 
					       &tmp));
		/* (unneeded since unused!) SWAP32_IF_NEEDED(byteswap, tmp);*/

		/* Read in the number of landmarks in this slot */
		ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					       stream, 
					       sizeof(textSegs->nLandmarks), 
					       1,
					        &(textSegs->nLandmarks)));
		SWAP32_IF_NEEDED(byteswap, textSegs->nLandmarks);

		/* Read in the number of special iterations used to get these
		 * landmarks */
		ERR_PASS_CLEANUP(wp_streamRead(pIPWState, 
					       stream, 
					       sizeof(textSegs->iterations), 
					       1, 
					       &(textSegs->iterations)));
		SWAP32_IF_NEEDED(byteswap, textSegs->iterations);

		/* Read in each landmark rectangle */
		prev = textSegs->landmarks = NULL;
		for (tmp = 0; tmp < textSegs->nLandmarks; tmp++)
		{
		    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
						   stream, 
						   sizeof(iorect.x),
						   1,
						    &(iorect.x)));

		    ERR_PASS_CLEANUP(wp_streamRead(pIPWState, 
						   stream,
						   sizeof(iorect.y), 
						   1,
						   &(iorect.y)));

		    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
						   stream,
						   sizeof(iorect.w),
						   1, 
						   &(iorect.w)));

		    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
						   stream,
						   sizeof(iorect.h),
						   1, 
						   &(iorect.h)));

		    
		    if (byteswap)
		    {
			iorect.x = BYTE_SWAP_INT(iorect.x);
			iorect.y = BYTE_SWAP_INT(iorect.y);
			iorect.w = BYTE_SWAP_INT(iorect.w);
			iorect.h = BYTE_SWAP_INT(iorect.h);
		    }

		    rect = NULL;
		    ERR_PASS_CLEANUP(w_createRect (pIPWState, 
						   iorect.x, iorect.y,
						   iorect.w, iorect.h, 
						   &rect));
		
		    /* Add the newly read rect in to be the new current end
		     * of the list.  (ie: textSegs->landmarks will point to
		     * the very first rect we read in from the DESCRIP)
		     */
		    if (prev == NULL)
		    {
			textSegs->landmarks = rect;
			rect->prev = NULL;
		    }
		    else
		    {
			prev->next = rect;
			rect->prev = prev;
		    }

		    rect->next = NULL;
		    prev = rect;
		}

		break;
	    }
	    default:
	    {
		unknownLM     *ulm;
	    
		/* Don't know what to do with this slot type, but read it in
		 * as a buffer, anyway, so we can write it out later untouched
		 * if necessary.
		 */
		ulm = (unknownLM *)IPW_CALLOC(1, sizeof(unknownLM), 
					      cAllocatorData);
		ERR_GEN_CLEANUP(ulm == NULL, cErrNomem);

		pageData->lmSlots[i].lms = (void *)ulm;
	    
		/* Read in the number of bytes (total!) this entire slot will
		 * take up */
		ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					       stream,
					       sizeof(ulm->nBytesInBuffer), 
					       1,
					       &(ulm->nBytesInBuffer)));
		
		SWAP32_IF_NEEDED(byteswap, ulm->nBytesInBuffer);
		
		if (ulm->nBytesInBuffer > 0)
		{
		    ulm->buffer = (UInt8 *)IPW_CALLOC(ulm->nBytesInBuffer, 
						      sizeof(UInt8),
						      cAllocatorData);
		    ERR_GEN_CLEANUP(ulm->buffer == NULL, cErrNomem);

		    /* Now just read this slot in as a buffer! */
		    ERR_PASS_CLEANUP(wp_streamRead(pIPWState, 
						   stream,
						   sizeof(UInt8),
						   ulm->nBytesInBuffer,
						   &(ulm->buffer)));
		}
		else
		    ulm->buffer = NULL;
	    }
	    }
	}
    }
    else
	pageData->lmSlots = NULL;
    
    NORMAL_RETURN;

ERR_CLEANUP:
    {
	/* Free any of the data we read and filled,
	 * Protecting any fields we did NOT touch here
	 */
	classType     *classes     = pageData->classes;
	instanceType  *instances   = pageData->instances;
	groupType     *groups      = pageData->groups;

	pageData->classes      = NULL;
	pageData->instances    = NULL;
	pageData->groups       = NULL;
	wp_destroyDescription( pIPWState, pageData );
	pageData->classes      = classes;
	pageData->instances    = instances;
	pageData->groups       = groups;
    }

    stream->ptr = savePtr;
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write "land marks" in a page's description to a serial bytestream.
 *
 * LongDescription:::
 *  On error, resets stream to state as when passed in.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageData::          Ptr to (single) page description data
 *
 * Output:::
 *  stream::            Ptr to stream where we'll add the data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_streamWriteLandMarks (
    IPWState        *pIPWState,
    pageDataType    *pageData,
    descripStream   *stream)
{
    PROLOGUE;
    Int32           i;
    UInt32          tmp;
    SLOTTYPEIO      type;
    unsigned char  *savePtr;

    /* In the event of error -- for cleanup */
    savePtr = stream->ptr;

    /* Write out the number of different kinds of landmarks there are
     */
    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pageData->nSlots),
				    sizeof(pageData->nSlots), 1, stream));

    for (i = 0; i < pageData->nSlots; i++)
    {
	type = (SLOTTYPEIO)pageData->lmSlots[i].type;
	
	/* Write out the type of this slot */
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &type, sizeof(SLOTTYPEIO), 
					1, stream));

	switch (pageData->lmSlots[i].type)
	{
	case TEXTSEGMENTS:
	{
	    textSegmentsType *textSegs = (textSegmentsType *)
		pageData->lmSlots[i].lms;
	    RECTANGL         *pRect;

	    /* Each landmark is a rectangle. 
	     * We have to save out the x,y,w,h
	     * of each landmark's rectangle
	     */
	    tmp =
		sizeof(textSegs->nLandmarks) + /* for "nLandmarks" */
		sizeof(textSegs->iterations);  /* for "iterations" */
	    if (textSegs->nLandmarks > 0)
		tmp += textSegs->nLandmarks * ( /* rectangles */
		    sizeof(textSegs->landmarks->x) +
		    sizeof(textSegs->landmarks->y) +
		    sizeof(textSegs->landmarks->w) +
		    sizeof(textSegs->landmarks->h));

	    /* Write out the number of bytes (total!) this entire slot will
	     * take up */
	    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, 
					    &tmp, sizeof(tmp), 1, stream));

	    /* Write out the number of landmarks in this slot */
	    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState,
					    &(textSegs->nLandmarks), 
					    sizeof(textSegs->nLandmarks),
					    1, stream));

	    /* Write out the number of special iterations used to get these
	     * landmarks */
	    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState,
					    &(textSegs->iterations), 
					    sizeof(textSegs->iterations), 
					    1, stream));

	    
	    /* Write out each landmark rectangle */
	    pRect = textSegs->landmarks;
	    for (tmp = 0; tmp < textSegs->nLandmarks; tmp++)
	    {
		ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pRect->x),
						sizeof(pRect->x), 1, stream));
		ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pRect->y),
						sizeof(pRect->y), 1, stream));
		ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pRect->w),
						sizeof(pRect->w), 1, stream));
		ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pRect->h), 
						sizeof(pRect->h), 1, stream));
		pRect = pRect->next;
	    }

	    break;
	}
	default:
	{
	    unknownLM    *ulm;
	    
	    /* Didn't know what to do with this unknown slot type, but
	     * we were smart about it!  we read in the number of bytes
	     * it took up into a buffer so we'll just dump it back out
	     * now.  Just because we don't look at it doesn't mean we
	     * can't transfer it 100% =)
	     */

	    ulm = (unknownLM *)(pageData->lmSlots[i].lms);

	    tmp = ulm->nBytesInBuffer;
	    
	    /* Write out the number of bytes this slot will take up */
	    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &tmp, sizeof(tmp), 
					    1, stream));

	    /* Write out the slot (the buffer we saved earlier!) */
	    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, ulm->buffer, 
					    sizeof(UInt8), tmp, stream));

	    
	    break;
	}
	}
    }
    

    NORMAL_RETURN;

ERR_CLEANUP:
    stream->ptr = savePtr;
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read "active objects" in a page's description from a serial bytestream
 *
 * LongDescription:::
 *  On error, resets stream to state as when passed in.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  stream::            Ptr to stream where we'll read the data
 *  byteswap::          0 to indicate contents in stream are not byteswapped;
 *                      O.w. stream contents are byteswapped.
 *
 * Output:::
 *  pageData::          Ptr to (single) page description data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_streamReadActiveObjList (
    IPWState         *pIPWState,
    descripStream    *stream,
    UInt8             byteswap,
    pageDataType     *pageData)
{
    PROLOGUE;
    Int32          i;
    Int32          tmp;
    classType     *pClass;
    instanceType  *pInstance;
    groupType     *pGroup;
    unsigned char *savePtr;

    /* In the event of error -- for cleanup */
    savePtr = stream->ptr;
    

    /* Read in the number of the page that follows */
    ERR_PASS(wp_streamRead(pIPWState,
			   stream,
			   sizeof(pageData->pageIndex),
			   1,
			   &(pageData->pageIndex)));
    SWAP16_IF_NEEDED(byteswap, pageData->pageIndex);
	

    /* Read in number of classes */
    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
				   stream,
				   sizeof(pageData->nClasses),
				   1,
				   &(pageData->nClasses)));
    SWAP32_IF_NEEDED(byteswap, pageData->nClasses);

#if DEBUG_PROGRESS
    fprintf(fp_out, "Number of classes = %ld\n", pageData->nClasses);
#endif /*DEBUG_PROGRESS*/

    if (pageData->nClasses > 0)
    {
	/* NOTE! Since we use calloc here, we know that in event of
	 * cleanup/freeing on error, the pointer fields
	 * model, srcPr, and extra in each classType struct will be
	 * NULL!
	 */
	pageData->classes =
	    (classType *)IPW_CALLOC(pageData->nClasses, sizeof(classType),
				    cAllocatorData);
	ERR_GEN_CLEANUP(pageData->classes == NULL, cErrNomem);

	/* Read in each of the classes in the file */
	for (i = 0; i < pageData->nClasses; i++)
	{
	    pClass = &(pageData->classes[i]);

	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(classKind),
					   1,
					   &(pClass->kind)));
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pClass->w),
					   1,
					   &(pClass->w)));
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pClass->h),
					   1,
					   &(pClass->h)));
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pClass->wThick),
					   1,
					   &(pClass->wThick)));
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pClass->hThick),
					   1,
					   &(pClass->hThick)));
	}

	if (byteswap)
	{
	    for (i = 0; i < pageData->nClasses; i++)
	    {
		pClass = &(pageData->classes[i]);

		pClass->kind   = BYTE_SWAP_INT(pClass->kind);
		pClass->w      = BYTE_SWAP_INT(pClass->w);
		pClass->h      = BYTE_SWAP_INT(pClass->h);
		pClass->wThick = BYTE_SWAP_INT(pClass->wThick);
		pClass->hThick = BYTE_SWAP_INT(pClass->hThick);
	    }
	}

#if DEBUG_PROGRESS
	fprintf(fp_out, "CLASSES: kind, width, height, wThick, hThick\n");
	for (i = 0; i < nClasses; i++)
	{
	    pClass = &(pageData->classes[i]);

	    fprintf(fp_out, "Class[%03ld] %6lu %6ld %6ld %6ld %6ld\n", i,
		    (UInt32)pClass->kind,
		    pClass->w,
		    pClass->h,
		    pClass->wThick,
		    pClass->wThick);
	}
#endif /*DEBUG_PROGRESS*/

    }
    else
	pageData->classes = NULL;
    



    /* Read in number of instances */
    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
				   stream,
				   sizeof(pageData->nInstances),
				   1,
				   &(pageData->nInstances)));
    SWAP32_IF_NEEDED(byteswap, pageData->nInstances);

#if DEBUG_PROGRESS
    fprintf(fp_out, "Number of instances = %ld\n", pageData->nInstances);
#endif /*DEBUG_PROGRESS*/

    if (pageData->nInstances > 0)
    {
	/* NOTE! Since we use calloc here, we know that in event of
	 * cleanup/freeing on error, the pointer fields
	 * "extra" and "name" in each instanceType struct will be 
	 * NULL!
	 */
	pageData->instances =
	    (instanceType *)IPW_CALLOC(pageData->nInstances, 
				       sizeof(instanceType),
				       cAllocatorData);
	ERR_GEN_CLEANUP(pageData->instances == NULL, cErrNomem);
	

#if DEBUG_PROGRESS
	fprintf(fp_out,
		"Class# guessx guessy  rectx  recty  rectw  recth rowColRims sampleScore name\n");
#endif /*DEBUG_PROGRESS*/
    
	/* Read in each of the instances in the file */
	for (i = 0; i < pageData->nInstances; i++)
	{
	    pInstance = &(pageData->instances[i]);
		

	    /* Get the user picked name of the object */

	    /* Get the number of bytes (until the NULL)/(in the string) */
	    ERR_PASS_CLEANUP( wp_streamUntil( pIPWState, stream,
					      NULL_CHAR, &tmp));
	    
	    pInstance->name = (char *)IPW_CALLOC(++tmp, sizeof(char),
						 cAllocatorData);
	    ERR_GEN_CLEANUP(pInstance->name == NULL, cErrNomem);

	    /* Get the string, leaving/tacking the NULL terminator
	     * on at the end!
	     */
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(unsigned char),
					   tmp,
					   pInstance->name));
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->classNo),
					   1,
					   &(pInstance->classNo))); 
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->guessx),
					   1,
					   &(pInstance->guessx))); 
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->guessy),
					   1,
					   &(pInstance->guessy))); 
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->rectx),
					   1,
					    &(pInstance->rectx))); 
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->recty),
					   1,
					   &(pInstance->recty))); 
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->rectw),
					   1,
					   &(pInstance->rectw))); 
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->recth),
					   1,
					   &(pInstance->recth))); 
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->rowColRims),
					   1,
					   &(pInstance->rowColRims))); 
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pInstance->sampleScore),
					   1,
					   &(pInstance->sampleScore))); 
#if DEBUG_PROGRESS
	    fprintf(fp_out,
		    "%6d %6ld %6ld %6ld %6ld %6ld %6ld        %2xx %11d ",
		    pInstance->classNo,
		    pInstance->guessx,
		    pInstance->guessy,
		    pInstance->rectx,
		    pInstance->recty,
		    pInstance->rectw,
		    pInstance->recth,
		    pInstance->rowColRims,
		    pInstance->sampleScore);
	    fprintf(fp_out, "%s\n", 
		    pInstance->name);

#endif /*DEBUG_PROGRESS*/
	    
	}

	if (byteswap)
	{
	    for (i = 0; i < pageData->nInstances; i++)
	    {
		pInstance = &(pageData->instances[i]);

		pInstance->classNo = BYTE_SWAP_SHORT(pInstance->classNo);
		pInstance->guessx  = BYTE_SWAP_INT  (pInstance->guessx);
		pInstance->guessy  = BYTE_SWAP_INT  (pInstance->guessy);
		pInstance->rectx   = BYTE_SWAP_INT  (pInstance->rectx);
		pInstance->recty   = BYTE_SWAP_INT  (pInstance->recty);
		pInstance->rectw   = BYTE_SWAP_INT  (pInstance->rectw);
		pInstance->recth   = BYTE_SWAP_INT  (pInstance->recth);
		pInstance->sampleScore =
		    BYTE_SWAP_SHORT(pInstance->sampleScore);
	    }
	}

#if DEBUG_PROGRESS
	fprintf(fp_out,
		"Class# guessx guessy  rectx  recty  rectw  recth rowColRims sampleScore name\n");
	for (i = 0; i < pageData->nInstances; i++)
	{
	    pInstance = &(pageData->instances[i]);

	    fprintf(fp_out,
		    "%6d %6ld %6ld %6ld %6ld %6ld %6ld        %2xx %11d ",
		    pInstance->classNo,
		    pInstance->guessx,
		    pInstance->guessy,
		    pInstance->rectx,
		    pInstance->recty,
		    pInstance->rectw,
		    pInstance->recth,
		    pInstance->rowColRims,
		    pInstance->sampleScore);
	    fprintf(fp_out, "%s\n",
		    pInstance->name);
	}
#endif /*DEBUG_PROGRESS*/

    }
    else
	pageData->instances = NULL;


    

    /* Read in number of groups */
    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
				   stream,
				   sizeof(pageData->nGroups),
				   1,
				   &(pageData->nGroups)));
    SWAP32_IF_NEEDED(byteswap, pageData->nGroups);

#if DEBUG_PROGRESS
    fprintf(fp_out, "Number of groups = %ld\n", pageData->nGroups);
#endif /*DEBUG_PROGRESS*/

    if (pageData->nGroups > 0)
    {
	/* NOTE! Since we use calloc here, we know that in event of
	 * cleanup/freeing on error, the pointer field
	 * instances in each groupType struct will be
	 * NULL!
	 */
	pageData->groups =
	    (groupType *)IPW_CALLOC(pageData->nGroups, sizeof(groupType),
				    cAllocatorData);
	ERR_GEN_CLEANUP(pageData->groups == NULL, cErrNomem);

	/* Read in each of the groups in the file */
	for (i = 0; i < pageData->nGroups; i++)
	{
	    pGroup = &(pageData->groups[i]);
		
	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
					   sizeof(pGroup->nInstances), 
					   1,
					   &(pGroup->nInstances)));
	    SWAP32_IF_NEEDED(byteswap, pGroup->nInstances);

	    if (pGroup->nInstances > 0)
	    {
		pGroup->instances = (Int16 *)IPW_CALLOC(
		    pGroup->nInstances,
		    sizeof(pGroup->instances), cAllocatorData);

		ERR_GEN_CLEANUP(pGroup->instances == NULL, cErrNomem);

		/* Read in the array of instances in this group */
		ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					       stream,
					       sizeof(pGroup->instances),
					       pGroup->nInstances,
					       &(pGroup->instances)));
	    }
	    else
		pGroup->instances = NULL;

	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
 					   sizeof(pGroup->minChecks),
					   1,
					   &(pGroup->minChecks)));

	    ERR_PASS_CLEANUP(wp_streamRead(pIPWState,
					   stream,
 					   sizeof(pGroup->maxChecks),
					   1,
					   &(pGroup->maxChecks)));
	}
	
	if (byteswap)
	{
	    Int32   j;
	
	    for (i = 0; i < pageData->nGroups; i++)
	    {
		pGroup = &(pageData->groups[i]);

		pGroup->minChecks = BYTE_SWAP_SHORT(pGroup->minChecks);
		pGroup->maxChecks = BYTE_SWAP_SHORT(pGroup->maxChecks);

		for (j = 0; j < pGroup->nInstances; j++)
		    pGroup->instances[j] =
			BYTE_SWAP_SHORT(pGroup->instances[j]);
	    }
	}
    }
    else
	pageData->groups = NULL;
    


    NORMAL_RETURN;

ERR_CLEANUP:
    {
	/* Free any of the data we read and filled,
	 * Protecting any fields we did NOT touch here
	 */
	landmarksSlot    *lmSlots = pageData->lmSlots;

	pageData->lmSlots      = NULL;
	wp_destroyDescription( pIPWState, pageData );
	pageData->lmSlots      = lmSlots;
    }
    
    stream->ptr = savePtr;
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write "active objects" in a page's description to a serial bytestream
 *
 * LongDescription:::
 *  On error, resets stream to state as when passed in.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageData::          Ptr to (single) page description data
 *
 * Output:::
 *  stream::            Ptr to stream where we'll add the data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_streamWriteActiveObjList (
    IPWState         *pIPWState,
    pageDataType     *pageData,
    descripStream    *stream)
{
    PROLOGUE;
    Int32           i;
    char            nullChar = NULL_CHAR;
    classType      *pClass;
    instanceType   *pInstance;
    groupType      *pGroup;
    unsigned char  *savePtr;

    /* In the event of error -- for cleanup */
    savePtr = stream->ptr;
    

    /* Write out the number of the page that will follow */
    ERR_PASS(wp_streamWrite(pIPWState, &(pageData->pageIndex),
			    sizeof(pageData->pageIndex), 
			    1, stream));

    /* Write out number of classes */
    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pageData->nClasses), 
				    sizeof(pageData->nClasses), 1, stream));
    
    /* Write out each of the classes to the file */
    for (i = 0; i < pageData->nClasses; i++)
    {
	pClass = &(pageData->classes[i]);

	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pClass->kind),
					sizeof(classKind), 1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pClass->w), 
					sizeof(pClass->w), 1,
					stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pClass->h),
					sizeof(pClass->h), 1,
					stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pClass->wThick),
					sizeof(pClass->wThick), 1,
					stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pClass->hThick),
					sizeof(pClass->hThick), 1,
					stream));
    }


    /* Write out number of instances */
    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pageData->nInstances),
				    sizeof(pageData->nInstances), 1, stream));

    /* Write out each of the instances to the file */
    for (i = 0; i < pageData->nInstances; i++)
    {
	pInstance = &(pageData->instances[i]);

	/* This is the user picked name of the object */
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, pInstance->name,
					sizeof(unsigned char),
					strlen(pInstance->name),
					stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &nullChar,
					sizeof(unsigned char), 1, stream));

	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->classNo),
					sizeof(pInstance->classNo),
					1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->guessx), 
					sizeof(pInstance->guessx), 1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->guessy),  
					sizeof(pInstance->guessy), 1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->rectx),   
					sizeof(pInstance->rectx), 1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->recty),
					sizeof(pInstance->recty), 1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->rectw),   
					sizeof(pInstance->rectw), 1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->recth),
					sizeof(pInstance->recth), 1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->rowColRims),
					sizeof(pInstance->rowColRims),
					1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pInstance->sampleScore), 
					sizeof(pInstance->sampleScore),
					1, stream));
    }


    /* Write out number of groups */
    ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pageData->nGroups),
				    sizeof(pageData->nGroups), 1, stream));

    /* Write out each of the groups to the file */
    for (i = 0; i < pageData->nGroups; i++)
    {
	pGroup = &(pageData->groups[i]);

	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pGroup->nInstances), 
					sizeof(pGroup->nInstances), 
					1, stream));
	/* Write out the array of instances in this group */
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pGroup->instances),
					sizeof(pGroup->instances), 
					pGroup->nInstances, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pGroup->minChecks),
					sizeof(pGroup->minChecks), 1, stream));
	ERR_PASS_CLEANUP(wp_streamWrite(pIPWState, &(pGroup->maxChecks),
					sizeof(pGroup->maxChecks), 1, stream));
    }
    
    NORMAL_RETURN;

ERR_CLEANUP:
    stream->ptr = savePtr;
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (Over)Write page description's serialized bytes header.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  nbytes::            The length of the serialized bytestream.
 *  byteswap::          0 to indicate contents in stream are not (or will
 *                      not need to be) byteswapped;
 *                      O.w. contents are (or will need to be) byteswapped.
 *  nPages::            Number of pages that will be serialized into this
 *                      array of description bytes.  Range is [0..255].
 *
 * Output:::
 *  descripBytes::      Writes header for description bytes serialization
 *                      starting at (valid location!) where this points to.
 *                      Doesn't modify ptr location.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_putSerializedDescriptionHeader (
    IPWState      *pIPWState,
    Int32          nbytes,
    UInt8          byteswap,
    UInt8          nPages,
    unsigned char *descripBytes)
{
    PROLOGUE;
    UInt8   version = CURRENT_DESCRIP_VERSION;
    UInt16  magic;

    ERR_GEN(descripBytes == NULL, cErrNullPtr);
    ERR_GEN(nbytes < DESCRIP_BYTES_HEAD_SIZE, cErrInvalidParm);

    magic = (byteswap ? BYTE_SWAP_SHORT(DESCRIP_MAGIC) : DESCRIP_MAGIC);

    memcpy(descripBytes,    &magic,   sizeof(UInt16));
    memcpy(descripBytes+2,  &version, sizeof(UInt8));
    memcpy(descripBytes+3,  &nbytes,  sizeof(Int32));
    memcpy(descripBytes+7,  &nPages,  sizeof(UInt8));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return the header information of a serialized page description bytestream
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  descripBytes::      Ptr to serialized description bytes
 *
 * Output:::
 *  nbytes::            Ptr to number of bytes in serialized bytestream
 *  byteswap::          Ptr to:
 *                      1--means any routine that reads 16 or 32 bit data
 *                      from the serialized bytes will need byteswaping.
 *                      0--means none needed.
 *  nPages::            Ptr to number of page descriptions in serialized bytes
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSerializedDescriptionHeader (
    IPWState        *pIPWState,
    unsigned char   *descripBytes,
    Int32           *nbytes,
    UInt8           *byteswap,
    UInt8           *nPages)
{
    PROLOGUE;
    UInt16       magic;
    UInt8        version;

    ERR_GEN(descripBytes == NULL  ||  nbytes == NULL  ||  byteswap == NULL  ||
	    nPages == NULL, cErrNullPtr);

    /*---HEADER---
     *
     * bytes[0..1]  magic# -- identifies as "description" bytestream.
     *                        also tells if we need to byte swap 16 and 32bits!
     * byte [2]     version# of these serialized bytes
     * bytes[3..6]  #bytes the entire bytestream is length wise (including
     *              this header)
     * byte [7]     #pages [0..255] stored in these serialized bytes
     *
     */


    /* Do absolute minimum -- read 2 bytes (must have 2 valid bytes!!)
     * and verify that we have an DESCRIP bytestream or punt.
     */
    memcpy(&magic,     descripBytes,     sizeof(UInt16));
    if (magic == DESCRIP_MAGIC)
	*byteswap = 0;
    else if (BYTE_SWAP_SHORT(magic) == DESCRIP_MAGIC)
	*byteswap = 1;
    else
	ERR_GEN(1, cErrInvalidParm);

    /* Now verify that it's an acceptable version */
    memcpy(&version, descripBytes+2, sizeof(UInt8));
    ERR_GEN(version != CURRENT_DESCRIP_VERSION, cErrInvalidParm);

    /* Now get the number of total bytes in the file, and do minimal
     * amount sanity checking.
     */
    memcpy(nbytes,   descripBytes+3, sizeof(Int32));
    if (*byteswap)
	*nbytes = BYTE_SWAP_INT(*nbytes);

    /* MUST have at least this many bytes! */
    ERR_GEN(*nbytes < DESCRIP_BYTES_HEAD_SIZE, cErrInvalidParm);

    /* Get the number of pages encoded in this DESCRIP bytestream */
    memcpy(nPages,   descripBytes+7, sizeof(UInt8));
    

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Merge two serialized bytes forms of page description(s)
 *
 * LongDescription:::
 *  Remember, we can only store [0..255] pages in any single serialized
 *  page description, so trying to merge descriptions where more than
 *  255 pages would result is considered an error.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  descripBytes1::     Ptr to serialized description bytes.  Routine will
 *                      not modify or destroy these bytes.
 *
 * Output:::
 *  descripBytes2::     Ptr to another serialized description bytes on input.
 *                      On return, descripBytes2 will be descripBytes2
 *                      followed sequentially by descripBytes1.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_mergeSerializedDescriptions (
    IPWState         *pIPWState,
    unsigned char    *descripBytes1,
    unsigned char   **descripBytes2)
{
    PROLOGUE;
    UInt8   byteswap1, byteswap2;
    UInt8   nPages1, nPages2;
    Int32   nbytes1, nbytes2;
	
    ERR_GEN(descripBytes1 == NULL  ||  descripBytes2 == NULL  ||
	    *descripBytes2 == NULL,
	    cErrNullPtr);
	
    ERR_PASS( w_getSerializedDescriptionHeader( pIPWState,
						descripBytes1,
						&nbytes1, 
						&byteswap1, 
						&nPages1));
    ERR_PASS( w_getSerializedDescriptionHeader( pIPWState,
						*descripBytes2,
						&nbytes2, 
						&byteswap2,
						&nPages2));

    /* We can only store 255 pages into a single FCF!
     * This would be true when a user tries to stuff another page in,
     * to store 256!
     */
    ERR_GEN((UInt32)nPages1 + (UInt32)nPages2 > 255,
	    cErrInvalidParm);

    /* This had better never happen! */
    ERR_GEN(byteswap1 != byteswap2, cErrInvalidParm);

    /* Here's how we're going to concatenate the two byte streams:
     * we'll rip off the new_fcfBytes header and realloc the fcfBytes
     * to be its original size plus the size of the new_fcfBytes minus 
     * it's ripped off header.
     */
    nbytes1 -= DESCRIP_BYTES_HEAD_SIZE;

    *descripBytes2 = (UInt8 *)IPW_REALLOC(*descripBytes2, nbytes1 + nbytes2,
					  cAllocatorData);
    ERR_GEN_CLEANUP(*descripBytes2 == NULL, cErrNomem);

    /* Append the new_fcfBytes to the end of the old fcfBytes */
    memcpy(*descripBytes2 + nbytes2,
	   descripBytes1 + DESCRIP_BYTES_HEAD_SIZE,
	   nbytes1);

    /* Overwrite the current header with the new one */
    ERR_PASS_CLEANUP( wp_putSerializedDescriptionHeader( pIPWState,
							 nbytes1 + nbytes2, 
							 byteswap1,
							 (UInt8)
							 (nPages1 + nPages2),
							 *descripBytes2));


    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}

    

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get number of bytes to serialize page description(s) internal format
 *
 * LongDescription:::
 *  Get the number of bytes required to dump the data for page description(s)
 *  to a serialized byte stream.  (If linked list of descriptions, return
 *  size for ALL).
 *
 *  See also w_serializeDescription().
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageDataLL::        Ptr to linked list of internal format page
 *                      descriptions, one node in list per page description.
 *
 * Output:::
 *  pSize::             Ptr to number of bytes to serialize data
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_serializeDescriptionSize (
    IPWState         *pIPWState,
    pageDataType     *pageDataLL,
    Int32            *pSize)
{
    PROLOGUE;
    Int32             size, i;
    classType        *classes;
    instanceType     *instances;
    groupType        *groups;
    landmarksSlot    *lmSlots;
    pageDataType     *pageData;

    ERR_GEN(pSize == NULL  ||  pageDataLL == NULL,
	    cErrNullPtr);

    /* Space for header info */
    size = DESCRIP_BYTES_HEAD_SIZE;
    if (sizeof(UInt16) + /* DESCRIP magic number */
	sizeof(UInt8) + /* version number */
	sizeof(Int32) + /* Number of bytes in entire stream (including bytes
			 * required for above!
			 */
	sizeof(UInt8) /* number of pages */
	!= DESCRIP_BYTES_HEAD_SIZE) {
	ERR_GEN(1, cErrInternal);
    }


    for (pageData = pageDataLL; pageData != NULL; pageData=pageData->pNext)
    {
	classes   = pageData->classes;
	instances = pageData->instances;
	groups    = pageData->groups;
	lmSlots   = pageData->lmSlots;

	/* Page number */
	size += sizeof(pageData->pageIndex);


	/* Landmarks */
	size += sizeof(pageData->nSlots);
	if (pageData->nSlots > 0)
	{
	    size += sizeof(SLOTTYPEIO) * pageData->nSlots;
	    for (i = 0; i < pageData->nSlots; i++)
	    {
		switch (lmSlots[i].type)
		{
		case TEXTSEGMENTS:
		{
		    textSegmentsType *textSegs =
			(textSegmentsType *)lmSlots[i].lms;

		    size +=
			sizeof(UInt32) +  /* for number bytes slot takes up */
			sizeof(textSegs->nLandmarks) + /* for "nLandmarks" */
			sizeof(textSegs->iterations); /* for "iterations" */
		    if (textSegs->nLandmarks > 0)
			size += textSegs->nLandmarks * ( /* rectangles */
			    sizeof(textSegs->landmarks->x) +
			    sizeof(textSegs->landmarks->y) +
			    sizeof(textSegs->landmarks->w) +
			    sizeof(textSegs->landmarks->h));
		    break;
		}
		default:
		{
		    unknownLM    *ulm;
	    
		    /* Didn't know what to do with this unknown slot type, but
		     * we were smart about it!  we read in the number of bytes
		     * it took up into a buffer so we'll just dump it back out
		     * now.  Just because we don't look at it doesn't mean we
		     * can't transfer it 100% =)
		     */
		
		    ulm = (unknownLM *)(lmSlots[i].lms);
		
		    size += 
			sizeof(UInt32) + /* for number bytes slot takes up */
			ulm->nBytesInBuffer; /* actual #bytes slot takes up */
		}
		}
	    }
	}
    
	/* Classes */
	size += sizeof(pageData->nClasses);
	if (pageData->nClasses > 0)
	    size += pageData->nClasses * (
		sizeof(classKind) + 
		sizeof(classes[0].w) +
		sizeof(classes[0].h) +
		sizeof(classes[0].wThick) + 
		sizeof(classes[0].hThick));
    
	/* Instances */
	size += sizeof(pageData->nInstances);
	if (pageData->nInstances > 0)
	{
	    for (i = 0; i < pageData->nInstances; i++)
		/* string length plus one for null terminator */
		size += (strlen(instances[i].name) + 1) *
		    sizeof(unsigned char);

	    size += pageData->nInstances * (
		sizeof(instances[0].classNo) +
		sizeof(instances[0].guessx) +
		sizeof(instances[0].guessy) +
		sizeof(instances[0].rectx) +
		sizeof(instances[0].recty) +
		sizeof(instances[0].rectw) +
		sizeof(instances[0].recth) +
		sizeof(instances[0].rowColRims) +
		sizeof(instances[0].sampleScore));	}
 
	/* Groups */
	size += sizeof(pageData->nGroups);
	if (pageData->nGroups > 0)
	{
	    for (i = 0; i < pageData->nGroups; i++) 
		size += sizeof(groups[i].instances) * 
		    groups[i].nInstances;
	
	    size += pageData->nGroups * (
		sizeof(groups[0].nInstances) +
		sizeof(groups[0].minChecks) +
		sizeof(groups[0].maxChecks));
	}
    }

    *pSize = size;

    NORMAL_RETURN;
}
 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert page description(s) internal format to its serialized bytes form
 *
 * LongDescription:::
 *  Inverse of w_deserializeDescription().
 *  The internal format fields that are NULL won't be serialized.
 *  (If linked list of descriptions, will serialize ALL).
 *  <BR><BR>
 *
 *  See also w_getSerializedDescriptionHeader() and 
 *  w_mergeSerializedDescriptions().
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageDataLL::        Ptr to linked list of internal format page
 *                      descriptions, one node in list per page description.
 *
 * Output:::
 *  descripBytes::      Ptr to ptr to (self-describing) serialized
 *                      description bytes, for all passed pages.
 *                      Ptr must be valid and *descripBytes must be NULL
 *                      on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_serializeDescription (
    IPWState         *pIPWState,
    pageDataType     *pageDataLL,
    unsigned char   **descripBytes)
{
    PROLOGUE;
    UInt8          nPages;
    descripStream *stream = NULL;
    pageDataType  *pageData;


    ERR_GEN(descripBytes == NULL  ||  pageDataLL == NULL,
	    cErrNullPtr);
    ERR_GEN(*descripBytes != NULL, cErrPtrNotNull);


    stream = (descripStream *)IPW_MALLOC(sizeof(descripStream), 
					 cAllocatorData);
    ERR_GEN(stream == NULL, cErrNomem);

    stream->base = NULL;

    ERR_PASS_CLEANUP(w_serializeDescriptionSize(pIPWState,
						pageDataLL, &(stream->size)));

    stream->base = (unsigned char *)IPW_CALLOC(stream->size,
					       sizeof(unsigned char),
					       cAllocatorData);
    ERR_GEN_CLEANUP(stream->base == NULL, cErrNomem);

    nPages = 0;
    for (pageData = pageDataLL; pageData != NULL; pageData=pageData->pNext)
    {
	/* Can serialize AT MOST 255 pages! */
	ERR_GEN_CLEANUP(nPages == 255, cErrInvalidParm);
	nPages++;
    }
    

    ERR_PASS_CLEANUP(wp_putSerializedDescriptionHeader(
	pIPWState,
	stream->size /* number bytes */,
	0, /* Won't be byteswapped! */
	nPages,
	stream->base));
    
    /* Point to just after the written out header */
    stream->ptr = stream->base + DESCRIP_BYTES_HEAD_SIZE;
    
    for (pageData = pageDataLL; pageData != NULL; pageData=pageData->pNext)
    {
	ERR_PASS_CLEANUP(wp_streamWriteActiveObjList(pIPWState,
						     pageData, stream));

	ERR_PASS_CLEANUP(wp_streamWriteLandMarks(pIPWState, pageData, stream));
    }


    /* Make sure we are EXACTLY where we think we should be -- that is,
     * exactly at the very end of the bytes.
     */
    ERR_GEN_CLEANUP(stream->ptr != stream->base + stream->size,
		    cErrInternal);


    *descripBytes = stream->base;
    IPW_FREE(stream, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (stream != NULL)
    {
	IPW_FREE(stream->base, cAllocatorData);
	IPW_FREE(stream, cAllocatorData);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert page description(s) serialized bytes form to its internal format
 *
 * LongDescription:::
 *  Inverse of w_serializeDescription().
 *  The internal format fields that weren't previously serialized 
 *  will be NULL on return.
 *  (If more than one page in serialized description, then create linked list
 *   of page descriptions).
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  descripBytes::      Ptr to (self-describing) serialized description bytes
 *
 * Output:::
 *  ppPageDataLL::      Ptr to ptr to linked list of internal format page
 *                      descriptions, one node in list per page description
 *                      stored in serialized bytes.
 *                      Ptr must be valid and *ppPageDataLL must be NULL
 *                      on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, serialization
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_deserializeDescription (
    IPWState         *pIPWState,
    unsigned char    *descripBytes,
    pageDataType    **ppPageDataLL)
{
    PROLOGUE;
    UInt8            byteswap, nPages, page;
    Int32            nbytes;
    descripStream    stream;
    pageDataType    *pageData, *pageDataLL = NULL, *pPrev = NULL;

    ERR_GEN(descripBytes == NULL  ||  ppPageDataLL == NULL,
	    cErrNullPtr);
    ERR_GEN(*ppPageDataLL != NULL, cErrPtrNotNull);
    
    ERR_PASS(w_getSerializedDescriptionHeader(pIPWState,
					      descripBytes, 
					      &nbytes, &byteswap, &nPages));

    /* Skip over header */
    stream.ptr  = &(descripBytes[DESCRIP_BYTES_HEAD_SIZE]);
    stream.size = nbytes;
    stream.base = descripBytes;

    for (page = 0; page < nPages; page++) 
    {
	/* We'll calloc so all the ptr fields are NULL */
	pageData = (pageDataType *)IPW_CALLOC(1, sizeof(pageDataType),
					      cAllocatorData);
	ERR_GEN_CLEANUP(pageData == NULL, cErrNomem);

	if (pPrev == NULL)
	    /* First node in linked list */
	    pageDataLL = pageData;
	else
	    /* Make previous-made node point to node to be read */
	    pPrev->pNext = pageData;
	

	/* Fill in the classes, instances, and groups */
	ERR_PASS_CLEANUP(wp_streamReadActiveObjList(pIPWState, 
						    &stream,
						    byteswap, 
						    pageData));

	/* Fill in the landmarks */
	ERR_PASS_CLEANUP(wp_streamReadLandMarks(pIPWState,
						&stream, 
						byteswap, 
						pageData));

	pPrev = pageData;
    }


    /* Make sure we are EXACTLY where we think we should be -- that is,
     * exactly at the very end of the bytes.
     */
    ERR_GEN_CLEANUP(stream.ptr != stream.base + stream.size,
		    cErrInternal);

    *ppPageDataLL = pageDataLL;

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyDescription(pIPWState, &pageDataLL);
            
    ERR_RETURN;
}



/*

 $Log: d_serial.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.3  1996/03/27 03:15:06  ejaquith
 ejaquith on Tue Mar 26 19:13:38 PST 1996

 d_serial.c -> /product/ipcore/ipwrap/src/RCS/d_serial.c,v

 Updated CLASSKIND to classKind (enumerated list to defined "list")
 LANDMARKTYPE to landmarktype (enum list to #defined "list"), and
 removed usage of now unnecessary KINDIO for reading/writing the
 type of class a special (check) mark.

 Revision 1.2  1996/02/22 04:03:25  ejaquith
 d_serial.c -> /product/ipcore/ipwrap/src/RCS/d_serial.c,v

 Fixed NT warning about param casting.

 Revision 1.1  1996/02/22 01:44:39  ejaquith
 Initial revision


 */
