/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  	descrip.c:   Wrappers for generating, using, and matching
 *                   "descriptions" of an image
 *
 *	Creating, managing, and destroying page descriptions:
 *		w_generateDescription()
 *		w_copyDescription()
 *		w_destroyDescription()
 *
 *	Creating, managing, and destroying a correlation between two
 *	page descriptions:
 *		w_correlateDescriptions()
 *		w_destroyCorrelation()
 *
 */

#include <stdio.h>

#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "defines.pub"
#include "hauscomp.h"
#include "descrip.h"
#include "imagerep.h"
#include "shrpatt.pub"

/* prototypes */
#include "props.pub"
#include "rectops.pub"
#include "itree.pub"
#include "descrip.pub"
#include "descrip.prv"
#include "hauscomp.pub"
#include "io.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: descrip.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");




#define HIGH_RECT_MATCH_MISS_PERCENT	0.5



/* These are ALWAYS 0 on checkin/release, and 1 only when checked out 
 * if some printed debugging information is desired.
 */
#define DEBUG_PROGRESS	0
#define DEBUG_TIME	0
#define DEBUG_INT_TREE	0



/* The tolerance here is 1/7th of an inch */
/* DPI: 100, 200, 300, 400  ==> PIXELS: 14, 28, 42, 57 */
#define MATCHTOL(res)  ((res) / 7)
#define PTOL           0.15

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a correlation between two page descriptions
 *
 * LongDescription:::
 *  This routine uses each description's landmarks which describe their source
 *  images, and correlates how these landmarks match to each other.
 *  We will produce a correlation for how the first page description relates
 *  to the second.
 *
 * PrivateDescription:::
 *  Original Version (hirsch)::
 *  Does a very simple matching by sorting the landmarks into approx distance
 *  from upper left corner and then matches using shape (width and height)
 *  that must be within a parameterized tolerance. This works fine in many
 *  cases.
 *
 *  Revision 2/14/94 (hirsch)::
 *  The original version begins to fail when there has been a fair degree
 *  of scaling or if the landmarks do not atomicize as on the original and 
 *  there are several landmarks of the same shape near each other. For
 *  example this could happend with a set of checkboxes in a column where 
 *  in one case two of them combine and the other they do not. The problem
 *  is that no account is taken of the positions and no checking is done
 *  to make sure that the translations and scalings encountered in doing the
 *  matches is reasonably consistent given the fact that we know we are on
 *  paper and only certain types of distortions happen (eg we don't worry
 *  about the paper being folded). So in this version I will try to keep
 *  track of the supposed shift and scaling going on so we can try to 
 *  spot potential matches that just do not make sense.<BR><BR>
 *
 *  Correlation is an array of ptrs to arrays of matching values, one array for
 *  each landmark slot.  Each array is the size of the number of
 *  landmarks in the "from" slot.
 *  The (index)/(element) of said array correspond to the
 *  ("from" landmark index)/("to" landmark index that it matches).
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageDataFrom::      Ptr to first page description
 *  pageDataTo::        Ptr to second page description
 *                      The page descriptions must have been made
 *                      from images with the same resolutions, and their
 *                      landmark types must be the same.
 *
 * Output:::
 *  correl::            Ptr to allocated (but not filled) correlation struct.
 *                      Will fill in with correlation between two page
 *                      page descriptions.
 *  nAttempted::        Ptr to total number of "from" landmarks we
 *                      tried to match to "to" landmarks.
 *  nMissed::           Ptr to number of "from" landmarks that ended up not
 *                      matching any "to" landmarks.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  template matching, page description
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  hirsch::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_matchLandMarks (
    IPWState        *pIPWState,
    pageDataType    *pageDataFrom,
    pageDataType    *pageDataTo,
    correlateType   *correl,
    Int32           *nAttempted,
    Int32           *nMissed)
{
    RECTANGL             *oldr, *newr;
    Int32                *takenarray = NULL, *maparray;
    Int32                 oldListIndex, newListIndex, slot, tolw, tolh;
    Int32                 scalex, scaley, shiftx, shifty;
    Int32                 deltax, deltay, deltaw, deltah;
    Int32                 myshiftx, myshifty, myscalex, myscaley;
    Int32                 first = cTrue;
    PROLOGUE;
    Int32                 numTaken = 0, nAttempts = 0, nMisses = 0;

    scalex = scaley = 100;
    shiftx = shifty = 0;


    ERR_GEN(pageDataTo->xres != pageDataFrom->xres  ||
	    pageDataTo->yres != pageDataFrom->yres, cErrInvalidParm);
    

    for (slot = 0; slot < pageDataTo->nSlots; slot++)
    {
	ERR_GEN_CLEANUP(pageDataTo->lmSlots[slot].type !=
			pageDataFrom->lmSlots[slot].type,
			cErrInvalidParm);

	switch (pageDataTo->lmSlots[slot].type)
	{
	case TEXTSEGMENTS:
	{
	    textSegmentsType   *textSegsTo, *textSegsFrom;
	    RECTANGL           *oldRect, *newRect;

	    textSegsTo = (textSegmentsType *)(pageDataTo->lmSlots[slot].lms);
	    textSegsFrom = (textSegmentsType *)(pageDataFrom->lmSlots[slot].lms);


	    takenarray = (Int32 *)IPW_CALLOC(textSegsTo->nLandmarks,
					     sizeof(Int32), cAllocatorData);
	    ERR_GEN_CLEANUP(takenarray == NULL,  cErrNomem);

	    /* Already been created */
	    maparray = correl->maparray[slot];


	    oldRect = textSegsFrom->landmarks;
	    newRect = textSegsTo->landmarks;

	    /* Historical note:
	     * This is David being really silly. How can you
	     * sort the rectangles and then expect the caller to know how your
	     * matches relate to the set of rects they passed you!!! So don't
	     * sorry. I also found out that the sort was not stable, so even
	     * though the caller sorted the rects the same way I was here the
	     * resulting list was not the same if two of the rects had
	     * identical x + ABS(y) which of course this one form did!! :(
	     * DEH 4/15/94
	     *
	     *  sortedOld = rectSort(oldRect, 
	     *  (ErrCode (CDECL *)(RECTANGL *, RECTANGL *))rectless);
	     *  sortedNew = rectSort(newRect, 
	     *  (ErrCode (CDECL *)(RECTANGL *, RECTANGL *))rectless);
	     */

	    for (oldListIndex = 0;
		 oldListIndex < textSegsFrom->nLandmarks; 
		 oldListIndex++) 
		maparray[oldListIndex] = NO_FEATURE_MATCHED;

	    for (newListIndex = 0;
		 newListIndex < textSegsTo->nLandmarks; 
		 newListIndex++) 
		takenarray[newListIndex] = 0;


	    if (textSegsTo->nLandmarks != textSegsFrom->nLandmarks)
	    {
#if DEBUG_PROGRESS
		i_printf (pIPWState, pIPWState->errorProc,
			  pIPWState->errorInfo,
			  "Warning! There are %ld new landmarks and %ld old landmarks\n",
			 textSegsTo->nLandmarks, textSegsFrom->nLandmarks);
#endif /* DEBUG_PROGRESS */
	    }
	    
	    for (oldListIndex = 0, oldr = oldRect; 
		 oldr != NULL;
		 oldListIndex++, oldr = oldr->next)
	    {
		tolw = MAX(MATCHTOL(pageDataTo->xres),
			   (Int32)(oldr->w * PTOL));

		tolh = MAX(MATCHTOL(pageDataTo->yres), 
			   (Int32)(oldr->h * PTOL)); 

		if ((oldr->x != 0) && (oldr->y != 0)) 
		{
		    for (newListIndex = 0, newr = newRect;
			 newr != NULL; 
			 newListIndex++, newr = newr->next)
		    {
			if ((!takenarray[newListIndex]) &&
			    (newr->x != 0) && (newr->y != 0))
			{
			    deltax = ROUND(newr->x - 
					   ((Float64)scalex/100 * oldr->x));
			    deltay = ROUND(newr->y -
					   ((Float64)scaley/100 * oldr->y));
			    deltaw = ABS(newr->w - oldr->w);
			    deltah = ABS(newr->h - oldr->h);
			    if (first)
			    {
				if ((deltaw <= tolw) &&
				    (deltah <= tolh))
				{
				    /* have a match */
				    takenarray[newListIndex] = 1;
				    maparray[oldListIndex] = newListIndex;
				     
				    /* Comment out here to stop scaling */
#if 0
				    scalex = ROUND(100 * (Float64)newr->w /
						   (Float64)oldr->w);

				    scaley = ROUND(100 * (Float64)newr->h /
						   (Float64)oldr->h);
#endif


				    shiftx = ROUND(newr->x -
						   ((Float64)scalex/100 * 
						    oldr->x));
				    shifty = ROUND(newr->y -
						   ((Float64)scaley/100 * 
						    oldr->y));
				    first = cFalse;
				    numTaken++;

#if DEBUG_PROGRESS
				    i_printf (pIPWState, pIPWState->errorProc,
					      pIPWState->errorInfo,
					      "\n Matched %li to %li", i, j);
				    i_printf (pIPWState, pIPWState->errorProc,
					      pIPWState->errorInfo,
					      "\n After %ld matches scale is (%ld %ld) shift is (%ld %ld)\n",
					      numTaken, scalex, scaley, shiftx, 
					      shifty);
#endif /*DEBUG_PROGRESS*/

				    break;
				}
			    } 
			    else
			    {
				/* not first time */
				if ((ABS(deltax - shiftx) <=
				     (2 * MATCHTOL(pageDataTo->xres))) &&

				    (ABS(deltay - shifty) <= 
				     (2 * MATCHTOL(pageDataTo->yres))) &&

				    (ABS(((Float64)scalex/100 * oldr->w) -
					 newr->w) <= tolw) &&

				    (ABS(((Float64)scaley/100 * oldr->h) -
					 newr->h) <= tolh))
				{
				    takenarray[newListIndex] = 1;
				    maparray[oldListIndex] = newListIndex;

#if DEBUG_PROGRESS
				    i_printf (pIPWState, pIPWState->errorProc,
					      pIPWState->errorInfo,
					      "\n Matched %li to %li", i, j);
				    i_printf (pIPWState, pIPWState->errorProc, 
					      pIPWState->errorInfo,
					      "Matched: dx is %d dy is %d\n",
					      ABS(deltax - shiftx), 
					      ABS(deltay - shifty));
#endif /*DEBUG_PROGRESS*/

				    myscalex = ROUND(100 * 
						     (Float64)newr->w /
						     (Float64)oldr->w);
				    myscaley = ROUND(100 * 
						     (Float64)newr->h /
						     (Float64)oldr->h);
/* Comment out here if want to stop scaling */
/*	      scalex = ((numTaken * scalex) + myscalex)/(numTaken + 1);
	      scaley = ((numTaken * scaley) + myscaley)/(numTaken + 1); */

				    myshiftx = ROUND(newr->x - 
						     ((Float64)scalex/100 *
						      oldr->x));
				    myshifty = ROUND(newr->y -
						     ((Float64)scaley/100 *
						      oldr->y));
				    shiftx = ((numTaken * shiftx) + myshiftx)/
					(numTaken + 1);
				    shifty = ((numTaken * shifty) + myshifty)/
					(numTaken + 1);
				    numTaken++;
#if DEBUG_PROGRESS
				    i_printf (pIPWState, pIPWState->errorProc,
					      pIPWState->errorInfo,
					      "\n After %ld matches myscale is (%ld %ld) myshift is (%ld %ld)\n",
					      numTaken, 
					      myscalex, myscaley,
					      myshiftx, myshifty);
				    i_printf (pIPWState, pIPWState->errorProc,
					      pIPWState->errorInfo,
					      "After %ld matches scale is (%ld %ld) shift is (%ld %ld)\n",
					      numTaken, 
					      scalex, scaley, 
					      shiftx, shifty);
#endif /*DEBUG_PROGRESS*/
				    break;
				}
			    }
			}
		    }
		}
	    }
	

	    IPW_FREE(takenarray, cAllocatorData);

	    /* Now accumulate some statistics to indicate performance */
	    nAttempts += textSegsFrom->nLandmarks;
	    for (oldListIndex = 0; 
		 oldListIndex < textSegsFrom->nLandmarks;
		 oldListIndex++)
		if (maparray[oldListIndex] == NO_FEATURE_MATCHED)
		    nMisses++;


#if DEBUG_PROGRESS
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "Matches are:\n");
	    for (i = 0; i < textSegsFrom->nLandmarks; i++)
		i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
			  " %ld", maparray[i]);
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "\n");
#endif /* DEBUG_PROGRESS */
	    
	    break;
	}
	default:
	    /* else we don't know how to process this other landmark type
	     * so ignore it!
	     */
	    break;
	}
    }
    
    correl->estScaleX = scalex;
    correl->estScaleY = scaley;

    *nAttempted = nAttempts;
    *nMissed    = nMisses;

    NORMAL_RETURN;  

ERR_CLEANUP:
    IPW_FREE(takenarray, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a correlation between two page descriptions
 *
 * LongDescription:::
 *  This routine uses each description's landmarks which describe their source
 *  images, and correlates how these landmarks match to each other.
 *  We will produce a correlation for how the first page description relates
 *  to the second.
 *
 * PrivateDescription:::
 *  This routine uses an interpretation-tree (full backtracking) search
 *  method for matching the landmark list of rectangles.<BR><BR>
 *
 *  Correlation is an array of ptrs to arrays of matching values, one array for
 *  each landmark slot.  Each array is the size of the number of
 *  landmarks in the "from" slot.
 *  The (index)/(element) of said array correspond to the
 *  ("from" landmark index)/("to" landmark index that it matches).
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageDataFrom::      Ptr to first page description
 *  pageDataTo::        Ptr to second page description
 *                      The page descriptions must have been made
 *                      from images with the same resolutions, and their
 *                      landmark types must be the same.
 *
 * Output:::
 *  correl::            Ptr to allocated (but not filled) correlation struct.
 *                      Will fill in with correlation between two page
 *                      page descriptions.
 *  nAttempted::        Ptr to total number of "from" landmarks we
 *                      tried to match to "to" landmarks.
 *  nMissed::           Ptr to number of "from" landmarks that ended up not
 *                      matching any "to" landmarks.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  template matching, page description
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_matchLandMarksITree (
    IPWState        *pIPWState,
    pageDataType    *pageDataFrom,
    pageDataType    *pageDataTo,
    correlateType   *correl,
    Int32           *nAttempted,
    Int32           *nMissed)
{
    PROLOGUE;
    Int32        slot, i, nAttempts = 0, nMisses = 0;

    ERR_GEN(pageDataFrom->xres != pageDataTo->xres  ||
	    pageDataFrom->yres != pageDataTo->yres, cErrInvalidParm);

    for (slot = 0; slot < correl->nSlots; slot++)
    {
	ERR_GEN(pageDataFrom->lmSlots[slot].type != pageDataTo->lmSlots[slot].type,
		cErrInvalidParm);

	switch (pageDataFrom->lmSlots[slot].type)
	{
	case TEXTSEGMENTS:
	{
	    textSegmentsType   *textSegsFrom, *textSegsTo;

	    textSegsFrom = (textSegmentsType *)(pageDataFrom->lmSlots[slot].lms);
	    textSegsTo = (textSegmentsType *)(pageDataTo->lmSlots[slot].lms);


	    ERR_PASS(w_interpTreeRectMatcher(pIPWState,
					     textSegsFrom->landmarks,
					     textSegsFrom->nLandmarks,
					     textSegsTo->landmarks,
					     textSegsTo->nLandmarks,
					     pageDataFrom->xres,
					     pageDataFrom->yres,
					     correl->maparray[slot]));



	    /* Now accumulate some statistics to indicate performance */
	    nAttempts += textSegsFrom->nLandmarks;
	    for (i = 0; i < textSegsFrom->nLandmarks; i++)
		if (correl->maparray[slot][i] == NO_FEATURE_MATCHED)
		    nMisses++;


#if DEBUG_INT_TREE
	    {
		RECTANGL    *pRectO, *pRectS;
		Int32        cnt;
		
		i_printf (pIPWState, pIPWState->errorProc,
			  pIPWState->errorInfo,
			  "\nFINAL INTERPRETATION TREE MATCHING (%ld model pts matched data pts):\n",
			  nAttempts - nMisses);
		for (i = 0; i < textSegsFrom->nLandmarks; i++)
		{
		    /* Get the "i"th rectangle in list */
		    for (pRectO = textSegsFrom->landmarks, cnt = 0;
			 cnt < i;
			 pRectO = pRectO->next, cnt++);

		    if (correl->maparray[slot][i] != NO_FEATURE_MATCHED)
		    {
			/* Get the "correl->maparray[slot][i]"th rectangle 
			 * in list
			 */
			for (pRectS = textSegsTo->landmarks, cnt = 0;
			     cnt < correl->maparray[slot][i];
			     pRectS = pRectS->next, cnt++);

			i_printf (pIPWState, pIPWState->errorProc,
				  pIPWState->errorInfo,
				  "M%2ld+D%2ld: MODEL rect [%6ld,%6ld,%6ld,%6ld] matched DATA rect [%6ld,%6ld,%6ld,%6ld]\n",
				  i, correl->maparray[slot][i],
				  pRectO->x, pRectO->y, pRectO->w, pRectO->h,
				  pRectS->x, pRectS->y, pRectS->w, pRectS->h);
		    }
		    else
			i_printf (pIPWState, pIPWState->errorProc,
				  pIPWState->errorInfo,
				  "M%2ld    : MODEL rect [%6ld,%6ld,%6ld,%6ld] no match\n",
				  i, 
				  pRectO->x, pRectO->y, pRectO->w, pRectO->h);
		}
	    }
	    

#endif /*DEBUG_INT_TREE*/
	  

#if DEBUG_PROGRESS
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "Matches are:\n");
	    for (i = 0; i < textSegsFrom->nLandmarks; i++)
		i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
			  " %ld", correl->maparray[slot][i]);
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "\n");
#endif /* DEBUG_PROGRESS */
	    
	    break;
	}
	default:
	    /* else we don't know how to process this other landmark type
	     * so ignore it!
	     */
	    break;
	}
    }

    /* Not too big on scaling yet */
    correl->estScaleX = correl->estScaleY = 100;

    *nAttempted = nAttempts;
    *nMissed    = nMisses;
    
    NORMAL_RETURN;  
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free a correlation (between two page descriptions)
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  correl::            Ptr to ptr to correlation.
 *                      Will destroy its allocated data and return ptr to
 *                      NULL on output.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyCorrelation (
    IPWState         *pIPWState,
    correlateType   **correl)
{
    PROLOGUE;
    ERR_GEN(correl == NULL, cErrInvalidParm);

    if (*correl != NULL)
    {
	Int32    i;

	for (i = 0; i < (*correl)->nSlots; i++)
	    IPW_FREE((*correl)->maparray[i], cAllocatorData);
	IPW_FREE((*correl)->maparray, cAllocatorData);

	IPW_FREE(*correl, cAllocatorData);
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the allocated fields in a page description and set ptrs to NULL.
 *
 * LongDescription:::
 *   Frees based on non-NULL fields in page description.  
 *   Any ptrs to client-allocated
 *   stuff in classes or instances "extra" fields must have already
 *   been freed, and the ptrs must point to NULL.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  pageData::          Ptr to page description on input.  
 *                      Will destroy all fields that allocated data and
 *                      set their ptrs to NULL on output.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_destroyDescription (
    IPWState        *pIPWState,
    pageDataType    *pageData)
{
    PROLOGUE;
    Int32           i;
    
    if (pageData != NULL)
    {
	if (pageData->classes != NULL)
	{
	    for (i = 0; i < pageData->nClasses; i++)
	    {
		if (pageData->classes[i].model)
		    ERR_PASS(w_destroyModelSet (
			pIPWState,
			&(pageData->classes[i].model)));

		if (pageData->classes[i].srcPr)
		    ERR_PASS(w_destroyPixr (pIPWState,
					    &pageData->classes[i].srcPr));
		
		
		if (pageData->classes[i].extra)
		    /* This field is supposed to already be freed
		     * (by however the client does so) and
		     * set to NULL!
		     */
		    ERR_GEN(1, cErrInvalidParm);
	    }
	    IPW_FREE(pageData->classes, cAllocatorData);
	    pageData->nClasses = 0;
	}
	

	if (pageData->instances != NULL)
	{
	    for (i = 0; i < pageData->nInstances; i++)
	    {
		IPW_FREE(pageData->instances[i].name, cAllocatorData);

		if (pageData->instances[i].extra)
		    /* This field is supposed to already be freed
		     * (by however the client does so) and
		     * set to NULL!
		     */
		    ERR_GEN(1, cErrInvalidParm);
	    }
	    IPW_FREE(pageData->instances, cAllocatorData);
	    pageData->nInstances = 0;
	}

	if (pageData->groups != NULL)
	{
	    for (i = 0; i < pageData->nGroups; i++)
	    {
		if (pageData->groups[i].instances)
		    IPW_FREE(pageData->groups[i].instances, cAllocatorData);
	    }
	    IPW_FREE(pageData->groups, cAllocatorData);
	    pageData->nGroups = 0;
	}
	    

	if (pageData->lmSlots != NULL)
	{
	    for (i = 0; i < pageData->nSlots; i++)
	    {
		switch (pageData->lmSlots[i].type)
		{
		case TEXTSEGMENTS:
		    ERR_PASS(w_destroyRectList (
			pIPWState,
			&((textSegmentsType *)
			  (pageData->lmSlots[i].lms))->landmarks));
		    IPW_FREE(pageData->lmSlots[i].lms, cAllocatorData);
		    break;
		default:
		{
		    unknownLM     *ulm;
		    ulm = (unknownLM *)(pageData->lmSlots[i].lms);
		
		    if (ulm != NULL  &&  ulm->buffer != NULL)
			IPW_FREE(ulm->buffer, cAllocatorData);

		    IPW_FREE(pageData->lmSlots[i].lms, cAllocatorData);
		
		    break;
		}
		    
		}
	    }

	    IPW_FREE(pageData->lmSlots, cAllocatorData);
	    pageData->nSlots = 0;
	}
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free a page description or list of them and set ptr to NULL.
 *
 * LongDescription:::
 *   Follows any "pNext" links in ptrs to page description structs and
 *   recursively destroys all nodes in linked list.  Frees based on
 *   non-NULL fields in page descriptions.  Any ptrs to client-allocated
 *   stuff in classes or instances "extra" fields must have already
 *   been freed, and the ptrs must point to NULL.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppPageDataLL::      Ptr to ptr to node (or linked list nodes) of page
 *                      description(s) on input.  Will destroy all allocated
 *                      associated data, and return ptr to NULL on output.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyDescription (
    IPWState        *pIPWState,
    pageDataType   **ppPageDataLL)
{
    PROLOGUE;
    pageDataType   *pageData, *nextPageData;

    
    ERR_GEN(ppPageDataLL == NULL, cErrInvalidParm);
    
    pageData = *ppPageDataLL;
    while (pageData != NULL)
    {
	nextPageData = pageData->pNext;

	ERR_PASS( wp_destroyDescription( pIPWState, pageData ));
	
	IPW_FREE(pageData, cAllocatorData);
	
	pageData = nextPageData;
    }
    *ppPageDataLL = NULL;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Allocate an appropriate correlation given a single page description
 *
 * LongDescription:::
 *  This just allocates and initializes a correlation based on the source
 *  pageData, but does not determine any actual correlation.  It constructs
 *  the appropriate correlation struct for matching given the source
 *  pageData's landmarks.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageData::          Ptr to page description
 *
 * Output:::
 *  ppCorrel::          Ptr to ptr to allocated correlation.  Note this
 *                      doesn't actually compute any correlation.
 *                      Ptr must be valid and *ppCorrel == NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_createCorrelation (
    IPWState        *pIPWState,
    pageDataType    *pageData,
    correlateType  **ppCorrel)
{
    PROLOGUE;
    Int32           i;
    correlateType  *correl = NULL;
    
    correl = (correlateType *)IPW_CALLOC(1, sizeof(correlateType),
					 cAllocatorData);
    ERR_GEN(correl == NULL, cErrNomem);

    correl->maparray = (Int32 **)IPW_CALLOC(pageData->nSlots,
					    sizeof(Int32 *), 
					    cAllocatorData);
    ERR_GEN_CLEANUP(correl->maparray == NULL, cErrNomem);
    correl->nSlots = pageData->nSlots;

    for (i = 0; i < pageData->nSlots; i++)
    {
	switch (pageData->lmSlots[i].type)
	{
	case TEXTSEGMENTS:
	{
	    textSegmentsType   *pTextSegs;

	    pTextSegs = (textSegmentsType *)(pageData->lmSlots[i].lms);

	    correl->maparray[i] = (Int32 *)IPW_CALLOC(
		pTextSegs->nLandmarks,
		sizeof(Int32), cAllocatorData);
	    ERR_GEN_CLEANUP(correl->maparray[i] == NULL,  cErrNomem);
	}
	default:
	    /* else we don't know how to process this other landmark type
	     * so ignore it, leaving NULL!
	     */
	    break;
	}
    }

    *ppCorrel = correl;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyCorrelation(pIPWState, &correl);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a correlation between two page descriptions
 *
 * LongDescription:::
 *  This routine uses each description's landmarks which describe their source
 *  images, and correlates how these landmarks match to each other.
 *  We will produce a correlation for how the first page description relates
 *  to the second.
 *
 * PrivateDescription:::
 *  We try to match the landmarks in two different ways.  The first is
 *  the fastest method, but less tolerant of errors.  If this does not
 *  produce enough matches, we try an interpretation-tree (full backtracking)
 *  search method, which is slower but has a much greater chance of
 *  getting accurate matches.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageDataFrom::      Ptr to first page description
 *  pageDataTo::        Ptr to second page description
 *                      The page descriptions must have been made
 *                      from images with the same resolutions, and their
 *                      landmark types must be the same.
 *
 * Output:::
 *  ppCorrel::          Ptr to ptr to computed correlation.
 *                      Ptr must be valid and *ppCorrel == NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, template matching
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_correlateDescriptions (
    IPWState        *pIPWState,
    pageDataType    *pageDataFrom,
    pageDataType    *pageDataTo,
    correlateType  **ppCorrel)
{
    PROLOGUE;
    Int32           i, nAttempted, nMissed;
    correlateType  *correl = NULL;
    correlateType  *correl2 = NULL;
#if DEBUG_TIME
    Float64         ttime;
#endif /* DEBUG_TIME */

    ERR_GEN(pageDataFrom == NULL  ||  pageDataTo == NULL  ||  
	    ppCorrel == NULL, cErrNullPtr);
    ERR_GEN(*ppCorrel != NULL, cErrPtrNotNull);
    

    ERR_GEN(pageDataFrom->xres != pageDataTo->xres  ||
	    pageDataFrom->yres != pageDataTo->yres, cErrInvalidParm);


    /* Consistency check first */
    ERR_GEN(pageDataFrom->nSlots != pageDataTo->nSlots, cErrInvalidParm);
    for (i = 0; i < pageDataFrom->nSlots; i++)
	ERR_GEN(pageDataFrom->lmSlots[i].type  !=  
		pageDataTo->lmSlots[i].type,
		cErrInvalidParm);
    
    ERR_PASS_CLEANUP(wp_createCorrelation(pIPWState, pageDataFrom, &correl));
    



#if DEBUG_TIME
    StartTime();
#endif
    ERR_PASS_CLEANUP(wp_matchLandMarks(pIPWState,
				       pageDataFrom, pageDataTo, correl,
				       &nAttempted, &nMissed));

#if DEBUG_TIME
    ttime = StopTime();
    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
	      "Matching the landmarks (reg) took    %f seconds\n",
	      ttime);
#endif

#if DEBUG_PROGRESS
    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
	      "Matched %4.2f percent of the landmarks\n",
	      ((Float32)(100*(nAttempted-nMissed)))/nAttempted);
#endif /*DEBUG_PROGRESS*/


    if ((Float32)nMissed / nAttempted < HIGH_RECT_MATCH_MISS_PERCENT)
	/* This is good enough */
	*ppCorrel = correl;
    else
    {
	Float32   avg = (Float32)nMissed / nAttempted;
	
	/* We missed a large percent of the landmarks above!
	 * Try the slower method that can backtrack out of a poorer previous
	 * match for a better one as it proceeds.
	 */
#if DEBUG_PROGRESS
	i_printf (pIPWState, pIPWState->errorProc,
		  pIPWState->errorInfo,
		  "Too few landmarks matched.  Trying interp. tree method!\n");
#endif /*DEBUG_PROGRESS*/

	ERR_PASS_CLEANUP(wp_createCorrelation(pIPWState,
					      pageDataFrom, &correl2));


#if DEBUG_TIME
	StartTime();
#endif /*DEBUG_TIME*/
	ERR_PASS_CLEANUP(wp_matchLandMarksITree(pIPWState,
						pageDataFrom, pageDataTo,
						correl2,
						&nAttempted, &nMissed));
#if DEBUG_TIME
	ttime = StopTime();
	i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		  "Matching the landmarks (interp tree) %f seconds\n",
		  ttime);
#endif /*DEBUG_TIME*/

#if DEBUG_PROGRESS
	i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		  "Matched %4.2f percent of the landmarks\n",
		  ((Float32)(100*(nAttempted-nMissed)))/nAttempted);
#endif /*DEBUG_PROGRESS*/
	
	if ((Float32)nMissed / nAttempted > avg)
	{
	    /* This means the interpretation tree method did WORSE.
	     * Use the previous results.
	     */
	    ERR_PASS_CLEANUP(w_destroyCorrelation(pIPWState, &correl2));
	    *ppCorrel = correl;
	}
	else
	{
	    /* We did better */
	    ERR_PASS_CLEANUP(w_destroyCorrelation(pIPWState, &correl));
	    *ppCorrel = correl2;
	}
    }


    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyCorrelation(pIPWState, &correl);
    w_destroyCorrelation(pIPWState, &correl2);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy a description of a page, copying all the original's fields
 *
 * LongDescription:::
 *  We copy the page description landmarks, classes, instances, and groups.
 *  In short all the useable fields.
 *  Note we don't copy the pNext field and 
 *  the ptr fields in the classes or instances will be set to NULL
 *  (except the "name" instance field will be copied)
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pageDataS::         Ptr to source page description to copy
 *
 * Output:::
 *  pageDataD::         Ptr to ptr to copied page description.
 *                      Ptr must be valid and *pageDataD == NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copyDescription (
    IPWState        *pIPWState,
    pageDataType    *pageDataS,
    pageDataType   **pageDataD)
{
    PROLOGUE;
    Int32           i;
    pageDataType   *pageData = NULL;

    ERR_GEN(pageDataS == NULL  ||  pageDataD == NULL, cErrNullPtr);
    ERR_GEN(*pageDataD != NULL, cErrPtrNotNull);
    

    pageData = (pageDataType *)IPW_CALLOC(1, sizeof(pageDataType), 
					  cAllocatorData);
    ERR_GEN(pageData == NULL, cErrNomem);

    /* Take care of all the easy fields first */
    pageData->xres      = pageDataS->xres;
    pageData->yres      = pageDataS->yres;
    pageData->nClasses  = pageDataS->nClasses;
    pageData->nInstances= pageDataS->nInstances;
    pageData->nGroups   = pageDataS->nGroups;
    pageData->nSlots    = pageDataS->nSlots;
    pageData->pageIndex = pageDataS->pageIndex;
    
    /* (Note how we leave the "pNext" field NULL --
     *  if we didn't, for one, then on error w_destroyDescription() would
     *  delete nodes from the _ORIGINAL_ pageData list!!)
     */


    /* Copy the classes */
    if (pageData->nClasses > 0) 
    {
	pageData->classes = (classType *)IPW_CALLOC(pageData->nClasses,
						    sizeof(classType),
						    cAllocatorData);
	ERR_GEN_CLEANUP(pageData->classes == NULL, cErrNomem);
	
	/* Copy the bytes since the memory is allocated */
	memcpy(pageData->classes, pageDataS->classes,
	       pageData->nClasses * sizeof(classType));
	
	/* Now adjust any struct ptrs. */
	for (i = 0; i < pageData->nClasses; i++)
	{
	    pageData->classes[i].model = NULL;
	    pageData->classes[i].srcPr = NULL;
	    pageData->classes[i].extra = NULL;
	}
    }
    


    /* Copy the instances */
    if (pageData->nInstances > 0)
    {
	pageData->instances = (instanceType *)IPW_CALLOC(pageData->nInstances,
							 sizeof(instanceType),
							 cAllocatorData);
	ERR_GEN_CLEANUP(pageData->instances == NULL, cErrNomem);
    
	/* Copy the bytes since the memory is allocated */
	memcpy(pageData->instances, pageDataS->instances,
	       pageData->nInstances * sizeof(instanceType));

	/* Now adjust any struct ptrs. */
	for (i = 0; i < pageData->nInstances; i++)
	{
	    pageData->instances[i].extra = NULL;
	    pageData->instances[i].name = (char *)IPW_CALLOC(
		strlen(pageDataS->instances[i].name) + 1, sizeof(char),
		cAllocatorData);
	    ERR_GEN_CLEANUP(pageData->instances[i].name == NULL, cErrNomem);
	
	    strcpy(pageData->instances[i].name, pageDataS->instances[i].name);
	}
    }
    

    /* Copy the groups */
    if (pageData->nGroups > 0)
    {
	pageData->groups = (groupType *)IPW_CALLOC(pageData->nGroups,
						   sizeof(groupType),
						   cAllocatorData);
	ERR_GEN_CLEANUP(pageData->groups == NULL, cErrNomem);

	/* Copy the bytes since the memory is allocated */
	memcpy(pageData->groups,
	       pageDataS->groups,
	       pageData->nGroups * sizeof(groupType));

	/* Now adjust any struct ptrs. */
	/* Note: Need to split these steps up in case of error and cleanup! */
	for (i = 0; i < pageData->nGroups; i++)
	    pageData->groups[i].instances = NULL;

	for (i = 0; i < pageData->nGroups; i++)
	{
	    if (pageData->groups[i].nInstances > 0)
	    {
		pageData->groups[i].instances = (Int16 *)IPW_CALLOC(
		    pageData->groups[i].nInstances,
		    sizeof(Int16), cAllocatorData);
	    
		ERR_GEN_CLEANUP(pageData->groups[i].instances == NULL, 
				cErrNomem);
		memcpy(pageData->groups[i].instances,
		       pageDataS->groups[i].instances,
		       pageData->groups[i].nInstances * sizeof(Int16));
	    }
	}
    }
    
    


    /* Copy the slots */
    if (pageData->nSlots > 0)
    {
	pageData->lmSlots = (landmarksSlot *)IPW_CALLOC
	    (pageData->nSlots, sizeof(landmarksSlot), cAllocatorData);
	ERR_GEN_CLEANUP(pageData->lmSlots == NULL, cErrNomem);

	for (i = 0; i < pageData->nSlots; i++)
	{
	    switch (pageDataS->lmSlots[i].type)
	    {
	    case TEXTSEGMENTS:
	    {
		/* scanned text segments slot */
		textSegmentsType   *textSegsS, *textSegs;
		
		pageData->lmSlots[i].lms = (void *)IPW_CALLOC
		    (1, sizeof(textSegmentsType), cAllocatorData);
		ERR_GEN_CLEANUP(pageData->lmSlots[i].lms == NULL, cErrNomem);
		pageData->lmSlots[i].type = TEXTSEGMENTS;

		textSegsS = (textSegmentsType *)pageDataS->lmSlots[i].lms;
		textSegs  = (textSegmentsType *)pageData ->lmSlots[i].lms;

		textSegs->iterations = textSegsS->iterations;
		textSegs->nLandmarks = textSegsS->nLandmarks;
	    
		ERR_PASS_CLEANUP( w_copyRectList( pIPWState,
						  textSegsS->landmarks,
						  &(textSegs->landmarks)));
	    
	    
		break;
	    }
	    default:
		/* We don't know how to process this type of landmark
		 * slot yet, so skip it
		 */
		pageData->lmSlots[i].type = pageDataS->lmSlots[i].type;
		pageData->lmSlots[i].lms = NULL;
		break;
	    }
	}
    }
    

    *pageDataD = pageData;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    w_destroyDescription( pIPWState, &pageData);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Given an image, generate a "description" of it
 *
 * LongDescription:::
 *  This routine attempts to find salient portions of an image and represent
 *  them in some non-bitmap way.  It is thus a way of describing the image
 *  without the raster bitmap.  We can have two "modes" here.  The first is
 *  to generate an initial description of some image.  The second is to
 *  generate an image description, basing its methodology on that of a 
 *  previous image description.  The latter is useful for generating a page
 *  description of a scanned page in a similar manner as a more ideal
 *  page description was generated from the original (pre-scan) image.
 *  We'll call these salient portions that we determine here landmarks
 *  and fill in just that field in the page description.
 *
 * PrivateDescription:::
 *  What find regions of text on the page, and return
 *  a list of rectangles that surround these regions.  We use the
 *  iterative processing of wp_findTextSegments() to do this, processing
 *  until we are satisfied with the list of rectangles to describe the page.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pr::                Ptr to source binary PIXR to make a page description
 *                      from
 *  xres::              Image resolution in X
 *  yres::              Image resolution in Y
 *  pageDataS::         If NULL, it means we want to generate an initial
 *                      page description, into pageDataD, for this image.
 *                      If non-NULL, it means we want to generate the page
 *                      description, into pageDataD, in the same "way" that
 *                      we generated this passed in one, 
 *                      using the same heuristics and algorithms and datatypes.
 *
 * Output:::
 *  pageDataD::         If ptr to NULL, it means allocate and fill in the
 *                      page description.
 *                      O.w., it means just fill in the relevant fields to the
 *                      existing data struct;  all of the relevant fields must
 *                      be cleared/NULL/0 in this case;  these fields are:
 *                      "lmSlots" and "nSlots".
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_generateDescription (
    IPWState        *pIPWState,
    PIXR            *pr,
    Int32            xres,
    Int32            yres,
    pageDataType    *pageDataS,
    pageDataType   **pageDataD)
{
    PROLOGUE;
    Int32          i;
    pageDataType  *pageData = NULL;

    /* [ pageDataS can be NULL or non-NULL] */
    /* [*pageDataD can be NULL or non-NULL] */
    ERR_GEN(pageDataD == NULL, cErrInvalidParm);

    ERR_GEN_PIXR(pr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* use calloc to set all fields to 0/NULL! */
    pageData = (pageDataType *)IPW_CALLOC(1, sizeof(pageDataType),
					  cAllocatorData);
    ERR_GEN(pageData == NULL,  cErrNomem);



    if (pageDataS != NULL)
    {
	/* Set up -- mirror the source slots. */
	pageData->lmSlots = (landmarksSlot *)IPW_CALLOC
	    (pageDataS->nSlots, sizeof(landmarksSlot), cAllocatorData);
	ERR_GEN_CLEANUP(pageData->lmSlots == NULL, cErrNomem);
	pageData->nSlots = pageDataS->nSlots;

	for (i = 0; i < pageData->nSlots; i++)
	{
	    switch (pageDataS->lmSlots[i].type)
	    {
	    case TEXTSEGMENTS:
		/* scanned text segments slot */
		
		pageData->lmSlots[i].lms = (void *)IPW_CALLOC
		    (1, sizeof(textSegmentsType), cAllocatorData);
		ERR_GEN_CLEANUP(pageData->lmSlots[i].lms == NULL, cErrNomem);
		pageData->lmSlots[i].type = TEXTSEGMENTS;
		break;
	    default:
		/* We don't know how to process this type of landmark
		 * slot yet so skip it
		 */
		pageData->lmSlots[i].type = pageDataS->lmSlots[i].type;
		pageData->lmSlots[i].lms = NULL;
		break;
	    }
	}
    }
    else
    {
	/* As of now we only have one type of slot of landmarks, TEXTSEGMENTS.
	 */

	pageData->lmSlots = (landmarksSlot *)IPW_CALLOC
	    (1, sizeof(landmarksSlot), cAllocatorData);
	ERR_GEN_CLEANUP(pageData->lmSlots == NULL, cErrNomem);
	pageData->nSlots = 1;

	pageData->lmSlots[0].lms = (void *)IPW_CALLOC
	    (1, sizeof(textSegmentsType), cAllocatorData);
	ERR_GEN_CLEANUP(pageData->lmSlots[0].lms == NULL, cErrNomem);
	pageData->lmSlots[0].type = TEXTSEGMENTS;
    }
    

    /* Now actually fill in the slots we made above */
    for (i = 0; i < pageData->nSlots; i++)
    {
	switch (pageData->lmSlots[i].type)
	{
	case TEXTSEGMENTS:
	{
	    /* Find scanned text segments */
	    textSegmentsType   *textSegs;

	    textSegs = (textSegmentsType *)(pageData->lmSlots[i].lms);

	    /* If we had a source page description, then we want to do as many
	     * iterations to the dest page description as it did.
	     * If we didn't have one (s.p.d.), then we want to find initial
	     * text segments and SAVE the number of iterations it took to
	     * find them.
	     */

	    ERR_PASS_CLEANUP(wp_findTextSegments(pIPWState, pr, xres, yres,
						 (pageDataS == NULL ? 0 :
						  ((textSegmentsType *)
						   pageDataS->lmSlots[i].lms)
						  ->iterations),
						 &(textSegs->iterations),
						 &(textSegs->landmarks),
						 &(textSegs->nLandmarks)));

	    break;
	}
	default:
	    /* We don't know how to process this type of landmark
	     * slot yet so skip it
	     */
	    break;
	}
    }


    pageData->xres = xres;
    pageData->yres = yres;
    

    if (*pageDataD == NULL)
    {
	*pageDataD = pageData;
    }
    else
    {
	/* Just transfer the relevant created slot fields
	 * to the existing page description.
	 * and destroy the leftover.
	 */
	ERR_GEN_CLEANUP((*pageDataD)->lmSlots != NULL  ||
			(*pageDataD)->nSlots != 0,
			cErrInvalidParm);

	(*pageDataD)->lmSlots = pageData->lmSlots;
	(*pageDataD)->nSlots = pageData->nSlots;
	
	/* Have to 0/NULL these so they don't get destroyed in the next call!
	 */
	pageData->lmSlots = NULL;
	pageData->nSlots = 0;
	ERR_PASS_CLEANUP(w_destroyDescription(pIPWState, &pageData));
    }
    
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyDescription(pIPWState, &pageData);
    ERR_RETURN;
}



/*

 $Log: descrip.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.5  1998/01/14 00:11:19  ddavies
 descrip.c -> /product/ipcore/ipwrap/src/RCS/descrip.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.4  1997/07/11 01:58:22  ddavies
 ddavies on Thu Jul 10 18:58:21 PDT 1997

 descrip.c -> /product/ipcore/ipwrap/src/RCS/descrip.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.3  1996/02/25 01:16:38  ejaquith
 descrip.c -> /product/ipcore/ipwrap/src/RCS/descrip.c,v

 Bug Fix.  Had confused array length with one of two too-similarly-named
 structures, so array was wrong length.  Fixed and updated names in
 routines to be more intuitive, and made code more intuitive in
 wp_matchLandMarks().

 Revision 1.2  1996/02/22 04:02:57  ejaquith
 ejaquith on Wed Feb 21 20:02:32 PST 1996

 descrip.c -> /product/ipcore/ipwrap/src/RCS/descrip.c,v

 Fixed a bunch of NT warnings about casts by using ROUND().

 Revision 1.1  1996/02/22 01:41:00  ejaquith
 Initial revision


 */
