/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  dtrans.c
 */
#include "ipwcore.prv"

#include "pixr.h"
#include "defines.pub"     /* for cTrue and cFalse */
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "prxform.pub"

#include "imageref.pub"

#include "shrip.pub"
#include "cmask.pub"
#include "ccsign.pub"

#include "wdtrans.prv"
#include "dtrans.pub"

#define DTSLOW 0


IP_RCSINFO(RCSInfo, "$RCSfile: dtrans.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Computes the 8 bit distance transform of a binary image.
 *
 * LongDescription:::
 *  Uses 4 or 8 neighbor "search" to compute the distance from either 1) a
 *  white pixel to the nearest black pixel, or 2) a black pixel to the nearest
 *  white pixel. Takes argument for assumed "color" (black or white) of the
 *  universe outside of the input image.  Note that the distance saturates
 *  at 255 in the output.
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  inPr::              Input binary image to be transformed
 *  connectType::       Either cDT4Connected or cDT8Connected for 4 or 8
 *                      neighbors, respectively
 *  fromBlack::         If cTrue, computes distances of all white pixels from
 *                      nearest black pixels. Otherwise computes distances of
 *                      all black pixels from nearest white pixels.
 *  universeColor::     Assumed color of universe outside of the input image (1
 *                      for black, 0 for white).
 *
 * Output:::
 *  outPr::             Pointer to pointer to output 8 bit image. If pointer to
 *                      NULL passed in, this image will be created.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  binary, depth conversion, gray, morphology, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *  See standard morphology texts/papers for detailed description.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_distanceTransform (
                     IPWState *pIPWState,
		     PIXR     *inPr,
		     Int32     connectType,
		     Int32     fromBlack,
		     Int32     universeColor,
		     PIXR    **outPr)
{
    PROLOGUE;
    Int32		 wS, hS, bplS, bplD; /* width, height, bpl
					      * of images */
    Bool		 destCreated = cFalse;
    ScaledInt1616	 xRes, yRes;
    Int32                rfc, bfc;
    UInt8		*iS, *iD;


    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*outPr != NULL)
        ERR_GEN_PIXR (*outPr, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xRes, &yRes));

    /* create the dest pixr if necessary */
    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS, hS, 8, xRes, yRes,
					      cFrameBits, outPr, &rfc, &bfc, 
					      &destCreated));


    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    ERR_PASS_CLEANUP (i_distanceTransform (pIPWState, iS, wS, hS,
					   bplS, connectType, 
                                           fromBlack, universeColor,
                                           iD, bplD));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    ERR_RETURN;

}  /* w_distanceTransform () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Calculate average color distance between 2 images.
 *
 * LongDescription:::
 *   This functon calculates average color distance between 2 images.
 *   In Yussi's paper, this distance is called low bound of EMD. It can
 *   speed up searching similar images.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  num1::           Number of color features (image 1)
 *  num2::           Number of color features (image 2)
 *  pFeat1::         Color features (image 1)
 *  pFeat2::         Color features (image 2)
 *
 * Output:::
 *  dis::            Difference between 2 images' color features
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   segmentation
 *
 * Keywords:::
 *   color, segmentation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Zhu
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_colorLowBoundDist (
    IPWState           *pIPWState,
    Int32               num1,
    Int32               num2,
    void               *pFeat1,
    void               *pFeat2,
    Int32              *dis)
{
    PROLOGUE;
    COLORSIGNATURE     *pF1, *pF2;
    Int32               d;

    ERR_GEN (!pFeat1, cErrNullPtr);
    ERR_GEN (!pFeat2, cErrNullPtr);
    ERR_GEN (!dis, cErrNullPtr);

    *dis =0;

    pF1 = (COLORSIGNATURE *) pFeat1;
    pF2 = (COLORSIGNATURE *) pFeat2;

    d = abs(((pF1+num1)->ll - (pF2+num2)->ll));
    *dis = d*d;
    d = abs(((pF1+num1)->aa - (pF2+num2)->aa));
    *dis += (d*d);
    d = abs(((pF1+num1)->bb - (pF2+num2)->bb));
    *dis += (d*d);
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Calculate color features' distance between 2 images.
 *
 * LongDescription:::
 *   Calculate color feature EMD. Using linear programming to find
 *   best solution.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  num1::           Number of color features (image 1)
 *  num2::           Number of color features (image 2)
 *  pFeat1::         Color features (image 1)
 *  pFeat2::         Color features (image 2)
 *
 * Output:::
 *  dis::            Difference between 2 images' color features
 *
 * Returns:::
 *
 * Notes:::
 *   An implementation of the Earth Movers Distance.
 *   Copyright (C) 1998 Yossi Rubner
 *   Computer Science Department, Stanford University
 *
 *   Here we rewrite it in our code style.
 *
 * FunctionCategory:::
 *   segmentation
 *
 * Keywords:::
 *   color, segmentation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_colorFeatureDist (
    IPWState           *pIPWState,
    Int32               num1,
    Int32               num2,
    void               *pFeat1,
    void               *pFeat2,
    Int32              *dis)
{
    PROLOGUE;
    DISTSTATE      *pState = NULL;
    Float32         totalCost, deltaMin;
    Int32           itr;
    nodeComp       *pNode;

    ERR_GEN (!pFeat1, cErrNullPtr);
    ERR_GEN (!pFeat2, cErrNullPtr);
    ERR_GEN (!dis, cErrNullPtr);

    *dis =0;

    ERR_PASS (wp_createDistanceState( pIPWState, (COLORSIGNATURE *)pFeat1,
				      (COLORSIGNATURE *)pFeat2,
				      num1, num2, &pState));

    pState->pointNode = pState->endNode ++;
    for(itr = 0; itr < ITRNUM; ++itr)
    {
	ERR_PASS (wp_lpDual (pIPWState, pState, &deltaMin) );
	if( deltaMin > (Float32) -Eps ) break;
	ERR_PASS (wp_lpNewSol (pIPWState, pState) );
    }

    totalCost = (Float32) 0;
    deltaMin = (Float32) 0;
    for(pNode = pState->nodeList; pNode != pState->endNode; pNode ++)
    {
	if( pNode == pState->pointNode) continue;
	if( pNode->i == pState->srcNum || pNode->j == pState->desNum)
	    continue;
	if( pNode->val > 0)
	{
	    deltaMin += pNode->val;
	    totalCost += pNode->val * pState->cost[pNode->i][pNode->j];
	}
    }

    *dis = (Int32)( totalCost / (Float32)pState->desCount );

    IPW_FREE(pState, cAllocatorData);

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Private function to init the distance state.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  pFeature1::      Color features (image 1)
 *  pFeature2::      Color features (image 2)
 *  num1::           Number of color features (image 1)
 *  num2::           Number of color features (image 2)
 *
 * Output:::
 *  ppState::        Point to point of state
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   segmentation
 *
 * Keywords:::
 *   color, segmentation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_createDistanceState (
    IPWState           *pIPWState,
    void               *pFeature1,
    void               *pFeature2,
    Int32               num1,
    Int32               num2,
    DISTSTATE         **ppState)
{
    PROLOGUE;
    COLORSIGNATURE  *pFeat1, *pFeat2,*srcF, *desF;
    Float32            count1, count2 , diff;
    Int32            idx, srcN, desN,  num, idx1, idx2;
    Float32          sumSrc, sumDes;
    Int32            dl, da, db;
    UInt8            l, a, b;
    DISTSTATE       *pState;

    pFeat1 = (COLORSIGNATURE *) pFeature1;
    pFeat2 = (COLORSIGNATURE *) pFeature2;

    pState = (DISTSTATE *) IPW_CALLOC(sizeof(DISTSTATE), 1, cAllocatorData);
    ERR_GEN( pState == NULL, cErrNomem);

    idx1 =0;
    for(idx=0; idx<num1; ++idx)
	idx1 += (pFeat1+idx)->num;
    idx2 =0;
    for(idx=0; idx<num2; ++idx)
	idx2 += (pFeat2+idx)->num;

    sumSrc = (Float32)(pFeat1+num1)->num;
    sumDes = (Float32)(pFeat2+num2)->num;

    count1 = (Float32) idx1/sumSrc;
    count2 = (Float32) idx2/sumDes;

    if(count2 > count1)
    {
	pState->srcNum = num2;
	pState->desNum = num1;
	pState->srcFeat = (void *)pFeat2;
	pState->desFeat = (void *)pFeat1;
	srcF = pFeat2;
	desF = pFeat1;
	pState->srcCount = count2;
	pState->desCount = count1;
	sumSrc = (Float32)(pFeat2+num2)->num;
	sumDes = (Float32)(pFeat1+num1)->num;
    }
    else
    {
	pState->srcNum = num1;
	pState->desNum = num2;
	pState->srcFeat = (void *)pFeat1;
	pState->desFeat = (void *)pFeat2;
	srcF = pFeat1;
	desF = pFeat2;
	pState->srcCount = count1;
	pState->desCount = count2;
    }

    srcN = pState->srcNum;
    desN = pState->desNum;
    for(idx=0; idx<srcN; ++idx)
    {
	pState->s[idx] = (Float32) (srcF+idx)->num/sumSrc;
    }

    for(idx=0; idx<desN; ++idx)
    {
	pState->d[idx] = (Float32) (desF+idx)->num/sumDes;
    }

    for(idx1=0; idx1<srcN; ++idx1)
    {
	num = (srcF + idx1)->num;
	l = (srcF + idx1)->ll;
	a = (srcF + idx1)->aa;
	b = (srcF + idx1)->bb;
	for(idx2=0; idx2<desN; ++idx2)
	{
	    dl = abs(l - (desF+idx2)->ll);
	    da = abs(a - (desF+idx2)->aa);
	    db = abs(b - (desF+idx2)->bb);
	    pState->cost[idx1][idx2] = 
		(Float32)( dl*dl + da*da + db*db );

	}
    }

    diff = pState->srcCount - pState->desCount;
    if( diff > 0 )
    {
	for(idx=0; idx<srcN; ++idx)
	    pState->cost[idx][desN] = (Float32) 0;
	pState->d[desN] = (Float32)( diff );
	++ pState->desNum;
    }

    pState->endNode = &pState->nodeList[0];

    ERR_PASS (wp_lpRussel ( pIPWState, pState ) );

    *ppState = pState;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Private function to find init variables.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::     State pointer
 *  pState::        Point to state
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   segmentation
 *
 * Keywords:::
 *   color, segmentation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_lpRussel (
    IPWState          *pIPWState,
    DISTSTATE         *pState)
{
    PROLOGUE;
    Int32 i, j, found, minI, minJ;
    Int32 Ucount, Vcount, offset;
    Float32 deltaMin, oldVal, diff, tmp, T;
    Float32 *Delta;

    struct node U[MAXLEN], UHead,  *CurU, *PrevU, *PrevUMinI, *Next;
    struct node V[MAXLEN], VHead,  *CurV, *PrevV, *PrevVMinJ;

    Delta = (Float32 *) IPW_CALLOC(sizeof(Float32), MAXLEN*MAXLEN, 
				   cAllocatorData);
    ERR_GEN( Delta == NULL, cErrNomem);

    Ucount= pState->srcNum;
    Vcount= pState->desNum;

    UHead.Next = CurU = U;
    for (i=0; i < Ucount; i++)
    {
	CurU->i = i;
	CurU->val = (Float32) -Inf;
	CurU->Next = CurU +1;
	++ CurU;
    }
    (--CurU)->Next = NULL;
  
    VHead.Next = CurV = V;
    for (j=0; j < Vcount; j++)
    {
	CurV->i = j;
	CurV->val = (Float32) -Inf;
	CurV->Next = CurV+1;
	CurV++;
    }
    (--CurV)->Next = NULL;
  
    for(i=0; i < Ucount ; i++)
    {
	for(j=0; j < Vcount ; j++)
	{
	    tmp = pState->cost[i][j];
	    if (U[i].val <= tmp) U[i].val = tmp;
	    if (V[j].val <= tmp) V[j].val = tmp;
	}
    }

    for(i=0, offset = 0; i < pState->srcNum ; i++, offset += MAXLEN)
	for(j=0; j < pState->desNum ; j++)
	    *(Delta+offset+j) = pState->cost[i][j] - U[i].val - V[j].val;

    minI = minJ =0;
    PrevUMinI = &UHead;
    PrevVMinJ = &VHead;
    do
    {
	found = 0; 
	deltaMin = (Float32) Inf; 
	PrevU = &UHead;
	for (CurU=UHead.Next; CurU != NULL; CurU=CurU->Next)
	{
	    i = CurU->i;
	    offset = i*MAXLEN;
	    PrevV = &VHead;
	    for (CurV=VHead.Next; CurV != NULL; CurV=CurV->Next)
	    {
		j = CurV->i;
		if (deltaMin > *(Delta+offset+j))
		{
		    deltaMin = *(Delta+offset+j);
		    minI = i;
		    minJ = j;
		    found = 1;
		    PrevUMinI = PrevU;
		    PrevVMinJ = PrevV;
		}
		PrevV = CurV;
	    }
	    PrevU = CurU;
	}

	if ( !found )
	    break;

	Next = PrevUMinI->Next;
	
	if( pState->s[minI] == pState->d[minJ] )
	{
	    T = pState->s[minI];
	    pState->s[minI] = (Float32) 0;
	    pState->d[minJ] = (Float32) 0;
	}
	else if( pState->s[minI] < pState->d[minJ])
	{
	    T = pState->s[minI];
	    pState->s[minI] = (Float32) 0;
	    pState->d[minJ] -= T;
	}
	else
	{
	    T = pState->d[minJ];
	    pState->s[minI] -= T;
	    pState->d[minJ] = (Float32) 0;
	}

	pState->mask[minI][minJ] = 1; 
	pState->endNode->val = T;
	pState->endNode->i = minI;
	pState->endNode->j = minJ;
	pState->endNode->nextCol = pState->rowList[minI];
	pState->endNode->nextRow = pState->colList[minJ];
	pState->rowList[minI] = pState->endNode;
	pState->colList[minJ] = pState->endNode;
	++ pState->endNode;

	if(pState->s[minI] == 0 && UHead.Next->Next != NULL )
	    PrevUMinI->Next = PrevUMinI->Next->Next;
	else
	    PrevVMinJ->Next = PrevVMinJ->Next->Next;

	if(Next == PrevUMinI->Next)
	{
	    for (CurV=VHead.Next; CurV != NULL; CurV=CurV->Next)
	    {
		j = CurV->i;
		if (CurV->val == pState->cost[minI][j])
		{
		    oldVal = CurV->val;
		    CurV->val = (Float32) -Inf;
		    for (CurU=UHead.Next; CurU != NULL; CurU=CurU->Next)
		    {
			i = CurU->i;
			if(CurV->val <= pState->cost[i][j])
			    CurV->val = pState->cost[i][j];
		    }
		    diff = oldVal - CurV->val;
		    if( diff != 0 )
		    {
			for(CurU = UHead.Next; CurU != NULL; CurU = CurU->Next)
			{
			    offset = CurU->i * MAXLEN;
			    *(Delta+offset+j) += diff;
			}
		    }
		}
	    }
	}
	else
	{
	    for (CurU=UHead.Next; CurU != NULL; CurU=CurU->Next)
	    {
		i = CurU->i;
		offset = i* MAXLEN;
		if (CurU->val == pState->cost[i][minJ])
		{
		    oldVal = CurU->val;
		    CurU->val = (Float32) -Inf;
		    for (CurV=VHead.Next; CurV != NULL; CurV=CurV->Next)
		    {
			j = CurV->i;
			if(CurU->val < pState->cost[i][j])
			    CurU->val = pState->cost[i][j];
		    }
		    diff = oldVal - CurU->val;
		    if( diff != 0 )
		    {
			for(CurV = VHead.Next; CurV != NULL; CurV = CurV->Next)
			{
			    *(Delta+offset+CurV->i) += diff;
			}
		    }
		}
	    }
	}
	  
    } while (UHead.Next != NULL || VHead.Next != NULL);

    IPW_FREE(Delta, cAllocatorData);
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Private function to find variables.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::     State pointer
 *  pState::        Point to state
 *
 * Output:::
 *  deltaMin::      Point to min delta
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   segmentation
 *
 * Keywords:::
 *   color, segmentation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_lpDual (
    IPWState          *pIPWState,
    DISTSTATE         *pState,
    Float32           *deltaMin)
{
    PROLOGUE;
    Int32 i, j, minI, minJ;
    Int32 UfoundNum=0, VfoundNum=0, found;
    Float32 delta, minDelta;
    struct node U[MAXLEN], U0Head,  U1Head, *CurU, *PrevU;
    struct node V[MAXLEN], V0Head,  V1Head, *CurV, *PrevV;
    
    U0Head.Next = CurU = U;
    for (i=0; i < pState->srcNum; i++)
    { 
	CurU->i = i;
	CurU->Next = CurU+1;
	CurU++;
    }
    (--CurU)->Next = NULL;
    U1Head.Next = NULL;

    V0Head.Next = CurV = V +1;
    for (j=1; j < pState->desNum; j++)
    {
	CurV->i = j;
	CurV->Next = CurV+1;
	CurV++;
    }
    (--CurV)->Next = NULL;
    V1Head.Next = NULL;

    V[0].val = (Float32) 0;
    V[0].i = 0;
    V1Head.Next = V;
    V1Head.Next->Next = NULL;

    while (UfoundNum < pState->srcNum || VfoundNum < pState->desNum)
    {
	found = 0;
	if (VfoundNum < pState->desNum)
	{
	    PrevV = &V1Head;
	    for (CurV=V1Head.Next; CurV != NULL; CurV=CurV->Next)
	    {
		j = CurV->i;
		PrevU = &U0Head;
		for (CurU=U0Head.Next; CurU != NULL; CurU=CurU->Next)
		{
		    i = CurU->i;
		    if (pState->mask[i][j] == 1)
		    {
			CurU->val = pState->cost[i][j] - CurV->val;
			PrevU->Next = CurU->Next;
			CurU->Next = U1Head.Next != NULL ? U1Head.Next : NULL;
			U1Head.Next = CurU;
			CurU = PrevU;
		    }
		    else
			PrevU = CurU;
		}
		PrevV->Next = CurV->Next;
		VfoundNum++;
		found = 1;
	    }
	}
	if (UfoundNum < pState->srcNum)
	{
	    PrevU = &U1Head;
	    for (CurU=U1Head.Next; CurU != NULL; CurU=CurU->Next)
	    {
		i = CurU->i;
		PrevV = &V0Head;
		for (CurV=V0Head.Next; CurV != NULL; CurV=CurV->Next)
		{
		    j = CurV->i;
		    if (pState->mask[i][j] == 1)
		    {
			CurV->val = pState->cost[i][j] - CurU->val;
			PrevV->Next = CurV->Next;
			CurV->Next = V1Head.Next != NULL ? V1Head.Next: NULL;
			V1Head.Next = CurV;
			CurV = PrevV;
		    }
		    else
			PrevV = CurV;
		}
		PrevU->Next = CurU->Next;
		UfoundNum++;
		found = 1;
	    }
	}
    }
    
    minDelta = (Float32) Inf;
    minI = minJ =0;
    for(i=0; i < pState->srcNum; i++)
    {
	for(j=0; j < pState->desNum; j++)
	{
	    if (pState->mask[i][j] == 0)
	    {
		delta = pState->cost[i][j] - U[i].val - V[j].val;
		if(minDelta > delta)
		{
		    minDelta = delta;
		    minI=i;
		    minJ=j;
		}
	    }
	}
    }
    if( minDelta == (Float32) Inf )
    {
	if( pState->srcNum > 1 && pState->desNum > 1)
	    ERR_GEN( minDelta == (Float32) Inf, cErrHeuristic_foundNothing );
    }

    *deltaMin = minDelta;
    pState->pointNode->i = minI;
    pState->pointNode->j = minJ;
   
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Private function to update variables.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::     State pointer
 *  pState::        Point to state
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   segmentation
 *
 * Keywords:::
 *   color, segmentation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_lpNewSol (
    IPWState          *pIPWState,
    DISTSTATE         *pState)
{
    PROLOGUE;
    Int32      i, j, k;
    Float32    xMin;
    Int32      steps;
    nodeComp  *Loop[2*MAXLEN], *CurX, *LeaveX;
 
    /* ENTER THE NEW BASIC VARIABLE */
    i = pState->pointNode->i;
    j = pState->pointNode->j;
    pState->mask[i][j] = 1;
    pState->pointNode->nextCol = pState->rowList[i];
    pState->pointNode->nextRow = pState->colList[j];
    pState->rowList[i] = pState->pointNode;
    pState->colList[j] = pState->pointNode;

    /* FIND A CHAIN REACTION */
    ERR_PASS (wp_lpFindLoop(pIPWState, pState, &Loop[0], &steps));

    /* FIND THE LARGEST VALUE IN THE LOOP */
    xMin = (Float32) Inf;
    LeaveX = NULL;
    for (k=1; k < steps; k+=2)
    {
	if (Loop[k]->val < xMin)
	{
	    LeaveX = Loop[k];
	    xMin = Loop[k]->val;
	}
    }

    /* UPDATE THE LOOP */
    for (k=0; k < steps; k+=2)
    {
	Loop[k]->val += xMin;
	Loop[k+1]->val -= xMin;
    }

    ERR_GEN (LeaveX == NULL, cErrHeuristic_foundNothing);

    /* REMOVE THE LEAVING BASIC VARIABLE */
    i = LeaveX->i;
    j = LeaveX->j;
    pState->mask[i][j] = 0;

    if (pState->rowList[i] == LeaveX )
    {
	pState->rowList[i] = LeaveX->nextCol;
    }
    else
    {
	for (CurX=pState->rowList[i]; CurX != NULL; CurX = CurX->nextCol)
	{
	    if (CurX->nextCol == LeaveX)
	    {
		CurX->nextCol = CurX->nextCol->nextCol;
		break;
	    }
	}
    }
    if (pState->colList[j] == LeaveX)
    {
	pState->colList[j] = LeaveX->nextRow;
    }
    else
    {
	for (CurX=pState->colList[j]; CurX != NULL; CurX = CurX->nextRow)
	{
	    if (CurX->nextRow == LeaveX)
	    {
		CurX->nextRow = CurX->nextRow->nextRow;
		break;
	    }
	}
    }

    /* SET _EnterX TO BE THE NEW EMPTY SLOT */
    pState->pointNode = LeaveX;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Private function to update variables.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  pState::         Point to state
 *  ppLoop::         Point to point of node list
 *
 * Output:::
 *  pSteps::         Point to steps
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   segmentation
 *
 * Keywords:::
 *   color, segmentation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_lpFindLoop (
    IPWState          *pIPWState,
    DISTSTATE         *pState,
    nodeComp         **ppLoop,
    Int32             *pSteps)
{
    PROLOGUE;
    Int32        i, steps, dis;
    nodeComp   **CurX, *NewX;
    Int8         IsUsed[2*MAXLEN]; 
 
    CurX = ppLoop;
    for (i=0; i < (pState->srcNum + pState->desNum); i++)
	IsUsed[i] = 0;

    NewX = *CurX = pState->pointNode;
    dis = pState->pointNode - pState->nodeList;
    IsUsed[dis] = 1;
    steps = 1;

    do
    {
	if (steps%2 == 1)
	{
	  /* FIND AN UNUSED X IN THE ROW */
	    NewX = pState->rowList[NewX->i];
	    while (NewX != NULL && IsUsed[NewX - pState->nodeList])
		NewX = NewX->nextCol;
	}
	else
	{
	  /* FIND AN UNUSED X IN THE COLUMN, OR THE ENTERING X */
	    NewX = pState->colList[NewX->j];
	    dis = NewX - pState->nodeList;
	    while (NewX != NULL && IsUsed[NewX - pState->nodeList] &&
		   NewX != pState->pointNode)
		NewX = NewX->nextRow;
	    if (NewX == pState->pointNode)
		break;
 	}

	if (NewX != NULL)  /* FOUND THE NEXT X */
	{
	    /* ADD X TO THE LOOP */
	    *++CurX = NewX;
	    dis = NewX - pState->nodeList;
	    IsUsed[dis] = 1;
	    steps++;
	}
	else  /* DIDN'T FIND THE NEXT X */
	{
	    /* BACKTRACK */
	    do
	    {
		NewX = *CurX;
		do 
		{
		    if (steps%2 == 1)
			NewX = NewX->nextRow;
		    else
			NewX = NewX->nextCol;
		} while (NewX != NULL && IsUsed[NewX-pState->nodeList]);
	     
		if (NewX == NULL)
		{
		    IsUsed[*CurX- pState->nodeList] = 0;
		    CurX--;
		    steps--;
		}
	    } while (NewX == NULL && CurX >= ppLoop);
	 
	    IsUsed[*CurX-pState->nodeList] = 0;
	    *CurX = NewX;
	    IsUsed[NewX-pState->nodeList] = 1;
	}     
    } while(CurX >= ppLoop);
  
    ERR_GEN (CurX == ppLoop, cErrHeuristic_foundNothing);
    *pSteps = steps;

    NORMAL_RETURN;
}

/*
 * $Log: dtrans.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 * Imported sources
 *
 * Revision 1.16  1998/11/05 00:42:40  xzhu
 * xzhu on Wed Nov 4 16:41:58 PST 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Use Eps instead of Inf.
 *
 * Revision 1.15  1998/10/28 23:07:52  xzhu
 * xzhu on Wed Oct 28 15:07:40 PST 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Fix mistyping bug.
 *
 * Revision 1.14  1998/10/02 16:27:34  xzhu
 * xzhu on Fri Oct 2 09:27:05 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Fixed NT waring.
 *
 * Revision 1.13  1998/10/01 17:01:07  xzhu
 * xzhu on Thu Oct 1 10:00:15 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Fixed bug and changed distance from l(1) to l(2).
 *
 * Revision 1.12  1998/09/16 19:15:46  xzhu
 * xzhu on Wed Sep 16 12:14:57 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Fixed bug.
 *
 * Revision 1.11  1998/09/09 22:26:22  xzhu
 * xzhu on Wed Sep 9 15:26:10 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Fixed bugs and added a new function.
 *
 * Revision 1.10  1998/09/03 18:47:08  xzhu
 * xzhu on Thu Sep 3 11:46:43 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Changed structure name and include ccsign.pub.
 *
 * Revision 1.9  1998/07/29 16:08:42  aber
 * aber on Wed Jul 29 09:08:24 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Typecast 0 to float32, for NT compiler warning.
 *
 * Revision 1.8  1998/07/24 20:51:36  xzhu
 * xzhu on Fri Jul 24 13:51:26 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Fixed bugs.
 *
 * Revision 1.7  1998/07/10 17:23:04  xzhu
 * xzhu on Fri Jul 10 10:22:43 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Fixed error reported by MAC (local data >32K).
 *
 * Revision 1.6  1998/07/09 16:36:08  xzhu
 * xzhu on Thu Jul 9 09:35:25 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Fixed complier warning.
 *
 * Revision 1.5  1998/07/08 23:39:00  xzhu
 * xzhu on Wed Jul 8 16:38:22 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Follow the change made in dtrans.pub.
 *
 * Revision 1.4  1998/07/08 18:41:02  xzhu
 * xzhu on Wed Jul 8 11:40:53 PDT 1998
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Added new functions.
 *
 * Revision 1.3  1998/01/14 00:11:21  ddavies
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.2  1997/04/15 21:46:43  ddavies
 * ddavies on Tue Apr 15 14:45:53 PDT 1997
 *
 * dtrans.c -> /product/ipcore/ipwrap/src/RCS/dtrans.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.1  1996/12/31 22:10:58  dferrell
 * Initial revision
 *
 *
 */

