/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* Data encoding functions
 *
 * This module provides all functions related to the encoding of
 * data, independent of layout and rendering information (although
 * generation of the key codeword necessarily requires some layout
 * information). There are no library entry points in this module,
 * since the encoding functions are, from the client's perspective,
 * transparent, being embedded within the layout and rendering
 * functions.
 *
 * Library entry points in this module:
 *
 *     None
 *
 * Internal entry points in this module:
 *
 *     wp_adjustEncoding()
 *     wp_calcEncoding()
 *     wp_encodeData()
 *     wp_getEncodingSize()
 *     wp_initEncoding()
 *
 * The functions in this module are organized alphabetically for easy
 * reference. Local (to this module) functions are defined first, then
 * internal (to the library) functions, then library entry points. Any
 * functions which are referenced before their definition within this
 * module have prototypes defined under "Forward References" below.
 *
 */

#define DATAGLYPH_VERSION1

#include <stdio.h>


#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "ipwerrs.pub"
#include "dgencdef.pub"
#include "imageref.pub"


#ifdef GLY_INFO
#define TIMING 1
#include <time.h>
#include "timing.h"
#endif

#include "shrutils.pub"
#include "dgcom.prv"
#include "dgencint.prv"
#include "dgecc.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif



IP_RCSINFO(RCSInfo, "$RCSfile: encode.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

/******************* Temporary debugging functions ********************/

#if 0

FILE * dumpfile = NULL;

static void OpenDumpFile(void)
{
	dumpfile = fopen("codeword.hex", "w");
}

static void DumpBytes(UInt8 * pBytes, UInt32 nLength)
{
	unsigned char asciiData[17];
	char *hc = "0123456789ABCDEF";
	int i, n;
	unsigned char c;

	asciiData[16] = '\n';
	asciiData[17] = '\0';
	while (nLength > 0) {
		n = MIN(nLength, 16);
		for (i = 0; i < n; i++) {
			c = *pBytes;
			fputc(hc[c >> 4], dumpfile);
			fputc(hc[c & 0x0F], dumpfile);
			if (i & 1) fputc(' ', dumpfile);
			asciiData[i] = (unsigned char)(isprint(c) ? c : '.');
			pBytes++;
			};
		if (n != 16) {
			for (i = 0; i < (40-n*5/2); i++) fputc(' ', dumpfile);
			asciiData[n] = '\n';
			asciiData[n+1] = '\0';
			};
		fputs("   ", dumpfile);
		fputs((char *)asciiData, dumpfile);
		nLength -= n;
		};
	fputs("\n", dumpfile);
}

static void CloseDumpFile(void)
{
	fclose(dumpfile);
}
#endif

/************************** Local functions ***************************/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Calculates the number and size of codewords to be used for data.
 *
 * LongDescription:::
 * The size is maximised, but must be less than MAX_CODEWORD_SIZE
 * (which is itself derived from the symbol size (i.e. byte) being
 * used for ECC coding). The number of parity bytes to be added in
 * each codeword is also calculated.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  enc::		encoding handle
 *
 * Notes:::
 *  The nCodewords, nCodewordSize, and nCodewordParity fields of the
 *  encoding handle are updated.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper	initial version
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_calcCodewordSize (
    IPWState    *pIPWState,
    PEncoding	enc)
{
    int nRequired, nCodewords, nCodewordSize, nBytesDataAndCRC;
    float parityFraction;

	/* keep compiler happy */
    pIPWState = pIPWState;

	/*
	 * Calculate total data + parity bytes, ensuring at least one
	 *	byte of parity for one byte of data.
	 */
    parityFraction =  (float)enc->eccLevel / 100;
    nRequired = MAX((int)((float)enc->nDataBytes / (1.0f - parityFraction)), 2);
	/* code called later e.g. wp_calcParity requires that there be at
 	 * least one byte of ecc.  there may be other places as well and
	 * instead of trying to find them, we'll just make sure that 
	 * there is at least one byte of ecc.  in addition, the code
	 * above sometimes gives no ecc even though the user asked for
	 * some.  for example, consider 2 data bytes with an ecc level
	 * of 25 percent.
	 */
    if (nRequired == enc->nDataBytes)
	nRequired++;

	/*
	 * Calculate number of codewords, each of size MAX_CODEWORD_SIZE
	 * or less, and including data, CRC and parity.
	 */
    nCodewords = (nRequired - 1) / MAX_CODEWORD_SIZE;
    do {
	    nCodewords++;
	    nCodewordSize = DIVUP(nRequired + nCodewords * CRC_SIZE, nCodewords);
    } while (nCodewordSize > MAX_CODEWORD_SIZE);
    enc->nCodewords = nCodewords;
    enc->nCodewordSize = nCodewordSize;

	/*
	 * Calculate the number of parity bytes per codeword.
	 */
    nBytesDataAndCRC =
	    (int)((enc->nDataBytes + nCodewords * CRC_SIZE + nCodewords - 1) /
	    nCodewords);
    enc->nCodewordParity = nCodewordSize - nBytesDataAndCRC;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Calculates the combined size of the key codeword including CRC and ecc
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *
 * Returns:::
 *  size of entire key codeword
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
UInt32 CDECL 
wp_calcHeaderSize (
    IPWState   *pIPWState)
{
    /* Make compiler happy */
    pIPWState = pIPWState;

    return(SIZEOF_KEYCODEWORD + CRC_SIZE + KEY_PARITY_SIZE);
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Generates ecc bytes for a codeword, using Reed-Solomon encoding
 *
 * PrivateDescription:::
 * A codeword (polynomial) is the concatenation of data bits and check
 * symbols.
 * First, we start with the data polynomial, D(x).  The data polynomial
 * can be defined in terms of a quotient and remainder:
 * D(x) = q(x) * g(x) + r(x) where q(x) is a quotient polynomial,
 * g(x) is the generator polynomial, and r(x) is the remainder polynomial.
 * Rearranging terms yields: ( D(x) + r(x) ) / g(x) = q(x).
 * This means that in order for the data polynomial to be divisible by
 * g(x), r(x) would have to be included somehow.  We do not wish to
 * modifiy D(x) so we premulitply D(x) by x^m so that r(x) (the check
 * symbols) can be or'd with D(x).
 * So, the codeword polynomial C(x) = D(x) * x^m + ( x^m * D(x) mod g(x) )
 * And, with the codeword generated in this manner, C(x) mod g(x) = 0.
 * The decoder can check the received codeword against the generator 
 * polynomial and determine if errors exist.
 * This can be implemented with a shift register that premultiplies by
 * x^m and divides by g(x).  This shift register is implemented in the
 * code below. For each data symbol, all the elements of the shift
 * register are recalculated.  First, the data symbol is exclusive or'd
 * with the highest shift register element.  This is 
 * multiplied with each of the coefficients of the generator polynomial.
 * Each shift register element is set to the contents of the previous
 * shift register element exclusive or'd with the results of the
 * multiplication of its corresponding coefficient.
 * Each shift register element corresponds to a check symbol.
 * How do galois fields fit into this?  Well each coefficient of the
 * polynomial is an element of the galois field.  
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  glog::		log table used by the galois code
 *  gexp::		antilog table used by the galois code
 *  logGenpoly::	table of log of coeffs of generator polynomial
 *  pData::		data bytes
 *  nBytesData::   	number of data bytes in each codeword  
 *  nBytesParity::   	number of bytes of ecc in each codeword
 *
 * Output:::
 *  pParity::           ecc bytes
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 **************************************************************************/
static void CDECL 
wp_calcParity (
    IPWState   *pIPWState,
    UInt32     *glog,
    UInt8      *gexp,
    UInt32     *logGenpoly,
    UInt8      *pData,
    Int32	nBytesData,
    UInt8      *pParity,
    Int		nBytesParity)
{
    UInt8	ecc[((MAX_CODEWORD_SIZE + 3)/4) * 4];
    Int32	i, j;		/* general indices */
    Int32	firstECCIndex;	/* index of rightmost word we use
				 * in ecc array */
    Int32	firstPBIndex;	/* the name given to any particular
				 * parity byte in the ecc array changes
				 * as explained below.  This is the
				 * name of of ecc[0].
				 */
    UInt32     *pLogGenpoly;	/* ptr to current logGenpoly byte. */
    UInt32	logTmp;		/* log(currentDataByte ^ lastParityByte).
				 * This is a UInt32 just so the compiler
				 * doesn't bother to try to toss the
				 * MS bit after adding logTmp to the
				 * log(lastParityByte). */
    UInt32	accumPoly;	/* 4 bytes; contains
				 * gexp[logTmp + logGenpoly[j]] for
				 * 4 consecutive values of j */

	/* keep compiler happy */
    pIPWState = pIPWState;

    memset((void *)ecc, 0, sizeof(*ecc) * MAX_CODEWORD_SIZE);
	/* This is the byte index of the rightmost word in the
	 * ecc array.  This is where each loop starts processing.
	 * We process from right to left.
	 */
    firstECCIndex = (nBytesParity - 1) & ~0x3;

	/* the first ecc bytes is number 0 when we begin */
    firstPBIndex = 0;
    for (i = 0; i < nBytesData; i++)
    {
	    /* the old code was pretty easy to understand, but not
	     * super fast.  It had several causes of slowness:
	     *   1.	The only thing the old loop did with tmp was multiply
	     *	genpoly[j] by it.  Multiplying is done by
	     *	adding the logarithms of the terms and taking the
	     *	exponent of the sum.  The old code computed the
	     *	log of tmp every time.  Since glog(tmp) doesn't
	     *	change, we just compute it once.
	     *   2.	Since the only thing we do with genpoly is multiply
	     *	by it, we store glog[genpoly], not genpoly itself.
	     *	That's whats passed in as logGenpoly.
	     *   3. The multiply was done using a subroutine call.
	     *	We should eliminate that.
	     *   4. It's awfully nice to do things 4 bytes at a time
	     *	instead of 1 byte at a time.  We could think about
	     *	doing all the multiplies in units of 4 bytes, but
	     *	there are carries from one byte to the next.  So,
	     *	we could do them 2 or 3 bytes at a time.  But then
	     *	we'd still have to take the 2 or 3 numbers apart
	     *	to calculate the exponent of each one.  That's a
	     *	pain.  Instead, we'll calculate each multiply
	     *	individually and cram the results into a vector
	     *	of four bytes.  We'll do the Galois add (XOR to us)
	     *	with the vector, so will retire 4 bytes at once.
	     *   5. Then there's that shift register thing.  The
	     *	algorithm shifts the parity bytes after/during
	     *	the Galois add.  Shifting the 4-byte vector is
	     *	a pain.  To compute word-aligned values, you have
	     *	to do left and right shifts and ORs.  Yuck.
	     *	Instead, we'll just change the names of the
	     *	ecc bytes.  What?  Say there are 6 parity
	     *	bytes.  Initially, they are numbered
	     *	  ecc array loc		0 1 2 3 4 5
	     *    holds parity byte num 0 1 2 3 4 5
	     *	After processing the first data byte, the same
	     *	byte positions (the contents have obviously
	     *	changed since that's what processing the first
	     *	data byte does) are named
	     *	  ecc array loc		0 1 2 3 4 5
	     *    holds parity byte num 1 2 3 4 5 0
	     *	That's because we pretend that we've shifted
	     *	towards the bigger numbers.  When we fetch
	     *	the MS byte at the beginning, we always fetch
	     *	the byte currently numbered "5".
	     *	We always compute 4 parity bytes at a time.
	     *	Yeah, but what happens when we start out
	     *	with something like:
	     *	  ecc array loc		0 1 2 3 4 5
	     *    holds parity byte num 5 0 1 2 3 4 
	     *	Well, we have to copy the first three words
	     *	of logGenpoly to the end of logGenpoly.  This
	     *	guarantees that if you start with the last
	     *	byte and process 4 bytes, you still get the
	     *	right four bytes.
	     *	Ok, but what about byte order?  The codewords
	     *	will be accessed a byte at a time.  Can we
	     *	update the parity bytes a word at a time?
	     *	Yes, if we're careful.  We still access the
	     *	logGenpoly words one at a time.  When composing
	     *	a vector of four feedback bytes to xor with
	     *	the existing parity bytes, we need to remember
	     *	byte order.  On big endian machines, the lower
	     *	numbered bytes are on the left.  On little
	     *	endian machines, the lower numbered bytes are
	     *	on the right.
	     *   6. And, we really want this to go fast on
	     *	x86 platforms.  That's the main platform of
	     *	the SmartPaper toolkit and will probably be the
	     *	platform for Taos/DAE.  So, we try to encourage
	     *	the compiler to generate good x86 code.
	     *  We do this by calculating two bytes in the
	     *	same statement, corresponding to the two
	     *	accessible bytes in an x86 32-bit register,
	     *	then shift by 16 and calculate two more bytes.
	     */

	    /* Find the ecc byte currently holding parity byte
	     * nBytesParity-1, or "5" in the example above.  At the beginning
	     * of time, this is ecc[nBytesParity-1].  After processing
	     * one data byte, which entails one fake shift,
	     * this is byte ecc[nBytesParity-2] and so on.
	     * The first ecc byte (ecc[0]) currently has the
	     * name in firstPBIndex.  Therefore
	     * the ecc byte with index (nBytesParity - 1 - firstPBIndex)
	     * holds parity byte (nBytesParity-1).
	     */
	    /* calculate the term we'll combine with every parity byte.
	     * This is combines the last current parity byte (which
	     * will be shifted off the end and the current data byte.
	     */
	logTmp = glog[pData[i] ^ ecc[nBytesParity - 1 - firstPBIndex]];

	    /* Point to the logGenpoly entry which corresponds to
	     * the byte held in ecc[firstECCIndex] this time through
	     * the loop.
	     */
	if (firstPBIndex + firstECCIndex < nBytesParity)
	    pLogGenpoly = &(logGenpoly[firstPBIndex + firstECCIndex]);
	else
	    pLogGenpoly = &(logGenpoly[firstPBIndex + firstECCIndex -
					nBytesParity]);

	    /* Process all the parity bytes in a loop that does
	     * four bytes at a time.  Process from high indices
	     * to low ones to make the loop sligtly more efficient.
	     */
	for (j = firstECCIndex; j >= 0; j -= 4)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

		/* do the multiplies noting that the quantities in
		 * logGenpoly[j] are already in exponent form.
		 * HEY!  We want to combine ecc[j-1] with
		 * a product of genpoly[j] to form ecc[j].
		 * If currentPBIndex = 0, then the first
		 * byte in the word is now 0, but will become
		 * byte 1.  That means that we are calculating
		 * byte 1, not byte 0!  Therefore, we should
		 * be reading from genpoly[currentPBIndex+1]
		 * for the first byte, not genpoly[currentPBIndex].
		 */
	    accumPoly = (gexp[pLogGenpoly[4] + logTmp] << 8) |
			(gexp[pLogGenpoly[3] + logTmp]);
	    accumPoly <<= 16;
	    accumPoly |= (gexp[pLogGenpoly[2] + logTmp] << 8) |
			 (gexp[pLogGenpoly[1] + logTmp]);

#elif _ALPACA_IMAGE_FMT_ == cBigEndianFmt

	    accumPoly = (gexp[pLogGenpoly[1] + logTmp] << 8) |
			(gexp[pLogGenpoly[2] + logTmp]);
	    accumPoly <<= 16;
	    accumPoly |= (gexp[pLogGenpoly[3] + logTmp] << 8) |
			 (gexp[pLogGenpoly[4] + logTmp]);

#endif

		/* Combine these four bytes with current four
		 * ecc bytes.
		 */
	    ULONG_ACCESS(&ecc[j]) = ULONG_ACCESS(&ecc[j]) ^ accumPoly;

		/* move pLogGenpoly right by 4 mod nBytesParity and continue */
	    pLogGenpoly -= 4;
	    if (pLogGenpoly < logGenpoly)
		pLogGenpoly += nBytesParity;
	}

	    /* We've now processed one more data byte.  The ecc
	     * array byte that used to have the highest name
	     * (it was referenced at the top of the loop above)
	     * now has the lowest name (0).  The data put into this location
	     * was incorrect since it was calculated using
	     * data from the other highest named parity byte
	     * instead of using no parity byte data.  So fix it.
	     */
	ecc[nBytesParity - 1 - firstPBIndex] = gexp[logGenpoly[0] + logTmp];

	    /* Now advance firstPBIndex.  Since we're pretending to
	     * have put data from ecc[x] into ecc[x+1] and we really
	     * didn't move the data, the name assigned to ecc[0]
	     * should increment by 1 MOD nBytesParity.  E.G. if ecc[0]
	     * used to hold parity byte 4, it now holds parity
	     * byte 5.  If there were only 5 parity bytes, it
	     * holds parity byte 0.
	     */
	firstPBIndex++;
	if (firstPBIndex >= nBytesParity)
	    firstPBIndex = 0;

    }

	/* Now we copy the ecc bytes, which are out of order, into
	 * the parity bytes.  Not only are they out of order because
	 * of the fake shifting we've been doing, they're out of
	 * order because they should be reversed in the codeword.
	 * Why?  Convention, I guess.  Anyway, we start the ecc
	 * index at (nParityBytes - 1 - firstPBIndex) since that's
	 * the ecc byte that holds the last parity bytes and
	 * move the ecc index down MOD nParityBytes.
	 */
    for (i = 0, j = nBytesParity - 1 - firstPBIndex;
	 i < nBytesParity;
	 i++)
    {
	pParity[i] = ecc[j];
	j--;
	if (j < 0)
	    j = nBytesParity - 1;
    }



#if 0	/* how we used to do it */

	/* get the index of the highest order shift register element */
	j = nBytesParity - 1; 
	/* premultiply by x^nBytesParity */
	tmp = pData[i] ^ ecc[j];
	/* now compute the value of each shift register element. */
	for (; j > 0; j--)
	    ecc[j] = ecc[j-1] ^ ERR_IGNORE(wp_gMultVal(pIPWState,
						       genpoly[j], tmp,
						       glog, gexp));
	ecc[0] = ERR_IGNORE(wp_gMultVal(pIPWState, genpoly[0],
					tmp, glog, gexp));
    }

    /* and move them into place in the codeword.  note this could
     * be done in the loop above.
     */
    for (i = 0; i < nBytesParity; i++)
	pParity[i] = (UInt8)ecc[nBytesParity - 1 - i];
#endif

}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Divides the data into codewords, and adds CRC and ecc to each codeword
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::		region handle
 *  pData::         	data to be encoded 
 *  nBytes::        	number of bytes of data 
 *
 * Output:::
 *  pCodewords::     	encoded data
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_generateCodewords (
    IPWState   *pIPWState,
    PDGRegion	pRegion,
    UInt8      *pData,
    UInt32	nBytes,
    UInt8      *pCodewords)
{
    UInt32 	 nDataSize, nCodeword;
    UInt16 	 crc;
    UInt8	*pRandom = NULL;
    UInt32	 randomOffset = 0;
#ifdef GLY_INFO
    Timer	 time;
#endif

#ifdef GLY_INFO
    /* start the timer */
    START_TIME(time);
#endif /* GLY_INFO */

     /* Calculate the number of data bytes per codeword. */
     nDataSize = pRegion->nCodewordSize -
		(CRC_SIZE + pRegion->nCodewordParity);

    /* we don't want codewords with only 0 in them since it makes the
     * synch stand out.  so get some random bytes if necessary
     */
    pRandom = pRegion->pRandom;

    /* Generate the codewords. */
    for (nCodeword = 0; nCodeword < pRegion->nCodewords; nCodeword++) 
    {

	/* Copy the data to the codeword */
	if (nBytes < nDataSize)
	{
	    /* copy pseudo-random bytes into data bytes that the user
	     * did not supply bytes for.
	     */
	    memcpy((void *)pCodewords, (void *)pData, nBytes);
	    /* note that the table of random entries is not infinite so
	     * make sure that we don't access off the end.
	     */
	    memcpy((void *)(pCodewords + nBytes), 
		   (void *)(pRandom + (randomOffset & RTABLE_MASK)), 
		   nDataSize - nBytes);
	    randomOffset += (nDataSize - nBytes);
	    nBytes = 0;
	}
	else
	{
	    memcpy((void *)pCodewords, (void *)pData, nDataSize);
	    nBytes -= nDataSize;
	}

	crc = ERR_IGNORE(wp_calcCRC(pIPWState, pCodewords, nDataSize));
	ERR_IGNORE(wp_storeCRC(pIPWState, (UInt32)crc, pCodewords + nDataSize));

	/* Add the parity */
	ERR_IGNORE(wp_calcParity(pIPWState, pRegion->glog, pRegion->gexp,
				 (UInt32 *)GET_LOG_DATA_GENPOLY(pRegion),
			      	 pCodewords, nDataSize + CRC_SIZE,
			      	 pCodewords + nDataSize + CRC_SIZE,
				 pRegion->nCodewordParity));
#if 0
	/* Dump the debugging data */
	if (getenv("DUMP_CODEWORDS"))
	    DumpBytes(pCodewords, pRegion->nCodewordSize);
#endif

	/* Increment the pointers */
	pCodewords += pRegion->nCodewordSize;
	pData += nDataSize;
    }


#ifdef GLY_INFO
    ELAPSED_TIME(time);
    if (getenv("EDB_PRINT_TIME"))
    {
	PRINT_ELAPSED(pRegion->errorProc, pRegion->errorInfo, 
		      "It took %3.3f seconds to calculate codewords\n", 
		      time);
    }
#endif /* GLY_INFO */
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Generates the key codeword, including CRC and parity bytes.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::        	region handle
 *  nBytes::		number of bytes of client data
 *
 * Output:::
 *  pCodeword::      	encoded key codeword
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_generateHeader (
   IPWState    *pIPWState,
   PDGRegion 	pRegion,
   UInt32	nBytes,
   UInt8       *pCodeword)
{
    UInt8 * pEncData = pCodeword;
    KeyCodeword keyCodeword;
    UInt16 crc;

    /* Set up the key codeword data itself */
    keyCodeword.nVersion = DG_LIBRARY_VERSION;
    keyCodeword.nCodewordSize = (UInt8)pRegion->nCodewordSize;
    keyCodeword.nCodewordParity = (UInt8)pRegion->nCodewordParity;
    keyCodeword.occlusionType = OCC_NONE;
    ERR_IGNORE(wp_store32(pIPWState, nBytes,
			  (UInt8 *)&(keyCodeword.nDataBytes)));
    memcpy((void *)pEncData, (void *)&keyCodeword, SIZEOF_KEYCODEWORD);
    pEncData += SIZEOF_KEYCODEWORD;

    /* Add the CRC */
    crc = ERR_IGNORE(wp_calcCRC(pIPWState, pCodeword, SIZEOF_KEYCODEWORD));
    ERR_IGNORE(wp_storeCRC(pIPWState, (UInt32)crc, pEncData));
    pEncData += CRC_SIZE;

    /* Add the parity */
    ERR_IGNORE(wp_calcParity(pIPWState, pRegion->glog, pRegion->gexp,
			     pRegion->logKeyGenpoly, pCodeword, 
			     SIZEOF_KEYCODEWORD + CRC_SIZE,
			     pEncData, KEY_PARITY_SIZE));
    pEncData += KEY_PARITY_SIZE;

#if 0
    /* Dump the debugging data */
    if (getenv("DUMP_CODEWORDS"))
	    DumpBytes(pCodeword,
		      SIZEOF_KEYCODEWORD + CRC_SIZE + KEY_PARITY_SIZE);
#endif

#if 0
    /* If there's no occlusion, we're done */
    if (keyCodeword.occlusionType == OCC_NONE)
	    return;

    /* Set up the occlusion codeword data itself */
    pCodeword = pEncData;
    memcpy((void *)pEncData, (void *)&(pLayout->rcOcclusion), SIZEOF_ROCODEWORD);
    pEncData += SIZEOF_ROCODEWORD;

    /* Add the CRC */
    crc = ERR_IGNORE(wp_calcCRC(pIPWState, pCodeword, SIZEOF_ROCODEWORD));
    ERR_IGNORE(wp_storeCRC(pIPWState, (UInt32)crc, pEncData));
    pEncData += CRC_SIZE;

    /* Add the parity */
    ERR_IGNORE(w_initGaloisTables(pIPWState, RO_PARITY_SIZE));
    ERR_IGNORE(wp_calcParity(pIPWState, pCodeword, SIZEOF_ROCODEWORD + CRC_SIZE,
			     pEncData, RO_PARITY_SIZE));
#endif

#if 0
    /* Dump the debugging data */
    if (getenv("DUMP_CODEWORDS"))
	    DumpBytes(pCodeword, SIZEOF_ROCODEWORD + CRC_SIZE + RO_PARITY_SIZE);
#endif
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  "Randomize" the encoded data before it is placed into the rectangle
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::        	region handle
 *  nTotalBytes::    	number of bytes to randomize
 *
 * Output:::
 *  pEncData::       	data to be randomized
 *
 * Returns:::
 *  cSuccessful		successfully randomized region
 *  eErrNomem		unable to allocate memory
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL 
wp_randomizeData (
    IPWState   *pIPWState,
    PDGRegion	pRegion,
    UInt8      *pEncData,
    UInt32	nTotalBytes)
{
    PROLOGUE;
    UInt32	i;
    UInt32	newPosition;
    UInt32	*rTable = NULL;
    UInt8	 temp;

    ERR_PASS(wp_generateRandomizeTable(pIPWState, pRegion->mallocProc,
				       pRegion->freeProc, 
				       pRegion->allocInfo,
				       &rTable, nTotalBytes, 
				       DG_RANDOM_SEED));
	
    for (i = 0; i < nTotalBytes; i++)
    {
	newPosition = rTable[i];
	temp = *(pEncData+newPosition);
	*(pEncData+newPosition) = *(pEncData+i);
	*(pEncData+i) = temp;
    }
    GLY_FREE(pRegion->freeProc, pRegion->allocInfo, rTable);
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Stores a 32-bit value in a byte stream
 *
 * LongDescription:::
 * Stores a 32-bit value in a byte stream, without making assumptions
 * about the alignment of the 32-bit location in the byte stream.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  val::    		value to be stored        
 *
 * Output:::
 *  data::         	byte stream  
 *
 * Notes:::
 * This function is written in such a way that the same byte stream order
 * is generated on both big-endian and little-endian systems.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL 
wp_store32 (
    IPWState   *pIPWState,
    UInt32	val,
    UInt8      *data)
{
	/* Keep compiler happy */
    pIPWState = pIPWState;

    data[0] = (UInt8)((val & 0xFF000000) >> 24);
    data[1] = (UInt8)((val & 0x00FF0000) >> 16);
    data[2] = (UInt8)((val & 0x0000FF00) >> 8);
    data[3] = (UInt8)(val & 0x000000FF);
}

/************************* Internal functions *************************/

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Recomputes the codeword size
 *
 * LongDescription:::
 * This procedure will be used either when calculating the capacity
 * of a rectangle or when adjusing the ecc level in order to make
 * better use of available space in the layout.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  selfDescribing::	is this a self-describing rectangle?
 *  nBytesAvailable::	total number of bytes in the glyph rectangle
 *  nDataBytes::	used when adjusting the eccLevel, number of data
 *			bytes to be placed in the rectangle.
 *			if 0, indicates that capacity should be calculated.
 *  eccLevel::		used when calulating capacity, ecc level
 *			requested by client
 *
 * Output:::
 *  numCodewords::	number of codewords in the rectangle
 *  codewordSize::	total number of bytes in each codeword
 *  codewordECC::	number of ecc bytes in each codeword
 *  codewordData::	number of data bytes in each codeword
 *
 * Note:::
 * Either nDataBytes or eccLevel should be non-zero before calling.
 * (However, this procedure does not check the values before attempting
 * to use them.) 
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper	initial version
 *  Julie Craig		try to keep number of ecc bytes less than 50
 *  Julie Craig		allow usage by capacity calculations
 *
 * EndDocumentation:::
 **************************************************************************/
void CDECL 
wp_calcEncoding (
    IPWState   *pIPWState,
    Bool	selfDescribing,
    UInt32	nBytesAvailable,
    UInt32	nDataBytes,
    UInt8	eccLevel,
    UInt32     *numCodewords,
    UInt8      *codewordSize,
    UInt8      *codewordECC,
    UInt8      *codewordData)
{
    UInt32	 nBytesDataAndCRC;
    UInt32	 nCodewords;
    UInt32	 nCodewordSize;
    UInt32	 nCodewordParity;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    /* initial guess of codeword size.  the loop below
     * always does an increment on the number of codewords.
     */
    if (selfDescribing == cTrue)
	nBytesAvailable -= ERR_IGNORE(wp_calcHeaderSize(pIPWState));
    nCodewords = (nBytesAvailable - 1) / MAX_CODEWORD_SIZE;

    /* what's with this loop with the 50 magic number?  well,
     * if the number of parity bytes is too large the amount of
     * time to calculate the tables need to encode (and decode)
     * is noticeable.  as a result we need to keep the number of
     * ecc bytes kinda small?  what is kinda small?  well, we should 
     * calculate the curve but since we haven't pick a number that
     * seems to do ok.
     */
    do 
    {
	do
	{
	    nCodewords++;
	    nCodewordSize = nBytesAvailable / nCodewords;
	} while (nCodewordSize > MAX_CODEWORD_SIZE);
	if (nDataBytes == 0)
	{
	    /* this calculation could be done without floating point
 	     * but I wanted to mirror the calc done in wp_calcCodewordSize
	     * initially.
	     */
	    nBytesDataAndCRC = 
		(UInt32)(nCodewordSize * 
			 (1.0f - ((float)eccLevel / 100.0f)));
	    nCodewordParity = nCodewordSize - nBytesDataAndCRC;
	    /* the code in wp_calcCodewordSize guarantees at least one
	     * byte of ecc.  make sure that we have at least one byte
	     * here as well.
	     */
	    if (nCodewordParity == 0)
		nCodewordParity++;
	}
	else
	{
	    nBytesDataAndCRC =
		DIVUP(nDataBytes + nCodewords * CRC_SIZE,
		      nCodewords);
	    nCodewordParity = nCodewordSize - nBytesDataAndCRC;
	}

	/* if the number of data bytes is small just give it up */
	if ((nCodewordSize - nCodewordParity) < 5)
	    break;
    } while (nCodewordParity > 50);

    *numCodewords = nCodewords;
    *codewordSize = (UInt8)nCodewordSize;
    *codewordECC = (UInt8)nCodewordParity;
    *codewordData = *codewordSize - *codewordECC - CRC_SIZE;
}

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Recomputes the codeword size
 *
 * LongDescription:::
 * Recomputes the codeword size, increasing the ecc bytes in
 * order to make better use of available space in the layout.
 *
 * PrivateDescription:::
 * This function is very similar to wp_calcCodewordSize, below, the
 * difference being that this function attempts to utilise as many
 * of the available bytes as possible by increasing the parity
 * fraction, whereas wp_calcCodewordSize uses the parity fraction
 * specified by the client.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pEncoding::     	encoding handle 
 *  nBytesAvailable::	total number of bytes in the glyph rectangle
 *
 * Notes:::
 * This function sets the following fields of the encoding structure
 * passed as a parameter:
 *
 *     nCodewords		number of codewords
 *     nCodewordSize		size of each codeword
 *     nCodewordParity		parity bytes for each codeword
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper	initial version
 *  Julie Craig		try to keep number of ecc bytes less than 50
 *
 * EndDocumentation:::
 **************************************************************************/
void CDECL 
wp_adjustEncoding (
    IPWState   *pIPWState,
    PEncoding  pEncoding,
    UInt32     nBytesAvailable)
{
    UInt8	codewordData;

    ERR_IGNORE(wp_calcEncoding(pIPWState,
			       pEncoding->selfDescribing, nBytesAvailable,
			       pEncoding->nDataBytes, pEncoding->eccLevel,
			       &pEncoding->nCodewords, 
			       &pEncoding->nCodewordSize,
			       &pEncoding->nCodewordParity, 
			       &codewordData));
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Creates an encoded byte sequence from the client-supplied data.
 *
 * LongDescription:::
 * The encoded byte sequence is composed of a data header, the
 * client-supplied data divided into codewords with CRC and parity added
 * to each codeword.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::        	region handle
 *  pData::          	client-supplied data
 *  nBytes::         	number of bytes of client data
 *
 * Returns:::
 *  cSuccessful		successfully encoded data
 *
 * Notes:::
 *  encodedData inside region handle computed
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
wp_encodeData (
    IPWState   *pIPWState,
    PDGRegion	pRegion,
    UInt8      *pData,
    UInt32	nBytes)
{
    PROLOGUE;
    UInt8 * pEncData;
#ifdef GLY_INFO
    Timer	 time;
#endif


#if 0
    /* Open the debugging file */
    if (getenv("DUMP_CODEWORDS"))
	    OpenDumpFile();
#endif

    /* Ensure that any previously encoded data is cleared */
    pEncData = GET_ENCODED_DATA(pRegion);
    memset((void *)pEncData, 0, (int)pRegion->nEncodedBytes);

    /* encode the key codeword, if necessary and the data */
    if (pRegion->selfDescribing == cTrue)
    {
	    /* Encode the header */
	    ERR_IGNORE(wp_generateHeader(pIPWState, pRegion, nBytes, pEncData));

	    /* Encode the data */
	    ERR_IGNORE(wp_generateCodewords(pIPWState, pRegion, pData, nBytes, 
		pEncData + ERR_IGNORE(wp_calcHeaderSize(pIPWState))));
    }
    else
	    /* Encode only the data */
	    ERR_IGNORE(wp_generateCodewords(pIPWState, pRegion, pData, nBytes,
					    pEncData));

#ifdef GLY_INFO
    /* start the timer */
    START_TIME(time);
#endif /* GLY_INFO */

    /* Randomize */
    ERR_PASS(wp_randomizeData(pIPWState, pRegion, pEncData,
			      pRegion->nEncodedBytes));

#ifdef GLY_INFO
    ELAPSED_TIME(time);
    if (getenv("EDB_PRINT_TIME"))
    {
	PRINT_ELAPSED(pRegion->errorProc, pRegion->errorInfo, 
		      "It took %3.3f seconds to randomize the encode data\n", 
		      time);
    }
#endif /* GLY_INFO */

#if 0
	/* Close the debugging file */
	if (getenv("DUMP_CODEWORDS"))
		CloseDumpFile();
#endif

	NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Calculates the total size (in bytes) of the encoded data stream
 *
 * LongDescription:::
 * Calculates the total size (in bytes) of the encoding, i.e. the data
 * header and the codewords, including all CRC and parity.
 * Calculation is based on the 
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pEncoding::      	encoding handle
 *
 * Output:::
 *  pSize::          	size of the encoded data
 *
 * Notes:::
 * This function sets the following fields of the encoding structure
 * passed as a parameter:
 *
 *     nCodewords		number of codewords
 *     nCodewordSize		size of each codeword
 *     nCodewordParity		parity bytes for each codeword
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
wp_getEncodingSize (
    IPWState    *pIPWState,
    PEncoding 	pEncoding,
    UInt32     *pSize)
{
    PROLOGUE;

    ERR_GEN(pEncoding->nDataBytes == 0, cErrDG_InvalidByteCount);
    ERR_IGNORE(wp_calcCodewordSize(pIPWState, pEncoding));
    if (pEncoding->selfDescribing == cTrue)
	*pSize = ERR_IGNORE(wp_calcHeaderSize(pIPWState)) +
	         pEncoding->nCodewords * pEncoding->nCodewordSize;
    else
	*pSize = pEncoding->nCodewords * pEncoding->nCodewordSize;
    NORMAL_RETURN;
}

/**************************** End of file *****************************/

/*
 $Log: encode.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.51  1998/01/14 00:11:22  ddavies
 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.50  1997/01/20 21:14:47  jcraig
 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 use number of bytes from client's call to w_setData when recording
 number of bytes in the key codeword.

 Revision 1.49  1996/03/24 21:16:24  ddavies
 ddavies on Sun Mar 24 13:16:10 PST 1996

 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 Fix auto-doc comments.

 Revision 1.48  1996/03/21 21:35:13  ddavies
 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 Improved the algorithm used to generate ECC parity bytes.  It's now about
 3x faster.  Also deleted wp_encryptData since it's never used.  Added
 optional timing calculations.

 Revision 1.47  1996/03/08 17:36:42  jcraig
 jcraig on Fri Mar 8 09:36:22 PST 1996

 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 add comments to wp_calcParity

 Revision 1.46  1996/03/01 00:05:07  ddavies
 ddavies on Thu Feb 29 16:05:07 PST 1996

 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 CommonCheck: Change all w_o_ and wp_o_ names to the original w_ and wp_ versions.  This
 leaves the original names associated with the original routines in the
 grand glyph toolkit update.

 Revision 1.45  1996/02/22 04:01:06  ddavies
 ddavies on Wed Feb 21 20:00:16 PST 1996

 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 Make wp_calcHeaderSize into a public procedure so wp_doComputeCapacity
 can use it.

 Revision 1.44  1996/02/13 21:09:20  ddavies
 ddavies on Tue Feb 13 13:07:46 PST 1996

 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 Add inclusion of new layout creation include files, dglaydef.pub and dglayout.pub.
 Also change names of all existing encoding routines by concatenating o_
 to all the prefixes.

 Revision 1.43  1996/01/04 20:22:03  ejaquith
 encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.42  1995/12/18  22:59:55  jcraig
 * jcraig on Mon Dec 18 14:58:07 PST 1995
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * in wp_generateCodewords, if the user data does not completely fill
 * the data needed, use a 256 byte table of generated data instead so
 * that the rectangle will not be filled with 0 (which makes the synch
 * stand out).
 *
 * Revision 1.41  1995/12/13  04:06:19  ejaquith
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.40  1995/12/12  18:47:59  jcraig
 * jcraig on Tue Dec 12 10:45:36 PST 1995
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * UInt => UInt32
 * raise cErrDG_InvalidByteCount if number of bytes have not been set
 * when trying to compute codeword size.
 * remove wp_initEncoding which is no longer used (since ecc level and
 * number of bytes to encode are not longer set when layout handle is
 * created.)
 *
 * Revision 1.39  1995/12/07  01:09:57  jcraig
 * jcraig on Wed Dec 6 17:08:21 PST 1995
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * add procedure wp_calcEncoding which is called by wp_adjustEncoding
 * when the ecc is increased to fill the spare bytes in the rectangle
 * and called by wp_doComputeCapacity to calculate the codeword information
 * and byte capacity of a rectangle.
 *
 * Revision 1.38  1995/11/23  02:07:13  jcraig
 * jcraig on Wed Nov 22 18:07:12 PST 1995
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: make Eric's scripts quit whining when I forgot to put :: after
 * parameter names in the comments
 *
 * Revision 1.37  1995/11/13  20:10:37  ejaquith
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.36  1995/11/09  01:18:57  jcraig
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: wrap ERR_IGNORE around all private procs which do not return an
 * ErrCode
 *
 * Revision 1.35  1995/11/07  21:30:20  jcraig
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * ecc tables are now passed as parameters instead of being global vars
 *
 * Revision 1.34  1995/10/13  21:59:41  ejaquith
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.33  1995/10/11  01:38:06  jcraig
 * jcraig on Tue Oct 10 18:35:48 PDT 1995
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * w_adjustEncoding => wp_adjustEncoding.
 * wp_dgiGetEncodingSize => wp_getEncodingSize.
 * wp_dgiInitEncoding => wp_initEncoding.
 * recompute the number of codewords each time wp_getEncodingSize is
 * called so that the layout handle can be reused after a valid layout
 * has been determined.  otherwise, wp_adjustEncoding is called at the
 * end of computeLayout and this may increase the number of bytes needed
 * since we try to use the extra space for ecc.
 *
 * Revision 1.32  1995/09/30  00:18:36  ejaquith
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: Converted private-ish glyph routines to be of "wp_" form and 100% private.
 *
 * Revision 1.31  1995/09/29  23:12:16  jcraig
 * jcraig on Fri Sep 29 16:11:23 PDT 1995
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * PENCODING => PEncoding.  PDGREGION => PDGRegion.
 *
 * Revision 1.30  1995/09/27  23:41:34  jcraig
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * new procedure comments
 *
 * Revision 1.29  1995/09/23  00:49:06  ejaquith
 * ejaquith on Fri Sep 22 17:48:23 PDT 1995
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * Removed forward declaration of wp_store32().
 * Changed wp_store32() from being a static function.
 *
 * Revision 1.28  1995/09/22  00:38:20  ejaquith
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: Fixed up wp_ issues.
 *
 * Revision 1.27  1995/09/21  22:52:30  ejaquith
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.26  1995/09/07  00:51:54  jcraig
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * dgi prefix becomes w_
 *
 * Revision 1.25  1995/09/06  20:18:23  ejaquith
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.24  1995/09/01  02:11:55  jcraig
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.23  1995/08/31  21:12:04  ejaquith
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.22  1995/08/17  22:29:05  jcraig
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.21  1994/12/05  21:18:41  jcraig
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * make sure that there is at least 1 ecc byte.
 *
 * Revision 1.20  1994/11/02  01:30:59  ejaquith
 * ejaquith on Tue Nov 1 17:28:52 PST 1994
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * Added inclusion of "types.pub" as "memory.pub" no longer includes it.
 *
 * Revision 1.19  1994/11/01  02:59:52  jcraig
 * jcraig on Mon Oct 31 18:59:40 PST 1994
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * include memory.pub since it is no longer in dgencode.pub
 *
 * Revision 1.18  1994/09/30  23:24:12  jcraig
 * jcraig on Fri Sep 30 16:23:09 PDT 1994
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * change eccLevel from an enumerated to a UInt8 which can range from
 * [0..100) and is a percentage of the total bytes given to ecc
 *
 * Revision 1.17  1994/09/30  01:55:17  jcraig
 * jcraig on Thu Sep 29 18:54:33 PDT 1994
 *
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 * replace EDD in handles and constants with D
 * uhhh.  make that DG.
 *
 * Revision 1.16  1994/09/28  21:10:49  jcraig
 * encode.c -> /product/ipcore/ipwrap/src/RCS/encode.c,v
 *
 *  change edd prefix to dg.
 *
 */
