/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  expand1.c
 *
 *      Power-of-2 replicative expansion.
 *
 *         General expansion procedure:
 *                  w_expand ()
 *
 */



#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "expand.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: expand1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Power of 2 (2, 4, 8, 16, 32) isomorphic expansion of 1, 2, 4, 8 bit PIXRs
 *
 * LongDescription:::
 *  Uses simple replicative expansion of rows and columns.  The source PIXR
 *  must have a standard frame.  Multi-channel PIXRs handled correctly.  If 
 *  a pointer to NULL is passed for the destination PIXR, a framed PIXR will 
 *  be allocated and returned here.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Input PIXR pointer
 *  factor::         Expansion factor (2, 4, 8, 16, 32)
 *
 * Output:::
 *  prd::            Output PIXR (allocated if necessary)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_expand (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               factor,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, hS, dS, 
                 i, channels, frame;
    Bool	 destCreated = cFalse;
    Int32        rfc, bfc;
    ScaledInt1616 xResS, yResS;
    PIXR        *prt = NULL,
                *subS = NULL,
		*subD = NULL;
    ColorType    type;
    ColorMap    *map;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepthMask, cChannelsMask, cFrame32);

    prt = NULL;

    ERR_GEN ((factor != 2 && factor != 4 && factor != 8 && 
              factor != 16 && factor != 32), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, prs, &dS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xResS, &yResS));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prs, &channels));
    ERR_PASS (w_pixrGetFrame (pIPWState, prs, &frame));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prs, &map));
    ERR_PASS (w_pixrGetType (pIPWState, prs, &type));

    ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState, 
						      wS * factor,
						      hS * factor, dS,
						      xResS * factor,
						      yResS * factor,
						      cFrameBits,
						      channels, type, map,
						      prd, &rfc, &bfc,
						      &destCreated));

    for (i = 0; i < channels; i++)
    {
	Int32        wSF, hSF, bplS, bplD, wTF, hTF, bplT;
	UInt8       *iS, *iD, *iT;

	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, i, &subS));
        ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, i, &subD));

	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subS, &wSF));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subS, &hSF));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subS, &bplS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subD, &bplD));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subS, &iS));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subD, &iD));

	switch(factor)
	{
	    case 2:
	        ERR_PASS_CLEANUP (i_expand2 (pIPWState, iD, bplD,
		                             iS, wSF, hSF, dS, bplS));
		break;

	    case 4:
	        ERR_PASS_CLEANUP (i_expand4 (pIPWState, iD, bplD,
		                             iS, wSF, hSF, dS, bplS));
		break;

	    case 8:
	        ERR_PASS_CLEANUP (i_expand8 (pIPWState, iD, bplD,
		                             iS, wSF, hSF, dS, bplS));
		break;

	    case 16:
	        ERR_PASS_CLEANUP (i_expand16 (pIPWState, iD, bplD,
		                              iS, wSF, hSF, dS, bplS));
		break;

	    case 32:
	        /* do a 2-stage expansion x4 and then x8 */
	        ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, 4*wS, 4*hS, dS,
		                              4 * xResS, 4 * yResS, cFrameBits,
					      &prt));

		ERR_PASS (w_pixrFramedWidth (pIPWState, prt, &wTF));
		ERR_PASS (w_pixrFramedHeight (pIPWState, prt, &hTF));
		ERR_PASS (w_pixrGetBpl (pIPWState, prt, &bplT));
		ERR_PASS (w_pixrGetImageAll (pIPWState, prt, &iT));

	        ERR_PASS_CLEANUP (i_expand4 (pIPWState, iT, bplT,
		                             iS, wSF, hSF, dS, bplS));
	        ERR_PASS_CLEANUP (i_expand8 (pIPWState, iD, bplD,
		                             iT, wTF, hTF, dS, bplT));

                ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prt));
		break;

	    default:
	        ERR_GEN (1, cErrInvalidParm);	        
		break;
	}
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &prt);
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_expand () */


/*
 *
 * $Log: expand1.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 * Imported sources
 *
 * Revision 1.31  1998/01/14 00:11:25  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.30  1997/04/15 21:47:05  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.29  1996/01/04 20:22:08  ejaquith
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.28  1995/12/13  04:06:27  ejaquith
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.27  1995/11/13  20:10:43  ejaquith
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.26  1995/10/13  21:59:47  ejaquith
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.25  1995/09/26  21:58:01  dferrell
 * dferrell on Tue Sep 26 14:57:27 PDT 1995
 *
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.24  1995/09/21  22:52:34  ejaquith
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.23  1995/09/19  21:08:52  ejaquith
 * ejaquith on Tue Sep 19 14:08:11 PDT 1995
 *
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 *
 * Revision 1.22  1995/09/15  22:38:00  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.21  1995/09/15  00:32:25  dferrell
 * dferrell on Thu Sep 14 17:31:24 PDT 1995
 *
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.20  1995/09/11  23:52:10  dferrell
 * dferrell on Mon Sep 11 16:51:53 PDT 1995
 *
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Changed destroyPixr to w_destroyPixr
 *
 * Revision 1.19  1995/09/11  17:52:27  dferrell
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.18  1995/09/07  01:30:39  ejaquith
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.17  1995/09/06  20:18:29  ejaquith
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.16  1995/09/01  18:18:43  dferrell
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.15  1995/08/31  21:12:10  ejaquith
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.14  1995/08/29  22:22:02  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Remove DEBUG_ON, was defined in debug.h.
 *
 * Revision 1.13  1995/08/29  21:54:08  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.12  1995/08/23  17:33:26  macleod
 * macleod on Wed Aug 23 10:32:38 PDT 1995
 *
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Revision 1.11  1995/08/17  21:52:46  dferrell
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.10  1995/02/03  01:32:07  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.9  1994/11/05  03:05:40  ddavies
 * ddavies on Fri Nov 4 19:04:50 PST 1994
 *
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Remove test to see if expanded image is greater than 33 pixels in width.
 * After inspecting and testing the code, I can't figure out why this was
 * in here in the first place.
 *
 * Revision 1.8  1994/10/03  20:35:29  lperry
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Changed to use expand.pub instead of scalepros.h
 *
 * Revision 1.7  1994/06/08  04:07:37  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.6  1994/04/12  01:13:02  macleod
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.5  1993/12/03  01:14:11  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.4  1993/11/06  03:11:08  ddavies
 * expand1.c -> /product/ipcore/ipwrap/src/RCS/expand1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

