/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  expand2.c
 *
 *      Functions for grayscale expansion, row/column expansion, and
 *      expand&mask.
 *
 *         General expand and Mask procedure:
 *                  w_expandAndMask ()
 *
 */



#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"
#include "logicops.h"
#include "logicops.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "expand.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: expand2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Mask operation which ANDs a smaller mask PIXR with a full sized source
 *
 * LongDescription:::
 *  ANDs the binary source and expanded binary mask PIXRs to form the
 *  binary output PIXR.  The output PIXR may be the same as the "big" source.
 *  The mask PIXR may be the same size as or smaller than the source PIXR
 *  with the additional limitations as follow:
 *  <BR>
 *  <PRE>
 *	If the required mask expansion is isomorphic, the valid
 *	expansion factors are limited to 1, 2, 4, 8, 16, and 32.
 *
 *	If the required mask expansion is anamorphic, the valid
 *	expansion factors must be a combination of one power of
 *	two anamorphic expansion (2, 4, 8, or 16) followed by an 
 *      isomorphic factor from the list above.
 *
 *	For example:  If the source PIXR has dimension 2000x1000 then
 *	the mask image with dimension:
 *
 *	2000 x 1000	is valid (no expansion)
 *	1000 x  500	is valid (2x2 isomorphic expansion)
 *	1000 x 1000	is valid (2x1 anamorphic expansion)
 *	500  x  500	is valid (2x1 anamorphic and 2x2 isomorphic)
 *	700  x  700	is not valid (not power of 2 expansion)
 *	250  x  500	is not valid (requires multiple anamorphic expansion)
 *
 *  </PRE>
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prsBig::         Full sized input PIXR to be masked
 *  prsSmall::       Mask PIXR (can be same size or smaller than prsBig)
 *
 * Output:::
 *  prDest::         Result PIXR (can be same PIXR as prsBig)
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: Does not handle multi-channel PIXRs
 *  (2):: All PIXRs must be binary and framed
 *
 * FunctionCategory:::
 *  masking
 *
 * Keywords:::
 *  masking, isomorphic, anamorphic, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_expandAndMask (
    IPWState           *pIPWState,
    PIXR               *prsBig,
    PIXR               *prsSmall,
    PIXR              **prDest)
{
    Int32        wS, wD, hS, hD, 
	         hRed, vRed,       /* horiz & vert reduction factors */
	         bplS, bplD,       /* stride */
		 dS;
    Bool         destCreated = cFalse;
    Int32        rfc, bfc;
    ScaledInt1616 xResS, yResS;
    UInt8       *iS, *iD;              /* image data pointers */
    PIXR        *prt = NULL,           /* intermediate result pixrs */
                *pri = NULL;
    PROLOGUE;

    /* pixrs OK? */
    ERR_GEN_PIXR (prsBig, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR (prsSmall, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prDest != NULL)
	ERR_GEN_PIXR (*prDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* check or create the dest (same size as source) */
    ERR_PASS (w_pixrGetWidth (pIPWState, prsBig, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prsBig, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, prsBig, &dS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prsBig, &xResS, &yResS));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS, hS, dS,
					      xResS, yResS,
					      cFrameBits, prDest, &rfc, &bfc, 
					      &destCreated));
    if (prsBig != (*prDest))
        ERR_PASS_CLEANUP (w_copy (pIPWState, prsBig, prDest));

    /* get the dimensions of the mask and dest PIXRs */
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prsSmall, &wS));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prsSmall, &hS));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, *prDest, &wD));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, *prDest, &hD));
    ERR_PASS (w_pixrGetDepth (pIPWState, prsSmall, &dS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prsSmall, &xResS, &yResS));

    if (wS == wD)
        hRed = 1;
    else
        if ((2*wS <= wD) && (2*wS + 2 > wD))
	    hRed = 2;
	else
	    if ((4*wS <= wD) && (4*wS + 4 > wD))
	        hRed = 4;
	    else
	        if ((8*wS <= wD) && (8*wS + 8 > wD))
		    hRed = 8;
		else
		    if ((16*wS <= wD) && (16*wS + 16 > wD))
		        hRed = 16;
		    else
		        if ((32*wS <= wD) && (32*wS + 32 > wD))
			    hRed = 32;
			else
			    ERR_GEN_CLEANUP (1, cErrDestDimensIncorrect);

    if (hS == hD)
        vRed = 1;
    else
        if ((2*hS <= hD) && (2*hS + 2 > hD))
	    vRed = 2;
	else
	    if ((4*hS <= hD) && (4*hS + 4 > hD))
	        vRed = 4;
	    else
	        if ((8*hS <= hD) && (8*hS + 8 > hD))
		    vRed = 8;
		else
		    if ((16*hS <= hD) && (16*hS + 16 > hD))
		        vRed = 16;
		    else
		        if ((32*hS <= hD) && (32*hS + 32 > hD))
			    vRed = 32;
			else
			    ERR_GEN_CLEANUP (1, cErrDestDimensIncorrect);

    /* If the expansion is anamorphic, pre-expand to isomorphic first */
    if (hRed != vRed) 
    {
        if (hRed > vRed)
	{
            ERR_PASS_CLEANUP (w_expandCols (pIPWState, prsSmall, 
	                                    hRed / vRed, &pri));
	    hRed = vRed;
	}
	else  /* vRed > hRed */
	{
            ERR_PASS_CLEANUP (w_expandRows (pIPWState, prsSmall, 
	                                    vRed / hRed, &pri));
	    vRed = hRed;
	}
    }
    else
    {
        pri = prsSmall;
    }

    /* pri now holds the "mask" image */

    /* Do the actual expand and mask step(s) */
    if (hRed == vRed)  
    {
	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, pri, &wS));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, pri, &hS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, pri, &bplS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prDest, &bplD));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, pri, &iS));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prDest, &iD));

        switch (hRed)
	{
	    case 1:
		ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pri, &wS));
		ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pri, &hS));
		/* can't use w_logicOp () here due to link order */
	        ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, pri,
		                              PIX_SRC & PIX_DST, 
					      0, 0, wS, hS, *prDest));
		break;

	    case 2:
	        ERR_PASS_CLEANUP (i_expandAndMask2 (pIPWState, iD, bplD,
		                            iS, wS, hS, bplS));
	        break;

	    case 4:
	        ERR_PASS_CLEANUP (i_expandAndMask4 (pIPWState, iD, bplD,
		                            iS, wS, hS, bplS));
	        break;

	    case 8:
	        ERR_PASS_CLEANUP (i_expandAndMask8 (pIPWState, iD, bplD,
		                            iS, wS, hS, bplS));
	        break;

	    case 16:
	        ERR_PASS_CLEANUP (w_expand (pIPWState, pri, 2, &prt));

		/* get info on intermediate image */
		ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prt, &wS));
		ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prt, &hS));
		ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prt, &bplS));
		ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prt, &iS));

		/* finish the operation */
		ERR_PASS_CLEANUP (i_expandAndMask8 (pIPWState, iD, bplD,
		                                    iS, wS, hS, bplS));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prt));
	        break;

	    case 32:
	        ERR_PASS_CLEANUP (w_expand (pIPWState, pri, 4, &prt));

		/* get info on intermediate image */
		ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prt, &wS));
		ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prt, &hS));
		ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prt, &bplS));
		ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prt, &iS));

		/* finish the operation */
		ERR_PASS_CLEANUP (i_expandAndMask8 (pIPWState, iD, bplD,
		                                    iS, wS, hS, bplS));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prt));
	        break;


	}

    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prDest,
						 rfc, bfc));

    if (pri != prsSmall)
        ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pri));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pri != prsSmall)
        w_destroyPixr (pIPWState, &pri);
    if (destCreated)
        w_destroyPixr (pIPWState, prDest);
    w_destroyPixr (pIPWState, &prt);
    ERR_RETURN;

}  /* w_expandAndMask () */


/*

 $Log: expand2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.28  1998/01/14 00:11:27  ddavies
 expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.27  1997/06/26 20:00:44  ddavies
 expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.26  1997/04/15 21:47:16  ddavies
 expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.25  1996/01/04 20:22:10  ejaquith
 expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.24  1995/12/13  04:06:30  ejaquith
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.23  1995/11/13  20:10:46  ejaquith
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.22  1995/10/13  21:59:50  ejaquith
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.21  1995/10/13  18:31:00  dferrell
 * dferrell on Fri Oct 13 11:27:03 PDT 1995
 *
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Improved comment block for w_expandAndMask ().  Replaced call to
 * w_logicOp () with w_rasterop () to avoid link order problems (scale
 * lib follows ip lib).  Improved performance for "in-place" masking
 * (prsBig == *prsDest) by conditionally copying as a first step.
 *
 * Revision 1.20  1995/09/26  21:58:10  dferrell
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.19  1995/09/21  22:52:37  ejaquith
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/19  21:09:10  ejaquith
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 *
 * Revision 1.17  1995/09/15  22:38:04  ddavies
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.16  1995/09/15  00:32:42  dferrell
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.15  1995/09/11  17:52:38  dferrell
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.14  1995/09/07  01:30:42  ejaquith
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:18:32  ejaquith
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  18:18:48  dferrell
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.11  1995/08/31  21:12:13  ejaquith
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/29  20:49:00  jcraig
 * jcraig on Tue Aug 29 13:47:42 PDT 1995
 *
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * use pri instead of prs when doing the actual expand and mask
 * fix typo setting reduction when expansion is anamorphic
 *
 * Revision 1.9  1995/08/17  21:52:54  dferrell
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.8  1995/06/01  17:04:32  dferrell
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Added logic and code to expandAndMask () so that it can now handle
 * anamorphic expansion (useful for FAX images) as well as the normal
 * unexpanded mask case (which should reduce logic in client code base).
 * Updated comments in expandAndMaskPixr () and expandAndMask () to
 * reflect these changes.  Note that no interface changes were necessary
 * and the functions will perform exactly as before.
 *
 * Revision 1.7  1995/02/03  01:32:18  ddavies
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.6  1994/10/03  20:35:38  lperry
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Changed to use expand.pub instead of scalepros.h
 *
 * Revision 1.5  1994/06/08  04:09:04  ddavies
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.4  1993/12/03  01:14:16  ddavies
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.3  1993/11/06  03:11:16  ddavies
 * expand2.c -> /product/ipcore/ipwrap/src/RCS/expand2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */


