/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  expand3.c
 *
 *      Power-of-2 replicative expansion of rows and columns.
 *
 *                  w_expandRows ()
 *                  w_expandCols ()
 *
 */

#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "expand.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: expand3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reproductively expand rows in a binary pixr by 2, 4, 8, or 16
 *
 * LongDescription:::
 *  Reproductively expand rows in a binary pixr by 2, 4, 8, or 16.
 *  Creates the output PIXR if necessary.
 *  
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  inPr::           Source PIXR
 *  factor::         Vertical expansion factor
 *
 * Output:::
 *  outPr::          Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, anamorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_expandRows (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               factor,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32        wS, hS, 
                 bplS, bplD;
    Bool	 destCreated = cFalse;
    Int32        rfc, bfc;
    UInt8       *iS, *iD;
    ScaledInt1616  xres, yres;

    /* pixrs OK? */
    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN ((factor != 2 && factor != 4 && factor != 8 && factor != 16),
              cErrInvalidParm);


    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xres, &yres));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS, hS * factor, 1, 
					      xres, yres * factor,
					      cFrameBits, outPr, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    ERR_PASS_CLEANUP (i_expandRows (pIPWState, iD, bplD,
				    iS, wS, hS, bplS, factor));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *outPr,
						 rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    ERR_RETURN;

}  /* w_expandRows () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reproductively expand columns in a binary pixr by 2, 4, 8, or 16
 *
 * LongDescription:::
 *  Reproductively expand columns in a binary pixr by 2, 4, 8, or 16.
 *  Creates the output PIXR if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  inPr::           Input PIXR to be expanded
 *  factor::         Horizontal expansion factor
 *
 * Output:::
 *  outPr::          Output PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  anamorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_expandCols (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               factor,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32        wS, hS, 
                 bplS, bplD;
    Bool	 destCreated = cFalse;
    Int32        rfc, bfc;
    UInt8       *iS, *iD;
    ScaledInt1616  xres, yres;

    /* pixrs OK? */
    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN ((factor != 2 && factor != 4 && factor != 8 && factor != 16),
             cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xres, &yres));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS * factor, hS, 1, 
					      xres * factor, yres,
					      cFrameBits, outPr, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    ERR_PASS_CLEANUP (i_expandCols (pIPWState, iD, bplD,
				    iS, wS, hS, bplS, factor));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *outPr,
						 rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    ERR_RETURN;

}  /* w_expandCols () */



/*
 $Log: expand3.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.21  1998/01/14 00:11:28  ddavies
 expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.20  1997/04/15 21:47:30  ddavies
 expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.19  1996/01/04 20:22:14  ejaquith
 expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.18  1995/12/13  04:06:34  ejaquith
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.17  1995/11/13  20:10:49  ejaquith
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/10/13  21:59:53  ejaquith
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/09/26  21:58:16  dferrell
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.14  1995/09/21  22:52:40  ejaquith
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/19  21:09:20  ejaquith
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 *
 * Revision 1.12  1995/09/15  22:38:09  ddavies
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.11  1995/09/15  00:32:54  dferrell
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.10  1995/09/11  17:52:48  dferrell
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.9  1995/09/07  01:30:45  ejaquith
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.8  1995/09/06  20:18:38  ejaquith
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.7  1995/09/01  18:18:54  dferrell
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.6  1995/08/31  21:12:15  ejaquith
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.5  1995/08/29  21:54:15  ddavies
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.4  1995/08/24  00:30:04  jcraig
 * jcraig on Wed Aug 23 17:28:37 PDT 1995
 *
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * in createPixr pass parameters in the correct order (depth is before
 * resolution not after).
 * in expandCols, the width is increased in the height
 *
 * Revision 1.3  1995/08/17  21:53:01  dferrell
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.2  1995/06/01  17:00:09  dferrell
 * dferrell on Thu Jun 1 09:58:04 PDT 1995
 *
 * expand3.c -> /product/ipcore/ipwrap/src/RCS/expand3.c,v
 *
 * Added more sanity checks of pixr sizes in expandRowsPr () and
 * expandColsPr ().
 *
 * Revision 1.1  1995/02/28  00:07:33  dferrell
 * Initial revision
 *

 */

