/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fill1.c:  includes subroutines (used by XIS)
 *
 *  --------------------   Fast Fill Clip Interface  ------------------------
 *
 *	  Int32	  fillClipSweep8()  8-connected version of fillClipSweep4
 *
 *	  Int32	  fillClipSweep2H() 2-connected horizontal fill operation
 *
 *	  Int32	  fillClipSweep2V() 2-connected vertical fill operation
 *
 *
 *
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "pixr.h"
#include "fill.h"
#include "geomadt.h"
#include "shrcc.pub"

/* Prototypes */
#include "props.pub"
#include "shrip.pub"
#include "fill.pub"




IP_RCSINFO(RCSInfo, "$RCSfile: fill1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");
#define  DEBUG_FILL       0

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Grows the seeds in the dest using multiple passes
 *
 * LongDescription:::
 *   takes dest, src, and clip pixrs, and a maxCount. Grows the seeds in 
 *   the dest using multiple  passes of i_fillClipSeed8Conn.  Stops when 
 *   either nothing changes or fillClipSweep8Conn has been called maxCount 
 *   times.  Does an 8-connected fill.  This operation is sometimes called
 *   "geodesic reconstruction" in the literature.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prs::            Seed image
 *  prc::            Clip image
 *  maxCount::       Max # of iterations
 *
 * Output:::
 *  pPrd::           Resulting image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 * 
 * FunctionCategory:::
 *  seed filling
 *
 * Keywords:::
 *  seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_fillClipSweep8 (
    IPWState           *pIPWState,
    PIXR               *prs,
    PIXR               *prc,
    Int32               maxCount,
    PIXR              **pPrd)
{
    PROLOGUE;
Int32		 i;
UInt8        	*destImage, *seedImage, *maskImage;
Int32		 destPitch, seedPitch, maskPitch;	/* in bytes */
Int32		 imageWidth, imageHeight;
Int32		 writeCount;
PIXR            *prd = NULL;

    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    if (*pPrd == NULL)
    {
	ERR_PASS (w_clonePixrNoInit (pIPWState, prs, &prd));
    }
    else
    {
	prd = *pPrd;
	ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);
    }

    ERR_GEN_PIXR (prc, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    ERR_GEN (maxCount <= 0, cErrInvalidParm);

	/* prepare args for i_fillClipSeedTwoPass */
    ERR_PASS (w_pixrGetImageAll(pIPWState, prd, &destImage));
    ERR_PASS (w_pixrGetImageAll(pIPWState, prs, &seedImage));
    ERR_PASS (w_pixrGetImageAll(pIPWState, prc, &maskImage));

    ERR_PASS (w_pixrGetBpl(pIPWState, prd, &destPitch));
    ERR_PASS (w_pixrGetBpl(pIPWState, prs, &seedPitch));
    ERR_PASS (w_pixrGetBpl(pIPWState, prc, &maskPitch));

    ERR_PASS (w_pixrFramedWidth(pIPWState, prd, &imageWidth));
    ERR_PASS (w_pixrFramedHeight(pIPWState, prd, &imageHeight));

	/* do the first pass, initially clipping the seed with the mask */
    ERR_PASS( i_fillClipSeed8Conn (pIPWState, destImage, seedImage, maskImage,
				   destPitch, seedPitch, maskPitch,
				   imageWidth, imageHeight, cTrue,
				   &writeCount));

	/* do remaining cycles with no initial clip.  We stop when either
	 * the routine fails, nothing has changed or we've called it
	 * the maximum number of times. */
    for (i = 1; (i < maxCount) && (writeCount > 0); i++)	
    {
	ERR_PASS( i_fillClipSeed8Conn (pIPWState, destImage,
				       destImage, maskImage,
				       destPitch, destPitch, maskPitch,
				       imageWidth, imageHeight, cFalse,
				       &writeCount));
    }

	/* It all worked.  Set the return image and return */
    *pPrd = prd;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Horizontal seed-fill in one direction
 *
 * LongDescription:::
 *  takes dest, src, and clip pixrs and direction.
 *  Grows the seeds into the dest using a horizontal fill. 
 *  The direction is specified by the fillDirection parameter
 *  since a pixel may be filled if the pixel to its left is
 *  on (FILL_RIGHT) or if the pixel to its right is 
 *  on (FILL_LEFT) or either the pixel to its left or its right
 *  is on (FILL_RIGHT_LEFT). This operation is sometimes called
 *  "geodesic reconstruction" in the literature.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prs::            Seed image
 *  prc::            Clip image
 *  fillDirection::  Fill direction, FILL_RIGHT or FILL_LEFT
 *
 * Output:::
 *  pPrd::           Filled image
 *
 * Returns:::
 *  error stats
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  seed filling
 *
 * Keywords:::
 *  seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_fillClipSweep2H (
    IPWState           *pIPWState,
    PIXR               *prs,
    PIXR               *prc,
    Int32               fillDirection,
    PIXR              **pPrd)
{
    PROLOGUE;
UInt8        	*destImage, *seedImage, *maskImage;
Int32		 destPitch, seedPitch, maskPitch;	/* in bytes */
Int32		 imageWidth, imageHeight;
PIXR             *prd = NULL;

    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);
    if (*pPrd == NULL)
    {
	ERR_PASS (w_clonePixrNoInit (pIPWState, prs, &prd));
    }
    else
    {
	prd = *pPrd;
	ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);
    }

    ERR_GEN_PIXR (prc, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);


	/* prepare args for i_fillClipSeedHorizontal */
    ERR_PASS (w_pixrGetImageAll(pIPWState, prd, &destImage));
    ERR_PASS (w_pixrGetImageAll(pIPWState, prs, &seedImage));
    ERR_PASS (w_pixrGetImageAll(pIPWState, prc, &maskImage));

    ERR_PASS (w_pixrGetBpl(pIPWState, prd, &destPitch));
    ERR_PASS (w_pixrGetBpl(pIPWState, prs, &seedPitch));
    ERR_PASS (w_pixrGetBpl(pIPWState, prc, &maskPitch));

    ERR_PASS (w_pixrFramedWidth(pIPWState, prd, &imageWidth));
    ERR_PASS (w_pixrFramedHeight(pIPWState, prd, &imageHeight));

	/* do the fill */
    ERR_PASS( i_fillClipSeedHorizontal (pIPWState, destImage,
					seedImage, maskImage,
					destPitch, seedPitch, maskPitch,
					imageWidth, imageHeight,
					fillDirection));

	/* It all worked.  Set the return image and return */
    *pPrd = prd;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Vertical seed-fill in one direction
 *
 * LongDescription:::
 *  Grows the seeds into the dest using a vertical fill. 
 *  The direction is specified by the fillDirection parameter
 *  since a pixel may be filled if the pixel above it is
 *  on (FILL_DOWN) or if the pixel below it is 
 *  on (FILL_UP) or either the pixel above or below it
 *  is on (FILL_UP_DOWN). This operation is sometimes called
 *  "geodesic reconstruction" in the literature.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prs::            Seed image
 *  prc::            Clip image
 *  fillDirection::  Fill direction, FILL_UP, FILL_DOWN, FILL_UP_DOWN
 *
 * Output:::
 *  pPrd::           Filled image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  seed filling
 *
 * Keywords:::
 *  seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_fillClipSweep2V (
    IPWState           *pIPWState,
    PIXR               *prs,
    PIXR               *prc,
    Int32               fillDirection,
    PIXR              **pPrd)
{
    PROLOGUE;
UInt8        	*destImage, *seedImage, *maskImage;
Int32		 destPitch, seedPitch, maskPitch;	/* in bytes */
Int32		 imageWidth, imageHeight;
PIXR            *prd = NULL;

    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);
    if (*pPrd == NULL)
    {
	ERR_PASS (w_clonePixrNoInit (pIPWState, prs, &prd));
    }
    else
    {
	prd = *pPrd;
	ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);
    }

    ERR_GEN_PIXR (prc, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);


	/* prepare args for i_fillClipSeedHorizontal */
    ERR_PASS (w_pixrGetImageAll(pIPWState, prd, &destImage));
    ERR_PASS (w_pixrGetImageAll(pIPWState, prs, &seedImage));
    ERR_PASS (w_pixrGetImageAll(pIPWState, prc, &maskImage));

    ERR_PASS (w_pixrGetBpl(pIPWState, prd, &destPitch));
    ERR_PASS (w_pixrGetBpl(pIPWState, prs, &seedPitch));
    ERR_PASS (w_pixrGetBpl(pIPWState, prc, &maskPitch));

    ERR_PASS (w_pixrFramedWidth(pIPWState, prd, &imageWidth));
    ERR_PASS (w_pixrFramedHeight(pIPWState, prd, &imageHeight));

	/* do the first pass, initially clipping the seed with the mask */
    ERR_PASS( i_fillClipSeedVertical (pIPWState, destImage,
				      seedImage, maskImage,
				      destPitch, seedPitch, maskPitch,
				      imageWidth, imageHeight, fillDirection));

	/* It all worked.  Set the return image and return */
    *pPrd = prd;
    NORMAL_RETURN;
}

/*

 $Log: fill1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.21  1998/01/14 00:11:30  ddavies
 fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.20  1997/04/15 21:47:42  ddavies
 fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.19  1997/01/14 02:46:25  ddavies
 fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v

 Heavens to Betsy!  FillClipSweep8, fillClipSweep2H and fillClipSweep2V
 all failed to return a destination image if they allocated one!  This
 is now fixed.

 Revision 1.18  1996/01/04 20:22:17  ejaquith
 fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.17  1995/12/13  04:06:38  ejaquith
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/13  20:10:52  ejaquith
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.15  1995/10/13  21:59:56  ejaquith
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.14  1995/10/03  21:12:53  macleod
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.13  1995/09/21  22:52:42  ejaquith
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/11  21:55:33  macleod
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.11  1995/09/07  01:30:47  ejaquith
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.10  1995/09/06  20:18:44  ejaquith
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  20:43:43  macleod
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * Added back #include files
 *
 * Revision 1.8  1995/08/31  21:12:18  ejaquith
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/25  01:36:10  ejaquith
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.6  1995/08/17  21:54:48  macleod
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.5  1994/10/06  20:53:27  lperry
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.4  1994/05/05  01:12:22  ddavies
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * Include adtpros.h instead of ippros.h since this is part of libadt.
 *
 * Revision 1.3  1993/12/03  01:14:19  ddavies
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:11:23  ddavies
 * fill1.c -> /product/ipcore/ipwrap/src/RCS/fill1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

