/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fill2.c:  includes subroutines
 *
 *  ---------------  Bounding Boxes of Connected Components  ---------------
 *        BOXES  *rabb():       returns bounding boxes is a list of
 *				RECTANGLs
 *
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "frames.pub"
#include "pixr.h"
#include "fill.h"
#include "geomadt.h"
#include "shrcc.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* Prototypes */
#include "props.pub"
#include "shrip.pub"
#include "intops.pub"
#include "rectops.pub"
#include "fill.pub"





IP_RCSINFO(RCSInfo, "$RCSfile: fill2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Construct a list of RECTANGLs bounding connected components in an image
 *
 * LongDescription:::
 *  Constructs list of RECTANGLs containing the bounding boxes of
 *  connected components from a bitmap.
 *  Uses 8-connected rules to determine whether components
 *  are connected.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State strucure
 *  prs::            Image to analyze
 *  minWidth::       Minimum width of a feature to consider
 *  minHeight::      Minimum height of a feature to consider
 *
 * Output:::
 *  ppRectList::     List of rectangles corresponding to connected components
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rabb (
    IPWState           *pIPWState,
    PIXR               *prs,
    UInt32              minWidth,
    UInt32              minHeight,
    RECTANGL          **ppRectList)
{
    PROLOGUE;
Int32	      rectReturnCode, bpl, clipW, clipH;
UInt8        *pImage;
AddCCPacket   packet;


    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    packet.rectList = NULL;
    ERR_PASS( w_pixrGetImageAll (pIPWState, prs, &pImage));
    ERR_PASS( w_pixrGetBpl (pIPWState, prs, &bpl));
    ERR_PASS( w_pixrGetWidth (pIPWState, prs, &clipW));
    ERR_PASS( w_pixrGetHeight (pIPWState, prs, &clipH));

	/* Do the real work.  The procedure addCCRectToList is called
	 * for each bounding box found.
	 */

    ERR_PASS_CLEANUP( i_findConnectedComponentBoundingBoxes (
	pIPState,
	pImage,
	bpl,
	cFrameBits, cFrameBits,	/* clipX, clipY */
	clipW,
	clipH,
	minWidth, minHeight,	/* min width & height
				 * of returned rects */
	1,			/* 8-connected */
	(void *)&packet,	/* packet which contains ptr to ptr to
				 * rects as well as the ipstate. */
	w_addCCRectToList,	/* callback routine */
	&rectReturnCode));	/* return code for callback proc*/

	/* if we get here, it must have worked */
    *ppRectList  = packet.rectList;
    NORMAL_RETURN;

ERR_CLEANUP:

    if (packet.rectList) w_destroyRectList(pIPWState, &packet.rectList);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Construct a list of RECTANGLs bounding connected components in an image
 *
 * LongDescription:::
 *  Constructs a list of RECTANGLs for bounding boxes of 
 *  connected components (bb of cc) from a bitmap.  Uses 4-connected rules 
 *  to determine whether components are connected.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to work from
 *  minWidth::       Min. width of an object to find as a feature
 *  minHeight::      Min. height of an object to find as a feature
 *
 * Output:::
 *  ppRectList::     List of RECTANGLs for bounding boxes
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rabb4Connected (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    UInt32              minWidth,
    UInt32              minHeight,
    RECTANGL          **ppRectList)
{
PROLOGUE;
Int32	      rectReturnCode, bpl, clipW, clipH;
UInt8        *pImage;
AddCCPacket   packet;
Bool          isZero;

    memset(&packet, 0, sizeof(packet));

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    ERR_PASS( w_isZero(pIPWState, prSrc, &isZero));
    if (!isZero)
    {
	packet.rectList = NULL;
	ERR_PASS( w_pixrGetImageAll (pIPWState, prSrc, &pImage));
	ERR_PASS( w_pixrGetBpl (pIPWState, prSrc, &bpl));
	ERR_PASS( w_pixrGetWidth (pIPWState, prSrc, &clipW));
	ERR_PASS( w_pixrGetHeight (pIPWState, prSrc, &clipH));
	
	/* Do the real work.  The procedure addCCRectToList is called
	 * for each bounding box found.
	 */
	ERR_PASS_CLEANUP( i_findConnectedComponentBoundingBoxes(
	    pIPState,
	    pImage,
	    bpl,
	    cFrameBits, cFrameBits,	/* clipX, clipY */
	    clipW,
	    clipH,
	    minWidth, minHeight,	/* min width & height
					 * of returned rects */
	    0,			        /* 4-connected */
	    (void *)&packet,	        /* packet which contains ptr to 
					 * rects as well as the ipstate. */
	    w_addCCRectToList,	        /* callback routine */
           &rectReturnCode));
    }

    *ppRectList  = packet.rectList;
    NORMAL_RETURN;

ERR_CLEANUP:

    if (packet.rectList) w_destroyRectList(pIPWState, &packet.rectList);
    ERR_RETURN;
}


/*

 $Log: fill2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.33  1998/01/14 00:11:32  ddavies
 fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.32  1997/08/02 01:02:36  aber
 aber on Fri Aug 1 18:02:28 PDT 1997

 fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v

 Must use ansiprot.h to use memset.

 Revision 1.31  1997/08/01 23:51:04  aber
 aber on Fri Aug 1 16:50:24 PDT 1997

 fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v

 Clear result before deciding to do nothing and return an uninitialized
 result (bug).

 Revision 1.30  1996/01/04 20:22:19  ejaquith
 fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.29  1995/12/13  04:06:42  ejaquith
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.28  1995/11/13  20:10:54  ejaquith
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.27  1995/10/13  21:59:59  ejaquith
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.26  1995/10/03  21:12:55  macleod
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.25  1995/09/21  22:52:45  ejaquith
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.24  1995/09/11  21:55:42  macleod
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.23  1995/09/06  20:18:47  ejaquith
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.22  1995/09/01  20:43:48  macleod
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Added back #include files
 *
 * Revision 1.21  1995/08/31  21:12:22  ejaquith
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.20  1995/08/25  01:36:13  ejaquith
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Fixed warnings
 *
 * Revision 1.19  1995/08/23  20:17:54  jcraig
 * jcraig on Wed Aug 23 13:16:01 PDT 1995
 *
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * fix packet that will given to callback so that it is initialized
 * properly.  changed so that it contains a pointer to a rectangle not
 * a pointer to a pointer to a rectangle
 *
 * Revision 1.18  1995/08/17  21:54:51  macleod
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.17  1995/07/14  02:48:21  ddavies
 * ddavies on Thu Jul 13 19:47:19 PDT 1995
 *
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Removed addCCRectToList.  It went to fill6.c so that the rabbLite could
 * use it without taking the rest of the stuff in fill2.c.
 *
 * Revision 1.16  1995/01/11  22:41:13  dferrell
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Changed include of shrip.pub to shrcc.pub.
 *
 * Revision 1.15  1994/10/21  03:50:11  ddavies
 * ddavies on Thu Oct 20 20:49:23 PDT 1994
 *
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Add nRuns, px and py parms to addCCToRectList.  This is the callback
 * procedure used by rabb.  These args are now returned by
 * i_findConnectedComponentBoundingBoxes.  These parm are recorded in the
 * new nRuns, px and py fields of each RECTANGL.
 *
 * Revision 1.14  1994/10/06  20:53:31  lperry
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.13  1994/09/16  17:35:32  jcraig
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * remove include of bitacces.h which was not needed.
 *
 * Revision 1.12  1994/06/08  04:10:03  ddavies
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Remove include of obsolete ipdef.h.
 * Also update the comments for rabb.  Remove the check to see if the entire
 * image is 0 - it hardly ever is.
 *
 * Revision 1.11  1994/03/31  03:05:05  ddavies
 * ddavies on Wed Mar 30 19:04:40 PST 1994
 *
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Add implementation of rabb4Connected.  This is just like rabb, except it defines
 * connected components using 4-connected rules (only horizontals and verticals
 * count as connections) instead of 8 connected rules (diagonals count too).
 *
 * Revision 1.10  1994/01/07  04:22:57  ddavies
 * ddavies on Thu Jan 6 20:21:40 PST 1994
 *
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Change addCCRectToList so it records the "mass", or the number of active
 * pixels in each connected component, to the RECTANGL made for each
 * connected component.
 *
 * Revision 1.9  1993/12/17  22:42:54  ddavies
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Completely change the implementation of rabb.  It used to use bbSeedFill and
 * nextOnPixel to turn off each pixel in each connected component in the process
 * of identifying connected component bounding boxes.  It now uses
 * i_findConnectedComponentBoundingBoxes and addCCRectToList.  The former is
 * an ipshared routine that finds the connected components of the image without
 * requiring a copy.  It's roughly 8x faster than the original implementation -
 * though that's highly dependent on the original image.  It also does not require
 * a new copy of the source image.  addCCRectToList is called by
 * i_findConnectedComponentBoundingBoxes each time a box of legal size is found.
 * addCCRectToList creates a RECTANGL and adds it to the list of rects.
 * This may be of little value, but the rects returned are sorted by the
 * Y coordinate of the *lowest* point in the blob (lowest y coords first).
 *
 * Revision 1.8  1993/12/03  01:14:24  ddavies
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.7  1993/11/29  23:00:47  ddavies
 * ddavies on Mon Nov 29 15:00:00 PST 1993
 *
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Fix memory leak in rabb.  The new function that discarded bounding boxes
 * that were too small didn't really discard them, it just forgot about them.
 *
 * Revision 1.6  1993/11/20  03:10:29  ddavies
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Move the allocating and deallocating of bbSeedFill's stack from
 * bbSeedFill (where it happens for every connected component) to rabb (where
 * it happens once per image).  Tidy up after error returns from rabb.
 * Modify rabb so it only returns rectangles that are at least minWidth
 * pixels wide and minHeight pixels tall.
 * Also change nextOnPixel so it uses i_nextBitToTheRight instead of a
 * custom piece of code to find the next 1 in the image.
 *
 * Revision 1.5  1993/11/06  03:11:30  ddavies
 * fill2.c -> /product/ipcore/ipwrap/src/RCS/fill2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

