/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fill3.c:  includes subroutines
 *
 *  --------------------   Fast Fill Clip Interface  ------------------------
 *
 *	  Int32	  fillClipSweep4() : faster version of fillClip operation
 *                                  than fillClipStack().  This operation
 *                                  is iterative and is not guaranteed to
 *                                  fill convoluted masks.
 *  ----------------   Fast Fill Clip Derivative Functions  ----------------
 *
 *	  Int32	  edgeFillSweep4():  version of edgeFill operation that
 *                                  uses fillClipSweep4().  Not guaranteed
 *                                  to fill arbitrarily convoluted masks.
 *
 *
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "frames.pub"
#include "pixr.h"
#include "fill.h"
#include "geomadt.h"
#include "shrcc.pub"

/* Prototypes */
#include "props.pub"
#include "rast.pub"
#include "shrip.pub"
#include "fill.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: fill3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");
#define  DEBUG_FILL       0



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Fast 4-connected seed fill
 *
 * LongDescription:::
 *  Grows the seeds in the dest using multiple passes of
 *  fillClipSweepTwoPass.  Stops when either nothing
 *  changes or fillClipSweepTwoPass has been called
 *  maxCount times. Does a 4 connected fill. This operation 
 *  is sometimes called "geodesic reconstruction" in the
 *  literature.
 * 
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prs::            Seed image
 *  prc::            Clip image
 *  maxCount::       Max # of iterations
 *
 * Output:::
 *  pPrd::           Destination image
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  seed filling
 *
 * Keywords:::
 *  seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_fillClipSweep4 (
    IPWState           *pIPWState,
    PIXR               *prs,
    PIXR               *prc,
    Int32               maxCount,
    PIXR              **pPrd)
{
    PROLOGUE;
Int32		 i;
UInt8        	*destImage, *seedImage, *maskImage;
Int32		 destPitch, seedPitch, maskPitch;	/* in bytes */
Int32		 imageWidth, imageHeight;
Int32		 writeCount;
PIXR            *prd = NULL;

    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);
    if (*pPrd == NULL)
    {
	ERR_PASS (w_clonePixrNoInit (pIPWState, prs, &prd));
    }
    else
    {
	prd = *pPrd;
	ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);
    }

    ERR_GEN_PIXR (prc, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    ERR_GEN (maxCount <= 0, cErrInvalidParm);


	/* prepare args for i_fillClipSeedTwoPass */
    ERR_PASS( w_pixrGetImageAll(pIPWState, prd, &destImage));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prs, &seedImage));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prc, &maskImage));

    ERR_PASS( w_pixrGetBpl(pIPWState, prd, &destPitch));
    ERR_PASS( w_pixrGetBpl(pIPWState, prs, &seedPitch));
    ERR_PASS( w_pixrGetBpl(pIPWState, prc, &maskPitch));

    ERR_PASS( w_pixrFramedWidth(pIPWState, prd, &imageWidth));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prd, &imageHeight));


	/* do the first pass, initially clipping the seed with the mask */
    ERR_PASS( i_fillClipSeedTwoPass (pIPWState, destImage, seedImage, maskImage,
				     destPitch, seedPitch, maskPitch,
				     imageWidth, imageHeight, cTrue,
				     &writeCount));

	/* do remaining cycles with no initial clip.  We stop when either
	 * the routine fails, nothing has changed or we've called it
	 * the maximum number of times. */
    for (i = 1; (i < maxCount) && (writeCount > 0); i++)	
    {
	ERR_PASS( i_fillClipSeedTwoPass (pIPWState, destImage,
					 destImage, maskImage,
					 destPitch, destPitch, maskPitch,
					 imageWidth, imageHeight, cFalse,
					 &writeCount));
    }

	/* It all worked.  Set the return image and return */
    *pPrd = prd;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Seed fill from the edges of an image
 *
 * LongDescription:::
 *  Flood fills into prd from the edge of the Frame.
 *  Grows the seeds in the dest using multiple passes of
 *  i_fillClipSeedTwoPass.  Stops when either nothing
 *  changes or i_fillClipSeedTwoPass has been called
 *  maxCount times.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::    State structure  
 *  prs::          Source image
 *  maxCount::     Max # of iterations
 *
 * Output:::
 *  pPrd::         Filled PIXR (can be the same as source PIXR; allocated if
 *                 NULL)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  seed filling
 *
 * Keywords:::
 *  seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_edgeFillSweep4 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               maxCount,
    PIXR              **pPrd)
{
    PROLOGUE;
Int32	     i;
UInt8       *seedImage, *maskImage;
Int32	     seedPitch, maskPitch;	/* in bytes */
Int32	     imageWidth, imageHeight;
Int32	     fullImageWidth, fullImageHeight;
PIXR        *prSeed = NULL, *prd = NULL;
Int32	     writeCount;
ScaledInt1616        xDPI, yDPI;


    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    if (*pPrd == NULL)
    {
	ERR_PASS (w_clonePixrNoInit (pIPWState, prs, &prd));
    }
    else
    {
	prd = *pPrd;
	ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);
    }

    ERR_GEN (maxCount <= 0, cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth(pIPWState, prs, &imageWidth));
    ERR_PASS( w_pixrGetHeight(pIPWState, prs, &imageHeight));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prs, &fullImageWidth));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prs, &fullImageHeight));
    ERR_PASS( w_pixrGetResolutionDPI (pIPWState, prs, &xDPI, &yDPI));

	/* construct a seed.  This is a pixrect with a 1 pixel line
	 * around the outside.
	 */
    ERR_PASS( w_createPixr (pIPWState, imageWidth, imageHeight, 1, 
			    xDPI, yDPI, cFrameBits, &prSeed));

	/* draw the border, here's the line across the top. */
    ERR_PASS_CLEANUP( w_rasterop (pIPWState, 
				  0, 0, NULL,   /* no src */
				  PIX_SET,	/* operation */
				  0, 0,	        /* x, y offsets */
				  imageWidth, 1,/* width, height */
				  prSeed));


	/* line across the bottom */
    ERR_PASS_CLEANUP( w_rasterop (pIPWState,
				  0, 0, NULL,	   /* no src */
				  PIX_SET,	   /* operation */
				  0, imageHeight-1,/* x, y offsets */
				  imageWidth, 1,   /* width, height */
				  prSeed));


	/* line down left side */
    ERR_PASS_CLEANUP( w_rasterop (pIPWState,
				  0, 0, NULL,
				  PIX_SET,
				  0, 0,	 
				  1, imageHeight,	
				  prSeed));

	/* line down right side */
    ERR_PASS_CLEANUP( w_rasterop (pIPWState, 
				  0, 0, NULL,           /* no src */
				  PIX_SET,	        /* operation */
				  imageWidth-1, 0,	/* x, y offsets */
				  1, imageHeight,	/* width, height */
				  prSeed));

	/* generate the mask image we'll actually use.  This is the inverse
	 * of the mask we're given. */
    ERR_PASS_CLEANUP( w_invert(pIPWState, prs, &prs));

	/* prepare args for i_fillClipSeedTwoPass */
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prSeed, &seedImage));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prs, &maskImage));

    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prSeed, &seedPitch));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prs, &maskPitch));

	/* do the first pass, initially clipping the seed with the mask */
    ERR_PASS_CLEANUP( i_fillClipSeedTwoPass (pIPWState, seedImage,
					     seedImage, maskImage,
					     seedPitch, seedPitch, maskPitch,
					     fullImageWidth, fullImageHeight, 
					     cTrue, &writeCount));

	/* do remaining cycles with no initial clip */
    for (i = 1; (i < maxCount) && (writeCount > 0); i++)	
    {
	ERR_PASS_CLEANUP( i_fillClipSeedTwoPass (pIPWState, seedImage,
						 seedImage, maskImage,
						 seedPitch, seedPitch,
						 maskPitch,
						 fullImageWidth,
						 fullImageHeight, cFalse,
						 &writeCount));
    }

	/* we've done as much filling as we're going to do.  Restore the
	 * seed image and OR it with the filled image to produce the
	 * final result. */
    ERR_PASS_CLEANUP( w_invert(pIPWState, prs, &prs));

    if (prs != prd)	/* copy prs to prd */
	ERR_PASS_CLEANUP( w_rasterop (pIPWState, 0, 0, prs, PIX_SRC,
				      0, 0, imageWidth, imageHeight, prd));

	/* insert the newly filled bits */
    ERR_PASS_CLEANUP( w_rasterop (pIPWState, 0, 0, prSeed, PIX_SRC | PIX_DST,
				 0, 0, imageWidth, imageHeight, prd));

    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prSeed));
    *pPrd = prd;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prSeed)
	w_destroyPixr(pIPWState, &prSeed);
    ERR_RETURN;
}

/*

 $Log: fill3.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.29  1998/01/14 00:11:33  ddavies
 fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.28  1997/06/26 20:00:49  ddavies
 fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.27  1997/04/15 21:47:53  ddavies
 fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.26  1997/02/26 19:28:49  jcraig
 fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v

 in w_edgeFillSweep4 when creating the 1 pixel line around the outside
 of the seed image, don't write into the bottom frame (and be
 consistent about how tall the line is on the left and right sides).
 the seed image is destroyed at the end of the procedure so it doesn't
 really matter that we've written into the frame (and the correct
 thing happens since we look at that line when filling the bottom
 line) but it's confusing so just clean it up....

 Revision 1.25  1997/01/14 02:47:18  ddavies
 fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v

 Heavens to Betsy!  FillClipSweep4 failed to return a destination image
 if it allocated one!  This is now fixed.

 Revision 1.24  1996/01/04 20:22:23  ejaquith
 fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.23  1995/12/13  04:06:46  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/17  01:04:58  macleod
 * macleod on Thu Nov 16 17:04:43 PST 1995
 *
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * w_edgeFillSweep4 was only setting seeds along the top and left edges
 *
 * Revision 1.21  1995/11/13  20:10:58  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.20  1995/10/13  22:00:02  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.19  1995/10/03  21:12:57  macleod
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.18  1995/10/02  19:17:34  jcraig
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * w_invert takes a pointer to a pointer to a pixr.
 *
 * Revision 1.17  1995/09/29  02:08:32  ejaquith
 * ejaquith on Thu Sep 28 19:08:24 PDT 1995
 *
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * w_invertIP() to w_invert()
 *
 * Revision 1.16  1995/09/21  22:52:48  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/19  19:20:05  macleod
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.14  1995/09/11  21:55:50  macleod
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.13  1995/09/07  01:30:53  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:18:49  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  20:44:00  macleod
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * Added back #include files
 *
 * Revision 1.10  1995/08/31  21:12:25  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/30  17:20:44  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * edgeFillSweep4() was calling createPixr() with non-null pixr pointer.
 *
 * Revision 1.8  1995/08/29  21:54:23  ddavies
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.7  1995/08/25  01:36:18  ejaquith
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * Fixed warnings
 *
 * Revision 1.6  1995/08/17  21:54:55  macleod
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.5  1994/10/06  20:53:36  lperry
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.4  1994/05/05  01:12:44  ddavies
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * Include adtpros.h instead of ippros.h since this is part of libadt.
 * >> .
 *
 * Revision 1.3  1993/12/03  01:14:29  ddavies
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:11:37  ddavies
 * fill3.c -> /product/ipcore/ipwrap/src/RCS/fill3.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

