/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fill4.c:  includes subroutines
 *
 *  ---------------  Bounding Boxes of Connected Components  ---------------
 *        BOXES  *rabbRuns():       gets the runs for each connected component
 *				    in a rectangle
 *
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "frames.pub"
#include "pixr.h"
#include "fill.h"
#include "geomadt.h"
#include "shrcc.pub"

/* Prototypes */
#include "props.pub"
#include "shrip.pub"
#include "fill.pub"




IP_RCSINFO(RCSInfo, "$RCSfile: fill4.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   gets the runs for each connected component in an image
 *
 * LongDescription:::
 *   Calls a client-supplied procedure for every connected
 *   component in an image.  The client gets a bounding box
 *   plus a list of all of the runs inside the CC. The connected
 *   components are filtered by size (bounding box): if the width or height
 *   is less than the minima passed, the callback is never called.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::           State structure
 *  prs::                 Image to analyze
 *  minWidth::            Min. width of connected component to consider
 *  minHeight::           Min. height of connected component to consider
 *  connectType::         Connectivity type: 
 *	                  Uses 4-connected rules for determining 
 *                        connectedness (only horizontal and vertical count)
 *                        when connectType = 0.
 *	                  Uses 8-connected rules (diagonals count too) 
 *                        when connectType equals 1.  Other values of 
 *                        connectType produces undefined outputs.
 *
 *  pCallbackData::       First argument to callback procedure
 *  callbackProc::        Callback procedure, which is called for each CC
 *
 * Output:::
 *  pCallbackReturnCode:: If any callback returns non-zero, we exit
 *                        immediately, and this gets stuffed with the 
 *                        return code from the callback.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis, run
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rabbRuns (
    IPWState           *pIPWState,
    PIXR               *prs,
    UInt32              minWidth,
    UInt32              minHeight,
    UInt32              connectType,
    void               *pCallbackData,
    Int32	(CDECL *callbackProc)(IPWState  *pIPWState,
				      void *, Int32 x, Int32 y,
				      Int32 w, Int32 h, Int32 m,
				      Int32 numRuns, TCScanLineRun *pRuns),
    Int32              *pCallbackReturnCode)
{
    PROLOGUE;
    UInt8    *pImage;
    Int32     bpl, clipW, clipH;
    

    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetImageAll (pIPWState, prs, &pImage));
    ERR_PASS( w_pixrGetBpl (pIPWState, prs, &bpl));
    ERR_PASS( w_pixrGetWidth (pIPWState, prs, &clipW));
    ERR_PASS( w_pixrGetHeight (pIPWState, prs, &clipH));

	/* Do the real work.  The client's callback is called
	 * for every connected component found.
	 */
    ERR_PASS( i_findConnectedComponentRuns(
				pIPWState,
				pImage,
				bpl,               	/* Bpl */
				cFrameBits, cFrameBits,	/* clipX, clipY */
				clipW,           	/* clipW */
				clipH,			/* clipH */
				minWidth, minHeight,	/* min width & height
							 * of returned rects */
				connectType,		/* 4- vs 8-connected */
				pCallbackData,		/* client's data */
				callbackProc,		/* callback routine */
				pCallbackReturnCode));	/* return code for */
							/* callback proc */
    NORMAL_RETURN;
}



/*

 $Log: fill4.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.19  1998/01/14 00:11:35  ddavies
 fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.18  1997/07/11 01:58:28  ddavies
 fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.17  1996/01/04 20:22:25  ejaquith
 fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.16  1995/12/13  04:06:50  ejaquith
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.15  1995/11/13  20:11:00  ejaquith
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.14  1995/10/13  22:00:05  ejaquith
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.13  1995/10/03  21:13:00  macleod
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.12  1995/09/21  22:52:51  ejaquith
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.11  1995/09/06  20:18:52  ejaquith
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  20:44:06  macleod
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * Added back #include files
 *
 * Revision 1.9  1995/08/31  21:12:30  ejaquith
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/25  01:36:21  ejaquith
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * Fixed warnings
 *
 * Revision 1.7  1995/08/17  21:54:58  macleod
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.6  1995/07/14  02:49:04  ddavies
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * Fix includes and a comment about callback procs.
 *
 * Revision 1.5  1995/06/08  21:08:30  ddavies
 * ddavies on Thu Jun 8 14:06:49 PDT 1995
 *
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * Add mass and number of runs to rabbRuns' callback proc.  This is necessary
 * since this is also a callback proc for i_findConnectedComponentRuns.
 *
 * Revision 1.4  1995/01/11  22:41:29  dferrell
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * Changed include of shrip.pub to shrcc.pub.
 *
 * Revision 1.3  1994/10/06  20:53:40  lperry
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.2  1994/09/16  17:35:46  jcraig
 * fill4.c -> /product/ipcore/ipwrap/src/RCS/fill4.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.1  1994/06/08  04:05:21  ddavies
 * Initial revision
 *

 */

