/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fill5.c:  includes subroutines
 *
 *	  Int32	 createRabbState()	create state used by rabbLite and
 *					rabbRunsLite.
 *
 *	  Int32	 destroyRabbState()	destroys state used by rabbLite and
 *					rabbRunsLite.
 *
 *	  RECTANGLE	*rabbLite()	lightweight version of rabb that
 *					uses preallocated heaps instead of
 *					creating and destroying its own
 *	  
 *        Int32  rabbRunsLite():	lightweight version of rabbRuns that
 *					uses preallocated heaps instead of
 *					creating and destroying its own
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "pixr.h"
#include "fill.h"
#include "geomadt.h"
#include "shrcc.pub"

/* Prototypes */
#include "props.pub"
#include "shrip.pub"
#include "rectops.pub"
#include "fill.pub"




IP_RCSINFO(RCSInfo, "$RCSfile: fill5.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/* Sometimes, it's nice to invoke the connected component finding code on
 * many small regions instead of finding all the CCs in the image in
 * one step.  XIS has talked about finding the runs for each character
 * individually!  Yow.  In cases like this, the time taken to create
 * and tear down the heaps used is significant.
 *
 * And another thing.  Images are almost always encoded as bitmaps.  However,
 * sometimes, they might be represented as XIS spans or something else.
 * In these routines, we allow the client to supply an ipshared level routine
 * that produces all the runs on on scan line.  When called again, it
 * should produce all the runs on the next scan line.  We also let the
 * client provide a state structure which is passed to the run-finding
 * routine.
 *
 * If the pointer to the run finding routine is NULL, we assume the client
 * wants to find runs in an ordinary bitmap and supply the routine and state
 * for that purpose.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create state structure used by lightweight connected component routines
 *
 * LongDescription::: 
 *  This routine creates the state structure used by the lightweight connected
 *  component routines, rabbLite and rabbRunsLite.  This saves the
 *  time needed to create, initialize and destroy the state on each call.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *
 * Output:::
 *  ppNewState::     Opaque state block created
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createRabbState (
    IPWState           *pIPWState,
    void              **ppNewState)
{
    PROLOGUE;
	/* if ppNewState is NULL, just quit */
    ERR_GEN (ppNewState == NULL, cErrInvalidParm);

	/* just do it */
    ERR_PASS( i_createCCState (pIPWState, ppNewState,
			       i_compressAlloc, i_compressFree));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroy the state struct used by lightweight connected component routines
 *
 * LongDescription:::
 *  Destroys the state structure used by the lightweight connected
 *  component routines.  Structure was created by createRabbState.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  ppNewState::     rabb state structure, created by createRabbState,
 *                   cleared on exit
 *
 * Output:::
 *  
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyRabbState (
    IPWState           *pIPWState,
    void              **ppNewState)
{
    PROLOGUE;
	/* if ppNewState is NULL, just quit */
    ERR_GEN (ppNewState == NULL, cErrInvalidParm);

	/* just do it */
    ERR_PASS( i_destroyCCState (pIPWState, ppNewState));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Find bounding rectangles of connected components in an image
 *
 * LongDescription:::
 *  Constructs list of RECTANGLs containing the bounding boxes of
 *  connected components from a bitmap.  Takes existing state so
 *  that the heaps used aren't created and destroyed on every call.
 *  <BR><BR>
 *  Uses 4-connected rules for determining connectedness (only
 *  horizontals and vertical count) when connectType = 0.
 *  Uses 8-connected rules (diagonals count too) when connectType
 *  equals 1.  Othervalues of connectType produces errors.
 *  <BR><BR>
 *  Also allows the client to provide a routine that finds all runs
 *  on each scan line.  This is useful if the image is in some format
 *  other than uncompressed bitmap.  If the runFindingProc pointer is
 *  NULL, we use a procedure for finding runs in uncompressed
 *  bitmaps.
 *  <BR><BR>
 *  If no runFindingProc (and no runData for it) is passed in, a
 *  PIXR and a clipping rectangle must be supplied.  The clipping
 *  rect's coordinates are relative to the inside of the image frame.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::        IPWrap state structure
 *  prS::              Image to analyze, can be NULL if run finding proc
 *                     gets data from somewhere else
 *  pState::           state structure created by w_createRabbState
 *  connectType::      Connectivity type, 0 = 4-connected, 1 = 8-connected
 *  clipX::            X-coordinate of upper left corner of clipping rectangle
 *  clipY::            Y-corrdinate of upper left corner of clipping rectangle
 *  clipW::            Width of clipping rectangle
 *  clipH::            Height of clipping rectangle
 *  minWidth::         Minimum width of connected component to consider
 *  minHeight::        Minimum height of connected component to consider
 *  pRunData::         Pointer to run data, passed to runFindingProc
 *  runFindingProc::   Client-supplied run-finding procedure. takes
 *                     "pRunData" as an argument.
 *
 * Output:::
 *  pRunReturnCode::   Return code from run-finding procedure if it failed
 *  ppRectList::       List of bounding rectangles of all connected components
 *                     in image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis, run
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rabbLite (
    IPWState           *pIPWState,
    PIXR               *prS,
    void               *pState,
    UInt32              connectType,
    Int32               clipX,
    Int32               clipY,
    Int32               clipW,
    Int32               clipH,
    UInt32              minWidth,
    UInt32              minHeight,
    void               *pRunData,
    Int32	(CDECL *runFindingProc)(IPWState *pIPWState,
					void *, TCScanLineRun *pFirst,
					TCBlockHeader *pRunHeap),
    Int32              *pRunReturnCode,
    RECTANGL          **ppRectList)
{
    PROLOGUE;
    Int32		 bpl, frameBits;
    UInt32		 runStateSize;
    UInt8           	*pImageAll;
    Int32		 runReturnCode;
    Int32		 rectReturnCode;
    UInt32		 allocatedHere;	    /* says whether we allocated the
					     * run-finder's data block */
    AddCCPacket   packet;

    ERR_GEN (pState == NULL, cErrInvalidParm);

    ERR_GEN (connectType > 1, cErrInvalidParm);

    allocatedHere = cFalse;

	/* If the client's run finding procedure is NULL, use the one
	 * for uncompressed bitmaps.
	 */
    if (runFindingProc == NULL)
    {
	ERR_GEN_PIXR (prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	ERR_PASS_CLEANUP( i_getFindRunsStateSize (pIPWState, &runStateSize));

	    /* allocate space for the standard run finding routine's state */
	pRunData = (void *) IPW_MALLOC(runStateSize, cAllocatorData);
	ERR_GEN_CLEANUP (pRunData == NULL, cErrNomem);
	allocatedHere = cTrue;

	    /* initialize that state block.  This is sort of tricky.
	     * The clipX and clipY coords specify an offset from
	     * inside the frame and i_findBoundingBoxesForSomeCCs
	     * wants offset from the outside of the frame, so we
	     * add the frame width to them now.
	     *
	     * The X coords specify a horizontal offset from the
	     * image pointer value on each scan line.  The clipY
	     * coord passed to i_findBoundingBoxesForSomeCCs below
	     * is just a label attached to the y coords of the bounding
	     * boxes.  i_findBoundingBoxesForSomeCCs does not move
	     * down by clipY lines before starting to look since it
	     * doesn't know what the image representation is anyway.
	     * Therefore, we have to offset the image pointer by the
	     * number of lines specified in clipY here.  The pointer
	     * we pass in points to the outside left edge of the frame
	     * on the scan line containing the top line of the clipping
	     * rectangle.
	     * We do this weird thing so that all bounding box coordinates
	     * will be relative to the outside of the frame as is
	     * conventional for ipshared routines.  These are the coords
	     * returned by rabb and rabbRuns.
	     */
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bpl));
	ERR_PASS_CLEANUP( w_pixrGetFrame(pIPWState, prS, &frameBits));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pImageAll));
	clipX += frameBits;
	clipY += frameBits;
	ERR_PASS_CLEANUP( i_initFindRunsState (pIPWState,
					       (FindRunsState *)pRunData,
						pImageAll + (clipY * bpl),
						clipX,
						clipX + clipW - 1,
						bpl));

	/* finally, set the pointer to the run finding routine */
	runFindingProc = i_findAllRunsInLineWithState;
	pRunReturnCode = &runReturnCode;
    }

	/* Do the real work.  The procedure addCCRectToList is called
	 * for each bounding box found.  First, initialize both callback
	 * return codes.  Remember to adjust the coordinates so they're
	 * relative to the OUTSIDE of the image frame.
	 */
    runReturnCode = rectReturnCode = cSuccessful;	/* assume the best */

    packet.rectList = NULL;
    ERR_PASS_CLEANUP( i_findBoundingBoxesForSomeCCs (
			       pIPState,
                      	       (CCState *)pState,	/* internal state */
				clipY,			/* clipY */
				clipH,
				minWidth, minHeight,	/* min width & height
							 * of returned rects */
				connectType,
				pRunData,		/* data for runfinder*/
				runFindingProc,		/* the run finder */
				pRunReturnCode,
				(void *)&packet,	/* packet which
							 * contains
							 * ptr to
							 * rects as well 
							 * as the ipstate.
							 */
				w_addCCRectToList,	/* callback routine */
				&rectReturnCode));

	/* if we get here, it must have worked.  Return the rect list.
	 */
    *ppRectList = packet.rectList;
    if (allocatedHere) IPW_FREE(pRunData, cAllocatorData);
    NORMAL_RETURN;

	/* Free the runfinder's data if we allocated it and return */
ERR_CLEANUP:
    if (allocatedHere) IPW_FREE(pRunData, cAllocatorData);
    if (packet.rectList) w_destroyRectList(pIPWState, &packet.rectList);
    *ppRectList = NULL;
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Call client-supplied procedure for every connected component in an image
 *
 * LongDescription:::
 *  Calls a client-supplied procedure for every connected
 *  component in an image.  The client gets a bounding box
 *  plus a list of all of the runs inside the CC.
 *  <BR><BR>
 *  
 *  Uses 4-connected rules for determining connectedness (only
 *  horizontals and vertical count) when connectType = 0.
 *  Uses 8-connected rules (diagonals count too) when connectType
 *  equals 1.  Othervalues of connectType produces errors.
 *  <BR><BR>
 *  
 *  callbackProc is called with pCallbackData as the first arg.  If it
 *  returns with something other than cSuccessful,
 *  *pCallbackReturnCode will contain callback proc's return
 *  code when this procedure returns.
 *  <BR><BR>
 *  
 *  Also allows the client to provide a routine that finds all runs
 *  on each scan line.  This is useful if the image is in some format
 *  other than uncompressed bitmap.  If the runFindingProc pointer is
 *  NULL, we use a procedure for finding runs in uncompressed
 *  bitmaps.
 *  <BR><BR>
 *  
 *  The Y coordinate (clipY) and height of the region of interest
 *  (clipH) must always be supplied.  The clipY coord is used to set
 *  the Y coordinates of rectangles found.  The clipH arg tells us
 *  how many scan lines to process.  Both clipY and clipX are relative 
 *  to the inside of the image frame to be consistent with ipwrap 
 *  convention.
 *  <BR><BR>
 *  
 *  HOWEVER, the x and y coordinates supplied to the callbackProc
 *  will be relative to the OUTSIDE of the frame to be consistent with
 *  ipwrap conventions.  Sorry!!  Just subtract the width of the frame.
 *  <BR><BR>
 *  
 *  If no runFindingProc (and no runData for it) is passed in, the 
 *  horizontal boundaries of the region must also be supplied.  The 
 *  coordinates are relative to the inside of the image frame.  If 
 *  the runFindingProc is passed in, the horizontal coordinates are
 *  ignore because it is assumed that they are somehow encoded in the
 *  pRunData.  If no runFindingProc is passed in, then the clipY
 *  coordinate also becomes the real upper y coordinate of the clipping
 *  rectangle used in the bitmap.  Otherwise, it's just a number that's
 *  incremented for each scan line and passed back in the callbackProc's
 *  y arg.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::           IPWrap state structure
 *  prS::                 Image to analyze, can be NULL if run finding proc
 *                        gets data from somewhere else
 *  pState::              state structure created by w_createRabbState
 *  connectType::         Connectivity type, 0 = 4-connected, 1 = 8-connected
 *  clipX::               X-coordinate of upper left corner of clipping 
 *                        rectangle
 *  clipY::               Y-corrdinate of upper left corner of clipping 
 *                        rectangle
 *  clipW::               Width of clipping rectangle
 *  clipH::               Height of clipping rectangle
 *  minWidth::            Minimum width of connected component to consider
 *  minHeight::           Minimum height of connected component to consider
 *  pRunData::            Pointer to run data, passed to runFindingProc
 *  runFindingProc::      Client-supplied run-finding procedure. Takes
 *                        "pRunData" as a first argument.
 *  pCallbackData::       Client-supplied thing to pass to callback proc
 *  callbackProc::        Callback procedure, called when a CC is found
 *
 * Output:::
 *  pRunReturnCode::      Return code from run-finding callback, if it failed
 *  pCallbackReturnCode:: Return code from connected-componet-found callback,
 *                        if it failed
 *
 * Returns:::
 *  error status.  An error is returned if both prS and the runFindingProc
 *  are NULL
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis, run
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rabbRunsLite (
    IPWState           *pIPWState,
    PIXR               *prS,
    void               *pState,
    UInt32              connectType,
    Int32               clipX,
    Int32               clipY,
    Int32               clipW,
    Int32               clipH,
    UInt32              minWidth,
    UInt32              minHeight,
    void               *pRunData,
    Int32	(CDECL *runFindingProc)(IPWState *pIPWState,
					void *, TCScanLineRun *pFirst,
					TCBlockHeader *pRunHeap),
    void               *pCallbackData,
    Int32	(CDECL *callbackProc)(IPWState *pIPWState,
				      void *, Int32 x, Int32 y,
				      Int32 w, Int32 h, Int32 m,
				      Int32 numRuns, TCScanLineRun *pRuns),
    Int32              *pRunReturnCode,
    Int32              *pCallbackReturnCode)
{
    PROLOGUE;
    Int32		 bpl, frameBits;
    UInt8           	*pImageAll;
    UInt32		 runStateSize;
    Int32		 runReturnCode;
    UInt32		 allocatedHere;		/* says whether we allocated the
					 * run-finder's data block */

    ERR_GEN ((prS == NULL) && (runFindingProc == NULL), cErrInvalidParm);
    ERR_GEN (pState == NULL, cErrInvalidParm);
    ERR_GEN (connectType > 1, cErrInvalidParm);

    allocatedHere = cFalse;

	/* If the client's run finding procedure is NULL, use the one
	 * for uncompressed bitmaps.
	 */
    if (runFindingProc == NULL)
    {
	ERR_GEN_PIXR (prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	ERR_PASS_CLEANUP( i_getFindRunsStateSize (pIPWState, &runStateSize));

	    /* allocate space for the standard run finding routine's state */
	pRunData = (void *)IPW_MALLOC(runStateSize, cAllocatorData);
	ERR_GEN_CLEANUP (pRunData == NULL, cErrNomem);
	allocatedHere = cTrue;

	    /* initialize that state block.  This is sort of tricky.
	     * The clipX and clipY coords specify an offset from
	     * inside the frame and i_findBoundingBoxesForSomeCCs
	     * wants offset from the outside of the frame, so we
	     * add the frame width to them now.
	     *
	     * The X coords specify a horizontal offset from the
	     * image pointer value on each scan line.  The clipY
	     * coord passed to i_findBoundingBoxesForSomeCCs below
	     * is just a label attached to the y coords of the bounding
	     * boxes.  i_findBoundingBoxesForSomeCCs does not move
	     * down by clipY lines before starting to look since it
	     * doesn't know what the image representation is anyway.
	     * Therefore, we have to offset the image pointer by the
	     * number of lines specified in clipY here.  The pointer
	     * we pass in points to the outside left edge of the frame
	     * on the scan line containing the top line of the clipping
	     * rectangle.
	     * We do this weird thing so that all bounding box coordinates
	     * will be relative to the outside of the frame as is
	     * conventional for ipshared routines.  These are the coords
	     * returned by rabb and rabbRuns.
	     */
	ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prS, &bpl));
	ERR_PASS_CLEANUP( w_pixrGetFrame (pIPWState, prS, &frameBits));
	ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prS, &pImageAll));
	clipX += frameBits;
	clipY += frameBits;
	ERR_PASS_CLEANUP( i_initFindRunsState (pIPWState,
					       (FindRunsState *)pRunData,
					       pImageAll + (clipY * bpl),
					       clipX,
					       clipX + clipW - 1,
					       bpl));

	    /* finally, set the pointer to the run finding routine */
	runFindingProc = i_findAllRunsInLineWithState;
	pRunReturnCode = &runReturnCode;
    }

	/* Do the real work.  Remember to adjust the coordinates so they're
	 * relative to the OUTSIDE of the image frame.
	 */
    runReturnCode = cSuccessful;			/* assume the best */
    ERR_PASS_CLEANUP( i_findRunsForSomeCCs (pIPWState,
			     (CCState *)pState,		/* internal state */
			     clipY,			/* clipY */
			     clipH,
			     minWidth, minHeight,	/* min width & height
							 * of returned rects */
			     connectType,
			     pRunData,			/* data for runfinder*/
			     runFindingProc,		/* the run finder */
			     pRunReturnCode,
			     pCallbackData,
			     callbackProc,		/* callback routine */
			     pCallbackReturnCode));

	/* if we get here, it must have worked. */
    if (allocatedHere) IPW_FREE(pRunData, cAllocatorData);
    NORMAL_RETURN;

	/* Free the runfinder's data if we allocated it and return */
ERR_CLEANUP:

    if (allocatedHere) IPW_FREE(pRunData, cAllocatorData);

    ERR_RETURN;
}


/*

 $Log: fill5.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.18  1998/01/14 00:11:36  ddavies
 fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.17  1997/07/11 01:58:31  ddavies
 fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.16  1996/05/21 19:11:35  ddavies
 ddavies on Tue May 21 12:10:49 PDT 1996

 fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v

 Change check for a NULL pixr in rabbRunsLite to a check for a NULL
 pixr AND a null runFindingProc.  One must be provided.

 Revision 1.15  1996/01/04 20:22:28  ejaquith
 fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.14  1995/12/13  04:06:54  ejaquith
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.13  1995/11/13  20:11:03  ejaquith
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.12  1995/10/13  22:00:08  ejaquith
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.11  1995/10/03  21:12:12  macleod
 * macleod on Tue Oct 3 14:11:57 PDT 1995
 *
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * Updated comment block; changed interface of rabbRunsLite so that both
 * output arguments were at the end.
 *
 * Revision 1.10  1995/09/21  22:52:53  ejaquith
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.9  1995/09/11  21:56:01  macleod
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.8  1995/09/06  20:18:55  ejaquith
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.7  1995/09/01  20:44:10  macleod
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * Revision 1.6  1995/08/31  21:12:35  ejaquith
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.5  1995/08/25  01:36:25  ejaquith
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * Fixed warnings
 *
 * Revision 1.4  1995/08/24  21:25:56  jcraig
 * jcraig on Thu Aug 24 14:25:11 PDT 1995
 *
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * fix memory leak in rabbLite and rabbRunsLite
 *
 * Revision 1.3  1995/08/23  20:18:41  jcraig
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * fix packet that will given to callback so that it is initialized
 * properly.  changed so that it contains a pointer to a rectangle not
 * a pointer to a pointer to a rectangle.
 *
 * Revision 1.2  1995/08/17  21:55:02  macleod
 * fill5.c -> /product/ipcore/ipwrap/src/RCS/fill5.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.1  1995/07/14  02:53:33  ddavies
 * Initial revision
 *

 */


