/*<HTML><HEAD><TITLE>fill8.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1998, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fill8.c:  includes subroutines
 *
 *      w_grayPropagateMask () - Geodesic grayscale propagation with mask
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "frames.pub"
#include "pixr.h"
#include "fill.h"
#include "geomadt.h"
#include "shrcc.pub"
#include "imageref.pub"

/* Prototypes */
#include "props.pub"
#include "prxform.pub"
#include "shrip.pub"
#include "fill.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: fill8.c,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:22 $ */

/* There are gray propagation routines in fill7.c.  Those routines require
 * the user to tag (set to 0) pixels in the image array where propagation
 * should occur.  The routines in this file require only that the user
 * supply a binary mask which indicates where propagation should happen.
 * These routines are faster.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Propagate grayscale values to unfilled neighbors, 4-connected, using mask
 *
 * LongDescription:::
 *  Given a grayscale or color image, propagate all seed pixels to non-seed
 *  neighbors using a geodesic dilation algorithm, such that seed pixels
 *  expand into neighborhoods evenly. The neighborhood is defined as
 *  the pixels left, right, above, and below.  The seed pixels are
 *  identified by pixels corresponding to 0's in a binary mask image.
 *
 * PrivateDescription:::
 *  This function uses an IPShared routine to do the real work.  It does
 *  expand the binary mask into an array of 8-bit or 16-bit masks before
 *  calling the ipshared routine.
 *
 * Input:::
 *  pIPWState::     State structure
 *  prMask::	    Pointer to binary mask image.  This image has 1's where
 *		    the seed pixels are, or 0's where the seed pixels are
 *		    depending on the value of onesAreSeed below.
 *  onesAreSeed::   Boolean which is set to cTrue if 1's in prMask indicate
 *		    seed pixels.  If cFalse, 0's in prMask indicate seed
 *		    pixels.
 *  maxDist::	    The maximum distance over which seeds will be
 *		    propagated strictly correctly.  At greater distances,
 *		    seeds will still be propagated but the final value
 *		    of a pixel may be the value of a seed pixel that's
 *		    pretty close as opposed to the strictly closest
 *		    seed pixel.  The only two legal values for this
 *		    parameter are 255 and 65535.
 *
 * Output:::
 *  prImage::       Image in which seeds will be propagated.
 *
 * Returns:::
 *  Error status
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *  This is descended from "Morphological Algorithms" by Luc Vincent,
 *  1991, Harvard Robotics Laboratory technical report no. 91-12.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_grayPropagate4Mask (
    IPWState *pIPWState,
    PIXR     *prMask,
    Bool      onesAreSeed,
    Int32     maxDist,
    PIXR     *prImage)
{
    PROLOGUE;
    Int32     wS, hS, wM, hM, bplS, bplDist, channels, i;
    PIXR     *prDistance = NULL;
    PIXR     *prChannel = NULL;
    UInt8    *pS0, *pS1, *pS2, *pDist;

    ERR_GEN (prImage == NULL, cErrNullPtr);
    ERR_GEN (prMask == NULL, cErrNullPtr);
    ERR_GEN ((maxDist != 255) && (maxDist != 65535), cErrInvalidParm);
    ERR_GEN_PIXR (prImage, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);
    ERR_GEN_PIXR (prMask, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, prMask, &wM));
    ERR_PASS (w_pixrGetHeight (pIPWState, prMask, &hM));
    ERR_GEN( (wS != wM) || (hS != hM), cErrSrcDimensNotEqual);

    ERR_PASS (w_pixrGetBpl (pIPWState, prImage, &bplS));

	/* Create a distance image.  We either have 8 bits/distance value
	 * or 16 bits/distance value.
	 * We want the distance image to have 0's where there are seed
	 * pixels, so invert the incoming mask if onesAreSeed is true.
	 */
    if (maxDist == 255)
    {
	if (onesAreSeed)
	    ERR_PASS (w_depth1To8 (pIPWState, prMask, cTrue, &prDistance));
	else
	    ERR_PASS (w_depth1To8 (pIPWState, prMask, cFalse, &prDistance));
    }
    else	/* maxDist must be 65535 */
    {
	if (onesAreSeed)
	    ERR_PASS (w_depth1To16 (pIPWState, prMask, cTrue, &prDistance));
	else
	    ERR_PASS (w_depth1To16 (pIPWState, prMask, cFalse, &prDistance));
    }

	/* get the info about the distance image */
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prDistance, &bplDist));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prDistance, &pDist));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prImage, &channels));

	/* If this is a 3-channel image, use the special color
	 * propagation function.  Otherwise, process each channel
	 * individually.
	 */
    if (channels == 3)
    {
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prImage, 0, 
					    &prChannel));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prChannel, &pS0));
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prImage, 1, 
					    &prChannel));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prChannel, &pS1));
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prImage, 2, 
					    &prChannel));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prChannel, &pS2));

	/* Do the propagation on the channel. */
	ERR_PASS_CLEANUP ( i_propagateColor4Mask(pIPWState,
						 pS0, pS1, pS2, wS, hS,
					         bplS, 8, pDist,
					         bplDist,
						 (maxDist == 255 ? 8 : 16)));
    }
    else
    {
	for (i = 0; i < channels; i++)
	{
		/* If this is not the first time through, re-initialize
		 * the distance image.
		 */
	    if (i > 0)
	    {
		if (maxDist == 255)
		{
		    ERR_PASS (w_depth1To8 (pIPWState, prMask, cFalse,
					   &prDistance));
		}
		else	/* maxDist must be 65535 */
		{
		    ERR_PASS (w_depth1To16 (pIPWState, prMask, cFalse,
					    &prDistance));
		}
	    }
	
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prImage, i, 
						&prChannel));
    
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prChannel, &pS0));
    
	    /* Do the propagation on the channel. */
	    ERR_PASS_CLEANUP (i_propagateGray4Mask (pIPWState, pS0, wS, hS,
						    bplS, 8, pDist,
						    bplDist,
						    (maxDist == 255 ? 8 : 16)));
    
	}
    }

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prDistance));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prDistance)
	w_destroyPixr (pIPWState, &prDistance);
    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Propagate wighted grayscale values to unfilled neighbors, 4-connected,
 *  using mask
 *
 * LongDescription:::
 *  Given a grayscale  image, propagate all seed pixels to non-seed
 *  neighbors using a geodesic dilation algorithm, such that seed pixels
 *  expand into neighborhoods evenly. The neighborhood is defined as
 *  the pixels left, right, above, and below.  The seed pixels are
 *  identified by pixels corresponding to 0's in a binary mask image.
 *
 * PrivateDescription:::
 *  This routine is same as w_grayPropagate4Mask in propagation. But we
 *  use distance as weights to modified filled values  and just work on
 *  one channel gray scal image. 
 *
 * Input:::
 *  pIPWState::     State structure
 *  prMask::	    Pointer to binary mask image.  This image has 1's where
 *		    the seed pixels are, or 0's where the seed pixels are
 *		    depending on the value of onesAreSeed below.
 *  onesAreSeed::   Boolean which is set to cTrue if 1's in prMask indicate
 *		    seed pixels.  If cFalse, 0's in prMask indicate seed
 *		    pixels.
 * Output:::
 *  prImage::       Image in which seeds will be propagated.
 *
 * Returns:::
 *  Error status
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  ddavies/ zhu::
 *
 * Citations:::
 *  This is descended from "Morphological Algorithms" by Luc Vincent,
 *  1991, Harvard Robotics Laboratory technical report no. 91-12.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_grayPropagate4MaskBi (
    IPWState *pIPWState,
    PIXR     *prMask,
    Bool      onesAreSeed,
    PIXR     *prImage)
{
    PROLOGUE;
    Int32     wS, hS, wM, hM, bplS, bplDist, channels;
    PIXR     *prDistance = NULL;
    UInt8    *pS, *pDist;

    ERR_GEN (prImage == NULL, cErrNullPtr);
    ERR_GEN (prMask == NULL, cErrNullPtr);
    ERR_GEN_PIXR (prImage, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);
    ERR_GEN_PIXR (prMask, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, prMask, &wM));
    ERR_PASS (w_pixrGetHeight (pIPWState, prMask, &hM));
    ERR_GEN( (wS != wM) || (hS != hM), cErrSrcDimensNotEqual);

    ERR_PASS (w_pixrGetChannelCount (pIPWState, prImage, &channels));
    ERR_GEN ( channels != 1, cErrPixr_channels );

	/* Create a distance image.  We either have 8 bits/distance value
	 * or 16 bits/distance value.
	 * We want the distance image to have 0's where there are seed
	 * pixels, so invert the incoming mask if onesAreSeed is true.
	 */

    if (onesAreSeed)
	ERR_PASS (w_depth1To16 (pIPWState, prMask, cTrue, &prDistance));
    else
	ERR_PASS (w_depth1To16 (pIPWState, prMask, cFalse, &prDistance));

	/* get the info about the distance image */
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prDistance, &bplDist));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prDistance, &pDist));

    ERR_PASS (w_pixrGetBpl (pIPWState, prImage, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prImage, &pS));
    
	    /* Do the propagation on the channel. */
    ERR_PASS_CLEANUP (i_propagateGray4Mask (pIPWState, pS, wS, hS,
					    bplS, 8, pDist,
					    bplDist, 16));

    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prDistance, &pDist));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prImage, &pS));
    ERR_PASS_CLEANUP (i_grayMinusDist (pIPWState, pS, wS, hS, bplS,
					   pDist, bplDist));
    
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prDistance));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prDistance)
	w_destroyPixr (pIPWState, &prDistance);
    ERR_RETURN;

}


/*
 * $Log: fill8.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:22  jlin
 * Imported sources
 *
 * Revision 1.3  1998/06/04 18:20:48  xzhu
 * xzhu on Thu Jun 4 11:20:39 PDT 1998
 *
 * fill8.c -> /product/ipcore/ipwrap/src/RCS/fill8.c,v
 *
 * Added a new function.
 *
 * Revision 1.2  1998/02/13 01:34:03  ddavies
 * fill8.c -> /product/ipcore/ipwrap/src/RCS/fill8.c,v
 *
 * Add the onesAreSeed parameter.  This indicates whether 1's in the binary
 * mask are seed pixels or non-seed pixels.
 *
 * Revision 1.1  1998/02/10 20:01:04  ddavies
 * Initial revision
 *
 */


