/*************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *	findtext.c:
 *		Private routines to find text segments in an image.
 *
 */


#include <stdio.h>
#include <stdlib.h>

#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "morph.pub"
#include "defines.pub"
#include "morphg.h"
#include "shrcc.pub"
#include "logicops.h"
#include "hauscomp.h"
#include "descrip.h"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "descrip.prv"
#include "rescale.pub"
#include "strel.pub"
#include "intops.pub"
#include "fill.pub"
#include "rectops.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif


#define SMALL_BOX_WIDTH		3
#define SMALL_BOX_HEIGHT	3
#define TEXTSEGMENT_LINEWIDTH	3

/* values at 300 dpi */
#define MIN_TSEG_WIDTH		100
#define MIN_TSEG_HEIGHT		40



/* These are ALWAYS 0 on checkin/release, and 1 only when checked out 
 * if some printed debugging information is desired.
 */
#define DEBUG_PROGRESS      0
#define DEBUG_WORD          0
#define DEBUG_TEXTSEGMENT   0
#define DEBUG_LANDMARKS     0
#define DEBUGPR_TYPE        TIFF_DEFAULT




/* Comments included from David Hirsch:
 *
 *  In this second attempt I want to try the following approach:
 *  1) Only box the connected components once
 *  2) Use histograms to determine interline spacing and perhaps
 *     significant vertical spacings (column breaks, etc.)
 *  3) Use these derived numbers to help group the set of boxes
 *     into clusters that will be the landmarks for the page.
 * Somehow need to figure out a way to avoid including abnormal boxes
 * with the others. That is, if there is a line that cuts across a bunch
 * of text we somehow want to see this as abnormal and thus leave its box
 * out of the clusters. Maybe looking at runlengths of interword and interline
 * spacings would help the outliers show up.
 */







/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arbitrary rescaling of single or multi-channel, 1, 4, or 8 bit PIXRs.
 *
 * LongDescription:::
 *  This uses rescale factors and calls w_scaleToSize() appropriately.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  prs::               Source PIXR to be scaled
 *  scaleX::            X-axis rescaling factor
 *  scaleY::            Y-axis rescaling factor
 *
 * Output:::
 *  pprd::              Scaled dest PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  hirsch::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_scaleByFactors (
    IPWState *pIPWState,
    PIXR     *prs,
    Float32   scaleX,
    Float32   scaleY,
    PIXR    **pprd)
{
    PROLOGUE;
    Int32         w, h, newW, newH;
    PIXR         *prd = NULL;

    ERR_GEN(prs == NULL  ||  pprd == NULL,  cErrInvalidParm);

    if ((scaleX == 1.0) && (scaleY == 1.0))
    {
        ERR_PASS( w_copy( pIPWState, prs, &prd));
    }
    else
    {
	ERR_PASS( w_pixrGetWidth( pIPWState, prs, &w));
	ERR_PASS( w_pixrGetHeight( pIPWState, prs, &h));

	newW = (Int32)(scaleX * (Float32)w);
	newH = (Int32)(scaleY * (Float32)h);

	ERR_PASS( w_scaleToSize( pIPWState, prs, newW, newH, &prd));
    }

    *pprd = prd;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dilate a PIXR by a STREL, adding result to dest PIXR.
 *
 * LongDescription:::
 *  Dilates a pixr by a STREL.  Then adds the result to the destination
 *  PIXR.  Differs from the usual w_dilate(), which simply puts the dilated
 *  result in the destination.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  prs::               Source binary PIXR
 *  strel::             Ptr to STREL
 *
 * Output:::
 *  prd::               Dest binary PIXR (must exist).
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_dilateAdd (
    IPWState *pIPWState,
    PIXR     *prs,
    STREL    *strel,
    PIXR     *prd)
{
    PROLOGUE;
    Int32           i, j, w, h, nx, ny, cx, cy;

    ERR_GEN(prs==NULL || prd==NULL || strel==NULL,  cErrInvalidParm);

    nx = strel->nx;
    ny = strel->ny;
    cx = strel->cx;
    cy = strel->cy;

    ERR_PASS( w_pixrGetDepth( pIPWState, prs, &i));
    ERR_PASS( w_pixrGetDepth( pIPWState, prd, &j));
    ERR_GEN(i != j, cErrSrcDimensNotEqual);

    ERR_PASS( w_pixrGetWidth( pIPWState, prs, &w));
    ERR_PASS( w_pixrGetHeight( pIPWState, prs, &h));
    

    for (j = 0; j < ny; j++)
	for (i = 0; i < nx; i++)
	    if (strel->data[j][i] == 1)
	    {
		ERR_PASS( w_rasterop( pIPWState, 
				      0, 0, prs, OP_HU,
				      i - cx, j - cy, w, h, prd));
	    }
	    else if (strel->data[j][i] == 2)
	    {
		ERR_PASS( w_rasterop( pIPWState, 
				      0, 0, prs, OP_MU,
				      i - cx, j - cy, w, h, prd));
	    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  8-connected bounding box morphological fill.
 *
 * LongDescription:::
 *           Iteratively fills into prd, starting with prs, and on each
 *           iteration tests for completion.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  prs::               Seed binary PIXR
 *  iter::              Maximum number of iterations.  (We'll abort earlier
 *                      when an iteration produces no change).
 *
 * Output:::
 *  prd::               Dest binary PIXR (must exist).
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, binary
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  hirsch::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_bbfill8Morph (
    IPWState *pIPWState,
    PIXR     *prs, 
    Int32     iter,
    PIXR     *prd)	
{
    PROLOGUE;
    Int32         i;
    Bool          equal;
    PIXR         *prt1= NULL;
    PIXR         *prt2= NULL;   /* put result from previous iteration here */
    STREL        *pStrel;



    ERR_PASS(w_clonePixrNoInit( pIPWState, prs, &prt1));
    ERR_PASS_CLEANUP( w_clonePixrNoInit( pIPWState, prs, &prt2));

    ERR_PASS_CLEANUP( w_copy( pIPWState, prs, &prd));
    ERR_PASS_CLEANUP( w_copy( pIPWState, prs, &prt2));

    for (i = 0; i < iter; i++)
    {
	pStrel = NULL;
	ERR_PASS_CLEANUP( w_getStrel( pIPWState, strel2dp, &pStrel));
	ERR_PASS_CLEANUP( w_erode( pIPWState, prd, pStrel, &prt1));
	
	pStrel = NULL;
	ERR_PASS_CLEANUP( w_getStrel( pIPWState, strel2dm, &pStrel));
	ERR_PASS_CLEANUP( wp_dilateAdd( pIPWState, prt1, pStrel, prd));
	
	pStrel = NULL;
	ERR_PASS_CLEANUP( w_getStrel( pIPWState, strel2dm, &pStrel));
	ERR_PASS_CLEANUP( w_erode( pIPWState, prd, pStrel, &prt1));

	pStrel = NULL;
	ERR_PASS_CLEANUP( w_getStrel( pIPWState, strel2dp, &pStrel));
	ERR_PASS_CLEANUP( wp_dilateAdd( pIPWState, prt1, pStrel, prd));
	
	ERR_PASS_CLEANUP( w_areEqual( pIPWState, prd, prt2, &equal));
	
	if (equal)
	    break;
	
	/* for comparison after next iteration */
	ERR_PASS_CLEANUP( w_copy( pIPWState, prd, &prt2));
    }

    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &prt1));
    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &prt2));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr( pIPWState, &prt1);
    w_destroyPixr( pIPWState, &prt2);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Find rectangle list of bounding boxes of "words" in an image.
 *
 * PrivateDescription:::
 *   Do a morphological close on the image.
 *   Call wp_bbfill8Morph().
 *   Get the connected component bounding boxes using w_rabb(),
 *   and return this as the rect list.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR
 *  xres::              Positive X-axis resolution of image
 *  yres::              Positive Y-axis resolution of image
 *
 * Output:::
 *  ppdrects::          Ptr to ptr to list of resulting rectangles that are
 *                      deemed to be "words" on the page.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, binary, image classification, segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  hirsch::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_words1bb (
    IPWState  *pIPWState,
    PIXR      *pixr, 
    Int32      xres,
    Int32      yres,
    RECTANGL **ppdrects)
{
    PROLOGUE;
    PIXR           *prT1=NULL, *prT2=NULL;
    RECTANGL       *bbRects=NULL;
    STREL          *strelh=NULL, *strelv=NULL;

  
    if (xres < 93)
    {
	ERR_PASS( w_getStrel( pIPWState, strel3h, &strelh));
    }
    else if (xres < 131)
    {
	ERR_PASS( w_getStrel( pIPWState, strel2h, &strelh));
    }
    else {
	ERR_PASS( w_getStrel( pIPWState, strel5h, &strelh));
    }

    if (yres < 93)
    {
	ERR_PASS( w_getStrel( pIPWState, strel3v, &strelv));
    } 
    else if (yres < 131)
    {
	ERR_PASS( w_getStrel( pIPWState, strel2v, &strelv));
    }
    else
    {
	ERR_PASS( w_getStrel( pIPWState, strel5v, &strelv));
    }

    ERR_PASS( w_copy( pIPWState, pixr, &prT1));

    ERR_PASS_CLEANUP( w_clonePixrNoInit( pIPWState, prT1, &prT2));

    /* Do rectangular morphological close on image */
    ERR_PASS_CLEANUP( w_close( pIPWState, prT1, strelh, &prT2));
    ERR_PASS_CLEANUP( w_close( pIPWState, prT2, strelv, &prT1));

    ERR_PASS_CLEANUP( wp_bbfill8Morph( pIPWState, prT1, 2, prT2));

    /* Get the connected component bounding boxes */
    ERR_PASS_CLEANUP( w_rabb( pIPWState, prT2, 0, 0, &bbRects));
  
  
    /* debug code: show word boxing */
#if  DEBUG_WORD
    {
	PIXR     *prt= NULL;
      
	ERR_PASS_CLEANUP( w_clonePixr( pIPWState, pixr, &prt));
      
	ERR_PASS_CLEANUP( w_displayRectList( pIPWState, bbRects, 2, prt));

#if 0
	ERR_PASS_CLEANUP( w_logicOp( pIPWState, pixr, XOR, prt));
#endif
	
	ERR_PASS_CLEANUP( w_write( pIPWState, "junk.word1b", prt,
				  DEBUGPR_TYPE));
	ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &prt));
    }
#endif  /* DEBUG_WORD */
  
    /* clean up */
    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &prT1));
    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &prT2));
  
    *ppdrects = bbRects;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyRectList(pIPWState, &bbRects);
    w_destroyPixr( pIPWState, &prT1);
    w_destroyPixr( pIPWState, &prT2);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Attempt to find regions of text areas on the page, and return their rects
 *
 * PrivateDescription:::
 *  Current method::
 *  <PRE>
 *   We reduce the image to 100dpi, then do a 2x2 dilation of it.
 *   Then we do so many iterations of processing as follows:
 *     Find words bounding boxes using wp_words1bb().
 *     Find rectangles of connected components using w_rabb().
 *     Merge overlapping rects.
 *     Remove small rects.
 *     If we are computing the number of iterations to do:
 *       If 10 rectangles or less are left, or the number of them changed
 *       by only 10% since the last iteration, we are done.
 *     Else we had some number in mind on entry, so stop when we reach it.
 *   Return the remaining rect list, sorted by 
 *     w_rectLessorUpperLeftPosition().
 *  </PRE>
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  sourcePr::          Ptr to binary PIXR to find text segments in.
 *  xres::              Positive X-axis resolution of image
 *  yres::              Positive Y-axis resolution of image
 *  iterations::        Number of times to try to iterate over image,
 *                      expanding and merging rough guesses of text rectangles.
 *                      Pass in value <= 0 for routine to determine this for
 *                      you.
 *
 * Output:::
 *  pIterations::       Ptr to number of iterations actually done, if routine
 *                      decided how many iterations to do.
 *  ppRects::           Ptr to ptr to list of resulting rectangles that are
 *                      deemed to be "text segments" on the page.
 *                      Ptr must be valid and *ppRects must be NULL on input.
 *  nRects::            Ptr to number of rects in rect list "ppRects".
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, binary, image classification, segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  hirsch::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_findTextSegments(
    IPWState  *pIPWState,
    PIXR      *sourcePr,
    Int32      xres, 
    Int32      yres, 
    Int32      iterations,
    Int32     *pIterations,
    RECTANGL **ppRects,
    Int32     *nRects)
{
    PROLOGUE;
    RECTANGL     *rl= NULL, *workRectList= NULL, *rabbl= NULL, *flatl= NULL;
    RECTANGL     *pRect;
    PIXR         *workPr = NULL, *prTemp = NULL;
    STREL        *pStrel;
    UInt8         computeIterations;
    Int32         i, lastnum = -1, num;
    Int32         smallBoxWidth, smallBoxHeight;
    Int32         linewidth, newresx, newresy, w, h;
    Float32       redx, redy;
#if DEBUG_LANDMARKS
    PIXR            *debugPr = NULL;
#endif

    ERR_GEN_PIXR(sourcePr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN(pIterations == NULL  ||  nRects == NULL  ||  ppRects == NULL,
	    cErrNullPtr);
    ERR_GEN(*ppRects != NULL, cErrPtrNotNull);
    ERR_GEN(xres <= 0, cErrInvalidParm);
    ERR_GEN(yres <= 0, cErrInvalidParm);
    
  

    /* Reduce to our "canonical" size image that we like to work with. */
    /* (want to reduce everything to 100 dpi) */
    redx = (Float32)xres / 100;
    redy = (Float32)yres / 100;
    ERR_PASS( wp_scaleByFactors( pIPWState, sourcePr, 
				 (Float32)1 / redx,
				 (Float32)1 / redy,
				 &workPr));
    
    /* Do a 2x2 dilation in-place */
    ERR_PASS_CLEANUP( w_clonePixrNoInit( pIPWState, workPr, &prTemp));

    pStrel = NULL;
    ERR_PASS_CLEANUP( w_getStrel( pIPWState, strel2h, &pStrel));
    ERR_PASS_CLEANUP( w_dilate( pIPWState, workPr, pStrel, &prTemp));
    
    pStrel = NULL;
    ERR_PASS_CLEANUP( w_getStrel( pIPWState, strel2v, &pStrel));
    ERR_PASS_CLEANUP( w_dilate( pIPWState, prTemp, pStrel, &workPr));

    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &prTemp));

    newresx = (Int32)(((Float32)xres)/redx);
    newresy = (Int32)(((Float32)yres)/redy);
    linewidth = (TEXTSEGMENT_LINEWIDTH * MIN(newresx, newresy))/75;
    smallBoxWidth  = (SMALL_BOX_WIDTH  * newresx)/75;
    smallBoxHeight = (SMALL_BOX_HEIGHT * newresy)/75;

#if DEBUG_TEXTSEGMENT
    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
	     "\n res(%ld,%ld), red(%f,%f), newres(%ld,%ld), lw %ld, sbw %ld, sbh %ld",
	     xres, yres, redx, redy, newresx, newresy, 
	     linewidth, smallBoxWidth, smallBoxHeight);
    ERR_PASS_CLEANUP( w_write( pIPWState, "jreduce.ras", workPr,
			       DEBUGPR_TYPE));
#endif /* DEBUG_TEXTSEGMENT */

    if (iterations <= 0)
    {
	/* User doesn't know the number of iterations we should do and
	 * instead wants us to compute it (and do it).
	 */
	computeIterations = 1;
	/* We'll break out of the loop below instead when the time is right".*/
	iterations = 0x7FFFFFFF /*MAXINT*/;
    }
    else 
	computeIterations = 0;
	
    for (i = 1; i <= iterations; i++)
    {
	ERR_PASS_CLEANUP( wp_words1bb(pIPWState,
				     workPr, newresx, newresy, 
				     &workRectList));

	/* Clear pixr */
	ERR_PASS_CLEANUP( w_pixrGetWidth( pIPWState, workPr, &w));
	ERR_PASS_CLEANUP( w_pixrGetHeight( pIPWState, workPr, &h));
	ERR_PASS_CLEANUP( w_rasterop( pIPWState, 0, 0, NULL, PIX_CLR,
				     0, 0, w, h, workPr));

	ERR_PASS_CLEANUP( w_pixrPutRectListBorder( pIPWState, 
						  workRectList,
						  linewidth, workPr));
	ERR_PASS_CLEANUP( w_destroyRectList( pIPWState, &workRectList));

	/* We want to do the following after above code when:
	 *   (a) computing the number of iterations (number of iterations
	 *       was not user-specified)
	 *   (b) we have completed doing the 
	 *       user-specified number of iterations
	 */
	if (computeIterations  ||  i == iterations)
	{
	    ERR_PASS_CLEANUP( w_rabb( pIPWState, workPr, 0, 0, &rabbl));
#if DEBUG_TEXTSEGMENT
	    for (num = 0, pRect = rabbl; 
		 pRect != NULL; 
		 pRect = pRect->next, num++);
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		     "\nfor %li rabbl count is %ld", i, num);
#endif /* DEBUG_TEXTSEGMENT */
	    ERR_PASS_CLEANUP( w_flatMergeRects(pIPWState, rabbl, &flatl));
	    
	    if (rl != NULL)
		/* need it to be around after last iteration */
		ERR_PASS_CLEANUP( w_destroyRectList( pIPWState, &rl)); 
	    ERR_PASS_CLEANUP( w_removeSmallRects(pIPWState, flatl, 
						smallBoxWidth, smallBoxHeight,
						&rl));

	    ERR_PASS_CLEANUP( w_destroyRectList( pIPWState, &rabbl));
	    ERR_PASS_CLEANUP( w_destroyRectList( pIPWState, &flatl));
	    
	    if (computeIterations)
	    {
		/* This stuff only needs to get done when we are
		 * computing the number of iterations.
		 */
		for (num = 0, pRect = rl; 
		     pRect != NULL; 
		     pRect = pRect->next, num++);

#if DEBUG_TEXTSEGMENT
		i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
			 "\nfor %li rl count is %ld\n", i, num);
		{
		    PIXR     *prt=NULL;
		    
		    ERR_PASS_CLEANUP( w_clonePixr( pIPWState, workPr, 
						      &prt));
		    ERR_PASS_CLEANUP( w_displayRectList( pIPWState,
							rl, 2, prt));
		    ERR_PASS_CLEANUP( w_write( pIPWState, "jfltrect.ras",
					      prt, DEBUGPR_TYPE)); 
		    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &prt));
		}
#endif /* DEBUG_TEXTSEGMENT */

		if (num <= 10)
		    /* If only 10 or less rectangles left, time to stop. */
		    break;

		/* If the change in the number of rectangles from last time
		 * to this time was less than 10%, time to stop.
		 */
		if (lastnum != -1  &&
		    (((float)(lastnum - num)/(float)lastnum) <= .10))
		    break;
		else
		    /* Save this current number of rectangles */
		    lastnum = num;
	    }
	}
    }

    /* Rescale back up to the scale of the inputted PIXR.
     * (Typically we reduced from 300-->100 DPI, so upscale by factor of 3)
     */
    ERR_PASS_CLEANUP( w_xformRectList( pIPWState, 0, 0, redx, redy, rl));

    ERR_PASS_CLEANUP( w_removeSmallRects( 
	pIPWState, rl,
	ROUND(((Float32)(MIN_TSEG_WIDTH  * xres))/300.0),
	ROUND(((Float32)(MIN_TSEG_HEIGHT * yres))/300.0),
	&workRectList));

    
    /* clean up */
    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &workPr));
    ERR_PASS_CLEANUP( w_destroyRectList( pIPWState, &rl));



#if DEBUG_LANDMARKS
    ERR_PASS_CLEANUP( w_clonePixr( pIPWState, sourcePr, &debugPr));
    ERR_PASS_CLEANUP( w_displayRectList( pIPWState, workRectList, 2, debugPr));
    
    ERR_PASS_CLEANUP( w_logicOp( pIPWState, sourcePr, XOR, debugPr));
    ERR_PASS_CLEANUP( w_write( pIPWState, "jscanlm.ras", debugPr, SUNRAS));
    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &debugPr));
#endif /* DEBUG_LANDMARKS */


    /* NOTICE we always sort the landmarks rect list before returning. */
    ERR_PASS_CLEANUP( w_rectSort( pIPWState, workRectList,
				  w_rectLessorUpperLeftPosition,
				  &rl));
    ERR_PASS_CLEANUP( w_destroyRectList( pIPWState, &workRectList));


#if DEBUG_PROGRESS
    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
	      "Rect for text segments are:\n");
    for (num = 0, pRect = rl; pRect != NULL; pRect = pRect->next, num++);
    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
	      " Number of rects: %ld\n", num);
    for (num = 1, pRect = rl; pRect != NULL; pRect = pRect->next, num++)
	i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		  "   Rect %ld:\n      (%ld,%ld,%ld,%ld)\n", 
		  num, pRect->x, pRect->y, pRect->w, pRect->h);
#endif /* DEBUG_PROGRESS */



    /* Return the number of rectangles in the returned list */
    for (num = 0, pRect = rl; pRect != NULL; pRect = pRect->next, num++);
    *nRects = num;

    /* Report the number of iterations that we did */
    if (computeIterations)
	*pIterations = i;
    else
	*pIterations = iterations;

    *ppRects = rl;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyRectList( pIPWState, &rl);
    w_destroyRectList( pIPWState, &workRectList);
    w_destroyRectList( pIPWState, &rabbl);
    w_destroyRectList( pIPWState, &flatl);
    w_destroyPixr( pIPWState, &workPr);
    w_destroyPixr( pIPWState, &prTemp);
#if DEBUG_LANDMARKS
    w_destroyPixr( pIPWState, &debugPr);
#endif

    ERR_RETURN;
}
