/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	floyd1.c:
 *
 *          Dithering code using Floyd-Steinberg algorithm
 *		From Nova.c by Rob Tow and tidied by Dan Bloomberg.
 *	    These are functions used by PerfectScan.
 *		
 *
 *          8 bit/pixel to 1 bit/pixel and 2 bit/pixel Floyd-Steinberg dither:
 *                  w_ditherFloyd()
 *          4 bit/pixel to 1 bit/pixel Floyd-Steinberg dither:
 *                  w_ditherFloyd4To1()
 *	    Misc:
 *		    w_cMapToLut () - makes a TRC table for above out
 *						of a cMap
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "rectops.pub"
#include "floyd.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: floyd1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/* The purpose of this module is to convert 8-bit gray scale images to
 * binary, dithered images.  This is not a simple binary threshold.  The
 * number of dots in a region of the binary image should be (roughly)
 * proportional to the gray value of the image.  The binary image should
 * be a faithful reproduction of the original.
 *
 * In addition, we will also apply dynamic range correction to the
 * gray pixels.  Each pixel's intensity will be mapped to some other
 * intensity using a table provided by the caller.  Doing the mapping
 * here saves the time required to write another complete gray image.
 *
 * There are many, many algorithms for converting gray images to binary.
 * The module ~bloomber/halftoning/Nova.c contains examples of 14
 * algorithms.  People invent more every day.  "Digital Halftoning"
 * by Robert Ulichney (MIT Press, 1987 call number T385.U45 1987 in the
 * PARC TIC) illustrates dozens more.
 *
 * We have chosen the Floyd-Steinberg method of dithering with error
 * diffusion.  It is simple to implement and gives very nice results.
 *
 * Like many diffusion algorithms, we start processing in the upper left
 * corner of an image.  There are two basic steps - threshold compare and
 * error diffusion.  The threshold compare is simple, we just use the most
 * significant bit of the pixel as the binary value.  Printing a "1" is
 * really only completely correct if the gray value is 255.  Printing a
 * 0 is only completely correct if the gray value is 0.  Using this
 * interpretation, we calculate the "error" as the difference between
 * the gray pixel and the "displayed" value.  For example, if the gray
 * pixel value p is less than 128, the pixel is actually p too large for
 * the value to be displayed.  Similarly, if p >= 128, the error is
 * p - 255 (a number less than or equal to 0).
 *
 * Now the magic comes in.  In Floyd-Steinberg, we apportion this error value
 * among 4 other gray pixels as shown below.  The "*" represents the
 * newly thresholded pixel.  The fractions represent the proportions of the
 * error value that is *added* to the pixel to the right and the three pixels
 * below.  The values of these fractions have been carefully chosen to make
 * the resulting images look nice.  I don't know why they work better than
 * some other set.  Nick Briggs says that the fractions are roughly
 * proportional to the distance from the center of the thresholded pixel to
 * to the other pixels.  Leigh Klotz says that FX says many combinations of
 * fractions work so long as there is no accumulated error.  Note that the
 * fractions do add to 1.
 *
 * |      |  *   | 7/16 |
 * | 3/16 | 5/16 | 1/16 |
 *
 * After the error is added to the original pixel values and previously
 * modified pixel values, we are supposed to truncate each addition back
 * to the range [0..255].  If a pixel value becomes greater than 255, we
 * set it to 255.  If it goes negative, we set it to 0.
 *
 * In the classic implementation, each line is processed from left to right
 * creating a new line of modified pixels below it.  This line of pixels
 * is processed in turn, leaving a new line and so on.  At least 4 write
 * operations are done for every pixel; more if truncation is required.
 *
 * How can we improve the speed of this beauty?  The first thing we notice
 * is the number of writes.  How can this be reduced?
 *
 * One approach is to calculate several lines of dithered image at once.
 * Assuming the error values can be held in processor registers until they
 * are binarized, we can avoid writing the error values to memory.  If we
 * calculate n lines of dithered image at once, we only have to write out
 * 1 line of partially-processed error values, so save writing n-1 lines of
 * pixels.
 *
 * A significant source of writes in some implementations is the binary
 * output.  For example, the code in Nova.c does a read, an AND, an OR, a
 * shift and a write for every bit of output image.  We'll just accumulate
 * 32 bits of output for each image line before writing.  We could mess
 * around with double writes, but it uses extra registers and has a very
 * small effect on overall performance (reducing the number of destination
 * writes by a factor of 32 makes the number of cycles spent on writes
 * insignificant).
 *
 * Another source of writes is the individual error diffusion calculations.
 * It would certainly be nice to avoid doing 4 separate READ-ADD-TRUNCATE-WRITE
 * sequences for each thresholded value.  Say we pack four pixels into
 * a register and allocate one register per line of image being calculated.
 * In addition, say we also add all three error values to the pixels in the
 * line below the thresholded pixel in one step.  This is done by fetching the
 * {3/16 | 5/16 | 1/16} triple from a table indexed by error value.  This
 * reduces the number of READ-ADD-TRUNCATE-WRITE operations from 4 to 2.
 * We explored schemes that packed all the pixels receiving error values
 * into one register so there would be only one READ-ADD-TRUNCATE-WRITE op.
 * The extra shifting and masking required wiped out the speed advantage.
 * The best of these schemes was very complex, and no faster.
 *
 * Truncation is also messy, especially if the pixels are packed.  We need
 * at least one bit between the pixels to catch the carries.  These buffer
 * bits can then be formed into masks used for truncation.  This is pretty
 * slow.  However, if we reserve 2 buffer bits between pixels, we can avoid
 * the truncation task altogether.  Say we shift each pixel right by 2 before
 * use.  We use the two upper bits for overflow as shown below:
 *
 * Value	Interpretation
 * 00xxxxxx	small positive number
 * 01xxxxxx	big positive number
 * 10xxxxxx	big negative number
 * 11xxxxxx	small negative number
 *
 * What would the range of pixel values be?  The biggest value would come from
 * a bunch of 127 pixels surrounding a 255.  Every 127 pixel would have an error
 * of 127.  The sum of all the weighted errors would be 127 * (7+1+3+5)/16
 * = 127.  The highest value a pixel could attain would be 255 + 127 = 382.
 * This can be represented in 9 bits plus a sign bit. Disregarding the bottom
 * 2 bits, it fits into 8 bits.  The most negative pixel value comes from
 * a 0 pixel surrounded by 128's.  It would similarly have a modified value of
 * 0 + (-128) = -128.  This fits in 8 bits, or 6 bits disregarding the lower
 * 2 bits. This means we can't generate a "big negative number," the only
 * attainable values are 00xxxxxx, 01xxxxxx and 11xxxxxx.
 *
 * Is is really ok to use the most significant 6 bits out of each pixel?  We'll
 * try it and see.  My reasoning is that the amount of information we
 * lose going from 256 gray levels to 2 gray levels is virtually
 * indistinguishable from that lost by going from 64 levels to 2 levels.
 * Even if we had the world's best scanner, humans would be very hard
 * pressed to distinguish 256 levels of dithered gray from 64 levels of
 * dithered gray.
 *
 * Where else can we improve things?  The binary values could be calculated by
 * comparing with 127>>2 and setting the appropriate bit in the output image
 * if the pixel was greater.  This would require a write per pixel. Yuck.
 * A faster way (at least in assembler, so we'll try it here and hope the
 * compiler is smart) is to subtract 127, invert the sign bit, shift it into
 * the carry (add reg to itself) and shift the carry into the accumulated
 * binary output.  We can skip inverting the sign bit by simply adding
 * 01100000 to the pixel value instead of subtracting 127>>2.  Since the pixel
 * values are restricted to 00xxxxxx, 01xxxxxx and 11xxxxxx; the sum will
 * have a 1 in the sign bit if and only if the original value was positive
 * and greater than 00011111 ( = 127 >> 2).
 *
 * Well, ooops, a bug already.  In gray scale land, big numbers mean BRIGHT
 * (like white).  In binary land, 1 means black.  The code being replaced
 * by this routine produces images that have to be inverted.  What we really
 * want to do is have the MS bit be a 0 if and only if the orignal value
 * was positive and greater than 00011111.  So add 11100000.  If the
 * pixel was 000xxxxx, we get 1 in the MS bit.  If the pixel was 001xxxxx,
 * 010xxxxx or 011xxxxx, we get a 0 in the MS bit.  If the pixel was
 * 110xxxxx or 111xxxxx (negative), we get a 1 in the MS bit.  Just what
 * we want.
 *
 * Since we're trying to explore how things might work in the assembly code,
 * we'll write C code that handles 3 images lines at a time - the top, middle
 * and bottom lines.  Assembly code will have more "middle" lines.  The table
 * below distinguishes them.  Gray Buffer is a private array of pixels one
 * line wide.
 *
 *	Line	Source		Binary Dest 	Gray Dest
 *	------  -----------	-----------	-----------
 *	Top	Gray Buffer	Dest Image	N.A.
 *	Middle	Gray Image	Dest Image	N.A.
 *	Bottom	Gray Image	N.A.		Gray Buffer
 *
 * The areas of image being processed are not quite vertically aligned.
 * Each line's locus of interest is down and 2 pixels to the right as
 * shown below.  Dan Bloomberg suggests we call this the "Knight Shift".
 *
 * |       |       |       |  T1  |  E11  |
 * |       |  T2   |  E21  |  E22 |  E23  |
 * |  E31  |  E32  |  E33  |      |       |
 *
 * The inner loop of the algorithm is:
 *
 *  1. Compute Binary value for T1.  Shift it into binary output reg for
 *	top line.  If that reg is full, write it to the destination image.
 *  2. Shift Top line left by 1 pixel, incoming pixel coming from Gray Buffer.
 *  3. Add 7/16 error value from T1 to E11.
 *  4. Add {3/16 | 5/16 | 1/16} of T1's error to {E21 | E22 | E23}.
 *  5. Compute Binary value for T2.  Shift into binary output reg for
 *	middle line.  If that reg is full, write it to the destination image.
 *  6. Shift middle line left by 1 pixel, incoming pixel from Source image.
 *  7. Add 7/16 error value from T2 to E21.
 *  8. Add {3/16 | 5/16 | 1/16} of T2's error to {E31 | E32 | E33}.
 *  9. Shift Bottom line left by 1 pixel.  Accumulate leftmost gray pixel
 *	in gray output reg.  If gray output reg is now full (only holds
 *	4 pixels) write it to gray buffer.
 *  10. Repeat 1-9 until top line has just processed the rightmost pixel
 *	in its line.
 *
 * Nice pipeline, but how do we start and end a line?
 *
 *  START OF LINE
 *  1. Top line reg is loaded with first four pixels from Gray Buffer.
 *	Leftmost pixel corresponds to leftmost pixel in the line.
 *  2. Middle line reg is loaded with {p[2,-2]:p[2,-1]:pl[2,0]:p[2,1]}
 *	where p[2,-2]:p[2,-1] are the two pixels immediately to the
 *	left of the second line and p[2,0]:p[2,1] are the leftmost two
 *	pixels in the second line.  Note that we explicitly use the
 *	area surrounding the halftone to generate edge values.  This seems
 *	more reasonable than blindly assuming white or black edges. 
 *  3. The bottom reg is loaded with {p[3,-4]:p[3,-3]:p[3,-2]:p[3,-1]}.
 *
 * The scheme for entering and exiting the pipeline should not add crud to
 * the inner loop.  Define DBLOCK_HEIGHT as the number of lines of binary
 * data produced on each pass across the image (i.e. doesn't include the
 * line of new error data).  We'll have two copies of the loop; one
 * for the left edge and one for the center and right edge of a line.
 * The left edge loop counts the number of horizontal steps taken by the
 * top line (= number of times we've gone through the loop).  The code for
 * the left loop is the same as the main inner loop shown above except no
 * binary or gray error data is written.  When the bottom line is filled
 * with valid pixel data (2*DBLOCK_HEIGHT times through the loop), we write
 * that data to GrayBuf (the buffer containing the error data) and fall into
 * the main line code.
 *
 * The next loop is executed long enough for the lowest line producing
 * dithered data (the middle line in our case) to have written out its
 * last word.  We assume DBLOCK_HEIGHT is less than 16, so the top line won't
 * try to write out an extra word on the right while we're waiting for the
 * last line.  How many cycles are required?  Say we start counting from the
 * left edge.  That's where the top line was aligned with the leftmost pixel.
 * The loop needs w cycles to for the top line to have computed all the valid
 * pixels.  It needs another (32 - (w MOD 32)) cycles to get to the next
 * word boundary.  Ooops, that's the right function unless w MOD 32 = 0.
 * Ok, it needs another ((32 - (w MOD 32)) & 0x1F) cycles to get to the
 * next word boundary.  Each line below the top line needs another 2 cycles
 * to get to the same point on its line, for a total of 2*DBLOCK_HEIGHT cycles.
 * Let's define
 *	cyclesOnRight = 2*DBLOCK_HEIGHT + ((32 - (w MOD 32)) & 0x1F)
 * We need to keep executing until loopCt (which started at 0
 * in the left loop and wasn't re-initialized) reaches (w + cyclesOnRight).
 *
 * We're not quite done yet.  The GrayBuf should be filled with enough data
 * to keep the top line happy.  How many bytes is that?  When we stop the loop,
 * the register accumulating GrayBuf's data holds at least 2 bytes off the
 * right side of the image.  The top line has just read an extra cyclesOnRight
 * bytes in the loop above, so we should keep accumulating, DRC'ing and storing
 * GrayBuf data until loopCt gets up to
 * (w + cyclesOnRight + (cyclesOnRight-2)).  After that, we keep going until
 * we hit a word (4 pixel) boundary to flush the last pixels.
 *
 * Much like the assembly version, the C code will have two routines:
 * ditherFloydPr and ditherFloydLine.  ditherFloydPr does all the high level
 * stuff (checing args for errors, allocating pixrects, etc) and calls
 * ditherFloydLine for every group of lines to be processed. In this C version
 * ditherFloydLine will put out 2 lines of dithered data at each pass.  The
 * assembly version will put out more lines per pass.   
 *
 * There's still a question about what to do at the bottom of the image.  If
 * we always put out several lines, we run the risk of writing into the bottom
 * frame of the dithered image.  There are (at least) four choices:
 * 1. write a version of ditherFloydLine that only does one line at a time and
 * call it appropriately at the bottom of the image; 2. Put a test and
 * branch into ditherFloydLine at each line boundary; 3. Allow ditherFloydLine
 * to mess up the bottom frame and clear it later.  By fiat, choice 1 is
 * too horrible to contemplate (especially in the assembly version).  I've
 * got enough trouble just getting one version to work right.  Besides, it's
 * slow.  Let's think about 2 vs 3.
 *
 * Say each frame-fixup word write takes 12 cycles and the test/branch pair
 * takes 2.  Since we produce one bit at a time, the test/branch pair will be
 * executed 32 times as often as the frame-fixup writes ( = 64 cycles/word).
 * Since 64 is bigger than 12, we'll do better to let ditherFloydLine smash
 * the bottom frame, then fix it up at the end.
 *
 * **********************  FIRST REWORK  ******************************
 *
 * Eventually (after debugging) that approach worked ok, but the image quality
 * wasn't quite as good as we would like.  The dynamic range was limited.
 * It was really only noticeable in the very light or very dark sections, but
 * it was noticeable there.
 *
 * We'd like to use more gray levels.  How can this be done without doing all
 * the error diffusion calculations in separate registers?  After some
 * contemplation, I had the brainstorm that we'd have lots of room if we only
 * kept *3* pixels in a register instead of 4.  Say we keep the image in
 * memory and in the GrayBuf packed at 4 pix/word, but spread things out
 * when pixels are processed.  We'll also have to rearrange the order in which
 * pixels are processed - adding the 7/16 error horizontally to a pixel before
 * adding the 1/16 error from the pixel above, but addition is communtative.
 * Here's the new inner loop with a new pixel diagram.  Note the inter-line
 * horizontal shift is 1 pixel, not 2.  At the start of the loop, the top
 * line's register holds E11:E12:E13.  The next line's reg holds
 * E21:E22:E23, etc.  The Ex3 pixels are held in bits 10:0.  Ex2's are held
 * in bits 21:11.  Ex1's are in bits 31:22.  That leaves at least 2 sign
 * bits between each pixel (plus one extra bit above the lower pixels).
 *
 * |       |       |  E11 |  E12  |  E13  |
 * |       |  E21  |  E22 |  E23  |       |
 * |  E31  |  E32  |  E33 |       |       |
 *
 *  1. Compute Binary value for E11.  Shift it into binary output reg for
 *	top line.  If that reg is full, write it to the destination image.
 *  2. Shift Top line left by 1 pixel, incoming pixel coming from Gray Buffer.
 *  3. Add 7/16 of E11's error value to E12.
 *  4. Add {3/16 | 5/16 | 1/16} of E11's error to {E21 | E22 | E23}.
 *  5. Compute Binary value for E21.  Shift into binary output reg for
 *	middle line.  If that reg is full, write it to the destination image.
 *  6. Shift middle line left by 1 pixel, incoming pixel from Source image.
 *  7. Add 7/16 of E21's error value to E22.
 *  8. Add {3/16 | 5/16 | 1/16} of E21's error to {E31 | E32 | E33}.
 *  9. Truncate leftmost gray pixel into the range [0..255], then accumulate
 *	it into the in gray output reg.  If gray output reg is now full
 *	(only holds 4 pixels) write it to gray buffer.
 *  10. Repeat 1-9 until top line has just processed the rightmost pixel
 *	in its line.
 *
 * How do we get started?  Picture the left edge of the rectangle to be
 * dithered as the line to the left of E11 above.  E11:E12:E13 get the
 * first three pixels from GrayBuf.  E21 is the pixel just to the left
 * of the rectangle, E22:E23 are the leftmost two gray pixels.  E31:E32 are
 * uninitialized (since they'd be discarded) and E33 is the leftmost pixel
 * in the third line.
 *
 * The GrayBuf is filled with processed pixels.  It is written every 4th pixel
 * starting from the very first pixel processed.  But, we see that the first
 * DBLOCK_HEIGHT pixels in the bottom line are trash, right?  No problem, we
 * just start reading the GrayBuf DBLOCK_HEIGHT pixels from the beginning.
 * This skips the trash and keeps the inner loop simple.
 *
 * How do we finish?  As before, we need to process enough lines so that
 * the lowest line producing dithered bits (the middle line in this case)
 * has processed its last pixel, then gets to a word boundary.  w cycles gets
 * the top line across the rectangle.  Set cyclesOnRight to
 * 	cyclesOnRight = DBLOCK_HEIGHT + ((32 - (w MOD 32)) & 0x1F)
 * So we do w + cyclesOnRight processing cycles.  At this point, the bottom
 * line has read 2 pixels to the right of the image.  So, as in the
 * first scheme, we read cyclesOnRight-2+4 more bytes and put them into
 * GrayBuf (the +4 term gets them into the output register.
 *
 * One nice feature of this scheme is that we no longer need to shift each
 * incoming pixel right before use.  A nasty feature is that we need to
 * truncate each modified pixel before storing it into GrayBuf.  As usual,
 * we'll do that with a table.
 *
 */



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dithers 8 bpp gray image to 1 bpp using Floyd-Steinberg error diffusion.
 *
 * LongDescription:::
 *  A simple way to convert a gray image to a binary image is to simply
 *  truncate the low order bits.  This is fast, but loses most of the
 *  gray levels.  A more complicated method that retains gray levels (but
 *  tends to fuzz sharp edges) is to use some sort of dithering, like
 *  halftoning or error diffusion.  There are books written on this stuff.
 *  This routine converts an 8 bit/pixel gray image into a binary image 
 *  using standard Floyd-Steinberg error diffusion.
 *<BR><BR>
 *  Since dithering should be done with the characteristics of the output
 *  device in mind, this routine requires the user to supply a Tone
 *  Reproduction Curve (TRC) table.  All gray pixels are mapped through
 *  this table before being processed further.
 *<BR><BR>
 *  Error diffusion is fairly expensive.  The portion of any page image
 *  occupied by a picture is often small.  To avoid unnecessary processing,
 *  this routine accepts a rectangle describing the portion of a gray image
 *  that should actually be diffused.
 *
 * PrivateDescription:::
 *  There is complex assembly code for the primitives of this operation.
 *  There is also a standard C implementation.  The outputs of the
 *  assembly code on the SPARC, assembly code on the x86 and standard
 *  C code are all slightly different.  The differences are only noticeable
 *  if you are counting bits.  They arise from differences in the number
 *  of lines processed before errors are truncated to 8 bits.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to the gray image to be modified.
 *  pRect::          Pointer to a RECTANGL that describes the portion
 *		     of the gray image to be dithered.  If this is NULL,
 *		     the entire image is dithered.
 *  pDRCTable::      Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map image pixels before dithering.
 *		     It is normally used to account for the fact that
 *		     printers may print too light or too dark.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  Clients should avoid using this routine to dither strips of an image.
 *  Restarting the error calculations at strip boundaries causes noticeable
 *  artifacts.  The routines w_ditherFloydInit, w_ditherFloydProcess
 *  and w_ditherFloydFinish can be used to seamlessly dither image
 *  strips.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, pixel mapping, gray
 *
 * Level:::
 *  deterministic, modulo roundoff differences
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  See "Digital Halftoning" by Robert Ulichney.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    UInt8              *pDRCTable,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 x, y, w, h, wS, hS, bplS, bplD;
    UInt8		*pS, *pD;
    ScaledInt1616	 xSdpi, ySdpi;
    Bool		 allocatedHere = cFalse;
    Int8		 valid;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pDRCTable == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* On entry, the rect (x, y, w, h) specifies a rectangle with respect
	 * to the upper left corner of the "proper" image (inside the frame).
	 * The output binary image is always justified to
	 * the upper left corner of its framed pixrect (i.e. inside the
	 * frame).   If the rect is NULL, the entire image is used.  */
    if (!pRect)
    {
	x = 0;
	y = 0;
	w = wS;
	h = hS;
    }
    else
    {
	ERR_PASS( w_isRectValid(pIPWState, prS, pRect, &valid));
	ERR_GEN( !valid, cErrInvalidParm);
	x = pRect->x;
	y = pRect->y;
	w = pRect->w;
	h = pRect->h;
    }

	/* check the width & height of the frameless gray image.  The gray
	 * frame is in units of PIXELS, not bits.
	 */
    ERR_GEN((x < 0) || (y < 0) || (w <= 0) || (h <= 0), cErrOutOfBounds);
    ERR_GEN(((x+w) > wS) || ((y+h) > hS), cErrOutOfBounds);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, w, h, 1, xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* Now that we have legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    
    ERR_PASS_CLEANUP( i_ditherFloyd (pIPWState, pS, bplS,
				     x+cFrameBits, y+cFrameBits,
				     w, h, pDRCTable,
				     pD, bplD));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames (pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dithers 4 bpp gray image to 1 bpp using Floyd-Steinberg error diffusion.
 *
 * LongDescription:::
 *  A simple way to convert a gray image to a binary image is to simply
 *  truncate the low order bits.  This is fast, but loses most of the
 *  gray levels.  A more complicated method that retains gray levels (but
 *  tends to fuzz sharp edges) is to use some sort of dithering, like
 *  halftoning or error diffusion.  There are books written on this stuff.
 *  This routine converts an 4 bit/pixel gray image into a binary image 
 *  using standard Floyd-Steinberg error diffusion.
 *<BR><BR>
 *  Since dithering should be done with the characteristics of the output
 *  device in mind, this routine requires the user to supply a Tone
 *  Reproduction Curve (TRC) table.  All gray pixels are mapped through
 *  this table before being processed further.  Since there aren't very
 *  many levels to start with, the contents of this table have a very
 *  strong effect on the binary output.
 *<BR><BR>
 *  Error diffusion is fairly expensive.  The portion of any page image
 *  occupied by a picture is often small.  To avoid unnecessary processing,
 *  this routine accepts a rectangle describing the portion of a gray image
 *  that should actually be diffused.
 *
 * PrivateDescription:::
 *  There is complex assembly code for the primitives of this operation.
 *  There is also a standard C implementation.  The outputs of the
 *  assembly code on the SPARC, assembly code on the x86 and standard
 *  C code are all slightly different.  The differences are only noticeable
 *  if you are counting bits.  They arise from differences in the number
 *  of lines processed before errors are truncated to 8 bits.
 *<BR><BR>
 *  In all cases, the 4 bit pixels are expanded up to 8 bits before being
 *  processed.  This allows us to retain as many diffused error bits
 *  as possible.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to the gray image to be modified.
 *  pRect::          Pointer to a RECTANGL that describes the portion
 *		     of the gray image to be dithered.  If this is NULL,
 *		     the entire image is dithered.
 *  pDRCTable::      Pointer to a table containing 16 UInt8's.  This
 *		     table is used to map image pixels before dithering.
 *		     It is normally used to account for the fact that
 *		     printers may print too light or too dark.
 *		     The 16 numbers should range between 0 and 255, not
 *		     0 and 15.  Proper selection of these numbers has a
 *		     critical effect on the binary output.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  Clients should avoid using this routine to dither strips of an image.
 *  Restarting the error calculations at strip boundaries causes noticeable
 *  artifacts.  The routines w_ditherFloydInit, w_ditherFloydProcess
 *  and w_ditherFloydFinish can be used to seamlessly dither image
 *  strips.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, pixel mapping, gray
 *
 * Level:::
 *  deterministic, modulo roundoff differences
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  See "Digital Halftoning" by Robert Ulichney.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd4To1 (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    UInt8              *pDRCTable,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 x, y, w, h, wS, hS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int8		 valid;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pDRCTable == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* On entry, the rect (x, y, w, h) specifies a rectangle with respect
	 * to the upper left corner of the "proper" image (inside the frame).
	 * The output binary image is always justified to
	 * the upper left corner of its framed pixrect (i.e. inside the
	 * frame).   If the rect is NULL, the entire image is used.  */
    if (!pRect) {
	x = 0;
	y = 0;
	w = wS;
	h = hS;
    }
    else {
	ERR_PASS( w_isRectValid(pIPWState, prS, pRect, &valid));
	ERR_GEN( !valid, cErrInvalidParm);
	x = pRect->x;
	y = pRect->y;
	w = pRect->w;
	h = pRect->h;
    }

	/* check the width & height of the frameless gray image.  The gray
	 * frame is in units of PIXELS, not bits.
	 */
    ERR_GEN((x < 0) || (y < 0) || (w <= 0) || (h <= 0), cErrOutOfBounds);
    ERR_GEN(((x+w) > wS) || ((y+h) > hS), cErrOutOfBounds);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, w, h, 1, xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* Now that we have legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    
    ERR_PASS_CLEANUP( i_ditherFloyd4to1 (pIPWState, pS, bplS,
					 x+cFrameBits, y+cFrameBits,
					 w, h, pDRCTable,
					 pD, bplD));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/* This function is here because PerfectScan will always want some sort of
 * color dither operation, though the implementation of that operation
 * may change.  The versions of color ops will be in individual files.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Produce a table used to map 16M or 4096 colors to 256 colors.
 *
 * LongDescription:::
 *  Some color images have 24 or 12 bits/pixel.  These have 8 or 4
 *  bits/pixel in each of three planes (Red, Green and Blue).  Other
 *  color images have only 8 bits/pixel where each 8-bit entry is mapped
 *  to a 24-bit color using a color map.  It's pretty easy to convert
 *  from color mapped image to a full color image, just replace each
 *  8-bit pixel with its 24-bit equivalent.  It's harder to map
 *  24-bit color images to indexed color images.  A glorious way is to
 *  dither the 24-bit image.  See w_ditherFloyd24To8.  A simpler way is
 *  to map each color pixel to the closest indexed color.  This routine
 *  is used to create the table used to map the pixels.
 *<BR><BR>
 *  The table created should be indexed with <MS 4 bits of Red><MS 4 bits of
 *  Green><MS 4 bits of Blue>  For that index, it contains the closest
 *  8 bit indexed color.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pCMap::          Pointer to indexed color map.
 *  isNew::          cFalse: i_cMapToLut, cTrue: i_cMapToLut1
 *
 * Output:::
 *  ppLUT::          Pointer to pointer to lookup table that will be
 *		     created.  The table is ALWAYS allocated here.
 *		     *ppLUT must be NULL.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine checks to see of *ppLUT
 *  is NULL.  If not, cErrPtrNotNull is returned.
 *
 * Notes:::
 *  The current version of this routine uses RGB color space to compute the
 *  "nearest" color to a color map color.  It might be better to use some
 *  other space like L*a*b*.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, color, format conversion, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cMapToLut (
    IPWState           *pIPWState,
    ColorMap           *pCMap,
    Bool                isNew,
    UInt8             **ppLUT)
{
    PROLOGUE;

    ERR_GEN(pCMap == NULL || ppLUT == NULL, cErrNullPtr);
    ERR_GEN(*ppLUT != NULL, cErrPtrNotNull);

    /* Yuck--a hard-coded constant. Well, we should really
       take a bit count, and figure out what to do. */
    
    *ppLUT = (UInt8 *) IPW_MALLOC (4096, cAllocatorData);
    ERR_GEN(*ppLUT == NULL, cErrNomem);

    /* Do the conversion */
    if( isNew == cFalse )   /* Mapping values around 0, 0x10, 0x20, .. 0xf0 */
	ERR_PASS_CLEANUP( i_cMapToLut (pIPWState,
				       (UInt8 *)&pCMap->map[0], pCMap->count,
				       *ppLUT));
    else /* Mapping values around 0, 0x11, 0x22, .. 0xff */
	ERR_PASS_CLEANUP( i_cMapToLut1 (pIPWState,
					(UInt8 *)&pCMap->map[0], pCMap->count,
					*ppLUT));

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(ppLUT, cAllocatorData);
    ERR_RETURN;
}

/*

 $Log: floyd1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.27  1998/01/14 00:11:45  ddavies
 floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.26  1997/11/26 19:33:15  xzhu
 xzhu on Wed Nov 26 11:32:17 PST 1997

 floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v

 Added new input argument isNew for w_cMapToLut to select using i_cMapToLut or i_cMapToLut1.

 Revision 1.25  1997/04/15 21:48:22  ddavies
 floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.24  1996/01/04 20:22:35  ejaquith
 floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.23  1995/12/13  04:07:09  ejaquith
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/13  20:11:09  ejaquith
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/13  22:00:15  ejaquith
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/10/05  04:28:01  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.19  1995/10/03  01:21:01  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.18  1995/09/21  22:52:58  ejaquith
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.17  1995/09/15  18:01:28  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.16  1995/09/11  22:12:02  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.15  1995/09/10  00:34:03  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.14  1995/09/07  01:31:03  ejaquith
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:19:00  ejaquith
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  01:15:35  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.11  1995/08/31  21:12:42  ejaquith
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/17  22:03:17  ejaquith
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.9  1995/07/29  04:37:13  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Checkin preliminary changes to new style - mostly making procs agree with protos.
 *
 * Revision 1.8  1994/10/07  19:18:42  lperry
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.7  1994/06/08  04:10:18  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.6  1994/04/12  01:13:11  macleod
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.5  1994/04/05  00:40:06  macleod
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Changed cMapToLut to take a ColorMap structure instead of a UInt8.
 *
 * Revision 1.4  1994/03/02  03:55:49  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Add IP_YIELD to give up control after creating dest pixr in dithering ops.
 *
 * Revision 1.3  1993/12/03  01:14:34  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:11:43  ddavies
 * floyd1.c -> /product/ipcore/ipwrap/src/RCS/floyd1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

