/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	floyd2.c:
 *
 *          Dithering code using Floyd-Steinberg algorithm
 *	    This is color dithering code that takes and consumes entire
 *	    images.
 *
 *	    24 bit/pixel to 8 bit/pixel Floyd-Steinberg dither:
 *		    w_ditherFloyd24To8()
 *		    w_ditherFloyd24ToCMap () - takes a color LUT
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "rectops.pub"
#include "floyd.pub"
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: floyd2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/* There are extensive comments about general dithering algorithms in
 * floyd1.c.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Error diffuses RGB image to indexed color image using linear color space.
 *
 * LongDescription:::
 *  Some color images have 24 bits/pixel.  These have 8
 *  bits/pixel in each of three planes (Red, Green and Blue).  Other
 *  color images have only 8 bits/pixel where each 8-bit entry is mapped
 *  to a 24-bit color using a color map.  It's pretty easy to convert
 *  from color mapped image to a full color image, just replace each
 *  8-bit pixel with its 24-bit equivalent.  It's harder to map
 *  24-bit color images to indexed color images.  A glorious way is to
 *  dither the 24-bit image.  This routine does just that.
 *<BR><BR>
 *  This routine also creates the color map.  It is given the number of
 *  red colors, green colors and blue colors.  It creates a color map
 *  containing redColors*greenColors*blueColors colors.  This product
 *  must be less than or equal to 256.  Since the color map can be easily
 *  recreated, it is not returned.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 24-bit RGB image to be dithered.
 *  pRect::          Pointer to a RECTANGL that describes the portion
 *		     of the gray image to be dithered.  If this is NULL,
 *		     the entire image is dithered.
 *  pRedDRCTable::   Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map red image pixels before dithering.
 *		     It is normally used to account for the fact that
 *		     displays or printers may print too light or too dark.
 *  pGreenDRCTable:: Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map green image pixels before dithering.
 *  pBlueDRCTable::  Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map blue image pixels before dithering.
 *  redColors::      Number of shades of red in the color map.
 *  greenColors::    Number of shades of green in the new color map.
 *  blueColors::     Number of shades of blue in the new color map.
 *  mapOffset::      The 8-bit numbers put out by this routine will start
 *		     at mapOffset and work up.  This allows the outputs
 *		     to fit into some other color map.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to an 8-bit, single channel
 *		     destination image.  As usual, if a destination image 
 *		     is supplied, its dimensions, depth, etc are checked.  
 *		     If no dest image is supplied, one is created.
 *
 * Returns:::
 *  In addition to the usual stuff, this routine checks to make sure that
 *  redColors*greenColors*blueColors + mapOffset <= 256.  That is, the
 *  number of colors in the color map plus the offset of the first entry
 *  should still fit in 8 bits.  This routine also returns cErrPtrNotNull
 *  if prS already has a color map.
 *
 * Notes:::
 *  The thing that makes this type of diffusion relatively straightforward
 *  is that each of the colors is linear, so you can just dither each
 *  plane independently and add the results together to get the final
 *  index.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  There are functions in Sandpiper that do about the same thing.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd24To8 (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    UInt8              *pRedDRCTable,
    UInt8              *pGreenDRCTable,
    UInt8              *pBlueDRCTable,
    UInt8               redColors,
    UInt8               greenColors,
    UInt8               blueColors,
    UInt8               mapOffset,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD;
    Int32		 x, y, w, h, wS, hS, bplD, bplS;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pR, *pG, *pB, *pD;
    PIXR		*subG, *subR, *subB;
    ColorMap		*cmap;
    Bool		 allocatedHere = cFalse;
    Int8		 valid;
    Int32		 totalColors;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

        /* checkout the args */
    ERR_GEN(prS == NULL || pRedDRCTable == NULL ||
	    pGreenDRCTable == NULL || pBlueDRCTable == NULL ||
	    pPrD == NULL, cErrNullPtr);

    totalColors = redColors * greenColors * blueColors;
    ERR_GEN( (totalColors == 0) || ((UInt32)totalColors + mapOffset > 256),
	     cErrInvalidParm);

	/* Note that the depth is the depth of ONE channel! */
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);

	/* make sure the pixr doesn't already have a color map
	 * since we're about to ignore it anyway.
	 */
    ERR_PASS( w_pixrGetColorMap (pIPWState, prS, &cmap));
    ERR_GEN(cmap != NULL, cErrPtrNotNull);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

        /* On entry, rect (x, y, w, h) specifies a rectangle with respect to
         * the upper left corner of the *image*, not the upper left
         * corner of the frame!  The output binary image is always justified to
         * the upper left corner of its framed pixr (i.e. inside the
	 * frame).   If the rect is NULL, the entire image is used.  */

    if (!pRect) {
	x = 0;
	y = 0;
	w = wS;
	h = hS;
    }
    else {
	ERR_PASS( w_isRectValid(pIPWState, prS, pRect, &valid));
	ERR_GEN( !valid, cErrInvalidParm);
	x = pRect->x;
	y = pRect->y;
	w = pRect->w;
	h = pRect->h;
    }

        /* check the width & height of the frameless gray image.  The gray
         * frame is in units of PIXELS, not bits.
	 */
    ERR_GEN((x < 0) || (y < 0) || (w <= 0) || (h <= 0), cErrOutOfBounds);
    ERR_GEN(((x+w) > wS) || ((y+h) > hS), cErrOutOfBounds);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, w, h, 8, xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* Now that we have legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    
    
    ERR_PASS_CLEANUP( i_ditherFloydRGBTo8 (pIPWState, pR, pG, pB, bplS,
					   x + cFrameBits, y + cFrameBits,
					   w, h,
					   pRedDRCTable,
					   pGreenDRCTable,
					   pBlueDRCTable,
					   redColors, greenColors, blueColors,
					   mapOffset,
					   pD, bplD));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Error diffuses RGB image to indexed color image using color map.
 *
 * LongDescription:::
 *  Some color images have 24 bits/pixel.  These have 8
 *  bits/pixel in each of three planes (Red, Green and Blue).  Other
 *  color images have only 8 bits/pixel where each 8-bit entry is mapped
 *  to a 24-bit color using a color map.  It's pretty easy to convert
 *  from color mapped image to a full color image, just replace each
 *  8-bit pixel with its 24-bit equivalent.  It's harder to map
 *  24-bit color images to indexed color images.  A glorious way is to
 *  dither the 24-bit image.  This routine does this.
 *<BR><BR>
 *  The difference between this routine and w_ditherFloyd24To8 is that
 *  the color indices produced by this routine correspond to a color
 *  map passed in by the client.  There are no guarantees about the
 *  contents of this color map, i.e. it is not necessarily linear.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 24-bit RGB image to be dithered.
 *  pRect::          Pointer to a RECTANGL that describes the portion
 *		     of the gray image to be dithered.  If this is NULL,
 *		     the entire image is dithered.
 *  pRedDRCTable::   Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map red image pixels before dithering.
 *		     It is normally used to account for the fact that
 *		     displays or printers may print too light or too dark.
 *  pGreenDRCTable:: Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map green image pixels before dithering.
 *  pBlueDRCTable::  Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map blue image pixels before dithering.
 *  cMap::           Pointer to existing ColorMap as defined in pixr.h.
 *		     This structure contains the number of colors in the
 *		     map (should be 256), the type of the colors (cColorRGB),
 *		     and a pointer to an array of 256 ColorRecords.  Each
 *		     ColorRecord is a 4 byte array containing R,G,B and X.
 *  colorLut::       Pointer to a color lookup table.  This table should
 *		     be created by w_cMapToLut.  It maps each 24-bit color
 *		     to the nearest color in cMap's color map.
 *  mapOffset::      This number will be added to all 8-bit indices produced.
 *		     Doing so will allow these outputs to be used with
 *		     some other color map.
 *  outputBits::     The number of bits/index in the output.  This must be
 *		     either 4 or 8 (16 colors or 256 colors).
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to an 8-bit, single channel
 *		     destination image.  As usual, if a destination image 
 *		     is supplied, its dimensions, depth, etc are checked.  
 *		     If no dest image is supplied, one is created.
 *
 * Returns:::
 *  In addition to the usual stuff, this routine checks to see if mapOffset +
 *  cMap->count is greater than or equal to (1 << outputBits)).  If so,
 *  we raise a cErrInvalidParm error because the largest output value would
 *  be too big to fit in outputBits bits.  This routine also returns
 *  cErrPtrNotNull if prS already has a color map.
 *
 * Notes:::
 *  The thing that makes this harder than w_ditherFloyd24To8 is that the
 *  colors are not linear.  We have to find the closest client color to the
 *  color in the color map by looking at all three channels at once, then
 *  distribute all three errors.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  Sandpiper contains routines like this one.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd24ToCMap (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    UInt8              *pRedDRCTable,
    UInt8              *pGreenDRCTable,
    UInt8              *pBlueDRCTable,
    ColorMap           *cMap,
    UInt8              *colorLut,
    UInt8               mapOffset,
    UInt8               outputBits,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD;
    Int32		 x, y, w, h, wS, hS, bplD, bplS;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pR, *pG, *pB, *pD;
    PIXR		*subG, *subR, *subB;
    ColorMap		*cmap;
    Bool		 allocatedHere = cFalse;
    Int8		 valid;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

        /* checkout the args */
    ERR_GEN(prS == NULL || pRedDRCTable == NULL ||
	    pGreenDRCTable == NULL || pBlueDRCTable == NULL ||
	    cMap == NULL || colorLut == NULL || pPrD == NULL, cErrNullPtr);
    ERR_GEN(outputBits != 4 && outputBits != 8, cErrInvalidParm);
    ERR_GEN((mapOffset + (UInt16)(cMap->count) >
		(1UL << outputBits)), cErrInvalidParm);

	/* Note that the depth is the depth of ONE channel! */
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);

	/* make sure the pixr doesn't already have a color map
	 * since we're about to ignore it anyway.
	 */
    ERR_PASS( w_pixrGetColorMap (pIPWState, prS, &cmap));
    ERR_GEN(cmap != NULL, cErrPtrNotNull);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

        /* On entry, rect (x, y, w, h) specifies a rectangle with respect to
         * the upper left corner of the *image*, not the upper left
         * corner of the frame!  The output binary image is always justified to
         * the upper left corner of its framed pixr (i.e. inside the
	 * frame).   If the rect is NULL, the entire image is used.  */

    if (!pRect) {
	x = 0;
	y = 0;
	w = wS;
	h = hS;
    }
    else {
	ERR_PASS( w_isRectValid(pIPWState, prS, pRect, &valid));
	ERR_GEN( !valid, cErrInvalidParm);
	x = pRect->x;
	y = pRect->y;
	w = pRect->w;
	h = pRect->h;
    }

        /* check the width & height of the frameless gray image.  The gray
         * frame is in units of PIXELS, not bits.
	 */
    ERR_GEN((x < 0) || (y < 0) || (w <= 0) || (h <= 0), cErrOutOfBounds);
    ERR_GEN(((x+w) > wS) || ((y+h) > hS), cErrOutOfBounds);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, w, h, outputBits,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* Now that we have legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));

    ERR_PASS_CLEANUP( i_ditherFloydRGBToCMap (pIPWState, pR, pG, pB, bplS,
					      x + cFrameBits, y + cFrameBits,
					      w, h,
					      (UInt8 *) &cMap->map[0],
					      cMap->count,
					      colorLut,
					      pRedDRCTable,
					      pGreenDRCTable,
					      pBlueDRCTable,
					      mapOffset,
					      pD, bplD,
					      outputBits));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/*

 $Log: floyd2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.27  1998/01/14 00:11:47  ddavies
 floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.26  1997/04/15 21:48:36  ddavies
 floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.25  1996/01/04 20:22:38  ejaquith
 floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.24  1995/12/13  04:07:13  ejaquith
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.23  1995/12/01  19:32:11  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Fix acc compiler warnings.
 *
 * Revision 1.22  1995/11/13  20:11:12  ejaquith
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/13  22:00:19  ejaquith
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/10/05  04:28:05  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.19  1995/10/03  18:17:09  ddavies
 * ddavies on Tue Oct 3 11:16:38 PDT 1995
 *
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Check to see if (mapOffset + cMap->count) > number of levels instead of
 * >= number of levels.
 *
 * Revision 1.18  1995/10/03  01:21:05  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.17  1995/09/21  22:53:01  ejaquith
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/15  18:01:33  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.15  1995/09/11  22:12:06  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.14  1995/09/10  00:34:06  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.13  1995/09/07  01:31:05  ejaquith
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:19:03  ejaquith
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  01:15:40  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:12:46  ejaquith
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/17  22:03:20  ejaquith
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.8  1995/07/29  04:37:33  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.7  1994/10/07  19:18:47  lperry
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.6  1994/06/08  04:10:26  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.5  1994/04/12  01:13:23  macleod
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.4  1994/04/05  00:40:28  macleod
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Changed color dithering code to take Color Pixrs rather than three
 * separate planes
 *
 * Revision 1.3  1993/12/03  01:14:39  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:11:51  ddavies
 * floyd2.c -> /product/ipcore/ipwrap/src/RCS/floyd2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

