/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	floyd3.c:
 *
 *          Dithering code using Floyd-Steinberg algorithm
 *		From Nova.c by Rob Tow and tidied by Dan Bloomberg.
 *	    This dithering function is not used by PerfectScan.
 *
 *          8 bit/pixel to 2 bit/pixel Floyd-Steinberg dither:
 *                  w_ditherFloyd2Pr()
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "rectops.pub"
#include "floyd.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: floyd3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/* Extensive comments on the general dithering algorithm are in floyd1.c.
 */

/* The monochrome dithering routines worked just fine.  However, our first
 * real printer takes 2 bit/pixel images, not monochrome.  We need a dither
 * routine that produces 2 bpp images from 8 bpp images.
 *
 * In the long run, it might be nicer to have one dither routine that was
 * passed an argument telling whether to produce a 1 or 2 bpp output.  That
 * would require changes to the current interface.  This is a very touchy
 * subject.  For now, we'll introduce a new routine name.  At some later point,
 * these may be consolidated.
 *
 * The changes aren't too bad.  Generating a bigger output image is trivial.
 * There are two more interesting changes.  First, each shift step producing
 * output bits gets two bits instead of one.  Output registers are written
 * to memory twice as often.
 *
 * The other interesting change is in the diffusion tables.  The three values
 * written are:
 *
 * Value	Intensity
 *  00		completely black
 *  01		1/3 white
 *  10		2/3 white
 *  11		completely white
 *
 * This changes the error values.  We still divide the input range into four
 * equally slices by the two upper bits.  However the error values are
 *
 * 	x in Range		Basic Error value	Digital Output
 * [-infinity, 00111111]	Max(0, x)		  00
 * [01000000,  01111111]	x - (1*255/3)		  01
 * [10000000,  10111111]	x - (2*255/3)		  10
 * [11000000, +infinity]	Min(x - 255, 0)		  11
 *
 * The first error value is always >= 0, the last always <= 0.  The
 * middle two might be positive or negative.  These basic values are
 * distributed using the familiar 7/16, 3/16, 5/16 and 1/16 fractions.
 *
 * How do we generate the correct 2 bit data?  As seen from the table above,
 * we can't just grab a pair of bits from the thresholded pixel because it
 * hasn't been truncated yet.  It would be unpleasant to add an explicit
 * truncation operation.
 *
 * Errors in the 7/16 table are normally added only to the middle pixel of
 * the register register containing the thresholded pixel.  For example,
 * we would add 7/16 of E11's basic error value from the table above to
 * E21.  Say we do this before shifting the register containing E11 and E21.
 * This means E11 is still in the most significant bits of the register.
 * In this code, we'll add some number to the upper pixel so that the
 * top two bits [31:30] of the result end up being the desired output value
 * for that pixel.  For example, if the thresholded pixel is negative
 * (bits [31:30] = 11), we add a number having 01 in bits [31:30].  After
 * the 7/16 value is added in, we use bits [31:30] as the truncated,
 * digitized output value of the pixel.
 *
 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dithers 8 bpp gray image to 2 bpp using Floyd-Steinberg error diffusion.
 *
 * LongDescription:::
 *  A simple way to convert a gray image to a 2 bit/pixel image is to simply
 *  truncate the low order bits.  This is fast, but loses most of the
 *  gray levels.  A more complicated method that retains gray levels (but
 *  tends to fuzz sharp edges) is to use some sort of dithering, like
 *  halftoning or error diffusion.  There are books written on this stuff.
 *  This routine converts an 8 bit/pixel gray image into a 2 bit/pixel image 
 *  using standard Floyd-Steinberg error diffusion.  Instead of keeping the
 *  top bit of each diffused pixel, we keep the top 2 bits.  The errors
 *  are correspondingly smaller since we can get closer to the real gray
 *  values.
 *<BR><BR>
 *  Since dithering should be done with the characteristics of the output
 *  device in mind, this routine requires the user to supply a Tone
 *  Reproduction Curve (TRC) table.  All gray pixels are mapped through
 *  this table before being processed further.
 *<BR><BR>
 *  Error diffusion is fairly expensive.  The portion of any page image
 *  occupied by a picture is often small.  To avoid unnecessary processing,
 *  this routine accepts a rectangle describing the portion of a gray image
 *  that should actually be diffused.
 *
 * PrivateDescription:::
 *  There is complex assembly code for the primitives of this operation.
 *  There is also a standard C implementation.  The outputs of the
 *  assembly code on the SPARC, assembly code on the x86 and standard
 *  C code are all slightly different.  The differences are only noticeable
 *  if you are counting bits.  They arise from differences in the number
 *  of lines processed before errors are truncated to 8 bits.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to the gray image to be modified.
 *  pRect::          Pointer to a RECTANGL that describes the portion
 *		     of the gray image to be dithered.  If this is NULL,
 *		     the entire image is dithered.
 *  pDRCTable::      Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map image pixels before dithering.
 *		     It is normally used to account for the fact that
 *		     printers may print too light or too dark.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a 2 bpp dest image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  Two bit/pixel images aren't very common.  They are sometimes used for
 *  printers that would like to try a little gray.
 *<BR><BR>
 *  There is no dither-by-strips version of this function.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, pixel mapping, gray
 *
 * Level:::
 *  deterministic, modulo roundoff differences
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  See "Digital Halftoning" by Robert Ulichney.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd2 (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    UInt8              *pDRCTable,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 x, y, w, h, wS, hS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int8		 valid;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pDRCTable == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* On entry, the rect (x, y, w, h) specifies a rectangle with respect
	 * to the upper left corner of the "proper" image (inside the frame).
	 * The output binary image is always justified to
	 * the upper left corner of its framed pixrect (i.e. inside the
	 * frame).   If the rect is NULL, the entire image is used.  */
    if (!pRect) {
	x = 0;
	y = 0;
	w = wS;
	h = hS;
    }
    else {
	ERR_PASS( w_isRectValid(pIPWState, prS, pRect, &valid));
	ERR_GEN( !valid, cErrInvalidParm);
	x = pRect->x;
	y = pRect->y;
	w = pRect->w;
	h = pRect->h;
    }

	/* check the width & height of the frameless gray image.  The gray
	 * frame is in units of PIXELS, not bits.
	 */
    ERR_GEN((x < 0) || (y < 0) || (w <= 0) || (h <= 0), cErrOutOfBounds);
    ERR_GEN(((x+w) > wS) || ((y+h) > hS), cErrOutOfBounds);


	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, w, h, 2, xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* Now that we have legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    
    ERR_PASS_CLEANUP( i_ditherFloyd2 (pIPWState, pS, bplS,
				      x + cFrameBits, y + cFrameBits,
				      w, h, pDRCTable,
				      pD, bplD));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/*

 $Log: floyd3.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.24  1998/01/14 00:11:49  ddavies
 floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.23  1997/04/15 21:48:47  ddavies
 floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.22  1996/01/04 20:22:40  ejaquith
 floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.21  1995/12/13  04:07:16  ejaquith
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:11:15  ejaquith
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:00:22  ejaquith
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/10/05  04:28:09  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.17  1995/10/03  01:21:09  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.16  1995/09/21  22:53:04  ejaquith
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/15  18:01:37  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.14  1995/09/11  22:12:09  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.13  1995/09/10  00:34:10  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.12  1995/09/07  01:31:08  ejaquith
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:19:07  ejaquith
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  01:15:45  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.9  1995/08/31  21:12:50  ejaquith
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/17  22:03:23  ejaquith
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.7  1995/07/29  04:37:55  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.6  1994/10/07  19:18:51  lperry
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.5  1994/06/08  04:10:33  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.4  1994/03/02  03:55:59  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * Add IP_YIELD to give up control after creating dest pixr in dithering ops.
 *
 * Revision 1.3  1993/12/03  01:14:44  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:11:59  ddavies
 * floyd3.c -> /product/ipcore/ipwrap/src/RCS/floyd3.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

