/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	floyd4.c:
 *
 *          Band-processed dithering code. To be used by PerfectScan.
 *
 *	    8 bit/pixel and 4 bit/pixel to 1 bit/pixel banded
 *	    Floyd-Steinberg dither
 *		    w_ditherFloydInit 
 *		    w_ditherFloydProcess
 *		    w_ditherFloydFinish
 *
 *	    8 bit/pixel and 4 bit/pixel to 1 bit/pixel single line
 *	    Floyd-Steinberg dither (to be used with above routines)
 *		    w_ditherFloydSingleProcess
 *
 *	    24 bit/pixel to 8 bit/pixel Banded Floyd-Steinberg Dither
 *		    w_ditherFloyd24To8Init
 *		    w_ditherFloyd24To8Process
 *		    w_ditherFloyd24To8Finish
 *
 *	    24 bit/pixel to 8 bit/pixel color-mapped banded Floyd-Steinberg
 *		    w_ditherFloyd24ToCMapInit
 *		    w_ditherFloyd24ToCMapProcess
 *		    w_ditherFloyd24ToCMapFinish
 *
 *
 *	
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "floyd.pub"
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: floyd4.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/* There are extensive comments about general dithering algorithms in
 * floyd1.c.
 */

/*
 * a dither state structure -- used for color dithering. 
 */

typedef struct
{
/*    RECTANGL	     cropBox;	*/    /* Crop box specified initially */
    Int32	     width;	    /* Width of resulting image */
    Int32	     currentLine;   /* Current line in processing */
    Int32	     linesPerCall;  /* max # of lines to do in a "process" */
    Int32	     lastLine;	    /* Last line to process */
/*    PIXR	    *sourceImage;  */  /* The image being processed */
    UInt8	    *colorMap;	    /* color color map */
    UInt8	    *colorLUT;	    /* The color LUT */
    ADitherState    *sharedState;   /* State thing from ipshared */
} DitherState;


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Prepare to dither RGB to indexed color by bands using linear color map.
 *
 * LongDescription:::
 *  Some color images have 24 bits/pixel.  These have 8
 *  bits/pixel in each of three planes (Red, Green and Blue).  Other
 *  color images have only 8 bits/pixel where each 8-bit entry is mapped
 *  to a 24-bit color using a color map.  It's pretty easy to convert
 *  from color mapped image to a full color image, just replace each
 *  8-bit pixel with its 24-bit equivalent.  It's harder to map
 *  24-bit color images to indexed color images.  A glorious way is to
 *  dither the 24-bit image.
 *<BR><BR>
 *  However, 24-bit color images are huge.  It's nice to be able to read
 *  in a horizontal slice, dither it, read another slice, dither that
 *  and so on.  In order to do this, we must retain state between the
 *  slices so that we don't introduces artifacts at the slice boundaries.
 *  This routine sets up the memory block used to retain that state.
 *  w_ditherFloyd24To8Process is called once for each band to do the real
 *  dithering and w_ditherFloyd24To8Finish tidies up at the end.
 *<BR><BR>
 *  This routine also creates the color map.  It is given the number of
 *  red colors, green colors and blue colors.  It creates a color map
 *  containing redColors*greenColors*blueColors colors.  This product
 *  must be less than or equal to 256.  Since the color map can be easily
 *  recreated, it is not returned.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pRedTRC::        Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map red image pixels before dithering.
 *		     It is normally used to account for the fact that
 *		     displays or printers may print too light or too dark.
 *  pGreenTRC::      Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map green image pixels before dithering.
 *  pBlueTRC::       Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map blue image pixels before dithering.
 *  redColors::      Number of shades of red in the color map.
 *  greenColors::    Number of shades of green in the new color map.
 *  blueColors::     Number of shades of blue in the new color map.
 *  mapOffset::      The 8-bit numbers put out by this routine will start
 *		     at mapOffset and work up.  This allows the outputs
 *		     to fit into some other color map.
 *  lineLength::     Number of pixels in the destination line.  This is
 *		     either the width of the image or the width of some
 *		     sub-rectangle of the image that will be processed.
 *  linesPerCall::   Pointer to the number of lines to be processed
 *		     in each band.  In the olden days, this was a matter of
 *		     negotiation between the client and this routine.  Now,
 *		     we always use the client's suggestion.
 *
 * Output:::
 *  ppDitherState::  Pointer to a pointer to the private state block produced
 *		     here.
 *
 * Returns:::
 *  In addition to the usual stuff, this routine checks to make sure that
 *  redColors*greenColors*blueColors + mapOffset <= 256.  That is, the
 *  number of colors in the color map plus the offset of the first entry
 *  should still fit in 8 bits.
 *
 * Notes:::
 *  This routine creates a private state block.  It's the client's job to
 *  hang onto this thing and pass it to w_ditherFloyd24To8Process and
 *  w_ditherFloyd24To8Finish.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  There are functions in Sandpiper that do about the same thing.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd24To8Init (
    IPWState           *pIPWState,
    UInt8              *pRedTRC,
    UInt8              *pGreenTRC,
    UInt8              *pBlueTRC,
    UInt8               redColors,
    UInt8               greenColors,
    UInt8               blueColors,
    UInt8               mapOffset,
    Int32               lineLength,
    Int32              *linesPerCall,
    void              **ppDitherState)
{
    PROLOGUE;
    Int32	     totalColors;
    DitherState	    *pState;

    ERR_GEN( pRedTRC == NULL || pGreenTRC == NULL || pBlueTRC == NULL ||
	     linesPerCall == NULL || ppDitherState == NULL, cErrNullPtr);

    totalColors = redColors * greenColors * blueColors;
    ERR_GEN( (totalColors == 0) || ((UInt32)totalColors + mapOffset > 256),
	     cErrInvalidParm);

    /* Get a state structure */
    *ppDitherState = NULL;
    pState = (DitherState *) IPW_MALLOC (sizeof(DitherState), cAllocatorData);
    ERR_GEN(pState == NULL, cErrNomem);

    /* Set up */
    pState->width = lineLength;

    /* For color dithering, we always do as many lines as requested,
       since we process one line at a time */
    pState->linesPerCall = *linesPerCall;

    /* Set up the ipshared stuff */
    ERR_PASS_CLEANUP( i_ditherFloydRGBTo8_Init (pIPWState, pRedTRC,
					        pGreenTRC,
					        pBlueTRC,
					        redColors,
					        greenColors,
					        blueColors,
					        mapOffset,
					        lineLength,
					        &(pState->sharedState)));
    *ppDitherState = (void *) pState;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pState, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dither one band of RGB to indexed color using linear color map.
 *
 * LongDescription:::
 *  Some color images have 24 bits/pixel.  These have 8
 *  bits/pixel in each of three planes (Red, Green and Blue).  Other
 *  color images have only 8 bits/pixel where each 8-bit entry is mapped
 *  to a 24-bit color using a color map.  It's pretty easy to convert
 *  from color mapped image to a full color image, just replace each
 *  8-bit pixel with its 24-bit equivalent.  It's harder to map
 *  24-bit color images to indexed color images.  A glorious way is to
 *  dither the 24-bit image.  This routine does just that for one horizontal
 *  band of the color image.
 *<BR><BR>
 *  Why one band?  Well, 24-bit color images are huge.  It's nice to be
 *  able to read in a horizontal slice, dither it, read another slice,
 *  dither that and so on.  In order to do this, we must retain state
 *  between the slices so that we don't introduces artifacts at the 
 *  slice boundaries.  That state was set up by w_ditherFloyd24To8Init.
 *  w_ditherFloyd24To8Process is called once for each band to do the real
 *  dithering and w_ditherFloyd24To8Finish tidies up at the end.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  destLine::       Starting line in destination buffer.
 *  prS::            Pointer to RGB image to be dithered.
 *  xOffset::        x offset into source image.
 *  yOffset::        y offset into source image.
 *  pState::         Pointer to state block generated by w_ditherFloyd24To8Init.
 *
 * Output:::
 *  prD::            Pointer to indexed color image.
 *
 * Notes:::
 *  This routine relies on clients to maintain the offsets into the source
 *  and destination images.  This is sort of a pain, but allows dithering
 *  to and from images that may not reside entirely in memory.  Clients
 *  can adjust the position of the images between calls.
 *<BR><BR>
 *  The thing that makes this type of diffusion relatively straightforward
 *  is that each of the colors is linear, so you can just dither each
 *  plane independently and add the results together to get the final
 *  index.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  There are functions in Sandpiper that do about the same thing.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd24To8Process (
    IPWState           *pIPWState,
    Int32               destLine,
    PIXR               *prS,
    Int32               xOffset,
    Int32               yOffset,
    void               *pState,
    PIXR               *prD)
{
    PROLOGUE;
    DitherState	*pDState;
    Int32	 bplD, bplS;
    PIXR	*subR, *subG, *subB;
    UInt8	*pD, *pR, *pG, *pB;

    ERR_GEN(prS == NULL || prD == NULL || pState == NULL, cErrNullPtr);

	/* Note that the depth is the depth of ONE channel! */
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    pDState = (DitherState *) pState;

    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cRedPlane, &subR));
    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cBluePlane, &subB));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subR, &pR));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subG, &pG));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subB, &pB));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS( w_pixrGetBpl(pIPWState, prD, &bplD));

    ERR_PASS( i_ditherFloydRGBTo8_Process (pIPWState, pDState->sharedState,
					   pR, pG, pB,
					   bplS,
					   pD + destLine * bplD,
					   bplD,
					   xOffset + cFrameBits,
					   yOffset + cFrameBits,
					   pDState->width,
					   pDState->linesPerCall));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free resources used by w_ditherFloyd24To8Process.
 *
 * LongDescription:::
 *  See w_ditherFloyd24to8Process for a description of banded dithering.
 *  That operation requires a private state block and some buffers.
 *  This routine frees those resources.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  ppState::        Pointer to pointer to the dithering state block
 *		     produced by w_ditherFloyd24To8Init.  *ppState will
 *		     be set to NULL by this routine.
 *
 * Notes:::
 *  Clients should not simply free the state block since the various little
 *  memory blocks that are attached to it would be lost.  Clients should
 *  use this routine to tidy up.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  There are functions in Sandpiper that do about the same thing.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd24To8Finish (
    IPWState           *pIPWState,
    void              **ppState)
{
    PROLOGUE;
    DitherState *pDState;

    ERR_GEN(ppState == NULL, cErrNullPtr);

    pDState = (DitherState *) *ppState;
    if (pDState != NULL)
    {
	ERR_PASS( i_ditherFloydRGBTo8_Finish (pIPWState, pDState->sharedState));
	IPW_FREE (pDState, cAllocatorData);
	*ppState = NULL;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Prepare to dither RGB to indexed color by bands using client's color map.
 *
 * LongDescription:::
 *  Some color images have 24 bits/pixel.  These have 8
 *  bits/pixel in each of three planes (Red, Green and Blue).  Other
 *  color images have only 8 bits/pixel where each 8-bit entry is mapped
 *  to a 24-bit color using a color map.  It's pretty easy to convert
 *  from color mapped image to a full color image, just replace each
 *  8-bit pixel with its 24-bit equivalent.  It's harder to map
 *  24-bit color images to indexed color images.  A glorious way is to
 *  dither the 24-bit image.
 *<BR><BR>
 *  However, 24-bit color images are huge.  It's nice to be able to read
 *  in a horizontal slice, dither it, read another slice, dither that
 *  and so on.  In order to do this, we must retain state between the
 *  slices so that we don't introduces artifacts at the slice boundaries.
 *  This routine sets up the memory block used to retain that state.
 *  w_ditherFloyd24ToCMapProcess is called once for each band to do the real
 *  dithering and w_ditherFloyd24ToCMapFinish tidies up at the end.
 *<BR><BR>
 *  The difference between this set of banded dithering routines
 *  and w_ditherFloyd24To8Init/w_ditherFloyd24To8Process is that
 *  the color indices produced by this routine correspond to a color
 *  map passed in by the client.  There are no guarantees about the
 *  contents of this color map, i.e. it is not necessarily linear.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pRedTRC::        Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map red image pixels before dithering.
 *		     It is normally used to account for the fact that
 *		     displays or printers may print too light or too dark.
 *  pGreenTRC::      Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map green image pixels before dithering.
 *  pBlueTRC::       Pointer to a table containing 256 UInt8's.  This
 *		     table is used to map blue image pixels before dithering.
 *  cMap::           Pointer to existing ColorMap as defined in pixr.h.
 *		     This structure contains the number of colors in the
 *		     map (should be 256), the type of the colors (cColorRGB),
 *		     and a pointer to an array of 256 ColorRecords.  Each
 *		     ColorRecord is a 4 byte array containing R,G,B and X.
 *  colorLUT::       Pointer to a color lookup table.  This table should
 *		     be created by w_cMapToLut.  It maps each 24-bit color
 *		     to the nearest color in cMap's color map.
 *  mapOffset::      This number will be added to all 8-bit indices produced.
 *		     Doing so will allow these outputs to be used with
 *		     some other color map.
 *  lineLength::     Number of pixels in the destination line.  This is
 *		     either the width of the image or the width of some
 *		     sub-rectangle of the image that will be processed.
 *  linesPerCall::   This is a pointer to the number of lines to be processed
 *		     in each band.  In the olden days, this was a matter of
 *		     negotiation between the client and this routine.  Now,
 *		     we always use the client's suggestion.
 *  destBits::       The number of bits/index in the output.  This must be
 *		     either 4 or 8 (16 colors or 256 colors).
 *
 * Output:::
 *  ppDitherState::  Pointer to a pointer to the private state block produced
 *		     here.
 *
 * Returns:::
 *  In addition to the usual stuff, this routine checks to see if mapOffset +
 *  cMap->count is greater than or equal to (1 << destBits)).  If so,
 *  we raise a cErrInvalidParm error because the largest output value would
 *  be too big to fit in destBits bits.
 *
 * Notes:::
 *  This routine creates a private state block.  It's the client's job to
 *  hang onto this thing and pass it to w_ditherFloyd24ToCMapProcess and
 *  w_ditherFloyd24ToCMapFinish.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  There are functions in Sandpiper that do about the same thing.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd24ToCMapInit (
    IPWState           *pIPWState,
    UInt8              *pRedTRC,
    UInt8              *pGreenTRC,
    UInt8              *pBlueTRC,
    ColorMap           *cMap,
    UInt8              *colorLUT,
    UInt8               mapOffset,
    Int32               lineLength,
    Int32              *linesPerCall,
    Int32               destBits,
    void              **ppDitherState)
{
    PROLOGUE;
    DitherState     *pDState;

    ERR_GEN( pRedTRC == NULL || pGreenTRC == NULL || pBlueTRC == NULL ||
	     cMap == NULL || colorLUT == NULL || linesPerCall == NULL ||
	     ppDitherState == NULL, cErrNullPtr);
    ERR_GEN((mapOffset + (UInt16)cMap->count >
		(1UL << destBits)), cErrInvalidParm);

    /* Get a state structure */
    *ppDitherState = NULL;
    pDState = (DitherState *) IPW_MALLOC (sizeof(DitherState), cAllocatorData);
    ERR_GEN(pDState == NULL, cErrNomem);

    /* Set up */
    pDState->width = lineLength;

    /* For color dithering, we always do as many lines as requested,
       since we process one line at a time */
    pDState->linesPerCall = *linesPerCall;
    pDState->colorMap = (UInt8 *) &cMap->map[0];

    /* Set up the ipshared stuff */
    ERR_PASS_CLEANUP( i_ditherFloydRGBToCMap_Init (pIPWState, pRedTRC,
						   pGreenTRC,
						   pBlueTRC,
						   (UInt8 *)&cMap->map[0],
						   cMap->count,
						   colorLUT,
						   mapOffset,
						   lineLength,
						   (UInt8) destBits,
						   &(pDState->sharedState)));
    *ppDitherState = (void *) pDState;

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pDState, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dither one band of RGB to indexed color using client's color map.
 *
 * LongDescription:::
 *  Some color images have 24 bits/pixel.  These have 8
 *  bits/pixel in each of three planes (Red, Green and Blue).  Other
 *  color images have only 8 bits/pixel where each 8-bit entry is mapped
 *  to a 24-bit color using a color map.  It's pretty easy to convert
 *  from color mapped image to a full color image, just replace each
 *  8-bit pixel with its 24-bit equivalent.  It's harder to map
 *  24-bit color images to indexed color images.  A glorious way is to
 *  dither the 24-bit image.  This routine does just that for one horizontal
 *  band of the color image.
 *<BR><BR>
 *  Why one band?  Well, 24-bit color images are huge.  It's nice to be
 *  able to read in a horizontal slice, dither it, read another slice,
 *  dither that and so on.  In order to do this, we must retain state
 *  between the slices so that we don't introduces artifacts at the 
 *  slice boundaries.  That state was set up by w_ditherFloyd24ToCMapInit.
 *  w_ditherFloyd24ToCMapProcess is called once for each band to do the real
 *  dithering and w_ditherFloyd24ToCMapFinish tidies up at the end.
 *<BR><BR>
 *  The difference between this routine and w_ditherFloyd24To8Process is that
 *  the color indices produced by this routine correspond to a color
 *  map passed in by the client.  There are no guarantees about the
 *  contents of this color map, i.e. it is not necessarily linear.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  destLine::       Starting line in destination buffer.
 *  prS::            Pointer to RGB image to be dithered.
 *  xOffset::        x offset into source image.
 *  yOffset::        y offset into source image.
 *  pState::         Pointer to state block generated by w_ditherFloyd24To8Init. *
 * Output:::
 *  prD::            Pointer to indexed color image.
 *
 * Notes:::
 *  This routine relies on clients to maintain the offsets into the source
 *  and destination images.  This is sort of a pain, but allows dithering
 *  to and from images that may not reside entirely in memory.  Clients
 *  can adjust the position of the images between calls.
 *<BR><BR>
 *  The thing that makes this harder than w_ditherFloyd24To8Process is that the
 *  colors are not linear.  We have to find the closest client color to the
 *  color in the color map by looking at all three channels at once, then
 *  distribute all three errors.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  Sandpiper contains routines like this one.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd24ToCMapProcess (
    IPWState           *pIPWState,
    Int32               destLine,
    PIXR               *prS,
    Int32               xOffset,
    Int32               yOffset,
    void               *pState,
    PIXR               *prD)
{
    PROLOGUE;
    DitherState	*pDState;
    Int32	 bplD, bplS;
    PIXR	*subR, *subG, *subB;
    UInt8	*pD, *pR, *pG, *pB;

    ERR_GEN(prS == NULL || prD == NULL || pState == NULL, cErrNullPtr);

	/* Note that the depth is the depth of ONE channel! */
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    pDState = (DitherState *) pState;

    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cRedPlane, &subR));
    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cBluePlane, &subB));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subR, &pR));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subG, &pG));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subB, &pB));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS( w_pixrGetBpl(pIPWState, prD, &bplD));

    ERR_PASS( i_ditherFloydRGBToCMap_Process (pIPWState, pDState->sharedState,
					      pR, pG, pB,
					      bplS,
					      pD + destLine * bplD,
					      bplD,
					      xOffset + cFrameBits,
					      yOffset + cFrameBits,
					      pDState->width,
					      pDState->linesPerCall));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free resources used by w_ditherFloyd24ToCMapProcess.
 *
 * LongDescription:::
 *  See w_ditherFloyd24toCMapProcess for a description of banded dithering.
 *  That operation requires a private state block and some buffers.
 *  This routine frees those resources.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  ppState::        Pointer to pointer to the dithering state block
 *		     produced by w_ditherFloyd24ToCMapInit.  *ppState will
 *		     be set to NULL by this routine.
 *
 * Notes:::
 *  Clients should not simply free the state block since the various little
 *  memory blocks that are attached to it would be lost.  Clients should
 *  use this routine to tidy up.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, color, pixel mapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  There are functions in Sandpiper that do about the same thing.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloyd24ToCMapFinish (
    IPWState           *pIPWState,
    void              **ppState)
{
    PROLOGUE;
    DitherState *pDState;
    
    ERR_GEN(ppState == NULL, cErrNullPtr);

    pDState = (DitherState *) *ppState;
    if (pDState != NULL)
    {
	ERR_PASS( i_ditherFloydRGBToCMap_Finish (pIPWState,
						 pDState->sharedState));
	IPW_FREE (pDState, cAllocatorData);
	*ppState = NULL;
    }

    NORMAL_RETURN;
}

    

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Prepare to dither 4 bpp or 8 bpp gray to binary by bands.
 *
 * LongDescription:::
 *  A simple way to convert a gray image to a binary image is to simply
 *  truncate the low order bits.  This is fast, but loses most of the
 *  gray levels.  A more complicated method that retains gray levels (but
 *  tends to fuzz sharp edges) is to use some sort of dithering, like
 *  halftoning or error diffusion.  There are books written on this stuff.
 *  w_ditherFloyd error diffuses an entire image.
 *<BR><BR>
 *  However, gray images are large.  It's nice to be able to read
 *  in a horizontal slice, dither it, read another slice, dither that
 *  and so on.  In order to do this, we must retain state between the
 *  slices so that we don't introduces artifacts at the slice boundaries.
 *  This routine sets up the memory block used to retain that state.
 *  w_ditherFloydProcess is called once for each band to do the real
 *  dithering and w_ditherFloydFinish tidies up at the end.
 *<BR><BR>
 *  Since dithering should be done with the characteristics of the output
 *  device in mind, this routine requires the user to supply a Tone
 *  Reproduction Curve (TRC) table.  All gray pixels are mapped through
 *  this table before being processed further.  In 4 bpp images, there
 *  aren't very many gray levels to begin with so dithered images can
 *  still look grainy unless care is taken to select the numbers in
 *  this table.
 *<BR><BR>
 *  Error diffusion is fairly expensive.  The portion of any page image
 *  occupied by a picture is often small.  To avoid unnecessary processing,
 *  this routine accepts the coords of the upper left corner and the width
 *  of a rectangle to be processed.  The height is determined by the
 *  number of bands processed and the height of each band.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 4 bpp or 8 bpp image to be dithered.
 *  xOffset::        X coord of upper left corner of source rect to be dithered.
 *  yOffset::        Y coord of upper left corner of source rect to be dithered.
 *  lineWidth::      Width in pixels of rectangle to be processed.
 *  pTRC::           Pointer to a table containing 16 or 256 UInt8's.  This
 *		     table is used to map image pixels before dithering.
 *		     It is normally used to account for the fact that
 *		     printers may print too light or too dark.  The table
 *		     contains 16 entries for 4 bpp source images and 256
 *		     entries for 8 bpp source images.  
 *		     The 16  for 4 bpp images numbers should range between
 *		     0 and 255, not 0 and 15.  Proper selection of these 
 *		     numbers has a critical effect on the binary output.
 *  linesPerCall::   Pointer to the number of lines to be processed
 *		     in each band.  In the olden days, this was a matter of
 *		     negotiation between the client and this routine.  Now,
 *		     we always use the client's suggestion.
 *
 * Output:::
 *  ppDitherState::  Pointer to a pointer to the private state block produced
 *		     here.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, pixel mapping, gray
 *
 * Level:::
 *  deterministic, modulo roundoff differences
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  See "Digital Halftoning" by Robert Ulichney.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloydInit (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               xOffset,
    Int32               yOffset,
    Int32               lineWidth,
    UInt8              *pTRC,
    Int32              *linesPerCall,
    void              **ppDitherState)
{
    PROLOGUE;
    Int32	 depth, bpl;
    UInt8	*pS;
    Int32	 stateSize;
    void	*pState;

    ERR_GEN(prS == NULL || pTRC == NULL || linesPerCall == NULL ||
	    ppDitherState == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &depth));
    ERR_PASS( w_pixrGetImageAll (pIPWState, prS, &pS));
    ERR_PASS( w_pixrGetBpl (pIPWState, prS, &bpl));

	/* First, get the size of the state and allocated it.  The
	 * same state block is used for 4 and 8 bit dithering.
	 */
    if (depth == 8)
    {
	ERR_PASS( i_getSizeOfDitherState (pIPWState, bpl,
					  lineWidth, &stateSize));
	pState = (void *) IPW_MALLOC( stateSize, cAllocatorData);
	ERR_GEN(pState == NULL, cErrNomem);
	ERR_PASS_CLEANUP( i_ditherFloydInit (pIPWState, pS, bpl,
					     xOffset + cFrameBits,
					     yOffset + cFrameBits,
					     lineWidth,
					     pTRC,
					     *linesPerCall, 
					     pState,
					     stateSize));
    }
    else	/* depth must be 4, since we checked it above */
    {
	ERR_PASS( i_getSizeOfDither4to1State (pIPWState, bpl,
					      lineWidth, &stateSize));
	pState = (void *) IPW_MALLOC( stateSize, cAllocatorData);
	ERR_GEN(pState == NULL, cErrNomem);
	ERR_PASS_CLEANUP( i_ditherFloyd4to1Init (pIPWState, pS, bpl,
						 xOffset + cFrameBits,
						 yOffset + cFrameBits,
						 lineWidth,
						 pTRC,
						 *linesPerCall, 
						 pState,
						 stateSize));
    }

	/* It worked, so return the state */
    *ppDitherState = pState;

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pState, cAllocatorData);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dither one band of 4 bpp or 8 bpp to binary.
 *
 * LongDescription:::
 *  A simple way to convert a gray image to a binary image is to simply
 *  truncate the low order bits.  This is fast, but loses most of the
 *  gray levels.  A more complicated method that retains gray levels (but
 *  tends to fuzz sharp edges) is to use some sort of dithering, like
 *  halftoning or error diffusion.  There are books written on this stuff.
 *  This routine convertsone band of a 4 bit/pixel or 8 bit/pixel gray
 *  image into a binary image using standard Floyd-Steinberg error diffusion.
 *<BR><BR>
 *  Why one band?  Well, gray images are large.  It's nice to be
 *  able to read in a horizontal slice, dither it, read another slice,
 *  dither that and so on.  In order to do this, we must retain state
 *  between the slices so that we don't introduces artifacts at the 
 *  slice boundaries.  That state was set up by w_ditherFloydInit.
 *  w_ditherFloydProcess is called once for each band to do the real
 *  dithering and w_ditherFloydFinish tidies up at the end.
 *
 * PrivateDescription:::
 *  There is complex assembly code for the primitives of this operation.
 *  There is also a standard C implementation.  The outputs of the
 *  assembly code on the SPARC, assembly code on the x86 and standard
 *  C code are all slightly different.  The differences are only noticeable
 *  if you are counting bits.  They arise from differences in the number
 *  of lines processed before errors are truncated to 8 bits.
 *<BR><BR>
 *  When starting with a 4-bit source image, the 4 bit pixels are
 *  expanded up to 8 bits before being processed.  This allows us to
 *  retain as many diffused error bits as possible.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  destLine::       Starting line in destination buffer.
 *  prS::            Pointer to RGB image to be dithered.
 *  srcXOffset::     x offset into source image.
 *  srcYOffset::     y offset into source image.
 *  pState::         Pointer to state block generated by w_ditherFloydInit.
 *
 * Output:::
 *  prD::            Pointer to binary output image.
 *
 * Notes:::
 *  This routine relies on clients to maintain the offsets into the source
 *  and destination images.  This is sort of a pain, but allows dithering
 *  to and from images that may not reside entirely in memory.  Clients
 *  can adjust the position of the images between calls.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, pixel mapping, gray
 *
 * Level:::
 *  deterministic, modulo roundoff differences
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  See "Digital Halftoning" by Robert Ulichney.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloydProcess (
    IPWState           *pIPWState,
    Int32               destLine,
    PIXR               *prS,
    Int32               srcXOffset,
    Int32               srcYOffset,
    void               *pState,
    PIXR               *prD)
{
    PROLOGUE;
    Int32     bplS, bplD;
    UInt8    *pS, *pD;

    ERR_GEN(prS == NULL || prD == NULL || pState == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS( w_pixrGetBpl (pIPWState, prS, &bplS));
    ERR_PASS( w_pixrGetBpl (pIPWState, prD, &bplD));

	/* i_ditherFloydProcess knows when to call 4to1process */
    ERR_PASS( i_ditherFloydProcess (pIPWState, pS, bplS,
				    srcXOffset + cFrameBits,
				    srcYOffset + cFrameBits,
				    pD + destLine * bplD,
				    bplD,
				    pState));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dither one scan line of 4 bpp or 8 bpp to binary.
 *
 * LongDescription:::
 *  See w_ditherFloydProcess for a description of banded dithering.
 *  This routine exists because clients want to set arbitrary band heights
 *  and the optimized routines like other band heights.  This routine
 *  is used to make up the difference between our optimized bands and
 *  the client's bands.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  destLine::       Starting line in destination buffer.
 *  prS::            Pointer to RGB image to be dithered.
 *  srcXOffset::     x offset into source image.
 *  srcYOffset::     y offset into source image.
 *  pState::         Pointer to state block generated by w_ditherFloydInit.
 *
 * Output:::
 *  prD::            Pointer to binary output image.
 *
 * Notes:::
 *  This routine relies on clients to maintain the offsets into the source
 *  and destination images.  This is sort of a pain, but allows dithering
 *  to and from images that may not reside entirely in memory.  Clients
 *  can adjust the position of the images between calls.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, pixel mapping, gray
 *
 * Level:::
 *  deterministic, modulo roundoff differences
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  See "Digital Halftoning" by Robert Ulichney.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloydSingleProcess (
    IPWState           *pIPWState,
    Int32               destLine,
    PIXR               *prS,
    Int32               srcXOffset,
    Int32               srcYOffset,
    void               *pState,
    PIXR               *prD)
{
    PROLOGUE;
    Int32     bplS, bplD;
    UInt8    *pS, *pD;

    ERR_GEN(prS == NULL || prD == NULL || pState == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS( w_pixrGetBpl (pIPWState, prS, &bplS));
    ERR_PASS( w_pixrGetBpl (pIPWState, prD, &bplD));

    /* i_ditherFloydSingleProcess knows when to call 4to1SingleProcess */
    ERR_PASS( i_ditherFloydSingleProcess (pIPWState, pS, bplS,
					  srcXOffset + cFrameBits,
					  srcYOffset + cFrameBits,
					  pD + destLine * bplD,
					  bplD,
					  pState));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free resources used by w_ditherFloydProcess.
 *
 * LongDescription:::
 *  See w_ditherFloydProcess for a description of banded dithering.
 *  That operation requires a private state block and some buffers.
 *  This routine frees those resources.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  w::              width of dest image, used to clear its frame.
 *  h::              height of dest image, used to clear its frame.
 *
 * Output:::
 *  ppState::        Pointer to pointer to the dithering state block
 *		     produced by w_ditherFloydInit.  *ppState will
 *		     be set to NULL by this routine.
 *  prD::            Pointer to destination image.  Its frame is cleared.
 *
 * Notes:::
 *  Clients should not simply free the state block since the various little
 *  memory blocks that are attached to it would be lost.  Clients should
 *  use this routine to tidy up.
 *
 * FunctionCategory:::
 *  dithering
 *
 * Keywords:::
 *  dithering, depth conversion, pixel mapping, gray
 *
 * Level:::
 *  deterministic, modulo roundoff differences
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  See "Digital Halftoning" by Robert Ulichney.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherFloydFinish (
    IPWState           *pIPWState,
    Int32               w,
    Int32               h,
    void              **ppState,
    PIXR               *prD)
{
    PROLOGUE;
    Int32   bpl;
    UInt8  *pD;

    ERR_GEN( prD == NULL || ppState == NULL || *ppState == NULL,
	     cErrNullPtr);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS( w_pixrGetBpl (pIPWState, prD, &bpl));
    
    /* ditherFloydFinish knows when to call 4to1finish */
    ERR_PASS( i_ditherFloydFinish (pIPWState, pD, bpl, w, h, *ppState));

	/* Free the state since we allocated it */
    IPW_FREE((*ppState), cAllocatorData);
    NORMAL_RETURN;
}


/*
 $Log: floyd4.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.36  1998/02/26 17:51:59  aber
 aber on Thu Feb 26 09:51:11 PST 1998

 floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v

 Update documentation to reflect that the frame will not be trashed
 after all. The user no longer is required to clear the frame, as it
 is now done at the IPShared level.

 Revision 1.35  1998/01/14 00:11:51  ddavies
 floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.34  1997/05/02 01:25:00  ddavies
 ddavies on Thu May 1 18:23:50 PDT 1997

 floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v

 Change linesPerCall arg passed to banded dither routines from pointer to
 band height to the band height itself.  In olden days, the band height was
 negotiated between client and lower level routines.  Now, the client
 just sets it.

 Revision 1.33  1997/04/23 21:43:53  ddavies
 ddavies on Wed Apr 23 14:39:31 PDT 1997

 floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v

 Change w_ditherFloydInit to pre-allocate the DitherState and hand it to
 ip_ditherFloydInit or ip_ditherFloyd4to1Init.
 Free the state in w_ditherFloydFinish.  This is nicely consistent with
 allocating it in w_ditherFloydInit.

 Revision 1.32  1996/02/13 17:28:33  dferrell
 dferrell on Tue Feb 13 09:27:43 PST 1996

 floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v

 Fixed typo/bug in checking for NULL pointers in w_ditherFloydFinish ().

 Revision 1.31  1996/01/04 20:22:43  ejaquith
 floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.30  1995/12/13  04:07:20  ejaquith
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.29  1995/12/01  19:32:20  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Fix acc compiler warnings.
 *
 * Revision 1.28  1995/11/13  23:55:20  ejaquith
 * ejaquith on Mon Nov 13 15:55:00 PST 1995
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Fixed Input::: parameter typo.
 *
 * Revision 1.27  1995/11/13  20:11:19  ejaquith
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.26  1995/10/13  22:00:25  ejaquith
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.25  1995/10/12  04:24:24  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.24  1995/10/05  04:28:12  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.23  1995/10/03  18:17:23  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Check to see if (mapOffset + cMap->count) > number of levels instead of
 * >= number of levels.
 *
 * Revision 1.22  1995/10/03  01:21:14  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.21  1995/09/21  22:53:07  ejaquith
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.20  1995/09/15  18:01:41  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.19  1995/09/11  22:12:13  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.18  1995/09/10  00:34:13  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.17  1995/09/07  01:31:11  ejaquith
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.16  1995/09/06  20:19:09  ejaquith
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.15  1995/09/01  01:15:50  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.14  1995/08/31  21:12:53  ejaquith
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.13  1995/08/17  22:03:25  ejaquith
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.12  1995/07/29  04:38:12  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.11  1994/10/07  19:18:55  lperry
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.10  1994/08/29  18:38:56  ejaquith
 * ejaquith on Mon Aug 29 11:37:57 PDT 1994
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Added ditherFloydSingleProcess which is used with the banded dithering
 * primitives, useful for dithering the last lines in an image (when the
 * band height size is greater than the remaining lines).
 *
 * Revision 1.9  1994/08/25  17:55:40  ejaquith
 * ejaquith on Thu Aug 25 10:53:58 PDT 1994
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Updated ditherFloydFinish by adding three needed parameters to clear
 * the image's right frame on clean up.
 *
 * Revision 1.8  1994/08/09  17:33:29  ejaquith
 * ejaquith on Tue Aug 9 10:33:13 PDT 1994
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Updated comments.  (No code changes)
 *
 * Revision 1.7  1994/06/08  04:10:40  ddavies
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.6  1994/04/27  22:34:09  macleod
 * macleod on Wed Apr 27 15:33:59 PDT 1994
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Added a destination line offset to banded dithering "process" calls.
 *
 * Revision 1.5  1994/04/19  01:17:51  macleod
 * macleod on Mon Apr 18 18:17:49 PDT 1994
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Another "unused" warning..
 *
 * Revision 1.4  1994/04/19  00:23:07  macleod
 * macleod on Mon Apr 18 17:22:18 PDT 1994
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Changed the banded dithering code so that the caller passes the source buffer
 * on each "process" routine, so that one doesn't need to keep the entire source
 * image in memory. the consequence is that the caller needs to keep track of the
 * current offset into the image.
 *
 * Revision 1.3  1994/04/14  01:05:50  macleod
 * macleod on Wed Apr 13 18:05:39 PDT 1994
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Took out a couple of unused variables
 *
 * Revision 1.2  1994/04/13  23:48:45  macleod
 * macleod on Wed Apr 13 16:47:44 PDT 1994
 *
 * floyd4.c -> /product/ipcore/ipwrap/src/RCS/floyd4.c,v
 *
 * Added ditherFloydInit, ditherFloydProcess, and ditherFloydFinish,
 * which are ipwrap interfaces to band-at-a-time processing.
 * They can handle 4- or 8-bit sources, and always dither to 1 bit
 * results (2 bit is not currently supported in band-at-a-time mode)
 *
 * Revision 1.1  1994/04/12  01:17:44  macleod
 * Initial revision
 *
 */
