/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  forms.c--includes
 *
 *		form field segmentation:
 *			ErrCode	w_findFormFields()
 *			ErrCode	w_destroyFormFieldArray()
 *			ErrCode w_findCheckField()
 *
 *		Private:
 *
 *                      ErrCode wp_vectorizeFields()
 *			ErrCode wp_reduceToRes()
 *                      ErrCode wp_sortFields()
 *                      ErrCode wp_groupFields()
 *                      ErrCode wp_appendToFormFieldContainer()
 *                      ErrCode wp_receiveFormFields()
 *                      ErrCode wp_filterHRuleCCs()
 *                      ErrCode wp_findFormFieldRules()
 *                      ErrCode wp_initFormFieldContainer()
 *			ErrCode wp_vectorizeVerticalRules()
 *			ErrCode wp_filterVRulesCCs()
 *			int compare_by_y()
 *			int compare_in_line_by_left_x()
 *                      ErrCode wp_removeStippleAndMakePepperMask()
 *			ErrCode CDECL wp_filterFields()
 */


#include <stdlib.h>
 

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "logicops.h"
#include "morphg.h"
#include "shrscgr.pub"
#include "frames.pub"
#include "defines.pub"
#include "alplib.h"
#include "shrcc.pub"
#include "runmorph.pub"
#include "memory.pub"

/* prototypes */
#include "forms.pub"
#include "forms.prv"
#include "clean.pub"
#include "props.pub"
#include "rast.pub"
#include "logicops.pub"
#include "halftone.pub"
#include "rectops.pub"
#include "morph.pub"
#include "expand.pub"
#include "reduce.pub"
#include "strel.pub"
#include "intops.pub"
#include "clean.prv"
#include "subsmpl.pub"
#include "runfilt.pub"
#include "fill.pub"
#include "shrcc.pub"
#include "prxform.pub"
#include "rotatop.pub"

#ifdef __GNUC__
#include "ansiprot.h"
#endif




IP_RCSINFO(RCSInfo, "$RCSfile: forms.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Finds blank form fields over rulings. 
 *
 * LongDescription:::
 *
 *  Finds form fields in a PIXR that is presumed to be the image of a document 
 *  in which form fields are designated by horizontal rulings. Does this by
 *  finding parts of horizontal rulings that have a sufficient margin of 
 *  white space top side to be considered a form field. Will split fields
 *  at interruptions in the white space. Interruptions could be caused by
 *  text or by vertical tick marks designating character fields. 
 *
 *   1 2 3 4 5 6 7 8 
 *  |_|_|_|_|_|_|_|_|
 *
 *  For example, the above configuration would result in 8 fields, furthermore, 
 *  because of the small and uniform width of all these fields, they would be 
 *  grouped  together, meaning they would all have the same non zero string_id 
 *  (see below).
 *  
 *
 *  Returns fields by assigning a pointer passed by reference to an array of  
 *  FORM_FIELD elements, the array is allocated internal to this routine
 *  using the client supplied ipwrap data malloc routine.
 *
 *  Fields clustered on approximately the same line (y slice) are ordered
 *  from left to right. Lines (where a line is not an image raster, but
 *  a cluster of fields having about the same y) are ordered from top to 
 *  bottom. So fields are raster ordered. 
 *
 *  The members of a FORM_FIELD consist of :
 *
 *      xleft      - start x coordinate of the field
 * 
 *      width      - width of a field 
 *
 *      ybase      - baseline, average y coordinate of the ruling under the 
 *                   field 
 *
 *      maxh       - ceiling of field, minimum height of all the white vertical 
 *                   runs that rest on top of the field. This value may be 
 *                   ridiculously large, since it just indicates the height of
 *                   uninterrupted white above the ruling (the minimum vertical
 *                   white run above each column that rests on the top of the 
 *                   line) 
 *
 *      line_id    - fields are grouped into text lines, text lines are numbered 
 *                   sequentially from 0
 *
 *      string_id  - 0 if the field is not a character field. Character fields
 *                   in a consecutive sequence are all given the same string_id.
 *                   string_ids are numbered sequentially from one, but there 
 *                   may be interruptions in the sequence due to deletion. To see 
 *                   if something is a character field, test the string_id to see
 *                   if it is non zero. The string will end when the string_id
 *                   makes a transition to another value.
 *                   
 *                   
 *
 *  Processing is done by automatically downsampling internally to a suitable
 *  resolution, so this routine accepts any res input image. Only requirement
 *  for decent results is that the res be > than 150 dpi. The highest res 
 *  internal processing will always be done at a resolution <= 250 dpi.
 *  Giving this routine a pixr with res > than 250 dpi is perfectly O.K., 
 *  but not necessary, since it will just be reduced internally.
 * 
 *  Can't tolerate much skew, so deskewing of the source pixr is recommended 
 *  prior to calling. Also removing pictures and stipple may sometimes help 
 *  prevent finding fields within the picture, but this is not really 
 *  necessary since there is some of this sort of thing built in.
 *
 *  This is only designed to work on blank forms !!! Won't work on forms
 *  that have already been filled in since it only looks for blank fields.
 *
 *  This routine may infrequently hallucinate lots of rulings 
 *  in stipple. Disabling finding dotted rulings will mitigate this.
 *
 *  Will find fields with dotted rulings if called with 
 *  enableFindDottedLines set to cTrue, but this may slightly increase  
 *  the number of false fields returned, so it is optional. The false
 *  fields introduced by enabling this option usually result from leader 
 *  dot and sometimes stipple that line up nicely. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Input PIXR
 *  enableFindDottedLines::   if non zero, find dotted lines 
 *  minRulingWidth::	      if non zero, min width of rulings (pixels)
 *  minFieldWidth::           if non zero, min width of any field (pixels)
 *  minIsolatedFieldWidth::   if non zero, min width of a lone field (pixels)
 *  minFieldHeight::          if non zero, min height of any field (pixels)
 *
 * Output:::
 *  pFormFieldArray:: Output, client provide pointer to a pointer to
 *                    a FormField, routine sets this pointer to array 
 *                    of form fields created by this routine. To free,
 *                    use ipwrap data malloc assigned to ipwrap state
 *                    used in this rotuine. 
 *
 *  pFormFieldCount:: Output, points to client's Int32 where a count of form 
 *                    fields in array will be written.  
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 *
 *  Array of form fields allocated internally uses ipwrap data malloc,
 *  can be freed using data free or w_destroyFormFieldArray. 
 *
 *
 *  Very important that resolutions of input pixr in dpi be correct and 
 *  that the ipwrap convention of multiplying the resolutions by 2<<16 be 
 *  followed, since the reductions and parameters in this routine are 
 *  entirely determined by the resolution. In other words resolutions must 
 *  be in accord with the ipwrap ScaledInt1616 convention/interpretation.  
 *  There are also a lot of heuristics built in to this thing that depend
 *  heavily on the resolution being correct.
 *
 *  A big flaw in the current design is that forms fields are found 
 *  without being linked to the ruling which they belong to. This
 *  can be a problem when sorting the fields, since if the ruling 
 *  was skewed, the fields on the ruling will end up with different
 *  y coordinates and my get assigned to different output lines in the
 *  clustering routine. If deskewing is used, this problem rarely arises. 
 *  Still, at some point it might be good to redesign this.
 * 
 *
 * FunctionCategory:::
 *   forms
 *
 * Keywords:::
 *   forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis 
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findFormFields (
    IPWState              *pIPWState,
    PIXR                  *prSrc,
    Bool                  enableFindDottedLines,
    UInt32		  minRulingWidth,
    UInt32		  minFieldWidth,
    UInt32		  minIsolatedFieldWidth,
    UInt32		  minFieldHeight,
    FormField             **pFormFieldArray,
    UInt32                 *pFormFieldCount)
{
    PROLOGUE;
    PIXR     *pPixrA=NULL, *pPixrB=NULL, *pPixrC=NULL;
    PIXR     *pPixrD=NULL;
    PIXR     *pPixr0=NULL;
    PIXR     *pPixrW=NULL;
    PIXR     *pPixrRules=NULL;

    STREL    *strel = NULL;
    ScaledInt1616 xres,yres,xres0,yres0,xresSrc,yresSrc;
    Int32 w0,h0;
    Int32 iredx,iredy;
    Int32 iredx_total,iredy_total;
    Int32 iredx_display,iredy_display;

    FormFieldContainer ffCont;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS(w_pixrGetResolutionDPI(pIPWState,prSrc,&xresSrc,&yresSrc) );

    if( (xresSrc<=0)||(yresSrc<=0) )
       return cErrPixr_resolution;

    /* initially reduce to res near 175x175 dpi */
    ERR_PASS_CLEANUP(wp_reduceToRes(pIPWState,prSrc,175,175,&iredx,&iredy,
    &pPixr0));


    iredx_total = iredx;
    iredy_total = iredy;
    minRulingWidth >>= iredx;


    /* Round up image width and height so that we will have alignment of 
       all the reduced resolution bitrects down to reduction by 1<<5 */

    ERR_PASS(w_pixrGetResolutionDPI(pIPWState,pPixr0,&xres0,&yres0) );
    ERR_PASS(w_pixrGetHeight(pIPWState,pPixr0,&h0) );
    ERR_PASS(w_pixrGetWidth(pIPWState,pPixr0,&w0) );
    w0= ((w0+31)>>5)<<5;
    h0= ((h0+31)>>5)<<5;
    ERR_PASS_CLEANUP(w_createPixr(pIPWState,w0,h0,1,xres0,yres0,cFrameBits,
    &pPixrA));
    /* according to internal ipwrap comments, O.K. use copy to copy a src 
     * PIXR into a dest PIXR whose width and height is within 1 frame width 
     * of the src, the rounded up images dimensions satisify this. */
    ERR_PASS_CLEANUP (w_copy(pIPWState, pPixr0, &pPixrA));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixr0));

    ERR_PASS_CLEANUP(wp_removeStippleAndMakePepperMask(pIPWState,
    pPixrA,&pPixrB));

    /* find horizontal rulings */
    ERR_PASS_CLEANUP(wp_findFormFieldRules(pIPWState,pPixrA,
    enableFindDottedLines,minRulingWidth,&pPixrRules));


    /* subtract form field rules from pepper since you don't want to filter
          out dotted rulings  */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrRules, SUBTRACT, pPixrB));
    /* subtract pepper  */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrB, SUBTRACT, pPixrA));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrB));

    /* can now reduce the resolution of the rulings to 50*100 dpi */
    ERR_PASS_CLEANUP(wp_reduceToRes(pIPWState,pPixrRules,50,100,&iredx,
    &iredy,&pPixrRules));
    iredx_display= iredx;
    iredy_display= iredy;


    /* 100x100 dpi map for fine stuff */
    ERR_PASS_CLEANUP(wp_reduceToRes(pIPWState,pPixrA,100,100,&iredx,&iredy,
    &pPixrA));

    iredx_total += iredx;
    iredy_total += iredy;

    /* 50x100 dpi map for coarse stuff that comes later (we need to save 
       pPixrA at this res ) */
    ERR_PASS_CLEANUP(wp_reduceToRes(pIPWState,pPixrA,50,100,&iredx,&iredy,
    &pPixrC));

    iredx_total += iredx;
    iredy_total += iredy;

    ERR_PASS(w_pixrGetResolutionDPI(pIPWState,pPixrA,&xres,&yres) );
    /* scale resolution for unscaled interpretation */
    xres=ScaledInt1616_TO_Int32(xres);
    yres=ScaledInt1616_TO_Int32(yres);

    /* find fields over rulings that are at least 1/8" wide, as long as they
       are only interrupted by thin strokes */ 

    /* 
     * Note, dilate-erode and erode-dilate sequences are used in the following
     * instead of w_close and w_open, which allocates an intermediate bitrect,
     * just to avoid the allocation of the intermediate. Not a big
     * deal one way or the other. I just started doing this by habit. Apologies
     * to the reader. Also note that I don't add padding to accomodate closes
     * so that things will be bogus within 1/2 of a strel width from the edge,
     * but I don't use very long strels, and most of the features I am looking 
     * for should be well away from the edges and not interfere with finding form 
     * fields. 
     */ 

    /* close by 1/10"  to glom words together */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, MAX(1,xres/10), 1, xres/20, 
    0, &strel));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrA, strel, &pPixrB));
    ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrB, strel, &pPixrA));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &strel));

    /* open by 1/16" to remove narrow strokes and noise  */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, MAX(3,xres/16), 1, 
    MAX(1,xres/32), 0, &strel));
    ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrA, strel, &pPixrB));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrB, strel, &pPixrA));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &strel));

    /* now that we've taken care of the fine stuff, we can reduce the res 
       some more */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrB));
    ERR_PASS_CLEANUP(wp_reduceToRes(pIPWState,pPixrA,50,100,&iredx,&iredy,
    &pPixrA));
    ERR_PASS(w_pixrGetResolutionDPI(pIPWState,pPixrA,&xres,&yres) );
    /* scale resolution for unscaled interpretation */
    xres=ScaledInt1616_TO_Int32(xres);
    yres=ScaledInt1616_TO_Int32(yres);

    /* close by 1/7" (old val = 1/4") to smear text lines completely together */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, MAX(1,xres/7), 1, xres/14, 0,
    &strel));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrA, strel, &pPixrB));
    ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrB, strel, &pPixrA));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &strel));

    /* rotate white space map by 90 degrees so that it can be used to find
       vertical runs later (using horizontal run finding routine) */
    ERR_PASS_CLEANUP (w_rotate270 (pIPWState, pPixrA, &pPixrW) );


    /* dilate rulings up vertically since we find fields by taking intersection
       of ruling map with white map. If we don't do this, there will be no 
       intersection. The amount we dilate here will trade off noise rejection
       for false fields. The more we dilate, the more jaggy the line can be,
       but we are also more likely to false alarm. Think of this as edge 
       finding, because we need a clean white on top of black juncture of 
       sufficient duration to get a ruling  */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1, MAX(1,yres/25), 0, 
    MAX(0,yres/25-1), &strel));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrRules, strel, &pPixrD));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrRules));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &strel));

    /* AND vertically dilated rules with the original, this is to get thin
       black things back (because they may be character field separators),
       but only if those thin things occur near the tops of the rulings, 
       otherwise they might just be noise or descenders from the text line 
       above. */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrD, AND, pPixrC));
    /* add thin stuff near the line to pPixrA */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrC, UNION, pPixrA));

    

    /* close by minimum width/height for fields (1/10"x 1/12") */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, MAX(1,xres/10), 
    MAX(1,yres/12), xres/20, yres/24, &strel));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrA, strel, &pPixrB));
    ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrB, strel, &pPixrA));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &strel));


    /* take intersection of thickened rulings map with white in pPixrA */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrA, SUBTRACT, pPixrD));

    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &pPixrA));
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &pPixrB));
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &pPixrC));

    /* initialize a form field container */
    ERR_PASS_CLEANUP (wp_initFormFieldContainer(pIPWState,&ffCont));

    /* vectorize the form fields */
    ERR_PASS_CLEANUP (
			wp_vectorizeFields (
					pIPWState,
					pPixrD,
					pPixrW,
					(1<<iredx_total),
					(1<<iredy_total),
				 	&ffCont	
					)
                     );
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &pPixrD));
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &pPixrW));

    /* sort the form fields */
    ERR_PASS_CLEANUP (
			wp_sortFields 	(
					pIPWState,
                                        ScaledInt1616_TO_Int32(xresSrc),
                                        ScaledInt1616_TO_Int32(yresSrc),
					&ffCont	
					)
                     );

    /* 
       A character field is field where the user is meant to type a
       single character, strings of such fields occur for things
       like phone numbers and social security numbers. Form fields
       that are believed to belong in the same string of character
       fields are assigned the same string ID.
    */
    ERR_PASS_CLEANUP (
			wp_groupFields 	(
					pIPWState,
                                        ScaledInt1616_TO_Int32(xresSrc),
                                        ScaledInt1616_TO_Int32(yresSrc),
					&ffCont	
					)
                     );

    if(minFieldWidth || minFieldHeight || minIsolatedFieldWidth)
    {
       ERR_PASS_CLEANUP (
			wp_filterFields	(
					pIPWState,
					minFieldWidth,
                                        minIsolatedFieldWidth,
					minFieldHeight,
					&ffCont
					)
                     );
       ERR_PASS_CLEANUP (
			wp_groupFields 	(
					pIPWState,
                                        ScaledInt1616_TO_Int32(xresSrc),
                                        ScaledInt1616_TO_Int32(yresSrc),
					&ffCont	
					)
                     );
    }

    *pFormFieldArray=ffCont.array;
    *pFormFieldCount= (UInt32) ffCont.count;



    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (pPixrA) w_destroyPixr (pIPWState, &pPixrA);
    if (pPixrB) w_destroyPixr (pIPWState, &pPixrB);
    if (pPixrC) w_destroyPixr (pIPWState, &pPixrC);
    if (pPixrD) w_destroyPixr (pIPWState, &pPixrD);
    if (pPixr0) w_destroyPixr (pIPWState, &pPixr0);
    if (pPixrW) w_destroyPixr (pIPWState, &pPixrW);
    if (pPixrRules) w_destroyPixr (pIPWState, &pPixrRules);
    if (strel) w_destroyStrel (pIPWState, &strel);

    ERR_RETURN;
}  /* w_findFormFields () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Finds the bounding box of a component near the input coordinate. 
 *
 * LongDescription:::
 *
 *  This is to assist checkmarking a form. Since we don't assume any 
 *  constraint for the shape of the component the end user wants to put a
 *  check in, we just grab the most plausible white or black connected 
 *  component (CC), solely based on the location and size of its bounding 
 *  box (BB), in the vicinity of input coordinate.
 *
 *  The idea is that the end user will click his mouse inside or near the 
 *  thing he wants to check and we'll search around that mouse click point
 *  for a reasonably sized blob near the mouse pointer.
 *
 *  Here are the precedence rules :
 *
 *    Smallest area white CC that meets the constaints, is totally
 *    within the search box, and whose BB contains the input point. 
 *
 *           and if that isn't satisifed .....
 *
 *    Hole nearest the input point that meets the constaints, 
 *    and is totally within the search box. Where a hole is defined as the 
 *    largest white component within the bounding box of a black component. 
 *     
 *  The idea is that, although this is far from perfect, the user should get the
 *  hang of it and be able to effectively use it to check away. Presumably the
 *  minW and minH will be set to filter out noise and the maxW and maxH will 
 *  be set so the user won't get something too big to possibly be a check 
 *  field. There are other parameters that dictate how "rectangular" the
 *  hole must be and minimum and maximum aspect ratios. So the user doesn't
 *  simply get the hole closest to his pick. But, if the closest component
 *  is what you want, just be very liberal about setting these things so 
 *  that you just filter out noise. 
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *
 *  pPixrS::	input Pixr
 *
 *  x::         x coordinate for center of search 
 *
 *  y::         y coordinate for center of search 
 *
 *  searchW::	search rectangle width in pixels
 *              default is 1" (using source pixr res) if set to 0
 *
 *  searchH::   search rectangle height in pixels
 *              default is 1" (using source pixr res) if set to 0
 *
 *  minW::      minimum width, in pixels, of the hole in a check component, 
 *              a hole is defined as the largest white component
 *              within the bounding box of a black component. 
 *              default is 1/30 (using source pixr res) if set to 0
 *
 *  maxW::      maximum width, in pixels,  of the hole in a check component
 *              default is 1/2" (using source pixr res) if set to 0
 *
 *  minH::      minimum height, in pixels, of the hole in a check component
 *              default is 1/30 (using source pixr res) if set to 0
 *
 *  maxH::      maximum height, in pixels, of the hole in a check component
 *              default is 1/2 (using source pixr res) if set to 0
 *
 *  maxpect::   (w*xres)/(h*yres) and (h*yres)/(w*xres) must be less than 
 *              this default is 4 if set to 0. Presumption is that check
 *              fields bounding boxes will not deviate that much from being 
 *              square.
 * 
 *  minRectangularity::
 *              On a scale of 0 to 100, how rectangular does the hole (white)
 *              in the component have to be ? 
 *  
 *              Perfect rectangle (remember, nothing is perfect ) - 100  
 *              perfect circle - pi/4, approx 75
 *  
 *              if set to 0, default is 65.
 *    
 *              I define rectangularity as:
 *            
 *                    100*mass/(bounding_box_width*bounding_box_height)
 *        
 *  
 *              This parameter only applies if the pick is outside the bounding
 *              box of the component. If the pick is inside the components
 *              bounding box, the rectangularity is ignored. 
 *               
 *
 *
 *  shapeWeight::
 *		Instead of just picking the component that minimizes the
 *              distance from the pick to a "valid" checkbox component, you 
 *              may also want to weight the rectangularity of the hole 
 *              (check field) in the component . The shapeWeight linearly
 *              scales "non rectangularity" to a "distance" as follows : 
 *               
 *              (shapeWeight/100)*xres*(1-mass/(boundbox_width*boundbox_height)) 
 * 
 *              default is 0.
 *    
 *              So shape weight of 100 would add a distance of a quarter of an 
 *              inch to a perfect circle, so if the pick were outside the 
 *              circle we'd favor a perfect rectangle whose distance  
 *              from the pick was within 1/4" of the distance from the 
 *              bounding box of the circle to the pick. 
 * 
 *              The idea is to favor more boxy candidates, even though they
 *              may be a little bit further from the pick. 
 * 		
 *              This parameter only applies if the pick is outside the bounding
 *              box of the component. If the pick is inside the components
 *              bounding box, the rectangularity is ignored. 
 *              
 *              Play with it ! 
 *              
 *
 * Output:::
 *
 *  pHit::	pointer to Bool where cTrue is returned by ref. if checkbox 
 *		found, otherwise cFalse is returned.	
 *
 *  pLeftX::	pointer to Int32 where left x coord of checkbox field is 
 *              returned by reference. 
 *
 *  pBotY::	pointer to Int32 where bottom y coord. of checkbox field is 
 *              returned by reference.
 *
 *  pWidth::    pointer to Int32 where width of checkbox field is returned 
 *		by reference. 
 *
 *  pHeight::	pointer to Int32 where height of checkbox field is returned by
 * 		reference.
 * 
 * 
 * Returns:::
 *  error status
 *
 * Notes:::
 * 
 *  Temporarily modifies source image (inverts the search rectangle), but 
 *  restores it before return. Source image could be left this way in 
 *  situations where an error in the inversion call prevents this routine 
 *  from proper termination. Also not suitable for multithreaded applications 
 *  in which another thread may be accessing the source image while this 
 *  routine is being called this scenario seems far fetched.
 *
 * 
 *
 * FunctionCategory:::
 *   forms
 *
 * Keywords:::
 *   forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findCheckField(
	IPWState *pIPWState,
	PIXR *pPixrS,
	Int32 x,
        Int32 y,
        UInt32 searchW,
        UInt32 searchH,
        UInt32 minW,
        UInt32 maxW,
        UInt32 minH,
        UInt32 maxH,
	UInt32 maxpect,
        UInt32 minRectangularity,
        UInt32 shapeWeight,
	Bool  *pHit,
        Int32 *pLeftX,
	Int32 *pBotY,
        Int32 *pWidth,
	Int32 *pHeight
	)
{
    PROLOGUE;
    CheckBoxParam param;
    CheckBoxPacket white,black; 
    Bool image_inverted=0;
    UInt8 *pIm;
    Int32 bpl,wIm,hIm;
    ScaledInt1616 xres,yres; 
    Int32 minX,maxX,minY,maxY;
    Int32 ret;
    Int32 leftSearchX,rightSearchX,topSearchY,botSearchY;

    *pHit=0;

    ERR_GEN_PIXR (pPixrS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* get pixr parameters we need */
    ERR_PASS (w_pixrGetImageAll (pIPWState, pPixrS, &pIm));
    ERR_PASS (w_pixrGetBpl (pIPWState, pPixrS, &bpl));
    ERR_PASS (w_pixrGetWidth (pIPWState, pPixrS, &wIm));
    ERR_PASS (w_pixrGetHeight (pIPWState, pPixrS, &hIm));
    ERR_PASS (w_pixrGetResolutionDPI(pIPWState,pPixrS,&xres,&yres) );

    /* unscale resolutions (so they can be interpreted as pixels) */
    xres=ScaledInt1616_TO_Int32(xres);
    yres=ScaledInt1616_TO_Int32(yres);


    /* set up defaults */
    if(!searchW)
        searchW=xres;

    if(!searchH)
        searchH=yres;

    if(!minW)
        minW=xres/30;

    if(!minH)
        minH=yres/30;

    if(!maxW)
        maxW=xres/2;

    if(!maxH)
        maxH=yres/2;

    if(!maxpect)
        maxpect= 4;

    if(!minRectangularity)
       minRectangularity=65;

    /* compute search rectangle */ 
    leftSearchX=MAX(0,x-((Int32)searchW)/2);
    rightSearchX=MIN(wIm-1,x+((Int32)searchW)/2);
    topSearchY=MAX(0,y-((Int32)searchH)/2);
    botSearchY=MIN(hIm-1,y+((Int32)searchH)/2);
    
    searchW=rightSearchX-leftSearchX+1;
    searchH=botSearchY-topSearchY+1;

    
    /* initialize parameter packet for callback */

    /* 
       Coordinates in terms of unframed image for connected
       component finding routine.
    */
    minX= leftSearchX+cFrameBits;
    maxX= rightSearchX+cFrameBits;
    minY= topSearchY+cFrameBits;
    maxY= botSearchY+cFrameBits;

    param.pPixrS=pPixrS;
    param.x =x+cFrameBits;
    param.y =y+cFrameBits; 

    white.white=1;
    black.white=0;


    param.maxX=maxX;
    param.maxY=maxY;
    param.minX=minX;
    param.minY=minY;


    param.xres=xres;
    param.yres=yres;

    param.maxW=maxW;
    param.maxH=maxH;

    param.minW=minW;
    param.minH=minH;
    param.maxpect=maxpect;
    param.minRectangularity=minRectangularity;
    param.shapeWeight=shapeWeight;

    white.dist=black.dist=0x7fffffff;
    white.pParam=black.pParam=&param;
    white.rectangularity=0;


    image_inverted=0;

    /* invert search rectangle in image */

    ERR_PASS_CLEANUP( w_rasterop(
				 pIPWState,
				 leftSearchX,
				 topSearchY,
				 pPixrS,
			         PIX_NOT(PIX_SRC),
				 leftSearchX,
				 topSearchY,
				 (Int32) searchW,
				 (Int32) searchH,
				 pPixrS
				 )
		    );
    image_inverted=1;
					
    /* find best white candidate */
    ERR_PASS_CLEANUP( 
	i_findConnectedComponentBoundingBoxes (pIPWState, pIm, bpl, minX, 
					       minY,
					       (Int32)searchW,(Int32)searchH,
					       minW, minH,
					       1, (void *)(&white),
					       wp_findCheckBoxCC, &ret));
    
 
    /* invert back */
    image_inverted=0;
    ERR_PASS( w_rasterop(
			 pIPWState,
			 leftSearchX,
			 topSearchY,
			 pPixrS,
			 PIX_NOT(PIX_SRC),
			 leftSearchX,
			 topSearchY,
			 (Int32)searchW,
			 (Int32)searchH,
			 pPixrS)
		    );
 
    /* if whiteHit, then finished */   
    if(white.dist==0)
    {
        *pHit=1;
        *pLeftX=white.x-cFrameBits;
        *pBotY=white.y-cFrameBits+white.h;
        *pWidth=white.w;
        *pHeight= white.h;
        NORMAL_RETURN;
    }

    /* find best black candidate */
    ERR_PASS_CLEANUP(
	i_findConnectedComponentBoundingBoxes (pIPWState, pIm, bpl, minX, 
					       minY,
					       (Int32)searchW, (Int32)searchH,
					       minW, minH,
					       1, (void *)(&black),
					       wp_findCheckBoxCC, &ret));


    if( white.dist<black.dist )
    {
        *pHit=1;
        *pLeftX=white.x-cFrameBits;
        *pBotY=white.y-cFrameBits+white.h;
        *pWidth=white.w;
        *pHeight= white.h;
    }
    else
    {
       if(black.dist<0x7fffffff)
       {
          *pHit=1;
          *pLeftX=black.x-cFrameBits;
          *pBotY=black.y-cFrameBits+black.h;
          *pWidth=black.w;
          *pHeight=black.h;
       }
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if(image_inverted)
    {
       w_rasterop(
				 pIPWState,
				 leftSearchX,
				 topSearchY,
				 pPixrS,
			         PIX_NOT(PIX_SRC),
				 leftSearchX,
				 topSearchY,
				 searchW,
				 searchH,
				 pPixrS
		  );
    }
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Frees array of form fields.
 *
 * LongDescription:::
 *  Just calls ipwrap data free routine on pointer. Paired with w_findFormFields
 *  since allocation of the array of form fields this routine returns is hidden
 *  within this routine.
 *
 *
 * Input:::
 *  pIPWState::        State structure
 *  pFormFieldArray::  pointer to array of form fields. 
 *
 *
 * FunctionCategory:::
 *   forms
 *
 * Keywords:::
 *   forms
 *
 * Level:::
 *  deterministic 
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyFormFieldArray(
    IPWState *pIPWState, 
    FormField **pFormFieldArray)
{
   PROLOGUE;
   if(*pFormFieldArray)
   {
      IPW_FREE(*pFormFieldArray,cAllocatorData);
      *pFormFieldArray=0;
   }
   NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Makes a horizontal ruling map for an input image, form field finding step.
 *
 * LongDescription:::
 *  Makes a horizontal ruling map for an input image as a step in finding form 
 *  fields. Only appropriate for finding rulings under blank fields. Also, if
 *  a ruling is deemed to be the top of a box, it will be filtered out, since
 *  it is rare for box tops to be fields. Actually, routine doesn't really look
 *  for boxes, it filters out any horizontal rulings that have tops of vertical 
 *  rulings near the left and right endpoints of the horizontal rulings.
 *  Optimized for resolution range from 150-200 dpi.
 *  prSrc and prDest must be the same size and depth.
 *
 * PrivateDescription:::
 *  Designed with 4 levels of fussiness. Long rulings (or, more accurately
 *  for rulings composed of long horizontal run lengths) only have to meet
 *  a liberal stoke thickness constraint. Shorter rulings have to meet a
 *  stricter thickness constraint under more severe horizontal agglomeration.
 *  Even shorter rulings have to have a clean margin underneath them (don't
 *  worry about the top side, since this is dealt with in the calling routine.
 *  Broken or dotted rulings must have a clean white horizontal stripe on both
 *  the bottom and the top. 
 * 
 *
 * Input:::
 *  pIPWState::      State structure
 *  pPixrS::         Input PIXR
 *  dotted::	     when nonzero, enables finding dotted rulings. 
 *  minRulingWidth:: minimum width for a horizontal ruling (if 0, use defaults). 
 *
 * Output:::
 *  ppPixrDest::     Output PIXR mask of horizontal rulings presumed to 
 *                   belong to form fields in a blank form.
 *
 * Returns:::
 *  error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 * Citations:::
 *  based on w_removeHVRules by dferrell
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_findFormFieldRules (
    IPWState           *pIPWState,
    PIXR               *pPixrS,
    Bool	       dotted,
    UInt32             minRulingWidth,
    PIXR              **ppPixrDest)
{
    PROLOGUE;
    Int32     ret, wSrc, hSrc, wDest, hDest, bplH, w, h;
    STREL    *thicken = NULL,
             *hStrelA = NULL,
             *hStrelB = NULL,
             *hStrelC = NULL,
             *hStrelDot = NULL,
             *vStrelA = NULL;

    PIXR     *pPixrA=NULL,
	     *pPixrB=NULL,
	     *pPixrC=NULL,
	     *pPixrD=NULL,
	     *pPixrOr=NULL,
	     *pPixrAnd = NULL,
#if 0 /* debug */
	     *pPixrTest = NULL,
#endif
	     *pPixrDest;
    FilterRulesCCPacket packet;
    UInt8    *pH = NULL;
    void     *runFilterState;
    Int32     minLinearSize;
    Int32     minRunLengthDotted;
    Int32     maxLinearSize;
    Int32     maxThicknessDotted;
    Int32     minThickness;

    Int32     minRunLengthThick,maxThicknessThin;
    Int32     maxThicknessThick;
    Int32     minRunLengthThin;
    Int32     minRunLengthThinAndClean;
    ScaledInt1616 xres,yres;
    Int32 max_width;
    Int32 xfactor;
    Int32 xshift;
    Int32 *index_array;
    Int32 iy,iy_prev,i;
    FormFieldContainer vRuleContainer;
    FormField *pFF;


    ERR_PASS (w_pixrGetWidth (pIPWState, pPixrS, &wSrc));
    ERR_PASS (w_pixrGetHeight (pIPWState, pPixrS, &hSrc));


    

    ERR_GEN_PIXR (pPixrS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    pPixrDest = *ppPixrDest;
    if(pPixrDest)
    {
       /* Sanity checks */
       ERR_GEN_PIXR (pPixrDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);
       ERR_PASS (w_pixrGetWidth (pIPWState, pPixrDest, &wDest));
       ERR_PASS (w_pixrGetHeight (pIPWState, pPixrDest, &hDest));
       ERR_GEN ( hSrc != hDest || wSrc != wDest, cErrInvalidParm);
    }

    ERR_PASS(w_pixrGetResolutionDPI(pIPWState,pPixrS,&xres,&yres) );
    /* scale resolution for unscaled interpretation */
    xres=ScaledInt1616_TO_Int32(xres);

    /* reduce horizontal res, one map with AND, other with OR */
    if(xres>80)
    {
       xfactor=4;
       xshift=2;
       ERR_PASS (w_reduceCols (pIPWState, pPixrS, xfactor, AND, &pPixrAnd));
       ERR_PASS (w_reduceCols (pIPWState, pPixrS, xfactor, OR, &pPixrOr));
    }
    else
    {
       xfactor=2;
       xshift=1;
       ERR_PASS (w_reduceCols (pIPWState, pPixrS, xfactor, AND, &pPixrAnd));
       ERR_PASS (w_reduceCols (pIPWState, pPixrS, xfactor, OR, &pPixrOr));
    }

    minRulingWidth >>= xshift;


    /* 
     * Find vertical rulings, for now only so we can filter out the tops of
     * vertical rulings.
     */
    ERR_PASS ( wp_vectorizeVerticalRules ( pIPWState, pPixrS, &vRuleContainer));

    /* sort the vertical rulings by top y */

    qsort(vRuleContainer.array,vRuleContainer.count,sizeof(FormField),
    compare_by_y);

    /* create an index into the the vertical rulings by y, scale the x 
       coordinates according to the scaling factors used to shrink the 
       feature pixrs */
    index_array= (Int32 *) IPW_MALLOC(hSrc*sizeof(Int32),cAllocatorData);
    memset(index_array,0,hSrc*sizeof(Int32));
  
    pFF=vRuleContainer.array;

    iy_prev=-1;

#if 0  /* debug */
    w_clonePixr(pIPWState,pPixrAnd,&pPixrTest);
#endif /* debug */

    /* compute the index array for faster access */

    for(i=0;i<vRuleContainer.count;++i)
    {
       iy=pFF->ybase;
       /* we'll recycle the width field to hold the right x coordinate */
       FF_RIGHTX_IN_WIDTH(pFF)= (pFF->xleft+pFF->width-1)>>xshift; 
       pFF->xleft>>=xshift;

#if 0  /* debug */
       w_pixrPutLine(pIPWState,pPixrTest,pFF->xleft,pFF->ybase,
       FF_RIGHTX_IN_WIDTH(pFF) ,pFF->ybase+pFF->maxh,1);
#endif /* debug */

       if ( (iy>iy_prev) && (iy<hSrc) && (iy>=0) )
       {
          index_array[iy]= i+1;
       }   
       pFF++;
       iy_prev=iy;
    }


    ERR_PASS(w_pixrGetResolutionDPI(pIPWState,pPixrAnd,&xres,&yres) );
    ERR_PASS(w_pixrGetWidth(pIPWState,pPixrAnd,&max_width) );
    /* scale resolution for unscaled interpretation (pixels) */
    xres=ScaledInt1616_TO_Int32(xres);
    yres=ScaledInt1616_TO_Int32(yres);

    /* set up strel pointers */
    ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel3, &thicken));
    ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel3h, &hStrelA));
    ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel5h, &hStrelB));
    ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel7h, &hStrelC));

    if(xres>40)
    {
       ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel3h, &hStrelDot));
    }
    else
    {
       ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel2h, &hStrelDot));
    }

    minThickness=1;
    maxLinearSize=max_width; /* the run filter routines need a max */ 

    if(minRulingWidth)
    {
       minLinearSize= minRulingWidth;
    }
    else
    {
       minLinearSize= xres/7;
    }

    /* min run length if over long white run */
    minRunLengthThinAndClean=MAX(1,xres/8);

    /* min run length if ruling is generally thin */
    minRunLengthThin=MAX(1,xres/4);

    /* min run length for thicker rulings */
    minRunLengthThick= MAX(1,xres/2); 

    /* min run length for dotted rulings */
    minRunLengthDotted= MAX(1,xres/4);

    maxThicknessThin= MAX(1,yres/30); /* probably can relax this to 1/25" */
    maxThicknessDotted= MAX(1,yres/30); /* probably can relax this to 1/25" */
    maxThicknessThick= MAX(1,yres/12);




    /** Find the horizontal rule-like things **/

    /* 
     * Note, dilate-erode and erode-dilate sequences are used in the following
     * instead of w_close and w_open, which allocates an intermediate bitrect,
     * just to avoid the allocation of the intermediate. Probably not a big
     * deal one way or the other. I just started doing this by habit. 
     *
     */ 

    /* get ready for run filtering */
    ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, pPixrAnd, cBothRuns,
                                       &runFilterState));

    /* Do a horizontal close, followed by an opening so that words with 
       joining characters, reverse video, and pictures will tend to become 
       thick and will be filtered out by the thickness filter, but we 
       want to ignore really thin, isolated things, hence the opening */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrOr, hStrelA, &pPixrA));
    ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrA, hStrelB, &pPixrB));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrB, hStrelA, &pPixrA));





    /* Get only things which are the right thickness */
    ERR_PASS_CLEANUP (w_filterVerticalRuns (pIPWState, runFilterState,
                                            pPixrA, 1, 0,
					    minThickness, maxThicknessThick, 
					    cPassRuns, &pPixrB));

    /* Keep only thin things */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrAnd, AND, pPixrB));

    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, runFilterState,
                                              pPixrB, 1, 0,
					      minRunLengthThick, 
				  	      maxLinearSize, 
					      cPassRuns, &pPixrC));

    /* Try to close small gaps between runs for larger run lengths */
    /* big horizontal closing (this is to keep long rulings in one
       piece, we don't risk this for the shorter rulings */
    ERR_PASS_CLEANUP(w_clear(pIPWState,pPixrA));
    ERR_PASS_CLEANUP(w_invert(pIPWState,pPixrA,&pPixrA));
    ERR_PASS_CLEANUP(w_filterHorizontalRuns(pIPWState,runFilterState,pPixrC,0,0,1,xres/2,
    cRejectRuns,&pPixrA)); 

    /* thicken larger run lengths them and clip to the reduced original */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrA, thicken, &pPixrC));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrOr, AND, pPixrC));


    /* special treatment for short rulings */

    /* Do a horizontal close, followed by an opening so that words with 
       joining characters, reverse video, and pictures will tend to become 
       thick and will be filtered out by the thickness filter, but we 
       want to ignore really thin, isolated things, hence the opening */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrOr, hStrelB, &pPixrA));
    ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrA, hStrelC, &pPixrB));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrB, hStrelA, &pPixrA));

    ERR_PASS_CLEANUP (w_filterVerticalRuns (pIPWState, runFilterState,
                                            pPixrA, 1, 0,
					    minThickness, maxThicknessThin,
					    cPassRuns, &pPixrA));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrAnd, AND, pPixrA));

    /* filter and union to pPixrC */
    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, runFilterState,
                                              pPixrA, 1, 0,
					      minRunLengthThin, maxLinearSize,
					      cPassRuns, &pPixrC));

    /* accept very short horizontal rulings, as long as they have a sufficient
       duration of white space underneath */

    /* invert pPixrOr */
    ERR_PASS_CLEANUP (w_invert (pIPWState, pPixrOr, &pPixrB));

    /* shorten runs symmetrically by 1/4" */
    ERR_PASS_CLEANUP (w_erodeHorizontalRuns( pIPWState, runFilterState,pPixrB,
					     0,xres/4,xres/8,&pPixrB));

    /* dilate  up by 1/20 of an inch */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1, MAX(1,yres/20), 0,
    MAX(0,yres/20-1), &vStrelA));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrB, vStrelA, &pPixrB));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &vStrelA));

    /* AND with pPixrA */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrA, AND, pPixrB));
    
    /* filter horizontal runs of duration < minRunLengthThinAndClean, union 
       the result to pPixrD */
    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, runFilterState,
                                              pPixrB, 1, 0,
					      minRunLengthThinAndClean, maxLinearSize,
					      cPassRuns, &pPixrB));

    /* seed pPixrA with pPixrB, then filter again, unioning to C */
    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, pPixrB, pPixrA, 3, &pPixrD));
    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, runFilterState,
                                              pPixrD, 1, 0,
					      minRunLengthThinAndClean, maxLinearSize,
					      cPassRuns, &pPixrC));

    if(dotted)
    {
       /* 
          Special treatment to recover dotted rulings without hallucinating 
          them in text. Only accept dotted rulings sandwiched between long
          white runs.
       */

       /* invert pPixrOr */
       ERR_PASS_CLEANUP (w_invert (pIPWState, pPixrOr, &pPixrA));

       /* shorten runs symmetrically by minRunLengthDotted */
       ERR_PASS_CLEANUP (w_erodeHorizontalRuns( pIPWState, runFilterState,pPixrA,
					     0,minRunLengthDotted,
                                             minRunLengthDotted/2,&pPixrA));

       /* vertical close by maxThicknessDotted */
       ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1, maxThicknessDotted, 0,
       maxThicknessDotted/2, &vStrelA));
       ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrA, vStrelA, &pPixrB));
       ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrB, vStrelA, &pPixrD));
       ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &vStrelA));

       /* close gaps */ 
       ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrOr, hStrelDot, &pPixrB));
       ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrB, hStrelDot, &pPixrA));

       /* shorten runs symmetrically by minRunLengthDotted */
       ERR_PASS_CLEANUP (w_erodeHorizontalRuns( pIPWState, runFilterState,pPixrA,
					     0,minRunLengthDotted,
                                             minRunLengthDotted/2,&pPixrA));

       /* AND with pPixrD */
       ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrD, AND, pPixrA));

       /* dilate symmetrically by minRunLengthDotted, union to C  */
       ERR_PASS_CLEANUP (w_dilateHorizontalRuns( pIPWState, runFilterState,pPixrA,
					     minRunLengthDotted,
                                             minRunLengthDotted/2,&pPixrC));

    }


    ERR_PASS_CLEANUP (w_clear (pIPWState, pPixrA));
    packet.pPixr = (void *)pPixrA;
    packet.minSize = minLinearSize;
    packet.pFF = vRuleContainer.array;
    packet.nFF = vRuleContainer.count;
    packet.pIndex=index_array;
    packet.lastIndex=hSrc-1;
    packet.ySearchRadius=4;
    packet.maxDistX=MAX(1,xres/16);

    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, pPixrC, &pH));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, pPixrC, &bplH));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pPixrC, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pPixrC, &h));

    /* Run CC-BB and eliminate non-rules based on bounding box widths */
    ERR_PASS_CLEANUP(
	i_findConnectedComponentBoundingBoxes (pIPWState, pH, bplH, 
                                               cFrameBits, cFrameBits, w, h,
					       1, 1,
					       1, (void *)(&packet),
					       wp_filterHRuleCCs, &ret));

    /* remove the vertical rulings data */
    IPW_FREE(vRuleContainer.array,cAllocatorData);
    IPW_FREE(index_array,cAllocatorData);
    
    /* grow the seeds of the rules */
    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, pPixrA, pPixrC, 3, &pPixrD));



    /* we're done with the run filtering */
    ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));


    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrOr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrAnd));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrB));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrC));


    /* expand back to input resolution */
    ERR_PASS (w_expandCols (pIPWState, pPixrD, xfactor, ppPixrDest));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrD));


    NORMAL_RETURN;
    
ERR_CLEANUP:

    if (pPixrA) w_destroyPixr (pIPWState, &pPixrA);
    if (pPixrB) w_destroyPixr (pIPWState, &pPixrB);
    if (pPixrC) w_destroyPixr (pIPWState, &pPixrC);
    if (pPixrD) w_destroyPixr (pIPWState, &pPixrD);

    if (pPixrAnd) w_destroyPixr (pIPWState, &pPixrAnd);
    if (pPixrOr) w_destroyPixr (pIPWState, &pPixrOr);

    if (vStrelA) w_destroyStrel (pIPWState, &vStrelA);

    ERR_RETURN;
}  /* w_findFormFieldRules () */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Vectorizes vertical rules in source pixr, puts them in a FormFieldContainer. 
 *
 * LongDescription:::
 *  Vectorizes vertical rules in source pixr, puts them in a FormFieldContainer. 
 *
 *
 * Input:::
 *  pIPWState::      State structure
 *  pPixrS::         Input PIXR
 *
 * Output:::
 *  *pContainer::    fills uninitialized FormFieldContainer that the caller 
 *                   passes a ref to. 
 *
 * Returns:::
 *  error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis 
 *
 * Citations:::
 *  based on w_removeHVRules by dferrell
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_vectorizeVerticalRules (
    IPWState           *pIPWState,
    PIXR               *pPixrS,
    FormFieldContainer *pContainer
    )
{
    PROLOGUE;
    Int32     ret, bplH, w, h;
    STREL    *pStrelA = NULL;

    PIXR     *pPixrA=NULL,
	     *pPixrB=NULL,
	     *pPixrC=NULL;

#if 0  /* debug code */
    PIXR     *pPixrD=NULL;
#endif /* debug code */

    FilterRulesCCPacket packet;
    UInt8    *pH = NULL;
    void     *runFilterState;
    Int32     minLinearSize;
    Int32     maxLinearSize;
    Int32     maxThickness;
    Int32     minThickness;

    ScaledInt1616 xres,yres;
    Int32 max_height;
    Int32 yfactor;
    Int32 yShift;

    ERR_GEN_PIXR (pPixrS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS(w_pixrGetResolutionDPI(pIPWState,pPixrS,&xres,&yres) );
    /* scale resolution for unscaled pixel interpretation */
    xres=ScaledInt1616_TO_Int32(xres);
    yres=ScaledInt1616_TO_Int32(yres);

    /* reduce res in y direction */
    if(yres>80)
    {
       yfactor=4;
       yShift=2;
       ERR_PASS (w_reduceRows (pIPWState, pPixrS, yfactor, AND, &pPixrA));
    }
    else
    {
       yfactor=2;
       yShift=1;
       ERR_PASS (w_reduceRows (pIPWState, pPixrS, yfactor, AND, &pPixrA));
    }

    yres >>= yShift;


    ERR_PASS(w_pixrGetHeight(pIPWState,pPixrA,&max_height) );
    minThickness=1;
    maxThickness= MAX(1,xres/12);
    maxLinearSize=max_height;
    minLinearSize= yres/8; 
    

    /** Find the vertical rule-like things **/


    /* get ready for run filtering */
    ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, pPixrA, cBothRuns,
                                       &runFilterState));


    /* Get only things which are the right thickness */
    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, runFilterState,
                                            pPixrA, 1, 0,
					    minThickness, maxThickness, 
					    cPassRuns, &pPixrB));
    ERR_PASS_CLEANUP (w_filterVerticalRuns  (pIPWState, runFilterState,
                                              pPixrA, 1, 0,
					      minLinearSize, maxLinearSize, 
					      cPassRuns, &pPixrC));
    /* Keep long AND thin things */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrB, AND, pPixrC));

    /* close small vertical gaps */
    ERR_PASS_CLEANUP (
                       w_makeBrickStrel ( 
						pIPWState, 
						3, 
						MAX(1,yres/8), 
						1, 
						(yres/16), 
						&pStrelA
                                       )
		     );
    ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrC, pStrelA, &pPixrB));
    ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrB, pStrelA, &pPixrC));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &pStrelA));

    /* AND with original */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrA, AND, pPixrC));

#if 0    /* debug code */
    ERR_PASS_CLEANUP (w_clear (pIPWState, pPixrA));
    packet.pPixr = (void *)pPixrA;
#endif   /* debug code */

    packet.minSize = minLinearSize;
    packet.minAspectNum =   yres*6;
    packet.minAspectDenom = xres;
    packet.yShift= yShift;  
    ERR_PASS_CLEANUP (wp_initFormFieldContainer( pIPWState, pContainer) );
    packet.pCont= pContainer;
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, pPixrC, &pH));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, pPixrC, &bplH));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pPixrC, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pPixrC, &h));

    /* we're done with the run filtering */
    ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    /* Run CC-BB, eliminate non-vertical rules and vectorize the vertical 
       rulings */
    ERR_PASS_CLEANUP(
	i_findConnectedComponentBoundingBoxes (pIPWState, pH, bplH, 
					       cFrameBits, 
                                               cFrameBits, w, h,
					       1, 1,
					       1, (void *)(&packet),
					       wp_filterVRuleCCs, &ret));

#if 0 /* debug code */    
    /* grow the seeds of the rules */
    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, pPixrA, pPixrC, 3, &pPixrD));
#endif /* debug code */


    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrB));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrC));

#if 0  /* debug code */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrD));
#endif /* debug code */


    NORMAL_RETURN;
    
ERR_CLEANUP:

    if (pPixrA) w_destroyPixr (pIPWState, &pPixrA);
    if (pPixrB) w_destroyPixr (pIPWState, &pPixrB);
    if (pPixrC) w_destroyPixr (pIPWState, &pPixrC);

    if (strel1) w_destroyStrel (pIPWState, &pStrelA);

    ERR_RETURN;
}  /* wp_vectorizeVerticalRules () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find CC bounding box nearest to a point.
 *
 * LongDescription:::
 *  Callback to find smallest connected component bounding box that encloses
 *  a point, or if there is no such box, box that is closest to the point.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  packet::         Generic callback struct
 *  x::              x location of this CC.
 *  y::              y location of this CC.
 *  w::              Width of this CC.
 *  h::              Height of this CC.
 *  m::              Mass of this CC.
 *  nRuns::          Number of runs in this CC.
 *  px::             Seed pixel x location
 *  py::             Seed pixel y location
 *
 *
 * Returns:::
 *  error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_findCheckBoxCC (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    CheckBoxParam *pParam;
    CheckBoxPacket *pPack; 
    CheckBoxPacket packWhite; 
    Int32 right,bot,xdist,ydist,dist;
    Int32 xres,yres;
    Int32 shapeWeighting;
    Int32 shapeDist;
    Int32 rectangularity;
    UInt8 *pIm; 
    Int32 bplIm; 
    Int32 ret;
    Bool image_inverted=0;

    /* keep compilers quiet */
    m = m;
    px=px;
    py=py;
    nRuns = nRuns;

    pPack   = (CheckBoxPacket *)packet;
    pParam  = pPack->pParam;

    xres=  pParam->xres ? pParam->xres: 1;
    yres=  pParam->yres ? pParam->yres: 1;
 
    right=x+w-1;
    bot=y+h-1;

    if (
          /* component is interior to box */
          (x>pParam->minX) && 
          (right<pParam->maxX) &&
          (y>pParam->minY) &&
          (bot<pParam->maxY) &&
          /* component is within max size limits */
          (w<pParam->maxW) &&
          (h<pParam->maxH) &&
          /* aspect ratio is within limits */
          ( (w*pParam->yres) < (h*pParam->xres*pParam->maxpect) ) && 
          ( (w*pParam->yres*pParam->maxpect) > (h*pParam->xres) )
       )
    {
       shapeWeighting=pParam->shapeWeight*xres*xres;
       if(pPack->white)
       {
          if(w&&h)
          {
             shapeDist= (shapeWeighting-((shapeWeighting*m)/(w*h)))/100;
             rectangularity=(100*m)/(w*h);
          }
          else
          {
             shapeDist=rectangularity=0;
          }
       }
       else
       {
          /* if black component, we have to find largest white component
          within, since it is the white we are testing for rectangularity,
          also, we will use the width and height of the black component
          bounding box since we expect a thin outline */
          /* invert rectangle */
          ERR_PASS_CLEANUP( w_rasterop(
		pIPWState,
		x-cFrameBits,
	  	y-cFrameBits,
	  	pParam->pPixrS,
          	PIX_NOT(PIX_SRC),
	  	x-cFrameBits,
	  	y-cFrameBits,
	  	w,
	  	h,
	  	pParam->pPixrS
          ));
          image_inverted=1;
          /* find most massive cc */
          packWhite.m=0;
          ERR_PASS_CLEANUP(w_pixrGetImageAll (pIPWState, 
	  pParam->pPixrS, &pIm));
          ERR_PASS_CLEANUP(w_pixrGetBpl (
                pIPWState, pParam->pPixrS, &bplIm));
          i_findConnectedComponentBoundingBoxes (
				pIPWState, 
				pIm, 
				bplIm, 
				x, 
				y, 
				w, 
				h,
			       	0, 
				0,
			       	1, 
				(void *)(&packWhite),
				wp_findMostMassiveCC, 
				&ret
          );
          image_inverted=0;
          /* reinvert rectangle */ 
          ERR_PASS_CLEANUP(w_rasterop(
				pIPWState,
	 			x-cFrameBits,
				y-cFrameBits,
				pParam->pPixrS,
				PIX_NOT(PIX_SRC),
				x-cFrameBits,
				y-cFrameBits,
				w,
				h,
				pParam->pPixrS
	  ));

          /* use "outer" bounding box width and height */
          m=packWhite.m;
          if(m)
          {
             w=packWhite.w;
             h=packWhite.h;
             x=packWhite.x;
             y=packWhite.y; 
          }
          if(m&&w&&h)
          {
             shapeDist= (shapeWeighting-((shapeWeighting*m)/(w*h)))/100;
             rectangularity=100*m/(w*h);
          }
          else
          {
             shapeDist=rectangularity=0;
          }
          if ( 
                (m==0) ||
                (w<pParam->minW) ||
                (h<pParam->minH) || 
                ( (w*pParam->yres) > (h*pParam->xres*pParam->maxpect) ) || 
                ( (w*pParam->yres*pParam->maxpect) < (h*pParam->xres) )
          )
             NORMAL_RETURN;
       } 



       /* pick is to the left of the cc bounding box */
       if ( pParam->x<x )
       {
          xdist= (x-pParam->x);
          xdist *= xdist;
       }
       /* pick is to the right of the c.c. bounding box */
       else if ( pParam->x>right )
       {
          xdist= (pParam->x-right);
          xdist *= xdist;
       }
       else /* pick is inside the x range of the c.c. bounding box */
       {
          xdist=0;
       }

       /* pick is above the c.c. bounding box */
       if ( pParam->y<y )
       {
          ydist= (y-pParam->y);
          ydist= (ydist*xres)/yres;
          ydist *= ydist;
       }
       /* pick is below the c.c. bounding box */
       else if ( pParam->y>bot )
       {
          ydist= (pParam->y-bot);
          ydist= (ydist*xres)/yres;
          ydist *= ydist;
       }
       /* pick is inside the y range of the c.c. bounding box */
       else
       {
          ydist=0;
       }

       dist=xdist+ydist;


       /* only weigh in shape distance if we are not dead on */
       if(dist)
          dist +=shapeDist;

       if( (dist>0) && (rectangularity<pParam->minRectangularity) )
          NORMAL_RETURN;

       if( 
           /* closer than current min */
           (dist<pPack->dist)
           /* 
             If distances are the same, pick the bounding box 
             with smallest area. 
           */
           || ( 
                 (dist==pPack->dist)&&
                 ( (w*h) < (pPack->w*pPack->h) )
              )
         )
       {
          pPack->x=x;
          pPack->y=y;
          pPack->w=w;
          pPack->h=h;
          pPack->dist=dist;
          pPack->rectangularity=rectangularity;
       }
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    if(image_inverted)
    {
       /* reinvert rectangle */ 
       w_rasterop(
                 pIPWState,
                 x-cFrameBits,
                 y-cFrameBits,
                 pParam->pPixrS,
                 PIX_NOT(PIX_SRC),
                 x-cFrameBits,
                 y-cFrameBits,
                 w,
                 h,
                 pParam->pPixrS
       );
    }

    ERR_RETURN;


}  /*  wp_findCheckBoxCC() */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find the connected component with the greatest number of pixels.
 *
 * Input:::
 *  pIPWState::	     Pointer to IPCore state
 *  packet::         Generic callback struct
 *  x::              x location of this CC.
 *  y::              y location of this CC.
 *  w::              Width of this CC.
 *  h::              Height of this CC.
 *  m::              Mass of this CC.
 *  nRuns::          Number of runs in this CC.
 *  px::             Seed pixel x location
 *  py::             Seed pixel y location
 *
 * Returns:::
 *  cSuccessful if successfully seeded CC 
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis 
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_findMostMassiveCC (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    CheckBoxPacket *pPack; 

    /* keep compilers quiet */
    pIPWState = pIPWState;
    nRuns = nRuns;
    px=px;
    py=py;

    pPack   = (CheckBoxPacket *)packet;


    if (m >= pPack->m)
    { 
       pPack->m=m;
       pPack->w=w;
       pPack->h=h;
       pPack->x=x;
       pPack->y=y;
    }

    NORMAL_RETURN;

}  /* wp_findMostMassiveCC () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Seed only "horizontal rule-like" connected components.
 *
 * LongDescription:::
 *  Seed only "rule-like" connected components. This is a callback to
 *  i_findConnectedComponentBoundingBoxes,
 *
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  packet::         Generic callback struct
 *  x::              x location of this CC.
 *  y::              y location of this CC.
 *  w::              Width of this CC.
 *  h::              Height of this CC.
 *  m::              Mass of this CC.
 *  nRuns::          Number of runs in this CC.
 *  px::             Seed pixel x location
 *  py::             Seed pixel y location
 *
 *
 * Returns:::
 *  cSuccessful if successfully seeded CC 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *   danis 
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_filterHRuleCCs (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    FilterRulesCCPacket *pack;
    Int32 potential_top,maxy,i,index,distX,maxDistX,xright;
    FormField *pFF; 

    /* keep compilers quiet */
    x = x;
    y = y;
    m = m;
    nRuns = nRuns;

    pack = (FilterRulesCCPacket *)packet;

    x=  x-cFrameBits;
    xright= x+w;
    y=  y-cFrameBits;

    if ( w >= pack->minSize )
    {
        /* verify that this is not a top in that the left and right
           endpoints do not occur near the top end points of any
           vertical rulings */

        potential_top=0;
        maxy= MIN(y+h+pack->ySearchRadius,pack->lastIndex);

        for(i=MAX(0,y-pack->ySearchRadius); i<maxy;++i)
        {
           if ((potential_top=pack->pIndex[i]))
              break;
        }

        /* 
           If any vertical rulings are on both the left and the right, this is
           probably not a field, but a box top, so dismiss.
        */
        maxDistX=pack->maxDistX;
        /* search for vertical ruling on left  */
        if(potential_top)
        {
           /* adjust offset, initially incremented by 1, since 0 was 
              used to indicate an empty cell  */
           index=potential_top-1;
           /* point to vertical ruling at that y location */
           pFF= pack->pFF+index;
           distX= maxDistX+1;
           while( (index<pack->nFF)&&(distX>maxDistX) && 
                  (pFF->ybase<= maxy) )
           {
              distX=MIN(ABS(x-pFF->xleft),ABS(x-(Int32)FF_RIGHTX_IN_WIDTH(pFF) ) );
              pFF++; 
              index++;
           }
           if(distX>maxDistX)
              potential_top=0; 
        }
        /* search for vertical ruling on right */
        if(potential_top)
        {
           /* adjust offset, initially incremented by 1, since 0 was 
              used to indicate an empty cell  */
           index=potential_top-1;
           /* point to vertical ruling at that y location */
           pFF= pack->pFF+index;
           distX= maxDistX+1;
           while( (index<pack->nFF)&&(distX>maxDistX) && 
                (pFF->ybase<= maxy) )
           {
              distX=MIN(ABS(xright-pFF->xleft),ABS(xright-
              (Int32)FF_RIGHTX_IN_WIDTH(pFF) ));
              pFF++;
              index++;
           }
           if(distX>maxDistX)
              potential_top=0; 
        }
        if(!potential_top)
        {
            ERR_PASS (w_pixrPutPixel (pIPWState, (PIXR *)(pack->pPixr),
	   	      px - cFrameBits, py - cFrameBits, 1));
        }
    }

    NORMAL_RETURN;

}  /* wp_filterHRuleCCs () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Vectorize only "vertical-rule-like" connected components.
 *
 * LongDescription:::
 *  Vectorize only "rule-like" connected components. This is a callback to
 *  i_findConnectedComponentBoundingBoxes,
 *
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  packet::         Generic callback struct
 *  x::              x location of this CC.
 *  y::              y location of this CC.
 *  w::              Width of this CC.
 *  h::              Height of this CC.
 *  m::              Mass of this CC.
 *  nRuns::          Number of runs in this CC.
 *  px::             Seed pixel x location
 *  py::             Seed pixel y location
 *
 *
 * Returns:::
 *  cSuccessful if successfully seeded CC 
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis 
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_filterVRuleCCs (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    FilterRulesCCPacket *pack;
    FormField ff;


    /* keep compilers quiet */
    m = m;
    nRuns = nRuns;
    px=px;
    py=py;

    pack = (FilterRulesCCPacket *)packet;

    if (
          ( h >= pack->minSize ) &&
          ( (h*pack->minAspectDenom) > (w*pack->minAspectNum) ) 
       )
    { 

#if 0  /* debug code */
        ERR_PASS (w_pixrPutPixel (pIPWState, pack->pPixr,
		  px - cFrameBits, py - cFrameBits, 1));
#endif /* debug code */ 

        ff.ybase=(y-cFrameBits)<<pack->yShift;
        ff.xleft=x-cFrameBits;
        ff.width=w;
        ff.maxh=h<<pack->yShift;
        ERR_PASS ( 
                    wp_appendToFormFieldContainer(
                    pIPWState,
                    &ff,
                    pack->pCont)
                 );
    }

    NORMAL_RETURN;

}  /* wp_filterVRuleCCs () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  reduce by factors of  2 to get resolution closest to desired resolution.
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pPixrS::          source pixr 
 *  xresD::           desired x resolution 
 *  yresD::           desired y resolution 
 *
 * Output:::
 *
 *  pRedx::	      reduction factor by which the input was reduced to the
 *                    output in the x direction (as a power of 2). 
 * 
 *  pRedy::           reduction factor by which the input was reduced to the
 *   		      output in the y direction (as a power of 2).
 *
 *  ppPixrD::         pointer to Pixr created by this routine. If *ppPixrD==
 *                    pPixrS, then the source Pixr is destroyed at the end of 
 *                    this routine, assume the intention was to replace the 
 *                    input by the output. 
 *
 * Returns:::
 *  error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_reduceToRes(
    		IPWState *pIPWState,
		PIXR *pPixrS,
		Int32 xresD, 
		Int32 yresD,
		Int32 *pRedx,
		Int32 *pRedy,
		PIXR **ppPixrD
	     )
{
    PROLOGUE;
    ScaledInt1616 xresS,yresS,xres,yres;
    Int32 ired,iredx,iredy,redf,niter16;
    Int32 width,height;
    PIXR *pPixrA=NULL,*pPixrB=NULL;
    Int32 i;

    ERR_PASS(w_pixrGetWidth(pIPWState,pPixrS,&width) );

    ERR_PASS(w_pixrGetHeight(pIPWState,pPixrS,&height) );

    ERR_PASS(w_pixrGetResolutionDPI(pIPWState,pPixrS,&xres,&yres) );
    /* scale resolution for unscaled interpretation */
    xresS=ScaledInt1616_TO_Int32(xres);
    yresS=ScaledInt1616_TO_Int32(yres);
 
    iredx=iredy=0; 
    while( ABS(xresS/2-xresD)<ABS(xresS-xresD) )
    {
       xresS>>=1;
       iredx++;
    }

    while( ABS(yresS/2-yresD)<ABS(yresS-yresD) )
    {
       yresS>>=1;
       iredy++;
    }

    if(iredx>iredy)
    {
       ired= iredy;
       iredx -= ired; 
       iredy=0;
    }
    else if (iredx<iredy)
    {
       ired=iredx;
       iredy -= ired;
       iredx=0;
    }
    else
    {
       ired=iredx;
       iredx=0;
       iredy=0;
    }

    niter16=ired/4;

    ired -= 4*niter16;

    pPixrA=pPixrS;
    for(i=0;i<niter16;++i)
    {
       xres/=16;
       yres/=16;
       width/=16;
       height/=16;
       ERR_PASS_CLEANUP(w_createPixr(pIPWState,width,height,1,xres,yres,
       cFrameBits,&pPixrB));
       ERR_PASS_CLEANUP( w_reduce(pIPWState,pPixrA,1,1,1,1,&pPixrB) );
       if(pPixrA != pPixrS)
          ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
       pPixrA=pPixrB;
       pPixrB=NULL;
    }

    switch (ired) {
	case 1:
           xres/=2;
           yres/=2;
           width/=2;
           height/=2;
           ERR_PASS_CLEANUP(w_createPixr(pIPWState,width,height,1,xres,yres,
           cFrameBits,&pPixrB));
           ERR_PASS_CLEANUP( w_reduce2(pIPWState,pPixrA,1,&pPixrB) );
           if(pPixrA != pPixrS)
               ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
           pPixrA=pPixrB;
           break;
	case 2:
           xres/=4;
           yres/=4;
           width/=4;
           height/=4;
           ERR_PASS_CLEANUP(w_createPixr(pIPWState,width,height,1,xres,yres,
           cFrameBits,&pPixrB));
           ERR_PASS_CLEANUP( w_reduce4(pIPWState,pPixrA,1,1,&pPixrB) );
           if(pPixrA != pPixrS)
               ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
           pPixrA=pPixrB;
           break;
	case 3:
           xres/=8;
           yres/=8;
           width/=8;
           height/=8;
           ERR_PASS_CLEANUP(w_createPixr(pIPWState,width,height,1,xres,yres,
           cFrameBits,&pPixrB));
           ERR_PASS_CLEANUP( w_reduce8(pIPWState,pPixrA,1,1,1,&pPixrB) );
           if(pPixrA != pPixrS)
               ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
           pPixrA=pPixrB;
           break;
    };

    pPixrB=NULL;

    if(iredx)
    {
       redf=1<<iredx;
       width/=redf;
       xres/=redf;
       ERR_PASS_CLEANUP(w_createPixr(pIPWState,width,height,1,xres,yres,
       cFrameBits,&pPixrB));
       ERR_PASS_CLEANUP( w_reduceCols(pIPWState,pPixrA,redf,UNION,&pPixrB) );
       if(pPixrA != pPixrS)
          ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
       pPixrA=pPixrB;
    };

    if (iredy) 
    {
       redf=1<<iredy;
       height/=redf;
       yres/=redf;
       ERR_PASS_CLEANUP(w_createPixr(pIPWState,width,height,1,xres,yres,
       cFrameBits,&pPixrB));
       ERR_PASS_CLEANUP( w_reduceRows(pIPWState,pPixrA,redf,UNION,&pPixrB) );
       if(pPixrA != pPixrS)
          ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
       pPixrA=pPixrB;
    };

    if(pPixrA==pPixrS)
    {
       /* no reduction, copy source to destination  */
       if(*ppPixrD==NULL)
       {
          ERR_PASS_CLEANUP(w_createPixr(pIPWState,width,height,1,xres,yres,
          cFrameBits,ppPixrD));
          ERR_PASS_CLEANUP (w_copy (pIPWState, pPixrS,ppPixrD));
       }
    }
    else if(pPixrS==*ppPixrD)
    {
       ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrS));
       *ppPixrD=pPixrA;
    }
    else
    {
       *ppPixrD=pPixrA;
    }
    
    *pRedx=ired+iredx;
    *pRedy=ired+iredy;
    NORMAL_RETURN;
    
ERR_CLEANUP:

    if (pPixrA && (pPixrA != pPixrS) ) w_destroyPixr (pIPWState, &pPixrA);
    if (pPixrB) w_destroyPixr (pIPWState, &pPixrB);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  vectorize PIXR form field feature map. 
 *
 * LongDescription:::
 *  vectorize PIXR form field feature map. putting each form field into  
 *  the form field container which this routine creates.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pPixrS::          source pixr 
 *  pPixrW::          pixr for white space analysis, rotated by 270 
 *  scalex::          x scaling factor to apply to go to full x resolution
 *  scaley::          y scaling factor to apply to go to full y resolution
 *
 * Output:::
 *
 *  pFormFieldContainer::	
 *		      initializes container of form fields, each form field is a  
 *                    vector description of each form field found in the image.
 *
 * Returns:::
 *  error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_vectorizeFields (
    			IPWState *pIPWState,
			PIXR     *pPixrS,
			PIXR     *pPixrW,
			Int32     scalex,
			Int32     scaley,
			FormFieldContainer *pFormFieldContainer
		)
{
   PROLOGUE;

   ScaledInt1616 xres,yres;
   Int32 minWidth;
   StateReceiveFormFields aux;
   UInt8    *pIm = NULL;
   Int32 bplIm,wIm,hIm;
   Int32     ret;

   aux.scalex=scalex;
   aux.scaley=scaley;
   aux.pCont= pFormFieldContainer;
   aux.pPixrWhite= pPixrW;
   

   ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, pPixrS, &pIm));
   ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, pPixrS, &bplIm));
   ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pPixrS, &wIm));
   ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pPixrS, &hIm));

   /* find the resolution of the PIXR */
   ERR_PASS(w_pixrGetResolutionDPI(pIPWState,pPixrS,&xres,&yres) );
   /* scale resolution for unscaled interpretation */
   xres=ScaledInt1616_TO_Int32(xres);
   minWidth=xres/10;

   /* Run CC-BB to vectorize the field map */
   ERR_PASS_CLEANUP( i_findConnectedComponentRuns (pIPWState, 
						   pIm, bplIm, cFrameBits,
						   cFrameBits, wIm, hIm,
						   minWidth, 1,
						   1, (void *)(&aux),
						   wp_receiveFormFields, &ret));

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  call back to i_findConnectedComponentBoundingBoxes to save form fields.
 *
 * LongDescription:::
 *  receive connected compnent bounding boxes from 
 *  i_findConnectedComponentBoundingBoxes and drop them into a FormField
 *  container.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  pAux::           Generic callback struct
 *  x::              x location of this CC.
 *  y::              y location of this CC.
 *  w::              Width of this CC.
 *  h::              Height of this CC.
 *  m::              Mass of this CC.
 *  nRuns::          Number of runs in this CC.
 *  pRuns::          pointer to the runs.
 *
 *
 * Returns:::
 *   error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_receiveFormFields (
    IPWState	       *pIPWState,
    void               *pAux,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    TCScanLineRun       *pRuns 
    )
{
    PROLOGUE;

    StateReceiveFormFields *pState;
    FormField ff;
 
    Int32 sum_bottom,av_bottom,count,max,irun,l,r;
    Int32 yline,left,right;

    Int32 *bot_array;
    Int32 i;
    TCScanLineRun       *pRun; 

    Int32 imH,bpl;
    Int32 fieldTop, scaledTop;
    UInt8 *pLine,*pImage;
    Int32 xoff;

    /* keep compilers quiet */
    m = m;
    nRuns = nRuns;

    pState = (StateReceiveFormFields *)pAux;


    /* 
       Find the y coordinate by averaging the bottom y
       coordinates across the width of the connected
       component.
    */ 

    bot_array= (Int32 *) IPW_MALLOC(w*sizeof(Int32),cAllocatorData);
    sum_bottom=0;
    count=0;
    memset(bot_array,0,w*sizeof(Int32));
    pRun=pRuns;
    max=w-1;

    yline=y+h-1;
    irun=nRuns;

    /* to avoid line increment first time through loop */ 
    left= pRun->runRight+2;

    /* find the bottom contour of the c.c. */
    while ( (count<w)&&(irun>0)&&(pRun!=NULL) )
    { 
        right=pRun->runRight;
        if(right >= (left-1) )
        {  
            /* new line */
            yline--;
        }
        left=pRun->runLeft;

        l=MIN(max,left-x);
        r=MIN(max,right-x); 
        for(i=l;i<=r;++i)
        {
           if(!bot_array[i])
           {
              bot_array[i]= yline;
              sum_bottom += yline;
              ++count;
           }
        }
        irun--;
        pRun = pRun->pNext;
    }      

    
    /* find minimum white height above the bottom contour */
    ERR_PASS(w_pixrGetImage(pIPWState,pState->pPixrWhite,&pImage));
    ERR_PASS(w_pixrGetHeight(pIPWState,pState->pPixrWhite,&imH));
    ERR_PASS(w_pixrGetBpl(pIPWState,pState->pPixrWhite,&bpl));

    xoff=MIN(5,w/3); 
    pLine= pImage+( imH-1-(x-cFrameBits)-xoff )*bpl;
    fieldTop= 0; 
    for(i= xoff ;i< (w-xoff) ;++i)
    {
       i_nextBitToTheLeft (pIPWState, pLine, fieldTop,
			   (UInt32)(bot_array[i]-cFrameBits),
                           &fieldTop);
       pLine -= bpl;
    }

    IPW_FREE(bot_array,cAllocatorData);

    av_bottom=( sum_bottom+(w/2) )/w;
    ff.ybase= pState->scaley*(av_bottom+YCORRECTION-cFrameBits);

    scaledTop= pState->scaley*(fieldTop+1);

    /* zero-height field can happen at top of page */
    if(scaledTop>=ff.ybase)
       NORMAL_RETURN;
    

    ff.maxh= ff.ybase-scaledTop;
    ff.xleft= pState->scalex*(x-cFrameBits);
    ff.width= pState->scalex*w;
    ff.line_id=0;
    ff.string_id=0;

    ERR_PASS (wp_appendToFormFieldContainer(pIPWState, &ff,pState->pCont));

    NORMAL_RETURN;

}  /* wp_receiveFormFields () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  sort the form fields, generally in raster order. 
 *
 * LongDescription:::
 *  sort the form fields, generally in raster order. but with allowance for   
 *  jitter, so it is not exactly raster order in that the y coordinate is 
 *  smeared a bit.
 *
 * PrivateDescription:::
 *  Big weakness is that form fields are separated from the rulings 
 *  they were found on top of and there is no information in a form field
 *  as to which ruling it belongs in. So if the ruling on which some form
 *  fields  were found is skewed, the fields may take on y coordinates such 
 *  that they end up on different lines after the y-clustering done in this 
 *  function, which is ugly, but seems to almost never happen is the image 
 *  is deskewed.
 *
 *  Will get back to this if it ever becomes a problem. Right now the 
 *  code is not structured so that linking a formfield to its ruling
 *  can be easily done.
 *   
 *
 * Input:::
 *  pIPWState::      State structure
 *  xres::          x res in DPI as a ordinary integer
 *  yres::          y res in DPI as an ordinary integer 
 *  pCont::	    container of form fields, each form field is a  vector 
 *
 *
 * Returns:::
 *   error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_sortFields(
   IPWState *pIPWState,
   Int32 xres,
   Int32 yres,
   FormFieldContainer *pCont)
{
   PROLOGUE;

   FormField *fieldArray;
   Int32 nFields;
   Int32 max_deltay;
   Int32 max_deltay_slope_check;
   Int32 xcenter,xcenter_new;
   Int32 i,j,k,l,iline;
   Int32 xleft,xright,y,xleftk,xrightk;
   Int32 deltax, deltay;
   Bool term_line;
   Int32 numLines;

   /* keep compiler happy, we may use this param in the future */
   pIPWState=pIPWState;

   nFields=pCont->count;

   if(nFields==0)
      NORMAL_RETURN;
   
   fieldArray= pCont->array;

   /* sort fields in place by y coordinate */
   qsort(fieldArray,nFields,sizeof(FormField),compare_by_y);

   /* break lines based on delta y */
   max_deltay = NUM_DELTA_Y_BREAK*yres;
   max_deltay_slope_check  = NUM_DELTA_Y_SLOPE_CHECK*yres;

   /* initial clustering by y coordinate */
   i=0;
   iline=0;
   while(i<nFields)
   {
      fieldArray[i].line_id=(Int16)iline;
      term_line=0;
      y=fieldArray[i].ybase;
      xleft=fieldArray[i].xleft;
      xright=xleft+fieldArray[i].width-1;
      xcenter = (xleft+xright)/2; 

      for( j=i+1;j<nFields;++j)
      {
         /* keep adding to line cluster until y out of range or 
            until x overlap occurs */ 
         xleft=fieldArray[j].xleft;
         xright=xleft+fieldArray[j].width-1;

         deltay=ABS(y-fieldArray[j].ybase);
         xcenter_new = (xleft+xright)/2; 
         deltax=ABS(xcenter-xcenter_new); 

         /* update */
         xcenter = xcenter_new; 
         y=fieldArray[j].ybase;

         /* y out of range */
         if( (DENOM_DELTA_Y_BREAK*deltay) >max_deltay )
            break;

         if( (DENOM_DELTA_Y_SLOPE_CHECK*deltay) > max_deltay_slope_check)
         {
            /* are fields spaced far enough apart w.r.t. x so that skew could
               explain delta y ? */ 
            if( (deltay*MAX_SLOPE_DENOM*xres)>(deltax*MAX_SLOPE_NUM*yres) )
               break; 
         }


         for(k=i;k<j;++k)
         {
            xleftk=  fieldArray[k].xleft;
            xrightk= xleftk+fieldArray[k].width-1;
            if(  
                /* x overlap */
                ( (xleft<xrightk)&&(xright>xleftk) )
              )
            {
               term_line=1;
               /* 
                  Back end of line up to field  where y coordinate 
                  is closer to predecessor than it is to the field  
                  that caused the line break. 
               */
               l=i+1;
               while( 
                       (l<j) && 
                       (
                          ABS(fieldArray[l].ybase-fieldArray[l-1].ybase)
                          < 
                          ABS(fieldArray[l].ybase-fieldArray[j].ybase)
                       )
                    )
               {
                  ++l;
               }
               j=l;
               /* end the line */
               break;
            }
         } 
         if(term_line)
         {
            break;
         }
         else
         {
            /* add j to current line cluster */
            fieldArray[j].line_id=(Int16)iline;
         }
      }
      iline++;
      i=j;
   }
 
   numLines=iline;

   /* sort by line cluster ID (which is already taken care of) then x  */
   qsort(fieldArray,nFields,sizeof(FormField),compare_in_line_by_left_x);


   NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Group fields believed to be character fields in the same string into 
 *  the same string by assigning them the same string ID.
 *
 * LongDescription:::
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::     State structure
 *  xres::          x res in DPI as a ordinary integer
 *  yres::          y res in DPI as an ordinary integer 
 *  pCont::	    container of form fields
 *
 *
 * Returns:::
 *   error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_groupFields(
    IPWState *pIPWState,
    Int32 xres,
    Int32 yres,
    FormFieldContainer *pCont)
{
   PROLOGUE;

   Int32 i,j;
   UInt16 id;
   Int32 nFields;
   Int32 width,strlen,sum_widths,rightx,line;
   Int32 deltax;
   FormField *pFF;
   Int32 last_string_id;
   Int32 last_line_id; 
   Int32 last_width; 

   /* keep compiler happy, we may use this param in the future */
   pIPWState=pIPWState;
   yres=yres;

   nFields=pCont->count;
   pFF= pCont->array;

   id=1;
   i=0;

   if(nFields==0)
      NORMAL_RETURN;
   
   while(i<nFields-1)
   {
      width=pFF[i].width;
      strlen=1;
      sum_widths=0;
      /* is field width small enough to be a character field ? */
      if( (MAX_CF_WIDTH_DENOM*width) < (MAX_CF_WIDTH_NUM*xres) )
      {
         sum_widths=width;
         j=i+1;
         line= pFF[i].line_id;
         rightx=pFF[i].xleft+width-1;
         while ( j<nFields )
         {
            deltax=pFF[j].xleft-rightx;
            rightx=pFF[j].xleft+pFF[j].width-1;
            width=pFF[j].width;
            if (
                 ( (MAX_CF_WIDTH_DENOM*width) < (MAX_CF_WIDTH_NUM*xres ) ) &&
                 (pFF[j].line_id==line) && 
                 (deltax>=0) && 
                 ( (10*deltax) <xres) && 
                 (
                    /* field is just so small, it probably is a character field */
                    ( ( (5*sum_widths)< (strlen*xres) ) && ( (5*width)<xres) )||
                    /* field width is in line with other fields, also weight by
                       the number of fields in the string  */
                    ( ( (3-MIN(2,strlen/5) )*ABS(strlen*width-sum_widths) )< 
                    sum_widths)
                 )
               )
            {
               sum_widths += width;
               strlen++;
               ++j;
            }
            else
            {
               break;
            }
         }
      }
      if( 
          /* 3 or more fields, average width < 1/6 of an inch */
          ( (strlen>2) && ( (6*sum_widths) < (xres*strlen) ) ) ||
          /* 4 or more fields */
          ( strlen>3 )
        )
      {
         for(j=i;j<(i+strlen);++j)
         {
            pFF[j].string_id=id;
         }
         i += strlen;
         ++id;
         /* if we ever encounter the ridiculous scenario that we 
            exceed 0x07fff strings, wrap around */
         if(id&(0x8000))
            id=1;
      }
      else if (strlen>1)
      /*  
       *  if string only has two or 3 fields, fill in top bit,
       *  marking it as undetermined, we'll check later to see if there
       *  are other strings with longer lengths surrounding it. 
       */ 
      {
         /* mark as undetermined */
         for(j=i;j<(i+strlen);++j)
         {
            pFF[j].string_id = (id|(0x8000));
         }
         i += strlen;
         ++id;

         /* if we ever encounter the ridiculous scenario that we 
            exceed 0x07fff strings, wrap around */
         if(id&(0x8000))
            id=0;
      }
      else
      {
         pFF[i].string_id=0;
         ++i;
      }
   }

   /* 
      Go forwards and backwards to resolve undetermined strings 
      that have lengths shorter than 4 fields, if there are longer 
      length strings surrounding them, label them as strings by 
      going back to a positive string id. 
   */

   /* iterate forwards (left context) to resolve undetermined strings */
   last_string_id=0;
   last_width=0;
   last_line_id=0;
   for(i=0;i<nFields;++i)
   {
      if( 
           (pFF[i].string_id&0x8000) && 
           (
             (
               /* last field on the same line was part of a string */
               ( (last_string_id>0) && !(last_string_id&0x8000) )  &&
               (pFF[i].line_id==last_line_id) &&
               /* smaller width than last, or approximately the same */
               ( ( ((Int32)pFF[i].width)-last_width)< (xres/10) )
             ) ||
             /* this is the next field in an undetermined string that
                that has already been accepted */
             ( (pFF[i].string_id&0x7fff) == last_string_id)
           )
        )
      {
         pFF[i].string_id= pFF[i].string_id&0x7fff;
      }
      last_string_id= pFF[i].string_id;
      last_line_id= pFF[i].line_id; 
      last_width= pFF[i].width; 
   }
   /* iterate backwards (right context) to resolve undetermined strings */
   last_string_id=0;
   last_width=0;
   last_line_id=0;
   for(i=nFields-1;i>=0;--i)
   {
      if( 
           (pFF[i].string_id&0x8000) && 
           (
             /* last field on the same line was part of a string */
             (
               ( (last_string_id>0) && !(last_string_id&0x8000) )  &&
               (pFF[i].line_id==last_line_id) &&
               /* smaller width than last, or approximately the same */
               ( ( ((Int32)pFF[i].width)-last_width)< (xres/10) )
             ) ||
             /* this is the next field in an undetermined string that
                that has already been accepted */
             ( (pFF[i].string_id&0x7fff) == last_string_id)
           )
        )
      {
         pFF[i].string_id= pFF[i].string_id&0x7fff;
      }
      last_string_id= pFF[i].string_id;
      last_line_id= pFF[i].line_id; 
      last_width= pFF[i].width; 
   }

   /* if an ID is still has a one in the MSb, label as non string */
   for(i=0;i<nFields;++i)
   {
      if (pFF[i].string_id&0x8000)
         pFF[i].string_id= 0;
   }
    

   NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Filter form fields based on user input parameters. 
 *
 * LongDescription:::
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pCont::	    container of form fields
 *  minFieldWidth:: fields with widths < than this are removed
 *  minFieldHeight:: fields with heights < than this are removed
 *  minIsolatedFieldWidth:: non string fields less than this are removed 
 *
 *
 * Returns:::
 *   error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_filterFields(
    IPWState *pIPWState,
    UInt32 minFieldWidth,
    UInt32 minIsolatedFieldWidth,
    UInt32 minFieldHeight,
    FormFieldContainer *pCont)
{
   PROLOGUE;

   int i;
   Int32 nFieldsIn,nFieldsOut;
   UInt32 width,height;
   FormField *pFFRead,*pFFWrite;

   /* keep compiler happy, we may use this param in the future */
   pIPWState=pIPWState;

   nFieldsIn=pCont->count;
   pFFRead= pFFWrite = pCont->array;

   nFieldsOut=0;

   for(i=0;i<nFieldsIn;++i)
   {
      width=pFFRead->width;
      height=pFFRead->maxh;
      if(minFieldWidth)
      {
         if(width<minFieldWidth)
         {
            pFFRead++;
            continue;
         }
      }
      if(minFieldHeight)
      {
         if(height<minFieldHeight)
         {
            pFFRead++;
            continue;
         }
      }

     if( minIsolatedFieldWidth&&(!pFFRead->string_id) )
     {
         if(width<minIsolatedFieldWidth)
         {
            pFFRead++;
            continue;
         }
     }
 
     ++nFieldsOut;
     *pFFWrite=*pFFRead;
     pFFRead++;
     pFFWrite++;
   }

   pCont->count=nFieldsOut;

   NORMAL_RETURN;
}

/* callbacks for qsort */
/*****************************************************************************
 *
 *  qsort callback to compare FormField type objects by y coordinate. 
 *
 *
 *  p1 - pointer to one FormField
 *  p2 - pointer to the other FormField
 *
 * Returns
 *  pFFa->ybase-pFFb->ybase  
 *
 *****************************************************************************/
static int compare_by_y (const void *p1,const void *p2)
{
   const FormField *pFFa,*pFFb;

   pFFa= (const FormField *)p1;
   pFFb= (const FormField *)p2;

   return (pFFa->ybase-pFFb->ybase);
}

/*****************************************************************************
 *
 *   qsort callback to sort within the same "line" by left x.
 *
 *   qsort callback to compare FormField type objects in the same "line" by
 *   left x coordinate. 
 *
 *
 *
 *   p1 - pointer to one FormField
 *   p2 - pointer to the other FormField
 *
 *
 *
 * Returns
 *   pFFa->line_id-pFFb->line_id if non zero, otherwise pFFa->xleft-pFFb->xleft 
 *
 *
 *****************************************************************************/
static int compare_in_line_by_left_x (const void *p1, const void *p2)
{
   const FormField *pFFa,*pFFb;
   Int32 diff;

   pFFa= (const FormField *)p1;
   pFFb= (const FormField *)p2;

   diff= pFFa->line_id-pFFb->line_id;

   if(diff)
      return diff;
   else
      return (pFFa->xleft-pFFb->xleft);

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  initializes an empty form field container.
 *
 * LongDescription:::
 *  initializes an empty form field container.
 *
 *
 * Input:::
 *  pIPWState::      State structure
 *  pCont::	     container of form fields
 *
 * Returns:::
 *  error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_initFormFieldContainer(
		IPWState *pIPWState, 
		FormFieldContainer *pCont
		)
{
   PROLOGUE;
   Int32 size;

   pCont->max_count = INITIAL_ARRAY_LEN;
   pCont->count = 0;
   size=INITIAL_ARRAY_LEN*sizeof(FormField);
   pCont->array= IPW_MALLOC(size,cAllocatorData);
   if(pCont->array==NULL)
      return cErrNullPtr;
   NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  appends a form field to a form field container.
 *
 * Input:::
 *  pIPWState::	      pointer to IPCore state
 *  pFF::             pointer to form field to be appended to container
 *  pCont::	      container of form fields
 *
 *
 * Returns:::
 *  error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
wp_appendToFormFieldContainer(
		IPWState *pIPWState,
		FormField *pFF,
		FormFieldContainer *pCont
		)
{
   PROLOGUE;
   Int32 count,new_count,size;
   char *ptr;

   count=pCont->count;
   new_count= pCont->count+1;
   if(new_count > pCont->max_count)
   {
      pCont->max_count += ARRAY_EXPANSION_INCREMENT;

      size= pCont->max_count*sizeof(FormField);
      ptr= (char *) IPW_MALLOC(size,cAllocatorData);
      if(ptr==NULL)
         return cErrNullPtr;
      size= count*sizeof(FormField);
      memcpy(ptr,(char *)pCont->array,size);
      IPW_FREE(pCont->array,cAllocatorData);
      pCont->array=(FormField *)ptr;
   }

   pCont->array[count]= *pFF;
   pCont->count=new_count;

   NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Aggressively remove stipple from pPixrS, put pepper noise in ppPixrDots. 
 *
 * LongDescription:::
 *  With respect to forms, we're mostly interested in finding rulings, so
 *  we can be pretty liberal about removing small dots, as long as we do
 *  not remove the dots from dotted lines. In this context, stipple removal
 *  just means removing dots that are clustered together such that the clusters 
 *  are thicker than the thickness of any dotted ruling we expect to find.
 *  We do this before finding rulings, since the ruling finding is susceptible
 *  to hallucinating on stipple. In the process of removing these clustered 
 *  dots, we make a map of the pepper, which we will subtract from the image  
 *  after finding the rulings, since at this point we want to remove any noise 
 *  that might make us miss the white margins we look for above the rulings 
 *  when looking for fields. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  pPixrS:: pointer to Pixr for which we find noise and remove stipple
 * Output:::
 *  ppPixrDots:: pointer to pixr containing small dots 
 *
 *
 * Returns:::
 *  error status
 *
 *
 * FunctionCategory:::
 *  forms
 *
 * Keywords:::
 *  forms
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  danis
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_removeStippleAndMakePepperMask(
    IPWState *pIPWState,
    PIXR *pPixrS,
    PIXR **ppPixrDots)
{
   PROLOGUE;
   STREL    *pStrel3x3=NULL;
   PIXR *pPixrA=NULL,*pPixrB=NULL,*pPixrStipple=NULL; 

   ERR_PASS_CLEANUP (w_removePepper (pIPWState, pPixrS, 3, ppPixrDots));
   ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrS, XOR, *ppPixrDots));

   ERR_PASS_CLEANUP( w_reduce8(pIPWState,*ppPixrDots,1,1,1,&pPixrA) );

   ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel3, &pStrel3x3));
   ERR_PASS_CLEANUP (w_erode (pIPWState, pPixrA, pStrel3x3, &pPixrB));
   ERR_PASS_CLEANUP (w_dilate (pIPWState, pPixrB, pStrel3x3, &pPixrA));
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrB));

   /* expand mask */
   ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, *ppPixrDots, pPixrA, &pPixrStipple));
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrA));
   ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPixrStipple, SUBTRACT, pPixrS));
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPixrStipple));

   NORMAL_RETURN;
    
ERR_CLEANUP:
   if (pPixrA) w_destroyPixr (pIPWState, &pPixrA);
   if (pPixrB) w_destroyPixr (pIPWState, &pPixrB);
   if (pPixrB) w_destroyPixr (pIPWState, &pPixrStipple);

   ERR_RETURN;
}

/*

 $Log: forms.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.8  1998/01/14 00:11:54  ddavies
 forms.c -> /product/ipcore/ipwrap/src/RCS/forms.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.7  1997/11/19 20:22:16  danis
 danis on Wed Nov 19 12:20:59 PST 1997

 forms.c -> /product/ipcore/ipwrap/src/RCS/forms.c,v

 Fix to remedy erroneous assignment of a negative value to
 the height member of a form field.

 Revision 1.6  1997/08/30 00:18:54  danis
 danis on Fri Aug 29 17:18:11 PDT 1997

 forms.c -> /product/ipcore/ipwrap/src/RCS/forms.c,v

 One of the Keywords tags had no entry for it.

 Revision 1.5  1997/08/22 18:24:57  danis
 danis on Fri Aug 22 11:24:20 PDT 1997

 forms.c -> /product/ipcore/ipwrap/src/RCS/forms.c,v

 Fixed things that caused WATCOM to complain about unreferenced args.

 Revision 1.4  1997/08/22 14:42:06  danis
 danis on Fri Aug 22 07:41:03 PDT 1997

 forms.c -> /product/ipcore/ipwrap/src/RCS/forms.c,v

 Put in test for NULL pointer in w_destroyFormFieldArray

 Revision 1.3  1997/08/21 20:24:16  danis
 danis on Thu Aug 21 13:24:00 PDT 1997

 forms.c -> /product/ipcore/ipwrap/src/RCS/forms.c,v

 Made mistake in last commit.

 Revision 1.2  1997/08/21 19:59:00  danis
 danis on Thu Aug 21 12:58:24 PDT 1997

 forms.c -> /product/ipcore/ipwrap/src/RCS/forms.c,v

 Fixes in syntax that coreCheckC clued me into.

 Revision 1.1  1997/08/20 22:22:32  danis
 Initial revision


 */
