/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* 
 *  frame1.c:  includes subroutines
 *
 *      for bordering (size change):
 *                w_addBorder()
 *
 *      for frames:
 *                w_addFrame()
 *                w_removeFrame()
 */



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "frames.pub"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "frame.pub"
#include "shrpixr.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: frame1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a PIXR from adding a border to another PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prs::            Ptr to source PIXR.
 *  pixels::         Number of border pixels to be added to each of 4 sides.
 *                   Must be >= 0.
 *  color::          Color of added border.
 *                   0 means make border white (minval for nonbinary);
 *                   1 means make border black (maxval for nonbinary).
 *
 * Output:::
 *  prd::            Ptr to dest PIXR with added border.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, rasterOp, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addBorder (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               pixels,
    Int32               color,
    PIXR              **prd)
{
    PROLOGUE;
    Int32          ws, hs, wd, hd, d, channels;
    ScaledInt1616  xdpi, ydpi;
    PIXR          *pixr = NULL;
    ColorType      type;
    ColorMap      *map;
   
    ERR_GEN_PIXR(prs, cByteOrderBig, cDepthAny, cChannels1, cFrame32);
    ERR_GEN(prd == NULL, cErrNullPtr);
    ERR_GEN(*prd != NULL, cErrPtrNotNull);
    ERR_GEN(pixels < 0, cErrInvalidParm);

    if (pixels == 0)  
    {
	/* just make a copy */
	ERR_PASS( w_copy(pIPWState, prs, &pixr));
    }
    else {
	/* make new pixr */
	ERR_PASS( w_pixrGetWidth(pIPWState, prs, &ws));
	ERR_PASS( w_pixrGetHeight(pIPWState, prs, &hs));
	ERR_PASS( w_pixrGetDepth(pIPWState, prs, &d));
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prs, &xdpi, &ydpi));
	ERR_PASS( w_pixrGetChannelCount(pIPWState, prs, &channels));
	ERR_PASS( w_pixrGetType(pIPWState, prs, &type));
	ERR_PASS( w_pixrGetColorMap(pIPWState, prs, &map));

	wd = ws + 2 * pixels;
	hd = hs + 2 * pixels;

	ERR_PASS( w_createPixrExtended(pIPWState, wd, hd, d, xdpi, ydpi,
				       cFrameBits, channels, type, map,
				       &pixr));

	/* if add'l border is black, make everything
	 * not including the Frame Bits black */
	if (color == 1)
	    ERR_PASS_CLEANUP( w_invert(pIPWState, pixr, &pixr));

	/* blit the old image, without Frame Bits, into the center of
	 * the new image.  We have to be careful, because the Frame Bits
	 * are always 0, but the color of the added border can be
	 * either 0 or 1 */
	ERR_PASS_CLEANUP( w_rasterop (pIPWState,
				      0, 0,
				      prs, 
				      PIX_SRC,
				      pixels, pixels, ws, hs,
				      pixr));
    }

    *prd = pixr;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &pixr);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a PIXR by adding frame pixels on all 4 sides to another PIXR.
 *
 * LongDescription:::
 *  Add cFrameBits of frame pixels to all 4 sides of PIXR.
 *  In the event of an existing frame in the source PIXR, the frame size
 *  of the dest PIXR will be that frame + cFrameBits.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prs::            Ptr to single-channel source PIXR.
 *
 * Output:::
 *  prd::            Ptr to ptr to created PIXR with frame.
 *                   prd is allocated and filled in.
 *                   *prd must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addFrame (
    IPWState           *pIPWState,
    PIXR               *prs,
    PIXR              **prd)
{
    PROLOGUE;
    Int32           ws, hs, bpls, bpld;
    Int32           depth, frame;
    UInt8          *pImages, *pImaged;
    ScaledInt1616   xdpi, ydpi;
    PIXR           *pixr = NULL;

    ERR_GEN_PIXR(prs, cByteOrderBig, cDepthAny, cChannels1, cFrameAny);
    ERR_GEN(prd == NULL, cErrNullPtr);
    ERR_GEN(*prd != NULL, cErrPtrNotNull);

    /* make new pixrect */
    ERR_PASS( w_pixrGetDepth(pIPWState, prs, &depth));
    ERR_PASS( w_pixrGetWidth(pIPWState, prs, &ws));
    ERR_PASS( w_pixrGetHeight(pIPWState, prs, &hs));
    ERR_PASS( w_pixrGetFrame(pIPWState, prs, &frame));
    ERR_PASS( w_pixrGetBpl(pIPWState, prs, &bpls));
    ERR_PASS( w_pixrGetImage(pIPWState, prs, &pImages));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prs, &xdpi, &ydpi));

    ERR_PASS( w_createPixrNoInit(pIPWState, ws, hs, depth, xdpi, ydpi,
				 frame + cFrameBits, &pixr));

    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pixr, &bpld));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, pixr, &pImaged));

    ERR_PASS_CLEANUP( i_addFrame (pIPWState, pImages, ws, hs, depth,
				  bpls, pImaged, bpld));
    *prd = pixr;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    w_destroyPixr(pIPWState, &pixr);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a PIXR by removing frame pixels on all 4 sides from another PIXR.
 *
 * LongDescription:::
 *    If the input PIXR has at least cFrameBits of frame
 *    pixels on each side, it makes a new PIXR that
 *    has cFrameBits fewer frame pixels on each side.
 *    If the input PIXR has no frame pixels, it returns
 *    a copy of the input Pixr.
 *    We assume that frame pixels come in units of 
 *    cFrameBits (i.e., 32).
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prs::            Ptr to single-channel source PIXR.
 *
 * Output:::
 *  prd::            Ptr to ptr to created PIXR with removed frame.
 *                   prd is allocated and filled in.
 *                   *prd must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_removeFrame (
    IPWState           *pIPWState,
    PIXR               *prs,
    PIXR              **prd)
{
    PROLOGUE;
    Int32           ws, hs, depth, frame, bpls, bpld;
    UInt8          *pImages, *pImaged;
    ScaledInt1616   xdpi, ydpi;
    PIXR           *pixr = NULL;

    ERR_GEN_PIXR(prs, cByteOrderBig, cDepthAny, cChannels1, cFrameAny);
    ERR_GEN(prd == NULL, cErrNullPtr);
    ERR_GEN(*prd != NULL, cErrPtrNotNull);

    /* if not framed */
    ERR_PASS( w_pixrGetFrame(pIPWState, prs, &frame));
    if (frame < cFrameBits)
    {
	ERR_PASS( w_copy(pIPWState, prs, &pixr));
    }
    else
    {
	/* else, make new Pixr with cFrameBits of pixels removed from
	 * each side of the frame */
	ERR_PASS( w_pixrGetWidth(pIPWState, prs, &ws));
	ERR_PASS( w_pixrGetHeight(pIPWState, prs, &hs));
	ERR_PASS( w_pixrGetDepth(pIPWState, prs, &depth));
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prs, &xdpi, &ydpi));

	ERR_PASS( w_createPixrNoInit(pIPWState, ws, hs, depth, xdpi, ydpi,
			             frame - cFrameBits, &pixr));

	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prs, &ws));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prs, &hs));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prs, &pImages));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, pixr, &pImaged));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prs, &bpls));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pixr, &bpld));

	/* blit the inner part of the src, exclusive of cFrameBits,
	 * into the dest.  */
	ERR_PASS_CLEANUP( i_removeFrame (pIPWState, pImages,
					 ws, hs, depth, bpls,
					 pImaged, bpld));
    }
    

    *prd = pixr;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    w_destroyPixr(pIPWState, &pixr);
    ERR_RETURN;
}

/*

 $Log: frame1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.24  1998/01/14 00:11:56  ddavies
 frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.23  1997/06/26 20:01:01  ddavies
 frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.22  1997/04/15 21:48:59  ddavies
 frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.21  1996/01/04 20:22:46  ejaquith
 frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.20  1995/12/13  04:07:25  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.19  1995/11/13  20:11:23  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.18  1995/10/13  22:00:29  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/10/02  19:17:57  jcraig
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * w_invert takes a pointer to a pointer to a destination pixr.
 *
 * Revision 1.16  1995/09/29  00:36:37  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.15  1995/09/21  22:53:10  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.14  1995/09/14  18:51:43  ejaquith
 * ejaquith on Thu Sep 14 11:51:16 PDT 1995
 *
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * Fixed types to be used with w_pixrGetResolutionDPI().
 *
 * Revision 1.13  1995/09/12  16:55:51  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.12  1995/09/08  00:15:47  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.11  1995/09/07  01:31:15  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.10  1995/09/06  20:19:12  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  02:21:22  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.8  1995/08/31  21:12:57  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/29  21:54:33  ddavies
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.6  1995/08/17  22:02:19  ejaquith
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.5  1994/09/30  15:52:28  lperry
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * Add include of utils.pub
 *
 * Revision 1.4  1994/09/29  23:20:58  lperry
 * lperry on Thu Sep 29 16:20:35 PDT 1994
 *
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * Change to use frame.pub instead of pixrpros.h
 *
 * Revision 1.3  1993/12/03  01:14:48  ddavies
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:12:06  ddavies
 * frame1.c -> /product/ipcore/ipwrap/src/RCS/frame1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

