/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* 
 *  frame2.c: 
 *        Includes C versions of procedures for which assembler
 *        accelerators have been written.
 *
 *      for frames (aka cFrameBits):
 *	       w_addFrameIP()
 *	       w_removeFrameIP()
 *
 */



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "frames.pub"

/* prototypes */
#include "props.pub"
#include "shrpixr.pub"
#include "frame.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: frame2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add cFrameBits of frame pixels to all 4 sides to a frameless PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  prS::            Ptr to single-channel source PIXR that has no frame bits,
 *                   but has enough memory allocated to hold image data+frame
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addFrameIP (
    IPWState           *pIPWState,
    PIXR               *prS)
{
    PROLOGUE;
    Int32		 sW, sH, depth;
    Int32		 sBpl;
    UInt8               *pImage;

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8, 
		 cChannels1, cFrame0);

    /* We assume the pixrect is actually big enough to hold the image
     * plus frame.  The image is crammed into the top portion
     * of the pixrect.  We copy it down to the bottom, fix the width
     * and height, then clear out the frame.
     */
    
    /* get pixrect dimensions */
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &sW));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &sH));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pImage));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
    
    /* do the real work */
    ERR_PASS( i_addFrameInPlace (pIPWState, pImage, sW, sH, depth, sBpl));

    /* fix the pixrect's dimensions.  Remember to use the legal
     * method of calculating bytes/line, not some local hack.
     */
    ERR_PASS( w_pixrSetFrame(pIPWState, prS, cFrameBits));
    ERR_PASS( w_pixrBpl(pIPWState, prS, &sBpl));
    ERR_PASS( w_pixrSetBpl(pIPWState, prS, sBpl));

    /* return the modified pixr */
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Remove cFrameBits of frame pixels from all 4 sides of framed PIXR.
 *
 * LongDescription:::
 *  We will return with the same PIXR with cFrameBits of frame removed
 *  on each side.  The image bits will be crammed into the top portion.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  prS::            Ptr to single-channel source PIXR that has cFrameBits
 *                   of frame, which will be removed.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_removeFrameIP (
    IPWState           *pIPWState,
    PIXR               *prS)
{
    PROLOGUE;
    Int32		 sW, sH, depth;
    Int32		 sBpl, dBpl;
    UInt8               *pImage;

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    /* get pixrect dimensions */
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &sW));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &sH));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pImage));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
    
    /* make the frameless image as small as possible */
    dBpl = (((sW - 2*cFrameBits) * depth) + 7)/8;
    ERR_PASS( i_removeFrameInPlace (pIPWState, pImage, sW, sH,
				    depth, sBpl, dBpl));

    /* fix the pixr's dimensions */
    ERR_PASS( w_pixrSetFrame(pIPWState, prS, 0));
    ERR_PASS( w_pixrSetBpl(pIPWState, prS, dBpl));

    /* return the modified pixrect */
    NORMAL_RETURN;
}

/*

 $Log: frame2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.22  1998/01/14 00:11:57  ddavies
 frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.21  1996/09/24 02:48:20  ddavies
 frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v

 Use w_pixrBpl to calculate the bytes/line of the destination image in
 w_addFrameInPlace instead of trying to calculate it here.  The bytes/line
 can depend on the processor, so it's a subtle calculation.

 Revision 1.20  1996/01/04 20:22:48  ejaquith
 frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.19  1995/12/13  04:07:28  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.18  1995/11/13  20:11:25  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.17  1995/10/13  22:00:32  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.16  1995/09/29  00:36:42  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.15  1995/09/21  22:53:13  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.14  1995/09/12  16:55:55  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.13  1995/09/07  01:31:17  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:19:15  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  02:21:25  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:13:01  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/29  21:54:38  ddavies
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.8  1995/08/17  22:02:22  ejaquith
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.7  1994/09/30  15:52:32  lperry
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * Add include of utils.pub
 *
 * Revision 1.6  1994/09/29  23:21:06  lperry
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * Change to use frame.pub instead of pixrpros.h
 *
 * Revision 1.5  1993/12/22  05:01:12  ddavies
 * ddavies on Tue Dec 21 21:00:36 PST 1993
 *
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * Change removeFrameInPlacePr so it produces frameless images of minimum size -
 * the lines are padded to byte boundaries only.
 *
 * Revision 1.4  1993/12/14  21:32:47  ddavies
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * Fix the calculation of the destination image's bytes/line in
 * addFrameInPlacePr because the source images's bytes/line might not be
 * a multiple of 4.
 *
 * Revision 1.3  1993/12/03  01:14:53  ddavies
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:12:13  ddavies
 * frame2.c -> /product/ipcore/ipwrap/src/RCS/frame2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

