/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fthresh.c -- Heuristics for finding thresholding values, etc
 *
 * Exports:
 *
 *	findBackgroundThreshold	()
 *	findTextThreshold ()
 *	threshTextOnGray ()
 *	adaptiveThreshold ()
 *	scaledSubtractPr ()
 *	makeBackgroundImage ()
 *	removeBackground ()
 *      findEdgeThreshold ()
 *
 */


#include <stdio.h> /* 4debug */
#include <stdlib.h> /* 4debug */

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "logicops.h"
#include "defines.pub"
#include "thrshold.h"
#include "morphg.h"
#include "frames.pub"
#include "imageref.pub"
#include "shrscgr.pub"
#include "shrip.pub"
#include "imagerep.h"

/* prototypes */
#include "props.pub"
#include "ccthresh.pub"
#include "grbinop.pub"
#include "rescale.pub"
#include "logicops.pub"
#include "expand.pub"
#include "morph.pub"
#include "thrshold.pub"
#include "intops.pub"
#include "strel.pub"
#include "fthresh.pub"
#include "color.pub"
#include "reduce.pub"
#include "morphg.pub"
#include "curves.pub"
#include "prxform.pub"
#include "io.pub"
#include "shrip.pub"

/*#include "runhist.pub"*/


IP_RCSINFO(RCSInfo, "$RCSfile: fthresh.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

#define cIntegrateBins		32
#define cIntegrateStepSize	 8
#define cHistoReduce		 4
#define cSlope8Cutoff		64  /* Was 32 */
#define cSlopeRCount		 6   /* Max repeating slope count */
#define cSlope4Cutoff	       800  /* Was 400 */
#define cConnectivityIndex      30  /* Was 10 */
/*#define DEBUG */


#ifdef DEBUG
void CDECL writeHisto (char *s, Int32 *histo)
#endif

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *      Find a threshold for the background. Also finds "paper white".
 *
 * LongDescription:::
 *	Find a threshold for the background. Also finds "paper white"
 *	point, which is defined as the peak of the background curve.
 *	We've also added a parameter for "black point", which will be
 *	set to zero at the moment, but might be filled with the darkest
 *	significant thing in the image.
 *
 * PrivateDescription:::
 *  Find the highest peak between the middle and the end, and threshold near 
 *  there, or maybe in between there and its lower valley,
 *  figuring that that will threshold below most of the image and text. 
 *  Then:
 *  reduce (1, 2, 0, 0) (more if >100dpi);
 *  open 9h and 9v to get rid of the text;
 *  XOR that into the original reduced thing to get a text mask;
 *  close that with strel3 or 4 to get a mask of the text;
 *  scale up the mask;
 *  mask out the zeros in that mask in the original;
 *  histogram with the masked copy, and throw away the masked bin (255);
 *
 *  The theory, then, is that the result will be a histogram of mostly text
 *  regions. There should then be one large peak, which we can march down 
 *  the side of.  Note that this assumes that text is dark, and won't work
 *  on pages where the text is lighter than the background. 
 *  That will eventually need to be fixed.
 *
 *  Note also that it is silly to do the opGrayBin thing, because it would
 *  be both less wasteful and more efficient to make a version of the
 *  histogrammer that took a mask.
 *
 *  This whole scheme is a bit of a mess. It was being relied 
 *  on to solve the problem of thresholding for the purpose of text/image
 *  segmentation in the presence of bleed-through, which is better
 *  solved using other methods.
 *
 * Input:::
 *  pIPWState::         State structure
 *  imageP::            Gray (4/8 bpp) image for whic to find threshold
 *  rect::              If non-null, look inside this rectangle
 *  generateTextMask::  If cTrue, keep the "mask" image were we thought
 *                      saw text. Note that this mask is not necessarily
 *                      all the text, just a good part of it that we
 *                      found for the purposes of looking for background.
 *
 * Output:::
 *  backIndex::         Text threshold computed  
 *  blackPoint::        "Black Point", which is currently always 0
 *  whitePoint::        "White Point" computed, i.e. paper white
 *  invert::            If cTrue, the image should be inverted, which is
 *                      currently never set
 *  probablyText::      If cTrue, we're pretty sure that the image
 *                      contains text
 *  pPrTextMask::       If generateTextMask was passed as true, this
 *                      will contain the text mask
 *
 * Returns:::
 *  error status
 * 
 * Notes:::
 *  This thresholder tends to emphasize bleed-through. There is a fix
 *  for that, but it was taken out, because the result of this is usually
 *  passed to segmentPage, and that function gets rid of the bleed-through,
 *  around rectangular images at least, fairly well, and the complaint
 *  was that if the bleed-through is suppressed here, so then is part
 *  of the images, which become un-rectangular.
 *
 * FunctionCategory:::
 *  threshold finding
 *
 * Keywords:::
 *  threshold finding, histogram
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod:: 
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findBackgroundThreshold (
    IPWState           *pIPWState,
    PIXR               *imageP,
    RECTANGL           *rect,
    Bool                generateTextMask,
    Int32              *backIndex,
    Int32              *blackPoint,
    Int32              *whitePoint,
    Bool               *invert,
    Bool               *probablyText,
    PIXR              **pPrTextMask)
{
    PROLOGUE;   

    Int32   *histo = NULL, *dx = NULL, returnCode;
    Int32   *intHisto = NULL, /* bins, */ *startIndex = NULL, *endIndex = NULL,
	     depth, i, /*j,*/ points, thresh, oneQuarter,
	     highestPoint, highestIndex, bits, slopeCutoff,
	     currentCount, currentRepeats, reduction, width, height, dpi;
    PIXR    *reduced = NULL, *result = NULL, *binImage = NULL,
	    *maskedImage = NULL;
    ScaledInt1616    xDpi, yDpi;
    RECTANGL lrect;
    STREL   *pStrel2 = NULL, *pStrel4 = NULL, *pStrel5 = NULL, 
	    *pStrel9h = NULL, *pStrel9v = NULL; 

    ERR_GEN_PIXR (imageP, cByteOrderBig, cDepth4 | cDepth8, 
		  cChannels1 | cChannels3, cFrame32);

    ERR_PASS (w_strel1Inits(pIPWState));
    ERR_PASS (w_getStrel (pIPWState, strel2, &pStrel2));
    ERR_PASS (w_getStrel (pIPWState, strel4, &pStrel4));
    ERR_PASS (w_getStrel (pIPWState, strel5, &pStrel5));
    ERR_PASS (w_getStrel (pIPWState, strel9h, &pStrel9h));
    ERR_PASS (w_getStrel (pIPWState, strel9v, &pStrel9v));

    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, imageP, &depth));

    if (rect == NULL)
    {
	rect = &lrect;
	rect->x = 0;
	rect->y = 0; 
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, imageP, &(rect->w)));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, imageP, &(rect->h)));
    }

    points = (1 << depth);
    slopeCutoff = (depth == 8) ? cSlope8Cutoff : cSlope4Cutoff;
    oneQuarter = points / 4;
    thresh = 1 << (depth - 1);

    returnCode = cErrAborted;

again:
    IP_YIELD;

    /* Threshold the picture */
    binImage = NULL;

    ERR_PASS (w_threshold (pIPWState, imageP, 0, thresh - 1, THRESH_NORM, 
			   &binImage));
    
    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, imageP, 
					      &xDpi, &yDpi));
    dpi = MAX (ScaledInt1616_TO_Int32(xDpi), ScaledInt1616_TO_Int32(yDpi));

    /* Make a reduced version */
    if (dpi <= 100)
    {
	ERR_PASS_CLEANUP (w_reduce (pIPWState, binImage, 1, 2, 0, 0, 
				      &reduced));
	reduction = 4;
    }
    else if (dpi <= 200)
    {
	ERR_PASS_CLEANUP (w_reduce (pIPWState, binImage, 1, 1, 2, 0, 
				      &reduced));
	reduction = 8;
    }
    else
    {
	ERR_PASS_CLEANUP (w_reduce (pIPWState, binImage, 1, 1, 1, 2, 
				      &reduced));
	reduction = 16;
    }

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &binImage));

    IP_YIELD;

    /* Get rid of isolated pixels */
    ERR_PASS_CLEANUP (w_open (pIPWState, reduced, pStrel2, &reduced));

    /* Open horizontals and verticals */
    result = NULL;
    ERR_PASS_CLEANUP (w_open (pIPWState, reduced, pStrel9h, &result));
    ERR_PASS_CLEANUP (w_open (pIPWState, result, pStrel9v, &result));

    /* XOR that with the reduced thing */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, reduced, XOR, result));

    /* Free anything you can */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &reduced));

    IP_YIELD;

    /* Close up the regions */
    ERR_PASS_CLEANUP (w_close (pIPWState, result, pStrel4, &result));

    /* Dilate it a bit to get some more background */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, result, pStrel5, &result));

    /* Count the pixels and make sure there are enough */
    ERR_PASS_CLEANUP (w_numberBits (pIPWState, result, &bits));

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, result, &width));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, result, &height));
    /* If we didn't get enough bits, try again.  */
#if 0
    if (bits)
	fprintf(stderr, "ratio is %f\n", (Float64) width * height / (Float64) bits);
    else
	fprintf(stderr, "ratio undefined\n");
#endif
    
    if (width * height > bits * 10)
    {
	/* We failed; reduce the threshold and try again */
	if (thresh > points/3)
	{
	    thresh -= (depth == 8) ? 16 : 1;
	    goto again;
	}
	/* We just use that ... */
	*probablyText = cFalse;
    }
    else
	*probablyText = cTrue;

    /* Expand the bin mask */

    if (width * reduction < 33)
    {
	/* Can't expand to <33 pixels; just do a scale-to-size */
	ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, result, width * reduction,
					   height * reduction, &binImage));
    }
    else
    {
	ERR_PASS_CLEANUP (w_expand(pIPWState, result, reduction, &binImage));
    }

    IP_YIELD;
    if (generateTextMask)
    {
	if (*pPrTextMask == NULL)
	    ERR_PASS_CLEANUP (w_copy (pIPWState, binImage, pPrTextMask));
	else
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, binImage, COPY, 
					 *pPrTextMask));
    }
	
    /* Don't need the result anymore */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &result));

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, imageP, &width));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, imageP, &height));
    
    if (rect->x == 0 && rect->y == 0 &&	rect->w == width && rect->h == height)
    {
	/* Right now, we only do masked histograms on full-size */
	ERR_PASS_CLEANUP (w_histoMask (pIPWState, imageP, binImage, NULL,
				       &histo));

	/* A hack: we get rid of the last bin. This 
	 * isn't "necessary" using masked histograms (it was
	 * when using opGrayBinPr) but the rest of the code
         * sort of assumes that it was zeroed. 
	 */
	if (depth == 8)
	{
	    histo[points - 1] = 0;
	}
    }
    else
    {
	/* Do the mask operation on the image */
	ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, imageP, binImage, 
				       PIX_SRC & PIX_DST, &maskedImage));
	
	/* Set it's resolution, so that text thresholder knows what to do */
	ERR_PASS_CLEANUP (w_pixrSetResolutionDPI (pIPWState, maskedImage, 
					     xDpi, yDpi));

	/* Now, do the histogram */
	if (depth == 4)
	{
	    ERR_PASS_CLEANUP (w_histoSubSampled (pIPWState, maskedImage, 
						     rect, cHistoReduce, 
						     &histo));
	}
	else
	{
	    ERR_PASS_CLEANUP (w_histo (pIPWState, maskedImage, rect, 
					   &histo));
	}
	
	/* Blow off white */
	if (depth == 8)
	{
	    histo[points - 1] = 0;
	}
    }

/*    DESTROY_PIXR (maskedImage); */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &binImage));
    binImage = NULL;

    /* There we have a histo. Let's just show it for now */
/*    plotPoints (histo, NULL, cPlotInt32, points, cPlotImpulse); */

    /* Now, do it again with the new histogram */

    /* Get rid of missing codes */
    ERR_PASS_CLEANUP (w_interpolateMissingCodes (pIPWState, histo, points));

    /* Take the first derivative */
    dx = (Int32 *) IPW_CALLOC (points, sizeof (Int32), cAllocatorData);
    ERR_GEN_CLEANUP (dx == NULL, cErrNomem);

    ERR_PASS_CLEANUP (w_numericalDiff (pIPWState, histo, points, 1,  
				     (depth == 8) ? 2 : 0, &dx));

    /* Find the peak */
    highestPoint = -1;
    highestIndex = 0;
    for (i = 0; i < points; i++)
    {
	if (histo[i] > highestPoint)
	{
	    highestPoint = histo[i];
	    highestIndex = i;
	}
    }

    /* Set the black/white points. The black point is always zero for now */
    *whitePoint = highestIndex;
    *blackPoint = 0;

    if (/*depth == 4*/ 1)
    {
	/* 4-bit is so screwy, we should get a range from the data itself */
	for (i = highestIndex; i > 0 && dx[i] < 0; i--)
	    ;
	highestPoint = 0;

	for (; i > 0 && dx[i] > 0; i--)
	{
	    if (dx[i] > highestPoint)
		highestPoint = dx[i];
	}

	slopeCutoff = highestPoint / 4;

    }

    if (highestIndex < oneQuarter)
    {
	/* Ugh: way too low. Just set it to some nominal value */
	*backIndex = (depth == 8 ? 128 : 8);
    }
    else
    {
	/* Make sure we're going downhill */
	for (i = highestIndex; i > oneQuarter; i--)
	{
	    if ((depth == 4 && dx[i] > 0) ||	
		(dx[i] > slopeCutoff + 2))
		break;
	}
	
	/* Search backwards from the peak until we find the cutoff slope */
	currentCount = -1;
	currentRepeats = 0;
	for ( ; i > oneQuarter; i--)
	{
	    if ((dx[i] < slopeCutoff) && (dx[i-1] < slopeCutoff) && (dx[i-2] < slopeCutoff))
		break;

	    /* psm 4/17/94: Don't let it get stuck in the same count forever */
	    if (dx[i] == currentCount)
	    {
		currentRepeats++;
		if (currentRepeats > cSlopeRCount)
		    break;
	    }
	    else
	    {
		currentCount = dx[i];
		currentRepeats = 0;
	    }
	}

	/* Take the point we passed by; dx tends to be 1 off, hence + 2 */
	*backIndex = i + 2;
    }

    IP_YIELD;

    if (maskedImage) 
	ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &maskedImage));

    *invert = cTrue;

    if (histo) IPW_FREE (histo, cAllocatorData);
    if (intHisto) IPW_FREE (intHisto, cAllocatorData);
    if (binImage) IPW_FREE (binImage, cAllocatorData);
    if (dx) IPW_FREE (dx, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (histo) IPW_FREE (histo, cAllocatorData);
    if (intHisto) IPW_FREE (intHisto, cAllocatorData);
    if (startIndex) IPW_FREE (startIndex, cAllocatorData);
    if (endIndex) IPW_FREE (endIndex, cAllocatorData);
    if (binImage) IPW_FREE (binImage, cAllocatorData);
    if (dx) IPW_FREE (dx, cAllocatorData);
    if (maskedImage)
	w_destroyPixr (pIPWState, &maskedImage);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find a threshold for the text in an image
 *
 * LongDescription:::
 *  Finds a global threshold for the gray image passed which, when
 *  applied, makes the text look good
 *
 * PrivateDescription:::
 *  Uses Luc Vincent's connected-component histogram code.
 *
 * Input:::
 *  pIPWState::      State structure
 *  imageP::         The grayscale (4- or 8-bit) image to analyze
 *
 * Output:::
 *  textIndex::      The threshold computed
 *  invert::         If cTrue, invert the image, which never happens
 *                   at the moment
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  threshold finding
 *
 * Keywords:::
 *  threshold finding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod:: wrapper to Luc's code, mostly
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findTextThreshold (
    IPWState           *pIPWState,
    PIXR               *imageP,
    Int32              *textIndex,
    Bool               *invert)
{
    PROLOGUE;

    /* Use connected-component method */
    ERR_PASS (w_ccTextThreshold (pIPWState, imageP, cConnectivityIndex,
			       cFalse, textIndex));

    /* We always invert, because we don't detect inversion */
    *invert = cTrue;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find thresholds for both background and text
 *
 * LongDescription:::
 *  This is a wrapper function which calls w_findTextThreshold and
 *  w_findBackgroundThreshold.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::         State structure
 *  imageP::            Gray (4/8 bpp) image for whic to find threshold
 *  rect::              If non-null, look inside this rectangle
 *  generateTextMask::  If cTrue, keep the "mask" image were we thought
 *                      saw text. Note that this mask is not necessarily
 *                      all the text, just a good part of it that we
 *                      found for the purposes of looking for background.
 *
 * Output:::
 *  backIndex::         Text threshold computed  
 *  textIndex::       
 *  invert::            If cTrue, the image should be inverted, which is
 *                      currently never set
 *  probablyText::      If cTrue, we're pretty sure that the image
 *                      contains text
 *  pPrTextMask::       If generateTextMask was passed as true, this
 *                      will contain the text mask
 *
 * Returns:::
 *
 * Notes:::
 *  Discards black/white point; if you want them, you'll have to 
 *  call w_findTextThreshold and w_findBackgroundThreshold individually.
 *
 * FunctionCategory:::
 *  threshold finding
 *
 * Keywords:::
 *  threshold finding
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findThresholds (
    IPWState           *pIPWState,
    PIXR               *imageP,
    RECTANGL           *rect,
    Bool                generateTextMask,
    Int32              *backIndex,
    Int32              *textIndex,
    Bool               *invert,
    Bool               *probablyText,
    PIXR              **pPrTextMask)
{
    PROLOGUE;
    Int32   white, black;

    ERR_PASS (w_findBackgroundThreshold (pIPWState,  imageP, rect, 
					 generateTextMask, backIndex, 
					 &black, &white, invert, probablyText,
				         pPrTextMask));

    ERR_PASS (w_findTextThreshold (pIPWState, imageP, textIndex, invert));

    NORMAL_RETURN;
}


#if 0
static Int32 counter = 1;  /* I am the ghost of debuggers past */
#endif

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Find a threshold for a gray image based on looking at edges.
 *
 * LongDescription:::
 *    Finds a threshold for a gray image by looking at edges.
 *    This allows one to build fairly stable adaptive thresholders.
 *
 * PrivateDescription:::
 *    The threshold finder based on connected components works pretty
 *    well when one has a full page of text. However, when attempting
 *    to do "accupage-esque" tiled thresholding, the c.c.-based thresholder
 *    was a bit too unstable.
 *
 *    This thresholding algorithm works on the following thought:
 *    if one looked at a typical gray edge, one would find that
 *    the black-white transitions started out at a small slope, increased
 *    in the middle, and then returned to a small slope near the end.
 *    The idea is that the point at which the slope is maximized is the
 *    "actual" edge. My theoretical reasoning, simplistic though it may
 *    be, is that if I take an infinite-frequency edge and apply a 
 *    symmetric blurring convolution, the "real" edge would
 *    indeed be at the point of maximum absolute-value slope in the 
 *    blurred result.
 *
 *    How do we implment this in a reasonably robust way? We probably 
 *    can't afford to do anything as fancy as trace the outlines of 
 *    connected components, and assume that edges lie along the normal.
 *    What we do instead is histogram the differences between adjacent
 *    horizontal pixels, and along with the histogram bits, maintain
 *    a parallel set of bins which record the values of both pixels
 *    which comprised the difference, i.e. if pixel1 is 100, and pixel2
 *    is 170, the difference is 70, so add 1 to bin #70 in the histogram,
 *    and add 270 (170 + 100) to bin #70 in the other array. When we're
 *    done, we look backwards down the histogram to find the peak difference.
 *    The idea is that the peak difference is most likely going to be 
 *    found on vertical edges, so the values that we counted will be
 *    valid as endpoints of the maximal-sloped difference. The value
 *    computed is then the value in the sum bin divided by twice the value
 *    in the histogram bin.
 *
 *    This has been tested quite a bit in the four-bit case, where I found
 *    that using 64x64 regions, text regions usually got quite a significant
 *    number of the same peak difference, thus making our statistic more
 *    meaningful. It remains to be seen whether the 8-bit case will
 *    do the same, or whether we'll want to truncate the lower bits in
 *    order to force slightly different values to the same bin to reduce
 *    noise in the result.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to analyze
 *  pRect::          Rectangle in which to search. If NULL, use the
 *                   entire image.
 *  minDiff::        Minimum difference between pixels to consider
 *                   as a threshold. If none is found >= this threshold,
 *                   the threshold computed is -1, indicating no threshold.
 *
 * Output:::
 *  pThresh::        Threshold computed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  threshold finding
 *
 * Keywords:::
 *  adaptive thresholding, threshold finding
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findEdgeThreshold (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *pRect,
    Int32               minDiff,
    Int32              *pThresh)
{
    PROLOGUE;
    UInt8 *pImage;
    Int32 w, h, x, y, bpl, depth;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1 | cChannels3,
		  cFrame32);

    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));
    
    if (pRect)
    {
	w = pRect->w;
	h = pRect->h;
	x = pRect->x + cFrameBits;
	y = pRect->y + cFrameBits;
    }
    else
    {
	ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &h));
	x = cFrameBits;
	y = cFrameBits;
    }

    ERR_PASS (w_pixrGetImageAll (pIPWState, prSrc, &pImage));
    ERR_PASS (w_pixrGetBpl (pIPWState, prSrc, &bpl));

    /* Call the ipshared thing to do the actual work */
    ERR_PASS_CLEANUP (i_findEdgeThreshold (pIPWState, pImage, depth, bpl, 
					   x, y, w, h, minDiff, pThresh));
    
    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Find thresholds for a gray image based on looking at edges.
 *
 * LongDescription:::
 *    Finds thresholds for a gray image by looking at edges.
 *    This allows one to build fairly stable adaptive thresholders.
 *
 * PrivateDescription:::
 *    Finf 2 thresholds, one for propagating, one for painting inside 
 *    rectangle. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to analyze
 *  pRect::          Rectangle in which to search. If NULL, use the
 *                   entire image.
 *  minDiff::        Minimum difference between pixels to consider
 *                   as a threshold. If none is found >= this threshold,
 *                   the threshold computed is 0, indicating no threshold.
 *
 * Output:::
 *  pThresh1::       Threshold computed
 *  pThresh2::       Threshold computed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  threshold finding
 *
 * Keywords:::
 *  adaptive thresholding, threshold finding
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findEdge2Thresholds (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    RECTANGL           *pRect,
    Int32               minDiff,
    Int32              *pThresh1,
    Int32              *pThresh2 )
{
    PROLOGUE;
    UInt8 *pImage;
    Int32 w, h, x, y, bpl, depth;
    PIXR  *prTmp = NULL;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1 | cChannels3,
		  cFrame32);

    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));

    if( depth == 4 )
	ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prSrc, &prTmp)); 
    else prTmp = prSrc;

    if (pRect)
    {
	w = pRect->w;
	h = pRect->h;
	x = pRect->x + cFrameBits;
	y = pRect->y + cFrameBits;
    }
    else
    {
	ERR_PASS (w_pixrGetWidth (pIPWState, prTmp, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prTmp, &h));
	x = cFrameBits;
	y = cFrameBits;
    }

    ERR_PASS (w_pixrGetImageAll (pIPWState, prTmp, &pImage));
    ERR_PASS (w_pixrGetBpl (pIPWState, prTmp, &bpl));

    /* Call the ipshared thing to do the actual work */
    ERR_PASS_CLEANUP (i_findEdge2Thresholds (pIPWState, pImage, bpl, 
					     x, y, w, h, minDiff, pThresh1,
					     pThresh2));
    if(prTmp != prSrc)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    
    NORMAL_RETURN;

ERR_CLEANUP:

    if( prTmp && (prTmp != prSrc ))
	w_destroyPixr (pIPWState, &prTmp);

    ERR_RETURN;
}

/*
 * $Log: fthresh.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 * Imported sources
 *
 * Revision 1.77  1998/06/30 17:40:28  xzhu
 * xzhu on Tue Jun 30 10:39:37 PDT 1998
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Moved function to fthresh1.
 *
 * Revision 1.76  1998/06/27 00:31:18  xzhu
 * xzhu on Fri Jun 26 17:30:04 PDT 1998
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Using textMask replacing morphology and fixed thresholding to mask text.
 *
 * Revision 1.75  1998/06/23 00:04:09  xzhu
 * xzhu on Mon Jun 22 17:03:47 PDT 1998
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added new function.
 *
 * Revision 1.74  1998/05/22 18:19:12  xzhu
 * xzhu on Fri May 22 11:19:03 PDT 1998
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added new function.
 *
 * Revision 1.73  1998/01/14 00:12:00  ddavies
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.72  1997/07/03 21:46:14  jcraig
 * jcraig on Thu Jul 3 14:45:59 PDT 1997
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * w_histoMask now takes a rectangle
 *
 * Revision 1.71  1997/04/22 18:20:02  xzhu
 * xzhu on Tue Apr 22 11:19:19 PDT 1997
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added IsNew when call w_ccTextThreshold().
 *
 * Revision 1.70  1997/04/15 21:50:09  ddavies
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Remove unused routines w_removeBackground and w_makeBackgroundImage.
 *
 * Revision 1.69  1996/07/19 15:34:17  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * w_findEdgeThreshold uses the ipshared routine rather than doing it
 * itself. This routine should probably be moved, because it doesn't
 * belong in appslib.
 *
 * Revision 1.68  1996/01/04 20:22:51  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.67  1995/12/13  04:07:33  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.66  1995/11/13  20:11:29  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.65  1995/10/13  22:00:36  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.64  1995/10/12  17:44:03  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * added explicit casts (float->int, long->short, long->char, etc) and
 * made floating constants the right size to shut up the NT compiler.
 *
 * Revision 1.63  1995/10/03  01:58:59  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.62  1995/09/21  22:53:17  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.61  1995/09/20  00:36:05  ddavies
 * ddavies on Tue Sep 19 17:35:54 PDT 1995
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Change type of allocatedHere from Int32 to Bool.
 *
 * Revision 1.60  1995/09/20  00:05:39  macleod
 * macleod on Tue Sep 19 17:05:19 PDT 1995
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Fixed bug in w_findBackgroundThreshold caused by change in w_threshold
 * semantics.
 *
 * Revision 1.59  1995/09/19  19:19:09  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.58  1995/09/11  21:53:50  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.57  1995/09/07  01:31:20  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.56  1995/09/06  20:19:17  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.55  1995/09/05  19:00:49  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Removed calls to GET_STREL macro, instead calling getStrel() function.
 * This required using local variables rather than invoking GET_STREL
 * as an argument to another function, since getStrel can return an error
 * code which should be checked.
 *
 * Revision 1.54  1995/09/01  21:02:17  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.53  1995/08/31  21:13:04  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.52  1995/08/29  21:54:43  ddavies
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.51  1995/08/25  01:36:29  ejaquith
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Fixed warnings
 *
 * Revision 1.50  1995/08/17  22:59:57  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Needed to comment out the #include of "shros.pub"
 *
 * Revision 1.49  1995/08/17  21:55:11  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.48  1995/05/24  17:54:50  macleod
 * macleod on Wed May 24 10:53:07 PDT 1995
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * In findEdgeThreshold, look in four directions rather than one when
 * figuring out adjacent pixel differences. This allows one to get a
 * better threshold when there are not many vertical components in a
 * particular region. It does add a bit of processing time. Currently,
 * this is only done in the 8 bit/pixel case! The 4 bit/pixel case still
 * looks only horizontally.
 *
 * Revision 1.47  1995/04/03  16:00:40  macleod
 * macleod on Mon Apr 3 08:54:27 PDT 1995
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added findEdgeThreshold. This is yet another thresholding algorithm,
 * which was developed to address the problem of thresholding on small
 * regions. The connected-component based thresholder proved too unstable
 * for this task. This thresholder is based on the following thought:
 * if one looked at a typical gray edge, one would find that the
 * black-white transitions started out at a small slope, increased
 * in the middle, and then returned to a small slope near the end.
 * The idea is that the point at which the slope is maximized is the
 * "actual" edge. Looking at it the other way, if we take an infinite-
 * frequency transition and apply a symmetrical smoothing convolution
 * to it, the point of maximal slope will be the point where the edge was
 * originally.
 * The algorithm works by histogramming the absolute value of horizontal
 * pixel differences. The peak difference is assumed to be the best
 * approximation of a vertical transition. We also keep track of both
 * intensities at the transition for each difference, so that we can then
 * guess at the threshold by taking the average value for the peak difference.
 *
 * Revision 1.46  1995/01/24  21:49:54  lperry
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Incorporate changes to use new strel manager interface.
 *
 * Revision 1.45  1994/12/21  02:53:08  macleod
 * macleod on Tue Dec 20 18:52:57 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Revert to the "old" background threshold finder.
 *
 * Revision 1.44  1994/11/23  00:24:28  macleod
 * macleod on Tue Nov 22 16:24:16 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Silenced WATCOM warning about unreferenced parameter.
 *
 * Revision 1.43  1994/11/22  18:53:52  macleod
 * macleod on Tue Nov 22 10:53:20 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Calls connected-component based background threshold finder.
 * The old version is still there, #if'ed out.
 *
 * Revision 1.42  1994/11/07  20:02:32  macleod
 * macleod on Mon Nov 7 11:59:44 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * in makeBackgroundImage, was using an even-sized strel, which resulted
 * in a slightly shifted result. Also added an erode to shrink the
 * background tint regions a bit, to avoid artifacts from roundoff errors
 * when expanding. This does mean that text butted right up against tints
 * can get an incorrect background.
 *
 * Revision 1.41  1994/11/07  19:20:11  macleod
 * macleod on Mon Nov 7 11:18:35 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * two things in removeBackground. First, maxLines is used to compute
 * the actual # of expanded lines, rather than the number of source lines.
 * Second, there was an egregious bug which caused the first line of
 * each chunk to be used as the background for the whole chunk.
 * (The precedence of '==' over '&' gets me once again)
 *
 * Revision 1.40  1994/11/04  18:59:43  macleod
 * macleod on Fri Nov 4 10:59:28 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Got rid of annoying warnings from acc by adding some casts.
 *
 * Revision 1.39  1994/11/04  03:16:53  macleod
 * macleod on Thu Nov 3 19:11:59 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added makeBackgroundImage, which takes an image and produces a "background"
 * image, which is a reduced version of the image (~25-50 dpi) which has
 * been closed so that small, foreground-y things are gone. Also added
 * removeBackground, which removes the background produced by
 * makeBackgroundImage from an image. The background image is scaled up
 * accordingly.
 * Both of these functions work on 4- or 8-bit images. makeBackgroundImage
 * can take a color image, and will produce a luminance image as its result.
 * removeBackground currently only works on grayscale images.
 * The result of having the background removed is useful when the image
 * is to be thresholded.
 * The background subtraction is scaled, so that, for example, text on a
 * dark background will end up about the same darkess as text on a
 * light background. This allows one to threshold the whole page with
 * one threshold, and have the text weights come out right.
 * These functions probably need more extensive testing.
 *
 * Revision 1.38  1994/10/18  22:21:55  macleod
 * macleod on Tue Oct 18 15:19:27 PDT 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * findBackgroundThreshold takes two additional arguments: black point and
 * white point. Currently, the black point is always set to zero.
 * The white point is the "paper white", i.e. the most common background
 * color around text.
 * Also made "findThresholds" call findBackgroundThreshold and then call
 * findTextThreshold, rather than having findThresholds do all the work,
 * which was a historical artifact.
 * Took out computeTextThreshold, threshTextOnGray, countConnectedComponents,
 * updateCounter, and scaledSubtractPr, all of which are, I believe,
 * not used.
 *
 * Revision 1.37  1994/10/11  17:40:46  lperry
 * lperry on Tue Oct 11 10:40:24 PDT 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Add include of shros.pub to get IP_YIELD definition.
 *
 * Revision 1.36  1994/10/07  19:06:04  lperry
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.35  1994/09/30  18:11:17  jcraig
 * jcraig on Fri Sep 30 11:11:05 PDT 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * go back to the previous version as requested by Macduff.
 *
 * Revision 1.33  1994/09/16  17:35:59  jcraig
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.32  1994/08/10  21:57:17  lperry
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Changes made to use new strel manager.
 *
 * Revision 1.31  1994/08/05  01:44:06  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Modified to compute a slope cutoff based on the maximum slope down from
 * the background peak. This results in (usually) more agressive
 * thresholds, which means that if there's a lot of bleed-through, the
 * result shouldn't be used unless one is segmenting with segmentTwice.
 *
 * Revision 1.30  1994/07/28  01:44:43  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * findThresholds calls ccBackgroundThreshold rather than using its
 * original algorithm. Warnings: 1. The rectangle passed is ignored.
 * The rectangle is always passed as NULL by XIPAPI (or so I'm told),
 * so this shouldn't be a problem, but we'll want to fix this in the
 * future. 2. There are some pictures (like a few in our image database)
 * where there is no good threshold at which stuff from the back of
 * the page doesn't bleed through. Sorry, we don't handle that yet, and
 * most likely won't until we move away from simple thresholding.
 *
 * Revision 1.29  1994/07/27  18:35:34  macleod
 * macleod on Wed Jul 27 11:34:34 PDT 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * This was a bug fix from a while ago that I never checked in:
 * expandPixr doesn't work if the result is <33 pixels wide, so if
 * that would be the case when expanding back the text mask,
 * use scaleToSizePixr instead
 *
 * Revision 1.28  1994/06/29  16:43:00  macleod
 * macleod on Wed Jun 29 09:42:53 PDT 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added some IP_YIELD calls
 *
 * Revision 1.27  1994/06/23  05:50:57  macleod
 * macleod on Wed Jun 22 22:47:31 PDT 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Fixed a bug: when creating the histogram using opGrayBinPr, the last bin
 * of the histogram needed to be chucked, because it counted every pixel not
 * in the mask (i.e. forced to white). Now, it uses histoMask, but that doesn't
 * have the problem. The trouble with this is that the heuristic which finds
 * the background threshold expects the high bin to be zeroed, so I just zero it.
 * A more correct thing is to fix the heuristic.
 * Also added scaledSubtractPr, which is used to subtract a background (determined
 * using grayscale morphology) from an image. It is "experimental", i.e. I'm just
 * checking it in so that (1) the bug fix gets in and (2) I don't lose track of it.
 *
 * Revision 1.26  1994/06/16  22:39:40  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * findTextThreshold and findThresholds use "ccTextThreshold", which is
 * Luc Vincent's connected-components-based thresholder, on 4-bit
 * as well as 8-bit images. findTextThreshold no longer takes a rectangle;
 * it also no longer calls findThresholds, since the text threshold
 * computation no longer depends on the background threshold computation.
 * Also, findThresholds uses histoMask to compute the histogram based on the
 * mask of text computed, rather than doing a histogram on a copy of the
 * image with the text regions smashed to white. The advantages of this
 * are twofold: first, we take less memory because we don't need to copy
 * the image, and second, it's quite a bit faster.
 *
 * Revision 1.25  1994/06/08  20:29:04  macleod
 * macleod on Wed Jun 8 13:28:54 PDT 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Uses thresholder in ccthresh.c to compute text threshold
 *
 * Revision 1.24  1994/05/17  18:38:57  macleod
 * macleod on Tue May 17 11:37:50 PDT 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Does text thresholding using a connected-component algorithm. This
 * algorithm is kind of slow, because it thresholds and analyzes the
 * image several times. Luc at XIS has a different algorithm, which we
 * intend to incorporate, which uses a similar tactic but is much faster.
 *
 * Revision 1.23  1994/04/03  02:08:32  ddavies
 * ddavies on Sat Apr 2 18:07:57 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Add the 8-connected parameter to call to i_findConnectedComponentBoundingBoxes.
 *
 * Revision 1.22  1994/03/24  23:03:59  macleod
 * macleod on Thu Mar 24 15:03:27 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * There were a couple of unused locals that WATCOM and gcc whined about
 *
 * Revision 1.21  1994/03/23  02:59:34  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * findThresholds uses "128" as its initial guess rather than being
 * fancy about it.
 *
 * Revision 1.20  1994/03/18  02:08:31  macleod
 * macleod on Thu Mar 17 18:08:14 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added sanity check for "initial guess", which should make it behave
 * a little better on uncropped images
 *
 * Revision 1.19  1994/03/17  03:02:01  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Changed call to findBackgroundThreshold, fixed a memory leak
 *
 * Revision 1.18  1994/03/10  02:03:33  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Took out a couple of unused variables
 *
 * Revision 1.17  1994/03/02  02:49:55  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Contains some experiments (#if0'd out) for finding optimal text thresholds.
 *
 * Revision 1.16  1994/02/19  02:54:41  macleod
 * macleod on Fri Feb 18 18:54:18 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added "findThresholds", which finds both text and background thresholds
 *
 * Revision 1.15  1994/02/18  22:21:46  macleod
 * macleod on Fri Feb 18 14:21:23 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Oops -- freed nil pointer
 *
 * Revision 1.14  1994/02/18  05:22:53  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * New version of findBackgroundThreshold. Makes a guess at the threshold,
 * then reduces the image, takes out small stuff, makes a mask out of the small
 * stuff, closes it, masks out that stuff in the original (sets it to white,
 * actually in a copy of the original) and then histograms it, the theory
 * being that it gets rid of the pictures and only histograms text regions.
 * It also does an iteration if it doesn't think it has enough text area.
 * Could be faster. There are several ways to speed it up.
 *
 * Revision 1.13  1994/02/15  01:43:22  macleod
 * macleod on Mon Feb 14 17:43:08 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Threshold routines take a rectangle, which can be NULL to indicate the whole image
 *
 * Revision 1.12  1994/02/08  23:35:44  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Text threshold backed off a bit from mean towards background thresh.
 *
 * Revision 1.11  1994/02/03  19:53:23  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added findTextThreshold. Took out a bunch of older routines which
 * weren't being used anywhere
 *
 * Revision 1.10  1994/01/31  20:58:07  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added findBackgroundThreshold, which finds a background threshold using
 * partitionCurve to partition the curve into regions, rather than using
 * an evenly-spaced partitioning.
 *
 * Revision 1.9  1994/01/17  20:28:41  macleod
 * macleod on Mon Jan 17 12:28:24 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Wasn't including thrshold.h
 *
 * Revision 1.8  1994/01/14  22:24:21  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Looks for max. power in an integrated histo region for background,
 * but looks for peak in a histo region for text
 *
 * Revision 1.7  1994/01/11  23:28:43  macleod
 * macleod on Tue Jan 11 15:28:05 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * added findThresholdsOnGray;
 * made inversion calculation depend on magnitudes of histogram
 * added parameter to findTextOnGray to return the histogram computed
 *
 * Revision 1.6  1994/01/08  04:36:39  macleod
 * macleod on Fri Jan 7 20:36:02 PST 1994
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Gray text/background finder and thresholder can work with
 * 4 bit images
 *
 * Revision 1.5  1993/12/03  01:14:57  ddavies
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.4  1993/11/24  22:00:59  macleod
 * macleod on Wed Nov 24 14:00:34 PST 1993
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Integrates into 32 bins rather than 16, to get better precision
 *
 * Revision 1.3  1993/11/13  02:06:09  macleod
 * macleod on Fri Nov 12 18:05:54 PST 1993
 *
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Added threshTextOnGray
 *
 * Revision 1.2  1993/11/13  00:50:52  macleod
 * fthresh.c -> /product/ipcore/ipwrap/src/RCS/fthresh.c,v
 *
 * Fixed a couple of warnings
 *
 * Revision 1.1  1993/11/10  01:23:48  macleod
 * Initial revision
 *
 */
