/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fthresh.c -- Heuristics for finding thresholding values, etc
 *
 * Exports:
 *
 *	findBgThreshAroundText	()
 *
 */


#include <stdio.h> /* 4debug */
#include <stdlib.h> /* 4debug */

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "logicops.h"
#include "defines.pub"
#include "thrshold.h"
#include "morphg.h"
#include "frames.pub"
#include "imageref.pub"
#include "shrscgr.pub"
#include "shrip.pub"
#include "imagerep.h"
#include "shrcc.pub"

/* prototypes */
#include "props.pub"
#include "ccthresh.pub"
#include "grbinop.pub"
#include "rescale.pub"
#include "logicops.pub"
#include "expand.pub"
#include "morph.pub"
#include "thrshold.pub"
#include "intops.pub"
#include "strel.pub"
#include "fthresh.pub"
#include "color.pub"
#include "reduce.pub"
#include "morphg.pub"
#include "curves.pub"
#include "prxform.pub"
#include "io.pub"
#include "shrip.pub"
#include "binarize.pub"
#include "imginfo.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: fthresh1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");

#define cSlope8Cutoff1		64  /* Was 32 */
#define cSlopeRCount1		 6   /* Max repeating slope count */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *      Find a threshold for the background around text.
 *
 * LongDescription:::
 *	Find a threshold for the background around text.
 *
 * PrivateDescription:::
 *      Using input text threshold as initial threshold to get text.
 *  Then analysis background threshold around text. 
 *
 * Input:::
 *  pIPWState::         State structure
 *  imageP::            Gray (8 bpp) image for which to find threshold
 *  rect::              If non-null, look inside this rectangle
 *  textThresh::        Initial threshold for analysis
 *
 * Output:::
 *  backIndex::         Text threshold computed  
 *
 * Returns:::
 *  error status
 * 
 * Notes:::
 *  Just work for 8 bit gray scale image.
 *
 * FunctionCategory:::
 *  threshold finding
 *
 * Keywords:::
 *  threshold finding, histogram
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod/zhu 
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findBgThreshAroundText (
    IPWState           *pIPWState,
    PIXR               *imageP,
    RECTANGL           *rect,
    Int32               textThresh,
    Int32              *backIndex)
{
    PROLOGUE;   

    Int32   *histo = NULL, *dx = NULL, returnCode;
    Int32   *intHisto = NULL, *startIndex = NULL, *endIndex = NULL,
	     depth, i, points, thresh, oneQuarter,
	     highestPoint, highestIndex, slopeCutoff,
	     currentCount, currentRepeats, width, height;
    PIXR    *result = NULL, *binImage = NULL,
	    *maskedImage = NULL;
    ScaledInt1616    xDpi, yDpi;
    RECTANGL lrect;
    STREL   *pStrel4h = NULL, *pStrel5 = NULL, *pStrel10v = NULL; 

    ERR_GEN_PIXR (imageP, cByteOrderBig, cDepth8, 
		  cChannels1 | cChannels3, cFrame32);

    ERR_PASS (w_strel1Inits(pIPWState));
    ERR_PASS (w_getStrel (pIPWState, strel4h, &pStrel4h));
    ERR_PASS (w_getStrel (pIPWState, strel5, &pStrel5));
    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel10v));

    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, imageP, &depth));
    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, imageP, 
					      &xDpi, &yDpi));

    if (rect == NULL)
    {
	rect = &lrect;
	rect->x = 0;
	rect->y = 0; 
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, imageP, &(rect->w)));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, imageP, &(rect->h)));
    }

    points = (1 << depth);
    slopeCutoff = cSlope8Cutoff1;
    oneQuarter = points / 4;
    thresh = textThresh;

    returnCode = cErrAborted;

    IP_YIELD;

    /* Threshold the picture */
    binImage = NULL;
    ERR_PASS_CLEANUP (w_binarize2 (pIPWState, imageP, 0, 28, &result));
    ERR_PASS_CLEANUP (w_textMask (pIPWState, result, &binImage, 0));
    ERR_PASS_CLEANUP (w_close (pIPWState, binImage, pStrel4h, &binImage));
    ERR_PASS_CLEANUP (w_close (pIPWState, binImage, pStrel10v, &binImage));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, binImage, pStrel5, &binImage));

    IP_YIELD;
	
    /* Don't need the result anymore */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &result));

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, imageP, &width));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, imageP, &height));
    
    if (rect->x == 0 && rect->y == 0 &&	rect->w == width && rect->h == height)
    {
	/* Right now, we only do masked histograms on full-size */
	ERR_PASS_CLEANUP (w_histoMask (pIPWState, imageP, binImage, NULL,
				       &histo));

	/* A hack: we get rid of the last bin. This 
	 * isn't "necessary" using masked histograms (it was
	 * when using opGrayBinPr) but the rest of the code
         * sort of assumes that it was zeroed. 
	 */
	histo[points - 1] = 0;
    }
    else
    {
	/* Do the mask operation on the image */
	ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, imageP, binImage, 
				       PIX_SRC & PIX_DST, &maskedImage));
	
	/* Set it's resolution, so that text thresholder knows what to do */
	ERR_PASS_CLEANUP (w_pixrSetResolutionDPI (pIPWState, maskedImage, 
					     xDpi, yDpi));

	ERR_PASS_CLEANUP (w_histo (pIPWState, maskedImage, rect, 
				   &histo));
	
	/* Blow off white */
	histo[points - 1] = 0;

    }

/*    DESTROY_PIXR (maskedImage); */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &binImage));
    binImage = NULL;

    /* Now, do it again with the new histogram */

    /* Get rid of missing codes */
    ERR_PASS_CLEANUP (w_interpolateMissingCodes (pIPWState, histo, points));

    /* Take the first derivative */
    dx = (Int32 *) IPW_CALLOC (points, sizeof (Int32), cAllocatorData);
    ERR_GEN_CLEANUP (dx == NULL, cErrNomem);

    ERR_PASS_CLEANUP (w_numericalDiff (pIPWState, histo, points, 1,  2, &dx));

    /* Find the peak */
    highestPoint = -1;
    highestIndex = 0;
    for (i = 0; i < points; i++)
    {
	if (histo[i] > highestPoint)
	{
	    highestPoint = histo[i];
	    highestIndex = i;
	}
    }


    for (i = highestIndex; i > 0 && dx[i] < 0; i--)
	;
    highestPoint = 0;
    
    for (; i > 0 && dx[i] > 0; i--)
    {
	if (dx[i] > highestPoint)
	    highestPoint = dx[i];
    }

    slopeCutoff = highestPoint / 4;

    if (highestIndex < oneQuarter)
    {
	/* Ugh: way too low. Just set it to some nominal value */
	*backIndex = textThresh;
    }
    else
    {
	/* Make sure we're going downhill */
	for (i = highestIndex; i > oneQuarter; i--)
	{
	    if ((depth == 4 && dx[i] > 0) ||	
		(dx[i] > slopeCutoff + 2))
		break;
	}
	
	/* Search backwards from the peak until we find the cutoff slope */
	currentCount = -1;
	currentRepeats = 0;
	for ( ; i > oneQuarter; i--)
	{
	    if ((dx[i] < slopeCutoff) && (dx[i-1] < slopeCutoff) && 
		(dx[i-2] < slopeCutoff))
		break;

	    /* psm 4/17/94: Don't let it get stuck in the same count forever */
	    if (dx[i] == currentCount)
	    {
		currentRepeats++;
		if (currentRepeats > cSlopeRCount1)
		    break;
	    }
	    else
	    {
		currentCount = dx[i];
		currentRepeats = 0;
	    }
	}

	/* Take the point we passed by; dx tends to be 1 off, hence + 2 */
	*backIndex = i + 2;
    }

    IP_YIELD;

    if (maskedImage) 
	ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &maskedImage));

    if (histo) IPW_FREE (histo, cAllocatorData);
    if (intHisto) IPW_FREE (intHisto, cAllocatorData);
    if (binImage) IPW_FREE (binImage, cAllocatorData);
    if (dx) IPW_FREE (dx, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (histo) IPW_FREE (histo, cAllocatorData);
    if (intHisto) IPW_FREE (intHisto, cAllocatorData);
    if (startIndex) IPW_FREE (startIndex, cAllocatorData);
    if (endIndex) IPW_FREE (endIndex, cAllocatorData);
    if (binImage) IPW_FREE (binImage, cAllocatorData);
    if (dx) IPW_FREE (dx, cAllocatorData);
    if (maskedImage)
	w_destroyPixr (pIPWState, &maskedImage);

    ERR_RETURN;
}

/*
 * $Log: fthresh1.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:22  jlin
 * Imported sources
 *
 * Revision 1.1  1998/06/30 17:36:58  xzhu
 * Initial revision
 *
 */
