/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * galois.c Multiplication and Arithmetic on Galois Field GF(256)
 *
 * From Mee, Daniel, "Magnetic Recording, Volume III", Ch. 5 by Patel.
 */

#include <stdio.h>
 

#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"


#define _GALOIS_C_DEFINED

#include "dgecc.prv"
#include "dgcom.prv"

#ifdef __GNUC__
#include "ansiprot.h"
#endif


IP_RCSINFO(RCSInfo, "$RCSfile: galois.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/* This code is dependent on the symbol size being a byte */

/****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * initialize galois tables needed before calling w_correctErrorsErasures
 *
 * PrivateDescription:::
 * Finite fields can be defined by polynomials with coefficients from
 * GF(2).  Alpha is a root of this polynomial.  Each element in a finite
 * field can be viewed as a power of alpha.  Furthermore each element 
 * can be represented as a combination of each of the eight powers of 
 * alpha.  These are represented in a binary vector with 1 indicating
 * this power of alpha is part of this element and 0 indicating that it 
 * is not.  We need tables that perform the log function - produce an
 * exponent from a vector (glog) and tables that perform the antilog
 * function - produce a vector from an exponent (gexp).  
 * This procedure initializes these tables.
 * Note that glog[0] should be negative infinity in that anything
 * added to negative infinity should equal negative infinity.
 * gexp[-infinity] = 0.  Initially, when the glog tables was made,
 * glog[0] was set to 0, which was bogus.  Every routine that was
 * going to use the glog table to see if it was about to take the
 * log of 0.  If so, it generally decided to do nothing at all since
 * it was probably going to do multiply that would produce 0.
 * In the new world, we set aside two bytes for each glog entry.
 * glog[0] is set to 511.  All the other glog entries range from 0 to
 * 255.  Therefore, the sum of any two normal glog table entries (this
 * is what we do when we multiply two numbers together) ranges from 0 to
 * 510.  Adding normal number to 511 produces numbers in the range 511 through
 * 766.  Adding glog[0] to glog[0] produces 1022.  The gexp table contains
 * 1023 bytes.  The first 255 are computed below.  The next 255 are
 * copies of the first 255 since we're doing remainder arithmentic.
 * The remaining entries are set to 0 so that multiplying any number
 * by 0 (adding its glog to glog[0]) produces 0.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  gexp::		antilog table: exponent to vector.
 *  glog::		log table: vector to exponent.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * EndDocumentation:::
 ***************************************************************************/
void CDECL
wp_initGaloisTables (
    IPWState	       *pIPWState,
    UInt8	       *gexp,
    UInt32	       *glog)
{
    Int32 i;
    UInt8 shiftReg;
    UInt8 feedback[2] = {0, 0x1D};	/* supplies x^4 + x^3 + x^2 + 1 term */
#define	cMinusInfinity	2*MAX_CODEWORD_SIZE + 1

	/* Keep compiler happy */
    pIPWState = pIPWState;

	/* Set glog of 0 to our equivalent of -infinity.  The maximum sum
	 * of any two normal logs is 2 * MAX_CODEWORD_SIZE.
	 */
    glog[0] = cMinusInfinity;

    /* Prime the shift register with:  00000001 */
    shiftReg = 1;

    for (i = 0; i < MAX_CODEWORD_SIZE; i++)
    {
	    /* This is one of 14 irreducible polynomials
	     * of degree 8 and cycle length 255.
	     * (Ch 5, pp. 275, Magnetic Recording)
	     * The high order 1 bit is implicit */
	    /* x^8 + x^4 + x^3 + x^2 + 1 */
	    /* Will need a new one for degree higher than 8!!!
	     * The "shiftReg << 1" term does a left shift.
	     * The array feedback has two elements, 0 and 00011101.
	     * We choose the correct feedback value by looking at the Most
	     * Significant bit of the shift register - JUST LIKE
	     * A REAL SHIFT REGISTER!
	     * Note that the gexp table is twice as big as necessary
	     * so that we can index it with glog[a] + glog[b]
	     * without doing MOD MAX_CODEWORD_SIZE (which is not
	     * quite a nice power of 2).
	     */
	gexp[i] = shiftReg;
	glog[shiftReg] = (UInt32) i;
	shiftReg = (shiftReg << 1) ^ feedback[shiftReg>>7];
    }
	/* At this point, we've initialized gexp entries 0 through
	 * MAX_CODEWORD_SIZE-1.  Since
	 * gexp[x] = gexp[x MOD MAX_CODEWORD_SIZE], duplicate every
	 * entry to fill gexp[MAX_CODEWORD_SIZE] through
	 * gexp[2*MAX_CODEWORD_SIZE - 1].
	 * gexp will be reference by the sum of at most two
	 * glog entries, so we need to duplicate the entries
	 * only once.
	 */
    memcpy((void *)(&gexp[MAX_CODEWORD_SIZE]), (void *)(&gexp[0]),
	   MAX_CODEWORD_SIZE);

	/* We've now filled gexp[0] through gexp[2*MAX_CODEWORD_SIZE - 1].
	 * Clear the remaining gexp entries.  This means that
	 * gexp[glog[anything] + glog[0]] = 0, or (anything * 0) = 0.
	 * All the remaining gexp table entries correspond to adding
	 * glog[0], or -infinity, to anything.
	 * How many remaining entries are there?  Well, there are
	 * 4 * (MAX_CODEWORD_SIZE + 1) entries in gexp.  We've done
	 * 2 * (MAX_CODEWORD_SIZE) of them so far.  That means
	 * there are 2*MAX_CODEWORD_SIZE + 4 left.
	 * 
	 */
    memset((void *)(&gexp[2*MAX_CODEWORD_SIZE]), 0,
			  2*MAX_CODEWORD_SIZE + 4);
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Polynomial multiplication
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  p1[]::       	multiplier polynomial    
 *  p2[]::              multiplicand polynomial
 *  len::            	order of the polynomial
 *  glog::		log table
 *  gexp::		antilog table.
 *
 * Output:::
 *  dst[]::          	product polynomial
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_multPolys (
    IPWState	       *pIPWState,
    UInt8               dst[],
    UInt8               p1[],
    UInt8               p2[],
    Int32               len,
    UInt32		glog[],
    UInt8		gexp[])
{
    UInt8	*dstPtr;
    UInt32	 p1Log[MAX_CODEWORD_SIZE+1], p2Log[MAX_CODEWORD_SIZE+1];
    Int32	 dIndex;
    Int32	 dstLen;
    Int32	 p1Index;
    UInt8	 result;
 
	/* Keep compiler happy */
    pIPWState = pIPWState;

    /* get the logs of the elements in the 2 polynomials to reduce the
     * number of reads in the inner loop
     */
    for (dIndex = 0; dIndex < len; dIndex++)
    {
	/* In the old code, we had to specially mark places where
	 * we tried to take the log of 0.  Since the glog table
	 * has been modified so that glog[0] = cMinusInfinity,
	 * we don't need to do that anymore.
	 */
	p1Log[dIndex] = glog[p1[dIndex]];
	p2Log[dIndex] = glog[p2[dIndex]];
    }
 
    /* the number of elements in the destination polynomial */
    dstLen = (len * 2) - 1;
    dstPtr = dst;
    /* index over the destination polynomial starting at the lowest
     * order term.  each element in the first half of the destination  
     * polynomial is contains in the product of the two source 
     * polynomials where one element is p1[0] (hence the inner loop
     * always starts at 0).
     */
    for (dIndex = 0; dIndex <= (len-1); dIndex++)
    {
        result = 0;
        for (p1Index = 0; p1Index <= dIndex; p1Index++)
            result ^= gexp[p1Log[p1Index] + p2Log[dIndex-p1Index]];
        *dstPtr++ = result;
    }
 
    /* finish looping over the destination polynomial.  unlike the above
     * loop, the inner loop here does not start referencing p1 at 0.
     * this is the reason for the difference in the two loop.s
     */
    for ( ; dIndex <= dstLen; dIndex++)
    {
        result = 0;
        for (p1Index = dIndex - (len-1); p1Index <= (len-1); p1Index++)
            result ^= gexp[p1Log[p1Index] + p2Log[dIndex-p1Index]];
        *dstPtr++ = result;
    }
}


/* multiplication using logarithms */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * multiplication using logarithms
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  a::             	multiplier 
 *  b::              	multiplicand
 *  glog::		log table
 *  gexp::		antilog table
 *
 * Returns:::
 *  product
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * EndDocumentation:::
 *****************************************************************************/
UInt8 CDECL
wp_gMultVal (
    IPWState	       *pIPWState,
    UInt8               a,
    UInt8               b,
    UInt32	       *glog,
    UInt8	       *gexp)
{
	/* Keep compiler happy */
    pIPWState = pIPWState;

    return( gexp[glog[a] + glog[b]] );
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the inverse of the given value
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  elt::		value to get inverse of
 *  glog::		log table
 *  gexp::		antilog table
 *
 * Returns:::
 *  inverse::		inverse of the given value
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * EndDocumentation:::
 *****************************************************************************/
UInt8 CDECL
wp_gInvVal (
    IPWState	       *pIPWState,
    UInt8               elt,
    UInt32	       *glog,
    UInt8	       *gexp)
{
	/* Keep compiler happy */
    pIPWState = pIPWState;

	/* This procedure computes M/elt in a finite field, where
	 * M is the biggest element in the finite field.
	 * This is equal to gexp[MAX_CODEWORD_SIZE - glog[elt]];
	 * Yeah, but what happens if elt is 0?  How to you express
	 * infinity in a finite field?  For the sake of consistency
	 * with previous code, we decided that MAX_CODEWORD_SIZE -
	 * glog[0] will be declared to be MAX_CODEWORD_SIZE.  This
	 * doesn't happen automatically since glog[0] is a special
	 * (big positive) value that normally indicates -infinity
	 * and subtracting it from MAX_CODEWORD_SIZE results in a 
	 * negative number that can't be used to address gexp.
	 * But, surprise! gexp[MAX_CODEWORD_SIZE] = gexp[0] because
	 * this is a finite field MOD MAX_CODEWORD_SIZE!  Therefore,
	 * we can do the simple thing of using MAX_CODEWORD_SIZE -
	 * glog[elt] unless this is negative - in which case we use
	 * 0.
	 */
    return(gexp[ MAX(0, ((Int32)MAX_CODEWORD_SIZE-(Int32)glog[elt])) ]);

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Add two polynomials
 *
 * LongDescription:::
 * Add destination polynomial to source polynomial and store result in
 * destination polynomial.  dst[idx] = src[idx] + dst[idx].
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  src::            	source polynomial
 *  len::            	order of polynomial
 *
 * Output:::
 *  dst::            	destination polynomial
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minski
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_addPolys (
    IPWState	       *pIPWState,
    UInt8              *dst,
    UInt8              *src,
    Int32               len)
{
    Int32	idx;

	/* Keep compiler happy */
    pIPWState = pIPWState;

    for (idx = 0; idx < len; idx++) dst[idx] ^= src[idx];
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Copy source polynomial to destination polynominal
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  src::            	source polynomial
 *  len::            	order of polynomial
 *
 * Output:::
 *  dst::            	destination polynomial
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_copyPoly (
    IPWState	       *pIPWState,
    UInt8              *dst,
    UInt8              *src,
    Int32               len)
{
	/* Keep compiler happy */
    pIPWState = pIPWState;

    memcpy((char *)dst, (char *)src, sizeof(*src)*len);
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Multiply a polynomial by a constant
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  k::              	scale factor
 *  len::            	order of polynomial
 *  glog::		log table
 *  gexp::		antilog table
 *
 * Output:::
 *  poly::           	polynomial to be scaled
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minski
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_scalePoly (
    IPWState	       *pIPWState,
    UInt8               k,
    UInt8              *poly,
    Int32               len,
    UInt32	       *glog,
    UInt8	       *gexp)
{
    Int32	idx;

	/* Keep compiler happy */
    pIPWState = pIPWState;

    for (idx = 0; idx < len; idx++) 
	poly[idx] = ERR_IGNORE(wp_gMultVal (pIPWState, k, poly[idx],
					    glog, gexp));
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set all values of a polynomial to zero
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  len::            	order of polynomial
 *
 * Output:::
 *  poly::           	polynomial to be cleared
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minski
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_zeroPoly (
    IPWState	       *pIPWState,
    UInt8              *poly,
    Int32               len)
{
	/* Keep compiler happy */
    pIPWState = pIPWState;

    memset((char *)poly, 0, sizeof(*poly)*len);
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Shift coefficients of polynomial one term to the right
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  len::            	order of polynomial
 *
 * Output:::
 *  poly::           	polynomial to be shifted
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Henry Minsky
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_shiftPoly (
    IPWState	       *pIPWState,
    UInt8              *poly,
    Int32               len)
{
    Int32	idx;

	/* Keep compiler happy */
    pIPWState = pIPWState;

    for (idx = len-1; idx > 0; idx--)
        poly[idx] = poly[idx-1];

    poly[0] = 0;
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Calculate generator polynomial for a Reed-Solomon code.
 *
 * PrivateDescription:::
 * Construct the generator polynomial of a Reed-Solomon code.  This
 * polynominal is (x + alpha^1) * (x + alpha^2) * ... (x + alpha^m)
 * where m is the number of check symbols for this codeword.
 * Note that Cain and Clark start at alpha^1 and continue to alpha^m while 
 * Glover and Dudley start at alpha^0 and continue to alpha^(m-1).
 * At this point I don't understand the difference and/or what the 
 * ramifications are.
 * For reasons of efficiency when the polynomial is used, this routine
 * really computes the log of each coefficient in the polynomial.  That's
 * why the returned values are UInt32's.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  nbytes::		number of bytes of polynomial
 *  glog::		log table
 *  gexp::		antilog table
 *
 * Output:::
 *  logGenpoly::	array of logs of coefficents of generator polynomial
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 * Practical Error Correction Design for Engineers, 
 *	Neal Glover and Trent Dudley
 * Error-Correction Coding for Digitial Communications
 *	George C. Clark, Jr. and J. Bibb Cain
 * 
 * EndDocumentation:::
 ***************************************************************************/
void CDECL
wp_computeGenpoly (
    IPWState	       *pIPWState,
    Int32               nbytes,
    UInt32	       *glog,
    UInt8	       *gexp,
    UInt32             *logGenpoly)
{
    Int32	i, j;
    UInt8	genpoly[MAX_CODEWORD_SIZE+1];

	/* Keep compiler happy */
    pIPWState = pIPWState;

    /* compute (x + a^1) * (x + a^2) * ...  * (x + a^nbytes) */
    /* To do the iterative multiply, we initialize genpoly to
     * 1.  Inititializing it to 0 would produce a product of 0,
     * right?
     */
    genpoly[0] = 1;

    for (i = 1; i <= nbytes; i++)
    {
	/* Multiplying a polynomial by x + a^i is pretty easy.
	 * Originally, we used wp_multPolys, which is enormous
	 * overkill.  It assumed that both polynomials were
	 * of the maximum length (nbytes + 1).  That's seldom
	 * true of the dest polynomial and never true of the
	 * source.
	 * So what do we do?  Well, multiplying a polynomial
	 * by x is easy, you just shift it left.  That is,
	 * dest[i] = src[i-1].  Multiplying a polynomial by
	 * a^i is also pretty easy, you just multiply every
	 * coefficient by a^i.  Therefore:
	 *	genpoly[j] = (genpoly[j] * a^i) + genpoly[j-1]
	 * How do we translate this into galois math?  Well,
	 * first, note that a^i is just gexp[i].  That's what
	 * the gexp table does, convert i into a^i.  Second,
	 * addition becomes xor, so we get
	 *	genpoly[j] = (genpoly[j] * gexp[i]) ^ genpoly[j-1]
	 * Multiplication is done by adding logarithms, then
	 * taking the exponent:
	 *	genpoly[j] = gexp[ glog[genpoly[j]] + glog[gexp[i]] ] ^
	 *			genpoly[j-1]
	 * Since glog[gexp[i]] = i, we can simplify this to:
	 *	genpoly[j] = gexp[ glog[genpoly[j]] + i ] ^ genpoly[j-1]
	 * When j == 0, we just discard the genpoly[j-1] term.
	 *
	 * The old code indexed the polynomials from the LS coefficients
	 * towards the MS coefficients.  That meant that
	 * coefficients were overwritten before they were used, so
	 * it used a temporary polynomial (well, it really used
	 * two).  If we do the computation starting with the MS
	 * coefficients and work down towards the LS coefficients,
	 * we won't write over any coefficients before we need them.
	 *
	 * Note also that there is no genpoly[i] term at this
	 * stage, so we can't multiply it by a^i.  Therefore,
	 * genpoly[i] simply equals genpoly[i-1] to account for the
	 * multiply by x.
	 *
	 */
	genpoly[i] = genpoly[i-1];
	for (j = i-1; j > 0; j--)
	    genpoly[j] = gexp[ glog[genpoly[j]] + i] ^ genpoly[j-1];
	genpoly[0] = gexp[ glog[genpoly[0]] + i];


#if 0	/* this is the way we used to do it */

	/* jac. 3/6/96. tp does not get modified by wp_multPolys so 
	 * there is no need to zero it out each time though the loop.
	 */
	/* ERR_IGNORE(wp_zeroPoly (pIPWState, tp, nbytes+1)); */
	tp[0] = gexp[i];  /* set up x+a^n */
	tp[1] = 1;

	/* note that wp_multPolys operates as it it were multiplying
	 * 2 polynomials of degree nbytes+1.  in this particular case
	 * we could speed up this code by noting the degrees of the
	 * polynomials that are actually being multiplied.
	 * jac. 3/6/96.
	 */
	ERR_IGNORE(wp_multPolys (pIPWState, genpoly, tp, tp1,
				 nbytes+1, glog, gexp));
	ERR_IGNORE(wp_copyPoly (pIPWState, tp1, genpoly, nbytes+1));
#endif

    }

    /* Optimization!  The only thing we do with genpoly is multiply
     * by it.  Therefore, we want to store glog[genpoly[j]]
     * instead of genpoly[j] so we can avoid applying glog to
     * each element before we use it later.
     * Therefore, we go through the array converting each
     * element once.  Note that we couldn't do this before
     * because we needed the real values when computing the
     * polynomial.
     * Note also that we never really access genpoly[nbytes]
     * again.  It was just there so we would have a place
     * to put the MS term when doing the last computation step.
     */
    for (i = 0; i < nbytes; i++)
    {
	logGenpoly[i] = glog[genpoly[i]];
    }

    /* And, the code that uses genpoly to encode codewords wants
     * to read it four bytes at a time starting one byte after any index.
     * Therefore, we need to copy the first four entries to just
     * beyond the last entry.  The array was allocated large enough
     * to account for this.
     */
    logGenpoly[nbytes] = logGenpoly[0];
    logGenpoly[nbytes+1] = logGenpoly[1];
    logGenpoly[nbytes+2] = logGenpoly[2];
    logGenpoly[nbytes+3] = logGenpoly[3];
}

/*

 $Log: galois.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.30  1998/01/14 00:12:04  ddavies
 galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.29  1996/03/24 21:16:58  ddavies
 galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v

 Fix auto-doc comments.  Add to comment explaining operation of
 wp_gInvVal, in particular when attempting to divide by 0.

 Revision 1.28  1996/03/21 21:39:18  ddavies
 galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v

 Changed all the glog tables to be UInt32s from UInt8s.  Put cMinusInfinity
 into glog[0] instead of putting 0 there.  Changed wp_gMultVal to a one-liner
 since we don't need to check for 0 now.  Changed wp_gInvVal so that it
 checks for 0 (now needed).  We do lots more multiplies than divides, so this
 is ok.  Changed the code that generates glog so that it generates UInt32s
 and makes a bigger gexp table that allows one to get
 gexp[glog[anything]+glog[0]] = 0 without any extra checking.
 Also generally changed all polynomial coefficients into UInt8s from Int32's.

 Revision 1.27  1996/03/13 21:41:09  ddavies
 ddavies on Wed Mar 13 13:39:24 PST 1996

 galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v

 Improved wp_initGaloisTables.  It now uses a real shift register with
 1 bit/bit (instead of 1 byte/bit).  It also computes glog and gexp at
 the same time instead of computing gexp, then recomputing glog.

 Revision 1.26  1996/03/08 22:23:38  jcraig
 jcraig on Fri Mar 8 14:22:52 PST 1996

 galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v

 update comments.  don't clear the factor polynomial each time thru
 the loop in wp_computeGenpoly

 Revision 1.25  1996/01/04 20:22:55  ejaquith
 galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.24  1995/12/13  04:07:37  ejaquith
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.23  1995/11/09  01:19:00  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * CommonCheck: wrap ERR_IGNORE around all private procs which do not return an
 * ErrCode
 *
 * Revision 1.22  1995/11/08  22:31:58  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * make all the code mucks with polynomials consistently return void
 *
 * Revision 1.21  1995/11/07  21:30:46  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * ecc tables are now passed as parameters instead of being global vars
 *
 * Revision 1.20  1995/10/13  22:00:40  ejaquith
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.19  1995/09/29  00:40:35  ejaquith
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * Fixed comment in comment typo
 *
 * Revision 1.18  1995/09/27  23:41:53  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * new procedure comments
 *
 * Revision 1.17  1995/09/22  00:38:23  ejaquith
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * CommonCheck: Fixed up wp_ issues.
 *
 * Revision 1.16  1995/09/21  22:53:20  ejaquith
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/07  00:52:34  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * dgi prefix becomes w_
 *
 * Revision 1.14  1995/09/06  20:19:21  ejaquith
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.13  1995/09/01  02:12:13  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.12  1995/08/31  21:13:10  ejaquith
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.11  1995/08/17  22:29:11  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.10  1994/10/12  20:47:07  jcraig
 * jcraig on Wed Oct 12 13:45:43 PDT 1994
 *
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * use Int32 instead of long in declarations
 *
 * Revision 1.9  1994/09/28  21:11:05  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 *  change edd prefix to dg.
 *
 * Revision 1.8  1994/05/02  20:26:10  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * get rid of warnings generated when compiling with gcc
 *
 * Revision 1.7  1994/03/25  23:27:00  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * rewrite eddMultPolys in order to speed it up.  loop over the
 * destination polynomial, accessing all the elements in the source
 * polynomial that are necessary to compute the current destination
 * index, writing each destination index only once.
 *
 * Revision 1.6  1994/01/27  00:50:50  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * include eddcom.h and eddecc.h instead of libglyph.prv.
 * functions internal to the library but that are called from other
 * modules have eddi not edd prepended.
 *
 * Revision 1.5  1994/01/15  03:57:31  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * removed include of libglyph.h - now only libglyph.prv is included.
 *
 * Revision 1.4  1993/12/03  01:15:03  ddavies
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.3  1993/11/10  02:36:57  jcraig
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * procedures that are not private to this module have edd prepended.
 *
 * Revision 1.2  1993/11/06  03:12:21  ddavies
 * galois.c -> /product/ipcore/ipwrap/src/RCS/galois.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

