/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * geoutil.c
 *
 *	A few utilities for doing geometric calculations
 *
 *	geo_pointOnLineSeg	Find a point on a line seg closest to the one passed
 *	geo_distanceToLineSeg	Calculate the min.distance between the line seg
 *				and the point passed
 *	geo_intersect		Find the intersection of two lines of 
 *				the form y = a + bx
 *	geo_colinear		Determine whether two lines are approximately
 *				colinear
 *	my_sqrt			For fun-with-gcc-and-purify
 *	solveL1			L1 solution to a system of linear equations
 *				subject to equality and inequality constraints
 *				(only sort of a "geometry" utility)
 *	linearFit		Interface to solveL1 for linear regression
 */

#include <math.h>


#include "ipwcore.prv"
#include "defines.pub"


#include "geoutil.pub"
#include "nrutil.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: geoutil.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

#ifndef M_PI_2
/* Watcom doesn't define any math constants! The following is from the SUN math.h */
#define	M_PI_2		1.57079632679489661923	/* pi/2 */
#endif /* M_PI_2 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * a square-root function to fake out purify w/gcc (ugh!)
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  n::              	number
 *
 * Output:::
 *  ans::             	square root of n
 *
 * Returns:::
 *  cSuccessful		successfully calculated square root
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_my_sqrt (
    IPWState           *pIPWState,
    Float64             n,
    Float64            *ans)

{
    PROLOGUE;
    pIPWState = pIPWState;
    *ans = sqrt(n);
    NORMAL_RETURN;
}

/*
 * geo_pointOnLineSeg
 *
 *	Finds the point on a line segment closest to the point passed.
 *	Does it by treating the segment as an equation for the line,
 *	and finds the perpindicular, clamping to the endpoints. Note that
 *	it computes the point on the segment, so one could return it if 
 *	one wanted. One should be warned that the point is returned as
 *	a pair of integers, so one might want to be careful about errors
 *	in small coordinate spaces.
 *
 * Arguments:
 *	x0, y0, x1, y1 =    Points describing line segment
 *	qx, qy =	    Point of interest
 *	px, py =	    Point computed.
 * 
 * Returns:
 *	cSuccessful
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Finds the point on a line segment closest to the point passed.
 *
 * LongDescription:::
 * Finds the point on a line segment closest to the given point, 
 * by treating the segment as an equation for the line,
 * and finds the perpindicular, clamping to the endpoints.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  x0::             	first x coordinate describing line segment
 *  y0::             	first y coordinate describing line segment
 *  x1::             	second x coordinate describing line segment
 *  y1::             	second y coordinate describing line segment
 *  qx::             	x coordinate of point of interest
 *  qy::             	y coordinate of point of interest
 *
 * Output:::
 *  px::             	x coordinate of closest point
 *  py::             	y coordinate of closest point
 *
 * Returns:::
 *
 * Notes:::
 * This proc computes the point on the segment, so one could return it if 
 * one wanted. One should be warned that the point is returned as
 * a pair of integers, so one might want to be careful about errors
 * in small coordinate spaces.
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_pointOnLineSeg (
    IPWState           *pIPWState,
    Int32               x0,
    Int32               y0,
    Int32               x1,
    Int32               y1,
    Int32               qx,
    Int32               qy,
    Int32              *px,
    Int32              *py)
{
    PROLOGUE;
    Float64 u; 
    Int32 un, ud;

    pIPWState = pIPWState;

    /*
     * What we want is the point perpindicular to the line and qx,qy.
     * If we consider the endpoints of the line as vectors from the origin,
     * say, p0, p1, and another vector from the origin to qx,qy to be q,
     * we get the minimum distance from q to the line at point p when:
     *
     *	    (p0 + u(p1 - p0) - q) . (p1 - p0) = 0
     *
     * This is expanded to the linear equation:
     * (x0 + u(x1 - x0) - qx)(x1 - x0) + (y0 + u(y1 - y0) - qy)(y1 - y0) = 0
     *
     * Then, solve for u and plug it back into the line to get the point
     * on the line. There are probably more efficient ways of doing this.
     */

    un = qx*x0 - x0*x0 - qx*x1 + x0*x1 + qy*y0 - y0*y0 - qy*y1 + y0*y1;
    ud = x0*x0 - 2*x0*x1 + x1*x1 + y0*y0 - 2*y0*y1 + y1*y1;

    /* 
     * I'm assuming that above product terms will be within 32-bit range.
     * That only holds true if coordinates are less than around 10 or 20 thousand,
     * so one might want to be careful for large rasters. If coordinates are
     * larger, use floats and pay the price.
     */

    u = -((Float64)un/ud);

    /*
     * If u is between 0 and 1, we're on the segment. If not, clamp to an endpoint 
     * so that we get the distance to the segment, not the (infinite) line 
     */
    if (u < 0.0) u = 0.0;
    if (u > 1.0) u = 1.0;

    *px = (Int32) (x0 + ((Float64)(x1 - x0) * u) + 0.5);
    *py = (Int32) (y0 + ((Float64)(y1 - y0) * u) + 0.5);
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Finds the point on a line segment closest to the point passed.
 *
 * LongDescription:::
 * This procedure is the same as w_geo_pointOnLineSeg except that it
 * doesn't clamp to the segment -- the points can line anywhere on the 
 * line described by the segment.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  x0::             	first x coordinate describing line segment
 *  y0::             	first y coordinate describing line segment
 *  x1::             	second x coordinate describing line segment
 *  y1::             	second y coordinate describing line segment
 *  qx::             	x coordinate of point of interest
 *  qy::             	y coordinate of point of interest
 *
 * Output:::
 *  px::             	x coordinate of closest point
 *  py::             	y coordinate of closest point
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_pointOnLine (
    IPWState           *pIPWState,
    Int32               x0,
    Int32               y0,
    Int32               x1,
    Int32               y1,
    Int32               qx,
    Int32               qy,
    Int32              *px,
    Int32              *py)
{
    PROLOGUE;
    Float64 u; 
    Int32 un, ud;

    pIPWState = pIPWState;
    /* See geo_pointOnLineSeg for a description */
    un = qx*x0 - x0*x0 - qx*x1 + x0*x1 + qy*y0 - y0*y0 - qy*y1 + y0*y1;
    ud = x0*x0 - 2*x0*x1 + x1*x1 + y0*y0 - 2*y0*y1 + y1*y1;
    u = -((Float64)un/ud);

    *px = (Int32) (x0 + ((Float64)(x1 - x0) * u) + 0.5);
    *py = (Int32) (y0 + ((Float64)(y1 - y0) * u) + 0.5);
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Computes distance between a point and the closest point on a line segment
 *
 * LongDescription:::
 * Finds the point on a line segment closest to the point passed.
 * Computes the distance between the point passed and that point on the
 * line segment.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  x0::             	first x coordinate describing line segment
 *  y0::             	first y coordinate describing line segment
 *  x1::             	second x coordinate describing line segment
 *  y1::             	second y coordinate describing line segment
 *  qx::             	x coordinate of point of interest
 *  qy::             	y coordinate of point of interest
 *
 * Output:::
 *  distance::       	distance to closest point on line segment
 *
 * Returns:::
 *
 * Notes:::
 * This proc computes the point on the segment, so one could return it if
 * one wanted. One should be warned that the point is returned as
 * a pair of integers, so one might want to be careful about errors
 * in small coordinate spaces.
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_distanceToLineSeg (
    IPWState           *pIPWState,
    Int32               x0,
    Int32               y0,
    Int32               x1,
    Int32               y1,
    Int32               qx,
    Int32               qy,
    Float64            *distance)
{
    PROLOGUE;
    Float64 dx, dy;
    Int32   px, py;

    ERR_PASS( w_geo_pointOnLineSeg(pIPWState, x0, y0, x1, y1, qx, qy, &px, &py));
    dx = px - qx;
    dy = py - qy;
    ERR_PASS( w_my_sqrt (pIPWState, dx*dx + dy*dy, distance));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Computes distance between a point and the closest point on the line
 *
 * LongDescription:::
 * This procedure is similar to w_geo_distanceToLineSeg except that it 
 * finds the distance to the line, not just a line segment.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  x0::             	first x coordinate describing line segment
 *  y0::             	first y coordinate describing line segment
 *  x1::             	second x coordinate describing line segment
 *  y1::             	second y coordinate describing line segment
 *  qx::             	x coordinate of point of interest
 *  qy::             	y coordinate of point of interest
 *
 * Output:::
 *  distance::       	distance to closest point on line
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_distanceToLine (
    IPWState           *pIPWState,
    Int32               x0,
    Int32               y0,
    Int32               x1,
    Int32               y1,
    Int32               qx,
    Int32               qy,
    Float64            *distance)
{
    PROLOGUE;
    Float64 dx, dy;
    Int32   px, py;

    ERR_PASS( w_geo_pointOnLine(pIPWState, x0, y0, x1, y1, qx, qy, &px, &py));
    dx = px - qx;
    dy = py - qy;
    ERR_PASS( w_my_sqrt(pIPWState, dx*dx + dy*dy, distance));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Computes the square of the distance to the closest point on a line segment
 *
 * LongDescription:::
 * Finds the point on a line segment closest to the point passed.
 * Computes the square of the distance between the point passed and that 
 * point on the line segment. We use the square rather than the distance 
 * itself in cases where we're doing a minimization, and we would just
 * square it again.
 * Does it by treating the segment as an equation for the line,
 * and finds the perpindicular, clamping to the endpoints.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  x0::             	first x coordinate describing line segment
 *  y0::             	first y coordinate describing line segment
 *  x1::             	second x coordinate describing line segment
 *  y1::             	second y coordinate describing line segment
 *  qx::             	x coordinate of point of interest
 *  qy::             	y coordinate of point of interest
 *
 * Output:::
 *  distance::       	distance squared to closest point on line
 *
 * Returns:::
 *
 * Notes:::
 * This proc computes the point on the segment, so one could return it if 
 * one wanted. One should be warned that the point is returned as
 * a pair of integers, so one might want to be careful about errors
 * in small coordinate spaces.
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_distanceToLineSegSquared (
    IPWState           *pIPWState,
    Int32               x0,
    Int32               y0,
    Int32               x1,
    Int32               y1,
    Int32               qx,
    Int32               qy,
    Float64            *distance)
{
    PROLOGUE;
    Float64 dx, dy;
    Int32   px, py;

    ERR_PASS(w_geo_pointOnLineSeg(pIPWState, x0, y0, x1, y1, qx, qy, &px, &py));
    dx = px - qx;
    dy = py - qy;
    *distance = dx*dx + dy*dy;
    NORMAL_RETURN;
}


#define EPS 1.0e-6
#define DET2X2(a,b,c,d) (a)*(d)-(b)*(c)
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Find the point of intersection of two lines.
 *
 * LongDescription:::
 * Find the point of intersection of two lines. The lines are
 * in the form ax + by = c.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  a1::             	x coefficient of first line equation
 *  b1::             	y coefficient of first line equation
 *  c1::             	constant of first line equation
 *  a2::             	x coefficient of second line equation
 *  b2::             	y coefficient of second line equation
 *  c2::             	constant of second line equation
 *
 * Output:::
 *  x::              	x coordinate of point of intersection
 *  y::              	y coordinate of point of intersection
 *  linesParallel::  	are lines parallel
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_intersect (
    IPWState           *pIPWState,
    Float64             a1,
    Float64             b1,
    Float64             c1,
    Float64             a2,
    Float64             b2,
    Float64             c2,
    Float64            *x,
    Float64            *y,
    Bool               *linesParallel)
{
    PROLOGUE;
    Float64 d1, d2, dd;

    pIPWState = pIPWState;
    dd = DET2X2 (a1, b1, a2, b2);

    if (fabs(dd) < EPS)
	/* Parallel */
	*linesParallel = cTrue;
    else
    {
    	d1 = DET2X2 (c1, b1, c2, b2);
    	d2 = DET2X2 (a1, c1, a2, c2);
    	/* Cramer's rule */
    	*x = d1 / dd;
    	*y = d2 / dd;
    	*linesParallel = cFalse;
    }
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Compute the rms distance between two lines at two points along the lines
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  a0::             	x coefficient of first line equation
 *  b0::             	y coefficient of first line equation
 *  c0::             	constant of first line equation
 *  a1::             	x coefficient of second line equation
 *  b1::             	y coefficient of second line equation
 *  c1::             	constant of second line equation
 *  p0::             	first ordinate to test
 *  p1::             	second ordinate to test
 *  vertical::       	are lines vertical?  if so, ordinates are the
 *			y axis.  if not, ordinates are on the x axis
 *
 * Output:::
 *  rmsDistance::    	rms distance
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_rmsDistance (
    IPWState           *pIPWState,
    Float64             a0,
    Float64             b0,
    Float64             c0,
    Float64             a1,
    Float64             b1,
    Float64             c1,
    Int32               p0,
    Int32               p1,
    Int32               vertical,
    Float64            *rmsDistance)
{
    PROLOGUE;
    Float64 q0line0, q0line1, q1line0, q1line1;
    Float64 distance0, distance1;

    if (vertical)
    {
	if (fabs(a0) < 1.0e-20 || fabs(a1) < 1.0e-20)
	{
	    /* fprintf(stderr, "huh? odd case # 1\n"); */
	    *rmsDistance = 1.0e+6;
	    NORMAL_RETURN;
	}

	/* Solving for x */
	q0line0 = (c0 - b0 * p0) / a0;
	q0line1 = (c1 - b1 * p0) / a1;
	q1line0 = (c0 - b0 * p1) / a0;
	q1line1 = (c1 - b1 * p1) / a1;
    }
    else
    {
	if (fabs(b0) < 1.0e-20 || fabs(b1) < 1.0e-20)
	{
	    /* fprintf(stderr, "huh? odd case # 2\n"); */
	    *rmsDistance = 1.0e+6;
	    NORMAL_RETURN;
	}
	/* Solving for y */
	q0line0 = (c0 - a0 * p0) / b0;
	q0line1 = (c1 - a1 * p0) / b1;
	q1line0 = (c0 - a0 * p1) / b0;
	q1line1 = (c1 - a1 * p1) / b1;
    }

    /* Compute the distances (or -distances) */
    distance0 = q0line1 - q0line0;
    distance1 = q1line1 - q1line0;

/* fprintf(stderr, "distance 0 %f, distance 1 %f\n", distance0, distance1); */

    /* Compute the RMS distance */
    ERR_PASS( w_my_sqrt (pIPWState, 
		      (distance0 * distance0 + distance1 * distance1) / 2.0,
		      rmsDistance));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Determine if two lines are colinear
 *
 * LongDescription:::
 * There are a lot of ways to define this. You could talk about the 
 * angle between lines, but since we're dealing with crop boxes
 * with usually nearly-parallel sides, lines would almost always
 * intersect *somewhere* at some small angle. So one way to tell
 * if two lines are colinear is to take the mean of the sum of the 
 * distance^2 between the lines at two strategic points, like the 
 * extent of the image when cropping.  The lines are in the form ax+by=c.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  a0::             	x coefficient of first line equation
 *  b0::             	y coefficient of first line equation
 *  c0::             	constant of first line equation
 *  a1::             	x coefficient of second line equation
 *  b1::             	y coefficient of second line equation
 *  c1::             	constant of second line equation
 *  p0::             	first ordinate to test
 *  p1::             	second ordinate to test
 *  vertical::       	are lines vertical?  if so, ordinates are the
 *			y axis.  if not, ordinates are on the x axis
    
 *  threshold::       	threshold of rms distance below which lines are
 *			considered colinear
 *
 * Output:::
 *  colinear::       	are the lines conlinear?
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_colinear (
    IPWState           *pIPWState,
    Float64             a0,
    Float64             b0,
    Float64             c0,
    Float64             a1,
    Float64             b1,
    Float64             c1,
    Int32               p0,
    Int32               p1,
    Int32               vertical,
    Float64             threshold,
    Bool               *colinear)
{
    PROLOGUE;
    Float64	 rmsDistance;

    ERR_PASS( w_geo_rmsDistance(pIPWState, a0, b0, c0, a1, b1, c1, p0, p1,
			     vertical, &rmsDistance));
    if (rmsDistance < threshold)
	*colinear = cTrue;
    else
	*colinear = cFalse;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Translate a line along its perpendicular.
 *
 * LongDescription:::
 * The distance moved is along a perpindicular from (0,0) to the line,
 * so a positive distance moves further from the origin, and a negative
 * distance moves closer to the origin.
 * This can be accomplished by realizing that the ratio of the 
 * distance moved to the distance from the origin to the line (the
 * perpendicular by definition) is equal to the ratio of the distance
 * between the y-intercept and the delta-y-intercept, which is what
 * we want to compute so that we can apply it. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  a::              	x coefficient of line equation
 *  b::              	y coefficient of line equation
 *  c::              	constant of line equation
 *  d::              	distance to move
 *
 * Output:::
 *  a1::             	x coefficient of new line equation
 *  b1::             	y coefficient of new line equation
 *  c1::             	constant of new line equation
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_geo_movePerpindicular (
    IPWState           *pIPWState,
    Float64             a,
    Float64             b,
    Float64             c,
    Float64             d,
    Float64            *a1,
    Float64            *b1,
    Float64            *c1)
{
    PROLOGUE;
    Float64 offset, slope, intercept, theta;

    pIPWState = pIPWState;
    /* Translate the line */
    if (a == 0.0)
    {
	/* by = c, so the line is vertical */
	*a1 = a;
	*b1 = b;
	*c1 = (c/b + d) * b;
    }
    else if (b == 0.0)
    {
	/* ay = c, so the line is horizontal */
	*a1 = a;
	*b1 = b;
	*c1 = (c/a + d) * a;
    }
    else
    {
	slope = -(a) / b;
	intercept = c / b;

	/* Compute the angle of the line in standard position */
	theta = atan (slope);

	/* Compute the offset relative to the y-intercept */
	offset = d * sin(M_PI_2 - theta);
	intercept = intercept + offset;

	/* Stuff it back */
	*a1 = -slope;
	*c1 = intercept;
	*b1 = 1.0;
    }

    NORMAL_RETURN;
}   

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Calculate a solution to system of linear equations with constraints
 *
 * LongDescription:::
 * This subroutine uses a modification of the simplex
 * method of linear programming to calculate an l1 solution
 * to a k by n system of linear equations ax=b
 * subject to l linear equality constraints cx=d
 * and m linear inequality constraints ex <= f
 *
 * PrivateDescription:::
 * The code has lots of gotos.  Many of them could be edited out fairly
 * easily; I didn't see a pressing need--the code is fairly "read-only" 
 * anyway.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  k::              	Number of rows of the matrix a (k >= 1)
 *  l::              	Number of rows of the matrix c (l >= 0).
 *  m::              	Number of rows of the matrix e (m >= 0).
 *  n::              	Number of columns of the matrices a,c,e (n >= 1).
 *  q::              	Two dimensional Float64 **array with klm2d rows and
 *			at least n+2 columns.
 *			NOTE: this is a "numerical-recipes"-style array,
 *			as allocated by w_nr_matrix(), which means that it
 *			is an array of pointers to 1D arrays. The first 
 *			index starts with 1, not zero. See linearFit below 
 *			for an example of how to allocate and use these 
 *			matrices. On entry the matrices a,c and e, and the 
 *			vectors b,d and f must be stored in the first k+l+m 
 *			rows and n+1 columns of q as follows
 *			    a b
 *			q = c d
 *		 	    e f
 *			these values are destroyed by the subroutine.
 *  toler::          	A small positive tolerance. Empirical
 *			evidence suggests toler = 10**(-d*2/3),
 *			where d represents the number of decimal
 *			digits of accuracy available. Essentially,
 *			the subroutine cannot distinguish between zero
 *			and any quantity whose magnitude does not exceed
 *			toler. In particular, it will not pivot on any
 *			number whose magnitude does not exceed toler.
 *
 * Output:::
 *  kode::           	A code used on entry to, and exit from, the
 *			subroutine.  on entry, this should normally be set 
 *			to 0. However, if certain nonnegativity constraints
 *			are to be included implicitly, rather than
 *			explicitly in the constraints ex <= f, then kode
 *			should be set to 1, and the nonnegativity
 *			constraints included in the arrays x and
 *			res (see below).
 *			on exit, kode has one of the
 *			following values
 *		    		0-	optimal solution found,
 *		    		1-	no feasible solution to the
 *					constraints,
 *		    		2-	calculations terminated
 *					prematurely due to rounding errors,
 *		    		3-  maximum number of iterations reached.
 *  iter::           	On entry iter must contain an upper bound on
 *			the maximum number of iterations allowed.
 *			A suggested value is 10*(k+l+m). On exit iter
 *			gives the number of simplex iterations.
 *  x::              	One dimensional Float64 array of size at least n+2.
 *			On exit this array contains a
 *			solution to the l1 problem. If kode=1
 *			on entry, this array is also used to include
 *			simple nonnegativity constraints on the
 *			variables. The values -1, 0, or 1
 *			for x(j) indicate that the j-th variable
 *			is restricted to be  <= 0, unrestricted,
 *			or  >= 0 respectively.
 *			NOTE: This vector is a "numerical recipes style" 
 *			vector, as allocated by nr_vector. The first index
 *			is 1 (which means that the 0th element is wasted).
 *			See w_linearFit() below for an example of how to
 *			create, use, and deallocate such a vector.
 *  res::            	One dimensional Float64 array of size at least 
 *			(k+l+m). On exit this contains the residuals b-ax
 *			in the first k components, d-cx in the
 *			next l components (these will be =0),and
 *			f-ex in the next m components. If kode=1 on
 *			entry, this array is also used to include simple
 *			nonnegativity constraints on the residuals
 *			b-ax. The values -1, 0, or 1 for res(i)
 *			indicate that the i-th residual (1 <= i <= k) is
 *			restricted to be  <= 0, unrestricted, or  >= 0
 *			respectively. "Numerical-recipes" style vector.
 *  error::          	the minimum sum of absolute values of the residuals.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *  This was machine translated from Fortran by f2c.
 *  The origin is TOMS #552. (from ACM TOMS 6 (1980), pp. 236-239, by 
 *  I. Barrodale and F.D.K. Roberts). The code was uploaded from
 *  netlib/toms.  It was called "cl1" there.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_solveL1 (
    IPWState           *pIPWState,
    Int32               k,
    Int32               l,
    Int32               m,
    Int32               n,
    Float64           **q,
    Float64             toler,
    Int32              *kode,
    Int32              *iter,
    Float64            *x,
    Float64            *res,
    Float64            *error)
{
    PROLOGUE;
    /* System generated locals */
    Float64         r__1;

    /* Local variables */
    Int32     iimn, nklm;
    Float64   xmin, xmax;
    Int32     iout = 0, i, j;
    Float64   z;
    Int32     iineg, maxit, n1, n2;
    Float64   pivot;
    Int32     ia, ii, kk, in = 0, nk, js;
    Float64   sn;
    Int32     iphase, kforce;
    Float64   zu, zv;
    Int32     nk1;
    Float64   tpivot;
    Int32     klm, jmn, nkl, jpn;
    Float64   cuv;
    Float64   sum;
    Int32     klm1, klm2, nkl1;
    Float64 **cu = NULL;
    Int32   **iu = NULL, *s = NULL;
    
    /* initialization. */
    ERR_PASS( w_nr_matrix(pIPWState, 1, 2, 1, n+k+l+m, &cu));
    ERR_PASS_CLEANUP( w_nr_imatrix(pIPWState, 1, 2, 1, n+k+l+m, &iu));
    ERR_PASS_CLEANUP( w_nr_ivector(pIPWState, 1, k+l+m, &s));

    /* Function Body */
    maxit = *iter;
    n1 = n + 1;
    n2 = n + 2;
    nk = n + k;
    nk1 = nk + 1;
    nkl = nk + l;
    nkl1 = nkl + 1;
    klm = k + l + m;
    klm1 = klm + 1;
    klm2 = klm + 2;
    nklm = n + klm;
    kforce = 1;
    *iter = 0;
    js = 1;
    ia = 0;

/* set up labels in q. */
    for (j = 1; j <= n; ++j)
    {
	q[klm2][j] = (Float64) j;
    }

    for (i = 1; i <= klm; ++i)
    {
	q[i][n2] = (Float64) (n + i);
	if (q[i][n1] >=  0.)
	{
	    goto L30;
	}

	for (j = 1; j <= n2; ++j)
	{
	    q[i][j] = -(Float64) q[i][j];
	}
L30:
	;
    }
/* set up phase 1 costs. */
    iphase = 2;

    for (j = 1; j <= nklm; ++j)
    {
	cu[1][j] =  0.;
	cu[2][j] =  0.;
	iu[1][j] = 0;
	iu[2][j] = 0;
    }
    if (l == 0)
    {
	goto L60;
    }

    for (j = nk1; j <= nkl; ++j)
    {
	cu[1][j] =  1.;
	cu[2][j] =  1.;
	iu[1][j] = 1;
	iu[2][j] = 1;
    }
    iphase = 1;
L60:
    if (m == 0)
    {
	goto L80;
    }

    for (j = nkl1; j <= nklm; ++j)
    {
	cu[2][j] =  1.;
	iu[2][j] = 1;
	jmn = j - n;
	if (q[jmn][n2] <  0.)
	{
	    iphase = 1;
	}
    }
L80:
    if (*kode == 0)
    {
	goto L150;
    }

    for (j = 1; j <= n; ++j)
    {
	if ((r__1 = x[j]) <  0.)
	{
	    goto L90;
	}
	else if (r__1 == 0)
	{
	    goto L110;
	}
	else
	{
	    goto L100;
	}
L90:
	cu[1][j] =  1.;
	iu[1][j] = 1;
	goto L110;
L100:
	cu[2][j] =  1.;
	iu[2][j] = 1;
L110:
	;
    }

    for (j = 1; j <= k; ++j)
    {
	jpn = j + n;
	if ((r__1 = res[j]) <  0.)
	{
	    goto L120;
	}
	else if (r__1 == 0)
	{
	    goto L140;
	}
	else
	{
	    goto L130;
	}
L120:
	cu[1][jpn] =  1.;
	iu[1][jpn] = 1;
	if (q[j][n2] >  0.)
	{
	    iphase = 1;
	}
	goto L140;
L130:
	cu[2][jpn] =  1.;
	iu[2][jpn] = 1;
	if (q[j][n2] <  0.)
	{
	    iphase = 1;
	}
L140:
	;
    }
L150:
    if (iphase == 2)
    {
	goto L500;
    }
/* compute the marginal costs. */
L160:

    for (j = js; j <= n1; ++j)
    {
	sum = 0.;

	for (i = 1; i <= klm; ++i)
	{
	    ii = (Int32) q[i][n2];
	    if (ii < 0)
	    {
		goto L170;
	    }
	    z = cu[1][ii];
	    goto L180;
    L170:
	    iineg = -ii;
	    z = cu[2][iineg];
    L180:
	    sum += (Float64) q[i][j] * (Float64) z;
	}
	q[klm1][j] = sum;
    }

    for (j = js; j <= n; ++j)
    {
	ii = (Int32) q[klm2][j];
	if (ii < 0)
	{
	    goto L210;
	}
	z = cu[1][ii];
	goto L220;
L210:
	iineg = -ii;
	z = cu[2][iineg];
L220:
	q[klm1][j] -= z;
    }
/* determine the vector to enter the basis. */
L240:
    xmax =  0.;
    if (js > n)
    {
	goto L490;
    }

    for (j = js; j <= n; ++j)
    {
	zu = q[klm1][j];
	ii = (Int32) q[klm2][j];
	if (ii > 0)
	{
	    goto L250;
	}
	ii = -ii;
	zv = zu;
	zu = -(Float64) zu - cu[1][ii] - cu[2][ii];
	goto L260;
L250:
	zv = -(Float64) zu - cu[1][ii] - cu[2][ii];
L260:
	if (kforce == 1 && ii > n)
	{
	    goto L280;
	}
	if (iu[1][ii] == 1)
	{
	    goto L270;
	}
	if (zu <= xmax)
	{
	    goto L270;
	}
	xmax = zu;
	in = j;
L270:
	if (iu[2][ii] == 1)
	{
	    goto L280;
	}
	if (zv <= xmax)
	{
	    goto L280;
	}
	xmax = zv;
	in = j;
L280:
	;
    }
    if (xmax <= toler)
    {
	goto L490;
    }
    if (q[klm1][in] == xmax)
    {
	goto L300;
    }

    for (i = 1; i <= klm2; ++i)
    {
	q[i][in] = -(Float64) q[i][in];
    }
    q[klm1][in] = xmax;
/* determine the vector to leave the basis. */
L300:
    if (iphase == 1 || ia == 0)
    {
	goto L330;
    }
    xmax =  0.;

    for (i = 1; i <= ia; ++i)
    {
	z = (r__1 = q[i][in], fabs(r__1));
	if (z <= xmax)
	{
	    goto L310;
	}
	xmax = z;
	iout = i;
L310:
	;
    }
    if (xmax <= toler)
    {
	goto L330;
    }

    for (j = 1; j <= n2; ++j)
    {
	z = q[ia][j];
	q[ia][j] = q[iout][j];
	q[iout][j] = z;
    }
    iout = ia;
    --ia;
    pivot = q[iout][in];
    goto L420;
L330:
    kk = 0;

    for (i = 1; i <= klm; ++i)
    {
	z = q[i][in];
	if (z <= toler)
	{
	    goto L340;
	}
	++kk;
	res[kk] = q[i][n1] / z;
	s[kk] = i;
L340:
	;
    }
L350:
    if (kk > 0)
    {
	goto L360;
    }
    *kode = 2;
    goto L590;
L360:
    xmin = res[1];
    iout = s[1];
    j = 1;
    if (kk == 1)
    {
	goto L380;
    }

    for (i = 2; i <= kk; ++i)
    {
	if (res[i] >= xmin)
	{
	    goto L370;
	}
	j = i;
	xmin = res[i];
	iout = s[i];
L370:
	;
    }
    res[j] = res[kk];
    s[j] = s[kk];
L380:
    --kk;
    pivot = q[iout][in];
    ii = (Int32) q[iout][n2];
    if (iphase == 1)
    {
	goto L400;
    }
    if (ii < 0)
    {
	goto L390;
    }
    if (iu[2][ii] == 1)
    {
	goto L420;
    }
    goto L400;
L390:
    iineg = -ii;
    if (iu[1][iineg] == 1)
    {
	goto L420;
    }
L400:
    ii = ABS(ii);
    cuv = cu[1][ii] + cu[2][ii];
    if (q[klm1][in] - pivot * cuv <= toler)
    {
	goto L420;
    }
/* bypass intermediate vertices. */
    for (j = js; j <= n1; ++j)
    {
	z = q[iout][j];
	q[klm1][j] -= z * cuv;
	q[iout][j] = -(Float64) z;
    }
    q[iout][n2] = -(Float64) q[iout][n2];
    goto L350;
/* gauss-jordan elimination. */
L420:
    if (*iter < maxit)
    {
	goto L430;
    }
    *kode = 3;
    goto L590;
L430:
    ++(*iter);

    for (j = js; j <= n1; ++j)
    {
	if (j != in)
	{
	    q[iout][j] /= pivot;
	}
    }

    for (j = js; j <= n1; ++j)
    {
	if (j == in)
	{
	    goto L460;
	}
	z = -(Float64) q[iout][j];

	for (i = 1; i <= klm1; ++i)
	{
	    if (i != iout)
	    {
		q[i][j] += z * q[i][in];
	    }
	}
L460:
	;
    }
    tpivot = -(Float64) pivot;

    for (i = 1; i <= klm1; ++i)
    {
	if (i != iout)
	{
	    q[i][in] /= tpivot;
	}
    }
    q[iout][in] =  1. / pivot;
    z = q[iout][n2];
    q[iout][n2] = q[klm2][in];
    q[klm2][in] = z;
    ii = (Int32) fabs(z);
    if (iu[1][ii] == 0 || iu[2][ii] == 0)
    {
	goto L240;
    }

    for (i = 1; i <= klm2; ++i)
    {
	z = q[i][in];
	q[i][in] = q[i][js];
	q[i][js] = z;
    }
    ++js;
    goto L240;
/* test for optimality. */
L490:
    if (kforce == 0)
    {
	goto L580;
    }
    if (iphase == 1 && q[klm1][n1] <= toler)
    {
	goto L500;
    }
    kforce = 0;
    goto L240;
/* set up phase 2 costs. */
L500:
    iphase = 2;

    for (j = 1; j <= nklm; ++j)
    {
	cu[1][j] =  0.;
	cu[2][j] =  0.;
    }

    for (j = n1; j <= nk; ++j)
    {
	cu[1][j] =  1.;
	cu[2][j] =  1.;
    }

    for (i = 1; i <= klm; ++i)
    {
	ii = (Int32) q[i][n2];
	if (ii > 0)
	{
	    goto L530;
	}
	ii = -ii;
	if (iu[2][ii] == 0)
	{
	    goto L560;
	}
	cu[2][ii] =  0.;
	goto L540;
L530:
	if (iu[1][ii] == 0)
	{
	    goto L560;
	}
	cu[1][ii] =  0.;
L540:
	++ia;

	for (j = 1; j <= n2; ++j)
	{
	    z = q[ia][j];
	    q[ia][j] = q[i][j];
	    q[i][j] = z;
/* L550: */
	}
L560:
	;
    }
    goto L160;
L570:
    if (q[klm1][n1] <= toler)
    {
	goto L500;
    }
    *kode = 1;
    goto L590;
L580:
    if (iphase == 1)
    {
	goto L570;
    }
/* prepare output. */
    *kode = 0;
L590:
    sum = 0.;

    for (j = 1; j <= n; ++j)
    {
	x[j] =  0.;
    }

    for (i = 1; i <= klm; ++i)
    {
	res[i] =  0.;
    }

    for (i = 1; i <= klm; ++i)
    {
	ii = (Int32) q[i][n2];
	sn =  1.;
	if (ii > 0)
	{
	    goto L620;
	}
	ii = -ii;
	sn =  -1.0F;
L620:
	if (ii > n)
	{
	    goto L630;
	}
	x[ii] = sn * q[i][n1];
	goto L640;
L630:
	iimn = ii - n;
	res[iimn] = sn * q[i][n1];
	if (ii >= n1 && ii <= nk)
	{
	    sum += (Float64) q[i][n1];
	}
L640:
	;
    }
    *error = sum;
    ERR_PASS_CLEANUP(w_nr_free_matrix (pIPWState, cu, 1, 2, 1, n+k+l+m));
    ERR_PASS_CLEANUP(w_nr_free_imatrix (pIPWState, iu, 1, 2, 1, n+k+l+m));
    ERR_PASS_CLEANUP(w_nr_free_ivector (pIPWState, s, 1, k+l+m));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (cu != NULL) w_nr_free_matrix (pIPWState, cu, 1, 2, 1, n+k+l+m);
    if (iu != NULL) w_nr_free_imatrix (pIPWState, iu, 1, 2, 1, n+k+l+m);
    if (s != NULL) w_nr_free_ivector (pIPWState, s, 1, k+l+m);
    ERR_RETURN;
}				/* cl1_ */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * do a linear least-squares fit of a set of points to a line.
 *
 * LongDescription:::
 * Given a set of (x,y) pair, do a linear least-squares fit to 
 * a straight line. This is just a simple interface to solveL1
 * for the linear regression case.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  pointsX::        	array of x-coordinates of points
 *  pointsY::        	array of y-coordinates of points
 *  nPoints::        	number of (x, y) points
 *
 * Output:::
 *  fit[3]::         	Fitted solution vector: fit[0]*x + fit[1] = y
 *			is the least-squares fit for the given data
 *  err::            	Error term
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_linearFit (
    IPWState           *pIPWState,
    Int32              *pointsX,
    Int32              *pointsY,
    Int32               nPoints,
    Float64             fit[3],
    Float64            *err)
{
    PROLOGUE;
    Float64 **q = NULL, *x = NULL, *res = NULL;
    Int32 i, kode, iter;

    /* Allocate the "q" matrix */
    ERR_PASS( w_nr_matrix(pIPWState, 1, nPoints + 2, 1, 4, &q));
    ERR_PASS_CLEANUP( w_nr_vector (pIPWState, 1, 4, &x));
    ERR_PASS_CLEANUP( w_nr_vector (pIPWState, 1, nPoints, &res));
    ERR_GEN_CLEANUP(!q || !x || !res, cErrNomem);

    /* Fill the "q" matrix */
    for (i = 1; i <= nPoints; i++)
    {
	q[i][1] = pointsX[i-1];
	q[i][2] = 1.0;
	q[i][3] = pointsY[i-1];
    }

    /* Solve it */
    kode = 0;
    iter = 10*nPoints;
    ERR_PASS_CLEANUP(w_solveL1(pIPWState, nPoints, 0, 0, 2, q, 1e-9, &kode, 
			       &iter, x, res, err));
    fit[0] = x[1];
    fit[1] = x[2];
    if (q) 
	ERR_PASS_CLEANUP(w_nr_free_matrix(pIPWState, q, 1, nPoints + 2, 1, 4));
    if (x)
	ERR_PASS_CLEANUP(w_nr_free_vector(pIPWState, x, 1, 4));
    if (res)
	ERR_PASS_CLEANUP(w_nr_free_vector(pIPWState, res, 1, nPoints));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (q) w_nr_free_matrix(pIPWState, q, 1, nPoints + 2, 1, 4);
    if (x) w_nr_free_vector(pIPWState, x, 1, 4);
    if (res) w_nr_free_vector(pIPWState, res, 1, nPoints);
    ERR_RETURN;
}

/*
$Log: geoutil.c,v $
Revision 1.1.1.1  1999/09/20 18:19:16  jlin
Imported sources

Revision 1.23  1996/09/26 21:15:50  dferrell
dferrell on Thu Sep 26 14:14:49 PDT 1996

geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v

Made w_geo_pointOnLineSeg () and w_geo_pointOnLine () round rather than
truncate when doing the final conversion from floats to ints.

Revision 1.22  1996/01/04 20:22:57  ejaquith
geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.21  1995/12/13  04:07:41  ejaquith
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/16  20:29:06  ejaquith
 * ejaquith on Thu Nov 16 12:28:47 PST 1995
 *
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * Fixed some parameter typos in some function's comment blocks.
 *
 * Revision 1.19  1995/11/15  22:43:20  ejaquith
 * ejaquith on Wed Nov 15 14:42:48 PST 1995
 *
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * Fixed reference to another IPWrap function typo (sniffed by hyperlinking).
 *
 * Revision 1.18  1995/11/13  20:11:35  ejaquith
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.17  1995/11/03  22:48:14  ddavies
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.16  1995/10/13  22:00:44  ejaquith
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/10/12  04:24:33  ddavies
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.14  1995/10/02  00:18:57  ejaquith
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * Fixed keyword typo
 *
 * Revision 1.13  1995/09/29  00:14:59  jcraig
 * jcraig on Thu Sep 28 17:14:58 PDT 1995
 *
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * CommonCheck: fixup procedure comments
 *
 * Revision 1.12  1995/09/21  22:53:23  ejaquith
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.11  1995/09/07  00:53:18  jcraig
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * add w_ to some of the free routines.
 * put ERR_PASS macros around a couple of procedures calls that were
 * missing it.
 *
 * Revision 1.10  1995/09/06  20:19:23  ejaquith
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  02:12:21  jcraig
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.8  1995/08/31  21:13:13  ejaquith
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/17  22:29:19  jcraig
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.6  1995/02/15  21:13:40  macleod
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * solveL1 used "abs"; changed it to use the ABS macro.
 *
 * Revision 1.5  1995/02/10  20:39:08  macleod
 * macleod on Fri Feb 10 12:35:21 PST 1995
 *
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * Added solveL1, which is a general-purpose linear least squares solver
 * which was converted from Fortran. The original came off of netlib.
 * It's not really a "geometry" utility in a strict sense, but one
 * could argue that it's only being used for problems of geometry, so ...
 * Also added linearFit, which is a simple interface to solveL1 for
 * linear regression problems.
 *
 * Revision 1.4  1994/10/28  02:01:57  macleod
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * Added my_sqrt, which is a gross hack for the sake of purify. Basically,
 * gcc generates an inline floating point instruction when one calls sqrt,
 * which would be fine, except that the assembler sticks a store below
 * the stack (st %fsr,[%sp-4]) before the square root, which purify
 * (rightly) catches as a write below the stack pointer. Anyway,
 * the purpose of this function is to isolate square root so that
 * one can suppress the message in purify.
 *
 * Revision 1.3  1994/09/29  00:47:17  lperry
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * Eliminate use of utilpros. Use utils.pub and geoutil.pub instead
 *
 * Revision 1.2  1994/03/02  02:50:26  macleod
 * geoutil.c -> /product/ipcore/ipwrap/src/RCS/geoutil.c,v
 *
 * Added geo_movePerpindicular, to move a line along its perpindicular
 *
 * Revision 1.1  1994/02/04  19:30:05  macleod
 * Initial revision
 *
*/


