/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  gesture2.c--Hilighter detection/extraction processing stuff.
 *
 *			PUBLIC
 *			w_extractHighlighter ()
 *
 */

#include "ipwcore.prv"
#include "pixr.h"
#include "frames.pub"
#include "defines.pub"   /* for cTrue and cFalse */
#include "morphg.h"
#include "alplib.h"
#include "logicops.h"
#include "geomadt.h"     /* for RECTANGL */
#include "shrcc.pub"     /* for TCScanLineRun and ConnectedComponent */

#include "props.pub"
#include "gesture.pub"
#include "strel.pub"
#include "morph.pub"
#include "fill.pub"
#include "intops.pub"
#include "logicops.pub"
#include "thrshold.h"    /* for THRESH_NORM */
#include "thrshold.pub"
#include "subsmpl.pub"
#include "reduce.pub"
#include "expand.pub"
#include "imginfo.pub"   /* for w_textmask prototype */

#define DEBUG_HL_F 0

#ifdef __GNUC__
#include "ansiprot.h"
#endif


IP_RCSINFO(RCSInfo, "$RCSfile: gesture2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Finds hilighted regions in a PIXR and extracts their contents.
 *
 * LongDescription:::
 *  Extracts hilighted text regions in a 4 or 8 bit gray image.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::           State structure
 *  inPr::                Input PIXR (4 or 8 bit gray)
 *  hlBandLow::           Lowest (darkest) value of hilighter
 *  hlBandHigh::          Highest (brightest) value of hilighter
 *
 * Output:::
 *  outPr::               Pointer to a pointer to a binary pixr which
 *                        contains the extracted text which was hilighted.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 * Caveats::
 *      Hilighted image (halftone or otherwise) and line art regions will
 *      not be extracted.
 *
 *      Some colors of hilighter on some scanners will produce images
 *      with very erratic and light hilighter.  Pink hilighter is the
 *      only color which scans in any consistent way.  Blue is sometimes
 *      readable while yellow and green are almost never detectable
 *      (especially given a common green lamp scanner).
 *
 * Bugs::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_extractHighlighter (
                      IPWState *pIPWState, 
		      PIXR *inPr, 
		      Int32 hlBandLow,
                      Int32 hlBandHigh, 
		      PIXR **outPr)
{
    PROLOGUE;
    Bool		destCreated = cFalse;
    Int32               depth;
    STREL              *cleanStrel = NULL;
    PIXR               *hlPr = NULL,
                       *hldPr = NULL,
                       *hllPr = NULL,
		       *textPr = NULL,
		       *tmpPr = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, inPr, &depth));

    /* This sequence tends to make things lighter (text/images) while
     * mostly leaving hilighter alone.
     */
    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, inPr, cTrue, &tmpPr));
    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, tmpPr, cTrue, &hllPr));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "rmax", hllPr, TIFF_LZW));
#endif

    /* Threshold this result leaving only regions in the hilighter band. */
    ERR_PASS_CLEANUP (w_threshold (pIPWState, hllPr, hlBandLow, hlBandHigh, 
                                   THRESH_NORM, &hlPr));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "rmaxthr", hlPr, TIFF_G4));
#endif

    /* This sequence tends to make things darker (text/images) while
     * leaving hilighter alone.
     */
    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, inPr, cFalse, &tmpPr));
    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, tmpPr, cFalse, &hldPr));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "rmin", hldPr, TIFF_LZW));
#endif

    /* Threshold this result leaving only regions in the hilighter band. */
    ERR_PASS_CLEANUP (w_threshold (pIPWState, hldPr, 
                                   MAX (0, hlBandLow - 1), hlBandHigh, 
                                   THRESH_NORM, &textPr));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "rminthr", textPr, TIFF_G4));
#endif

    /* Create a hilighter region mask by combining the two thresholded
     * min/max reductions.
     */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, textPr, AND, hlPr));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "hlmask1", hlPr, TIFF_G4));
#endif

    /* Text will have left gaps in the hilighter region mask, so attempt
     * to at least partially close these.  We will be seed filling later
     * so it is important that we at least cover part of each letter
     * which was hilighted.
     */

    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 5, 5, 2, 2, &cleanStrel));
    ERR_PASS_CLEANUP (w_close (pIPWState, hlPr, cleanStrel, &hlPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &cleanStrel));
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 3, 3, 1, 1, &cleanStrel));
    ERR_PASS_CLEANUP (w_open (pIPWState, hlPr, cleanStrel, &hlPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &cleanStrel));
    /* This assumes horizontal text line orientation */
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 7, 1, 3, 0, &cleanStrel));
    ERR_PASS_CLEANUP (w_close (pIPWState, hlPr, cleanStrel, &hlPr));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "hlmask2", hlPr, TIFF_G4));
#endif

    /* Threshold for text at full-scale.  Note that this step should
     * use a more robust method for thresholding.
     */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &textPr));
    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr, 0, (depth == 4 ? 5 : 80), 
                                   THRESH_NORM, &textPr));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "binthr", textPr, TIFF_G4));
#endif

    /* Find the text regions.  This does not currently work well for
     * kanji/kana.
     */
    ERR_PASS_CLEANUP (w_textMask (pIPWState, textPr, &tmpPr, 0));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "txtmask", tmpPr, TIFF_G4));
#endif

    /* Create a set of seeds for hilighted text regions */
    ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, tmpPr, hlPr, outPr));

#if DEBUG_HL_F
    ERR_PASS_CLEANUP (w_write (pIPWState, "hlseeds", *outPr, TIFF_G4));
#endif

    /* Seed fill the text which is covered by any hilghter */
    ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, *outPr, textPr, 3, outPr));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &hlPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &hllPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &hldPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &textPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &cleanStrel));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &hlPr);
    w_destroyPixr (pIPWState, &hllPr);
    w_destroyPixr (pIPWState, &hldPr);
    w_destroyPixr (pIPWState, &tmpPr);
    w_destroyPixr (pIPWState, &textPr);
    w_destroyStrel (pIPWState, &cleanStrel);
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    ERR_RETURN;

}  /* w_extractHighlighter () */


/*
 *
 * $Log: gesture2.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 * Imported sources
 *
 * Revision 1.2  1997/03/29 01:26:12  xzhu
 * gesture2.c -> /product/ipcore/ipwrap/src/RCS/gesture2.c,v
 *
 * CommonCheck: Add a input parameter IsTextSeg in w_textMask(). If want keep more text, turn on this parameter, otherwise set it 0.
 *
 * Revision 1.1  1996/10/03 21:29:26  dferrell
 * Initial revision
 *
 *
 */

