/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	Gradient finding wrapper (8 and 24 bit images)
 *
 *		w_findMagnitudeOfGradient ()
 *			
 */

#include "ipwcore.prv"

#include "pixr.h"
#include "defines.pub"		/* for cTrue and cFalse */
#include "geomadt.h"		/* for RECTANGL */
#include "frames.pub"
#include "props.pub"

#include "shrip.pub"
#include "gradient.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: gradient.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Wrapper for all 8 and 24 bit (RBG) gradient finders
 *
 * LongDescription:::
 *  
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  inPr::           Input image to be processed (must be 8 or 24 bit, 3 plane)
 *  opType::         Gradient operator type specifier (from shrip.pub):
 *			cGradOpMaxL1FromCenter: max L-1 intensity or color
 *				distance from center.
 *			cGradOpMaxL2FromCenter: max L-2 intensity or color
 *				distance from center.
 *			cGradOpMaxLInfFromCenter: max L-infinity intensity or 
 *				color distance from center.
 *  neighborhood::   Gradient neighborhood specifier (from shrip.pub):
 *			cGradNeighborhoodDownRight: use pixel directly to
 *				the right and the pixel directly below the
 *				center pixel to compute the gradient.
 *
 * Output:::
 *  gradientPr::     8-bit PIXR which will contain the gradient image.  If
 *                   NULL, this will be allocated, otherwise it must have the
 *                   same width and height as the input image.
 *
 * Returns:::
 *  cSuccessful      if it worked
 *  cErrNoImpl       if requested operation is not yet implemented
 *  cErrInvalidParm  if opType or neighborhood specifier is invalid
 *	
 * Notes:::
 *  For 8-bit gradients, the L-1, l-2, and L-infinity operators are equivalent,
 *  so there is actually only one underlying implementation.
 *  For 24-bit L-1 and L-2 gradients, the values returned are approximations
 *  due to the computational expense of computing them accurately to 8-bits.
 *  The 24-bit L-infinity gradient implementation is exact.
 *
 * FunctionCategory:::
 *  filtering
 *
 * Keywords:::
 *  color, gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_findMagnitudeOfGradient (
                 IPWState  *pIPWState,
                 PIXR      *inPr,
		 Int32      opType,
		 Int32      neighborhood,
                 PIXR      **gradientPr)
{
    PROLOGUE;
    Int32     w, h, created = cFalse,
              fws, fhs, 
	      fwd, fhd,
	      bpls, bpld,
	      d, channels;
    ColorType ct;
    ScaledInt1616 xdpi, ydpi;
    UInt8    *isr = NULL,
             *isg = NULL,
	     *isb = NULL,
	     *id = NULL;
    PIXR     *cr = NULL,
             *cg = NULL,
	     *cb = NULL;



    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth8, cChannels1 | cChannels3, 
                  cFrame32);

    ERR_PASS (w_pixrGetType (pIPWState, inPr, &ct));
    ERR_GEN ((ct != cColorRGB) && (ct != cColorGray), cErrTypeNotSupported);

    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &h));
    ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &d));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, inPr, &channels));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xdpi, &ydpi));

    if (*gradientPr == NULL)
    {
        ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, w, h, 8,
					      xdpi, ydpi,
					      cFrameBits, gradientPr));
	created = cTrue;
    }

    if (channels == 3)
    {
        /* 24-bit color image */
        ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, inPr, cRedPlane, &cr));
        ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, inPr, cGreenPlane, &cg));
        ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, inPr, cBluePlane, &cb));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, cr, &isr));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, cg, &isg));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, cb, &isb));
        ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &fws));
        ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &fhs));
        ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bpls));

	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *gradientPr, &id));
        ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *gradientPr, &fwd));
        ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *gradientPr, &fhd));
        ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *gradientPr, &bpld));

	ERR_PASS_CLEANUP (i_findMagnitudeOfGradientRGB (
					     pIPWState, 
	                                     isr, isg, isb, fws, fhs, bpls, 
	                                     opType, neighborhood,
					     id, fwd, fhd, bpld));

    }
    if (channels == 1)
    {
	/* gray image */
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &isr));
        ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &fws));
        ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &fhs));
        ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bpls));

	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *gradientPr, &id));
        ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *gradientPr, &fwd));
        ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *gradientPr, &fhd));
        ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *gradientPr, &bpld));

	ERR_PASS_CLEANUP (i_findMagnitudeOfGradient8bpp (
					      pIPWState, 
	                                      isr, fws, fhs, bpls, 
					      opType, neighborhood,
	                                      id, fwd, fhd, bpld));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (created)
        w_destroyPixr (pIPWState, gradientPr);
    ERR_RETURN;

}  /* w_findMagnitudeOfGradient () */


/*
 * $Log: gradient.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 * Imported sources
 *
 * Revision 1.5  1998/01/14 00:12:53  ddavies
 * gradient.c -> /product/ipcore/ipwrap/src/RCS/gradient.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.4  1997/07/25 01:20:54  jcraig
 * jcraig on Thu Jul 24 18:20:27 PDT 1997
 *
 * gradient.c -> /product/ipcore/ipwrap/src/RCS/gradient.c,v
 *
 * add gradient to the list of keywords for w_findMagnitudeOfGradient
 *
 * Revision 1.3  1997/04/15 21:53:01  ddavies
 * gradient.c -> /product/ipcore/ipwrap/src/RCS/gradient.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.2  1996/07/25 16:49:37  dferrell
 * dferrell on Thu Jul 25 09:49:26 PDT 1996
 *
 * gradient.c -> /product/ipcore/ipwrap/src/RCS/gradient.c,v
 *
 * Changed "operator" parameter name in gradient code to "opType" since
 * in C++ "operator" is a reserved word.
 *
 * Revision 1.1  1996/07/01 17:13:04  dferrell
 * Initial revision
 *
 *
 */

