/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  grayproj.c
 *		w_projectGrayPixels
 *		w_backprojectGrayPixels
 *		w_normalizeProjection
 *		w_writeProjection
 *
 */

#include <stdio.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"

#include "pixr.h"
#include "frames.pub"    /* for cFrameBits */
#include "defines.pub"   /* for cTrue and cFalse */
#include "geomadt.h"
#include "props.pub"
#include "prxform.pub"
#include "imageref.pub"

#include "shrip.pub"
#include "grayproj.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: grayproj.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Creates horizontal or vertical projection histogram of a gray image
 *
 * LongDescription:::
 *	Sums pixel intensities in rows (for horizontal projection) or 
 *	columns (for vertical projection) into an array.
 *
 * Input:::
 *	pIPWState::	Ptr to IP struct for memory alloc/freeing routines
 *	inPr::		Input gray image to be projected
 *	invert::	If cTrue, invert the intensity when projecting
 *	direction::	Direction in which to project pixels.  Can be
 *			cHorizontal or cVertical
 *
 * Output:::
 *	projection::	Pointer to pointer to projection array.  If points
 *			to NULL, the array will be allocated.
 *
 * Returns:::
 *	cSuccessful	if completes successfully
 *	cErrNomem	if unable to allocate array
 *	pixr errors	if pixr parameters incorrect
 *
 * Notes:::
 *	Currently only supports 8 bit images
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_projectGrayPixels (
                    IPWState *pIPWState, 
		    PIXR *inPr,
		    Bool  invert,
		    Int32 direction,
		    Int32 **projection)
{
    PROLOGUE;
    Bool	 projectionCreated = cFalse;
    Int32        w, h, d, channels, c, bpl, retCode;
    UInt8	*iS;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth4 | cDepth8, 
                  cChannelsAny, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &h));
    ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &d));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, inPr, &channels));

    if ((*projection) == NULL)
    {
	if (direction == cHorizontal)
	    *projection = (Int32 *)IPW_CALLOC (h, sizeof (Int32), 
	                                       cAllocatorData);
	else
	    *projection = (Int32 *)IPW_CALLOC (w, sizeof (Int32), 
	                                       cAllocatorData);
	ERR_GEN ((*projection) == NULL, cErrNomem);
	projectionCreated = cTrue;
    }



    if (channels > 1)
    {
        PIXR  *subPr = NULL;

	ERR_PASS (w_pixrFramedHeight (pIPWState, inPr, &h));
	ERR_PASS (w_pixrFramedWidth (pIPWState, inPr, &w));
	ERR_PASS (w_pixrGetBpl (pIPWState, inPr, &bpl));

	for (c = 0; c < channels; c++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, inPr, c, &subPr));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subPr, &iS));

	    if (d == 8)
	    {
		if (direction == cHorizontal)
		    retCode = i_projectHGrayPixels8 (pIPWState, iS, w, h, bpl, 
		                                     invert, *projection);
		else
		    retCode = i_projectVGrayPixels8 (pIPWState, iS, w, h, bpl, 
		                                     invert, *projection);
		ERR_GEN_CLEANUP ((retCode != cSuccessful), retCode);
	    }
	    else
	    {
		ERR_GEN_CLEANUP (cTrue, cErrNoImpl);
	    }
	}
    }
    else
    {
	ERR_PASS (w_pixrFramedHeight (pIPWState, inPr, &h));
	ERR_PASS (w_pixrFramedWidth (pIPWState, inPr, &w));
	ERR_PASS (w_pixrGetBpl (pIPWState, inPr, &bpl));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));

        if (d == 8)
	{
	    if (direction == cHorizontal)
		retCode = i_projectHGrayPixels8 (pIPWState, iS, w, h, bpl, 
		                                 invert, *projection);
            else
		retCode = i_projectVGrayPixels8 (pIPWState, iS, w, h, bpl, 
		                                 invert, *projection);
            ERR_GEN_CLEANUP ((retCode != cSuccessful), retCode);
	}
	else
	{
	    ERR_GEN_CLEANUP (cTrue, cErrNoImpl);
	}
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    if (projectionCreated)
    {
        IPW_FREE (*projection, cAllocatorData);
        *projection = NULL;
    }
    ERR_RETURN;

}  /* w_projectGrayPixels () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Redistributes two orthogonal projections into an image using "op".
 *
 * LongDescription:::
 *	The value of the result image pixel at location (x, y) is: 
 *	vProjection[x] <op> hProjection[y] where op is minimum, maximum,
 *	average, or absolute value of difference of the two operands.
 *	Note that the projection arrays should be normalized before this
 *	operation.
 *
 * Input:::
 *	pIPWState::	Ptr to IP struct for memory alloc/freeing routines
 *	inPr::		Input gray image to be projected
 *	invert::	If cTrue, invert the intensity when projecting
 *	normalize::	If cTrue, normalize before backprojecting
 *	op::		Operator used to combine vertical and horizontal
 *			projections at each pixel location.  Can be
 *			cMin, cMax, cDiff, or cAvg.
 *	hProjection::	The horizontal projection array
 *	vProjection::	The vertical projection array
 *
 * Output:::
 *	outPr::		Pointer to pointer to result PIXR.  If pointer to
 *			NULL passed, this will be allocated.
 *
 * Returns:::
 *	cSuccessful	if completes successfully
 *	cErrNomem	if unable to allocate PIXR
 *	pixr errors	if pixr parameters incorrect
 *
 * Notes:::
 *	Currently only supports 8 bit images
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_backprojectGrayPixels (
                    IPWState *pIPWState,
		    PIXR     *inPr,
		    Bool      invert,
		    Bool      normalize,
		    Int32     op,
		    Int32    *hProjection,
		    Int32    *vProjection,
		    PIXR    **outPr)
{
    PROLOGUE;
    Int32        w, h, d, bpl;
    ScaledInt1616 xRes, yRes;
    Bool         outPrCreated = cFalse;
    UInt8	*iD;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth4 | cDepth8, 
                  cChannelsAny, cFrame32);

    ERR_GEN ((hProjection == NULL) || (vProjection == NULL), cErrInvalidParm);

    ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &d));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &h));

    if (*outPr == NULL)
    {
	ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, inPr, 
	                                          &xRes, &yRes));

        ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, w, h, d, xRes, yRes,
					      cFrameBits, outPr));
	outPrCreated = cTrue;
    }

    ERR_PASS (w_pixrFramedWidth (pIPWState, *outPr, &w));
    ERR_PASS (w_pixrFramedHeight (pIPWState, *outPr, &h));
    ERR_PASS (w_pixrGetBpl (pIPWState, *outPr, &bpl));
    ERR_PASS (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    switch (op)
    {
        case cMin:
	    if (d == 4)
	        ERR_GEN_CLEANUP (cTrue, cErrNoImpl);
	    else
	        ERR_PASS_CLEANUP (i_backprojectGrayPixelsMin8 (pIPWState,
							       invert,
		                                               normalize,
							       hProjection,
							       vProjection,
							       w, h, bpl, iD));
	    break;

	case cMax:
	    if (d == 4)
	        ERR_GEN_CLEANUP (cTrue, cErrNoImpl);
	    else
	        ERR_PASS_CLEANUP (i_backprojectGrayPixelsMax8 (pIPWState,
							       invert,
		                                               normalize,
							       hProjection,
							       vProjection,
							       w, h, bpl, iD));
	    break;

	case cDiff:
	    if (d == 4)
	        ERR_GEN_CLEANUP (cTrue, cErrNoImpl);
	    else
	        ERR_PASS_CLEANUP (i_backprojectGrayPixelsDiff8 (pIPWState,
								invert,
		                                                normalize,
							        hProjection,
							        vProjection,
							        w, h, bpl, iD));
	    break;

	case cAvg:
	    if (d == 4)
	        ERR_GEN_CLEANUP (cTrue, cErrNoImpl);
	    else
	        ERR_PASS_CLEANUP (i_backprojectGrayPixelsAvg8 (pIPWState,
							       invert,
		                                               normalize,
							       hProjection,
							       vProjection,
							       w, h, bpl, iD));
	    break;

	default:
	    ERR_RETURN;
	    break;
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (outPrCreated)
        w_destroyPixr (pIPWState, outPr);
    ERR_RETURN;

}  /* w_backprojectGrayPixels () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Normalizes a projection histogram by a normalization factor.
 *
 * LongDescription:::
 *
 * Input:::
 *	pIPWState::	Ptr to IP struct for memory alloc/freeing routines
 *	projection::	Projection array to be normalized
 *	length::	Length of array
 *	depth::		Depth of image that was projected
 *	invert::	If cTrue, invert the intensity sense when normalizing
 *	normalizationFactor::  The number by which to divide each projection
 *			array value
 *
 * Output:::
 *
 * Returns:::
 *  cSuccessful
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_normalizeProjection (
                   IPWState *pIPWState,
		   Int32    *projection,
		   Int32     length,
		   Int32     depth,
		   Bool      invert,
                   Int32     normalizationFactor)
{
    PROLOGUE;

    ERR_PASS (i_normalizeProjection (pIPWState, projection,
				     length, depth, invert,
                                     normalizationFactor));

    NORMAL_RETURN;

}  /* w_normalizeProjection () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Dumps a projection histogram to a file for plotting using GNUPlot
 *
 * LongDescription:::
 *
 * Input:::
 *	pIPWState::	Ptr to IP struct for memory alloc/freeing routines
 *	projection::	Projection array to be output
 *	length::	Length of array
 *      fileName::	Filename to which to write the array
 *
 * Output:::
 *
 * Returns:::
 *  cSuccessful
 *  cErrInternal	if file I/O fails
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeProjection (
                   IPWState *pIPWState,
		   Int32    *projection,
		   Int32     length,
		   char     *fileName)
{
    PROLOGUE;
    Int32    b;
    FILE    *fp;

    fp = fopen (fileName, "w");
    ERR_GEN ((fp == NULL), cErrInternal);

    fprintf (fp, "#\tProjection histogram\n\n");

    for (b = 0; b < length; b++)
        fprintf (fp, "%ld\t%ld\n", b, projection[b]);

    fclose (fp);

    NORMAL_RETURN;

}  /* w_writeProjection () */


/*
 * $Log: grayproj.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 * Imported sources
 *
 * Revision 1.4  1998/01/14 00:12:55  ddavies
 * grayproj.c -> /product/ipcore/ipwrap/src/RCS/grayproj.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.3  1997/04/15 21:53:15  ddavies
 * grayproj.c -> /product/ipcore/ipwrap/src/RCS/grayproj.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.2  1997/03/04 17:35:02  dferrell
 * dferrell on Tue Mar 4 09:34:08 PST 1997
 *
 * grayproj.c -> /product/ipcore/ipwrap/src/RCS/grayproj.c,v
 *
 * Added missing one line comments to three function headers.
 * I will always comment my code.  I will always comment my code.  I will...
 *
 * Revision 1.1  1997/03/03 22:32:13  dferrell
 * Initial revision
 *
 */

