/*<HTML><HEAD><TITLE>grbinop.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  grbinop.c: contains ops combining gray and binary pixrs
 *
 *         Combine Gray PIXR with Binary pixr:
 *                  w_opGrayBin()
 *                  w_opGrayBinIP()
 *
 *         Select between two gray images based on a binary image
 *                  w_graySelect()
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "grbinop.pub"
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: grbinop.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");

/* These routines combine binary and "gray" pixrs of the same width and
 * height.  The gray is in quotes because the gray pixr may also be binary.
 * Two operations are currently supported: mask and paint.  The operation
 * selectors are the equal to the mask and paint selectors used by the
 * Sun Pixrect library routines (mask = PIX_SRC & PIX_DST, paint =
 * PIX_SRC | PIX_DST).
 *
 * If both pixrs are binary, the mask operation returns a pixr
 * that's the AND of the two source pixrects.
 *
 * If both pixrs are binary, the paint operation returns a pixr that's
 * the OR of the two source pixrs.
 *
 * Unfortunately for image processing brains everywhere, the mapping between
 * [white <-> black] and [small pixel values <-> big pixel values] works
 * one way for binary images and the opposite way for gray images.  In the
 * binary world, 0 = white; 1 = black.  It came from computer people drawing
 * pictures on screens and putting 1's where they wanted something to appear
 * (all real computers use white blackgrounds, right?).  In the gray world,
 * 0 = black and 1111...11 = white.  The value represent the number of photons
 * hitting a sensor.
 *
 * The mask operation wants to force white to appear in the gray image
 * everywhere there's a 0 in the binary image.  We do this by inverting the
 * binary image, expanding it to cover the gray image, and ORing the
 * results.  Every 0 in the binary image forces the corresponding gray pixel
 * to be all 1's.
 *
 * The paint operation is complementary - we invert the binary image, expand
 * it and AND the results.  Every 1 in the binary image forces the
 * corresponding gray pixel to be 0 (black).
 *
 * The same expansion tables used in the fast expansion procedures (see
 * fexpand.c) are used to expand the binary pixr to cover the gray pixr.
 * Note that our expansion is purely horizontal, not vertical.
 *
 * opGrayBinPr combines two source pixrs into a new gray pixr.
 * opGraybinPrIP combines two source pixrs.  The result is placed in the
 * original gray pixr.
 */

/* The difference between the accelerated and non-accelerated versions is
 * that each line is processed with assembly code in the accelerated version.
 * We don't bother accelerating the 1 bit/pixel version since it's using the
 * standard Sun Pixrect libraries already (and is pretty fast anyway).
 */


/*****************************************************************************
 * BeginDocumentation::: "w_opGrayBin"
 *
 * OneLineDescription:::
 *  Paint or mask a gray image with a binary image.
 *
 * LongDescription:::
 *  One would often like to select or obliterate some arbitrary region
 *  of a gray image.  The region is described by a mask in a binary image.
 *  This operator can perform one of two functions.  The "paint" option
 *  finds all gray pixels that corresponding to black binary pixels.  These
 *  are forced to black.  All other gray pixels are unaffected.  The
 *  "mask" option identifies all gray pixels that correspond to white pixels
 *  in the binary image.  These are set to white.  All other pixels are
 *  unaffected.
 *<BR><BR>
 *  It is best to think in terms of black/white, as opposed to 0/1,
 *  here because the association between pixel values and black/white
 *  changes between binary and gray images.  In binary images, 0 = white.
 *  In gray images, 0 = black.  This routine deals with all that.
 *<BR><BR>
 *  The "gray" image can in fact be binary.  In this case, paint is
 *  a simple OR and mask is a simple AND.  Even though this is not
 *  the w_rasterop operation used when operating on gray images, we always
 *  say the paint option is specified by (PIX_SRC | PIX_DST) and the
 *  mask option is specified by (PIX_SRC & PIX_DST).  The constants
 *  PIX_SRC and PIX_DST are defined in ipshared/include/defines.pub.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prG::            Pointer to gray source image.  This can have only
 *		     one channel, but may be 1, 2, 4 or 8 bits/pixel.
 *  prB::            Pointer to binary mask image.  This must be
 *		     approximately the same size and the same resolution
 *		     as the gray image.
 *  op::             The operation desired.  To paint the gray image with
 *		     the binary image, specify (PIX_SRC | PIX_DST).  To
 *		     mask the gray image with the binary image, specify
 *		     (PIX_SRC & PIX_DST).  All other values are illegal.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.  If *pPrD == prG, the operation will
 *		     be done in place.
 *
 * Returns:::
 *  Checks the normal pointer stuff, plus checks to see if op is legal.
 *  If the source image is too big or too small (or otherwise weird),
 *  cErrSrcDimensNotEqual is returned.
 *
 * Notes:::
 *  Sometimes people find it useful to remember what this routine really does.
 *  It expands the each binary pixel to cover the gray pixel then operates.
 *  If the "gray" image is really binary, then the operation is the exact
 *  one specified.  If the gray image is really gray, then "paint" turns
 *  into (~PIX_SRC & PIX_DST) and "mask" turns into (~PIX_SRC | PIX_DST).
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, document drycleaning, segmentation, masking, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation::: <A NAME = "w_opGrayBin"></A>
 *****************************************************************************/
ErrCode CDECL
w_opGrayBin (
    IPWState           *pIPWState,
    PIXR               *prG,
    PIXR               *prB,
    Int32               op,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    Int32	 dG, bplB, bplG, bplD, wG, hG;
    ScaledInt1616 xdpi, ydpi;
    UInt8	*pB, *pG, *pD;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prG == NULL || prB == NULL || pPrD == NULL, cErrNullPtr);
    ERR_GEN((op != (PIX_SRC | PIX_DST)) && (op != (PIX_SRC & PIX_DST)),
	    cErrInvalidParm);

    ERR_GEN_PIXR(prG, cByteOrderBig, cDepth1 | cDepth2| cDepth4 | cDepth8,
		 cChannels1, cFrame32);
    ERR_GEN_PIXR(prB, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* We have to check the binary image to see if it's about the
	 * right size and all.  The checking proc takes lots of
	 * args.
	 */
    ERR_PASS(w_pixrGetWidth(pIPWState, prG, &wG));
    ERR_PASS(w_pixrGetHeight(pIPWState, prG, &hG));
    ERR_PASS(w_pixrGetResolutionDPI(pIPWState, prG, &xdpi, &ydpi));
    ERR_CHECK(w_checkDestPixr (pIPWState, wG, hG, 1, xdpi, ydpi,
			        cFrameBits, 1, cColorGray, NULL,
			        prB, &rightFrameColsToClear,
			        &bottomFrameRowsToClear))
    {
	    /* if the mask image was the wrong size, fix the error code. */
	if (__error__ == cErrDestDimensIncorrect)
	    __error__ = cErrSrcDimensNotEqual;
	goto ERR_CLEANUP;
    }

	/* If we survive that, the mask pixr is ok.  Just ignore
	 * the stuff in rightFrameColsToClear and bottomFrameRowsToClear,
	 * the mask doesn't cause stuff to be written into the gray
	 * if (or the frame is cleared if it does).
	 */

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prG, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/* we`ve got legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prG, &wG));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prG, &hG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prG, &bplG));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prB, &bplB));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prG, &dG));

    ERR_PASS_CLEANUP( i_opGrayBin (pIPWState, pG, wG, hG, bplG, dG,
				   pB, bplB, op,
				   pD, bplD));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation::: "w_graySelect"
 *
 * OneLineDescription:::
 *  Toggle between two gray (or color) images based on a selection image
 *
 * LongDescription:::
 *  This function uses the pixels of a selection mask to toggle between
 *  pixels in two gray source images. If the selection image is binary,
 *  its pixels are used to toggle between the two source images.
 *  If the selection image is gray (4-bit only, at the moment) then
 *  the selection pixels are used to weight the resulting pixels linearly
 *  between all from first image to all from the second image.
 *  <BR><BR>
 *  By convention one gray image is called the background image and the
 *  other is called the foreground image.  Conventionally, the background
 *  image represents the light background of a page (paper or tint) and
 *  the foreground represents the darker color of text or other blocks
 *  laid on top of the background.
 *  <BR><BR>
 *  This routine implements four input options.  The user may supply
 *  neither gray input image, just the foreground, just the background
 *  or both foreground and background.  If the foreground is omitted, we
 *  substitute a virtual black image for it (no memory is actually allocated).
 *  Similarly, if the background is omitted, we substitute a virtual
 *  white image for it.  An image is omitted by supplying a NULL pointer.
 *  <BR><BR>
 *  A gray selector is usually made by scaling a binary selector using
 *  a scale to gray op.  Our convention for binary and gray images says
 *  the 0 = white in binary and 0 = black in gray.  Therefore, the scale
 *  to gray op inverts the sense of the gray pixels.  Rather than force
 *  the user to reinvert the binary or gray selector image to make them
 *  consistent, we reverse our interpretation of gray selector images.
 *  Therefore, a binary selector pixel having a value of 0 => use pixel
 *  from background image.  A gray selector pixel having a value of 0 =>
 *  use pixel from foreground image.
 *  <BR><BR>
 *  All the images (two gray inputs if provided, one selector, one output)
 *  must be the same size.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prBack::         Pointer to gray background image.
 *                   8 bits/pixel is the only depth supported at the moment.
 *		     If NULL, a virtual white image is substituted.
 *  prFore::         Pointer to gray foreground image.
 *                   8 bits/pixel is the only depth supported at the moment.
 *		     If NULL, a virtual black image is substituted.
 *  prT::            Pointer to toggle image.  This must be
 *		     exactly the same size and the same resolution
 *		     as the source gray images and the dest image.
 *		     The result for dest pixel i is ((pFore[i] & pT[i]) |
 *		     (pBack[i] & ~pT[i])) if the toggle image is binary
 *		     and ((pBack[i] * pT[i]) + (pFore[i] * (15 - pT[i])) when
 *		     the toggle image is 4-bit gray.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.  If a foreground and/or background and/or dest
 *		     image is supplied, we check to see that the number
 *		     of channels is consistent among them and use that number
 *		     of channels for the dest.  If none of these are supplied,
 *		     we produce a dest image having one channel.
 *
 * Returns:::
 *  Checks the normal pointer stuff.
 *  If images aren't all the same size, cErrSrcDimensNotEqual is returned.
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, segmentation, masking, gray, binary, comparison, page rendering
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. MacLeod, D. Davies
 *
 * EndDocumentation::: <A NAME = "w_graySelect"></A>
 *****************************************************************************/
ErrCode CDECL
w_graySelect (
    IPWState           *pIPWState,
    PIXR               *prBack,
    PIXR               *prFore,
    PIXR               *prT,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL, *prForeChan = NULL, *prBackChan = NULL, *prDChan;
    Int32	 bplT, bplFore, bplBack, bplD,
		 wFore, hFore, wBack, hBack, wT, hT, 
	         chanFore, chanBack, numChannels, depthT, i;
    ScaledInt1616 xdpi, ydpi;
    UInt8	*pT, *pFore, *pBack, *pD;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear = 0, bottomFrameRowsToClear = 0;
    ColorMap    *map = NULL;
    ColorType    type;

    ERR_GEN(prT == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prT, cByteOrderBig, cDepth1 | cDepth4, cChannels1, cFrame32);

	/* We have to check the toggle image to see if it's about the
	 * right size and all.  The checking proc takes lots of
	 * args.
	 */
    ERR_PASS(w_pixrGetWidth(pIPWState, prT, &wT));
    ERR_PASS(w_pixrGetHeight(pIPWState, prT, &hT));

    if (prFore != NULL)
    {
	ERR_GEN_PIXR(prFore, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);
	ERR_PASS(w_pixrGetWidth(pIPWState, prFore, &wFore));
	ERR_GEN (wFore != wT, cErrSrcDimensNotEqual);
	ERR_PASS(w_pixrGetHeight(pIPWState, prFore, &hFore));
	ERR_GEN (hFore != hT, cErrSrcDimensNotEqual);
	ERR_PASS(w_pixrGetChannelCount(pIPWState, prFore, &chanFore));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prFore, &bplFore));
    }
    else
    {
	wFore = 0;
	hFore = 0;
	chanFore = 0;
	bplFore = 0;
    }

    if (prBack != NULL)
    {
	ERR_GEN_PIXR(prBack, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);
	ERR_PASS(w_pixrGetWidth(pIPWState, prBack, &wBack));
	ERR_GEN (wBack != wT, cErrSrcDimensNotEqual);
	ERR_PASS(w_pixrGetHeight(pIPWState, prBack, &hBack));
	ERR_GEN (hBack != hT, cErrSrcDimensNotEqual);
	ERR_PASS(w_pixrGetChannelCount(pIPWState, prBack, &chanBack));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prBack, &bplBack));
    }
    else
    {
	wBack = 0;
	hBack = 0;
	chanBack = 0;
	bplBack = 0;
    }

	/* Decide on the number of channels in the output image.
	 * If we got two input images, make sure they match and
	 * use that number of channels.
	 */
    if ((prBack != NULL) && (prFore != NULL))
    {
	ERR_GEN (chanFore != chanBack, cErrChannelsNotEqual);
	ERR_GEN (bplFore != bplBack, cErrSrcDimensNotEqual);
	numChannels = chanFore;
	ERR_PASS(w_pixrGetType (pIPWState, prFore, &type));
	ERR_PASS(w_pixrGetColorMap (pIPWState, prFore, &map));
    }
    else	/* if there's only foreground, use its channel count */
    if (prFore != NULL)
    {
	numChannels = chanFore;
	ERR_PASS(w_pixrGetType (pIPWState, prFore, &type));
	ERR_PASS(w_pixrGetColorMap (pIPWState, prFore, &map));
    }
    else	/* if there's only background, use its channel count */
    if (prBack != NULL)
    {
	numChannels = chanBack;
	ERR_PASS(w_pixrGetType (pIPWState, prBack, &type));
	ERR_PASS(w_pixrGetColorMap (pIPWState, prBack, &map));
    }
    else	/* if there's only a dest, use its channel count */
    if (*pPrD != NULL)
    {
	ERR_PASS(w_pixrGetChannelCount (pIPWState, *pPrD, &numChannels));
	ERR_PASS(w_pixrGetType (pIPWState, *pPrD, &type));
	ERR_PASS(w_pixrGetColorMap (pIPWState, *pPrD, &map));
    }
    else	/* there's no foreground, no background, no dest, so
		 * make a single channel dest.
		 */
    {
	numChannels = 1;
	type = cColorGray;
	map = NULL;
    }


    ERR_PASS(w_pixrGetDepth(pIPWState, prT, &depthT));
    ERR_PASS(w_pixrGetResolutionDPI(pIPWState, prT, &xdpi, &ydpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrExtendedConditionalNoInit (pIPWState, wT, hT, 8,
						      xdpi, ydpi, cFrameBits,
						      numChannels,
						      type, map, &prD,
						      &rightFrameColsToClear,
						      &bottomFrameRowsToClear,
						      &allocatedHere));

    /* we`ve got legal inputs, do the real work */
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prT, &wT));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prT, &hT));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prT, &pT));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prT, &bplT));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));

    for (i = 0; i < numChannels; i++)
    {
	prForeChan = prBackChan = prDChan = NULL;
	if (prFore != NULL)
	{
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prFore, i,
						&prForeChan));
	    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prForeChan, &pFore));
	}
	else
	    pFore = NULL;

	if (prBack != NULL)
	{
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prBack, i,
						&prBackChan));
	    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prBackChan, &pBack));
	}
	else
	    pBack = NULL;

	ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, i, &prDChan));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prDChan, &pD));

	ERR_PASS_CLEANUP( i_graySelect (pIPWState, pBack, pFore, wT, hT,
				        MAX(bplFore, bplBack), 8,
				        pT, bplT, depthT,
				        pD, bplD));
    }


    /* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*

 $Log: grbinop.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.24  1998/01/14 00:12:57  ddavies
 grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.23  1997/05/16 18:15:47  ddavies
 grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v

 It no longer scales the input gray images.
 It can take either foreground or background or both or neither.  It is now
 consistent in how it treats the selector pixel values (white always selects
 background instead of 0 always selecting background).
 It derives the number of channels in the output image by looking at the
 input images (if any), then the output image provide (if any).  If none
 are provided, it assumes 1 channel (gray, not color).
 The args have been renamed Back and Fore instead of G1 and G2.
 use wp_createPixrExtendedConditionalNoInit instead of doing the checking
 for a dest PIXR internally.
 All image sizes and bytes/line and resolutions are now checked to make
 sure they're the same.  Only one set of width,height,depth/bpl is provided
 to i_graySelect for all the gray inputs.

 Revision 1.22  1997/04/15 21:53:29  ddavies
 grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.21  1996/01/08 16:00:38  macleod
 macleod on Mon Jan 8 07:59:58 PST 1996

 grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v

 Allow a 4-bit selection mask in addition to a binary selection mask

 * Revision 1.20  1995/12/13  18:55:52  macleod
 * macleod on Wed Dec 13 10:55:22 PST 1995
 *
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * Added w_graySelect, which is a wrapper to i_graySelect. It allows one
 * to select between two gray (or color) images based on a binary mask
 *
 * Revision 1.19  1995/12/13  04:08:59  ejaquith
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.18  1995/11/13  20:12:40  ejaquith
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.17  1995/10/13  22:01:52  ejaquith
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.16  1995/10/05  04:28:21  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.15  1995/10/03  01:21:21  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.14  1995/09/21  22:53:33  ejaquith
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/15  18:01:48  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.12  1995/09/11  22:12:20  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.11  1995/09/10  00:34:20  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.10  1995/09/07  01:32:22  ejaquith
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.9  1995/09/06  20:20:21  ejaquith
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  01:16:21  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.7  1995/08/31  21:14:17  ejaquith
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/17  22:03:38  ejaquith
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.5  1995/07/29  04:40:21  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.4  1994/10/07  19:19:05  lperry
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.3  1993/12/03  01:16:15  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:12:59  ddavies
 * grbinop.c -> /product/ipcore/ipwrap/src/RCS/grbinop.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

/*</PRE></BODY>*/
