/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	gsrot1.c:
 *
 *          Gray Scale rotation routines, SMALL ANGLES ONLY! (< 10 degrees)
 *
 *          Gray Scale rotation routines:
 *                  w_rotateGray
 *		    w_rotateGrayIP
 * 
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "alplib.h"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "rotatop.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: gsrot1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");

/* These routines are wrappers for the ipshared routines that do the
 * real work of rotating gray image using interpolation. */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate a gray image by a small angle, output pixels are interpolated.
 *
 * LongDescription:::
 *  There are a couple of ways to rotation and scaling.  At issue is the
 *  way output pixels are calculated when the math tell you that there is
 *  no 1 to 1 mapping from input pixels to output pixels.  Conceptually,
 *  say you rotate or stretch the input pixels so that they occupy exactly
 *  the correct positions, though the grids may not be rectangular or strictly
 *  horizontal and vertical anymore.  You then lay the real output grid
 *  down on top of this rotated image.  Each output pixel overlaps some
 *  set of input pixels.  In the simplest rotation scheme, you set each 
 *  output pixel to the exact value of the input pixel with the largest
 *  overlap.  This is called non-interpolated rotation.  A nicer (& slower)
 *  version of rotation computes the value of each output pixel from the
 *  weighted average of all input pixels which intersect the output pixel.
 *  The weighting factor for one input pixel is the fraction of the area
 *  of the output pixel that lies on top of the input pixel.  This is
 *  called interpolated rotation.  w_rotateGray does interpolated rotation.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be rotated.
 *  rotAngle::       Rotation angle in radians.  A positive number causes
 *		     clockwise rotation, a negative number causes
 *		     counterclockwise rotation.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.  It is not possible to set the source and
 *		     dest to be the same.
 *
 * Returns:::
 *  In addition to the usual conditions, this routine tests to see if the
 *  input angle is too big (> 0.12 radians).  If so, cErrInvalidParm is
 *  returned.  This routine also returns cErrInvalidParm if the source
 *  and dest are equal.
 *
 * Notes:::
 *  The output image is always the same size as the input image, so
 *  some input pixels will be shifted off the edge and lost.  New pixels
 *  that were shifted in from off the edge are set to white.
 *<BR><BR>
 *  The optimized rotation method used does make approximations.  For small
 *  angles, these are normally not noticeable.  A result of one of the
 *  approximations is that pixels in gray images get a little darker.  For
 *  example, the result of rotating a white area (all pixels = 0xFF) is
 *  a region where the whitest pixel has a value of 0xFC.  If this is
 *  objectionable, the client may get brighter brights and whiter whites
 *  by applying a mild TRC using w_map.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, interpolated, small angle, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotateGray (
    IPWState           *pIPWState,
    PIXR               *prS,
    Float32             rotAngle,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL, *prChanS = NULL, *prChanD = NULL;
    Int32		 wS, hS;	/* width and height of gray Pr */
    Int32		 dS, bplD, bplS, channels, i;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    ERR_GEN(prS == *pPrD, cErrInvalidParm);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    /* angle must be between -10 and 10 degrees (0.17 rads) */
    ERR_GEN(ABS(rotAngle) > 10.0F * 3.1415926F / 180.0F, cErrInvalidParm);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

    ERR_PASS_CLEANUP( w_pixrGetChannelCount (pIPWState, prS, &channels));
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS_CLEANUP( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

    for (i = 0; i < channels; i++)
    {
	/* now that we have legal operands, do the operation */
	prChanS = prChanD = NULL;
	ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, i, &prChanS));
	ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, i, &prChanD));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prChanS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prChanS, &bplS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prChanD, &pD));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prChanD, &bplD));

	ERR_PASS_CLEANUP( i_rotateGray (pIPWState, pS, wS, hS, bplS, dS,
				        rotAngle,
				        pD, bplD,
				        (Int16) ScaledInt1616_TO_Int32(xSdpi),
				        (Int16) ScaledInt1616_TO_Int32(ySdpi)));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform small-angle, interpolated rotation, leave result in source image.
 *
 * LongDescription:::
 *  See w_rotateGray for a discussion of interpolated rotation vs.
 *  non-interpolated rotation.  The difference between this routine and
 *  w_rotateGray is that this routine saves memory by leaving the output
 *  bits in the same image as the input bits.
 *
 * Private Description:::
 *  A smallish memory buffer is allocated and freed internally.  This buffer
 *  holds portions of rotated image before they are put back into the
 *  input image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  rotAngle::       Rotation angle in radians.  A positive number causes
 *		     clockwise rotation, a negative number causes
 *		     counterclockwise rotation.
 *
 * Output:::
 *  prS::            Pointer to image that is used as both source and
 *		     destination.
 *
 * Returns:::
 *  In addition to the usual conditions, this routine tests to see if the
 *  input angle is too big (> 0.12 radians).  If so, cErrInvalidParm is
 *  returned.
 *
 * Notes:::
 *  The output image is always the same size as the input image, so
 *  some input pixels will be shifted off the edge and lost.  New pixels
 *  that were shifted in from off the edge are set to white.
 *<BR><BR>
 *  The optimized rotation method used does make approximations.  For small
 *  angles, these are normally not noticeable.  A result of one of the
 *  approximations is that pixels in gray images get a little darker.  For
 *  example, the result of rotating a white area (all pixels = 0xFF) is
 *  a region where the whitest pixel has a value of 0xFC.  If this is
 *  objectionable, the client may get brighter brights and whiter whites
 *  by applying a mild TRC using w_map.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, interpolated, small angle, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotateGrayIP (
    IPWState           *pIPWState,
    Float32             rotAngle,
    PIXR               *prS)
{
    PROLOGUE;
    Int32		 wS, hS;       /* width and height of gray Pr */
    Int32		 depth, bplS, channels, i;
    ScaledInt1616	 xdpi, ydpi;
    UInt8		*pS;
    PIXR                *prChanS = NULL;
	
    ERR_GEN(prS == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xdpi, &ydpi));

    /* angle must be between -10 and 10 degrees (0.17 rads) */
    ERR_GEN(ABS(rotAngle) > 10.0F * 3.1415926F / 180.0F, cErrInvalidParm);

    ERR_PASS( w_pixrGetChannelCount (pIPWState, prS, &channels));

    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &hS));

    for (i = 0; i < channels; i++)
    {
	prChanS = NULL;
	ERR_PASS( w_pixrGetChannel(pIPWState, prS, i, &prChanS));
	ERR_PASS( w_pixrGetImageAll(pIPWState, prChanS, &pS));
	ERR_PASS( w_pixrGetBpl(pIPWState, prChanS, &bplS));

	ERR_PASS( i_rotateGrayIP (pIPWState, pS, wS, hS, bplS, depth,
				  rotAngle,
				  (Int16) ScaledInt1616_TO_Int32(xdpi),
				  (Int16) ScaledInt1616_TO_Int32(ydpi)));
    }

    NORMAL_RETURN;
}

/*

 $Log: gsrot1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.6  1998/01/14 00:12:59  ddavies
 gsrot1.c -> /product/ipcore/ipwrap/src/RCS/gsrot1.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.5  1997/04/15 21:53:44  ddavies
 gsrot1.c -> /product/ipcore/ipwrap/src/RCS/gsrot1.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.4  1996/02/15 02:03:03  macleod
 macleod on Wed Feb 14 18:02:18 PST 1996

 gsrot1.c -> /product/ipcore/ipwrap/src/RCS/gsrot1.c,v

 Modified w_rotateGray and w_rotateGrayIP to take multi-channel images

 Revision 1.3  1996/01/04 20:23:54  ejaquith
 gsrot1.c -> /product/ipcore/ipwrap/src/RCS/gsrot1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.2  1995/12/13  04:09:08  ejaquith
 * gsrot1.c -> /product/ipcore/ipwrap/src/RCS/gsrot1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.1  1995/12/13  02:09:38  ddavies
 * Initial revision
 *
 * Revision 1.22  1995/10/13  22:01:55  ejaquith
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.21  1995/10/12  04:24:59  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.20  1995/10/05  04:27:57  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.19  1995/10/03  01:20:57  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.18  1995/09/21  22:53:36  ejaquith
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.17  1995/09/15  18:01:24  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.16  1995/09/11  22:11:58  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.15  1995/09/10  00:33:59  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.14  1995/09/07  01:32:24  ejaquith
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:20:23  ejaquith
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  01:16:26  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.11  1995/08/31  21:14:20  ejaquith
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/17  22:03:14  ejaquith
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.9  1995/07/29  04:40:39  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.8  1995/02/03  01:34:05  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Pass x and y resolutions down to the gray rotations routines so they can
 * shear anisotropic images appropriately.  The shear is needed to account
 * for the natural shear present in those images.  Also set the resolution
 * of the output image created here from the input image being rotated.
 *
 * Revision 1.7  1994/10/07  19:18:36  lperry
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.6  1994/06/08  04:11:07  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.5  1994/05/05  01:13:46  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Change return value returned by abortI from NULL to 1.
 *
 * Revision 1.4  1994/03/02  03:56:38  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Added IP_YIELD to give up control after creating dest pixr in gray rotates.
 *
 * Revision 1.3  1993/12/03  01:16:19  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:13:05  ddavies
 * gsrot.c -> /product/ipcore/ipwrap/src/RCS/gsrot.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

