/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	gsrot2.c:
 *
 *          Strip-based Gray Scale rotation routines,
 *		SMALL ANGLES ONLY! (< 10 degrees)
 *
 *          Gray Scale rotation routines:
 *		    w_getInputLinesForRotatedOutputLines()
 *		    w_getSizeOfGrayRotateInfo()
 *                  w_rotateGrayInit()
 *		    w_rotateGrayProcess()
 * 
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "alplib.h"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "rotatop.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: gsrot2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");

/* There are times when it's useful to be able to rotate images in strips.
 * This module contains the wrappers for routines that do just that.
 * If a client wishes to rotate an image in strips, there are several
 * pieces of data needed.  First, the client will have to allocate and
 * initialize a block of private data that's used to rotate each strip.
 * The routines w_getSizeOfGrayRotateInfo and w_rotateGrayInit are used
 * for these two tasks.  w_getSizeOfGrayRotateInfo is used to obtain the
 * size of the private data block.  It's the client's responsibility to
 * actually allocate it.
 * Next, the client will have to know the number of input lines
 * required to produce an output block of a given height.
 * The routine w_getInputLinesForRotatedOutputLines performs this task.
 * Finally, the client should know which input lines are needed to produce
 * a strip of output lines.  The normal way this routine will be used is
 * to have the client call w_getInputLinesForRotatedOutputLines to get the
 * number of input lines needed, say S, to produce D output lines.  The client
 * supplies the S source lines and a buffer capable of holding D output
 * lines to w_rotateGrayProcess.
 * On completion, w_rotateGrayProcess will return the number of lines
 * actually rotated (less than or equal to D) and the number of the input
 * line that should be presented at the top of the next input buffer assuming
 * the client wishes to rotate the next strip of output lines.  The client
 * is free to present some other set of input and output lines, but some
 * extra time will be required to recalculate coordinates if this is done.
 *
 * We assume the client provides the source and dest buffers in either
 * framed or frameless PIXRs.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return the number of input lines required for each rotated output line.
 *
 * LongDescription:::
 *  In rotation, each output line contains pixels from some number of
 *  input lines.  When rotating by strips, it's useful to know the number
 *  of input line required when composing input and output buffers.  This
 *  function returns that number.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  rotAngle::       Rotation angle in radians.  A positive number causes
 *		     clockwise rotation, a negative number causes
 *		     counterclockwise rotation.
 *  width::	     Width of input and output images in pixels.
 *  height::	     Total height of input image.
 *  xResolution::    Resolution of input and output images in DPI or some
 *		     other unit.  The units of xResolution and yResolution
 *		     do not matter except that xResolution and yResolution
 *		     must have the same units.  In both cases, the binary
 *		     point is between bits 15 and 16 of the 32 bit words.
 *  yResolution::    Resolution of input and output images.
 *  numOutputLinesDesired::
 *		     The number of output lines the client wishes to produce.
 *		     The answer returned will correspond to the minimum
 *		     of prD's height and numOutputLinesDesired.
 *
 * Output:::
 *  pLines::         Pointer to the number of input lines required for
 *		     to fill the output strip.  Note that we assume the
 *		     output strip is in the middle of the output image.
 *		     If it is near the top or bottom of the output image,
 *		     fewer input lines are actually used.  In no case will
 *		     the client be required to supply lines above the top
 *		     or below the bottom of the input image.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, interpolated, non-interpolated, small angle, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getInputLinesForRotatedOutputLines (
    IPWState           *pIPWState,
    Float32             rotAngle,
    Int32		width,
    Int32		height,
    ScaledInt1616	xResolution,
    ScaledInt1616	yResolution,
    Int32		numOutputLinesDesired,
    Int32              *pLines)
{
    PROLOGUE;
    Int32	lines;

    ERR_GEN(pLines == NULL, cErrNullPtr);

	/* ask the ipshared routine to do the work */
    ERR_PASS(i_getInputLinesForRotatedOutputLines (
					pIPWState, 
					width, height, numOutputLinesDesired,
					rotAngle,
					ScaledInt1616_TO_Int32(xResolution),
					ScaledInt1616_TO_Int32(yResolution),
					&lines));

	/* things seem to have gone well, so return the value calculated */
    *pLines = lines;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return the size of the private block needed for gray rotation in strips.
 *
 * LongDescription:::
 *  Gray images are big.  It is sometimes convenient to rotate them a strip
 *  at a time.  This function returns the size of a private data block
 *  that the client should allocate.  This data block will be used by
 *  each rotation operation.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prD::            Pointer to destination strip.
 *
 * Output:::
 *  pSize::          Pointer to the returned size of the private data block
 *		     in bytes.
 *
 * Notes:::
 *  The client should allocate a block of this size.  The rotation operation
 *  is implemented with one call to w_rotateGrayProcess() for each strip.
 *  When the client is done, the private data block should be freed by
 *  the client.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, interpolated, non-interpolated, small angle, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSizeOfGrayRotateInfo (
    IPWState           *pIPWState,
    PIXR               *prD,
    Int32              *pSize)
{
    PROLOGUE;
    Int32	depth;
    Int32	size;

    ERR_GEN(prD == NULL || pSize == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels1, cFrameAny);

    ERR_PASS(w_pixrGetDepth(pIPWState, prD, &depth));

	/* ask the ipshared routine to do the work */
    ERR_PASS(i_getSizeOfGrayRotateInfo (pIPWState, depth, &size));

	/* it worked, so return the size */
    *pSize = size;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Fill in the private data block needed for gray rotation by strips.
 *
 * LongDescription:::
 *  Gray images are big.  It is sometimes convenient to rotate them a strip
 *  at a time.  However, some state and tables are needed for every strip
 *  and it is silly to recreate them each time.  Clients should call
 *  w_getSizeOfGrayRotateInfo() to obtain the size of the private data block
 *  we'll use.  A block of at least this size should be passed to this
 *  routine where it will be filled with the necessary data.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prD::            Pointer to destination strip.
 *  rotAngle::	     The rotate angle in radians.  A positive number causes
 *		     clockwise rotation, a negative number causes
 *		     counterclockwise rotation.
 *  rotateInfoSize:: The size of the block being supplied in bytes.
 *  pRotateInfo::    Pointer to a block of memory containing rotateInfoSize
 *		     bytes.
 *
 * Notes:::
 *  This routine fills in the block, it does not allocate it.  That's the
 *  client's job.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, interpolated, non-interpolated, small angle, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotateGrayInit (
    IPWState           *pIPWState,
    PIXR               *prD,
    Float64		rotAngle,
    Int32		rotateInfoSize,
    void               *pRotateInfo)
{
    PROLOGUE;
    Int32		depth;
    ScaledInt1616	xRes, yRes;

    ERR_GEN(prD == NULL || pRotateInfo == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels1, cFrameAny);

    ERR_PASS(w_pixrGetDepth(pIPWState, prD, &depth));
    ERR_PASS(w_pixrGetResolutionDPI(pIPWState, prD, &xRes, &yRes));

	/* ask the ipshared routine to do the work */
    ERR_PASS(i_rotateGrayInit (pIPWState, depth, rotAngle,
			       (Int16)ScaledInt1616_TO_Int32(xRes),
			       (Int16)ScaledInt1616_TO_Int32(yRes),
			       rotateInfoSize, pRotateInfo));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform interpolated rotation of one image strip.
 *
 * LongDescription:::
 *  Gray images are big.  It is sometimes convenient to rotate them one
 *  horizontal strip at a time.  This routine performs a small-angle,
 *  interpolated rotation of one full-width strip.  We assume that the
 *  source and dest images have the same width, so the strips have the
 *  same width.  We also assume the source and dest have the same depth
 *  and height.
 *<BR><BR>
 *  This routine requires private data supplied by w_rotateGrayInit().
 *  Since each output line contains pixels from more than one input line,
 *  more input lines are needed than output lines produced.  The function
 *  w_getInputLinesForRotatedOutputLines() tells the user the number of
 *  inputs lines needed to when filling a given output block.
 *<BR><BR>
 *  The easiest and highest performance method of doing the rotation is
 *  to start from the top of the input and output images, rotating strips
 *  in order from top to bottom.  This routine will return the number of
 *  the topmost input line to be provided in the next input buffer and
 *  the number of output lines actually produced.  This routine stops
 *  when it runs out of either input or output.
 *  The number of output lines actually produced is also recorded.
 *<BR><BR>
 *  This routine tries to fill the entire output strip as defined by
 *  prD.  It stops early only if a) it runs out of input bits or b)
 *  it has produced line hS-1 (assuming the the first line in the
 *  output buffer is outputImageOffset).
 *  
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pRotateInfo::    Pointer to private rotation state block.
 *  totalHeight::    Height of the entire source image (same as dest height).
 *  prS::            Pointer to PIXR containing strip of source image.
 *  inputStripOffset::  Number of first input line of interest relative to
 *		     top of prS.  This allows us to extract a strip out
 *		     of a larger PIXR.
 *  inputImageOffset:: Number of first line in prS relative to the
 *		     top of the entire input image.  Line 0 is
 *		     the top line of the entire input image.
 *  numInputLines::  The number of valid lines in the input buffer.
 *  outputStripOffset::  The number of the first interesting line in prD
 *		     relative to the top line of prD.  This allows us to
 *		     produce a strip inside a larger output image.
 *  outputImageOffset:: The number of the first interesting line in prD
 *		     relative to the entire output image.  As before,
 *		     line 0 is the first line in the entire output image.
 *  numOutputLines:: The maximum number of output lines to produce.
 *
 * Output:::
 *  prD::	     Pointer to strip of destination image.
 *  pNumLinesProduced:: Pointer to the number of valid lines inserted
 *			into prD.
 *  pNextInputLine:: Pointer to the number of the input line that
 *		     should be placed at the top of the next source
 *		     buffer if the next strip will continue where this
 *		     one left off.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, interpolated, non-interpolated, small angle, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotateGrayProcess (
    IPWState           *pIPWState,
    void               *pRotateInfo,
    Int32		totalHeight,
    PIXR               *prS,
    Int32		inputStripOffset,
    Int32		inputImageOffset,
    Int32		numInputLines,
    Int32		outputStripOffset,
    Int32		outputImageOffset,
    Int32		numOutputLines,
    PIXR	       *prD,
    Int32	       *pNumLinesProduced,
    Int32	       *pNextInputLine)
{
    PROLOGUE;
    Int32		sDepth, dDepth, sWidth, dWidth, sBpl, dBpl;
    Int32		numLinesProduced, nextInputLine;
    UInt8	       *pS, *pD;

    ERR_GEN(prS == NULL || prD == NULL || pRotateInfo == NULL ||
	    pNumLinesProduced == NULL || pNextInputLine == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels1, cFrameAny);
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels1, cFrameAny);

	/* get and check a wad of input parameters */
    ERR_PASS(w_pixrGetDepth(pIPWState, prS, &sDepth));
    ERR_PASS(w_pixrGetDepth(pIPWState, prD, &dDepth));
    ERR_GEN(sDepth != dDepth, cErrDestDimensIncorrect);

    ERR_PASS(w_pixrGetWidth(pIPWState, prS, &sWidth));
    ERR_PASS(w_pixrGetWidth(pIPWState, prD, &dWidth));
    ERR_GEN(sWidth != dWidth, cErrDestDimensIncorrect);

    ERR_PASS(w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS(w_pixrGetBpl(pIPWState, prD, &dBpl));

    ERR_PASS(w_pixrGetImage(pIPWState, prS, &pS));
    ERR_PASS(w_pixrGetImage(pIPWState, prD, &pD));

	/* ask the ipshared routine to do the work */
    ERR_PASS(i_rotateGrayProcess (pIPWState, pRotateInfo,
				  pS + (inputStripOffset * sBpl), sWidth,
				  totalHeight, sBpl, sDepth,
				  pD + (outputStripOffset * dBpl), dBpl,
			          inputImageOffset, numInputLines,
				  outputImageOffset, numOutputLines,
				  &numLinesProduced, &nextInputLine));

	/* If we're still happy, update the args and return */
    *pNumLinesProduced = numLinesProduced;
    *pNextInputLine = nextInputLine;

    NORMAL_RETURN;
}



/*

 $Log: gsrot2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.5  1998/01/14 00:13:01  ddavies
 gsrot2.c -> /product/ipcore/ipwrap/src/RCS/gsrot2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.4  1996/10/05 02:35:58  ddavies
 ddavies on Fri Oct 4 19:35:45 PDT 1996

 gsrot2.c -> /product/ipcore/ipwrap/src/RCS/gsrot2.c,v

 Add height parameter to w_getInputLinesForRotatedOutputLines.  It's needed
 by the ipshared code underneath.

 Revision 1.3  1996/01/04 20:23:56  ejaquith
 gsrot2.c -> /product/ipcore/ipwrap/src/RCS/gsrot2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.2  1995/12/13  04:09:12  ejaquith
 * gsrot2.c -> /product/ipcore/ipwrap/src/RCS/gsrot2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.1  1995/12/13  02:11:40  ddavies
 * Initial revision
 *

 */

