/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1998-1998, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	gsrot3.c:
 *
 *          Strip-based Gray Scale rotation routines,
 *		SMALL ANGLES ONLY! (< 10 degrees)
 *
 *          Gray Scale rotation routines:
 *		    w_getInputLinesForRotatedOutputLines()
 *		    w_getSizeOfGrayRotateInfo()
 *                  w_rotateGrayInit()
 *		    w_rotateGrayProcess()
 * 
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "alplib.h"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "rotatop.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: gsrot3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");


/* This module contains interpolated rotation front-ends to image-only routines
 * located elsewhere (IPShared).
 *
 *
 *	w_rotateInterpolated()	Interpolated rotation
 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform a full-image rotation, using interpolation, on gray/color images.
 *
 * LongDescription:::
 *  An 8-bit gray or multi-channeled set of 8-bit gray images are rotated
 *  about their centers, an arbitrary angle.  The rotation uses interpolation
 *  of 1, 2, or 4 points, depending on the nearest or coinciding pixels that
 *  contribute from input to output.
 *  The image sizes of source and destination are not required to be the
 *  same; however, all points in the destination image are filled.
 *  The interpolation is currently written to work in floating-point, for
 *  correctness, but may be sped up significantly.
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  srcPixr::		Source image
 *  rotAngle::		Rotation angle in radians.  A positive number causes
 *			clockwise rotation, a negative number causes
 *			counterclockwise rotation.
 *  fillValue::		Pixel value to use when rotating from outside of the
 *			source image
 *
 * Output:::
 *  pdstPixr::		Pointer to destination image, if contents is NULL, then
 *			an image of the same size is created.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, interpolated, small angle, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotateInterpolated (
    IPWState		*pIPWState,
    PIXR		*srcPixr,
    Float32             rotAngle,
    Int32		fillValue,
    PIXR		**pdstPixr)
{
    PROLOGUE;
    PIXR *pSubPixr;			/* Multi-channel sub-PIXR */
    Int32 srcChannels, dstChannels;	/* Channel counts */
    Int32 channel;			/* Individual channel */
    ScaledInt1616 xRes, yRes;		/* Resolutions */
    Float32 fxRes, fyRes;		/* Source resolutions (floats) */
    Float32 dxRes, dyRes;		/* Destination resolutions (floats) */
    UInt8 *srcImage, *dstImage;		/* Image pointers */
    Int32 wS, hS, wD, hD;		/* Image dimensions */
    Int32 sBpl, dBpl;			/* Bytes per line */
    Int32 depth;


    ERR_GEN_PIXR(srcPixr, cByteOrderBig, cDepth8, cChannels1 |cChannels3,
		 cFrame32);
    ERR_PASS(w_pixrGetChannelCount(pIPWState, srcPixr, &srcChannels));
    if (*pdstPixr != NULL) {
	ERR_GEN_PIXR(*pdstPixr, cByteOrderBig, cDepth8, cChannels1 |cChannels3,
		     cFrame32);
	ERR_PASS(w_pixrGetChannelCount(pIPWState, *pdstPixr, &dstChannels));
	ERR_GEN(srcChannels != dstChannels, cErrChannelsNotEqual);
    } else {
	ERR_PASS(w_clonePixrNoInit(pIPWState, srcPixr, pdstPixr));
    }
    depth = 8;

    if (srcChannels > 1) {
	for (channel = 0; channel < srcChannels; channel++) {
	    pSubPixr = NULL;
	    ERR_PASS(w_pixrGetChannel(pIPWState, srcPixr, channel, &pSubPixr));
	    ERR_PASS(w_pixrGetImage(pIPWState, pSubPixr, &srcImage));
	    ERR_PASS(w_pixrGetWidth(pIPWState, pSubPixr, &wS));
	    ERR_PASS(w_pixrGetHeight(pIPWState, pSubPixr, &hS));
	    ERR_PASS(w_pixrGetBpl(pIPWState, pSubPixr, &sBpl));

	    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, pSubPixr, &xRes, &yRes));
	    fxRes = (Float32)ScaledInt1616_TO_Int32(xRes);
	    fyRes = (Float32)ScaledInt1616_TO_Int32(yRes);

	    ERR_PASS(w_pixrGetChannel(pIPWState, *pdstPixr, channel, &pSubPixr));
	    ERR_PASS(w_pixrGetImage(pIPWState, pSubPixr, &dstImage));
	    ERR_PASS(w_pixrGetWidth(pIPWState, pSubPixr, &wD));
	    ERR_PASS(w_pixrGetHeight(pIPWState, pSubPixr, &hD));
	    ERR_PASS(w_pixrGetBpl(pIPWState, pSubPixr, &dBpl));

	    /* ask the ipshared routine to do the work */
	    ERR_PASS(i_rotateInterpolated (
		pIPWState, 
		srcImage, wS, hS, sBpl, fxRes, fyRes,
		depth, fillValue, rotAngle,
		dstImage, wD, hD, dBpl, &dxRes, &dyRes
		));

	    /* Set the output resolution according to the rotation shift */
	    xRes = Int32_TO_ScaledInt1616((Int32)(dxRes + 0.5));
	    yRes = Int32_TO_ScaledInt1616((Int32)(dyRes + 0.5));
	    ERR_PASS( w_pixrSetResolutionDPI(pIPWState, *pdstPixr, xRes, yRes));
	}
    } else {

	ERR_PASS(w_pixrGetImage(pIPWState, srcPixr, &srcImage));
	ERR_PASS(w_pixrGetWidth(pIPWState, srcPixr, &wS));
	ERR_PASS(w_pixrGetHeight(pIPWState, srcPixr, &hS));
	ERR_PASS(w_pixrGetBpl(pIPWState, srcPixr, &sBpl));
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, srcPixr, &xRes, &yRes));
	fxRes = (Float32)ScaledInt1616_TO_Int32(xRes);
	fyRes = (Float32)ScaledInt1616_TO_Int32(yRes);

	ERR_PASS(w_pixrGetImage(pIPWState, *pdstPixr, &dstImage));
	ERR_PASS(w_pixrGetWidth(pIPWState, *pdstPixr, &wD));
	ERR_PASS(w_pixrGetHeight(pIPWState, *pdstPixr, &hD));
	ERR_PASS(w_pixrGetBpl(pIPWState, *pdstPixr, &dBpl));

	/* ask the ipshared routine to do the work */
	ERR_PASS(i_rotateInterpolated (
					pIPWState, 
					srcImage, wS, hS, sBpl, fxRes, fyRes,
					depth, fillValue, rotAngle,
					dstImage, wD, hD, dBpl, &dxRes, &dyRes
					));

	/* Set the output resolution according to the rotation shift */
	xRes = Int32_TO_ScaledInt1616((Int32)(dxRes + 0.5));
	yRes = Int32_TO_ScaledInt1616((Int32)(dyRes + 0.5));
	ERR_PASS( w_pixrSetResolutionDPI(pIPWState, *pdstPixr, xRes, yRes));

    }

    /* things seem to have gone well, so return */
    NORMAL_RETURN;
}


/*
 $Log: gsrot3.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.1  1998/01/16 00:34:00  aber
 Initial revision

 */

