/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  halftone.c--includes
 *
 *		w_removeHeadlines()
 *              w_segmentPageMask ()
 *		w_segmentPage ()
 *              w_segmentPagePixrs ()
 *		w_endpointsOnLines ()
 *               
 */



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "morphg.h"
#include "logicops.h"
#include "defines.pub"
#include "shrscgr.pub"
#include "shrcc.pub"
#include "thrshold.h"
#include "iaerror.pub"
#include "frames.pub"

/* prototypes */
#include "clean.pub"
#include "props.pub"
#include "rast.pub"
#include "curves.pub"
#include "fthresh.pub"
#include "prxform.pub"
#include "fill.pub"
#include "thrshold.pub"
#include "imginfo.pub"
#include "runhist.pub"
#include "halftone.pub"
#include "strel.pub"
#include "morph.pub"
#include "expand.pub"
#include "rectops.pub"
#include "logicops.pub"
#include "reduce.pub"
#include "intops.pub"
#include "ccs.pub"
#include "runmorph.pub"
#include "runfilt.pub"
#include "subsmpl.pub"
#include "binarize.pub"

#include "halftone.prv"

/* DEBUG */
#include <stdio.h>
#include "imagerep.h"
#include "io.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: halfton1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");

#define	DEBUG_HT_THRESH	0

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Attempt to suppress headlines in a segmented image
 *
 * LongDescription:::
 *   Try to keep from making pictures out of headlines: what we're
 *   going to do is look for short (less than 1" or so?) pictures
 *   and analyze their histograms. If the histogram is strongly two-peaked
 *   with not much in between, then the image is probably headline, so
 *   we'll try to convert it back to text.
 *
 * PrivateDescription:::
 *  Well, that worked allright for some images, but the problem with using
 *  variance is that it's too vulnerable to bleed-through, which raises
 *  the variance enough to make headlines appear as text. Let's try
 *  another tactic--take the run-length histogram, hoping that we'll
 *  get lots of large lines and not many small ones, i.e. 'thick' strokes.
 *  We should also expect a small variance about the peak stroke width.
 *  <BR>---<BR>
 *  So that was ok, but sometimes it gets rid of little pictures.
 *  One thing that seems to differentiate pictures & headline text
 *  is that headline text almost never has very short runs. So, I
 *  added a heuristic which sums three runs around one, sums three
 *  runs around the peak, and compares them. On images, there are usually
 *  more runs around one, even if there happens to be a peak further down.
 *  This can still fail on images which turn solid when thresholded best
 *  for text.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prGray::         The original (gray-scale) image
 *  prMask::         The mask of the alleged pictures
 *  rectP::          The list of rectangles, as found by halftoneThresh
 *
 * Output:::
 *
 * Returns:::
 *  error status
 * Notes:::
 *  warning:: I don't think anyone's using this.
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_removeHeadlines (
    IPWState           *pIPWState,
    PIXR               *prGray,
    PIXR               *prMask,
    RECTANGL           *rectP)
{
    PROLOGUE;

Int32	 peak, runHist[256], txt, firstSum, secondSum, i;
PIXR	*prTmp = NULL, *prTmp1 = NULL;
Bool	 invert;
ScaledInt1616 xDPI, yDPI;
Int32    dpi;

    /* Validate the PIXR passed */
    ERR_GEN_PIXR (prGray, cByteOrderBig, cDepth4 | cDepth8, cChannelsAny, 
		  cFrame32);
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prGray, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    for (; rectP;  rectP = rectP->next)
    {
	/* 1.5 inches? */
	if (rectP->h < dpi * 3 / 2) 
	{
	    /* Copy it out so that we can threshold it */
	    prTmp = NULL; prTmp1 = NULL;
	    ERR_PASS_CLEANUP (w_rectOf (pIPWState, prGray, rectP, &prTmp));
	    ERR_PASS_CLEANUP (w_findTextThreshold (pIPWState, prTmp, &txt,
	    &invert));
	    ERR_PASS_CLEANUP (w_threshold (pIPWState, prTmp, 0, txt, 
					   THRESH_NORM, &prTmp1));

	    /* Compute run length histo: whether we should count "log" runs
	       is questionable...most headline text won't have runs
	       that long.
	     */
	    
	    ERR_PASS_CLEANUP (w_horizRunLengthHistogram (pIPWState, prTmp1, 
							 0, NULL, cRunOfOnes,
							 256, 0, 0, 0, 
							 &runHist[0]));

	    /* Compute the peak */  
	    ERR_PASS_CLEANUP (w_findGlobalMaximum (pIPWState, runHist, 1, 255, 
						   &peak));

	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

	    if (peak >= cMinRunPeak)
	    {
		firstSum = secondSum = 0;
		for (i = 1; i < 4; i++)
		{
		    firstSum += runHist[i];
		    secondSum += runHist[i + peak - 1];
		}

		/* Take it out if it looks like it should go.. */
		if (firstSum < secondSum)
		{
		    ERR_PASS_CLEANUP (w_rasterop (
			pIPWState, 
			0, 0, NULL,
			PIX_CLR,
			rectP->x, rectP->y, rectP->w, rectP->h, prMask));
		}
	    }
	}

    }

    if (prTmp)	
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    if (prTmp1)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prTmp)	w_destroyPixr (pIPWState, &prTmp);
    if (prTmp1)	w_destroyPixr (pIPWState, &prTmp1);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  A text/image segmenter which uses two thresholds and a grayscale image
 *
 * LongDescription:::
 *  This is a similar function to w_halftoneThresh(), but it works from
 *  a grayscale image, and takes two thresholds, a "best for text" threshold
 *  and a "best for background" threshold. The images found in the
 *  "best for text" page are used to determine where we're pretty sure
 *  that there is image; the background page is used to fill out
 *  images that are often left incomplete in the text version. 
 *  The thresholds can be found by calling w_findThresholds().
 *
 * PrivateDescription:::
 * Another "Twice Cooked" segmenter. Takes a grayscale image and 2 thresholds,
 * along with the usual stuff that halftoneThreshold needs. Runs
 * halftoneThreshold twice, first on an image thresholded using the
 * text threshold passed, second on the image thresholded using the 
 * background threshold passed. The mask found using the text-thresholded
 * image is then seed-filled into the mask found using the background
 * thresholded image, thus preserving only those images which have some
 * part of them found by the text thresholder. The idea is that if
 * the background thresholded image has significant bleed-through problems,
 * the seeds in the text-thresholded thing will eliminate those
 * regions. 
 * <BR>---<BR>
 * This version of the segmenter has some additional stuff to deal with
 * bleed-through and other such problems. First, it calls textMask to
 * get a mask for the text, dilates it vertically to get blocks of text,
 * and then subtracts that from the mask. Second, it tries to find "gutters"
 * in the text-thresholded image, where a gutter is defined as a long
 * vertical or horizontal run of white. The gutters are then restricted
 * to only those whose ends touch a gutter going in the perpindicular 
 * direction, and then subtracted from the mask. If the gutters are not
 * restricted, a lot of chopping happens on images with a lot of light
 * stuff around the edges.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prs::            The source (grayscale, 4 or 8 bits/pixel) image
 *  textThresh::     "best for text" threshold for above image
 *  backThresh::     "best for background" threshold for above image
 *  redFactor::      Reduction factor (see w_halftoneThresh() ). This should
 *                   probably be inferred from the resolution.
 *  ruleWidth::      Min. Rule width (see w_halftoneThresh() )
 *  redThreshold::   Reduction threshold, 1, 2, 3 or 4 (see w_halftoneThresh())
 *  minWidth::       Discard objects smaller than this width
 *  minHeight::      ...or this height
 *  dilate::         Dilate the mask by this amount. Adds a margin
 *                   of "slop" around the mask, of the number of
 *                   pixels passed. The number can range from 0 to 4.
 *                   This is used so that we have some margin of
 *                   roundoff error for including images in masks.
 *
 * Output:::
 *  pPrText::        Text mask, if it is NULL, don't output text mask
 *  pPrResultMask::  Result mask, where a "1" in a pixel indicates imagel
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, seed filling, morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_segmentPageMask (
    IPWState            *pIPWState,
    PIXR                *prs,
    Int32                textThresh,
    Int32                backThresh,
    UInt32               redFactor,
    UInt32               ruleWidth,
    UInt32               redThreshold,
    UInt32               minWidth,
    UInt32               minHeight,
    Int32                dilate,
    PIXR               **pPrText,
    PIXR               **pPrResultMask)
{
    PROLOGUE;
    PIXR	*prText = NULL, *prReveText = NULL,
		*prTextMask = NULL,
		*prBack = NULL, *prTmp = NULL, *prTmp1 = NULL,
		*prBackMask = NULL, *prHRules = NULL, *prVRules = NULL,
		*prVGutters = NULL, *prHGutters = NULL,
		*prResultMask = NULL, *prTextOnly = NULL;
    RECTANGL	*textRects = NULL, *backRects = NULL, *rectList = NULL, *rectP;
    STREL       *pStrel10v = NULL, *pStrel20h = NULL, *pStrel20v = NULL,
	        *pStrel50h = NULL, *pStrel50v = NULL,
	        *pHDStrel = NULL, *pVDStrel = NULL;
    STREL       *pStrel5v = NULL, *pStrel5h = NULL;
    Int32        dpi, depth;
    StrelID      dilateH = strel3h, dilateV = strel3v;
    ScaledInt1616 xDPI, yDPI;
    ConnectedComponent *pCCList = NULL, *p;
	    
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));
    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prs, &depth));
    ERR_GEN (dilate < 0 || dilate > 4, cErrOutOfBounds);

    ERR_PASS (w_initStrelByName (pIPWState, strel5v));
    ERR_PASS (w_initStrelByName (pIPWState, strel5h));
    ERR_PASS (w_initStrelByName (pIPWState, strel10v));
    ERR_PASS (w_initStrelByName (pIPWState, strel20h));
    ERR_PASS (w_initStrelByName (pIPWState, strel20v));
    ERR_PASS (w_initStrelByName (pIPWState, strel50h));
    ERR_PASS (w_initStrelByName (pIPWState, strel50v));

    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel5v));
    ERR_PASS (w_getStrel (pIPWState, strel20h, &pStrel5h));
    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel10v));
    ERR_PASS (w_getStrel (pIPWState, strel20h, &pStrel20h));
    ERR_PASS (w_getStrel (pIPWState, strel20v, &pStrel20v));
    ERR_PASS (w_getStrel (pIPWState, strel50h, &pStrel50h));
    ERR_PASS (w_getStrel (pIPWState, strel50v, &pStrel50v));

    /* Create a dilation strel */
    if (dilate == 1)
    {
	dilateH = strel3h;
	dilateV = strel3v;
    }
    if (dilate == 2)
    {
	dilateH = strel5h;
	dilateV = strel5v;
    }
    else if (dilate == 3)
    {
	dilateH = strel7h;
	dilateV = strel7v;
    }
    else if (dilate == 4)
    {
	dilateH = strel9h;
	dilateV = strel9v;
    }

    ERR_PASS (w_initStrelByName (pIPWState, dilateH));
    ERR_PASS (w_initStrelByName (pIPWState, dilateV));

    ERR_PASS (w_getStrel (pIPWState, dilateH, &pHDStrel));
    ERR_PASS (w_getStrel (pIPWState, dilateV, &pVDStrel));
    
    /* Threshold for text */
    ERR_PASS (w_threshold (pIPWState, prs, 0, textThresh,
			    THRESH_NORM, &prText));

    /* Find the text */
    ERR_PASS_CLEANUP (w_textMask (pIPWState, prText, &prTextOnly, 0));

    /* Find reverse vidio text */
    prTmp = NULL;
    ERR_PASS_CLEANUP (w_invert (pIPWState, prTextOnly, &prTmp));
    if(depth == 8)
    {
	prTmp1 = NULL;
	ERR_PASS_CLEANUP (w_binarize2 (pIPWState, prs, 0, 28, &prTmp1));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp1, AND, prTmp)); 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1)); 
    }
    else
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prText, AND, prTmp)); 
    ERR_PASS_CLEANUP (w_reversetextMask(pIPWState, prTmp, &prReveText));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp)); 
    prTmp = NULL;

    /* Close vertically (a rather arbitrary amount, really) */
    ERR_PASS_CLEANUP (w_close (pIPWState, prTextOnly, pStrel10v, &prTextOnly));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prReveText, pStrel5v, &prReveText));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prReveText, pStrel5h, &prReveText));

    /* Find the image mask */
    ERR_PASS_CLEANUP (w_copy (pIPWState, prText, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReveText, SUBTRACT, prTmp));
    ERR_PASS_CLEANUP (w_halftoneThresh (pIPWState, prTmp, redFactor,
					ruleWidth, redThreshold, minWidth, 
					minHeight, &textRects, &prTextMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp)); 
    if (!prTextMask)
    {
	/* NULL back from halftoneThresh may mean that there is
	   just no text at that threshold. Just return NULL */
	goto unworked;
    }

    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &textRects));

    /* Threshold for background */
    ERR_PASS_CLEANUP (w_threshold (pIPWState, prs, 0, backThresh, 
				   THRESH_NORM, &prBack));

    /* Find a mask for the background-thresholded thing */
    ERR_PASS_CLEANUP (w_halftoneThresh (pIPWState, prBack, redFactor,
				      ruleWidth, redThreshold, minWidth, 
				      minHeight, &backRects, &prBackMask));
    if (prBackMask == NULL)
	goto unworked;

    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &backRects));

    /* Find the rectangles in the text-thresholded thing */
    ERR_PASS_CLEANUP (w_reduce2 (pIPWState, prText, 1, &prHRules));
    ERR_PASS_CLEANUP (w_open (pIPWState, prHRules, pStrel20v, &prVRules));
    ERR_PASS_CLEANUP (w_open (pIPWState, prHRules, pStrel20h, &prHRules));
    ERR_PASS_CLEANUP (w_endpointsOnLines (pIPWState, prHRules,  prVRules,
					  &prHRules));

    /*
     * Find the connected components in that, and clear lightweight ones,
     * figuring that they weren't *really* bounding the image, as in
     * forms and the like.
     */
    rectList = NULL;
    ERR_PASS_CLEANUP (w_rabb (pIPWState, prHRules, 30, 30, &rectList));
    for (rectP = rectList; rectP; rectP = rectP->next)
    {
	if (rectP->mass * 5 < rectP->w * rectP->h)
	{
	    /* Bah. Clear it out */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
					  rectP->x, rectP->y, 
					  rectP->w, rectP->h, prHRules));
	}
    }

    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &rectList));

    ERR_PASS_CLEANUP (w_expand (pIPWState, prHRules, 2, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prHRules));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prVRules));
    ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prTmp, &prTmp1));
    ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, prTextMask, 
					prTmp, 10, &prTmp1));

    /* Stick those rules into the background mask */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp1, OR, prBackMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

    /* To close image edge then filled them after take out text */
    if( dpi > 150 )
	ERR_PASS_CLEANUP (w_reduce4 (pIPWState, prBackMask, 2, 2, &prTmp));
    else
	ERR_PASS_CLEANUP (w_reduce2 (pIPWState, prBackMask, 2, &prTmp));
    ERR_PASS_CLEANUP (w_clonePixr(pIPWState, prTmp, &prTmp1));
    ERR_PASS_CLEANUP (w_findConnectedComponents (pIPWState,
						 prTmp, 10, 10,
						 0, &pCCList));
    for (p = pCCList; p != NULL ;  p = p->next)
    {
	ERR_PASS_CLEANUP (w_CCSet (pIPWState, p, 0, 0, prTmp1));
	ERR_PASS_CLEANUP (w_close (pIPWState, prTmp1, pStrel20v, &prTmp1));
	ERR_PASS_CLEANUP (w_close (pIPWState, prTmp1, pStrel20h, &prTmp1));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp1, OR, prTmp));
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR, 
				      (p->x - cFrameBits), (p->y - cFrameBits),
				      p->width, p->height, prTmp1));
    }
    ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCList));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

    if( dpi > 150 )
	ERR_PASS_CLEANUP (w_expand (pIPWState, prTmp, 4, &prTmp1));
    else
	ERR_PASS_CLEANUP (w_expand (pIPWState, prTmp, 2, &prTmp1));

    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp1, OR, prBackMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

   /* Subtract out the text & reverse text */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTextOnly, SUBTRACT, prBackMask));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReveText, SUBTRACT, prBackMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReveText));

    /* dilate big text hole */
    prTmp = NULL;
    ERR_PASS_CLEANUP (w_edgeFillSweep4 (pIPWState, prBackMask, 10, &prTmp));
    ERR_PASS_CLEANUP (w_invert (pIPWState, prTmp, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTextOnly, AND, prTmp)); 

    /* Output text mask */
    if ( pPrText != NULL )
	*pPrText = prTextOnly;
    else
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextOnly));

    ERR_PASS_CLEANUP (w_open (pIPWState, prTmp, pStrel20v, &prTmp));
    ERR_PASS_CLEANUP (w_open (pIPWState, prTmp, pStrel50h, &prTmp)); 
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmp, pStrel50h, &prTmp));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmp, pStrel20v, &prTmp));
    ERR_PASS_CLEANUP (w_invert (pIPWState, prTmp, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, AND, prBackMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

    /* Find the gutters */
    ERR_PASS_CLEANUP (w_reduce4 (pIPWState, prText, 1, 2, &prVGutters));
    ERR_PASS_CLEANUP (w_invert (pIPWState, prVGutters, &prVGutters));
    ERR_PASS_CLEANUP (w_open (pIPWState, prVGutters, pStrel50h, &prHGutters));
    ERR_PASS_CLEANUP (w_open (pIPWState, prVGutters, pStrel50v, &prVGutters));

    prTmp = NULL;
    ERR_PASS_CLEANUP (w_endpointsOnLines (pIPWState, prHGutters, 
					  prVGutters, &prHGutters));
    ERR_PASS_CLEANUP (w_expand (pIPWState, prHGutters, 4, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prHGutters));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prVGutters));

    /* Subtract the gutters */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, prBackMask));

    /* Do a seed fill from the text-thresholded thing so we get rid
       of isolated light stuff */
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prBack, &prResultMask));
    ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, prTextMask, prBackMask, 
					10, &prResultMask));

    prTmp1 = NULL;
    ERR_PASS_CLEANUP (w_getFrameMask( pIPWState, prResultMask,
				       minWidth, minHeight*2,
				       &prTmp1));

    /* Take out frame before filling hole */
    if( prTmp1 != NULL )
	ERR_PASS_CLEANUP(w_logicOp (pIPWState, prTmp1, SUBTRACT, 
				    prResultMask));

    /* Fill the holes */
    ERR_PASS_CLEANUP (w_edgeFillSweep4 (pIPWState, prResultMask, 10, &prTmp));
    ERR_PASS_CLEANUP (w_invert (pIPWState, prTmp, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, OR, prResultMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

    if( prTmp1 )
    {
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp1, OR, prResultMask));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
    }

    /* Clean up */
    if (prText) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prText));
    if (prTextMask) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));
    if (prBack) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBack));
    if (prBackMask) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBackMask));

    /* Dilate the mask, if necessary */
    if (dilate)
    {
	prTmp = NULL;
	ERR_PASS_CLEANUP (
	    w_dilate (pIPWState, prResultMask, pHDStrel, &prTmp));
	ERR_PASS_CLEANUP (
	    w_dilate (pIPWState, prTmp, pVDStrel, &prResultMask));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    }
	
    *pPrResultMask = prResultMask;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prText) w_destroyPixr (pIPWState, &prText);
    if (prBack) w_destroyPixr (pIPWState, &prBack);
    if (prTextMask) w_destroyPixr (pIPWState, &prTextMask);
    if (prBackMask) w_destroyPixr  (pIPWState, &prBackMask);
    if (prResultMask) w_destroyPixr (pIPWState, &prResultMask);
    if (prTextOnly && pPrText ) w_destroyPixr (pIPWState, &prTextOnly); 
    if (prReveText) w_destroyPixr (pIPWState, &prReveText);
    if (rectList) w_destroyRectList (pIPWState, &rectList); 
    if (backRects) w_destroyRectList (pIPWState, &backRects);
    if (textRects) w_destroyRectList (pIPWState, &textRects);
    if (prTmp) w_destroyPixr (pIPWState, &prTmp);
    if (prTmp1) w_destroyPixr (pIPWState, &prTmp1);
    *pPrResultMask = NULL;
    ERR_RETURN;

/*
 * This is where it didn't actually find anything, but we got in trouble
 * for saying it "failed" because stupid regression test thinks that 
 * that's a dirty word. So don't call abortP at all.
 */
unworked:
    if (prText) w_destroyPixr (pIPWState, &prText);
    if (prBack) w_destroyPixr (pIPWState, &prBack);
    if (prTextMask) w_destroyPixr (pIPWState, &prTextMask);
    if (prBackMask) w_destroyPixr  (pIPWState, &prBackMask);
    if (prResultMask) w_destroyPixr (pIPWState, &prResultMask);
    if (prTextOnly) w_destroyPixr (pIPWState, &prTextOnly);
    if (prReveText) w_destroyPixr (pIPWState, &prReveText);
    if (rectList) w_destroyRectList (pIPWState, &rectList); 
    if (backRects) w_destroyRectList (pIPWState, &backRects);
    if (textRects) w_destroyRectList (pIPWState, &textRects);
    *pPrResultMask = NULL;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  A text/image segmenter which uses two thresholds and a grayscale image
 *
 * LongDescription:::
 *  This is a wrapper to w_segmentPageMask() which makes a list of rectangles
 *  and fills in the rectangles if the region they cover is "almost"
 *  rectangular. 
 *
 * PrivateDescription:::
 * 
 * Input:::
 *  pIPWState::      State structure
 *  prs::            The source (grayscale, 4 or 8 bits/pixel) image
 *  textThresh::     "best for text" threshold for above image
 *  backThresh::     "best for background" threshold for above image
 *  redFactor::      Reduction factor (see w_halftoneThresh()). This should
 *                   probably be inferred from the resolution.
 *  ruleWidth::      Min. Rule width (see w_halftoneThresh())
 *  redThreshold::   Reduction threshold, 1, 2, 3 or 4 (see w_halftoneThresh())
 *  minWidth::       Discard objects smaller than this width
 *  minHeight::      ...or this height
 *  dilate::         Dilate the mask by this amount. Adds a margin
 *                   of "slop" around the mask, of the number of
 *                   pixels passed. The number can range from 0 to 4.
 *                   This is used so that we have some margin of
 *                   roundoff error for including images in masks.
 *
 * Output:::
 *  pRects::         List of rectangles, one for each CC in result mask
 *  pPrResultMask::  Result mask, where a "1" in a pixel indicates image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, seed filling, morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_segmentPage (
    IPWState            *pIPWState,
    PIXR                *prs,
    Int32                textThresh,
    Int32                backThresh,
    UInt32               redFactor,
    UInt32               ruleWidth,
    UInt32               redThreshold,
    UInt32               minWidth,
    UInt32               minHeight,
    Int32                dilate,
    RECTANGL           **pRects,
    PIXR               **pPrResultMask)
{
    PROLOGUE;
    RECTANGL *rectP, *rectList = NULL;
    Bool      isRect;
    PIXR     *prResultMask = NULL, *prText = NULL;

    /* Segment the page */
    ERR_PASS (w_segmentPageMask (pIPWState, prs, textThresh, backThresh,
				 redFactor, ruleWidth, redThreshold,
				 minWidth, minHeight, dilate,
				 &prText, &prResultMask));

    /* If we got a segment mask, find bounding boxes in it.  We might
     * not get a segment mask.  If that happens, just recList will stay
     * null and we'll fall through. */
    rectList = NULL;
    if (prResultMask != NULL)
	ERR_PASS_CLEANUP (w_rabb (pIPWState, prResultMask, minWidth, minHeight,
				  &rectList));
	
    /* not finding a rectlist is not considered an error. */
    if (rectList)
    {
	/* Fill in rectangles in which the image is rectangular */
	for (rectP = rectList; rectP; rectP = rectP->next)
	{
	    ERR_PASS_CLEANUP (w_maskIsRect (pIPWState, rectP, prResultMask, 
					    0, 0, prText, 
					    cIsRectCutoff, &isRect));

	    if (isRect)
	    {
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_SET,
					      rectP->x, rectP->y, 
					      rectP->w, rectP->h,
					      prResultMask));
	    }
	}
    }
    if (prText) ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prText));

    *pRects = rectList;
    *pPrResultMask = prResultMask;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (rectList) w_destroyRectList (pIPWState, &rectList);
    if (prResultMask) w_destroyPixr (pIPWState, &prResultMask);
    if (prText) w_destroyPixr (pIPWState, &prText);

    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  A text/image segmenter which uses two thresholds and a grayscale image
 *
 * LongDescription:::
 *  This is a wrapper to w_segmentPageMask() which makes an array of
 *  PixrRef structures, each of which contains an (x,y) location and
 *  a pointer to a PIXR. Each PIXR is the mask of a single image (connected
 *  component). This also fills in the masks if they're "almost" rectangles. 
 *  The (x,y) offsets are relative to the IMAGE, not the FRAME, which means
 *  that they can be passed to ipwrap rasterop-type routines, but need to
 *  have the frame added back in if using ipshared routines.
 *
 * PrivateDescription:::
 * 
 * Input:::
 *  pIPWState::      State structure
 *  prs::            The source (grayscale, 4 or 8 bits/pixel) image
 *  textThresh::     "best for text" threshold for above image
 *  backThresh::     "best for background" threshold for above image
 *  redFactor::      Reduction factor (see w_halftoneThresh()). This should
 *                   probably be inferred from the resolution.
 *  ruleWidth::      Min. Rule width (see w_halftoneThresh())
 *  redThreshold::   Reduction threshold, 1, 2, 3 or 4 (see w_halftoneThresh())
 *  minWidth::       Discard objects smaller than this width
 *  minHeight::      ...or this height
 *  dilate::         Dilate the mask by this amount. Adds a margin
 *                   of "slop" around the mask, of the number of
 *                   pixels passed. The number can range from 0 to 4.
 *                   This is used so that we have some margin of
 *                   roundoff error for including images in masks.
 *
 * Output:::
 *  ppImageMasks::   Array of PixrRef structures
 *  pMaskCount::     Number of image masks created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, seed filling, morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_segmentPagePixrs (
    IPWState            *pIPWState,
    PIXR                *prs,
    Int32                textThresh,
    Int32                backThresh,
    UInt32               redFactor,
    UInt32               ruleWidth,
    UInt32               redThreshold,
    UInt32               minWidth,
    UInt32               minHeight,
    Int32                dilate,
    PixrRef            **ppImageMasks,
    Int32               *pMaskCount)
{
    PROLOGUE;
    Bool                isRect;
    PIXR               *prResultMask = NULL, *prTmp = NULL, *prText = NULL;
    PixrRef            *pImageMasks = NULL, *pNewMasks = NULL;
    ConnectedComponent *pCCList = NULL, *p;
    Int32               count = 0, i, w, h, j, x, y;
    Int32               newCount, *whList = NULL;
    ScaledInt1616       xDPI, yDPI;

    /* Segment the page */
    ERR_PASS (w_segmentPageMask (pIPWState, prs, textThresh, backThresh,
				 redFactor, ruleWidth, redThreshold,
				 minWidth, minHeight, dilate, 
				 &prText, &prResultMask));

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xDPI, &yDPI));

    pCCList = NULL;
    count   = 0;
    pImageMasks = NULL;

    if( prResultMask != NULL)
    {
    /* Find a list of connected components */
	ERR_PASS_CLEANUP (w_findConnectedComponents (pIPWState,
						 prResultMask,
						 minWidth, minHeight,
						 1, &pCCList));
	if (pCCList)
	{
	    for (count = 0, p = pCCList; p; p = p->next)
	    {
		count++;
	    }
	
	    /* Create the pixr array */
	    pImageMasks = (PixrRef *) 
		IPW_MALLOC (count * sizeof (PixrRef), cAllocatorData);
	    ERR_GEN_CLEANUP (pImageMasks == NULL, ia_nomem);

	    whList = (Int32 *) 
		IPW_MALLOC (count * sizeof (Int32)*3, cAllocatorData);
	    ERR_GEN_CLEANUP (whList == NULL, ia_nomem);

	    /* Create and paint the PIXRs */
	    for (i = 0, p = pCCList; i < count; i++, p = p->next)
	    {
		pImageMasks[i].xOffset = p->x - cFrameBits;
		pImageMasks[i].yOffset = p->y - cFrameBits;
		whList[i] = p->width + pImageMasks[i].xOffset;
		whList[count+i]=  p->height + pImageMasks[i].yOffset;
		prTmp = NULL;
		ERR_PASS_CLEANUP (w_createPixr (pIPWState, p->width, p->height,
					    1, xDPI, yDPI, cFrameBits,
					    &prTmp));

		ERR_PASS_CLEANUP (w_CCSet (pIPWState, p, -(p->x) + cFrameBits,
				       -(p->y) + cFrameBits, prTmp));
		pImageMasks[i].pPixr = prTmp;

		/* See if it's rectangular */
		ERR_PASS_CLEANUP (w_maskIsRect (pIPWState, NULL, prTmp, 
						pImageMasks[i].xOffset,
						pImageMasks[i].yOffset, prText,
						cIsRectCutoff, &isRect));
		whList[count+count+i] = (Int32)isRect;
		if (isRect)
		{
		    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prTmp, &w));
		    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prTmp, &h));
		    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL,
						  PIX_SET, 0, 0, w, h, prTmp));
		}
	    }
	}
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prResultMask));
	ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCList));
    }
    if(prText) ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prText));

/* 4-23-98 To check small rectangle inside big rectangle */
    newCount = count;
    for (i = 0; i <count; i++)
    {
	w = whList[i];
	h = whList[i+count];
	x = pImageMasks[i].xOffset;
	y = pImageMasks[i].yOffset;
	for(j= 0; j< count; ++j)
	{
	    if( j == i ) continue;
	    if( j < i && pImageMasks[j].pPixr == NULL)
		continue;
	    if( whList[count+j+count] && 
		pImageMasks[j].xOffset <= x && 
		pImageMasks[j].yOffset <= y &&
		whList[j] >= w && 
		whList[j+count] >= h)
	    {
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, 
						 &pImageMasks[i].pPixr));
		pImageMasks[i].pPixr = NULL;
		--newCount;
		break;
	    }
	}
    }
    IPW_FREE(whList, cAllocatorData);
    if( newCount < count )
    {
	pNewMasks = pImageMasks;
	pImageMasks = (PixrRef *) 
	    IPW_MALLOC (newCount * sizeof (PixrRef), cAllocatorData);
	ERR_GEN_CLEANUP (pImageMasks == NULL, ia_nomem);
	for(i=0, j=0; i<count; ++i)
	{
	    if( pNewMasks[i].pPixr == NULL ) continue;
	    pImageMasks[j] = pNewMasks[i];
	    ++ j;
	}
	IPW_FREE(pNewMasks, cAllocatorData);
	count = newCount;
    }

    *ppImageMasks = pImageMasks;
    *pMaskCount = count;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prResultMask) w_destroyPixr (pIPWState, &prResultMask);
    if (pCCList) w_destroyConnectedComponents (pIPWState, &pCCList);
    if (prText) w_destroyPixr (pIPWState, &prText);
    if (whList) IPW_FREE(whList, cAllocatorData); 
    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Supress rules which are not abutted by orthogonal rules
 *
 * LongDescription:::
 *  This routine takes a map of horizontal and vertical runs, and
 *  combines them, suppressing runs whose endpoints are not touched by
 *  perpindicular lines.
 * 
 * PrivateDescription:::
 *  Works by filling into the horizontal left and right
 *  using the vertical as a seed, ANDing the result, filling the
 *  vertical up and down using the horizontals as a seed, ANDing
 *  that result, and then ORing those two results. What we end up
 *  getting is lines whose endpoints touched perpindicular lines.
 *
 * Input:::
 *  pIPWState::      The state structure
 *  prHRules::       Bitmap containing horizontal rules 
 *  prVRules::       Bitmap containing vertical rules
 *
 * Output:::
 *  pPrResult::      Combined h/v rules which touch orthogonal rules on
 endpoints
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This is used by segmentPage; it is exported because I figured that
 *  others might find it useful.
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_endpointsOnLines (
    IPWState           *pIPWState,
    PIXR               *prHRules,
    PIXR               *prVRules,
    PIXR              **pPrResult)
{
    PROLOGUE;
    PIXR    *prLeft = NULL, *prRight = NULL, *prUp = NULL, *prDown = NULL;

    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prHRules, &prLeft));
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prHRules, &prRight));
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prHRules, &prUp));
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, prHRules, &prDown));

    /* Fill left and right using vertical as a seed */
    ERR_PASS_CLEANUP (w_fillClipSweep2H (pIPWState, prVRules, prHRules, 
					 FILL_LEFT, &prLeft));
    ERR_PASS_CLEANUP (w_fillClipSweep2H (pIPWState, prVRules, prHRules,
					 FILL_RIGHT, &prRight));
    
    /* Take the intersection to get only those lines which touch 
       verticals at both endpoints */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prRight, AND, prLeft));

    /* Do the same for the horizontals */
    ERR_PASS_CLEANUP (w_fillClipSweep2V (pIPWState, prHRules, prVRules, 
					 FILL_UP, &prUp));
    ERR_PASS_CLEANUP (w_fillClipSweep2V (pIPWState, prHRules, prVRules, 
					 FILL_DOWN, &prDown));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prDown, AND, prUp));

    /* Combine them */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prUp, OR, prLeft));

    if (*pPrResult == NULL)
	*pPrResult = prLeft;
    else
    {
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prLeft, COPY, *pPrResult));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prLeft));
    }

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRight));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prUp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prDown));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prLeft) w_destroyPixr (pIPWState, &prLeft);
    if (prRight) w_destroyPixr (pIPWState, &prRight);
    if (prUp) w_destroyPixr (pIPWState, &prUp);
    if (prDown) w_destroyPixr (pIPWState, &prDown);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  A image painting mask segmenter which uses two thresholds and 
 *  a grayscale image
 *
 * LongDescription:::
 *  This is a similar function to w_segmentPageMask(), but it works for
 *  generating image mask which was painted to binary image.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prs::            The source (grayscale, 4 or 8 bits/pixel) image
 *  textThresh::     "best for text" threshold for above image
 *  backThresh::     "best for background" threshold for above image
 *  redFactor::      Reduction factor (see w_halftoneThresh() ). This should
 *                   probably be inferred from the resolution.
 *  ruleWidth::      Min. Rule width (see w_halftoneThresh() )
 *  redThreshold::   Reduction threshold, 1, 2, 3 or 4 (see w_halftoneThresh())
 *  minWidth::       Discard objects smaller than this width
 *  minHeight::      ...or this height
 *
 * Output:::
 *  pPrText::        Text mask, if it is NULL, don't output text mask
 *  pPrResultMask::  Result mask, where a "1" in a pixel indicates imagel
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, seed filling, morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_segmentPagePaintingMask (
    IPWState            *pIPWState,
    PIXR                *prs,
    Int32                textThresh,
    Int32                backThresh,
    UInt32               redFactor,
    UInt32               ruleWidth,
    UInt32               redThreshold,
    UInt32               minWidth,
    UInt32               minHeight,
    PIXR               **pPrText,
    PIXR               **pPrResultMask)
{
    PROLOGUE;
    PIXR	*prText = NULL, *prReveText = NULL,
		*prTextMask = NULL,
		*prBack = NULL, *prTmp = NULL, *prTmp1 = NULL,
		*prBackMask = NULL,
		*prVGutters = NULL, *prHGutters = NULL,
		*prResultMask = NULL, *prTextOnly = NULL;
    RECTANGL	*textRects = NULL, *backRects = NULL, *rectList = NULL;
    STREL       *pStrel10v = NULL, *pStrel20h = NULL, *pStrel20v = NULL,
	        *pStrel50h = NULL, *pStrel50v = NULL;
    STREL       *pStrel5v = NULL, *pStrel5h = NULL;
    Int32        dpi, depth;
    ScaledInt1616 xDPI, yDPI;
    ConnectedComponent *pCCList = NULL, *pCCMask = NULL, *p, *p1;
    Int32        x0, x1, y0, y1;
    Int32        mind, idx, d;

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));
    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prs, &depth));

    ERR_PASS (w_initStrelByName (pIPWState, strel5v));
    ERR_PASS (w_initStrelByName (pIPWState, strel5h));
    ERR_PASS (w_initStrelByName (pIPWState, strel10v));
    ERR_PASS (w_initStrelByName (pIPWState, strel20h));
    ERR_PASS (w_initStrelByName (pIPWState, strel20v));
    ERR_PASS (w_initStrelByName (pIPWState, strel50h));
    ERR_PASS (w_initStrelByName (pIPWState, strel50v));

    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel5v));
    ERR_PASS (w_getStrel (pIPWState, strel20h, &pStrel5h));
    ERR_PASS (w_getStrel (pIPWState, strel10v, &pStrel10v));
    ERR_PASS (w_getStrel (pIPWState, strel20h, &pStrel20h));
    ERR_PASS (w_getStrel (pIPWState, strel20v, &pStrel20v));
    ERR_PASS (w_getStrel (pIPWState, strel50h, &pStrel50h));
    ERR_PASS (w_getStrel (pIPWState, strel50v, &pStrel50v));

    /* Threshold for text */
    if( depth == 8)
	ERR_PASS_CLEANUP (w_binarize2 (pIPWState, prs, 0, 28, &prText));
    else
	ERR_PASS_CLEANUP (w_threshold (pIPWState, prs, 0, textThresh, 
				       THRESH_NORM, &prText));
    /* Find the text */
    ERR_PASS_CLEANUP (w_textMask (pIPWState, prText, &prTextOnly, 0));

    /* Find reverse video text */
    prTmp = NULL;
    ERR_PASS_CLEANUP (w_invert (pIPWState, prTextOnly, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prText, AND, prTmp)); 
    ERR_PASS_CLEANUP (w_reversetextMask(pIPWState, prTmp, &prReveText));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp)); 

    /* Close vertically (a rather arbitrary amount, really) */
    ERR_PASS_CLEANUP (w_close (pIPWState, prTextOnly, pStrel20v, &prTextOnly));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prReveText, pStrel5v, &prReveText));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prReveText, pStrel5h, &prReveText));

    /* Find the image mask */
    ERR_PASS_CLEANUP (w_copy (pIPWState, prText, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReveText, SUBTRACT, prTmp));
    ERR_PASS_CLEANUP (w_halftoneThresh (pIPWState, prTmp, redFactor,
					ruleWidth, redThreshold, minWidth, 
					minHeight, &textRects, &prTextMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp)); 
    if (!prTextMask)
    {
	/* NULL back from halftoneThresh may mean that there is
	   just no text at that threshold. Just return NULL */
	goto unworked;
    }

    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &textRects));

    /* Threshold for background */
    ERR_PASS_CLEANUP (w_threshold (pIPWState, prs, 0, backThresh, 
				   THRESH_NORM, &prBack));

    /* Find a mask for the background-thresholded thing */
    ERR_PASS_CLEANUP (w_halftoneThresh (pIPWState, prBack, redFactor,
				      ruleWidth, redThreshold, minWidth, 
				      minHeight, &backRects, &prBackMask));
    if (prBackMask == NULL)
	goto unworked;

    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &backRects));

   /* Subtract out the text & reverse text */
    ERR_PASS_CLEANUP (w_copy (pIPWState, prBackMask, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTextOnly, AND, prTmp));
    ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, prTmp, prBackMask, 
					10, &prTmp1));

    ERR_PASS_CLEANUP (w_dilate(pIPWState, prTextOnly, pStrel10v, &prTextOnly));
    ERR_PASS_CLEANUP (w_dilate(pIPWState, prTextOnly, pStrel20h, &prTextOnly));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTextOnly, SUBTRACT, prBackMask));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReveText, SUBTRACT, prBackMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReveText));

    /* Output text mask */
    if ( pPrText != NULL )
	*pPrText = prTextOnly;
    else
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextOnly));

    /* Clearn up the text inside background mask */
    ERR_PASS_CLEANUP (w_findConnectedComponents (pIPWState,
						 prTmp, 50, 20,
						 0, &pCCList));
    ERR_PASS_CLEANUP (w_findConnectedComponents (pIPWState,
						 prTmp1, 50, 20,
						 0, &pCCMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

    for (p = pCCList; p != NULL ;  p = p->next)
    {
	x0 = p->x;
	y0 = p->y;
	x1 = p->x + p->width;
	y1 = p->y + p->height;
	for(p1 = pCCMask; p1 != NULL; p1 = p1->next)
	{
	    if( (p1->x >= x1) || (p1->y >= y1) ||
	       ((p1->x + p1->width) <= x0) || ((p1->y+p1->height)<=y0))
		continue;

	    mind = x0- p1->x;
	    idx = 0;
	    if( (d = p1->x + p1->width - x1) < mind )
	    {
		mind = d;
		idx = 1;
	    }
	    if( (d = y0 - p1->y ) < mind )
	    {
		mind = d;
		idx = 2;
	    }
	    if( (d = p1->y + p1->height - y1) < mind )
	    {
		mind = d;
		idx = 3;
	    }
	    if(mind > 0)
	    {
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
					      NULL, PIX_CLR, 
					      (x0 - cFrameBits), 
					      (y0 - cFrameBits),
					      p->width, p->height, 
					      prBackMask));
		switch (idx )
		{
		case 0:
		    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
						  NULL, PIX_CLR, 
						  (p1->x - cFrameBits), 
						  (p->y - cFrameBits),
						  mind+1, p->height, 
						  prBackMask));
		    break;
		case 1:
		    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
						  NULL, PIX_CLR, 
						  (x1 - cFrameBits), 
						  (p->y - cFrameBits),
						  mind+1, p->height, 
						  prBackMask));
		    break;
		case 2:
		    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
						  NULL, PIX_CLR, 
						  (p->x - cFrameBits), 
						  (p1->y - cFrameBits),
						  p->width, mind+1, 
						  prBackMask));
		    break;
		case 3:
		    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
						  NULL, PIX_CLR, 
						  (p->x - cFrameBits), 
						  (y1 - cFrameBits),
						  p->width, mind+1, 
						  prBackMask));
		    break;
		}
	    }
	    break;
	}

    }
    ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCList));
    ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCMask));

    /* Find the gutters */
    ERR_PASS_CLEANUP (w_reduce4 (pIPWState, prText, 1, 2, &prVGutters));
    ERR_PASS_CLEANUP (w_invert (pIPWState, prVGutters, &prVGutters));
    ERR_PASS_CLEANUP (w_open (pIPWState, prVGutters, pStrel50h, &prHGutters));
    ERR_PASS_CLEANUP (w_open (pIPWState, prVGutters, pStrel50v, &prVGutters));

    prTmp = NULL;
    ERR_PASS_CLEANUP (w_endpointsOnLines (pIPWState, prHGutters, 
					  prVGutters, &prHGutters));
    ERR_PASS_CLEANUP (w_expand (pIPWState, prHGutters, 4, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prHGutters));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prVGutters));

    /* Subtract the gutters */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, prBackMask));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

    ERR_PASS_CLEANUP (w_findConnectedComponents (pIPWState,
						 prBackMask, 40, 20,
						 0, &pCCList));
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, prBackMask, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBackMask));

    for (p = pCCList; p != NULL ;  p = p->next)
    {
	ERR_PASS_CLEANUP (w_CCSet (pIPWState, p, 0, 0, prTmp));
    }
    ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCList));
    prBackMask = prTmp;
    prTmp = NULL;

    /* Do a seed fill from the text-thresholded thing so we get rid
       of isolated light stuff */
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, prBack, &prResultMask));
    ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, prTextMask, prBackMask, 
					10, &prResultMask));

     /* Clean up */
    if (prText) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prText));
    if (prTextMask) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));
    if (prBack) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBack));
    if (prBackMask) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBackMask));

    *pPrResultMask = prResultMask;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prText) w_destroyPixr (pIPWState, &prText);
    if (prBack) w_destroyPixr (pIPWState, &prBack);
    if (prTextMask) w_destroyPixr (pIPWState, &prTextMask);
    if (prBackMask) w_destroyPixr  (pIPWState, &prBackMask);
    if (prResultMask) w_destroyPixr (pIPWState, &prResultMask);
    if (prTextOnly && pPrText ) w_destroyPixr (pIPWState, &prTextOnly); 
    if (prReveText) w_destroyPixr (pIPWState, &prReveText);
    if (rectList) w_destroyRectList (pIPWState, &rectList); 
    if (backRects) w_destroyRectList (pIPWState, &backRects);
    if (textRects) w_destroyRectList (pIPWState, &textRects);
    if (prTmp) w_destroyPixr (pIPWState, &prTmp);
    if (prTmp1) w_destroyPixr (pIPWState, &prTmp1);
    *pPrResultMask = NULL;
    ERR_RETURN;

/*
 * This is where it didn't actually find anything, but we got in trouble
 * for saying it "failed" because stupid regression test thinks that 
 * that's a dirty word. So don't call abortP at all.
 */
unworked:
    if (prText) w_destroyPixr (pIPWState, &prText);
    if (prBack) w_destroyPixr (pIPWState, &prBack);
    if (prTextMask) w_destroyPixr (pIPWState, &prTextMask);
    if (prBackMask) w_destroyPixr  (pIPWState, &prBackMask);
    if (prResultMask) w_destroyPixr (pIPWState, &prResultMask);
    if (prTextOnly) w_destroyPixr (pIPWState, &prTextOnly);
    if (prReveText) w_destroyPixr (pIPWState, &prReveText);
    if (rectList) w_destroyRectList (pIPWState, &rectList); 
    if (backRects) w_destroyRectList (pIPWState, &backRects);
    if (textRects) w_destroyRectList (pIPWState, &textRects);
    *pPrResultMask = NULL;
    NORMAL_RETURN;
}

/*
 $Log: halfton1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.3  1998/10/05 18:17:34  xzhu
 xzhu on Mon Oct 5 11:15:10 PDT 1998

 halfton1.c -> /product/ipcore/ipwrap/src/RCS/halfton1.c,v

 Fixed memory initial problem.

 Revision 1.2  1998/10/02 21:30:54  xzhu
 xzhu on Fri Oct 2 14:30:45 PDT 1998

 halfton1.c -> /product/ipcore/ipwrap/src/RCS/halfton1.c,v

 Added a new function.

 Revision 1.1  1998/08/05 18:11:53  xzhu
 Initial revision

 Revision 1.75  1998/08/04 23:18:55  xzhu
 xzhu on Tue Aug 4 16:18:33 PDT 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Added wp_getFrameMask.

 Revision 1.74  1998/08/04 00:22:13  xzhu
 xzhu on Mon Aug 3 17:21:46 PDT 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Follow change of w_maskIsRect.

 Revision 1.73  1998/04/24 00:51:09  xzhu
 xzhu on Thu Apr 23 17:50:01 PDT 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Added checking for masks. If mask is inside other mask, clean up this mask.

 Revision 1.72  1998/03/19 03:01:53  xzhu
 xzhu on Wed Mar 18 19:00:28 PST 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Improved image segmentation by closing small holes.

 Revision 1.71  1997/06/26 20:01:21  ddavies
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.70  1997/05/14 21:04:06  xzhu
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 CommonCheck: If we just check edge to determine whether the image mask is rectangle, we may delete inside text. So we add something to check whether there is text inside to reduce segmentation errors.

 Revision 1.69  1997/04/15 21:54:02  ddavies
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.68  1997/03/29 01:25:55  xzhu
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 CommonCheck: Add a input parameter IsTextSeg in w_textMask(). If want keep more text, turn on this parameter, otherwise set it 0.

 Revision 1.67  1997/03/26 03:19:54  ddavies
 ddavies on Tue Mar 25 19:19:15 PST 1997

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Eliminate call to w_maskIsRect in w_segmentBinaryPagePixrs.  It was taking
 almost half of the total time and not really doing anything useful.

 Revision 1.66  1997/03/25 02:40:13  ddavies
 ddavies on Mon Mar 24 18:39:28 PST 1997

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Another iteration on w_segmentBinaryPagePixrs.  This one does a better
 job on text inside image blocks, reverse video text and text-on-tint.
 Well, what's left anyway?

 Revision 1.65  1997/01/31 00:53:30  ddavies
 ddavies on Thu Jan 30 16:51:34 PST 1997

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Add new procedure w_segmentBinaryPagePixrs.  This is a routine for segmenting
 binary images.  It attempts to find pictures in images, then remove text
 when it's on stipple and remove reverse video text from the images.
 The list of images returned should be XORed with the original image to
 remove them.  This converts the reverse video text to positive video text
 for the text plane.
 Unlike w_segmentPagePixrs, this routine returns the real image bits, not
 the image masks.

 Revision 1.64  1996/04/29 15:05:04  macleod
 macleod on Mon Apr 29 08:04:19 PDT 1996

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Added "dilate" parameter to segmentPage, etc. to support a slight
 dilation of the resulting masks. This is used to get a bit of "slop"
 around the mask to deal with the problems of multiple-resolution
 segmentation.

 Revision 1.63  1996/03/27 00:39:27  xzhu
 xzhu on Tue Mar 26 16:38:53 PST 1996

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Fixed memory leak reported from PURIFY.

 Revision 1.62  1996/03/25 17:30:26  xzhu
 xzhu on Mon Mar 25 09:29:53 PST 1996

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Use w_reversetextMask in w_segmentPageMask().

 Revision 1.61  1996/01/27 22:37:41  ejaquith
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 updated call to w_xformRectList()

 * Revision 1.60  1996/01/20  00:33:48  xzhu
 * xzhu on Fri Jan 19 16:31:33 PST 1996
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added protection in w_segmentPagePirxs. When w_segmentPageMask return
 nothing,
 * w_segmentPagePirxs will crash. Modified w_segmentPageMask to improve
 performance.
 *
 * Revision 1.59  1996/01/04  20:24:00  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.58  1995/12/13  04:09:16  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.57  1995/11/30  17:18:04  macleod
 * macleod on Thu Nov 30 09:17:52 PST 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Changed comment headers, to shut up the nag script
 *
 * Revision 1.56  1995/11/16  19:47:22  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Made comments agree with comments in header file
 *
 * Revision 1.55  1995/11/13  20:12:46  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.54  1995/11/04  01:35:55  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing
 ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.53  1995/10/17  23:48:34  ddavies
 * ddavies on Tue Oct 17 16:47:36 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fix bug in w_segmentPage in which an error was detected when
 w_segmentPageMask
 * decided that there weren't any segments (so returned prResultMask == NULL),
 * and we called rabb on prResultMask anyway.
 *
 * Revision 1.52  1995/10/13  22:01:59  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.51  1995/10/12  14:17:44  macleod
 * macleod on Thu Oct 12 07:16:16 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added segmentPagePixrs, which segments a page like segmentPage did,
 * only outputs an array of PIXRs and (x,y) offsets, one for each
 * segment, rather than a single bitmap and rectangle list, like
 * segmentPage did. Also split segmentPage into two modules, segmentPage
 * and segmentPageMask; the former is the original interface, the latter
 * only puts out a mask, and is called by both segmentPage and
 * segmentPagePixrs.
 *
 * Revision 1.50  1995/10/03  01:58:53  macleod
 * macleod on Mon Oct 2 18:58:53 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in
 interfaces.
 *
 * Revision 1.49  1995/09/21  22:53:39  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.48  1995/09/20  00:05:56  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fixed bug in segmentPage caused by change in w_threshold semantics
 *
 * Revision 1.47  1995/09/19  19:19:02  macleod
 * macleod on Tue Sep 19 12:19:02 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi
 parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.46  1995/09/11  21:53:41  macleod
 * macleod on Mon Sep 11 14:53:40 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.45  1995/09/07  01:32:27  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's
 declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.44  1995/09/06  20:20:26  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.43  1995/09/05  19:00:59  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Removed calls to GET_STREL macro, instead calling getStrel()
 function.
 * This required using local variables rather than invoking GET_STREL
 * as an argument to another function, since getStrel can return an error
 * code which should be checked.
 *
 * Revision 1.42  1995/09/01  21:02:10  ejaquith
 * ejaquith on Fri Sep 1 14:02:10 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.41  1995/08/31  21:14:23  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.40  1995/08/29  00:10:44  ejaquith
 * ejaquith on Mon Aug 28 17:08:46 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fixed a ton o' bugs/typos/etc that "segment" target in ipwtest was
 * generating.
 * zeroPr() was being used incorrectly in 2 instances.
 * rabb() wasn't being error checked in 2 instances.
 * result pixr was getting set to NULL when it shouldn't be resulting in
 * leaks and errors.
 *
 * Revision 1.39  1995/08/28  22:43:00  jcraig
 * jcraig on Mon Aug 28 15:42:47 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * removed include of debug.h - which is not necessary
 *
 * Revision 1.38  1995/08/25  01:36:42  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fixed warnings
 *
 * Revision 1.37  1995/08/25  01:08:38  jcraig
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * get rid of pc compiler warnings
 *
 * Revision 1.36  1995/08/17  23:00:23  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Needed to comment out the #include of "shros.pub"
 *
 * Revision 1.35  1995/08/17  21:55:20  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.34  1995/07/27  20:58:14  macleod
 * macleod on Thu Jul 27 13:57:21 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * When halftoneThresh and segmentDocument found no segments, they
 * returned an "error", which caused our purify regression test to think
 * that it didn't work. Changed it so that it doesn't generate an error,
 * since it's not really an erroneous result.
 *
 * Revision 1.33  1995/01/24  21:59:32  lperry
 * lperry on Tue Jan 24 13:59:09 PST 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Incorporate changes to use new strel manager interface.
 * Only allocate actual strels required.
 *
 * Revision 1.32  1995/01/11  22:42:50  dferrell
 * dferrell on Wed Jan 11 14:42:34 PST 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Changed include of shrip.pub to shrcc.pub.
 *
 * Revision 1.31  1994/12/08  04:41:55  ddavies
 * ddavies on Wed Dec 7 20:41:10 PST 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Move the routines segmentGrays and segmentTwice (and it's callback proc)
 * to /product/ipcore/ipwrap/obsolete/halftone.obs.  These are no longer used.
 *
 * Revision 1.30  1994/11/04  03:17:17  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * textMask takes a DPI parameter. It is only used when the image is grayscale.
 *
 * Revision 1.29  1994/10/28  01:59:31  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * There were a couple of places which called strel manager but didn't
 * terminate. There was also a place in segmentDocument where the
 * rectangle list wasn't freed. It no longer leaks.
 *
 * Revision 1.28  1994/10/20  02:07:51  macleod
 * macleod on Wed Oct 19 19:06:57 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added a scheme to find rectangular outlines and keep them in the
 * image mask (in segmentDocument) so that images with outlines get their
 * edges preserved.
 *
 * Revision 1.27  1994/10/19  22:10:54  macleod
 * macleod on Wed Oct 19 15:09:59 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Set strels created in halftoneThresh to NULL (return value of destroyStrel)
 * after destroying them, so that they aren't destroyed twice.
 *
 * Revision 1.26  1994/10/14  19:47:23  macleod
 * macleod on Fri Oct 14 12:41:04 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added yet-another two-threshold segmenter, called segmentDocument.
 * It makes two calls to halftoneThresh, at the two thresholds passed
 * ("text" and "background"). It finds the text by calling textMask, and
 * chops that out of the background mask. It then finds the gutters (long
 * vertical or horizontal white runs), and eliminates all gutters that
 * don't have a perpindicular gutter at their endpoint. The gutters are
 * subtracted from the "background" mask, which has the effect of
 * chopping images away from bleed-through and other failures. The
 * final mask is constructed by seed-filling using the text mask as
 * a seed and the background mask as a clip region. The resulting mask
 * gets its holes plugged, and then regions in the mask which are
 * nearly rectangular are filled in.
 *
 * Revision 1.25  1994/10/07  19:05:42  lperry
 * lperry on Fri Oct 7 12:04:59 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.24  1994/08/29  22:51:21  macleod
 * macleod on Mon Aug 29 15:50:22 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fixed a stupidity in segmentTwice where if the dpi passed was <100,
 * the minimum box size passed to rabbRuns became zero, due to misplaced
 * parenthesis.
 * (er, parentheses)
 *
 * Revision 1.23  1994/08/29  21:56:49  macleod
 * macleod on Mon Aug 29 14:56:24 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * added error checks after PIXR creation in segmentTwice
 *
 * Revision 1.22  1994/08/17  21:45:01  macleod
 * macleod on Wed Aug 17 14:45:00 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added a heuristic ("hack") to the headline remover so that it doesn't think
 * quite so many small pictures are headlines: it compares the number of 1, 2,
 * and 3-pixel runs to the number of runs at the peak of the run-length
 * histogram. In images, there are usually more runs near the bottom, even
 * if the peak happens to be higher, but in headline text, there are usually
 * very few short runs. Dark, solid imaes can sometimes still be called
 * "headline" and removed from the mask..
 *
 * Revision 1.21  1994/08/16  21:51:07  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Silenced a warning.
 *
 * Revision 1.20  1994/08/16  06:39:56  macleod
 * macleod on Mon Aug 15 23:37:49 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added removeHeadlines, which is a fairly simple headline detector, called
 * from the currently 'experimental' segmenter, segmentTwice.
 * It decides what's really a headline (and therefore to be left alone)
 * by first only checking things <= 1.5 inches high, and then by
 * taking a run-length histogram (horizontal) and looking at the peak index.
 * It hasn't been tested very thoroughly (7 documents).
 *
 * Revision 1.19  1994/08/12  00:34:15  macleod
 * macleod on Thu Aug 11 17:30:56 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Changed segmentTwice to use a fairly brain-dead scheme which seems
 * to make people happy--just fill in the bounding rectangle of the
 * best-for-text-thresholded pictures, then subtract off stuff that we
 * find as text in the original best-for-text image - the image mask (not
 * filled). The result is that rectangular pictures almost never turn out
 * grungy, even when there's lots of bleed-through. There are
 * probably better solutions, but finding a "right" threshold below which
 * one magically gets all of the pictures and none of the background or
 * bleed-through or tint isn't one of them.
 *
 * Revision 1.18  1994/08/10  21:59:01  lperry
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Changes made to use new strel manager.
 *
 * Revision 1.17  1994/08/05  01:42:40  macleod
 * macleod on Thu Aug 4 18:38:54 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added segmentTwice, which takes a grayscale image and two thresholds,
 * and does a "Twice Cooked" segmentation, where the "text" thresholded
 * version is segmented, and the result is used as a seed to fill into
 * a mask consisting of the result of the segmentation of the "background"
 * thresholded image. This does an OK job of filling in more image in
 * the resulting mask without filling in all of the bleed-through that
 * might have been thresholded into the background image. Note that
 * there are still cases where bleed-through will come through, such
 * as when the image on the front of the page and the image bleeding
 * through from the back of the page overlap. A better solution is to
 * put black paper behind the scanned object: starting with good data
 * is always better than trying to "fix" bad data, because your "fix"
 * is always wrong some of the time.
 *
 * Revision 1.16  1994/03/02  18:49:38  ddavies
 * ddavies on Wed Mar 2 10:47:29 PST 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fix bug in segmentGrays in which the source mask and source gray images
 * were being freed at 1x resolution and the dest image was a copy of the
 * source mask image.  What I really meant to say was the dest image was
 * just another pointer to the source mask image, so the caller would get
 * confused when it tried to free both of them.
 *
 * Revision 1.15  1994/03/02  03:57:13  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added lots of IP_YIELDs to give up control during halftoneThresh and
 graySegment.
 *
 * Revision 1.14  1994/02/16  03:59:57  ddavies
 * ddavies on Tue Feb 15 19:59:49 PST 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fix up comments in segmentGrays.
 *
 * Revision 1.13  1994/02/16  03:57:29  ddavies
 * ddavies on Tue Feb 15 19:57:02 PST 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added implementation of segmentGrays.  This routine attempts to find "image"
 * portions of an image based on whether there were largish blocks of gray
 * stuff.  halftoneThresh tries to find images based purely on size.
 * segmentGrays is better about not classifying large text as image.
 * It probably needs more tuning.
 *
 * Revision 1.12  1993/12/17  22:49:07  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * halftoneThresh now returns a linked list of RECTANGLs instead of a BOXES
 thing.
 * Most of the changes revolve around this.  We now fill in the holes in the
 * mask before calling rabb since the new rabb deals happily with large black
 areas
 * and somewhat less happily with noisy edges.
 * We now transform the rectlist in place instead of making it into a BOXES
 thing,
 * transforming that into another BOXES thing and throwing the first one away.
 * "Transform" means scale the box coordinates so they're work on the full-size
 * mask.
 *
 * Revision 1.11  1993/12/03  01:16:23  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.10  1993/11/20  03:20:53  ddavies
 * ddavies on Fri Nov 19 19:18:56 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Oh yeah, also change the meaning of ruleWidth.  It used to specify the
 * width of lines to be eliminated from the mask at full scale.  These sizes
 * were adjusted in the reduced mask according to the reduction factor.
 * However, the EZScan folks really used the reduction factor to give all
 * images approximately the same resolution, so rule widths should be
 * roughly constant in the reduced images.  ruleWidth now specifies the
 * width of lines in the reduced image directly.
 *
 * Revision 1.9  1993/11/20  03:17:06  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Modify this routine to take minWidth and minHeight.  Bounding boxes that
 * are thinner than minWidth or shorter than minHeight are not returned in
 * the box list.
 * Also change the algorithm a little.  The strels used to remove horizontal
 * and vertical rules are now (2*ruleWith)+1 x 5 and 5 x (2*ruleWidth)+1 now.
 * Limiting the length of the rule section being matched to 5 was done to
 * avoid very long delays when segmenting large images with a reduction
 * factor of 1.  We now also remove 1x1, 2x1, 1x2 and 2x2 dots in the reduced
 * mask before calling rabb in the hope that this will make life easier.  It
 * didn't have a noticeable affect in my tests, but we'll let the EZScan guys
 * try it.
 *
 * Revision 1.8  1993/11/18  23:42:07  ddavies
 * ddavies on Thu Nov 18 15:41:43 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Also put strel1Inits back in so strel5h and strel5v are defined.  How did
 * it disappear??
 *
 * Revision 1.7  1993/11/18  23:39:39  ddavies
 * ddavies on Thu Nov 18 15:38:06 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * In halftoneThresh, only reduce the seed to 1/4 the size of the reduced mask
 * instead of always reducing the seed by 16x.  With small images that aren't
 * reduced, the seed was getting too small and seed pixels were disappearing.
 * This fixes a bug found by EZScan.
 *
 * Revision 1.6  1993/11/12  01:01:06  ddavies
 * ddavies on Thu Nov 11 17:00:16 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * If the mask is not expanded because it's already at full size, return it
 * directly instead of returning a null image.
 *
 * Revision 1.5  1993/11/11  23:06:28  ddavies
 * ddavies on Thu Nov 11 15:05:51 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Don't expand the mask back to full size if it's already full size.  Trying
 * to do so caused an error in expandPixr.
 *
 * Revision 1.4  1993/11/06  03:13:12  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

