/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  halftone.c--includes
 *
 *              w_getFrameMask()
 *		w_halftoneThresh()
 *		w_segmentBinaryPagePixrs ()
 *               
 */



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "morphg.h"
#include "logicops.h"
#include "defines.pub"
#include "shrscgr.pub"
#include "shrcc.pub"
#include "thrshold.h"
#include "iaerror.pub"
#include "frames.pub"

/* prototypes */
#include "clean.pub"
#include "props.pub"
#include "rast.pub"
#include "curves.pub"
#include "fthresh.pub"
#include "prxform.pub"
#include "fill.pub"
#include "thrshold.pub"
#include "imginfo.pub"
#include "runhist.pub"
#include "halftone.pub"
#include "strel.pub"
#include "morph.pub"
#include "expand.pub"
#include "rectops.pub"
#include "logicops.pub"
#include "reduce.pub"
#include "intops.pub"
#include "ccs.pub"
#include "runmorph.pub"
#include "runfilt.pub"
#include "subsmpl.pub"

/* DEBUG */
#include <stdio.h>
#include "imagerep.h"
#include "io.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: halftone.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");

#define	DEBUG_HT_THRESH	0

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get frame mask.
 *
 * LongDescription:::
 *  This is a private function, which get frame mask. Before doing hole 
 *  filling, it is necessary to check if there is frame. Filling big 
 *  frame hole will cause the big mistake in segmentation. 
 *
 * PrivateDescription:::
 * 
 * Input:::
 *  pIPWState::      State structure
 *  prMask::         Point to mask image
 *  minWidth::       Min. width of component to find
 *  minHeight::      Min. height of component to find
 *
 * Output:::
 *  pPrFrameMask::   Point to point of frame mask image 
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, seed filling, morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getFrameMask (
    IPWState            *pIPWState,
    PIXR                *prMask,
    Int32                minWidth,
    Int32                minHeight,
    PIXR               **pPrFrameMask)
{
    PROLOGUE;
    ConnectedComponent *pCCList = NULL, *p;
    Int32               w, h, i, len, bits;
    Int32               testH, centW, left, right;
    ScaledInt1616       xDPI, yDPI;
    PIXR               *prFill = NULL;

    ERR_GEN (prMask == NULL, cErrNullPtr);
    ERR_GEN (*pPrFrameMask, cErrPtrNotNull);

    ERR_PASS (w_pixrGetWidth (pIPWState, prMask, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prMask, &h));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prMask, &xDPI, &yDPI));

    ERR_PASS (w_findConnectedComponents (pIPWState, prMask,
					 minWidth, minHeight,
					 0, &pCCList));
    if (pCCList)
    {
	for ( p = pCCList; p; p = p->next)
	{
	    if( p->width < (w/2) || p->height < (h/2) ||
		p->runCount < p->height)
		continue;
	    bits = testH = 0;
	    centW = p->width >> 1;
	    left = 0x0fffffff;
	    right = p->runs[0].runRight;
	    for(i=0; i < p->runCount; ++i)
	    {
		if( p->runs[i].runRight >= left )
		{
		    if( left > centW ) ++testH;
		    if( right < centW ) ++testH;
		    right = p->runs[i].runRight;
		}
		left = p->runs[i].runLeft;
		len = p->runs[i].runRight - left;
		bits += len;
	    }
	    len  = bits /p->runCount; 
	    if( (bits << 2) < (p->width * p->height) &&
		len < (w/4) && testH < minHeight)
	    {
		if(prFill == NULL)
		    ERR_PASS (w_createPixr (pIPWState, w, h,
						    1, xDPI, yDPI, cFrameBits,
						    &prFill));

		ERR_PASS_CLEANUP (w_CCSet (pIPWState, p, 0, 0, prFill));
	    }
	}

	ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCList));
    }

    *pPrFrameMask = prFill;

    NORMAL_RETURN;

ERR_CLEANUP:

    if( pCCList ) w_destroyConnectedComponents (pIPWState, &pCCList);
    if( prFill) w_destroyPixr(pIPWState, &prFill);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  A segmentation scheme based on a seed/mask method. 
 *
 * LongDescription:::
 *  Creates an image separation mask and a list of rectangles, one for each
 *  connected component in the mask.
 *<BR>
 *  If the separation mask is empty, returns NULL.
 *  Otherwise, the pRects ptr is written to the given
 *  address, and the separation mask pixr is
 *  returned.
 *<BR>
 *  This method takes the reduction factor as a
 *  parameter.  It may be one of 1x, 2x or 4x.
 *<BR>
 *  This method also takes the width of rules 
 *  to be eliminated from the mask.  The widths
 *  are expressed in pixels at full size.
 *  The rule width is specified by the caller for the
 *  reduced image.  This is because the EZScan folks
 *  use the reduction step as a way of converting
 *  all images to a common resolution.
 *<BR>
 *  When reducing the mask by 2x, we use thresholds of
 *  1, 2, 3, or 4.  A threshold of n means that if
 *  at least n of the bits in the source 2x2 square
 *  are 1's, the output is a 1.  Lower numbers tend
 *  to connects things, which closes places where the
 *  mask is spotty.  Higher numbers tend to avoid
 *  connecting things, which keeps image and text
 *  separate.  Since we don't know apriori which
 *  threshold to use, we'll let the user decide!
 *  redThrshold can be 1, 2, 3 or 4.
 *
 * PrivateDescription:::
 *
 * method::
 *    (1)  Make the seed at 1 / (4*redFactor) then expand to 1/redFactor.
 *    (2)  Make the mask with given reduction factor
 *    (3)  Remove the rules from the mask
 *    (4)  Fill from the seed into the mask
 *    (5)  Edge fill to get holes in the mask
 *    (6)  Plug the holes
 *    (7)  In the places that were plugged, see if there is text.
 *	   If so, take the text mask out of the filled mask.
 *    (8)  Remove components that are too small from the mask.
 *    (9)  Make a list of boxes, one for each connected component in the mask
 *   (10)  Expand the separation mask to f.s.
 *
 * Input:::
 *  pIPWState::      IP State Structure
 *  prSrc::          Image to segment
 *  redFactor::      Reduction factor, power of 2 <16
 *  ruleWidth::      Width of rules to be removed from mask (full size pixels)
 *  redThreshold::   Reduction threshold, 1, 2, 3 or 4
 *  minWidth::       Min. width of features to find
 *  minHeight::      Min. height of features to find
 *
 * Output:::
 *  ppRects::        List of bounding rectangles for images found
 *  pPrResultMask::  Bit map containing 1 where image pixels were foudn

 * Returns:::        Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, seed filling, morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::	ancient version
 *  davies::	current version
 *  macleod::	a few hacks on current version
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_halftoneThresh (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    UInt32              redFactor,
    UInt32              ruleWidth,
    UInt32              redThreshold,
    UInt32              minWidth,
    UInt32              minHeight,
    RECTANGL          **ppRects,
    PIXR              **pPrResultMask)
{
    PIXR            *prTemp, *prSeed, *prMask;
    RECTANGL        *rectList;
    STREL           *horizRule, *vertRule;
    STREL           *pStrel5h = NULL, *pStrel5v = NULL, *pStrel2 = NULL,
	            *pS_pepper_1 = NULL, *pS_pepper_2 = NULL;
    Bool             isZero;
    PROLOGUE;

    prTemp = prSeed = prMask = NULL;	/* initialize pointers so we
					 * can call destroyPixr on them
					 * without fear. */
    rectList = NULL;			/* ditto */
    horizRule = vertRule = NULL;	/* ditto */


    /* Validate the PIXR passed */
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel5h));
    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel5h, &pStrel5h));
    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel5v));
    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel5v, &pStrel5v));
    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel2));
    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel2, &pStrel2));
    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, s_pepper_1));
    ERR_PASS_CLEANUP (w_getStrel(pIPWState, s_pepper_1, &pS_pepper_1));
    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, s_pepper_2));
    ERR_PASS_CLEANUP (w_getStrel(pIPWState, s_pepper_2, &pS_pepper_2));

	/* test the value of the redution factor */
    ERR_GEN_CLEANUP ((redFactor != 1) && (redFactor != 2) &&
	(redFactor != 4), cErrBadReduction);

    IP_YIELD;

	/* We generally want to make the seed at about 1/4 the scale
	 * of the mask.  So, if redFactor is 1, we reduce the seed
	 * by 4 and so on.
	 */
    prTemp = NULL;
    switch (redFactor)
    {
    case 1:	/* only reduce the seed by 4x */
	ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, 1, 4, 0, 0, &prTemp));
	break;

    case 2:	/* reduce the seed by 8x */
	ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, 1, 1, 4, 0, &prTemp));
	break;

    case 4:	/* reduce the seed by 16x */
	ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, 1, 1, 4, 4, &prTemp));
	break;
    }

    ERR_PASS_CLEANUP (w_open (pIPWState, prTemp, pStrel5h, &prTemp));
    ERR_PASS_CLEANUP (w_open (pIPWState, prTemp, pStrel5v, &prTemp));
    prSeed = NULL;
    ERR_PASS_CLEANUP (w_expand(pIPWState, prTemp, 4, &prSeed));

    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prTemp));

    IP_YIELD;

	/* test for NULL result */
    ERR_PASS_CLEANUP (w_isZero(pIPWState, prSeed, &isZero));
    if (isZero)
    {
	goto nullFinish;
    }

	/* Here's where things get a little tricky.  We want to either
	 * not reduce at all, reduce by 2x, 4x or 8x.  Since the reduction
	 * routines take a variable number of args depending on the
	 * reduction factor, we just have three cases.  We also use
	 * a threshold of 3, instead of 1.  With EZScan's fancy background
	 * correction code, they seem to find pretty solid blocks for
	 * halftones.  Using a threshold of 1 means we tend to connect
	 * text to the images, which is no good.
	 */
    prMask = NULL;
    switch (redFactor)
    {
    case 1:
	    /* just make a copy of the source image */
	ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prSrc, &prMask));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prMask));
	break;

    case 2: /* use 2x reduction */
	ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, redThreshold, 
				   0, 0, 0, &prMask));
	break;

    case 4:
	ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, redThreshold, 
				   redThreshold, 0, 0, &prMask));
	break;

    }

    IP_YIELD;

	/* Prepare to remove horizontal and vertical rules */
    prTemp = NULL;
    ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prMask, &prTemp));

	/* To remove the horizontal and vertical rules from the mask,
	 * we have to create a couple of strels, one for horizontal
	 * stuff and one for vertical stuff.  The strels have hits down
	 * the centers and misses along the sides (or along the top
	 * and bottom for horizontal strels).  These select things that
	 * are 1/2 the distance between the rows or columns of misses.
	 * Therefore, we should make the strels twice the width of
	 * the rules at the current resolution.
	 */
    if (ruleWidth > 0)
    {
	horizRule = NULL;
	ERR_PASS_CLEANUP (w_createHorizRuleStrel(pIPWState, ruleWidth, 
						 &horizRule));

	vertRule = NULL;
	ERR_PASS_CLEANUP (w_createVertRuleStrel(pIPWState, ruleWidth,
						&vertRule));
    
	    /* get rid of horizontal and vertical lines */
	ERR_PASS_CLEANUP (w_open (pIPWState, prMask, horizRule, &prTemp));
	IP_YIELD;

	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTemp, SUBTRACT, prMask));
	IP_YIELD;

	ERR_PASS_CLEANUP (w_open (pIPWState, prMask, vertRule, &prTemp));
	IP_YIELD;

	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTemp, SUBTRACT, prMask));
	IP_YIELD;

	    /* we're done with the strels, so delete them */
	ERR_PASS_CLEANUP (w_destroyStrel(pIPWState, &horizRule));
	horizRule = NULL;
	ERR_PASS_CLEANUP (w_destroyStrel(pIPWState, &vertRule));
	vertRule = NULL;
    }

	/* expand seed into mask, max of 10 filling iterations */
    ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, prSeed, prMask, 10, 
					&prTemp));
    IP_YIELD;

	/* test for NULL result */
    ERR_PASS_CLEANUP (w_isZero(pIPWState, prTemp, &isZero));
    if (isZero)      
    {
	/* This is not an "error", but a common result */
	goto nullFinish;
    }

	/* Get rid of the isolated 1x1 and 2x2 dots.  This greatly reduces the
	 * amount of work rabb will do on most images.
	 */

        /* find the single dots */
    ERR_PASS_CLEANUP (w_erode (pIPWState, prTemp, pS_pepper_1, &prMask)); 

	/* get rid of them */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prMask, SUBTRACT, prTemp));
    IP_YIELD;

        /* get rid of 2x1, 1x2, 2x2 */
    ERR_PASS_CLEANUP (w_erode (pIPWState, prTemp, pS_pepper_2, &prMask));

	/* prSeed was free */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prMask, pStrel2, &prSeed));

	/* get rid of those dots too */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSeed, SUBTRACT, prTemp));
    IP_YIELD;


    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prSeed));
    prSeed = NULL;
    ERR_PASS_CLEANUP (w_getFrameMask( pIPWState, prTemp,
				       minWidth, minHeight*2,
				       &prSeed));

    /* Take out frame before filling hole */
    if( prSeed != NULL )
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSeed, SUBTRACT, prTemp));

	/* make hole overlay image and plug the "holes" */
        /* fill from outside */    
    ERR_PASS_CLEANUP (w_edgeFillSweep4 (pIPWState, prTemp, 1, &prMask));
    IP_YIELD;

        /* get interior pixels (holes) */
    ERR_PASS_CLEANUP (w_invert(pIPWState, prMask, &prMask));

          /* plug the holes */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTemp, OR, prMask));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prTemp));
    IP_YIELD;

    /* Put frame back to mask */
    if( prSeed )
    {
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSeed, OR, prMask));
	ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prSeed));
    }

    /* make and save list of rectangles for the connected components.*/
    rectList = NULL;
    ERR_PASS_CLEANUP (w_rabb(pIPWState, prMask, minWidth, minHeight, 
			     &rectList));
    if (rectList == NULL)
    {
	/* This isn't an error, so don't complain. */
	goto nullFinish;
    }
    IP_YIELD;

    ERR_PASS_CLEANUP (w_xformRectList (pIPWState, 0, 0,
				       (Float32) redFactor, 
				       (Float32) redFactor, 
				       rectList));

	/* return the list of rectangles */
    *ppRects = rectList;

	/* expand mask to full size */
    if (redFactor > 1)
    {
	prTemp = NULL;
	ERR_PASS_CLEANUP (w_expand(pIPWState, prMask, redFactor, &prTemp));
	ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prMask));
	prMask = NULL;
	*pPrResultMask = prTemp;
    }
    else	/* redFactor == 1 */
    {
	*pPrResultMask = prMask;
    }
    NORMAL_RETURN;

#if  DEBUG_HT_THRESH
    {
    PIXR     *prt11;
    if ((prt11 = copyPixr(prTemp)) == NULL)
	return abortP("prt11 not made", procName, NULL);
    opPrPr(prt11, prSrc, AND);  /* halftone image part */
    writePr("c:\\temp\\junkhalftone", prt11, 1);
    opPrPr(prt11, prSrc, XOR);  /* non-image part */
    writePr("c:\\temp\\junkother", prt11, 1);
    DESTROY_PIXR(prt11);
    }
#endif  /* DEBUG_HT_THRESH */

        /* 
	 * In this case, we didn't really failed, we just didn't get
	 * a result. 
	 */
nullFinish:
	/* clean up everything */
    if (horizRule != NULL)
	ERR_PASS_CLEANUP (w_destroyStrel(pIPWState, &horizRule));
    if (vertRule != NULL)
	ERR_PASS_CLEANUP (w_destroyStrel(pIPWState, &vertRule));

    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prSeed));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prMask));
    ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prTemp));
    if (rectList != NULL)
	ERR_PASS_CLEANUP (w_destroyRectList(pIPWState, &rectList));
    *ppRects = NULL;
    *pPrResultMask = NULL;
    NORMAL_RETURN;

ERR_CLEANUP:
	/* clean up everything */
    if (horizRule != NULL)
	w_destroyStrel (pIPWState, &horizRule);
    if (vertRule != NULL)
	w_destroyStrel (pIPWState, &vertRule);

    w_destroyPixr (pIPWState, &prSeed);
    w_destroyPixr (pIPWState, &prMask);
    w_destroyPixr (pIPWState, &prTemp);
    if (rectList != NULL)
	w_destroyRectList(pIPWState, &rectList);
    *ppRects = NULL;

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  A text/image segmenter which takes a single binary image
 *
 * LongDescription:::
 *  This routine is similar to w_segmentPagePixrs except that it takes
 *  a single binary input image instead of a gray image and two thresholds.
 *  This routine is also different from w_segmentPagePixrs in that it returns
 *  cleaned up images themselves instead of a mask for image regions on the
 *  page.  The line art and text in each image have been removed (we hope).
 *  There is no real use for the mask itself in that combining two binary
 *  images is trivial (OR) while combining a gray or color image with a binary
 *  image may require a mask to properly define the edges.<BR>
 *<BRL>
 *  The output of this routine is an array of pointer to PIXRs, each of
 *  which contains one picture from the original page.  Reverse video
 *  text is normally mistaken for an image.  In these images, the
 *  reverse video text blocks have been turned completely black.  We
 *  recommend that the caller remove images from input pages by XORing
 *  rather than anding with the inverse of the image.  In this way, the
 *  reverse video text will become normal, which is better for OCR and
 *  compression.
 *
 * PrivateDescription:::
 *  The heuristic goes basically like this:<BR>
 *  1. Use w_textMask to find all the text on the image.  Save this mask.<BR>
 *  2. Use w_halftoneThresh to identify areas of potential image and stipple
 *     regions.  Toss the rectlist returned by w_halftoneThresh.
 *  3. Use the mask returned by w_halftoneThresh to isolate only these regions.
 *  4. Use w_cleanStipple to extract the non-stipple portions of these
 *     regions.
 *  5. Use the text mask found in step 1 to identify text from these regions.
 *     This also identifies any headlines that may have been mistaken
 *     for images.
 *  6. Subtract this text stuff from the images found in step 2.  This
 *     leaves a PIXR that contains only image pixels and a mask.
 *  7. Follow the method of w_segmentPagePixrs to create an array of
 *     PIXR pointers and create the associated PIXRs.  If a PIXR has very
 *     few black pixels, toss it.
 * 
 * Input:::
 *  pIPWState::      State structure
 *  prs::            The source binary image
 *  redFactor::      Reduction factor (see w_halftoneThresh()). This should
 *                   probably be inferred from the resolution.
 *  ruleWidth::      Min. Rule width (see w_halftoneThresh())
 *  redThreshold::   Reduction threshold, 1, 2, 3 or 4 (see w_halftoneThresh())
 *  minWidth::       Discard objects smaller than this width
 *  minHeight::      ...or this height
 *
 * Output:::
 *  ppImages::	     Array of PixrRef structures
 *  pImageCount::    Number of images created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, seed filling, morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_segmentBinaryPagePixrs (
    IPWState            *pIPWState,
    PIXR                *prs,
    UInt32               redFactor,
    UInt32               ruleWidth,
    UInt32               redThreshold,
    UInt32               minWidth,
    UInt32               minHeight,
    PixrRef            **ppImages,
    Int32               *pImageCount)
{
    PROLOGUE;
    Bool                isZero;
    STREL	       *pStrel2 = NULL, *pStrel3 = NULL, *pStrel5 = NULL;
    STREL	       *pStrel11h = NULL, *pStrel3v = NULL;
    PIXR               *prHalftoneMask = NULL, *prTmp = NULL;
    PIXR	       *prTextMask = NULL, *prImages = NULL;
    PIXR	       *prRevText = NULL;
    PIXR	       *prTmp1 = NULL, *prTmp2 = NULL, *prTmpRed = NULL;
    PIXR	       *prTmp3 = NULL;
    PixrRef            *pImageArray = NULL;
    ConnectedComponent *pCCList = NULL, *p;
    RECTANGL	       *rectList = NULL;
    Int32               count, i, dpi, width, height;
    Int32		compW, compH;
    ScaledInt1616       xDPI, yDPI;

    ERR_PASS (w_initStrelByName (pIPWState, strel2));
    ERR_PASS (w_initStrelByName (pIPWState, strel3));
    ERR_PASS (w_initStrelByName (pIPWState, strel3v));
    ERR_PASS (w_initStrelByName (pIPWState, strel5));
    ERR_PASS (w_initStrelByName (pIPWState, strel11h));

    ERR_PASS (w_getStrel (pIPWState, strel2, &pStrel2));
    ERR_PASS (w_getStrel (pIPWState, strel3, &pStrel3));
    ERR_PASS (w_getStrel (pIPWState, strel3, &pStrel3v));
    ERR_PASS (w_getStrel (pIPWState, strel5, &pStrel5));
    ERR_PASS (w_getStrel (pIPWState, strel5h, &pStrel11h));

	/* halftoneThresh likes to know lots of stuff based on
	 * the incoming picture's resolution.
	 */
    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, prs, &xDPI, &yDPI));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prs, &width));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prs, &height));
    dpi = MAX (ScaledInt1616_TO_Int32 (xDPI),
	       ScaledInt1616_TO_Int32 (yDPI));

	/* Segment the page */

	/* Use halftone thresh to find the regions of the page where
	 * there may dithered images.  This also finds text on stipple
	 * and reverse video text.
	 * It may also confuse some big headline text with images.  However,
	 * that stuff should also be in the text mask (we hope).
	 */
    ERR_PASS_CLEANUP (w_halftoneThresh (pIPWState, prs,
					redFactor, ruleWidth,
					redThreshold, minWidth, minHeight,
					&rectList, &prHalftoneMask));

    ERR_PASS_CLEANUP (w_destroyRectList(pIPWState, &rectList));

	/* If there were no images, return happily now. */
    if (prHalftoneMask == NULL)
	goto noImages;

	/* Create an image that contains only these regions.  That's so we
	 * don't mistake small text for stipple.  We'll only look for stipple
	 * in dithered regions.
	 */
    ERR_PASS_CLEANUP (w_copy (pIPWState, prs, &prImages));
    ERR_PASS_CLEANUP (w_logicOp(pIPWState, prHalftoneMask,
				INTERSECTION, prImages));

	/* Captions are fairly often attached to the image masks.
	 * However, not all letters in the captions may be attached.
	 * Do two reductions, one with threshold 4 and one with threshold
	 * 1.  Open the thresh4 reduced image to clip off all the stuff
	 * around the outside.  Dilate the thresh1 reduced image to 
	 * grow to include all the stuff around the outside, even
	 * parts of captions that weren't connected to the image mask.
	 * We want all the captions because w_testMask isn't happy
	 * about isolated letters.  Subtract the thresh4 image from
	 * the thresh1 image so only the stuff around the outside is
	 * left.  Mask the original image with that and look for text.
	 * If we find any, remove it from the image.
	 * And one more thing - w_textMask gets fussy about thin strips
	 * around images.  Fill in the middle of the image with and
	 * expanded version of the interior mask.  This convinces
	 * w_testMask that it should ignore the images and concentrate
	 * on the text.
	 */
    ERR_PASS_CLEANUP (w_reduce (pIPWState, prHalftoneMask, 4, 4, 4, 4,
				&prTmpRed));
    ERR_PASS_CLEANUP (w_erode (pIPWState, prTmpRed, pStrel3, &prTmpRed));
    ERR_PASS_CLEANUP (w_reduce (pIPWState, prHalftoneMask, 1, 1, 1, 1,
				&prTmp1));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmp1, pStrel3, &prTmp2));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
	/* get the parts that aren't image now */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmpRed, SUBTRACT, prTmp2));
    ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, prHalftoneMask, prTmp2,
				       &prTmp1));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp2));
	/* prTmp1 has a mask of just the little stuff, like text */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prImages, INTERSECTION, prTmp1));
	/* Fill in the middle of the images to keep w_textMask on the
	 * straight and narrow.
	 */
    ERR_PASS_CLEANUP (w_expand (pIPWState, prTmpRed, 16, &prTmp2));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmpRed));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp2, UNION, prTmp1));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp2));

	/* Find the captions, if any, around the outside of images */
    ERR_PASS_CLEANUP (w_textMask (pIPWState, prTmp1, &prTextMask, 0));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

	/* Now, did we find any text?  If not skip messing around
	 * with the text mask and removing the text.
	 */
    ERR_PASS_CLEANUP (w_isZero (pIPWState, prTextMask, &isZero));
    if (!isZero)
    {
	    /* Because w_textMask sometimes chops off punctuation at the
	     * ends of lines, make the masks bigger.
	     */
	ERR_PASS_CLEANUP (w_reduce (pIPWState, prTextMask, 1, 1, 1, 1,
			  &prTmpRed));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));
	ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmpRed, pStrel3, &prTmp1));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmpRed));
	ERR_PASS_CLEANUP (w_invert (pIPWState, prTmp1, &prTmp1));
	ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, prImages, prTmp1,
					   &prImages));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
    }
    else
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));


	/* Now look at the remaining part of the image to see if there's
	 * text on tint.
	 * Use plain old open to remove the stipple from these regions.
	 * That's so we can isolate the text cleanly.  The stipple-removing
	 * routines remove too much text and you have to fill it back up,
	 * all of which takes time.
	 */
    if (dpi >= 400)
    {
	ERR_PASS_CLEANUP (w_open(pIPWState, prImages, pStrel3, &prTmp));
    }
    else
    {
	ERR_PASS_CLEANUP (w_open(pIPWState, prImages, pStrel2, &prTmp));
    }

	/* In some images with really dense stipple/halftone under
	 * text, we still have a bunch of pepper inside the text
	 * which causes w_textMask to ignore the text.  Remove it here.
	 */
    ERR_PASS_CLEANUP (w_removePepper (pIPWState, prTmp, 2, &prTmp));
	/* All this processing ate parts of the characters.  Grow
	 * them a touch.
	 */
    ERR_PASS_CLEANUP (w_close (pIPWState, prTmp, pStrel2, &prTmp));


	/* Find the positive and reverse text in the de-stippled images.  We
	 * use these images because the text-finding stuff is based
	 * on connected components, so we want as few of them as
	 * possible.
	 *
	 * This is harder than you might think.  w_textMask does a good
	 * job of finding positive (black on white) text, but it also
	 * finds some image pieces.  w_reversetextMask does a great
	 * job of not finding any image pieces, but in doing so,
	 * it misses some inverted text.  For example, it works fine
	 * on nice clean horizontal bars that contain some reversed text.
	 * However, it doesn't like the bars to be attached to images.
	 * When that happens, it decides that the union of the black
	 * block that surrounds the inverted text and some big chunk
	 * of image is very big and strangely shaped, hence is not an
	 * inverted text block.  It also doesn't like the bars to
	 * share a bounding box with something else that might look
	 * like reverse text.  It also doesn't like small inverted
	 * text regions (looks like noise).
	 * I spent a few days figuring this out empirically.
	 *
	 * Here's the solution.  The basic idea to find stuff that both
	 * looks like regular text when its inverted and looks like
	 * inverse video text when it's left alone.
	 *
	 * You'd think we could just invert the entire image and look
	 * for text in that, but Noooo!  An image that's mostly a
	 * giant black blob really upsets w_textMask.  So you have to
	 * be careful about what you invert.  Also, there are some fools
	 * who make their reverse video text extend all they way through
	 * the edge of their surrounding blocks.
	 *
	 * We solve this problem by reducing the image horizontally,
	 * dilating it horizontally to close holes and try to attach
	 * random dilated glop to big blobs (so they don't look like
	 * text).  We then erode this slightly vertically to chop off
	 * noise that would otherwise grow on the top and bottom of
	 * the blocks.  We expand this thing back up vertically and
	 * SUBTRACT the original image from it.  This leaves very nice
	 * positive video text, and other glop in images.  w_textMask
	 * operates on this.
	 *
	 * Once you've found positive text (and some pieces of inverted
	 * image) in the carefully inverted image, You might also
	 * think, as I did, that you could just narrow it down to 
	 * just inverted text by inverting the sections where the
	 * positive text was found and using w_reversetextMask.
	 * But Nooooooooooooo!  Those regions are too small for
	 * w_reversetextMask.  Ok, you might just try making the
	 * masks bigger.  Disaster.  Not only do you grab
	 * surrounding pieces of image (which w_reversetextMask doesn't
	 * like) but you start merging some of the masks, which causes
	 * the bounding boxes for some of the text blobs to fall inside
	 * other text blob bb's, which also upsets w_reversetextMask.
	 * Well, the reason we like w_reversetextMask is because it's
	 * so picky - it does a really good job of ruling out image.
	 *
	 * The real problem is that each blob of text is surrounded by
	 * some unknown trash, like image or other text blobs.  This
	 * surrounding trash influences w_reversetextMask.  So, we
	 * we extract each text blob, put it into it's own image and
	 * ask w_reversetextMask about it individually.  This sounds
	 * nasty, but I think it's actually faster most of the time
	 * (as compared to calling w_reversetextMask on the entire
	 * image) since the size of inverted text regions inside image
	 * regions is usually pretty tiny.
	 *
	 * The plan is:
	 * 1. Find positive text in the destippled image.
	 * 2. If there is any, see if it's also reverse video text.
	 *    The positive text finder sometimes gets confused about that.
	 * 3. Remove the positive text from the image, and set any reverse
	 *    video sections black.  The client will XOR the image with the
	 *    original page to remove the image and convert reverse video
	 *    to positive video.
	 * 4. Find the big black blocks in the de-stippled image.  These
	 *    will include the reverse text blocks as well as big image
	 *    blocks.  These blocks are found by reducing horizontally,
	 *    dilating horizontally, slightly eroding vertically, and
	 *    re-expanding.
	 * 5. SUBTRACT the original image from these blocks.  This inverts the
	 *    image and reverse text blocks.  It also puts a big fat
	 *    border on the left and right of these regions because of
	 *    the dilation done in step 4.  Without the dilation, we get
	 *    a thin, unpredicable border, which w_textMask sometimes
	 *    thinks is text.  That confuses it a lot.  It never confuses
	 *    the big border with text, so is happy.
	 * 6. Do a tiny close to consolidate halftoned regions without
	 *    doing anything too bad to text.  w_textMask has a tendency
	 *    to think dithered stuff is text, this helps with that.
	 * 7. Use w_textMask to find all the text regions in this partially
	 *    inverted image.  It will also find some inverted dithered
	 *    regions, which we don't want.
	 * 8. Find all the connected components in the text mask.
	 * 9. For each connected component:
	 *    a. Create a bitmap the size of the CC.  If the CC is too
	 *	 small, make a bigger bitmap.
	 *    b. Paint the bitmap black.
	 *    c. Subtract the region of step 6's bitmap specified by
	 *	 the CC (which is always a rectangle) from the new
	 *	 bitmap.  This creates a region of inverted text.
	 *    d. Ask w_reversetextMask if this is inverted text.  If
	 *	 so, it will create a black bitmap.  Otherwise, it
	 *	 will create a white bitmap.
	 *    e. If there are any black pixels in the returned bitmap,
	 *	 OR this CC with the original image in prImages.  This has
	 *	 the effect of filing in all the inverted text.  The
	 *	 client should later remove the dithered images from
	 *	 the page using XOR.  This will convert the inverted
	 *	 text to positive text, which is better for symbol-based
	 *	 compression (and OCR, I hope).
	 */
	 
	/* First, find the positive text on the page.  It might be slightly
	 * more reliable to find the text in the original page,
	 * but that takes a long time.  We'll find the text in the
	 * de-stippled image.
	 */
    ERR_PASS_CLEANUP (w_textMask (pIPWState, prTmp, &prTextMask, 0));

	/* See if we found any text in the image regions.
	 */
    ERR_PASS_CLEANUP (w_isZero (pIPWState, prTextMask, &isZero));
    if (!isZero)
    {
	    /* Because w_textMask sometimes chops off punctuation at the
	     * ends of lines, make the masks bigger.
	     */
	ERR_PASS_CLEANUP (w_reduce (pIPWState, prTextMask, 1, 1, 1, 1,
			  &prTmpRed));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));
	ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmpRed, pStrel3, &prTmp1));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmpRed));
	ERR_PASS_CLEANUP (w_expand (pIPWState, prTmp1, 16, &prTextMask));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

	    /* This finds the positive text.  However, if reverse
	     * video text is embedded in long, thin horizontal bars,
	     * the entire bar is often returned.
	     * We can sort of live with this, but it would be pleasant
	     * to find the reverse text.  We'll try just a little
	     * harder by letting w_reversetextMask have a crack
	     * at the text we found.  Anything it finds will be
	     * blacked out in prImages below.
	     */
	ERR_PASS_CLEANUP (w_copy (pIPWState, prTextMask, &prTmp1));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, INTERSECTION,
				     prTmp1));
	ERR_PASS_CLEANUP (w_reversetextMask (pIPWState, prTmp1, &prTmp2));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

	    /* Subtract the text from the original image and/or stipple
	     * sections.
	     * This removes the text from the de-stippled image.
	     * This is a good thing since there are cases where text on
	     * stipple is found as both positive and reverse text, which
	     * confuses everyone.
	     */
	ERR_PASS_CLEANUP (w_copy (pIPWState, prTmp, &prTmp1));

	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTextMask, INTERSECTION,
				     prTmp1));
	    /* The text was a little beat up by the de-stippler, try
	     * to help a little.
	     */
	ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmp1, pStrel3, &prTmp3));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp3, SUBTRACT, prImages));

	    /* After having subtracted out the text, and possibly
	     * black bars of reverse text, OR the preliminary
	     * reverse text mask back in.  This sets regions containing
	     * reverse text to all black.  The client should XOR the
	     * image with the original to produce the text plane.
	     * The XOR eliminates the image parts and inverts the
	     * inverted text parts.
	     */
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp2, UNION, prImages));
	   /* Also subtract the positive text and reverse text blocks
	    * from the de-stippled image.  We don't want to find either
	    * of them when looking for reverse text below.  We obviously
	    * don't want to find positive text and don't need to
	    * find the reverse text again.
	    */
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp3, SUBTRACT, prTmp));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp3));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp2));
    }
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));

	/* Now we start looking for reverse video text in the remaining
	 * destippled regions.
	 * This is where we generate the black blobs.  We do this by
	 * reducing the image by 16 horizontally, ORing the rows,
	 * closing horizontally, eroding vertically to get rid of
	 * any lines on the top and bottom that appeared because of
	 * noise, then expanding back up.
	 */
    ERR_PASS_CLEANUP (w_reduceCols (pIPWState, prTmp, 16, OR, &prTmpRed));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmpRed, pStrel11h, &prTmpRed));
    ERR_PASS_CLEANUP (w_erode (pIPWState, prTmpRed, pStrel3v, &prTmpRed));
    ERR_PASS_CLEANUP (w_expandCols (pIPWState, prTmpRed, 16, &prTmp1));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmpRed));

	/* Subtract the initial image from this mask.  This should leave
	 * just the reverse video characters, even if they extend to
	 * the top and bottom edge of the reverse text block.
	 * Ooops, minor bug fix.  One test image had very fine reverse
	 * video text.  All the processing above closed it up in
	 * chunks.  We simply AND the destippled image with the
	 * original image before looking for reverse video.  This
	 * will clear out the original reverse video text while not
	 * adding any black stuff to the de-stippled image.
	 */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prImages, INTERSECTION, prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, prTmp1));

	/* Try to solify halftoned stuff without doing too much damage
	 * to thin text.
	 */
    ERR_PASS_CLEANUP (w_close (pIPWState, prTmp1, pStrel2, &prTmp1));

	/* We now make our first attempt to find inverted text */
    ERR_PASS_CLEANUP (w_textMask(pIPWState, prTmp1, &prRevText, 0));

	/* Now we go through all the CCs in this text mask, seeing
	 * if each one covers real text.
	 */
	/* Find a list of connected components */
    ERR_PASS_CLEANUP (w_findConnectedComponents (pIPWState,
						 prRevText,
						 8, 8,
						 1, &pCCList));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRevText));

	/* Go through the CC's */
    if (pCCList)
    {
	    /* Create and paint the PIXRs */
	for (p = pCCList; p != NULL; p = p->next)
	{
	    compW = 3 * p->width + 8;
	    compH = 2 * p->height;
	        /* Create a PIXR to hold the image bits */
	    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, 
	    					  compW, compH,
						  1, xDPI, yDPI, cFrameBits,
						  &prTmp2));
		/* I just want to paint it black */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState,
					  0, 0, NULL,
					  PIX_SET,
					  0, 0, compW, compH,
					  prTmp2));
		/* Subtract the portion of the inverted image that
		 * was recognized as text.
		 */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState,
					  p->x - cFrameBits,
					  p->y - cFrameBits, prTmp1,
					  PIX_NOT(PIX_SRC),
					  p->width / 2, p->height / 2,
					  p->width, p->height,
					  prTmp2));

		/* See if it's reverse text */
	    ERR_PASS_CLEANUP (w_reversetextMask (pIPWState, prTmp2,
						 &prTmp3));
	    ERR_PASS_CLEANUP (w_isZero (pIPWState, prTmp3, &isZero));
	    if (!isZero)	/* there's a mask here, so that was
				   reversed text. */
	    {
		    /* fill in the reverse text chars in the dithered
		     * image.  Note that we leave a little space around
		     * the image so that the characters inside will be
		     * separated from any remaining mask around the outside.
		     */
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL,
					      PIX_SET,
					      p->x - cFrameBits - 1,
					      p->y - cFrameBits - 1,
					      p->width + 2, p->height + 2,
					      prImages));
	    }
	    
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp2));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp3));
	}
    }

	/* We're done with prTmp (for awhile). */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));
    ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCList));

	/* We now have a PIXR that contains only the dithered images.
	 * We've eliminated the text.
	 * Create a bunch of sub-PIXRs, one for each individual
	 * mask in prHalftoneMask.  But first, seed fill the dithered
	 * stuff back into prHalftoneMask so we eliminate the
	 * masks that didn't cover images (covered headline text
	 * or some othere dreck).
	 */
    ERR_PASS_CLEANUP( w_fillClipSweep8(pIPWState, prImages, prHalftoneMask, 3,
				       &prTmp));
    ERR_PASS_CLEANUP( w_destroyPixr (pIPWState, &prHalftoneMask));
    prHalftoneMask = prTmp;
    prTmp = NULL;

    pCCList = NULL;
    count   = 0;
    pImageArray = NULL;

	/* Find a list of connected components */
    ERR_PASS_CLEANUP (w_findConnectedComponents (pIPWState,
						 prHalftoneMask,
						 minWidth, minHeight,
						 1, &pCCList));

	/* We're done with prHalftoneMask temporarily.  All of the
	 * masks are in the list of connected components.
	 */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prHalftoneMask));

    if (pCCList)
    {
	    /* See how many masks there are. */
	for (count = 0, p = pCCList; p != NULL; p = p->next)
	{
	    count++;
	}
    
	    /* Create the pixr array */
	pImageArray = (PixrRef *) 
	    IPW_MALLOC (count * sizeof (PixrRef), cAllocatorData);
	ERR_GEN_CLEANUP (pImageArray == NULL, ia_nomem);

	    /* Create and paint the PIXRs */
	for (i = 0, p = pCCList; i < count; i++, p = p->next)
	{
	    pImageArray[i].xOffset = p->x - cFrameBits;
	    pImageArray[i].yOffset = p->y - cFrameBits;
	    prTmp = NULL;
	        /* Create a PIXR to hold the image bits */
	    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
						  p->width, p->height, 1,
						  xDPI, yDPI, cFrameBits,
						  &prTmp));
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState,
					  p->x - cFrameBits,
					  p->y - cFrameBits, prImages,
					  PIX_SRC, 0, 0, p->width, p->height,
					  prTmp));
		/* Create a PIXR to hold the halftone mask */
	    ERR_PASS_CLEANUP (w_createPixr (pIPWState, p->width, p->height,
					    1, xDPI, yDPI, cFrameBits,
					    &prHalftoneMask));
	    ERR_PASS_CLEANUP (w_CCSet (pIPWState, p, -(p->x) + cFrameBits,
				   -(p->y) + cFrameBits, prHalftoneMask));

		/* In w_segmentPagePixrs, we spent time here seeing if
		 * we could make the mask rectangular.  This takes an
		 * AMAZIMG amount of time, almost half of the total time
		 * in the routine.  It seems like a waste, so we don't do
		 * it here.
		 */

		/* Select just the bits from this image.  This step really
		 * only does something useful if the rectangle containing this
		 * halftone mask overlaps another halftone mask so the
		 * w_rasterop above moved too many bits.
		 */
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prHalftoneMask,
					 INTERSECTION, prTmp));
	    ERR_PASS_CLEANUP (w_destroyPixr( pIPWState, &prHalftoneMask));
	    pImageArray[i].pPixr = prTmp;
	}
    }
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prImages));
    ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCList));

    *ppImages = pImageArray;
    *pImageCount = count;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (rectList != NULL) w_destroyRectList (pIPWState, &rectList);
    if (prTmp != NULL) w_destroyPixr (pIPWState, &prTmp);
    if (prTextMask != NULL) w_destroyPixr (pIPWState, &prTextMask);
    if (prImages != NULL) w_destroyPixr (pIPWState, &prImages);
    if (prHalftoneMask != NULL) w_destroyPixr (pIPWState, &prHalftoneMask);
    if (prRevText != NULL) w_destroyPixr (pIPWState, &prRevText);
    if (prTmp1 != NULL) w_destroyPixr (pIPWState, &prTmp1);
    if (prTmp2 != NULL) w_destroyPixr (pIPWState, &prTmp2);
    if (prTmp3 != NULL) w_destroyPixr (pIPWState, &prTmp3);
    if (prTmpRed != NULL) w_destroyPixr (pIPWState, &prTmpRed);
    if (pCCList != NULL) w_destroyConnectedComponents (pIPWState, &pCCList);

    ERR_RETURN;

/* This is where we go if there are no images.  This is not a "failure,"
 * so we don't return an error condition and don't print error messages.
 */
noImages:
    if (rectList != NULL) w_destroyRectList (pIPWState, &rectList);
    if (prTmp != NULL) w_destroyPixr (pIPWState, &prTmp);
    if (prTextMask != NULL) w_destroyPixr (pIPWState, &prTextMask);
    if (prImages != NULL) w_destroyPixr (pIPWState, &prImages);
    if (prHalftoneMask != NULL) w_destroyPixr (pIPWState, &prHalftoneMask);
    if (prRevText != NULL) w_destroyPixr (pIPWState, &prRevText);
    if (prTmp1 != NULL) w_destroyPixr (pIPWState, &prTmp1);
    if (prTmp2 != NULL) w_destroyPixr (pIPWState, &prTmp2);
    if (prTmp3 != NULL) w_destroyPixr (pIPWState, &prTmp3);
    if (prTmpRed != NULL) w_destroyPixr (pIPWState, &prTmpRed);
    if (pCCList != NULL) w_destroyConnectedComponents (pIPWState, &pCCList);

    *ppImages = NULL;
    *pImageCount = 0;

    NORMAL_RETURN;
}


/*

 $Log: halftone.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.76  1998/08/05 18:10:54  xzhu
 xzhu on Wed Aug 5 11:09:17 PDT 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Seperated into 2 source halftone and halfton1. Moved w_removeHeadlines and 5 functions to new source halfton1.

 Revision 1.75  1998/08/04 23:18:55  xzhu
 xzhu on Tue Aug 4 16:18:33 PDT 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Added wp_getFrameMask.

 Revision 1.74  1998/08/04 00:22:13  xzhu
 xzhu on Mon Aug 3 17:21:46 PDT 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Follow change of w_maskIsRect.

 Revision 1.73  1998/04/24 00:51:09  xzhu
 xzhu on Thu Apr 23 17:50:01 PDT 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Added checking for masks. If mask is inside other mask, clean up this mask.

 Revision 1.72  1998/03/19 03:01:53  xzhu
 xzhu on Wed Mar 18 19:00:28 PST 1998

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Improved image segmentation by closing small holes.

 Revision 1.71  1997/06/26 20:01:21  ddavies
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.70  1997/05/14 21:04:06  xzhu
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 CommonCheck: If we just check edge to determine whether the image mask is rectangle, we may delete inside text. So we add something to check whether there is text inside to reduce segmentation errors.

 Revision 1.69  1997/04/15 21:54:02  ddavies
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.68  1997/03/29 01:25:55  xzhu
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 CommonCheck: Add a input parameter IsTextSeg in w_textMask(). If want keep more text, turn on this parameter, otherwise set it 0.

 Revision 1.67  1997/03/26 03:19:54  ddavies
 ddavies on Tue Mar 25 19:19:15 PST 1997

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Eliminate call to w_maskIsRect in w_segmentBinaryPagePixrs.  It was taking
 almost half of the total time and not really doing anything useful.

 Revision 1.66  1997/03/25 02:40:13  ddavies
 ddavies on Mon Mar 24 18:39:28 PST 1997

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Another iteration on w_segmentBinaryPagePixrs.  This one does a better
 job on text inside image blocks, reverse video text and text-on-tint.
 Well, what's left anyway?

 Revision 1.65  1997/01/31 00:53:30  ddavies
 ddavies on Thu Jan 30 16:51:34 PST 1997

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Add new procedure w_segmentBinaryPagePixrs.  This is a routine for segmenting
 binary images.  It attempts to find pictures in images, then remove text
 when it's on stipple and remove reverse video text from the images.
 The list of images returned should be XORed with the original image to
 remove them.  This converts the reverse video text to positive video text
 for the text plane.
 Unlike w_segmentPagePixrs, this routine returns the real image bits, not
 the image masks.

 Revision 1.64  1996/04/29 15:05:04  macleod
 macleod on Mon Apr 29 08:04:19 PDT 1996

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Added "dilate" parameter to segmentPage, etc. to support a slight
 dilation of the resulting masks. This is used to get a bit of "slop"
 around the mask to deal with the problems of multiple-resolution
 segmentation.

 Revision 1.63  1996/03/27 00:39:27  xzhu
 xzhu on Tue Mar 26 16:38:53 PST 1996

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Fixed memory leak reported from PURIFY.

 Revision 1.62  1996/03/25 17:30:26  xzhu
 xzhu on Mon Mar 25 09:29:53 PST 1996

 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 Use w_reversetextMask in w_segmentPageMask().

 Revision 1.61  1996/01/27 22:37:41  ejaquith
 halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v

 updated call to w_xformRectList()

 * Revision 1.60  1996/01/20  00:33:48  xzhu
 * xzhu on Fri Jan 19 16:31:33 PST 1996
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added protection in w_segmentPagePirxs. When w_segmentPageMask return
 nothing,
 * w_segmentPagePirxs will crash. Modified w_segmentPageMask to improve
 performance.
 *
 * Revision 1.59  1996/01/04  20:24:00  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.58  1995/12/13  04:09:16  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.57  1995/11/30  17:18:04  macleod
 * macleod on Thu Nov 30 09:17:52 PST 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Changed comment headers, to shut up the nag script
 *
 * Revision 1.56  1995/11/16  19:47:22  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Made comments agree with comments in header file
 *
 * Revision 1.55  1995/11/13  20:12:46  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.54  1995/11/04  01:35:55  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing
 ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.53  1995/10/17  23:48:34  ddavies
 * ddavies on Tue Oct 17 16:47:36 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fix bug in w_segmentPage in which an error was detected when
 w_segmentPageMask
 * decided that there weren't any segments (so returned prResultMask == NULL),
 * and we called rabb on prResultMask anyway.
 *
 * Revision 1.52  1995/10/13  22:01:59  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.51  1995/10/12  14:17:44  macleod
 * macleod on Thu Oct 12 07:16:16 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added segmentPagePixrs, which segments a page like segmentPage did,
 * only outputs an array of PIXRs and (x,y) offsets, one for each
 * segment, rather than a single bitmap and rectangle list, like
 * segmentPage did. Also split segmentPage into two modules, segmentPage
 * and segmentPageMask; the former is the original interface, the latter
 * only puts out a mask, and is called by both segmentPage and
 * segmentPagePixrs.
 *
 * Revision 1.50  1995/10/03  01:58:53  macleod
 * macleod on Mon Oct 2 18:58:53 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in
 interfaces.
 *
 * Revision 1.49  1995/09/21  22:53:39  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.48  1995/09/20  00:05:56  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fixed bug in segmentPage caused by change in w_threshold semantics
 *
 * Revision 1.47  1995/09/19  19:19:02  macleod
 * macleod on Tue Sep 19 12:19:02 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi
 parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.46  1995/09/11  21:53:41  macleod
 * macleod on Mon Sep 11 14:53:40 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.45  1995/09/07  01:32:27  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's
 declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.44  1995/09/06  20:20:26  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.43  1995/09/05  19:00:59  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Removed calls to GET_STREL macro, instead calling getStrel()
 function.
 * This required using local variables rather than invoking GET_STREL
 * as an argument to another function, since getStrel can return an error
 * code which should be checked.
 *
 * Revision 1.42  1995/09/01  21:02:10  ejaquith
 * ejaquith on Fri Sep 1 14:02:10 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.41  1995/08/31  21:14:23  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.40  1995/08/29  00:10:44  ejaquith
 * ejaquith on Mon Aug 28 17:08:46 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fixed a ton o' bugs/typos/etc that "segment" target in ipwtest was
 * generating.
 * zeroPr() was being used incorrectly in 2 instances.
 * rabb() wasn't being error checked in 2 instances.
 * result pixr was getting set to NULL when it shouldn't be resulting in
 * leaks and errors.
 *
 * Revision 1.39  1995/08/28  22:43:00  jcraig
 * jcraig on Mon Aug 28 15:42:47 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * removed include of debug.h - which is not necessary
 *
 * Revision 1.38  1995/08/25  01:36:42  ejaquith
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fixed warnings
 *
 * Revision 1.37  1995/08/25  01:08:38  jcraig
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * get rid of pc compiler warnings
 *
 * Revision 1.36  1995/08/17  23:00:23  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Needed to comment out the #include of "shros.pub"
 *
 * Revision 1.35  1995/08/17  21:55:20  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.34  1995/07/27  20:58:14  macleod
 * macleod on Thu Jul 27 13:57:21 PDT 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * When halftoneThresh and segmentDocument found no segments, they
 * returned an "error", which caused our purify regression test to think
 * that it didn't work. Changed it so that it doesn't generate an error,
 * since it's not really an erroneous result.
 *
 * Revision 1.33  1995/01/24  21:59:32  lperry
 * lperry on Tue Jan 24 13:59:09 PST 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Incorporate changes to use new strel manager interface.
 * Only allocate actual strels required.
 *
 * Revision 1.32  1995/01/11  22:42:50  dferrell
 * dferrell on Wed Jan 11 14:42:34 PST 1995
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Changed include of shrip.pub to shrcc.pub.
 *
 * Revision 1.31  1994/12/08  04:41:55  ddavies
 * ddavies on Wed Dec 7 20:41:10 PST 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Move the routines segmentGrays and segmentTwice (and it's callback proc)
 * to /product/ipcore/ipwrap/obsolete/halftone.obs.  These are no longer used.
 *
 * Revision 1.30  1994/11/04  03:17:17  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * textMask takes a DPI parameter. It is only used when the image is grayscale.
 *
 * Revision 1.29  1994/10/28  01:59:31  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * There were a couple of places which called strel manager but didn't
 * terminate. There was also a place in segmentDocument where the
 * rectangle list wasn't freed. It no longer leaks.
 *
 * Revision 1.28  1994/10/20  02:07:51  macleod
 * macleod on Wed Oct 19 19:06:57 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added a scheme to find rectangular outlines and keep them in the
 * image mask (in segmentDocument) so that images with outlines get their
 * edges preserved.
 *
 * Revision 1.27  1994/10/19  22:10:54  macleod
 * macleod on Wed Oct 19 15:09:59 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Set strels created in halftoneThresh to NULL (return value of destroyStrel)
 * after destroying them, so that they aren't destroyed twice.
 *
 * Revision 1.26  1994/10/14  19:47:23  macleod
 * macleod on Fri Oct 14 12:41:04 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added yet-another two-threshold segmenter, called segmentDocument.
 * It makes two calls to halftoneThresh, at the two thresholds passed
 * ("text" and "background"). It finds the text by calling textMask, and
 * chops that out of the background mask. It then finds the gutters (long
 * vertical or horizontal white runs), and eliminates all gutters that
 * don't have a perpindicular gutter at their endpoint. The gutters are
 * subtracted from the "background" mask, which has the effect of
 * chopping images away from bleed-through and other failures. The
 * final mask is constructed by seed-filling using the text mask as
 * a seed and the background mask as a clip region. The resulting mask
 * gets its holes plugged, and then regions in the mask which are
 * nearly rectangular are filled in.
 *
 * Revision 1.25  1994/10/07  19:05:42  lperry
 * lperry on Fri Oct 7 12:04:59 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.24  1994/08/29  22:51:21  macleod
 * macleod on Mon Aug 29 15:50:22 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fixed a stupidity in segmentTwice where if the dpi passed was <100,
 * the minimum box size passed to rabbRuns became zero, due to misplaced
 * parenthesis.
 * (er, parentheses)
 *
 * Revision 1.23  1994/08/29  21:56:49  macleod
 * macleod on Mon Aug 29 14:56:24 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * added error checks after PIXR creation in segmentTwice
 *
 * Revision 1.22  1994/08/17  21:45:01  macleod
 * macleod on Wed Aug 17 14:45:00 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added a heuristic ("hack") to the headline remover so that it doesn't think
 * quite so many small pictures are headlines: it compares the number of 1, 2,
 * and 3-pixel runs to the number of runs at the peak of the run-length
 * histogram. In images, there are usually more runs near the bottom, even
 * if the peak happens to be higher, but in headline text, there are usually
 * very few short runs. Dark, solid imaes can sometimes still be called
 * "headline" and removed from the mask..
 *
 * Revision 1.21  1994/08/16  21:51:07  macleod
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Silenced a warning.
 *
 * Revision 1.20  1994/08/16  06:39:56  macleod
 * macleod on Mon Aug 15 23:37:49 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added removeHeadlines, which is a fairly simple headline detector, called
 * from the currently 'experimental' segmenter, segmentTwice.
 * It decides what's really a headline (and therefore to be left alone)
 * by first only checking things <= 1.5 inches high, and then by
 * taking a run-length histogram (horizontal) and looking at the peak index.
 * It hasn't been tested very thoroughly (7 documents).
 *
 * Revision 1.19  1994/08/12  00:34:15  macleod
 * macleod on Thu Aug 11 17:30:56 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Changed segmentTwice to use a fairly brain-dead scheme which seems
 * to make people happy--just fill in the bounding rectangle of the
 * best-for-text-thresholded pictures, then subtract off stuff that we
 * find as text in the original best-for-text image - the image mask (not
 * filled). The result is that rectangular pictures almost never turn out
 * grungy, even when there's lots of bleed-through. There are
 * probably better solutions, but finding a "right" threshold below which
 * one magically gets all of the pictures and none of the background or
 * bleed-through or tint isn't one of them.
 *
 * Revision 1.18  1994/08/10  21:59:01  lperry
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Changes made to use new strel manager.
 *
 * Revision 1.17  1994/08/05  01:42:40  macleod
 * macleod on Thu Aug 4 18:38:54 PDT 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added segmentTwice, which takes a grayscale image and two thresholds,
 * and does a "Twice Cooked" segmentation, where the "text" thresholded
 * version is segmented, and the result is used as a seed to fill into
 * a mask consisting of the result of the segmentation of the "background"
 * thresholded image. This does an OK job of filling in more image in
 * the resulting mask without filling in all of the bleed-through that
 * might have been thresholded into the background image. Note that
 * there are still cases where bleed-through will come through, such
 * as when the image on the front of the page and the image bleeding
 * through from the back of the page overlap. A better solution is to
 * put black paper behind the scanned object: starting with good data
 * is always better than trying to "fix" bad data, because your "fix"
 * is always wrong some of the time.
 *
 * Revision 1.16  1994/03/02  18:49:38  ddavies
 * ddavies on Wed Mar 2 10:47:29 PST 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fix bug in segmentGrays in which the source mask and source gray images
 * were being freed at 1x resolution and the dest image was a copy of the
 * source mask image.  What I really meant to say was the dest image was
 * just another pointer to the source mask image, so the caller would get
 * confused when it tried to free both of them.
 *
 * Revision 1.15  1994/03/02  03:57:13  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added lots of IP_YIELDs to give up control during halftoneThresh and
 graySegment.
 *
 * Revision 1.14  1994/02/16  03:59:57  ddavies
 * ddavies on Tue Feb 15 19:59:49 PST 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Fix up comments in segmentGrays.
 *
 * Revision 1.13  1994/02/16  03:57:29  ddavies
 * ddavies on Tue Feb 15 19:57:02 PST 1994
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Added implementation of segmentGrays.  This routine attempts to find "image"
 * portions of an image based on whether there were largish blocks of gray
 * stuff.  halftoneThresh tries to find images based purely on size.
 * segmentGrays is better about not classifying large text as image.
 * It probably needs more tuning.
 *
 * Revision 1.12  1993/12/17  22:49:07  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * halftoneThresh now returns a linked list of RECTANGLs instead of a BOXES
 thing.
 * Most of the changes revolve around this.  We now fill in the holes in the
 * mask before calling rabb since the new rabb deals happily with large black
 areas
 * and somewhat less happily with noisy edges.
 * We now transform the rectlist in place instead of making it into a BOXES
 thing,
 * transforming that into another BOXES thing and throwing the first one away.
 * "Transform" means scale the box coordinates so they're work on the full-size
 * mask.
 *
 * Revision 1.11  1993/12/03  01:16:23  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.10  1993/11/20  03:20:53  ddavies
 * ddavies on Fri Nov 19 19:18:56 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Oh yeah, also change the meaning of ruleWidth.  It used to specify the
 * width of lines to be eliminated from the mask at full scale.  These sizes
 * were adjusted in the reduced mask according to the reduction factor.
 * However, the EZScan folks really used the reduction factor to give all
 * images approximately the same resolution, so rule widths should be
 * roughly constant in the reduced images.  ruleWidth now specifies the
 * width of lines in the reduced image directly.
 *
 * Revision 1.9  1993/11/20  03:17:06  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Modify this routine to take minWidth and minHeight.  Bounding boxes that
 * are thinner than minWidth or shorter than minHeight are not returned in
 * the box list.
 * Also change the algorithm a little.  The strels used to remove horizontal
 * and vertical rules are now (2*ruleWith)+1 x 5 and 5 x (2*ruleWidth)+1 now.
 * Limiting the length of the rule section being matched to 5 was done to
 * avoid very long delays when segmenting large images with a reduction
 * factor of 1.  We now also remove 1x1, 2x1, 1x2 and 2x2 dots in the reduced
 * mask before calling rabb in the hope that this will make life easier.  It
 * didn't have a noticeable affect in my tests, but we'll let the EZScan guys
 * try it.
 *
 * Revision 1.8  1993/11/18  23:42:07  ddavies
 * ddavies on Thu Nov 18 15:41:43 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Also put strel1Inits back in so strel5h and strel5v are defined.  How did
 * it disappear??
 *
 * Revision 1.7  1993/11/18  23:39:39  ddavies
 * ddavies on Thu Nov 18 15:38:06 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * In halftoneThresh, only reduce the seed to 1/4 the size of the reduced mask
 * instead of always reducing the seed by 16x.  With small images that aren't
 * reduced, the seed was getting too small and seed pixels were disappearing.
 * This fixes a bug found by EZScan.
 *
 * Revision 1.6  1993/11/12  01:01:06  ddavies
 * ddavies on Thu Nov 11 17:00:16 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * If the mask is not expanded because it's already at full size, return it
 * directly instead of returning a null image.
 *
 * Revision 1.5  1993/11/11  23:06:28  ddavies
 * ddavies on Thu Nov 11 15:05:51 PST 1993
 *
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Don't expand the mask back to full size if it's already full size.  Trying
 * to do so caused an error in expandPixr.
 *
 * Revision 1.4  1993/11/06  03:13:12  ddavies
 * halftone.c -> /product/ipcore/ipwrap/src/RCS/halftone.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

