/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  	hauscall.c:   Some Hausdorff comparator callback routines
 *
 *  	    Code to find single best match via callback
 *  	    	  w_best1Callback()
 *
 *  	    Code to find best single 'ideal' match via callback
 *  	    	  w_bestIdeal1Callback()
 *
 *  	    Code to find best 'N' matches (N == some integral number)
 *  	    via callback
 *  	    	  w_initBestNCallback()
 *  	    	  w_bestNCallback()
 *  	    	  w_averageBestNCallback()
 *  	    	  w_finishBestNCallback()
 *
 *
 * A general description of the Hausdorff method and some general searching
 * routines are included in haussrch.c
 */


#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "alplib.h"

/* prototypes */
#include "hauscall.pub"
#include "props.pub"




IP_RCSINFO(RCSInfo, "$RCSfile: hauscall.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");


/* Given two points, calculate the Euclidean distance, squared.
 * Useful for comparisons (without the costly sqrt() ). */
#define DISTANCE_SQD(ptax,ptay,ptbx,ptby) \
    (((ptax)-(ptbx))*((ptax)-(ptbx)) + ((ptay)-(ptby))*((ptay)-(ptby)))



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Hausdorff callback routine to find single best match
 *
 * LongDescription:::
 *  Record the currently reported match as the new best match.
 *  Return and constrain all following reported matches be AT LEAST as good
 *  as currently reported.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pData::          Ptr to internal information to use in this routine
 *  tx::             Hausdorff returned match X location value
 *  ty::             Hausdorff returned match Y location value
 *  percAtoB::       A to B match percent at this location
 *  percBtoA::       B to A match percent at this location
 *
 * Output:::
 *  pNextAsGood::    This routine will set this to point to 1.
 *                   1 to make all further routine invocations
 *                     report matches AT LEAST as good as the match
 *                     passed to this current invocation.
 *                   (0 would keep same matching criterion as before
 *                     this invocation.)
 *
 * Returns:::
 *
 * Notes:::
 *  pData param::
 *   This void ptr is to have been casted to a ptr to a "matchData" struct.
 *   If you want to check if this routine was invoked during the entire search
 *   while attempting to find Hausdorff matches, then initialize the AtoB
 *   field of the struct the void * points to to be -1. before the search is
 *   begun.  If this field is still -1. after the search is over, then we know
 *   this routine was never invoked.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_best1Callback (
    IPWState           *pIPWState,
    void               *pData,
    Int32               tx,
    Int32               ty,
    Float64             percAtoB,
    Float64             percBtoA,
    Int8               *pNextAsGood)
{
    PROLOGUE;
    matchData    *mdata = (matchData *)pData;

    ERR_GEN(pData == NULL  ||  pNextAsGood == NULL, cErrNullPtr);

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    /* A new better match, record it */
    mdata->tx       = tx;
    mdata->ty       = ty;
    mdata->percAtoB = percAtoB;
    mdata->percBtoA = percBtoA;

    /* Returning this value will allow the Hausdorff module's internal
     * criterion to be updated so that any future w_best1Callback()
     * invocations will have its:
     *   percAtoB >= this current percAtoB
     *   percBtoA >= this current percBtoA
     * This not only means less calls to w_best1Callback(), but it also means
     * more of the transation area may be pruned without full image-to-image
     * comparison and thus a faster search of the remaining area.
     */
    *pNextAsGood = 1;
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Hausdorff callback routine to find single best ideal match
 *
 * LongDescription:::
 *  Will return the "best ideal" match in a search.  The "best ideal" match
 *  is defined as the match that has the highest Hausdorff percents when the
 *  percents are less than the ideal Hausdorff percents and as the match
 *  that is closest to an ideal location when the percents are at least the
 *  ideal Hausdorff percents.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pData::          Ptr to internal information to use in this routine
 *  tx::             Hausdorff returned match X location value
 *  ty::             Hausdorff returned match Y location value
 *  percAtoB::       A to B match percent at this location
 *  percBtoA::       B to A match percent at this location
 *
 * Output:::
 *  pNextAsGood::    This routine will set this to point to 0 or 1.
 *                   1 to make all further routine invocations
 *                     report matches AT LEAST as good as the match
 *                     passed to this current invocation.
 *                   0 to keep same matching criterion as before
 *                     this invocation.
 *
 * Returns:::
 *
 * Notes:::
 *  pData param::
 *   This void ptr is to have been casted to a ptr to an "idealData" struct.
 *   Before a search may call back this routine, the client must have
 *   initialized the fields: txIdeal and tyIdeal, the ideal location, and
 *   percAtoBIdeal and percBtoAIdeal, the ideal Hausdorff percents.
 *   Additionally, the fields percAtoB and percBtoA should be -1.
 *   Field percBtoAIdeal should be inited to -1. when 1-Way matching is to
 *   be done.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_bestIdeal1Callback (
    IPWState           *pIPWState,
    void               *pData,
    Int32               tx,
    Int32               ty,
    Float64             percAtoB,
    Float64             percBtoA,
    Int8               *pNextAsGood)
{
    PROLOGUE;
    Int32       tmp;
    idealData  *idata = (idealData *)pData;
    
    ERR_GEN(pData == NULL  ||  pNextAsGood == NULL, cErrNullPtr);

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    if (percAtoB >= idata->percAtoBIdeal &&
	percBtoA >= idata->percBtoAIdeal)
    {
	/* Finding matches that have at-least-the-ideal percents,
	 * so continue keeping the match that is closest to the
	 * ideal location.
	 */
	
	if (idata->percAtoB < idata->percAtoBIdeal ||
	    idata->percBtoA < idata->percBtoAIdeal)
	{
	    /* Percents at this match are at-least-the-ideal.  This is 
	     * the first match where this is true.
	     */
	    idata->tx = tx;
	    idata->ty = ty; 
	    idata->percAtoB = percAtoB;
	    idata->percBtoA = percBtoA;
	    idata->distSqdToIdeal = DISTANCE_SQD(tx, ty, idata->txIdeal,
						 idata->tyIdeal);
	}
	else
	{
	    tmp = DISTANCE_SQD(tx, ty, idata->txIdeal, idata->tyIdeal);
	    if (tmp < idata->distSqdToIdeal)
	    {
		/* Percents at this match are at-least-the-ideal and this
		 * match is closer than the previous at-least-the-ideal match,
		 * so record it!
		 */
		idata->tx = tx;
		idata->ty = ty; 
		idata->percAtoB = percAtoB;
		idata->percBtoA = percBtoA;
		idata->distSqdToIdeal = tmp;
	    }
	}

	*pNextAsGood = 0;
	NORMAL_RETURN;
    }
    else
    {
	/* Still finding less-than-ideal matches, so continue keeping the
	 * match with the best percents */

	if (percAtoB >= idata->percAtoB &&
	    percBtoA >= idata->percBtoA)
	{
	    /* Keep the match with the best percents, or if they tie,
	     * keep the match that is closest to the ideal location.
	     */
	    if (percAtoB > idata->percAtoB ||
		percBtoA > idata->percBtoA ||
		(DISTANCE_SQD(tx, ty, idata->txIdeal, idata->tyIdeal) <
		 DISTANCE_SQD(idata->tx, idata->ty,
			      idata->txIdeal, idata->tyIdeal)))
	    {
		idata->tx = tx;
		idata->ty = ty; 
		idata->percAtoB = percAtoB;
		idata->percBtoA = percBtoA;
	    }
	}
	
	/* Can update the search criterion because we know we never want
	 * this routine invoked with a match with lessor percents than
	 * this!
	 */
	*pNextAsGood = 1;
	NORMAL_RETURN;
    }
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set up internal data (void *) for use with w_bestNCallback()
 *
 * LongDescription:::
 *  Allocate space for what's necessary for use with callback routine to keep
 *  track of the best 'N' matches, and initialize it.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  N::              Number of matches to try to find, [0..255]
 *  widthA::         Width of A model bitmap excluding frame and pad
 *  heightA::        Height of A model bitmap excluding frame and pad
 *  avgThresh::      Maximum "Manhattan" distance two overlapping matches
 *                   may be apart when we should record average
 *                   statistics.  (See notes).
 *
 * Output:::
 *  ppData::         Ptr to ptr to internal data for w_bestNCallback().
 *                   ppData is allocated and filled in.
 *                   *ppData must point to NULL on input.
 *                   *ppData will then point to the finished info on return.
 *
 * Returns:::
 *
 * Notes:::
 *  Overlapping matches::
 *    Returned match locations will NOT be allowed to overlap at all (where
 *    "overlap" is defined as where one or more pixel in the two
 *    widthA * heightA bitmaps are the same).
 *  Averaging matches::
 *    Since overlapping matches aren't allowed, whenever two matches with equal
 *    Hausdorff percents are found AND are overlapping, we will arbitrarily
 *    pick one.  However, if the two upper left corners are Manhattan
 *    distance of avgThresh or less apart, then we will record information
 *    for both matches that may be averaged at the end, by calling
 *    w_averageBestNCallback().
 *    Example, if we made A from a clipped portion of B, made dilated BD from B
 *    by STREL 1 1 1
 *             1 1c1
 *             1 1 1, and then looked for the best 1 match of A in BD, we
 *    would find exactly 9 match locations with the same percent.  If
 *    avgThresh was 4, then all 9 overlapping locations would be averaged
 *    using w_averageBestNCallback() at the end of the search, and we would
 *    get the single best match location as the center of the 9 locations.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initBestNCallback (
    IPWState           *pIPWState,
    UInt8               N,
    Int32               widthA,
    Int32               heightA,
    Int32               avgThresh,
    void              **ppData)
{
    PROLOGUE;
    matchNData  *mndata;

    ERR_GEN(ppData == NULL, cErrNullPtr);
    ERR_GEN(*ppData != NULL, cErrPtrNotNull);

    mndata = (matchNData *)IPW_MALLOC(sizeof(matchNData), cAllocatorData);
    ERR_GEN(!mndata, cErrNomem);
    mndata->matches = NULL;
    mndata->avgInfo = NULL;

    mndata->nLook = N;
    mndata->nFound = 0;
    mndata->wA = widthA;
    mndata->hA = heightA;
    mndata->avgThresh = avgThresh;
    mndata->lowestABIndex = N;
    mndata->lowestBAIndex = N;
    mndata->matches = (matchData *)IPW_CALLOC(N, sizeof(matchData),
					      cAllocatorData);
    mndata->avgInfo = (avgData *)IPW_CALLOC(N, sizeof(avgData),
					    cAllocatorData);

    ERR_GEN_CLEANUP(!mndata->matches || !mndata->avgInfo, cErrNomem);
    
    *ppData = (void *)mndata;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(mndata->matches, cAllocatorData);
    IPW_FREE(mndata->avgInfo, cAllocatorData);
    IPW_FREE(mndata, cAllocatorData);
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Average locations where multiple close matches were found
 *
 * LongDescription:::
 *  Now that the search for the best 'N' matches is over, go through the
 *  list of best 'N' matches.  Matches where there were many equally good
 *  matches to pick from had one arbitrary location assigned as the best match.
 *  Replace this location instead with the average x and average y
 *  values from these many equally good locations.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  pData::          Ptr to internal data for w_bestNCallback
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_averageBestNCallback (
    IPWState           *pIPWState,
    void               *pData)
{
    PROLOGUE;
    matchNData   *mndata;
    matchData    *mdata;
    UInt8         index;

    ERR_GEN(pData == NULL, cErrNullPtr);

    mndata = (matchNData *)pData;
    mdata  = mndata->matches;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    for (index = 0; index < mndata->nFound; index++)
    {
	mdata[index].tx = (Int32)(((Float32)mndata->avgInfo[index].sumx / 
				   mndata->avgInfo[index].n) + 0.5);
	mdata[index].ty = (Int32)(((Float32)mndata->avgInfo[index].sumy /
				   mndata->avgInfo[index].n) + 0.5);
	
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free w_bestNCallback information and set ptr to NULL
 *
 * LongDescription:::
 *  Given the ptr to the internal data, free its allocated memory. 
 *  If already NULL, this routine
 *  does nothing; o.w. it will set ptr to NULL upon exiting.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppData::         Ptr to ptr to internal data for w_bestNCallback
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_finishBestNCallback (
    IPWState           *pIPWState,
    void              **ppData)
{
    PROLOGUE;

    ERR_GEN(ppData == NULL, cErrNullPtr);

    if (*ppData != NULL) {
	matchNData   **mndata = (matchNData **)ppData;

	IPW_FREE((*mndata)->matches, cAllocatorData);
	IPW_FREE((*mndata)->avgInfo, cAllocatorData);
	IPW_FREE(*mndata, cAllocatorData);

	*ppData = NULL;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Hausdorff callback routine to find best 'N' matches
 *
 * LongDescription:::
 *  Callback to feed to Hausdorff search routine that will find and keep track
 *  of the best 'N' "distinct" matches.  Two match locations are distinct if
 *  the A bitmap at each location do not overlap.  When overlap occurs and
 *  additionally if the locations are "close", keep track of the average
 *  location.  "close" is defined by the avgThresh parameter passed
 *  in to w_initBestNCallback().
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pData::          Ptr to internal information to use in this routine
 *  tx::             Hausdorff returned match X location value
 *  ty::             Hausdorff returned match Y location value
 *  percAtoB::       A to B match percent at this location
 *  percBtoA::       B to A match percent at this location
 *
 * Output:::
 *  pNextAsGood::    This routine will set this to point to 0 or 1.
 *                   1 to make all further routine invocations
 *                     report matches AT LEAST as good as the match
 *                     passed to this current invocation.
 *                   0 to keep same matching criterion as before
 *                     this invocation.
 *
 * Returns:::
 *
 * Notes:::
 *  Updating slots::
 *    Once 'N' slots have been found with acceptable minimum Hausdorff
 *    percents, the question becomes how do we find even better matches if
 *    they exist?  Well, we obviously will have to start overwriting slots
 *    at this point, and the criterion is as follows:
 *      Overwrite the slot with the least Hausdorff A to B percent
 *    Notice this DOES NOT care what the B to A percent is.
 *  Setting *pNextAsGood to 1::
 *          This is done when N' distinct matches have been
 *  	    found (this return value will then constrain any following routine
 *  	    invocations to report matches AT LEAST as good as this match
 *  	    in the current invocation (this also enables faster searching!))
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_bestNCallback (
    IPWState           *pIPWState,
    void               *pData,
    Int32               tx,
    Int32               ty,
    Float64             percAtoB,
    Float64             percBtoA,
    Int8               *pNextAsGood)
{
    PROLOGUE;
    UInt8         index;
    matchNData   *mndata;
    matchData    *mdata;

    ERR_GEN(pData == NULL  ||  pNextAsGood == NULL, cErrNullPtr);

    mndata = (matchNData *)pData;
    mdata = mndata->matches;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    /* The lowestABIndex and lowestBAIndex fields used below may seem a bit
     * confusing.  We simply want to know when it is okay to return '1'
     * (via pNextAsGood)
     * from this routine.  We can do so if all the slots are filled AND
     * if the passed in match location's percents are worse than any that
     * appear in the slots.
     */

    for (index = 0; index < mndata->nFound; index++)
    {
	if (MIN(ty, mdata[index].ty) + mndata->hA >= MAX(ty, mdata[index].ty)
	    &&
	    MIN(tx, mdata[index].tx) + mndata->wA >= MAX(tx, mdata[index].tx))
	{
	    /* Means this new match overlaps this previously recorded match
	     * in this slot/index */
	    
	    if (percAtoB >= mdata[index].percAtoB &&
		percBtoA >= mdata[index].percBtoA)
	    {
		if (percAtoB > mdata[index].percAtoB ||
		    percBtoA > mdata[index].percBtoA)
		{
		    /* Our new match overlaps this match AND has greater
		     * match percents, so replace slot with new match.
		     */
		    mdata[index].tx = tx;
		    mdata[index].ty = ty;
		    mdata[index].percAtoB = percAtoB;
		    mdata[index].percBtoA = percBtoA;
		    
		    mndata->avgInfo[index].sumx = tx;
		    mndata->avgInfo[index].sumy = ty;
		    mndata->avgInfo[index].n = 1;

		    if (mndata->nFound == mndata->nLook &&
			mndata->lowestABIndex == index)
		    {
			/* Update markers to slot(s) with the lowest percents
			 */
			mndata->lowestABIndex = mndata->lowestBAIndex = 0;
			for (index = 1; index < mndata->nFound; index++)
			{
			    if (mdata[index].percAtoB <=
				mdata[mndata->lowestABIndex].percAtoB)
				mndata->lowestABIndex = index;
			    if (mdata[index].percBtoA <=
				mdata[mndata->lowestBAIndex].percBtoA)
				mndata->lowestBAIndex = index;
			}
		    }
		}
		else if (ABS(tx - mdata[index].tx) + ABS(ty - mdata[index].ty)
			 <= mndata->avgThresh)
		{
		    /* Our new match overlaps this match, has the same
		     * match percents, AND is quite close to this match,
		     * so add to averaging statistics.
		     */
		    mndata->avgInfo[index].sumx += tx;
		    mndata->avgInfo[index].sumy += ty;
		    mndata->avgInfo[index].n++;
		}
		/* else this new match overlaps this match, has the same
		 * match percents, BUT is further away from this match than
		 * the threshold averaging distance, so chuck it by doing
		 * nothing.
		 */
	    }
	    /* else this new match isn't as good as the match
	     * it overlaps, so chuck it by doing nothing.
	     */

	    /* If all the slots are full and the match reported to this
	     * callback had percents no greater than the slot with the
	     * lowest percents, then return 1 (via pNextAsGood)
	     * to tighten up the matching
	     * so the next time we are called we will be guaranteed to
	     * be at least as good as this current reported one!
	     */
	    *pNextAsGood = ((mndata->nFound == mndata->nLook &&
		     mdata[mndata->lowestABIndex].percAtoB >= percAtoB &&
		     mdata[mndata->lowestBAIndex].percBtoA >= percBtoA)
		    ? 1 : 0);
	    NORMAL_RETURN;
	}

    }

    /* At this point, we know the new match did not overlap any presently
     * recorded matches!
     */

    /* If there is a match slot still open, fill it with this new match.
     * Otherwise, replace the slot that had the
     *      LOWEST A TO B PERCENT   <-- NOTE not ignores B to A percent!!
     * with this new match. */
    if (mndata->nFound == mndata->nLook)
	index = mndata->lowestABIndex;
    else
	/* NOTE! If mndata->nFound == mndata->nLook-1 at this point
	 * then that means we will fill the slots with this new match.
	 * We'll need to keep track of the lowest slot(s) from here on out
	 */
	index = mndata->nFound++;


    mdata[index].tx = tx;
    mdata[index].ty = ty;
    mdata[index].percAtoB = percAtoB;
    mdata[index].percBtoA = percBtoA;
    
    mndata->avgInfo[index].sumx = tx;
    mndata->avgInfo[index].sumy = ty;
    mndata->avgInfo[index].n = 1;

    if (mndata->nFound == mndata->nLook)
    {
	/* We either overwrote a slot above ('cause they were all full before)
	 * or we just filled the last slot.
	 * Now (re)calculate the slot(s) with the lowest percents */
	mndata->lowestABIndex = mndata->lowestBAIndex = 0;
	for (index = 1; index < mndata->nFound; index++)
	{
	    if (mdata[index].percAtoB <= mdata[mndata->lowestABIndex].percAtoB)
		mndata->lowestABIndex = index;
	    if (mdata[index].percBtoA <= mdata[mndata->lowestBAIndex].percBtoA)
		mndata->lowestBAIndex = index;
	}

	*pNextAsGood = ((mdata[mndata->lowestABIndex].percAtoB >= percAtoB &&
		 mdata[mndata->lowestBAIndex].percBtoA >= percBtoA) ? 1 : 0);
	NORMAL_RETURN;
    }
    else {
	*pNextAsGood = 0;
	NORMAL_RETURN;
    }
}


/*

 $Log: hauscall.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.18  1996/01/04 20:24:03  ejaquith
 hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.17  1995/12/13  04:09:20  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/13  20:12:50  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.15  1995/11/10  02:04:05  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.14  1995/10/30  17:56:58  ejaquith
 * ejaquith on Mon Oct 30 09:56:35 PST 1995
 *
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * Updated function block comments who named other IPWrap routines.
 *
 * Revision 1.13  1995/10/13  22:02:02  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.12  1995/09/29  00:34:57  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.11  1995/09/21  22:53:43  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.10  1995/09/12  16:54:43  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.9  1995/09/06  20:20:30  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  02:20:31  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.7  1995/08/31  21:14:26  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/22  02:21:04  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.5  1995/08/17  22:01:11  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.4  1995/07/24  22:26:10  ejaquith
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * Bug Fix.
 * In bestNCallback(), used with Hausdorff matching routines, we look to fill
 * N "slots" with matches.  This routine should begin finding matches, and
 * when very close but better matches are found, the slots should begin
 * to "move" toward the local maximum in the area.  However, in some instances
 * they were not.  This was because I had made the routine return 1 to
 * start increasing the minimum Hausdorff percents in the search once all
 * slots were full.  However!  What we really have to do is, once all the
 * slots are full, find the one(s) with the lowest Hausdorff percents.
 * Then, if a potential match is reported to this callback from the search and
 * its percents are worse than those in the lowest slot(s), THEN we can
 * return 1 and increase the min Hausdorff percents (but not until then).
 * Added/updated some comments.
 * Needed to add 2 fields to the public "matchNData" data struct to
 * accomodate this bug fix.
 *
 * Revision 1.3  1995/07/14  20:31:30  ejaquith
 * ejaquith on Fri Jul 14 13:30:50 PDT 1995
 *
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * Uh, not exactly a bug fix.  Let's just a
 * say that one of my inline if statements was making me nervous
 * so I expanded it.
 *
 * Revision 1.2  1995/03/30  01:35:33  ejaquith
 * ejaquith on Wed Mar 29 17:32:13 PST 1995
 *
 * hauscall.c -> /product/ipcore/ipwrap/src/RCS/hauscall.c,v
 *
 * Added bestIdeal1Callback() routine.
 * This callback will return the "best ideal" match in a search.  The
 * "best ideal" match is defined here as the match with the highest Hausdorff
 * percents when the percents are less than the ideal Hausdorff percents,
 * and as the match that is closest to an ideal location when the percents
 * are at least the ideal Hausdorff percents.  You just initialize the
 * ideal location and the ideal percents.
 * This is useful in instances when a subrect may contain multiple instances
 * of "A" but the client is really looking for one, the one near this
 * ideal location.
 *
 * Revision 1.1  1995/03/29  02:05:25  ejaquith
 * Initial revision
 *
 
 */
