/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  	hausdbug.c:
 *
 *  	    Hausdorff matching routines (mainly for debugging)
 *			w_pixrHausdorff2WayMatches()
 *			w_checkHausdorff2WayMatches()
 *
 *  w_pixrHausdorff2WayMatches():
 *  	    a strictly PIXR and slower version of
 *  	    w_hausdorffMatches(), performing the same calculations
 *
 *  w_checkHaudorff2WayMatches():
 *  	    calls both w_pixrHausdorff2WayMatches() and
 *  	    w_hausdorffMatches(), and ensures they return the same
 *  	    results!
 *
 * Extensive comments regarding the Hausdorff comparator module can be found
 * at the top of file haussrch.c.
 *
 */

#include <stdio.h>

#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "alplib.h"
#include "imagerep.h"
#include "hausconf.pub"
#include "logicops.h"
#include "hauscomp.h"

/* prototypes */
#include "intops.pub"
#include "logicops.pub"
#include "hauscomp.pub"
#include "props.pub"
#include "io.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif



IP_RCSINFO(RCSInfo, "$RCSfile: hausdbug.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");

#define DEBUG_OUT_FILES  0

typedef struct {
    PIXR   *AS;
    PIXR   *AD;
    PIXR   *BS;
    PIXR   *BD;
    Int32   nptsAS;
    Int32   nptsBS;
    UInt8   searchMeth;
} rtnChk;




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Place model over imagery and compute Hausdorff percent.
 *
 * LongDescription:::
 *  	Take the model and the imagery and compare them to see if they match,
 *  	at the appropriate passed in offset of the model over the imagery.
 *  	Do this by taking the model bits and ANDing them with the imagery bits
 *  	and determine if some suitable percent of the model points overlap
 *  	imagery points for this one way match.
 *  	No early termination.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pM::             Ptr to PIXR to use as model
 *  pI::             Ptr to PIXR to use as imagery (place model over)
 *  x::              Location w.r.t. imagery's upper left corner to place
 *  y::              model over
 *  nptsM::          Number of "on"/1 bits in model
 *  justUseOverlap:: calculate percent as:
 *                   if 0,  #ANDED pts / #pts in entire model;
 *                   o.w.   #ANDED pts / #pts in model region that
 *                   overlaps imagery.
 *  debugDumpPixrs:: a debug flag, 1 to output relevant PIXRs, 0 to do nothing
 *
 * Output:::
 *  pCalcPercent::   ptr to returned Hausdorff percent at this overlapping
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrHausdorffPercent (
    IPWState           *pIPWState,
    PIXR               *pM,
    PIXR               *pI,
    Int32               x,
    Int32               y,
    Int32               nptsM,
    UInt8               justUseOverlap,
    Int8                debugDumpPixrs,
    Float64            *pCalcPercent)
{
    PROLOGUE;
    RECTANGL rectM, rectI;
    PIXR    *overM = NULL, *overI = NULL;
    Int32    nAND, wM, wI, hM, hI;
    char     fileout[80];

    /* pM and pI have already been verified -- this routine is private */

    if (nptsM == 0)
    {
	*pCalcPercent = 0.0;
	NORMAL_RETURN;
    }

    ERR_PASS( w_pixrGetWidth(pIPWState, pM , &wM) );
    ERR_PASS( w_pixrGetWidth(pIPWState, pI , &wI) );
    ERR_PASS( w_pixrGetHeight(pIPWState, pM, &hM) );
    ERR_PASS( w_pixrGetHeight(pIPWState, pI, &hI) );

    /* Set up the two rects so they are the same height and width and
     * their (x,y) values point to the first overlapping pixels in both
     */
    if (x < 0)
    {
	rectM.x = -x;
	rectI.x =  0;
	rectM.w = rectI.w = MIN(wM + x, wI);
    }
    else
    {
	rectM.x = 0;
	rectI.x = x;
	rectM.w = rectI.w = MIN(wI - x, wM);
    }
    if (y < 0)
    {
	rectM.y = -y;
	rectI.y =  0;
	rectM.h = rectI.h = MIN(hM + y, hI);
    }
    else
    {
	rectM.y = 0;
	rectI.y = y;
	rectM.h = rectI.h = MIN(hI - y, hM);
    }

    if (rectM.w <= 0  ||  rectM.h <= 0)
    {
	/* Means no portion of the imagery and model overlapped! */
	*pCalcPercent = 0.0;
	NORMAL_RETURN;
    }

    /* Clip the overlapping region of the model */
    ERR_PASS_CLEANUP( w_rectOf( pIPWState, pM, &rectM, &overM ));

    /* If we only want the Hausdorff percent for the number of OVERLAPPING
     * pixels in the model (and not the whole model), then readjust the
     * number of model points to be just those overlapping */
    if (justUseOverlap)
    {
	ERR_PASS_CLEANUP( w_numberBits( pIPWState, overM, &nptsM ));
        if (nptsM == 0)
	{
	    *pCalcPercent = 0.0;
	    NORMAL_RETURN;
	}
    }

    /* Clip the overlapping region of the imagery */
    ERR_PASS_CLEANUP( w_rectOf( pIPWState, pI, &rectI, &overI ));

    if (debugDumpPixrs)
    {
	sprintf(fileout, "%s", "model.overlap.pbm");
	ERR_PASS_CLEANUP( w_write(pIPWState, fileout, overM, PNM));
	sprintf(fileout, "%s", "imagery.overlap.pbm");
	ERR_PASS_CLEANUP( w_write(pIPWState, fileout, overI, PNM));
    }

    /* Bitwise AND the two portions and place result in overM */
    ERR_PASS_CLEANUP( w_logicOp ( pIPWState, overI, INTERSECTION, overM));

    if (debugDumpPixrs)
    {
	sprintf(fileout, "%s", "anded.overlap.pbm");
	ERR_PASS_CLEANUP( w_write(pIPWState, fileout, overM, PNM));
    }

    ERR_PASS_CLEANUP( w_numberBits( pIPWState, overM, &nAND ));

    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &overM ));
    ERR_PASS_CLEANUP( w_destroyPixr( pIPWState, &overI ));
    
    *pCalcPercent = (Float64) nAND / nptsM;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr( pIPWState, &overM );
    w_destroyPixr( pIPWState, &overI );
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (slow) Find all 2-Way matches of A in subrect of B
 *
 * LongDescription:::
 *  Given A, a model bitmap AS and its dilation AD, match it to B, a model
 *  bitmap BS and its dilation BD in an array of location translation offsets
 *  to try overlapping A and B at.
 *  When the Hausdorff percents of correlation between A to B and B to A
 *  are at least the passed in thresholds, report this location as a
 *  match invoking the client's callback routine.
 *  <BR><BR>
 *  Does the 2 way matching of matching model AS to imagery BD,
 *  and model BS to imagery BD.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  AS::             ptr to source A binary pixr
 *  AD::             ptr to A-dilated binary pixr, dimensions as AS
 *  BS::             ptr to source B binary pixr
 *  BD::             ptr to B-dilated binary pixr, dimensions as BS
 *  x::              Rectangular area in B where we search
 *  y::              for A to match.
 *  w::              Must be entirely inside B.
 *  h::              
 *  skipx::          Amounts to skip.  Check for a match    
 *  skipy::          every skipx pixel in X, every skipy pixel
 *                   in Y in subrect translation area.  (1 or greater)
 *  searchMeth::     Desired search method. One of:           
 *                   {SPARSE_FILTER, FLY_FILTER, EVERY_TRANS} 
 *                   May be OR-ed with USE_FULL_B_FOR_BTOA_PERCENT
 *  pCallbackData::     Ptr to client supplied data that we      
 *                      include as an argument when calling the  
 *                      client's callback proc.
 *  callbackProc::      Ptr to the client's callback routine,
 *                      called with the match's translation x
 *                      and y offset, the A to B match percent,
 *                      the B to A match percent, (ignored if 1-Way)
 *                      the client supplied argument, and a pointer to a
 *                      value to shape further invocations (see Notes).
 *  pMinPercentAtoB::           The match(es) must have A to B Hausdorff 
 *                              percent(s) at least the value this points to,
 *                              [0..1].  See notes below.
 *  pMinPercentBtoA::           The 2-Way match(es) must additionally
 *                              have B to A Hausdorff percent(s) at least
 *                              the value this points to, [0..1].
 *                              See notes below
 *
 * Output:::
 *  pCallbackReturnCode::       Ptr to code returned by client's
 *                              callback routine
 *
 * Returns:::
 *
 * Notes:::
 *  If skipx == skipy == 1::
 *    will try every translation in the subrect.
 *  Match locations::
 *    are in the frame of the full B image.
 *  callbackRoutine pNextAsGood value::
 *    Return 1 to make all further routine invocations report matches AT LEAST
 *      as good as the match in the current invocation.
 *    Return 0 to keep same matching criterion as before invocation.
 *  pMinPercentAtoB, pMinPercentBtoA::
 *    On completion, these will point to the min percents last used in the
 *    search.  (One can inspect if they had been increased by the callback).
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrHausdorff2WayMatches (
    IPWState           *pIPWState,
    PIXR               *AS,
    PIXR               *AD,
    PIXR               *BS,
    PIXR               *BD,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               skipx,
    Int32               skipy,
    UInt8               searchMeth,
    void               *pCallbackData,
    ErrCode	(CDECL *callbackProc)( IPWState *pIPWState, void *,
				       Int32 tx, Int32 ty,
				       Float64 percentAtoB,
				       Float64 percentBtoA,
				       Int8 *pNextAsGood),
    Float64            *pMinPercentAtoB,
    Float64            *pMinPercentBtoA,
    ErrCode            *pCallbackReturnCode)
{
    PROLOGUE;
    Int32    i, j, hAS, wAS;
    Float64  percAtoB, percBtoA;
    Float64  minPercentAtoB, minPercentBtoA;
    Int32    nptsAS, nptsBS;
    Int8     nextAsGood;

    ERR_GEN_PIXR(AS, cByteOrderBig, cDepth1, cChannels1, cFrame32 | cFrame0);
    ERR_GEN_PIXR(AD, cByteOrderBig, cDepth1, cChannels1, cFrame32 | cFrame0);
    ERR_GEN_PIXR(BS, cByteOrderBig, cDepth1, cChannels1, cFrame32 | cFrame0);
    ERR_GEN_PIXR(BD, cByteOrderBig, cDepth1, cChannels1, cFrame32 | cFrame0);

    ERR_GEN(callbackProc == NULL  ||
	    pMinPercentAtoB == NULL  ||  pMinPercentBtoA == NULL  ||
	    pCallbackReturnCode == NULL, cErrNullPtr);

    ERR_PASS( w_numberBits(pIPWState, AS, &nptsAS) );
    ERR_PASS( w_numberBits(pIPWState, BS, &nptsBS) );

    ERR_GEN(*pMinPercentAtoB < 0. ||
	    *pMinPercentBtoA < 0. ||
	    *pMinPercentAtoB > 1. ||
	    *pMinPercentBtoA > 1., cErrInvalidParm);
    ERR_GEN(skipx < 1 || skipy < 1, cErrInvalidParm);

    minPercentAtoB = *pMinPercentAtoB;
    minPercentBtoA = *pMinPercentBtoA;

    ERR_PASS( w_pixrGetWidth(pIPWState, AS, &wAS) );
    ERR_PASS( w_pixrGetHeight(pIPWState, AS, &hAS) );

    for (j = y; j < y+h-hAS; j += skipy)
    {
	for (i = x; i < x+w-wAS; i += skipx)
	{
	    /* Get the A to B 1-way match percent for placing the A model over
	     * the B imagery at this location */
	    ERR_PASS( wp_pixrHausdorffPercent(pIPWState, AS, BD,
					      i, j, nptsAS,
					      0,
					      0, &percAtoB));
	    
	    if (percAtoB >= minPercentAtoB)
	    {
		/* Get the B to A 1-way match percent for placing the B model
		 * over the A imagery at the corresponding location */
		ERR_PASS( wp_pixrHausdorffPercent(pIPWState, BS, AD,
						  -i, -j, nptsBS,
						  (UInt8)((searchMeth &
							   USE_FULL_B_FOR_BTOA_PERCENT)
							  ? 0 : 1),
						  0, &percBtoA));

		if (percBtoA >= minPercentBtoA)
		{
		    /* Found a 2-way match for the models and imagerys at this
		     * location, so report it to the caller's routine */
		    ERR_CHECK(callbackProc(pIPWState, pCallbackData,
					   i, j,
					   percAtoB, percBtoA,
					   &nextAsGood)) 
			{
			    *pCallbackReturnCode = __error__;
			    ERR_GEN(1, cErrCallbackError);
			}
		    if (nextAsGood)
		    {
			minPercentAtoB = percAtoB;
			minPercentBtoA = percBtoA;
		    }
		}
	    }
	}
    }

    *pMinPercentAtoB = minPercentAtoB;
    *pMinPercentBtoA = minPercentBtoA;

    /* This to indicate no error found in callback */
    *pCallbackReturnCode = cSuccessful;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Double check the IPCore Hausdorff calculations at this location
 *
 * LongDescription:::
 *  This routine gets invoked whenever a pair of Hausdorff percents gets
 *  returned in w_checkHausdorff2WayMatches()
 *  It recalculates the pair of Hausdorff percents at this location
 *  by calling the PIXR based Hausdorff percent routine and ensuring the
 *  percents are the same.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pMatchData::     Ptr to internal information to use in this routine
 *  tx::             Hausdorff returned match X location value
 *  ty::             Hausdorff returned match Y location value
 *  percentAtoB::    A to B match percent at this location
 *  percentBtoA::    B to A match percent at this location
 *
 * Output:::
 *  pNextAsGood::    This routine will set this to point to 0 or 1.
 *                   1 to make all further routine invocations
 *                     report matches AT LEAST as good as the match
 *                     passed to this current invocation.
 *                   0 to keep same matching criterion as before
 *                     this invocation.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_checkMatchReturn (
    IPWState           *pIPWState,
    void               *pMatchData,
    Int32               tx,
    Int32               ty,
    Float64             percentAtoB,
    Float64             percentBtoA,
    Int8               *pNextAsGood)
{
    PROLOGUE;
    Float64   checkPercAtoB, checkPercBtoA;
    rtnChk *data = (rtnChk *)pMatchData;


    ERR_PASS( wp_pixrHausdorffPercent(pIPWState, data->AS, data->BD,
				      tx,  ty, data->nptsAS,
				      0, 0, &checkPercAtoB));
    ERR_PASS( wp_pixrHausdorffPercent(pIPWState, data->BS, data->AD,
				      -tx, -ty, data->nptsBS,
				      (UInt8)((data->searchMeth & USE_FULL_B_FOR_BTOA_PERCENT)
					      ? 0:1), 0, &checkPercBtoA));
    
    if (percentAtoB != checkPercAtoB)
	fprintf(stderr,
		"--Error x,y=(%ld,%ld)   AtoB%% was <%g> but check/slower AtoB%% was <%g>.\n",
		tx, ty, percentAtoB, checkPercAtoB);

    
    if (percentBtoA != checkPercBtoA)
	fprintf(stderr,
		"--Error x,y=(%ld,%ld)   BtoA%% was <%g> but check/slower BtoA%% was <%g>.\n",
		tx, ty, percentBtoA, checkPercBtoA);


	
#if DEBUG_OUT_FILES
    if (percentAtoB != checkPercAtoB)
    {
	/* Write out the relevant PIXRs for inspection to try to figure what
	 * went wrong */
	ERR_PASS( wp_pixrHausdorffPercent(data->AS, data->BD,  tx,  ty, 
					  data->nptsAS, 0, &checkPercAtoB, 1));
	fprintf(stderr, "  Wrote files 'model.overlap', 'imagery.overlap', and 'anded.overlap'\n");
    }
    
    if (percentBtoA != checkPercBtoA)
    {
	/* Write out the relevant PIXRs for inspection to try to figure what
	 * went wrong */
	ERR_PASS( wp_pixrHausdorffPercent(data->BS, data->AD, -tx, -ty, 
					  data->nptsBS,
					  (UInt8) ((data->searchMeth & USE_FULL_B_FOR_BTOA_PERCENT)
						   ? 0:1),
					  &checkPercBtoA, 1));
	fprintf(stderr, "  Wrote files 'model.overlap', 'imagery.overlap', and 'anded.overlap'\n");
    }
#endif /*DEBUG_OUT_FILES*/


    *pNextAsGood = 0;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Debugging/Sanity Hausdorff matching checker
 *
 * LongDescription:::
 *  This routine is mainly for debugging purposes.
 *  For every translation to try checking at, invoke both
 *  w_hausdorffMatches() and w_pixrHausdorff2WayMatches()
 *  and return the calculated pair of Hausdorff percents from A to B and
 *  from B to A.  Then, compare these two pairs and ensure that they are
 *  absolutely equal.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  AS::             ptr to source A binary pixr
 *  AM::             Ptr to A's model set
 *  AD::             ptr to A-dilated binary pixr, dimensions as AS
 *  BS::             ptr to source B binary pixr
 *  BM::             Ptr to B's model set
 *  BD::             ptr to B-dilated binary pixr, dimensions as BS
 *  BDTemp::         ptr to B temp binary pixr, dimensions as BD
 *                   (NULL to not use multi-res search)
 *  x::              Rectangular area in B where we search
 *  y::              for A to match.
 *  w::              Must be entirely inside B.
 *  h::              
 *  skipx::          Amounts to skip.  Check for a match    
 *  skipy::          every skipx pixel in X, every skipy pixel
 *                   in Y in subrect translation area.  (1 or greater)
 *  minPercentAtoB:: The match(es) must have A to B Hausdorff 
 *                   percent(s) at least the value this points to,
 *                   [0..1].
 *  minPercentBtoA:: The 2-Way match(es) must additionally
 *                   have B to A Hausdorff percent(s) at least
 *                   the value this points to, [0..1].
 *  pMultResInfo::   ptr to multi-res info (NULL to not use multi-res search)
 *  searchMeth::     Desired search method. One of:           
 *                   {SPARSE_FILTER, FLY_FILTER, EVERY_TRANS} 
 *                   May be OR-ed with USE_FULL_B_FOR_BTOA_PERCENT
 *
 * Returns:::
 *
 * Notes:::
 *  If skipx == skipy == 1::
 *    will try every translation in the subrect.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching, multi-resolution
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_checkHausdorff2WayMatches (
    IPWState           *pIPWState,
    PIXR               *AS,
    modelType          *AM,
    PIXR               *AD,
    PIXR               *BS,
    modelType          *BM,
    PIXR               *BD,
    PIXR               *BDTemp,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               skipx,
    Int32               skipy,
    Float64             minPercentAtoB,
    Float64             minPercentBtoA,
    void               *pMultResInfo,
    UInt8               searchMeth)
{
    PROLOGUE;
    void     *pMatchData;
    rtnChk   *data;
    Int32     matchReturnCode;

    ERR_GEN_PIXR(AS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR(AD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR(BS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR(BD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (BDTemp != NULL)
	ERR_GEN_PIXR(BDTemp, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN(AM == NULL  ||  BM == NULL, cErrNullPtr);

    ERR_GEN((pMultResInfo == NULL  &&  BDTemp != NULL) ||
	    (pMultResInfo != NULL  &&  BDTemp == NULL), cErrInvalidParm);

    ERR_GEN(skipx < 1 || skipy < 1, cErrInvalidParm);

    /* Fill the match data struct */
    data = (rtnChk *)IPW_MALLOC(sizeof(rtnChk), cAllocatorData);
    ERR_GEN(!data, cErrNomem);
    data->AS = AS;
    data->AD = AD;
    data->BS = BS;
    data->BD = BD;
    ERR_PASS_CLEANUP( w_numberBits(pIPWState, AS, &(data->nptsAS)));
    ERR_PASS_CLEANUP( w_numberBits(pIPWState, BS, &(data->nptsBS)));
    data->searchMeth = searchMeth;
    pMatchData = (void *)data;

    /* Run the normal 2Way Matching (with minimum percents at 0% and 0%
     * to return match percents for all translations.)
     * The match checking routine we pass in will then actually call
     * another way entirely for the same computation (slower) to ensure
     * the percents are the same at each translation, and thus that
     * the Hausdorff routines are working properly!
     */
    if (pMultResInfo == NULL)
    {
	ERR_PASS_CLEANUP( w_hausdorffMatches (pIPWState,
					      AM,
					      BD,
					      BM,
					      AD,
					      x, y, w, h,
					      skipx, skipy,
					      searchMeth,
					      pMatchData,
					      wp_checkMatchReturn,
					      &minPercentAtoB,
					      &minPercentBtoA,
					      &matchReturnCode ));
    }
    else
    {
	ERR_PASS_CLEANUP( w_hausdorffMultResMatches(pIPWState,
						    AM,
						    BD,
						    BDTemp,
						    BM,
						    AD,
						    x, y, w, h,
						    searchMeth,
						    pMultResInfo,
						    pMatchData,
						    wp_checkMatchReturn,
						    &minPercentAtoB,
						    &minPercentBtoA,
						    &matchReturnCode ));
    }



    IPW_FREE(data, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(data, cAllocatorData);
    ERR_RETURN;
}


/*

 $Log: hausdbug.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.30  1997/07/11 01:58:43  ddavies
 hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.29  1997/06/30 23:10:23  ddavies
 hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v

 Delete inclusion of Sandpiper's internal.h.

 Revision 1.28  1996/01/04 20:24:06  ejaquith
 hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.27  1995/12/13  04:09:23  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.26  1995/11/13  20:12:53  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.25  1995/11/10  02:04:09  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.24  1995/10/30  17:57:21  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Added inclusion of "hauscomp.h"
 * Updated function block comments who named other IPWrap routines.
 *
 * Revision 1.23  1995/10/13  22:02:06  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.22  1995/10/13  03:29:29  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Casted some params to shut up C++ compiler.
 *
 * Revision 1.21  1995/10/07  03:11:17  ddavies
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Add stdio.h since it's not in ipwcore.prv anymore.
 *
 * Revision 1.20  1995/09/29  00:35:02  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.19  1995/09/21  22:53:46  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/12  16:54:47  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.17  1995/09/08  00:14:43  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.16  1995/09/07  01:32:34  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.15  1995/09/06  20:20:34  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.14  1995/09/01  02:20:34  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.13  1995/08/31  21:09:25  ejaquith
 * ejaquith on Thu Aug 31 14:09:25 PDT 1995
 *
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.12  1995/08/25  01:36:46  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Fixed warnings
 *
 * Revision 1.11  1995/08/17  22:01:14  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.10  1995/07/24  22:20:03  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Updated Hausdorff ...Matches() routine calls to fit with new
 * parameters in library.  (Pointers to minimum Hausdorff percents).
 * These changes were needed for hausdorffReducedMatches().
 *
 * Revision 1.9  1995/04/22  21:52:51  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Updated all routines to use new interface ipwrap level Hausdorff routines.
 *
 * Revision 1.8  1995/03/29  02:03:40  ejaquith
 * ejaquith on Tue Mar 28 18:03:25 PST 1995
 *
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.7  1995/03/21  19:42:39  ejaquith
 * ejaquith on Tue Mar 21 11:42:26 PST 1995
 *
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Updated comments.
 *
 * Revision 1.6  1995/03/15  01:34:36  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Updated indentation.
 *
 * Revision 1.5  1995/01/07  21:57:05  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Made calculating where the imagery and model overlap a bit faster and
 * cleaner.
 *
 * Revision 1.4  1995/01/04  19:05:39  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Modified checkHausdorff2WayMatches() to allow checking for multi-resolution
 * searching (in addition to regular searching).
 *
 * Revision 1.3  1994/12/07  23:01:38  ejaquith
 * ejaquith on Wed Dec 7 15:01:12 PST 1994
 *
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Added skipx and skipy translation option to pixrHausdorff2WayMatches().
 *
 * Revision 1.2  1994/12/06  23:51:59  ejaquith
 * hausdbug.c -> /product/ipcore/ipwrap/src/RCS/hausdbug.c,v
 *
 * Updated debugging Hausdorff routines to conform to the newly built
 * Hausdorff modules.
 *

 */
