/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  	hausmres.c: 	Hausdorff binary image comparator using
 *  	    	    	multi-resolution searching wrappers
 *  	
 *  	    Code to construct/destruct multi-resolutional information
 *  	    		w_makeMultResInfo()
 *  	    		w_destroyMultResInfo()
 *
 *  	    Hausdorff multi-resolution matching routine
 *  	    		w_hausdorffMultResMatches()
 *
 *
 * A general description of the Hausdorff method and notes on
 * multi-resolution searching are included in haussrch.c
 */



#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "alplib.h"
#include "hauscomp.h"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */

/* prototypes */
#include "hauscomp.pub"
#include "props.pub"
#include "shrpatt.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: hausmres.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");





/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make the multi-resolution info
 *
 * LongDescription:::
 *  Given an array of Translation Area Sizes, make the multi-resolution info
 *  from it, for use in the multi-resolutional searching routines.
 *
 * PrivateDescription:::
 *  The size of the array determines the number of levels to search.
 *  Each level searches a finer resolution than the previous, and rules out
 *  areas equal to the array elements squared.  For example, if 8, 4, 2, 1 is
 *  used, later search routines remove 8x8 translation regions (from a
 *  subrect of possible translations) first, remove 4x4 regions next, remove
 *  2x2 regions lastly, and finally check all remaining subrect translations.
 *
 * Input:::
 *  pIPWState::       Ptr to IP struct for memory alloc/freeing routines
 *  pTransAreaSizes:: Ptr to array of integers specifying translation area
 *                    sizes to use for filtering.  See Notes.
 *
 * Output:::
 *  ppMultResInfo::   Ptr to ptr to internal representation of multi res info.
 *                    ppMultResInfo is allocated and filled in.
 *                    *ppMultResInfo must be NULL on input.
 *                    *ppMultResInfo will point to the finished multi res info.
 *
 * Returns:::
 *
 * Notes:::
 *  pTransAreaSizes array (of integers) must::
 *   contain at least 2 elements.
 *   contain elements in strictly decreasing order of multiples of "2".
 *   end with element "1".
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching, multi-resolution
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeMultResInfo (
    IPWState           *pIPWState,
    Int32              *pTransAreaSizes,
    void              **ppMultResInfo)
{
    PROLOGUE;
    UInt32  size;
    void   *pMultResInfo = NULL;

    ERR_GEN(pTransAreaSizes == NULL, cErrNullPtr);
    ERR_GEN(ppMultResInfo == NULL, cErrNullPtr);
    ERR_GEN(*ppMultResInfo != NULL, cErrPtrNotNull);

    ERR_PASS( i_getMultResInfoSize (pIPWState, pTransAreaSizes, &size ));

    pMultResInfo = (void *)IPW_MALLOC(size, cAllocatorData);
    ERR_GEN(!pMultResInfo, cErrNomem);

    ERR_PASS_CLEANUP( i_makeMultResInfo (pIPWState,
					 pTransAreaSizes, pMultResInfo ));

    *ppMultResInfo = pMultResInfo;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pMultResInfo, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the multi-resolution info memory and set ptr to NULL
 *
 * LongDescription:::
 *  Given the ptr to the internal representation of the multi-res info,
 *  free the allocated memory.  If already NULL, this routine
 *  does nothing; o.w. it will set ptr to NULL upon exiting.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppMultResInfo::  Ptr to ptr to multi-resolution info
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching, multi-resolution
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyMultResInfo (
    IPWState           *pIPWState,
    void              **ppMultResInfo)
{
    PROLOGUE;

    ERR_GEN(ppMultResInfo == NULL, cErrNullPtr);
    
    IPW_FREE(*ppMultResInfo, cAllocatorData);
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find all 1-Way or 2-Way matches of A and B
 *
 * LongDescription:::
 *  Given A's model set, place it over/at every location where A fits inside a
 *  passed in subrect of B's dilated pixr, and calculate the A to B Hausdorff
 *  percent.  If 2-Way, then take the corresponding B's model set over A's
 *  dilated pixr and calculate the B to A Hausdorff percent.
 *
 *  Report subrect location(s) with the Hausdorff percent(s)
 *  at least the passed in value(s), as matches of A and B.
 *
 * Input:::
 *	pIPWState::	Ptr to IP struct for memory alloc/freeing routines
 *	AM::		Ptr to A's model set
 *	BD::		Ptr to B-dilated binary pixr
 *	BDTemp::	Ptr to binary pixr of same dimens as BD
 *			image. Used internally for calculations
 *	BM::		Ptr to B's model set or NULL if 1-Way
 *	AD::		Ptr to A-dilated binary pixr if 2-Way.
 *			If NULL, indicates do only 1-Way.
 *	x::		Rectangular area in B where we search
 *	y::		for A to match.
 *	w::                                                         
 *	h::
 *	searchMeth::	0 or USE_FULL_B_FOR_BTOA_PERCENT
 *	pMultResInfo::	Ptr to the multi-resolution set-up info
 *	pCallbackData::	Ptr to client supplied data that we      
 *			include as an argument when calling the  
 *			client's callback proc.
 *	callbackProc::	Ptr to the client's callback routine,
 *			called with the match's translation x
 *			and y offset, the A to B match percent,
 *			the B to A match percent, (ignored if 1-Way)
 *			the client supplied argument, and a pointer to a
 *			value to shape further invocations (see Notes).
 *	pMinPercentAtoB::	The match(es) must have A to B Hausdorff 
 *				percent(s) at least the value this points to,
 *				[0..1].  See notes below.
 *	pMinPercentBtoA::	The 2-Way match(es) must additionally
 *				have B to A Hausdorff percent(s) at least
 *				the value this points to, [0..1].
 *				Ignored if 1-Way.  See notes below.
 *
 * Output:::
 *	pCallbackReturnCode::	Ptr to code returned by client's
 *				callback routine.
 *
 * Returns:::
 *
 * Notes:::
 *  Multi-resolution::
 *   This does multi-resolutional searching to rule out large translation
 *   areas that can't possibly contain a translation that will match.
 *  BDTemp::
 *   The image BDTemp points to is overwritten and its passed in bitmap
 *   information is not used.
 *  subrect::
 *   The subrect of B may extend outside of B.
 *  match locations::
 *   Match locations are in the frame of the full B image.
 *  callbackRoutine pNextAsGood value::
 *    Return 1 to make all further routine invocations report matches AT LEAST
 *      as good as the match in the current invocation.
 *    Return 0 to keep same matching criterion as before invocation.
 *  pMinPercentAtoB, pMinPercentBtoA::
 *    On completion, these will point to the min percents last used in the
 *    search.  (One can inspect if they had been increased by the callback).
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching, multi-resolution
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_hausdorffMultResMatches (
    IPWState           *pIPWState,
    modelType          *AM,
    PIXR               *BD,
    PIXR               *BDTemp,
    modelType          *BM,
    PIXR               *AD,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    UInt8               searchMeth,
    void               *pMultResInfo,
    void               *pCallbackData,
    ErrCode	(CDECL *callbackProc) (IPWState *pIPWState,
				       void *,
				       Int32 tx,
				       Int32 ty,
				       Float64 percentAtoB,
				       Float64 percentBtoA,
				       Int8 *pNextAsGood),
    Float64            *pMinPercentAtoB,
    Float64            *pMinPercentBtoA,
    ErrCode            *pCallbackReturnCode)
{
    PROLOGUE;
    Int8         nextAsGood;
    Int32        tx, ty;
    void        *pSearchInfo = NULL;
    UInt8        done;
    Float64      percentAtoB, percentBtoA;
    Float64      minPercentAtoB, minPercentBtoA;
    UInt8       *pImageAD, *pImageBD, *pImageBDT;
    Int32        fwAD, fwBD, fhAD, fhBD, bplAD, bplBD;
    MallocType	 ipAllocator;
    void	*allocInfo;

    ERR_GEN_PIXR(BD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN((BM == NULL  &&  AD != NULL) ||
	    (BM != NULL  &&  AD == NULL), cErrInvalidParm);
    ERR_GEN_PIXR(BDTemp, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (AD != NULL)
    {
	ERR_GEN_PIXR(AD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    }
	/* Check to make sure we can pass the IPWrap allocator down
	 * to the ipshared procedures.
	 */
    ERR_GEN((pIPWState == NULL) ||
	    (pIPWState->mallocVec[cAllocatorData] == NULL), cErrInvalidParm);
    ipAllocator = pIPWState->mallocVec[cAllocatorData];
    allocInfo = pIPWState->dataVec[cAllocatorData];

    ERR_GEN(AM == NULL  ||  callbackProc == NULL  ||
	    pMinPercentAtoB == NULL  ||  pMinPercentBtoA == NULL  ||
	    pCallbackReturnCode == NULL  ||  pMultResInfo == NULL,
	    cErrNullPtr);

    ERR_GEN(*pMinPercentAtoB < 0. || *pMinPercentAtoB > 1. ||
	    *pMinPercentBtoA < 0. || *pMinPercentBtoA > 1.,
	    cErrInvalidParm);

    minPercentAtoB = *pMinPercentAtoB;
    minPercentBtoA = *pMinPercentBtoA;


    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, BD, &pImageBD) );
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, BDTemp, &pImageBDT) );
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, BD, &fwBD) );
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, BD, &fhBD) );
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, BD, &bplBD) );

    if (AD)
    {
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, AD, &pImageAD) );
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, AD, &fwAD) );
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, AD, &fhAD) );
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, AD, &bplAD) );
    }
    else {
	pImageAD = NULL;
	fwAD = fhAD = bplAD = 0;
    }
    
    /* Multi-resolution filter out all the obvious nonmatching locations.
     */
    ERR_PASS( i_hausdorffMultResFilterSearchInit (pIPWState, AM->pIMS,
						  AM->ppBitmaps,
						  AM->nBitmaps,
						  pImageBD,
						  pImageBDT,
						  bplBD, fwBD, fhBD,
						  BM ? BM->pIMS : NULL,
						  pImageAD,
						  bplAD, fwAD, fhAD,
						  pMultResInfo,
						  x, y, w, h,
						  searchMeth,
						  minPercentAtoB, 
						  ipAllocator,
						  allocInfo,
						  &pSearchInfo));

    do
    {
	ERR_PASS_CLEANUP( i_hausdorffSearch (pIPWState, AM->pIMS,
					     AM->ppBitmaps,
					     AM->nBitmaps,
					     pImageBD,
					     BM ? BM->pIMS : NULL,
					     BM ? BM->ppBitmaps : NULL,
					     (UInt8) (BM ? BM->nBitmaps : 0),
					     pImageAD,
					     pSearchInfo,
					     minPercentAtoB, 
					     AD ? minPercentBtoA : -1.0F,
					     &tx, &ty,
					     &percentAtoB, &percentBtoA,
					     &done) );

	if (!done)
	{
	    ERR_CHECK(callbackProc(pIPWState, pCallbackData, tx, ty,
				   percentAtoB,
				   percentBtoA, &nextAsGood))
		{
		    *pCallbackReturnCode = __error__;
		    ERR_GEN_CLEANUP(1, cErrCallbackError);
		}
	    if (nextAsGood)
	    {
		minPercentAtoB = percentAtoB;
		minPercentBtoA = percentBtoA;
	    }
	}
    } while (!done);

    IPW_FREE(pSearchInfo, cAllocatorData);

    *pMinPercentAtoB = minPercentAtoB;
    
    /* We have to do this to ensure that value did not _decrease_ from how
     * it was passed in -- specifically, if we were doing 1way matching only,
     * we would always be getting -1. back in the above percentBtoA value
     * from the i_hausdorffSearch() calls!
     */
    *pMinPercentBtoA = MAX(*pMinPercentBtoA, minPercentBtoA);

    /* This to indicate no error found in callback */
    *pCallbackReturnCode = cSuccessful;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pSearchInfo, cAllocatorData);
    ERR_RETURN;
}


/*

 $Log: hausmres.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.26  1998/01/14 00:13:04  ddavies
 hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.25  1997/07/11 01:58:47  ddavies
 hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.24  1996/03/26 23:27:23  ddavies
 ddavies on Tue Mar 26 15:24:49 PST 1996

 hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v

 Change the type of the allocator passed to i_hausdorffMultResSearchInit to
 MallocType.  This is the type of allocator stored in IPWState.  Changing this
 type (and passing the IPWrap allocator and its data) means that we can
 use the client's memory management more consistently and free the
 table with IPW_FREE legally.  Also remove inclusion of reduce.prv since
 it was only needed for wp_makeScaleMalloc, which is no longer used.

 Revision 1.23  1996/01/04 20:24:08  ejaquith
 hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.22  1995/12/13  04:09:27  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.21  1995/11/16  02:17:39  ejaquith
 * ejaquith on Wed Nov 15 18:17:12 PST 1995
 *
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Using w_makeScaleMalloc() becomes wp_makeScaleMalloc().
 *
 * Revision 1.20  1995/11/13  20:12:56  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/11/10  02:04:12  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.18  1995/10/30  17:57:45  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Added inclusion of "hauscomp.h"
 *
 * Revision 1.17  1995/10/13  22:02:09  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.16  1995/10/13  03:28:59  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Casted some params to shut up C++ compiler.
 *
 * Revision 1.15  1995/09/29  00:34:46  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.14  1995/09/22  04:30:44  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * fixed logic in w_hausdorffMultResMatches () to work properly in
 * w_hausdorffReducedMatches().  Specifically, made the passed back B to A
 * percent be at least 0 (when 1way matching, it would return -1)
 *
 * Revision 1.13  1995/09/14  18:52:59  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Fixed checking of BM parameter in w_hausdorffMultResMatches().
 *
 * Revision 1.12  1995/09/12  16:54:37  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.11  1995/09/08  00:14:34  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.10  1995/09/06  20:20:36  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  02:20:26  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.8  1995/08/31  21:04:27  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/17  22:01:04  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.6  1995/07/24  22:14:28  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 *   Updated hausdorffMultResMatches() routine to
 * use _pointers_ to minimum Hausdorff percents, so it's possible to pass
 *  back the last used/tightened percents that were used in the subrect
 *  /list search.   This was needed for hausdorffReducedMatches() and will
 * provide the client more information about what happened during the
 * invocation.
 *
 * Revision 1.5  1995/07/18  02:21:55  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Added some soon-to-be-standard function header comments for testing
 * auto-doc module.
 *
 * Revision 1.4  1995/04/25  17:51:45  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Fixed (valid!) acc compiler warning of where I had (conditionally)
 *  passed in NULL to an integer param of i_hausdorffSearch() call.
 *
 * Revision 1.3  1995/04/22  21:52:46  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Removed getMultResInfoSize().
 * Updated hausdorffMultResMatches() to use search method interface
 * instead of Hausdorff bit flags interface.  Internally updated to call
 * ipshared multi-resolution filter routine and then use new ipshared
 * iterative interface of i_hausdorffSearch().
 * Routine now requires that the already created model set be
 * passed in compared to the previous interface which allowed either the
 * model set or source pixr be passed in.
 * Added all modifications to comply with usage of new ipshared level as
 * necessary.
 *
 * Revision 1.2  1995/03/29  02:04:02  ejaquith
 * hausmres.c -> /product/ipcore/ipwrap/src/RCS/hausmres.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.1  1995/03/21  19:45:39  ejaquith
 * Initial revision
 *
 
 */
