/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/



/*
 *  	hausrduc.c: 	Hausdorff binary image comparator using
 *  	    	    	image reduced searching wrappers
 *  	
 *  	    Code to construct/destruct reduction information
 *  	    		w_makeReducedInfoA()
 *  	    		w_makeReducedInfoB()
 *  	    		w_destroyReducedInfoA()
 *  	    		w_destroyReducedInfoB()
 *
 *  	    Hausdorff reduced matching routine
 *  	    		w_hausdorffReducedMatches()
 *
 * A general description of the Hausdorff method is included in haussrch.c
 * Notes on this specific image reduction searching method are included below.
 */



#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "alplib.h"
#include "shrscgr.pub"
#include "hauscomp.h"

/* prototypes */
#include "hauscomp.pub"
#include "props.pub"
#include "rast.pub"
#include "shrpatt.pub"
#include "reduce.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: hausrduc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");


/*____________________________________________________________________________
 *
 *
 * This code is a simple extension of the basic Hausdorff search.
 * It reduces both the A model's bitmap and the B image bitmap (by some
 * client specified amount) and then does a normal client-specified search.
 * (Where the client may pick any of the basic Hausdorff search styles.)
 * When a match is found in compared reduced A to reduced B, the region
 * correponding to that area in the full resolution bitmaps is used in a
 * full resolution A and B search, whereupon any matches found here are
 * returned via client callback as the true actual Hausdorff matches.
 *
 * Note, however, that due to the nature of the reduction of the images, this
 * search method is not guarranteed to find _all_ Hausdorff matches as the
 * other search methods are.
 *
 *____________________________________________________________________________
 */




    /* A reduction info */
typedef struct redTypeA {
    /* Reduced stuff */
    modelType  *redAM;
    Int32       reductionFactor;

    /* Full resolution matching stuff */
    modelType  *AM;
    Int32       widthAM;
    Int32       heightAM;
    
    Int8        destroyRedAM; /* Destroy redAM upon freeing? */
    Int8        destroyAM;    /* Destroy AM upon freeing? */
} redTypeA;

    /* B reduction info */
typedef struct redTypeB {
    /* Reduced stuff */
    PIXR    *redBD;
    PIXR    *redBDTemp;
    Int32    reductionFactor;

    /* Full resolution matching stuff */
    modelType  *AM;
    PIXR       *AD;
    modelType  *BM;
    PIXR       *BD;
    Int32       widthAM;
    Int32       heightAM;
    Float64     minAtoB;
    Float64     minBtoA;
    void       *pCallbackData;
    ErrCode (CDECL *callbackProc)( IPWState *pIPWState,
				   void *, Int32 tx, Int32 ty,
				   Float64 percentAtoB, Float64 percentBtoA,
				   Int8 *pNextAsGood);
    ErrCode     callbackReturnCode;

    Int8        destroyRedBD; /* Destroy redBD upon freeing? */
    Int8        first;
} redTypeB;


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Turn on an (x,y) pixel in a binary PIXR.
 *
 * LongDescription:::
 *  Caller must know binary PIXR exists, and x,y is in bounds.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prs::            Ptr to pixr to set pixel in
 *  x::              x-coordinate in prs of pel to set/read
 *  y::              y-coordinate in prs of pel to set/read
 *
 * Output:::
 *  pWasClear::      Set to 0 on return if pixel was _already_ set; 1 o.w.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_mySetPixel (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               x,
    Int32               y,
    Int8               *pWasClear)
{
    PROLOGUE;
    Int32    wpl;
    UInt8   *data;
    UInt32  *ldata;

    /* prs has already been verified -- this routine is private */
    
    ERR_PASS( w_pixrGetImage(pIPWState, prs, &data));
    ERR_PASS( w_pixrGetBpl(pIPWState, prs, &wpl));

    ldata = (UInt32 *)data;
    wpl >>= 2;

    if (ldata[wpl*y + (x >> 5)] & ((UInt32)0x80000000 >> (x & 31))) {
	*pWasClear = 0;
    }
    else {
	ldata[wpl*y + (x >> 5)] |= ((UInt32)0x80000000 >> (x & 31));
	*pWasClear = 1;
    }
    
    NORMAL_RETURN;
}

    
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a reduction information for A
 *
 * Input:::
 *  pIPWState::       Ptr to IP struct for memory alloc/freeing routines
 *  AS::              Ptr to source A binary pixr
 *  nBitmapsA::       0,1,or 32 (see w_makeModelSet() for details)
 *  AM::              Ptr to A's model set.  If NULL, we make it
 *                    internally (and destroy it upon cleanup)
 *  reductionFactor:: How much to reduce AS: 2,4,8, or 16
 *  redAS::           Ptr to AS-dilated reduced binary pixr.  If NULL,
 *                    this means client wants us to reduce AS in this routine
 *                    by the reductionFactor.  If non-NULL, this means
 *                    client has already reduced AS by reductionFactor, and
 *                    no further reduction is needed.
 *  nBitmapsRedA::    0,1,or 32 (see w_makeModelSet() for details)
 *  redAM::           Ptr to reduced A's model set.  If NULL, we
 *                    make it internally from redAS (and destroy
 *                    it upon cleanup)
 *
 * Output:::
 *  ppReducedInfoA::  Ptr to ptr to reduced A information.  Info will get
 *                    allocated and filled in this routine and passed back.
 *                    *ppReducedInfoA must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Reduction::
 *   For internal reduction, for reduction .. we simply call ..
 *    2    w_reduce(AS, 1, 0, 0, 0);
 *    4    w_reduce(AS, 1, 1, 0, 0);
 *    8    w_reduce(AS, 1, 1, 1, 0);
 *   16    w_reduce(AS, 1, 1, 1, 1);
 *  redAS::
 *   If NULL, we will destroy the reduced AS image later upon cleanup.
 *   If non-NULL, client is responsible for destroying it after cleanup.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeReducedInfoA (
    IPWState           *pIPWState,
    PIXR               *AS,
    UInt8               nBitmapsA,
    modelType          *AM,
    Int32               reductionFactor,
    PIXR               *redAS,
    UInt8               nBitmapsRedA,
    modelType          *redAM,
    void              **ppReducedInfoA)
{
    PROLOGUE;
    redTypeA  *pRedInfoA = NULL;
    Int8       destroyRedAS = 0;
    Int32      h;

    ERR_GEN(ppReducedInfoA == NULL, cErrNullPtr);
    ERR_GEN(*ppReducedInfoA != NULL, cErrPtrNotNull);
    ERR_GEN_PIXR(AS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (redAS != NULL)
	ERR_GEN_PIXR(redAS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
	

    pRedInfoA = (redTypeA *)IPW_MALLOC(sizeof(redTypeA), cAllocatorData);
    ERR_GEN(!pRedInfoA, cErrNomem);
    pRedInfoA->AM = NULL;
    pRedInfoA->redAM = NULL;
    
    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, AS, &(pRedInfoA->widthAM)));
    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, AS, &(pRedInfoA->heightAM)));

    if (AM == NULL)
    {
	/* Client wants us to make AM from AS */
	pRedInfoA->destroyAM = 1;

	ERR_PASS_CLEANUP( w_makeModelSet( pIPWState, AS,
				       pRedInfoA->heightAM, nBitmapsA,
				       &AM));
    }
    else
	pRedInfoA->destroyAM = 0;

    pRedInfoA->AM = AM;
	

    if (redAM == NULL && redAS == NULL)
    {
	/* This means that the client wants us to reduce the AS PIXR for
	 * hir.
	 */
	destroyRedAS = 1;

	switch (reductionFactor)
	{
	case 2:
	    ERR_PASS_CLEANUP( w_reduce(pIPWState, AS, 1, 0, 0, 0, &redAS));
	    break;
	case 4:
	    ERR_PASS_CLEANUP( w_reduce(pIPWState, AS, 1, 1, 0, 0, &redAS));
	    break;
	case 8:
	    ERR_PASS_CLEANUP( w_reduce(pIPWState, AS, 1, 1, 1, 0, &redAS));
	    break;
	case 16:
	    ERR_PASS_CLEANUP( w_reduce(pIPWState, AS, 1, 1, 1, 1, &redAS));
	    break;
	default:
	    ERR_GEN(1, cErrInvalidParm);
	}
    }

    if (redAM == NULL)
    {
	/* Client wants us to make redAM from redAS */
	pRedInfoA->destroyRedAM = 1;

	ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, redAS, &h));
	ERR_PASS_CLEANUP( w_makeModelSet( pIPWState, redAS, h,
				       nBitmapsRedA, &redAM ));
    }
    else
	pRedInfoA->destroyRedAM = 0;

    if (destroyRedAS)
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &redAS));
    
    pRedInfoA->redAM = redAM;
    pRedInfoA->reductionFactor = reductionFactor;
    
    *ppReducedInfoA = (void *)pRedInfoA;
    NORMAL_RETURN;

    
ERR_CLEANUP:
    if (pRedInfoA != NULL && pRedInfoA->destroyAM)
	w_destroyModelSet(pIPWState, &pRedInfoA->AM);
    if (pRedInfoA != NULL && pRedInfoA->destroyRedAM)
	w_destroyModelSet(pIPWState, &pRedInfoA->redAM);
    if (destroyRedAS)
	w_destroyPixr(pIPWState, &redAS);
    IPW_FREE(pRedInfoA, cAllocatorData);
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a reduction information for B
 *
 * Input:::
 *  pIPWState::       Ptr to IP struct for memory alloc/freeing routines
 *  BD::              Ptr to B-dilated binary pixr (from BS)
 *  reductionFactor:: Factor to reduce BD by: 2,4,8, or 16.
 *  redBD::           Ptr to B-dilated reduced binary pixr.  If NULL,
 *                    this means client wants us to reduce BD in this routine
 *                    by the reductionFactor.  If non-NULL, this means
 *                    client has already reduced BD by reductionFactor, and
 *                    no further reduction is needed.
 *
 * Output:::
 *  ppReducedInfoB::  Ptr to ptr to reduced B information.  Info will get
 *                    allocated and filled in this routine and passed back.
 *                    *ppReducedInfoB must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Reduction::
 *   For internal reduction, for reduction .. we simply call ..
 *    2    w_reduce(BD, 1, 0, 0, 0);
 *    4    w_reduce(BD, 1, 1, 0, 0);
 *    8    w_reduce(BD, 1, 1, 1, 0);
 *   16    w_reduce(BD, 1, 1, 1, 1);
 *  redBD::
 *   If NULL, we will destroy the reduced BD image later upon cleanup.
 *   If non-NULL, client is responsible for destroying it after cleanup.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeReducedInfoB (
    IPWState           *pIPWState,
    PIXR               *BD,
    Int32               reductionFactor,
    PIXR               *redBD,
    void              **ppReducedInfoB)
{
    PROLOGUE;
    redTypeB  *pRedInfoB = NULL;

    ERR_GEN(ppReducedInfoB == NULL, cErrNullPtr);
    ERR_GEN(*ppReducedInfoB != NULL, cErrPtrNotNull);
    ERR_GEN_PIXR(BD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (redBD != NULL)
	ERR_GEN_PIXR(redBD, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    pRedInfoB = (redTypeB *)IPW_MALLOC(sizeof(redTypeB),
				       cAllocatorData);
    ERR_GEN(!pRedInfoB, cErrNomem);
    pRedInfoB->redBD = NULL;
    pRedInfoB->redBDTemp = NULL;
    
    if (redBD == NULL)
    {
	/* This means that the client wants us to reduce the BD PIXR for
	 * hir.
	 */
	pRedInfoB->destroyRedBD = 1;

	switch (reductionFactor)
	{
	case 2:
	    ERR_PASS_CLEANUP( w_reduce(pIPWState, BD, 1, 0, 0, 0, &redBD));
	    break;
	case 4:
	    ERR_PASS_CLEANUP( w_reduce(pIPWState, BD, 1, 1, 0, 0, &redBD));
	    break;
	case 8:
	    ERR_PASS_CLEANUP( w_reduce(pIPWState, BD, 1, 1, 1, 0, &redBD));
	    break;
	case 16:
	    ERR_PASS_CLEANUP( w_reduce(pIPWState, BD, 1, 1, 1, 1, &redBD));
	    break;
	default:
	    ERR_GEN(1, cErrInvalidParm);
	}
    }
    else
	pRedInfoB->destroyRedBD = 0;

    pRedInfoB->BD = BD;
    pRedInfoB->redBD = redBD;
    pRedInfoB->reductionFactor = reductionFactor;

    /* This gets used to keep track of the translation areas we have
     * searched at the full resolution (and also by the multi-resolution
     * searching if it is used).
     */
    ERR_PASS_CLEANUP( w_clonePixr(pIPWState, pRedInfoB->redBD,
				  &pRedInfoB->redBDTemp));
    
    /* Fill in later if want to use 2-Way searching */
    pRedInfoB->BM = NULL;
    
    *ppReducedInfoB = (void *)pRedInfoB;
    NORMAL_RETURN;
    

ERR_CLEANUP:
    if (pRedInfoB != NULL && pRedInfoB->destroyRedBD)
	w_destroyPixr(pIPWState, &pRedInfoB->redBD);
    if (pRedInfoB != NULL)
	w_destroyPixr(pIPWState, &pRedInfoB->redBDTemp);
    IPW_FREE(pRedInfoB, cAllocatorData);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroy A's reduction information, and set ptr to NULL.
 *
 * Input:::
 *  pIPWState::       Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppReducedInfoA::  Ptr to ptr to A's Hausdorff reduction info.
 *                    Will be freed if non-NULL and set to NULL on return.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyReducedInfoA (
    IPWState           *pIPWState,
    void              **ppReducedInfoA)
{
    PROLOGUE;

    ERR_GEN(ppReducedInfoA == NULL, cErrNullPtr);

    if (*ppReducedInfoA != NULL)
    {
	redTypeA  **ppRedInfoA = (redTypeA **)ppReducedInfoA;

	if ((*ppRedInfoA)->destroyRedAM)
	    ERR_PASS( w_destroyModelSet(pIPWState, &((*ppRedInfoA)->redAM)));
	
	if ((*ppRedInfoA)->destroyAM)
	    ERR_PASS( w_destroyModelSet(pIPWState, &((*ppRedInfoA)->AM)));

	IPW_FREE(*ppRedInfoA, cAllocatorData);
	
	*ppReducedInfoA = NULL;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroy B's reduction information, and set ptr to NULL
 *
 * Input:::
 *  pIPWState::       Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppReducedInfoB::  Ptr to ptr to B's Hausdorff reduction info.
 *                    Will be freed if non-NULL and set to NULL on return.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyReducedInfoB (
    IPWState           *pIPWState,
    void              **ppReducedInfoB)
{
    PROLOGUE;

    ERR_GEN(ppReducedInfoB == NULL, cErrNullPtr);

    if (*ppReducedInfoB != NULL)
    {
	redTypeB  **ppRedInfoB = (redTypeB **)ppReducedInfoB;

	if ((*ppRedInfoB)->destroyRedBD)
	    ERR_PASS( w_destroyPixr(pIPWState, &((*ppRedInfoB)->redBD)));

	ERR_PASS( w_destroyPixr(pIPWState, &((*ppRedInfoB)->redBDTemp)));

	IPW_FREE(*ppRedInfoB, cAllocatorData);

	*ppReducedInfoB = NULL;
    }
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Fill in the remaining B reduction info fields
 *
 * Input:::
 *  pIPWState::           Ptr to IP struct for memory alloc/freeing routines
 *  pRedInfoA::           Ptr to A's Hausdorff reduction info.
 *  AD::                  Ptr to A-dilated binary pixr (from AS) if 2-Way.
 *                        If NULL, indicates do only 1-Way.
 *  BM::                  Ptr to B's model set (from BS)
 *                        or NULL if 1-Way searching.
 *  minPercentAtoB::      The match(es) must have A to B Hausdorff 
 *                        percent(s) at least this value, [0..1].
 *  minPercentBtoA::      The 2-Way match(es) must additionally
 *                        have B to A Hausdorff percent(s) at least
 *                        this value, [0..1].  Ignored if 1-Way.
 *  pCallbackData::       Ptr to client supplied data that we
 *                        include as an argument when calling the
 *                        client's callback proc.
 *  callbackProc::        Ptr to the client's callback routine,
 *                        called with the match's translation x
 *                        and y offset, the A to B match percent,
 *                        the B to A match percent, (ignored if 1-Way)
 *                        the client supplied argument, and a pointer to a
 *                        value to shape further invocations (see Notes).
 *
 * Output:::
 *  pRedInfoB::           Ptr to B's Hausdorff reduction info.
 *                        Its remaining fields will be filled in here
 *                        from the other arguments.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_fillInReductionInfo (
    IPWState           *pIPWState,
    redTypeA           *pRedInfoA,
    redTypeB           *pRedInfoB,
    PIXR               *AD,
    modelType          *BM,
    Float64             minPercentAtoB,
    Float64             minPercentBtoA,
    void               *pCallbackData,
    ErrCode	(CDECL *callbackProc)( IPWState *pIPWState,
				       void *,
				       Int32 tx,
				       Int32 ty,
				       Float64 percentAtoB,
				       Float64 percentBtoA,
				       Int8 *pNextAsGood))
{
    PROLOGUE;

    /* AD has already been verified -- this is a private routine */

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    if (AD && BM)
    {
	/* Means we want 2-Way searching */
	pRedInfoB->AD = AD;
	pRedInfoB->BM = BM;
    }
    else
    {
	pRedInfoB->AD = NULL;
	pRedInfoB->BM = NULL;
    }

    /* We will need all of these fields whenever we find a match in the
     * reduced resolution image to then expand to check for a match in the
     * corresponding area of the expanded full-resolution image */
    pRedInfoB->pCallbackData = pCallbackData;
    pRedInfoB->callbackProc = callbackProc;
    pRedInfoB->callbackReturnCode = cSuccessful;
    pRedInfoB->minAtoB = minPercentAtoB;
    pRedInfoB->minBtoA = minPercentBtoA;
    pRedInfoB->first = 1;

    /* Transfer appropriate A reduced info that will need later to the
     * B reduced info */
    pRedInfoB->AM = pRedInfoA->AM;
    pRedInfoB->widthAM = pRedInfoA->widthAM;
    pRedInfoB->heightAM = pRedInfoA->heightAM;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reduced A and B images match at this location -- now try full resolution
 *
 * PrivateDescription:::
 *  We have found a match from reduced A over reduced B at this location.
 *  Now we need to search the corresponding space at the full resolution
 *  for matches that meet our current full A to full B criterion.
 *  If any are found at the full resolution, then we'll call the ACTUAL
 *  client's callback routine because these will be ACTUAL match candidates.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  data::           Ptr to the client's supplied argument to be used in
 *                   this callback routine
 *  tx::             Match's X translation location of A in B
 *  ty::             Match's Y translation location of A in B
 *  percentAtoB::    Match's A to B Hausdorff match percent
 *  percentBtoA::    Match's B to A Hausdorff match percent
 *
 * Output:::
 *  pNextAsGood::    Will point to 0 on return to indicate to NOT tighten
 *                   up the matching criterion (via Hausdorff percents)
 *                   for any ensuing invocations of this routine
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching, multi-resolution
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_foundReducedMatch (
    IPWState           *pIPWState,
    void               *data,
    Int32               tx,
    Int32               ty,
    Float64             percentAtoB,
    Float64             percentBtoA,
    Int8               *pNextAsGood)
{
    PROLOGUE;
    redTypeB   *pRedInfoB = (redTypeB *)data;
    UInt8       check;
    Int8        wasClear;

    /* Keep silly compilers hapy */
    percentAtoB = percentAtoB;
    percentBtoA = percentBtoA;

    if (pRedInfoB->first == 1)
    {
	Int32 w,h;
	/* This means this is the first invocation of this function.
	 * If multi-resolution searching, we know that this function can
	 * only be called once we are at the lowest level, and therefore
	 * pRedInfoB->redBDTemp will no longer be needed there.
	 * Thus, wipe it out to all zeros so we may use it to tally
	 * the translation areas we have searched at the fullest
	 * resolution!
	 *
	 * Using this PIXR as a tally as such will allow two tres important
	 * things:
	 *  -we will only try matches as the full resolution once and once
	 *   only
	 *  -we won't return a match location (w.r.t. full resolution)
	 *   to client's callback routine multiple times
	 */
	pRedInfoB->first = 0;

	/* Clear the PIXR */
	ERR_PASS( w_pixrGetWidth(pIPWState, pRedInfoB->redBDTemp, &w));
	ERR_PASS( w_pixrGetHeight(pIPWState, pRedInfoB->redBDTemp, &h));
	ERR_PASS( w_rasterop (pIPWState,
			      0, 0, NULL,
			      PIX_CLR, 
			      0, 0, w, h,
			      pRedInfoB->redBDTemp));
    }
    

    /* Notice we can be sloppy about expanding back up to the full resolution
     * because we use the SEARCH_EVERY_TRANS search method which allows for
     * A to be placed only partially overlapping B (the box-filtering search
     * methods constrain A to lie wholly inside B!)
     */

    /* We know (tx,ty) is in bounds of pRedBD image. */

    /* Set bit 0 if redBDTemp(x,y) is "0"
     * Set bit 1 if redBDTemp(x-1,y) is valid and is "0"
     * Set bit 2 if redBDTemp(x-1,y-1) is valid and is "0"
     * Set bit 3 if redBDTemp(x,y-1) is valid and is "0"
     * Set all the redBDTemp() bits, too, as mySetPixel() checks they are
     * "0" before setting them to "1".
     */
    ERR_PASS( wp_mySetPixel(pIPWState, pRedInfoB->redBDTemp,
			    tx, ty, &wasClear));
    check = wasClear ? 0x1 : 0x0;
    tx--;

    if (tx >= 0) {
	ERR_PASS( wp_mySetPixel(pIPWState, pRedInfoB->redBDTemp,
				tx, ty, &wasClear));
	check |= wasClear ? 0x2 : 0x0;
    }
    ty--;

    if (tx >= 0 && ty >= 0) {
	ERR_PASS( wp_mySetPixel(pIPWState, pRedInfoB->redBDTemp,
				tx, ty, &wasClear));
	check |= wasClear ? 0x4 : 0x0;
    }
    tx++;

    if (ty >= 0) {
	ERR_PASS( wp_mySetPixel(pIPWState, pRedInfoB->redBDTemp,
				tx, ty, &wasClear));
	check |= wasClear ? 0x8 : 0x0;
    }
    ty++;

    if (check == 0xF)
    {
	/* Means check all four areas
	 * The four areas are the full resolution areas that correspond
	 * to the expanded locations in reduced BD:
	 * (tx,ty), (tx-1,ty), (tx,ty-1), and (tx-1,ty-1)
	 */
	ERR_PASS( w_hausdorffMatches (pIPWState, pRedInfoB->AM, pRedInfoB->BD,
				      pRedInfoB->BM, pRedInfoB->AD,
				      (tx-1) * pRedInfoB->reductionFactor,
				      (ty-1) * pRedInfoB->reductionFactor,
				      pRedInfoB->widthAM +
				      pRedInfoB->reductionFactor * 2 - 1,
				      pRedInfoB->heightAM +
				      pRedInfoB->reductionFactor * 2 - 1,
				      1, 1,
				      EVERY_TRANS,
				      pRedInfoB->pCallbackData,
				      pRedInfoB->callbackProc,
				      &(pRedInfoB->minAtoB),
				      &(pRedInfoB->minBtoA),
				      &(pRedInfoB->callbackReturnCode)));
    }
    else
    {
	if (check & 0x1)
	{
	    /* Check the (tx,ty) area */
	    ERR_PASS( w_hausdorffMatches (pIPWState, 
					  pRedInfoB->AM, pRedInfoB->BD,
					  pRedInfoB->BM, pRedInfoB->AD,
					  tx * pRedInfoB->reductionFactor,
					  ty * pRedInfoB->reductionFactor,
					  pRedInfoB->widthAM +
					  pRedInfoB->reductionFactor - 1,
					  pRedInfoB->heightAM +
					  pRedInfoB->reductionFactor - 1,
					  1, 1,
					  EVERY_TRANS,
					  pRedInfoB->pCallbackData,
					  pRedInfoB->callbackProc,
					  &(pRedInfoB->minAtoB),
					  &(pRedInfoB->minBtoA),
					  &(pRedInfoB->callbackReturnCode)));
	}
	if (check & 0x2)
	{
	    /* Check the (tx-1,ty) area */
	    ERR_PASS( w_hausdorffMatches (pIPWState,
					  pRedInfoB->AM, pRedInfoB->BD,
					  pRedInfoB->BM, pRedInfoB->AD,
					  (tx-1) * pRedInfoB->reductionFactor,
					  ty * pRedInfoB->reductionFactor,
					  pRedInfoB->widthAM +
					  pRedInfoB->reductionFactor - 1,
					  pRedInfoB->heightAM +
					  pRedInfoB->reductionFactor - 1,
					  1, 1,
					  EVERY_TRANS,
					  pRedInfoB->pCallbackData,
					  pRedInfoB->callbackProc,
					  &(pRedInfoB->minAtoB),
					  &(pRedInfoB->minBtoA),
					  &(pRedInfoB->callbackReturnCode)));
	}
	if (check & 0x4)
	{
	    /* Check the (tx-1,ty-1) area */
	    ERR_PASS( w_hausdorffMatches (pIPWState,
					  pRedInfoB->AM, pRedInfoB->BD,
					  pRedInfoB->BM, pRedInfoB->AD,
					  (tx-1) * pRedInfoB->reductionFactor,
					  (ty-1) * pRedInfoB->reductionFactor,
					  pRedInfoB->widthAM +
					  pRedInfoB->reductionFactor - 1,
					  pRedInfoB->heightAM +
					  pRedInfoB->reductionFactor - 1,
					  1, 1,
					  EVERY_TRANS,
					  pRedInfoB->pCallbackData,
					  pRedInfoB->callbackProc,
					  &(pRedInfoB->minAtoB),
					  &(pRedInfoB->minBtoA),
					  &(pRedInfoB->callbackReturnCode)));
	}
	if (check & 0x8)
	{
	    /* Check the (tx,ty-1) area */
	    ERR_PASS( w_hausdorffMatches (pIPWState,
					  pRedInfoB->AM, pRedInfoB->BD,
					  pRedInfoB->BM, pRedInfoB->AD,
					  tx * pRedInfoB->reductionFactor,
					  (ty-1) * pRedInfoB->reductionFactor,
					  pRedInfoB->widthAM +
					  pRedInfoB->reductionFactor - 1,
					  pRedInfoB->heightAM +
					  pRedInfoB->reductionFactor - 1,
					  1, 1,
					  EVERY_TRANS,
					  pRedInfoB->pCallbackData,
					  pRedInfoB->callbackProc,
					  &(pRedInfoB->minAtoB),
					  &(pRedInfoB->minBtoA),
					  &(pRedInfoB->callbackReturnCode)));
	}
    }

    /* Always return 0 -- we don't want to tighten matching criterion! */
    *pNextAsGood = 0;
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find (most) 1-Way or 2-Way matches of A and B
 *
 * LongDescription:::
 *  Report full resolution subrect location(s) with the Hausdorff percent(s)
 *  at least the passed in value(s), as matches of A and B.
 *
 * PrivateDescription:::
 *  Place the reduced A's model set over/at every location where reduced A fits
 *  inside a passed in subrect of reduced B's dilated pixr,
 *  and calculate the A to B Hausdorff percent.  If 2-Way, then take the
 *  corresponding reduced B's model set over reduced A's dilated pixr
 *  and calculate the B to A Hausdorff percent.
 *  At subrect location(s) with the Hausdorff percent(s)
 *  at least the passed in value(s), search the corresponding area in the
 *  full resolution A and B in a simlar way.
 *
 * Input:::
 *  pIPWState::           Ptr to IP struct for memory alloc/freeing routines
 *  AD::                  Ptr to A-dilated binary pixr (from AS) if 2-Way.
 *                        If NULL, indicates do only 1-Way.
 *  BM::                  Ptr to B's model set (from BS)
 *                        or NULL if 1-Way searching.
 *  pReducedInfoA::       Ptr to A's Hausdorff reduction info.
 *  pReducedInfoB::       Ptr to B's Hausdorff reduction info.
 *  x::                   Rectangular area in B where we search
 *  y::                   for A to match.
 *  w::                   Must be entirely inside B.
 *  h::
 *  minReducedAtoB::      The reduced A and B must have an A to B
 *                        Hausdorff percent at least this value,
 *                        [0..1], to invoke searching in the
 *                        corresp. area at the full resolution.
 *  searchMeth::          Desired search method. One of:           
 *                        {SPARSE_FILTER, FLY_FILTER, EVERY_TRANS} 
 *                        May be OR-ed with USE_FULL_B_FOR_BTOA_PERCENT
 *  pMultResInfo::        Ptr to the multi-resolution set-up info
 *                        if multi-resolution searching; o.w. NULL.
 *  pCallbackData::       Ptr to client supplied data that we
 *                        include as an argument when calling the
 *                        client's callback proc.
 *  callbackProc::        Ptr to the client's callback routine,
 *                        called with the match's translation x
 *                        and y offset, the A to B match percent,
 *                        the B to A match percent, (ignored if 1-Way)
 *                        the client supplied argument, and a pointer to a
 *                        value to shape further invocations (see Notes).
 *  pMinPercentAtoB::     The match(es) must have A to B Hausdorff 
 *                        percent(s) at least the value this points to,
 *                        [0..1].  See notes below.
 *  pMinPercentBtoA::     The 2-Way match(es) must additionally
 *                        have B to A Hausdorff percent(s) at least
 *                        the value this points to, [0..1].
 *                        Ignored if 1-Way.  See notes below.
 *
 * Output:::
 *  pCallbackReturnCode:: Ptr to code returned by client's
 *                        callback routine
 *
 * Returns:::
 *
 * Notes:::
 *  Multi-resolution::
 *    This can do multi-resolutional searching to rule out large translation
 *    areas that can't possibly contain a translation that will match.
 *  Match locations::
 *    are in the frame of the full B image.
 *  callbackRoutine pNextAsGood value::
 *    Return 1 to make all further routine invocations report matches AT LEAST
 *      as good as the match in the current invocation.
 *    Return 0 to keep same matching criterion as before invocation.
 *  pMinPercentAtoB, pMinPercentBtoA::
 *    On completion, these will point to the min percents last used in the
 *    search.  (One can inspect if they had been increased by the callback).
 *    NOTICE: These percents are the ones used at the _full_ resolution.
 *  minReducedAtoB::
 *    This value will never get increased, regardless if the above ones
 *    do at the full resolution.  Think of it as a "trigger" percent, where
 *    we will always invoke full resolution searching when regions in the
 *    reduced world this good are found.
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, multi-resolution, template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_hausdorffReducedMatches (
    IPWState           *pIPWState,
    PIXR               *AD,
    modelType          *BM,
    void               *pReducedInfoA,
    void               *pReducedInfoB,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Float64             minReducedAtoB,
    UInt8               searchMeth,
    void               *pMultResInfo,
    void               *pCallbackData,
    ErrCode	(CDECL *callbackProc)( IPWState *pIPWState,
				       void *,
				       Int32 tx,
				       Int32 ty,
				       Float64 percentAtoB,
				       Float64 percentBtoA,
				       Int8 *pNextAsGood),
    Float64            *pMinPercentAtoB,
    Float64            *pMinPercentBtoA,
    ErrCode            *pCallbackReturnCode)
{
    PROLOGUE;
    redTypeA    *pRedInfoA = (redTypeA *)pReducedInfoA;
    redTypeB    *pRedInfoB = (redTypeB *)pReducedInfoB;
    Int32        wRedBD, hRedBD;
    /* This is the error code returned by wp_foundReducedMatch() */
    ErrCode      FRMcallbackRetCode;
    
    Float64      tmp;

    ERR_GEN((BM == NULL  &&  AD != NULL) ||
	    (BM != NULL  &&  AD == NULL), cErrInvalidParm);
    ERR_GEN(pReducedInfoA == NULL  ||  pReducedInfoB == NULL  ||
	    pMinPercentAtoB == NULL  ||  pMinPercentBtoA == NULL  ||
	    pCallbackReturnCode == NULL  ||  callbackProc == NULL,
	    cErrNullPtr);
    ERR_GEN(pRedInfoA->reductionFactor != pRedInfoB->reductionFactor,
	    cErrInvalidParm);
    ERR_GEN(*pMinPercentAtoB < 0. || *pMinPercentAtoB > 1. ||
	    *pMinPercentBtoA < 0. || *pMinPercentBtoA > 1. ||
	    minReducedAtoB < 0. || minReducedAtoB > 1., cErrInvalidParm);
    /* Means we want to do 2-Way searching */
    if (BM && AD)
    {
	ERR_GEN_PIXR(AD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    }


    /* Fill in reduction information */
    ERR_PASS( wp_fillInReductionInfo(pIPWState, pRedInfoA,
				     pRedInfoB,
				     AD,
				     BM,
				     *pMinPercentAtoB,
				     *pMinPercentBtoA,
				     pCallbackData,
				     callbackProc));

    
    /* Now scale the BD subrect x,y,w,h down to the equivalent subrect
     * in the reduced image, redBD.
     */
    w = (x+w-1) / pRedInfoB->reductionFactor;
    h = (y+h-1) / pRedInfoB->reductionFactor;
    x /= pRedInfoB->reductionFactor;
    y /= pRedInfoB->reductionFactor;
    w -= x-1;
    h -= y-1;

    /* Ensure the subrect is now completely within redBD image */
    ERR_PASS( w_pixrGetWidth(pIPWState, pRedInfoB->redBD, &wRedBD));
    ERR_PASS( w_pixrGetHeight(pIPWState, pRedInfoB->redBD, &hRedBD));
    w = MIN(w, wRedBD - x);
    h = MIN(h, hRedBD - y);

    /* minReducedBtoA */
    tmp = 0.;

    if (pMultResInfo)
    {
	/* Do 1-Way multi-resolution searching on the initial pass
	 * on the reduced images.  Once matches are found, then
	 * expand to the full resolution and do the desired 1-Way or 2-Way
	 * matching. 
	 */
	ERR_CHECK( w_hausdorffMultResMatches(pIPWState, pRedInfoA->redAM,
					     pRedInfoB->redBD,
					     pRedInfoB->redBDTemp,
					     NULL, NULL,
					     x, y, w, h,
					     searchMeth,
					     pMultResInfo,
					     (void *)pRedInfoB,
					     wp_foundReducedMatch,
					     &minReducedAtoB, &tmp,
					     &FRMcallbackRetCode))
	    {
		/* This logic is actually pretty tricky (at least I found it
		 * so).  There are two levels of callbacks here, one which
		 * calls the other.  The former callback is our private
		 * wp_foundReducedMatch() routine, which in turn, calls
		 * the actual client's callback.  So we have to figure out
		 * what level returned what error, so when the client 
		 * gets an error from this routine, sie will know whether
		 * or not hir routine actually returned the error.
		 */
		if (__error__ == cErrCallbackError) 
		{
		    if (FRMcallbackRetCode == cErrCallbackError)
		    {
			/* This means the CLIENT's callback routine
			 * returned an error
			 */
			*pCallbackReturnCode = pRedInfoB->callbackReturnCode;
			ERR_GEN(1, cErrCallbackError);
		    }
		    else
		    {
			/* This means wp_foundReducedMatch() returned an
			 * error (and the client's routine was okay)
			 */
			*pCallbackReturnCode = cSuccessful;
			ERR_GEN(1, FRMcallbackRetCode);
		    }
		}
		else
		{
		    /* This means that both levels of callbacks returned
		     * okay, but another error turned up elsewhere
		     */
		    *pCallbackReturnCode = cSuccessful;
		    ERR_GEN(1, __error__);
		}
	    }
    }
    else
    {
	/* Do 1-Way searching on the initial pass
	 * on the reduced images.  Once matches are found, then
	 * expand to the full resolution and do the desired 1-Way or 2-Way
	 * matching. 
	 */
	ERR_CHECK( w_hausdorffMatches (pIPWState, pRedInfoA->redAM,
				       pRedInfoB->redBD,
				       NULL, NULL,
				       x, y, w, h,
				       1, 1,
				       searchMeth,
				       (void *)pRedInfoB,
				       wp_foundReducedMatch,
				       &minReducedAtoB, &tmp,
				       &FRMcallbackRetCode))
	    {
		/* This logic is actually pretty tricky (at least I found it
		 * so).  There are two levels of callbacks here, one which
		 * calls the other.  The former callback is our private
		 * wp_foundReducedMatch() routine, which in turn, calls
		 * the actual client's callback.  So we have to figure out
		 * what level returned what error, so when the client 
		 * gets an error from this routine, sie will know whether
		 * or not hir routine actually returned the error.
		 */
		if (__error__ == cErrCallbackError) 
		{
		    if (FRMcallbackRetCode == cErrCallbackError)
		    {
			/* This means the CLIENT's callback routine
			 * returned an error
			 */
			*pCallbackReturnCode = pRedInfoB->callbackReturnCode;
			ERR_GEN(1, cErrCallbackError);
		    }
		    else
		    {
			/* This means wp_foundReducedMatch() returned an
			 * error (and the client's routine was okay)
			 */
			*pCallbackReturnCode = cSuccessful;
			ERR_GEN(1, FRMcallbackRetCode);
		    }
		}
		else
		{
		    /* This means that both levels of callbacks returned
		     * okay, but another error turned up elsewhere
		     */
		    *pCallbackReturnCode = cSuccessful;
		    ERR_GEN(1, __error__);
		}
	    }
    }

    /* Pass back the last values we used at the full res. search */
    *pMinPercentAtoB = pRedInfoB->minAtoB;
    *pMinPercentBtoA = pRedInfoB->minBtoA;

    /* This to indicate no error found in (client's!) callback */
    *pCallbackReturnCode = cSuccessful;
    NORMAL_RETURN;
}


/*

 $Log: hausrduc.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.25  1997/07/11 01:58:50  ddavies
 hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.24  1997/06/26 20:04:06  ddavies
 ddavies on Thu Jun 26 13:04:05 PDT 1997

 hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.23  1997/04/15 21:54:28  ddavies
 hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v

 Change name w_clonePixrZero to w_clonePixr.

 Revision 1.22  1996/01/04 20:24:11  ejaquith
 hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.21  1995/12/13  04:09:31  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:12:59  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/11/10  02:04:18  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.18  1995/10/30  17:58:13  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Updated function block comments who named other IPWrap routines.
 *  Added inclusion of "hauscomp.h"
 *
 * Revision 1.17  1995/10/13  22:02:13  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.16  1995/09/29  00:34:52  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.15  1995/09/22  04:29:27  ejaquith
 * ejaquith on Thu Sep 21 21:28:21 PDT 1995
 *
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Fixed bug in wp_mySetPixel() where it was setting the wrong
 * bits in the PIXR because I hadn't used the LONG pointer and was using
 * the SHORT pointer.
 *
 * Revision 1.14  1995/09/14  18:53:42  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Fixed checking of BM parameter in w_hausdorffReducedMatches().
 * Added pointer checks to all public IPWrap routines.
 *
 * Revision 1.13  1995/09/12  16:54:40  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.12  1995/09/08  00:14:37  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.11  1995/09/07  01:32:39  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.10  1995/09/06  20:20:39  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  02:20:29  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.8  1995/08/31  21:14:30  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/22  02:21:08  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.6  1995/08/17  22:01:07  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.5  1995/07/24  22:19:05  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Updated hausdorffReducedMatches() routine to
 * use _pointers_ to minimum Hausdorff percents, so it's possible to pass
 * back the last used/tightened percents that were used in the subrect
 * search.  This will allow us to tighten the minimum percents at the
 * FULL RESOLUTION regions.  Added another Hausdorff fraction parameter
 * to the routine.  This will server as the "trigger" percent in the
 * reduced regions -- that is, if reducedA matches reducedB with a Hausdorff
 * A to B percent of this amount or more, then it will search the corresponding
 * locations at the full resolution, using the other 2 minimum Hausdorff
 * percents as targets.  These latter 2 percents can then be updated by
 * the callback routine returning '1'.  This allows for efficient and less
 * confusing matching.  Previously without these changes, the routines
 * were erratic and unpredictable -- therefore this is an efficiency gain
 * and a BUG FIX.
 *
 * Revision 1.4  1995/04/22  21:52:48  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Removed writeReducedImages().
 * Updated all routines to properly use ipwrap level hausdorffMatches() and
 * hausdorffMultResMatches().
 *
 * Revision 1.3  1995/03/29  02:04:09  ejaquith
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.2  1995/03/24  02:19:53  ejaquith
 * ejaquith on Thu Mar 23 18:19:31 PST 1995
 *
 * hausrduc.c -> /product/ipcore/ipwrap/src/RCS/hausrduc.c,v
 *
 * Fixed a couple minor bugs in freeing the reduction information
 * structures.
 *
 * Revision 1.1  1995/03/21  19:48:44  ejaquith
 * Initial revision
 *
 
 */
