/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  	haussrch.c:   Hausdorff binary image comparator wrappers
 *
 *  	    Code to construct/deconstruct a set of models
 *  	    		w_getModelSetSize()
 *  	    		w_makeModelSet()
 *  	    		w_destroyModelSet()
 *
 *  	    Hausdorff matching routines
 *  	    		w_hausdorffMatches()
 *  	    		w_hausdorffListMatches()
 */

#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "alplib.h"
#include "hauscomp.h"

/* prototypes */
#include "hauscomp.pub"
#include "props.pub"
#include "shrpatt.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: haussrch.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");

/*____________________________________________________________________________
 *
 *
 * This purpose of this code is to implement comparison of two binary
 * bitmaps using the Hausdorff metric method.  The comparison can be either
 * 1-way or 2-way.  The method can be thought of as a "dilate and correlate"
 * method.
 *
 * To try it out, run the "haustest" program in ipshared/testsrc.
 *
 * High-level comments concerning the Hausdorff technique and ipshared
 * interface are included in ipshared/src/i_hc32ms.c.
 * This will go over only specific noteable comments in ipwrap and will
 * assume the reader has the ipshared level of understanding.
 *
 *METHODS:
 * At this level there are three methods of matching (once the model
 * set(s) have been made):
 *
 *   i)  hausdorffMatches()
 *
 *  ii) *makeMultResInfo()
 *       hausdorffMultResMatches()
 *      *deleteMultResInfo()
 *
 * iii)  makeReducedInfoA()
 *       makeReducedInfoB()
 *       hausdorffReducedMatches()
 *       deleteReducedInfoA()
 *       deleteReducedInfoB()
 *
 *  iv)  hausdorffListMatches()
 *
 * All four of these methods use a callback routine interface.  They each
 * set up the ipshared level search and call i_hausdorffSearch()
 * iteratively.  Every time a match is found, they invoke the callback.
 * Some sample/useable callbacks are in ipwrap/src/hauscall.c.
 *
 *   i) Regular Hausdorff matching
 *
 *  ii) Multi-resolutional search matching
 *      The routines with a '*' next to them indicate that they only need to be
 *      called once for each set of levels desired.  For example, if you want
 *      to try matching twenty A images into a full page B image and decide on
 *      levels 16,8,4,2,1, then the '*' routines are called once, and
 *      hausdorffMultResMatches() is called twenty times.
 *
 * iii) Reduce the A and B images and look for matches.  Where each match is
 *      found at this reduced resolution, try the corresponding area at the
 *      full resolution and return matches found here as the actual matches.
 *      Note this particular method is not guaranteed to find ALL matches.
 *
 *  iv) Identical to i) except instead of using the standard "check a set
 *      of valid translations in a subrect" interface, it uses an interface
 *      that just tries every translation in a list of passed in translations.
 *      (Thus no filtering is done, every translation is tried).
 *____________________________________________________________________________
 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determine the space required to make a model set
 *
 * LongDescription:::
 *  Returns the number of bytes that will be needed to make a model
 *  set given the specified arguments.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           binary PIXR desired to be the model bitmap
 *  blockHeight::    Indicates how to divide up and order the model bitmap
 *                   (should always be 0 when creating a _B_ model set)
 *  nBitmaps::       The number of desired bitmaps to use in constructing
 *                   the model set: 0, 1, or 32.
 *
 * Output:::
 *  pModelSetSize::  Ptr to returned number of bytes needed to make model set
 *
 * Returns:::
 *
 * PrivateNotes:::
 *  blockHeight::
 *   If 0, it means don't order the model rows, and later always sequentially
 *     compare the model row by row. (slowest)
 *   If >= sHeight, it means order all the model rows densest first.
 *   Otherwise, divide the model into blocks of blockHeight rows.  Within each
 *     block, order the rows densest first.  Compare the densest blocks first.
 *     This is to help avoid cache misses.  Thus, blockHeight should be set
 *     with the cache size and the width of the model bitmap AND the imagery
 *     bitmap in mind.
 * Notes:::
 *  Differences with ipshared::
 *   This function is different from IPShared's i_getModelSetSize fn. in that
 *   we are computing the space here for the entire model set including the
 *   model bitmaps (unlike ipshared's which does not include the bitmaps).
 *  Main Usage::
 *   Despite knowing the size, w_makeModelSet() still is responsible for
 *   allocating (and filling) the model set.  Thus, the main use for this
 *   routine will be to aid in determining which type to use based on the
 *   size a given type.
 *  nBitmaps::
 *   If 0, means make the model set data a list of points
 *     (instead of using bitmap(s) in comparisons.
 *   If 1, means use the ptr to pixr bitmap (no bitmaps copied)
 *   If 32, means copy the pixr bitmap 32 times
 *     (one copy per right shift 0..31),
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getModelSetSize (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               blockHeight,
    UInt8               nBitmaps,
    UInt32             *pModelSetSize)
{
    PROLOGUE;
    Int32     size, bitmapSize;
    Int32     w, h, bpl;
    UInt8    *pImage;

    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN(pModelSetSize == NULL, cErrNullPtr);

    ERR_GEN(nBitmaps != 0  &&  nBitmaps != 1  &&  nBitmaps != 32,
	    cErrInvalidParm);

    ERR_PASS( w_pixrGetImageAll(pIPWState, pixr, &pImage) );
    ERR_PASS( w_pixrFramedWidth(pIPWState, pixr, &w) );
    ERR_PASS( w_pixrFramedHeight(pIPWState, pixr, &h) );
    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &bpl) );

    ERR_PASS( i_getModelSetSize (pIPWState,  pImage, bpl, w, h, blockHeight,
				 nBitmaps, &bitmapSize, &size ));

    switch (nBitmaps)
    {
    case 0:
	size += 
	    sizeof(modelType);   /* Struct to store model */
	break;
    case 1:
	size +=
	    sizeof(modelType) +
	    sizeof(UInt8 *); /* 1 bitmap pointer */
	break;
    default: /* case 32: */
	size +=
	    sizeof(modelType) +
	    32 * sizeof(UInt8 *) + /* 32 bitmap pointers */
	    32 * bitmapSize;  /* we'll need 32 of these bitmaps */
	break;
    }

    *pModelSetSize = size;
    NORMAL_RETURN;
}

    
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Construct an A or B model set to use for Hausdorff matching.
 *
 * LongDescription:::
 *  Given a binary PIXR, this routine will return a filled model set.
 *  The model set type created is determined from the passed in nBitmaps.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           binary PIXR desired to be the model bitmap
 *  blockHeight::    Indicates how to divide up and order the model bitmap
 *                   (should always be 0 when creating a _B_ model set)
 *  nBitmaps::       The number of desired bitmaps to use in constructing
 *                   the model set: 0, 1, or 32.
 *
 * Output:::
 *  ppM::            Ptr to ptr to internal representation of model set.
 *                   ppM is allocated and filled in. 
 *                   *ppM must point to NULL on input.
 *                   *ppM will then point to the finished model set.
 *
 * Returns:::
 *
 * PrivateNotes:::
 *  blockHeight::
 *   If 0, it means don't order the model rows, and later always sequentially
 *     compare the model row by row. (slowest)
 *   If >= sHeight, it means order all the model rows densest first.
 *   Otherwise, divide the model into blocks of blockHeight rows.  Within each
 *     block, order the rows densest first.  Compare the densest blocks first.
 *     This is to help avoid cache misses.  Thus, blockHeight should be set
 *     with the cache size and the width of the model bitmap AND the imagery
 *     bitmap in mind.
 * Notes:::
 *  nBitmaps::
 *   If 0, means make the model set data a list of points
 *     (instead of using bitmap(s) in comparisons.
 *   If 1, means use the ptr to pixr bitmap (no bitmaps copied)
 *   If 32, means copy the pixr bitmap 32 times
 *     (one copy per right shift 0..31),
 *  zero border::
 *   It may be desirable for the image to have the ADDITIONAL
 *   'dilateRadius' zeroed border around the image.
 *   (see note on 'dilateRadius' at top of ipwrap/haussrch.c)
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeModelSet (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               blockHeight,
    UInt8               nBitmaps,
    modelType         **ppM)
{
    PROLOGUE;
    Int32        size, bitmapSize;
    Int32        fw, fh, bpl;
    UInt8       *pImage;
    modelType   *pM = NULL;

    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN(ppM == NULL, cErrNullPtr);
    ERR_GEN(*ppM != NULL, cErrPtrNotNull);


    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &bpl) );
    ERR_PASS( w_pixrFramedHeight(pIPWState, pixr, &fh) );
    ERR_PASS( w_pixrFramedWidth(pIPWState, pixr, &fw) );
    ERR_PASS( w_pixrGetImageAll(pIPWState, pixr, &pImage) );

    ERR_PASS( i_getModelSetSize (pIPWState,  pImage, bpl, fw, fh, blockHeight,
				 nBitmaps, &bitmapSize, &size ));

    pM = (modelType *)IPW_MALLOC(sizeof(modelType), cAllocatorData);
    ERR_GEN(!pM, cErrNomem);
    pM->pIMS = NULL;
    pM->ppBitmaps = NULL;
    pM->pBitmapData = NULL;

    pM->pIMS = (void *)IPW_MALLOC(size, cAllocatorData);
    ERR_GEN_CLEANUP(!pM->pIMS, cErrNomem);

    if (nBitmaps == 32)
    {
	Int32   i;
			    
	pM->nBitmaps = 32;
	pM->ppBitmaps = (UInt8 **)IPW_CALLOC(32, sizeof(UInt8 *),
					     cAllocatorData);
	pM->pBitmapData = (UInt8 *)IPW_CALLOC(bitmapSize * 32,
					      sizeof(UInt8),
					      cAllocatorImage);
	ERR_GEN_CLEANUP(!pM->ppBitmaps  ||  !pM->pBitmapData,
			cErrNomem);

	/* Assign the 32 bitmap pointers */
	pM->ppBitmaps[0] = pM->pBitmapData;
	for (i = 1; i < 32; i++)
	    pM->ppBitmaps[i] = pM->ppBitmaps[i-1] + bitmapSize;
    }
    else
    {
	ERR_GEN_CLEANUP(nBitmaps != 0  &&  nBitmaps != 1, cErrInvalidParm);
	pM->nBitmaps = nBitmaps;
	
	if (nBitmaps == 1)
	{
	    pM->ppBitmaps = (UInt8 **)IPW_MALLOC(sizeof(UInt8 *),
						     cAllocatorData);
	    ERR_GEN_CLEANUP(!pM->ppBitmaps, cErrNomem);

	    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, pixr, 
					  &(pM->ppBitmaps[0])));
	}
    }


    ERR_PASS_CLEANUP( i_makeModelSet (pIPWState,  pImage,
				      bpl, fw, fh, blockHeight, 
				      pM->ppBitmaps,
				      nBitmaps, pM->pIMS ));

    *ppM = pM;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pM != NULL)
    {
	IPW_FREE(pM->pIMS, cAllocatorData);
	IPW_FREE(pM->ppBitmaps, cAllocatorData);
	IPW_FREE(pM->pBitmapData, cAllocatorImage);
	IPW_FREE(pM, cAllocatorData);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free model set and set ptr to NULL
 *
 * LongDescription:::
 *  Given the ptr to the internal representation of the model set,
 *  free the model set's allocated memory.  If already NULL, this routine
 *  does nothing; o.w. it will set ptr to NULL upon exiting.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppM::            Ptr to ptr to internal representation of model set
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyModelSet (
    IPWState           *pIPWState,
    modelType         **ppM)
{
    PROLOGUE;

    ERR_GEN(ppM == NULL, cErrNullPtr);

    if (*ppM != NULL)
    {
	IPW_FREE((*ppM)->pIMS, cAllocatorData);
	IPW_FREE((*ppM)->ppBitmaps, cAllocatorData);
	IPW_FREE((*ppM)->pBitmapData, cAllocatorImage);
	IPW_FREE(*ppM, cAllocatorData);
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find Hausdorff matches of A and B by placing A at a list of locations.
 *
 * LongDescription:::
 *  Given A's model set, place it over B's dilated pixr at every
 *  location in a list of translations and calculate the A to B Hausdorff
 *  percent.  If 2-Way, then take the corresponding B's model set over A's
 *  dilated pixr and calculate the B to A Hausdorff percent.
 *
 *  Report subrect location(s) with the Hausdorff percent(s)
 *  at least the passed in value(s), as matches of A and B.
 *
 * Input:::
 *  pIPWState::           Ptr to IP struct for memory alloc/freeing routines
 *  AM::                  Ptr to A's model set
 *  BD::                  Ptr to B-dilated binary pixr
 *  BM::                  Ptr to B's model set or NULL if 1-Way
 *  AD::                  Ptr to A-dilated binary pixr if 2-Way.
 *                        If NULL, indicates do only 1-Way
 *  pListX::              Array of X and Y components of desired
 *  pListY::              translations to search
 *  nTrans::              Size of above arrays; number of
 *                        translations to search
 *  searchMeth::          Desired search method. 0 or 
 *                        USE_FULL_B_FOR_BTOA_PERCENT
 *  pCallbackData::       Ptr to client supplied data that we
 *                        include as an argument when calling the
 *                        client's callback proc.
 *  callbackProc::        Ptr to the client's callback routine,
 *                        called with the match's translation x
 *                        and y offset, the A to B match percent,
 *                        the B to A match percent, (ignored if 1-Way)
 *                        the client supplied argument, and a pointer to a
 *                        value to shape further invocations (see Notes).
 *  pMinPercentAtoB::     The match(es) must have A to B Hausdorff
 *                        percent(s) at least the value this
 *                        points to, [0..1].  See notes below.
 *  pMinPercentBtoA::     The 2-Way match(es) must additionally
 *                        have B to A Hausdorff percent(s) at least
 *                        the value this points to, [0..1].
 *                        Ignored if 1-Way.  See notes below.
 *
 * Output:::
 *  pCallbackReturnCode:: Ptr to code returned by client's
 *                        callback routine
 *
 * Returns:::
 *
 * Notes:::
 *  Match locations::
 *    they are in the frame of the full B image.
 *  callbackRoutine pNextAsGood value::
 *    Return 1 to make all further routine invocations report matches AT LEAST
 *      as good as the match in the current invocation.
 *    Return 0 to keep same matching criterion as before invocation.
 *  pMinPercentAtoB, pMinPercentBtoA::
 *    On completion, these will point to the min percents last used in the
 *    search.  (One can inspect if they had been increased by the callback).
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_hausdorffListMatches (
    IPWState           *pIPWState,
    modelType          *AM,
    PIXR               *BD,
    modelType          *BM,
    PIXR               *AD,
    Int32              *pListX,
    Int32              *pListY,
    Int32               nTrans,
    UInt8               searchMeth,
    void               *pCallbackData,
    ErrCode	(CDECL *callbackProc) (IPWState *pIPWState,
				       void *,
				       Int32 tx,
				       Int32 ty,
				       Float64 percentAtoB,
				       Float64 percentBtoA,
				       Int8 *pNextAsGood),
    Float64            *pMinPercentAtoB,
    Float64            *pMinPercentBtoA,
    ErrCode            *pCallbackReturnCode)
{
    PROLOGUE;
    UInt32   searchInfoSize;
    void    *pSearchInfo = NULL;
    Int32    tx, ty;
    UInt8    done;
    Float64  percentAtoB, percentBtoA;
    Float64  minPercentAtoB, minPercentBtoA;
    Int8     nextAsGood;
    UInt8   *pImageAD, *pImageBD;
    Int32    fwAD, fwBD, fhAD, fhBD, bplAD, bplBD;
    

    ERR_GEN_PIXR(BD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN((BM == NULL  &&  AD != NULL) ||
	    (BM != NULL  &&  AD == NULL), cErrInvalidParm);
    if (AD != NULL)
	ERR_GEN_PIXR(AD, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN(AM == NULL  ||  callbackProc == NULL  ||
	    pMinPercentAtoB == NULL  ||  pMinPercentBtoA == NULL  ||
	    pCallbackReturnCode == NULL, cErrNullPtr);

    ERR_GEN(*pMinPercentAtoB < 0. || *pMinPercentAtoB > 1. ||
	    *pMinPercentBtoA < 0. || *pMinPercentBtoA > 1.,
	    cErrInvalidParm);

    minPercentAtoB = *pMinPercentAtoB;
    minPercentBtoA = *pMinPercentBtoA;

    ERR_PASS( i_hausdorffListSearchSize (pIPWState, nTrans, &searchInfoSize) );

    pSearchInfo = (void *)IPW_MALLOC(searchInfoSize, cAllocatorData);
    ERR_GEN(!pSearchInfo, cErrNomem);

    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, BD, &pImageBD) );
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, BD, &fwBD) );
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, BD, &fhBD) );
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, BD, &bplBD) );

    if (AD)
    {
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, AD, &pImageAD) );
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, AD, &fwAD) );
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, AD, &fhAD) );
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, AD, &bplAD) );
    }
    else {
	pImageAD = NULL;
	fwAD = fhAD = bplAD = 0;
    }

    ERR_PASS_CLEANUP( i_hausdorffListSearchInit (pIPWState, AM->pIMS,
						 pImageBD,
						 bplBD, fwBD, fhBD,
						 BM ? BM->pIMS : NULL,
						 pImageAD,
						 bplAD, fwAD, fhAD,
						 pListX, pListY, nTrans,
						 searchMeth, pSearchInfo) );
    
    do 	
    {
	ERR_PASS_CLEANUP( i_hausdorffSearch (pIPWState, AM->pIMS,
					     AM->ppBitmaps,
					     AM->nBitmaps,
					     pImageBD,
					     BM ? BM->pIMS : NULL,
					     BM ? BM->ppBitmaps : NULL,
					     (UInt8) (BM ? BM->nBitmaps : 0),
					     pImageAD,
					     pSearchInfo,
					     minPercentAtoB,
					     AD ? minPercentBtoA : -1.0F,
					     &tx, &ty,
					     &percentAtoB, &percentBtoA,
					     &done) );
	if (!done)
	{
	    ERR_CHECK(callbackProc(pIPWState, pCallbackData, tx, ty,
				   percentAtoB,
				   percentBtoA, &nextAsGood)) 
		{
		    *pCallbackReturnCode = __error__;
		    ERR_GEN_CLEANUP(1, cErrCallbackError);
		}
	    if (nextAsGood)
	    {
		minPercentAtoB = percentAtoB;
		minPercentBtoA = percentBtoA;
	    }
	}
    } while (!done);

    
    IPW_FREE(pSearchInfo, cAllocatorData);

    *pMinPercentAtoB = minPercentAtoB;
    *pMinPercentBtoA = minPercentBtoA;

    /* This to indicate no error found in callback */
    *pCallbackReturnCode = cSuccessful;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pSearchInfo, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find all 1-Way or 2-Way matches of A in subrect of B
 *
 * LongDescription:::
 *  Given A's model set, place it over/at every location where A fits inside a
 *  passed in subrect of B's dilated pixr, and calculate the A to B Hausdorff
 *  percent.  If 2-Way, then take the corresponding B's model set over A's
 *  dilated pixr and calculate the B to A Hausdorff percent.
 *
 *  Report subrect location(s) with the Hausdorff percent(s)
 *  at least the passed in value(s), as matches of A and B.
 *
 * Input:::
 *      pIPWState::	Ptr to IP struct for memory alloc/freeing routines
 *	AM::		Ptr to A's model set
 *	BD::		Ptr to B-dilated binary pixr
 *	BM::		Ptr to B's model set or NULL if 1-Way
 *	AD::		Ptr to A-dilated binary pixr if 2-Way.
 *			If NULL, indicates do only 1-Way.
 *	x::		Rectangular area in B where we search
 *	y::		for A to match.
 *	w::		Must be entirely inside B.
 *	h::
 *	skipx::		Amounts to skip.  Check for a match    
 *	skipy::		every skipx pixel in X, every skipy pixel
 *			in Y in subrect translation area.  (1 or greater)
 *	searchMeth::	Desired search method. One of:           
 *			{SPARSE_FILTER, FLY_FILTER, EVERY_TRANS} 
 *			May be OR-ed with USE_FULL_B_FOR_BTOA_PERCENT
 *	pCallbackData::	Ptr to client supplied data that we      
 *			include as an argument when calling the  
 *			client's callback proc.
 *	callbackProc::	Ptr to the client's callback routine,
 *			called with the match's translation x
 *			and y offset, the A to B match percent,
 *			the B to A match percent, (ignored if 1-Way)
 *			the client supplied argument, and a pointer to a
 *			value to shape further invocations (see Notes).
 *	pMinPercentAtoB::	The match(es) must have A to B Hausdorff 
 *				percent(s) at least the value this points to,
 *				[0..1].  See notes below.
 *	pMinPercentBtoA::	The 2-Way match(es) must additionally
 *				have B to A Hausdorff percent(s) at least
 *				the value this points to, [0..1].
 *				Ignored if 1-Way.  See notes below
 *
 * Output:::
 *	pCallbackReturnCode::	Ptr to code returned by client's
 *				callback routine
 *
 * Returns:::
 *
 * Notes:::
 *  If skipx == skipy == 1::
 *    will try every translation in the subrect.
 *  Match locations::
 *    are in the frame of the full B image.
 *  callbackRoutine pNextAsGood value::
 *    Return 1 to make all further routine invocations report matches AT LEAST
 *      as good as the match in the current invocation.
 *    Return 0 to keep same matching criterion as before invocation.
 *  pMinPercentAtoB, pMinPercentBtoA::
 *    On completion, these will point to the min percents last used in the
 *    search.  (One can inspect if they had been increased by the callback).
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_hausdorffMatches (
    IPWState           *pIPWState,
    modelType          *AM,
    PIXR               *BD,
    modelType          *BM,
    PIXR               *AD,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               skipx,
    Int32               skipy,
    UInt8               searchMeth,
    void               *pCallbackData,
    ErrCode	(CDECL *callbackProc) (IPWState *pIPWState,
				       void *,
				       Int32 tx,
				       Int32 ty,
				       Float64 percentAtoB,
				       Float64 percentBtoA,
				       Int8 *pNextAsGood),
    Float64            *pMinPercentAtoB,
    Float64            *pMinPercentBtoA,
    ErrCode            *pCallbackReturnCode)
{
    PROLOGUE;
    UInt32   searchInfoSize;
    void    *pSearchInfo = NULL;
    Int32    tx, ty;
    UInt8    done;
    Float64  percentAtoB, percentBtoA;
    Float64  minPercentAtoB, minPercentBtoA;
    Int8     nextAsGood;
    UInt8   *pImageAD, *pImageBD;
    Int32    fwAD, fwBD, fhAD, fhBD, bplAD, bplBD;

    ERR_GEN_PIXR(BD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN((BM == NULL  &&  AD != NULL) ||
	    (BM != NULL  &&  AD == NULL), cErrInvalidParm);
    if (AD != NULL) {
	ERR_GEN_PIXR(AD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    }
    
    ERR_GEN(AM == NULL  ||  callbackProc == NULL  ||
	    pMinPercentAtoB == NULL  ||  pMinPercentBtoA == NULL  ||
	    pCallbackReturnCode == NULL, cErrNullPtr);
        
    ERR_GEN(*pMinPercentAtoB < 0. || *pMinPercentAtoB > 1. ||
	    *pMinPercentBtoA < 0. || *pMinPercentBtoA > 1.,
	    cErrInvalidParm);
    ERR_GEN(skipx < 1 || skipy < 1, cErrInvalidParm);

    minPercentAtoB = *pMinPercentAtoB;
    minPercentBtoA = *pMinPercentBtoA;


    ERR_PASS( i_hausdorffRectSearchSize (pIPWState, AM->pIMS, w, searchMeth, 
					 &searchInfoSize) );

    pSearchInfo = (void *)IPW_MALLOC(searchInfoSize, cAllocatorData);
    ERR_GEN(!pSearchInfo, cErrNomem);

    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, BD, &pImageBD) );
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, BD, &fwBD) );
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, BD, &fhBD) );
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, BD, &bplBD) );

    if (AD)
    {
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, AD, &pImageAD) );
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, AD, &fwAD) );
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, AD, &fhAD) );
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, AD, &bplAD) );
    }
    else {
	pImageAD = NULL;
	fwAD = fhAD = bplAD = 0;
    }

    ERR_PASS_CLEANUP( i_hausdorffRectSearchInit (pIPWState, AM->pIMS,
						 pImageBD,
						 bplBD, fwBD, fhBD,
						 BM ? BM->pIMS : NULL,
						 pImageAD,
						 bplAD, fwAD, fhAD,
						 x, y, w, h,
						 skipx, skipy,
						 searchMeth, pSearchInfo) );

    do 	
    {
	ERR_PASS_CLEANUP( i_hausdorffSearch (pIPWState, AM->pIMS,
					     AM->ppBitmaps,
					     AM->nBitmaps,
					     pImageBD,
					     BM ? BM->pIMS : NULL,
					     BM ? BM->ppBitmaps : NULL,
					     (UInt8) (BM ? BM->nBitmaps : 0),
					     pImageAD,
					     pSearchInfo,
					     minPercentAtoB,
					     AD ? minPercentBtoA : -1.0F,
					     &tx, &ty,
					     &percentAtoB, &percentBtoA,
					     &done) );
	if (!done)
	{
	    ERR_CHECK(callbackProc(pIPWState, pCallbackData, tx, ty,
				   percentAtoB,
				   percentBtoA, &nextAsGood)) 
		{
		    *pCallbackReturnCode = __error__;
		    ERR_GEN_CLEANUP(1, cErrCallbackError);
		}
	    if (nextAsGood)
	    {
		minPercentAtoB = percentAtoB;
		minPercentBtoA = percentBtoA;
	    }
	}
    } while (!done);

    IPW_FREE(pSearchInfo, cAllocatorData);

    *pMinPercentAtoB = minPercentAtoB;

    /* We have to do this to ensure that value did not _decrease_ from how
     * it was passed in -- specifically, if we were doing 1way matching only,
     * we would always be getting -1. back in the above percentBtoA value
     * from the i_hausdorffSearch() calls!
     */
    *pMinPercentBtoA = MAX(*pMinPercentBtoA, minPercentBtoA);

    /* This to indicate no error found in callback */
    *pCallbackReturnCode = cSuccessful;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pSearchInfo, cAllocatorData);
    ERR_RETURN;
}


/*

 $Log: haussrch.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 Imported sources

 Revision 1.30  1998/01/14 00:13:06  ddavies
 haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.29  1997/07/11 01:58:53  ddavies
 haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.28  1996/02/29 03:28:06  ejaquith
 ejaquith on Wed Feb 28 19:26:10 PST 1996

 haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v

 Updated routines w_getModelSetSize() and w_makeModelSet()
 to use newly modified i_getModelSetSize().  They also now use the
 returned size-per-bitmap parameter for calculating the memory block
 size for the 32-bitmap Hausdorff model set type.
 (This is more robust and simpler).

 Revision 1.27  1996/01/04 20:24:15  ejaquith
 haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.26  1995/12/13  04:09:35  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.25  1995/11/13  23:56:01  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Fixed 4 Input::: parameter typos.
 *
 * Revision 1.24  1995/11/13  20:13:03  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.23  1995/11/10  02:04:22  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.22  1995/10/30  17:58:31  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Added inclusion of "hauscomp.h"
 * Updated function block comments who named other IPWrap routines.
 *
 * Revision 1.21  1995/10/13  22:02:17  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/10/13  03:28:48  ejaquith
 * ejaquith on Thu Oct 12 20:28:18 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Casted some params to shut up C++ compiler.
 *
 * Revision 1.19  1995/09/29  00:34:40  ejaquith
 * ejaquith on Thu Sep 28 17:34:39 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.18  1995/09/22  04:31:04  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * fixed logic in w_hausdorffMatches () to work properly in
 * w_hausdorffReducedMatches().  Specifically, made the passed back B to A
 * percent be at least 0 (when 1way matching, it would return -1)
 *
 * Revision 1.17  1995/09/14  18:54:18  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Fixed checking of BM parameter in w_hausdorffListMatches() and
 * w_hausdorffMatches().
 *
 * Revision 1.16  1995/09/12  16:54:33  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.15  1995/09/08  00:14:31  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.14  1995/09/07  01:32:42  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:20:43  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  02:20:23  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.11  1995/08/31  21:04:24  ejaquith
 * ejaquith on Thu Aug 31 14:04:23 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/24  02:51:10  ejaquith
 * ejaquith on Wed Aug 23 19:49:57 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Changed pixrGetWidth/Height to framed versions in makeModelSet() as they
 * should have been.
 *
 * Revision 1.9  1995/08/17  22:00:59  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.8  1995/07/24  22:13:23  ejaquith
 * ejaquith on Mon Jul 24 15:10:41 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Updated hausdorffMatches() and hausdorffListMatches() routines to
 * use _pointers_ to minimum Hausdorff percents, so it's possible to pass
 * back the last used/tightened percents that were used in the subrect
 * /list search.   This was needed for hausdorffReducedMatches() and will
 * provide the client more information about what happened during the
 * invocation.
 *
 * Revision 1.7  1995/07/18  02:21:43  ejaquith
 * ejaquith on Mon Jul 17 19:21:21 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Added some soon-to-be-standard function header comments for testing
 * auto-doc module.
 *
 * Revision 1.6  1995/05/11  20:12:54  ejaquith
 * ejaquith on Thu May 11 13:09:58 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Added routine getModelSetSize().  This will determine the space that
 * makeModelSet() will allocate in the total number of bytes to construct
 * the IPWRAP level model set.  Note that this number is much larger than
 * the number returned by the IPSHARED level i_getModelSetSize() for
 * nBitmaps parameter values of 1 and 32 because this number includes all
 * the space needed for the bitmaps created in makeModelSet() (but not
 * i_makeModelSet()).
 *
 * Revision 1.5  1995/04/25  17:51:19  ejaquith
 * ejaquith on Tue Apr 25 10:50:18 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Fixed two (valid!) acc compiler warnings of where I had (conditionally)
 * passed in NULL to an integer param of i_hausdorffSearch() call.
 *
 * Revision 1.4  1995/04/22  21:52:43  ejaquith
 * ejaquith on Sat Apr 22 14:52:42 PDT 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Removed getModelSetSize() and pixrPutModel().
 * Added hausdorffListMatches().
 * Updated general ipwrap level Hausdorff module comments.
 * Dropped usage of Hausdorff bit flags in routines.
 * makeModelSet() now has client pass in the number of bitmaps they want to have
 * the model set stored in, 0, 1, or 32.
 * makeModelSet() now allocates space for these bitmaps and complies with the
 * new ipshared interface.  It fill the bitmaps and stores in a new modelType
 * struct.
 * Updated deleteModelSet() to properly delete this new modleType struct.
 * Updated hausdorffMatches() to use search method interface instead of Hausdorff
 * bit flags interface.
 * Added hausdorffListMatches() to search a list of translations of A in B.
 * Both ..Matches() routines use the new ipshared iteratively searching interface
 * but keep the ipwrap level callback interface.  Why both?  Well, as of now
 * there are four nice choices for callbacks, and four ipwrap search routines.
 * To make ipwrap drop callbacks, an incredible amount of code would need to
 * be replicated, unless we used callbacks internally.  Realizing this, it seems
 * silly to not allow the client to continue to use callbacks.  Since the
 * functionality is exactly the same, I have not added simple wrappers that
 * allow the iterative search calling style as in ipshared.
 * Both ..Matches() routines now require the already created model set be
 * passed in compared to the previous interface which allowed either the
 * model set or source pixr be passed in.
 * Added all modifications to comply with usage of new ipshared level as
 * necessary.
 *
 * Revision 1.3  1995/03/31  22:17:24  ejaquith
 * ejaquith on Fri Mar 31 14:16:48 PST 1995
 *
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Removed probFalseMatch() wrap to ipshared as this routine was not
 * returning good confidence values 100% of the time (and is therefore
 * unusable).
 *
 * Revision 1.2  1995/03/29  02:04:17  ejaquith
 * haussrch.c -> /product/ipcore/ipwrap/src/RCS/haussrch.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.1  1995/03/21  19:44:41  ejaquith
 * Initial revision
 *
 
 */
