/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * histeq.c	Histogram equalization wrapper
 *
 */

#include "ipwcore.prv"
#include "pixr.h"
#include "defines.pub"

#include "geomadt.h"
#include "props.pub"
#include "shrip.pub"

#include "histeq.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: histeq.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Histogram equalization of 8-bit gray images
 *
 * LongDescription:::
 *   Equalizes the histogram in the gray range between "lo" and "hi"
 *   (inclusive).  Gray values which were the same in the input will be
 *   the same in the output, but may be shifted up or down in value.
 *   This algorithm can be useful to increase contrast in low contrast
 *   images.  It will attempt to evenly spread the gray values throughout
 *   the range.  For text (or other high contrast images), the result
 *   will be lower in contrast (with lots of intermediate gray values).
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  inPr::		Image to be histogram equalized
 *  lo::		lowest gray value in band to be equalized [0-254]
 *  hi::		highest gray value in band to be equalized [1-255]
 *
 * Output:::
 *  outPr::		Resulting histogram equalized image.
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram analysis
 *
 * Keywords:::
 *  automatic enhancement, histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *  Digital Image Processing, Gonzalez and Woods, Addison-Wesley, 1992.
 *  pp 173-180.  ISBN 0-201-50803-6.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_histogramEqualizeBand (
			IPWState  *pIPWState,
			PIXR      *inPr,
			Int32      lo,
			Int32      hi,
			PIXR     **outPr)
{
    PROLOGUE;
    Int32     w, h, inBpl, outBpl;
    Bool      allocatedHere = cFalse;
    UInt8     *pS, *pD;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    ERR_GEN (lo >= hi, cErrInvalidParm);
    ERR_GEN ((lo < 0) || (hi > 255), cErrInvalidParm);

    if ((*outPr) != NULL)
	ERR_GEN_PIXR (*outPr, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    else
    {
        ERR_PASS (w_clonePixrNoInit (pIPWState, inPr, outPr));
	allocatedHere = cTrue;
    }

    ERR_PASS_CLEANUP (w_pixrFramedWidth(pIPWState, inPr, &w));
    ERR_PASS_CLEANUP (w_pixrFramedHeight(pIPWState, inPr, &h));
    ERR_PASS_CLEANUP (w_pixrGetImageAll(pIPWState, inPr, &pS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll(pIPWState, *outPr, &pD));
    ERR_PASS_CLEANUP (w_pixrGetBpl(pIPWState, inPr, &inBpl));
    ERR_PASS_CLEANUP (w_pixrGetBpl(pIPWState, *outPr, &outBpl));

    ERR_PASS_CLEANUP (i_histogramEqualizeBand8 (pIPWState, lo, hi, pS,
						w, h, inBpl,
                                                pD, outBpl));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
        w_destroyPixr (pIPWState, outPr);

    ERR_RETURN;

}  /* w_histogramEqualizeBand () */


/*
 * $Log: histeq.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 * Imported sources
 *
 * Revision 1.3  1998/01/14 00:13:09  ddavies
 * histeq.c -> /product/ipcore/ipwrap/src/RCS/histeq.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.2  1997/04/15 21:54:47  ddavies
 * histeq.c -> /product/ipcore/ipwrap/src/RCS/histeq.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.1  1996/08/05 21:13:33  dferrell
 * Initial revision
 *
 *
 */

