/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/*
 * File:        hutuls.c
 *
 * Description: Basic routines on histograms
 * NewHisto()             Allocate new histogram
 * NewBlankHisto()        Allocate new histogram and set to 0
 * NewHistoLikeTemplate() Allocate new histogram using provided
 *                        histogram as template
 * DestroyHisto()         Free histogram and set pointer to NULL
 * ClearHisto()           Set histogram to 0
 * CopyHisto()            Copy source histogram into destination
 * WriteHisto()           Write histogram to a file
 * SmoothHisto()          moving average over histogram

 * Author: Luc Vincent
 * Stuck into IPWRAP and edited a bit by P. MacLeod
*/


#include <stdio.h>

#include "ipwcore.prv"
#include "defines.pub"
#include "histo.h"



#include "hutils.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif


IP_RCSINFO(RCSInfo, "$RCSfile: hutils.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:17 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Allocate a new histogram
 *
 * LongDescription:::
 * Allocates histogram array, sets minval and maxval and makes
 * histogram data point to bin 0.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  minval::         	minimal value represented in histogram
 *  maxval::         	maximal value represented in histogram
 *  padding::        	how many more bins to allocate on either side
 *			of array
 *
 * Output:::
 *  pphisto::        	new histogram
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Luc Vincent
 *  Peter Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newHisto (
    IPWState           *pIPWState,
    Int32               minval,
    Int32               maxval,
    Int32               padding,
    Histo             **pphisto)
{
    PROLOGUE;
    Histo     *histo = NULL;
    Int32     *histo_array = NULL;
    Int32      histo_size;


    ERR_GEN(minval > maxval, cErrInvalidParm);
    ERR_GEN(padding < 0, cErrInvalidParm);

    histo = (Histo *)IPW_MALLOC(sizeof(Histo), cAllocatorData);
    ERR_GEN(histo == NULL, cErrNomem);

    histo_size = maxval - minval + 1 + (2 * padding);
    histo_array =
	 (Int32 *)IPW_MALLOC(histo_size * sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP(histo_array == NULL, cErrNomem);

    HISTO_SET_MINVAL(histo, minval);
    HISTO_SET_MAXVAL(histo, maxval);
    HISTO_SET_ARRAY (histo, histo_array);
    HISTO_SET_SIZE  (histo, histo_size);
    HISTO_SET_DATA  (histo, histo_array - minval + padding);

    *pphisto = histo;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (histo != NULL) IPW_FREE(histo, cAllocatorData);
    ERR_RETURN;
} /* NewHisto */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Allocate new histogram and set bins to 0.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  minval::         	minimal value represented in histogram
 *  maxval::         	maximal value represented in histogram
 *  padding::        	how many more bins to allocate on either side
 *			of array
 *
 * Output:::
 *  pphisto::        	new histogram
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Luc Vincent
 *  Peter Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newBlankHisto (
    IPWState           *pIPWState,
    Int32               minval,
    Int32               maxval,
    Int32               padding,
    Histo             **pphisto)
{
    PROLOGUE;
    Histo     *histo;

    ERR_PASS( w_newHisto(pIPWState, minval, maxval, padding, &histo));

    ERR_PASS_CLEANUP( w_clearHisto(pIPWState, histo));

    *pphisto = histo;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyHisto(pIPWState, &histo);
    ERR_RETURN;

} /* NewBlankHisto */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Allocate new histogram using provided one as template
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  histoS::         	template histogram
 *
 * Output:::
 *  pHistoD::        	new histogram
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Luc Vincent
 *  Peter Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newHistoLikeTemplate (
    IPWState           *pIPWState,
    Histo              *histoS,
    Histo             **pHistoD)
{
    PROLOGUE;
    Histo     *histoD;
    Int32     *histoD_array;
    Int32      histoD_size;


    histoD = (Histo *)IPW_MALLOC(sizeof(Histo), cAllocatorData);
    ERR_GEN(histoD == NULL, cErrNomem);

    histoD_size = HISTO_GET_SIZE(histoS);
    histoD_array =
	(Int32 *)IPW_MALLOC(histoD_size * sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP(histoD_array == NULL, cErrNomem);

    HISTO_SET_MINVAL(histoD, HISTO_GET_MINVAL(histoS));
    HISTO_SET_MAXVAL(histoD, HISTO_GET_MAXVAL(histoS));
    HISTO_SET_ARRAY (histoD, histoD_array);
    HISTO_SET_SIZE  (histoD, histoD_size);
    HISTO_SET_DATA  (histoD,
		     histoD_array + (HISTO_GET_DATA(histoS) -
				     HISTO_GET_ARRAY(histoS)));
    *pHistoD = histoD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (histoD != NULL) IPW_FREE(histoD, cAllocatorData);
    ERR_RETURN;

} /* NewHistoLikeTemplate */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Free histogram structure and data. Set to NULL.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  pphisto::        	histogram to free
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Luc Vincent
 *  Peter Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyHisto (
    IPWState           *pIPWState,
    Histo             **pphisto)
{
    PROLOGUE;

    if (*pphisto != NULL) {
	if (HISTO_GET_ARRAY(*pphisto))
	    IPW_FREE(HISTO_GET_ARRAY(*pphisto), cAllocatorData);
 	IPW_FREE(*pphisto, cAllocatorData);
    }
    *pphisto = NULL;
    NORMAL_RETURN;
} /* DestroyHisto */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set all histogram bins to 0, including "padding"
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *
 * Output:::
 *  histo::          	cleared histogram
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Luc Vincent
 *  Peter Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_clearHisto (
    IPWState           *pIPWState,
    Histo              *histo)
{
    PROLOGUE;
/*    Int32 minval = HISTO_GET_MINVAL(histo); */

    pIPWState = pIPWState;
    memset ((void *)(HISTO_GET_ARRAY(histo)), 0,
	    HISTO_GET_SIZE(histo) * sizeof(Int32));

    NORMAL_RETURN;

} /* ClearHisto */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Copy source histogram into destination histogram one as template
 *
 * LongDescription:::
 * Copies source histogram to destination histogram.  If destination
 * is NULL on entry, a new histogram is allocated.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  histoS::         	source histogram
 *
 * Output:::
 *  histoD::         	destination histogram 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Luc Vincent
 *  Peter Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copyHisto (
    IPWState           *pIPWState,
    Histo              *histoS,
    Histo             **histoD)
{
    PROLOGUE;
    Histo       *dstHisto;
/*    Int32     *histoD_array; */
/*    Int32      histoD_size; */


    ERR_GEN(histoS == NULL, cErrInvalidParm);
    if (histoS == *histoD)
	NORMAL_RETURN;
    dstHisto = *histoD;

    if (dstHisto != NULL)
    {
   	ERR_GEN((HISTO_GET_MINVAL(histoS) != HISTO_GET_MINVAL(dstHisto)) ||
		(HISTO_GET_MAXVAL(histoS) != HISTO_GET_MAXVAL(dstHisto)),
	        cErrInvalidParm);
  	ERR_GEN(HISTO_GET_SIZE(histoS) != HISTO_GET_SIZE(dstHisto),
		cErrInvalidParm);
	ERR_GEN((HISTO_GET_DATA(histoS) - HISTO_GET_ARRAY(histoS)) !=
	        (HISTO_GET_DATA(dstHisto) - HISTO_GET_ARRAY(dstHisto)),
		cErrInvalidParm);
    }
    else ERR_PASS( w_newHistoLikeTemplate(pIPWState, histoS, &dstHisto));

    memcpy((void *)(HISTO_GET_ARRAY(dstHisto)),
	   (void *)(HISTO_GET_ARRAY(histoS)),
	   HISTO_GET_SIZE(histoS) * sizeof(Int32));

    *histoD = dstHisto;
    NORMAL_RETURN;

} /* CopyHisto */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Write histogram to a file.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  histo::          	histogram
 *  filename::       	name of file in which to write histogram
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Luc Vincent
 *  Peter Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeHisto (
    IPWState           *pIPWState,
    Histo              *histo,
    char               *filename)
{
    PROLOGUE;
    FILE   *histo_stream;
    Int32   maxval, i;
    Int32  *ptrdata;


        /* Open stream where to spit data */
    histo_stream = fopen(filename, "w");
    ERR_GEN(histo_stream == NULL, cErrIO_WriteError);

        /* Write file header */
    fprintf(histo_stream, "#\n#\tHistogram: %s\n#\n\n", filename);

        /* Write file data */
    maxval  = HISTO_GET_MAXVAL(histo);
    i = HISTO_GET_MINVAL(histo);
    ptrdata = HISTO_GET_DATA(histo) + i;
    for (; i <= maxval; i++, ptrdata++)
	fprintf(histo_stream, "%ld\t%ld\n", i, *ptrdata);

    fclose(histo_stream);
    NORMAL_RETURN;
} /* WriteHisto */



/* ===================================================================
**  SmoothHisto()  --  Smoothing of a histogram with moving average
**
**  ARGUMENTS:	histoS - source histogram
**              n - number of pixels in moving window
**              histoD - destination histogram or NULL
**  OPERATION:	As described.
**              Note that if this n is even, then the center
**                  of the window is shifted a bit left (n/2-th pixel)
**  RETURNS:	histoD - destination histogram
**  SEE ALSO:   
**  NOTES:      Extreme values of the histogram are replicated on
**              the sides to minimize aberrations.
=================================================================== */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Smooth a histogram with moving average
 *
 * LongDescription:::
 * Smooth the source histogram with a moving average and place it in
 * the destination histogram.  If the destination histogram is NULL on
 * entry, a new histogram will be allocated.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  histoS::         	source histogram
 *  n::              	number of pixels in moving window
 *
 * Output:::
 *  pHistoD::        	smoothed histogram
 *
 * Returns:::
 *
 * Notes:::
 * Extreme values of the histogram are replicated on the sides to minimize
 * aberrations.
 * If n is even, then the center of the window is shifted a bit left 
 * (n/2-th pixel).
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Luc Vincent
 *  Peter Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_smoothHisto (
    IPWState           *pIPWState,
    Histo              *histoS,
    Int32               n,
    Histo             **pHistoD)
{
    PROLOGUE;
    Int32          *hdataS;
    register Int32 *endS, *endD, *ptr1S, *ptr2S, *ptrD;
    register Int32  left_value, right_value;
    Histo          *histoL = NULL, *histoD;
    register Int32  current_sum;
    Int32           p, q;


        /*----- Check input parameters -----*/
    ERR_GEN(n < 1, cErrInvalidParm);
    ERR_GEN(histoS == NULL, cErrInvalidParm);
    histoD = *pHistoD;
    if (histoD != NULL)
    {
   	ERR_GEN((HISTO_GET_MINVAL(histoS) != HISTO_GET_MINVAL(histoD)) ||
		(HISTO_GET_MAXVAL(histoS) != HISTO_GET_MAXVAL(histoD)),
	        cErrInvalidParm);
  	ERR_GEN(HISTO_GET_SIZE(histoS) != HISTO_GET_SIZE(histoD),
		cErrInvalidParm);
	ERR_GEN((HISTO_GET_DATA(histoS) - HISTO_GET_ARRAY(histoS)) !=
	        (HISTO_GET_DATA(histoD) - HISTO_GET_ARRAY(histoD)),
		cErrInvalidParm);
    }
    else ERR_PASS( w_newHistoLikeTemplate(pIPWState, histoS, &histoD));

        /*----- Allocate temporary histogram if needed -----*/
    if (histoS == histoD)
	ERR_PASS( w_newHistoLikeTemplate(pIPWState, histoS, &histoL));
    else histoL = histoD;

        /*----- Initialize some parameters -----*/
    hdataS = HISTO_GET_DATA(histoS) + HISTO_GET_MINVAL(histoS);
    endS   = HISTO_GET_DATA(histoS) + HISTO_GET_MAXVAL(histoS);

    ptrD   = HISTO_GET_DATA(histoL) + HISTO_GET_MINVAL(histoL);
    endD   = HISTO_GET_DATA(histoL) + HISTO_GET_MAXVAL(histoL);


        /*----- Determines the position of the center of
	        the window. When n is odd, there is no ambiguity.
		However, with even values of n, the center is
		chosen to be on the center-left pixel, i.e:
		            c
		        * * * * * *
		p = number of pixels strictly left of center
		q = number of pixels strictly right of center -----*/
    q = (n >> 1);
    p = (n % 2 ? q : q - 1);


        /*----- Determines start value of current_sum -----*/
    left_value  = *hdataS;
    right_value = *endS;

    current_sum = (p + 1) * left_value;
    ptr2S = MIN(hdataS + q, endS);
    for (ptr1S = hdataS; ptr1S < ptr2S; ptr1S++)
	current_sum += *ptr1S;
    if (ptr2S == endS)
	current_sum += ((hdataS + q - endS) * right_value);


        /*----- Main loop -----*/
    ptr1S = hdataS - p - 1;
    ptr2S = hdataS + q;

    for (; ptrD <= endD; ptrD++, ptr1S++, ptr2S++) {

	current_sum -= (ptr1S < hdataS ? left_value : *ptr1S);
	current_sum += (ptr2S > endS   ? right_value : *ptr2S);

	*ptrD = (current_sum / n);
    }


        /*----- Final clean-up -----*/
    if (histoS == histoD) {
	ERR_PASS_CLEANUP( w_copyHisto(pIPWState, histoL, &histoD));
	ERR_PASS( w_destroyHisto(pIPWState, &histoL));
    }
    *pHistoD = histoD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if ((histoL != NULL) && (histoS == histoD))
	w_destroyHisto(pIPWState, &histoL);
    ERR_RETURN;

}  /* SmoothHisto */

/*
 * $Log: hutils.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:17  jlin
 * Imported sources
 *
 * Revision 1.20  1996/01/04 20:24:18  ejaquith
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.19  1995/12/13  04:09:39  ejaquith
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.18  1995/12/01  19:32:29  ddavies
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Fix acc compiler warnings.
 *
 * Revision 1.17  1995/11/13  20:13:07  ejaquith
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/11/03  22:48:18  ddavies
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.15  1995/10/13  22:02:21  ejaquith
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.14  1995/10/07  01:21:52  ddavies
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Include stdio.h since ipwcore.prv no longer includes it (will be true soon).
 *
 * Revision 1.13  1995/09/29  19:09:35  ejaquith
 * ejaquith on Fri Sep 29 12:09:06 PDT 1995
 *
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * "file i/o" keyword has become "file io"
 *
 * Revision 1.12  1995/09/29  00:15:03  jcraig
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: fixup procedure comments
 *
 * Revision 1.11  1995/09/21  22:53:58  ejaquith
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.10  1995/09/07  01:00:49  jcraig
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * naming convention changes - first letter after w_ is not capitialized
 *
 * Revision 1.9  1995/09/06  20:20:46  ejaquith
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  02:14:20  jcraig
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.7  1995/08/31  21:14:34  ejaquith
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/25  01:36:50  ejaquith
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * Fixed warnings
 *
 * Revision 1.5  1995/08/17  22:31:54  jcraig
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.4  1994/09/30  15:43:16  lperry
 * lperry on Fri Sep 30 08:42:48 PDT 1994
 *
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * Add include of utils.pub to get _abortX defs.
 *
 * Revision 1.3  1994/09/29  00:47:37  lperry
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * Eliminate use of utilpros. Use hutils.pub
 *
 * Revision 1.2  1994/06/08  23:59:31  macleod
 * hutils.c -> /product/ipcore/ipwrap/src/RCS/hutils.c,v
 *
 * Fixed "unused variable" warning.
 *
 * Revision 1.1  1994/06/08  20:32:56  macleod
 * Initial revision
 * 
 */
