/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  intops1.c--procedures that extract integer information from images,
 *             and that require only the pixrect library.
 *
 *           Pixel count:
 *		w_isZero()		determines if image is all zeros
 *		w_isOne()		determines if image is all ones
 *		w_isAllBlack()		determines if image is all black
 *		w_isAllWhite()		determines if image is all white
 *		w_isAllOneColor()	determines if image is all white
 *		w_regionIsZero()	is image region all zeros
 *		w_regionIsOne()		is image region all ones
 *		w_regionIsAllBlack()	is image region all black
 *		w_regionIsAllWhite()	is image region all white
 *		w_regionIsAllOneColor()	is image region all one color
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "intops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: intops1.c,v $; $Revision: 1.2 $; $Date: 2000/10/17 20:31:12 $");


/*************************************************************
 **************            Pixel Count           *************
 ************************************************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image is filled with only 0's
 *
 * LongDescription:::
 *  Goes through an image looking for any non-zero bits.  Returns as soon
 *  as one is found.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be tested.
 *
 * Output:::
 *  pIsZero::        Pointer to Bool that will be set to cTrue if the entire
 *		     image is filled with 0's.  Otherwise, it is set to
 *		     cFalse.
 *
 * Notes:::
 *  Note that this routine does not care about black vs white, only 1's
 *  vs 0's.  It works on gray images as well as black and white images,
 *  but always looks for 1's.  An all 0 binary image is (normally) white.
 *  An all 0 gray image is (normally) all black. See w_isAllBlack() and
 *  w_isAllWhite() for this functionality. See also w_regionIsZero().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_isZero (
    IPWState           *pIPWState,
    PIXR               *prS,
    Bool               *pIsZero)
{
    PROLOGUE;
    Int32             h, w, d, bpl;
    register Int32    imageWordsPerLine;    /* Whole words per line */
    register Int32    totalWordsPerLine;    /* Words per line (frame incl) */
    register UInt32  *pRow, *pWord, inputData;
    UInt8            *pImage;

    ERR_GEN(prS == NULL || pIsZero == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, prS, &pImage));

    totalWordsPerLine = bpl / 4;
    imageWordsPerLine = ((w * d) + 31) / 32;
    pRow = (UInt32 *) pImage;
    inputData = 0;

    for (; h > 0; pRow += totalWordsPerLine, h--)
    {
        pWord = pRow + imageWordsPerLine;	/* Start at right edge */
        while (--pWord >= pRow)			/* End at left edge */
            inputData |= *pWord;

        if (inputData != 0)
	{
	    *pIsZero = cFalse;
            NORMAL_RETURN;
	}
    }

    *pIsZero = cTrue;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image is filled with only 1's (binary) or 0xff (gray)
 *
 * LongDescription:::
 *  Goes through an image looking for any zero bits.  Returns as soon
 *  as one is found.
 *
 * Input:::
 *  pIPWState::     Pointer to standard IPWrap state block.
 *  prS::           Pointer to image to be tested.
 *
 * Output:::
 *  pIsOne::	    Pointer to Bool that will be set to cTrue if the entire
 *		    image is filled with 1 bits.  Otherwise, it is set to
 *		    cFalse.
 *
 * Notes:::
 *  Note that this routine does not care about black vs white, only 1's
 *  vs 0's.  It works on gray images as well as binary images, but always
 *  searches for 0 bits.  An all 0 binary image is (normally) white.
 *  An all 0 gray image is (normally) all black. See w_isAllBlack() and
 *  w_isAllWhite() for this functionality. See also w_regionIsOne().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_isOne (
    IPWState           *pIPWState,
    PIXR               *prS,
    Bool               *pIsOne)
{
    PROLOGUE;
    Int32             h, w, d, bpl;
    register Int32    imageWordsPerLine;    /* Whole words per line */
    register Int32    totalWordsPerLine;    /* Physical words (frame incl) */
    register UInt32  *pRow, *pWord, inputData;
    UInt8            *pImage;
    UInt32	      trailingBits;
    UInt32	      mask;

    ERR_GEN(prS == NULL || pIsOne == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, prS, &pImage));

    pRow = (UInt32 *) pImage;
    inputData = 0xffffffff;
    trailingBits = (w * d) % 32;
    mask = (0xffffffff >> (32 - trailingBits)) << (32 - trailingBits);
    totalWordsPerLine = bpl / 4;
    imageWordsPerLine = ((w * d) + 31) / 32;

    for (; h > 0; pRow += totalWordsPerLine, h--)
    {
        pWord = pRow + imageWordsPerLine;	/* Start at right edge */
	if (trailingBits)
	{
	    if ((*--pWord & mask) != mask)
	    {
		*pIsOne = cFalse;
		NORMAL_RETURN;
	    }
	}
        while (--pWord >= pRow)			/* End at left edge */
            inputData &= *pWord;

        if (inputData != 0xffffffff)
	{
	    *pIsOne = cFalse;
            NORMAL_RETURN;
	}

    }

    *pIsOne = cTrue;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image is filled with only a single color (0/1's if binary)
 *
 * LongDescription:::
 *  Picks the first color found in an image, then goes through the rest of the
 *  image looking for any other color.  Returns as soon as one is found, or
 *  returns with the color if none other found.
 *
 * Input:::
 *  pIPWState::     Pointer to standard IPWrap state block.
 *  prS::           Pointer to image to be tested.
 *
 * Output:::
 *  pIsOne::	    Pointer to Bool that will be set to cTrue if the entire
 *		    image is filled with a single color.  Otherwise, it is
 *		    set to cFalse.
 *  ppColor::	    The color (must be pointer to array of 1 or 3 if
 *		    binary/gray or RGB, respectively).
 *
 * Notes:::
 *  Note that this routine works on color, gray and binary images.
 *  See also w_regionIsAllOneColor().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_isAllOneColor (
    IPWState           *pIPWState,
    PIXR               *prS,
    Bool               *pIsOneColor,
    UInt8	      **ppColor)
{
    PROLOGUE;
    Int32             h, w, d, bpl, c;
    register Int32    imageWordsPerLine;    /* Whole words per line */
    register Int32    totalWordsPerLine;    /* Physical words (frame incl) */
    register UInt32  *pRow, *pWord;
    UInt8            *pImage;
    UInt32	      trailingBits;
    UInt32	      mask;
    UInt32	      colorWord = 0UL;
    UInt8	      color = 0;

    ERR_GEN(prS == NULL || pIsOneColor == NULL || ppColor == NULL || *ppColor == NULL,
	cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1 | cChannels3, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    if (d == 1) {
	ERR_PASS( w_isOne(pIPWState, prS, pIsOneColor));
	if (*pIsOneColor) {
	    **ppColor = 1;
	    NORMAL_RETURN;
	}
	ERR_PASS( w_isZero(pIPWState, prS, pIsOneColor));
	if (*pIsOneColor) {
	    **ppColor = 0;
	}
	NORMAL_RETURN;
    }

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &c));
    if (c > 1) {
	Int32 channel = 0;
	PIXR *prSub;
	UInt8 *colorPtr = *ppColor;

	while (channel < c) {
	    ERR_PASS( w_pixrGetChannel(pIPWState, prS, channel, &prSub));
	    ERR_PASS( w_isAllOneColor (pIPWState, prSub, pIsOneColor, &colorPtr));
	    if (*pIsOneColor == 0) break;
	    channel++;
	    colorPtr++;
	}

	NORMAL_RETURN;
    }

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, prS, &pImage));

    pRow = (UInt32 *) pImage;
    trailingBits = (w * d) % 32;
    mask = (0xffffffff >> (32 - trailingBits)) << (32 - trailingBits);
    totalWordsPerLine = bpl / 4;
    imageWordsPerLine = ((w * d) + 31) / 32;

    if (w && h) {
	/* Pick off the first color */
	color = (UInt8)(*pRow >> (32 - d));

	/* Replicate it to fill out a word */
	{
	    Int32 bitsFilled = 0;

	    while (bitsFilled < 32) {
		colorWord = (colorWord << d) | color;
		bitsFilled += d;
	    }
	}
    }

    for (; h > 0; pRow += totalWordsPerLine, h--)
    {
        pWord = pRow + imageWordsPerLine;	/* Start at right edge */
	if (trailingBits)
	{
	    if ((*--pWord & mask) != (colorWord & mask))
	    {
		*pIsOneColor = cFalse;
		NORMAL_RETURN;
	    }
	}
        while (--pWord >= pRow)			/* End at left edge */
	{
	    if (*pWord != colorWord)
	    {
		*pIsOneColor = cFalse;
		NORMAL_RETURN;
	    }
	}
    }

    *pIsOneColor = cTrue;
    **ppColor = color;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image is all black
 *
 * LongDescription:::
 *  Uses w_isZero() and w_isOne() to determine if an image is all black.
 *  Does take into account that binary black is 1's, and non-binary black
 *  is 0's.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be tested.
 *
 * Output:::
 *  pIsAllBlack::    Pointer to Bool that will be set to cTrue if the entire
 *		     image is black, otherwise it is set to cFalse.
 *
 * Notes:::
 *  See also w_regionIsAllBlack().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_isAllBlack (
    IPWState           *pIPWState,
    PIXR               *prS,
    Bool               *pIsAllBlack)
{
    Int32	d;
    PROLOGUE;

    ERR_GEN(prS == NULL || pIsAllBlack == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));

    switch (d) {
    case 1:
	ERR_PASS( w_isOne(pIPWState, prS, pIsAllBlack));
	break;
    case 2:
    case 4:
    case 8:
	ERR_PASS( w_isZero(pIPWState, prS, pIsAllBlack));
	break;
    }
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image is all white
 *
 * LongDescription:::
 *  Uses w_isZero() and w_isOne() to determine if an image is all white.
 *  Does take into account that binary white is 0's, and non-binary white
 *  is 1's.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be tested.
 *
 * Output:::
 *  pIsAllWhite::    Pointer to Bool that will be set to cTrue if the entire
 *		     image is white, otherwise it is set to cFalse.
 *
 * Notes:::
 *  See also w_regionIsAllWhite().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_isAllWhite (
    IPWState           *pIPWState,
    PIXR               *prS,
    Bool               *pIsAllWhite)
{
    Int32	d;
    PROLOGUE;

    ERR_GEN(prS == NULL || pIsAllWhite == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));

    switch (d) {
    case 1:
	ERR_PASS( w_isZero(pIPWState, prS, pIsAllWhite));
	break;
    case 2:
    case 4:
    case 8:
	ERR_PASS( w_isOne(pIPWState, prS, pIsAllWhite));
	break;
    }
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image region is filled with only 0's
 *
 * LongDescription:::
 *  Goes through an image looking for any non-zero bits in a region of an
 *  image.  Returns as soon as one is found.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be tested.
 *  xOffset::	     X offset of first pixel in ROI (region-of-interest)
 *  yOffset::	     Y offset of first pixel in ROI
 *  width::	     Width of ROI
 *  height::	     Height of ROI
 *
 * Output:::
 *  pIsZero::        Pointer to Bool that will be set to cTrue if the entire
 *		     image is filled with 0's.  Otherwise, it is set to
 *		     cFalse.
 *
 * Notes:::
 *  Note that this routine does not care about black vs white, only 1's
 *  vs 0's.  It works on gray images as well as black and white images,
 *  but always looks for 1's.  An all 0 binary image is (normally) white.
 *  An all 0 gray image is (normally) all black. See w_regionIsAllBlack() and
 *  w_regionIsAllWhite() for this functionality. See also w_isZero().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_regionIsZero (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32		xOffset,
    Int32		yOffset,
    Int32		width,
    Int32		height,
    Bool               *pIsZero)
{
    PROLOGUE;
    Int32             h, w, d, bpl;
    register Int32    imageWordsPerLine;    /* Whole words per line */
    register Int32    totalWordsPerLine;    /* Words per line (frame incl) */
    register UInt32  *pRow, *pWord, inputData;
    UInt8            *pImage;
    Int32	      bitStart;		    /* Bit offset of start */
    Int32	      bitEnd;		    /* Bit offset just past the end */
    Int32	      leadingBits, trailingBits;
    UInt32	      leadingMask, trailingMask;

    ERR_GEN(prS == NULL || pIsZero == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);
    ERR_GEN(xOffset < 0 || yOffset < 0, cErrOutOfBounds);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, prS, &pImage));

    ERR_GEN(xOffset + width > w || yOffset + height > h, cErrOutOfBounds);

    /* Only look at bits in the region */
    bitStart = xOffset * d;
    bitEnd = bitStart + width * d;
    leadingBits = (bitStart & 0x1f)? 32 - (bitStart & 0x1f): 0;
    leadingMask = (0xffffffff << (32 - leadingBits)) >> (32 - leadingBits);
    trailingBits = bitEnd & 0x1f;
    trailingMask = (0xffffffff >> (32 - trailingBits)) << (32 - trailingBits);

    /* If start and stop bits cut into the same word, adjust trailing mask,
     * and avoid the leading mask. */
    if (leadingBits + trailingBits > width * d) {
	trailingBits = width * d;
	leadingBits = 0;
	trailingMask &= leadingMask;
	leadingMask = 0;
	imageWordsPerLine = 1;
    } else {
	imageWordsPerLine = (width * d - trailingBits - leadingBits) / 32 +
	    (leadingBits? 1: 0) + (trailingBits? 1: 0);
    }
    totalWordsPerLine = bpl / 4;
    pRow = (UInt32 *) pImage + yOffset * totalWordsPerLine + bitStart / 32;

    inputData = 0;

    for (; height > 0; pRow += totalWordsPerLine, height--)
    {

        pWord = pRow + imageWordsPerLine;	/* Start at right edge */
	if (trailingBits)		/* Check partial word at end */
	    inputData |= *--pWord & trailingMask;
        while (--pWord > pRow)			/* End before left edge */
            inputData |= *pWord;

	if (leadingBits) {		/* Check partial word at start */
	    inputData |= *pWord & leadingMask;
	} else if (pWord == pRow) {
            inputData |= *pWord;	/* else do whole word at edge */
	}

        if (inputData != 0)
	{
	    *pIsZero = cFalse;
            NORMAL_RETURN;
	}
    }

    *pIsZero = cTrue;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image region is filled with only 1's (binary) or
 *  0xff (gray)
 *
 * LongDescription:::
 *  Goes through an image region looking for any zero bits.  Returns as soon
 *  as one is found.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::	     Pointer to image to be tested.
 *  xOffset::	     X offset of first pixel in ROI (region-of-interest)
 *  yOffset::	     Y offset of first pixel in ROI
 *  width::	     Width of ROI
 *  height::	     Height of ROI
 *
 * Output:::
 *  pIsOne::	     Pointer to Bool that will be set to cTrue if any of the
 *		     image region is filled with 1 bits.  Otherwise, it is
 *		     set to cFalse.
 *
 * Notes:::
 *  Note that this routine does not care about black vs white, only 1's
 *  vs 0's.  It works on gray images as well as binary images, but always
 *  searches for 0 bits.  An all 0 binary image is (normally) white.
 *  An all 0 gray image is (normally) all black. See w_regionIsAllBlack() and
 *  w_regionIsAllWhite() for this functionality. See also w_isOne().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_regionIsOne (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32		xOffset,
    Int32		yOffset,
    Int32		width,
    Int32		height,
    Bool               *pIsOne)
{
    PROLOGUE;
    Int32             h, w, d, bpl;
    register Int32    imageWordsPerLine;    /* Whole words per line */
    register Int32    totalWordsPerLine;    /* Physical words (frame incl) */
    register UInt32  *pRow, *pWord, inputData;
    UInt8            *pImage;
    Int32	      bitStart;		    /* Bit offset of start */
    Int32	      bitEnd;		    /* Bit offset just past the end */
    Int32	      leadingBits, trailingBits;
    UInt32	      leadingMask, trailingMask;

    ERR_GEN(prS == NULL || pIsOne == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);
    ERR_GEN(xOffset < 0 || yOffset < 0, cErrOutOfBounds);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, prS, &pImage));

    ERR_GEN(xOffset + width > w || yOffset + height > h, cErrOutOfBounds);

    /* Only look at bits in the region */
    bitStart = xOffset * d;
    bitEnd = bitStart + width * d;
    leadingBits = (bitStart & 0x1f)? 32 - (bitStart & 0x1f): 0;
    leadingMask = (0xffffffff << (32 - leadingBits)) >> (32 - leadingBits);
    trailingBits = bitEnd & 0x1f;
    trailingMask = (0xffffffff >> (32 - trailingBits)) << (32 - trailingBits);

    /* If start and stop bits cut into the same word, adjust trailing mask,
     * and avoid the leading mask. */
    if (leadingBits + trailingBits > width * d) {
	trailingBits = width * d;
	leadingBits = 0;
	trailingMask &= leadingMask;
	leadingMask = 0;
	imageWordsPerLine = 1;
    } else {
	imageWordsPerLine = (width * d - trailingBits - leadingBits) / 32 +
	    (leadingBits? 1: 0) + (trailingBits? 1: 0);
    }
    totalWordsPerLine = bpl / 4;
    pRow = (UInt32 *) pImage + yOffset * totalWordsPerLine + bitStart / 32;

    inputData = 0xffffffff;

    for (; height > 0; pRow += totalWordsPerLine, height--)
    {

        pWord = pRow + imageWordsPerLine;	/* Start at right edge */
	if (trailingBits) {	/* Check partial word at end */
	    if ((*--pWord & trailingMask) != trailingMask)
	    {
		*pIsOne = cFalse;
		NORMAL_RETURN;
	    }
	}
        while (--pWord > pRow)			/* End before left edge */
            inputData &= *pWord;

	if (leadingBits) {		/* Check partial word at start */
	    if ((*pWord & leadingMask) != leadingMask)
	    {
		*pIsOne = cFalse;
		NORMAL_RETURN;
	    }
	} else if (pWord == pRow) {
            inputData &= *pWord;	/* else do whole word at edge */
	}

        if (inputData != 0xffffffff)
	{
	    *pIsOne = cFalse;
            NORMAL_RETURN;
	}
    }

    *pIsOne = cTrue;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image is all black
 *
 * LongDescription:::
 *  Uses w_isZero() and w_isOne() to determine if an image is all black.
 *  Does take into account that binary black is 1's, and non-binary black
 *  is 0's.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be tested.
 *  xOffset::	     X offset of first pixel in ROI (region-of-interest)
 *  yOffset::	     Y offset of first pixel in ROI
 *  width::	     Width of ROI
 *  height::	     Height of ROI
 *
 * Output:::
 *  pIsAllBlack::    Pointer to Bool that will be set to cTrue if the entire
 *		     image is black, otherwise it is set to cFalse.
 *
 * Notes:::
 *  See also w_isAllBlack().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_regionIsAllBlack (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32		xOffset,
    Int32		yOffset,
    Int32		width,
    Int32		height,
    Bool               *pIsAllBlack)
{
    Int32	d;
    PROLOGUE;

    ERR_GEN(prS == NULL || pIsAllBlack == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));

    switch (d) {
    case 1:
	ERR_PASS( w_regionIsOne(pIPWState, prS, xOffset, yOffset,
			  width, height, pIsAllBlack));
	break;
    case 2:
    case 4:
    case 8:
	ERR_PASS( w_regionIsZero(pIPWState, prS, xOffset, yOffset,
			   width, height, pIsAllBlack));
	break;
    }
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if an image is all white
 *
 * LongDescription:::
 *  Uses w_isZero() and w_isOne() to determine if an image is all white.
 *  Does take into account that binary white is 0's, and non-binary white
 *  is 1's.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be tested.
 *  xOffset::	     X offset of first pixel in ROI (region-of-interest)
 *  yOffset::	     Y offset of first pixel in ROI
 *  width::	     Width of ROI
 *  height::	     Height of ROI
 *
 * Output:::
 *  pIsAllWhite::    Pointer to Bool that will be set to cTrue if the entire
 *		     image is white, otherwise it is set to cFalse.
 *
 * Notes:::
 *  See also w_isAllWhite().
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  histogram, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_regionIsAllWhite (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32		xOffset,
    Int32		yOffset,
    Int32		width,
    Int32		height,
    Bool               *pIsAllWhite)
{
    Int32	d;
    PROLOGUE;

    ERR_GEN(prS == NULL || pIsAllWhite == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));

    switch (d) {
    case 1:
	ERR_PASS( w_regionIsZero(pIPWState, prS, xOffset, yOffset,
				 width, height, pIsAllWhite));
	break;
    case 2:
    case 4:
    case 8:
	ERR_PASS( w_regionIsOne(pIPWState, prS, xOffset, yOffset,
			  width, height, pIsAllWhite));
	break;
    }
    NORMAL_RETURN;
}

/*

 $Log: intops1.c,v $
 Revision 1.2  2000/10/17 20:31:12  dabercro
 New function to check if a PIXR is all one color.

 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.27  1998/03/03 23:38:38  aber
 aber on Tue Mar 3 15:37:02 PST 1998

 intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v

 Fixed edge conditions in image check routines. In the case of a partial
 word check, the check was going off the edge of the region.
 Also, in the case of a partial word check that didn't extend to the
 word boundaries, a special case needed to be added.

 Revision 1.26  1998/02/18 21:45:52  aber
 aber on Wed Feb 18 13:44:14 PST 1998

 intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v

 Add more comments. Fix logic causing first word to be checked twice in
 w_isRegionAll One/Zero. The first check wasn't using mask, used for end
 conditions (bug).

 Revision 1.25  1998/01/12 22:43:14  aber
 aber on Mon Jan 12 14:42:22 PST 1998

 intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v

 Boolean expression fix in w_regionIsOne(), OR should have been AND.

 Revision 1.24  1997/11/24 23:58:19  aber
 aber on Mon Nov 24 15:58:05 PST 1997

 intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v

 Add new functions to take into account an image region: w_regionIsZero(),
 w_regionIsOne(), w_regionIsAllBlack(), w_regionIsAllWhite().

 Revision 1.23  1997/08/27 00:15:56  aber
 intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v

 Update w_isZero to handle the right-edge boundary condition correctly.
 Added three new functions: w_isOne, which checks for all 1's instead
 of 0's; w_isAllBlack, which correctly calls w_isZero or w_isOne;
 and w_isAllWhite, which correctly calls w_isOne or w_isZero. Oops,
 in the middle of this note, Dan corrects me, so I've change the
 boundary check back on w_isZero. All frames must always have 0's.

 Revision 1.22  1996/01/04 20:24:23  ejaquith
 intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.21  1995/12/13  04:09:47  ejaquith
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:13:15  ejaquith
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:02:28  ejaquith
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/10/05  04:27:00  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.17  1995/10/03  01:19:53  ddavies
 * ddavies on Mon Oct 2 18:19:52 PDT 1995
 *
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.16  1995/09/21  22:54:05  ejaquith
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/15  18:00:20  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.14  1995/09/11  22:11:00  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.13  1995/09/10  00:33:03  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.12  1995/09/06  20:20:52  ejaquith
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  01:16:39  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:14:40  ejaquith
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/25  01:36:59  ejaquith
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.8  1995/08/17  22:02:32  ejaquith
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.7  1995/07/29  04:41:29  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.6  1994/10/06  20:54:24  lperry
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Change to use new .pub header file instead of ippros.h
 *
 * Revision 1.5  1994/06/08  04:11:17  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.4  1994/02/09  20:32:36  lperry
 * lperry on Wed Feb 9 12:30:42 PST 1994
 *
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Fixed offset problem. Was looking one 32 bit word to the
 * right of where it should look.
 *
 * Revision 1.3  1993/12/03  01:16:28  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:13:20  ddavies
 * intops1.c -> /product/ipcore/ipwrap/src/RCS/intops1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

