/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  intops2.c--procedures that extract integer information from images,
 *             and that require only the pixrect library.
 *
 *           Pixel count:
 *		w_numberBits()	returns number of bits in an
 *				image with or without frame.
 *
 *		w_numberBitsRect() returns number of bits inside a rectangle
 *				(relative to the upper left corner of the
 *				image) of an image.
 *
 *		w_numberBitsFrameRect() returns number of bits inside a 
 *				rectangle (relative to the upper left 
 *				corner of the frame) of an 
 *				image. 
 *
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "rectops.pub"
#include "intops.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: intops2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Count number of 1's in an image.
 *
 * LongDescription:::
 *  This procedure counts the 1's in an entire image.  It does not
 *  try to distinguish between gray and binary images.  It does not
 *  worry about the values of white and black - it just counts the
 *  1's.  Among other things, it's a cheap (and not very accurate)
 *  way of comparing two images.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be inspected.
 *
 * Output:::
 *  pNumBits::       Pointer to returned number of 1's in prS.
 *
 * Notes:::
 *  This function can deal with framed or frameless images.  In either case,
 *  it counts the number of bits in the image portion only (not in
 *  the frame).
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_numberBits (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32              *pNumBits)
{
    PROLOGUE;
    Int32    w, h, d, bpl, f, num, numBits = 0, channel, channels;
    UInt8   *pImage;
    PIXR    *ch;

    ERR_GEN(prS == NULL || pNumBits == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepthAny,
		 cChannels1 | cChannels3, cFrameAny);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &f));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));

    for (channel = 0; channel < channels; channel++)
    {
	ERR_PASS( w_pixrGetChannel (pIPWState, prS, channel, &ch));
	ERR_PASS( w_pixrGetImage(pIPWState, ch, &pImage));

	/* Images without frames might not have pitches that are
	 * multiples of 4.  If the image doesn't have a frame,
	 * pretend it only has one very wide line and count it
	 * all at once.
	 * BUG: But images without frames MIGHT still be padded
	 * (with non-zero bits) to multiples of 4 or 8, so
	 * that doesn't always work.  Only do the "all one line"
	 * trick if there's no frame and the bpl is not a 
	 * multiple of 4.
	 */
	if ((f != 0)  || ((bpl % 4) == 0))
	{
	    /* let i_countBits do the real work */
	    ERR_PASS( i_countBits (pIPWState, pImage, 0, 0, w, h,
				   d, bpl, &num));
	}
	else	/* this is a weird image */
	{
	    /* let i_countBits do the real work */
	    ERR_PASS( i_countBits (pIPWState, pImage, 0, 0,
				   bpl*h, /*w in bytes*/
				   1, /*h*/
				   8, /*d*/
				   0, /*bpl*/
				   &num));
	}

	numBits += num;
    }
    
    *pNumBits = numBits;
    NORMAL_RETURN; 
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Count the number of 1's in a sub-rectangle of an image.
 *
 * LongDescription:::
 *  This procedure counts the 1's in a sub-rectangle of an image.  It does not
 *  try to distinguish between gray and binary images.  It does not
 *  worry about the values of white and black - it just counts the
 *  1's.  Among other things, it's a cheap (and not very accurate)
 *  way of comparing portions of images.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be inspected.
 *  pCountRect::     Pointer to RECTANGL struct describing sub-rect of
 *		     prS.  Only pixels inside this rectangle will be
 *		     inspected.
 *
 * Output:::
 *  pNumBits::       Pointer to returned number of 1's inside the sub-
 *		     rectangle of prS.
 *
 * Notes:::
 *  This function can deal with framed images only.  The entire rectangle
 *  must lie within the frame.  See w_numberBitsFrameRect for to count
 *  bits both inside and outside a frame.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies?
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_numberBitsRect (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pCountRect,
    Int32              *pNumBits)
{
    PROLOGUE;
    Int32	 frameBits, d, bpl, num, numBits = 0, channel, channels;
    UInt8	*pImage;
    Int8	 valid;
    PIXR        *ch;

    ERR_GEN(prS == NULL || pCountRect == NULL || pNumBits == NULL,
	    cErrPixr_null);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepthAny,
		 cChannels1 | cChannels3, cFrame32);
    ERR_PASS( w_isRectValid (pIPWState, prS, pCountRect, &valid));
    ERR_GEN( !valid, cErrOutOfBounds);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &frameBits));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));

    for (channel = 0; channel < channels; channel++)
    {
	ERR_PASS( w_pixrGetChannel (pIPWState, prS, channel, &ch));
	ERR_PASS( w_pixrGetImageAll(pIPWState, ch, &pImage));

	/* let i_countBits do the real work */
	ERR_PASS( i_countBits (pIPWState, pImage,
			       pCountRect->x + frameBits,   /* x offset */
			       pCountRect->y + frameBits,   /* y offset */
			       pCountRect->w, pCountRect->h,/* width / height */
			       d, bpl,
			       &num));
	numBits += num;
    }
    
    *pNumBits = numBits;
    NORMAL_RETURN; 
}


/* numberBitsFrameRect(): Count the number of 1's inside the specified
 *			rectangle in ANY image.  The
 *			rectangle is relative to the upper left corner
 *			of the frame.
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Count the number of 1's in a sub-rectangle of an image.
 *
 * LongDescription:::
 *  This procedure counts the 1's in a sub-rectangle of an image.  It does not
 *  try to distinguish between gray and binary images.  It does not
 *  worry about the values of white and black - it just counts the
 *  1's.  Among other things, it's a cheap (and not very accurate)
 *  way of comparing portions of images.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be inspected.
 *  pCountRect::     Pointer to RECTANGL struct describing sub-rect of
 *		     prS.  Only pixels inside this rectangle will be
 *		     inspected.
 *
 * Output:::
 *  pNumBits::       Pointer to returned number of 1's inside the sub-
 *		     rectangle of prS.
 *
 * Notes:::
 *  Like w_numberBitsRect, this function can deal with framed images only.
 *  Unlike w_numberBitsRect, the rectangle may include frame bits.
 *  This routine is used primarily to detect erroneous writes to the frame.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies?
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_numberBitsFrameRect (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pCountRect,
    Int32              *pNumBits)
{
    PROLOGUE;
    Int32	 d, bpl, num, numBits = 0, channel, channels;
    UInt8	*pImage;
    PIXR        *ch;

    ERR_GEN(prS == NULL || pCountRect == NULL || pNumBits == NULL,
	    cErrPixr_null);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepthAny,
		 cChannels1 | cChannels3, cFrame32);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));

    for (channel = 0; channel < channels; channel++)
    {
	ERR_PASS( w_pixrGetChannel (pIPWState, prS, channel, &ch));
	ERR_PASS( w_pixrGetImageAll(pIPWState, ch, &pImage));

	/* let i_countBits do the real work */
	ERR_PASS(i_countBits (pIPWState, pImage,
			      pCountRect->x, pCountRect->y,/* x and y offset */
			      pCountRect->w, pCountRect->h,/* width / height */
			      d, bpl,
			      &num));

	numBits += num;
    }
    
    *pNumBits = numBits;
    NORMAL_RETURN; 
}



/*

 $Log: intops2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.33  1998/02/10 19:58:00  ddavies
 ddavies on Tue Feb 10 11:57:03 PST 1998

 intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v

 Change depth checks in bit counting routines from individually checking
 for depths 1, 2, 4 and 8 to checking for cDepthAny.  This was done to
 allow 16 bit/pixel images.

 Revision 1.32  1998/01/14 00:13:21  ddavies
 intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.31  1997/04/15 22:11:31  ddavies
 intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v

 Fix bug in w_numberBits.  This routine now only reduces an image to one
 lone scanline (for the purposes of counting bits) if it has no frame
 AND it's bytes/line is not a multiple of 4.  It used to only look
 at the frame size to decide.  However, frameless, padded images confused
 it.

 Revision 1.30  1996/01/04 20:24:26  ejaquith
 intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.29  1995/12/13  04:09:51  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.28  1995/11/13  20:13:19  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.27  1995/10/13  22:02:31  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.26  1995/10/12  04:25:15  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.25  1995/10/05  04:27:03  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.24  1995/10/03  01:19:58  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.23  1995/09/21  22:54:07  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.22  1995/09/14  18:56:26  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Made routines w_numberBits(), w_numberBitsRect(), and
 * w_numberBitsFrameRect() all be able to properly take multi-channeled
 * images -- that is, they now count the number of bits in ALL the planes
 * instead of just the "primary" one.  (These weren't originally intended to
 * be used for multi-channel images, but we are using them as such in the
 * regression tests, so it made sense to add the functionality).  Updated
 * comments and the top of and throughout the file.
 *
 * Revision 1.21  1995/09/13  18:30:00  ddavies
 * ddavies on Wed Sep 13 11:29:49 PDT 1995
 *
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Reindent prologues.
 *
 * Revision 1.20  1995/09/13  00:21:30  ejaquith
 * ejaquith on Tue Sep 12 17:21:11 PDT 1995
 *
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Pulled w_isRectValid() check for when using frame.
 *
 * Revision 1.19  1995/09/11  22:11:04  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.18  1995/09/06  20:20:55  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.17  1995/09/01  01:16:44  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.16  1995/08/31  21:14:44  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.15  1995/08/25  01:37:02  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Fixed warnings
 *
 * Revision 1.14  1995/08/17  22:02:34  ejaquith
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.13  1995/07/29  04:41:53  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.12  1994/10/06  20:54:27  lperry
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Revision 1.11  1994/08/30  00:26:15  jcraig
 * jcraig on Mon Aug 29 17:25:30 PDT 1994
 *
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * add numberBitsFrameRect which is like numberBitsRects except that
 * the rectangle is relative to the upper left corner of the frame not the
 * image.
 *
 * Revision 1.10  1994/08/29  18:53:49  jcraig
 * jcraig on Mon Aug 29 11:52:28 PDT 1994
 *
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * numberBitsRect now takes a RECTANGL instead of x, y, w, and h
 * separately.  the rectangle is relative to the upper left corner of
 * the image NOT the frame.
 *
 * Revision 1.9  1994/06/08  04:11:28  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.8  1994/05/26  19:21:42  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Add #ifndef PRODUCTION around declarations of procName in numberBits and
 * numberBitsRect.
 *
 * Revision 1.7  1993/12/22  05:01:47  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Change numberBits so that it works properly on images whose bytes/line
 * is not a multiple of 4.  This is used to check frameless images.
 *
 * Revision 1.6  1993/12/03  01:16:35  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.5  1993/11/16  22:00:17  ddavies
 * ddavies on Tue Nov 16 14:00:03 PST 1993
 *
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Include shrip.pub to get prototype for i_countBits.
 *
 * Revision 1.4  1993/11/16  03:16:38  ddavies
 * ddavies on Mon Nov 15 19:16:22 PST 1993
 *
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Add implementation of numberBitsRect.  This routine counts the 1's in a
 * rectangle of a PIXR.  The coordinates are relative to the upper left corner
 * of the (outside of the) frame.
 *
 * Revision 1.3  1993/11/13  04:20:02  ddavies
 * ddavies on Fri Nov 12 20:19:15 PST 1993
 *
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Remove makeSumPixelTab since we don't make tables for counting bits anymore,
 * we include a premade one.  Convert numberBits into a really dumb wrapper for
 * i_countBits in ipshared.
 *
 * Revision 1.2  1993/11/06  03:13:27  ddavies
 * intops2.c -> /product/ipcore/ipwrap/src/RCS/intops2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

