/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  intops3.c--includes procedures that extract integer information
 *             from images, and that require morphological or logical
 *             operations.
 *		This is stuff that is used by XIS or for testing images.
 *
 *         Equality between two images:
 *	            w_areEqual()   returns cTrue if two images are identical
 *
 *	   Difference between two images:
 *		    w_absValDifference()	conditionally creates dest
 *						image that contains the
 *						absolute value of the 
 *						difference between two gray 
 *						or binary images.
 *
 *	   Produce binary image marking where pixels of one gray image
 *	   are greater than pixels of another gray image
 *		     w_compareGrayImages()
 *
 *	   Min/Max of two images at each pixel:
 *		     w_min()			Take the minimum 
 *		     w_max()			Take the maximum
 *
 *	    Image arithmetic:
 *		    w_addGray()			Produce an image by
 *						adding the pixels in one
 *						image from those in another.
 *						Pixels are clipped to [0..255].
 *		    w_addGrayNoClip()		Like addGray except there is
 *						to check for pixels > 255..
 *		    w_subtractGray()		Produce an image by
 *						subtracting the pixels in one
 *						image from those in another.
 *						Negative pixels values are
 *						set to 0.
 *		    w_subtractGrayNoClip()	Like subtractGray except
 *						negative pixels values could
 *						be returned.
 *		    w_averageGray()		Produce pixel-by-pixel average
 *						of two images.
 *
 *	    Compute range (lowest, highest):
 *		    w_findRange()		Find minimum and maximum
 *						pixel values in an image.
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "intops.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: intops3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determines if two images are identical.
 *
 * LongDescription:::
 *  This procedure goes through two images, comparing pixels.  If they are
 *  identical, the routine returns cTrue, otherwise cFalse.
 *  Images may be any number of channels/planes, each depth 1,2,4, or 8.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pr1::            Pointer to first input image.
 *  pr2::            Pointer to second input image.
 *
 * Output:::
 *  pAreEqual::      Pointer to returned Bool.  The Bool is set to cTrue
 *		     if the images are equal and cFalse if they are not
 *		     equal.
 *
 * Notes:::
 *  This routine stops as soon as it finds a difference.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg?  Modified by D. Davies to compare entire lines at once
 *  instead of branching after each word compared.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_areEqual (
    IPWState           *pIPWState,
    PIXR               *pr1,
    PIXR               *pr2,
    Bool               *pAreEqual)
{
    PROLOGUE;
    Int32         eq, w, currentOffset, lastWordOffset, h, d, wpl, tmp;
    Int32         channel, nChannels;
    UInt32       *pLine1, *pLine2;
    UInt8        *pImage1, *pImage2;
    PIXR         *ch1, *ch2;



    ERR_GEN(pr1 == NULL || pr2 == NULL || pAreEqual == NULL, cErrNullPtr);
    ERR_GEN_PIXR(pr1, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);
    ERR_GEN_PIXR(pr2, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

	/* get the geometry of the images */
    ERR_PASS( w_pixrGetWidth(pIPWState, pr1, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, pr1, &h));
    ERR_PASS( w_pixrGetDepth(pIPWState, pr1, &d));
    ERR_PASS( w_pixrGetBpl(pIPWState, pr1, &tmp));
    ERR_PASS( w_pixrGetChannelCount (pIPWState, pr1, &nChannels));

    wpl = tmp / 4;

	/* do the geometries match? */
    ERR_PASS( w_pixrGetWidth(pIPWState, pr2, &tmp));
    if (tmp != w)
    {
	*pAreEqual = cFalse;
	NORMAL_RETURN;
    }
    
    
    ERR_PASS( w_pixrGetHeight(pIPWState, pr2, &tmp));
    if (tmp != h)
    {
	*pAreEqual = cFalse;
	NORMAL_RETURN;
    }
    
    ERR_PASS( w_pixrGetDepth(pIPWState, pr2, &tmp));
    if (tmp != d)
    {
	*pAreEqual = cFalse;
	NORMAL_RETURN;
    }
    
    ERR_PASS( w_pixrGetBpl(pIPWState, pr2, &tmp));
    if (tmp/4 != wpl)
    {
	*pAreEqual = cFalse;
	NORMAL_RETURN;
    }

    ERR_PASS( w_pixrGetChannelCount (pIPWState, pr2, &tmp));
    if (tmp != nChannels)
    {
	*pAreEqual = cFalse;
	NORMAL_RETURN;
    }
    
	

	/* if the width are 0, they match.  We need to eliminate
	 * this case for the calculations below to avoid negative
	 * numbers.
	 */
    if (w == 0)
    {
	*pAreEqual = cTrue;
	NORMAL_RETURN;
    }

	/* set the offset of the last image word in each line */
    lastWordOffset = ((w * d) - 1) / 32;

    for (channel = 0; channel < nChannels; channel++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, pr1, channel, &ch1));
	ERR_PASS( w_pixrGetChannel(pIPWState, pr2, channel, &ch2));

	/* Look at each line of the two images.  Return at the end of the
	 * first line with a difference */
	ERR_PASS( w_pixrGetImage(pIPWState, ch1, &pImage1));
	ERR_PASS( w_pixrGetImage(pIPWState, ch2, &pImage2));

	for (pLine1 = (UInt32 *)pImage1, pLine2 = (UInt32 *)pImage2, 
		 tmp = h, eq = 0;
	     tmp > 0;
	     tmp--, pLine1 += wpl, pLine2 += wpl)
	{
	    /* search the line from right to left to save a variable */
	    for (currentOffset = lastWordOffset;
		 currentOffset >= 0;
		 currentOffset--)
	    {
		/* accumulate the differences in a single line */
		eq |= pLine1[currentOffset] ^ pLine2[currentOffset];
	    }

	    if (eq != 0)	/* we found a difference, they don't match */
	    {
		*pAreEqual = cFalse;
		NORMAL_RETURN;
	    }
	}
    }
    

	/* we didn't find any differences, so return AreEqual */
    *pAreEqual = cTrue;
    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute the absolute value of the pixel difference between two images.
 *
 * LongDescription:::
 *  Compare two images.  The result is the absolute value of the difference
 *  between the two images at every pixel location.  This routine is used
 *  to find differences in gray images.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image.
 *  prS2::           Pointer to second source image.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to dest image.  If a dest image is
 *		     supplied, it is checked for proper dimensions before
 *		     being used.  If no dest image is supplied, one is
 *		     created.
 *
 * Notes:::
 *  This routine is primarily used by people trying to figure out where
 *  some version of an algorithm differed from from some previous version.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable, though there seems to be bug involving pixels close to 127.
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_absValDifference (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 w1, h1, d1;   /* width, height and depth of images */
    Int32		 w2, h2, d2;   /* width, height and depth of images */
    ScaledInt1616	 xdpi1, ydpi1;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepthAny, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepthAny, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS1, &d1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));

    ERR_PASS( w_pixrGetDepth(pIPWState, prS2, &d2));
    ERR_GEN(d2 != d1, cErrSrcDimensNotEqual);

	/* Compute the size of the output image as the minimum of
	 * the width and height of the input images.  This is done
	 * since sometimes we want to compare images of slightly
	 * different sizes (with the same upper left corner).
	 */

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.  NOTE: we create a new pixr instead of
	 * cloning one of the old ones because we've modified w1 and h1
	 * above.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, d1,
					     xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_absValDiff (pIPWState, pImage1, bpl1, pImage2, bpl2,
				   w1 + 2*cFrameBits, h1 + 2*cFrameBits, d1,
				   pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compares gray images, producing binary image indicating where img1 > img2.
 *
 * LongDescription:::
 *  This routine compares two gray (8 bpp) images.  It returns a binary
 *  image.  If image1[x,y] * mult1 > image2[x,y] * mult2, then dest[x,y] = 1,
 *  otherwise dest[x,y] = 0..
 *  
 * PrivateDescription:::
 *  This routine is used by Luc Vincent at XIS to find horizontal and
 *  vertical gutters in images.  The first step is to tag each pixel with
 *  the length of the horizontal run and the length of the vertical run
 *  of which it's a part.  This routine is then used to find all pixels
 *  which are primarly horizontal (that's what the weights are for) or
 *  primarily vertical.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image.
 *  prS2::           Pointer to second source image.
 *  mult1::          Weighting factor for pixels in first source image.
 *  mult2::          Weighting factor for pixels in second source image.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to binary dest image.  If a dest
 *		     image is supplied, it is checked for proper depth,
 *		     resolution and dimensions before being used.  If no 
 *		     dest image is supplied, one is created.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison, run length
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies, L. Vincent.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_compareGrayImages (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    UInt32              mult1,
    UInt32              mult2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 w1, h1, d1;     /* width, height and depth of images */
    Int32		 w2, h2, d2;
    ScaledInt1616	 xdpi1, ydpi1;
    ScaledInt1616	 xdpi2, ydpi2;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;


    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepthAny, cChannels1, cFrame32);

	/* Both source images must be the same size.  The output image
	 * has the same dimensions, but is binary.
	 */

    ERR_PASS( w_pixrGetDepth(pIPWState, prS1, &d1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));

    ERR_PASS( w_pixrGetDepth(pIPWState, prS2, &d2));
    ERR_GEN(d2 != d1, cErrSrcDimensNotEqual);

    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS2, &xdpi2, &ydpi2));

    ERR_GEN(w1 != w2  ||  h1 != h2, cErrSrcDimensNotEqual);
    ERR_GEN((ABS(xdpi1 - xdpi2) > 0x10000) ||
	    (ABS(ydpi1 - ydpi2) > 0x10000), cErrSrcDimensNotEqual);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, 1, xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_compareGrayImages (pIPWState, pImage1, bpl1, mult1,
					  pImage2, bpl2, mult2,
					  w1 + 2*cFrameBits, h1 + 2*cFrameBits,
					  d1, pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute the pixel-wise minimum of two images.
 *
 * LongDescription:::
 *  Compare two images.  At each pixel position, output the minimum value
 *  of the pixels in the two source images.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image.
 *  prS2::           Pointer to second source image.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to dest image.  If a dest image is
 *		     supplied, it is checked for proper dimensions before
 *		     being used.  If no dest image is supplied, one is
 *		     created.
 *
 * Notes:::
 *  This isn't a wonderfully speedy operation, but it's really useful
 *  sometimes.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_min (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 w1, h1, d1;     /* width, height and depth of images */
    Int32		 w2, h2, d2;
    ScaledInt1616	 xdpi1, ydpi1;
    ScaledInt1616	 xdpi2, ydpi2;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepthAny, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepthAny, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS1, &d1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));

    ERR_PASS( w_pixrGetDepth(pIPWState, prS2, &d2));
    ERR_GEN(d2 != d1, cErrSrcDimensNotEqual);

    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS2, &xdpi2, &ydpi2));

    ERR_GEN((ABS(xdpi1 - xdpi2) > 0x10000) ||
	    (ABS(ydpi1 - ydpi2) > 0x10000), cErrSrcDimensNotEqual);

	/* Compute the size of the output image as the minimum of
	 * the width and height of the input images.  This is done
	 * since sometimes we want to compare images of slightly
	 * different sizes (with the same upper left corner).
	 */
    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.  NOTE: we create a new pixr instead of
	 * cloning one of the old ones because we've modified w1 and h1
	 * above.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, d1,
					     xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_min (pIPWState, pImage1, bpl1,
			    pImage2, bpl2,
			    w1 + 2*cFrameBits, h1 + 2*cFrameBits, d1,
			    pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute the pixel-wise maximum of two images.
 *
 * LongDescription:::
 *  Compare two images.  At each pixel position, output the maximum value
 *  of the pixels in the two source images.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image.
 *  prS2::           Pointer to second source image.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to dest image.  If a dest image is
 *		     supplied, it is checked for proper dimensions before
 *		     being used.  If no dest image is supplied, one is
 *		     created.
 * Notes:::
 *  This isn't a wonderfully speedy operation, but it's really useful
 *  sometimes.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_max (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 w1, h1, d1;     /* width, height and depth of images */
    Int32		 w2, h2, d2;
    ScaledInt1616	 xdpi1, ydpi1;
    ScaledInt1616	 xdpi2, ydpi2;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepthAny, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepthAny, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS1, &d1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));

    ERR_PASS( w_pixrGetDepth(pIPWState, prS2, &d2));
    ERR_GEN(d2 != d1, cErrSrcDimensNotEqual);

    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS2, &xdpi2, &ydpi2));

    ERR_GEN((ABS(xdpi1 - xdpi2) > 0x10000) ||
	    (ABS(ydpi1 - ydpi2) > 0x10000), cErrSrcDimensNotEqual);

	/* Compute the size of the output image as the minimum of
	 * the width and height of the input images.  This is done
	 * since sometimes we want to compare images of slightly
	 * different sizes (with the same upper left corner).
	 */
    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, d1,
					     xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_max (pIPWState, pImage1, bpl1,
			    pImage2, bpl2,
			    w1 + 2*cFrameBits, h1 + 2*cFrameBits, d1,
			    pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute pixel-wise sum of two 8-bit images, clip values at 255
 *
 * LongDescription:::
 *  Go through two 8-bit images.  At each pixel location, compute
 *  image1[x,y] + image2[x,y].  If the result is greater than 255,
 *  clip it to 255.
 *  Write the output in a destination image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image.
 *  prS2::           Pointer to second source image.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to dest image.  If a dest image is
 *		     supplied, it is checked for proper dimensions before
 *		     being used.  If no dest image is supplied, one is
 *		     created.
 *
 * Notes:::
 *  This function computes prS1 + prS2.  It does spend time on the clipping
 *  function.  If you know that prS1 + prS2 <= 255 at every pixel location, use
 *  w_addGrayNoClip for a performance improvement.
 *  <BR><BR>
 *  Unlike w_subtractGray, this function only works on 8-bit pixels.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison, threshold finding
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod, D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addGray (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 w1, h1, d1;   /* width, height, depth of images */
    Int32		 w2, h2, d2;   /* width, height, depth of images */
    ScaledInt1616	 xdpi1, ydpi1;
    ScaledInt1616	 xdpi2, ydpi2;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS1, &d1));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS2, &d2));

	/* Check the depths */
    ERR_GEN( (d1 != d2), cErrSrcDimensNotEqual);

    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS2, &xdpi2, &ydpi2));

    ERR_GEN((ABS(xdpi1 - xdpi2) > 0x10000) ||
	    (ABS(ydpi1 - ydpi2) > 0x10000), cErrSrcDimensNotEqual);

	/* Compute the size of the output image as the minimum of
	 * the width and height of the input images.  This is done
	 * since sometimes we want to compare images of slightly
	 * different sizes (with the same upper left corner).
	 */
    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, d1,
					     xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_addGray (pIPWState, pImage1, bpl1,
			        pImage2, bpl2,
			        w1 + 2*cFrameBits, h1 + 2*cFrameBits, d1,
			        pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute pixel-wise sum of two 8-bit images, do NOT clip values at 255.
 *
 * LongDescription:::
 *  Go through two images.  At each pixel location, compute
 *  image1[x,y] + image2[x,y].  Do not check for overflow.
 *  Write the output in a destination image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image.
 *  prS2::           Pointer to second source image.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to dest image.  If a dest image is
 *		     supplied, it is checked for proper dimensions before
 *		     being used.  If no dest image is supplied, one is
 *		     created.
 *
 * Notes:::
 *  This function is like w_addGray except that it does not spend time
 *  clipping pixels differeces at 255.  Since overflowed pixel values are not
 *  used elsewhere, this routine is really only useful when the client knows
 *  that prS1[x,y] + prS2[x,y] <= 255 for all x and y.
 *  <BR><BR>
 *  Unlike w_subtractGrayNoClip, this routine only works on 8-bit images.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod, D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addGrayNoClip (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 d1, w1, h1;   /* depth, width, height of images */
    Int32		 d2, w2, h2;   /* depth, width, height of images */
    ScaledInt1616	 xdpi1, ydpi1;
    ScaledInt1616	 xdpi2, ydpi2;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS1, &d1));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS2, &d2));

	/* Check the depths */
    ERR_GEN( (d1 != d2), cErrSrcDimensNotEqual);

    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS2, &xdpi2, &ydpi2));

    ERR_GEN((ABS(xdpi1 - xdpi2) > 0x10000) ||
	    (ABS(ydpi1 - ydpi2) > 0x10000), cErrSrcDimensNotEqual);

	/* Compute the size of the output image as the minimum of
	 * the width and height of the input images.  This is done
	 * since sometimes we want to compare images of slightly
	 * different sizes (with the same upper left corner).
	 */
    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, d1,
					     xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_addGrayNoClip (pIPWState, pImage1, bpl1,
				      pImage2, bpl2,
				      w1 + 2*cFrameBits, h1 + 2*cFrameBits, d1,
				      pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute pixel-wise difference of two images, clip values at 0.
 *
 * LongDescription:::
 *  Go through two images.  At each pixel location, compute
 *  image1[x,y] - image2[x,y].  If the result is negative, set it to 0.
 *  Write the output in a destination image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image (minuend).
 *  prS2::           Pointer to second source image (subtrahend).
 *
 * Output:::
 *  pPrD::           Pointer to pointer to dest image.  If a dest image is
 *		     supplied, it is checked for proper dimensions before
 *		     being used.  If no dest image is supplied, one is
 *		     created.
 *
 * Notes:::
 *  This function computes prS1 - prS2.  It does spend time on the clipping
 *  function.  If you know that prS1 >= prS2 at every pixel location, use
 *  w_subtractGrayNoClip for a performance improvement.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_subtractGray (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 w1, h1, d1;   /* width, height and depth of images */
    Int32		 w2, h2, d2;   /* width, height and depth of images */
    ScaledInt1616	 xdpi1, ydpi1;
    ScaledInt1616	 xdpi2, ydpi2;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS1, &d1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));

    ERR_PASS( w_pixrGetDepth(pIPWState, prS2, &d2));
    ERR_GEN(d2 != d1, cErrSrcDimensNotEqual);

    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS2, &xdpi2, &ydpi2));

    ERR_GEN((ABS(xdpi1 - xdpi2) > 0x10000) ||
	    (ABS(ydpi1 - ydpi2) > 0x10000), cErrSrcDimensNotEqual);

	/* Compute the size of the output image as the minimum of
	 * the width and height of the input images.  This is done
	 * since sometimes we want to compare images of slightly
	 * different sizes (with the same upper left corner).
	 */
    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, d1,
					     xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_subtractGray (pIPWState, pImage1, bpl1,
				     pImage2, bpl2,
				     w1 + 2*cFrameBits, h1 + 2*cFrameBits, d1,
				     pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute pixel-wise difference of two images, do NOT clip values at 0.
 *
 * LongDescription:::
 *  Go through two images.  At each pixel location, compute
 *  image1[x,y] - image2[x,y].  Write the output in a destination image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image (minuend).
 *  prS2::           Pointer to second source image (subtrahend).
 *
 * Output:::
 *  pPrD::           Pointer to pointer to dest image.  If a dest image is
 *		     supplied, it is checked for proper dimensions before
 *		     being used.  If no dest image is supplied, one is
 *		     created.
 *
 * Notes:::
 *  This function is like w_subtractGray except that it does not spend time
 *  clipping pixels differeces at 0.  Since negative pixel values are not
 *  used elsewhere, this routine is really only useful when the client knows
 *  that prS1[x,y] >= prS2[x,y] for all x and y.  This can happen if prS1
 *  is the max of two images, and/or prS2 is the min.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_subtractGrayNoClip (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 w1, h1, d1;   /* width, height and depth of images */
    Int32		 w2, h2, d2;   /* width, height and depth of images */
    ScaledInt1616	 xdpi1, ydpi1;
    ScaledInt1616	 xdpi2, ydpi2;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS1, &d1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));

    ERR_PASS( w_pixrGetDepth(pIPWState, prS2, &d2));
    ERR_GEN(d2 != d1, cErrSrcDimensNotEqual);

    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS2, &xdpi2, &ydpi2));

    ERR_GEN((ABS(xdpi1 - xdpi2) > 0x10000) ||
	    (ABS(ydpi1 - ydpi2) > 0x10000), cErrSrcDimensNotEqual);

	/* Compute the size of the output image as the minimum of
	 * the width and height of the input images.  This is done
	 * since sometimes we want to compare images of slightly
	 * different sizes (with the same upper left corner).
	 */
    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, d1,
					     xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_subtractGrayNoClip (pIPWState, pImage1, bpl1,
					   pImage2, bpl2,
					   w1 + 2*cFrameBits, h1 + 2*cFrameBits,
					   d1,
					   pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute pixel-wise average of two 8-bit images.
 *
 * LongDescription:::
 *  Go through two images.  At each pixel location, compute
 *  (image1[x,y] + image2[x,y])/2.
 *  Write the output in a destination image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS1::           Pointer to first source image.
 *  prS2::           Pointer to second source image.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to dest image.  If a dest image is
 *		     supplied, it is checked for proper dimensions before
 *		     being used.  If no dest image is supplied, one is
 *		     created.
 *
 * Notes:::
 *  This function simply averages two images pixel by pixel.  To save
 *  time, we disregard the LS bits, so it doesn't produce exact values.
 *  <BR><BR>
 *  This routine only works on 8-bit images.
 *
 * FunctionCategory:::
 *  comparison
 *
 * Keywords:::
 *  comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_averageGray (
    IPWState           *pIPWState,
    PIXR               *prS1,
    PIXR               *prS2,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 w1, h1;   /* width, height of images */
    Int32		 w2, h2;   /* width, height of images */
    ScaledInt1616	 xdpi1, ydpi1;
    ScaledInt1616	 xdpi2, ydpi2;
    Int32		 bpl, bpl1, bpl2;
    UInt8		*pImage, *pImage1, *pImage2;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS1 == NULL || prS2 == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS1, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prS2, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS1, &w1));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS2, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS1, &h1));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS2, &h2));

    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS1, &xdpi1, &ydpi1));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS2, &xdpi2, &ydpi2));

    ERR_GEN((ABS(xdpi1 - xdpi2) > 0x10000) ||
	    (ABS(ydpi1 - ydpi2) > 0x10000), cErrSrcDimensNotEqual);

	/* Compute the size of the output image as the minimum of
	 * the width and height of the input images.  This is done
	 * since sometimes we want to compare images of slightly
	 * different sizes (with the same upper left corner).
	 */
    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrConditionalNoInit(pIPWState, w1, h1, 8,
					     xdpi1, ydpi1,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we were given a dest, compute intersection with it too. */
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w2));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &h2));

    w1 = MIN(w1, w2);
    h1 = MIN(h1, h2);

	/* it seems to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS1, &pImage1));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prS2, &pImage2));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  &bpl ));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS1, &bpl1));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prS2, &bpl2));

    ERR_PASS_CLEANUP(i_averageGray (pIPWState, pImage1, bpl1,
				    pImage2, bpl2,
				    w1 + 2*cFrameBits, h1 + 2*cFrameBits, 8,
				    pImage, bpl));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/***********************************************************************
 *
 * findRange ()
 *
 *	Compute the (min, max) in a grayscale image 
 *
 * Arguments:
 *	pImage	    -	    Image on which to compute range
 *	pLowest	    -	    Lowest value
 *	pHighest    -	    Highest value
 *
 * Returns:
 *	cSuccessful or error code
 *
 ***********************************************************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find minimum and maximum pixels values in an image.
 *
 * LongDescription:::
 *  This routine looks through an entire image.  It finds the minimum
 *  pixel value and the maximum pixel value, returning just these.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be inspected.
 *
 * Output:::
 *  pLowest::        Pointer to byte that will contain minimum pixel value.
 *  pHighest::       Pointer to byte that will contain maximum pixel value.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findRange (
    IPWState           *pIPWState,
    PIXR               *prS,
    UInt8              *pLowest,
    UInt8              *pHighest)
{
PROLOGUE;
Int32   depth, w, h, bpl;
UInt8  *pImage;

    ERR_GEN(prS == NULL || pLowest == NULL || pHighest == NULL,
	    cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &depth));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pImage));

    ERR_PASS( i_range (pIPWState, pImage, bpl, w, h, depth, pLowest, pHighest));
    
    NORMAL_RETURN;
}
	    
/*

 $Log: intops3.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.35  1998/01/14 00:13:23  ddavies
 intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.34  1997/07/31 00:55:30  ddavies
 intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v

 Add wrappers w_addGray (adds pixels in two images, clips results at 0xFF),
 w_addGrayNoClip (same without clipping), and w_averageGray (averages pixels
 in two images).

 Revision 1.33  1997/04/15 22:12:15  ddavies
 intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.32  1996/05/21 03:37:23  ejaquith
 ejaquith on Mon May 20 20:36:29 PDT 1996

 intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v

 w_areEqual() was only doing single plane PIXR checking.
   And we were using it for 3-plane RGB.
 hmm.  Anyway, added multi-plane checking.

 Revision 1.31  1996/01/04 20:24:28  ejaquith
 intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.30  1995/12/13  04:09:55  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.29  1995/11/13  20:13:22  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.28  1995/11/03  22:48:24  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.27  1995/10/13  22:02:34  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.26  1995/10/05  04:27:07  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.25  1995/10/03  01:20:04  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.24  1995/09/21  22:54:11  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.23  1995/09/18  23:44:44  ejaquith
 * ejaquith on Mon Sep 18 16:43:36 PDT 1995
 *
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Made w_areEqual() not modify output pointers until returning.
 *
 * Revision 1.22  1995/09/18  18:32:49  ddavies
 * ddavies on Mon Sep 18 11:29:59 PDT 1995
 *
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Add 2*cFrameBits to all widths and heights in calls to ipshared routines.
 * Also add dimensions check so the area operated on is the intersection of
 * the two source images AND the dest image is one is supplied.
 *
 * Revision 1.21  1995/09/15  18:00:25  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.20  1995/09/11  22:11:08  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.19  1995/09/10  00:33:07  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.18  1995/09/07  01:32:56  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.17  1995/09/06  20:20:57  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.16  1995/09/01  01:16:49  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.15  1995/08/31  21:14:48  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.14  1995/08/25  01:37:05  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Fixed warnings
 *
 * Revision 1.13  1995/08/24  00:31:37  jcraig
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * in compareGrayImages, do depth checks correctly.
 * pass source depth to i_compareGrayImages
 *
 * Revision 1.12  1995/08/17  22:02:37  ejaquith
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.11  1995/07/29  04:42:10  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.10  1994/11/04  03:17:37  macleod
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Added rangePr, which computes the (min,max) range of an image.
 *
 * Revision 1.9  1994/10/06  20:54:34  lperry
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Change to use new .pub header file instead of ippros.h
 *
 * Revision 1.8  1994/07/29  01:29:39  macleod
 * macleod on Thu Jul 28 18:29:17 PDT 1994
 *
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Added subtractGrayPr, subtractGrayPixr, subtractGrayNoClipPr,
 * and subtractGrayNoClipPixr
 *
 * Revision 1.7  1994/07/27  18:27:29  macleod
 * macleod on Wed Jul 27 11:27:03 PDT 1994
 *
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Added wrappers for i_min and i_max, called minPr, minPixr, maxPr, and
 * maxPixr.
 *
 * Revision 1.6  1994/06/08  04:11:38  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.5  1993/12/08  21:56:24  ddavies
 * ddavies on Wed Dec 8 13:55:37 PST 1993
 *
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Fix a bug in equalPr that Art Medlar noticed.  It was comparing one word
 * too many in each line when the image width was a multiple of 32 bits.
 *
 * Revision 1.4  1993/12/03  01:16:39  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.3  1993/11/06  03:13:34  ddavies
 * intops3.c -> /product/ipcore/ipwrap/src/RCS/intops3.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

