/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  ioutils.c:  Generally useful IO utility functions
 *              to transfer Intel format data from and to
 *              an in-memory data buffer on both
 *              386 and non-386 machines
 *
 *		    w_getIntel16()
 *		    w_getIntel32()
 *		    w_putIntel16()
 *		    w_putIntel32()
 *	 Byte-swap a pixr (IFF little endian)
 *		    w_flip()
 */

#include <stdio.h>

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "defines.pub"

/* prototypes */
#include "props.pub"
#include "io.pub"
#include "shrpixr.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: ioutils.c,v $; $Revision: 1.1.1.1 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get a short integer that is stored in Intel format at a specified location.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cp::             Ptr to short-aligned byte location to get integer from
 *
 * Output:::
 *  dp::             Ptr to returned short int value.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  byteswapping
 *
 * Keywords:::
 *  byteswapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getIntel16 (
    IPWState           *pIPWState,
    UInt8              *cp,
    UInt16             *dp)
{
    PROLOGUE;

#ifndef _386_ARCH_

    Int16      intel;

    ERR_GEN(cp == NULL  ||  dp == NULL, cErrNullPtr);

    intel   = *(cp+1);  /* msb is stored in 2nd byte */
    intel <<=  8;       /* shift it top msb of intel */
    intel  |= *cp;      /* OR in the lsb             */

    *dp = intel;

#else     /* _386_ARCH_ */

    Int16     *intel;

    ERR_GEN(cp == NULL  ||  dp == NULL, cErrNullPtr);

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    intel = (Int16 *)cp;
    *dp = *intel;

#endif  /* _386_ARCH_ */

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Store a short integer in Intel format at a specified location.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  value::          Integer to store
 *
 * Output:::
 *  cp::             Ptr to short-aligned byte location to store integer
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  byteswapping
 *
 * Keywords:::
 *  byteswapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_putIntel16 (
    IPWState           *pIPWState,
    Int16               value,
    UInt8              *cp)
{
    PROLOGUE;

#ifndef _386_ARCH_

    ERR_GEN(cp == NULL, cErrNullPtr);

    *cp++   = value & 0x00ff;
    *cp     = (value >> 8) & 0x00ff;

#else   /* if _386_ARCH_ */

    Int16    *intel;

    ERR_GEN(cp == NULL, cErrNullPtr);

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    intel  = (Int16 *)cp;
    *intel = value;

#endif  /* _386_ARCH_ */

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get an integer that is stored in Intel format at a specified location.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cp::             Ptr to int-aligned byte location to get integer from
 *
 * Output:::
 *  dp::             Ptr to returned int value.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  byteswapping
 *
 * Keywords:::
 *  byteswapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getIntel32 (
    IPWState           *pIPWState,
    UInt8              *cp,
    Int32              *dp)
{
    PROLOGUE;

#ifndef _386_ARCH_

    Int32   intel;

    ERR_GEN(cp == NULL  ||  dp == NULL, cErrNullPtr);

    intel   = *(cp+3);  /* msb is stored in 4th byte */
    intel <<=  8;       /* shift it to msb of intel */
    intel  |= *(cp+2);  /* next msb */
    intel <<=  8;       
    intel  |= *(cp+1);  
    intel <<=  8;
    intel  |= *cp;      /* OR in the lsb             */

    *dp = intel;

#else  /* for _386_ARCH_ */

    Int32    *intel;

    ERR_GEN(cp == NULL  ||  dp == NULL, cErrNullPtr);

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    intel = (Int32 *)cp;
    *dp = *intel;

#endif  /* _386_ARCH_ */

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Store a 32-bit integer in Intel format at a specified location.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  value::          Integer to store
 *
 * Output:::
 *  cp::             Ptr to int-aligned byte location to store integer
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  byteswapping
 *
 * Keywords:::
 *  byteswapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_putIntel32 (
    IPWState           *pIPWState,
    Int32               value,
    UInt8              *cp)
{
    PROLOGUE;

#ifndef  _386_ARCH_

    ERR_GEN(cp == NULL, cErrNullPtr);

    *cp++   = value & 0x000000ff;  /* store the lsb */
    value >>= 8;
    *cp++   = value & 0x000000ff;
    value >>= 8;
    *cp++   = value & 0x000000ff;
    value >>= 8;
    *cp     = value & 0x000000ff;  /* store the msb */

#else  /* _386_ARCH_ */

    Int32       *intel;

    ERR_GEN(cp == NULL, cErrNullPtr);

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    intel  = (Int32 *)cp;
    *intel = value;

#endif  /* _386_ARCH_ */

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Byte swap a PIXR;  Becomes a no-op if on a big-endian platform.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  pixr::           Ptr to PIXR, any depth, may be multi-channel.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  byteswapping
 *
 * Keywords:::
 *  byteswapping, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_flip (
    IPWState           *pIPWState,
    PIXR               *pixr)
{
    PROLOGUE;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
 
#if _ALPACA_IMAGE_FMT_ != cBigEndianFmt
    {
	Int32   i, bpl, channels;

	ERR_GEN(pixr == NULL, cErrPixr_null);

	ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &bpl));
	ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

	ERR_GEN(bpl % 4, cErrBplNotMult4);
    
	/* Do the real work */
	if (channels == 1) {
	    UInt8  *pImage;
	    Int32   fh;
	
	    ERR_PASS( w_pixrGetImageAll(pIPWState, pixr, &pImage));
	    ERR_PASS( w_pixrFramedHeight(pIPWState, pixr, &fh));

	    ERR_PASS( i_swapImageBytes (pIPWState, pImage, bpl, fh));
	}
	else {
	    PIXR  *sub;
	
	    for (i = 0; i < channels; i++) {
		ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &sub));

		ERR_PASS( w_flip(pIPWState, sub));
	    }
	}
    }
#endif  /* _ALPACA_IMAGE_FMT_ */

    NORMAL_RETURN;
}


/*
 $Log: ioutils.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.21  1998/01/14 00:13:25  ddavies
 ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.20  1997/06/30 23:10:32  ddavies
 ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v

 Delete inclusion of Sandpiper's internal.h.

 Revision 1.19  1996/01/04 20:24:31  ejaquith
 ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.18  1995/12/13  04:09:59  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.17  1995/11/13  20:04:33  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * Updated PROLOGUE usage.
 *
 * Revision 1.16  1995/10/13  22:02:38  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/10/07  03:11:41  ddavies
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * Add stdio.h since it's not in ipwcore.prv anymore.
 *
 * Revision 1.14  1995/09/29  00:35:40  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.13  1995/09/23  00:50:19  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * Removed the double comment blocks.
 *
 * Revision 1.12  1995/09/21  22:54:15  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.11  1995/09/13  23:31:34  ddavies
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * Remove invertImage parameter from i_swapImageBytes.  I backed out of that
 * change.
 *
 * Revision 1.10  1995/09/13  00:20:42  ddavies
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * Set new invertImage parameter in i_swapImageBytes to cFalse.
 *
 * Revision 1.9  1995/09/12  16:55:12  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.8  1995/09/07  01:32:59  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.7  1995/09/06  20:21:02  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.6  1995/09/01  02:20:55  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.5  1995/08/31  21:14:51  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.4  1995/08/22  02:21:12  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.3  1995/08/17  22:01:43  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.2  1994/12/19  23:26:35  ejaquith
 * ioutils.c -> /product/ipcore/ipwrap/src/RCS/ioutils.c,v
 *
 * Updated flip_pixr() to allow for multi-channeled PIXR byte swapping.
 * Now byte-swaps every plane if a multi-plane PIXR is passed in.
 *
 * Revision 1.1  1994/10/10  21:40:48  lperry
 * Initial revision
 *
*/
