/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  	itree.c:   Interpretation tree-based matching wrappers
 *
 *	Code to match two lists of rectangles:
 *		w_interpTreeRectMatcher()
 *
 */

#include <stdio.h>

#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "defines.pub"
#include "imagerep.h"
#include "frames.pub"

/* prototypes */
#include "props.pub"
#include "rectops.pub"
#include "io.pub"
#include "itree.pub"
#include "shrpatt.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: itree.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");



/* Allow up to 1/8" mismatch for width and height
 * for unary constraints in interpretation tree rectangle matching.
*/
#define SIZETOL(res)   ((res) / 8)

/* Allow up to 1/4" total translational movement for both rects
 * for binary constraints in interpretation tree rectangle matching.
 */
#define TRANSTOL(res)  ((res) / 4)


/* This is ALWAYS 0 on checkin/release, and 1 only when checked out 
 * if some printed debugging information is desired.
 */
#define DEBUG_I_TREE   0


/* This is the type the ptr to the interpretation tree stuff will point
 * to that is passed to the unary and binary constraint tests and to
 * the model test, for rectangle list interpretation tree matching
 */
typedef struct interpTreeRectType
{
    IPWState    *pIPWState;
    Int32       *wmin, *wmax, *hmin, *hmax;
    RECTANGL    *mrect, *drect;
    Int16      **minDx, **minDy, **maxDx, **maxDy;
    Int32       *matchesDtoM;
    Int32        nMatchesDtoM;
} interpTreeRectType;



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Used for unary constraints of an interpretation tree rect matcher
 *
 * LongDescription:::
 *  See if it is possible for the single data rect "dataFeatureNum" to match
 *  the single model rect "modelFeatureNum".  This match test is context-free
 *  outside of these two rectangles.
 *
 * PrivateDescription:::
 *  Here our unary constraint is that the width and height of the data
 *  feature's rect must be "close enough" to that of the model feature's rect.
 *
 * Input:::
 *  pIPWState::		Ptr to IPCore state
 *  pITStuff::          Ptr to interpretation tree stuff for rect data
 *  dataFeatureNum::    Index value of data rect we are considering
 *  modelFeatureNum::   Index value of model rect we are considering
 *
 * Output:::
 *  pPassed::           Ptr to cTrue if this new data-model feature pairing
 *                      passed this test.  Ptr to cFalse, o.w.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_interpTreeRectUnaryConstraints (
    IPWState *pIPWState,
    void     *pITStuff,
    Int32     dataFeatureNum,
    Int32     modelFeatureNum,
    Bool     *pPassed)
{
    PROLOGUE;
    interpTreeRectType    *pITRect = NULL;

    pITRect = (interpTreeRectType *)pITStuff;

    ERR_GEN(pITStuff == NULL  ||  pPassed == NULL,  cErrNullPtr);

    *pPassed = ((pITRect->drect[dataFeatureNum].w >= 
		 pITRect->wmin[modelFeatureNum] &&

		 pITRect->drect[dataFeatureNum].w <=
		 pITRect->wmax[modelFeatureNum] &&

		 pITRect->drect[dataFeatureNum].h >=
		 pITRect->hmin[modelFeatureNum] &&

		 pITRect->drect[dataFeatureNum].h <= 
		 pITRect->hmax[modelFeatureNum])
		? cTrue : cFalse);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Used for the binary constraints of an interpretation tree rect matcher
 *
 * LongDescription:::
 *  Matching array, matchesDtoM, has been filled with values
 *  from matchesDtoM[0..d-1].
 *  The value, v, at each index, i, in the array corresponds to the
 *  "i"th data rect matching the "v"th model rect.
 *  The value may be NO_FEATURE_MATCHED.<BR><BR>
 *  
 *  Given this array, the new data-model matching is matchesDtoM[d].
 *  We want to see if the new match we are considering between data feature "d"
 *  and model feature "matchesDtoM[d]" is pairwise consistent with every
 *  match in the list so far.
 *
 * PrivateDescription:::
 *  The binary constaint test is:
 *  for each data-model pairing, i in [0..d-1], in the array,
 *  to determine if the difference in the X(Y) positions between
 *  the upper left corners of the two data rects "i" and "d" is
 *  "similar" to the difference in the
 *  X(Y) positions between the upper left corners of two model rects
 *  matchesDtoM[i] and matchesDtoM[d].
 *
 * Input:::
 *  pIPWState::		Ptr to IPCore state
 *  pITStuff::          Ptr to interpretation tree stuff for rect data
 *  d::                 This is the current "depth" we are at in the
 *                      tree, and simultaneously (or equivalently) the data
 *                      rect index that we are want to check for a match
 *                      with the "matchesDtoM[d]"th model rect.
 *  matchesDtoM::       Ptr to array of data rect matching list.
 *                      Element 0 corresponds the 0th data rect, and so on.
 *                      Each element's value will be a number, i,
 *                      [0..nModelFeatures-1] corresponding to it matching
 *                      the the "i"th model rect, or NO_FEATURE_MATCHED
 *                      to indicate no suitable matching model rect was found.
 *
 * Output:::
 *  pPassed::           Ptr to cTrue if this new data-model feature pairing
 *                      passed this test.  Ptr to cFalse, o.w.
 *
 * Returns:::
 *
 * Notes:::
 *  matchesDtoM[d] .. values::
 *   These array values have not been filled in yet (and will therefore
 *   be NO_FEATURE_MATCHED) and should not be used/considered.
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_interpTreeRectBinaryConstraints (
    IPWState   *pIPWState,
    void       *pITStuff,
    Int32       d,
    Int32      *matchesDtoM,
    Bool       *pPassed)
{
    PROLOGUE;
    Int32                  i;
    interpTreeRectType    *pITRect = NULL;

    pITRect = (interpTreeRectType *)pITStuff;

    ERR_GEN(pITStuff == NULL  ||  pPassed == NULL  ||  matchesDtoM == NULL,
	    cErrNullPtr);
    

    for (i = 0; i < d; i++)
    {
	if (matchesDtoM[i] == NO_FEATURE_MATCHED)
	    /* Null branch always consistent */
	    continue;

	/* Here we check that the relationship between data features
	 * (i) and (d) in terms of the manhattan distance between both
	 * their upper left corners (they're rectangles remember)
	 * is pretty close (ie. within a specified range) to this distance
	 * between both the model features (one we've already matched and one
	 * we're trying to determine if the model-data match is consistent
	 * in this routine).
	 */
	if (pITRect->drect[i].x - pITRect->drect[d].x < 
	    pITRect->minDx[matchesDtoM[d]][matchesDtoM[i]]  ||

	    pITRect->drect[i].x - pITRect->drect[d].x >
	    pITRect->maxDx[matchesDtoM[d]][matchesDtoM[i]]  ||

	    pITRect->drect[i].y - pITRect->drect[d].y < 
	    pITRect->minDy[matchesDtoM[d]][matchesDtoM[i]]  ||

	    pITRect->drect[i].y - pITRect->drect[d].y >
	    pITRect->maxDy[matchesDtoM[d]][matchesDtoM[i]])
	{
	    /* Not consistent with previous matching model-data features */
	    *pPassed = cFalse;
	    NORMAL_RETURN;
	}
    }
    
    /* The current suggested model-data pairing is pairwise consistent 
     * with all previous model-data pairings so far!
     */

    *pPassed = cTrue;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Used for the model test step of an interpretation tree rectangle matcher
 *
 * LongDescription:::
 *  Used in w_interpTreeRectMatcher() as the model test routine for
 *  interpretation tree matching.  i_interpTreeMatcher() invokes this
 *  routine  whenever we hit a leaf in the search space, doing any
 *  additional testing on the data to model pairing  and 
 *  recording good pairings.
 *
 * PrivateDescription:::
 *  Presently, this only looks for more data rects matching model rects
 *  in a leaf, and records pairings when this is the case.  It also
 *  instructs the matcher to increase the pruning value if possible when
 *  good pairings are found.
 *
 * Input:::
 *  pIPWState::		Ptr to IPCore state
 *  nDataFeatures::     Number of rectangles in data rect list
 *  nModelFeatures::    Number of rectangles in model rect list
 *  matchesDtoM::       Ptr to array of "nDataFeatures" elements.
 *                      Element 0 corresponds the 0th data rect, and so on.
 *                      Each element's value will be a number, i,
 *                      [0..nModelFeatures-1] corresponding to it matching
 *                      the the "i"th model rect, or NO_FEATURE_MATCHED
 *                      to indicate no suitable matching model rect was found.
 *  nMatchesDtoM::      Number of elements in above array that matched
 *                      actual model rects.
 *
 * Output:::
 *  pITStuff::          Ptr to interpretation tree stuff for rect data and
 *                      for recording good pairings.
 *  pNextNMatchesDtoMAsGood::  Ptr to cTrue to tell searching module to
 *                             increase pruning level to this level;
 *                             cFalse to leave level the same.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  template matching
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_interpTreeRectModelTest (
    IPWState   *pIPWState,
    Int32       nDataFeatures,
    Int32       nModelFeatures,
    Int32      *matchesDtoM,
    Int32       nMatchesDtoM,
    void       *pITStuff,
    Bool       *pNextNMatchesDtoMAsGood)
{
    PROLOGUE;
    Int32                  i;
    interpTreeRectType    *pITRect = NULL;

    pITRect = (interpTreeRectType *)pITStuff;

    ERR_GEN(pITStuff == NULL  ||  matchesDtoM == NULL,  cErrNullPtr);

    /* Keep anal compilers happy */
    nModelFeatures = nModelFeatures;
    
    /* We're really only concerned now with matching the most rectangles.
     * If we got more matches now, consider it better.
     * If there are more matches present, fill in/overwrite the
     * matchesDtoM field with the passed in data and model features pairing.
     */

    if (nMatchesDtoM >= pITRect->nMatchesDtoM)
    {
	/* Means we have a better (or equally as good) matching
	 * of data features to model features.
	 */

	/* Copy/record the matches */
	for (i = 0; i < nDataFeatures; i++)
	    pITRect->matchesDtoM[i] = matchesDtoM[i];
	pITRect->nMatchesDtoM = nMatchesDtoM;

	*pNextNMatchesDtoMAsGood = cTrue;
    }
    else
    {
	*pNextNMatchesDtoMAsGood = cFalse;
    }


#if DEBUG_I_TREE
    i_printf (pIPWState, pIPWState->errorProc,
	      pIPWState->errorInfo,
	      "%ld of the %ld data features matched model features.\n",
	      nMatchesDtoM, nDataFeatures);


    for (i = 0; i < nDataFeatures; i++)
    {
	if (matchesDtoM[i] != NO_FEATURE_MATCHED) {
	    i_printf (pIPWState, pIPWState->errorProc,
		      pIPWState->errorInfo,
		      "D%2ld+M%2ld: Data rect [%6ld,%6ld,%6ld,%6ld] matched rect [%6ld,%6ld,%6ld,%6ld]\n",
		      i, matchesDtoM[i],
		      pITRect->drect[i].x, pITRect->drect[i].y, 
		      pITRect->drect[i].w, pITRect->drect[i].h,
		      pITRect->mrect[matchesDtoM[i]].x,
		      pITRect->mrect[matchesDtoM[i]].y,
		      pITRect->mrect[matchesDtoM[i]].w,
		      pITRect->mrect[matchesDtoM[i]].h);
	else
	    i_printf (pIPWState, pIPWState->errorProc,
		      pIPWState->errorInfo,
		      "D%2ld    : Data rect [%6ld,%6ld,%6ld,%6ld] null branch match\n",
		      i, 
		      pITRect->drect[i].x, pITRect->drect[i].y,
		      pITRect->drect[i].w, pITRect->drect[i].h);
	}
    }

#endif /*DEBUG_I_TREE*/


    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Match up a two lists rectangles via an interpretation tree matcher
 *
 * LongDescription:::
 *  Match up a list of "data" rectangles to a list of "model" rectangles.
 *  This routine tries to match these lists in a related way, in that we
 *  consider each list to be on a single page, and that the best matching
 *  should try to match keeping this in mind.  For example, data rects with
 *  widths and heights 50x50 and 100x100, where the 50x50 is _above_ the
 *  100x100 should not match model rects 50x50 and 100x100, where the 50x50
 *  is _below_ the 100x100!   <BR><BR>
 *  
 *  See the general description for routine i_interpTreeMatcher() for
 *  more details on the general matching method, which uses unary and binary
 *  constraints for general matching and a more exhaustive "model test"
 *  at each leaf explored to find matching correspondences.
 *
 * PrivateDescription::
 *  The method we use here is to use wp_interpTreeRectUnaryConstraints()
 *  for the unary constraints, wp_interpTreeRectBinaryConstraints() for
 *  the binary constraints, and wp_interpTreeRectModelTest() for the
 *  model test.<BR><BR>
 *
 *  We precompute arrays to use for checking "similarity" of widths and
 *  heights in the unary constraints, and precompute arrays for the
 *  binary constraints checking.  Then, we attempt to match at least
 *  1/2 of the data rectangles to model rectangles.  If that fails,
 *  we try 1/4.  If that fails, we'll take ANY!
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  drectsList::        Ptr to list of rectangles to use for data features
 *  nDataFeatures::     Number of rectangles in data rect list
 *  mrectsList::        Ptr to list of rectangles to use for model features
 *  nModelFeatures::    Number of rectangles in model rect list
 *  xres::              X resolution of rectangles
 *  yres::              Y resolution of rectangles
 *                      
 * Output:::
 *  matchesDtoM::       Ptr to created array of "nDataFeatures" elements.
 *                      Element 0 corresponds the 0th data rect, and so on.
 *                      On output, each element's value will be a number, i,
 *                      [0..nModelFeatures-1] corresponding to it matching
 *                      the the "i"th model rect, or NO_FEATURE_MATCHED
 *                      to indicate no suitable matching model rect was found.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  template matching
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_interpTreeRectMatcher (
    IPWState        *pIPWState,
    RECTANGL        *drectsList,
    Int32            nDataFeatures,
    RECTANGL        *mrectsList,
    Int32            nModelFeatures,
    Int32            xres,
    Int32            yres,
    Int32           *matchesDtoM)
{
    PROLOGUE;
    Int32               i, j;
    Bool                passed;
    interpTreeRectType  stuffIT;

    ERR_GEN(mrectsList == NULL  ||
	    drectsList == NULL  ||
	    matchesDtoM == NULL, cErrNullPtr);
    ERR_GEN(nModelFeatures <= 0  ||  nDataFeatures <= 0  ||  xres <= 0  ||
	    yres <= 0, cErrInvalidParm);

    /* Initialize pointers in event of cleanup */
    stuffIT.wmin  = NULL;
    stuffIT.mrect = NULL;
    stuffIT.drect = NULL;
    stuffIT.minDx = NULL;
    stuffIT.minDy = NULL;
    stuffIT.maxDx = NULL;
    stuffIT.maxDy = NULL;
    stuffIT.pIPWState = pIPWState;

    /* Set up the model's width and height ranges arrays.
     * We'll use these arrays for the unary constraint test, to determine
     * if a data rect's width and height are "similar" to that of a model 
     * rect's.
     */
    stuffIT.wmin = (Int32 *)IPW_CALLOC(4 * nModelFeatures, sizeof(Int32),
				       cAllocatorData);
    ERR_GEN(stuffIT.wmin == NULL,  cErrNomem);
    stuffIT.wmax = stuffIT.wmin + nModelFeatures;
    stuffIT.hmin = stuffIT.wmax + nModelFeatures;
    stuffIT.hmax = stuffIT.hmin + nModelFeatures;
    stuffIT.matchesDtoM = matchesDtoM;


    ERR_PASS_CLEANUP(w_createRectArrayList ( pIPWState,
					     nDataFeatures, 
					     &(stuffIT.drect)));
    ERR_PASS_CLEANUP(w_createRectArrayList ( pIPWState,
					     nModelFeatures,
					     &(stuffIT.mrect)));


    for (i = 0; i < nModelFeatures; i++)
    {
	stuffIT.mrect[i].x = mrectsList->x;
	stuffIT.mrect[i].y = mrectsList->y;
	stuffIT.mrect[i].w = mrectsList->w;
	stuffIT.mrect[i].h = mrectsList->h;
	mrectsList = mrectsList->next;
    }
    mrectsList = NULL;

    for (i = 0; i < nDataFeatures; i++)
    {
	stuffIT.drect[i].x = drectsList->x;
	stuffIT.drect[i].y = drectsList->y;
	stuffIT.drect[i].w = drectsList->w;
	stuffIT.drect[i].h = drectsList->h;
	drectsList = drectsList->next;
    }
    drectsList = NULL;

    /* Fill in the width and height ranges arrays */
    for (i = 0; i < nModelFeatures; i++)
    {
	stuffIT.wmin[i] = stuffIT.mrect[i].w - SIZETOL(xres);
	stuffIT.wmax[i] = stuffIT.mrect[i].w + SIZETOL(xres);
	stuffIT.hmin[i] = stuffIT.mrect[i].h - SIZETOL(yres);
	stuffIT.hmax[i] = stuffIT.mrect[i].h + SIZETOL(yres);
    }

    /* Make the four nModelFeatures by nModelFeatures arrays.
     * These arrays will be used for the binary constaint test, to determine
     * if the difference in the X(Y) positions between the upper left
     * corners of two data rects is "similar" to the difference in the
     * X(Y) positions between the upper left corners of two model rects.
     */

    /* IDEA: compress these into 2 tables where they are split into
     * the dX components on the upper diagonal, and
     * the dY on the lower.
     * Then, macros like MIN_DIFF_IN_X(min, a, b) which will reorder
     * a and b as needed and multiply by -1 as needed!
     */
    stuffIT.minDx = (Int16 **)IPW_CALLOC(nModelFeatures*4, sizeof(Int16 *),
					 cAllocatorData);
    ERR_GEN_CLEANUP(stuffIT.minDx == NULL,  cErrNomem);
    stuffIT.maxDx = stuffIT.minDx + nModelFeatures;
    stuffIT.minDy = stuffIT.maxDx + nModelFeatures;
    stuffIT.maxDy = stuffIT.minDy + nModelFeatures;
    
    stuffIT.minDx[0] = (Int16 *)IPW_CALLOC
	(nModelFeatures*nModelFeatures, sizeof(Int16), cAllocatorData);
    stuffIT.minDy[0] = (Int16 *)IPW_CALLOC
	(nModelFeatures*nModelFeatures, sizeof(Int16), cAllocatorData);
    stuffIT.maxDx[0] = (Int16 *)IPW_CALLOC
	(nModelFeatures*nModelFeatures, sizeof(Int16), cAllocatorData);
    stuffIT.maxDy[0] = (Int16 *)IPW_CALLOC
	(nModelFeatures*nModelFeatures, sizeof(Int16), cAllocatorData);
    ERR_GEN_CLEANUP(stuffIT.minDx[0] == NULL || stuffIT.minDy[0] == NULL ||
		    stuffIT.maxDx[0] == NULL || stuffIT.maxDy[0] == NULL,
		    cErrNomem);

    for (i = 1; i < nModelFeatures; i++)
    {
	stuffIT.minDx[i] = stuffIT.minDx[i-1] + nModelFeatures;
	stuffIT.minDy[i] = stuffIT.minDy[i-1] + nModelFeatures;
	stuffIT.maxDx[i] = stuffIT.maxDx[i-1] + nModelFeatures;
	stuffIT.maxDy[i] = stuffIT.maxDy[i-1] + nModelFeatures;
    }

    /* Fill in the {min,max}{Dx,Dy} arrays. */
    for (i = 0; i < nModelFeatures; i++)
    {
	for (j = 0; j < nModelFeatures; j++)
	{
	    stuffIT.minDx[i][j] = 
		stuffIT.mrect[j].x - stuffIT.mrect[i].x - TRANSTOL(xres);

	    stuffIT.maxDx[i][j] = 
		stuffIT.mrect[j].x - stuffIT.mrect[i].x + TRANSTOL(xres);

	    stuffIT.minDy[i][j] = 
		stuffIT.mrect[j].y - stuffIT.mrect[i].y - TRANSTOL(yres);

	    stuffIT.maxDy[i][j] =
		stuffIT.mrect[j].y - stuffIT.mrect[i].y + TRANSTOL(yres);
	}
    }



    /* Initialize to no matches */
    stuffIT.nMatchesDtoM = 0;

    /* Try first for a fast/"works most of the time" check for at least
     * 1/2 the data rects matching model rects.
     */
    ERR_PASS_CLEANUP(
	i_interpTreeMatcher (pIPWState, nModelFeatures, nDataFeatures,
			     nDataFeatures / 2, /* Min data feature matches! */
			     wp_interpTreeRectUnaryConstraints,
			     wp_interpTreeRectBinaryConstraints,
			     wp_interpTreeRectModelTest,
			     (void *)&stuffIT, 
			     &passed));
    if (!passed)
    {
	/* Initialize to no matches */
	stuffIT.nMatchesDtoM = 0;

	/* Try at least 1/4 */
	ERR_PASS_CLEANUP(
	    i_interpTreeMatcher (pIPWState, nModelFeatures, nDataFeatures,
				 nDataFeatures / 4, 
				 wp_interpTreeRectUnaryConstraints,
				 wp_interpTreeRectBinaryConstraints,
				 wp_interpTreeRectModelTest,
				 (void *)&stuffIT, 
				 &passed));

	if (!passed)
	{
	    /* Initialize to no matches */
	    stuffIT.nMatchesDtoM = 0;

	    /* We're desparate -- report any matching! */
	    ERR_PASS_CLEANUP(
		i_interpTreeMatcher (pIPWState, nModelFeatures, nDataFeatures,
				     0,
				     wp_interpTreeRectUnaryConstraints,
				     wp_interpTreeRectBinaryConstraints,
				     wp_interpTreeRectModelTest,
				     (void *)&stuffIT, 
				     &passed));
	}

    }
    

    /* Clean up */
    IPW_FREE(stuffIT.wmin, cAllocatorData );
    IPW_FREE(stuffIT.mrect, cAllocatorData);
    IPW_FREE(stuffIT.drect, cAllocatorData);
    IPW_FREE(stuffIT.minDx[0], cAllocatorData);
    IPW_FREE(stuffIT.minDy[0], cAllocatorData);
    IPW_FREE(stuffIT.maxDx[0], cAllocatorData);
    IPW_FREE(stuffIT.maxDy[0], cAllocatorData);
    IPW_FREE(stuffIT.minDx, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(stuffIT.wmin, cAllocatorData );
    IPW_FREE(stuffIT.mrect, cAllocatorData);
    IPW_FREE(stuffIT.drect, cAllocatorData);
    IPW_FREE(stuffIT.minDx[0], cAllocatorData);
    IPW_FREE(stuffIT.minDy[0], cAllocatorData);
    IPW_FREE(stuffIT.maxDx[0], cAllocatorData);
    IPW_FREE(stuffIT.maxDy[0], cAllocatorData);
    IPW_FREE(stuffIT.minDx, cAllocatorData);

    ERR_RETURN;
}


/*

 $Log: itree.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.2  1998/01/14 00:13:27  ddavies
 itree.c -> /product/ipcore/ipwrap/src/RCS/itree.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.1  1996/02/22 01:47:30  ejaquith
 Initial revision


 */
