/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2000, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	iw_cjpeg.c:	JPEG compression wrappers:
 *
 *	Code to compress PIXRs to JPEG compressed (JFIF formatted) files:
 *		w_jCompress()
 *		w_jCompress_callback()
 *
 *	Or, you may call the following:
 *		w_jcInit() - begin
 *        ** optional routine to help input control **
 *		w_jcSetDPI() - to set DPI
 *              w_jcGetImageLinesPerMCU - to get lines in one MCU row
 *              w_jcGetMCURowsPerImage - to get MCU rows in whole image
 *              w_jcGetErrorStr() - to get error description string
 *              w_jcFreeErrorStr() - to free memory stroing error message
 *		w_jcAllocMCURowBuffer() - buffer pointers for one MCU row
 *              w_jcMCURowFree() - free MCU buffer points
 *        **                                       **
 *		w_jcMCURow() - compress one MCU row
 *		w_jcTerm() - clean up
 */


#define _IW_CJPEG_C_

#ifndef _WIN32_WCE
#include <stdio.h>
#endif

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "jpeglib.h"
#include "imageref.pub"
#include "defines.pub"

/* prototypes */
#include "shrpixr.pub"
#include "props.pub"
#include "jpeg.pub"
#include "iw_jpeg.pub"

IP_RCSINFO(iw_cjpeg_c_RCSInfo, "$RCSfile: iw_cjpeg.c,v $; $Revision: 1.12 $; $Date: 2002/04/16 23:01:15 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a JPEG table to convert 4 bit pixels to 8 bits
 *
 * LongDescription:::
 *  Create a LUT (Look Up Table) to convert 4 bit pixels to eight bits
 *  Converts two nibbles (one byte) to two bytes (UInt16)
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  pLut::           Ptr to ptr to lookup table.
 *                   If ptr to table is NULL on input, table is allocated
 *                   and filled.
 *                   Else table is just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcBuild4To8Lut (
    IPWState             *pIPWState,
    UInt16              **pLut)
{
    PROLOGUE;
    register UInt16 data;
    register UInt16 index;
    UInt16 *lut;

    ERR_GEN(pLut == NULL, cErrNullPtr);

    if (*pLut == NULL)
    {
	lut = (UInt16 *)IPW_CALLOC(sizeof(UInt16), 256, cAllocatorData);
	ERR_GEN_CLEANUP(lut == NULL, cErrNomem);
    }
    else
        lut = *pLut;

    for (index = 0; index < 256; index++)
    {
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt  
/* Reverse the order for PC's */
        data = (index >> 4) | ((index & 0x000F) << 8);
#else
        data = ((index & 0x00F0) << 4) | (index & 0x000F);
#endif
        lut[index] = data | (data << 4);
    }

    *pLut = lut;
    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Expand a single line of a 4 bit buffer to 8 bits/pixel.
 *
 * LongDescription:::
 *  Expand a single line of a 4 bit buffer to 8 bits/pixel.
 *  Both source and dest are in memory order.
 *  The caller must create the 4 bit to 8 bit look up table lut4To8.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pS::             Ptr to source line in 4 bit buffer.
 *  w::              Image width
 *  lut4To8::        Ptr to 4 bit to 8 bit look up table.
 *
 * Output:::
 *  pD::             Ptr to dest line in 8 bit buffer.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcCopyLineMem4ToMem8 (
    IPWState           *pIPWState,
    UInt8              *pS,
    UInt32              w,
    UInt16             *lut4To8,
    UInt8              *pD)
{
    PROLOGUE;
    UInt32   wOver2;
    register UInt8  *psByte;
    register UInt16 *pdWord;
    register UInt32 count;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    wOver2 = (w+1)/2;

    psByte = pS;
    pdWord = (UInt16 *) pD;
    count = wOver2 + 1;

    while (--count)
    {
        /* Look up each byte in lut4To8, which expands the two nibbles
         * into two words
         */
        *pdWord++ = lut4To8[*psByte];
        psByte++;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Expand a single line of a 4 bit PIXR to 8 bits/pixel.
 *
 * LongDescription:::
 *  Expand a single line of a 4 bit PIXR to 8 bits/pixel.
 *  The destination is in memory order (not PIXR).
 *  The caller must create the 4 bit to 8 bit look up table lut4To8.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pS::             Ptr to source line in 4 bit PIXR
 *  w::              Image width
 *  lut4To8::        Ptr to 4 bit to 8 bit look up table.
 *
 * Output:::
 *  pD::             Ptr to dest line in 8 bit buffer
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcCopyLine4ToMem8 (
    IPWState           *pIPWState,
    UInt8              *pS,
    UInt32              w,
    UInt16             *lut4To8,
    UInt8              *pD)
{
    PROLOGUE;
    UInt32   wOver2;
    register UInt8  *psByte;
    register UInt16 *pdWord;
    register UInt32 count;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    wOver2 = (w+1)/2;

    psByte = pS;
    pdWord = (UInt16 *) pD;
    count = wOver2 + 1;

    while (--count)
    {
        /* Look up each byte in lut4To8, which expands the two nibbles
         * into two words. Access the src in PIXR order with UCHAR_ACCESS
         */
        *pdWord++ = lut4To8[UCHAR_ACCESS(psByte)];
        psByte++;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert data from BSQ format (PIXR order) into interleaved RGB.
 *
 * LongDescription:::
 *  Convert the user's data from BSQ format (PIXR order) into interleaved RGB.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  src::            Ptr to 3D array of RGB data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *  xzhu :: changed Deinterleave into Interleave.
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcInterleavePixrMCURow (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
     UInt8              ***src)
{
    PROLOGUE;

    UInt8  *rptr, *gptr, *bptr, *buf;
    Int32  idx, w, row, nrows;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    w = cinfo->j_cinfo.image_width;
    ERR_PASS(jc_get_linesToRead(pIPWState, &cinfo->j_cinfo, &nrows));

    for(row=0; row<nrows; row++)
    {
	buf = cinfo->jpegBuf[0][row];
	rptr = src[0][row];
	gptr = src[1][row];
	bptr = src[2][row];

	for ( idx =0; idx < w; ++idx )
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	    *buf++ = UCHAR_ACCESS(rptr++);
	    *buf++ = UCHAR_ACCESS(gptr++);
	    *buf++ = UCHAR_ACCESS(bptr++);
#else
	    *buf++ = *rptr++;
	    *buf++ = *gptr++;
	    *buf++ = *bptr++;
#endif
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert data from BSQ format into interleaved RGB.
 *
 * LongDescription:::
 *  Convert the user's data from BSQ format into interleaved RGB.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  src::            Ptr to 3D array of RGB data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *  xzhu :: changed Deinterleave into Interleave.
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcInterleaveMCURow (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
     UInt8              ***src)
{
    PROLOGUE;

    UInt8  *rptr, *gptr, *bptr, *buf;
    Int32  idx, w, row, nrows;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    w = cinfo->j_cinfo.image_width;
    ERR_PASS(jc_get_linesToRead(pIPWState, &cinfo->j_cinfo, &nrows));

    for(row=0; row<nrows; row++)
    {
	buf = cinfo->jpegBuf[0][row];
	rptr = src[0][row];
	gptr = src[1][row];
	bptr = src[2][row];

	for ( idx =0; idx < w; ++idx )
	{
	    *buf++ = *rptr++;
	    *buf++ = *gptr++;
	    *buf++ = *bptr++;
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy the user's buffer pointers to the jpegBuf pointers.

 * LongDescription:::
 *  Copy the user's buffer pointers to the jpegBuf pointers,
 *  i.e just point to the user's data rather than copy it.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  src::            Ptr to 2D array of user's data buffer pointers
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcCopyRowPtrs (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
    UInt8              ***src)
{
    PROLOGUE;
    Int32 row;
    Int32 nrows;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

/* Get the image size and band count */
    ERR_PASS(jc_get_linesToRead(pIPWState, &cinfo->j_cinfo, &nrows));

    for (row=0; row<nrows; row++)
	cinfo->jpegBuf[0][row] = src[0][row];

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Upconvert line of 4 bit gray buffer pixels to 8 bit grays.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  src::            Ptr to 2D array of user's data buffer pointers
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcCvtGray4ToGray8 (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
    UInt8            ***src)
{
    PROLOGUE;
    Int32 row, nrows, nsamps;

/* Get the image size */
    ERR_PASS(jc_get_linesToRead(pIPWState, &cinfo->j_cinfo, &nrows));
    nsamps = cinfo->j_cinfo.image_width;

/* Upconvert 4 bit pixels to 8 bits */
    for (row=0; row<nrows; row++)
        ERR_PASS( wp_jcCopyLineMem4ToMem8(pIPWState,
					 src[0][row],
					 nsamps,
					 cinfo->plut,
					 cinfo->jpegBuf[0][row]));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Upconvert line of 4 bit gray PIXR pixels to 8 bit grays.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  src::            Ptr to 2D array of user's data buffer pointers
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcCvtGray4ToMem (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
    UInt8            ***src)
{
    PROLOGUE;
    Int32 row, nrows, nsamps;

/* Get the image size and band count */
    ERR_PASS(jc_get_linesToRead(pIPWState, &cinfo->j_cinfo, &nrows));
    nsamps = cinfo->j_cinfo.image_width;

    for (row=0; row<nrows; row++)
        ERR_PASS( wp_jcCopyLine4ToMem8(pIPWState,
				       src[0][row],
				       nsamps,
				       cinfo->plut,
				       cinfo->jpegBuf[0][row]));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Convert line of PIXR pixels to buffers of pixels in memory format.
 *
 * LongDescription:::
 *  Do byteswapping as necessary (ie, if on PC platform).
 *  Else just copy the row pointers.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  src::            Ptr to 2D array of user's data buffer pointers
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion, byteswapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcCvtToMem (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
    UInt8              ***src)
{
    PROLOGUE;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt  
    Int32 row, samp;
    Int32 nrows, nsamps;

    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(src == NULL, cErrNullPtr);

/* Get the image size and band count */
    ERR_PASS(jc_get_linesToRead(pIPWState, &cinfo->j_cinfo, &nrows));
    nsamps = cinfo->j_cinfo.image_width;

/*
  Do the byte swap - note that fast swap is in 32 bit words.
  We'll do the full 32 bit words with the fast swap
  and do the stragglers with UCHAR_ACCESS.
*/
    for (row=0; row<nrows; row++)
    {
	cinfo->byteSwapProc(pIPWState, (UInt32 *)(src[0][row]), 
                                (UInt32 *)(cinfo->jpegBuf[0][row]),
                                (Int32) (nsamps/4));
	for (samp=nsamps&(~3); samp<nsamps; samp++)
	    cinfo->jpegBuf[0][row][samp] =
                UCHAR_ACCESS(src[0][row]+samp);
    }

#else
/* We're on a big-endian machine; just copy ptrs */

    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(src == NULL, cErrNullPtr);

    ERR_PASS( wp_jcCopyRowPtrs(pIPWState, cinfo, src));

#endif  

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set new Q-table.
 *
 * LongDescription:::
 *  Set new Q-table according to selection. It must be called after w_jcInit.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *  quality::        Compression quality setting (0 ... 100).
 *  jcOptions::      Options for progressive and qtable defined 
 *                   in iw_jpeg.pub. JC_QTABLE_IMPACT is to set impact
 *                   Qtable lu_qtabel = ch_qtable = NULL and 
 *                   JC_QTABLE_INPUT is to set Qtable in lu_qtable & ch_qtable.
 *  lu_qtable::      Ptr to luminance q-table
 *  ch_qtable::      Ptr to chrominance q-table
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcSetQtable(
   IPWState           *pIPWState,
   CompressInfo       *cinfo, 
   Int32               quality,
   Int32               jcOptions,
   UInt16             *lu_qtable,
   UInt16             *ch_qtable)
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    if( jcOptions & JC_QTABLE_IMPACT )
    {
	UInt16  luTab[64] = {16,  24,  28,  32,  36,  40,  52,  72,
			    24,  24,  28,  32,  36,  40,  60,  68,
			    28,  28,  32,  36,  40,  58,  68,  56,
			    32,  36,  40,  48,  52,  68,  80,  62,
			    36,  46,  52,  56,  68,  110, 104, 78,
			    40,  52,  56,  68,  82,  104, 114, 92,
			    52,  64,  68,  88,  104, 122, 120, 102,
			    72,  68,  96,  98,  112, 100, 104, 100};
	UInt16  chTab[64] = {36,  38,  40,  48,  100, 100, 100, 100,
			    38,  38,  40,  66,  100, 100, 100, 100,
			    40,  38,  56,  100, 100, 100, 100, 100,
			    48,  66,  100, 100, 100, 100, 100, 100,
			    100, 100, 100, 100, 100, 100, 100, 100,
			    100, 100, 100, 100, 100, 100, 100, 100,
			    100, 100, 100, 100, 100, 100, 100, 100,
			    100, 100, 100, 100, 100, 100, 100, 100};

	ERR_PASS( jc_set_luminance_qtable (pIPWState, &cinfo->j_cinfo,
					   quality, luTab));
	ERR_PASS( jc_set_chrominance_qtable (pIPWState, &cinfo->j_cinfo,
					     quality, chTab));

    }
    else if( lu_qtable != NULL && ch_qtable != NULL &&
	     (jcOptions & JC_QTABLE_INPUT) )
    {
	ERR_PASS( jc_set_luminance_qtable (pIPWState, &cinfo->j_cinfo,
					   quality, lu_qtable));
	ERR_PASS( jc_set_chrominance_qtable (pIPWState, &cinfo->j_cinfo,
					     quality, ch_qtable));
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Extract the tables, as a separate stream.
 *
 * LongDescription:::
 *  Writes the Q-table(s) and Huffman tables, in a separate data stream,
 *  and marks the subsequent compression call as not needing these tables.
 *  This is useful when writing TIFF files with the JPEGTables field.
 *  The output is written to the output stream.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *  quality::        Compression quality setting (0 ... 100).
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcExtractTables(
   IPWState           *pIPWState,
   CompressInfo       *cinfo, 
   Int32               quality)
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    ERR_PASS( jc_extract_tables (pIPWState, &cinfo->j_cinfo, quality));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get line number in one MCU row for compression.
 *
 * LongDescription:::
 *  Get line number in one MCU row for compression.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  lines::          lines in one MCU row.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcGetImageLinesPerMCU(
   IPWState           *pIPWState,
   CompressInfo       *cinfo, 
   Int32              *lines)
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    *lines = cinfo->j_cinfo.max_v_samp_factor * DCTSIZE;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get MCU rows in the image for compression.
 *
 * LongDescription:::
 *  Get MCU rows in the image for compression.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  rows::           MCU rows in image.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcGetMCURowsPerImage(
   IPWState           *pIPWState,
   CompressInfo       *cinfo, 
   Int32              *rows)
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    *rows = cinfo->j_cinfo.MCU_rows_in_scan;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Wrapper for JPEG compression initialization.
 *
 * LongDescription:::
 *    Performs the initialization steps for JPEG compression.
 *    Callers specify the memory order of the input image.
 *    This interface was written for EasyScan/Amber, so that
 *    data could be processed from non-pixr sources. The write
 *    callback will almost always be an fwrite statement, unless
 *    the compressed data is to go to a memory buffer. The difficulty
 *    with the latter is that the size of the compressed image
 *    is not known until AFTER the compression has been performed.
 *    (JPEG is an open-loop compressor. There is currently no
 *    target rate driven version implemented).
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::             Ptr to compression info
 *  imageWidth::        Image width in pixels
 *  imageHeight::       Number of scan lines in image
 *  numComponents::     Number of color channels in image
 *  quality::           Compression quality setting (0 ... 100).
 *                      Higher is better. Settings of less
 *                      than 50 may degrade quality too much.
 *  jcOptions::         Options for progressive and qtable defined 
 *                      in iw_jpeg.pub. JC_DEFAULT is no progressive and 
 *                      use default Qtable.
 *  XDpi::              Horizontal resolution dots per inch. If 0, unknow
 *  YDpi::              Vertical resolution dots per inch. If 0, unknow
 *  WriteCallback::     Callback function to write out compressed data.
 *  userData::          Ptr to hook for passing in user data (only
 *                      used to pass FILE desriptor now).
 *  qTab::              Point to Qtable, the first 64 is luminance Qtable,
 *                      and the next 64 is chrominance Qtable.
 *  memOrg::            Memory organization of image to be compressed.
 *                       May be one of:
 *                       JMEMORG_BIP         -Interleaved RGB;
 *                       JMEMORG_BSQ_MEM     -Separate bands - memory order;
 *                       JMEMORG_BSQ_PIXR    -Separate bands - pixr order;
 *                       JMEMORG_GRAY8_MEM   -8 bit gray - memory order;
 *                       JMEMORG_GRAY8_PIXR  -8 bit gray - pixr order;
 *                       JMEMORG_GRAY4_MEM   -4 bit gray - memory order;
 *                       JMEMORG_GRAY4_PIXR  -4 bit gray - pixr order;
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcInit (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
    Int32               imageWidth,
    Int32               imageHeight,
    Int32               numComponents,
    Int32               quality,
    Int32               jcOptions,
    Int32               XDpi,
    Int32               YDpi,
    ErrCode	(CDECL *WriteCallback) (IPWState *pIPWState,
					void * userData, UInt8 *buf,
					UInt32 nbytes, UInt32 *nCount),
    void               *userData,
    UInt16             *qTab,
    UInt32              memOrg)
{
    PROLOGUE;
    Bool                progressive = cFalse;

    if( jcOptions & JC_PROGRESSIVE )
	progressive = cTrue;

/* Do a sanity check on the arguments */
    ERR_GEN(imageWidth <=0 || imageWidth > 32768, cErrInvalidParm);
    ERR_GEN(imageHeight <=0 || imageHeight > 32768, cErrInvalidParm);
    ERR_GEN(numComponents <=0 || numComponents > 3 ||
	    quality <0 || quality > 100, cErrInvalidParm);
    ERR_GEN(WriteCallback == NULL, cErrNullPtr);
    ERR_GEN(cinfo == NULL, cErrNullPtr);


    cinfo->userMemOrg = memOrg;

/* Use this funny cast to get around compiler grumblings in the call
 * to jcInit.
 */
    cinfo->LocalJCSpecialOrders = NULL;

    cinfo->plut = NULL;
    cinfo->jpegBuf = NULL;

/* Call the ipshared init routine */
    ERR_PASS( jc_init(pIPWState, &cinfo->j_cinfo, 
		    imageWidth,
		    imageHeight,
		    numComponents,
		    quality,
		    progressive,
		    XDpi,
		    YDpi,
		    WriteCallback,
		    userData));
    
    if ( (jcOptions & JC_QTABLE_IMPACT) ||
	 ((jcOptions & JC_QTABLE_INPUT) && qTab != NULL ))
    {
	UInt16 *pTab = NULL;
	if( qTab != NULL) pTab = qTab + 64;
	ERR_PASS( wp_jcSetQtable( pIPWState, cinfo,
				  quality, jcOptions,
				  qTab, pTab));
    }

    /* Mark as special color space */
    if (jcOptions & (JC_CIELAB | JC_ITULAB)) {
	/* Normally, the IJG library uses the UNKNOWN colorspace for "other"
	 * colorspaces, but they assume that the "other" colorspaces do not
	 * include chrominance components.  I've added a case where the first
	 * component is luminance, and the rest are chrominance, since the
	 * UNKNOWN type triggers: 1) no chrominance subsampling, and 2) the
	 * first set of tables are used (the second set are for the chrominance
	 * components). */
	ERR_PASS( jc_set_colorspace(pIPWState, &cinfo->j_cinfo,
	    JCS_UNKNOWN_W_CHROM));
    } else if (jcOptions & (JC_UNKNOWN)) {
	/* Could be any other color space: RGB, CMYK, etc. */
	ERR_PASS( jc_set_colorspace(pIPWState, &cinfo->j_cinfo,
	    JCS_UNKNOWN));
    }

    /* (un)set chrominance channel subsampling */
    if (jcOptions & JC_NO_CHROMA_SS) {
	ERR_PASS(jc_set_sampling_factors (pIPState, &cinfo->j_cinfo,
	    1, 1, 1, 1, 1, 1));
    }

    /* use one set of tables */
    if (jcOptions & JC_JPEGTABLES) {
	ERR_PASS(w_jcExtractTables(pIPWState, cinfo, quality));
    }

    ERR_PASS(jc_startCompress(pIPWState, &cinfo->j_cinfo));

/* Allocate the memory buffer or pointer array */
    switch (memOrg)
    {
	/*
	  In these two cases the user's memory organization
	  is identical to JPEG's on any CPU, so we won't need to copy
	  the data to jpegBuf. Therefore we won't actually
	  allocate any memory (cFalse arg), but just set the jpegBuf
	  pointers to point to the user's data. This
	  optimization should save both time and memory.
	  */
        case JMEMORG_BIP:
        case JMEMORG_GRAY8_MEM:
	    ERR_PASS( w_jcAllocMCURowBuffer(pIPWState, cinfo, cFalse,
					     &(cinfo->jpegBuf)));
	    ERR_GEN(cinfo->jpegBuf == NULL, cErrNomem);
	    cinfo->jpegBufLineMem = cFalse;
	    break;

    /*
      In the following cases, data conversions MUST be made on any
      CPU, so we'll need to actually allocate some line buffers
      (cTrue arg) to hold the converted data.  There is a
      separately allocated buffer for each line of each channel in
      the jpegBuf.
    */
        case JMEMORG_BSQ_MEM:
        case JMEMORG_BSQ_PIXR:
        case JMEMORG_GRAY4_MEM:
        case JMEMORG_GRAY4_PIXR:
	    ERR_PASS( w_jcAllocMCURowBuffer(pIPWState, cinfo, cTrue, 
						 &(cinfo->jpegBuf)));
	    ERR_GEN(cinfo->jpegBuf == NULL, cErrNomem);
	    cinfo->jpegBufLineMem = cTrue;
	    break;

    /*
      In the following cases, data conversions MAY be made,
      depending on the CPU. For Big-endian machines (Sun, Mac),
      PIXR and natural memory order are identical, so we can just
      alloc line pointers. For little-endian machines (PC) we need
      to do byte swapping between Pixrs and memory buffers, so real
      line buffers (cTrue arg) are required to hold the converted
      data.  There is a separately allocated buffer for each line of
      each channel in the jpegBuf.  
    */

        case JMEMORG_GRAY8_PIXR:
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt  
	    ERR_PASS( w_jcAllocMCURowBuffer(pIPWState, cinfo, cTrue, 
						 &(cinfo->jpegBuf)));
	    cinfo->jpegBufLineMem = cTrue;
#else
	    ERR_PASS( w_jcAllocMCURowBuffer(pIPWState, cinfo, cFalse, 
						 &(cinfo->jpegBuf)));
	    cinfo->jpegBufLineMem = cFalse;
#endif
	    ERR_GEN(cinfo->jpegBuf == NULL, cErrNomem);
	    break;

        default:
	    ERR_GEN(1, cErrInvalidParm);
    }

    /*
      These cases require data conversion.
      Set a function pointer to the conversion routine
    */
    switch (memOrg)
    {
        case JMEMORG_BIP:   /* Band interleaved by pixel */
	    cinfo->cvtCompImage = wp_jcCopyRowPtrs;
	    break;
        case JMEMORG_BSQ_MEM: /* Band sequential - memory order */
	    cinfo->cvtCompImage = wp_jcInterleaveMCURow;
	    break;
        case JMEMORG_BSQ_PIXR: /* Band sequential - PIXR order */
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt  
	    i_getDIBPixrLineProc (pIPWState, cFalse, cPixrToPixr, cFalse,
			     NULL, NULL, &cinfo->byteSwapProc, NULL);
#endif
	    cinfo->cvtCompImage = wp_jcInterleavePixrMCURow;
	    break;
        case JMEMORG_GRAY8_PIXR: /* 8 bpp - PIXR order */
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt  
	    i_getDIBPixrLineProc (pIPWState, cFalse, cPixrToPixr, cFalse,
			     NULL, NULL, &cinfo->byteSwapProc, NULL);
#endif
	    cinfo->cvtCompImage = wp_jcCvtToMem;
	    break;
        case JMEMORG_GRAY8_MEM: /* 8 bpp gray - memory order */
	    cinfo->cvtCompImage = wp_jcCopyRowPtrs;
	    break;
        case JMEMORG_GRAY4_MEM: /* 4 bpp - memory order */
	    ERR_PASS( wp_jcBuild4To8Lut(pIPWState, &(cinfo->plut)));
	    cinfo->cvtCompImage = wp_jcCvtGray4ToGray8;
	    break;
        case JMEMORG_GRAY4_PIXR: /* 4 bpp - PIXR order */
	    ERR_PASS( wp_jcBuild4To8Lut(pIPWState, &(cinfo->plut)));
	    cinfo->cvtCompImage =  wp_jcCvtGray4ToMem;
	    break;
        default:
	    break;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Perform JPEG compression on one MCU row of an image.
 *
 * LongDescription:::
 *    Performs JPEG compression on one MCU (minimum coded unit)
 *    row of the image. This is 8 lines for a gray scale image
 *    or 16 lines for an RGB image. The output goes to the file
 *    or buffer desrribed by the userData parameter to w_jcInit,
 *    using the writeCallback function.
 *  
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  userBuf::        The buffer containing the data to be compressed.
 *                   (Ptr to 2D array of user's data buffer pointers)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcMCURow (
    IPWState             *pIPWState,
    Compress_info_ptr     cinfo,
    void                 *userBuf)
{
    PROLOGUE;
    Int32       lineCount;

    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(userBuf == NULL, cErrNullPtr);

/*
  Convert the input data into the form the JPEG compressor wants.
  JPEG wants the data arranged in memory order (not PIXR). It also
  wants each row in a separate memory buffer, allocated as a
  JSAMPIMAGE, i.e. UInt8 * jpegBuf[nbands][nrows]

  We've already assigned the proper procedure to the
  function pointer cvtCompImage, so just call that function
*/
    ERR_PASS( cinfo->cvtCompImage(pIPWState, cinfo, (unsigned char ***)userBuf));

/*
  Everything is now set up as the ipshared JPEG code expects it,
  so we can now call the function to compress the MCU row into
  the JFIF file format
*/
    ERR_PASS(jc_get_linesToRead(pIPWState, &cinfo->j_cinfo, &lineCount));

    ERR_PASS(jc_write_scanline(pIPWState, &cinfo->j_cinfo,
			       cinfo->jpegBuf[0], lineCount));
 
    
   NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform cleanup operations for termination of JPEG compression.
 *
 * LongDescription:::
 *    Performs cleanup operations for JPEG compression,
 *    including file trailer writing and closing and memory freeing.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcTerm (
    IPWState             *pIPWState,
    Compress_info_ptr     cinfo)
{
    PROLOGUE;
 
    ERR_GEN(cinfo == NULL, cErrNullPtr);

    if(cinfo->jpegBuf != NULL)
    {
	ERR_PASS(w_jcMCURowFree(pIPWState, cinfo,cinfo->jpegBufLineMem, 
				cinfo->jpegBuf));
	cinfo->jpegBuf = NULL;
    }
    if(cinfo->plut != NULL)
    {
	IPW_FREE(cinfo->plut, cAllocatorData);
	cinfo->plut = NULL;
    }
    ERR_PASS( jc_term(pIPWState, &cinfo->j_cinfo));
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Retrive error string for display error message.
 *
 * LongDescription:::
 *       Allows an application to retrieve a string containing
 *       an error string for display in an error message box
 *       or other such purpose.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  cinfo::          Ptr to compression info
 *
 * Output:::
 *  string::         Ptr to error string address
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcGetErrorStr(
    IPWState		 *pIPWState,
    Compress_info_ptr     cinfo,
    char                  **string)
{
    PROLOGUE;
 
    ERR_CHECK_NO_MSG( jc_get_error_string(pIPWState, &cinfo->j_cinfo, string) )
	ERR_RETURN;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Free memory stroing error meaasge.
 *
 * LongDescription:::
 *       In order to retrieve an error string, the error struct hasn't
 *    been freed in error exit. Provide this function to free memory
 *    storing all error messages and function points.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  cinfo::          Ptr to compression info
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcFreeErrorStr(
    IPWState		 *pIPWState,
    Compress_info_ptr     cinfo)
{
    PROLOGUE;

    ERR_CHECK_NO_MSG(jc_free_error_string(pIPWState,  &cinfo->j_cinfo ))
	ERR_RETURN;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Provide interface wrapper for allocating JPEG compression row buffer.
 *
 * LongDescription:::
 *    Provide a general interface wrapper for allocating the row buffer.
 *    This will selectively use either the normal ipshared function
 *    or the special function for BIP data, depending on the
 *    setting of cinfo.userMemOrg.  Thus this function MUST be called
 *    only AFTER w_jcInit is called.
 *  
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  allocLineMem::   Boolean. cTrue if memory is actually to
 *                   be allocated to hold the uncompressed image
 *                   data. If cFalse, a set of row pointers is
 *                   returned. The caller is expected to set
 *                   the pointers to hir image buffers.  cFalse
 *                   is to be used only when there is no
 *                   memory organization conversion needed
 *                   between the caller's buffer and the JPEG
 *                   compressor's internal buffer. The JPEG
 *                   internal buffer is in the CPU's  natural 
 *                   memory order, with a separate buffer for
 *                   each band (BSQ). Thus, if the data is in PIXR
 *                   order, cFalse can be used for the Sun, but
 *                   not for the PC.
 *
 * Output:::
 *  pData::          Ptr to the returned row buffer.
 *                   The buffers (allocLineMem=cTrue)
 *                   or row pointers (allocLineMem=cFalse)
 *                   allocated to hold (or point to) the image data.
 *                   The JSAMPIMAGE typedef is equivalent to UInt8 ***, and
 *                   holds data organized as buf[colorBand][line][column].
 *
 * Returns:::
 *
 * Notes:::
 *           Please remember to free MCU budder by calling
 *           w_jcMCUFree().
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcAllocMCURowBuffer (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
    Bool                allocLineMem,
    JSAMPIMAGE         *pData)
{
    PROLOGUE;
    JSAMPIMAGE  data = NULL;
    Int32 band, row, nlines, nsamps, nbands;
    
    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(pData == NULL, cErrNullPtr);
    ERR_GEN(*pData != NULL, cErrPtrNotNull); 

    ERR_PASS( w_jcGetImageLinesPerMCU(pIPWState, cinfo, &nlines) );
    nsamps = cinfo->j_cinfo.image_width;
    switch (cinfo->userMemOrg)
    {
        case JMEMORG_BIP:
        case JMEMORG_GRAY8_MEM:
        case JMEMORG_GRAY4_MEM:
        case JMEMORG_GRAY4_PIXR:
        case JMEMORG_GRAY8_PIXR:
	    nbands =1;
	    break;

        case JMEMORG_BSQ_MEM:
        case JMEMORG_BSQ_PIXR:
	    nbands = 3;
	    break;

        default:
	    nbands = 1; /* Force to a single band (with 24 bit pixels) */
    }

/* Allocate the channel pointers */
    data = (JSAMPIMAGE) IPW_CALLOC(sizeof(JSAMPARRAY), nbands, cAllocatorData);
    ERR_GEN_CLEANUP(data == NULL, cErrNomem);

    for (band=0; band < nbands; band++)
    {
        /* Allocate the row pointers */
        data[band] = (JSAMPARRAY) IPW_CALLOC(sizeof(JSAMPROW), 
					     nlines, cAllocatorData);
	ERR_GEN(data[band] == NULL, cErrNomem);

        for (row=0; row<nlines; row++)
        {
            if (allocLineMem) /* Alloc 24 bit row mem in 32 bit chunks */
            {
                data[band][row] = (JSAMPROW)IPW_CALLOC(sizeof(Int32), 
				      ((3*nsamps+3)/4), cAllocatorData);
		ERR_GEN(data[band][row] == NULL, cErrNomem);
            }
            else   /* Just set the row pointers to NULL */
                data[band][row] = (JSAMPROW) 0;
        }
 
    }

    *pData = data;
 
    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Provide interface wrapper for free JPEG compression row buffer.
 *
 * LongDescription:::
 *    Provide a general interface wrapper for free the row buffer.
 *  
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to compression info
 *  allocLineMem::   Boolean. cTrue if memory is actually to
 *                   be allocated to hold the uncompressed image
 *                   data. If cFalse, a set of row pointers is
 *                   returned. The caller is expected to set
 *                   the pointers to hir image buffers.  cFalse
 *                   is to be used only when there is no
 *                   memory organization conversion needed
 *                   between the caller's buffer and the JPEG
 *                   compressor's internal buffer. The JPEG
 *                   internal buffer is in the CPU's  natural 
 *                   memory order, with a separate buffer for
 *                   each band (BSQ). Thus, if the data is in PIXR
 *                   order, cFalse can be used for the Sun, but
 *                   not for the PC.
 *  pData::          Ptr to the row buffer.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jcMCURowFree (
    IPWState           *pIPWState,
    CompressInfo       *cinfo,
    Bool                allocLineMem,
    JSAMPIMAGE         pData)
{
    PROLOGUE;
    Int32 band, row, nlines, nsamps, nbands;
    
    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(pData == NULL, cErrNullPtr);
 
    ERR_PASS( w_jcGetImageLinesPerMCU(pIPWState, cinfo, &nlines) );
    nsamps = cinfo->j_cinfo.image_width;
    switch (cinfo->userMemOrg)
    {
        case JMEMORG_BIP:
        case JMEMORG_GRAY8_MEM:
        case JMEMORG_GRAY4_MEM:
        case JMEMORG_GRAY4_PIXR:
        case JMEMORG_GRAY8_PIXR:
	    nbands =1;
	    break;

        case JMEMORG_BSQ_MEM:
        case JMEMORG_BSQ_PIXR:
	    nbands = 3;
	    break;

        default:
	    nbands = 1; /* Force to a single band (with 24 bit pixels) */
    }

    for (band=0; band < nbands; band++)
    {
	if (allocLineMem) /* Alloc 24 bit row mem in 32 bit chunks */
	{
	    for (row=0; row<nlines; row++)
	    {
                IPW_FREE(pData[band][row], cAllocatorData);
	    }
        }
	IPW_FREE(pData[band], cAllocatorData);
    }
    IPW_FREE(pData, cAllocatorData);

    NORMAL_RETURN;

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write callback for outputting JPEG compressed data.
 *
 * LongDescription:::
 *  Write callback for use in w_jCompress. The userData
 *  structure element must have been loaded with the file
 *  pointer to the file to hold the compresssed image.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  userData::       Ptr to FILE ( save compressed data )
 *  buf::            Ptr to byte buffer to write out.
 *  nbytes::         Number of bytes to write.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
  *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jcWriteCallback (
    IPWState		 *pIPWState,
    void                 *userData,
    UInt8                *buf,
    UInt32                nbytes,
    UInt32 		  *ncount)
{
/* keep core check and compiler happy */
    PROLOGUE;
    pIPWState = pIPWState;
    __error__ = __error__;
 
    *ncount = (UInt32)fwrite((UInt8 *)buf, (size_t)1, (size_t)nbytes, (FILE *)userData); 
/*    ERR_GEN( (*ncount != (UInt32) nbytes), cErrIO_WriteError); */
    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Compress gray or RGB PIXR to a JFIF formatted file.
 *
 * LongDescription:::
 *    Compress a gray or RGB image in three PIXRs to a JFIF formatted file.
 *    The source PIXRs must be either 4 or 8 bpp. They must all be
 *    of the same bit depth. 4 bpp PIXRs are scaled up to 8 bpp
 *    before compression.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPixr::        Source 4 or 8 bit PIXR to be compressed (gray or RGB)
 *  fdJFIF::         FILE ptr for the compressed file
 *  quality::        Quality factor (0 ... 100). Higher is better quality.
 *  jcOptions::      Options for progressive and qtable defined 
 *                   in iw_jpeg.pub. JC_DEFAULT is no progressive and 
 *                   use default Qtable.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jCompress (
    IPWState           *pIPWState,
    PIXR               *srcPixr,
    FILE               *fdJFIF,
    Int32               quality,
    Int32               jcOptions)
{
    PROLOGUE;
    CompressInfo       cinfoStruct;
    CompressInfo *     cinfo = &cinfoStruct;
    PIXR *             pixrBand;
    UInt8 ***          userBuf = NULL;
    UInt8 *            image[3];
    Int32              w, h, d, bpl, xDpi, yDpi;
    UInt32             row, chan, line, jMemOrg;
    Int32              ncomps, nlines;
    ScaledInt1616      xRes, yRes;
    
/* Verify arguments */
    ERR_GEN(fdJFIF == NULL, cErrNullPtr);
    ERR_GEN_PIXR(srcPixr, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels1 | cChannels3, cFrame32);

/* clear out the cinfo struct */
   memset ((void *)cinfo, 0, sizeof (CompressInfo));
   ERR_PASS( w_pixrGetChannelCount(pIPWState, srcPixr, &ncomps));


/* 
  Load each of the pixr component addresses into the band array
  and get the address of the image data from each band
*/
    for (chan=0; chan<ncomps; chan++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, srcPixr, chan, &pixrBand));
        ERR_GEN(pixrBand == NULL, cErrInvalidParm);
	ERR_PASS( w_pixrGetImage(pIPWState, pixrBand, &(image[chan])));
    }

/* Determine the image size, start address and line pitch */
/* We'll assume here that all of the components are the same */
    ERR_PASS( w_pixrGetWidth(pIPWState, srcPixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, srcPixr, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, srcPixr, &bpl););
    ERR_PASS( w_pixrGetResolutionDPI (pIPWState, srcPixr, &xRes, &yRes));
    xDpi = MAX(ScaledInt1616_TO_Int32 (xRes), 0); 
    yDpi = MAX(ScaledInt1616_TO_Int32 (yRes), 0);

/* Verify that this is an 8 or 4 bit image */
    ERR_PASS( w_pixrGetDepth(pIPWState, srcPixr, &d));
    ERR_GEN(! (d==8 || d==4), cErrDepthNotSupported);

/* Set the proper memory organization indicator */
    if (ncomps > 1)
        jMemOrg = JMEMORG_BSQ_PIXR;
    else if (d == 8)
        jMemOrg = JMEMORG_GRAY8_PIXR;
     else
	 jMemOrg = JMEMORG_GRAY4_PIXR;

/* Perform the compression initialization steps */
    ERR_PASS_CLEANUP( w_jcInit(pIPWState, cinfo, 
		       w, h, ncomps,
		       quality,
		       jcOptions,
		       xDpi, yDpi,
		       wp_jcWriteCallback,
		       (void *) fdJFIF, NULL,
		       jMemOrg));

/* Allocate a pointer array (but no memory) to point to user's data */
    ERR_PASS(w_jcAllocMCURowBuffer(pIPWState, cinfo, cFalse, &userBuf));
    ERR_GEN(userBuf == NULL, cErrNomem);

/* Compress the image an MCU row at a time */
    ERR_PASS_CLEANUP( w_jcGetImageLinesPerMCU(pIPWState, cinfo, &nlines) );
    row =0;
    while ( row<h )
    {
        for (chan=0; chan<ncomps; chan++)
        {
	    for(line=0; line<nlines; ++line)
	    {
		userBuf[chan][line] = image[chan];
		image[chan] += bpl;
	    }
        }
        /* Now compress the MCU row */
	ERR_PASS_CLEANUP( w_jcMCURow(pIPWState, cinfo, userBuf));
	row+=nlines;
	if( (h - row) <nlines) nlines = h- row;
    }

    if(userBuf != NULL)
    {
	ERR_PASS_CLEANUP( w_jcMCURowFree(pIPWState, cinfo, cFalse, userBuf));
	userBuf = NULL;
    }

/* Finish up and close things down */
    ERR_PASS( w_jcTerm(pIPWState, cinfo));

    NORMAL_RETURN;

ERR_CLEANUP:
    if(userBuf != NULL)
    {
	w_jcMCURowFree(pIPWState, cinfo, cFalse, userBuf);
	userBuf = NULL;
    }
    if(cinfo->jpegBuf != NULL)
    {
	w_jcMCURowFree(pIPWState, cinfo,cinfo->jpegBufLineMem, cinfo->jpegBuf);
	cinfo->jpegBuf = NULL;
    }
    jpeg_destroy_compress(pIPWState, &cinfo->j_cinfo);

    w_jcFreeErrorStr(pIPWState, cinfo);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Compress gray or RGB PIXR to a JFIF formatted file.
 *
 * LongDescription:::
 *    Compress a gray or RGB image in three PIXRs to a JFIF formatted file.
 *    The source PIXRs must be either 4 or 8 bpp. They must all be
 *    of the same bit depth. 4 bpp PIXRs are scaled up to 8 bpp
 *    before compression. Write JFIF format compressed image through
 *    callback function
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPixr::        Source 4 or 8 bit PIXR to be compressed (gray or RGB)
 *  WriteCallback::  Callback function to write out compressed data
 *  userData::       Hook for passing in user data 
 *  quality::        Quality factor (0 ... 100). Higher is better quality.
 *  jcOptions::      Options for progressive and qtable defined 
 *                   in iw_jpeg.pub. JC_DEFAULT is no progressive and 
 *                   use default Qtable.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jCompress_callback (
    IPWState           *pIPWState,
    PIXR               *srcPixr,
    ErrCode	(CDECL *WriteCallback) (IPWState *pIPWState,
					void *userData, UInt8 *buf, 
					UInt32 nbytes, UInt32 *nCount),
    void                *userData,
    Int32               quality,
    Int32               jcOptions)
{
    PROLOGUE;
    CompressInfo       cinfoStruct;
    CompressInfo *     cinfo = &cinfoStruct;
    PIXR *             pixrBand;
    UInt8 ***          userBuf = NULL;
    UInt8 *            image[3];
    Int32              w, h, d, bpl, xDpi, yDpi;
    UInt32             row, chan, line, jMemOrg;
    Int32              ncomps, nlines;
    ScaledInt1616      xRes, yRes;
    
/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN_PIXR(srcPixr, cByteOrderBig, cDepth4 | cDepth8,
		 cChannels1 | cChannels3, cFrame32);

/* clear out the cinfo struct */
   memset ((void *)cinfo, 0, sizeof (CompressInfo));
   ERR_PASS( w_pixrGetChannelCount(pIPWState, srcPixr, &ncomps));


/* 
  Load each of the pixr component addresses into the band array
  and get the address of the image data from each band
*/
    for (chan=0; chan<ncomps; chan++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, srcPixr, chan, &pixrBand));
        ERR_GEN(pixrBand == NULL, cErrInvalidParm);
	ERR_PASS( w_pixrGetImage(pIPWState, pixrBand, &(image[chan])));
    }

/* Determine the image size, start address and line pitch */
/* We'll assume here that all of the components are the same */
    ERR_PASS( w_pixrGetWidth(pIPWState, srcPixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, srcPixr, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, srcPixr, &bpl););
    ERR_PASS( w_pixrGetResolutionDPI (pIPWState, srcPixr, &xRes, &yRes));
    xDpi = MAX(ScaledInt1616_TO_Int32 (xRes), 0); 
    yDpi = MAX(ScaledInt1616_TO_Int32 (yRes), 0);

/* Verify that this is an 8 or 4 bit image */
    ERR_PASS( w_pixrGetDepth(pIPWState, srcPixr, &d));
    ERR_GEN(! (d==8 || d==4), cErrDepthNotSupported);

/* Set the proper memory organization indicator */
    if (ncomps > 1)
        jMemOrg = JMEMORG_BSQ_PIXR;
    else if (d == 8)
        jMemOrg = JMEMORG_GRAY8_PIXR;
     else
	 jMemOrg = JMEMORG_GRAY4_PIXR;

/* Perform the compression initialization steps */
    ERR_PASS_CLEANUP( w_jcInit(pIPWState, cinfo, 
		       w, h, ncomps,
		       quality,
		        jcOptions,
			xDpi, yDpi,
			WriteCallback,
			(void *) userData, 
		       NULL, jMemOrg));

/* Allocate a pointer array (but no memory) to point to user's data */
    ERR_PASS( w_jcAllocMCURowBuffer(pIPWState, cinfo, cFalse, &userBuf));
    ERR_GEN(userBuf == NULL, cErrNomem);

/* Compress the image an MCU row at a time */
    ERR_PASS_CLEANUP( w_jcGetImageLinesPerMCU(pIPWState, cinfo, &nlines) );
    row =0;
    while ( row<h )
    {
        for (chan=0; chan<ncomps; chan++)
        {
	    for(line=0; line<nlines; ++line)
	    {
		userBuf[chan][line] = image[chan];
		image[chan] += bpl;
	    }
        }
        /* Now compress the MCU row */
	ERR_PASS_CLEANUP( w_jcMCURow(pIPWState, cinfo, userBuf));
	row+=nlines;
	if( (h - row) <nlines) nlines = h- row;
    }

    if(userBuf != NULL)
    {
	ERR_PASS_CLEANUP( w_jcMCURowFree(pIPWState, cinfo, cFalse, userBuf));
	userBuf = NULL;
    }

/* Finish up and close things down */
    ERR_PASS( w_jcTerm(pIPWState, cinfo));

    NORMAL_RETURN;

ERR_CLEANUP:
    if(userBuf != NULL)
    {
	w_jcMCURowFree(pIPWState, cinfo, cFalse, userBuf);
	userBuf = NULL;
    }
    if(cinfo->jpegBuf != NULL)
    {
	w_jcMCURowFree(pIPWState, cinfo,cinfo->jpegBufLineMem, cinfo->jpegBuf);
	cinfo->jpegBuf = NULL;
    }
    jpeg_destroy_compress(pIPWState, &cinfo->j_cinfo);

    w_jcFreeErrorStr(pIPWState, cinfo);
    ERR_RETURN;
}

#undef _IW_CJPEG_C_

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*
$Log: iw_cjpeg.c,v $
Revision 1.12  2002/04/16 23:01:15  dabercro
Allow 'unknown' colorspace (LAB).

Revision 1.11  2001/11/14 00:10:38  dabercro
Use the mode for UNKNOWN colorspace which include chrominance
components.

Revision 1.10  2001/07/05 21:43:34  dabercro
Add an option to extract the tables separatetly for the  JPEGTables TIFF field.

Revision 1.9  2001/06/27 19:13:43  dabercro
Remove colorspace support in JPEG, Lab is supported by identifying the
colorspace as unknown.

Revision 1.8  2001/01/18 00:34:56  dabercro
New function for setting the gammut range.
New option to not subsample.

Revision 1.7  2000/08/01 21:56:33  dabercro
Copyright notice; WinCE add. Changes from Mark Tobin.

Revision 1.6  2000/06/09 23:15:15  jlin
fix read write callback functions return type for JPEG

Revision 1.4  2000/06/07 00:28:26  jlin
change wp_jcWriteCallback

Revision 1.3  2000/04/07 23:59:48  dabercro
Added support for CIELAB and ICCLAB color spaces.

Revision 1.2  2000/02/01 00:39:04  dabercro
Added ITULAB color space.

Revision 1.55  1998/07/23 20:49:17  xzhu
xzhu on Thu Jul 23 13:47:46 PDT 1998

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

Added set Qtable in w_jcInit. JPEG compression can use different Qtable according to the setting.

Revision 1.54  1998/07/13 23:58:16  xzhu
xzhu on Mon Jul 13 16:56:51 PDT 1998

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

Added new function w_jcSetQtable and put IMPACT Qtable in the function. SetQtable must be called after w_jcInit.

Revision 1.53  1998/01/15 01:49:33  ddavies
iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

CommonCheck: Fix WATCOM warnings about unused procedure args.

Revision 1.52  1998/01/14 00:13:29  ddavies
iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.51  1997/11/20 00:07:49  xzhu
xzhu on Wed Nov 19 15:56:26 PST 1997

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

Added two new inputs xDpi and yDpi in w_jcInit, the resolution info will be saved in the header of compressed file. Deleted w_setDPI, it will not work since w_jcInit has handle all of init. procedure.

Revision 1.50  1997/07/11 01:58:57  ddavies
iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.49  1997/06/10 21:17:57  xzhu
xzhu on Tue Jun 10 14:16:45 PDT 1997

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

Added progressiv selection in w_jcInit, w_jCompress and w_jCompress_callback to do pregressive jpeg compression.

Revision 1.48  1996/10/29 02:19:40  xzhu
xzhu on Mon Oct 28 18:19:40 PST 1996

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

CommonCheck: Added new functions to retrive error string and free error message memory.

Revision 1.47  1996/09/23 16:25:43  xzhu
xzhu on Mon Sep 23 09:24:48 PDT 1996

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

Fixed core check warning.

Revision 1.46  1996/09/20 18:44:07  xzhu
xzhu on Fri Sep 20 11:42:21 PDT 1996

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

Added new function w_jCompress_callback to write JFIF by user's callback function.

Revision 1.45  1996/04/23 17:49:10  xzhu
xzhu on Tue Apr 23 10:49:09 PDT 1996

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

CommonCheck: Fixed PC compiler warning.

Revision 1.44  1996/04/23 00:20:50  xzhu
xzhu on Mon Apr 22 17:20:50 PDT 1996

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

CommonCheck: Fixed core check warning by removing comment about input and changing input parameter position and so on.

Revision 1.43  1996/04/20 01:47:56  xzhu
xzhu on Fri Apr 19 18:47:05 PDT 1996

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

Modified functions for new jpeg and added some new functions.

Revision 1.42  1996/02/08 01:21:29  ejaquith
ejaquith on Wed Feb 7 17:21:09 PST 1996

iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

Fixed initialization bug that was killing a routine.

Revision 1.41  1996/01/04 20:24:35  ejaquith
iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.40  1995/12/13  04:10:03  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.39  1995/11/13  20:13:26  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.38  1995/10/18  04:16:29  ddavies
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Add extra pointers to i_getDIBPixrLineProc call to reflect the four
 * types of procedures it could return.
 *
 * Revision 1.37  1995/10/13  22:02:41  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.36  1995/10/12  04:25:29  ddavies
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.35  1995/10/07  01:21:56  ddavies
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Include stdio.h since ipwcore.prv no longer includes it (will be true soon).
 *
 * Revision 1.34  1995/10/02  00:18:44  ejaquith
 * ejaquith on Sun Oct 1 17:18:24 PDT 1995
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Filled in standardized routine header comment blocks.
 *
 * Revision 1.33  1995/09/22  00:38:29  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Fixed up wp_ issues.
 *
 * Revision 1.32  1995/09/21  22:54:18  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.31  1995/09/20  01:19:22  dferrell
 * dferrell on Tue Sep 19 18:18:38 PDT 1995
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Fixed uninitialized memory access (struct on stack) from purify test.
 *
 * Revision 1.30  1995/09/18  22:48:23  dferrell
 * dferrell on Mon Sep 18 15:46:51 PDT 1995
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * fixed logic problem in wp_jBuild4To8Lut () which did not allow for
 * previously allocated tables to be used.  Now it creates one if it needs
 * to and uses the one passed in otherwise.  In both cases it fills in the
 * table with new data.
 *
 * Revision 1.29  1995/09/12  16:54:27  ejaquith
 * ejaquith on Tue Sep 12 09:54:26 PDT 1995
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.28  1995/09/08  00:14:24  ejaquith
 * ejaquith on Thu Sep 7 17:14:23 PDT 1995
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.27  1995/09/07  01:33:02  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.26  1995/09/06  20:21:05  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.25  1995/09/01  02:20:17  ejaquith
 * ejaquith on Thu Aug 31 19:20:16 PDT 1995
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.24  1995/08/31  21:14:54  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.23  1995/08/24  02:52:32  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Make writeCallback routine return number of bytes written.
 * Added NULL error checking to buffer allocations.
 * Added missing pIPWState argument when calling routine pointed to by
 * cinfo's field cvtImageRow.
 *
 * Revision 1.22  1995/08/22  02:21:19  ejaquith
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.21  1995/08/17  22:00:40  ejaquith
 * ejaquith on Thu Aug 17 15:00:32 PDT 1995
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.20  1994/10/25  20:58:23  lperry
 * lperry on Tue Oct 25 13:55:12 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Add public function iw_jcSetDPI which provides a way to
 * specify the resolution of a compressed image. This function
 * must be called BEFORE iw_jcInit() is called.
 *
 * Also added complete parameter specs for all callback
 * procedures.
 *
 * Revision 1.19  1994/10/11  01:04:53  lperry
 * lperry on Mon Oct 10 18:03:48 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Added include of shrpixr.pub to get prototypes for
 * i_getDIBPixrLineProc(). This is only needed on PC.
 *
 * Revision 1.18  1994/10/03  23:29:18  lperry
 * lperry on Mon Oct 3 16:28:31 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Change to use of iw_jpeg.pub header instead of jpegpro.h
 *
 * Revision 1.17  1994/09/08  00:28:19  lperry
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Remove Date from RCS macro. Test for redundant types.pub includes.
 *
 * Revision 1.16  1994/08/01  17:33:24  lperry
 * lperry on Mon Aug 1 10:31:56 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Fix pixr to memory order copy routine to correctly deal with
 * line lengths not a multiple of 4 pixels.
 *
 * Revision 1.15  1994/07/21  18:41:54  lperry
 * lperry on Thu Jul 21 11:40:30 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Add return code debugging abortI() function to trace errors.
 *
 * Revision 1.14  1994/07/21  00:19:23  lperry
 * lperry on Wed Jul 20 17:19:20 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Change error handling from old setjmp/longjmp to function return codes.
 *
 * Revision 1.13  1994/07/15  23:57:08  lperry
 * lperry on Fri Jul 15 16:55:59 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Added proper argument casts for fwrite in the
 * callback function to write JPEG out JPEG data.
 *
 * Revision 1.12  1994/06/27  17:39:33  lperry
 * lperry on Mon Jun 27 10:35:40 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Changed the function which allocated 4 to 8 bit depth
 * conversion table, since it now needs to use the cinfo
 * structure in calling the memory allocator.
 *
 * Revision 1.11  1994/06/01  01:10:01  lperry
 * lperry on Tue May 31 18:07:16 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Fixed bug in cases where row pointers are copied rather
 * than actual data. jcCopyRow pointers now always copies
 * to jpegBuf, so interface is same for all functions
 * referred to by function pointer cvtimageRow.
 *
 * Revision 1.10  1994/05/26  22:23:26  lperry
 * lperry on Thu May 26 15:19:28 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Change for Easyscan/Amber. Treat band-interleaved RGB data
 * line by line instead of in a packed buffer. Essentially its
 * like a single band 24 bit image.
 * Also expose all the required MCU row processing interfaces
 * (some had been static in the PIXR-only implementation).
 *
 * Revision 1.9  1994/05/24  22:08:36  lperry
 * lperry on Tue May 24 15:06:17 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Changed interface to compression init function (iw_jcInit) to
 * be consistent with decompression init. Caller now passes
 * cinfo structure by reference.
 *
 * Revision 1.8  1994/05/24  00:50:19  lperry
 * lperry on Mon May 23 17:47:40 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Wrappers for new JPEG interface. Supports multiple user
 * data organizations (interleaved, pixr, natural memory order
 * and 8 or 4 bpp). Ipshared now has deals with only memory
 * order images. Ipwrap provides the conversion layer.
 *
 * Revision 1.7  1994/04/13  01:29:21  lperry
 * lperry on Tue Apr 12 18:27:57 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Add modifications to use new color pixr type.
 * There is now only one version of pixr compression
 * to handle gray, color, 4 bit and 8 bit pixrs.
 *
 * Revision 1.6  1994/04/08  00:44:00  lperry
 * lperry on Thu Apr 7 17:37:15 PDT 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Added functions for:
 *   building 4 to 8 bit luts
 *   copying 4 bit pixrs to 8 bit pixrs a line at a time
 * Removed several of the wrappers for the ipshared functions
 * which didn't add any functionality or value. Just call
 * the ipshared function directly.
 *
 * Added code to convert between pixrs and buffers, dealing
 * with buffer pointers rather than intermediate buffering.
 *
 * Revision 1.5  1994/03/29  01:38:52  ddavies
 * ddavies on Mon Mar 28 17:37:41 PST 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Added "const" to return value for jerrorString.  This is required since
 * jerrorString returns a "const char *" and acc warns about tossing the "const." when jerrorString returns whatever jerrorString returned.
 *
 * Revision 1.4  1994/03/24  20:52:31  lperry
 * lperry on Thu Mar 24 12:49:28 PST 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Changed interface to iw_jcInit to allow a procedure
 * to be specified for setting non-standard options
 * The initial usage is for RGB to gray decompression,
 * but it will be the general interface for any special changes
 * Also substituted UCHAR_ACCESS memory r/w for forthe
 * previous byte swapping of the whole pixr.
 *
 * Revision 1.3  1994/03/05  03:04:07  lperry
 * lperry on Fri Mar 4 19:02:08 PST 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Added routines iw_jCompressGrayPixr and iw_jDecompressPixr
 * which take either a single 8 bit gray pixr or three 8 bpp
 * R, G and B pixrs and produce a compressed JPEG file
 * in JFIF format
 *
 * Revision 1.2  1994/03/02  02:07:51  lperry
 * lperry on Tue Mar 1 18:05:45 PST 1994
 *
 * iw_cjpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_cjpeg.c,v
 *
 * Added wrappers for functions to query the number of lines
 * in an MCU (minimum compressible unit) and to retrieve the
 * userData pointer, used as a hook to pass FILE stream ptrs
 * or buffer addresses
 *
 * Revision 1.1  1994/02/27  02:33:43  lperry
 * Initial revision
 *
*/
