/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	iw_dhvq.c:	Hierarchical Vector Quantization compression wrappers:
 *
 *	Code to compress PIXRs to HVQ compressed PIXRs:
 *		w_HVQCompress()
 *
 */

#define _IW_DHVQ_C_

#include <stdio.h>
#include <math.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imageref.pub"
#include "defines.pub"
#include "memory.pub"

/* prototypes */
#include "shrpixr.pub"
#include "props.pub"
#include "hvq.pub"

/* extern HVQ LUTs */
#include "hvq.h"

IP_RCSINFO(iw_dhvq_c_RCSInfo, "$RCSfile: iw_dhvq.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a single channel of 8-bit HVQ indices.
 *
 * LongDescription:::
 *  Decompress a single channel of 8-bit HVQ indices to a preallocated
 *  output PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPIXR::        Source 8-bit single channel PIXR to be compressed
 *  hvqLevels::      Number of the four HVQ levels to use. Note: the
 *                   compression achieved per level is: 2:1, 4:1, 8:1, & 16:1
 *  expansions::      1 - normal expansions, 2 - expansions by two in both
 *                   directions (halftoning option for hvqLevels = 4 only!)
 *
 * Output:::
 *  dstPIXR::        Preallocated destination 8-bit PIXR after decompression.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  HVQ
 *
 * Keywords:::
 *  HVQ, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_hvqd_channel(
	     IPWState	*pIPWState,
	     PIXR	*srcPIXR,
	     Int32	hvqLevels,
	     Int32	expansions,
	     PIXR	*dstPIXR
	     )
{
    register Int32 vecrow;	/* Rows left in vector */
    register UInt8 *cbvec;	/* Pointer to codebook vector */
    register Int32 row;		/* Rows left to decode */
    register UInt32	index;	/* Codebook index resulting from HVQ */
    Int32	src_bpl;	/* Source bytes/line */
    Int32	src_width;	/* Source width */
    Int32	src_height;	/* Source height */
    Int32	src_depth;	/* Source depth */
    UInt8	*src_row;	/* Source image swath pointer */
    Int32	dst_bpl;	/* Destination bytes/line */
    Int32	dst_width;	/* Destination width */
    Int32	dst_height;	/* Destination height */
    Int32	dst_depth;	/* Destination depth */
    UInt32	vecs_across;	/* How many vectors along the horizontal */
    UInt32	vec_num;	/* Which vector along the horizontal */
    UInt32	rows_down;	/* How many rows down the image */
    UInt8	*vec;		/* Pointer to source vector */
    UInt8	*dst_swath;	/* Pointer to destination row */
    Int32	xexpansions;	/* Expansions factor in X direction */
    Int32	yexpansions;	/* Expansions factor in Y direction */
    Int32	vecsize;	/* Vector size */
    UInt32	big_index;	/* Buffered index */
    register Int32	indices;	/* Number of indices left in buffer */
    Int32	dst_lpl;	/* Destination longs per line */
    UInt32	*src_longptr;	/* Source long pointer */
    UInt32	*dst_longptr;	/* Destination long pointer */
    UInt32	*cb_longptr;	/* Codebook long pointer */
    Int32	dst_spl;	/* Destination shorts per line */
    UInt16	*dst_shortptr;	/* Destination short pointer */
    UInt16	*cb_shortptr;	/* Codebook short pointer */

    PROLOGUE;
    ERR_PASS(w_pixrGetBpl(pIPWState, srcPIXR, &src_bpl));
    ERR_PASS(w_pixrGetWidth(pIPWState, srcPIXR, &src_width));
    ERR_PASS(w_pixrGetHeight(pIPWState, srcPIXR, &src_height));
    ERR_PASS(w_pixrGetDepth(pIPWState, srcPIXR, &src_depth));
    ERR_PASS(w_pixrGetImage(pIPWState, srcPIXR, &src_row));

    ERR_PASS(w_pixrGetBpl(pIPWState, dstPIXR, &dst_bpl));
    ERR_PASS(w_pixrGetWidth(pIPWState, dstPIXR, &dst_width));
    ERR_PASS(w_pixrGetHeight(pIPWState, dstPIXR, &dst_height));
    ERR_PASS(w_pixrGetDepth(pIPWState, dstPIXR, &dst_depth));
    ERR_PASS(w_pixrGetImage(pIPWState, dstPIXR, &dst_swath));

    /* Compute the output size based on the number of HVQ levels */
    switch (hvqLevels) {
    case 1:
	xexpansions = 2, yexpansions = 1;
	break;
    case 2:
	xexpansions = 2, yexpansions = 2;
	break;
    case 3:
	xexpansions = 4, yexpansions = 2;
	break;
    case 4:
	if (expansions == 2) {
	    xexpansions = 1, yexpansions = 8;
	} else {
	    xexpansions = 4, yexpansions = 4;
	}
	break;
    default:
	ERR_RETURN;
    }
    vecs_across = src_width;
    rows_down = src_height;
    vecsize = xexpansions * yexpansions;

    /* Destination size checking */
    if (expansions == 2) {
	ERR_PASS( dst_depth != 1);
    } else {
	ERR_PASS( dst_depth != 8);
    }
    ERR_PASS( dst_width < src_width * xexpansions);
    ERR_PASS( dst_height < src_height * yexpansions);

    /* Loop according to compression ratio (number of HVQ levels) */
    switch (hvqLevels) {
    case 1:

	/* Loop through every row of vectors */
	for (row = 0; row < rows_down; row++, src_row += src_bpl,
	     dst_swath += dst_bpl) {

	    /* Loop through the vectors within the swath */
	    src_longptr = (UInt32 *)src_row;
	    big_index = *src_longptr++;
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    dst_shortptr = (UInt16 *)dst_swath;
#else
	    dst_shortptr = (UInt16 *)dst_swath + 1;
#endif
	    indices = 0;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Extract the index */
		index = (big_index & 0xff000000) >> 24;
		big_index <<= 8;
		if (++indices == 4) {
		    big_index = *src_longptr++;
		    indices = 0;
		}

		/* Decode the individual vector
		 */
		cb_shortptr = hvqdlut_2x1_1 + index;
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		*dst_shortptr++ = *cb_shortptr++;
#else
		*dst_shortptr = *cb_shortptr++;
		if (vec_num & 1) {
		    dst_shortptr += 3;
		} else {
		    dst_shortptr--;
		}
#endif
	    } /* For every vector in the row */
	} /* For every row of vectors in the image */
	break;

    case 2:

	dst_spl = dst_bpl >> 1;

	/* Loop through every row of vectors */
	for (row = 0; row < rows_down; row++, src_row += src_bpl,
	     dst_swath += dst_bpl * 2) {

	    /* Loop through the vectors within the swath */
	    src_longptr = (UInt32 *)src_row;
	    big_index = *src_longptr++;
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    vec = dst_swath;
#else
	    vec = dst_swath + 2;
#endif
	    indices = 0;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Extract the index */
		index = (big_index & 0xff000000) >> 24;
		big_index <<= 8;
		if (++indices == 4) {
		    big_index = *src_longptr++;
		    indices = 0;
		}

		/* Decode the individual vector
		 */
		cb_shortptr = hvqdlut_2x2_1 + index * 2;
		dst_shortptr = (UInt16 *)vec;

		vecrow = 2;
		while (vecrow--) {
		    *dst_shortptr = *cb_shortptr++;
		    dst_shortptr += dst_spl;
		}
	    	    
		/* Go to the next vector */
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
		vec += 2;
#else
		if (vec_num & 1) {
		    vec += 6;
		} else {
		    vec -= 2;
		}
#endif
	    } /* For every vector in the row */
	} /* For every row of vectors in the image */
	break;

    case 3:
	dst_lpl = dst_bpl >> 2;

	/* Loop through every row of vectors */
	for (row = 0; row < rows_down; row++, src_row += src_bpl,
	     dst_swath += dst_bpl * yexpansions) {

	    /* Loop through the vectors within the swath */
	    src_longptr = (UInt32 *)src_row;
	    big_index = *src_longptr++;
	    vec = dst_swath;
	    indices = 0;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Extract the index */
		index = (big_index & 0xff000000) >> 24;
		big_index <<= 8;
		if (++indices == 4) {
		    big_index = *src_longptr++;
		    indices = 0;
		}

		/* Decode the individual vector
		 */
		cb_longptr = hvqdlut_4x2_1 + index * 2;
		dst_longptr = (UInt32 *)vec;

		vecrow = 2;
		while (vecrow--) {
		    *dst_longptr = *cb_longptr++;
		    dst_longptr += dst_lpl;
		}
	    	    
		/* Go to the next vector */
		vec += 4;
	    } /* For every vector in the row */
	} /* For every row of vectors in the image */
	break;

    case 4:
	if (expansions == 2) {
	    UInt32 row1 = 0, row2 = 0, row3 = 0, row4= 0;
	    UInt32 row5 = 0, row6 = 0, row7 = 0, row8 = 0;
	    UInt32 outbyte = 0;
	    {
		UInt8 *bitptr = hvqdlut_4x4_halftone_1;

		for (vec_num = 0; vec_num < 256; vec_num++) {
		    for (row = 0; row < 8; row++) {
			*bitptr = ~*bitptr;
			bitptr++;
		    }
		}
	    }

	    dst_lpl = dst_bpl >> 2;
	    ERR_PASS(dst_bpl & 3);

	    /* Loop through every row of vectors */
	    for (row = 0; row < rows_down; row++, src_row += src_bpl,
		 dst_swath += dst_bpl * 8) {

		/* Loop through the vectors within the swath */
		src_longptr = (UInt32 *)src_row;
		big_index = *src_longptr++;
		vec = dst_swath;
		indices = 0;
		for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		    /* Extract the index */
		    index = (big_index & 0xff000000) >> 24;
		    big_index <<= 8;
		    if (++indices == 4) {
			big_index = *src_longptr++;
			indices = 0;
		    }

		    /* Decode the individual vector
		     */
		    cbvec = hvqdlut_4x4_halftone_1 + index * 8;
		    row1 = (row1 << 8) | *cbvec++;
		    row2 = (row2 << 8) | *cbvec++;
		    row3 = (row3 << 8) | *cbvec++;
		    row4 = (row4 << 8) | *cbvec++;
		    row5 = (row5 << 8) | *cbvec++;
		    row6 = (row6 << 8) | *cbvec++;
		    row7 = (row7 << 8) | *cbvec++;
		    row8 = (row8 << 8) | *cbvec;

		    if (++outbyte == 4) {
			dst_longptr = (UInt32 *)vec;
			*dst_longptr = row1, dst_longptr += dst_lpl;
			*dst_longptr = row2, dst_longptr += dst_lpl;
			*dst_longptr = row3, dst_longptr += dst_lpl;
			*dst_longptr = row4, dst_longptr += dst_lpl;
			*dst_longptr = row5, dst_longptr += dst_lpl;
			*dst_longptr = row6, dst_longptr += dst_lpl;
			*dst_longptr = row7, dst_longptr += dst_lpl;
			*dst_longptr = row8;
			outbyte = 0;
			vec += 4;
		    }
		} /* For every vector in the row */

		/* If we are in the middle of a 'long', flush it */
		if (outbyte != 0) {
		    Int32 shift = (4 - outbyte) * 8;

		    dst_longptr = (UInt32 *)vec;
		    row1 <<= shift;
		    row2 <<= shift;
		    row3 <<= shift;
		    row4 <<= shift;
		    row5 <<= shift;
		    row6 <<= shift;
		    row7 <<= shift;
		    row8 <<= shift;
		    *dst_longptr = row1, dst_longptr += dst_lpl;
		    *dst_longptr = row2, dst_longptr += dst_lpl;
		    *dst_longptr = row3, dst_longptr += dst_lpl;
		    *dst_longptr = row4, dst_longptr += dst_lpl;
		    *dst_longptr = row5, dst_longptr += dst_lpl;
		    *dst_longptr = row6, dst_longptr += dst_lpl;
		    *dst_longptr = row7, dst_longptr += dst_lpl;
		    *dst_longptr = row8;
		    outbyte = 0;
		}
	    } /* For every row of vectors in the image */

	} else {

	    dst_lpl = dst_bpl >> 2;

	    /* Loop through every row of vectors */
	    for (row = 0; row < rows_down; row++, src_row += src_bpl,
		 dst_swath += dst_bpl * yexpansions) {

		/* Loop through the vectors within the swath */
		src_longptr = (UInt32 *)src_row;
		big_index = *src_longptr++;
		vec = dst_swath;
		indices = 0;
		for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		    /* Extract the index */
		    index = (big_index & 0xff000000) >> 24;
		    big_index <<= 8;
		    if (++indices == 4) {
			big_index = *src_longptr++;
			indices = 0;
		    }

		    /* Decode the individual vector
		     */
		    cb_longptr = hvqdlut_4x4_1 + index * 4;
		    dst_longptr = (UInt32 *)vec;

		    vecrow = 4;
		    while (vecrow--) {
			*dst_longptr = *cb_longptr++;
			dst_longptr += dst_lpl;
		    }
	    	    
		    /* Go to the next vector */
		    vec += 4;
		} /* For every vector in the row */
	    } /* For every row of vectors in the image */
	}
	break;
    default:
	ERR_RETURN;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress an image containing one or more channels of 8-bit HVQ indices.
 *
 * LongDescription:::
 *  Decompress an image containing one or more channels of 8-bit HVQ indices
 *  to a preallocated output PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPIXR::        Source 8-bit PIXR to be compressed (gray or RGB)
 *  hvqLevels::      Number of the four HVQ levels to use. Note: the
 *                   compression achieved per level is: 2:1, 4:1, 8:1, & 16:1
 *  expansions::      1 - normal expansions, 2 - expansions by two in both
 *                   directions (halftoning option for hvqLevels = 4 only!)
 *
 * Output:::
 *  dstPIXR::        Preallocated destination 8-bit PIXR after decompression.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  HVQ
 *
 * Keywords:::
 *  HVQ, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_hvqd_image(
	   IPWState	*pIPWState,
	   PIXR		*srcPIXR,
	   Int32	hvqLevels,
	   Int32	expansions,
	   PIXR		*dstPIXR
	  )
{
    Int32	channel, channels;
    Int32	width;
    Int32	height;
    Int32	depth;
    PIXR	*src_ch_pixr;
    PIXR	*dst_ch_pixr;

    PROLOGUE;
    ERR_PASS(w_pixrGetChannelCount(pIPWState, dstPIXR, &channels));
    if (expansions == 1) {
	Int32 dst_ch_pixrannels;
	ERR_PASS(w_pixrGetChannelCount(pIPWState, dstPIXR, &dst_ch_pixrannels));

	ERR_PASS(dst_ch_pixrannels != channels);
    }

    for (channel = 0; channel < channels; channel++) {

	/* Get the PIXR for that channel */
	ERR_PASS(w_pixrGetChannel(pIPWState, srcPIXR, channel, &src_ch_pixr));
	ERR_PASS(w_pixrGetChannel(pIPWState, dstPIXR, channel, &dst_ch_pixr));

	/* Get the size of the image */
	ERR_PASS(w_pixrGetWidth(pIPWState, src_ch_pixr, &width));
	ERR_PASS(w_pixrGetHeight(pIPWState, src_ch_pixr, &height));
	ERR_PASS(w_pixrGetDepth(pIPWState, src_ch_pixr, &depth));

	/* de-HVQ that channel */
	ERR_PASS(wp_hvqd_channel(pIPWState, src_ch_pixr, hvqLevels, expansions,
				 dst_ch_pixr));
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  HVQ Decompress gray or RGB HVQ PIXR.
 *
 * LongDescription:::
 *  Decompress a gray image in one (or RGB images in three PIXRs) from an (a
 *  set of) output PIXR(s) containing 8-bit HVQ indices.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPIXR::        Source 8-bit PIXR to be compressed (gray or RGB)
 *  hvqLevels::      Number of the four HVQ levels to use. Note: the
 *                   compression achieved per level is: 2:1, 4:1, 8:1, & 16:1
 *
 * Output:::
 *  dstPIXRptr::     Pointer to destination 8-bit PIXR pointer after compression
 *                   (gray or RGB), which is generated and returned.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  HVQ
 *
 * Keywords:::
 *  HVQ, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_HVQDecompress (
		 IPWState	*pIPWState,
		 PIXR		*srcPIXR,
		 Int32		hvqLevels,
		 PIXR		**dstPIXRptr)
{
    PROLOGUE;
    Int32	src_width;	/* Source width */
    Int32	src_height;	/* Source height */
    Int32	src_depth;	/* Source depth */
    ColorType	src_type;	/* Source color type */
    ScaledInt1616  src_xdpi;	/* Source X resolution in dpi */
    ScaledInt1616  src_ydpi;	/* Source Y resolution in dpi */
    PIXR	*dstPIXR;	/* Destination PIXR */
    Int32	dst_width;	/* Destination width */
    Int32	dst_height;	/* Destination height */
    Int32	dst_depth;	/* Destination depth */
    ColorType	dst_type;	/* Destination color type */
    ScaledInt1616  dst_xdpi;	/* Destination X resolution in dpi */
    ScaledInt1616  dst_ydpi;	/* Destination Y resolution in dpi */
    Int32	channels;	/* Number of channels in PIXR */
    UInt32	size;		/* Size of input PIXR */
    UInt32	outsize;	/* Size of output PIXR */
    Int32	xexpansions;	/* Expansions factor in X direction */
    Int32	yexpansions;	/* Expansions factor in Y direction */

    /* Default to return NULL */
    *dstPIXRptr = NULL;

/* Verify arguments */
    ERR_GEN_PIXR(srcPIXR, cByteOrderBig, cDepth8,
		 cChannels1 | cChannels3, cFrame32);

    ERR_PASS(w_pixrGetChannelCount(pIPWState, srcPIXR, &channels));

/* Determine the image size, start address and line pitch */
/* We'll assume here that all of the components are the same */
    ERR_PASS(w_pixrGetWidth(pIPWState, srcPIXR, &src_width));
    ERR_PASS(w_pixrGetHeight(pIPWState, srcPIXR, &src_height));
    ERR_PASS(w_pixrGetDepth(pIPWState, srcPIXR, &src_depth));
    ERR_PASS(w_pixrGetType(pIPWState, srcPIXR, &src_type));
    ERR_PASS(w_pixrGetResolutionDPI(pIPWState, srcPIXR, &src_xdpi, &src_ydpi));

/* Verify that this is an 8-bit image */
    ERR_GEN(! (src_depth==8), cErrDepthNotSupported);

/* Calculate compressed image size */
    switch (hvqLevels) {
    case 1:
	xexpansions = 2, yexpansions = 1;
	break;
    case 2:
	xexpansions = 2, yexpansions = 2;
	break;
    case 3:
	xexpansions = 4, yexpansions = 2;
	break;
    case 4:
	xexpansions = 4, yexpansions = 4;
	break;
    default:
	ERR_RETURN;
    }
    dst_width = src_width * xexpansions;
    dst_height = src_height * yexpansions;
    dst_depth = 8;
    dst_type = src_type;
    dst_xdpi = ScaledInt1616_TO_Int32(src_xdpi) * xexpansions;
    dst_ydpi = ScaledInt1616_TO_Int32(src_ydpi) * yexpansions;
 
/* Create a PIXR for the output HVQ image */
    ERR_PASS_CLEANUP( w_createPixrExtendedNoInit(
					pIPWState, dst_width, dst_height,
					dst_depth,
					Int32_TO_ScaledInt1616(dst_xdpi),
					Int32_TO_ScaledInt1616(dst_ydpi),
					32, channels, dst_type, NULL,
					&dstPIXR));

/* Compress source PIXR into destination PIXR */
    ERR_PASS(wp_hvqd_image(pIPWState, srcPIXR, hvqLevels, 1, dstPIXR));

    size = src_width * src_height * channels;
    outsize = dst_width * dst_height * channels;

    *dstPIXRptr = dstPIXR;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (dstPIXR) {
	w_destroyPixr(pIPWState, &dstPIXR);
    }
	
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  HVQ Decompress gray or RGB HVQ PIXR into a binary using halftoning.
 *
 * LongDescription:::
 *  Decompress to binary a gray image in one (or RGB images in three PIXRs)
 *  from an (a set of) output PIXR(s) containing 8-bit HVQ indices.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPIXR::        Source 8-bit PIXR to be compressed (gray or RGB)
 *  hvqLevels::      Number of the four HVQ levels to use. Note: the
 *                   compression achieved per level is: 2:1, 4:1, 8:1, & 16:1
 *  expansions::     number of spatial expansions in both X and Y directions
 *                   (currently only expansions = 2 supported)
 *
 * Output:::
 *  dstPIXRptr::     Pointer to destination 8-bit PIXR pointer after compression
 *                   (gray or RGB), which is generated and returned.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  HVQ
 *
 * Keywords:::
 *  HVQ, dithering, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_HVQHalftoneDecompress (
			 IPWState	*pIPWState,
			 PIXR		*srcPIXR,
			 Int32		hvqLevels,
			 Int32		expansions,
			 PIXR		**dstPIXRptr)
{
    PROLOGUE;
    Int32	src_channels;	/* Number of channels in source */
    Int32	src_width;	/* Source width */
    Int32	src_height;	/* Source height */
    Int32	src_depth;	/* Source depth */
    ColorType	src_type;	/* Source color type */
    ScaledInt1616  src_xdpi;	/* Source X resolution in dpi */
    ScaledInt1616  src_ydpi;	/* Source Y resolution in dpi */
    PIXR	*dstPIXR;	/* Destination PIXR */
    Int32	dst_channels;	/* Number of channels in destination */
    Int32	dst_width;	/* Destination width */
    Int32	dst_height;	/* Destination height */
    Int32	dst_depth;	/* Destination depth */
    ColorType	dst_type;	/* Destination color type */
    ScaledInt1616  dst_xdpi;	/* Destination X resolution in dpi */
    ScaledInt1616  dst_ydpi;	/* Destination Y resolution in dpi */
    UInt32	size;		/* Size of input PIXR */
    UInt32	outsize;	/* Size of output PIXR */
    Int32	xexpansions;	/* Expansions factor in X direction */
    Int32	yexpansions;	/* Expansions factor in Y direction */

    /* Default to return NULL */
    *dstPIXRptr = NULL;

    /* Verify arguments */
    ERR_GEN_PIXR(srcPIXR, cByteOrderBig, cDepth8,
		 cChannels1 | cChannels3, cFrame32);
    ERR_PASS(expansions != 1 && expansions != 2);
    ERR_PASS(w_pixrGetChannelCount(pIPWState, srcPIXR, &src_channels));
    /* Can only do halftoning on single channel PIXRs */
/*    ERR_PASS( expansions > 1 && src_channels != 1); */

    /* Determine the image size, start address and line pitch */
    /* We'll assume here that all of the components are the same */
    ERR_PASS(w_pixrGetWidth(pIPWState, srcPIXR, &src_width));
    ERR_PASS(w_pixrGetHeight(pIPWState, srcPIXR, &src_height));
    ERR_PASS(w_pixrGetDepth(pIPWState, srcPIXR, &src_depth));
    ERR_PASS(w_pixrGetType(pIPWState, srcPIXR, &src_type));
    ERR_PASS(w_pixrGetResolutionDPI(pIPWState, srcPIXR, &src_xdpi, &src_ydpi));

    /* Verify that this is an 8-bit image */
    ERR_GEN(! (src_depth==8), cErrDepthNotSupported);

    /* Calculate compressed image size */
    switch (hvqLevels) {
    case 1:
	xexpansions = 2, yexpansions = 1;
	break;
    case 2:
	xexpansions = 2, yexpansions = 2;
	break;
    case 3:
	xexpansions = 4, yexpansions = 2;
	break;
    case 4:
	xexpansions = 4, yexpansions = 4;
	break;
    default:
	ERR_RETURN;
    }

    dst_depth = 8, dst_channels = src_channels;
    dst_type = src_type;
    if (expansions > 1) {
	xexpansions *= expansions;
	yexpansions *= expansions;
	dst_depth = 1;
	dst_channels = 1;
	dst_type = cColorNone;
    }
    dst_width = src_width * xexpansions;
    dst_height = src_height * yexpansions;
    dst_xdpi = ScaledInt1616_TO_Int32(src_xdpi) * xexpansions;
    dst_ydpi = ScaledInt1616_TO_Int32(src_ydpi) * yexpansions;
 
    /* Create a PIXR for the output HVQ image */
    ERR_PASS_CLEANUP( w_createPixrExtendedNoInit(
					pIPWState, dst_width, dst_height,
					dst_depth,
					Int32_TO_ScaledInt1616(dst_xdpi),
					Int32_TO_ScaledInt1616(dst_ydpi),
					32, dst_channels, dst_type, NULL,
					&dstPIXR));

    /* Decompress source PIXR into destination PIXR */
    ERR_PASS(wp_hvqd_image(pIPWState, srcPIXR, hvqLevels, expansions, dstPIXR));

    size = src_width * src_height * src_channels * src_depth / 8;
    outsize = dst_width * dst_height * dst_channels * dst_depth / 8;

    *dstPIXRptr = dstPIXR;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (dstPIXR) {
	w_destroyPixr(pIPWState, &dstPIXR);
    }
	
    ERR_RETURN;
}


#undef _IW_DHVQ_C_

/*
$Log: iw_dhvq.c,v $
Revision 1.1.1.1  1999/09/20 18:19:18  jlin
Imported sources

Revision 1.4  1997/04/15 22:12:46  ddavies
iw_dhvq.c -> /product/ipcore/ipwrap/src/RCS/iw_dhvq.c,v

Use non-initialized PIXR creation where possible.

Revision 1.3  1996/11/25 19:53:03  aber
aber on Mon Nov 25 11:52:28 PST 1996

iw_dhvq.c -> /product/ipcore/ipwrap/src/RCS/iw_dhvq.c,v

Initialize variables (even though uninitialized part is shifted away).

Revision 1.2  1996/11/23 00:11:48  aber
iw_dhvq.c -> /product/ipcore/ipwrap/src/RCS/iw_dhvq.c,v

CommonCheck: Fix comments to make coreCheckC happy. Update halftoning for portability.

Revision 1.1  1996/11/21 19:20:03  aber
Initial revision

*/
