/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2000, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*	iw_djpeg.c:	JPEG decompression wrappers:
 *
 *	Code to decompress 3 types of JPEG compressed (JFIF formatted) files:
 *		w_jDecompressToGray()
 *		w_jDecompressToGray4()
 *		w_jDecompressToRGB()
 *		w_jDecompressToGray_callback()
 *		w_jDecompressToGray4_callback()
 *		w_jDecompressToRGB_callback()
 *		w_jDecompressToGray_SIDCT()
 *		w_jDecompressToRGB_SIDCT()
 *
 *	Or, you may call the following:
 *		w_jdInit() - begin
 *        ** optional routine to help output control **
 *              w_jdGetImageLinesPerMCU - to get lines in one MCU row
 *              w_jdGetMCURowsPerImage - to get MCU rows in whole image
 *              w_jdGetErrorStr() - to get error description string
 *              w_jdFreeErrorStr() - to free memory stroing error message
 *		w_jdAllocMCURowBuffer() - buffer pointers for one MCU row
 *              w_jdMCURowFree() - free MCU buffer points
 *              w_jdSmartIDCTSwitch() - set smart IDCT flag or delete it
 *        **                                         **      
 *		w_jdMCURow() - decompress one MCU row
 *		w_jdTerm() - clean up
 *
 *	JFIF information:
 *		w_jdQueryJFIF()
 *              w_jdQueryJFIF_callback()
 *		w_jdJfifGetWidth()
 *		w_jdJfifGetHeight()
 *		w_jdJfifGetChannels()
 *              w_jdEndQueryJFIF()
 */

#define _IW_DJPEG_C_

#ifndef _WIN32_WCE
#include <stdio.h>
#endif

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "jpeglib.h"
#include "defines.pub"
#include "imageref.pub"
#include "frames.pub"

/* prototypes */
#include "shrpixr.pub"
#include "props.pub"
#include "jpeg.pub"
#include "iw_jpeg.pub"


IP_RCSINFO(iw_djpeg_c_RCSInfo, "$RCSfile: iw_djpeg.c,v $; $Revision: 1.9 $; $Date: 2001/11/14 00:10:06 $");


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/* Define a set of nibble masks - Little-Endian (PC) version */

#define cNib0MaskLE 0x000000f0
#define cNib1MaskLE 0x0000000f
#define cNib2MaskLE 0x0000f000
#define cNib3MaskLE 0x00000f00
#define cNib4MaskLE 0x00f00000
#define cNib5MaskLE 0x000f0000
#define cNib6MaskLE 0xf0000000
#define cNib7MaskLE 0x0f000000


/* Define a set of nibble masks - Big-Endian (Sun, Mac) version */

#define cNib0MaskBE 0xf0000000
#define cNib1MaskBE 0x0f000000
#define cNib2MaskBE 0x00f00000
#define cNib3MaskBE 0x000f0000
#define cNib4MaskBE 0x0000f000
#define cNib5MaskBE 0x00000f00
#define cNib6MaskBE 0x000000f0
#define cNib7MaskBE 0x0000000f



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Do a 8 bpp buffer to 4 bpp PIXR copy of line of image data
 *
 * LongDescription:::
 *  Do a 8 bpp buffer to 4 bpp PIXR copy of a line of image data
 *  for the two cases of big and little-endian machines
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcBuf::         Ptr to source line in 8 bpp buffer
 *  npix::           Number pixels/line (image width)
 *
 * Output:::
 *  dstBuf::         Ptr to dest line in 4 bpp PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jdCopyLine8To4 (
    IPWState           *pIPWState,
    UInt8              *srcBuf,
    UInt32              npix,
    UInt8              *dstBuf)
{
    PROLOGUE;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    UInt32 *dst;
    UInt32 *src;
    UInt32 iword0, iword1, oword;
    UInt32 nextra, nbytes, nnibs;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    oword = 0;          /* keep the compiler happy */
 
/**************************************************************
*
*      Little-endian 8 bit buffer to 4 bit buffer conversion
*      -----------------------------------------------------
*
*                     Source byte layout
*                     ------------------
*
*                  Even word         Odd word
*              |  D   C   B   A |  H   G   F   E  |
*  Pixel MSB     31  23  15   7   31  23  15   7
*
*
*
*                     Destination nibble layout
*                     -------------------------
*
*              |  G   H   E   F   C   D   A   B  |
*  Pixel MSB     31  27  23  19  15  11   7   3
*
****************************************************************/

/* 
  Handle the bulk of the transfers, moving two 32 bit words into one 
  while converting from 8 bpp to 4 bpp (same byte order) 
  Note somewhat dangerous pointer cast. This shouldn't be a problem
  as long as the memory is malloc'ed line by line. Malloc is
  advertised to guarantee alignment to the largest data types 
*/
    for (src=(UInt32 *)srcBuf, dst=(UInt32 *)dstBuf; 
         src<(UInt32 *)(srcBuf+(npix/8)*8); dst++)
    {

/* Move the 4 bytes of the even word to the 4 MS nibbles of oword */
        oword = 0;
        iword0 = *src++; /* Get even-numbered word */
        oword = iword0 & cNib0MaskLE;
        oword |= (iword0 >> (15-3)) & cNib1MaskLE;
        oword |= (iword0 >> (23-15)) & cNib2MaskLE;
        oword |= (iword0 >> (31-11)) & cNib3MaskLE;

/* Move the 4 bytes of the odd word to the 4 LS nibbles of oword */
        iword1 = *src++; /* Get odd-numbered word */
        oword |= (iword1 << (23-7)) & cNib4MaskLE;
        oword |= (iword1 << (19-15)) & cNib5MaskLE;
        oword |= (iword1 << (31-23)) & cNib6MaskLE;
        oword |= (iword1 >> (31-27)) & cNib7MaskLE;

        *dst = oword;
    }

/* Now deal with any remaining bytes */
    nextra = npix % 8;
    if (nextra > 0)
    {
        nbytes = nextra/2;      /* Whole output bytes */
        nnibs = nextra & 0x1;   /* Last nibble */

    /* If main part was executed, start where it left off */
        if (npix > 8)
        {
            srcBuf = (UInt8 *) src;
            dstBuf = (UInt8 *) dst;
        }

    /* Move 4 MSBs of each src byte into one dst byte */
        while (nbytes--)
            *dstBuf++ = ((*srcBuf++) & 0xF0) | ((*srcBuf++) >> 4);
    
    /* 
      Move 4 MSBs of last odd src byte (if any) to 4 MSBs 
      of dst byte, preserving original contents of dst LSBs
    */
        if (nnibs)
            *dstBuf = (*dstBuf & 0x0F) | ((*srcBuf) & 0xF0);
    }

#else       /* Big-endian version */
    UInt32 *dst;
    UInt32 *src;
    UInt32 iword0, iword1, oword;
    UInt32 nextra, nbytes, nnibs;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    oword = 0;          /* keep the compiler happy */
 
/**********************************************************************
*
*    Big-endian 8 bit buffer to 4 bit buffer (or Pixr) conversion
*    ------------------------------------------------------------
*
*                     Source byte layout
*                     ------------------
*
*                  Even word         Odd word
*              |  A   B   C   D |  E   F   G   H  |
*  Pixel MSB     31  23  15   7   31  23  15   7
*
*
*
*                     Destination nibble layout
*                     -------------------------
*
*              |  A   B   C   D   E   F   G   H  |
*  Pixel MSB     31  27  23  19  15  11   7   3
*
****************************************************************/
/* 
  Handle the bulk of the transfers, moving two 32 bit words into one 
  while converting from 8 bpp to 4 bpp (same byte order) 
  Note somewhat dangerous pointer cast. This shouldn't be a problem
  as long as the memory is malloc'ed line by line. Malloc is
  advertised to guarantee alignment to the largest data types 
*/
    for (src=(UInt32 *)srcBuf, dst=(UInt32 *)dstBuf; 
         src<(UInt32 *)(srcBuf+(npix/8)*8); dst++)
    {

/* Move the 4 bytes of the even word to the 4 MS nibbles of oword */
        oword = 0;
        iword0 = *src++; /* Get even-numbered word */
        oword = iword0 & cNib0MaskBE;
        oword |= (iword0 << (27-23)) & cNib1MaskBE;
        oword |= (iword0 << (23-15)) & cNib2MaskBE;
        oword |= (iword0 << (19-7)) & cNib3MaskBE;

/* Move the 4 bytes of the odd word to the 4 LS nibbles of oword */
        iword1 = *src++; /* Get odd-numbered word */
        oword |= (iword1 >> (31-15)) & cNib4MaskBE;
        oword |= (iword1 >> (23-11)) & cNib5MaskBE;
        oword |= (iword1 >> (15-7)) & cNib6MaskBE;
        oword |= (iword1 >> (7-3)) & cNib7MaskBE;

        *dst = oword;
    }

/* Now deal with any remaining bytes */
    nextra = npix % 8;
    if (nextra > 0)
    {
        nbytes = nextra/2;      /* Whole output bytes */
        nnibs = nextra & 0x1;   /* Last nibble */

    /* If main part was executed, start where it left off */
        if (npix > 8)
        {
            srcBuf = (UInt8 *) src;
            dstBuf = (UInt8 *) dst;
        }

    /* Move 4 MSBs of each src byte into one dst byte */
        while (nbytes--)
            *dstBuf++ = ((*srcBuf++) & 0xF0) | ((*srcBuf++) >> 4);
    
    /* 
      Move 4 MSBs of last odd src byte (if any) to 4 MSBs 
      of dst byte, preserving original contents of dst LSBs
    */
        if (nnibs)
            *dstBuf = (*srcBuf) & 0xF0;
    }
#endif

    NORMAL_RETURN;
}


#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Do a 8 bpp buffer to 4 bpp PIXR copy of line of image data
 *
 * LongDescription:::
 *  Do a 8 bpp buffer to 4 bpp PIXR copy of a line of image data.
 *  This case only applies to little endian machines, since 
 *  buffer and PIXR order are identical for big-endian machines,
 *  allowing wp_jdCopyLine8To4() to be used for buffers or PIXRs.<BR><BR>
 *
 *  Here the destination has its data in PIXR order, so we 
 *  need to use the Big-endian masks, regardless of the CPU
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcBuf::         Ptr to source line in 8 bpp buffer
 *  npix::           Number pixels/line (image width)
 *
 * Output:::
 *  dstBuf::         Ptr to dest line in 4 bpp PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jdCopyLine8To4Pixr (
    IPWState           *pIPWState,
    UInt8              *srcBuf,
    UInt32              npix,
    UInt8              *dstBuf)
{
    PROLOGUE;
    UInt32 *dst;
    UInt32 *src;
    UInt32 iword0, iword1, oword;
    UInt32 nextra, nbytes;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    oword = 0;          /* keep the compiler happy */
 
/**************************************************************
*
*      Little-endian 8 bit buffer to 4 bit Pixr conversion
*      -----------------------------------------------------
*
*                     Source byte layout
*                     ------------------
*
*                  Even word         Odd word
*              |  D   C   B   A |  H   G   F   E  |
*  Pixel MSB     31  23  15   7   31  23  15   7
*
*
*
*                     Destination nibble layout
*                     -------------------------
*
*              |  A   B   C   D   E   F   G   H  |
*  Pixel MSB     31  27  23  19  15  11   7   3
*
****************************************************************/
/* 
  Handle the bulk of the transfers, moving two 32 bit words into one 
  while converting from 8 bpp to 4 bpp and reversing the order
  Note somewhat dangerous pointer cast. This shouldn't be a problem
  as long as the memory is malloc'ed line by line. Malloc is
  advertised to guarantee alignment to the largest data types 
*/
    src=(UInt32 *)srcBuf; 
    dst=(UInt32 *)dstBuf;
    for ( ; src<(UInt32 *)(srcBuf+(npix/8)*8); dst++)
    {

/* Move the 4 bytes of the even word to the 4 MS nibbles of oword */
        oword = 0;
        iword0 = *src++; /* Get even-numbered word */
        oword = (iword0 << (31-7)) & cNib0MaskBE;
        oword |= (iword0 << (27-15)) & cNib1MaskBE;
        oword |= (iword0 ) & cNib2MaskBE;
        oword |= (iword0 >> (31-19)) & cNib3MaskBE;

/* Move the 4 bytes of the odd word to the 4 LS nibbles of oword */
        iword1 = *src++; /* Get odd-numbered word */
        oword |= (iword1 << (15-7)) & cNib4MaskBE;
        oword |= (iword1 >> (15-11)) & cNib5MaskBE;
        oword |= (iword1 >> (23-7)) & cNib6MaskBE;
        oword |= (iword1 >> (31-3)) & cNib7MaskBE;

        *dst = oword;
    }

/* Now deal with any remaining bytes */
    nextra = npix % 8;
    if (nextra > 0)
    {
	srcBuf = (UInt8 *) src;
	dstBuf = (UInt8 *) dst;

	nbytes = nextra;
	if(nbytes > 3)
	{
	    *dstBuf++ = (*(srcBuf+3)&0xf0) | ((*(srcBuf+2)>>4)&0x0f);
	    *dstBuf++ = (*(srcBuf+1)&0xf0) | ((*srcBuf >>4)&0x0f);
	    nbytes -= 4;
	    srcBuf += 4;
	}
	switch( nbytes)
	{
	case 1:
	    *dstBuf++ = ((*srcBuf >>4)&0x0f);
	    break;
	case 2:
	    *dstBuf++ = (*(srcBuf+1)&0xf0) | ((*srcBuf >>4)&0x0f);
	    break;
	case 3:
	    *dstBuf++ = ((*srcBuf++ >>4)&0x0f);
	    *dstBuf++ = (*(srcBuf+1)&0xf0) | ((*srcBuf >>4)&0x0f);
	    break;
	default:
	    break;
	}
    }


    NORMAL_RETURN;
}
#endif

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Convert line buffers of pixels in memory format to PIXR pixels.
 *
 * LongDescription:::
 *  This routine is only used/called on the PC.  Therefore, it will always
 *  do byteswapping since PC is little endian, but PIXR memory order is
 *  big endian.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  userBuf::        Ptr to 2D array of user's PIXR data buffer pointers
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion, byteswapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jdCvtMemToPixr (
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    JSAMPIMAGE          userBuf)
{
    PROLOGUE;
    UInt32 band, row, samp;
    UInt32 nbands, nrows, nsamps;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

/* Get the image size and band count */
    nrows = cinfo->lines;
    nbands = cinfo->j_dinfo.output_components;
    nsamps = cinfo->j_dinfo.image_width;

/*
  Do the byte swap - note that fast swap is in 32 bit words.
  We'll do the full 32 bit words with the fast swap
  and do the stragglers with UCHAR_ACCESS.
*/
    for (band=0; band<nbands; band++)
        for (row=0; row<nrows; row++)
	{
            cinfo->byteSwapProc(pIPWState,
				(UInt32 *) (cinfo->jpegBuf[band][row]),
                                (UInt32 *) (userBuf[band][row]),
                                (Int32) (nsamps/4));
            for (samp=nsamps&(~3); samp<nsamps; samp++)
                UCHAR_ACCESS(userBuf[band][row]+samp) =
                cinfo->jpegBuf[band][row][samp];
        }

    NORMAL_RETURN;
}
#endif

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Convert line of 8 bit gray buffer pixels to 4 bit grays.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  userBuf::        Ptr to 2D array of user's data buffer pointers
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jdCvtGray8ToGray4 (
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    JSAMPIMAGE          userBuf)
{
    PROLOGUE;
    UInt32 band, row;
    UInt32 nbands, nrows, nsamps;

/* Get the image size and band count */
    nrows = cinfo->lines;
    nbands = cinfo->j_dinfo.output_components;
    nsamps = cinfo->j_dinfo.image_width;

/* Do the depth conversion */
    for (band=0; band<nbands; band++)
        for (row=0; row<nrows; row++)
            ERR_PASS( wp_jdCopyLine8To4 (pIPWState, cinfo->jpegBuf[band][row],
					nsamps,
					userBuf[band][row]));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Convert line of 8 bit gray buffer pixels to 4 bit gray PIXR pixels.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  userBuf::        Ptr to 2D array of user's PIXR data buffer pointers
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jdCvtGray8ToGray4Pixr (
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    JSAMPIMAGE          userBuf)
{
    PROLOGUE;
    UInt32 band, row;
    UInt32 nbands, nrows, nsamps;

/* Get the image size and band count */
    nrows = cinfo->lines;
    nbands = cinfo->j_dinfo.output_components;
    nsamps = cinfo->j_dinfo.image_width;

/* Do the byte swap - note that its in 32 bit words */
    for (band=0; band<nbands; band++)
        for (row=0; row<nrows; row++)
	{
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
            ERR_PASS( wp_jdCopyLine8To4Pixr(pIPWState,
					   cinfo->jpegBuf[band][row],
					   nsamps,
					   userBuf[band][row]));
#else
	    ERR_PASS( wp_jdCopyLine8To4 (pIPWState, cinfo->jpegBuf[band][row],
					nsamps,
					userBuf[band][row]));
#endif
	}
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert data from the interleaved RGB into BSQ format.
 *
 * LongDescription:::
 *  Convert the data from user's interleaved RGB to BSQ format.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  userBuf::        Ptr to 2D array of RGB data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jdDeinterleaveMCURow (
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    UInt8            ***userBuf)

{
    PROLOGUE;
    UInt32 row, count;
    UInt32 nrows, nsamps;
    UInt8 *outRPtr, *outGPtr, *outBPtr;
    UInt8 *inPtr;
 
    /* Keep anal compilers happy */
    pIPWState = pIPWState;

/* Get the image size and band count */
    nrows = cinfo->lines;
    nsamps = cinfo->j_dinfo.image_width;

/* 
  Do the interleaving into packed RGBRGBRGB... 
  one row at a time.
*/
    
    for (row=0; row<nrows; row++)
    {
        outRPtr = userBuf[0][row];
        outGPtr = userBuf[1][row];
        outBPtr = userBuf[2][row];

        inPtr = cinfo->jpegBuf[0][row];

        count = nsamps + 1;
        while (--count)
        {
            *outRPtr++ = *inPtr++;
            *outGPtr++ = *inPtr++;
            *outBPtr++ = *inPtr++;
        }
    }


    NORMAL_RETURN;
}

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt  
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert data from the interleaved RGB into BSQ format (PIXR order).
 *
 * LongDescription:::
 *  Convert the data from user's interleaved RGB to BSQ format (PIXR order).
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  userBuf::        Ptr to 2D array of RGB data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jdDeinterleavePixrMCURow (
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    UInt8            ***userBuf)
{
    PROLOGUE;
    Int32 row, count;
    Int32 nrows, nsamps;
    UInt8 *outRPtr, *outGPtr, *outBPtr;
    UInt8 *inPtr;
 
    /* Keep anal compilers happy */
    pIPWState = pIPWState;

/* Get the image size and band count */
    nrows = cinfo->lines;
    nsamps = cinfo->j_dinfo.image_width;

/* 
  Do the interleaving into packed RGBRGBRGB... 
  one row at a time.
*/
    for (row=0; row<nrows; row++)
    {
        outRPtr = cinfo->jpegBuf[1][0];
        outGPtr = cinfo->jpegBuf[1][1];
        outBPtr = cinfo->jpegBuf[1][2];

        inPtr = cinfo->jpegBuf[0][row];
        count = nsamps + 1;
        while (--count)
        {
            *outRPtr++ = *inPtr++;
            *outGPtr++ = *inPtr++;
            *outBPtr++ = *inPtr++;
        }
        outRPtr = userBuf[0][row];
        outGPtr = userBuf[1][row];
        outBPtr = userBuf[2][row];

	cinfo->byteSwapProc(pIPWState, (UInt32 *)(cinfo->jpegBuf[1][0]),
			    (UInt32 *)(outRPtr),(Int32) (nsamps/4));
	for (count = nsamps&(~3); count<nsamps; count++)
	    UCHAR_ACCESS(outRPtr+count) = *(cinfo->jpegBuf[1][0]+count);

	cinfo->byteSwapProc(pIPWState, (UInt32 *)(cinfo->jpegBuf[1][1]),
			    (UInt32 *)(outGPtr),(Int32) (nsamps/4));
	for (count = nsamps&(~3); count<nsamps; count++)
	    UCHAR_ACCESS(outGPtr+count) = *(cinfo->jpegBuf[1][1]+count);

	cinfo->byteSwapProc(pIPWState, (UInt32 *)(cinfo->jpegBuf[1][2]),
			    (UInt32 *)(outBPtr),(Int32) (nsamps/4));
	for (count = nsamps&(~3); count<nsamps; count++)
	    UCHAR_ACCESS(outBPtr+count) = *(cinfo->jpegBuf[1][2]+count);
    }


    NORMAL_RETURN;
}
#endif

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Wrapper for JPEG decompression initialization
 *
 * LongDescription:::
 *    Performs the initialization steps for JPEG decompression.
 *    Callers specify the memory order of the output image.
 *    This interface was written for EasyScan/Amber, so that
 *    data could be processed to/from non-pixr sources. The read
 *    callback will almost always be an fread statement, unless
 *    the compressed data comes from a memory buffer. 
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::             Ptr to (de)compression info
 *  readCallback::      Callback function to read in compressed data
 *  userData::          Hook for passing in user data 
 *  jdSpecialOrders::   Optional function to execute during decompression
 *                      (For future usage: should be NULL now)
 *  memOrg::            Memory organization of image to be decompressed
 *                       May be one of:
 *                       JMEMORG_BIP         -Interleaved RGB;
 *                       JMEMORG_BSQ_MEM     -Separate bands - memory order;
 *                       JMEMORG_BSQ_PIXR    -Separate bands - pixr order;
 *                       JMEMORG_GRAY8_MEM   -8 bit gray - memory order;
 *                       JMEMORG_GRAY8_PIXR  -8 bit gray - pixr order;
 *                       JMEMORG_GRAY4_MEM   -4 bit gray - memory order;
 *                       JMEMORG_GRAY4_PIXR  -4 bit gray - pixr order;
 *  scale::             the output scaling factor to produce an output
 *                      image of dimensions 1, 1/2, 1/4 or 1/8 of full image.
 * Output:::
 *  pImageWidth::       Ptr to image width in pixels
 *  pImageHeight::      Ptr to number of scan lines in image
 *  pNumComponents::    Ptr to number of color channels in image
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdInit (
    IPWState               *pIPWState,
    Decompress_info_ptr     cinfo,
    ErrCode	    (CDECL *readCallback) (IPWState *pIPWState,
					   void *userData, UInt8 *buf, 
				           UInt32 nbytes, UInt32 *ncount),
    void                   *userData,
    ErrCode	    (CDECL *jdSpecialOrders)(DecompressInfo *),
    UInt32                  memOrg,
    Float64                 scale,
    Int32                  *pImageWidth,
    Int32                  *pImageHeight,
    Int32                  *pNumComponents,
    UInt32		    options)
{
    PROLOGUE;
    Int32           imageW, imageH, nComponents;
    Bool            progressive;
 
/* Do a sanity check on the (non-return) arguments */
    ERR_GEN(readCallback == NULL, cErrNullPtr);
    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(pImageWidth == NULL, cErrNullPtr);
    ERR_GEN(pImageHeight == NULL, cErrNullPtr);
    ERR_GEN(pNumComponents == NULL, cErrNullPtr);

/*
  If the "Special Orders" function ptr is non-null, use it as is.
  Otherwise, check if grayscale output is required and force it to
  be gray even if the compressed image is RGB.
*/
    cinfo->LocalJDSpecialOrders = jdSpecialOrders;
    cinfo->userMemOrg = memOrg;

/* Call the ipshared routine */
    ERR_PASS( jd_init (pIPWState, &cinfo->j_dinfo, readCallback, userData));

/* The init routine returns info about the compressed image sizes, etc */
/* So do a sanity check on the returned image information */
    ERR_PASS( jd_get_input_width (pIPWState, &cinfo->j_dinfo, &imageW));
    ERR_PASS( jd_get_input_height (pIPWState, &cinfo->j_dinfo, &imageH));
    ERR_PASS( jd_get_components (pIPWState, &cinfo->j_dinfo, &nComponents));

/* Set the output scale and get the effective output image size */
    ERR_PASS(jd_set_output_scale (pIPWState,
				  &cinfo->j_dinfo, scale, &imageW, &imageH));

    ERR_GEN(imageW <=0 || imageW > 32768 ||
	    imageH <=0 || imageH > 32768 ||
	    nComponents <=0 || nComponents > 3,
	    cErrInvalidParm);

    /*
      This switch assigns the data conversion routine 
      (when one is required) to a function pointer.
    */
    switch (memOrg)
    {
 /* Need interleaving conversion */
        case JMEMORG_BIP:   /* Band interleaved by pixel */
	    cinfo->cvtDecompImage = NULL;
	    break;

 /* Need byte swaping on PCs only */
        case JMEMORG_BSQ_PIXR: /* Band sequential - PIXR order */
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt  
	    ERR_PASS( i_getDIBPixrLineProc (pIPWState, cFalse,
					    cPixrToPixr, cFalse,
					    NULL, NULL,
					    &cinfo->byteSwapProc, NULL));
	    cinfo->cvtDecompImage = wp_jdDeinterleavePixrMCURow;
#else
	    cinfo->cvtDecompImage = wp_jdDeinterleaveMCURow;
#endif
	    break;

        case JMEMORG_GRAY8_PIXR: /* 8 bpp - PIXR order */
	    ERR_PASS(jd_force_grayscale_output (pIPWState, &cinfo->j_dinfo));
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt  
	    ERR_PASS( i_getDIBPixrLineProc (pIPWState, cFalse,
					    cPixrToPixr, cFalse,
					    NULL, NULL,
					    &cinfo->byteSwapProc, NULL));
	    cinfo->cvtDecompImage = wp_jdCvtMemToPixr;
#else
	    cinfo->cvtDecompImage = NULL;
#endif
	    break;

/* No conversions required */
        case JMEMORG_BSQ_MEM: /* Band sequential - memory order */
	    cinfo->cvtDecompImage = wp_jdDeinterleaveMCURow;
	    break;

        case JMEMORG_GRAY8_MEM: /* 8 bpp gray - memory order */
	    ERR_PASS(jd_force_grayscale_output (pIPWState, &cinfo->j_dinfo));
	    cinfo->cvtDecompImage = NULL;;
	    break;

/* Need depth conversion */
        case JMEMORG_GRAY4_MEM: /* 4 bpp - memory order */
	    ERR_PASS(jd_force_grayscale_output (pIPWState, &cinfo->j_dinfo));
	    cinfo->cvtDecompImage = wp_jdCvtGray8ToGray4;
	    break;

/* Need depth conversion with byte swapping on PCs */
        case JMEMORG_GRAY4_PIXR: /* 4 bpp - PIXR order */
	    ERR_PASS(jd_force_grayscale_output (pIPWState, &cinfo->j_dinfo));
	    cinfo->cvtDecompImage =  wp_jdCvtGray8ToGray4Pixr;
	    break;

        default:
	    cinfo->cvtDecompImage = NULL;
	    break;
    }

/* clearn smart IDCT flag (default is no SIDCT since it is 5-6 times slower */
    ERR_PASS( jd_set_smart_idct (pIPWState,  &cinfo->j_dinfo, 0 ));

/* Check progressive mode. If progDisplay is true, set buffer image flag */
    ERR_PASS( jd_check_pogressive (pIPWState,
				   &cinfo->j_dinfo, cinfo->progDisplay,
				    &progressive ));

    /* Mark as unknown colorspace, if it's not YCbCr */
    if (options & (JD_CIELAB | JD_ITULAB)) {
	ERR_PASS(jd_set_colorspace(pIPWState, &cinfo->j_dinfo, JCS_UNKNOWN));
    }

/* Set all of decompression table and data, no any change can be made */
    ERR_PASS( jd_start_decompress (pIPWState, &cinfo->j_dinfo));

    cinfo->jpegBuf = NULL;
    ERR_PASS( w_jdGetImageLinesPerMCU(pIPWState, cinfo, &cinfo->lines));

    ERR_PASS(w_jdAllocMCURowBuffer(pIPWState,cinfo, cTrue, 
				       &cinfo->jpegBuf));
    ERR_GEN (cinfo->jpegBuf == NULL, cErrNomem);
    cinfo->jpegBufLineMem = cTrue;

    *pImageWidth = imageW;
    *pImageHeight = imageH;
    *pNumComponents = nComponents;
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform JPEG decompression on one MCU row of an image.
 *
 * LongDescription:::
 *    Performs JPEG decompression on one MCU (minimum coded unit)
 *    row of the image. This is 8 lines for a gray scale image
 *    or 16 lines for an RGB image. The input comes from the file
 *    or buffer desrribed by the userData parameter to w_jdInit,
 *    using the readCallback function.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  userBuf::        The buffer containing the data to be decompressed.
 *                   (Ptr to 2D array of user's data buffer pointers).
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdMCURow (
    IPWState               *pIPWState,
    Decompress_info_ptr     cinfo,
    JSAMPIMAGE              userBuf)
{
    PROLOGUE;
    Int32 lineCount;

    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(userBuf == NULL, cErrNullPtr);

/* 
  Allow write-through directly to the user's
  buffer in those cases where the JPEG internal memory
  organization is identical to userMemOrg.
  Copy the userBuf line pointers into jpegBuf, rather
  than performing an intermediate write to jpegBuf.
*/
    ERR_PASS(jd_get_linesToRead (pIPWState, &cinfo->j_dinfo, &lineCount));
    cinfo->lines = lineCount;
    if( cinfo->cvtDecompImage == NULL)
    {
	ERR_PASS(jd_read_scanline (pIPWState, &cinfo->j_dinfo,
				   userBuf[0], lineCount));
    }
    else
    {
	ERR_PASS(jd_read_scanline (pIPWState, &cinfo->j_dinfo, 
				      cinfo->jpegBuf[0], lineCount));

	ERR_PASS( cinfo->cvtDecompImage(pIPWState, cinfo, userBuf));
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get line number in one MCU row for decompression.
 *
 * LongDescription:::
 *  Get line number in one MCU row for decompression.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  lines::         lines in one MCU row.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdGetImageLinesPerMCU(
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    Int32              *lines )
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    *lines = cinfo->j_dinfo.max_v_samp_factor * DCTSIZE;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set progressive display for decomposition.
 *
 * LongDescription:::
 *  Set progressive display model. Default model is false (no progressive
 *  output). This setting should be called before w_jdInit.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  cinfo::          Ptr to (de)compression info
 *  progDisplay::    True: output progressive pass, False: just final output. 
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdSetProgressiveDisplay(
    IPWState		   *pIPWState,
    Decompress_info_ptr     cinfo,
    Bool                    progDisplay)
{
    PROLOGUE;

	/* Keep compiler happy */
    pIPWState = pIPWState;

    cinfo->progDisplay = progDisplay;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Init progressive pass.
 *
 * LongDescription:::
 *  If set progressive flag, output will be multi-pass images. In every
 *  pass, it is necessary to init paas.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *  pass::           Pass number ( 1, 2, ....... )
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdProgPassStart(
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    Int32               pass )
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    ERR_PASS( jd_progStart_output (pIPWState, &cinfo->j_dinfo, pass) );  
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Term progressive pass.
 *
 * LongDescription:::
 *  If set progressive flag, output will be multi-pass images. In every
 *  pass, it is necessary to term paas.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdProgPassFinish(
    IPWState           *pIPWState,
    DecompressInfo     *cinfo)
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    ERR_PASS( jd_progFinish_output (pIPWState, &cinfo->j_dinfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check input data to determine whether progressive decompression is
 *  finished.
 *
 * LongDescription:::
 *  Check input data to determine whether progressive decompression is
 *  finished.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *  final_pass::     Ptr to boolean of final pass flag
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdCheckInputCompleted(
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    Bool               *final_pass)
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    ERR_PASS( jd_input_complete (pIPWState, &cinfo->j_dinfo, final_pass));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get number of output pass and number of output lines in current pass. 
 *
 * LongDescription:::
 *  Get number of output pass and number of output lines in current pass. 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *  pass::           Ptr to number of output pass
 *  lines::          Ptr to number of output lines in current pass
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  testing
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdGetOutputScanInfo(
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    Int32              *pass,
    Int32              *lines)
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    ERR_PASS( jd_get_output_pass_lines (pIPWState,
					&cinfo->j_dinfo, pass, lines));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get MCU rows in the image for decompression.
 *
 * LongDescription:::
 *  Get MCU rows in the image for decompression.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Output:::
 *  rows::           MCU rows in image.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdGetMCURowsPerImage(
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    Int32              *rows )
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    *rows = cinfo->j_dinfo.MCU_rows_in_scan;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Select smart IDCT or not.
 *
 * LongDescription:::
 *   If select smart IDCT, decompressed data will pass a
 *   filter based post processing. Otherwise, it will not.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *   flag::          Smart IDCT flag, 1: switch on, 0: switch off
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdSmartIDCTSwitch(
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    Int32              flag )
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    ERR_PASS( jd_set_smart_idct (pIPWState, &cinfo->j_dinfo, flag ));

    NORMAL_RETURN;
}
 
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Perform cleanup operations for termination of JPEG decompression.
 *
 * LongDescription:::
 *    Performs cleanup operations for JPEG decompression,
 *    including file trailer reading and closing and memory freeing.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdTerm (
    IPWState               *pIPWState,
    Decompress_info_ptr     cinfo)
{
    PROLOGUE;

    ERR_GEN(cinfo == NULL, cErrNullPtr);

    if(cinfo->jpegBuf != NULL)
    {
	ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cinfo->jpegBufLineMem, 
				cinfo->jpegBuf));
	cinfo->jpegBuf = NULL;
    }

    ERR_PASS( jd_term (pIPWState, &cinfo->j_dinfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Retrive error string for display error message.
 *
 * LongDescription:::
 *       Allows an application to retrieve a string containing
 *       an error string for display in an error message box
 *       or other such purpose.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  cinfo::          Ptr to compression info
 *
 * Output:::
 *  string::         Ptr to error string address
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdGetErrorStr(
    IPWState		   *pIPWState,
    Decompress_info_ptr     cinfo,
    char                  **string)
{
    PROLOGUE;

    ERR_CHECK_NO_MSG(jd_get_error_string (pIPWState, &cinfo->j_dinfo, string ))
	ERR_RETURN;
 
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Free memory stroing error meaasge.
 *
 * LongDescription:::
 *       In order to retrieve an error string, the error struct hasn't
 *    been freed in error exit. Provide this function to free memory
 *    storing all error messages and function points.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  cinfo::          Ptr to compression info
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdFreeErrorStr(
    IPWState		   *pIPWState,
    Decompress_info_ptr     cinfo)
{
    PROLOGUE;

    ERR_CHECK_NO_MSG(jd_free_error_string (pIPWState, &cinfo->j_dinfo ))
	ERR_RETURN;

    NORMAL_RETURN;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Provide interface wrapper for allocating JPEG decompression row buffer
 *
 * LongDescription:::
 *    Provide a general interface wrapper for allocating the row buffer
 *    This will selectively use either the normal ipshared function
 *    or the special function for BIP data, depending on the
 *    setting of cinfo.userMemOrg. Thus this function MUST be called
 *    only AFTER w_jdInit is called.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to (de)compression info
 *  allocLineMem::   Boolean.  cTrue if memory is actually to
 *                   be allocated to hold the decompressed image
 *                   data. If cFalse, a set of row pointers is
 *                   returned.  The caller is expected to set
 *                   the pointers to hir image buffers. cFalse
 *                   is to be used only when there is no
 *                   memory organization conversion needed
 *                   between the caller's buffer and the JPEG
 *                   decompressor's internal buffer. The JPEG
 *                   internal buffer is in the CPU's  natural
 *                   memory order, with a separate buffer for
 *                   each band (BSQ).  Thus, if the data goes to PIXR
 *                   order, cFalse can be used for the Sun, but
 *                   not for the PC.
 *
 * Output:::
 *  pData::          Ptr to the returned row buffer.
 *                   The buffers (allocLineMem=cTrue)
 *                   or row pointers (allocLineMem=cFalse)
 *                   allocated to hold (or point to) the image data.
 *                   The JSAMPIMAGE typedef is equivalent to UInt8 ***, and
 *                   holds data organized as buf[colorBand][line][column].
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdAllocMCURowBuffer (
    IPWState           *pIPWState,
    DecompressInfo     *cinfo,
    Bool                allocLineMem,
    JSAMPIMAGE         *pData)
{
    PROLOGUE;
    Int32 band,row, nlines, nsamps, nbands;
    JSAMPIMAGE  data = NULL;

    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(pData == NULL, cErrNullPtr);
    ERR_GEN(*pData != NULL, cErrPtrNotNull);

    ERR_PASS( w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines) );
    nsamps = cinfo->j_dinfo.image_width;
    switch (cinfo->userMemOrg)
    {
        case JMEMORG_BIP:
        case JMEMORG_GRAY8_MEM:
        case JMEMORG_GRAY4_MEM:
        case JMEMORG_GRAY4_PIXR:
        case JMEMORG_GRAY8_PIXR:
	    nbands =1;
	    break;

        case JMEMORG_BSQ_MEM:
        case JMEMORG_BSQ_PIXR:
	    nbands = 3;
	    break;

        default:
	    nbands = 1; /* Force to a single band (with 24 bit pixels) */
    }


/* Allocate the channel pointers */
    data = (JSAMPIMAGE) IPW_CALLOC(sizeof(JSAMPARRAY), nbands, cAllocatorData);
    ERR_GEN(data == NULL, cErrNomem);

    for (band=0; band < nbands; band++)
    {
        /* Allocate the row pointers */
        data[band] = (JSAMPARRAY) IPW_CALLOC(sizeof(JSAMPROW), 
					     nlines, cAllocatorData);
	ERR_GEN(data[band] == NULL, cErrNomem);

        for (row=0; row<nlines; row++)
        {
            if (allocLineMem) /* Alloc 24 bit row mem in 32 bit chunks */
            {
                data[band][row] = (JSAMPROW)IPW_CALLOC(sizeof(Int32), 
				      ((3*nsamps+3)/4), cAllocatorData);
		ERR_GEN(data[band][row] == NULL, cErrNomem);
            }
            else   /* Just set the row pointers to NULL */
                data[band][row] = (JSAMPROW) 0;
        }
 
    }

    *pData = data;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Provide interface wrapper for free JPEG compression row buffer.
 *
 * LongDescription:::
 *    Provide a general interface wrapper for free the row buffer.
 *  
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  cinfo::          Ptr to decompression info
 *  allocLineMem::   Boolean. cTrue if memory is actually to
 *                   be allocated to hold the uncompressed image
 *                   data. If cFalse, a set of row pointers is
 *                   returned. The caller is expected to set
 *                   the pointers to hir image buffers.  cFalse
 *                   is to be used only when there is no
 *                   memory organization conversion needed
 *                   between the caller's buffer and the JPEG
 *                   compressor's internal buffer. The JPEG
 *                   internal buffer is in the CPU's  natural 
 *                   memory order, with a separate buffer for
 *                   each band (BSQ). Thus, if the data is in PIXR
 *                   order, cFalse can be used for the Sun, but
 *                   not for the PC.
 *  pData::          Ptr to the row buffer.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdMCURowFree (
    IPWState           *pIPWState,
    DecompressInfo       *cinfo,
    Bool                allocLineMem,
    JSAMPIMAGE         pData)
{
    PROLOGUE;
    Int32 band, row, nlines, nsamps, nbands;
    
    ERR_GEN(cinfo == NULL, cErrNullPtr);
    ERR_GEN(pData == NULL, cErrNullPtr);
 
    ERR_PASS( w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    nsamps = cinfo->j_dinfo.image_width;
    switch (cinfo->userMemOrg)
    {
        case JMEMORG_BIP:
        case JMEMORG_GRAY8_MEM:
        case JMEMORG_GRAY4_MEM:
        case JMEMORG_GRAY4_PIXR:
        case JMEMORG_GRAY8_PIXR:
	    nbands =1;
	    break;

        case JMEMORG_BSQ_MEM:
        case JMEMORG_BSQ_PIXR:
	    nbands = 3;
	    break;

        default:
	    nbands = 1; /* Force to a single band (with 24 bit pixels) */
    }

    for (band=0; band < nbands; band++)
    {
	if (allocLineMem) /* Alloc 24 bit row mem in 32 bit chunks */
	{
	    for (row=0; row<nlines; row++)
	    {
                IPW_FREE(pData[band][row], cAllocatorData);
	    }
        }
	IPW_FREE(pData[band], cAllocatorData);
    }
    IPW_FREE(pData, cAllocatorData);

    NORMAL_RETURN;

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read callback for getting JPEG compressed data.
 *
 * LongDescription:::
 *  Read callback for use in w_jDecompress. The userData
 *  structure element must have been loaded with the file
 *  pointer to the file containing the compresssed image.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  userData::       Ptr to FILE ( save decompressed data )
 *  buf::            Ptr to byte buffer to read into.
 *  nbytes::         Number of bytes to read.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_jdReadCallback (
    IPWState	    *pIPWState,
    void            *userData,
    UInt8           *buf,
    UInt32           nbytes,
    UInt32           *ncount)
{
/* keep core check and compiler happy */
    PROLOGUE;
    pIPWState = pIPWState;
    __error__ = __error__;
 
    *ncount = (UInt32 )fread((void *)buf, (size_t)1, (size_t)nbytes, (FILE *) userData);
    NORMAL_RETURN;
}

 
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF formatted file to 8 bit gray PIXR.
 *
 * LongDescription:::
 *    Decompress a a stored JFIF format compressed image to
 *    an 8 bit gray pixr. If the compressed image is RGB, only
 *    the luminance will be used.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fdJFIF::         FILE ptr for the compressed file
 *  scale::          The output scaling factor to produce an output
 *                   image of dimensions 1, 1/2, 1/4 or 1/8 of full image.
 *
 * Output:::
 *  dstPixrPtr::     Ptr to ptr to 8 bit gray dest PIXR.
 *                   If *dstPixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressToGray (
    IPWState           *pIPWState,
    FILE               *fdJFIF,
    Float64             scale,
    PIXR              **dstPixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image;
    Int32               w, h, bpl, ncomps;
    Int32               row, line;
    Int32              nlines;
    PIXR               *dstPixr = NULL;
    JSAMPIMAGE          userBuf =NULL;

/* Verify arguments */
    ERR_GEN(fdJFIF == NULL, cErrNullPtr);
    ERR_GEN(dstPixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Initialize */
    ERR_PASS_CLEANUP(w_jdInit (pIPWState, cinfo, 
			wp_jdReadCallback,
			(void *) fdJFIF, 
			NULL,
			JMEMORG_GRAY8_PIXR, scale,
			&w, &h, &ncomps, 0));

/*
  If dstPixrPtr is NULL, assume that the caller wants one created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
/* Use a local pixr variable to avoid all the indirection */
    dstPixr = *dstPixrPtr;

    if (dstPixr == NULL)
	ERR_PASS_CLEANUP( w_createPixrNoInit(pIPWState, w, h, cDepth8,
					     300, 300, cFrameBits, &dstPixr));
    else
    {
	ERR_GEN_PIXR(dstPixr, cByteOrderBig, cDepth8, 
		     cChannels1, cFrame32);
    }

    *dstPixrPtr = dstPixr;

/* Get the location of the image and the line pitch */
    ERR_PASS( w_pixrGetImage(pIPWState, dstPixr, &image));
    ERR_PASS( w_pixrGetBpl(pIPWState, dstPixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, 
				       &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);
    
/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row<h; ++row)
    {
	userBuf[0][line] = image;
	image += bpl;
	++ line;
	if(line >= nlines || row == (h-1))
	{    
	    /* Decompress an MCU row into userBuf */
	    ERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *dstPixrPtr = dstPixr;
    NORMAL_RETURN;

ERR_CLEANUP:
    if(userBuf != NULL)
	w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF formatted file to 4 bit gray PIXR.
 *
 * LongDescription:::
 *    Decompress a a stored JFIF format compressed image to
 *    an 4 bit gray pixr. If the compressed image is RGB, only
 *    the luminance will be used.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fdJFIF::         FILE ptr for the compressed file
 *  scale::          The output scaling factor to produce an output
 *                   image of dimensions 1, 1/2, 1/4 or 1/8 of full image.
 *
 * Output:::
 *  dstPixrPtr::     Ptr to ptr to 4 bit gray dest PIXR.
 *                   If *dstPixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressToGray4 (
    IPWState           *pIPWState,
    FILE               *fdJFIF,
    Float64             scale,
    PIXR              **dstPixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image;
    JSAMPIMAGE          userBuf = NULL;
    Int32               w, h, bpl, ncomps;
    Int32               row, line;
    Int32              nlines;
    PIXR               *dstPixr = NULL;

/* Verify arguments */
    ERR_GEN(fdJFIF == NULL, cErrNullPtr);
    ERR_GEN(dstPixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Initialize */
    ERR_PASS_CLEANUP(w_jdInit (pIPWState, cinfo, 
			wp_jdReadCallback,
			(void *) fdJFIF, 
			NULL,
			JMEMORG_GRAY4_PIXR, scale,
			&w, &h, &ncomps, 0));

/*
  If dstPixrPtr is NULL, assume that the caller wants one created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
/* Use a local pixr variable to avoid all the indirection */
    dstPixr = *dstPixrPtr;

    if (dstPixr == NULL)
	ERR_PASS_CLEANUP( w_createPixrNoInit(pIPWState, w, h, cDepth4,
					     300, 300,
					     cFrameBits, &dstPixr));
    else
    {
	ERR_GEN_PIXR(dstPixr, cByteOrderBig, cDepth4, 
		     cChannels1, cFrame32);
    }

    *dstPixrPtr = dstPixr;

/* Get the location of the image and the line pitch */
    ERR_PASS( w_pixrGetImage(pIPWState, dstPixr, &image));
    ERR_PASS( w_pixrGetBpl(pIPWState, dstPixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, 
				       &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);
    
/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row<h; ++row)
    {
	userBuf[0][line] = image;
	image += bpl;
	++ line;
	if(line >= nlines || row == (h-1))
	{    
	    /* Decompress an MCU row into userBuf */
	    ERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *dstPixrPtr = dstPixr;
    NORMAL_RETURN;

ERR_CLEANUP:

    if(userBuf != NULL)
	w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF formatted file to a 3 channel x 8 bit RGB PIXR.
 *
 * LongDescription:::
 *    Decompress a stored JFIF format compressed image to
 *    an RGB pixr. 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fdJFIF::         FILE ptr for the compressed file
 *  scale::          The output scaling factor to produce an output
 *                   image of dimensions 1, 1/2, 1/4 or 1/8 of full image.
 *
 * Output:::
 *  pixrPtr::        Ptr to ptr to 3 channel x 8 bit RGB dest PIXR.
 *                   If *pixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressToRGB (
    IPWState           *pIPWState,
    FILE               *fdJFIF,
    Float64             scale,
    PIXR              **pixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image[3];
    JSAMPIMAGE          userBuf = NULL;
    Int32               w,h, bpl, ncomps;
    Int32               row, chan, line;
    Int32              nlines;
    PIXR               *pixr = NULL;
    Int32              xResDpi,yResDpi;
    ScaledInt1616      xRes,yRes;
    Int32              resUnit;
    

/* Verify arguments */
    ERR_GEN(fdJFIF == NULL, cErrNullPtr);
    ERR_GEN(pixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Perform the decompression initialization steps */
/* The init call returns the image size and # of components */
    ERR_PASS_CLEANUP( w_jdInit (pIPWState, cinfo, 
			wp_jdReadCallback,
			(void *) fdJFIF, 
			NULL,
			JMEMORG_BSQ_PIXR, scale,
			&w, &h, &ncomps, 0));

    ERR_PASS_CLEANUP( w_jdJfifGetResolutions(
			pIPWState,
                        cinfo,
                        &xResDpi,
                        &yResDpi,
                        &resUnit) ); 

    xRes = Int32_TO_ScaledInt1616(xResDpi);
    yRes = Int32_TO_ScaledInt1616(yResDpi);

    ERR_GEN(ncomps != 3, cErrInvalidParm);

/*
  If destination pixr is NULL, assume that the caller wants it created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
    if (*pixrPtr == NULL)
	ERR_PASS_CLEANUP(w_createPixrExtendedNoInit(pIPWState, w, h, cDepth8, 
			    xRes, yRes, cFrameBits, 3, cColorRGB, NULL, &pixr));
    else
    {
	pixr = *pixrPtr;
	ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth8, 
		     cChannels3, cFrame32);
    }

    for (chan=0; chan<3; chan++)
    {
 	PIXR  *subPixr;
	
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, chan, &subPixr));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subPixr, &(image[chan])));

    }
    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);
    
    
/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row<h; ++row)
    {
        for (chan=0; chan<3; chan++)
	{
	    userBuf[chan][line] = image[chan];
	    image[chan] += bpl;
	}
	++ line;
	if(line >= nlines || row == (h-1))
	{
        /* Decompress an MCU row into jpegBuf */
	    ERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *pixrPtr = pixr;
    NORMAL_RETURN;

ERR_CLEANUP:

    if(userBuf != NULL)
	IPW_FREE(userBuf, cAllocatorData);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF formatted file to 8 bit gray PIXR.
 *
 * LongDescription:::
 *    Decompress a a stored JFIF format compressed image to
 *    an 8 bit gray pixr. If the compressed image is RGB, only
 *    the luminance will be used.Read JFIF file through
 *    callback function.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  readCallback::   Callback function to read compressed data
 *  userData::       Hook for passing in user data 
 *  scale::          The output scaling factor to produce an output
 *                   image of dimensions 1, 1/2, 1/4 or 1/8 of full image.
 *
 * Output:::
 *  dstPixrPtr::     Ptr to ptr to 8 bit gray dest PIXR.
 *                   If *dstPixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressToGray_callback (
    IPWState           *pIPWState,
    ErrCode	(CDECL *readCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes, UInt32 *ncount),
    void                *userData,
    Float64             scale,
    PIXR              **dstPixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image;
    Int32               w, h, bpl, ncomps;
    Int32               row, line;
    Int32              nlines;
    PIXR               *dstPixr = NULL;
    JSAMPIMAGE          userBuf =NULL;

/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN(dstPixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Initialize */
    ERR_PASS_CLEANUP(w_jdInit (pIPWState, cinfo, 
			readCallback,
			(void *) userData, 
			NULL,
			JMEMORG_GRAY8_PIXR, scale,
			&w, &h, &ncomps, 0));

/*
  If dstPixrPtr is NULL, assume that the caller wants one created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
/* Use a local pixr variable to avoid all the indirection */
    dstPixr = *dstPixrPtr;

    if (dstPixr == NULL)
	ERR_PASS_CLEANUP( w_createPixrNoInit(pIPWState, w, h, cDepth8,
					     300, 300,
					     cFrameBits,  &dstPixr));
    else
    {
	ERR_GEN_PIXR(dstPixr, cByteOrderBig, cDepth8, 
		     cChannels1, cFrame32);
    }

    *dstPixrPtr = dstPixr;

/* Get the location of the image and the line pitch */
    ERR_PASS( w_pixrGetImage(pIPWState, dstPixr, &image));
    ERR_PASS( w_pixrGetBpl(pIPWState, dstPixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, 
				       &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);
    
/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row<h; ++row)
    {
	userBuf[0][line] = image;
	image += bpl;
	++ line;
	if(line >= nlines || row == (h-1))
	{    
	    /* Decompress an MCU row into userBuf */
	    ERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *dstPixrPtr = dstPixr;
    NORMAL_RETURN;

ERR_CLEANUP:
    if(userBuf != NULL)
	w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF formatted file to 4 bit gray PIXR.
 *
 * LongDescription:::
 *    Decompress a a stored JFIF format compressed image to
 *    an 4 bit gray pixr. If the compressed image is RGB, only
 *    the luminance will be used. Read JFIF compressed image
 *    through callback function.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  readCallback::   Callback function to read compressed data
 *  userData::       Hook for passing in user data 
 *  scale::          The output scaling factor to produce an output
 *                   image of dimensions 1, 1/2, 1/4 or 1/8 of full image.
 *
 * Output:::
 *  dstPixrPtr::     Ptr to ptr to 4 bit gray dest PIXR.
 *                   If *dstPixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressToGray4_callback (
    IPWState           *pIPWState,
    ErrCode	(CDECL *readCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes, UInt32 *ncount),
    void                *userData,
    Float64             scale,
    PIXR              **dstPixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image;
    JSAMPIMAGE          userBuf = NULL;
    Int32               w, h, bpl, ncomps;
    Int32               row, line;
    Int32              nlines;
    PIXR               *dstPixr = NULL;

/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN(dstPixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Initialize */
    ERR_PASS_CLEANUP(w_jdInit (pIPWState, cinfo, 
			readCallback,
			(void *) userData, 
			NULL,
			JMEMORG_GRAY4_PIXR, scale,
			&w, &h, &ncomps, 0));

/*
  If dstPixrPtr is NULL, assume that the caller wants one created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
/* Use a local pixr variable to avoid all the indirection */
    dstPixr = *dstPixrPtr;

    if (dstPixr == NULL)
	ERR_PASS_CLEANUP( w_createPixrNoInit(pIPWState, w, h, cDepth4,
					     300, 300,
					     cFrameBits, &dstPixr));
    else
    {
	ERR_GEN_PIXR(dstPixr, cByteOrderBig, cDepth4, 
		     cChannels1, cFrame32);
    }

    *dstPixrPtr = dstPixr;

/* Get the location of the image and the line pitch */
    ERR_PASS( w_pixrGetImage(pIPWState, dstPixr, &image));
    ERR_PASS( w_pixrGetBpl(pIPWState, dstPixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, 
				       &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);
    
/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row<h; ++row)
    {
	userBuf[0][line] = image;
	image += bpl;
	++ line;
	if(line >= nlines || row == (h-1))
	{    
	    /* Decompress an MCU row into userBuf */
	    ERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *dstPixrPtr = dstPixr;
    NORMAL_RETURN;

ERR_CLEANUP:

    if(userBuf != NULL)
	w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF formatted file to a 3 channel x 8 bit RGB PIXR.
 *
 * LongDescription:::
 *    Decompress a stored JFIF format compressed image to
 *    an RGB pixr. Read JFIF format compressed image through
 *    callback function.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  readCallback::   Callback function to read compressed data
 *  userData::       Hook for passing in user data 
 *  scale::          The output scaling factor to produce an output
 *                   image of dimensions 1, 1/2, 1/4 or 1/8 of full image.
 *
 * Output:::
 *  pixrPtr::        Ptr to ptr to 3 channel x 8 bit RGB dest PIXR.
 *                   If *pixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressToRGB_callback (
    IPWState           *pIPWState,
    ErrCode	(CDECL *readCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes, UInt32 *ncount),
    void                *userData,
    Float64             scale,
    PIXR              **pixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image[3];
    JSAMPIMAGE          userBuf = NULL;
    Int32               w,h, bpl, ncomps;
    Int32               row, chan, line;
    Int32              nlines;
    PIXR               *pixr = NULL;

/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN(pixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Perform the decompression initialization steps */
/* The init call returns the image size and # of components */
    ERR_PASS_CLEANUP( w_jdInit (pIPWState, cinfo, 
			readCallback,
			(void *) userData, 
			NULL,
			JMEMORG_BSQ_PIXR, scale,
			&w, &h, &ncomps, 0));

    ERR_GEN(ncomps != 3, cErrInvalidParm);

/*
  If destination pixr is NULL, assume that the caller wants it created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
    if (*pixrPtr == NULL)
	ERR_PASS_CLEANUP(w_createPixrExtendedNoInit(pIPWState, w, h, cDepth8, 
			     300, 300, cFrameBits, 3, cColorRGB, NULL, &pixr));
    else
    {
	pixr = *pixrPtr;
	ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth8, 
		     cChannels3, cFrame32);
    }

    for (chan=0; chan<3; chan++)
    {
 	PIXR  *subPixr;
	
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, chan, &subPixr));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subPixr, &(image[chan])));

    }
    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);
    
    
/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row<h; ++row)
    {
        for (chan=0; chan<3; chan++)
	{
	    userBuf[chan][line] = image[chan];
	    image[chan] += bpl;
	}
	++ line;
	if(line >= nlines || row == (h-1))
	{
        /* Decompress an MCU row into jpegBuf */
	    ERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *pixrPtr = pixr;
    NORMAL_RETURN;

ERR_CLEANUP:

    if(userBuf != NULL)
	IPW_FREE(userBuf, cAllocatorData);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF formatted file to 8 bit gray PIXR with
 *  selection of smart IDCT and without scale selection.
 *
 * LongDescription:::
 *    Decompress a a stored JFIF format compressed image to
 *    an 8 bit gray pixr with SIDCT switch and without scale down 
 *    selection. SIDCT works with(scale =1.0). If the compressed images
 *    is RGB, only the luminance will be used.Read JFIF file through
 *    callback function.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  readCallback::   Callback function to read compressed data
 *  userData::       Hook for passing in user data 
 *  SIDCT_flag::     Smart IDCT flag, 1: switch on , 0: switch off 
 *
 * Output:::
 *  dstPixrPtr::     Ptr to ptr to 8 bit gray dest PIXR.
 *                   If *dstPixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressToGray_SIDCT (
    IPWState           *pIPWState,
    ErrCode	(CDECL *readCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes, UInt32 *ncount),
    void                *userData,
    Int32               SIDCT_flag,
    PIXR              **dstPixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image;
    Int32               w, h, bpl, ncomps;
    Int32               row, line;
    Int32              nlines;
    PIXR               *dstPixr = NULL;
    JSAMPIMAGE          userBuf =NULL;

/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN(dstPixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Initialize */
    ERR_PASS_CLEANUP(w_jdInit (pIPWState, cinfo, 
			readCallback,
			(void *) userData, 
			NULL,
			JMEMORG_GRAY8_PIXR, 1.0,
			&w, &h, &ncomps, 0));

/*
  If dstPixrPtr is NULL, assume that the caller wants one created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
/* Use a local pixr variable to avoid all the indirection */
    dstPixr = *dstPixrPtr;

    if (dstPixr == NULL)
	ERR_PASS_CLEANUP( w_createPixrNoInit(pIPWState, w, h, cDepth8,
					     300, 300,
					     cFrameBits,  &dstPixr));
    else
    {
	ERR_GEN_PIXR(dstPixr, cByteOrderBig, cDepth8, 
		     cChannels1, cFrame32);
    }

    *dstPixrPtr = dstPixr;

/* Get the location of the image and the line pitch */
    ERR_PASS( w_pixrGetImage(pIPWState, dstPixr, &image));
    ERR_PASS( w_pixrGetBpl(pIPWState, dstPixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, 
				       &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);

/* select smart IDCT */
    ERR_PASS_CLEANUP(w_jdSmartIDCTSwitch(pIPWState, cinfo, SIDCT_flag));

/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row<h; ++row)
    {
	userBuf[0][line] = image;
	image += bpl;
	++ line;
	if(line >= nlines || row == (h-1))
	{    
	    /* Decompress an MCU row into userBuf */
	    ERR_PASS( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *dstPixrPtr = dstPixr;
    NORMAL_RETURN;

ERR_CLEANUP:
    if(userBuf != NULL)
	w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF formatted file to a 3 channel x 8 bit RGB PIXR.
 *  with smart IDCT selection and without scale size selection.
 *
 * LongDescription:::
 *    Decompress a stored JFIF format compressed image to
 *    an RGB pixr with the selection of smart IDCT and without the 
 *    selection of scaling size. Read JFIF format compressed image through
 *    callback function.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  readCallback::   Callback function to read compressed data
 *  userData::       Hook for passing in user data 
 *  SIDCT_flag::     Smart IDCT flag, 1: switch on , 0: switch off 
 *
 * Output:::
 *  pixrPtr::        Ptr to ptr to 3 channel x 8 bit RGB dest PIXR.
 *                   If *pixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressToRGB_SIDCT (
    IPWState           *pIPWState,
    ErrCode	(CDECL *readCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes, UInt32 *ncount),
    void                *userData,
    Int32               SIDCT_flag,
    PIXR              **pixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image[3];
    JSAMPIMAGE          userBuf = NULL;
    Int32               w,h, bpl, ncomps;
    Int32               row, chan, line;
    Int32              nlines;
    PIXR               *pixr = NULL;

/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN(pixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Perform the decompression initialization steps */
/* The init call returns the image size and # of components */
    ERR_PASS_CLEANUP( w_jdInit (pIPWState, cinfo, 
			readCallback,
			(void *) userData, 
			NULL,
			JMEMORG_BSQ_PIXR, 1.0,
			&w, &h, &ncomps, 0));

    ERR_GEN(ncomps != 3, cErrInvalidParm);

/*
  If destination pixr is NULL, assume that the caller wants it created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
    if (*pixrPtr == NULL)
	ERR_PASS_CLEANUP(w_createPixrExtendedNoInit(pIPWState, w, h, cDepth8, 
			     300, 300, cFrameBits, 3, cColorRGB, NULL, &pixr));
    else
    {
	pixr = *pixrPtr;
	ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth8, 
		     cChannels3, cFrame32);
    }

    for (chan=0; chan<3; chan++)
    {
 	PIXR  *subPixr;
	
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, chan, &subPixr));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subPixr, &(image[chan])));

    }
    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);
    
/* select smart IDCT */
    ERR_PASS_CLEANUP(w_jdSmartIDCTSwitch(pIPWState, cinfo, SIDCT_flag));
    
/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row<h; ++row)
    {
        for (chan=0; chan<3; chan++)
	{
	    userBuf[chan][line] = image[chan];
	    image[chan] += bpl;
	}
	++ line;
	if(line >= nlines || row == (h-1))
	{
        /* Decompress an MCU row into jpegBuf */
	    ERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *pixrPtr = pixr;
    NORMAL_RETURN;

ERR_CLEANUP:

    if(userBuf != NULL)
	IPW_FREE(userBuf, cAllocatorData);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Decompress a JFIF or progressive JFIF formatted file to pixr.
 *
 * LongDescription:::
 *  (JPEG) Decompress a JFIF or progressive JFIF formatted file to pixr.
 *   Output image's components and depth are same as compressed image.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  readCallback::   Callback function to read compressed data
 *  userData::       Hook for passing in user data 
 *  scale::          The output scaling factor to produce an output
 *                   image of dimensions 1, 1/2, 1/4 or 1/8 of full image.
 *
 * Output:::
 *  pixrPtr::        Ptr to dest PIXR. 
 *                   If *pixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *   If want pass dest Pixr in, get compressed image info first, then allocate
 *   Pixr and pass it in. The Pixr's component, depth, and so on have to be
 *   same as compressed image.
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, format conversion, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jDecompressProgressive (
    IPWState           *pIPWState,
    ErrCode	(CDECL *readCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes, UInt32 *ncount),
    void                *userData,
    Float64             scale,
    PIXR              **pixrPtr)
{
    PROLOGUE;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8              *image[3], *pimg[3];
    JSAMPIMAGE          userBuf = NULL;
    Int32               w,h, bpl, ncomps;
    Int32               row, chan, line;
    Int32              nlines, pass;
    PIXR               *pixr = NULL;
    Bool                final_pass;

/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN(pixrPtr == NULL, cErrNullPtr);

/*  Cleanup cinfo */
    memset ((void *)cinfo, 0, sizeof (DecompressInfo));

/* Set progressive display */
    ERR_PASS( w_jdSetProgressiveDisplay (pIPWState, cinfo, cTrue));

/* Perform the decompression initialization steps */
/* The init call returns the image size and # of components */
    ERR_PASS_CLEANUP( w_jdInit (pIPWState, cinfo, 
			readCallback,
			(void *) userData, 
			NULL,
			JMEMORG_BSQ_PIXR, scale,
			&w, &h, &ncomps, 0));

    ERR_GEN(ncomps != 3, cErrInvalidParm);

/*
  If destination pixr is NULL, assume that the caller wants it created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
    if (*pixrPtr == NULL)
    {
      if( ncomps == 1)
      	ERR_PASS_CLEANUP( w_createPixr(pIPWState, w, h, cDepth8,
					     300, 300,
					     cFrameBits,  &pixr));
      else
   	ERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, w, h, cDepth8, 
			    300, 300, cFrameBits, 3, cColorRGB, NULL, &pixr));
    }
    else
    {
	pixr = *pixrPtr;
	ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth8, 
		     cChannels3|cChannels1, cFrame32);
    }

    for (chan=0; chan<ncomps; chan++)
    {
 	PIXR  *subPixr;
	
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, chan, &subPixr));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subPixr, &(pimg[chan])));

    }
    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &bpl));

    ERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState,cinfo, cFalse, &userBuf));
    ERR_GEN (userBuf == NULL, cErrNomem);
    
    
/* Decompress the image an MCU row at a time */
    ERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));

    final_pass = cFalse;
    for( pass=1; !final_pass; ++pass)
    {
	ERR_PASS_CLEANUP( w_jdProgPassStart (pIPWState, cinfo, pass) );  
	for(chan=0; chan<3; ++chan)
	    image[chan] = pimg[chan];

	for (row = 0; row<h; row += nlines)
	{
	    for(line =0; line < nlines && (row+line) < h; ++line)
	    {
		for (chan=0; chan<ncomps; chan++)
		{
		    userBuf[chan][line] = image[chan];
		    image[chan] += bpl;
		}
	    }

	    /* Decompress an MCU row into jpegBuf */
	    ERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
            /*  Display image here */
	}
	ERR_PASS_CLEANUP( w_jdProgPassFinish (pIPWState, cinfo));
	ERR_PASS_CLEANUP( w_jdCheckInputCompleted (pIPWState, cinfo, 
						   &final_pass));
    }

    ERR_PASS(w_jdMCURowFree(pIPWState,cinfo, cFalse, userBuf));

    ERR_PASS( w_jdTerm(pIPWState, cinfo));

    *pixrPtr = pixr;
    NORMAL_RETURN;

ERR_CLEANUP:

    if(userBuf != NULL)
	IPW_FREE(userBuf, cAllocatorData);

    jpeg_destroy_decompress (pIPWState, &cinfo->j_dinfo);

    w_jdFreeErrorStr (pIPWState, cinfo);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Query a JFIF file to get its width, height, # of components etc.
 *
 * LongDescription:::
 *    Performs only enough of the initial JPEG decompression process
 *    to get header information. The file is opened, read and
 *    then closed again. This function must be called before
 *    any of the specific query functions w_jdJfifGetWidth(),
 *    w_jdJfifGetHeight(), or w_jdJfifGetChannels().
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fdJFIF::         FILE ptr to compressed file.
 *
 * Output:::
 *  jQuery::         Ptr to allocated querying info state structure.  Will be
 *                   filled in with info pulled from the compressed file.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdQueryJFIF (
    IPWState               *pIPWState,
    FILE                   *fdJFIF,
    Decompress_info_ptr     jQuery)
{
    PROLOGUE;

    ERR_GEN(jQuery == NULL, cErrNullPtr);
    ERR_GEN(fdJFIF == NULL, cErrNullPtr);

    jQuery->jpegBuf = NULL;
    ERR_PASS( jd_init (pIPWState, &jQuery->j_dinfo, wp_jdReadCallback, 
		       (void *)fdJFIF));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (JPEG) Query a JFIF file to get its width, height, # of components etc.
 *
 * LongDescription:::
 *    Performs only enough of the initial JPEG decompression process
 *    to get header information. The file is opened and read through
 *    assigned callback function. This function must be called before
 *    any of the specific query functions w_jdJfifGetWidth(),
 *    w_jdJfifGetHeight(), or w_jdJfifGetChannels().
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  readCallback::      Callback function to read in compressed data
 *  userData::          Hook for passing in user data 
 *
 * Output:::
 *  jQuery::         Ptr to allocated querying info state structure.  Will be
 *                   filled in with info pulled from the compressed file.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdQueryJFIF_callback (
    IPWState               *pIPWState,
    ErrCode	    (CDECL *readCallback) (IPWState *pIPWState,
					   void *userData, UInt8 *buf, 
					   UInt32 nbytes, UInt32 *ncount),
    void                   *userData,
    Decompress_info_ptr     jQuery)
{
    PROLOGUE;

    ERR_GEN(jQuery == NULL, cErrNullPtr);
    ERR_GEN(userData == NULL, cErrNullPtr);

    jQuery->jpegBuf = NULL;
    ERR_PASS( jd_init (pIPWState, &jQuery->j_dinfo,
		       readCallback, (void *)userData));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  End (JPEG) Query a JFIF file. 
 *
 * LongDescription:::
 *  After you call w_jdQueryJFIF or w_jdQueryJFIF_callback and 
 *  get width, height and channel,  you have to
 *  call w_jdEndQueryJFIF before do decompression. Otherwise the decompression 
 *  will be failed.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  jQuery::         Ptr to allocated querying info state structure.  Will be
 *                   filled in with info pulled from the compressed file.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdEndQueryJFIF (
    IPWState               *pIPWState,
    Decompress_info_ptr     jQuery)
{
    PROLOGUE;

    ERR_GEN(jQuery == NULL, cErrNullPtr);
 
    ERR_PASS( jd_endQueryJFIF (pIPWState, &jQuery->j_dinfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
*    Get the image width in pixels
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  jQuery::         Ptr to state made in w_jdQueryJFIF().
 *
 * Output:::
 *  pWidth::         Ptr to the image width in pixels.
 *
 * Returns:::
 *
 * Notes:::
 *  
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdJfifGetWidth (
    IPWState               *pIPWState,
    Decompress_info_ptr     jQuery,
    Int32                  *pWidth)
{
    PROLOGUE;

    ERR_GEN(jQuery == NULL, cErrNullPtr);
    ERR_GEN(pWidth == NULL, cErrNullPtr);

    ERR_PASS(jd_get_input_width (pIPWState, &jQuery->j_dinfo, pWidth));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Get the image height in number of lines
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  jQuery::         Ptr to state made in w_jdQueryJFIF().
 *
 * Output:::
 *  pHeight::        Ptr to image height in number of line.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdJfifGetHeight (
    IPWState               *pIPWState,
    Decompress_info_ptr     jQuery,
    Int32                  *pHeight)
{
    PROLOGUE;

    ERR_GEN(jQuery == NULL, cErrNullPtr);
    ERR_GEN(pHeight == NULL, cErrNullPtr);

    ERR_PASS(jd_get_input_height (pIPWState, &jQuery->j_dinfo, pHeight));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Get the number of channels in the compressed image
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  jQuery::         Ptr to state made in w_jdQueryJFIF().
 *
 * Output:::
 *  pChannels::      Ptr to number of channels in the compressed image
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdJfifGetChannels (
    IPWState               *pIPWState,
    Decompress_info_ptr     jQuery,
    Int32                  *pChannels)
{
    PROLOGUE;

    ERR_GEN(jQuery == NULL, cErrNullPtr);
    ERR_GEN(pChannels == NULL, cErrNullPtr);

    ERR_PASS(jd_get_components (pIPWState, &jQuery->j_dinfo, pChannels));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Check progressive model.
 *
 * LongDescription:::
 *   Check whether it is progressive jpeg compression. 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  jQuery::         Ptr to state made in w_jdQueryJFIF().
 *
 * Output:::
 *  progressive::    Ptr to boolean flag
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io, color, multi-channel, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdCheckProgressive(
    IPWState               *pIPWState,
    Decompress_info_ptr     jQuery,
    Bool                   *progressive)
{
    PROLOGUE;

  /* Keep compilers happy */
    pIPWState = pIPWState;

    ERR_PASS( jd_check_pogressive (pIPWState,  &jQuery->j_dinfo, (Bool)cFalse,
				   progressive ));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Get the resolution of the compressed image
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  jQuery::         Ptr to state made in w_jdQueryJFIF().
 *
 * Output:::
 *  pXRes::          Ptr to the horizontal resolution of compressed image
 *  pYRes::          Ptr to the vertical resolution of compressed image
 *  pUnit::          Ptr to the unit, 0:unkown resolution, 
 *                      1: resolution, dots/inch, 2: resolution, dots/cm
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  JPEG
 *
 * Keywords:::
 *  JPEG, compression, file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry::
 *
 * Citations:::
 *  wrappers for Independent JPEG Group's software
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_jdJfifGetResolutions (
    IPWState               *pIPWState,
    Decompress_info_ptr     jQuery,
    Int32                  *pXRes,
    Int32                  *pYRes,
    Int32                  *pUnit)
{
    PROLOGUE;

    ERR_GEN(jQuery == NULL, cErrNullPtr);
    ERR_GEN(pXRes == NULL || pYRes == NULL, cErrNullPtr);

    ERR_PASS(jd_get_resolution (pIPWState,
				&jQuery->j_dinfo, pXRes, pYRes, pUnit));

    NORMAL_RETURN;
}

#undef _IW_DJPEG_C_

/*
$Log: iw_djpeg.c,v $
Revision 1.9  2001/11/14 00:10:06  dabercro
Fix the resolution designation to that of the actual image.

Revision 1.8  2001/06/27 19:13:43  dabercro
Remove colorspace support in JPEG, Lab is supported by identifying the
colorspace as unknown.

Revision 1.7  2001/01/18 00:35:20  dabercro
Comment fixes.

Revision 1.6  2000/08/01 21:56:50  dabercro
Copyright notice; WinCE add. Changes from Mark Tobin.

Revision 1.5  2000/06/09 23:15:16  jlin
fix read write callback functions return type for JPEG

Revision 1.4  2000/04/07 23:59:48  dabercro
Added support for CIELAB and ICCLAB color spaces.

Revision 1.3  2000/02/26 00:12:29  dabercro
Allow gammut specification for ITULAB conversion.

Revision 1.2  2000/02/01 00:39:04  dabercro
Added ITULAB color space.

Revision 1.61  1998/01/15 01:49:41  ddavies
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

CommonCheck: Fix WATCOM warnings about unused procedure args.

Revision 1.60  1998/01/14 00:13:36  ddavies
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.59  1997/11/19 23:36:45  xzhu
xzhu on Wed Nov 19 15:34:55 PST 1997

iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Added new function w_jdJfifGetResolutions. It will be used in the same way  as w_jdJfifgetChannels.

Revision 1.58  1997/07/11 01:59:05  ddavies
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.57  1997/07/02 01:03:57  xzhu
xzhu on Tue Jul 1 18:03:21 PDT 1997

iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Added progressive jpeg decompression functions.

Revision 1.56  1997/04/15 22:13:11  ddavies
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Use non-initialized PIXR creation where possible.

Revision 1.55  1997/02/20 22:31:31  xzhu
xzhu on Thu Feb 20 14:29:04 PST 1997

iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Fix PC swaping bug in wp_jdDeinterleavePixerMCURow(), reported by Dave.

Revision 1.54  1996/10/29 02:19:46  xzhu
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

CommonCheck: Added new functions to retrive error string and free error message memory.

Revision 1.53  1996/09/23 18:02:07  xzhu
xzhu on Mon Sep 23 11:00:52 PDT 1996

iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Added short comments at the begining.

Revision 1.52  1996/09/20 21:52:19  xzhu
xzhu on Fri Sep 20 14:51:08 PDT 1996

iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

IPWrap function for smart IDCT.

Revision 1.51  1996/09/17 19:57:27  xzhu
xzhu on Tue Sep 17 12:56:03 PDT 1996

iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Added cFrameBit in all createPixr function to replace 0 -- no frame.

Revision 1.50  1996/08/30 22:13:35  xzhu
xzhu on Fri Aug 30 15:10:11 PDT 1996

iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Added new functions: w_jdQureJFIF_callback, w_jDecompressToGray_callback, w__jDecompressToGray4_callback and w_jDecompressToRGB_callback. These functions use callback function and userdat to get compressed data instead of using FILE *. Sometimes FILE* is hard for user to give.

Revision 1.49  1996/04/23 17:49:16  xzhu
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

CommonCheck: Fixed PC compiler warning.

Revision 1.48  1996/04/23 00:20:56  xzhu
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

CommonCheck: Fixed core check warning by removing comment about input and changing input parameter position and so on.

Revision 1.47  1996/04/20 01:48:08  xzhu
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Revision 1.46  1996/02/08 01:21:35  ejaquith
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

Fixed initialization bug that was killing a routine.

Revision 1.45  1996/01/04 20:24:38  ejaquith
iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.44  1996/01/04  19:33:37  dferrell
 * dferrell on Thu Jan 4 11:32:40 PST 1996
 *
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Use pIPWState variable in both big and little endian sections so
 * compilers don't needlessly warn.
 *
 * Revision 1.43  1995/12/13  04:10:09  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.42  1995/11/13  20:13:30  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.41  1995/11/04  01:35:59  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.40  1995/11/03  22:48:30  ddavies
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.39  1995/10/18  04:16:40  ddavies
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Add extra pointers to i_getDIBPixrLineProc call to reflect the four
 * types of procedures it could return.
 *
 * Revision 1.38  1995/10/13  22:02:45  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.37  1995/10/12  04:25:43  ddavies
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.36  1995/10/07  01:22:01  ddavies
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Include stdio.h since ipwcore.prv no longer includes it (will be true soon).
 *
 * Revision 1.35  1995/10/02  00:18:49  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Filled in standardized routine header comment blocks.
 *
 * Revision 1.34  1995/09/23  00:50:39  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Removed the double comment block for "wp_jCopyLine8To4()"
 *
 * Revision 1.33  1995/09/22  00:38:33  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Fixed up wp_ issues.
 *
 * Revision 1.32  1995/09/21  22:54:21  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.31  1995/09/12  18:29:51  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Fixed PC compiler caught errors/warnings.
 *
 * Revision 1.30  1995/09/12  16:54:30  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.29  1995/09/08  00:14:27  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.28  1995/09/07  01:33:05  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.27  1995/09/06  20:21:09  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.26  1995/09/01  02:20:20  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.25  1995/08/31  21:14:57  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.24  1995/08/30  20:50:59  ddavies
 * ddavies on Wed Aug 30 13:49:44 PDT 1995
 *
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Replace vars of type COLOR_SPACE with defines to get around non-portability
 * of enumerated types.
 *
 * Revision 1.23  1995/08/24  02:53:13  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Made readCallback routine return number of bytes read.
 * Added missing pIPWState argument when calling routine pointed to by
 * cinfo's field cvtImageRow.
 *
 * Revision 1.22  1995/08/22  02:21:23  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.21  1995/08/17  22:00:55  ejaquith
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.20  1994/10/25  20:58:41  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * added complete parameter specs for all callback procedures
 *
 * Revision 1.19  1994/10/11  01:05:18  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Added include of shrpixr.pub to get prototypes for
 * i_getDIBPixrLineProc(). This is only needed on PC.
 *
 * Revision 1.18  1994/10/03  23:29:51  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Change to use of iw_jpeg.pub amd iw_jpeg.prv headers
 * instead of jpegpro.h
 *
 * Revision 1.17  1994/09/08  00:28:26  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Remove Date from RCS macro. Test for redundant types.pub includes.
 *
 * Revision 1.16  1994/08/01  17:33:53  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Fix pixr to memory order copy routine to correctly deal with
 * line lengths not a multiple of 4 pixels.
 *
 * Revision 1.15  1994/07/21  18:42:10  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Add return code debugging abortI() function to trace errors.
 *
 * Revision 1.14  1994/07/21  00:19:29  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Change error handling from old setjmp/longjmp to function return codes.
 *
 * Revision 1.13  1994/07/15  23:57:47  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Added proper arg casts for fread in the callback
 * function to read in JPEG data.
 *
 * Revision 1.12  1994/07/11  18:00:59  lperry
 * lperry on Mon Jul 11 10:57:03 PDT 1994
 *
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Fix 8 to 4 bit conversion code so it won't read entire
 * 32 bit words at the end of an image row. This was causing
 * purify to report an uninitialized memory read when the
 * image width was not a multiple of 8. End of row pixels
 * are now handled one at a time.
 *
 * Revision 1.11  1994/06/03  00:33:05  lperry
 * lperry on Thu Jun 2 17:20:23 PDT 1994
 *
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Fixed bug in decompressing to 4 bit memory order buffers on PC.
 * This needed a separate depth conversion routine.
 * Also corrected behavior when an RGB image was being decompressed
 * to a grayscale image in memory order. Init function now looks
 * at callers memory org argument and forces conversion of RGB to
 * Luminance if its one of the grayscale options. Previous code
 * ended was ending up with the red channel.
 *
 * Revision 1.10  1994/05/26  22:24:19  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Change for Easyscan/Amber. Treat band-interleaved RGB data
 * line by line instead of in a packed buffer. Essentially its
 * like a single band 24 bit image.
 * Also expose all the required MCU row processing interfaces
 * (some had been static in the PIXR-only implementation).
 *
 * Revision 1.9  1994/05/24  22:09:13  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Ifdef'ed a couple of functions used only on PC to
 * mollify gcc.
 *
 * Revision 1.8  1994/05/24  00:51:03  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Wrappers for new JPEG interface. Supports multiple user
 * data organizations (interleaved, pixr, natural memory order
 * and 8 or 4 bpp). Ipshared now has deals with only memory
 * order images. Ipwrap provides the conversion layer.
 *
 * Revision 1.7  1994/05/05  01:42:43  ddavies
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Mess with tabs.  Initialize oword in two procs to keep gcc happy.
 *
 * Revision 1.6  1994/04/13  01:30:53  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Add modifications to use new color pixrs.
 * Call syntax is unchanged, since caller needs to
 * tell JPEG what kind of output is wanted - gray, color, 4 or 8 bit.
 *
 * Revision 1.5  1994/04/08  00:48:23  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Added functions for
 *   8 bit buffer to 4 bit pixr conversions line by line
 *   Decompression to 4 bit gray and 4 bit RGB pixrs
 *   Compressed file querying for size and # of components
 * Added code to deal with new buffer pointer method of
 * accessing external buffers
 *
 * Revision 1.4  1994/03/24  20:53:52  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Changed interface to iw_jdInit to allow a procedure
 * to be specified for setting non-standard options
 * The initial usage is for RGB to gray decompression,
 * but it will be the general interface for any special changes
 * Also substituted UCHAR_ACCESS memory r/w for for the
 * previous byte swapping of the whole pixr.
 *
 * Revision 1.3  1994/03/05  03:05:44  lperry
 * lperry on Fri Mar 4 19:04:28 PST 1994
 *
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Added routines iw_jDecompressToGrayPixr and iw_jDecompressToRGBPixr
 * which decode a JPEG JFIF file  to a pixr or pixrs
 *
 * Revision 1.2  1994/03/02  02:08:53  lperry
 * iw_djpeg.c -> /product/ipcore/ipwrap/src/RCS/iw_djpeg.c,v
 *
 * Added wrappers for functions to query the number of lines
 * in an MCU (minimum compressible unit) and to retrieve the
 * userData pointer, used as a hook to pass FILE stream ptrs
 * or buffer addresses
 *
 * Revision 1.1  1994/02/27  02:34:07  lperry
 * Initial revision
 *
*/
