/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * iw_scale.c - Wrappers for IPCore's plethora of PIXR scaling
 *              and depth conversion functions.
 *
 * Functions :
 *
 *    w_scale ()
 *    w_depthConvertGray ()
 *    w_rescaleOnLadder ()
 *    w_isOnRescaleLadder ()
 *    wp_getLadderValue ()
 *
 */

#include "ipwcore.prv"
#include "shrscgr.pub"   /* for filter junk in reduce.pub */
#include "geomadt.h"
#include "pixr.h"
#include "alplib.h"
#include "appsinc.h"
#include "defines.pub"
#include "frames.pub"
#include "thrshold.h"
#include "logicops.h"    /* for SAMPLE */

/* prototypes */
#include "props.pub"
#include "expand.pub"
#include "reduce.pub"
#include "iw_scale.pub"
#include "floyd.pub"
#include "prxform.pub"
#include "subsmpl.pub"
#include "grbinop.pub"
#include "resconv.pub"
#include "rescale.pub"
#include "bctable.pub"
#include "thrshold.pub"

/* ladder sizes for each depth--used for table access */
#define LADDER_SIZE_1      9
#define LADDER_SIZE_2      5
#define LADDER_SIZE_4     12
#define LADDER_SIZE_8     13


IP_RCSINFO(RCSInfo, "$RCSfile: iw_scale.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Rescale by calling other ipwrap functions 
 *
 * LongDescription:::
 *   Takes an input pixr and performs the
 *   optimal sequence of IPCore operations to produce
 *   an output pixr of the desired size.
 *   Uses the quality parameter to determine how to
 *   perform certain operations.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::        State structure
 *  inPr::             PIXR to convert
 *  outWidth::         Requested width of output PIXR
 *  outHeight::        Requested height of output PIXR 
 *  invertIntensity::  If cTrue, invert image. Note that the "normal"
 *                     black = 1, white = 0 mapping is the default when 
 *                     dealing with binary images.
 *  outQuality::       Requested "quality" level of processing.  This is
 *                     intended to allow the caller to trade quality for speed
 *                     and vice-versa.  There are three valid quality levels:
 *                     IP_LO_QUALITY : for fastest, but worst looking output;
 *                     IP_MED_QUALITY : fast and reasonable looking output;
 *                     IP_HI_QUALITY : sometimes slow, but best looking output.
 *
 * Output:::
 *  outPr::            Pointer to PIXR pointer for output.  Note that this
 *                     PIXR is created within this function--no pre-allocated
 *                     PIXR pointer should be passed in.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function should not be used as it will be removed after mid-1996
 *  
 * FunctionCategory:::
 *  resolution conversion
 *
 * Keywords:::
 *  resolution conversion
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scale (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               outWidth,
    Int32               outHeight,
    Int32               invertIntensity,
    Int32               outQuality,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32  inWidth,
	   inHeight,
	   inDepth,
	   channel,
	   channels = 0;
    PIXR  *inSub = 0;
    PIXR  *outSub[cMaxColorPlanes];
    Bool   isOnLadder;

    *outPr = NULL;
    for (channel = 0; channel < cMaxColorPlanes; channel++)
	outSub[channel] = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1|cDepth2|cDepth4|cDepth8,
		  cChannels1|cChannels3, cFrame32);


    ERR_PASS (w_pixrGetChannelCount (pIPWState, inPr, &channels));
    ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &inDepth));

    ERR_GEN ((outWidth < 1) || (outHeight < 1), cErrInvalidParm);
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &inWidth));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &inHeight));

    for (channel = 0; channel < channels; channel++)
    {
        outSub[channel] = NULL;

        ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, inPr, channel, &inSub));
	ERR_PASS_CLEANUP (w_isOnRescaleLadder (pIPWState, inWidth, inHeight, 
					       inDepth, outWidth, outHeight, 
					       &isOnLadder));
	if (isOnLadder)
	{
	    ERR_PASS_CLEANUP (w_rescaleOnLadder (pIPWState, inSub, 
	                                         outWidth, outHeight,
						 &(outSub[channel])));
	}
	else
	{
	    switch (inDepth)
	    {
		case 1:
		    ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, inSub, 
						     outWidth, outHeight, 
						     &(outSub[channel])));
		    break;
    
		case 2:
		    ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
		    break;
    
		case 4:
		case 8:
		    if (outQuality == IP_HI_QUALITY)
			ERR_PASS_CLEANUP (w_scaleLinearSize (
			    pIPWState, inSub, outWidth, outHeight, 
			    &(outSub[channel])));

		    else
			ERR_PASS_CLEANUP (w_scaleToSize (
			    pIPWState, inSub, outWidth, outHeight,
			    &(outSub[channel])));
		    break;

		default:
		    ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
		    break;
	    }
        }

	if (invertIntensity)
	    ERR_PASS_CLEANUP (w_invert (pIPWState, outSub[channel], 
					&outSub[channel]));

    }   /* scale each channel */

    /* compose the color pixr struct if necessary */
    if (channels > 1)
    {
	ColorType type;
	ERR_PASS_CLEANUP (w_pixrGetType (pIPWState, inPr, &type));
	ERR_PASS_CLEANUP (w_createPixrComposite (pIPWState, outSub, 
						 channels, type, outPr));
    }
    else
    {
	*outPr = outSub[0];
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    for (channel = 0; channel < channels; channel++)
    {
	if (outSub[channel]) 
	    w_destroyPixr(pIPWState, &(outSub[channel]));
    }
    ERR_RETURN;

}  /* w_scale () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Perform depth conversion by calling other ipwrap functions
 *
 * LongDescription:::
 *  Takes an input pixr and produces an output pixr of the same size which
 *  has the requested depth. Uses the quality parameter to determine how to
 *  perform certain operations (most notably, converting gray pixrs to binary).
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::        State structure
 *  inPr::             Input PIXR to convert
 *  outDepth::         Requested depth of output PIXR
 *  invertIntensity::  If cTrue, invert image. Note that the "normal"
 *                     black = 1, white = 0 mapping is the default when 
 *                     dealing with binary images.
 *  outQuality::       Requested "quality" level of processing.  This is
 *                     intended to allow the caller to trade quality for speed
 *                     and vice-versa.  There are three valid quality levels:
 *                     IP_LO_QUALITY : for fastest, but worst looking output;
 *                     IP_MED_QUALITY : fast and reasonable looking output;
 *                     IP_HI_QUALITY : sometimes slow, but best looking output.
 *
 * Output:::
 *  outPr::            Pointer to PIXR pointer for output.  Note that this
 *                     PIXR is created within this function--no pre-allocated
 *                     PIXR pointer should be passed in.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function should not be used as it will be removed after mid-1996.
 *     Low quality output is intended for quick and dirty use.  It uses
 *     thresholding to reduce depth and bit replication to expand depth.
 *     Medium quality output is the same as low quality for this function.
 *     High quality output is intended for use when speed is not as
 *     important as quality.  Depth reduction operations use dithering
 *     when it is available.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_depthConvertGray (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               outDepth,
    Int32               invertIntensity,
    Int32               outQuality,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32  inDepth,
           tmpPrCreated, inWidth, inHeight;
    ScaledInt1616 xDPI, yDPI;
    UInt8 *DRCtable = NULL;
    PIXR  *tmpPr = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepthAny, cChannels1, cFrame32);
    ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &inDepth));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &inWidth));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &inHeight));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xDPI, &yDPI));
    *outPr = NULL;
    tmpPr = NULL;
    tmpPrCreated = cFalse;
    
    switch (outDepth)
    {
        case 8:
	    if (invertIntensity)   
	    /* invert the binary image rather than the 8-bit gray one */
	    {
	        ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, inPr, &tmpPr));
		tmpPrCreated = cTrue;
		ERR_PASS_CLEANUP (w_invert (pIPWState, inPr, &tmpPr));
		inPr = tmpPr;
		tmpPr = NULL;
	    }

	    if (inDepth == 1)
	    {
		ERR_PASS_CLEANUP (w_depth1To8 (pIPWState, inPr, cTrue, &tmpPr));
		*outPr = tmpPr;
	    }
	    else
	    {
		if (inDepth == 4)
		{
		    ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, inPr, &tmpPr));
		    *outPr = tmpPr;
		}
		else
		{
		    ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
		}
	    }

            /* If we previously inverted the input image, free the
	     * pixr we created in the process.
	     */
	    if (tmpPrCreated)
	        ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &inPr));
	    break;

	case 4:
	    if (inDepth == 1)
	    {
	        /* mask or paint from binary to gray */
		ERR_PASS_CLEANUP (w_createPixr (pIPWState, inWidth, inHeight,
					      outDepth, xDPI, yDPI, 
					      cFrameBits, &tmpPr));
		if (invertIntensity)
		    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, tmpPr, inPr, 
						   PIX_SRC | PIX_DST, 
						   &tmpPr));
		else
		    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, tmpPr, inPr, 
						   PIX_SRC & PIX_DST, 
						   &tmpPr));

		*outPr = tmpPr;

	    }
	    else
	    {
		if (inDepth == 8)
		{
		    /* threshold 8->4 */
		    ERR_PASS_CLEANUP (w_scale8To4 (pIPWState, inPr, &tmpPr));

		    if (invertIntensity)
			ERR_PASS_CLEANUP (w_invert (pIPWState, tmpPr, &tmpPr));
		    *outPr = tmpPr;
		}
		else
		    ERR_GEN (1, cErrDepthNotSupported);
	    }
	    break;

	case 2:
	    if (inDepth == 1)
	    {
	        /* mask and paint */
		ERR_PASS_CLEANUP (w_createPixr (pIPWState, inWidth, inHeight, 
					       outDepth, xDPI, yDPI, 
					      cFrameBits, &tmpPr));
		if (invertIntensity)
		    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, tmpPr, inPr, 
						   PIX_SRC | PIX_DST, 
						   &tmpPr));
                else
		    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, tmpPr, inPr, 
						   PIX_SRC & PIX_DST,
						   &tmpPr));
		*outPr = tmpPr;
	    }
	    else
	    {
		if (inDepth == 8)
		{
		    DRCtable = NULL;
		    if (invertIntensity)
			ERR_PASS_CLEANUP (w_makeBrightnessContrastTable (
			    pIPWState, 1.1, 0.0, 256, 256, cTrue, &DRCtable));
		    else
			ERR_PASS_CLEANUP (w_makeBrightnessContrastTable (
			    pIPWState, 1.1, 0.0, 256, 256, cFalse, &DRCtable));

		    ERR_PASS_CLEANUP (w_ditherFloyd2 (pIPWState, inPr, NULL, 
						      DRCtable, &tmpPr));
		    *outPr = tmpPr;
		    IPW_FREE (DRCtable, cAllocatorData);
		}
		else
		    ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
	    }
	    break;

	case 1:  /* to dither or to threshold, that is the question... */
	    if (outQuality == IP_HI_QUALITY)
	    {
		if ((inDepth == 4) || (inDepth == 8))
		{
		    DRCtable = NULL;

		    if (inDepth == 4)
		    {
			if (invertIntensity)
			    ERR_PASS_CLEANUP (w_makeBrightnessContrastTable (
				pIPWState, 1.1, 0.0, 16, 256, cTrue,
				&DRCtable));
			else
			    ERR_PASS_CLEANUP (w_makeBrightnessContrastTable (
				pIPWState, 1.1, 0.0, 16, 256, cFalse, 
				&DRCtable));

			ERR_PASS_CLEANUP (w_ditherFloyd4To1 (
			    pIPWState, inPr, NULL, DRCtable, &tmpPr));
			
		    }
		    else
		    {
			if (invertIntensity)
			    ERR_PASS_CLEANUP (w_makeBrightnessContrastTable (
				pIPWState, 1.1, 0.0, 256, 256, cTrue, 
				&DRCtable));
			else
			    ERR_PASS_CLEANUP (w_makeBrightnessContrastTable (
				pIPWState, 1.1, 0.0, 256, 256, cFalse, 
				&DRCtable));

			ERR_PASS_CLEANUP (w_ditherFloyd (pIPWState, inPr, NULL,
							 DRCtable, &tmpPr));
			
		    }
		    
		    IPW_FREE (DRCtable, cAllocatorData);
		}
		else
		{
		    if (inDepth == 2)  /* Threshold it! */
		    {
			/* remember that gray and binary images have
			 * opposite notions of black and white!
			 */
			if (invertIntensity)
			    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr, 
			                                   2, 3, THRESH_NORM,
							   &tmpPr));
			else
			    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr,
			                                   0, 1, THRESH_NORM,
							   &tmpPr));
		    }
		    else
		        ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
		}
	    }
	    else  /* med or low quality, so threshold for speed */
	    {
	        /* remember that gray and binary images have
		 * opposite notions of black and white!
		 */
		switch (inDepth)
		{
		    case 2:
			if (invertIntensity)
			    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr, 
			                                   2, 3, THRESH_NORM,
							   &tmpPr));
			else
			    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr,
			                                   0, 1, THRESH_NORM,
							   &tmpPr));
		        break;

		    case 4:
			if (invertIntensity)
			    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr, 
			                                   8, 15, THRESH_NORM,
							   &tmpPr));
			else
			    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr,
			                                   0, 7, THRESH_NORM,
							   &tmpPr));
		        break;

		    case 8:
			if (invertIntensity)
			    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr, 
			                                   128, 255, THRESH_NORM,
							   &tmpPr));
			else
			    ERR_PASS_CLEANUP (w_threshold (pIPWState, inPr,
			                                   0, 127, THRESH_NORM,
							   &tmpPr));
		        break;

		    default:
			ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
			break;

		}

	    }

	    *outPr = tmpPr;
	    break;

	default:
	    ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
	    break;

    }  /* switch on output depth */

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &tmpPr);
    if (DRCtable != NULL) 
        IPW_FREE (DRCtable, cAllocatorData);
    ERR_RETURN;

}  /* w_depthConvertGray () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  "on-the-ladder" resolution conversion
 * 
 * LongDescription:::
 *   Takes an input pixr and a width and height for the resulting output 
 *   pixr.  Performs a reduce, expand or resolution conversion operation to
 *   produce the requested output size. The caller should have checked to 
 *   determine whether the desired result was on the "ladder of zoom" 
 *   using w_isOnRescaleLadder () before calling this function. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           Input PIXR 
 *  outWidth::       Requested width of output PIXR
 *  outHeight::      Requested height of output PIXR
 *
 * Output:::
 *  pPrOut::         The resulting rescaled image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function should not be used as it will be removed after mid-1996
 *
 * FunctionCategory:::
 *  resolution conversion
 *
 * Keywords:::
 *  resolution conversion
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rescaleOnLadder (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               outWidth,
    Int32               outHeight,
    PIXR              **pPrOut)
{
    PROLOGUE;
    PIXR   *tmpPr = NULL;
    Int32   inWidth,
	    inHeight,
	    inDepth;
    Bool    isOnLadder;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1|cDepth2|cDepth4|cDepth8,
		  cChannelsAny, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &inWidth));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &inHeight));
    ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &inDepth));

    ERR_PASS (w_isOnRescaleLadder (pIPWState, inWidth, inHeight, 
					    inDepth, outWidth, outHeight,
					    &isOnLadder));
    
    ERR_GEN (!isOnLadder, cErrInvalidParm);

    switch (inDepth)
    {
	case 1:
            if ((inWidth / 16 == outWidth) && (inHeight / 16 == outHeight))
                ERR_PASS (w_reduce (pIPWState, inPr, 2, 2, 2, 2, &tmpPr));
	    else if ((inWidth / 8 == outWidth) && (inHeight / 8 == outHeight))
                ERR_PASS (w_reduce (pIPWState, inPr, 2, 2, 2, 0, &tmpPr));
	    else if ((inWidth / 4 == outWidth) && (inHeight / 4 == outHeight))
                ERR_PASS (w_reduce (pIPWState, inPr, 2, 2, 0, 0, &tmpPr));
	    else if ((inWidth / 2 == outWidth) && (inHeight / 2 == outHeight))
                ERR_PASS (w_reduce (pIPWState, inPr, 2, 0, 0, 0, &tmpPr));
	    else if ((inWidth == outWidth) && (inHeight == outHeight))
                ERR_PASS (w_copy (pIPWState, inPr, &tmpPr));
	    else if ((inWidth * 2 == outWidth) && (inHeight * 2 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 2, &tmpPr));
	    else if ((inWidth * 4 == outWidth) && (inHeight * 4 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 4, &tmpPr));
	    else if ((inWidth * 8 == outWidth) && (inHeight * 8 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 8, &tmpPr));
	    else if ((inWidth * 16 == outWidth) && (inHeight * 16 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 16, &tmpPr));

	    else if ((inWidth == outWidth) && (inHeight * 2 == outHeight))
	        ERR_PASS (w_expandRows (pIPWState, inPr, 2, &tmpPr));
	    else if ((inWidth == outWidth) && (inHeight == outHeight * 2))
	        ERR_PASS (w_reduceRows (pIPWState, inPr, 2, SAMPLE, &tmpPr));
	    else if ((inWidth * 2 == outWidth) && (inHeight == outHeight))
	        ERR_PASS (w_expandCols (pIPWState, inPr, 2, &tmpPr));
	    else if ((inWidth == outWidth * 2) && (inHeight == outHeight))
	        ERR_PASS (w_reduceCols (pIPWState, inPr, 2, SAMPLE, &tmpPr));
	    else
	        ERR_GEN (1, cErrDepthNotSupported);
	    break;

	case 2:
	    if ((inWidth * 2 == outWidth) && (inHeight * 2 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 2, &tmpPr));
	    else if ((inWidth * 4 == outWidth) && (inHeight * 4 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 4, &tmpPr));
	    else if ((inWidth * 8 == outWidth) && (inHeight * 8 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 8, &tmpPr));
	    else if ((inWidth * 16 == outWidth) && (inHeight * 16 == outHeight))
		ERR_PASS (w_expand (pIPWState, inPr, 16, &tmpPr));
	    else
	        ERR_GEN (1, cErrDepthNotSupported);
	    break;

	case 4:
            if ((inWidth / 32 == outWidth) && (inHeight / 32 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 32, &tmpPr));
	    else if ((inWidth / 16 == outWidth) && (inHeight / 16 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 16, &tmpPr));
	    else if ((inWidth / 8 == outWidth) && (inHeight / 8 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 8, &tmpPr));
	    else if ((inWidth / 4 == outWidth) && (inHeight / 4 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 4, &tmpPr));
	    else if ((inWidth / 2 == outWidth) && (inHeight / 2 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 2, &tmpPr));
	    else if (((inWidth * 2) / 3 == outWidth) && 
	             ((inHeight * 2) / 3 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 1, &tmpPr));
	    else if ((inWidth == outWidth) && (inHeight == outHeight))
                ERR_PASS (w_copy (pIPWState, inPr, &tmpPr));
	    else if (((inWidth * 3) / 2  == outWidth) && 
	             ((inHeight * 3) / 2 == outHeight))
                ERR_PASS (w_resConv (pIPWState, inPr, 2, 3, &tmpPr));
	    else if ((inWidth * 2 == outWidth) && (inHeight * 2 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 2, &tmpPr));
	    else if ((inWidth * 4 == outWidth) && (inHeight * 4 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 4, &tmpPr));
	    else if ((inWidth * 8 == outWidth) && (inHeight * 8 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 8, &tmpPr));
	    else if ((inWidth * 16 == outWidth) && (inHeight * 16 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 16, &tmpPr));
	    else
	        ERR_GEN (1, cErrDepthNotSupported);
	    break;

	case 8:
            if ((inWidth / 32 == outWidth) && (inHeight / 32 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 32, &tmpPr));
	    else if ((inWidth / 16 == outWidth) && (inHeight / 16 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 16, &tmpPr));
	    else if ((inWidth / 8 == outWidth) && (inHeight / 8 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 8, &tmpPr));
	    else if ((inWidth / 4 == outWidth) && (inHeight / 4 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 4, &tmpPr));
	    else if ((inWidth / 3 == outWidth) && (inHeight / 3 == outHeight))
                ERR_PASS (w_resConv (pIPWState, inPr, 3, 1, &tmpPr));
	    else if ((inWidth / 2 == outWidth) && (inHeight / 2 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 2, &tmpPr));
	    else if (((inWidth * 2) / 3 == outWidth) && 
	             ((inHeight * 2) / 3 == outHeight))
                ERR_PASS (w_reduceS (pIPWState, inPr, 1, &tmpPr));
	    else if ((inWidth == outWidth) && (inHeight == outHeight))
                ERR_PASS (w_copy (pIPWState, inPr, &tmpPr));
	    else if (((inWidth * 3) / 2  == outWidth) && 
	             ((inHeight * 3) / 2 == outHeight))
                ERR_PASS (w_resConv (pIPWState, inPr, 2, 3, &tmpPr));
	    else if ((inWidth * 2 == outWidth) && (inHeight * 2 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 2, &tmpPr));
	    else if ((inWidth * 4 == outWidth) && (inHeight * 4 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 4, &tmpPr));
	    else if ((inWidth * 8 == outWidth) && (inHeight * 8 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 8, &tmpPr));
	    else if ((inWidth * 16 == outWidth) && (inHeight * 16 == outHeight))
	        ERR_PASS (w_expand (pIPWState, inPr, 16, &tmpPr));
	    else
	        ERR_GEN (1, cErrDepthNotSupported);
	    break;

	default:
	    ERR_GEN (1, cErrDepthNotSupported);
	    break;
    }

    *pPrOut = tmpPr;
    NORMAL_RETURN;

}  /* w_rescaleOnLadder () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Check whether an image can be rescaled using a fast rescaling routine
 *
 * LongDescription:::
 *  Takes input width, height, and depth as well as output width and 
 *  height and determines whether a rescale operation with these parameters 
 *  exists which is one of the very fast reduction/expansion/res conversion 
 *  types.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inWidth::        Input image width
 *  inHeight::       Input image height
 *  inDepth::        Input image depth
 *  outWidth::       Output image width
 *  outHeight::      Output image height
 *
 * Output:::
 *  pIsOnLadder::    If cTrue, the conversion is "on the ladder", and can
 *                   be done quickly
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function should not be used as it will be removed after mid-1996
 *
 * FunctionCategory:::
 *  resolution conversion  
 *
 * Keywords:::
 *  resolution conversion
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_isOnRescaleLadder (
    IPWState           *pIPWState,
    Int32               inWidth,
    Int32               inHeight,
    Int32               inDepth,
    Int32               outWidth,
    Int32               outHeight,
    Bool               *pIsOnLadder)
{
    PROLOGUE;
    Int32   retCode;

    pIPWState = pIPWState;
    switch (inDepth)
    {
	case 1:
            if ((inWidth / 16 == outWidth) && (inHeight / 16 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 8 == outWidth) && (inHeight / 8 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 4 == outWidth) && (inHeight / 4 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 2 == outWidth) && (inHeight / 2 == outHeight))
                retCode = cTrue;
	    else if ((inWidth == outWidth) && (inHeight == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 2 == outWidth) && (inHeight * 2 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 4 == outWidth) && (inHeight * 4 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 8 == outWidth) && (inHeight * 8 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 16 == outWidth) && (inHeight * 16 == outHeight))
                retCode = cTrue;
	    else if ((inWidth == outWidth) && (inHeight * 2 == outHeight))
                retCode = cTrue;
	    else if ((inWidth == outWidth) && (inHeight == outHeight * 2))
                retCode = cTrue;
	    else if ((inWidth * 2 == outWidth) && (inHeight == outHeight))
                retCode = cTrue;
	    else if ((inWidth == outWidth * 2) && (inHeight == outHeight))
                retCode = cTrue;
	    else
	        retCode = cFalse;
	    break;

	case 2:
	    if ((inWidth * 2 == outWidth) && (inHeight * 2 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 4 == outWidth) && (inHeight * 4 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 8 == outWidth) && (inHeight * 8 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 16 == outWidth) && (inHeight * 16 == outHeight))
                retCode = cTrue;
	    else
	        retCode = cFalse;
	    break;

	case 4:
            if ((inWidth / 32 == outWidth) && (inHeight / 32 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 16 == outWidth) && (inHeight / 16 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 8 == outWidth) && (inHeight / 8 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 4 == outWidth) && (inHeight / 4 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 2 == outWidth) && (inHeight / 2 == outHeight))
                retCode = cTrue;
	    else if (((inWidth * 2) / 3 == outWidth) && 
	            ((inHeight * 2) / 3 == outHeight))
                retCode = cTrue;
	    else if ((inWidth == outWidth) && (inHeight == outHeight))
                retCode = cTrue;
	    else if (((inWidth * 3) / 2  == outWidth) && 
	             ((inHeight * 3) / 2 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 2 == outWidth) && (inHeight * 2 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 4 == outWidth) && (inHeight * 4 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 8 == outWidth) && (inHeight * 8 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 16 == outWidth) && (inHeight * 16 == outHeight))
                retCode = cTrue;
	    else
	        retCode = cFalse;
	    break;

	case 8:
            if ((inWidth / 32 == outWidth) && (inHeight / 32 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 16 == outWidth) && (inHeight / 16 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 8 == outWidth) && (inHeight / 8 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 4 == outWidth) && (inHeight / 4 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 3 == outWidth) && (inHeight / 3 == outHeight))
                retCode = cTrue;
	    else if ((inWidth / 2 == outWidth) && (inHeight / 2 == outHeight))
                retCode = cTrue;
	    else if (((inWidth * 2) / 3 == outWidth) && 
	            ((inHeight * 2) / 3 == outHeight))
                retCode = cTrue;
	    else if ((inWidth == outWidth) && (inHeight == outHeight))
                retCode = cTrue;
	    else if (((inWidth * 3) / 2  == outWidth) && 
	             ((inHeight * 3) / 2 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 2 == outWidth) && (inHeight * 2 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 4 == outWidth) && (inHeight * 4 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 8 == outWidth) && (inHeight * 8 == outHeight))
                retCode = cTrue;
	    else if ((inWidth * 16 == outWidth) && (inHeight * 16 == outHeight))
                retCode = cTrue;
	    else
	        retCode = cFalse;
	    break;

	default:
	    retCode = cFalse;
	    break;
    }

    *pIsOnLadder = retCode;
    NORMAL_RETURN;

}  /* w_isOnRescaleLadder () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get ladder value (scale factor) for closest fast resolution conversion
 *
 * LongDescription:::
 *  Takes input width, height, and depth and returns 
 *  the next larger or smaller scale factor which would
 *  use one of the very fast reduction/expansion/res 
 *  conversion types.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  larger::         Boolean flag used to specify whether to get the next
 *                   larger (cTrue) or smaller (cFalse) ladder point.
 *  depth::          depth of input PIXR (or alternately the output PIXR if
 *                   a depth conversion is to be performed).
 *  realWidth::      The width of the "real" and original source PIXR.
 *  realHeight::     The height of the "real" and original source PIXR.
 *  currentWidth::   The width of the current scaled copy of the original
 *                   PIXR.
 *  currentHeight::  The height of the current scaled copy of the original
 *                   PIXR.
 *
 * Output:::
 *  resultWidth::    The (output) width corresponding the the next adjacent
 *                   ladder point.
 *  resultHeight::   The (output) height corresponding the the next adjacent
 *                   ladder point.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function should not be used as it will be removed after mid-1996
 *
 * FunctionCategory:::
 *  resolution conversion
 *
 * Keywords:::
 *  resolution conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getLadderValue (
    IPWState           *pIPWState,
    Int32               larger,
    Int32               depth,
    Int32               realWidth,
    Int32               realHeight,
    Int32               currentWidth,
    Int32               currentHeight,
    Int32              *resultWidth,
    Int32              *resultHeight)
{
    PROLOGUE;
    Int32  lsize, i;
    static Float32 ladder1[] = {0.06250F, 0.12500F, 0.25000F, 0.50000F,
                                1.0000F, 2.0000F, 4.0000F, 8.0000F, 16.000F};
    static Float32 ladder2[] = {1.0000F, 2.0000F, 4.0000F, 8.0000F, 16.000F};
    static Float32 ladder4[] = {0.031250F, 0.062500F, 0.12500F, 0.25000F, 
				0.50000F,  0.66667F, 1.0000F, 1.5000F, 
				2.0000F, 4.0000F, 8.0000F, 16.000F};
    static Float32 ladder8[] = {0.03125F, 0.062500F, 0.12500F, 0.25000F, 
				0.33334F, 0.50000F, 0.66667F, 1.0000F, 
				1.5000F, 2.0000F, 4.0000F, 8.0000F, 16.000F};
    Float32 *pl;
    Float32  inScale, outScale;

    *resultWidth = 0;
    *resultHeight = 0;
    pl = NULL;

    ERR_GEN (((realWidth <= 0) || (realHeight <= 0) || 
        (currentWidth <= 0) || (currentHeight <= 0)), cErrInvalidParm);


    /* should check that in:out width and height ratios are very close
     * in value!
     */

    inScale = (Float32)currentWidth / (Float32)realWidth;
    outScale = 1.0F;

    switch (depth)
    {
	case 1:
	    lsize = LADDER_SIZE_1;
	    pl = ladder1;  
	    break;

	case 2:
	    lsize = LADDER_SIZE_2;
	    pl = ladder2;  
	    break;

	case 4:
	    lsize = LADDER_SIZE_4;
	    pl = ladder4;  
	    break;

	case 8:
	    lsize = LADDER_SIZE_8;
	    pl = ladder8;  
	    break;

	default:
            lsize = 0;
	    break;
    }

    if (lsize)
    {
        if (larger)
	{
	    for (i = 0; i < lsize; i++)
	    {
	        if (pl[i] > inScale)
		{
		    outScale = pl[i];
		    break;
		}
	    }

	    /* if we went off the end, just return the same w & h */
	    if (i == lsize)
	        outScale = pl[i - 1];
	}
	else
	{
	    for (i = lsize - 1; i >= 0; i--)
	    {
	        if (pl[i] < inScale)
		{
		    outScale = pl[i];
		    break;
		}
	    }
	    /* if we went off the end, just return the same w & h */
	    if (i < 0)
	        outScale = pl[0];

	}

	*resultWidth = (Int32)((Float32)realWidth * outScale);
	*resultHeight = (Int32)((Float32)realHeight * outScale);
    }
    else
        ERR_GEN (1, cErrDepthNotSupported);

    NORMAL_RETURN;

}  /* wp_getLadderValue () */



/*
$Log: iw_scale.c,v $
Revision 1.1.1.1  1999/09/20 18:19:18  jlin
Imported sources

Revision 1.32  1997/04/15 22:13:41  ddavies
iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v

Use non-initialized PIXR creation where possible.

Revision 1.31  1996/01/04 20:24:41  ejaquith
iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.30  1995/12/13  04:10:15  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.29  1995/11/15  16:37:36  dferrell
 * dferrell on Wed Nov 15 08:36:35 PST 1995
 *
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Fixed typo in a **comment** which caused our (nazi) scripts to complain.
 *
 * Revision 1.28  1995/11/13  20:13:34  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.27  1995/11/10  02:03:55  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.26  1995/11/08  19:55:11  dferrell
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Removed unused functions.
 * Fixed several bugs (thresholding, not assigning the output PIXR
 * properly, etc.)
 * Added use of fast anamorphic scaling when applicable.
 *
 * Revision 1.25  1995/11/01  23:58:19  dferrell
 * dferrell on Wed Nov 1 15:56:58 PST 1995
 *
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Fixed bug in w_depthConvertGray () when expanding from binary to
 * 8-bit gray.  The output PIXR was not being set properly before the
 * function returned.
 *
 * Revision 1.24  1995/11/01  22:17:56  ddavies
 * ddavies on Wed Nov 1 14:16:54 PST 1995
 *
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Change calls to w_createPixrComposite so that it takes the array outSub and
 * the number of channels instead of always taking 3 PIXR pointers.  This
 * allows us to operate on arbitrary multi-plane images.
 *
 * Revision 1.23  1995/10/16  23:26:31  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Added fwd declarations for static routines:
 * wp_doDepthExpansion () and wp_doDepthReduction (
 * ).
 *
 * Revision 1.22  1995/10/13  22:02:49  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.21  1995/10/12  17:44:15  macleod
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * added explicit casts (float->int, long->short, long->char, etc) and
 * made floating constants the right size to shut up the NT compiler.
 *
 * Revision 1.20  1995/10/05  19:16:31  macleod
 * macleod on Thu Oct 5 12:15:50 PDT 1995
 *
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Added some fields which were missing in function headers,
 * and there were a couple of cases where all of the function maturity
 * types were still there, etc.
 *
 * Revision 1.19  1995/10/03  01:59:23  macleod
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.18  1995/10/02  19:18:55  jcraig
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * w_invert takes a pointer to a pointer to a destination pixr.
 *
 * Revision 1.17  1995/09/29  20:55:16  ddavies
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Change calls to w_opGrayBinIP to equivalent calls to w_opGrayBin.
 *
 * Revision 1.16  1995/09/29  01:41:33  ejaquith
 * ejaquith on Thu Sep 28 18:41:08 PDT 1995
 *
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * changed TRUE to cTrue and FALSE to cFalse.
 * updated invocations of w_resConv().
 *
 * Revision 1.15  1995/09/21  22:54:25  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.14  1995/09/19  19:19:29  macleod
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.13  1995/09/11  21:54:21  macleod
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.12  1995/09/07  01:33:08  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:21:12  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  21:02:40  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.9  1995/08/31  21:15:00  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/25  01:37:09  ejaquith
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Fixed warnings
 *
 * Revision 1.7  1995/08/25  01:08:48  jcraig
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * get rid of pc compiler warnings
 *
 * Revision 1.6  1995/08/17  21:55:29  macleod
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.5  1995/04/01  01:43:10  ddavies
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Remove the factor 2/3 from the ladder for binary images.  We do not have
 * a fast 1.5x reduction routine for binary images.
 *
 * Revision 1.4  1994/10/10  22:40:37  lperry
 * lperry on Mon Oct 10 15:40:05 PDT 1994
 *
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Remove use of pcpros.h
 *
 * Revision 1.3  1994/10/07  19:06:42  lperry
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.2  1994/09/16  17:36:16  jcraig
 * iw_scale.c -> /product/ipcore/ipwrap/src/RCS/iw_scale.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.1  1994/07/12  00:36:14  dferrell
 * Initial revision
 *

*/

