/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* Layout functions
 *
 * This module provides all functions related to the layout of data,
 * independent of encoding and rendering information. The library
 * entry points allow the client to specify a number of constraints
 * on glyph layout, and then solve for that layout. Most of these
 * constraints are optional, and the system will use defaults for
 * those not specified.
 *
 * The simplest case for layout is where the client specifies only
 * the amount of data to be encoded, and no constraints on layout.
 * In this case, the library will generate a rectangle of arbitrary
 * shape large enough to contain the data. The most complex case is
 * where an occlusion is specified in addition to other constraints.
 * The placement of an occlusion alone adds significant complexity
 * to the problem.
 *
 * Library entry points in this module:
 *
 *     None
 *
 * Internal entry points in this module:
 *
 *     wp_doComputeCapacity()
 *     wp_doComputeLayout()
 *     wp_doCreateLayout()
 *     wp_doDestroyLayout()
 *     wp_doSetAspectRatio()
 *     wp_doSetDataSize()
 *     wp_doSetDimensions()
 *     wp_doSetECCLevel()
 *     wp_doSetExpansion()
 *     wp_doSetGlyphSize()
 *     wp_doSetOcclusion()
 *
 *
 * The functions in this module are organized alphabetically for easy
 * reference. Local (to this module) functions are defined first, then
 * internal (to the library) functions, then library entry points. Any
 * functions which are referenced before their definition within this
 * module have prototypes defined under "Forward References" below.
 *
 */

#define	DATAGLYPH_VERSION1

#include <math.h>


#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "ipwerrs.pub"
#include "dgencdef.pub"


#include "shrutils.pub"
#include "dgcom.prv"
#include "dgencint.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif



IP_RCSINFO(RCSInfo, "$RCSfile: layout.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

/************************* Forward declarations ***********************/

static UInt32 CDECL 
wp_getFramesOccluded (
    IPWState   *pIPWState,
    UInt32	nRows,
    UInt32	nCols,
    LayoutRect *fRect);


/************************** Local functions ***************************/

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Check the occlusion for size constraints.
 *
 * LongDescription:::
 * Any occlusion must meet 2 size related constraints.  First, the 
 * occlusion must not cut the rectangle into 2 separate regsions.  
 * Seconds, the occlusion must be at least MIN_OCCL_FRAMES frames wide.
 * Check to make sure that these constraints are being met.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  nRows::		number of rows of frames
 *  nCols::		number of columns of frames
 *  frameWidth::	width of each frame in glyphs
 *  frameHeight::	height of each frame in glyphs
 *  fOcclusion::	occlusion info in terms of frames
 *  gOcclusion::	occlusion info in terms of glyphs
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL 
wp_checkOcclusion (
    IPWState   *pIPWState,
    UInt32	nRows,
    UInt32	nCols,
    UInt32	frameWidth,
    UInt32	frameHeight,
    LayoutRect *fOcclusion,
    LayoutRect *gOcclusion)
{
    PROLOGUE;

    /* check to make sure that any occlusion does not cut the rectangle
     * into 2 separate regions.
     */
    ERR_GEN((fOcclusion->w == nCols) || (fOcclusion->h == nRows),
	    cErrDG_OcclusionSpansRect);

    /* is there an occlusion and is it large enough?  each occlusion
     * must be MIN_OCCL_FRAMES wide.  each frame is frameWidth+1 glyphs
     * wide and then in addition we must add another glyph for the
     * final synch.
     */
    ERR_GEN((gOcclusion->w < (MIN_OCCL_FRAMES*(frameWidth+1) + 1)) &&
	    (gOcclusion->h < (MIN_OCCL_FRAMES*(frameHeight+1) + 1)), 
	    cErrDG_OcclusionTooSmall);

    NORMAL_RETURN;
}

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Compute the occlusion rectangle in terms of frames.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  frameWidth::     	width of frame in glyphs
 *  frameHeight::    	height of frame in glyphs
 *  glyphRect::      	position and size of occlusion in glyphs
 *
 * Output:::
 *  frameRect::      	position and size of occlusion in frames
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 **************************************************************************/
static void CDECL 
wp_computeFrameRect (
    IPWState   *pIPWState,
    UInt32	frameWidth,
    UInt32	frameHeight,
    LayoutRect *glyphRect,
    LayoutRect *frameRect)
{
    UInt32	 firstBadFrameRow, firstBadFrameCol;
    UInt32	 firstGoodFrameRow, firstGoodFrameCol;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    /* find out which frames are bad.  besides enough space for
     * the synch around the occlusion lets also leave another 1 glyph
     * border just to give a good a chance as possible to read the synch.
     * hence, the subtraction of 2.  watch out for the UInts going negative.
     * as a result  we make sure the subtraction gets a 0 if the x or y
     * are 0 or 1.  this used to cast to an integer but cc on solaris
     * was having problem with this line....
     */
    firstBadFrameRow = (MAX(glyphRect->y, 2) - 2) / (frameHeight + 1);
    firstBadFrameCol = (MAX(glyphRect->x, 2) - 2) / (frameWidth + 1);

    /* the first good frame is computed by finding the last bad glyph
     * (firstBadYGlyph + glyphHeight - 1) adding 2 (one for the synch
     * bit and one glyph of pad) rounding (frameHeight + 1 - 1) and 
     * dividing by the frame size (frameHeight + 1)
     */
    firstGoodFrameRow = (glyphRect->y + glyphRect->h + 1 + frameHeight) / 
		        (frameHeight + 1);
    firstGoodFrameCol = (glyphRect->x + glyphRect->w + 1 + frameWidth) / 
		        (frameWidth + 1);

    frameRect->x = firstBadFrameCol;
    frameRect->y = firstBadFrameRow;
    frameRect->w = firstGoodFrameCol - firstBadFrameCol;
    frameRect->h = firstGoodFrameRow - firstBadFrameRow;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Compute the min and max frames per row and col and total frames required
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 * This takes a lot of parameters and pointers to varibles to be filled 
 * in.  I would rather avoid it but these lines get called several times.
 *
 * Input:::
 *  pIPWState::			pointer to IPCore state
 *  frameWidth::       		frame width in glyphs
 *  frameHeight::      		frame height in glyphs
 *  nMinGlyphsPerRow:: 		min number of glyphs per row
 *  nMaxGlyphsPerRow:: 		max number of glyphs per row
 *  nMinGlyphsPerCol:: 		min number of glyphs per col
 *  nMaxGlyphsPerCol:: 		max number of glyphs per col
 *  nEncodingBytes::   		number of encoded bytes
 *
 * Output:::
 *  nMinFramesPerRow:: 		min number of frames per row
 *  nMaxFramesPerRow:: 		max number of frames per row
 *  nMinFramesPerCol:: 		min number of frames per col
 *  nMaxFramesPerCol:: 		max number of frames per col
 *  nFrameBytes::      		number of bytes in each frame
 *  nTotalFrames::     		total number of frames required to
 *				hold the data.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_computeMinMaxFrames (
    IPWState   *pIPWState,
    UInt32	frameWidth,
    UInt32	frameHeight,
    UInt32	nMinGlyphsPerRow,
    UInt32	nMaxGlyphsPerRow,
    UInt32	nMinGlyphsPerCol,
    UInt32	nMaxGlyphsPerCol,
    UInt32	nEncodingBytes,
    UInt32     *nMinFramesPerRow,
    UInt32     *nMaxFramesPerRow,
    UInt32     *nMinFramesPerCol,
    UInt32     *nMaxFramesPerCol,
    UInt32     *nFrameBytes,
    UInt32     *nTotalFrames)
{
    /* Keep compiler happy */
    pIPWState = pIPWState;

    /* Compute min and max dimensions, in frames */
    *nMinFramesPerRow = GLYPHS_TO_FRAMES(nMinGlyphsPerRow, frameWidth);
    *nMaxFramesPerRow = GLYPHS_TO_FRAMES(nMaxGlyphsPerRow, frameWidth);
    *nMinFramesPerCol = GLYPHS_TO_FRAMES(nMinGlyphsPerCol, frameHeight);
    *nMaxFramesPerCol = GLYPHS_TO_FRAMES(nMaxGlyphsPerCol, frameHeight);

    *nFrameBytes = frameWidth * frameHeight / 8;
    *nTotalFrames = DIVUP(nEncodingBytes, *nFrameBytes);
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * compute the number of rows and cols available given the aspect ratio
 *
 * Input:::
 *  pIPWState::			pointer to IPCore state
 *  frameWidth::      		width of each frame in glyphs
 *  frameHeight::     		height of each frame in glyphs
 *  nMinHorz::        		min mica width of each rectangle
 *  nMaxHorz::        		max mica width of each rectangle
 *  nMinVert::        		min mica height of each rectangle
 *  nMaxVert::        		max mica height of each rectangle
 *  nMinGlyphsPerRow:: 		min number of glyphs per row
 *  nMaxGlyphsPerRow:: 		max number of glyphs per row
 *  nMinGlyphsPerCol:: 		min number of glyphs per col
 *  nMaxGlyphsPerCol:: 		max number of glyphs per col
 *  nMinFramesPerRow:: 		min number of frames per row
 *  nMaxFramesPerRow:: 		max number of frames per row
 *  nMinFramesPerCol:: 		min number of frames per col
 *  nMaxFramesPerCol:: 		max number of frames per col
 *  horzAspect::      		horizontal component of aspect ratio
 *  vertAspect::      		vertical component of aspect ratio
 *
 * Output:::
 *  nRows::           		number of rows needed to contain data
 *  nCols::           		number of cols needed to contain data
 *  nPadHorz::        		number of pad glyphs in each row
 *  nPadVert::        		number of pad glyph in each col
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_computeRowCol (
    IPWState   *pIPWState,
    UInt32	frameWidth,
    UInt32	frameHeight,
    UInt32	nMinHorz,
    UInt32	nMaxHorz,
    UInt32	nMinVert,
    UInt32	nMaxVert,
    UInt32	nMinGlyphsPerRow,
    UInt32	nMaxGlyphsPerRow,
    UInt32	nMinGlyphsPerCol,
    UInt32	nMaxGlyphsPerCol,
    UInt32	nMinFramesPerRow,
    UInt32	nMaxFramesPerRow,
    UInt32	nMinFramesPerCol,
    UInt32	nMaxFramesPerCol,
    UInt32	horzAspect,
    UInt32	vertAspect,
    UInt32     *nRows,
    UInt32     *nCols,
    UInt32     *nPadHorz,
    UInt32     *nPadVert)
{

    /* Keep compiler happy */
    pIPWState = pIPWState;

    if (nMaxFramesPerRow == 0)
    {
	nMaxGlyphsPerRow = frameWidth + 2;
	nMaxFramesPerRow = 1;
	nMaxGlyphsPerCol = (UInt32)((Float64)vertAspect / 
	                            (Float64)horzAspect * nMaxGlyphsPerRow);
	nMaxFramesPerCol = GLYPHS_TO_FRAMES(nMaxGlyphsPerCol, frameHeight);
    }
    if (nMaxFramesPerCol == 0)
    {
	nMaxGlyphsPerCol = frameHeight + 2;
	nMaxFramesPerCol = 1;
	nMaxGlyphsPerRow = (UInt32)((Float64)horzAspect / 
	                            (Float64)vertAspect * nMaxGlyphsPerRow);
	nMaxFramesPerRow = GLYPHS_TO_FRAMES(nMaxGlyphsPerCol, frameWidth);
    }

    if (nMinHorz == nMaxHorz && nMinVert == nMaxVert) 
    {
	*nCols = nMaxFramesPerRow;
	*nPadHorz =  
	    nMaxGlyphsPerRow -
	    FRAMES_TO_GLYPHS(nMaxFramesPerRow, frameWidth);
	*nRows = nMaxFramesPerCol;
	*nPadVert = 
	    nMaxGlyphsPerCol -	
	    FRAMES_TO_GLYPHS(nMaxFramesPerCol, frameHeight);
    }
    else if (nMinHorz == nMaxHorz) 
    {
    /* If the horizontal dimension is specified exactly, then maintain
     * the aspect ratio in the vertical dimension.
     */
    UInt32 nGlyphsPerCol;
    
	*nCols = nMaxFramesPerRow;
	*nPadHorz = 
	    nMaxGlyphsPerRow -
	    FRAMES_TO_GLYPHS(nMaxFramesPerRow, frameWidth);
	nGlyphsPerCol =
	    (UInt32)((Float64)nMaxGlyphsPerRow *
		     (Float64)vertAspect / (Float64)horzAspect);
	*nRows = GLYPHS_TO_FRAMES(nGlyphsPerCol, frameHeight);
	*nPadVert =
	    nGlyphsPerCol - FRAMES_TO_GLYPHS(*nRows, frameHeight);
    }
    else if (nMinVert == nMaxVert) 
    {
    /* If the vertical dimension is specified exactly, then maintain
     * the aspect ratio in the horizontal dimension.
     */
    UInt32 nGlyphsPerRow;
    
	*nRows = nMaxFramesPerCol;
	*nPadVert = 
	    nMaxGlyphsPerCol -
	    FRAMES_TO_GLYPHS(nMaxFramesPerCol, frameHeight);
	nGlyphsPerRow =
	    (UInt32)((Float64)nMaxGlyphsPerCol *
		     (Float64)horzAspect / (Float64)vertAspect);
	*nCols = GLYPHS_TO_FRAMES(nGlyphsPerRow, frameWidth);
	*nPadHorz = 
	    nGlyphsPerRow - FRAMES_TO_GLYPHS(*nCols, frameWidth);
    }
    else 
    {
    /* we have made sure that the minimum meets the aspect
     * ratio so use the minimum as a default.
     */
	*nRows = nMinFramesPerCol;
	*nPadVert = 
	    nMinGlyphsPerCol -
	    FRAMES_TO_GLYPHS(nMinFramesPerCol, frameHeight);
	*nCols = nMinFramesPerRow;
	*nPadHorz = 
	    nMinGlyphsPerRow -
	    FRAMES_TO_GLYPHS(nMinFramesPerRow, frameWidth);
    }
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * calculate the rectangle dimensions.
 *
 * LongDescription:::
 * Given the aspect ratio and the number of frames necessary to encode
 * the data, calculate the rectangle dimensions.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  horzAspect::     	horizontal component of aspect ratio
 *  vertAspect::     	vertical component of aspect ratio
 *  nTotalFrames::   	number of frames needed to hold data
 *  frameWidth::     	width of each frame in glyphs
 *  frameHeight::    	height of each frame in glyphs
 *
 * Output:::
 *  fRect::          	occlusion rectangle (in frames)
 *  nPadHorz::       	number of pad glyphs in each row
 *  nPadVert::       	number of pad glyphs in each col
 *  nRows::          	number of rows of frames
 *  nCols::          	number of cols of frames
 *
 * Notes:::
 * There will be some number of frames in the x and y directions and 
 * probably some pad glyphs less than a frame in size to maintain the ratio.
 * This routine is only called when the rectangle computed from the
 * min/max dimensions is insufficient.
 * fRect is modified to conform to the dimensions of the rectangle in
 * case it has been specified as being much larger.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_dimensionsFromAspectRatio (
    IPWState   *pIPWState,
    UInt32	horzAspect,
    UInt32	vertAspect,
    UInt32	nTotalFrames,
    UInt32	frameWidth,
    UInt32	frameHeight,
    LayoutRect *fRect,
    UInt32     *nPadHorz,
    UInt32     *nPadVert,
    UInt32     *nRows,
    UInt32     *nCols)
{
    Bool	 addToHorzFirst;
    UInt32	 framesAvailable;
    UInt32	 glyphsNeeded;

    framesAvailable = (*nRows * *nCols) - 
		      ERR_IGNORE(wp_getFramesOccluded(pIPWState,
						      *nRows, *nCols, fRect));

    /* in order to stay under the max dimensions as well as possible
     * add frames to the direction that will require the smallest
     * number of frames to be added in the other dimension.
     */
    addToHorzFirst = ((Float64)horzAspect / (Float64)vertAspect) > 1.0;

    while (framesAvailable < nTotalFrames)
    {
	if (addToHorzFirst)
	{
	    /* then add to the horizontal first */ 
	    *nCols += 1;
	    *nPadHorz = 0;
	    glyphsNeeded = (UInt32)(FRAMES_TO_GLYPHS(*nCols, frameWidth) *
			           (Float64)vertAspect / (Float64)horzAspect);
	    *nRows = GLYPHS_TO_FRAMES(glyphsNeeded, frameHeight);
	    *nPadVert = glyphsNeeded - 
			FRAMES_TO_GLYPHS(*nRows, frameHeight);
	}
	else
	{
	    /* add to the vertical first */
	    *nRows += 1;
	    *nPadVert = 0;
	    glyphsNeeded = (UInt32)(FRAMES_TO_GLYPHS(*nRows, frameHeight) *
			           (Float64)horzAspect / (Float64)vertAspect);
	    *nCols = GLYPHS_TO_FRAMES(glyphsNeeded, frameWidth);
	    *nPadHorz = glyphsNeeded - 
			FRAMES_TO_GLYPHS(*nCols, frameWidth);
	}

   	/* now how many frames are available? */
        framesAvailable = (*nRows * *nCols) - 
		          ERR_IGNORE(wp_getFramesOccluded(pIPWState,
							  *nRows, *nCols,
							  fRect));
    }

    /* we require at least one row of good frames either above or 
     * below and on one side of the occlusion.
     */
    if ((fRect->y == 0) && (fRect->h >= *nRows))
    {
	*nRows += fRect->h - *nRows + 1;
	*nPadVert = 0;
	glyphsNeeded = (UInt32)(FRAMES_TO_GLYPHS(*nRows, frameHeight) *
		               (Float64)horzAspect / (Float64)vertAspect);
	*nCols = GLYPHS_TO_FRAMES(glyphsNeeded, frameWidth);
	*nPadHorz = glyphsNeeded - 
		    FRAMES_TO_GLYPHS(*nCols, frameWidth);
    }
    if ((fRect->x == 0) && (fRect->w >= *nCols))
    {
	*nCols += fRect->w - *nCols + 1;
	*nPadHorz = 0;
	glyphsNeeded = (UInt32)(FRAMES_TO_GLYPHS(*nCols, frameWidth) *
		               (Float64)vertAspect / (Float64)horzAspect);
	*nRows = GLYPHS_TO_FRAMES(glyphsNeeded, frameHeight);
	*nPadVert = glyphsNeeded - 
		    FRAMES_TO_GLYPHS(*nRows, frameHeight);
    }
    /* if the occlusion would extend over the glyph rectangle, modify
     * the size to reach the edge of the rectangle.
     * if the occluion is outside the rectangle, leave it alone.
     */
    if ((fRect->x < *nCols) && ((fRect->x + fRect->w) > *nCols))
	fRect->w = *nCols - fRect->x;
    if ((fRect->y < *nRows) && ((fRect->y + fRect->h) > *nRows))
        fRect->h = *nRows - fRect->y;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Find the number of rows and columns need for horizontal first expansion
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  nTotalFrames::   		number of frames needed to hold data
 *  nMaxGlyphsPerRow:: 		max number of glyphs per row
 *  nMinGlyphsPerCol:: 		min number of glyphs per col
 *  nMaxFramesPerRow:: 		max number of frames per row
 *  nMinFramesPerCol:: 		max number of frames per col
 *  frameWidth::     		width of each frame in glyphs
 *  frameHeight::    		height of each frame in glyphs
 *
 * Output:::
 *  frameRect::          	occlusion rectangle (in frames)
 *  nRows::          		number of rows of frames
 *  nCols::          		number of cols of frames
 *  nPadHorz::       		number of pad glyphs in each row
 *  nPadVert::       		number of pad glyphs in each col
 *
 * Notes:::
 * Modifies frameRect if the original computed frames extend beyond
 * the edge of the rectangle.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_doHorizExpansion (
    IPWState   *pIPWState,
    UInt32	nTotalFrames,
    UInt32	nMaxGlyphsPerRow,
    UInt32	nMinGlyphsPerCol,
    UInt32	nMaxFramesPerRow,
    UInt32	nMinFramesPerCol,
    UInt32	frameWidth,
    UInt32	frameHeight,
    LayoutRect *frameRect,
    UInt32     *nCols,
    UInt32     *nRows,
    UInt32     *nPadHorz,
    UInt32     *nPadVert)
{
    UInt32	 currRow;
    UInt32	 framesPlaced;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    if (nMaxFramesPerRow == 0)
    {
	nMaxGlyphsPerRow = frameWidth + 2;
	nMaxFramesPerRow = 1;
    }

    for (currRow = 0, framesPlaced = 0; framesPlaced < nTotalFrames; currRow++)
    {
	/* determine the number of frames in this row */
    	if ((currRow >= frameRect->y) && 
	    (currRow < (frameRect->y + frameRect->h)))
    	{
	    if (frameRect->x >= nMaxFramesPerRow)
		/* then the occlusion starts outside the region */
		framesPlaced += nMaxFramesPerRow;
	    /* else we are in a row with an occlusion */
	    else 
		/* add the amount of the left of the occlusion, if any,
		 * and the amount to the right of the occlusion, if any.
		 */
		framesPlaced += frameRect->x +
				(nMaxFramesPerRow - 
				 MIN(nMaxFramesPerRow, 
				     frameRect->x + frameRect->w));
#if 0
	    framesPlaced += nMaxFramesPerRow - frameRect->w;
#endif
    	}
    	else	/* there is no occlusion in this row */
	    framesPlaced += nMaxFramesPerRow;
    }

    *nCols = nMaxFramesPerRow;
    *nRows = MAX(currRow, nMinFramesPerCol);
    /* we require at least one row of good frames either above or 
     * below and on one side of the occlusion.
     */
    if ((frameRect->y == 0) && (frameRect->h >= *nRows))
	*nRows += frameRect->h - *nRows + 1;
    /* since horizontal is important to client do not add columns.  if the
     * occlusion spans the rectangle an error code will be returned
     * later.
     */
#if 0
    if ((frameRect->x == 0) && (frameRect->w >= *nCols))
    {
	*nPadHorz = 0;
	*nCols += frameRect->w - *nCols + 1;
    }
    else 
#endif
        *nPadHorz = nMaxGlyphsPerRow -
	            FRAMES_TO_GLYPHS(nMaxFramesPerRow, frameWidth);
    /* If an exact height was specified, and we haven't already
     * overflowed the layout, pad out vertically. we also must check
     * nRows against nMaxFramesPerCol because we don't want to get
     * a negative number of pad glyphs.
     * This used to only compute when nMinVert was equal to nMaxVert
     * but this ignored the case that required pad glyphs to be added
     * to get the minimum.  For example, say that we only need 1 row in
     * order to hold the data but the minimum vertical requires 2 rows
     * and 2 pad glyphs.  We need to add the pad glyphs.
     * Note that by changing the use of nMaxFramesPerCol to 
     * nMinFramesPerCol we still what we want when the min and max are
     * equal.
     */
    if (*nRows == nMinFramesPerCol) 
	*nPadVert = nMinGlyphsPerCol -
	            FRAMES_TO_GLYPHS(nMinFramesPerCol, frameHeight);
    else *nPadVert = 0;

    /* if the occlusion would extend over the glyph rectangle, modify
     * the size to reach the edge of the rectangle.
     * if the occluion is outside the rectangle, leave it alone.
     */
    if ((frameRect->x < *nCols) && ((frameRect->x + frameRect->w) > *nCols))
	frameRect->w = *nCols - frameRect->x;
    if ((frameRect->y < *nRows) && ((frameRect->y + frameRect->h) > *nRows))
        frameRect->h = *nRows - frameRect->y;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Find the number of rows and columns need for vertical first expansion
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  nTotalFrames::   		number of frames needed to hold data
 *  nMinGlyphsPerRow::		min number of glyphs per row
 *  nMaxGlyphsPerCol::		max number of glyphs per col
 *  nMinFramesPerRow::		min number of frame per row
 *  nMaxFramesPerCol::		max number of frames per col
 *  frameWidth::      		width of each frame in glyphs
 *  frameHeight::    		height of each frame in glyphs
 *
 * Output:::
 *  frameRect::          	occlusion rectangle (in frames)
 *  nCols::          		number of cols of frames
 *  nRows::          		number of rows of frames
 *  nPadHorz::       		number of pad glyphs in each row
 *  nPadVert::       		number of pad glyphs in each col
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_doVertExpansion (
    IPWState   *pIPWState,
    UInt32	nTotalFrames,
    UInt32	nMinGlyphsPerRow,
    UInt32	nMaxGlyphsPerCol,
    UInt32	nMinFramesPerRow,
    UInt32	nMaxFramesPerCol,
    UInt32	frameWidth,
    UInt32	frameHeight,
    LayoutRect *frameRect,
    UInt32     *nCols,
    UInt32     *nRows,
    UInt32     *nPadHorz,
    UInt32     *nPadVert)
{
    UInt32 	 currCol;
    UInt32	 framesPlaced;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    if (nMaxFramesPerCol == 0)
    {
	nMaxGlyphsPerCol = frameHeight + 2;
	nMaxFramesPerCol = 1;
    }

    /* determine the minimum number of columns needed to hold the data */
    for (currCol = 0, framesPlaced = 0; framesPlaced < nTotalFrames; currCol++)
    {
	/* determine the number of frames in this col.  note that if we
	 * have no occlusion, then all the frameRect elements will be 0
	 * and so we will never take the then part of the if statement.
	 */
    	if ((currCol >= frameRect->x) && 
	    (currCol < (frameRect->x + frameRect->w)))
    	{
	    if (frameRect->y >= nMaxFramesPerCol)
		/* then the occlusion starts outside the region */
	        framesPlaced += nMaxFramesPerCol;
	    /* else we are in a column with an occlusion */
	    else
		/* add the amount of the top of the occlusion, if any,
		 * and the amount to the bottom of the occlusion, if any.
		 */
		framesPlaced += frameRect->y +
				(nMaxFramesPerCol - 
				 MIN(nMaxFramesPerCol, 
				     frameRect->y + frameRect->h));
    	}
    	else	/* there is no occlusion in this column */
	    framesPlaced += nMaxFramesPerCol;
    }

    *nRows = nMaxFramesPerCol;
    *nCols = MAX(currCol, nMinFramesPerRow);
    /* we require at least one row of good frames either above or 
     * below and on one side of the occlusion.
     */
    if ((frameRect->x == 0) && (frameRect->w >= *nCols))
	*nCols += frameRect->w - *nCols + 1;
    /* since vertical is important to client do not add rows.  if the
     * occlusion spans the rectangle an error code will be returned
     * later.
     */
#if 0
    if ((frameRect->y == 0) && (frameRect->h >= *nRows))
    {
	*nPadVert = 0;
	*nRows += frameRect->h - *nRows + 1;
    }
    else 
#endif
        *nPadVert = nMaxGlyphsPerCol -
	            FRAMES_TO_GLYPHS(nMaxFramesPerCol, frameHeight);
    /* If an exact width was specified, and we haven't already
     * overflowed the layout, pad out horizontally. we don't want to get
     * a negative number of pad glyphs.
     * This used to only compute when nMinHorz was equal to nMaxHorz
     * but this ignored the case that required pad glyphs to be added
     * to get the minimum.  For example, say that we only need 1 col in
     * order to hold the data but the minimum horizontal requires 2 cols
     * and 2 pad glyphs.  We need to add the pad glyphs.
     * Note that by changing the use of nMaxFramesPerRow to 
     * nMinFramesPerRow we still what we want when the min and max are
     * equal.
     */
    if (*nCols == nMinFramesPerRow) 
	*nPadHorz = nMinGlyphsPerRow -
	            FRAMES_TO_GLYPHS(nMinFramesPerRow, frameWidth);
    else *nPadHorz = 0;

    /* if the occlusion would extend over the glyph rectangle, modify
     * the size to reach the edge of the rectangle.
     * if the occluion is outside the rectangle, leave it alone.
     */
    if ((frameRect->x < *nCols) && ((frameRect->x + frameRect->w) > *nCols))
	frameRect->w = *nCols - frameRect->x;
    if ((frameRect->y < *nRows) && ((frameRect->y + frameRect->h) > *nRows))
        frameRect->h = *nRows - frameRect->y;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Determine the number of occluded frames in the given rectangle
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  nRows::			number of rows in the rectangle
 *  nCols::			number of cols in the rectangle
 *  fRect::			occluded rectangle in frames
 *
 * Returns:::
 *  number of occluded frames
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static UInt32 CDECL 
wp_getFramesOccluded (
    IPWState   *pIPWState,
    UInt32	nRows,
    UInt32	nCols,
    LayoutRect *fRect)
{
    UInt32	 colsOccl;
    UInt32	 rowsOccl;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    if (fRect->w > 0)
    {
	/* then we have an occlusion */
	if ((fRect->x + fRect->w) <= nCols)
	    /* then the occlusion ends before the edge of the rect */
	    colsOccl = fRect->w;
	else if (fRect->x >= nCols)
	    /* the occlusion is not in the region */
	    colsOccl = 0;
	/* else the occlusion is partially in the region */
	else colsOccl = nCols - fRect->x;

  	if ((fRect->y + fRect->h) <= nRows)
	    /* then the occlusion ends before the edge of the rect */
	    rowsOccl = fRect->h;
	else if (fRect->y >= nRows)
	    /* the occlusion is not in the region */
	    rowsOccl = 0;
	else rowsOccl = nRows - fRect->y;
    }
    else
	/* there is no occlusion */
	colsOccl = rowsOccl = 0;
    return colsOccl * rowsOccl;
}


/***************** Functions called by other library modules *************/

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Compute the capacity of the glyph region.
 *
 * LongDescription:::
 * Computes the byte capacity of the given rectangle.  Also returns
 * codeword information.  This is useful if this is a non-self describing
 * rectangle.  If the rectangle is filled to capacity and the ecc level
 * and rectangle size are the same as given to this routine, the 
 * codeword information should be the same as that returned by 
 * wp_doComputeLayout.
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  selfDescribing:: 		does this rectangle have a key codeword?
 *  hasOcclusion::		does this rectangle have an occlusion?
 *  glyphSize::			size of glyphs
 *  eccLevel::			ecc percentage
 *  rcOcclusion::		occlusion rectangle
 *  rectWidth::      		width of rectangle in micas
 *  rectHeight::     		height of rectangle in micas
 *
 * Output:::
 *  pMaxBytes::			ptr to capacity of rectangle in bytes
 *  pBytesPerCodeword::   	ptr to size of a codeword
 *  pECCBytesPerCodeword::    	ptr to number of ecc byte in each codeword
 *  pNumCodewords::		ptr to number of codewords
 *  pGOcclusion::		ptr to occlusion in units of glyphs
 *  pFOcclusion::		ptr to occlusion in units of frames
 *  pFrameWidth::		ptr to width of each frame
 *  pFrameHeight::		ptr to height of each frame
 *  pFrameCols::		ptr to number of frames in each row
 *  pFrameRows::		ptr to number of frames in each column
 *  pPadHorz::			ptr to number of horizontal pad glyphs
 *  pPadVert::			ptr to number of vertical pad glyphs
 *  pNumUnusedFrameBytes::	ptr to number of bytes of glyphs that contain
 *				no user data, ECC, CRC or header info.
 *
 * Returns:::
 *  cErrDG_InvalidGlyphSize	invalid glyph handle
 *  cErrDG_OcclusionSpansRect	occlusion spans either the vertical or
 *				horizontal width of the rectangle
 *  cErrDG_OcclusionTooSmall	occlusion is too small
 *  cErrInvalidParm		width and height parameters invalid
 *  cSuccessful			successful completion
 *
 * Notes:::
 * The units of the rectWidth, rectHeight and glyphSize must be the same.
 * However, this procedure does not care what the units are.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig			initial version
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
wp_doComputeCapacity (
    IPWState   *pIPWState,
    Bool	selfDescribing,
    Bool	hasOcclusion,
    Float64	glyphSize,
    UInt8	eccLevel,
    LayoutRect *rcOcclusion,
    UInt32	rectWidth,
    UInt32	rectHeight,
    UInt32     *pMaxBytes,
    UInt8      *pBytesPerCodeword,
    UInt8      *pECCBytesPerCodeword,
    UInt32     *pNumCodewords,
    LayoutRect *pGOcclusion,
    LayoutRect *pFOcclusion,
    UInt32     *pFrameWidth,
    UInt32     *pFrameHeight,
    UInt32     *pFrameCols,
    UInt32     *pFrameRows,
    UInt32     *pPadHorz,
    UInt32     *pPadVert,
    UInt32     *pNumUnusedFrameBytes)
{
PROLOGUE;
UInt32		 capacity = 0;
UInt8		 codewordSize, codewordECC, codewordData;
LayoutRect	 fOcclusion, gOcclusion, occlusion;
UInt32		 frameWidth = 0, frameHeight = 0;
UInt32		 nBytesPerFrame;
UInt32		 nCodewords;
UInt32		 nCols = 0, nRows = 0;
UInt32		 nFramesPerRow, nFramesPerCol;
UInt32		 nGlyphsPerRow, nGlyphsPerCol;
UInt32		 nTotalFrames;
UInt32		 tCapacity;
UInt8		 tCodewordSize, tCodewordECC, tCodewordData;
UInt32		 tCodewords;
UInt32		 totalEncodedBytes;

    ERR_GEN(eccLevel >= 100, cErrDG_InvalidECCLevel);
    ERR_GEN(glyphSize == 0, cErrDG_InvalidGlyphSize);
    ERR_GEN(rectWidth == 0, cErrInvalidParm);
    ERR_GEN(rectHeight == 0, cErrInvalidParm);

    /* convert the occlusion into glyphs occluded */
    if (hasOcclusion)
    {
	gOcclusion.x = (UInt32)((Float64)rcOcclusion->x / glyphSize);
	gOcclusion.y = (UInt32)((Float64)rcOcclusion->y / glyphSize);
	gOcclusion.w = (UInt32)ceil((Float64)rcOcclusion->w / glyphSize);
	gOcclusion.h = (UInt32)ceil((Float64)rcOcclusion->h / glyphSize);
    }
    else
    {
	gOcclusion.x = gOcclusion.y = 0;
	gOcclusion.w = gOcclusion.h = 0;
    }

    fOcclusion.x = fOcclusion.y = 0;
    fOcclusion.w = fOcclusion.h = 0;

    /* convert the rectangle width and height dimensions into glyphs */
    nGlyphsPerRow = (UInt32)((Float64)rectWidth / glyphSize);
    nGlyphsPerCol = (UInt32)((Float64)rectHeight / glyphSize);

    /* check to make sure that we can get at least one frame */
    ERR_GEN((nGlyphsPerRow < (THIN_FRAME_WIDTH + 2)), cErrInvalidParm);
    ERR_GEN((nGlyphsPerCol < (SHORT_FRAME_HEIGHT + 2)), cErrInvalidParm);
    /* check to see if the width makes it a 4x16 frame but the
     * height is insufficient.
     */
    ERR_GEN(((nGlyphsPerRow < (OCC_FRAME_WIDTH + 2)) && 
	     (nGlyphsPerCol < (FRAME_HEIGHT + 2))),
	    cErrInvalidParm);
    /* check to see if the height makes it a 16x4 frame but the
     * width is insufficient.
     */
    ERR_GEN(((nGlyphsPerRow < (FRAME_HEIGHT + 2)) && 
	     (nGlyphsPerCol < (OCC_FRAME_WIDTH + 2))),
	    cErrInvalidParm);

    /* first check if 4x16 frames are possible.  since we are just
     * starting out this is the best so far.
     */
    if ((nGlyphsPerRow < (OCC_FRAME_WIDTH + 2)) && 
	(hasOcclusion == cFalse))
    {
	nFramesPerRow = GLYPHS_TO_FRAMES(nGlyphsPerRow, THIN_FRAME_WIDTH);
	nFramesPerCol = GLYPHS_TO_FRAMES(nGlyphsPerCol, THIN_FRAME_HEIGHT);

	if (nFramesPerRow == 1)
	{
    	    nBytesPerFrame = THIN_FRAME_WIDTH * THIN_FRAME_HEIGHT / 8;

	    ERR_IGNORE(wp_calcEncoding(pIPWState, selfDescribing, 
				       nFramesPerCol * nBytesPerFrame, 0,
				       eccLevel, &nCodewords, &codewordSize,
				       &codewordECC, &codewordData));
	    capacity = codewordData * nCodewords;
	    frameWidth = THIN_FRAME_WIDTH;
	    frameHeight = THIN_FRAME_HEIGHT;
	    nRows = nFramesPerCol;
	    nCols = nFramesPerRow;
	}
    }

    /* check if 16x4 frames are possible. */
    if ((nGlyphsPerCol < (OCC_FRAME_HEIGHT + 2)) && 
	(hasOcclusion == cFalse))
    {
	nFramesPerRow = GLYPHS_TO_FRAMES(nGlyphsPerRow, SHORT_FRAME_WIDTH);
	nFramesPerCol = GLYPHS_TO_FRAMES(nGlyphsPerCol, SHORT_FRAME_HEIGHT);

	if (nFramesPerCol == 1)
	{
    	    nBytesPerFrame = SHORT_FRAME_WIDTH * SHORT_FRAME_HEIGHT / 8;

	    ERR_IGNORE(wp_calcEncoding(pIPWState, selfDescribing, 
				       nFramesPerRow * nBytesPerFrame, 0,
				       eccLevel, &tCodewords,
				       &tCodewordSize, &tCodewordECC, 
				       &tCodewordData));

	    /* check to see if this rectangle has a greater capacity 
	     * than any previous rectangle.
	     */
	    tCapacity = tCodewordData * tCodewords;
	    if (tCapacity > capacity)
	    {
		capacity = tCapacity;
		nCodewords = tCodewords;
		codewordSize = tCodewordSize;
		codewordECC = tCodewordECC;
		codewordData = tCodewordData;
	        frameWidth = THIN_FRAME_WIDTH;
	        frameHeight = THIN_FRAME_HEIGHT;
	        nRows = nFramesPerCol;
	        nCols = nFramesPerRow;
	    }
	}
    }

    /* check if 8x8 frames are possible. */
    if ((nGlyphsPerCol >= (OCC_FRAME_HEIGHT + 2)) && 
	(nGlyphsPerRow >= (OCC_FRAME_WIDTH + 2)))
    {
	nFramesPerRow = GLYPHS_TO_FRAMES(nGlyphsPerRow, OCC_FRAME_WIDTH);
	nFramesPerCol = GLYPHS_TO_FRAMES(nGlyphsPerCol, OCC_FRAME_HEIGHT);

    	/* change the occlusion rectangle from micas to frames */
	if (hasOcclusion == cTrue)
            ERR_IGNORE(wp_computeFrameRect(pIPWState, OCC_FRAME_WIDTH,
					   OCC_FRAME_HEIGHT, 
					   &gOcclusion, &fOcclusion));

    	nBytesPerFrame = OCC_FRAME_WIDTH * OCC_FRAME_HEIGHT / 8;
	nTotalFrames = (nFramesPerRow * nFramesPerCol) -
	    ERR_IGNORE(wp_getFramesOccluded(pIPWState,
					    nFramesPerRow, nFramesPerCol,
					    &fOcclusion));
	ERR_IGNORE(wp_calcEncoding(pIPWState, selfDescribing, 
				   nTotalFrames * nBytesPerFrame, 0,
				   eccLevel, &tCodewords,
				   &tCodewordSize, &tCodewordECC, 
				   &tCodewordData));

	/* check to see if this rectangle has a greater capacity 
	 * than any previous rectangle.
	 */
	tCapacity = tCodewordData * tCodewords;
	if (tCapacity > capacity)
	{
	    capacity = tCapacity;
	    nCodewords = tCodewords;
	    codewordSize = tCodewordSize;
	    codewordECC = tCodewordECC;
	    codewordData = tCodewordData;
	    frameWidth = THIN_FRAME_WIDTH;
	    frameHeight = THIN_FRAME_HEIGHT;
	    nRows = nFramesPerCol;
	    nCols = nFramesPerRow;
	    occlusion = fOcclusion;
	}
    }


    /* check if 16x16 frames are possible. */
    if ((nGlyphsPerCol >= (FRAME_HEIGHT + 2)) && 
	(nGlyphsPerRow >= (FRAME_WIDTH + 2)))
    {
	nFramesPerRow = GLYPHS_TO_FRAMES(nGlyphsPerRow, FRAME_WIDTH);
	nFramesPerCol = GLYPHS_TO_FRAMES(nGlyphsPerCol, FRAME_HEIGHT);

    	/* change the occlusion rectangle from micas to frames */
	if (hasOcclusion == cTrue)
            ERR_IGNORE(wp_computeFrameRect(pIPWState,
					   FRAME_WIDTH, FRAME_HEIGHT, 
					   &gOcclusion, &fOcclusion));

    	nBytesPerFrame = FRAME_WIDTH * FRAME_HEIGHT / 8;
	nTotalFrames = (nFramesPerRow * nFramesPerCol) -
	    ERR_IGNORE(wp_getFramesOccluded(pIPWState,
					    nFramesPerRow, nFramesPerCol,
					    &fOcclusion));
	ERR_IGNORE(wp_calcEncoding(pIPWState, selfDescribing, 
				   nTotalFrames * nBytesPerFrame, 0,
				   eccLevel, &tCodewords,
				   &tCodewordSize, &tCodewordECC, 
				   &tCodewordData));

	/* check to see if this rectangle has a greater capacity 
	 * than any previous rectangle.
	 */
	tCapacity = tCodewordData * tCodewords;
	if (tCapacity > capacity)
	{
	    capacity = tCapacity;
	    nCodewords = tCodewords;
	    codewordSize = tCodewordSize;
	    codewordECC = tCodewordECC;
	    codewordData = tCodewordData;
	    frameWidth = THIN_FRAME_WIDTH;
	    frameHeight = THIN_FRAME_HEIGHT;
	    nRows = nFramesPerCol;
	    nCols = nFramesPerRow;
	    occlusion = fOcclusion;
	}
    }

    if (hasOcclusion == cTrue)
	ERR_PASS(wp_checkOcclusion(pIPWState, nRows, nCols,
				   frameWidth, frameHeight,
				   &occlusion, &gOcclusion));

    if (pMaxBytes != NULL)
	*pMaxBytes = capacity;
    if (pBytesPerCodeword != NULL)
	*pBytesPerCodeword = codewordSize;
    if (pECCBytesPerCodeword != NULL)
	*pECCBytesPerCodeword = codewordECC;
    if (pNumCodewords != NULL)
	*pNumCodewords = nCodewords;
    if ((pGOcclusion != NULL) && hasOcclusion)
	*pGOcclusion = gOcclusion;
    if ((pFOcclusion != NULL) && hasOcclusion)
	*pFOcclusion = occlusion;
    if (pFrameWidth != NULL)
	*pFrameWidth = frameWidth;
    if (pFrameHeight != NULL)
	*pFrameHeight = frameHeight;
    if (pFrameCols != NULL)
	*pFrameCols = nCols;
    if (pFrameRows != NULL)
	*pFrameRows = nRows;
    if (pPadHorz != NULL)
	*pPadHorz = nCols - FRAMES_TO_GLYPHS(nCols, frameWidth);
    if (pPadVert != NULL)
	*pPadVert = nRows - FRAMES_TO_GLYPHS(nRows, frameHeight);
    if (pNumUnusedFrameBytes != NULL)
    {
	    /* Calculate the number of bytes encoded.  This includes the
	     * number of user data bytes + number of ECC bytes +
	     * CRC + the keycodeword.  That's only used if this is
	     * a self-describing rectangle.
	     */
	if (selfDescribing == cTrue)
	    totalEncodedBytes = ERR_IGNORE(wp_calcHeaderSize(pIPWState));
	else
	    totalEncodedBytes = 0;
	totalEncodedBytes += nCodewords * codewordSize;

	    /* Calculate a number used to get the efficiency of the
	     * encoding. This is the number of bytes of glyphs minus
	     * the number of bytes containing useful stuff.
	     */
	*pNumUnusedFrameBytes =
	    (nRows * nCols * (frameWidth * frameHeight /8)) -
	    totalEncodedBytes;
    }


    NORMAL_RETURN;
}


/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Compute the actual layout of the glyph region.
 *
 * LongDescription:::
 * Computes the actual layout of the glyph region, in terms of frames
 * and pad glyphs, and the outside dimensions of the region, in micas.
 * Constraints are imposed by the glyph size, the frame size, the
 * expansion type, minimum and maximum dimensions, and an aspect ratio.
 *
 * The glyph size is immutable, and is set by the client. The frame
 * size is determined based on characteristics of the region. If
 * the region is very thin (i.e. less than one 8x8 frame) in either
 * the horizontal or vertical dimension, then 4x16 or 16x4 frames are
 * used.  Otherwise, the frame that uses the space most efficiently is
 * used.
 *
 * Note that padding is accomplished by increasing the number of rows
 * or columns until no more will fit in the specified dimensions, and
 * then using pad glyphs beyond that. This means that there are never
 * enough pad glyphs in either dimension to constitute a complete frame.
 * The extra frames are taken advantage of by increasing the parity
 * fraction beyond that specified by the client, thus utilising as much
 * of the glyph region as possible for increased robustness.
 *
 * The frame size and number of pad glyphs in each dimension are stored
 * in the layout structure. The dimensions of the region to be used are
 * returned to the client via the rcLayout parameter. If the size of
 * the required region is larger than the area specified by the client,
 * then an error (cErrDG_LayoutOverflow) is returned, and rcLayout contains
 * the minimum region size required, according to the expansion type.
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  pDGLayout::      		layout handle
 *  selfDescribing:: 		does this rectangle have a key codeword?
 *
 * Output:::
 *  rectWidth::      		width of rectangle in micas
 *  rectHeight::     		height of rectangle in micas
 *  codewordSize::   		size of a codeword
 *  codewordECC::    		number of ecc byte in each codeword
 *
 * Returns:::
 *  cErrDG_InvalidLayout	invalid layout handle
 *  cErrDG_InvalidGlyphSize	invalid glyph handle
 *  cErrDG_OcclusionSpansRect	occlusion spans either the vertical or
 *				horizontal width of the rectangle
 *  cErrDG_OcclusionTooSmall	occlusion is too small
 *  cErrDG_LayoutOverflow	data would not fit in specified dimensions
 *  cSuccessful			successful completion
 *
 * Notes:::
 * This function sets the following fields of the layout structure
 * passed as a parameter:
 *
 *     frameCols		number of rows of frames
 *     frameRows		number of columns of frames
 *     padHorz			number of horizontal pad glyphs
 *     padVert			number of vertical pad glyphs
 *     fOcclusion		occlusion in frame units
 *     gOcclusion		occlusion in glyph units
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper		initial version
 *  Julie Craig			bug fixes and occlusions
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doComputeLayout (
    IPWState   *pIPWState,
    PDGLayout	pDGLayout,
    Bool	selfDescribing,
    UInt32     *rectWidth,
    UInt32     *rectHeight,
    UInt8      *codewordSize,
    UInt8      *codewordECC)
{
    PROLOGUE;
    PLayout pLayout;
    UInt32 nEncodingBytes;
    UInt32 nMinHorz, nMaxHorz, nMinVert, nMaxVert;
    UInt32 nMinGlyphsPerRow, nMaxGlyphsPerRow;
    UInt32 nMinFramesPerRow, nMaxFramesPerRow = 0;
    UInt32 nMinGlyphsPerCol, nMaxGlyphsPerCol;
    UInt32 nMinFramesPerCol, nMaxFramesPerCol = 0;
    UInt32 nRows = 0xFFFFFFFF, nCols = 0xFFFFFFFF;
    UInt32 nPadHorz = 0, nPadVert = 0;
    UInt32 nGlyphsInRect = 0xFFFFFFFF, nUnusedGlyphs = 0xFFFFFFFF;
    Float64 glyphSize;
    UInt32 medFrameWidth, medFrameHeight;
    UInt32 normFrameWidth, normFrameHeight;
    UInt32 tMinFramesPerRow, tMaxFramesPerRow;
    UInt32 tMinFramesPerCol, tMaxFramesPerCol;
    UInt32 tRows, tCols;
    UInt32 tPadHorz = 0, tPadVert = 0;
    UInt32 tFrameBytes, tTotalFrames;
    UInt32 tGlyphsInRect, tUnusedGlyphs;
    LayoutRect fOcclusion;
    UInt32 horzAspect = 0, vertAspect = 0;

    ERR_GEN(pDGLayout == NULL, cErrDG_InvalidLayout);

    pLayout = &(pDGLayout->lyt);
    pDGLayout->enc.selfDescribing = selfDescribing;
    fOcclusion.x = fOcclusion.y = 0;
    fOcclusion.w = fOcclusion.h = 0;

    /* Get the real glyph size to be used */
    glyphSize = pLayout->glyphSize;
    ERR_GEN(glyphSize == 0, cErrDG_InvalidGlyphSize);

    /* Check for defaults */
    nMinHorz = (pLayout->minHorz == DG_DEFAULT) ? DIM_MIN_HORZ : pLayout->minHorz;
    nMaxHorz = (pLayout->maxHorz == DG_DEFAULT) ? DIM_MAX_HORZ : pLayout->maxHorz;
    nMinVert = (pLayout->minVert == DG_DEFAULT) ? DIM_MIN_VERT : pLayout->minVert;
    nMaxVert = (pLayout->maxVert == DG_DEFAULT) ? DIM_MAX_VERT : pLayout->maxVert;

    switch (pLayout->expansionType) 
    {
    case cExpHorzFirst:
	/* if the expansion type is horizontal first, then the min and
	 * max should be the same since the client has specified that
	 * the horizontal should be first priority.  this eliminates
	 * the confusing case of expansion type of horizontal first,
	 * the min and max horizontal different and the min and max
	 * vertical the same.  this should just basically fill the max
	 * rectangle.
	 */
	nMinHorz = nMaxHorz;
	break;
    case cExpVertFirst:
	nMinVert = nMaxVert;
	break;
    case cExpAspectRatio:
    default:
        /*
         * The minima are related when an aspect ratio is being used,
         * so adjust these if necessary.
         */
    	{
	Float64 facHorz, facVert, factor;
	UInt32 newMinHorz, newMinVert;

	    horzAspect = pLayout->horzAspect;
	    vertAspect = pLayout->vertAspect;
	    facHorz = (Float64)nMinHorz / horzAspect;
 	    facVert = (Float64)nMinVert / vertAspect;
	    factor = MAX(facHorz, facVert);
	    /* watch out for roundoff problems.  if the new minimum 
	     * calculated isn't at least a glyph bigger then forget it.
	     */
    	    newMinHorz = (UInt32)(factor * horzAspect);
	    if (newMinHorz > (nMinHorz + glyphSize))
	        nMinHorz = newMinHorz;
	    newMinVert = (UInt32)(factor * vertAspect);
	    if (newMinVert > (nMinVert + glyphSize))
	        nMinVert = newMinVert;
    	}
	break;
    }


    /* determine number of bytes required */
    ERR_PASS(wp_getEncodingSize(pIPWState, &(pDGLayout->enc), &nEncodingBytes));

    /* if we have an occlusion, decide which glyphs cannot be used.
     * note we are relying on the truncate from float to int here.
     * we want to truncate instead of round because if the occlusion
     * hits a glyph anywhere, the glyph should be considered to be bad.
     */
    if (pLayout->hasOcclusion)
    {
	pLayout->gOcclusion.x = (UInt32)((Float64)pLayout->rcOcclusion.x / 
	                                  glyphSize);
	pLayout->gOcclusion.y = (UInt32)((Float64)pLayout->rcOcclusion.y / 
	                                  glyphSize);

     	/* and the width in glyphs. */
	pLayout->gOcclusion.w = (UInt32)ceil((Float64)pLayout->rcOcclusion.w /
	                                      glyphSize);
	pLayout->gOcclusion.h = (UInt32)ceil((Float64)pLayout->rcOcclusion.h / 
	                                      glyphSize);
    }

    /* Compute min and max dimensions, in glyphs */
    nMinGlyphsPerRow = (UInt32)((Float64)nMinHorz / glyphSize);
    nMaxGlyphsPerRow = (UInt32)((Float64)nMaxHorz / glyphSize);
    nMinGlyphsPerCol = (UInt32)((Float64)nMinVert / glyphSize);
    nMaxGlyphsPerCol = (UInt32)((Float64)nMaxVert / glyphSize);

    /* Compute the layout according to requested expansion type */
    medFrameWidth = (UInt32)((Float64)(OCC_FRAME_WIDTH + 2) * glyphSize);
    medFrameHeight = (UInt32)((Float64)(OCC_FRAME_HEIGHT + 2) * glyphSize);
    normFrameWidth = (UInt32)((Float64)(FRAME_WIDTH + 2) * glyphSize);
    normFrameHeight = (UInt32)((Float64)(FRAME_HEIGHT + 2) * glyphSize);

    /* the following contains alot of duplicated code.  this should be
     * cleaned up but until it is be careful when editting since you
     * may need to change mulitple places.  this would be cleaner if
     * there were less parameters to pass to the procedures - maybe a
     * structure?
     */

    /* first check if 4x16 frames are possible.  since we are just
     * starting out this is the best so far.
     */
    if ((nMinHorz < medFrameWidth) && (pLayout->hasOcclusion == cFalse))
    {
	ERR_IGNORE(
	    wp_computeMinMaxFrames(pIPWState,
				   THIN_FRAME_WIDTH, THIN_FRAME_HEIGHT,
			    	   nMinGlyphsPerRow, nMaxGlyphsPerRow,
			    	   nMinGlyphsPerCol, nMaxGlyphsPerCol, 
				   nEncodingBytes, &nMinFramesPerRow, 
			    	   &nMaxFramesPerRow, &nMinFramesPerCol, 
			    	   &nMaxFramesPerCol, &tFrameBytes,
				   &tTotalFrames));

	switch (pLayout->expansionType)
	{
	case cExpHorzFirst:
	    ERR_IGNORE(
		wp_doHorizExpansion(pIPWState, tTotalFrames, nMaxGlyphsPerRow, 
			     	    nMinGlyphsPerCol, nMaxFramesPerRow, 
			     	    nMinFramesPerCol, THIN_FRAME_WIDTH,  
			     	    THIN_FRAME_HEIGHT, &fOcclusion,  
			     	    &nCols, &nRows, &nPadHorz, &nPadVert));
	    break;

	case cExpVertFirst:
	    ERR_IGNORE(
	        wp_doVertExpansion(pIPWState, tTotalFrames, nMinGlyphsPerRow,
			    	   nMaxGlyphsPerCol, nMinFramesPerRow,
			    	   nMaxFramesPerCol, THIN_FRAME_WIDTH,
			    	   THIN_FRAME_HEIGHT, &fOcclusion,
			    	   &nCols, &nRows, &nPadHorz, &nPadVert));
	    break;

        default:
	case cExpAspectRatio:
	    ERR_IGNORE(
	        wp_computeRowCol(pIPWState, THIN_FRAME_WIDTH, THIN_FRAME_HEIGHT,
			     	 nMinHorz, nMaxHorz, nMinVert, nMaxVert,
			    	 nMinGlyphsPerRow, nMaxGlyphsPerRow,
			     	 nMinGlyphsPerCol, nMaxGlyphsPerCol,
			     	 nMinFramesPerRow, nMaxFramesPerRow,
			     	 nMinFramesPerCol, nMaxFramesPerCol,
			     	 horzAspect, vertAspect, &nRows, &nCols,
			     	 &nPadHorz, &nPadVert));
	    ERR_IGNORE(
	        wp_dimensionsFromAspectRatio(pIPWState, horzAspect, vertAspect,
					     tTotalFrames, THIN_FRAME_WIDTH,
				      	     THIN_FRAME_HEIGHT, &fOcclusion,
				      	     &nPadHorz, &nPadVert, &nRows,
					     &nCols));
	    break;
	}
	if (nCols == 1)
	{
	    nGlyphsInRect = (FRAMES_TO_GLYPHS(nRows, THIN_FRAME_HEIGHT) + tPadVert) *
			     (THIN_FRAME_WIDTH + 2 + tPadHorz);
	    nUnusedGlyphs = 
		(nPadHorz * FRAMES_TO_GLYPHS(nRows, THIN_FRAME_HEIGHT)) +
		(nPadVert * (THIN_FRAME_WIDTH + 2 + nPadHorz)) +
		((nRows - tTotalFrames) * tFrameBytes * 8);
	    pLayout->frameWidth = THIN_FRAME_WIDTH;
	    pLayout->frameHeight = THIN_FRAME_HEIGHT;
	    pLayout->fOcclusion = fOcclusion;
	    pLayout->nUnusedFrameBytes = 
		(nRows * tFrameBytes) - nEncodingBytes;
	}
	else 
	{   /* make sure the following code will find this as too big */
	    nRows = nCols = 0xFFFFFFFF; 
	    nMaxFramesPerCol = nMaxFramesPerRow = 0;
	}
    }

    /* check if 16x4 frames are possible. */
    if ((nMinVert < medFrameHeight) && (pLayout->hasOcclusion == cFalse))
    {
	ERR_IGNORE(
	    wp_computeMinMaxFrames(pIPWState,
				   SHORT_FRAME_WIDTH, SHORT_FRAME_HEIGHT,
			      	   nMinGlyphsPerRow, nMaxGlyphsPerRow,
			       	   nMinGlyphsPerCol, nMaxGlyphsPerCol,   
			       	   nEncodingBytes, &tMinFramesPerRow, 
			       	   &tMaxFramesPerRow, &tMinFramesPerCol,  
			       	   &tMaxFramesPerCol, &tFrameBytes,
			       	   &tTotalFrames));

	switch (pLayout->expansionType)
	{
	case cExpHorzFirst:
	    ERR_IGNORE(
	        wp_doHorizExpansion(pIPWState, tTotalFrames, nMaxGlyphsPerRow, 
			            nMinGlyphsPerCol, tMaxFramesPerRow, 
			            tMinFramesPerCol, SHORT_FRAME_WIDTH, 
			            SHORT_FRAME_HEIGHT, &fOcclusion,  
			            &tCols, &tRows, &tPadHorz, &tPadVert));
	    break;

	case cExpVertFirst:
	    ERR_IGNORE(
	        wp_doVertExpansion(pIPWState, tTotalFrames, nMinGlyphsPerRow, 
			      	   nMaxGlyphsPerCol, tMinFramesPerRow,
			       	   tMaxFramesPerCol, SHORT_FRAME_WIDTH, 
			           SHORT_FRAME_HEIGHT, &fOcclusion,
			           &tCols, &tRows, &tPadHorz, &tPadVert));
	    break;

        default:
	case cExpAspectRatio:
	    ERR_IGNORE(
	        wp_computeRowCol(pIPWState,
				 SHORT_FRAME_WIDTH, SHORT_FRAME_HEIGHT,
			     	 nMinHorz, nMaxHorz, nMinVert, nMaxVert,
			     	 nMinGlyphsPerRow, nMaxGlyphsPerRow,
			     	 nMinGlyphsPerCol, nMaxGlyphsPerCol,
			     	 tMinFramesPerRow, tMaxFramesPerRow,
			     	 tMinFramesPerCol, tMaxFramesPerCol,
			     	 horzAspect, vertAspect, &tRows, &tCols,
			     	 &tPadHorz, &tPadVert));
	    ERR_IGNORE(
	        wp_dimensionsFromAspectRatio(pIPWState, horzAspect, vertAspect,
					     tTotalFrames,
				             SHORT_FRAME_WIDTH, 
					     SHORT_FRAME_HEIGHT,
				             &fOcclusion, &tPadHorz, 
				             &tPadVert, &tRows, &tCols));
	    break;
	}
    /* if the best layout that we found so far was below the maximum size
     * and this layout is below the maximum size then compare them to
     * see which is the best.  if the best layout that we found so far
     * is over the maximum size then compare the layouts and pick the most
     * efficient one.
     */
    if (((nRows <= nMaxFramesPerCol) && (nCols <= nMaxFramesPerRow) &&
     	 (tRows <= tMaxFramesPerCol) && (tCols <= tMaxFramesPerRow)) ||
	((nRows > nMaxFramesPerCol) || (nCols > nMaxFramesPerRow)))
	/* there can only be one row of 16x4 frames */
	if (tRows == 1)
	{
	    tGlyphsInRect = (FRAMES_TO_GLYPHS(tCols, SHORT_FRAME_WIDTH) + tPadHorz) *
			     (SHORT_FRAME_HEIGHT + 2 + tPadVert);

	    tUnusedGlyphs = 
		(tPadHorz * FRAMES_TO_GLYPHS(tRows, SHORT_FRAME_HEIGHT)) +
		(tPadVert * (FRAMES_TO_GLYPHS(tCols, SHORT_FRAME_WIDTH) + tPadHorz)) +
		((tCols - tTotalFrames) * tFrameBytes * 8);
            if ((tGlyphsInRect < nGlyphsInRect) ||
	        ((tGlyphsInRect == nGlyphsInRect) &&
	         (((Float64)tUnusedGlyphs / (Float64)tGlyphsInRect) <
	          ((Float64)nUnusedGlyphs / (Float64)nGlyphsInRect))))
	    {
		pLayout->frameWidth = SHORT_FRAME_WIDTH;
		pLayout->frameHeight = SHORT_FRAME_HEIGHT;
		pLayout->fOcclusion = fOcclusion;
	    	pLayout->nUnusedFrameBytes = 
		    (tCols * tFrameBytes) - nEncodingBytes;
		nCols = tCols;
		nRows = tRows;
		nPadHorz = tPadHorz;
		nPadVert = tPadVert;
		nUnusedGlyphs = tUnusedGlyphs;
		nGlyphsInRect = tGlyphsInRect;
		nMaxFramesPerCol = tMaxFramesPerCol;
		nMaxFramesPerRow = tMaxFramesPerRow;
	    }
	}
    }

    /* check if 8x8 frames are possible. */
    if (((pLayout->expansionType == cExpVertFirst) &&
	 ((nMaxVert >= medFrameHeight) || (nMaxVert == 0))) ||
	((pLayout->expansionType == cExpHorzFirst) &&
	 ((nMaxHorz >= medFrameWidth) || (nMaxHorz == 0))) |
	(pLayout->expansionType == cExpAspectRatio))
    {
	ERR_IGNORE(
            wp_computeMinMaxFrames(pIPWState,
				   OCC_FRAME_WIDTH, OCC_FRAME_HEIGHT, 
			   	   nMinGlyphsPerRow, nMaxGlyphsPerRow,
			    	   nMinGlyphsPerCol, nMaxGlyphsPerCol, 
			   	   nEncodingBytes, &tMinFramesPerRow, 
			    	   &tMaxFramesPerRow, &tMinFramesPerCol, 
			    	   &tMaxFramesPerCol, &tFrameBytes,
				   &tTotalFrames));

    	/* change the rectangle from micas to frames */
	if (pLayout->hasOcclusion)
            ERR_IGNORE(wp_computeFrameRect(pIPWState, OCC_FRAME_WIDTH,
					   OCC_FRAME_HEIGHT, 
		             		   &pLayout->gOcclusion,
					   &fOcclusion));

	switch (pLayout->expansionType)
	{
	case cExpHorzFirst:
	    ERR_IGNORE(
	        wp_doHorizExpansion(pIPWState, tTotalFrames, nMaxGlyphsPerRow, 
			     	    nMinGlyphsPerCol, tMaxFramesPerRow, 
			     	    tMinFramesPerCol, OCC_FRAME_WIDTH, 
			     	    OCC_FRAME_HEIGHT, &fOcclusion,  
			     	    &tCols, &tRows, &tPadHorz, &tPadVert));
	    break;

	case cExpVertFirst:
	    ERR_IGNORE(
	        wp_doVertExpansion(pIPWState, tTotalFrames, nMinGlyphsPerRow, 
			    	   nMaxGlyphsPerCol, tMinFramesPerRow,
			    	   tMaxFramesPerCol, OCC_FRAME_WIDTH, 
			    	   OCC_FRAME_HEIGHT, &fOcclusion,
			    	   &tCols, &tRows, &tPadHorz, &tPadVert));
	    break;

        default:
	case cExpAspectRatio:
	    ERR_IGNORE(
	        wp_computeRowCol(pIPWState, OCC_FRAME_WIDTH, OCC_FRAME_HEIGHT,
			  	 nMinHorz, nMaxHorz, nMinVert, nMaxVert,
			  	 nMinGlyphsPerRow, nMaxGlyphsPerRow,
			  	 nMinGlyphsPerCol, nMaxGlyphsPerCol,
			  	 tMinFramesPerRow, tMaxFramesPerRow,
			  	 tMinFramesPerCol, tMaxFramesPerCol,
			  	 horzAspect, vertAspect, &tRows, &tCols,
			  	 &tPadHorz, &tPadVert));
	    ERR_IGNORE(
	        wp_dimensionsFromAspectRatio(pIPWState, horzAspect, vertAspect,
					     tTotalFrames, OCC_FRAME_WIDTH,
				      	     OCC_FRAME_HEIGHT, &fOcclusion,
				      	     &tPadHorz, &tPadVert,
				      	     &tRows, &tCols));
	    break;
	}

        if (((nRows <= nMaxFramesPerCol) && (nCols <= nMaxFramesPerRow) &&
     	     (tRows <= tMaxFramesPerCol) && (tCols <= tMaxFramesPerRow)) ||
	    ((nRows > nMaxFramesPerCol) || (nCols > nMaxFramesPerRow)))
        {
            tGlyphsInRect = (FRAMES_TO_GLYPHS(tCols, OCC_FRAME_WIDTH) + tPadHorz) *
		            (FRAMES_TO_GLYPHS(tRows, OCC_FRAME_HEIGHT) + tPadVert);
            tUnusedGlyphs = 
	        (tPadHorz * FRAMES_TO_GLYPHS(tRows, OCC_FRAME_HEIGHT)) +
	        (tPadVert * (FRAMES_TO_GLYPHS(tCols, OCC_FRAME_WIDTH) + tPadHorz)) +
	        (((tRows * tCols) - tTotalFrames) * tFrameBytes * 8);
            if ((tGlyphsInRect < nGlyphsInRect) ||
	        ((tGlyphsInRect == nGlyphsInRect) &&
	         (((Float64)tUnusedGlyphs / (Float64)tGlyphsInRect) <
	          ((Float64)nUnusedGlyphs / (Float64)nGlyphsInRect))))
            {
	        pLayout->frameWidth = OCC_FRAME_WIDTH;
	        pLayout->frameHeight = OCC_FRAME_HEIGHT;
	        pLayout->fOcclusion = fOcclusion;
	    	pLayout->nUnusedFrameBytes = 
		    (tCols * tRows * tFrameBytes) - nEncodingBytes;
	        nCols = tCols;
	        nRows = tRows;
	        nPadHorz = tPadHorz;
	        nPadVert = tPadVert;
	        nUnusedGlyphs = tUnusedGlyphs;
	        nGlyphsInRect = tGlyphsInRect;
	        nMaxFramesPerCol = tMaxFramesPerCol;
	        nMaxFramesPerRow = tMaxFramesPerRow;
            }
        }
    }


    /* check if 16x16 frames are possible. */
    if (((pLayout->expansionType == cExpVertFirst) &&
	 ((nMaxVert >= normFrameHeight) || (nMaxVert == 0))) ||
	((pLayout->expansionType == cExpHorzFirst) &&
	 ((nMaxHorz >= normFrameWidth) || (nMaxHorz == 0))) |
	(pLayout->expansionType == cExpAspectRatio))
    {
	ERR_IGNORE(
            wp_computeMinMaxFrames(pIPWState, FRAME_WIDTH, FRAME_HEIGHT, 
				   nMinGlyphsPerRow,
			    	   nMaxGlyphsPerRow, nMinGlyphsPerCol,  
			    	   nMaxGlyphsPerCol, nEncodingBytes,  
			    	   &tMinFramesPerRow, &tMaxFramesPerRow,  
			    	   &tMinFramesPerCol, &tMaxFramesPerCol, 
			    	   &tFrameBytes, &tTotalFrames));

	/* change the rectangle from micas to frames */
        if (pLayout->hasOcclusion)
	    ERR_IGNORE(wp_computeFrameRect(pIPWState,
					   FRAME_WIDTH, FRAME_HEIGHT, 
		             		   &pLayout->gOcclusion, 
					   &fOcclusion));

	switch (pLayout->expansionType)
	{
	case cExpHorzFirst:
	    ERR_IGNORE(
	        wp_doHorizExpansion(pIPWState, tTotalFrames, nMaxGlyphsPerRow, 
			     	    nMinGlyphsPerCol, tMaxFramesPerRow, 
			     	    tMinFramesPerCol, FRAME_WIDTH, 
			     	    FRAME_HEIGHT, &fOcclusion,  
			     	    &tCols, &tRows, &tPadHorz, &tPadVert));
	    break;

	case cExpVertFirst:
	    ERR_IGNORE(
	        wp_doVertExpansion(pIPWState, tTotalFrames, nMinGlyphsPerRow, 
			    	   nMaxGlyphsPerCol, tMinFramesPerRow,
			    	   tMaxFramesPerCol, FRAME_WIDTH, 
			    	   FRAME_HEIGHT, &fOcclusion, &tCols,  
			    	   &tRows, &tPadHorz, &tPadVert));
	    break;

        default:
	case cExpAspectRatio:
	    ERR_IGNORE(
	        wp_computeRowCol(pIPWState, FRAME_WIDTH, FRAME_HEIGHT,
			  	 nMinHorz, nMaxHorz, nMinVert, nMaxVert,
			  	 nMinGlyphsPerRow, nMaxGlyphsPerRow,
			  	 nMinGlyphsPerCol, nMaxGlyphsPerCol,
			  	 tMinFramesPerRow, tMaxFramesPerRow,
			  	 tMinFramesPerCol, tMaxFramesPerCol,
			  	 horzAspect, vertAspect, &tRows, &tCols,
			  	 &tPadHorz, &tPadVert));
	   ERR_IGNORE(
		wp_dimensionsFromAspectRatio(pIPWState, horzAspect, vertAspect,
					     tTotalFrames, FRAME_WIDTH,
					     FRAME_HEIGHT, &fOcclusion, 
				      	     &tPadHorz, &tPadVert, &tRows, 
					     &tCols));
	    break;
	}

        if (((nRows <= nMaxFramesPerCol) && (nCols <= nMaxFramesPerRow) &&
     	     (tRows <= tMaxFramesPerCol) && (tCols <= tMaxFramesPerRow)) ||
	    ((nRows > nMaxFramesPerCol) || (nCols > nMaxFramesPerRow)))
        {
            tGlyphsInRect = (FRAMES_TO_GLYPHS(tCols, FRAME_WIDTH) + tPadHorz) *
		            (FRAMES_TO_GLYPHS(tRows, FRAME_HEIGHT) + tPadVert);
            tUnusedGlyphs = 
	        (tPadHorz * FRAMES_TO_GLYPHS(tRows, FRAME_HEIGHT)) +
	        (tPadVert * (FRAMES_TO_GLYPHS(tCols, FRAME_WIDTH) + tPadHorz)) +
	        (((tRows * tCols) - tTotalFrames) * tFrameBytes * 8);
            if ((tGlyphsInRect < nGlyphsInRect) ||
	        ((tGlyphsInRect == nGlyphsInRect) &&
	         (((Float64)tUnusedGlyphs / (Float64)tGlyphsInRect) <
	          ((Float64)nUnusedGlyphs / (Float64)nGlyphsInRect))))
	    {
	        pLayout->frameWidth = FRAME_WIDTH;
	        pLayout->frameHeight = FRAME_HEIGHT;
	        pLayout->fOcclusion = fOcclusion;
	    	pLayout->nUnusedFrameBytes = 
		    (tCols * tRows * tFrameBytes) - nEncodingBytes;
	        nCols = tCols;
	        nRows = tRows;
	        nPadHorz = tPadHorz;
	        nPadVert = tPadVert;
	        nMaxFramesPerCol = tMaxFramesPerCol;
	        nMaxFramesPerRow = tMaxFramesPerRow;
	        nUnusedGlyphs = tUnusedGlyphs;
	        nGlyphsInRect = tGlyphsInRect;
	    }
        }
    }

    /* check occlusion for size constraints */
    if (pLayout->hasOcclusion)
	ERR_PASS(wp_checkOcclusion(pIPWState, nRows, nCols, pLayout->frameWidth,
				   pLayout->frameHeight, 
				   &pLayout->fOcclusion,
				   &pLayout->gOcclusion));

    /* almost certainly this will have some fractional value and
     * changing the float to an int will truncate the value.  this
     * will leave us 1 mica too small and if the next call uses these 
     * values it will not suceed either.  so, add 1 to the result.
     * in the extremely rare case that there was no fraction component
     * will will be returning 1 mica more than we need.  I claim 1
     * mica is so small that this doesn't matter.
     */
    *rectWidth = (UInt32)(glyphSize *
	    (Float64)(FRAMES_TO_GLYPHS(nCols, (UInt32)pLayout->frameWidth) + nPadHorz)) + 1;
    *rectHeight = (UInt32)(glyphSize *
	    (Float64)(FRAMES_TO_GLYPHS(nRows, (UInt32)pLayout->frameHeight) + nPadVert)) + 1;

    /* Check for overflow */
    ERR_GEN((nRows > nMaxFramesPerCol) || (nCols > nMaxFramesPerRow) || 
	    (nMinHorz > nMaxHorz) || (nMinVert > nMaxVert),
	    cErrDG_LayoutOverflow);

    /* Set the layout fields */
    pLayout->frameCols = nCols;
    pLayout->frameRows = nRows;
    pLayout->padHorz = nPadHorz;
    pLayout->padVert = nPadVert;

    /* Adjust the encoding size for best fit.  if there is an occlusion
     * inside the rectangle that space must not be used but if the 
     * occlusion is outside the rectangle, all the space can be used.
     * if there is not occlusion, all the space can be used.
     */
    if (pLayout->hasOcclusion && (pLayout->fOcclusion.x < nCols) &&
        (pLayout->fOcclusion.y < nRows))
        ERR_IGNORE(
	    wp_adjustEncoding(pIPWState, &(pDGLayout->enc), 
            	((nRows * nCols) - 
		(pLayout->fOcclusion.w * pLayout->fOcclusion.h)) * 
	    	pLayout->frameWidth * pLayout->frameHeight / 8));
    else
        ERR_IGNORE(wp_adjustEncoding(pIPWState, &(pDGLayout->enc), 
            (nRows * nCols) * pLayout->frameWidth * pLayout->frameHeight / 8));

    if (codewordSize != NULL) *codewordSize = pDGLayout->enc.nCodewordSize;
    if (codewordECC != NULL) *codewordECC = pDGLayout->enc.nCodewordParity;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Creates/init the handle used for data encoding and layout manipulation
 *
 * Input:::
 *  pIPWState::		pointer to IP Core state
 *  mallocProc::     	memory allocation procedure
 *  callocProc::     	memory allocation and clearing procedure
 *  freeProc::       	memory deallocation procedure
 *  allocInfo::      	client ptr to pass to memory routines
 *
 * Output:::
 *  phLayout::       	layout handle
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successful completion
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doCreateLayout (
    IPWState   *pIPWState,
    MallocType	mallocProc,
    CallocType	callocProc,
    FreeType	freeProc,
    void       *allocInfo,
    HDGLayout  *phLayout)
{
    PROLOGUE;
    PDGLayout pLayout;

    ERR_GEN(phLayout == NULL, cErrInvalidParm);
    ERR_GEN(*phLayout != NULL, cErrInvalidParm);

    pLayout = (PDGLayout)GLY_CALLOC(callocProc, allocInfo,
				    sizeof(DGLayout), sizeof(UInt8));
    ERR_GEN(pLayout == NULL, cErrNomem);
    *phLayout = (HDGLayout)pLayout;

    pLayout->enc.nDataBytes = 0;
    pLayout->enc.nCodewords = 0;
    pLayout->enc.nCodewordSize = 0;
    pLayout->enc.nCodewordParity = 0;
    pLayout->enc.eccLevel = DEFAULT_ECC_LEVEL;
    pLayout->enc.selfDescribing = cTrue;	/* seems like a good default */

    pLayout->lyt.horzAspect = 1;
    pLayout->lyt.vertAspect = 1;
    pLayout->lyt.minHorz = DIM_MIN_HORZ;
    pLayout->lyt.maxHorz = DIM_MAX_HORZ;
    pLayout->lyt.minVert = DIM_MIN_VERT;
    pLayout->lyt.maxVert = DIM_MAX_VERT;
    pLayout->lyt.expansionType = cExpAspectRatio;
    pLayout->lyt.glyphSize = 0.0;
    pLayout->lyt.hasOcclusion = cFalse;
    pLayout->lyt.rcOcclusion.x = 0;
    pLayout->lyt.rcOcclusion.y = 0;
    pLayout->lyt.rcOcclusion.w = 0;
    pLayout->lyt.rcOcclusion.h = 0;
    pLayout->lyt.fOcclusion.x = 0;
    pLayout->lyt.fOcclusion.y = 0;
    pLayout->lyt.fOcclusion.w = 0;
    pLayout->lyt.fOcclusion.h = 0;
    pLayout->lyt.framesOccluded = 0;
    pLayout->lyt.gOcclusion.x = 0;
    pLayout->lyt.gOcclusion.y = 0;
    pLayout->lyt.gOcclusion.w = 0;
    pLayout->lyt.gOcclusion.h = 0;

    pLayout->lyt.isTransparent = cFalse;

    pLayout->lyt.frameWidth = FRAME_WIDTH;
    pLayout->lyt.frameHeight = FRAME_HEIGHT;
    pLayout->lyt.frameRows = 0;
    pLayout->lyt.frameCols = 0;
    pLayout->lyt.padHorz = 0;
    pLayout->lyt.padVert = 0;
    pLayout->lyt.nUnusedFrameBytes = 0;

    pLayout->mallocProc = mallocProc;
    pLayout->callocProc = callocProc;
    pLayout->freeProc = freeProc;
    pLayout->allocInfo = allocInfo;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Releases the structures created by dgCreateLayout
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  hLayout::   		layout handle     
 *
 * Returns:::
 *  cErrDG_InvalidLayout	layout handle is invalid
 *  cSuccessful			successful completion
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doDestroyLayout (
    IPWState   *pIPWState,
    HDGLayout  *hLayout)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)(*hLayout);

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);

    GLY_FREE(pLayout->freeProc, pLayout->allocInfo, pLayout);
    *hLayout = NULL;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set an aspect ratio
 *
 * LongDescription:::
 * Set an aspect ratio that will be used to constrain the shape of the
 * glyph rectangle.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  hLayout::        		layout handle
 *  horz::           		horizontal component of aspect ratio
 *  vert::           		vertical component of aspect ratio
 *
 * Returns:::
 *  cErrDG_InvalidLayout	invalid layout handle
 *  cSuccessful			successful completion
 *
 * Notes:::
 * This function sets the following fields of the layout structure
 * passed as a parameter:
 *
 *     horzAspect		horizontal component of aspect ratio
 *     vertAspect		vertical component of aspect ratio
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doSetAspectRatio (
    IPWState   *pIPWState,
    HDGLayout	hLayout,
    UInt32	horz,
    UInt32	vert)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)hLayout;

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);

    /* assume that if the client specified 0 as the aspect ratio
     * what they really set it to a default value.
     */
    pLayout->lyt.horzAspect = (horz == 0) ? 1 : horz;
    pLayout->lyt.vertAspect = (vert == 0) ? 1 : vert;

    NORMAL_RETURN;
}

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set the maximum number of bytes to be encoded in the glyph rectangle.
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  hLayout::			layout handle
 *  nBytes::			number of data bytes
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *************************************************************************/
ErrCode CDECL 
wp_doSetDataSize (
    IPWState   *pIPWState,
    HDGLayout	hLayout,
    UInt32	nBytes)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)hLayout;

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);
    ERR_GEN(nBytes == 0, cErrDG_InvalidByteCount);
    pLayout->enc.nDataBytes = nBytes;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set min and max horizontal and vertical dimensions
 *
 * LongDescription:::
 * Set min and max horizontal and vertical dimensions that will be used
 * to contrain the size of the glyph rectangle.
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  hLayout::        		layout handle
 *  minHorz::			minimum width of glyph rectangle
 *  maxHorz::			maximum width of glyph rectangle
 *  minVert::			minimum height of glyph rectangle
 *  maxVert::			maximum height of glyph rectangle 
 *
 * Returns:::
 *  cErrDG_InvalidLayout	invalid layout handle
 *  cErrInvalidParm		min dimension is greater than max
 *  cSuccessful			successful completion
 *
 *
 * Notes:::
 * This function sets the following fields of the layout structure
 * passed as a parameter:
 *
 *     minHorz			minimum width of glyph rectangle
 *     maxHorz			maximum width of glyph rectangle
 *     minVert			minimum height of glyph rectangle
 *     maxVert			maximum height of glyph rectangle 
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doSetDimensions (
    IPWState   *pIPWState,
    HDGLayout	hLayout,
    UInt32	minHorz,
    UInt32	maxHorz,
    UInt32	minVert,
    UInt32	maxVert)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)hLayout;
    UInt32 nMinHorz, nMaxHorz, nMinVert, nMaxVert;

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);

    /* Check for defaults */
    nMinHorz = (minHorz == DG_DEFAULT) ? DIM_MIN_HORZ : minHorz;
    nMaxHorz = (maxHorz == DG_DEFAULT) ? DIM_MAX_HORZ : maxHorz;
    nMinVert = (minVert == DG_DEFAULT) ? DIM_MIN_VERT : minVert;
    nMaxVert = (maxVert == DG_DEFAULT) ? DIM_MAX_VERT : maxVert;

    /* Check for inside out dimensions */
    ERR_GEN((nMinHorz > nMaxHorz) || (nMinVert > nMaxVert),
	    cErrInvalidParm);

    /* Set original (not default-adjusted) values */
    pLayout->lyt.minHorz = minHorz;
    pLayout->lyt.minVert = minVert;
    pLayout->lyt.maxHorz = maxHorz;
    pLayout->lyt.maxVert = maxVert;

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set the amount of error correction in each codeword.
 *
 * LongDescription:::
 * The error correction level is set as a percentage of the area, or
 * percentage of the number of encoded bytes, or percentage of the
 * number of bytes in each codeword.  An error correction level
 * of x% will correct for x% erasures (glyphs we know that we can't
 * derender) or (x/2)% errors (glyphs that were derendered incorrectly)
 * or any combination such that (percent erasures + 2*percent errors) <= x%.
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  hLayout::			layout handle
 *  eccLevel::			percentage of area given to error correction
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL 
wp_doSetECCLevel (
    IPWState   *pIPWState,
    HDGLayout	hLayout,
    UInt8	eccLevel)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)hLayout;

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);
    ERR_GEN(eccLevel >= 100, cErrDG_InvalidECCLevel);
    pLayout->enc.eccLevel = eccLevel;

    NORMAL_RETURN;
}


/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set the expansion type
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  hLayout::        		layout handle
 *  expansionType::  		expansion type
 *
 * Returns:::
 *  cErrDG_InvalidLayout	invalid layout handle
 *  cSuccessful			successful completion
 *
 * Notes:::
 * This function sets the following fields of the layout structure
 * passed as a parameter: expansionType (which defines how the rectangle
 * will be filled with glyphs).
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doSetExpansion (
    IPWState   *pIPWState,
    HDGLayout	hLayout,
    RectExpansion expansionType)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)hLayout;

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);

    pLayout->lyt.expansionType = expansionType;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the physical size of the glyphs to be used in the rectangle.
 *
 * LongDescription:::
 * This procedure sets the physical size of the glyphs to be encoded.
 * The glyph toolkit was written assuming physical dimensions are specified
 * in micas, but any other physical unit can be used so long as it is
 * used consistently in all toolkit calls.  In particular, the physical
 * units used to specify the glyph size, the occlusions and glyph block
 * boundaries must be the same.
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  hLayout::        		layout handle
 *  glyphSize::      		glyph size in micas
 *
 * Returns:::
 *  cErrDG_InvalidLayout	invalid layout handle
 *  cSuccessful			successful completion
 *
 * Notes:::
 * This function sets the following fields of the layout structure
 * passed as a parameter: glyphSize
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doSetGlyphSize (
    IPWState   *pIPWState,
    HDGLayout	hLayout,
    Float64	glyphSize)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)hLayout;

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);

    pLayout->lyt.glyphSize = glyphSize;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Specifies an occluded area
 *
 * LongDescription:::
 * Specifies an area within the glyph rectangle which should not be used
 * for valid glyph data.
 *
 * Input:::
 *  pIPWState::			pointer to IP Core state
 *  hLayout::        		layout handle
 *  occlXOffset::    		x offset in micas of the occlusion
 *  occlYOffset::    		y offset in micas of the occlusion
 *  occlWidth::      		width in micas of the occlusion
 *  occlHeight::     		height in micas of the occlusion
 *  horzAlign::      		how to place the occlusion horizontally
 *  vertAlign::      		how to place the occlusion vertically
 *  isTransparent::  		are there glyphs in the occlusion?
 *
 * Returns:::
 *  cErrDG_InvalidLayout	invalid layout handle
 *  cErrDG_NotSupported		horzAlign or vertAlign has incorrect value
 *  cErrDG_OcclusionAlreadySet	an occlusion has already been set for
 *				this rectangle.
 *  cSuccessful			successful completion
 *
 * Notes:::
 * The only alignment currently implemented in cAlignLow (or relative to
 * the upper left corner).
 * There can be only one occlusion per glyph rectangle.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doSetOcclusion (
    IPWState   *pIPWState,
    HDGLayout	hLayout,
    UInt32	occlXOffset,
    UInt32	occlYOffset,
    UInt32	occlWidth,
    UInt32	occlHeight,
    OcclAlignment horzAlign,
    OcclAlignment vertAlign,
    Bool	isTransparent)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)hLayout;

    ERR_GEN((horzAlign != cAlignLow) || (vertAlign != cAlignLow),
	    cErrDG_NotSupported);

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);
    ERR_GEN(pLayout->lyt.hasOcclusion, cErrDG_OcclusionAlreadySet);

    pLayout->lyt.hasOcclusion = cTrue;
    pLayout->lyt.rcOcclusion.x = occlXOffset;
    pLayout->lyt.rcOcclusion.y = occlYOffset;
    pLayout->lyt.rcOcclusion.w = occlWidth;
    pLayout->lyt.rcOcclusion.h = occlHeight;
    pLayout->lyt.isTransparent = isTransparent;

    NORMAL_RETURN;
}


/**************************** End of file *****************************/

/*
 $Log: layout.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.59  1998/01/14 00:13:43  ddavies
 layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.58  1996/09/20 00:13:03  jcraig
 jcraig on Thu Sep 19 17:12:44 PDT 1996

 layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v

 Alignment => OcclAlignment and Expansion => RectExpansion.
 (rename because of clash with SmartPaper exported name)

 Revision 1.57  1996/03/21 21:51:59  ddavies
 layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v

 Change a comment to wp_doGetLayoutHandleSize.

 Revision 1.56  1996/03/01 01:25:12  ddavies
 ddavies on Thu Feb 29 17:23:35 PST 1996

 layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v

 Fixed some comments.  Changes all the w
 gack, try again.  Changed all the wp_o_ names to the original wp_
 names.  This leaves the old routine names associated with the old
 routines in the grand glyph toolkit conversion.  New routine names
 will have V2 appended to them.

 Revision 1.55  1996/02/22 04:08:33  ddavies
 layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v

 Remove inclusion of defunct files dglaydef.pub and dglayout.prv.
 Add a bunch of output pointer args to wp_doComputeCapacity so that when
 these are non-NULL, the parameters associated with the returned capacity
 are written to these places.  This is intended to be done when a client
 is using w_computeCapacity to just verify that the desired amount of data
 fits into a known block.  In particular, the client doesn't want to fuss
 with wp_doComputeLayout, which tries harder to build a block.
 Also add a section at the bottom of wp_doComputeCapacity that actually
 updates all the output locations.

 Revision 1.54  1996/02/13 21:12:46  ddavies
 layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v

 Add inclusion of new layout creation include files, dglaydef.pub,
 dglayout.pub and dglayout.prv.  Change names of all existing encoding routines
 by adding
 o_ to their name prefixes.  The corresponding new routines will not have
 the o_.

 Revision 1.53  1996/01/04 20:24:45  ejaquith
 layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.52  1995/12/13  04:44:13  ejaquith
 * ejaquith on Tue Dec 12 20:44:05 PST 1995
 *
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * Fixed a couple documentation typos.
 *
 * Revision 1.51  1995/12/13  04:10:19  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.50  1995/12/12  19:37:57  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * rectangle and occlusion width and height parameters are now UInt32
 * instead of UInt16.
 * add wp_doSetECCLevel which sets the ecc level of the glyph rectangle.
 * ad wp_doSetDataSize which sets the max number of bytes to be
 * encoded in the glyph rectangle.
 * createLayout no longer takes the max bytes and ecc level.
 *
 * Revision 1.49  1995/12/07  01:12:07  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * add new procedure wp_doComputeCapacity which calculates the capacity
 * and codeword information of a non-selfdescribing rectangle.
 * add procedure wp_checkOcclusion which is called to check any
 * occlusion to make sure that it meets any size related constraints
 *
 * Revision 1.48  1995/11/23  02:08:02  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: make Eric's scripts quit whining when I forgot to put :: after
 * parameter names in the comments
 *
 * Revision 1.47  1995/11/13  20:13:38  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.46  1995/11/09  01:19:39  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: wrap ERR_IGNORE around all private procs which do not return an
 * ErrCode
 *
 * Revision 1.45  1995/11/08  22:39:38  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * add comment to wp_getFramesOccluded
 *
 * Revision 1.44  1995/11/07  21:51:33  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * include dgcom.prv for defs required now that the region handle
 * contains the ecc tables
 *
 * Revision 1.43  1995/10/16  22:57:12  dferrell
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.42  1995/10/13  22:02:52  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.41  1995/10/11  01:38:33  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * w_adjustEncoding => wp_adjustEncoding
 * wp_dgiGetEncodingSize => wp_getEncodingSize
 * wp_dgiInitEncoding => wp_initEncoding
 *
 * Revision 1.40  1995/09/30  00:19:33  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: Converted private-ish glyph routines to be of "wp_" form and 100% private.
 *
 * Revision 1.39  1995/09/29  23:16:01  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * PDGLAYOUT => PDGLayout.  get rid of wp_doGetLayoutHandleSize and as
 * a result, the layout handle must be NULL in wp_doGetLayoutHandle
 *
 * Revision 1.38  1995/09/29  01:21:11  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * remove some old procedure comments
 *
 * Revision 1.37  1995/09/29  00:43:52  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * Changed keyword  DataG;l to DataGlyph
 *
 * Revision 1.36  1995/09/27  23:44:08  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * new procedure comments
 *
 * Revision 1.35  1995/09/22  00:38:36  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: Fixed up wp_ issues.
 *
 * Revision 1.34  1995/09/21  22:54:29  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.33  1995/09/19  00:47:47  jcraig
 * jcraig on Mon Sep 18 17:46:05 PDT 1995
 *
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * in wp_doDestroyLayout, hLayout is a pointer to a handle not the handle
 * itself
 *
 * Revision 1.32  1995/09/07  01:01:27  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * HDGLAYOUT => HDGLayout. HDGREGION => HDGRegion.
 *
 * Revision 1.31  1995/09/06  20:21:15  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.30  1995/09/01  02:14:38  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.29  1995/08/31  21:15:04  ejaquith
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.28  1995/08/31  01:06:32  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * since Expansion and Alignment are no longer enums, the names have
 * changed slightly
 *
 * Revision 1.27  1995/08/17  22:32:01  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.26  1995/01/18  02:19:05  jcraig
 * jcraig on Tue Jan 17 18:18:41 PST 1995
 *
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * handle occlusions outside of the glyph rectangle correctly
 *
 * Revision 1.25  1994/12/05  21:18:26  jcraig
 * jcraig on Mon Dec 5 13:16:36 PST 1994
 *
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * fix compiler warnings about unused variables.
 * fix bug in DoHorzExpansion and wp_doVertExpansion when the occlusion
 * was larger than the current max height/width.
 * use the occlusion info stored in the layout handle (not the temp
 * variable) when determining if the occlusion spans the rectangle.
 *
 * Revision 1.24  1994/11/30  18:23:12  jcraig
 * jcraig on Wed Nov 30 10:02:07 PST 1994
 *
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * fix bug in DoHorzExpansion and wp_doVertExpansion so that pad glyphs will
 * not be ignored if the minimum size of the rectangle is larger than the
 * space actually needed.
 * calculate nUnusedFrameBytes correctly.
 *
 * Revision 1.22  1994/11/01  03:02:39  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * include memory.pub since it is no longer in dgencode.pub.
 * remove nTotalFrames and nTotalBytes and use tTotalFrames and tTotalBytes
 * instead.
 * keep nUnusedFrameBytes in the layout handle (from dgComputeLayout).
 *
 * Revision 1.21  1994/10/12  20:55:31  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * remove use of LayoutRect from dgComputeLayout, dgComputeNonSDLayout
 * and DoComputeLayout, dgSetOcclusion.  add FAR to pointers in the
 * api parameters.
 *
 * Revision 1.20  1994/10/12  01:04:27  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * removed procedures (SetFrameSize and dgSetFrameSize) that allowed
 * the frame width and height to be set even though it was ignored.
 *
 * Revision 1.19  1994/09/30  23:27:13  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * check for valid ecc level in dgCreateLayout.  ecc level is not a UInt8
 * not an enumerated.
 *
 * Revision 1.18  1994/09/30  02:00:11  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * replace EDD in handles and constants with DG
 *
 * Revision 1.17  1994/09/28  21:12:57  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 *  change edd prefix to dg.
 *
 * Revision 1.16  1994/09/23  00:24:29  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * add eddGetLayoutHandleSize which returns size of the handle so that
 * client may allocate it instead of relying on eddCreateHandle to
 * allocate the space.  add eddComputeNonSDLayout to allow rectangles
 * without a key codeword (non selfdescribing rectangle).
 *
 * Revision 1.15  1994/08/16  23:58:45  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * calculate number of glyphs in rectangle correctly in eddComputeLayout.
 * don't let the width and height in micas overflow the UInt16
 *
 * Revision 1.14  1994/08/02  00:43:25  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * ooops.  the glybox changes are really the layout changes.
 * fix cc on solaris compiler warnings.
 * previous version busted occlusions if the fill type was aspect ratio.
 * don't recompute occlusion in terms of glyphs in wp_computeFrameRect.
 * don't recompute min/max glyphs per row and col in wp_computeMinMaxFrames.
 * the changes to glybox should have been that isTextRegion is not called
 * if the possible region is very small (what seems to be 4x16 frames
 * with <= 1 line of pad).
 *
 * Revision 1.13  1994/07/29  00:34:32  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * rewrite eddComputeLayout to pick the most efficient rectangle that it
 * can given the sizing constraints.
 *
 * Revision 1.12  1994/06/30  18:57:19  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * in wp_doHorizExpansion, currRow and framesPlaced are UInt32 not Int32
 *
 * Revision 1.11  1994/05/27  02:04:17  jcraig
 * jcraig on Thu May 26 19:03:29 PDT 1994
 *
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * no pLayout parameter to eddiAdjustEncoding and eddiGetEncodingSize.
 *
 * Revision 1.10  1994/05/24  23:27:43  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * rework horizontal first, vertical first and aspect ratio fill.
 * allow occlusions given a x, y distance (in micas) from the upper left
 * corner of the rectangle and a width and height (also in micas).
 *
 * Revision 1.9  1994/04/15  02:23:01  jcraig
 * layout.c -> /product/ipcore/ipwrap/src/RCS/layout.c,v
 *
 * rework eddComputeLayout so that if there are less than (in either
 * the horizontal or vertical direction) 3 frames, 8x8 frames are used
 * instead of 16x16 frames.
 *
 */
