/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   ldutl.c:
 *
 *	Utility routines for Lossy Dither compression.
 *	
 *		
 *	w_getSizeOfLDCompressionInfo():		returns size of memory block
 *						needed for opaque state block
 *						used by LD
 *						compression.
 *
 *	w_getSizeOfLDDecompressionInfo():	returns size of memory block
 *						needed for opaque state block
 *						used by LD
 *						decompression.
 *
 *	w_initializeLDCompressionInfo():	initializes opaque LD
 *						state block
 *
 *	w_initializeLDDecompressionInfo():	initializes opaque LD
 *						state block
 *
 *	w_writeLDEOL():				write EndOfLine code to
 *						pipeline
 *
 *	w_flushLD():				flushes LD compression
 *						pipeline
 *
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "ld.pub"
#include "shrcmprs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: ldutl.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

/* This file contains wrappers for routines used to compress images
 * into Lossy Dither bytestreams.
 *
 * The compressed data stream is always in the natural byte order of the
 * current platform.
 *
 * To avoid interactions with client memory allocators, we allow the
 * client to allocate and free all memory structures.  The routine
 * w_getSizeOfLDCompressionInfo is used by the client to get the
 * size of the opaque state block used in LD compression.
 * Similarly, w_getSizeOfLDDecompressionInfo returns the size of the
 * opaque state block used in decompression.
 *
 * The routines w_initializeLDCompressionInfo and
 * w_initializeLDDecompressionInfo are used to initialize these state blocks.
 * They sets up the things that you wouldn't want to calculate on every
 * compression call.
 *
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Returns size of private Lossy Dither compression state block.
 *
 * LongDescription:::
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform LD
 *	compression.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block
 *  w::              Width of image not including frame.
 *
 * Output:::
 *  pSize::          Pointer to returned size in bytes.
 *
 * Notes:::
 *  LD compression uses a private state block that is allocated either
 *  locally or by the client.  This procedure returns the size of this
 *  block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSizeOfLDCompressionInfo (
    IPWState           *pIPWState,
    UInt32              w,
    UInt32             *pSize)
{
    PROLOGUE;

    ERR_GEN( pSize == NULL, cErrNullPtr);

    ERR_PASS( i_getSizeOfLDCompressionInfo (pIPWState,
					    w + 2*cFrameBits, pSize));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Returns size of private Lossy Dither decompression state block.
 *
 * LongDescription:::
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform LD
 *	decompression.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block
 *  w::              Width of image not including frame.
 *
 * Output:::
 *  pSize::          Pointer to returned size in bytes.
 *
 * Notes:::
 *  LD decompression uses a private state block that is allocated either
 *  locally or by the client.  This procedure returns the size of this
 *  block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how Modified Huffman
 *  compression really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSizeOfLDDecompressionInfo (
    IPWState           *pIPWState,
    UInt32              w,
    UInt32             *pSize)
{
    PROLOGUE;

    ERR_GEN( pSize == NULL, cErrNullPtr);

    ERR_PASS( i_getSizeOfLDDecompressionInfo (pIPWState,
					      w + 2*cFrameBits, pSize));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize Lossy Dither compression state block.
 *
 * LongDescription:::
 *  This procedure fills in a previously allocated state block with
 *  the information required for LD compression.
 *
 * Input:::
 *  pIPWState::         Pointer to standard IPWrap state block.
 *  pLDCompInfoIn::     Pointer to private LD state block.
 *  w::                 Width of image to be compressed not including frame.
 *  compInfoBufSize::   Size of the private state block in bytes.
 *
 * Returns:::
 *  In addition to the normal codes, this routine returns cErrInvalidParm if
 *  the private structure's size (compInfoBufSize) is too small.
 *
 * Notes:::
 *  LD compression uses a private state block that is
 *  allocated either locally or by the client.  This procedure fills
 *  in the contents of this block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initializeLDCompressionInfo (
    IPWState           *pIPWState,
    void               *pLDCompInfoIn,
    UInt32              w,
    UInt32              compInfoBufSize)
{
    PROLOGUE;

    ERR_GEN( pLDCompInfoIn == NULL, cErrNullPtr);

    ERR_PASS( i_initializeLDCompressionInfo (pIPWState, pLDCompInfoIn,
					     w + 2*cFrameBits,
					     compInfoBufSize));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize Lossy Dither decompression state block.
 *
 * LongDescription:::
 *  This procedure fills in a previously allocated state block with
 *  the information required for LD decompression.
 *
 * Input:::
 *  pIPWState::           Pointer to standard IPWrap state block.
 *  pLDDecompInfoIn::     Pointer to private LD state block.
 *  w::                   Width of image to be compressed not including frame.
 *  decompInfoBufSize::   Size of the private state block in bytes.
 *
 * Returns:::
 *  In addition to the normal codes, this routine returns cErrInvalidParm if
 *  the private structure's size (decompInfoBufSize) is too small.
 *
 * Notes:::
 *  LD decompression uses a private state block that is
 *  allocated either locally or by the client.  This procedure fills
 *  in the contents of this block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initializeLDDecompressionInfo (
    IPWState           *pIPWState,
    void               *pLDDecompInfoIn,
    UInt32              w,
    UInt32              decompInfoBufSize)
{
    PROLOGUE;

    ERR_GEN( pLDDecompInfoIn == NULL, cErrNullPtr);

    ERR_PASS( i_initializeLDDecompressionInfo (pIPWState, pLDDecompInfoIn,
					       w + 2*cFrameBits,
					       decompInfoBufSize));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Flushes the Lossy Dither compression pipeline to the compressed data stream.
 *
 * LongDescription:::
 *	The LD compression routines maintain a pipeline used to
 *	accumulate codes into a bytestream.  This routine is used to
 *	flush the bits remaining in the pipeline to the compressed data
 *	stream.  It is used during LD compression when the
 *	pipeline is not flushed at the end of individual scan lines.
 *
 * Input:::
 *  pIPWState::		    Pointer to standard IPWrap state block
 *  pLDCompInfoIn::	    Pointer to private LD state block.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    written.  If several calls
 *			    to this routine are needed to create
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *
 * Output:::
 *  pBytesWritten::         Number of compressed bytes actually
 *			    written to the compressed data
 *			    block.  If flushPipeline is false,
 *			    some bytes required to complete
 *			    a scan line may still be in the
 *			    pipeline and are not included in
 *			    this count.
 *
 * Notes:::
 *  LD compression uses a private state block that is allocated either
 *  locally or by the client.  The pipeline is contained in that state
 *  block.  The pipeline is normally flushed at the end of a strip or tile,
 *  or at the end of the entire image if that was written in one block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_flushLD (
    IPWState           *pIPWState,
    void               *pLDCompInfoIn,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32             *pBytesWritten)
{
    PROLOGUE;

    ERR_GEN( pLDCompInfoIn == NULL || pCompData == NULL ||
	     pBytesWritten == NULL, cErrNullPtr);

    ERR_PASS( i_flushLD (pIPWState, pLDCompInfoIn, pCompData, compDataBufSize,
			 (Int32 *)pBytesWritten));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Writes an EndOfLine code to a Lossy Dither bytestream.
 *
 * LongDescription:::
 *	The LD compression routines maintain a pipeline used to
 *	accumulate codes into a bytestream.  This routine is used to
 *	insert an EndOfLine code into this bytestream.
 *
 * Input:::
 *  pIPWState::		    Pointer to standard IPWrap state block
 *  pLDCompInfoIn::	    Pointer to private LD state block.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    written.  If several calls
 *			    to this routine are needed to create
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *
 * Output:::
 *  pBytesWritten::         Number of compressed bytes actually
 *			    written to the compressed data
 *			    block.  If flushPipeline is false,
 *			    some bytes required to complete
 *			    a scan line may still be in the
 *			    pipeline and are not included in
 *			    this count.
 *
 * Notes:::
 *  LD compression uses a private state block that is allocated either
 *  locally or by the client.  The pipeline is contained in that state
 *  block.  This routine may be used to insert an EOL at the end of the
 *  entire strip or tile.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeLDEOL (
    IPWState           *pIPWState,
    void               *pLDCompInfoIn,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32             *pBytesWritten)
{
    PROLOGUE;

    ERR_GEN( pLDCompInfoIn == NULL || pCompData == NULL ||
	     pBytesWritten == NULL, cErrNullPtr);

    ERR_PASS( i_writeLDEOL (pIPWState, pLDCompInfoIn, 
			    pCompData, compDataBufSize,
			    (Int32 *)pBytesWritten));
    
    NORMAL_RETURN;
}



/*

 $Log: ldutl.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.2  1998/01/14 00:13:51  ddavies
 ldutl.c -> /product/ipcore/ipwrap/src/RCS/ldutl.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.1  1997/03/17 20:13:12  ddavies
 Initial revision


 */


