/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/



/*
 *  logicops.c--basic library operations
 *
 *           w_logicOp()	six operations: union, intersect, xor,
 *                              subtraction, copy, invert
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "geomadt.h"
#include "pixr.h"
#include "logicops.h"

/* PROTOTYPES */
#include "props.pub"
#include "rast.pub"
#include "logicops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: logicops.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Combine two images using a logic function.
 *
 * LongDescription:::
 *  This function uses w_rasterop to combine two images.  It does not
 *  shift the images, it just combines them.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to one of the source images.
 *  op::             Used to specify which operation should be applied.
 *		     Legal choices are found in ipwrap/include/logicops.h.
 *		     They include UNION (same as OR), INTERSECTION
 *		     (same as AND), XOR, SUBTRACT (~prS & prD), COPY
 *		     and INVERT.  This is what happens when you let
 *		     people who excelled in the New Math name things.
 *
 * Output:::
 *  prD::            Pointer to the second input image, which also serves
 *		     as the output image.  This pointer cannot be NULL.
 *
 * Notes:::
 *  This function is commonly used to mask images, for example.  You can
 *  think of it as a pleasant way to call w_rasterop.<BR>
 *  Note that for subtraction, prS is subtracted from prD.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_logicOp (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               op,
    PIXR               *prD)
{
    PROLOGUE;
    Int32         ws, hs, wd, hd, ds, dd, chS, chD;

    ERR_GEN(prS == NULL || prD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &ds));
    ERR_PASS( w_pixrGetDepth(pIPWState, prD, &dd));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &ws));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hs));
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &wd));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &hd));

    ERR_GEN(dd != ds, cErrSrcDimensNotEqual);

    ERR_PASS (w_pixrGetChannelCount (pIPWState, prS, &chS));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prD, &chD));
    ERR_GEN (chS != chD, cErrChannelsNotEqual);

    if (op == UNION) {
	ERR_PASS( w_rasterop(pIPWState, 0, 0, prS, PIX_SRC | PIX_DST, 0, 0,
			     ws, hs, prD));
    }
    else if (op == INTERSECTION) {
	ERR_PASS( w_rasterop(pIPWState, 0, 0, prS, PIX_SRC & PIX_DST, 0, 0,
			     ws, hs, prD));
    }
    else if (op == XOR) {
	ERR_PASS( w_rasterop(pIPWState, 0, 0, prS, PIX_SRC ^ PIX_DST, 0, 0,
			     ws, hs, prD));
    }
    else if (op == SUBTRACT)  /* prD setminus prS */
    {
	ERR_PASS( w_rasterop(pIPWState, 0, 0, prS, PIX_NOT(PIX_SRC) & PIX_DST,
			     0, 0, ws, hs, prD));
    }
    else if (op == COPY)
    {
	if (ws < wd || hs < hd)  /* untouched areas of dest first */
	{
	    ERR_PASS( w_rasterop(pIPWState, 0, 0, NULL, PIX_CLR,
				 ws, 0, wd-ws, hd, prD));
	    ERR_PASS( w_rasterop(pIPWState, 0, 0, NULL, PIX_CLR,
				 0, hs, wd, hd-hs, prD));
	}
	ERR_PASS( w_rasterop(pIPWState, 0, 0, prS, PIX_SRC,
			     0, 0, ws, hs, prD));
    }
    else if (op == INVERT)
	ERR_PASS( w_rasterop(pIPWState, 0, 0, prS, PIX_NOT(PIX_SRC),
			     0, 0, ws, hs, prD));
    else
	ERR_GEN(1, cErrInvalidParm);
    
    NORMAL_RETURN;
}

/*

 $Log: logicops.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.24  1997/06/26 20:04:15  ddavies
 logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.23  1996/05/21 23:34:42  macleod
 logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v

 w_logicOp was only working on single-channel images, whereas
 w_rasterop, on which it is based, works on multiple-channel images.
 Changed so that w_logicOp can take multi-channel images, requiring
 that the source and destination images have the same number of channels.

 Revision 1.22  1996/01/04 20:24:48  ejaquith
 logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.21  1995/12/13  04:10:24  ejaquith
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:13:42  ejaquith
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:02:55  ejaquith
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/10/05  04:27:34  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.17  1995/10/03  01:20:31  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.16  1995/09/21  22:54:34  ejaquith
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/15  18:00:55  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.14  1995/09/11  22:11:33  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.13  1995/09/10  00:33:34  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.12  1995/09/07  01:33:15  ejaquith
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:21:18  ejaquith
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  01:17:02  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.9  1995/08/31  21:15:09  ejaquith
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/17  22:02:48  ejaquith
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.7  1995/07/29  04:43:14  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.6  1994/10/07  19:14:39  lperry
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.5  1993/12/03  01:16:54  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.4  1993/11/06  03:13:56  ddavies
 * logicops.c -> /product/ipcore/ipwrap/src/RCS/logicops.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

