/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  	hausmtch.c:   Image/template matching heuristics using the
 *                    Hausdorff comparator and support routines
 *
 *	Code to find a page description "instance" (a special mark that can
 *	be described in non-bitmap terms) in an image bitmap:
 *		w_matchInstance()
 *
 *	Code to determine "best" Hausdorff model storage type for a page
 *	description "class":
 *		w_pickModelDataType()
 *
 *	Code to check PIXR area surrounding a page description "class" for
 *	sparseness:
 *  
 *		w_sparseRimsCheck()
 *
 */

#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "hauscomp.h"
#include "descrip.h"
#include "alplib.h"
#include "frames.pub"
#include "morphg.h"
#include "hauscall.pub"
#include "defines.pub"
#include "match.h"
#include "hausconf.pub"
#include "logicops.h"
#include "iaerror.pub"

/* prototypes */
#include "shrcc.pub"	/* i_findConnectedComponentBoundingBoxes() */
#include "shrip.pub"	/* i_countBits() */
#include "morph.pub"
#include "hauscomp.pub"
#include "props.pub"
#include "rast.pub"
#include "shrutils.pub"
#include "strel.pub"
#include "logicops.pub"
#include "intops.pub"
#include "match.pub"
#include "rectops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: match.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

#define DEBUG_VERIFY          0  /* 1 to write out a debugging PIXRs */



 

/* The dilation STREL for use with Hausdorff routines */
#define HAUSDORFF_STREL     (strel3)


#if DEBUG_VERIFY
int nDebug = 0;
char debugOutName[80];
#endif /* DEBUG_VERIFY */



typedef struct verifyType 
{
    classType     *class;
    Int32          wImage;
    Int32          hImage;
} verifyType;





/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determine "best" Hausdorff model storage type for a page description class
 *
 * LongDescription:::
 *  Uses the PIXR dimensions as well as some basic class information to
 *  determine a good Hausdorff model type to use in w_makeModelSet().<BR><BR>
 *
 *  This is for **classes**, thus not necessarily the best heuristic for
 *  all applications.  The basic thing about classes is that they are
 *  quite sparse -- usually just thin segments around the outsides of the
 *  bitmaps.
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to the PIXR	that contains or will contain the
 *                      rendered bitmap that represents the class.
 *                      (Note the PIXR just needs to be set up and can be
 *                       blank).
 *  wThick::            The "width" of the thickness for the vertical line
 *                      segments of the class bitmap.  Must be > 0.
 *  hThick::            The "height" of the thickness for the horizontal line
 *                      segments of the class bitmap.  Must be > 0.
 *
 * Output:::
 *  pType::             Ptr to the number of bitmaps to use in
 *                      w_makeModelSet(), the "type" of the model set to use.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Hausdorff
 *
 * Keywords:::
 *  Hausdorff, template matching, page description
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pickModelDataType (
    IPWState *pIPWState,
    PIXR     *pixr,
    Int32     wThick,
    Int32     hThick,
    UInt8    *pType)
{
    PROLOGUE;
    UInt32   size;
    Int32    w, h;

    ERR_GEN(pType == NULL, cErrNullPtr);
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* Keep anal compilers happy */
    wThick = wThick;
    
    /* Check how much space it would take to use the fastest 32 bitmaps model
     * data format.
     */
    ERR_PASS(w_pixrGetHeight (pIPWState, pixr, &h));
    ERR_PASS(w_getModelSetSize (pIPWState, pixr, h+1 /*blockHeight*/,
				32, &size));

    /* If the size required to store it in the fastest data format is
     * less than 256K bytes, then use it
     */
    if (size < 262144)
    {
	*pType = 32;
    }
    else
    {
	/* If the width is significantly wide (many longwords to construct each
	 * row) then we know the model will be very very sparse, so use the
	 * data format optimal for this
	 */
	ERR_PASS(w_pixrGetWidth (pIPWState, pixr, &w));
	
	if ((w / 32) > 25  &&  h > 10 * hThick)
	    *pType = 0;
	else
	{
	    /* Else, 32 models format uses too much space, and while the model
	     * is big, later row ordering using this format should be 
	     * faster than the list of points format.
	     */
	    *pType = 1;
	}
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check PIXR area surrounding a page description "class" for sparseness
 *
 * LongDescription:::
 *  Given the specifications of a page description "class" at some location
 *  in a PIXR, look at the four subrectangles that would surround the
 *  class' bitmap rectangle (we'll call them rims here) and see if they
 *  are sparse (have few "on" bits).
 *
 *  <PRE>
 *  It will check these four rims:
 *        +-------------+       -
 *        | top rim     |       | rim dimension is determined as multiple
 *        |             |       |            of class's hthickness
 *        +-------------+       -  -
 *                                 | skip dimension is determined as
 *   +--+ +-------------+ +--+     -        multiple of class's hthickness
 *   |le| | class       | |ri|
 *   | f| |             | | g|
 *   | t| |             | |th|
 *   +--+ +-------------+ +--+
 *         
 *        +-------------+
 *        | bottom rim  |
 *        |             |
 *        +-------------+
 *  The location and height of the top and bottom rims (the width of the
 *  left and right) are determined as a multiple of the class's hthickness
 *  (wthickness).
 *  The width of the top and bottom rims (the height of the left and right)
 *  is the same as that of the instance.
 *
 *  Go ahead and read the parameter descriptions now and then read this:
 *   Eg:  If checkRims == xxxx1010, checkAll4 ==1 on input, it means check
 *        the top and left rims for sparseness.  On output:
 *        If *pResRims == 00000010, it means the left rim NOT sparse.
 *        If *pResRims == 00001010, it means the left and top rims NOT sparse.
 *        If *pResRims == 00000000, it means left and top rims BOTH sparse.
 *  </PRE>
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  checkRims::
 *    Indicates which of the four rims to check for sparseness by whether
 *    or not each of the LSB (rightmost) four bits are on.  If the MSB of the
 *    4 bits is "1", it means to check the top rim;  the next bit, check the
 *    bottom;  check the left;  the last (LSB) bit, check the right.
 *    The MSB (leftmost) four bits are ignored.
 *  checkAll4::         cFalse to return immediately once a rim fails
 *                      (is not found to be sparse); cTrue to check all
 *                      4 rims regardless
 *  sparseRimMaxPercent::  The maximum percent of pixels in each rim that can
 *                         be "on" and still be considered sparse.
 *                         Range [0..1]
 *  classw::            The positive width of the class' bitmap
 *  classh::            The positive height of the class' bitmap
 *  classwThick::       The "width" of the thickness for the vertical line
 *                      segments of the class bitmap.  Must be > 0.
 *  classhThick::       The "height" of the thickness for the horizontal line
 *                      segments of the class bitmap.  Must be > 0.
 *  pixr::              Ptr to binary PIXR
 *  x::                 Upper left corner location where class sits
 *  y::                 in PIXR
 *
 * Output:::
 *  pResRims::
 *    Ptr to sparse checking results in LSB (rightmost) four bits;
 *    the MSB (leftmost) four bits will always be 0.  The four LSB bits will
 *    indicate which of the four rims are sparse with "0", in a manner similar
 *    to the passed in "checkRims".
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page description
 *
 * Keywords:::
 *  page description, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_sparseRimsCheck (
    IPWState    *pIPWState,
    UInt8        checkRims,
    Bool         checkAll4,
    Float32      sparseRimMaxPercent,
    Int32        classw,
    Int32        classh,
    Int32        classwThick,
    Int32        classhThick,
    PIXR        *pixr,
    Int32        x,
    Int32        y,
    UInt8       *pResRims)
{
    PROLOGUE;
    Int32     cnt;
    Int32     wRim;
    Int32     hRim;
    Int32     wSkip;
    Int32     hSkip;
    Int32     bpl;        /* Bytes per line of image */
    Int32     wPixr;        /* Dimensions of image, excluding frame and pad */
    Int32     hPixr;
    UInt8     resRims = 0;
    UInt8    *pImage;        /* Ptr to upper left image corner, excluding
			      * frame */
    Int32     hRimMaxCnt; /* Maximum denseness a horizontal rim may be
			   * that we'll still consider sparse */
    Int32     vRimMaxCnt; /* Maximum denseness a vertical rim may be
			   * that we'll still consider sparse */
#if DEBUG_SPARSE
    RECTANGL *rect = NULL;
#endif /*DEBUG_SPARSE*/

    ERR_GEN(pResRims == NULL, cErrNullPtr);
    ERR_GEN(sparseRimMaxPercent < 0. ||  sparseRimMaxPercent > 1.,
	    cErrInvalidParm);
    ERR_GEN(classw <= 0  ||  classh <= 0  ||
	    classwThick <= 0  ||  classhThick <= 0, 
	    cErrInvalidParm);
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);



    if (!(checkRims & 0x0F))
    {
	/* Nothing to check! */
	*pResRims = 0;
	NORMAL_RETURN;
    }

    hRimMaxCnt = ROUND(((RIM_THICKNESS_MULTIPLE * classhThick) * classw) *
		       sparseRimMaxPercent);
    vRimMaxCnt = ROUND(((RIM_THICKNESS_MULTIPLE * classwThick)* classh) *
		       sparseRimMaxPercent);

    


    /* NOTICE!!  PIXR MUST be framed, and wRim and hRim must be less than
     * or equal to the frame size so we don't run off the PIXR bitmap.
     */
    wRim  = classwThick * RIM_THICKNESS_MULTIPLE;
    hRim  = classhThick * RIM_THICKNESS_MULTIPLE;
    wSkip = classwThick * SKIP_THICKNESS_MULTIPLE;
    hSkip = classhThick * SKIP_THICKNESS_MULTIPLE;


    ERR_PASS(w_pixrGetImage (pIPWState, pixr, &pImage));
    ERR_PASS(w_pixrGetWidth (pIPWState, pixr, &wPixr));
    ERR_PASS(w_pixrGetHeight (pIPWState, pixr, &hPixr));
    ERR_PASS(w_pixrGetBpl (pIPWState, pixr, &bpl));



    /*
     *  It will check these four rims:
     *        +-------------+       -
     *        | top rim     |       | rim dimension is determined as multiple
     *        |             |       |            of class's hthickness
     *        +-------------+       -  -
     *                                 | skip dimension is determined as
     *   +--+ +-------------+ +--+     -        multiple of class's hthickness
     *   |le| | match       | |ri|
     *   | f| |             | | g|
     *   | t| |             | |th|
     *   +--+ +-------------+ +--+
     *         
     *        +-------------+
     *        | bottom rim  |
     *        |             |
     *        +-------------+
     * The location and height of the top and bottom rims (the width of the
     * left and right) are determined as a multiple of the class's hthickness
     * (wthickness).
     * The width of the top and bottom rims (the height of the left and right)
     * is the same as that of the match class.
     */

    if (checkRims & 0x01)
    {
	/* This bit means that we should check if the right rim 
	 * is completely or nearly sparse.
	 */
	if (x + classw + wSkip <= wPixr - wRim)
	{
	    /* Check the right rim */
	    ERR_PASS(i_countBits (pIPWState, pImage,
				  x + classw + wSkip, y, wRim, classh,
				  1, bpl, &cnt));
#if DEBUG_SPARSE
	    ERR_PASS(w_createRect(pIPWState,
				  x + classw + wSkip, y, wRim, classh, &rect));
	    fprintf(fp_out,
		    "Rect: %4ld,%4ld,%4ld,%4ld    Right rim: %4ld,%4ld,%4ld,%4ld   On bits:%4ld\n",
		    x, y, classw, classh, 
		    rect->x, rect->y, rect->w, rect->h,cnt);
	    if (THEDEBUGPIXR != NULL)
		ERR_PASS(w_pixrPutRect(pIPWState, THEDEBUGPIXR, rect, 1, 1));
	    ERR_PASS(w_destroyRect(pIPWState, &rect));
#endif /*DEBUG_SPARSE*/
	    if (cnt > vRimMaxCnt)
	    {
		resRims |= 0x01;
		if (!checkAll4) 
		{
		    *pResRims = resRims;
		    NORMAL_RETURN;
		}
	    }
	}
	else if (x + classw + wSkip < wPixr)
	{
	    /* Check reduced right rim with appropriately scaled maxCnt */
	    Int32   lx = x + classw + wSkip;
	    Int32   rx = MIN(lx + wRim - 1, wPixr-1);

	    ERR_PASS(i_countBits (pIPWState, pImage,
				  lx, y, rx-lx+1, classh, 
				  1, bpl, &cnt));
#if DEBUG_SPARSE
	    ERR_PASS(w_createRect(pIPWState,
				  lx, y, rx-lx+1, classh, &rect));
	    fprintf(fp_out,
		    "Rect: %4ld,%4ld,%4ld,%4ld    Right rim: %4ld,%4ld,%4ld,%4ld   On bits:%4ld\n",
		    x, y, classw, classh, 
		    rect->x, rect->y, rect->w, rect->h, cnt);
	    if (THEDEBUGPIXR != NULL)
		ERR_PASS(w_pixrPutRect(pIPWState, THEDEBUGPIXR, rect, 1, 1));
	    ERR_PASS(w_destroyRect(pIPWState, &rect));
#endif /*DEBUG_SPARSE*/
	    if (cnt > ROUND(((rx-lx+1) * classh) * sparseRimMaxPercent))
	    {
		resRims |= 0x01;
		if (!checkAll4)
		{
		    *pResRims = resRims;
		    NORMAL_RETURN;
		}
	    }
	}
    }
    
    
    if (checkRims & 0x04)
    {
	/* This bit means that we should check if the bottom rim 
	 * is completely or nearly sparse.
	 */
	if (y + classh + hSkip <= hPixr - hRim)
	{
	    /* Check the bottom rim */
	    ERR_PASS(i_countBits (pIPWState, pImage,
				  x, y + classh + hSkip, classw, hRim, 
				  1, bpl, &cnt));
#if DEBUG_SPARSE
	    ERR_PASS(w_createRect(pIPWState,
				  x, y + classh + hSkip, classw, hRim, &rect));
	    fprintf(fp_out,
		    "Rect: %4ld,%4ld,%4ld,%4ld    Bottm rim: %4ld,%4ld,%4ld,%4ld   On bits:%4ld\n",
		    x, y, classw, classh, 
		    rect->x, rect->y, rect->w, rect->h, cnt);
	    if (THEDEBUGPIXR != NULL)
		ERR_PASS(w_pixrPutRect(pIPWState, THEDEBUGPIXR, rect, 1, 1));
	    ERR_PASS(w_destroyRect(pIPWState, &rect));
#endif /*DEBUG_SPARSE*/
	    if (cnt > hRimMaxCnt)
	    {
		resRims |= 0x04;
		if (!checkAll4)
		{
		    *pResRims = resRims;
		    NORMAL_RETURN;
		}
	    }
	}
	else if (y + classh + hSkip < hPixr)
	{
	    /* Check reduced bottom rim with appropriately scaled maxCnt */
	    Int32   ty = y + classh + hSkip;
	    Int32   by = MIN(ty + hRim - 1, hPixr-1);

	    ERR_PASS(i_countBits (pIPWState, pImage,
				  x, ty, classw, by-ty+1,
				  1, bpl, &cnt));
#if DEBUG_SPARSE
	    ERR_PASS(w_createRect(pIPWState,
				  x, ty, classw, by-ty+1, &rect));
	    fprintf(fp_out,
		    "Rect: %4ld,%4ld,%4ld,%4ld    Bottm rim: %4ld,%4ld,%4ld,%4ld   On bits:%4ld\n",
		    x, y, classw, classh, 
		    rect->x, rect->y, rect->w, rect->h, cnt);
	    if (THEDEBUGPIXR != NULL)
		ERR_PASS(w_pixrPutRect(pIPWState, THEDEBUGPIXR, rect, 1, 1));
	    ERR_PASS(w_destroyRect(pIPWState, &rect));
#endif /*DEBUG_SPARSE*/
	    if (cnt > ROUND(((by-ty+1) * classw) * sparseRimMaxPercent))
	    {
		resRims |= 0x04;
		if (!checkAll4)
		{ 
		    *pResRims = resRims;
		    NORMAL_RETURN;
		}
	    }
	}
    }
    

    if (checkRims & 0x02)
    {
	/* This bit means that we should check if the left rim 
	 * is completely or nearly sparse.
	 */
	if (x - wRim - wSkip >= 0)
	{
	    /* Check the left rim */
	    ERR_PASS(i_countBits (pIPWState, pImage,
				  x - wRim - wSkip, y, wRim, classh,
				  1, bpl, &cnt));
#if DEBUG_SPARSE
	    ERR_PASS(w_createRect(pIPWState,
				  x - wRim - wSkip, y, wRim, classh, &rect));
	    fprintf(fp_out,
		    "Rect: %4ld,%4ld,%4ld,%4ld    Left  rim: %4ld,%4ld,%4ld,%4ld   On bits:%4ld\n",
		    x, y, classw, classh, 
		    rect->x, rect->y, rect->w, rect->h, cnt);
	    if (THEDEBUGPIXR != NULL)
		ERR_PASS(w_pixrPutRect(pIPWState, THEDEBUGPIXR, rect, 1, 1));
	    ERR_PASS(w_destroyRect(pIPWState, &rect));
#endif /*DEBUG_SPARSE*/
	    if (cnt > vRimMaxCnt)
	    {
		resRims |= 0x02;
		if (!checkAll4)
		{
		    *pResRims = resRims;
		    NORMAL_RETURN;
		}
	    }
	}
	else if (x - wSkip - 1 >= 0)
	{
	    /* Check reduced left rim with appropriately scaled maxCnt */
	    /* lx is 0 here! */
	    Int32   rx = x - wSkip - 1;

	    ERR_PASS(i_countBits (pIPWState, pImage,
				  0, y, rx+1, classh,
				  1, bpl, &cnt));
#if DEBUG_SPARSE
	    ERR_PASS(w_createRect(pIPWState,
				  0, y, rx+1, classh, &rect));
	    fprintf(fp_out,
		    "Rect: %4ld,%4ld,%4ld,%4ld    Left  rim: %4ld,%4ld,%4ld,%4ld   On bits:%4ld\n",
		    x, y, classw, classh, 
		    rect->x, rect->y, rect->w, rect->h, cnt);
	    if (THEDEBUGPIXR != NULL)
		ERR_PASS(w_pixrPutRect(pIPWState, THEDEBUGPIXR, rect, 1, 1));
	    ERR_PASS(w_destroyRect(pIPWState, &rect));
#endif /*DEBUG_SPARSE*/
	    if (cnt > ROUND(((rx+1) * classh) * sparseRimMaxPercent))
	    {
		resRims |= 0x02;
		if (!checkAll4)
		{
		    *pResRims = resRims;
		    NORMAL_RETURN;
		}
	    }
	}
    }
    

    if (checkRims & 0x08)
    {
	/* This bit means that we should check if the top rim 
	 * is completely or nearly sparse.
	 */
	if (y - hRim - hSkip >= 0)
	{
	    /* Check the top rim */
	    ERR_PASS(i_countBits (pIPWState, pImage,
				  x, y - hRim - hSkip, classw, hRim,
				  1, bpl, &cnt));
#if DEBUG_SPARSE
	    ERR_PASS(w_createRect(pIPWState,
				  x, y - hRim - hSkip, classw, hRim, &rect));
	    fprintf(fp_out,
		    "Rect: %4ld,%4ld,%4ld,%4ld    Top   rim: %4ld,%4ld,%4ld,%4ld   On bits:%4ld\n",
		    x, y, classw, classh, 
		    rect->x, rect->y, rect->w, rect->h, cnt);
	    if (THEDEBUGPIXR != NULL)
		ERR_PASS(w_pixrPutRect(pIPWState, THEDEBUGPIXR, rect, 1, 1));
	    ERR_PASS(w_destroyRect(pIPWState, &rect));
#endif /*DEBUG_SPARSE*/
	    if (cnt > hRimMaxCnt)
	    {
		resRims |= 0x08;
		if (!checkAll4)
		{
		    *pResRims = resRims;
		    NORMAL_RETURN;
		}
	    }
	}
	else if (y - hSkip - 1 >= 0)
	{
	    /* Check reduced top rim with appropriately scaled maxCnt */
	    /* ty is 0 here! */
	    Int32   by = y - hSkip - 1;

	    ERR_PASS(i_countBits (pIPWState, pImage, 
				  x, 0, classw, by+1,
				  1, bpl, &cnt));
#if DEBUG_SPARSE
	    ERR_PASS(w_createRect(pIPWState,
				  x, 0, classw, by+1, &rect));
	    fprintf(fp_out,
		    "Rect: %4ld,%4ld,%4ld,%4ld    Top   rim: %4ld,%4ld,%4ld,%4ld   On bits:%4ld\n",
		    x, y, classw, classh, 
		    rect->x, rect->y, rect->w, rect->h, cnt);
	    if (THEDEBUGPIXR != NULL)
		ERR_PASS(w_pixrPutRect(pIPWState, THEDEBUGPIXR, rect, 1, 1));
	    ERR_PASS(w_destroyRect(pIPWState, &rect));
#endif /*DEBUG_SPARSE*/
	    if (cnt > ROUND(((by+1) * classw) * sparseRimMaxPercent))
	    {
		resRims |= 0x08;
		if (!checkAll4)
		{
		    *pResRims = resRims;
		    NORMAL_RETURN;
		}
	    }
	}
    }
    

    /* All rims have been checked for sparseness! */
    *pResRims = resRims;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  i_findConnectedComponentBoundingBoxes' callback in wp_notVerifiable()
 *
 * LongDescription:::
 *  Callback routine for whenever i_findConnectedComponentBoundingBoxes()
 *  finds a connected component ("CC"), where
 *  i_findConnectedComponentBoundingBoxes() is called from wp_notVerifiable().
 *
 * PrivateDescription:::
 *  See wp_notVerifiable() for why we are interested in CCs.
 *  We inspect what type of page description instance (see w_matchInstance())
 *  we are looking for, and base our criterion on whether or not the
 *  returned CC is too big for it.  We are trying to ensure that large portions
 *  of the instance's bitmap are not not overlapping some "on" bits in another
 *  bitmap (because we are trying to Hausdorff match the instance in another
 *  bitmap!)
 *
 * Input:::
 *  pIPWState::		Ptr to IPCore state
 *  data::              Ptr to our information
 *  x::                 Location of the CC, in the frame of the ENTIRE PIXR
 *  y::                 including frame. ie: (0,0)==(cFrameBits,cFrameBits)
 *  w::                 width of CC
 *  h::                 height of CC
 *  m::                 mass of CC
 *  nRuns::             number of runs in the CC
 *  px::                point on the CC (to be used for seed fills normally)
 *  py::                (point y coordinate)
 *
 * Returns:::
 *
 * Notes:::
 *  Non-standard Return value!!::
 *   Now here's the interesting part, we return nonzero in this routine
 *   to trigger the CC finding routine to barf and die immediately.
 *   wp_notVerifiable() knows what to do...
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  template matching, page description, comparison
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static Int32 CDECL
wp_isCCTooBig (
    IPWState	   *pIPWState,
    void           *data,
    Int32           x,
    Int32           y,
    Int32           w,
    Int32           h,
    Int32           m,
    Int32           nRuns,
    Int32           px,
    Int32           py)
{
    verifyType  *veri = (verifyType *)data;
    Bool         tooBig;

    /* Avoid compiler warnings */
    pIPWState = pIPWState;
    px = px;
    py = py;
    nRuns = nRuns;

    /* We know that for any invocation for this routine :
     * w >= class->wThick
     * h >= class->hThick
     *
     * Thus, the CC is considered fairly big, and most likely "too big".
     */
    switch (veri->class->kind)
    {
    case CHECKBOX_BOTRITE:
    case CHECKBOX_BOTRITE_CHECKED:
	/* Model is a CHECKBOX, like a horizontally flipped "L".
	 * If this CC is at the very top right or the very bottom left,
	 * we assume it is just the model being rendered slightly
	 * too long in a dimension.  Give it up to twice the rendered
	 * thickness for being off.
	 */

	/* If it isn't either corner, or this CC at the corner is too big
	 * to just be slop from a model being rendered slightly too
	 * big, then fail --> Verification method failed!
	 * Cause i_findCC..() to abort early with a callback "error".
	 */
	if (x == cFrameBits)
	    tooBig = (w > 2*veri->class->wThick ? cTrue : cFalse);
	else if (y == cFrameBits)
	    tooBig = (h > 2*veri->class->hThick ? cTrue : cFalse);
	else
	    tooBig = cTrue;
	break;
    case CHECKBOX_BOTLEFT:
    case CHECKBOX_BOTLEFT_CHECKED:
	/* Like above case, except we have to check a different corner. */
	if (y == cFrameBits)
	    tooBig = (h > 2*veri->class->hThick ? cTrue : cFalse);
	else if (x - cFrameBits + w == veri->wImage)
	    tooBig = (w > 2*veri->class->wThick ? cTrue : cFalse);
	else
	    tooBig = cTrue;
	break;
    case CHECKBOX_TOPRITE:
    case CHECKBOX_TOPRITE_CHECKED:
	/* Like above case, except we have to check a different corner. */
	if (x == cFrameBits)
	    tooBig = (w > 2*veri->class->wThick ? cTrue : cFalse);
	else if (y - cFrameBits + w == veri->wImage)
	    tooBig = (h > 2*veri->class->hThick ? cTrue : cFalse);
	else
	    tooBig = cTrue;
	break;
    case CHECKBOX_TOPLEFT:
    case CHECKBOX_TOPLEFT_CHECKED:
	/* Like above case, except we have to check a different corner. */
	if (x - cFrameBits + w == veri->wImage)
	    tooBig = (w > 2*veri->class->wThick ? cTrue : cFalse);
	else if (y - cFrameBits + w == veri->wImage)
	    tooBig = (h > 2*veri->class->hThick ? cTrue : cFalse);
	else
	    tooBig = cTrue;
	break;
    case BUBBLE:
    case BUBBLE_CHECKED:
	/* We want to find a "clump" of pixels at least (xthick)x(ythick)
	 * and very dense.  For a bubble, there could be 1 pixel thick
	 * contour CC that has large overall (w)x(h) dimensions.
	 * Thus, there must be at least (w)-(xthick) + (h)-(ythick) pixels
	 * in this CC outside of any such clump.  Approximately scale
	 * down the mass, and see if it is still big enough to be a dense
	 * at least (xthick)x(ythick) clump.
	 */
	m -= (MAX(0, w - veri->class->wThick) +
	      MAX(0, h - veri->class->hThick));
	    
	/* If it's still pretty darn big for a bubble, then fail. */
	tooBig = (m >= veri->class->wThick * veri->class->hThick ?
		  cTrue : cFalse);
	break;
    default:
	/* Verification method failed!
	 * Cause i_findCC..() to abort early with a "callback"
	 * error */
	tooBig = cTrue;
	break;
    }

    return (Int32)tooBig;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Verification technique for page description "instance" potential match
 *
 * PrivateDescription:::
 *  We assume we just found some acceptable location in the image where
 *  we believe the instance matches by using Hausdorff A to B percents
 *  from Hausdorff matching.  This routine
 *  is used in lieu of B to A Hausdorff check, which is not suitable
 *  for finding these types of bitmaps in an image.  However, if we only
 *  passed the A to B Hausdorff check alone, we could be matching any
 *  random black blob on the page.  (See top of file haussrch.c for more
 *  details).  Thus, we need a way to ensure this isn't the case.<BR><BR>
 *
 *  Technique::
 *   <PRE>
 *  -If the bitmap is rather large, we'll assume A to B percents are adequate
 *   and just "rubber stamp" verify it.
 *  -If, when we looked at the instance bitmap on the perfect page, any of
 *   the four rectangles immediately surrounding the bitmap were very sparse
 *   (as indicated by the instances "rowColRims" field), we can look at the
 *   corresponding four rectangles around this match and check to ensure that
 *   they are sparse too.  If they aren't sparse, fail to verify this match.
 *  -When the A to B percents were not extremely high, there may be
 *   "significant" portions of the model's bitmap that don't overlap at this
 *   match location.  Use morphology to determine this, and if there are
 *   significant nonoverlapping portions, fail to verify.
 *   </PRE>
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  class::             Ptr to class we are looking for
 *  instance::          Ptr to instance we are mlooking for
 *  x::                 The location in the "B" image/PIXR where we found
 *  y::                 this A to B Hausdorff match.
 *  clipOffsetX::       The offset the upper left corner of the "B" PIXR was
 *  clipOffsetY::       at as a subPIXR of a larger PIXR.  This is used only
 *                      for outputting statistics to give the match coord.s
 *                      in the coord.s of the _full_ image we are really using.
 *  percentAtoB::       The Hausdorff A to B matching percent at this location
 *                      (calculated by a Hausdorff routine like 
 *                       w_hausdorffMatches()).
 *  BS::                Ptr to the "B" image PIXR (using Hausdorff terms)
 *  BD::                Ptr to the "B" dilated image PIXR (using Hausdorff 
 *                      terms) (which is just "BS" dilated by a STREL).
 *
 * Output:::
 *  pNotVerified::      Ptr to returned value of how it did (not) verify.
 *                      0 if it verifies the match.
 *                      1 if the right  rim is too dense.
 *                      2 if the left   rim is too dense.
 *                      4 if the bottom rim is too dense.
 *                      8 if the top    rim is too dense.
 *                      255 if a significant model portion doesn't overlap.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  page description, template matching, Hausdorff, comparison
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_notVerifiable (
    IPWState     *pIPWState,
    classType    *class,
    instanceType *instance,
    Int32         x,
    Int32         y,
    Int32         clipOffsetX,
    Int32         clipOffsetY,
    Float64       percentAtoB,
    PIXR         *BS,
    PIXR         *BD,
    UInt8        *pNotVerified)
{
    PROLOGUE;
    PIXR        *notover = NULL, *clip = NULL;
    STREL       *pStrel = NULL, *pStrelC = NULL;
    RECTANGL     rect;
    Int32        tmpi, callbackRetCode, w, h;
    UInt8        resRims;
    Bool         zero;

    /* For very large bitmaps, most likely the Hausdorff percent will have
     * taken care of the verification concerns.  Psyche!
     */
    if (VERIFIABLY_BIG(class->w, class->h))
    {
	*pNotVerified = 0;
	NORMAL_RETURN;
    }

    if (instance->rowColRims & 0x0F)
    {
	/* We can check one or more of the rims to see if they are sparse
	 * as they should be.
	 */

	ERR_PASS( w_sparseRimsCheck( pIPWState,
				     instance->rowColRims, cFalse, 
				     SPARSE_RIM_MAX_PERCENT,
				     class->w, class->h,
				     class->wThick, class->hThick,
				     BS,
				     x, y, 
				     &resRims));
	if (resRims != 0)
	{
	    /* Sorry!  that match didn't verify.  */

	    if (PRINTABLE(pIPWState)) {
		i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
			  "JQ: NOT VERIFIED at %4ld, %4ld, percent %8.7f!!",
			  x + clipOffsetX, y + clipOffsetY, percentAtoB);
		if (resRims & 0x01)
		    i_printf (pIPWState,
			      pIPWState->errorProc, pIPWState->errorInfo,
			      " (dense right rim)\n");
		else if (resRims & 0x02)
		    i_printf (pIPWState, pIPWState->errorProc,
			      pIPWState->errorInfo,
			      " (dense left rim)\n");
		else if (resRims & 0x04)
		    i_printf (pIPWState, pIPWState->errorProc,
			      pIPWState->errorInfo,
			      " (dense bottom rim)\n");
		else
		    i_printf (pIPWState, pIPWState->errorProc,
			      pIPWState->errorInfo,
			      " (dense top rim)\n");
	    }

	    *pNotVerified = resRims;
	    NORMAL_RETURN;
	}
    }

    
    /* If the percent match was 100%, then we don't have to bother with
     * checking for large nonoverlapping model portions!
     */
    if (percentAtoB == 1.)
    {
	*pNotVerified = 0;
	NORMAL_RETURN;
    }

    rect.x = x;
    rect.y = y;
    rect.w = class->w;
    rect.h = class->h;

    if (class->wThick > 2 || class->hThick > 2)
    {
	/* Dilate the dilated image by another 1.4! */
	ERR_PASS(w_rectOf (pIPWState, BD, &rect, &clip));
	
	ERR_PASS_CLEANUP(w_clonePixrNoInit (pIPWState, clip, &notover));

	ERR_PASS_CLEANUP(w_getStrel(pIPWState, HAUSDORFF_STREL, &pStrel));
	
	ERR_PASS_CLEANUP(w_dilate (pIPWState, clip, pStrel, &notover));
    }
    else
    {
	ERR_PASS(w_rectOf (pIPWState, BD, &rect, &notover));
	
#if DEBUG_VERIFY
	ERR_PASS_CLEANUP(w_clonePixr (pIPWState, notover, &clip));
#else
	ERR_PASS_CLEANUP(w_clonePixrNoInit (pIPWState, notover, &clip));
#endif /* DEBUG_VERIFY */
    }
    
#if DEBUG_VERIFY
    sprintf(debugOutName, "dil.%d", nDebug);
    ERR_PASS_CLEANUP(w_write (pIPWState, debugOutName, clip, DEBUG_FILETYPE));
#endif /* DEBUG_VERIFY */

    
    /* Overlay the model at the location to verify in the dilated dilated
     * image, and keep only the pixels that are in the model that don't
     * overlay the image */

    /* Clear the clip PIXR, and add the model bitmap */
    ERR_PASS_CLEANUP(w_pixrGetWidth (pIPWState, clip, &w));
    ERR_PASS_CLEANUP(w_pixrGetHeight (pIPWState, clip, &h));
    ERR_PASS_CLEANUP(w_rasterop (pIPWState, 0, 0, NULL, 
				 PIX_CLR, 0, 0, w, h, clip));

    ERR_PASS_CLEANUP(w_logicOp(pIPWState, class->srcPr, OR, clip));

#if DEBUG_VERIFY
    sprintf(debugOutName, "dildil.%d", nDebug);
    ERR_PASS_CLEANUP(w_write (pIPWState, debugOutName, notover,
			      DEBUG_FILETYPE));
    sprintf(debugOutName, "model.%d", nDebug);
    ERR_PASS_CLEANUP(w_write (pIPWState, debugOutName, clip,
			      DEBUG_FILETYPE));
#endif /* DEBUG_VERIFY */
    
    ERR_PASS_CLEANUP(w_invert (pIPWState, notover, &notover));
    ERR_PASS_CLEANUP(w_logicOp(pIPWState, clip, AND, notover));
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &clip));

#if DEBUG_VERIFY
    sprintf(debugOutName, "modelxnotdildil.%d", nDebug);
    ERR_PASS_CLEANUP(w_write (pIPWState, debugOutName, notover, 
			      DEBUG_FILETYPE));
    nDebug++;
#endif /* DEBUG_VERIFY */


    /* If the checkmark is a bubble, we don't want to do an open on it
     * because chances are that it would remove most everything since very
     * often there aren't many (xthick)x(ythick) chunks in the entire
     * regularly rendered checkmark bitmap!
     */
    if (class->kind != BUBBLE && class->kind != BUBBLE_CHECKED)
    {
	/* Perform a morphological open (on the bitmap containing the portion
	 * of the model that is "not over" the dilated dilated image) that
	 * is a brick (xthick)x(ythick).
	 */

	pStrel = NULL;
	if (class->wThick == class->hThick  &&  class->wThick <= 5) {
	    switch (class->wThick)
	    {
	    case 2:
		ERR_PASS_CLEANUP(w_getStrel (pIPWState, strel2, &pStrel));
		break;
	    case 3:
		ERR_PASS_CLEANUP(w_getStrel (pIPWState, strel3, &pStrel));
		break;
	    case 4:
		ERR_PASS_CLEANUP(w_getStrel (pIPWState, strel4, &pStrel));
		break;
	    case 5:
		ERR_PASS_CLEANUP(w_getStrel (pIPWState, strel5, &pStrel));
		break;
	    default:
		/* If 1x1 brick, don't do the open! */
		pStrel = NULL;
	    }

	    if (pStrel != NULL)
		ERR_PASS_CLEANUP(w_open (pIPWState, notover, pStrel,
					 &notover));
	}
	else {
	    ERR_PASS_CLEANUP(w_makeSolidCenterStrel (pIPWState, 
						     class->wThick,
						     class->hThick, 
						     &pStrelC));

	    ERR_PASS_CLEANUP(w_open (pIPWState, notover, pStrelC, &notover));

	    ERR_PASS_CLEANUP(w_destroyStrel (pIPWState, &pStrelC));
	}
    }

    /* Determine if we should fail given what is left in notover and
     * what type of mark it is.
     */
    switch (class->kind)
    {
    case BUBBLE:
    case BUBBLE_CHECKED:
    case CHECKBOX_BOTRITE:
    case CHECKBOX_BOTLEFT:
    case CHECKBOX_TOPRITE:
    case CHECKBOX_TOPLEFT:
    case CHECKBOX_BOTRITE_CHECKED:
    case CHECKBOX_BOTLEFT_CHECKED:
    case CHECKBOX_TOPRITE_CHECKED:
    case CHECKBOX_TOPLEFT_CHECKED:
    {
	verifyType  veri;
	Int32       bpl;
	UInt8      *pImage;

	veri.wImage = w;
	veri.hImage = h;
	veri.class = class;

	ERR_PASS_CLEANUP(w_pixrGetImageAll (pIPWState, notover, &pImage));
	ERR_PASS_CLEANUP(w_pixrGetBpl (pIPWState, notover, &bpl));
	
	/* Look for CCs in this "not model" bitmap and check each one out */
	tmpi = i_findConnectedComponentBoundingBoxes (pIPWState, pImage,
						      bpl,
						      cFrameBits, cFrameBits,
						      w, h,
				                      /* Next 2 params are min
						       * w,h of rects returned*/
						      class->wThick,
						      class->hThick,
						      0, /* 0 == 4cc, 1== 8cc*/ 
						      (void *)(&veri),
						      wp_isCCTooBig,
						      &callbackRetCode);
	/* Now this is interesting.
	 * Given our callback above, we designed it so it would return
    	 * a callback error code if it found a CC that was too big.
	 * Thus, if the above routine completed with a callback error,
	 * it means a CC was too big;  another error means the routine had
	 * a genuine error/failure!
	 */
	ERR_GEN_CLEANUP( tmpi != cSuccessful  &&  tmpi != ia_callbackError,
			 cErrInternal);
	
	tmpi = (tmpi == ia_callbackError ? 1 : 0);
	break;
    }
    default:
	/* For any other checkmark type, we simply check if there are
	 * any ON bits left in notover.  If so, it meant a chunk of the
	 * orignal model was at least (xthick)x(ythick) big and didn't
	 * overlap, so fail verification.
	 */
	ERR_PASS_CLEANUP(w_isZero (pIPWState, notover, &zero));
	tmpi = (zero ? 0 : 1);
	break;
    }
    
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &notover));


    if (tmpi)
    {
	/* Sorry!  that match didn't verify.  */
	if (PRINTABLE(pIPWState)) {
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "JQ: NOT VERIFIED at %4ld, %4ld, percent %8.7f!!",
		      x + clipOffsetX, y + clipOffsetY, percentAtoB);
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      " (large model piece doesn't overlap)\n");
	}

	*pNotVerified = 255;
	NORMAL_RETURN;
    }
    else
    {
	*pNotVerified = 0;
	NORMAL_RETURN;
    }

ERR_CLEANUP:
    w_destroyStrel (pIPWState, &pStrelC);
    w_destroyPixr (pIPWState, &notover);
    w_destroyPixr (pIPWState, &clip);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Fast first effort to find a page description "instance" in an image bitmap
 *
 * PrivateDescription:::
 *  The theory is this.  We only have a guess at where this instance may
 *  be.  We'd like to think that most of the time, we are pretty close.
 *  Thus, it makes sense to do a faster, less exhaustive technique first,
 *  and then follow up (if this doesn't find it) with a slower, more error-
 *  tolerant method like wp_hausdorffMatchIdeal().<BR>
 *  Here, we just go for the best overall match in the subrect using the
 *  comparing the Hausdorff A to B percents as we find might-be-matches,
 *  by using the routine w_best1Callback()
 *  as our Hausdorff callback routine.
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary image to search in.
 *  pMultResInfo::      Ptr to the multi-resolution set-up info, created
 *                      by w_makeMultResInfo().
 *                      If NULL, it means don't use multi-resolution searching,
 *                      because we believe the non-use method will be faster.
 *  clipRect::          Ptr to subrect of pixr where we will search
 *                      for matches.
 *  minPerc::           Any match found must have a Hausdorff A to B percent
 *                      of at least this much, [0..1].
 *  class::             Ptr to class information (see w_matchInstance).
 *
 * Output:::
 *  instance::          Ptr to instance information (see w_matchInstance).
 *
 * Returns:::
 *
 * Notes:::
 *
 * PrivateNotes:::
 *  The types of marks we are looking for are not conducive to using 2-Way
 *  Hausdorff matching, so only 1-Way is used.  Thus, the percetBtoA instance
 *  field will not be filled.  (However we will use another verification
 *  technique heuristic to help ensure it is a good match).
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  page description, template matching, Hausdorff, multi-resolution
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_hausdorffMatchBest (
    IPWState      *pIPWState,
    PIXR          *pixr,
    void          *pMultResInfo,
    RECTANGL      *clipRect,
    Float64        minPerc,
    classType     *class,
    instanceType  *instance)
{
    PROLOGUE;
    Int32      result = 0, w, h;
    UInt8      notVerified;
    PIXR      *BS = NULL, *BD = NULL, *BDTmp = NULL;
    matchData  matchi;
    STREL     *pStrel = NULL;
    Float64    percBtoA = 0.;
    
    ERR_PASS(w_rectOf (pIPWState, pixr, clipRect, &BS));
    ERR_PASS_CLEANUP(w_clonePixrNoInit (pIPWState, BS, &BD));

    ERR_PASS_CLEANUP(w_getStrel(pIPWState, HAUSDORFF_STREL, &pStrel));
    ERR_PASS_CLEANUP(w_dilate (pIPWState, BS, pStrel, &BD));

    matchi.percAtoB = -1.;

    ERR_PASS_CLEANUP(w_pixrGetWidth (pIPWState, BD, &w));
    ERR_PASS_CLEANUP(w_pixrGetHeight (pIPWState, BD, &h));

    if (pMultResInfo)
    {
	ERR_PASS_CLEANUP(w_clonePixr (pIPWState, BD, &BDTmp));

	/* Search the full subrect of BD for matches */
	ERR_PASS_CLEANUP(w_hausdorffMultResMatches (pIPWState,
						    class->model, BD, BDTmp,
						    NULL, NULL,
						    0,
						    0,
						    w, h,
						    0,
						    pMultResInfo,
						    (void *)&matchi,
						    w_best1Callback,
						    &minPerc, &percBtoA,
						    &result));
	ERR_GEN_CLEANUP(result, cErrInternal);

	ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &BDTmp));
    }
    else
    {
	/* w_write (pIPWState, "BD", BD, TIFF_G4);*/
	ERR_PASS_CLEANUP(w_hausdorffMatches (pIPWState, class->model, BD,
					     NULL, NULL,
					     0,
					     0,
					     w, h,
					     1, 1,
					     SPARSE_FILTER,
					     (void *)&matchi,
					     w_best1Callback,
					     &minPerc, &percBtoA,
					     &result));
	ERR_GEN_CLEANUP (result, cErrInternal);
    }
    

    if (matchi.percAtoB >= 0.)
    {
	ERR_PASS_CLEANUP(wp_notVerifiable(pIPWState, class, instance,
					  matchi.tx, matchi.ty,
					  clipRect->x, clipRect->y,
					  matchi.percAtoB, BS, BD, 
					  &notVerified));
	
	if (notVerified)
	    matchi.percAtoB = -1.;
    }

    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &BS));
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &BD));

    if (matchi.percAtoB >= 0.)
    {
	instance->status = HAUSDORFF_FOUND;
	/* The final match location will be indicated by the translation 
	 * difference from the instance's guess location */
	instance->transx = (Int16)(matchi.tx + MAX(0, clipRect->x) -
				   instance->guessx);
	instance->transy = (Int16)(matchi.ty + MAX(0, clipRect->y) -
				   instance->guessy);
	instance->percentAtoB = matchi.percAtoB;
	/* The match exhibits no scaling with this current class bitmap */
	instance->scalex = 1.;
	instance->scaley = 1.;
    }

    
    if (PRINTABLE(pIPWState)) {
	if (matchi.percAtoB >= 0.)
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "JQ: Matched (%3ldx%3ld) at %5ld, %5ld (Percent:%8.7f)\n",
		      class->w, class->h,
		      instance->guessx + instance->transx,
		      instance->guessy + instance->transy,
		      instance->percentAtoB);
	else
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "JQ: No best match was found in rect (%ld,%ld,%ld,%ld)\n",
		      clipRect->x, clipRect->y, clipRect->w, clipRect->h);
    }
    
    
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &BS);
    w_destroyPixr (pIPWState, &BD);
    w_destroyPixr (pIPWState, &BDTmp);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Last effort to find a page description "instance" in an image bitmap
 *
 * LongDescription:::
 *  We say "ideal" here, because this technique is slower and more prone
 *  to finding a match (than say wp_hausdorffMatchBest()) but we also want
 *  to keep some "ideal" match in mind and not jump at the first 
 *  okay-but-not-so-great match that we see.
 *
 * PrivateDescription:::
 *  The theory is this.  We only have a guess at where this instance may
 *  be.  Maybe the guess is way off, yet the bitmap is intact.  Or perhaps
 *  the guess is right on, but the bitmap in the image is really badly
 *  degraded.  So what to do?  Here we think to look pretty far from the
 *  guess, but to also pay attention to the Hausdorff A to B percents
 *  as we find might-be-matches, by using the routine w_bestIdeal1Callback()
 *  as our Hausdorff callback routine.
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary image to search in.
 *  pMultResInfo::      Ptr to the multi-resolution set-up info, created
 *                      by w_makeMultResInfo().
 *  clipRect::          Ptr to subrect of pixr where we will search
 *                      for matches.
 *  minPerc::           Any match found must have a Hausdorff A to B percent
 *                      of at least this much, [0..1].  This will define how
 *                      desperate we are to find a match.
 *  txIdeal::           Where we would ideally expect to find the instance.
 *  tyIdeal::           As in, if the form were perfect, it would be here.
 *  percIdeal::         A higher percent (than "minPerc") where any match at
 *                      least this will be considered "ideal" and will
 *                      follow the criteria in w_bestIdeal1Callback().
 *  class::             Ptr to class information (see w_matchInstance).
 *
 * Output:::
 *  instance::          Ptr to instance information (see w_matchInstance).
 *
 * Returns:::
 *
 * Notes:::
 *
 * PrivateNotes:::
 *  The types of marks we are looking for are not conducive to using 2-Way
 *  Hausdorff matching, so only 1-Way is used.  Thus, the percetBtoA instance
 *  field will not be filled.  (However we will use another verification
 *  technique heuristic to help ensure it is a good match).
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  page description, template matching, Hausdorff, multi-resolution
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_hausdorffMatchIdeal (
    IPWState      *pIPWState,
    PIXR          *pixr,
    void          *pMultResInfo,
    RECTANGL      *clipRect,
    Float64        minPerc,
    Int32          txIdeal,
    Int32          tyIdeal,
    Float64        percIdeal,
    classType     *class,
    instanceType  *instance)
{
    PROLOGUE;
    Int32      result = 0, w, h;
    UInt8      notVerified;
    PIXR      *BS = NULL, *BD = NULL, *BDTmp = NULL;
    STREL     *pStrel = NULL;
    idealData  matchi;
    Float64    percBtoA = 0.;
    

    ERR_PASS(w_rectOf (pIPWState, pixr, clipRect, &BS));
    ERR_PASS_CLEANUP(w_clonePixrNoInit (pIPWState, BS, &BD));
    
    ERR_PASS_CLEANUP(w_getStrel(pIPWState, HAUSDORFF_STREL, &pStrel));
    ERR_PASS_CLEANUP(w_dilate (pIPWState, BS, pStrel, &BD));

    ERR_PASS_CLEANUP(w_clonePixr (pIPWState, BD, &BDTmp));

    /* Put in the coordinate frame of the full rectangle we are
     * searching, BD */
    matchi.txIdeal = txIdeal - clipRect->x;
    matchi.tyIdeal = tyIdeal - clipRect->y;
    matchi.percAtoBIdeal = percIdeal;
    matchi.percBtoAIdeal = matchi.percAtoB = matchi.percBtoA = -1.;

    ERR_PASS_CLEANUP(w_pixrGetWidth (pIPWState, BD, &w));
    ERR_PASS_CLEANUP(w_pixrGetHeight (pIPWState, BD, &h));

    /* Search the full subrect of BD for matches */
    ERR_PASS_CLEANUP(w_hausdorffMultResMatches (pIPWState, class->model,
						BD, BDTmp,
						NULL, NULL,
						0, 0, w, h,
						0,
						pMultResInfo,
						(void *)&matchi,
						w_bestIdeal1Callback,
						&minPerc, &percBtoA,
						&result));
    ERR_GEN_CLEANUP(result, cErrInternal);
    

    if (matchi.percAtoB >= 0.)
    {
	ERR_PASS_CLEANUP(wp_notVerifiable(pIPWState, 
					  class, instance, 
					  matchi.tx, matchi.ty,
					  clipRect->x, clipRect->y,
					  matchi.percAtoB, BS, BD,
					  &notVerified));
	
	if (notVerified)
	    matchi.percAtoB = -1.;
    }

    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &BS));
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &BD));
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &BDTmp));


    if (matchi.percAtoB >= 0.)
    {
	instance->status = HAUSDORFF_FOUND;
	/* The final match location will be indicated by the translation 
	 * difference from the instance's guess location */
	instance->transx = (Int16)(matchi.tx + MAX(0, clipRect->x) -
				   instance->guessx);
	instance->transy = (Int16)(matchi.ty + MAX(0, clipRect->y) -
				   instance->guessy);
	instance->percentAtoB = matchi.percAtoB;
	/* The match exhibits no scaling with this current class bitmap */
	instance->scalex = 1.;
	instance->scaley = 1.;
    }

    if (PRINTABLE(pIPWState)) {
	if (matchi.percAtoB >= 0.)
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "JQ: Matched (%3ldx%3ld) at %5ld, %5ld (Percent:%8.7f)\n",
		      class->w, class->h,
		      instance->guessx + instance->transx,
		      instance->guessy + instance->transy,
		      instance->percentAtoB);
	else
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "JQ: No ideal match was found in rect (%ld,%ld,%ld,%ld)\n",
		      clipRect->x, clipRect->y, clipRect->w, clipRect->h);
    }
    
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &BS);
    w_destroyPixr (pIPWState, &BD);
    w_destroyPixr (pIPWState, &BDTmp);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find a page description "instance" in an image bitmap
 *
 * PrivateDescription:::
 *  Given our best guess where we think the instance is located, create a
 *  rect that perfectly fits around its bitmap.  Then expand this rect
 *  by 1/2 its width and height, capping at a certain "reasonable" expansion
 *  size.  This is our first rectangle to search the image for matches.<BR><BR>
 *
 *  The second rectangle is that which is described already by the
 *  instance fields: rect{x,y,w,h}, the maximal searching subrect (and we
 *  typically expect it to be larger -- although if it isn't we won't do
 *  the second check).<BR><BR>
 *
 *  Given these two rectangles that contain areas we believe the instance
 *  can be found, we search the first rectangle using wp_hausdorffMatchBest().
 *  Observing data about the bitmap, we try to come up with some good guess
 *  for what the lowest Hausdorff A to B percent should be here, which outght
 *  to typically be higher than the "minPerc" parameter.
 *  If it's not found, we search the second rectangle using
 *  wp_hausdorffMatchIdeal(), a slower, more "desperate" method.
 *  The key to this is the first check is speedy and we use the fastest
 *  routine.  In instances where one's guess is way off, we search the entire
 *  larger rect (slower -- more translations to check).  We assume/hope that
 *  this larger rect search won't be necessary most of the time!<BR><BR>
 * 
 *  If the second rectangle doesn't contain the first, we just do the
 *  second check.
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary image to search in.
 *  xres::              X resolution of image
 *  yres::              Y resolution of image
 *  pMultResInfo::      Ptr to the multi-resolution set-up info, created
 *                      by w_makeMultResInfo().
 *  minPerc::           Any match found using the last and slowest check
 *                      must have a Hausdorff A to B percent of at
 *                      least this much, [0..1].  This will define how
 *                      desperate we are to find a match.
 *  class::             Ptr to class information (see w_matchInstance).
 *
 * Output:::
 *  instance::          Ptr to instance information (see w_matchInstance).
 *
 * Returns:::
 *
 * Notes:::
 *  The instance's rect{x,y,w,h} fields are expected to be cropped to the pixr!
 *
 * PrivateNotes:::
 *  The types of marks we are looking for are not conducive to using 2-Way
 *  Hausdorff matching, so only 1-Way is used.  Thus, the percetBtoA instance
 *  field will not be filled.  (However we will use another verification
 *  technique heuristic to help ensure it is a good match).
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  page description, template matching, Hausdorff, multi-resolution
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_hausdorffMatch (
    IPWState        *pIPWState,
    PIXR            *pixr,
    Int16            xres,
    Int16            yres,
    void            *pMultResInfo,
    Float64          minPerc,
    classType       *class,
    instanceType    *instance)
{
    PROLOGUE;
    Float64    percIdeal;
    Int32      txIdeal, tyIdeal, w, h;
    RECTANGL   rect, maxrect;

    if (instance->status == HAUSDORFF_FOUND ||
	instance->status == HAUSDORFF_MISSED)
	/* Already been here */
	NORMAL_RETURN;
    
    if (instance->rectw <= 0 || instance->recth <= 0)
    {
	if (PRINTABLE(pIPWState)) {
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "JQ: No match found (entire search out of bounds)!!\n");
	}
	NORMAL_RETURN;
    }

    /* We'll only be doing 1-Way matching here */
    instance->percentAtoB = -1.;
    instance->percentBtoA = -1.;

    /* Initialize to not found */
    instance->status = HAUSDORFF_MISSED;

    /* This is our ideal world.  We probe at this (x,y) location and get
     * this ideal percent!
     * Calculate a suitable minimum Hausdorff percent given the size of this
     * class, for finding a match */
    if (((Float64) class->w / xres) > ((Float64)class->h / yres))
    {
	percIdeal = (Float64)class->w / xres;
	percIdeal = MIN_HAUS_PERCENT((Float64)class->h / yres, percIdeal);
    }
    else
    {
	percIdeal = (Float64)class->h / yres;
	percIdeal = MIN_HAUS_PERCENT((Float64)class->w / xres, percIdeal);
    }

    /* As of now, the user's best guess is here, the ideal location */
    txIdeal = instance->guessx + instance->guessTransx;
    tyIdeal = instance->guessy + instance->guessTransy;

    /* Expand the rect where we will search the image for matches.
     * Allow class to translate by 1/2 its width and height,
     * capping at a certain "reasonable" expansion size
     */
    rect.x = instance->guessx + instance->guessTransx -
	MIN(class->w/2, (Int32)(MAX_TRANS_INCR_INCHES*xres));
    rect.w = class->w + 2 * MIN(class->w/2,
				(Int32)(MAX_TRANS_INCR_INCHES*xres));
    rect.y = instance->guessy + instance->guessTransy -
	MIN(class->h/2, (Int32)(MAX_TRANS_INCR_INCHES*yres));
    rect.h = class->h + 2 * MIN(class->h/2,
				(Int32)(MAX_TRANS_INCR_INCHES*yres));

    maxrect.x = instance->rectx + instance->guessTransx;
    maxrect.y = instance->recty + instance->guessTransy;
    maxrect.w = instance->rectw;
    maxrect.h = instance->recth;

    ERR_PASS(w_pixrGetWidth (pIPWState, pixr, &w));
    ERR_PASS(w_pixrGetHeight (pIPWState, pixr, &h));
    ERR_PASS(w_cropRect( pIPWState, 0, 0, w, h, &maxrect));
    
    ERR_PASS(w_cropRect( pIPWState, 
			 maxrect.x, maxrect.y, maxrect.w, maxrect.h, &rect));

    /* If the two rects aren't equal AND "rect" has positive width/height...*/
    if (!(rect.x == maxrect.x  &&
	  rect.y == maxrect.y  &&
	  rect.w == maxrect.w  &&
	  rect.h == maxrect.h) &&
	rect.w > 0  &&  rect.h > 0)
    {
	ERR_PASS(wp_hausdorffMatchBest(pIPWState,
				       pixr,
				       /*NULL*/ /*WART*/ pMultResInfo,
				       &rect,
				       percIdeal,
				       class,
				       instance));
    }

    if (instance->status != HAUSDORFF_FOUND)
    {
	ERR_PASS(wp_hausdorffMatchIdeal(pIPWState,
					pixr,
					pMultResInfo,
					&maxrect,
					minPerc,
					txIdeal,
					tyIdeal,
					percIdeal,
					class,
					instance));
    }
	

    NORMAL_RETURN;
}


/* To allow up to 1/16" extra movement below */
#define XTRA_BR(dpi)  (dpi / 16)
/* To allow small pixel movements below */
#define XTRA_TRANS 2

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Split a rectangular page description instance up and try matching it
 *
 * PrivateDescription:::
 *  We tried to match a rectangular based bitmap but could not find one
 *  previously.  Thus, try breaking up this rigid model into two corner
 *  pieces and try to match in the original rectangle we tried to match at.
 *  If they both match, then construct a new model
 *  based on where they each matched, and try the new whole match.<BR>
 *  The basic philosophy is that:
 *  when rectangles are drawn, in instances where the rectangle
 *  is large, if the render spec is even a little different from the actual
 *  rectangle in the image, it may need to be scaled in X and/or Y.
 *  (Things like faxes can cause weird scaling, etc).
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary image to search in.
 *  xres::              X resolution of image
 *  yres::              Y resolution of image
 *  pMultResInfo::      Ptr to the multi-resolution set-up info, created
 *                      by w_makeMultResInfo().
 *  class::             Ptr to class information (see w_matchInstance).
 *
 * Output:::
 *  instance::          Ptr to instance information (see w_matchInstance).
 *
 * Returns:::
 *
 * Notes:::
 *  The instance's rect{x,y,w,h} fields are expected to be cropped to the pixr!
 *
 * PrivateNotes:::
 *  The types of marks we are looking for are not conducive to using 2-Way
 *  Hausdorff matching, so only 1-Way is used.  Thus, the percetBtoA instance
 *  field will not be filled.  (However we will use another verification
 *  technique heuristic to help ensure it is a good match).
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  page description, template matching, Hausdorff, multi-resolution
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_hausdorffSplitMatch (
    IPWState        *pIPWState,
    PIXR            *pixr,
    Int16            xres,
    Int16            yres,
    void            *pMultResInfo,
    classType       *class,
    instanceType    *instance)
{
    PROLOGUE;
    classType     splitClass;
    instanceType  splitInstance;
    UInt8         nBitmaps = 0;
    Int32         w, h;
    Int16         spRes0tx, spRes0ty, spRes1tx, spRes1ty;
    RECTANGL      splitMaxrect;
    Bool          rendered;


    /* If need be later, since I no longer cache the split models,
     * one could pass in an array of instances that need to be matched,
     * and each model piece rendered could be tried for each instance
     * in the array.
     */

    w = class->w;
    h = class->h;

    /* The split will look like this:
     * ____________________
     * | piece0
     * |                  |
     * |                  |
     * |                  |
     * |                  |
     *             piece1 |
     * --------------------
     */


    if (PRINTABLE(pIPWState)) {
	i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		  "JQ: Splitting rect\n");
    }


    /* Now try matching each of the split models until one of them
     * doesn't match, or all match */
    




    /* TRY TOP LEFT, piece 0 */
    splitClass.kind = CHECKBOX_TOPLEFT;
    splitClass.w = w;
    splitClass.h = h;
    splitClass.wThick = class->wThick;
    splitClass.hThick = class->hThick;
    splitClass.model = NULL;
    splitClass.srcPr = NULL;
    
    ERR_PASS(w_createPixr (pIPWState, w, h, 1, 0, 0,
			   32, &splitClass.srcPr));
    
    ERR_PASS_CLEANUP(w_pickModelDataType(pIPWState, splitClass.srcPr,
					 splitClass.wThick, splitClass.hThick,
					 &nBitmaps));

    ERR_PASS_CLEANUP( w_renderClass( pIPWState, splitClass.kind, w, h,
				     splitClass.wThick, splitClass.hThick, 0,
				     &rendered, &(splitClass.srcPr)));

    ERR_GEN_CLEANUP(!rendered, cErrInternal);

    ERR_PASS_CLEANUP(w_makeModelSet (pIPWState, splitClass.srcPr, 
				     h+1/* blockHeight */, nBitmaps,
				     &(splitClass.model)));


    /* Search in the same area as the full model */
    splitInstance.guessx = instance->guessx;
    splitInstance.guessy = instance->guessy;
    splitInstance.guessTransx = instance->guessTransx;
    splitInstance.guessTransy = instance->guessTransy;
    splitInstance.guessScalex = instance->guessScalex;
    splitInstance.guessScaley = instance->guessScaley;
    splitInstance.rectx = instance->rectx;
    splitInstance.recty = instance->recty;
    splitInstance.rectw = instance->rectw;
    splitInstance.recth = instance->recth;
    splitInstance.classNo = 0;
    splitInstance.rowColRims = instance->rowColRims;
    splitInstance.status = USER_GUESS;
    
    ERR_PASS_CLEANUP(wp_hausdorffMatch(pIPWState,
				       pixr,
				       xres,
				       yres,
				       pMultResInfo,
				       ANY_MATCH,
				       &splitClass,
				       &splitInstance));
    

    ERR_PASS_CLEANUP(w_destroyModelSet (pIPWState, &(splitClass.model)));

    if (splitInstance.status != HAUSDORFF_FOUND)
    {
	/* Means couldn't match the sub piece */
	if (PRINTABLE(pIPWState)) {
	    i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
		      "JQ: Split--couldn't match top left piece\n");
	}
    }
    else 
    {
	
	/* 0th split model's resulting translation from the original guess */
	spRes0tx = splitInstance.transx;
	spRes0ty = splitInstance.transy;


    
    
    
	/* TRY BOTTOM RIGHT, piece 1 */
	splitClass.kind = CHECKBOX_BOTRITE;
	splitClass.model = NULL;
    
	/* clear pixr */
	ERR_PASS_CLEANUP(w_rasterop (pIPWState, 0, 0, NULL,
				     PIX_CLR, 0, 0, w, h,
				     splitClass.srcPr));
	ERR_PASS_CLEANUP(w_pickModelDataType(pIPWState, splitClass.srcPr,
					     splitClass.wThick, 
					     splitClass.hThick,
					     &nBitmaps));

	ERR_PASS_CLEANUP(w_renderClass(pIPWState, splitClass.kind, w, h,
				       splitClass.wThick, splitClass.hThick, 0,
				       &rendered, &(splitClass.srcPr)));
    
	ERR_GEN_CLEANUP (!rendered, cErrInternal);
    
	ERR_PASS_CLEANUP(w_makeModelSet (pIPWState, splitClass.srcPr,
					 h+1/* blockHeight */, nBitmaps,
					 &(splitClass.model)));


	/* Make the search VERY constrained to just allow up to 1/16" movement
	 * either way from where we would normally expect it to be.
	 * Addition:  At the same time, we need to sanity check for small
	 * checkmarks.  So allow the movement to be AT MOST 1/4 width and 1/4
	 * height in x and y respectively.
	 */
	splitInstance.guessTransx = spRes0tx;
	splitInstance.guessTransy = spRes0ty;
	splitInstance.rectx = splitInstance.guessx - MIN(XTRA_BR(xres), w/4);
	splitInstance.rectw = w + 2*MIN(XTRA_BR(xres), w/4);
	splitInstance.recty = splitInstance.guessy - MIN(XTRA_BR(yres), h/4);
	splitInstance.recth = h + 2*MIN(XTRA_BR(yres), h/4);
	splitInstance.status = USER_GUESS;

/* WART This seems kind of silly/stupid interface to hMB()! */
	splitMaxrect.x = splitInstance.rectx + splitInstance.guessTransx;
	splitMaxrect.y = splitInstance.recty + splitInstance.guessTransy;
	splitMaxrect.w = splitInstance.rectw;
	splitMaxrect.h = splitInstance.recth;
    

	/* WART this could make *some* difference, but shouldn't... */
#if 0
	ERR_PASS_CLEANUP(w_cropRect( pIPWState,
				     maxrect->x, maxrect->y, 
				     maxrect->w, maxrect->h,
				     &splitMaxrect));
#endif

	/* We'll look for the best in "splitMaxrect" */
	ERR_PASS_CLEANUP(wp_hausdorffMatchBest(pIPWState,
					       pixr,
					       /*NULL*/ /*WART*/pMultResInfo,
					       &splitMaxrect,
					       ANY_MATCH,
					       &splitClass,
					       &splitInstance));
    

	ERR_PASS_CLEANUP(w_destroyModelSet (pIPWState, &(splitClass.model)));

	if (splitInstance.status != HAUSDORFF_FOUND)
	{
	    /* Means couldn't match the sub piece */
	    if (PRINTABLE(pIPWState)) {
		i_printf (pIPWState, pIPWState->errorProc, pIPWState->errorInfo,
			  "JQ: Split--couldn't match bottom right piece\n");
	    }
	}
	else
	{
	

	    /* 1st split model's resulting translation from the
	     * original guess
	     */
	    spRes1tx = splitInstance.transx;
	    spRes1ty = splitInstance.transy;




    
	    /* NOW TRY A GLUE TOGETHER OF THE ABOVE TWO */

	    /* Now assemble the two matches together and see if we can get a
	     * general match with some newly rendered approximation (a "whole")
	     * to them! */

	    /* This is the resized new dimensions of the whole */
	    splitClass.w = w + spRes1tx - spRes0tx;
	    splitClass.h = h + spRes1ty - spRes0ty;
	    splitClass.kind = RECTANGLE;

	    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &(splitClass.srcPr)));
	    splitClass.srcPr = NULL;
	    ERR_PASS_CLEANUP(w_renderClass(pIPWState, RECTANGLE, 
					   splitClass.w, splitClass.h,
					   splitClass.wThick, 
					   splitClass.hThick,
					   0, &rendered, &(splitClass.srcPr)));

	    ERR_GEN_CLEANUP (!rendered, cErrInternal);

    
	    ERR_PASS_CLEANUP(w_pickModelDataType(pIPWState,
						 splitClass.srcPr,
						 splitClass.wThick, 
						 splitClass.hThick,
						 &nBitmaps));
    
	    splitClass.model = NULL;
	    ERR_PASS_CLEANUP(w_makeModelSet (pIPWState, splitClass.srcPr, 
					     splitClass.h+1/* blockHeight */,
					     nBitmaps, &(splitClass.model)));


	    /* Make the search VERY constrained to just allow a SMALL movement
	     * either way from where we would normally expect it to be */

	    /* NOTE Already set:
	     * splitInstance.guessTransx = spRes0tx;
	     * splitInstance.guessTransy = spRes0ty;
	     */

	    splitInstance.rectx = splitInstance.guessx - XTRA_TRANS;
	    splitInstance.rectw = splitClass.w + 2*XTRA_TRANS;
	    splitInstance.recty = splitInstance.guessy - XTRA_TRANS;
	    splitInstance.recth = splitClass.h + 2*XTRA_TRANS;
	    splitInstance.status = USER_GUESS;

	    /* WART This seems kind of silly/stupid interface to hMB()! */
	    splitMaxrect.x = splitInstance.rectx + splitInstance.guessTransx;
	    splitMaxrect.y = splitInstance.recty + splitInstance.guessTransy;
	    splitMaxrect.w = splitInstance.rectw;
	    splitMaxrect.h = splitInstance.recth;

	    /* WART this could make *some* difference, but shouldn't... */
#if 0
	    ERR_PASS_CLEANUP(w_cropRect(pIPWState, 
					maxrect->x, maxrect->y,
					maxrect->w, maxrect->h,
					&splitMaxrect));
#endif

	    /* Now get the best possible match we can in the small subarea */
	    /* We'll look for the best in here "splitMaxrect" */
	    ERR_PASS_CLEANUP(wp_hausdorffMatchBest(pIPWState,
						   pixr,
						   /*NULL*/ /*WART*/ pMultResInfo,
						   &splitMaxrect,
						   ANY_MATCH,
						   &splitClass,
						   &splitInstance));
   
	    ERR_PASS_CLEANUP(w_destroyModelSet (pIPWState, 
						&(splitClass.model)));

	    if (splitInstance.status == HAUSDORFF_FOUND)
	    {
		instance->status = HAUSDORFF_FOUND;
		instance->transx = splitInstance.transx;
		instance->transy = splitInstance.transy;
		instance->percentAtoB = splitInstance.percentAtoB;
	
		/* This is the scaling information that we can 
		 * decipher from this
		 * match ... how far the class width and/or 
		 * height stretched at this
		 * particular instance from the ideal class dimensions.
		 */
		instance->scalex = (Float64)splitClass.w / w;
		instance->scaley = (Float64)splitClass.h / h;
	    }
    

	    if (PRINTABLE(pIPWState)) {
		if (splitInstance.status != HAUSDORFF_FOUND)
		    /* No match was found */
		    i_printf (pIPWState, pIPWState->errorProc,
			      pIPWState->errorInfo,
			      "JQ: Split--couldn't match whole.\n");
		else
		    i_printf (pIPWState, pIPWState->errorProc,
			      pIPWState->errorInfo,
			      "JQ: Split--successfully matched whole.  Width/height scale factors: %6.4f/%6.4f\n",
			     instance->scalex, instance->scaley);
	    }
	}
    }
    
    
    
    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &(splitClass.srcPr)));
    NORMAL_RETURN;


ERR_CLEANUP:
    w_destroyPixr (pIPWState, &(splitClass.srcPr));
    w_destroyModelSet (pIPWState, &(splitClass.model));
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find a page description "instance" in an image bitmap
 *
 * LongDescription:::
 *  We are looking for a special mark on a page, given a partial page
 *  description (some specification of an image that describes its contents
 *  in a recoverable way that is much smaller than the image itself).
 *  This routine deals with finding special marks that are specifications
 *  that can be converted to an actual bitmap using w_renderClass() and then
 *  w_makeModelSet() to make a Hausdorff model from this bitmap.
 *  The best way to describe what this routine does is based on the
 *  parameters.<BR>
 *
 *  <PRE>
 *  For the ptr to the instance's struct:
 *   on input:
 *    guess{x,y}:      location where client believes instance should be
 *                     in ideal page
 *    guessTrans{x,y}: passed in best guess at translation in x and y,
 *                     relative to guess{x,y}.
 *    guessScale{x,y}: optional passed in best guess at scaling in x,y.
 *                     [presently ignored]
 *    name, extra:     for client's use; ignored.
 *    rect{x,y,w,h}:   maximal subrect around above guess location that we
 *                     should search for a match in.
 *                     (rectx <= guessx, recty <= guessy
 *                      rectw >= class' w, recth >= class' h)
 *    rowColRims:      Tells us which of the four rectangles immediately
 *                     surrounding the instance bitmap on the perfect page
 *                     were very sparse.
 *   on output:
 *    status:          HAUSDORFF_FOUND or HAUSDORFF_MISSED based on whether
 *                     or not we found it.
 *    when found:
 *     trans{x,y}:     Translation determined relative to guess{x,y}.
 *     scale{x,y}:     Scaling (if) determined.
 *     percent{AtoB,BtoA}:  Hausdorff percents of matched location.
 *                          (like degree of confidence in match -- see
 *                           more general comments in "haussrch.c".)
 *  For the ptr to the class' struct:
 *   All fileds must be filled and accurate.  This is the exact representation
 *   of what we are looking for.  The "instance" above only tells us where
 *   to look, and where it was found.  This struct gives us a bitmap and
 *   Hausdorff model to use to find the instance.
 *  </PRE><BR>
 *
 *  Thus, we use the best guess up to now where we believe
 *  the instance's bitmap can be found.
 *  We use Hausdorff matching techniques to find this more
 *  precise match location.
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary image to search in.
 *  xres::              X resolution of image
 *  yres::              Y resolution of image
 *  pMultResInfo::      Ptr to the multi-resolution set-up info, created
 *                      by w_makeMultResInfo().
 *  class::             Ptr to class information (see above description).
 *
 * Output:::
 *  instance::          Ptr to instance information (see above description).
 *
 * Returns:::
 *
 * Notes:::
 *  The instance's rect{x,y,w,h} fields are expected to be cropped to the pixr!
 *
 * PrivateNotes:::
 *  The types of marks we are looking for are not conducive to using 2-Way
 *  Hausdorff matching, so only 1-Way is used.  Thus, the percetBtoA instance
 *  field will not be filled.  (However we will use another verification
 *  technique heuristic to help ensure it is a good match).
 *
 * FunctionCategory:::
 *  template matching
 *
 * Keywords:::
 *  page description, template matching, Hausdorff, multi-resolution
 *
 * Level:::
 *  directive
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_matchInstance (
    IPWState          *pIPWState,
    PIXR              *pixr,
    Int16              xres,
    Int16              yres,
    void              *pMultResInfo,
    classType         *class,
    instanceType      *instance)
{
    PROLOGUE;

    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN(class == NULL  ||  instance == NULL  ||  pMultResInfo == NULL,
	    cErrNullPtr);
    /* The maximal subrect must surround the instance at least!! */
    ERR_GEN(instance->rectx > instance->guessx  ||  
	    instance->recty > instance->guessy, cErrInvalidParm);
    ERR_GEN(instance->rectw < class->w  ||
	    instance->recth < class->h, cErrInvalidParm);

    
    /* Means we haven't rendered a bitmap for the class that this instance
     * belongs to.
     * Thus, it can't be Hausdorff matched!
     */
    ERR_GEN(class->model == NULL, cErrInvalidParm);


    ERR_PASS( wp_hausdorffMatch( pIPWState,
				 pixr,
				 xres,
				 yres,
				 pMultResInfo,
				 ANY_MATCH,
				 class,
				 instance));
	    
    if (instance->status != HAUSDORFF_FOUND  &&
	(class->kind == RECTANGLE ||
	 class->kind == CLIPREGION ||
	 class->kind == BOXSIGNATURE))
    {
	/* Means we tried to match a rectangular model but couldn't
	 * find a match above ... so try splitting the model up.
	 */
	ERR_PASS( wp_hausdorffSplitMatch( pIPWState,
					  pixr,
					  xres,
					  yres,
					  pMultResInfo,
					  class,
					  instance));
    }

    NORMAL_RETURN;
}


/*

 $Log: match.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.6  1998/01/15 01:49:50  ddavies
 match.c -> /product/ipcore/ipwrap/src/RCS/match.c,v

 CommonCheck: Fix WATCOM warnings about unused procedure args.

 Revision 1.5  1998/01/14 00:13:53  ddavies
 match.c -> /product/ipcore/ipwrap/src/RCS/match.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.4  1997/06/26 20:04:19  ddavies
 match.c -> /product/ipcore/ipwrap/src/RCS/match.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.3  1997/04/15 22:13:57  ddavies
 match.c -> /product/ipcore/ipwrap/src/RCS/match.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.2  1996/02/26 20:05:16  ddavies
 ddavies on Mon Feb 26 12:05:04 PST 1996

 match.c -> /product/ipcore/ipwrap/src/RCS/match.c,v

 Add casts to make Visual C/C++ happy.

 Revision 1.1  1996/02/25 01:12:07  ejaquith
 Initial revision


 */
